# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for ESX Host based events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ClientEvents::ESXHostEvents;

use ParamsMgr;
use Tharn;
use San::SanUtil;
use strict;
use warnings;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ClientEvents",
                         "EventMan::EventModules");
    # registering the svmotion event
    $self->EventUtil::register_event(-event_name => "esx_svmotion",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX Storage vMotion Event",
                                     -exec => sub { &esx_svmotion(@_) },
                                     -validate => sub { &_esx_svmotion_validate(@_) },
                                     -arg_defaults => { esx_svmotion_lun_path_regex => undef,
                                                        esx_svmotion_destination_datastores => undef,
                                                        esx_svmotion_restore_sleep_time => 60,
                                                        esx_svmotion_interval_sleep_time => 60,
                                                        esx_svmotion_timeout => 5400 });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    # registering the vmotion event
    $self->EventUtil::register_event(-event_name => "esx_vmotion",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX vMotion Event",
                                     -exec => sub { &esx_vmotion(@_) },
                                     -validate => sub { &_esx_vmotion_validate(@_) },
                                     -arg_defaults => { esx_vmotion_destination_esx_server => "",
                                                        esx_vmotion_priority => undef,
                                                        esx_vmotion_state => undef,
                                                        esx_vmotion_retry => 10,
                                                        esx_vmotion_restore_sleep_time => 60 });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    # registering the snapshot event
    $self->EventUtil::register_event(-event_name => "esx_snapshot",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX Snapshot Event",
                                     -exec => sub { &esx_snapshot(@_) },
                                     -validate => sub { &_esx_snapshot_validate(@_) },
                                     -arg_defaults => {esx_snapshot_name => "",
													   esx_snapshot_memory => 0,
													   esx_snapshot_quiesce => 0});
	
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    # registering the esx_vxdmp_rotate_io_policies event
    $self->EventUtil::register_event(-event_name => "esx_vxdmp_rotate_io_policies",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX Event that rotates VxDMP IO Policies ",
                                     -exec => sub { &esx_vxdmp_rotate_io_policies(@_) },
                                     -validate => sub { &_esx_vxdmp_rotate_io_policies_validate(@_) },
                                     -arg_defaults => { esx_vxdmp_io_policies => [] });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    
    # registering the Lun Target Reset event
    $self->EventUtil::register_event(-event_name => "esx_lun_target_reset",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX Lun Target Reset Event",
                                     -exec => sub { &esx_lun_target_reset(@_) },
                                     -validate => sub { &_esx_lun_target_reset_validate(@_) },
                                     -arg_defaults => { esx_lun_target_reset_count => 10,
                             							esx_lun_target_reset_type => "lunreset" });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # registering the ESX add/remove vmdk event
    $self->EventUtil::register_event(-event_name => "esx_add_remove_vmdk",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX Add/Remove VMDK Event",
                                     -exec => sub { &esx_add_remove_vmdk(@_)});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # registering the ESX VAAI/unmap script event
    $self->EventUtil::register_event(-event_name => "esx_vaai_and_unmap",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX VAAI/UNMAP Event",
                                     -exec => sub { &esx_vaai_and_unmap(@_)});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # registering the ESX VAAI/unmap/xcopy script event
    $self->EventUtil::register_event(-event_name => "esx_vaai_unmap_xcopy",
                                     -event_modules => \@event_modules,
                                     -required_resources => ["sanclients"],
                                     -description => "ESX VAAI/UNMAP/XCOPY Event",
                                     -exec => sub { &esx_vaai_unmap_xcopy(@_)});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults => { virtual_machines => [],
                                                                        esx_vcenter_server => "",
                                                                        esx_vcenter_datacenter => undef });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "virtual_machines",
                                           -package => __PACKAGE__,
                                           -default => [],
                                           -format => "ARRAY Reference",
                                           -example => "-virtual_machines => ['vm1', 'vm2']",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies the virtual machines to " .
                                                           "use for a given event");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vcenter_server",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String",
                                           -example => "-esx_vcenter_server => 'myvcenter'",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies the hostname of the vcenter server, " .
                                                           "must resolve to a tharnhost file name.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vcenter_datacenter",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String",
                                           -example => "-esx_vcenter_datacenter => 'ha-datacenter'",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies the datacenter of the vcenter server.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_svmotion_lun_path_regex",
                                           -package => __PACKAGE__,
                                           -default => "undef",
                                           -format => "Regexp",
                                           -example => "-esx_svmotion_lun_path_regex => qr/mysvmotion_lun/i",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Controls what luns to perform svmotions to and from. " .
                                               "This parameter is mutually exclusive with 'esx_svmotion_destination_datastores'.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_svmotion_destination_datastores",
                                           -package => __PACKAGE__,
                                           -default => "undef",
                                           -format => "ARRAY Reference",
                                           -example => "-esx_svmotion_destination_datastores => ['mydatastore1','mydatastore2']",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Controls what VMFS datastores to perform svmotions to and from. " .
                                               "This parameter is mutually exclusive with 'esx_svmotion_lun_path_regex'.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_svmotion_timeout",
                                           -package => __PACKAGE__,
                                           -default => "5400",
                                           -format => "seconds",
                                           -example => "-esx_svmotion_timeout => 5400",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Controls the timeout for the svmotion operation");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vmotion_destination_esx_server",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String",
                                           -example => "-esx_vmotion_destination_esx_server => 'destination_esx_server'",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies the destination Host name to migrate the virtual machines.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vmotion_priority",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String",
                                           -example => "-esx_vmotion_priority => 'highPriority'",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "The priority of the migration task: defaultPriority, highPriority, lowPriority");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vmotion_state",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String",
                                           -example => "-esx_vmotion_state => 'poweredOn'",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "State of the virtual machine: poweredOn, poweredOff, suspended");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vmotion_retry",
                                           -package => __PACKAGE__,
                                           -default => "10",
                                           -format => "Integer",
                                           -example => "-esx_vmotion_retry => 15",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "retry parameter for the virtual machine: any positive number");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_vmotion_restore_sleep_time",
                                           -package => __PACKAGE__,
                                           -default => "60",
                                           -format => "integer (seconds)",
                                           -example => "-esx_vmotion_restore_sleep_time => 120",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates the amount of seconds to sleep before restoring " .
                                                           "the state of an vmotion operation");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_svmotion_restore_sleep_time",
                                           -package => __PACKAGE__,
                                           -default => "60",
                                           -format => "integer (seconds)",
                                           -example => "-esx_svmotion_restore_sleep_time => 120",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates the amount of seconds to sleep before restoring " .
                                                           "the state of an svmotion operation");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_svmotion_interval_sleep_time",
                                           -package => __PACKAGE__,
                                           -default => "60",
                                           -format => "integer (seconds)",
                                           -example => "-esx_svmotion_interval_sleep_time => 120",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates the amount of seconds to sleep before moving on to the next " .
                                                           "svmotion operation");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_snapshot_name",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String",
                                           -example => "-esx_snapshot_name => 'snapshot name' ",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Name for new snapshot ");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
	
    $self->EventUtil::document_event_param(-param => "esx_snapshot_memory",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "integer",
                                           -example => "-esx_snapshot_memory => 1",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates whether the snapshot has to be created with memory or not ");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
	
    $self->EventUtil::document_event_param(-param => "esx_snapshot_quiesce",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "integer",
                                           -example => "-esx_snapshot_quiesce => 1 ",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates whether disk I/O has to be quiesced before snapshot creation. ");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
	
    $self->EventUtil::document_event_param(-param => "esx_vxdmp_io_policies",
                                           -package => __PACKAGE__,
                                           -default => "60",
                                           -format => "integer (seconds)",
                                           -example => "-esx_vxdmp_io_policies => ['round-robin','minimumq']",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates the list of VxDMP IO policies to choose from");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    
    $self->EventUtil::document_event_param(-param => "esx_lun_target_reset_count",
                                           -package => __PACKAGE__,
                                           -default => "10",
                                           -format => "integer (count)",
                                           -example => "-esx_lun_target_reset_count => 10",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates the number of iterations to run target|lun reset for a lun");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "esx_lun_target_reset_type",
                                           -package => __PACKAGE__,
                                           -default => "lunreset",
                                           -format => "string",
                                           -example => "-esx_lun_target_reset_type => targetreset",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates the type of reset targetreset|lunreset for a lun");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    # get a list of esxi sanclient objects
    my @sanclient_objs = &_get_esx_vma_sanclients($self);
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # verify that each host object has an MPDevMgr object created, otherwise build one.
    foreach my $esx_host_obj (@sanclient_objs) {
        if (not defined($esx_host_obj->mpdevmgr()) or ref($esx_host_obj->mpdevmgr()) !~ /MPDevMgr/i) {
            # build a mpdevmgr for this host
            my $esx_host_name = $esx_host_obj->hostname();

            # get the resource object for this sanclient
            my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                                       -type => "sanclients");
            if ($self->error()) {
                $self->_set_err("Could not obtain resource '$esx_host_name' of type 'sanclients'");
                return $EventUtil::FALSE;
            }

            # get the 7mode and Cmode components
            my @sanfilers7m_objs = $self->EventUtil::get_sanfiler_objs(-type => "7mode");
            if ($self->error()) {
                $self->_set_err("Unable to obtain sanfilers7m objects");
                return $EventUtil::FALSE;
            }

            my @cluster_objs = $self->EventUtil::get_nacl_cluster_cm_objs();
            if ($self->error()) {
                $self->_set_err("Unable to obtain controllercluster objects");
                return $EventUtil::FALSE;
            }

            # get the names of the vservers that are in scope to this sanclient resource, otherwise default
            # to the environment's value of vservers
            my @vserver_names = $self->EventUtil::get_vservers(-objects => [$resource],
                                                               -default_to_env => 1);
            if ($self->error()) {
                $self->_set_err("Unable to obtain vserver names");
                return $EventUtil::FALSE;
            }

            # create the esx host mpdevmgr
            $esx_host_obj->create_mpdevmgr(-sanfilers7m_objs => \@sanfilers7m_objs,
                                           -nacl_cluster_objs => \@cluster_objs,
                                           -vservers => \@vserver_names);
            if ($esx_host_obj->error()) {
                $self->_set_err($esx_host_obj->errmsg());
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method: esx_svmotion()
# Objective: Perform a Storage vMotion operation against the virtual machines
#            specified in an ESX host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_svmotion {
    my ($self) = @_;

    # get a list of esxi sanclient objects
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # kick off svmotion subtests per esx host object in parallel
    my @subtests = ();
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $runid = "svmotion__" . $esx_host_name;
        my $subtest = subtest(
            sub {
                $self->_esx_svmotion_per_esx_host(-esx_host_obj => $esx_host_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }

                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        push(@subtests, $subtest);
    }

    # reap the subtests
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to svmotion all ESX clients");
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: esx_vmotion()
# Objective: Perform a vMotion operation against the virtual machines
#            specified in an ESX host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_vmotion {
    my ($self) = @_;

    # get a list of esxi sanclient objects
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # kick off vmotion subtests per esx host object in parallel
    my @subtests;
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $runid = "vmotion__" . $esx_host_name;
        my $subtest = subtest(
            sub {
                $self->_esx_vmotion_per_esx_host(-esx_host_obj => $esx_host_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);
    }

    # reap the subtests
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to vmotion all ESX clients");
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: esx_snapshot()
# Objective: Perform a Snapshot operation against the virtual machines
#            specified in an ESX host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_snapshot {
    my ($self) = @_;
    # get a list of esxi sanclient objects
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }
    # kick off snapshot subtests per esx host object in parallel
    my @subtests = ();
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $runid = "snapshot__" . $esx_host_name;
        my $subtest = subtest(
            sub {
                $self->_esx_snapshot_per_esx_host(-esx_host_obj => $esx_host_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        push(@subtests, $subtest);
    }

    # reap the subtests
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to create snapshot for all VMs ");
		return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: esx_vxdmp_rotate_io_policies()
# Objective: Rotate VxDMP IO policies
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_vxdmp_rotate_io_policies {
    my ($self) = @_;

    # get a list of esxi sanclient objects
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # kick off subtests per esx host object in parallel
    my @subtests;
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $runid = "vxdmp_rotate_io_policies__" . $esx_host_name;
        my $subtest = subtest(
            sub {
                $self->_esx_vxdmp_rotate_io_policies_per_esx_host(-esx_host_obj => $esx_host_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);
    }

    # reap the subtests
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to change all VxDMP IO Policies all ESX clients");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: esx_lun_target_reset()
# Objective: Perform a target|lun reset operation on all Netapp luns in an ESX host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_lun_target_reset {
    my ($self) = @_;
    # get a list of esx sanclient objects
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }
    # kick off  target|lun reset  subtests per esx host object in parallel
    my @subtests;
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $runid = "reset__" . $esx_host_name;
        my $subtest = subtest(
            sub {
                $self->_esx_lun_target_reset_per_esx_host(-esx_host_obj => $esx_host_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);
    }
    # reap the subtests
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to reset all ESX clients");
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: esx_add_remove_vmdk()
# Objective: Perform the add/remove operations on a group of esx machines.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_add_remove_vmdk {
    my ($self) = @_;
    while (1) {
        $self->esx_run_event_on_clients(event_function => "_esx_add_remove_vmdk");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

}
###############################################################################
# Event Method: esx_vaai_and_unmap()
# Objective: Perform the VAAI and unmap operations on a group of esx machines.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_vaai_and_unmap {
    my ($self) = @_;

    $self->esx_run_event_on_clients(event_function => "_esx_vaai_and_unmap");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

}
###############################################################################
# Event Method: esx_vaai_umap_xcopy()
# Objective: Perform the vaai/unmap/xcopy operations on a group of esx machines.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_vaai_unmap_xcopy {
    my ($self) = @_;

    $self->esx_run_event_on_clients(event_function => "_esx_vaai_unmap_xcopy");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

}
###############################################################################
# Event Method: esx_run_event_on_clients()
# Objective: Perform the specified event operations on a group of esx machines.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_run_event_on_clients {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_function = &ParamsMgr::get_arg(-arg => "event_function",
                                             -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
    # get a list of esxi sanclient objects
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # kick off running an event subtests per esx host object in parallel
    my @subtests;
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $runid = $event_function . "__" . $esx_host_name;
        my $subtest = subtest(
            sub {
                $self->$event_function(-esx_host_obj => $esx_host_obj);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);
    }

    # reap the subtests
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to $event_function all ESX clients");
    }

    return $EventUtil::TRUE;
}
###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Event Method: _esx_svmotion_per_esx_host()
# Objective: Perform svmotion operations on a single esx host
###############################################################################
sub _esx_svmotion_per_esx_host {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $esx_host_name = $esx_host_obj->hostname();

    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the list of virtual machines that the user wants us to svmotion
    my $virtual_machines = $self->EventUtil::get_arg(-arg => "virtual_machines",
                                                     -objects => [$resource],
                                                     -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get list of virtual machines for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the destination lun path regex that the user wants us to svmotion to
    my $esx_svmotion_lun_path_regex = $self->EventUtil::get_arg(-arg => "esx_svmotion_lun_path_regex",
                                                                -objects => [$resource],
                                                                -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_svmotion_lun_path_regex parameter");
        return $EventUtil::FALSE;
    }

    # get the destination datastore names that the user wants us to svmotion to
    my $esx_svmotion_destination_datastores = $self->EventUtil::get_arg(-arg => "esx_svmotion_destination_datastores",
                                                                        -objects => [$resource],
                                                                        -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_svmotion_destination_datastores parameter");
        return $EventUtil::FALSE;
    }

    # get the vcenter server name to use for this svmotion
    my $vcenter_server = $self->EventUtil::get_arg(-arg => "esx_vcenter_server",
                                                   -objects => [$resource],
                                                   -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vcenter_server parameter");
        return $EventUtil::FALSE;
    }

    # get the datacenter name to use for this svmotion
    my $datacenter = $self->EventUtil::get_arg(-arg => "esx_vcenter_datacenter",
                                               -objects => [$resource],
                                               -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vcenter_datacenter parameter");
        return $EventUtil::FALSE;
    }

    # build the destination datastores array
    my @destination_datastores;

    # discover the datastores based on the lun path regex
    if (defined($esx_svmotion_lun_path_regex)) {
        my @mpdev_objs = $esx_host_obj->mpdevmgr()->get_filtered_devices(-ontap_lun_path_regex => $esx_svmotion_lun_path_regex);
        if ($esx_host_obj->mpdevmgr()->error()) {
            $self->_set_err($esx_host_obj->mpdevmgr()->errmsg());
            return $EventUtil::FALSE;
        }

        foreach my $mpdev_obj (@mpdev_objs) {
            my $destination_datastore = $mpdev_obj->get_os_specific_data("datastore_label");
            if (not defined($destination_datastore)) {
                # issue a warning saying that this is not a VMFS lun
                &SanUtil::nate_log('-warn', "Cannot svmotion to a non-VMFS LUN " .
                                       $mpdev_obj->mp_device_name() . " (" . $mpdev_obj->ontap_vserver() .
                                           ":" . $mpdev_obj->ontap_lun_path() . ")");
                next;
            }
            push (@destination_datastores, $destination_datastore);
        }
    } else {
        # otherwise use the destination datastore list specified by the event
        @destination_datastores = @{$esx_svmotion_destination_datastores};
    }

    # validate that we actually have datastores to use
    if (not scalar(@destination_datastores)) {
        $self->_set_err("Could not find any destination VMFS datastores to use for svmotion on host $esx_host_name");
        return $EventUtil::FALSE;
    }

    # foreach virtual machine, perform the svmotion
    foreach my $virtual_machine (@{$virtual_machines}) {

        # walk the destination datastore list
        foreach my $destination_datastore (@destination_datastores) {
			
            # determine where all the disks are currently located for this virtual machine
            my $old_vmdks = $esx_host_obj->get_vmdks_from_vm(-vm_name => $virtual_machine,
                                                             -return_hash => "scsi_id");
            if ($esx_host_obj->error()) {
                $self->_set_err("Failed to obtain disks from $virtual_machine", $esx_host_obj->errmsg());
                return $EventUtil::FALSE;
            }

            # build the vmdk hash to keep track of where each disk went and where it came from
            my $vmdk_hash;
            foreach my $scsi_id (keys %{$old_vmdks}) {
                my $vmdk_path = $old_vmdks->{$scsi_id};
                my ($datastore, $subdir_path) = ($vmdk_path =~ /^\s*\[(.+?)\]\s+(.+)/);
                $vmdk_hash->{$scsi_id}->{original_subdir_path} = $subdir_path;
                $vmdk_hash->{$scsi_id}->{original_datastore} = $datastore;
            }

            # build the inputs for the svmotion command
            my %svmotion_args = ();
            $svmotion_args{-vcenter_server} = $vcenter_server;
            $svmotion_args{-destination_datastore} = $destination_datastore;
            $svmotion_args{-vm_name} = $virtual_machine;
            if (defined($datacenter)) {
                $svmotion_args{-datacenter} = $datacenter;
            }
            $svmotion_args{-timeout} = $self->EventUtil::get_arg(-arg => "esx_svmotion_timeout",
                                                                 -objects => [$resource],
                                                                 -default_to_env => 1);

            # run svmotion
            my $vmx_hash = $esx_host_obj->svmotion(%svmotion_args);
            if ($esx_host_obj->error()) {
                $self->_set_err("Failed to run Storage vmotion against $virtual_machine", $esx_host_obj->errmsg());
                return $EventUtil::FALSE;
            }

            $self->EventUtil::sleep_arg(-arg => "esx_svmotion_restore_sleep_time");

            # now svmotion everything back to where it came from
            # determine where all the disks are currently located for this virtual machine
            my $new_vmdks = $esx_host_obj->get_vmdks_from_vm(-vm_name => $virtual_machine,
                                                             -return_hash => "scsi_id");
            if ($esx_host_obj->error()) {
                $self->_set_err("Failed to obtain disks from $virtual_machine", $esx_host_obj->errmsg());
                return $EventUtil::FALSE;
            }

            # build the vmdk spec to revert all the disks back to their original locations
            my $vmdk_spec;
            foreach my $scsi_id (keys %{$new_vmdks}) {
                my $vmdk_path = $new_vmdks->{$scsi_id};
                my ($datastore, $subdir_path) = ($vmdk_path =~ /^\s*\[(.+?)\]\s+(.+)/);
                $vmdk_hash->{$scsi_id}->{new_subdir_path} = $subdir_path;
                $vmdk_hash->{$scsi_id}->{new_datastore} = $datastore;
                $vmdk_spec->{"$vmdk_path"} = $vmdk_hash->{$scsi_id}->{original_datastore};
            }

            # rebuild the svmotion args and restore the state of the vm's svmotion.
            # need to specify the destination datastore to be the old datastore for this virtual machine
            $svmotion_args{-destination_datastore} = $vmx_hash->{$virtual_machine}->{old_datastore};

            # plug in the vmdk spec
            $svmotion_args{-vmdk_spec} = $vmdk_spec;

            # svmotion everything back to where it came from
            $esx_host_obj->svmotion(%svmotion_args);
            if ($esx_host_obj->error()) {
                $self->_set_err("Failed to run Storage vmotion against $virtual_machine to restore state", $esx_host_obj->errmsg());
                return $EventUtil::FALSE;
            }

            # for good measure, display to the logs that the virtual machine has been restored
            $esx_host_obj->get_vmdks_from_vm(-vm_name => $virtual_machine,
                                             -return_hash => "scsi_id");
            if ($esx_host_obj->error()) {
                $self->_set_err("Failed to obtain disks from $virtual_machine", $esx_host_obj->errmsg());
                return $EventUtil::FALSE;
            }

            $self->EventUtil::sleep_arg(-arg => "esx_svmotion_interval_sleep_time");
        }
    }

    return $EventUtil::TRUE;
}
###############################################################################
# Event Method: _esx_vmotion_per_esx_host()
# Objective: Perform vmotion operations on a single esx host
###############################################################################
sub _esx_vmotion_per_esx_host {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get source server object
    my $esx_host_name = $esx_host_obj->hostname();
    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the list of virtual machines that the user wants us to vmotion
    ## get the list of virtual machines related to  source server
    my $virtual_machines = $self->EventUtil::get_arg(-arg => "virtual_machines",
                                                     -objects => [$resource],
                                                     -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get list of virtual machines for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the destination esx server that the user wants us to vmotion to
    my $esx_vmotion_destination_esx_server = $self->EventUtil::get_arg(-arg => "esx_vmotion_destination_esx_server",
                                                                       -objects => [$resource],
                                                                       -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vmotion_destination_esx_server parameter");
        return $EventUtil::FALSE;
    }

    # get the priority
    my $esx_vmotion_priority = $self->EventUtil::get_arg(-arg => "esx_vmotion_priority",
                                                         -objects => [$resource],
                                                         -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vmotion_priority parameter");
        return $EventUtil::FALSE;
    }

    # get the state
    my $esx_vmotion_state = $self->EventUtil::get_arg(-arg => "esx_vmotion_state",
                                                      -objects => [$resource],
                                                      -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vmotion_state parameter");
        return $EventUtil::FALSE;
    }

    # get the retry parameter
    my $esx_vmotion_retry = $self->EventUtil::get_arg(-arg => "esx_vmotion_retry",
                                                      -objects => [$resource],
                                                      -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vmotion_retry parameter");
        return $EventUtil::FALSE;
    }
    # get the vcenter server name to use for this vmotion
    my $vcenter_server = $self->EventUtil::get_arg(-arg => "esx_vcenter_server",
                                                   -objects => [$resource],
                                                   -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vcenter_server parameter");
        return $EventUtil::FALSE;
    }

    # foreach virtual machine, perform the vmotion
    my @subtests;
    foreach my $virtual_machine (@{$virtual_machines}) {
        my %vmotion_args = ();
        $vmotion_args{-vcenter_server} = $vcenter_server;
        $vmotion_args{-target_host} = $esx_vmotion_destination_esx_server;
        $vmotion_args{-source_host} = $esx_host_name;
        $vmotion_args{-vm_name} = $virtual_machine;
        if (defined($esx_vmotion_priority)) {
            $vmotion_args{-vmotion_priority} = $esx_vmotion_priority;
        }
        if (defined($esx_vmotion_state)) {
            $vmotion_args{-vm_powerstate} = $esx_vmotion_state;
        }

        my $runid = "migrate_" . $virtual_machine;
        my $retry_count;
        my $subtest_vmotion = subtest(
            sub {
                for($retry_count=1 ;$retry_count<=$esx_vmotion_retry ;$retry_count++){
                    $esx_host_obj->reconnect();
                    if ($esx_host_obj->error()) {
                                        &SanUtil::nate_log('-error', "Unable to reconnect to " .
                                                                    $esx_host_obj->hostname()."\n".
                                                                    $esx_host_obj->errmsg);
                                        return $EventUtil::FALSE;
                    }
                    $esx_host_obj->vmotion(%vmotion_args);
                    if ($esx_host_obj->error()) {
                        my $retry_count_left = $esx_vmotion_retry - $retry_count;
                        &SanUtil::nate_log('-warn', $esx_host_obj->errmsg() . "Iteration [$retry_count_left left of $esx_vmotion_retry]");
                        if($retry_count < $esx_vmotion_retry){
                            $esx_host_obj->_clear_err();
                            next;
                        }
                        $self->_set_err("Failed to vmotion $virtual_machine", $esx_host_obj->errmsg() . "Retried [$retry_count of $esx_vmotion_retry] times");
                        $self->SubtestComm::subtest_fail();
                        return $EventUtil::FALSE;
                    }
                    last;
                }
                &SanUtil::nate_log('-comment', "$virtual_machine Migrated Successfully from $esx_host_name to $esx_vmotion_destination_esx_server");
                $self->EventUtil::sleep_arg(-arg => "esx_vmotion_restore_sleep_time");

                ## Reverting Prcess start here for the virtual machine
                $vmotion_args{-target_host} = $esx_host_name;
                $vmotion_args{-source_host} = $esx_vmotion_destination_esx_server;
                for($retry_count=1 ;$retry_count<=$esx_vmotion_retry ;$retry_count++){
                    $esx_host_obj->vmotion(%vmotion_args);
                    if ($esx_host_obj->error()) {
                        my $retry_count_left = $esx_vmotion_retry - $retry_count;
                        &SanUtil::nate_log('-warn', $esx_host_obj->errmsg() . "Iteration [$retry_count_left left of $esx_vmotion_retry]");
                        if($retry_count < $esx_vmotion_retry){
                            $esx_host_obj->_clear_err();
                            next;
                        }
                        $self->_set_err("Failed to run vmotion against $virtual_machine to restore state", $esx_host_obj->errmsg() . "Retried [$retry_count of $esx_vmotion_retry] times");
                        $self->SubtestComm::subtest_fail();
                        return $EventUtil::FALSE;
                    }
                    last;
                }
                &SanUtil::nate_log('-comment', "$virtual_machine reverted Successfully from $esx_vmotion_destination_esx_server to $esx_host_name");
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        push(@subtests, $subtest_vmotion);
    }
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to vmotion all virtual machines");
    }

    return $EventUtil::TRUE;
}
###############################################################################
# Event Method: _esx_snapshot_per_esx_host()
# Objective: Perform snapshot operations on a single esx host
###############################################################################
sub _esx_snapshot_per_esx_host {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
	
	#get source server object
    my $esx_host_name = $esx_host_obj->hostname();
    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the list of virtual machines that the user wants to take snapshot
    my $virtual_machines = $self->EventUtil::get_arg(-arg => "virtual_machines",
                                                     -objects => [$resource],
                                                     -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get list of virtual machines for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }
	
    #get the snapshot name
    my $esx_snapshot_name = $self->EventUtil::get_arg(-arg => "esx_snapshot_name",
                                                      -objects => [$resource],
                                                      -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_snapshot_name parameter");
        return $EventUtil::FALSE;
    }
		
    #get the memory parameter for snapshpt creation
    my $esx_snapshot_memory = $self->EventUtil::get_arg(-arg => "esx_snapshot_memory",
                                                         -objects => [$resource],
                                                         -default_to_env => 0);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_snapshot_memory parameter");
        return $EventUtil::FALSE;
    }

    #get the quiesce parameter for snapshpt creation
    my $esx_snapshot_quiesce = $self->EventUtil::get_arg(-arg => "esx_snapshot_quiesce",
                                                         -objects => [$resource],
                                                         -default_to_env => 0);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_snapshot_quiesce parameter");
        return $EventUtil::FALSE;
    }
	my @subtests;
    foreach my $virtual_machine (@{$virtual_machines}) {
        my %snapshot_args = ();
		$snapshot_args{-vm_name} = $virtual_machine;
		my $retval = time();
		my $local_time = gmtime( $retval);
		$local_time =~ s/\s+/_/g;
		$snapshot_args{-snapshot_name} = $virtual_machine."_".$local_time;
		$snapshot_args{-memory} = $esx_snapshot_memory;
		$snapshot_args{-quiesce} = $esx_snapshot_quiesce;
        my $runid = "snapshot_".$virtual_machine;
        my $subtest_snapshot = subtest(
            sub {                
				$esx_host_obj->reconnect();
				if ($esx_host_obj->error()) {
									&SanUtil::nate_log('-error', "Unable to reconnect to " .
																$esx_host_obj->hostname()."\n".
																$esx_host_obj->errmsg);
									return $EventUtil::FALSE;
				}
				$esx_host_obj->create_virtual_machine_snapshot(%snapshot_args);
				if ($esx_host_obj->error()) {
					$self->_set_err("Failed to create snapshot $virtual_machine", $esx_host_obj->errmsg());
					$self->SubtestComm::subtest_fail();
					return $EventUtil::FALSE;
				}
                &SanUtil::nate_log('-comment', "$virtual_machine snapshot creation successful");
                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            }, -runid => $runid, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        push(@subtests, $subtest_snapshot);
    }
	$self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Failed to create snapshot all virtual machines");
		return $EventUtil::FALSE;
    }
    return $EventUtil::TRUE;  
}
###############################################################################
# Event Method: _esx_lun_target_reset_per_esx_host()
# Objective: Perform lun_target_reset operations on a single esx host
###############################################################################
sub _esx_lun_target_reset_per_esx_host {
    my ($self, %args) = @_;
    &ParamsMgr::process_args(%args);
    my $esx_host_obj = &ParamsMgr::get_arg( -arg => "esx_host_obj",
                                            -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
    # get source server object
    my $esx_host_name = $esx_host_obj->hostname();
    my $resource = $self->ResMgr::get_resource( -name => $esx_host_name,
                                                -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }
    # get the reset count
    my $esx_lun_target_reset_count = $self->EventUtil::get_arg(-arg => "esx_lun_target_reset_count",
                                                               -objects => [$resource],
                                                               -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get reset count parameter");
        return $EventUtil::FALSE;
    }
    # get reset type
    my $esx_lun_target_reset_type = $self->EventUtil::get_arg(-arg => "esx_lun_target_reset_type",
                                                              -objects => [$resource],
                                                              -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get reset count parameter");
        return $EventUtil::FALSE;
    }
    # get lun naa ids of all NETAPP devices

     $esx_host_obj->create_mpdevmgr();
     
     if($esx_host_obj->error()) {
        &SanUtil::nate_log('-error',"Unable to create mpdevmgr object for" .
        $esx_host_obj->hostname()."\n".
         $esx_host_obj->errmsg);
     }

     my @lun_naa_ids = $esx_host_obj->mpdevmgr()->get_device_names();

    if (scalar @lun_naa_ids){
        # Perform reset operation
        my @subtests;
        my %reset_args = ();
        $reset_args{-lun_naa_ids} = \@lun_naa_ids ;
        $reset_args{-reset_count} = $esx_lun_target_reset_count;
        $reset_args{-reset_type} = $esx_lun_target_reset_type;
        $esx_host_obj->reconnect();
        if ($esx_host_obj->error()) {
            $self->set_err('-error', "Unable to reconnect to " .
            $esx_host_obj->hostname()."\n".
            $esx_host_obj->errmsg);
            return $EventUtil::FALSE;
        }
        $esx_host_obj->lun_target_reset(%reset_args);
        if ($esx_host_obj->error()) {
            $self->_set_err("Failed to reset : " . $esx_host_obj->errmsg());
            return $EventUtil::FALSE;
        }
        &SanUtil::nate_log('-comment', "Reset operation completed successfully for host $esx_host_name");
    }else{
            $self->_set_err("No NETAPP devices found.");
            return $EventUtil::FALSE;
    }
    return $EventUtil::TRUE;
}
###############################################################################
# Event Method: _esx_vxdmp_rotate_io_policies_per_esx_host()
# Objective: Perform vxdmp rotate io policy operations on a single esx host
###############################################################################
sub _esx_vxdmp_rotate_io_policies_per_esx_host {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $esx_host_name = $esx_host_obj->hostname();

    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    my $esx_vxdmp_io_policies = $self->EventUtil::get_arg(-arg => "esx_vxdmp_io_policies",
                                                          -objects => [$resource],
                                                          -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get esx_vxdmp_io_policies parameter");
        return $EventUtil::FALSE;
    }

    # check all the devices, if none are VxDMP owned, then return immediately.
    my $enclosures = {};
    my $found = 0;
    my @mpdev_objs = $esx_host_obj->mpdevmgr()->get_device_objs();
    foreach my $device_obj (@mpdev_objs) {
        next if ($device_obj->mp_type() !~ /VxDMP/i);
        $found = 1;
        my $mp_device_name = $device_obj->mp_device_name();
        my $enclosure_name = $device_obj->get_os_specific_data("vxdmp_enclosure_name");
        if (not $enclosure_name) {
            &SanUtil::nate_log('-warn', "Could not determine VxDMP enclosure name for $mp_device_name, " .
                               "skipping this device.");
            next;
        }

        # store the enclosures in a hash to get rid of dups
        $enclosures->{$enclosure_name}++;
    }

    # if no VxDMP devices were found, then return
    if (not $found) {
        &SanUtil::nate_log('-comment', "Skipping $esx_host_name because it does not have any LUNs under VxDMP control");
        return $EventUtil::TRUE;
    }

    # foreach VxDMP enclosure that we find, change the io policy to the ones specified in
    # esx_vxdmp_io_policies, loop around the list of policies to choose from.
    if (keys %{$enclosures}) {
        my $i = 0;
        foreach my $enclosure (keys %{$enclosures}) {
            # pick the next policy in the list
            my $policy = $esx_vxdmp_io_policies->[$i % scalar(@{$esx_vxdmp_io_policies})];
            $esx_host_obj->mp()->set_vxdmp_io_policy(-enclosure => $enclosure,
                                                     -io_policy => $policy);
            if ($esx_host_obj->mp()->error()) {
                $self->_set_err("Failed to set IO policy for VxDMP enclosure $enclosure to");
                return $EventUtil::FALSE;
            }
            $i++;
        }
    } else {
        # no enclosures were found, but this is VxDMP host... abort
        $self->_set_err("Did not find Any VxDMP enclosures for any of the VxDMP devices on the system. " .
                        "Are you sure you installed vxdmpadm on this vMA?");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _esx_add_remove_vmdk()
# Objective: Perform add_remove_vmdk operations on a single esx host
###############################################################################
sub _esx_add_remove_vmdk {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get source server object
    my $esx_host_name = $esx_host_obj->hostname();
    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the file name
    my $esx_add_remove_vmdk_name = $self->EventUtil::get_arg(-arg => "esx_add_remove_vmdk_name",
                                                             -objects => [$resource],
                                                             -default_to_env => 1);
    if ($self->error() or not defined($esx_add_remove_vmdk_name)) {
            $self->_set_err("Failed to get list ESX file name.");
            return $EventUtil::FALSE;
    }
    # get the file size
    my $esx_host_ops_file_size = $self->EventUtil::get_arg(-arg => "esx_host_ops_file_size",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($self->error() or not defined($esx_host_ops_file_size)) {
            $self->_set_err("Failed to get list ESX file size.");
            return $EventUtil::FALSE;
    }

    my $esx_host_ops_sleep_time = $self->EventUtil::get_arg(-arg => "esx_host_ops_sleep_time",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($self->error() or not defined($esx_host_ops_sleep_time)) {
            $self->_set_err("Failed to get ESX host ops sleep time.");
            return $EventUtil::FALSE;
    }    
    
    #Check to see if datastore is overridden by a per client param
    my $index = 0;
    # clean datastore
    $esx_add_remove_vmdk_name =~ m!(/.*/.*/)!;
    my $clean_dir = $1;
    my $clean_cmd =  "rm -rf $clean_dir/*";

    my ($status, @output) = $esx_host_obj->esxi_direct_execute(-cmd => $clean_cmd);
    if ($status > 0) {
        $self->_set_err("Error cleaning datastore ");
        return $EventtUtil::FALSE;
    }
        
    while ($index < 10) {
        my $filename = $esx_add_remove_vmdk_name . "_" . $index . ".vmdk";
        $esx_host_obj->create_vmdk(-vmdk_file_path => $filename,
                                   -vmdk_size_in_kbytes => $esx_host_ops_file_size);
        if ($esx_host_obj->error()) {
            $self->_set_err("Could not create vmdk $filename");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_sleep($esx_host_ops_sleep_time);

        $esx_host_obj->delete_vmdk(vmdk_file => $filename);
        if ($esx_host_obj->error()) {
            $self->_set_err("Error deleting vmdk $filename");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_sleep($esx_host_ops_sleep_time);
        $index++;
        &SanUtil::nate_log('-info', "Completed $index passes of vaai and unmap.");
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _esx_vaai_and_unmap()
# Objective: Perform VAAI and unmap operations on a single esx host
###############################################################################
sub _esx_vaai_and_unmap {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get source server object
    my $esx_host_name = $esx_host_obj->hostname();
    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

    # get the file name
    my $esx_add_remove_vmdk_name = $self->EventUtil::get_arg(-arg => "esx_add_remove_vmdk_name",
                                                             -objects => [$resource],
                                                             -default_to_env => 1);
    if ($self->error() or not defined($esx_add_remove_vmdk_name)) {
            $self->_set_err("Failed to get list ESX file name.");
            return $EventUtil::FALSE;
    }

    # get the file size
    my $esx_host_ops_file_size = $self->EventUtil::get_arg(-arg => "esx_host_ops_file_size",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($self->error() or not defined($esx_host_ops_file_size)) {
            $self->_set_err("Failed to get list ESX file size.");
            return $EventUtil::FALSE;
    }

    my $esx_host_ops_sleep_time = $self->EventUtil::get_arg(-arg => "esx_host_ops_sleep_time",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($self->error() or not defined($esx_host_ops_sleep_time)) {
            $self->_set_err("Failed to get ESX host ops sleep time.");
            return $EventUtil::FALSE;
    }

    $esx_add_remove_vmdk_name =~ m!(/.*/.*/)!;
    my $clean_dir = $1;
    my $clean_cmd =  "rm -rf $clean_dir/*";

    my ($status, @output) = $esx_host_obj->esxi_direct_execute(-cmd => $clean_cmd);
    if ($status > 0) {
        $self->_set_err("Error cleaning datastore ");
        return $EventtUtil::FALSE;
    }

    my $pass_count = 1;
    while (1) { 
        my $local_index = 0;
        my $count = 10; 
        while ($local_index < $count) {
            my $filename = $esx_add_remove_vmdk_name . "_" . $local_index . ".vmdk";
            $esx_host_obj->create_vmdk(-vmdk_file_path => $filename,
                                       -vmdk_size_in_kbytes => $esx_host_ops_file_size);
            if ($esx_host_obj->error()) {
                $self->_set_err("Could not create vmdk $filename");
                return $EventUtil::FALSE;
            }
            $local_index++;
        }

        &SanUtil::nate_sleep($esx_host_ops_sleep_time);
    
        $local_index = 0;
        while ($local_index < $count) {
            my $filename = $esx_add_remove_vmdk_name . "_" . $local_index . ".vmdk";
            $esx_host_obj->delete_vmdk(vmdk_file => $filename);
            if ($esx_host_obj->error()) {
                $self->_set_err("Error deleting vmdk $filename");
                return $EventUtil::FALSE;
            }
            $local_index++;
        }

        $esx_add_remove_vmdk_name  =~ m!(/.*/.*/)!;
        my $unmap_directory = $1;
        $esx_host_obj->unmap(reclaim_percentage => '60',
                             unmap_directory    => $unmap_directory);

        if ($esx_host_obj->error()) {
            $self->_set_err("Error running command esx unmap", $esx_host_obj->errmsg());
            return $EventUtil::FALSE;
        }
        &SanUtil::nate_log('-info', "Completed $pass_count passes of vaai and unmap.");
        $pass_count++;
    }
    return $EventUtil::TRUE;
}
###############################################################################
# Event Method: _esx_vaai_unmap_xcopy()
# Objective: Perform VAAI, unmap, and xcopy operations on a single esx host
###############################################################################
sub _esx_vaai_unmap_xcopy {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $esx_host_obj = &ParamsMgr::get_arg(-arg => "esx_host_obj",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get source server object
    my $esx_host_name = $esx_host_obj->hostname();
    my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
        return $EventUtil::FALSE;
    }

     # get the file name
    my $esx_add_remove_vmdk_name = $self->EventUtil::get_arg(-arg => "esx_add_remove_vmdk_name",
                                                             -objects => [$resource],
                                                             -default_to_env => 1);
    if ($self->error() or not defined($esx_add_remove_vmdk_name)) {
            $self->_set_err("Failed to get  ESX file name.");
            return $EventUtil::FALSE;
    }

    # get the file size
    my $esx_host_ops_file_size = $self->EventUtil::get_arg(-arg => "esx_host_ops_file_size",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($self->error() or not defined($esx_host_ops_file_size)) {
            $self->_set_err("Failed to get ESX file size.");
            return $EventUtil::FALSE;
    }

    my $esx_host_ops_sleep_time = $self->EventUtil::get_arg(-arg => "esx_host_ops_sleep_time",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($self->error() or not defined($esx_host_ops_sleep_time)) {
            $self->_set_err("Failed to get ESX host ops sleep time.");
            return $EventUtil::FALSE;
    }

    my $esx_xcopy_src_name = $self->EventUtil::get_arg(-arg => "esx_xcopy_src_name",
                                                       -objects => [$resource],
                                                       -default_to_env => 1);
    if ($self->error()) {
            $self->_set_err("Failed to get ESX host ops xcopy src.");
            return $EventUtil::FALSE;
    }

    my $esx_xcopy_thin_dest_name = $self->EventUtil::get_arg(-arg => "esx_xcopy_thin_dest_name",
                                                             -objects => [$resource],
                                                             -default_to_env => 1);
    if ($self->error()) {
            $self->_set_err("Failed to get ESX host ops xcopy thin destination.");
            return $EventUtil::FALSE;
    }

    my $esx_xcopy_thick_dest_name = $self->EventUtil::get_arg(-arg => "esx_xcopy_thick_dest_name",
                                                              -objects => [$resource],
                                                              -default_to_env => 1);
    if ($self->error()) {
            $self->_set_err("Failed to get ESX host ops xcopy thick destination.");
            return $EventUtil::FALSE;
    }

    if (not defined($esx_add_remove_vmdk_name)
        or not defined($esx_xcopy_src_name)
        or not defined($esx_xcopy_thin_dest_name)
        or not defined($esx_xcopy_thick_dest_name)) {
            $self->_set_err("Failed to get ESX host ops filenames.");
            return $EventUtil::FALSE;
    }

    # clean datastore
    $esx_add_remove_vmdk_name =~ m!(/.*/.*/)!;
    my $clean_dir = $1;
    my $clean_cmd =  "rm -rf $clean_dir/*";

    my ($status, @output) = $esx_host_obj->esxi_direct_execute(-cmd => $clean_cmd);
    if ($status > 0) {
        $self->_set_err("Error cleaning datastore ");
        return $EventUtil::FALSE;
    }
    my $index = 0;
    while (1) {
        
        ($status) = $self->_esx_vaai_and_unmap(%args);
        if ($status ne $EventUtil::TRUE) {
            $self->_set_err("Could not perform vaai and unmap.");
            return $EventUtil::FALSE;
        }

        #create a VMDK source for cloning
        $esx_host_obj->create_vmdk(-vmdk_file_path => $esx_xcopy_src_name,
                                   -vmdk_size_in_kbytes => $esx_host_ops_file_size);
        if ($esx_host_obj->error()) {
            $self->_set_err("Could not create vmdk $esx_xcopy_src_name");
            return $EventUtil::FALSE;
        }

        my $clone_dest_thin_filename = $esx_xcopy_thin_dest_name . "_clone_dest_thin.vmdk";
        my $clone_dest_thick_filename = $esx_xcopy_thin_dest_name . "_clone_dest_thick.vmdk";

        # create clones on the two datastores
        $esx_host_obj->clone_vmdk(source_vmdk_file      => $esx_xcopy_src_name,
                                  destination_vmdk_file => $clone_dest_thin_filename,
                                  disk_format           => 'thin');
        if ($esx_host_obj->error()) {
            $self->_set_err("Could not clone vmdk $esx_xcopy_src_name");
            return $EventUtil::FALSE;
        }
    
        $esx_host_obj->clone_vmdk(source_vmdk_file      => $esx_xcopy_src_name,
                                  destination_vmdk_file => $clone_dest_thick_filename,
                                  disk_format           => 'eagerzeroedthick');
        if ($esx_host_obj->error()) {
            $self->_set_err("Could not clone vmdk $esx_xcopy_src_name");
            return $EventUtil::FALSE;
        }

        $esx_host_obj->delete_vmdk(vmdk_file => $clone_dest_thin_filename);
        if ($esx_host_obj->error()) {
            $self->_set_err("Could not delete clone $clone_dest_thin_filename");
            return $EventUtil::FALSE;
        }

        $esx_host_obj->delete_vmdk(vmdk_file => $clone_dest_thick_filename);
        if ($esx_host_obj->error()) {
            $self->_set_err("Could not delete clone $clone_dest_thick_filename");
            return $EventUtil::FALSE;
        }

        $esx_add_remove_vmdk_name =~ m!(/.*/.*/)!;
        my $unmap_directory = $1;
        
        $esx_host_obj->unmap(reclaim_percentage => '60',
                             unmap_directory    => $unmap_directory);
        
        if ($esx_host_obj->error()) {       
            $self->_set_err("Error running command esx unmap", $esx_host_obj->errmsg());
            return $EventUtil::FALSE;
        }   
    } 
    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _get_esx_vma_sanclients()
# Objective: Obtain sanclient objs for all ESX vMA hosts
###############################################################################
sub _get_esx_vma_sanclients {
    my ($self) = @_;

    # get the list of sanclients resources
    my @sanclient_objs = $self->EventUtil::get_sanclient_objs();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return wantarray() ? () : [];
    }

    # filter the list down to ESX hosts that are under vMA control ONLY
    @sanclient_objs = grep {
        $_->hosttype() =~ /vmware/i and $_->has_vma_host();
    } @sanclient_objs;
	# if there are no objs in the list fail
    if (not scalar(@sanclient_objs)) {
        $self->_set_err("There are no ESX SanClient resources under vMA control available");
        return wantarray() ? () : [];
    }

    return wantarray() ? @sanclient_objs : \@sanclient_objs;
}

###############################################################################
# Event Method: _esx_svmotion_validate()
# Objective: Validate parameters for the svmotion event
###############################################################################
sub _esx_svmotion_validate {
    my ($self) = @_;

    # get the list of sanclients resources
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # foreach sanclient (filter esx only)
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }

        # get the list of virtual machines that the user wants us to svmotion
        my $virtual_machines = $self->EventUtil::get_arg(-arg => "virtual_machines",
                                                         -objects => [$resource],
                                                         -default_to_env => 1);
        if ($self->error() or not defined($virtual_machines)) {
            $self->_set_err("Failed to get list of virtual machines for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }

        # get the destination datastore name that the user wants us to svmotion to
        my $esx_svmotion_lun_path_regex = $self->EventUtil::get_arg(-arg => "esx_svmotion_lun_path_regex",
                                                                -objects => [$resource],
                                                                -default_to_env => 1);
        if ($self->error()) {
            $self->_set_err("Failed to get esx_svmotion_lun_path_regex parameter");
            return $EventUtil::FALSE;
        }

        # get the destination datastore name that the user wants us to svmotion to
        my $esx_svmotion_destination_datastores = $self->EventUtil::get_arg(-arg => "esx_svmotion_destination_datastores",
                                                                        -objects => [$resource],
                                                                        -default_to_env => 1);
        if ($self->error()) {
            $self->_set_err("Failed to get esx_svmotion_destination_datastores parameter");
            return $EventUtil::FALSE;
        }

        # lun path regex and datastore destinations are mutually exclusive, however one must be present
        if (defined($esx_svmotion_lun_path_regex) ^ defined($esx_svmotion_destination_datastores)) {
            # then at least one of them was specified
            # verify that esx_svmotion_lun_path_regex is a regular expression
            if (defined($esx_svmotion_lun_path_regex) and ref($esx_svmotion_lun_path_regex) ne "Regexp") {
                $self->_set_err("esx_svmotion_lun_path_regex is not a valid regular expression");
                return $EventUtil::FALSE;
            }

            # verify that svmotion_datastore_destinations was properly specified
            if (defined($esx_svmotion_destination_datastores)) {
                if (ref($esx_svmotion_destination_datastores) ne "ARRAY") {
                    $self->_set_err("esx_svmotion_destination_datastores is not a valid array reference");
                    return $EventUtil::FALSE;
                }

                if (not scalar(@{$esx_svmotion_destination_datastores})) {
                    $self->_set_err("esx_svmotion_destination_datastores must be a non empty array reference");
                    return $EventUtil::FALSE;
                }
            }
        } else {
            $self->_set_err("esx_svmotion_lun_path_regex and svmotion_datastore_destinations are mutually exclusive, " .
                            "yet one must be specified. Only specify one at a time");
            return $EventUtil::FALSE;
        }

        # get the vcenter server name to use for this svmotion
        my $vcenter_server = $self->EventUtil::get_arg(-arg => "esx_vcenter_server",
                                                       -objects => [$resource],
                                                       -default_to_env => 1);
        if ($self->error() or not defined($vcenter_server)) {
            $self->_set_err("Failed to get esx_vcenter_server parameter");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _esx_snapshot_validate()
# Objective: Validate parameters for the snapshot event
###############################################################################
sub _esx_snapshot_validate {
    my ($self) = @_;

    # get the list of sanclients resources
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available  validate : ".$self->errmsg() );
        return $EventUtil::FALSE;
    }

    # foreach sanclient (filter esx only)
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }

        # get the list of virtual machines that the user wants us to take snapshots
        my $virtual_machines = $self->EventUtil::get_arg(-arg => "virtual_machines",
                                                         -objects => [$resource],
                                                         -default_to_env => 1);
        if ($self->error() or not defined($virtual_machines)) {
            $self->_set_err("Failed to get list of virtual machines for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }
    
	}
	return $EventUtil::TRUE;
}
###############################################################################
# Event Method: _esx_vxdmp_rotate_io_policies_validate
# Objective: Validate parameters for the vxdmp io policy rotate event
###############################################################################
sub _esx_vxdmp_rotate_io_policies_validate {
    my ($self) = @_;

    # get the list of sanclients resources
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # foreach sanclient (filter esx only)
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }

        my $esx_vxdmp_io_policies = $self->EventUtil::get_arg(-arg => "esx_vxdmp_io_policies",
                                                              -objects => [$resource],
                                                              -default_to_env => 1);
        if ($self->error()) {
            $self->_set_err("Failed to get esx_vxdmp_io_policies parameter");
            return $EventUtil::FALSE;
        }

        # verify that vxdmp io policies was specified properly
        if (defined($esx_vxdmp_io_policies)) {
            if (ref($esx_vxdmp_io_policies) ne "ARRAY") {
                $self->_set_err("esx_vxdmp_io_policies is not a valid array reference");
                return $EventUtil::FALSE;
            }
            if (not scalar(@{$esx_vxdmp_io_policies})) {
                $self->_set_err("esx_vxdmp_io_policies is empty. You must specify a list of VxDMP io policies");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _esx_vmotion_validate()
# Objective: Validate parameters for the vmotion event
###############################################################################
sub _esx_vmotion_validate {
    my ($self) = @_;

    # get the list of sanclients resources
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # foreach sanclient (filter esx only)
    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }

        # get the list of virtual machines that the user wants us to vmotion
        my $virtual_machines = $self->EventUtil::get_arg(-arg => "virtual_machines",
                                                         -objects => [$resource],
                                                         -default_to_env => 1);
        if ($self->error() or not defined($virtual_machines)) {
            $self->_set_err("Failed to get list of virtual machines for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }

        # get the destination esx server that the user wants us to  vmotion to
        my $esx_vmotion_destination_esx_server = $self->EventUtil::get_arg(-arg => "esx_vmotion_destination_esx_server",
                                                                           -objects => [$resource],
                                                                           -default_to_env => 1);
        if ($self->error()) {
            $self->_set_err("Failed to get esx_vmotion_destination_esx_server parameter");
            return $EventUtil::FALSE;
        }

        if ($esx_host_name eq $esx_vmotion_destination_esx_server) {
            $self->_set_err("$esx_host_name should be different from $esx_vmotion_destination_esx_server");
            return $EventUtil::FALSE;
        }

        # get the vcenter server name to use for this vmotion
        my $vcenter_server = $self->EventUtil::get_arg(-arg => "esx_vcenter_server",
                                                       -objects => [$resource],
                                                       -default_to_env => 1);
        if ($self->error() or not defined($vcenter_server)) {
            $self->_set_err("Failed to get esx_vcenter_server parameter");
            return $EventUtil::FALSE;
        }
    }
    return $EventUtil::TRUE;
}
###############################################################################
# Event Method: _esx_lun_target_reset_validate()
# Objective: Validate parameters for the vmotion event
###############################################################################
sub _esx_lun_target_reset_validate {
    my ($self) = @_;

    # get the list of sanclients resources
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
        $self->_set_err("There are no SanClient resources available");
        return $EventUtil::FALSE;
    }

    # foreach sanclient (filter esx only)
    if(!@sanclient_objs) {
         $self->_set_err("No SanClient provided");
    }

    foreach my $esx_host_obj (@sanclient_objs) {
        my $esx_host_name = $esx_host_obj->hostname();
        my $resource = $self->ResMgr::get_resource(-name => $esx_host_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Failed to get resource for (${esx_host_name} : sanclients)");
            return $EventUtil::FALSE;
        }
    }
    return $EventUtil::TRUE;
}
1;

=cut
=pod

=head1 Package Name

ESXHostEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for ESX (VMware) client side events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ClientEvents

=item EventMan::EventModules

=back

=head1 Utility Module Usage

None

=over 4

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-virtual_machines>

=over 4

Description: An array reference of virtual machine names. This
specifies the virtual machines to use for a given event

Per Resource Capable: No

Default: []

=back

=head2 I<-esx_vcenter_server>

=over 4

Description: Specifies the hostname of the vcenter server, this must resolve to a tharnhost file name.

Example vcenter tharnhost file:

 hosttype=vcenter
 default_ip=10.228.254.2
 default_username="root"
 default_password="vmware"

Per Resource Capable: No

Default: ""

=back

=head2 I<-esx_vcenter_datacenter>

=over 4

Description: Specifies the datacenter of the vcenter server

Per Resource Capable: No

Default: ""

=back

=head2 I<-esx_svmotion_lun_path_regex>

=over 4

Description: Controls what luns to perform svmotions to and from.
This parameter is mutually exclusive with 'esx_svmotion_destination_datastores'

Per Resource Capable: No

Default: undef

Example: qr/mysvmotion_luns/

=back

=head2 I<-esx_svmotion_destination_datastores>

=over 4

Description: Controls what VMFS datastores to perform svmotions to and from.
This parameter is mutually exclusive with 'esx_svmotion_lun_path_regex'

Per Resource Capable: No

Default: undef

Example: ['mydatastore1', 'mydatastore2']

=back

=head2 I<-esx_vmotion_destination_esx_server>

=over 4

Description: Specifies the destination Host name to migrate the virtual machines.

Per Resource Capable: No

Default: ""

=back

=head2 I<-esx_vmotion_priority>

=over 4

Description: The priority of the migration task: defaultPriority, highPriority, lowPriority.

Per Resource Capable: No

Default: ""

Example: "-esx_vmotion_priority => 'highPriority'"

=back

=head2 I<-esx_vmotion_state>

=over 4

Description: State of the virtual machine: poweredOn, poweredOff, suspended.

Per Resource Capable: No

Default: ""

Example: "-esx_vmotion_state => 'poweredOn'"

=back

=head2 I<-esx_vmotion_retry>

=over 4

Description: retry parameter for the virtual machine: any positive number.

Per Resource Capable: No

Default: 10

Example: "-esx_vmotion_retry => 15"

=back

=head2 I<-esx_vmotion_restore_sleep_time>

=over 4

Description: Indicates the amount of seconds to sleep before restoring
the state of an vmotion operation.

Per Resource Capable: No

Default: 60

=back

=head2 I<-esx_svmotion_restore_sleep_time>

=over 4

Description: Indicates the amount of seconds to sleep before restoring
the state of an svmotion operation.

Per Resource Capable: No

Default: 60

=back

=head2 I<-esx_svmotion_interval_sleep_time>

=over 4

Description: Indicates the amount of seconds to sleep moving on to the
next svmotion operation.

Per Resource Capable: No

Default: 60

=back

=head2 I<-esx_svmotion_timeout>

=over 4

Description: Indicates the timeout for the svmotion command

Per Resource Capable: yes

Default: 5400 (seconds)

=back

=head1 B<Event Keywords>

=head2 B<esx_svmotion>

Performs Storage vMotion operations against the virtual_machines owned by ESXi SanClient resources.
Each virtual machine will have all their storage migrated to each datastore specified by either
'esx_svmotion_lun_path_regex' xor 'esx_svmotion_destination_datastores' and restored to their original
locations. After each svmotion, it will sleep 'esx_svmotion_restore_sleep_wait' seconds, before restoring
the virtual machines storage to its original location. After the storage for a virtual machine has been
restored, it will then sleep 'esx_svmotion_interval_sleep_time' before moving on to the next lun/datastore to
svmotion to. Once all luns/datastores have been used, it will move on to the next virtual machine and repeat the cycle.
Each ESXi host will be operated against in parallel.

=over 4

=item Event Parameters

Required Event Parameters:
 esx_svmotion_lun_path_regex XOR esx_svmotion_destination_datastores
 virtual_machines
 esx_vcenter_server

Optional Event Parameters:
 esx_vcenter_datacenter
 esx_svmotion_interval_sleep_time
 esx_svmotion_restore_sleep_time

=item Examples

 # see EventMan::add_event() for API details
 $EventMan_obj->add_event(-event_name => "esx_svmotion",
                          -esx_svmotion_lun_path_regex => "qr/svmotion_lun/");
 if ($EventMan_obj->error()) {
     ...
 }

=back

=head2 B<esx_vmotion>

Performs vMotion operations against the virtual_machines owned by ESXi SanClient resources.
Each virtual machine will be migrated to destination_esx_server specified and restored to their original
esx_servers.After each vmotion, it will sleep 'esx_vmotion_restore_sleep_wait' seconds, before restoring
the virtual machines to its original esx server.Migration of virtual machines related to an esx_server is invoked parallely.
Each ESXi host will be operated against in parallel.

=over 4

=item Event Parameters

Required Event Parameters:
 esx_vmotion_destination_esx_server
 virtual_machines
 esx_vcenter_server

Optional Event Parameters:
 esx_vmotion_priority
 esx_vmotion_state
 esx_vmotion_retry
 esx_vmotion_restore_sleep_time

=item Examples

 # see EventMan::add_event() for API details
 $EventMan_obj->add_event(-event_name => "esx_vmotion",
                          -esx_vmotion_destination_esx_server => 'destination_esx_server');
 if ($EventMan_obj->error()) {
     ...
 }

=back

=head2 B<esx_snapshot>

Performs create snapshot operations against the virtual_machines owned by ESXi SanClient resources.
Snapshot creation of virtual machines related to an esx_server is invoked parallely.
Each ESXi host will be operated against in parallel.

=over 4

=item Event Parameters

Required Event Parameters:
 esx_snapshot_name
 virtual_machines
 esx_vcenter_server

Optional Event Parameters:
 esx_snapshot_quiesce
 esx_snapshot_memory

=item Examples

 # see EventMan::add_event() for API details
 $EventMan_obj->add_event(-event_name => "esx_snapshot",
                          -esx_snapshot_name => 'esx_snapshot_name');
 if ($EventMan_obj->error()) {
     ...
 }
=back


=head2 B<esx_vxdmp_rotate_io_policies>

This will take all VxDMP capable ESXi hosts and rotate the IO policies for all enclosures.
Each ESXi host will be operated against in parallel.

=over 4

=item Event Parameters

Required Event Parameters:
 esx_vxdmp_io_policies

=item Examples

 # this will rotate all vxdmp io policies for every VxDMP enabled ESXi host registered with EventMan
 # to use either minimumq or round-robin
 # see EventMan::add_event() for API details
 $EventMan_obj->add_event(-event_name => "esx_vxdmp_rotate_io_policies",
                          -esx_vxdmp_io_policies => ["minimumq", "round-robin"]
 if ($EventMan_obj->error()) {
     ...
 }

=back





=item Event Parameters

    Required Event Parameters:
    esx_lun_target_reset()

=item Examples

  # this function get a list of esx sanclient objects and will register with EventMan
  # my @sanclient_objs = $self->_get_esx_vma_sanclients();
  # it will get the LUN naa.id of each ESX hosts , reset type ( target reset or lunreset )and reset count also
  # get the reset count
   my $esx_lun_target_reset_count = $self->EventUtil::get_arg(-arg => "esx_lun_target_reset_count",
                                                  -objects => [$resource],
                                                 -default_to_env => 1);
  # get reset type
    my $esx_lun_target_reset_type = $self->EventUtil::get_arg(-arg => "esx_lun_target_reset_type",
                                                                    -objects => [$resource],
  # get lun naa ids of all NETAPP devices
    my @lun_naa_ids = $esx_host_obj->get_netapp_lun_naa_ids();
  # kick off  target or lun reset  subtests per esx host object in parallel 
    foreach my $esx_host_obj (@sanclient_objs) {
    my $esx_host_name = $esx_host_obj->hostname();
    my $runid = "reset__" . $esx_host_name;
    my $subtest = subtest(
         sub {
             $self->_esx_lun_target_reset_per_esx_host(-esx_host_obj => $esx_host_obj);
  # it will perform target reset or lunreset operation and validate the output
   $esx_host_obj->lun_target_reset(%reset_args);
  # validate the output
   sub _esx_lun_target_reset_validate {
    my ($self) = @_;
  # get the list of sanclients resources
    my @sanclient_objs = $self->_get_esx_vma_sanclients();
    if ($self->error()) {
     $self->_set_err("There are no SanClient resources available");
    return $EventUtil::FALSE;
    }
}
 
=back

=cut
