# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for generic Controller Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents;

use Tie::IxHash;
use San::SanUtil;
use NACL::STask::Node;
use NATE::BaseException qw(:try);
use NACL::Exceptions::HighAvailabilityError;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "collect_controller_stats",
                                     -event_modules => \@event_modules,
                                     -description => "This event collect controller statistics over a period of time.",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["nacl_nodes_cm"],
                                     -exec => sub { &collect_controller_stats(@_) },
                                     -arg_defaults => { controller_stats_iteration_count => 1,
                                                        controller_stats_interval_time => 30 });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    collect_controller_stats()
# Objective: Routinely collect controller statistics over specified interval
#            times.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub collect_controller_stats {
    my ($self, %args) = @_;

    my $iteration_count = $self->EventUtil::get_arg(-arg => "controller_stats_iteration_count");

    # collect stats per interval period
    for (my $i = 0; $i < $iteration_count; $i++) {
        &SanUtil::nate_log('-comment', "Collecting controller stats (iteration $i of $iteration_count)");

        # collect sysstats
        $self->SysStats::collect_sysstats_on_controllers(-type => "all");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        # collect memstats
        $self->SysStats::collect_memstats_on_controllers(-type => "all");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # collect cluster statistics
        $self->SysStats::collect_cluster_statistics(-type => "all");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        $self->EventUtil::sleep_arg(-arg => "controller_stats_interval_time",
                                    -handle_stop_messages => $EventUtil::TRUE);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    parse_node_pairs()
# Objective: Generate a hash relationship that maps Node HA Pairs (7m/Cm)
###############################################################################
sub parse_node_pairs {
    my ($self, @node_names) = @_;

    my (%primaries, %secondaries);
    my $partner_relationships = {};

    # retain insertion order
    tie(%primaries, "Tie::IxHash");
    tie(%secondaries, "Tie::IxHash");
    tie(%{$partner_relationships}, "Tie::IxHash");

    # determine which of the node_names are partners, designate one member
    # of each pair to be the master and one to be the slave
    foreach my $node_name (@node_names) {

        # if this node is already designated as a slave, then skip it
        if (exists($secondaries{$node_name})) {
            next;
        }

        my $node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            return ();
        }

        try {
            # obtain the partner name
            my $partner_name = NACL::STask::Node->get_partner(command_interface => $node_obj);

            # add this node to the primaries hash
            $primaries{$node_name}++;

            # add this partner to the secondaries hash
            $secondaries{$partner_name}++;

            # set up HA pair relationships for all the controllers in the HA pair
            $partner_relationships->{$node_name} = $partner_name;
            $partner_relationships->{$partner_name} = $node_name;

        } catch NACL::Exceptions::HighAvailabilityError with {
            # This exception gets thrown when the partner name is "-",
            # which means this node does not have an HA partner
            # issue a warning that this node doesn't have an HA partner
            &SanUtil::nate_log('-warn', "The node '$node_name' does not have an HA partner accessible");

            # add this node to the primaries list
            $primaries{$node_name}++;
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };

        if ($self->error()) {
            return ();
        }
    }

    return ([keys(%primaries)], [keys(%secondaries)], $partner_relationships);
}

###############################################################################
#                               Private Methods
###############################################################################

1;

=pod

=head1 Package Name

ControllerEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for Generic Controller based Events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::SysStats

=back

=head1 B<Event Parameters>

None

=head1 B<Event Keywords>

=head2 B<collect_controller_stats>

Routinely collects controller statistics (Sysstats/Memory Stats) on c-mode / 7-mode controllers

=over 4

=item Event Module Inheritance

default

=item Event Parameters

=item *I<-controller_stats_iteration_count>

=over 4

Description: Specifies the number of iterations to perform when collecting controller stats.

Per Resource Capable: No

Default: 1

=back

=item *I<-controller_stats_interval_time>

=over 4

Description: Specifies the interval to wait before collecting another sample of controller stats.

Per Resource Capable: No

Default: 30

=back

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode NACL::C::Node resources to
operate against.

Default: all CMode NACL::C::Node resources in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "collect_controller_stats",
                         -controller_stats_iteration_count => 10,
                         -controller_stats_interval_time => 60,
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head1 B<Public Methods>

Note: All public methods exported by this package are only visible in
Event Modules that inherit this Event Module.

=head2 B<parse_node_pairs(@args)>

=over 4

Creates a hash that maps Node HA pairs (both Cmode & 7mode)

=back

=item Inputs

=item *I<@args>

Required: An array of node names to use when generating the HA relationship hash

=item Output

($primaries, $secondaries, $relationships)
$primaries = ARRAY reference of primary nodes
$secondaries = ARRAY reference of secondary nodes
$relationships = HASH reference mapping primary nodes as keys and secondary nodes as values

=item Examples

 # parse the specified node controllers
 my ($primaries, $secondaries, $relationships) = $EventManobj->parse_node_pairs("fas3070-rtp60",
                                                                                "fas3070-rtp61",
                                                                                "fas2040-rtp50",
                                                                                "fas2040-rtp51");
 if ($EventManobj->error()) {
     ...
 }

 # The output should look like:
 # $primaries   == ["fas3070-rtp60", "fas2040-rtp50"]
 # $secondaries == ["fas3070-rtp61", "fas2040-rtp51"]
 # $relationships == { "fas3070-rtp60" => "fas3070-rtp61",
 #                     "fas2040-rtp50" => "fas2040-rtp51" }

=back

=cut
