# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for generic CMode Controller Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode;

use NACL::C::StorageFailover;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use ParamsMgr;
use strict;
use warnings;

use vars qw($Log);

$Log = log_global();

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { toggle_scsit_fault_injection_panic_cdb => undef });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    # toggle scsit fault injection on all nodes
    &toggle_scsit_fault_injection_panic_cdb($self);
    if ($self->error()) {
        $self->_set_err("Failed to toggle scsit fault injection settings on all nodes");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Event Method: toggle_auto_giveback()
# Objective: Toggle auto giveback on all clusters
# Details:   See POD documentation at the end of this file.
###############################################################################
sub toggle_auto_giveback {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $state = &ParamsMgr::get_arg(-arg => "state",
                                    -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                      -faulted => $EventUtil::FALSE,
                                                      -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Failed to get nacl nodes");
        return $EventUtil::FALSE;
    }

    # adjust the input based on the state
    $state = "true"  if ($state =~ /enable/i);
    $state = "false" if ($state =~ /disable/i);

    foreach my $node_name (@nacl_nodes) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl node object for $node_name");
            return $EventUtil::FALSE;
        }

        my $attempts_left = 5;
        my $command_failed_flag;
        my $warned_for_failure = 0;

        while ($attempts_left > 0) {
            $command_failed_flag = 0;

            # toggle auto giveback, also make sure onpanic and onreboot are enabled
            try {
                NACL::C::StorageFailover->modify(command_interface => $nacl_node_obj,
                                                 node              => $node_name,
                                                 'auto-giveback'   => $state,
                                                 'onpanic'         => "true",
                                                 'onreboot'        => "true");
            } otherwise {
                my $e = shift;

                if ($e->text() =~ /Failed to set option cf\.giveback\.auto\.enable/
                    and not $warned_for_failure) {
                    $Log->warn("Encountered ONTAP BUG 715249." .
                               "Please update the burt with relevant information " .
                               "to indicate that you hit this issue.\n".
                               "Retrying 'storage failover modify' $attempts_left " .
                               "retry attempts left");
                    $warned_for_failure = 1;
                }

                $attempts_left--;
                if ($attempts_left == 0) {
                    $self->_set_err("Exception caught: " . $e->text());
                }

                # set flag showing the command failed.
                $command_failed_flag = 1;
            };

            # if the command failed loop again, otherwise break out of the loop
            if ($command_failed_flag and $attempts_left > 0) {
                sleep 30;
                next;
            } else {
                last;
            }
        }
        if ($self->error()) {
            $self->_set_err("Could not modify storage failover settings on " .
                            "node N after 5 attempts");
            return $EventUtil::FALSE;
        }

        # toggle auto giveback after panic (only works in RR.2+ & SN.0+)
        try {
            NACL::C::StorageFailover->modify(command_interface => $nacl_node_obj,
                                             node => $node_name,
                                             'auto-giveback-after-panic' => $state);
        } otherwise {
            # ignore the error because this option doesn't exist in RR.0 & RR.1
        };
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: disable_autosupport()
# Objective: Disable autosupport on all clusers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable_autosupport {
    my ($self) = @_;

    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                      -faulted => $EventUtil::FALSE,
                                                      -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_se_err("Failed to get nacl nodes");
        return $EventUtil::FALSE;
    }

    foreach my $node_name (@nacl_nodes) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl node object for $node_name");
            return $EventUtil::FALSE;
        }

        try {
            # ignore any exceptions thrown
            NACL::C::SystemNodeAutosupport->modify(command_interface => $nacl_node_obj,
                                                   node => "$node_name",
                                                   support => "disable",
                                                   state => "disable",
                                                   'local-collection' => "false");
        } otherwise {

        };
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: toggle_scsit_fault_injection_panic_cdb()
# Objective: Toggle SCSIT inject fault cdb options on all nodes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub toggle_scsit_fault_injection_panic_cdb {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @nacl_nodes = &ParamsMgr::get_arg(-arg => "nodes",
                                         -default => [],
                                         -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return -1;
    }

    my $toggle_scsit_fi_panic_cdb = $self->EventUtil::get_arg(-arg => "toggle_scsit_fault_injection_panic_cdb",
                                                              -default => undef);

    # return if the event parameter was not specified for either enable or disable
    if (not defined($toggle_scsit_fi_panic_cdb) or
        $toggle_scsit_fi_panic_cdb !~ /enable|disable/i) {
        return $EventUtil::FALSE;
    }

    # set the value param that we will use to enable/disable the option
    my $value;
    if ($toggle_scsit_fi_panic_cdb =~ /enable/i) {
        $value = "1";
    } else {
        $value = "0";
    }

    # fetch all the nacl nodes if the user didnt pass in a list
    unless (scalar(@nacl_nodes)) {
        @nacl_nodes = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                       -faulted => $EventUtil::FALSE,
                                                       -fall_back_on_nacl_clusters => $EventUtil::TRUE);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl nodes");
            return $EventUtil::FALSE;
        }
    }

    # toggle the sysvar option on all nodes
    foreach my $node_name (@nacl_nodes) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl node object for ${node_name}");
            return $EventUtil::FALSE;
        }

        try {
            # open a connection to the systemshell to execute the command
            my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                interface => "CLI",
                                                set => "Systemshell");
            my $cmd = "sudo sysctl sysvar.scsit.debug.enable_fi_panic=$value";
            $apiset->execute_raw_command(command => $cmd);
        } otherwise {
            my $e = shift;
            $self->_set_err("Failed to ${toggle_scsit_fi_panic_cdb} " .
                            "sysvar.scsit.debug.enable_fi_panic on ${node_name}");
        };

        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

1;

=pod

=head1 Package Name

CMode.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for Generic CMode Controller based Events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

None

=back

=head1 B<Event Parameters>

=head2 I<-toggle_scsit_fault_injection_panic_cdb>

=over 4

Description: Specifies whether to 'enable', 'disable' with the scsit fault injection
option on cmode controllers. If this option is not set, then it will not modify the setting.

Per Resource Capable: No

Default: undef

=back

=head1 B<Public Methods>

Note: All public methods exported by this package are only visible in
Event Modules that inherit this Event Module.

=head2 B<toggle_auto_giveback(%args)>

Toggle auto-giveback on all CMode nacl_node_cm resources in scope.

=over 4

=item Inputs

=item *I<-state>

Required: enabled or disabled

=item Output

$EventUtil::TRUE on success, $EventUtil::FALSE and error set on failure

=item Examples

 # toggle auto giveback
 $EventManobj->toggle_auto_giveback();
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<disable_autosupport()>

Disable autosupport on all CMode nacl_node_cm resources in scope.

=over 4

=item Inputs

None

=item Output

$EventUtil::TRUE on success, $EventUtil::FALSE and error set on failure

=item Examples

 # disable autosupport
 $EventManobj->disable_autosupport();
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<toggle_scsit_fault_injection_panic_cdb(%args)>

Toggles the setting of the sysvar 'sysvar.scsit.debug.enable_fi_panic' to 1 or 0 depending
on the event parameter 'toggle_scsit_fault_injection_panic_cdb'.

=over 4

=item Inputs

=item *I<-nodes>

Optional: Specifies which nacl_nodes_cm resources to use

Default: all nacl_nodes_cm resources in scope.

=item Output

$EventUtil::TRUE on success, $EventUtil::FALSE and error set on failure

=item Event Parameter Dependencies

 toggle_scsit_fault_injection_panic_cdb

=item Examples

 # disable autosupport
 $EventManobj->toggle_scsit_fault_injection_panic_cdb(-nodes => ['node1','node2']);
 if ($EventManobj->error()) {
     ...
 }

=back

=cut
