# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for Aggr Events
## @author dalmond@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::AggrEvents;

use Data::Dumper;
use San::SanUtil;
use San::SanClient::ClientUtil;
use ParamsMgr;
use strict;
use warnings;
use NACL::CS::StorageAggregate;
use NACL::STask::Node;
use NACL::C::StorageAggregateRelocation;
use NATE::BaseException qw(:try);

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;
    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    # Aggr Move Event Modules
    $self->EventUtil::register_event(-event_name => "aggr_reloCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a CMode event that performs aggr relos across HA pairs",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { aggr_relo_aggr_list => [],
                                                        aggr_relo_aggr_count => "",
                                                        aggr_relo_command_timeout => "600",
                                                        aggr_relo_aggr_filter => "",
                                                        aggr_relo_interval_time => 600},
                                     -resource_counts => { "nacl_clusters_cm" => "1"},
                                     -exec => sub { &aggr_reloCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    aggr_reloCm()
# Objective: Perform Aggr Relo operation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub aggr_reloCm {
    my ($self, %args) = @_;

    # get the cluster to operate against
    my ($cluster) = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        $self->_set_err("Unable to obtain controller cluster list");
        return $EventUtil::FALSE;
    }

    # obtain the name of nodes in this cluster (that are not faulted)
    my @nodes = $self->EventUtil::get_nacl_node_objs(-clusters => [$cluster],
                                                     -faulted => $EventUtil::FALSE,
                                                     -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain a node from cluster $cluster");
        return $EventUtil::FALSE;
    }

    # get the aggr count to relo or
    # get the aggr names or regex to relo
    # and get the aggr filter
    my $aggr_list = $self->EventUtil::get_arg(-arg => "aggr_relo_aggr_list");
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $EventUtil::FALSE;
    }

    my $aggr_count = $self->EventUtil::get_arg(-arg => "aggr_relo_aggr_count");
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $EventUtil::FALSE;
    }

    my $aggr_filter = $self->EventUtil::get_arg(-arg => "aggr_relo_aggr_filter");
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $EventUtil::FALSE;
    }

    my $aggr_relo_command_timeout = $self->EventUtil::get_arg(-arg => "aggr_relo_command_timeout");
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $EventUtil::FALSE;
    }

    my @aggr_objs;
    # first grab all the aggr object components
    foreach my $node (@nodes) {
        try {
            push(@aggr_objs, NACL::C::StorageAggregate->find(command_interface => $node,
                                                             filter=>{ "ha-policy" => "sfo", nodes => [ $node->node() ] }));
        } otherwise {
            my $e = shift;

            $self->_set_err("Exception caught while fetching aggregate objects on node $node." . $e->text());
        };
        if ($self->error()) {
            # An exception is thrown if no aggrs match your filter
            # In this case, we will search on the next node.
            unless ($self->errmsg() =~ /No matching/i) {
                return $EventUtil::FALSE;
            }
            $self->_clear_err();
        }
    }

    # if an aggr_filter/aggr_list/aggr_count are not default values, filter the list of aggr objects now
    # aggr_list will take precedence, then aggr_filter, and aggr_count
    # if aggr_count and aggr_list are default values, move
    # all the sfo aggrs
    my @filtered_aggr_objects = ();
    if ((not ref($aggr_list) eq "ARRAY") or not scalar(@{$aggr_list})) {
        if ($aggr_filter ne "") {
            @filtered_aggr_objects =  grep {$_->aggregate() =~ $aggr_filter}  @aggr_objs;
        } else {
            @filtered_aggr_objects = @aggr_objs;
        }

        # checking to see if aggr count is defined, if so
        # check to see if its a number and if the number of
        # acceptable aggrs is at least equal to the number
        # of aggrs to relocate
        if ($aggr_count =~ /\d+/) {
            if (scalar(@filtered_aggr_objects) < $aggr_count){
                $self->_set_err("The number of aggrs in the cluster that meet the filter " .
                            "is less than the amount of aggrs to relocate.");
                return $EventUtil::FALSE;
            } elsif (scalar(@filtered_aggr_objects) > $aggr_count){
                # trim filtered_aggr_objects to size
                while (scalar (@filtered_aggr_objects) > $aggr_count){
                    pop(@filtered_aggr_objects);
                }
            }
        }
    } else {
        # if a list is defined parse it and verify all the aggrs exist in the cluster

        foreach my $aggr_name (@{$aggr_list}) {
            my $found_aggr = 0;
            foreach my $aggr_obj (@aggr_objs) {
                if ($aggr_name eq $aggr_obj->aggregate()){
                    push(@filtered_aggr_objects, $aggr_obj);
                    $found_aggr = 1;
                    last;
                }
            }
            unless ($found_aggr) {
                $self->_set_err("The aggr $aggr_name does not exist in the cluster but was requested".
                                " to be relocated in the param aggr_relo_aggr_list.\n");
                return $EventUtil::FALSE;
            }
        }
    }

    # going node by node, relocate all the aggrs on that node,
    # sleep, validdate pathing, relocate all the moved aggrs back,
    # sleep, validate pathing and move to the next node
    # if we want to do this in parallel in the future, we can
    # make a subtest for each node then have them call back to
    # here for pathvalidation and sleeping.

    foreach my $node (@nodes) {
        my @relocated_aggrs;
        foreach my $aggr_object (@filtered_aggr_objects) {
            if ($aggr_object->get_local_node_name() ne $node->node()) {
                next;
            }

            # collect stats
            $self->_collect_stats();
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
            # relo aggr
            $self->_relo_aggr(-aggr_object               => $aggr_object,
                              -aggr_relo_command_timeout => $aggr_relo_command_timeout);
            # check for error
            if ($self->error()) {
                $self->_set_err("Unable to relocate " . $aggr_object->aggregate());
                return $EventUtil::FALSE;
            }
            push(@relocated_aggrs, $aggr_object);
        }

        $self->_sleep_and_validate_paths();
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # relo aggr back
        foreach my $aggr_object (@relocated_aggrs){
            $self->_relo_aggr(-aggr_object => $aggr_object,
                              -aggr_relo_command_timeout => $aggr_relo_command_timeout);
            # check for error
            if ($self->error()) {
                $self->_set_err("Unable to restore relocated aggr  " . $aggr_object->aggregate());
                return $EventUtil::FALSE;
            }
            #continue to next aggr
        }

        # collect stats
        $self->_collect_stats();
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        #continue to next node
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################
###############################################################################
# Event Method:    _collect_stats()
# Objective: collect system stats
#
###############################################################################
sub _collect_stats {
    my ($self) = @_;
    # collect node stats
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster memeory statistics
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Event Method:    _sleep_and_validate_paths()
# Objective: sleep the required time and then validate pathing
#
###############################################################################
sub _sleep_and_validate_paths {
    my ($self) = @_;

    my $aggr_relo_interval_time = $self->EventUtil::sleep_arg(-arg => "aggr_relo_interval_time");

    # validate controller after the aggr relos relo
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension      => "aggr_relo",
                                                                    -last_timeout      => $aggr_relo_interval_time,
                                                                    -last_timeout_name => "aggr_relo_interval_time");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after the aggr relo
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "aggr_relo",
                                                         -last_timeout      => $aggr_relo_interval_time,
                                                         -last_timeout_name => "aggr_relo_interval_time");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate paths after all aggr have completed their first round of relos
    $self->PathValidation::validate_paths_after_giveback(-id_extension => "aggr_relo",
                                                         -last_timeout => $aggr_relo_interval_time,
                                                         -last_timeout_name => "aggr_relo_interval_time");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}
##############################################################################
# Event Method:    _relo_aggr()
# Objective: relocate an aggr from a node to its partner node while
#            verbosely reporting success or errors
###############################################################################
sub _relo_aggr {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $aggr_object = &ParamsMgr::get_arg(-arg => "aggr_object",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $aggr_relo_command_timeout = &ParamsMgr::get_arg(-arg => "aggr_relo_command_timeout",
                                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
    my $node_name = $aggr_object->get_local_node_name();
    my $ci  = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
    my $aggr_name = $aggr_object->aggregate();
    my $destination;
    try {
        $destination = NACL::STask::Node->get_partner(command_interface => $ci);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while fetching partner node ." . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log("-comment", "Relocating aggr $aggr_name from $node_name to $destination.\n");

    my $aggr_relo;
    try {
        $aggr_relo = NACL::C::StorageAggregateRelocation->new(command_interface => $ci,
                                                              'node' => $node_name);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while creating storage aggr relo for node $node_name." . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    try {
        $aggr_relo->start(command_interface => $ci,
                          'node' => $node_name,
                          'aggregate-list' => [ $aggr_name ],
                          'destination' => $destination);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while starting aggr relo for aggr $aggr_name ." . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    my $relo_end_time = $aggr_relo_command_timeout + time();
    while (time() < $relo_end_time) {
        &SanUtil::nate_sleep(1);
        my $aggr_relo_state = $aggr_relo->state();
        my $status = $aggr_relo_state->relocation_status();
        if ($status eq "Done") {
            &SanUtil::nate_log("-comment", "Completed relocating aggr $aggr_name from $node_name to $destination.\n");
            return $EventUtil::TRUE;
        } elsif ($status =~ /Failed/i) {
            &SanUtil::nate_log("-comment", "Failed relocating aggr $aggr_name from $node_name to $destination.\n");
            return $EventUtil::FALSED;
        }
    }

    &SanUtil::nate_log("-comment", "Timed out relocating aggr $aggr_name from $node_name to $destination.\n");
    return $EventUtil::FALSE;
}

=pod

=head1 Package Name

AggrEvents.pm

=head1 Author

Dan Almond

=head1 Description

Event Module for CMode Aggr Events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::AggrEvents

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::PathValidation

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=over 4

None

=back

=head1 B<Event Keywords>

=head2 B<aggr_reloCm>

Perform Aggr Relocations on a single CMode Cluster using NACL::StorageAggregateRelocation.
This event adds some additional logic around the library.

=over 4

=item Resources

=item *I<-nacl_clusters_cm>

Required: An array reference of nacl_clusters_cm resources to operate against. Only one
cluster is supported for this event.

Default: all nacl_clusters_cm resources in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation.

Default: all sanclients resources in the Event scope will be used

=item Event Parameters

=item *I<-aggr_relo_aggr_list>

Optional:  An array reference of aagr names to perform Aggr relocations against. This
parameter and aggr_relo_aggr_filter are mutually exclusive.

Default: ""

=item *I<-aggr_relo_aggr_filter>

Optional: A perl regular expression matching of aggrs to perform Aggr relocations against. This
parameter and aggr_relo_aggr_list are mutually exclusive.

Default: qr/.+/

=item *I<-aggr_relo_aggr_count>

Optional: If aggr_relo_aggr_filter is used, this will parameter allows the user to limit how
many aggrs to operate against. So if 10 aggrs match the filter, but count is only set to 2, then
the first two aggrs that matched will be operated against.

=item *I<-aggr_relo_interval_time>

Optional: Specifies the time (in seconds) to rest in between aggr relos. This will also
control how long to wait before running validation routines.

Default: 600

=item *I<-aggr_relo_command_timeout>

Optional: The number of seconds to wait for an aggr relo to complete. If it doesn't
complete within this time, the event will fail.

Default: 600

=item Examples

 # Perform aggr relocates on ten aggrs that match qr/testaggr\d+/
 # on 'mycluster'. Each aggr will be relocated to the partner node and back once.
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "aggr_reloCm",
                         -nacl_clusters_cm => ["mycluster"],
                         -aggr_relo_aggr_filter => qr/testaggr\d+/,
                         -aggr_relo_aggr_count => 10);
 if ($EventManobj->error()) {
     ...
 }

 # Perform aggr relocates on aggr1, aggr2, and aggr3
 # on 'mycluster'. Each aggr will be relocated to the partner node and back once.
 # aggr relo interval time is set to wait 500 seconds after each aggr relo operation.
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "aggr_reloCm",
                         -nacl_clusters_cm => ["mycluster"],
                         -aggr_relo_aggr_list => ['aggr1','aggr2','aggr3'],
                         -aggr_relo_interval_time => 500);
 if ($EventManobj->error()) {
     ...
 }

=back

=over 4

=item Resources

=item *I<-nacl_clusters_cm>

Required: an Array reference of controller clusters to use. No more than ONE cluster can be specified.

Resource Count: 1

=back

=cut

1;
