# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for CMode Cluster Interconnect events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::ClusterICEvents;

use Data::Dumper;
use San::SanUtil;
use Tharn;
use strict;
use warnings;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules::SwitchEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "cluster_ic_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that blocks Cluster Interconnect switch ports, " .
                                                     "one port at a time per node",
                                     -require_at_least_one_resource_types => ["nacl_nodes_cm", "nacl_clusters_cm"],
                                     -required_resource_types => ["sanswitches"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanswitches => "1+" },
                                     -exec => sub { &cluster_ic_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "cluster_ic_switch_disruption",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that blocks all Cluster Interconnect switch ports " .
                                                     "on a switch at a time. This simulates a switch reboot or firmware upgrade.",
                                     -required_resource_types => ["sanswitches", "nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanswitches => "2+",
                                                           nacl_nodes_cm => "2+,even" },
                                     -arg_hardcoded => { switch_port_block_algorithm => "parallel",
                                                         switch_port_block_cmode_target_type => "node",
                                                         switch_port_block_sort_switch_names => $EventUtil::TRUE,
                                                         switch_port_block_conserve_switch_connections => $EventUtil::TRUE,
                                                         switch_port_block_sync_parallel_blocks => $EventUtil::TRUE,
                                                         switch_port_down_time =>
                                                             sub {
                                                                 # tie this event parameter to the value specified by cluster_ic_switch_port_down_time
                                                                 my ($self) = @_;
                                                                 return $self->EventUtil::get_arg(-arg => "cluster_ic_switch_port_down_time",
                                                                                                  -default => 300);
                                                             },
                                                         switch_port_up_time =>
                                                             sub {
                                                                 # tie this event parameter to the value specified by cluster_ic_switch_port_up_time
                                                                 my ($self) = @_;
                                                                 return $self->EventUtil::get_arg(-arg => "cluster_ic_switch_port_up_time",
                                                                                                  -default => 300);
                                                             }
                                                         },
                                     -arg_defaults => { cluster_ic_switch_port_down_time => 300,
                                                        cluster_ic_switch_port_up_time   => 300 },
                                     -exec => sub { &cluster_ic_switch_disruption(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    @event_modules = (__PACKAGE__,
                      "EventMan::EventModules::ControllerEvents::CMode::SFOEvents",
                      "EventMan::EventModules::ControllerEvents::CMode",
                      "EventMan::EventModules::ControllerEvents",
                      "EventMan::EventModules::SwitchEvents",
                      "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "cluster_ic_node_out_of_quorum",
                                     -event_modules => \@event_modules,
                                     -enabled => $EventUtil::TRUE,
                                     -description => "This is a Cmode event that blocks all Cluster Interconnect switch " .
                                                     "ports, per node at a time, to trigger an Out of Quorum (OOQ) event on the cluster.",
                                     -require_at_least_one_resource_types => ["nacl_nodes_cm", "nacl_clusters_cm"],
                                     -required_resource_types => ["sanswitches"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanswitches => "1+",
                                                           nacl_nodes_cm => "2+,even" },
                                     -exec => sub { &cluster_ic_node_out_of_quorum(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { cluster_ic_switch_port_down_time => 90,
                                                           cluster_ic_switch_port_up_time   => 90,
                                                           controller_down_time             => 600 }
                                                     );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "cluster_ic_switch_port_down_time",
                                           -package => __PACKAGE__,
                                           -default => "30",
                                           -format => "integer (seconds)",
                                           -example => "-cluster_ic_switch_port_down_time => 60",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how long to keep a cluster " .
                                                           "interconnect port block down after " .
                                                           "the port is disabled on the switch");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "cluster_ic_switch_port_up_time",
                                           -package => __PACKAGE__,
                                           -default => "30",
                                           -format => "integer (seconds)",
                                           -example => "-cluster_ic_switch_port_up_time => 60",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how long to wait after enabling a " .
                                                           "cluster interconnect port on the switch");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    cluster_ic_port_blocks()
# Objective: Perform CMode Cluster Interconnect port blocks on Cluster switches
#            Disables one port per node at a time.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cluster_ic_port_blocks {
    my ($self, %args) = @_;

    my (@nacl_nodes) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                        -fall_back_on_controller_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Could not obtain nacl_nodes_cm resources.");
        return $EventUtil::FALSE;
    }

    # loop through each cluster port on sanfilesCm and disable the cluster ports

    # or loop through all the nodes that belong to the controllerclusters
    # and block the cluster ports

    my (@sanswitches) = $self->EventUtil::get_sanswitches();

    # extract the switch_info resource
    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");
    if ($self->error()) {
        $self->_set_err("Could not switch info resource");
        return $EventUtil::FALSE;
    }

    my $performed_blocks = 0;

    # for each cmode controller, disable the cluster ports on the switches
    foreach my $node_name (@nacl_nodes) {
        foreach my $mac (keys %{$switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}}) {
            if (exists($switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{switch_name})) {
                my $switch_name = $switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{switch_name};
                my $port_numbers = $switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{port_number};

                # check to see if this switch is in the sanswitch list that we are operating against.
                next if (not grep { /^$switch_name$/ }  @sanswitches);
                $performed_blocks = 1;

                foreach my $port_number (@{$port_numbers}) {

                    # now we have the switch and port this cluster port is plugged into, perform the port disable/enable
                    # disable it
                    $self->disable_switch_port(-switch_name => $switch_name,
                                               -port_number => $port_number,
                                               -resource_name => $node_name,
                                               -resource_type => "nacl_nodes_cm");
                    if ($self->error()) {
                        $self->_set_err("Failed to disable cluster interconnect port $port_number on switch $switch_name for controller $node_name");
                        return $EventUtil::FALSE;
                    }

                    # sleep switch_port_down_time before re-enabling them
                    $self->EventUtil::sleep_arg(-arg => "cluster_ic_switch_port_down_time");
                    if ($self->error()) {
                        $self->_set_err("Unable to sleep cluster_ic_switch_port_down_time");
                        return $EventUtil::FALSE;
                    }

                    # enable it
                    $self->enable_switch_port(-switch_name => $switch_name,
                                              -port_number => $port_number,
                                              -resource_name => $node_name,
                                              -resource_type => "nacl_nodes_cm");
                    if ($self->error()) {
                        $self->_set_err("Failed to enable cluster interconnect port $port_number on switch $switch_name for controller $node_name");
                        return $EventUtil::FALSE;
                    }

                    # sleep switch_port_up_time before re-enabling them
                    $self->EventUtil::sleep_arg(-arg => "cluster_ic_switch_port_up_time");
                    if ($self->error()) {
                        $self->_set_err("Unable to sleep cluster_ic_switch_port_up_time");
                        return $EventUtil::FALSE;
                    }
                }
            }
        }
    }

    if ($performed_blocks == 0) {
        $self->_set_err("No port blocks were performed due to lack of switch resources");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    cluster_ic_switch_disruption()
# Objective: Disable all ports on a single Cluster Interconnect switch at a
#            time. This will try to simulate the case where a cluster ic switch
#            loses all connectivity (reboot or a firmware upgrade simulation)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cluster_ic_switch_disruption {
    my ($self, %args) = @_;

    # perform port blocks for all cluster port mac ports on all switches
    # in the environment
    $self->switch_port_block_abstracted(-port_block_types => ["cluster_port_mac"]);
    if ($self->error()) {
        $self->_set_err("Cluster Interconnect Switch disruption port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    cluster_ic_node_out_of_quorum()
# Objective: Perform CMode cluster interconnect port blocks on cluster switches.
#            Disable all cluster ports per node simultaneously to simulate
#            a node being isloated from the cluster. As a result of injecting
#            this fault, the isolated node will either be 1) (in a 4+ node cluster)
#            taken down by its HA partner via an SFO, or 2) (in a two node
#            cluster) the isloated node or the ha parter of said node, will issue
#            the SFO.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cluster_ic_node_out_of_quorum {
    my ($self, %args) = @_;

    my (@nacl_nodes) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                        -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Could not obtain nacl_nodes_cm resources.");
        return $EventUtil::FALSE;
    }

    # split up all the nodes into HA pairs
    my (undef, undef, $pairs) = $self->parse_node_pairs(@nacl_nodes);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # for each cmode controller, disable all cluster ports on the switches at the same time
    # to segment the node from the cluster
    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");
    if ($self->error()) {
        $self->_set_err("Could not switch info resource");
        return $EventUtil::FALSE;
    }

    my @sanswitches = $self->EventUtil::get_sanswitches();
    my $performed_blocks = 0;
    my %processed_nodes;
    
    # keep a history of which switches have been reconnected
    my $switch_reconnect_history = {};

    foreach my $node_name (@nacl_nodes) {

        my $ha_partner = $pairs->{$node_name};

        # keep track of which HA pairs have been operated against for this event
        $processed_nodes{$node_name}++;
        $processed_nodes{$ha_partner}++;

        # if not requested to do symmetrical_events, then if this HA pair has already been processed,
        # skip it
        if (not $self->EventUtil::get_arg(-arg => "symmetrical_events") and
            exists($processed_nodes{$node_name})) {
            next;
        }

        # toggle auto giveback on this node and it's HA partner
        if ($self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") eq $EventUtil::TRUE) {
            $self->toggle_auto_giveback(-state => "enable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            $self->toggle_auto_giveback(-state => "disable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }

        # collect stats on all the nodes
        $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # collect cluster statistics
        $self->SysStats::collect_cluster_statistics(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # log to the syslogs what we are about to do
        my $sfo_takeover_start_sentinel = "takeover start " . time;

        $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Inducing Cluster Node Out of Quorum takeover against HA pair '$node_name' '$ha_partner'",
                                                       $sfo_takeover_start_sentinel],
                                          -async => $EventUtil::TRUE);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # store this takeover start sentinel on both of the node resources as an attribute
        $self->ResMgr::set_resource_attribute(-name => $node_name,
                                              -type => "nacl_nodes_cm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $ha_partner,
                                              -type => "nacl_nodes_cm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # disable all cluster ports at the same time for this controller
        foreach my $mac (keys %{$switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}}) {
            if (exists($switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{switch_name})) {
                my $switch_name = $switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{switch_name};
                my $port_numbers = $switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{port_number};

                # check to see if this switch is in the sanswitch list that we are operating against.
                next if not grep(/^$switch_name$/, @sanswitches);
                $performed_blocks = 1;

                if ((not (defined $switch_reconnect_history->{$switch_name})) or 
                    $switch_reconnect_history->{$switch_name} ne  "reconnected") {
                    $self->EventUtil::reconnect_disconnected_sanswitches(-switch_names => [$switch_name]);
                    if ($self->error()) {
                        $self->_set_err("Unable to reconnect to $switch_name");
                        return $EventUtil::FALSE;
                    }
                    $switch_reconnect_history->{$switch_name} = "reconnected";
                }

                foreach my $port_number (@{$port_numbers}) {

                    # now we have the switch and port this cluster port is plugged into, perform the port disable/enable
                    # disable it
                    $self->disable_switch_port(-switch_name => $switch_name,
                                               -port_number => $port_number,
                                               -resource_name => $node_name,
                                               -resource_type => "nacl_nodes_cm");
                    if ($self->error()) {
                        $self->_set_err("Failed to disable cluster interconnect port $port_number on switch $switch_name for controller $node_name");
                        return $EventUtil::FALSE;
                    }
                }
            }
        }

        # if we havnt done any blocks then skip to the next controller
        if ($performed_blocks == 0) {
            next;
        }

        my $event_obj = $self->EventUtil::get_current_event_obj();
        my $runid_prefix = $event_obj->increment_subtest_runid_counter();

        # perform SFO detection on $node_name and its HA partner since
        # the node that gets takenover is not deterministic in a two node cluster,
        # this is determined by the actual system id's of the nodes themselves.
        # where the lower system id node will take epsilon.
        # in a 4+ node cluster, the isolated node will always be the node
        # that gets taken over regardless of system id order.
        my $subtest1 = subtest(sub {
                                   $self->_spawn_sfo_detection(-node_name => $node_name,
                                                               -partner_name => $ha_partner);
                                   if ($self->error()) {
                                       $self->_clear_err();
                                       $self->SubtestComm::subtest_return_data(-data => { node => $node_name,
                                                                                          result => "failed" });
                                   } else {
                                       $self->SubtestComm::subtest_return_data(-data => { node => $node_name,
                                                                                          result => "completed" });
                                   }

                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => "${runid_prefix}_sfo_detection_on_${node_name}", -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        my $subtest2 = subtest(sub {
                                   $self->_spawn_sfo_detection(-node_name => $ha_partner,
                                                               -partner_name => $node_name);
                                   if ($self->error()) {
                                       $self->_clear_err();
                                       $self->SubtestComm::subtest_return_data(-data => { node => $ha_partner,
                                                                                          result => "failed" });
                                   } else {
                                       $self->SubtestComm::subtest_return_data(-data => { node => $ha_partner,
                                                                                          result => "completed" });
                                   }

                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => "${runid_prefix}_sfo_detection_on_${ha_partner}", -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        # reap the two sfo detection subtests, when one passes, the other will be reaped.
        my $return_data = $self->wait_for_subtests(-subtests => [$subtest1, $subtest2],
                                                   -dependence_type => "stop_after_one_completes",
                                                   -propagate_wait => 0,
                                                   -propagate_events => 1,
                                                   -enable_sync => 0,
                                                   -check_cmds_from_parent => 0,
                                                   -forward_cmds_from_parent => 0,
                                                   -async => 0);
        if ($self->error()) {
            $self->_set_err("SFO Detection subtests have failed");
            return $EventUtil::FALSE;
        }

        my $hash;
        $hash->{completed} = 0;
        my $take_over_node;
        my $taken_over_node;

        # check to see which node performed the takeover
        foreach my $message_obj (@{$return_data}) {
            my $msg_node_name = $message_obj->{data}->{node};
            my $result = $message_obj->{data}->{result};
            $hash->{$result}++;
            if ($result eq "completed") {
                $take_over_node = $msg_node_name;
            } else {
                $taken_over_node = $msg_node_name;
            }
        }
        # it's possible that we only received one return data message
        if (defined($take_over_node)) {
            if ($take_over_node eq $node_name) {
                $taken_over_node = $ha_partner;
            } else {
                $taken_over_node = $node_name;
            }
        }

        if ($hash->{"completed"} != 1) {
            # both failed, this is an error
            $self->_set_err("Expected Out of Quorum SFO did not occur between $node_name and $ha_partner " .
                            "after disabling cluster interconnect ports.");
            return $EventUtil::FALSE;
        }

        # at this point, $take_over_node should contain the node that performed the SFO.
        # mark that $taken_over_node is now in a faulted state
        $self->ResMgr::set_faulted_resource(-name => $taken_over_node,
                                            -type => "nacl_nodes_cm",
                                            -propagate => $EventUtil::TRUE);

        # sleep controller down time
        my $controller_down_time = $self->EventUtil::sleep_arg(-arg => "controller_down_time_after_takeover");

        # synchronize before running path validation
        $self->EventUtil::synchronize();

        $self->PathValidation::validate_paths_after_takeover(-id_extension => "sfo_takeover_on_${take_over_node}",
                                                             -path_validation_rules => {validate_takeover_giveback => 1},
                                                             -last_timeout => $controller_down_time,
                                                             -last_timeout_name => "controller_down_time_after_takeover");
        if ($self->error()) {
            $self->_set_err("Failed path validation");
            return $EventUtil::FALSE;
        }

        # log to all resources that the takeover completed
        my $sfo_takeover_end_sentinel = "takeover end " . time;
        $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Cluster Out of Quorum induced takeover against node '$taken_over_node' has completed",
                                                       $sfo_takeover_end_sentinel],
                                          -async => $EventUtil::TRUE);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # store this takeover end sentinel on both of the node resources as an attribute
        $self->ResMgr::set_resource_attribute(-name => $take_over_node,
                                              -type => "nacl_nodes_cm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                              -type => "nacl_nodes_cm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # reenable all cluster ports for this controller.
        foreach my $mac (keys %{$switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}}) {
            if (exists($switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{switch_name})) {
                my $switch_name = $switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{switch_name};
                my $port_numbers = $switch_info->{obj}->{nacl_nodes_cm}->{$node_name}->{cluster_port_mac}->{$mac}->{port_number};

                # check to see if this switch is in the sanswitch list that we are operating against.
                next if not grep(/^$switch_name$/, @sanswitches);

                foreach my $port_number (@{$port_numbers}) {
                    # enable this port
                    $self->enable_switch_port(-switch_name => $switch_name,
                                              -port_number => $port_number,
                                              -resource_name => $node_name,
                                              -resource_type => "nacl_nodes_cm");
                    if ($self->error()) {
                        $self->_set_err("Failed to enable cluster interconnect port $port_number on switch $switch_name for controller $node_name");
                        return $EventUtil::FALSE;
                    }
                }
            }
        }

        $self->EventUtil::sleep_arg(-arg => "cluster_ic_switch_port_up_time");
        if ($self->error()) {
            $self->_set_err("Unable to sleep cluster_ic_switch_port_up_time");
            return $EventUtil::FALSE;
        }

        # issue a giveback on the node that issued the SFO takeover
        $self->sfo_giveback_generic(-giveback_node => $take_over_node,
                                    -taken_over_node => $taken_over_node,
                                    -forcegive => $EventUtil::FALSE,
                                    -use_nacl_apis => $EventUtil::TRUE);
        if ($self->error()) {
            $self->_set_err("giveback failed");
            return $EventUtil::FALSE;
        }

        # sleep controller down time
        $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

        # synchronize before running path validation
        $self->EventUtil::synchronize();

        # need to perform giveback since an sfo should have occurred
        $self->PathValidation::validate_paths_after_giveback(-id_extension => "sfo_giveback_on_${take_over_node}",
                                                             -path_validation_rules => {validate_takeover_giveback => 1},
                                                             -last_timeout => $controller_down_time,
                                                             -last_timeout_name => "controller_down_time_after_giveback");
        if ($self->error()) {
            $self->_set_err("Failed validation paths");
            return $EventUtil::FALSE;
        }
    }

    if ($performed_blocks == 0) {
        $self->_set_err("No cluster interconnect port blocks were performed due " .
                        "to lack of cluster interconnect switch resources. Did you forget to include them?");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

sub _spawn_sfo_detection {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);
    my $partner_name = &ParamsMgr::get_arg(-arg => "partner_name",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # check if a takeover is occurring, we need to set an alarm signal
    # because we do not know for certain which node will perform the takeover
    my $old_sig_alrm = $SIG{ALRM};
    $SIG{ALRM} = sub { die "timeout" };

    eval {
        # set the alarm to be slightly longer than the max timeout
        alarm($self->EventUtil::get_arg(-arg => "sfo_max_timeout") + 30);
        $self->detect_sfo_takeover(-sfo_timeout     => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                   -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                   -take_over_node  => $node_name,
                                   -taken_over_node => $partner_name);
        alarm(0);
    };

    # restore the original alarm handler
    $SIG{ALRM} = $old_sig_alrm;

    # check if the alarm was tripped
    if ($@ =~ /timeout/i) {
        # The sfo detection failed (probably because this node was taken over)
        $self->_set_err("The node $node_name was taken over by $partner_name");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

=pod

=head1 Package Name

ClusterICEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for CMode Cluster Interconnect Events.

=head1 B<Event Keywords>

=head2 B<cluster_ic_port_blocks>

Performs port blocks on cluster interconnect switches. This will only block one
cluster port on each node at a time.

=over 4

=item Event Parameters

=item *I<-nacl_nodes_cm> or I<-nacl_clusters_cm>

Required: An array reference of a CMode nodes resources or CMode
clusters to operate against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # perform a single port blocks on the cluster ic ports for these two cmode
 # controllers. Use the two switches specified when blocking.
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "cluster_ic_single_port_blocks",
                         -nacl_nodes_Cm => ["fas3070-rtp40", "fas3070-rtp41"],
                         -sanswitches => ["cisco5020-rtp60", "cisco5020-rtp61"],
                         -cluster_ic_switch_port_down_time => 50,
                         -cluster_ic_switch_port_up_time => 20);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<cluster_ic_switch_disruption>

Performs port blocks on cluster interconnect switches in such a way that each switch
has all its ports blocked at a time. This will simulate the scenario where a switch goes
down in a disruptive manner, such as a switch reboot/power cycle/ or a firmware upgrade
procedure. After all ports are blocked and the duration of cluster_ic_switch_port_down_time
has been waited, all the ports will be re-enabled. It will then move on to the
next cluster interconnect switch resource that is available.

=over 4

=item Resources

=item *I<-nacl_nodes_cm> or I<-nacl_clusters_cm>

Required: An array reference of a CMode nodes resources or CMode
clusters to operate against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.
There must be at least 1 cluster interconnect switch resources specified in this
list

=item Event Parameters

=item *I<-cluster_ic_switch_port_down_time>

How long to wait after disabling all the ports
Default: 300 seconds

=item *I<-cluster_ic_switch_port_up_time>

How long to wait after enabling all the ports
Default: 300 seconds

=item Examples

 # block all cluster IC ports on cisco5020-rtp60 for both nodes, wait 400 seconds
 # reenable the ports, wait 400 seconds, then perform the same thing for cisco5020-rtp61
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "cluster_ic_switch_disruption",
                         -nacl_nodes_Cm => ["fas3070-rtp40", "fas3070-rtp41"],
                         -sanswitches => ["cisco5020-rtp60", "cisco5020-rtp61"],
                         -cluster_ic_switch_port_down_time => 400,
                         -cluster_ic_switch_port_up_time => 400);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<cluster_ic_node_out_of_quorum>

Performs port blocks on cluster interconnect switches. This will block all
cluster ports of a single node at a time. One Node at a time. It is expected
that after a node is cut off from the cluster an SFO will occur on either
the isolated node or against the isolated node by its HA partner.

=over 4

=item Event Parameters

=item *I<-nacl_nodes_cm> or <-nacl_clusters_cm>

Required: An array reference of a CMode Node resources or CMode
clusters to operate against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # perform port blocks on cluster ic ports for the entire cluster
 # named 'mycmodecluster' to isolate all nodes from the cluster, one node at a time.
 # Use the two sanswitch resources specified when blocking.
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "cluster_ic_node_out_of_quorum",
                         -nacl_clusters_cm => ["mycmodecluster"],
                         -sanswitches => ["cisco5020-rtp60", "cisco5020-rtp61"],
                         -cluster_ic_switch_port_down_time => 30,
                         -cluster_ic_switch_port_up_time => 40);
 if ($EventManobj->error()) {
     ...
 }

=back

=cut

1;
