# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for CMode Disruptive Node Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::DisruptiveNodeEvents;

use Tharn;
use San::SanUtil;
use ParamsMgr;
use Data::Dumper;
use strict;
use warnings;
use NATE::BaseException qw(:try);
use NACL::Transit;
use NACL::C::StorageFailover;
use NACL::STask::Node;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "disruptive_node_rebootCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that reboots a single node (with no SFO)" .
                                                     " and waits for it to come back up",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &disruptive_node_rebootCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "disruptive_node_haltCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that halts a single node (with no SFO)" .
                                                     " and waits for it to come back up",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &disruptive_node_haltCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "disruptive_node_panicCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that panics a single node (with no SFO)" .
                                                     " and waits for it to come back up",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &disruptive_node_panicCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "disruptive_cluster_rebootCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that reboots the entire cluster (no SFOs)" .
                                                     " and waits for it to come back up",
                                     -required_resource_types => ["nacl_clusters_cm", "nacl_nodes_cm"],
                                     -exec => sub { &disruptive_cluster_rebootCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { -symmetrical_events => 1,
                                                           -disruptive_control_node => "",
                                                           -disruptive_node_event_post_wait_time => 600,
                                                           -disruptive_node_action_sleep_time => 10 }
                                                     );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    disruptive_node_rebootCm()
# Objective: Reboot single C-mode node dispruptively (no SFO)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disruptive_node_rebootCm {
    my ($self, %args) = @_;

    $self->disruptive_action_abstracted(-action => "reboot");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    disruptive_node_panicCm()
# Objective: Panic a single C-mode node dispruptively (no SFO)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disruptive_node_panicCm {
    my ($self, %args) = @_;

    $self->disruptive_action_abstracted(-action => "panic");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    disruptive_node_haltCm()
# Objective: Halt a single C-mode node dispruptively (no SFO)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disruptive_node_haltCm {
    my ($self, %args) = @_;

    $self->disruptive_action_abstracted(-action => "halt");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    disruptive_cluster_rebootCm()
# Objective: Reboot the entire cluster at the same time.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disruptive_cluster_rebootCm {
    my ($self, %args) = @_;

    my @clusters = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        $self->_set_err("Failed to obtain clusters for cluster reboot");
        return $EventUtil::FALSE;
    }

    foreach my $cluster_name (@clusters) {
        # get all the nacl nodes in this cluster
        my @nacl_nodes = $self->EventUtil::get_nacl_nodes(type => 'cmode',
                                                          clusters => [$cluster_name]);

        # reboot all the nodes in this cluster
        $self->simultaneous_node_reboot(nacl_nodes => \@nacl_nodes);
        if ($self->error()) {
            $self->_set_err("Failed to reboot cluster $cluster_name");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    disruptive_action_abstracted()
# Objective: Issue disruptive actions across the node list
#            one at a time
###############################################################################
sub disruptive_action_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $action = &ParamsMgr::get_arg(-arg => "action",
                                     -required => 1,
                                     -format => qr/^(reboot|panic|halt)$/i);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get unfaulted nodes
    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                      -faulted => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@nacl_nodes);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # initialize the previous_node to be the last node in the list
    my $previous_node = $nacl_nodes[-1];
    my %processed_pairs;

    foreach my $node_name (@nacl_nodes) {
        # unless a control node was specified, then use the previous node
        my $control_node = $self->EventUtil::get_arg(-arg => "disruptive_control_node");
        my $has_ha_partner = 1;
        if ($control_node eq "") {
            # set the control node to be the previous node one we worked on
            $control_node = $previous_node;
        }

        # if this node has an HA partner, use it as the control node regardless of any
        # override
        if (exists($pairs->{$node_name}) and defined($pairs->{$node_name})) {
            $control_node = $pairs->{$node_name};
            $has_ha_partner = 1;

            # if this HA pair has already been processed and symmetrical_events has been disabled,
            # then skip this node
            if (exists($processed_pairs{$node_name}) and
                $self->EventUtil::get_arg(-arg => "symmetrical_events") == 0) {
                next;
            } else {
                # add the ha partner to the processed pair list, so that it can get checked on later
                $processed_pairs{$control_node}++;
            }
        } else {
            $has_ha_partner = 0;
        }

        if ($action =~ /reboot/i) {
            # reboot this node via the control node
            $self->disruptive_reboot_generic(-reboot_node => $node_name,
                                             -control_node => $control_node);
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } elsif ($action =~ /panic/i) {
            # panic this node via the control node (or ha partner)
            $self->disruptive_panic_generic(-panic_node => $node_name,
                                            -control_node => $control_node,
                                            -has_ha_partner => $has_ha_partner);
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } elsif ($action =~ /halt/i) {
            # halt this node via the control node
            $self->disruptive_halt_generic(-halt_node => $node_name,
                                           -control_node => $control_node);
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid action '$action'");
            return $EventUtil::FALSE;
        }

        # sleep the disruptive node event post wait time
        my $value = $self->EventUtil::sleep_arg(-arg => "disruptive_node_event_post_wait_time");

        # perform controller validation after 'giveback'
        $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "disruptive_${action}_on_" . $node_name,
                                                                        -last_timeout => $value,
                                                                        -last_timeout_name => "disruptive_node_event_post_wait_time");
        if ($self->error()) {
            $self->_set_err("Controller validation failed after a disruptive '$action' on $node_name");
            return $EventUtil::FALSE;
        }

        # perform path validation after 'giveback'
        $self->PathValidation::validate_paths_after_giveback(-id_extension => "disruptive_${action}_on_" . $node_name,
                                                             -last_timeout => $value,
                                                             -last_timeout_name => "disruptive_node_event_post_wait_time");
        if ($self->error()) {
            $self->_set_err("Path validation failed after a disruptive '$action' on $node_name");
            return $EventUtil::FALSE;
        }

        # re-set the previous node to be the one we just used
        $previous_node = $node_name;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disruptive_panic_generic()
# Objective: Issue a panic against a single c-mode Node
###############################################################################
sub disruptive_panic_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $panic_node = &ParamsMgr::get_arg(-arg => "panic_node",
                                          -required => 1);
    my $control_node = &ParamsMgr::get_arg(-arg => "control_node",
                                           -required => 1);
    my $has_ha_partner = &ParamsMgr::get_arg(-arg => "has_ha_partner",
                                             -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $panic_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $panic_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $control_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $control_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a panic takeover is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_disruptive_panic",
                                          -data => {panic_node => $panic_node,
                                                    control_node => $control_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Disruptively Panicking '$panic_node'"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action
    $self->EventUtil::synchronize();

    # need to disable takeover onpanic on HA partner of panic_node
    if ($has_ha_partner == 1) {
        # make sure takeover onpanic is DISABLED on the ha partner
        try {
            NACL::C::StorageFailover->modify(command_interface => $control_nacl_node_obj,
                                             node => $control_node,
                                             onpanic => "false");
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    # mark that the panicked node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $panic_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # panic the the node
    try {
        # this will panic the node and wait for it to reboot
        NACL::STask::Node->panic(command_interface => $panic_nacl_node_obj,
                                 node => $panic_node,
                                 nacltask_reboot => $EventUtil::TRUE);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that the panic completed
    $self->SubtestComm::log_event_message(-event_message => "completed_disruptive_panic",
                                          -data => {panic_node => $panic_node,
                                                    control_node => $control_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # log to all resources that the panic takeover completed
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Panic on '$panic_node' has completed"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # need to reenable takeover onpanic on HA partner of panic_node
    # if this node has an ha partner
    if ($has_ha_partner == 1) {
        try {
            NACL::C::StorageFailover->modify(command_interface => $control_nacl_node_obj,
                                             node => $control_node,
                                             onpanic => "true");
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    # mark that the panicked node is no longer in a faulted state
    $self->ResMgr::unset_faulted_resource(-name => $panic_node,
                                          -type => "nacl_nodes_cm",
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disruptive_reboot_generic()
# Objective: Issue a reboot against a single c-mode Node
###############################################################################
sub disruptive_reboot_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $reboot_node = &ParamsMgr::get_arg(-arg => "reboot_node",
                                          -required => 1);
    my $control_node = &ParamsMgr::get_arg(-arg => "control_node",
                                           -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $reboot_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $reboot_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $control_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $control_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a reboot takeover is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_disruptive_reboot",
                                          -data => {reboot_node => $reboot_node,
                                                    control_node => $control_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Disruptively rebooting '$reboot_node'"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action
    $self->EventUtil::synchronize();

    # mark that the rebooted node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $reboot_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # reboot the the node and inhibit the takeover from occuring
    try {
        my $transit_obj = NACL::Transit->new(name => $reboot_node);
        $transit_obj->reboot(inhibit_takeover => $EventUtil::TRUE);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that the reboot completed
    $self->SubtestComm::log_event_message(-event_message => "completed_disruptive_reboot",
                                          -data => {reboot_node => $reboot_node,
                                                    control_node => $control_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # log to all resources that the reboot completed
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Disruptive reboot on '$reboot_node' has completed"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # mark that the rebooted node is no longer in a faulted state
    $self->ResMgr::unset_faulted_resource(-name => $reboot_node,
                                          -type => "nacl_nodes_cm",
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disruptive_halt_generic()
# Objective: Issue a halt against a single c-mode Node
###############################################################################
sub disruptive_halt_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $halt_node = &ParamsMgr::get_arg(-arg => "halt_node",
                                        -required => 1);
    my $control_node = &ParamsMgr::get_arg(-arg => "control_node",
                                           -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $halt_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $halt_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $control_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $control_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a halt is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_disruptive_halt",
                                          -data => {halt_node => $halt_node,
                                                    control_node => $control_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Disruptively halting '$halt_node'"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action
    $self->EventUtil::synchronize();

    # halt the node and inhibit any sfo takeover from occuring
    my $transit_obj;
    try {
        $transit_obj = NACL::Transit->new(name => "$halt_node");

        # halt this node and prevent any takeover
        $transit_obj->change_state(to => "FIRMWARE",
                                   inhibit_takeover => $EventUtil::TRUE);
    } otherwise {
        my $e = shift;
        $self->_set_err("Failed to halt node $halt_node, " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # mark that the halted node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $halt_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # boot the node back up
    try {
        $transit_obj->change_state(to => "UP_OR_TAKEN_OVER");
    } otherwise {
        my $e = shift;
        $self->_set_err("Failed to boot node $halt_node, " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that the halt completed
    $self->SubtestComm::log_event_message(-event_message => "completed_disruptive_halt",
                                          -data => {halt_node => $halt_node,
                                                    control_node => $control_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # log to all resources that the halt completed
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Disruptive halt on '$halt_node' has completed"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # mark that the halted node is no longer in a faulted state
    $self->ResMgr::unset_faulted_resource(-name => $halt_node,
                                          -type => "nacl_nodes_cm",
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    simultaneous_node_reboot()
# Objective: Reboot all Nodes specified
###############################################################################
sub simultaneous_node_reboot {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @nacl_nodes = &ParamsMgr::get_arg(-arg => "nacl_nodes",
                                         -default => [],
                                         -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get the nacl node objs
    my %get_args = ();
    if (scalar(@nacl_nodes)) {
        $get_args{-nacl_nodes} = \@nacl_nodes;
    }

    my @nacl_node_objs = $self->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE,
                                                              %get_args);
    if ($self->error()) {
        $self->_set_err("Error searching for NACL node objects");
        return $EventUtil::FALSE;
    }

    # reboot all nodes simultaneously
    my @subtests;
    foreach my $nacl_node_obj (@nacl_node_objs) {
        my $nodename = $nacl_node_obj->node();

        my $subtest = subtest(
            sub {
                # reboot this node (prevent the takeover)
                try {
                    $nacl_node_obj->reboot('inhibit-takeover' => "true");
                } otherwise {
                    # ignore any exceptions thrown
                };

                sleep 20;

                # transit to the CLI
                try {
                    my $transit_obj = NACL::Transit->new(name => $nodename);
                    $transit_obj->change_state(to => 'CLI');
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught during Node reboot: " . $e->text());
                };

                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }

                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            },
            -runid => "reboot_node_${nodename}", -bg,
            "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        push(@subtests, $subtest);
    }

    # reap the install subtests
    $self->wait_for_subtests(-subtests => \@subtests);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################

###############################################################################
#                               Private Methods
###############################################################################

1;

=pod

=head1 Package Name

DisruptiveNodeEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for disruptive node events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::DisruptiveNodeEvents

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::PathValidation

=item EventMan::UtilityModules::SysLogging

=item EventMan::UtilityModules::SysStats

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-disruptive_node_event_post_wait_time>

=over 4

Description: Specifies the amount of time (in seconds) to sleep after performing a disruptive node event.

Per Resource Capable: No

Default: 600 (seconds)

=back

=head2 I<-disruptive_node_action_sleep_time>

=over 4

Description: Specifies the amount of time (in seconds) to sleep briefly before checking to see the status of a node that has disruptively been taken down.

Per Resource Capable: No

Default: 10 (seconds)

=back

=head2 I<-disruptive_control_node>

=over 4

Description: Specifies the control node to send commands to when performing disruptive node events.
If not specified, will automatically select a control node for you. If a node has an HA partner,
it's HA partner will be its control node. If a node does not have an HA partner, one of the nodes in this
operated list will be selected to be used as the control node.

Per Resource Capable: No

Default: ""

=back

=head2 I<-symmetrical_events>

=over 4

Description: Indicates whether to perform symmetrically styled events (A -> B then B -> A)

Per Resource Capable: No

Default: ""

=back

=head1 B<Event Keywords>

=head2 B<disruptive_node_haltCm>

Disruptively Halts (no SFO) a list of nodes in sequence one a time.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resources to
operate against.

Default: all CMode NACL nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "disruptive_node_haltCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<disruptive_node_rebootCm>

Disruptively Reboots (no SFO) a list of nodes in sequence one a time.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resources to
operate against.

Default: all CMode NACL Nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "disruptive_node_rebootCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<disruptive_node_panicCm>

Disruptively Panics (no SFO) a list of nodes in sequence one a time.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resource to
operate against.

Default: all CMode NACL Nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "disruptive_node_panicCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<disruptive_cluster_rebootCm>

Disruptively reboots all the nodes of each cluster at the same time. Only
one cluster will be rebooted at a time.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resource to
operate against.

Default: all CMode NACL Nodes in the Event scope will be used

=item *I<-nacl_clusters_cm>

Required: An array reference of CMode cluster resources to operate against.

Default: all CMode NACL Clusters in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 # reboot every cluster known to eventman
 # each cluster will be rebooted sequentially
 $EventManobj->add_event(-event_name => "disruptive_cluster_rebootCm")
 if ($EventManobj->error()) {
     ...
 }

 # only reboot cluster 'mycluster'
 $EventManobj->add_event(-event_name => "disruptive_cluster_rebootCm",
                         -nacl_clusters_cm => ['mycluster'])
 if ($EventManobj->error()) {
     ...
 }

 # only reboot nodes node1, node2, node3 in 'mycluster'
 $EventManobj->add_event(-event_name => "disruptive_cluster_rebootCm",
                         -nacl_clusters_cm => ['mycluster'],
                         -nacl_nodes_cm => ['node1','node2','node3'])
 if ($EventManobj->error()) {
     ...
 }

 # reboot nodes that are in 'mycluster1' and 'mycluster2'.
 # the list of nodes specifies which nodes in the clusters will get
 # rebooted. Each cluster gets rebooted sequentially
 $EventManobj->add_event(-event_name => "disruptive_cluster_rebootCm",
                         -nacl_clusters_cm => ['mycluster1','mycluster2'],
                         -nacl_nodes_cm => ['node1_on_mycluster1','node2_on_mycluster1','node3_on_mycluster2'])
 if ($EventManobj->error()) {
     ...
 }

=back

=head1 B<Public Methods>

=head2 B<disruptive_action_abstracted(%args)>

Performs a generic disruptive action against a list of nacl_nodes_cm nodes

=over 4

=item Inputs

=item *I<-action>

Required: Specifies what action to take, valid actions are 'reboot', 'panic', 'halt'

=item Examples

 # issue a disruptive reboot action against the c-mode nodes in the event
 $EventManObj->disruptive_action_abstracted(-action => "reboot");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<disruptive_panic_generic(%args)>

Panic a single c-mode node disruptively.

=over 4

=item Inputs

=item *I<-panic_node>

Required: Specifies the node to panic

=item *I<-control_node>

Required: Specifies the node to use when panicking -panic_node

=item *I<-has_ha_partner>

Required: Specifies whether panic node has an HA partner. If set to 1, then
-control_node must be the HA partner of -panic_node.

=item Examples

 # issue a disruptive panic against a panic node via its ha partner
 $EventManObj->disruptive_panic_generic(-panic_node => "fas3070-1",
                                        -control_node => "fas3070-2",
                                        -has_ha_partner => 1);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<disruptive_halt_generic(%args)>

Halt a single c-mode node disruptively.

=over 4

=item Inputs

=item *I<-halt_node>

Required: Specifies the node to halt

=item *I<-control_node>

Required: Specifies the node to use when halting -halt_node

=item Examples

 # issue a disruptive halt against halt_node
 $EventManObj->disruptive_halt_generic(-halt_node => "fas3070-1",
                                       -control_node => "fas3070-2");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<disruptive_reboot_generic(%args)>

Reboot a single c-mode node disruptively.

=over 4

=item Inputs

=item *I<-reboot_node>

Required: Specifies the node to reboot

=item *I<-control_node>

Required: Specifies the node to use when rebooting -reboot_node

=item Examples

 # issue a disruptive reboot against reboot_node
 $EventManObj->disruptive_reboot_generic(-reboot_node => "fas3070-1",
                                       -control_node => "fas3070-2");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<simultaneous_node_reboot(%args)>

Reboots all Nodes in scope (or specified) in parallel

=over 4

=item Inputs

=item *I<-nacl_node_objs>

Optional: Specifies which node objects to reboot in parallel

=item Examples

 # reboot all nodes in the current Event scope simultaneously
 $EventManObj->simultaneous_node_reboot()
 if ($EventManObj->error()) {
     ...
 }

=back

=cut
