# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for CMode Lif events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::LifEvents;

use Data::Dumper;
use Tharn;
use POSIX qw(ceil);
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);

# NACL packages
use NACL::CS::NetworkInterface;
use NACL::Exceptions::NoElementsFound;
use NACL::C::NetworkInterface;
use NACL::APISet::Exceptions::ConnectionFailedException;

use warnings;
use strict;

use vars qw($Log);

$Log = log_global();

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "data_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring Data lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &data_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "san_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring SAN lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &san_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nas_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring NAS lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &nas_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "fc_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring FCP lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &fc_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "iscsi_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring iSCSI lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &iscsi_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "cifs_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring CIFS lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &cifs_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nfs_lif_down_up_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "This event will bring NFS lifs administratively down and then up",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -exec => sub { &nfs_lif_down_up_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { lif_down_time => 30,
                                                           lif_up_time => 30,
                                                           lif_down_up_total_count => -1,
                                                           lif_down_up_execution_style => "parallel",
                                                           lif_down_up_name_regex_filter => undef,
                                                           lif_down_up_home_nodes => [],
                                                           lif_down_up_parallel_vserver_count => 10 }
                                                     );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    data_lif_down_up_Cm()
# Objective: Administratively bring all Data Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub data_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["iscsi", "fcp", "nfs", "cifs"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    san_lif_down_up_Cm()
# Objective: Administratively bring SAN Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub san_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["iscsi", "fcp"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nas_lif_down_up_Cm()
# Objective: Administratively bring NAS Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nas_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["nfs", "cifs"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    fc_lif_down_up_Cm()
# Objective: Administratively bring FCP Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fc_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["fcp"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    iscsi_lif_down_up_Cm()
# Objective: Administratively bring iSCSI Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["iscsi"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nfs_lif_down_up_Cm()
# Objective: Administratively bring NFS Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nfs_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["nfs"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    cifs_lif_down_up_Cm()
# Objective: Administratively bring CIFS Lifs down and up in a Cmode cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cifs_lif_down_up_Cm {
    my ($self, %args) = @_;

    $self->lif_down_up_abstracted(-protocols => ["cifs"]);
    if ($self->error()) {
        $self->_set_err("Lif down up operations failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    lif_down_up_abstracted()
# Objective: Administratively bring Lifs down and up in all c-mode clusters
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lif_down_up_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $data_protocols = &ParamsMgr::get_arg(-arg => "protocols",
                                             -required => 1,
                                             -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
    # get the $lif_down_up_name_regex_filter
    my $lif_down_up_name_regex_filter = $self->EventUtil::get_arg(-arg      => "lif_down_up_name_regex_filter",
                                                                  -required => 0,
                                                                  -default  => undef);
    if ($self->error()) {
        $self->_set_err("Error getting lif_down_up_name_regex_filter");
        return $EventUtil::FALSE;
    }

    # get the $lif_down_up_home_nodes
    my @lif_down_up_home_nodes = @{$self->EventUtil::get_arg(-arg      => "lif_down_up_home_nodes",
                                                             -required => 0,
                                                             -default  => [] )};

    if ($self->error()) {
        $self->_set_err("Error getting lif_down_up_home_nodes");
        return $EventUtil::FALSE;
    }

    # for every cmode cluster on this event
    my @nacl_clusters = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        return $EvenUtil::FALSE;
    }

    my @data_protocols = @{$data_protocols};
    my $execution_style = $self->EventUtil::get_arg(-arg => "lif_down_up_execution_style");

    foreach my $cluster_name (@nacl_clusters) {
	# find vservers in this cluster
        my @vservers = $self->EventUtil::get_vservers(-clusters => [$cluster_name]);
        if ($self->error()) {
            $self->_set_err("Failed to find vserver resources in the cluster $cluster_name");
            return $EvenUtil::FALSE;
        }

        # get a non faulted node command interface that lives in this cluster
        my @nacl_node_objs = $self->EventUtil::get_nacl_node_objs(-type => "cmode",
                                                                  -faulted => $EventUtil::FALSE,
                                                                  -clusters => [$cluster_name],
                                                                  -fall_back_on_nacl_clusters => $EventUtil::TRUE);
        if ($self->error()) {
            $self->_set_err("Failed to obtain any nonfaulted nacl nodes for cluster $cluster_name");
            return $EventUtil::FALSE;
        }

        # some vservers may not have lifs that match the data protocol that we wish to
        # perform operations against, filter out those vservers here.
        my @vservers_filtered_by_protocol;
        foreach my $vserver_name (@vservers) {
            my @lif_objs;
            try {
                # use the first node in the list as the command interface
                @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $nacl_node_objs[0],
                                                              filter => { vserver => $vserver_name,
                                                                          role => "data"});
            } catch NACL::Exceptions::NoElementsFound with {
                @lif_objs = ();
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught while trying to fetch Lif objects: " . $e->text());
            };
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # filter out lifs that don't match the requested protocol from the caller
            my @filtered_lif_objs = ();
            foreach my $lif_obj (@lif_objs) {
                my @lif_protocols = $lif_obj->data_protocol();
                foreach my $proto (@data_protocols) {
                    if (grep(/$proto/, @lif_protocols)) {
                        # lif matched, break out
                        push(@filtered_lif_objs, $lif_obj);
                        last;
                    }
                }
            }
            @lif_objs = @filtered_lif_objs;

            # filter luns based on name regex filter
            if (defined $lif_down_up_name_regex_filter) {
                my @name_filtered_lif_objs = ();
                foreach my $lif_obj (@lif_objs) {
                    my $lif_name = $lif_obj->lif();
                    if ($lif_name =~ $lif_down_up_name_regex_filter) {
                        push(@name_filtered_lif_objs, $lif_obj);
                    }
                }
                @lif_objs = @name_filtered_lif_objs;
            }

            # filter luns based on home node
            if (scalar(@lif_down_up_home_nodes)) {
                my @home_node_filtered_lif_objs = ();
                foreach my $lif_obj (@lif_objs) {
                    my $home_node = $lif_obj->home_node();
                    if (grep(/$home_node/, @lif_down_up_home_nodes)) {
                        push(@home_node_filtered_lif_objs, $lif_obj);
                    }
                }
                @lif_objs = @home_node_filtered_lif_objs;
            }

            # if we matched lifs, then keep this vserver
            if (scalar(@lif_objs)) {
                push(@vservers_filtered_by_protocol, $vserver_name);
            }
        }

        # if there are no vservers that matched the filters, this is an error
        if (not scalar(@vservers_filtered_by_protocol)) {
            $self->_set_err("There are no vservers that have LIFs that match the " .
                            "requested data protocols '@data_protocols'");
            return $EventUtil::FALSE;
        }

        # calculate how many lif down up operations we should perform for each vserver.
        # each vserver should get an equal portion of the total # of down/up operations
        # that was requested to perform. For example, if there are 2 vservers and
        # the event parameter lif_down_up_total_count=100, then both vservers will
        # perform 50 lif down up operations. Round up the number. If the user
        # requested -1 total lif down up operations, then perform enough operations per
        # vserver such that each lif for each vserver gets blocked exactly once.
        my $lif_down_up_count_per_vserver;
        if ($self->EventUtil::get_arg(-arg => "lif_down_up_total_count") > 0) {
            $lif_down_up_count_per_vserver =
                ceil($self->EventUtil::get_arg(-arg => "lif_down_up_total_count") / scalar(@vservers_filtered_by_protocol));
        } else {
            $lif_down_up_count_per_vserver = -1;
        }

        # perform lif down/up operations on all vservers that have made it this far
        if ($execution_style =~ /parallel/i) {
            $self->parallel_lif_down_up_operations(-cluster_name => $cluster_name,
                                                   -lif_down_up_count => $lif_down_up_count_per_vserver,
                                                   -data_protocols => $data_protocols,
                                                   -filtered_by_protocol => \@vservers_filtered_by_protocol);
        } else {
            foreach my $vserver (@vservers_filtered_by_protocol) {
                # spawn a thread for each vserver (if sequential)
                $self->lif_down_up_per_vserver(-vserver_name => $vserver,
                                               -cluster_name => $cluster_name,
                                               -lif_down_up_count => $lif_down_up_count_per_vserver,
                                               -execution_style => "sequential",
                                               -data_protocols => $data_protocols);
                if ($self->error()) {
                    $self->_set_err("Lif down/up failed on vserver '$vserver'");
                    return $EventUtil::FALSE;
                }
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    parallel_lif_down_up_operations
# Objective: Administratively bring Lifs down and up for this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub parallel_lif_down_up_operations {

    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $cluster_name = &ParamsMgr::get_arg(-arg => "cluster_name",
                                           -required => 1);
    my $data_protocols = &ParamsMgr::get_arg(-arg => "data_protocols",
                                             -ref_type => "ARRAY",
                                             -required => 1);
    my $lif_down_up_count_per_vserver = &ParamsMgr::get_arg(-arg => "lif_down_up_count",
                                                -default => -1);
    my $vservers_filtered_by_protocol_ref = &ParamsMgr::get_arg(-arg => "filtered_by_protocol",
                                                                -ref_type => "ARRAY",
                                                                -required => 1);
    my @vservers_filtered_by_protocol = @{$vservers_filtered_by_protocol_ref};

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @subtests;

    # initialize path validation and controller validation type to lif down
    my $path_validation_type = "lif_down";
    my $controller_validation_type = "lif_down";
    my $port_validation_type = "lif_down";
    my $lif_down_up_count = 0;

    # build the controller validation method callback that will be executed when
    # all subtests hit a sync point
    my $controller_validation_method = sub {
        my ($self) = @_;

        # We only validate lifs when they come back up
        if ($controller_validation_type eq "lif_up") {
            $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "lif_up_" . $lif_down_up_count,
                                                                            -last_timeout => $self->EventUtil::get_arg(-arg => "lif_up_time"),
                                                                            -last_timeout_name => "lif_up_time");
        }

        if ($self->error()) {
            $self->_set_err("Controller validation failed after ${controller_validation_type} operation " .
                            "(count=$lif_down_up_count)");
            return $EventUtil::FALSE;
        }

        # flip the controller_validation_type so the next invocation of this method gets the value right
        if ($controller_validation_type eq "lif_down") {
            $controller_validation_type = "lif_up";
        } else {
            $controller_validation_type = "lif_down";
        }

        return $EventUtil::TRUE;
    };

    # build the port validation method callback that will be executed when
    # all subtests hit a sync point
    my $port_validation_method = sub {
        my ($self) = @_;

        if ($port_validation_type eq "lif_up") {
            $self->PortValidation::validate_ports_after_lif_up(-id_extension => "lif_up_" . $lif_down_up_count,
                                                               -last_timeout => $self->EventUtil::get_arg(-arg => "lif_up_time"),
                                                               -last_timeout_name => "lif_up_time");
        } elsif ($port_validation_type eq "lif_down") {
            $self->PortValidation::validate_ports_after_lif_down(-id_extension => "lif_down_" . $lif_down_up_count,
                                                                 -last_timeout => $self->EventUtil::get_arg(-arg => "lif_down_time"),
                                                                 -last_timeout_name => "lif_down_time");
        }

        if ($self->error()) {
            $self->_set_err("Port validation failed after ${port_validation_type} operation (count=$lif_down_up_count)");
            return $EventUtil::FALSE;
        }

        # flip the port_validation_type so the next invocation of this method gets the value right
        if ($port_validation_type eq "lif_down") {
                $port_validation_type = "lif_up";
            } else {
                $port_validation_type = "lif_down";
            }

            return $EventUtil::TRUE;
        };

    # build the path validation method callback that will be executed when
    # all subtests hit a sync point
    my $path_validation_method = sub {
        my ($self) = @_;

        if ($path_validation_type eq "lif_down") {
            $self->PathValidation::validate_paths_after_lif_down(-id_extension => "lif_down_" . $lif_down_up_count,
                                                                 -path_validation_rules => {validate_path_counts => 1,
                                                                                            path_down_count => 1},
                                                                 -last_timeout => $self->EventUtil::get_arg(-arg => "lif_down_time"),
                                                                 -last_timeout_name => "lif_down_time");
        } elsif ($path_validation_type eq "lif_up") {
            $self->PathValidation::validate_paths_after_lif_up(-id_extension => "lif_up_" . $lif_down_up_count,
                                                               -last_timeout => $self->EventUtil::get_arg(-arg => "lif_up_time"),
                                                               -last_timeout_name => "lif_up_time");
        }

        if ($self->error()) {
            $self->_set_err("Path validation failed after ${path_validation_type} operation (count=$lif_down_up_count)");
            return $EventUtil::FALSE;
        }

        # flip the path_validation_type so the next invocation of this method gets the value right
        if ($path_validation_type eq "lif_down") {
            $path_validation_type = "lif_up";
        } else {
            # increment the lif down_up count
            $lif_down_up_count++;

            $path_validation_type = "lif_down";
        }

        return $EventUtil::TRUE;
    };

    # build a sysstats callback method to collect controller data during sync points
    my $sysstats_method = sub {
        my ($self) = @_;

        $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # collect cluster statistics
        $self->SysStats::collect_cluster_statistics(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        return $EventUtil::TRUE;
    };

    # perform lif down/up operations on all vservers that have made it this far
    my $parallel_vserver_count = $self->EventUtil::get_arg(-arg => "lif_down_up_parallel_vserver_count",
                                                           -required => 1);
    my $subtest_count = 0;
    foreach my $vserver (@vservers_filtered_by_protocol) {
        # spawn a thread for each vserver (if parallel)
        my $subtest = subtest(
            sub {
                # perform lif down up operations on this vserver
                $self->lif_down_up_per_vserver(-vserver_name => $vserver,
                                               -cluster_name => $cluster_name,
                                               -lif_down_up_count => $lif_down_up_count_per_vserver,
                                               -execution_style => "parallel",
                                               -data_protocols => $data_protocols);
                if ($self->error()) {
                    $self->_set_err("Lif down/up failed on vserver '$vserver'");
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                } else {
                    $self->SubtestComm::subtest_complete();
                    return $EventUtil::TRUE;
                }
            },
            -runid => "lif_down_up_vserver__$vserver",
            -bg,
            "-onexit" => sub {my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);

        $subtest_count++;

        if (scalar(@subtests) == $parallel_vserver_count || $subtest_count == scalar(@vservers_filtered_by_protocol)) {
            # reap all the lif down up subtests, and pass in the callback
            # methods to be executed every time the subtests hit a synchronization point
            $self->wait_for_subtests(-subtests => \@subtests,
                                     -dependence_type => "dependent",
                                     -propagate_wait => 0,
                                     -propagate_events => 1,
                                     -enable_sync => 1,
                                     -sync_callbacks => [$controller_validation_method,
                                                         $port_validation_method,
                                                         $path_validation_method,
                                                         $sysstats_method],
                                     -check_cmds_from_parent => 0,
                                     -forward_cmds_from_parent => 0,
                                     -async => 0);

            if ($self->error()) {
                $self->_set_err("Error encountered while waiting for lif/down subtests to complete");
                return $EventUtil::FALSE;
            }

            @subtests = (); #reset subtests array
        }
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    lif_down_up_per_vserver()
# Objective: Administratively bring Lifs down and up for this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lif_down_up_per_vserver {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $vserver_name = &ParamsMgr::get_arg(-arg => "vserver_name",
                                           -required => 1);
    my $cluster_name = &ParamsMgr::get_arg(-arg => "cluster_name",
                                           -required => 1);
    my $data_protocols = &ParamsMgr::get_arg(-arg => "data_protocols",
                                             -ref_type => "ARRAY",
                                             -required => 1);
    my $lif_down_up_count = &ParamsMgr::get_arg(-arg => "lif_down_up_count",
                                                -default => -1);
    my $execution_style = &ParamsMgr::get_arg(-arg => "execution_style",
                                           -required => 1,
                                           -format => qr/^(sequential|parallel)$/i);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
    my $lif_down_up_name_regex_filter = $self->EventUtil::get_arg(-arg      => "lif_down_up_name_regex_filter",
                                                                  -required => 0,
                                                                  -default  => undef);
    if ($self->error()) {
        $self->_set_err("Error getting lif_down_up_name_regex_filter");
        return $EventUtil::FALSE;
    }

    my @lif_down_up_home_nodes = @{$self->EventUtil::get_arg(-arg      => "lif_down_up_home_nodes",
                                                             -required => 0,
                                                             -default  => [] )};
    if ($self->error()) {
        $self->_set_err("Error getting lif_down_up_name_regex_filter");
        return $EventUtil::FALSE;
    }
    my $count = 0;

    # get a non faulted node command interface that lives in this cluster
    my @nacl_node_objs = $self->EventUtil::get_nacl_node_objs(-type => "cmode",
                                                              -faulted => $EventUtil::FALSE,
                                                              -clusters => [$cluster_name],
                                                              -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Failed to obtain any nonfaulted nacl nodes for cluster $cluster_name");
        return $EventUtil::FALSE;
    }

    # grep out lif objects that do not match the requested protocols
    my @data_protocols = @{$data_protocols};
    my @lif_objs;
    my $retry_connection_count = 5;
    while (1) {
        my $connection_failed_flag = 0;

        try {
            @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $nacl_node_objs[0],
                                                          filter            => { vserver => $vserver_name,
                                                                                 role    => "data"});
        } catch NACL::Exceptions::NoElementsFound with {
            @lif_objs = ();
        } otherwise {
            my $e = shift;
            $connection_failed_flag = 1;
            if (--$retry_connection_count <= 0) {
                $self->_set_err("Execption " . ref($e) . " caught while trying to make connection: " . $e->text());
            } else {
                $Log->comment("Connection failed. Retrying Connection. $retry_connection_count retries left.");
                sleep(10);
                try {
                    #Refresh the command interface to rerun the login commands
                    $nacl_node_objs[0]->refresh_command_interface(no_version_change => $EventUtil::TRUE);
                } otherwise {
                    # ignore any error that happens
                };
            }
        };

        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        redo if $connection_failed_flag;
        last;
    }
    # filter out lifs that don't match the requested protocol from the caller
    my @filtered_lif_objs = ();
    foreach my $lif_obj (@lif_objs) {
        my @lif_protocols = $lif_obj->data_protocol();
        foreach my $proto (@data_protocols) {
            if (grep(/$proto/, @lif_protocols)) {
                # lif matched, break out
                push(@filtered_lif_objs, $lif_obj);
                last;
            }
        }
    }
    @lif_objs = @filtered_lif_objs;

    # if at this point there are no lif objects, then this is a serious error
    if (not scalar(@lif_objs)) {
        $self->_set_err("There are no LIFs on vserver '$vserver_name' that " .
                        "match the requested protocols '@data_protocols'");
        return $EventUtil::FALSE;
    }

    if (defined $lif_down_up_name_regex_filter) {
        my @name_filtered_lif_objs = ();
        foreach my $lif_obj (@lif_objs) {
            my $lif_name = $lif_obj->lif();
            if ($lif_name =~ $lif_down_up_name_regex_filter) {
                push(@name_filtered_lif_objs, $lif_obj);
            }
        }
        @lif_objs = @name_filtered_lif_objs;

        # if at this point there are no lif objects, then this is a serious error
        if (not scalar(@lif_objs)) {
            $self->_set_err("There are no LIFs on vserver '$vserver_name' that " .
                            "match the lif filter '$lif_down_up_name_regex_filter'");
            return $EventUtil::FALSE;
        }
    }

    # filter luns based on home node
    if (scalar(@lif_down_up_home_nodes)) {
        my @home_node_filtered_lif_objs = ();
        foreach my $lif_obj (@lif_objs) {
            my $home_node = $lif_obj->home_node();
            if (grep(/$home_node/, @lif_down_up_home_nodes)) {
                push(@home_node_filtered_lif_objs, $lif_obj);
            }
        }
        @lif_objs = @home_node_filtered_lif_objs;
        # if at this point there are no lif objects, then this is a serious error
        if (not scalar(@lif_objs)) {
            $self->_set_err("There are no LIFs on vserver '$vserver_name' that " .
                            "match the lif home nodes filter");
            return $EventUtil::FALSE;
        }
    }

    # keep performing lif down/up operations until we hit the total lif down up count for this vserver
    while (1) {

        foreach my $lif_obj (@lif_objs) {
            # Get the lif name.  We need this to be able to add it
            # and remove it to/from the downed_lifs list on the vserver object
            my $lif_name = $lif_obj->lif();
            if (not defined($lif_name) or not length($lif_name)) {
                $self->_set_err("Unable to get the lif name from NACL::CS::NetworkInterface object");
                return $EventUtil::FALSE;
            }

            $self->SubtestComm::log_event_message(-event_message => "starting_lif_down_${vserver_name}");

            # if we have hit the max requested number of lif down up operations,
            # then set the lif down up count to -1 and break out
            if ($count == $lif_down_up_count) {
                $lif_down_up_count= -1;
                last;
            }

            # bring the lif down
            try {
                NACL::C::NetworkInterface->modify(command_interface => $nacl_node_objs[0],
                                                  server => $vserver_name,
                                                  lif => $lif_name,
                                                  'status-admin' => "down");
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught while attempting bring lif $lif_name ($vserver_name) down");
            };
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # Add the lif to the "downed_lifs" key on the vserver object
            $self->_add_downed_lif(-vserver => $vserver_name,
                                   -cluster => $cluster_name,
                                   -lif => $lif_name);
            if ($self->error()) {
                $self->_set_err("Failed to add lif '$lif_name' of vserver '$vserver_name' from downed lif list");
                return $EventUtil::FALSE;
            }

            $self->EventUtil::sleep_arg(-arg => "lif_down_time");

            $self->EventUtil::synchronize();

            # if execution style is sequential, then perform path validation now
            if ($execution_style =~ /sequential/i) {
                # validate ports after the lif up occurred
                $self->PortValidation::validate_ports_after_lif_down(-id_extension => "lif_down_" . $count,
                                                                     -last_timeout => $self->EventUtil::get_arg(-arg => "lif_down_time"),
                                                                     -last_timeout_name => "lif_down_time");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                $self->PathValidation::validate_paths_after_lif_down(-id_extension => "lif_down_" . $count,
                                                                     -path_validation_rules => {validate_path_counts => 1,
                                                                                                path_down_count => 1},
                                                                     -last_timeout => $self->EventUtil::get_arg(-arg => "lif_down_time"),
                                                                     -last_timeout_name => "lif_down_time");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }
                $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                # collect cluster statistics
                $self->SysStats::collect_cluster_statistics(-type => "cmode");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }
            }

            $self->SubtestComm::log_event_message(-event_message => "starting_lif_up_${vserver_name}");

            # bring the lif up
            try {
                NACL::C::NetworkInterface->modify(command_interface => $nacl_node_objs[0],
                                                  server => $vserver_name,
                                                  lif => $lif_name,
                                                  'status-admin' => "up");
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught while attempting bring lif $lif_name ($vserver_name) up");
            };
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # Remove the lif from the "downed_lifs" key on the vserver object
            $self->_remove_downed_lif(-vserver => $vserver_name,
                                      -cluster => $cluster_name,
                                      -lif => $lif_name);
            if ($self->error()) {
                $self->_set_err("Failed to remove lif '$lif_name' of vserver '$vserver_name' from downed lif list");
                return $EventUtil::FALSE;
            }

            $self->EventUtil::sleep_arg(-arg => "lif_up_time");

            $self->EventUtil::synchronize();

            # if execution style is sequential, then perform path validation now
            if ($execution_style =~ /sequential/i) {
                # validate lifs after the lif up occurred
                $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "lif_up_" . $count,
                                                                                -last_timeout => $self->EventUtil::get_arg(-arg => "lif_up_time"),
                                                                                -last_timeout_name => "lif_up_time");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                # validate ports after the lif up occurred
                $self->PortValidation::validate_ports_after_lif_up(-id_extension => "lif_up_" . $count,
                                                                   -last_timeout => $self->EventUtil::get_arg(-arg => "lif_up_time"),
                                                                   -last_timeout_name => "lif_up_time");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                $self->PathValidation::validate_paths_after_lif_up(-id_extension => "lif_up_" . $count,
                                                                   -last_timeout => $self->EventUtil::get_arg(-arg => "lif_up_time"),
                                                                   -last_timeout_name => "lif_up_time");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }
                $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                # collect cluster statistics
                $self->SysStats::collect_cluster_statistics(-type => "cmode");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }
            }

            # incrememnt the lif down up count
            $count++;
        }

        # if the lif down up count is -1, then we are done
        if ($lif_down_up_count == -1) {
            last;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

################################################################################
# Method Name: _add_downed_lif
# Objective: Add a lif to the list of downed lifs for a vserver resource
################################################################################
sub _add_downed_lif {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $lif     = &ParamsMgr::get_arg(-arg => "lif", -required => 1);
    my $vserver = &ParamsMgr::get_arg(-arg => "vserver", -required => 1);
    my $cluster = &ParamsMgr::get_arg(-arg => "cluster", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Get the current value for the downed lifs for the vserver resource
    my $downed_lifs_hash = $self->ResMgr::get_resource_attribute(-name      => $vserver,
                                                                 -type      => "vservers",
                                                                 -parent_name => $cluster,
                                                                 -parent_type => "nacl_clusters_cm",
                                                                 -attribute => "downed_lifs");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Add the lif if
    $downed_lifs_hash->{$lif} = 1;

    # Set the value of for the attribute back on the resource
    $self->ResMgr::set_resource_attribute(-name      => $vserver,
                                          -type      => "vservers",
                                          -attribute => "downed_lifs",
                                          -parent_name => $cluster,
                                          -parent_type => "nacl_clusters_cm",
                                          -value     => $downed_lifs_hash,
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to set resource attribute 'downed_lifs' on resource '$vserver' " .
                        "for type 'vservers'");
        return $EventUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _remove_downed_lif
# Objective: Remove a lif from the list of downed lifs for a vserver resource
################################################################################
sub _remove_downed_lif {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $lif     = &ParamsMgr::get_arg(-arg => "lif", -required => 1);
    my $vserver = &ParamsMgr::get_arg(-arg => "vserver", -required => 1);
    my $cluster = &ParamsMgr::get_arg(-arg => "cluster", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Get the current value for the downed lifs for the vserver resource
    my $downed_lifs_hash = $self->ResMgr::get_resource_attribute(-name      => $vserver,
                                                                 -type      => "vservers",
                                                                 -parent_name => $cluster,
                                                                 -parent_type => "nacl_clusters_cm",
                                                                 -attribute => "downed_lifs");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Delete the lif from the downed_lifs_hash if it exists
    if (exists($downed_lifs_hash->{$lif})) {
        delete($downed_lifs_hash->{$lif});
    }

    # Set the value of for the attribute back on the resource
    $self->ResMgr::set_resource_attribute(-name      => $vserver,
                                          -type      => "vservers",
                                          -attribute => "downed_lifs",
                                          -parent_name => $cluster,
                                          -parent_type => "nacl_clusters_cm",
                                          -value     => $downed_lifs_hash,
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to set resource attribute 'downed_lifs' on resource '$vserver' " .
                        "for type 'vservers'");
        return $EventUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

=pod

=head1 Package Name

LifEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for CMode Lif Events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::LifEvents

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::PathValidation

=item EventMan::UtilityModules::SysLogging

=item EventMan::UtilityModules::SysStats

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-lif_down_time>

=over 4

Description: Specifies the amount of time (in seconds) to sleep after performing a lif down operation.

Per Resource Capable: No

Default: 30 (seconds)

=back

=head2 I<-lif_up_time>

=over 4

Description: Specifies the amount of time (in seconds) to sleep after performing a lif up operation.

Per Resource Capable: No

Default: 30 (seconds)

=back

=head2 I<-lif_down_up_total_count>

=over 4

Description: Specifies the total amount of lif down up operations to perform. Note: this is an approximation, not an absolute value.
By default, each lif in the event scope will be brought down and up exactly once.

Per Resource Capable: No

Default: -1

=back

=head2 I<-lif_down_up_execution_style>

=over 4

Description: Specifies how to perform the lif down/up operations, 'parallel' or 'sequential'. Parallel lif down up operations
will perform each vserver in parallel, where at most one lif per vserver gets operated at a time

Per Resource Capable: No

Default: parallel

=back

=head2 I<-lif_down_up_name_regex_filter>

=over 4

Description: A regular expression that a lif name must match in order for the down/up event to be run on it.

Per Resource Capable: No

Default: none

=back

=head2 I<-lif_down_up_home_nodes>

=over 4

Description: A list of lif home nodes. The lif up down event will only be run on lifs that have a home
node that is in this list.

Per Resource Capable: No

Default: all nodes

=back

=head1 B<Event Keywords>

=head2 B<data_lif_down_up_Cm>

Administratively brings Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all data lifs owned by vs0 and vs1 in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "data_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"],
                         -vservers => ["vs0", "vs1"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<san_lif_down_up_Cm>

Administratively brings SAN Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all SAN data lifs in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "san_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nas_lif_down_up_Cm>

Administratively brings NAS Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all NAS data lifs in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nas_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<fc_lif_down_up_Cm>

Administratively brings FC Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all FC data lifs in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "fc_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<iscsi_lif_down_up_Cm>

Administratively brings iSCSI Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all iSCSI data lifs in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "iscsi_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nfs_lif_down_up_Cm>

Administratively brings NFS Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all NFS data lifs in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nfs_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<cifs_lif_down_up_Cm>

Administratively brings CIFS Data Lifs down and up

=over 4

=item Event Parameters

=item *I<-controllerclusters>

Required: An array reference of cmode controller cluster resources to operate against

=item *I<-vservers>

Optional: An array reference of cmode vserver resources to operate against. By default
all vservers in the controller clusters specified will be operated against.

=item Examples

 # Perform data lif down/up operations against all CIFS data lifs in cluster 'mycluster'
 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "cifs_lif_down_up_Cm",
                         -controllerclusters => ["mycluster"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head1 B<Public Methods>

=head2 B<lif_down_up_abstracted(%args)>

Performs the generic lif down up operation against a set of controller clusters & vservers

=over 4

=item Inputs

=item *I<-protocols>

Required: Specifies what LIF protocols to perform down/up operations against

=item Examples

 # play with san lifs
 $EventManObj->lif_down_up_abstracted(-protocols => ["iscsi", "fcp"]);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<lif_down_up_per_vserver(%args)>

Performs the generic lif down up operations against all specified lifs in the specified vserver

=over 4

=item Inputs

=item *I<-vserver_name>

Required: Vserver to operate against

=item *I<-cluster_name>

Required: The name of the cluster that is being operated against

=item *I<-data_protocols>

Required: An array reference of protocols to block

=item *I<-lif_down_up_count>

Optional: The exact number of down up operations to perform, by default each lif gets
operated on at exactly once.
Default: -1

=item *I<-execution_style>

Required: specifies whether this vserver is being blocked in 'parallel'
or 'sequential' order relative to its sibling vservers

=item Examples

 # perform 10 down up operations against myvs0's SAN lifs
 $EventManObj->lif_down_up_per_vserver(-data_protocols => ["iscsi", "fcp"],
                                       -vserver_name => "myvs0",
                                       -cluster_name => "myclus",
                                       -lif_down_up_count => 10,
                                       -execution_style => "sequential");
 if ($EventManObj->error()) {
     ...
 }

=back

=cut

1;
