# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for CMode Node Upgrade Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::NodeUpgradeEvents;

use ParamsMgr;
use NATE::BaseException qw(:try);
use NACL::STask::SystemNodeImage;
use NACL::CS::SystemNodeImage;
use NACL::MTask::Revert;
use San::SanUtil;
use Data::Compare;

use Tharn;
use Data::Dumper;

use strict;
use warnings;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode::DisruptiveNodeEvents",
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "disruptive_upgradeCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that performs a disruptive upgrade on a cluster",
                                     -required_resource_types => ["nacl_clusters_cm","nacl_nodes_cm"],
                                     -validate => sub { &_validate_commonCm(@_); },
                                     -exec => sub { &disruptive_upgradeCm(@_); });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    @event_modules = (__PACKAGE__,
                      "EventMan::EventModules::ControllerEvents::CMode::SFOEvents",
                      "EventMan::EventModules::ControllerEvents::CMode",
                      "EventMan::EventModules::ControllerEvents",
                      "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "nduCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that performs an NDU on a cluster",
                                     -required_resource_types => ["nacl_clusters_cm","nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+,even" },
                                     -validate => sub { &_validate_nduCm(@_); },
                                     -exec => sub { &nduCm(@_); });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "minor_nduCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that performs a Minor NDU on a cluster",
                                     -required_resource_types => ["nacl_clusters_cm","nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+,even" },
                                     -exec => sub { &nduCm(@_); },
                                     -validate => sub { &_validate_nduCm(@_); },
                                     -arg_hardcoded => {
                                         ndu_type => "minor"
                                     });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "major_nduCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that performs a Major NDU on a cluster",
                                     -required_resource_types => ["nacl_clusters_cm","nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+,even" },
                                     -exec => sub { &nduCm(@_); },
                                     -validate => sub { &_validate_nduCm(@_); },
                                     -arg_hardcoded => {
                                         ndu_type => "major"
                                     });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    @event_modules = (__PACKAGE__,
                      "EventMan::EventModules::ControllerEvents::CMode",
                      "EventMan::EventModules::ControllerEvents",
                      "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "revertCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that performs a revert on a CMode cluster",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -resource_counts => { nacl_nodes_cm => "2+,even" },
                                     -validate => sub { &_validate_revertCm(@_); },
                                     -exec => sub { &revertCm(@_); });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults => {
                                                         ndu_force_giveback => $EventUtil::FALSE,
                                                         ndu_type => "",
                                                         revert_ontap_release => "",
                                                         install_ontap_architecture => "x86_64",
                                                         install_ontap_destination => "",
                                                         install_ontap_timeout => 7200,
                                                         install_ontap_build_options => ["optimize"],
                                                         install_ontap_release_web_server => "http://10.60.132.98/engineering",
                                                         controller_down_time => "300"
                                                     }
                                                 );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "ndu_force_giveback",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::FALSE,
                                           -format => "0|1",
                                           -example => "-ndu_force_giveback => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to perform forced " .
                                                           "givebacks during the NDU procedure.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "ndu_type",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "minor|major",
                                           -example => "-ndu_type => 'major'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies what kind of NDU to perform. " .
                                                           "Only applicable to the event 'nduCm'");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "revert_ontap_release",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "8.1|8.0",
                                           -example => "-ndu_type => '8.1'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies what kind major ONTAP release to revert to");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }


    $self->EventUtil::document_event_param(-param => "install_ontap_architecture",
                                           -package => __PACKAGE__,
                                           -default => "x86_64",
                                           -format => "x86_64|x86",
                                           -example => "-install_ontap_architecture => 'x86_64'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies what architecture of ONTAP to install");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "install_ontap_destination",
                                           -package => __PACKAGE__,
                                           -default => "x86_64",
                                           -format => "URL or path to ontap release",
                                           -example => "-install_ontap_destination => '/x/eng/rlse/DOT/8.1/final/bedrock/export/'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies what ONTAP to install");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "install_ontap_timeout",
                                           -package => __PACKAGE__,
                                           -default => "7200",
                                           -format => "time in seconds",
                                           -example => "-install_ontap_timeout => 10000",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the timeout to set when waiting for ONTAP to install");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "install_ontap_build_options",
                                           -package => __PACKAGE__,
                                           -default => "['normal']",
                                           -format => "array reference of build options e.g 'normal','debug','ibm','optimize', etc...",
                                           -example => "-install_ontap_destination => ['optimize','ibm']",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies what build options to select from whent installing ONTAP");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "install_ontap_release_web_server",
                                           -package => __PACKAGE__,
                                           -default => "http://10.60.132.98/engineering",
                                           -format => "URL to a web server that hosts ontap releases",
                                           -example => "-install_ontap_destination => 'http://eng-web.rtp.netapp.com/engineering'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the server to download ONTAP from.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    revertCm()
# Objective: Revert Downgrade for all c-mode resources
# Details:   See POD documentation at the end of this file.
###############################################################################
sub revertCm {
    my ($self, %args) = @_;

    # install ONTAP on every unfaulted node using NACL
    $self->install_ontap_on_all_nodes();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # revert the nodes
    my (@nodes_ci) = $self->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Error searching for NACL node objects");
        return $EventUtil::FALSE;
    }

    try {
        NACL::MTask::Revert->revert(command_interface => $nodes_ci[0],
                                    revert_to => $self->EventUtil::get_arg(-arg => "revert_ontap_release"));
    } otherwise {
        my ($e) = shift;
        $self->_set_err("Error while reverting: " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Event Method:    nduCm()
# Objective: Non-Disruptive Upgrade for all c-mode resources
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nduCm {
    my ($self, %args) = @_;

    # install the new version of ontap on all the nodes
    $self->install_ontap_on_all_nodes();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # determine what type of NDU we are going to do
    my $allow_version_mismatch;
    if ($self->EventUtil::get_arg(-arg => "ndu_type") =~ /major/i) {
        $allow_version_mismatch = $EventUtil::TRUE;
    } else {
        $allow_version_mismatch = $EventUtil::FALSE;
    }

    # perform takegives on all HA pairs using sfo_takegive_abstracted();
    $self->sfo_takegive_abstracted(-allow_version_mismatch => $allow_version_mismatch,
                                   -forcegive => $self->EventUtil::get_arg(-arg => "ndu_force_giveback"),
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # make sure ontap was installed properly
    $self->compare_current_ontap_images();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    disruptive_upgradeCm()
# Objective: Disruptive Upgrade for all c-mode resources
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disruptive_upgradeCm {
    my ($self, %args) = @_;

    # install the new version of ontap on all the nodes
    $self->install_ontap_on_all_nodes();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # bounce all the controllers (one at a time...)
    $self->simultaneous_node_reboot();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # reconnect to all nacl nodes (assume a version change occurred)
    $self->EventUtil::reconnect_all_nacl_nodes(no_version_change => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # make sure ontap was installed properly
    $self->compare_current_ontap_images();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    install_ontap_on_all_nodes()
# Objective: install ONTAP on all Nodes
###############################################################################
sub install_ontap_on_all_nodes {
    my ($self) = @_;

    # install ONTAP on every unfaulted node using NACL
    my (@nodes_ci) = $self->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Error searching for NACL node objects");
        return $EventUtil::FALSE;
    }

    # determine which version of ONTAP are currently installed
    my @image_state_objs = $self->fetch_current_ontap_images(-nodes_ci => \@nodes_ci);
    if ($self->error()) {
        $self->_set_err("Failed to obtain current ontap images");
        return $EventUtil::FALSE;
    }

    # update the ResMgr (so we can compare the what version is installed post upgrade)
    $self->update_current_ontap_images(-nodes_ci => \@nodes_ci,
                                       -image_state_objs => \@image_state_objs);
    if ($self->error()) {
        $self->_set_err("Failed to update the ontap image stats in the ResMgr");
        return $EventUtil::FALSE;
    }

    # get the next padded runid from the event object
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $padded_runid = $event_obj->increment_subtest_runid_counter();

    my @subtests;

    # install ontap across the nodes in parallel
    foreach my $node_ci (@nodes_ci) {
        my $nodename = $node_ci->hostrec()->hostname();

        my $subtest = subtest(
            sub {
                $self->install_ontap_on_node(-node_ci => $node_ci);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }

                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            },
            -runid => "${padded_runid}_install_ontap_on_node_${nodename}", -bg,
            "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

        push(@subtests, $subtest);
    }

    # reap the install subtests
    $self->wait_for_subtests(-subtests => \@subtests);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }


    return $EventUtil::TRUE;
}

###############################################################################
# Method:    install_ontap_on_node()
# Objective: install ONTAP on a single node
###############################################################################
sub install_ontap_on_node {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node_ci = &ParamsMgr::get_arg(-arg => "node_ci",
                                      -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # obtain the destination and build types
    my $ontap_destination = $self->EventUtil::get_arg(-arg => "install_ontap_destination");
    my @build_types = @{$self->EventUtil::get_arg(-arg => "install_ontap_build_options",
                                                  -default => [])};
    my $web_server = $self->EventUtil::get_arg(-arg => "install_ontap_release_web_server");
    my $arch = $self->EventUtil::get_arg(-arg => "install_ontap_architecture");
    my $install_ontap_timeout = $self->EventUtil::get_arg(-arg => "install_ontap_timeout");

    my $package_url;

    # if its not a URL, build out the URL for them
    if ($ontap_destination !~ /^http:\/\//) {

        # get rid of trailing spaces
        $ontap_destination =~ s/\s+$//;

        # build out the URL package
        $package_url = $web_server . $ontap_destination;

        # if the package url doesn't  already specify  a tgz file at this point, the add it dynamically
        if ($package_url !~ /\.tgz$/) {

            # add a trailing slash if there isnt one
            if ($package_url !~ /\/\s*$/) {
                $package_url .= "/";
            }

            my $build = $arch;

            # if requesting a normal build type, then clean out the build types array
            if (grep(/normal/, @build_types) and scalar(@build_types) == 1) {
                @build_types = ();
            }

            # optimize and debug are mutually exclusive
            if (grep(/optimize/, @build_types)) {
                $build .= ".optimize";
            } elsif (grep(/debug/, @build_types)) {
                $build .= ".debug";
                if (grep(/sim/, @build_types)) {
                    $build .= ".sim";
                }
            } elsif (grep(/sim/, @build_types)) {
                $build .= ".sim";
            }

            # ibm rebrand option comes last
            if (grep(/ibm/, @build_types)) {
                $build .= ".ibm";
            }

            # build the final url
            if ($package_url =~ /\/promo\//) {
                # promo builds
                $package_url .= $build . "/image.tgz";
            } else {
                # non-promo builds
                $package_url .= $build . "/tarball/image.tgz";
            }
        }
    } else {
        # otherwise, its a URL
        $package_url = $ontap_destination;
    }

    &SanUtil::nate_log(-comment, "Installing ONTAP ($package_url) on node " . $node_ci->name());

    # install the version of ONTAP
    try {
        my $update_obj = NACL::STask::SystemNodeImage->update(command_interface => $node_ci,
                                                              node              => $node_ci->hostrec()->hostname(),
                                                              package           => $package_url,
                                                              'method-timeout'  => $install_ontap_timeout);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught during SystemNodeImage update: " . $e->text());
    };

    if ($self->error()) {
        return $EventUtil::FALSE;
    } else {
        return $EventUtil::TRUE;
    }
}

###############################################################################
# Method:    fetch_current_ontap_images()
# Objective: Determine what ontap versions are currently installed on all nodes
#            in scope.
###############################################################################
sub fetch_current_ontap_images {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes_ci = &ParamsMgr::get_arg(-arg => "nodes_ci",
                                       -ref_type => "ARRAY",
                                       -default => []);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    if (not scalar(@nodes_ci)) {
        (@nodes_ci) = $self->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Error searching for NACL node objects");
            return (wantarray()) ? () : [];
        }
    }

    # obtain the current version of ontap that is running on the systems
    my @image_states;
    try {
        @image_states = NACL::CS::SystemNodeImage->fetch(command_interface => $nodes_ci[0]);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while obtaining Image details", $e->text());
    };
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @image_states : \@image_states;
}

###############################################################################
# Method:    update_current_ontap_images()
# Objective: Update the current ontap images installed for the nodes in ResMgr
###############################################################################
sub update_current_ontap_images {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my @nodes_ci = &ParamsMgr::get_arg(-arg => "nodes_ci",
                                       -ref_type => "ARRAY",
                                       -default => []);
    my @image_state_objs = &ParamsMgr::get_arg(-arg => "image_state_objs",
                                               -ref_type => "ARRAY",
                                               -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    if (not scalar(@nodes_ci)) {
        (@nodes_ci) = $self->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Error searching for NACL node objects");
            return $EventUtil::FALSE;
        }
    }

    foreach my $node_ci (@nodes_ci) {
        my $node_name = $node_ci->hostrec()->hostname();
        # find the current running ontap image on this system
        my ($curr_image) = grep {
            $_->{node} =~ /^$node_name$/ and $_->{iscurrent} eq "true";
        } @image_state_objs;

        # delete the command interface aspect of the curr image object before we store it
        delete($curr_image->{command_interface});

        # store this image on the ResMgr for this resource
        $self->ResMgr::set_resource_attribute(-attribute => "current_ontap_image",
                                              -value => $curr_image,
                                              -name => $node_name,
                                              -type => "nacl_nodes_cm");
        if ($self->error()) {
            $self->_set_err("Failed to store current image object in ResMgr for $node_name");
            return $EventUtil::FALSE;
        }

        # display to the log what version of ONTAP is currently installed
        &SanUtil::nate_log(-comment => Dumper($curr_image));
    }

    return $EventUtil::FALSE;
}

###############################################################################
# Method:    compare_current_ontap_images()
# Objective: Compare the current ontap images installed for the nodes in ResMgr
#            against the ontap images passed in. If any of them compare to true
#            than we didnt successfully install.
###############################################################################
sub compare_current_ontap_images {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my @nodes_ci = &ParamsMgr::get_arg(-arg => "nodes_ci",
                                       -ref_type => "ARRAY",
                                       -default => []);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    if (not scalar(@nodes_ci)) {
        (@nodes_ci) = $self->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Error searching for NACL node objects");
            return $EventUtil::FALSE;
        }
    }

    # collect the current_ontap image stats for all the nodes
    my @current_image_state_objs = $self->fetch_current_ontap_images(-nodes_ci => \@nodes_ci);
    if ($self->error()) {
        $self->_set_err("Failed to obtain current ontap images");
        return $EventUtil::FALSE;
    }

    my $return = 0;

    foreach my $node_ci (@nodes_ci) {
        my $node_name = $node_ci->hostrec()->hostname();
        # find the current running ontap image on this system
        my ($curr_image_state) = grep {
            $_->{node} =~ /^$node_name$/ and $_->{iscurrent} eq "true";
        } @current_image_state_objs;

        # delete the command interface aspect of the curr image object before we store it
        delete($curr_image_state->{command_interface});

        # store this image on the ResMgr for this resource
        my ($old_image_state) = $self->ResMgr::get_resource_attribute(-attribute => "current_ontap_image",
                                                                      -name => $node_name,
                                                                      -type => "nacl_nodes_cm");
        if ($self->error()) {
            $self->_set_err("Failed to obtain the old image object in ResMgr for $node_name");
            return $EventUtil::FALSE;
        }

        # store this image on the ResMgr for this resource
        $self->ResMgr::set_resource_attribute(-attribute => "current_ontap_image",
                                              -value => $curr_image_state,
                                              -name => $node_name,
                                              -type => "nacl_nodes_cm");
        if ($self->error()) {
            $self->_set_err("Failed to store current image object in ResMgr for $node_name");
            return $EventUtil::FALSE;
        }

        # compare $old_image_state and $curr_image_state
        if (&Data::Compare::Compare($old_image_state, $curr_image_state)) {
            # then they are identical, warn that the ontap image did not change on this node
            &SanUtil::nate_log('-warn', "ONTAP Image on $node_name did not change after the upgrade\nPrevious Version:" .
                               Dumper($old_image_state) . "\nCurrent Version:" . Dumper($curr_image_state));
            $return++;
        } else {
            # display to the log what version of ONTAP is currently installed
            &SanUtil::nate_log(-comment => Dumper($curr_image_state));

        }
    }

    # return true or false if the compare succeeded or failed
    if ($return != 0) {
        return $EventUtil::TRUE;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _validate_nduCm()
# Objective: Validate parameters into this event
###############################################################################
sub _validate_nduCm {
    my ($self) = @_;

    # make sure that ndu_type was specified
    my $ndu_type = $self->EventUtil::get_arg(-arg => "ndu_type");
    if ($ndu_type !~ /^(major|minor)$/i) {
        $self->_set_err("Invalid value for Event Parameter: 'ndu_type' ($ndu_type). Must be either 'major' or 'minor'");
        return $EventUtil::FALSE;
    }

    # make sure that ndu_force_giveback is either 0 or 1
    my $ndu_force_giveback = $self->EventUtil::get_arg(-arg => "ndu_force_giveback");
    if ($ndu_force_giveback !~ /^(1|0)$/i) {
        $self->_set_err("Invalid value for Event Parameter: 'ndu_force_giveback' ($ndu_force_giveback). Must be either '1' or '0'");
        return $EventUtil::FALSE;
    }

    $self->_validate_commonCm();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_revertCm()
# Objective: Validate parameters into this event
###############################################################################
sub _validate_revertCm {
    my ($self) = @_;

    # make sure that revert_ontap_release was specified
    my $revert_ontap_release = $self->EventUtil::get_arg(-arg => "revert_ontap_release");
    if ($revert_ontap_release !~ /^(8\.[0-9]|8\.[1-9][0-9]*)$/i) {
        $self->_set_err("Invalid value for Event Parameter: 'revert_ontap_release' ($revert_ontap_release). " .
                        "It must be a valid major ONTAP release in the 8.0 family Eg. (8.0, 8.1, 8.2, etc...)'");
        return $EventUtil::FALSE;
    }

    $self->_validate_commonCm();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_commonCm()
# Objective: Validate Common parameters for all events
###############################################################################
sub _validate_commonCm {
    my ($self) = @_;

    # make sure that install_ontap_architecture is either x86_64 or x86
    my $install_ontap_architecture = $self->EventUtil::get_arg(-arg => "install_ontap_architecture");
    if ($install_ontap_architecture !~ /^(x86|x86_64)$/i) {
        $self->_set_err("Invalid value for Event Parameter: 'install_ontap_architecture' ($install_ontap_architecture). Must be either 'x86' or 'x86_64'");
        return $EventUtil::FALSE;
    }

    # make sure that install_ontap_destination was specified
    my $install_ontap_destination = $self->EventUtil::get_arg(-arg => "install_ontap_destination");
    if (not defined($install_ontap_destination) or not length($install_ontap_destination)) {
        $self->_set_err("Event Parameter 'install_ontap_destination' is a required parameter.");
        return $EventUtil::FALSE;
    }

    # make sure that install_ontap_release_web_server was specified
    my $install_ontap_release_web_server = $self->EventUtil::get_arg(-arg => "install_ontap_release_web_server");
    if (not defined($install_ontap_release_web_server) or not length($install_ontap_release_web_server)) {
        $self->_set_err("Event Parameter 'install_ontap_release_web_server' is a required parameter.");
        return $EventUtil::FALSE;
    }

    # make sure install_ontap_build_options is empty or contains only valid entries
    # normal, optimize, ibm, debug, sim
    my $install_ontap_build_options = $self->EventUtil::get_arg(-arg => "install_ontap_build_options",
                                                                -default => []);
    if (ref($install_ontap_build_options) eq "ARRAY") {
        foreach my $opt (@{$install_ontap_build_options}) {
            if ($opt !~ /^(normal|optimize|debug|sim|ibm)$/) {
                $self->_set_err("Event Parameter 'install_ontap_build_options' ($opt), valid options are 'normal', 'optimize', 'debug', 'sim', 'ibm'");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}


1;

=pod

=head1 Package Name

NodeUpgradeEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for CMode Controller ONTAP Upgrade Events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::SFOEvents

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::ControllerValidation

=item EventMan::UtilityModules::PortValidation

=item EventMan::UtilityModules::PathValidation

=item EventMan::UtilityModules::SRAMDetection

=item EventMan::UtilityModules::SysLogging

=item EventMan::UtilityModules::SysStats

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-ndu_force_giveback>

=over 4

Description: Specifies whether to perform a forced giveback or not while
performing the NDU procedure.

Per Resource Capable: No

Default: 0 (off)

=back

=head2 I<-ndu_type>

=over 4

Description: Specifies what type of NDU to perform: 'minor' or 'major'.

Per Resource Capable: No

Required by Events: nduCm

=back

=head2 I<-controller_down_time>

=over 4

Description: Specifies how long to wait in between takeovers/givebacks during the NDU procedure

Per Resource Capable: No

Default: 300 seconds (5 minutes)

=back

=head2 I<-revert_ontap_release>

=over 4

Description: Specifies what major ONTAP release to revert to

Examples: 8.2, 8.1, 8.0,

Per Resource Capable: No

Required by Events: revertCm

=back

=head2 I<-install_ontap_architecture>

=over 4

Description: Specifies the Controller architecture of the cluster to upgrade.

Per Resource Capable: No

Default: x86_64

=back

=head2 I<-install_ontap_destination>

=over 4

Description: Specifies the path to the ONTAP build to install on the cluster.
This can either be a build path via /x/eng/rlse/... or an HTTP link to a image.tgz file.
If specifiying a UNIX path and not an HTTP link, the path must be to the directory
where the architecutre and build subdirectories are. In other words the directory must
contain the subdirectories that look like this:
   x86_64.optimize/
   x86_64.debug/
   x86_64.ibm/
   x86/
   x86_64/
   etc...
For NON-promoted: this is typically found in the final/bedrock/export/ subdirectory
For Promoted Builds: this is found in the promo/{release}/ subdirectory
A user can also specify the UNIX path directly to an image.tgz file.

Per Resource Capable: No

Required by Events: major_installCm, minor_installCm, installCm, disruptive_upgradeCm, revertCm

Examples:
 install_ontap_destination => "/x/eng/rlse/DOT/8.1/final/bedrock/export/" (non-promo build)
 install_ontap_destination => "/x/eng/rlse/DOT/8.1/promo/8.1/"            (promo build)
 install_ontap_destination => "/u/hrivera/mybuild/final/bedrock/export"   (custom user build)
 install_ontap_destination => "/u/hrivera/image.tgz"                      (custom user image.tgz file)
 install_ontap_destination => "http://somehost.netapp.com/image.tgz"      (custom http image.tgz link)

=back

=head2 I<-install_build_options>

=over 4

Description: Controls which options to apply to the build installation. Must be
An array reference of build options. e.g: ["optimize","ibm"] will install an
IBM rebranded optimized build.

Per Resource Capable: No

Default: ["optimize"]

=back

=head2 I<-install_ontap_release_web_server>

=over 4

Description: The web server that hosts the ontap build destination path.

Per Resource Capable: No

Default: http://10.60.132.98/engineering (http://eng-web.rtp.netapp.com/engineering)

=back

=head2 I<-install_ontap_timeout>

=over 4

Description: Specifies the timeout to use when installing ONTAP on the nodes.

Per Resource Capable: No

Default: 7200 (seconds)

=back

=head1 B<Event Keywords>

=head2 B<nduCm>

Performs a Non-Disruptive Upgrade based on the Event Parameter (ndu_type)

=over 4

=item Event Module Inheritance

default

=item Resources

=over

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resources to
operate against.

Resource Count: An even numbered list of Nodes, of size >= 2.

Default: all CMode nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=back

=item Required Event Parameters

=over

=item *I<-ndu_type>

=item *I<-install_ontap_destination>

=back

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nduCm",
                         -ndu_type => "minor",
                         -install_ontap_destination => "/x/eng/rlse/DOT/RrollingrockN/final/bedrock/export",
                         -install_ontap_build_options => ["optimize"],
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<minor_nduCm>

Performs a Minor Non-Disruptive Upgrade against C-mode Nodes.

=over 4

=item Event Module Inheritance

default

=item Resources

=over

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode nacl_nodes_cm resources to
operate against.

Resource Count: An even numbered list of Nodes, of size >= 2.

Default: all CMode nacl_nodes_cm in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=back

=item Required Event Parameters

=over

=item *I<-install_ontap_destination>

=back

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "minor_nduCm",
                         -install_ontap_destination => "/x/eng/rlse/DOT/RrollingrockN/final/bedrock/export",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41", "fas3070-rtp42", "fas3070-rtp43"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<major_nduCm>

Performs a Major Non-Disruptive Upgrade against C-mode Nodes.

=over 4

=item Event Module Inheritance

default

=item Resources

=over

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resources to
operate against.

Resource Count: An even numbered list of Nodes, of size >= 2.

Default: all CMode nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=back

=item Required Event Parameters

=over

=item *I<-install_ontap_destination>

=back

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "major_nduCm",
                         -install_ontap_destination => "/x/eng/rlse/DOT/RsierranevadaN/final/bedrock/export",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41", "fas3070-rtp42", "fas3070-rtp43"]);
 if ($EventManobj->error()) {
     ...
 }

=back


=head2 B<disruptive_upgradeCm>

Performs a Disruptive upgrade against C-Mode Nodes.

=over 4

=item Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerEvents::CMode::DisruptiveNodeEvents

=item EventMan::EventModules::ControllerEvents::CMode

=item EventMan::EventModules::ControllerEvents

=item EventMan::EventModules

=back

=item Resources

=over

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resources to
operate against.

Resource Count: An even numbered list of Nodes, of size >= 2.

Default: all CMode nodes in the Event scope will be used

=back

=item Required Event Parameters

=over

=item *I<-install_ontap_destination>

=back

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "disruptive_upgradeCm",
                         -install_ontap_destination => "/x/eng/rlse/DOT/RsierranevadaN/final/bedrock/export",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41", "fas3070-rtp42", "fas3070-rtp43"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<revertCm>

Performs a revert of ONTAP on a cluster of nodes.

=over 4

=item Event Module Inheritance

default

=item Resources

=over

=item *I<-nacl_nodes_cm>

Required: An array reference of a CMode node resources to
operate against.

Resource Count: An even numbered list of Nodes, of size >= 2.

Default: all CMode nodes in the Event scope will be used

=item Required Event Parameters

=over

=item *I<-revert_ontap_release>

=item *I<-install_ontap_destination>

=back

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "revertCm",
                         -revert_ontap_release => "8.1",
                         -install_ontap_destination => "/x/eng/rlse/DOT/RrollingrockN/final/bedrock/export",
                         -install_ontap_build_options => ["optimize"],
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head1 B<Public Methods>

=head2 B<install_ontap_on_all_nodes()>

Installs ONTAP on every node (in scope) in parallel

=over 4

=item Inputs

None

=item Examples

 # Install ONTAP on every node in scope
 $EventManObj->install_ontap_on_all_nodes();
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<install_ontap_on_node(%args)>

Installs ONTAP on a specific node

=over 4

=item Inputs

=item *I<-node_ci>

Required: A NACL::C::Node object

=item Examples

 # Install ontap on a single Node
 ($node) = NACL::C::Node->find();
 $EventManObj->install_ontap_on_node(-node_ci => $node);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<fetch_current_ontap_images(%args)>

Obtain the current ONTAP image objects

=over 4

=item Inputs

=item *I<-nodes_ci>

Optional: An array reference of NACL::C::Node objects

=item Examples

 # get the image states for all nodes in scope
 my @image_states = $EventManObj->fetch_current_ontap_images();
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<update_current_ontap_images(%args)>

Store SystemNodeImage objects for each Node in the ResMgr

=over 4

=item Inputs

=item *I<-nodes_ci>

Optional: An array reference of NACL::C::Node objects

=item *I<-image_state_objs>

Optional: An array reference of NACL::CS::SystemNodeImage objects

=item Examples

 # store the image state objs
 $EventManObj->update_current_ontap_images(-image_state_objs => \@objs);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<compare_current_ontap_images(%args)>

Compare current SystemNodeImage objects for each Node with those stored in the ResMgr

=over 4

=item Inputs

=item *I<-nodes_ci>

Optional: An array reference of NACL::C::Node objects

=item *I<-image_state_objs>

Optional: An array reference of NACL::CS::SystemNodeImage objects

=item Examples

 # store the image state objs
 if (not $EventManObj->compare_current_ontap_images(-image_state_objs => \@objs)) {
     if ($EventManObj->error()) {
         # handle case where eventman encountered error
     } else {
         # handle case where compare failed
     }
 } else {
     # compare succeeded
 }

=back

=cut
