# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for 60 Second SLA Events
## @author derricks@netapp.com
## @status private
## @pod here

# $Id:$

package EventMan::EventModules::ControllerEvents::CMode::SixtySecondSLA;

use Tharn;
use San::SanUtil;
use ParamsMgr;
use Tie::IxHash;
use Data::Dumper;
use FileHandle;
use DirHandle;
use Date::Manip;
use strict;
use warnings;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name              => "sixty_second_sla_Cm",
                                     -event_modules           => \@event_modules,
                                     -description             => "60-Second SLA Event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -resource_counts         => { nacl_nodes_cm => "2+" },
                                     -validate                => sub { &_validate_sixty_second_sla_Cm(@_) },
                                     -exec                    => sub { &sixty_second_sla_Cm(@_) });

    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { sla_io_delay               => 1200,
                                                           sla_cycles                 => 10,
                                                           sla_enable_dedup           => 0,
                                                           sla_active_dedup_ops       => 0,
                                                           sla_enable_compression     => 0,
                                                           sla_active_compression_ops => 0,
                                                           sla_inline_compression     => 0,
                                                           sla_fobench_title          => "SAN_Interop_60_Second_SLA",
                                                           sla_fobench_parse_path     => "/x/eng/rtpperf/fo_perf/CFO-TO-GB-Measurements/CFO_DATA",
                                                           sla_fobench_destination    => "",
                                                           sla_fobench_type           => "",
                                                           sla_fobench_prefix         => "nate",
                                                           sla_giveback_type          => "forced",
                                                           sla_takeover_type          => "nvpanic",
                                                           sla_io_type                => "fcp",
                                                           sla_ontap_version          => "8.1",
                                                           sla_panic_threshold_max    => "95",
                                                           sla_panic_threshold_min    => "95",
                                                           sla_client_os              => "mixed",
                                                         }
                                                     );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    # document each parameter used in this module
    $self->EventUtil::document_event_param(-param => "sla_io_delay",
                                           -package => __PACKAGE__,
                                           -default => "1200",
                                           -format => "integer in seconds",
                                           -example => "-sla_io_delay => 600",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Sets the amount of time in seconds to sleep prior to " .
                                               "beginning a takeover cycle",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_cycles",
                                           -package => __PACKAGE__,
                                           -default => "10",
                                           -format => "integer greater than 0",
                                           -example => "-sla_cycles => 6",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Sets the number of takeover/giveback cycles to perform.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_enable_dedup",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-sla_enable_dedup => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Turns dedup on or off during the test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_active_dedup_ops",
                                           -package => __PACKAGE__,
                                           -default => "4",
                                           -format => "integer between 0 and 8",
                                           -example => "-sla_active_dedup_ops => 5",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If dedup is enabled, sets the maximum number of dedup " .
                                                           "operations to perform in parallel.  The sum total of " .
                                                           "active dedup operations and active compression " .
                                                           "operations must be less than or equal to 8.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_enable_compression",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-sla_enable_compression => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Turns compression on or off during the test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_active_compression_ops",
                                           -package => __PACKAGE__,
                                           -default => "4",
                                           -format => "integer between 0 and 8",
                                           -example => "-sla_active_compression_ops => 5",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If compression is enabled, sets the maximum number of " .
                                                           "compression operations to perform in parallel.  The sum " .
                                                           "total of active dedup operations and active compression " .
                                                           "operations must be less than or equal to 8.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_inline_compression",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-sla_inline_compression => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Turns inline compression  on or off during the test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_fobench_title",
                                           -package => __PACKAGE__,
                                           -default => "SAN_Interop_60_Second_SLA",
                                           -format => "String without spaces",
                                           -example => "-sla_fobench_title => 'this_is_my_title'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Sets the title for the test SLA test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_fobench_parse_path",
                                           -package => __PACKAGE__,
                                           -default => "/x/eng/rtpperf/fo_perf/CFO-TO-GB-Measurements/CFO_DATA",
                                           -format => "Fully qualified directory",
                                           -example => "-sla_fobench_parse_path => '/fully/qualified/directory'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Sets the output directory for the parsed data",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_fobench_destination",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "Fully qualified directory",
                                           -example => "-sla_fobench_destination => '/x/eng/rtpperf/fo_perf/CFO-TO-GB-Measurements/auto-fas3170-03_04'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Sets the output directory for SLA test.  This should be " .
                                                           "benchmark testbed directory.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_fobench_type",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "String without spaces"
                                           -example => "-sla_fobench_type => 'san_interop'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Keyword to describe the sla test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_fobench_prefix",
                                           -package => __PACKAGE__,
                                           -default => "nate",
                                           -format => "String without spaces"
                                           -example => "-sla_fobench_prefix => 'my_test'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Prefix to prepend to all output files during the sla test",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_giveback_type",
                                           -package => __PACKAGE__,
                                           -default => "forced",
                                           -format => "forced or graceful"
                                           -example => "-sla_giveback_type => 'graceful'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Type of giveback to use during the sla test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_takeover_type",
                                           -package => __PACKAGE__,
                                           -default => "nvpanic",
                                           -format => "nvpanic, panic, graceful, rpower, panicd, or watchdog",
                                           -example => "-sla_takeover_type => 'panic'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Type of takeover to use during the sla test.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_io_type",
                                           -package => __PACKAGE__,
                                           -default => "fcp",
                                           -format => "fcp, iscsi, or nfs",
                                           -example => "-sla_io_type => 'iscsi'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Type of IO that is being run in the sla test",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_ontap_version",
                                           -package => __PACKAGE__,
                                           -default => "8.1",
                                           -format => "8.1, 8.1.0",
                                           -example => "-sla_ontap_version => '8.1'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "The version of ontap running on the controllers",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_panic_threshold_max",
                                           -package => __PACKAGE__,
                                           -default => "95",
                                           -format => "Percentage between 0 and 100",
                                           -example => "-sla_panic_threshold_max => 85",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "How full NV RAM should be (max), before we panic, " .
                                                           "nvpanic tests only",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_panic_threshold_min",
                                           -package => __PACKAGE__,
                                           -default => "95",
                                           -format => "Percentage between 0 and 100",
                                           -example => "-sla_panic_threshold_max => 85",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "How full NV RAM should be (min), before we panic, " .
                                                           "nvpanic tests only. Cannot be greater than " .
                                                           "sla_panic_threshold_min.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "sla_client_os",
                                           -package => __PACKAGE__,
                                           -default => "mixed",
                                           -format => "mixed|windows|linux",
                                           -example => "-sla_client_os => linux",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "OS that is running on the clients.  Valid values are " .
                                                           "mixed, windows, and linux",
                                           -resource_types => ["nacl_nodes_cm"]);


    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################
###############################################################################
# Event Method: sixty_second_sla_Cm()
# Objective: Perform 60-Second SLA Event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sixty_second_sla_Cm {
    my ($self) = @_;

    # collect cluster statistics
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # make sure auto giveback is disabled before starting the benchmark
    &SanUtil::nate_log("-comment", "Disabling auto giveback on all controllers");
    $self->toggle_auto_giveback(-state => "disable");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # make sure autosupport is disabled before starting the benchmark
    &SanUtil::nate_log("-comment", "Disabling Auto Support on all controllers");
    $self->disable_autosupport();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Save a time hack prior to starting the test.  We need to do this so
    # when we copy out the sla logs that we only copy out the logs for this
    # run.  Since the SLA io threads are run outside eventman and it is
    # possible that multiple 60 second sla events may be called with different
    # run options, we have to make sure that we grab the correct log entries.
    my $event_start_time = time();

    # Run the first pass with parsing turned off
    $self->_run_fo_bench(-run_type => "run_no_parse");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Copy the SLA logs from the TESTDIR
    $self->_copy_client_sla_logs(-start_time => $event_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # refresh nacl node connections after running the fo_bench tool
    # it's possible that the fo bench tool performed takeovers and givebacks
    # so our connections may have been destroyed.
    $self->EventUtil::reconnect_all_nacl_nodes(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate controllers
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "post_sla_test");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports
    $self->PortValidation::validate_ports_after_giveback(-id_extension => "post_sla_test");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate paths
    $self->PathValidation::validate_paths_after_giveback(-id_extension          => "post_sla_test",
                                                         -path_validation_rules => {validate_takeover_giveback => 1});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Run the second pass with parsing turned on
    $self->_run_fo_bench(-run_type => "parse_only");
    if ($self->error()) {
        # If the "parse_only" pass of fo_bench fails, we don't want the whole test to tank.  Print out a
        # warning instead of failing.
        &SanUtil::nate_log('-warn', "fo_bench parse only process failed.  See the log for details. " .
                                    "You can email 'dl-perf-cfo-measurement\@netapp.com' for assistance " .
                                    "parsing this test run.\n" . $self->errmsg());
    } else {
        &SanUtil::nate_log('-comment', "inject_fault.thpl: fo_bench 'parse only' process completed successfully.\n" .
                                       "Please email dl-perf-cfo-measurement\@netapp.com with your fo_bench " .
                                       "testbed name so that they can verify your test run\'s parsed data.");
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Event Method: _copy_client_sla_logs()
# Objective: Copy the client SLA logs from the testdir
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _copy_client_sla_logs {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_start_time = &ParamsMgr::get_arg(-arg      => "start_time",
                                               -default  => "",
                                               -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $destination = $self->EventUtil::get_arg(-arg => "sla_fobench_destination");
    my $prefix      = $self->EventUtil::get_arg(-arg => "sla_fobench_prefix");

    # Get the set of environment parameters from ParamsMgr
    my $logdir = &Tharn::param("TOP_LOGDIR");
    unless (defined($logdir) and $logdir ne "") {
        $self->_set_err("Unable to determine the nate TOP_LOGDIR");
        return $EventUtil::FALSE;
    }

    my $sla_dir = $logdir . "/SLA";

    # Make sure the sla directory exists
    unless (-d $sla_dir) {
        $self->_set_err("Client SLA Directory '$sla_dir' does not exist or is not accessible");
        return $EventUtil::TRUE;
    }

    # Create a dir handle for the sla log directory
    my $dh_sla_logs = new DirHandle;
    $dh_sla_logs->open("$sla_dir");

    unless (defined($dh_sla_logs)) {
        $self->_set_err("Unable to create a directory handle for '$sla_dir'");
        return $EventUtil::TRUE;
    }

    # Iterate over the log files and process each one.  We want to pull out
    # the output lines that are dated and create a new log file under the
    # destination.  The file name will be <prefix>.client_sla.<filename>.
    # Make sure you only copy in logs that started on or after $event_start_time.
    my @errors_found = ();
    while(defined(my $dir_item = $dh_sla_logs->read())) {
        my $client_name;
        if ($dir_item =~ /^(\S+)_availability_check/i) {
            $client_name = $1;
        } else {
            next;
        }

        my $full_log    = $sla_dir . "/" . $dir_item;
        my $output_file = $destination . "/" . $prefix . ".client_sla." . $client_name;

        &SanUtil::nate_log("-comment", "Processing SLA Log '$full_log' file for client '$client_name'");

        # Create file handles for the current log and the new output log
        my $fh_log = new FileHandle "$full_log";
        unless (defined($fh_log)) {
            $self->_set_err("Unable to open log file '$dir_item' for read");
            return $EventUtil::FALSE;
        }

        my $fh_output = new FileHandle "> $output_file";
        unless (defined($fh_log)) {
            $self->_set_err("Unable to open log file '$output_file' for write");
            return $EventUtil::FALSE;
        }

        # Parse through the output and look for the timestamped lines that look like:
        #
        # [auto-rhel5u6-02:linux:Fri Sep 16 20:32:08 2011] Avail check on LUN /mnt/qa/n_vg_f/nvol-1-f/auto-rhel5u6-02/sla_nate_test took 68 seconds
        while (my $line = $fh_log->getline()) {
            chomp($line);
            if ($line =~ /^\[\S+\:\S+\:(.+)\]\s+Avail\s+.*took (\d+) seconds/i) {
                # Grab the time, convert it to seconds since the epoch and compare it to
                # the event start time.
                my $date_time = $1;
                my $seconds = $2;
                my $epoch_time = Date::Manip::UnixDate(($date_time), "%s");
                if ($epoch_time >= $event_start_time) {
                    # Looks like this line belongs to the current event
                    print $fh_output "$line\n";
                }

                # Check to see if the seconds are higher than 60
                if ($seconds > 60) {
                    push (@errors_found, $line);
                }
            }
        }
        $fh_log->close();
        $fh_output->close();
    }

    $dh_sla_logs->close();

    # Check to see if we found any errors (greater than 60 seconds) in this test run
    # We are only going to print a warning.  We don't want to fail the test.  This is more
    # to just bring it to the user's attention.
    if (scalar(@errors_found)) {
        my $errmsg = join("\n", @errors_found);
        unshift(@errors_found, "The following clients were found to have outages greater than 60 seconds " .
                               "during SLA log processing:\n");
        &SanUtil::nate_log("-warn", $errmsg);
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _verify_fo_bench_config()
# Objective: Verify that the fo_config.reg file exists and the version matches
#            the version in fo_bench.pl
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _verify_fo_bench_config {
    my ($self) = @_;

    my $options_hash = {};

    # Grab the destination
    $options_hash->{'destination'} = $self->EventUtil::get_arg(-arg => "sla_fobench_destination");

    # define the scripts directory
    my $scripts_dir  = $options_hash->{'destination'} . "/scripts_default/fo_bench";

    # Define the fo_bench command and the fo_config.reg
    my $fo_bench_cmd  = "fo_bench.pl";
    my $fo_config_reg = "fo_config.reg";

    # Make sure the scripts directory exists
    unless (-d "$scripts_dir") {
        $self->_set_err("Scripts directory '$scripts_dir' doesn't exist");
        return $EventUtil::FALSE;
    }

    # Change into the destination directory
    chdir $scripts_dir;

    # Make sure the fo_bench script exists
    unless (-f "$fo_bench_cmd") {
        $self->_set_err("Unable to find file '$fo_bench_cmd'");
        return $EventUtil::FALSE;
    }

    # Make sure the fo_config.reg file exists
    unless (-f "$fo_config_reg") {
        $self->_set_err("Unable to find file '$fo_config_reg'");
        return $EventUtil::FALSE;
    }

    # Get the fo_bench script version
    my @fo_bench_output = `./$fo_bench_cmd --help`;
    my $fo_bench_vers = "";
    foreach my $line (@fo_bench_output) {
        if ($line =~ /^FO Bench -- Version (\d+):/i) {
            $fo_bench_vers = $1;
            last;
        }
    }

    # Make sure we found the version
    if ($fo_bench_vers eq "") {
        $self->_set_err("Unable to determine the value of 'fo_bench_vers' in the script '$fo_bench_cmd'");
        return $EventUtil::FALSE;
    }

    # Open the config read only
    my $fh_foconfig = new FileHandle "$fo_config_reg";
    unless(defined($fh_foconfig)) {
        $self->_set_err("Unable to open file '$fo_config_reg' for read");
        return $EventUtil::FALSE;
    }

    # Process the fo_config.reg file and look for the "fo_bench_vers =" line
    my $fo_config_vers = "";
    while (my $line = $fh_foconfig->getline()) {
        chomp($line);
        if ($line =~ /^fo_bench_vers = (\S+)$/) {
            $fo_config_vers = $1;
            last;
        }
    }
    $fh_foconfig->close();

    # Make sure we found the config version
    if ($fo_config_vers eq "") {
        $self->_set_err("Unable to determine the value of 'fo_bench_vers' in the config file '$fo_config_reg'");
        return $EventUtil::FALSE;
    }

    # Make sure the two versions match
    unless ($fo_bench_vers eq $fo_config_vers) {
        $self->_set_err("Version '$fo_bench_vers' found in the script file '$fo_bench_cmd' does not match " .
                        "version '$fo_config_vers' found in the config file '$fo_config_reg'");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _run_fo_bench()
# Objective: Run the fo_bench script
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _run_fo_bench {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $run_type = &ParamsMgr::get_arg(-arg      => "run_type",
                                       -default  => "run_parse",
                                       -format   => qr/^(run_parse|run_no_parse|parse_only)$/);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Define the fo_bench command
    my $fo_bench_cmd = "fo_bench.pl";

    # Start loading up all of the command line options in a hash
    my $options_hash = {};

    # Set options that are set based on run_type
    if ($run_type eq "parse_only") {
        # Options if we are only parsing
        $options_hash->{'parse'}      = "0";
        $options_hash->{'mp_parse'}   = "1";
        $options_hash->{'parse_only'} = "1";
        $options_hash->{'parse_exec'} = "cfo_parse";
        $options_hash->{'parse_path'} = $self->EventUtil::get_arg(-arg => "sla_fobench_parse_path");
    } else {
        # These run types involve actually running the benchmark
        if ($run_type eq "run_no_parse") {
            # Options if we are running the bench test but not
            # parsing the output
            $options_hash->{'parse'}          = "0";
            $options_hash->{'mp_parse'}       = "0";
            $options_hash->{'parse_only'}     = "0";
        } else {
            # Full bench test with parsing at the end
            $options_hash->{'parse'}      = "1";
            $options_hash->{'mp_parse'}   = "1";
            $options_hash->{'parse_only'} = "0";
            $options_hash->{'parse_exec'} = "cfo_parse";
            $options_hash->{'parse_path'} = $self->EventUtil::get_arg(-arg => "sla_fobench_parse_path");
        }
    }

    # Options that have nothing to do with run type
    $options_hash->{'reboot'}         = "0";
    $options_hash->{'client'}         = "0";
    $options_hash->{'fork'}           = "0";
    $options_hash->{'install_ontap'}  = "0";
    $options_hash->{'cli_mode'}       = "c";
    $options_hash->{'console'}        = "0";
    $options_hash->{'mpath'}          = "1";
    $options_hash->{'panic_fallback'} = "600";
    $options_hash->{'max_event'}      = "900";

    # Pull the rest of the options from ParamsMgr
    $options_hash->{'cycles'}                     = $self->EventUtil::get_arg(-arg => "sla_cycles");
    $options_hash->{'io_delay'}                   = $self->EventUtil::get_arg(-arg => "sla_io_delay");
    $options_hash->{'compression_enabled'}        = $self->EventUtil::get_arg(-arg => "sla_enable_compression");
    $options_hash->{'active_compression_ops'}     = $self->EventUtil::get_arg(-arg => "sla_active_compression_ops");
    $options_hash->{'inline_compression_enabled'} = $self->EventUtil::get_arg(-arg => "sla_inline_compression");
    $options_hash->{'dedup_enabled'}              = $self->EventUtil::get_arg(-arg => "sla_enable_dedup");
    $options_hash->{'active_dedup_ops'}           = $self->EventUtil::get_arg(-arg => "sla_active_dedup_ops");
    $options_hash->{'panic_threshold_max'}        = $self->EventUtil::get_arg(-arg => "sla_panic_threshold_max");
    $options_hash->{'panic_threshold_min'}        = $self->EventUtil::get_arg(-arg => "sla_panic_threshold_min");
    $options_hash->{'giveback_type'}              = $self->EventUtil::get_arg(-arg => "sla_giveback_type");
    $options_hash->{'takeover_type'}              = $self->EventUtil::get_arg(-arg => "sla_takeover_type");
    $options_hash->{'ontap_version'}              = $self->EventUtil::get_arg(-arg => "sla_ontap_version");
    $options_hash->{'io_type'}                    = $self->EventUtil::get_arg(-arg => "sla_io_type");
    $options_hash->{'client_os'}                  = $self->EventUtil::get_arg(-arg => "sla_client_os");
    $options_hash->{'prefix'}                     = $self->EventUtil::get_arg(-arg => "sla_fobench_prefix");
    $options_hash->{'type'}                       = $self->EventUtil::get_arg(-arg => "sla_fobench_type");
    $options_hash->{'title'}                      = $self->EventUtil::get_arg(-arg => "sla_fobench_title");
    $options_hash->{'destination'}                = $self->EventUtil::get_arg(-arg => "sla_fobench_destination");

    # Some additional command line options to keep the fo_bench.pl script from
    # having uninitialized variables.
    $options_hash->{'upgrade_version'}   = $self->EventUtil::get_arg(-arg => "sla_ontap_version");
    $options_hash->{'downgrade_version'} = $self->EventUtil::get_arg(-arg => "sla_ontap_version");

    # Convert any spaces to underscores for title, prefix, and type
    $options_hash->{'prefix'} =~ s/\s+/_/g;
    $options_hash->{'title'}  =~ s/\s+/_/g;
    $options_hash->{'type'}   =~ s/\s+/_/g;

    # Dump the options
    &SanUtil::nate_log("-comment", "Global 60 Second SLA Parameters:\n" . Dumper($options_hash));

    # Combine all of the options in the hash into one massive option string to pass
    # in on the command line to fo_bench
    my $fo_bench_options = "";
    foreach my $key (sort keys %$options_hash) {
        $fo_bench_options = $fo_bench_options . " -" . $key . " " . $options_hash->{$key};
    }

    # We will now run the fo_bench script.  We need to do this via a plain old file
    # handle versus a nate exec call.  We need to write the output out to the stdout
    # as well as a file at the same time.  We also need to watch for the string "rstatus="
    # to be able to catch the return code.  We can't capture the output into an array
    # and process it since it is a huge amount of output and the parser that gets run as
    # part of the script expects to find the output file as well.

    # Define the scripts directory now that we have the destination
    my $scripts_dir  = $options_hash->{'destination'} . "/scripts_default/fo_bench";

    # Change into the destination directory
    chdir $scripts_dir;

    # Create a filehandle for the output.  This is dependent on whether
    # or not we are parsing in the run.  If we are not, we will write to parse_output
    # without a prefix.
    my $outfile;
    if ($run_type eq "parse_only") {
        $outfile = $options_hash->{'destination'} . "/parse_output";
    } else {
        $outfile = $options_hash->{'destination'} . "/" . $options_hash->{'prefix'} . ".test_output";
    }

    my $fh_out = new FileHandle "> $outfile";
    unless(defined($fh_out)) {
        $self->_set_err("Unable to open file '$outfile' for write");
        return $EventUtil::FALSE;
    }

    $fh_out->autoflush(1);

    # Create a filehandle for the command
    my $cmd     = "./" . ${fo_bench_cmd} . " " . $fo_bench_options . " --reg";
    my $fh_cmd  = new FileHandle "${cmd}; echo rstatus=\$?|";
    unless(defined($fh_cmd)) {
        $self->_set_err("Unable to run command '$cmd'");
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log("-comment", "Running command '$cmd'");

    # Loop over the command output and process the data
    my $test_id = "";
    while (my $line = $fh_cmd->getline()) {
        if ($line =~ /^rstatus=(\d+)$/) {
            if ($1 == 0) {
                # Looks like the test completed successfully.  Print out a message
                # and then run the validation routines.
                $fh_cmd->close();
                $fh_out->close();
                &SanUtil::nate_log("-comment", "Command '$fo_bench_cmd' completed successfully");
                if ($test_id ne "") {
                    # We found the test id in the output
                    my $output_message = "\nFound Test ID '$test_id' while parsing the output. To look at the results" .
                                         "\nusing the Performance Team's web tool, point a browser at the address" .
                                         "\nhttp://grover.rtp.netapp.com, click 'Lab Results' drop down menu on the left" .
                                         "\nand then choose 'FO Runid Search'.  Enter Runid '$test_id' in the Run ID" .
                                         "\nsearch box and hit the search button.";
                    &SanUtil::nate_log("-comment", $output_message);
                }
                return $EventUtil::TRUE;
            } else {
                # Looks like the test had issues print a message and also set error
                $fh_cmd->close();
                $fh_out->close();
                &SanUtil::nate_log("-comment", "Command '$fo_bench_cmd' failed");
                $self->_set_err("Command '$fo_bench_cmd' failed.  Check the output logs from the test");
                return $EventUtil::FALSE;
            }
        } elsif ($line =~ /Using test ID: (\S+)$/i) {
            # Grab the new test id so we can print it out in the log file.
            # It may have a period at the end so strip that off.
            $test_id = $1;
            $test_id =~ s/\.$//g;
        } else {
            # Print the output to stdout and to the fh_out filehandle.  We need to use print
            # versus nate logging mechanisms to handle this.
            print $fh_out $line;
            print $line;
        }
    }
    $fh_cmd->close();
    $fh_out->close();

    # If we made it out this far, then we didn't find the "rstatus=" line.  We have no
    # idea at this point if the test passed or failed so set an error and bail.
    $self->_set_err("Test ended and we didn't find the message flag 'rstatus=' in the output.  Check the " .
                    "output logs to see if there is any indication of what happened.");
    return $EventUtil::FALSE;
}


###############################################################################
# Event Method:    _validate_sixty_second_sla_Cm()
# Objective: Validate that parameters are set properly for the sixty_second_slaCm
#            Event.
###############################################################################
sub _validate_sixty_second_sla_Cm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $event_name = $event_obj->get_name();

    # Store any errors in an array.  We will print out an error summary
    # so the user doesn't have to run multiple times to find all problems.
    my @errors_found = ();

    my $sla_enable_compression     = $self->EventUtil::get_arg(-arg => "sla_enable_compression");
    my $sla_enable_dedup           = $self->EventUtil::get_arg(-arg => "sla_enable_dedup");
    my $sla_active_compression_ops = $self->EventUtil::get_arg(-arg => "sla_active_compression_ops");
    my $sla_active_dedup_ops       = $self->EventUtil::get_arg(-arg => "sla_active_dedup_ops");

    # Make sure that compression operations and dedup operations are in the range of 0-8
    unless ($sla_active_compression_ops =~ /[0-8]/) {
        push (@errors_found, "(Event: $event_name) 'sla_active_compression_ops' must be between 0 and 8");
    }

    unless ($sla_active_dedup_ops =~ /[0-8]/) {
        push (@errors_found, "(Event: $event_name) 'sla_active_dedup_ops' must be between 0 and 8");
    }

    # If both compression and dedup are enabled, make sure that the sum of the active compression
    # and active dedup operations is less than or equal to 8.  This is a requirement of the fo_bench.pl script.
    if ($sla_enable_compression and $sla_enable_dedup) {
        my $total_ops = $sla_active_compression_ops + $sla_active_dedup_ops;
        if ($total_ops > 8) {
            push (@errors_found, "(Event: $event_name) The sum of sla_dedup_operations and " .
                                 "sla_compression_operations must be less than or equal to 8");
        }
    }

    # Make sure that panic threshold min is not greater than panic threshold max
    my $sla_panic_threshold_max = $self->EventUtil::get_arg(-arg => "sla_panic_threshold_max");
    my $sla_panic_threshold_min = $self->EventUtil::get_arg(-arg => "sla_panic_threshold_min");

    if ($sla_panic_threshold_max < $sla_panic_threshold_min) {
        push (@errors_found, "(Event: $event_name) Panic threshold max $sla_panic_threshold_max' can " .
                             "not be less than panic threshold min '$sla_panic_threshold_min'");
    }

    # Make sure the parse path directory exists and is not a relative path
    my $sla_fobench_parse_path  = $self->EventUtil::get_arg(-arg => "sla_fobench_parse_path");

    if ($sla_fobench_parse_path eq "" or $sla_fobench_parse_path !~ /^\/.+/) {
        push (@errors_found, "(Event: $event_name) Invalid value for 'sla_fobench_parse_path'.  The directory " .
                             "must be fully qualified and cannot be '/'");
    } else {
        unless (-d $sla_fobench_parse_path) {
            push (@errors_found, "(Event: $event_name) 'sla_fobench_parse_path' directory '$sla_fobench_parse_path' " .
                                 "does not exist or is not accessible");
        }
    }

    # Make sure the destination directory exists and is not a relative path
    my $sla_fobench_destination = $self->EventUtil::get_arg(-arg => "sla_fobench_destination");

    if ($sla_fobench_destination eq "" or $sla_fobench_destination !~ /^\/.+/) {
        push (@errors_found, "(Event: $event_name) Invalid value for 'sla_fobench_destination'.  The directory " .
                             "must be fully qualified and cannot be '/'");
    } else {
        if (-d $sla_fobench_destination) {
            # Make sure the fo_bench.pl script exists and is executable
            my $scripts_dir = $sla_fobench_destination . "/scripts_default/fo_bench";
            unless (-x "$scripts_dir/fo_bench.pl") {
                push (@errors_found, "(Event: $event_name) Unable to find the fo_bench.pl script in the " .
                                     "directory '$scripts_dir'");
            }
        } else {
            push (@errors_found, "(Event: $event_name) 'sla_fobench_destination' directory " .
                                 "'$sla_fobench_destination' does not exist or is not accessible");
        }
    }

    # Make sure the value for sla_fobench_type is not empty
    my $sla_fobench_type = $self->EventUtil::get_arg(-arg => "sla_fobench_type");
    if ($sla_fobench_type eq "") {
        push (@errors_found, "(Event: $event_name) 'sla_fobench_type' cannot be an empty string");
    }

    # Make sure the value for sla_fobench_prefix is not empty
    my $sla_fobench_prefix = $self->EventUtil::get_arg(-arg => "sla_fobench_prefix");
    if ($sla_fobench_prefix eq "") {
        push (@errors_found, "(Event: $event_name) 'sla_fobench_prefix' cannot be an empty string");
    }

    # Make sure the value for sla_fobench_title is not empty
    my $sla_fobench_title = $self->EventUtil::get_arg(-arg => "sla_fobench_title");
    if ($sla_fobench_title eq "") {
        push (@errors_found, "(Event: $event_name) 'sla_fobench_title' cannot be an empty string");
    }

    # Make sure the value for sla_io_type is either fcp or iscsi
    my $sla_io_type = $self->EventUtil::get_arg(-arg => "sla_io_type");
    unless ($sla_io_type =~ /^(fcp|iscsi)$/) {
        push (@errors_found, "(Event: $event_name) 'sla_io_type' must be either 'fcp' or 'iscsi'");
    }

    # Make sure the value for sla_client_os is either windows, linux, or mixed
    my $sla_client_os = $self->EventUtil::get_arg(-arg => "sla_client_os");
    unless ($sla_client_os =~ /^(windows|linux|mixed)$/) {
        push (@errors_found, "(Event: $event_name) 'sla_client_os' must be either 'mixed', 'windows', or 'linux'");
    }

    # Make sure the version of the config file matches the version of fo_bench.pl
    $self->_verify_fo_bench_config();
    if ($self->error()) {
        push (@errors_found, $self->errmsg());
        $self->_clear_err();
    }

    # Check to see if we found any errors. If we did, print out all of the error messages and
    # then set error and return false.
    if (scalar(@errors_found)) {
        my $errmsg = join("\n", @errors_found);
        &SanUtil::nate_log("-warn", $errmsg);
        $self->_set_err("There were errors validating event '$event_name'");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

SixtySecondSLA.pm

=head1 Author

Derrick Shultz

=head1 Description

Event Module for Sixty Second SLA events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::SixtySecondSLA

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-sla_io_delay>

=over 4

Description: Sets the amount of time in seconds to sleep prior to beginning a takeover cycle

Per Resource Capable: No

Default: 1200 (seconds)

=back

=head2 I<-sla_cycles>

=over 4

Description: Sets the number of takeover/giveback cycles to perform

Per Resource Capable: No

Default: 10

=back

=head2 I<-sla_enable_dedup>

=over 4

Description: Turns dedup on or off during the test

Per Resource Capable: No

Default: 1

=back

=head2 I<-sla_active_dedup_ops>

=over 4

Description: If dedup is enabled, sets the maximum number of dedup
operations to perform in parallel.  The sum total of active dedup operations
and active compression operations must be less than or equal to 8.

Per Resource Capable: No

Default: 4

=back

=head2 I<-sla_enable_compression>

=over 4

Description: Turns compression on or off during the test

Per Resource Capable: No

Default: 1

=back

=head2 I<-sla_active_compression_ops>

=over 4

Description: If compression is enabled, sets the maximum number of compression
operations to perform in parallel.  The sum total of active dedup operations
and active compression operations must be less than or equal to 8.

Per Resource Capable: No

Default: 4

=back

=head2 I<-sla_inline_compression>

=over 4

Description: Turns inline compression on or off during the test

Per Resource Capable: No

Default: 1

=back

=head2 I<-sla_fobench_title>

=over 4

Description: Sets the title for the test SLA test

Per Resource Capable: No

Default: "SAN_Interop_60_Second_SLA"

=back

=head2 I<-sla_fobench_parse_path>

=over 4

Description: Sets the output directory for the parsed data

Per Resource Capable: No

Default: "/x/eng/rtpperf/fo_perf/CFO-TO-GB-Measurements/CFO_DATA"

=back

=head2 I<-sla_fobench_destination>

=over 4

Description: Sets the output directory for SLA test.  This should be benchmark testbed directory.

Per Resource Capable: No

Default: Empty String ("")

=back

=head2 I<-sla_fobench_type>

=over 4

Description: Keyword to describe the sla test

Per Resource Capable: No

Default: Empty String ("")

=back

=head2 I<-sla_fobench_prefix>

=over 4

Description: Prefix to prepend to all output files during the sla test

Per Resource Capable: No

Default: Empty String ("nate")

=back

=head2 I<-sla_giveback_type>

=over 4

Description: Type of giveback to use during the sla test.  Acceptable values
are "forced" and "graceful"

Per Resource Capable: No

Default: "forced"

=back

=head2 I<-sla_takeover_type>

=over 4

Description: Type of takeover to use during the sla test.  Acceptable values
are "nvpanic", "panic", "graceful", "rpower", "panicd", or "watchdog"

Per Resource Capable: No

Default: "nvpanic"

=back

=head2 I<-sla_io_type>

=over 4

Description: Type of IO that is being run in the sla test.  Acceptable values
are "fcp", "iscsi", or "nfs".

Per Resource Capable: No

Default: "fcp"

=back

=head2 I<-sla_ontap_version>

=over 4

Description: The version of ontap running on the controllers

Per Resource Capable: No

Default: "8.1"

=back

=head2 I<-sla_panic_threshold_max>

=over 4

Description: How full NV RAM should be (max), before we panic, nvpanic tests only.

Per Resource Capable: No

Default: 95

=back

=head2 I<-sla_panic_threshold_min>

=over 4

Description: How full NV RAM should be (min), before we panic, nvpanic tests only.

Per Resource Capable: No

Default: 95

=back

=head2 I<-sla_client_os>

=over 4

Description: OS that is running on the clients.  Valid values are mixed, windows, and linux.

Per Resource Capable: No

Default: "mixed"

=back

=head1 B<Event Keywords>

=head2 B<sixty_second_sla_Cm>

Performs the 60-Second SLA event for C-Mode.

=over 4

=item Event Parameters

None

=item Examples

 # see EventMan::add_event()
 # perform 60 second sla
 $EventManobj->add_event(-event_name => "sixty_second_sla_Cm");
 if ($EventManobj->error()) {
     ...
 }

=back

=cut
