# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for Volume Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::VolumeEvents;

use Data::Dumper;
use San::SanUtil;
use San::SanClient::ClientUtil;
use ParamsMgr;
use strict;
use warnings;
use NACL::STask::VolumeMove;
use NACL::CS::StorageAggregate;
use NACL::CS::Volume;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use Tharn;

my $Log = log_global();

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    # Volume Move Event Modules
    $self->EventUtil::register_event(-event_name => "vol_moveCm",
                                     -enabled => $EventUtil::FALSE,
                                     -event_modules => \@event_modules,
                                     -description => "This is a CMode event that performs parallel vol moves in the same cluster",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { "nacl_clusters_cm" => "1"},
                                     -arg_defaults => { vol_move_interval_time => 45,
                                                        vol_move_flexvol_names => "",
                                                        vol_move_destination_aggrs => [],
                                                        vol_move_algorithm => "load_balance",
                                                        vol_move_parallel_count => "all",
                                                        vol_move_cutover_window => undef,
                                                        vol_move_cutover_attempts => undef,
                                                        vol_move_cutover_action => undef,
                                                        vol_move_trigger_cutover => "off" },
                                     -validate => sub { &_validate_vol_moveCm(@_) },
                                     -exec => sub { &vol_moveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_vol_moveCm",
                                     -event_modules => \@event_modules,
                                     -description => "This is a CMode event that performs parallel vol moves in the same cluster (using NACL::STask::VolumeMove)",
                                     -required_resource_types => ["nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { "nacl_clusters_cm" => "1"},
                                     -arg_defaults => { vol_move_interval_time => 45,
                                                        vol_move_flexvol_names => "",
                                                        vol_move_destination_aggrs => [],
                                                        vol_move_algorithm => "load_balance",
                                                        vol_move_timeout => 5400,
                                                        vol_move_poll_time => 30,
                                                        vol_move_cutover_window => undef,
                                                        vol_move_cutover_attempts => undef,
                                                        vol_move_cutover_action => undef,
                                                        vol_move_trigger_cutover => "off"},
                                     -validate => sub { &_validate_nacl_vol_moveCm(@_) },
                                     -exec => sub { &nacl_vol_moveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "volops_move_wrapper",
                                     -event_modules => \@event_modules,
                                     -enabled => $EventUtil::FALSE,
                                     -description => "Event Wrapper for the VolOps::Move library",
                                     -required_resource_types => ["nacl_clusters_cm", "nacl_nodes_cm"],
                                     -resource_counts => { "nacl_clusters_cm" => 1 },
                                     -arg_defaults => { volops_global_move_params => {},
                                                        volops_per_vol_move_params => [] },
                                     -validate => sub { &_validate_volops_moveCm(@_) },
                                     -exec => sub { &volops_move_wrapper(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Volume Copy Event Modules
    $self->EventUtil::register_event(-event_name => "volops_copy_wrapper",
                                     -event_modules => \@event_modules,
                                     -description => "Event Wrapper for the VolOps::Copy library",
                                     -required_resource_types => ["nacl_clusters_cm", "nacl_nodes_cm"],
                                     -resource_counts => { "nacl_clusters_cm" => 1 },
                                     -arg_defaults => { volops_global_copy_params => {},
                                                        volops_per_vol_copy_params => [] },
                                     -validate => sub { &_validate_volops_copyCm(@_) },
                                     -exec => sub { &volops_copy_wrapper(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    vol_moveCm()
# Objective: Perform Vol Move operations using the VolOps::Move Engine
# Details:   See POD documentation at the end of this file.
###############################################################################
sub vol_moveCm {
    my ($self, %args) = @_;

    require VolOps::Move;
    require VolOps::BaseException;

    # get the cluster to operate against
    my ($cluster) = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        $self->_set_err("Unable to obtain controller cluster list");
        return $EventUtil::FALSE;
    }

    # obtain the name of a node in this cluster (that is not faulted)
    my ($primary_node) = $self->EventUtil::get_nacl_nodes(-clusters => [$cluster],
                                                          -faulted => $EventUtil::FALSE,
                                                          -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain a node from cluster $cluster");
        return $EventUtil::FALSE;
    }

    # get the nacl node obj for $primary node
    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $primary_node);
    if ($self->error()) {
        $self->_set_err("Unable to obtain a nacl node object for node $primary_node");
        return $EventUtil::FALSE;
    }

    # get the volume name to move
    my $flexvol_name = $self->EventUtil::get_arg(-arg => "vol_move_flexvol_names");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_flexvol_name");
        return $EventUtil::FALSE;
    }

    # get the destination aggrs to play with
    my @aggr_names = @{$self->EventUtil::get_arg(-arg => "vol_move_destination_aggrs")};
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_destination_aggrs");
        return $EventUtil::FALSE;
    }

    # build list of volumes to move
    my @final_volume_objs = $self->_build_list_of_vol_objs(-flexvol_names => $flexvol_name,
                                                           -cluster_name => $cluster);
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $EventUtil::FALSE;
    }
    # if there are no volume objects to move, then bail out
    if (scalar(@final_volume_objs) == 0) {
        $self->_set_err("There are no volumes to move that match the requested specifications");
        return $EventUtil::FALSE;
    }

    # first grab all the Aggr object component states
    my @aggr_objs;
    try {
        @aggr_objs = NACL::CS::StorageAggregate->fetch(command_interface => $nacl_node_obj);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while fetching aggregate objects. " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $vol_move_count = 0;
    my $vol_move_history = {};

    # generate the vol move history for each volume
    foreach my $vol_obj (@final_volume_objs) {
        my $vol_name = $vol_obj->volume();
        my $vs_name = $vol_obj->vserver();

        # if requested, randomize @aggr_names so that its not so predictable
        if ($self->EventUtil::get_arg(-arg => "vol_move_algorithm") =~ /random/i) {
            &ClientUtil::randomize_array(\@aggr_names);
        }

        # copy @aggr_names to {aggr_move_list}
        my @tmp_aggrs = @aggr_names;
        $vol_move_history->{$vs_name}->{$vol_name}->{aggr_move_list} = \@tmp_aggrs;

        # if requested, rotate @aggr_names so each volume load balances across the aggr list
        if ($self->EventUtil::get_arg(-arg => "vol_move_algorithm") =~ /load_balance/i) {
            push(@aggr_names, shift(@aggr_names));
        }
    }

    # perform scalar(@aggr_names) rounds of vol move operations
    for (my $i = 0 ; $i < scalar(@aggr_names) ; $i++) {

        # collect node stats
        $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # collect cluster statistics
        $self->SysStats::collect_cluster_statistics(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # keep track of all the volumes that will get moved this round
        my @per_vol_move_params = ();
        my @volumes_to_be_moved = ();
        my @vol_move_pv_history = ();

        # perform vol moves in parallel (on all volumes)
        foreach my $vol_obj (@final_volume_objs) {
            my $vol_name = $vol_obj->volume();
            my $vs_name = $vol_obj->vserver();

            # pull out the aggr move list for this volume
            my $aggr_move_list = $vol_move_history->{$vs_name}->{$vol_name}->{aggr_move_list};

            # if vol currently lives in @{$aggr_move_list[0]}, then shift it off
            # this aggr and put it to the end of the move list for this volume
            my $curr_aggr_name = $vol_obj->aggregate();
            if ($curr_aggr_name eq @{$aggr_move_list}[0]) {
                # if the aggr_move_list only has one aggregate, then this means this volume has no
                # where to be moved since it already lives in the destination aggregate
                # skip this aggr
                if (scalar(@{$aggr_move_list}) == 1) {
                    next;
                } else {
                    # rotate the aggr move list
                    push(@{$aggr_move_list}, shift @{$aggr_move_list});
                }
            }

            # if there are no moves to make for this volume go to next vol
            if (not scalar(@{$aggr_move_list})) {
                next;
            }

            # obtain the destination aggregate from the front of the move list
            my $destination_aggr = shift @{$aggr_move_list};

            # determine which node owns $curr_aggr_name, and $destination_aggr
            my @dest_nodes;
            my @source_nodes;

            foreach my $aggr_obj (@aggr_objs) {
                my $aggr_name = $aggr_obj->aggregate();
                if ($destination_aggr eq $aggr_name) {
                    # this is the dest node
                    (@dest_nodes) = $aggr_obj->nodes();
                } elsif ($curr_aggr_name eq $aggr_name) {
                    # this is the source node
                    (@source_nodes) = $aggr_obj->nodes();
                }
            }

            # verify that we were able to determine the aggr owners
            if (not scalar(@dest_nodes)) {
                $self->_set_err("Could not determine aggr owner for aggregate $destination_aggr");
                return $EventUtil::FALSE;
            }
            if (not scalar(@source_nodes)) {
                $self->_set_err("Could not determine aggr owner for aggregate $curr_aggr_name.");
                return $EventUtil::FALSE;
            }

            # keep track of all the vol move operations we are doing for this phase
            # so that path validation knows how to validate paths afterwards.
            my $vol_move_item;
            $vol_move_item->{vserver_name} = $vs_name;
            $vol_move_item->{vol_name} = $vol_name;
            $vol_move_item->{source_nodes} = \@source_nodes;
            $vol_move_item->{dest_nodes} = \@dest_nodes;
            push(@vol_move_pv_history, $vol_move_item);

            # build the vol hash for this volume,
            # this will be the input passed into VolOps::Move() for this volume.
            my $vol_hash;
            $vol_hash->{volume} = $vol_name;
            $vol_hash->{vserver} = $vs_name;
            $vol_hash->{aggr_list} = [$destination_aggr];
            $vol_hash->{aggrtype_affinity} = "any";
            $vol_hash->{aggr_loc_affinity} = "any";
            push(@per_vol_move_params, $vol_hash);

            push(@volumes_to_be_moved, $vol_name);

            # increment the move count
            $vol_move_count++;

            # rotate this volume's aggr move list so that it walks the aggr list
            push(@{$aggr_move_list}, $destination_aggr);
        }

        # if there are any vol moves to perform, do them now
        if (scalar(@per_vol_move_params)) {

            # perform the vol moves using the VolOps::Move infrastructure
            my $volops_obj;

            my $vol_move_parallel_count = $self->EventUtil::get_arg(-arg => "vol_move_parallel_count");
            if ($self->error()) {
                $self->_set_err("Unable to obtain vol_move_parallel_count");
                return $EventUtil::FALSE;
            }

            my $vol_move_trigger_cutover = $self->EventUtil::get_arg(-arg => "vol_move_trigger_cutover");
            if ($self->error()) {
                $self->_set_err("Unable to obtain vol_move_trigger_cutover");
                return $EventUtil::FALSE;
            }
            my %extra_args = ();
            my $vol_move_cutover_window = $self->EventUtil::get_arg(-arg => "vol_move_cutover_window");
            if ($self->error()) {
                $self->_set_err("Unable to obtain vol_move_cutover_window");
                return $EventUtil::FALSE;
            }
            if (defined($vol_move_cutover_window)) {
                $extra_args{cutover_window} = $vol_move_cutover_window;
            }

            my $vol_move_cutover_attempts = $self->EventUtil::get_arg(-arg => "vol_move_cutover_attempts");
            if ($self->error()) {
                $self->_set_err("Unable to obtain vol_move_cutover_attempts");
                return $EventUtil::FALSE;
            }
            if (defined($vol_move_cutover_attempts)) {
                $extra_args{cutover_attempts} = $vol_move_cutover_attempts;
            }

            my $vol_move_cutover_action = $self->EventUtil::get_arg(-arg => "vol_move_cutover_action");
            if ($self->error()) {
                $self->_set_err("Unable to obtain vol_move_cutover_action");
                return $EventUtil::FALSE;
            }
            if (defined($vol_move_cutover_action)) {
                $extra_args{cutover_action} = $vol_move_cutover_action;
            }

            # if vol_move_parallel_count eq 'all', then make it equal to
            # the of volumes we are working with
            if ($vol_move_parallel_count =~ /all/) {
                $vol_move_parallel_count = scalar(@final_volume_objs);
            }

            eval {
                my $global_move_params = { iteration_mode => "iteration",
                                           iteration_value => "1",
                                           compare_attrs => $EventUtil::FALSE,
                                           parallel_count => $vol_move_parallel_count,
                                           bg_mode => "fg",
                                           %extra_args,
                                           auto_trigger => $vol_move_trigger_cutover,
                                           volumes => \@volumes_to_be_moved };

                &SanUtil::nate_log("-comment", "Vol Move Phase '$i':\n\n\tGlobal Move Params:\n" . Dumper($global_move_params) .
                                               "\n\tPer Vol Move Params:\n" . Dumper(@per_vol_move_params));

                # Instantiate a volops move object
                $volops_obj = new VolOps::Move(node => $primary_node,
                                               global_move_params => $global_move_params,
                                               per_vol_move_params => \@per_vol_move_params);

                # perform the moves (this will block)
                $volops_obj->move();

                # obtain the status of the test and check if any did not pass
                my $status = $volops_obj->get_status();
                $volops_obj->{subtested} = 1;
                if ($status->{"total_failed"} != 0) {
                    $self->_set_err("Some vol move operations failed.");
                    return $EventUtil::FALSE;
                }
            };

            # catch any exceptions thrown by VolOps::Move
            if ($@) {
                $self->_set_err("$@");
                return $EventUtil::FALSE;
            }

            # sleep for this specified time after all the vol moves complete
            my $vol_move_interval_time = $self->EventUtil::sleep_arg(-arg => "vol_move_interval_time");

            # validate controller after the vol move phase
            $self->ControllerValidation::validate_controller_after_giveback(-id_extension      => "vol_move_phase_$i",
                                                                            -last_timeout      => $vol_move_interval_time,
                                                                            -last_timeout_name => "vol_move_interval_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # Validate ports after the vol move phase
            $self->PortValidation::validate_ports_after_giveback(-id_extension      => "vol_move_phase_$i",
                                                                 -last_timeout      => $vol_move_interval_time,
                                                                 -last_timeout_name => "vol_move_interval_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # validate paths after all volumes have completed their first round of moves
            $self->PathValidation::validate_paths_after_giveback(-id_extension => "vol_move_phase_$i",
                                                                 -path_validation_rules => {validate_vol_move => 1,
                                                                                            validate_takeover_giveback => 1,
                                                                                            vol_move_history => \@vol_move_pv_history},
                                                                 -last_timeout => $vol_move_interval_time,
                                                                 -last_timeout_name => "vol_move_interval_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }

        # move on to the next round of vol moves
        # update the vol object component states since their attributes have changed
        @final_volume_objs = $self->_build_list_of_vol_objs(-flexvol_names => $flexvol_name,
                                                            -cluster_name => $cluster);
        if ($self->error()) {
            $self->_set_err($self->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # if we did not perform any moves, then that is a failure
    if ($vol_move_count == 0) {
        $self->_set_err("Did not perform any vol move operations because all " .
                        "volumes currently reside on the specified destination aggregates");
        return $EventUtil::FALSE;
    }

    # collect node stats
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_vol_moveCm()
# Objective: Perform Vol Move operations using NACL::STask::VolumeMove
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_vol_moveCm {
    my ($self, %args) = @_;

    # get the volume names to move
    my $flexvol_names = $self->EventUtil::get_arg(-arg => "vol_move_flexvol_names");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_flexvol_name");
        return $EventUtil::FALSE;
    }

    # get the destination aggrs to play with
    my $aggr_names = $self->EventUtil::get_arg(-arg => "vol_move_destination_aggrs");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_destination_aggrs");
        return $EventUtil::FALSE;
    }

    # check if flexvol names and aggr names are hash groups
    if (ref($flexvol_names) eq "HASH" and ref($aggr_names) eq "HASH") {
        # process each group serially
        foreach my $group (sort keys %{$flexvol_names}) {
            my $flexvols = $flexvol_names->{$group};
            my $aggrs = $aggr_names->{$group};
            $self->_nacl_vol_move_helper(-flexvol_names => $flexvols,
                                         -aggr_names => $aggrs,
                                         -group_name => $group);
            if ($self->error()) {
                $self->_set_err("Vol Move Group $group failed");
                return $EventUtil::FALSE;
            }
        }
    } else {
        # single group move
        $self->_nacl_vol_move_helper(-flexvol_names => $flexvol_names,
                                     -aggr_names => $aggr_names);
        if ($self->error()) {
            $self->_set_err("Vol Moves failed");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    volops_move_wrapper()
# Objective: Event Wrapper for VolOps::Move library
# Details:   See POD documentation at the end of this file.
###############################################################################
sub volops_move_wrapper {
    my ($self, %args) = @_;

    my $status = $EventUtil::TRUE;
    my $volops_global_move_params  = $self->EventUtil::get_arg(-arg => "volops_global_move_params");
    my $volops_per_vol_move_params = $self->EventUtil::get_arg(-arg => "volops_per_vol_move_params");

    # Call the event wrapper specifying this vol operation is 'vol move'
    $status = $self->_volops_wrapper(volops_global_params  => $volops_global_move_params,
                                     volops_per_vol_params => $volops_per_vol_move_params,
                                     volops_type           => 'move');

    return $status;
}

###############################################################################
# Event Method: volops_copy_wrapper()
# Objective:    Event Wrapper for VolOps::Copy library
# Details:      See POD documentation at the end of this file.
###############################################################################
sub volops_copy_wrapper {
    my ($self, %args) = @_;

    my $status = $EventUtil::TRUE;
    my $volops_global_copy_params  = $self->EventUtil::get_arg(-arg => "volops_global_copy_params");
    my $volops_per_vol_copy_params = $self->EventUtil::get_arg(-arg => "volops_per_vol_copy_params");

    # Call the event wrapper specifying this vol operation is a vol copy
    $status = $self->_volops_wrapper(volops_global_params  => $volops_global_copy_params,
                                     volops_per_vol_params => $volops_per_vol_copy_params,
                                     volops_type           => 'copy');

    return $status;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Private Method: _nacl_vol_move_helper()
# Objective:   Perform a single vol move / aggr name group move
###############################################################################
sub _nacl_vol_move_helper {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $group_name = &ParamsMgr::get_arg(-arg => "group_name",
                                         -default => undef);
    my $flexvol_name = &ParamsMgr::get_arg(-arg => "flexvol_names",
                                           -required => 1);
    my $aggr_names = &ParamsMgr::get_arg(-arg => "aggr_names",
                                         -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # aggr names should be an array
    my @aggr_names = @{$aggr_names};

    # get the cluster to operate against
    my ($cluster) = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        $self->_set_err("Unable to obtain controller cluster list");
        return $EventUtil::FALSE;
    }

    # obtain the name of a node in this cluster (that is not faulted)
    my ($primary_node) = $self->EventUtil::get_nacl_nodes(-clusters => [$cluster],
                                                          -faulted => $EventUtil::FALSE,
                                                          -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain a node from cluster $cluster");
        return $EventUtil::FALSE;
    }

    # get the nacl node obj for $primary node
    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $primary_node);
    if ($self->error()) {
        $self->_set_err("Unable to obtain a nacl node object for node $primary_node");
        return $EventUtil::FALSE;
    }

    my $vol_move_timeout = $self->EventUtil::get_arg(-arg => "vol_move_timeout");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_timeout");
        return $EventUtil::FALSE;
    }

    my $vol_move_trigger_cutover = $self->EventUtil::get_arg(-arg => "vol_move_trigger_cutover");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_trigger_cutover");
        return $EventUtil::FALSE;
    }

    my $vol_move_cutover_window = $self->EventUtil::get_arg(-arg => "vol_move_cutover_window");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_cutover_window");
        return $EventUtil::FALSE;
    }

    my $vol_move_cutover_attempts = $self->EventUtil::get_arg(-arg => "vol_move_cutover_attempts");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_cutover_attempts");
        return $EventUtil::FALSE;
    }

    my $vol_move_cutover_action = $self->EventUtil::get_arg(-arg => "vol_move_cutover_action");
    if ($self->error()) {
        $self->_set_err("Unable to obtain vol_move_cutover_action");
        return $EventUtil::FALSE;
    }
    my %extra_vol_move_args;
    if (defined($vol_move_cutover_window)) {
        $extra_vol_move_args{'cutover-window'} = $vol_move_cutover_window;
    }
    if (defined($vol_move_cutover_attempts)) {
        $extra_vol_move_args{'cutover-attempts'} = $vol_move_cutover_attempts;
    }
    if (defined($vol_move_cutover_action)) {
        $extra_vol_move_args{'cutover-action'} = $vol_move_cutover_action;
    }

    # build list of volumes to move
    my @final_volume_objs = $self->_build_list_of_vol_objs(-flexvol_names => $flexvol_name,
                                                           -cluster_name => $cluster);
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $EventUtil::FALSE;
    }
    # if there are no volume objects to move, then bail out
    if (scalar(@final_volume_objs) == 0) {
        $self->_set_err("There are no volumes to move that match the requested specifications");
        return $EventUtil::FALSE;
    }

    # first grab all the Aggr object component states
    my @aggr_objs;
    try {
        @aggr_objs = NACL::CS::StorageAggregate->fetch(command_interface => $nacl_node_obj);
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while fetching aggregate objects. " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $vol_move_count = 0;
    my $vol_move_history = {};

    # generate the vol move history for each volume
    foreach my $vol_obj (@final_volume_objs) {
        my $vol_name = $vol_obj->volume();
        my $vs_name = $vol_obj->vserver();

        # if requested, randomize @aggr_names so that its not so predictable
        if ($self->EventUtil::get_arg(-arg => "vol_move_algorithm") =~ /random/i) {
            &ClientUtil::randomize_array(\@aggr_names);
        }

        # copy @aggr_names to {aggr_move_list}
        my @tmp_aggrs = @aggr_names;
        $vol_move_history->{$vs_name}->{$vol_name}->{aggr_move_list} = \@tmp_aggrs;

        # if requested, rotate @aggr_names so each volume load balances across the aggr list
        if ($self->EventUtil::get_arg(-arg => "vol_move_algorithm") =~ /load_balance/i) {
            push(@aggr_names, shift(@aggr_names));
        }
    }

    # perform scalar(@aggr_names) rounds of vol move operations
    for (my $i = 0 ; $i < scalar(@aggr_names) ; $i++) {

        # collect node stats
        $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # collect cluster statistics
        $self->SysStats::collect_cluster_statistics(-type => "cmode");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # keep track of all the volumes that will get moved this round
        my @vol_move_tasks = ();

        # perform vol moves in parallel (on all volumes)
        foreach my $vol_obj (@final_volume_objs) {
            my $vol_name = $vol_obj->volume();
            my $vs_name = $vol_obj->vserver();

            # pull out the aggr move list for this volume
            my $aggr_move_list = $vol_move_history->{$vs_name}->{$vol_name}->{aggr_move_list};

            # if vol currently lives in @{$aggr_move_list[0]}, then shift it off
            # this aggr and put it to the end of the move list for this volume
            my $curr_aggr_name = $vol_obj->aggregate();
            if ($curr_aggr_name eq @{$aggr_move_list}[0]) {
                # if the aggr_move_list only has one aggregate, then this means this volume has no
                # where to be moved since it already lives in the destination aggregate
                # skip this aggr
                if (scalar(@{$aggr_move_list}) == 1) {
                    next;
                } else {
                    # rotate the aggr move list
                    push(@{$aggr_move_list}, shift @{$aggr_move_list});
                }
            }

            # if there are no moves to make for this volume go to next vol
            if (not scalar(@{$aggr_move_list})) {
                next;
            }

            # obtain the destination aggregate from the front of the move list
            my $destination_aggr = shift @{$aggr_move_list};

            # determine which node owns $curr_aggr_name, and $destination_aggr
            my @dest_nodes;
            my @source_nodes;

            foreach my $aggr_obj (@aggr_objs) {
                my $aggr_name = $aggr_obj->aggregate();
                if ($destination_aggr eq $aggr_name) {
                    # this is the dest node
                    (@dest_nodes) = $aggr_obj->nodes();
                } elsif ($curr_aggr_name eq $aggr_name) {
                    # this is the source node
                    (@source_nodes) = $aggr_obj->nodes();
                }
            }

            # verify that we were able to determine the aggr owners
            if (not scalar(@dest_nodes)) {
                $self->_set_err("Could not determine aggr owner for aggregate $destination_aggr");
                return $EventUtil::FALSE;
            }
            if (not scalar(@source_nodes)) {
                $self->_set_err("Could not determine aggr owner for aggregate $curr_aggr_name.");
                return $EventUtil::FALSE;
            }

            # build the vol move item
            my $vol_move_item;
            $vol_move_item->{vserver_name} = $vs_name;
            $vol_move_item->{vol_name} = $vol_name;
            $vol_move_item->{source_nodes} = \@source_nodes;
            $vol_move_item->{dest_nodes} = \@dest_nodes;
            $vol_move_item->{destination_aggr} = $destination_aggr;
            $vol_move_item->{source_aggr} = $curr_aggr_name;
            push(@vol_move_tasks, $vol_move_item);

            # increment the move count
            $vol_move_count++;

            # rotate this volume's aggr move list so that it walks the aggr list
            push(@{$aggr_move_list}, $destination_aggr);
        }

        # if there are any vol moves to perform, do them now
        if (scalar(@vol_move_tasks)) {

            if (defined($group_name)) {
                &SanUtil::nate_log("-comment", "Starting Vol Move Phase '$i' (group: $group_name) :\n\n\tUpcoming vol move tasks:\n" . Dumper(@vol_move_tasks));
            } else {
                &SanUtil::nate_log("-comment", "Starting Vol Move Phase '$i' :\n\n\tUpcoming vol move tasks:\n" . Dumper(@vol_move_tasks));
            }
            # perform the vol move tasks one by one
            foreach my $vol_move_task (@vol_move_tasks){
                my $volume = $vol_move_task->{vol_name};
                my $vserver = $vol_move_task->{vserver_name};
                my $destination_aggr = $vol_move_task->{destination_aggr};
                my $source_aggr = $vol_move_task->{source_aggr};
                my $vol_move_job;
                my $vol_move_obj;

                # Redirect NATE Logs
                #my $prefix = $self->EventUtil::get_current_event_obj()->increment_subtest_runid_counter();
                #$Log->redirect(path => (Tharn::param('LOGDIR') . "/${prefix}_vol_move_task__vs_${vserver}__vol_${volume}__from_aggr_${source_aggr}__to_aggr_${destination_aggr}.log"));

                try {
                    &SanUtil::nate_log("-comment", "Starting vol move task: '$vserver: $volume' is moving to aggr '$destination_aggr' from aggr '$source_aggr'");
                    $vol_move_obj = NACL::STask::VolumeMove->start(command_interface => $nacl_node_obj,
                                                                   volume => $volume,
                                                                   vserver => $vserver,
                                                                   'destination-aggregate' => $destination_aggr,
                                                                   foreground => "false",
                                                                   nacltask_wait => 0,
                                                                   job_component => \$vol_move_job,
                                                                   %extra_vol_move_args);
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Volume move task '$vserver: $volume' to aggr '$destination_aggr' " .
                                    "failed.\nReason: " . $e->text());
                };
                if ($self->error()) {
                    # reset the log
                    #$Log->reset_path();
                    return $EventUtil::FALSE;
                }

                # calculate the vol move timeout timestamp
                my $fail_timeout = time + $vol_move_timeout;
                my $complete = 0;

                # wait for the vol move to complete
                while (not $complete) {

                    # sleep the polling time
                    $self->EventUtil::sleep_arg(-arg => "vol_move_poll_time");

                    # check if we timed out
                    if (time >= $fail_timeout) {
                        # we timed out set an error
                        $self->_set_err("Volume move task '$vserver: $volume' to aggr '$destination_aggr' timed out (timeout: $vol_move_timeout sec)");
                        # reset the log
                        #$Log->reset_path();
                        return $EventUtil::FALSE;
                    }

                    # monitor the status of the vol move job
                    try {
                        my $vol_move_cs_state = $vol_move_job->state();
                        if ($vol_move_cs_state->state() =~ /Failure|Error|Quit|Dead/i) {
                            $self->_set_err("Volume move task '$vserver: $volume' to aggr '$destination_aggr' failed");
                        } elsif ($vol_move_cs_state->state() =~ /Success/i) {
                            # the job passed
                            $complete = 1;
                        } elsif (!$vol_move_cs_state->state()) {
                            # I'm not sure about this case, but it basically means that the vol move job does not have
                            # a state which can mean that the vol move completed. Tell the user to email us
                            # so we can better understand this case.
                            $Log->warn("Send email to dl-san-automation about this : " . Dumper($vol_move_cs_state));
                            $complete = 1;
                        } else {
                            # still running ...
                        }

                        # check the state of the vol move object
                        my ($vol_move_cs_obj) = NACL::CS::VolumeMove->fetch(command_interface => $nacl_node_obj,
                                                                            filter => {vserver => $vserver,
                                                                                       volume => $volume});

                        # abort if the job was hard deferred
                        if ($vol_move_cs_obj->phase() =~ /cutover_hard_deferred/i) {
                            if ($vol_move_trigger_cutover =~ /^noforce$/i) {
                                # trigger a non-forced cutover
                                &SanUtil::nate_log("-comment", "Vol move task: '$vserver: $volume' to aggr '$destination_aggr' was hard-deferred, " .
                                                       "triggering cutover ($vol_move_trigger_cutover)");
                                NACL::C::VolumeMove->trigger_cutover(command_interface => $nacl_node_obj,
                                                                     vserver => $vserver,
                                                                     volume => $volume,
                                                                     force => 'false');
                            } elsif ($vol_move_trigger_cutover =~ /^force$/i) {
                                # trigger a forced cutover
                                &SanUtil::nate_log("-comment", "Vol move task: '$vserver: $volume' to aggr '$destination_aggr' was hard-deferred, " .
                                                       "triggering cutover ($vol_move_trigger_cutover)");
                                NACL::C::VolumeMove->trigger_cutover(command_interface => $nacl_node_obj,
                                                                     vserver => $vserver,
                                                                     volume => $volume,
                                                                     force => 'true');
                            } else {
                                # abort on hard deferrals
                                $self->_set_err("Volume move task '$vserver: $volume' to aggr '$destination_aggr' was hard-deferred, aborting, " .
                                                "set vol_move_trigger_cutover to force/noforce to " .
                                                "have this event manually cutover.");
                            }
                        }
                    } catch NACL::Exceptions::NoElementsFound with {
                        # VolumeMove fetch had no contents, this could mean that the job has just ended between the
                        # job show and vol move show
                        # do nothing
                    } otherwise {
                        my $e = shift;
                        $self->_set_err("Encountered an error during Volume move task '$vserver: $volume' to aggr '$destination_aggr'\nReason: " . $e->text());
                    };
                    if ($self->error()) {
                        # reset the log
                        #$Log->reset_path();
                        return $EventUtil::FALSE;
                    }
                }

                # verify the vol move job completed successfully
                try {
                    $vol_move_obj->verify();
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Volume move task '$vserver: $volume' to aggr '$destination_aggr' failed post-move verification\nReason: " . $e->text());
                };
                if ($self->error()) {
                    # reset the log
                    #$Log->reset_path();
                    return $EventUtil::FALSE;
                }

                &SanUtil::nate_log("-comment", "Vol move task: '$vserver: $volume' was moved to aggr '$destination_aggr' from aggr '$source_aggr'");

                # reset the log
                #$Log->reset_path();
            }

            # sleep for this specified time after all the vol moves complete
            my $vol_move_interval_time = $self->EventUtil::sleep_arg(-arg => "vol_move_interval_time");

            # initialize the group name
            my $id_extension;
            if (defined($group_name)) {
                $id_extension = "vol_move_group_${group_name}_phase_${i}";
            } else {
                $id_extension = "vol_move_phase_${i}";
            }

            # validate controller after the vol move phase
            $self->ControllerValidation::validate_controller_after_giveback(-id_extension      => $id_extension,
                                                                            -last_timeout      => $vol_move_interval_time,
                                                                            -last_timeout_name => "vol_move_interval_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # Validate ports after the vol move phase
            $self->PortValidation::validate_ports_after_giveback(-id_extension      => $id_extension,
                                                                 -last_timeout      => $vol_move_interval_time,
                                                                 -last_timeout_name => "vol_move_interval_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # validate paths after all volumes have completed their first round of moves
            $self->PathValidation::validate_paths_after_giveback(-id_extension => $id_extension,
                                                                 -path_validation_rules => {validate_vol_move => 1,
                                                                                            validate_takeover_giveback => 1,
                                                                                            vol_move_history => \@vol_move_tasks},
                                                                 -last_timeout => $vol_move_interval_time,
                                                                 -last_timeout_name => "vol_move_interval_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }

        # move on to the next round of vol moves
        # update the vol object component states since their attributes have changed
        @final_volume_objs = $self->_build_list_of_vol_objs(-flexvol_names => $flexvol_name,
                                                            -cluster_name => $cluster);
        if ($self->error()) {
            $self->_set_err($self->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # if we did not perform any moves, then that is a failure
    if ($vol_move_count == 0) {
        $self->_set_err("Did not perform any vol move operations because all " .
                        "volumes currently reside on the specified destination aggregates");
        return $EventUtil::FALSE;
    }

    # collect node stats
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _volops_wrapper()
# Objective:    Event wrapper for either VolOps::Copy or VolOps:Move
###############################################################################
sub _volops_wrapper {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $volops_global_params  = &ParamsMgr::get_arg(-arg => "volops_global_params",
                                                    -required => 1);
    my $volops_per_vol_params = &ParamsMgr::get_arg(-arg => "volops_per_vol_params",
                                                    -required => 1);
    my $volops_type           = &ParamsMgr::get_arg(-arg => "volops_type",
                                                    -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get the cluster to operate against
    my ($cluster) = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        $self->_set_err("Unable to obtain controller cluster list");
        return $EventUtil::FALSE;
    }

    # obtain the name of a node in this cluster (that is not faulted)
    my ($primary_node) = $self->get_nacl_nodes(-clusters => [$cluster],
                                               -faulted => $EventUtil::FALSE,
                                               -fall_back_on_nacl_clusters => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain a node from cluster $cluster");
        return $EventUtil::FALSE;
    }

    # Collect stats on all nodes
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Change the 'bg_mode' attribute to 'foreground'. Using 'background' would
    # cause the event (parent process) to finish before the volume operation
    # starts. This applies to both Move and Copy

    require VolOps::Move;
    require VolOps::BaseException;

    $volops_global_params->{bg_mode} = 'fg';

    if ($volops_type =~ /^move$/i) {
        eval {
            my $vol_move = new VolOps::Move(node                => $primary_node,
                                            global_move_params  => $volops_global_params,
                                            per_vol_move_params => $volops_per_vol_params);
            $vol_move->move();
            my $status = $vol_move->get_status();
            $vol_move->{subtested} = $EvenUtil::TRUE;
            if ($status->{"total_failed"} == 0) {
                return $EventUtil::TRUE;
            } else {
                $self->_set_err("not all moves passed");
                return $EventUtil::FALSE;
            }
        };

        if ($@) {
            my $errmsg = "$@";
            $self->_set_err($errmsg);
            return $EventUtil::FALSE;
        }
    } elsif ($volops_type =~ /^copy$/i) {
        eval {
            my $vol_copy = new VolOps::Copy(node                => $primary_node,
                                            global_copy_params  => $volops_global_params,
                                            per_vol_copy_params => $volops_per_vol_params);
            $vol_copy->copy();
            my $status = $vol_copy->get_status();
            if ($status->{"total_failed"} == 0) {
                return $EventUtil::TRUE;
            } else {
                $self->_set_err("not all copies passed");
                return $EventUtil::FALSE;
            }
        };

        if ($@) {
            my $errmsg = "$@";
            $self->_set_err($errmsg);
            return $EventUtil::FALSE;
        }
    }

    # Collect stats on all nodes
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    _validate_vol_moveCm()
# Objective: Validate that parameters are set properly for the vol_moveCm
#            Event.
###############################################################################
sub _validate_vol_moveCm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $event_name = $event_obj->get_name();

    # verify that vol_move_flexvol_names is non-null
    my $vol_move_flexvol_names = $self->EventUtil::get_arg(-arg => "vol_move_flexvol_names");
    if (ref($vol_move_flexvol_names) eq "ARRAY") {
        if (not scalar(@{$vol_move_flexvol_names})) {
            $self->_set_err("(Event: $event_name) Missing required value for -vol_move_flexvol_names.");
            return $EventUtil::FALSE;
        }
    } elsif ($vol_move_flexvol_names =~ /^\s*$/) {
        $self->_set_err("(Event: $event_name) Missing required value for -vol_move_flexvol_names.");
        return $EventUtil::FALSE;
    }

    # verify that vol_move_destination_aggr is non-null
    my $vol_move_dest_aggrs = $self->EventUtil::get_arg(-arg => "vol_move_destination_aggrs");
    if ($self->error() or (ref($vol_move_dest_aggrs) ne "ARRAY" and not scalar(@{$vol_move_dest_aggrs}))) {
        $self->_set_err("(Event: $event_name) Missing required value or badly formatted " .
                        "value for -vol_move_destination_aggr.");
        return $EventUtil::FALSE;
    }

    # verify that there are no dups
    $self->_check_for_aggr_dups($vol_move_dest_aggrs);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    _validate_nacl_vol_moveCm()
# Objective: Validate that parameters are set properly for the nacl_vol_moveCm
#            Event.
###############################################################################
sub _validate_nacl_vol_moveCm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $event_name = $event_obj->get_name();
    my @flexvol_keys = ();
    my @dest_aggr_keys = ();

    # verify that vol_move_flexvol_names is non-null
    my $vol_move_flexvol_names = $self->EventUtil::get_arg(-arg => "vol_move_flexvol_names");
    if (ref($vol_move_flexvol_names) eq "ARRAY") {
        if (not scalar(@{$vol_move_flexvol_names})) {
            $self->_set_err("(Event: $event_name) Missing required value for -vol_move_flexvol_names.");
            return $EventUtil::FALSE;
        }
    } elsif (ref($vol_move_flexvol_names) eq "HASH") {
        # verify that both flexvol names and dest aggrs have the same number of phase groups
        @flexvol_keys = sort keys %{$vol_move_flexvol_names};
    } elsif ($vol_move_flexvol_names =~ /^\s*$/) {
        $self->_set_err("(Event: $event_name) Missing required value for -vol_move_flexvol_names.");
        return $EventUtil::FALSE;
    }

    # verify that vol_move_destination_aggr is non-null
    my $vol_move_dest_aggrs = $self->EventUtil::get_arg(-arg => "vol_move_destination_aggrs");
    if (ref($vol_move_dest_aggrs) eq "HASH") {
        @dest_aggr_keys = sort keys %{$vol_move_dest_aggrs};
    } elsif (ref($vol_move_dest_aggrs) ne "ARRAY" and not scalar(@{$vol_move_dest_aggrs})) {
        $self->_set_err("(Event: $event_name) Missing required value or badly formatted " .
                            "value for -vol_move_destination_aggr.");
        return $EventUtil::FALSE;
    }

    # calculate the diff between the two arrays if the hash format was used
    if (scalar(@flexvol_keys) eq scalar(@dest_aggr_keys) and scalar(@flexvol_keys)) {
        my (@diff) = ();
        my (%count) = ();
        foreach my $e (@flexvol_keys, @dest_aggr_keys) {$count{$e}++}
        foreach my $val (values %count) {
            if ($val != 2) {
                # detected a diff between the two arrays
                $self->_set_err("(Event: $event_name) Detected a group ($val) that does not have " .
                                "a peer group in both vol_move_destination_aggrs and vol_move_flexvol_names");
                return $EventUtil::FALSE;
            }
        }

        # verify that there are no duplicate aggr names in any of the groups
        foreach my $aggr_list (values %{$vol_move_dest_aggrs}) {
            $self->_check_for_aggr_dups($event_name, $aggr_list);
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }
    } else {
        # verify that there are no duplicate aggr names
        $self->_check_for_aggr_dups($event_name, $vol_move_dest_aggrs);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Private Method:    _check_for_aggr_dups
# Objective: Validate that there are no aggr duplicates
###############################################################################
sub _check_for_aggr_dups {
    my ($self, $event_name, $vol_move_dest_aggrs) = @_;


    # verify that there are no duplicate aggr names in the aggr lists
    if (ref($vol_move_dest_aggrs) eq "ARRAY") {
        my %aggr_hash;
        foreach my $aggr (@{$vol_move_dest_aggrs}) {
            $aggr_hash{$aggr}++;
        }

        foreach my $a (keys %aggr_hash) {
            if ($aggr_hash{$a} > 1) {
                $self->_set_err("(Event: $event_name) Found duplicate aggr names in -vol_move_destination_aggrs");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Private Method:    _validate_volops_moveCm()
# Objective: Validate that parameters are set properly for the volops_moveCm
#            Event.
###############################################################################
sub _validate_volops_moveCm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $event_name = $event_obj->get_name();

    # verify that the following Event Parameters were specified properly
    # using ParamsMgr
    my %arg_hash = (-volops_global_move_params => $self->EventUtil::get_arg(-arg => "volops_global_move_params"),
                    -volops_per_vol_move_params => $self->EventUtil::get_arg(-arg => "volops_per_vol_move_params"));
    &ParamsMgr::process_args(%arg_hash);
    &ParamsMgr::get_arg(-arg => "volops_global_move_params",
                        -required => 1,
                        -ref_type => "HASH");
    &ParamsMgr::get_arg(-arg => "volops_per_vol_move_params",
                        -required => 0,
                        -default  => undef,
                        -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err("(Event: $event_name) Event Parameter validation failed: " . &ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    _validate_volops_copyCm()
# Objective: Validate that parameters are set properly for the volops_copyCm
#            Event.
###############################################################################
sub _validate_volops_copyCm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $event_name = $event_obj->get_name();

    # verify that the following Event Parameters were specified properly
    # using ParamsMgr
    my %arg_hash = (-volops_global_copy_params  => $self->EventUtil::get_arg(-arg => "volops_global_copy_params"),
                    -volops_per_vol_copy_params => $self->EventUtil::get_arg(-arg => "volops_per_vol_copy_params"));
    &ParamsMgr::process_args(%arg_hash);
    &ParamsMgr::get_arg(-arg => "volops_global_copy_params",
                        -required => 1,
                        -ref_type => "HASH");
    &ParamsMgr::get_arg(-arg => "volops_per_vol_copy_params",
                        -required => 1,
                        -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err("(Event: $event_name) Event Parameter validation failed: " . &ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method: _build_list_of_vol_objs()
# Objective: Builds a list of volume objects to either move or copy
###############################################################################
sub _build_list_of_vol_objs {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $flexvol_names = &ParamsMgr::get_arg(-arg => "flexvol_names",
                                            -required => 1);
    my $cluster_name = &ParamsMgr::get_arg(-arg => "cluster_name",
                                           -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @final_volume_objs = ();

    # get the vservers in this cluster
    my @vservers = $self->EventUtil::get_vservers(-clusters => [$cluster_name]);
    if ($self->error()) {
        $self->_set_err("Unable to obtain vserver list for cluster $cluster_name");
        return (wantarray()) ? () : [];
    }

    # get nacl_clusters_cm object
    my $nacl_cluster_obj = $self->EventUtil::get_nacl_cluster_cm_obj(-cluster_name => $cluster_name);
    if ($self->error()) {
        $self->_set_err("Failed to obtain cluster object for $cluster_name");
        return (wantarray()) ? () : [];
    }

    # build list of volumes to copy
    foreach my $vserver (@vservers) {
        # get all volumes on this vserver
        my @volume_objs;
        try {
            @volume_objs = NACL::CS::Volume->fetch(command_interface => $nacl_cluster_obj,
                                                   filter => {vserver => $vserver});
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught while fetching volumes for vserver $vserver " . $e->text());
        };
        if ($self->error()) {
            return (wantarray()) ? () : [];
        }

        # filter out the volumes that dont match the regex
        @volume_objs = grep {
            my $vol_name = $_->volume();

            # $flexvol_name can either be an ARRAY of absolute vol names, or
            # a regular expression matching, handle each differently
            if (ref($flexvol_names) eq "ARRAY") {
                # filter out the volumes that are not in the array to filter
                grep {
                    if ($vol_name eq $_) {
                        # include this volume
                        1;
                    } else {
                        # dont include this volume
                        0;
                    }
                } @{$flexvol_names};
            } else {
                # regex filter
                if ($vol_name =~ $flexvol_names) {
                    # include this volume
                    1;
                } else {
                    # dont include this volume
                    0;
                }
            }
        } @volume_objs;

        push(@final_volume_objs, @volume_objs);
    }

    return (wantarray()) ? @final_volume_objs : \@final_volume_objs;
}

=pod

=head1 Package Name

VolumeEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for CMode Volume Events

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::VolumeEvents

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::PathValidation

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=over 4

None

=back

=head1 B<Event Keywords>

=head2 B<nacl_vol_moveCm>

Perform Vol Moves on a single CMode Cluster using NACL::STask::VolMove. This event adds some additional
logic around the library to load balance the aggr destination choices for each vol move operation. One
vol move will be done at a time.

=over 4

=item Resources

=item *I<-vservers>

Optional: An array reference of vservers to use for the vol move operation

Default: All vservers in the Event scope will be used.

=item *I<-nacl_clusters_cm>

Required: An array reference of controller cluster. Must be exactly one cluster.

Default: All controller clusters in the Event scope will be used.

Resource Count: 1

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation.

Default: all SanClients in the Event scope will be used

=item Event Parameters

=item *I<-vol_move_flexvol_names>

Required: (1) An array reference of volume names to operate against, OR
(2) a perl regular expression matching of volumes to operate against, OR
(3) a hash reference of array reference/regex groups. If a hash reference is used,
then the keys to the hash (the group names) must match the keys (group names) in
the hash reference vol_move_destination_aggrs. See the examples later.

Default: ""

=item *I<-vol_move_destination_aggrs>

Required: An array reference of aggregate names to use as vol move destinations,
or a Hash reference of array reference groups. See the examples later.

Default: []

=item *I<-vol_move_interval_time>

Optional: Specifies the time (in seconds) to rest in between vol move phases. This will also
control how long to wait before running path validation.

Default: 45

=item *I<-vol_move_algorithm>

Optional: The algorithm to use when choosing aggr destinations for each vol move phase.

Valid Values: "load_balance" (which will attempt to load balance the aggr selections),
or "random" (which will randomly select aggr destinations from the list)

Default: load_balance

=item *I<-vol_move_trigger_cutover>

Options: Specifies whether or not the trigger-cutover command should be automatically executed when a volume move job enters a hard-deferred state.
Valid values::
      off     - Do not automatically trigger a cutover
      force   - Automatically trigger cutover with the force
                option set to true
      noforce - Automatically trigger cutover with the force
                option set to false

Default: off


=item *I<-vol_move_timeout>

Optional: Specifies the maximum amount of time to wait for the vol move to complete. The event will fail if any vol move
exceeds this timeout.

Default: 5400 (seconds)

=item *I<-vol_move_poll_time>

Optional: Specifies how often to poll a vol move job to see if it has completed yet.

Default: 30 (seconds)

=item *I<-vol_move_cutover_window>

Optional: Specifies the cutover window in seconds

Default: (implied)

=item *I<-vol_move_cutover_attempts>

Optional: Specifies the cutover attempts

Default: (implied)

=item *I<-vol_move_cutover_action>

Optional: Specifies the cutover action.

Default: (implied)

=item Examples

 # Perform vol moves on all volumes that match qr/testvol\d+/ for the two vservers
 # specified on 'mycluster'. Each volume will be moved to each aggregate in the list exactly once
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "nacl_vol_moveCm",
                         -nacl_clusters_cm => ["mycluster"]
                         -vservers => ["vs0", "vs1"],
                         -vol_move_flexvol_names => qr/testvol\d+/,
                         -vol_move_destination_aggrs => ["aggr1", "aggr2", "aggr3"]);
 if ($EventManobj->error()) {
     ...
 }

 # Perform vol moves on all volumes named ["v1", "v2", "v3] in the cluster "mycluster",
 # Each volume will be moved to each aggregate in the list exactly once
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "nacl_vol_moveCm",
                         -nacl_clusters_cm => ["mycluster"]
                         -vol_move_flexvol_names => qr/testvol\d+/,
                         -vol_move_destination_aggrs => ["aggr1", "aggr2", "aggr3"]);
 if ($EventManobj->error()) {
     ...
 }

 # Perform group based vol moves,
 # all volumes in group1 (all vols that match 'foo') will be moved around between aggr1 and aggr2
 # all volumes in group2 (volumes 'bar1' and 'bar2') will be moved around between aggr3 and aggr1
 # there is no limit to the number of vol move groups that can be specified. The only requirement
 # is that both vol_move_flexvol_names and vol_move_destination_aggrs have the exact same group names specified.
 # Each volume will be moved to each aggregate in the list exactly once
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "nacl_vol_moveCm",
                         -nacl_clusters_cm => ["mycluster"]
                         -vservers => ["vs0","vs1"],
                         -vol_move_flexvol_names => { group1 => qr/foo/i, group2 => ['bar1','bar2']},
                         -vol_move_destination_aggrs => { group1 => ["aggr1", "aggr2"], group2 => ["aggr1", "aggr3"] });
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<vol_moveCm> I<deprecated>

Perform Vol Moves on a single CMode Cluster using VolOps::Move. This event adds some additional
logic around the library to load balance the aggr destination choices for each vol move operation.

=over 4

=item Resources

=item *I<-vservers>

Optional: An array reference of vservers to use for the vol move operation

Default: All vservers in the Event scope will be used.

=item *I<-nacl_clusters_cm>

Required: An array reference of controller cluster. Must be exactly one cluster.

Default: All controller clusters in the Event scope will be used.

Resource Count: 1

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation.

Default: all SanClients in the Event scope will be used

=item Event Parameters

=item *I<-vol_move_flexvol_names>

Required: (1) An array reference of volume names to operate against, OR
(2) a perl regular expression matching of volumes to operate against

Default: ""

=item *I<-vol_move_destination_aggrs>

Required: An array reference of aggregate names to use as vol move destinations.

Default: []

=item *I<-vol_move_interval_time>

Optional: Specifies the time (in seconds) to rest in between vol move phases. This will also
control how long to wait before running path validation.

Default: 45

=item *I<-vol_move_algorithm>

Optional: The algorithm to use when choosing aggr destinations for each vol move phase.

Valid Values: "load_balance" (which will attempt to load balance the aggr selections),
or "random" (which will randomly select aggr destinations from the list)

Default: load_balance

=item *I<-vol_move_parallel_count>

Optional: Controls how many vol moves to issue in parallel (vol move operations will be queued as necessary)

Valid Values: <integers> or "all"

Default: all

=item *I<-vol_move_trigger_cutover>

Options: Specifies whether or not the trigger-cutover command should be automatically executed when a volume move job enters a deferred state.
Valid values::
      off     - Do not automatically trigger a cutover
      force   - Automatically trigger cutover with the force
                option set to true
      noforce - Automatically trigger cutover with the force
                option set to false

Default: off

=item Examples

 # Perform vol moves on all volumes that match qr/testvol\d+/ for the two vservers
 # specified on 'mycluster'. Each volume will be moved to each aggregate in the list exactly once
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "vol_moveCm",
                         -nacl_clusters_cm => ["mycluster"]
                         -vservers => ["vs0", "vs1"],
                         -vol_move_flexvol_names => qr/testvol\d+/,
                         -vol_move_destination_aggrs => ["aggr1", "aggr2", "aggr3"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<volops_move_wrapper> I<deprecated>

A passthrough event wrapper for the VolOps::Move library

=over 4

=item Resources

=item *I<-nacl_clusters_cm>

Required: an Array reference of controller clusters to use. No more than ONE cluster can be specified.

Resource Count: 1

=item Event Parameters

=item *I<-volops_global_move_params>

Required: Hash reference to control global volops parameters.

See Pod for VolOps::Move::new(). Parameter = global_move_params

=item *I<-volops_per_vol_move_params>

Optional Array reference of Hash references to control per volume move parameters.

See Pod for VolOps::Move::new(). Parameter = per_vol_move_params

=item Examples

 # Wrapper for the VolOps::Move automation suite.
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "volops_move_wrapper",
                         -nacl_clusters_cm => ["mycmodecluster"],
                         -volops_global_move_params => {},
                         -volops_per_vol_move_params => []);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<volops_copy_wrapper> I<deprecated>

An passthrough event wrapper for the VolOps::Copy library

=over 4

=item Resources

=item *I<-nacl_clusters_cm>

Required: an Array reference of controller clusters to use. No more than ONE cluster can be specified.

Resource Count: 1

=item Event Parameters

=item *I<-volops_global_copy_params>

Required: Hash reference to control global volops parameters.

See Pod for VolOps::Copy::new(). Parameter = global_copy_params

=item *I<-volops_per_vol_copy_params>

Optional: Array reference of Hash references to control per volume copy parameters.

See Pod for VolOps::Copy::new(). Parameter = per_vol_copy_params

=item Examples

 # Wrapper for the VolOps::Copy automation suite.
 # see EventMan::add_event() for parameter inputs
 $EventManobj->add_event(-event_name => "volops_copy_wrapper",
                         -nacl_clusters_cm => ["mycmodecluster"],
                         -volops_global_copy_params => {},
                         -volops_per_vol_copy_params => []);
 if ($EventManobj->error()) {
     ...
 }

=back

=cut

1;
