# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for generic FC Fabric Events
## @author dl-san-automation@netapp.com
## @status private
## @pod here

package EventMan::EventModules::FabricEvents;

use strict;
use warnings;
use ParamsMgr;
use NATE::Events;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::SwitchEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "fabric_reboot",
                                     -event_modules => \@event_modules,
                                     -description => "Perform fabric reboots to simulate fabric outages.",
                                     -required_resource_types => ["sanswitches", "fabrics"],
                                     -resource_counts => {"fabrics" => "1+"},
                                     -exec => sub { &fabric_reboot(@_) });

    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults => { fabric_reboot_wait_time => 300,
                                                                        fc_gs_collection => 1 });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    fabric_reboot()
# Objective: Reboot all fabrics one by one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fabric_reboot {
    my ($self, %args) = @_;

    # obtain the list of fabrics to work with
    my @fabrics = $self->EventUtil::get_fabrics(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Failed to obtain fabric resources");
        return $EventUtil::FALSE;
    }

    # verify we have fabrics to work with
    if (not scalar(@fabrics)) {
        $self->_set_err("There are no unfaulted fabric resources to work with");
        return $EventUtil::FALSE;
    }

    # run the FC_GS Commands now
    $self->ControllerValidation::collect_fc_gs_information();
    if ($self->error()) {
        $self->_set_err("Failed to collect FC_GS Information on nodes");
        return $EventUtil::FALSE;
    }

    # dump the topology of the switches
    $self->dump_switch_topology();
    if ($self->error()) {
        $self->_set_err("Failed to dump switch topology");
        return $EventUtil::FALSE;
    }

    # walk through the list of fabrics
    foreach my $fabric_name (@fabrics) {
        # obtain the list of switches in this fabric
        my @sanswitches = $self->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE,
                                                            -fabrics => [$fabric_name]);
        if ($self->error()) {
            $self->_set_err("Failed to obtain sanswitches list for fabric $fabric_name");
            return $EventUtil::FALSE;
        }

        # reboot every switch in this fabric
        $self->reboot_switches(-sanswitches => \@sanswitches);
        if ($self->error()) {
            $self->_set_err("Failed to reboot switches @sanswitches");
            return $EventUtil::FALSE;
        }

        # wait 10 minutes
        my $time = $self->EventUtil::sleep_arg(-arg => "fabric_reboot_wait_time");

        my $id_extension = "reboot_on_fabric_${fabric_name}";

        # collect FC_GS information
        $self->ControllerValidation::collect_fc_gs_information();
        if ($self->error()) {
            $self->_set_err("Failed to collect FC_GS Information on nodes");
            return $EventUtil::FALSE;
        }

        # dump the topology of the switches
        $self->dump_switch_topology();
        if ($self->error()) {
            $self->_set_err("Failed to dump switch topology");
            return $EventUtil::FALSE;
        }

        # controller validation
        $self->ControllerValidation::validate_controller_after_giveback(-id_extension => $id_extension,
                                                                        -last_timeout => $time,
                                                                        -last_timeout_name => "fabric_reboot_wait_time");
        if ($self->error()) {
            $self->_set_err("Controller validation failed after fabric reboot on fabric $fabric_name");
            return $EventUtil::FALSE;
        }

        # validate paths
        $self->PathValidation::validate_paths_after_giveback(-id_extension => $id_extension,
                                                             -path_validation_rules => {validate_takeover_giveback => 1},
                                                             -last_timeout => $time,
                                                             -last_timeout_name => "fabric_reboot_wait_time");
        if ($self->error()) {
            $self->_set_err("Path validation failed after fabric reboot on fabric $fabric_name");
            return $EventUtil::FALSE;
        }

    }

    return $EventUtil::TRUE;
}

1;
