# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for Miscellaneous Events
## @author derricks@netapp.com
## @status private
## @pod here

# $Id:$

package EventMan::EventModules::MiscEvents;

use Tharn;
use ParamsMgr;
use Sys::Hostname qw(hostname);
use NATE::Util::Sockets qw(get_address);
use NATE::STAF;
use NATE::Events qw(listener_add eventloop_drain);

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "manual_event",
                                     -event_modules => \@event_modules,
                                     -description => "Manual Event",
                                     -exec => sub { &manual_event(@_) });

    $self->EventUtil::register_event(-event_name => "path_validation_on_giveback",
                                     -event_modules => \@event_modules,
                                     -description => "Path Validation Event",
                                     -required_resource_types => ["sanclients"],
                                     -exec => sub { &path_validation_on_giveback(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults => {});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################
###############################################################################
# Event Method: manual_event()
# Objective: Perform manual fault
# Details:   See POD documentation at the end of this file.
###############################################################################
sub manual_event {
    my ($self) = @_;

    # Get the STAF object so we can get the staf handle
    my $staf = NATE::STAF::staf_global();
    unless (defined($staf) and ref($staf) =~ /staf/i) {
        $self->_set_err("Unable to get the STAF object");
        return $EventUtil::FALSE;
    }

    # Get the hostname for the current process so we can tell the user
    # what endpoint to use
    my $endpoint = hostname();
    unless (defined($endpoint) and $endpoint ne "") {
        $self->_set_err("Unable to get the hostname where our STAF process is running");
        return $EventUtil::FALSE;
    }

    # Get the STAF handle
    my $staf_handle = $staf->handle();
    unless (defined($staf_handle) and $staf_handle ne "") {
        $self->_set_err("Unable to get handle for our STAF process");
        return $EventUtil::FALSE;
    }

    # Start up the listeners
    unless ($self->_start_event_listeners(-object => $self)) {
        $self->_set_err("Unable to start the listeners");
        return $EventUtil::FALSE;
    }

    # Print out a message to the comment log explaining to the user
    # how to inject messages into the manual event
    &SanUtil::nate_log(-comment, "STAF Handle: $staf_handle   EndPoint: $endpoint");
    &SanUtil::nate_log(-comment, "In order to inject messages into the manual event, run the script\n" .
                                 "SANInterop/tools/bin/manual_event_control and provide the\n" .
                                 "STAF handle '$staf_handle' and EndPoint '$endpoint' when prompted");

    # Wait on the event messages
    $self->_monitor_event_messages();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
## Event Method: path_validation_on_giveback()
## Objective: Perform path validation
## Details:   See POD documentation at the end of this file.
################################################################################
sub path_validation_on_giveback {
    my ($self) = @_;

    # perform path validation on all initiators
    $self->PathValidation::validate_paths_after_giveback(-id_extension => "pv_event",
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => 1,
                                                         -last_timeout_name => "timeout");
    if ($self->error()) {
        $self->_set_err("Path validation failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _monitor_event_messages
# Objective: Wait for event messages
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _monitor_event_messages {
    my ($self) = @_;

    # Wait on the eventloop
    while (1) {
        # wait on an event
        while (scalar(@{$self->{command_queue}}) > 0) {
            my $command_hash = shift @{$self->{command_queue}};

            my $command = $command_hash->{command};
            my $message = $command_hash->{message};

            # Check the job_control attribute on the object.
            if ($command eq "stop") {
                # Looks like they want to stop the event
                &SanUtil::nate_log(-comment, "***EventMan*** Fault Stop Message Received");
                return $EventUtil::TRUE;
            } elsif ($command eq "validate") {
                # Looks like they want to perform validation
                &SanUtil::nate_log(-comment, "***EventMan*** Performing Validation Routines");

                # Perform Port Validation
                $self->PortValidation::validate_ports_after_port_up(-id_extension => "manual_event_port_validation");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                # Perform Controller Validation
                $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "manual_event_controller_validation");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                # Perform Path Validation
                $self->PathValidation::validate_paths_after_giveback(-id_extension => "manual_event_path_validation");
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                # Perform SRAM Dump Detection
                $self->SRAMDetection::check_for_new_sram_dumps();
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }

                next;
            } elsif ($command eq "abort") {
                # Looks like they want to abort the event
                $self->_set_err("***EventMan*** Fault Abort Message Received");
                return $EventUtil::FALSE;
            } elsif ($command eq "comment") {
                # Looks like they want to log a comment in the NATE log.  Make sure we have a
                # mesage to use.
                &SanUtil::nate_log(-comment, "***EventMan*** " . $message);
            } elsif ($command eq "syslog") {
                # Looks like they want to send a message to syslog
                $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** " . $message]);
            } else {
                # We didn't find a valid message type so log a message but don't error.
                &SanUtil::nate_log(-comment,
                                "***EventMan*** Invalid command '$command' recieved");
                next;
            } ##END if ($command eq "stop") {
        } ##END while (scalar(@{$self->{command_queue}}) > 0) {

        sleep 10;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _start_event_listeners
# Objective: Start all of the listeners
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _start_event_listeners {
    my ($self) = @_;

    # Create the fault listeners
    $self->_create_event_listener(-type => "fault");
    if ($self->error) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _create_event_listener
# Objective: Add a STAF listener.  Valid types are eventman, fault, and io
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _create_event_listener {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # Get some args
    my $type = &ParamsMgr::get_arg(-arg => "type", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Make sure type is lowercase
    $type = lc($type);

    # Start the requested listner
    if ($type eq "fault") {
        &NATE::Events::listener_add( {callback => sub {$self->_process_fault_message(@_)},
                                      filter   => sub {$self->_check_fault_message(@_)},
                                      object   => $self});
    } else {
        $self->_set_err("Invalid type '$type' specified.  Value must be 'fault'");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _check_fault_message
# Objective: check to see if we received a fault message
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _check_fault_message {
    my ($self, $msg) = @_;

    # Make sure we received the required parameter
    unless (defined($msg)  and $msg ne "") {
        &SanUtil::nate_log('-warn', "Missing paramater 'msg'");
        return $EventUtil::FALSE;
    }

    # Check to see if we receieved an fault msg type
    if (defined($msg->{type}) and $msg->{type} eq "fault") {
        return $EventUtil::TRUE;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    _process_fault_message
# Objective: Handle the message we received for fault.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _process_fault_message {
    my ($self, $event, $listener) = @_;

    # Process the event hash
    &ParamsMgr::process_args(%$event);

    my $message_hash  = &ParamsMgr::get_arg(-arg      => "message",
                                            -ref_type => "HASH",
                                            -required => 1,
                                            -default  => {});

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Process the message hash
    &ParamsMgr::process_args(%$message_hash);

    my $command = &ParamsMgr::get_arg(-arg => "command", -required => 1, -default => "");
    my $message = &ParamsMgr::get_arg(-arg => "message", -required => 0, -default => "");

    # Take the command and message (with defaults if necessary) and create the command hash.
    # We can't use message_hash outright in cases they didn't pass in a message.
    my $command_hash = {'command' => $command, 'message' => $message};

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Add the command to the command queue for self so it can be processed by
    # the monitor routine.
    push(@{$self->{command_queue}}, $command_hash);

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

MiscEvents.pm

=head1 Author

Derrick Shultz

=head1 Description

Event Module for miscellaneous events

=head1 B<Event Keywords>

=head2 B<manual_event>

Performs a manual event.  Control of the event can be achieved by the use of an external
control script called test/SANInterop/tools/bin/manual_event_control.  This script can be
run with options or interactive and send "event stop", "event abort", "event message", and
"syslog message" type commands to the manual event.

=over 4

=item Event Parameters

None

=item Examples

 # see EventMan::add_event()
 # perform manual event
 $EventManobj->add_event(-event_name => "manual_event");
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<path_validation_on_giveback>

Runs path validation as an event.

=over 4

=item Event Parameters

None

=item Examples

 # see EventMan::add_event()
 # perform path validation event
 $EventManobj->add_event(-event_name => "path_validation_on_giveback");
 if ($EventManobj->error()) {
     ...
 }

=back

=cut
