# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for generic FC & Ethernet Switch Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::SwitchEvents;

use Data::Dumper;
use Tharn;
use ParamsMgr;
use San::FabricMgr;
use strict;
use warnings;

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "dump_switch_info",
                                     -event_modules => \@event_modules,
                                     -description => "Dump switch info",
                                     -exec => sub { &dump_switch_info(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "initiator_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all Initiators",
                                     -required_resource_types => ["sanswitches", "sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "parallel" },
                                     -exec => sub { &initiator_switch_port_blocks(@_) });

    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "fc_initiator_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on FC Initiators",
                                     -required_resource_types => ["sanswitches", "sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "parallel" },
                                     -exec => sub { &fc_initiator_switch_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "iscsi_initiator_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on iSCSI initiators",
                                     -required_resource_types => ["sanswitches", "sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "parallel" },
                                     -exec => sub { &iscsi_initiator_switch_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "target_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["nacl_nodes_cm", "sanfilers7m", "nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &target_switch_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "target_switch_port_blocks_7m",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all 7mode Targets",
                                     -required_resource_types => ["sanswitches", "sanfilers7m"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &target_switch_port_blocks_7m(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "target_switch_port_blocks_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all Cmode Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["nacl_nodes_cm", "nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &target_switch_port_blocks_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "fc_target_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all FC Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["sanfilers7m", "nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &fc_target_switch_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "iscsi_target_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all iSCSI Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["sanfilers7m", "nacl_nodes_cm", "nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &iscsi_target_switch_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "fc_target_switch_port_blocks_7m",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all 7mode FC Targets",
                                     -required_resource_types => ["sanswitches", "sanfilers7m"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &fc_target_switch_port_blocks_7m(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "iscsi_target_switch_port_blocks_7m",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all 7mode iSCSI Targets",
                                     -required_resource_types => ["sanswitches", "sanfilers7m"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &iscsi_target_switch_port_blocks_7m(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "fc_target_switch_port_blocks_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all Cmode FC Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &fc_target_switch_port_blocks_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "iscsi_target_switch_port_blocks_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all Cmode iSCSI Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["nacl_nodes_cm", "vservers", "nacl_clusters_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &iscsi_target_switch_port_blocks_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "apcon_initiator_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Apcon Port blocks on FC Initiators",
                                     -required_resource_types => ["sanswitches", "sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "parallel" },
                                     -exec => sub { &apcon_initiator_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "apcon_target_port_blocks_7m",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Apcon Port blocks on all 7mode FC Targets",
                                     -required_resource_types => ["sanswitches", "sanfilers7m"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &apcon_target_port_blocks_7m(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "apcon_target_port_blocks_Cm",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Switch Port blocks on all Cmode FC Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &apcon_target_port_blocks_Cm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "apcon_target_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform Apcon Port blocks on all FC Targets",
                                     -required_resource_types => ["sanswitches"],
                                     -require_at_least_one_resource_types => ["sanfilers7m", "nacl_clusters_cm", "vservers"],
                                     -optional_resource_types => ["sanclients"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &apcon_target_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "fc_isl_switch_port_blocks",
                                     -event_modules => \@event_modules,
                                     -description => "Perform E_Port switch port blocks",
                                     -required_resource_types => ["sanswitches"],
                                     -optional_resource_types => ["sanclients", "sanfilers7m", "nacl_clusters_cm", "vservers"],
                                     -arg_defaults => { switch_port_block_algorithm => "sequential" },
                                     -exec => sub { &fc_isl_switch_port_blocks(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { switch_port_block_abort_on_error => 1,
                                                           switch_port_block_miscompare_abort_on_error => 1,
                                                           switch_port_block_dump_topology => 1,
                                                           switch_port_block_dump_info_on_failure => $EventUtil::FALSE,
                                                           switch_port_block_fcoe_block_type => "both",
                                                           switch_port_down_time => 30,
                                                           switch_port_up_time => 30,
                                                           switch_port_block_algorithm => "sequential",
                                                           switch_port_block_conserve_switch_connections => $EventUtil::FALSE,
                                                           switch_port_block_sync_parallel_blocks => $EventUtil::FALSE,
                                                           switch_port_total_block_count => 0,
                                                           switch_port_block_cycle_count => 1,
                                                           switch_port_block_pv_type => "port",
                                                           switch_port_block_cmode_target_type => "node",
                                                           switch_port_block_sort_switch_names => $EventUtil::FALSE});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "switch_port_down_time",
                                           -package => __PACKAGE__,
                                           -default => "30",
                                           -format => "integer (seconds)",
                                           -example => "-switch_port_down_time => 60",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how long to port down after " .
                                                           "the port is disabled on the switch.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_up_time",
                                           -package => __PACKAGE__,
                                           -default => "30",
                                           -format => "integer (seconds)",
                                           -example => "-switch_port_up_time => 60",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how long to wait after enabling a " .
                                                           "port on the switch.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_dump_info_on_failure",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::FALSE,
                                           -format => "$EventUtil::FALSE|$EventUtil::TRUE",
                                           -example => "-switch_port_block_dump_info_on_failure => $EventUtil::TRUE",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to dump port information " .
                                                           "when an error is discovered during port blocks");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_dump_topology",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0|1"
                                           -example => "0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether the event should dump the fabric topology");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_algorithm",
                                           -package => __PACKAGE__,
                                           -default => "sequential",
                                           -format => "sequential|parallel"
                                           -example => "-switch_port_block_algorithm => \"parallel\"",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to perform parallel " .
                                                           "or sequential resource port blocks");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_total_block_count",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "integer"
                                           -example => "-switch_port_total_block_count => 100",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how many total port blocks to perform. " .
                                                           "If set to '0', will perform at least one block per " .
                                                           "port specified/discovered. Only applicable when " .
                                                           "-switch_port_block_algorithm => \"sequential\"");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_cycle_count",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "integer"
                                           -example => "-switch_port_block_cycle_count => 10",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how many cycles of port blocks to perform " .
                                                           "per resource. Only applicable when " .
                                                           "-switch_port_block_algorithm => \"parallel\"");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_cmode_target_type",
                                           -package => __PACKAGE__,
                                           -default => "node",
                                           -format => "node|vservers|nacl_clusters_cm"
                                           -example => "-switch_port_block_cmode_target_type => \"vservers\"",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies from which point of view Cmode targets " .
                                                           "should be chosen when performing target port blocks, at " .
                                                           "the Node, Vserver, or ControllerCluster level");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_pv_type",
                                           -package => __PACKAGE__,
                                           -default => "port",
                                           -format => "port|cycle"
                                           -example => "-switch_port_block_pv_type => \"port\"",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether path validation should be done " .
                                                           "after every port block or after every cycle");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0|1"
                                           -example => "0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether the event should error out after a " .
                                                           "port block failure");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_miscompare_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0|1"
                                           -example => "0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether the event should error out after a " .
                                                           "miscompare of type or speed following a port enable");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_fcoe_block_type",
                                           -package => __PACKAGE__,
                                           -default => "both",
                                           -format => "physical|virtual|both"
                                           -example => "virtual",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies the type of port block to do with FCoE ports");

    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "switch_port_block_sort_switch_names",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::FALSE,
                                           -format => "0|1"
                                           -example => "1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to sort the list of switch names before " .
                                                           "performing port blocks");

    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    # build the switch info hash.  If already exists in ResMgr, then it won't
    # be rebuilt.
    $self->EventUtil::build_switch_info_hash();
    if ($self->error()) {
        $self->_set_err("Unable to build switch info hash with WWPN and MAC address data");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    dump_switch_info()
# Objective: Dump the switch info hash
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_switch_info {
    my ($self, %args) = @_;

    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");

    print Dumper($switch_info);

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    initiator_switch_port_blocks()
# Objective: Perform initiator switch ports (FC & Ethernet protocols)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_switch_port_blocks {
    my ($self, %args) = @_;

    # perform FCP/iSCSI initiator port blocks
    ## walk through each sanclient and block wwpns/macs
    $self->switch_port_block_abstracted(-port_block_types => ["fc_initiator_wwpn", "iscsi_initiator_mac"]);
    if ($self->error()) {
        $self->_set_err("Initiator port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    fc_initiator_switch_port_blocks()
# Objective: Perform FC initiator switch port blocks
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fc_initiator_switch_port_blocks {
    my ($self, %args) = @_;

    $self->switch_port_block_abstracted(-port_block_types => ["fc_initiator_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC initiator port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    apcon_initiator_port_blocks()
# Objective: Perform FC initiator apcon port blocks
# Details:   See POD documentation at the end of this file.
###############################################################################
sub apcon_initiator_port_blocks {
    my ($self, %args) = @_;

    $self->switch_port_block_abstracted(-port_block_types => ["apcon_initiator_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC initiator port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    iscsi_initiator_switch_port_blocks()
# Objective: Peform iSCSI initiator switch port blocks
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_initiator_switch_port_blocks {
    my ($self, %args) = @_;

    # perform iSCSI initiator port blocks
    ## walk through each sanclient and block macs
    $self->switch_port_block_abstracted(-port_block_types => ["iscsi_initiator_mac"]);
    if ($self->error()) {
        $self->_set_err("iSCSI initiator port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    target_switch_port_blocks()
# Objective: Perform target (FC & Ethernet) switch port blocks
# Details:   See POD documentation at the end of this file.
###############################################################################
sub target_switch_port_blocks {
    my ($self, %args) = @_;

    # perform FC port blocks
    # walk through each sanfiler7m/nacl_nodes_cm/nacl_clusters_cm/vserver object
    # and block wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["fc_target_wwpn", "fc_target_lif_wwpn",
                                                              "iscsi_target_mac","iscsi_target_data_port_mac"]);
    if ($self->error()) {
        $self->_set_err("Target port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    target_switch_port_blocks_7m()
# Objective: Perform target (FC & Ethernet) switch port blocks on 7Mode
#            controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub target_switch_port_blocks_7m {
    my ($self, %args) = @_;

    # perform FC & iSCSI port blocks
    ## walk through each sanfiler7m object and block wwpns/macs
    $self->switch_port_block_abstracted(-port_block_types => ["fc_target_wwpn", "iscsi_target_mac"]);
    if ($self->error()) {
        $self->_set_err("Target (7mode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    target_switch_port_blocks_Cm()
# Objective: Perform target (FC & Ethernet) switch port blocks on CMode
#            controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub target_switch_port_blocks_Cm {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each nacl_clusters_cm object and block lif wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["fc_target_lif_wwpn", "iscsi_target_data_port_mac"]);
    if ($self->error()) {
        $self->_set_err("Target (Cmode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    fc_target_switch_port_blocks()
# Objective: Perform FC target switch port blocks on all controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fc_target_switch_port_blocks {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each sanfiler7m object and block wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["fc_target_wwpn", "fc_target_lif_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC target (7mode/Cmode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    apcon_target_port_blocks()
# Objective: Perform FC target apcon port blocks on all controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub apcon_target_port_blocks {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each sanfiler7m object and controllercluster object and block wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["apcon_target_wwpn", "apcon_target_lif_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC target (7mode/Cmode) apcon port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    iscsi_target_switch_port_blocks()
# Objective: Perform iSCSI target switch port blocks on all controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_target_switch_port_blocks {
    my ($self, %args) = @_;

    # perform iSCSI port blocks
    ## walk through each sanfiler7m object and block macs
    $self->switch_port_block_abstracted(-port_block_types => ["iscsi_target_mac", "iscsi_target_data_port_mac"]);
    if ($self->error()) {
        $self->_set_err("iSCSI target (7mode/Cmode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    fc_target_switch_port_blocks_7m()
# Objective: Perform FC target switch port blocks on all 7Mode controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fc_target_switch_port_blocks_7m {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each sanfiler7m object and block wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["fc_target_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC target (7mode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    apcon_target_port_blocks_7m()
# Objective: Perform FC target apcon port blocks on all 7Mode controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub apcon_target_port_blocks_7m {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each sanfiler7m object and block wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["apcon_target_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC target (7mode) apcon port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    iscsi_target_switch_port_blocks_7m()
# Objective: Perform iSCSI target switch port blocks on all 7Mode controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_target_switch_port_blocks_7m {
    my ($self, %args) = @_;

    # perform iSCSI port blocks
    ## walk through each sanfiler7m object and block macs
    $self->switch_port_block_abstracted(-port_block_types => ["iscsi_target_mac"]);
    if ($self->error()) {
        $self->_set_err("iSCSI target (7mode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    fc_target_switch_port_blocks_Cm()
# Objective: Perform FC target switch port blocks on all CMode controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fc_target_switch_port_blocks_Cm {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each controllercluster object and block lif wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["fc_target_lif_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC target (Cmode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    apcon_target_port_blocks_Cm()
# Objective: Perform FC target apcon port blocks on all CMode controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub apcon_target_port_blocks_Cm {
    my ($self, %args) = @_;

    # perform FC port blocks
    ## walk through each controllercluster object and block lif wwpns
    $self->switch_port_block_abstracted(-port_block_types => ["apcon_target_lif_wwpn"]);
    if ($self->error()) {
        $self->_set_err("FC target (Cmode) apcon port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    iscsi_target_switch_port_blocks_Cm()
# Objective: Perform iSCSI target switch port blocks on all CMode controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_target_switch_port_blocks_Cm {
    my ($self, %args) = @_;

    # perform iSCSI port blocks
    ## walk through each nacl_nodes_cm object and block macs,
    ## if no nacl_nodes_cm objects, then walk through controllercluster objects
    ## and blocks macs of each node in each cluster
    $self->switch_port_block_abstracted(-port_block_types => ["iscsi_target_data_port_mac"]);
    if ($self->error()) {
        $self->_set_err("iSCSI target (Cmode) port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    fc_isl_switch_port_blocks() {
# Objective: Perform FC E-port switch port blocks
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fc_isl_switch_port_blocks {
    my ($self, %args) = @_;

    # Perform eport blocks
    $self->switch_port_block_abstracted(-port_block_types => ["fc_isl_wwpn"]);
    if ($self->error()) {
        $self->_set_err("E-Port port blocks failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    switch_port_block_abstracted()
# Objective: Perform switch port blocks generically for all port block types
# Details:   See POD documentation at the end of this file.
###############################################################################
sub switch_port_block_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @port_block_types = &ParamsMgr::get_arg(-arg => "port_block_types",
                                               -ref_type => "ARRAY",
                                               -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $resources_to_block = ();

    # handle CMode Targets
    if (grep(/fc_target_lif_wwpn|apcon_target_lif_wwpn|iscsi_target_data_port_mac|cluster_port_mac/, @port_block_types)) {
        # choose the cmode targets of the specified type and add them to the
        # list of resources that will be used for port blocks
        my $cmode_target_type = $self->EventUtil::get_arg(-arg => "switch_port_block_cmode_target_type");
        if ($self->error()) {
            $self->_set_err("Unable to determine cmode target_type");
            return $EventUtil::FALSE;
        }

        # Depending on what scope the user specified "node|vservers|nacl_clusters_cm",
        # this code will scope the port block based on that type of CMode resources
        if ($cmode_target_type =~ /node/i) {
            # perform port blocks at the Node level, IE: block all ports represented by each of
            # the physical nodes in the current event scope
            my @nacl_node_names = $self->EventUtil::get_nacl_nodes(-type => "cmode");
            if ($self->error()) {
                $self->_set_err("Could not obtain Cmode NACL Node resources");
                return $EventUtil::FALSE;
            }
            if (scalar(@nacl_node_names)) {
                $resources_to_block->{nacl_nodes_cm} = \@nacl_node_names;
            }
        } elsif ($cmode_target_type =~ /vservers/i) {
            # perform port blocks at the vserver level, IE: block all ports represented by
            # each of the vservers in the current event scope

            # get all current clusters in scope
            my @cluster_names = $self->EventUtil::get_nacl_clusters_cm();
            if ($self->error()) {
                $self->_set_err("Could not obtain Cmode Nacl_clusters_cm");
                return $EventUtil::FALSE;
            }

            # get all vservers in the current event scope and filter them to
            # just the clusters that are in scope
            my @vservers = $self->EventUtil::get_vservers(-clusters => \@cluster_names);
            if ($self->error()) {
                $self->_set_err("Could not obtain Vservers");
                return $EventUtil::FALSE;
            }
            if (scalar(@vservers)) {
                $resources_to_block->{vservers} = \@vservers;
            }
        } elsif ($cmode_target_type =~ /controllerclusters/i) {
            # perform port blocks at the controllercluster level, IE: block all ports represented by
            # each of the clusters in the current event scope
            my @cluster_names = $self->EventUtil::get_nacl_clusters_cm();
            if ($self->error()) {
                $self->_set_err("Could not obtain Cmode Nacl_clusters_cm");
                return $EventUtil::FALSE;
            }
            if (scalar(@cluster_names)) {
                $resources_to_block->{nacl_clusters_cm} = \@cluster_names;
            }
        }
    }

    # handle 7Mode targets
    if (grep(/fc_target_wwpn|apcon_target_wwpn|iscsi_target_mac/, @port_block_types)) {
        my @sanfiler_names = $self->EventUtil::get_sanfilers(-type => "7mode");
        if ($self->error()) {
            $self->_set_err("Could not obtain 7mode SanFilers");
            return $EventUtil::FALSE;
        }
        if (scalar(@sanfiler_names)) {
            $resources_to_block->{sanfilers7m} = \@sanfiler_names;
        }
    }

    # handle FC & iSCSI initiators
    if (grep(/fc_initiator_wwpn|apcon_initiator_wwpn|iscsi_initiator_mac/, @port_block_types)) {
        # get all FC capable sanclients
        my @fc_sanclient_names = $self->EventUtil::get_sanclients(-type => "7mode",
                                                                  -protocol => qr/fc/i);
        if ($self->error()) {
            $self->_set_err("Could not obtain FC SanClients");
            return $EventUtil::FALSE;
        }

        # get all iSCSI capable sanclients
        my @iscsi_sanclient_names = $self->EventUtil::get_sanclients(-type => "7mode",
                                                                     -protocol => qr/iscsi/i);
        if ($self->error()) {
            $self->_set_err("Could not obtain iSCSI SanClients");
            return $EventUtil::FALSE;
        }

        my @sanclients;
        # only obtain the resources that are doing the protocols that we intend to block
        if (grep(/fc_initiator_wwpn/, @port_block_types) and
            grep(/iscsi_initiator_mac/, @port_block_types)) {
            # then we are doing both FC and iSCSI, include both
            @sanclients = (@fc_sanclient_names, @iscsi_sanclient_names);
        } elsif (grep(/fc_initiator_wwpn/, @port_block_types)) {
            # then we are doing only FC initiators
            @sanclients = (@fc_sanclient_names);
        } elsif (grep(/apcon_initiator_wwpn/, @port_block_types)) {
            # then we are doing only FC initiators
            @sanclients = (@fc_sanclient_names);
        } else {
            # then we are doing only iSCSI initiators
            @sanclients = (@iscsi_sanclient_names);
        }

        if (scalar(@sanclients)) {
            # its possible sanclients can be doing more than one protocol
            # which could lead to duplicates in @sanclients
            # dedupe the list by converting to hash and then back
            my %tmp;
            foreach my $client_name (@sanclients) {
                $tmp{$client_name}++;
            }
            @sanclients = keys %tmp;
            $resources_to_block->{sanclients} = \@sanclients;
        }
    }

    # handle switch ISLs
    if (grep(/fc_isl_wwpn/, @port_block_types)) {
        my @sanswitch_names = $self->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Could not obtain switches");
            return $EventUtil::FALSE;
        }
        if (scalar(@sanswitch_names)) {
            $resources_to_block->{sanswitches} = \@sanswitch_names;
        }
    }

    # error out if there are no resources to block
    if (not scalar(keys %{$resources_to_block})) {
        $self->_set_err("There are no resources to perform port blocks for.");
        return $EventUtil::FALSE;
    }

    # determine what algorithm to use (serial/parallel)
    my $switch_port_block_algorithm = $self->EventUtil::get_arg(-arg => "switch_port_block_algorithm");
    if ($self->error()) {
        $self->_set_err("Could not obtain value for switch_port_block_algorithm");
        return $EventUtil::FALSE;
    }

    # Dump the switch topology.
    $self->dump_switch_topology();
    if ($self->error()) {
        # We don't want the errors to cause any issues
        my $errmsg = join("\n", $self->errmsg());
        &SanUtil::nate_log('-comment', "Switch topology dump failed for one or more switches:\n" . $errmsg);
        $self->_clear_err();
    }

    # perform parallel port blocks
    if ($switch_port_block_algorithm =~ /parallel/i) {
        $self->parallel_switch_port_blocks(-resources_to_block => $resources_to_block,
                                           -port_block_types => \@port_block_types);
        if ($self->error()) {
            $self->_set_err("Parallel @port_block_types port blocks failed");
            return $EventUtil::FALSE;
        }
    } elsif ($switch_port_block_algorithm =~ /sequential/i) {
        # otherwise perform sequential port blocks
        $self->sequential_switch_port_blocks(-resources_to_block => $resources_to_block,
                                             -port_block_types => \@port_block_types);
        if ($self->error()) {
            $self->_set_err("Sequential @port_block_types port blocks failed");
            return $EventUtil::FALSE;
        }
    } else {
        $self->_set_err("Unsupported switch port block algorithm '$switch_port_block_algorithm'");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sequential_switch_port_blocks()
# Objective: Disable or Enable a given port on a switch for a set of resources
#            in sequential order
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sequential_switch_port_blocks {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resources_to_block = &ParamsMgr::get_arg(-arg => "resources_to_block",
                                                 -ref_type => "HASH",
                                                 -required => 1);
    my @port_block_types = &ParamsMgr::get_arg(-arg => "port_block_types",
                                               -ref_type => "ARRAY",
                                               -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");

    # collect the sanswitches that will be used for the block
    my (@sanswitches) = $self->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain all SanSwitch resource names");
        return $EventUtil::FALSE;
    }
    
    # initialize the block count variables
    my $current_block_count = 0;
    my $switch_port_total_block_count = $self->EventUtil::get_arg(-arg => "switch_port_total_block_count");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Get the validation type to use
    my $switch_port_block_pv_type = $self->EventUtil::get_arg(-arg => "switch_port_block_pv_type");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # keep a history of which ports have been blocked
    my $port_block_history = {};

    # keep a history of which switches have been reconnected
    my $switch_reconnect_history = {};

    # keep looping until we hit the switch port total block count
    while (1) {

        # perform blocks on all resource_types in $resources_to_block sequentially
        foreach my $resource_type (keys %{$resources_to_block}) {

            # perform blocks on all resources of this resource type
            foreach my $resource_name (@{$resources_to_block->{$resource_type}}) {

                # perform blocks for each port block type that this resource is capable of doing
                foreach my $port_block_type (keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}}) {
                    if (not grep(/$port_block_type/, @port_block_types)) {
                        # skip this port block type if it doesnt match one of the requested types to block
                        next;
                    }

                    my @mac_or_wwpn_list = keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}};

                    if ($self->EventUtil::get_arg(-arg => "switch_port_block_sort_switch_names")) {
                        @mac_or_wwpn_list = $self->_sort_mac_wwpn_list_by_switch_names(
                            -port_block_type_hash => $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type});
                        if ($self->error()) {
                            $self->_set_err("Failed to sort mac/wwpn list by switch names for port block type " .
                                            "$port_block_type and resource $resource_name : $resource_type");
                            return $EventUtil::FALSE;
                        }
                    }

                    # perform blocks for each mac or wwpn this resource has (for the given port block type)
                    foreach my $mac_or_wwpn (@mac_or_wwpn_list) {

                        # if this mac or wwpn was discovered on a specific switch, perform the block
                        if (exists($switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{switch_name})) {
                            my $switch_name = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{switch_name};
                            my $port_numbers = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{port_number};
                            my $vf_context_id = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{vf_context_id};

                            # check if we have hit the total block count (and that the total block count mechanism was enabled >0)
                            if ($current_block_count >= $switch_port_total_block_count and $switch_port_total_block_count > 0) {
                                &SanUtil::nate_log('-comment', "Total block count ($switch_port_total_block_count) has been achieved.");
                                return $EventUtil::TRUE;
                            }

                            # if the switch that this port lives on is not one of the switches in the current scope,
                            # then skip this port
                            if (not grep(/^$switch_name$/, @sanswitches)) {
                                next;
                            }

                            if ((not (defined $switch_reconnect_history->{$switch_name})) or 
                                    $switch_reconnect_history->{$switch_name} ne  "reconnected") {
                                $self->EventUtil::reconnect_disconnected_sanswitches(-switch_names => [$switch_name]);
                                if ($self->error()) {
                                    $self->_set_err("Unable to reconnect to $switch_name");
                                    return $EventUtil::FALSE;
                                }
                                $switch_reconnect_history->{$switch_name} = "reconnected";
                            }

                            foreach my $port_number (@{$port_numbers}) {
                                # check in our table if we have already blocked this port
                                # this can happen if we have multiple resources sharing the same physical port on a switch
                                if (defined $port_block_history->{$switch_name}->{$port_number} and
                                    $port_block_history->{$switch_name}->{$port_number} > 0) {
                                    # then skip this port since we have already blocked it.
                                    &SanUtil::nate_log('-comment', "Skipping port block on switch ($switch_name) for port " .
                                                                   "($port_number) because it was previously blocked/unblocked.");
                                    next;
                                }

                                # perform the disable/enable port block
                                $self->switch_port_block_generic_disable_enable(-switch_name => $switch_name,
                                                                                -port_number => $port_number,
                                                                                -vf_context_id => $vf_context_id,
                                                                                -resource_name => $resource_name,
                                                                                -resource_type => $resource_type);
                                if ($self->error()) {
                                    $self->_set_err("Failed to disable/enable switch port");
                                    return $EventUtil::FALSE;
                                }

                                # keep track of how many total blocks we have done
                                $current_block_count++;

                                # store in the port_block_history that we performed this block
                                $port_block_history->{$switch_name}->{$port_number}++;

                                # get the up time value so it can be passed into path validation
                                # as the minimum path validation min wait time
                                my $switch_port_up_time = $self->EventUtil::get_arg(-arg => "switch_port_up_time");
                                if ($self->error()) {
                                    return $EventUtil::FALSE;
                                }

                                my $id_extension = "switch_port_${port_number}_block_on_${switch_name}__blkcount_${current_block_count}";
                                $id_extension =~ s/\//-/g;

                                # Check to see how they want to run path validation
                                if ($switch_port_block_pv_type =~ /port/i) {
                                    # Check to see if we did FC type blocks.  If so, we can
                                    # do port validation.
                                    if ($port_block_type =~ /^fc_/i) {
                                        $self->PortValidation::validate_ports_after_port_up(-id_extension      => $id_extension,
                                                                                            -switch            => $switch_name,
                                                                                            -ports             => [$port_number],
                                                                                            -last_timeout      => $switch_port_up_time,
                                                                                            -last_timeout_name => "switch_port_up_time");
                                        if ($self->error()) {
                                            $self->_set_err("Port validation failed after port block on " .
                                                            "port '$port_number' on switch '$switch_name'.");
                                            return $EventUtil::FALSE;
                                        }
                                    } else {
                                        &SanUtil::nate_log('-comment', "Skipping port validation on switch " .
                                                           "($switch_name) for port ($port_number) " .
                                                           "because it is not an FC port");
                                    }

                                    # check if we did a target or initiator port block
                                    # to decide how we should handle path validation
                                    if ($port_block_type !~ /initiator/i) {
                                        # Check to see if we are blocking a vserver or nacl_nodes_cm
                                        # resource type
                                        if ($resource_type =~ /nacl_nodes_cm|vserver/i) {
                                            $self->ControllerValidation::validate_controller_after_giveback(-id_extension => $id_extension,
                                                                                                            -last_timeout => $switch_port_up_time,
                                                                                                            -last_timeout_name => "switch_port_up_time");
                                            if ($self->error()) {
                                                $self->_set_err("Controller validation failed after port block on port $port_number on " .
                                                                "switch $switch_name.");
                                                return $EventUtil::FALSE;
                                            }
                                        }

                                        # perform path validation on all initiators since we should
                                        # have just done a single target port block (giveback rules)
                                        $self->PathValidation::validate_paths_after_switch_port_enable(-id_extension => $id_extension,
                                                                                                       -last_timeout => $switch_port_up_time,
                                                                                                       -last_timeout_name => "switch_port_up_time");
                                        if ($self->error()) {
                                            $self->_set_err("Path validation failed after port block on port $port_number on " .
                                                            "switch $switch_name.");
                                            return $EventUtil::FALSE;
                                        }
                                    } else {
                                        # perform path validation on only the initiator resource that was blocked (giveback rules)
                                        $self->PathValidation::validate_paths_after_switch_port_enable(-id_extension => $id_extension,
                                                                                                       -sanclients => [$resource_name],
                                                                                                       -last_timeout => $switch_port_up_time,
                                                                                                       -last_timeout_name => "switch_port_up_time");
                                        if ($self->error()) {
                                            $self->_set_err("Path validation failed after port block on port $port_number on " .
                                                            "switch $switch_name for resource ($resource_name : $resource_type)");
                                            return $EventUtil::FALSE;
                                        }
                                    }
                                }
                            } # end foreach port number
                        }
                    } # end foreach mac_or_wwpn
                } # end foreach capable port block type
            } # end foreach resource_name
        } # end foreach resource type

        # Check to see how they want to run path validation
        if ($switch_port_block_pv_type =~ /cycle/i) {
            my $id_extension = "switch_port_block_cycle";

            # get the up time value so it can be passed into path validation
            # as the validation min wait time
            my $switch_port_up_time = $self->EventUtil::get_arg(-arg => "switch_port_up_time");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }

            # See if we have nacl_nodes_cm or vserver port block types.
            # If we do, then call lif validation
            if (grep(/nacl_nodes_cm|vserver/i, @port_block_types)) {
                $self->ControllerValidation::validate_controller_after_giveback(-id_extension => $id_extension,
                                                                                -last_timeout => $switch_port_up_time,
                                                                                -last_timeout_name => "switch_port_up_time");
                if ($self->error()) {
                    $self->_set_err("Controller validation failed after port block cycle");
                    return $EventUtil::FALSE;
                }
            }

            # Since we are validating after a complete cycle, we will
            # validate all ports.
            $self->PortValidation::validate_ports_after_port_up(-id_extension      => $id_extension,
                                                                -last_timeout      => $switch_port_up_time,
                                                                -last_timeout_name => "switch_port_up_time");
            if ($self->error()) {
                $self->_set_err("Port validation failed after port block cycle");
                return $EventUtil::FALSE;
            }

            # perform path validation on all initiators since we should
            # are testing after a cycle of blocks
            $self->PathValidation::validate_paths_after_switch_port_enable(-id_extension => $id_extension,
                                                                           -last_timeout => $switch_port_up_time,
                                                                           -last_timeout_name => "switch_port_up_time");
            if ($self->error()) {
                $self->_set_err("Path validation failed after port block cycle");
                return $EventUtil::FALSE;
            }
        } ##END if ($switch_port_block_pv_type =~ /cycle/i) {

        # reset the port block history since have just completed an iteration through
        # all resources types
        $port_block_history = {};

        # if no port blocks have been done, then we didnt find anything to block
        # this should be an error
        if ($current_block_count == 0) {
            $self->_set_err("No port blocks were performed, could not find any ports for the given resources");
            return $EventUtil::FALSE;
        }

        # if the total block count was not being used, then break out now
        if ($switch_port_total_block_count <= 0) {
            last;
        }
    } # end while(1)

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    parallel_switch_port_blocks()
# Objective: Disable or Enable a given port on a switch for a set of resources
#            in parallel
# Details:   See POD documentation at the end of this file.
###############################################################################
sub parallel_switch_port_blocks {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resources_to_block = &ParamsMgr::get_arg(-arg => "resources_to_block",
                                                 -ref_type => "HASH",
                                                 -required => 1);
    my @port_block_types = &ParamsMgr::get_arg(-arg => "port_block_types",
                                               -ref_type => "ARRAY",
                                               -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");


    # see if the caller wants us to conserve switch connections, this controls whether we
    # reconnect & disconnect after disabling ports and enabling ports
    # this should allow better conservation of connections when dealing with multiple
    # parallel threads working against the same switch
    my $conserve_connections = $self->EventUtil::get_arg(-arg => "switch_port_block_conserve_switch_connections");

    # perform blocks on all resource_types in parallel
    my @subtests = ();
    foreach my $resource_type (keys %{$resources_to_block}) {

        # perform blocks on each resource of the specified resource type
        foreach my $resource_name (@{$resources_to_block->{$resource_type}}) {

            # spawn off a subtest to perform the port blocks for this resource
            # in the background
            my $subtest = Tharn::subtest(
                sub {

                    # collect the sanswitches that will be used for the block
                    my (@sanswitches) = $self->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE);
                    if ($self->error()) {
                        $self->_set_err("Unable to obtain all SanSwitch resource names");
                        $self->SubtestComm::subtest_fail();
                        return $EventUtil::FALSE;
                    }

                    # keep track of which switches we will be using for blocks, only reconnect and disconnect from
                    # this list of switches once
                    my %switch_connection_hash;

                    # keep track of whether we have performed blocks for this resource
                    my $performed_blocks = 0;

                    # keep track of whether we intend to perform blocks for this resource
                    my $performing_blocks = 0;

                    my $switch_port_block_cycle_count = $self->EventUtil::get_arg(-arg => "switch_port_block_cycle_count");

                    # perform the specified number of port block cycles as denoted by 'switch_port_block_cycle_count'
                    for (my $current_block_count = 1; $current_block_count <= $switch_port_block_cycle_count; $current_block_count++) {

                        # perform blocks for each port block type that this resource is capable of doing
                        foreach my $port_block_type (keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}}) {
                            # if this port block type is not being requested in the list of port block types to block,
                            # skip to the next one in the list
                            if (not grep(/$port_block_type/, @port_block_types)) {
                                next;
                            } else {
                                # indicate that we intend to perform blocks for this initiator
                                $performing_blocks = 1;
                            }

                            my @mac_or_wwpn_list = keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}};

                            if ($self->EventUtil::get_arg(-arg => "switch_port_block_sort_switch_names")) {
                                @mac_or_wwpn_list = $self->_sort_mac_wwpn_list_by_switch_names(
                                    -port_block_type_hash => $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type});
                                if ($self->error()) {
                                    $self->_set_err("Failed to sort mac/wwpn list by switch names for port block type " .
                                                        "$port_block_type and resource $resource_name : $resource_type");
                                    return $EventUtil::FALSE;
                                }
                            }

                            # block the ports for this resource
                            foreach my $mac_or_wwpn (@mac_or_wwpn_list) {
                                if (exists($switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{switch_name})) {
                                    my $switch_name = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{switch_name};
                                    my $port_numbers = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{port_number};
                                    my $vf_context_id = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_block_type}->{$mac_or_wwpn}->{vf_context_id};

                                    # if the switch that this port lives on is not one of the switches in the current scope,
                                    # then skip this port
                                    if (not grep(/^$switch_name$/, @sanswitches)) {
                                        next;
                                    } else {
                                        # otherwise we are now certain that we will perform blocks on this initiator
                                        # set a flag that we will be doing them
                                        $performed_blocks = 1;
                                    }

                                    # walk the list of port numbers
                                    foreach my $port_number (@{$port_numbers}) {

                                        # if we have not already reconnected to this switch, do so here
                                        if (not exists($switch_connection_hash{$switch_name})) {
                                            # indicate that we have reconnected to this switch, so we dont do it again
                                            $switch_connection_hash{$switch_name} = 1;

                                            # if we are not conserving connections, then reconnect right now
                                            if ($conserve_connections eq $EventUtil::FALSE) {
                                                # now reconnect to this switch, leave the original connection intact
                                                # so we dont affect our parent's connection to the switch

                                                $self->EventUtil::reconnect_all_sanswitches(-disconnect_existing_conn => $EventUtil::FALSE,
                                                                                            -switch_names => [$switch_name]);
                                                if ($self->error()) {
                                                    # if we failed, cleanup the siwtches we connected to
                                                    $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                                 -ignore_errors => $EventUtil::TRUE);
                                                    $self->_set_err("Unable to create new connection to SanSwitch '$switch_name'");
                                                    $self->SubtestComm::subtest_fail();
                                                    return $EventUtil::FALSE;
                                                }
                                            }
                                        }

                                        # perform the disable/enable port block on this switch at the specified port
                                        $self->switch_port_block_generic_disable_enable(-switch_name => $switch_name,
                                                                                        -port_number => $port_number,
                                                                                        -vf_context_id => $vf_context_id,
                                                                                        -resource_name => $resource_name,
                                                                                        -resource_type => $resource_type,
                                                                                        -conserve_connections => $conserve_connections);
                                        if ($self->error()) {
                                            $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                         -ignore_errors => $EventUtil::TRUE);
                                            $self->_set_err("Failed to disable/enable switch port '$port_number' on " .
                                                            "switch '$switch_name'");
                                            $self->SubtestComm::subtest_fail();
                                            return $EventUtil::FALSE;
                                        }

                                        # get the up time value so it can be passed into path validation
                                        # as the minimum path validation min wait time
                                        my $switch_port_up_time = $self->EventUtil::get_arg(-arg => "switch_port_up_time");
                                        if ($self->error()) {
                                            $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                         -ignore_errors => $EventUtil::TRUE);
                                            $self->SubtestComm::subtest_fail();
                                            return $EventUtil::FALSE;
                                        }

                                        my $id_extension = "switch_port_${port_number}_block_on_${switch_name}__blkcount_${current_block_count}";
                                        $id_extension =~ s/\//-/g;

                                        # Check to see if we did FC type blocks. If so, we can
                                        # do port validation.
                                        if ($port_block_type =~ /^fc_/i) {
                                            # reconnect to the switch to perform port operation operation
                                            if ($conserve_connections eq $EventUtil::TRUE) {
                                                $self->EventUtil::reconnect_all_sanswitches(-switch_names => [$switch_name]);
                                            }
                                            $self->PortValidation::validate_ports_after_port_up(-id_extension      => $id_extension,
                                                                                                -switch            => $switch_name,
                                                                                                -ports             => [$port_number],
                                                                                                -last_timeout      => $switch_port_up_time,
                                                                                                -last_timeout_name => "switch_port_up_time");
                                            if ($self->error()) {
                                                $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                             -ignore_errors => $EventUtil::TRUE);
                                                $self->_set_err("Port validation failed after port blocks on switch '$switch_name' " .
                                                                "port '$port_number'");
                                                $self->SubtestComm::subtest_fail();
                                                return $EventUtil::FALSE;
                                            }
                                            # disconnect from the switch to conserve connections
                                            if ($conserve_connections eq $EventUtil::TRUE) {
                                                $self->EventUtil::disconnect_all_sanswitches(-switch_names => [$switch_name]);
                                            }
                                        } else {
                                            &SanUtil::nate_log('-comment', "Skipping port validation for port '$port_number' " .
                                                                           "on switch '$switch_name' because  it " .
                                                                           "is not an FC port");
                                        }

                                        # check if we did a target or initiator port block
                                        # to decide how we should handle path validation
                                        if ($port_block_type !~ /initiator/i) {
                                            # Check to see if we are blocking a vserver or nacl_nodes_cm
                                            # resource type
                                            if ($resource_type =~ /nacl_nodes_cm|vserver/i) {
                                                $self->ControllerValidation::validate_controller_after_giveback(-id_extension => $id_extension,
                                                                                                                -last_timeout => $switch_port_up_time,
                                                                                                                -last_timeout_name => "switch_port_up_time");
                                                if ($self->error()) {
                                                    $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                                 -ignore_errors => $EventUtil::TRUE);
                                                    $self->_set_err("Controller validation failed after port block " .
                                                                    "on port $port_number on switch $switch_name.");
                                                    $self->SubtestComm::subtest_fail();
                                                    return $EventUtil::FALSE;
                                                }
                                            }

                                            # perform path validation on all initiators since we should
                                            # have just done a single target port block (giveback rules
                                            $self->PathValidation::validate_paths_after_switch_port_enable(-id_extension => $id_extension,
                                                                                                           -last_timeout => $switch_port_up_time,
                                                                                                           -last_timeout_name => "switch_port_up_time");
                                            if ($self->error()) {
                                                $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                             -ignore_errors => $EventUtil::TRUE);
                                                $self->_set_err("Path validation failed after port block on port $port_number on " .
                                                                "switch $switch_name.");
                                                $self->SubtestComm::subtest_fail();
                                                return $EventUtil::FALSE;
                                            }
                                        } else {
                                            # perform path validation on only the initiator resource that was blocked (giveback rules)
                                            $self->PathValidation::validate_paths_after_switch_port_enable(-id_extension => $id_extension,
                                                                                                           -sanclients => [$resource_name],
                                                                                                           -last_timeout => $switch_port_up_time,
                                                                                                           -last_timeout_name => "switch_port_up_time");
                                            if ($self->error()) {
                                                $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                                             -ignore_errors => $EventUtil::TRUE);
                                                $self->_set_err("Path validation failed after port block on port $port_number on " .
                                                                "switch $switch_name for resource ($resource_name : $resource_type)");
                                                $self->SubtestComm::subtest_fail();
                                                return $EventUtil::FALSE;
                                            }
                                        }
                                    } # end foreach port number
                                }
                            } # end foreach mac_or_wwpn
                        } # end foreach port_block_type

                        # if we intended to perform blocks, but none were done, then warn and break
                        if ($performing_blocks == 1 and $performed_blocks == 0) {
                            &SanUtil::nate_log('-warn', "No blocks were performed for resource ('$resource_name' : $resource_type)");
                            last;
                        } elsif ($performing_blocks == 0) {
                            # if we never intended to perform blocks, then just break out
                            last;
                        }
                    } # end for (1..$switch_port_block_cycle_count)

                    # disconnect from the switches that we reconnected to during the port blocking
                    if (scalar(keys %switch_connection_hash)) {
                        $self->EventUtil::disconnect_all_sanswitches(-switch_names => [keys %switch_connection_hash],
                                                                     -ignore_errors => $EventUtil::TRUE);
                    } else {
                        # otherwise there is no way we could have performed any blocks because
                        # we never reconnected to any switch, this is a failure
                        $self->_set_err("No switch port blocks were performed for the resource ('$resource_name' : $resource_type)");
                        $self->SubtestComm::subtest_fail();
                        return $EventUtil::FALSE;
                    }

                    # we have completed port blocks for this resource, tell the parent we are done
                    $self->SubtestComm::subtest_complete();
                    return $EventUtil::TRUE;
                }, -runid => "switch_port_blocks_for_${resource_name}", -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

            # push all parallel block subtests on the list
            push(@subtests, $subtest);
        } # end foreach resource
    } # end foreach resource type

    # reap all the parallel port block subtests. Keep them independent of one another so
    # that a failure in one does not interrupt any of the others
    # enable sync only if requested to
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => $EventUtil::FALSE,
                             -propagate_events => $EventUtil::TRUE,
                             -enable_sync => $self->EventUtil::get_arg(-arg => "switch_port_block_sync_parallel_blocks"),
                             -check_cmds_from_parent => $EventUtil::FALSE,
                             -forward_cmds_from_parent => $EventUtil::FALSE,
                             -async => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Failed while reaping switch port block subtests");
        return $EventUtil::TRUE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    switch_port_block_generic_disable_enable()
# Objective: Disable and enable a port for a particular resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub switch_port_block_generic_disable_enable {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $port_number = &ParamsMgr::get_arg(-arg => "port_number",
                                          -required => 1);
    my $switch_name = &ParamsMgr::get_arg(-arg => "switch_name",
                                          -required => 1);
    my $resource_name = &ParamsMgr::get_arg(-arg => "resource_name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "resource_type",
                                            -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Grab the SanSwitch object
    my $sanswitch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanSwitch object for switch '$switch_name'");
        return $EventUtil::FALSE;
    }

    # Grab the PortMgr object
    my $pm_obj = $sanswitch_obj->portmgr();
    unless (ref($pm_obj) =~ /PortMgr/i) {
        # Looks like we don't have a portmgr object on the sanswitch object.  Try creating one.
        $sanswitch_obj->create_portmgr(-load_nameserver => 1);
        if ($sanswitch_obj->error()) {
            $self->_set_err("Unable to create a portmgr object for switch '$switch_name':\n",
                            $sanswitch_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # Grab the nameserver information
    my $ns = $pm_obj->get_nameserver("1");
    if ($pm_obj->error) {
        $self->_set_err("Unable to grab the nameserver information for switch '$switch_name'\n",
                        $pm_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # Define the block types based on whether or not we have an FCoE switch
    my @block_types = ();
    # check for per resource switch_port_block_fcoe_block_type
    my @param_resources;
    my $resource = $self->ResMgr::get_resource(-name => $resource_name,
                                               -type => $resource_type);
    if ($self->error()) {
        $self->_set_err("Could not get resource '$resource_name' of type '$resource_type'");
        return $EventUtil::FALSE;
    }
    push(@param_resources, $resource);
    # Looks like an FCoE Switch.  Determine what kind of blocks are requested.
    # Get the resource for the switch
    my $switch_resource = $self->ResMgr::get_resource(-name => $switch_name,
                                                      -type => "sanswitches");
    if ($self->error()) {
        $self->_set_err("Could not get resource '$switch_name' of type 'sanswitches'");
        return $EventUtil::FALSE;
    }
    push(@param_resources, $switch_resource);

    # Grab the value for switch_port_block_fcoe_block_type
    my $switch_port_block_fcoe_block_type = $self->EventUtil::get_arg(-arg => "switch_port_block_fcoe_block_type",
                                                                      -objects => \@param_resources,
                                                                      -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Could not retrieve parameter 'switch_port_block_fcoe_block_type' for switch " .
                        "resource '$switch_name' nor for the environment");
        return $EventUtil::FALSE;
    }

    # burt737851 (only block ethernet switches ports at the physical layer, otherwise we will block the port twice)
    if (ref($sanswitch_obj) =~ /Ethernet/i) {
        $switch_port_block_fcoe_block_type = "physical";
    }

    if ($switch_port_block_fcoe_block_type eq "both") {
        &SanUtil::nate_log("-comment", "Both physical and virtual blocks requested for '$switch_name' " .
                                       "port '$port_number''");
        push(@block_types, "physical");
        push(@block_types, "virtual");
    } elsif ($switch_port_block_fcoe_block_type =~ /^(virtual|physical)/i) {
        &SanUtil::nate_log("-comment", "$switch_port_block_fcoe_block_type blocks requested for '$switch_name' " .
                                      "port '$port_number''");
        push(@block_types, lc($switch_port_block_fcoe_block_type));
    } else {
        if ($self->error()) {
            $self->_set_err("Invalid fcoe block type of '$switch_port_block_fcoe_block_type'");
            return $EventUtil::FALSE;
        }
    }
    foreach my $block_type (@block_types) {
        if ($block_type eq "physical") {
            # They want physical blocks so set the fcoe_ethernet_block value to 1
            $sanswitch_obj->set_fcoe_ethernet_block(1);
        } else {
            # They want physical blocks so set the fcoe_ethernet_block value to 0
            $sanswitch_obj->set_fcoe_ethernet_block(0);
        }
        if ($sanswitch_obj->error()) {
            $self->_set_err("Unable to set the fcoe_ethernet_block attribute on switch '$switch_name':\n",
                            $sanswitch_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # disable the port
        $self->disable_switch_port(%args);
        if ($self->error()) {
            $self->_set_err("Failed to disable switch port '$port_number' on switch " .
                            "'$switch_name' for resource ('$resource_name' : $resource_type)");
            return $EventUtil::FALSE;
        }

        # synchronize so that any parallel port blocks perform blocks in lock step
        # (if enable_sync is turned on)
        $self->EventUtil::synchronize();

        # wait the specified port down time
        $self->EventUtil::sleep_arg(-arg => "switch_port_down_time");
        if ($self->error()) {
            $self->_set_err("Failed to sleep 'switch_port_down_time'");
            return $EventUtil::FALSE;
        }

        # enable the port
        $self->enable_switch_port(%args);
        if ($self->error()) {
            $self->_set_err("Failed to enable switch port '$port_number' on switch " .
                            "'$switch_name' for resource ('$resource_name' : $resource_type)");
            return $EventUtil::FALSE;
        }

        # synchronize so that any parallel port blocks perform blocks in lock step
        # (if enable_sync is turned on)
        $self->EventUtil::synchronize();

        # wait the specified port up time
        $self->EventUtil::sleep_arg(-arg => "switch_port_up_time");
        if ($self->error()) {
            $self->_set_err("Failed to sleep 'switch_port_up_time'");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disable_switch_port()
# Objective: Disable a port on the specified switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable_switch_port {
    my ($self, %args) = @_;

    # invoke the generic disable method
    $self->_switch_port_disable_enable(%args,
                                       -action => "disable");
    if ($self->error()) {
        $self->_set_err("Failed to disable port");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    enable_switch_port()
# Objective: Enable a port on the specified switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable_switch_port {
    my ($self, %args) = @_;

    # invoke the generic enable method
    $self->_switch_port_disable_enable(%args,
                                       -action => "enable");
    if ($self->error()) {
        $self->_set_err("Failed to enable port");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    dump_switch_topology()
# Objective: Dump the switch topology from all switches
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_switch_topology {
    my ($self) = @_;

    # skip dump switch topology if it was disabled
    if ($self->EventUtil::get_arg(-arg => "switch_port_block_dump_topology") == $EventUtil::FALSE) {
        return $EventUtil::FALSE;
    }

    # get the next padded run id in the event
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $runid_padded = $event_obj->increment_subtest_runid_counter();

    # Dump the switch topology.
    my $subtest = subtest(sub {
                              $self->_dump_switch_topology();
                              if ($self->error()) {
                                  $self->_set_err("Switch topology dump failed for one or more switches");

                                  # perform a graceful failure since we don't want this to kill the test
                                  $self->SubtestComm::subtest_fail(-graceful => 1);
                                  return $EventUtil::FALSE;
                              }
                              $self->SubtestComm::subtest_complete();
                              return $EventUtil::TRUE;
                          },
                          -bg, -runid => "${runid_padded}_switch_topology",
                          "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

    # wait for all the switch validation subtests to complete
    $self->wait_for_subtests(-subtests => [$subtest],
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        my $errmsg = join("\n", $self->errmsg());
        &SanUtil::nate_log('-comment', "Switch topology dump failed for one or more switches:\n" . $errmsg);
        $self->_set_err("Switch topology dump failed for one or more switches");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reboot_switches()
# Objective: Reboot all the switches provided
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reboot_switches {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @sanswitches = &ParamsMgr::get_arg(-arg => "sanswitches",
                                          -ref_type => "ARRAY",
                                          -required => 1);
    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $padded_runid = $event_obj->increment_subtest_runid_counter();

    # fork off threads for each sanswitch object and reboot the nodes
    # return all threads to the caller
    my @subtests;
    foreach my $switch (@sanswitches) {
        my $sub = Tharn::subtest(
            sub {
                # reboot the switch
                my $switch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch);
                if ($self->error()) {
                    $self->_set_err("Failed to obtain switch object");
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }

                $switch_obj->reboot();
                if ($switch_obj->error()) {
                    $self->_set_err("Failed to reboot $switch: " . $switch_obj->errmsg());
                    $self->SubtestComm::subtest_fail();
                    return $EventUtil::FALSE;
                }

                $self->SubtestComm::subtest_complete();
                return $EventUtil::TRUE;
            },
            -runid => $padded_runid . "_reboot_switch__${switch}",
            -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $sub);
    }

    if ($async eq $EventUtil::FALSE) {
        # wait for the switch reboot subtests to complete
        $self->wait_for_subtests(-subtests => \@subtests);
        if ($self->error()) {
            $self->_set_err("Encountered failure while reaping switch reboot subtests");
            return ();
        }
    }

    return @subtests;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _switch_port_disable_enable()
# Objective: Disable or Enable a given port on a switch
###############################################################################
sub _switch_port_disable_enable {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $port_number = &ParamsMgr::get_arg(-arg => "port_number",
                                          -required => 1);
    my $switch_name = &ParamsMgr::get_arg(-arg => "switch_name",
                                          -required => 1);
    my $resource_name = &ParamsMgr::get_arg(-arg => "resource_name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "resource_type",
                                            -required => 1);
    my $vf_context_id = &ParamsMgr::get_arg(-arg => "vf_context_id",
                                            -default => undef);
    my $action = &ParamsMgr::get_arg(-arg => "action",
                                     -required => 1);
    my $conserve_connections = &ParamsMgr::get_arg(-arg => "conserve_connections",
                                                   -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $sanswitch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanSwitch object for switch '$switch_name'");
        return $EventUtil::FALSE;
    }

    # reconnect to the switch to perform the operation
    if ($conserve_connections eq $EventUtil::TRUE) {
        $self->EventUtil::reconnect_all_sanswitches(-switch_names => [$switch_name]);
    }

    # if virtual fabrics are enabled, then set the context to the vf context id
    # of this port.
    if ($sanswitch_obj->is_vf_enabled() and defined($vf_context_id)) {
        $sanswitch_obj->set_vf_context($vf_context_id);
        if ($sanswitch_obj->error()) {
            $self->_set_err($sanswitch_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # disable or enable this port
    if ($action eq "enable") {
        &SanUtil::nate_log("-comment", "Enabling port ($port_number) on switch ($switch_name) for resource ($resource_name : $resource_type)");

        # enable the port
        $sanswitch_obj->portmgr()->port_enable($port_number);

        # disconnect from the switch to conserve connections
        if ($conserve_connections eq $EventUtil::TRUE) {
            $self->EventUtil::disconnect_all_sanswitches(-switch_names => [$switch_name]);
        }

        $self->SysLogging::syslog_msg_all(-message => ["**EventMan** Enabled port ($port_number) " .
                                          "on switch ($switch_name) for resource ($resource_name : $resource_type)"],
                                          -resource_name => $resource_name,
                                          -resource_type => $resource_type,
                                          -async => $EventUtil::TRUE);
        if ($self->error()) {
            $self->_set_err("Syslogging failed");
            return $EventUtil::FALSE;
        }

    } elsif ($action eq "disable") {
        $self->SysLogging::syslog_msg_all(-message => ["**EventMan** Disabling port ($port_number) " .
                                          "on switch ($switch_name) for resource ($resource_name : $resource_type)"],
                                          -resource_name => $resource_name,
                                          -resource_type => $resource_type,
                                          -async => $EventUtil::TRUE);
        if ($self->error()) {
            $self->_set_err("Syslogging failed");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log("-comment", "Disabling port ($port_number) on switch ($switch_name) for resource ($resource_name : $resource_type)");

        # disable the port
        $sanswitch_obj->portmgr()->port_disable($port_number);

        # disconnect from the switch if requested to reconnect during the enable/disable
        if ($conserve_connections eq $EventUtil::TRUE) {
            $self->EventUtil::disconnect_all_sanswitches(-switch_names => [$switch_name]);
        }

    } else {
        $self->_set_err("Invalid action '$action', use 'disable' or 'enable'");
        return $EventUtil::FALSE;
    }

    # check if there was an error during the enable/disable
    if ($sanswitch_obj->portmgr()->error()) {
        # check to see if this error was a miscopare error and handle it appropriately.
        my @portmgr_errmsg = $sanswitch_obj->portmgr()->errmsg();
        if (grep(/Miscompare on/i, @portmgr_errmsg) > 0) {
            # Looks like a miscompare
            &SanUtil::nate_log('-warn', "Port came back online after the port_enable, but there was a miscompare of " .
                                        "port speed or type.\n" .
                                        "Port: $port_number  Switch: $switch_name  Resource: $resource_name\n" .
                                        "Reason: @portmgr_errmsg");

            # dump switch port info if requested
            $self->_dump_switch_port_info(-switch_name => $switch_name, -port_number => $port_number);

            # See if the user wants to ignore miscompare errors. If so, then
            if ($self->EventUtil::get_arg(-arg => "switch_port_block_miscompare_abort_on_error") == $EventUtil::FALSE) {
                # User doesn't want to abort on miscompare. Clear the portmgr error and return true
                $sanswitch_obj->portmgr()->_clear_err();
                return $EventUtil::TRUE;
            } else {
                # User wants to abort on miscompare
                $self->_set_err(@portmgr_errmsg);

                return $EventUtil::FALSE;
            }
        } else {
            # Not a miscompare
            &SanUtil::nate_log('-warn', "Failed to $action port ($port_number) on switch ($switch_name)" .
                                        "for resource ($resource_name)\n" .
                                        "Reason: @portmgr_errmsg");

            # dump switch port info if requested
            $self->_dump_switch_port_info(-switch_name => $switch_name, -port_number => $port_number);

            # Check to see if switch_port_block_abort_on_error was set
            if ($self->EventUtil::get_arg(-arg => "switch_port_block_abort_on_error") == $EventUtil::FALSE) {
                # User doesn't want to abort on switch port block errors.
                # Clear the portmgr error and return true
                $sanswitch_obj->portmgr()->_clear_err();
                return $EventUtil::TRUE;
            } else {
                # User does want to abort on switch port block errors
                $self->_set_err(@portmgr_errmsg);

                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _dump_switch_port_info()
# Objective: Given a switch and a port, dump the info on that port.
###############################################################################
sub _dump_switch_port_info {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $switch_name = &ParamsMgr::get_arg(-arg => "switch_name",
                                          -required => 1);
    my $port_number = &ParamsMgr::get_arg(-arg => "port_number",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if requested to dump switch port info, do so now
    if ($self->EventUtil::get_arg(-arg => "switch_port_block_dump_info_on_failure") == $EventUtil::TRUE) {
        my $sanswitch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
        my $mod_pn = $port_number;
        $mod_pn =~ s/\//_/g;
        $mod_pn =~ s/\\/_/g;
        my $runid = "${switch_name}_port${mod_pn}_dump";

        # subtest off the task to dump the port information in a foreground subtest
        Tharn::subtest(sub { $sanswitch_obj->portmgr()->port_dump_info($port_number) } ,
                       -runid, $runid);
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _dump_switch_topology()
# Objective: Given a switch, dump the topology info
###############################################################################
sub _dump_switch_topology {
    my ($self) = @_;

    # collect the sanswitches that will be used for the block
    my (@sanswitches) = $self->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain all SanSwitch resource names");
        return $EventUtil::FALSE;
    }

    my @subtests = ();
    foreach my $switch_name (@sanswitches) {
        my $runid = $switch_name . "_topology_dump";

        my $subtest = subtest(sub {
                                  # Reconnect the switch
                                  $self->EventUtil::reconnect_all_sanswitches(-disconnect_existing_conn => $EventUtil::FALSE,
                                                                              -switch_names => [$switch_name]);
                                  if ($self->error()) {
                                      $self->_set_err("Switch reconnect failed for " .
                                                          "switch '$switch_name'");

                                      # perform a graceful failure since we don't want this to kill the test.
                                      $self->SubtestComm::subtest_fail(-graceful => 1);
                                      return $EventUtil::FALSE;
                                  }

                                  # Get the switch object
                                  my $sanswitch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
                                  if ($self->error() or ref($sanswitch_obj) !~ /sanswitch/i) {
                                      $self->_set_err("Failed to retrieve the switch object " .
                                                      "for switch '$switch_name'");

                                      # Disconnect the new connection we created
                                      $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch_name],
                                                                                   -ignore_errors => $EventUtil::TRUE);

                                      # perform a graceful failure since we don't want this to kill the test.
                                      $self->SubtestComm::subtest_fail(-graceful => 1);
                                      return $EventUtil::FALSE;
                                  }

                                  # Dump the topology
                                  $sanswitch_obj->dump_topology();
                                  if ($sanswitch_obj->error()) {
                                      my $errmsg = join("\n", $sanswitch_obj->errmsg());
                                      &SanUtil::nate_log('-comment', "Switch topology dump failed for " .
                                                             "switch '$switch_name':\n" . $errmsg);

                                      # Disconnect the new connection we created
                                      $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch_name],
                                                                                   -ignore_errors => $EventUtil::TRUE);

                                      # perform a graceful failure since we don't want this to kill the test.
                                      $self->SubtestComm::subtest_fail(-graceful => 1);
                                      return $EventUtil::FALSE;

                                  }

                                  # Disconnect the new connection we created
                                  $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch_name],
                                                                               -ignore_errors => $EventUtil::TRUE);

                                  $self->SubtestComm::subtest_complete();
                                  return $EventUtil::TRUE;
                              },
                              -runid => $runid, -bg,
                              "-onexit" => sub{my ($self) = @_; $self->exithandler(1);}
                          );

        push(@subtests, $subtest);
    }

    # wait for all the switch validation subtests to complete
    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        $self->_set_err("Switch topology dump failed for one or more switches\n");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _sort_mac_wwpn_list_by_switch_names()
# Objective: Given a port block type hash, sort the mac or wwpn list based on
# the switch names.
###############################################################################
sub _sort_mac_wwpn_list_by_switch_names {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $port_block_type_hash = &ParamsMgr::get_arg(-arg => "port_block_type_hash",
                                                   -ref_type => "HASH",
                                                   -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }


    # grab all the mac or wwpns that we intend to block
    my %tmp_hash;
    my @mac_or_wwpn_list = keys %{$port_block_type_hash};
    # create a switch hash that we can use to sort the switch names
    foreach my $mac_or_wwpn (@mac_or_wwpn_list) {
        if (exists($port_block_type_hash->{$mac_or_wwpn}->{switch_name})) {
            my $switch_name = $port_block_type_hash->{$mac_or_wwpn}->{switch_name};
            push(@{$tmp_hash{$switch_name}}, $mac_or_wwpn);
        }
    }
    # recreate the mac or wwpn list array so that the order is based on the sorted switches
    @mac_or_wwpn_list = ();
    my @sorted_switches = sort(keys %tmp_hash);
    foreach my $sorted_switch (@sorted_switches) {
        foreach my $mac_or_wwpn (@{$tmp_hash{$sorted_switch}}) {
            push(@mac_or_wwpn_list, $mac_or_wwpn);
        }
    }

    return (wantarray()) ? @mac_or_wwpn_list : \@mac_or_wwpn_list;
}

=pod

=head1 Package Name

SwitchEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for Switch Events (port blocks & switch failures)


=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-switch_port_down_time>

=over 4

Description: Specifies how many seconds to leave the port down after the port is disabled on the switch

Resource Types: sanswitches

Per Resource Capable: No

Default: 30

=back

=head2 I<-switch_port_up_time>

=over 4

Description: Specifies how many seconds to wait after enabling the port on the switch

Resource Types: sanswitches

Per Resource Capable: No

Default: 30

Format: Integer

=back

=head2 I<-switch_port_block_dump_info_on_failure>

=over 4

Description: Specifies whether to dump port information when an error is discovered during port blocks

Resource Types: sanswitches

Per Resource Capable: No

Default: 0 (off)

Format: 0|1

=back

=head2 I<-switch_port_block_algorithm>

=over 4

Description: Specifies whether to perform parallel or sequential switch port blocks.

Resource Types: sanswitches

Per Resource Capable: No

Default: sequential

Format: sequential|parallel

=back

=head2 I<-switch_port_total_block_count>

=over 4

Description: Specifies how many total port blocks to perform. If set to '0', will perform at least one block
per port specified/discovered. Only applicable when -switch_port_block_algorithm => "sequential"

Resource Types: sanswitches

Per Resource Capable: No

Default: 0

Format: Integer

=back

=head2 I<-switch_port_block_cycle_count>

=over 4

Description: Specifies how many cycles of port blocks to perform per resource. Only applicable when
-switch_port_block_algorithm => "parallel"

Resource Types: sanswitches

Per Resource Capable: No

Default: 0

Format: Integer

=back

=head2 I<-switch_port_block_cmode_target_type>

=over 4

Description: Specifies from which point of view Cmode targets should be chosen when performing target port
blocks, at the Node, Vserver, or ControllerCluster level.

Resource Types: sanswitches

Per Resource Capable: No

Default: node

Format: node|vservers|nacl_clusters_cm

=back

=head2 I<-switch_port_block_pv_type>

=over 4

Description: Specifies whether path validation should be done after every port block or after every cycle

Resource Types: sanswitches

Per Resource Capable: No

Default: port

Format: port|cycle

=back

=head2 I<-switch_port_block_abort_on_error>

=over 4

Description: Specifies whether the event should error out after a port block failure

Resource Types: sanswitches

Per Resource Capable: No

Default: 1 (on)

Format: 0|1

=back

=head2 I<-switch_port_block_miscompare_abort_on_error>

=over 4

Description: Specifies whether the event should error out miscompare of type or speed after a port block

Resource Types: sanswitches

Per Resource Capable: No

Default: 1 (on)

Format: 0|1

=back

=head2 I<-switch_port_block_dump_topology>

=over 4

Description: Specifies whether the event should dump the switch fabric topology before each event

Resource Types: sanswitches

Per Resource Capable: No

Default: 1 (on)

Format: 0|1

=back

=head2 I<-switch_port_block_conserve_switch_connections>

=over 4

Description: Specifies whether switch connections should be preserved while performing parallel port blocks.
Connections to switches will be opened and closed when they are needed.

Per Resource Capable: No

Default: 0 (off)

Format: 0|1

=back

=head2 I<-switch_port_block_sync_parallel_blocks>

=over 4

Description: Specifies whether to perform parallel port blocks in lock step with one another. Before and after
enabling a port, all parallel port blocks will synchronize with one another.

Per Resource Capable: No

Default: 0 (off)

Format: 0|1

=back

=head2 I<-switch_port_block_fcoe_block_type>

=over 4

Description: Specifies the type of port block to do with FCoE ports. Valid options are physical, virtual, or both.

Resource Types: sanswitches, sanclients, nacl_clusters_cm, sanfilersCm

Per Resource Capable: Yes

Default: both

Format: physical|virtual|both

=back

=head2 I<-switch_port_block_sort_switch_names>

=over 4

Description: Specifies whether to sort the list of switch names before performing port blocks.

Per Resource Capable: No

Default: 0

Format: 0|1

=back

=head1 B<Event Keywords>

=head2 B<initiator_switch_port_blocks>

Performs initiator switch port blocks (all SAN protocols)

=over 4

=item Event Parameters

=item *I<-sanclients>

Required: An array reference of SanClient resources to block

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform initiator port blocks on the two sanclient resources specified,
 # use the discovered ports on the four sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "initiator_switch_port_blocks",
                         -sanclients => ["ibm3550-rtp60", "ibm3550-rtp61"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02", "cisco4948-rtp01", "cisco4948-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<apcon_target_port_blocks_Cm>

Performs Apcon target switch port blocks for Cmode

=over 4

=item Event Parameters

=item *I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform Apcon target port blocks on the four vserver resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "apcon_target_port_blocks_Cm",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanswitches => ["apcon10g-01""]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform Apcon target port blocks on the two cluster resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "apcon_target_port_blocks_Cm",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanswitches => ["apcon10g-01""]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<apcon_target_port_blocks_7m>

Performs Apcon target switch port blocks for 7mode

=over 4

=item Event Parameters

=item *I<-sanfilers7m>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform Apcon target port blocks on the two 7mode filer  resources specified.
 # Use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "apcon_target_port_blocks_7m",
                         -sanfilers7m => ["filer1", "filer2"],
                         -sanswitches => ["apcon10g-01""]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<apcon_target_port_blocks>

Performs Apcon target switch port blocks for Cmode and 7mode combined

=over 4

=item Event Parameters

=item *I<-nacl_clusters_cm> or I<-vservers> or I<-sanfilers7m>

Required: An array reference of Target resources to perform apcon port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform Apcon target port blocks on the four vserver resources and sanfilers7m resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "apcon_target_port_blocks",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanfilers7m => ["filer1", "filer2"],
                         -sanswitches => ["apcon10g-01""]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform Apcon target port blocks on the two cluster resources and 2 sanfilers7m resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "apcon_target_port_blocks",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanfilers7m => ["filer1", "filer2"],
                         -sanswitches => ["apcon10g-01""]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<apcon_initiator_port_blocks>

Performs Apcon initiator port blocks

=over 4

=item Event Parameters

=item *I<-sanclients>

Required: An array reference of Initiator resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform Apcon initiator port blocks on the three sanclient resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "apcon_initiator_port_blocks",
                         -vservers => ["client1", "client2", "client3"],
                         -sanswitches => ["apcon10g-01""]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<fc_initiator_switch_port_blocks>

Performs FC initiator switch port blocks

=over 4

=item Event Parameters

=item *I<-sanclients>

Required: An array reference of SanClient resources to block

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform FC initiator port blocks on the two sanclient resources specified,
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_initiator_switch_port_blocks",
                         -sanclients => ["ibm3550-rtp60", "ibm3550-rtp61"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<iscsi_initiator_switch_port_blocks>

Performs iSCSI initiator switch port blocks

=over 4

=item Event Parameters

=item *I<-sanclients>

Required: An array reference of SanClient resources to block

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform iSCSI initiator port blocks on the two sanclient resources specified,
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_initiator_switch_port_blocks",
                         -sanclients => ["ibm3550-rtp60", "ibm3550-rtp61"],
                         -sanswitches => ["cisco4948-rtp01", "cisco4948-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<target_switch_port_blocks>

Performs target switch port blocks (all SAN protocols, both CMode & 7Mode targets)

=over 4

=item Event Parameters

=item *I<-nacl_nodes_cm> or I<-sanfilers7m> or I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform target port blocks on the four nacl_nodes_cm resources
 # perform target port blocks on a single sanfilers7M resource
 # use the discovered ports on the four sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "target_switch_port_blocks",
                         -sanfilers7m => ["fas960-rtp10"],
                         -nacl_nodes_cm => ["fas3070-rtp60", "fas3070-rtp61", "fas3170-rtp62", "fas3170-rtp63"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02", "cisco4948-rtp01", "cisco4948-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # perform target port blocks on the two vservers specified
 $EventManobj->add_event(-event_name => "target_switch_port_blocks",
                         -vservers => ["vs0", "vs1"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02", "cisco4948-rtp01", "cisco4948-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # perform target port blocks on all vservers in the cluster 'myclus'
 $EventManobj->add_event(-event_name => "target_switch_port_blocks",
                         -nacl_clusters_cm => ["myclus"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02", "cisco4948-rtp01", "cisco4948-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<target_switch_port_blocks_7m>

Performs FC & iSCSI target switch port blocks for 7mode

=over 4

=item Event Parameters

=item *I<-sanfilers7m>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform FC & iSCSI 7mode target port blocks on the three sanfiler7m resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "target_switch_port_blocks_7m",
                         -sanfilers7m => ["fas960-rtp01", "fas960-rtp02", "fas2050-rtp50"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<target_switch_port_blocks_Cm>

Performs FC & iSCSI target switch port blocks for Cmode

=over 4

=item Event Parameters

=item *I<-nacl_nodes_cm> or I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the two nacl_nodes_cm resources
 # use the discovered ports on the four sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "target_switch_port_blocks_Cm",
                         -nacl_nodes_cm => ["fas3070-rtp60", "fas3070-rtp61"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02", "cisco4948-rtp01", "cisco4948-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform FC & iSCSI target port blocks on the four vserver resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "target_switch_port_blocks_Cm",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform FC & iSCSI target port blocks on the two cluster resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "target_switch_port_blocks_Cm",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<fc_target_switch_port_blocks>

Performs FC target switch port blocks for both Cmode and 7mode

=over 4

=item Event Parameters

=item *I<-sanfilers7m> or I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform FC target port blocks on the three sanfiler7m resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_target_switch_port_blocks",
                         -sanfilers7m => ["fas960-rtp01", "fas960-rtp02", "fas2050-rtp50"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform FC target port blocks on the four vserver resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_target_switch_port_blocks",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform FC target port blocks on the two cluster resources specified
 # as well as the 7mode controller specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_target_switch_port_blocks",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanfilers7m => ["fas960-rtp01"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<fc_target_switch_port_blocks_7m>

Performs FC target switch port blocks for 7mode

=over 4

=item Event Parameters

=item *I<-sanfilers7m>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform FC 7mode target port blocks on the three sanfiler7m resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_target_switch_port_blocks_7m",
                         -sanfilers7m => ["fas960-rtp01", "fas960-rtp02", "fas2050-rtp50"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<fc_target_switch_port_blocks_Cm>

Performs FC target switch port blocks for Cmode

=over 4

=item Event Parameters

=item *I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform FC target port blocks on the four vserver resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_target_switch_port_blocks_Cm",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform FC target port blocks on the two cluster resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "fc_target_switch_port_blocks_Cm",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<iscsi_target_switch_port_blocks>

Performs iSCSI target switch port blocks for both Cmode and 7mode

=over 4

=item Event Parameters

=item *I<-sanfilers7m> or I<-nacl_nodes_cm> or I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the three sanfilers7m resources specified
 # perform iSCSI target port blocks on the nacl_nodes_cm resource specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks",
                         -sanfilers7m => ["fas960-rtp01", "fas960-rtp02", "fas2050-rtp50"],
                         -nacl_nodes_cm => ["fas3070-rtp50"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the four vserver resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the two cluster resources specified
 # as well as the 7mode controller specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanfilers7m => ["fas960-rtp01"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<iscsi_target_switch_port_blocks_7m>

Performs FC target switch port blocks for 7mode

=over 4

=item Event Parameters

=item *I<-sanfilers7m>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform iSCSI 7mode target port blocks on the three sanfiler7m resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks_7m",
                         -sanfilers7m => ["fas960-rtp01", "fas960-rtp02", "fas2050-rtp50"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<iscsi_target_switch_port_blocks_Cm>

Performs iSCSI target switch port blocks for Cmode

=over 4

=item Event Parameters

=item *I<-nacl_nodes_cm> or I<-nacl_clusters_cm> or I<-vservers>

Required: An array reference of Target resources to perform port blocks against.

=item *I<-sanswitches>

Required: An array reference of SanSwitches to use when performing port blocks.

=item Examples

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the nacl_nodes_cm resource specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks_Cm",
                         -nacl_nodes_cm => ["fas3070-rtp50"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the four vserver resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks_Cm",
                         -vservers => ["vs0", "vs1", "vs2", "vs3"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

 # see EventMan::add_event()
 # perform iSCSI target port blocks on the two cluster resources specified
 # use the discovered ports on the two sanswitch resources specified when
 # performing the blocks.
 $EventManobj->add_event(-event_name => "iscsi_target_switch_port_blocks_Cm",
                         -nacl_clusters_cm => ["mycluster1", "mycluster2"],
                         -sanswitches => ["brocade300-rtp01", "brocade300-rtp02"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head1 B<Public Methods>

Note: All public methods exported by this package are only visible in
Event Modules that inherit this Event Module.

=head2 B<switch_port_block_abstracted(%args)>

Perform Switch port blocks generically for the specified port block type

=over 4

=item Event Parameters

=item *I<-port_block_types>

Required: An array reference of port block types to perform blocks against

=item Output

None

=item Examples

 # perform blocks against FC and iSCSI initiators
 $EventManobj->switch_port_block_abstracted(-port_block_types => ["iscsi_initiator_mac", "fc_initiator_wwpn"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<sequential_switch_port_blocks(%args)>

Perform Sequential switch port blocks on the specified resources.

=over 4

=item Inputs

=item *I<-resources_to_block>

Required: A hash reference of resource type keys, whose values are
array references containing the list of resources to block for the
respective resource type.

=item *I<-port_block_types>

Required: An array reference of port block types to perform blocks against

=item Output

None

=item Examples

 # perform sequential blocks against FC initiators and FC 7mode targets
 $EventManobj->sequential_switch_port_blocks(-port_block_types => ["fc_initiator_wwpn", "fc_target_wwpn"],
                                             -resources_to_block => { sanclients => ["ibm3550-rtp70", "ibm3550-rtp71"],
                                                                      sanfilers7m => ["fas270-rtp60", "fas270-rtp61"] });
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<parallel_switch_port_blocks(%args)>

Perform Parallel switch port blocks on the specified resources.

=over 4

=item Inputs

=item *I<-resources_to_block>

Required: A hash reference of resource type keys, whose values are
array references containing the list of resources to block for the
respective resource type.

=item *I<-port_block_types>

Required: An array reference of port block types to perform blocks against

=item Output

None

=item Examples

 # perform parallel blocks against FC & iSCSI initiators
 $EventManobj->parallel_switch_port_blocks(-port_block_types => ["fc_initiator_wwpn", "iscsi_initiator_mac"],
                                           -resources_to_block => { sanclients => ["ibm3550-rtp70", "ibm3550-rtp71"] });
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<switch_port_block_generic_disable_enable(%args)>

Perform a port block (disable/enable) against the specified port number and switch resource

=over 4

=item Inputs

=item *I<-port_number>

Required: The port identifier to perform the block against

=item *I<-switch_name>

Required: The name of the SanSwitch resource to use when performing the block

=item *I<-resource_name>

Required: The name of the resource that will be affected by the port block

=item *I<-resource_type>

Required: The type of the resource that will be affected by the port block

=item *I<-vf_context_id>

Optional: The virtual fabric context id to use on the switch when performing the port block

=item Output

None

=item Examples

 # perform a port block on switch cisco5020-rtp40 at port GigabitEthernet1/20
 # this describes that the affecting cmode target is the node 'fas3070-rtp40'
 $EventManobj->switch_port_block_generic_disable_enable(-switch_name => "cisco5020-rtp40",
                                                        -port_number => "GigabitEthernet1/20",
                                                        -resource_name => "fas3070-rtp40",
                                                        -resource_type => "nacl_nodes_cm");
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<disable_switch_port(%args)>

Disable the specified port on a switch

=over 4

=item Inputs

=item *I<-port_number>

Required: The port identifier to perform the disable

=item *I<-switch_name>

Required: The name of the SanSwitch resource to use when performing the disable

=item *I<-resource_name>

Required: The name of the resource that will be affected by the port disable

=item *I<-resource_type>

Required: The type of the resource that will be affected by the port disable

=item *I<-vf_context_id>

Optional: The virtual fabric context id to use on the switch when performing the port disable

=item Output

None

=item Examples

 # perform a port disable on switch cisco5020-rtp40 at port GigabitEthernet1/20
 # this describes that the affecting cmode target is the node 'fas3070-rtp40'
 $EventManobj->disable_switch_port(-switch_name => "cisco5020-rtp40",
                                   -port_number => "GigabitEthernet1/20",
                                   -resource_name => "fas3070-rtp40",
                                   -resource_type => "nacl_nodes_cm");
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<enable_switch_port(%args)>

Enable the specified port on a switch

=over 4

=item Inputs

=item *I<-port_number>

Required: The port identifier to enable

=item *I<-switch_name>

Required: The name of the SanSwitch resource to use to enable the port

=item *I<-resource_name>

Required: The name of the resource that will be affected by the port enable

=item *I<-resource_type>

Required: The type of the resource that will be affected by the port enable

=item *I<-vf_context_id>

Optional: The virtual fabric context id to use on the switch when performing the port enable

=item Output

None

=item Examples

 # perform a port enable on switch cisco5020-rtp40 at port GigabitEthernet1/20
 # this describes that the affecting cmode target is the node 'fas3070-rtp40'
 $EventManobj->enable_switch_port(-switch_name => "cisco5020-rtp40",
                                  -port_number => "GigabitEthernet1/20",
                                  -resource_name => "fas3070-rtp40",
                                  -resource_type => "nacl_nodes_cm");
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<build_switch_info_hash()>

Discover and build a hash containing all discovered MACs and WWPNs
for all resources in the current Event scope.

=over 4

=item Inputs

None

=item Output

A hash reference containing all the data discovered on the FC & Ethernet fabrics

=item Examples

 # discover all items
 my $switch_info_hash = $EventManobj->build_switch_info_hash();
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<reboot_switches()>

Reboot all switches specified

=over 4

=item Inputs

=item *I<-sanswitches>

Required: An array reference of switches to reboot.

=item *I<-async>

Optional: Whether to reboot asynchronously or not

Default: $EventUtil::FALSE (synchronous reboot / blocking)

=item Output

An array of subtest handles each representing the reboots

=item Examples

 # discover all items
 $EventManObj->reboot_switches(-sanswitches => ["switch1", "switch2"]);
 if ($EventManObj->error()) {
    # fail
 }

=back

=cut

1;
