# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An EventSet object container
## @author hrivera@netapp.com
## @status public
## @pod here

package EventMan::EventSet;

use Error;
use San::SanUtil;
use ParamsMgr;

use vars qw(@ISA);
@ISA = qw(Error);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: EventSet object Constructor
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # parse out the required event set name
    $self->{event_set_name} = &ParamsMgr::get_arg(-arg => "event_set_name",
                                                  -required => 1);
    delete($args{event_set_name});

    # parse out the optional parameters
    $self->{unique_id} = &ParamsMgr::get_arg(-arg => "unique_id",
                                             -default => 0);
    delete($args{unique_id});

    # run_style
    $self->{run_style} = &ParamsMgr::get_arg(-arg => "run_style",
                                             -default => "sequential");
    delete($args{run_style});

    $self->{cycle_count} = &ParamsMgr::get_arg(-arg => "cycle_count",
                                               -default => 1);
    delete($args{cycle_count});

    # initialize the event list to empty
    $self->{event_list} = [];

    # initialize the event set group list to empty
    $self->{groups} = [];

    # store the remaining args in the args datastructure
    $self->{args} = {%args};

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    add_event_to_set()
# Objective: Add the specified Event object to this EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_event_to_set {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    push(@{$self->{event_list}}, $event_obj);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    remove_event_from_set()
# Objective: Remove the specified Event object from this EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_event_from_set {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $event_obj = &ParamsMgr::get_arg(-arg => "event_obj",
                                        -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # filter the old event object list by removing the event object
    # that matches the unique id we wish to remove
    my @old_event_obj_list = $self->get_event_list();

    # grep out the event object that matches the unique id of the event
    # we want to delete
    my @new_event_obj_list = grep {
        if ($_->get_unique_id() == $event_obj->get_unique_id()) {
            0;
        } else {
            1;
        }
    } @old_event_obj_list;

    # if nothing was removed set an error
    if (scalar(@new_event_obj_list) == scalar(@old_event_obj_list)) {
        $self->_set_err("Could not remove event object from event set because it does not exist");
        return $EventUtil::FALSE;
    }

    # now store the new event list on the event set object
    $self->set_event_list(-event_list => \@new_event_obj_list);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    process_event_list_order()
# Objective: Process the event list order (randomize if necessary)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub process_event_list {
    my ($self) = @_;

    $self->_clear_err();

    # randomize the event list if the run style is random
    if ($self->get_run_style() =~ /random/i) {
        my (@event_list) = $self->get_event_list();

        # randomize the array
        $self->_randomize_array(\@event_list);

        # reset the event list
        $self->set_event_list(-event_list => \@event_list);

        # display the new order of this event set's event list
        my @summary;
        push(@summary, "EventSet '" . $self->get_name() . "': The event list has been randomized and will be:");
        foreach my $event ($self->get_event_list()) {
            push(@summary, "\t\t" . $event->get_name() . "(x" . $event->get_cycle_count() . ")");
        }

        &SanUtil::nate_log(-comment => join("\n", @summary));
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_event_list()
# Objective: Obtain the list of Event objects in this EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_event_list {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? @{$self->{event_list}} : $self->{event_list};
}

###############################################################################
# Method:    set_event_list()
# Objective: Modify the list of Event objects in this EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_event_list {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $event_list = &ParamsMgr::get_arg(-arg => "event_list",
                                         -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    $self->{event_list} = $event_list;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    join_group()
# Objective: Join this EventSet into the specified EventSetGroup name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub join_group {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $group_name = &ParamsMgr::get_arg(-arg => "group_name",
                                         -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    push(@{$self->{groups}}, $group_name);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    unjoin_group()
# Objective: Remove this EventSet from the specified EventSetGroup name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unjoin_group {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $group_name = &ParamsMgr::get_arg(-arg => "group_name",
                                         -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # filter out the group that we want to delete
    @{$self->{groups}} = grep {
        if ($_ eq $group_name) {
            0;
        } else {
            1;
        }
    } @{$self->{groups}};

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_group_list()
# Objective: Obtain the list of EventSetGroups that this EventSet is a member of
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_group_list {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? @{$self->{groups}} : $self->{groups};
}

###############################################################################
# Method:    get_name()
# Objective: Obtain the name of the EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{event_set_name};
}

###############################################################################
# Method:    get_run_style()
# Objective: Obtain the run style of the EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_run_style {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{run_style};
}

###############################################################################
# Method:    get_unique_id()
# Objective: Obtain the unique identifier for this EventSet object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unique_id {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{unique_id};
}

###############################################################################
# Method:    get_cycle_count()
# Objective: Obtain the cycle count for this EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cycle_count {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{cycle_count};
}

###############################################################################
# Method:    set_cycle_count()
# Objective: Modify the cycle count for this EventSet
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_cycle_count {
    my ($self, $count) = @_;

    $self->_clear_err();

    $self->{cycle_count} = $count;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    display()
# Objective: Display a summary of what this EventSet will do
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_summary {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # parse out the EventMan object
    my $event_man_obj = &ParamsMgr::get_arg(-arg => "event_man_obj",
                                            -required => 1);
    delete($args{-event_man_obj});

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # display a summary of what this eventset is going to do
    my @summary = ();

    push(@summary, "Event Set: " . $self->get_name());
    push(@summary, "\tAttributes:");
    push(@summary, "\t\t-cycle_count : " . $self->get_cycle_count());
    push(@summary, "\t\t-run_style : " . $self->get_run_style());
    push(@summary, "\t\t-unique_id : " . $self->get_unique_id());

    push(@summary, "");
    push(@summary, "\tEvent List:");
    foreach my $event ($self->get_event_list()) {
        push(@summary, "\t\t" . $event->get_name() . "(x" . $event->get_cycle_count() . ")");
    }
    push(@summary, "");

    # display the event parameters in this eventset's scope
    push(@summary, $event_man_obj->EventUtil::display_event_parameters(%args));

    my $str = join("\n", @summary);
    &SanUtil::nate_log('-comment', $str);

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

################################################################################
# Subroutine Name: _randomize_array
# Objective:       Randomizes an array in place (fisher yates shuffle)
################################################################################
sub _randomize_array {
    my ($self, $array) = @_;

    my ($i);

    for ($i = scalar(@{$array}) - 1; $i >= 0; $i--) {
        # generate a random index between [0, $i+1)
        my $j = int rand ($i + 1);

        # dont perform the swap if they are the same
        next if ($i == $j);

        # swap the element in $i with the element in $j
        # this will update the array reference in place.
        @$array[$i, $j] = @$array[$j, $i];
    }
}

1;

=pod

=head1 Package Name

EventSet.pm

=head1 Author

Robert Hunter

=head1 Description

Object interface for an EventSet, an ordered collection of Event objects.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for this EventSet to use.

=head1 B<Public Methods>

=head2 B<new(%args)>

Creates an EventSet object handle. Do not call this method directly, use
EventMan::create_event_set() to create an EventSet object

=over 4

=item Event Param Capable?

Yes (EventSet level)

=item Inputs

=item *I<-event_set_name>

Required: A valid uniqie EventSet name to associate this object with

=item *I<-unique_id>

Optional: a unique string to identify this EventSet object
Default: 0

=item *I<-run_style>

Optional: Specifies the runstyle of the eventset (random or sequential)
Default: sequential

=item *I<-cycle_count>

Optional: Denotes how many times to run the EventSet back to back.
Default: 1

=item Outputs

Returns an EventSet object

=item Examples

 # see EventMan::create_event_set()

=back

=head2 B<get_name()>

EventSet name accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the name of the EventSet (keyword)

=item Examples

 my $event_set_name = $event_set_obj->get_name();

=back

=head2 B<get_unique_id()>

Accessor method for unique_id

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the unique identifier for this EventSet object

=item Examples

 my $id = $event_set_obj->get_unique_id();

=back

=head2 B<get_run_style()>

Accessor method for run_style

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

The run style of this eventset object

=item Examples

 if ($event_set_obj->get_run_style() eq "random") {
    ...
 }

=back

=head2 B<get_cycle_count()>

Accessor method for cycle_count

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the cycle_count for this EventSet object

=item Examples

 my $cycle_count = $event_set_obj->get_cycle_count();

=back

=head2 B<set_cycle_count($cycle_count)>

Set method for cycle_count

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-cycle_count>

Required: The new cycle count value

=item Outputs

Set the cycle_count for this EventSet object

=item Examples

 $event_set_obj->set_cycle_count(10);

=back

=head2 B<add_event_to_set(%args)>

Adds the specified Event object to this EventSet's list

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_obj>

Required: The Event object to add to this EventSet

=item Outputs

Returns $EventUtil::TRUE, does not set error

=item Examples

 $event_set_obj->add_event_to_set(-event_obj => $event_obj);

=back

=head2 B<remove_event_from_set(%args)>

Removes the specified Event object from this EventSet's list

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_obj>

Required: The Event object to remove from this EventSet

=item Outputs

Returns $EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $event_set_obj->remove_event_from_set(-event_obj => $event_obj);
 if ($event_set_obj->error()) {
     ...
 }

=back

=head2 B<process_event_list()>

Process the EventSet's Event list (randomize the list if necessary)

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

$EventUtil::TRUE

=item Examples

 $event_set_obj->process_event_list();

=back

=head2 B<get_event_list()>

Obtain the Event list for this EventSet

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns an array or array reference of Event objects

=item Examples

 my (@event_objects) = $event_set_obj->get_event_list();

=back

=head2 B<set_event_list(%args)>

Sets the Event list for this EventSet

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_list>

Required: An array reference of one or more Event objects.

=item Outputs

Returns $EventUtil::TRUE

=item Examples

 $event_set_obj->set_event_list(-event_list => [$event_obj1, $event_obj2]);

=back

=head2 B<join_group(%args)>

Joins this EventSet into the specified EventSetGroup

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-group_name>

Required: A valid EventSetGroup name

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE and error if not

=item Examples

 $event_set_obj->join_group(-group_name => "my_fset_group");
 if ($event_set_obj->error()) {
     ...
 }

=back

=head2 B<unjoin_group(%args)>

Unjoins this EventSet from the specified EventSetGroup

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-group_name>

Required: A valid EventSetGroup name

=item Outputs

Returns $EventUtil::TRUE if successful, $EventUtil::FALSE and error if not

=item Examples

 $event_set_obj->unjoin_group(-group_name => "my_fset_group");
 if ($event_set_obj->error()) {
     ...
 }

=back

=head2 B<get_group_list()>

Accessor method for list of groups this EventSet is a member of

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns an array or array reference of group names

=item Examples

 my (@group_list) = $event_set_obj->get_group_list();

=back

=cut
