# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An EventSetGroup object container
## @author hrivera@netapp.com
## @status public
## @pod here

package EventMan::EventSetGroup;

use Error;
use ParamsMgr;

use vars qw(@ISA);
@ISA = qw(Error);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new()
# Objective: EventSetGroup object Constructor
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless($self, $class);

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # parse out the required event set group name
    $self->{group_name} = &ParamsMgr::get_arg(-arg => "group_name",
                                              -required => 1);
    delete($args{-group_name});

    # parse out the optional parameters
    $self->{run_style} = &ParamsMgr::get_arg(-arg => "run_style",
                                             -default => "sequential");
    delete($args{-run_style});

    $self->{dependence_type} = &ParamsMgr::get_arg(-arg => "dependence_type",
                                                   -default => "independent");
    delete($args{-dependence_type});

    $self->{enable_sync} = &ParamsMgr::get_arg(-arg => "enable_sync",
                                               -default => 0);
    delete($args{-enable_sync});

    # store the remaining args in the args datastructure
    $self->{args} = {&SanUtil::strip_dash(%args)};

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    return $self;
}

###############################################################################
# Method:    add_event_set_to_group()
# Objective: Add the specified EventSet object to this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_event_set_to_group {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $event_set_obj = &ParamsMgr::get_arg(-arg => "event_set_obj",
                                            -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # add this groupname the this eventset's membership attribute
    $event_set_obj->join_group(-group_name => $self->get_group_name());
    if ($event_set_obj->error()) {
        $self->_set_err($event_set_obj->errmsg());
        return $EventUtil::FALSE;
    }

    push(@{$self->{event_set_objs}}, $event_set_obj);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    remove_event_set_from_group()
# Objective: Remove the specified EventSet object from this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_event_set_from_group {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $event_set_obj = &ParamsMgr::get_arg(-arg => "event_set_obj",
                                            -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # filter the old eventset object list by removing the eventset object
    # that matches the unique id we wish to remove
    my @old_event_set_objs_list = $self->get_event_sets();

    # grep out the eventset object that matches the unique id of the eventset
    # we want to delete
    my @new_event_set_objs_list = grep {
        if ($_->get_unique_id() == $event_set_obj->get_unique_id()) {
            0;
        } else {
            1;
        }
    } @old_event_set_objs_list;

    # if nothing was removed set an error
    if (scalar(@new_event_set_objs_list) == scalar(@old_event_set_objs_list)) {
        $self->_set_err("Could not remove EventSet object from EventSetGroup because it does not exist");
        return $EventUtil::FALSE;
    }

    # remove this group from this eventset's group membership list
    $event_set_obj->unjoin_group(-group_name => $self->get_group_name());

    # now store the new eventset list on the group object
    $self->set_event_sets(-event_set_objs => \@new_event_set_objs_list);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_group_name()
# Objective: Obtain the EventSetGroup name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_group_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{group_name};
}

###############################################################################
# Method:    get_event_sets()
# Objective: Obtain the EventSets that are members of this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_event_sets {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? @{$self->{event_set_objs}} : $self->{event_set_objs};
}

###############################################################################
# Method:    set_event_sets()
# Objective: Modify the list of EventSets that are members of this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_event_sets {
    my ($self, %args) = @_;

    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $event_set_objs = &ParamsMgr::get_arg(-arg => "event_set_objs",
                                             -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    $self->{event_set_objs} = $event_set_objs;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_run_style()
# Objective: Obtain the run style of this EventSetGroup (parallel/sequential)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_run_style {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{run_style};
}

###############################################################################
# Method:    set_run_style()
# Objective: Modify the run style of the group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_run_style {
    my ($self, $style) = @_;

    $self->_clear_err();

    if ($style !~ /sequential|parallel/) {
        $self->_set_err("Invalid run style '$style', use 'sequential' or 'parallel'");
        return $EventUtil::FALSE;
    }

    $self->{run_style} = $style;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_dependence_type()
# Objective: Obtain the dependence type of the group (independent/dependent)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_dependence_type {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{dependence_type};
}

###############################################################################
# Method:    set_dependence_type()
# Objective: Modify the dependence type for this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_dependence_type {
    my ($self, $type) = @_;

    $self->_clear_err();

    if ($type !~ /independent|dependent/i) {
        $self->_set_err("Invalid dependence type '$type', use 'dependent' or 'independent'");
        return $EventUtil::FALSE;
    }

    $self->{dependence_type} = $type;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    is_sync_enabled()
# Objective: Determine if EventSet Synchronization is enabled for this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_sync_enabled {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{enable_sync};
}

###############################################################################
# Method:    enable_sync()
# Objective: Enable EventSet Synchronization for this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable_sync {
    my ($self) = @_;

    $self->_clear_err();

    $self->{enable_sync} = $EventUtil::TRUE;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disable_sync()
# Objective: Disable EventSet Synchronization for this group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable_sync {
    my ($self) = @_;

    $self->_clear_err();

    $self->{enable_sync} = $EventUtil::FALSE;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    destroy()
# Objective: Cleanly destroy this EventSetGroup by removing dependencies
#            from all EventSet objects.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;

    $self->_clear_err();

    # destroy this event set group

    # remove the group membership of this event set group
    # from all the event sets that are a part of it
    my @event_sets = $self->get_event_sets();

    foreach my $event_set_obj (@event_sets) {
        $self->remove_event_set_from_group(-event_set_obj => $event_set_obj);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

EventSetGroup.pm

=head1 Author

Robert Hunter

=head1 Description

Object interface for an EventSetGroup, a collection of EventSet objects.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for this EventSetGroup to use.

=head1 B<Public Methods>

=head2 B<new(%args)>

Creates an EventSetGroup object handle. Do not call this method directly, use
EventMan::create_event_set_group() to create an EventSetGroup object

=over 4

=item Event Param Capable?

Yes (EventSetGroup level)

=item Inputs

=item *I<-group_name>

Required: A valid uniqie EventSetGroup name to associate this object with

=item *I<-run_style>

Optional: a unique string to identify this EventSetGroup object
Default: "sequential"

=item *I<-dependence_type>

Optional: Defines the dependence relationship between all EventSets in the EventSetGroup
Default: "independent"

=item *I<-enable_sync>

Optional: If set to '1', will enable EventSet synchronization.
Default: 0 (off)

=item Outputs

Returns an EventSetGroup object

=item Examples

 # see EventMan::create_event_set_group()

=back

=head2 B<get_group_name()>

EventSetGroup name accessor method

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Returns the name of the EventSetGroup

=item Examples

 my $group_name = $event_set_group_obj->get_group_name();

=back

=head2 B<get_event_sets()>

Accessor method for EventSet objects in this group

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

An array or array reference of EventSet objects in this group

=item Examples

 my @eventset_objs = $event_set_group_obj->get_event_sets();

=back

=head2 B<set_event_sets(%args)>

Setter method for EventSet member list

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_set_objs>

Array reference of EventSet objects

=item Outputs

$EventUtil::TRUE if successful

=item Examples

$event_set_group_obj->set_event_sets(-event_set_objs => [$fset1, $fset2]);

=back

=head2 B<get_run_style()>

Accessor method for run style

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the run style (parallel/sequential)

=item Examples

 my $run_style = $event_set_group_obj->get_run_style();

=back

=head2 B<set_run_style(run_style)>

Set method for the run style attribute

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-run_style>

Required: The new run_style

=item Outputs

Sets the run_style attribute to the specified value

=item Examples

 $event_set_group_obj->set_run_style("parallel");

=back

=head2 B<get_dependence_type()>

Accessor method for the dependence type

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

Get the dependence_type (independent/dependent)

=item Examples

 my $dependence_type = $event_set_group_obj->get_dependence_type();

=back

=head2 B<set_dependence_type(dependence_type)>

Set method for the run style attribute

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-dependence_type>

Required: The new dependence_type

=item Outputs

Sets the dependence_type attribute to the specified value, sets error if failure

=item Examples

 $event_set_group_obj->set_dependence_type("dependent");

=back

=head2 B<is_sync_enabled()>

Get method for enable_sync attribute

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

$EventUtil::TRUE if enabled, $EventUtil::FALSE if not enabled

=item Examples

 if ($event_set_group_obj->is_sync_enabled() == $EventUtil:FALSE) {
    $event_set_group_obj->enable_sync();
 }

=back

=head2 B<enable_sync()>

Enables synchronization for this EventSetGroup

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

$EventUtil::TRUE

=item Examples

 if ($event_set_group_obj->is_sync_enabled() == $EventUtil:FALSE) {
    $event_set_group_obj->enable_sync();
 }

=back

=head2 B<disable_sync()>

Disables synchronization for this EventSetGroup

=over 4

=item Event Param Capable?

No

=item Inputs

None

=item Outputs

$EventUtil::TRUE

=item Examples

 if ($event_set_group_obj->is_sync_enabled() == $EventUtil:TRUE) {
    $event_set_group_obj->disable_sync();
 }

=back

=head2 B<add_event_set_to_group(%args)>

Adds the specified EventSet object to this EventSetGroup

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_set_obj>

Required: The EventSet object to add

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and error

=item Examples

 $event_set_group_obj->add_event_set_to_group(-event_set_obj => $fset_obj);
 if ($event_set_group_obj->error()) {
    ...
 }

=back

=head2 B<remove_event_set_from_group(%args)>

Removes the specified EventSet object from this EventSetGroup

=over 4

=item Event Param Capable?

No

=item Inputs

=item *I<-event_set_obj>

Required: The EventSet object to remove

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and error

=item Examples

 $event_set_group_obj->remove_event_set_from_group(-event_set_obj => $fset_obj);
 if ($event_set_group_obj->error()) {
    ...
 }

=back

=cut
