# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing Clam Validation for Cmode Controllers.
## @author mhester@netapp.com
## @status private
## @pod here

package ClamValidationUtil;

# Perl Modules
use Data::Dumper;

# NATE Modules
use San::FIT::ClamValidation;
use ParamsMgr;
use Tharn;

###############################################################################
#                               Public Methods
###############################################################################

################################################################################
# Method:     collect_clam_master()
# Objective:       Collect CLAM Master on c-mode controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_clam_master {
    my ($self, %args) = @_;

    # skip clam master collection if not requested
    my $validate_clam = $self->EventUtil::get_arg(-arg => "clam_validation");
    unless ($validate_clam) {
        &SanUtil::nate_log('-comment', "Skipping clam master collection");
        return "";
    }

    &ParamsMgr::process_args(%args);

    my $sanfilerCm = &ParamsMgr::get_arg(-arg      => "sanfilerCm",
                                         -required => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return "";
    }

    my $cv_abort_on_error = $self->EventUtil::get_arg(-arg => "clam_validation_abort_on_error");

    my @active_sanfilerCms = $self->ClamValidationUtil::_get_active_sanfilersCm(-sanfilerCm => $sanfilerCm);
    if ($self->error()) {
        return "";
    }

    my $clam_master = &ClamValidation::get_clam_master(-nodes => \@active_sanfilerCms);
    unless ($clam_master) {
        if ($cv_abort_on_error) {
            $self->_set_err("No sanfilersCM were returned as clam master.");
        } else {
            &SanUtil::nate_log('-warn',
                               "No sanfilersCM were returned as clam master." .
                               "The test will continue as the event flag " .
                               "\'clam_validation_abort_on_error\' is ".
                               " disabled, howevever, all remaining clam " .
                               " validation for this event " .
                               " will be void.");
        }
        return "";
    }

    return $clam_master;
} # End sub collect_clam_master

################################################################################
# Method:     validate_clam_after_giveback()
# Objective:       Collect CLAM Master on c-mode controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub validate_clam_after_giveback {
    my ($self, %args) = @_;

    # skip clam validation after giveback if not requested
    my $clam_validation = $self->EventUtil::get_arg(-arg => "clam_validation");
    my $clam_validation_on_giveback = $self->EventUtil::get_arg(-arg => "clam_validation_on_giveback");
    unless ($clam_validation and $clam_validation_on_giveback ) {
        &SanUtil::nate_log('-comment', "Skipping validation of clam on giveback");
        return {};
    }

    &ParamsMgr::process_args(%args);

    my $taken_over_node = &ParamsMgr::get_arg(-arg      => "taken_over_node",
                                              -required => $EventUtil::TRUE);
    my $original_master = &ParamsMgr::get_arg(-arg      => "original_master",
                                              -required => $EventUtil::TRUE);
    my $give_start_time = &ParamsMgr::get_arg(-arg      => "sfo_giveback_start_time",
                                              -required => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my $cv_abort_on_error = $self->EventUtil::get_arg(-arg => "clam_validation_abort_on_error");
    unless ($original_master) {
        if ($cv_abort_on_error) {
            $self->_set_err("Original clam master is required to validate " .
                            "clam on giveback.");
        } else {
            &SanUtil::nate_log('-warn',
                               "Original clam master is required to validate " .
                               "clam on giveback, however, the event flag " .
                               "\'clam_validation_abort_on_error\' is ".
                               " disabled so all remaining clam " .
                               " validation for this event will be void.");
        }
        return {};
    }

    # get the next padded runid from the event
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $clam_validation_runid_padded = $event_obj->increment_subtest_runid_counter();

    my $subtest = subtest(
        sub {
            $self->EventUtil::reconnect_all_sanfilers(-disconnect_existing_conn => $EventUtil::FALSE,
                                                      -faulted                  => $EventUtil::FALSE,
                                                      -type                     => "cmode");
            if ($self->error()) {
                # perform a graceful failure
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }

            my $cv_abort_on_error = $self->EventUtil::get_arg(-arg => "clam_validation_abort_on_error");
            my @active_sanfilersCm = $self->ClamValidationUtil::_get_active_sanfilersCm(-sanfilerCm => $original_master);
            if ($self->error()) {
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }
            # We need to add in the node that will be given
            #  back as it is still faulted when this
            #  function is called.
            push(@active_sanfilersCm, $taken_over_node);
            unless (&ClamValidation::validate_clam_after_giveback(-nodes           => \@active_sanfilersCm,
                                                                  -original_master => $original_master,
                                                                  -give_start_time => $give_start_time)) {
                $self->_set_err("Failed to validate all clam instances after giveback "
                                . "of $fault_node.");
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }
            if ($self->error()) {
                # perform a graceful failure
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }
            $self->SubtestComm::subtest_complete();
            return $EventUtil::TRUE;
        },
        -runid => $clam_validation_runid_padded . "_validate_clam_instances_after_giveback_of_$taken_over_node",
        -bg,
        "-onexit" => sub {my ($self) = @_; $self->exithandler(1);}
    );

    return $subtest;
} # End sub validate_clam_after_giveback

################################################################################
# Method:     validate_clam_after_takeover()
# Objective:       Collect CLAM Master on c-mode controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub validate_clam_after_takeover {
    my ($self, %args) = @_;

    # skip clam validation after takeover if not requested
    my $clam_validation = $self->EventUtil::get_arg(-arg => "clam_validation");
    my $clam_validation_on_takeover = $self->EventUtil::get_arg(-arg => "clam_validation_on_takeover");
    unless ($clam_validation and $clam_validation_on_takeover) {
        &SanUtil::nate_log('-comment', "Skipping validation of clam instances on takeover event.");
        return {};
    }

    &ParamsMgr::process_args(%args);

    my $original_master = &ParamsMgr::get_arg(-arg      => "original_master",
                                              -required => $EventUtil::TRUE);
    my $taken_over_node = &ParamsMgr::get_arg(-arg      => "taken_over_node",
                                              -required => $EventUtil::TRUE);
    my $take_start_time = &ParamsMgr::get_arg(-arg      => "sfo_takeover_start_time",
                                              -required => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # get the next padded runid from the event
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $clam_validation_runid_padded = $event_obj->increment_subtest_runid_counter();

    my $subtest = subtest(
        sub {
            $self->EventUtil::reconnect_all_sanfilers(-disconnect_existing_conn => $EventUtil::FALSE,
                                                      -faulted                  => $EventUtil::FALSE,
                                                      -type                     => "cmode");
            if ($self->error()) {
                # perform a graceful failure
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }

            my $cv_abort_on_error = $self->EventUtil::get_arg(-arg => "clam_validation_abort_on_error");
            my @active_sanfilersCm = $self->ClamValidationUtil::_get_active_sanfilersCm(-sanfilerCm => $original_master);
            if ($self->error()) {
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }
            unless (&ClamValidation::validate_clam_after_takeover(-nodes           => \@active_sanfilersCm,
                                                                  -fault_node      => $taken_over_node,
                                                                  -original_master => $original_master,
                                                                  -take_start_time => $take_start_time)) {
                $self->_set_err("Failed to validate all clam instances after takeover "
                                . "of $fault_node.");
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }

            if ($self->error()) {
                # perform a graceful failure
                $self->SubtestComm::subtest_fail(-graceful => not $cv_abort_on_error);
                return $EventUtil::FALSE;
            }
            $self->SubtestComm::subtest_complete();
            return $EventUtil::TRUE;
        },
        -runid => $clam_validation_runid_padded . "_validate_clam_instances_after_takeover_of_$taken_over_node",
        -bg,
        "-onexit" => sub {my ($self) = @_; $self->exithandler(1);}
    );

    return $subtest;
} # End sub validate_clam_after_takeover

################################################################################
# Method:     reap_clamvalidation_subtest()
# Objective:       Wait for all clam validation subtests to complete.
# Details:         See POD documentation at the end of this file
################################################################################
sub reap_clamvalidation_subtest {
    my ($self,%args) = @_;

    # skip clam validation subtest reap routines if not requested
    my $validate_clam = $self->EventUtil::get_arg(-arg => "clam_validation");
    if (not $validate_clam) {
        return $EventUtil::TRUE;
    }

    &ParamsMgr::process_args(%args);

    my @subtests = &ParamsMgr::get_arg(-arg      => "subtests",
                                       -required => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # filter out any invalid subtests from the passed in list
    my @final_subtest_list;
    foreach my $sub (@subtests) {
        if (ref($sub) =~ /subtest/i) {
            push(@final_subtest_list, $sub);
        }
    }

    # reassign the subtests array with the valid ones
    @subtests = @final_subtest_list;

    # verify that the subtests passed in contain values
    # if there are no subtests, just return.
    if (not scalar(@subtests)) {
        return $EventUtil::TRUE;
    }

    $self->wait_for_subtests(-subtests => \@subtests,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0);
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    return $EventUtil::TRUE;
} # End sub reap_clamvalidation_subtest

###############################################################################
#                               Private Methods
###############################################################################
sub _get_active_sanfilersCm {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $sanfilerCm = &ParamsMgr::get_arg(-arg      => "sanfilerCm",
                                         -required => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return wantarray() ? () : [];
    }

    my $sanfilerCm_name = $sanfilerCm->get_name();
    if ($sanfilerCm->error()) {
        $self->_set_err($sanfilerCm->errmsg());
        return wantarray() ? () : [];
    }

    my @faulted_sf_objs = $self->EventUtil::get_sanfiler_objs(-faulted => $EventUtil::TRUE,
                                                              -type    => 'cmode');
    if ($self->error()) {
        # Error set, just return.
        return wantarray() ? () : [];
    }

    # In case there are multiple ControllerClusters, find the specific one with
    # this node member.
    my $cluster_name = $self->ResMgr::get_resource_attribute(-name      => $sanfilerCm_name,
                                                             -type      => "sanfilersCm",
                                                             -attribute => "controllercluster_name");
    if ($self->error()) {
        return wantarray() ? () : [];
    }
    my $clus_obj = $self->EventUtil::get_controllercluster_obj(-cluster_name => $cluster_name);
    if ($self->error()) {
        return wantarray() ? () : [];
    }
    my @sanfiler_objs = $clus_obj->get_node_objs();
    if ($clus_obj->error()) {
            $self->_set_err($clus_obj->errmsg());
            return wantarray() ? () : [];
    }

    # Remove the faulted sanfilers from the list of all sanfilers in cluster.
    my @unfaulted_sf_objs = @sanfiler_objs;
    foreach my $sf_obj (@faulted_sf_objs) {
        @unfaulted_sf_objs = &ClamValidation::_remove_node_from_list(-remove_node => $sf_obj,
                                                                     -nodes       => \@unfaulted_sf_objs);
    }

    return wantarray() ? @unfaulted_sf_objs : \@unfaulted_sf_objs;
} # End sub _get_active_sanfilersCm

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    $self->EventUtil::register_utility_module_defaults(
        -utility_module => __PACKAGE__,
        -arg_defaults   => {
            clam_validation                  => 0,
            clam_validation_on_takeover      => 1,
            clam_validation_on_giveback      => 1,
            clam_validation_abort_on_error   => 0
        }
    );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # Currently not used.

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    # Currently not used.

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "clam_validation",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::TRUE,
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-clam_validation => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Enable or disable clam validation");

    $self->EventUtil::document_event_param(-param => "clam_validation_on_takeover",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::TRUE,
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-validate_clam_on_takeover => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Enable or disable clam validation on takeover events");

    $self->EventUtil::document_event_param(-param => "clam_validation_on_giveback",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::TRUE,
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-validate_clam_on_giveback => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Enable or disable clam validation on giveback events");

    $self->EventUtil::document_event_param(-param => "clam_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::FALSE,
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-clam_validation_abort_on_error => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Enable or disable clam validation abort on error");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

ClamValidationUtil.pm

=head1 Author

Michael Hester

=head1 Description

Issue Clam Validation Before/After Cmode Events.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-clam_validation>

=over 4

Description: When enabled, will validate clam instances within the cluster.

Per Resource Capable: No

Default: 0 (off)

=back

=head2 I<-clam_validation_on_giveback>

=over 4

Description: When enabled, will validate clam instances within the cluster
             after giveback.

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-clam_validation_on_takeover>

=over 4

Description: When enabled, will validate clam instances within the cluster
             after takeover.

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-clam_validation_abort_on_error>

=over 4

Description: When enabled, will cause events to exit Fatal on validation error.

Per Resource Capable: No

Default: 1 (on)

=back

=head1 B<Public Methods>

=head2 B<collect_clam_master()>

Get the current cluster clam master.

=over 4

=item Inputs

=item *I<-sanfilerCm>

Required: C-mode resource that is a member of the cluster to collect clam master

Default: None

=item Outputs

sanfilerCm object that is the current clam master.

=item Event Parameter Dependencies

 -clam_validation

=item Examples

    # Collect current clam master before executing Event.
    my $clam_master = $self->ClamValidationUtil::collect_clam_master(
        -sanfilerCm => $sanfiler_obj
    );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

=back

=head2 B<validate_clam_after_giveback()>

Validate all clam instances are in good health after a giveback is issued.

=over 4

=item Inputs

=item *I<-taken_over_node>

Required: C-mode resource that is the node that was taken over
          and is being given back.

Default: None

=item *I<-original_master>

Required: C-mode resource that is the clam master of the cluster
          before issuing giveback command.

Default: None

=item *I<-sfo_giveback_start_time>

Optional: the exact time the sfo giveback command completed at the cli.

Default:  Perl Built-In 'time' this subroutined is called.

=item Outputs

$subtest handle and sets error if they occur

=item Event Parameter Dependencies

    -clam_validation
    -clam_validation_on_giveback

=item Examples

    # Store the exact time giveback call exited.
    my $sfo_giveback_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_giveback(-original_master         => $clam_master,
                                                                             -sfo_giveback_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

=back

=head2 B<validate_clam_after_takeover()>

When enabled, validate the state of clam on all unfaulted sanfilers (Cmode) in
the current Event scope.

=over 4

=item Inputs

=item *I<-original_master>

Required: C-mode resource that is the clam master of the cluster
          before issuing takeover command

Default: None

=item *I<-fault_node>

Required: C-mode resource that is the taken over node in the cluster
          during the takeover

=item *I<-sfo_takeover_start_time>

Optional: the exact time the sfo takeover command completed at the cli.

Default:  Perl Built-In 'time' this subroutined is called.

=item Outputs

$subtest handle and sets error if they occur

=item Event Parameter Dependencies

    -clam_validation
    -clam_validation_on_takeover

=item Examples

    # Store the exact time takeover call exited.
    my $sfo_takeover_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_takeover(-original_master         => $clam_master,
                                                                             -fault_node              => $taken_over_node,
                                                                             -sfo_takeover_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

=back

=head2 B<reap_clamvalidation_subtest()>

Waits for all clam validation subtests to finish before exiting the parent
subroutine.

=over 4

=item Inputs

=item *I<-subtests>

Required: Reference to an array reference of subtest references.
(Can be a single element array ref)

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Examples

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

=back

=cut
