# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing LIF Validation
## @author derricks@netapp.com
## @status private
## @pod here

package ControllerValidation;

use strict;
use warnings;

# NATE packages
use Tharn;

# San packages
use San::SanClient::ClientUtil;
use San::SanUtil;
use ParamsMgr;

# NACL packages
use NACL::CS::VserverFcpInitiator;
use NACL::CS::VserverFcpInterface;
use NACL::CS::NetworkInterface;
use NACL::CS::NetworkPort;
use NACL::CS::SystemNode;
use NACL::CS::NetworkFcpAdapter;
use NACL::APISet::Node::CLI::CMode;
use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound;

use NATE::Log qw(log_global);

$Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    validate_controller_before_takeover()
# Objective: Issue controller validation before a takeover is issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_controller_before_takeover {
    my ($self, %args) = @_;

    # determine if controller validation on takeover is enabled in the current event/eventset
    my $controller_validation_before_takeover = $self->EventUtil::get_arg(-arg => "controller_validation_before_takeover");
    my $controller_validation                 = $self->EventUtil::get_arg(-arg => "controller_validation");

    if (!$controller_validation_before_takeover or !$controller_validation) {
        return $EventUtil::TRUE;
    }

    $self->ControllerValidation::_validate_before_failover(%args);
    if ($self->error()) {
        $self->_set_err("Controller validation before takeover failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_controller_after_takeover()
# Objective: Issue controller validation assuming a takeover was issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_controller_after_takeover {
    my ($self, %args) = @_;

    # determine if controller validation on takeover is enabled in the current event/eventset
    my $controller_validation_on_takeover = $self->EventUtil::get_arg(-arg => "controller_validation_on_takeover");
    my $controller_validation             = $self->EventUtil::get_arg(-arg => "controller_validation");

    if (!$controller_validation_on_takeover or !$controller_validation) {
        return $EventUtil::TRUE;
    }

    $self->ControllerValidation::_validate_on_failover(%args);
    if ($self->error()) {
        $self->_set_err("Controller validation after takeover failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_controller_after_giveback()
# Objective: Issue controller validation assuming a giveback was issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_controller_after_giveback {
    my ($self, %args) = @_;

    # determine if lif validation on giveback is enabled in the current event/eventset
    my $controller_validation_on_giveback = $self->EventUtil::get_arg(-arg => "controller_validation_on_giveback");
    my $controller_validation             = $self->EventUtil::get_arg(-arg => "controller_validation");

    if (!$controller_validation_on_giveback or !$controller_validation) {
        return $EventUtil::TRUE;
    }

    $self->ControllerValidation::_validate_on_failover(%args);
    if ($self->error()) {
        $self->_set_err("Controller validation after giveback failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    collect_fc_gs_information()
# Objective: Run commands to collect FC_GS data from the controllers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub collect_fc_gs_information {
    my ($self) = @_;

    my $fc_gs_collection = $self->EventUtil::get_arg(-arg => "fc_gs_collection");
    if (not $fc_gs_collection) {
        return $EventUtil::TRUE;
    }

    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $runid_padded = $event_obj->increment_subtest_runid_counter();

    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                      -type => "cmode");
    if ($self->error()) {
        $self->_set_err("Failed to obtain C-mode nacl nodes");
        return $EventUtil::FALSE;
    }

    my @subtests = ();

    foreach my $node_name (@nacl_nodes) {
        my $runid = $runid_padded . "_fc_gs_collection_on_${node_name}";
        my $subtest = subtest(sub {
                                  # Get the node object
                                  my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
                                  if ($self->error()) {
                                      $self->_set_err("Unable to retrieve the object for node '$node_name'");
                                      $self->SubtestComm::subtest_fail();
                                      return $EventUtil::FALSE;
                                  }

                                  # FC_GS commands
                                  my @cmds = ("run -node $node_name -command \"sysconfig -v\"",
                                              "run -node $node_name -command \"fcp topology show\"",
                                              "run -node $node_name -command \"fcp zone show\"",
                                              "run -node $node_name -command \"fcp nameserver show\"");

                                  foreach my $cmd (@cmds) {
                                      try {
                                          my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                                              interface => "CLI",
                                                                              set => "CMode");
                                          $apiset->execute_raw_command(command => $cmd);
                                      } otherwise {
                                          # warn if any of the commands failed for this host
                                          my $e = shift;
                                          $Log->warn("Exception " . ref($e) . " caught: " . $e->text());
                                      };
                                  }
                                  $self->SubtestComm::subtest_complete();
                                  return $EventUtil::TRUE;

                              }, -runid => $runid, -bg,
                              "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);
    }

    # reap all subtests
    if (scalar(@subtests)) {
        $self->wait_for_subtests(-subtests => \@subtests,
                                 -dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 1,
                                 -enable_sync => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -async => 0);
        if ($self->error()) {
            $self->_set_err("Encountered a failure during FC_GS collection");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _get_lif_info
# Objective: Build a hash of lif information for each node passed in.
#            Format of lif hash:
#            $lif_hash =>
#                    <node_name> =>
#                       <vserver_name> =>
#                           <lif_name> =>
#                                current_node  => Node where the lif currently resides
#                                current_port  => Port where the lif currently resides
#                                home_node     => Node where the lif should reside
#                                home_port     => Name of home port
#                                data_protocol => Data protocol of lif
#                                status_oper   => Operational status of lif
#                                status_admin  => Admin status of lif
#                                is_home       => Is the lif home
###############################################################################
sub _get_lif_info {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $data_protocols = &ParamsMgr::get_arg(-arg      => "protocols",
                                             -required => 0,
                                             -default  => [],
                                             -ref_type => "ARRAY");

    my $nacl_nodes_cm  = &ParamsMgr::get_arg(-arg      => "nacl_nodes_cm",
                                             -required => 0,
                                             -default  => [],
                                             -ref_type => "ARRAY");

    my $reconnect      = &ParamsMgr::get_arg(-arg      => "reconnect",
                                             -required => 0,
                                             -default  => 0);

    if (&ParamsMgr::error()) {
        $self->set_err(&ParamsMgr::get_error());
        return {};
    }

    my $lv_lif_name_regex_filter = $self->EventUtil::get_arg(-arg => "lif_validation_lif_name_regex_filter");

    # Create the lif hash.  The structure of the hash will look like the following:
    my $lif_hash = {};
    foreach my $node_name (sort @$nacl_nodes_cm) {
        my $reconnected = 0;

        # See if we were requested to reconnect
        if ($reconnect == 1) {
            $reconnected = 1;
        }

        # Get the node object
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the object for node '$node_name'");
            return {};
        }

        # Load all the lifs that are homed to the current node
        my @lif_objs;
        try {
            @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $nacl_node_obj,
                                                          filter => {'curr-node' => $node_name,
                                                                     'role' => "data"});
        } catch NACL::Exceptions::NoElementsFound with {
            # no lifs found on this node
            @lif_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return {};
        }

        # go to next node if no data lifs were found on this node
        if (not scalar(@lif_objs)) {
            next;
        }

        # Walk through the lif objects and add them to the lif_hash
        foreach my $lif_obj (@lif_objs) {
            # Get the lif name, protocol list and vserver name
            my ($lif_name, @curr_data_protocols, $vserver_name);
            try {
                $lif_name = $lif_obj->lif();
                @curr_data_protocols = $lif_obj->data_protocol();
                $vserver_name = $lif_obj->vserver();
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }

            unless (defined($lif_name) and $lif_name ne "") {
                $self->_set_err("Could not obtain lif name for lif on node '$node_name'");
                return {};
            }

            # Check to see if the lif passes the regex filter
            unless (defined($lv_lif_name_regex_filter) and $lif_name =~ $lv_lif_name_regex_filter) {
                # Lif doesn't match the regex filter so skip it.
                next;
            }
            if (not scalar(@curr_data_protocols)) {
                $self->_set_err("Could not obtain data protocol for lif '$lif_name' on node '$node_name'");
                return {};
            }
            unless (defined($vserver_name) and $vserver_name ne "") {
                $self->_set_err("Could not obtain vserver name for lif '$lif_name' on node '$node_name'");
                return {};
            }

            # Check to see if we need to filter by data protocol
            if (scalar(@$data_protocols) > 0) {
                # We need to filter, so make sure at least one of the data protocols for this lif
                # is found in the data_protocols array reference that was passed in.
                my $matched = 0;
                foreach my $protocol (@curr_data_protocols) {
                    if (grep(/$protocol/, @$data_protocols)) {
                        # The protocol was found in @$data_protocols
                        $matched = 1;
                        last;
                    }
                }
                # if we didnt match a protocol, skip this lif
                if ($matched == 0) {
                    my $protocol_list = join(" ", @$data_protocols);
                    $Log->comment("Skipping lif '$lif_name' since it isn't one of '$protocol_list'");
                    next;
                }
            }

            try {
                # Get the lif attributes and store them in the lif hash
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{current_node} = $lif_obj->curr_node();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{current_port} = $lif_obj->curr_port();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{home_port} = $lif_obj->home_port();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{home_node} = $lif_obj->home_node();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{status_oper} = $lif_obj->status_oper();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{status_admin} = $lif_obj->status_admin();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{is_home} = $lif_obj->is_home();
                $lif_hash->{$node_name}->{$vserver_name}->{$lif_name}->{data_protocol} = join(",", @curr_data_protocols);
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }
        }
    }

    return $lif_hash;
}

###############################################################################
# Method:    _get_network_port_info
# Objective: Build a hash of network port information for each node passed in.
#            $network_port_hash =>
#                    <node_name> =>
#                           <port> =>
#                                link          => Link status
#                                role/ipspace  => Role/IPSpace of the port
#                                mtu           => MTU of the port
#                                speed_admin   => Speed Admin
#                                speed_oper    => Speed Oper
#                                duplex_admin  => Duplex Admin
#                                duplex_oper   => Duplex Oper
###############################################################################
sub _get_network_port_info {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $nacl_nodes_cm        = &ParamsMgr::get_arg(-arg      => "nacl_nodes_cm",
                                                   -required => 0,
                                                   -default  => [],
                                                   -ref_type => "ARRAY");

    my $reconnect          = &ParamsMgr::get_arg(-arg      => "reconnect",
                                                 -required => 0,
                                                 -default  => 0);

    if (&ParamsMgr::error()) {
        $self->set_err(&ParamsMgr::get_error());
        return {};
    }

    # Get the regex filter
    my $npv_port_name_regex_filter = $self->EventUtil::get_arg(-arg => "net_port_validation_port_name_regex_filter");

    # Create the network port hash.
    my $network_port_hash = {};
    foreach my $node_name (sort @$nacl_nodes_cm) {
        my $reconnected = 0;

        # See if we were requested to reconnect
        if ($reconnect == 1) {
            $reconnected = 1;
        }

        # Get the node object
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the object for node '$node_name'");
            return {};
        }

        # Load all the network ports that on the current node
        my @port_objs;
        try {
            @port_objs = NACL::CS::NetworkPort->fetch(command_interface => $nacl_node_obj,
                                                      filter => {'node' => $node_name});
        } catch NACL::Exceptions::NoElementsFound with {
            # no ports found on this node
            @port_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return {};
        }

        # go to next node if no ports were found on this node
        if (not scalar(@port_objs)) {
            next;
        }

        foreach my $port_obj (@port_objs) {
            # Get the node for the port
            my ($node, $port);
            try {
                $node = $port_obj->node();
                $port = $port_obj->port();
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }

            unless (defined($node) and $node ne "") {
                $self->_set_err("Invalid data returned for 'node'");
                return {};
            }

            # Get the port
            unless (defined($port) and $port ne "") {
                $self->_set_err("Invalid data returned for 'port' for node '$node'");
                return {};
            }

            # Check to see if the port passes the regex filter
            unless ($port =~ $npv_port_name_regex_filter) {
                # Port doesn't match the regex filter so skip it.
                next;
            }

            # Process the rest of the fields and store the data in the $network_port_obj hash
            try {
                $network_port_hash->{$node}->{$port}->{speed_admin}  = $port_obj->speed_admin();
                $network_port_hash->{$node}->{$port}->{speed_oper}   = $port_obj->speed_oper();
                $network_port_hash->{$node}->{$port}->{duplex_admin} = $port_obj->duplex_admin();
                $network_port_hash->{$node}->{$port}->{duplex_oper}  = $port_obj->duplex_oper();
                $network_port_hash->{$node}->{$port}->{mtu}          = $port_obj->mtu();
                $network_port_hash->{$node}->{$port}->{link}         = $port_obj->link();

                #burt667979 ipspaces change
                if ($nacl_node_obj->version_manager->has_uichange(uichange => 'ipspaces-ms4',
                                                                  throw_exception => 0)) {
                    $network_port_hash->{$node}->{$port}->{ipspace}  = $port_obj->ipspace();
                } else {
                    $network_port_hash->{$node}->{$port}->{role}  = $port_obj->role();
                }
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }
            my ($key, $value);
            while (($key, $value) = each %{$network_port_hash->{$node}->{$port}}) {
                unless (defined($value and $value ne "")) {
                    $self->_set_err("Invalid data returned for '$key' on node '$node' port '$port'");
                    return {};
                }
            }
        }
    }

    return $network_port_hash;
}

###############################################################################
# Method:    _get_fcp_port_info
# Objective: Build a hash of fcp adapter information for each node passed in.
#            $fcp_port_hash =>
#                    <node_name> =>
#                           <adapter> =>
#                                fabric_established  => Fabric Established
#                                status              => Status Field
#                                state               => Link State
#                                conn_established    => Connection Established
###############################################################################
sub _get_fcp_port_info {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $nacl_nodes_cm = &ParamsMgr::get_arg(-arg      => "nacl_nodes_cm",
                                            -required => 0,
                                            -default  => [],
                                            -ref_type => "ARRAY");

    my $reconnect     = &ParamsMgr::get_arg(-arg      => "reconnect",
                                            -required => 0,
                                            -default  => 0);

    if (&ParamsMgr::error()) {
        $self->set_err(&ParamsMgr::get_error());
        return {};
    }

    # Get the regex filter
    my $fpv_port_name_regex_filter = $self->EventUtil::get_arg(-arg => "fcp_port_validation_port_name_regex_filter");

    # Create the fcp adapter hash.
    my $fcp_port_hash = {};
    # 30 Minute timeout to handle parallel faults (Burt#746465)
    my $timeout = 1800;
    foreach my $node_name (sort @$nacl_nodes_cm) {
        my $reconnected = 0;

        # See if we were requested to reconnect
        if ($reconnect == 1) {
            $reconnected = 1;
        }

        # Get the node object
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the object for node '$node_name'");
            return {};
        }

        # Load all the fcp adapter objs that are on this current node
        my @fcp_adapter_objs;
        try {
            @fcp_adapter_objs = NACL::CS::NetworkFcpAdapter->fetch(command_interface => $nacl_node_obj,
                                                                   filter => {'node' => $node_name},
                                                                   'method-timeout' => $timeout);
        } catch NACL::Exceptions::NoElementsFound with {
            # no adapters found on this node
            @fcp_adapter_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return {};
        }

        # go to next node if no fcp adapters were found on this node
        if (not scalar(@fcp_adapter_objs)) {
            next;
        }

        foreach my $fcp_adapter_obj (@fcp_adapter_objs) {
            my ($node, $adapter);
            # Get the node for the adapter
            try {
                $node = $fcp_adapter_obj->node();
                $adapter = $fcp_adapter_obj->adapter();
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }

            unless (defined($node) and $node ne "") {
                $self->_set_err("Invalid data returned for 'node'");
                return {};
            }
            unless (defined($adapter) and $adapter ne "") {
                $self->_set_err("Invalid data returned for 'adapter' for node '$node'");
                return {};
            }

            # Check to see if the adapter passes the regex filter
            unless ($adapter =~ $fpv_port_name_regex_filter) {
                # Port doesn't match the regex filter so skip it.
                next;
            }

            try {
                # Process the rest of the fields and store the data in $fcp_port_hash
                $fcp_port_hash->{$node}->{$adapter}->{fabric_established} = $fcp_adapter_obj->fabric_established();
                $fcp_port_hash->{$node}->{$adapter}->{status}             = $fcp_adapter_obj->status();
                $fcp_port_hash->{$node}->{$adapter}->{conn_established}   = $fcp_adapter_obj->conn_established();
                $fcp_port_hash->{$node}->{$adapter}->{state}              = $fcp_adapter_obj->state();
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }

           my ($key, $value);
            while (($key, $value) = each %{$fcp_port_hash->{$node}->{$adapter}}) {
                unless (defined($value and $value ne "")) {
                    $self->_set_err("Invalid data returned for '$key' on node '$node' port '$adapter'");
                    return {};
                }
            }
        }

        # burt645617: display various fcp data
        try {
            NACL::CS::VserverFcpInitiator->fetch(command_interface => $nacl_node_obj,
                                                 filter => {'node' => $node_name},
                                                 allow_empty       => 1);
            NACL::CS::VserverFcpInterface->fetch(command_interface => $nacl_node_obj,
                                                 filter => {'curr-node' => $node_name},
                                                 allow_empty       => 1);
        } otherwise {
            # burt776530: ignore errors
        };
    }

    return $fcp_port_hash;
}

###############################################################################
# Method:    _get_cna_port_info
# Objective: Build a hash of cna information for each node passed in.
#            Format of lif hash:
#            $cna_port_hash =>
#                    <node_name> =>
#                           <adapter> =>
#                                fc_status   => FC status
#                                link_speed  => Link Speed
#                                link_status => Link Status
###############################################################################
sub _get_cna_port_info {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $nacl_nodes_cm = &ParamsMgr::get_arg(-arg      => "nacl_nodes_cm",
                                            -required => 0,
                                            -default  => [],
                                            -ref_type => "ARRAY");

    my $reconnect     = &ParamsMgr::get_arg(-arg      => "reconnect",
                                            -required => 0,
                                            -default  => 0);

    if (&ParamsMgr::error()) {
        $self->set_err(&ParamsMgr::get_error());
        return {};
    }

    # Get the regex filter
    my $cpv_port_name_regex_filter = $self->EventUtil::get_arg(-arg => "cna_port_validation_port_name_regex_filter");

    # Create the cna port hash.
    my $cna_port_hash = {};
    foreach my $node_name (sort @$nacl_nodes_cm) {
        my $reconnected = 0;

        # See if we were requested to reconnect
        if ($reconnect == 1) {
            $reconnected = 1;
        }

        # Get the node object
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the object for node '$node_name'");
            return {};
        }

        my $output;
        try {
            my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                interface => "CLI",
                                                set => "CMode");
            my $cmd = "run -node $node_name -command \"priv set diag; cna show\"";
            $output = $apiset->execute_raw_command(command => $cmd);
        } otherwise {
            my $e = shift;
            $self->_set_err("Could not obtain CNA adapter information for node '$node_name'. " . $e->text());
        };
        if ($self->error()) {
            return {};
        }

        # split the cna show output on newlines
        my @cna_show;
        foreach my $line (split(/\n+/, $output)) {
            # Remove any leading and trailing whitespace
            $line =~ s/(^\s+)|\s+$//g;
            push (@cna_show, $line);
        }

        # Process the output from cna show
        my $adapter     = undef;
        my $fc_status   = undef;
        my $link_speed  = undef;
        my $link_status = undef;

        foreach my $line (@cna_show) {
            chomp($line);
            next if ($line =~ /^\s*$/);

            if ($line =~ /^FC Adapter Name\:\s+(.*)$/i) {
                if (not defined($adapter)) {
                    $adapter = $1;
                } else {
                    $self->_set_err("Error while processing the output from 'cna show'.  Adapter '$adapter' " .
                                    "is already being processed and a new adapter was found in the output");
                    return {};
                }
            } elsif ($line =~ /^FC Status\:\s+(.*)$/i) {
                if (not defined($fc_status)) {
                    $fc_status = $1;
                } else {
                    $self->_set_err("Error while processing the output from 'cna show'.  FC Status '$fc_status' " .
                                    "is already being processed and a new FC Status was found in the output");
                    return {};
                }
            } elsif ($line =~ /^Link Speed\:\s+(.*)$/i) {
                if (not defined($link_speed)) {
                    $link_speed = $1;
                } else {
                    $self->_set_err("Error while processing the output from 'cna show'.  Link Speed '$link_speed' " .
                                    "is already being processed and a new link speed was found in the output");
                    return {};
                }
            } elsif ($line =~ /^Link Status\:\s+(.*)$/i) {
                if (not defined($link_status)) {
                    $link_status = $1;
                } else {
                    $self->_set_err("Error while processing the output from 'cna show'.  Link Status '$link_status' " .
                                    "is already being processed and a new link status was found in the output");
                    return {};
                }
            }

            # Check to see if we have everything we are looking for
            if (defined($adapter) and defined($fc_status) and defined($link_speed) and defined($link_status)) {
                # We have everything we need so store it and reset the variables
                # Check to see if the adapter passes the regex filter
                if ($adapter =~ $cpv_port_name_regex_filter) {
                    $cna_port_hash->{$node_name}->{$adapter}->{'fc_status'}   = $fc_status;
                    $cna_port_hash->{$node_name}->{$adapter}->{'link_speed'}  = $link_speed;
                    $cna_port_hash->{$node_name}->{$adapter}->{'link_status'} = $link_status;
                }

                # Reset the variables we store everything in
                $adapter     = undef;
                $fc_status   = undef;
                $link_speed  = undef;
                $link_status = undef;
            }
        }

        # check to see if any of the variables are defined.  If they are, then
        # that would indicate that we found another adapter but didn't find all
        # of the information.  If that is the case, then throw an error.
        if (defined($adapter) or defined($fc_status) or defined($link_speed) or defined($link_status)) {
            # Looks like we found something but not everything so this is an error
            $self->_set_err("Error while processing the output from 'cna show'.  It looks like we found " .
                            "some data for a port but not all of the expected data.  Perhaps the output " .
                            "was truncated.");
            return {};
        }
    }

    return $cna_port_hash;
}

###############################################################################
# Method:    _validate_before_failover()
# Objective: Validate the specified nacl_nodes_cm before failover
###############################################################################
sub _validate_before_failover {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $id_extension = &ParamsMgr::get_arg(-arg => "id_extension", -required => 1);

    my @nacl_nodes_cm = @{&ParamsMgr::get_arg(-arg => "nacl_nodes_cm", -ref_type => "ARRAY",-default => [])};

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine if validation checks are enabled in the current event/eventset
    my $bcomd_info_validation = $self->EventUtil::get_arg(-arg => "bcomd_info_validation");

    # if no validation checks are enabled, then return immediately
    if (!$bcomd_info_validation){
        $Log->comment("Skipping controller validation before failover becase it was disabled.");
        return $EventUtil::TRUE;
    }

    my $controller_abort_on_error = $self->EventUtil::get_arg(-arg => "controller_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $bcomd_info_validation_abort_on_error = $self->EventUtil::get_arg(-arg => "bcomd_info_validation_abort_on_error",
                                                              -default_to_env => 1);

    my @healthy_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE, -type => "cmode");
    if ($self->error()) {
        $self->_set_err("Failed to find healthy nodes to use for validation.");
        return $EventUtil::FALSE;
    }

    # get the next padded runid from the event
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $validation_runid_padded = $event_obj->increment_subtest_runid_counter();

    # if there are no nacl_nodes to perform validation on at this point, then return immediately
    unless (scalar(@healthy_nodes)) {
        $Log->comment("Skipping Validation before failover because there " .
                      "are no healthy nacl_nodes_cm resources to operate against");
        return $EventUtil::TRUE;
    }

     # Process each healthy node and make sure everything matches
     my @subtests;
     foreach my $node (sort @healthy_nodes) {
        # need to run a command on each node before doing validation, so that versioning occurs in the parent subtest
        # this is a known NACL limitation, otherwise it will open many connections to the controllers in each child subtest
        my $node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node);
        try {
            NACL::CS::SystemNode->fetch(command_interface => $node_obj);
        } otherwise {
        };
        $self->_clear_err();

        # If the user passed in a list of nacl_nodes_cm then see if this node
        # is in that list and skip it if it is not.
        if (scalar(@nacl_nodes_cm)) {
            unless (grep(/^$node$/, @nacl_nodes_cm)) {
                next;
            }
        }

        my ($subtest, $id);

        # run bcomd_info_validation for the node if bcomd validation is enabled
        if ($bcomd_info_validation) {
            $id = $validation_runid_padded . "_verify_bcomd_info_${node}_${id_extension}";
            $subtest = subtest(sub {
                                    $self->ControllerValidation::_validate_bcomd_info(-node => $node);
                                    if ($self->error()) {
                                        $self->_set_err("bcomd_info validation failed for node: '$node'");
                                        # perform a graceful failure if this client
                                        # has $bcomd_bringup_info_validation_abort_on_error == 0
                                        # or if $controller_abort_on_error == 0
                                        my $graceful = (!$controller_abort_on_error or !$bcomd_info_validation_abort_on_error) ? 1 : 0;

                                        $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                        return $EventUtil::FALSE;
                                   }
                                  $self->SubtestComm::subtest_complete();
                                  return $EventUtil::TRUE;
                               }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            $Log->comment("Skipping bcomd_bringup_info validation for this event because it is disabled");
        }
     }

    # wait for all the validation subtests to complete
    if (scalar(@subtests)) {
        $self->wait_for_subtests(-subtests => \@subtests,
                                 -dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 1,
                                 -enable_sync => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -async => 0);
        if ($self->error()) {
            $self->_set_err("Controller validation failed on a node");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_on_failover()
# Objective: Validate the specified nacl_nodes_cm during failover
###############################################################################
sub _validate_on_failover {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $id_extension = &ParamsMgr::get_arg(-arg => "id_extension", -required => 1);

    my @nacl_nodes_cm = @{&ParamsMgr::get_arg(-arg      => "nacl_nodes_cm",
                                              -ref_type => "ARRAY",
                                              -default  => [])};

    # the caller may have requested that we wait a minimum timeout
    # before performing lif validation
    my $last_timeout = &ParamsMgr::get_arg(-arg => "last_timeout", -default => 0);
    my $last_timeout_name = &ParamsMgr::get_arg(-arg => "last_timeout_name", -default => "timeout");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine if validation checks are enabled in the current event/eventset
    my $lif_validation = $self->EventUtil::get_arg(
            -arg => "lif_validation");
    my $net_validation = $self->EventUtil::get_arg(
            -arg => "net_port_validation");
    my $fcp_validation = $self->EventUtil::get_arg(
            -arg => "fcp_port_validation");
    my $cna_validation = $self->EventUtil::get_arg(
            -arg => "cna_port_validation");
    my $coredump_validation = $self->EventUtil::get_arg(
            -arg => "coredump_validation");
    my $bcomd_bringup_info_validation = $self->EventUtil::get_arg(
            -arg => "bcomd_bringup_info_validation");

    # if validation checks are not enabled for this current event, then return immediately
    if (!$lif_validation and !$net_validation and !$fcp_validation and !$cna_validation and
        !$coredump_validation and !$bcomd_bringup_info_validation) {
        $Log->comment("Skipping controller validation for this event because all validation checks are disabled");
        return $EventUtil::TRUE;
    }

    # determine if we want to abort on error
    my $controller_abort_on_error = $self->EventUtil::get_arg(-arg => "controller_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $lv_abort_on_error         = $self->EventUtil::get_arg(-arg => "lif_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $npv_abort_on_error        = $self->EventUtil::get_arg(-arg => "net_port_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $fpv_abort_on_error        = $self->EventUtil::get_arg(-arg => "fcp_port_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $cpv_abort_on_error        = $self->EventUtil::get_arg(-arg => "cna_port_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $coredump_abort_on_error   = $self->EventUtil::get_arg(-arg => "coredump_validation_abort_on_error",
                                                              -default_to_env => 0);
    my $bcomd_bringup_info_validation_abort_on_error =
                                    $self->EventUtil::get_arg(-arg => "bcomd_bringup_info_validation_abort_on_error",
                                                              -default_to_env => 1);

    my @healthy_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE, -type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # if there are no nacl_nodes to perform validation on at this point, then return immediately
    unless (scalar(@healthy_nodes)) {
        $Log->comment("Skipping Validation for this event because there " .
                      "are no healthy nacl_nodes_cm resources to operate against");
        return $EventUtil::TRUE;
    }

    # wait controller validation min wait time if a value for
    # last timeout was passed in
    if ($last_timeout > 0) {
        $self->ControllerValidation::_wait_for_validate(-last_timeout => $last_timeout,
                                                        -last_timeout_name => $last_timeout_name);
    }

    # get the next padded runid from the event
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $validation_runid_padded = $event_obj->increment_subtest_runid_counter();

    # Process each healthy node and make sure everything matches
    my @subtests;
    foreach my $node (sort @healthy_nodes) {
        # need to run a command on each node before doing validation, so that versioning occurs in the parent subtest
        # this is a known NACL limitation, otherwise it will open many connections to the controllers in each child subtest
        my $node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node);
        try {
            NACL::CS::SystemNode->fetch(command_interface => $node_obj);
        } otherwise {
        };
        $self->_clear_err();

        # If the user passed in a list of nacl_nodes_cm then see if this node
        # is in that list and skip it if it is not.
        if (scalar(@nacl_nodes_cm)) {
            unless (grep(/^$node$/, @nacl_nodes_cm)) {
                next;
            }
        }

        my ($subtest, $id);
        # run lif validation for the node if lif validation is enabled
        if ($lif_validation) {
            $id = $validation_runid_padded . "_verify_lif_info_${node}_${id_extension}";
            $subtest = subtest(sub {
                                   $self->ControllerValidation::_validate_lifs_per_node(-node => $node);
                                   if ($self->error()) {
                                       $self->_set_err("LIF validation failed for node '$node'");
                                       # perform a graceful failure if this client has lv_abort_on_error == 0
                                       # or if $controller_abort_on_error == 0
                                       my $graceful = (!$controller_abort_on_error or !$lv_abort_on_error) ? 1 : 0;

                                       $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                       return $EventUtil::FALSE;
                                   }
                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            $Log->comment("Skipping LIF validation for this event because it is disabled");
        }

        # run network port validation for the node if network port validation is enabled
        if ($net_validation) {
            $id = $validation_runid_padded . "_verify_network_port_info_${node}_${id_extension}";
            $subtest = subtest(sub {
                                   $self->ControllerValidation::_validate_network_ports_per_node(-node => $node);
                                   if ($self->error()) {
                                       $self->_set_err("Network port validation failed for node '$node'");
                                       # perform a graceful failure if this client has npv_abort_on_error == 0
                                       # or if $controller_abort_on_error == 0
                                       my $graceful = (!$controller_abort_on_error or !$npv_abort_on_error) ? 1 : 0;

                                       $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                       return $EventUtil::FALSE;
                                   }
                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            $Log->comment("Skipping Network Port validation for this event because it is disabled");
        }

        # run fcp adapter validation for the node if fcp adapter validation is enabled
        if ($fcp_validation) {
            $id = $validation_runid_padded . "_verify_fcp_port_info_${node}_${id_extension}";
            $subtest = subtest(sub {
                                   $self->ControllerValidation::_validate_fcp_ports_per_node(-node => $node);
                                   if ($self->error()) {
                                       $self->_set_err("FCP adapter validation failed for node '$node'");
                                       # perform a graceful failure if this client has fpv_abort_on_error == 0
                                       # or if $controller_abort_on_error == 0
                                       my $graceful = (!$controller_abort_on_error or !$fpv_abort_on_error) ? 1 : 0;

                                       $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                       return $EventUtil::FALSE;
                                   }
                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            $Log->comment("Skipping FCP adapter validation for this event because it is disabled");
        }

        # run cna adapter validation for the node if cna adapter validation is enabled
        if ($cna_validation) {
            $id = $validation_runid_padded . "_verify_cna_port_info_${node}_${id_extension}";
            $subtest = subtest(sub {
                                   $self->ControllerValidation::_validate_cna_ports_per_node(-node => $node);
                                   if ($self->error()) {
                                       $self->_set_err("CNA adapter validation failed for node '$node'");
                                       # perform a graceful failure if this client has cpv_abort_on_error == 0
                                       # or if $controller_abort_on_error == 0
                                       my $graceful = (!$controller_abort_on_error or !$cpv_abort_on_error) ? 1 : 0;

                                       $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                       return $EventUtil::FALSE;
                                   }
                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            $Log->comment("Skipping CNA adapter validation for this event because it is disabled");
        }

        if ($coredump_validation) {
            $id = $validation_runid_padded . "_verify_coredump_${node}_${id_extension}";
            $subtest = subtest(sub {
                                   $self->ControllerValidation::_validate_coredumps_per_node(-node => $node);
                                   if ($self->error()) {
                                       $self->_set_err("Coredump validation failed for node '$node'");
                                       # perform a graceful failure if this client has coredump_abort_on_error == 0
                                       # or if $controller_abort_on_error == 0
                                       my $graceful = (!$controller_abort_on_error or !$coredump_abort_on_error) ? 1 : 0;

                                       $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                       return $EventUtil::FALSE;
                                   }
                                   $self->SubtestComm::subtest_complete();
                                   return $EventUtil::TRUE;
                               }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            $Log->comment("Skipping Coredump validation for this event because it is disabled");
        }

         # run bcomd timeout validation for the node if bcomd validation is enabled
         # This block is designed to validate that the latency for the nodes in the
         # bcomd_bringup_info table, with the state 'VS_JOIN_POLL,' does not exceed
         # 5 minutes. This block can be extended later to reflect validation
         # for other states or fields within the table.
        if ($bcomd_bringup_info_validation) {
            $id = $validation_runid_padded . "_verify_bcomd_bringup_info_${node}_${id_extension}";
            my $is_release_SN_or_after;

            try {
                $is_release_SN_or_after = $node_obj->version_manager()->is_version_ge(release =>'8.2');
            } otherwise {
                my $e =shift;
                $self->_set_err("Failed to determine if the release is 8.2 or later", $e->text());
            };
            if ($self->error()) {
               return $EventUtil::FALSE;
            }

           # Check if this version of ONTAP can be used to validate.
           # bcomd_bringup_info_table is only available in ONTAP versions >= 8.2
           if ($is_release_SN_or_after) {
               $subtest = subtest(sub {
                                       $self->ControllerValidation::_validate_bcomd_bringup_info_timeout_per_node(-node => $node);
                                       if ($self->error()) {
                                           $self->_set_err("bcomd_bringup_info validation failed for node: '$node'");
                                           # perform a graceful failure if this client
                                           # has $bcomd_bringup_info_validation_abort_on_error == 0
                                           # or if $controller_abort_on_error == 0
                                           my $graceful = (!$controller_abort_on_error or !$bcomd_bringup_info_validation_abort_on_error) ? 1 : 0;

                                           $self->SubtestComm::subtest_fail(-graceful => $graceful);
                                           return $EventUtil::FALSE;
                                      }
                                      $self->SubtestComm::subtest_complete();
                                      return $EventUtil::TRUE;
                                   }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
           }
           else {
               $Log->comment("Skipping bcomd_bringup_info validation for this event. Node (" . $node .
                      ") is not running Data ONTAP v8.2 or above.");
           }
        } else {
            $Log->comment("Skipping bcomd_bringup_info validation for this event because it is disabled");
        }
    }

    # wait for all the path validation subtests to complete
    if (scalar(@subtests)) {
        $self->wait_for_subtests(-subtests => \@subtests,
                                 -dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 1,
                                 -enable_sync => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -async => 0);
        if ($self->error()) {
            $self->_set_err("Controller validation failed on a node");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}
###############################################################################
# Method:    _validate_coredumps_per_node()
# Objective: Validate the paths for the given node.
###############################################################################
sub _validate_coredumps_per_node {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }
    # Get the node object
    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node);
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the object for node '$node'");
        return $EventUtil::FALSE;
    }
    my $no_new_coredump = $self->ControllerValidation::check_node_for_coredump(-node => $nacl_node_obj);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    if (not $no_new_coredump) {
        # error already set
        return $EventUtil::FALSE;
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_lifs_per_node()
# Objective: Validate the paths for the given node.
###############################################################################
sub _validate_lifs_per_node {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Grab the lif info hash resource object if the steady state has been loaded
    # otherwise error out
    unless ($self->ControllerValidation::_is_steady_state_loaded(-resource => "lif_info")) {
        $self->_set_err("LIF steady state information was not loaded into Resource Manager");
        return $EventUtil::FALSE;
    }

    my $resource_hash = $self->ResMgr::get_resource(-name => "lif_info",
                                                    -type => "_GLOBALS");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $steady_state_hash = $resource_hash->{obj};

    # Build out the lif hash for the current state of the healthy nodes and faulted nodes
    my $node_lif_state = $self->ControllerValidation::_get_lif_info(-nacl_nodes_cm => [$node], -reconnect => 1);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # LIF Info Hash Structure:
    #
    #            $lif_hash =>
    #                    <node_name> =>
    #                       <vserver_name> =>
    #                           <lif_name> =>
    #                                current_node  => Node where the lif currently resides
    #                                current_port  => Port where the lif currently resides
    #                                home_node     => Node where the lif should reside
    #                                home_port     => Name of home port
    #                                data_protocol => Data protocol of lif
    #                                status_oper   => Operational status of lif
    #                                status_admin  => Admin status of lif
    #                                is_home       => Is the LIF at home

    # log an event that we are validating paths for this client
    $self->SubtestComm::log_event_message(-event_message => "lv_validating_lifs_on_${node}");

    # Determine how they want to handle new resources that might
    # be found during validation
    my $lv_new_resource_behavior = $self->EventUtil::get_arg(-arg => "lif_validation_new_resource_behavior");

    my @errors_found = ();
    foreach my $vserver (sort keys %{$node_lif_state->{$node}}) {
        foreach my $lif (sort keys %{$node_lif_state->{$node}->{$vserver}}) {
            my $lif_is_home = $node_lif_state->{$node}->{$vserver}->{$lif}->{is_home};

            if (exists($steady_state_hash->{$node}->{$vserver}->{$lif})) {
                # We have steady state data for the lif.
                # Make sure all the pertinent values match.  Save any errors in @errors_found array
                # Check to see if this lif is home.  If it isn't, then we will get a mismatch
                # on keys. If the protocol is nfs or cifs, we don't want to compare the current node and current
                # port to steady state since it might have migrated.
                my @keys_to_verify = ();
                if ($node_lif_state->{$node}->{$vserver}->{$lif}->{data_protocol} =~ /(nfs|cifs|fcache|none)/i) {
                    @keys_to_verify = ("status_oper", "status_admin", "data_protocol", "home_node", "home_port");
                } else {
                    @keys_to_verify = (keys %{$node_lif_state->{$node}->{$vserver}->{$lif}});
                }

                # Check the lif keys to see if they match.  If they don't, add the key name to the bad_keys array.
                my @bad_keys = ();
                foreach my $key (sort @keys_to_verify) {
                    unless ($node_lif_state->{$node}->{$vserver}->{$lif}->{$key} eq $steady_state_hash->{$node}->{$vserver}->{$lif}->{$key}) {
                        push(@bad_keys, $key);
                    }
                }

                # If we have a bad lif status, then add the lif information to the errors_found array.
                if (scalar(@bad_keys)) {
                    my $mismatched_keys = join("|", @bad_keys);
                    push(@errors_found,  sprintf("%-20s %-10s %-20s %-12s %-20s %-25s %-8s %-12s %-12s %-8s %s",
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'home_node'},
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'home_port'},
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'current_node'},
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'current_port'},
                                                 $vserver,
                                                 $lif,
                                                 $lif_is_home,
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'status_oper'},
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'status_admin'},
                                                 $node_lif_state->{$node}->{$vserver}->{$lif}->{'data_protocol'},
                                                 $mismatched_keys));
                }
            } else {
                # We don't have the steady state info for this lif/vserver combination.  It
                # is possible that it was created after initialization or that it is an NFS/CIFS
                # LIF that migrated.
                if ($node_lif_state->{$node}->{$vserver}->{$lif}->{'data_protocol'} =~ /(nfs|cifs|fcache|none)/i) {
                    # The LIF meets the following criteria:
                    # 1.  We don't have steady state
                    # 2.  The lif is an nfs or cifs lif
                    # 3.  It could have migrated from the other node.
                    #
                    # Just make sure that if status_admin is up and that status_oper is also up.
                    # If status_admin is down, then we don't have any way to know if it is supposed
                    # to be up or down since we don't have the steady state information.
                    if ($node_lif_state->{$node}->{$vserver}->{$lif}->{'status_admin'} eq "up") {
                        # Status Admin is down.  Make sure that the status oper is also up.
                        if ($node_lif_state->{$node}->{$vserver}->{$lif}->{'status_oper'} eq "down") {
                            push(@errors_found, sprintf("%-20s %-10s %-20s %-12s %-20s %-25s %-8s %-12s %-12s %-8s %s",
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'home_node'},
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'home_port'},
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'current_node'},
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'current_port'},
                                                        $vserver,
                                                        $lif,
                                                        $lif_is_home,
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'status_oper'},
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'status_admin'},
                                                        $node_lif_state->{$node}->{$vserver}->{$lif}->{'data_protocol'},
                                                        "status_oper"));
                        }
                    } else {
                        # Status Admin is down.  We don't know if it is supposed to be down.
                        $Log->warn("Unable to find steady state information for Node '$node' " .
                                   "Vserver '$vserver' Lif '$lif'.\n" .
                                   "This LIF is not currently home and the status_admin state is " .
                                   "currently not up");
                        next;
                    } ##END if ($node_lif_state->{$node}->{$vserver}->{$lif}->{'status_admin'} eq "up") {
                } else {
                    # 1. We don't have steady state
                    # 2. The LIF  not an nfs/cifs lif
                    # 3. The LIF shouldn't have showed up during a migration.
                    if ($lv_new_resource_behavior eq "fatal") {
                        $self->_set_err("Unable to find steady state information for Node '$node' " .
                                        "Vserver '$vserver' Lif '$lif'");
                        return $EventUtil::FALSE;
                    } elsif ($lv_new_resource_behavior eq "warning") {
                        $Log->warn("Unable to find steady state information for Node '$node' " .
                                   "Vserver '$vserver' Lif '$lif'");
                        next;
                    } elsif ($lv_new_resource_behavior eq "ignore") {
                        $Log->comment("Unable to find steady state information for Node '$node' " .
                                      "Vserver '$vserver' Lif '$lif'");
                        next;
                    } else {
                        $self->_set_err("Invalid value specified for attribute 'lif_validation_new_resource_behavior'. " .
                                        "Valid values are 'fatal', 'warning', and 'ignore'");
                        return $EventUtil::FALSE;
                    } ##END if ($lv_new_resource_behavior eq "fatal") {
                } ##END if ($node_lif_state->{$node}->{$vserver}->{$lif}->{'is_home'} == $EventUtil::FALSE) {
            } ##END if (exists($steady_state_hash->{$node}->{$vserver}->{$lif})) {
        } ##END foreach my $lif (sort keys %{$node_lif_state->{$node}->{$vserver}}) {
    } ##END foreach my $vserver (sort keys %{$node_lif_state->{$node}}) {

    # See if we found errors.  If we did, set error and bail.
    if (scalar(@errors_found)) {
        unshift(@errors_found, sprintf("%-20s %-10s %-20s %-12s %-20s %-25s %-8s %-12s %-12s %-8s %s",
                                       "====================",
                                       "==========",
                                       "====================",
                                       "============",
                                       "====================",
                                       "=========================",
                                       "========",
                                       "============",
                                       "============",
                                       "========",
                                       "===================="));
        unshift(@errors_found, sprintf("%-20s %-10s %-20s %-12s %-20s %-25s %-8s %-12s %-12s %-8s %s",
                                       "Home Node", "Home Port", "Current Node", "Current Port",
                                       "Vserver", "LIF", "Is Home", "Status Oper", "Status Admin",
                                       "Protocol", "Mismatched Keys"));

        unshift(@errors_found, "The following LIFs were found with errors during validation:\n");
        push (@errors_found, "\n");
        my $errmsg = join("\n", @errors_found);
        $Log->comment($errmsg);
        $self->_set_err("LIF Validation failed for one or more nodes");
        return $EventUtil::FALSE;
    } else {
        $Log->comment("All LIFs that belong to node '$node' have been validated");
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_network_ports_per_node()
# Objective: Validate the network ports per node
###############################################################################
sub _validate_network_ports_per_node {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Grab the network info hash resource object if the steady state has been loaded
    # otherwise error out
    unless ($self->ControllerValidation::_is_steady_state_loaded(-resource => "network_port_info")) {
        $self->_set_err("Network Port steady state information was not loaded into Resource Manager");
        return $EventUtil::FALSE;
    }

    my $resource_hash = $self->ResMgr::get_resource(-name => "network_port_info",
                                                    -type => "_GLOBALS");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $steady_state_hash = $resource_hash->{obj};

    # Build out the network port hash for the current state of the healthy nodes and faulted nodes
    #            $network_port_hash =>
    #                    <node_name> =>
    #                           <port> =>
    #                                link          => Link status
    #                                role/ipspace  => Role/IPspace of the port
    #                                mtu           => MTU of the port
    #                                speed_admin   => Speed Admin
    #                                speed_oper    => Speed Oper
    #                                duplex_admin  => Duplex Admin
    #                                duplex_oper   => Duplex Oper
    #
    my $network_port_hash = $self->ControllerValidation::_get_network_port_info(-nacl_nodes_cm => [$node], -reconnect => 1);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node);
    if ($self->error()) {
        $self->_set_err("Could not obtain nacl node object for $node");
        return $EventUtil::FALSE;
    }

    # log an event that we are validating the network ports for this node
    $self->SubtestComm::log_event_message(-event_message => "lv_validating_network_ports_on_${node}");

    my @errors_found = ();
    foreach my $port (sort keys %{$network_port_hash->{$node}}) {
        my @bad_keys = ();
        if (exists($steady_state_hash->{$node})) {
            if (exists($steady_state_hash->{$node}->{$port})) {
                # We have steady state data for the port.
                # Make sure all the pertinent values match.
                foreach my $key (sort keys %{$network_port_hash->{$node}->{$port}}) {
                    unless (defined($steady_state_hash->{$node}->{$port}->{$key}) and
                            $network_port_hash->{$node}->{$port}->{$key} eq $steady_state_hash->{$node}->{$port}->{$key}) {
                        push(@bad_keys, $key);
                    }
                }

                # If we have a bad status, then add the port information to the errors_found array.
                if (scalar(@bad_keys)) {
                    my $mismatched_keys = join("|", @bad_keys);
                    if ($nacl_node_obj->version_manager->has_uichange(uichange => 'ipspaces-ms4',
                                                                      throw_exception => 0)) {
                        push(@errors_found, sprintf("%-20s %-8s %-8s %-8s %-13s %-12s %-12s %-12s %-12s %s",
                                                    $node,
                                                    $port,
                                                    $network_port_hash->{$node}->{$port}->{'mtu'},
                                                    $network_port_hash->{$node}->{$port}->{'link'},
                                                    $network_port_hash->{$node}->{$port}->{'ipspace'},
                                                    $network_port_hash->{$node}->{$port}->{'speed_admin'},
                                                    $network_port_hash->{$node}->{$port}->{'speed_oper'},
                                                    $network_port_hash->{$node}->{$port}->{'duplex_admin'},
                                                    $network_port_hash->{$node}->{$port}->{'duplex_oper'},
                                                    $mismatched_keys));
                    } else {
                        push(@errors_found, sprintf("%-20s %-8s %-8s %-8s %-13s %-12s %-12s %-12s %-12s %s",
                                                    $node,
                                                    $port,
                                                    $network_port_hash->{$node}->{$port}->{'mtu'},
                                                    $network_port_hash->{$node}->{$port}->{'link'},
                                                    $network_port_hash->{$node}->{$port}->{'role'},
                                                    $network_port_hash->{$node}->{$port}->{'speed_admin'},
                                                    $network_port_hash->{$node}->{$port}->{'speed_oper'},
                                                    $network_port_hash->{$node}->{$port}->{'duplex_admin'},
                                                    $network_port_hash->{$node}->{$port}->{'duplex_oper'},
                                                    $mismatched_keys));
                    }
                } ##END if (scalar(@bad_keys)) {
            } else {
                # We don't have the steady state info for this node/port combo
                $self->_set_err("Unable to find steady state information for Node '$node' Port '$port'");
                return $EventUtil::FALSE;
            } ##END if (exists($steady_state_hash->{$node}->{$port})) {
        } else {
            # We don't have the steady state info for this node combo
            $self->_set_err("Unable to find steady state information for Node '$node'");
            return $EventUtil::FALSE;
        } ##END if (exists($steady_state_hash->{$node})) {
    } ##END foreach my $port (sort keys %{$network_port_hash->{$node}}) {

    # See if we found errors.  If we did, set error and bail.
    if (scalar(@errors_found)) {
        unshift(@errors_found, sprintf("%-20s %-8s %-8s %-8s %-13s %-12s %-12s %-12s %-12s %s",
                                       "====================",
                                       "========",
                                       "========",
                                       "========",
                                       "==========",
                                       "============",
                                       "============",
                                       "============",
                                       "============",
                                       "===================="));
        unshift(@errors_found, sprintf("%-20s %-8s %-8s %-8s %-13s %-12s %-12s %-12s %-12s %s",
                                       "Node",
                                       "Port",
                                       "MTU",
                                       "Link",
                                       "Role/IPSpace",
                                       "Speed_Admin",
                                       "Speed_Oper",
                                       "Duplex_Admin",
                                       "Duplex_Oper",
                                       "Mismatched Keys"));

        unshift(@errors_found, "The following network ports were found with errors during validation:\n");
        push (@errors_found, "\n");
        my $errmsg = join("\n", @errors_found);
        $Log->comment($errmsg);
        $self->_set_err("Network Port Validation failed for one or more nodes");
        return $EventUtil::FALSE;
    } else {
        $Log->comment("All network ports that belong to node '$node' have been validated");
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_fcp_ports_per_node()
# Objective: Validate the fcp adapters per node
###############################################################################
sub _validate_fcp_ports_per_node {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Grab the fcp adapter info hash resource object if the steady state has been loaded
    # otherwise error out
    unless ($self->ControllerValidation::_is_steady_state_loaded(-resource => "fcp_port_info")) {
        $self->_set_err("FCP Adapter steady state information was not loaded into Resource Manager");
        return $EventUtil::FALSE;
    }

    my $resource_hash = $self->ResMgr::get_resource(-name => "fcp_port_info",
                                                    -type => "_GLOBALS");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $steady_state_hash = $resource_hash->{obj};

    # Build out the fcp adapter hash for the current state of the healthy nodes and faulted nodes
    #            $fcp_port_hash =>
    #                    <node_name> =>
    #                           <adapter> =>
    #                                fabric_established  => Fabric Established
    #                                status              => Status Field
    #                                state               => Link State
    #                                conn_established    => Connection Established
    #                                lif_cable           => Whether adapter is lif capable
    #
    my $fcp_port_hash = $self->ControllerValidation::_get_fcp_port_info(-nacl_nodes_cm => [$node], -reconnect => 1);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # log an event that we are validating the fcp adapters for this node
    $self->SubtestComm::log_event_message(-event_message => "lv_validating_fcp_ports_on_${node}");

    my @errors_found = ();
    foreach my $port (sort keys %{$fcp_port_hash->{$node}}) {
        my @bad_keys = ();
        if (exists($steady_state_hash->{$node})) {
            if (exists($steady_state_hash->{$node}->{$port})) {
                # We have steady state data for the adapter.
                # Make sure all the pertinent values match.
                foreach my $key (sort keys %{$fcp_port_hash->{$node}->{$port}}) {
                    unless (defined($steady_state_hash->{$node}->{$port}->{$key}) and
                            $fcp_port_hash->{$node}->{$port}->{$key} eq $steady_state_hash->{$node}->{$port}->{$key}) {
                        push(@bad_keys, $key);
                    }
                }

                # If we have a bad status, then add the port information to the errors_found array.
                if (scalar(@bad_keys)) {
                    my $mismatched_keys = join("|", @bad_keys);
                    push(@errors_found, sprintf("%-20s %-8s %-18s %-8s %-25s %-10s %s",
                                                $node,
                                                $port,
                                                $fcp_port_hash->{$node}->{$port}->{'fabric_established'},
                                                $fcp_port_hash->{$node}->{$port}->{'state'},
                                                $fcp_port_hash->{$node}->{$port}->{'status'},
                                                $fcp_port_hash->{$node}->{$port}->{'conn_established'},
                                                $mismatched_keys));
                } ##END if (scalar(@bad_keys)) {
            } else {
                # We don't have the steady state info for this node/port combo
                $self->_set_err("Unable to find steady state information for Node '$node' Port '$port'");
                return $EventUtil::FALSE;
            } ##END if (exists($steady_state_hash->{$node}->{$port})) {
        } else {
            # We don't have the steady state info for this node combo
            $self->_set_err("Unable to find steady state information for Node '$node'");
            return $EventUtil::FALSE;
        } ##END if (exists($steady_state_hash->{$node})) {
    } ##END foreach my $port (sort keys %{$fcp_port_hash->{$node}}) {

    # See if we found errors.  If we did, set error and bail.
    if (scalar(@errors_found)) {
        unshift(@errors_found, sprintf("%-20s %-8s %-18s %-8s %-25s %-10s %s",
                                       "====================",
                                       "========",
                                       "==================",
                                       "========",
                                       "=========================",
                                       "==========",
                                       "===================="));
        unshift(@errors_found, sprintf("%-20s %-8s %-18s %-8s %-25s %-10s %s",
                                       "Node",
                                       "Adapter",
                                       "Fabric Established",
                                       "State",
                                       "Status",
                                       "Conn Type",
                                       "Mismatched Keys"));

        unshift(@errors_found, "The following fcp adapters were found with errors during validation:\n");
        push (@errors_found, "\n");
        my $errmsg = join("\n", @errors_found);
        $Log->comment($errmsg);
        $self->_set_err("FCP Validation failed for one or more nodes");
        return $EventUtil::FALSE;
    } else {
        $Log->comment("All fcp adapters that belong to node '$node' have been validated");
    }
    return $EventUtil::TRUE;
}

##############################################################################
# Method:    _validate_cna_ports_per_node()
# Objective: Validate the cna ports per node
###############################################################################
sub _validate_cna_ports_per_node {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Grab the cna adapter info hash resource object if the steady state has been loaded
    # otherwise error out
    unless ($self->ControllerValidation::_is_steady_state_loaded(-resource => "cna_port_info")) {
        $self->_set_err("CNA Port steady state information was not loaded into Resource Manager");
        return $EventUtil::FALSE;
    }

    my $resource_hash = $self->ResMgr::get_resource(-name => "cna_port_info",
                                                    -type => "_GLOBALS");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $steady_state_hash = $resource_hash->{obj};

    # Create the CNA Port Hash
    #            $cna_port_hash =>
    #                    <node_name> =>
    #                           <adapter> =>
    #                                fc_status   => FC status
    #                                link_speed  => Link Speed
    #                                link_status => Link Status
    my $cna_port_hash = $self->ControllerValidation::_get_cna_port_info(-nacl_nodes_cm => [$node], -reconnect => 1);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # log an event that we are validating the cna ports for this node
    $self->SubtestComm::log_event_message(-event_message => "lv_validating_cna_ports_on_${node}");

    my @errors_found = ();
    foreach my $port (sort keys %{$cna_port_hash->{$node}}) {
        my @bad_keys = ();
        if (exists($steady_state_hash->{$node})) {
            if (exists($steady_state_hash->{$node}->{$port})) {
                # We have steady state data for the adapter.
                # Make sure all the pertinent values match.
                foreach my $key (sort keys %{$cna_port_hash->{$node}->{$port}}) {
                    unless (defined($steady_state_hash->{$node}->{$port}->{$key}) and
                            $cna_port_hash->{$node}->{$port}->{$key} eq $steady_state_hash->{$node}->{$port}->{$key}) {
                        push(@bad_keys, $key);
                    }
                }

                # If we have a bad status, then add the port information to the errors_found array.
                if (scalar(@bad_keys)) {
                    my $mismatched_keys = join("|", @bad_keys);
                    push(@errors_found, sprintf("%-20s %-8s %-20s %-20s %-15s %s",
                                                $node,
                                                $port,
                                                $cna_port_hash->{$node}->{$port}->{'fc_status'},
                                                $cna_port_hash->{$node}->{$port}->{'link_status'},
                                                $cna_port_hash->{$node}->{$port}->{'link_speed'},
                                                $mismatched_keys));
                } ##END if (scalar(@bad_keys)) {
            } else {
                # We don't have the steady state info for this node/port combo
                $self->_set_err("Unable to find steady state information for Node '$node' Port '$port'");
                return $EventUtil::FALSE;
            } ##END if (exists($steady_state_hash->{$node}->{$port})) {
        } else {
            # We don't have the steady state info for this node combo
            $self->_set_err("Unable to find steady state information for Node '$node'");
            return $EventUtil::FALSE;
        } ##END if (exists($steady_state_hash->{$node})) {
    } ##END foreach my $port (sort keys %{$cna_port_hash->{$node}}) {

    # See if we found errors.  If we did, set error and bail.
    if (scalar(@errors_found)) {
        unshift(@errors_found, sprintf("%-20s %-8s %-20s %-20s %-15s %s",
                                       "====================",
                                       "========",
                                       "====================",
                                       "====================",
                                       "===============",
                                       "===================="));
        unshift(@errors_found, sprintf("%-20s %-8s %-20s %-20s %-15s %s",
                                       "Node",
                                       "Adapter",
                                       "FC Status",
                                       "Link Status",
                                       "Link Speed",
                                       "Mismatched Keys"));

        unshift(@errors_found, "The following cna adapters were found with errors during validation:\n");
        push (@errors_found, "\n");
        my $errmsg = join("\n", @errors_found);
        $Log->comment($errmsg);
        $self->_set_err("CNA Validation failed for one or more nodes");
        return $EventUtil::FALSE;
    } else {
        $Log->comment("All cna adapters that belong to node '$node' have been validated");
    }
    return $EventUtil::TRUE;
}

##############################################################################
# Method:    _validate_bcomd_bringup_info_timeout_per_node()
# Objective: Validate the latency for the nodes in the bcomd_bringup_info table,
#            with state 'VS_JOIN_POLL,' has not exceeded 5 minutes
###############################################################################
sub _validate_bcomd_bringup_info_timeout_per_node {
    my ($self, %args) = @_;

    my $timeout_value = 300000;
    my $table = 'bcomd_bringup_info';
    my $desired_state = 'VS_JOIN_POLL';

    &ParamsMgr::process_args(%args);

    my $node_name = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $node = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
    if ($self->error()) {
        $self->_set_err("Failed to get nacl node objects for: " . $node_name);
        return $EventUtil::FALSE;
    }

    my $bcomd_bringup_info_table;

    # This line will return the parsed output of the bcomd_bringup_info table
    # The APISET Layer was used instead of the Component layer.
    # debugsmdb support at the Component level is not functional.
    # burt780133: need to force CLI since there are no ZAPIs for debug smdb commands
    try {
        $node->apiset_must({interface => 'CLI'});
        $bcomd_bringup_info_table = $node->apiset()->debug_smdb_table_show(table => 'bcomd_bringup_info',
                                                                           node  => $node_name)->{parsed_output};

    } otherwise {
        my $e = shift;
        $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Get latency state for the node
    my $node_latency_state;
    foreach my $hash (@{$bcomd_bringup_info_table}) {
        if($hash->{'init_state'} eq $desired_state){
            $node_latency_state = $hash->{'latency'};
       }
    }

    # Check if node latency state exceeds timeout value in milliseconds (5mins)
    if($node_latency_state > $timeout_value){
        $self->_set_err("The following node has timed out: " . $node->node);
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

##############################################################################
# Method:    _validate_bcomd_info()
# Objective: Validate the init_state for each node is at 'IDLE' state.
###############################################################################
sub _validate_bcomd_info {
    my ($self, %args) = @_;

    my $controller_abort_on_error = $self->EventUtil::get_arg(-arg => "controller_validation_abort_on_error",
                                                              -default_to_env => 1);
    my $bcomd_info_validation_abort_on_error = $self->EventUtil::get_arg(-arg => "bcomd_info_validation_abort_on_error",
                                                              -default_to_env => 1);

    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $table = 'bcomd_info';
    my $desired_state = 'IDLE';

    &ParamsMgr::process_args(%args);

    my $node_name = &ParamsMgr::get_arg(-arg => "node", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $node = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
    if ($self->error()) {
        $self->_set_err("Failed to get nacl node objects for: " . $node_name);
        return $EventUtil::FALSE;
    }

    my $bcomd_info_table;

    # This line will return the parsed output of the bcomd_info table
    # The APISET Layer was used instead of the Component layer.
    # debugsmdb support at the Component level is not functional.
    try {
        # burt780133: need to force CLI since there are no ZAPIs for debug smdb commands
        $node->apiset_must({interface => 'CLI'});
        $bcomd_info_table = $node->apiset()->debug_smdb_table_show(table => 'bcomd_info',
                                                                   node  => $node_name)->{parsed_output};

    } otherwise {
        my $e = shift;
        $self->_set_err("Exception " . ref($e) . " caught: " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # This will evaluate one hash holding values for the node that was passed in.
    # If the node state is something other than 'IDLE' a warning message is displyed
    # suggesting the user file a burt. The test will then continue.
    my $state;
    foreach my $hash (@{$bcomd_info_table}) {
        $state = $hash->{'init_state'};

        if($state eq $desired_state){
            ; # do nothing
        } else {
            if($state eq 'CTRAN_INIT') {
                $Log->warn("Node (" . $node->node . ") state is " .
                              $state . " instead of 'IDLE.' A burt should be filed " .
                              "with SW/kcs_ctran");
            } elsif ($state eq 'VS_JOIN_POLL') {
                $Log->warn("Node (" . $node->node . ") state is " .
                              $state . " instead of 'IDLE.' A burt should be filed " .
                              "with SW/san_bcomka");
            } else {
                $Log->warn("Node (" . $node->node . ") is in a state " .
                              "other than 'IDLE.' A burt should be filed " .
                              "with SW/san_bcom.");
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _wait_for_validate()
# Objective: Given a previous timeout, wait the minimum wait time needed
#            which is the difference between the previous time out and the
#            min timeout
###############################################################################
sub _wait_for_validate {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # the caller may have requested that we wait a minimum timeout
    # before performing lif validation
    my $last_timeout = &ParamsMgr::get_arg(-arg => "last_timeout",
                                           -default => 0);
    my $last_timeout_name = &ParamsMgr::get_arg(-arg => "last_timeout_name",
                                                -default => "timeout");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine the min wait time from the current event/eventset
    my $cv_min_wait_time = $self->EventUtil::get_arg(-arg => "controller_validation_min_wait_time",
                                                     -default => 0);

    my ($delta);

    # make sure we have slept AT LEAST cv_min_wait_time
    # seconds. If we have not, sleep the delta of cv_min_wait_time
    # and the passed in value.
    if ($last_timeout <= $cv_min_wait_time) {
        # calculate the additional amount of time we must sleep
        $delta = $cv_min_wait_time - $last_timeout;

        $Log->comment("Sleeping an additional ($delta) seconds to allow paths to settle\n" .
                      "before performing lif validation. This is in addition to the\n" .
                      "${last_timeout_name}=${last_timeout} seconds we already waited.");
        &SanUtil::nate_sleep($delta);
    } else {
        $Log->comment("No additional sleep time is needed since we have already waited " .
                      "${last_timeout_name}=${last_timeout} seconds\nwhich is greater than " .
                      "-controller_validation_min_wait_time ($cv_min_wait_time)");
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _is_steady_state_loaded()
# Objective: Determine if steady state is loaded on the specified SanClient
#            resource
###############################################################################
sub _is_steady_state_loaded {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource = &ParamsMgr::get_arg(-arg => "resource",
                                       -default => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $steady_state_loaded = $self->ResMgr::get_resource_attribute(-name => "$resource",
                                                                    -type => "_GLOBALS",
                                                                    -attribute => "steady_state_loaded");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # check if steady state is set on this client
    if (defined($steady_state_loaded) and $steady_state_loaded == $EventUtil::TRUE) {
        return $steady_state_loaded;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    check_for_coredump()
# Objective: check all nodes for coredumps
###############################################################################
sub check_for_coredump {
    my ($self, %args) = @_;

    my $coredump_validation_name_regex_filter = $self->EventUtil::get_arg(-arg => "coredump_validation_name_regex_filter");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    # get all the nodes that in this event's scope
    # and that are not in a faulted state
    my @node_objs = ();
    (@node_objs) = $self->EventUtil::get_nacl_node_objs(-type => "cmode",
                                                        -faulted => $EventUtil::FALSE);

    foreach my $node_obj (@node_objs) {
        $self->check_node_for_coredump(node            => $node_obj,
                                       coredump_validation_name_regex_filter => $coredump_validation_name_regex_filter);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    check_node_for_coredump()
# Objective: Check for coredump on a single Node
###############################################################################
sub check_node_for_coredump {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node = &ParamsMgr::get_arg(-arg      => "node",
                                   -required => 1);

    my $coredump_validation_name_regex_filter = &ParamsMgr::get_arg(-arg      => "coredump_validation_name_regex_filter",
                                                         -required => 0,
                                                         -default  => undef);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $node_name = $node->node();

    # get the names of the preexisting coredumps
    my @original_coredump_names = $self->ResMgr::get_resource_attribute(
                                              -name => "coredump_info",
                                              -type => "_GLOBALS",
                                              -attribute => "existing_coredumps");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    # check for coredump.
    my @current_coredump_names = $self->ControllerValidation::_get_coredump_names(nodes => [$node]);
    foreach my $current_coredump_name (@current_coredump_names) {
        if( not grep(/$current_coredump_name/, @original_coredump_names)) {
            # found new coredump!!!!
            # but, lets check if the name matches the filter
            if ((defined $coredump_validation_name_regex_filter) and
                    (not $current_coredump_name =~ /$coredump_validation_name_regex_filter/)) {
                $self->_set_err("Found new coredump $current_coredump_name on $node_name");
                last;
            }
        }
    }
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _get_coredump_names()
# Objective: Get the names of the coredumps saved on the nodes
###############################################################################
sub _get_coredump_names {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @node_objs = &ParamsMgr::get_arg(-arg      => "nodes",
                                        -required => 1);
    my $error_text = undef;
    my @systemnodecoredump_states;
    my @coredump_names;
    #grab the list of cores generated prior to the test.
    foreach my $node_obj (@node_objs) {
        try {
            @systemnodecoredump_states = NACL::CS::SystemNodeCoredump->fetch(command_interface => $node_obj,
                                                                             filter => {'node' => $node_obj->node()});
        } otherwise {
            my $exception = shift;
            $error_text = $exception->text();
        };
        if (defined $error_text) {
            if (not ($error_text =~ /No matching/)) {
                $self->_set_err($error_text);
            }
            return ();
        }
        foreach my $systemnodecoredump_state (@systemnodecoredump_states) {
            push (@coredump_names, $systemnodecoredump_state->corename());
        }
    }
    return @coredump_names;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    # define the default values for Event parameters
    $self->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                       -arg_defaults =>
                                                           { controller_validation => 1,
                                                             controller_validation_abort_on_error => 1,
                                                             controller_validation_before_takeover => 1,
                                                             controller_validation_on_takeover => 1,
                                                             controller_validation_on_giveback => 1,
                                                             controller_validation_min_wait_time => 0,
                                                             lif_validation => 1,
                                                             lif_validation_abort_on_error => 1,
                                                             lif_validation_lif_name_regex_filter => qr/.*/,
                                                             lif_validation_new_resource_behavior => "fatal",
                                                             net_port_validation => 1,
                                                             net_port_validation_abort_on_error => 1,
                                                             net_port_validation_port_name_regex_filter => qr/.*/,
                                                             fcp_port_validation => 1,
                                                             fcp_port_validation_abort_on_error => 1,
                                                             fcp_port_validation_port_name_regex_filter => qr/.*/,
                                                             cna_port_validation => 1,
                                                             cna_port_validation_abort_on_error => 1,
                                                             cna_port_validation_port_name_regex_filter => qr/.*/,
                                                             coredump_validation => 1,
                                                             coredump_validation_abort_on_error => 1,
                                                             coredump_validation_name_regex_filter => undef,
                                                             coredump_validation_delete_existing_coredumps => 0,
                                                             bcomd_bringup_info_validation => 1,
                                                             bcomd_bringup_info_validation_abort_on_error => 1,
                                                             bcomd_info_validation => 1,
                                                             bcomd_info_validation_abort_on_error => 0,
                                                             fc_gs_collection => 0
                                                           },
                                                       );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # determine if we are doing controller validation as a whole
    my $controller_validation_on_giveback = $self->EventUtil::get_arg(-arg => "controller_validation_on_giveback");
    my $controller_validation_on_takeover = $self->EventUtil::get_arg(-arg => "controller_validation_on_takeover");
    my $controller_validation             = $self->EventUtil::get_arg(-arg => "controller_validation");

    if ($controller_validation == 0) {
        $Log->comment("Skipping initialization of Controller validation since controller " .
                      "validation is disabled");
        return $EventUtil::TRUE;
    } elsif ($controller_validation_on_giveback == 0 and $controller_validation_on_takeover == 0) {
        $Log->comment("Skipping initialization of Controller validation since both controller " .
                      "validation on takeover and controller validation on giveback are disabled");
        return $EventUtil::TRUE;
    }

    # determine if lif validation is enabled in the current event/eventset
    my $lif_validation = $self->EventUtil::get_arg(-arg => "lif_validation");

    # Get a list of all nodes
    my @nacl_nodes_cm = $self->EventUtil::get_nacl_nodes(-faulted                          => $EventUtil::FALSE,
                                                         -fall_back_on_controller_clusters => $EventUtil::TRUE,
                                                         -type                             => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # If lif validation is not enabled, then don't bother continuing
    if ($lif_validation == 1) {
        # build the lif hash
        my $lif_hash = $self->ControllerValidation::_get_lif_info(-reconnect => 0,
                                                                  -nacl_nodes_cm => \@nacl_nodes_cm);
        if ($self->error()) {
            $self->_set_err("Unable to build the lif hash");
            return $EventUtil::FALSE;
        }

        # Before adding the lif_hash to Resource Manager, check to see if there are any
        # lifs that are down.
        # LIF Info Hash Structure:
        #
        #            $lif_hash =>
        #                    <node_name> =>
        #                       <vserver_name> =>
        #                           <lif_name> =>
        #                                current_node  => Node where the lif currently resides
        #                                home_port     => Name of home port
        #                                data_protocol => Data protocol of lif
        #                                status_oper   => Operational status of lif
        #                                status_admin  => Admin status of lif
        my @errors_found = ();
        foreach my $node (keys %$lif_hash) {
            foreach my $vserver ( keys %{$lif_hash->{$node}}) {
                foreach my $lif_name ( keys %{$lif_hash->{$node}->{$vserver}}) {
                    if (lc($lif_hash->{$node}->{$vserver}->{$lif_name}->{'status_oper'}) eq "down") {
                        my $errmsg = sprintf("%-20s %-20s %-20s %-10s", $node, $vserver, $lif_name, "Oper");
                        push(@errors_found, $errmsg);
                    }
                    if (lc($lif_hash->{$node}->{$vserver}->{$lif_name}->{'status_admin'}) eq "down") {
                        my $errmsg = sprintf("%-20s %-20s %-20s %-10s", $node, $vserver, $lif_name, "Admin");
                        push(@errors_found, $errmsg);
                    }
                }
            }
        }

        if (scalar(@errors_found)) {
            unshift(@errors_found, sprintf("%-20s %-20s %-20s %-10s", "====================", "====================",
                                           "====================", "=========="));
            unshift(@errors_found, sprintf("%-20s %-20s %-20s %-10s", "Node Name", "VServer", "LIF", "ADMIN/OPER"));
            unshift(@errors_found, "The following LIFs were found down and not excluded from the test:");
            push (@errors_found, "\n");
            my $errmsg = join("\n", @errors_found);
            $Log->warn($errmsg);
            $self->_set_err("Initialization of LIF validation failed.");
            return $EventUtil::FALSE;
        }

        # register the lif_hash datastructure as a "_GLOBALS" resource type
        # this will allow events to access this datastructure and make persistent modifications to it
        $self->ResMgr::add_resource(-name => "lif_info",
                                    -type => "_GLOBALS",
                                    -obj  => $lif_hash);
        if ($self->error()) {
            $self->_set_err("Unable to add resource 'lif_info' to the Resource Manager");
            return $EventUtil::FALSE;
        }

        $self->ResMgr::set_resource_attribute(-name => "lif_info",
                                              -type => "_GLOBALS",
                                              -attribute => "steady_state_loaded",
                                              -value => $EventUtil::TRUE,
                                              -propagate => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Unable to set resource  attribute 'steady_state_loaded' on resource 'lif_info' " .
                            "for type '_GLOBALS'");
            return $EventUtil::FALSE;
        }
    } else {
        $Log->comment("Skipping initialization of LIF Validation since it is disabled.");
    }

    # build the network port hash as long as network port validation is not disabled
    my $net_validation = $self->EventUtil::get_arg(-arg => "net_port_validation");
    if ($net_validation == 1) {
        my $network_port_hash = $self->ControllerValidation::_get_network_port_info(-reconnect => 0,
                                                                                    -nacl_nodes_cm => \@nacl_nodes_cm);
        if ($self->error()) {
            $self->_set_err("Unable to build the network port hash");
            return $EventUtil::FALSE;
        }

        # register the network_port_hash datastructure as a "_GLOBALS" resource type
        # this will allow events to access this datastructure and make persistent modifications to it
        $self->ResMgr::add_resource(-name => "network_port_info",
                                    -type => "_GLOBALS",
                                    -obj  => $network_port_hash);
        if ($self->error()) {
            $self->_set_err("Unable to add resource 'network_port_info' to the Resource Manager");
            return $EventUtil::FALSE;
        }

        $self->ResMgr::set_resource_attribute(-name => "network_port_info",
                                              -type => "_GLOBALS",
                                              -attribute => "steady_state_loaded",
                                              -value => $EventUtil::TRUE,
                                              -propagate => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Unable to set resource  attribute 'steady_state_loaded' on resource 'network_port_info' " .
                            "for type '_GLOBALS'");
            return $EventUtil::FALSE;
        }
    } else {
        $Log->comment("Skipping initialization of Network Port Validation since it is disabled.");
    }

    # build the fcp adapter hash
    my $fcp_validation = $self->EventUtil::get_arg(-arg => "fcp_port_validation");
    if ($fcp_validation == 1) {
        my $fcp_port_hash = $self->ControllerValidation::_get_fcp_port_info(-reconnect => 0,
                                                                            -nacl_nodes_cm => \@nacl_nodes_cm);
        if ($self->error()) {
            $self->_set_err("Unable to build the fcp adapter hash");
            return $EventUtil::FALSE;
        }

        # register the fcp_port_info datastructure as a "_GLOBALS" resource type
        # this will allow events to access this datastructure and make persistent modifications to it
        $self->ResMgr::add_resource(-name => "fcp_port_info",
                                    -type => "_GLOBALS",
                                    -obj  => $fcp_port_hash);
        if ($self->error()) {
            $self->_set_err("Unable to add resource 'fcp_port_info' to the Resource Manager");
            return $EventUtil::FALSE;
        }

        $self->ResMgr::set_resource_attribute(-name => "fcp_port_info",
                                              -type => "_GLOBALS",
                                              -attribute => "steady_state_loaded",
                                              -value => $EventUtil::TRUE,
                                              -propagate => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Unable to set resource  attribute 'steady_state_loaded' on resource 'fcp_port_info' " .
                            "for type '_GLOBALS'");
            return $EventUtil::FALSE;
        }
    } else {
        $Log->comment("Skipping initialization of FCP Port Validation since it is disabled.");
    }

    # build the cna adapter hash
    my $cna_validation = $self->EventUtil::get_arg(-arg => "cna_port_validation");
    if ($cna_validation == 1) {
        my $cna_port_hash = $self->ControllerValidation::_get_cna_port_info(-reconnect => 0,
                                                                            -nacl_nodes_cm => \@nacl_nodes_cm);
        if ($self->error()) {
            $self->_set_err("Unable to build the cna port hash");
            return $EventUtil::FALSE;
        }

        # register the cna_port_info datastructure as a "_GLOBALS" resource type
        # this will allow events to access this datastructure and make persistent modifications to it
        $self->ResMgr::add_resource(-name => "cna_port_info",
                                    -type => "_GLOBALS",
                                    -obj  => $cna_port_hash);
        if ($self->error()) {
            $self->_set_err("Unable to add resource 'cna_port_info' to the Resource Manager");
            return $EventUtil::FALSE;
        }

        $self->ResMgr::set_resource_attribute(-name => "cna_port_info",
                                              -type => "_GLOBALS",
                                              -attribute => "steady_state_loaded",
                                              -value => $EventUtil::TRUE,
                                              -propagate => $EventUtil::FALSE);
        if ($self->error()) {
            $self->_set_err("Unable to set resource  attribute 'steady_state_loaded' on resource 'cna_port_info' " .
                            "for type '_GLOBALS'");
            return $EventUtil::FALSE;
        }
    } else {
        $Log->comment("Skipping initialization of CNA Port Validation since it is disabled.");
    }

    my $coredump_validation = $self->EventUtil::get_arg(-arg => "coredump_validation");
    if ($coredump_validation) {
        my $coredump_validation_delete_existing_coredumps =
            $self->EventUtil::get_arg(-arg => "coredump_validation_delete_existing_coredumps");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        my (@node_objs) = $self->EventUtil::get_nacl_node_objs(-type => "cmode",
                                                               -faulted => $EventUtil::FALSE);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        my $delete_error = undef;
        foreach my $nacl_node_obj (@node_objs) {
            my $node_name = $nacl_node_obj->node();
            if ($coredump_validation_delete_existing_coredumps) {
                try {
                    NACL::C::SystemNodeCoredump->delete_all(
                                        command_interface => $nacl_node_obj,
                                        'node'            => $node_name,
                                        'type'            => 'all');
                } otherwise {
                    my $exception = shift;
                    $delete_error = $exception->text();
                };
                if (defined $delete_error){
                    $self->set_err($delete_error);
                    return $EventUtil::FALSE;
                }
            }
        }

        # register the coredump_info datastructure as a "_GLOBALS" resource type
        # this will allow events to access this datastructure and make persistent modifications to it
        my $coredump_info_hash = {};
        $self->ResMgr::add_resource(-name => "coredump_info",
                                    -type => "_GLOBALS",
                                    -obj  => $coredump_info_hash);
        if ($self->error()) {
            $self->_set_err("Unable to add resource 'coredump_info' to the Resource Manager");
            return $EventUtil::FALSE;
        }
        my @systemnodecoredump_names = $self->ControllerValidation::_get_coredump_names(nodes => \@node_objs);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => "coredump_info",
                                              -type => "_GLOBALS",
                                              -attribute => "existing_coredumps",
                                              -value => \@systemnodecoredump_names,
                                              -propagate => $EventUtil::TRUE);
        if ($self->error()) {
            $self->_set_err("Unable to set resource attribute 'systemnodecoredump_names'");
            return $EventUtil::FALSE;
        }
    }
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    # document each parameter used in this module
    $self->EventUtil::document_event_param(-param => "controller_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-controller_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable controller validation " .
                                               "on all CMode Nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"],
                                           -overrides => ["controller_validation_on_giveback", "controller_validation_on_takeover"]);

    $self->EventUtil::document_event_param(-param => "controller_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-controller_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if a controller " .
                                                           "validation error is detected during intial setup or " .
                                                           "during validation",
                                           -resource_types => ["nacl_nodes_cm"],
                                           -overrides => ["lif_validation_abort_on_error",
                                                          "net_port_validation_abort_on_error",
                                                          "fcp_port_validation_abort_on_error",
                                                          "cna_port_validation_abort_on_error"]);

    $self->EventUtil::document_event_param(-param => "controller_validation_before_takeover",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-controller_validation_before_takeover => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable controller validation before takeover");

    $self->EventUtil::document_event_param(-param => "controller_validation_on_takeover",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-controller_validation_on_takeover => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable controller validation during takeover");

    $self->EventUtil::document_event_param(-param => "controller_validation_on_giveback",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-controller_validation_on_giveback => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable controller validation after giveback");

    $self->EventUtil::document_event_param(-param => "controller_validation_min_wait_time",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "integer, in seconds",
                                           -example => "-controller_validation_min_wait_time => 300",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the minimum amount of time (seconds) that must be waited " .
                                                           "before performing lif validation");

    $self->EventUtil::document_event_param(-param => "lif_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-lif_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable lif validation " .
                                               "on all CMode Nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "lif_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-lif_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if a lif validation error is " .
                                               "detected during intial setup or during validation",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "lif_validation_lif_name_regex_filter",
                                           -package => __PACKAGE__,
                                           -default => "qr/.*/",
                                           -format => "regular expression",
                                           -example => "-lif_validation_lif_name_regex_filter => qr/^(?!.*dummy).*$/",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies regex filter to determine what lifs to validate");

    $self->EventUtil::document_event_param(-param => "lif_validation_new_resource_behavior",
                                           -package => __PACKAGE__,
                                           -default => "fatal",
                                           -format => "string of fatal, warning, or ignore",
                                           -example => "-lif_validation_new_resource_behavior => 'warning'",
                                           -per_resource_capable => "0",
                                           -environment_capable => "0",
                                           -description => "Specifies how to handle when a new lif is found during " .
                                                           "validation");

    $self->EventUtil::document_event_param(-param => "net_port_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-net_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable network port validation " .
                                                           "on all CMode Nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "net_port_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-net_port_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if a network port " .
                                                           "validation error is detected during intial setup " .
                                                           "or during validation",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "net_port_validation_port_name_regex_filter",
                                           -package => __PACKAGE__,
                                           -default => "qr/.*/",
                                           -format => "regular expression",
                                           -example => "-net_port_validation_port_name_regex_filter => qr/^(?!.*dummy).*$/",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies regex filter to determine what network ports " .
                                                           "to validate");

    $self->EventUtil::document_event_param(-param => "fcp_port_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-fcp_port_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable fcp port validation " .
                                                            "on all CMode Nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "fcp_port_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-fcp_port_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if an fcp port validation " .
                                                           "error is detected during intial setup or during validation",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "fcp_port_validation_port_name_regex_filter",
                                           -package => __PACKAGE__,
                                           -default => "qr/.*/",
                                           -format => "regular expression",
                                           -example => "-fcp_port_validation_port_name_regex_filter => qr/^(?!.*dummy).*$/",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies regex filter to determine what fcp ports " .
                                                           "to validate");

    $self->EventUtil::document_event_param(-param => "cna_port_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-cna_port_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable cna port validation " .
                                                            "on all CMode nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "cna_port_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-cna_port_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if an cna port validation " .
                                                           "error is detected during intial setup or during validation",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "cna_port_validation_port_name_regex_filter",
                                           -package => __PACKAGE__,
                                           -default => "qr/.*/",
                                           -format => "regular expression",
                                           -example => "-cna_port_validation_port_name_regex_filter => qr/^(?!.*dummy).*$/",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies regex filter to determine what cna ports " .
                                                           "to validate");

    $self->EventUtil::document_event_param(-param => "bcomd_bringup_info_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-bcomd_bringup_info_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable bcomd_bringup_info_validation " .
                                                            "on all CMode Nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"]);

    $self->EventUtil::document_event_param(-param => "bcomd_bringup_info_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-bcomd_bringup_info_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if an" .
                                                           "bcomd_bringup_info_validation error is " .
                                                           "detected during intial setup or during validation.",
                                           -resource_types => ["nacl_nodes_cm"]);
    $self->EventUtil::document_event_param(-param => "bcomd_info_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-bcomd_info_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable bcomd_bringup_info_validation " .
                                                            "on all CMode Nodes that support the methods to do so.",
                                           -resource_types => ["nacl_nodes_cm"]);
    $self->EventUtil::document_event_param(-param => "bcomd_info_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-bcomd_info_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if an" .
                                                           "bcomd_info_validation error is " .
                                                           "detected during intial setup or during validation",
                                           -resource_types => ["nacl_nodes_cm"]);

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

ControllerValidation.pm

=head1 Author

Derrick Shultz

=head1 Description

An EventMan internal utility module designed to provide helper
methods to perform Controller Validation for CMode Nodes
during takeover / giveback and port block scenarios.  The validation
includes LIF, network port, fcp adapter, cna adapter, and coredumps.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-controller_validation>

=over 4

Description: If set to 1, will enable controller validation
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-controller_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a controller validation error is
detected during intial setup or during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-controller_validation_before_takeover>

=over 4

Description: If set to 1, will enable controller validation before takeover
on healthy Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Default: 1 (on)

Per Resource Capable: No

=back

=head2 I<-controller_validation_on_takeover>

=over 4

Description: If set to 1, will enable controller validation during takeover
on surviving Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Default: 1 (on)

Per Resource Capable: No

=back

=head2 I<-controller_validation_on_giveback>

=over 4

Description: If set to 1, will enable controllerlif validation after giveback
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-controller_validation_min_wait_time>

=over 4

Description: Specifies the minimum amount of time (seconds) that
must be waited before performing lif validation

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-lif_validation>

=over 4

Description: If set to 1, will enable lif validation
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-lif_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a lif validation error is
detected during intial setup or during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-lif_validation_new_resource_behavior>

=over 4

Description: Describes how to behave if new LIFs are discovered during the validation
phase that didn't exist during the steady state discovery phase.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: "fatal"

Valid Values: "fatal", "warning", "ignore"

=back

=head2 I<-lif_validation_lif_name_regex_filter>

=over 4

Description: Specifies the regex filter for the list of lifs to validate.

Per Resource Capable: No

Default: qr/.*/

Regex Examples:

    Do not validate lifs containing the word "dummy"

        qr/^(?!.*dummy).*$/

    Validate lifs starting with "fc" or "iscsi"

        qr/^(fc|iscsi).*$/

=back

=head2 I<-net_port_validation>

=over 4

Description: If set to 1, will enable network port validation
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-net_port_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a network port validation error is
detected during intial setup or during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-net_port_validation_port_name_regex_filter>

=over 4

Description: Specifies the regex filter for the list of network ports to validate.

Per Resource Capable: No

Default: qr/.*/

Regex Examples:

    Do not validate lifs containing the word "dummy"

        qr/^(?!.*dummy).*$/

    Validate ports e1a and e4b

        qr/^(e1a|e4b)$/

=back

=head2 I<-fcp_port_validation>

=over 4

Description: If set to 1, will enable fcp port validation
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-fcp_port_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if an fcp port validation error is
detected during intial setup or during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-fcp_port_validation_port_name_regex_filter>

=over 4

Description: Specifies the regex filter for the list of fcp ports to validate.

Per Resource Capable: No

Default: qr/.*/

Regex Examples:

    Do not validate lifs containing the word "dummy"

        qr/^(?!.*dummy).*$/

    Validate ports 0a and 4b

        qr/^(0a|4b)$/

=back

=head2 I<-cna_port_validation>

=over 4

Description: If set to 1, will enable cna port validation
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-cna_port_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a cna port validation error is
detected during intial setup or during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-cna_port_validation_port_name_regex_filter>

=over 4

Description: Specifies the regex filter for the list of cna ports to validate.

Per Resource Capable: No

Default: qr/.*/

Regex Examples:

    Do not validate lifs containing the word "dummy"

        qr/^(?!.*dummy).*$/

    Validate ports 1a and 1b

        qr/^(1a|1b)$/

=back

=head2 I<-fcp_port_validation>

=over 4

Description: If set to 1, will enable fcp port validation
on all Nodes that support the methods to do so.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-coredump_validation>

=over 4

Description: If set to 1, perform check to see that there are no new coredumps since starting
 the test.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-coredump_validation_abort_on_error>

=over 4

Description: If set to 1, will abort the test if a new coredump is discovered.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-coredump_validation_name_regex_filter>

=over 4

Description: If set to a regex, disregard coredumps with names that match the regex when
checking for new coredumps.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: undef

=back

=head2 I<-coredump_validation_delete_existing_coredumps>

=over 4

Description: If set to 1, delete all existing coredumps from the cluster nodes.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1

=back

=head2 I<-bcomd_bringup_info_validation>

=over 4

Description: If set to 1, perform check to see that the information in the
bcomd_bringup_info table matches the desired results. Current desired results
are latency per node should be < 5 mins.

Per Resource Capable: No

Default: 1

=back

=head2 I<-bcomd_bringup_info_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a bcomd_bringup_info error is
detected during intial setup or during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 1

=back


=head2 I<-bcomd_info_validation>

=over 4

Description: If set to 1, perform check to verify the state of the node is 'IDLE.'

Per Resource Capable: No

Default: 1

=back

=head2 I<-bcomd_info_abort_on_error>

=over 4

Description: If set to 1, will abort events if a bcomd_info error is
detected during validation.

Resource Types: nacl_nodes_cm

Per Resource Capable: No

Default: 0 

=back

=head2 I<-fc_gs_collection>

=over 4

Description: If set to 1, will enable FC_GS command collection on ontap controllers

Per Resource Capable: No

Default: 0 (off)

=back

=head1 B<Public Methods>

=head2 B<validate_controller_before_takeover>

Performs controller validation before takeover for all NodeCM resources in the
current Event/EventSet/EventMan resource scope.

=over 4

=item Inputs

=item *I<-nacl_nodes_cm>

Optional: A reference to an array containing a list of nacl_nodes_cm resources to check.
Default: Empty array reference

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -controller_validation
 -controller_validation_abort_on_error
 -controller_validation_before_takeover
 -bcomd_info_validation
 -bcomd_info_validation_abort_on_error

=item Examples

 $sanfaults_obj->ControllerValidation::validate_controller_before_takeover();
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<validate_controller_after_takeover()>

Performs controller validation for all NodeCM resources in the current
Event/EventSet/EventMan resource scope assuming controllers are in a
takeover state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off lif validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do lif validation. This
will be used to calculate additional lif validation wait time that needs
to be applied before performing lif validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: "timeout"

=item *I<-nacl_nodes_cm>

Optional: A reference to an array containing a list of nacl_nodes_cm resources to check.
Default: Empty array reference

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -controller_validation
 -controller_validation_abort_on_error
 -controller_validation_on_takeover
 -controller_validation_min_wait_time
 -lif_validation
 -lif_validation_abort_on_error
 -lif_validation_new_resource_behavior
 -lif_validation_lif_name_regex_filter
 -net_port_validation
 -net_port_validation_abort_on_error
 -net_port_validation_port_name_regex_filter
 -fcp_port_validation
 -fcp_port_validation_abort_on_error
 -fcp_port_validation_port_name_regex_filter
 -cna_port_validation
 -cna_port_validation_abort_on_error
 -cna_port_validation_port_name_regex_filter
 -coredump_validation
 -coredump_validation_abort_on_error
 -coredump_validation_name_regex_filter
 -coredump_validation_delete_existing_coredumps

=item Examples

 $sanfaults_obj->ControllerValidation::validate_controller_after_takeover(-id_extension => "takeover_on_node_A",
                                                                          -last_timeout => "600",
                                                                          -last_timeout_name => "filer_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<validate_controller_after_giveback()>

Performs controller validation for all NodeCM resources in the current
Event/EventSet/EventMan resource scope assuming controllers are in a
giveback state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off lif validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do lif validation. This
will be used to calculate additional lif validation wait time that needs
to be applied before performing lif validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: ""

=item *I<-nacl_nodes_cm>

Optional: A reference to an array containing a list of nacl_nodes_cm resources to check.
Default: Empty array reference

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -controller_validation
 -controller_validation_abort_on_error
 -controller_validation_on_giveback
 -controller_validation_min_wait_time
 -lif_validation
 -lif_validation_abort_on_error
 -lif_validation_new_resource_behavior
 -lif_validation_lif_name_regex_filter
 -net_port_validation
 -net_port_validation_abort_on_error
 -net_port_validation_port_name_regex_filter
 -fcp_port_validation
 -fcp_port_validation_abort_on_error
 -fcp_port_validation_port_name_regex_filter
 -cna_port_validation
 -cna_port_validation_abort_on_error
 -bcomd_bringup_info_validation
 -bcomd_bringup_info_validation_abort_on_error

=item Examples

 $sanfaults_obj->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_node_A",
                                                                          -last_timeout => "600",
                                                                          -last_timeout_name => "controller_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<collect_fc_gs_information()>

Runs FC_GS commands on the ontap controllers in the current event scope.

=over 4

=item Inputs

None

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -fc_gs_collection

=item Examples

 $em_obj->ControllerValidation::collect_fc_gs_information();
 if ($em_obj->error()) {
     ...
 }

=back

=cut
