# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A general purpose internal utility class for the EventMan package
## @author hrivera@netapp.com
## @status private
## @pod here

package EventUtil;

# San packages
use San::SanUtil;
use San::FabricMgr;
use ParamsMgr;
use Data::Dumper;
use Tharn;
use NACL::C::Node;
use NACL::C::Lun;
use NACL::C::NetworkInterface;
use NACL::C::Volume;
use NACL::C::NetworkPortIfgrp;
use NACL::C::NetworkPortVlan;
use NACL::C::Vserver;
use NACL::C::LunIgroup;;
use NATE::BaseException qw (:try);
use strict;
use warnings;

# Exported Globals
$EventUtil::TRUE = 1;
$EventUtil::FALSE = 0;
$EventUtil::SUBTEST_EXIT_SLEEP_TIME = 10;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    get_arg()
# Objective: Obtain the specified Event Parameter value given a list of
#            implicit or explicit -objects to search through.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_arg {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the objects to search through in this order:
    # 1: hardcoded event params from current event & current event module,
    # 2: current event obj
    # 3: current event set obj,
    # 4: and eventman obj
    my (@objects) = &ParamsMgr::get_arg(-arg => "objects",
                                        -ref_type => "ARRAY",
                                        -default => [$self->{_current_event}->{event_module_hardcoded_args},
                                                     $self->{_current_event}->{event_obj},
                                                     $self->{_current_event}->{event_set_obj},
                                                     $self]);
    my $arg = &ParamsMgr::get_arg(-arg => "arg",
                                  -required => 1);
    my $default = &ParamsMgr::get_arg(-arg => "default",
                                      -default => undef);
    my $default_to_env = &ParamsMgr::get_arg(-arg => "default_to_env",
                                             -default => $EventUtil::FALSE);
    my $no_event_module_defaults = &ParamsMgr::get_arg(-arg => "no_event_module_defaults",
                                                       -default => $EventUtil::FALSE);
    my $no_utility_module_defaults = &ParamsMgr::get_arg(-arg => "no_utility_module_defaults",
                                                         -default => $EventUtil::FALSE);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return undef;
    }

    # search through all the objects in order.
    # return the first instance of the arg that we find.
    foreach my $obj (@objects) {
        if (defined($obj) and exists $obj->{args}) {
            my %args_tmp = &SanUtil::strip_dash(%{$obj->{args}});

            if (exists($args_tmp{$arg})) {
                # evaluate the value of the arg that we found if it is a subroutine reference
                if (ref($args_tmp{$arg}) eq "CODE") {
                    return $args_tmp{$arg}->($self);
                } else {
                    # otherwise just return the value as it is
                    return $args_tmp{$arg};
                }
            }
        }
    }

    # implicitly using the current event/eventset/eventman environment
    # if still no value is found, then $default will be selected as the value
    if ($default_to_env == 1) {
        return $self->EventUtil::get_arg(-arg => $arg,
                                         -default => $default);
    }

    # if there is a default event module loaded then look through its event
    # parameter default value.
    if (defined($self->{_current_event}->{event_module_default_args}) and $no_event_module_defaults == $EventUtil::FALSE) {
        # if at this point the argument still can't be found, then
        # look for the argument in the current event module's defaults list,
        # but dont use recursion if it can't be found, otherwise we will go into an endless
        # loop
        return $self->EventUtil::get_arg(-arg => $arg,
                                         -objects => [$self->{_current_event}->{event_module_default_args}],
                                         -default => $default,
                                         -no_event_module_defaults => $EventUtil::TRUE,
                                         -no_utility_module_defaults => $no_utility_module_defaults);
    }

    # if there is still no value found, then look through the default event
    # parameters in the Utility modules
    if (defined($self->{_utility_module_default_args}) and $no_utility_module_defaults == $EventUtil::FALSE) {
        # disable lookup on both event modules and utility modules or else we will
        # go into an endless loop.
        return $self->EventUtil::get_arg(-arg => $arg,
                                         -objects => [$self->{_utility_module_default_args}],
                                         -default => $default,
                                         -no_event_module_defaults => $EventUtil::TRUE,
                                         -no_utility_module_defaults => $EventUtil::TRUE);
    }

    # if at this point the parameter was not found then return the default value that was
    # explicitly passed into the get_arg() method by the caller
    if (ref($default) eq "CODE") {
        return $default->($self);
    } else {
        return $default;
    }
}

###############################################################################
# Method:    get_current_event_obj()
# Objective: Obtain the event object of the current running event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_event_obj {
    my ($self) = @_;

    return $self->{_current_event}->{event_obj};
}

###############################################################################
# Method:    get_current_eventset_obj()
# Objective: Obtain the eventset object of the current running eventset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_eventset_obj {
    my ($self) = @_;

    return $self->{_current_event}->{eventset_obj};
}

###############################################################################
# Method:    display_event_parameters()
# Objective: Given a list of objects to work with, build a list of all Event
#            Parameters and their respective values.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_event_parameters {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the object event params to display to the EventMan object
    my (@objects) = &ParamsMgr::get_arg(-arg => "objects",
                                        -ref_type => "ARRAY",
                                        -default => [$self]);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # need to reverse this list when we call get_current_resources off of ResMgr
    # because the precendence order for object lookup is reveresed
    my @reversed_objects = reverse @objects;

    # get current resources that are in use by the current Event/EventSet/EventMan environment
    # this is returned as a hash where the key is the resource type
    # and the value is an array reference of the resource names of that type
    my $resources = $self->ResMgr::get_current_resources(-objects => \@reversed_objects);
    if ($self->error()) {
        $self->_set_err("Unable to get current resource list");
        return (wantarray()) ? () : [];
    }

    my @summary;

    # display the resources that are being used in this scope
    push(@summary, "\tResources:");
    if (scalar(keys %{$resources}) > 0) {
        foreach my $arg (sort keys %{$resources}) {
            if (ref($resources->{$arg}) eq "ARRAY") {
                # handle array references
                push(@summary, "\t\t* $arg : " . join(", ", @{$resources->{$arg}}));
            } elsif (ref($resources->{$arg}) eq "CODE") {
                # handle code values
                push(@summary, "\t\t* $arg : " . $resources->{$arg}->($self));
            } elsif (ref($resources->{$arg}) eq "HASH") {
                my $tmp = Dumper($resources->{$arg});
                $tmp =~ s/\$VAR\d\s*=\s+//;
                $tmp =~ s/\n/ /g;
                $tmp =~ s/\s+/ /g;
                push(@summary, "\t\t* $arg : " . $tmp);
            } else {
                # handle strings
                if (length($resources->{$arg})) {
                    push(@summary, "\t\t* $arg : " . $resources->{$arg});
                } else {
                    # handle empty strings
                    push(@summary, "\t\t* $arg : \"\"");
                }
            }
        }
    } else {
        push(@summary, "\t\t* (none)");
    }

    push(@summary, "");

    # get current sub resources that are in use by the current Event/EventSet/EventMan environment
    # this is returned as a hash where the key is the resource type
    # and the value is an array reference of the resource names of that type
    my $sub_resources = $self->ResMgr::get_current_sub_resources(-objects => \@reversed_objects);
    if ($self->error()) {
        $self->_set_err("Unable to get current sub resource list");
        return (wantarray()) ? () : [];
    }

    # display the sub_resources that are being used in this scope
    push(@summary, "\tSub Resources:");
    if (scalar(keys %{$sub_resources}) > 0) {
        foreach my $arg (sort keys %{$sub_resources}) {
            if (ref($sub_resources->{$arg}) eq "ARRAY") {
                # handle array references
                push(@summary, "\t\t* $arg : " . join(", ", @{$sub_resources->{$arg}}));
            } elsif (ref($sub_resources->{$arg}) eq "CODE") {
                # handle code values
                push(@summary, "\t\t* $arg : " . $sub_resources->{$arg}->($self));
            } elsif (ref($sub_resources->{$arg}) eq "HASH") {
                my $tmp = Dumper($sub_resources->{$arg});
                $tmp =~ s/\$VAR\d\s*=\s+//;
                $tmp =~ s/\n/ /g;
                $tmp =~ s/\s+/ /g;
                push(@summary, "\t\t* $arg : " . $tmp);
            } else {
                # handle strings
                if (length($sub_resources->{$arg})) {
                    push(@summary, "\t\t* $arg : " . $sub_resources->{$arg});
                } else {
                    # handle empty strings
                    push(@summary, "\t\t* $arg : \"\"");
                }
            }
        }
    } else {
        push(@summary, "\t\t* (none)");
    }

    push(@summary, "");

    # if hardcoded event module event parameters are stored, push it to
    # the end of the list (highest precendence)
    if (defined($self->{_current_event} and defined($self->{_current_event}->{event_module_hardcoded_args}))) {
        push(@objects, $self->{_current_event}->{event_module_hardcoded_args});
    }

    # if default event module event parameters are stored, unshift it to
    # the front of the list
    if (defined($self->{_current_event} and defined($self->{_current_event}->{event_module_default_args}))) {
        unshift(@objects, $self->{_current_event}->{event_module_default_args});
    }

    # if default utility module event parameters are stored, unshift it to
    # the front of the list
    if (defined($self->{_utility_module_default_args})) {
        unshift(@objects, $self->{_utility_module_default_args});
    }

    # load in all the Event Parameters into the event_params hash, allow
    # overwrites of duplicated entries, this means that elements at the
    # end of @objects have more precendence over elements at the beginning
    # of the list.
    my $event_params = {};
    foreach my $obj (@objects) {
        if (exists($obj->{args})) {
            foreach my $arg (keys %{$obj->{args}}) {
                my $value = $obj->{args}->{$arg};
                $event_params->{$arg} = $value;
            }
        }
    }

    # delete all resource parameters from the event_params hash so that they dont
    # appear twice
    foreach my $resource_type (keys %{$resources}) {
        delete($event_params->{$resource_type});
    }

    # display the parameters in Event params with their specified value
    push(@summary, "\tEvent Parameters:");
    if (scalar(keys %{$event_params}) > 0) {
        foreach my $arg (sort keys %{$event_params}) {
            if (ref($event_params->{$arg}) eq "ARRAY") {
                # handle array references
                push(@summary, "\t\t* $arg : " . join(", ", @{$event_params->{$arg}}));
            } elsif (ref($event_params->{$arg}) eq "CODE") {
                # handle code values
                push(@summary, "\t\t* $arg : " . $event_params->{$arg}->($self));
            } elsif (ref($event_params->{$arg}) eq "HASH") {
                my $tmp = Dumper($event_params->{$arg});
                $tmp =~ s/\$VAR\d\s*=\s+//;
                $tmp =~ s/\n/ /g;
                $tmp =~ s/\s+/ /g;
                push(@summary, "\t\t* $arg : " . $tmp);
            } else {
                # handle strings
                if (defined($event_params->{$arg}) and length($event_params->{$arg})) {
                    push(@summary, "\t\t* $arg : " . $event_params->{$arg});
                } else {
                    # handle empty strings
                    push(@summary, "\t\t* $arg : \"\"");
                }
            }
        }
    } else {
        push(@summary, "\t\t* (none)");
    }

    return (wantarray()) ? @summary : \@summary;
}

###############################################################################
# Method:    get_sanclients()
# Objective: Obtain the list of SanClient Resource names in use by the current
#            Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanclients {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_state = &ParamsMgr::get_arg(-arg => "faulted",
                                          -default => -1);

    my $protocol = &ParamsMgr::get_arg(-arg => "protocol",
                                       -default => undef);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # determine which sanclients to operate on, obtain the list from the current
    # event/eventset
    my $sanclients = $self->EventUtil::get_arg(-arg => "sanclients");

    my @sanclients = ();
    if (ref($sanclients) eq "ARRAY" and scalar(@{$sanclients})) {
        @sanclients = @{$sanclients};
    } elsif (ref($sanclients) eq "HASH") {
        # then allow the value to be null because the user explicitly set it that way
    } else {
        # default to the full list of clients otherwise
        @sanclients = $self->ResMgr::get_resource_names_by_type(-type => "sanclients");
    }

    # if there are no sanclients just return
    if (not scalar(@sanclients)) {
        return (wantarray()) ? () : [];
    }

    # if requested to filter by event state, do that now
    if ($event_state != -1) {
        # filter through the sanclients that match the specified event state
        @sanclients = $self->ResMgr::filter_by_attribute(-types => ["sanclients"],
                                                         -names => \@sanclients,
                                                         -attribute => "faulted",
                                                         -value => $event_state);
        if ($self->error()) {
            $self->_set_err("Unable to filter resources by attribute faulted => '$event_state'");
            return (wantarray()) ? () : [];
        }
    }

    # if requested to filter by protocol, do so now
    if (defined($protocol)) {
        my @tmp_sanclients;
        foreach my $sanclient (@sanclients) {
            my $sanclient_obj = $self->EventUtil::get_sanclient_obj(-client_name => $sanclient);
            if ($self->error()) {
                $self->_set_err("Unable to obtain SanClient object for client '$sanclient'");
                return (wantarray()) ? () : [];
            }

            if ($sanclient_obj->get_protocol() =~ $protocol) {
                push(@tmp_sanclients, $sanclient);
            }
        }

        # overwrite sanclients with the protocol filtered version
        @sanclients = @tmp_sanclients;
    }

    return (wantarray()) ? @sanclients : \@sanclients;
}

###############################################################################
# Method:    get_sanswitches()
# Objective: Obtain the list of SanSwitch Resource names in use by the current
#            Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanswitches {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $event_state = &ParamsMgr::get_arg(-arg => "faulted",
                                          -default => -1);

    my @fabrics_filter = &ParamsMgr::get_arg(-arg => "fabrics",
                                             -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # determine which sanclients to operate on, obtain the list from the current
    # event/eventset
    my $sanswitches = $self->EventUtil::get_arg(-arg => "sanswitches");

    my @sanswitches = ();
    if (ref($sanswitches) eq "ARRAY" and scalar(@{$sanswitches})) {
        @sanswitches = @{$sanswitches};
    } elsif (ref($sanswitches) eq "HASH") {
        # then allow the value to be null because the user explicitly set it that way
    } else {
        # default to the full list of switches otherwise
        @sanswitches = $self->ResMgr::get_resource_names_by_type(-type => "sanswitches");
    }

    # if there are no sanswitches just return
    if (not scalar(@sanswitches)) {
        return (wantarray()) ? () : [];
    }

    # if requested to filter by fabric name, do so now
    if (scalar(@fabrics_filter)) {
        my @final_switch_list;
        foreach my $fabric_name (@fabrics_filter) {
            my @switches_filtered = $self->ResMgr::filter_by_event_parameter(-names => \@sanswitches,
                                                                             -types => ["sanswitches"],
                                                                             -event_parameter => "fabrics",
                                                                             -value => $fabric_name);
            if ($self->error()) {
                $self->_set_err("Unable to filter resources by event parameter fabrics => '$fabric_name'");
                return (wantarray()) ? () : [];
            }
            push(@final_switch_list, @switches_filtered);
        }
        @sanswitches = @final_switch_list;
    }

    # if requested to filter by event state, do that now
    if ($event_state != -1) {
        # filter through the sanwitches that match the specified event state
        @sanswitches = $self->ResMgr::filter_by_attribute(-types => ["sanswitches"],
                                                          -names => \@sanswitches,
                                                          -attribute => "faulted",
                                                          -value => $event_state);
        if ($self->error()) {
            $self->_set_err("Unable to filter resources by attribute faulted => '$event_state'");
            return (wantarray()) ? () : [];
        }
    }

    return (wantarray() ? @sanswitches : \@sanswitches);
}

###############################################################################
# Method:    get_fabrics()
# Objective: Obtain the list of Fabric resource names in the current
#            Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fabrics {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $event_state = &ParamsMgr::get_arg(-arg => "faulted",
                                          -default => -1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # determine which fabrics to operate on, obtain the list from the current
    # event/eventset
    my $fabrics = $self->EventUtil::get_arg(-arg => "fabrics");

    my @fabrics = ();
    if (ref($fabrics) eq "ARRAY" and scalar(@{$fabrics})) {
        @fabrics = @{$fabrics};
    } elsif (ref($fabrics) eq "HASH") {
        # then allow the value to be null because the user explicitly set it that way
    } else {
        # default to the full list of fabrics otherwise
        @fabrics = $self->ResMgr::get_resource_names_by_type(-type => "fabrics");
    }

    # if there are no fabrics just return
    if (not scalar(@fabrics)) {
        return (wantarray()) ? () : [];
    }

    # if requested to filter by event state, do that now
    if ($event_state != -1) {
        # filter through the fabrics that match the specified event state
        @fabrics = $self->ResMgr::filter_by_attribute(-types => ["fabrics"],
                                                      -names => \@fabrics,
                                                      -attribute => "faulted",
                                                      -value => $event_state);
        if ($self->error()) {
            $self->_set_err("Unable to filter fabric resources by attribute faulted => '$event_state'");
            return (wantarray()) ? () : [];
        }
    }

    return (wantarray()) ? @fabrics : \@fabrics;
}

###############################################################################
# Method:    get_sanfilers()
# Objective: Obtain the list of SanFiler Resource names in use by the current
#            Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanfilers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the sanfiler types to all (cmode/7mode)
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $event_state = &ParamsMgr::get_arg(-arg => "faulted",
                                          -default => -1);

    # if no sanfilerCm resources are found, then fallback to the controller cluster resources node objects
    my $fall_back_on_controller_clusters = &ParamsMgr::get_arg(-arg => "fall_back_on_controller_clusters",
                                                               -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # determine which sanfilers to operate on, obtain the list from the current
    # event/eventset/eventman environment
    my $sanfilers7m = $self->EventUtil::get_arg(-arg => "sanfilers7m");
    my $sanfilersCm = $self->EventUtil::get_arg(-arg => "sanfilersCm");

    my @sanfilers = ();

    # obtain all sanfilers 7m and Cm
    if ($type =~ /^all$/i) {
        # get 7mode sanfilers
        if (ref($sanfilers7m) eq "ARRAY" and scalar(@{$sanfilers7m})) {
            push(@sanfilers, (@{$sanfilers7m}));
        } elsif (ref($sanfilers7m) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of sanfiler7m resources if none were found
            # in the environment
            push(@sanfilers, $self->ResMgr::get_resource_names_by_type(-type => "sanfilers7m"));
        }

        # get cmode sanfilers
        if (ref($sanfilersCm) eq "ARRAY" and scalar(@{$sanfilersCm})) {
            push(@sanfilers, (@{$sanfilersCm}));
        } elsif (ref($sanfilersCm) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of sanfilerCm resources if none were found
            # in the environment
            push(@sanfilers, $self->ResMgr::get_resource_names_by_type(-type => "sanfilersCm"));
        }
    } elsif ($type =~ /^7mode$/i) {
        # get 7mode sanfilers
        if (ref($sanfilers7m) eq "ARRAY" and scalar(@{$sanfilers7m})) {
            push(@sanfilers, (@{$sanfilers7m}));
        } elsif (ref($sanfilers7m) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of sanfiler7m resources if none were found
            # in the environment
            push(@sanfilers, $self->ResMgr::get_resource_names_by_type(-type => "sanfilers7m"));
        }
    } elsif ($type =~ /^cmode$/i) {
        # get cmode sanfilers
        if (ref($sanfilersCm) eq "ARRAY" and scalar(@{$sanfilersCm})) {
            push(@sanfilers, (@{$sanfilersCm}));
        } elsif (ref($sanfilersCm) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of sanfilerCm resources if none were found
            # in the environment
            push(@sanfilers, $self->ResMgr::get_resource_names_by_type(-type => "sanfilersCm"));
        }
    }

    # if type was cmode or "all" and $sanfilersCm is empty,
    # then if the "fall_back_on_controllerclusters" flag
    # is set, then do so now
    if ($type =~ /cmode|all/ and
        ref($sanfilersCm) eq "ARRAY" and
        scalar(@{$sanfilersCm}) == 0 and
        $fall_back_on_controller_clusters == $EventUtil::TRUE) {

        # fall back on the controller clusters
        my @controller_clusters = $self->EventUtil::get_controllerclusters();
        if ($self->error()) {
            $self->_set_err("Could not obtain Controller clusters");
            return (wantarray()) ? () : [];
        }

        # add all nodes from each controller cluster in the current scope
        foreach my $controller_cluster (@controller_clusters) {
            my $controller_cluster_obj = $self->EventUtil::get_controllercluster_obj(-cluster_name => $controller_cluster);
            if ($self->error()) {
                $self->_set_err("Could not obtain Controller clusters obj for $controller_cluster");
                return (wantarray()) ? () : [];
            }

            # get the list of nodes in this cluster
            push(@sanfilers, $controller_cluster_obj->get_node_members());
            if ($controller_cluster_obj->error()) {
                $self->_set_err($controller_cluster_obj->errmsg());
                return (wantarray()) ? () : [];
            }
        }
    }

    # if there are no sanfiler resources to work with just return none
    if (not scalar(@sanfilers)) {
        return (wantarray()) ? () : [];
    }

    # if requested to filter by event state, do that now
    if ($event_state != -1) {
        # filter through the sanfilers that match the specified event state
        @sanfilers = $self->ResMgr::filter_by_attribute(-types => ["sanfilers7m", "sanfilersCm"],
                                                        -names => \@sanfilers,
                                                        -attribute => "faulted",
                                                        -value => $event_state);
        if ($self->error()) {
            $self->_set_err("Unable to filter resources by attribute faulted: '$event_state'");
            return (wantarray()) ? () : [];
        }
    }

    return (wantarray() ? @sanfilers : \@sanfilers);
}

###############################################################################
# Method:    get_nacl_nodes()
# Objective: Obtain the list of nacl_nodes from the current
#            Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_nodes {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the nacl_node types to all (cmode/7mode)
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $event_state = &ParamsMgr::get_arg(-arg => "faulted",
                                          -default => -1);

    # allow user to filter nacl nodes by cluster_name
    my @clusters_filter = &ParamsMgr::get_arg(-arg => "clusters",
                                              -default => []);

    # if no nacl_node_cm resources are found, then fallback to the controller cluster resources node objects
    my $fall_back_on_nacl_clusters = &ParamsMgr::get_arg(-arg => "fall_back_on_nacl_clusters",
                                                         -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    my $nacl_nodes_cm = $self->EventUtil::get_arg(-arg => "nacl_nodes_cm");
    my $nacl_nodes_7m = $self->EventUtil::get_arg(-arg => "nacl_nodes_7m");

    my @nacl_nodes;

    # obtain all nacl_nodes 7m and Cm
    if ($type =~ /^all$/i) {
        # get 7mode nodes
        if (ref($nacl_nodes_7m) eq "ARRAY" and scalar(@{$nacl_nodes_7m})) {
            push(@nacl_nodes, (@{$nacl_nodes_7m}));
        } elsif (ref($nacl_nodes_7m) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of nacl_node7m resources if none were found
            # in the environment
            push(@nacl_nodes, $self->ResMgr::get_resource_names_by_type(-type => "nacl_nodes_7m"));
        }

        # get cmode nodes
        if (ref($nacl_nodes_cm) eq "ARRAY" and scalar(@{$nacl_nodes_cm})) {
            push(@nacl_nodes, (@{$nacl_nodes_cm}));
        } elsif (ref($nacl_nodes_cm) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of resources if none were found
            # in the environment
            push(@nacl_nodes, $self->ResMgr::get_resource_names_by_type(-type => "nacl_nodes_cm"));
        }
    } elsif ($type =~ /^7mode$/i) {
        # get 7mode nodes
        if (ref($nacl_nodes_7m) eq "ARRAY" and scalar(@{$nacl_nodes_7m})) {
            push(@nacl_nodes, (@{$nacl_nodes_7m}));
        } elsif (ref($nacl_nodes_7m) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of resources if none were found
            # in the environment
            push(@nacl_nodes, $self->ResMgr::get_resource_names_by_type(-type => "nacl_nodes_7m"));
        }
    } elsif ($type =~ /^cmode$/i) {
        # get cmode nodes
        if (ref($nacl_nodes_cm) eq "ARRAY" and scalar(@{$nacl_nodes_cm})) {
            push(@nacl_nodes, (@{$nacl_nodes_cm}));
        } elsif (ref($nacl_nodes_cm) eq "HASH") {
            # then allow the value to be null because the user explicitly set it that way
        } else {
            # default to the full list of resources if none were found
            # in the environment
            push(@nacl_nodes, $self->ResMgr::get_resource_names_by_type(-type => "nacl_nodes_cm"));
        }
    }

    # if type was cmode or "all" and $nacl_nodes_cm is empty,
    # then if the "fall_back_on_controllerclusters" flag
    # is set, then do so now
    if ($type =~ /cmode|all/ and
        ref($nacl_nodes_cm) eq "ARRAY" and
        scalar(@{$nacl_nodes_cm}) == 0 and
        $fall_back_on_nacl_clusters == $EventUtil::TRUE) {

        # fall back on the controller clusters
        my @nacl_clusters_cm = $self->EventUtil::get_nacl_clusters_cm();
        if ($self->error()) {
            $self->_set_err("Could not obtain Controller clusters");
            return (wantarray()) ? () : [];
        }

        # add all nodes from each controller cluster in the current scope
        foreach my $nacl_clusters_cm (@nacl_clusters_cm) {
            my $nacl_cluster_cm_obj = $self->EventUtil::get_nacl_cluster_cm_obj(-cluster_name => $nacl_clusters_cm);
            if ($self->error()) {
                $self->_set_err("Could not obtain Controller clusters obj for $nacl_clusters_cm");
                return (wantarray()) ? () : [];
            }

            # get the list of nodes in this cluster
            try {
                push(@nacl_nodes, $nacl_cluster_cm_obj->nodes());
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught: " . $e->text());
            };
        }
    }

    # if there are no NACL node resources to work with just return none
    if (not scalar(@nacl_nodes)) {
        return (wantarray()) ? () : [];
    }

    # if requested to filter by event state, do that now
    if ($event_state != -1) {
        # filter through the nacl_nodes that match the specified event state
        @nacl_nodes = $self->ResMgr::filter_by_attribute(-types => ["nacl_nodes_7m", "nacl_nodes_cm"],
                                                         -names => \@nacl_nodes,
                                                         -attribute => "faulted",
                                                         -value => $event_state);
        if ($self->error()) {
            $self->_set_err("Unable to filter resources by attribute faulted: '$event_state'");
            return (wantarray()) ? () : [];
        }
    }

    # if requested to filter by cluster name, do so now
    if (scalar(@clusters_filter)) {
        my @final_node_list;
        foreach my $cluster_name (@clusters_filter) {
            my @nodes_filtered = $self->ResMgr::filter_by_attribute(-types => ["nacl_nodes_cm"],
                                                                    -names => \@nacl_nodes,
                                                                    -attribute => "controllercluster_name",
                                                                    -value => $cluster_name);
            if ($self->error()) {
                $self->_set_err("Unable to filter resources by attribute 'cluster_name' : $cluster_name");
                return (wantarray()) ? () : [];
            }
            push(@final_node_list, @nodes_filtered);
        }
        @nacl_nodes = @final_node_list;
    }

    return (wantarray() ? @nacl_nodes : \@nacl_nodes);
}

###############################################################################
# Method:    get_controllerclusters()
# Objective: Obtain the list of ControllerCluster Resource names in use by
#            the current Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_controllerclusters {
    my ($self) = @_;

    # determine which sanfilers to operate on, obtain the list from the current
    # event/eventset/eventman scope
    my $controllerclusters = $self->EventUtil::get_arg(-arg => "controllerclusters");

    my @controllerclusters = ();
    if (ref($controllerclusters) eq "ARRAY" and scalar(@{$controllerclusters})) {
        @controllerclusters = @{$controllerclusters};
    } elsif (ref($controllerclusters) eq "HASH") {
        # then allow the value to be null because the user explicitly set it that way
    } else {
        # default to the full list of controller clusters otherwise
        @controllerclusters = $self->ResMgr::get_resource_names_by_type(-type => "controllerclusters");
    }

    # if there are no controllerclusters just return
    if (not scalar(@controllerclusters)) {
        return (wantarray()) ? () : [];
    }

    return (wantarray() ? @controllerclusters : \@controllerclusters);
}

###############################################################################
# Method:    get_nacl_clusters_cm()
# Objective: Obtain the list of NACL::C::Cluster resource names that are in
#            the current Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_clusters_cm {
    my ($self) = @_;

    # determine which sanfilers to operate on, obtain the list from the current
    # event/eventset/eventman scope
    my $nacl_clusters_cm = $self->EventUtil::get_arg(-arg => "nacl_clusters_cm");

    my @nacl_clusters_cm = ();
    if (ref($nacl_clusters_cm) eq "ARRAY" and scalar(@{$nacl_clusters_cm})) {
        @nacl_clusters_cm = @{$nacl_clusters_cm};
    } elsif (ref($nacl_clusters_cm) eq "HASH") {
        # then allow the value to be null because the user explicitly set it that way
    } else {
        # default to the full list of controller clusters otherwise
        @nacl_clusters_cm = $self->ResMgr::get_resource_names_by_type(-type => "nacl_clusters_cm");
    }

    # if there are no nacl_clusters_cm just return
    if (not scalar(@nacl_clusters_cm)) {
        return (wantarray()) ? () : [];
    }

    return (wantarray() ? @nacl_clusters_cm : \@nacl_clusters_cm);
}

###############################################################################
# Method:    get_vservers()
# Objective: Obtain the list of Vserver names in use in the
#            the current Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vservers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # allow user to filter nacl nodes by cluster_name
    my @clusters_filter = &ParamsMgr::get_arg(-arg => "clusters",
                                              -default => []);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # get the list of vservers in the current scoped environment
    my $vservers = $self->EventUtil::get_arg(%args,
                                             -arg => "vservers");

    my @vservers = ();
    if (ref($vservers) eq "ARRAY" and scalar(@{$vservers})) {
        # dereference the list if its an array
        @vservers = @{$vservers};
    } elsif (ref($vservers) eq "HASH") {
        # then allow the value to be null because the user explicitly set it that way
    } else {
        # default to the full list of vservers registered in the ResMgr
        # if there are none in the environment
        @vservers = $self->ResMgr::get_sub_resource_names_by_type(-type => "vservers",
                                                                  -parent_type => 'nacl_clusters_cm');
    }

    # if there are no vservers just return
    if (not scalar(@vservers)) {
        return (wantarray()) ? () : [];
    }

    # if requested to filter by cluster name, do so now
    if (scalar(@clusters_filter)) {
        my @final_vserver_list;
        foreach my $cluster_name (@clusters_filter) {
            my @vservers_filtered = $self->ResMgr::filter_by_attribute(-names => \@vservers,
                                                                       -types => ['vservers'],
                                                                       -parent_types => ['nacl_clusters_cm'],
                                                                       -attribute => 'controllercluster_name',
                                                                       -value => $cluster_name);
            if ($self->error()) {
                $self->_set_err("Failed to filter vservers by attribute 'controllercluster_name'");
                return (wantarray()) ? () : [];
            }
            push(@final_vserver_list, @vservers_filtered);
        }

        @vservers = @final_vserver_list;
    }

    return (wantarray() ? @vservers : \@vservers);
}

###############################################################################
# Method:    get_sanclient_obj()
# Objective: Obtain the SanClient object for the specified client name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanclient_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        return {};
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_sanclient_objs()
# Objective: Obtain all sanclient objs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanclient_objs {
    my ($self, %args) = @_;

    my @sanclient_names = $self->EventUtil::get_sanclients(%args);
    if ($self->error()) {
        $self->_set_err("Failed to get sanclient names");
        return (wantarray()) ? () : [];
    }

    my @sanclient_objs;

    foreach my $client_name (@sanclient_names) {
        my $sanclient_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
        if ($self->error()) {
            $self->_set_err("Failed to get sanclient object for $client_name");
            return (wantarray()) ? () : [];
        }

        push(@sanclient_objs, $sanclient_obj);
    }

    return (wantarray()) ? @sanclient_objs : \@sanclient_objs;
}

###############################################################################
# Method:    get_sanswitch_obj()
# Objective: Obtain the SanSwitch object for the specified switch name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanswitch_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $switch_name = &ParamsMgr::get_arg(-arg => "switch_name",
                                          -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my $resource = $self->ResMgr::get_resource(-name => $switch_name,
                                               -type => "sanswitches");
    if ($self->error()) {
        return {};
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_sanswitch_objs()
# Objective: Obtain all sanswitch objs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanswitch_objs {
    my ($self, %args) = @_;

    my @sanswitch_names = $self->EventUtil::get_sanswitches(%args);
    if ($self->error()) {
        $self->_set_err("Failed to get sanswitch names");
        return (wantarray()) ? () : [];
    }

    my @sanswitch_objs;

    foreach my $switch_name (@sanswitch_names) {
        my $sanswitch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
        if ($self->error()) {
            $self->_set_err("Failed to get sanswitch object for $switch_name");
            return (wantarray()) ? () : [];
        }

        push(@sanswitch_objs, $sanswitch_obj);
    }

    return (wantarray()) ? @sanswitch_objs : \@sanswitch_objs;
}

###############################################################################
# Method:    get_sanfiler_obj()
# Objective: Obtain the SanFiler object for the specified controller name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanfiler_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $filer_name = &ParamsMgr::get_arg(-arg => "filer_name",
                                         -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # first look through sanfilers7m
    my $resource = $self->ResMgr::get_resource(-name => $filer_name,
                                               -type => "sanfilers7m");
    if ($self->error()) {
        $self->_clear_err();

        # try looking through sanfilersCm if we couldnt find it
        $resource = $self->ResMgr::get_resource(-name => $filer_name,
                                                -type => "sanfilersCm");
        if ($self->error()) {
            $self->_set_err("Unable to find resource '$filer_name' in both 'sanfilers7m' and 'sanfilersCm'");
            return {};
        }
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_sanfiler_objs()
# Objective: Obtain all sanfiler objs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanfiler_objs {
    my ($self, %args) = @_;

    my @sanfiler_names = $self->EventUtil::get_sanfilers(%args);
    if ($self->error()) {
        $self->_set_err("Failed to get sanfiler names");
        return (wantarray()) ? () : [];
    }

    my @sanfiler_objs;

    foreach my $filer_name (@sanfiler_names) {
        my $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $filer_name);
        if ($self->error()) {
            $self->_set_err("Failed to get sanfiler object for $filer_name");
            return (wantarray()) ? () : [];
        }

        push(@sanfiler_objs, $sanfiler_obj);
    }

    return (wantarray()) ? @sanfiler_objs : \@sanfiler_objs;
}

###############################################################################
# Method:    get_nacl_node_obj()
# Objective: Obtain the nacl node object specified
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_node_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # look at the 7mode resources first
    my $resource = $self->ResMgr::get_resource(-name => $node_name,
                                               -type => "nacl_nodes_7m");
    if ($self->error()) {
        $self->_clear_err();

        # try looking through nacl_nodes_cm if we couldnt find it
        $resource = $self->ResMgr::get_resource(-name => $node_name,
                                                -type => "nacl_nodes_cm");
        if ($self->error()) {
            $self->_set_err("Unable to find resource '$node_name' in both 'nacl_nodes_7m' and 'nacl_nodes_cm'");
            return {};
        }
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_nacl_node_objs()
# Objective: Obtain the nacl node objects in the current
#            Event/EventSet/EventMan environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_node_objs {
    my ($self, %args) = @_;

    my @nacl_node_names = $self->EventUtil::get_nacl_nodes(%args);
    if ($self->error()) {
        $self->_set_err("Failed to get nacl node names");
        return (wantarray()) ? () : [];
    }

    my @nacl_node_objs;

    foreach my $node_name (@nacl_node_names) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl node object for $node_name");
            return (wantarray()) ? () : [];
        }

        push(@nacl_node_objs, $nacl_node_obj);
    }

    return (wantarray()) ? @nacl_node_objs : \@nacl_node_objs;
}

###############################################################################
# Method:    get_controllercluster_obj()
# Objective: Obtain the Controller Cluster object for the specified cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_controllercluster_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $cluster_name = &ParamsMgr::get_arg(-arg => "cluster_name",
                                           -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my $resource = $self->ResMgr::get_resource(-name => $cluster_name,
                                               -type => "controllerclusters");
    if ($self->error()) {
        return {};
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_controllercluster_objs()
# Objective: Obtain all controllercluster objs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_controllercluster_objs {
    my ($self, %args) = @_;

    my @controllercluster_names = $self->EventUtil::get_controllerclusters(%args);
    if ($self->error()) {
        $self->_set_err("Failed to get controllercluster names");
        return (wantarray()) ? () : [];
    }

    my @controllercluster_objs;

    foreach my $cluster_name (@controllercluster_names) {
        my $controllercluster_obj = $self->EventUtil::get_controllercluster_obj(-cluster_name => $cluster_name);
        if ($self->error()) {
            $self->_set_err("Failed to get controllercluster object for $cluster_name");
            return (wantarray()) ? () : [];
        }

        push(@controllercluster_objs, $controllercluster_obj);
    }

    return (wantarray()) ? @controllercluster_objs : \@controllercluster_objs;
}

###############################################################################
# Method:    get_nacl_cluster_cm_obj()
# Objective: Obtain the NACL Cluster CM object for this resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_cluster_cm_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $cluster_name = &ParamsMgr::get_arg(-arg => "cluster_name",
                                           -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my $resource = $self->ResMgr::get_resource(-name => $cluster_name,
                                               -type => "nacl_clusters_cm");
    if ($self->error()) {
        return {};
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_nacl_cluster_cm_objs()
# Objective: Obtain all nacl_cluster_cm objects
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nacl_cluster_cm_objs {
    my ($self, %args) = @_;

    my @cluster_names = $self->EventUtil::get_nacl_clusters_cm(%args);
    if ($self->error()) {
        $self->_set_err("Failed to get nacl_clusters_cm names");
        return (wantarray()) ? () : [];
    }

    my @nacl_cluster_cm_objs;

    foreach my $cluster_name (@cluster_names) {
        my $nacl_cluster_cm_obj = $self->EventUtil::get_nacl_cluster_cm_obj(-cluster_name => $cluster_name);
        if ($self->error()) {
            $self->_set_err("Failed to get nacl_clusters_cm object for $cluster_name");
            return (wantarray()) ? () : [];
        }

        push(@nacl_cluster_cm_objs, $nacl_cluster_cm_obj);
    }

    return (wantarray()) ? @nacl_cluster_cm_objs : \@nacl_cluster_cm_objs;
}

###############################################################################
# Method:    get_vserver_obj()
# Objective: Obtain the NACL::C::Vserver object for the specified cluster:vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver_obj {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $vserver_name = &ParamsMgr::get_arg(-arg => "vserver_name",
                                           -required => 1);
    my $cluster_name = &ParamsMgr::get_arg(-arg => "cluster_name",
                                           -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my $resource = $self->ResMgr::get_resource(-name => $vserver_name,
                                               -type => "vservers",
                                               -parent_name => $cluster_name,
                                               -parent_type => 'nacl_clusters_cm');
    if ($self->error()) {
        $self->_set_err("Failed to get sub resource for vserver '$vserver_name' on cluster '$cluster_name'");
        return {};
    }

    return $resource->{obj};
}

###############################################################################
# Method:    get_vserver_objs()
# Objective: Obtain all vserver objs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver_objs {
    my ($self, %args) = @_;


    &ParamsMgr::process_args(%args);

    # filter by cluster name
    my @clusters = &ParamsMgr::get_arg(-arg => "clusters",
                                       -ref_type => 'ARRAY',
                                       -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    my @vserver_objs;

    foreach my $cluster (@clusters) {
        my @vserver_names = $self->EventUtil::get_vservers(%args,
                                                           -clusters => [$cluster]);
        if ($self->error()) {
            $self->_set_err("Failed to get vserver names");
            return (wantarray()) ? () : [];
        }

        foreach my $vserver_name (@vserver_names) {
            my $vserver_obj = $self->EventUtil::get_vserver_obj(-vserver_name => $vserver_name,
                                                                -cluster_name => $cluster);
            if ($self->error()) {
                $self->_set_err("Failed to get vserver object for $vserver_name on cluster $cluster");
                return (wantarray()) ? () : [];
            }

            push(@vserver_objs, $vserver_obj);
        }
    }

    return (wantarray()) ? @vserver_objs : \@vserver_objs;
}

###############################################################################
# Method:    reconnect_all_sanclients()
# Objective: Create new connections to all SanClient objects in the
#            current Event environment.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_sanclients {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # if a specific set of clients are specified use that instead
    my @client_names = &ParamsMgr::get_arg(-arg => "client_names",
                                           -ref_type => "ARRAY",
                                           -default => []);

    # if not sanfilerCm resources are found, then fallback to the controller cluster resources node objects
    my $disconnect_existing_conn = &ParamsMgr::get_arg(-arg => "disconnect_existing_conn",
                                                       -default => $EventUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not requested a specific set of clients, use the
    # list currently in the environment
    if (not scalar(@client_names)) {
        @client_names = $self->EventUtil::get_sanclients(-faulted => $faulted);
        if ($self->error()) {
            $self->_set_err("Failed to get SanClients");
            return $EventUtil::FALSE;
        }
    }

    # reconnect all sanclient objects that are not currently in a faulted state
    foreach my $client_name (@client_names) {
        my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
        if ($self->error()) {
            $self->_set_err("Failed to get sanclient object '$client_name'");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log(-comment, "Reconnecting to '$client_name'");

        $sc_obj->reconnect(-disconnect_existing_conn => $disconnect_existing_conn);
        if ($sc_obj->error()) {
            $self->_set_err($sc_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disconnect_all_sanclients()
# Objective: Disconnect all SanClient connections in the current Event
#            environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_all_sanclients {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # if a specific set of clients are specified use that instead
    my @client_names = &ParamsMgr::get_arg(-arg => "client_names",
                                           -ref_type => "ARRAY",
                                           -default => []);

    # allow this function to ignore all errors set on the object if requested
    my $ignore_errors = &ParamsMgr::get_arg(-arg => "ignore_errors",
                                            -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error() and not $ignore_errors) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if ignore_errors is set, then save any errors that may be on the object, and restore later
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_save_error();
    }

    # if not requested a specific set of clients, use the
    # list currently in the environment
    if (not scalar(@client_names)) {
        @client_names = $self->EventUtil::get_sanclients(-faulted => $faulted);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get SanClients");
            return $EventUtil::FALSE;
        }
    }

    foreach my $client_name (@client_names) {
        my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get sanclient object '$client_name'");
            return $EventUtil::FALSE;
        }

        if ($sc_obj->can("get_connection")) {
            &Tharn::disconnect($sc_obj->get_connection(), '-ok');
        }
    }

    # if ignore_errors is set, then restore any error that may be saved away
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_restore_error();
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_disconnected_sanswitches()
# Objective: Create new connections to disconnected SanSwitch objects in the
#            current Event environment. Basically go through a list of switch
#            names, get the objects, find out which are disconnnected, and pass
#            that list along to reconnect_all_sanswitches
###############################################################################
sub reconnect_disconnected_sanswitches {
    my ($self, %args) = @_;
    my $switch_names = $args{'-switch_names'};
    my @filtered_switch_names = ();
    foreach my $switch_name (@{$switch_names}) {
        my $switch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
        if ($self->error()) {
            $self->_set_err("Failed to get sanswitch object '$switch_name'");
            return $EventUtil::FALSE;
        }
        my $switch_conn = $switch_obj->{conn};
        my $disconnected = 1;
        if (defined $switch_conn) {        
            $disconnected = $switch_obj->{conn}->is_disconnected(verify => "1");
        }

        if ($switch_obj->error()) {
            $self->_set_err($switch_obj->errmsg());
            return $EventUtil::FALSE;
        }

        if ($disconnected) {
            push(@filtered_switch_names, $switch_name);
        } else {
            &SanUtil::nate_log(-comment, "Switch  '$switch_name' has a good connection, not reconnecting");
        }
    }
    if(@filtered_switch_names) {
        $args{switch_names} = \@filtered_switch_names;
        return $self->EventUtil::reconnect_all_sanswitches(%args);
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_all_sanswitches()
# Objective: Create new connections to all SanSwitch objects in the
#            current Event environment.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_sanswitches {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # if not sanfilerCm resources are found, then fallback to the controller cluster resources node objects
    my $disconnect_existing_conn = &ParamsMgr::get_arg(-arg => "disconnect_existing_conn",
                                                       -default => $EventUtil::TRUE);

    # if a specific set of switches are specified use that instead of the environment scope
    my @switch_names = &ParamsMgr::get_arg(-arg => "switch_names",
                                           -ref_type => "ARRAY",
                                           -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not requested a specific set of switches, use the
    # list currently in the environment
    if (not scalar(@switch_names)) {
        @switch_names = $self->EventUtil::get_sanswitches(-faulted => $faulted);
        if ($self->error()) {
            $self->_set_err("Failed to get SanSwitches");
            return $EventUtil::FALSE;
        }
    }

    # reconnect all sanclient objects that are not currently in a faulted state
    foreach my $switch_name (@switch_names) {
        my $switch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
        if ($self->error()) {
            $self->_set_err("Failed to get sanswitch object '$switch_name'");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log(-comment, "Reconnecting to '$switch_name'");
        
        $switch_obj->reconnect(-disconnect_existing_conn => $disconnect_existing_conn);
        if ($switch_obj->error()) {
            $self->_set_err($switch_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disconnect_all_sanswitches()
# Objective: Disconnect all SanSwitch connections in the current Event
#            environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_all_sanswitches {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # if a specific set of switches are specified use that instead of the environment scope
    my @switch_names = &ParamsMgr::get_arg(-arg => "switch_names",
                                           -ref_type => "ARRAY",
                                           -default => []);

    # allow this function to ignore all errors set on the object if requested
    my $ignore_errors = &ParamsMgr::get_arg(-arg => "ignore_errors",
                                            -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error() and not $ignore_errors) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if ignore_errors is set, then save any errors that may be on the object, and restore later
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_save_error();
    }

    # if not requested a specific set of switches, use the
    # list currently in the environment
    if (not scalar(@switch_names)) {
        @switch_names = $self->EventUtil::get_sanswitches(-faulted => $faulted);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get SanSwitches");
            return $EventUtil::FALSE;
        }
    }

    foreach my $switch_name (@switch_names) {
        my $switch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch_name);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get sanswitch object '$switch_name'");
            return $EventUtil::FALSE;
        }

        # only disconnect if we can fetch the connection
        if ($switch_obj->can("get_connection")) {
            &Tharn::disconnect($switch_obj->get_connection(), '-ok');
        }
    }

    # if ignore_errors is set, then restore any error that may be saved away
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_restore_error();
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_all_cservers();
# Objective: Create new connections to all cserver objects in the
#            current Event environment.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_cservers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # if not sanfilerCm resources are found, then fallback to the controller cluster resources node objects
    my $disconnect_existing_conn = &ParamsMgr::get_arg(-arg => "disconnect_existing_conn",
                                                       -default => $EventUtil::TRUE);

    # if a specific set of controller are specified use that instead of the environment scope
    my @cluster_names = &ParamsMgr::get_arg(-arg => "cluster_names",
                                            -ref_type => "ARRAY",
                                            -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not requested a specific set of clusters, use the
    # list currently in the environment
    if (not scalar(@cluster_names)) {
        @cluster_names = $self->EventUtil::get_controllerclusters();
        if ($self->error()) {
            $self->_set_err("Failed to get Clusters");
            return $EventUtil::FALSE;
        }
    }

    # reconnect all cserver objects
    foreach my $cluster_name (@cluster_names) {
        my $cluster_obj = $self->EventUtil::get_controllercluster_obj(-cluster_name => $cluster_name);
        if ($self->error()) {
            $self->_set_err("Failed to get Controller cluster object '$cluster_name'");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log(-comment, "Reconnecting to cserver on cluster '$cluster_name'");

        my $cserver_obj = $cluster_obj->get_cserver_obj();
        if ($cluster_obj->error()) {
            $self->_set_err($cluster_obj->errmsg());
            return $EventUtil::FALSE;
        }

        $cserver_obj->reconnect(-disconnect_existing_conn => $disconnect_existing_conn);
        if ($cserver_obj->error()) {
            $self->_set_err($cserver_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disconnect_all_cservers()
# Objective: Disconnect all Cserver connections in the current Event
#            environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_all_cservers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # if a specific set of controller are specified use that instead of the environment scope
    my @cluster_names = &ParamsMgr::get_arg(-arg => "cluster_names",
                                            -ref_type => "ARRAY",
                                            -default => []);

    # allow this function to ignore all errors set on the object if requested
    my $ignore_errors = &ParamsMgr::get_arg(-arg => "ignore_errors",
                                            -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error() and not $ignore_errors) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if ignore_errors is set, then save any errors that may be on the object, and restore later
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_save_error();
    }

    # if not requested a specific set of clusters, use the
    # list currently in the environment
    if (not scalar(@cluster_names)) {
        @cluster_names = $self->EventUtil::get_controllerclusters();
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get Clusters");
            return $EventUtil::FALSE;
        }
    }

    foreach my $cluster_name (@cluster_names) {
        my $cluster_obj = $self->EventUtil::get_controllercluster_obj(-cluster_name => $cluster_name);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get Controller cluster object '$cluster_name'");
            return $EventUtil::FALSE;
        }

        my $cserver_obj = $cluster_obj->get_cserver_obj();
        if ($cluster_obj->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err($cluster_obj->errmsg());
            return $EventUtil::FALSE;
        }

        if ($cserver_obj->can("connection")) {
            &Tharn::disconnect($cserver_obj->connection(), '-ok');
        }
    }

    # if ignore_errors is set, then restore any error that may be saved away
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_restore_error();
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_all_sanfilers()
# Objective: Create new connections to all SanFiler objects in the
#            current Event environment.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_sanfilers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # if not sanfilerCm resources are found, then fallback to the controller cluster resources node objects
    my $disconnect_existing_conn = &ParamsMgr::get_arg(-arg => "disconnect_existing_conn",
                                                       -default => $EventUtil::TRUE);

    # default sanfiler type to all (cmode/7mode)
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");

    # if a specific set of controller are specified use that instead of the environment scope
    my @controller_names = &ParamsMgr::get_arg(-arg => "controller_names",
                                               -ref_type => "ARRAY",
                                               -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not requested a specific set of sanfilers, use the
    # list currently in the environment
    if (not scalar(@controller_names)) {
        @controller_names = $self->EventUtil::get_sanfilers(-type => $type,
                                                            -faulted => $faulted);
        if ($self->error()) {
            $self->_set_err("Failed to get SanFilers");
            return $EventUtil::FALSE;
        }
    }

    # reconnect all sanfiler objects that are not currently in a faulted state
    foreach my $filer_name (@controller_names) {
        my $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $filer_name);
        if ($self->error()) {
            $self->_set_err("Failed to get SanFiler object '$filer_name'");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log(-comment, "Reconnecting to '$filer_name'");

        $sanfiler_obj->reconnect(-disconnect_existing_conn => $disconnect_existing_conn);
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_all_nacl_nodes()
# Objective: Create new connections to all NACL::C::Node objects
#            current Event environment.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_nacl_nodes {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # default sanfiler type to all (cmode/7mode)
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");

    my $no_version_change = &ParamsMgr::get_arg(-arg => "no_version_change",
                                                -default => $EventUtil::TRUE);

    # if a specific set of controller are specified use that instead of the environment scope
    my @controller_names = &ParamsMgr::get_arg(-arg => "controller_names",
                                               -ref_type => "ARRAY",
                                               -default => []);
    my @clusters_filter = &ParamsMgr::get_arg(-arg => "clusters",
                                              -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not requested a specific set of nacl nodes, use the
    # list currently in the environment
    if (not scalar(@controller_names)) {
        @controller_names = $self->EventUtil::get_nacl_nodes(-type => $type,
                                                             -faulted => $faulted,
                                                             -clusters => \@clusters_filter);
        if ($self->error()) {
            $self->_set_err("Failed to get NACL Nodes");
            return $EventUtil::FALSE;
        }
    }

    # reconnect all nacl node objects that are not currently in a faulted state
    foreach my $node_name (@controller_names) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Failed to get NACL::C::Node object '$node_name'");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log(-comment, "Refreshing connection to NACL::C::Node '$node_name'");

        # this won't actually reconnect to the controller, it will set a bit on the controller that
        # tells NACL to reconnect the connection the next time it actually uses it.
        try {
            $nacl_node_obj->refresh_command_interface(no_version_change => $no_version_change);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_all_nacl_clusters()
# Objective: Create new connections to all NACL::C::Cluster objects
#            current Event environment.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_nacl_clusters {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @cluster_names = &ParamsMgr::get_arg(-arg => "cluster_names",
                                            -ref_type => "ARRAY",
                                            -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not requested a specific set of nacl nodes, use the
    # list currently in the environment
    if (not scalar(@cluster_names)) {
        @cluster_names = $self->EventUtil::get_nacl_clusters_cm();
        if ($self->error()) {
            $self->_set_err("Failed to get NACL Clusters");
            return $EventUtil::FALSE;
        }
    }

    # reconnect all nacl node objects that are not currently in a faulted state
    foreach my $cluster_name (@cluster_names) {
        my $nacl_cluster_obj = $self->EventUtil::get_nacl_cluster_cm_obj(-cluster_name => $cluster_name);
        if ($self->error()) {
            $self->_set_err("Failed to get NACL::STask::Node object '$cluster_name'");
            return $EventUtil::FALSE;
        }

        &SanUtil::nate_log(-comment, "Refreshing connection to NACL::STask::Cluster '$cluster_name'");

        # this won't actually reconnect to the controller, it will set a bit on the controller that
        # tells NACL to reconnect the connection the next time it actually uses it.
        try {
            $nacl_cluster_obj->refresh_command_interface();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disconnect_all_sanfilers()
# Objective: Disconnect all SanFilers connections in the current Event
#            environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_all_sanfilers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # default sanfiler type to all (cmode/7mode)
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");

    # if a specific set of controller are specified use that instead of the environment scope
    my @controller_names = &ParamsMgr::get_arg(-arg => "controller_names",
                                               -ref_type => "ARRAY",
                                               -default => []);

    # allow this function to ignore all errors set on the object if requested
    my $ignore_errors = &ParamsMgr::get_arg(-arg => "ignore_errors",
                                            -default => $EventUtil::FALSE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error() and not $ignore_errors) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if ignore_errors is set, then save any errors that may be on the object, and restore later
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_save_error();
    }

    # if not requested a specific set of sanfilers, use the
    # list currently in the environment
    if (not scalar(@controller_names)) {
        @controller_names = $self->EventUtil::get_sanfilers(-type => $type,
                                                            -faulted => $faulted);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get SanFilers");
            return $EventUtil::FALSE;
        }
    }

    foreach my $filer_name (@controller_names) {
        my $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $filer_name);
        if ($self->error() and not $ignore_errors) {
            $self->EventUtil::_restore_error();
            $self->_set_err("Failed to get sanfiler object '$filer_name'");
            return $EventUtil::FALSE;
        }

        if ($sanfiler_obj->can("connection")) {
            &Tharn::disconnect($sanfiler_obj->connection(), '-ok');
        }
    }

    # if ignore_errors is set, then restore any error that may be saved away
    if ($ignore_errors == $EventUtil::TRUE) {
        $self->EventUtil::_restore_error();
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_to_previously_faulted_resources()
# Objective: Reconnect to resources that have been previously faulted
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_to_previously_faulted_resources {
    my ($self) = @_;

    my $resource_hash = $self->ResMgr::get_current_resources();

    foreach my $type (keys %{$resource_hash}) {
        foreach my $name (@{$resource_hash->{$type}}) {
            my $faulted_state = $self->ResMgr::get_resource_attribute(-name => $name,
                                                                      -type => $type,
                                                                      -attribute => "faulted");
            if ($self->error()) {
                $self->_set_err("Failed to obtain faulted state attribute from resource ($name : $type)");
                return $EventUtil::FALSE;
            }

            # dont reconnect to devices that are still faulted
            next if ($faulted_state eq $EventUtil::TRUE);

            my $resource = $self->ResMgr::get_resource(-name => $name,
                                                       -type => $type);

            # obtain the transitioning attribute
            my $transitioned_state = $self->ResMgr::get_resource_attribute(-name => $name,
                                                                           -type => $type,
                                                                           -attribute => "transitioned_faulted_unfaulted");
            if ($self->error()) {
                $self->_set_err("Failed to obtain transitioned faulted unfaulted attribute from resource ($name : $type)");
                return $EventUtil::FALSE;
            }
            # check if this resource is currently in transition
            if (defined($transitioned_state) and $transitioned_state eq 1) {
                # then this resource needs to be reconnected to, obtain the object for it
                # and reconnect to it
                my $resource = $self->ResMgr::get_resource(-name => $name,
                                                           -type => $type);
                if ($self->error()) {
                    $self->_set_err("Failed to obtain resource ($name : $type)");
                    return $EventUtil::FALSE;
                }

                # reconnect to this resource if possible (dont reconnect to nacl objects)
                if ($type =~ /nacl/) {
                    if ($resource->{obj}->can("refresh_command_interface")) {
                        try {
                            $resource->{obj}->refresh_command_interface();
                        };
                    }
                } else {
                    if ($resource->{obj}->can("reconnect")) {
                        $resource->{obj}->reconnect();
                    }
                }

                # Unset the transitioning faulted/unfaulted attribute on this resource
                # burt542328: set propagate to false, otherwise sibling EventSet's may not reconnect properly to
                # faulted/unfaulted resources. The reason why we don't propagate is because the parent may
                # believe (and tell other child subtests) that this resource never transitioned between faulted/unfaulted
                # and the child subtests will get stale connections to some resources
                $self->ResMgr::set_resource_attribute(-attribute => "transitioned_faulted_unfaulted",
                                                      -value => 0,
                                                      -name => $name,
                                                      -type => $type,
                                                      -propagate => $EventUtil::FALSE);
                if ($self->error()) {
                    $self->_set_err("Failed to set transitioned faulted unfaulted attribute for resource ($name : $type)");
                    return $EventUtil::FALSE;
                }
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    create_new_resource_connections()
# Objective: Create fresh resource connections to all SanFilers and SanClients
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_new_resource_connections {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # create fresh connections for all sanfiler objects, dont disconnect
    # the original connection
    $self->EventUtil::reconnect_all_sanfilers(-disconnect_existing_conn => 0,
                                              -faulted => $faulted);
    if ($self->error()) {
        $self->_set_err("Create reconnect all sanfilers");
        return $EventUtil::FALSE;
    }

    # create fresh connections for all sanclient objects, dont disconnect
    # the original connection
    $self->EventUtil::reconnect_all_sanclients(-disconnect_existing_conn => 0,
                                               -faulted => $faulted);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # create fresh connections for all sanswitch objects, dont disconnect
    # the original connection
    $self->EventUtil::reconnect_all_sanswitches(-disconnect_existing_conn => 0,
                                                -faulted => $faulted);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # reconnect to all nacl nodes (perform cdef versioning)
    $self->EventUtil::reconnect_all_nacl_nodes(-no_version_change => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->EventUtil::reconnect_all_nacl_clusters();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    reconnect_all_resource_connections()
# Objective: Reconnect all SanFiler and SanClient connections
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect_all_resource_connections {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the event state type to either of the faulted states (faulted|unfaulted)
    my $faulted = &ParamsMgr::get_arg(-arg => "faulted",
                                      -default => -1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # reconnect all sanfiler object connections
    # disconnect the previous connections
    $self->EventUtil::reconnect_all_sanfilers(-disconnect_existing_conn => 1,
                                              -faulted => $faulted);
    if ($self->error()) {
        $self->_set_err("Failed to reconnect all SanFilers");
        return $EventUtil::FALSE;
    }

    # reconnect all sanclient connections,
    # disconnect the previous connections
    $self->EventUtil::reconnect_all_sanclients(-disconnect_existing_conn => 1,
                                               -faulted => $faulted);
    if ($self->error()) {
        $self->_set_err("Failed to reconnect all SanClients");
        return $EventUtil::FALSE;
    }

    # reconnect all sanswitch connections,
    # disconnect the previous connections
    $self->EventUtil::reconnect_all_sanswitches(-disconnect_existing_conn => 1,
                                                -faulted => $faulted);
    if ($self->error()) {
        $self->_set_err("Failed to reconnect all SanSwitches");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    disconnect_all_resource_connections()
# Objective: Disconnect all SanFiler and SanClient connections
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_all_resource_connections {
    my ($self, %args) = @_;

    # disconnect all sanfiler object connections
    $self->EventUtil::disconnect_all_sanfilers(%args);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # disconnect all sanclient connections
    $self->EventUtil::disconnect_all_sanclients(%args);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # disconnect all sanswitch connections
    $self->EventUtil::disconnect_all_sanswitches(%args);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    synchronize()
# Objective: Synchronize this subtest with sibling subtests
# Details:   See POD documentation at the end of this file.
###############################################################################
sub synchronize {
    my ($self) = @_;

    # collect all the ResMgr resources attributes values and send
    # them as payload to subtest_wait()
    my $resmgr_attr_state = $self->ResMgr::get_resmgr_attrs();
    if ($self->error()) {
        $self->_set_err("Failed to get ResMgr attrs");
        return undef;
    }

    # issue a subtest wait, this will return how long this subtest waited
    # in seconds
    return $self->SubtestComm::subtest_wait(-data => $resmgr_attr_state,
                                            -data_type => "resmgr_update");
}

###############################################################################
# Method:    synchronize_and_sleep()
# Objective: Synchronize this subtest with sibling subtests then sleep the
#            time based on the arg passed in.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub synchronize_and_sleep {
    my ($self, %args) = @_;

    my $arg = defined($args{-arg}) ? $args{-arg} : $args{arg};
    # arg is a required parameter.
    unless (defined $arg and length $arg) {
        $self->_set_err("Failed to get ResMgr attrs");
        return undef;
    }

    # This will return how long this subtest waited in seconds
    my $return = $self->EventUtil::synchronize();

    # return the total time that was waited including the sleep in seconds.
    return ($self->EventUtil::sleep_arg(-arg => $arg) + $return);
}

################################################################################
# Subroutine Name: document_event_param()
# Objective:       Document a specific Event Parameter
# Details:   See POD documentation at the end of this file.
################################################################################
sub document_event_param {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse out the parameter to document and the package that is documenting it
    my $param = &ParamsMgr::get_arg(-arg => "param",
                                    -required => 1);
    my $package = &ParamsMgr::get_arg(-arg => "package",
                                      -required => 1);

    # parse out which resources are affected by this param
    my @resource_types = &ParamsMgr::get_arg(-arg => "resource_types",
                                             -ref_type => "ARRAY",
                                             -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # remove the args that we parsed out
    delete($args{-param});
    delete($args{-package});

    # remove all dashes
    %args = &SanUtil::strip_dash(%args);

    # categorize which package is documenting this param and store the documentation (%args)
    # in _event_param_documentation
    if ($package =~ /EventMan::EventModules/) {
        $package =~ s/EventMan::EventModules/EM/;
        $self->{_event_param_documentation}->{event_modules_params}->{$package}->{param}->{$param} = \%args;
    } else {
        $self->{_event_param_documentation}->{utility_modules_params}->{$package}->{param}->{$param} = \%args;
        $self->{_event_param_documentation}->{utility_modules}->{$package} = {};
    }

    # also keep track of all the params that affect each resource type
    foreach my $type (@resource_types) {
        $self->{_event_param_documentation}->{resource_types}->{$type}->{param}->{$param} = {};
    }

    return $EventUtil::TRUE;
}

################################################################################
# Subroutine Name: register_event()
# Objective:       Register an Event with the EventMan library
# Details:   See POD documentation at the end of this file.
################################################################################
sub register_event {
    my ($self, %args) = @_;

    # invoke the register method in EventMan.pm
    $self->_register_event(%args);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

################################################################################
# Subroutine Name: register_event_module_defaults()
# Objective:       Register an Event with the EventMan library
# Details:   See POD documentation at the end of this file.
################################################################################
sub register_event_module_defaults {
    my ($self, %args) = @_;

    # invoke the register method in EventMan.pm
    $self->_register_event_module_default_args(%args);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

################################################################################
# Subroutine Name: register_utility_module_defaults()
# Objective:       Register a utility module's event parameter defaults with
#                  the EventMan library
# Details:   See POD documentation at the end of this file.
################################################################################
sub register_utility_module_defaults {
    my ($self, %args) = @_;

    # invoke the register method in EventMan.pm
    $self->_register_utility_module_default_args(%args);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sleep_arg()
# Objective: Obtain the specified Event Parameter value and sleep for its
#            value.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sleep_arg {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $arg = &ParamsMgr::get_arg(-arg => "arg",
                                  -required => 1);
    my $handle_stop_messages = &ParamsMgr::get_arg(-arg => "handle_stop_messages",
                                                   -default => $EventUtil::FALSE);
    my $poll_interval = &ParamsMgr::get_arg(-arg => "poll_interval",
                                            -default => 2);
    my $exit_callback_list = &ParamsMgr::get_arg(-arg => "exit_callback_list",
                                                 -ref_type => "ARRAY",
                                                 -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return undef;
    }

    # sleep the amount of time specified by the value
    my $value = $self->EventUtil::get_arg(%args);
    if ($self->error() or not defined($value)) {
        $self->_set_err("Unable to get value for Event Parameter '$arg'");
        return undef;
    }

    if ($handle_stop_messages eq $EventUtil::FALSE) {
        &SanUtil::nate_log('-comment', "Sleeping ${arg}=${value} seconds");
        &SanUtil::nate_sleep($value);
    } else {
        # calculate the end time.
        my $stime = time();
        my $end_time = $stime + $value;
        &SanUtil::nate_log('-comment', "Sleeping ${arg}=${value} seconds (Checking for stop messages)");
        my $stop = 0;

        # poll until he parent told us to stop
        while (time < $end_time) {
            # see if the parent has told us to stop
            if ($self->SubtestComm::has_parent_requested_to_stop()) {
                &SanUtil::nate_log('-comment', "Stop/Abort/Timeout message was received from parent");
                # if there are any exit callbacks, execute them in the order specified
                if (scalar(@{$exit_callback_list})) {
                    foreach my $callback (@{$exit_callback_list}) {
                        &SanUtil::nate_log('-comment', "Executing exit callback '$callback'");
                        my $args_ref = $callback->{args};
                        my $code_ref = $callback->{coderef};
                        $code_ref->($args_ref);
                    }
                }
                # enter the stopped phase.
                # this will tell the parent that it is ok to terminate this
                # subtest.
                $self->SubtestComm::subtest_stopped("Stopping Event");
                last;
            }

            &SanUtil::nate_sleep($poll_interval);
        }

    }

    return $value;
}

###############################################################################
# Method:    build_switch_info_hash()
# Objective: Gather and Populate the switch info hash
# Details:   See POD documentation at the end of this file.
###############################################################################
sub build_switch_info_hash {
    my ($self) = @_;

    # Check to see if the resource hash has already been built and loaded
    # into ResourceMgr
    my $resource_hash = $self->ResMgr::get_resource(-name => "switch_info",
                                                    -type => "_GLOBALS");
    unless ($self->error()) {
        # The resource hash has already been built and loaded into
        # ResMgr
        return $EventUtil::TRUE;
    }

    # The switch info hash has not been built yet so clear the error from ResMgr and
    # start building it.
    $self->_clear_err();

    # initialize every SanSwitch resource
    my (@sanswitches) = $self->EventUtil::get_sanswitch_objs(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain all SanSwitch resource names");
        return $EventUtil::FALSE;
    }

    # get all sanclients
    my @sanclients = $self->EventUtil::get_sanclient_objs(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Failed to get all SanClients");
        return $EventUtil::FALSE;
    }

    # get all cmode cluster objects
    my @cmode_clusters = $self->EventUtil::get_nacl_cluster_cm_objs();
    if ($self->error()) {
        $self->_set_err("Failed to get all CMode Clusters");
        return $EventUtil::FALSE;
    }

    # get all of the vserver names
    my @cmode_vservers = $self->EventUtil::get_vservers();
    if ($self->error()) {
        $self->_set_err("Failed to get all vserver names");
        return $EventUtil::FALSE;
    }

    # get all 7mode sanfilers
    my @sanfilers_7m = $self->EventUtil::get_sanfiler_objs(-type => "7mode");
    if ($self->error()) {
        $self->_set_err("Failed to get all 7Mode SanFilers");
        return $EventUtil::FALSE;
    }

    my $fm_obj = new San::FabricMgr(-client_objs       => \@sanclients,
                                    -nacl_cluster_objs => \@cmode_clusters,
                                    -vservers          => \@cmode_vservers,
                                    -sanfiler7m_objs   => \@sanfilers_7m,
                                    -switch_objs       => \@sanswitches,
                                    -fabric            => "eventman",
                                    -used_for_zoning   => 0);
    if ($fm_obj->error()) {
        $self->_set_err("Failed to create FabricMgr obj", $fm_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # this will force the build of the switch info hash and return it
    my $hash = $fm_obj->get_switch_info_hash(-refresh => 1);
    if ($fm_obj->error()) {
        $self->_set_err("Failed to get switch info hash", $fm_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # register the switch_info datastructure as a "_GLOBALS" resource type
    # this will allow events to access this datastructure and make persistent modifications to it
    $self->ResMgr::add_resource(-name => "switch_info",
                                -type => "_GLOBALS",
                                -obj  => $hash);
    if ($self->error()) {
        $self->_set_err("Unable to add resource 'switch_info' to the Resource Manager");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    background_execute()
# Objective: Invoke an arbitrary piece of code in the background repeatedly
#            for a specified amount of time.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub background_execute {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $runid = &ParamsMgr::get_arg(-arg => "runid",
                                    -required => 1);
    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => 0);
    my $message = &ParamsMgr::get_arg(-arg => "message",
                                      -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
    return undef;
    }

    # obtain the current running event, and update the runid pad
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $runid_prefix = $event_obj->increment_subtest_runid_counter();

    if (length($message)) {
        &SanUtil::nate_log(-comment, "$message");
    }

    my $subtest = subtest(
        sub {

            my ($fault_state, $function_ptr);
            my ($rc, $start_time, $execute_count, $temp_var) = (0, time(), 0);

            &ParamsMgr::process_args(%args);

            my $total_exec_time = &ParamsMgr::get_arg(-arg => "total_exec_time",
                                                      -required => 1);
            my $execute_frequency = &ParamsMgr::get_arg(-arg => "execute_frequency",
                                                        -required => 1);
            my $initial_sleep_time = &ParamsMgr::get_arg(-arg => "initial_sleep_time",
                                                         -default => 0);
            my $exec = &ParamsMgr::get_arg(-arg => "exec",
                                           -required => 1,
                                           -ref_type => "CODE");
            my $init = &ParamsMgr::get_arg(-arg => "init",
                                           -default => undef,
                                           -ref_type => "CODE");
            my $complain_exec = &ParamsMgr::get_arg(-arg => "complain_exec",
                                                    -default => undef,
                                                    -ref_type => "CODE");
            if (&ParamsMgr::error()) {
                $self->_set_err(&ParamsMgr::get_error());
                $self->SubtestComm::subtest_fail();
                return $EventUtil::FALSE;
            }

            if ($initial_sleep_time > 0) {
                # sleep the initial sleep time if specified
                &SanUtil::nate_log('-comment', "Sleeping $initial_sleep_time seconds before running '$runid'");
                &SanUtil::nate_sleep($initial_sleep_time);

                # subtract the sleep time we already took from the total exec time
                # only if not asked to run forever
                if ($total_exec_time !~ /forever/i) {
                    $total_exec_time -= $initial_sleep_time;
                }
                $temp_var = time() - $start_time;
            }

            # if an init method was specified, run it here
            if (defined($init)) {
                &$init();
            }

            # if requested to loop forever set the flag
            my $loop_forever = 0;
            if ($total_exec_time =~ /forever/i) {
                $loop_forever = 1;
                $total_exec_time = $execute_frequency;
            }

            while ($total_exec_time > 0 or $loop_forever) {
                # Sleep 1 second and remove that amount from TOTAL_EXEC_TIME
                &SanUtil::nate_sleep(1);
                $total_exec_time--;

                # Check to see if it is time to execute the function given to us
                if ($total_exec_time % $execute_frequency == 0) {
                    $temp_var = time();
                    my $rc_local = &$exec();
                    $rc += $rc_local;

                    # check to see if the function they specified failed or not
                    if ($rc_local == $EventUtil::FALSE) {
                        # execute the complain exec method if specified
                        if (defined($complain_exec)) {
                            &$complain_exec();
                        }
                        $self->_set_err("Encountered failure in '$runid' method");
                        $self->SubtestComm::subtest_fail();
                    }

                    # increment the total number of times we invoked this method
                    $execute_count++;

                    # Calculate the time it took to run function, then subtract
                    # that from total_exec_time
                    $total_exec_time -= (time() - $temp_var);

                    my $time_elapsed = time() - $start_time;
                    if ($loop_forever) {
                        &SanUtil::nate_log('-comment', "Elapsed time since call of '$runid': '$time_elapsed' second(s)\n" .
                                                       "Time remaining: infinite\n" .
                                                       "Return Code Count: $rc\n" .
                                                       "Execution Count: $execute_count");
                    } else {
                        &SanUtil::nate_log('-comment', "Elapsed time since call of '$runid': '$time_elapsed' second(s)\n" .
                                                       "Time remaining: '$total_exec_time' second(s)\n" .
                                                       "Return Code Count: $rc\n" .
                                                       "Execution Count: $execute_count");
                    }
                }
            }

            # calculate how long it took
            $temp_var = time() - $start_time;

            &SanUtil::nate_log('-comment', "Method Complete\n" .
                                           "Elapsed time since call of '$runid': '$temp_var' second(s)\n" .
                                           "Return Code Count: $rc\n" .
                                           "Execution Count: $execute_count");

            $self->SubtestComm::subtest_complete();
        }, -runid, "${runid_prefix}_${runid}", -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

    # reap the subtest if async is off
    if ($async == 0) {
        # wait for the background subtest method to terminate
        # if not specified for async access
        $self->wait_for_subtests(-subtests => [$subtest],
                                 -propagate_wait => 0,
                                 -propagate_events => 1,
                                 -enable_sync => 0,
                                 -check_cmds_from_paranet => 0,
                                 -forward_cmds_from_parent => 0,
                             -async => 0);
        if ($self->error()) {
            return $subtest;
        }
    }

    return $subtest;
}

###############################################################################
# Method:    gather_resource_info
# Objective: Spawn off subtests to gather data on all eventman resources
# Details:   See POD documentation at the end of this file.
###############################################################################
sub gather_resource_info {
    my ($self) = @_;

    # skip gather resource info if not requested to do so by the user
    if ($self->EventUtil::get_arg(-arg => "gather_resource_info") eq $EventUtil::FALSE) {
        return $EventUtil::TRUE;
    }

    # ignore any errors encountered by the gather subtest
    my $gather_subtest =
        subtest(sub {
                    my @subtests;
                    # get all sanclients
                    my @sanclients = $self->EventUtil::get_sanclients(-faulted => $EventUtil::FALSE);
                    if ($self->error()) {
                        $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                        return $EventUtil::FALSE;
                    }
                    # get all sanswitches
                    my @sanswitches = $self->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE);
                    if ($self->error()) {
                        $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                        return $EventUtil::FALSE;
                    }

                    my @nacl_clusters = $self->EventUtil::get_nacl_clusters_cm();
                    if ($self->error()) {
                        $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                        return $EventUtil::FALSE;
                    }
                    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE);
                    if ($self->error()) {
                        $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                        return $EventUtil::FALSE;
                    }

                    foreach my $cluster_name (@nacl_clusters) {
                        # spawn off subtests for every nacl node cm resource registered
                        my $resource = $self->ResMgr::get_resource(-name => $cluster_name,
                                                                   -type => "nacl_clusters_cm");
                        if ($self->error()) {
                            $self->_set_err("Could not obtain resource '$cluster_name : nacl_clusters_cm'");
                            return $EventUtil::FALSE;
                        }
                        next if (not $self->EventUtil::get_arg(-arg => "gather_resource_info",
                                                               -objects => [$resource],
                                                               -default_to_env => 1));

                        my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                                          -cluster_filter => ['$cluster_name']);

                        if ($self->error()) {
                            $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                            return $EventUtil::FALSE;
                        }
                        my $subtest = subtest(
                            sub {
                                $self->EventUtil::_gather_resource_info_nacl_cluster(cluster_name => $cluster_name,
                                                                                     nodes        => \@nacl_nodes);

                                if ($self->error()) {
                                    $self->_set_err("Failed to gather resource info on nacl_cluster_cm resource '$cluster_name'");
                                    $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                                    return $EventUtil::FALSE;
                                }
                                $self->SubtestComm::subtest_complete();
                                return $EventUtil::TRUE;
                            },
                            -runid, "cluster__${cluster_name}", -bg,
                            "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
                        push(@subtests, $subtest);
                    }

                    # spawn off subtests for every sanclient resource registered
                    foreach my $client_name (@sanclients) {
                        # skip this client if gather_resource_info was set to false
                        my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                                                   -type => "sanclients");
                        if ($self->error()) {
                            $self->_set_err("Could not obtain resource '$client_name : sanclients'");
                            return $EventUtil::FALSE;
                        }
                        next if (not $self->EventUtil::get_arg(-arg => "gather_resource_info",
                                                               -objects => [$resource],
                                                               -default_to_env => 1));

                        my $subtest = subtest(
                            sub {
                                $self->EventUtil::_gather_resource_info_sanclient(-client_name => $client_name);
                                if ($self->error()) {
                                    $self->_set_err("Failed to gather resource info on 'sanclients' resource '$client_name'");
                                    $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                                    return $EventUtil::FALSE;
                                }
                                $self->SubtestComm::subtest_complete();
                                return $EventUtil::TRUE;
                            },
                            -runid, "client__${client_name}", -bg,
                            "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
                        push(@subtests, $subtest);
                    }

                    # spawn off subtests for every data switch registered
                    foreach my $switch_name (@sanswitches) {
                        # skip this switch if gather_resource_info was set to false
                        my $resource = $self->ResMgr::get_resource(-name => $switch_name,
                                                                   -type => "sanswitches");
                        if ($self->error()) {
                            $self->_set_err("Could not obtain resource '$switch_name : sanswitches'");
                            return $EventUtil::FALSE;
                        }
                        next if (not $self->EventUtil::get_arg(-arg => "gather_resource_info",
                                                               -objects => [$resource],
                                                               -default_to_env => 1));

                        my $subtest = subtest(
                            sub {
                                $self->EventUtil::_gather_resource_info_sanswitch(-switch_name => $switch_name);
                                if ($self->error()) {
                                    $self->_set_err("Failed to gather resource info on 'sanswitches' resource '$switch_name'");
                                    $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                                    return $EventUtil::FALSE;
                                }
                                $self->SubtestComm::subtest_complete();
                                return $EventUtil::TRUE;
                            },
                            -runid, "switch__${switch_name}", -bg,
                            "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

                        push(@subtests, $subtest);
                    }

                    # reap the subtests
                    $self->wait_for_subtests(-subtests => \@subtests);
                    if ($self->error()) {
                        $self->_set_err("Failed to gather resource info");
                        $self->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);
                        return $EventUtil::FALSE;
                    }

                    $self->SubtestComm::subtest_complete();
                    return $EventUtil::TRUE;
                },
                -runid, "resource_info", -bg,
                "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

    # reap the gather info subtest
    $self->wait_for_subtests(-subtests => [$gather_subtest]);
    if ($self->error()) {
        $self->_set_err("Failed to gather resource info");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _gather_resource_info_sanclient
# Objective: display info for a single sanclient obj
###############################################################################
sub _gather_resource_info_sanclient {
    my ($self, %args) = @_;

    # get the sanclient object
    my $obj = $self->EventUtil::get_sanclient_obj(%args);
    if ($self->error()) {
        $self->_set_err("Failed to obtain sanclient object");
        return $EventUtil::FALSE;
    }

    # run display info (ignore errors)
    $obj->display_info();
    $obj->_clear_err();

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _gather_resource_info_sanswitch
# Objective: display info for a single sanswitch obj
###############################################################################
sub _gather_resource_info_sanswitch {
    my ($self, %args) = @_;

    # get the sanclient object
    my $obj = $self->EventUtil::get_sanswitch_obj(%args);
    if ($self->error()) {
        $self->_set_err("Failed to obtain sanswitch object");
        return $EventUtil::FALSE;
    }

    # run display info (ignore errors)
    $obj->display_info();
    $obj->_clear_err();

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _gather_resource_info_nacl_cluster
# Objective: display info for a single cluster object
###############################################################################
sub _gather_resource_info_nacl_cluster {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $cluster_name = &ParamsMgr::get_arg(-arg      => 'cluster_name',
                                           -required => 1);
    my @node_names   = &ParamsMgr::get_arg(-arg      => 'nodes',
                                           -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return undef;
    }
    # get the nacl node object
    my $cluster = $self->EventUtil::get_nacl_cluster_cm_obj(%args);
    if ($self->error()) {
        $self->_set_err("Failed to obtain cluster object");
        return $EventUtil::FALSE;
    }

    my @aggregates;
    try {
        @aggregates = NACL::C::StorageAggregate->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @aggregates = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $msg .= "    Cluster summary for $cluster_name\n\n";
    $msg .= "    Cluster aggregate count: " .  @aggregates . "\n\n";

    my @vservers;
    try {
        @vservers = NACL::C::Vserver->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @vservers = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "    Cluster vserver count: " .  @vservers . "\n\n";

    my @luns;
    try {
        @luns = NACL::C::Lun->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @luns = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "    Cluster LUN count: " .  @luns . "\n\n";

    my @volumes;
    try {
        @volumes = NACL::C::Volume->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @volumes = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "    Cluster volume count: " .  @volumes . "\n\n";

    my @cluster_lifs;
    my %lifs_by_node;
    foreach my $node_name (@node_names) {
        try {
            my @node_lifs = NACL::C::NetworkInterface->find(command_interface => $cluster,
                                                            filter => {'curr-node' => $node_name});
            $lifs_by_node{$node_name} = @node_lifs;

        } otherwise {
            my $exception = shift;
            if ($exception->text() =~ /No matching/i) {
                $lifs_by_node{$node_name} = ();
            } else {
                $self->_set_err("Exception caught: " . $exception->text());
            }
        };
    }
    try {
        @cluster_lifs = NACL::C::NetworkInterface->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @cluster_lifs = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "    Cluster Lif count: " .  @cluster_lifs . "\n";
    foreach my $node_name (@node_names) {
        $msg .= "    Node $node_name lif count: " .  $lifs_by_node{$node_name} . "\n";
        if (@cluster_lifs > 0) {
            my $lif_percent = sprintf '%.0f', 100 * $lifs_by_node{$node_name} / @cluster_lifs;
            $msg .= "    Node $node_name percentage of cluster lifs " .  $lif_percent . "\n";
        }
    }

    my @cluster_ifgrps;
    my %ifgrps_by_node;
    foreach my $node_name (@node_names) {
        try {
            my @node_ifgrps = NACL::C::NetworkPortIfgrp->find(command_interface => $cluster,
                                                              filter => {'node' => $node_name});
            $ifgrps_by_node{$node_name} = @node_ifgrps;
        } otherwise {
            my $exception = shift;
            if ($exception->text() =~ /No matching/i) {
                $ifgrps_by_node{$node_name} = ();
            } else {
                $self->_set_err("Exception caught: " . $exception->text());
            }
        };
    }
    try {
        @cluster_ifgrps = NACL::C::NetworkPortIfgrp->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @cluster_ifgrps = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "\n    Cluster Ifgrp count: " .  @cluster_ifgrps . "\n";
    foreach my $node_name (@node_names) {
        my $node_ifgrp_count;
        if (! defined($ifgrps_by_node{$node_name})) {
            $node_ifgrp_count = 0;
        } else {
            $node_ifgrp_count = $ifgrps_by_node{$node_name};
        }
        $msg .= "    Node $node_name ifgrp count: " .  $node_ifgrp_count . "\n";
        if (@cluster_ifgrps > 0) {
            my $ifgrp_percent = sprintf '%.0f', 100 * $node_ifgrp_count / @cluster_ifgrps;
            $msg .= "    Node $node_name percentage of cluster ifgrps " .  $ifgrp_percent . "%\n";
        }
    }

    my @cluster_vlans;
    my %vlans_by_node;
    foreach my $node_name (@node_names) {
        try {
            my @node_vlans = NACL::C::NetworkPortVlan->find(command_interface => $cluster,
                                                            filter => {'node' => $node_name});
            $vlans_by_node{$node_name} = @node_vlans;
        } otherwise {
            my $exception = shift;
            if ($exception->text() =~ /No matching/i) {
                $vlans_by_node{$node_name} = ();
            } else {
                $self->_set_err("Exception caught: " . $exception->text());
            }
        };
    }
    try {
        @cluster_vlans = NACL::C::NetworkPortVlan->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @cluster_vlans = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "\n    Cluster Vlan count: " .  @cluster_vlans . "\n";
    foreach my $node_name (@node_names) {
        my $node_vlan_count;
        if (! defined($vlans_by_node{$node_name})) {
            $node_vlan_count = 0;
        } else {
            $node_vlan_count = $vlans_by_node{$node_name};
        }
        $msg .= "    Node $node_name vlan count: " .  $node_vlan_count . "\n";
        if (@cluster_vlans > 0) {
            my $vlan_percent = sprintf '%.0f', 100 * $node_vlan_count / @cluster_vlans;
            $msg .= "    Node $node_name percentage of cluster vlans " .  $vlan_percent . "%\n";
        }
    }

    my @cluster_igroups;
    try {
        @cluster_igroups = NACL::C::LunIgroup->find(command_interface => $cluster);
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching/i) {
            @cluster_igroups = ();
        } else {
            $self->_set_err("Exception caught: " . $exception->text());
        }
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $msg .= "\n    Cluster Igroup count: " .  @cluster_igroups . "\n";

    # log the information
    &SanUtil::nate_log(-comment, $msg);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _save_error()
# Objective: Save the error on the object (in otherwords, reset the error but
#            dont delete it.
###############################################################################
sub _save_error {
    my ($self) = @_;

    if ($self->error()) {
        $self->{_saved_error} = delete($self->{error});
        push(@{$self->{_saved_errmsg}}, @{delete($self->{errmsg})});
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _restore_error()
# Objective: Restore the saved error that may or may not be on the object
###############################################################################
sub _restore_error {
    my ($self) = @_;

    if (exists($self->{_saved_error})) {
        $self->{error} = delete($self->{_saved_error});
        $self->{errmsg} = delete($self->{_saved_errmsg});
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    # set up default event parameter values for this utility module
    $self->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                       -arg_defaults => {
                                                           -gather_resource_info => $EventUtil::TRUE
                                                       });
    if ($self->error()) {
        $self->_set_err("Failed to register utility module defaults for " . __PACKAGE__);
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __init {
    my ($self) = @_;

    # create fresh connections for all our resources
    # we need to do this because we dont want to trample on
    # the resource connections that we did not create.
    $self->EventUtil::create_new_resource_connections(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Failed to create new resource connections");
        return $EventUtil::FALSE;
    }

    # spawn off subtests to gather data for each resource
    $self->EventUtil::gather_resource_info();
    if ($self->error()) {
        $self->_set_err("Failed to gather resource info");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __uninit {
    my ($self) = @_;

    # disconnect all resource connections
    $self->EventUtil::disconnect_all_resource_connections();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

EventUtil.pm

=head1 Author

Robert Hunter

=head1 Description

An EventMan Internal Utility Module

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head1 B<Public Methods>

=head2 B<get_arg(%args)>

Returns the first Event Parameter value found among the list of
EventMan/Event/EventSet/EventSetGroup/Resource objects specified

=over 4

=item Inputs

=item *I<-objects>

Optional: An array reference of EventMan/Event/EventSet/EventSetGroup/Resource objects
to search through when looking for the specified Event parameter.
Default: The current environment, ie: [$curr_event_obj, $curr_eventset_obj, $curr_eventman_obj]

=item *I<-arg>

Required: The Event parameter to search for.

=item *I<-default>

Optional: The default value to fall back on if no value is found (in this order)
(1) among -objects,
(2) if no default value is found among the environment (when -default_to_env is set)
(3) if no value is found among the current event module hierarchy event param defaults.
Default: undef

=item *I<-default_to_env>

Optional: If set to $EventUtil::TRUE, will default to the current environment
[$curr_event_obj, $curr_eventset_obj, $curr_eventman_obj] when searching for -arg
Default: $EventUtil::FALSE (off)

=item *I<-no_event_module_defaults>

Optional: Disables event module defaults lookup and returns the value of
-default if no value for -arg is found
Default: $EventUtil::FALSE (off)

=item Outputs

Returns the value of the Event parameter if found, undef otherwise

=item Examples

 # get the value of path validation from the current environment, default to off if not found
 my $path_validation_value = $eventman_obj->EventUtil::get_arg(-arg => "path_validation",
                                                                -default => $EventUtil::FALSE);

 ...

 # get the resource object from the resource manager
 my $resource = $self->ResMgr::get_resource(-name => "fas3070-rtp07",
                                            -type => "sanfilersCm");
 if ($self->error()) {
    return $EventUtil::FALSE;
 }

 # if sram dump detection is enabled on this controller, then set it up
 my $sram_detection = $self->EventUtil::get_arg(-arg => "sram_detection",
                                                -objects => [$resource],
                                                -default => $EventUtil::TRUE,
                                                -default_to_env => $EventUtil::TRUE);

=back

=head2 B<display_event_parameters()>

Return a summary of all the event parameters/resources loaded in the current environment

=over 4

=item Inputs

=item *I<-objects>

Optional: An array reference of EventMan/Event/EventSet/EventSetGroup/Resource objects
to display Event Parameters
Default: [$curr_eventman_obj]

=item Outputs

Returns an array or array reference containing the Summary in a pretty format

=item Examples

 my @summary = $eventman_obj->EventUtil::display_event_parameters(-objects => [$eventman_obj, $curr_event_set, $curr_event]);

=back

=head2 B<get_sanclients()>

Return a list of all sanclients in the current environment.

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing SanClient resource names in the current environment

=item Examples

 my @unfaulted_sanclient_names = $eventman_obj->EventUtil::get_sanclients(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_sanfilers()>

Return a list of all sanfilers in the current environment.

=over 4

=item Inputs

=item *I<-type>

Optional: Specifies which type of SanFilers to get (7mode,cmode, or all)
Default: all (both cmode/7mode resources)

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-fall_back_on_controller_clusters>

Optional: If set to $EventUtil::TRUE, will search the controllercluster objects for sanfiler resources if none are found
Default: $EventUtil::FALSE

=item Outputs

Returns an array or array reference containing SanFiler resource names in the current environment

=item Examples

 # get unfaulted sanfilers running 7mode
 my @unfaulted_sanfiler7m_names = $eventman_obj->EventUtil::get_sanfilers(-faulted => $EventUtil::FALSE,
                                                                           -type => "7mode");

=back

=head2 B<get_nacl_nodes()>

Return a list of all nacl_nodes in the current environment

=over 4

=item Inputs

=item *I<-type>

Optional: Specifies which type of nacl_node to get (7mode, cmode, or all)
Default: all (both cmode/7mode resources)

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-fall_back_on_nacl_clusters>

Optional: If set to $EventUtil::TRUE, will search the nacl_clusters_cm objects for nacl_nodes_cm resources if none are found
Default: $EventUtil::FALSE

=item *I<-clusters>

Optional: An array reference of cluster names to filter the node list by
Default: []

=item Outputs

Returns an array or array reference containing nacl_nodes_cm resource names in the current environment

=item Examples

 # get unfaulted nacl_nodes running cmode
 my @unfaulted_nacl_node_names = $eventman_obj->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                                          -type => "cmode");

=back

=head2 B<get_nacl_node_objs()>

Return a list of all NACL::C::Node objects in the current event scope

=over 4

=item Inputs

=item *I<-type>

Optional: Specifies which type of nacl_node to get (7mode, cmode, or all)
Default: all (both cmode/7mode resources)

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-fall_back_on_nacl_clusters>

Optional: If set to $EventUtil::TRUE, will search the nacl_clusters_cm objects for nacl_nodes_cm resources if none are found
Default: $EventUtil::FALSE

=item *I<-clusters>

Optional: An array reference of cluster names to filter the node list by
Default: []

=item Outputs

Returns an array or array reference containing NACL::C::Node objects in the environemnt

=item Examples

 # get unfaulted NACL::C::Node objects in the cluster "mycluster"
 my @unfaulted_nacl_nodes = $eventman_obj->EventUtil::get_nacl_node_objs(-faulted => $EventUtil::FALSE,
                                                                         -clusters => ['mycluster']);

=back

=head2 B<get_controllerclusters()>

Return a list of all Controller Clusters in the current environment.

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference containing ControllerCluster resource names in the current environment

=item Examples

 my @cluster_names = $eventman_obj->EventUtil::get_controllerclusters();

=back

=head2 B<get_nacl_clusters_cm()>

Return a list of all nacl_clusters_cm resources in the current environment.

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference containing nacl_clusters_cm resource names in the current environment

=item Examples

 my @cluster_names = $eventman_obj->EventUtil::get_nacl_clusters_cm();

=back

=head2 B<get_vservers()>

Return a list of all Vserver names in the current environment

=over 4

=item Inputs

=item *I<-clusters>

Optional: An array reference of cluster names to filter the vserver list by
Default: []

=item Outputs

Returns an array or array reference containing Vserver resource names in the current environment

=item Examples

 # get the list of vserver resources in this cluster mycluster
 my @vserver_names = $eventman_obj->EventUtil::get_vservers(-clusters => ['mycluster']);

=back

=head2 B<get_sanswitches()>

Return a list of all sanswitches in the current environment.

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing SanSwitch resource names in the current environment

=item Examples

 my @unfaulted_sanswitch_names = $eventman_obj->EventUtil::get_sanswitches(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_fabrics()>

Return a list of all fabrics in the current environment

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing Fabric resource names in the current environment

=item Examples

 my @unfaulted_fabric_names = $eventman_obj->EventUtil::get_fabrics(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_sanclient_obj(%args)>

Return the SanClient object specified by -client_name

=over 4

=item Inputs

=item *I<-client_name>

Required: The name of the sanclient object to retrieve from the resource manager.

=item Outputs

Returns a SanClient object if successful, sets error if not.

=item Examples

 my $sc_obj = $eventman_obj->EventUtil::get_sanclient_obj(-client_name => "myclient1");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_sanfiler_obj(%args)>

Return the SanFiler object specified by -filer_name

=over 4

=item Inputs

=item *I<-filer_name>

Required: The name of the sanfiler object to retrieve from the resource manager.

=item Outputs

Returns a SanFiler object if successful, sets error if not.

=item Examples

 my $sf_obj = $eventman_obj->EventUtil::get_sanfiler_obj(-filer_name => "myfiler1");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_nacl_node_obj(%args)>

Return the nacl_nodes object specified by -node_name

=over 4

=item Inputs

=item *I<-node_name>

Required: The name of the nacl_node object to retrieve from the resource manager.

=item Outputs

Returns a NACL::C::Node object if successful, sets error if not.

=item Examples

 my $nacl_node_obj = $eventman_obj->EventUtil::get_nacl_node_obj(-node_name => "my_node1");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_controllercluster_obj(%args)>

Return the ControllerCluster object specified by -cluster_name

=over 4

=item Inputs

=item *I<-cluster_name>

Required: The name of the ControllerCluster object to retrieve from
the resource manager.

=item Outputs

Returns a ControllerCluster object if successful, sets error if not.

=item Examples

 my $cluster_obj = $eventman_obj->EventUtil::get_controllercluster_obj(-cluster_name => "3070_cluster");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_nacl_cluster_cm_obj(%args)>

Return the nacl_clusters_cm object specified by -cluster_name

=over 4

=item Inputs

=item *I<-cluster_name>

Required: The name of the nacl_clusters_cm object to retrieve from
the resource manager.

=item Outputs

Returns a nacl_clusters_cm object if successful, sets error if not.

=item Examples

 my $cluster_obj = $eventman_obj->EventUtil::get_nacl_cluster_cm_obj(-cluster_name => "3070_cluster");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_vserver_obj(%args)>

Return the Vserver object specified by -vserver_name

=over 4

=item Inputs

=item *I<-vserver_name>

Required: The name of the Vserver object to retrieve from
the resource manager.

=item *I<-cluster_name>

Required: The name of the cluster that the vserver belongs to.

=item Outputs

Returns a Vserver object if successful, sets error if not.

=item Examples

 my $vserver_obj = $eventman_obj->EventUtil::get_vserver_obj(-vserver_name => "vserver0",
                                                             -cluster_name => "mycluster");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_sanswitch_obj(%args)>

Return the SanSwitch object specified by -switch_name

=over 4

=item Inputs

=item *I<-switch_name>

Required: The name of the sanswitch object to retrieve from the resource manager.

=item Outputs

Returns a SanSwitch object if successful, sets error if not.

=item Examples

 my $sc_obj = $eventman_obj->EventUtil::get_sanswitch_obj(-switch_name => "myswitch1");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_sanclient_objs()>

Return a list of sanclient objects

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing SanClient objects in the current environment

=item Examples

 my @unfaulted_sanclient_objs = $eventman_obj->EventUtil::get_sanclient_objs(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_sanfiler_objs()>

Return a list of sanfiler objects

=over 4

=item Inputs

=item *I<-type>

Optional: Specifies which type of SanFilers to get (7mode,cmode, or all)
Default: all (both cmode/7mode resources)

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing SanFiler objects in the current environment

=item Examples

 my @unfaulted_sanfiler_objs = $eventman_obj->EventUtil::get_sanfiler_objs(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_sanswitch_objs()>

Return a list of sanswitch objects

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing SanSwitch objects in the current environment

=item Examples

 my @unfaulted_sanswitch_objs = $eventman_obj->EventUtil::get_sanswitch_objs(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_controllercluster_objs()>

Return a list of controllercluster objects

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

Returns an array or array reference containing ControllerCluster objects in the current environment

=item Examples

 my @unfaulted_controllercluster_objs = $eventman_obj->EventUtil::get_controllercluster_objs(-faulted => $EventUtil::FALSE);

=back

=head2 B<get_nacl_cluster_cm_objs()>

Return a list of nacl_clusters_cm objects

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference containing nacl_clusters_cm objects in the current environment

=item Examples

 my @nacl_cluster_cm_objs = $eventman_obj->EventUtil::get_nacl_cluster_cm_objs();

=back

=head2 B<get_vserver_objs()>

Return a list of vserver objects

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not return resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-clusters>

Required: an array reference of cluster names to get vserver objects from

=item Outputs

Returns an array or array reference containing Vserver objects in the current environment

=item Examples

 my @unfaulted_vserver_objs = $eventman_obj->EventUtil::get_vserver_objs(-faulted => $EventUtil::FALSE, -clusters => ['clus1', 'clus2']);

=back

=head2 B<reconnect_all_sanclients(%args)>

Reconnect all SanClient resources

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not reconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-client_names>

Optional: An array hash of client names to dreconnect connections for
Default: All client names in current Event Scope (that meet -faulted criteria)

=item *I<-disconnect_existing_conn>

Optional: If set to $EventUtil::TRUE, will disconnect any existing connection on the resource objects.
Setting this to $EventUtil::FALSE will leave the previous connection intact.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::reconnect_all_sanclients(-faulted => $EventUtil::FALSE);
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<reconnect_all_sanfilers(%args)>

Reconnect all SanFiler resources

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not reconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-type>

Optional: Specifies which type of SanFilers to get (7mode,cmode, or all)
Default: all (both cmode/7mode resources)

=item *I<-controller_names>

Optional: An array reference of controller names to reconnect connections for
Default: All controller names in current Event Scope (that meet -type/-faulted criteria)

=item *I<-disconnect_existing_conn>

Optional: If set to $EventUtil::TRUE, will disconnect any existing connection on the resource objects.
Setting this to $EventUtil::FALSE will leave the previous connection intact.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::reconnect_all_sanfilers(-faulted => $EventUtil::FALSE);
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<reconnect_all_cservers(%args)>

Reconnect all Cserver resources

=over 4

=item Inputs

=item *I<-cluster_names>

Optional: An array reference of cluster names to reconnect connections for
Default: All cluster names in current Event Scope (that meet -type/-faulted criteria)

=item *I<-disconnect_existing_conn>

Optional: If set to $EventUtil::TRUE, will disconnect any existing connection on the resource objects.
Setting this to $EventUtil::FALSE will leave the previous connection intact.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::reconnect_all_cservers();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<reconnect_all_sanswitches(%args)>

Reconnect all SanFiler resources

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not reconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-switch_names>

Optional: An array reference of switch names to reconnect connections for
Default: All switch names in current Event Scope (that meet -faulted criteria)

=item *I<-disconnect_existing_conn>

Optional: If set to $EventUtil::TRUE, will disconnect any existing connection on the resource objects.
Setting this to $EventUtil::FALSE will leave the previous connection intact.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::reconnect_all_sanswitches(-faulted => $EventUtil::FALSE);
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<disconnect_all_sanclients(%args)>

Disconnect all SanClient resources

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not disconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-client_names>

Optional: An array reference of client names to disconnect connections for
Default: All switch names in current Event Scope

=item *I<-ignore_errors>

Optional: Indicates whether to ignore errors within the function
Default: $EventUtil::FALSE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::disconnect_all_sanclients();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<disconnect_all_sanfilers(%args)>

Disconnect all SanFiler resources

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not disconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-type>

Optional: Specifies which type of SanFilers to get (7mode,cmode, or all)
Default: all (both cmode/7mode resources)

=item *I<-controller_names>

Optional: An array reference of controller names to disconnect connections for
Default: All controller names in current Event Scope (that meet -type/-faulted criteria)

=item *I<-ignore_errors>

Optional: Indicates whether to ignore errors within the function
Default: $EventUtil::FALSE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::disconnect_all_sanfilers();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<disconnect_all_cservers(%args)>

Disconnect all Cserver resources

=over 4

=item Inputs

=item *I<-cluster_names>

Optional: An array reference of cluster names to disconnect connections for
Default: All cluster names in current Event Scope (that meet -type/-faulted criteria)

=item *I<-ignore_errors>

Optional: Indicates whether to ignore errors within the function
Default: $EventUtil::FALSE

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::disconnect_all_cserverss();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<disconnect_all_sanswitches(%args)>

Disconnect all SanSwitch resources

=over 4

=item Inputs

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not disconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item *I<-ignore_errors>

Optional: Indicates whether to ignore errors within the function
Default: $EventUtil::FALSE

=item *I<-switch_names>

Optional: An array reference of switch names to disconnect connections for
Default: All switch names in current Event Scope

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::disconnect_all_sanswitches();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<create_new_resource_connections(%args)>

Create fresh resource connections in the ResMgr without destroying the previous connection
on the resources.

=over 4

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not connect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::create_new_resource_connections(-faulted => $EventUtil::FALSE);
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<reconnect_to_previously_faulted_resources()>

Reconnect all resource connections in the ResMgr that have been previously faulted

=over 4

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::reconnect_to_previously_faulted_resources();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<reconnect_all_resource_connections(%args)>

Reconnect all resource connections in the ResMgr

=over 4

=item *I<-faulted>

Optional: If set to $EventUtil::FALSE, will not reconnect resources that are in faulted states, $EventUtil::TRUE is the opposite
Default: -1 (both faulted/nonfaulted)

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::reconnect_all_resource_connections(-faulted => $EventUtil::FALSE);
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<disconnect_all_resource_connections(%args)>

Disconnect all resource connections in the ResMgr

=over 4

=item Inputs

See inputs for:
EventUtil::disconnect_all_sanfilers(),
EventUtil::disconnect_all_sanclients(),
EventUtil::disconnect_all_sanswitches()

=item Outputs

$EventUtil::TRUE if succesful, $EventUtil::FALSE if not and sets error

=item Examples

 $eventman_obj->EventUtil::disconnect_all_resource_connections();
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<synchronize()>

Synchronize the current running subtest with All sibling subtests that are currently running.

=over 4

=item Inputs

None

=item Outputs

The time (in seconds) that it took to synchronize this subtest with its siblings.

=item Examples

 $eventman_obj->EventUtil::synchronize();

=back

=head2 B<document_event_param(%args)>

Document a specific Event Parameter used in the EventMan package.

=over 4

=item Inputs

=item *I<-param>

The name of the parameter to document

=item *I<-package>

The name of the package that is documenting this parameter

=item Outputs

None

=item Examples

 $eventman_obj->EventUtil::document_event_param(-param => "param1",
                                                 -package => "San::EventMan::...",
                                                 -description => "this is param1",
                                                 -example => "...");

=back

=head2 B<register_event(%args)>

Register an Event with the EventMan package

=over 4

=item Inputs

=item *I<-event_name>

Required: The event keyword (must be unique)

=item *I<-description>

Optional: A brief description of what this event will do.
Default: ""

=item *I<-enabled>

Optional: Denotes whether this event is enabled or disabled by the Event Module
Default: $EventUtil::TRUE

=item *I<-arg_defaults>

Optional: A hash reference of key value pairs representing default event parameter
values specifically for this event.
Default: {}

=item *I<-required_resource_types>

Optional: An array reference of resource types that this event requires to
successfully execute.
Default: []

=item *I<-require_at_least_one_resource_types>

Optional: An array reference of resource types that at least one of them
must be represented in order for the event to successfully execute.
Default: []

=item *I<-optional_resource_types>

Optional: An array reference of resource types that  are
not required, but are used by this event if available.
Default: []

=item *I<-aliases>

Optional: An array reference of event name aliases to attach to this event.
Default: []

=item *I<-event_modules>

Required: An array reference of event modules that will dynamically be pushed
onto the @ISA array of the EventMan object when the event is executed. Once
the event completes, the @ISA array will be reverted to its original value.

=item *I<-exec>

Required: A subroutine reference that will be invoked when the event is executed

=item *I<-init>

Optional: A subroutine reference that will be invoked before the prejob and exec methods.
This cannot be overridden by the user of the library.

=item *I<-prejob>

Optional: A subroutine reference that will be invoke after the init method but before the exec method.
This can be overridden by the user of the library and can be enabled or disabled.

=item *I<-cojob>

Optional: A subroutine reference that will be invoked in parallel with the exec method
This can be overridden by the user of the library and can be enabled or disabled.

=item *I<-postjob>

Optional: A subroutine reference that will be invoked after the exec method but before the uninit method.
This can be overridden by the user of the library and can be enabled or disabled.

=item *I<-uninit>

Optional: A subroutine reference that will be invoked after the postjob method.
This cannot be overridden by the user of the library.

=item *I<-validate>

Optional: A subroutine reference that will be invoked once an Event object for
this event is created. This method can be enabled or disabled, but not overridden
by the user of the library.

=item *I<-pre_validate>

Optional: A subroutine reference that will be invoked once an Event object for
this event is created. This method runs before the validation methods are
run.

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Examples

 # register this event
 $eventman_obj->EventUtil::register_event(-event_name => "myfault",
                                           -event_modules => [__PACKAGE__, "EventMan::EventModules::..."],
                                           -description => "this is my event",
                                           -aliases => ['myfault_alias1', 'myfault_alias2'],
                                           -optional_resource_types => ["sanclients"];
                                           -required_resource_types => ["nacl_nodes_cm", "sanfilers7m"];
                                           -exec => sub { &myfault(@_); },
                                           -prejob => sub { print "prejob"; },
                                           -cojob => sub { print "cojob"; },
                                           -postjob => sub { print "postjob"; },
                                           -init => sub { print "init"; },
                                           -uninit => sub { print "uninit"; },
                                           -pre_validate => sub { print "pre_validated"; });
                                           -validate => sub { print "validated"; });
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<register_event_module_defaults(%args)>

Define an Event Module's Event Parameter defaults.

=over 4

=item Inputs

=item *I<-event_module>

Required: The Event Module to define the defaults for, recommended to set this to
__PACKAGE__.

=item *I<-arg_defaults>

Optional: A hash reference of key value pairs where the key is the
event parameter, and the value is the default value for that parameter.

=item *I<-arg_harcoded>

Optional: A hash reference of key value pairs where the key is the
event parameter, and the value is a hardcoded value for the parameter

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Examples

 # register this event modules default event parameters
 $eventman_obj->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                           -arg_defaults =>
                                                               { -some_param1 => "value1",
                                                                 -some_param2 => "value1"},
                                                           -arg_hardcoded =>
                                                               { -some_param3 => "value1",
                                                                 -some_param4 => "value1"});
;
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<register_utility_module_defaults(%args)>

Define a Utility Module's Event Parameter defaults.

=over 4

=item Inputs

=item *I<-utility_module>

Required: The Utility Module to define the defaults for, recommended to set this to
__PACKAGE__.

=item *I<-arg_defaults>

Required: A hash reference of key value pairs where the key is the
event parameter, and the value is the default value for that parameter.

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Examples

 # register this utility module's default event parameters
 $eventman_obj->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                             -arg_defaults =>
                                                               { -some_param1 => "value1",
                                                                 -some_param1 => "value1"});
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<sleep_arg(%args)>

Sleep for the value specified by the requested Event Parameter

=over 4

=item Inputs

See EventUtil::get_arg() for additional inputs

=item *I<-handle_stop_messages>

Optional: If set to 1, will allow this method to poll for stop messages from the parent.
If any stop/abort/timeout messages are received, then this function will invoke all callback
methods specified in the '-exit_callback_list' input, and then gracefully terminate
the current running event.
Default: $EventUtil::FALSE

=item *I<-poll_interval>

Optional: The polling interval to use when polling for stop messages
Default: 2 (seconds)

=item *I<-exit_callback_list>

Optional: An array reference of hashes where each hash contains a required 'coderef' key and
an optional 'args' key. This callbacks will be executed in the order that they are specified
if a stop/abort/timeout message is passed in. Requires -handle_stop_messages => 1
Default: []
Example:
 # this will execute two coderefs, one with args [1,2] and theo ther with no args.
 -exit_callback_list => [ { coderef => sub {my ($args) = @_; print Dumper $args;},
                            args => [1,2]},
                          { coderef => sub {print "Hi"}}]

=item Outputs

Returns the amount of time slept if successful, error if not

=item Examples

 # sleep the amount of time given by controller_down_time
 $eventman_obj->EventUtil::sleep_arg(-arg => "controller_down_time");
 if ($eventman_obj->error()) {
     ...
 }

=back

=head2 B<get_current_event_obj()>

Obtain the current running Event object

=over 4

=item Inputs

None

=item Outputs

Returns the current running Event object

=item Examples

 # obtain the current event object that is running
 my $event_obj = $eventman_obj->EventUtil::get_current_event_obj();

=back

=head2 B<get_current_eventset_obj()>

Obtain the current running EventSet object

=over 4

=item Inputs

None

=item Outputs

Returns the current running EventSet object

=item Examples

 # obtain the current eventset object that is running
 my $eventset_obj = $eventman_obj->EventUtil::get_current_eventset_obj();

=back

=head2 B<background_execute()>

Invoke an arbitrary method in the background periodically for a specified
amount of time.

=over 4

=item Inputs

=item *I<-runid>

Required: The name to use for the subtest runid

=item *I<-async>

Optional: Control whether the method is asynchronous or not
Default: 0 (synchronous)

=item *I<-message>

Optional: Display a message to the logs prior to invoking the background process
Default: ""

=item *I<-total_exec_time>

Required: Total amount of time to execute the arbitrary method

=item *I<-execute_frequency>

Required: How often to execute the arbitrary method

=item *I<-initial_sleep_time>

Optional: How long to wait before executing the arbitrary method
Default: 0 seconds

=item *I<-init>

Optional: CODE reference (Initialization code to execute exactly once after sleeping
-initial_sleep_time and before executing -exec)

=item *I<-exec>

Required: CODE reference (The arbitrary method that will get executed in the background every so often)

=item *I<-complain_exec>

Optional: CODE reference, code to execution if the arbitrary method returns $EventUtil::FALSE

=item Outputs

Returns the the subtest object created.

=item Examples

 # print hello every 20 seconds for 100 seconds in a background subtest
 $em_obj->EventUtil::background_execute(-runid => "hello_world",
                                        -total_exec_time => 100,
                                        -execute_frequency => 20,
                                        -exec => sub { print "hello\n" });

=back

=head2 B<gather_resource_info()>

Spawn off subtests to gather information about all EventMan resources that are not faulted.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 # gather resource info
 $em_obj->EventUtil::gather_resource_info();

=back

=cut
