# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing SAN Path Validation
## @author hrivera@netapp.com
## @status private
## @pod here

package PathValidation;

# NATE packages
use Tharn;
use NATE::Log qw(log_global);

# NACL
use NACL::CS::Volume;
use NACL::CS::Lun;

# San packages
use San::SanClient::ClientUtil;
use San::SanClient::MPDevMgr;
use San::SanUtil;
use ParamsMgr;
use Data::Dumper;

my $Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    validate_paths_after_giveback()
# Objective: Issue path validation assuming a giveback was issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_paths_after_giveback {
    my ($self, %args) = @_;

    # determine if path validation on giveback is enabled in the current event/eventset
    my $path_validation_on_giveback = $self->EventUtil::get_arg(-arg => "path_validation_on_giveback");
    if ($path_validation_on_giveback == 0) {
        return $EventUtil::TRUE;
    }

    $self->PathValidation::_validate_paths(%args,
                                           -takeover_state => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Path validation after giveback failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_paths_after_takeover()
# Objective: Issue path validation assuming a takeover was issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_paths_after_takeover {
    my ($self, %args) = @_;

    # determine if path validation on takeover is enabled in the current event/eventset
    my $path_validation_on_takeover = $self->EventUtil::get_arg(-arg => "path_validation_on_takeover");
    if ($path_validation_on_takeover == 0) {
        return $EventUtil::TRUE;
    }

    $self->PathValidation::_validate_paths(%args,
                                           -takeover_state => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Path validation after takeover failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_paths_after_lif_up()
# Objective: Issue path validation assuming a lif up was issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_paths_after_lif_up {
    my ($self, %args) = @_;

    # determine if path validation after lif up is enabled in the current event/eventset
    my $path_validation_after_lif_up = $self->EventUtil::get_arg(-arg => "path_validation_after_lif_up");
    if ($path_validation_after_lif_up == 0) {
        return $EventUtil::TRUE;
    }

    # hardcode validate takeover giveback path validation rules
    $args{-path_validation_rules}->{validate_takeover_giveback} = 1;

    $self->PathValidation::_validate_paths(%args,
                                           -takeover_state => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Path validation after lif up failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_paths_after_lif_down()
# Objective: Issue path validation assuming a lif down was issued on SAN
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_paths_after_lif_down {
    my ($self, %args) = @_;

    # determine if path validation on takeover is enabled in the current event/eventset
    my $path_validation_after_lif_down =
        $self->EventUtil::get_arg(-arg => "path_validation_after_lif_down");
    if ($path_validation_after_lif_down == 0) {
        return $EventUtil::TRUE;
    }

    $self->PathValidation::_validate_paths(%args,
                                           -takeover_state => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Path validation after lif down failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_paths_after_switch_port_enable()
# Objective: Issue path validation assuming a switch port was just enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_paths_after_switch_port_enable {
    my ($self, %args) = @_;

    # determine if path validation on takeover is enabled in the current event/eventset
    my $path_validation_after_switch_port_enable =
        $self->EventUtil::get_arg(-arg => "path_validation_after_switch_port_enable");
    if ($path_validation_after_switch_port_enable == 0) {
        return $EventUtil::TRUE;
    }

    # hardcode validate takeover giveback path validation rules
    $args{-path_validation_rules}->{validate_takeover_giveback} = 1;

    $self->PathValidation::_validate_paths(%args,
                                           -takeover_state => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Path validation after port block enable failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _load_pathing_information()
# Objective: Load the initial pathing information for all SanClients
###############################################################################
sub _load_pathing_information {
    my ($self) = @_;

    my @subtests = ();

    # get the sanclient objects to work on
    my @sanclients = $self->EventUtil::get_sanclients();

    my @cluster_objs = $self->EventUtil::get_nacl_cluster_cm_objs();
    if ($self->error()) {
        $self->_set_err("Unable to obtain NACL::C::Cluster objects");
        return $EventUtil::FALSE;
    }

    # foreach client, get the names of the vservers they are related to:
    my %vs_hash = ();
    foreach my $client_name (@sanclients) {
        # get the resource object for this sanclient
        my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Could not obtain resource '$client_name' of type 'sanclients'");
            return $EventUtil::FALSE;
        }

        # get the names of the vservers that are in scope for this sanclient resource, otherwise default
        # to the environment's value of vservers
        my @vserver_names = $self->EventUtil::get_vservers(-objects => [$resource],
                                                           -default_to_env => 1);
        if ($self->error()) {
            $self->_set_err("Unable to obtain vserver names");
            return $EventUtil::FALSE;
        }
        foreach (@vserver_names) {
            $vs_hash{$_}++;
        }
    }

    # fetch volumes and luns from all clusters up front
    # and store in a datastructure that will get inherited by all the child
    # path validation load routines
    my $nacl_cluster_content = {};
    if (scalar(keys %vs_hash) and scalar(@sanclients)) {
        my @vservers = keys %vs_hash;

        # scope each fetch operation to no more than 15 vservers at a time due
        # to ONTAP/NACL/NATE getting upset when you feed it a potentially large command
        my @vs_scope_lists = &SanUtil::break_up_array(array => \@vservers, size => 15);
        foreach my $cluster_obj (@cluster_objs) {
            my $cluster_name = $cluster_obj->name();

            foreach my $vs_scope_list (@vs_scope_lists) {
                my $vs_filter = {filter => {vserver => join(",", @{$vs_scope_list})}};
                try {
                    my @vol_objs = NACL::CS::Volume->fetch(command_interface => $cluster_obj,
                                                           allow_empty => 1,
                                                           requested_fields => ["vserver","volume","aggregate","size"],
                                                           %{$vs_filter});
                    my @lun_objs = NACL::CS::Lun->fetch(command_interface => $cluster_obj,
                                                        allow_empty => 1,
                                                        requested_fields => ["vserver","path","volume","qtree","lun","size","serial","ostype"],
                                                        %{$vs_filter});
                    my @lun_mapped_objs = NACL::CS::LunMapped->fetch(command_interface => $cluster_obj,
                                                                     allow_empty => 1,
                                                                     requested_fields => ["vserver","igroup","lun-id","path","initiators"],
                                                                     %{$vs_filter});
                    push(@{$nacl_cluster_content->{$cluster_name}->{vol_objs}}, @vol_objs);
                    push(@{$nacl_cluster_content->{$cluster_name}->{lun_objs}}, @lun_objs);
                    push(@{$nacl_cluster_content->{$cluster_name}->{lun_mapped_objs}}, @lun_mapped_objs);
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught: " . $e->text());
                };
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }
            }
        }
    }

    # load the initial pathing information for all sanclient objects that
    # have path validation enabled
    foreach my $client_name (@sanclients) {

        # get the resource object for this sanclient
        my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            $self->_set_err("Could not get resource '$client_name' of type 'sanclients'");
            return $EventUtil::FALSE;
        }

        # determine if this client has path validation enabled on it
        # default the value of path validation to 'on' for all hosts
        my $path_validation = $self->EventUtil::get_arg(-arg => "path_validation",
                                                        -objects => [$resource],
                                                        -default_to_env => 1);
        if ($path_validation == 0) {
            &SanUtil::nate_log('-comment', "Skipping load initial pathing state for client " .
                                           "($client_name) because path validation is not enabled for this host");
            next;
        }

        # determine if path validation abort on error is set
        my $pv_abort_on_error = $self->EventUtil::get_arg(-arg => "path_validation_abort_on_error",
                                                          -objects => [$resource],
                                                          -default_to_env => 1);
        # determine if path validation init abort on error is set
        my $pv_init_abort_on_error = $self->EventUtil::get_arg(-arg => "path_validation_init_state_abort_on_error",
                                                               -objects => [$resource],
                                                               -default_to_env => 1);

        # extract the sanclient object from the sanclients datastructure
        my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
        if ($self->error()) {
            $self->_set_err("Could not obtain SanClient object for client '$client_name'");
            return $EventUtil::FALSE;
        }

        # parse out the mptype
        my $mp_type = $sc_obj->get_mp_type();

        # load pathing information for all sc objects that are running multipathing
        if ($mp_type !~ /NONE/i) {
            if ($sc_obj->mp()->can("load_path_info") or $sc_obj->San::SanClient::MPDevMgr::is_supported()) {
                my $st_id = "load_initial_pathing_info_" . $client_name;
                my $subtest = subtest(sub {
                                          if ($sc_obj->San::SanClient::MPDevMgr::is_supported()) {
                                              # load the MPDevMgr for each SanClient object
                                              $self->PathValidation::_load_mpdevmgr_pathing_information_per_sanclient(-client_name => $client_name,
                                                                                                                      -nacl_cluster_content => $nacl_cluster_content);
                                              if ($self->error()) {
                                                  $self->_set_err("Could not load mpdevmgr pathing information for client '$client_name'");
                                                  $self->SubtestComm::subtest_fail(-graceful => not ($pv_abort_on_error or $pv_init_abort_on_error));
                                                  return $EventUtil::FALSE;
                                              }
                                          } else {
                                              # load pathing information for this host
                                              $self->PathValidation::_load_legacy_pathing_information_per_sanclient(-client_name => $client_name);
                                              if ($self->error()) {
                                                  $self->_set_err("Could not load pathing information for client '$client_name'");
                                                  $self->SubtestComm::subtest_fail(-graceful => not ($pv_abort_on_error or $pv_init_abort_on_error));
                                                  return $EventUtil::FALSE;
                                              }
                                          }

                                          # wait for the parent to reap
                                          $self->SubtestComm::subtest_complete();
                                          return $EventUtil::TRUE;
                                      }, -runid => $st_id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
                push(@subtests, $subtest);
            }
        } else {
            &SanUtil::nate_log('-comment', "Skipping load initial pathing state for client " .
                                           "($client_name) because this host is not running " .
                                           "any multipathing software (mp_type == NONE)");
            next;
        }
    }

    # if there are any subtests that we need to reap do so now
    if (scalar(@subtests)) {
        # reap all children, once the steady states have been collected, store
        # the steady state hash on the sanclient's data structure for each client
        my $return_data = $self->wait_for_subtests(-subtests => \@subtests,
                                                   -dependence_type => "independent",
                                                   -propagate_wait => 0,
                                                   -propagate_events => 1,
                                                   -enable_sync => 0,
                                                   -check_cmds_from_parent => 0,
                                                   -forward_cmds_from_parent => 0,
                                                   -async => 0);
        if ($self->error()) {
            $self->_set_err("Failed to load pathing information for all clients");
            return $EventUtil::FALSE;
        }

        # process any return data from the children subtests
        $self->_process_event_return_data(-return_data => $return_data,
                                          -propagate => $EventUtil::TRUE);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # parse the return data and store the steady state of each client in the sanclients datastructure
        foreach my $message_obj (@{$return_data}) {
            next if ($message_obj->{msg} !~ /steady_state_update/);

            my $client_name = $message_obj->{data}->{client_name};
            my $steady_state = $message_obj->{data}->{steady_state};

            # get the sanclient object for this client
            my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
            if ($self->error()) {
                $self->_set_err("Could not obtain SanClient object for client '$client_name'");
                return $EventUtil::FALSE;
            }

            # set the steady state on the mp object for the SanClient object
            $sc_obj->mp()->set_steady_state($steady_state);

            # set an attribute on this client resource object to indicate that steady state is loaded
            # don't propagate the resource attribute change to the parent subtest since we should
            # be at the top subtest anyway
            $self->ResMgr::set_resource_attribute(-name => $client_name,
                                                  -type => "sanclients",
                                                  -attribute => "steady_state_loaded",
                                                  -value => $EventUtil::TRUE,
                                                  -propagate => $EventUtil::FALSE);
            if ($self->error()) {
                $self->_set_err("Could not set attribute 'steady_state_loaded' on resource '$client_name' of type 'sanclients'");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _load_mpdevmgr_pathing_information_per_sanclient()
# Objective: Load the initial pathing information for a specific SanClient
###############################################################################
sub _load_mpdevmgr_pathing_information_per_sanclient {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);
    my $nacl_cluster_content = &ParamsMgr::get_arg(-arg => "nacl_cluster_content",
                                                   -required => 1);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get the resource object for this sanclient
    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Could not obtain resource '$client_name' of type 'sanclients'");
        return $EventUtil::FALSE;
    }

    # extract the sanclient object from the sanclients datastructure
    my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanClient object for client '$client_name'");
        return $EventUtil::FALSE;
    }

    # get all the nacl cluster / sanfiler7m objects in scope
    my @sanfilers7m_objs = $self->EventUtil::get_sanfiler_objs(-type => "7mode");
    if ($self->error()) {
        $self->_set_err("Unable to obtain sanfilers7m objects");
        return $EventUtil::FALSE;
    }

    my @cluster_objs = $self->EventUtil::get_nacl_cluster_cm_objs();
    if ($self->error()) {
        $self->_set_err("Unable to obtain NACL::C::Cluster objects");
        return $EventUtil::FALSE;
    }

    # get the names of the vservers that are in scope for this sanclient resource, otherwise default
    # to the environment's value of vservers
    my @vserver_names = $self->EventUtil::get_vservers(-objects => [$resource],
                                                       -default_to_env => 1);
    if ($self->error()) {
        $self->_set_err("Unable to obtain vserver names");
        return $EventUtil::FALSE;
    }

    # log an event that load pathing information started for this client
    $self->SubtestComm::log_event_message(-event_message => "pv_loading_mpdevmgr_on_${client_name}");

    $sc_obj->create_mpdevmgr(-sanfilers7m_objs => \@sanfilers7m_objs,
                             -nacl_cluster_objs => \@cluster_objs,
                             -nacl_cluster_content => $nacl_cluster_content,
                             -vservers => \@vserver_names);
    if ($sc_obj->error()) {
        $self->_set_err($sc_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # update the sanclient object for this sanclient
    $self->ResMgr::modify_resource_obj(-value => $sc_obj,
                                       -name => $client_name,
                                       -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Failed to modify resource obj for ($client_name : sanclients)");
        return $EventUtil::FALSE;
    }

    # display a pathing report of all the devices in the MPDevMgr
    my $reports = $sc_obj->mpdevmgr()->generate_device_reports();
    unshift(@{$reports}, "");
    unshift(@{$reports}, "Multi-Pathed Device Report");
    unshift(@{$reports}, ".");
    my $report_str = join("\n. ", @{$reports});
    $Log->comment("$report_str");

    # check pathing steady state for the mpdevmgr object
    # verify the steady state of this sanclient host
    $self->PathValidation::_check_pathing_steady_state_per_sanclient(-client_name  => $client_name,
                                                                     -mpdevmgr => $sc_obj->mpdevmgr());
    if ($self->error()) {
        $self->_set_err("Check pathing steady state failed for client '$client_name'");
        return $EventUtil::FALSE;
    }

    # set an attribute so that we know whether mpdevmgr is loaded
    $self->ResMgr::set_resource_attribute(-name => $client_name,
                                          -type => "sanclients",
                                          -attribute => "mpdevmgr_loaded",
                                          -value => $EventUtil::TRUE,
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Could not set attribute 'mpdevmgr_loaded' on resource '$client_name' of type 'sanclients'");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _load_legacy_pathing_information_per_sanclient()
# Objective: Load the initial pathing information for a specific SanClient
###############################################################################
sub _load_legacy_pathing_information_per_sanclient {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # extract the sanclient object from the sanclients datastructure
    my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanClient object for client '$client_name'");
        return $EventUtil::FALSE;
    }

    # log an event that load pathing information started for this client
    $self->SubtestComm::log_event_message(-event_message => "pv_loading_pathing_information_on_${client_name}");

    # run load_path_info() on mp object
    $sc_obj->mp()->load_path_info();
    if ($sc_obj->mp()->error()) {
        $self->_set_err($sc_obj->mp()->errmsg());
        return $EventUtil::FALSE;
    }

    # dump the pathing information to the logs
    $sc_obj->mp()->dump_path_info();
    if ($sc_obj->mp()->error()) {
        $self->_set_err($sc_obj->mp()->errmsg());
        return $EventUtil::FALSE;
    }

    # obtain the steady state hash from the mp object
    my $steady_state = $sc_obj->mp()->get_steady_state();

    # verify the steady state of this sanclient host
    $self->PathValidation::_check_pathing_steady_state_per_sanclient(-client_name  => $client_name,
                                                                     -steady_state => $steady_state);
    if ($self->error()) {
        $self->_set_err("Check pathing steady state failed for client '$client_name'");
        return $EventUtil::FALSE;
    }

    # return the subtest data for the parent to process
    $self->SubtestComm::subtest_return_data(-msg => "steady_state_update",
                                            -data => {client_name  => $client_name,
                                                      steady_state => $steady_state});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _check_pathing_steady_state_per_sanclient()
# Objective: Verify the steady state of the specified SanClient
###############################################################################
sub _check_pathing_steady_state_per_sanclient {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);

    my $steady_state = &ParamsMgr::get_arg(-arg => "steady_state",
                                           -default => undef);
    my $mpdevmgr = &ParamsMgr::get_arg(-arg => "mpdevmgr",
                                       -default => undef);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get the resource object for this sanclient
    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # determine this client's init state parameters
    my $pv_init_state = $self->EventUtil::get_arg(-arg => "path_validation_init_state",
                                                  -objects => [$resource],
                                                  -default_to_env => 1);

    my $pv_fcp_init_state = $self->EventUtil::get_arg(-arg => "path_validation_fcp_init_state",
                                                      -objects => [$resource],
                                                      -default_to_env => 1);

    my $pv_iscsi_init_state = $self->EventUtil::get_arg(-arg => "path_validation_iscsi_init_state",
                                                        -objects => [$resource],
                                                        -default_to_env => 1);

    # parse the init state parameters
    my ($init_num_paths_agnostic, $init_alive_paths_agnostic, $init_dead_paths_agnostic) =
        split(/:/, $pv_init_state);
    my ($init_num_paths_fcp, $init_alive_paths_fcp, $init_dead_paths_fcp) =
        split(/:/, $pv_fcp_init_state);
    my ($init_num_paths_iscsi, $init_alive_paths_iscsi, $init_dead_paths_iscsi) =
        split(/:/, $pv_iscsi_init_state);

    # if any of the init state variables are defined, enable init state checking
    my $init_state_check = 0;
    if ($pv_init_state ne "" or
        $pv_fcp_init_state ne "" or
        $pv_iscsi_init_state ne "") {
        &SanUtil::nate_log('-comment',
                           "'path_validation_init_state' checking is enabled " .
                           "and will be used to validate steady state.\n" .
                           " 'path_validation_init_state' ($pv_init_state)\n" .
                           " 'path_validation_fcp_init_state' ($pv_fcp_init_state)\n" .
                           " 'path_validation_iscsi_init_state' ($pv_iscsi_init_state)");
        $init_state_check = 1;
    } else {
        # otherwise disable init state checking
        &SanUtil::nate_log('-comment', "Init State checking is not enabled. The current state will be considered 'steady state'");
        $init_state_check = 0;
    }

    my $path_errs = 0;
    my $dead_paths_exist = 0;
    my $not_enough_paths = 0;
    my $errors = 0;

    # choose between MP and MPDevMgr
    my @devices;
    if (defined($steady_state) and (scalar(keys %{$steady_state}) != 0)) {
        @devices = keys %{$steady_state};
    } elsif (defined($mpdevmgr)) {
        # find only netapp devices when doing path validation
        my @device_objs = $mpdevmgr->get_filtered_devices(-device_vendor => "NETAPP");
        if ($mpdevmgr->error()) {
            $self->_set_err("The client ($client_name) has no NETAPP LUNs/Paths found in steady state.");
            return $EventUtil::FALSE;
        }
        @devices = map {$_->mp_device_name();} @device_objs;
    } else {
        # if steady state does not have any LUNs/Paths, this is really bad
        $self->_set_err("The client ($client_name) has no LUNs/Paths found in steady state.");
        return $EventUtil::FALSE;
    }

    # Look through each client and grab state information and path info.
    foreach my $device (@devices) {

        my $cur_num_paths;
        my $cur_alive_paths;
        my $cur_dead_paths;
        my $cur_protocol;

        # use legacy sanclient MP steady state
        if (defined($steady_state)) {
            $cur_num_paths = $steady_state->{$device}->{num_paths};
            $cur_alive_paths = $steady_state->{$device}->{num_paths_alive};
            $cur_dead_paths = $steady_state->{$device}->{num_paths_dead};
            $cur_protocol = $steady_state->{$device}->{protocol};
        } else {
            # use MPDevMgr APIs to get steady state
            my $devobj = $mpdevmgr->get_device(-mp_device_name => $device);
            my $path_state = $devobj->get_current_path_state();
            $cur_num_paths = $path_state->total_paths();
            $cur_alive_paths = $path_state->total_alive_paths();
            $cur_dead_paths = $path_state->total_dead_paths();
            $cur_protocol = $devobj->protocol();
        }

        # per LUN, determine the appropriate init state to use
        # based on the protocol of the LUN
        my $init_state = $pv_init_state;
        my $init_state_check_str = "path_validation_init_state";
        my $init_num_paths = $init_num_paths_agnostic;
        my $init_alive_paths = $init_alive_paths_agnostic;
        my $init_dead_paths = $init_dead_paths_agnostic;
        if ($cur_protocol =~ /iscsi/i) {
            # iscsi init state
            if ($pv_iscsi_init_state ne "") {
                $init_state = $pv_iscsi_init_state;
                $init_state_check_str = "path_validation_iscsi_init_state";
                $init_num_paths = $init_num_paths_iscsi;
                $init_alive_paths = $init_alive_paths_iscsi;
                $init_dead_paths = $init_dead_paths_iscsi;
            }
        } elsif ($cur_protocol =~ /fcp/i) {
            # fcp init state
            if ($pv_fcp_init_state ne "") {
                $init_state = $pv_fcp_init_state;
                $init_state_check_str = "path_validation_fcp_init_state";
                $init_num_paths = $init_num_paths_fcp;
                $init_alive_paths = $init_alive_paths_fcp;
                $init_dead_paths = $init_dead_paths_fcp;
            }
        } else {
            # mixed protocols
            $Log->warn("Device ($device) is using a mixed protocol, Path validation does not support this at this time");
            next;
        }

        # if init state check is enabled, perform the comparisions for (total:alive:dead) paths for this LUN
        if ($init_state_check == 1) {
            # compare the current num paths with the init state num paths, warn if they dont match
            if ($cur_num_paths != $init_num_paths) {
                $Log->warn("($client_name) Steady state number of paths ($cur_num_paths) for device ($device)\n" .
                           "does not match number passed in the variable $init_state_check_str ($init_num_paths)");
                $path_errs++;
            }

            # compare the number of alive paths with the init state alive paths, warn if they dont match
            if ($cur_alive_paths != $init_alive_paths) {
                $Log->warn("($client_name) Steady state number of alive paths ($cur_alive_paths) for device ($device)\n" .
                           "does not match number passed in the variable $init_state_check_str ($init_alive_paths)");
                $path_errs++;
            }

            # compare the number of dead paths with the init state dead paths, warn if they dont match
            if ($cur_dead_paths != $init_dead_paths) {
                $Log->warn("($client_name) Steady state number of dead paths ($cur_dead_paths) for device ($device)\n" .
                           "does not match number passed in the variable $init_state_check_str ($init_dead_paths)");
                $path_errs++;
            }

        } else {
            # otherwise init state checking is disabled, so perform some other validation

            # make sure there is at least more than 2 paths for this lun
            if ($cur_num_paths < 2) {
                $Log->warn("($client_name) Steady state found only ($cur_num_paths) paths to the device ($device)\n" .
                           "With multipathing, there should be two or more paths per LUN.\n" .
                           "Please make sure this is was intended (or correct the configuration)");
                $not_enough_paths++;
            }

            # report if we detect any dead paths to this lun
            if ($cur_dead_paths != 0) {
                $Log->warn("($client_name) Steady state found ($cur_dead_paths) dead path(s) for device ($device)\n" .
                           "Please make sure this was intended (or correct the configuration)");
                $dead_paths_exist++;
            }
        }
    }

    $errors = 0;

    if ($init_state_check == 1) {
        # If init state path comparison errors exist, report this as an error
        if ($path_errs > 0) {
            $errors = 1;
        } else {
            &SanUtil::nate_log('-comment', "($client_name) Passed - init state path validation succeeded");
        }
    } else {
        # report if dead paths were found
        if ($dead_paths_exist > 0) {
            $Log->warn("$client_name) Failed - dead paths were recorded in steady state");
            $errors = 1;
        } else {
            &SanUtil::nate_log('-comment', "($client_name) Passed - no dead paths were recorded in steady state for any device");
        }

        # if not enough paths to any LUN was found
        if ($not_enough_paths > 0) {
            $Log->warn("($client_name) Failed - not enough paths (2 or greater) were recorded in steady state");
            $errors = 1;
        } else {
            &SanUtil::nate_log('-comment', "($client_name) Passed - 2 or more paths were recorded in steady state for all devices");
        }
    }

    # if there were any errors during the path checking, set an error on the object
    if ($errors == 1) {
        $self->_set_err("Client ($client_name) encountered errors during init state path validation.");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_paths()
# Objective: Validate the paths for all SanClients or specified sanclients
###############################################################################
sub _validate_paths {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);


    # parse out the id extension when creating the subtest for path validation
    my $id_extension = &ParamsMgr::get_arg(-arg => "id_extension",
                                           -required => 1);

    # default the takeover mode for path validation to off, which is the equivalent
    # of invoking the wrapper validate_paths_after_giveback()
    my $takeover_state = &ParamsMgr::get_arg(-arg => "takeover_state",
                                             -default => 0);

    # these are a set of rules defined by whomever wants path validation to occur for mpdevmgr
    my $pv_rules = &ParamsMgr::get_arg(-arg => "path_validation_rules",
                                       -default => undef);

    # the caller may have requested to perform path validation on a specific set of
    # sanclients
    my @sanclients = @{&ParamsMgr::get_arg(-arg => "sanclients",
                                           -ref_type => "ARRAY",
                                           -default => [])};

    # the caller may have requested that we wait a minimum timeout
    # before performing path validation
    my $last_timeout = &ParamsMgr::get_arg(-arg => "last_timeout",
                                           -default => 0);
    my $last_timeout_name = &ParamsMgr::get_arg(-arg => "last_timeout_name",
                                                -default => "timeout");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine if path validation is enabled in the current event/eventset
    my $path_validation = $self->EventUtil::get_arg(-arg => "path_validation");

    # if path validation is not enabled for this current event, then return immediately
    if ($path_validation == 0) {
        &SanUtil::nate_log('-comment', "Skipping Path Validation for this event because it was disabled");
        return $EventUtil::TRUE;
    }

    # if not requested to only perform path validation on a specific list of hosts by the calling function
    # then overwrite @sanclients with the value in the current event environemnt
    if (not scalar(@sanclients)) {
        @sanclients = $self->EventUtil::get_sanclients();
    }

    # if there are no sanclients to perform path validation on at this point, then return immediately
    if (not scalar(@sanclients)) {
        &SanUtil::nate_log('-comment', "Skipping Path Validation for this event because there are no SanClient resources to operate against");
        return $EventUtil::TRUE;
    }

    # get the next padded runid from the event
    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $path_validation_runid_padded = $event_obj->increment_subtest_runid_counter();

    # wait path validation min wait time if a value for
    # last timeout was passed in
    if ($last_timeout > 0) {
        $self->PathValidation::_wait_for_validate_paths(-last_timeout => $last_timeout,
                                                        -last_timeout_name => $last_timeout_name);
    }

    my @subtests = ();

    # for each client, run path validation
    foreach my $client_name (@sanclients) {

        # extract the sanclient object from the sanclients datastructure
        my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
        if ($self->error()) {
            $self->_set_err("Could not obtain a SanClient object for client '$client_name'");
            return $EventUtil::FALSE;
        }

        # get the resource object for this sanclient
        my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                                   -type => "sanclients");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # determine if path validation is enabled for this client, if it doesnt have a specified value
        # for this client, default to the value defined in the current event/eventset
        $path_validation = $self->EventUtil::get_arg(-arg => "path_validation",
                                                     -objects => [$resource],
                                                     -default_to_env => 1);
        # skip path validation on this host if its not enabled on the client
        if ($path_validation == 0) {
            &SanUtil::nate_log("-comment", "Skipping Path Validation on client ($client_name) because " .
                                           "path validation is not enabled for this host.");
            next;
        }

        # determine if this client has a valid steady state loaded
        # skip path validation if it doesnt have it loaded
        my $steady_state_loaded = $self->PathValidation::_is_steady_state_loaded(-client_name => $client_name);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        my $mpdevmgr_loaded = $self->PathValidation::_is_mpdevmgr_loaded(-client_name => $client_name);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # skip this client if we don't have a steady state to validate against
        if ($steady_state_loaded == $EventUtil::FALSE and $mpdevmgr_loaded == $EventUtil::FALSE) {
            &SanUtil::nate_log("-comment", "Skipping Path Validation on client ($client_name) ".
                                           "because it does not have steady state loaded.");
            next;
        }

        # determine if path validation abort on error is set for this client, if it isnt, then
        # default to the value defined in the current event/eventset
        my $pv_abort_on_error = $self->EventUtil::get_arg(-arg => "path_validation_abort_on_error",
                                                          -objects => [$resource],
                                                          -default_to_env => 1);

        # run path validation only if this client is running multipathing software
        my $mp_type = $sc_obj->get_mp_type();
        if ($mp_type =~ /NONE/i) {
            &SanUtil::nate_log('-comment', "Skipping Path Validation on client ($client_name) because it " .
                                           "is not running multipathing software (mp_type == NONE)");
            next;
        }

        # and if path validation is specifically enabled for this client
        my $id = $path_validation_runid_padded . "_verify_pathing_info_${client_name}_${id_extension}";
        my $subtest = subtest(sub {
                                  # first reconnect to the client
                                  $sc_obj->reconnect();
                                  if ($sc_obj->error()) {
                                      $self->_set_err($sc_obj->errmsg());
                                      $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                      return $EventUtil::FALSE;
                                  }

                                  my (@errors, @warnings);

                                  if ($mpdevmgr_loaded) {
                                      # perform mpdevmgr path validation
                                      my ($return_hsh1) =
                                          $self->PathValidation::_validate_mpdevmgr_paths_per_sanclient(-client_name => $client_name,
                                                                                                        -path_validation_rules => $pv_rules,
                                                                                                        -takeover_state => $takeover_state);
                                      # check for config errors found during the collection
                                      if ($self->error()) {
                                          # perform a graceful failure if this client has pv_abort_on_error == 0
                                          $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                          return $EventUtil::FALSE;
                                      }
                                      push(@warnings, @{$return_hsh1->{warnings}});
                                      push(@errors, @{$return_hsh1->{errors}});
                                  } elsif ($steady_state_loaded) {
                                      # perform legacy path validation otherwise
                                      my ($return_hsh2) =
                                          $self->PathValidation::_validate_legacy_paths_per_sanclient(-client_name => $client_name,
                                                                                                      -id_extension => $id_extension,
                                                                                                      -takeover_state => $takeover_state);
                                      # check for config errors found during the collection
                                      if ($self->error()) {
                                          # perform a graceful failure if this client has pv_abort_on_error == 0
                                          $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                          return $EventUtil::FALSE;
                                      }
                                      push(@warnings, @{$return_hsh2->{warnings}});
                                      push(@errors, @{$return_hsh2->{errors}});
                                  } else {
                                      # We were never able to collect steady state for this client, can't perform
                                      # path validation
                                      push(@warnings, "Can not perform path validation because initial " .
                                                      "steady state was not loaded.");
                                  }

                                  # Check for path validation warnings & errors
                                  if (scalar(@warnings)) {
                                      # log the warning, but continue
                                      my $warnings = join("\n", @warnings);
                                      &SanUtil::nate_log('-warn', "Detected warnings during path validation for '$client_name'");
                                      &SanUtil::nate_log('-comment', "Path Validation Warnings for client '$client_name':\n\n$warnings");
                                  }
                                  if (scalar(@errors)) {
                                      my $errors = join("\n", @errors);
                                      &SanUtil::nate_log('-comment', "Path Validation Errors for client '$client_name':\n\n$errors");

                                      # if there were errors run any optional user path validation abort routines here:
                                      $self->PathValidation::_run_path_validation_abort_scripts(-client_name => $client_name);
                                      if ($self->error()) {
                                          $self->_set_err("Encountered aproblem when running path validation abort " .
                                                          "scripts for client '$client_name'");
                                      }

                                      # set an error and abort
                                      $self->_set_err("Path Validation failed for client '$client_name'");
                                      $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                      return $EventUtil::FALSE;
                                  } else {
                                      &SanUtil::nate_log('-comment', "No path validation errors found for client '$client_name'");
                                  }

                                  # extract the sanclient object from the sanclients datastructure
                                  my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
                                  if ($self->error()) {
                                      $self->_set_err("Could not obtain SanClient object for client '$client_name'");
                                      $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                      return $EventUtil::FALSE;
                                  }

                                  # update the sanclient object on the ResMgr
                                  $self->ResMgr::modify_resource_obj(-value => $sc_obj,
                                                                     -name => $client_name,
                                                                     -type => "sanclients");
                                  if ($self->error()) {
                                      $self->_set_err("Failed to modify resource obj for ($client_name : sanclients)");
                                      $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                      return $EventUtil::FALSE;
                                  }

                                  # path validation succeeded
                                  $self->SubtestComm::subtest_complete();
                                  return $EventUtil::TRUE;
                              }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
        push(@subtests, $subtest);
    }

    # wait for all the path validation subtests to complete
    if (scalar(@subtests)) {
        my $return_data = $self->wait_for_subtests(-subtests => \@subtests,
                                                   -dependence_type => "independent",
                                                   -propagate_wait => 0,
                                                   -propagate_events => 1,
                                                   -enable_sync => 0,
                                                   -check_cmds_from_parent => 0,
                                                   -forward_cmds_from_parent => 0,
                                                   -async => 0);
        if ($self->error()) {
            $self->_set_err("Path validation failed on a client");
            return $EventUtil::FALSE;
        }

        # process any return data from the children subtests
        if (defined($return_data) and ref($return_data) eq "HASH" and scalar(keys %{$return_data})) {
            $self->_process_event_return_data(-return_data => $return_data,
                                              -propagate => $EventUtil::TRUE);
            if ($self->error()) {
                $self->_set_err("Failed to process event return data");
                return $EventUtil::FALSE;
            }
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _run_path_validation_abort_scripts()
# Objective: Run the path validation abort scripts for the given client
###############################################################################
sub _run_path_validation_abort_scripts {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get the resource object for this sanclient
    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Could not obtain resource '$client_name' of type 'sanclients'");
        return $EventUtil::FALSE;
    }

    # extract the sanclient object from the sanclients datastructure
    my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanClient object for client '$client_name'");
        return $EventUtil::FALSE;
    }

    # Gather data from the client if path validation failed
    $sc_obj->gather_data();
    if ($sc_obj->error) {
        @errmsg = $sc_obj->errmsg();
        &SanUtil::nate_log('-warn',
                           "Failed to Gather data from " .
                               "client ($client_name)\n@errmsg");
        $sc_obj->_clear_err();
    }

    # handle the client path validation abort script in the event that path validation fails
    my $pv_abort_script = $self->EventUtil::get_arg(-arg => "path_validation_abort_script",
                                                    -objects => [$resource],
                                                    -default_to_env => 1);
    if ($pv_abort_script ne "") {
        # determine whether the script should execute on the client or on the control server
        my $pv_abort_script_execute_on = $self->EventUtil::get_arg(-arg => "path_validation_abort_script_execute_on",
                                                                   -objects => [$resource],
                                                                   -default => "device",
                                                                   -default_to_env => 1);

        my $connection = $sc_obj->get_connection();
        my $cmd = $pv_abort_script;
        if ($pv_abort_script_execute_on eq "device") {
            # execute on client
            &ClientUtil::execute($connection, $cmd, 600);
        } elsif ($pv_abort_script_execute_on eq "nate_control_server") {
            # execute on control server
            &SanUtil::nate_server_execute(-cmd => $cmd,
                                          -timeout => 600);
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_legacy_paths_per_sanclient()
# Objective: Validate paths for a specific SanClient
###############################################################################
sub _validate_legacy_paths_per_sanclient {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);
    my $id_extension = &ParamsMgr::get_arg(-arg => "id_extension",
                                           -required => 1);
    my $takeover_state = &ParamsMgr::get_arg(-arg => "takeover_state",
                                             -required => 1);

    # initialize a few things
    my $return_hash = {};
    $return_hash->{warnings} = [];
    $return_hash->{errors} = [];

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $return_hash;
    }

    # get the resource object for this sanclient
    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Could not obtain resource '$client_name' of type 'sanclients'");
        return $return_hash;
    }

    # log an event that we are validating paths for this client
    $self->SubtestComm::log_event_message(-event_message => "pv_validating_paths_on_${client_name}");

    # extract the sanclient object from the sanclients datastructure
    my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanClient object for client '$client_name'");
        return $return_hash;
    }

    # determine what the validation rule type for this resource are (7mode/Cmode)
    my $pv_rule_type = $self->EventUtil::get_arg(-arg => "path_validation_rule_type",
                                                 -objects => [$resource],
                                                 -default_to_env => 1);

    # validate the paths using the specified rule type
    $sc_obj->mp()->validate_paths($client_name,
                                  $takeover_state,
                                  $pv_rule_type);
    if ($sc_obj->mp()->error()) {
        # bubble up the error on the mp object
        push(@{$return_hash->{errors}}, $sc_obj->mp()->errmsg());
        return $return_hash;
    }

    return $return_hash;
}

###############################################################################
# Method:    _validate_mpdevmpgr_per_sanclient()
# Objective: Validate paths for a specific SanClient via the mpdevmgr
###############################################################################
sub _validate_mpdevmgr_paths_per_sanclient {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);
    my $path_validation_rules = &ParamsMgr::get_arg(-arg => "path_validation_rules",
                                                    -default => undef);
    my $takeover_state = &ParamsMgr::get_arg(-arg => "takeover_state",
                                             -required => 1);

    # initialize a few things
    my $return_hash = {};
    $return_hash->{warnings} = [];
    $return_hash->{errors} = [];

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $return_hash;
    }

    # get the resource object for this sanclient
    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        $self->_set_err("Could not obtain resource '$client_name' of type 'sanclients'");
        return $return_hash;
    }

    # extract the sanclient object from the sanclients datastructure
    my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanClient object for client '$client_name'");
        return $return_hash;
    }

    # skip mpdevmgr path validation if there is no MPDevMgr defined for this host
    my $mpdevmgr = $sc_obj->mpdevmgr();
    if (not defined($mpdevmgr)) {
        &SanUtil::nate_log('-comment', "Skipping MPDevMgr path validation for '$client_name' " .
                                       "because no MPDevMgr object is loaded");
        return $return_hash;
    }

    # log an event that we are validating paths for this client
    $self->SubtestComm::log_event_message(-event_message => "pv_validating_mpdevmgr_on_${client_name}");

    # first refresh the mpdevmgr
    $mpdevmgr->refresh();
    if ($mpdevmgr->error()) {
        $self->_set_err($mpdevmgr->errmsg());
        return $return_hash;
    }

    # notify the user if no path validation rules were specified.
    if (not defined($path_validation_rules)) {
        &SanUtil::nate_log('-comment', "No path validation rules defined for '$client_name'");
    }

    # If requested to do vol move path validation, we need to regenerate lun relationship
    # data now since it might be stale, (aggr information may be stale in the mpdevmgr)
    if ($path_validation_rules and $path_validation_rules->{validate_vol_move} == 1) {
        # get all the cluster objects
        my @cluster_objs = $self->EventUtil::get_nacl_cluster_cm_objs();
        if ($self->error()) {
            $self->_set_err("Failed to obtain cluster objs");
            return $return_hash;
        }

        # get the names of the vservers that are in scope to this sanclient resource, otherwise default
        # to the environment's value of vservers
        my @vserver_names = $self->EventUtil::get_vservers(-objects => [$resource],
                                                           -default_to_env => 1);
        if ($self->error()) {
            $self->_set_err("Unable to obtain vserver names");
            return $EventUtil::FALSE;
        }

        # first and foremost, we have to rerun generate lun relationships on
        # the mpdevmgr because luns have moved around between nodes due to vol move
        # operations and the current mpdevmgr may not be accurate
        $mpdevmgr->generate_lun_relationships(-nacl_cluster_objs => \@cluster_objs,
                                              -vservers => \@vserver_names);
        if ($mpdevmgr->error()) {
            $self->_set_err("Failed to regenerate LUN relationships");
            return $return_hash;
        }
    }

    my @errors;
    my @warnings;

    # Find any devices that are missing from the mpdevmgr after refreshing
    if (my @missing_devices = $mpdevmgr->find_missing_devices()) {
        my @arr = ();
        foreach my $mpdev_obj (@missing_devices) {
            push(@arr, $mpdev_obj->mp_device_name() . " -> " . $mpdev_obj->ontap_controller() . ": " . $mpdev_obj->ontap_lun_path());
        }
        push(@errors, "The following MP Devices are missing on this client:\n\t" . join("\n\t", @arr));
    }

    # perform smart takeover giveback path validation
    # determine all the nodes that are in faulted states in the resmgr
    my @faulted_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to get faulted NACL node resources");
        return $return_hash;
    }

    # determine all the nodes that are in faulted states in the resmgr
    my @unfaulted_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to get unfaulted NACL node resources");
        return $return_hash;
    }

    # validate paths on all netapp devices
    foreach my $dev_obj ($mpdevmgr->get_filtered_devices(-device_vendor => "NETAPP")) {
        my @errors_for_device;
        my @warnings_for_device;

        # if this device has no paths to it, that's an error.
        if ($dev_obj->get_current_path_state()->total_alive_paths() eq 0) {
            push(@errors_for_device, "There are no alive paths for this device.");
        }

        # perform custom path count validation
        # this validation allows the caller to specify a specific path down count
        # to validate against in the pv rules
        if ($path_validation_rules->{validate_path_counts}) {
            # default the path down count to 0, which basically means that no paths should be missing
            my $path_down_count = exists($path_validation_rules->{path_down_count}) ? $path_validation_rules->{path_down_count} : 0;
            my $curr_alive_paths = $dev_obj->get_current_path_state()->total_alive_paths();
            my $prev_alive_paths = $dev_obj->get_previous_path_state()->total_alive_paths();
            my $expected_alive_paths = $prev_alive_paths - $path_down_count;
            if ($curr_alive_paths < ($prev_alive_paths - $path_down_count)) {
                # the number of alive paths is less than the expected alive count
                push(@errors_for_device, "Was expecting there to be $expected_alive_paths alive paths, " .
                     "but there are only $curr_alive_paths paths. (path_down_count=${path_down_count})");
            }
        }

        # perform takeover/giveback path validation
        if ($path_validation_rules->{validate_takeover_giveback} == 1) {

            # basic takeover/giveback path validation:
            if ($takeover_state eq $EventUtil::FALSE) {
                # basic path validation while in giveback state:
                # no paths should be dead or missing
                # compare the current path state with the previous path state to
                # see if any paths are missing or are dead
                push(@errors_for_device, $dev_obj->find_missing_or_dead_paths());
            } else {
                # basic path validation while in a takeover state:
                # compare the current path state with the previous path state to
                # see that each lun has at least half the number of alive paths that they did before
                my $curr_alive_paths = $dev_obj->get_current_path_state()->total_alive_paths();
                my $prev_alive_paths = $dev_obj->get_previous_path_state()->total_alive_paths();
                if ($curr_alive_paths < int($prev_alive_paths / 2)) {
                    push(@errors_for_device, "Did not find enough alive paths after the " .
                         "takeover, usually there should be at least half. " .
                         "current_count=$curr_alive_paths previous_count=$prev_alive_paths");
                }
            }

            # If this device has a lun relationship defined, then we can do some
            # additional validations.
            # Scan the path state of this device to see if the proper paths are missing
            # if takeover mode: for every path to each mpdevice, validate that the only paths that are
            #                   down are the nodes in faulted states
            # in giveback mode: verify that all paths are accessible
            if ($dev_obj->has_lun_relationship()) {
                my $path_state_obj = $dev_obj->get_current_path_state();

                # walk through all path objects
                foreach my $path_obj ($path_state_obj->get_path_objs()) {

                    # skip paths that dont have lif relationships
                    if (not $path_obj->has_lif_relationship()) {
                        # issue a warning that we cannot verify this path unless it doesn't care about lif relationships
                        unless ($path_obj->ignore_lif_relationship()) {
                            push(@warnings_for_device, "The path: '" . $path_obj->path_name() .
                                 "' does not have a lif associated to it, which means we did not see " .
                                 "it when path states were initialized. Is this a newly added path?");
                        }
                        next;
                    }

                    # obtain the home node for this lif
                    my $node = $path_obj->lif_obj()->home_node();

                    # verify that all paths to faulted nodes are not accessible
                    if (grep(/^$node$/, @faulted_nodes)) {
                        # verify that this path is down because this lif should not be accessible
                        if ($path_obj->path_state() =~ /active|standby|passive/i) {

                            # if the lif owner is a reboot node then it is possible that an auto giveback occurred.
                            # this is not necessarily a warning
                            if (defined($path_validation_rules->{reboot_node}) and
                                    $path_validation_rules->{reboot_node} eq $node and
                                        $self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") == 1) {
                                # then this case is not an error
                            } else {
                                # realy it as a warning
                                push(@warnings_for_device, "The path: '" . $path_obj->path_name() .
                                     "' is active when it should be dead. It is a path to node $node, " .
                                     "which is in a faulted state.");
                            }
                        }
                    }

                    # verify that all paths to unfaulted nodes are accesible
                    if (grep(/^$node$/, @unfaulted_nodes)) {
                        # verify that this path is up because this lif should be accessible
                        if ($path_obj->path_state() !~ /active|standby|passive/i) {
                            push(@errors_for_device, "The path: '" . $path_obj->path_name() .
                                 "' is dead when it should be active. It is a path to node $node, " .
                                 "which is not in a faulted state.");
                        }
                    }

                    # verify alua states:
                    # (1) if this path is to a Lif on the HA partner, and the aggr owner is faulted, or
                    # (2) if this path is to a Lif on the aggr owner and the aggr owner is not faulted,
                    # then:
                    #      verify that the path to this lif is AO
                    if ($self->EventUtil::get_arg(-arg => "path_validation_verify_alua_states",
                                                  -objects => [$resource],
                                                  -default_to_env => 1) == 1) {
                        my $hapartner = $dev_obj->ontap_home_node_hapartner();
                        my $aggr_owner = $dev_obj->ontap_home_node();

                        # if this path is a lif on the aggr owner and we know the aggr owner is not
                        # in a faulted state, then the path must be AO
                        my $path_must_be_ao = 0;
                        if ($node eq $aggr_owner and grep(/^$aggr_owner$/, @unfaulted_nodes)) {
                            $path_must_be_ao = 1;
                        }

                        if ((($node eq $hapartner) and grep(/^$aggr_owner$/, @faulted_nodes)) or
                            $path_must_be_ao) {
                            # verify that this path is active optimized
                            if ($path_obj->alua_state() ne "AO") {
                                # if the aggr owner is a reboot node then it is possible that an auto giveback occurred.
                                # this may have caused this path's alua state to switch back to ANO because the reboot node is alive
                                # if this happened, then we can not deterministically know what the correct alua state for this path is
                                if (not $path_must_be_ao and (defined($path_validation_rules->{reboot_node}) and
                                                              $path_validation_rules->{reboot_node} eq $aggr_owner and
                                                              $self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") == 1)) {
                                    # we can't assume that this path's alua state be AO or ANO, skip any validation on it
                                } else {
                                    # we are certain this path must be AO
                                    if ($self->EventUtil::get_arg(-arg => "path_validation_verify_alua_states_abort_on_error",
                                                                  -objects => [$resource],
                                                                  -default_to_env => 1) == 1) {
                                        # otherwise treat it like an error
                                        push(@errors_for_device, "The path: '" . $path_obj->path_name() .
                                             "' should have alua state 'AO', but it is " . $path_obj->alua_state());
                                    } else {
                                        # otherwise treat it like an warning
                                        push(@warnings_for_device, "The path: '" . $path_obj->path_name() .
                                             "' should have alua state 'AO', but it is " . $path_obj->alua_state());
                                    }
                                }
                            }
                        }
                    }
                }
            }

            # if there are any path validation errors for this device, load them in
            if (scalar(@errors_for_device)) {
                my $error_string = join("\n\t", @errors_for_device);
                push(@errors, "MP Device: " . $dev_obj->mp_device_name() . "\n\t$error_string\n");
            }

            if (scalar(@warnings_for_device)) {
                my $warning_string = join("\n\t", @warnings_for_device);
                push(@warnings, "MP Device: " . $dev_obj->mp_device_name() . "\n\t$warning_string\n");
            }
        }
    }

    # If requested to do vol move path validation, do it here.
    if ($path_validation_rules->{validate_vol_move} == 1 and
        $self->EventUtil::get_arg(-arg => "path_validation_verify_alua_states") == 1) {
        my @vol_move_history = @{$path_validation_rules->{vol_move_history}};

        # walk through all the volumes
        foreach my $vol_move_item (@vol_move_history) {
            my $vserver_name = $vol_move_item->{vserver_name};
            my (@source_nodes) = $vol_move_item->{source_nodes};
            my (@dest_nodes) = $vol_move_item->{dest_nodes};
            my $vol_name = $vol_move_item->{vol_name};

            # find all LUNs in the mpdevmgr that live in this volume that have match home node of $dest node
            my @dev_objs = $mpdevmgr->get_filtered_devices(-vservers => [$vserver_name],
                                                           -ontap_home_nodes => \@dest_nodes,
                                                           -ontap_mode => "cmode",
                                                           -ontap_lun_path_regex => qr/\/vol\/${vol_name}\//);
            if ($mpdevmgr->error()) {
                $mpdevmgr->_clear_err();
                # this means that this client does not have any LUNs in the specified volume, this is ok
                next;
            }

            # verify alua states for each LUN in this volume/vserver
            foreach my $dev_obj (@dev_objs) {
                my $path_state_obj = $dev_obj->get_current_path_state();
                my $aggr_name = $dev_obj->ontap_aggregate_name();
                my $aggr_owner = $dev_obj->ontap_home_node();

                my @errors_for_device;

                # validate that the alua states switched properly for local/remote paths
                foreach my $path_obj ($path_state_obj->get_path_objs()) {
                    my $home_node = $path_obj->lif_obj()->home_node();

                    # if the home node for this path is on the list of destination nodes,
                    # then this path should be a local path
                    if (grep(/^$home_node$/, @dest_nodes)) {
                        # verify that this path's alua state is AO (local path to lun)
                        if ($path_obj->alua_state() ne "AO") {
                            push(@errors_for_device, $path_obj->path_name() . " should have ALUA state 'AO' after " .
                                 "vol ($vol_name) moved to node:aggr (${aggr_owner}:${aggr_name}). The current ALUA state is " . $path_obj->alua_state());
                        }
                    } else {
                        # this path is not the destination node list, therefore it should not have alua state AO
                        # since it is a remote path
                        if ($path_obj->alua_state() eq "AO") {
                            push(@errors_for_device, $path_obj->path_name() . " should not have ALUA state 'AO' (should be 'ANO') after " .
                                 "vol ($vol_name) moved to node:aggr (${aggr_owner}:${aggr_name})");
                        }
                    }
                }

                # check for any errors found for this device
                if (scalar(@errors_for_device)) {
                    my $error_string = join("\n\t", @errors_for_device);
                    push(@errors, uc($dev_obj->get_protocol()) . " Device: " . $dev_obj->mp_device_name() . "\n\t$error_string");
                }
            }
        }
    }

    # perform OS Specific validation
    my ($os_specific_validation_errors, $os_specific_validation_warnings) = $mpdevmgr->os_specific_validation();
    if ($mpdevmgr->error()) {
        $self->_set_err("Failed to perform os specific validation", $mpdevmgr->errmsg());
        return $return_hash;
    }

    # check if there were any warnings found
    if (scalar(@warnings)) {
        push(@{$return_hash->{warnings}}, @warnings);
    }
    if (scalar(@{$os_specific_validation_warnings})) {
        push(@{$return_hash->{warnings}}, @{$os_specific_validation_warnings});
    }

    # check if there were any errors found
    if (scalar(@errors)) {
        push(@{$return_hash->{errors}}, @errors);
    }
    if (scalar(@{$os_specific_validation_errors})) {
        push(@{$return_hash->{errors}}, @{$os_specific_validation_errors});
    }

    return $return_hash;
}

###############################################################################
# Method:    _wait_for_validate_paths()
# Objective: Given a previous timeout, wait the minimum wait time needed
#            which is the difference between the previous time out and the
#            min timeout
###############################################################################
sub _wait_for_validate_paths {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # the caller may have requested that we wait a minimum timeout
    # before performing path validation
    my $last_timeout = &ParamsMgr::get_arg(-arg => "last_timeout",
                                           -default => 0);
    my $last_timeout_name = &ParamsMgr::get_arg(-arg => "last_timeout_name",
                                                -default => "timeout");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine the min wait time from the current event/eventset
    my $pv_min_wait_time = $self->EventUtil::get_arg(-arg => "path_validation_min_wait_time",
                                                     -default => 0);

    my ($delta);

    # make sure we have slept AT LEAST pv_min_wait_time
    # seconds. If we have not, sleep the delta of pv_min_wait_time
    # and the passed in value.
    if ($last_timeout <= $pv_min_wait_time) {
        # calculate the additional amount of time we must sleep
        $delta = $pv_min_wait_time - $last_timeout;

        &SanUtil::nate_log('-comment',
                           "Sleeping an additional ($delta) seconds to allow paths to settle\n" .
                           "before performing path validation. This is in addition to the\n" .
                           "${last_timeout_name}=${last_timeout} seconds we already waited.");
        &SanUtil::nate_sleep($delta);
    } else {
        &SanUtil::nate_log('-comment',
                           "No additional sleep time is needed since we have already waited " .
                           "${last_timeout_name}=${last_timeout} seconds\nwhich is greater than " .
                           "-path_validation_min_wait_time ($pv_min_wait_time)");
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _is_steady_state_loaded()
# Objective: Determine if steady state is loaded on the specified SanClient
#            resource
###############################################################################
sub _is_steady_state_loaded {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $steady_state_loaded = $self->ResMgr::get_resource_attribute(-name => $client_name,
                                                                    -type => "sanclients",
                                                                    -attribute => "steady_state_loaded");
    if ($self->error()) {
        $self->_clear_err();
        return $EventUtil::FALSE;
    }

    # check if steady state is set on this client
    if (defined($steady_state_loaded) and $steady_state_loaded == $EventUtil::TRUE) {
        return $steady_state_loaded;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    _is_mpdevmgr_loaded()
# Objective: Determine if mpdevmgr is loaded on the specified SanClient
#            resource
###############################################################################
sub _is_mpdevmgr_loaded {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $mpdevmgr_loaded = $self->ResMgr::get_resource_attribute(-name => $client_name,
                                                                -type => "sanclients",
                                                                -attribute => "mpdevmgr_loaded");
    if ($self->error()) {
        $self->_clear_err();
        return $EventUtil::FALSE;
    }

    # check if steady state is set on this client
    if (defined($mpdevmgr_loaded) and $mpdevmgr_loaded == $EventUtil::TRUE) {
        return $mpdevmgr_loaded;
    } else {
        return $EventUtil::FALSE;
    }
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    # define the default values for Event parameters
    $self->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                       -arg_defaults =>
                                                           { path_validation => 1,
                                                             path_validation_on_takeover => 1,
                                                             path_validation_on_giveback => 1,
                                                             path_validation_after_lif_down => 1,
                                                             path_validation_after_lif_up => 1,
                                                             path_validation_after_switch_port_enable => 1,
                                                             path_validation_verify_alua_states => 1,
                                                             path_validation_abort_on_error => 1,
                                                             path_validation_init_state_abort_on_error =>
                                                                 sub { shift->EventUtil::get_arg(-arg => "path_validation_abort_on_error", -default => 1); },
                                                             path_validation_verify_alua_states_abort_on_error => 0,
                                                             path_validation_abort_script => "",
                                                             path_validation_abort_script_execute_on => "device",
                                                             path_validation_min_wait_time => 0,
                                                             path_validation_init_state => "",
                                                             path_validation_rule_type => "Cmode",
                                                             path_validation_fcp_init_state => "",
                                                             path_validation_iscsi_init_state => "" }
                                                       );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # load pathing information here on all sanclient objects
    $self->PathValidation::_load_pathing_information();
    if ($self->error()) {
        $self->_set_err("Failed to load pathing information for all sanclient objects");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    # document each parameter used in this module
    $self->EventUtil::document_event_param(-param => "path_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-path_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable path validation " .
                                               "on all SanClients that support the methods to do so.",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "path_validation_mpdevmgr_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-path_validation_mpdevmgr_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable MPDevMgr path validation " .
                                               "on all SanClients that support the methods to do so.",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "path_validation_on_takeover",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-path_validation_on_takeover => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable path validation during takeover",
                                           -overrides => ["path_validation"]);

    $self->EventUtil::document_event_param(-param => "path_validation_on_giveback",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-path_validation_on_giveback => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable path validation during takeover",
                                           -overrides => ["path_validation"]);

    $self->EventUtil::document_event_param(-param => "path_validation_after_lif_down",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-path_validation_after_lif_down => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable path validation after lif down operations",
                                           -overrides => ["path_validation"]);

    $self->EventUtil::document_event_param(-param => "path_validation_after_lif_up",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-path_validation_after_lif_up => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable path validation after lif up operationsr",
                                           -overrides => ["path_validation"]);

    $self->EventUtil::document_event_param(-param => "path_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-path_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if a path validation error is " .
                                               "detected during intial setup or during validation",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "path_validation_abort_script",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "any executable command",
                                           -example => "-path_validation_abort_script => \"perl /u/hrivera/path_validation_error_data_collection.pl\"",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies a command to run when a path validation error is encountered. " .
                                               "This does not apply to errors found during initial setup, only during validation.",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "path_validation_abort_script_execute_on",
                                           -package => __PACKAGE__,
                                           -default => "device",
                                           -format => "device|nate_control_server",
                                           -example => "-path_validation_abort_script_execute_on => \"nate_control_server\"",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Indicates whether to run the abort script on the device/system that had the ".
                                               "failure, or on the NATE control server.",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "path_validation_min_wait_time",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "integer, in seconds",
                                           -example => "-path_validation_min_wait_time => 300",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the minimum amount of time (seconds) that must be waited " .
                                               "before performing path validation");

    $self->EventUtil::document_event_param(-param => "path_validation_init_state",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "x:y:z (where x is the total number of paths to each LUN, y is the total number of alive paths, and z is the total number of dead paths)",
                                           -example => "-path_validation_init_state => \"4:4:0\"",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies what the pathing steady state should match during the initial load for all LUNs.",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "path_validation_rule_type",
                                           -package => __PACKAGE__,
                                           -default => "Cmode",
                                           -format => "7mode|Cmode",
                                           -example => "-path_validation_rule_type => \"7mode\"",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies what validation rules to use when validating paths.\n" .
                                               "\t7mode rules:\n" .
                                                   "\t\tFCP (at least half the paths to each LUN must be present during takeover mode).\n" .
                                                   "\t\tFCP (all paths to each must LUN must be present after a giveback).\n" .
                                                   "\t\tiSCSI (all paths to each must LUN must be present during takeover mode).\n" .
                                                   "\t\tiSCSI (all paths to each must LUN must be present after a giveback).\n" .
                                               "\tCmode rules:\n" .
                                                   "\t\tFCP (at least half the paths to each LUN must be present during takeover mode).\n" .
                                                   "\t\tFCP (all paths to each must LUN must be present after a giveback).\n" .
                                                   "\t\tiSCSI (at least  half the paths to each LUN must be present during takeover mode).\n" .
                                                   "\t\tiSCSI (all paths to each must LUN must be present after a giveback).\n",
                                           -resource_types => ["sanclients"]);


    $self->EventUtil::document_event_param(-param => "path_validation_fcp_init_state",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "x:y:z (where x is the total number of paths to each LUN, y is the total number of alive paths, and z is the total number of dead paths)",
                                           -example => "-path_validation_fcp_init_state => \"4:2:2\"",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies what the pathing steady state should match during the initial load for all FCP LUNs.",
                                           -resource_types => ["sanclients"],
                                           -overrides => ["path_validation_init_state"]);

    $self->EventUtil::document_event_param(-param => "path_validation_iscsi_init_state",
                                           -package => __PACKAGE__,
                                           -default => "",
                                           -format => "x:y:z (where x is the total number of paths to each LUN, y is the total number of alive paths, and z is the total number of dead paths)",
                                           -example => "-path_validation_iscsi_init_state => \"2:2:0\"",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Specifies what the pathing steady state should match during the initial load for all iSCSI LUNs.",
                                           -resource_types => ["sanclients"],
                                           -overrides => ["path_validation_init_state"]);

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

PathValidation.pm

=head1 Author

Robert Hunter

=head1 Description

An EventMan internal utility module designed to provide helper
methods to perform iSCSI & FCP Path Validation for SAN clients
during takeover / giveback scenarios

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-path_validation>

=over 4

Description: If set to 1, will enable path validation
on all SanClients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_on_takeover>

=over 4

Description: If set to 1, will enable path validation during takeover
on all SanClients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_on_giveback>

=over 4

Description: If set to 1, will enable path validation after giveback
on all SanClients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_after_lif_down>

=over 4

Description: If set to 1, will enable path validation after lif down operations
on all SanClients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_after_lif_up>

=over 4

Description: If set to 1, will enable path validation after lif up operations
on all SanClients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_after_switch_port_enable>

=over 4

Description: If set to 1, will enable path validation after switch port block enable
on all SanClients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_verify_alua_states>

=over 4

Description: If set to 1, will enable alua state verification on all paths.
ALUA state verification involves verifying all device paths have the expected
ALUA state after performing some sort of fault. The ALUA states are
AO (active-optimized) or ANO (active non-optimized). They indicate whether
a path is a local or remote path respectively.
Only applicable to SanClient resources that implement the MPDevMgr package.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a path validation error is
detected during intial setup or during validation.

Per Resource Capable: Yes

Resource Types: sanclients

Default: 1 (on)

=back

=head2 I<-path_validation_verify_alua_states_abort_on_error>

=over 4

Description: If set to 1, will abort if alua state errors are encountered.
Only applicable to SanClient resources that are supported by L<C<San::SanClient::MPDevMgr>>.
If set to 0, it will emit a wawrning if any alua state errors are encountered

Per Resource Capable: Yes

Resource Types: sanclients

Default: This defaults to the value of 'path_validation_abort_on_error'

=back

=head2 I<-path_validation_init_state_abort_on_error>

=over 4

Description: If set to 1, will abort if a client fails the initial path state check.
If set to 0, it will emit a warning during the init state check and move on. The clients
that fail the init state check will have all subsequent path validation checks disabled.
A useful way to use this parameter is if you want to have the ability to abort the test
on a misconfigured system, but you don't want to abort the test if path validation errors
are encountered during fault injection. To get that behavior you'll want to set
the following parameters in this way:

 path_validation_init_state_abort_on_error => 1
 path_validation_abort_on_error => 0

Per Resource Capable: Yes

Resource Types: sanclients

Default: If left unset, this defaults to the value of 'path_validation_abort_on_error'

=back

=head2 I<-path_validation_abort_script>

=over 4

Description: Specifies a command to run when a path validation
error is encountered. This does not apply to errors found during
initial setup, only during validation.

Dependent on: -path_validation_abort_on_error => 1

Per Resource Capable: Yes

Resource Types: sanclients

Default: <empty string>

=back

=head2 I<-path_validation_abort_script_execute_on>

=over 4

Description: Indicates whether to run the abort script on
the device/system that had the failure, or on the NATE
control server. Valid values are "device | nate_control_server".

Dependent on: -path_validation_abort_on_error => 1, -path_validation_abort_script => "<command>"

Per Resource Capable: Yes

Resource Types: sanclients

Default: "device"

=back

=head2 I<-path_validation_min_wait_time>

=over 4

Description: Specifies the minimum amount of time (seconds) that
must be waited before performing path validation

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-path_validation_rule_type>

=over 4

Description: Specifies what validation rules to use when validating paths. Only applicable
to legacy path validation routines (hpux,aix,solaris)
 7mode rules:
   FCP (at least half the paths to each LUN must be present during takeover mode).
   FCP (all paths to each must LUN must be present after a giveback).
   iSCSI (all paths to each must LUN must be present during takeover mode).
   iSCSI (all paths to each must LUN must be present after a giveback).
 Cmode rules:
   FCP (at least half the paths to each LUN must be present during takeover mode).
   FCP (all paths to each must LUN must be present after a giveback).
   iSCSI (at least  half the paths to each LUN must be present during takeover mode).
   iSCSI (all paths to each must LUN must be present after a giveback).

Per Resource Capable: Yes

Resource Types: sanclients

Format: "7mode | Cmode"

Default: "Cmode"

=back

=head2 I<-path_validation_init_state>

=over 4

Description: Specifies what the pathing steady state should
match during the initial load for all LUNs.

Per Resource Capable: Yes

Resource Types: sanclients

Format: x:y:z (where x is the total number of paths to each LUN,
y is the total number of alive paths, and z is the total number of dead paths)

Default: <empty string>

=back

=head2 I<-path_validation_fcp_init_state>

=over 4

Description: Specifies what the pathing steady state should
match during the initial load for all FCP LUNs.

Per Resource Capable: Yes

Resource Types: sanclients

Overrides: -path_validation_init_state

Format: x:y:z (where x is the total number of paths to each LUN,
y is the total number of alive paths, and z is the total number of dead paths)

Default: <empty string>

=back

=head2 I<-path_validation_iscsi_init_state>

=over 4

Description: Specifies what the pathing steady state should
match during the initial load for all iSCSI LUNs.

Per Resource Capable: Yes

Resource Types: sanclients

Overrides: -path_validation_init_state

Format: x:y:z (where x is the total number of paths to each LUN,
y is the total number of alive paths, and z is the total number of dead paths)

Default: <empty string>

=back

=head1 B<Public Methods>

=head2 B<validate_paths_after_takeover()>

Performs path validation for all SanClient resources in the current
Event/EventSet/EventMan resource scope assuming controllers are in a
takeover state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off path validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do path validation. This
will be used to calculate additional path validation wait time that needs
to be applied before performing path validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: ""

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -path_validation
 -path_validation_on_takeover
 -path_validation_on_giveback
 -path_validation_abort_on_error
 -path_validation_abort_script
 -path_validation_abort_script_execute_on
 -path_validation_min_wait_time
 -path_validation_init_state
 -path_validation_rule_type
 -path_validation_fcp_init_state
 -path_validation_iscsi_init_state

=item Examples

 $sanfaults_obj->PathValidation::validate_paths_after_takeover(-id_extension => "takeover_on_node_A",
                                                               -last_timeout => "600",
                                                               -last_timeout_name => "filer_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<validate_paths_after_giveback()>

Performs path validation for all SanClient resources in the current
Event/EventSet/EventMan resource scope assuming controllers are in a
giveback state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off path validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do path validation. This
will be used to calculate additional path validation wait time that needs
to be applied before performing path validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: ""

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -path_validation
 -path_validation_on_takeover
 -path_validation_on_giveback
 -path_validation_abort_on_error
 -path_validation_abort_script
 -path_validation_abort_script_execute_on
 -path_validation_min_wait_time
 -path_validation_init_state
 -path_validation_rule_type
 -path_validation_fcp_init_state
 -path_validation_iscsi_init_state

=item Examples

 $sanfaults_obj->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_node_A",
                                                               -last_timeout => "600",
                                                               -last_timeout_name => "filer_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=cut
