# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing Port Validation
## @author derricks@netapp.com
## @status private
## @pod here

package PortValidation;

# NATE packages
use Tharn;
use NATE::BaseException qw(:try);

# NACL packages
use NACL::CS::NetworkInterface;
use NACL::Exceptions::NoElementsFound;

# San packages
use San::SanClient::ClientUtil;
use San::SanUtil;
use ParamsMgr;
use Data::Dumper;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    validate_ports_after_giveback()
# Objective: Issue port validation assuming a giveback has taken place
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_ports_after_giveback {
    my ($self, %args) = @_;

    # determine if port validation on port up is enabled in the current event/eventset
    my $port_validation_on_giveback = $self->EventUtil::get_arg(-arg => "port_validation_on_giveback");
    if ($port_validation_on_giveback == 0) {
        return $EventUtil::TRUE;
    }

    $self->PortValidation::_validate_ports(%args, -type => "giveback");
    if ($self->error()) {
        $self->_set_err("Port validation after giveback failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_ports_after_port_up()
# Objective: Issue port validation assuming a port was blocked to an sfo or switch event
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_ports_after_port_up {
    my ($self, %args) = @_;

    # determine if port validation on port up is enabled in the current event/eventset
    my $port_validation_on_port_up = $self->EventUtil::get_arg(-arg => "port_validation_on_port_up");
    if ($port_validation_on_port_up == 0) {
        return $EventUtil::TRUE;
    }

    $self->PortValidation::_validate_ports(%args, -type => "port_up");
    if ($self->error()) {
        $self->_set_err("Port validation after port up failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_ports_after_lif_down()
# Objective: Issue port validation assuming a lif down has occurred
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_ports_after_lif_down {
    my ($self, %args) = @_;

    # determine if port validation on lif down is enabled in the current event/eventset
    my $port_validation_on_lif_down = $self->EventUtil::get_arg(-arg => "port_validation_on_lif_down");
    if ($port_validation_on_lif_down == 0) {
        return $EventUtil::TRUE;
    }

    $self->PortValidation::_validate_ports(%args, -type => "lif_down");
    if ($self->error()) {
        $self->_set_err("Port validation after lif down failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    validate_ports_after_lif_up()
# Objective: Issue port validation assuming a lif up has occurred
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_ports_after_lif_up {
    my ($self, %args) = @_;

    # determine if port validation on lif up is enabled in the current event/eventset
    my $port_validation_on_lif_up = $self->EventUtil::get_arg(-arg => "port_validation_on_lif_up");
    if ($port_validation_on_lif_up == 0) {
        return $EventUtil::TRUE;
    }

    $self->PortValidation::_validate_ports(%args, -type => "lif_up");
    if ($self->error()) {
        $self->_set_err("Port validation after lif up failed");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _get_port_info
# Objective: Build a hash of port information for each node passed in.
#            Format of port hash:
#
#                 <switch> =>
#                       <port> =>
#                           resource_type => Type of resource (fc_target_wwpn, etc)
#                           wwpns         => hash of wwpns logged in
#
###############################################################################
sub _get_port_info {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @resource_types      = &ParamsMgr::get_arg( -arg      => "resource_types",
                                                   -required => 0,
                                                   -default  => [],
                                                   -ref_type => "ARRAY");

    my @switches            = &ParamsMgr::get_arg( -arg      => "switches",
                                                   -required => 0,
                                                   -default  => [],
                                                   -ref_type => "ARRAY");

    my $reconnect_switches  = &ParamsMgr::get_arg( -arg      => "reconnect_switches",
                                                   -default  => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->set_err(&ParamsMgr::get_error());
        return {};
    }

    # Check to see if we were passed in a list of resource types.
    if (scalar(@resource_types) > 0) {
        # Make sure we don't have any invalid resource types
        foreach my $resource_type (sort @resource_types) {
            # Check for an illegal resource type
            if ($resource_type =~ /apcon/i or $resource_type !~ /^fc_/i) {
                $self->_set_err("Invalid resource type specified.  Only 'non-apcon' and 'fc_' resource types " .
                                "will work with port validation");
                return {};
            }
        }
    } else {
        # We weren't given any resource types so default to all of the supported types
        @resource_types = ("fc_target_wwpn",
                           "fc_target_lif_wwpn",
                           "fc_isl_wwpn",
                           "fc_initiator_wwpn");
    }

    # Grab the switch info hash from ResMgr
    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");
    if ($self->error()) {
        $self->_set_err("Unable to get resource 'switch_info' from the Resource Manager");
        return {};
    }

    unless(defined($switch_info->{obj})) {
        $self->_set_err("Unable to find the attribue 'obj' on the switch info hash retrieved from ResMgr");
        return {};
    }
    my $switch_info_hash = $switch_info->{obj};

    # Make sure we have at least one switch in the hash
    unless (exists($switch_info_hash->{sanswitches}) and scalar(%{$switch_info_hash->{sanswitches}}) > 0) {
        $self->_set_err("Unable to find any switch resources in the switch info hash");
        &SanUtil::nate_log('-comment', "Dumping the switch info hash:\n" . Dumper($switch_info_hash));
        return {};
    }

    # Get the lif and resource_name regex filters
    my $pv_lif_name_regex_filter = $self->EventUtil::get_arg(-arg => "port_validation_lif_name_regex_filter");
    my $pv_resource_name_regex_filter = $self->EventUtil::get_arg(-arg => "port_validation_resource_name_regex_filter");

    # Build out the list of switches we are going to iterate over. If we were
    # given a list of switchs, then use that; otherwise, use the list of switches
    # defined in the switch info hash.
    if (scalar(@switches) > 0) {
        foreach my $switch (@switches) {
            # Make sure the switch exists in the switch info hash
            unless (exists($switch_info_hash->{sanswitches}->{$switch})) {
                $self->_set_err("Unable to find switch '$switch' in the switch info hash");
                return {};
            }
        }
    } else {
        foreach my $switch (sort keys %{$switch_info_hash->{sanswitches}}) {
            push (@switches, $switch);
        }
    }

    # Build out the port hash with the following format
    #    $port_hash =>
    #        <switch> =>
    #             <port> =>
    #                 resource_type => Type of resource (fc_target_wwpn, etc)
    #                 wwpns         => hash of wwpns logged in
    #
    my $port_hash = {};

    foreach my $switch (sort @switches) {
        # Grab the object for the switch
        my $sanswitch_obj = $self->EventUtil::get_sanswitch_obj(-switch_name => $switch);
        if ($self->error()) {
            $self->_set_err("Could not obtain SanSwitch object for switch '$switch'");
            return {};
        }

        # Get the hosttype of the switch
        my $hosttype = $sanswitch_obj->hosttype();
        if ($sanswitch_obj->error()) {
            $self->_set_err("Could not obtain hosttype for switch '$switch':", $sanswitch_obj->errmsg());
            return {};
        }

        # Check to see if this is an ethernet switch.  If it is,
        # then skip it.  Port Validation is only for FC switches.
        if ($hosttype =~ /ethernet/i) {
            # We have an ethernet switch so skip it
            &SanUtil::nate_log('-comment', "Skipping Port Validation for switch '$switch' because it " .
                                           "has hosttype of '$hosttype'");
            next;
        }

        # Reconnect the switches
        if ($reconnect_switches == $EventUtil::TRUE) {
            $self->EventUtil::reconnect_all_sanswitches(-switch_names             => [$switch],
                                                        -disconnect_existing_conn => $EventUtil::FALSE);
            if ($self->error()) {
                return {};
            }
        }

        # Grab the portmgr object
        my $portmgr_obj = $sanswitch_obj->portmgr();
        unless (ref($portmgr_obj) =~ /portmgr/i) {
            # Looks like we weren't able to get a portmgr object so try creating one
            # without the load_nameserver flag since we are just going to invalidate
            # the cache anyway.
            $portmgr_obj = $sanswitch_obj->create_portmgr(-load_nameserver => 0);
            if ($sanswitch_obj->error()) {
                # Disconnect the switch we reconnected.  We want to
                # do this because port validation failures may not result in
                # a failure...but instead a warning
                if ($reconnect_switches == $EventUtil::TRUE) {
                    $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch],
                                                                 -ignore_errors => $EventUtil::TRUE);
                }
                $self->_set_err("Unable to retrieve or create a PortMgr object for switch '$switch':\n",
                                $sanswitch_obj->errmsg());
                return {};
            }
        }

        # Invalidate the nameserver cache so it gets the data fresh
        $sanswitch_obj->invalidate_items_cache_data("nameserver");

        # Grab the nameserver info
        my $ns = $portmgr_obj->get_nameserver("1");
        if ($portmgr_obj->error) {
            # Disconnect the switch we reconnected.  We want to
            # do this because port validation failures may not result in
            # a failure...but instead a warning
            if ($reconnect_switches == $EventUtil::TRUE) {
                $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch],
                                                             -ignore_errors => $EventUtil::TRUE);
            }
            $self->_set_err("Unable to retrieve the nameserver info for switch '$sanswitch':\n", $portmgr_obj->errmsg());
            return {};
        }

        foreach my $resource_type (sort @resource_types) {
            # Check to make sure the resource type exists in the info hash.  If it doesn't,
            # it isn't necessarily an illegal type.
            next unless (exists($switch_info_hash->{sanswitches}->{$switch}->{$resource_type}));

            # Walk all of the resources within the given type
            foreach my $wwpn (sort keys %{$switch_info_hash->{sanswitches}->{$switch}->{$resource_type}}) {
                # Make sure the port_number is defined before continuing
                my $ports = $switch_info_hash->{sanswitches}->{$switch}->{$resource_type}->{$wwpn}->{'port_number'};

                unless (ref($ports) eq "ARRAY" and $ports->[0] ne "") {
                    # Disconnect the switch we reconnected.  We want to
                    # do this because port validation failures may not result in
                    # a failure...but instead a warning
                    if ($reconnect_switches == $EventUtil::TRUE) {
                        $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch],
                                                                     -ignore_errors => $EventUtil::TRUE);
                    }

                    $self->_set_err("Unable to determine the port number for wwpn '$wwpn' on switch '$switch' " .
                                    "with resource type '$resource_type'");
                    return {};
                }

                # pull out the first port in the list (this is ok since fibre channel should only ever have one port)
                my $port = $ports->[0];

                # Store the resource type for the port in the port hash
                $port_hash->{$switch}->{$port}->{"resource_type"} = $resource_type;

                # Special case for VDX switches as fcoe port numbers can change, except for isls
                if ($sanswitch_obj->isa("San::SanSwitch::FC::Switch::BrocadeVDXFCSwitch") and
                    not ($resource_type =~ /isl/i)) {
                    $port = $portmgr_obj->lookup_fcoe_port($port);
                    if ($self->error()) {
                        $self->_set_err("Unable to lookup port $port");
                        return {};
                    }
                }

                # Look up the port in the nameserver data
                unless (exists($ns->{$port})) {
                    # Disconnect the switch we reconnected.  We want to
                    # do this because port validation failures may not result in
                    # a failure...but instead a warning
                    if ($reconnect_switches == $EventUtil::TRUE) {
                        $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch],
                                                                     -ignore_errors => $EventUtil::TRUE);
                    }

                    $self->_set_err("Unable to find port '$port' in the nameserver information for switch '$switch'");
                    return {};
                }

                # Look up the wwpn data for the port
                unless (exists($ns->{$port}->{wwpn})) {
                    # If a port is offline or everything is logged out then there won't be any
                    # wwpn field in the nameserver hash for the port;
                    next;
                }

                # If the wwpns hash hasn't been initialized yet, go ahead and
                # initialize it. We want to do this so the key will exist even
                # if we end up not having any wwpns logged in for that port.
                unless (exists ($port_hash->{$switch}->{$port}->{wwpns})) {
                    $port_hash->{$switch}->{$port}->{wwpns} = {};
                }

                # Split the namserver wwpns into an array.
                my @ns_wwpns = split(/\;/, $ns->{$port}->{wwpn});

                # We need to iterate over all of the wwpns on the nameserver
                # port data and add them to the port hash.  We are storing the
                # wwpn data in a hash instead of an array so it is easier to
                # compare later.
                foreach my $ns_wwpn (@ns_wwpns) {
                    # If this is a qlogic target (starts with 50:), we need to skip
                    # it if we are in cmode because  physical targets are have random
                    # wwpns generated  at boot so they will likely change when the node boots.
                    # We don't want them in the output.
                    if ($ns_wwpn =~ /^50/ and grep(/^2/, @ns_wwpns) > 0) {
                        # Looks like this is cmode physical card and we have at least
                        # one lif so we can skip this wwpn.
                        next;
                    }

                    # Lowercase the wwpn and add colons
                    $ns_wwpn = lc($ns_wwpn);
                    $ns_wwpn = &SanUtil::wwn_add_colons(-wwn => $ns_wwpn);

                    # Add it to the port hash
                    $port_hash->{$switch}->{$port}->{wwpns}->{$ns_wwpn} = 1;
                } ##END foreach my $ns_wwpn (@ns_wwpns) {
            } ##END foreach my $wwpn (sort keys %{$switch_info_hash->{sanswitches}->{$switch}->{$resource_type}}) {
        } ##END foreach my $resource_type (sort @resource_types) {

        # Disconnect the switch we reconnected.
        if ($reconnect_switches == $EventUtil::TRUE) {
            $self->EventUtil::disconnect_all_sanswitches(-switch_names  => [$switch],
                                                         -ignore_errors => $EventUtil::TRUE);
        }
    } ##END foreach my $switch (sort @switches) {

    # We now need to see if any filtering needs to happen
    foreach my $switch (sort keys %$port_hash) {
        foreach my $port (sort keys %{$port_hash->{$switch}}) {
            foreach my $wwpn (sort keys %{$port_hash->{$switch}->{$port}->{wwpns}}) {
                # Try to get the lif and resource name for the wwpn
                my ($lif_name, $resource_name) = $self->PortValidation::_get_wwpn_name(-wwpn => $wwpn);
                if ($self->error()) {
                    return {};
                }

                if (defined($pv_lif_name_regex_filter) and $lif_name ne "") {
                    # Try to match the filter
                    unless ($lif_name =~ $pv_lif_name_regex_filter) {
                        # The filter didn't match so remove the wwpn from the
                        # port hash
                        delete($port_hash->{$switch}->{$port}->{wwpns}->{$wwpn});
                        next;
                    }
                }
                if (defined($pv_resource_name_regex_filter)) {
                    unless ($resource_name =~ $pv_resource_name_regex_filter) {
                        # The filter didn't match so remove the wwpn from the
                        # port hash
                        delete($port_hash->{$switch}->{$port}->{wwpns}->{$wwpn});
                        next;
                    }
                }
                # Get the lif name
            } #END foreach my $wwpn (sort keys %{$port_hash->{$switch}->{$port}->{wwpns}}) {
        } ##END foreach my $port (sort keys %{$port_hash->{$switch}}) {
    } ##END foreach my $switch (sort keys %$port_hash) {

    return $port_hash;
}

###############################################################################
# Method:    _get_wwpn_name()
# Objective: Get the name of the lif and  resource name associated with the wwpn.  If no name is
#            found, return an empty string but don't set an error.
###############################################################################
sub _get_wwpn_name {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $wwpn_to_find = &ParamsMgr::get_arg(-arg => "wwpn", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ("", "");
    }

    # Lowercase the wwpn and add colons
    $wwpn_to_find = lc($wwpn_to_find);
    $wwpn_to_find = &SanUtil::wwn_add_colons(-wwn => $wwpn_to_find);

    # Grab the switch info hash from ResMgr
    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");
    if ($self->error()) {
        $self->_set_err("Unable to get resource 'switch_info' from the Resource Manager");
        return ("", "");
    }

    unless(defined($switch_info->{obj})) {
        $self->_set_err("Unable to find the attribue 'obj' on the switch info hash retrieved from ResMgr");
        return ("", "");
    }

    foreach my $switch_name (sort keys %{$switch_info->{obj}->{sanswitches}}) {
        foreach my $port_type (sort keys %{$switch_info->{obj}->{sanswitches}->{$switch_name}}) {
            if (exists($switch_info->{obj}->{sanswitches}->{$switch_name}->{$port_type}->{$wwpn_to_find})) {
                # Hoooray, we found it.  See if there is a lif name or resource
                # based on what name type was requested.

                # Grab the hash for the wwpn for easier readability
                my $wwpn_sub_hash = $switch_info->{obj}->{sanswitches}->{$switch_name}->{$port_type}->{$wwpn_to_find};

                my $lif_name      = "";
                my $resource_name = "";

                if (defined($wwpn_sub_hash->{'lif_name'})) {
                    $lif_name = $wwpn_sub_hash->{'lif_name'};
                }

                if (defined($wwpn_sub_hash->{'resource_name'})) {
                    $resource_name = $wwpn_sub_hash->{'resource_name'};
                }

                return ($lif_name, $resource_name);
            } ##END if (exists($switch_info->{obj}->{sanswitches}->{$switch_name}->{$port_type}->{$wwpn_to_find})) {
        } ##END foreach my $port_type (sort keys %{$switch_info->{obj}->{sanswitches}->{$switch_name}}) {
    } ##END foreach my $resource_name (sort keys %{$switch_info->{obj}->{sanswitches}}) {

    # Looks like we didn't find it so return an empty string
    return ("", "");
}

###############################################################################
# Method:    _validate_ports()
# Objective: Validate all of the ports on a given set of switches against the
#            steady state.
###############################################################################
sub _validate_ports {
    my ($self, %args) = @_;

    # Check to make sure we have switch objects in the environment.  If we
    # don't then bypass port validation.
    my (@sanswitches) = $self->EventUtil::get_sanswitch_objs(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain all SanSwitch resource objects");
        return $EventUtil::FALSE;
    }

    if (scalar(@sanswitches) == 0) {
        &SanUtil::nate_log('-comment', "Skipping port validation since there are no switches " .
                                       "in the environment");
        return $EventUtil::TRUE;
    }

    &ParamsMgr::process_args(%args);

    my $id_extension      = &ParamsMgr::get_arg(-arg => "id_extension", -required => 1);
    my $type              = &ParamsMgr::get_arg(-arg => "type", -required => 1);
    my $last_timeout      = &ParamsMgr::get_arg(-arg => "last_timeout", -default => 0);
    my $last_timeout_name = &ParamsMgr::get_arg(-arg => "last_timeout_name", -default => "timeout");
    my @resource_types    = &ParamsMgr::get_arg(-arg => "resource_type",
                                                -ref_type => "ARRAY",
                                                -required => 0,
                                                -default => ["nacl_nodes_cm", "sanfilers7m"]);

    # If the caller wants to specify a list of ports on a single switch,
    # the can do so.
    my $caller_switch     = &ParamsMgr::get_arg(-arg => "switch", -default => "");
    my @caller_ports      = &ParamsMgr::get_arg(-arg => "ports",
                                                -ref_type => "ARRAY",
                                                -required => 0,
                                                -default => []);

    # The caller has the option to pass in a list of swtiches to narrow the scope.  This may be done
    # if they are doing sequential port blocks and only want to validate the ports on the switch
    # they blocked.
    my @switches = &ParamsMgr::get_arg(-arg => "switches",
                                       -ref_type => "ARRAY",
                                       -required => 0,
                                       -default => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Make sure we received a valid type of "port_up", "lif_up", or "lif_down"
    unless ($type eq "port_up" or $type eq "lif_up" or $type eq "lif_down" or $type eq "giveback") {
        $self->_set_err("Invalid type '$type' specified.  Valid types are 'port_up', 'giveback', 'lif_up', " .
                        "and 'lif_down'");
        return $EventUtil::FALSE;
    }

    # determine if port validation is enabled in the current event/eventset
    my $port_validation = $self->EventUtil::get_arg(-arg => "port_validation");

    # if port validation is not enabled for this current event, then return immediately
    if ($port_validation == 0) {
        &SanUtil::nate_log('-comment', "Skipping Port Validation for this event because it was disabled");
        return $EventUtil::TRUE;
    }

    # determine if we want to abort on error
    my $pv_abort_on_error = $self->EventUtil::get_arg(-arg => "port_validation_abort_on_error");

    # Grab the port hash from ResMgr
    my $resource_hash = $self->ResMgr::get_resource(-name => "port_info",
                                                    -type => "_GLOBALS");
    if ($self->error()) {
        $self->_set_err("Unable to get resource 'port_info' from the Resource Manager");
        return $EventUtil::FALSE;
    }

    unless(defined($resource_hash->{obj})) {
        $self->_set_err("Unable to find the attribue 'obj' on the port info hash retrieved from ResMgr");
        return $EventUtil::FALSE;
    }

    my $steady_state_port_hash = $resource_hash->{obj};

    # If we were passed in a single switch, make sure it is in the
    # steady state hash.
    if ($caller_switch ne "") {
        # Make sure we have steady state data for the switch
        unless (exists ($steady_state_port_hash->{$caller_switch})) {
            $self->_set_err("Unable to find switch '$caller_switch' in the steady state port hash");
            return $EventUtil::FALSE;
        }
    }

    # If we were passed in a list of switches, make sure they
    # are in the steady state hash.
    if (scalar(@switches) > 0) {
        foreach my $switch (@switches) {
            unless (exists ($steady_state_port_hash->{$switch})) {
                $self->_set_err("Unable to find switch '$switch' in the steady state port hash");
                return $EventUtil::FALSE;
            }
        }
    }

    # wait port validation min wait time if a value for
    # last timeout was passed in
    if ($last_timeout > 0) {
        $self->PortValidation::_wait_for_validate_ports(-last_timeout      => $last_timeout,
                                                        -last_timeout_name => $last_timeout_name);
    }

    # get the next padded runid from the event
    my $event_obj = $self->EventUtil::get_current_event_obj();

    my @subtests = ();
    if ($type eq "lif_down") {
        # Get list of wwpns that are down
        my $switch_wwpn_hash = $self->PortValidation::_get_lif_wwpn_hash(-state => "down");
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the lif objects for the lifs that are down");
            return $EventUtil::FALSE;
        }

        # Make sure we have some lifs that are down
        unless (scalar(keys %$switch_wwpn_hash) > 0) {
            &SanUtil::nate_log("-comment", "No fcp lifs were found in the down state.  Port validation " .
                                           "of downed lifs will be skipped.");
            return $EventUtil::TRUE;
        }

        my $port_validation_runid_padded = $event_obj->increment_subtest_runid_counter();

        # run port validation as a subtest for each switch
        foreach my $switch (sort keys %$switch_wwpn_hash) {
            my $id = $port_validation_runid_padded . "_verify_port_info_${switch}_${id_extension}";
            my $subtest = subtest(sub {
                                        $self->PortValidation::_validate_ports_by_wwpn(-switch    => $switch,
                                                                                       -state     => "down",
                                                                                       -port_hash => $switch_wwpn_hash->{$switch});
                                        if ($self->error()) {
                                            $self->_set_err("Port validation failed for switch '$switch'");
                                            # perform a graceful failure if this client has pv_abort_on_error == 0
                                            $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                            return $EventUtil::FALSE;
                                        }
                                        $self->SubtestComm::subtest_complete();
                                        return $EventUtil::TRUE;
                                        }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        }
    } elsif ($type eq "lif_up") {
        # Get list of wwpns that are up
        my $switch_wwpn_hash = $self->PortValidation::_get_lif_wwpn_hash(-state => "up");
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the lif objects for the lifs that are up");
            return $EventUtil::FALSE;
        }

        # Make sure we have some lifs that are up
        unless (scalar(keys %$switch_wwpn_hash) > 0) {
            &SanUtil::nate_log("-comment", "No fcp lifs were found in the up state.  Port validation " .
                                           "of downed lifs will be skipped.");
            return $EventUtil::TRUE;
        }

        my $port_validation_runid_padded = $event_obj->increment_subtest_runid_counter();

        # run port validation as a subtest for each switch
        foreach my $switch (sort keys %$switch_wwpn_hash) {
            my $id = $port_validation_runid_padded . "_verify_port_info_${switch}_${id_extension}";
            my $subtest = subtest(sub {
                                        $self->PortValidation::_validate_ports_by_wwpn(-switch    => $switch,
                                                                                       -state     => "up",
                                                                                       -port_hash => $switch_wwpn_hash->{$switch});
                                        if ($self->error()) {
                                            $self->_set_err("Port validation failed for switch '$switch'");
                                            # perform a graceful failure if this client has pv_abort_on_error == 0
                                            $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                            return $EventUtil::FALSE;
                                        }
                                        $self->SubtestComm::subtest_complete();
                                        return $EventUtil::TRUE;
                                        }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        }
    } else {
        # See if we were given a single switch by the caller.  If we were, then we
        # only want to process that switch and not all switches.
        if ($caller_switch ne "") {
            # run port validation for the switch
            my $port_validation_runid_padded = $event_obj->increment_subtest_runid_counter();
            my $id = $port_validation_runid_padded . "_verify_port_info_${caller_switch}_${id_extension}";
            my $subtest = subtest(sub {
                                        $self->PortValidation::_validate_ports_by_switch(-switch => $caller_switch,
                                                                                         -ports  => \@caller_ports);
                                        if ($self->error()) {
                                            $self->_set_err("Port validation failed for switch '$caller_switch'");
                                            # perform a graceful failure if this client has pv_abort_on_error == 0
                                            $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                            return $EventUtil::FALSE;
                                        }
                                        $self->SubtestComm::subtest_complete();
                                        return $EventUtil::TRUE;
                                    }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
            push(@subtests, $subtest);
        } else {
            if (scalar(@switches) == 0) {
                # We weren't given any switches.  If this is a giveback validation
                # then get a list of switches for only the cluster nodes; otherwise,
                # get a list of all switchesl
                if ($type eq "giveback") {
                    # Grab the switch info hash from ResMgr
                    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                                  -type => "_GLOBALS");
                    if ($self->error()) {
                        $self->_set_err("Unable to get resource 'switch_info' from the Resource Manager");
                        return $EventUtil::FALSE;
                    }

                    # Get a list of switches and ports for each resource type
                    # Use a hash to store the switches we find.  We do this
                    # to avoid duplicate switch names getting added.
                    my $switch_hash = {};
                    foreach my $resource_type (sort @resource_types) {
                        if (exists($switch_info->{obj}->{$resource_type})) {
                            foreach my $resource_name (sort keys %{$switch_info->{obj}->{$resource_type}}) {
                                foreach my $port_type (sort keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}}) {
                                    # We only care about fc_target_wwpn and fc_target_lif_wwpn
                                    # so skip any others
                                    unless ($port_type =~ /^fc_target/) {
                                        next;
                                    }

                                    foreach my $wwpn (sort keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_type}}) {
                                        # Get the wwpn sub hash for easier readability
                                        my $wwpn_sub_hash = $switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_type}->{$wwpn};

                                        # Get the switch name for the resource
                                        if (defined($wwpn_sub_hash->{'switch_name'})
                                            and $wwpn_sub_hash->{'switch_name'} ne "") {
                                            $switch_hash->{$wwpn_sub_hash->{'switch_name'}} = 1;
                                        }
                                    } ##END foreach my $wwpn (sort keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}->{$port_type}}) {
                                } ##END foreach my $port_type (sort keys %{$switch_info->{obj}->{$resource_type}->{$resource_name}}) {
                            } ##END foreach my $resource_name (sort keys %{$switch_info->{obj}->{$resource_type}}) {
                        } ##END if (exists($switch_info->{obj}->{$resource_type})) {
                    } ##END foreach my $resource_type (sort @resource_types) {

                    # Parse the switch hash and use it to build the @switches array
                    foreach my $switch (sort keys %$switch_hash) {
                        push(@switches, $switch);
                    }
                } else {
                    # This isn't a giveback so generate a list of switches
                    # from the steady state hash
                    foreach my $switch (sort keys %$steady_state_port_hash) {
                        push(@switches, $switch);
                    }
                }
            }

            my $port_validation_runid_padded = $event_obj->increment_subtest_runid_counter();

            # Process each switch in @switches array
            foreach my $switch (sort @switches) {
                # run port validation for the switch
                my $id = $port_validation_runid_padded . "_verify_port_info_${switch}_${id_extension}";
                my $subtest = subtest(sub {
                                            $self->PortValidation::_validate_ports_by_switch(-switch => $switch);
                                            if ($self->error()) {
                                                $self->_set_err("Port validation failed for switch '$switch'");
                                                # perform a graceful failure if this client has pv_abort_on_error == 0
                                                $self->SubtestComm::subtest_fail(-graceful => not $pv_abort_on_error);
                                                return $EventUtil::FALSE;
                                            }
                                            $self->SubtestComm::subtest_complete();
                                            return $EventUtil::TRUE;
                                        }, -runid => $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});
                push(@subtests, $subtest);
            } ##END foreach my $switch (sort @switches) {
        } ##END if ($caller_switch ne "") {
    } ##END if ($type eq "lif_down") {

    # wait for all the path validation subtests to complete
    if (scalar(@subtests)) {
        $self->wait_for_subtests(-subtests => \@subtests,
                                 -dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 1,
                                 -enable_sync => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -async => 0);
        if ($self->error()) {
            $self->_set_err("Port validation failed on a switch");
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _get_lif_wwpn_hash()
# Objective: Get a list of wwpns for the fcp lifs in a certain state
###############################################################################
sub _get_lif_wwpn_hash {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $state = &ParamsMgr::get_arg(-arg => "state", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Check that we received a valid state
    unless ($state eq "up" or $state eq "down") {
        $self->_set_err("Invalid state of '$state' provided.  Valid states are 'up' and 'down'");
        return {};
    }

    # first get all cluster names that are in scope
    my @cluster_names = $self->EventUtil::get_nacl_clusters_cm();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the nacl_cluster_cm resource names");
        return {};
    }

    my @wwpns = ();

    foreach my $cluster_name (@cluster_names) {
        # Grab all of the vserver objects in this cluster.
        # We need to generate a list of all
        # wwpns for all the lifs that are down for the vservers
        my @vserver_objs = $self->EventUtil::get_vserver_objs(clusters => [$cluster_name]);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the vserver objects");
            return {};
        }

        # Process each vserver and find the lif objects of the requested type
        foreach my $vserver_obj (@vserver_objs) {
            # Get the vserver name
            my $vserver_name = $vserver_obj->vserver();
            if (not defined($vserver_name) or not length($vserver_name)) {
                $self->_set_err("Unable to retrieve vserver name for the vserver object");
                return {};
            }

            # Get the list of downed lifs for the vserver.  We need this list to make sure
            # that we are returning the correct state
            my @downed_lifs = $self->PortValidation::_get_downed_lifs(-vserver => $vserver_name,
                                                                      -cluster => $cluster_name);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve downed lif list for vserver $vserver_name on cluster $cluster_name");
                return {};
            }

            # Obtain a non faulted node object that is in the cluster that this vserver belongs to
            my ($nacl_node_obj) = $self->EventUtil::get_nacl_node_objs(-type => "cmode",
                                                                       -faulted => $EventUtil::FALSE,
                                                                       -clusters => [$cluster_name],
                                                                       -fall_back_on_nacl_clusters => $EventUtil::TRUE);
            if ($self->error()) {
                $self->_set_err("Failed to obtain any nonfaulted nacl nodes for cluster '$cluster_name'");
                return {};
            }

            $nacl_node_obj->refresh_command_interface();

            # Load all the lifs for the current vserver that are type fcp
            my @lif_objs;
            try {
                @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $nacl_node_obj,
                                                              filter => { vserver => $vserver_name,
                                                                          role => "data",
                                                                          'data-protocol' => ["fcp"]});
            } catch NACL::Exceptions::NoElementsFound with {
                # no fcp lifs were found on this vserver
                @lif_objs = ();
            } otherwise {
                my $e = shift;
                $self->_set_err("Exception caught while trying to fetch Lif objects: " . $e->text());
            };
            if ($self->error()) {
                return {};
            }

            # go to the next vserver if there are no fcp lifs
            next if (not scalar(@lif_objs));

            # Walk through the lif objects.
            foreach my $lif_obj (@lif_objs) {
                # Get the lif name
                my $lif_name = $lif_obj->lif();
                if (not defined($lif_name) or not length($lif_name)) {
                    $self->_set_err("Could not retrieve the lif name from the lif object");
                    return {};
                }

                if ($state eq "down") {
                    # The caller wants downed lifs
                    if (grep(/^${lif_name}$/, @downed_lifs)) {
                        # Get the wwpn for the lif
                        my $wwpn = $lif_obj->wwpn();
                        if (not defined($wwpn) or not length($wwpn)) {
                            $self->_set_err("Could not retrieve the wwpn for lif '$lif_name'");
                            return {};
                        }

                        # This lif is in the down list so add the wwpn to the list
                        push (@wwpns, $wwpn);
                    } else {
                        # This lif is not in the down list and the caller wants lifs in the down state
                        # so skip this lif.
                        next;
                    }
                } else {
                    # The caller wants lifs that are up
                    if (grep(/^${lif_name}$/, @downed_lifs)) {
                        # This lif is in the down list and the caller wants lifs in the up
                        # state so skip this lif.
                        next;
                    } else {
                        # Get the wwpn for the lif
                        my $wwpn = $lif_obj->wwpn();
                        if (not defined($wwpn) or not length($wwpn)) {
                            $self->_set_err("Could not retrieve the wwpn for lif '$lif_name'");
                            return {};
                        }

                        # This lif is not in the down list so add the wwpn to the list
                        push (@wwpns, $wwpn);
                    }
                }
            }
        } # end for each vserver obj
    } # end for each cluster

    # Grab the switch info hash from ResMgr
    my $switch_info = $self->ResMgr::get_resource(-name => "switch_info",
                                                  -type => "_GLOBALS");
    if ($self->error()) {
        $self->_set_err("Unable to get resource 'switch_info' from the Resource Manager");
        return {};
    }

    my $switch_info_hash = $switch_info->{obj};

    # We now need to walk the switch info hash to find the wwpn.  We can't assume what kind
    # of device it is so we will walk all non-apcon "fc" types under the sanswitches tree.
    # We need to make a hash by switch and port to validate against the steady state hash.
    # Format will be as follows:
    # $switch_port_hash->{<switch_name>}->{<port_number>}->[array of wwpns];
    my $switch_port_hash = {};
    foreach my $wwpn (@wwpns) {
        my $found = 0;
        foreach my $switch (keys %{$switch_info_hash->{sanswitches}}) {
            foreach my $resource_type (keys %{$switch_info_hash->{sanswitches}->{$switch}}) {
                unless ($resource_type =~ /^fc_/) {
                    # This isn't an fc resource so we won't find any wwpns
                    next;
                }

                unless (exists($switch_info_hash->{sanswitches}->{$switch}->{$resource_type}->{$wwpn})) {
                    # The wwpn we are looking for isn't found in this resource type
                    next;
                }

                my $port_numbers = $switch_info_hash->{sanswitches}->{$switch}->{$resource_type}->{$wwpn}->{port_number};
                foreach my $port_number (@{$port_numbers}) {
                    push(@{$switch_port_hash->{$switch}->{$port_number}}, $wwpn);
                }

                $found = 1;
                last;
            } ##END foreach my $resource_type

            last unless ($found == 0);
        } ##END foreach my $switch
    } ##END foreach my $wwpn (@wwpns) {

    return $switch_port_hash;
}

###############################################################################
# Method:    _validate_ports_by_wwwpn()
# Objective: Validate that the wwpn(s) exist or don't exist in the nameserver data
###############################################################################
sub _validate_ports_by_wwpn {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the port validation state to "up" which will validate and
    # assume all ports are in steady state.  Valid states are "down" and "up".
    # If state is "down", then a switch and port must also be passed in.
    my $state     = &ParamsMgr::get_arg(-arg => "state", -required => 1);
    my $switch    = &ParamsMgr::get_arg(-arg => "switch", -required => 1);
    my $port_hash = &ParamsMgr::get_arg(-arg => "port_hash", -required => 1, -ref_type => "HASH");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Make sure the state passed in is valid
    unless ($state eq "up" or $state eq "down") {
        $self->_set_err("Invalid value for arg '-state'.  Valid values are 'up' and 'down'");
        return $EventUtil::FALSE;
    }

    my @errors_found = ();

    # We can now process the port hash and get the current state
    my $current_state_hash = $self->PortValidation::_get_port_info(-switches => [$switch]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # log an event that we are validating paths for this client
    $self->SubtestComm::log_event_message(-event_message => "pv_validating_ports_on_${switch}");

    # Current State Hash Structure:
    #    $current_state_hash =>
    #        <switch> =>
    #             <port> =>
    #                 resource_type => Type of resource (fc_target_wwpn, etc)
    #                 wwpns         => hash of wwpns logged in
    #
    foreach $port_number (sort keys %{$port_hash}) {
        unless (exists ($current_state_hash->{$switch}->{$port_number})) {
            # We didn't find port in the current state hash for the switch
            # so this is an error.
            $self->_set_err("Unable to find port '$port_number' in the current state hash for switch '$switch'");
            return $EventUtil::FALSE;
        }

        # Walk the list of wwpns in the port hash and check their state against
        # the current state hash
        if ($state eq "up") {
            # make sure the wwpn IS included
            foreach my $wwpn (@{$port_hash->{$port_number}}) {
                # Lowercase the wwpn and add colons
                $wwpn = lc($wwpn);
                $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);

                unless (exists ($current_state_hash->{$switch}->{$port_number}->{wwpns}->{$wwpn})) {
                    # We didn't find the wwpn in the current state hash so it is missing.
                    my $errmsg = sprintf("%-20s %-8s %-25s %s", $switch, $port_number, $wwpn, "Missing");
                    push(@errors_found, $errmsg);
                }
            } ##END  foreach my $wwpn (@{$port_hash->{$port_number}}) {
        } else {
            # make sure the wwpn IS NOT included
            foreach my $wwpn (@{$port_hash->{$port_number}}) {
                # Lowercase the wwpn and add colons
                $wwpn = lc($wwpn);
                $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);

                if (exists ($current_state_hash->{$switch}->{$port_number}->{wwpns}->{$wwpn})) {
                    # We found the wwpn in the current state hash but it shouldn't be there
                    my $errmsg = sprintf("%-20s %-8s %-25s %s", $switch, $port_number, $wwpn, "Not Logged Out");
                    push(@errors_found, $errmsg);
                }
            } ##END  foreach my $wwpn (@{$port_hash->{$port_number}}) {
        } ##END if ($state eq "up") {
    } ##END foreach $port_number (sort keys %{$port_hash}) {

    # See if we found errors.  If we did, set error and bail.
    if (scalar(@errors_found)) {
        unshift(@errors_found, sprintf("%-20s %-8s %-25s %s", "====================",
                                                              "========",
                                                              "=========================",
                                                              "==============="));
        unshift(@errors_found, sprintf("%-20s %-8s %-25s %s", "Switch Name", "Port", "WWPN", "State"));
        unshift(@errors_found, "The following errors were found during port validation:\n");
        push (@errors_found, "\n");
        my $errmsg = join("\n", @errors_found);

        &SanUtil::nate_log("-comment", $errmsg);
        $self->_set_err("Port Validation failed for one or more nodes");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _validate_ports_by_switch()
# Objective: Validate the ports for the given switch.
###############################################################################
sub _validate_ports_by_switch {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the port validation state to "up" which will validate and
    # assume all ports are in steady state.  Valid states are "down" and "up".
    # If state is "down", then a switch and port must also be passed in.
    my $switch = &ParamsMgr::get_arg(-arg => "switch", -required => 1);
    my $state  = &ParamsMgr::get_arg(-arg => "state", -required => 0, -default => "up");

    my @ports  = &ParamsMgr::get_arg(-arg      => "ports",
                                     -required => 0,
                                     -ref_type => "ARRAY",
                                     -default  => []);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # Make sure that state is either up or down
    unless ($state eq "up" or $state eq "down") {
        $self->_set_err("Invalid state '$state' specified.  Valid values are 'up' and 'down'");
        return $EventUtil::FALSE;
    }

    # make sure that if the state is "down" that we were given a list of ports
    # to validate against since ports isn't a required parameter.
    if ($state eq "down") {
        unless (scalar(@ports) > 0) {
            $self->_set_err("You must specify a list of ports to validate (-ports) if validation " .
                            "state is 'down'");
            return $EventUtil::FALSE;
        }
    }

    # Grab the port info hash resource object if the steady state has been loaded
    # otherwise error out
    unless ($self->PortValidation::_is_steady_state_loaded()) {
        $self->_set_err("Port steady state information was not loaded into Resource Manager");
        return $EventUtil::FALSE;
    }

    my $resource_hash = $self->ResMgr::get_resource(-name => "port_info",
                                                    -type => "_GLOBALS");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $steady_state_hash = $resource_hash->{obj};

    # Build out the port hash for the current state of the switch
    my $current_state_hash = $self->PortValidation::_get_port_info(-switches => [$switch]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Port Info Hash Structure:
    #    $port_hash =>
    #        <switch> =>
    #             <port> =>
    #                 resource_type => Type of resource (fc_target_wwpn, etc)
    #                 wwpns         => hash of wwpns logged in

    # log an event that we are validating paths for this client
    $self->SubtestComm::log_event_message(-event_message => "pv_validating_ports_on_${switch}");

    # If the caller passed in a list of ports for this switch, then update the port
    # hash so we can only perform validations on those ports.
    my $port_hash = {};
    if (scalar(@ports) > 0) {
        foreach my $port (@ports) {
            $port_hash->{$port} = 1;
        }
    }

    my @errors_found = ();
    foreach my $port (sort keys %{$current_state_hash->{$switch}}) {
        if (scalar(keys %$port_hash) > 0 and not exists($port_hash->{$port})) {
            &SanUtil::nate_log("-comment", "Skipping port '$port' on switch '$switch' since it was not " .
                                           "included in the list of ports passed in by the caller");
            next;
        }

        # Grab the hash for the port itself to make comparisons easier to read
        # in the operations below
        unless (scalar(keys %{$steady_state_hash->{$switch}->{$port}})) {
            $self->_set_err("Unable to find steady state information for Node '$node' " .
                            "Vserver '$vserver' Port '$port'");
            return $EventUtil::FALSE;
        }

        my $port_current_state_hash = $current_state_hash->{$switch}->{$port};
        my $port_steady_state_hash  = $steady_state_hash->{$switch}->{$port};

        # Make sure all the wwpns logged back in if "state" is up
        if ($state eq "up") {
            foreach my $wwpn (sort keys %{$port_steady_state_hash->{wwpns}}) {
                # Check to make sure that the wwpns listed for the port in
                # the steady state hash DO exist in the current state hash
                unless (exists($port_current_state_hash->{wwpns}->{$wwpn})) {
                    my $errmsg = sprintf("%-20s %-8s %-25s %s", $switch, $port, $wwpn, "Missing");
                    push(@errors_found, $errmsg);
                }
            }
        } else {
            # State must be down
            foreach my $wwpn (sort keys %{$port_steady_state_hash->{wwpns}}) {
                # Check to make sure that the wwpns listed for the port in
                # the steady state hash DO NOT exist in the current state hash
                if (exists($port_current_state_hash->{wwpns}->{$wwpn})) {
                    my $errmsg = sprintf("%-20s %-8s %-25s %s", $switch, $port, $wwpn, "Not logged out");
                    push(@errors_found, $errmsg);
                }
            }
        }
    }

    # See if we found errors.  If we did, set error and bail.
    if (scalar(@errors_found)) {
        unshift(@errors_found, sprintf("%-20s %-8s %-25s %s", "====================",
                                                              "========",
                                                              "=========================",
                                                              "==============="));
        unshift(@errors_found, sprintf("%-20s %-8s %-25s %s", "Switch Name", "Port", "WWPN", "State"));
        unshift(@errors_found, "The following errors were found during port validation:\n");
        push (@errors_found, "\n");
        my $errmsg = join("\n", @errors_found);

        &SanUtil::nate_log("-comment", $errmsg);
        $self->_set_err("Port Validation failed for one or more nodes");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _wait_for_validate_ports()
# Objective: Given a previous timeout, wait the minimum wait time needed
#            which is the difference between the previous time out and the
#            min timeout
###############################################################################
sub _wait_for_validate_ports {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # the caller may have requested that we wait a minimum timeout
    # before performing port validation
    my $last_timeout = &ParamsMgr::get_arg(-arg => "last_timeout",
                                           -default => 0);
    my $last_timeout_name = &ParamsMgr::get_arg(-arg => "last_timeout_name",
                                                -default => "timeout");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # determine the min wait time from the current event/eventset
    my $pv_min_wait_time = $self->EventUtil::get_arg(-arg => "port_validation_min_wait_time",
                                                     -default => 0);

    my ($delta);

    # make sure we have slept AT LEAST pv_min_wait_time
    # seconds. If we have not, sleep the delta of pv_min_wait_time
    # and the passed in value.
    if ($last_timeout <= $pv_min_wait_time) {
        # calculate the additional amount of time we must sleep
        $delta = $pv_min_wait_time - $last_timeout;

        &SanUtil::nate_log('-comment',
                           "Sleeping an additional ($delta) seconds to nameserver to settle\n" .
                           "before performing port validation. This is in addition to the\n" .
                           "${last_timeout_name}=${last_timeout} seconds we already waited.");
        &SanUtil::nate_sleep($delta);
    } else {
        &SanUtil::nate_log('-comment',
                           "No additional sleep time is needed since we have already waited " .
                           "${last_timeout_name}=${last_timeout} seconds\nwhich is greater than " .
                           "-port_validation_min_wait_time ($pv_min_wait_time)");
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _is_steady_state_loaded()
# Objective: Determine if steady state is loaded on the specified SanClient
#            resource
###############################################################################
sub _is_steady_state_loaded {
    my ($self) = @_;

    my $steady_state_loaded = $self->ResMgr::get_resource_attribute(-name => "port_info",
                                                                    -type => "_GLOBALS",
                                                                    -attribute => "steady_state_loaded");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # check if steady state is set on this client
    if (defined($steady_state_loaded) and $steady_state_loaded == $EventUtil::TRUE) {
        return $steady_state_loaded;
    } else {
        return $EventUtil::FALSE;
    }
}

################################################################################
# Method Name: _get_downed_lifs
# Objective: get a list of downed lifs for a vserver resource
################################################################################
sub _get_downed_lifs {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $vserver = &ParamsMgr::get_arg(-arg => "vserver", -required => 1);
    my $cluster = &ParamsMgr::get_arg(-arg => "cluster", -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? (): [];
    }

    my $downed_lifs_hash = $self->ResMgr::get_resource_attribute(-name      => $vserver,
                                                                 -type      => "vservers",
                                                                 -parent_name => $cluster,
                                                                 -parent_type => 'nacl_clusters_cm',
                                                                 -attribute => "downed_lifs");
    if ($self->error()) {
        return (wantarray()) ? (): [];
    }

    unless (scalar(keys %$downed_lifs_hash) > 0) {
        # Looks like this resource is empty so just return
        return (wantarray()) ? (): [];
    }

    my @lifs = ();
    foreach my $lif (sort keys %$downed_lifs_hash) {
        push(@lifs, $lif);
    }

    return (wantarray()) ? @lifs: \@lifs;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    # define the default values for Event parameters
    $self->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                       -arg_defaults =>
                                                           { port_validation                            => 1,
                                                             port_validation_on_giveback                => 1,
                                                             port_validation_on_lif_down                => 1,
                                                             port_validation_on_lif_up                  => 1,
                                                             port_validation_on_port_up                 => 1,
                                                             port_validation_abort_on_error             => 1,
                                                             port_validation_min_wait_time              => 0,
                                                             port_validation_lif_name_regex_filter      => qr/.*/,
                                                             port_validation_resource_name_regex_filter => qr/.*/,
                                                           },
                                                       );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # determine if port validation is enabled in the current event/eventset
    my $port_validation = $self->EventUtil::get_arg(-arg => "port_validation");

    # If port_validation is not enabled, then don't bother continuing
    if ($port_validation == 0) {
        return $EventUtil::TRUE;
    }

    # Check to make sure we have switch objects in the environment.  If we
    # don't then bypass port validation.
    my (@sanswitches) = $self->EventUtil::get_sanswitch_objs(-faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain all SanSwitch resource objects");
        return $EventUtil::FALSE;
    }

    if (scalar(@sanswitches) == 0) {
        &SanUtil::nate_log('-comment', "Skipping initialization of port validation since there are no switches " .
                                       "in the environment");
        return $EventUtil::TRUE;
    }

    # build the switch info hash.  If already exists in ResMgr, then it won't
    # be rebuilt.
    $self->EventUtil::build_switch_info_hash();
    if ($self->error()) {
        $self->_set_err("Unable to build switch info hash with WWPN and MAC address data");
        return $EventUtil::FALSE;
    }

    # build the port hash (dont reconnect to switches)
    my $port_hash = $self->PortValidation::_get_port_info(-reconnect_switches => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to build the port hash");
        return $EventUtil::FALSE;
    }

    # register the port_hash datastructure as a "_GLOBALS" resource type
    # this will allow events to access this datastructure and make persistent modifications to it
    $self->ResMgr::add_resource(-name => "port_info",
                                -type => "_GLOBALS",
                                -obj  => $port_hash);
    if ($self->error()) {
        $self->_set_err("Unable to add resource 'port_info' to the Resource Manager");
        return $EventUtil::FALSE;
    }

    $self->ResMgr::set_resource_attribute(-name => "port_info",
                                          -type => "_GLOBALS",
                                          -attribute => "steady_state_loaded",
                                          -value => $EventUtil::TRUE,
                                          -propagate => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to set resource  attribute 'steady_state_loaded' on resource 'port_info' " .
                        "for type '_GLOBALS'");
        return $EventUtil::FALSE;
    }


    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    # document each parameter used in this module
    $self->EventUtil::document_event_param(-param => "port_validation",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-port_validation => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable port validation " .
                                               "on all fibre channel switch ports support the methods to do so.");

    $self->EventUtil::document_event_param(-param => "port_validation_on_giveback",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-port_validation_on_giveback => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable port validation after giveback",
                                           -overrides => ["port_validation"]);

    $self->EventUtil::document_event_param(-param => "port_validation_on_lif_down",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-port_validation_on_lif_down => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable port validation after lif down",
                                           -overrides => ["port_validation"]);

    $self->EventUtil::document_event_param(-param => "port_validation_on_lif_up",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-port_validation_on_lif_up => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable port validation after lif up",
                                           -overrides => ["port_validation"]);

    $self->EventUtil::document_event_param(-param => "port_validation_on_port_up",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-port_validation_on_port_up => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will enable port validation after port up",
                                           -overrides => ["port_validation"]);

    $self->EventUtil::document_event_param(-param => "port_validation_abort_on_error",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (dont abort) | 1 (abort)",
                                           -example => "-port_validation_abort_on_error => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "If set to 1, will abort events if a port validation error is " .
                                               "detected during intial setup or during validation",
                                           -resource_types => ["sanclients"]);

    $self->EventUtil::document_event_param(-param => "port_validation_min_wait_time",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "integer, in seconds",
                                           -example => "-port_validation_min_wait_time => 300",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the minimum amount of time (seconds) that must be waited " .
                                               "before performing port validation");

    $self->EventUtil::document_event_param(-param => "port_validation_lif_name_regex_filter",
                                           -package => __PACKAGE__,
                                           -default => "qr/.*/",
                                           -format => "regular expression",
                                           -example => "-port_validation_lif_name_regex_filter => qr/^(?!.*dummy).*$/",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies regex filter to determine what lif ports to validate");

    $self->EventUtil::document_event_param(-param => "port_validation_resource_name_regex_filter",
                                           -package => __PACKAGE__,
                                           -default => "qr/.*/",
                                           -format => "regular expression",
                                           -example => "-port_validation_resource_name_regex_filter => qr/^(?!.*das_dummy_vs).*$/",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies regex filter to determine what resource names to validate");


    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

PortValidation.pm

=head1 Author

Derrick Shultz

=head1 Description

An EventMan internal utility module designed to provide helper
methods to perform Port Validation for CMode SanFilers
during takeover / giveback, lif events, and port block scenarios

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-port_validation>

=over 4

Description: If set to 1, will enable port validation
on all fibre channel SanFilers, SanClients, and SanClients
that support the methods to do so

Resource Types: sanfilersCM, SanClients, SanSwitches

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-port_validation_on_lif_down>

=over 4

Description: If set to 1, will enable port validation during after
a lif down event on all SanFilers that support the methods to do so.

Resource Types: sanfilersCM

Default: 1 (on)

Per Resource Capable: No

=back

=head2 I<-port_validation_on_lif_up>

=over 4

Description: If set to 1, will enable port validation after lif
up events all SanFilers that support the methods to do so.

Resource Types: sanfilersCM

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-port_validation_on_port_up>

=over 4

Description: If set to 1, will enable port validation after a
port is brought up through either a port unblock occurs.
This will happen on all SanFilers, SanClients, and SanSwitches
that support the methods to do so.

Resource Types: sanfilersCM, sanclients, sanswitches

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-port_validation_on_giveback>

=over 4

Description: If set to 1, will enable port validation after a
giveback ocurrs.  This will happen on all SanFilers, SanSwitches
that support the methods to do so.

Resource Types: sanfilersCm, sanfilers7m, sanswitches

Per Resource Capable: No

Default: 1 (on)


=head2 I<-port_validation_abort_on_error>

=over 4

Description: If set to 1, will abort events if a port validation error is
detected during intial setup or during validation.

Resource Types: sanfilersCM, sanclients, sanswitches

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-port_validation_min_wait_time>

=over 4

Description: Specifies the minimum amount of time (seconds) that
must be waited before performing port validation

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-port_validation_lif_name_regex_filter>

=over 4

Description: Specifies the regex filter for the list of lifs to validate.

Per Resource Capable: No

Default: qr/.*/

Regex Examples:

    Do not validate lifs containing the word "dummy"

        qr/^(?!.*dummy).*$/

    Validate lifs starting with "fc" or "iscsi"

        qr/^(fc|iscsi).*$/

=back

=head2 I<-port_validation_resource_name_regex_filter>

=over 4

Description: Specifies the regex filter for the list of resource names to validate.

Per Resource Capable: No

Default: qr/.*/

Regex Examples:

    Do not validate resource names containing the word "windows"

        qr/^(?!.*windows).*$/

    Validate the verservers whose names start with 'das_' and end in '_vs'

        qr/^das_.*_vs$/

=back

=head1 B<Public Methods>

=head2 B<validate_ports_after_lif_down()>

Performs port validation for all SanFilerCM fibre channel LIF resources
in the current Event/EventSet/EventMan resource scope assuming LIFs are
in a down state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off port validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do port validation. This
will be used to calculate additional port validation wait time that needs
to be applied before performing port validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: "timeout"

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -port_validation
 -port_validation_on_lif_down
 -port_validation_abort_on_error
 -port_validation_min_wait_time

=item Examples

 Port Validation following a lif down:

 $sanfaults_obj->PortValidation::validate_ports_after_lif_down(-id_extension => "lif_down_node_A",
                                                               -last_timeout => "30",
                                                               -last_timeout_name => "lif_up_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<validate_ports_after_lif_up()>

Performs port validation for all SanFilerCM fibre channel LIF resources
in the current Event/EventSet/EventMan resource scope assuming LIFs are
in a up state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off port validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do port validation. This
will be used to calculate additional port validation wait time that needs
to be applied before performing port validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: "timeout"

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -port_validation
 -port_validation_on_lif_up
 -port_validation_abort_on_error
 -port_validation_min_wait_time

=item Examples

 Port Validation following a lif up:

 $sanfaults_obj->PortValidation::validate_ports_after_lif_up(-id_extension => "lif_up_node_A",
                                                             -last_timeout => "30",
                                                             -last_timeout_name => "lif_up_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<validate_ports_after_port_up()>

Performs port validation for all SanFilerCM and SanClient resources in
the current Event/EventSet/EventMan resource scope. Controllers, Switches,
and Clients are assumed to be in the steady state.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off port validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do port validation. This
will be used to calculate additional port validation wait time that needs
to be applied before performing port validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: ""

=item *I<-switch>

Optional: switch to perform validation against
Default: Empty string

=item *I<-ports>

Optional: Array of ports to perform validation against.  If
ports list is specified, then the switch must also be specified.
Default: Empty array reference

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -port_validation
 -port_validation_on_port_up
 -port_validation_abort_on_error
 -port_validation_min_wait_time

=item Examples

 $sanfaults_obj->PortValidation::validate_ports_after_port_up(-id_extension => "switch_port_${port_number}_block_on_${switch_name}",
                                                              -last_timeout => "30",
                                                              -last_timeout_name => "switch_port_up_time",
                                                              -switch => $switch_name,
                                                              -ports => [$port_number]);
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<validate_ports_after_giveback()>

Performs port validation for all sanfilerCm and sanfiler7m resources in
the current Event/EventSet/EventMan resource scope following a giveback.

=over 4

=item Inputs

=item *I<-id_extension>

Required: The runid prefix to use when subtesting off port validation subtests

=item *I<-last_timeout>

Optional: The last timeout value used before waiting to do port validation. This
will be used to calculate additional port validation wait time that needs
to be applied before performing port validation.
Default: 0

=item *I<-last_timeout_name>

Optional: the name of the timeout that is specified in -last_timeout
Default: ""

=item *I<-switch>

Optional: switch to perform validation against
Default: Empty string

=item *I<-ports>

Optional: Array of ports to perform validation against.  If
ports list is specified, then the switch must also be specified.
Default: Empty array reference

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -port_validation
 -port_validation_on_giveback
 -port_validation_abort_on_error
 -port_validation_min_wait_time

=item Example

 $sanfaults_obj->PortValidation::validate_ports_on_giveback(-id_extension => "giveback_on_node_A",
                                                            -last_timeout => "600",
                                                            -last_timeout_name => "filer_down_time");
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=cut
