# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing Resource Management APIs
## @author hrivera@netapp.com
## @status private
## @pod here

package ResMgr;

use warnings;
use strict;

use ParamsMgr;
use Data::Dumper;
use Time::HiRes qw(gettimeofday);
use Tie::IxHash;
use Storable;
use NACL::STask::Cluster;
use NACL::C::Vserver;
use NATE::BaseException qw (:try);
use NACL::Exceptions::NoElementsFound;
use NATE::Log qw(log_global);
use San::SanClient;
use San::SanSwitch;
use Data::Dump;
use San::SanUtil;

use vars qw($Log);

$Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    add_resource()
# Objective: Add a specific resource into the ResMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_resource {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_obj = &ParamsMgr::get_arg(-arg => "obj",
                                           -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    delete($args{-name});
    delete($args{-obj});
    delete($args{-type});

    # verify that there is not already a resource registered by that name and type
    $self->ResMgr::get_resource(-type => $resource_type,
                                -name => $resource_name);
    if (not $self->error()) {
        # then a resource already exists by this name/type,
        $self->_set_err("A resource already exists with the name '$resource_name' and type '$resource_type'");
        $Log->exit();
        return $EventUtil::FALSE;
    } else {
        # clear the error
        $self->_clear_err();
    }

    # validate that the blessed objects match the type
    # that was requested
    if ($resource_type =~ /^sanclients$/) {
        if (ref($resource_obj) !~ /^San::SanClient/) {
            $self->_set_err("-obj is an invalid SanClient object");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^sanfilers7m$/) {
        if (ref($resource_obj) !~ /^San::SanFiler7M$/) {
            $self->_set_err("-obj is an invalid SanFiler7M object");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^sanfilersCm$/) {
        if (ref($resource_obj) !~ /^San::ControllerCluster::Filer10M$/) {
            $self->_set_err("-obj is an invalid Filer10M object");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^controllerclusters$/) {
        if (ref($resource_obj) =~ /^San::ControllerCluster$/) {
            # initialize this Controller Cluster object by populating resources for its
            # individual node objects

            # get all the sanfiler nodes in this cluster
            my @sanfiler_node_objs = $resource_obj->get_node_objs();
            if ($resource_obj->error()) {
                $self->_set_err($resource_obj->errmsg());
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # delete the -vservers param if it exists so that we dont add it as an event param
            # to the node resource
            my %node_args = %args;
            delete($node_args{-vservers});
            delete($node_args{vservers});

            # add each node as a resource to the ResMgr
            foreach my $node_obj (@sanfiler_node_objs) {
                my $node_name = $node_obj->get_name();
                $self->ResMgr::add_resource(-name => $node_name,
                                            -obj => $node_obj,
                                            -type => "sanfilersCm",
                                            %node_args);
                if ($self->error()) {
                    $Log->exit();
                    return $EventUtil::FALSE;
                }

                # store the controller cluster resource name with this sanfilerCm resource
                # to allow reverse lookups.
                $self->ResMgr::set_resource_attribute(-name => $node_name,
                                                      -type => "sanfilersCm",
                                                      -attribute => "controllercluster_name",
                                                      -value => $resource_name,
                                                      -propagate => $EventUtil::FALSE);
                if ($self->error()) {
                    $Log->exit();
                    $self->_set_err("Unable to set resource attribute 'controllercluster_name' for " .
                                    "resource ($node_name : sanfilersCm)");
                    return $EventUtil::FALSE;
                }
            }

            # if a nacl_clusters_cm resource for $resource_name doesnt exist, then discover it now.
            # add a NACL::STask::Cluster object that represents this cluster (use the first node)
            $self->ResMgr::get_resource(-name => $resource_name,
                                        -type => "nacl_clusters_cm");
            if ($self->error()) {
                # add this nacl_clusters_cm resource to eventman
                $self->_clear_err();
                my $nacl_cluster_obj;
                try {
                    ($nacl_cluster_obj) = NACL::STask::Cluster->new(node => $sanfiler_node_objs[0]->get_name());
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught while creating NACL::STask::Cluster - " . $e->text());
                };
                if ($self->error()) {
                    $Log->exit();
                    return $EventUtil::FALSE;
                }

                $self->ResMgr::add_resource(-name => $resource_name,
                                            -obj => $nacl_cluster_obj,
                                            -type => "nacl_clusters_cm",
                                            %args);
                if ($self->error()) {
                    $self->_set_err("Failed to add NACL::STask::Cluster object resource for $resource_name");
                    $Log->exit();
                    return $EventUtil::FALSE;
                }
            } else {
                # there already is a nacl_clusters_cm resource registered, dont
                # add it again
            }
        } else {
            $self->_set_err("-obj is an invalid ControllerCluster object");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^sanswitches$/) {
        if (ref($resource_obj) !~ /^San::SanSwitch/) {
            $self->_set_err("-obj is an invalid SanSwitch object");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^nacl_nodes_7m$/) {
        unless (ref($resource_obj) and $resource_obj->isa("NACL::C::Node")) {
            $self->_set_err("-obj is an invalid NACL::C::Node object (" . ref($resource_obj) . ")");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^nacl_nodes_cm$/) {
        unless (ref($resource_obj) and $resource_obj->isa("NACL::C::Node")) {
            $self->_set_err("-obj is an invalid NACL::C::Node object (" . ref($resource_obj) . ")");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } elsif ($resource_type =~ /^vservers$/) {
        $self->_set_err("Cannot add NACL::C::Vserver objects as resources, invoke add_resource() ".
                        "with a NACL::STask::Cluster object instead.");
        $Log->exit();
        return $EventUtil::FALSE;
    } elsif ($resource_type =~ /^nacl_clusters_cm$/) {
        unless (ref($resource_obj) and $resource_obj->isa("NACL::STask::Cluster")) {
            $self->_set_err("-obj is an invalid NACL::STask::Cluster object (" . ref($resource_obj) . ")");
            $Log->exit();
            return $EventUtil::FALSE;
        }

        # discover the individual nodes and register them as resources
        my @node_objs;
        try {
            @node_objs = $resource_obj->node_objs();
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            $Log->exit();
            return $EventUtil::FALSE;
        }

        my %node_args = %args;
        delete($node_args{-vservers});
        delete($node_args{vservers});

        # add each node as a resource to the ResMgr
        foreach my $node_obj (@node_objs) {
            my $node_name = $node_obj->node();
            $self->ResMgr::add_resource(-name => $node_name,
                                        -obj => $node_obj,
                                        -type => "nacl_nodes_cm",
                                        %node_args);
            if ($self->error()) {
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # store the controller cluster resource name with this nacl node resource
            # to help reverse lookups.
            $self->ResMgr::set_resource_attribute(-name => $node_name,
                                                  -type => "nacl_nodes_cm",
                                                  -attribute => "controllercluster_name",
                                                  -value => $resource_name,
                                                  -propagate => $EventUtil::FALSE);
            if ($self->error()) {
                $self->_set_err("Unable to set resource attribute 'controllercluster_name' for " .
                                "resource ($node_name : nacl_nodes_cm)");
                $Log->exit();
                return $EventUtil::FALSE;
            }
        }

        # if a vservers filter was passed in, scope the command to this filter
        my @vs_filters = ();
        if (defined($args{-vservers}) or defined($args{vservers})) {
            my $global_vserver_list;
            $global_vserver_list = $args{-vservers} if (defined($args{-vservers}));
            $global_vserver_list = $args{vservers} if (defined($args{vservers}));

            # break up the vserver scope list into multiple arrays
            my @vserver_scope_lists = &SanUtil::break_up_array(array => $global_vserver_list,
                                                               size => 10);
            foreach my $vserver_scope_list (@vserver_scope_lists) {
                if (ref($vserver_scope_list) eq "ARRAY" and scalar(@{$vserver_scope_list})) {
                    my %vs_filter = (filter => { vserver => join(",", @{$vserver_scope_list}) });
                    push(@vs_filters, \%vs_filter);
                }
            }
        }

        # find all vserver components and add them as resources
        my @vserver_objs = ();
        if (scalar(@vs_filters)) {
            foreach my $vs_filter (@vs_filters) {
                try {
                    # use the regular find method if there are vserver filters passed in
                    push(@vserver_objs, NACL::C::Vserver->find(command_interface => $resource_obj,
                                                               %{$vs_filter}));
                } catch NACL::Exceptions::NoElementsFound with {
                    # do nothing if there are no elements found
                } otherwise {
                    my $e = shift;
                    $self->_set_err("NACL Exception caught: " . $e->text());
                };
            }
        } else {
            try {
                # use the find_data_vservers method if there are no vserver filters passed in
                @vserver_objs = NACL::C::Vserver->find_data_vservers(command_interface => $resource_obj);
            } catch NACL::Exceptions::NoElementsFound with {
                @vserver_objs = ();
            } otherwise {
                my $e = shift;
                $self->_set_err("NACL Exception caught: " . $e->text());
            };
        }
        if ($self->error()) {
            $Log->exit();
            return $EventUtil::FALSE;
        }

        my %vserver_args = %args;
        delete($vserver_args{-vservers});
        delete($vserver_args{vservers});

        foreach my $vserver_obj (@vserver_objs) {
            # add this vserver as a sub resource to the cluster resource
            my $vserver_name = $vserver_obj->vserver();
            $self->ResMgr::add_sub_resource(-name => $vserver_name,
                                            -obj => $vserver_obj,
                                            -type => "vservers",
                                            -parent_type => "nacl_clusters_cm",
                                            -parent_name => $resource_name,
                                            %vserver_args);
            if ($self->error()) {
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # store the controller cluster resource name with this vserver resource
            # to allow reverse lookups.
            $self->ResMgr::set_resource_attribute(-name => $vserver_name,
                                                  -type => "vservers",
                                                  -attribute => "controllercluster_name",
                                                  -value => $resource_name,
                                                  -parent_name => $resource_name,
                                                  -parent_type => 'nacl_clusters_cm',
                                                  -propagate => $EventUtil::FALSE);
            if ($self->error()) {
                $self->_set_err("Unable to set resource attribute 'controllercluster_name' for " .
                                "resource ($vserver_name : vservers)");
                $Log->exit();
                return $EventUtil::FALSE;
            }
        }
    }

    # retain hash insertion order for this resource type if this is the first
    # resource of this type
    if (not exists($self->{_resources}->{$resource_type})) {
        tie(%{$self->{_resources}->{$resource_type}}, "Tie::IxHash");
    }

    # populate the resource entry
    $self->{_resources}->{$resource_type}->{$resource_name}->{obj} = $resource_obj;
    $self->{_resources}->{$resource_type}->{$resource_name}->{args} = {&SanUtil::strip_dash(%args)};

    # mark this device as not faulted
    $self->ResMgr::unset_faulted_resource(-name => $resource_name,
                                          -type => $resource_type,
                                          -propagate => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to mark resource '$resource_name' as not faulted");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    add_sub_resource()
# Objective: Add a specific sub resource into the ResMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add_sub_resource {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_obj = &ParamsMgr::get_arg(-arg => "obj",
                                           -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);
    my $parent_resource_type = &ParamsMgr::get_arg(-arg => "parent_type",
                                                   -required => 1);
    my $parent_resource_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                                   -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    delete($args{-name});
    delete($args{-obj});
    delete($args{-type});
    delete($args{-parent_type});
    delete($args{-parent_name});

    # verify that there is not already a resource registered by that name and type
    $self->ResMgr::get_resource(-type => $resource_type,
                                -name => $resource_name,
                                -parent_type => $parent_resource_type,
                                -parent_name => $parent_resource_name);
    if (not $self->error()) {
        # then a resource already exists by this name/type,
        $self->_set_err("A resource already exists with the name '$resource_name' and type '$resource_type'");
        $Log->exit();
        return $EventUtil::FALSE;
    } else {
        # clear the error
        $self->_clear_err();
    }

    if ($resource_type =~ /^vservers$/) {
        unless (ref($resource_obj) and $resource_obj->isa("NACL::C::Vserver")) {
            $self->_set_err("-obj is an invalid NACL::C::Vserver object (" . ref($resource_obj) . ")");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    }

    # retain hash insertion order for this resource type if this is the first
    # resource of this type
    if (not exists($self->{_resources}->{$parent_resource_type}->{$parent_resource_name}->{sub_resources}->{$resource_type})) {
        tie(%{$self->{_resources}->{$parent_resource_type}->{$parent_resource_name}->{sub_resources}->{$resource_type}}, "Tie::IxHash");
    }

    # populate the resource entry as a sub resource of the parent resource
    $self->{_resources}->{$parent_resource_type}->{$parent_resource_name}->{sub_resources}->{$resource_type}->{$resource_name}->{obj}  = $resource_obj;
    $self->{_resources}->{$parent_resource_type}->{$parent_resource_name}->{sub_resources}->{$resource_type}->{$resource_name}->{args} = {&SanUtil::strip_dash(%args)};

    # mark this device as not faulted
    $self->ResMgr::unset_faulted_resource(-name => $resource_name,
                                          -type => $resource_type,
                                          -parent_type => $parent_resource_type,
                                          -parent_name => $parent_resource_name,
                                          -propagate => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to mark resource '$resource_name' as not faulted");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    remove_resource()
# Objective: Remove a specific resource from the ResMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove_resource {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);
    my $parent_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                          -default => undef);
    my $parent_type = &ParamsMgr::get_arg(-arg => "parent_type",
                                          -default => undef);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    delete($args{-name});
    delete($args{-type});
    delete($args{-parent_name});
    delete($args{-parent_type});

    # remove sub resources if specified
    if ($parent_type and $parent_name) {
        if (exists($self->{_resources}->{$parent_type}) and
            exists($self->{_resources}->{$parent_type}->{$parent_name}) and
            exists($self->{_resources}->{$parent_type}->{$parent_name}->{sub_resources}) and
            exists($self->{_resources}->{$parent_type}->{$parent_name}->{sub_resources}->{$resource_type})) {
            # delete the sub resource
            delete($self->{_resources}->{$parent_type}->{$parent_name}->{sub_resources}->{$resource_type}->{$resource_name});

            # delete the resource type if there is none left
            if (not scalar(keys %{$self->{_resources}->{$parent_type}->{$parent_name}->{sub_resources}->{$resource_type}})) {
                delete($self->{_resources}->{$parent_type}->{$parent_name}->{sub_resources}->{$resource_type});
                delete($self->{_resources}->{$parent_type}->{$parent_name}->{sub_resources});
            }
        } else {
            $self->_set_err("Could not find sub resource '$resource_name' of type '$resource_type' ($parent_type:$parent_name)");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    } else {
        # remove regular resources if otherwise
        if (exists($self->{_resources}->{$resource_type})) {
            if (exists($self->{_resources}->{$resource_type}->{$resource_name})) {
                # delete this resource from the resources hash
                $self->{_resources}->{$resource_type}->{$resource_name} = {};
                delete($self->{_resources}->{$resource_type}->{$resource_name});

                # if this is a controller cluster resource type, then we need to remove all
                # sanfilersCm/nacl_nodes_cm resources that belong to this cluster
                my @removal_list;
                if ($resource_type eq "controllerclusters") {
                    @removal_list = ("sanfilersCm");
                } elsif ($resource_type eq "nacl_clusters_cm") {
                    @removal_list = ("nacl_nodes_cm");
                } else {
                    @removal_list = ();
                }

                foreach my $type (@removal_list) {
                    my @resources = $self->ResMgr::get_resource_names_by_type(-type => $type);
                    if ($self->error()) {
                        $self->_set_err("Unable to get resources of type ($type)");
                        $Log->exit();
                        return $EventUtil::FALSE;
                    }
                    next if (not scalar(@resources));
                    my @resources_to_remove = $self->ResMgr::filter_by_attribute(-types => [$type],
                                                                                 -names => \@resources,
                                                                                 -attribute => "controllercluster_name",
                                                                                 -value => $resource_name);
                    if ($self->error()) {
                        $self->_set_err("Failed to filter resources by attribute 'controllercluster_name'" .
                                            "='$resource_name', for resources of type ($type)");
                        $Log->exit();
                        return $EventUtil::FALSE;
                    }
                    foreach my $name (@resources_to_remove) {
                        $self->ResMgr::remove_resource(-name => $name,
                                                       -type => $type);
                        if ($self->error()) {
                            $self->_set_err("Unable to remove resource ($name : $type)");
                            $Log->exit();
                            return $EventUtil::FALSE;
                        }
                    }
                }
            } else {
                $self->_set_err("Could not find resource '$resource_name' of type '$resource_type'");
                $Log->exit();
                return $EventUtil::FALSE;
            }
        } else {
            $self->_set_err("Could not find resource type '$resource_type'");
            $Log->exit();
            return $EventUtil::FALSE;
        }
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    refresh_resources()
# Objective: Go through all resource types and refresh them
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh_resources {
    $Log->enter();
    my ($self) = @_;

    # get all non resource types
    my @types = $self->ResMgr::get_all_resource_types();
    if ($self->error()) {
        $self->_set_err("Failed to get all resource types");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # now remove and re add controller cluster resources
    if (grep(/controllerclusters/, @types)) {
        # remove and add the controller cluster resources back into the ResMgr
        my @clusters = $self->ResMgr::get_resource_names_by_type(-type => "controllerclusters");
        if ($self->error()) {
            $self->_set_err("Failed to get all controllercluster resources");
            $Log->exit();
            return $EventUtil::FALSE;
        }

        foreach my $cluster (@clusters) {
            # extract all the data for this cluster
            my $res = $self->ResMgr::get_resource(-name => $cluster,
                                                  -type => "controllerclusters");
            if ($self->error()) {
                $self->_set_err("Failed to obtain controllercluster resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }
            my $clus_obj = $res->{obj};
            my %args = %{$res->{args}};
            my %attrs = %{$res->{attrs}};

            # remove the resource
            $self->ResMgr::remove_resource(-name => $cluster,
                                           -type => "controllerclusters");
            if ($self->error()) {
                $self->_set_err("Failed to remove controllercluster resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # add the resource back with all its original data
            $self->ResMgr::add_resource(-name => $cluster,
                                        -type => "controllerclusters",
                                        -obj => $clus_obj,
                                        %args);
            if ($self->error()) {
                $self->_set_err("Failed to add controllercluster resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }

            $res = $self->ResMgr::get_resource(-name => $cluster,
                                               -type => "controllerclusters");
            if ($self->error()) {
                $self->_set_err("Failed to obtain controllercluster resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # reassign attributes that may have been erased
            $res->{attrs} = \%attrs if (scalar(keys %attrs));
        }
    }

    # now remove and re add controller cluster resources
    if (grep(/nacl_clusters_cm/, @types)) {
        # remove and add the controller cluster resources back into the ResMgr
        my @clusters = $self->ResMgr::get_resource_names_by_type(-type => "nacl_clusters_cm");
        if ($self->error()) {
            $self->_set_err("Failed to get all controllercluster resources");
            $Log->exit();
            return $EventUtil::FALSE;
        }

        foreach my $cluster (@clusters) {
            # extract all the data for this cluster
            my $res = $self->ResMgr::get_resource(-name => $cluster,
                                                  -type => "nacl_clusters_cm");
            if ($self->error()) {
                $self->_set_err("Failed to obtain nacl_clusters_cm resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }
            my $clus_obj = $res->{obj};
            my %args = %{$res->{args}};
            my %attrs = %{$res->{attrs}};
            # store all the sub resources for restoration later on
            my $sub_resources = $res->{sub_resources};

            # remove the resource
            $self->ResMgr::remove_resource(-name => $cluster,
                                           -type => "nacl_clusters_cm");
            if ($self->error()) {
                $self->_set_err("Failed to remove nacl_clusters_cm resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # add the resource back with all its original data
            $self->ResMgr::add_resource(-name => $cluster,
                                        -type => "nacl_clusters_cm",
                                        -obj => $clus_obj,
                                        %args);
            if ($self->error()) {
                $self->_set_err("Failed to add nacl_clusters_cm resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }

            $res = $self->ResMgr::get_resource(-name => $cluster,
                                               -type => "nacl_clusters_cm");
            if ($self->error()) {
                $self->_set_err("Failed to obtain nacl_clusters_cm resource $cluster");
                $Log->exit();
                return $EventUtil::FALSE;
            }

            # reassign attributes that may have been erased
            $res->{attrs} = \%attrs if (scalar(keys %attrs));

            # reassign all the attributes and args for each sub resource type
            if ($sub_resources and ref($sub_resources) eq 'HASH') {
                foreach my $sub_resource_type (keys %{$sub_resources}) {
                    foreach my $sub_resource_name (keys %{$sub_resources->{$sub_resource_type}}) {
                        my $sub_res = $self->ResMgr::get_resource(-name => $sub_resource_name,
                                                                  -type => $sub_resource_type,
                                                                  -parent_name => $cluster,
                                                                  -parent_type => "nacl_clusters_cm");
                        if ($self->error()) {
                            $self->_clear_err();
                            next;
                        }
                        # restore the attr fields and merge the args fields on the sub resource
                        $sub_res->{attrs} = $sub_resources->{$sub_resource_type}->{$sub_resource_name}->{attrs};
                        foreach my $arg (keys %{$sub_resources->{$sub_resource_type}->{$sub_resource_name}->{args}}) {
                            $sub_res->{args}->{$arg} = $sub_resources->{$sub_resource_type}->{$sub_resource_name}->{args}->{$arg};
                        }
                    }
                }
            }
        }
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_resource_names_by_type()
# Objective: Get all resources of a specific type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource_names_by_type {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    my @resource_names = ();

    # find the for the resouce type directly
    # return an empty array if this resource type does not exist
    if (not exists($self->{_resources}->{$resource_type})) {
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    @resource_names = keys %{$self->{_resources}->{$resource_type}};

    $Log->exit();
    return (wantarray()) ? @resource_names : \@resource_names;
}

###############################################################################
# Method:    get_sub_resource_names_by_type()
# Objective: Get all sub resources of a specific type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sub_resource_names_by_type {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);
    my $parent_type = &ParamsMgr::get_arg(-arg => "parent_type",
                                          -default => undef);
    my $parent_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                          -default => undef);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    my @resource_names = ();

    # if a parent type was specified, scan through all the resources of that type
    if ($parent_type) {
        # return an empty array if this resource type does not exist
        if (not exists($self->{_resources}->{$parent_type})) {
            $Log->exit();
            return (wantarray()) ? () : [];
        }

        foreach my $pname (keys %{$self->{_resources}->{$parent_type}}) {
            # skip this resource if it doesn't match the parent name that may have
            # been specified
            next if ($parent_name and $pname ne $parent_name);

            # extract the sub resources
            if (exists($self->{_resources}->{$parent_type}->{$pname}->{sub_resources}) and
                exists($self->{_resources}->{$parent_type}->{$pname}->{sub_resources}->{$resource_type})) {
                push(@resource_names, keys %{$self->{_resources}->{$parent_type}->{$pname}->{sub_resources}->{$resource_type}});
            }
        }
    } else {
        # find all subresource names of this type
        foreach my $rtype (keys %{$self->{_resources}}) {
            foreach my $rname (keys %{$self->{_resources}->{$rtype}}) {
                next if (!$self->{_resources}->{$rtype}->{$rname}->{sub_resources});
                foreach my $sub_rtype (keys %{$self->{_resources}->{$rtype}->{$rname}->{sub_resources}}) {
                    next if ($sub_rtype ne $resource_type);
                    push(@resource_names, keys %{$self->{_resources}->{$rtype}->{$rname}->{sub_resources}->{$sub_rtype}});
                }
            }
        }
    }

    $Log->exit();
    return (wantarray()) ? @resource_names : \@resource_names;
}

###############################################################################
# Method:    get_all_resource_types()
# Objective: Get all available resource types
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_all_resource_types {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $exclude_private = &ParamsMgr::get_arg(-arg => "exclude_private",
                                              -default => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # return back all the resource types that are currently registered in
    # in the resource manager
    my @resource_types = keys %{$self->{_resources}};

    # exclude private resource types (those that start with an underscore char)
    # if requested to do so
    if ($exclude_private == $EventUtil::TRUE) {
        @resource_types = grep { !/^_/ } @resource_types;
    }

    $Log->exit();
    return (wantarray()) ? @resource_types : \@resource_types;
}

###############################################################################
# Method:    get_all_sub_resource_types()
# Objective: Get all available sub resource types
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_all_sub_resource_types {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $exclude_private = &ParamsMgr::get_arg(-arg => "exclude_private",
                                              -default => $EventUtil::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # return back all the resource types that are currently registered in
    # in the resource manager
    my @resource_types = keys %{$self->{_resources}};

    # exclude private resource types (those that start with an underscore char)
    # if requested to do so
    if ($exclude_private == $EventUtil::TRUE) {
        @resource_types = grep { !/^_/ } @resource_types;
    }

    # find any sub resource types as well
    my %sub_resource_types;
    foreach my $type (@resource_types) {
        my @resource_names = $self->ResMgr::get_resource_names_by_type(-type => $type);
        foreach my $name (@resource_names) {
            my $res = $self->ResMgr::get_resource(-type => $type,
                                                  -name => $name);
            if (exists $res->{sub_resources}) {
                foreach my $sub_type (keys %{$res->{sub_resources}}) {
                    $sub_resource_types{$sub_type}++;
                }
            }
        }
    }

    my @sub_resource_types = ();
    # add sub resource types to the list if there are any
    if (scalar(keys %sub_resource_types)) {
        push(@sub_resource_types, keys %sub_resource_types);
    }

    $Log->exit();
    return (wantarray()) ? @sub_resource_types : \@sub_resource_types;
}

###############################################################################
# Method:    get_current_resources()
# Objective: Obtain all resources that are in use in the current environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_resources {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the object event params to nothing
    my (@objects) = &ParamsMgr::get_arg(-arg => "objects",
                                        -ref_type => "ARRAY",
                                        -default => []);
    my $exclude_private = &ParamsMgr::get_arg(-arg => "exclude_private",
                                              -default => $EventUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # return back all the resources that are being used in the current event/eventset
    my @resource_types = $self->ResMgr::get_all_resource_types(-exclude_private => $exclude_private);

    my $resource_hash = {};

    # we need to pass in -objects into the get arg method but we
    # dont want it to be non-null, so only pass in the argument
    # if it contains elements
    my %passthru_args = ();
    if (scalar(@objects)) {
        $passthru_args{-objects} = \@objects;
    }

    # obtain the currently used resources and store them in a hash where
    # the key is the resource name
    foreach my $type (@resource_types) {
        my $resource_list = $self->EventUtil::get_arg(-arg => $type,
                                                      %passthru_args);

        # if no resources were found then default the list to every resource by that type
        if (not defined($resource_list)) {
            $resource_list = $self->ResMgr::get_resource_names_by_type(-type => $type);
            if ($self->error()) {
                $self->_set_err("Could not obtain resource names by type '$type'");
                $Log->exit();
                return {};
            }
        }

        # store this list with its respective type
        $resource_hash->{$type} = $resource_list;
    }
    $Log->exit();
    return $resource_hash;
}

###############################################################################
# Method:    get_current_sub_resources()
# Objective: Obtain all sub_resources that are in use in the current environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_sub_resources {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # default the object event params to nothing
    my (@objects) = &ParamsMgr::get_arg(-arg => "objects",
                                        -ref_type => "ARRAY",
                                        -default => []);
    my $exclude_private = &ParamsMgr::get_arg(-arg => "exclude_private",
                                              -default => $EventUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # return back all the resources that are being used in the current event/eventset
    my @sub_resource_types = $self->ResMgr::get_all_sub_resource_types(-exclude_private => $exclude_private);

    my $sub_resource_hash = {};

    # we need to pass in -objects into the get arg method but we
    # dont want it to be non-null, so only pass in the argument
    # if it contains elements
    my %passthru_args = ();
    if (scalar(@objects)) {
        $passthru_args{-objects} = \@objects;
    }

    # obtain the currently used resources and store them in a hash where
    # the key is the resource name
    foreach my $type (@sub_resource_types) {
        my $sub_resource_list = $self->EventUtil::get_arg(-arg => $type,
                                                          %passthru_args);

        # if no sub resources were found then default the list to every sub resource by that type
        if (not defined($sub_resource_list)) {
            $sub_resource_list = $self->ResMgr::get_sub_resource_names_by_type(-type => $type);
            if ($self->error()) {
                $self->_set_err("Could not obtain sub resource names by type '$type'");
                $Log->exit();
                return {};
            }
        }

        # store this list with its respective subtype
        $sub_resource_hash->{$type} = $sub_resource_list;
    }

    $Log->exit("Sub resources " . Dumper($sub_resource_hash));
    return $sub_resource_hash;
}

###############################################################################
# Method:    get_current_resources_by_type()
# Objective: Obtain all resources by the specified type that are in use in the
#            current environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_current_resources_by_type {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    my $resource_list = $self->EventUtil::get_arg(-arg => $resource_type);

    # if no resources were found then default the list to every resource by that type
    if (not defined($resource_list)) {
        my $curr_resource_hsh = $self->ResMgr::get_current_resources(%args);
        if ($self->error()) {
            $self->_set_err("Could not obtain resource names by type '$resource_type'");
            $Log->exit();
            return (wantarray()) ? () : [];
        }

        my $curr_sub_resource_hsh = $self->ResMgr::get_current_sub_resources(%args);
        if ($self->error()) {
            $self->_set_err("Could not obtain sub resource names by type '$resource_type'");
            $Log->exit();
            return (wantarray()) ? () : [];
        }
        # pull in the resources into the single resource_list hash and filter out those that are not the right type
        foreach my $restype (keys %{$curr_resource_hsh}) {
            if ($restype eq $resource_type) {
                push(@{$resource_list}, @{$curr_resource_hsh->{$restype}});
                last;
            }
        }
        foreach my $restype (keys %{$curr_sub_resource_hsh}) {
            if ($restype eq $resource_type) {
                push(@{$resource_list}, @{$curr_sub_resource_hsh->{$restype}});
                last;
            }
        }
    }

    # if the resource list is an empty hash reference, then this is a special way of denoting
    # that there should be no resources of this type used.
    if (ref($resource_list) eq "HASH") {
        $Log->exit();
        return (wantarray()) ? [] : ();
    }

    # default the resource list to an empty array
    # if it is undefined
    if (!$resource_list) {
        $resource_list = [];
    }

    $Log->exit("Resource list: " . Dumper($resource_list));
    return (wantarray()) ? @{$resource_list} : $resource_list;
}

###############################################################################
# Method:    get_resource()
# Objective: Obtain the resource given a name and type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -default => "");
    my $parent_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                          -default => undef);
    my $parent_type = &ParamsMgr::get_arg(-arg => "parent_type",
                                          -default => undef);
    my @parent_types = &ParamsMgr::get_arg(-arg => "parent_types",
                                           -ref_type => "ARRAY",
                                           -default => []);
    my @resource_types = &ParamsMgr::get_arg(-arg => "types",
                                             -ref_type => "ARRAY",
                                             -default => []);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    if (not scalar(@resource_types) and not defined($resource_type)) {
        $self->_set_err("Invalid value for -types or -type");
        $Log->exit();
        return {};
    }

    # if a single resource type was specified push it on the types array
    # that will be used to iterate over
    if (defined($resource_type)) {
        push(@resource_types, $resource_type);
    }

    if ($parent_name) {
        if ($parent_type) {
            push(@parent_types, $parent_type);
        }

        # search for the resource in the parent's sub types
        foreach my $ptype (@parent_types) {
            if (exists($self->{_resources}->{$ptype}) and
                exists($self->{_resources}->{$ptype}->{$parent_name})) {
                foreach my $type (@resource_types) {
                    if (exists($self->{_resources}->{$ptype}->{$parent_name}->{sub_resources}->{$type}) and
                        exists($self->{_resources}->{$ptype}->{$parent_name}->{sub_resources}->{$type}->{$resource_name})) {
                        $Log->exit();
                        return $self->{_resources}->{$ptype}->{$parent_name}->{sub_resources}->{$type}->{$resource_name};
                    }
                }
            }
        }
    } else {
        # search for the resource in all the specified resource types
        foreach my $type (@resource_types) {
            if (exists($self->{_resources}->{$type}) and
                exists($self->{_resources}->{$type}->{$resource_name})) {
                $Log->exit();
                return $self->{_resources}->{$type}->{$resource_name};
            }
        }
    }

    # at this point we couldnt find any resource of the specified types,
    $self->_set_err("Could not find resource '$resource_name' under the following type(s) '@resource_types'");
    $Log->exit();
    return {};
}

###############################################################################
# Method:    get_resource_event_param_list()
# Objective: Obtain the per resource event parameters that are set on the
#            specified resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource_event_param_hash {
    $Log->enter();
    my ($self, %args) = @_;

    my $resource = $self->ResMgr::get_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get resource");
        $Log->exit();
        return {};
    }

    $Log->exit();
    return $resource->{args};
}

###############################################################################
# Method:    set_resource_event_param_list()
# Objective: Modify the per resource event parameters hash for the specified
#            resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_resource_event_param_hash {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $args = &ParamsMgr::get_arg(-arg => "args",
                                   -default => {},
                                   -ref_type => "HASH");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # obtain the resource that we will modify
    my $resource = $self->ResMgr::get_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get resource");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # update the event param hash for this resource
    $resource->{args} = $args;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    filter_by_attribute()
# Objective: Given a list of resources of mixed set of types, filter the
#            resources that match the specified attribute filter.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub filter_by_attribute {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @resource_names = &ParamsMgr::get_arg(-arg => "names",
                                             -ref_type => "ARRAY",
                                             -required => 1);
    my @resource_types = &ParamsMgr::get_arg(-arg => "types",
                                             -ref_type => "ARRAY",
                                             -required => 1);
    my @parent_names = &ParamsMgr::get_arg(-arg => "parent_names",
                                           -ref_type => "ARRAY",
                                           -default => []);
    my @parent_types = &ParamsMgr::get_arg(-arg => "parent_types",
                                           -ref_type => "ARRAY",
                                           -default => []);
    my $attribute = &ParamsMgr::get_arg(-arg => "attribute",
                                        -required => 1);
    my $value = &ParamsMgr::get_arg(-arg => "value",
                                    -default => undef);
    my $regex = &ParamsMgr::get_arg(-arg => "regex",
                                    -default => undef);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # verify that these parameters have proper values set
    if (not scalar(@resource_types)) {
        $self->_set_err("Invalid value for -types");
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    if (not scalar(@resource_names)) {
        $self->_set_err("Invalid value for -names");
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    if ((not defined($value) or $value =~ /^\s*$/) and (not defined($regex))) {
        $self->_set_err("Invalid value for -value or -regex");
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    my @final_resource_names = ();

    # filter all the resources that were passed in by the attr / value pair
    # cycle through all the resource types that were specified
    if (@parent_types) {
        foreach my $ptype (@parent_types) {
            # get all the parent names by this type
            if (not scalar(@parent_names)) {
                @parent_names = $self->ResMgr::get_resource_names_by_type(type => $ptype);
                if ($self->error()) {
                    $self->_set_err("Failed to get resource names by type '$ptype'");
                    return (wantarray()) ? () : [];
                }
            }
            foreach my $pname (@parent_names) {
                foreach my $type (@resource_types) {
                    my @new_resource_names = grep {
                        my $resource_name = $_;
                        my $attribute_value = $self->ResMgr::get_resource_attribute(-name => $resource_name,
                                                                                    -type => $type,
                                                                                    -parent_name => $pname,
                                                                                    -parent_type => $ptype,
                                                                                    -attribute => $attribute);
                        if ($self->error()) {
                            # filter out the resource if there was an error getting the attr value
                            # most likely will happen if there is more than one resource type
                            $self->_clear_err();
                            0;
                        } else {
                            # check if this resource attribute matches the value requested
                            # filter out the ones that dont match the value or dont match the regex
                            if (((defined($value) and $attribute_value eq $value) or
                                 (defined($regex) and $attribute_value =~ $regex))) {
                                1;
                            } else {
                                0;
                            }
                        }
                    } @resource_names;

                    if (scalar(@new_resource_names)) {
                        # accumulate the final resource name list
                        push(@final_resource_names, @new_resource_names);
                    }
                }
            }
        }
    } else {
        foreach my $type (@resource_types) {
            my @new_resource_names = grep {
                my $resource_name = $_;
                my $attribute_value = $self->ResMgr::get_resource_attribute(-name => $resource_name,
                                                                            -type => $type,
                                                                            -attribute => $attribute);
                if ($self->error()) {
                    # filter out the resource if there was an error getting the attr value
                    # most likely will happen if there is more than one resource type
                    $self->_clear_err();
                    0;
                } else {
                    # check if this resource attribute matches the value requested
                    # filter out the ones that dont match the value or dont match the regex
                    if (((defined($value) and $attribute_value eq $value) or
                         (defined($regex) and $attribute_value =~ $regex))) {
                        1;
                    } else {
                        0;
                    }
                }
            } @resource_names;

            if (scalar(@new_resource_names)) {
                # accumulate the final resource name list
                push(@final_resource_names, @new_resource_names);
            }
        }
    }

    $Log->exit();
    return (wantarray()) ? @final_resource_names : \@final_resource_names;
}

###############################################################################
# Method:    filter_by_event_parameter()
# Objective: Given a list of resources of mixed set of types, filter the
#            resources that match the specified event parameter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub filter_by_event_parameter {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @resource_names = &ParamsMgr::get_arg(-arg => "names",
                                             -ref_type => "ARRAY",
                                             -required => 1);
    my @resource_types = &ParamsMgr::get_arg(-arg => "types",
                                             -ref_type => "ARRAY",
                                             -required => 1);
    my @parent_names = &ParamsMgr::get_arg(-arg => "parent_names",
                                           -ref_type => "ARRAY",
                                           -default => []);
    my @parent_types = &ParamsMgr::get_arg(-arg => "parent_types",
                                           -ref_type => "ARRAY",
                                           -default => []);
    my $event_parameter = &ParamsMgr::get_arg(-arg => "event_parameter",
                                        -required => 1);
    my $value = &ParamsMgr::get_arg(-arg => "value",
                                    -default => undef);
    my $regex = &ParamsMgr::get_arg(-arg => "regex",
                                    -default => undef);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    # verify that these parameters have proper values set
    if (not scalar(@resource_types)) {
        $self->_set_err("Invalid value for -types");
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    if (not scalar(@resource_names)) {
        $self->_set_err("Invalid value for -names");
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    if ((not defined($value) or $value =~ /^\s*$/) and (not defined($regex))) {
        $self->_set_err("Invalid value for -value or -regex");
        $Log->exit();
        return (wantarray()) ? () : [];
    }

    my @final_resource_names = ();

    # filter all the resources that were passed in by the event_param / value pair
    # cycle through all the resource types that were specified
    if (@parent_types) {
        foreach my $ptype (@parent_types) {
            # get all the parent names by this type
            if (not scalar(@parent_names)) {
                @parent_names = $self->ResMgr::get_resource_names_by_type(type => $ptype);
                if ($self->error()) {
                    $self->_set_err("Failed to get resource names by type '$ptype'");
                    return (wantarray()) ? () : [];
                }
            }
            foreach my $pname (@parent_names) {
                foreach my $type (@resource_types) {
                    my @new_resource_names = grep {
                        my $resource_name = $_;
                        # grab the resource
                        my $resource =  $self->ResMgr::get_resource(-name => $resource_name,
                                                                    -type => $type,
                                                                    -parent_name => $pname,
                                                                    -parent_type => $ptype);
                        if ($self->error()) {
                            $self->_clear_err();
                            0;
                        }

                        my $event_param_value = $self->EventUtil::get_arg(-objects => [$resource],
                                                                          -arg => $event_parameter);
                        if ($self->error()) {
                            # filter out the resource if there was an error getting the attr value
                            # most likely will happen if there is more than one resource type
                            $self->_clear_err();
                            0;
                        } else {
                            # if the event param value is undefined, then return 1 only if
                            # value and regex are not undefined as well
                            if (not defined($event_param_value)) {
                                if (not defined($value) and not defined($regex)) {
                                    1;
                                } else {
                                    0;
                                }
                            } else {
                                # check if this resources event param matches the value requested
                                if (ref($event_param_value) eq "ARRAY") {
                                    if (defined($regex)) {
                                        # grep through the array using the regex
                                        grep($regex, @{$event_param_value});
                                    } elsif (defined($value)) {
                                        if (ref($value) eq "ARRAY") {
                                            # deref both values and compare the two arrays
                                            if (scalar(@{$event_param_value}) eq scalar(@{$value})) {
                                                my $match = 0;
                                                for (my $i = 0; $i < scalar(@{$event_param_value}); $i++) {
                                                    if ($event_param_value->[$i] eq $value->[$i]) {
                                                        # matched
                                                        $match = 1;
                                                    } else {
                                                        # no match, break out
                                                        $match = 0;
                                                        last;
                                                    }
                                                }
                                                # return the match value
                                                $match;
                                            } else {
                                                # arrays are not equal, no match
                                                0;
                                            }
                                        } else {
                                            # value is a scalar, if event param value has exactly one element and
                                            # it is equal to value, then we have a match
                                            if (scalar(@{$event_param_value}) == 1 and $event_param_value->[0] eq $value) {
                                                # matched
                                                1;
                                            } else {
                                                # no match
                                                0;
                                            }
                                        }
                                    } else {
                                        # no match
                                        0;
                                    }
                                } else {
                                    # event param value is a scalar, do the basic comparison against value & regex
                                    # filter out the ones that dont match the value or dont match the regex
                                    if (((defined($value) and $event_param_value eq $value) or
                                             (defined($regex) and $event_param_value =~ $regex))) {
                                        1;
                                    } else {
                                        0;
                                    }
                                }
                            }
                        }
                    } @resource_names;

                    # accumulate the final resource name list
                    if (scalar(@new_resource_names)) {
                        push(@final_resource_names, @new_resource_names);
                    }
                }
            }
        }
    } else {
        foreach my $type (@resource_types) {
            my @new_resource_names = grep {
                my $resource_name = $_;
                # grab the resource
                my $resource =  $self->ResMgr::get_resource(-name => $resource_name,
                                                            -type => $type);
                if ($self->error()) {
                    $self->_clear_err();
                    0;
                }

                my $event_param_value = $self->EventUtil::get_arg(-objects => [$resource],
                                                                  -arg => $event_parameter);
                if ($self->error()) {
                    # filter out the resource if there was an error getting the attr value
                    # most likely will happen if there is more than one resource type
                    $self->_clear_err();
                    0;
                } else {
                    # if the event param value is undefined, then return 1 only if
                    # value and regex are not undefined as well
                    if (not defined($event_param_value)) {
                        if (not defined($value) and not defined($regex)) {
                            1;
                        } else {
                            0;
                        }
                    } else {
                        # check if this resources event param matches the value requested
                        if (ref($event_param_value) eq "ARRAY") {
                            if (defined($regex)) {
                                # grep through the array using the regex
                                grep($regex, @{$event_param_value});
                            } elsif (defined($value)) {
                                if (ref($value) eq "ARRAY") {
                                    # deref both values and compare the two arrays
                                    if (scalar(@{$event_param_value}) eq scalar(@{$value})) {
                                        my $match = 0;
                                        for (my $i = 0; $i < scalar(@{$event_param_value}); $i++) {
                                            if ($event_param_value->[$i] eq $value->[$i]) {
                                                # matched
                                                $match = 1;
                                            } else {
                                                # no match, break out
                                                $match = 0;
                                                last;
                                            }
                                        }
                                        # return the match value
                                        $match;
                                    } else {
                                        # arrays are not equal, no match
                                        0;
                                    }
                                } else {
                                    # value is a scalar, if event param value has exactly one element and
                                    # it is equal to value, then we have a match
                                    if (scalar(@{$event_param_value}) == 1 and $event_param_value->[0] eq $value) {
                                        # matched
                                        1;
                                    } else {
                                        # no match
                                        0;
                                    }
                                }
                            } else {
                                # no match
                                0;
                            }
                        } else {
                            # event param value is a scalar, do the basic comparison against value & regex
                            # filter out the ones that dont match the value or dont match the regex
                            if (((defined($value) and $event_param_value eq $value) or
                                     (defined($regex) and $event_param_value =~ $regex))) {
                                1;
                            } else {
                                0;
                            }
                        }
                    }
                }
            } @resource_names;

            # accumulate the final resource name list
            if (scalar(@new_resource_names)) {
                push(@final_resource_names, @new_resource_names);
            }
        }
    }

    $Log->exit();
    return (wantarray()) ? @final_resource_names : \@final_resource_names;
}

###############################################################################
# Method:    set_resource_attribute()
# Objective: Set an attribute on a specific resource and propagate the change
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_resource_attribute {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my ($time, $utime) = gettimeofday();

    my $attribute = &ParamsMgr::get_arg(-arg => "attribute",
                                        -required => 1);
    my $value = &ParamsMgr::get_arg(-arg => "value",
                                    -required => 1);
    my $name = &ParamsMgr::get_arg(-arg => "name",
                                   -required => 1);
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -required => 1);
    my $propagate = &ParamsMgr::get_arg(-arg => "propagate",
                                        -default => $EventUtil::TRUE);
    my $timestamp = &ParamsMgr::get_arg(-arg => "timestamp",
                                        -default => $time . "." . $utime);
    my $parent_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                          -default => undef);
    my $parent_type = &ParamsMgr::get_arg(-arg => "parent_type",
                                          -default => undef);

    # default the propagate data value to be a regular resource attribute change
    my $propagate_data = &ParamsMgr::get_arg(-arg => "propagate_data",
                                             -default => { -data => { name => $name,
                                                                      type => $type,
                                                                      attribute => $attribute,
                                                                      value => $value,
                                                                      parent_name => $parent_name,
                                                                      parent_type => $parent_type,
                                                                      timestamp => $timestamp},
                                                           -msg => "set_resource_attribute" });
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # get the resource object
    my $resource = $self->ResMgr::get_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get resource");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # set the attribute on this resource object
    $resource->{attrs}->{$attribute}->{value} = $value;

    # store the timestamp for this attribute
    $resource->{attrs}->{$attribute}->{timestamp} = $timestamp;

    # if requested to propagate this resource change to the parent subtest,
    # do so here
    if ($propagate == $EventUtil::TRUE) {
        $self->SubtestComm::subtest_return_data(%{$propagate_data});
        if ($self->error()) {
            $Log->exit();
            $self->_set_err("Failed to return subtest data");
            return $EventUtil::FALSE;
        }
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_resource_attribute()
# Objective: Obtain the value of an attribute for a particular resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource_attribute {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $attribute = &ParamsMgr::get_arg(-arg => "attribute",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return undef;
    }

    # get the resource object
    my $resource = $self->ResMgr::get_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get resource");
        $Log->exit();
        return undef;
    }

    # get the attribute value on this resource object
    if (not exists($resource->{attrs}->{$attribute})) {
        $Log->exit();
        return undef;
    } else {
        $Log->exit();
        return $resource->{attrs}->{$attribute}->{value};
    }
}

###############################################################################
# Method:    set_faulted_resource()
# Objective: Mark that a resource is now in a faulted state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_faulted_resource {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # set the faulted attribute to true on this resource
    $self->ResMgr::set_resource_attribute(%args,
                                          -attribute => "faulted",
                                          -value => $EventUtil::TRUE);

    if ($self->error()) {
        $self->_set_err("Unable to set resource '$resource_name' as faulted");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # indicate that this resource is transitioning between a faulted/unfaulted state
    $self->ResMgr::set_resource_attribute(%args,
                                          -attribute => "transitioned_faulted_unfaulted",
                                          -value => $EventUtil::TRUE);
    if ($self->error()) {
        $self->_set_err("Unable to set resource '$resource_name' transition_faulted_unfaulted");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # if this is a nacl_nodes_cm resource, try to run this method on the sanfilersCm resource
    # as well, if it exists
    if ($resource_type eq "nacl_nodes_cm") {
        $self->ResMgr::set_faulted_resource(%args,
                                            -name => $resource_name,
                                            -type => "sanfilersCm");
        if ($self->error()) {
            # ignore Unable to get resource errors
            if ($self->errmsg() =~ /Unable to get resource/i) {
                $self->_clear_err();
            } else {
                $Log->exit();
                return $EventUtil::FALSE;
            }
        }
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    unset_faulted_resource()
# Objective: Mark that a resource is no longer in a faulted state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unset_faulted_resource {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # set the faulted attribute to true on this resource
    $self->ResMgr::set_resource_attribute(%args,
                                          -attribute => "faulted",
                                          -value => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to set resource '$resource_name' as unfaulted");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # if this is a nacl_nodes_cm resource, try to run this method on the sanfilersCm resource
    # as well, if it exists
    if ($resource_type eq "nacl_nodes_cm") {
        $self->ResMgr::unset_faulted_resource(%args,
                                              -name => $resource_name,
                                              -type => "sanfilersCm");
        if ($self->error()) {
            # ignore Unable to get resource errors
            if ($self->errmsg() =~ /Unable to get resource/i) {
                $self->_clear_err();
            } else {
                $Log->exit();
                return $EventUtil::FALSE;
            }
        }
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    modify_resource_obj()
# Objective: Modify the underlying object of the specified resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify_resource_obj {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $resource_name = &ParamsMgr::get_arg(-arg => "name",
                                            -required => 1);
    my $resource_type = &ParamsMgr::get_arg(-arg => "type",
                                            -required => 1);
    my $value = &ParamsMgr::get_arg(-arg => "value",
                                    -required => 1);
    my $reconnect = &ParamsMgr::get_arg(-arg => "reconnect",
                                        -default => $EventUtil::FALSE);
    my $propagate = &ParamsMgr::get_arg(-arg => "propagate",
                                        -default => $EventUtil::TRUE);
    my $serialized = &ParamsMgr::get_arg(-arg => "serialized",
                                         -default => $EventUtil::FALSE);
    my $parent_name = &ParamsMgr::get_arg(-arg => "parent_name",
                                          -default => undef);
    my $parent_type = &ParamsMgr::get_arg(-arg => "parent_type",
                                          -default => undef);

    # default the propagate data value to be a regular resource object modification
    my $propagate_data = &ParamsMgr::get_arg(-arg => "propagate_data",
                                             -default => { -data => { name => $resource_name,
                                                                      type => $resource_type,
                                                                      value => $value,
                                                                      parent_type => $parent_type,
                                                                      parent_name => $parent_name,
                                                                      serialized => $serialized},
                                                           -msg => "modify_resource_obj" });

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # get the resource object
    my $resource = $self->ResMgr::get_resource(%args);
    if ($self->error()) {
        $self->_set_err("Unable to get resource");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # propagate the resource change if requested to do so
    if ($propagate == $EventUtil::TRUE) {
        # if this is a san object, disconnect the connection and serialize the object
        # unless it's already been serialized.
        if ($resource_type =~ /sanclients|sanswitches/i and $serialized == $EventUtil::FALSE) {
            # destroy the connection on the object before we serialize it
            my $obj = $propagate_data->{-data}->{value};

            $Log->debug("Disconnecting connection to resource ($resource_name : $resource_type)");
            $obj->disconnect_connection();

            # serialize the object
            my $serialized_obj = Data::Dump::dump($obj);
            $propagate_data->{-data}->{value} = $serialized_obj;

            if ($Log->may_debug()) {
                $Log->debug("Serializing object for resource ($resource_name : $resource_type) : $serialized_obj");
            }

            # specify that the value of the data has been serialized
            $propagate_data->{-data}->{serialized} = $EventUtil::TRUE;
        }

        $self->SubtestComm::subtest_return_data(%{$propagate_data});
    }

    # if the value of the object came in serialized, then unserialize it now
    if ($serialized == $EventUtil::TRUE) {
        $Log->debug("Deserializing resource ($resource_name : $resource_type)");
        if ($resource_type =~ /sanclients/i) {
            $value = SanClient->new(-obj_template => eval($value));
            if ($value->error()) {
                $self->_set_err("Failed to instantiate SanClient obj", $value->errmsg());
                $Log->exit();
                return $EventUtil::FALSE;
            }
        } elsif ($resource_type =~ /sanswitches/i) {
            $value = SanSwitch->new(-obj_template => eval($value));
            if ($value->error()) {
                $self->_set_err("Failed to instantiate SanSwitch obj", $value->errmsg());
                $Log->exit();
                return $EventUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid serialized resource ($resource_name : $resource_type)");
            $Log->exit();
            return $EventUtil::FALSE;
        }

        if ($Log->may_debug()) {
            $Log->debug("Deserialized value of resource ($resource_name : $resource_type) is now: " . Data::Dump::dump($value));
        }
    }

    # if requested to reconnect this object, do so now so long as $value is
    # not a HASH, ARRAY, or SCALAR.
    if (ref($value) !~ /HASH|ARRAY|SCALAR/ and $reconnect == $EventUtil::TRUE) {
        if ($value->can("reconnect")) {
            $Log->debug("Reconnecting to resource ($resource_name : $resource_type)");
            $value->reconnect();
        }
    }

    # update the resource obj
    $resource->{obj} = $value;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    display_event_parameters_for_current_resources()
# Objective: Modify the underlying object of the specified resource
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_event_parameters_for_current_resources {
    $Log->enter();
    my ($self) = @_;

    my ($resource_hash) = $self->ResMgr::get_current_resources();
    if ($self->error()) {
        $self->_set_err("Unable to get current resources");
        $Log->exit();
        return $EventUtil::FALSE;
    }

    my @summary = ();
    push(@summary, "Event Parameters per Resource");

    foreach my $type (keys %{$resource_hash}) {
        push(@summary, "\tResource Type: '$type'");
        if (ref($resource_hash->{$type}) eq "HASH") {
            push(@summary, "\t\t(none)");
            next;
        }
        foreach my $resource_name (@{$resource_hash->{$type}}) {
            push(@summary, "\t\t'${resource_name}':");
            my $resource = $self->ResMgr::get_resource(-name => $resource_name,
                                                       -type => $type);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve resource '$resource_name' of type '$type'");
                $Log->exit();
                return $EventUtil::FALSE;
            }
            foreach my $event_param (keys %{$resource->{args}}) {
                my $event_param_value = $resource->{args}->{$event_param};
                if (ref($event_param_value) eq "ARRAY") {
                    push(@summary, "\t\t\t* ${event_param} : " . join(", ", @{$event_param_value}));
                } else {
                    push(@summary, "\t\t\t* ${event_param} : ${event_param_value}");
                }
            }
            if (scalar(keys %{$resource->{args}}) == 0) {
                push(@summary, "\t\t\t* (none)");
            }

            push(@summary, "");

            # find sub resources
            if ($resource->{sub_resources} and scalar keys %{$resource->{sub_resources}}) {
                foreach my $sub_resource_type (keys %{$resource->{sub_resources}}) {
                    push(@summary, "\t\t\tSub-Resource Type: '$sub_resource_type'");

                    foreach my $sub_resource_name (keys %{$resource->{sub_resources}->{$sub_resource_type}}) {
                        push(@summary, "\t\t\t\t'${sub_resource_name}':");
                        my $sub_resource = $self->ResMgr::get_resource(-name => $sub_resource_name,
                                                                       -type => $sub_resource_type,
                                                                       -parent_name => $resource_name,
                                                                       -parent_type => $type);
                        if ($self->error()) {
                            $self->_set_err("Unable to retrieve sub resource '$sub_resource_name' of " .
                                            "type '$sub_resource_type' from resource ($resource_name:$type)");
                            $Log->exit();
                            return $EventUtil::FALSE;
                        }
                        foreach my $event_param (keys %{$sub_resource->{args}}) {
                            my $event_param_value = $sub_resource->{args}->{$event_param};
                            if (ref($event_param_value) eq "ARRAY") {
                                push(@summary, "\t\t\t\t* ${event_param} : " . join(", ", @{$event_param_value}));
                            } else {
                                push(@summary, "\t\t\t\t* ${event_param} : ${event_param_value}");
                            }
                        }
                        if (scalar(keys %{$sub_resource->{args}}) == 0) {
                            push(@summary, "\t\t\t\t* (none)");
                        }
                    }
                    push(@summary, "");
                }
            }
        }
        push(@summary, "");
    }

    my $string = join("\n", @summary);

    &SanUtil::nate_log("-comment", "$string");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    get_resmgr_attrs
# Objective: Obtain the ResMgr Attr states
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resmgr_attrs {
    $Log->enter();
    my ($self) = @_;

    # obtain all the current resources that are in scope
    my $resource_hash = $self->ResMgr::get_current_resources(-exclude_private => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain current resources");
        $Log->exit();
        return undef;
    }

    # obtain all the current sub resources that are in scope
    my $sub_resource_hash = $self->ResMgr::get_current_sub_resources(-exclude_private => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to obtain current sub resources");
        $Log->exit();
        return undef;
    }

    # iterate through the list of resources and build a datastructure
    # to hold the attribute values for all resources
    my $resmgr_attr_state = {};
    foreach my $type (keys %{$resource_hash}) {
        foreach my $name (@{$resource_hash->{$type}}) {
            my $resource = $self->ResMgr::get_resource(-name => $name,
                                                       -type => $type);
            $resmgr_attr_state->{$type}->{$name}->{attrs} = $resource->{attrs};

            # get all sub resources from this resource, that are currently in scope
            # and add them to the resmgr attr state hash
            if ($resource->{sub_resources}) {
                foreach my $sub_rtype (keys %{$resource->{sub_resources}}) {
                    foreach my $sub_rname (keys %{$resource->{sub_resources}->{$sub_rtype}}) {
                        # skip this sub resource if its not in the current sub resource scope
                        # XXX: It's possible that we cause a race condition when two sub resources that share the same name
                        # are getting updated from separate event sets.
                        # There is a very small chance this will happen, but if:
                        # eventset X is using resources A and B, and resource A's sub resource N is being used
                        # and a parallel eventset Y is using resources A and B, and resource B's sub resource N` is
                        # being used, the attribute update procedure may cause a race condition because
                        # we are going think we have to update the attrs of N and N` in both eventsets. This
                        # will cause one of the eventsets to win the attr update over the other and not properly
                        # merge attr updates between eachother.
                        # To fix this, we need to be able to differentiate here whether $sub_rname truly is being
                        # used in the current event resource scope. Unfortunately there is no way to do that here,
                        # so the following line is the only thing we can do for now, unless you can figure out
                        # a better way. Fortunately there are not any events that will suffer from this problem
                        # yet.
                        next if (not grep(/^$sub_rname$/, @{$sub_resource_hash->{$sub_rtype}}));

                        $resmgr_attr_state->{$type}->{$name}->{sub_resources}->{$sub_rtype}->{$sub_rname}->{attrs} =
                            $resource->{sub_resources}->{$sub_rtype}->{$sub_rname}->{attrs};
                    }
                }
            }
        }
    }

    $Log->exit();
    return $resmgr_attr_state;
}

###############################################################################
# Method:    update_resmgr_attrs
# Objective: Update ResMgr attributes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub update_resmgr_attrs {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # parse out the ResMgr attribute datastructures that need to get applied
    # to the current ResMgr
    my @resmgr_updates = &ParamsMgr::get_arg(-arg => "resmgr_attrs",
                                             -required => 1,
                                             -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # apply each ResMgr attr datastructure update  to all the current resources.
    foreach my $state (@resmgr_updates) {
        foreach my $type (keys %{$state}) {
            foreach my $name (keys %{$state->{$type}}) {
                my $resource = $self->ResMgr::get_resource(-name => $name,
                                                           -type => $type);
                if ($self->error()) {
                    $self->_set_err("Unable to obtain resource for '$name' : '$type'");
                    $Log->exit();
                    return $EventUtil::FALSE;
                }

                foreach my $attr (keys %{$state->{$type}->{$name}->{attrs}}) {
                    # if the timestamp on this attr is higher than the timestamp on the
                    # current resource, or if this attr doesnt currently exist,
                    # update the attribute now
                    if ((not exists($resource->{attrs}->{$attr})) or
                        ($state->{$type}->{$name}->{attrs}->{$attr}->{timestamp} >=
                         $resource->{attrs}->{$attr}->{timestamp})) {
                        $resource->{attrs}->{$attr} = $state->{$type}->{$name}->{attrs}->{$attr};
                    }
                }

                # for all the sub resources owned by this resource, update their attrs as well
                if ($state->{$type}->{$name}->{sub_resources}) {
                    foreach my $stype (keys %{$state->{$type}->{$name}->{sub_resources}}) {
                        foreach my $sname (keys %{$state->{$type}->{$name}->{sub_resources}->{$stype}}) {
                            foreach my $attr (keys %{$state->{$type}->{$name}->{sub_resources}->{$stype}->{attrs}}) {
                                # if the timestamp of this attr is higher than the timestamp on the current sub resource,
                                # or if the attr doesnt currently exist on the sub resource, update the attr now
                                if ((not exists($resource->{sub_resources}->{$stype}->{$sname}->{attrs}->{$attr})) or
                                    ($state->{$type}->{$name}->{sub_resources}->{$stype}->{$sname}->{attrs}->{$attr}->{timestamp} >=
                                     $resource->{sub_resources}->{$stype}->{$sname}->{attrs}->{$attr}->{timestamp})) {
                                    $resource->{sub_resources}->{$stype}->{$sname}->{attrs}->{$attr} = $state->{$type}->{$name}->{sub_resources}->{$stype}->{$sname}->{attrs}->{$attr};
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    $Log->enter();
    my ($self) = @_;

    # document each known resource type
    $self->EventUtil::document_event_param(-param => "sanfilers7m",
                                           -package => __PACKAGE__,
                                           -default => "[]",
                                           -format => "ARRAY reference of 7mode SanFiler names",
                                           -example => "-sanfilers => [\"node_A\", \"node_B\"]",
                                           -per_resource_capable => "0",
                                           -is_a_resource => "1",
                                           -environment_capable => "1",
                                           -description => "Defines which 7mode SanFiler resources to use");

    $self->EventUtil::document_event_param(-param => "sanclients",
                                           -package => __PACKAGE__,
                                           -default => "[]",
                                           -format => "ARRAY reference of SanClient names",
                                           -example => "-sanclients => [\"client_A\"]",
                                           -per_resource_capable => "0",
                                           -is_a_resource => "1",
                                           -environment_capable => "1",
                                           -description => "Defines which SanClient resources to use");

    $self->EventUtil::document_event_param(-param => "sanswitches",
                                           -package => __PACKAGE__,
                                           -default => "[]",
                                           -format => "ARRAY reference of SanSwitch names",
                                           -example => "-sanswitches => [\"switch_A\", \"switch_B\"]",
                                           -per_resource_capable => "0",
                                           -is_a_resource => "1",
                                           -environment_capable => "1",
                                           -description => "Defines which SanSwitch resources to use");

    $self->EventUtil::document_event_param(-param => "controllerclusters",
                                           -package => __PACKAGE__,
                                           -default => "[]",
                                           -format => "ARRAY reference of ControllerCluster names",
                                           -example => "-controllerclusters => [\"clusterA\", \"clusterB\"]",
                                           -per_resource_capable => "0",
                                           -is_a_resource => "1",
                                           -environment_capable => "1",
                                           -description => "Defines which ControllerCluster resources to use");

    if ($self->error()) {
        $Log->exit();
        return $EventUtil::FALSE;
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

ResMgr.pm

=head1 Author

Robert Hunter

=head1 Description

The Resource Manager Utility for the EventMan library.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head1 B<Public Methods>

=head2 B<add_resource(%args)>

Registers a resource with the resource manager.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-obj>

Required: A blessed object representing the resource.

=item *I<-type>

Required: The type of resource to associate this object with.

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets error.

=item Examples

 $em_obj->ResMgr::add_resource(-name => "fas3070-rtp07",
                               -type => "sanfilersCm",
                               -obj  => $fas3070_rtp07_sanfiler_obj);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<add_sub_resource(%args)>

Registers a sub resource with the resource manager.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-obj>

Required: A blessed object representing the resource.

=item *I<-type>

Required: The type of resource to associate this object with.

=item *I<-parent_name>

Required: The name of parent resource to associate this object with.

=item *I<-parent_type>

Required: The type of parent resource to associate this object with.

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets error.

=item Examples

 $em_obj->ResMgr::add_sub_resource(-name => "myvserver",
                                   -type => "vservers",
                                   -parent_name => 'mycluster',
                                   -parent_type => 'nacl_clusters_cm',
                                   -obj  => $vserver_obj);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<remove_resource(%args)>

Unregisters a resource with the resource manager.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item *I<-parent_name>

Optional: The parent name of the resource.

=item *I<-parent_type>

Optional: The parent type of the resource

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE if not and sets error.

=item Examples

 $em_obj->ResMgr::remove_resource(-name => "fas3070-rtp07",
                                  -type => "sanfilersCm");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_resource_names_by_type(%args)>

Obtains all resource names that match the specified resource type.

=over 4

=item Inputs

=item *I<-type>

Required: The type of resources to collect.

=item Outputs

An array or array reference of resource names that match the resource
type -type. Sets error if not successful.

=item Examples

 my (@sanfiler_names) = $em_obj->ResMgr::get_resource_names_by_type(-type => "sanfilersCm");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_sub_resource_names_by_type(%args)>

Obtains all resource names that match the specified resource type.

=over 4

=item Inputs

=item *I<-type>

Required: The type of resources to collect.

=item *I<-parent_type>

Optional: The parent type to look through

=item *I<-parent_name>

Optional: The parent name to look through

=item Outputs

An array or array reference of sub resource names that match the resource
type -type. Sets error if not successful.

=item Examples

 my (@vserver_names) = $em_obj->ResMgr::get_sub_resource_names_by_type(-type => "vservers");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_all_resource_types(%args)>

Obtains all registered resource types

=over 4

=item Inputs

=item *I<-exclude_private>

Optional: If set to $EventUtil::TRUE, will exclude all private resource
types (resource types beginning with the underscore character "_")
Default: $EventUtil::TRUE

=item Outputs

An array or array reference of resource types

=item Examples

 my (@resource_types) = $em_obj->ResMgr::get_all_resource_types();
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_current_resources()>

Fetches all resources that are being used in the current
Event Environment.

=over 4

=item Inputs

=item *I<-objects>

Optional: An array reference of Event Parameter holding objects to search for resources.
Default: The current Event Parameter environment.

=item Outputs

A hash reference of resource array references categorized by resource types.

=item Examples

 my ($resource_hash) = $em_obj->ResMgr::get_current_resources();
 if ($em_obj->error()) {
     ...
 }

 foreach my $type (keys %{$resource_hash}) {
     foreach my $resource_name (@{$resource_hash->{$type}}) {
         ...
     }
 }

=back

=head2 B<get_current_sub_resources()>

Fetches all resources that are being used in the current
Event Environment.

=over 4

=item Inputs

=item *I<-objects>

Optional: An array reference of Event Parameter holding objects to search for resources.
Default: The current Event Parameter environment.

=item Outputs

A hash reference of resource array references categorized by resource types.

=item Examples

 my ($resource_hash) = $em_obj->ResMgr::get_current_sub_resources();
 if ($em_obj->error()) {
     ...
 }

 foreach my $type (keys %{$resource_hash}) {
     foreach my $resource_name (@{$resource_hash->{$type}}) {
         ...
     }
 }

=back

=head2 B<get_resource(%args)>

Returns the resource object for the specified resource name and type.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type> or I<-types>

Required: The -type of the resource or an array reference of -types

=item *I<-parent_name>

Required: The name of the parent resource.

=item *I<-parent_type> or I<-parent_types>

Required: The parent_type of the resource or an array reference of parent_types

=item Outputs

A hash reference representation of a resource

=item Examples

 # if the type of the resource is certain
 my ($resource) = $em_obj->ResMgr::get_resource(-name => "fas3070-rtp07",
                                                -type => "sanfilersCm");
 if ($em_obj->error()) {
     ...
 }

 # if the type of the resource is not certain
 my ($resource) = $em_obj->ResMgr::get_resource(-name => "fas3070-rtp07",
                                                -types => ["sanfilersCm", "sanfilers7m"]);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_resource_event_param_hash(%args)>

Returns the event parameter hash for the specified resource.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type> or I<-types>

Required: The -type of the resource or an array reference of -types

=item *I<-parent_name>

Required: The name of the parent resource.

=item *I<-parent_type> or I<-parent_types>

Required: The parent_type of the resource or an array reference of parent_types

=item Outputs

A hash reference of event parameter and value pairs that are set for this resource.

=item Examples

 # get the event params for this resource
 my ($event_params) = $em_obj->ResMgr::get_resource_event_param_hash(-name => "fas3070-rtp07",
                                                                     -type => "sanfilersCm");
 if ($em_obj->error()) {
     ...
 }

 $event_params->{sram_detection} = 1;

 # set the new event params for this resource
 $em_obj->ResMgr::set_resource_event_param_hash(-name => "fas3070-rtp07",
                                                -type => "sanfilersCm",
                                                -args => $event_params);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<set_resource_event_param_hash(%args)>

Modify the event parameter hash for the specified resource.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type> or I<-types>

Required: The -type of the resource or an array reference of -types

=item *I<-parent_name>

Required: The name of the parent resource.

=item *I<-parent_type> or I<-parent_types>

Required: The parent_type of the resource or an array reference of parent_types

=item *I<-args>

Optional: The new value for the event param hash for this resource.
Default: undef

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 # get the event params for this resource
 my ($event_params) = $em_obj->ResMgr::get_resource_event_param_hash(-name => "fas3070-rtp07",
                                                                     -type => "sanfilersCm");
 if ($em_obj->error()) {
     ...
 }

 $event_params->{sram_detection} = 1;
 $event_params->{enable_syslog_messages} = 0;

 # set the new event params for this resource
 $em_obj->ResMgr::set_resource_event_param_hash(-name => "fas3070-rtp07",
                                                -type => "sanfilersCm",
                                                -args => $event_params);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<filter_by_attribute(%args)>

Given a list of resource names of a specific set of types, filter out
the list of resources that do not contain an -attribute that
matches the specified -value or -regex.

=over 4

=item Inputs

=item *I<-names>

Required: An array reference of resource names (of the same type)

=item *I<-types>

Required: An array reference of resource types to search through

=item *I<-parent_names>

Optional: The names of the parent resource for the resource to filter

=item *I<-parent_types>

Optional: An array reference of parent_types for the resources to filter

=item *I<-attribute>

Required: The attribute to filter by

=item *I<-value> or I<-regex>

Required: The value or regex match that -attribute must be set to to
fufill the filter requirements. One or the other must be passed in.

=item Outputs

An array or array reference containing a subset of the resource names
specified in -names

=item Examples

 # only get sanfiler resources that are cmode capable
 my (@filtered_resource_names) = $em_obj->ResMgr::filter_by_attribute(-names => \@unfiltered_resource_names,
                                                                      -types => ["sanfilers7m", "sanfilersCm"],
                                                                      -attribute => "cmode_capable",
                                                                      -value => $EventUtil::TRUE);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<filter_by_event_parameter(%args)>

Given a list of resource names of a specific set of types, filter out
the list of resources that do not contain an event_parameter that
matches the specified -value or -regex.

=over 4

=item Inputs

=item *I<-names>

Required: An array reference of resource names (of the same type)

=item *I<-types>

Required: An array reference of resource types to search through

=item *I<-parent_names>

Optional: The names of the parent resource for the resource to filter

=item *I<-parent_types>

Optional: An array reference of parent types for the resources to filter

=item *I<-event_parameter>

Required: The event parameter to filter by

=item *I<-value> or I<-regex>

Required: The value or regex match that -event_parameter must be set to to
fulfill the filter requirements. One or the other must be passed in.

=item Outputs

An array or array reference containing a subset of the resource names
specified in -names

=item Examples

 # only get sanswitches that belong to the fabric "some fabric name"
 my (@filtered_resource_names) = $em_obj->ResMgr::filter_by_attribute(-names => \@unfiltered_resource_names,
                                                                      -types => ["sanswitches"],
                                                                      -event_parameter => "fabrics",
                                                                      -value => "some fabric name");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_current_resources_by_type(%args)>

Get the current resources in use of the specified type.

=over 4

=item Inputs

=item *I<-type>

Required: The type of the resource

=item Outputs

An array or array reference of resource names that match the type.

=item Examples

 # obtain the resources that are currently in use by this event by the specified type
 @sanfilersCm = $em_obj->ResMgr::get_current_resources_by_type(-type => "sanfilersCm");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<set_resource_attribute(%args)>

Set or modify an attribute value of the given resource name/type.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item *I<-parent_name>

Optional: The parent name of the resource.

=item *I<-parent_type>

Optional: The parent type of the resource

=item *I<-attribute>

Required: The attribute to change

=item *I<-propagate>

Optional: When set to $EventUtil::TRUE, will push the resource change to the parent subtests
so that the change is persistent between Events and EventSets.
Default: $EventUtil::TRUE

=item *I<-propagate_data>

Optional: When -propagate is set to $EventUtil::TRUE, this is the inputs into the
SubtestComm::subtest_return_data() method. For advanced users only. Note that
EventMan.pm must be modified to handle individual return data messages.
Default: A callback return data message to invoke this method recursively in the parent.

=item Outputs

The value of the attribute requested

=item Examples

 # Indicate that this controller is in a takeover state, propagate this
 # resource modification to the parent subtests so that it remains persistent
 # in the ResMgr
 $value = $em_obj->ResMgr::set_resource_attribute(-name => "fas3070-rtp07",
                                                  -type => "sanfilers7m",
                                                  -attribute => "in_takeover_mode",
                                                  -value => $EventUtil::TRUE,
                                                  -propagate => $EventUtil::TRUE);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_resource_attribute(%args)>

Get an attribute value of the given resource name/type.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item *I<-parent_name>

Optional: The parent name of the resource.

=item *I<-parent_type>

Optional: The parent type of the resource

=item *I<-attribute>

Required: The attribute to get

=item Outputs

The value of the attribute requested

=item Examples

 # determine if this resource is in a takeover state by checking its attribute
 $value = $em_obj->ResMgr::get_resource_attribute(-name => "fas3070-rtp07",
                                                  -type => "sanfilers7m",
                                                  -attribute => "in_takeover_mode");

 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<set_faulted_resource(%args)>

Mark that a resource is in a faulted state.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item *I<-parent_name>

Optional: The parent name of the resource.

=item *I<-parent_type>

Optional: The parent type of the resource

=item *I<-propagate>

Optional: When set to $EventUtil::TRUE, will push the resource change
to the parent subtests so that the change is persistent between
Events and EventSets.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 # mark this resource as faulted
 $em_obj->ResMgr::set_faulted_resource(-name => "fas3070-rtp07",
                                       -type => "sanfilers7m");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<unset_faulted_resource(%args)>

Mark that a resource is no longer in faulted state.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item *I<-parent_name>

Optional: The parent name of the resource.

=item *I<-parent_type>

Optional: The parent type of the resource

=item *I<-propagate>

Optional: When set to $EventUtil::TRUE, will push the resource change
to the parent subtests so that the change is persistent between
Events and EventSets.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 # mark this resource as unfaulted
 $em_obj->ResMgr::unset_faulted_resource(-name => "fas3070-rtp07",
                                         -type => "sanfilers7m");
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<modify_resource_obj(%args)>

Modifies the "obj" attribute of a resource to the new specified value.
WARNING: This does not work if the "obj" attribute is a blessed object.

=over 4

=item Inputs

=item *I<-name>

Required: The name of the resource.

=item *I<-type>

Required: The type of the resource

=item *I<-parent_name>

Optional: The parent name of the resource.

=item *I<-parent_type>

Optional: The parent type of the resource

=item *I<-value>

Required: The new value of the resource object

=item *I<-reconnect>

Optional: Reconnect the object after updating it
Default: $EventUtil::FALSE

=item *I<-propagate>

Optional: When set to $EventUtil::TRUE, will push the resource change
to the parent subtests so that the change is persistent between
Events and EventSets. Cannot propagate a object value change if the
new value is a blessed object.
Default: $EventUtil::TRUE

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 # change the object value of the switch_info resource.
 $em_obj->ResMgr::modify_resource_obj(-name => "switch_info",
                                      -type => "_GLOBALS",
                                      -value => $new_switch_info_obj);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<get_resmgr_attrs()>

Obtain the current ResMgr attribute datastructure.

=over 4

=item Inputs

None

=item Outputs

The current ResMgr attribute state datastructure.

=item Examples

 # obtain the current resmgr attr state
 my $resmgr_attr_state = $em_obj->ResMgr::get_resmgr_attrs();
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<update_resmgr_attrs(%args)>

Given a list of ResMgr update datastructures, update the current ResMgr Attr
state.

=over 4

=item Inputs

=item *I<-resmgr_attrs>

Required: An array reference of resmgr update attribute datastructures

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 # Update the current ResMgr Attribute state with the given attr change array
 $em_obj->ResMgr::update_resmgr_attrs(-resmgr_attrs => \@resmgr_attr_changes);
 if ($em_obj->error()) {
     ...
 }

=back

=head2 B<display_event_parameters_for_current_resources()>

=over 4

=item Inputs

None

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not.

=item Examples

 # change the object value of the switch_info resource.
 $em_obj->ResMgr::display_event_parameters_for_current_resources();
 if ($em_obj->error()) {
     ...
 }

=back

=cut
