# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing FC Target SRAM Coredump detection
## @author hrivera@netapp.com
## @status private
## @pod here

package SRAMDetection;

use strict;
use warnings;

use San::SanUtil;
use ParamsMgr;

use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound;
use NACL::CS::NetworkFcpAdapter;

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    check_for_new_sram_dumps()
# Objective: Check if any nacl_nodes_cm resources in the current environment
#            have generated any new SRAM dumps.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub check_for_new_sram_dumps {
    my ($self) = @_;

    # get all nacl nodes that are not in a faulted state
    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                      -type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    foreach my $node_name (@nacl_nodes) {
        # get the resource object from the resource manager
        my $resource = $self->ResMgr::get_resource(-name => $node_name,
                                                   -types => ["nacl_nodes_cm"]);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # if sram dump detection is enabled on this controller, then perform the validation
        my $sram_detection = $self->EventUtil::get_arg(-arg => "sram_detection",
                                                       -objects => [$resource],
                                                       -default_to_env => 1);
        if ($sram_detection == 0) {
            # skip sram detection on this controller
            &SanUtil::nate_log('-comment', "Skipping SRAM Detection on ($node_name) " .
                                           "because it is disabled for this controller");
            next;
        }

        # copy off the original sram list that is stored in the ResMgr
        my $original_sram_list = $self->SRAMDetection::_get_sram_dump_list(-node_name => $node_name);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # load in the current sram list (this will update the ResMgr with the new list)
        $self->SRAMDetection::_load_sram_dump_list(-node_name => $node_name);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # get the current list from the ResMgr
        my $current_sram_list = $self->SRAMDetection::_get_sram_dump_list(-node_name => $node_name);

        # compare the the original and new sram lists
        my $sram_hash = {};
        foreach my $sram_file (@{$original_sram_list}, @{$current_sram_list}) {
            $sram_hash->{$sram_file}++;
        }

        # verify that every key in the sram hash has a value of 2
        # flag sram files that only have a value of 1
        my @new_sram_dumps = ();
        foreach my $sram_file (keys %{$sram_hash}) {
            if ($sram_hash->{$sram_file} == 1) {
                # flag this sram file
                push(@new_sram_dumps, $sram_file);
            }
        }

        # if there are any new SRAM files, issue a warning
        if (scalar(@new_sram_dumps)) {
            my $sram_files = join("\n\t", @new_sram_dumps);
            &SanUtil::nate_log('-warn', "New SRAM dumps detected on controller '${node_name}'\n\t${sram_files}");
        } else {
            &SanUtil::nate_log('-comment', "No new SRAM dumps detected on controller '${node_name}'");
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _setup_sram_dump_detection()
# Objective: Initialize the SRAM Dump list on all nacl_nodes_cm objects
###############################################################################
sub _setup_sram_dump_detection {
    my ($self) = @_;

    # get all nacl_node objects that are not in a faulted state
    my @nacl_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                      -type => "cmode");

    foreach my $node_name (@nacl_nodes) {
        # get the resource object from the resource manager
        my $resource = $self->ResMgr::get_resource(-name => $node_name,
                                                   -types => ["nacl_nodes_cm"]);
        if ($self->error()) {
            $self->_set_err("Unable to setup SRAM Detection setup on $node_name");
            return $EventUtil::FALSE;
        }

        # if sram dump detection is enabled on this controller, then set it up
        my $sram_detection = $self->EventUtil::get_arg(-arg => "sram_detection",
                                                       -objects => [$resource],
                                                       -default_to_env => 1);
        if ($sram_detection == 0) {
            # skip sram detection on this controller
            &SanUtil::nate_log('-comment', "Skipping SRAM Detection setup on ($node_name) " .
                                           "because it is disabled for this controller");
            next;
        }

        # enable sram dumps on all node fcp target ports
        $self->SRAMDetection::_enable_fcp_adapter_sram_dumps(-node_name => $node_name);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # next task is to load the initial sram dump list for each node and
        # store each one in the ResMgr
        $self->SRAMDetection::_load_sram_dump_list(-node_name => $node_name);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _enable_fcp_adapter_sram_dumps()
# Objective: Turn on sram dumps on all fcp adapters for this node
###############################################################################
sub _enable_fcp_adapter_sram_dumps {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain NACL Node object for node '$node_name'");
        return $EventUtil::FALSE;
    }

    # turn on sram dumps on each fcp adapter for this node
    try {
        my @fcp_adapter_objs = NACL::CS::NetworkFcpAdapter->fetch(command_interface => $nacl_node_obj,
                                                                  filter => {node => $node_name});
        my $apiset = $nacl_node_obj->apiset(category => "Node",
                                            interface => "CLI",
                                            set => "CMode");
        foreach my $fcp_adapter_obj (@fcp_adapter_objs) {
            my $cmd = "fcp dump -e " . $fcp_adapter_obj->adapter();
            $apiset->execute_raw_command(command => "run -node $node_name -command \"$cmd\"");
        }
    } catch NACL::Exceptions::NoElementsFound with {
        # no fcp adapters on this node
    } otherwise {
        my $e = shift;
        $self->_set_err("Failed to enable SRAM dumps for fcp adapters on $node_name, Exception: " . $e->text());
    };

    if ($self->error()) {
        return $EventUtil::FALSE
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _load_sram_dump_list()
# Objective: Discover the current sram dump list from this node and store it
###############################################################################
sub _load_sram_dump_list {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain NACL Node object for node '$node_name'");
        return $EventUtil::FALSE;
    }

    # fetch the sram dump list from the controller
    my @sram_list = ();
    try {
        my $apiset = $nacl_node_obj->apiset(category => "Node",
                                            interface => "CLI",
                                            set => "Systemshell");
        my $cmd = "ls /mroot/etc/log | cat | grep ispfct";
        my $output = $apiset->execute_raw_command(command => $cmd);
        @sram_list = split(/\n/, $output);
    } otherwise {
        my $e = shift;
        $self->_set_err("Failed to obtain SRAM dump list from $node_name, Exception: " . $e->text());
    };
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # store the sram list in the ResMgr for this node
    $self->SRAMDetection::_set_sram_dump_list(%args,
                                              -sram_list => \@sram_list);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _set_sram_dump_list()
# Objective: store the sram dump list of this node in the ResMgr
###############################################################################
sub _set_sram_dump_list {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);
    my $sram_list = &ParamsMgr::get_arg(-arg => "sram_list",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE
    }

    # update the sram list on each node resource (propagate the resource change up the stack)
    $self->ResMgr::set_resource_attribute(-name => $node_name,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "sram_list",
                                          -value => $sram_list,
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _get_sram_dump_list()
# Objective: Get the current sram dump list for this node from the ResMgr
###############################################################################
sub _get_sram_dump_list {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE
    }

    # fetch the sram_list attribute from ResMgr
    my $sram_list = $self->ResMgr::get_resource_attribute(-name => $node_name,
                                                          -type => "nacl_nodes_cm",
                                                          -attribute => "sram_list");
    if ($self->error()) {
        return $EventUtil::FALSE
    }

    return $sram_list;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    $self->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                       -arg_defaults =>
                                                           { sram_detection => 1 }
                                                       );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # load SRAM dump data on all nacl_node objects
    $self->SRAMDetection::_setup_sram_dump_detection();
    if ($self->error()) {
        $self->_set_err("Failed to load SRAM data on all nacl_nodes_cm objects");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    # document each parameter used
    $self->EventUtil::document_event_param(-param => "sram_detection",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-sram_detection => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Enable or disable SRAM Dump detection on all nacl_nodes_cm resources",
                                           -resource_types => ["nacl_nodes_cm"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

SRAMDetection.pm

=head1 Author

Robert Hunter

=head1 Description

An EventMan internal utility module designed to provide helper
methods to perform SRAM Dump detection on NetApp Controllers running
CMode

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-sram_detection>

=over 4

Description: When enabled, will perform sram dump detection on Netapp controllers

Per Resource Capable: yes

Resource Types: nacl_nodes_cm

Default: 1 (on)

=back

=head1 B<Public Methods>

=head2 B<check_for_new_sram_dumps()>

Performs SRAM dump detection on all nacl_nodes_cm resources in the current
EventMan/EventSet/Event environment.

=over 4

=item Inputs

None

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -sram_detection

=item Examples

 $EventMan_obj->SRAMDetection::check_for_new_sram_dumps();
 if ($EventMan_obj->error()) {
     ...
 }

=back

=cut
