# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing APIs for Subtest Communication
## @author hrivera@netapp.com
## @status private
## @pod here

package SubtestComm;

use ParamsMgr;
use Data::Dumper;
use Data::Dump;
use strict;
use warnings;
use NATE::STAF::Peer qw($Parent);
use NATE::Log qw(log_global);
use NATE::Events;

use vars qw ($Log);

$Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    subtest_fail()
# Objective: Tells a parent subtest that this subtest has encountered a failure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_fail {
    $Log->enter();
    my ($self, %args) = @_;

    # dont clear the error because it should be set in this case.
    # clearing it will cause this method to misbehave.

    &ParamsMgr::process_args(%args);

    # default behavior is that this method treats all failures as non graceful
    my $graceful = &ParamsMgr::get_arg(-arg => "graceful",
                                       -default => $EventUtil::FALSE);

    # default the error string to be the error message on self
    my $error_strings = &ParamsMgr::get_arg(-arg => "error",
                                            -default => [$self->errmsg()]);

    # ignoring ParamsMgr errors here because this method has to complete due
    # to it being an error handler

    my $msg = "";

    # determine if the error strings is an array ref or if it is a string
    if (ref($error_strings) eq "ARRAY") {
        # join all the error strings into a single string by new lines
        $msg = join("\n", @{$error_strings});
    } else {
        # otherwise it is a single string
        $msg = $error_strings;
    }

    # determine if this subtest failure should be graceful or non graceful
    if (not $graceful) {
        # issue a failure if not graceful
        &SanUtil::nate_log(-fail => "$msg", -abortive => 0);
    } else {
        # otherwise gracefully warn
        &SanUtil::nate_log(-warn => "$msg");
    }

    # build the message object to send to the parent
    my $message_obj = {};
    $message_obj->{type} = "subtest_msg";

    # determine what type of message to send to the parent
    if ($graceful == $EventUtil::FALSE) {
        # the parent should treat this as a non graceful failure
        $message_obj->{msg} = "fail";
    } else {
        # the parent should not take any action since it is a graceful failure
        $message_obj->{msg} = "fail_ignore";
    }
    $message_obj->{data}->{"error_string"} = $msg;
    $message_obj->{timestamp} = time();

    # send the error message
    $Parent->message_put(type => "EventMan/" . $message_obj->{type},
                         message => $message_obj);

    # spin lock on a sleep, this function should never exit on it's own,
    # the parent subtest will terminate this process.
    while (1) {
        NATE::Events::eventloop_drain();
        &SanUtil::nate_sleep($EventUtil::SUBTEST_EXIT_SLEEP_TIME);
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_stopped()
# Objective: Tells a parent subtest that this subtest has entered the stop state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_stopped {
    $Log->enter();
    my ($self, $msg) = @_;

    if ($msg) {
        &SanUtil::nate_log(-comment => "$msg");
    }

    $self->SubtestComm::_subtest_state_change("stopped");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_aborted()
# Objective: Tells a parent subtest that this subtest has entered the aborted
#            state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_aborted {
    $Log->enter();
    my ($self, $msg) = @_;

    &SanUtil::nate_log(-warn => "$msg", -abortive => 0);

    $self->SubtestComm::_subtest_state_change("aborted");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_timedout()
# Objective: Tells a parent subtest that this subtest has entered the timedout
#            state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_timedout {
    $Log->enter();
    my ($self, $msg) = @_;

    $self->SubtestComm::_subtest_state_change("timedout");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_complete()
# Objective: Tells a parent subtest that this subtest has completed
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_complete {
    $Log->enter();
    my ($self) = @_;

    $self->SubtestComm::_subtest_state_change("complete");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_wait()
# Objective: Tells a parent that this subtest is waiting to be continued
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_wait {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    # optional payload to send with subtest_wait
    my $data = &ParamsMgr::get_arg(-arg => "data",
                                   -default => undef);
    my $data_type = &ParamsMgr::get_arg(-arg => "data_type",
                                        -default => undef);
    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return undef;
    }

    my $message_obj = {};
    $message_obj->{type} = "subtest_msg";
    $message_obj->{msg} = "waiting";
    $message_obj->{data} = $data;
    $message_obj->{data_type} = $data_type;
    $message_obj->{timestamp} = time();

    $Parent->message_put(type => "EventMan/" . $message_obj->{type},
                         message => $message_obj);

    &SanUtil::nate_log('-comment', "Waiting for parent to continue this subtest");

    # time how long we are in a wait state
    my $stime = time();

    # wait for a parent to tell us to continue
    while (1) {
        # this will block
        my $message_obj = $Parent->message_get(filter => sub {
                                                   my $m = shift;
                                                   return $m->{type} =~ /EventMan/;
                                               });

        # if the message is shit, then start the loop again
        if (not defined($message_obj)) {
            &SanUtil::nate_sleep(1);
            next;
        }

        # handle subtest messages while in the waiting state
        if ($message_obj->{type} eq "subtest_msg") {
            if ($message_obj->{msg} eq "stop") {
                # change the current state if we received a stop command
                $self->_set_current_state($EventMan::STATE_STOPPED);
                &SanUtil::nate_log('-comment', "Parent sent message 'stop' while in 'waiting' state, " .
                                       "changing current state to 'stopped'");
            } elsif ($message_obj->{msg} eq "abort") {
                # change the current state if we received an abort command
                $self->_set_current_state($EventMan::STATE_ABORTED);
                &SanUtil::nate_log('-comment', "Parent sent message 'abort' while in 'waiting' state, " .
                                       "changing current state to 'aborted'");
            } elsif ($message_obj->{msg} eq "timeout") {
                # change the current state if we received a timeout command
                $self->_set_current_state($EventMan::STATE_TIMEDOUT);
                &SanUtil::nate_log('-comment', "Parent sent message 'timeout' while in 'waiting' state, " .
                                       "changing current state to 'timedout'");
            } elsif ($message_obj->{msg} eq "continue") {
                &SanUtil::nate_log('-comment', "Parent sent message 'continue'");

                # if a data payload was including in the message, process it
                if (defined($message_obj->{data_type})) {
                    if ($message_obj->{data_type} eq "update_resmgr_attrs") {
                        # update ResMgr attributes
                        $self->ResMgr::update_resmgr_attrs(-resmgr_attrs => [$message_obj->{data}]);
                        if ($self->error()) {
                            $self->_set_err("Failed to update ResMgr attributes");
                            $Log->exit();
                            return undef;
                        }
                    }
                }

                # break out of the wait loop once we received a continue command
                last;
            }
        } else {
            # unknown message type case
            # since we can't process this message now, store it in the message queue
            # so that it can be processed later
            push(@{$self->{_message_queue}}, $message_obj);
        }

    }

    # return how long we were waiting for
    $Log->exit();
    return time() - $stime;
}

###############################################################################
# Method:    subtest_continue()
# Objective: Tells the specified subtest to continue execution
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_continue {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);
    my $data = &ParamsMgr::get_arg(-arg => "data",
                                   -default => undef);
    my $data_type = &ParamsMgr::get_arg(-arg => "data_type",
                                        -default => undef);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "Telling subtest (" . $subtest->runid() . ") to continue");

    $self->SubtestComm::_subtest_command(-subtest => $subtest,
                                         -command => "continue",
                                         -data => $data,
                                         -data_type => $data_type);

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_exit()
# Objective: Tells the specified subtest to exit
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_exit {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "Telling subtest (" . $subtest->runid() . ") to exit");

    $self->SubtestComm::_subtest_command(-subtest => $subtest,
                                         -command => "exit");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_stop()
# Objective: Tells the specified subtest to stop all operations when it can
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_stop {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "Telling subtest (" . $subtest->runid() . ") to stop gracefully");

    $self->SubtestComm::_subtest_command(-subtest => $subtest,
                                         -command => "stop");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_abort()
# Objective: Tells the specified subtest to abort all operations when it can
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_abort {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "Telling subtest (" . $subtest->runid() . ") to abort");

    $self->SubtestComm::_subtest_command(-subtest => $subtest,
                                         -command => "abort");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_timeout()
# Objective: Tells the specified subtest to timeout when it can
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_timeout {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    &SanUtil::nate_log('-comment', "Telling subtest (" . $subtest->runid() . ") to timeout");

    $self->SubtestComm::_subtest_command(-subtest => $subtest,
                                         -command => "timeout");

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_return_data()
# Objective: Return customized data back to the parent subtest for it to
#            process
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_return_data {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $data_hash = &ParamsMgr::get_arg(-arg => "data",
                                        -default => {},
                                        -ref_type => "HASH");
    my $message = &ParamsMgr::get_arg(-arg => "msg",
                                      -default => "");

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    my $message_obj = {};
    $message_obj->{type} = "return_data";
    $message_obj->{msg} = $message;
    $message_obj->{timestamp} = time();
    $message_obj->{data} = $data_hash;

    $Parent->message_put(type => "EventMan/" . $message_obj->{type},
                         message => $message_obj);

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    subtest_passthrough()
# Objective: Broadcast a message and data to the specified subtest.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub subtest_passthrough {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $data = &ParamsMgr::get_arg(-arg => "data",
                                   -default => {},
                                   -ref_type => "HASH");
    my $msg = &ParamsMgr::get_arg(-arg => "msg",
                                  -required => 1);
    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # build the passthrough message object
    my $message_obj = {};
    $message_obj->{type} = "passthrough";
    $message_obj->{timestamp} = time();
    $message_obj->{msg} = $msg;
    $message_obj->{data} = $data;

    # send the message object to the child subtest specified
    $subtest->message_put(type => "EventMan/" . $message_obj->{type},
                          message => $message_obj);

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    log_event_message()
# Objective: Log an event message to the parent subtest for it to store in
#            the event messages queue
# Details:   See POD documentation at the end of this file.
###############################################################################
sub log_event_message {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $data_hash = &ParamsMgr::get_arg(-arg => "data",
                                        -default => {},
                                        -ref_type => "HASH");
    my $event_message = &ParamsMgr::get_arg(-arg => "event_message",
                                            -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    # build the event message object
    my $message_obj = {};
    $message_obj->{type} = "event_message";
    $message_obj->{msg} = "$event_message";
    $message_obj->{timestamp} = time();

    # indicate that this event has not been read yet
    $message_obj->{read} = 0;

    # store the user data and event data hashes into the message
    $message_obj->{data} = $data_hash;
    $message_obj->{event_data} = $self->_get_current_event_data();

    $Parent->message_put(type => "EventMan/" . $message_obj->{type},
                         message => $message_obj);

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    can_read()
# Objective: Determine if any messages are available on the message queue
# Details:   See POD documentation at the end of this file.
###############################################################################
sub can_read {
    $Log->enter();
    my ($self) = @_;

    # return true if there are any messages in the message queue or if the parent subtest
    # has sent us a new message of type EventMan
    if (defined($self->{_message_queue}) and scalar(@{$self->{_message_queue}}) > 0) {
        $Log->debug("Internal Message queue has something on it");
        $Log->exit();
        return $EventUtil::TRUE;
    }

    $Log->debug("Checking for presence of message on queue from parent");

    # this should not block for more than 100ms
    my $msg;
    if ($msg = $Parent->message_get(filter => sub { my $m = shift; return $m->{type} =~ /EventMan/i; },
                                    timeout => 0.1,
                                    on_timeout => sub { return 0; })) {
        if ($Log->may_debug()) {
            $Log->debug("Storing message on internal queue: " . Data::Dump::dump($msg));
        }

        # store the message that we may have pulled onto the internal msg queue
        push(@{$self->{_message_queue}}, $msg);

        $Log->exit();
        return $EventUtil::TRUE;
    } else {
        $Log->exit();
        return $EventUtil::FALSE;
    }
}

###############################################################################
# Method:    read_message_obj()
# Objective: Read a message from the internal message queue
# Details:   See POD documentation at the end of this file.
###############################################################################
sub read_message_obj {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $shift_message = &ParamsMgr::get_arg(-arg => "shift_message",
                                            -default => $EventUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    # if there is a message waiting (internal or on subtest handle)
    if ($self->SubtestComm::can_read()) {

        # grab the earliest message from the queue
        my $message_obj = @{$self->{_message_queue}}[0];

        if ($shift_message == $EventUtil::TRUE) {
            if ($Log->may_debug()) {
                $Log->debug("Shifting message $message_obj off internal message queue: " . Data::Dump::dump($message_obj));
            }
            # shift the message
            shift(@{$self->{_message_queue}});
        } else {
            if ($Log->may_debug()) {
                $Log->debug("Leaving message $message_obj on internal message queue: " . Data::Dump::dump($message_obj));
            }
        }

        $Log->exit();
        return $message_obj;
    } else {
        # otherwise we cant read off the queue
        $Log->exit();
        return {};
    }
}

###############################################################################
# Method:    has_parent_requested_to_stop()
# Objective: Determine if the parent subtest is asking this subtest to stop
# Details:   See POD documentation at the end of this file.
###############################################################################
sub has_parent_requested_to_stop {
    $Log->enter();
    my ($self) = @_;

    # while we still have messages sent to us from the parent
    while ($self->SubtestComm::can_read()) {
        my $message_obj = $self->SubtestComm::read_message_obj();

        # if the parent wants us to stop, abort, or timeout, then return true
        if ($message_obj->{type} eq "subtest_msg") {
            if ($message_obj->{msg} eq "stop"  or
                $message_obj->{msg} eq "abort" or
                $message_obj->{msg} eq "timeout") {
                $Log->exit();
                return $EventUtil::TRUE;
            }
        } else {
            # restore the message on the message queue
            if ($self->SubtestComm::restore_message_obj_on_queue(-message_obj => $message_obj) eq $EventUtil::FALSE) {
                # this means we have seen this message before on the queue, which means we
                # probably wrapped the message queue
                last;
            }

        }
    }

    # parent is not asking for us to stop or abort
    $Log->exit();
    return $EventUtil::FALSE;
}

###############################################################################
# Method:    restore_message_obj_on_queue()
# Objective: Given a message obj, restore it on the internal message queue
# Details:   See POD documentation at the end of this file.
###############################################################################
sub restore_message_obj_on_queue {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $message_obj = &ParamsMgr::get_arg(-arg => "message_obj",
                                          -required => $EventUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    push(@{$self->{_message_queue}}, $message_obj);

    # this function is needed because
    # we might get caught in an infinite loop since we just pushed this message
    # back onto the message queue, keep a record of this message, and if we hit
    # the exact same message again, return false to the parent caller
    if (defined($self->{_restored_messages}) and
            ref($self->{_restored_messages}) eq "ARRAY") {
        # scan the restored messages list and break out of the loop if the current message
        # is listed in the queue
        foreach my $restored_message_obj (@{$self->{_restored_messages}}) {
            if ($message_obj eq $restored_message_obj) {
                # clear the restored messages queue
                delete($self->{_restored_messages});
                # we found a match, break out now, this will inform the caller that
                # they should stop processing the message queue
                $Log->exit();
                return $EventUtil::FALSE;
            }
        }
    }
    # we havn't restored this message before so
    # store the message on the restored messages queue
    push(@{$self->{_restored_messages}}, $message_obj);

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _subtest_command()
# Objective: Tells the specified subtest to perform a specific command
###############################################################################
sub _subtest_command {
    $Log->enter();
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $subtest = &ParamsMgr::get_arg(-arg => "subtest",
                                      -required => 1);
    my $command = &ParamsMgr::get_arg(-arg => "command",
                                      -required => 1);
    my $data = &ParamsMgr::get_arg(-arg => "data",
                                   -default => undef);
    my $data_type = &ParamsMgr::get_arg(-arg => "data_type",
                                        -default => undef);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $Log->exit();
        return $EventUtil::FALSE;
    }

    my $message_obj = {};
    $message_obj->{type} = "subtest_msg";
    $message_obj->{msg} = $command;
    $message_obj->{timestamp} = time();
    $message_obj->{data} = $data;
    $message_obj->{data_type} = $data_type;

    if ($Log->may_debug()) {
        $Log->debug("Sending message to " . $subtest->runid() . ": " . Data::Dump::dump($message_obj));
    }

    $subtest->message_put(type => "EventMan/" . $message_obj->{type},
                          message => $message_obj);

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _subtest_state_change()
# Objective: Changes the current state of this running subtest
###############################################################################
sub _subtest_state_change {
    $Log->enter();
    my ($self, $state) = @_;

    my $message_obj = {};
    $message_obj->{type} = "subtest_msg";
    $message_obj->{msg} = $state;
    $message_obj->{timestamp} = time();

    if ($Log->may_debug()) {
        $Log->debug("Sending message to parent " . Data::Dump::dump($message_obj));
    }

    # tell the parent we are changing state
    $Parent->message_put(type => "EventMan/" . $message_obj->{type},
                         message => $message_obj);

    $Log->debug("Waiting for parent to kill us, entering eventloop_drain loop");

    while (1) {
        NATE::Events::eventloop_drain();
        &SanUtil::nate_sleep($EventUtil::SUBTEST_EXIT_SLEEP_TIME);
    }

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
#                               Internal Method
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    $Log->enter();
    my ($self) = @_;

    $Log->exit();
    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

SubtestComm

=head1 Author

Robert Hunter

=head1 Description

An EventMan internal utility module designed to provide helper
methods to abstract subtest communication.

=head1 B<Public Methods>

=head2 B<subtest_fail(%args)>

Sends a message to the parent subtest that this child subtest has
encountered a failure and must terminate now.

=over 4

=item Inputs

=item *I<-graceful>

Optional: Indicates how the parent subtest should handle the error. If
set to 0, then the parent should treat this error as a failure and attempt
to stop all event activities. If set to 1, then the parent will set acknoledge
the error but not treat it as a true failure.
Default: 0 (non graceful)

=item *I<-error>

Optional: The error string to send with the failure message
Default: (The current error on the eventman object)

=item Outputs

None

=item Examples

 # terminate this subtest non gracefully
 $sanfaults_obj->SubtestComm::subtest_fail();

 # terminate this subtest gracefully
 $sanfaults_obj->SubtestComm::subtest_fail(-graceful => $EventUtil::TRUE);

=back

=head2 B<subtest_stopped()>

Sends a message to the parent subtest that this child subtest has
entered the stopped state due to being asked to stop, the parent subtest
should handle this message and terminate the child subtest.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 # Tell the parent that we are in a stopped state
 $sanfaults_obj->SubtestComm::subtest_stopped();

=back

=head2 B<subtest_timedout()>

Sends a message to the parent subtest that this child subtest has
entered the timedout state due to being asked to timeout or if a
timer has expired the parent subtest should handle this message
and terminate the child subtest.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 # Tell the parent that we are in a timed out state
 $sanfaults_obj->SubtestComm::subtest_timedout();

=back

=head2 B<subtest_complete()>

Sends a message to the parent subtest that this child subtest has
successfully completed. The parent subtest should handle this message
and terminate the child subtest gracefully.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 # Tell the parent that we are ready to exit
 $sanfaults_obj->SubtestComm::subtest_complete();

=back

=head2 B<subtest_wait()>

Sends a message to the parent subtest that this child subtest is
waiting for confirmation to continue.
The parent subtest should handle this message by issuing a subtest_continue()
message when it wishes to continue the child's activities.

=over 4

=item Inputs

None

=item Outputs

The time (in seconds) this subtest waited.

=item Examples

 # Tell the parent we are waiting, this function will block
 # until the parent tells us to continue
 $sanfaults_obj->SubtestComm::subtest_wait();

=back

=head2 B<subtest_continue(%args)>

Sends a message to the child subtest that it can now continue activities.

=over 4

=item Inputs

=item *I<-subtest>

Required: A NATE Subtest object to send the message to.

=item *I<-data>

Optional: A data payload to send along with the continue message

=item *I<-data_type>

Optional: The type of data payload that is being sent

=item Outputs

None

=item Examples

 # Tell the child subtest to continue from a subtest_wait() call
 $em_obj->SubtestComm::subtest_continue(-subtest => $child_subtest);

=back

=head2 B<subtest_exit(%args)>

Sends a message to a child subtest that it can now exit gracefully.

=over 4

=item Inputs

=item *I<-subtest>

Required: A NATE Subtest object to send the message to.

=item Outputs

None

=item Examples

 # Tell the $child_subtest to exit
 $em_obj->SubtestComm::subtest_exit(-subtest => $child_subtest);

=back

=head2 B<subtest_stop(%args)>

Sends a message to a child subtest telling it to stop.

=over 4

=item Inputs

=item *I<-subtest>

Required: A NATE Subtest object to send the message to.

=item Outputs

None

=item Examples

 # Tell the $child_subtest to stop
 $em_obj->SubtestComm::subtest_stop(-subtest => $child_subtest);

=back

=head2 B<subtest_abort(%args)>

Sends a message to a child subtest telling it to abort.

=over 4

=item Inputs

=item *I<-subtest>

Required: A NATE Subtest object to send the message to.

=item Outputs

None

=item Examples

 # Tell the $child_subtest to abort
 $em_obj->SubtestComm::subtest_abort(-subtest => $child_subtest);

=back

=head2 B<subtest_timeout(%args)>

Sends a message to a child subtest telling it to timeout.

=over 4

=item Inputs

=item *I<-subtest>

Required: A NATE Subtest object to send the message to.

=item Outputs

None

=item Examples

 # Tell the $child_subtest to timeout
 $em_obj->SubtestComm::subtest_timeout(-subtest => $child_subtest);

=back

=head2 B<subtest_return_data(%args)>

Sends a custom return_data message to the parent subtest. The
Parent subtest can then process the message outside of the normal
subtest reaping methods (EventMan::wait_for_subtests())

=over 4

=item Inputs

=item *I<-msg>

Required: A message to associate the data with

=item *I<-data>

Required: A hash reference of key/value pairs to pass along with
the message

=item Outputs

None

=item Examples

 # Send customized data to the parent subtest to process
 $sanfaults_obj->SubtestComm::subtest_return_data(-msg => "handle_this_data",
                                                  -data => {key1 => "value1",
                                                            key2 => "value2"});

=back

=head2 B<subtest_passthrough(%args)>

Sends a passthrough message to the child subtest specified.

=over 4

=item Inputs

=item *I<-msg>

Required: A message to associate the passthrough message with

=item *I<-data>

Required: A hash reference of key/value pairs to pass along with
the message

=item *I<-subtest>

Required: A NATE Subtest object to send the passthrough message to.

=item Outputs

None

=item Examples

 # Send customized passthrough message to a child subtest
 $sanfaults_obj->SubtestComm::subtest_passthrough(-msg => "do_something_with_this",
                                                  -subtest => $child_subtest,
                                                  -data => {key1 => "value1",
                                                            key2 => "value2"});

=back

=head2 B<log_event_message(%args)>

Sends an event message up the stack to the Top level parent subtest.
Use EventMan::read_event_message() to read these events. In addition
To the custom -data hash that can be specified, Current event data is
embedded in the event indicated what resources are currently being used
in this event's resource and environment scope. Other data such as what
eventset/event, eventset iteration, event count, event set sequence iteration
are stored in the event so that the event can be uniquely identified
when examined.

=over 4

=item Inputs

=item *I<-event_message>

Required: A message to associate the event message with

=item *I<-data>

Optional: A hash reference of key/value pairs to pass along with
the message
Default: undef

=item Outputs

None

=item Examples

 # Log an event in the Event messages table
 $sanfaults_obj->SubtestComm::log_event_message(-msg => "Issued takeover on NODE A",
                                                -data => {node_name => "A",
                                                          someparam => "value2"});

=back

=head2 B<can_read()>

Determine if there is a message from a parent subtest that needs
to be read. This is in relation to messages sent via the SubtestComm::subtest_passthrough()
command.

=over 4

=item Inputs

None

=item Outputs

None

=item Examples

 # see if any passthrough messages are available  and process them all
 while ($self->SubtestComm::can_read()) {
    my $msg_obj = $self->SubtestComm::read_message_obj();
    ...
 }

=back

=head2 B<read_message_obj(%args)>

Read a message object off of the passthrough message queue

=over 4

=item Inputs

=item *I<-shift_message>

Optional: If set to $EventUtil::TRUE, will shift off the earliest unread message
If set to $EventUtil::FALSE, it will leave the message so that it can be read again
Default: $EventUtil::TRUE

=item Outputs

A message hash reference sent by SubtestComm::subtest_passthrough()

=item Examples

 # see if any passthrough messages are available  and process them all
 while ($self->SubtestComm::can_read()) {
    my $msg_obj = $self->SubtestComm::read_message_obj(-shift_message => $EventUtil::TRUE);
    if ($msg_obj->{msg} eq "time to do something") {
        # do something
    } elsif ($msg_obj->{msg} eq "sleep 100 seconds") {
        sleep 100;
    }
 }

=back

=head2 B<has_parent_requested_to_stop()>

See if the parent has sent a stop/abort/timeout message to the current process

=over 4

=item Inputs

None

=item Outputs

$EventUtil::TRUE or $EventUtil::FALSE

=item Examples

 # see if the parent has resquested to stop
 if ($self->SubtestComm::has_parent_requested_to_stop()) {
    # perform some kind of cleanup code and terminate the event
    return $EventUtil::TRUE;
 }

=back

=head2 B<restore_message_obj_on_queue(%args)>

Given a message object obtained from SubtestComm::read_message_obj(), push it back on the
message queue.

=over 4

=item Inputs

=item *I<-message_obj>

Required: The message object to restore

=item Outputs

Returns $EventUtil::TRUE if the message object was restored, returns $EventUtil::FALSE if
the message object has been restored more than once. This is used to prevent the caller
from entering into an infinite loop while reading the message queue and restoring messages
onto the queue.

=item Examples

 # see if any passthrough messages are available  and process them all
 while ($self->SubtestComm::can_read()) {
    my $msg_obj = $self->SubtestComm::read_message_obj(-shift_message => $EventUtil::TRUE);
    if ($msg_obj->{msg} eq "time to do something") {
        # do something
    } else {
        # this message is not for us, we should put it back on the queue
        if (not $self->SubtestComm::restore_message_obj_on_queue(-message_obj => $msg_obj)) {
            # we've wrapped the message queue, break out to prevent infinite loops
            last;
        }
    }
 }

=back

=cut
