# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing System Logging APIs to all resources
## @author hrivera@netapp.com
## @status private
## @pod here

package SysLogging;

use strict;
use warnings;
use Data::Dumper;
use Tharn;
use ParamsMgr;
use NATE::Events;
use NATE::STAF::Peer;
use NATE::Log qw(log_global);

my $Log = log_global();

###############################################################################
#                               Public Methods
###############################################################################

################################################################################
# Method:     syslog_msg_all()
# Objective:       Will send message to syslog if platform supports
#                  for all clients/controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg_all {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @message = &ParamsMgr::get_arg(-arg => "message",
                                      -ref_type => "ARRAY",
                                      -required => 1);
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");
    my $resource_name = &ParamsMgr::get_arg(-arg => "resource_name",
                                            -default => undef);
    my $resource_type = &ParamsMgr::get_arg(-arg => "resource_type",
                                            -default => undef);

    my $asynchronous_bool = &ParamsMgr::get_arg(-arg => "async",
                                                -default => 0);

    my $quiet = &ParamsMgr::get_arg(-arg => "quiet",
                                    -default => 0);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # if not asked to syslog in the current environment, return
    my $enable_syslog_messages = $self->EventUtil::get_arg(-arg => "enable_syslog_messages");
    if ($enable_syslog_messages == 0) {
        $Log->comment("Syslog messaging is disabled for this event");
        return $EventUtil::TRUE;
    }

    my @sanclients = ();
    my @nacl_nodes = ();

    # determine who to syslog to
    if ($type =~ /all|clients/i) {
        # get the sanclient objects to work on
        if (defined($resource_type) and defined($resource_name) and $resource_type =~ /sanclients/i) {
            # scope the resource to syslog to be just the client
            @sanclients = ($resource_name);
        } else {
            @sanclients = $self->EventUtil::get_sanclients(-faulted => $EventUtil::FALSE);
        }
    }
    if ($type =~ /all|controllers/i) {
        # get the sanfiler objects to work on
        @nacl_nodes = $self->EventUtil::get_nacl_nodes(-faulted => $EventUtil::FALSE,
                                                       -type => "cmode");
    }

    # send message to NATE logs before we syslog it
    $Log->comment("@message") if ($quiet ne $EventUtil::FALSE);

    if (not $asynchronous_bool) {
        # write to syslog (in subtests to not clutter event log)
        # also, background subtests to not cause large timeout
        # delays if host is hung and we are trying to write to syslog
        my $id = "syslog_messages";
        $self->{sys_logging_runid_counter}++;

        my $syslog_subtest = subtest(
            sub {
                $self->SysLogging::_syslogging_helper(sanclients => \@sanclients,
                                                      nacl_nodes => \@nacl_nodes,
                                                      message    => \@message);
                if ($self->error()) {
                    $self->SubtestComm::subtest_fail(-graceful => 1);
                    return $EventUtil::TRUE;
                }
                $self->SubtestComm::subtest_complete();
                return $EventUtil::FALSE;
            },
            -runid, $id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);}
        );

        # reap the main subtest for syslogging to clients
        $self->wait_for_subtests(-subtests => [$syslog_subtest],
                                 -dependence_type => "independent",
                                 -propagate_wait => 0,
                                 -propagate_events => 1,
                                 -enable_sync => 0,
                                 -check_cmds_from_parent => 0,
                                 -forward_cmds_from_parent => 0,
                                 -async => 0,
                                 -poll_time => 1);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        return $EventUtil::TRUE;
    } else {
        eval {
            my $peer = NATE::STAF::Peer->new(handle   => $self->{sys_logging_handle},
                                             endpoint => $self->{sys_logging_endpoint});

            $peer->message_put(type => "EventMan/Syslogging",
                               message => {
                                   msg     => \@message,
                                   clients => \@sanclients,
                                   nodes   => \@nacl_nodes
                               });
        };
    }
    return $EventUtil::TRUE;
}

################################################################################
# Method:     syslog_msg_client()
# Objective:       Will send message to client syslog if platform supports
#                  for all clients.
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg_client {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @message = &ParamsMgr::get_arg(-arg => "message",
                                      -ref_type => "ARRAY",
                                      -required => 1);
    my $client_name = &ParamsMgr::get_arg(-arg => "client_name",
                                          -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # extract the sanclient object from the sanclients datastructure
    my $sc_obj = $self->EventUtil::get_sanclient_obj(-client_name => $client_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain SanClient object for client '$client_name'");
        return $EventUtil::FALSE;
    }

    # reconnect this sanclient object
    $sc_obj->reconnect(-disconnect_existing_conn => $EventUtil::FALSE);
    if ($sc_obj->error()) {
        $self->_set_err($sc_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # get the resource object from the resource manager
    my $resource = $self->ResMgr::get_resource(-name => $client_name,
                                               -type => "sanclients");
    if ($self->error()) {
        &Tharn::disconnect($sc_obj->get_connection(), '-ok');
        return $EventUtil::FALSE;
    }

    # determine if this client has syslog enabled on it
    my $enable_syslog_messages = $self->EventUtil::get_arg(-arg => "enable_syslog_messages",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($enable_syslog_messages == 0) {
        $Log->comment("Syslog messages disabled for $client_name");
        &Tharn::disconnect($sc_obj->get_connection(), '-ok');
        return $EventUtil::TRUE;
    }

    # issue the syslog command for this client
    if ($sc_obj->can("syslog_msg")) {
        $sc_obj->syslog_msg(@message);
        if ($sc_obj->error) {
            $Log->warn("Could not write to syslog on $client_name");
            $sc_obj->_clear_err();
        } else {
            $Log->comment("Successfully wrote to syslog on $client_name");
        }
    }

    &Tharn::disconnect($sc_obj->get_connection(), '-ok');

    return $EventUtil::TRUE;
}

################################################################################
# Method:     syslog_msg_controller()
# Objective:       Will send message to the controller's syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg_controller {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @message = &ParamsMgr::get_arg(-arg => "message",
                                      -ref_type => "ARRAY",
                                      -required => 1);

    my $node_name = &ParamsMgr::get_arg(-arg => "node_name",
                                        -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
    if ($self->error()) {
        $self->_set_err("Could not obtain NACL::C::Node object for controller '$node_name'");
        return $EventUtil::FALSE;
    }

    # get the resource object from the resource manager
    my $resource = $self->ResMgr::get_resource(-name => $node_name,
                                               -types => ["nacl_nodes_cm"]);
    if ($self->error()) {
        $self->_set_err("Could not obtain resource object for '$node_name' of type 'nacl_nodes_cm'");
        return $EventUtil::FALSE;
    }

    # determine if this filer has syslog enabled on it
    my $enable_syslog_messages = $self->EventUtil::get_arg(-arg => "enable_syslog_messages",
                                                           -objects => [$resource],
                                                           -default_to_env => 1);
    if ($enable_syslog_messages == 0) {
        $Log->comment("Syslog messages disabled for $node_name");
        return $EventUtil::TRUE;
    }

    # issue the syslog command for this controller
    try {
        my $apiset = $nacl_node_obj->apiset(category => "Node",
                                            interface => "CLI",
                                            set => "CMode");
        foreach my $msg (@message) {
            my $cmd = "run -node $node_name -command \"logger ${msg}\"";
            $apiset->execute_raw_command(command => $cmd);
        }
        $Log->comment("Successfully wrote to syslog on $node_name");
    } otherwise {
        my $e = shift;
        $Log->warn("Could not write to syslog on $node_name\n" . $e->text());
    };

    return $EventUtil::TRUE;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    $self->EventUtil::register_utility_module_defaults(
                                                       -utility_module => __PACKAGE__,
                                                       -arg_defaults =>
                                                           { enable_syslog_messages => 1 }
                                                       );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # initialize the runid counter for syslogging
    $self->{sys_logging_runid_counter} = 0;

    #fork off subtest in background
    my $subtest = subtest(sub { $self->SysLogging::_listener() },
                          -runid, ".syslog_listener",
                          -bg, "-onexit" => sub { my ($self) = @_; $self->exithandler(1) });

    #while still in parent, obtain -endpoint and -handle to the subttest.
    #Store that on $self. May need to store subtest object on to self.
    $self->{sys_logging_handle} = $subtest->handle();
    $self->{sys_logging_endpoint} = $subtest->endpoint();
    $self->{sys_logging_subtest} = $subtest;

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    _listener()
# Objective:
###############################################################################
sub _listener {
   my ($self) = @_;

   my @workload = ();

   my $listener = {
       filter => sub { my ($entry) = @_; return ($entry->{type} eq "EventMan/Syslogging") },
       callback => sub {
           my ($msg) = @_;
           # store this message on the workload array
           push(@workload, $msg);
       },
   };
   NATE::Events::listener_add($listener);

   while (1) {
       # Poll for listener/timer events
       NATE::Events::eventloop_drain();

       # if there are any messages in the workload list, process them all now
       while (scalar(@workload)) {
           # read the latest workload message and process it
           my $msg = shift @workload;
           $self->{sys_logging_runid_counter}++;
           my @sanclients = @{$msg->{message}->{clients}};
           my @nacl_nodes = @{$msg->{message}->{nodes}};
           my @message = @{$msg->{message}->{msg}};
           $self->SysLogging::_syslogging_helper(sanclients => \@sanclients,
                                                 nacl_nodes => \@nacl_nodes,
                                                 message    => \@message);
           # ignore all errors
           $self->_clear_err();
       }
   }
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    #if subtest object exists on self, kill it
    $self->{sys_logging_subtest}->kill(message => ["-comment", "Stopping SysLogging Listener"]);

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    # document each parameter used in this module
    $self->EventUtil::document_event_param(-param => "enable_syslog_messages",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-enable_syslog_messages => 0",
                                           -per_resource_capable => "1",
                                           -environment_capable => "1",
                                           -description => "Enable or disable event syslogging on all SanClients/SanFilers",
                                           -resource_types => ["sanclients", "sanfilersCm", "sanfilers7m", "controllerclusters"]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

1;

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _syslogging_helper()
# Objective: subtests client's and filer's syslog threads
# Details:   see pod
###############################################################################
sub _syslogging_helper {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my @sanclients = &ParamsMgr::get_arg(-arg => "sanclients",
                                         -ref_type => "ARRAY",
                                         -required => 1);
    my @nacl_nodes = &ParamsMgr::get_arg(-arg => "nacl_nodes",
                                         -ref_type => "ARRAY",
                                         -required => 1);
    my @message = &ParamsMgr::get_arg(-arg => "message",
                                      -ref_type => "ARRAY",
                                      -required => 1);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $runid_padded = $self->{sys_logging_runid_counter};
    if (length($runid_padded) == 1) {
        $runid_padded = "00" . $runid_padded;
    } elsif (length($runid_padded) == 2) {
        $runid_padded = "0" . $runid_padded;
    }

    my @sts;
    # subtest off each client's syslog thread
    foreach my $client_name (@sanclients) {
        my $st_id = "client_syslog_messages_" . $client_name;
        my $c_st = subtest(sub {
                                # write to the syslog on this client
                                $self->SysLogging::syslog_msg_client(-client_name => $client_name,
                                                                     -message => \@message);
                                if ($self->error()) {
                                   $self->_set_err("Failed to write to syslog on client '$client_name'");
                                   $self->SubtestComm::subtest_fail(-graceful => 1);
                                   return $EventUtil::FALSE;
                                }

                                $self->SubtestComm::subtest_complete();
                                return $EventUtil::TRUE;
                            },
                           -runid, $runid_padded . "_" . $st_id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1)});
        push(@sts, $c_st);
    }

    # subtest off each filer's syslog thread
    foreach my $node_name (@nacl_nodes) {
        my $st_id = "controller_syslog_messages_" . $node_name;
        my $f_st = subtest(sub {
                                # write to the syslog on this controller
                                $self->SysLogging::syslog_msg_controller(-node_name => $node_name,
                                                                         -message => \@message);
                                if ($self->error()) {
                                    $self->_set_err("Failed to write to syslog on controller '$node_name'");
                                    $self->SubtestComm::subtest_fail(-graceful => 1);
                                    return $EventUtil::FALSE;
                                }

                                $self->SubtestComm::subtest_complete();
                                return $EventUtil::TRUE;
                            },
                           -runid, $runid_padded . "_" . $st_id, -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1)});
        push(@sts, $f_st);
    }

    # reap all the client syslogging subtests
    $self->wait_for_subtests(-subtests => \@sts,
                             -dependence_type => "independent",
                             -propagate_wait => 0,
                             -propagate_events => 1,
                             -enable_sync => 0,
                             -check_cmds_from_parent => 0,
                             -forward_cmds_from_parent => 0,
                             -async => 0,
                             -poll_time => 1);
    if ($self->error()) {
        $self->_set_err("Failed to write to syslogs on all resources");
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

=pod

=head1 Package Name

SysLogging.pm

=head1 Author

Robert Hunter

=head1 Description

An EventMan internal utility module designed to provide helper
methods to perform SysLogging on SanClients and SanFiler methods

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-enable_syslog_messages>


=over 4

Description: If set to '1', will enable syslog messages for all sanfilers and
sanclients that support the methods to do so.

Per Resource Capable: Yes

Resource Types: sanclients, sanfilersCm, sanfilers7m, controllerclusters

Default: 1 (on)

=back

=head1 B<Public Methods>

=head2 B<syslog_msg_all()>

Writes the specified -message to the syslogs of all
SanFiler, SanClient resources in the current EventMan/EventSet/Event
environment

=over 4

=item Inputs

=item *I<-message>

An array reference of message strings to submit to the syslogs of
all available resources

=item *I<-async>

A boolean that allows syslogging to occur synchronously by default ('0')
or asynchronously ('1')

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -enable_syslog_messages

=item Examples

 $sanfaults_obj->SRAMDetection::syslog_msg_all(-message => ["hello world", "hello world2"]);
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<syslog_msg_client()>

Writes the specified -message to the syslogs of all
SanClient resources in the current EventMan/EventSet/Event
environment

=over 4

=item Inputs

=item *I<-message>

An array reference of message strings to submit to the syslogs of
all available client resources

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -enable_syslog_messages

=item Examples

 $sanfaults_obj->SRAMDetection::syslog_msg_client(-message => ["hello world", "hello world2"]);
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=head2 B<syslog_msg_controller()>

Writes the specified -message to the syslogs of all
SanFiler resources in the current EventMan/EventSet/Event
environment

=over 4

=item Inputs

=item *I<-message>

An array reference of message strings to submit to the syslogs of
all available SanFiler resources

=item *I<-node_name>

The node to perform syslogging against.

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -enable_syslog_messages

=item Examples

 $sanfaults_obj->SRAMDetection::syslog_msg_controller(-message => ["hello world", "hello world2"]);
 if ($sanfaults_obj->error()) {
     ...
 }

=back

=cut
