# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing System Statistics Logging for Resources
## @author hrivera@netapp.com
## @status private
## @pod here

package SysStats;

use ParamsMgr;
use Data::Dumper;
use NACL::C::Statistics;
use NATE::BaseException qw (:try);

###############################################################################
#                               Public Methods
###############################################################################

################################################################################
# Method:     collect_cluster_statistics()
# Objective:       Collect System Statistics on c-mode clusters
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_cluster_statistics {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my @clusters = &ParamsMgr::get_arg(-arg      => "clusters",
                                       -default  => [],
                                       -ref_type => "ARRAY");

    my $type = &ParamsMgr::get_arg(-arg => "type", -default => "all");

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # skip statistics collection if not requested
    my $collect_cluster_statistics = $self->EventUtil::get_arg(-arg => "collect_cluster_statistics");
    if ($collect_cluster_statistics == $EventUtil::FALSE) {
        &SanUtil::nate_log('-comment', "Skipping cluster statistics collection");
        return $EventUtil::TRUE;
    }

    if ($type =~ /cmode|all/i) {
        # Get the controller cluster objects
        if (not scalar(@clusters)) {
            # The caller didn't give us any clusters so find them all
            @clusters = $self->EventUtil::get_nacl_clusters_cm();
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the NACL::C::Cluster objects");
                return $EventUtil::FALSE;
            }
        }

        # Now run the statistics show-periodic cdef for each cluster
        foreach my $cluster_name (@clusters) {
            # get the nacl nodes in scope that live in this cluster
            my @nacl_nodes_cm = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                                 -faulted => $EventUtil::FALSE,
                                                                 -clusters => [$cluster_name]);
            if ($self->error()) {
                $self->_set_err("Failed to get nacl nodes from $cluster_name");
                return $EventUtil::FALSE;
            }
            # Make sure we actually got back something
            if (not scalar(@nacl_nodes_cm)) {
                $self->_set_err("No node names were found for cluster '$cluster_name'");
                return $EventUtil::FALSE;
            }

            # Grab one node, get its object, and use it to run commands
            my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $nacl_nodes_cm[0]);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the NACL::C::Node object for '$nacl_nodes_cm[0]'");
                return $EventUtil::FALSE;
            }

            try {
                # collect a 10 second sample
                my $nacl_cs_stat_obj = NACL::C::Statistics->show_periodic(command_interface => $nacl_node_obj,
                                                                          iterations => "10",
                                                                          interval => "1",
                                                                          'method-timeout' => 35);

                # parse out cluster traffic stats
                my $clv_code1 = ":CLV: cluster=$cluster_name, " .
                                     "avg_cluster_busy=" .$nacl_cs_stat_obj->averages()->cluster_busy() . ", " .
                                     "max_cluster_busy=" .$nacl_cs_stat_obj->maximums()->cluster_busy() . ", " .
                                     "avg_cluster_sent=" .$nacl_cs_stat_obj->averages()->cluster_sent() . ", " .
                                     "max_cluster_sent=" .$nacl_cs_stat_obj->maximums()->cluster_sent() . ", " .
                                     "avg_cluster_recv=" .$nacl_cs_stat_obj->averages()->cluster_recv() . ", " .
                                     "max_cluster_recv=" .$nacl_cs_stat_obj->maximums()->cluster_recv();

                # parse out disk/cpu utilization for the cluster
                my $clv_code2 = ":CLV: filer=$cluster_name, cpu="  . $nacl_cs_stat_obj->averages()->cpu_busy()  . ", " .
                                                           "disk=" . $nacl_cs_stat_obj->averages()->data_busy() . ", mem=,";

                # due to burt570989, NACL::C::Statistics may parse corrupted data,
                # only print out the clv code if the data is not corrupted (in this case it should start with numbers)
                if ($nacl_cs_stat_obj->averages()->cluster_busy() =~ /^\d/) {
                    &SanUtil::nate_log('-comment', "$clv_code1");
                    &SanUtil::nate_log('-comment', "$clv_code2");
                }
            } otherwise {
                # This will happen because of burt576372,
                # not issuing a warning because it'll happen all the time
                my $e = shift;
                &SanUtil::nate_log('-comment', "Collect cluster statistics on $cluster_name hit an exception\n" . $e->stacktrace());

                # burt608235 need to flush the connection on the command interface $nacl_node_obj
                # discard all of the buffers
                eval {$nacl_node_obj->apiset()->get_connection()->discardbuf();};
            };
        }
    }

    return $EventUtil::TRUE;
}

################################################################################
# Method:     collect_sysstats_on_controllers()
# Objective:       Collect System Statistics on c-mode/7-mode controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysstats_on_controllers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");
    my @nacl_nodes_cm = &ParamsMgr::get_arg(-arg => "nacl_nodes_cm",
                                            -default => [],
                                            -ref_type => "ARRAY");
    my @sanfilers7m = &ParamsMgr::get_arg(-arg => "sanfilers7m",
                                          -default => [],
                                          -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # skip sysstat collection if not requested
    my $collect_controller_sysstats = $self->EventUtil::get_arg(-arg => "collect_controller_sysstats");
    if ($collect_controller_sysstats == $EventUtil::FALSE) {
        &SanUtil::nate_log('-comment', "Skipping controller sysstat collection");
        return $EventUtil::TRUE;
    }

    # collect sysstats on cmode controllers
    if ($type =~ /cmode|all/i) {

        # get cmode resources from the environment if none were specified
        if (not scalar(@nacl_nodes_cm)) {
            (@nacl_nodes_cm) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                                -faulted => $EventUtil::FALSE);
            if ($self->error()) {
                $self->_set_err("Unable to get NACL::C::Node resources");
                return $EventUtil::FALSE;
            }
        }

        &SanUtil::nate_log('-comment', "Collecting System Stats on C-mode Controllers");

        # collect CMode controller SAN Stats
        foreach my $node_name (@nacl_nodes_cm) {
            my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
            if ($self->error()) {
                $self->_set_err("Unable to get NACL::C::Node object for '$node_name'");
                return $EventUtil::FALSE;
            }

            try {
                my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                    interface => "CLI",
                                                    set => "CMode");
                my $cmd = "run -node $node_name -command \"sysstat -b -c 1 10\"";
                my $output = $apiset->execute_raw_command(command => $cmd);

                # Example sysstat output
                #  CPU     FCP   iSCSI Partner  Total     FCP   kB/s   iSCSI   kB/s Partner   kB/s    Disk   kB/s    CP  CP  Disk
                #                                          in    out      in    out      in    out    read  write  time  ty  util
                #  16%       0       0       0      0       0      0       0      0       0      0      94  16729   32%  Zs   23%
                my @header = ("cpu", "fcp", "iscsi", "partner", "total", "fcp_in", "fcp_out", "iscsi_in",
                              "iscsi_out", "partner_in", "partner_out", "disk_read", "disk_write", "cp_time", "cp_ty", "disk_util");
                my @output = split("\n", $output);

                # Get rid of any empty lines
                @output = grep(!/^\s*$/, @output);

                # Get rid of the command line
                @output = grep(!/priv set/, @output);

                # Pull off the header
                @output = grep(!/^\s*CPU/, @output);
                @output = grep(!/\s*in\s+out/, @output);

                my $data_line = shift(@output);
                # Strip off white space at the beginning of the data
                $data_line =~ s/^\s*//g;
                my @data = split(/\s+/, $data_line);

                # Build out the stats hash mapping one-to-one header to data
                my $stats;
                for (my $i = 0; $i < @header; $i++) {
                    $data[$i] =~ s/\%//g;
                    $stats->{$header[$i]} = $data[$i];
                }

                my ($cpu, $disk_util, $cp_time, $cp_ty) = ($stats->{cpu}, $stats->{disk_util}, $stats->{cp_time}, $stats->{cp_ty});
                my $clv_code = ":CLV: filer=$node_name, cpu=$cpu, disk=$disk_util, mem=, cp_time=$cp_time, cp_type=$cp_ty";
                &SanUtil::nate_log('-comment', "$clv_code");
            } otherwise {
                # sysstat collection failed, issue a warning
                &SanUtil::nate_log('-warn', "Failed to collect sysstats on $node_name");
            };
        }
    }

    # collect sysstats on 7mode controllers
    if ($type =~ /7mode|all/i) {
        # get 7mode resources from the environment if none were specified
        if (not scalar(@sanfilers7m)) {
            (@sanfilers7m) = $self->EventUtil::get_sanfilers(-faulted => $EventUtil::FALSE,
                                                             -type => "7mode");
            if ($self->error()) {
                $self->_set_err("Unable to get SanFiler7m resources");
                return $EventUtil::FALSE;
            }
        }

        &SanUtil::nate_log('-comment', "Collecting System Stats on 7-mode Controllers");

        # collect 7mode controller SAN stats
        foreach my $controller_name (@sanfilers7m) {
            my $sf_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $controller_name);
            if ($self->error()) {
                $self->_set_err("Unable to get SanFiler object for '$controller_name'");
                return $EventUtil::FALSE;
            }

            # collect sysstats for san on this node for 10 seconds
            $sf_obj->sysstat("-b 10");
            if ($sf_obj->error()) {
                &SanUtil::nate_log('-warn', "Failed to collect sysstats on $controller_name");
                $sf_obj->_clear_err();
                next;
            }
        }
    }

    return $EventUtil::TRUE;
}

################################################################################
# Method:     collect_memstats_on_controllers()
# Objective:       Collect Memory Statistics on c-mode/7-mode controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_memstats_on_controllers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $type = &ParamsMgr::get_arg(-arg => "type",
                                   -default => "all");
    my @nacl_nodes_cm = &ParamsMgr::get_arg(-arg      => "nacl_nodes_cm",
                                            -default  => [],
                                            -ref_type => "ARRAY");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # skip memstat collection if not requested
    my $collect_controller_memstats = $self->EventUtil::get_arg(-arg => "collect_controller_memstats");
    if ($collect_controller_memstats == $EventUtil::FALSE) {
        &SanUtil::nate_log('-comment', "Skipping controller memstat collection");
        return $EventUtil::TRUE;
    }

    # collect memory stats on cmode controllers
    if ($type =~ /cmode|all/i) {

        # get cmode resources from the environment if none were specified
        if (not scalar(@nacl_nodes_cm)) {
            (@nacl_nodes_cm) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                                -faulted => $EventUtil::FALSE);
            if ($self->error()) {
                $self->_set_err("Unable to get SanFiler resources");
                return $EventUtil::FALSE;
            }
        }

        &SanUtil::nate_log('-comment', "Collecting Memory Stats on C-mode Controllers");

        # collect CMode controller Memory Stats
        foreach my $node_name (@nacl_nodes_cm) {
            my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
            if ($self->error()) {
                $self->_set_err("Unable to get NACL::C::Node object for '$node_name'");
                return $EventUtil::FALSE;
            }
            try {
                my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                    interface => "CLI",
                                                    set => "CMode");
                # collect memstats on this node
                my $cmd = "run -node $node_name -command \"priv set diag; mem_stats -e\"";
                my $stats = $apiset->execute_raw_command(command => $cmd);

                # if memstats is not enabled, re-enable it
                if ($stats =~ /No saved copy of memory statistics/i) {
                    # re-enable mem stats
                    $cmd = "run -node $node_name -command \"priv set diag; mem_stats -b\"";
                    $apiset->execute_raw_command(command => $cmd);

                    # rerun memstats collection
                    $cmd = "run -node $node_name -command \"priv set diag; mem_stats -e\"";
                    $stats = $apiset->execute_raw_command(command => $cmd);
                }

                # parse the memstats and display it in a format that CLV can parse
                my @stats = split(/\n/, $stats);
                my @final_stats;
                foreach my $line (@stats) {
                    my ($allocations, $memory, $module_function);
                    if ($line =~ /^\s*(\d+)\s+(\d+)\s+0x\S+\s+\[(\S+)\]/) {
                        $allocations = $1;
                        $memory = $2;
                        $module_function = $3;
                    } elsif ($line =~ /^\s*(\d+)\s+(\d+)\s+(\S+)/) {
                        $allocations = $1;
                        $memory = $2;
                        $module_function = $3;
                    } else {
                        next;
                    }

                    # filter out functions that don't match the user defined filter
                    next if ($module_function !~ $self->EventUtil::get_arg(-arg => "clv_memstats_module_regex_filter"));
                    push(@final_stats, "module_function=${module_function},allocations=${allocations},memory=${memory}");
                }

                my $str = join(";", @final_stats);
                &SanUtil::nate_log('-comment', ":CLV: ONTAP Memory Usage on node=$node_name : '$str'");
            } otherwise {
                &SanUtil::nate_log('-warn', "Failed to collect memstats on $node_name");
            };
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Method:    __register_utility_module()
# Objective: Register this Utility module with the EventMan package
# Details:   See POD documentation at the end of this file.
###############################################################################
sub __register_utility_module {
    my ($self) = @_;

    $self->EventUtil::register_utility_module_defaults(-utility_module => __PACKAGE__,
                                                       -arg_defaults =>
                                                           { collect_controller_sysstats      => $EventUtil::TRUE,
                                                             collect_controller_memstats      => $EventUtil::TRUE,
                                                             collect_cluster_statistics       => $EventUtil::TRUE,
                                                             clv_memstats_module_regex_filter => qr/.*/ }
                                                       );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __init()
# Objective: Initialize this Utility Module
###############################################################################
sub __init {
    my ($self) = @_;

    # skip memstat collection if not requested
    my $collect_controller_memstats = $self->EventUtil::get_arg(-arg => "collect_controller_memstats");
    if ($collect_controller_memstats == $EventUtil::FALSE) {
        &SanUtil::nate_log('-comment', "Skipping controller memstat initialization");
        return $EventUtil::TRUE;
    }

    &SanUtil::nate_log('-comment', "Initializing Mem stats on C-mode Controllers");

    my (@nacl_nodes_cm) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                           -faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to get NACL::C::Node resources");
        return $EventUtil::FALSE;
    }

    # collect CMode controller Memory Stats
    foreach my $node_name (@nacl_nodes_cm) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Unable to get NACL::C::Node object for '$node_name'");
            return $EventUtil::FALSE;
        }
        try {
            my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                interface => "CLI",
                                                set => "CMode");
            # collect memstats on this node
            my $cmd = "run -node $node_name -command \"priv set diag; mem_stats -c; mem_stats -b\"";
            $apiset->execute_raw_command(command => $cmd);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught while collecting mem stats on node $node_name. " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __uninit()
# Objective: Uninitialize this Utility Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    # skip memstat collection if not requested
    my $collect_controller_memstats = $self->EventUtil::get_arg(-arg => "collect_controller_memstats");
    if ($collect_controller_memstats == $EventUtil::FALSE) {
        &SanUtil::nate_log('-comment', "Skipping controller memstat initialization");
        return $EventUtil::TRUE;
    }

    &SanUtil::nate_log('-comment', "Uninitializing Mem stats on C-mode Controllers");

    $self->EventUtil::reconnect_all_nacl_nodes(-faulted => $EventUtil::FALSE,
                                               -type => "cmode");
    if ($self->error()) {
        $self->_set_err("Failed to reconnect to all nacl nodes");
        return $EventUtil::FALSE;
    }

    my (@nacl_nodes_cm) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                           -faulted => $EventUtil::FALSE);
    if ($self->error()) {
        $self->_set_err("Unable to get NACL::C::Node resources");
        return $EventUtil::FALSE;
    }

    # collect CMode controller Memory Stats
    foreach my $node_name (@nacl_nodes_cm) {
        my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node_name);
        if ($self->error()) {
            $self->_set_err("Unable to get NACL::C::Node object for '$node_name'");
            return $EventUtil::FALSE;
        }
        try {
            my $apiset = $nacl_node_obj->apiset(category => "Node",
                                                interface => "CLI",
                                                set => "CMode");
            # collect memstats on this node
            my $cmd = "run -node $node_name -command \"priv set diag; mem_stats -c\"";
            $apiset->execute_raw_command(command => $cmd);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught while uninitializing mem stats on node $node_name. " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    __generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "collect_controller_sysstats",
                                           -package => __PACKAGE__,
                                           -default => $EventUtil::TRUE,
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-collect_controller_sysstats => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Enable or disable controller sysstat collection");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

SysStats.pm

=head1 Author

Robert Hunter

=head1 Description

Collect System Statistics for resources.

=head1 Event Parameters

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-collect_controller_sysstats>

=over 4

Description: When enabled, will collect controller sysstats during events
that utilize this Utility Module.

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-collect_controller_memstats>

=over 4

Description: When enabled, will collect controller memstats during events
that utilize this Utility Module.

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-clv_memstats_module_regex_filter>

=over 4

Description: Allows for finer granularity filtering when collecting ONTAP memstat
information for viewing in CLV (consolidated log viewer). Format for this parameter
is a perl regular expression defined. By default, no filter is defined (qr/.*/)

Per Resource Capable: No

Default: qr/.*/

Example: qr/bcomka/i

=back

=head1 B<Public Methods>

=head2 B<collect_sysstats_on_controllers()>

Collect controller sysstats for all sanfilers (7mode/Cmode) in the current
Event scope.

=over 4

=item Inputs

=item *I<-type>

Optional: Specify what controller types to operate against "7mode|cmode|all"

Default: "all"

=item *I<-nacl_nodes_cm>

Optional: Array reference of C-mode resources to operate against

Default: Empty Array (operate against all availble c-mode resources)

=item *I<-sanfilers7m>

Optional: Array reference of 7-mode resources to operate against

Default: Empty Array (operate against all availble 7-mode resources)

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -collect_controller_sysstats

=item Examples

 # Collect sysstats on all cmode controllers available in the current scope
 $EventManObj->SysStats::collect_sysstats_on_controllers(-type => "cmode");
 if ($self->error()) {
     ...
 }

 # Collect sysstats only on the specified 7mode controllers
 $EventManObj->SysStats::collect_sysstats_on_controllers(-nacl_nodes_cm => ["fas3070-rtp07", "fas3070-rtp08"],
                                                         -type => "7mode");
 if ($self->error()) {
     ...
 }

=back

=head2 B<collect_memstats_on_controllers()>

Collect controller memstats for all nacl_nodes_cm resources in the current
Event scope.

=over 4

=item Inputs

=item *I<-type>

Optional: Specify what controller types to operate against "|cmode|all"

Default: "all"

=item *I<-nacl_nodes_cm>

Optional: Array reference of C-mode resources to operate against

Default: Empty Array (operate against all availble c-mode resources)

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -collect_controller_memstats
 -clv_memstats_module_regex_filter

=item Examples

 # Collect memstats on all cmode controllers available in the current scope
 $EventManObj->Memstats::collect_memstats_on_controllers();
 if ($self->error()) {
     ...
 }

 # Collect memstats only on the specified Cmode controllers
 $EventManObj->Memstats::collect_memstats_on_controllers(-nacl_nodes_cm => ["fas3070-rtp07", "fas3070-rtp08"]);
 if ($self->error()) {
     ...
 }

=back

=head2 B<collect_cluster_statistics()>

Collect cluster statistics for all clusters (Cmode) in the current
Event scope.

=over 4

=item Inputs

=item *I<-type>

Optional: Specify what controller types to operate against "cmode|all"

Default: "all"

=item *I<-clusters>

Optional: Array reference of C-mode clusters to operate against

Default: Empty Array (operate against all availble c-mode clusters)

=item Outputs

$EventUtil::TRUE if successful, $EventUtil::FALSE and sets error if not

=item Event Parameter Dependencies

 -collect_cluster_statistics
 -clv_memstats_module_regex_filter

=item Examples

 # Collect statistics on all cmode clusters available in the current scope
 $EventManObj->SysStats::collect_cluster_statistics();
 if ($self->error()) {
     ...
 }

 # Collect statistics only on the specified Cmode clusters
 $EventManObj->SysStats::collect_cluster_statistics(-clusters => ["cluster1"]);
 if ($self->error()) {
     ...
 }

=back

=cut
