# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/EventMgr.pm#1 $

## @summary     Class for sequencing arbitrary pieces of code (code refs or
## @summary     thpls).
## @author      dl-afd@netapp.com
## @status      private
## @pod here

package EventMgr;

# Compiler directives.
use strict;

# Module import.
use IO::Select;
use Exporter;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN HASHREF ARRAYREF CODEREF OBJECT UNDEF);
use base qw(Exporter);
use TracePkg;
use Data::Dumper;

# Constants
use constant STANDARD_EVENT => 'std';         # Standard event
use constant POST_EVENT     => 'post_event';  # Run after std event
use constant DELAY_EVENT    => 'delay';       # Sleep arbitrary amount of time

# Globals
use vars qw(
	    $IO_mon_period
	    $Results
	    $Subtest_alarm
	    $Subtest_map
	    $Trace_obj
	    @EXPORT_OK
	    @Run_queue
	    );

@EXPORT_OK     = qw(STANDARD_EVENT POST_EVENT DELAY_EVENT);
@Run_queue     = ();
$Subtest_map   = {};
$Subtest_alarm = 2;  # Interval for checking subtest status.
$Results       = {   # Overall results structure.
    pass  => 0,      # Pass is a count, all others are the result objects.
    warn  => [],
    fail  => [],
    other => [],
};
$IO_mon_period = 5;  # Monitor I/O every 5 seconds.
$Trace_obj     = new TracePkg;

sub new {
    my $class = shift;
    my %params = 
	validate_mixed(@_,
		       [],
		       {
			   clients => { type => ARRAYREF, optional => 1 },
			   args    => { type => ARRAYREF, optional => 1 },
		       });
    my(
       $self,
       );

    $self = {};
    $self->{clients}   = $params{clients} if $params{clients};

    # Caller can specify arbitrary arg list (in the form of an array ref)
    # to be passed to each event.
    $self->{args} = exists $params{args} ? $params{args} : [];

    # Caller can specify multiple I/O objects to monitor during event
    # processing.
    $self->{monitors} = [];

    bless $self, $class;

    return $self;
} # end sub new


sub create_sequence {
    my $self = shift;
    my %params =
	validate_mixed(@_,
		       [],
		       {
			   name     => { type => SCALAR,  optional => 0 },
			   data_ref => { type => HASHREF | ARRAYREF,
				    	   optional => 0 },
			   bg       => { type => BOOLEAN, optional => 1 },
			   loop     => { type => SCALAR,  optional => 1 },
			   sleep    => { type => SCALAR,  optional => 1 },
			   notify   => { type => CODEREF, optional => 1 },
		       });
    my(
       $seq_obj,
       );

    # Create sequence object.  This sequence object is added to
    # the global run queue array and returned to the caller.
    # Caller will presumably add events to the sequence object
    # (otherwise, it's an empty sequence object, which is legal,
    # but useless).

    $seq_obj = EventMgr::EventSequence->new(%params);
    push @Run_queue, $seq_obj;

    return $seq_obj;
} # end sub create_sequence


sub add_monitor {
    my($self, $mon_obj) = @_;

    unless (defined $mon_obj) {
	die "EventMgr::add_monitor requires a monitorable object";
    }

    push @{$self->{monitors}}, $mon_obj;

    return 1;
} # end sub add_monitor


sub event_loop {
    my $self = shift;
    my(
       $alarm_ipc_handle,
       $event_obj,
       $event_type,
       $fh,
       $handler_method,
       $iosel_obj,
       $rc,
       $runid,
       $st_ipc_handle,
       $st_obj,
       $stest,
       $stest_alarm_ipc,
       $stest_alarm_obj,
       @event_args,
       @ready,
       @remove,
       );

    # All events are added at this point - tell each event sequence to
    # get everything in order and and prep for execution (account for
    # POST_EVENTs etc.)
    foreach my $evseq_obj (@Run_queue) {
	$evseq_obj->finalize;
    }

    # Loop through the run queue and determine if we have any
    # sequences that are foreground.  If there are, any subsequent
    # sequences must be blocked from initially running (instantiated
    # event sequences are initially set to runnable).
    for (my $i = 0; $i <= $#Run_queue; $i++) {
	unless ($Run_queue[$i]->is_background) {
	    # Set remaining sequences (that following this foreground
	    # sequence) as non runnable.
	    for (my $j = $i + 1; $j <= $#Run_queue; $j++) {
		$Run_queue[$j]->set_runnable(0);
	    }

	    last;  # Remaining sequences have been properly deferred.
	}
    }

    # Main event loop - keep going until we have exhausted all
    # event sequences, or something has failed and we need to
    # jump out.
    #
    # Basic logic is:
    #
    # 1.  Start any I/O monitors that have been defined.
    #
    # 2.  Get the IPC handle from the I/O monitors - add to list of
    #     sockets to listen for message from.
    #
    # 3.  Start next runnable event (via subtest) from as many sequences
    #     as we can (all sequences that are marked runnable), using the
    #     IPC option to subtest.
    #
    # 4.  Get the IPC handle (socket) from the newly started subtest and
    #     add it to the IO::Select wait queue.
    #
    # 5.  Wait for one of the subtests to send a message back on the IPC
    #     pipe.
    #
    # 6.  When a message arrives, determine which subtest the message
    #     is from, and the sequence to which it belongs.
    #
    # 7.  Message can be one of 'update sequence data ref', subtest failed,
    #     or subtest completed successfully.
    #
    #     If message is 'update sequence data ref', update the dataref
    #     for the event sequence associated with this subtest and wait
    #     for more messages (nothing else marked runnable at this point).
    #
    #     If message is subtest failed, mark all remaining sequences as
    #     done, jump out of event loop, wait for oustanding subtests to
    #     complete and deal with failures.
    #
    #     If message is subtest completed successfully, grab the next
    #     runnable event from the associated event sequence the completed
    #     subtest came from and go back to waiting for subtests to send
    #     messages.
    #
    # 8. Once all event sequences have been exhausted, reap the results
    #    and return them back to the caller.

    $iosel_obj = new IO::Select;

    # Start monitor alarm - this is a subtest that will poke
    # the parent every n seconds to check the status of monitors.
    $runid = "monitor_alarm";
    $stest = &Tharn::subtest(\&EventMgr::_monitor_alarm,
			     -runid  => $runid,
			     -onexit => sub {},
			     "-bg",
			     "-ipc",
			     "--",
			     $IO_mon_period);
	
    #################################################
    # Need to handle failure in launching subtest
    #################################################

    $alarm_ipc_handle = $stest->ipc;
    $fh = $alarm_ipc_handle->fileno;

    # We're interested in message back from I/O monitors.
    $iosel_obj->add($alarm_ipc_handle);
    $Subtest_map->{private}{$fh} = {
	subtest_obj   => $stest,
	subtest_runid => $runid,
	sequence_obj  => undef,   # not associated w/ a sequence
	event_type    => undef,   # monitor, not event.
	msg_handler   => "_mon_alarm_msg_handler",
    };

    while ($self->_keep_going) {

	foreach my $evseq_obj (@Run_queue) {
	    if ($evseq_obj->runnable) {
		# This event sequence has a runnable event.
		$event_obj = $evseq_obj->get_runnable_event;

		# Run id is set to the id of the event sequence.
		# This is done to avoid collisions between parallel
		# event sequences, but minimize the number of subtest logs.
		($runid = $evseq_obj->get_id) =~ s/\s/_/g;


		# If this is a delay event, only pass delay time.
		@event_args = ();
		$event_type = $event_obj->get_type;
		if ($event_type eq DELAY_EVENT) {
		    push @event_args, ("--", $event_obj->get_delay);
		}
		else {
		    # Pass any caller specify args to event.
		    if (@{$self->{args}}) {
			push @event_args, ("--", @{$self->{args}});
		    }
		}

		$Trace_obj->comment(msg => "Starting event: " .
				    "[Loop: " . $evseq_obj->get_cur_loop .
				     " of " . $evseq_obj->get_num_loop . "]" .
				    "  [Sequence: " . "\"" .
				    $evseq_obj->get_id . "\"]  [Event: " .
				    "\"" . $event_obj->get_id . "\"]");
		$stest = &Tharn::subtest($event_obj->get_code,
					 -runid  => $runid,
					 -onexit =>  sub {},
					 "-bg",
					 "-ipc",
					 @event_args);

		# Deal with failure in ability to launch subtest here...
		# For now, we're going to assume it subtested properly.

		# Tell this event sequence that we're currently running
		# one of the events in it's run queue (will lock down
		# the sequence).
		$evseq_obj->event_started;

		# Need IPC handle for this subtest to send data and
		# wait for return messages.
		$st_ipc_handle = $stest->ipc;
		$fh = $st_ipc_handle->fileno;

		# Send data structure for this event's sequence to
		# the newly subtested event.
		$st_ipc_handle->write_obj($evseq_obj->get_dataref);

		# We're interested in messages back on this IPC handle.
		$iosel_obj->add($st_ipc_handle);
		$Subtest_map->{event}{$fh} = {
		    subtest_obj   => $stest,
		    subtest_runid => $runid,
		    sequence_obj  => $evseq_obj,
		    event_type    => $event_type,
		    msg_handler   => "_event_msg_handler",
		};

		# What other things do we need to do here before waiting
		# for messages back from subtests?

	    } # if runnable

	} # foreach evseq_obj

	# Wait for msgs back from subtest or i/o monitors.
	@ready = $iosel_obj->can_read();
	foreach my $ipc_sock (@ready) {
	    $fh = $ipc_sock->fileno;

	    if (exists $Subtest_map->{event}{$fh}) {
		# Call the appropriate msg handler - pass the subtest obj.
		$st_obj = $Subtest_map->{event}{$fh}{subtest_obj};
		$handler_method = $Subtest_map->{event}{$fh}{msg_handler};
		$self->$handler_method($st_obj, $iosel_obj);
	    }
	    elsif (exists $Subtest_map->{private}{$fh}) {
		# Call the appropriate msg handler - pass the subtest obj.
		$st_obj = $Subtest_map->{private}{$fh}{subtest_obj};
		$handler_method = $Subtest_map->{private}{$fh}{msg_handler};
		$self->$handler_method($st_obj, $iosel_obj);
	    }
	    else {
		# Msg came back from subtest object that does not exist
		# in Subtest map.
		$Trace_obj->warn(msg => "!! Object with runid: \"$runid\" " .
				 "and filehandle $fh is signaling ready in " .
				 "iowait list, but does not exist in " .
				 "Subtest map");
	    }
	}
    } # while keep_going

    # Event loop comleted either because we have completed all events
    # successfully, or something failed.  Shutdown all monitors and
    # wait on any outstanding events.
    $Trace_obj->comment(msg => "Event sequence complete, stopping monitors");
    foreach my $fh (keys %{$Subtest_map->{private}}) {
	$stest = $Subtest_map->{private}{$fh}{subtest_obj};
	$alarm_ipc_handle = $stest->ipc;
	if ($stest->is_alive) {
	    $alarm_ipc_handle->write_message("stop\n");
	}
	# Not interested in msgs from this montor anymore.
	$iosel_obj->remove($alarm_ipc_handle);
	push @remove, $fh;
    }

    foreach my $fh (@remove) {
	delete $Subtest_map->{private}{$fh};
    }

    # For now, kill off all subtests.  Need to have a cleaner way of
    # shutting down (waiting for the subtest to finish, sending
    # the appropriate msg back, etc.
    #$Trace_obj->comment(msg => "Waiting for outstanding events to finish");
    $Trace_obj->comment(msg => "Killing off outstanding events");
    foreach my $fh (keys %{$Subtest_map->{event}}) {
	$stest = $Subtest_map->{event}{$fh}{subtest_obj};
	$st_ipc_handle = $stest->ipc;
	if ($stest->is_alive) {
	    # The only reason we would kill a subtest is because I/O
	    # failed.  Otherwise, we won't get here because all event
	    # subtests will have completed.
	    $stest->kill(type => "comment",
			 "Killed due to external failure");
# 	    $Trace_obj->debug(msg => "Calling wait_finish on subtest with " .
# 			      "filehandle $fh");
# 	    $stest->wait_finish;
	}
	# Not interested in msgs from this subtest anymore.
	$iosel_obj->remove($st_ipc_handle);
	push @remove, $fh;
    }

    foreach my $fh (@remove) {
	delete $Subtest_map->{event}{$fh};
    }
    
    $rc = @{$Results->{fail}} ? 0 : 1;
    return $rc;
} # end sub event_loop


sub get_fail_results {
    my $self = shift;

    return @{$Results->{fail}};
} # end sub get_fail_results


sub get_warn_results {
    my $self = shift;

    return @{$Results->{warn}};
} # end sub get_warn_results


sub get_pass_count {
    my $self = shift;

    return $Results->{pass};
} # end sub get_pass_results


sub get_other_results {
    my $self = shift;

    return @{$Results->{other}};
} # end sub get_other_results


##################################
# Private methods
##################################

sub _shutdown {
    my($self)  = @_;

    # Force a shutdown of each of the event sequences (does not
    # kill off subtests that are in progress).
    foreach my $evseq_obj (@Run_queue) {
	$evseq_obj->set_runnable(0);
	$evseq_obj->set_sequence_complete(1);
	$evseq_obj->set_forcibly_shutdown(1);
    }

    return 1;
} # end sub shutdown


sub _keep_going {
    my $self = shift;
    my(
       $keep_going,
       );

    $keep_going = 0;

    # Loop through all event sequence objects and see if we still have
    # some that are in process.
    foreach my $ev_seq_obj (@Run_queue) {
	unless ($ev_seq_obj->sequence_complete) {
	    $keep_going = 1;

	    # Found an incomplete sequence - don't need to keep checking.
	    last;  
	}
    }

    return $keep_going;
} # end sub _keep_going


sub _monitor_alarm {
    my($alarm_interval) = @_;
    my(
       $ipc_sock,
       $msg,
       );

    $Trace_obj->debug(msg => "Monitor alarm: alarm interval is " .
		      $alarm_interval);

    # Wake up every $alarm_interval seconds and poke the parent
    # to force it to check the status of all it's monitors (i/o, etc.)

    # This routine is launched as a subtest - retrieve IPC handle.
    $ipc_sock = Subtest::ipc;

    # Loop until the parent tells us to stop.
  ALARMLOOP:
    while (1) {
	# Check to see if parent told us to stop.
	if ($ipc_sock->can_read) {
	    $msg = $ipc_sock->read_message;
	    last ALARMLOOP if $msg =~ /^stop/i;
	}

	select(undef, undef, undef, $alarm_interval); 	# Reliable sleep.
	$Trace_obj->debug(msg => "Poking parent with alarm msg");
	$ipc_sock->write_message("monitor alarm\n");

	# To prevent the parent from getting backed up, wait 
	# for ack of alarm from parent.
	eval {
	    $msg = $ipc_sock->read_message;
	};
	if ($@) {
	    if ($@ =~ /(unexpected eof)|(connection reset by peer)/i) {
		$msg = "IPC connection from parent closed unexpectedly";
		$Trace_obj->warn(msg => $msg);
		Tharn::addresult(type => "WARN", $msg);
		return;
	    }
	}

	# Msg received cleanly from parent - we're expecting an ack in
	# response to the poke, but in some cases, we could receive a
	# stop request from parent at this point as well.
	last ALARMLOOP if $msg =~ /^stop/i;
	if ($msg =~ /^ok/i) {
	    $Trace_obj->debug(msg => "Received proper alarm ack from parent");
	}
	else {
	    $Trace_obj->debug(msg => "Received strange msg from parent: $msg");
	}
    }

    $Trace_obj->comment(msg => "Monitor alarm: received stop msg from parent");
    Tharn::addresult(type => "PASS",
		     "monitor alarm completed successfully");
    return 1;
} # end sub _monitor_alarm


sub _event_msg_handler {
    my($self, $subtest_obj, $iosel_obj) = @_;
    my(
       $dataref,
       $event_obj,
       $evseq_obj,
       $fail,
       $fh,
       $ipc_sock,
       $notify_cb,
       $st_finished,
       $st_map_entry,
       @results,
       );

    $st_finished = 0;

    # There's 2 ways we can get a msg from an event.  First, it has
    # ended without sending us any data back (pass or fail), and
    # we're being notified because the other end of the socket has
    # shut down.  Secondly, we may be sent a data message from the event
    # subtest that we need to handle.
    $ipc_sock = $subtest_obj->ipc;
    $fh = $ipc_sock->fileno;
    $st_map_entry = $Subtest_map->{event}{$fh};
    $evseq_obj = $st_map_entry->{sequence_obj};
    if ($subtest_obj->is_alive) {
	# Subtest reports that it's still running, however, it might
	# be in shutdown mode where the other end of the IPC socket
	# has been closed, but it's still reporting a child process.
	if ($ipc_sock->can_read) {
	    eval {
		$dataref = $ipc_sock->read_obj;
	    };
	    if ($@) {
		if ($@ =~ /(unexpected eof)|(connection reset by peer)/i) {
		    $st_finished = 1;
		}
		else {
		    die "Strange socket error while attempting to read: $@";
		}
	    }

	    unless ($st_finished) {
		unless ($dataref) {
		    # Replace with better error handling?
		    die "Could not get updated data reference from event " .
			"$st_map_entry->{subtest_runid}";
		}

		$evseq_obj->set_dataref($dataref);

		# Ack caller.  This part was added so we won't run into the
		# case where the caller would post a message and then exit
		# before the parent could read it.
		$ipc_sock->write_message('data received');
	    }
	}
    }
    else {
	$st_finished = 1;
    }

    if ($st_finished) {
	# Ensure subtest is finished with it's shutdown, etc.  Clean up
	# and get results from subtest.
	$subtest_obj->wait_finish;
	$fail = 0;
	@results = $subtest_obj->results;

	# If we have anything other than a pass, we fail and stop
	# the whole event sequencing engine.
	foreach my $result_obj (@results) {
	    # Anything above a warn result will cause everything to
	    # shutdown.
	    if ($result_obj->severity > NATE::Result::Warn->severity) {
		push @{$Results->{fail}}, $result_obj;
		$fail = 1;
	    }
	    else {
		if ($result_obj->type =~ /^pass$/i) {
		    # Don't store pass results - array can grow too large.
		    # Also, users might be confused by the fact that sleep
		    # is implemented as a delay event - don't count those.
		    $Results->{pass}++
			unless ($st_map_entry->{event_type} eq DELAY_EVENT);
		    $Trace_obj->debug(msg => "Event completed successfully " .
				      "with the following messages:");
		    foreach my $msg ($result_obj->messages) {
			$Trace_obj->debug(msg => $msg);
		    }
		}
		elsif ($result_obj->type =~ /^warn$/i) {
		    push @{$Results->{warn}}, $result_obj;
		}
		else {
		    # non-fail result type.
		    push @{$Results->{other}}, $result_obj;
		}
	    }
	}


	# Not monitoring this subtest any more.
	delete $Subtest_map->{event}{$fh};
	$iosel_obj->remove($ipc_sock);

	# Cleanly shutdown the socket.
	$ipc_sock->shutdown(0);
	close($ipc_sock);
	
	# If we didn't fail, notify associated event sequence that
	# this event is complete.
	if ($fail) {
	    $self->_shutdown;
	    return;
	}
	else {
	    # Need to get the current event object now, since calling
	    # event_completed will increment it and we'd get the
	    # next available one to run instead of the one that just
	    # completed.
	    $event_obj = $evseq_obj->get_cur_event;
	    $evseq_obj->event_completed;

	    # Did caller specify a notification callback?  Don't
	    # invoke callback for delay events.
	    unless ($st_map_entry->{event_type} eq DELAY_EVENT) {
		if ($notify_cb = $evseq_obj->get_notify_cb) {
		    &{$notify_cb}($event_obj->get_id,
				  $evseq_obj->get_id,
				  $evseq_obj->get_cur_loop,
				  $evseq_obj->get_duration,
				  $evseq_obj->get_dataref);
		}
	    }
	}
    }

    return 1;
} # end sub _event_msg_handler


sub _mon_alarm_msg_handler {
    my($self, $subtest_obj, $iosel_obj) = @_;
    my(
       $cl_stopped,
       $fh,
       $ipc_sock,
       $msg,
       $st_finished,
       @status,
       @stopped,
       );

    # Normally, we'll get a msg from the monitor alarm subtest waking
    # us up every n seconds to check on the monitors.  If, for some
    # reason, the subtest ends, we'll also get a message from the socket
    # closing.
    $ipc_sock = $subtest_obj->ipc;
    $fh = $ipc_sock->fileno;
    if ($subtest_obj->is_alive) {
	# Subtest is still alive - we're receiving a msg from monitor alarm.
	if ($ipc_sock->can_read) {
	    eval {
		$msg = $ipc_sock->read_message;
	    };
	    if ($@) {
		if ($@ =~ /(unexpected eof)|(connection reset by peer)/i) {
		    $st_finished = 1;
		}
		else {
		    die "Strange socket error while attempting to read: $@";
		}
	    }

	    unless ($st_finished) {
		if ($msg =~ /^monitor alarm/i) {
		    $Trace_obj->debug(msg => "Received alarm msg from " .
				      "monitor alarm subtest");

		    # Loop through our monitors and check status.
		    foreach my $mon_obj (@{$self->{monitors}}) {
			@status = $mon_obj->status;
			foreach my $client_status (@status) {
			    if (@{$client_status->{stopped}}) {
				# One or more clients have stopped - shut
				# everything down.
				$cl_stopped =
				    join("\n", @{$client_status->{stopped}});
				$Trace_obj->warn(msg => "The following " .
						 "clients have stopped: \n" .
						 $cl_stopped . "\n" .
						 "shutting down");
				$self->_shutdown;
				return;
			    }
			}
		    }

		    # alarm subtest is waiting for ack.
		    $ipc_sock->write_message("ok\n");
		    $Trace_obj->debug(msg => "All clients still running");
		}
		else {
		    # unexpected msg from I/O monitor.
		    $Trace_obj->warn(msg => "Unexpected message from I/O " .
				     "monitor: $msg");

		    # alarm subtest is waiting for ack.
		    $ipc_sock->write_message("improper msg from alarm\n");
		}
	    }
	}

    }
    else {
	$st_finished = 1;
    }

    if ($st_finished) {
	# Ensure subtest is finished with shutdown, etc.
	$subtest_obj->wait_finish;

	# Not monitoring this subtest anymore.
	delete $Subtest_map->{private}{$fh};
	$iosel_obj->remove($ipc_sock);

	# Cleanly shutdown socket.
	$ipc_sock->shutdown(0);
	close($ipc_sock);

	# Monitor alarm should not have finished on it's own.
	$Trace_obj->warn(msg => "Monitor alarm finished unexpectedly, " . 
			 "shutting down");
	$self->_shutdown;
	return;
    }
    
    return 1;
} # end sub _iomon_msg_handler

1;


# Users of the EventMgr class do not instantiate EventSequence objects
# directly, but rather through the create_sequence call.  Because of
# this, it's included directly in the EventMgr file rather than 
# separately.

package EventMgr::EventSequence;

use strict;

# Module import.
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN HASHREF ARRAYREF CODEREF OBJECT UNDEF);


sub new {
    my ($class, %params) = @_;
    my(
       $self,
       );

    # Since EventSequence::new is only called by EventMgr::create_sequence,
    # which has already validated the parameters, we do not need to validate
    # them again.
    $self = {};
    
    # Is this sequence runnable as a background task (in parallel with
    # other sequences?)  Default is true - caller can override.
    $self->{bg} = exists $params{bg} ? $params{bg} : 1;

    # Other sequence attributes (loop, sleep, etc.)
    $self->{name}       = $params{name};
    $self->{dataref}  	= $params{data_ref};
    $self->{num_loop} 	= exists $params{loop} ? $params{loop} : 1;
    $self->{loop_delay} = exists $params{sleep} ? $params{sleep} : 0;
    $self->{notify_cb}  = exists $params{notify} ? $params{notify} : undef;

    # Each event sequence has it's own run queue of events.
    $self->{run_queue} = [];

    # Might have post events (events that run after each std event) as well.
    $self->{post_queue} = [];

    # Internal attributes used to manage the run queue.
    $self->{runnable}        = 1;  # Initially, we can run an event.
    $self->{cur_loop}        = 1;
    $self->{cur_event}       = 0;  # Always start with first event in queue.
    $self->{ev_start_time}   = 0;
    $self->{ev_end_time}     = 0;
    $self->{ev_duration}     = 0;  # Duration for most recent completed event.
    $self->{seq_complete}    = 0;  # Are we there yet?
    $self->{forced_shutdown} = 0;  # If true, runnable an seq_complete can't
                                   # be changed.
    bless $self, $class;

    return $self;
} # end sub new


sub add_event {
    my $self = shift;
    my %params = 
	validate_mixed(@_,
		       [],
		       {
			   name    => { type => SCALAR,   optional => 0 },
			   code    => { type => CODEREF | SCALAR, 
					optional  => 0 },
			   type    => { type => SCALAR,   optional => 1 },
			   clients => { type => ARRAYREF, optional => 1 },
		       });

    my(
       $ev_obj,
       );

    $ev_obj = EventMgr::Event->new(%params);

    if ($ev_obj->{type} eq EventMgr::STANDARD_EVENT) {
	push @{$self->{run_queue}}, $ev_obj;
    }
    elsif ($ev_obj->{type} eq EventMgr::POST_EVENT) {
	push @{$self->{post_queue}}, $ev_obj;
    }
    else {
	# Weird event type specified.
	die "Invalid event type specified: $ev_obj->{type}";
    }

    return 1;
} # end sub add_event


sub add_delay {
    my($self, $delay) = @_;
    my(
       $ev_obj,
       );
    
    unless ($delay) {
	# Change to better error handling?
	die "add_delay: delay (in seconds) must be specified and > 0";
    }

    $ev_obj = EventMgr::Event->new(name  => 'delay event',
				   type  => EventMgr::DELAY_EVENT,
				   delay => $delay,
				   code  =>
				   \&EventMgr::EventSequence::_do_delay);
    push @{$self->{run_queue}}, $ev_obj;

    return 1;
} # end sub add_delay


sub is_background {
    my $self = shift;

    return $self->{bg};
} # end sub is_background


sub runnable {
    my $self = shift;
    
    return $self->{runnable};
} # end sub runnable


sub set_runnable {
    my $self = shift;

    # Don't allow it to be modified if we're in a forced shutdown.
    unless ($self->{forced_shutdown}) {
	$self->{runnable} = shift;
    }

    return 1;
} # end sub set_runnable


sub set_forcibly_shutdown {
    my $self = shift;

    $self->{forced_shutdown} = 1;

    return 1;
} # end sub set_forcibly_shutdown


sub get_cur_event {
    my $self = shift;
    
    return $self->{run_queue}[$self->{cur_event}];
} # end sub get_cur_event


sub get_cur_event_id {
    my $self = shift;
    my(
       $event_obj,
       );

    $event_obj = $self->{run_queue}[$self->{cur_event}];

    return $event_obj->get_id;
} # end sub get_cur_event_id


sub get_num_loop {
    my $self = shift;

    return $self->{num_loop};
} # end sub get_num_loop


sub get_cur_loop {
    my $self = shift;

    return $self->{cur_loop};
} # end sub get_cur_loop


sub get_id {
    my $self = shift;

    return $self->{name};
} # end sub get_id


sub get_duration {
    my $self = shift;

    # This is the duration of the most recently run event.
    return $self->{ev_duration};
} # end sub get_duration


sub get_dataref {
    my $self = shift;
    
    return $self->{dataref};
} # end sub get_dataref


sub set_dataref {
    my ($self, $dataref) = @_;

    $self->{dataref} = $dataref;

    return 1;
} # end sub set_dataref


sub get_runnable_event {
    my $self = shift;

    return $self->{run_queue}[$self->{cur_event}];
} # end sub get_runnable_event


sub get_notify_cb {
    my $self = shift;
    
    return $self->{notify_cb};
} # end sub get_notify_cb


sub finalize {
    my $self = shift;
    my(
       $sleep_event,
       @run_q,
       );

    # Intersperse post events.
    if (@{$self->{post_queue}}) {
	@run_q = @{$self->{run_queue}};
	$self->{run_queue} = [];

	# Caller can specify more than 1 post event.  For every standard
	# event specified, add all post_events to the run queue.
	foreach my $event (@run_q) {
	    push @{$self->{run_queue}}, $event;
	    
	    foreach my $post_event (@{$self->{post_queue}}) {
		push @{$self->{run_queue}}, $post_event;
	    }
	}
    }

    # Add loop delay to end of run queue as a sleep event.
    if ($self->{loop_delay}) {
	$self->add_delay($self->{loop_delay});
    }	

    return 1;
} # end sub finalize


sub event_started {
    my $self = shift;

    # Lock down this sequence until the currently running event is
    # complete.
    $self->{runnable} = 0;
    $self->{ev_start_time} = time;

    return 1;
} # end sub event_started


sub event_completed {
    my $self = shift;

    $self->{ev_end_time} = time;
    $self->{ev_duration} = $self->{ev_end_time} - $self->{ev_start_time};

    # Unlock sequence.
    if ($self->{forced_shutdown}) {
	$self->{runnable} = 0;  # Just in case.
	return;
    }
    else {
	$self->{runnable} = 1;
    }

    $self->{cur_event}++;
    if ($self->{cur_event} > $#{$self->{run_queue}}) {
	$self->{cur_event} = 0;   # Back to the first event.
	$self->{cur_loop}++;
	if ($self->{cur_loop} > $self->{num_loop}) {
	    # This sequence is done.
	    $self->{runnable} = 0;
	    $self->{seq_complete} = 1;
	}
    }

    return 1;
} # end sub event_completed


sub sequence_complete {
    my $self = shift;

    return $self->{seq_complete};
} # end sub sequence_complete


sub set_sequence_complete {
    my($self, $is_complete) = @_;
    
    # Don't allow it to be set if we're in a forced shutdown.
    unless ($self->{forced_shutdown}) {
	$self->{seq_complete} = $is_complete;
    }
    
    return 1;
} # end sub set_sequence_complete


sub _do_delay {
    my($delay) = @_;

    # Sleep specified amount of time and return pass result object.
    # This code will always be called as a subtest.
    select(undef, undef, undef, $delay);

    Tharn::addresult(type => "PASS",
		     "Delay of $delay seconds successful");

    return 1;
} # end sub _do_delay

1;


# Users of the EventMgr class do not instantiate Event objects.  It
# is used only by EventMgr::EventSequence to add events to and
# event sequence.
# Because of this, it's included directly in the EventMgr file rather
# than separately.

package EventMgr::Event;

use strict;

sub new {
    my($class, %params) = @_;
    my(
       $self,
       );

    $self = {};
    # Since Event::new is only called by EventMgr::EventSequence::add_event,
    # which has already validated the parameters, we do not need to validate
    # them again.

    $self->{delay} = undef;
    $self->{code}  = $params{code};  # Code ref or thpl name.
    $self->{name}  = $params{name};

    # By default, it's a standard event (as opposed to a POST_EVENT);
    $self->{type} =
	exists $params{type} ? $params{type} : EventMgr::STANDARD_EVENT;

    # If clients were specified for this event, it is a remote
    # event to be run on all specified clients.  Otherwise, it's
    # a local event.
    if (exists $params{clients}) {
	$self->{remote}  = 1;
	$self->{clients} = $params{clients};
    }
    else {
	$self->{remote}  = 0;
	$self->{clients} = undef;
    }

    # If this is a delay event, caller must specify a delay amount.
    if ($self->{type} eq EventMgr::DELAY_EVENT) {
	unless (exists $params{delay}) {
	    die "Must specify delay amount when creating delay event";
	}

	$self->{delay} = $params{delay};
    }

    bless $self, $class;

    return $self;
} # end sub new


sub is_remote {
    my $self = shift;

    return $self->{remote};
} # end sub is_remote


sub get_id {
    my $self = shift;
    
    return $self->{name};
} # end sub get_id


sub get_type {
    my $self = shift;
    
    return $self->{type};
} # end sub get_id


sub get_delay {
    my $self = shift;

    return $self->{delay};
} # end sub get_delay


sub get_code {
    my $self = shift;

    return $self->{code};
} # end sub get_code


sub get_clients {
    my $self = shift;

    return $self->{clients};
} # end sub get_clients

1;

__END__

#################################
#  Pod
#################################

=pod

=head1 NAME

EventMgr

=head1 SYNOPSIS

 use EventMgr qw(POST_EVENT);
 use MyCodePackage;

 # Create objects you want to pass to eventmgr (Server, Trace, etc.)
 $Server = new Server;
 $Trace = new TracePkg;

 # Create EventMgr object, passing it other objects that you want it to
 # pass to every event in a sequence.
 $eventmgr = new EventMgr(args => [ $Server, $Trace ]);

 # Event sequences can use a data structure (reference) to pass data
 # between events.  Data ref is a freeform data structure - content is
 # up to you, the dataref shown below is just an example.
 $dataref = {
     vols    => [],
     node    => <filer>,
     timeout => 180,
     count   => 0,
 };

 # Create an event sequence.
 $sequence_obj = $eventmgr->create_sequence(name     => "sequence 1",
					    data_ref => $dataref,
					    loop     => 5,
					    sleep    => 2);

 # Add events to the sequence.
 $sequence_obj->add_event(name => "code a",
			  \&MyCodePackage::first_event);
 $sequence_obj->add_event(name => "code b",
			  \&MyCodePackage::second_event);

 # Add an event that will run after every other event in the sequence.
 $sequence_obj->add_event(type => POST_EVENT,
			  name => 'sysstat',
			  code => \&my_sysstat);  # Local subroutine.

 # Run the event loop.
 $rc = $eventmgr->event_loop;

 $pass_count = $eventmgr->get_pass_count;
 if ($rc) {
     # All events passed successfully.
     # log pass count.
 }
 else {
     # There was a failure in an event or monitor.
     @fail = $eventmgr->get_fail_results;
     @warn = $eventmgr->get_warn_results;
     @other = $eventmgr->get_other_results;
     # log failures, etc.
 }

=head1 DESCRIPTION

EventMgr provides the capability to create sequences of events to be run
in a specified order.  Events are code references (local subroutines or
references to code in an included package) or thpls that are meant to
operate independently.

Any number of event sequences can be constructed and run in parallel, since all
the events are executed as subtests.  Caller can specify that a given sequence
should be run in a loop any number of times.  Loop count is specified per
sequence and can be different across sequences.

EventMgr provides a mechanism to pass data between events in a sequence
using NATE::IPC.  Data can be retrieved and udpated within event code and
then sent back to the EventMgr to be passed on to the next event in
the sequence.

Additionally, caller can specify a monitor object for the EventMgr to
periodically poll (an I/O load object, for example).  In this mode, the
continued execution of events is dependent on the monitored object reporting
success each polling cycle.  If the monitored object reports a failure,
the sequencing engine will be shut down, results gathered and control
passed back to the caller.

In short, the EventMgr is ideal for situations where the caller needs to
execute a number of operations (possibly in parallel) with background load
(e.g. client I/O) applied, and wants to have the operations stop if
there is a failure (corruption, for example) in the background load.
EventMgr does not provide a mechanism for specifying and starting the
background load, but that can be easily generated using the B<ClientMgr>
library.

=head1 EXPORTED CONSTANTS (@EXPORT_OK)

event type: B<STANDARD_EVENT>, B<POST_EVENT>, B<DELAY_EVENT>

=head1 METHODS

=head2 new

Create an EventMgr object.

=over

=item Synopsis

 $eventmgr_obj = EventMgr->new(%args);

=item Arguments

=over

=item clients => I<$arrayref>

(Optional)

A reference to an array of client names.  Can specify a list of clients
on the I<new> call to be used for any remote events.

=item args => I<$arrayref>

(Optional)

A reference to an array of arguments to be passed to each event as it is
executed.  Arguments can be scalars, references, objects, etc.

=back

=item Returns

This method returns an EventMgr object.

=back

=head2 create_sequence

Create a new sequence container, to which events can be added.

=over

=item Synopsis

 $sequence_obj = $eventmgr_obj->create_sequence(%args)

=item Arguments

=over 4

=item name => I<string>

(Required)

Specify a unique identifier for the sequence.  Will be used for logging and
as a parameter in the notify callback.

=item data_ref => I<hashref or arrayref>

(Required)

This provides a mechanism for passing data between events.  It can be
specified as either a reference to a hash or an array.  The contents are
freeform and completely up to the user.  Events should use the DataStore
library functions B<get_data> and B<write_data> to retrieve the data and
optionally write it back, if contents of the data structure are modified
by the event.

=item bg => 1 | 0

(Optional)

Specifies whether the events in the sequence should be run in the foreground
or the background.  Default value is 1, meaning that all events will be
run in the background.  If foreground is specified (bg => 0), then any
subsequent sequences will be blocked until this sequence has completed.

=item loop => I<integer>

(Optional)

Specifies how many times the entire sequence should run.  Default is 1.

=item sleep => I<seconds>

(Optional)

Specifies a delay that will occur at the end of the sequence.  Probably
only useful if sequence is looping.

=item notify => I<code reference>

(Optional)

Provides a mechanism for the caller to be notified upon event completion.
After each event is complete, the specified code ref will be called with
the following parameters (in the order shown below):

=over 2

=item *

event_id (name specified during event object creation)

=item *

sequence_id (name specified durin sequence object creation)

=item *

current loop number for this sequence

=item *

duration (seconds) of the event that just finished

=item *

dataref associated with this sequence

=back

=back

=item Returns

This method returns an I<Event Sequence> object, from which events can
be created.

=back

=head2 add_event

Adds events to the I<Event Sequence> object.  Events will be executed in the
order in which they are added to the sequence.

=over

=item Synopsis

 $sequence_obj->add_event(%args);

=item Arguments

=over

=item name => I<string>

(Required)

Unique identifier for this event.  Spaces are allowed.  Used for logging and
as a parameter to the notification callback.

=item code => I<coderef or thpl>

(Required)

This is the event code that will be executed.  Can be a local coderef
(reference to local subroutine), package coderef (collection of similar
events), or the name of a thpl.

=item type => <STANDARD_EVENT | POST_EVENT>

(Optional)

By default, events are I<standard>, meaning they will run the specified code
in the order in which they are added to the sequence.  If an event is
added with the I<POST_EVENT> type, it will be run after every other event
in the sequence.  This is useful if there is a bit of code that the caller
would like to run as a verification routine after every other event.  The
I<POST_EVENT> is provided as a convenience.

=item clients => I<$arrayref>

(Optional)

List of clients that this event should operate on.  This is only valid
if the event code specified is a thpl.

Not operational at this time (i.e., the event manager does not currently
farm off thpl events to be run on remote clients.)

=back

=item Returns

1 on successful creation/addition of the event.  Will die if params are
not specified properly, or an invalid event type is specified.

=item Notes

This method is called from the I<Event Sequence> object created by the
I<create_sequence> call, not from the event manager object.

=back

=head2 add_monitor

Tell the EventMgr that the specified object should be monitored on a periodic
basis.  Any failures reported in the monitored object will cause the
event sequences to shut down.  Any events that are in-flight are killed.

=over

=item Synopsis

 $eventmgr_obj->add_monitor(I<$monitor_obj>);

=item Arguments

=over

=item $monitor_obj

(Required)

Object that can be monitored (provides an API that has a monitor or status
method).  Currently only supports ClientMgr (I/O load) type objects.

=back

=item Returns

Returns 1 on success.  Will die if no monitor object is specified.

=back

=head2 event_loop

Loops over all events in all sequences (in order specified) until all sequences
have completed, or an event or monitored object fails.

=over

=item Synopsis

 $rc = $eventmgr_obj->event_loop;

B<or>

 if ($eventmgr_obj->event_loop) {
     # Returns 1 on success.
     $eventmgr_obj->get_pass_count;
     # log pass count
 }
 else {
     # Returns 0 if there were failures.
     @fail = $eventmgr_obj->get_fail_results;
     @warn = $eventmgr_obj->get_warn_results;
     @other = $eventmgr_obj->get_other_results;
     $pass = $eventmgr_obj->get_pass_count;
     # log failures, warn, etc.
 }

=item Arguments

None.

=item Returns

Returns 1 if all sequences completed successfully and all monitors ran
without error.  Returns 0 if any events failed, or if any of the monitors
reported an error.  Using other methods specified below, the caller can
get the details of failed results, warn results, other results and the
pass count.

=back

=head2 get_fail_results

Return any NATE::Result::Fail objects that occurred in the event loop.

=over

=item Synopsis

 @fail = $eventmgr_obj->get_fail_results

=item Arguments

None.

=item Returns

Array of NATE::Result::Fail objects.  Empty array is returned if all events
completed successfully.

=back

=head2 get_warn_results

Return any NATE::Result::Warn objects that occurred in the event loop.

=over

=item Synopsis

 @warn = $eventmgr_obj->get_warn_results;

=item Arguments

None.

=item Returns

Array of NATE::Result::FAIL objects.  Empty array is returned if all events
completed without warnings.

=back

=head2 get_pass_count

Return a count of events that passed.

=over

=item Synopsis

 $pass_count = $eventmgr->get_pass_count;

=item Arguments

None.

=item Returns

Integer representing the number of events that completed successfully in
the event loop.

=back

=head2 get_other_results

Return any NATE::Result::Other objects that occurred in the event loop.

=over

=item Synopsis

 @other = $eventmgr_obj->get_fail_results;

=item Arguments

None.

=item Returns

Array of Nate::Result::Other objects.  Empty array is returned if there are
no Nate::Result::Other results.

=back

=head1 AUTHOR/MAINTAINER

Kevin Lambright (kevinl@netapp.com)

=cut
