#!/bin/sh /usr/software/bin/perl 

#
#
package ExecTC::Object;

use strict;
use Data::Dumper;

use TracePkg;
use ExecTC::Step;

my $t = TracePkg->new();



sub new {

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    #  a hash.
    my ($self, %arg) = @_;

    $t->entry( msg=>"Initializing Object: $arg{Id}, cmds to host: $arg{Host}" );

    # Fill the member vars for this instance
    my $object = {
	
	_id     => $arg{Id},
	_cmdDir => $arg{CmdDir},
	_keys   => $arg{Keys},
	_fields => $arg{Fields},
	_preDel => $arg{PreDelete},
	_deleteAction => undef,
	_preDelAction => undef,
	_getCurrentAction => undef,
	_state  => undef,
	_testCreated  => undef,
	_host   => $arg{Host},
	_deltaStack => undef,
    };

    # This tags the hash pointed to by $object as a 
    #  type of Table
    bless $object, $self;

    $t->exit( msg=>"Object defined: $arg{Id}\n" );
    # And give the caller the referece to the bless object.
    return $object;
}


sub getId {
    my ($self, %arg) = @_;
    return $self->{_id};
}


sub setState {
    my ($self, %arg) = @_;
    $self->{_state} = $arg{state};
}

sub getState {
    my ($self, %arg) = @_;
    return $self->{_state};
}

sub setTestCreated {
    my ($self, %arg) = @_;
    $self->{_testCreated} = 1;    
    return;
}

sub wasTestCreated {
    my ($self, %arg) = @_;
    return $self->{_testCreated};
}

sub procPreDelete {
    my ($self, %arg) = @_;

    if ( !defined $self->{_preDel} ) {
	$t->trace(msg=>"No PreDelete section defined for: $self->{_id}");
	return;
    }

    # Section defined in test TestCase, so set it up.
    $t->entry(msg=>"Building pre-delete cmd for: $self->{_id}\n");

    my ($cmd, $key, $value);

    $self->{_preDelAction} = 
	ExecTC::Action->new( Cmd    => $self->{_preDel}{Cmd},
			     RefId  => $self->{_id},
			     Fields => $self->{_preDel}{Fields},
			     Verify => $self->{_preDel}{Verify},
			     TC     => $arg{TC},
			     Host   => $self->{_host},
			     );    

    $t->exit(msg=>"Action built: " . $self->{_preDelAction}->getFullCmd() .
	     ", meth: " . $self->{_preDelAction}{_cliMethod});
}


sub buildModify {
    my ($self, %arg) = @_;

    $t->entry(msg=>"Building modify cmd for: $self->{_id}\n");

    my ($cmd, $key, $value);

    $self->{_modifyAction} = 
	ExecTC::Action->new( Cmd    => "modify",
			     RefId  => $self->{_id},
			     Verify => { State => "exists" },
			     TC     => $arg{TC},
			     Host   => $self->{_host},
			     );    

    $t->exit(msg=>"Action built: " . $self->{_modifyAction}->getFullCmd() . 
	     ", meth: " . $self->{_modifyAction}{_cliMethod});
}


sub getModifyAction {
    my ($self, %arg) = @_;

    if ( defined $self->{_modifyAction} ) {
	return $self->{_modifyAction};
    } else {
	$t->debug(msg=>"Object has no modify action: $self->{_id}");
	return undef;
    }
}


sub buildGetCurrent {
    my ($self, %arg) = @_;

    $t->entry(msg=>"Building the Action to get the current state for: $self->{_id}\n");

    my ($cmd, $key, $value);

    $self->{_getCurrentAction} = 
	ExecTC::Action->new( Cmd    => "show",
			     RefId  => $self->{_id},
			     TC     => $arg{TC},
			     Host   => $self->{_host},
			     );    

    $t->exit(msg=>"Action built: " . $self->{_getCurrentAction}->getFullCmd());
}


sub getDelCmd {
    my ($self, %arg) = @_;
    return $self->{_deleteCmd};
}


sub getCmdDir {
    my ($self, %arg) = @_;
    return $self->{_cmdDir};
}


sub getKeys {
    my ($self, %arg) = @_;
    return $self->{_keys};
}


sub getFields {
    my ($self, %arg) = @_;
    if ( exists $self->{_fields} ) {
	return $self->{_fields};
    }
    return undef;
}


# Passed a the output from a show command, -showallfields true  and
# -showseparator ",".  Parse it into a _fields
# hash, then remove and hash elements that are already represented 
# as _keys (so we don't get dup fields when commands are built).
sub hydrate {
    my ($self, %arg) = @_;

    $t->entry( msg=>"Updating object: $self->{_id}\n" );

    my $text = $arg{text};
    my $key;

    $self->{_fields} = _parseFields( $text );

    # Remove the Keys from the Fields
    for $key ( keys %{ $self->{_keys} } ) {
	$t->debug( msg=>"Deleting field: $key\n" );
	delete $self->{_fields}{$key};
    }
    $t->exit( msg=>"Updated object: $self->{_id}\n" );
}


sub _parseFields {
    my $output = shift;

    # field,field1,field2,\n<descriptions>,\nval1,val2,val3,
    my (@lines, $key, $value, %fields);
    my (@args, @values, @headers);

    $output =~ s/\r//g;
    @lines = split /\n/, $output; 
    @args = split /X;X/, $lines[1];
    @headers = split /X;X/, $lines[2];
    @values = split /X;X/, $lines[3];

    
    for $key ( 0 .. $#args ) {
	$fields{$args[$key]} = $values[$key];
    }
    return \%fields;    
}


sub hasKeys {
    my ($self, %arg) = @_;

    if ( defined $self->{_keys} ) {
	return 1;
    } else {
	$t->debug(msg=>"Object: $self->{_id} has no keys");
	return undef;
    }
}


sub setDelAction {    
    my ($self, %arg) = @_;

    $self->{_deleteAction} = $arg{action};
}    


sub getDelAction {    
    my ($self, %arg) = @_;

    if ( defined $self->{_deleteAction} ) {
	return $self->{_deleteAction};
    } else {
	$t->debug(msg=>"Object has no delete action: $self->{_id}");
	return undef;
    }
}    


sub getPreDelAction {    
    my ($self, %arg) = @_;

    return $self->{_preDelAction};
}    


sub getCurrentAction {    
    my ($self, %arg) = @_;

    return $self->{_getCurrentAction};
}    


sub pushDelta {
    my ($self, %arg) = @_;
    my @deltaStack = $self->{_deltaStack};
    push @{ $self->{_deltaStack} }, $arg{delta};
    $t->trace(msg=>"Delta pushed on to stack, now: " . Dumper($self->{_deltaStack}));
}
    

sub popDelta {
    my ($self) = @_;
    if ( defined $self->{_deltaStack} ) {
	return pop @{ $self->{_deltaStack} };
    }
    $t->trace(msg=>"Nothing to pop, array empty");
    return undef;
}


# Save the object's state based on the passed set of fields.  If there
# are none passed used the locally defined fields.  If there are none, this is
# basically a nop.
# Return the fields that where actually saved.
sub saveState {    
    my ($self, %arg) = @_;

    my ($currFields, $requestFields);

    $t->entry(msg=>"Asked to save the state for fields: " . 
	      join(",", keys %{$arg{fields}}) );

    if ( scalar keys %{$arg{fields}} == 0 ) {
	$t->debug(msg=>"No fields passed, using local fields");
	
	if ( scalar keys %{$self->{_fields}} != 0 ) {
	    $t->debug(msg=>"Using locally defined fields");
	    $requestFields = $self->{_fields};
	} else {
	    $t->exit(msg=>"No fields defined locally, skipping save");
	    return undef;
	}
    } else {
	$requestFields = $arg{fields};
    }

    # Get the state of the passed fields
    $currFields = $self->getCurrent( fields => $requestFields);

    # Prune the list of fields.  Don't save the state of fields
    # that match the destination state.  We won't be trying
    # to modify fields to the same state.
    foreach my $key ( keys %{$requestFields} ) {
	if ( $requestFields->{$key} eq $currFields->{$key} ) {
	    delete $currFields->{$key};
	    $t->debug(msg=>"Deleted field: $key.  " .
		      "It's already at value: $requestFields->{$key}");
	} 
    }    
    # Save the state for later cleanup, if needed
    if ( scalar keys %{$currFields} > 0 ) {
	$self->pushDelta( delta => $currFields );
    }

    # Save the last exist/modify state
    $self->{_oldState} = $arg{state};

    $t->exit();
    return $currFields;
}


# For the passed fields, return the current values.
sub getCurrent {
    my ($self, %arg) = @_;
    my ($outHash, @outHash, %outHash);

    $t->entry(msg=>"Getting the current values for fields: " . 
	      join(",", keys %{$arg{fields}}) );

    my $requestFields = $arg{fields};

    # First get all the current field values
    $self->{_getCurrentAction}->execute();
    $outHash = $self->{_getCurrentAction}->getOutHash();
    @outHash = @{$outHash};
    %outHash = %{$outHash[0]};
    for my $key ( keys %outHash ) {
	# Delete all fields that weren't asked for
	if ( !defined $arg{fields}->{$key} ) {
	    delete $outHash{$key};
	}
    }

    # For each requested field, build a matching hash of the current values.
    $t->exit(msg=>"Got the current values for fields: " . 
	     join(",", keys %outHash) );
    return \%outHash;
}


1;
