#Hey emacs, this is basically -*- perl -*-


#
#
package ExecTC::Step;

use strict;
use Getopt::Long;
use Data::Dumper;

use TracePkg;
use ExecTC::Step;
use ExecTC::Action;

my $t = TracePkg->new();


sub new {

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    #  a hash.
    my ($self, %arg) = @_;

    $t->entry( msg=>"Initializing Step: $arg{step}{Name}, on host: $arg{host}\n" );

    # Fill the member vars for this instance
    my $object = {
	
	_name     => $arg{step}{Name},
	_desc     => $arg{step}{Desc},
	_parent   => $arg{tc},
	_defObjects => $arg{step}{Objects},
	_defAction => $arg{step}{Action},
	_defOutput => $arg{step}{Output},
	_defCleanUp => $arg{step}{CleanUp},
	_defHost => $arg{step}{Host},
	_host    => $arg{host},
	_action  => undef,
	_output  => undef,
	_errMsg  => "",
    };

    $t->debug(msg=>"name: $object->{_name}, desc: $object->{_desc}\n");
    # This tags the hash pointed to by $object as a 
    #  type of Table
    bless $object, $self;

    # Objects defined in this step (add to test case)
    if ( $object->{_defObjects} ) {
	_procObjs( $object );
    }
    # Action to be performed in this step
    if ( $object->{_defAction} ) {
	_procAction( $object );
    }
    # New objects retrieved from the system in this step (add to tc)
    if ( $object->{_defOutput} ) {
	_procOutput( $object );
    }
    # If this a CleanUp step, gen the Action that walks the objects
    # in the order created.
    if ( defined $object->{_defCleanUp} ) {
	_procCleanUp( $object );
    }


    $t->exit( msg=>"Step init complete: $arg{step}{Name}\n" );
    # And give the caller the referece to the bless object.
    return $object;
}


sub _procObjs {
    $t->entry( msg=>"Defining objects\n");

    my ($self, %arg) = @_;
    my ($objIdx, @objs, $objsRef);
    my $tc = $self->{_parent};

    # Process all the objects defined in this step
    $objsRef = $self->{_defObjects};
    @objs = @$objsRef;

    for $objIdx ( 0 .. $#objs) {
	$t->debug( msg=>"Working on: $objs[$objIdx]{Id}\n");
	$tc->addObject( objDef => $objs[$objIdx], host => $self->{_host} );	
    }
    $t->exit( msg=>"Objects defined\n");
}


sub _procAction {
    my ($self, %arg) = @_;
    $t->entry( msg=>"Defining action: $self->{_defAction}{Cmd}\n");

    # Add the new action objet for this step.
    $self->{_action} = 
	ExecTC::Action->new( Cmd    => $self->{_defAction}{Cmd},
			     RefId  => $self->{_defAction}{RefId},
			     Host   => $self->{_host},
			     Fields => $self->{_defAction}{Fields},
			     Verify => $self->{_defAction}{Verify},
			     Delete => $self->{_defAction}{Delete},
			     OnError => $self->{_defAction}{OnError},
			     TimeOut => $self->{_defAction}{TimeOut},
			     Revert  => $self->{_defAction}{Revert},
			     TC     => $self->{_parent},
			     );    
    
    my $msg = "Action defined: " . $self->{_action}->getName() . "\n";
    $t->exit( msg=>$msg);
}


sub _procOutput {
# "Output" is really just another type of Action.  Some command is needed 
# to fill in fields for some object.  That object is is based on the 
# keys as defined by some refObj.  
# The created Output Objects share the same structure as 
# objects defined by user input.  The difference is that they are defined 
# by retrieving info from the system.
    my ($self, %arg) = @_;

    $t->entry( msg=>"Add output object: $self->{_defOutput}{Id}\n");

    # Create and Action to fill in this Output's Object
    $self->{_output} = 
	ExecTC::Action->new( Cmd    => $self->{_defOutput}{Cmd},
			     RefId  => $self->{_defOutput}{RefId},
			     TC     => $self->{_parent},
			     );    
    # Now add the Object to the test case
    $self->{_parent}->addObject( 
				 objDef => {
				     Id => $self->{_defOutput}{Id},
				     CmdDir => $self->{_parent}->getObject($self->{_defOutput}{RefId})->getCmdDir(),
				     Keys => $self->{_parent}->getObject($self->{_defOutput}{RefId})->getKeys(),
				 } );
    $t->exit( msg=>"Output object added: $self->{_defOutput}{Id}\n");
}


sub _procCleanUp {
# Generate an Action for this step.  
    my ($self, %arg) = @_;

    $t->entry( msg=>"Add CleanUp Action: $self->{_defCleanUp}{Id}\n");
    
    # Create and Action to fill in this Output's Object
    $self->{_action} = 
	ExecTC::Action->new( CleanUp => $self->{_defCleanUp},
			     TC      => $self->{_parent},
			     );    
    $t->exit( msg=>"Output object added: $self->{_defOutput}{Id}\n");
}


sub getName {
    my ($self, %arg) = @_;
    return $self->{_name};
}


sub execute {
    my ($self, %arg) = @_;
    my $tc = $self->{_parent};
    my $action = $self->{_action};
    my $output = $self->{_output};

    my $result = "PASSED";
    my ($fields, $tmp);

    $t->entry(msg=>"Running step: $self->{_name}\n");

    my $conn = $arg{conn};

  EXEC: {
      # Run this step's Action
      if ( defined $self->{_action} ) {
	  $t->debug(msg=>"Processing action\n");
	  $result = $action->execute( conn => $arg{conn} );
	  
	  if ( $result eq "FAILED" ) {
	      $self->{_errMsg} = $action->getOutput();
	  } 
      }  
      if ( $result eq "FAILED" ) { last EXEC; }

      # Run this step's Output
      if ( defined $self->{_output} ) {
	  $t->debug(msg=>"Processing output\n");
	  $result = $output->execute( conn => $arg{conn} );
	  
	  if ( $result eq "FAILED" ) { last EXEC; }
    
	  # Update the output object with the command's key and values.
	  $tc->getObject($output->getRefId())->hydrate( text => $action->getOutput() );
	  
      }
  }     

    $t->exit(msg=>"Step $self->{_name}: $result\n");
    return $result;
}


sub isCleanUp {
   my ($self, %arg) = @_;
   if ( defined $self->{_defCleanUp} ) {
       return 1;
   } else {
       return undef;
   }
}


sub getErrMsg {
   my ($self, %arg) = @_;
   return $self->{_errMsg};
}


1;
