#Hey emacs, this is basically -*- perl -*-

#
#
package ExecTC::TestCase;

use strict;
use Getopt::Long;

use TracePkg;
use ExecTC::Step;
use ExecTC::Object;

my $t = TracePkg->new();


sub new {

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    #  a hash.
    my ($self, %arg) = @_;

    $t->entry( msg=>"Create new TestCase: $arg{name}\n" );

    # Fill the member vars for this instance
    my $object = {
	
	_name  => $arg{name},
	_desc  => $arg{desc},
	_hosts => $arg{hosts},
	_conns => $arg{connections},
	_steps => undef,
	_objects => undef,
	_createdObjsOrder => undef,
	_engine => $arg{engine},
    };

    # This tags the hash pointed to by $object as a 
    #  type of Table
    bless $object, $self;

    $t->exit( msg=>"TestCase created\n" );
    # And give the caller the referece to the bless object.
    return $object;
}


sub addSteps {
    $t->entry( msg=>"Adding steps\n" );
    my ($self, %arg) = @_;

    my $stepsRef = $arg{steps};
    my @steps = @$stepsRef;

    my ($stepIdx, $server);

    # Steps are defined as an array of hashes.  Step through the 
    # elements of the array creating a new step and adding it to the
    # _steps array in this TestCase.  Steps will always be processed
    # sequencially
    for $stepIdx ( 0 .. $#steps ) {
	$t->debug(msg=>"Adding: $steps[$stepIdx]{Name}\n");

	# If they defined a host for this step, use it.
	if ( defined $steps[$stepIdx]{Host} ) {
	    $t->debug(msg=>"Host defined in step");
	    $server = $steps[$stepIdx]{Host};
	} else {
	    # Else, use some host defined at the TestCase level
	    $t->debug(msg=>"Host not defined in step");
	    $server = $self->getSomeHost();
	}
	$t->debug(msg=>"Actions to: $server");

	push( @{ $self->{_steps} }, 
	      ExecTC::Step->new( step => $steps[$stepIdx],
				 tc => $self,
				 host => $server) );
    }

    $t->exit( msg=>"Steps added\n" );

    return;
}


sub getStep {
    my ($self, $step) = @_;

    return $self->{_steps}[$step];
}


sub getStepCount {
    my ($self, %arg) = @_;
    return $#{$self->{_steps}}+1;
}


sub addObject {
    my ($self, %arg) = @_;
    my $objDef = $arg{objDef};
    my $id = $objDef->{Id};
    my $object = undef;

    $t->entry( msg=>"Adding object: $id\n" );

    $object = 
	ExecTC::Object->new( Id     => $objDef->{Id},
			     CmdDir => $objDef->{CmdDir},
			     Keys   => $objDef->{Keys},
			     Fields => $objDef->{Fields},
			     PreDelete => $objDef->{PreDelete},
			     Host   => $arg{host},
			     );
    $self->{_objects}{$id} = $object;
    # Create the default delete Action for this object, as long
    # as we have some keys
    if ( defined( $object->hasKeys() )) {
	$t->trace( msg=>"Create delete Action for object: $id");
    	$object->setDelAction( action => 
			       _createDelAction( $self, object => $object ) );
    } else {
	$t->trace(msg=>"Object: $arg{Id} has no keys, thus no delete Action");
    }
    
    # Create a PreDeleteAction for this object (if one is defined).
    $object->procPreDelete( TC => $self );

    # Create a getCurrAction for pre-modify state saves
    $object->buildGetCurrent( TC => $self );
	
    # Create a getCurrAction for pre-modify state saves
    $object->buildModify( TC => $self );
	
    $t->exit( msg=>"Object added\n" );
}


sub _createDelAction {
    my ($self, %arg) = @_;
    my $object = $arg{object};

    my ($cmd, $key, $value, $action);

    $t->entry(msg=>"Building delete action for: ". $object->getId());

    # Build Action for deleting this Object
    $action = 
	ExecTC::Action->new( Cmd    => "delete",
			     RefId  => $object->getId(),
			     Verify => {
				 State => "missing",
			     },
			     Host   => $self->getSomeHost(),
			     TC     => $self,
			     );    
    
    $t->exit(msg=>"Delete Action built: " . $action->getFullCmd() );
    return $action;
}


sub getObject {
    my ($self, $id) = @_;
    if ( exists $self->{_objects}{$id} ) {
	return $self->{_objects}{$id};
    } else {
	$t->trace(msg=>"Object $id not found");
	return undef;
    }
}


sub pushCreatedObj {
    my ($self, %arg) = @_;
    
    my @createdObjsOrder = $self->{_createdObjsOrder};
#    push @createdObjsOrder, $arg{object};
    push @{ $self->{_createdObjsOrder} }, $arg{object};
    $t->trace(msg=>"Pushing obj: " . $self->getObject($arg{object})->getId() .
	      " on to objs: @createdObjsOrder");
}


sub popCreatedObj {
    my ($self) = @_;
    
    if ( defined $self->{_createdObjsOrder} ) {
	return pop @{ $self->{_createdObjsOrder} };
    }
    $t->trace(msg=>"Nothing to pop, array empty");
    return undef;
}


sub getCreatedObjs {
    my ($self) = @_;
    
    return @{$self->{_createdObjsOrder}};
}


sub getSomeHost {
    my ($self, %arg) = @_;

    if ( ref($self->{_hosts}) eq "ARRAY" ) {
	my @hostArray = @{$self->{_hosts}};
	return  $hostArray[int rand(@hostArray)];
    } else {
	return $self->{_hosts};
    }
}


1;
