# -*- perl -*-
#  $Id: //depot/prod/ontap/shasta/test/lib/ExportFS.pm#2 $

## @summary Manipulate the data exports of a filer
## @author dl-central-libs
## @cf     false
## @status public
## @pod here
##
## @change 11-08-226 anishv: added write_etc_exports to the nfs tag list



=head1 NAME

ExportFS - Exporting functions for NFS and CIFS

=head1 SYNOPSIS

    use ExportFS qw(:nfs);

    my $h = host($FILER);
    my $conn = connect("#rsh");

    write_etc_exports ($conn, $path, $options);
    export_path ($conn, $path, %options);
    unexport_path ($conn, $path, %options);
    list_exports ($conn);


    use ExportFS qw(:cifs);

    my $h = host($FILER);
    my $conn = connect("#rsh");

    $ret = share_exists($conn, $share);
    $ret = share_add($conn, $share, $path, %options);
    $ret = share_delete ($conn, $sharename, %options);


=head1 DESCRIPTION

This package defines routines which manipulate the set of
data paths which the filer makes available to clients.
Hence, export/unexport for NFS and share/unshare paths for CIFS.

=head1 FUNCTIONS

Following is a description of the subroutines in this library.

=cut

package ExportFS;

use strict;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
	nfs => [ qw( write_etc_exports export_path unexport_path list_exports) ],
	cifs => [ qw(share_exists share_add share_delete) ]
  );
}

use SayWrap qw(sayEOF);
use Utils qw(make_conn_and_complain);
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate_pos SCALAR BOOLEAN);
use CLI::FileIO qw(rdfile wrfile);
use Carp qw(croak);
use Filer::Version qw(are_we_ontapng);
use filer qw(filer_vfiler_command);
use Admin::License qw(license);

=head2 write_etc_exports ($conn, $path, $options)

Using the given connection, writes out the /etc/exports file,
and overwrites the entry for $path.

Arguments:

=over 4

=item conn

(Required) : Connection Object to filer.

=item path

(Required) : Path to be exported

=item options

(Optional) : If no options are specified, ($options eq undef)
removes the entry for $path instead.
Note, this is different from specifying an empty options string
(options eq ""), which will overwrite the entry for $path
with a blank entry.  Throws an exception on error.

=back

=cut

sub write_etc_exports($$;$) {
  my ($conn, $path, $options) = @_;

  # Read in /etc/exports file.
  my @input = split(/\n/, rdfile($conn, "/etc/exports"));
  die $@ if ($@);

  # Skip our path if it is in the file,
  # since we will overwrite it or remove it.
  my $pathexp = quotemeta($path);
  my @output = grep(!/^$pathexp(\s|\z)/, @input);
  my $data = join("\n", @output) . "\n";

  if (defined $options) {
    # Add in our path (possibly with no options).
    my $optstring = ($options ? "-" . $options : "");
    $data .= sprintf("%s\t%s\n", $path, $optstring);
  }
  # else, we just removed the path.

  # Write out the file.
  die $@ unless wrfile($conn, "/etc/exports", $data);
}

=head2 export_path ($conn, $path, %options)

Exports the given path for NFS clients to use.

Returns sucess or failure, with an error message in $@ on failure.

Note that using this function to set an empty options string
(OPTIONS => "") does not yield the same behavior as unexporting
the path.


Arguments

=over 4

=item conn

(Required) : Connection Object to filer.

=item path

(Required) : path to be exported
    
=item OPTIONS

(Optional) : export options string (default="anon=0")

=item WRITE_ETC

(Optional) : write changes to /etc/exports (default=0)

=back

=cut

sub export_path($$;%) {
  my $conn = make_conn_and_complain(shift);
  my ($path, %param) = validate_mixed(
	@_,
	[ { type => SCALAR } ],
	{
	 OPTIONS => { type => SCALAR, optional => 1, default => "anon=0" },
	 LICENSE_NFS => { type => BOOLEAN, optional => 1, default => 1 },
	 WRITE_ETC => { type => BOOLEAN, optional => 1, default => 0 },
	}
  );

  if (are_we_ontapng($conn)) {
      return 1;
  }
  return eval {
    if ($param{LICENSE_NFS}) {
        license($conn, 'nfs');
    }
    # Export the path.
    my $optstring = ($param{'OPTIONS'} eq "") ? '""' : $param{'OPTIONS'};
    my $cmd = sprintf("exportfs -i -o %s %s", $optstring, $path);
    my $text = sayEOF($conn, $cmd);
    unless ($text eq "") {
      $text =~ s/^exportfs:\s*(\S+)/$1/;
      die "$text\n";
    }

    if ($param{'WRITE_ETC'}) {
      write_etc_exports($conn, $path, $param{'OPTIONS'});
    }

    return 1;
  };
}

=head2 unexport_path ($conn, $path, %options)

Unexports the path on the filer.

Returns success or failure with an error message in $@ on failure.

Arguments

=over 4

=item conn

(Required) : Connection Object to filer.

=item path

(Required) : path to be exported
    
=item EXISTENT

(Optional) : only unexport the path if it is currently exported (default=1)

=item WRITE_ETC

(Optional) : write changes to /etc/exports (default=0)

=back


=cut

sub unexport_path($$;%) {
  my $conn = make_conn_and_complain(shift);
  my ($path, %param) = validate_mixed(
	@_,
	[ { type => SCALAR } ],
	{
	 EXISTENT => { type => BOOLEAN, optional => 1, default => 1 },
	 WRITE_ETC => { type => BOOLEAN, optional => 1, default => 0 }
	}
  );

  return eval {

    if ($param{'EXISTENT'}) {
      # Return successfully if the path is not already exported.
      my $exports = list_exports($conn);
      return 1 unless ($exports->{$path});
    }

    # Unexport the path.
    my $cmd = sprintf("exportfs -u %s", $path);
    my $text = sayEOF($conn, $cmd);
    die "$text\n" unless ($text eq "");

    if ($param{'WRITE_ETC'}) {
      write_etc_exports($conn, $path);
    }

    return 1;
  };
}

=head2 list_exports ($conn)

Returns a hash reference of the exported paths on the given filer
The hash keys are the exported paths, and they hash to their
export options strings.  On failure, returns undef with an error
message in $@.


Arguments

=over 4

=item conn

(Required) : Connection Object to filer.

=back

=cut

sub list_exports($) {
  my $conn = make_conn_and_complain(shift);

  return eval {
    my $text = sayEOF($conn, "exportfs");

    my %exports = ();
    foreach my $line ( split(/\n/, $text) ) {
      if ($line =~ /^(\S+)\s+\-(\S+)\s*$/) {
	my ($path, $opts) = ($1, $2);
	$exports{$path} = $opts;
      } elsif ($line =~ /^(\S+)\s*$/) {
	# Ignore lines that explicitly "unexport a path",
	# i.e. export a path with no options.
	next;
      } else {
	# Unexpected/unparsable text is an error message.
	die "$line\n";
      }
    }

    return \%exports;
  };
}




=head2 share_exists ($conn, $share)

Checks presence or absence of CIFS share on filer

Returns

=over 4

=item 1
if share exists

=item 0
if share does not exist

=item dies
with $@ if error occured (e.g cifs is not setup)

=back


Arguments

=over 4

=item conn

(Required) :  Connection Object to filer.


=item share

(Required) : Name of share

=back

=cut

sub share_exists {

    my $conn = make_conn_and_complain(shift);
    my ($share) = validate_pos(@_, { type => SCALAR });

    # Convert command if simulating a vfiler context
    my $cmd = filer_vfiler_command($conn, "cifs shares $share");
    
    my $text = sayEOF($conn, $cmd);

    croak("$text\n") if ($text =~ /CIFS (is )?not (licensed|running|configured)/i);

    return( $text =~ /There is no share named \Q$share\E|No share is matching that name/i ? 0 : 1);
}

=head2 share_add ($conn, $share, $path, %options)

Creates CIFS share on filer.
This routine will not complain if the share was already existing
if purge_existing is 1 the it will remove existing share 
and create it again.
It will also validate if the share was *really* created.

Returns 1 on success otherwise die with $@.

Arguments

=over 4

=item conn

(Required) : Connection Object to filer.

=item share

(Required) : name of share

=item path

(Required) : path of share

=item force

(Optional) : boolean denoting "-f" option

=item comment

(Optional) : comment string for the share

=item maxusers

(Optional) : max users

=item groupname

(Optional) : group name

=item widelink

(Optional) : boolean denoting -widelink option

=item novscan

(Optional) : boolean denoting -novscan option

=item novscanread

(Optional) : boolean denoting -novscanread option

=item nosymlink_strict_security

(Optional) : boolean denoting -nosymlink_strict_security option

=item umask

(Optional) : mask to be used

=item caching

(Optional) : caching type to be used

Possible values : no / auto_document / auto_program

=item purge_existing

(Optional) : purge existing share

=item verify

(Optional) : will verify if the share addition really worked (default : 1)

=back

=cut

sub share_add {

    my $conn = make_conn_and_complain(shift);
    my ($share, $path, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR }, { type => SCALAR } ],
                    {
                     force => { type=>BOOLEAN, optional=>1 },
                     comment => { type=>SCALAR, optional=>1 },
                     maxusers => { type=>SCALAR, optional=>1, regex=>qr/\d+/ },
                     groupname => { type=>SCALAR, optional=>1 },
                     widelink => { type=>BOOLEAN, optional=>1 },
                     novscan => { type=>BOOLEAN, optional=>1 },
                     novscanread => { type=>BOOLEAN, optional=>1 },
                     nosymlink_strict_security => { type=>BOOLEAN, optional=>1 },
                     umask => { type=>SCALAR, optional=>1 },
                     caching => { type=>SCALAR, optional=>1, regex=>qr/no|auto_document|auto_program/ },
                     purge_existing => { type=>BOOLEAN, optional=>1 },
                     verify => { type=>BOOLEAN, optional=>1, default=>1 },
                    }
                );

    if ($args{purge_existing} && share_exists($conn, $share) ) {

        share_delete($conn, $share)
            or croak("$@\n");
    }
        
    my $cmd = "cifs shares -add $share $path";
    $cmd .= " -f" if ($args{force});
    $cmd .= " -comment \"$args{comment}\"" if ($args{comment});
    $cmd .= " -maxusers $args{maxusers}" if ($args{maxusers});
    $cmd .= " -forcegroup \"$args{groupname}\"" if ($args{groupname});
    $cmd .= " -widelink" if ($args{widelink});
    $cmd .= " -novscan" if ($args{novscan});
    $cmd .= " -novscanread" if ($args{novscanread});
    $cmd .= " -nosymlink_strict_security" if ($args{nosymlink_strict_security});
    $cmd .= " -umask $args{umask}" if ($args{umask});
    $cmd .= " -$args{caching}".'_caching' if ($args{caching});

    # Convert command if simulating a vfiler context
    $cmd = filer_vfiler_command($conn, $cmd);

    my $text = sayEOF($conn, $cmd);

    croak("$text\n") if ($text =~ /CIFS (is )?not (licensed|running|configured)/i);

    return(1) if ($text =~ /already a share named \Q$share\E/i);

    # Remove vfiler header
    $text =~ s/(?:.|\n)*=====\s+\S+//;

    #ok if not acessible by some MS-DOS workstations
    $text=~ s/\A(\n)?The share name \'\Q$share\E\' will not be accessible by some MS-DOS workstations\s*(\n|\z)//;

    # no output to verify ; die if we get something
    die("$text\n") if ($text);

    # verify if share got created if asked to
    croak("Verification : Addition of share $args{share} failed.\n")
        if ($args{verify} && (!share_exists($conn, $share)) );

    return(1);
}


=head2 share_delete ($conn, $sharename, %options)

Delete CIFS share on filer

Will not fail if the share to be deleted was non existant.

Returns 1 on success otherwise die with $@.

Arguments

=over 4

=item conn

(Required) : Connection Object to filer.

=item share

(Required) : name of share to delete.

=item force

(Optional) : Boolean whether to force the deletion or not

=item die_non_existing

(Optional) : Boolean, if set will die if the share to be deleted
does not exist.

=item verify

(Optional) : Boolean, if set will verify if the delete really happened

=back


=cut

sub share_delete {

    my $conn = make_conn_and_complain(shift);
    my ($share, %args) = validate_mixed( @_,
                            [ { type => SCALAR } ],
                            {
                                force => { type=>BOOLEAN, optional=>1 },
                                die_non_existing => { type=>BOOLEAN,
                                                    optional=>1, default=>0 },
                                verify => { type=>BOOLEAN, optional=>1, default=>1 }
                            }
                            );


    my $cmd = "cifs shares -delete ";
    $cmd .= "-f " if ($args{force});
    $cmd .= "$share";
  
    # Convert command if simulating a vfiler context
    $cmd = filer_vfiler_command($conn, $cmd);
    
    my $text = sayEOF($conn, $cmd);

    croak("$text\n") if ($text =~ /CIFS (is )?not (licensed|running|configured)/i);

    if ($text =~ /no share named \Q$share\E/i) {
        if ($args{die_non_existing}) {
            croak("$text\n");
        } else {
            return(1);
        } 
    } 

    # Remove vfiler header
    $text =~ s/(?:.|\n)*=====\s+\S+//;

    # no output to verify ; die if we get something
    croak("$text\n") if ($text);

    # verify if share got deleted, if asked to
    croak("Verification : Deletion of share $args{share} failed.\n")
        if ($args{verify} && share_exists($conn, $share)); 

    return(1);
}


=head1 BUGS

    The following routines will be added/updated when needed
        - to get details of a / all CIFS share(s) on filer.
        - share_change

=head1 ONTAPCLI

exportfs, cifs shares,

=cut

1;

