# -*- perl -*-
#  $Id: $


## @summary Library of FPolicy CLI
## @author  sureshb@netapp.com
## @status  private
## @pod     here

package FPolicy::CLI;

use strict;

#use warnings;
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF OBJECT);
use Tharn qw(say logcomment logresult host );
use Carp qw(croak);
use SayWrap qw(sayEOF);
use Subtest;
use FPolicy::Utils qw(make_NV PreProcess_InputParams Is_policy_exists make_connection make_arrayhash make_hash_fromArrayHash trim_str);
use FPolicy::Grammer qw(GenerateCmd ValidateParams);
use FPolicy::Parser qw( Get_LocalBlkdPolicyStat Get_Monitored Get_PServer Get_Extension Get_PolicyStat Get_PolicyState Get_Volume Get_CifsState Get_SnapDetails CLI_Execute_Validate_Outputstr get_STATES CLI_Verify_NV_Generic);
use FPolicy::Constants qw(%OpsTemplate);

use FPolicy::Generator qw(CLI_Expected_Str);
use FPolicy::Utils qw(stats_parse);

=head1 NAME

new - this is the object instantiation routine for CLI interface
This module should be instantiated by FPolicy module.
Certain features like version control etc will not work well when directly used.

=head1 SYNOPSIS
	use FPolicy::CLI;

	my $pol = new FPolicy("policy=POLICY", "vfiler=vfiler0", "filer=FILER", 
			"monitor=read,write,open,close", 
			"vol_inc=vol0", "vol_exc=", 
			"ext_inc=txt", "ext_exc=ppt", 
			"pserver_pri=10.10.1.1", "pserver_pri=20.1.1.1",
			"client=$CLIENT"
		);

	$pol->connect();
	$pol->create();
	$pol->enable();
	$pol->disable();
	$pol->monitor(
	
=head1 DESCRIPTION

This package provides the CLI interface, instantiated and destroyed by the FPolicy object.
Data structures, parent object are tightly coupled to this module hence should not be directly instantiated.
One or more of the functionality may not work when directly instantiated.

=head1 ROUTINES

=head2 new

	my $pol = new FPolicy("policy=POLICY", "vfiler=vfiler0", "filer=FILER", 
			"monitor=read,write,open,close", 
			"vol_inc=vol0", "vol_exc=", 
			"ext_inc=txt", "ext_exc=ppt", 
			"pserver_pri=10.10.1.1", "pserver_pri=20.1.1.1",
			"client=$CLIENT"
		);

B<new> causes all parameters to be processed and stored for further processing.
These name value pairs are optional. However policy name is mandatory but not imposed at this object level.
optional name value pairs can be provided later with relevant API's if required to change.


The following named arguments are accepted:

=over

=item policy
(Mandatory) : Name of the policy

=item vfiler
(Optional) : Name of the virtual filer to be used

Default is vfiler0
To use the non-default vfiler, multistore license should be provisioned

=item filer
(Mandatory) : Name of the Filer

=item monitor
(Optional) : Operations to be monitored

Values = read,write,open,close,create,delete,rename,link,symlink,getattr,setattr,lookup,dircreate,dirdelete,dirrename

=item vol_inc
(Optional) : comma seperated list of Volume(s) in the inclusion list

=item vol_exc
(Optional) : comma seperated list of Volume(s) in the exclusion list

=item ext_inc
(Optional) : comma seperated list of Extensions(s) in the inclusion list

=item ext_exc
(Optional) : comma seperated list of Extensions(s) in the exclusion list

=item pserver_pri
(Optional) : comma seperated list of IP address(es) of primary policy server(s)

=item pserver_sec
(Optional) : comma seperated list of IP address(es) of secondary policy server(s)

=item client
(Mandatory) : Name of the NFS client to mount the filer and perform file operations

=back

=cut


sub new ($;@)
{
	my $object = shift;
	my $class = ref($object) || $object;

	my %NV = make_NV(@_);

	my $CLI = {
	       INTF => "cli",
		FILER => {
			NAME => "$NV{filer}",
			IP => "$NV{filer_ip}",
			VERSION => $NV{version} || ""
			},

		POLICY => {
			NAME => "$NV{policy}",
			TYPE => "$NV{type}" || "screen",
			MONITOR => "$NV{monitor}",
			PROTOCOL => "$NV{proto}",
			EXT_INC => "$NV{ext_inc}",
			EXT_EXC => "$NV{ext_exc}",
			VOL_INC => "$NV{vol_inc}",
			VOL_EXC => "$NV{vol_exc}",
			PSERVER_PRI => "$NV{pserver_pri}",
			PSERVER_SEC => "$NV{pserver_sec}",
			},

		VFILER => {NAME => "$NV{vfiler}" || "vfiler0"},

		CLIENT => "$NV{client}",

		CONN => undef,
		HOST => undef,

	};

	logcomment("Initializing CLI Interface
		INTF = cli
		FILER{NAME} = $NV{filer}
		FILER{VERSION} = $NV{version}
		FILER{IP} = $NV{filer_ip}
		POLICY{NAME} = $NV{policy}
		POLICY{TYPE} = $NV{type}
		POLICY{MONITOR} = $NV{monitor}
		POLICY{PROTOCOL} = $NV{proto}
		POLICY{EXT_INC} = $NV{ext_inc}
		POLICY{EXT_EXC} = $NV{ext_exc}
		POLICY{VOL_INC} = $NV{vol_inc}
		POLICY{VOL_EXC} = $NV{vol_exc}
		POLICY{PSERVER_PRI} = $NV{pserver_pri}
		POLICY{PSERVER_SEC} = $NV{pserver_sec} ");


	bless($CLI, $class); 

	return $CLI;
}


=head1 NAME
create - This API is to create the policy on the filer

=head2 DESCRIPTION
This API is used to create policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->create();
	$pol->{cli}->create("result=FAIL");

Disable result logging mechanism by this module
	$pol->{cli}->create("result=noresult");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut

sub create ($;@)
{
	my $self = shift;
	my %parms = validate
	(
		@_, 
		{
		conn => { 	type => SCALAR|OBJECT, 
				default => $self->{CONN}, 
				optional => 1 },

		policy => { 	type => SCALAR, 
				default => "$self->{POLICY}{NAME}", 
#				regex => qr/(\w+)/, 
				optional => 1 },

		type => { 	type => SCALAR, 
				default => "$self->{POLICY}{TYPE}", 
				optional => 1},

		result => { 	type => SCALAR, 
				default => "PASS", 
				regex =>  qr/(\w+)/, 
				optional => 1 },

		vfiler => { 	type => SCALAR, 
				default => "$self->{VFILER}{NAME}", 
				regex => qr/(\w+)/, 
				optional => 1},
		
		force => {	type => SCALAR, 
					default => "no", 
					optional => 1},
					
		expected => { 	type => SCALAR|HASHREF, 
				optional => 1 }
		}
	);

	if($parms{force} eq "yes")
	{
		if(Is_policy_exists(conn => $parms{conn}, intf => "cli", vfiler => $parms{vfiler}, policy => $parms{policy}))
		{
			$self->destroy(policy=>"$parms{policy}");
		}
	}
	# fallback to default values when expected values are not passed
	# forming expected {success => "string", failure => [array of strings]}
	if($parms{expected} eq "")
	{
		$parms{expected} = CLI_Expected_Str( 
			version => $self->{FILER}{VERSION},  # versioning for differences of output across releases
			cmd_type => "policy_create",	# keyword
			policy => $parms{policy},
			cmd_code => "already_exists" ); # cmd_code is currently not used, for future use

	}else {

		my %str = ( success => $parms{expected} );
		$parms{expected} = \%str;

	}

	#logcomment("CLI Expected, success=${$parms{expected}}{success}, failure=${$parms{expected}}{failure}");


	# result is in the form of {result => true/false/0/1 , output => "obtained output string" }
	my %ptr =  
	CLI_Execute_Validate_Outputstr( 

		conn => $parms{conn}, 
		intf => "cli",
		vfiler => "$parms{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd(
				cmd_type => "create",
				vfiler => "$parms{vfiler}",
				policy => $parms{policy}, 
				version => $self->{FILER}{VERSION}, 
				type=> ( $parms{type} ne "")? 
					$parms{type} : 
					$self->{POLICY}{TYPE} 
			),
	
		expected => $parms{expected},

		result => "$parms{result}"
	);


	# decide result based on user 'result' and obtained result
#	if( $parms{result} =~ /noresult/i )
#	{ logcomment("No Result"); }  # module does not log result
#
#	elsif($parms{result} =~ /fail/i)
#	{
#		my $tmp = ($ptr{result} =~ /pass/i)? 1 : 0; # negate the result
#
#		if($tmp) { logresult("fail", "Failure"); } 
#
#		else { logresult("pass", "Success");  return 0; }
#
#	} elsif($parms{result} =~ /pass/i and $ptr{result} == 0)
#	{
#
#		logresult("pass", "Success");  
#		return 0; 
#
#	} else 
#	{ logresult("fail", "Failure"); }
#
#	return $ptr{result};
	return decideResult($ptr{result}, $parms{result});

}



=head1 NAME
enable - This API is to enable the policy on the filer

=head2 DESCRIPTION
This API is used to enable policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->enable();
	$pol->{cli}->enable("result=FAIL");

Disable result logging mechanism by this module
	$pol->{cli}->enable("result=noresult");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub enable ($;@)
{
	my $self = shift;
	my %NV = @_;
	my %parms = validate 
	(
		@_,
		{
		conn => { 	type => SCALAR|OBJECT, 
				default => $self->{CONN}
			},

		policy => { 	type => SCALAR, 
				default => "$self->{POLICY}{NAME}", 
				regex =>  qr/(\w+)/, 
				optional => 1
			 },

		vfiler => { 	type => SCALAR, 
				default => "$self->{VFILER}{NAME}", 
				regex => qr/(\w+)/, 
				optional => 1
			},

		result => { 	type => SCALAR, 
				default => "PASS", 
				regex =>  qr/[pass|fail]/i, 
				optional => 1
			},

		expected => { 	type => SCALAR|HASHREF, 
				default => CLI_Expected_Str(
					version => $self->{FILER}{VERSION}, 
					cmd_type => "policy_enable",
					policy => $NV{policy} || $self->{POLICY}{NAME}, 
					cmd_code => "already_enabled" # not used at this point
					 ),
				optional => 1 
			}
		}
	);

	# result is in the form of {result => true/false/0/1 , output => "obtained output string" }
	my %ptr = 
	CLI_Execute_Validate_Outputstr( 

		conn => $parms{conn}, 
		intf => "cli",
		vfiler => "$parms{vfiler}",
		filer => "$self->{FILER}{NAME}",
		command => GenerateCmd(
				vfiler => "$parms{vfiler}",
				cmd_type => "enable",
				version => $self->{FILER}{VERSION}, 
				policy => $parms{policy} ),
	
		expected => $parms{expected},
		result => $parms{result}
	);

	# decide result based on user 'result' and obtained result
#	if( $parms{result} =~ /noresult/i )
#	 { logcomment("No Result"); } 
#
#	elsif( $parms{result} =~ /fail/i )
#	{
#		my $tmp = ($ptr{result} =~ /pass/i)? 1 : 0;
#
#		if($tmp) 
#		{ logresult("fail", "Failure"); }
#
#		else 
#		{ 
#			logresult("pass", "Success");  
#			return 0;
#		}
#
#	}
#	elsif( $parms{result} =~ /pass/i and $ptr{result} == 0 )
#	{
#		logresult("pass", "Success");
#		return 0; 
#	} else
#	{ logresult("fail", "Failure"); }
#
#	return $ptr{result};
	return decideResult($ptr{result}, $parms{result});

}


=head1 NAME
disable - This API is to disable the policy on the filer

=head2 DESCRIPTION
This API is used to disable policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->disable();
	$pol->{cli}->disable("result=FAIL");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub disable ($;@)
{
	my $self = shift;
	my %NV = @_;
	my %parm = validate
	(
		@_, 
		{
			conn => { 	type => SCALAR|OBJECT, 
					default => $self->{CONN} 
				},

			policy => { 	type => SCALAR, 
					default => "$self->{POLICY}{NAME}", 
					regex =>  qr/(\w+)/ 
				},
			vfiler => { 	type => SCALAR, 
					default => "$self->{VFILER}{NAME}", 
					regex => qr/(\w+)/, 
					optional => 1
				},
			result => { 	type => SCALAR, 
					default => "PASS", 
					regex => qr/[pass|fail]/i, 
					optional => 1 
				},
			expected => { 	type => SCALAR|HASHREF, 

					default => CLI_Expected_Str(
						version => $self->{FILER}{VERSION}, 
						cmd_type => "policy_disable",
						policy => $NV{policy} || $self->{POLICY}{NAME}, 
	
					cmd_code => "already_disabled" ),
		
					qr/((\s)+(\w+)(\s)+)*/i, optional => 1 
				}
		}
	);

	my %ptr = 
	CLI_Execute_Validate_Outputstr
	( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd(
				vfiler => "$parm{vfiler}",
				version => $self->{FILER}{VERSION}, 
				cmd_type => "disable",
				policy => $parm{policy} 
			),
	
		expected => $parm{expected},

		result => $parm{result}
	);


#	if($parm{result} =~ /noresult/i)
#	 { logcomment("No Result"); } 
#
#	elsif( $parm{result} =~ /fail/i )
#	{
#		my $tmp = ($ptr{result} =~ /pass/i)? 1 : 0;
#
#		if($tmp)
#		{ logresult("fail", "Failure"); }
#
#		else
#		{ 
#			logresult("pass", "Success");  
#			return 0; 
#		}
#	}
#	elsif( $parm{result} =~ /pass/i and $ptr{result} == 0 ) 
#	{
#		logresult("pass", "Success");  
#		return 0; 
#	}
#	else
#	{ logresult("fail", "Failure"); }
#
#	return $ptr{result};
	return decideResult($ptr{result}, $parm{result});

}


=head1 NAME
destroy - This API is to destroy the policy on the filer

=head2 DESCRIPTION
This API is used to destroy policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->destroy();
	$pol->{cli}->destroy("result=FAIL");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub destroy ($;@)
{
	my $self = shift;
	my %NV = @_;
	my %parm = validate
	(
		@_,
		{
			conn => {	type => SCALAR|OBJECT , 
					default => $self->{CONN}
				},

			policy => { 	type => SCALAR, 
					default => "$self->{POLICY}{NAME}", 
#					regex =>  qr/(\w+)/ 
				},

			vfiler => { 	type => SCALAR, 
					default => "$self->{VFILER}{NAME}", 
					regex => qr/(\w+)/, 
					optional => 1
				},

			result => { 	type => SCALAR, 
					default => "PASS", 
					regex =>  qr/[pass|fail]/i, 
					optional => 1 
				},

			expected => { 	type => SCALAR|HASHREF, 

					default => CLI_Expected_Str(
						version => $self->{FILER}{VERSION}, 
						cmd_type => "policy_destroy",
						policy => $NV{policy} || $self->{POLICY}{NAME}, 
						cmd_code => "already_destroy" # not in use
						),

					qr/((\s)+(\w+)(\s)+)*/i,
					optional => 1 
			}
		}
	);

	
	if($parm{result} ne "fail")
	{
		$self->disable(policy=>"$parm{policy}");
		
	}
	
	my %ptr = 
	CLI_Execute_Validate_Outputstr
	( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",
		command => GenerateCmd (
				vfiler => "$parm{vfiler}",
				version => $self->{FILER}{VERSION}, 
				cmd_type => "destroy",
				policy => "$parm{policy}" 
			),
		expected => $parm{expected},
		result => "$parm{result}"
	);

	my $status = decideResult($ptr{result}, $parm{result});


	# Verify whether the policy has been destroyed. 
	# calculate the expected result for destroy verify operation
	
	
	my $exp = CLI_Expected_Str ( 
				cmd_type => "policy_destroyverify",
				version => $self->{FILER}{VERSION}, 
				policy => $NV{policy} || $self->{POLICY}{NAME}
		 );


	my %ptrv = 
	CLI_Execute_Validate_Outputstr( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd(
				vfiler => "$parm{vfiler}",
				version => $self->{FILER}{VERSION}, 
				cmd_type => "destroy",
				policy => "$parm{policy}"
			 ),

		expected => $exp,
		result => "pass"  			# since we are verifying if the policy has been destroyed. Expected to PASS.
	);

	return decideResult($ptrv{result}, "pass");	# since we are verifying if the policy has been destroyed. Expected to PASS.

}


=head1 NAME
show - This API is to show the policy on the filer

=head2 DESCRIPTION
This API is used to show policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->show();
	$pol->{cli}->show("result=FAIL");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub show ($;@)
{
	my $self = shift;
	my %NV = @_;
	my %parm = validate
	(
		@_, 
		{
			conn => { 	type => SCALAR|OBJECT, 
					default => $self->{CONN}
				 },

			policy => { 	type => SCALAR, 
					default => "$self->{POLICY}{NAME}", 
					regex =>  qr/(\w+)/ 
				},

			vfiler => { 	type => SCALAR, 
					default => "$self->{VFILER}{NAME}", 
					regex => qr/(\w+)/, 
					optional => 1
				},

			result => { 	type => SCALAR, 
					default => "PASS", 
					regex =>  qr/[pass|fail]/i, 
					optional => 1 
				},

			expected => { 	type => SCALAR|HASHREF, 

					default => CLI_Expected_Str (
						version => $self->{FILER}{VERSION}, 
						policy => $NV{policy} || $self->{POLICY}{NAME}, 
						policystate => $NV{policystate}, 
						cmd_type => "policy_show_default"
					),

					qr/((\s)+(\w+)(\s)+)*/i, 
					optional => 1 
				}
		});

	my %ptr = 
	CLI_Execute_Validate_Outputstr
	(
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd(
				vfiler => "$parm{vfiler}",
				version => $self->{FILER}{VERSION}, 
				cmd_type => "show",
				policy => "$parm{policy}", 
			),

		expected => $parm{expected},
		result => "$parm{result}"
	);


#	if($parm{result} =~ /noresult/i)
#	{ logcomment("No Result"); } 
#
#	elsif($parm{result} =~ /fail/i)
#	{
#		my $tmp = ($ptr{result} =~ /pass/i)? 1 : 0;
#
#		if($tmp) 
#		{ logresult("fail", "Failure"); }
#
#		else 
#		{ 
#			logresult("pass", "Success");  
#			return 0; 
#		}
#	}
#	elsif( $parm{result} =~ /pass/i and $ptr{result} == 0 ) 
#	{
#		logresult("pass", "Success");  
#		return 0; 
#	}
#	else 
#	{ logresult("fail", "Failure"); }
#
#	return $ptr{result};
	
	return decideResult($ptr{result}, $parm{result});
}

=head1 NAME
verify - This API is to verify the policy on the filer

=head2 DESCRIPTION
This API is used to verify policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->verify();
	$pol->{cli}->verify("result=FAIL");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut

sub verify ($;@)
{
	my $self = shift;

	logcomment("Verifying - @_");
	my %parm = validate
	(
		@_, 
		{
		conn => { 	type => SCALAR|OBJECT, 
				default=> $self->{CONN}, 
				optional => 1 
			},
		policy => { 	type => SCALAR, 
				default => $self->{POLICY}{NAME}, 
				regex =>  qr/(\w+)/ 
			},
		policystate => { 	
				type => SCALAR, 
				optional => 1
			},
		policy_type => { 
				type => SCALAR, 
				optional => 1
			},
		monitor => { 	type => SCALAR, 
				optional => 1
			},
		vol_inc => { 	type => SCALAR, 
				optional => 1
			},
		vol_exc => { 	type => SCALAR, 
				optional => 1
			},
		ext_inc => { 	type => SCALAR, 
				optional => 1
			},
		ext_exc => { 	type => SCALAR,
				optional => 1
			},
		pserver_ip => { 
				type => SCALAR, 
				optional => 1
			},
		sserver_ip => { 
				type => SCALAR, 
				optional => 1
			},
		pstat_screen => { 
				type => SCALAR, 
				optional => 1
			},
		pstat_screenfail => { 
				type => SCALAR, 
				optional => 1
			},
		# Added by Prashant For verifying localblock count ####
		
		pstat_localblock => { 
				type => SCALAR, 
				optional => 1
			},			
		cifs_state => { 
				type => SCALAR, 
				optional => 1
			},
        opt_show => {
                type => SCALAR,
                optional => 1
            },			
		opt_show_monitor_ads => {               #### Added by Rajaram
                type => SCALAR,
                optional => 1
            },
		result => { 
				type => SCALAR, 
				default => "pass", 
				regex =>  qr/[pass|fail]/i, 
				optional => 1 
			},
		expected => { 
				type => SCALAR|HASHREF, 
#				default => CLI_Expected_Str(
#					cmd_type => "policy_verify"),
				 optional => 1 }
		});


	# retrieve the policy under test
	my %tmp = (success => "Number of");

	my %ptr = 
	CLI_Execute_Validate_Outputstr
	( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd (
				cmd_type => "show",
				version => $self->{FILER}{VERSION}, 
				vfiler => "$parm{vfiler}" || $self->{VFILER}{NAME},
				policy => $parm{policy}
			),

		expected => \%tmp,
		#result => "$parm{result}"  # use the provided result definition 
	 	result => "pass" 	 #here we are only getting the output of the fpolicy and hence expecting to pass. This result is not same as the expected result passed from the test case.
	);

	my %output = (); # result hash to compare
	my %expected; # = ($parm{expected}); # store expected name value pairs to compare result

	# verify monitored operations, if expected
	if(defined $parm{monitor}) {
		$output{monitor} = Get_Monitored($ptr{output});
		logcomment("CLI:Verify: received Monitor operations = \"$output{monitor}\"");
		$expected{monitor} = "";
		my $version = "";
		foreach my $k (split/,/, $parm{monitor}){ 
#			$expected{monitor} .= ucfirst(lc($k)) .",";  
			$version = $self->{FILER}{VERSION};
			# version control for monitoring ops lookup from template
			my $ref = ( $version ne "" )? 
					$OpsTemplate{$version} :
					$OpsTemplate{default} ;
		
			# map the Operation from template
			my $ops = ${$ref}{$k};
			
			$expected{monitor} .= $ops .",";  
		} #for
		logcomment("Expected monitor Operations from Template for version $version: $expected{monitor}");
	}

	# verify policy server IP address, if expected
	if(defined $parm{pserver_ip}) 
	{
		my $pserver = Get_PServer($ptr{output});

		my ($pserver_ip, $pserver_name, $pserver_state, 
			$pserver_time, $pserver_screen, $pserver_screenfail) =
			 split /\,/, $pserver;

		foreach my $key ("pserver_ip", "pserver_name", "pserver_state", "pserver_time", "pserver_screen", "pserver_screenfail")
		{
			if(defined $parm{$key})
			{ 
				$output{$key} = ${$key};
				$expected{$key} = $parm{$key};
			}
			 
		}
		# store only the results for values defined in the expected list
		#$output{pserver_ip} = $pserver_ip if( defined $parm{pserver_ip} );
		#$output{pserver_name} = $pserver_name if( defined $parm{pserver_name} );
		#$output{pserver_state} = $pserver_state if( defined $parm{pserver_state} );
		#$output{pserver_time} = $pserver_time if( defined $parm{pserver_time} );
		#$output{pserver_screen} = $pserver_screen if( defined $parm{pserver_screen} );
		#$output{pserver_screenfail} = $pserver_screenfail if( defined $parm{pserver_screenfail} );

		logcomment("CLI:Verify: received policy server: pserver_ip=$pserver_ip, pserver_name=$pserver_name, 
				pserver_state=$pserver_state, pserver_time=$pserver_time,
				pserver_screen=$pserver_screen, pserver_screenfail=$pserver_screenfail");
		
	}

	#verify extension (both inclusion and exclusion), if expected
	if( defined $parm{ext_exc} or defined $parm{ext_inc} ) 
	{
		my $extension = Get_Extension($ptr{output});

		logcomment("CLI:Verify: received extension = $extension");

		my ($extn_inc, $extn_exc) = split /\;/, $extension;

		#include only the required fields based on expected
		$output{ext_inc} = trim_str($extn_inc) if( defined $parm{ext_inc} );
		$output{ext_exc} = trim_str($extn_exc) if( defined $parm{ext_exc} );
		
		$expected{ext_inc} = uc($parm{ext_inc}) if( defined $parm{ext_inc} );
		$expected{ext_exc} = uc($parm{ext_exc}) if( defined $parm{ext_exc} );
		if (($expected{ext_inc} eq "") and (defined $parm{ext_inc})) 
		{ $expected{ext_inc} = "Extensions-to-screen list is empty.";}
		if (($expected{ext_exc} eq "") and (defined $parm{ext_exc}))
		{ $expected{ext_exc} = "Extensions-not-to-screen list is empty.";}
		
	}

	### Added by Prashant(prash@netapp.com)
	#verify policy statistics - Number of operations locally blocked, if expected
	if( defined $parm{pstat_localblock})
	{
		my $policystat = Get_LocalBlkdPolicyStat($ptr{output});
		
		logcomment("CLI:Verify: received policy stat, Locally blocked = $policystat");
		#my ($scr, $fail) = split /\,/, $policystat;
		my $fail = $policystat;
		
		#$output{pstat_screen} = $scr if( defined $parm{pstat_screen} );
		$output{pstat_localblock} = $fail if( defined $parm{pstat_localblock} );
		#$expected{pstat_screen} = $parm{pstat_screen} if( defined $parm{pstat_screen} );
		$expected{pstat_localblock} = $parm{pstat_localblock} if( defined $parm{pstat_localblock} );
	}
		
	#verify policy statistics - file screened and screen failures, if expected
	if( defined $parm{pstat_screen} or defined $parm{pstat_screenfail} )
	{
		my $policystat = Get_PolicyStat($ptr{output});

		logcomment("CLI:Verify: received policy stat = $policystat");

		my ($scr, $fail) = split /\,/, $policystat;

		$output{pstat_screen} = $scr if( defined $parm{pstat_screen} );
		$output{pstat_screenfail} = $fail if( defined $parm{pstat_screenfail} );
		$expected{pstat_screen} = $parm{pstat_screen} if( defined $parm{pstat_screen} );
		$expected{pstat_screenfail} = $parm{pstat_screenfail} if( defined $parm{pstat_screenfail} );
	}

	# verify policy state (enabled/disabled), if expected
	if( defined $parm{policystate} )
	{
		$output{policystate} = Get_PolicyState($ptr{output});
		$expected{policystate} = $parm{policystate};
		logcomment("CLI:Verify: received policy state = $output{policystate}");
	}

	# verify cifs state (disable/enabled), if expected
	if(defined $parm{cifsstate})
	{
		$output{cifsstate} = Get_CifsState($ptr{output});
		$expected{cifsstate} = $parm{cifsstate};
		logcomment("CLI:Verify: received cifs state = $output{cifsstate}");
	}

	my $status = 0;

	### Verify Local file blocking status..
	### Added by Prashant
	
	if(defined $parm{opt_show})
	{
		if($parm{opt_show} eq "on"){
			my $optr = $self->options( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				method => "required", 
				value => " ",
				expected => { success => "fpolicy options $parm{policy} required: on" }
			);
			$status |= $optr;			
		}else{
			my $optr = $self->options( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				method => "required", 
				value => " ",
				expected => { success => "fpolicy options $parm{policy} required: off" }
			);
		$status |= $optr;
		}
	}

	### Verify monitor ADS status..
	### Added by Rajaram
	if(defined $parm{opt_show_monitor_ads})
	{
		if($parm{opt_show_monitor_ads} eq "on"){
			my $optr = $self->options( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				method => "monitor_ads", 
				value => " ",
				expected => { success => "fpolicy options $parm{policy} monitor_ads: on" }
			);
			$status |= $optr;			
		}else{
			my $optr = $self->options( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				method => "monitor_ads", 
				value => " ",
				expected => { success => "fpolicy options $parm{policy} monitor_ads: off" }
			);
		$status |= $optr;
		}
	}
	
	# verify volume inclusion, if expected
	if(defined $parm{vol_inc})
	{
		if($parm{vol_inc} eq ""){
			my $vptr = $self->volume( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				oper => "show",
				volume => $parm{vol_inc}, 
				method => "inc", 
				expected => { success => "List empty - All volumes are being monitored." }
				);
			$status |= $vptr;			
		}else{
			my $vptr = $self->volume( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				oper => "show",
				volume => $parm{vol_inc}, 
				method => "inc", 
				expected => { success => $parm{vol_inc} }
				);
			$status |= $vptr;
		}
	}

	# verify volume exclusion, if expected
	
	# 
	if(defined $parm{vol_exc}) {
		if($parm{vol_exc} eq ""){
			my $vptr = $self->volume( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				oper => "show",
				method => "exc", 
				volume => $parm{vol_exc}, 
				expected => { success => "List empty - Not excluding any volume." } 
#sandeeppTBD: use if/? and eliminate the below else by giving both success conditions above.				
				);
		$status |= $vptr;

		}else{
			my $vptr = $self->volume( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				oper => "show",
				method => "exc", 
				volume => $parm{vol_exc}, 
				expected => { success => $parm{vol_exc} } 
				);
			$status |= $vptr;
		}
	}
	
	# Verify Secondary Server IP 
#	if(defined $parm{sserver_ip})
#	{
#		my $optr = $self->options( 
#				conn => $parm{conn}, 
#				policy => $parm{policy}, 
#				result => $parm{result}, 
#				method => "secondary_servers", 
#				value => " ",
#				expected => { success => $parm{sserver_ip} }
#			);
#	}	
	if(defined $parm{sserver_ip})
	{	
		if($parm{sserver_ip} eq ""){
			my $optr = $self->options( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				method => "secondary_servers", 
				value => " ",
				expected => { success => "No secondary file screening servers list"}
			);
			$status |= $optr;

		}else{
			my $optr = $self->options( 
				conn => $parm{conn}, 
				policy => $parm{policy}, 
				result => $parm{result}, 
				method => "secondary_servers", 
				value => " ",
				expected => { success => $parm{sserver_ip} }
			);
			$status |= $optr;

		}
	}
	
	my $str = "Obtained output:\n";
	foreach my $key (keys %output)
	{ $str .= "Key = $key, value = $output{$key}\n"; }

	$str .= "Expected output:\n";
	foreach my $key (keys %expected)
	{ $str .= "Key = $key, value = $expected{$key}\n"; }

	logcomment($str);
	# compare the expected list and the obtained output for result analysis
	$status |= CLI_Verify_NV_Generic(\%output, \%expected);

	#logcomment("CLI Verify status  = $status");

#	if( $parm{result} =~ /noresult/i) 
#	{ logcomment("No Result"); }
#	
#	elsif( $parm{result} =~ /FAIL/i) 
#	{
#		$status = ($status)? 0 : 1; # negate the obtained result
#
#		if($status) # match failure with expected fail state
#		{
#			logresult("fail", "Output did not match expected results. 
#				refer to above logs for further details, Details : $self->{FLAGS}{result}, $status");
#			return 1;
#		} 
#		else # output failed as expected result
#		{
#			logresult("pass", "Output did NOT match expected results, 
#				Expected to $self->{FLAGS}{result}, $status");
#		}
#	}
#
##	elsif ( $parm{result} =~ /noresult/i )
##	{ logcomment("No Result"); } 
##
#	else 
#	{
#		if($status) # failure condition
#		{
#			logresult("fail", "Output did not match expected results. 
#						refer to above logs for further details, Details : $self->{FLAGS}{result}, $status");
#			return 1;
#		} 
#		else # output passed as expected result
#		{
#			logresult("pass", "Output matched expected results, 
#				Expected to $self->{FLAGS}{result}, $status");
#		}
#	}
#	return $status;
	return decideResult($status, $parm{result});
}


=head1 NAME
verify_system_defaults - This API is to verify system default settings with respect to file policy on the filer

=head2 DESCRIPTION
This API is used to verify system default settings on the filer using CLI
This should be performed when the filer is first boot up with new image to verify the default settings
for license, cifs settings etc

Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->verify_system_defaults();
	$pol->{cli}->verify_system_defaults("result=FAIL");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
#sub verify_system_defaults ($@)
#{
#	my $self = shift;
#	my @options = PreProcess_InputParams($self, @_) ;
#	my %opt = make_NV(@options);
#
#	my $cifs_state = "";
#	my $fpolicy_state = "";
#
# 	if(defined $opt{cifs}) {
#		$cifs_state = get_STATES("FPOLICY_STATES", 
#					"CIFS_NOT_LICENSED");
#		foreach my $s (@{$cifs_state}) { push @options, $s; }
#	}
#
#	if(defined $opt{fpolicy}) {
#		$fpolicy_state = get_STATES("CIFS_STATES", 
#					"NOT_LICENSED");
#		foreach my $s (@{$fpolicy_state}) { push @options, $s; }
#	}
#
#	logcomment("FPolicy:CLI: Verify system defaults : \"@{$cifs_state}\", \"@{$fpolicy_state}\"");
#
#	my $stat = $self->verify( @options);
#	if($stat) {
#		logresult("fail", "REASON = OUTPUT not as expected - $self->{expected}");
#	}
#	return $stat;
#}

=head1 NAME
monitor - This API is to perform monitor operations for the policy on the filer

=head2 DESCRIPTION
This API is used to perform monitor operations for the policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->monitor("oper=add", "result=FAIL");
	$pol->{cli}->monitor("oper=add", "opspec=link,symlink", "protocol=nfs");
=over

=item oper
operation to be performed
Values = add/remove/set/reset/show

=item opspec
operations specification for file monitoring

this value is mandatory, if no value specified during initialization

Values = read,write,open,close,create,delete,rename,link,symlink,getattr,setattr,lookup,dircreate,dirdelete,dirrename

If this value is initialized during object initialization, then it is optional
otherwise it is mandatory

opspec if unspecified, attempt will be made to use the values from policy object and proceed further

if opspec provided and object value are null then error will be thrown and test failed

=item protocol
Protocol type specification
Values = nfs, cifs

if no value provided, then filer defaults will be assumed

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub monitor ($@)
{
	my $self = shift;
	my %parm = @_;

	# this is to keep the validate function happy
	if($parm{opspec} eq "" and $self->{POLICY}{MONITOR} ne "") { $parm{opspec} = $self->{POLICY}{MONITOR}; }
	if($parm{policy} eq "" and $self->{POLICY}{NAME} ne "") { $parm{policy} = $self->{POLICY}{NAME}; }

	if($parm{result} eq "") { $parm{result} = "pass"; }
	if($parm{conn} eq "") { $parm{conn} = $self->{CONN}; }
	if($parm{vfiler} eq "") { $parm{vfiler} = $self->{VFILER}{NAME}; }

	my @opspec = (); 
	foreach my $k (split /\,/, $parm{opspec}) 
	{ push @opspec, "$k";}
	$parm{opspec} = [@opspec];
	#############################################

#	eval {
#	%parm = validate(@options, {
#		conn => { type => SCALAR|OBJECT, optional => 1 },
#		policy => { type => SCALAR, regex =>  qr/(\w+)/, optional => 1 },
#		oper => { type => SCALAR, regex => qr/[add|remove|show|set|reset]/i },
#		opspec => { type => SCALAR|ARRAYREF, regex => qr/[((\w)+(\,(\w)+)+)?]/, optional => 1 },
##		opspec => { type => SCALAR|ARRAYREF, optional => 1 },
#		protocol => { type => SCALAR, regex => qr/[nfs|cifs]/i, optional => 1 },
#		vfiler => { type => SCALAR, regex => qr/(\w)+/, optional => 1},
#		result => { type => SCALAR, default => "PASS", regex =>  qr/[pass|fail]/i, optional => 1 },
#		expected => { type => SCALAR|ARRAYREF, optional => 1 }
#		});
#	};

#	if(!ValidateParams(@_)) {
#		logresult("fail", "Error at Monitor during validating input parameters:- $@ ");
#		return 1;
#	}

	my $str =  CLI_Expected_Str( 
		cmd_type => "policy_monitor", 
		version => $self->{FILER}{VERSION}, 
		policy => $parm{policy}, 
		cmd_code => "policy_monitor_noserver"
	);

	logcomment("CLI: expected = $parm{expected}");

	my %ptr = 
	CLI_Execute_Validate_Outputstr( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}" || $self->{VFILER}{NAME},
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd(
				vfiler => $parm{vfiler} || "",
				version => $self->{FILER}{VERSION}, 
				cmd_type => "monitor",
				policy => $parm{policy} || "", 
				oper => $parm{oper} || "", 
				opspec => $parm{opspec} || "", 
				proto => $parm{protocol} || $self->{POLICY}{PROTOCOL} 
			),

		expected => $parm{expected} || $str,
		result => $parm{result} || ""
	);



#	if( $parm{result} =~ /noresult/i )
#	{ logcomment("No Result"); } 

#	elsif($parm{result} =~ /FAIL/i) 
#	{
#		my $tmp = ($ptr{result} =~ /pass/i or $ptr{result} == 0)? 1 : 0;
#
#		if($tmp) 
#		{ 
#			# output and expected result: PASS-FAIL condition
#			logresult("fail", "Output did not match expected results. Expected to FAIL but PASSED
#						refer to above logs for further details, Details : $ptr{result}, $tmp");
#			return 1;
#		
#		}
#		else  # output failed as expected result
#		{
#			# output and expected result: FAIL-FAIL condition
#			logresult("pass", "Output did NOT match, expected to FAIL hence NEGATIVE case passed as expected, 
#				Expected to $ptr{result}, $tmp");
#		}
#
#	}
#	else 
#	{
#		if($ptr{result} =~ /fail/i or $ptr{result} == 1) #failure condition
#		{
#			# output and expected result: FAIL-PASS condition
#			logresult("fail", "Output did not match expected results. 
#							refer to above logs for further details, Details : $ptr{result}");
#				return 1;
#		} 
#		else # output passed as expected result 
#		{  
#			# output and expected result: PASS-PASS condition
#			logresult("pass", "Output matched expected results, 
#				Expected to $self-> {$ptr{result}");
#		}
#	}
#	return $ptr{result};
	#decide final result based on output validation and expected result as passed from test case
	return decideResult($ptr{result}, $parm{result});
}

sub decideResult
{

	my ($outputResult, $expectedResult) = @_;
	my $retResult = 0;
	
		if( $expectedResult =~ /noresult/i )
		{ 
			logcomment("No Result"); 
		} 
	
		elsif($expectedResult =~ /FAIL/i) 
		{
			my $tmp = ($outputResult =~ /pass/i or $outputResult == 0)? 1 : 0;
	
			if($tmp) 
			{ 
				# output and expected result: PASS-FAIL condition
			#	logresult("fail", "Output did not match expected results. Expected to FAIL but PASSED
			#				refer to above logs for further details, Details : $outputResult, $tmp");
                               logcomment("FAIL: Output did not match expected results. Expected to FAIL but PASSED
                                                        refer to above logs for further details, Details : $outputResult, $tmp");			
				return 1;
			
			}
			else  # output failed as expected result
			{
				# output and expected result: FAIL-FAIL condition
                        #       logresult("pass", "Output did NOT match, expected to FAIL hence NEGATIVE case passed as expected,
                        #               Expected to $outputResult, $tmp");
                                logcomment("PASS: Output did NOT match, expected to FAIL hence NEGATIVE case passed as expected,
                                        Expected to $outputResult, $tmp");

			}
	
		}
		else 
		{
			if($outputResult =~ /fail/i or $outputResult == 1) #failure condition
			{
				# output and expected result: FAIL-PASS condition
                        #       logresult("fail", "Output did not match expected results. Expected to PASS.
                        #                                       refer to above logs for further details, Details : $outputResult");
                                logcomment("FAIL: Output did not match expected results. Expected to PASS.
                                                        refer to above logs for further details, Details : $outputResult");

					return 1;
			} 
			else # output passed as expected result 
			{  
				# output and expected result: PASS-PASS condition
                        #       logresult("pass", "Output matched expected results.");
                                logcomment("PASS: Output matched expected results.");
				#Expected to $self-> {$outputResult}");
			}
		}
	
	return $retResult;
}

sub finalResult
{
        shift @_;
        my $res = shift;
#        logcomment ("result value is in CLI module is $res");
        if($res) {
                logresult("fail", "Test case failed. Refer to the logs to debug the issue.");
        } else {
                logresult("pass", "Test case passed.");
        }
}


=head1 NAME
connect - This API is used to make connection to the filer

=head2 DESCRIPTION
create connection based on the connection type

=head2 ROUTINES
	parameter definitions for type and file are used from the initialized object FPolicy/CLI
	$pol->{cli}->connect();
	
	$pol->{cli}->connect($type, $filer);

=over

=item type
Connection type to the filer
Mandatory parameter, when not initialized during new object

=item filer
Name of the filer
Mandatory parameter, when not initialized during new object

=cut

sub connect 
{
	my $self = shift;
	
	my $type = $self->{INTF};
	my $filer = $self->{FILER}{NAME};

	my $ptr = FPolicy::Utils::make_connection("conn_type=$type", "filer=$filer");

	#Save the information
	$self->{CONN} = ${$ptr}{conn};
	$self->{HOST} = ${$ptr}{host};
	logcomment("Host = $self->{HOST}, CONNREC = $self->{CONN}");
	
}


=head1 NAME
volume - This API is provision the volume configuration for policy on the filer

=head2 DESCRIPTION
This API is used to provision the volume definition for policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->volume("oper=add", "method=include", "result=FAIL");
	$pol->{cli}->volume("oper=add", "method=exclude", "volume=vol1", "result=PASS");
=over

=item oper
Mandatory parameter, indicates the type of operation to be performed on the volume
values = add,remove,set,reset,show

=item method
mandatory parameter, indicates the inclusion / exclusion list to work on for the volume
values = inclusion/exclusion

=item volume
Specifies the name of the volume 
optional, if the values was initialized during object instantiation

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub volume ($;@)
{
	my $self = shift;

	my %parm;
	eval {
	%parm = validate
	(
		@_, 
		{
		conn => { 	type => SCALAR|OBJECT, 
				default => $self->{CONN} 
			},

		policy => { 	type => SCALAR, 
				default => "$self->{POLICY}{NAME}", 
				regex =>  qr/(\w+)/ 
			},

		oper => { 	type => SCALAR, 
				regex => qr/[add|remove|show|set|reset]/i 
			},

		method => { 	type => SCALAR, 
				regex => qr/[include|exclude]/i 
			},

		volume => { 	type => SCALAR|ARRAYREF, 
#				regex => qr/(\w+)(,(\w)+)*/, 
				optional => 1 
			},

		vfiler => { 	type => SCALAR, 
				default => "$self->{VFILER}{NAME}", 
				regex => qr/(\w+)/, 
				optional => 1
			},

		result => { 	type => SCALAR, 
				default => "PASS", 
				regex => qr/[pass|fail]/i, 
				optional => 1 
			},

		expected => { 	type => SCALAR|HASHREF, 
				optional => 1 
			}
		}
	);

	}; # eval

	# check if there was a validation error
	if($@) { 
		logresult("fail", "Input parameter validation failed - $@"); 
		return 1; 
	};

	if($parm{expected} eq "") 
	{
		my $out = CLI_Expected_Str
			( 
				cmd_type => "policy_volume",
				version => $self->{FILER}{VERSION}, 
			);
		$parm{expected} = $out;
	}

	logcomment("CLI: expected = $parm{expected}{success}, $parm{expected}{failure}");


	my %ptr = 
	CLI_Execute_Validate_Outputstr
	( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd
			(
				vfiler => "$parm{vfiler}",
				cmd_type => "volume",
				policy => "$parm{policy}", 
				version => $self->{FILER}{VERSION}, 
				oper => "$parm{oper}", 
				method => "$parm{method}", 
				volume => "$parm{volume}" 
			),

		expected => $parm{expected},
		result => "$parm{result}"
	);



#	if( $parm{result} =~ /noresult/i )
#	 { logcomment("No Result"); } 
#
#	elsif( $parm{result} =~ /fail/i ) 
#	{
#		my $tmp = ($ptr{result} =~ /pass/i)? 1 : 0;
#
#		if( $tmp ) 
#		{ logresult("fail", "Failure"); }
#
#		else 
#		{ 
#			logresult("pass", "Success");  
#			return 0; 
#		}
#
#	}
#	elsif( $parm{result} =~ /pass/i and $ptr{result} == 0 ) 
#	{
#		logresult("pass", "Success");  
#		return 0; 
#	}
#	else 
#	{ logresult("fail", "Failure"); }
#
#	return $ptr{result};
	return decideResult($ptr{result}, $parm{result});

}





=head1 NAME
extension - This API is to provision the file extensions for the policy on the filer

=head2 DESCRIPTION
This API is used to provision file extension for the policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->extension("oper=add", "method=include");
	$pol->{cli}->extension("oper=add", "method=include", "result=FAIL");
	$pol->{cli}->extension("oper=add", "method=exclude", "extension=txt", "result=PASS");
=over

=item oper
Mandatory parameter to specify the type of operation performed on the file extension for the policy
Values add,remove,show,set,reset

=item method
Mandatory parameter to specify the inclusion/exclusion list on which the operation need to be performed for the policy
Values - include, exclude

=item extension
Specifies the value(s) of file extensions.
optional, if the values was initialized during object instantiation
Values - comma seperated list of extensions

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub extension ($;@)
{
	my $self = shift;
	my %parm = validate
	(
		@_, 
		{
		conn => { 	type => SCALAR|OBJECT, 
				default => $self->{CONN} 
			},

		policy => { 	type => SCALAR, 
				default => "$self->{POLICY}{NAME}", 
				regex =>  qr/(\w+)/ 
			},

		oper => { 	type => SCALAR, 
				regex => qr/[add|remove|show|set|reset]/i 
			},

		method => { 	type => SCALAR, 
				regex => qr/[inclusion|exclusion]/i 
			},

		extension => { 	type => SCALAR|ARRAYREF, 
				regex => qr/(\w+)(,(\w)+)*/, 
				optional => 1 
			},

		vfiler => { 	type => SCALAR, 
				default => "$self->{VFILER}{NAME}", 
				regex => qr/(\w+)/, 
				optional => 1
			},

		result => { 	type => SCALAR, 
				default => "PASS", 
				regex => qr/[pass|fail]/i, 
				optional => 1 
			},

		expected => { 	type => SCALAR|HASHREF, 
				default => CLI_Expected_Str
					(
					version => $self->{FILER}{VERSION}, 
					cmd_type => "policy_extension"
					),
				optional => 1 
			}
		}
	);



	# fill in the values to extension based on the type	
	if( $parm{extension} eq "" and 
		$parm{method} =~ /inc/i and 
		$self->{POLICY}{EXT_INC} ne "" )
	{ 
		$parm{extension} = "$self->{POLICY}{EXT_INC}"; 
	} 

	elsif ( $parm{extension} eq "" and 
		$parm{method} =~ /exc/i and 
		$self->{POLICY}{EXT_EXC} ne "" )
	{   
		$parm{extension} = "$self->{POLICY}{EXT_EXC}";   
	}

	# this should never happen
	elsif(	$parm{extension} eq "" and 
		 ( $self->{POLICY}{EXT_INC} eq "" and 
			( $self->{FLAGS}{result} =~ /FAIL/i  or $parm{result} =~ /fail/i) ) or
		 	( $self->{POLICY}{EXT_INC} eq "" and 
				( $self->{FLAGS}{result} =~ /FAIL/i or $parm{result} =~ /fail/i ) ) )
  	{ 
		logcomment("Invalid command code $parm{oper} with expected 
			FLAG $self->{FLAGS}{RESULT},$parm{result}");  
	}


	my %ptr = 
	CLI_Execute_Validate_Outputstr
	( 
		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",

		command => GenerateCmd
			(
				vfiler => "$parm{vfiler}",
				cmd_type => "extension",
				policy => "$parm{policy}", 
				version => $self->{FILER}{VERSION}, 
				oper => "$parm{oper}", 
				method => "$parm{method}", 
				extension => "$parm{extension}" 
			),
		expected => $parm{expected},
		result => "$parm{result}"
	);

#	if( $parm{result} =~ /noresult/i )
#	{ logcomment("No Result"); } 
#
#	elsif( $parm{result} =~ /fail/i ) 
#	{
#		my $tmp = ($ptr{result} =~ /pass/i)? 1 : 0;
#
#		if($tmp) 
#		{ logresult("fail", "Failure"); }
#
#		else 
#		{ 
#			logresult("pass", "Success");  
#			return 0; 
#		}
#
#	}
#	elsif( $parm{result} =~ /pass/i and $ptr{result} == 0 ) 
#	{
#		logresult("pass", "Success");  
#		return 0; 
#	} 
#	else 
#	{ logresult("fail", "Failure"); }
#
#	return $ptr{result};
	return decideResult($ptr{result}, $parm{result});

}

=head1 NAME
generic - This is a Generic API to perform all the CLI operations for the policy on the filer

=head2 DESCRIPTION
This API is used to perform all types of operations for policy on the filer using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	interface is still being worked on
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
#sub generic ($@)
#{
#	my $ref = shift;
#	my @options = @_;
#
#	#ordering is very much important here. Do not alter the ordering
#	@options = PreProcess_InputParams($ref, @options) ;
#
#	my %parm = make_NV(@options);
#
#	my $oper = "fpolicy_$parm{oper}_$parm{cmdtype}";
#
#	## overridding policy name from the arguments
#	## should be used only when policy name was not 
#	## provided during initialization
#	if (!defined $parm{policy} and $parm{policy} ne "") 
#	{
#		logcomment("Policy name was [$ref->{POLICY}{NAME}], now changed to $parm{policy}");
#		$ref->{POLICY}{NAME} = $parm{policy};
#	}
#
#
#	my $cmd = GenerateCmd("cmd_type=$parm{cmd_type}", @options);
#
#	# Input command is stored for further processing
#	$ref->{CMD} = $cmd;
#
#	# expected output is gathered from the user input
#	# otherwise, default generated expected string is used
#	if($parm{expected} ne "") {
#		logcomment("Overriding User defined Expected value $parm{expected}");
#		$ref->{expected} = $parm{expected};
#	} else {
#		my @expected = CLI_Expected_Str($oper, $ref);
#		$ref->{expected} = "@expected";
#	}
#
#
#	my $stat = CLI_Execute_Validate_Outputstr($ref);
#	if($stat) {
#		logresult("fail", "REASON = OUTPUT not as expected - $ref->{expected}");
#	}
#	return $stat;
#}


=head1 NAME
testme - This API is to test all aspects of this module

=head2 DESCRIPTION
This API should be used only to test the internal APIs and its integrity with other APIs or external modules using a seperate script

=head2 ROUTINES
	$pol->{cli}->testme()

	ntest FPolicy/test_cli.thpl FILER=nss-u50 CLIENT=nss-s37 FILER_IP=172.17.180.112 PKGNAME=CLI SUBNAME=testme POLICY=pol1

=over

=cut
sub testme () 
{

	package main;
	our ($FILER, $FILER_CONN, $CLIENT, $POLICY, $VFILER);

	param('FILER',       '-mesg',    "Filer to test");
	param('FILER_CONN','-default',"#rsh");
        param('POLICY',   '-default', 'pol1');
        param('VFILER',   '-default', 'vfiler0');
        param('CLIENT',   '-mesg', 'CLIENT');
	use FPolicy;
	my $pol = new FPolicy("policy=pol1", "vfiler=vfiler0", "filer=$FILER", 
#			"monitor=read,write,open,close", 
#			"vol_inc=vol0", "vol_exc=", 
#			"ext_inc=txt", "ext_exc=ppt", 
#			"pserver_pri=10.10.1.1", "pserver_pri=20.1.1.1",
			"client=$CLIENT"
		);

	$pol->{cli}->connect();
	$pol->{cli}->create();
	$pol->{cli}->create(result=>"FAIL");
	$pol->{cli}->create(vfiler => "vfiler1", policy => "one");
	$pol->{cli}->enable();
	$pol->{cli}->enable(result => "PASS");
	$pol->{cli}->verify(policystate => "enabled");
	$pol->{cli}->verify(policystate => "enabled", result => "PASS");
	$pol->{cli}->disable();
	$pol->{cli}->disable(result => "PASS");
	$pol->{cli}->verify(policystate => "Disabled", result => "FAIL");
	$pol->{cli}->show();
#	$pol->{cli}->monitor(result => "FAIL");
#	$pol->{cli}->monitor(oper => "add");
#	$pol->{cli}->monitor(oper => "add", result => "PASS");
	$pol->{cli}->monitor(oper => "add", opspec => "link,symlink");
	$pol->{cli}->verify(monitor => "link,symlink");
	$pol->{cli}->monitor(oper => "add", opspec => "link,symlink", protocol => "nfs");
	$pol->{cli}->verify(monitor => "link,symlink");
	$pol->{cli}->monitor(oper => "remove", opspec => "link,symlink", protocol => "nfs");
	$pol->{cli}->verify(monitor => "link,symlink");
	$pol->{cli}->monitor(oper => "set", opspec => "link,symlink", protocol => "nfs");
	$pol->{cli}->verify(monitor => "link,symlink");
	#$pol->{cli}->monitor(oper => "reset");
	$pol->{cli}->monitor(oper => "show");

#	$pol->{cli}->volume(oper => "add", method => "include");
	$pol->{cli}->volume(oper => "add", method => "include", volume => "vol0");
	$pol->{cli}->verify(vol_inc => "vol0");
	$pol->{cli}->volume(oper => "add", method => "include", volume => "vol0", result => "PASS");
	$pol->{cli}->volume(oper => "add", method => "exclude", volume => "vol1");
	$pol->{cli}->verify(vol_exc => "vol1");
	$pol->{cli}->volume(oper => "add", method => "exclude", volume => "vol1", result => "PASS");
	$pol->{cli}->volume(oper => "remove", method => "exclude", volume => "vol1");
	$pol->{cli}->verify(vol_exc => "-blind");
	$pol->{cli}->volume(oper => "remove", method => "exclude", volume => "vol1", result => "fail");
	$pol->{cli}->volume(oper => "set", method => "exclude", volume => "vol1", result => "pass");
	$pol->{cli}->verify(vol_exc => "vol1");
	$pol->{cli}->volume(oper => "set", method => "exclude", volume => "vol1");
	$pol->{cli}->volume(oper => "reset", method => "exclude");
	$pol->{cli}->verify(vol_exc => "-blind");
	$pol->{cli}->volume(oper => "reset", method => "include", result => "pass");
	$pol->{cli}->verify(vol_exc => "-blind");
	$pol->{cli}->volume(oper => "show", method => "include");
	$pol->{cli}->volume(oper => "show", method => "include", result => "pass");
	$pol->{cli}->volume(oper => "show", method => "exclude", result => "pass");
	$pol->{cli}->volume(oper => "show", method => "exclude");
	
	
#	$pol->{cli}->extension(oper => "add", method => "include");
#	$pol->{cli}->extension(oper => "add", method => "include", result => "pass");
	$pol->{cli}->extension(oper => "add", method => "include", extension => "ppt");
	$pol->{cli}->verify(ext_inc => "ppt");
	$pol->{cli}->extension(oper => "add", method => "include", extension => "ppt", result => "PASS");
	$pol->{cli}->extension(oper => "add", method => "exclude", extension => "txt");
	$pol->{cli}->verify(ext_exc => "txt");
	$pol->{cli}->extension(oper => "add", method => "exclude", extension => "txt", result => "PASS");

#	$pol->{cli}->extension(oper => "remove", method => "include");
#	$pol->{cli}->extension(oper => "remove", method => "include", result => "pass");
	$pol->{cli}->extension(oper => "remove", method => "include", extension => "ppt");
	$pol->{cli}->verify(ext_inc => "");
	$pol->{cli}->extension(oper => "remove", method => "include", extension => "ppt", result => "fail");
	$pol->{cli}->extension(oper => "set", method => "include", extension => "ppt");
	$pol->{cli}->verify(ext_inc => "ppt");
	$pol->{cli}->extension(oper => "set", method => "include", extension => "ppt", result => "fail");
	$pol->{cli}->extension(oper => "reset", method => "include", extension => "ppt");
	$pol->{cli}->verify(ext_inc => "");
	$pol->{cli}->extension(oper => "reset", method => "include", extension => "ppt", result => "pass");

	$pol->{cli}->destroy();
	$pol->{cli}->destroy(result => "PASS");

	my $pol1 = new FPolicy("policy=pol2", "vfiler=vfiler0", "filer=$FILER", 
			"monitor=read,write,open,close", 
			"vol_inc=vol0", "vol_exc=", 
			"ext_inc=txt", "ext_exc=ppt", 
			"pserver_pri=10.10.1.1", "pserver_pri=20.1.1.1",
			"client=$CLIENT"
		);

	$pol1->{cli}->connect();
	$pol1->{cli}->create();
#	$pol1->{cli}->create(result=>"FAIL");
#	$pol1->{cli}->create(vfiler => "vfiler1", policy => "one");
#	$pol1->{cli}->enable();
#	$pol1->{cli}->enable(result => "PASS");
#	$pol1->{cli}->verify(policystate => "enabled");
#	$pol1->{cli}->verify(policystate => "enabled", result => "PASS");
#	$pol1->{cli}->disable();
#	$pol1->{cli}->disable(result => "PASS");
#	$pol1->{cli}->verify(policystate => "Disabled", result => "FAIL");
#
#	$pol1->{cli}->monitor(result => "FAIL");
	$pol1->{cli}->monitor(oper => "add");
	$pol1->{cli}->monitor(oper => "add", result => "PASS");
	$pol1->{cli}->monitor(oper => "add", opspec => "link,symlink");
	$pol1->{cli}->monitor(oper => "add", opspec => "link,symlink", protocol => "nfs");
	$pol1->{cli}->monitor(oper => "remove", opspec => "link,symlink", protocol => "nfs");
	$pol1->{cli}->monitor(oper => "set", opspec => "link,symlink", protocol => "nfs");
	#$pol1->{cli}->monitor(oper => "reset");
	$pol1->{cli}->monitor(oper => "show");

	$pol1->{cli}->volume(oper => "add", method => "include");
	$pol1->{cli}->volume(oper => "add", method => "include", volume => "vol0");
	$pol1->{cli}->volume(oper => "add", method => "include", volume => "vol0", result => "PASS");
	$pol1->{cli}->volume(oper => "add", method => "exclude", volume => "vol1");
	$pol1->{cli}->volume(oper => "add", method => "exclude", volume => "vol1", result => "PASS");
	$pol1->{cli}->volume(oper => "remove", method => "exclude", volume => "vol1");
	$pol1->{cli}->volume(oper => "remove", method => "exclude", volume => "vol1", result => "fail");
	$pol1->{cli}->volume(oper => "set", method => "exclude", volume => "vol1", result => "pass");
	$pol1->{cli}->volume(oper => "set", method => "exclude", volume => "vol1");
	$pol1->{cli}->volume(oper => "reset", method => "exclude");
	$pol1->{cli}->volume(oper => "reset", method => "include", result => "pass");
	$pol1->{cli}->volume(oper => "show", method => "include");
	$pol1->{cli}->volume(oper => "show", method => "include", result => "pass");
	$pol1->{cli}->volume(oper => "show", method => "exclude", result => "pass");
	$pol1->{cli}->volume(oper => "show", method => "exclude");
	
	
	$pol1->{cli}->extension(oper => "add", method => "include");
	$pol1->{cli}->extension(oper => "add", method => "include", result => "pass");
	$pol1->{cli}->extension(oper => "add", method => "include", extension => "ppt");
	$pol1->{cli}->extension(oper => "add", method => "include", extension => "ppt", result => "PASS");
	$pol1->{cli}->extension(oper => "add", method => "exclude", extension => "txt");
	$pol1->{cli}->extension(oper => "add", method => "exclude", extension => "txt", result => "PASS");

#	$pol1->{cli}->extension(oper => "remove", method => "include");
#	$pol1->{cli}->extension(oper => "remove", method => "include", result => "pass");
	$pol1->{cli}->extension(oper => "remove", method => "include", extension => "ppt");
	$pol1->{cli}->extension(oper => "remove", method => "include", extension => "ppt", result => "fail");
	$pol1->{cli}->extension(oper => "set", method => "include", extension => "ppt");
	$pol1->{cli}->extension(oper => "set", method => "include", extension => "ppt", result => "fail");
	$pol1->{cli}->extension(oper => "reset", method => "include", extension => "ppt");
	$pol1->{cli}->extension(oper => "reset", method => "include", extension => "ppt", result => "pass");

	$pol1->{cli}->destroy();
	$pol1->{cli}->destroy(result => "PASS");

	# Another Variation of the usage


	my $cpol  =  new FPolicy::CLI("policy=pol3", "vfiler=vfiler0", "filer=$FILER", "client=$CLIENT");
	$cpol->connect();
	$cpol->create();
	$cpol->create(result => "FAIL");
#	$cpol->create(VFILER => "vfiler1", policy => "one");
	$cpol->enable();
	$cpol->enable(result => "PASS");
#	$cpol->verify(policystate => "enabled");
#	$cpol->verify(policystate => "enabled", result => "PASS");
	$cpol->disable();
	$cpol->disable(result => "PASS");
#	$cpol->verify(policystate => "Disabled", result => "FAIL");
#
##	$cpol->monitor(result => "FAIL");
#	$cpol->monitor(oper => "add");
#	$cpol->monitor(oper => "add", result => "PASS");
	$cpol->monitor(oper => "add", opspec => "link,symlink");
	$cpol->monitor(oper => "add", opspec => "link,symlink", protocol => "nfs");
	$cpol->monitor(oper => "remove", opspec => "link,symlink", protocol => "nfs");
	$cpol->monitor(oper => "set", opspec => "link,symlink", protocol => "nfs");
	$cpol->monitor(oper => "reset");
	$cpol->monitor(oper => "show");
#
#	$cpol->volume(oper => "add", method => "include");
	$cpol->volume(oper => "add", method => "include", volume => "vol0");
	$cpol->volume(oper => "add", method => "include", volume => "vol0", result => "PASS");
	$cpol->volume(oper => "add", method => "exclude", volume => "vol1");
	$cpol->volume(oper => "add", method => "exclude", volume => "vol1", result => "PASS");
	$cpol->volume(oper => "remove", method => "exclude", volume => "vol1");
	$cpol->volume(oper => "remove", method => "exclude", volume => "vol1", result => "fail");
	$cpol->volume(oper => "set", method => "exclude", volume => "vol1", result => "pass");
	$cpol->volume(oper => "set", method => "exclude", volume => "vol1");
	$cpol->volume(oper => "reset", method => "exclude");
	$cpol->volume(oper => "reset", method => "include", result => "pass");
	$cpol->volume(oper => "show", method => "include");
	$cpol->volume(oper => "show", method => "include", result => "pass");
	$cpol->volume(oper => "show", method => "exclude", result => "pass");
	$cpol->volume(oper => "show", method => "exclude");
#	
#	
#	$cpol->extension(oper => "add", method => "include");
#	$cpol->extension(oper => "add", method => "include", result => "pass");
	$cpol->extension(oper => "add", method => "include", extension => "ppt");
	$cpol->extension(oper => "add", method => "include", extension => "ppt", result => "PASS");
	$cpol->extension(oper => "add", method => "exclude", extension => "txt");
	$cpol->extension(oper => "add", method => "exclude", extension => "txt", result => "PASS");
#
#	$cpol->extension(oper => "remove", method => "include");
#	$cpol->extension(oper => "remove", method => "include", result => "pass");
	$cpol->extension(oper => "remove", method => "include", extension => "ppt");
	$cpol->extension(oper => "remove", method => "include", extension => "ppt", result => "fail");
	$cpol->extension(oper => "set", method => "include", extension => "ppt");
	$cpol->extension(oper => "set", method => "include", extension => "ppt", result => "fail");
	$cpol->extension(oper => "reset", method => "include", extension => "ppt");
	$cpol->extension(oper => "reset", method => "include", extension => "ppt", result => "pass");
#
	$cpol->destroy();
	$cpol->destroy(result => "FAIL");


}

### Added by prashant for Native file blocking

=head1 NAME
enable - This API is to enable the native file blocking on the specified policy

=head2 DESCRIPTION
This API is used to enable native file blocking on the policy using CLI
Connection to filer will be initiated only for the first API usage

=head2 ROUTINES
	$pol->{cli}->options(mthod => "required", value => "on");
	$pol->{cli}->options("result=FAIL");

Disable result logging mechanism by this module
	$pol->{cli}->options("result=noresult");
=over

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut

sub options ($;@)
{
	my $self = shift;
	my %NV = @_;
	my %parm = validate 
	(
		@_,
		{
		conn => { 	type => SCALAR|OBJECT, 
				default => $self->{CONN}
			},

		policy => { 	type => SCALAR, 
				default => "$self->{POLICY}{NAME}", 
				regex =>  qr/(\w+)/, 
				optional => 1
			 },

		vfiler => { 	type => SCALAR, 
				default => "$self->{VFILER}{NAME}", 
				regex => qr/(\w+)/, 
				optional => 1
			},
			
		method => {	type => SCALAR,
				#regex => qr/[required|secondary_servers|monitor_ads]/i  # Appended monitor_ads by Rajaram
			  },
			   
	        value  => {	type => SCALAR,
	        		#regex => qr/[on|off|*+]/i
	        		qr/(\w+)/, 
	        	  },
	        	  	

		result => { 	type => SCALAR, 
				default => "PASS", 
				regex =>  qr/[pass|fail]/i, 
				optional => 1
			},

		expected => { 	type => SCALAR|HASHREF, 
				default => CLI_Expected_Str(
					version => $self->{FILER}{VERSION}, 
					cmd_type => "policy_options"
					 ),
				optional => 1 
			}
		}
	);

	# result is in the form of {result => true/false/0/1 , output => "obtained output string" }
	my %ptr = 
	CLI_Execute_Validate_Outputstr( 

		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",
		command => GenerateCmd(
				vfiler => "$parm{vfiler}",
				cmd_type => "options",
				version => $self->{FILER}{VERSION}, 
				policy => $parm{policy},
				method => $parm{method},
				value => $parm{value}
				),
	
		expected => $parm{expected},
		result => $parm{result}
	);

	return decideResult($ptr{result}, $parm{result});

}

sub snapId ($;@)
{
	my $self = shift;
	my %NV = @_;
	my %parm = validate 
	(
		@_,
		{
		conn => { 	type => SCALAR|OBJECT, 
				default => $self->{CONN}
			},

		vfiler => { 	type => SCALAR, 
				default => "$self->{VFILER}{NAME}", 
				regex => qr/(\w+)/, 
				optional => 1
			},
			
#		method => {	type => SCALAR,
#				regex => qr/[status]/i
#			  },
		volume => {	type => SCALAR, 
				regex => qr/(\w+)/,
				default => "vol0"			
			},

		result => { 	type => SCALAR, 
				default => "PASS", 
				regex =>  qr/[pass|fail]/i, 
				optional => 1
			},

		expected => { 	type => SCALAR|HASHREF, 
				default => CLI_Expected_Str(
					version => $self->{FILER}{VERSION}, 
					cmd_type => "snap_status"
					 )
#				optional => 1 
			}
		}
	);

	my %ptr = 
	CLI_Execute_Validate_Outputstr( 

		conn => $parm{conn}, 
		intf => "cli",
		vfiler => "$parm{vfiler}",
		filer => "$self->{FILER}{NAME}",
#		command => GenerateCmd(
#				vfiler => "$parm{vfiler}",
#				command => "snap",
#				cmd_type => "status",
#				version => $self->{FILER}{VERSION}, 
#				volume => $parm{volume},
#				),	
		command => " priv set advanced; snap status -V $parm{volume}",
		expected => $parm{expected},
		result => $parm{result}
	);

	return ( Get_SnapDetails($ptr{output}) );

}

sub statsShow()
{
	my $self = shift;
	my %args = @_;
	
	my $filer = $self->{CONN};
	my (undef, undef, $stats) = say($filer, "priv set diag; stats show $args{counter}");
	my $ret = stats_parse($stats);
	
	return $ret;
}

1;
