# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/FPolicy/MinWin.pm#1 $
##
## @summary Functions to interact with a minwin-compatible test client machine.
##
## @description
##	This library interacts with "minwin" clients.
##
## @synopsis
##	use minwin;
##
##	minwin_compatible($CLIENT);
##
## @usage
##
## <hr><b>minwin_compatible()</b></hr>
##
## <p>This function determines whether a given windows client is
## minwin-compatible.
##
## @status	in-progress
## @author	tjt@netapp.com
## @change  migrate from main

package FPolicy::MinWin;

use strict;
use Tharn;
use Socket;
use Sys::Hostname;
use Module::Common;

BEGIN
{
	Module::Common::setup_export_tags(
		'basic' => [ qw(
			minwin_compatible
			map_drive
			unmap_drive
			execute_cmd
			execute_cmd_special
		) ]
	);
}

my $cifs_timeout = $main::TIMEOUT || 30;

#----------------------------
#   functions
#---------------------------

## @name ::minwin_xshare;
## @usage minwin_xshare()
## @synopsis
##    Returns the UNC name for the X: drive of minwin clients.
##    The value depends on which network you are in.
sub minwin_xshare {

	my ($winconnspec) = @_;
	my $ip = $winconnspec->{hostp}->{default_ip};
	my($xshare) = "";
    my $xuser = undef;
    my $xpassword = undef;

	if ( $ip =~ /^10\.97\.*/ ) {        # NANE
		$xshare = "\\\\orca7.nane.netapp.com\\localtest\\arch\\win32";
	} elsif ( $ip =~ /^10\.60\.*/ ) {   # RTP
		$xshare = "\\\\agate.hq.netapp.com\\localtest\\arch\\win32";
        $xuser = "agate\\minwin";
        $xpassword = "qa4testing";
	} elsif ( $ip =~ /^10\.98\.*/ ) {   # Pittsburgh
		$xshare = "\\\\spin-filer.spin.eng.netapp.com\\localtest\\arch\\win32";
        $xuser = "spin-filer\\minwin";
        $xpassword = "qa4testing";
	} else {
		# Xshare defaults to Bangalore's maglite i.e., btcinf02
		$xshare = "\\\\btcinf02.eng.btc.netapp.in\\localtest\\arch\\win32";
        $xuser = "btcinf02\\minwin";
        $xpassword = "qa4testing";
	}
	return($xshare, $xuser, $xpassword);
}

## @name ::minwin_compatible;
## @usage minwin_compatible($clientName)
## @synopsis
##    Returns 1 if the named windows client machine is minwin-compatible,
##    which means that it has an rshd server running, and now has
##    the X: share mounted containing various test utilities.
sub minwin_compatible {
    
	my($winclient) = shift;

	if ( ! defined($winclient) ) {
		logcomment("minwin_compatible warning: Undefined value for winclient argument");
		return undef;
	}

	my($spec) = $winclient;
	$spec .= ";hosttype=windows"
		unless ($winclient =~ /;hosttype=windows/);
		
	my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

	# Here's where all the magic information about the X drive is derived
	my $xdrive = "X";
	my($output);
	my($net_use_status);
	my($xshare, $xuser, $xpassword) = minwin_xshare($c);
	$xuser = "netapp\\ncqa" if (!defined($xuser));
	$xpassword = "netcache" if (!defined($xpassword));

	if ( $xshare eq "" ) {
		my $err = "minwin_compatible() doesn't know what "
			."share to use for the $xdrive: drive, in your network";
		logcomment("minwin_compatible warning: $err");
		return undef;
	}

	my($num) = rand(999);	# We just need a unique string of some sort
	(undef, undef, $output) = say($c, "cmd /c echo $num", "-timeout", $cifs_timeout, "-eval_error");
	if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("minwin_compatible warning: TIMEOUT during echo");
	}

	if ( $output !~ /$num/ ) {
		logcomment("minwin_compatible warning: The client $winclient is "
			."not responding to rsh, its rshd needs resetting.");
		return undef;
	}

	# This is an attempt to address the situation for burt 116148.
    # The basic problem is that attempting to do a net use for a device that 
    # is already configured, will cause a prompt that we cannot answer.
    # The previous approach was to delete any existing X: drive
    # This got us in trouble when more than one subtest was running
    # since the second one would delete the configuration 
    # out from under the first one.
    # The new approach is to query the existing definition using 'net use'
    (undef, undef, $net_use_status) = say($c, "net use", "-eval_error");
    if ($@ && $@ =~ /Err=TIMEOUT/i) {
        logcomment("minwin_compatible warning: "
            ."TIMEOUT during net use $xdrive:");
    }
    # Then, only if the existing connection is disconnected or Unavailable
    # we delete it.
    my $need_delete = 0;
    my $need_connect = 1;
    my $did_stale_map_reconnect = 0;
	if ( $net_use_status =~ /Unavailable\s+$xdrive:\s+|Disconnected\s+$xdrive:\s+/i ) {
        # The drive is configured, but it is presently disconnected
        $need_delete = 1;
	logcomment("minwin_compatible warning: "
	    . "'$xdrive' is in a bad state and will be unmapped.");
    } elsif ( $net_use_status !~ /\s+$xdrive:\s+(\S+)\\localtest\\arch\\win32/i ) {
        $need_connect = 1;
	logcomment("minwin_compatible warning: "
	    . "'$xdrive' is mapped to '$1' which doesn't match '$xshare'.\n");
    } else {
	# if the drive map actually does look good, don't make the call to mount it
	$need_connect = 2;
    }
    
    while ($need_delete or $need_connect) {
        if ($need_delete) {
            $need_delete = 0;
            (undef, undef, $output) = say($c, "net use $xdrive: /DELETE", "-timeout", $cifs_timeout, "-eval_error");
            if ($@ && $@ =~ /Err=TIMEOUT/i) {
                logcomment("minwin_compatible warning: "
                    ."TIMEOUT during net use xdrive /DELETE");
            }
            $need_connect = 1;   # Force a reconnect next
        }
        # Next, we (re)establish the net use
        # if the drive is not presently configured, or it is and we just want to check it
        if ($need_connect == 1) {
            $need_connect = 0;
	    (undef, undef, $output) = say($c, "net use $xdrive: $xshare $xpassword /persistent:NO /USER:$xuser", "-timeout", $cifs_timeout, "-eval_error");
	    if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("minwin_compatible warning: "
		    ."TIMEOUT during net use xdrive");
	    }

            # The output may or may not contain the "local device name is
            # already in use" error - oddly, it's inconsistent.
            # But this problem always contains the "error.*85", so we key off that.

            # The 1909 error is intermittent, with increasing likelihood on busy systems
            # However, 3 bad attempts may lock out users - don't loop on retry here
            if ( $output =~ /System.*error.*1909/ ) {
                # Let's try it without the defined password in case access is already allowed
                # (This should be available to real user ids)
                (undef, undef, $output) = say($c, "net use $xdrive: $xshare /persistent:NO", "-timeout", $cifs_timeout, "-eval_error");
                if ($@ && $@ =~ /Err=TIMEOUT/i) {
                    logcomment("minwin_compatible warning: "
                        ."TIMEOUT during net use xdrive");
                }
                if ($output =~ /System.*error.*1909/) {
                    logcomment("minwin_compatible warning: "
                        ."Unable to map drive=$xdrive to "
                        ." share=$xshare : $output");
                    return undef;
                }
            }
            if ( $output =~ /successfully/ ) {
                # Verify the new mapping
                (undef, undef, $net_use_status) = say($c, "net use","-timeout", $cifs_timeout, "-eval_error");
                # fall through
            } elsif ( $output =~ /System.*error.*85/ ) {
                # Somehow, this drive became mapped since we looked a minute ago!
                # Let's verify it now has the right mapping
                (undef, undef, $net_use_status) = say($c, "net use","-timeout", $cifs_timeout, "-eval_error");
                if ($@ && $@ =~ /Err=TIMEOUT/i) {
                    logcomment("minwin_compatible warning: "
                        ."TIMEOUT during net use $xdrive:");
                }

                if ( $net_use_status !~ /\s+$xdrive:\s+\S+/ ) {
                    # It went away? Try to map the drive again
                    (undef, undef, $output) = say($c, "net use $xdrive: $xshare $xpassword /USER:$xuser", "-timeout", $cifs_timeout, "-eval_error");
                    if ($@ && $@ =~ /Err=TIMEOUT/i) {
                        logcomment("minwin_compatible warning: "
                            ."TIMEOUT during net use $xdrive $xshare");
                    }
                    if ( $output !~ /successfully/ ) {
                        logcomment("minwin_compatible warning: "
                            ."Unable to map $xdrive "
                            ."share: $output");
                        return undef;
                    }
                    # Verify the new mapping
                    (undef, undef, $net_use_status) = say($c, "net use","-timeout", $cifs_timeout, "-eval_error");
                }


            } elsif ( $output =~ /System.*error.*53/ ) {
                logcomment("minwin_compatible warning: "
                    ."Unable to map drive=$xdrive to "
                    ." share=$xshare : $output");
                return undef;
            } else {
                logcomment("minwin_compatible warning: "
                    ."Unexpected output when mapping "
                    ."$xdrive share: $output");
                return undef;
            }
        } 
	# if the status is OK or simply blank 
        if ( $net_use_status =~ /(:?OK\s+|\s+)$xdrive:\s+(\S+\\localtest\\arch\\win32)/m ) {
            my $present_map = $1;
            # We have verified that the directory has really been mapped
            # Let's make sure we can find some expected directories there
            # (There have been times where they do not appear!)
            (undef, undef, $output) = say($c, "cmd /c dir /on $xdrive:\\", "-timeout", $cifs_timeout, "-eval_error");
            if ($@ && $@ =~ /Err=TIMEOUT/i && (!defined($output) or ($output !~ /\s+bin\s+/ and $output !~ /\s+nate\s+/ and $output !~ /\s+perl\s+/)) ) {
                   logcomment("minwin_compatible warning: TIMEOUT during dir");
            }
            logcomment("Windows client $winclient has been declared "
                ."minwin-compatible and the $xdrive: drive is mapped to $present_map");
	    if ( $output =~ /The specified network name is no longer available/i ) {
		logcomment("minwin_compatible warning: "
		    . "Drive '$xdrive' was mapped and listed as OK, but really isn't.  Forcing a Delete and remap.\n");
		$need_delete = 1;
		$need_connect = 1;
		next; # make sure we don't return yet
	    } elsif ( $output !~ /\s+bin\s+/ and $output !~ /\s+nate\s+/ and $output !~ /\s+perl\s+/) {
                logcomment("minwin_compatible warning: "
                    . "Drive $xdrive was mapped but it does not have the expected directories"
                    );
                return undef;
            } else {
                (undef, undef, $output) = say($c, "cmd /c dir /on $xdrive:\\bin\\sio.exe", "-timeout", $cifs_timeout, "-eval_error");
                if ($@ && $@ =~ /Err=TIMEOUT/i && (!defined($output) or ($output !~ /\s+\d+\s+sio.exe/i )) ) {
                       logcomment("minwin_compatible warning: TIMEOUT during dir");
                } elsif($output =~ /\s+(\d+)\s+sio.exe/i ) {
                    if ($1 eq "0") {
                        # We have a stale copy. Drop the map and try again, but only once
                        if (!$did_stale_map_reconnect) {
                            $did_stale_map_reconnect = 1;
                            $need_connect = 1;
                        }
                    }
                }
            }
            return(1);
        } elsif ( $net_use_status !~ /\s+$xdrive:\s+\S+/ ) {
            # Second attempt to map it also failed, perhaps there is a script error here?
            logcomment("minwin_compatible warning: "
                ."Drive $xdrive could not be mapped: $net_use_status"
                );
            return undef;
        } else {
            # It is mapped somewhere that is not compatible
            logcomment("minwin_compatible warning: "
                ."Drive $xdrive is already mapped: $net_use_status"
                );
            return undef;
        }
    }
}

## @name ::map_drive;
## @usage map_drive($clientName,$driveLetter,$sharePath, %opts)
## @synopsis
##    Map the named share to the named drive on the named windows client.
## @param clientName required
## @param driveLetter required
## @param sharePath required
## @param opts optional
##	user and passwd credentials used to map drive maybe specified
##	user => "myuser", passwd => "mypasswd"
##	mapped drive can be PERSISTENT, defaults to NO
##	persistent => "YES" or "NO"
sub map_drive {
	
	my($winclient, $drive, $share, %opts) = @_;

	my($spec) = $winclient;
	$spec .= ";hosttype=windows"
		unless ($winclient =~ /;hosttype=windows/);
	my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

	my($output);
	(undef, undef, $output) = say($c, "cmd /c net use", "-eval_error");
	if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("map_drive error: "
			."TIMEOUT during net use in map_drive");
	}
	logcomment("currently mapped drives: $output");


	if ( $drive !~ /^[*A-Za-z]:?$/ ) {
		logresult('WARN',"Bad drive value ($drive) given to map_drive()");
		return undef;
	}

	# Add ":" if it's not there.
	if ( $drive =~ /^[A-Za-z]$/ ) {
		$drive = $drive.":";
	}

	my $cmd = "cmd /c net use $drive $share ";
	$cmd .= $opts{passwd} . " /USER:" . $opts{user}
		if ($opts{passwd} && $opts{user});

	$opts{persistent} = $opts{persistent} =~ /YES/i ? "YES" : "NO";
	$cmd .= " /PERSISTENT:".$opts{persistent};

	(undef, undef, $output) = say($c, $cmd, "-eval_error");
	if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("map_drive error: "
			."TIMEOUT during net use drive in map_drive");
	}
	logresult('WARN', $output)
		if ($output =~ /err|invalid/i);

	# This is silly, but works.
	$output =~ s/\r//g;
	$output =~ s/\n//g;

	# Different output from different versions of Windows.

	if ( $output !~ /successfully/ && $output !~ /now connected/ ) {
		logresult('WARN',"Unable to map $drive from $share : $output");
		return undef;
	} elsif ( $output =~ /now connected/ ) {
		if ( "$drive" eq "*" ) {
			# If we've let the windows client pick the drive
			# letter, we need to figure out which one was selected.

			# This is silly, but works.
			$output =~ s/.*Drive //;
			$output =~ s/:.*//;
			$drive = $output;
		}
	} else {
		if ( "$drive" eq "*" ) {
			# If we've let the windows client pick the drive
			# letter, we need to figure out which one was selected.
			$output =~ /Drive ([A-Z]):.*/;
			$drive = $output;
		}
	}
	$drive =~ s/://;
	logcomment("minwin_map has mapped $share as drive $drive");
	return($drive);
}

## @name ::unmap_drive;
## @usage unmap_drive($clientName,$driveLetter)
## @synopsis
##    Unmap the named drive from the named windows client.
sub unmap_drive {
	
	my($winclient) = shift;
	my($drive) = shift;

	if ( $drive !~ /^[A-Za-z]:?$/ ) {
		logresult('WARN',"Bad drive value ($drive) given to unmap_drive()");
		return undef;
	}

	my($spec) = $winclient;
	$spec .= ";hosttype=windows"
		unless ($winclient =~ /;hosttype=windows/);
	my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

	# Add ":" if it's not there.
	if ( $drive =~ /^[A-Za-z]$/ ) {
		$drive = $drive.":";
	}

	my($output);
	(undef, undef, $output) = say($c, "cmd /c net use $drive /delete", "-eval_error");
	if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("TIMEOUT during net use /delete in unmap_drive");
	}

	if ( $output !~ /successfully/ ) {
		logresult('WARN',"Unexpected output when unmapping "
				."$drive: $output");
		return("");
	} else {
		logcomment("minwin_mount has unmapped drive $drive");
		return($drive);
	}
}

## @name ::execute_cmd;
## @usage execute_cmd($clientName,$cmd)
## @synopsis
##    Run a command on the named minwin-compatible client.
sub execute_cmd {
	
	my($winclient) = shift;
	my($cmd) = shift;

	my($spec) = $winclient;
	$spec .= ";hosttype=windows"
		unless ($winclient =~ /;hosttype=windows/);
	my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

	my($output);
	
	# We use the runcmd.pl utility in order to make sure
	# that the stderr output of the command is completely captured.

	(undef, undef, $output) = say($c,
			"x:\\perl\\bin\\perl.exe x:\\bin\\runcmd.pl $cmd",
			"-eval_error");
	if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("TIMEOUT during command execution: $cmd");
	}
	return($output);
}

## @name ::execute_cmd_special;
## @usage execute_cmd_special($clientName,$cmd)
## @synopsis
##    Run a command on the named minwin-compatible client. It is identical to 
##    execute_cmd except that the output of the command(stdout, stderr) is NOT 
##    redirected to special files for printing. 
##    Commands like echo "some_text" > filename, do not have the desired
##    output with the execute_cmd. This is useful for testing appending, 
##    overwriting to files.
sub execute_cmd_special {
	
	my($winclient) = shift;
	my($cmd) = shift;

	my($spec) = $winclient;
	$spec .= ";hosttype=windows"
		unless ($winclient =~ /;hosttype=windows/);
	my($c) = connect("$spec#rsh", "-timeout", $cifs_timeout);

	my($output);
	
	(undef, undef, $output) = say($c,
			"cmd.exe /c $cmd");
	if ($@ && $@ =~ /Err=TIMEOUT/i) {
		logcomment("TIMEOUT during command execution: $cmd");
	}
	return($output);
}
1;
