# -*- perl -*-
#  $Id: //depot/prod/test/main/lib/Parser.pm#1 $


## @summary Library of FPolicy routines
## @author  sureshb (DRG FPolicy QA)
## @status  private
## @pod     here

package FPolicy::Parser;

use strict;

use warnings;
use Utils qw(make_host make_conn);
use Tharn qw(say logcomment logresult);
use Carp qw(croak);
use SayWrap qw(sayEOF);
use Subtest;
use FPolicy::Utils;
use Params::Validate qw(validate_pos validate BOOLEAN OBJECT SCALAR ARRAYREF HASHREF);
use FPolicy::Execute qw(ExecuteAndExpect);

require Exporter;
use base qw(Exporter);

our @EXPORT_OK = qw(
			Get_Monitored
			Get_PServer
			Get_LocalBlkdPolicyStat
			Get_Extension
			Get_PolicyStat
			Get_PolicyState
			Get_Volume
			Get_CifsState
			Get_SnapDetails
			CLI_Execute_Validate_Outputstr
			CLI_Verify_NV_Generic
			get_STATES
		);

# this need to be moved to Constants definitions
my %CIFS_STATES = (
	NOT_LICENSED => ["CIFS is not licensed\.", "\(Use the \"license\" command to license it\.\)"],
	NOT_RUNNING => ["CIFS not running\.  Use \" cifs restart\" to restart"],
	RUNNING => ["The following commands are available\; for more information"],
	);

# this need to be moved to Constants definitions
my %FPOLICY_STATES = (
	CIFS_NOT_LICENSED => ["CIFS is not licensed\.", "\(Use the \"license\" command to license it\.\)", "Can't run file policy unless CIFS is licensed\."],
	CIFS_NOT_RUNNING => ["Can't access file policies unless CIFS is running\."],
	CIFS_RUNNING_NO_POLICY => ["CIFS file policy is enabled\."],
	CIFS_RUNNING_POLICY_EXISTS => [
		"CIFS file policy is enabled\.",
		"File policy pol1 \(file screening\) is disabled\.",
		"",
		"No file policy servers are registered with the filer\.",
		"",
		"Operations monitored:",
		" No operations are monitored\.",
		"",
		"List of extensions to screen:",
		" \?\?\?",
		"",
		"List of extensions not to screen:",
		" Extensions-not-to-screen list is empty\.",
		"",
		"Number of files screened:   0",
		"Number of screen failures:  0"
		],
	);


# SHOW Policy output
#		########################
#		#
#		#CIFS file policy is enabled.
#		#
#		#
#		#File policy abc (file screening) is enabled.
#		#
#		#File screen servers              P/S Connect time (dd:hh:mm)  Reqs    Fails
#		#----------------------------------------------------------------------------
#		# 172.17.180.64   \\DELTA         Pri    00:00:01                 0        0
#		#
#		#Operations monitored:
#		#File open,File create,File rename,File close,File delete,File read,File write
#		#Directory rename,Directory delete,Directory create
#		# Above operations are monitored for NFS and CIFS
#		#
#		#Lookup, Getattr, Setattr, Link, Symlink
#		# Above operations are monitored for NFS only
#		#
#		#List of extensions to screen:
#		# MP3,TXT
#		#
#		#List of extensions not to screen:
#		# Extensions-not-to-screen list is empty.
#		#
#		#Number of files screened:   0
#		#Number of screen failures:  4
#		#
#		########################################

=head1 NAME
GetMonitored

=head2 DESCRIPTION
retrieve the monitored file operations from the policy output

=over

=item INPUT
array of received output strings from the policy output

=item Return values
returns the list of comma seperated monitored operations from the output

=cut
sub Get_Monitored
{
	my @opts = @_;

#		Operations monitored:
#		File open,File create,File rename,File close,File delete,File read,File write
#		Directory rename,Directory delete,Directory create
#		 Above operations are monitored for NFS and CIFS
#		
#		Lookup, Getattr, Setattr, Link, Symlink
#		 Above operations are monitored for NFS only
#		

	my $mon = "";
	foreach my $line (@opts) 
	{
		foreach my $subline (split /\n/, "$line") 
		{
			if (	$subline =~ /File open/ or 
				$subline =~ /File close/ or 
				$subline =~ /File read/ or 
				$subline =~ /File write/ or 
				$subline =~ /File create/ or 
				$subline =~ /File delete/ or 
				$subline =~ /File rename/ or 
				$subline =~ /Directory create/ or 
				$subline =~ /Directory delete/ or 
				$subline =~ /Directory rename/ or 
				$subline =~ /Getattr/ or 
				$subline =~ /Setattr/ or 
				$subline =~ /Lookup/ or 
				$subline =~ /Link/ or 
				$subline =~ /SymLink/ 
			) 
			{
				$subline =~ s/^(\s)+$//g;

				# remove the extra spaces around the key words
				foreach my $tok (split /\,/,$subline) 
				{
					$tok =~ s/^(\s)+//g;
					$tok =~ s/(\s)+$//g;
					$mon .= "$tok,";
				}
			} # if
		} #foreach
	} #foreach

	chop $mon;

	#logcomment("Parser:Get_Monitored: return value = $mon");
	return $mon;
}

=head1 NAME
Get_Pserver

=head2 DESCRIPTION
Retrieve the policy server details from the policy output

=item Input parameters
array of output recieved lines

=item Return values
comma seperated string of policy server values

=cut

# enhancements that can be made here
# accomodate the multiple policy server cases and return ARRAY of policy server informations
sub Get_PServer
{
	my @opts = @_;

	# FORMAT
	# 172.17.180.64   \\DELTA         Pri    00:00:01                 0        0

	my @mon = ();
	my $val = "";

	foreach my $line (@opts) 
	{
		foreach my $subline (split /\n/, "$line") 
		{
			if ( $subline =~ /(\d+).(\d+).(\d+).(\d+)(\s)+\\\\(\w+)(\s)+(\w+)(\s)+(\d+):(\d+):(\d+)(\s)+(\d+)(\s)+(\d+)/ ) 
			{
				$val = "$1.$2.$3.$4,$6,$8,$10:$11:$12:$13,$14,$16";
			}
		}
	}
	return $val;
}

=head1 NAME
Get_Extension

=head2 DESCRIPTION
Retrieves the extension list from the policy output

=item Input values
Array of output recieved lines

=item Return value
semi-colon seperated inclusion/exclusion list of extensions

format: inclusion;exclusion

=cut
sub Get_Extension
{
	my @opts = @_;
#		#List of extensions to screen:
#		# MP3,TXT
#		#
#		#List of extensions not to screen:
#		# Extensions-not-to-screen list is empty.

	my @mon = ();
	my $inc = 0;
	my $exc = 0;
	my $incval = "";
	my $excval = "";
	my $done = 0;
	foreach my $line (@opts) 
	{
		foreach my $subline (split /\n/, "$line") 
		{
			if ( $subline =~ /List of extensions to screen/ ) 
			{ 	$inc = 1; 	}

			elsif ( $subline =~ /List of extensions not to screen/ ) 
			{
				$inc = 0;
				$exc = 1;
			}
			elsif ( $subline =~ /Number of requests screened/ ) 
			{
				$inc = 0;
				$exc = 0;
				$done = 1;
			}
			elsif( $inc ) 
			{
				chomp $subline;
				$incval .= "$subline";
			} 
			elsif( $exc ) 
			{
				chomp $subline;
				$excval .= "$subline";
			}
			last if($done);
		}
		last if($done);
	}
	#logcomment("extension = $incval;$excval");
	return "$incval;$excval";
}

#### Added by Prashant for verifying number of ops locally blocked on IC releases..

=head1 NAME
Get_LocalBlkdPolicyStat
	
=head2 DESCRIPTION
Retrieve the policy statistics (Number of files blocked locally) from the policy output
	
=item Input value
Array of recieved output for the policy 
	
=item Return Values
Number of files blocked locally "failure" statistics
	
=cut
	
	sub Get_LocalBlkdPolicyStat
	{
	####   Number of requests screened          :  0
	####   Number of screen failures            :  0
	####   Number of requests blocked locally   :  0 ===> this value will be extracted and returned

	
		my @opts = @_;
		my $screened = 0;
		my $failures = 0;
	
		foreach my $line ( @opts ) 
		{
			foreach my $subline (split /\n/, "$line") 
			{
				#if( $subline =~ /Number of requests blocked locally:(\s)+(\d+)/ ) 
				if( $subline =~ /Number of requests blocked locally[ ]*:[ ]*([0-9]+)/ ) 
				{	$failures = $1;		 }
				
				#elsif( $subline =~ /Number of files screened:(\s)+(\d+)/ ) 
				#{	$screened = $2;		 }
			}
		}
	
		return "$failures";
	}


=head1 NAME
Get_PolicyStat

=head2 DESCRIPTION
Retrieve the policy statistics (files screened / screen failures) from the policy output

=item Input value
Array of recieved output for the policy 

=item Return Values
comma seperated "screened" and "failure" statistics

=cut

sub Get_PolicyStat
{
#		#
#		#Number of files screened:   0
#		#Number of screen failures:  4
#		#

	my @opts = @_;
	my $screened = 0;
	my $failures = 0;

	foreach my $line ( @opts ) 
	{
		foreach my $subline (split /\n/, "$line") 
		{
			if( $subline =~ /Number of files screened:(\s)+(\d+)/ ) 
			{	 $screened = $2;	 }
			elsif( $subline =~ /Number of screen failures:(\s)+(\d+)/ ) 
			{	$failures = $2;		 }
		}
	}

	return "$screened,$failures";
}

=head1 NAME
Get_PolicyState

=head2 DESCRIPTION
Retrieve the policy state (enabled/disabled) from the policy output

=item Input Value
Array of recieved policy output

=item Return Value
Policy state

=cut

sub Get_PolicyState
{
	my @opts = @_;
	my $state = "";
#		#
#		#File policy abc (file screening) is enabled.
#		#

	# Get the state irrespective of output list is scalar or ARRAY
	foreach my $line (@opts) 
	{
		foreach my $subline (split /\n/, "$line") 
		{
			if($subline =~ /screening\) is (\w+)/) 
			{	 $state = $1;		 }
		}
	}
	return $state;
}

=head1 NAME
get_STATES

=head2 DESCRIPTION
Lookup the output string for the object with a particular state
from the table 

=item object
object type for which the state need to be retrieved

=item state

=item Return Value
returns the string corresponding to the object and its state

=cut
sub get_STATES
{
	my ($object, $state) = @_;

	if ( $object eq "CIFS_STATES" ) 
	{	 return $CIFS_STATES{$state};	 } 

	elsif ( $object eq "FPOLICY_STATES" ) 
	{	 return $FPOLICY_STATES{$state};	 } 

	else # state is not defined
	{	return ("");	 }
}

=head1 NAME
Get_CifsState

=head2 DESCRIPTION
Retrieves the CIFS state from the policy output

=item Input Values
Array of recieved output from the policy output

=item Return Values
CIFS state 

=cut
sub Get_CifsState
{
	my @opts = @_;
	my $state = "";
#		#
#		#CIFS file policy is enabled.
#		#
	
	foreach my $line (@opts) 
	{
		foreach my $subline (split /\n/, "$line") 
		{
			foreach my $key (keys %CIFS_STATES) 
			{
				if($subline =~ /$CIFS_STATES{$key}/) 
				{	 $state = $1;		 }
			}
		}
	}

	return $state;
}

=head1 NAME
Get_SnapDetails

=head2 DESCRIPTION
Retrieve the Snap ID and Name from the snap status output

=item Input Values
Outout of snap status -V <vol-name> command

=item Return Value
Hash containing Snap ID and Snap Name values

=cut
sub Get_SnapDetails
{
	my @opts = @_;
    my %snapStatus = ();
	# FORMAT
	# 12     complete   Aug 01 08:00       115     7.3 19736 hourly.0

	my $next = 0;
	foreach my $status (@opts) 
	{
		foreach my $subline (split /\n/, "$status") 
		{
			if( $next && ($subline =~ /\s(\d+)/)) 
			{	
				my @tmp_arr = split (' ', "$subline");
				$snapStatus{$tmp_arr[8]} = $tmp_arr[0];
				$next = 1;			
			}
			if( $subline =~ /--------/ )
			{	 $next = 1;	 } 

			elsif( $subline =~ /(\w+)(\*)?\>/ ) 
			{
				$next = 0;
				last;
			}
		}
	}

	return (%snapStatus);
}

=head1 NAME
Get_Volume

=head2 DESCRIPTION
Retrieve the volume list from the volume output

=item Input Values
Array of recieved volume output 

=item Return Value
comma seperated list of volume names

=cut
sub Get_Volume
{
	my @opts = @_;
	my $volume = "";


	my $next = 0;
	foreach my $vol (@opts) 
	{
		foreach my $subline (split /\n/, "$vol") 
		{
			if( $subline =~ /List of volume specifications/ )
			{	 $next = 1;	 } 

			elsif( $subline =~ /(\w+)(\*)?\>/ ) 
			{
				$next = 0;
				last;
			}

			if( $next ) 
			{	 $volume .= "$subline,";	 }
		}
	}

	chop $volume;
	return $volume;
}



# not in use, needs more clarity and work
sub CLI_Verify_NV_Generic
{
	my $output = shift; # hash reference to the received output
	my $expected = shift; # hash reference to the expected output

	my $error = 0;
	foreach my $exp (keys %{$expected})
	{
		logcomment("Key - $exp, value = ${$expected}{$exp}");
		# incase of monitor, extension, volume where multiple values exists
		# (comma) is the record seperator
		my @evalues = split /\,/, ${$expected}{$exp};
		my @ovalues = split /\,/, ${$output}{$exp};

		#if( $exp ne "" and
		#	( ${$output}{$exp} eq ${$expected}{$exp} or 
		#	${$output}{$exp} =~ /^(\s)+${$expected}{$exp}(\s)+$/ )  or
		#	( ${$output}{$exp} eq "" and ${$expected}{$exp} =~ /default/ )
		#) {
		#	logcomment("$exp: ${$output}{$exp} FOUND as expected");
		#} else {
		#	logcomment("$exp: ${$output}{$exp} NOT as expected value ${$expected}{$exp}");
		#	$error = 1;
		#}
		my $found = 0;
		foreach my $ekey (@evalues) 
		{
			foreach my $okey (@ovalues) 
			{
				if( $ekey eq $okey )
				{ $found = 1; last; }
				else 
				{ $found = 0;}
			}
			if(!$found) { last;}
		}
		if(${$expected}{$exp} eq ${$output}{$exp})
		{ 	$found = 1; } #when expected and cli output is NULL 
		
		if( $found )
		{
			logcomment("$exp: ${$output}{$exp} FOUND as expected");
		} else {
			logcomment("$exp: ${$output}{$exp} NOT as expected value ${$expected}{$exp}");
			$error = 1;
		}
	}

	return $error;
}


=head1 NAME
CLI_Execute_Validate_Outputstr

=head2 DESCRIPTION
Execute the Command and validate the expected output vs recieved output

=item conn
connection object to connect to the filer

=item intf
interface type definition (CLI/ZAPI)
ZAPI - not yet implemented, low priority

=item expected
hash of expected values

=item vfiler
Name of the vfiler 

=item filer
Name of the filer

=item result
Expected result as compared to the operation performed
(pass/fail)

=item Return Value
hash of result and output received
result = 0/1 (pass/fail)
output = received output from the command executed

=cut

sub CLI_Execute_Validate_Outputstr
{
	my %parms = validate
	(
		@_, 
		{
		conn => {	type => SCALAR|OBJECT|HASHREF 	},

		intf => {	type => SCALAR, 
				regex => qr/[cli|zapi]/i
			},

		command => {	type => SCALAR, 
				regex => qr/(\w+)((\s)+(\w+))*/
			},

		expected => {	type => SCALAR|HASHREF		},

		vfiler => {	type => SCALAR		},

		filer => {	type => SCALAR		},

		result => {	type => SCALAR, 
				regex => qr/[pass|fail]/i
			}
	});
			

	my $ptr;
	# This can be enabled when connection need to be made dynamically
	# when commands are executed without need to use seperate connect API
	# see some issues depending on the type of usage, will be clear as we go forward
	# let this be un-used for now
#	if ( !defined $parms{conn} or $parms{conn} eq "" or ref($parms{conn}) eq "SCALAR")
#	{
#		$ptr = FPolicy::Utils::make_connection("conn_type=$parms{intf}", "filer=$parms{filer}");
#		$parms{conn} = ${$ptr}{conn};
#		logcomment("Created new CLI conn = $parms{conn}");
#	}

	logcomment("ExecuteAndExpect: $parms{conn}, $parms{command}, 
		$parms{expected}, success=${$parms{expected}}{success}, 
		failure=${$parms{expected}}{failure}, expected result=$parms{result}");

	my %status = ExecuteAndExpect(
				$parms{conn}, 
				$parms{command}, 
				$parms{expected},
				$parms{vfiler}		### Added by Rajaram
			);

	my $str = "Received status : ";
	$str .= (!$status{result})? "SUCCESS" : "FAILURE";

	if( $parms{result} eq "FAIL" )
	{
		# negate the result
		my $tmp = ($status{result})? 0 : 1; 

		# using the return status will validate the script issues
		 $str .= " , Expected to FAIL, returning $status{result}"; #returning $tmp";

		$status{result} = $tmp;
	}
	else 
	{ # using the return status will validate the script issues
		$str .= " , Expected to PASS, returning $status{result}"; 
	}

	# do not comment this line, will add clarity when script fails
	logcomment($str);

	return ( 
# connection parameters need to be sent when using dynamic connection
# as specified from above
#		host => ${$ptr}{host},
#		conn=> $parms{conn}, 
		result => $status{result}, 
		output => $status{output} 
	);
}


1;

