# -*- perl -*-
#  $Id: //depot/prod/test/main/lib/Traffic.pm#1 $


## @summary Library of FPolicy routines
## @author  sureshb,sandeepp (DRG FPolicy QA)
## @status  private
## @pod     here

package FPolicy::Traffic;

use strict;

#use warnings;
use Switch;
use Utils qw(make_host make_conn);
use TestEnv::MinWin qw(:all);
use Tharn;
use Carp qw(croak);
use SayWrap qw(sayEOF);
use Subtest;
use MapShare qw(:all);

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use nfs::NfsDef qw(nfs_error NFREG EXCLUSIVE NFLNK FATTR4_MODE UNCHECKED FILE_SYNC NFLNK FATTR4_TIME_ACCESS_SET FATTR4_TIME_ACCESS FATTR4_TIME_MODIFY_SET FATTR4_TIME_MODIFY SET_TO_SERVER_TIME NFDIR );
use nfs::NfsOps qw(set_env clean_env scalar_validate);
use nfs::NfsClient;
use Time::Local qw(timelocal timegm);
use FPolicy::Utils qw(make_NV make_hash_fromArrayHash);
use FPolicy::Constants qw($POLICY_SERVER_TRANSFER_EXE $DELETE_NT_STREAMS_PATH $DELETE_NT_STREAMS $CIFS_DOMAIN_NETBIOS $CIFS_DOMAIN_USERNAME $CIFS_DOMAIN_PASSWORD $CIFS_SHARE $CIFS_WORKGROUP);


my %CIFS_FILE_OPERATIONS = (
	READ => 1,
	WRITE => 1,
	CREATE => 1, 
	DELETE => 1, 
	RENAME => 1, 
	DIRCREATE => 1, 
	DIRRENAME => 1, 
	DIRDELETE => 1,
	OFFLINE	=> 1,
	DELETE_ON_CLOSE => 1,
);

my %NFS_FILE_OPERATIONS = (
	READ => 1, 
	WRITE => 1,
	CREATE => 1, 
	DELETE => 1, 
	RENAME => 1, 
	SETATTR => 1,
	GETATTR => 1,
	LOOKUP => 1,
	DIRCREATE => 1, 
	DIRRENAME => 1, 
	DIRDELETE => 1
);

sub new ($;@)
{
	my $object = shift;
	my $class = ref($object) || $object;
	my %args = make_NV(@_);

	my $tmp;
	my $host;
	my $conn;
	my $share_drive;
	my $cifs_share;
	
	
	if($args{traffic_type} =~ /nfs/i)
	{
		eval {
		my $hostname = $args{filer_ip};
		my $version = $args{nfsversion} || "3";	# 2|3|4
		my $proto = $args{protocol} || "tcp"; # tcp|udp
		my $timeout = $args{timeout} || "30";
		my $localaddr = $args{nfsclient};

		$tmp = new nfs::NfsClient(
				hostname => $hostname,
				version => $version,
				proto => $proto,
				timeout => $timeout,
				localaddr => $localaddr
			);
		};

		if($@) {
			logresult("config", "Errors during NFS Initialization with parms [@_], error = $@");
		} else {
		}

	}
	if($args{traffic_type} =~ /cifs/i)
	{
		logcomment("Initializing CIFS client");
		#To check for MinWin compliance
#		setup_cifs($args{cifsclient});
#		($share_drive, $cifs_share) = setup_cifs(%args);
#		$self->{cifsclient} = $args{cifsclient};
		eval {
			$host = make_host("$args{cifsclient};hosttype=windows");
			$conn = make_conn($host);	
		
		};
	}

	my $self = {	
		traffic_type => $args{traffic_type},
		volume => $args{volume},
		nfs => $tmp,
		cifs_conn => $conn, 
		cifs_client => $args{cifsclient},
		filer_ip => $args{filer_ip},
		nfs_version => ($args{nfsversion} || "3")
	#	cifs_share => $cifs_share,
	#	cifs_share_drive => $share_drive
	};


	logcomment("Traffic Agent is initialized with
			type 		= $args{traffic_type}
			hostname 	= $args{hostname}
			nfs_version 	= $args{nfs_version}
			proto 		= $args{protocol} 
			nfs object 	= $tmp
			cifs connection = $conn
	#		cifs share 	= $cifs_share
	#		cifs share drive = $share_drive
		");

       bless($self, $class); 
       return $self;
}

=head1 NAME
setup

=head2 DESCRIPTION
This API is used to configure default options
Mount the path to volume under test 

=head2 ROUTINES
	$pol->{traffic}->setup(path => "<mount path>");

=over

=item result
mount handle is stored in the object 
returns PASS / FAIL (0/1)

=cut

sub setup ($;%)
{
	my $self = shift;
#	my %opts = @_;
        my %opts = validate(@_, {
			traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}},
	                cifs_client => {type => OBJECT|SCALAR, default => $self->{cifs_client}},
			filer_ip => {type => OBJECT|SCALAR, default => $self->{filer_ip}},
			volume => {type => OBJECT|SCALAR, default => $self->{volume}}
                });
	
	
	logcomment("In setup: $self->{traffic_type} @_");
	
	if($opts{traffic_type} =~ /nfs/i){
	
		eval {
	#		my $path = "/vol/". ($opts{volume}|| $self->{volume}) ;
			my $path = "/vol/". $opts{volume};
			logcomment("Traffic setup for NFS with ( path => $path )");
			$self->{rpcobject} = 
				$self->{nfs}->mount( 
					path => "$path"
				);

			$self->{nfs}->{rpc}->auth_sys_set( uid => 0, gid => 0 );

		};

		if ($@)
		{
			logresult("config", "Errors during NFS setup:$@") if ($@);
		} else {
			logcomment("Setup complete for NFS: mnt_path = $self->{nfs}->{mnt_path}, rpcobject = $self->{rpcobject}, mnt_fh = $self->{nfs}->{mnt_fh}");
		}

		$self->{nfs}->{mnt_fh} = $self->{rpcobject}->{fhandle};
	} #if nfs
	
	if($opts{traffic_type} =~ /cifs/i){
		my $share_drive;
		my $cifs_share;
		eval {
			logcomment("Traffic setup for CIFS on client $opts{cifs_client}");
			($share_drive, $cifs_share) = setup_cifs(%opts);
			$self->{cifs_share} = $cifs_share;
			$self->{cifs_share_drive} = $share_drive;	
		};
		if ($@)
		{
			logresult("config", "Errors during CIFS setup:$@") if ($@);
		} else {
			logcomment("Setup complete for CIFS: cifs share = $cifs_share cifs share drive = $share_drive");
		}
		
	} #if cifs

}


sub setup_cifs($;%)
{	
#	my $self = shift;
	
#	my ( $winClient, $temp) = @_;
	my %setup_args = @_;
	my $winClient = $setup_args{cifs_client};
	my $cifs_share = $CIFS_SHARE;		#HOME
	my $shareDrive;
	
	logcomment("Debug - Params in setup_cifs: @_");
	
	# Bail out if the Windows Client is not minwin_compatible
	if ( ! minwin_compatible($winClient) ) {
	        logresult('FAIL',"CLIENT ($winClient) is not minwin-compatible");
	        return();
	}else{	
		# Create the required share on the filer for tests.
		# $cifs_share = create_a_share_on_filer();
			
		# Map the share on Windows client to the next available drive letter and list share contents
			#if domain name is provided then workgroup will be ignored. 
		
		if($CIFS_DOMAIN_NETBIOS ne ""){ 
			$shareDrive = map_cifs_share(server=>$setup_args{filer_ip},
				share=>$cifs_share,
				client=>$winClient,
				domain=>$CIFS_DOMAIN_NETBIOS, 		#$CIFS_DOMAIN_NETBIOS,
				username=>$CIFS_DOMAIN_USERNAME, 	#$cifsuser,
				password=>$CIFS_DOMAIN_PASSWORD, 		#$cifspasswd,
				drive=>'*');
	
			if (!$shareDrive) {
			     die("Unable to map the cifs share $cifs_share on Windows Client $winClient");
			}
			logcomment("Cifs share was mapped using Domain credentials.");
		}else{
			$shareDrive = map_cifs_share(server=>$setup_args{filer_ip},
				share=>$cifs_share,
				client=>$winClient,
				workgroup=>$CIFS_WORKGROUP, 		#$CIFS_WORKGROUP,
				drive=>'*');			
			
			if (!$shareDrive) {
				     die("Unable to map the cifs share $cifs_share on Windows Client $winClient");
			}
			logcomment("Cifs share was mapped using Workgroup authentication.");
		}
	
		#logresult('pass',"CLIENT ($winClient) is minwin-compatible and $shareDrive is mapped to $cifs_share");
		logcomment("PASS: CLIENT ($winClient) is minwin-compatible and $shareDrive is mapped to $cifs_share");
	}
	return ($shareDrive, $cifs_share);
}


=head1 NAME
cleanup

=head2 DESCRIPTION
This API is used to unmount the mounted filesystem during setup for NFS and delete mounted shares
on CIFS clients.

=head2 ROUTINES
	$pol->{traffic}->cleanup();

=over

=item result
no return value

=cut

sub cleanup ($;%)
{
	my $self = shift;
#	my %opts = @_;
        my %opts = validate(@_, {
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}},
                cifs_client => {type => OBJECT|SCALAR, default => $self->{cifs_client}},
		filer_ip => {type => OBJECT|SCALAR, default => $self->{filer_ip}}
        });	


	if($opts{traffic_type} =~ /nfs/i){
		logcomment("Clean-up...Un-mounting path $opts{volume}");
		eval {
			my $fh_mnt = $self->{nfs}->umount( path => $opts{volume}|| $self->{volume} );
		};
	
		logresult("fail", "Errors during NFS clean-up: $@") if($@);	
	}#if nfs
	
	if($opts{traffic_type} =~ /cifs/i){
		logcomment("Clean-up...Deleteing share mount $self->{cifs_share_drive} on CIFS client $self->{cifs_client}.");
		say($self->{cifs_conn}, "net use $self->{cifs_share_drive} \/DELETE", "-wait", qr/successfully/i, "-fail", "Errors during cleanup on CIFS client $self->{cifs_client}.");
	}
	
#Added by Avinash K, to clean up nfpc_transfer.exe
# my $par = {
#                pserver_pri => $NV{pserver_pri},
#                pserver_sec => $NV{pserver_sec}
#        };
#
#my $pserver = $par->{pserver_pri};
 #     logcomment("PServer  Transfer - INSIDE Cleanup");
#my $host = make_host("$pserver;hosttype=windows");
#my $conn = make_conn($host);

#	eval {
	# close the session

#	logcomment("Check presence of $POLICY_SERVER_TRANSFER_EXE");
#	my(undef,undef,$msgc) = say($conn}, "cmd /c tasklist /FI \"IMAGENAME eq $POLICY_SERVER_TRANSFER_EXE\"", "dummy", -ok);
 #               logcomment("Msg Returned here: $msgc");

#	if($msgc =~ /No tasks/i) { #No tasks are running
#		logcomment("Msg Returned $msgc");
#	}
#	else
#	{
#		logcomment("Msg Returned $msgc Kill the process $POLICY_SERVER_TRANSFER_EXE");
#		my(undef,undef,$msgc) = say($conn}, "cmd /c taskkill \/F \/IM $POLICY_SERVER_TRANSFER_EXE", "dummy", -ok);
#	}

#	};

#	if($@) {
#		logcomment("Error during cleanup: $@");
#	}

}


=head1 NAME
Verify

=head2 DESCRIPTION
This API is used to compare the expected vs obtained values for 
SetAttributes

Every call to SetAttribute can vary with provided optional parameters
This API tries to verify all the attributes under test before and after SetAttribute is called.

=head2 ROUTINES
	$pol->{traffic}->Verify( attribute => <type of attribute under test>, 
				get => <get object>, 
				set => <set object> );

=over

=item result
mount handle is stored in the object 
returns PASS / FAIL (0/1)

=cut

sub Verify ($;@)
{
	my $self = shift;
	my %hash = validate(@_, {
		attribute => { type => SCALAR, optional => 1 },
		get => { type => HASHREF },
		set => { type => HASHREF }
		});

	my $set = $hash{set};
	my $get = $hash{get};

	logcomment("Verify 
		attribute = $hash{attribute}, 
		shash = $set, ${$set}{size},
		ghash = $get, $get->{attributes}->{size}"
	);

	my $error = 0;
	# compare attribute

	# use mtime, atime to verify the SetAttribute
	foreach my $attr (split /\|/, $hash{attribute})
	{
		logcomment("Compare Attribute = $attr");
		SWITCH: for ($attr) {
			/atime|ctime|mtime/i && do {
			logcomment(" get nseconds = $get->{attributes}{$attr}{nseconds}, set = ${$set}{$attr}{nseconds}, get seconds = $get->{attributes}{$attr}{seconds}, set = ${$set}{$attr}{seconds}");

			if( $get->{attributes}{$attr}{nseconds} != ${$set}{$attr}{nseconds} ||
			$get->{attributes}{$attr}{seconds} != ${$set}{$attr}{seconds} )
			{
				$error = 1;
				logcomment("error - $attr is incorrect");
			}

			};

			/size|mode/i && do {
			logcomment(" get $attr = $get->{attributes}{$attr}, set=${$set}{$attr}" );
			if( $get->{attributes}{$attr} != ${$set}{$attr} )
			{
				$error = 1;
				logcomment("error - $attr is incorrect");
			}
			};
		}
	}

	return $error;
}

=head1 NAME
Lookup

=head2 DESCRIPTION
This API is used to generate a 'lookup' file event by performing the lookup operation

=head2 ROUTINES
	$pol->{traffic}->Lookup(filename => "");

=over

=item result
Lookup handle for the file is captured in the object for further operations on the same file
returns status of the operation (1/0)

=cut

sub Lookup
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR },
		dhandle => {type => SCALAR|OBJECT, optional => 1 },
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1}
		});

	my $filename = $args{filename};
	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $res;
	eval {
	$res = $self->{nfs}->lookup(
			name => $filename, 
			dhandle => $self->{nfs}->{mnt_fh}
		);
	$self->{lookup_fhandle} = ${$res}{fhandle};
	};

	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Lookup" );
	

	#return ${$nfsRes}{status};
	return ( status => $nfsRes );
}

# atime, ctime, mtime
# chgrp, chmod, chown, set current size, incr file size,
# decr file size, file truncate, 
sub SetAttribute
{
	my $self = shift;
	my %args = validate(@_, {
		atime => {type => SCALAR, optional => 1},
		mtime => {type => SCALAR, optional => 1},
		ctime => {type => SCALAR, optional => 1},
		size => {type => SCALAR, optional => 1},
		mode => {type => SCALAR, optional => 1},
		dhandle => {type => OBJECT|SCALAR, optional => 1},
		filename => {type => SCALAR},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1},
                #Added by Avinash K for NfsV4
                nfsver => {type => SCALAR, optional => 1}
		});

	
	my $filename = $args{filename};
	my $dirhandle = ""; # $args{dhandle} || $self->{nfs}->{cur_fh};

	if($dirhandle eq "")
	{
		$self->Lookup( filename => $filename );
		$dirhandle = $self->{lookup_fhandle};
	}

	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $res;
        my $version = $args{nfsver};
	my %FILE_ATTRIBUTES = ( 
			atime => $args{atime},
			ctime => $args{ctime},
			mtime => $args{mtime},
			mode => $args{mode},
			size => $args{size} 
	);
	
my $secs = timelocal((localtime)[0..5]) - (int(rand 50)+10);
my $mode_key =  FATTR4_TIME_MODIFY_SET();
# $expected{$mode_key} = { seconds    => $secs,
 #                                $usec_nsec => 0 }; 
# $expected{FATTR4_TIME_MODIFY()} = $expected{FATTR4_TIME_MODIFY_SET()};
 #           delete $expected{FATTR4_TIME_MODIFY_SET()};
#$res = $self->{nfs}->setattr( attr => { $mode_key => SET_TO_SERVER_TIME } );

my %expected;
my $usec_nsec;

if ($version != 4)
{
	eval {
	# for the attribute specified
	
	foreach my $key (keys %FILE_ATTRIBUTES)
	{
		next if( !defined $FILE_ATTRIBUTES{$key} );

		logcomment("dirhandle = $dirhandle, ATTR = $key, VAL = $FILE_ATTRIBUTES{$key}");


		SWITCH: for ($key) {
		
		/[atime|size|mtime]/i && do {
				$res = $self->{nfs}->setattr ( fhandle => $dirhandle, attr => { $key => "$args{$key}" } );
			};
		/mode/i && do {
				$self->{nfs}->{rpc}->auth_sys_set(uid=>0, gid=>0);
				$res = $self->{nfs}->setattr ( fhandle => $dirhandle, attr => {size => $args{$key}});
			};
		}
        }
        };
}
else {
	eval {
         $expected{$mode_key} = { seconds    => $secs,
                                 $usec_nsec => 0 };
         $expected{FATTR4_TIME_MODIFY()} = $expected{FATTR4_TIME_MODIFY_SET()};
         delete $expected{FATTR4_TIME_MODIFY_SET()};
         $res = $self->{nfs}->setattr( attr => { $mode_key => SET_TO_SERVER_TIME } );
	};
}

	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Setattr" );
	return ( status => $nfsRes );
}

sub GetAttribute
{
	my $self = shift;
	my %args = validate(@_, {
		dhandle => {type => OBJECT|SCALAR, optional => 1},
		filename => {type => SCALAR},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1}
		});
	my $filename = $args{filename};
	my $dirhandle = $args{dhandle} || $self->{nfs}->{cur_fh};

	if($dirhandle eq "")
	{
		$self->Lookup( filename => $filename );
		$dirhandle = $self->{lookup_fhandle};
	}
	
	$args{result} = "pass" if(!defined $args{result});

	my $res;
	my $nfsRes;
	
	eval {
	$res = $self->{nfs}->getattr ( fhandle => $dirhandle	);
	};
	
	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Create" );


	return { 
		#status => ${$res}{status},
		status => $nfsRes, 
		object => $res, 
		attributes => ${$res}{fattr}
	};
}

sub Delete
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		dhandle => {type => OBJECT|SCALAR, optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                ## Added By Prashant(prash@netapp.com) for negative testing
                result => {type => SCALAR, optional => 1}
		});
		
	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $cifsRes;
	my $res;
	my $filename = $args{filename};
	
	if($args{traffic_type} =~ /nfs/i) {
	
		my $dirhandle = $args{dhandle} || $self->{nfs}->{mnt_fh} || $self->{create_fhandle};

		eval {
			$res = $self->{nfs}->remove
			( 
				dhandle => $dirhandle,
				name => $filename,
			);
		}; # eval
		
		
		$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Delete" );
	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $filename;
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "del \/F $cifs_filename");
		if ($args{result} =~ /fail/i){
			my(undef,undef,$clientRes) = say($self->{cifs_conn}, "dir \/F $cifs_filename");
			if ($cifsRes =~ /$cifs_filename/i){
				# Delete not succeded..
				$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Delete" );
			}
		}else{
			$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Delete" );
		}
		
	}#if cifs
	
	my $finalRes = $cifsRes | $nfsRes;
	
	return (status => $finalRes );
}


sub Create
{
        my $self = shift;
        my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
                dhandle => {type => OBJECT|SCALAR, optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                ## Added By Prashant(prash@netapp.com) for negative testing
                result => {type => SCALAR, optional => 1},
                nfsver => {type => SCALAR, optional => 1, default => 3}
                });

	$args{result} = "pass" if(!defined $args{result});
	
        my $res;
        my $nfsRes;
        my $cifsRes;
        my $spec; #avi
        
        if($args{traffic_type} =~ /nfs/i) {
        
#Added by Avinash K

	        my $filename = $args{filename};
	        my $dirhandle = $args{dhandle} || $self->{nfs}->{mnt_fh};
	        my $version = $args{nfsver} || "3";
                my $mode = EXCLUSIVE;
                my $verf = time;
                

		
	        my $filename = $args{filename};
	        my $dirhandle = $args{dhandle} || $self->{nfs}->{mnt_fh};
		
	        $args{result} = "pass" if(!defined $args{result});
	        logcomment("version is $version ");
	        
	        
	if ( $version != 4 ) {
	        eval {
	        $res = $self->{nfs}->create
	                (
	                        dhandle => $dirhandle,
	                        name => $filename,
	                );
        	};
        	
        } else {
        
        	eval {
        	$res = $self->{nfs}->open(
		                      name    => $filename,
		                      cr_mode => $mode,  #OPEN4_CREATE
		                      verf    => $verf,
		                      attr    => { FATTR4_MODE() => 0777 }
		                     );
		
		 };
	}	  	
        	
        	#logcomment("Status after NFS create ops is : ${$res}{status}");
        	
        	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Create" );
		
	}
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo sometext > $cifs_filename");
			
		#my(undef,undef,$clientRes) = say($self->{cifs_conn}, "del $filename", "dummy", -ok);
		
	
		#if ($clientRes ne ""){
		#	if($args{result} =~ /fail/i) {
		#		# you expected test to fail and it failed
		#		logresult("pass", "CIFS file create Operation expected to Fail");
                #        	#${$res}{status} = 0;
                #        	$cifsRes = 0;
                #	}else{
		#		logresult("fail", "CIFS File Create operation did NOT complete on the client");
		#		#${$res}{status} |= 1;
		#		$cifsRes = 1;
		#	}

		#}else{
		#	if($args{result} =~ /pass/i){
		#		logcomment("CIFS File Create operation completed successfully on the client");	
		#		# ${$res}{status} |= 0;
		#		$cifsRes = 0;
		#		
		#	}else{
		#		logresult("fail", "CIFS Create operation is expected to Fail but it passed");
		#		$cifsRes = 1;
		#	}
		#}
		
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Create" );
		
        	
	}


        #if ($@)
        #{
        #        ${$res}{status} = 1; # indicate failure
        #        ## Added By Prashant(prash@netapp.com) for negative testing
        #        if($args{result} =~ /fail/i) {
        #                # you expected test to fail and it failed
        #                logresult("pass", "Operation \"create\" expected to Fail");
        #               ${$res}{status} = 0;
        #       }else
        #       {
        #                logresult("fail", "Errors during Create:$@");
        #        }
     	#}else{
     	#	logresult("pass", "File Create success");
     	#}
         
        my $finalRes = $cifsRes | $nfsRes;

        #return { status => ${$res}{status}, ${$res}{fhandle} };
        return (status => $finalRes , ${$res}{fhandle} );
}


#To be deleted. This code is obsolete Create() takes care of CIFS create as well.
#sub Create_cifs
#{	
#	my $self = shift;
#	my $filename = "c:\\newfi1.txt";
##	say($self->{cifs_conn}, "echo somedata > c:\\newfi1.txt");
#	
#	my(undef,undef,$clientRes) = say($self->{cifs_conn}, "del $filename", "dummy", -ok);
#
##	my $clientRes = say($self->{cifs_conn}, "del $filename", "-wait", qr".+");
#	if ($clientRes ne ""){
#		logcomment("Here$clientRes Operation did NOT complete on the client");
#	}else{
#		logcomment("Here$clientRes Operation completed successfully on the client");
#	}
#}


#to be deleted. This is the original NFS only create() function.
#sub Create
#{
#	my $self = shift;
#	my %args = validate(@_, {
#		filename => {type => SCALAR},
#		dhandle => {type => OBJECT|SCALAR, optional => 1}
#		});
#	my $filename = $args{filename};
#	my $dirhandle = $args{dhandle} || $self->{nfs}->{mnt_fh};
#
#	my $res;
#
#	eval {
#	$res = $self->{nfs}->create
#		( 
#			dhandle => $dirhandle,
#			name => $filename,
#		);
#	};
#
#	if ($@)
#	{
#		${$res}{status} = 1; # indicate failure
#		logresult("fail", "Errors during Create:$@");
#	#}else{
#	#logresult("pass", "Create success");
#	}
#
#	return { status => ${$res}{status}, ${$res}{fhandle} };
#}

sub Read
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		dhandle => {type => OBJECT|SCALAR, optional => 1},
		count => {type => SCALAR, default => 1024, optional => 1},
		offset => {type => SCALAR, default => 0, optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		result => {type => SCALAR, optional => 1}
		});

	my $filename = $args{filename};
	
	my $nfsRes;
	my $cifsRes;
	
	my $res;
	
	$args{result} = "pass" if(!defined $args{result});

	
	if($args{traffic_type} =~ /nfs/i) {	
		my $dhandle = $args{dhandle} || $self->{read_fhandle} || $self->{nfs}->{mnt_fh} ;
		my $offset = $args{offset};
		my $count = $args{count};
		my $stateid = $args{stateid} || 0;
	
	
		print "\nReading file handle=$dhandle for length=$count bytes from offset $offset:\n";
	
		my @parms = ( 
			offset => $offset,
			count => $count,
			stateid => $stateid
		);
	
		$self->{lookup_fhandle} = $dhandle;
		eval {
		$self->Lookup( filename => $filename );
	
		$res = $self->{nfs}->read
				(
					fhandle => $self->{lookup_fhandle},
					@parms
				);
		};
	
		$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-File Read" );

	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) {
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "type $cifs_filename");		
	
	#Read op will echo back the read data from the file. Hence error when access is denied or cannot access.
	#hence matching with keyword access.
	
		if (!($clientRes =~ /Access/i)){    
			$clientRes ="";
		}
			
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Read" );
		
        	#logcomment("Status after CIFS Read ops is : ${$res}{status}");
	
	}#if cifs

	
	my $finalRes = $cifsRes | $nfsRes;

	return ( status => ${$res}{status}, data => ${$res}{output} );
}

sub Open
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		result => {type => SCALAR, optional => 1},
                nfsver => {type => SCALAR, optional => 1}
		});
	#TBD To support NFSv4 Opens in future
        
	my $cifsRes;
	my $nfsRes;
	my $finalRes;
	my $res;
	$args{result} = "pass" if(!defined $args{result});
	
	my $filename = $args{filename};
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
#		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo readmoretext > $cifs_filename");		
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "type $cifs_filename");		
	
		if (!($clientRes =~ /Access/i)){
			$clientRes = "";
		}
		
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Open" );
			
        	#logcomment("Status after CIFS Open ops is : ${$res}{status}");
	        #return { status => $cifsRes, data => ${$res}{output} };
	}
    
	###Added by Avinash K to handle nfs v4 open
    if($args{traffic_type} =~ /nfs/i)
	{

        my $filename = $args{filename};
        my $dirhandle = $args{dhandle} || $self->{nfs}->{mnt_fh};
        my $version = $args{nfsver} || "3";

        $args{result} = "pass" if(!defined $args{result});

		logcomment("version is $version ");
		if ( $version != 4 )
		{
			logcomment("File open not supported in NFS version $version");
			logresult("fail", "File open failed coz of unsupported version");
		} else
		{
			logcomment ("@@@@@@@ file name is $filename");
			$res = $self->{nfs}->open(
                      name  => $filename,
                      dhandle  => $self->{nfs}->{mnt_fh}
                     );
            $nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Open" );
                                 
		}
        $nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Open" );
		
	}
	
	$finalRes = $cifsRes | $nfsRes;
    return (status => $finalRes, output => $res);	
}


sub Close
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		nfsver => {type => SCALAR, optional => 1},
		fhandle => {type => SCALAR, default => $self->{nfs}->{mnt_fh}, optional => 1},
		stateid => {type => HASHREF, optional => 1},
		result => {type => SCALAR, optional => 1}
		});
	#TBD To support NFSv4 close in future
	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $cifsRes;
	my $finalRes;
	my $res;
	
	my $filename = $args{filename};
	my $version = $args{nfsver} || "3";
	
	if($args{traffic_type} =~ /cifs/i)
	{ 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "type $cifs_filename");		
	
		if (!($clientRes =~ /Access/i))
		{
			$clientRes = "";
		}
		
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Close" );
        #logcomment("Status after CIFS Close ops is : ${$res}{status}");
	}
	
	#return ( status => $cifsRes, data => ${$res}{output} );
	
	### Added by Rajaram for NFsv4 
	if($args{traffic_type} =~ /nfs/i)
	{
		if ( $version != 4 )
		{
			logcomment("File Close not supported in NFS version $self->{version}");
			logresult("fail","File close failed because of unsupported version");
		}
		
		#logresult("fail","Please provide fhandle of the file to close") if(!defined $args{fhandle});
		logresult("fail","Please provide stateid of the file to close") if(!defined $args{stateid});
		eval {
		$self->Lookup( filename => $filename );
		$res = $self->{nfs}->close(stateid => $args{stateid}, fhandle => $self->{lookup_fhandle});
		};
		
		$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-File Close" );
	}
	
	$finalRes = $cifsRes | $nfsRes;
	return ( status => $finalRes );
}

sub Write
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		data => {type => SCALAR},
		offset => {type => SCALAR, default => 0, optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1}
		});


	my $nfsRes;
	my $cifsRes;
	
	my $res;
	$args{result} = "pass" if(!defined $args{result});

	my $filename = $args{filename};
	if($args{traffic_type} =~ /nfs/i) {	
		$res = $self->Lookup( filename => $args{filename}, dhandle => $self->{nfs}->{mnt_fh} );
		#logcomment ("Before the write");
		eval {
		$res = $self->{nfs}->write(
					fhandle => $self->{lookup_fhandle},
                                       # stable  => FILE_SYNC,
					offset  => $args{offset},
                                        stateid => 0,
					data    => $args{data}
			);
			
		};
	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-File write" );

	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo moretext > $cifs_filename");		
	
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-File write" );
		
        	
	}#if cifs
	
	
	my $finalRes = $cifsRes | $nfsRes;
	#return { status => $finalRes, data => ${$res}{output} };
	#return { status => $finalRes);
	
}

sub Rename
{
	my $self = shift;
	my %args = validate(@_, {
		fromname => {type => SCALAR, default => "auto_test.txt"},
		toname => {type => SCALAR, default => "auto_test_1.txt"},
		fromdir => {type => SCALAR},
		todir => {type => SCALAR},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR,optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1}
		});

	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $cifsRes;
	my $res;

	if($args{traffic_type} =~ /nfs/i) {
		eval {
			$self->Lookup( filename => "$args{fromname}", dhandle => $self->{nfs}->{mnt_fh} );
		
			#logcomment ("@@@ $args{fromname},$args{toname},$self->{nfs}->{mnt_fh} @@@");
			$res = $self->{nfs}->rename(
					fromname => $args{fromname},
					toname => $args{toname},
					fromdir => $self->{nfs}->{mnt_fh},
					todir => $self->{nfs}->{mnt_fh}
			);
	
		};

	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-File Rename" );

	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) { 
	
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{fromname};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "ren $cifs_filename $args{toname}", "dummy", -ok);
				
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-File Rename" );
		
	}#if cifs	
	
	my $finalRes = $cifsRes | $nfsRes;
	
	#return { status => $finalRes, data => ${$res}{output} };
	
	return (status => $finalRes);
}


sub Link
{
	my $self = shift;
	my %args = validate(@_, {
		fromname => {type => SCALAR},
		todir=> {type => SCALAR},
		name => {type => SCALAR},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1}
		});

	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $res;

	eval {
	$self->Lookup( filename => $args{fromname});

	$res = $self->{nfs}->link(
		fhandle => $self->{lookup_fhandle},
		todir => $self->{nfs}->{mnt_fh},
		name => $args{name},
		);

	};

	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Link" );

	#return { status => $nfsRes, data => ${$res}{output} };
	return  ( status => $nfsRes );
}


sub Symlink
{
	my $self = shift;
	my %args = validate(@_, {
		name => {type => SCALAR},
		data => {type => SCALAR},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1},
                nfsver => {type => SCALAR, optional => 1},
		});

	$args{result} = "pass" if(!defined $args{result});
	my $nfsRes;
	my $res;

        my $version = $args{nfsver};
if ($version != 4) {
	eval {
	$res = $self->{nfs}->symlink(
				dhandle => $self->{nfs}->{mnt_fh},
				name => $args{name},
				data => $args{data},
		);
            };
}
else {          
       eval {
            $res = $self->{nfs}->create( 
                                 name => $args{name},
                                 type => NFLNK,
                                 data => $args{data},
                                 dhandle  => $self->{nfs}->{mnt_fh}
               );
            };
}

	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Symbolic Link" );

	#return { status => ${$res}{status}, data => ${$res}{output} };
	return ( status => $nfsRes );
}


sub DirRename
{
	my $self = shift;
	my %args = validate(@_, {
		fromname => {type => SCALAR},
		toname => {type => SCALAR},
		fromdir => {type => SCALAR},
		todir => {type => SCALAR},
		## Added By Prashant(prash@netapp.com) for negative testing
                result => {type => SCALAR, optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1}
		});

	$args{result} = "pass" if(!defined $args{result});
	
	my $nfsRes;
	my $cifsRes;
	my $res;
	
	#$res = $self->Lookup( filename => "$args{fromname}", dhandle => $self->{nfs}->{mnt_fh} );
	if($args{traffic_type} =~ /nfs/i) {
		eval {	
			$res = $self->{nfs}->rename(
					fromname => $args{fromname},
					toname => $args{toname},
					fromdir => $self->{nfs}->{mnt_fh},
					todir => $self->{nfs}->{mnt_fh},
			);
		
		};
		
		$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Directory rename" );
	
	
	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_dirname = $self->{cifs_share_drive} . "\\" . $args{fromname};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "ren $cifs_dirname $args{toname}", "dummy", -ok);
				
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Directory rename" );
	}#if cifs

	my $finalRes = $cifsRes | $nfsRes;
	
	return ( status => $finalRes, data => ${$res}{output} );
}

sub DirCreate
{
	my $self = shift;
	my %args = validate(@_, {
		directory => {type => SCALAR},
		type    => { type => SCALAR, default => NFDIR, optional => 1 },
		attr => {type => SCALAR, default => 0777, optional => 1},
		## Added By Prashant(prash@netapp.com) for negative testing
                result => {type => SCALAR, optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		nfsver => {type => SCALAR, optional => 1}
		} );

	$args{result} = "pass" if(!defined $args{result});
	my $nfsRes;
	my $cifsRes;
	my $res;
	
	my $version = $args{nfsver};
	if($args{traffic_type} =~ /nfs/i) {
	if ($version != 4) {
		eval {
			#$res = $self->Lookup( filename => $args{directory}, dhandle => $self->{nfs}->{mnt_fh} );
			$res = $self->{nfs}->mkdir(
					dhandle => $self->{nfs}->{mnt_fh},
					name => $args{directory},
					attr => { mode => $args{attr} }
				);
		}; 
	} else {
                eval {
			$res = $self->{nfs}->create(
                  			     dhandle => $self->{nfs}->{mnt_fh},
                                             name    => $args{directory}, #$dirs[0],
                                             type    => NFDIR,
                                             #cr_mode => EXCLUSIVE, #$mode,
                                             #verf    => time,
                                             attr    => { FATTR4_MODE() => 0777 }
                                            );
			}; 
	}

	$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Directory create" );

	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_dirname = $self->{cifs_share_drive} . "\\" . $args{directory};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "mkdir $cifs_dirname");
				
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Directory Create" );
	}#if cifs	
	my $finalRes = $cifsRes | $nfsRes;
	return ( status => $finalRes, data => ${$res}{output} );
}


sub DirDelete
{
	my $self = shift;
	my %args = validate(@_, {
		directory => {type => SCALAR},
		type    => { type => SCALAR, default => NFDIR, optional => 1 },
		attr => {type => SCALAR, default => 0777, optional => 1},
		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		nfsver => {type => SCALAR, optional => 1}
		} );

	$args{result} = "pass" if(!defined $args{result});
	my $nfsRes;
	my $cifsRes;
	my $res;
	my $version = $args{nfsver};
	
	if($args{traffic_type} =~ /nfs/i) { 
	
	if ( $version != 4 ) {
		eval {
			#$res = $self->Lookup( dhandle => $args{directory} );
			$res = $self->{nfs}->rmdir(
					dhandle => $self->{nfs}->{mnt_fh},
					name => $args{directory},
			);
		}; 
	} else {
		
		logcomment("@@@@@@ I am into the version 4");
		eval {
		$res = $self->{nfs}->remove(name => $args{directory},
                           dhandle => $self->{nfs}->{mnt_fh}
                           );
		};
	}

		$nfsRes  = TrafficResult ($@ , $args{result}, "NFS-Directory Delete" );
	}#if nfs
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_dirname = $self->{cifs_share_drive} . "\\" . $args{directory};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "rmdir $cifs_dirname", "dummy", -ok);
				
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Directory Delete" );
	}#if cifs
	
	my $finalRes = $cifsRes | $nfsRes;
		
	return ( status => $finalRes, data => ${$res}{output} );
}


sub GenerateAllOps
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "this_file.txt"},
		newfilename => {type => SCALAR, default => "this_file_1.txt"},
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		directory => {type => SCALAR, default => "thisDir"},
		newdirectory => {type => SCALAR, default => "thisDir1"},
#		type    => { type => SCALAR, default => NFDIR, optional => 1 },
#		attr => {type => SCALAR, default => 0777, optional => 1},
#		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added prashat to support i2p.
		i2p_path => {type => SCALAR, default => "", optional => 1},
		is_sao => {type => SCALAR, default => "", optional => 1},
		is_npc => {type => SCALAR, default => "", optional => 1},
		is_sparse => {type => SCALAR, default => "", optional => 1},
		nfsver => {type => SCALAR, default => "", optional => 1},
		snapid => {type => SCALAR, optional => 1},
	} );	
	 
       
       my $result = 0;
       my $stateid;
	   
       ## used for new pathname for i2p..
       my $newfilename_1 = "";
       my $newdirectory_1 = "";
       
              
#-----Create-----	

	my %res = $self->Create( filename => $args{filename}, traffic_type => $args{traffic_type},nfsver => $args{nfsver});
	$result |= $res{status};
		
	if($args{is_sparse} ne "") {
		### Set the NTFS sprse bit on the file.
		%res = $self->SetSparse( filename => $args{filename},result=> "pass",traffic_type => "cifs");
		$result |= $res{status};
	}
	
	# Added by prashant(prash) to support i2p.
	if($args{i2p_path} ne "") {
		$newfilename_1 = "$args{i2p_path}\/$args{filename}";
		%res = $args{polObj}->{pserver}->Verify( type => "create", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
		$result |= $res{status};
	} else {
		%res = $args{polObj}->{pserver}->Verify( type => "create", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
		$result |= $res{status};
	}

	
	#$args{polObj}->{pserver}->Verify( type => "create", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});

#----Open-------
#Open being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 
		%res = $self->Open(filename => $args{filename}, traffic_type => "cifs");
		$result |= $res{status};
		%res = $args{polObj}->{pserver}->Verify( type => "open", filename => $args{filename}, traffic_type => "cifs", result => $args{result}, is_sao => $args{is_sao});	
		$result |= $res{status};
	}
	
### For Nfsv4 open ops is supported.

	if($args{nfsver} ne "") {
		
		#logcomment ("#### Inside the Nfs version 4");
		logcomment("####### version information is $args{nfsver}");
		## Call the nfsv4 open ops.
		%res = $self->Open(filename => $args{filename}, traffic_type => "nfs",nfsver => $args{nfsver});
		$result |= $res{status};

		# Save the open information (stateid, fhandle)
		$stateid = $res{output}->{stateid};
		#my $fhandle = $res{nfsoutput}->{fhandle};
		
		## Decide on whether i2p has to be verified or not.
		if($args{i2p_path} ne "") {
			$newfilename_1 = "$args{i2p_path}\/$args{filename}";
			%res = $args{polObj}->{pserver}->Verify( type => "open", filename => $newfilename_1, traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
			$result |= $res{status};
		} else {
			%res = $args{polObj}->{pserver}->Verify( type => "open", filename => "", traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
			$result |= $res{status};
		}
	}
	
	
	
#-----Write------

	my $buf = "12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";

	%res = $self->Write (filename => $args{filename},
			 	data    => $buf,
				offset  => 0,
				traffic_type => $args{traffic_type}
			);
			
	$result |= $res{status};
	
	# Added by prashant(prash) to support i2p.
	if($args{i2p_path} ne "") {
		$newfilename_1 = "$args{i2p_path}\/$args{filename}";
			
		%res = $args{polObj}->{pserver}->Verify( type => "write", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
		$result |= $res{status};
	} else {
		#$args{filename} = "";
		%res = $args{polObj}->{pserver}->Verify( type => "write", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
		$result |= $res{status};
	}

	#$args{polObj}->{pserver}->Verify( type => "write", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});
	
	
	
#-----Read---------
#	$self->Read( filename => $args{filename} );
#	$args{polObj}->{pserver}->Verify( type => "lookup", filename => $args{filename}, result => $args{result});

	%res = $self->Read( filename => $args{filename}, traffic_type => $args{traffic_type} );
	$result |= $res{status};
	
	## Added by prashant(prash) to support i2p
	if($args{i2p_path} ne "") {
		$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				
		%res = $args{polObj}->{pserver}->Verify( type => "read", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
		$result |= $res{status};
	} else {
		#$args{filename} = "";
		%res = $args{polObj}->{pserver}->Verify( type => "read", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
		$result |= $res{status};
	}

	#$args{polObj}->{pserver}->Verify( type => "read", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result} );


	
#-----Close-------
#Close being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 	
		%res = $self->Close(filename => $args{filename}, traffic_type => "cifs");
		$result |= $res{status};
		%res = $args{polObj}->{pserver}->Verify( type => "close", filename => $args{filename} , traffic_type => "cifs", result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	}	

### For Nfsv4 close op.
	if($args{nfsver} eq "4")
	{
		logcomment("### Version information is $args{nfsver}");
		## Call the nfsv4 close op.
		%res = $self->Close(filename => $args{filename}, traffic_type => "nfs", nfsver => $args{nfsver},  stateid => $stateid);
		$result |= $res{status};

		## Decide on whether i2p has to be verified or not.
		if($args{i2p_path} ne "") {
			$newfilename_1 = "$args{i2p_path}\/$args{filename}";
			%res = $args{polObj}->{pserver}->Verify( type => "close", filename => $newfilename_1, traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
			$result |= $res{status};
		} else {
			%res = $args{polObj}->{pserver}->Verify( type => "close", filename => "", traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
			$result |= $res{status};
		}
	}
	
#-----getattr and setattr-----
		if($args{traffic_type} =~ /nfs/i) { 
			my %pghash = $self->GetAttribute (filename => $args{filename});
			$result |= $pghash{status};
			
			if($args{i2p_path} ne "") {
				$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				$args{polObj}->{pserver}->Verify( type => "getattr", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );			
				$result |= $res{status};
				
			} else {
				#$args{filename} = "";
				$args{polObj}->{pserver}->Verify( type => "getattr", filename => "", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
				$result |= $res{status};
			}
	
			#$args{polObj}->{pserver}->Verify( type => "getattr", filename => $args{filename}, result => $args{result} );
	
			my %shash = (
				atime => {
					nseconds => $pghash{object}{fattr}{atime}{nseconds} + 1,
					seconds => $pghash{object}{fattr}{atime}{seconds} + 1
				}
			);
	
			%res = $args{polObj}->{traffic}->SetAttribute (filename => $args{filename},
								atime => $shash{atime}{nseconds},nfsver => $args{nfsver}
					);
			$result |= $res{status};
			
			## Added by prahant(prash) to support i2p
			
			if($args{i2p_path} ne "") {
				$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				%res = $args{polObj}->{pserver}->Verify( type => "setattr", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
				$result |= $res{status};
			} else {
				#$args{filename} = "";
				%res = $args{polObj}->{pserver}->Verify( type => "setattr", filename => "", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
				$result |= $res{status};
			}
			
			#$args{polObj}->{pserver}->Verify( type => "setattr", filename => $args{filename}, result => $args{result} );
	
			my $ghash = $self->GetAttribute (filename => $args{filename});
			
			if($args{i2p_path} ne "") {
				#$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				#$args{polObj}->{pserver}->Verify( type => "getattr", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao});
			} else {
				#$args{filename} = "";
				#$args{polObj}->{pserver}->Verify( type => "getattr", filename => "", result => $args{result}, is_sao => $args{is_sao});
			}
			
			#$args{polObj}->{pserver}->Verify( type => "getattr", filename => $args{filename}, result => $args{result});
	 		
	 		
			my $status = $self->Verify( attribute => "atime",
					set => \%shash, get => $ghash );
	}	

	#-----Lookup---------
			## Added by prashant(prash) to support i2p
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Lookup(filename => $args{filename});
				$result |= $res{status};
				
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{filename}";
					%res = $args{polObj}->{pserver}->Verify( type => "lookup", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
					$result |= $res{status};
				} else {
				
					#$args{filename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "lookup", filename => "", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
					$result |= $res{status};
				}
				
				
				#$args{polObj}->{pserver}->Verify( type => "lookup", filename => $args{filename}, result => $args{result} );
			}
		#-----Symlink--------
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Symlink (	name => $args{newfilename}, data => $args{filename}, nfsver => $args{nfsver});
				### Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "symlink", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "symlink", filename => "", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "symlink", filename => $args{newfilename}, result => $args{result} );
			}
		
		#----Delete Symlink---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs" );
				$result |= $res{status};
				#setting the expected value to fail. But the once the RFE is fixed the symlink delete nitification should also be reported
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => $args{result}  );	
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => "fail"  );	
			}
		#----Link------------	
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Link ( fromname => $args{filename}, todir => $args{todir}, name => $args{newfilename});
				$result |= $res{status};
				
				### Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "link", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "link", filename => "", result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "link", filename => $args{newfilename}, result => $args{result} );
			}
		#----Delete link---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs");
				
				## Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => $newfilename_1, traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}  );	
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => "", traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}  );	
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => $args{result}  );	
			}
		
		#-----Rename-------
				%res = $self->Rename (fromname => $args{filename}, toname => $args{newfilename} ,
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
				$result |= $res{status};
							
				## Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {		
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "rename", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "rename", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "rename", filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{result});
				
			#-----Delete------
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => $args{traffic_type});
				$result |= $res{status};
				
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});	
					$result |= $res{status};
				} else {
					#$args{newfilename}= "";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid});	
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{result});	
				
				%res = $args{polObj}->{cli}->show();
				
			#-----Directory Ops----	
				
				%res = $self->DirCreate (	directory => $args{directory}, traffic_type => $args{traffic_type},nfsver => $args{nfsver});
				$result |= $res{status};
				
				## Added by Prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newdirectory_1 = "$args{i2p_path}\/$args{directory}";
					%res = $args{polObj}->{pserver}->Verify( type => "dircreate", filename => $newdirectory_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} , snapid => $args{snapid});
					$result |= $res{status};
				} else {
					%res = $args{polObj}->{pserver}->Verify( type => "dircreate", filename => $args{directory}, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "dircreate", filename => $args{directory}, traffic_type => $args{traffic_type}, result => $args{result} );
			
				%res = $self->DirRename (	fromname => $args{directory}, toname => $args{newdirectory},
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
							
				$result |= $res{status};
							
				## Added by Prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newdirectory_1 = "$args{i2p_path}\/$args{newdirectory}";
					%res = $args{polObj}->{pserver}->Verify( type => "dirrename", filename => $newdirectory_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				} else {
					%res = $args{polObj}->{pserver}->Verify( type => "dirrename", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "dirrename", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result} );
			
				%res = $self->DirDelete (	directory => $args{newdirectory}, traffic_type => $args{traffic_type}, nfsver => $args{nfsver});
				$result |= $res{status};
				
				## Added by Prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newdirectory_1 = "$args{i2p_path}\/$args{newdirectory}";
					%res = $args{polObj}->{pserver}->Verify( type => "dirdelete", filename => $newdirectory_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );
					$result |= $res{status};
				} else {
					%res = $args{polObj}->{pserver}->Verify( type => "dirdelete", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao}, snapid => $args{snapid} );	
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "dirdelete", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result} );	
				return ( status => $result );
}



sub GenerateAllFileOps
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "this_file.txt"},
		newfilename => {type => SCALAR, default => "this_file_1.txt"},
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added prashat to support i2p.
		i2p_path => {type => SCALAR, default => "", optional => 1},
		is_sao => {type => SCALAR, default => "", optional => 1},
		is_npc => {type => SCALAR, default => "", optional => 1},
		is_sparse => {type => SCALAR, default => "", optional => 1},
		nfsver => {type => SCALAR, default => "", optional => 1}

	} );	
	
	my $result = 0;
	## used for new pathname for i2p..
	my $newfilename_1 = "";
       	my $newdirectory_1 = "";
		
#-----Create-----	

	my %res = $self->Create( filename => $args{filename}, traffic_type => $args{traffic_type},nfsver => $args{nfsver});
	$result |= $res{status};
	
	if($args{is_sparse} ne "") {
		### Set the NTFS sprse bit on the file.
		%res = $self->SetSparse( filename => $args{filename},result=> "pass",traffic_type => "cifs");
		$result |= $res{status};
	}
	
	# Added by prashant(prash) to support i2p.
	if($args{i2p_path} ne "") {
		$newfilename_1 = "$args{i2p_path}\/$args{filename}";
		
		%res = $args{polObj}->{pserver}->Verify( type => "create", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	} else {
		
		%res = $args{polObj}->{pserver}->Verify( type => "create", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	}

	
	#$args{polObj}->{pserver}->Verify( type => "create", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});

#----Open-------
#Open being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 
		%res = $self->Open(filename => $args{filename}, traffic_type => "cifs");
		$result |= $res{status};
		%res = $args{polObj}->{pserver}->Verify( type => "open", filename => $args{filename}, traffic_type => "cifs", result => $args{result}, is_sao => $args{is_sao});	
		$result |= $res{status};
	}
	
### For Nfsv4 open ops is supported.

	if($args{nfsver} ne "") {
		
		#logcomment ("#### Inside the Nfs version 4");
		logcomment("####### version information is $args{nfsver}");
		## Call the nfsv4 open ops.
		%res = $self->Open(filename => $args{filename}, traffic_type => "nfs",nfsver => $args{nfsver});
		$result |= $res{status};
		
		#$self->Open(filename => $args{filename}, traffic_type => "nfs", nfsver => $ver);
		
		## Decide on whether i2p has to be verified or not.
		if($args{i2p_path} ne "") {
			$newfilename_1 = "$args{i2p_path}\/$args{filename}";
			%res = $args{polObj}->{pserver}->Verify( type => "open", filename => $newfilename_1, traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao});
			$result |= $res{status};
		} else {
			%res = $args{polObj}->{pserver}->Verify( type => "open", filename => "", traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao});
			$result |= $res{status};
		}
	}
	
	
	
#-----Write------

	my $buf = "12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";

	my %res = $self->Write (filename => $args{filename},
			 	data    => $buf,
				offset  => 0,
				traffic_type => $args{traffic_type}
			);
	$result |= $res{status};
	
	# Added by prashant(prash) to support i2p.
	if($args{i2p_path} ne "") {
		$newfilename_1 = "$args{i2p_path}\/$args{filename}";
			
		%res = $args{polObj}->{pserver}->Verify( type => "write", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	} else {
		#$args{filename} = "";
		%res = $args{polObj}->{pserver}->Verify( type => "write", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	}

	#$args{polObj}->{pserver}->Verify( type => "write", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});
	
	
	
#-----Read---------
#	$self->Read( filename => $args{filename} );
#	$args{polObj}->{pserver}->Verify( type => "lookup", filename => $args{filename}, result => $args{result});

	%res = $self->Read( filename => $args{filename}, traffic_type => $args{traffic_type} );
	$result |= $res{status};
	
	## Added by prashant(prash) to support i2p
	if($args{i2p_path} ne "") {
		$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				
		%res = $args{polObj}->{pserver}->Verify( type => "read", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	} else {
		#$args{filename} = "";
		%res = $args{polObj}->{pserver}->Verify( type => "read", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	}

	#$args{polObj}->{pserver}->Verify( type => "read", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result} );


	
#-----Close-------
#Close being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 	
		%res = $self->Close(filename => $args{filename}, traffic_type => "cifs");
		$result |= $res{status};
		%res = $args{polObj}->{pserver}->Verify( type => "close", filename => $args{filename} , traffic_type => "cifs", result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	}	
	
#-----getattr and setattr-----
		if($args{traffic_type} =~ /nfs/i) { 
			my %pghash = $self->GetAttribute (filename => $args{filename});
			$result |= $pghash{status};
			
			if($args{i2p_path} ne "") {
				$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				$args{polObj}->{pserver}->Verify( type => "getattr", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao} );			
				$result |= $res{status};
				
			} else {
				#$args{filename} = "";
				$args{polObj}->{pserver}->Verify( type => "getattr", filename => "", result => $args{result}, is_sao => $args{is_sao} );
				$result |= $res{status};
			}
	
			#$args{polObj}->{pserver}->Verify( type => "getattr", filename => $args{filename}, result => $args{result} );
	
			my %shash = (
				atime => {
					nseconds => $pghash{object}{fattr}{atime}{nseconds} + 1,
					seconds => $pghash{object}{fattr}{atime}{seconds} + 1
				}
			);
	
			%res = $args{polObj}->{traffic}->SetAttribute (filename => $args{filename},
								atime => $shash{atime}{nseconds},nfsver => $args{nfsver}
					);
			$result |= $res{status};
			
			## Added by prahant(prash) to support i2p
			
			if($args{i2p_path} ne "") {
				$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				%res = $args{polObj}->{pserver}->Verify( type => "setattr", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao} );
				$result |= $res{status};
			} else {
				#$args{filename} = "";
				%res = $args{polObj}->{pserver}->Verify( type => "setattr", filename => "", result => $args{result}, is_sao => $args{is_sao} );
				$result |= $res{status};
			}
			
			#$args{polObj}->{pserver}->Verify( type => "setattr", filename => $args{filename}, result => $args{result} );
	
			my $ghash = $self->GetAttribute (filename => $args{filename});
			
			if($args{i2p_path} ne "") {
				#$newfilename_1 = "$args{i2p_path}\/$args{filename}";
				#$args{polObj}->{pserver}->Verify( type => "getattr", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao});
			} else {
				#$args{filename} = "";
				#$args{polObj}->{pserver}->Verify( type => "getattr", filename => "", result => $args{result}, is_sao => $args{is_sao});
			}
			
			#$args{polObj}->{pserver}->Verify( type => "getattr", filename => $args{filename}, result => $args{result});
	 		
	 		
			my $status = $self->Verify( attribute => "atime",
					set => \%shash, get => $ghash );
	}	

	#-----Lookup---------
			## Added by prashant(prash) to support i2p
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Lookup(filename => $args{filename});
				$result |= $res{status};
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{filename}";
					%res = $args{polObj}->{pserver}->Verify( type => "lookup", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao});
					$result |= $res{status};
				} else {
				
					#$args{filename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "lookup", filename => "", result => $args{result}, is_sao => $args{is_sao});
					$result |= $res{status};
				}
				
				
				#$args{polObj}->{pserver}->Verify( type => "lookup", filename => $args{filename}, result => $args{result} );
			}
		#-----Symlink--------
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Symlink (	name => $args{newfilename}, data => $args{filename}, nfsver => $args{nfsver});
				### Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "symlink", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "symlink", filename => "", result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "symlink", filename => $args{newfilename}, result => $args{result} );
			}
		
		#----Delete Symlink---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs" );
				$result |= $res{status};
				#setting the expected value to fail. But the once the RFE is fixed the symlink delete nitification should also be reported
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => $args{result}  );	
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => "fail"  );	
			}
		#----Link------------	
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Link ( fromname => $args{filename}, todir => $args{todir}, name => $args{newfilename});
				
				### Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "link", filename => $newfilename_1, result => $args{result}, is_sao => $args{is_sao} );
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "link", filename => "", result => $args{result}, is_sao => $args{is_sao} );
				}
				
				#$args{polObj}->{pserver}->Verify( type => "link", filename => $args{newfilename}, result => $args{result} );
			}
		#----Delete link---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs");
				$result |= $res{status};
				
				## Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => $newfilename_1, traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}  );	
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => "", traffic_type => "nfs", result => $args{result}, is_sao => $args{is_sao}  );	
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => $args{result}  );	
			}
		
		#-----Rename-------
				%res = $self->Rename (fromname => $args{filename}, toname => $args{newfilename} ,
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
				$result |= $res{status};
							
				## Added by prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {		
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "rename", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
					$result |= $res{status};
				} else {
					#$args{newfilename} = "";
					%res = $args{polObj}->{pserver}->Verify( type => "rename", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "rename", filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{result});
				
			#-----Delete------
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => $args{traffic_type});
				$result |= $res{status};
				
				if($args{i2p_path} ne "") {
					$newfilename_1 = "$args{i2p_path}\/$args{newfilename}";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});	
					$result |= $res{status};
				} else {
					#$args{newfilename}= "";
					%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});	
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{result});	
				
				%res = $args{polObj}->{cli}->show();
				$result |= $res{status};
				return ( status => $result );
}


sub GenerateAllDirOps
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		directory => {type => SCALAR, default => "thisDir"},
		newdirectory => {type => SCALAR, default => "thisDir1"},
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added prashat to support i2p.
		i2p_path => {type => SCALAR, default => "", optional => 1},
		is_sao => {type => SCALAR, default => "", optional => 1},
		is_npc => {type => SCALAR, default => "", optional => 1},
		nfsver => {type => SCALAR, default => "", optional => 1}
	} );
	
	my $result = 0;
	## used for new pathname for i2p..
	my $newfilename_1 = "";
    my $newdirectory_1 = "";
		
		
#-----Directory Ops----	
				logcomment ("@@@@@@@ Version is $args{nfsver}");
				my %res = $self->DirCreate (	directory => $args{directory}, traffic_type => $args{traffic_type},nfsver =>$args{nfsver});
				$result |= $res{status};
				
				## Added by Prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newdirectory_1 = "$args{i2p_path}\/$args{directory}";
					%res = $args{polObj}->{pserver}->Verify( type => "dircreate", filename => $newdirectory_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				} else {
					%res = $args{polObj}->{pserver}->Verify( type => "dircreate", filename => $args{directory}, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "dircreate", filename => $args{directory}, traffic_type => $args{traffic_type}, result => $args{result} );
			
				%res = $self->DirRename (	fromname => $args{directory}, toname => $args{newdirectory},
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
				$result |= $res{status};
							
				## Added by Prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newdirectory_1 = "$args{i2p_path}\/$args{newdirectory}";
					%res = $args{polObj}->{pserver}->Verify( type => "dirrename", filename => $newdirectory_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				} else {
					%res = $args{polObj}->{pserver}->Verify( type => "dirrename", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "dirrename", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result} );
				
				
				%res = $self->DirDelete (directory => $args{newdirectory}, traffic_type => $args{traffic_type}, nfsver => $args{nfsver});
				$result |= $res{status};
				
				## Added by Prashant(prash) to support i2p.
				if($args{i2p_path} ne "") {
					$newdirectory_1 = "$args{i2p_path}\/$args{newdirectory}";
					%res = $args{polObj}->{pserver}->Verify( type => "dirdelete", filename => $newdirectory_1, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} );
					$result |= $res{status};
				} else {
					%res = $args{polObj}->{pserver}->Verify( type => "dirdelete", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao} );	
					$result |= $res{status};
				}
				
				#$args{polObj}->{pserver}->Verify( type => "dirdelete", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result} );	
				
				return ( status => $result );

}

sub GenerateAllOps_only
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "this_file.txt"},
		newfilename => {type => SCALAR, default => "this_file_1.txt"},
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		directory => {type => SCALAR, default => "thisDir"},
		newdirectory => {type => SCALAR, default => "thisDir1"},
#		type    => { type => SCALAR, default => NFDIR, optional => 1 },
#		attr => {type => SCALAR, default => 0777, optional => 1},
#		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added prashat to support i2p.
		nfsver => {type => SCALAR, default => "", optional => 1}
	} );	
	 
       
       my $result = 0;
       
       ## used for new pathname for i2p..
       my $newfilename_1 = "";
       my $newdirectory_1 = "";
       
              
#-----Create-----	

	my %res = $self->Create( filename => $args{filename}, traffic_type => $args{traffic_type},nfsver => $args{nfsver}, result => $args{filename});
	$result |= $res{status};
	
#----Open-------
#Open being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 
	
		%res = $self->Open(filename => $args{filename}, traffic_type => "cifs", result => $args{filename});
		$result |= $res{status};	
	}
	
### For Nfsv4 open ops is supported.

	if($args{nfsver} ne "") {
		
		logcomment ("#### Inside the Nfs version 4");
		logcomment("####### version information is $args{nfsver}");
		## Call the nfsv4 open ops.
		%res = $self->Open(filename => $args{filename}, traffic_type => "nfs",nfsver => $args{nfsver}, result => $args{filename});
		$result |= $res{status};
	}
			
#-----Write------

	my $buf = "12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";

	%res = $self->Write (filename => $args{filename},
			 	data    => $buf,
				offset  => 0,
				traffic_type => $args{traffic_type}, result => $args{filename}
			);
			
	$result |= $res{status};
		
#-----Read---------

	%res = $self->Read( filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{filename} );
	$result |= $res{status};
	
	

	
#-----Close-------
#Close being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 	
		%res = $self->Close(filename => $args{filename}, traffic_type => "cifs", result => $args{filename});
		$result |= $res{status};		
	}	
	
#-----getattr and setattr-----
		if($args{traffic_type} =~ /nfs/i) { 
			my %pghash = $self->GetAttribute (filename => $args{filename}, result => $args{filename});
			$result |= $pghash{status};
			
			my %shash = (
				atime => {
					nseconds => $pghash{object}{fattr}{atime}{nseconds} + 1,
					seconds => $pghash{object}{fattr}{atime}{seconds} + 1
				}
			);
	
			%res = $args{polObj}->{traffic}->SetAttribute (filename => $args{filename},
								atime => $shash{atime}{nseconds},nfsver => $args{nfsver}, result => $args{filename}
					);
			$result |= $res{status};
				
			my $ghash = $self->GetAttribute (filename => $args{filename}, result => $args{filename});		
	}	

	#-----Lookup---------
			## Added by prashant(prash) to support i2p
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Lookup(filename => $args{filename}, result => $args{filename});
				$result |= $res{status};
				
			}
		#-----Symlink--------
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Symlink (	name => $args{newfilename}, data => $args{filename}, nfsver => $args{nfsver}, result => $args{filename});				
			}
		
		#----Delete Symlink---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs" );
				$result |= $res{status};	
			}
		#----Link------------	
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Link ( fromname => $args{filename}, todir => $args{todir}, name => $args{newfilename}, result => $args{filename});
				$result |= $res{status};
			}
		#----Delete link---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs");
				
			}
		
		#-----Rename-------
				%res = $self->Rename (fromname => $args{filename}, toname => $args{newfilename} ,
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} , result => $args{filename});
				$result |= $res{status};
							
				
			#-----Delete------
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{filename});
				$result |= $res{status};
								
				%res = $args{polObj}->{cli}->show();
				
			#-----Directory Ops----	
				
				%res = $self->DirCreate (	directory => $args{directory}, traffic_type => $args{traffic_type},nfsver => $args{nfsver}, result => $args{filename});
				$result |= $res{status};
				
				
				%res = $self->DirRename (	fromname => $args{directory}, toname => $args{newdirectory},
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type}, result => $args{filename} );
							
				$result |= $res{status};		
				%res = $self->DirDelete (	directory => $args{newdirectory}, traffic_type => $args{traffic_type}, nfsver => $args{nfsver}, result => $args{filename});
				$result |= $res{status};				
				return { status => $result };
}

sub GenerateAllDirOpsOnly
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},	
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		directory => {type => SCALAR, default => "thisDir"},
		newdirectory => {type => SCALAR, default => "thisDir1"},
#		type    => { type => SCALAR, default => NFDIR, optional => 1 },
#		attr => {type => SCALAR, default => 0777, optional => 1},
#		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added prashat to support i2p.
		nfsver => {type => SCALAR, default => "", optional => 1}
	} );	
	 
       
       my $result = 0;
       
       ## used for new pathname for i2p..
       my $newdirectory_1 = "";
       my %res = "";
              
	   #-----Directory Ops----	
				
	   %res = $self->DirCreate (	directory => $args{directory}, traffic_type => $args{traffic_type},nfsver => $args{nfsver});
	   $result |= $res{status};		
	   %res = $self->DirRename (	fromname => $args{directory}, toname => $args{newdirectory},
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
							
	   $result |= $res{status};		
	   %res = $self->DirDelete (	directory => $args{newdirectory}, traffic_type => $args{traffic_type}, nfsver => $args{nfsver});
	   $result |= $res{status};				
	   return { status => $result };
}

sub GenerateAllFileOpsonly
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "this_file.txt"},
		newfilename => {type => SCALAR, default => "this_file_1.txt"},
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		directory => {type => SCALAR, default => "thisDir"},
		newdirectory => {type => SCALAR, default => "thisDir1"},
#		type    => { type => SCALAR, default => NFDIR, optional => 1 },
#		attr => {type => SCALAR, default => 0777, optional => 1},
#		## Added By Prashant(prash@netapp.com) for negative testing
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		## Added prashat to support i2p.
		nfsver => {type => SCALAR, default => "", optional => 1}
	} );	
	 
       
       my $result = 0;
       
       ## used for new pathname for i2p..
       my $newfilename_1 = "";
       my $newdirectory_1 = "";
       
              
#-----Create-----	

	my %res = $self->Create( filename => $args{filename}, traffic_type => $args{traffic_type},nfsver => $args{nfsver});
	$result |= $res{status};
	
#----Open-------
#Open being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 
	
		%res = $self->Open(filename => $args{filename}, traffic_type => "cifs");
		$result |= $res{status};	
	}
	
### For Nfsv4 open ops is supported.

	if($args{nfsver} ne "") {
		
		logcomment ("#### Inside the Nfs version 4");
		logcomment("####### version information is $args{nfsver}");
		## Call the nfsv4 open ops.
		%res = $self->Open(filename => $args{filename}, traffic_type => "nfs",nfsver => $args{nfsver});
		$result |= $res{status};
	}
			
#-----Write------

	my $buf = "12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";

	%res = $self->Write (filename => $args{filename},
			 	data    => $buf,
				offset  => 0,
				traffic_type => $args{traffic_type}
			);
			
	$result |= $res{status};
		
#-----Read---------

	%res = $self->Read( filename => $args{filename}, traffic_type => $args{traffic_type} );
	$result |= $res{status};
	
	

	
#-----Close-------
#Close being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
	if($args{traffic_type} =~ /cifs/i) { 	
		%res = $self->Close(filename => $args{filename}, traffic_type => "cifs");
		$result |= $res{status};		
	}	
	
#-----getattr and setattr-----
		if($args{traffic_type} =~ /nfs/i) { 
			my %pghash = $self->GetAttribute (filename => $args{filename});
			$result |= $pghash{status};
			
			my %shash = (
				atime => {
					nseconds => $pghash{object}{fattr}{atime}{nseconds} + 1,
					seconds => $pghash{object}{fattr}{atime}{seconds} + 1
				}
			);
	
			%res = $args{polObj}->{traffic}->SetAttribute (filename => $args{filename},
								atime => $shash{atime}{nseconds},nfsver => $args{nfsver}
					);
			$result |= $res{status};
				
			my $ghash = $self->GetAttribute (filename => $args{filename});		
	}	

	#-----Lookup---------
			## Added by prashant(prash) to support i2p
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Lookup(filename => $args{filename});
				$result |= $res{status};
				
			}
		#-----Symlink--------
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Symlink (	name => $args{newfilename}, data => $args{filename}, nfsver => $args{nfsver});				
			}
		
		#----Delete Symlink---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs" );
				$result |= $res{status};	
			}
		#----Link------------	
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Link ( fromname => $args{filename}, todir => $args{todir}, name => $args{newfilename});
				$result |= $res{status};
			}
		#----Delete link---
			if($args{traffic_type} =~ /nfs/i) { 
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs");
				
			}
		
		#-----Rename-------
				%res = $self->Rename (fromname => $args{filename}, toname => $args{newfilename} ,
							fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
				$result |= $res{status};
							
				
			#-----Delete------
				%res = $self->Delete(filename => $args{newfilename}, traffic_type => $args{traffic_type});
				$result |= $res{status};				
				return { status => $result };
}


sub testme
{
	package main;
	our ($FILER, $FILER_IP, $FILER_CONN, $NFSCLIENT, $NFSCLIENT_TYPE, $POLICY, $PROTOCOL, $NFSVERSION, $VOLUME, $CMDTIMEOUT, $MOUNT);

	param('FILER',       '-mesg',    "Filer to test");
	param('FILER_IP',       '-mesg',    "Filer to test_IP");
	param('FILER_CONN','-default',"#rsh");
        param('POLICY',   '-default', 'pol1');
        param('NFSCLIENT',   '-mesg', 'NFSCLIENT');
        param('NFSCLIENT_TYPE',   '-mesg', 'NFSCLIENT_TYPE');
        param('PROTOCOL',   '-mesg', 'PROTOCOL');
        param('NFSVERSION',   '-mesg', 'NFSVERSION');
        param('MOUNT',   '-mesg', 'MOUNT PATH');
        param('VOLUME',   '-mesg', 'VOLUME');
        param('CMDTIMEOUT',   '-default', '30');

	use FPolicy::Traffic;

	my $traffic = new FPolicy::Traffic(
				"traffic_type=nfs",
				"nfsclient=$NFSCLIENT",
				"nfsclienttype=$NFSCLIENT_TYPE",
				"filer_ip=$FILER_IP",
				"filer=$FILER",
				"protocol=$PROTOCOL",
				"nfs_version=$NFSVERSION",
				"volume=$VOLUME",
				"resvport=0",
				"mount=$MOUNT" );


	my %result;

	$traffic->setup();	

	my $buf = "1234567890";
	%result = $traffic->Create (	filename => "this_file.txt"	);

	logcomment("Create status = $result{status}");
	%result = $traffic->Read (	filename => "this_file.txt"	);
	logcomment("Create status = $result{status}");
	%result = $traffic->Lookup(	filename => "this_file.txt"	);
	logcomment("Create status = $result{status}");
	%result = $traffic->Write (	filename => "this_file.txt",
			 	data    => $buf,
				offset  => 0
			);
	logcomment("Create status = $result{status}");

	my %pghash = $traffic->GetAttribute (	filename => "this_file.txt"	);
	logcomment("Create status = $pghash{status}");

#	my %shash = (
#		atime => {	
#			nseconds => $pghash->{object}{fattr}{atime}{nseconds} + 1,
#			seconds => $pghash->{object}{fattr}{atime}{seconds} + 1
#			},
#		ctime => {
#			nseconds => $pghash->{object}{fattr}{ctime}{nseconds} + 1,
#			seconds => $pghash->{object}{fattr}{ctime}{seconds} + 1
#			},
#		mtime => {	
#			nseconds => $pghash->{object}{fattr}{mtime}{nseconds} + 1,
#			seconds => $pghash->{object}{fattr}{mtime}{seconds} + 1
#			},
#		size => $pghash->{object}{fattr}{size} + 1,
#		mode => "0711"
#	);
#
#	foreach my $key (keys %shash)
#	{
#		$pghash = $traffic->GetAttribute (filename => "this_file.txt");
#	%shash = (
#		atime => {	
#			nseconds => $pghash->{object}{fattr}{atime}{nseconds} + 1,
#			seconds => $pghash->{object}{fattr}{atime}{seconds} + 1
#			},
#		ctime => {
#			nseconds => $pghash->{object}{fattr}{ctime}{nseconds} + 1,
#			seconds => $pghash->{object}{fattr}{ctime}{seconds} + 1
#			},
#		mtime => {	
#			nseconds => $pghash->{object}{fattr}{mtime}{nseconds} + 1,
#			seconds => $pghash->{object}{fattr}{mtime}{seconds} + 1
#			},
#		size => $pghash->{object}{fattr}{size} + 1,
#		mode => "0777"
#		);
#
#
#
#		$traffic->SetAttribute (	filename => "this_file.txt",
#						$key => $shash{$key}
#				);
#
#		my $ghash = $traffic->GetAttribute (filename => "this_file.txt"	);
#
#		my $status = $traffic->Verify( attribute => $key,
#			set => \%shash, get => $ghash );
#
#		if($status) {
#			logresult("fail", "one / more of the attributes could not be set");
#		}else {
#			logresult("pass", "one / more of the attributes could not be set");
#		}
#	}

#	$traffic->SetAttribute (	filename => "this_file.txt",
#					mtime => $pghash->{object}{fattr}{mtime}{nseconds} + 1
#			);

	%result = $traffic->Rename (	fromname => "this_file.txt", toname => "this_file_1.txt",
				fromdir => "./", todir => "./" );
	logcomment("Create status = $result{status}");

	%result = $traffic->Delete (	filename => "this_file_1.txt"	);
	logcomment("Create status = $result{status}");

	%result = $traffic->DirCreate (	directory => "this_dir"	);
	logcomment("Create status = $result{status}");

	%result = $traffic->DirRename (	fromname => "this_dir", toname => "this_dir_1",
				fromdir => ".", todir => "." );
	logcomment("Create status = $result{status}");

	%result = $traffic->DirDelete (	directory => "this_dir_1"	);
	logcomment("Create status = $result{status}");


	%result = $traffic->DirCreate( directory => "this_dir/dir1" );
	logcomment("Create status = $result{status}");
	$traffic->cleanup();	

}

sub TrafficResult
{

	my ($outputResult, $expectedResult, $ops) = @_;
	my $retResult = 0;
	
	if( $expectedResult =~ /noresult/i ){ 
		logcomment("No Result"); 
		return 0;
	} 
	
	if ($outputResult) {
	
		if($expectedResult =~ /fail/i){
			# you expected test to fail and it failed
			#logresult("pass","$ops expected to fail");
			logcomment("PASS: $ops expected to fail");
	       		$retResult = 0;
	        }elsif($expectedResult =~ /pass/i){
	        	#logresult("fail", "$ops expected pass but failed with error $outputResult");
	        	logcomment("FAIL: $ops expected pass but failed with error $outputResult");
	        	$retResult = 1;
		}
	
	}else{
		if( $expectedResult =~ /pass/i){
			$retResult = 0;	
			#logresult("pass","$ops passed");
			logcomment("PASS: $ops passed");
		}elsif($expectedResult =~ /fail/i){
			#logresult("fail", "$ops expected to fail but passed");
			logcomment("FAIL: $ops expected to fail but passed");
			$retResult = 1;		
		}
	} #End of else
	
	return $retResult;

} #End of proc

sub GenerateNPC
{
        my $self = shift;
        my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
                dhandle => {type => OBJECT|SCALAR, optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                ## Added By Prashant(prash@netapp.com) for negative testing
                result => {type => SCALAR, optional => 1},
                optype => {type => SCALAR,optional => 1}
                });

	$args{result} = "pass" if(!defined $args{result});
	
        
        my $cifsRes;
        my $command_line;
        
        my $cifs_filename;
         
        #$cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
        #my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
        #$command_line = "SetACL.exe -on $cifs_filename /-ot file /-actn ace /-ace n/:$CIFS_DOMAIN_NETBIOS/\Administrator\;p/:change";
        
        
        my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
        
        # Construct the command line according to the optype
        
	if($args{optype} =~ /sacl/i) { 
	
		## Constrcut the commandline
		$command_line = "C:\\bin\\SetACL.exe -on $cifs_filename -ot file -actn ace -ace n:$CIFS_DOMAIN_NETBIOS\\Administrator\;p:change -ace n:$CIFS_DOMAIN_NETBIOS\\Administrator\;p:full\;m:aud_fail\;w:sacl";
		
	} elsif ($args{optype} =~ /dacl/i) { 
	
		## Constrcut the commandline
		$command_line = "C:\\bin\\SetACL.exe -on $cifs_filename -ot file -actn ace -ace n:$CIFS_DOMAIN_NETBIOS\\Administrator\;p:change";
		
	} elsif ($args{optype} =~ /owner_chane/i) {
	
		## Constrcut the commandline
		$command_line = "C:\\bin\\SetACL.exe -on $cifs_filename -ot file  -actn setowner -ownr n:$CIFS_DOMAIN_NETBIOS\\Administrator";
		
	} elsif ($args{optype} =~ /group_change/i) {
	
		## Construct the commandline.
		$command_line = "C:\\bin\\SetACL.exe -on $cifs_filename -ot file  -actn setgroup -grp n:$CIFS_DOMAIN_NETBIOS\\Administrator";
		
	} 

	
	my(undef,undef,$clientRes) = say($self->{cifs_conn}, "$command_line");
	
	logcomment("######### $clientRes");
	
	## This has particular pattern to be matched to decide on the result
	## Hence not using the library call to decide the result.
			
	#my(undef,undef,$clientRes) = say($self->{cifs_conn}, "del $filename", "dummy", -ok);
	
	if(($clientRes =~ /SetACL finished successfully/) && !($clientRes =~ /Access is denied/)) {
		$cifsRes = "pass";
		
	} elsif (($clientRes =~ /SetACL finished successfully/i) && ($clientRes =~ /Access is denied/)) {
		$cifsRes = "fail";
	}
	
	if(($args{result} =~ /fail/i) && ($cifsRes =~ /fail/i)) {
	
		logresult("pass", "You expected to fail $args{optype} op and it failed");
		
	} elsif (($args{result} =~ /pass/i) && ($cifsRes =~ /pass/i)) {
	
		logresult("pass", "You expected to pass $args{optype} op and it passed");
		
	} elsif (($args{result} =~ /pass/i) && ($cifsRes =~ /fail/i)) {
	
		logresult("fail", "You expected to pass $args{optype} op and it failed");
		
	} elsif (($args{result} =~ /fail/i) && ($cifsRes =~ /pass/i)) {
	
		logresult("fail", "You expected to fail $args{optype} op and it passed");
		
	} 
	
	#$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Create" );
		
        	
}

sub SetSparse
{
        my $self = shift;
        my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
                dhandle => {type => OBJECT|SCALAR, optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                ## Added By Prashant(prash@netapp.com) for negative testing
                result => {type => SCALAR, optional => 1}
                });

	$args{result} = "pass" if(!defined $args{result});
	
        
        my $cifsRes;
        my $command_line;
        my $clientRes;
        
        my $cifs_filename;
         
        #$cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
        #my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
        
        my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
        
        # Construct the command line according to the optype
        
        $command_line = "C:\\bin\\setsparseattr.exe $cifs_filename";
        
        
	my(undef,undef,$clientRes) = say($self->{cifs_conn}, "$command_line");
	
	logcomment("######### $clientRes");
	
			
	if ($clientRes =~ /Attributes AFTER setting the sparse bit = 0x00000220/i) {
		$cifsRes = "pass";
	} else {
		$cifsRes = "fail";
	}
	
	if(($args{result} =~ /fail/i) && ($cifsRes =~ /fail/i)) {
	
		logresult("pass", "You expected to fail setting sparse attribute and it failed");
		
	} elsif (($args{result} =~ /pass/i) && ($cifsRes =~ /pass/i)) {
	
		logresult("pass", "You expected to pass setting sparse attribute and it pass");
		
	} elsif (($args{result} =~ /pass/i) && ($cifsRes =~ /fail/i)) {
	
		logresult("fail", "You expected to pass setting sparse attribute and it failed");
		
	} elsif (($args{result} =~ /fail/i) && ($cifsRes =~ /pass/i)) {
	
		logresult("fail", "You expected to fail setting sparse attribute and it passed");
		
	} 
	
	#$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-Create" );
		        	
}

	
sub GenerateAllOpsMultipleServer
{
	my $self = shift;
	my %args = validate(@_, {
		pserverObjA => {type => OBJECT},
		pserverObjB => {type => OBJECT, optional => 1},
		samePolicy => {type => SCALAR, default => "true", optional => 1},
		filename => {type => SCALAR, default => "this_file.txt"},
		newfilename => {type => SCALAR, default => "this_file_1.txt"},
		fromdir => {type => SCALAR, default => "./"},
		todir => {type => SCALAR, default => "./"},
		directory => {type => SCALAR, default => "thisDir"},
		newdirectory => {type => SCALAR, default => "thisDir1"},
#		type    => { type => SCALAR, default => NFDIR, optional => 1 },
#		attr => {type => SCALAR, default => 0777, optional => 1},
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1}
	} );	
	my $result = 0;	
#-----Create-----	

	my %res =  $self->Create( filename => $args{filename}, traffic_type => $args{traffic_type} );
	$result |= $res{status};
	%res = $args{pserverObjA}->Verify( type => "create", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});
	my $res1 = $res{status};
	%res = $args{pserverObjB}->Verify( type => "create", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});
	if($args{samePolicy} =~ /true/i){
		$res1 &= $res{status};
	}
	else{
		$res1 |= $res{status};
	}
	$result |= $res1;
	
##----Open-------
##Open being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
#	if($args{traffic_type} =~ /cifs/i) { 
#		%res = $self->Open(filename => $args{filename}, traffic_type => "cifs");
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "open", filename => $args{filename}, traffic_type => "cifs", result => $args{result});	
#		$result |= $res{status};
#	}

	
##-----Write------
#
#	my $buf = "12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";
#
#	%res = $self->Write (filename => $args{filename},
#			data    => $buf,
#			offset  => 0,
#			traffic_type => $args{traffic_type}
#		);
#	$result |= $res{status};
#	%res = $args{pserverObjA}->Verify( type => "write", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result});
#	$result |= $res{status};
	
##-----Read---------
##	%res = $self->Read( filename => $args{filename} );
##	%res = $args{pserverObjA}->Verify( type => "lookup", filename => $args{filename}, result => $args{result});
#
#	%res = $self->Read( filename => $args{filename}, traffic_type => $args{traffic_type} );
#	$result |= $res{status};
#	%res = $args{pserverObjA}->Verify( type => "read", filename => $args{filename}, traffic_type => $args{traffic_type}, result => $args{result} );
#	$result |= $res{status};
	
##-----Close-------
##Close being tested only for CIFS. To be changed to support NFSv4 Open and Close Ops.
#	if($args{traffic_type} =~ /cifs/i) { 	
#		%res = $self->Close(filename => $args{filename}, traffic_type => "cifs");
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "close", filename => $args{filename} , traffic_type => "cifs", result => $args{result});
#		$result |= $res{status};
#	}	
	
##-----getattr and setattr-----
#	if($args{traffic_type} =~ /nfs/i) { 
#		my %pghash = $self->GetAttribute (filename => $args{filename});
#		$result |= $pghash{status};
#		%res = $args{pserverObjA}->Verify( type => "getattr", filename => $args{filename}, result => $args{result} );
#		$result |= $res{status};
#		
#		my %shash = (
#			atime => {
#				nseconds => $pghash{object}{fattr}{atime}{nseconds} + 1,
#				seconds => $pghash{object}{fattr}{atime}{seconds} + 1
#			}
#		);
#
#		%res = $args{polObj}->{traffic}->SetAttribute (filename => $args{filename},
#							atime => $shash{atime}{nseconds});
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "setattr", filename => $args{filename}, result => $args{result} );
#		$result |= $res{status};
#		my %ghash = $self->GetAttribute (filename => $args{filename});
#		$result |= $ghash{status};
#		%res = $args{pserverObjA}->Verify( type => "getattr", filename => $args{filename}, result => $args{result});
#		$result |= $res{status};
#		my $status = $self->Verify( attribute => "atime",
#			#	set => \%shash, get => $ghash );
#				set => \%shash, get => \%ghash );
#	}		
##-----Lookup---------
#	if($args{traffic_type} =~ /nfs/i) { 
#		%res = $self->Lookup(filename => $args{filename});
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "lookup", filename => $args{filename}, result => $args{result} );
#		$result |= $res{status};
#	}
##-----Symlink--------
#	if($args{traffic_type} =~ /nfs/i) { 
#		%res = $self->Symlink (	name => $args{newfilename}, data => $args{filename});
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "symlink", filename => $args{newfilename}, result => $args{result} );
#		$result |= $res{status};
#	}
#
##----Delete Symlink---
#	if($args{traffic_type} =~ /nfs/i) { 
#		%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs" );
#		$result |= $res{status};
#		#setting the expected value to fail. But the once the RFE is fixed the symlink delete nitification should also be reported
#		#%res = $args{pserverObjA}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => $args{result}  );	
#		%res = $args{pserverObjA}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => "fail"  );	
#		$result |= $res{status};
#	}
##----Link------------	
#	if($args{traffic_type} =~ /nfs/i) { 
#		%res = $self->Link ( fromname => $args{filename}, todir => $args{todir}, name => $args{newfilename});
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "link", filename => $args{newfilename}, result => $args{result} );
#		$result |= $res{status};
#	}
##----Delete link---
#	if($args{traffic_type} =~ /nfs/i) { 
#		%res = $self->Delete(filename => $args{newfilename}, traffic_type => "nfs");
#		$result |= $res{status};
#		%res = $args{pserverObjA}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => "nfs", result => $args{result}  );	
#		$result |= $res{status};
#	}	
#-----Rename-------
	%res = $self->Rename (fromname => $args{filename}, toname => $args{newfilename} ,
				fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
	$result |= $res{status};
	%res = $args{pserverObjA}->Verify( type => "rename", filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{result});
	$result |= $res{status};
#-----Delete------
	%res = $self->Delete(filename => $args{newfilename}, traffic_type => $args{traffic_type});
	$result |= $res{status};
	%res = $args{pserverObjA}->Verify( type => "delete", filename => $args{newfilename}, traffic_type => $args{traffic_type}, result => $args{result});	
	$result |= $res{status};
	
##	$args{polObj}->{cli}->show();
	
##-----Directory Ops----	
#	
#	%res = $self->DirCreate (	directory => $args{directory}, traffic_type => $args{traffic_type}	);
#	$result |= $res{status};
#	%res = $args{pserverObjA}->Verify( type => "dircreate", filename => $args{directory}, traffic_type => $args{traffic_type}, result => $args{result} );
#	$result |= $res{status};
#	%res = $self->DirRename (	fromname => $args{directory}, toname => $args{newdirectory},
#				fromdir => $args{fromdir}, todir => $args{todir}, traffic_type => $args{traffic_type} );
#	$result |= $res{status};
#	%res = $args{pserverObjA}->Verify( type => "dirrename", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result} );
#	$result |= $res{status};
#	%res = $self->DirDelete (	directory => $args{newdirectory}, traffic_type => $args{traffic_type}	);
#	$result |= $res{status};
#	%res = $args{pserverObjA}->Verify( type => "dirdelete", filename => $args{newdirectory}, traffic_type => $args{traffic_type}, result => $args{result} );	
#	$result |= $res{status};
#	
	return ( status => $result );
}

###  Added by Rajaram
=head1 NAME
Create_ADS - This API is used to create an ADS on a file.

=head2 DESCRIPTION
This API creates an ADS(alternate data stream) on a file given as input.

=head2 ROUTINES
	$pol->{traffic}->Create_ADS("filemane=newfile.txt", "ADSname=newADSfile.doc", "traffic_type=cifs");
	$pol->{traffic}->Create_ADS("filemane=newfile.txt", "ADSname=newADSfile.doc", "traffic_type=cifs", "result=fail");
=over

=item filename
Mandatory parameter, indicates the name of the file on which the ADS has to be created
values = string

=item ADSname
mandatory parameter, indicates the name of the ADS
values = inclusion/exclusion

=item traffic_type
Specifies the type of traffic, whether nfs or cifs
optional, if the values was initialized during object instantiation
values = cifs  #To be supported for NFS later

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub Create_ADS
{
    my $self = shift;
    my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
				ADSname => {type => SCALAR, default => "auto_test_ads.txt"},
				traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                result => {type => SCALAR,default => "pass", optional => 1},
				});

    my $cifsRes;
    my $cifs_ADSname = $args{ADSname};
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo sometext > $cifs_filename" . ":" . "$cifs_ADSname");
			
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-CREATE-ADS" );
	}	
    
	return (status => $cifsRes) ; # Comment: Return fhandle
}

##  Added by Rajaram
sub Read_ADS
{
    my $self = shift;
    my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
				ADSname => {type => SCALAR, default => "auto_test_ads.txt"},
				traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                result => {type => SCALAR,default => "pass", optional => 1},
                });

    my $cifsRes;
    my $cifs_ADSname = $args{ADSname};
	my $temp_clientRes;
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "more < $cifs_filename" . ":" . "$cifs_ADSname");
		$temp_clientRes = $clientRes;
		
		if (!($clientRes =~ /Access is denied/i)){    
			$clientRes = 0; # Pass
		}

		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-READ-ADS" );
	}	
    
	return (status => $cifsRes, data => "$temp_clientRes") ; # Comment: return data
}

##  Added by Rajaram
sub Open_ADS
{
    my $self = shift;
    my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
				ADSname => {type => SCALAR, default => "auto_test_ads.txt"},
				traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                result => {type => SCALAR,default => "pass", optional => 1},
                });

    my $cifsRes;
    my $cifs_ADSname = $args{ADSname};
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "more < $cifs_filename" . ":" . "$cifs_ADSname");
			
		if (!($clientRes =~ /Access is denied/i)){    
			$clientRes = 0; # Pass
		}
		
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-OPEN-ADS" );
	}	
    
	return (status => $cifsRes) ;
}


##  Added by Rajaram
sub Write_ADS
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		ADSname => {type => SCALAR, default => "auto_test_ads.txt"},
		data => {type => SCALAR, default => "1234567890qwertyuiopasdfghjklzxcvbnm", optional => 1 },
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		result => {type => SCALAR,default => "pass", optional => 1}
		});


	my $cifsRes;
	my $buf = $args{data};

	my $filename = $args{filename};
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo $buf > $cifs_filename" . ":" . "$args{ADSname}");		
	
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-WRITE-ADS" ); # Comment: CIFS-Write-ADS
		
        	
	}#if cifs
		
	return (status => $cifsRes );
}

##  Added by Rajaram
sub Close_ADS
{
    my $self = shift;
    my %args = validate(@_, {
                filename => {type => SCALAR, default => "auto_test.txt"},
				ADSname => {type => SCALAR, default => "auto_test_ads.txt"},
				traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
                result => {type => SCALAR,default => "pass", optional => 1},
                });


    my $cifsRes;
    my $cifs_ADSname = $args{ADSname};
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $args{filename};
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "more < $cifs_filename" . ":" . "$cifs_ADSname");
			
		if (!($clientRes =~ /Access is denied/i)){    
			$clientRes = 0; # Pass
		}
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-CLOSE-ADS" );
	}	
    
	return (status => $cifsRes) ;
}

##  Added by Rajaram
sub Delete_ADS
{
	my $self = shift;
	my %args = validate(@_, {
		filename => {type => SCALAR, default => "auto_test.txt"},
		ADSname => {type => SCALAR, default => "auto_test_ads.txt"},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
        result => {type => SCALAR, default => "pass", optional => 1}
		});
		
	#$args{result} = "pass" if(!defined $args{result});
	
	my $cifsRes;
	my $filename = $args{filename};
	
	# streams.exe should be present in C:\bin\
	my $commandline = $DELETE_NT_STREAMS_PATH . "\\" . $DELETE_NT_STREAMS . " -d ";
	
	if($args{traffic_type} =~ /cifs/i) { 
	
		my $cifs_filename = $self->{cifs_share_drive} . '\\' . $filename; 
		my(undef,undef,$clientRes) = say($self->{cifs_conn}, "$commandline" . "$cifs_filename");
		
		#logcomment ( "$clientRes");
		if($clientRes =~ /Deleted :$args{ADSname}:/i )
		{
			$clientRes  = 0 #Comment: pass 0 or 1
		}
		else
		{
			$clientRes  = 1
		}
		
		$cifsRes  = TrafficResult ($clientRes , $args{result}, "CIFS-DELETE-ADS" );
	}
	return (status => $cifsRes);
}


###  Added by Rajaram
=head1 NAME
Generate_Verify_AllADSOps - This API is used to perform operations on ADS attached to files or directories and verify its response from the server.

=head2 DESCRIPTION
This API creates an ADS(alternate data stream) on a file given as input and performs operations like reading, writing, closing, and deleting it..

=head2 ROUTINES
	$pol->{traffic}->Generate_Verify_AllADSOps("polObj=pol1","filemane=newfile.txt", "ADSname=newADSfile.doc", "traffic_type=cifs");
	$pol->{traffic}->Generate_Verify_AllADSOps("polObj=pol1","filemane=newfile.txt", "ADSname=newADSfile.doc", "traffic_type=cifs", "result=fail");
=over

=item filename
Mandatory parameter, indicates the name of the file on which the ADS has to be created
values = string

=item ADSname
mandatory parameter, indicates the name of the ADS
values = inclusion/exclusion

=item traffic_type
Specifies the type of traffic, whether nfs or cifs
optional, if the values was initialized during object instantiation
values = cifs  #To be supported for NFS later

=item result
(Optional) - PASS / FAIL, expected result of the API.
Required when the API is expected to fail

=cut
sub Generate_Verify_AllADSOps
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "this_file.txt"},
		ADSname => {type => SCALAR, default => "this_file_ADS.txt"},
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		is_sao => {type => SCALAR, default => "", optional => 1},
	} );	
	
	my $result = 0;
	
	#Comment: Check if traffic type is CIFS 
	if($args{traffic_type} =~ /cifs/i) {
	
		
	#-----Create-----	

		my %res = $self->Create_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type});
		$result |= $res{status};
		
		%res = $args{polObj}->{pserver}->Verify( type => "create", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};

	#------Open------
		%res = $self->Open_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type});
		$result |= $res{status};
			
		%res = $args{polObj}->{pserver}->Verify( type => "open", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
		
	#-----Write------

		my $buf = "12345678";

		%res = $self->Write_ADS(filename => $args{filename},
					ADSname => $args{ADSname},
					data    => $buf,
					traffic_type => $args{traffic_type}
				);
		$result |= $res{status};
		
		%res = $args{polObj}->{pserver}->Verify( type => "write", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	
	#------Read------
	%res = $self->Read_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type});
	$result |= $res{status};
		
	%res = $args{polObj}->{pserver}->Verify( type => "read", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
	$result |= $res{status};
		
	#-----Close-------
		%res = $self->Close_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type});
		$result |= $res{status};
			
		%res = $args{polObj}->{pserver}->Verify( type => "close", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};

	#-----Delete------
		%res = $self->Delete_ADS(filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type});
		$result |= $res{status};

		%res = $args{polObj}->{pserver}->Verify( type => "delete", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_sao => $args{is_sao});
		$result |= $res{status};
	}
	return ( status => $result );
}

sub Generate_AllADSOps_Only
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "this_file.txt"},
		ADSname => {type => SCALAR, default => "this_file_ADS.txt"},
		result => {type => SCALAR, default => "pass", optional => 1},
		traffic_type => {type => SCALAR, default => $self->{traffic_type}, optional => 1},
		is_sao => {type => SCALAR, default => "", optional => 1},
	} );	
	
	my $result = 0;
	
	#Comment: Check if traffic type is CIFS 
	if($args{traffic_type} =~ /cifs/i) {
	
		
	#-----Create-----	

		my %res = $self->Create_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type},result => $args{result});
		$result |= $res{status};


	#------Read------
		%res = $self->Read_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type},result => $args{result});
		$result |= $res{status};

		
	#------Open------
		%res = $self->Open_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type},result => $args{result});
		$result |= $res{status};

		
	#-----Write------

		my $buf = "12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";

		%res = $self->Write_ADS(filename => $args{filename},
					ADSname => $args{ADSname},
					data    => $buf,
					result => $args{result},
					traffic_type => $args{traffic_type}
				);
		$result |= $res{status};

				
	#-----Close-------
		%res = $self->Close_ADS( filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type}, result => $args{result});
		$result |= $res{status};


	#-----Delete------
		%res = $self->Delete_ADS(filename => $args{filename},ADSname => $args{ADSname}, traffic_type => $args{traffic_type}, result => $args{result});
		$result |= $res{status};

	}
	return ( status => $result );
}

sub CreateFile
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "rrdfile.txt"},
		filesize => {type => SCALAR, default => 10000, optional => 1},
		data => {type => SCALAR, default => "1234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890",optional => 1},
		volume => {type => SCALAR, default => $self->{volume}, optional => 1},
		result => {type => SCALAR, optional => 1}
		});
	
	my $buf = $args{data};
	my $data;
	my $totaldatasize = 0;
	my $offset = 0;
	my $count = length($buf);
	
	#Create the file
	my $filer_conn = $args{polObj}->{cli}->{CONN};
	my $clientRes = sayEOF($filer_conn, "priv set diag;mkfile $args{filesize} \/vol\/$args{volume}\/home\/$args{filename}");
	if (($clientRes =~ /file creation failed/i)){    
		$clientRes = 1; # Fail
	}
	my $result |= $clientRes;
	
	#Set the sparse bit on the file.
	my %res = $self->SetSparse( filename => $args{filename},result=> "pass",traffic_type => "cifs");
	$result |= $res{status};
		
	#Create a copy of the  file (nfpc would read from this file)
	my %res = $self->Create( filename => $args{filename}."_copy", traffic_type => "nfs");
	$result |= $res{status};
	

	#Write data into the file
	# while ($offset < $args{filesize}) 
	# {
		# $data = $data . $buf; 
		# $offset = $offset + $count;
	# }
	
	# my %res = $self->Write(filename => ($args{filename}."_copy"), data => $data, traffic_type => "nfs");
	# $result |= $res{status};
	
	while ($offset < ($args{filesize} - length($buf))) 
	{
		$data = "";
		#group data into chunks of 30k
		while(length($data) < 30000 && $totaldatasize < ($args{filesize} - length($buf)))
		{
			$data = $data . $buf;
			$totaldatasize = $totaldatasize + length($buf);
		}
		my %res = $self->Write(filename => ($args{filename}."_copy"), data => $data, offset => $offset, traffic_type => "nfs");
		$result |= $res{status};
		
		$offset = $offset + length($data);
	}
	
	return $result;
}

sub ReadFile
{
	my $self = shift;
	my %args = validate(@_, {
		polObj => {type => OBJECT},
		filename => {type => SCALAR, default => "rrdfile.txt"},
		dhandle => {type => OBJECT|SCALAR, optional => 1},
		length => {type => SCALAR, default => 4096, optional => 1},
		offset => {type => SCALAR, default => 0, optional => 1},
		i2p_path => {type => SCALAR, default => "", optional => 1},
		is_sao => {type => SCALAR, default => "", optional => 1},
		is_rrd => {type => SCALAR, default => "", optional => 1},
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		result => {type => SCALAR, default=> "pass", optional => 1}
		});

	my $filename = $args{filename};
	my $tmpNfsFile = "tmp_nfs_".$args{filename};
	my $tmpCifsFile = "tmp_cifs_".$args{filename};
	my %res;
	my $final_res;
	my $nfsRes;
	my $cifsRes;	
	my $nfs_read_data;
	my $result = 0;
	
	if($args{traffic_type} =~ /nfs/i) 
	{	
		my $dhandle = $args{dhandle} || $self->{read_fhandle} || $self->{nfs}->{mnt_fh} ;
		my $offset = $args{offset};
		my $length = $args{length};
		my $stateid = 0;
	    
		print "\nReading file handle=$dhandle for length=$length bytes from offset $offset:\n";
	
		#Create a temp file to store data read through NFS
		%res = $self->Create( filename => $tmpNfsFile, traffic_type => "nfs", nfsver => $self->{nfs_version});
		$result |= $res{status};
		
		my $nfs_EOF = 0;
		$self->{lookup_fhandle} = $dhandle;
		while (($nfs_EOF == 0))      # Do till EOF is reached
		{
			# Read a block
			eval {
			$self->Lookup( filename => $filename );
			$nfs_read_data = $self->{nfs}->read
				(
					fhandle => $self->{lookup_fhandle},
					offset => $offset,
					count => $length,
					stateid => $stateid
				);
			};

			# Verify the read data	
			if($args{i2p_path} ne "")
			{
				my $newfilename_1 = "$args{i2p_path}\/$args{filename}";
				
				%res = $args{polObj}->{pserver}->Verify( type => "read", filename => $newfilename_1, traffic_type => $args{traffic_type}, result => $args{result},is_rrd => $args{is_rrd}, is_sao => $args{is_sao});
				$result |= $res{status};
			} 
			else
			{
				%res = $args{polObj}->{pserver}->Verify( type => "read", filename => "", traffic_type => $args{traffic_type}, result => $args{result}, is_rrd => $args{is_rrd}, is_sao => $args{is_sao});
				$result |= $res{status};
			}
		
			#Write the block into a temp file
			%res = $self->Write( filename => $tmpNfsFile, data => ($nfs_read_data->{data}), offset  => $offset, traffic_type => "nfs");
			$result |= $res{status};
			
			$nfs_EOF = $nfs_read_data->{eof};
			$offset = $offset + $length;
		} # End of while
	
	}#if nfs
	$nfsRes = $result;
	
	# CIFS read
	if($args{traffic_type} =~ /cifs/i)
	{
		my $cifs_filename = $self->{cifs_share_drive} . "\\" . $filename;
		my $tmp_cifs_file = $self->{cifs_share_drive} . "\\" . $tmpCifsFile;
		my(undef,undef,$cifs_read_data) = say($self->{cifs_conn}, "type $cifs_filename > $tmp_cifs_file",-timeout, 1200);		
		
		my %res = $args{polObj}->{pserver}->Verify( type => "read", filename => $filename, traffic_type => "cifs", result => $args{result}, is_rrd => $args{is_rrd}, is_sao => $args{is_sao});
		$result |= $res{status};
		
		if (!($cifs_read_data =~ /Access/i)){    
			my $cifsRes = 0; #Pass
		}
		else
		{
			$cifsRes = 1; # Fail
		}
		$cifsRes  = TrafficResult($cifsRes , $args{result}, "CIFS-Read" );
	
	}#if cifs
	$cifsRes = $result;
	
	$final_res = $cifsRes | $nfsRes;
	return $final_res;
}


sub Comparefiles
{		
		my $self = shift;
		my %args = validate(@_, {
		filename => {type => SCALAR, default => "rrdfile.txt"},
		tmpfilename => {type => SCALAR, optional => 1 },
		traffic_type => {type => OBJECT|SCALAR, default => $self->{traffic_type}, optional => 1},
		result => {type => SCALAR, default => "pass", optional => 1}
		});
		
		my $filename1;
		my $filename2;
		my $tmpfile;
		my $result;
		
		if($args{traffic_type} =~ /nfs/i) 
		{
			if(!defined $args{tmpfilename})
			{
				$tmpfile = "tmp_nfs_".$args{filename};
			}
			else
			{
				$tmpfile = $args{tmpfilename};
			}	
			$filename1 = $self->{cifs_share_drive} . "\\" . $args{filename} . "_copy";
			$filename2 = $self->{cifs_share_drive} . "\\" . $tmpfile;

			my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo n | comp $filename1 $filename2 /A", -timeout, 1200 );
		
			if ($clientRes =~ /Files compare OK/i){    
				$clientRes = 0; 
				logcomment("Files Match");
			}
			else
			{
				$clientRes = 1;
				logcomment("Files dont Match");
			}
			
			$result |= TrafficResult ($clientRes , $args{result}, "FILE MATCHING");
			my %res = $self->Delete(filename => $tmpfile, traffic_type => "nfs");
			$result |= $res{status};
		}
		
		if($args{traffic_type} =~ /cifs/i) 
		{
			if(!defined $args{tmpfilename})
			{
				$tmpfile = "tmp_cifs_".$args{filename};
			}
			else
			{
				$tmpfile = $args{tmpfilename};
			}
		
			$filename1 = $self->{cifs_share_drive} . "\\" . $args{filename} . "_copy";
			$filename2 = $self->{cifs_share_drive} . "\\" . $tmpfile;

			my(undef,undef,$clientRes) = say($self->{cifs_conn}, "echo n | comp $filename1 $filename2 /A", -timeout, 1200 );
		
			if ($clientRes =~ /Files compare OK/i){    
				$clientRes = 0; 
				logcomment("Files Match");
			}
			else
			{
				$clientRes = 1;
				logcomment("Files dont Match");
			}
			
			$result |= TrafficResult ($clientRes , $args{result}, "FILE MATCHING");
			my %res = $self->Delete(filename => $tmpfile, traffic_type => "cifs");
			$result |= $res{status};
		}
		
		# Delete the _copy file
		my %res = $self->Delete(filename => ($args{filename} . "_copy"), traffic_type => "cifs");
		$result |= $res{status};
		
	return $result;
}

sub launch_fastfilewalk()
{
	my $self = shift;
	my %args = validate(@_, {
                dirname => {type => SCALAR, default => "fpdir"},
                });
				
	my $cmd = "c:\\bin\\fastfilewalk.exe " . "$self->{cifs_share_drive}" . "\\" . "$args{dirname}" . " -m all";
	say($self->{cifs_conn}, "cmd /c $cmd", '-blind');
	sleep(5);
	logcomment("Check presence of Fastfilewalk");
	my (undef,undef,$status) = say($self->{cifs_conn}, "cmd /c tasklist /FI \"IMAGENAME eq fastfilewalk.exe\"", "dummy", -ok);
	
	if($status =~ /No tasks/i)
	{
		logcomment("FAIL: Fastfilewalk was not launched");
		return 1;
	}
	else
	{
		logcomment("PASS: Launched fastfilewalk");
		return 0;
	}
	
}

sub cleanup_fastfilewalk()
{
	my $self = shift;
	
	logcomment("Check presence of Fastfilewalk");
	my(undef,undef,$msgc) = say($self->{cifs_conn}, "cmd /c tasklist \/FI \"IMAGENAME eq fastfilewalk.exe\"", "dummy", -ok);

	if($msgc =~ /No tasks/i) {   #sandeepp No tasks are running
		;
	}
	else {
		logcomment("Kill the process fastfilewalk.exe");
		(undef,undef,$msgc) = say($self->{cifs_conn}, "cmd /c taskkill \/F \/IM fastfilewalk.exe", "", -ok);
	}
	
}

1;
