# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Cmode/Utils.pm#1 $
#
#Copyright (c) 2009 Network Appliance, Inc.
#All rights reserved.
## @summary Routines for performing basic Cmode operations .
##
## @description
##    General utilities for C-mode automation.
##
## @synopsis
##    use FS::Cmode::Utils;
##
## @author nraju (neelkanth.raju@netapp.com), prashanc (prashanth.chennagiri@netapp.com)
## @status review
#

package FS::Cmode::Utils;

use strict;
use warnings;

use Data::Dumper;
use Exporter qw(import);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF GLOBREF BOOLEAN);
use Tharn;

use Utils::Units qw(:constants);

use constant DF_SPACE   => 'space';
use constant DF_INODES  => 'inodes';
use constant DF_PRIVATE  => 'private inodes';
use constant DF_RESERVE => 'reserve';

my %DF_REPORT_TYPE_OPTS = (
        UNIT_KB() => '-kb',
        UNIT_MB() => '-mb',
        UNIT_GB() => '-gb',
        UNIT_TB() => '-tb',
);
#  Note: 'UNIT_B' is available for (but only for) DF_INODES space type, handled in code.

my %DF_SPACE_TYPE_OPTS = (
        DF_SPACE()    => '',
        DF_INODES()   => '-i',
        DF_PRIVATE()  => '-i -P',
        DF_RESERVE()  => '-r'
);

BEGIN {
    our @EXPORT = qw( &cmd_wait_or_check_state
        &snapshot_create
        &mount_setup
        &mount_delete
        &disk_free
        &vol_modify_and_rename
      );

    use vars qw(
        $Destination
    );

    $Destination = "0.0.0.0/0";

} ## end BEGIN

#======== cmd_wait_or_check_state ==============================================
# name ::       cmd_wait_or_check_state
# summary  :    Executes a particular type of show cmd and polls or check for the given 'field'
#               to reach the given 'value' depending on value of parameter 'wait' .
# Required options:
#               server => Server object
#               cmd    => command to execute
#               cmd implemented :
#                       aggregate_show => storage_aggregate_show
#                       volume_show    => volume_show
#                       wafliron_show  => storage_aggregate_wafliron_show
#
#               options=> reference of arguments to be passed to the show function
#               field  => attribute of show for which polling has to be done
#               value  => state of field to wait for..
#               wait   => 1 if polling required
#                         0 if only check required.(default)
#-------------------------------------------------------------------------------
sub cmd_wait_or_check_state (%) {
    my %opts    = @_;
    my $server  = $opts{server};
    my $field   = $opts{field};
    my $value   = $opts{value};
    my $cmd     = $opts{cmd};
    my $options = $opts{options};
    my ($t_val, $output, $flag) = (0, "", 1);
    my ($timeout);
    my @option = %$options;

    if ($opts{wait}) {
        $timeout = $opts{timeout} ? $opts{timeout} : $main::TIMEOUT;
    } else {
        $timeout = 0;
    }
    if ($cmd =~ /aggregate_show/) {
        $output = $server->storage_aggregate_show(@option,-Timeout => 100);
    } elsif ($cmd =~ /volume_show/) {
        $output = $server->volume_show(@option);
    } elsif ($cmd =~ /wafliron_show/) {
        $output = $server->storage_aggregate_wafliron_show(@option);
    } elsif ($cmd =~ /volume_quota_show/) {
        $output = $server->volume_quota_show(@option);
    }


    $t_val = time;
    $flag = 0 if (${$$output[0]}{$field} =~ /$value/);
    $timeout += $t_val;

    while ($flag and $t_val < $timeout) {
        sleep 5;    # delay between polling
        if ($cmd =~ /aggregate_show/) {
            $output = $server->storage_aggregate_show(@option, -Timeout => 100);
        } elsif ($cmd =~ /volume_show/) {
            $output = $server->volume_show(@option, -Timeout => 100);
        } elsif ($cmd =~ /wafliron_show/) {
            $output = $server->storage_aggregate_wafliron_show(@option);
        } elsif ($cmd =~ /volume_quota_show/) {
            $output = $server->volume_quota_show(@option);
        }

        $flag = 0 if (${$$output[0]}{$field} =~ /$value/);
        $t_val = time;
    } ## end while ($flag and $t_val <...

    if ($flag and $t_val >= $timeout) {
        die "Waiting for $field to reach $value Timed Out";
    }
    return 1;
} ## end sub cmd_wait_or_check_state (%)
################################################################################

#============== snapshot_create ================================================
# create a snapshot in a given volume
# return hash reference to the created snapshot's properties
#-------------------------------------------------------------------------------
sub snapshot_create(%) {
    my %opts = validate(@_,
                        {"server"    => { type=>OBJECT, isa=>'Server' },
                         "snapshot" => {type => SCALAR},
                         "volume" => {type => SCALAR, default => $main::FILER_VOL, optional=> 1},
                         "vserver" => {type => SCALAR, default => $main::VSERVER, optional=> 1},
                        });
    my $server = $opts{server};
    delete $opts{server};
    my $timeout = $main::TIMEOUT ? $main::TIMEOUT : 60;
    $server->volume_snapshot_create(%opts, -Timeout => $timeout);
    sleep 10;
    my @ret_val = $server->volume_snapshot_show(%opts);
    logcomment(" Snapshot $opts{snapshot} Created successfully ");
    return $ret_val[0];
} ## end sub snapshot_create
################################################################################

#=========== mount_setup =======================================================
# This sub routine performes mounting a volume on the client.
#-------------------------------------------------------------------------------
sub mount_setup(%) {
    my %opts = @_;
    my @out;
    my ($routing_group, @response, @cmd_output);

    my $server    = $opts{server};
    my $vserver   = $opts{vserver};
    my @nodes     = $server->cluster_show;
    my $Node      = $opts{node} ? $opts{node} : $nodes[0]->{node};
    my $Host      = host($Node) or die($@);
    my $Address   = $opts{address} ? $opts{address} : $Host->{DATA_IP1};
    my $Port      = $opts{port} ? $opts{port} : $Host->{DATA_PORT1};
    my $Netmask   = $opts{netmask} ? $opts{netmask} : $Host->{DATA_MASK1};
    my $Gateway   = $opts{gateway} ? $opts{gateway} : $Host->{DATA_GATEWAY1};
    my $lif       = $opts{lif} ? $opts{lif} : "data1";
    $Destination  = $opts{destination} if( $opts{destination} );

    logcomment("\$Node ====> $Node\n");
    logcomment("\$Port ====> $Port\n");
    logcomment("\$Address ====> $Address\n");
    logcomment("\$Netmask ====> $Netmask\n");
    logcomment("\$Gateway ====> $Gateway\n");

    logcomment(  "====== Verifying existance of lif data1 ."
                  . " if present deleting the same.");
    eval {
        $server->network_interface_show(server =>$vserver,
                                        lif => $lif,
                                        'home-node'=> $Node,
                                        'home-port'=> $Port);
    };
    if ($@) {
        ###creating network interface
        eval {
            $server->network_interface_create(server         => $vserver,
                                              lif            => $lif,
                                              role           => "data",
                                              'home-node'    => $Node,
                                              'home-port'    => $Port,
                                              address        => $Address,
                                              netmask        => $Netmask,
                                              'status-admin' => "up");

        };
        #if unable to create interface with data_ip1 then try using data_ip2.
        if ($@) {
            $Address = $Host->{DATA_IP2};
            $Port    = $Host->{DATA_PORT1};
            $server->network_interface_create(server         => $vserver,
                                              lif            => $lif,
                                              role           => "data",
                                              'home-node'    => $Node,
                                              'home-port'    => $Port,
                                              address        => $Address,
                                              netmask        => $Netmask,
                                              'status-admin' => "up");
        } ## end if ($@)
    } ## end of external if($@)

    @response = $server->network_routing_groups_show(server => $vserver);
    $routing_group = ${$response[0]}{'routing-group'};
    logcomment("======The routing group is $routing_group");

    #### Creating Routing-Group route.

    eval {
        $server->network_routing_groups_route_show(server => $vserver,
                                                   'routing-group' => $routing_group,
                                                   destination     => $Destination,
                                                   gateway         => $Gateway);
    };
    if ($@) {
        ##If routing-groups-route not present then creating it
        $server->network_routing_groups_route_create(
                                            server          => $vserver,
                                            'routing-group' => $routing_group,
                                            destination     => $Destination,
                                            gateway         => $Gateway);
    }

    #####creating export policy
    eval {
        @out =
            $server->vserver_export_policy_rule_show(vserver    => $vserver,
                                                     policyname => "default");
    };
    if ($@) {
        $server->vserver_export_policy_rule_create(vserver    => $vserver,
                                                   policyname => "default",
                                                   clientmatch  => $Destination,
                                                   rorule       => "any",
                                                   rwrule       => "any",
                                                   superuser    => "any",
                                                   anon         => "anon",
                                                   'allow-suid' => "true",
                                                   'allow-dev'  => "true");
    } else {
        $server->vserver_export_policy_rule_modify(vserver    => $vserver,
                                                   ruleindex  => $out[0]->{ruleindex},
                                                   policyname => "default",
                                                   clientmatch  => $Destination,
                                                   rorule       => "any",
                                                   rwrule       => "any",
                                                   superuser    => "any",
                                                   anon         => "anon",
                                                   'allow-suid' => "true",
                                                   'allow-dev'  => "true");
    }

    # Adding Base license with hardcoded value
    $server->system_license_add('license-code' => "VMZIVZGKWVMADF");

    # Adding NFS license with hardcoded value depending on the base license added above
    $server->system_license_add('license-code' => "BAKJZZGKWVMADF");

    eval { $server->vserver_nfs_show(vserver => $opts{vserver} ); };
    if ($@) {
       $server->vserver_nfs_create(vserver => $vserver, access => "true");
    }

    return $Address;

} ## end sub mount_setup
###############################################################################

####### mount_delete ##########################################################
sub mount_delete(%) {
    my %opts   = @_;
    my $server = $opts{server};
    my $lif    = $opts{lif} ? $opts{lif} : "data1";
    my (@response, $routing_group,$policy_name );

    ###deleting network interface
    $server->network_interface_delete(server => $opts{vserver},
                                      lif    => $lif);

    ##### Getting the value of Routing-group
    @response =
          $server->network_routing_groups_show(server => $opts{vserver});
    $routing_group = ${$response[0]}{'routing-group'};
    logcomment("======The routing group is $routing_group");

    #### Deleting Route
    $server->network_routing_groups_route_delete(
                                          server          => $opts{vserver},
                                          'routing-group' => $routing_group,
                                          destination     => $Destination );

    #### Deleting Routing-group
    $server->network_routing_groups_delete(server => $opts{vserver},
                                           'routing-group' => $routing_group
                                          );

    $server->vserver_nfs_delete(vserver => $opts{vserver});

    ###### Getting the rulesets of a vserver
    @response = $server->vserver_export_policy_rule_show(vserver => $opts{vserver});
    logcomment("Vserver show output is:\n" . Dumper(\@response));
    $policy_name = ${$response[0]}{'policyname'};
    if ($policy_name =~ /default/){
        $server->vserver_export_policy_rule_delete(vserver    => $opts{vserver},
                                                   policyname => $policy_name,
                                                   ruleindex  => ${$response[0]}{'ruleindex'});
    }

    return 1;
} ## end sub mount_delete
###############################################################################

#============== disk_free =====================================================
#
#   Runs df command on the specified volume/path. Below is the description of the
#   parameters of this subroutine.
#
#    server
#        Server object to the node.
#    node
#        Filer on which the 'df' command should be run
#    type
#       type of conatiner to run on,  default:vol
#    vol
#        name of volume on which 'df' needs to be run
#    aggr
#        name of aggregate on which to run df.
#    path
#        This is the name of the volume or path.
#    space_type
#        The types of statistics to be reported. There are three types:
#        DF_SPACE, DF_INODES, DF_PRIVATE and DF_RESERVE. Default is DF_SPACE.
#    report_type
#        Units in which the above type of statistics are reported.
#        Currently, there are following units: UNIT_KB, UNIT_MB, UNIT_GB and UNIT_TB.
#        The default is UNIT_MB.
#    snapshot
#        If the value of this field is non-zero, the statistics are
#        calculated for the snapshot of the requested volume/path.  The default is 0.
#    sync
#        'sync' option is either 0 or 1. If 0, no delay or sync is performed
#        If 1, then delay for sync_delay seconds is done.
#    sync_delay
#        delay in seconds to wait after issuing 'sync' to the node.
#
#================================================================================
#
sub disk_free(%) {
    my %args = validate(@_, {
            server      => { type=>OBJECT, isa=>'Server' },
            node        => { type => SCALAR,   default => 'local' ,  optional=>1 },
            snapshot    => { type => SCALAR,   default => 0     ,    optional=>1 },
            type        => { type => SCALAR,   default => 'vol' ,    optional=>1 },
            path        => { type => SCALAR,   default => ''    ,    optional=>1 },
            vol         => { type => SCALAR,   default => $::FILER_VOL, optional=>1 },
            aggr        => { type => SCALAR,   default => ''    ,    optional=>1 },
            space_type  => { type => SCALAR,   default => DF_SPACE,  optional=>1 },
            report_type => { type => SCALAR,   default => UNIT_MB,   optional=>1 },
            snapshot    => { type => BOOLEAN,  default => 0,         optional=>1 },
            sync        => { type => BOOLEAN,  default => 1,         optional=>1 },
            sync_delay  => { type => SCALAR,   default => $::DF_SYNC_DELAY,  optional=>1 },
        } );

    my $path = $args{path} || $args{aggr} ;
    my $sync = $args{sync};
    my $sync_delay = $args{sync_delay};
    my $cmd= "df";
    my $node = $args{node};
    my $server = $args{server};

    unless ($path) {
        $path = '/vol/'.$args{vol};
    }
    if ($args{snapshot}) {
        if ($path =~ m{/$}) {
            $path = $path . ".snapshot";
        } else {
            $path = $path . "/.snapshot";
        }
    }

    # If sync is requested, issue 'sync' command and sleep
    # for 30 seconds to force CP and then take df output.
    # Default for 'sync' is now 1. See BURT 306218
    if($sync || $sync_delay) {
        $server->system_node_run(node    => $node,
                                 command => "sync");
        sleep($sync_delay);
    }

    my $space_type = $DF_SPACE_TYPE_OPTS{ $args{space_type} };
    if(!defined($space_type)) {
        die "Unrecognized space type: $args{space_type}\n";
    }

    my $report = $DF_REPORT_TYPE_OPTS{ $args{report_type} };

    # 'UNIT_B' only availailable for 'DF_INODES' space type.
    if ($args{space_type} eq DF_INODES  &&  $args{report_type} eq UNIT_B) {
        $report = '';
    }
    my ($rt) = $report =~ /-(\w+)/;
    $rt = uc($rt);
    $report = $report . ' -A' if (($args{type} eq "aggr") or $args{aggr});

    if(!defined($report)) {
        die "Unrecognized report type: $args{report_type}\n";
    }

    # Run the df command

    my $text = $server->system_node_run(node    => $node,
                                        command => "$cmd $space_type $report $path");

    if ($text =~ m/No such file or directory/) {
            die("cmd_disk_free Error -- : $2");
    }
    if ($text =~ /invalid option/i) {
            die("cmd_disk_free Error -- $text");
    }

    # Parse the output recieved form df

    if ($args{space_type} eq DF_INODES) {
         # match header first, and then look for the entry
         if ($text !~ s|^\s*\w+\s+iused\s+ifree\s+%iused||m) {
             die "Error parsing df output (new format?)\n" .
                 'Expected ^\s*\w+\s+iused\s+ifree\s+%iused' .
                 ", but got $text";
         }

         if ($text =~ m|^\s*$path/?\s+(\d+)\s+(\d+)\s+(\d+)%|m) {
             return wantarray ?
                 ($1, $2, $3) :
                 { iused => $1, ifree => $2, percent_iused => $3 };
        }
        die "Could not find $path in df output";

    } elsif ($args{space_type} eq DF_PRIVATE) {
        # match header first, and then look for the entry
        if ($text !~ s|^\s*\w+\s+iused\s+ifree\s+%iused||m) {
            die "Error parsing df output (new format?)\n" .
                'Expected ^\s*\w+\s+iused\s+ifree\s+%iused'.
                ", but got $text";
        }

        if ($text =~ m|^\s*$path/?\s+(\d+)\s+(\d+)\s+(\d+)%|m) {
            return wantarray ?
                   ($1, $2, $3) :
                   { iused => $1, ifree => $2, percent_iused => $3 };
        }
        die "Could not find $path in df output";

    } elsif ($args{space_type} eq DF_RESERVE) {
        # match header first, and then look for the entry
        if ($text !~ s|^\s*\w+\s+total\s+used\s+avail\s+reserved||m) {
            die "Error parsing df output (new format?)\n" .
                'Expected ^\s*\w+\s+total\s+used\s+avail\s+reserved' .
                ", but got $text";
        }

        if ($text =~ m|^\s*$path/?\s+(\d+)$rt\s+(\d+)$rt\s+(\d+)$rt\s+(\d+)$rt|m) {
            return wantarray ?
                ($1, $2, $3, $4) :
                { total => $1, used => $2, avail => $3, reserved => $4 };
        }
        die "Could not find $path in df output";

    } else {
        # match header first, and then look for the entry
        if ($text !~ s|^\s*\w+\s+total\s+used\s+avail\s+capacity||m) {
            die "Error parsing df output (new format?)\n" .
                'Expected ^\s*\w+\s+total\s+used\s+avail\s+capacity' .
                ", but got $text";
        }
        if ($text =~ m|\s*$path/?\s+(\d+)$rt\s+(\d+)$rt\s+(\d+)$rt\s+(\d+)|m) {
            return wantarray ?
                ($1, $2, $3, $4) :
                { total => $1, used => $2, avail => $3, capacity => $4 };
        }
        die "Could not find $path in df output";
    }
}
##################################################################################

#============== vol_modify_and_rename ============================================
# Executes a volume modify function for given state
# Required options:
#               state => Flex volume state to modify
#               vol_name => Volume name to modify
#               new_vol => New volume name to rename the old vol
#=================================================================================

sub vol_modify_and_rename(%) {

    my %opts = validate(@_, {
        server   => { type => OBJECT,   isa => 'Server' },
        vserver  => { type => SCALAR,   default => $main::VSERVER  , optional=>1 },
        volume   => { type => SCALAR,   default => $main::FILER_VOL, optional=>1 },
        state    => { type => SCALAR,   default => ''              , optional=>1 },
        newname  => { type => SCALAR,   default => ''              , optional=>1 },
        size     => { type => SCALAR,   default => ''              , optional=>1 },
     } );
    my $server   = $opts{server};
    my $vserver  = $opts{vserver};
    my $vol_name = $opts{volume};
    my $new_vol  = $opts{newname};
    my ($value, $field );

    if ( $opts{state} ) {
        $value = $opts{state};
        $field = 'state';
    } elsif ($opts{size}) {
        $value = $opts{size};
        $field = 'size';
    }

    # Modify flex volume to required state
    logcomment("Modify $field of flex volume to $value");

    $server->volume_modify(volume  => $vol_name,
                           vserver => $vserver,
                           $field  => $value);

    # Wait for flex volume to be given state
    if ($opts{state}) {
        logcomment("Wait for flex volume to be given state");
        &cmd_wait_or_check_state(server  => $server,
                                 cmd     => "volume_show",
                                 options => {volume => $vol_name, vserver => $vserver},
                                 wait    => 1,
                                 field   => $field,
                                 value   => $value
                                );
    }
    # Renaming flex volume
    if($new_vol) {

        logcomment("Renaming flex volume");
        $server->volume_rename(volume  => $vol_name,
                               vserver => $vserver,
                               newname => $new_vol);

        # Validation: Verify that flex volume is renamed
        logcomment("Validation : Verify that flex volume is renamed");

        $server->volume_show(vserver => $vserver, volume => $new_vol, vserver => $vserver);
    }

    return 1;

}    # end of sub vol_modify_and_rename
################################################################################

1;
