# $Id: //depot/prod/test/Rsierranevada/lib/FS/Dumpblock.pm#1 $

## @name Dumpblock.pm
## @summary Dumpblock having the subroutine used by
## REP(Raid Error Propagation) Tests
## @status public
## @author Hardikkumar.Darji@netapp.com
## @pod here

=head1 NAME  
                 Dumpblock.pm    
                

=head1 DESCRIPTION
                Dumpblock -   Library that supports  dumpblock operation
                with different options using NACL APISet

=head1 ROUTINES
                Dumpblock
=cut


# Compiler directives.
use strict;
use warnings;

# =============================================================
# Module Import
# =============================================================

require Exporter;

package FS::Dumpblock;
use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF OBJECT BOOLEAN UNDEF);
use NACL::APISet;
use vars qw (@ISA @EXPORT_OK);
use Tharn qw($Log);
use Data::Dumper;
use POSIX;
use NACL::APISet::Exceptions::ResponseException qw(:try);

# =============================================================
# Components Used
# =============================================================


@ISA       = qw(Exporter);
@EXPORT_OK = qw(dumpblock);

# =============================================================
# Declare Global Variables in BEGIN
# =============================================================

BEGIN {
    use vars qw(
        $Api_Set_Obj
        );
}



#######################################################

## @name ::dumpblock(api_obj       => $api_obj,
##                   block_type => $block_type,
##                   volume     => $volume,
##                   type       => 1,
##                   file       => $file,
##                   fileid     => $fileid,
##                   private    => 1,
##                   blockno    => $blockno,
##                   level      => $level,
##                   snapname   => $snapname,
##                   snapid     => $snapid,
##                   dbn        => $dbn,
##                   vbn        => $vbn,
##                   diskno     => $diskno,
##                   cksum      => 1,
##                   block      => $block)
##
## @summary
##
##     Execute the dumpblock command on the filer and returns the output
##
## @description
##
##     This function runs the dumpblock command on the filer. The command
##     usage for dumpblock is:
##
##     dumpblock ifile [-A | -V] <vol-name> <fileid> [<snapid> [-b]]
##       - dump block of inofile containing that inode
##     dumpblock file [-A | -V] <vol-name> <fileid> [-private]  <blockno> [-no_abort] [<snapid>]
##       - dump data block <blockno> from file inode # <fileid> <snapid>
##     dumpblock vbn [-A | -V] <vol-name> <vbn>
##       - dump block with the specified vbn on the specified volume/aggregate
##     dumpblock dbn [-c] <disk_name> <dbn>
##       - dump block with the specified dbn on the specified disk
##     dumpblock info [-A | -V] <vol-name> <fileid> [-private] <blockno>
##                                                  [<level>] [<snapid>]
##       - dump information about the specified block
##
##     The output of the command is parsed and stored in hash. A reference to
##     the hash is returned on success.
##
## @arg %opts Required
##
##     This hash contains all the parameters that are passed to the dumpblock
##     command. Individual parameters are specified below:
##
##     api_obj => ApiSet Object Required
##     This is the NACL APISet object to get the node name and run the system_node_run command
##
##     block_type => 'dumpblock type' Required
##     The dumpblock type that has to be executed. Valid values are ifile, file,
##     vbn, dbn or info.
##
##     volume => 'volume or aggr name' Optional
##     The volume name on which the dumpblock command has to be run.
##
##     type => 'aggr|vol' Optional
##     This value denotes if the volume parameter is an aggregate or volume.
##     If the value of type is 'aggr' then -A will be prefixed to the name of aggregate . Default is 'vol'
##
##     file => 'filename' Optional
##     The filename of the file against which dumpblock has to be run.
##     If filename is provided and fileid is not provided, the fileid will
##     be obtained by calling filer_showfh.
##
##     fileid => 'fileid' Optional
##     The fileid of the file against which dumpblock has to be run.
##     The filename parameter can be provided instead of this, and the function
##     will take care of finding the fileid.
##
##     noabort  => '-no_abort' Optional
##     Option to run dumpblock to trigger non-abortable messages.
##     Only this option works on RR or later builds.
##
##     private  => '-private' Optional
##     Private option to access the private/meta level information.
##     This option only works on BR.1 or later builds.
##
##     blockno => 'blockno' Optional
##     Block number of the file for which dumpblock is being run.
##
##     level => 'level' Optional
##     Block level of the file being used for dumpblock
##
##     snapname => 'snapshot name' Optional
##     Name of the snapshot, for which the snapid will be passed to dumpblock.
##     This function will find out the snapid of the provided snapname and pass
##     it to dumpblock
##
##     snapid => 'snapid' Optional
##     Snapid that will be passed to dumpblock. Snapname parameter can be used
##     instead of this, the function will get the required snapid and pass
##     it to dumpblock
##
##     dbn => 'dbn' Optional
##     The value of dbn that is to be passed to dumpblock
##
##     vbn => 'virtual block number' Optional
##     The virtual block number which will be passed to dumpblock
##
##     diskno => 'disk number' Optional
##     The disk identifier that will be passed to dumpblock
##
##     cksum => '0|1' Optional
##     The '-c' option will be passed to dumpblock if this parameter is set to 1
##     This is valid only for 'dumpblock dbn'
##
##     block => '0|1' Optional
##     Setting this parameter to 1 will set the flag '-b' while calling
##     dumpblock. This is valid only for
##     'dumpblock ifile'
##
## @return value
##     Returns errcode = 0 - Success
##                      < 0 - Failure
##                      a hashref that contains the content of the block
##                      that was read using dumpblock. Also contains diskid
##                      ,VBN and other values that are returned by dumpblock.
##
## @example
##     my ($errcode,$result) = dumpblock(api_obj=>$CLI_Obj,
##                                       block_type => 'info',
##                                       volume => 'testvol',
##                                       fileid => '101',
##                                       blockno => '0',
##                                       level => '0');
##     The above example will run the following command on the filer:
##     dumpblock info testvol 101 0 0


sub dumpblock(%) {

    my %opts = validate(@_,
                        {api_obj => { type => OBJECT },
                         block_type => {type  => SCALAR,
                                        regex => qr/^(ifile|file|vbn|dbn|info)$/},
                         volume     => {type => SCALAR, optional => 1},
                         type       => {type => SCALAR, optional => 1, 
                                        default => 'vol'},
                         file       => {type => SCALAR, optional => 1},
                         fileid     => {type => SCALAR, optional => 1},
                         noabort    => {type => SCALAR, optional => 1},
               	         private    => {type => SCALAR, optional => 1},
                         blockno    => {type => SCALAR, optional => 1},
                         level      => {type => SCALAR, optional => 1},
                         snapname   => {type => SCALAR, optional => 1},
                         snapid     => {type => SCALAR, optional => 1},
                         dbn        => {type => SCALAR, optional => 1},
                         vbn        => {type => SCALAR, optional => 1},
                         diskno     => {type => SCALAR, optional => 1},
                         cksum      => {type => SCALAR, optional => 1},
                         block      => {type => SCALAR, optional => 1}});

    my ($Api_Set_Obj,$block_type, $cmd,$volume, $fileid, $blockno, $level, $status, $noabort,
        $snapid, $dbn,  $vbn,  $diskno, $cksum,  @info_arr, $block, $dump_out, $private, $errcode);
    
    $Api_Set_Obj =  $opts{api_obj};
    $block_type = $opts{block_type};
    $volume     = ($opts{type} eq 'aggr') ? '-A ' . $opts{volume} : $opts{volume};
    $fileid     = $opts{fileid};
    $noabort    = ($opts{noabort} == 1) ? '-no_abort' : undef;
    $private    = ($opts{private} == 1) ? '-private' : undef;
    $blockno    = $opts{blockno};
    $level      = $opts{level};
    $snapid     = $opts{snapid};
    $dbn        = $opts{dbn};
    $vbn        = $opts{vbn};
    $diskno     = $opts{diskno};
    $cksum      = ($opts{cksum} == 1) ? '-c' : '';
    $block      = ($opts{block} == 1) ? '-b' : '';
    $errcode = 0;
    ## Getting the fileid, if filename is supplied instead of fileid
    if (!defined($opts{fileid}) && defined($opts{file})) {
       my %showfh = &_filer_showfh($Api_Set_Obj, $opts{file});
       if(!defined($showfh{error}))
       {
            $fileid = hex($showfh{fileid});
       }
       else
       {
            $errcode = -1;
            $dump_out="File path provided is not correct";
            return ($errcode,$dump_out);
       }
    }

    # Getting the snapid, if snapname is specified instead of snapid
    if (!defined($opts{snapid}) && defined($opts{snapname})) {
        my @snaps = &_snapshot_status($Api_Set_Obj, $volume);
        foreach my $snaprec (@snaps) {
            if ($opts{snapname} eq $$snaprec{name}) {
                $snapid = $$snaprec{snapid};
                last;
            }
        }
    }

    # Building the command
    $cmd = "dumpblock $block_type ";

    if ($block_type eq 'info') {
       if(defined($private)) {
          $cmd .= "$volume $fileid $private $blockno $level $snapid";
        }
        else {
             $cmd .= "$volume $fileid $blockno $level $snapid";
        }
    } elsif ($block_type eq 'dbn') {
        $cmd .= "$cksum $diskno $dbn";
    } elsif ($block_type eq 'vbn') {
        $cmd .= "$volume $vbn";
    } elsif ($block_type eq 'file') {
         if (defined($noabort) || defined($snapid) || defined($private)) {
            $cmd .= "$volume $fileid $private $blockno $noabort $snapid";
         }
         else {
            $cmd .= "$volume $fileid $blockno";
         }
    } elsif ($block_type eq 'ifile') {
        $cmd .= "$volume $fileid $snapid $block";
    }
    
    my $resp = $Api_Set_Obj->system_node_run(node => "local", command => "\" priv set -q diag; $cmd \"");
    
    $dump_out= Dumper($resp->get_raw_output());
    
    # If there is a syntax error, return 0.
    if ($dump_out =~ m/dumpblock|block not found/) {
        $@ = $dump_out;
        $errcode = -1;
        return ($errcode,$dump_out);
    }
    
    # supressing ($VAR1 = ') (;) and (') from the output of dumpblock command
    $dump_out=~s/\$VAR1 = '//;
    $dump_out=~s/'//;
    $dump_out=~s/;//;
    
    
    # Parse the output based on the $block_type
    if ($block_type eq 'info') {

        my %dump_hash;
        
        # Storing the output of dumpblock into an array
        my @arr = split(/\n/, $dump_out);
       
        ## removing array elelemts that are null
        # Case 1: the $dump_out starts with following 
        # $VAR1 = '  
        # ok
        #
        # Case 2: the $dump_out starts with following 
        # $VAR1 = 'Ok 
        
        # In the above Case 1 when we substitute the white space
        # for the string "$VAR1 = '" we get first array elelemt as null.
        # in Case 2 after substitution, the first array element is ("ok")
        # which is non-null string.
        
        # in the following logic we will find out first
        # non-null elelmet of array which will be used as $status
        
        
        my ($cnt,$unused);
        $cnt=0;
        foreach(@arr)
        {
            if($_ ne "")
            {
                last;
            }
            else
            {
                $cnt=$cnt+1;
            }
        }
        
        for (my $i=0;$i<$cnt;$i++)
        {
            # removing elelements which are null
            # from the start of the array
            $unused =shift @arr;
        }
        
        # saving the first not null element as status of dumpblock
        $status =shift @arr;
        
        $Log->comment("status=$status");
        
        foreach my $str (@arr)
        {            
            push(@info_arr, split(/ /, $str));                        
        }
        
        %dump_hash = @info_arr;
        $dump_hash{status} = $status;

        return ($errcode,\%dump_hash);
        

    } elsif ($block_type =~ m/dbn|vbn|file|ifile/) {
        my $dump_hash = _parse_dumpblock($dump_out);
        
        return ($errcode,$dump_hash);
    }
} ## end sub dumpblock(%)

sub _parse_dumpblock($) {
    my ($dump_out) = @_;
    my (%dump_hash, @dmp);
        
    # Storing the output of dumpblock into an array
    @dmp = split(/\n/, $dump_out);

    # Replacing ^M characters with ''
    @dmp = map { s/\cM//g; $_ } @dmp;

    # Parsing the first line. This line contains diskno, dbn, vbn etc.
    # These values will be placed into the hash dump_hash
    
    # Looking for comma separated list of parameters
    if ($dmp[0] =~ m/,/) {
        
        # Using comma as the delimiter for parameter-value pairs
        my @spl = split(/,/, $dmp[0]);
        foreach (@spl) {
            m/(\w+)\s+=?\s*([\w\.]+)/;
            $dump_hash{$1} = $2;
        }
    }
    elsif ($dmp[1] =~ m/,/) {
        
        # Using comma as the delimiter for parameter-value pairs
        my @spl = split(/,/, $dmp[1]);
        foreach (@spl) {
            m/(\w+)\s+=?\s*([\w\.]+)/;
            $dump_hash{$1} = $2;
        }
    }

    # Saving the contents of the block into the hash
    $dump_hash{block_contents} = \@dmp;
    
    return (\%dump_hash);
} ## end sub _parse_dumpblock($)


=head2  _filer_showfh($$)
 @summary
     get information about file

 @description
     executes showfh command on filer and get information about file
     like fileid,snapid,flags,gen,fsid
     
 @arg $Api_set_Obj Required:
      $Api_set_Obj => NACL APISet object.
     
 @arg $file Required:
      $file => path of file for which info. is required.
 @return
      %fh if file exist then returns hash  of (key , value) of file related information
          if file do not exist the hash contains only one pair ('error','error description')
=cut
sub _filer_showfh($$)
{
    my $Api_Set_Obj = shift;
    my($file) = shift;

    my $cmd = "showfh $file"; 
    my $resp = $Api_Set_Obj->system_node_run(node => "local", command => "\" priv set -q diag; $cmd \"");
    
    my $text= Dumper($resp->get_raw_output());
    
    my %fh;
    if ($text =~ /flags=(\S+)\s+snapid=(\S+)\s+fileid=(\S+)\s+gen=(\S+)\s+fsid=(\S+)/)
    {
       $fh{flags}=$1;
       $fh{snapid}=$2;
       $fh{fileid}=$3;
       $fh{gen}=$4;
       $fh{fsid}=$5;
   
    }
    else
    {
        $fh{error}=$text;
        
    }
    return %fh;
}

    
sub _snapshot_status($$) {
	my ($api_obj, $volume) = @_;
	my @snapshots = ();
        my $resp = $api_obj->system_node_run(node => "local", command =>"\" priv set -q diag;snap status $volume\""); 
	my $text = Dumper($resp->get_raw_output());
	croak($text) if( $text =~ /does not exist or is not online/ );	

	my @hkeys = ();
	my $sncnt=0;
	foreach( split( '[\n\r]',$text) ) {
		chomp; s/^\s+//;
		s/(\w\w\w)\s+(\d\d)\s+(\d\d:\d\d)/$1_$2_$3/;	# work around dates
		@hkeys = split('\s+',$_) if( /snapid/ );
		if( /^[0-9]/ ) {
			my $i=-1;
			foreach( split('\s+',$_) ) {
				s/_/ /g if($hkeys[++$i] =~ /date/ );	# restore space in dates
				$snapshots[$sncnt]->{ $hkeys[$i] } = $_ if($hkeys[$i]);
			}
			$sncnt++;
		}
	}
	return(@snapshots);
}
1;
