# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/EA/CifsSetup.pm#1 $

## @summary CIFS Setup routines
## @author dl-rakesh-automation@netapp.com
## @status Shared
## @pod here

=head1 NAME

CifsSetup - Filer setup routines for CIFS.

=head1 COPYRIGHT

Copyright 2011, NetApp Inc., All Rights Reserved.


=head1 SYNOPSIS

    use CifsSetup;
    my $cifs_setup = new CifsSetup( hostname => $Node->name() );
    my $ret = $cifs_setup->is_cifs_setup();
    $cifs_setup->cifs_setup(local_admins=>.....);


=head1 DESCRIPTION

    This library contains CIFS setup routines for 7-mode only.

=head1 DEPENDENCIES

    Since component support is currently not available for
    7-mode cifs cmds, execute_raw_command is used instead of
    component. See Burt : 477353


=head1 SEE-ALSO

    Network::Nis::nis_configure

=head1 ROUTINES

=cut

package CifsSetup;

use strict;
use warnings;
use NACL::APISet;
use Tharn qw(sleep $Log host connect);
use FS::commonParams;
use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF BOOLEAN UNDEF OBJECT);
use Admin::License qw(license);
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NATE::BaseException qw(:try);

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => 'hostname',
    scalar => 'root_password',
    scalar => 'set_passwd',
    scalar => [ { -type => 'NACL::APISet' }, 'filer_api_console' ],
    scalar => [ { -type => 'NACL::APISet' }, 'filer_api_rsh' ]
];

sub init {
    my ( $self, %opts ) = @_;
    my %v_args = validate_with(
        params => \%opts,
        spec   => { hostname => { type => SCALAR } }
    );
    my $host_obj = host( $v_args{hostname} );
    my $connection_rsh = $host_obj->connect( template => "rsh" );
    $connection_rsh->timeout($main::CMDTIMEOUT)
        if ( defined $main::CMDTIMEOUT );
    $self->filer_api_rsh(
        NACL::APISet->new(
            category   => "Node",
            interface  => "CLI",
            set        => "7Mode",
            connection => $connection_rsh
        )
    );
    my $connection_console = $host_obj->connect( template => "console" );
    $connection_console->timeout($main::CMDTIMEOUT)
        if ( defined $main::CMDTIMEOUT );
    $self->filer_api_console(
        NACL::APISet->new(
            category   => "Node",
            interface  => "CLI",
            set        => "7Mode",
            connection => $connection_console
        )
    );
    $self->root_password( $connection_console->hostp()->default_password() );
    license( $connection_console, "cifs" )
        or NACL::APISet::Exceptions::CommandFailedException->throw(
        "Failed to add cifs license");
} ## end sub init

=head2 is_cifs_setup

    use CifsSetup;
    my $cifs_setup_obj = new CifsSetup( hostname => $Node->name() );
    $cifs_steup_obj->is_cifs_setup()

Check if CIFS is already setup on filer. This is only for 7-Mode.
The routine returns 1 if CIFS is setup already and 0 otherwise.

=cut

sub is_cifs_setup {

    my $pkg = shift;
    $Log->comment("Checking if CIFS is already configured on filer");

    #$pkg->filer_api_console()->set_privilege( "privilege-level" => "test" );
    #my $out = $pkg->filer_api_console()
    #    ->execute_raw_command( command => "cifs sessions" );
    my $out = $pkg->filer_api_console()
        ->execute_raw_command( command => "cifs sessions" );
    return ( ( $out =~ /(not licensed|not running)/s ) ? 0 : 1 );
} ## end sub is_cifs_setup

=head2 cifsshare_delete

    use CifsSetup;
    my $cifs_setup_obj = new CifsSetup( hostname => $Node->name() );
    $cifs_steup_obj->cifsshare_delete( $apiset, $sharename );

Delete a cifs share by using apiset. If a cifsshare does not exist, do
nothing. 

=over

=item apiset

(Required) : APISet used to run CIFS operations on node.

=item sharename

(Required) : User specified CIFS share name to be deleted.

=back

=cut

sub cifsshare_delete {

    my ( $pkg, %opts ) = @_;

    my %args = validate_with(
        params => \%opts,
        spec   => {
            apiset      => { type => OBJECT, optional => 0 },
            sharename   => { type => SCALAR, optional => 0 },
        }
    );
    my $apiset      = $args{apiset};
    my $sharename   = $args{sharename};
    $Log->comment("Deleing CIFS share $sharename");

    try {
        $apiset->cifs_shares(
            delete      => 1,
            "sharename" => $sharename
        );
    }
    catch NATE::BaseException with {
        my $exception_object = shift;
        if ($exception_object->text() =~ /There is no share named $sharename/) {
            $Log->comment("$sharename does not exist");
        }
        else {
            $exception_object->throw();
        }
    };

} ## end sub is_cifs_setup

=head2 cifs_setup

    use CifsSetup;
    my $cifs_setup_obj = new CifsSetup( hostname => $Node->name() );
    $cifs_steup_obj->cifs_setup( %opts );

A routine to setup CIFS on the filer. This is only for 7-Mode.
The routine returns 1 if CIFS setup is successfully done and 0 otherwise.


This test uses, then disconnects, the "#console" connection. (non-interactive
"cifs setup" exists, but it is way too hidden and limited and not expected to
be used by customers, so it cannot be trusted)

Some of the things to be done for this routine

add support for nt4

migrate to using the zapi style. Basically

    * change domain to domain-name
    * change domain_user to login-user
    * change domain_pass to login-password
    * change root_pass to default-root-password
    * implement pdc-ip-address
    * implement server-name
    * in ZAPI workgroup name is not specified - how do they do that ???
      - The domain-name field serves as the workgroup name. ~~ Garrett

    * Actually, you should just use the ZAPIs themselves, eh? ~~ Garrett

The routine takes the following named arguments

=over

=item auth-type

(Optional) : The authentication style that determines the
method by which clients will be authenticated when connecting
to this CIFS server.
Possible values: 'ad'(Active Directory), 'workgroup'(Workgroup),
'passwd'(/etc/passwd).  Default: 'workgroup'.
This routine does not support nt4 as yet. When it does 'nt4' will be
a possible value.

Note: The domain, domain_user and domain_pass need to be specified
if auth-type is ad.

=item workgroup

(Optional) : The workgroup of which the filer would be part of.
Defaults to "WORKGROUP" if not specified and auth-type is 'workgroup'.

=item domain

(Optional) : The domain name of which the filer would be a part of.
If the domain name is specified then the domain_user and domain_pass
must also be specified.

=item domain_user

(Optional) : The username on the domain with administrative
privileges to add a filer to it.

=item domain_pass

(Optional) : The password of the "domain_user".

=item root_pass

(Optional) : A scalar specifying the root password of the filer.
Defaults to the hostspec "default_password".

=item cifs_pass

(Required) : The password for the cifs administrative user.
Deafults cifs password used is cifspass123

=item dns_enable

(Optional) : If defined will configure DNS when prompted by cifs setup.
If defined must provided 'dns_domainname' and 'dns_server_list' also.

=item dns_domainname

(Optional) : Filer's domain name. Must be defined if 'dns_enable' is defined.

=item dns_server_list

(Optional) : Comma separted list of DNS server IP(s).
Must be defined if dns_enable is defined.

=item timed_enable

(Optional) : If defined will configure time services when prompted by cifs setup.
If defined must provided 'timed_server_list' also.

=item timed_server_list

(Optional) : Comma separted list of time server names or IP(s).
Must be defined if timed_enable is defined.

=item passwd_group

(Optional) : If set then create the default /etc/passwd and/or /etc/group files.
Default is 'no'.

=item vfiler

(Optional) : If variable vfiler is defined, then cifs  setup is done on the vfiler

=item server-name

(Optional) : CIFS Server name.

=item security-style

(Optional) : The security style determines wheter or not the CIFS service will
support multiprotocol access. These styles are subject to change in a future
release. Possible values : 'ntfs','multiprotocol'. Default: 'multiprotocol'

=item create_admin

(Optional) : If set to 1 will create the CIFS administrator account if prompted.
Possible values : 'yes' / 'no'.  Default : 'yes'

=item ou-name

(Optional) : The distinguished name of the organizational unit that the CIFS
service will become a member of. If cifs.new_setup.enable = on, the filer
will join the 'CN=Computers' OU if ou-name is not found in the list of
available OUs. Note: This information is only used when the auth-type is 'ad'.

=item select_site_name

(Optional) : If set to 'yes' will enable selection of the AD site name if prompted.
Possible values : 'yes' / 'no'. Default : 'no'.

=item site-name

(Optional) : The name of the site to choose when select_site_name is yes. If
select_site_name = yes and site-name is not defined the Default-First-Site-Name
will be selected.

=item wins_option

(Optional) : Not supported when cifs.new_setup.enable = on. If set to 1 WINS
will be enabled using the existing WINS servers if they exist otherwise the
wins servers will be set to wins_server_list. If set to 2, WINS will be
enabled and any existing WINS servers will be overwritten with
wins_server_list. If set to 3 WINS will be disabled.
Possible values : 1,2,3. Default : 3.";

=item wins_servers

(Optional) : Not supported when cifs.new_setup.enable = on. A comma seperated
list of WINS servers.

=item local_admins

(Optional) : User or group ID to add to the BUILTIN\adminisrators group

=back

=cut

sub cifs_setup {

    my ( $pkg, %opts ) = @_;

    # while adding args to cifs_setup make sure to add them to cifs_setup_old
    # or remove the args from the list before calling cifs_setup_old
    my %args = validate_with(
        params => \%opts,
        spec   => {
            local_admins => { type => SCALAR, optional => 1 },
            wins_option  => { type => SCALAR, optional => 1, default => 3 },
            wins_server_list => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/
            },
            vfiler      => { type => SCALAR, optional => 1 },
            'auth-type' => {
                type     => SCALAR,
                optional => 1,                             # used ZAPI style
                regex    => qr/^(ad|workgroup|passwd)$/,
                default  => 'workgroup'
            },
            'server-name' => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/\w{1,15}/
            },
            workgroup =>
                { type => SCALAR, optional => 1, default => 'workgroup' },
            domain => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/\w+/,
                depends  => [ 'domain_user', 'domain_pass' ]
            },
            domain_user =>
                { type => SCALAR, regex => qr/\w+/, optional => 1 },
            domain_pass => { type => SCALAR, optional => 1 },
            'ou-name' =>
                { type => SCALAR, optional => 1, default => 'CN=computers' },
            root_pass => { type => SCALAR, default => $pkg->root_password() },
            cifs_pass => { type => SCALAR, default => 'cifspass123' },
            'security-style' => {
                type     => SCALAR,
                optional => 1,                         # used ZAPI style
                default  => 'multiprotocol',
                regex    => qr/^ntfs|multiprotocol$/
            },
            create_admin => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^(yes|no)$/,
                default  => 'yes'
            },
            dns_enable => {
                type     => SCALAR,
                optional => 1,
                depends  => [ 'dns_server_list', 'dns_domainname' ]
            },
            dns_domainname  => { type => SCALAR, optional => 1 },
            dns_server_list => {
                type     => SCALAR,
                optional => 1,       # comma separated list of DNS server IP's
                regex => qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/
            },
            timed_enable => {
                type     => SCALAR,
                optional => 1,
                depends  => ['timed_server_list']
            },
            timed_server_list => { type => SCALAR, optional => 1 }
            ,    # comma separated list of time server name or IP's
            passwd_group =>
                { type => SCALAR, regex => qr/^(yes|no)$/, default => 'yes' },
            select_site_name =>
                { type => SCALAR, regex => qr/^(yes|no)$/, default => 'yes' },
            'site-name' => { type => SCALAR, optional => 1, default => '' },
        }
    );
    $Log->comment("Configuring CIFS on Filer: $pkg->hostname()");

    # Set the user authentication selction based on auth-type
    my $setup_passed = 1;
    my $auth_type    = $pkg->_compute_auth_type(%args);

    # check whether to execute the old or new style cifs routine
    $Log->comment("Remove administrative user");
    $pkg->_cmd_execute(
        vfiler     => $args{vfiler},
        cmd_string => "useradmin user delete administrator"
    );
    my $output_buffer = $pkg->_cmd_execute(
        vfiler     => $args{vfiler},
        cmd_string => "options cifs.new_setup.enable"
    );
    if (   ( $output_buffer =~ /No such option cifs.new_setup.enable/ )
        || ( $output_buffer =~ /cifs.new_setup.enable\s+off/ ) )
    {

        # execute old style of cifs setup
        $Log->comment("Using OLD style of cifs setup");
        my %old_args = %args;

        # delete the args that are not supported by cifs_setup_old
        # if needed these args need to be coded in cifs_setup_old and tested
        delete $old_args{'server-name'};
        delete $old_args{dns_enable};
        delete $old_args{dns_domainname};
        delete $old_args{dns_server_list};
        delete $old_args{timed_enable};
        delete $old_args{timed_server_list};
        delete $old_args{'security-style'};
        delete $old_args{passwd_group};
        delete $old_args{select_site_name};
        delete $old_args{'site-name'};
        delete $old_args{wins_option};
        delete $old_args{wins_server_list};
        delete $old_args{local_admins};

        $pkg->_cifs_setup_old(%old_args)
            or NACL::APISet::Exceptions::CommandFailedException->throw(
            "Failed to configure cifs");
        return $setup_passed;
    } ## end if ( ( $output_buffer ...
    $Log->comment("Using NEW style of cifs setup");

    # Checking the Filer security option field values
    undef($output_buffer);
    $output_buffer = $pkg->_cmd_execute(
        vfiler     => $args{vfiler},
        cmd_string => "options security.passwd.rules.history"
    );
    $output_buffer =~ /history\s+(\d+)/;
    my $out_store = "$1";

    # Temporarily Resetting security.passwd.rules.history to 0
    $pkg->_cmd_execute(
        vfiler     => $args{vfiler},
        cmd_string => "options security.passwd.rules.history 0"
    );

    # Terminate and license CIFS
    $pkg->_cmd_execute(
        vfiler     => $args{vfiler},
        cmd_string => "cifs terminate -t 0"
    );

    # Create a replytable.
    my $flag       = 0;
    my @replytable = (
        "Do you want to make the system visible via WINS\\? .*" =>
            $args{wins_option} =~ /[1-2]/ ? "yes" : "no",
        "\\(1\\)(?:.|\n)Keep the current WINS(?:.|\n)*Selection \\(1-3\\)\?"
            => $args{wins_option},
        "IPv4 address\\(es\\) of your WINS name server\\(s\\).*" =>
            $args{wins_server_list},
        "Would you like to specify additional WINS name servers" => "no",
        "Enter the password for the root user" => $args{root_pass},
        "Enter the password for root"          => $args{root_pass},
        "Retype the password:"                 => sub {
            ( $flag == 1 ) ? $args{cifs_pass} : $args{root_pass};
        },
        "(?s:The default name of this filer will be .*"
            . "Do you want to modify this name\\?)" => "no",
        "(?s:This filer will be a member of the Windows Workgroup .*"
            . "Do you want to modify this name\\?)" =>
            exists( $args{workgroup} ) ? "yes" : "no",
        "Workgroup name for the filer\\?.*:" => $args{workgroup},
        "Do you want to modify this list\\?" => "no",
        "Do you want to delete the existing filer account information\\?" =>
            "yes",
        "create the .*\\\\administrator account\\?" => $args{create_admin},
        "Please enter the Windows.*user.*:"         => $args{domain_user},
        "Enter the name of the Windows user"        => $args{domain_user},
        "Password for.*:"                           => $args{domain_pass},
        "Do you wish to continue and join.*in NT4 mode\\?" => "",
        "Active Directory container for filer account\\?" => $args{'ou-name'},
        "Do you want to continue and change the current filer account"
            . " information\\?" => 'yes',
        "Would you like to reconfigure this filer to be an NTFS-only filer\\?"
            => ( ( $args{'security-style'} eq 'ntfs' ) ? 'yes' : 'no' ),
        "Do you want to configure this filer as a NTFS-only filer" =>
            ( ( $args{'security-style'} eq 'ntfs' ) ? 'yes' : 'no' ),
        "Data ONTAP CIFS services support four styles of user"
            . " authentication(?:.|\n)*Selection \\(1-4\\)\?" => $auth_type,
        "\\(1\\) Multiprotocol filer[\r|\n]*\\(2\\) NTFS-only filer[\r|\n]*"
            . "Selection \\(1-2\\)\?" =>
            ( ( $args{'security-style'} eq 'multiprotocol' ) ? 1 : 2 ),
        "\\(1\\) NTFS-only filer[\r|\n]*\\(2\\) Multiprotocol filer[\r|\n]*"
            . "Selection \\(1-2\\)\?" =>
            ( ( $args{'security-style'} eq 'multiprotocol' ) ? 2 : 1 ),
        "\\(1\\) Ignore Active Directory Site information when choosing domain"
            . " controllers[\r|\n]*\\(2\\) Choose a site name from the list of"
            . " available sites[\r|\n]*Selection \\(1-2\\)\?" =>
            ( ( $args{select_site_name} eq 'no' ) ? 1 : 2 ),

        # site-name
        "Choose a Site name from the following list of Sites(?:.|\n)*Selection"
            . " \\(1-\\d+\\)\?" => '',    # default : Default-first-site-name
                                          # ou-name
        "The user that you specified has permission to create(?:.|\n)*\\(1\\)"
            . " CN=computers(?:.|\n)*Selection \\(1-\\d+\\)" =>
            $args{'ou-name'},
        "\\(3\\) Choose from a subset of containers by specifying a search"
            . " filter(?:.|\n)*Selection \\(1-3\\)" =>
            exists( $args{'ou-name'} ) ? 3 : 1,
        "Please enter some or all of a container specification to limit"
            . " the(?:.|\n)*Enter the search string" => $args{'ou-name'},
        "\\(1\\) $args{'ou-name'}(?:.|\n)*Selection \\(1-2\\)" => 1,
        "Would you like to change this name\\?" =>
            exists( $args{'server-name'} ) ? 'yes' : 'no',
        "Would you like to specify a user or group that can administer CIFS"
            => exists( $args{'local_admins'} ) ? 'yes' : 'no',
        "Enter the name of a user or group that will administer CIFS on the filer"
            => $args{'local_admins'},
        "Enter the CIFS server name for the filer" => $args{'server-name'},
        "Active Directory Domain name\\?"          => $args{domain},
        "What is the name of the Active Directory domain" => $args{domain},
        "Do you want to re-use this AD account\\?"   => "",    # default : yes
        "Do you want to re-use this machine account" => "",    # default : yes
        "Enter the new password for .*\\\\administrator" => sub {
            $flag = 1;
            $args{cifs_pass};
        },
        "Would you like to specify a user or group that can administer CIFS\\?"
            => 'no',
        "What is the name of the Workgroup\\?" => "",
        "Should CIFS create a default /etc/passwd file" =>
            $args{passwd_group},
        "Should CIFS create a default /etc/group file" => $args{passwd_group},
        "Should CIFS create default /etc/passwd and /etc/group files" =>
            $args{passwd_group},
        "Would you like to enable NIS group caching" => "",    # default : yes
        "Enter the hour\\(s\\) when NIS should update the group cache" =>
            "",    # default : 24hrs
        "Would you like to specify additional hours" => "",    # default : no
        "Do you want to configure the filer\\'s DNS resolver service" =>
            exists( $args{dns_enable} ) ? 'yes' : 'no',
        "What is the filer\\'s DNS domain name" => $args{dns_domainname},
        "What are the IPv4 address\\(es\\) of your authoritative DNS"
            . " name server\\(s\\)" => $args{dns_server_list},
        "What are the IPv4/IPv6 address\\(es\\) of your authoritative DNS"
            . " name server\\(s\\)" => $args{dns_server_list},
        "Would you like to specify additional DNS name servers" => "no",
        "Would you like to configure time services\\?" =>
            exists( $args{timed_enable} ) ? 'yes' : 'no',
        "Enter the time server host\\(s\\) and\\/or address\\(es\\)" =>
            $args{timed_server_list},
        "Would you like to specify additional time servers\\?" => 'no',
        "Would you like to reconfigure this filer to be a multiprotocol filer\\?"
            => (
            ( $args{'security-style'} eq 'multiprotocol' ) ? 'yes' : 'no' ),
        "Do you want to create the filer's machine account in" =>
            "",    #default : yes
    );

    # wait for "cifs setup" prompts in %replytable, and act on
    # them, until the prompt coming back is one of the built-in
    # say prompts (i.e. the filer's normal prompt, showing that
    # cifs setup is done)
    for ( my $tries = 3; $tries; $tries-- ) {

        $output_buffer = $pkg->filer_api_console()->execute_raw_command(
            command                  => "cifs setup",
            'connectrec-match_table' => \@replytable
        );
        if ( $output_buffer
            =~ m[CIFS local server is running|CIFS Setup can not be run while CIFS is active]i
            )
        {
            sleep 10;
            last;
        } ## end if ( $output_buffer =~...
    } ## end for ( my $tries = 3; $tries...

    $output_buffer
        = $pkg->filer_api_console()->execute_raw_command( command => "cifs" );
    if ( $output_buffer =~ /commands are available/i ) {
        $Log->comment(
            "CIFS successfully configured on Filer: " . $pkg->hostname() );
    } else {
        NACL::APISet::Exceptions::CommandFailedException->throw(
            "setup::cifs():cifs setup unsuccessful :" . $output_buffer );
    }

    # Reverting back the security.passwd.rules.history value to actual value
    $pkg->_cmd_execute(
        vfiler     => $args{vfiler},
        cmd_string => "options security.passwd.rules.history $out_store"
    );

    return $setup_passed;
} ## end sub cifs_setup

sub _cifs_setup_old() {
    my $pkg  = shift;
    my %args = validate_with(
        params => @_,
        spec   => {
            'auth-type' => {
                type     => SCALAR,
                optional => 1,                             # used ZAPI style
                regex    => qr/^(ad|workgroup|passwd)$/,
                default  => 'workgroup'
            },
            workgroup => { type => SCALAR, optional => 1 },
            domain    => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/\w+/,
                depends  => [ 'domain_user', 'domain_pass' ]
            },
            domain_user =>
                { type => SCALAR, regex => qr/\w+/, optional => 1 },
            domain_pass => { type => SCALAR, optional => 1 },
            'ou-name'   => { type => SCALAR, optional => 1 },
            root_pass => { type => SCALAR, default => $pkg->root_password() },
            cifs_pass => { type => SCALAR, default => 'cifspass123' },
            create_admin =>
                { type => SCALAR, optional => 1, default => 'yes' },
            vfiler => { type => SCALAR, optional => 1 }
        }
    );
    $Log->comment("Configuring CIFS on Filer: $pkg->hostname()");
    my $setup_passed = 1;
    my $cmd_prefix = ( $args{vfiler} ) ? "vfiler run $args{vfiler}" : "";

    $Log->comment("Remove administrative user");
    $pkg->filer_api_rsh()
        ->execute_raw_command(
        command => "$cmd_prefix useradmin user delete administrator" );

    # Checking the Filer security option field security.passwd.rules.history
    # values and storing it
    my $output_buffer
        = $pkg->filer_api_rsh()
        ->execute_raw_command(
        command => "$cmd_prefix options security.passwd.rules.history" );

    $output_buffer =~ /(\d+)/;
    my $out_store = "$1";

    # Disabling the NIS options for old setup to work
    $pkg->filer_api_rsh()
        ->execute_raw_command(
        command => "cmd_prefix options nis.enable off" );

    # Disabling the LDAP options for old setup to work
    $pkg->filer_api_rsh()
        ->execute_raw_command(
        command => "$cmd_prefix options ldap.enable off" );

    # Temporarily Resetting security.passwd.rules.history to 0
    $pkg->filer_api_rsh()
        ->execute_raw_command(
        command => "$cmd_prefix options security.passwd.rules.history 0" );

    # Set the user authentication selction based on auth-type
    my $auth_type = _compute_auth_type(%args);

    # Terminate and license CIFS
    $pkg->filer_api_rsh()
        ->execute_raw_command( command => "$cmd_prefix cifs terminate -t 0" );

    # Create a replytable.
    my @replytable = (
        "Do you want to make the system visible via WINS\\?"          => "no",
        "Do you wish to make the system visible via WINS\\?"          => "no",
        "Do you want to configure this filer as a NTFS-only filer\\?" => "no",
        "Enter the password for root" => defined( $args{root_pass} )
        ? &_password_set(%args)
        : $args{root_pass},
        "Set the password for .*\\\\administrator" =>
            defined( $args{cifs_pass} ) ? &_password_set(%args)
        : $args{cifs_pass},
        "Retype the password:" => ( $pkg->set_passwd() ) ? $args{root_pass}
        : $args{cifs_pass},
        "(?s:The default name of this filer will be .*"
            . "Do you want to modify this name\\?)" => "no",
        "(?s:This filer will be a member of the Windows Workgroup .*"
            . "Do you want to modify this name\\?)" =>
            exists( $args{workgroup} ) ? "yes" : "no",
        "(?s:This filer will be a member of the Workgroup .*"
            . "Do you want to modify this name\\?)" =>
            exists( $args{workgroup} ) ? "yes" : "no",
        "Enter the Workgroup for the filer.*:" => $args{workgroup},
        "Do you want to modify this list\\?"   => "no",
        "What type of authentication will this filer use\\?" => $auth_type,
        "Do you want to delete the existing filer account information\\?" =>
            "yes",
        "Create the .*\\\\administrator account\\?" => $args{create_admin},
        "Enter the Windows Domain for the filer.*:" => $args{domain},
        "Please enter the Windows.*user.*:"         => $args{domain_user},
        "Password for.*:"                           => $args{domain_pass},
        "Do you wish to continue and join.*in NT4 mode\\?" => "",
        "Active Directory container for filer account\\?" => $args{'ou-name'},
        "Would you like to reconfigure this filer to be a multiprotocol"
            . " filer\\?" => 'no',
    );

    # wait for "cifs setup" prompts in %replytable, and act on
    # them, until the prompt coming back is one of the built-in
    # say prompts (i.e. the filer's normal prompt, showing that
    # cifs setup is done)
    for ( my $tries = 3; $tries; $tries-- ) {
        $output_buffer = $pkg->filer_api_console()->execute_raw_command(
            command                  => "cifs setup",
            'connectrec-match_table' => \@replytable
        );
        if ( $output_buffer = /CIFS local server is running/ ) {
            sleep 10;
            last;
        }
    } ## end for ( my $tries = 3; $tries...
    $output_buffer = $pkg->filer_api_rsh()
        ->execute_raw_command( command => "$cmd_prefix cifs" );
    if ( $output_buffer =~ /The following commands are available/ ) {
        $Log->comment(
            "CIFS successfully configured on Filer: " . $pkg->hostname() );
    } else {
        NACL::APISet::Exceptions::CommandFailedException->throw(
            "setup::cifs():cifs setup unsuccessful : \"$output_buffer\"\n");
    }

    # Reverting back the security.passwd.rules.history value to actual value
    $Log->comment("the resetting of original values of security");
    $pkg->filer_api_rsh()
        ->execute_raw_command( command =>
            "$cmd_prefix options security.passwd.rules.history $out_store" );

    return $setup_passed;
} ## end sub _cifs_setup_old()

=head2 _compute_auth_type

A routine to compute authentication type (WORKGROUP/Domain).

This is only for 7-Mode.

The routine returns authentication type in integers
1   for Domain
2/3 for Workgroup (Depending upon cifs new setup)

=cut

sub _compute_auth_type {
    my $pkg  = shift;
    my %args = @_;
    my $flag = 0;
    my $auth_type;
    my $cmd_prefix = ( $args{vfiler} ) ? "vfiler run $args{vfiler}" : "";
    my $out_compute
        = $pkg->filer_api_console()
        ->execute_raw_command(
        command => "$cmd_prefix options cifs.new_setup.enable" );

    $flag = 1
        if ( ( $out_compute =~ /No such option cifs.new_setup.enable/ )
        || ( $out_compute =~ /cifs.new	_setup.enable\s+off/ ) );
    if ( $args{'auth-type'} eq 'passwd' ) {

        # Dont know of anything that depends on this auth style
        $auth_type = ($flag) ? 3 : 4;
    } elsif ( ( $args{'auth-type'} eq 'workgroup' )
        || ( !$args{'workgroup'} ) )
    {

        # Check if workgroup is specified - else default it to WORKGROUP
        #if (!$args{'workgroup'}) {
        $Log->comment("Setting workgroup to 'WORKGROUP'");

        # Get rid of defaulting to "WORKGROUP"
        $args{workgroup} = 'WORKGROUP';

        # }
        $auth_type = ($flag) ? 2 : 3;
    } elsif ( $args{'auth-type'} eq 'nt4' ) {

        # This is not supported as yet
        $Log->warn("This user auth style is not supported by this routine");
        $auth_type = ($flag) ? "" : 2;
    } else {

        # Check if domain name etc are also specified
        $Log->warn( "'ad' user auth style has been chosen, and "
                . "domain, domain_user, domain_pass also need to be specified"
            )
            if ( !$args{domain}
            || !$args{domain_user}
            || !$args{domain_pass} );
        $auth_type = 1;
    } ## end else [ if ( $args{'auth-type'...

    return ($auth_type);
} ## end sub _compute_auth_type

=head2 _cmd_execute

Helper routine to execute filer side cmds. This is only for 7-Mode.

The routine returns the cmd output.

The routine takes the following named arguments

=over

=item vfiler

(Optional) : If variable vfiler is defined, then cifs setup is done on the vfiler

=item cmd_string

(Required) : command with options

=back

=cut

sub _cmd_execute {
    my ( $pkg, %opts ) = @_;
    my %args = validate_with(
        params => \%opts,
        spec   => {
            vfiler     => { type => SCALAR | UNDEF, optional => 1 },
            cmd_string => { type => SCALAR }
        }
    );
    my $vfiler     = $args{vfiler};
    my $cmd_string = $args{cmd_string};
    $cmd_string = "vfiler run $vfiler $cmd_string" if ($vfiler);

AGAIN: {
        my (@cmd_outputs)
            = $pkg->filer_api_console()
            ->execute_raw_command( command => "$cmd_string" );

        if ( defined( $cmd_outputs[2] )
            && $cmd_outputs[2] =~ /vfiler cannot be run/ )
        {
            redo AGAIN;
        }
        return wantarray ? @cmd_outputs : $cmd_outputs[0];
    } ## end AGAIN:
} ## end sub _cmd_execute

=head2 _password_set

Helper routine to set/reset password. Returns root passwd if password
is already set else returns cifs administrative user password.

=cut

sub _password_set {
    my ( $pkg, %args ) = @_;
    if ( !$pkg->set_passwd() ) {
        $pkg->set_passwd(1);
        return ( $args{root_pass} );
    } else {
        $pkg->set_passwd(0);
        return ( $args{cifs_pass} );
    }
}    #end of sub _password_set

1;
