#$Id: //depot/prod/test/Rsierranevada/lib/FS/EA/SpaceAggrSnapshot.pm#2 $

#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

# @summary Space Specific Aggregate Snapshot Library
# @author Hardikkumar.Darji@netapp.com
# @status shared
# @pod here

=head1 NAME

FS::EA::SpaceAggrSnapshot

=head1 DESCRIPTION

This module provides the methods for creating, deleting, reserving and
checking status of the Aggregate level Snapshots
aggr snapshot support is available on August 15,2011 so need to update
the library to use component level calls (see BURT 526536)

Unittest script is at: lib/FS/EA/unittest/SpaceAggrSnapshot_unit_test.thpl
=cut

package FS::EA::SpaceAggrSnapshot;

use strict;
use warnings;

require Exporter;

use Tharn qw($Log);
use Params::Validate qw(validate SCALAR OBJECT);
use vars qw (@ISA @EXPORT_OK);
use Data::Dumper;
use NACL::APISet::Exceptions::UnexpectedOutputException qw (:try);
use NACL::APISet::Response::CLI::ParserUtils;
use NATE::BaseException;
@ISA       = qw(Exporter);

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(aggr_snapshot_create aggr_snapshot_delete
    aggr_snapshot_status aggr_snapshot_restore aggr_snap_reserve_set
    aggr_snap_reserve_get aggr_snapshot_autodelete_base_exists);
};



=head1 METHODS

=head2 aggr_snapshot_create(%opt)

    aggr_snapshot_create(
                    node           => $node,
                    "aggregate-name" => $aggr_name,
                    snapshot         => $snapshot);

This method will create a snapshot on a specific aggregate.

=over

=item Options

=over

=item C<<node=>$node>>

Required: NACL node object.

=item C<<aggregate-name=>$aggr_name>>

Required: This is the name of the Aggregate.

=item C<<snapshot=>$snapshot>>

Required: This is the name of the Snapshot.

=item C<<wait=>$wait>>

Optional: wait = 0 then do not wait for complete creation of snapshot
          wait = 1 then wait for complete creation of snapshot
          (default is wait = 0)

=item C<<timeout=>$timeout>>

Optional: this is the timeout while waiting for the snapshot to be fully created
          default time out value = 600 if wait=1

=cut

sub aggr_snapshot_create
{

    $Log->enter();
    my (%opts) =
      validate(@_,
               {node              => {type => OBJECT},
                "aggregate-name"  => {type => SCALAR},
                snapshot          => {type => SCALAR},
                wait         => {type => SCALAR,optional => 1},
                timeout      => {type => SCALAR,optional => 1,default => 600}});

    my ($node, $aggr_name, $snapshot, $wait, $timeout)=(
                                $opts{node},
                                $opts{"aggregate-name"},
                                $opts{snapshot},
                                $opts{wait},
                                $opts{timeout});

    my ($snapshots, $snap_obj);

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    # if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    # create aggregate level snapshot
    my $response_create = $apiset->snap_create(
        aggregate        => 1,
        "aggregate-name" => $aggr_name,
        snapshot         => $snapshot
    );

    $snapshots = aggr_snapshot_status(
                                node           =>$node,
                                "aggregate-name" =>$aggr_name);

    $snap_obj  = $snapshots->[0]->{'snapshot_status'};

    if(!defined($opts{wait}) || $wait == 0){
        if($snap_obj->[0]->{'name'} !~ /$snapshot/){
            NATE::BaseException->throw(
            text => "Snapshot creation Failed\n" );
        }
    } else {
           $Log->debug("Wait until the snapshot $snapshot is created");
        do {
            Tharn::sleep(15);
            $timeout -= 15;
            $snapshots = aggr_snapshot_status(
                            node           =>$node,
                            "aggregate-name" =>$aggr_name);
            $snap_obj = $snapshots->[0]->{'snapshot_status'};
        } while ($timeout > 0 && $snap_obj->[0]->{'name'} ne $snapshot);
        if ($timeout <= 0) {
            NATE::BaseException->throw(
            text =>"$snapshot not created completely, timeout occured");
        }
    }
    $Log->exit();
    return;
}


=head2 aggr_snapshot_delete(%opt)

    aggr_snapshot_delete(
                    node               => $node,
                    "aggregate-name"   => $aggr_name);

This method call will delete all the snapshots from the specified aggregate.

OR
    aggr_snapshot_delete(
                    node              => $node,
                    "aggregate-name"  => $aggr_name
                    snapshot          => $snapshot);

This method call will delete the specified snapshots from
the specified aggregate.
=over

=item Options

=over

=item C<<node=>$node>>

Required: NACL node object.

=item C<<"aggregate-name"=>$aggr_name>>

Required: The name of the Aggregate.

=item C<<snapshot=>$snapshot>>

Optional: The snapshot to be deleted.
          if this parameter is not provided then
          all the snapshots will be deleted
=cut
sub aggr_snapshot_delete
{
    $Log->enter();
    my (%opts) =
      validate(@_,
               {node              => {type => OBJECT},
                "aggregate-name"  => {type => SCALAR},
                snapshot          => {type => SCALAR,optional => 1}});

    my ($node, $aggr_name, $snapshot)=(
                            $opts{node},
                            $opts{"aggregate-name"},
                            $opts{snapshot});

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    # if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    # aggr snapshot is not defined for NACL, see BURT 510594
    # delete specified snapshots if snapshot is provided
    # otherwise remove all snapshots
    if ( defined($snapshot) ) {
        $apiset->snap_delete(
            aggregate  => 1,
            "vol-name" => $aggr_name,
            snapshot   => $snapshot
        );
    } else {
        $apiset->snap_delete(
            aggregate  =>1,
            "vol-name" =>$aggr_name,
            all        =>1,
            force      =>1
        );
    }
    $Log->exit();
    return;
}


=head2 aggr_snapshot_status(%opt)

    $snapshots = aggr_snapshot_status(
                                node          => $node,
                                "aggregate-name" => $aggr_name);

This method will run "snapshot status -A <aggr_name>" on given aggregate.
It will return an array reference containing snapshot status information

=over

=item Options

=over

=item C<<node=>$node>>

Required: NACL node object.

=item C<<"aggregate_name"=>$aggr_name>>

Required: The name of the Aggregate.

=cut
sub aggr_snapshot_status {
    $Log->enter();
    my (%opts) =
      validate(@_,
               {node            => {type => OBJECT},
                "aggregate-name"  => {type => SCALAR}});

    my ($node, $aggr_name)=(
                            $opts{node},
                            $opts{"aggregate-name"});
    my $result;

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    # if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    # calling snap status -A <aggregate-name>
    my $response_status = $apiset->snap_status(
        aggregate  => 1,
        "vol-name" => $aggr_name
    );

    my $raw_output   = $response_status->get_processed_output();

    # Now we can parse the output
    $result = _snap_status_parser(
        raw_output      => $raw_output,
        response_object => $response_status
    );
    $Log->debug("parsed_output:\n" .Dumper($result));
    $Log->exit();
    #return the array with snapshot related info.
    return $result;
}

=head2 _snap_status_parser(%opt)

    $snapshots = _snap_status_parser(raw_output      => $raw_output,
                                     response_object => $response_object);

This is a private method which parses the output of aggr_snapshot_status()
It will return an array reference containing snapshot status information
We have to have it for now because NACL doesn't provide us with a way to
get and parse the aggregate snapshot status.  See Burt 510594

=over

=item Options

=over

=item C<<raw_output = >$raw_output>>

Required: Raw output generated by aggr_snapshot_status() .

item C<<response_object=>$response_object>>

Required: response object created in aggr_snapshot_status().

=cut
sub _snap_status_parser
{
    my %args       = @_;
    my $raw_output = $args{"raw_output"};
    my $response   = $args{"response_object"};

    my $first_line = "Aggregate";
    my $can_throw  = 0;
    my @lines      = get_lines(
        'raw_output' => $raw_output,
        'first_line' => $first_line,
        'can_throw'  => $can_throw
    );
    my @return_array;

    while ( my $line = shift(@lines) ) {
        my $info_hash = {};
        if ( $line =~ /^Aggregate (\w+)(?:\s+)?(?:\(.*\))?$/ ) {
            $info_hash->{'volume'} = $1;
            $line = shift(@lines);
            if ( $line =~ /No snapshots exist./ ) {
                $info_hash->{'snapshot_status'} = [];
                push( @return_array, $info_hash );
                next;
            }
            my @keys;
            my $vol_log_flag = 0;
            if ( $line =~ /Aggregate.*Logical/ ) {
                $vol_log_flag = 1;
                @keys         = (
                    'volume_snapid', 'logical_snapid',
                    'status',        'date',
                    'ownblks',       'release',
                    'fsrev',         'name'
                );
                shift(@lines)
                    ; #shifting the line which has snapid status and such column names
            } else {
                @keys = (
                    'snapid',  'status', 'date', 'ownblks',
                    'release', 'fsrev',  'name'
                );
            }
            shift(@lines);    #skipping the dashes
            $line = shift(@lines);
            while ( $line !~ /Aggregate/ ) {
                my $table_hash = {};
                my @values;
                my $date_format = '\w{3}\s+\d{1,2}\s+\d{2}:\d{2}';
                if ($vol_log_flag) {
                    $line
                        =~ /(\d+)(?:\s+)?(\d+)(?:\s+)?(\w+)?(?:\s+)?($date_format)(?:\s+)?(\d+)(?:\s+)?(\d+\.\d+)(?:\s+)(\d+)(?:\s+)(.*)$/;
                    push( @values, $1, $2, $3, $4, $5, $6, $7, $8 );
                } else {
                    $line
                        =~ /(\d+)(?:\s+)?(\w+)?(?:\s+)?($date_format)(?:\s+)?(\d+)(?:\s+)?(\d+\.\d+)(?:\s+)(\d+)(?:\s+)(.*)$/;
                    push( @values, $1, $2, $3, $4, $5, $6, $7 );
                }
                NACL::APISet::Exceptions::UnexpectedOutputException->throw(
                    text            => "Unexpected Output while parsing",
                    response_object => $response
                ) unless ( scalar(@keys) == scalar(@values) );

                for ( my $i = 0; $i < scalar(@keys); $i++ ) {
                    $table_hash->{ $keys[$i] } = $values[$i];
                }
                push( @{ $info_hash->{'snapshot_status'} }, $table_hash );
                last unless ( scalar(@lines) );
                $line = shift(@lines);
            }
            unshift( @lines, $line ) if ( scalar(@lines) );
        } else {
            NACL::APISet::Exceptions::UnexpectedOutputException->throw(
                text            => "Unexpected Output while parsing",
                response_object => $response
            );
        }
        push( @return_array, $info_hash );
    }

    return \@return_array;
}

=head2 aggr_snap_reserve_set(%opt)

    aggr_snap_reserve_set(
                node             => $node,
                "aggregate-name" => $aggr_name,
                percent          => $percent);

This method is used to set ASR to user defined percentage

=over

=item Options

=over


=item C<<node=>$node>>

Required: NACL node object.

=item C<<"aggregate-name"=>$aggr_name>>

Required: The name of the Aggregate.

item C<<percent=>$percent>>

Required: user defined ASR percentage

=cut
sub aggr_snap_reserve_set
{
    $Log->enter();
    my (%opts) =
      validate(@_,
               {node              => {type => OBJECT},
                "aggregate-name"  => {type => SCALAR},
                percent           => {type => SCALAR}});

    my ($node, $aggr_name, $percent)=(
                            $opts{node},
                            $opts{"aggregate-name"},
                            $opts{percent});

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    # if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    # setting Aggregate Snapshot Reserve (ASR) to user-defined %
    $apiset->aggr_options(
                aggregate       => $aggr_name,
                "option-name"   => "percent_snapshot_space",
                "option-value"  => "$percent");
    $Log->exit();
    return;

}
=head2 aggr_snap_reserve_get(%opt)

    $asr_hash = aggr_snap_reserve_get(
                            node           => $node,
                            "aggregate-name" => $aggr_name);

This method is used to get the current ASR value.
It will return a hash with $percent, the ASR value in percentage, and
$size, the size in K-Bytes ( which is ONTAP default).

=over

=item Options

=over

=item C<<node=>$node>>

Required: NACL node object.

=item C<<"aggregate_name"=>$aggr_name>>

Required: The name of the Aggregate.

=cut
sub aggr_snap_reserve_get {

    $Log->enter();
    my (%opts) =
      validate(@_,
               {node             => {type => OBJECT},
                "aggregate-name" => {type => SCALAR}});

    my ($node, $aggr_name)=(
                        $opts{node},
                        $opts{"aggregate-name"});

    my (%asr_hash, $asr_output, $parsed_result);

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    # if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    # Get Aggregate Snapshot Reserve (ASR)
    $asr_output = $apiset->snap_reserve(aggregate => 1,
                                       'vol-name'=>$aggr_name);

    $parsed_result = $asr_output->get_parsed_output();
    $Log->debug(" parsed_output:\n" . Dumper($parsed_result));

    if($parsed_result->[0]->{'reserve'} =~ /(\d+)%/){
        $asr_hash{'percent'} = $1;
        $Log->debug("ASR in %(percentage) = $1");
    } else {
        NATE::BaseException->throw(
            text =>"Unexpected Aggregate Snap Reserve percentage value");
    }

    if($parsed_result->[0]->{'size'} =~ /(\d+)/){
        $asr_hash{'size'} = $1;
        $Log->debug("ASR in k-bytes = $1 ");
    } else {
        NATE::BaseException->throw(
            text =>"Unexpected Aggregate Snap Reserve value");
    }

    $Log->exit();
    return \%asr_hash;

}
=head2 aggr_snapshot_restore(%opt)

    aggr_snapshot_restore(
                    node             => $node,
                    "aggregate-name" => $aggr_name,
                    snapshot         => $snapshot,
                    type             => $type);

This method will restore the Active file System to a previous snaspshot.
(This is a 7-Mode only feature)

=over

=item Options

=over

=item C<<node=>$node>>

Required: NACL node object.

=item C<<"aggregate-name"=>$aggr_name>>

Required: The name of the Aggregate.

=item C<<snapshot=>$snapshot>>

Required: The name of the Snapshot.

=item C<<type=>$type>>

Required: Type can be 'vol' or 'file'
=cut

sub aggr_snapshot_restore
{
    $Log->enter();
    my (%opts) =
      validate(@_,
               {node              => {type => OBJECT},
                "aggregate-name"  => {type => SCALAR},
                snapshot          => {type => SCALAR},
                type              => {type => SCALAR}});

    my ($node, $aggr_name, $snapshot, $type)=(
                        $opts{node},
                        $opts{"aggregate-name"},
                        $opts{snapshot},
                        $opts{type});

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    #if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    # create aggregate level snapshot
    my $response_create = $apiset->snap_restore(
        aggregate        => 1,
        force            => 1,
        type             => $type,
        "snapshot-name"  => $snapshot,
        "aggregate-name" => $aggr_name
    );
    my $raw_result = $response_create->get_raw_output();
    # Currently NACL do not have parser for the command "snap restore -A"
    # Burt 510594 is logged for Aggr snapshot operations which should provide
    # proper parser utility for snap restore -A
    $Log->debug(" raw_output:" . Dumper($raw_result));
    if ($raw_result =~ /Aggregate\s+(.*):\s+revert\s+successful/) {
        return;
    } else {
        NATE::BaseException->throw(
            text =>"Snap restore Operation not sucessful");
    }
}

=head2 aggr_snapshot_autodelete_base_exists(%opt)

This method is to check if "autodelete_base" snapshot exists for a given
aggregate. Return 1 if found or 0 for not found.

Example:
    if ( aggr_snapshot_autodelete_base_exists(
            node       => $Node,
            aggregate  => $FILER_AGGR,
         ) ) {
        return $TCD::FAILED;
    };
=over

=item Options

=over

=item C<<node=>$node>>

Required: The Node command interface.

=item C<<aggregate=>$aggregate>>

Required: The name of the Aggregate.

=item C<<timeout=>$timeout>>

Optional: Timeout (in seconds) to check aggregate snapshot "autodelete_base".
Default is 300 (5 minutes).

=cut
sub aggr_snapshot_autodelete_base_exists
{
    $Log->enter();
    my $rc = 1;

    my (%opts) = validate(@_, {
         node       => {type => OBJECT, optional => 0 },
         aggregate  => {type => SCALAR, optional => 0 },
         timeout    => {type => SCALAR, optional => 1, default => 300 },
    });

    my $found = 0;
    my $check_interval = 10; # seconds
    my $timer = 0;
    while (1) {
        # Set return code to 0 if timeout value is hit
        if ( $timer >= $opts{timeout} ) {
            $Log->comment("Hit timeout after $opts{timeout} minutes");
            $rc = 0;
            last;
        }
        # Get aggr snapshots
        my $snapshots = aggr_snapshot_status(
            node                => $opts{node},
            "aggregate-name"    => $opts{aggregate},
        );
        # Look for autodelete_base in available aggr snapshots
        my $snapshots = $snapshots->[0]->{'snapshot_status'};
        foreach my $snapshot (@$snapshots) {
            $Log->comment("Snapshot: " . $snapshot->{'name'});
            if ($snapshot->{'name'} eq "autodelete_base") {
                $found = 1;
                last;
            };
        }

        # Exit while loop if found "autodelete_base"
        last if ( $found );

        # If not found, re-check after every 10 seconds
        $timer += $check_interval;
        sleep ($check_interval);
    }

    $Log->exit();
    return $rc;
}

1;
