#$Id: //depot/prod/test/Rsierranevada/lib/FS/EA/SpaceUsage.pm#1 $

#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

# @summary Space Specific Aggregate / Volume Populate Library
# @author Hardikkumar.Darji@netapp.com
# @status shared
# @pod here

package FS::EA::SpaceUsage;

use strict;
use warnings;

require Exporter;

use Tharn qw($Log);
use Params::Validate qw(validate SCALAR OBJECT);
use vars qw (@ISA @EXPORT_OK);
use Data::Dumper;
use WaflSpaceLib  qw( convert_size );
use NACL::STask::Volume;
use NACL::C::Mkfile;
use NACL::CS::Aggregate;
use NACL::CS::Volume;
use NACL::APISet::Exceptions::ResponseException qw(:try);
# BURT 510594 is filed to support all Aggregate snapshot related operations in
# NACL
use FS::EA::SpaceAggrSnapshot qw(aggr_snapshot_create aggr_snapshot_status );
@ISA = qw(Exporter);

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(aggr_snap_usage_populate volume_populate vol_size_get
    aggr_size_get aggr_snap_usage_get);
};

=cut
=head1 NAME

FS::EA::SpacePopulate

=head1 DESCRIPTION

This module provides the methods for populating aggregate and volumes and
checking the volume's or aggregate's space utilization

Unittest script is at: lib/FS/EA/unittest/SpaceUsage_unit_test.thpl

=head1 METHODS

=head2 aggr_snap_usage_populate(%opt)

    aggr_snap_usage_populate(
                    node                => $node,
                    "aggregate-name"    => $aggr_name,
                    "aggr-total-inkb"   => $aggr_total_inkb,
                    vol                 => $vol,
                    percent             => $percent,
                    "num-of-snaps"      => $num_of_snaps,
                    "num-of-files"      => $num_of_files,
                    "populate-type"     => $populate_type,
                    timeout             => $timeout);

This method will populate the aggregate's Snapshot usage by user defined
percentage independent of the amount of space currently in use. For example,
if $percent is 50, and snapshots already consume 25% of the aggregates space,
the resulting usage will be 75%.

we are not checking the provided volume's size, so be careful to have enough
size of the volume and volume should be empty.


The steps followed in the Method are:
1) calculate percentage size of aggregate
2) calculate data size and divide the data into files
3) create files and snapshots
4) remove the provided volume to trap the blocks only in aggr level snapshots
5) check that volume snapshot usage got increased

=over

=item Options

=over

=item C<<node => $node>>

Required: NACL node object for the node

=item C<<"aggregate-name"=>$aggr_name>>

Required: This is the name of the Aggregate.

=item C<<"aggr-total-inkb"=>$aggr_total_inkb>>

Required: This is total size of aggregate in K-bytes.

=item C<<vol=>$vol>>

Required: Volume object

=item C<<percent=>$percent>>

Required: Aggregate's Snapshot usage will be $percent of total size
 of aggregate, it should be an integer value.

=item C<<"num-of-snaps"=>$num_of_snaps>>

Required: Number of Snapshots to be created

=item C<<"num-of-files"=>$num_of_files>>

Required: total number of files to be created per snapshot

=item C<<"populate-type"=>$populate_type>>

Required: Method to populate data could be either "mkfile" or "unsparse",
 populate-type defualts to "mkfile" 

=item C<<timeout=>$timeout>>

Optional: Amount of time to wait for data population as well
          as snapshot usage

=cut

sub aggr_snap_usage_populate
{
    $Log->enter();

    my (%opts) =
      validate(@_,
               {node               => {type => OBJECT,optional => 0},
                "aggregate-name"   => {type => SCALAR,optional => 0},
                "aggr-total-inkb"  => {type => SCALAR,optional => 0},
                vol                => {type => OBJECT,optional => 0},
                percent            => {type => SCALAR,optional => 0},
                "num-of-snaps"     => {type => SCALAR,optional => 0},
                "num-of-files"     => {type => SCALAR,optional => 1,default => 2},
                "populate-type"    => {type => SCALAR,optional => 1,default => "mkfile"},
                timeout            => {type => SCALAR,optional => 1,default => 600}});

    my ($node, $aggr_name, $aggr_total_inkb, $vol, $percent, $num_of_snaps,
        $num_of_files, $populate_type, $timeout)=(
                                $opts{node},
                                $opts{"aggregate-name"},
                                $opts{"aggr-total-inkb"},
                                $opts{vol},
                                $opts{percent},
                                $opts{"num-of-snaps"},
                                $opts{"num-of-files"},
                                $opts{"populate-type"},
                                $opts{timeout});

    my ($percent_aggr_size);
    my ($file_size,$data_size,$snap_ind,$snap_name);
    my ($file_ind,$start_ind,$stop_ind,$vol_used_size,$used_size);

    # find the total size of aggr to populate based on percent requested
    $percent_aggr_size = int($aggr_total_inkb *($percent / 100));

    $Log->debug(" $percent% size of aggr is "
                  . $percent_aggr_size."kb");


    # Determine amount of data to create before taking each snapshot
    $data_size = int($percent_aggr_size / $num_of_snaps);
    $Log->debug(" data size of each snapshot in aggr is "
                  . $data_size."kb");

    # Split data over the desired number of files
    $file_size = int($data_size / $num_of_files);

    $Log->debug(" File size = ". $file_size."kb");

    $snap_ind = 1;
    for($file_ind = 1 ; $file_ind <= ($num_of_files*$num_of_snaps);
                $file_ind = $file_ind + $num_of_files) {
        $start_ind = $file_ind;
        $stop_ind = ($file_ind + $num_of_files)-1;

        $Log->comment(" populating volume by creating files");
        volume_populate(
            node            => $node,
            "vol-name"      => $vol->volume(),
            "num-of-files"  => $num_of_files,
            "file-size"     => $file_size,
            "populate-type" => $populate_type,
            "start-ind"     => $start_ind,
            "stop-ind"      => $stop_ind,
            timeout         => $timeout
        );

        $snap_name = "Aggr_Snap_".$snap_ind;
        $Log->comment(" Create snapshot $snap_name on $aggr_name");

        aggr_snapshot_create(
                    node             => $node,
                    "aggregate-name" => $aggr_name,
                    snapshot         => $snap_name,
                    wait             => 1);
        $snap_ind++;
    }

    $Log->debug(" Displaying snapshots");
    aggr_snapshot_status(
                    node           => $node,
                    "aggregate-name" => $aggr_name);

    # Determine space used on volume, before deleting it.
    $vol_used_size = vol_size_get(
                            vol         => $vol,
                            "size-type" => "used",
                        );

    $Log->comment(" remove one volume.");
    NACL::STask::Volume->purge(
        command_interface => $node,
        volume            => $vol->volume(),
        vserver           => $vol->vserver()
    );

    # After the volume has been deleted, the filer takes time
    # to recognize the increase in space used by the snapshot

    $Log->comment("Snapshot usage should grow to at least $vol_used_size");
    # Increased snapshot usage for aggregate can be 17 to 20MB
    # greater than $vol_used_size
    $used_size = aggr_snap_usage_get(
                node          => $node,
                "aggregate-name" => $aggr_name,
                "usage-type"    => "used");

    my $snap_usage_counter = 0;
    while (($used_size <= $vol_used_size) && ($snap_usage_counter < $timeout)) {
        $used_size = aggr_snap_usage_get(
                node          => $node,
                "aggregate-name" => $aggr_name,
                "usage-type"    => "used");

        aggr_snapshot_status(
                        node          => $node,
                        "aggregate-name" => $aggr_name);
        $snap_usage_counter++;
    }

    $Log->exit();
    if ($snap_usage_counter >= $timeout) {
        NATE::BaseException->throw(
            "timeout occured before Aggregate Snapshot Usage ".
            "becomes equal or greater than $vol_used_size");
    }
    return;
}


=head2 volume_populate(%opt)

        volume_populate(
            node            => $node,
            "vol-name"      => $vol->volume(),
            "num-of-files"  => $num_of_files,
            "file-size"     => $file_size,
            "populate-type" => $populate_type,
            "start-ind"     => $start_ind,
            "stop-ind"      => $stop_ind,
            timeout         => $timeout
        );

This method will populate the volume by creating files using either
"mkfile" or "unsparse" tools.

=over

=item Options

=over

=item C<<node=>$node>>

Required: NACL Node object for the node on which test need to be run.

=item C<<"vol-name"=>$vol_name>>

Required: The name of the volume to be populated

=item C<<"num-of-files"=>$num_of_files>>

Required: total number of files to be created

=item C<<"file-size"=>$file_size>>

Required: Size of each file in Kilobytes to be created

=item C<<"populate-type"=>$populate_type>>

Required: Method to populate data could be either "mkfile" or "unsparse"
 It will use "mkfile" as default mechanism 

=item C<<"start-ind"=>$start_ind>>

Required: start index of files (first file in case of "unsparse")

=item C<<"stop-ind"=>$stop_ind>>

Required: stop index of files (last file in case of "unsparse")

=item C<<timeout=>$timeout>>

Optional: Amount of time to wait for data population
=cut
sub volume_populate
{
    $Log->enter();
    my (%opts) =
      validate(@_,
               {node          => {type => OBJECT,optional => 0},
                "vol-name"      => {type => SCALAR,optional => 0},
                "num-of-files"  => {type => SCALAR,optional => 0},
                "file-size"     => {type => SCALAR,optional => 0},
                "populate-type" => {type => SCALAR,optional => 1,default => "mkfile"},
                "start-ind"     => {type => SCALAR,optional => 1},
                "stop-ind"      => {type => SCALAR,optional => 1},
                timeout         => {type => SCALAR,optional => 1,default => 600}});

    my ($node, $vol, $num_of_files, $file_size, $populate_type,
        $start_ind, $stop_ind, $timeout) =(
                                $opts{node},
                                $opts{"vol-name"},
                                $opts{"num-of-files"},
                                $opts{"file-size"},
                                $opts{"populate-type"},
                                $opts{"start-ind"},
                                $opts{"stop-ind"},
                                $opts{timeout});

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    # if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();
    my ($Filename,$cmd);
    $Filename= "/vol/$vol";
    $file_size = $file_size . "k";

    # based on populate_type, use appropriate call to mkfile or unsparse
    if($populate_type eq "mkfile")
    {
        $Log->debug(" Using mkfile to create $num_of_files files");
        NACL::C::Mkfile->mkfile(
            command_interface   => $node,
            size                => $file_size,
            filename            => $Filename . "/file_" . $start_ind,
            filecount           => $num_of_files,
            'method-timeout'    => $timeout
        );
    }
    elsif ($populate_type eq "unsparse")
    {
        $Log->debug(" Using unsparse to create $num_of_files files");
        if($num_of_files > 1){
            $cmd = "priv set -q test; unsparse -v -l $file_size -d " .
            "$Filename $start_ind ... $stop_ind";
        } else {
            $cmd = "priv set -q test; unsparse -v -l $file_size -d " .
            "$Filename $start_ind";
        }
        my $Mode = $node->mode();
        if($Mode eq "CMode"){
            # CMode requires the extra escaped quotes.
            $cmd = "\"$cmd\"";
            $node->run( command => $cmd );
        } elsif ($Mode eq "7Mode") {
            $apiset->execute_raw_command(command => $cmd );
        } else {
            NATE::BaseException->throw("Unexpected value \"$Mode\" " .
            "for Mode");
        }
    } else {
        NATE::BaseException->throw("Invalid value \"$populate_type\" for populate_type");
    }

    $Log->exit();
    return;

}


=head2 vol_size_get(%opt)

        $vol_used_size = vol_size_get(
                            vol         => $vol,
                            "size-type" => "used",
                        );

This method will find the size of volume from
based on "size-type". It returns size in K-bytes.

=over

=item Options

=over

=item C<<"size-type"=>$size_type >>

Required: size_type can be "used", "size" or "total"

=item C<<vol=>$vol>>

Required: volume object

=cut
sub vol_size_get {
    $Log->enter();
    my (%opts) =
      validate(@_,
               {"size-type"  => {type => SCALAR,optional => 0},
                "vol"      => {type => OBJECT,optional => 0}});

    my ($size_type, $vol) =(
                        $opts{"size-type"},
                        $opts{"vol"});

    my ($size, $vol_size, $vol_size_inkb);
    # finding value of "size" attribute from volume Component State
    $size = $vol->get_one_state_attribute($size_type);

    # converting value of "size" attribute to Kilobytes
    $Log->debug("Vol Size before conversion = " . $size);
    $vol_size = convert_size($size);
    $vol_size_inkb = int($vol_size->{'Kb'});

    $Log->debug("  vol $size_type size after conversion : " .
                  "$vol_size_inkb kb ");

    $Log->exit();
    return $vol_size_inkb;

}

=head2 aggr_size_get(%opt)

         $aggr_used_size = aggr_size_get(
                            aggr         => $vol,
                            "size-type" => "usedsize",
                        );
This method will find the size of aggregate
based on size_type. It returns size in K-bytes.

=over

=item Options

=over

=item C<<aggr=>$aggr>>

Required: aggregate object

=item C<<""size-type"=>$size_type>>

Required: size-type can be  "usedsize", "size" or "availsize"

=cut
sub aggr_size_get
{
    $Log->enter();
    my (%opts) =
      validate(@_,
               {"size-type"  => {type => SCALAR,optional => 0},
                "aggr"      =>  {type => OBJECT,optional => 0}});

    my ($size_type, $aggr) =(
                        $opts{"size-type"},
                        $opts{"aggr"});

    my ($size, $aggr_size, $aggr_size_inkb);

    # finding value of "size" attribute based on "size-type"
    # size-type can be  "usedsize", "size" or "availsize"
    $size = $aggr->get_one_state_attribute($size_type);
    $Log->comment(" Aggr Size before conversion = $size");

    # converting aggregate's size into Kilo-bytes
    $aggr_size=convert_size($size);
    $aggr_size_inkb = int($aggr_size->{'Kb'});

    $Log->comment(" Aggr Size ($size_type) After conversion = "
                  . $aggr_size_inkb ."kb");

    $Log->exit();
    return $aggr_size_inkb;

}

=head2 aggr_snap_usage_get(%opt)

    $usage_inkb = aggr_snap_usage_get(
                node          => $node,
                "aggregate-name" => $aggr_name,
                "usage-type"    => $usage_type);

This method will find out the space used by aggregate level snapshots
from df output based on usage-type. It returns size in K-bytes.

=over

=item Options

=over

=item C<<node=>$node>>

Required: Node object for the node on which volume resides.

=item C<<"aggregate-name"=>$aggr_name>>

Required: The name of the aggregate

=item C<<"usage-type"=>$usage_type >>

Required: usage-type can be snap "used"

=cut
sub aggr_snap_usage_get
{

    $Log->enter();
    my (%opts) =
      validate(@_,
               {node             => {type => OBJECT,optional => 0},
                "aggregate-name" => {type => SCALAR,optional => 0},
                "usage-type"     => {type => SCALAR,optional => 0}});

    my ($node, $aggr_name, $usage_type) =(
                                $opts{node},
                                $opts{"aggregate-name"},
                                $opts{"usage-type"});
    my ($usage, $usage_temp, $aggr_snap_usage_inkb);

    # Create / get cached APIset object which will get
    # 7Mode Api set for 7-Mode filer or node scope API set
    #if the filer is C-Mode
    my $apiset = $node->get_7m_or_nodescope_apiset();

    #affected due to df-A change(1797089) does not show aggregate spill
    #alternative is to use counter manager for aggregate object
    #snap usage is wv_fsinfo_blks_used - wv_fsinfo_blks_used_by_plane0

    my $aggr_fsinfo_blks_used = NACL::CS::Statistics->fetch(
        command_interface => $opts{node},
        filter            => {
            object   => 'aggregate',
            instance => $opts{'aggregate-name'},
            counter => "wv_fsinfo_blks_used",
            raw => "true",
        }
    );
    $Log->comment("aggr-blks-used=$aggr_fsinfo_blks_used->{value}");
    my $aggr_fsinfo_blks_used_plane_0 = NACL::CS::Statistics->fetch(
        command_interface => $opts{node},
        filter            => {
           object   => 'aggregate',
           instance => $opts{'aggregate-name'},
           counter => "wv_fsinfo_blks_used_by_plane0",
           raw => "true",
        }
    );
    $Log->comment("aggr-blks-used-plane0=$aggr_fsinfo_blks_used_plane_0->{value}");

    my $snap_usage = ($aggr_fsinfo_blks_used->{value} - $aggr_fsinfo_blks_used_plane_0->{value}) * 4096;
    $aggr_snap_usage_inkb = NACL::C::UnitNormalization::to_kilobytes($snap_usage);
    $aggr_snap_usage_inkb =~ s/KB//g;

    $Log->comment(" Aggr Snap usage ($usage_type): $aggr_snap_usage_inkb kb");

    $Log->exit();
    return int($aggr_snap_usage_inkb);
}

1;
