#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/FijiError.pm#1 $

#this library is intended for Cmode Raid error injection

=head1 NAME

  FS::FijiError- Library functions for simulating different kind of RAID errors

=head1 DESCRIPTION

  The RAID error library consists of functions to perform following tasks:

     1) raid_fijierror_clear() : Clear any Fiji error points
     2) raid_fijierror_inject() : Inject the requested RAID error
     

=head1 SYNOPSIS

  use FS::FijiError;

  my $resp = raid_fijierror_inject(api_obj => $APISet_obj, vol => $vol, 
						level=>0);
  my $resp = raid_fijierror_inject(api_obj => $APISet_obj, vol => $vol, 
                                 		path=>"/vol/$vol/f1", 
						snapid=>1); 

=head1 STATUS
 @status Development
 
=head1 AUTHOR
 @author Hardikkumar.Darji@netapp.com
 
=head1 DISCLAIMER

 @disclaimer This library does NOT support duplicate vol names in the same aggr. 
             In short no vserver checking is available.
=cut


# Compiler directives.
use strict;
use warnings;

# =============================================================
# Module Import
# =============================================================

require Exporter;

package FS::FijiError;
use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF OBJECT BOOLEAN UNDEF);
use NACL::APISet;
use vars qw (@ISA @EXPORT_OK);
use Data::Dumper;
use Tharn qw($Log);
use POSIX;
use Carp qw(croak);

# =============================================================
# Components Used
# =============================================================

use NACL::C::Node;

@ISA       = qw(Exporter);
@EXPORT_OK = qw(raid_fijierror_clear raid_fijierror_inject);

# =============================================================
# Declare Global Variables in BEGIN
# =============================================================

BEGIN {
    use vars qw(
        $Api_Set_Obj
        );
}



# =============================================================
=head1 METHODS
=head2 raid_fijierror_clear( api_obj=> $APISet_obj)

@summary
     Library function for clearing any Fiji error points

@description
     raid_fijierror_clear - Clear any Fiji error points

@arg $APISet_obj Required:
      api_obj => APISet object.

@return value
      Returns nothing
                          
      
=cut
# =============================================================
sub raid_fijierror_clear(%)
{
    my %opts = validate(
	@_,
	{
		api_obj => { type => OBJECT },
	}
    );

    my $Api_Set_Obj = $opts{api_obj};

    ## Clear any fiji error injection points
    my $fijicmd = "fiji clear all";

    my $resp = $Api_Set_Obj->system_node_run(node => "local", command =>$fijicmd);
    my $output = Dumper($resp->get_raw_output());
    if (!($output =~ /filter cleared/i)) {
	croak("fiji clear: $output");
    }
    else
    {
        $Log->comment("All fiji points cleared");
    }

    return 0; 
}


# ===============================================================
=head2 raid_fijierror_inject(api_obj=> $APISet_obj, vol => $vol [, %opts])

@summary

  Simulate Fiji error point as specified

@description
     raid_fijierror_inject - Inject the requested RAID error

 @arg $APISet_obj Required:
      api_obj => APISet object.

 @arg $vol Required: 
      vol => name of aggregate, traditional volume or flexible volume

 @arg %opts Optional: options can be set:

      fileid => fileid of the file to cause corruption on
      fileop => fileid operator "!=", "<", "<=", "==", ">=" or ">"
      path   => file path if corruption is desired on a given file
                This is ignored if a fileid is specified. 
      snapid => snapid of snapshot on which corruption is requested. 
                Default is 0 (Active FS)
      level  => block level for which error has to be injected, default is 0
      fbn    => Specific FBN to cause corruption on
      vbn    => Specific VBN to cause corruption on
      count  => Number of errors to inject, default is 999
      is_priv=> Inject private/public inode space.Applicable to dual inode concept.
      unverified => run fiji with unverified flag    
 @return value
      Returns $errcode = 0 if command successful
                       < 0 if command fail
                       
              $result = 0            if command successful
                      = error string if command fail
                          
 @example my ($errcode,$result) = raid_fijierror_inject(api_obj => $APISet_obj, vol => $vol, 
						level=>0);
          my ($errcode,$result) = raid_fijierror_inject(api_obj => $APISet_obj, vol => $vol, 
                                 		path=>"/vol/$vol/f1", 
						snapid=>1); 

=cut
# =============================================================
sub raid_fijierror_inject (%) 
{
    my %opts = validate( 
	@_,
	{
		api_obj    => { type => OBJECT},
                vol     => { type => SCALAR},
		fileid	=> { type => SCALAR, optional => 1 },
		fileop	=> { type => SCALAR, optional => 1, default => "==" },
		path	=> { type => SCALAR, optional => 1 },
		snapid	=> { type => SCALAR, optional => 1 },
		level	=> { type => SCALAR, optional => 1 },
		fbn	=> { type => SCALAR, optional => 1 },
		vbn	=> { type => SCALAR, optional => 1 },
		count	=> { type => SCALAR, optional => 1, default => 999 },
		is_priv => { type => SCALAR, optional => 1, default =>0 },
                unverified => { type => SCALAR, optional => 1, default =>0 },
		
	}
    );

    my $Api_Set_Obj = $opts{api_obj};
    my $vol = $opts{vol};

    my ($errcode,$result,$fileid, $path, $snapid, $level, $fbn, $vbn, $count, $is_priv);

    my $fileidstr = "";
    my $snapidstr = "";
    my $levelstr = "";
    my $fbnstr = "";
    my $vbnstr = "";
    my $countstr = "";
    my $privstr = "";
    my $unverified = "";
    $is_priv=$opts{is_priv};
    $privstr = "(== \@is_priv $is_priv)";
    
    if (defined($opts{fileid})) {
        $fileid = $opts{fileid};
        $fileidstr = "($opts{fileop} \@fileid $fileid)";
    } 

    if (defined($opts{path})) {
        $path = $opts{path};
    }

    # Use path only if fileid is not defined
    if (!defined($opts{fileid}) && defined($opts{path})) {
	my %showfh = &_filer_showfh($Api_Set_Obj, $opts{path});
        if(!defined $showfh{error})
        {
            $fileid = hex($showfh{fileid});
            $fileidstr = "($opts{fileop} \@fileid $fileid)";
        }
        else
        {
            $errcode =-1;
            $result= "File Path is not correct";
            return ($errcode,$result);
        }
    }

    if (defined($opts{snapid})) {
        $snapid = $opts{snapid};
        $snapidstr = "(== \@snapid $snapid)";
    } 

    if (defined($opts{level})) {
        $level = $opts{level};
        $levelstr = "(== \@level $level)";
    }

    if (defined($opts{fbn})) {
        $fbn = $opts{fbn};
        $fbnstr = "(== \@fbn $fbn)";
    }

    if (defined($opts{vbn})) {
        $vbn = $opts{vbn};
        $vbnstr = "(== \@vbn $vbn)";
    }

    $unverified = $opts{unverified};
    $count = $opts{count};
    $countstr = "repeat $count";

    ## Introduce data corruption for the volume
    $Log->trace("Introduce data corruption on volume");
    my $fijicmd;
    
    ## Introducing raid error based on type ()
    if($unverified==0)
    {
       $fijicmd = "fiji filter \"( $countstr (raid.return.pseudo.bad.error (&& (eq \@vol $vol) $levelstr $fileidstr $privstr $snapidstr $fbnstr $vbnstr)(return (true))))\"\\"; 
    }
    else
    {
        $fijicmd = "fiji filter \"( $countstr (raid.return.unverified.blk.error (&& (eq \@vol $vol) $levelstr $fileidstr $privstr $snapidstr $fbnstr $vbnstr)(return (true))))\"\\";
    }
    

    my $resp = $Api_Set_Obj->system_node_run(node => "local", command =>$fijicmd);
    my $output = Dumper($resp->get_raw_output());
    
    if (!($output =~ /filter set/i))
    {
	$result="fiji filter: $output";
        $errcode =-1;
    }
    else
    {
        $Log->comment("fiji point injected properly");
        $result=0;
        $errcode = 0;
    }

  return ($errcode,$result); 
}


sub _filer_showfh($$) {
    my $Api_Set_Obj = shift;
    my($file) = shift;

    my $cmd = "showfh $file"; 
    my $resp = $Api_Set_Obj->system_node_run(node => "local", command => "\" priv set -q diag; $cmd \"");
    
    my $text= Dumper($resp->get_raw_output());
    my %fh;
    if ($text =~ /flags=(\S+)\s+snapid=(\S+)\s+fileid=(\S+)\s+gen=(\S+)\s+fsid=(\S+)/)
    {
       $fh{flags}=$1;
       $fh{snapid}=$2;
       $fh{fileid}=$3;
       $fh{gen}=$4;
       $fh{fsid}=$5;
   
    }
    else
    {
        $fh{error}=$text;
        
    }
    return %fh;
}
1;
