#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Rep_Utils.pm#1 $

#this library is intended for C-mode Raid error propagation scripts

=head1 NAME

  FS::Raid_Utils- This library contains the Utility Functions used
  by Raid Error Propagation Scripts

=head1 DESCRIPTION

  The Raid_Utils library consists of function to perform following tasks:

  access_file_blocks() : Routines for reading the file based on message type
  access_file_blocks_and_verify_ems() : wrapper on access_file_blocks to
                                        check proper ems is generated
  

=head1 SYNOPSIS

  
    use FS::Rep_Utils;
    
    my ($errcode,$result) = wreck(api_obj =>$APISet_obj,
                           type =>'aggr', vol =>'myaggr',
                           inject_script =>'inode_type inum=64 type=1'
                           error_type =>"unverified");
                           
    if ($errcode < 0) {
        $Log->comment("wreck failed with eroor: $result");
    }
    
         
=head1 STATUS
 @status private
 
=head1 AUTHOR
 @author Hardikkumar.Darji@netapp.com

=head1 DISCLAIMER

 @disclaimer This library does NOT support duplicate vol names in the same aggr. 
             In short no vserver checking is available.
=cut


# Compiler directives.
use strict;
use warnings;

# =============================================================
# Module Import
# =============================================================

require Exporter;

package FS::Rep_Utils;

use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF OBJECT BOOLEAN UNDEF);
use NACL::APISet;
use FS::EA::EventlogTask;
use FS::Dumpblock qw(dumpblock);
use vars qw (@ISA @EXPORT_OK);
use Data::Dumper;
use Tharn;
use POSIX;
use Carp qw(croak);



# =============================================================
# Components Used
# =============================================================

use NACL::C::Node;
use NACL::C::Aggregate;
use NACL::C::Volume;
@ISA       = qw(Exporter);
@EXPORT_OK = qw(access_file_blocks access_file_blocks_and_verify_ems);

# =============================================================
# Declare Global Variables in BEGIN
# =============================================================

BEGIN {
    use vars qw(
        $Api_Set_Obj
       
        );
    #$node = NACL::C::Node->find();
}


# ===============================================================
=head2 access_file_blocks(api_set_obj=>ApiSet_obj, path => $path, cmd => $cmd, 
                        msgtype => $msgtype, nfs_client => $client)

 @summary
     Return the output by accessing  the user data from nfs, or execute
     the command on the filer and return the output depending on the msgtype.

 @description
     access_file_blocks - Verify the msgtype and execute the command either in filer or in nfs and
     return the output to the caller.
 

 @arg $api_obj Required:
      api_obj => APISet object.

 @arg $path Required:
      path => path for the file to verify corruption for
              path is reuired to find out snapid in case of
              file is in snapshot (also to verify file really exist or not)

 @arg $msgtype Required:
      msgtype => The type of the message (abortable or non-abortable)

 @arg $cmd Optional:
      cmd => Any cmd is acceptable which is having a vol_path as an argument.
      Default: rdfile => if msgtype => 'non-abortable' 
      Default: cat => if msgtype => 'abortable' 

 @arg $client Optional:
      nfs_client => Client to work on.
      Default: localhost
      
 @arg $timeout Optional:
      timeout => Timeout for rdfile.
      Default: 300

 @arg $private Optional:
      private => To access a private file throgh dumpblock.
      Default: undef

 @arg $blockno Optional:
      blockno => Blockno of the file to access throgh dumpblock.
      Default: 0

 @arg $fileid Optional:
      fileid => Inode number of the file to pass to dumpblock.
      Default: Finds and uses the fileid from the path provided.

 @arg $snapid Optional:
      snapid => snapid of the volume to pass to dumpblock. 
      Default: Finds and uses the snapid from the path provided. 
 
 @return values
      Returns the message returned by client if the msg type is 'abortable'
                      or
      Returns the output of the filer in case of 'non-abortable'.

 @example my ($errcode,$msg) = access_file_blocks(api_obj=>ApiSet_obj,
                           	           path => "/vol/testvol/testfile",
					   msgtype => "abortable",
                                         cmd => "ls");

=cut
sub access_file_blocks(%) {
    my (%opts, $api_obj, $api_set_client_obj, $host,$client_conn, $resp, $vol, $path, 
         $nfs_client, $cmd, $cmd_to_exec,$path_of_file, $msgtype, $abort_msg, 
         $non_abort_msg, $mnt, $tmout, $snap,%fh,$private, $blockno, $fileid, 
         $snapid, $str,$result,$errcode,$volume_state,$volume, $junctionpath);

    %opts = validate(
        @_,
        {
            api_obj       => { type => OBJECT,optional => 0},
            path => {type => SCALAR,optional => 0},
            nfs_client => {type => SCALAR|OBJECT, optional => 1, default => 'localhost'},
            msgtype => {type => SCALAR, default => 'abortable'},
            cmd => {type => SCALAR, optional => 1},
	    timeout => {type => SCALAR, optional => 1},
            private => {type =>SCALAR, optional =>1},
            blockno => {type =>SCALAR, optional =>1},
            fileid  => {type =>SCALAR, optional =>1},
            snapid  => {type =>SCALAR, optional =>1}, 
        }
    );
    
    $api_obj = $opts{api_obj};
    #$Log->comment("NACL Node Object creation");
    my $node = NACL::C::Node->find();
    $path = $opts{path};
    $msgtype = $opts{msgtype};
    $nfs_client = $opts{nfs_client};
    $cmd = $opts{cmd};
    $tmout = $opts{timeout};
    $blockno = $opts{blockno};
    $private = ($opts{private} == 1) ? '1' : undef;
    $fileid = $opts{fileid};
    $snapid = $opts{snapid}; 

    if (!defined($tmout)) {
	    $tmout = 300;
    }
   
    if (!defined($blockno)) {
            $blockno = 0;
    }
    if ($nfs_client =~ m!Connectrec=HASH!) {
       $client_conn = $nfs_client;
    }
    else { 
       $host = host($nfs_client);
       $client_conn = connect('#rsh');
    }
    
    # checking that path provided by user is correct or not
    # by calling filer_showfh to find fileid
    %fh=&_filer_showfh($api_obj,$path);
                     
    if(!defined($fh{error}))
    {
        # assigning fileid found using show_fh
        $fileid=hex($fh{fileid});
        
        if ($msgtype eq 'abortable')
        {
            $cmd = defined $opts{cmd} ? $opts{cmd} : 'cat';
            
            # seperating volume name and file path from the $path
            # which will be useful for NFS read
            if ($path =~ m!/vol/(.*?)/(.*)!) {
               $vol = $1;
               $path_of_file = $2;
            }
                        
            if($cmd eq "rdfile")        
            {
                $cmd_to_exec = "$path";          
                $resp = $api_obj->system_node_run(node => "local", "connectrec-timeout"=>$tmout, command => "\" priv set -q test; $cmd $cmd_to_exec \"");
                $abort_msg=Dumper($resp->get_raw_output());
                $Log->comment("abortable message");                
                $Log->comment($abort_msg);
                if($abort_msg=~ /\$VAR1\s+=\s+'\s+(.*)\s+'/)
                {
                    $abort_msg=$1;
                    if($abort_msg !~ /Unrecoverable RAID error/)
                    {
                        $abort_msg="NO Expected Error Message found";
                        $errcode = -1;
                    }
                    else
                    {
                         $errcode=0;
                    }
                }
               
                return ($errcode,$abort_msg);
            }
            else
            {   
                ### mounting the volume on client
                $volume = NACL::C::Volume->find( 
                        command_interface => $node,
                        #command_interface => "local",
                        filter            => {volume => $vol}, 
                        );
                
                $volume_state = $volume->state();                  
                $junctionpath = $volume_state->junction_path();
                
                if ($volume_state->junction_active() ne 'true'
                    or $volume_state->junction_path() !~ /\w+/ )
                {
                    $errcode =-1;
                    $result ="Junction path of the volume is not set or juction_active is not set to true";
                    return ($errcode,$result);
                }
                
                $api_set_client_obj = NACL::APISet->new(connection => $client_conn,
                                                        category  => "Host",
                                                        interface => "CLI",
                                                        set       => "Linux",
                                                        );
                my $mount_name = "/tmp/" . $vol . $$;
                
                $api_set_client_obj->execute_raw_command(
                                 'command' => "mkdir " . $mount_name );
                
                $api_set_client_obj->mount(
                                 'source-path' =>
                                 #"local"->{'hostrec'}->{'DATA_IP1'}. ":$junctionpath",
                                 $node->{'hostrec'}->{'DATA_IP1'}. ":$junctionpath",
                                 'mount-point' => $mount_name );
                
                my $resp = $api_set_client_obj->execute_raw_command( 
                'command' => "cat $mount_name/$path_of_file" );            
                
                $result= $resp;
                
                $api_set_client_obj->execute_raw_command(
                'command' => "/bin/umount $mount_name" );
                
                $api_set_client_obj->execute_raw_command(
                'command' => "/bin/rmdir $mount_name" );
                
                $errcode=0;
                return ($errcode,$result);
            
            }
        
        }
        elsif ($msgtype eq 'non-abortable')
        {
            # finding snapid from the correct path provided in snapshot
            # Examples: /vol/vv_1/metadir/.snapshot/snap1/file1
            #           /vol/vv_1/.snapshot/snap1/file
            if ($path =~ m!/vol/(.*?)/(metadir\/|).snapshot/(.*?)/(.*)!) {
               $vol = $1;
               $snap =  $3;
               $path_of_file = $4;
            }
            # finding snapid from the correct path provided in AFS
            # Examples: /vol/vv_1/metadir/file1
            #           /vol/vv_1/file
            elsif ($path =~ m!/vol/(.*?)/(metadir\/|)(.*)!) {
               $vol = $1;
               $path_of_file = $2;
            }
                
            my $no_abort = 1;
            my $block_type="file";
            
            # creating dumpblock() call to pass to eval ( which will evaluate the string as a statment)
            $str = 'dumpblock(api_obj=>$api_obj, block_type => $block_type, volume => $vol, fileid  => $fileid, blockno => $blockno, noabort => $no_abort';
            
    
            if(!defined($snapid) && defined($snap)) {
               $str .= ',snapname => $snap'; 
            }
            elsif (defined($snapid) && defined($snap)) {
               $str .= ',snapid => $snapid';
            }
            elsif (defined($snapid)) {
               $str .= ',snapid => $snapid';
            }
                
            if (defined($private)) {  
               $str .= ',private => $private';
            }
    
            $str .= ');';
            
            
            ($errcode,$non_abort_msg) = eval $str;
            
            return ($errcode,$non_abort_msg);
           
        } 
        else
        {
            $Log->comment("Here Message type is $msgtype");
            $errcode= -1;
            $result="Invalid msgtype: $msgtype";
            return ($errcode,$result);
        }
    }
    else
    {
        $Log->comment("Path provided is not correct or not exists");
        $errcode= -1;
        $result="Invalid filepath: $path";
        return ($errcode,$result);
    }
    
    
}
=head2  access_file_blocks_and_verify_ems(%)

 @summary
     Wrapper function of access_file_blocks which will provide ems error message for 
     the pattern provided along with abortable and non-abortable messages. 

 @description
     Function is only applicable for the ems triggering activities.
     Returns ems error messages along with abortable and non-abortable messages

 @arg api_obj Required:
      api_obj => NACL APISet object.
 
 @arg $pattern Required:
      pattern => pattern to match with ems log.

 @arg $path Required:
      path => path for the file to verify corruption for

 @arg $msgtype Required:
      msgtype => The type of the message (abortable or non-abortable)

 @arg $cmd Optional:
      cmd => cmd to trigger execute.
      Default: dumpblock => if msgtype => 'non-abortable'
      Default: cat or rdfile=> if msgtype => 'abortable'

 @arg $client Optional:
      nfs_client => Client to work on.
      Default: localhost

 @arg $timeout Optional:
      timeout => Timeout for rdfile.
      Default: 300
      
 @return 
      Returns the ems error message for the pattern provided and also,
      returns the abortable/non-abortable error message from the ems 
      triggering event.

 @example my ($errcode,$msg,$response) = access_file_blocks(api_obj=>ApiSet_obj,
                           	           path => "/vol/testvol/testfile",
					   msgtype => "abortable",
                                         cmd => "cat");
                                         
      $errcode is the flag to indicate pass (0) or fail (-1)
      $msg is the abortable/non abortable message
      $response is the EMS
=cut
sub access_file_blocks_and_verify_ems(%) {

    my(%opts, $api_obj, $pattern, $emsLog, $ems_msg,
        $msg, $tmout, $debug, $errcode, $response);
 
    %opts = validate(
         @_,
         {
             api_obj       => { type => OBJECT,optional => 0},
             path => {type => SCALAR},
             nfs_client => {type => SCALAR|OBJECT, optional => 1, default => 'localhost'},
             msgtype => {type => SCALAR, default => 'abortable'},
             cmd => {type => SCALAR, optional => 1},
             pattern => {type => SCALAR},
	     timeout => {type => SCALAR, optional => 1},
             
         }
       );
    
    $api_obj = $opts{api_obj};
    $pattern = $opts{pattern};
    $tmout = $opts{timeout};
    #$Log->comment("NACL Node Object creation");
    my $node = NACL::C::Node->find();
    
    #create Eventlog Task object
    my $Elogobj = EventlogTask->new( 'command_interface' => $node );
    #my $Elogobj = EventlogTask->new( 'command_interface' => "local" );
    
    if (!defined($tmout)) {
                $tmout = 300;
    }

    $Elogobj->get_time_stamp();
    if (defined $opts{cmd}) {
       ($errcode,$msg) = &access_file_blocks(api_obj => $api_obj,
                               path => $opts{path},
                               msgtype => $opts{msgtype},
                               nfs_client => $opts{nfs_client},
                               cmd => $opts{cmd},
                               timeout => $tmout,
			       );
       
    }
    else {
       ($errcode,$msg) = &access_file_blocks(api_obj => $api_obj,
                               path => $opts{path},
                               msgtype => $opts{msgtype},
                               nfs_client => $opts{nfs_client},
                               timeout => $tmout,
			       );
    }
    $Log->comment("Message from access_file_blocks: $msg");
    if($errcode == 0)
    {
        
        
        ## Stop capturing the EMS (wafl.raid.incons.userdata)
        $response = $Elogobj->capture_all_event_logs
                         ( messagename => $pattern );
    
        if (defined($response)) {
            $Log->comment("response=".$response);
            $Log->comment("Lib Detected data corruption");
        }
        else {
            $Log->comment("Lib Did not detect data corruption");
            $response ="no corruption detected";
            $errcode = -1;
        }
        
        return ($errcode,$msg,$response);
    }
    else
    {
        $response="Error accessing file using access_file_blocks() ";
        return ($errcode,$msg,$response);
    }
}

=head2  _filer_showfh($$)
 @summary
     get information about file

 @description
     executes showfh command on filer and get information about file
     like fileid,snapid,flags,gen,fsid
     
 @arg $Api_set_Obj Required:
      $Api_set_Obj => NACL APISet object.
     
 @arg $file Required:
      $file => path of file for which info. is required.
 @return
      %fh if file exist then returns hash  of (key , value) of file related information
          if file do not exist the hash contains only one pair ('error','error description')
=cut

sub _filer_showfh($$) {
    my $Api_Set_Obj = shift;
    my($file) = shift;

    my $cmd = "showfh $file"; 
    my $resp = $Api_Set_Obj->system_node_run(node => "local", command => "\" priv set -q diag; $cmd \"");
    
    my $text= Dumper($resp->get_raw_output());
    my %fh;
    
    if ($text =~ /flags=(\S+)\s+snapid=(\S+)\s+fileid=(\S+)\s+gen=(\S+)\s+fsid=(\S+)/)
    {
    
       $fh{flags}=$1;
       $fh{snapid}=$2;
       $fh{fileid}=$3;
       $fh{gen}=$4;
       $fh{fsid}=$5;
   
       return %fh;
    }
    else
    {
        $fh{error}=$text;
        return %fh;
    }
}

1;
