# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Retention.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
# --------------------------------------------------------------------------
## @summary library of functions and constants for SnapLock retention testing.
##
## @description
##    Many of these functions explicitely use a data structure that I'm
##    calling 'RETEN_CHECK'. It is an array or arrays. Each inner array
##    contains the following 4 fields (in order):
##
##      filename --  filename relative to the volume/directory,
##
##      set_atime -- 'touch -t' format date to set the atime to,
##
##      test_tstamp -- unix timestamp that it should be after commit,
##
##      delta -- the accuracy of the match, or the allowed slop.
##
##      reten_check = (['filename1', set_atime1, test_tstamp1, delta1],
##                      ['filename2', set_date2,  test_atime2, delta2],
##                      ...);
##
## @notes ()!%*@#$ CIFS... There doesn't seem to be any way to GET the
##        atime to 1-second accuracy with CIFS. Therefore the kludge
##        to read the atimes for comparison using NFS, even when
##        testing CIFS setting.
##
## @keywords snaplock, compliance, retention
##
## @status public
##
## @author dl-johny@netapp.com (FSS WAFL Quality)

BEGIN {
    use vars qw( $VOL_SIZE $SNAPLOCK_TYPE $SETUP_DONE
                 $SLCC_STRICT $CMDTIMEOUT $VOL_GUARANTEE
                 $I_WANT_TO_WASTE_DISKS $mount_point $VCC_VALUE);
    param("VOL_GUARANTEE", '-default', 'none');
    param("VOL_SIZE", '-default', '20m');
    param("SNAPLOCK_TYPE", '-default', "snaplock_enterprise");
    param("SETUP_DONE", '-default', 0); # override
    param("SLCC_STRICT", '-default', 0); # allow compliance clock to be set
    param("CMDTIMEOUT", '-default', 600); # override
    param("I_WANT_TO_WASTE_DISKS", '-default', 0); # see below...
    param("VCC_VALUE", '-default', undef );

    require "FS/commonParams.pl"
};

package FS::Retention;
use strict;
use vars qw( $VERSION @ISA @EXPORT @EXPORT_OK );
use Data::Dumper;               # useful for debugging
use Carp;
use SPARSE::Util::Logger qw( logWarn logTrc logStep logDbg logInfo );
use Tharn qw( connect host say logresult
              config_result na_result fatal_result fail_result script_result );
use snaplock qw( compliance_clock_initialize compliance_clock_status compliance_clock_status_verbose  set_retention_periods
                 parse_retention snaplock_unlicense dates_delta_in_secs
                 system_time_get client_set_atime client_get_atime
                 client_commit_file client_get_ctime );
use Utils qw( make_conn );
use Volume qw( vol_exists vol_purge );
use SayWrap qw( sayEOF );
use FS::Util::Setup qw( create_vol_and_mount create_vol umount_vol
                        common_cleanup create_and_map_cifs_share );
use FS::Util::Utilities qw( nfs_cifs_mkdir nfs_cifs_mkfile );
use Filer::Version qw( get_wafl_version WV_TSINGTAO_MIN get_ontap_version WV_ROLLINGROCK_MIN);
use Aggregate qw( aggr_exists aggr_options );
use Date::Manip;
use setup qw(is_cifs_setup);
use CIFS qw(:file);
use ExportFS qw(share_add share_delete share_exists);
use MapShare qw(map_cifs_share unmap_cifs_share);
use TestEnv::MinWin qw(:all);
use File::Spec;
use Admin::License qw(license);
use Wreck;
use Volume qw(vol_online);
require Exporter;
$VERSION = '0.5';

# globals for CIFS testing
use vars qw($new_drive $cifs_client $share_path);


@ISA = ('Exporter' );
@EXPORT_OK = qw(
                WRAP_END
                WRAP_END_STAMP
                WRAP_END_M1
                WRAP_END_M1_STAMP
                WRAP_BEGIN
                WRAP_BEGIN_STAMP
                WRAP_BEGIN_P1
                WRAP_BEGIN_P1_STAMP
                FIXED_END
                FIXED_END_STAMP
                FIXED_END_M1
                FIXED_END_M1_STAMP
                INFINITE_STAMP
                reten_can_run_test
                reten_setup
                reten_process_many
                map_to_datestr
                map_to_timestamp
                reten_cleanup
                reten_set_and_commit_one
                reten_check_one
                reten_check_one_and_log
                update_reten_check
                $new_drive
                $cifs_client
                $share_path
);

# constants for the border cases
use constant WRAP_END          => '200212312359.59'; # GMT -- touch -t format
use constant WRAP_END_STAMP    => 0x3E122F7F;        # Unix timestamp
use constant WRAP_END_M1       => '200212312359.58'; # GMT -- touch -t format
use constant WRAP_END_M1_STAMP => 0x3E122F7E;        # Unix timestamp
use constant WRAP_BEGIN        => '197001010000.00'; # GMT -- touch -t format
use constant WRAP_BEGIN_STAMP  => 0x00000000;        # Unix timestamp
use constant WRAP_BEGIN_P1     => '197001010000.01'; # GMT -- touch -t format
use constant WRAP_BEGIN_P1_STAMP => 0x00000001;      # Unix timestamp
use constant FIXED_END         => '203801190314.06'; # GMT -- touch -t format
use constant FIXED_END_STAMP   => 0x7ffffffe;        # Unix timestamp
use constant FIXED_END_M1      => '203801190314.05'; # GMT -- touch -t format
use constant FIXED_END_M1_STAMP => 0x7ffffffd;       # Unix timestamp
use constant INFINITE_STAMP    => 0x7fffffff;        # Unix timestamp

## New global variables to save retention periods, retention time stamp and volume name.

my $gbl_ret_periods;
my $gbl_ret_stamp;
my $gbl_volume_name;

# functions follow

## @name ::reten_can_run_test
##
## @description
##  Verify that the test can run due to environment stuff
##
## @synopsis
##    reten_setup(%options);
##
## @arg %options -- some are required, others default to environment
##
##    snaplock -- snaplock license type, defaults to $::SNAPLOCK_TYPE
##
##    maximum -- maximum retention period (required) -- filer format

sub reten_can_run_test (%)
{
    my %opts = @_;

    my $snaplock = defined($opts{snaplock}) ? $opts{snaplock}
      : $::SNAPLOCK_TYPE;
    my $maximum = defined($opts{maximum}) ? $opts{maximum}
      : die script_result("reten_setup: maximum is a required parameter");
    my $min_vers = defined($opts{min_version}) ? $opts{min_version}
      : WV_TSINGTAO_MIN;

    ## @step check valid release
    logStep("check valid release");
    my $host = host($::FILER);
    my $wvers = get_wafl_version($host);
    if ($wvers < $min_vers) {
        my $overs = get_ontap_version($host);
        die na_result("this test not valid for version $overs, try 7.1 or later");
    }

    ## @step check that aggregate conforms to desired type
    logStep("check that aggregate conforms to desired type: $::SNAPLOCK_TYPE");
    my $fc = make_conn($host);
    my %options;
    if (aggr_exists($fc, $::WORM_AGGR)) {
        %options = aggr_options($fc, $::WORM_AGGR);
        if ($options{snaplock_enterprise} and
            $::SNAPLOCK_TYPE ne 'snaplock_enterprise') {
            die config_result("Aggregate type mismatch: want compliance, have enterprise");
        } elsif ($options{snaplock_compliance} and
                 $::SNAPLOCK_TYPE ne "snaplock") {
            die config_result("Aggregate type mismatch: Want enterprise, have compliance");
        }
    }
    ## @step force confirmation if compliance and hardware...
    logStep("force confirmation if compliance and hardware...");
    if ($host->hosttype() !~ /simulator/ and $snaplock eq "snaplock") {
        unless($::I_WANT_TO_WASTE_DISKS) {
            die config_result("This test requires a SNAPLOCK_ENTERPRISE aggregate if run on hardware");
        } else {
            logWarn("aggregate $::WORM_AGGR will contain a volume with long expiration WORM files, and cannot be destroyed.");
        }
    }

    if ($::FS_PROTO and $::FS_PROTO eq 'cifs') {
        logStep("Verify that CIFS_CLIENT has been specified at command line");
        if (!($::CIFS_CLIENT)){
            logDbg("Input paramater CIFS_CLIENT not defined for cifs protocol");
            die config_result("Input paramater CIFS_CLIENT not defined for cifs protocol");
        }
        logStep("Verify that the windows client $::CIFS_CLIENT is MinWin Compatible");
        minwin_compatible($::CIFS_CLIENT)
          or die config_result("Windows client $::CIFS_CLIENT is not minwin compatible");
    }

}


## @name ::reten_setup
##
## @description
##  All the retention tests need the same setup. dies on failure with
##    result object;
##
## @synopsis
##  ($conn, $volname, $mountpoint, %reten_stamps) =
##      reten_setup(%options);
##
## @arg %options -- some are required, others default to environment
##
##    snaplock -- snaplock license type, defaults to $::SNAPLOCK_TYPE
##
##    minimum -- minimum retention period (required) -- filer format
##
##    default -- default retention period (required) -- filer format
##
##    maximum -- maximum retention period (required) -- filer format
##
##    subdir  -- subdir to create test files under (optional)

use POSIX qw(strftime);
sub reten_setup (%)
{
    my %opts = @_;
    $gbl_ret_periods = \%opts;

    my $snaplock = defined($opts{snaplock}) ? $opts{snaplock}
      : $::SNAPLOCK_TYPE;
    my $minimum = defined($opts{minimum}) ? $opts{minimum}
      : die script_result("reten_setup: minimum is a required parameter");
    my $default = defined($opts{default}) ? $opts{default}
      : die script_result("reten_setup: default is a required parameter");
    my $maximum = defined($opts{maximum}) ? $opts{maximum}
      : die script_result("reten_setup: maximum is a required parameter");
    my $subdir = defined($opts{subdir}) ? $opts{subdir}
      : strftime "SL%Y%m%d_%H%M%S", localtime();

    my ($conn, $vol_name, %reten_stamps, $prefix);
    my $host = host($::FILER);

    ## @step License snaplock (default to snaplock_enterprise).
    ## @step Create compliance aggregate and volume, or trad volume.
    ## @step Mount the volume via NFS.
    logStep("create compliance vol and mount");
    if ($snaplock eq 'snaplock_enterprise') {
        ($conn, $vol_name, $::mount_point) =
          create_vol_and_mount(snaplock_enterprise=>1,
                               guarantee=>$::VOL_GUARANTEE)
              or { ref($@) ? die $@ : die config_result($@) };
    } elsif ($snaplock eq 'snaplock') {
        ($conn, $vol_name, $::mount_point) =
          create_vol_and_mount(snaplock=>1,
                               guarantee=>$::VOL_GUARANTEE)
              or { ref($@) ? die $@ : die config_result($@) };
    } else {
        die config_result("invalid value for SNAPLOCK_TYPE: $snaplock");
    }
 
    $gbl_volume_name = $vol_name;

    ## @step Intialize the compliance clock, if needed.
    logStep("Intialize the compliance clock, if needed.");
    my ($init_cc_time, $inited) = compliance_clock_initialize($conn, reinit_ok=>1);

    my $wvers = get_wafl_version($host);

    ## Get CC for IC and below releases and VCC for RR and above releases.	
    my ($cc_time) = compliance_clock_status($conn, $vol_name);
    logTrc("Compliance clock time = '$cc_time' (wall clock time=". gmtime(time()) .")");

    ## Verify the retention of WORM file by changing the VCC to higher dates.
    ## These tests are applicable only for RR and above releases.
    if ( $::VCC_VALUE and $wvers >= WV_ROLLINGROCK_MIN ) {

        ## Corrupt the vcc using wreck, Use the library developed to change the VCC of the volume.
	## 'wafl cache flush' has to be called after wreck, corrupt_volinfo_secure_clock_metadata() takes care of everything.
        
        snaplock::corrupt_volinfo_secure_clock_metadata(conn => $conn , vcc_time=>$::VCC_VALUE , scc_time => 0 , scc_id => 0 , node_id => 0 , vol_name => $vol_name , vol_type => 'flex' );

        ## Verify that wreck has set VCC properly
        ## Get VCC from the filer.
        ## Since 'snaplock clock status' will not show CC time if it is greater that 2038
	## Using ;compliance_clock_status_-v' to get the CC time.
        my $cc_hash = compliance_clock_status_verbose($conn, $vol_name);
	logStep(Dumper($cc_hash));
        if ( abs( $cc_hash->{in_core_vcc} - $::VCC_VALUE ) > 15 ) {
	    logresult('FATAL' , "VCC is not set properly");
        }
        $cc_time = $cc_hash->{in_core_vcc};
    }


    ## @step compute compliance clock offset time.
    logStep("compute compliance clock offset time.");
    my $sys_time = &UnixDate(system_time_get($conn), "%s");
    logDbg("sys time='$sys_time'; compliance clock='$cc_time'");
    my $cclockdiff = $cc_time - $sys_time;
    logTrc("compliance clock offset = $cclockdiff");
    if (defined($::LARGE_DELTA)
        and abs($cclockdiff) >= parse_retention($::LARGE_DELTA)) {
        ## Test should return failure if there is a large skew b/w filer clock and compliance clock
        logWarn("Compliance clock offset ($cclockdiff) > $::LARGE_DELTA ");
        logStep("Compliance clock offset ($cclockdiff) > $::LARGE_DELTA ");
    }

    ## Changing now to use Filer compliance clock time instead of Client clock time.
    my $now = $cc_time;  # unix timestamp seconds
    my $wrap = 0x7fffffff - $now;
    logDbg("compliance clock now = time() = $now");

    ## @step Set the minimum, default, and maximum retention periods to
    ##       easily distinguishable distinct times.
    logStep("Set the retention periods.");
    $reten_stamps{now} = $now;
    my ($min_period, $default_period, $max_period) =
      set_retention_periods($conn, $vol_name, minimum=>$minimum,
                           default=>$default, maximum=>$maximum);

    logDbg("retention periods: (min,default,max)=($min_period,$default_period,$max_period)");

    $reten_stamps{minimum} = parse_retention($min_period);
    $reten_stamps{default} = parse_retention($default_period);
    $reten_stamps{maximum} = parse_retention($max_period);

    $gbl_ret_stamp = \%reten_stamps;

    logDbg("retention times: (min,default,max)=($reten_stamps{minimum},$reten_stamps{default},$reten_stamps{maximum})");

    ## @step do cifs setup if fs_proto is cifs
    if ($::FS_PROTO and $::FS_PROTO eq 'cifs') {
        logStep("Setup CIFS on the filer");
        license($conn, 'cifs');
        $new_drive = create_and_map_cifs_share(vol_name => $::WORM_VOL,
                                               filerconn => $conn,
                                               cifs_client => $::CIFS_CLIENT);
	if(!$new_drive){
        $new_drive = create_and_map_cifs_share(vol_name => $::WORM_VOL,
                                               filerconn => $conn,
                                               cifs_client => $::CIFS_CLIENT)
          or die ref $@ ? $@ : config_result $@;
	}
	logStep("Created cifs share, drive letter is set to : $new_drive : verify thats not null");

        $share_path = "$new_drive\\$subdir";
    }
    # always generate NFS paths (used for verification during CIFS tests)
    $prefix = "$::mount_point/$subdir";

    return ($conn, $vol_name, $prefix, %reten_stamps);
}

## @name ::reten_process_many
##
## @description
##  All the retention tests use the same process. dies on failure with
##    result object. logs PASS and FAIL as progresses.
##
## @synopsis
##      reten_process_many($conn, $mountpoint, @reten_check);
##
## @arg $conn -- filer connection object
##
## @arg $mountpoint -- worm volume mount point (on client)
##
## @arg @reten_check -- array of arrays of test records.
##
##      reten_check = (['filename1', set_atime1, test_tstamp1, delta1],
##                     ['filename2', set_date2,  test_atime2, delta2],
##                      ...);
##    where filename is the filename relative to the volume/directory,
##      set_atime is the 'touch -t' format date to set the atime to,
##      test_tstamp is the unix timestamp that it should be after commit,
##      delta is the accuracy of the match, or the allowed slop.

sub reten_process_many ($$%)
{
    my ($conn, $mount_point, @reten_check) = @_;

    my ($mount_path, $client);
    my $cc = make_conn($::CLIENT);

    if (defined($::FS_PROTO) and $::FS_PROTO eq 'cifs') {
        $client = $::CIFS_CLIENT;
        $mount_path = $share_path;
    } else {
        $client = $cc;
        $mount_path = $mount_point;
    }

    logStep("create the working subdir");
    my $ret = nfs_cifs_mkdir($mount_path,
                             client=>$client,
                             protocol=>$::FS_PROTO);

    my ($filename, $cifs_filename, $path);
    my $ver = get_wafl_version($conn->{hostp}) ;
    if (  $::VCC_VALUE  and $ver >= WV_ROLLINGROCK_MIN ) {
           ## Dont run existing test sets.
    } else {
        for my $f (0..$#reten_check) {
            ## @step Create file.
            $filename = $mount_point.'/'.$reten_check[$f][0];
            if (defined($::FS_PROTO) and $::FS_PROTO eq 'cifs') {
                $cifs_filename = $share_path.'\\'.$reten_check[$f][0];
                $path = $cifs_filename;
            } else {
                $path = $filename;
            }
            logStep("Create file $filename");
            my $result = nfs_cifs_mkfile($path,
                                     protocol=>$::FS_PROTO,
                                     client=>$client);
            ## @step Set the atime
            ## @step Commit the files (chmod -w files).
            my $atime = reten_set_and_commit_one($cc,
                                             $filename,
                                             $reten_check[$f][1],
                                             $cifs_filename);
            ## @step Verify that the atimes are set correctly
            # NOTE: always done via NFS
            reten_check_one_and_log($cc, $filename, $reten_check[$f][2],
                                $reten_check[$f][3], $cifs_filename);
        }
    }
  
    ## New tests added for RR. 
    if ( $ver >= WV_ROLLINGROCK_MIN ) {
        ## Verify the retention by commiting the file to default retention of the volume.
        ## Make the minimum/maximum  retention as default retention.
        my @new_ret_check = ( "minimum" , "maximum", "default" ) ;
        foreach my $period ( @new_ret_check  ) {
            my $def_value; 
            if ( $period eq "minimum" ) {$def_value = "min"; } 
            elsif($period eq "maximum" ) { $def_value = "max" ; }
            else { $def_value = $gbl_ret_periods->{$period} ; }
            set_retention_periods($conn, $gbl_volume_name , minimum=> $gbl_ret_periods->{minimum},default=>$def_value , maximum =>$gbl_ret_periods->{maximum}  );

            $filename = $mount_point.'/'.$period."_vol_ret";
            if (defined($::FS_PROTO) and $::FS_PROTO eq 'cifs') {
                $cifs_filename = $share_path.'\\'.$period."_vol_ret";
                $path = $cifs_filename;
            } else {
                $path = $filename;
            }  
            logStep("Create file $filename");
            my $result = nfs_cifs_mkfile($path,
                                     protocol=>$::FS_PROTO,
                                     client=>$client);

            # Commit the files (chmod -w files).
            logStep("Commit the file (chmod -w $filename).");
            $result = client_commit_file($client, $path, protocol=>$::FS_PROTO)
                and logWarn("commiting file: $result");

            if ( $gbl_ret_periods->{$period} eq "infinite" ) {
		$period = INFINITE_STAMP;
            } 

            ### Verify the retention time stamp of committed files.
            my $expected = map_to_timestamp($period ,%$gbl_ret_stamp );
            logStep("Printing expected value $expected");
            reten_check_one_and_log($cc, $filename, $expected , $::SMALL_DELTA);
        } 
    }
}

## @name ::reten_set_and_commit_one
##
## @description
##  Set the retention date on one file, verify it was set,
##  and commit it.
##
## @synopsis
##    $atime = reten_set_and_commit_one($conn, $filename, $atime);
##
##   The returned atime is in unix timestamp format.
##
## @arg $conn -- client connection object
##
## @arg $filename -- client-relative path of file
##
## @arg $atime -- the atime to set, in 'touch -t' format

sub reten_set_and_commit_one ($$$;$)
{
    my ($cc, $nfs_path, $set_atime, $cifs_path) = @_;

    my ($result, $client, $filename);

    if (defined($::FS_PROTO) and $::FS_PROTO eq 'cifs') {
        $client = $::CIFS_CLIENT;
        $filename = $cifs_path;
    } else {
        $client = $cc;
        $filename = $nfs_path;
    }
    # Set the atime
    logStep("Set the atime of $filename to $set_atime");
    if ($set_atime ne 'unset') {
        $result = client_set_atime($client, $filename, $set_atime,
                                   protocol=>$::FS_PROTO)
          and logWarn("setting atime: $result");
    }
    # check the atime
    my $atime = client_get_atime($cc, $nfs_path);
    $atime =~ s/^\D*(\d+)\D*\r*\n*$/$1/;
    logTrc("set atime before commit for $filename = $atime");

    # Commit the files (chmod -w files).
    logStep("Commit the file (chmod -w $filename).");
    $result = client_commit_file($client, $filename, protocol=>$::FS_PROTO)
      and logWarn("commiting file: $result");

    return $atime;
}

## @name ::reten_check_one
##
## @description
##   Get the atime of $filename, and compare it to the $expected
##   within the $delta, return 1/0 as pass/fail
##
## @synopsis
##      reten_check_one($conn, $filename, $expected, $delta);
##
## @arg $conn -- client connection object
##
## @arg $filename -- client-relative path of file
##
## @arg $expected -- expected atime (unix timestamp format)
##
## @arg $delta -- allowed deviation between actual and expected

sub reten_check_one ($$$$)
{
    my ($cc, $filename, $expected, $delta) = @_;

    my $basename = $filename;
    $basename =~ s,^.*/,,;

    if ($expected =~ /^\+(-?\d+)/) {
        my $incr = $1 + 0;
        my $ref = client_get_ctime($cc, $filename);
        $expected = _wrap_clip($ref + $incr);
    }
    ## @step Verify that the atimes are set correctly
    logStep("Verify atime of $filename set to $expected");
    my $atime = client_get_atime($cc, $filename);
    if ($atime =~ s/\D*(\d+)\D*\r*\n*/$1/) {
        logTrc("committed atime for $filename: $atime (compare: $expected)");
        if (abs($expected - $atime) <= $delta) {
            return 1;
        } else {
            return 0;
        }
    } else {
        die "error on client $::CLIENT: $atime";
    }
    # NOTREACHED
}

## @name ::reten_check_one_and_log
##
## @description
##   Get the atime of $filename, and compare it to the $expected
##   within the $delta, log the PASS/FAIL result, and continue
##
## @synopsis
##      reten_check_one_and_log($conn, $filename, $expected, $delta);
##
## @arg $conn -- client connection object
##
## @arg $filename -- client-relative path of file
##
## @arg $expected -- expected atime (unix timestamp format)
##
## @arg $delta -- allowed deviation between actual and expected

sub reten_check_one_and_log ($$$$)
{
    my ($cc, $filename, $expected, $delta) = @_;

    my $basename = $filename;
    $basename =~ s,^.*/,,;

    if ($expected =~ /^\+(-?\d+)/) {
        my $incr = $1 + 0;
        my $ref = client_get_ctime($cc, $filename);
        $expected = _wrap_clip($ref + $incr);
    }
    ## @step Verify that the atimes are set correctly
    logStep("Verify atime of $filename set to $expected");
    my $atime = client_get_atime($cc, $filename);
    if ($atime =~ s/\D*(\d+)\D*\r*\n*/$1/) {
        logTrc("committed atime for $filename: $atime (compare: $expected)");
        if ( abs($expected - $atime) <= $delta ) {
            logresult("PASS", "setting $basename passed");
        } else {
            logresult("FAIL", "setting $basename failed");
        }
    } else {
        die "error on client $::CLIENT: $atime";
    }
}

## @name ::update_reten_check
##
## @description
##   The reten_check array has fields that need to get mapped to
##   date- and time-stamps at runtime. This updates them.
##
## @synopsis
##   update_reten_check(\%reten_stamps, \@reten_table);
##
## @arg %reten_stamps -- reference to table of retention periods
##                       as timestamps and the offset in seconds.
##
## @arg @reten_table -- the table of retention tests

sub update_reten_check(\%\@)
{
    my ($stamps, $check) = @_;

    my %reten_stamps = %$stamps;
    my @reten_check = @$check;

    for my $i (0..$#reten_check) {
        $reten_check[$i][1] = map_to_datestr($reten_check[$i][1],
                                             %reten_stamps);
        $reten_check[$i][2] = map_to_timestamp($reten_check[$i][2],
                                               %reten_stamps);

        # check the settings...
        logDbg("row $i: (@{$reten_check[$i]})");
    }
}

## @name ::_wrap_clip
##
## @description
##  take a timestamp value, wrap it at INFINITE_STAMP, and truncate to
##    WRAP_END_M1_STAMP. internal function

sub _wrap_clip
{
    my ($val) = @_;
    if ($val > INFINITE_STAMP) {
        $val -= INFINITE_STAMP;
        $val = ($val >= WRAP_END_STAMP) ? WRAP_END_M1_STAMP : $val;
    }
    return $val;
}


## @name ::map_to_datestr
##
## @description
##  convert an intent string to a 'touch -t' datestring. returns
##   datestrings untouched.
##
## @synopsis
##    $datestr = map_to_datestr($object, %reten);
##
## @arg $obj -- object to convert
##
## @arg %reten -- retention timestamps and offset
##
##    minimum -- minimum retention period, seconds (required)
##
##    default -- default retention period, seconds (required)
##
##    maximum -- maximum retention period, seconds (required)
##
##    offset  -- set atime offset in seconds (required)

sub map_to_datestr ($%)
{
    my ($object, %reten) = @_;
    my $value;

    if ($object =~ /^min.*plus/) {
        # using $reten{now} instead of time() handles large compliance
        # clock offsets better
        $value = _wrap_clip($reten{now} + $reten{minimum} + $reten{offset});
    } elsif ($object =~ /^min.*minus/) {
        $value = _wrap_clip($reten{now} + $reten{minimum} - $reten{offset});
    } elsif ($object =~ /^max.*plus/) {
        $value = _wrap_clip($reten{now} + $reten{maximum} + $reten{offset});
    } elsif ($object =~ /^max.*minus/) {
        $value = _wrap_clip($reten{now} + $reten{maximum} - $reten{offset});
    } elsif ($object =~ /^now.*plus/) {
        $value = _wrap_clip($reten{now} + $reten{offset});
    } else {
        return $object;
    }
    my @t = gmtime($value);
    return sprintf("%04d%02d%02d%02d%02d.%02d",
                   $t[5]+1900,$t[4]+1,$t[3],$t[2],$t[1],$t[0]);
}

## @name ::map_to_timestamp
##
## @description
##  convert an intent string to a unix timestamp. returns
##   timestamps untouched.
##
## @synopsis
##    $timstamp = map_to_timestamp($object, %reten);
##
## @arg $obj -- object to convert
##
## @arg %reten -- retention timestamps and offset
##
##    minimum -- minimum retention period, seconds (required)
##
##    default -- default retention period, seconds (required)
##
##    maximum -- maximum retention period, seconds (required)
##
##    offset  -- set atime offset in seconds (required)

sub map_to_timestamp ($%)
{
    my ($object, %reten) = @_;
    my $value;

    if ($object =~ /^min.*plus/) {
        $value = $reten{minimum} + $reten{offset};
        if (_wrap_clip(time() + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object =~ /^min.*minus/) {
        $value = $reten{minimum} - $reten{offset};
        if (_wrap_clip(time() + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object =~ /^max.*plus/) {
        $value = $reten{maximum} + $reten{offset};
        if (_wrap_clip(time() + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object =~ /^max.*minus/) {
        $value = $reten{maximum} - $reten{offset};
        if (_wrap_clip(time() + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object eq 'minimum') {
        $value = $reten{minimum};
        # using $reten{now} instead of time() handles large compliance
        # clock offsets better
        if (_wrap_clip($reten{now} + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object eq 'maximum') {
        $value = $reten{maximum};
        if (_wrap_clip($reten{now} + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object eq 'default') {
        $value = $reten{default};
        if (_wrap_clip($reten{now} + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } elsif ($object =~ /^now.*plus/) {
        $value = $reten{offset};
        if (_wrap_clip(time() + $value) == WRAP_END_M1_STAMP) {
            $value = WRAP_END_M1_STAMP
        } else {
            $value = sprintf('+%d',$value );
        }
    } else {
        return $object;
    } 
    return $value;
}

## @name ::reten_cleanup
##
## @description
##  standard cleanup for retention tests.
##
## @synopsis
##   reten_cleanup($conn, $vol_name);
##
## @arg $conn -- filer connection object
##
## @arg $vol_name -- name of WORM volume

sub reten_cleanup($$;$)
{
    my ($conn, $vol_name) = @_;

    ## @step Cleanup by destroying the volume.
    logStep("Cleaning up: CLEANUP == $::CLEANUP");
    if (! defined($conn)) {
        my $host = host($::FILER);
        $conn = connect($::FILER_CONN, -timeout, $::TIMEOUT);
        $conn->timeout($::CMDTIMETOUT);
    }
    my $vol = defined($vol_name) ? $vol_name : $::WORM_VOL;
    if ($::CLEANUP ne 'none') {
        # cleanup options are limited --
        # can't wait_file_expiry() here: files are effectively permanent
        umount_vol($::mount_point) if defined($::mount_point);
        snaplock_unlicense($conn) or logWarn $@;
        if ($::FS_PROTO and $::FS_PROTO eq 'cifs') {
            if ($new_drive) {
                unmap_cifs_share(client	=> $::CIFS_CLIENT,
                                 drive	=> $new_drive);
            }
            if (share_exists($conn, $::CIFS_SHARE)) {
                share_delete($conn, $::CIFS_SHARE);
            }
        }
    }
    common_cleanup($conn, snaplock=>1, snaplock_enterprise=>1)
      or logWarn($@);
}

1;
