
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/CIFS_File_Session.pm#1 $

## @summary     FS::Util::CIFS_File_Session - NATE library for testing CIFS file attribute changes.
##
## @author      mcrae@netapp.com
##
## @status      private (in development, needs up-to-date doc)
##
## @pod here

=head1 NAME

    FS::Util::CIFS_File_Session - NATE library for testing CIFS file attribute changes.

=head1 SYNOPSIS

    use FS::Util::CIFS_File_Session;

    # create a new FS::Util::CIFS_File_Session object here
    my $file = FS::Util::CIFS_File_Session->new(
	host => $CIFS_CLIENT,
	file => $path,
	open_args => {
	    create_disposition => FS::Util::CIFS_File_Session::FILE_OPEN_IF,
	},
    );

    my $result;

    $result = $file->get_size();

    $result = $file->get_file_size();

    $result = $file->get_alloc_size();

    $result = $file->set_size(
	attr => 's',
	size => 23,
    );

    $result = $file->set_size(
	attr => 's',
	size => 23,
	keep_open => 0,
    );

    $result = $file->set_size(
	attr => 's',
	size => 23,
	keep_open => 1,
    );

=head1 STATUS

  Private, in development.

=head1 DESCRIPTION

  This module provides session-based file manipulation methods for testing CIFS file attributes
  via the supporting .exe files found under /x/eng/localtest/arch/win32/WAFL/bin, including support
  for holding a file open in one session while other operations are performed in another.

=cut

package FS::Util::CIFS_File_Session;

use 5.008008;
use strict;
use warnings;

require Exporter;

use Params::Validate qw(validate BOOLEAN SCALAR HASHREF ARRAYREF);
use Carp qw(croak);
use Data::Dumper;

# need to use an import list here because we define a disconnect() method
use Tharn qw(say);

use FS::Util::CIFS_Session;

# these constants map the indices of the components within the list that say() returns.
use constant SAY_MATCH_IDX  => 0;
use constant SAY_BEFORE_IDX => 1;
use constant SAY_AFTER_IDX  => 2;
use constant SAY_PROMPT_IDX => 3;

# change this when our version is incorporated into the shared MinWin.pm (in "phase 2")
# also, note other MinWin package references to be changed below
use dblade::wafl::cifs::temp::TestEnv::MinWin qw(minwin_compatible execute_cmd_interactive);
use SayWrap qw(sayEOF);

BEGIN {
    Params::Validate::validation_options(
        # allow presence of options other than those named in the validation specification
        allow_extra => 1,
   );
}

our @ISA = qw(Exporter FS::Util::CIFS_Session);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use vol_db ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = (
    'all' => [ qw() ]
);

our @EXPORT_OK = (
    @{ $EXPORT_TAGS{'all'} },
    # allow these constant definitions to be imported into the calling package
    'FILE_SUPERSEDE',
    'FILE_OPEN',
    'FILE_CREATE',
    'FILE_OPEN_IF',
    'FILE_OVERWRITE',
    'FILE_OVERWRITE_IF',
);

our @EXPORT = qw();

our $VERSION = '0.01';

=head1 METHODS

=over

=item new() - create a new instance of this class, connected to the given (Windows) host.

=over

=item host

[Required] hostname or NATE HostRec reference indicating the target CIFS client system.

=item file

[Optional] client system path to the target file. if this value is not provided, then it
will be set by calling the open() method with a target file name. this value must be set  
before calling any methods other than new() or open().

=item open_args

[Optional] should be a hash of arguments suitable for passing to the open() method, to be
used when the target file is intially opened. (note that the 'file' arg must be defined
if 'open_args' is specified)

=item drive_letter

[Optional] the drive letter to use for mounting the shared Win32 repository, i.e. /x/eng/localtest/arch/win32. defaults to 'Y'. Note that the MS telnet service requires drives to be mapped freshly in each session, even if such maps already exist on the system.

=back

=cut

sub new {
    my $class = shift;
    $class = ref($class) || $class;

    # check args
    my (%args) = validate(
	@_,
	{
	    host => {
		type => SCALAR | HASHREF,
	    },
	    file => {
		type => SCALAR,
		optional => 1,
	    },
	    # these args get passed to the open() sub
	    open_args => {
		type => HASHREF,
		default => undef,
	    },
	    drive_letter => {
		type => SCALAR,
		default => '*',
	    },
	}
    );
    die "open requires a file name" if (defined($args{open_args}) && (!defined($args{file})));

    # initialize self
    my $self = $class->SUPER::new(%args);

    $self->{file}            = "$args{file}" if (exists $args{file});
    $self->{fsetattr_prompt} = 'Enter the command:';

    # this dies if it something goes wrong
    $self->open(%{$args{open_args}}) if (defined($args{open_args}));

    # done
    return $self;
}

=item fopen() - open the given file from the target client, via the fopen.exe command.

=over

=item create_disposition

[Optional] this argument determines how the file is opened, default values is FILE_OPEN. For details, see http://wikid.netapp.com/w/User:Raghavc/Test_plan/SMB_file_allocation_size#fopen. this module defines the following constants for use as values for this arg -

=over

=item FILE_SUPERSEDE

=item FILE_CREATE

=item FILE_OPEN

=item FILE_OPEN_IF

=item FILE_OVERWRITE

=item FILE_OVERWRITE_IF

=back

=item alloc_size

[Optional] a number indicating the desired file allocation size, default is 1.

=item keep_open

[Optional] a boolean indicating whether to hold the file open or not. default is not (0, or false). if this value is true, then the file should be closed (when ready) via the press_any_key() method.

=item expected_error_code

[Optional] a number indicating an 'expected' error code, to be interpreted as a successful open.

=back

=cut

# create_disposition values for file open: these were obtained from
#     http://wikid.netapp.com/w/User:Raghavc/Test_plan/SMB_file_allocation_size#Test_Cases
# and
#     http://msdn.microsoft.com/en-us/library/bb432380(VS.85).aspx
#
# If the file already exists, replace it with the given file. If it does not, create the given file.
use constant FILE_SUPERSEDE => 0;
#
# If the file already exists, open it instead of creating a new file. If it does not, fail the request and do not create a new file.
use constant FILE_OPEN => 1;
# 
# If the file already exists, fail the request and do not create or open the given file. If it does not, create the given file.
use constant FILE_CREATE => 2;
# 
# If the file already exists, open it. If it does not, create the given file.
use constant FILE_OPEN_IF => 3;
# 
# If the file already exists, open it and overwrite it. If it does not, fail the request.
use constant FILE_OVERWRITE => 4;
# 
# If the file already exists, open it and overwrite it. If it does not, create the given file.
use constant FILE_OVERWRITE_IF => 5;
 
# regular expressions for determining file open status
use constant MATCH_OPEN_FAILURE => qr^Failed to open/create the file\.\s+Error code = (.*)\s+Error = (.*)\s*^;
use constant MATCH_OPEN_SUCCESS => qr^Successfully opened/created the file\.^;

sub open {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    file => {
		type => SCALAR,
		optional => 1,
	    },
	    create_disposition => {
		type => SCALAR,
		callbacks => {
		    'matches_defined' => sub {
			my $value = shift;
			
			($value eq FILE_SUPERSEDE) or 
			($value eq FILE_CREATE) or 
			($value eq FILE_OPEN) or 
			($value eq FILE_OPEN_IF) or 
			($value eq FILE_OVERWRITE) or 
			($value eq FILE_OVERWRITE_IF);
		    },
		},
		# by default, only open the file if it already exists
		default => FILE_OPEN,
	    },
	    alloc_size => {
		type => SCALAR,
		default => 1,
	    },
	    keep_open => {
		type => BOOLEAN,
		default => undef,
	    },
	    # for negative test cases, pass in the expected error code
	    expected_error_code => {
		type => SCALAR,
		default => undef,
	    },
	}
    );

    # error checking
    if (defined($args{file})) {
	if (defined($self->{file})) {
	    croak "file arg provided, but a file value is already set for this instance" if ($self->{file} ne $args{file});
	} else {
	    $self->{file} = $args{file};
	}
    } else {
	croak "no file value set, don't know what to open" if (!defined($self->{file}));
    }

    # make sure an interactive command is not already running
    if (defined($self->{session_owner})) {
	croak "file '$self->{file}' is already open (via '$self->{session_owner}'), you must call close() first.";
    }

    # define this as a separate value (used in multiple places)
    my $exe = $self->{minwin_xshare} . "\\WAFL\\bin\\fopen.exe";

    # build the command and execute it
    my $cmd = "$exe \"$self->{file}\" $args{create_disposition} $args{alloc_size}";
    my $response = execute_cmd_interactive($self->{session_conn}, $cmd, -prompt, 'Press any key to close the file:', -swallow);

    # check the result
    if ($response->[SAY_AFTER_IDX] =~ MATCH_OPEN_FAILURE) {
	my $error_code = $1;
	my $error = $2;
	
	$Tharn::Log->trace("error_code is $error_code, error is $error\n");

	# is this an expected failure?
	if (defined($args{expected_error_code}) && ($args{expected_error_code} eq $error_code)) {
	    $Tharn::Log->comment("found expected error code when trying to open the file '$self->{file}' (error code '$error_code' - '$error'): full response is '@$response'");
	    return 1;
	}

	# a real error, apparently
	croak "failed to open the file '$self->{file}' (error code '$error_code' - '$error'): full response is '@$response'";
    } elsif ($response->[SAY_AFTER_IDX] =~ MATCH_OPEN_SUCCESS) {
	# this exe now owns the session
	$self->{session_owner} = $exe;

	unless (defined($args{keep_open}) && $args{keep_open}) {
	    # close if keep_open not requested (this clears session_owner, among other things)
	    $self->close();
	}

	return 1;
    } else {
	croak "failed to recognize client response '" . $response->[SAY_AFTER_IDX] . "'";
    }

    # how do we get here?
    croak "unrecognized response (1): '@$response'";
}

sub _get_fsetattr_session {
    my $self = shift;

    my $new_session = 0;

    #my $exe = $self->{minwin_xshare} . '\WAFL\bin\fsetattr.exe';
    #my $exe = $self->{minwin_xshare} . '\WAFL\bin\new-fsetattr.exe';
    my $exe = $self->{minwin_xshare} . '\WAFL\bin\rel-fsetattr.exe';

    croak "no file value set" if (!defined($self->{file}));

    if (defined($self->{session_owner})) {
	if ($self->{session_owner} ne $exe) {
	    croak "file '$self->{file}' is already open (via '$self->{session_owner}'), you must call close() first.";
	}
    } else {
	$new_session = 1;
	my $cmd = "$exe \"$self->{file}\"";
	my $response = execute_cmd_interactive($self->{session_conn}, $cmd, -prompt, $self->{fsetattr_prompt}, -swallow);

	if ($response->[SAY_PROMPT_IDX] ne $self->{fsetattr_prompt}) {
	    croak "unrecognized response (2):\n" . Dumper($response) . "\n";
	}

	# add the fsetattr prompt to the list of prompts for this connection
	my @prompts = $self->{session_conn}->prompts();
	$self->{session_conn}->prompts($self->{fsetattr_prompt}, @prompts);

	$self->{session_owner} = $exe;
    }

    return ($self->{session_conn}, $new_session);
}

=item get_size() - retrieve the file size from an ongoing fsetattr.exe session. returns a hash containing the file size (file_size) and the file allocation size (alloc_size).

=over

=item keep_open

[Optional] a boolean indicating whether to hold the file open or not. default is not (0, or false). if this value is true, then the file should be closed (when ready) via the press_any_key() method.

=back

=cut

# regex used to extract values from get size operations
use constant MATCH_SIZE => qr/Getting attributes of the file..\s+EOF = (\d+)\s+Allocation Size = (\d+)\s+/;

sub get_size {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    keep_open => {
		type => BOOLEAN,
		default => undef,
	    },
	}
    );

    my %values;

    my ($session, $new) = $self->_get_fsetattr_session();

    my ($match, $before, $after, $prompt) = say($session, 'g');

    if ($new) {
	unless (defined($args{keep_open}) && $args{keep_open}) {
	    $self->close();
	}
    }

    if ($after =~ MATCH_SIZE) {
        %values = (
	    file_size   => $1,
	    alloc_size => $2,
	);
    } else {
	croak "unrecognized response (3): '$after'";
    }

    return \%values;
}

=item get_file_size() - same as get_size(), but returns just the file size as a scalar.

=cut

sub get_file_size {
    my $self = shift;

    my $result_hash = $self->get_size(@_);

    return $result_hash->{file_size};
}

=item get_alloc_size() - same as get_size(), but returns just the file allocation size as a scalar.

=cut

sub get_alloc_size {
    my $self = shift;

    my $result_hash = $self->get_size(@_);

    return $result_hash->{alloc_size};
}

=item set_size() - execute the fsetattr.exe command on the given file from the target client, in order to set either the file size or the file allocation size. returns a hash containing the requested size (req_size), the file size (file_size) and the file allocation size (alloc_size).

=over

=item attr

[Required] which file attribute to set: use 's' to set the file size and 'a' to set the file allocation size.

=item size

[Required] a number indicating the desired size. note that file allocation sizes are set according to the underlying block size of the target filesystem, so the value provided here may be rounded up when setting the file allocation size.

=item keep_open

[Optional] a boolean indicating whether to hold the file open or not. default is not (0, or false). if this value is true, then the file should be closed (when ready) via the press_any_key() method.

=item expected_error_code

[Optional] a number indicating an 'expected' error code, to be interpreted as a successful operation.

=back

=cut

# regex used to extract values from set size output
use constant MATCH_SIZE_CHANGE => qr/Changing (?:the|allocation) size of the file to (\d+)\.\.\.\s+Current attributes of the file:\s+EOF = (\d+)\s+Allocation Size = (\d+)\s+/;
use constant MATCH_SIZE_FAILURE => qr/^Failed to change allocation size of the file: Error code (\d+) - (.*)$/m;

sub set_size {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    attr => {
		type => SCALAR,
	    },
	    size => {
		type => SCALAR,
	    },
	    keep_open => {
		type => BOOLEAN,
		default => undef,
	    },
	    expected_error_code => {
		type => SCALAR,
		default => undef,
	    },
	}
    );

    croak "no file value set" if (!defined($self->{file}));

    my ($session, $new) = $self->_get_fsetattr_session();

    #my ($match, $before, $after, $prompt) = say($session, "$args{attr} $args{size}", -swallow);
    my @response = say($session, "$args{attr} $args{size}", -swallow);

    if ($new) {
	unless (defined($args{keep_open}) && $args{keep_open}) {
	    $self->close();
	}
    }

    my %values;

    if ($response[SAY_AFTER_IDX] =~ MATCH_SIZE_CHANGE) {
        %values = (
	    req_size   => $1,
	    file_size  => $2,
	    alloc_size => $3
	);
    } elsif ($response[SAY_AFTER_IDX] =~ MATCH_SIZE_FAILURE) {
	my $error_code = $1;
	my $error = $2;

	if ($args{expected_error_code} eq $error_code) {
	    $Tharn::Log->comment("found expected error code when trying to open the file '$self->{file}' (error code '$error_code' - '$error'): full response is '@response'");
	    return 1;
	} else {
	    croak "failed to open the file '$self->{file}' (error code '$error_code' - '$error'): full response is '@response'";
	}
    } else {
	croak "unrecognized response (4): '@response'";
    }

    return \%values;
}

=item set_file_size() - same as set_size(attr => 'f', ...). any args (other than 'attr') are just passed through to the set_size() method.

=cut

sub set_file_size {
    my $self = shift;

    return $self->set_size(
	@_,
        attr => 's',
    );
}

=item set_alloc_size() - same as set_size(attr => 'a', ...). any args (other than 'attr') are just passed through to the set_size() method.

=cut

sub set_alloc_size {
    my $self = shift;

    return $self->set_size(
	@_,
        attr => 'a',
    );
}

=item close() - closes down the current session (if any). call this to close down a session started with keep_open=1 (see set_size() and open() args)

=cut

sub close {
    my $self = shift;

    if (!defined($self->{session_owner})) {
        croak "no session to close";
    }

    my ($match, $before, $after, $prompt);
    if ($self->{session_owner} =~ /fopen.exe$/) {
	($match, $before, $after, $prompt) = say($self->{session_conn}, "\n");
    } elsif ($self->{session_owner} =~ /fsetattr.exe$/) {
	($match, $before, $after, $prompt) = say($self->{session_conn}, 'c', -prompt, 'Closing the file...');
    } else {
        croak "unrecognized 'session_owner' value: '$self->{session_owner}'";
    }
    delete $self->{session_owner};
    return 1;
}

=item fwrite()

=over

=item offset

[Required] where to begin the write within the target file - for details, see http://wikid.netapp.com/w/User:Raghavc/Test_plan/SMB_file_allocation_size#fwrite.

=item offset

[Optional] where to start writing (how many bytes into the file)

=item length

[Optional] how many bytes to write.

=back

=cut

# regular expressions for capturing values from write operation output
use constant MATCH_WRITE_FAILURE => qr^Failed\s.*\sError code (\d+)\s+-\s+(.*)\s+$^;
use constant MATCH_WRITE_SUCCESS => qr^Attributes of the file after write:\s+EOF = (\d+)\s+Allocation Size = (\d+)\s+^;

sub write {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    offset => {
		type => SCALAR,
	    },
	    length => {
		type => SCALAR,
	    },
	},
    );

    my ($session, $new) = $self->_get_fsetattr_session();

    my ($match, $before, $after, $prompt) = say($session, "w $args{offset} $args{length}");

    if ($new) {
	unless (defined($args{keep_open}) && $args{keep_open}) {
	    $self->close();
	}
    }

    my %values;

    if ($after =~ MATCH_WRITE_SUCCESS) {
        %values = (
	    file_size  => $1,
	    alloc_size => $2
	);
    } else {
	croak "unrecognized response (5): '$after'";
    }

    return \%values;
}

# make sure to close everything out upon object destruction
sub DESTROY {
    my $self = shift;

    # interesting corner case - if we die during construction, then the
    # object goes out of scope immediately and DESTROY is called before
    # the exception is propagated. this means we need to preserve any eval
    # error before proceeding.
    my $eval_error = $@ if ($@);

    eval {
	$self->close()      if (defined($self->{session_owner}));
	$self->disconnect() if (defined($self->{session_conn}));
    };
    # tack on any exceptions thrown above
    $eval_error .= $@ if ($@);

    # now, reset the eval error
    $@ = $eval_error;
}

=head1 SEE ALSO

  BURT 324637
  http://wikid.netapp.com/w/User:Ymichael/BURT324637
  http://wikid.netapp.com/w/User:Raghavc/Test_plan/SMB_file_allocation_size
  dblade/wafl/cifs/test_cifs_file_module.thpl (unit test script for this library, not complete or comprehensive)
  dblade/wafl/cifs/temp/TestEnv/MinWin.pm (temporary)
  TestEnv/MinWin.pm
  NATE documentation

=head1 AUTHOR

  Chris McRae (mcrae@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/CIFS_File_Session.pm#1 $
  Last Revised $Date: 2013/01/22 $

=head1 COPYRIGHT AND LICENSE

  Copyright (c) 2008 Network Appliance, Inc.
  All rights reserved.

=cut

1;

