
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/CIFS_Session.pm#1 $

## @summary     FS::Util::CIFS_Session - NATE library for managing persistent Windows sessions, for
##              driving interactive CLI programs.
##
## @author      mcrae@netapp.com
##
## @status      private (in development, needs up-to-date doc)
##
## @pod here

=head1 NAME

    FS::Util::CIFS_Session - NATE class library for managing persistent Windows sessions,
                             for driving interactive CLI programs.

=head1 SYNOPSIS

    use FS::Util::CIFS_Session;

    # create a new FS::Util::CIFS_Session object here
    my $file = FS::Util::CIFS_Session->new(
	host => $CIFS_CLIENT,
	drive_letter => 'Q',
	max_conn => 42,
    );

    ...

=head1 STATUS

  Private, in development.

=head1 DESCRIPTION

  This module provides basic Windows session connectivity and setup support, including:
  - establishing telnet sessions to a given CIFS client.
  - mapping and unmapping shares within a telnet session.
  This functionality is built on that provided by the TestEnv::MinWin module.

  This module serves as a common parent class to the FS::Util::CIFS_Client_Session and
  FS::Util::CIFS_File_Session classes.

=cut

package FS::Util::CIFS_Session;

use 5.008008;
use strict;
use warnings;

require Exporter;

use Params::Validate qw(validate BOOLEAN SCALAR HASHREF ARRAYREF OBJECT);
use Carp qw(croak);
use Data::Dumper;

use Tharn qw();

# this is handy when debugging
#BEGIN { @INC = grep !ref($_), @INC; }

# change this when our version is incorporated into the shared MinWin.pm (in "phase 2")
# also, note other MinWin package references to be changed below
use dblade::wafl::cifs::temp::TestEnv::MinWin;

BEGIN {
    Params::Validate::validation_options(
        # allow presence of options other than those named in the validation specification
        allow_extra => 1,
   );
}

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use vol_db ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = (
    'all' => [ qw() ]
);

our @EXPORT_OK = (
    @{ $EXPORT_TAGS{'all'} },
);

our @EXPORT = qw();

our $VERSION = '0.01';

# note this is a class-wide value (each instance of this class sees the same, shared data)
our %Telnet_Connections_By_Client;

=head1 METHODS

=cut

=over

=item new() - create a new instance of this class, connected to the given (Windows) host.

=over

=item host

[Required] hostname or NATE HostRec reference indicating the target CIFS client system.

=item drive_letter

[Optional] the drive letter to use for mounting the shared Win32 repository, i.e. /x/eng/localtest/arch/win32. defaults to '*'. Note that the MS telnet service requires drives to be mapped freshly in each session, even if such maps already exist on the system.

=item max_conn

[Optional] the MS telnet server on the target host is configured to allow maxconn simultaneous sessions. the default value is 99, but other values can be provided via this arg.

=back

=cut

sub new {
    my $class = shift;
    $class = ref($class) || $class;

    my (%args) = validate(
	@_,
	{
	    host => {
		type => SCALAR | HASHREF,
	    },
	    drive_letter => {
		type    => SCALAR,
		default => '*',
	    },
	    max_conn => {
		type    => SCALAR,
		default => 99,
	    },
	}
    );

    # initialize self
    my $self = bless {
	host => $args{host},
    }, $class;

    croak 'target host ($self->{host}) not minwin compatible'
	unless (dblade::wafl::cifs::temp::TestEnv::MinWin::minwin_compatible($self->{host}));

    my $output = dblade::wafl::cifs::temp::TestEnv::MinWin::execute_cmd_special($self->{host}, "tlntadmn config maxconn=$args{max_conn}");
    if ($output !~ /The settings were successfully updated./) {
	croak "failed to set max telnet connections for host '$self->{host}'"
    }

    ($self->{session_conn}, $self->{minwin_xshare}) = $self->new_session(
	drive_letter => $args{drive_letter},
    );

    # done
    return $self;
}

=item new_session() - establish a new session to the target client, using the given drive letter
for the session's minwin share. returns the list returned by the TestEnv::MinWin::get_telnet_conn()
function, see that module for details.

=over

=item drive_letter

[Optional] the drive letter to use for mounting the shared Win32 repository, i.e. /x/eng/localtest/arch/win32. defaults to '*'. Note that the MS telnet service requires drives to be mapped freshly in each session, even if such maps already exist on the system.

=back

=cut

sub new_session {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    drive_letter => {
		type    => SCALAR,
		default => '*',
	    },
	}
    );

    # initialize per-client connection count
    if (!exists $Telnet_Connections_By_Client{$self->{host}}) {
	$Telnet_Connections_By_Client{$self->{host}} = 0;
    }

    my @result = dblade::wafl::cifs::temp::TestEnv::MinWin::get_telnet_conn(
	$self->{host},
	$Telnet_Connections_By_Client{$self->{host}},
	$args{drive_letter},
    );

    # increment *after* the call completes
    $Telnet_Connections_By_Client{$self->{host}}++;

    # apply global CMDTIMEOUT value (this should probably be an arg instead...)
    if (defined($::CMDTIMEOUT)) {
        $result[0]->timeout($::CMDTIMEOUT);
    }

    return @result;
}

=item map_drive() - within the default session, map the given share onto the given drive letter.

=over

=item drive_letter

[Required] the drive letter to use for mounting the shared Win32 repository.

=item share

[Required] the share name to be mapped.

=back

=cut

sub map_drive {
    my $self = shift;
    my $drive = shift;
    my $share = shift;

    # Add ":" if it's not already there.
    if (($drive ne '*') && ($drive !~ /:$/)) {
	$drive .= ':';
    }

    # just call MinWin::map_drive() with our telnet conn and whatever
    # args are given by the caller
    return dblade::wafl::cifs::temp::TestEnv::MinWin::map_drive(
	$self->{session_conn},
	$drive,
	$share,
	say_opts => [ '-swallow' ],
	@_,
    );
}

=item unmap_drive() - unmap the given drive.

=over

=item drive_letter

[Required] the drive letter to be unmapped.

=back

=cut

sub unmap_drive {
    my $self = shift;
    my $drive = shift;

    # Add ":" if it's not there.
    if (($drive ne '*') && ($drive !~ /:$/)) {
	$drive .= ':';
    }

    # just call TestEnv::MinWin::unmap_drive() with our session conn
    return dblade::wafl::cifs::temp::TestEnv::MinWin::unmap_drive(
	$self->{session_conn},
	$drive,
	force => 1,
    );
}

=item disconnect() - disconnect the current session to the target client. returns true or dies.

=over

=item conn

[Optional] the ConnectRec object to be disconnected, defaults to the default session.

=back

=cut

sub disconnect {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    conn => {
		type => OBJECT,
		default => $self->{session_conn},
	    },
	}
    );

    if (defined($args{conn})) {
	# close it down and clear it out
	Tharn::disconnect($args{conn}, -logout);
	delete $self->{session_conn} if ($args{conn} eq $self->{session_conn});
    }

    return 1;
}

# make sure to close everything out upon object destruction
sub DESTROY {
    my $self = shift;

    # interesting corner case - if we die during construction, then the
    # object goes out of scope immediately and DESTROY is called before
    # the exception is propagated. this means we need to preserve any eval
    # error before proceeding.
    my $eval_error = $@ if ($@);

    if (defined($self->{session_conn})) {
	eval { $self->disconnect(); };
	$eval_error .= "disconnect error: $@" if ($@);
    }

    # now, reset the eval error
    $@ = $eval_error;
}

=head1 SEE ALSO

  BURT 324637
  http://wikid.netapp.com/w/User:Ymichael/BURT324637
  http://wikid.netapp.com/w/User:Raghavc/Test_plan/SMB_file_allocation_size
  dblade/wafl/cifs/test_cifs_file_module.thpl (unit test script for this library, not complete or comprehensive)
  dblade/wafl/cifs/temp/TestEnv/MinWin.pm (temporary)
  TestEnv/MinWin.pm
  NATE documentation
  MS telnet service reference - http://technet.microsoft.com/en-us/library/cc772789.aspx

=head1 AUTHOR

  Chris McRae (mcrae@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/CIFS_Session.pm#1 $
  Last Revised $Date: 2013/01/22 $

=head1 COPYRIGHT AND LICENSE

  Copyright (c) 2008 Network Appliance, Inc.
  All rights reserved.

=cut

1;

