# Let emacs know that we are editing -*-perl-*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/LowLevelUtils.pm#1 $
##
## @summary 'Low-level' utilities library -- created to break cyclic library dependencies
## @description
## A few low-level utility functions. Functions were moved here from 'Setuu.pm', 'Utilities.pm',
## etc. to break depenency cycles.
## @usage
##  use FS::Util::LowLevelUtils;
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public

package FS::Util::LowLevelUtils;

use strict;
no strict qw(refs);

use Exporter qw(import);
BEGIN {
    our @EXPORT = qw(
		     &uSystem

		     &util_chmod
                     &util_unlink
                     &util_rmdir
                     &util_mkdir
                     &util_rename
                     &util_mkfile
                     &util_link
                     &util_filesize
                     &util_stat
                     &util_exists
                     &disk_zero_spares

                     &write_file

		     &snapshot_status
		     &get_vfiler_root_path
		     &deltree
		     &delfile
		     &recursive_rm_path_check
		     &acceptable_rm_path
		     &mount_vol
		     &umount_vol
		     );
}

use Carp qw(croak);
use Data::Dumper;
use Tharn;
use POSIX;
use Module::Common qw(validate_mixed);

use Params::Validate qw(validate BOOLEAN SCALAR HASHREF ARRAYREF SCALARREF CODEREF OBJECT UNDEF);

use Admin::License qw(license unlicense is_licensed);
use Utils qw(make_conn make_host make_conn_and_complain);
use ExportFS qw(:nfs export_path);
use mount qw(mount umount);
use SayWrap qw(sayEOF);
use File::Find qw(find);
use Data::Dumper;

my $sudo_cmd = defined($main::SUDO_PATH) ? $main::SUDO_PATH : "/usr/local/bin/sudo";


my $RM_NOTIFY_LIST = [qw(dl-waflqa@netapp.com)];




#= Run a shell command and report its exit status or verify its output
#= :command        command to run
#= :[cond]        output success expr "=~ /good/" or "!~ /fail/"
#= >undef|error string
sub uSystem($;$)
{
        my ($cmd, $exp) = @_;

	# check the command for attempts to recursively delete root (dies
	# if such an attempt is detected)
	recursive_rm_path_check($cmd);

        logTrc("Running command: $cmd");

        if (!defined($exp)) {
                my $pid = fork();
                croak("fork failed: $!") unless defined($pid);

                if ($pid == 0) {        # child
			exec($cmd);
                        #croak("exec($cmd): $!");
                }

                if ((my $p = waitpid($pid, 0)) != $pid) {
                        croak("waitpid found wrong child: $p  expect $pid");
                }

                return uStatus($?);
        }

        # $exp is defined from here on.
        # grab stout and stderr
        open(RES, "$cmd 2>&1 |")        or croak("could not start $cmd: $!");
        my @res = <RES>;
        close(RES);
        logDbg("$cmd: \"@res\"");

        # check against expected output
        my $exc = $@;        # preserve any current exception
        my $match = 0;
        if (!@res) {
                croak("bad result from \"$cmd\": \"$_\"");
        }
        for (@res) {
                if (eval("q{$_} $exp")) {
                        $match = 1;
                        last;
                }
        }
        croak("bad result from \"$cmd\": \"@res\"") if $match == 0;

        $@ = $exc;
        return undef;
}



## @name ::util_chmod
## @arg $mode - required: the mode bits to change to
## @arg $path - required: the target path
## @arg conn - optional: a named parameter referencing a connection
## object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
## this parameter is only meaningful when the 'conn' parameter is also
## given.
## @description 'chmod' on the given path on the given client.
## returns 1 on success, 0 and $@ on failure.

sub util_chmod($$;%)
{
    my ($mode, $path, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    if (! defined($cc)) {
        logDbg("changing to mode $mode for file $path");
        return chmod($mode, $path);
    } else {
        $mode = sprintf("0%o",$mode);
        logDbg("changing to mode $mode for file $path");
        my ($match, undef, $after) = say($cc, "chmod '$mode' $path");
        if ($after =~ /permission|denied|error|fail|cannot|can't change/i) {
            $@ = $after;
            return 0;
        } else {
            return 1;
        }
    }
}

## @name ::util_unlink
## @arg $srcpath - required: the target path to the file
## @arg conn - optional: a named parameter referencing a connection
##  object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
##  This parameter is only meaningful when the 'conn' parameter is also
##  given.
## @description create a link between the two files on the given client.
## returns 1 on success, 0 and $@ on failure.

sub util_unlink($;%)
{
    my ($srcpath, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    my ($bs, $count) = (1, 0);

    my $ret;
    if (! defined($cc)) {
        $ret = uSystem("unlink $srcpath");
        return $@ ? 0 : 1;
    } else { # use client connection
            my ($match, undef, $after) =
                say($cc, "unlink $srcpath");
            if ($after =~ /permission|denied|error|fail|cannot/i) {
                $@ = $after;
                return 0;
            }
    }
    return 1;
}

## @name ::util_rmdir
## @arg $path - required: the directory path to be created
## @arg conn - optional: a named parameter referencing a connection
## object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
## this parameter is only meaningful when the 'conn' parameter is also
## given.
## @description 'rmdir -p' on the given path on the given client.
## returns 1 on success, 0 and $@ on failure.

sub util_rmdir($;%)
{
    my ($path, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    if (! defined($cc)) {
        return !uSystem("$sudo_cmd rmdir -p $path");
    } else {
        my ($match, undef, $after) = say($cc, "$sudo_cmd rmdir -p $path");
        if ($after =~ /permission|denied|error|fail|cannot/i) {
            $@ = $after;
            return 0;
        } else {
            return 1;
        }
    }
}

## @name ::util_mkdir
## @arg $path - required: the directory path to be created
## @arg conn - optional: a named parameter referencing a connection
## object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
## this parameter is only meaningful when the 'conn' parameter is also
## given.
## @description 'mkdir -p' on the given path on the given client.
## returns 1 on success, 0 and $@ on failure.

sub util_mkdir($;%)
{
    my ($path, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    if (! defined($cc)) {
        return !uSystem("mkdir -p $path");
    } else {
        my ($match, undef, $after) = say($cc, "mkdir -p $path");
        if ($after =~ /permission|denied|error|fail|cannot/i) {
            $@ = $after;
            return 0;
        } else {
            return 1;
        }
    }
}

## @name ::util_rename
## @arg $path1 - required: the source path
## @arg $path2 - required: the destination path
## @arg conn - optional: a named parameter referencing a connection
## object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
## this parameter is only meaningful when the 'conn' parameter is also
## given.
## @description 'mv $path1 $path2' on the given path on the given client.
## returns 1 on success, 0 and $@ on failure.
## @notes this function only correctly works with plain files and
## directories.

sub util_rename($$;%)
{
    my ($path1, $path2, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    if (! defined($cc)) {
        return rename($path1, $path2);
    } else {
        my (undef, undef, $ret) = say($cc, "test -f '$path2' -o -d '$path2' && echo target exists");
        if ($ret =~ /target exists/) {
            logWarn("rename target '$path2' exists before rename");
        }
        my ($match, undef, $after) = say($cc, "mv $path1 $path2",
                                         -timeout, $timeo);
        (undef, undef, $ret) = say($cc, "test -f '$path1' -o -d '$path1' && echo not moved");
        if ($ret =~ /not moved/) {
            $@ = $after;
            return 0;
        } else {
            return 1;
        }
    }
}


## @name ::util_mkfile
## @arg $path - required: the target path to create file
## @arg size - optional: a named parameter giving the desired size.
##  created contents are all zeros (or random). exclusive with content
## @arg content - optional: a named parameter with the initial contents
##  of the file. Exclusive with size, and overrides.
## @arg mode - optional: a named parameter with the initial mode bits.
## @arg conn - optional: a named parameter referencing a client connection
##  object through which the create/write file should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
##  This parameter is only meaningful when the 'conn' parameter is also
##  given.
## @arg random - optional: if true, copy $size bytes from /dev/urandom.
##  Will fail if /dev/urandom doesn't exist.
## @description create a file on the given path on the given client.
##  If neither size nor content is supplied, default is that the initial
##  content is the path.
## returns 1 on success, 0 and $@ on failure.

sub util_mkfile($;%)
{
    my ($path, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;
    my $size = $opts{size};
    my $mode = $opts{mode};
    my $content = $opts{content} || (defined($size) ? undef : $path);
    my $random = $opts{random} ? "/dev/urandom" : "/dev/zero";

    my ($bs, $count) = (1, 0);
    if (defined($size)) {
        if ($size == 0) {
            $count = 0;
            $bs = 1;
        } elsif ($size % 4096 == 0) {
            $bs = 4096;
            $count = $size/4096;
        } elsif ($size % 512 == 0) {
            $bs = 512;
            $count = $size/512;
        } else {
            $count = 1;
            $bs = $size;
        }
    }
    my $ret;
    if (! defined($cc)) {
        if ($content or $size == 0) {
            open(OUT, ">$path") or $@ = $!, return 0;
            print OUT $content if $content;
            close(OUT);
            if ($mode) {
                return chmod($mode, $path);
            }
        } else {
            $ret = uSystem("dd if=$random of=$path count=$count bs=$bs");
            $@ = $ret unless $ret =~ /records out/;
        }
        return $@ ? 0 : 1;
    } else { # use client connection
        if ($content) {
            my ($match, undef, $after) = say($cc, "echo '$content' > $path");
            if ($after =~ /permission|denied|error|fail|cannot/i) {
                $@ = $after;
                return 0;
            }
        } elsif ($size == 0) {
            my ($match, undef, $after) =
                say($cc, "touch $path");
            if ($after =~ /permission|denied|error|fail|cannot/i) {
                $@ = $after;
                return 0;
            }
        } else {
            my ($match, undef, $after) =
                say($cc, "dd if=$random of=$path count=$count bs=$bs");
            if ($after =~ /permission|denied|error|fail|cannot/i) {
                $@ = $after;
                return 0;
            }
        }
        if ($mode) {
            return util_chmod($mode, $path, conn=>$cc);
        }
    }
    return 1;
}

## @name ::util_link
## @arg $srcpath - required: the target path to create file
## @arg $refpath - required: a named parameter giving the desired size.
##  created contents are all zeros (or random). exclusive with content
## @arg $type - Option: Describes the type of link "s" for symbolic or "h" for hard
## @arg conn - optional: a named parameter referencing a connection
##  object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
##  This parameter is only meaningful when the 'conn' parameter is also
##  given.
## @description create a link between the two files on the given client.
## returns 1 on success, 0 and $@ on failure.

sub util_link($$;%)
{
    my ($srcpath, $refpath, %opts) = @_;

    my $type = $opts{type} || "h";
    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    my ($bs, $count) = (1, 0);

    # if user specified "h" for a hardlink then just blank it out
    #  if they specified an "s" for sym link then leave it alone
    $type =~ s/h//i;
    my $ret;
    if (! defined($cc)) {
        $ret = uSystem("ln $type $srcpath $refpath");
        return $@ ? 0 : 1;
    } else { # use client connection
            my ($match, undef, $after) =
                say($cc, "ln $srcpath $refpath");
            if ($after =~ /permission|denied|error|fail|cannot/i) {
                $@ = $after;
                return 0;
            }
    }
    return 1;
}

## @name ::util_filesize
## @arg $path - required: the target path to create file
## @arg conn - optional: a named parameter referencing a connection
##  object through which the delete should be done.
## @description return the size of the file at path on (optional) client.
## returns size on success, die on failure.

sub util_filesize($;%)
{
    my ($path, %opts) = @_;

    my $size;
    my $cc = $opts{conn};
    if ($cc) {
        (undef, undef, $size) = say($cc, "/usr/local/bin/perl -le 'print 0+(stat(qq($path)))[7];'");
    } else {
        $size = (stat($path))[7];
    }
    chomp($size);
    return $size;
}

## @name ::util_stat
## @arg $path - required: the target path to create file
## @arg conn - optional: a named parameter referencing a connection
##  object through which the delete should be done.
## @description return the stat structure of the file at path on
## (optional) client as an array.
## returns stat array on success, die on failure.
## For non-existent file, an empty array is returned.
## Array structure:
## ($dev, $ino. $mode, $nlink, $uid, $gid, $rdev, $size, $atime $mtime, $ctime,
##   $blksize, $blocks)
sub util_stat($;%)
{
    my ($path, %opts) = @_;
    my @arr;
    my $cc = $opts{conn};
    if ($cc) {
        my $str;
        my @s;
        (undef, undef, $str) = say($cc, "/usr/local/bin/perl -le '\@s=stat(qq($path));print join(qq( ),\@s);'");
        @arr = split(/\s+/,$str);
    } else {
        @arr = stat($path);
    }
    chomp(@arr);
    return @arr;
}

## @name ::util_exists
## @arg $path - required: the target path to create file
## @arg conn - optional: a named parameter referencing a connection
##  object through which the delete should be done.
## @description Returns "1" if the given path points to an existing file/dir/link/etc.
## (optional) client as an array.
## returns stat "1" or "0" on success, die on failure.

sub util_exists($;%)
{
    my ($path, %opts) = @_;
    my @arr = util_stat($path, %opts);
    my $exists = ((scalar(@arr) > 0)?1:0);
    logDbg("util_exists  Line: ".__LINE__."   Path: \"$path\"   Conn: \"$opts{conn}\"   Returning: \"$exists\"");
    return($exists);
}


## @name ::disk_zero_spares
## @arg $conn - required: a connection object to filer
## @description Starts a disk zeroing process; does NOT wait for completion

sub disk_zero_spares($)
{
    my ($conn) = @_;
    my $ret = sayEOF($conn, "disk zero spares");
    return;
}



## @name ::acceptable_rm_path
## @arg path - required: the path to validate
## @description This routine returns false if the given path contains any bare
## slashes or repeated slases. It is meant to be used to validate a path before
## attempting to recursively remove it.
sub acceptable_rm_path($) {
    my ($path) = @_;

    my @statinfo = stat($path);

    # per Tap (tapl@netapp.com): the inode number of a root directory of a
    # filesystem is 2 for Linux's ext2/ext3, Solaris's UFS and IBM's JFS.'
    return ((($path ne '/') && ($path !~ m%//+%)) ||  $statinfo[1] > 3);

}

## @name ::_parse_recursive_rm_target_paths
## @arg $str - required: the command string to parse
## @description This routine considers the given string as a unix shell command line
## and examines it for any embedded recursive rm commands. For each such command found,
## the target paths are collected and the overall list returned to the caller.
sub _parse_recursive_rm_target_paths($) {
    my ($str) = @_;

    croak script_result(__PACKAGE__ . "::_parse_recursive_rm_target_paths: cannot check an empty string.") if (!$str);

    my @target_paths;

    # remove any embedded newlines (and any preceeding backslashes)
    $str =~ s/\\?\n//g;

    # split the line into separate commands, using three
    # command delimiters: '&&', '||', ';'
    my @commands = split(/\&\&|\|\||;/, $str);
    @commands = ($str) if (!scalar(@commands));

    # trim spaces from end of each command
    chomp(@commands);

    # trim spaces from beginning of each command
    @commands = grep($_ =~ s/^\s*//, @commands);

    my $orig_value = $";
    $" = "\n\t";
    logDbg("commands found:\n\t@commands\n");
    $" = $orig_value;

    # this sub-pattern matches a space, a slash or the beginnning of a line.
    #   -> the matched text is NOT captured.
    my $rm_prefix = '(?:/|^|\s)';

    # this sub-pattern matches any group of command line args which are preceeded
    # by either one or two hyphens and which don't contain the letter 'r'.
    #   -> the matched text is NOT captured.
    my $arg_group_pattern = '(?:\s+-{1,2}[^r\s]+)*';

    foreach my $cmd (@commands) {
	# check whether the command includes 'rm -r'
	#   note: the regular expression below says
	#     - look for the $rm_prefix pattern (the matched text is NOT captured)
	#     - followed by the string 'rm'
	#     - optionally followed by an 'arg_group_pattern' as defined above
	#     - followed by an arg group pattern that DOES contain the
	#     letter 'r'
	#     - optionally followed by an 'arg_group_pattern' as defined above
	#     - followed by some whitespace
	#     - followed by a string of characters which is the LIST OF PATHS
	#     TARGETED BY THE EMBEDDED 'rm' COMMAND AND IS CAPTURED IN $1.
	#     - optionally followed by some whitespace
	#     - followed by the END-OF-LINE
	if ($cmd =~ m%${rm_prefix}rm\s+${arg_group_pattern}-{1,2}\S*r\S*${arg_group_pattern}\s+(.+)\s*$%) {
	    # get the paths targeted by the current recursive rm command
	    my @paths = split(/\s+/, $1);

	    # remove outermost pair of enclosing quotes from each path (if any)
	    my @trimmed;
	    while (my $path = pop(@paths)) {
		my $tmp = $path;

		# remove enclosing single quotes?
		$tmp =~ s/^'(.*)'/$1/;

		# or, remove enclosing double quotes?
		$tmp =~ s/^"(.*)"/$1/ if (length($tmp) == length($path));

		push(@trimmed, $tmp);
	    }

	    # add the target paths to our list
	    push(@target_paths, @trimmed);
	}
    }

    return @target_paths;
}


## @name ::deltree
## @arg $path - required: the path to be removed
## @arg conn - optional: a named parameter referencing a connection
## object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
## this parameter is only meaningful when the 'conn' parameter is also
## given.
## @description This routine recursively removes the given path using
## system(). The target path is first verified to make sure that it
## doesn't map to the filesystem root (/). The function dies with a
## NATE script result if the path doesn't exist or is unacceptable.
## If the system() command is actually executed, then the exit value,
## signal_num and dumped_core values found in $? are returned. For
## more information about these values, see 'perldoc -f system'.
##
## Options:
##	cc:	'client connection'	Connection to NFS client
##	   If given, then this connection is used to execute the 'rm'
##	   command. Otherwise client is assumed to be 'localhost',
##	   and the 'rm' command command is executed here.
##
##	timeo:	'command timeout'	Timeout for 'rm' command.
##	    Default is 300 sec.
##
##	retain_dir:  'retain_directory'
##	    If '1', then all directories and files under the given
##	    directory are deleted, but the directory itself is
##	    retained. Otherwise	the directory is deleted also --
##	    (equiv. to 'rm -rf $DIR/*'). Default is '0', so directory
##	    is deleted.
##
sub deltree($;%) {
    my ($path, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;
    my $retain_dir = $opts{retain_dir};

    if (!defined($cc) && (! -e $path)) {
	croak script_result(__PACKAGE__ . "::deltree: no such path: '$path'");
    }

    if ($retain_dir) {
	if ($path =~ /[^\/\*]$/) {
	    $path .= '/*';
	} elsif ($path =~ /\/$/) {
	    $path .= '*';
	}
    }

    my $cmd = "rm -rf \"$path\"";

    # this call dies if the path maps to the root directory
    recursive_rm_path_check($cmd);

    if (!defined($cc)) {
	my $failed = system("$sudo_cmd $cmd");
	return(!$failed);
    } else {
	my $match = (say($cc, $cmd, '-timeout', $timeo))[2];
        # ignore 'stty:' messages as not errors
        $match =~ s/^stty: standard input: .*$//m;
        chomp($match);
	croak("deltree: client rm command failed: $match")
          if ($match =~ /error|permission|denied/);
	return 1;
    }
}

## @name ::delfile
## @arg $path - required: the path to be removed
## @arg conn - optional: a named parameter referencing a connection
## object through which the delete should be done.
## @arg timeout - optional: a named parameter specifying the timeout.
## this parameter is only meaningful when the 'conn' parameter is also
## given.
## @description This routine removes the file specified by the given path.
sub delfile($;%) {
    my ($path, %opts) = @_;

    my $cc = $opts{conn};
    my $timeo = $opts{timeout} || 300;

    if (!defined($cc) && (! -e $path)) {
	croak script_result(__PACKAGE__ . "::delfile: no such path: '$path'");
    }

    my $cmd = "$sudo_cmd rm -f \"$path\"";

    if (!defined($cc)) {
	my $failed = system($cmd);
	return(!$failed);
    } else {
	my $match = (say($cc, $cmd, '-timeout', $timeo))[2];
	croak("delfile: client rm command failed: $match") if ($match);
	return 1;
    }
}

## @name ::recursive_rm_path_check
## $cmd - required: a unix command line string which may contain a recursive
## rm command.
## @description This routine checks the given string for instances of recursive
## rm's. Each path targeted by such a command is checked using acceptable_rm_path()
## to see if it is acceptable or not. When an unacceptable path is found, the
## function dies with a script result.
sub recursive_rm_path_check($) {
    my ($cmd) = @_;

    # get the list of paths targeted by any embedded recursive rm commands
    my @recursive_rm_target_paths = _parse_recursive_rm_target_paths($cmd);

    # log a trace message
    my $orig_value = $";
    $" = "\n\t";
    logTrc("recursive rm paths found:\n\t@recursive_rm_target_paths\n")
					    if (@recursive_rm_target_paths);
    $" = $orig_value;

    # check the recursive rm paths to (hopefully) prevent attempts
    # to remove the root
    foreach my $path (@recursive_rm_target_paths) {
	if (acceptable_rm_path($path)) {
	    logTrc("acceptable rm path: '$path'");
	} else {
	    # build a stack backtrace
	    my $backtrace;
	    for (my $i=1; 1; $i++) {
		my ($package, $file, $line) = caller $i;

		# bail when we get to the bottom of the stack
		last if (!defined($package));

		$backtrace .= ${file} . " line $line\n";
	    }

	    # get the current hostname
	    my $host = hostname();

	    # build our message
	    my $msg = "\nA script you were running on host '$host' passed an unacceptable path ($path) for a recursive remove command. Refer to the stacktrace below.\n\n\n$backtrace\n";
	    my $subject = 'WARNING: A SCRIPT ATTEMPTED TO RECURSIVELY DELETE ROOT';
	    my $tester = "$ENV{LOGNAME}\@netapp.com";

	    # send mail
	    my %mail = (
		To      => $tester,
		Cc      => join(',', @{$RM_NOTIFY_LIST}),
		From    => $tester,
		Subject => $subject,
		Message => "$msg"
	    );

	    logDbg("sending notification email:\n" . Dumper(\%mail) . "\n");

	    sendmail(%mail) or warn $Mail::Sendmail::error;

	    croak script_result(__PACKAGE__ . "::recursive_rm_path_check: can't continue - unacceptable recursive rm path: '$path'");
	}
    }
}





#  get_vfiler_root_path
sub get_vfiler_root_path($$) {
    my ($h, $vfiler) = @_;
    my $pathstat = sayEOF($h, "vfiler status -a $vfiler");
    my @paths = ($pathstat =~ m#Path: (\S+)#mg);
    return $paths[0];
}


## @name ::snapshot_status
## @usage snapshot_list($filer, $volume)
## @synopsis
##    Returns a list of all the snapshots in the specified volume
##    returned by advance command 'snap status' as array of hash
##    references, keys are taken from header
##
##    Volume vol0
##    snapid  status     date           ownblks release fsRev name
##    ------  ------     ------------   ------- ------- ----- --------
##    245     complete   Nov 03 20:00      6941     7.2    71 hourly.0
##    243     complete   Nov 03 16:00      2141     7.2    71 hourly.1
##
##    Volume vl00327_874b_c95a8134 (cleaning summary map)
##    snapid  status     date           ownblks release fsRev name
##    ------  ------     ------------   ------- ------- ----- --------
##      1     creating   Nov 03 21:53         0     7.2    71 fsr-u60(0033604436)_vl00327_874b_c95a8134.1 (no map)
##

sub snapshot_status($$) {
	my ($filer, $volume) = @_;
	my @snapshots = ();
	my $text = sayEOF($filer, "priv set -q advanced;snap status $volume");
	croak($text) if( $text =~ /does not exist or is not online/ );	

	my @hkeys = ();
	my $sncnt=0;
	foreach( split( '[\n\r]',$text) ) {
		chomp; s/^\s+//;
		s/(\w\w\w)\s+(\d\d)\s+(\d\d:\d\d)/$1_$2_$3/;	# work around dates
		@hkeys = split('\s+',$_) if( /snapid/ );
		if( /^[0-9]/ ) {
			my $i=-1;
			foreach( split('\s+',$_) ) {
				s/_/ /g if($hkeys[++$i] =~ /date/ );	# restore space in dates
				$snapshots[$sncnt]->{ $hkeys[$i] } = $_ if($hkeys[$i]);
			}
			$sncnt++;
		}
	}
	return(@snapshots);
}


## @name ::mount_vol
##
## @summary Mounts the volume on localhost (or CLIENT), when 'dir' option is omittted.
##	    If 'dir' option is given, then mounts the given directory for the given volume.
##          Returns: mount point
##
## @arg %opts. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER
##		May also be either a Hostrec or Connectrec object.
##
##         client - hostspec of client. defaults to $main::CLIENT
##		May also be either a Hostrec or Connectrec object.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         vol          - Required. Name of the filer volume to mount, or else path (on filer)
##                         to the volume (or volume/dir) to be mounted. Defaults to $main::FILER_VOL.
##                        Examples:
##                                     vol1   /vol/vol1    /vol/vol1/qtree1
##
##         vol_path, filer_vol  -- 'alias' options for the 'vol' option.
##
##	   dir - name of directory which is a direct decsendent of the parent volume.
##               If given, this directory will be mounted rather than the entire volume.
##		 default is undef, so entire volume is mounted. The resulting 'vol_path'
##	         will then be the concatenation:  "/" . "vol" . <VOLUME> . "/" <DIR>
##
##         filer_vol_mount - filer mount target. defaults to
##                           $main::FILER_VOL_MOUNT
##
##         mountpoint - local mount point. defaults to $main::MOUNTPOINT
##
##         timeout   - Timeout in secs. for mount to complete. Defaults to 300 (5 min).
##
##
sub mount_vol(%)
{
    my (%opts) = @_;
    my %raw_opts = %opts;

    $raw_opts{filer}         = ref($raw_opts{filer})         if (ref($raw_opts{filer})); 
    $raw_opts{filer_conn}    = ref($raw_opts{filer_conn})    if (ref($raw_opts{filer_conn})); 
    $raw_opts{client}        = ref($raw_opts{client})        if (ref($raw_opts{client})); 
    logDbg("Enter mount_vol   Line: ".__LINE__."    Opts:\n\n".Dumper(\%raw_opts));

    my $FILER	 	= $opts{fc}  || $opts{fh}  ||  $opts{filer} ||
			      $main::FILER;  # Can be HostSpec, Hostrec, or Connectrec.
    my $FILER_CONN 	= $opts{filer_conn}    || $main::FILER_CONN;
    my $VOL_PATH	= $opts{vol_path}      || $opts{vol} 
                               || $opts{filer_vol}  ||  $main::FILER_VOL ;
    my $MOUNTPOINT 	= $opts{mountpoint};
    my $CLIENT 		= $opts{client}        || $main::CLIENT || 'localhost';
    my $FILER_IP 	= $opts{filer_ip}      || $main::FILER_IP;
    my $MOUNTOPTS 	= $opts{mountopts}     || $main::MOUNTOPTS
                               || 'actimeo=0,noac,hard,intr,noquota';

    my $TIMEOUT         = $opts{timeout}       || $main::TIMEOUT;
    my $CMDTIMEOUT 	= $opts{cmdtimeout}    || $main::CMDTIMEOUT;
    my $CHMOD 		= defined($opts{chmod}) ? $opts{CHMOD} : '0777';
    my $DIR             = $opts{dir};


    return eval {
	my %opts_copy = (%opts);
	if ( ref($opts_copy{filer}) ) {
		$opts_copy{filer} = sprintf("%s", $opts_copy{filer});
	}
	if ( ref($opts_copy{client}) ) {
		$opts_copy{client} = sprintf("%s", $opts_copy{client});
	}
	logDbg("mount_vol  Line: ".__LINE__."\n\nOpts:\n\n".Dumper(\%opts_copy));

	if ($opts{filer_conn} && ref($opts{filer_conn}) =~ /connectrec/i) {
		my $logmsg = "mount_vol   Line: ".__LINE__.
		             "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n".
		             "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n".
		             "Will set 'filer' option to this object, and delete 'filer_conn' option";
#		logDbg($logmsg);
		$FILER = $opts{filer_conn};
		$opts{filer} = $opts{filer_conn};
		$FILER_CONN = $main::FILER_CONN;
		$opts{filer_conn} = $main::FILER_CONN;
	}

        logDbg("mount_vol  Line: ".__LINE__."  Establish filer connection\n".
		 "  --  FILER: \"$FILER\"");
        # filer connection
        my ($fc, $fh);
        my $conn_timeout = $TIMEOUT;
        if (ref($FILER) =~ /connectrec/i) {
            $fc = $FILER;
        } else {
            $fc = make_conn($FILER, $FILER_CONN);
	    if (!$fc) {
	        logDbg("mount_vol  Line: ".__LINE__."   Could not establish filer connection: $@");
	        die("mount_vol:  Could not establish filer connection: $@");
	    }
        }
        $fh = $fc->hostp();
	my $hostname = $fh->hostname();
	my $filer_ip = $FILER_IP || $fh->default_ip();
        if (defined($CMDTIMEOUT)) {
    	    $fc->timeout($CMDTIMEOUT);
        }
	# client connection/host
	# XXX/tapl Arvind advocate to always use #rsh for clients
	# -- we never use '$cc', so comment out next few lines
	#my $cc = make_conn($CLIENT, '#rsh');
	#$cc->timeout($CMDTIMEOUT);
	# We can use host object to pass to mount::mount() (but not a Connectrec).
	#my $ch = $cc->hostp();

        logDbg("mount_vol  Line: ".__LINE__."  Establish client connection\n".
	    	"  --  CLIENT: \"$CLIENT\"");
        # client connection
	#
	# Use either 'localhost' or $ch (HOSTREC) as arg to mount( client => ... )
	#
        my ($cc, $ch);
	if ($CLIENT eq 'localhost') {
		$ch = 'localhost';
	} else {
	        if (ref($CLIENT) =~ /connectrec/i) {
	            $cc = $CLIENT;
	        } else {
	            $cc = make_conn($CLIENT, '#rsh');
		    if (!$cc) {
		        logDbg("mount_vol  Line: ".__LINE__."   Could not establish client connection: $@");
		        die("mount_vol:  Could not establish client connection: $@");
		    }
	        }
	        $ch = $cc->hostp();
	        if (defined($CMDTIMEOUT)) {
	    	    $cc->timeout($CMDTIMEOUT);
	        }
	}

	if (!is_licensed($fc, "nfs")) {
	    logTrc("license nfs on filer");
	    license($fc, "nfs") or croak $@;
	}

	my ($mntpt, $vol, $vol_path, $dir);
	if (!$VOL_PATH) {
		croak("mount_vol:  One of 'vol_path', or 'vol' options must be defined\n".
                        "with name or path of volume to be mounted");
	}

	if ($VOL_PATH =~ m%^([^/]+)$%) {
	    # 'vol name' format
	    $vol = $1;
	    $vol_path = "/vol/$vol";
	    $vol_path .= "/$DIR" if ($DIR);
	    $dir = $DIR;
	} elsif ($VOL_PATH =~ m%^(/vol/([^/]+))$%) {
	    # 'vol path' format, no subdir
	    $vol_path = $1;
	    $vol = $2;
	    $vol_path .= "/$DIR" if ($DIR);
	    $dir = $DIR;
	} elsif ($VOL_PATH =~ m%^(/vol/([^/]+)/([^/]+))$%) {
	    # 'vol path' format, with subdir
	    $vol_path = $1;
	    $vol = $2;
	    $dir = $3;
	} else {
	    croak("new_mount_root_vol:  Invalid format for vol/vol_path option: \"$VOL_PATH\"");
	}

	my $vol_path_tail = $vol_path;
	$vol_path_tail =~ s%^/vol/%%;
	$mntpt = $MOUNTPOINT || "/t/$filer_ip/$vol_path_tail";

	logDbg("mount_vol   Line: ".__LINE__."\nVOL_PATH: \"$VOL_PATH\"\nvol_path: \"$vol_path\"\n".
	         "vol: \"$vol\"\nDIR: \"$DIR\"\ndir: \"$dir\"\n".
		 "vol_path_tail:\"$vol_path_tail\"\nmntpt: \"$mntpt\"");

	logDbg("mount_vol  Line: ".__LINE__."\n\nvol_path: \"$vol_path\"");
	my $qtpath;
	if ($dir) {
	    $qtpath = $vol_path;
	    if (! eval {qtreePresent($fc, $vol_path)} ) {
		$qtpath = "/vol/$vol";
	    }
	} else {
	    $qtpath = "/vol/$vol";
	}
	my $sec = eval { qtreeSecurity($fc, $qtpath) };
	if (($sec ne 'unix') && ($sec ne 'mixed')) {
	    my $rtn = eval { qtreeSecurity($fc, $qtpath, 'mixed') };
	    logWarn("can't change $qtpath access to 'mixed: $@") unless ($rtn);
	}
	logDbg("mount_vol  Line: ".__LINE__."\n\nvol_path: \"$vol_path\"\nqtree path: \"$qtpath\"\n".
	        "qtree sec: \"". eval { qtreeSecurity($fc, $qtpath) } ."\"");

	export_path($fc, "$vol_path")
	    or croak($@);

	logTrc("mount_vol: \"$vol\"  mountpoint: \"$mntpt\"");
	my %mount_subr_opts = ( filer_ip => $filer_ip, vol_path => $vol_path,
				     client=>$ch,
				     mountpoint=>$mntpt,
				     mountopts=>$MOUNTOPTS,
				     vers=>$main::MOUNT_VERS,
				     chown=>1,
				     chmod=>$CHMOD, );
	logDbg("mount_vol:   Calling mount::mount\n\n".Dumper(\%mount_subr_opts));
	my $mntpt_returned =
	              mount::mount($filer_ip, $vol_path,
				     client=>$ch,
				     mountpoint=>$mntpt,
				     mountopts=>$MOUNTOPTS,
				     vers=>$main::MOUNT_VERS,
				     chown=>1,
				     chmod=>$CHMOD,
			   );

	logDbg("mount_vol \"$vol_path\" returned error: $@") if ($@ || !$mntpt_returned);
	croak("mount_vol \"$vol_path\" returned error: $@") if ($@ || !$mntpt_returned);
	logDbg("mount_vol:   Returned from mount::mount");
	logTrc("mount_vol returning $mntpt_returned");
	return ($mntpt_returned);
    };
}


## @name ::umount_vol
## @summary unmount volume. returns 1 on success.
## @arg local_mnt_pt - what to unmount
## @arg %args. Optional. If not provided, values will be those of NATE command
##             line params.
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##	NOT_MOUNTED_OKAY:	If '1', then don't complain if the volume is not
##				mounted when this function is called. If '0', then
##				issue a warning if the volume is not currently mounted.
##				and return (0).
##				Default is '1', unless NOT_MOUNTED_OKAY is set.
##				-- return (1) silently if the volume is not mounted.
##
##	NOT_MOUNTED_TEST:	Mutually exclusive with NOT_MOUNTED_OKAY. If '1' and if volume
##				is not mounted at start of test, return (0). Unlike NOT_MOUNTED_OKAY,
##				no warning is issued in this case. Default is '0'.
##
sub umount_vol($;%)
{
    my ($local_mnt_pt, %opts) = @_;
    my $CLIENT 		= $opts{client} || $main::CLIENT || 'localhost';
    my $FORCE		= defined($opts{force}) ? $opts{force} : 1;
    my $NOT_MOUNTED_TEST  = defined($opts{not_mounted_test}) ? $opts{not_mounted_test} : 0;
    my $NOT_MOUNTED_OKAY  = defined($opts{not_mounted_okay}) ? $opts{not_mounted_okay} : !$NOT_MOUNTED_TEST;

    return eval {
	umount($local_mnt_pt,client=>$CLIENT, force=>$FORCE);
	my $msg = $@;
	if (!$msg) {
		return 1;
	} elsif ($msg =~ /not mounted|not in mnt/) {
		if ($NOT_MOUNTED_OKAY) {
			return 1;
		} elsif ($NOT_MOUNTED_TEST) {
			return 0;
		} else {
			logWarn("umount_vol: $local_mnt_pt is not mounted");
			return 0;
		}
	} else {
		croak("umount_vol: $@");
	}
    };
}





















## @name ::mount_vol
##
## @summary Mounts the volume on localhost (or CLIENT), when 'dir' option is omittted.
##	    If 'dir' option is given, then mounts the given directory for the given volume.
##          Returns: mount point
##
## @arg %opts. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER
##		May also be either a Hostrec or Connectrec object.
##
##         client - hostspec of client. defaults to $main::CLIENT
##		May also be either a Hostrec or Connectrec object.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         vol          - Required. Name of the filer volume to mount, or else path (on filer)
##                         to the volume (or volume/dir) to be mounted. Defaults to $main::FILER_VOL.
##                        Examples:
##                                     vol1   /vol/vol1    /vol/vol1/qtree1
##
##         vol_path, filer_vol  -- 'alias' options for the 'vol' option.
##
##	   dir - name of directory which is a direct decsendent of the parent volume.
##               If given, this directory will be mounted rather than the entire volume.
##		 default is undef, so entire volume is mounted. The resulting 'vol_path'
##	         will then be the concatenation:  "/" . "vol" . <VOLUME> . "/" <DIR>
##
##         filer_vol_mount - filer mount target. defaults to
##                           $main::FILER_VOL_MOUNT
##
##         mountpoint - local mount point. defaults to $main::MOUNTPOINT
##
##         timeout   - Timeout in secs. for mount to complete. Defaults to 300 (5 min).
##
##
sub mount_vol(%)
{
    my (%opts) = @_;
    my %raw_opts = %opts;

    $raw_opts{filer}         = ref($raw_opts{filer})         if (ref($raw_opts{filer})); 
    $raw_opts{filer_conn}    = ref($raw_opts{filer_conn})    if (ref($raw_opts{filer_conn})); 
    $raw_opts{client}        = ref($raw_opts{client})        if (ref($raw_opts{client})); 
    logDbg("Enter mount_vol   Line: ".__LINE__."    Opts:\n\n".Dumper(\%raw_opts));

    my $FILER	 	= $opts{fc}  || $opts{fh}  ||  $opts{filer} ||
			      $main::FILER;  # Can be HostSpec, Hostrec, or Connectrec.
    my $FILER_CONN 	= $opts{filer_conn}    || $main::FILER_CONN;
    my $VOL_PATH	= $opts{vol_path}      || $opts{vol} 
                               || $opts{filer_vol}  ||  $main::FILER_VOL ;
    my $MOUNTPOINT 	= $opts{mountpoint};
    my $CLIENT 		= $opts{client}        || $main::CLIENT || 'localhost';
    my $FILER_IP 	= $opts{filer_ip}      || $main::FILER_IP;
    my $MOUNTOPTS 	= $opts{mountopts}     || $main::MOUNTOPTS
                               || 'actimeo=0,noac,hard,intr,noquota';

    my $TIMEOUT         = $opts{timeout}       || $main::TIMEOUT;
    my $CMDTIMEOUT 	= $opts{cmdtimeout}    || $main::CMDTIMEOUT;
    my $CHMOD 		= defined($opts{chmod}) ? $opts{CHMOD} : '0777';
    my $DIR             = $opts{dir};


    return eval {
	my %opts_copy = (%opts);
	if ( ref($opts_copy{filer}) ) {
		$opts_copy{filer} = sprintf("%s", $opts_copy{filer});
	}
	if ( ref($opts_copy{client}) ) {
		$opts_copy{client} = sprintf("%s", $opts_copy{client});
	}
	logDbg("mount_vol  Line: ".__LINE__."\n\nOpts:\n\n".Dumper(\%opts_copy));

	if ($opts{filer_conn} && ref($opts{filer_conn}) =~ /connectrec/i) {
		my $logmsg = "mount_vol   Line: ".__LINE__.
		             "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n".
		             "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n".
		             "Will set 'filer' option to this object, and delete 'filer_conn' option";
#		logDbg($logmsg);
		$FILER = $opts{filer_conn};
		$opts{filer} = $opts{filer_conn};
		$FILER_CONN = $main::FILER_CONN;
		$opts{filer_conn} = $main::FILER_CONN;
	}

        logDbg("mount_vol  Line: ".__LINE__."  Establish filer connection\n".
		 "  --  FILER: \"$FILER\"");
        # filer connection
        my ($fc, $fh);
        my $conn_timeout = $TIMEOUT;
        if (ref($FILER) =~ /connectrec/i) {
            $fc = $FILER;
        } else {
            $fc = make_conn($FILER, $FILER_CONN);
	    if (!$fc) {
	        logDbg("mount_vol  Line: ".__LINE__."   Could not establish filer connection: $@");
	        die("mount_vol:  Could not establish filer connection: $@");
	    }
        }
        $fh = $fc->hostp();
	my $hostname = $fh->hostname();
	my $filer_ip = $FILER_IP || $fh->default_ip();
        if (defined($CMDTIMEOUT)) {
    	    $fc->timeout($CMDTIMEOUT);
        }
	# client connection/host
	# XXX/tapl Arvind advocate to always use #rsh for clients
	# -- we never use '$cc', so comment out next few lines
	#my $cc = make_conn($CLIENT, '#rsh');
	#$cc->timeout($CMDTIMEOUT);
	# We can use host object to pass to mount::mount() (but not a Connectrec).
	#my $ch = $cc->hostp();

        logDbg("mount_vol  Line: ".__LINE__."  Establish client connection\n".
	    	"  --  CLIENT: \"$CLIENT\"");
        # client connection
	#
	# Use either 'localhost' or $ch (HOSTREC) as arg to mount( client => ... )
	#
        my ($cc, $ch);
	if ($CLIENT eq 'localhost') {
		$ch = 'localhost';
	} else {
	        if (ref($CLIENT) =~ /connectrec/i) {
	            $cc = $CLIENT;
	        } else {
	            $cc = make_conn($CLIENT, '#rsh');
		    if (!$cc) {
		        logDbg("mount_vol  Line: ".__LINE__."   Could not establish client connection: $@");
		        die("mount_vol:  Could not establish client connection: $@");
		    }
	        }
	        $ch = $cc->hostp();
	        if (defined($CMDTIMEOUT)) {
	    	    $cc->timeout($CMDTIMEOUT);
	        }
	}

	if (!is_licensed($fc, "nfs")) {
	    logTrc("license nfs on filer");
	    license($fc, "nfs") or croak $@;
	}

	my ($mntpt, $vol, $vol_path, $dir);
	if (!$VOL_PATH) {
		croak("mount_vol:  One of 'vol_path', or 'vol' options must be defined\n".
                        "with name or path of volume to be mounted");
	}

	if ($VOL_PATH =~ m%^([^/]+)$%) {
	    # 'vol name' format
	    $vol = $1;
	    $vol_path = "/vol/$vol";
	    $vol_path .= "/$DIR" if ($DIR);
	    $dir = $DIR;
	} elsif ($VOL_PATH =~ m%^(/vol/([^/]+))$%) {
	    # 'vol path' format, no subdir
	    $vol_path = $1;
	    $vol = $2;
	    $vol_path .= "/$DIR" if ($DIR);
	    $dir = $DIR;
	} elsif ($VOL_PATH =~ m%^(/vol/([^/]+)/([^/]+))$%) {
	    # 'vol path' format, with subdir
	    $vol_path = $1;
	    $vol = $2;
	    $dir = $3;
	} else {
	    croak("new_mount_root_vol:  Invalid format for vol/vol_path option: \"$VOL_PATH\"");
	}

	my $vol_path_tail = $vol_path;
	$vol_path_tail =~ s%^/vol/%%;
	$mntpt = $MOUNTPOINT || "/t/$filer_ip/$vol_path_tail";

	logDbg("mount_vol   Line: ".__LINE__."\nVOL_PATH: \"$VOL_PATH\"\nvol_path: \"$vol_path\"\n".
	         "vol: \"$vol\"\nDIR: \"$DIR\"\ndir: \"$dir\"\n".
		 "vol_path_tail:\"$vol_path_tail\"\nmntpt: \"$mntpt\"");

	logDbg("mount_vol  Line: ".__LINE__."\n\nvol_path: \"$vol_path\"");
	my $qtpath;
	if ($dir) {
	    $qtpath = $vol_path;
	    if (! eval {qtreePresent($fc, $vol_path)} ) {
		$qtpath = "/vol/$vol";
	    }
	} else {
	    $qtpath = "/vol/$vol";
	}
	my $sec = eval { qtreeSecurity($fc, $qtpath) };
	if (($sec ne 'unix') && ($sec ne 'mixed')) {
	    my $rtn = eval { qtreeSecurity($fc, $qtpath, 'mixed') };
	    logWarn("can't change $qtpath access to 'mixed: $@") unless ($rtn);
	}
	logDbg("mount_vol  Line: ".__LINE__."\n\nvol_path: \"$vol_path\"\nqtree path: \"$qtpath\"\n".
	        "qtree sec: \"". eval { qtreeSecurity($fc, $qtpath) } ."\"");

	export_path($fc, "$vol_path")
	    or croak($@);

	logTrc("mount_vol: \"$vol\"  mountpoint: \"$mntpt\"");
	my %mount_subr_opts = ( filer_ip => $filer_ip, vol_path => $vol_path,
				     client=>$ch,
				     mountpoint=>$mntpt,
				     mountopts=>$MOUNTOPTS,
				     vers=>$main::MOUNT_VERS,
				     chown=>1,
				     chmod=>$CHMOD, );
	logDbg("mount_vol:   Calling mount::mount\n\n".Dumper(\%mount_subr_opts));
	my $mntpt_returned =
	              mount::mount($filer_ip, $vol_path,
				     client=>$ch,
				     mountpoint=>$mntpt,
				     mountopts=>$MOUNTOPTS,
				     vers=>$main::MOUNT_VERS,
				     chown=>1,
				     chmod=>$CHMOD,
			   );

	logDbg("mount_vol \"$vol_path\" returned error: $@") if ($@ || !$mntpt_returned);
	croak("mount_vol \"$vol_path\" returned error: $@") if ($@ || !$mntpt_returned);
	logDbg("mount_vol:   Returned from mount::mount");
	logTrc("mount_vol returning $mntpt_returned");
	return ($mntpt_returned);
    };
}


## @name ::umount_vol
## @summary unmount volume. returns 1 on success.
## @arg local_mnt_pt - what to unmount
## @arg %args. Optional. If not provided, values will be those of NATE command
##             line params.
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##	NOT_MOUNTED_OKAY:	If '1', then don't complain if the volume is not
##				mounted when this function is called. If '0', then
##				issue a warning if the volume is not currently mounted.
##				and return (0).
##				Default is '1', unless NOT_MOUNTED_OKAY is set.
##				-- return (1) silently if the volume is not mounted.
##
##	NOT_MOUNTED_TEST:	Mutually exclusive with NOT_MOUNTED_OKAY. If '1' and if volume
##				is not mounted at start of test, return (0). Unlike NOT_MOUNTED_OKAY,
##				no warning is issued in this case. Default is '0'.
##
sub umount_vol($;%)
{
    my ($local_mnt_pt, %opts) = @_;
    my $CLIENT 		= $opts{client} || $main::CLIENT || 'localhost';
    my $FORCE		= defined($opts{force}) ? $opts{force} : 1;
    my $NOT_MOUNTED_TEST  = defined($opts{not_mounted_test}) ? $opts{not_mounted_test} : 0;
    my $NOT_MOUNTED_OKAY  = defined($opts{not_mounted_okay}) ? $opts{not_mounted_okay} : !$NOT_MOUNTED_TEST;

    return eval {
	umount($local_mnt_pt,client=>$CLIENT, force=>$FORCE);
	my $msg = $@;
	if (!$msg) {
		return 1;
	} elsif ($msg =~ /not mounted|not in mnt/) {
		if ($NOT_MOUNTED_OKAY) {
			return 1;
		} elsif ($NOT_MOUNTED_TEST) {
			return 0;
		} else {
			logWarn("umount_vol: $local_mnt_pt is not mounted");
			return 0;
		}
	} else {
		croak("umount_vol: $@");
	}
    };
}



#= write_file()
#=
#= Replacement for "mkfile", both "filer CLI" and "client/NFS based"
#=
#=  All args are 'keyword' args
#=
#=  Returns: status: 1 => success, 0 => failed.
#=
#=  This function strives to be "backwards compatible" with several other functions which it is
#=  intended to replace. This has resulted in a long options list, and several alternative "styles"
#=  of use. Also, there are a few "aliases" for option keywords, such as "path"/"file", and
#=  "content"/"contents".
#=
#=  The library fcn. FileManip::write_file is used to do all 'client-based' (NFS or CIFS) write
#=  operations. Only 'filer-based' writes are done directly by this fcn.
#=
#= :filer               Optional. Filer HOSTSPEC, HOSTREC, or CONNECTREC.
#=
#= :client              Optional. NFS Client HOSTSPEC, HOSTREC, or CONNECTREC (or string 'localhost').
#=
#= :cifs_client         Optional. CIFS Client HOSTSPEC, HOSTREC, or CONNECTREC
#=
#= :agent		Optional. One of "localhost", "client", "cifs_client", or "filer".
#=			Allows more than one of filer, client, and/or cifs_client options to be specified.
#=
#=  Either a 'path' or 'file' argument must be specified
#= :path                Required. Path to file to be created/wriiten
#= :file                Synonnym for 'path'
#=
#= :timeout		Optional.
#=
#= :size		Required. Size of files. e.g. "10m", "100k" "50g"
#=
#= :numfiles		Optionsal. If given, then number of files to create. Default is 1.
#=			Files will be named <PATH>_1, <PATH>_2, <PATH>_3
#=
#= :dont_overwrite	Optional. If '1', then new files are created, but existing files are not
#=			overwritten. Default is '1'.
#=
#= :content	     - optional: a named parameter with the initial contents
#=  			of the file. Exclusive with size, and overrides.
#=
#= :contents	     - A synonym for 'content'
#=
#= :debug_verbose    - Optional.  If '1', then more debug/diagnostic messages are given.
#=			  Default is '0'.
#=			      
#=			      
#= At most one of "filer" or "client" or "cifs_client" should be specified, unless the "agent"
#= option is also speciified:
#=
#=	Option		Agent Option		Effect
#=	--------------------------------------------------------------------------------------------------
#=	filer		"filer"		==>	Filer CLI "mkfile" will be used to create/write to file(s)
#=	client		"client"	==>	NFS client will be used to create/write to file(s)
#=	cifs_ client	"cifs_cleint"	==>	CIFS client will be used to create/write to file(s)
#=	(neither)	"localhost"	==>	Local NFS client will be used to create/write to file(s)
#=
#= If "filer" is specified, then the "path" should be of form "/vol/<VOLNAME>/..." (filer relative).
#= Otherwise, "path" should be for the mounted path on the specified client (or local client).
#=
#= As an alternative to the "path" option, a "vol_path" option may be specified instead. For a "filer"
#= type invocation, this has the same meaning as the "path" option. Howver, with any other type of
#= invocation, a "mountpoint" option must also be given in addition to the "vol_path" option. The
#= "client relative" path is then determined by stripping the "/vol/<VOL_NAME>" prefix from the vol_path,
#= and appending the result to the mountpoint. Not that for cifs clients, the mountpoint is a mapped drive,
#= e.g. "D:". The motivation for this is to easily switch between "filer" and non-filer agents, without having
#= to change the form of the path option to adopt to/from "filer relative" vs. "client relative" paths.
#= The "mountpoint" option is ignored for "filer" type invocations.
#=
#=	Option		Agent Option		Path options allowed
#=	--------------------------------------------------------------------------------------------------
#=	filer		"filer"			path || vol_path (mountpoint allowed but is ignored)
#=	client		"client"		path || (vol_path && mountpoint)
#=	cifs_client	"cifs_client"		path || (vol_path && mountpoint)
#=	(none)		"localhost"		path || (vol_path && mountpoint)
#=
#=
#=
#=         "Filer Only" Args -- The following args are ONLY valid if 'filer' option is specified.
#=
#= :step_size 	   - optioneal. The step-size-in-blocks specifies that mkfile should create files
#= 			with holes in them. mkfile writes one block of zeros for each
#=			step-size-in-blocks blocks. For example, if you use a step-size-in-blocks
#=			of 3, the resulting file will have one block of zeros followed by two empty
#=			(unallocated) blocks, followed by a block of zeros and 2 more empty blocks,
#=			and so on. The default step-size-in-blocks is 1. The step_size option can
#=			only be used in "filer-type" invocations, and in conjunction with the
#=			"numfiles" option.
#=
#=
#=
#=         "Client Only" Args -- The following args are NOT valid if 'filer' option is specified.
#=
#= :buffer_size	     - Optional. Size of data buffer to be used for writing to the file.
#=			Large files need to be written in multiple operations, due to
#=			limitatitations on max string length imposed by memory constraints.
#=			If 'buffer_size' is given, and file size is larger than buffer_size,
#=			then multiple write operations will be used to write to the file,
#=			and all operations will be in increments of "buffer_size" or less.
#=
#=			The 'write_data_fcn' option may be used in conjunction with the
#=			'buffer_size' option, to allow for more flexible data string
#=			creation. If used with 'contents' option, and if buffer_size is
#=			less than size of "contents" string, then multiple write operations
#=			will be used to write the file.
#=
#=			Defualt is 10Mbyte.
#=
#=
#= :file_list	     - optional. If given, an ARRAYREF. The paths of all files written will be
#=                       appended to this list. The list should be initialized (e.g to "empty list") 
#=                       prior to calling this fcn.
#=
#= :uid		     - Optional. If given, this UID will be assigned to the file(s).
#=
#= :gid		     - Optional. If given, this GID will be assigned to the file(s).
#=
## :mode             - optional: a named parameter with the initial mode bits.
#=		       If cifs is being used, the mode bits - +R | -R
#=		       (R - readonly)
#=
#= :fstat	     - Optional.
#=			  	If '0', then scalar status is returned --
#=					1 -> success, 0 => failure.
#=			  	If '1', then list of attrs is returned (@attrs)
#=			  	Default is '0'.
#=
#= :verify	     - Optional.  If '1', then verify existance and size of files written.
#=			  Default is '0'.
#=
#= :catch_errors     - Optionsal.  If '1', then some errors (those that occur during the 'write' operation)
#=				         are caught, and '$@' is set with an error message.
#=					 Errors such as invalid parameter values are not caught.
#=				   Otherwise, the routine will 'die' on all errors.
#=				   Default: '0' (die on error).
#=			      
#=			      
#= :write_data_fcn   - optional: Valid only if 'buffer_size' option is also given.
#=			If given, a "function ref" for a function which takes
#=			the following keyword options/args, and returns a "data string"
#=			which will be written to the file. The write_data_fcn is just
#=			responsible for creating the "data string" to be written, it
#=			does not itself perform any I/O.
#=
#=			If no "write_data_fcn" is specified, then a simple "default" fcn.
#=			is used. The behavior of the "default" fcn is not specified, it
#=			may simply return a block of identical chars.
#=
#=			"content", "contents", and "write_data_fcn" are all mutually
#=			exclusive options.
#=
#=			The following options will be passed to the fcn, it can ignore any except 
#=			"write_len". The other options are useful for passing 'state' to the
#=			write_data_fcn, to enable it to return different values of "data string"
#=			when called multiple times (but this is not required).
#=
#=				write_len:	Length of data to write, in bytes. The "write_data_fcn"
#=					        is required to honor this option.
#=
#=				'Convenience Options' -- may be ignored.
#=
#=				buffer_size:	Buffer size, in bytes. For all write
#=					        operations except the last one, 'write_len'
#=					        will equal 'buffer_size'. For the last write,
#=						'write_len' may be less than the buffer size
#=					        (file_length mod buffer_size).
#=
#=				file_size:	Size (in bytes) of file to be created.
#=
#=				file_path:	Absoluete path of file to be created.
#=
#=				count:		Number of write operations prior to this one
#=						(i.e. zero-based 'invocation' count)
#=						    first call to 'write_data_fcn'   ==> 0,
#=						    second call "     "   "    "     ==> 1, etc.
#=
#=				end_count:	The count value for the last write operation.
#=
#=				offset:		The starting address of the file into which the
#=						data is to be written for this write operation. (bytes).
#=
#=				start_offset:	The offset into the file for the first write operation.
#=
#=				special_volume_options:	 See 'special_volume_options' arg for 'write_file'
#=						         below.
#=
#=				io_size:	The 'prefered I/O size' (block size) of the
#=						filesystem hosting the file(s) to be written.
#=
#=				write_data_args  Only applies if 'write_data_args' option was provided for
#=						 'write_file'.  User-defined data. Is commonly a HASHREF
#=						 for user-defined info to be used by the write_data_fcn,
#=						 byt may be any scalar value chosen by the user.
#=
#=				scramble	 (see below)
#=
#=
#= :write_data_args  - Optional. Ignored unless 'write_data_fcn' option is also given. If supplied, this
#=		       value is passed to the 'write_data_fcn' via the keyword 'write_data_args'.
#=			This arg is not used by the 'write_file' fcn. itself. It may be any scalar
#=			value, but would typically be a HASHREF containing various args to be
#=			used by the 'write_data_fcn'.
#=
#=
#= :scramble         - optional:  
#=				1) For user-defined write_data_fcn, the behavior is determined by that fcn.
#=				   -- the fcn may simply ignore this arg.
#=
#=				2. For 'default_write_data_fcn', the behavior is the following:
#=			 	    'Approximates'  behavior of 'SCRAMBLE' parameter of DataManip::populate"
#=				    function.   See: http://wikid.netapp.com/w/DataManip for more info on
#=				    "populate" use of "SCRAMBLE". The main difference is that for 'scramble'
#=				    values > 1, 'populate' treats this value as an absolute number, whereas
#=				    here the number is treated as a percent value.
#=
#=				    > If scramble is '0.0' (default):
#=					Then every block written is identical. (100% de-dupable and compressible)
#=
#=				    > If scramble is in range: 0.0 - 1.0:
#=					Then 'scramble' represents the fraction of blocks that will be unique
#=					(diffferent from every other block of the given file). The fraction of
#=					"common" (de-dupable) blocks in the file is (1 - scramble)
#=						e.g.    scramble=0.15  ==>  15% of blocks will be unique
#=					-- Compressible, de-dup resistance depends on 'scramble' value
#=
#=				    > If scramble is greater than 1.0:
#=					Then the 'int' part of scramble gives the percentage of random bytes
#=					in each block.
#=						e.g.    scramble=30    ==>  30% of bytes in each block will be random
#=					-- De-dup resistant, compressiblity depends on 'scramble' value.
#=
#=			      
#= :special_volume_options  - Optional.  If given, this is a HASHREF with zero or more
#=				"special volume options" specified. If not supplied,
#=				then the corresponding global parameter values are used.
#=				That is:
#=					$default_special_volume_options = {
#=						compression_enable    => $::COMPRESSION_ENABLE,
#=						asis_enable           => $::ASIS_ENABLE,
#=						encryption_enable     => $::ENCRYPTION_ENABLE,
#=						large_aggr	      => $::LARGE_AGGR,
#=					};
#=				If a 'special_volume_options' HASHREF arg is supplied, then the
#=				'default_special_volume_options' above is used to supply any
#=				fields not set in the HASHREF.
#=
#=				These options are not used by the 'write_fcn' function itself,
#=				but are passed on to the 'write_data_fcn'.
#=
sub write_file(%) {

       my (%args) = validate (
                            @_,
                            {
				#=========  'Common' (filer or client) options:  ==========
                                filer            => {type=>SCALAR|HASHREF|OBJECT|UNDEF,     optional=>1},
                                client           => {type=>SCALAR|HASHREF|OBJECT|UNDEF,     optional=>1},
                                cifs_client      => {type=>SCALAR|HASHREF|OBJECT|UNDEF,     optional=>1},
                                agent            => {type=>SCALAR,            optional=>1},
                                path             => {type=>SCALAR,            optional=>1},
                                file             => {type=>SCALAR,            optional=>1},
                                vol_path         => {type=>SCALAR|UNDEF,      optional=>1},
                                mountpoint       => {type=>SCALAR|UNDEF,      optional=>1},
				#
				# Exactly 1 of 'size', 'contents', or 'content' must be given
                                size             => {type=>SCALAR,	      optional=>1},
                                content          => {type=>SCALAR,            optional=>1},
                                contents         => {type=>SCALAR,            optional=>1},
				#
                                numfiles         => {type=>SCALAR|UNDEF,      optional=>1},
                                dont_overwrite   => {type=>SCALAR|UNDEF,      optional=>1,   default=>0},
                                timeout          => {type=>SCALAR|UNDEF,      optional=>1},
                                debug_verbose    => {type=>BOOLEAN,           optional=>1,   default=>0},

				#=========  'Client Only' options:  ==========
                                catch_errors     => {type=>BOOLEAN,           optional=>1,   default=>0},
                                mode             => {type=>SCALAR,            optional=>1},
                                uid              => {type=>SCALAR,            optional=>1},
                                gid              => {type=>SCALAR,            optional=>1},
                                offset           => {type=>SCALAR,            optional=>1},
                                buffer_size      => {type=>SCALAR,            optional=>1,   default=>"2m"},
                                file_list        => {type=>ARRAYREF,          optional=>1},
                                user             => {type=>SCALAR,            optional=>1},
                                fstat            => {type=>BOOLEAN,           optional=>1,   default=>0},
                                verify           => {type=>BOOLEAN,           optional=>1,   default=>0},
                                scramble         => {type=>SCALAR,            optional=>1,   default=>undef},
				#
				# 'write_data_args' if given, can be any type. Used with 'write_data_fcn'.
                                write_data_fcn   => {type=>CODEREF,           optional=>1},
                                write_data_args  => { optional=>1 },
				#
				#
				#  'random' option is deprecated -- use 'scramble' option instead
                                random           => {type=>BOOLEAN,          optional=>1},  # Deprecated
				#
                                special_volume_options
						 => {type=>HASHREF,          optional=>1,   default=>{}},

				#=========  'Filer Only' options:  ==========
                                step_size       => {type=>SCALAR,          optional=>1},
                            }
                     );
	#===========================================================================================
	my $default_write_data_fcn =
		sub(%) {
			my (%args)      	= @_;

			# -------------- 'generic' write_data_fcn args -------------
			my $write_len   	= $args{write_len};	# Length of data to write
			#
			#    Contract:  write_data_fcn must return string/buffer of
			#		length 'write_len' bytes
			#	        'write_len' will always be <= 'buffer_size'
			#
			#    #  If 'file_size' is greater than 'buffer_size', then this fcn. will
			#       will be invoked multiple times.
			#
			#    *  'write_len' will be equal to 'buffer_size' for all invocations
			#       except (possibly) the last, when 'buffer_size' will be equal to
			#       'file_size' mod 'buffer_size.
			#
			#    #  Total length wriiten prior to this invocation is equal to
			#       'count' times 'buffer_size'
			#
			#    *  Following args are for 'convenience' of user-defined 'write_data_fcn',
			#       they may be ignored.
			#
			my $buffer_size      	= $args{buffer_size};	# user specified 'buffer_size'
			my $file_size      	= $args{file_size};	# user specified 'file_size'
			my $offset      	= $args{offset};	# offset into file (bytes)
			my $start_offset      	= $args{start_offset};	# offset into file of first invocation
			my $count	      	= $args{count};		# "invocation count" (zero based) for this invocation
			my $end_count	      	= $args{end_count};	# "invocation count" of final invocation.
			my $special_volume_options
					        = $args{special_volume_options};  # See 'special_volume_options' param, above.
										  #   Currently: asis_enable, compression_enable,
										  #   encryption_enable, large_aggr.
			my $io_size     	= $args{io_size};	# 'prefered I/O size'
			my $scramble     	= $args{scramble};	# (see info for 'scramble' param above)
			my $file_path    	= $args{file_path};	# Path of file to be written.

			# ------- 'custom' (user-defined) write_data_fcn args -------
			my $write_data_args  	= $args{write_data_args};
			my $contents 		= $write_data_args->{contents};


			#  Most of following code is for implementing behavior related to 'scramble'
			#  parameter.
			my ($msg, $msg1);
			my $num_key_char     = 2;
			#  'orig_blk_size' gives size of (filer) data block
			#      (except for small values of 'write_len')
			#  'curr_blk_size' equals 'orig_blk_size', except possibly for last
			#      pass thru 'for' loop starting around line 4205.
			my $orig_blk_size    = (($io_size < $write_len) ? $io_size : $write_len);
			my $curr_blk_size    = $orig_blk_size;
			my $blk_buff_len;
			my $blk_buff = '';
			my $rtn_buff = '';
			my $RAND = undef;
			my $common_key;
			my $n_rand;
			my $n_width;
			my $key;
			my $n_read;
			my $saved_data_args;
			my $saved_data_args_copy;
			if ($count == 0) {
				delete($write_data_args->{saved_data_args});
			}

			$scramble = 0 if ($contents);
			if (!defined($scramble)) {
				if ($special_volume_options->{compression_enable}) {
					$scramble = 20;
			    		logDbg("default_write_data_fcn  Line: ".__LINE__.
					          "\n\n'compression_enable' special_volume_option is set --\n".
						  "   Setting (undefined) 'scramble' value to \"$scramble\"");
				} elsif ($special_volume_options->{asis_enable}) {
					$scramble = 0.4;
			    		logDbg("default_write_data_fcn  Line: ".__LINE__.
					          "\n\n'asis_enable' special_volume_option is set --\n".
						  "   Setting (undefined) 'scramble' value to \"$scramble\"");
				} elsif ($args{random}) {
					$scramble = 30;
			    		logDbg("default_write_data_fcn  Line: ".__LINE__.
					          "\n\n'random' option is set --\n".
						  "   Setting (undefined) 'scramble' value to \"$scramble\"");
				} else {
					$scramble = 0;
				}
			} elsif (!$scramble) {
				$scramble = 0;
			} elsif (int($scramble) > 1.0) {
				$scramble = int($scramble);
				$scramble = 100 if ($scramble > 100);
			} elsif ($scramble < 0.0) {
				$scramble = 0;
			}

			my $args_copy = { %args };
			$args_copy->{write_data_args} = "{  write_data_args ... }";
			logDbg("default_write_data_fcn  Line: ".__LINE__."   Args:\n\n".Dumper($args_copy).
			           "\nScramble (final value):\t$scramble");

			#  I.  if 'contents' is given, then just return the appropriate substring of 'contents'....
			if ($contents) {
				my $contents_offset;
				my $contents_len;
				if ($write_data_args->{saved_data_args} && (Y$count > 0)) {
					#  Get saved values -- set during previous invocation
					$saved_data_args 	= $write_data_args->{saved_data_args};
					$contents_offset 	= $saved_data_args->{contents_offset};
					$contents_len 		= $saved_data_args->{contents_len};
					$contents 		= $saved_data_args->{contents};
				} else {
					$contents_offset = 0;
					$saved_data_args->{contents} = $contents;
					$saved_data_args->{contents_offset} 	= $contents_offset;
					$contents_len = length($contents);
					$saved_data_args->{contents_len} 	= $contents_len;
				}
				if ($contents_offset == $contents_len) {
					return("");
				} elsif ($contents_offset > $contents_len) {
					die("default_write_data_fcn  Line: ".__LINE__.
				                  "\n\nContents_offset=$contents_offset\n".
						  "    is greater than \nContents_len=$contents_len");
				} elsif (($write_len + $contents_offset) > $contents_len) {
					$write_len = $contents_len - $contents_offset;
				}
				$rtn_buff = substr($contents, $contents_offset, $write_len);
				$saved_data_args->{contents_offset} += $write_len;
				return( $rtn_buff );
			}

			#  II.  if 'contents' is not given, then we need to generate appropriate data,
			#        including "scrambling" when appropriate.

			#  First Initialize local args.
			#  Open 'random' filehandle 'RAND' for use later.
			sysopen($RAND, "/dev/urandom", O_RDONLY)
			    or die("default_write_data_fcn  Line: ".__LINE__.
				     "\nCould not open '/dev/urandom' for reading: $!");

			unless (defined($RAND)) {
				die("default_write_data_fcn  Line: ".__LINE__.
				       "\n\n'/dev/urandom' filehandle 'RAND' is undefined");
			}
			unless (stat(*$RAND)) {
				die("default_write_data_fcn  Line: ".__LINE__.
				       "\n\n'Stat' failed on '/dev/urandom' filehandle 'RAND'");
			}

			if ($write_data_args->{saved_data_args} && ($count > 0)) {
				#  Get saved values -- set during previous invocation
				$saved_data_args = $write_data_args->{saved_data_args};
				$common_key	= $saved_data_args->{common_key};
				$n_rand		= $saved_data_args->{n_rand};
				$n_width	= $saved_data_args->{n_width};
				$blk_buff	= $saved_data_args->{blk_buff};
			} else {
				#  Compute values and store these in 'saved_data_args' for
				#  subsequent incocations.
				$saved_data_args = {};
				$write_data_args->{saved_data_args} = $saved_data_args;
				$saved_data_args->{orig_blk_size} = $orig_blk_size;
				$saved_data_args->{curr_blk_size} = $curr_blk_size;
				$saved_data_args->{file_path} = $file_path;

				#  Initialize remaining vars --
				#     a. Vars used for 'scramble' value 0 < x < 1
				#     b. Vars used for 'scramble' values x > 1
				if ($scramble && ($scramble > 0.0)  && ($scramble <= 1.0)) {
					#  Vars for scrable values  x < 1
					$n_read = sysread(*$RAND, ($common_key=''), $num_key_char);
					if ($n_read != $num_key_char) {
						die("default_write_data_fcn  Line: ".__LINE__.
						         "\n\nError reading from /dev/urandom\n".
						         "Expected to read $num_key_char bytes, ".
						         "actually read $n_read bytes:\n> $!\n");
					}
					$saved_data_args->{num_key_char} = $num_key_char;
					$saved_data_args->{common_key} = $common_key;
				} elsif ($scramble && ($scramble > 1.0)) {
					#  Vars for scrable values  x > 1
					$n_rand = ($curr_blk_size  * int($scramble)) / 100;
					if ( ($n_rand - int($n_rand) > .4) && ($n_rand < $curr_blk_size)) {
						$n_rand += 1;
					}
					$n_rand = int($n_rand);
					$saved_data_args->{n_rand} = $n_rand;
					$n_width = int(($curr_blk_size -1)/ $n_rand);
					$saved_data_args->{n_width} = $n_width;
				}

				#  Generate single 'random' block 'blk_buff' which will be used as
				#  starting point for all blocks to be written, with changes to
				#  individual blocks determined by value of 'scramble' parameter.
				$blk_buff = '';
				$rtn_buff = '';
				for (   ( my $rem_to_read  = $curr_blk_size,
				          my $total_read   = 0 );
				        ( $rem_to_read > 0 );
				        ( $rem_to_read -= $n_read,
				          $total_read += $n_read )
				    ) {
					my $tmp_buff = '';
					# $n_read = sysread(*$RAND, substr($blk_buff, $total_read), $rem_to_read);
					$n_read = sysread(*$RAND,$tmp_buff, $rem_to_read);
					unless (defined($n_read) && ($n_read > 0)) {
						$msg = "default_write_data_fcn  Line: ".__LINE__.
						         "\n\nError reading from /dev/urandom\n".
						         "Expected to read $curr_blk_size bytes, ".
						         "actually read $n_read bytes:\n> $!\n";
						logDbg($msg);
						die($msg);
					}
					$blk_buff .= $tmp_buff;
				}
				if (length($blk_buff) != $curr_blk_size) {
					die("default_write_data_fcn  Line: ".__LINE__.
					       "\n\nExpected  length(blk_buff)=".length($blk_buff)."\n".
					       "    to equal curr_ blk_size=$curr_blk_size\n");
				}
				#  Store copy of 'blk_buff' (data/string) in 'saved_data_args'
				$saved_data_args->{blk_buff}  = $blk_buff;
				#  -- Finished generating 'blk_buff'
			}

			#  -- Finished initializing args, or reading them from 'saved_data_args'

			$blk_buff_len			    = length($blk_buff);
			$saved_data_args->{blk_buff_len}    = $blk_buff_len;
			$saved_data_args_copy 		    = { %$saved_data_args };
			$saved_data_args_copy->{blk_buff}   = "< BLOCK BUFFER  ($blk_buff_len bytes) .... >";
			if ($args{debug_verbose}) {
				logDbg("default_write_data_fcn  Line: ".__LINE__.
				         "\n\nSaved Data Args:\n\n".Dumper($saved_data_args_copy));
				logDbg("default_write_data_fcn  Line: ".__LINE__.
				          "\n\nI/O Size:\t$io_size\nWrite Len:\t$write_len\n".
					  "Orig blk size:\t$orig_blk_size\nCurr blk size:\t$curr_blk_size\n".
				          "length(blk_buff):\t".length($blk_buff)."\n".
				          "N_rand:\t$n_rand\nN_width:\t$n_width\nNum key char:\t$num_key_char\n".
				          "Scramble:\t$scramble");
			}
			if (length($blk_buff) != $curr_blk_size) {
				die("default_write_data_fcn  Line: ".__LINE__.
				       "\n\nExpected  length(blk_buff)=".length($blk_buff)."\n".
				       "    to equal curr_ blk_size=$curr_blk_size\n");
			}

			#   Modify 'blk_buff' as determined by 'scramble' value -- three cases:
			#     a. 'scramble' value 0 < x < 1
			#     b. 'scramble' values x > 1
			#     c. 'scramble' value == 0.0
			#
			#   Loop for inserting 'random' bytes in blocks
			#   Handle data in 'curr_blk_size' chunks
			for (      ( my $rem_chars = $write_len,
				       my $blkcnt       = 0,
				       my $rand_blkcnt  = 0,
				       my $wrlen        = $curr_blk_size) ;
				   ( $wrlen > 0 ) ;
				   ( $rem_chars -= $wrlen,
				       $blkcnt++,
				       $wrlen = (($io_size < $rem_chars) ? $io_size : $rem_chars) )
			    ) {

			        $curr_blk_size = ($wrlen < $curr_blk_size) ? $wrlen : $curr_blk_size; 

				if ($args{debug_verbose}) {
					logDbg("default_write_data_fcn  Line: ".__LINE__.
					          "\n\nTop-of-loop\n\nBlk cnt:\t$blkcnt\nRem chars:\t$rem_chars\n".
					          "Write len\t$write_len\nOrig blk size:\t$orig_blk_size\n".
					          "Wrlen:\t$wrlen\nCurr blk size:\t$curr_blk_size\n".
						  "Rand blk cnt:\t$rand_blkcnt");
				}

				if ($curr_blk_size < $orig_blk_size) {
					#  Set 'curr_blk_size' for last pass thru loop
					my $old_blk_buff_len = length($blk_buff);
					# $blk_buff = substr($blk_buff, $curr_blk_size) = '';
					$blk_buff = substr($blk_buff, 0, $curr_blk_size);
					$blk_buff_len = length($blk_buff);
					if ($args{debug_verbose}) {
						logDbg("default_write_data_fcn  Line: ".__LINE__.
						          "\n\nCalling 'substr' on:\n".
						          "String:     blk_buff\n".
						          "Length:     $old_blk_buff_len\n".
							  "Offest:     $curr_blk_size (curr_blk_size)\n".
						          "Orig blk size: $orig_blk_size\n\n".
							  "New blk_buff len: $blk_buff_len");
					}
					# $saved_data_args->{blk_buff} = $blk_buff;
					$saved_data_args->{curr_blk_size} = $curr_blk_size;
				}

				if ($blk_buff_len != $curr_blk_size) {
					die("default_write_data_fcn  Line: ".__LINE__.
					       "\n\nExpected length(blk_buff)=$blk_buff_len\n".
					           "   to equal  curr_blk_size=$curr_blk_size\n");
				}
				if ($args{debug_verbose}) {
					logDbg("default_write_data_fcn  Line: ".__LINE__.
					         "\n\nBlk_cnt:\t$blkcnt\nRand_blkcnt:\t$rand_blkcnt\n".
					         "Rem_chars:\t$rem_chars\nWrlen:\t$wrlen\n");
				}

				if ($scramble && ($scramble > 0.0)  && ($scramble <= 1.0)) {
					#  a. 'scramble' value 0 < x < 1
					last unless ($wrlen > $num_key_char);

					# my $rand_val = rand();
					my $rand_val = (ord(getc(*$RAND)) / 255);
					if ($rand_val < $scramble) {
						$rand_blkcnt++;
						$n_read = sysread(*$RAND, ($key=''), $num_key_char);
						if ($n_read != $num_key_char) {
							$msg = "default_write_data_fcn  Line: ".__LINE__.
							         "\n\nError reading from /dev/urandom\n".
							         "Expected to read $num_key_char bytes, ".
							         "actually read $n_read bytes:\n> $!\n";
							logDbg($msg);
							die($msg);
						}
						if ($args{debug_verbose}) {
							logDbg("default_write_data_fcn  Line: ".__LINE__.
							        "\n'Random' block added...");
						}

					} else {
						$key = $common_key;
					}
					substr($blk_buff, 0, $num_key_char) = $key;

					if ($args{debug_verbose}) {
						logDbg("default_write_data_fcn  Line: ".__LINE__.
						        "\n\nScramble:\t$scramble\nBlk_cnt:\t$blkcnt\n".
						        "Rand_blkcnt: $rand_blkcnt\nRand Val:\t$rand_val\n".
						        "Rand ratio: ".sprintf("%.3f", ($rand_blkcnt / (1+$blkcnt))) );
					}

					#  End of case: 'scramble' value 0 < x < 1
				} elsif ($scramble && ($scramble > 1.0)) {
					#  b. 'scramble' values x > 1
					if ($args{debug_verbose}) {
						logDbg("default_write_data_fcn  Line: ".__LINE__."\n\n".
						           "Scramble:\t$scramble\nWrite Len:\t$write_len\n".
						           "I/O Size:\t$io_size\nCurr Block size:\t$curr_blk_size\n".
							   "Wrlen:\t$wrlen\nN_rand:\t$n_rand\nN_width:\t$n_width");
					}
					if ($curr_blk_size < $orig_blk_size) {
						$n_rand = ($curr_blk_size * int($scramble)) / 100;
						if ( ($n_rand - int($n_rand) > .4) && ($n_rand < $curr_blk_size)) {
							$n_rand += 1;
						}
						$n_rand = int($n_rand);
					}
					if ($n_rand >= 1) {
						$n_width = int(($curr_blk_size -1)/ $n_rand);
						if ($args{debug_verbose}) {
							logDbg("default_write_data_fcn  Line: ".__LINE__.
							          "\n\nN_rand:\t$n_rand\nN_width:\t$n_width\n");
						}
						for ( (my $ofst=0, my $rndcnt=0 );
						      ( $rndcnt < $n_rand && $ofst < $curr_blk_size );
					      	      ( $rndcnt++,  $ofst+=$n_width )
						    ) {
							if ($args{debug_verbose}) {
								# logDbg("default_write_data_fcn  Line: ".__LINE__.
								#         "\n\nN_rand:\t$n_rand\nRnd_cnt:\t$rndcnt\n".
								#         "Wrtlen:\tOffset:\t$ofst\nN_width:\t$n_width");
							}
						my $chr = getc(*$RAND);
						substr($blk_buff, $ofst, 1) = $chr;
						}
					}

				} else {
					#  c. 'scramble' value == 0.0
					#  End of case: 'scramble' values x > 1
					#  No aditional action needed for 'non-scrambled' case.
					#  -- just use 'blk_buff' as-is
				}

				$rtn_buff .= $blk_buff;
				if ($args{debug_verbose}) {
					logDbg("default_write_data_fcn  Line: ".__LINE__.
					         "\n\n  Blk cnt:\t$blkcnt\nLength(rtn_buff):\t".
					         length($rtn_buff)." bytes");
				}

			}  # for (...)

			#  Close 'RAND' filehandle before exiting, each invocation re-opens it.
			if ($count == $end_count) {
				close(*$RAND) if ($RAND);
			}
			logDbg("default_write_data_fcn  Line: ".__LINE__."\n\n Returning:  ".
			           length($rtn_buff)." bytes");
			return($rtn_buff);
	         };
	#===========================================================================================


        #=================================================================================================
        my $write_file_aux = sub() {
	    #
	    #       %args 
	    #	          offset, $buffer_size, file_list, fstat, rtn_fstat, special_volume_options
	    #	          contents, size_bytes, io_size, write_data_fcn, write_data_args,
	    #


            return eval {
	    	my ($argsref) = @_;
		my ($msg, $msg1, $rtn_status);

		my $write_file_subr_args    = $argsref->{write_file_subr_args};
		my $file_path               = $write_file_subr_args->{file};

		my $dont_overwrite          = $argsref->{dont_overwrite};
		my $size_bytes              = $argsref->{size_bytes};
		my $raw_offset              = $argsref->{offset} || 0;
		my $raw_buffer_size         = $argsref->{buffer_size},
		my $file_list               = $argsref->{file_list};
		my $contents                = $argsref->{contents};
		my $scramble                = $argsref->{scramble};
		my $io_size                 = $argsref->{io_size};
		my $write_data_fcn          = $argsref->{write_data_fcn};
		my $write_data_args  	    = $argsref->{write_data_args};
		my $rtn_fstat        	    = $argsref->{rtn_fstat};
		my $special_volume_options  = $argsref->{special_volume_options};

		# my $fstat         	    = $argsref->{fstat};
		my $fstat           	    = 1;
		my @attrs	     	    = ();
		my $truncate          	    = 0;


		my $write_file_subr_args_copy = { %$write_file_subr_args };
		delete($write_file_subr_args->{contents});
		my $size_bytes_aux = $size_bytes;
		if ($size_bytes_aux == 0) {
			logDbg("Enter write_file_aux  Line: ".__LINE__.
			         "\nZero 'size_bytes' given, will use '1' instead, and then truncate to zero");
			$size_bytes_aux = 1;
			$truncate = 1;
		}

		if (!$raw_buffer_size) {
			die("Utilities::write_file_aux:  Line: ".__LINE__.
			         "\n'buffer_size' must be defined, and greater than zero");
		}
		my $buffer_size = convert_to_bytes($raw_buffer_size);
		if (!$buffer_size || $buffer_size <= 0) {
			die("Utilities::write_file_aux  Line: ".__LINE__.
			        "\nInvalid 'buffer_size' value: \"$raw_buffer_size\"");
		}
		my $start_offset = convert_to_bytes($raw_offset);
		if ( !defined($start_offset) ||
		      ($start_offset ne '0' && !$start_offset) || 
		      ($start_offset < 0) ) {
			die("Utilities::write_file_aux  Line: ".__LINE__.
			        "\nInvalid 'offset' value: \"$raw_offset\"");
		}

		my $end_count        = int($size_bytes_aux / $buffer_size) - 1;
		my $rem_size         = $size_bytes_aux % $buffer_size;
		$end_count++         if ($rem_size);

		$msg1 =   "    File Path:        \"$file_path\"\n".
			  "    dont_overwrite:   ".($dont_overwrite ? 1 : 0)."\n".
			  "    File Size:        $size_bytes_aux\n\nOpts:\n".
			  "    (Start) Offset:   $raw_offset\n".
			  "    Buffer Size:      $raw_buffer_size\n".
			  "    End Count:        $end_count\n".
			  "    Rem Size:         $rem_size";

		if ($argsref->{debug_verbose}) {
			logDbg("Enter write_file_aux  Line: ".__LINE__.
				 "\n\n$msg1 \n\n".
			           Dumper($write_file_subr_args));
		}

		my $len_to_write = ( $rem_size ? (($end_count * $buffer_size) + $rem_size) :
				          (($end_count + 1) * $buffer_size) );
		if ($len_to_write != $size_bytes_aux) {
		        $msg = "write_file_aux  Line: ".__LINE__.
				 "\n\n$msg1 \n\n".
			         "Internal 'write_file' error:\n".
				 "'Length_to_be_written' =\n".
				 "       ( (end_count=$end_count * buffer_size=$buffer_size)  +".
				 "  rem_size=$rem_size )\n".
				 "                       = $len_to_write\n\n".
				 "     does not equal 'File_Size'=$size_bytes_aux";
			logcomment($msg);
			die($msg);
		}

		my %write_data_fcn_args = (
						special_volume_options	=> $special_volume_options,
						io_size           	=> $io_size,
						scramble          	=> $scramble,
						write_data_args   	=> $write_data_args,
						start_offset		=> $start_offset,
						end_count		=> $end_count,
						buffer_size		=> $buffer_size,
						file_size		=> $size_bytes_aux,
						file_path		=> $file_path,
					  );

		if ($dont_overwrite && util_exists($file_path)) {
			logDbg("Utilities::write_file_aux  Line: ".__LINE__.
			            "\n'dont_overwrite' is set and file ".
			            "$file_path exists, not overwriting");
			return(1);
		} else {
			my $wrlen;
			my $data_str_ref   = \$write_file_subr_args->{contents};
			my $count          = 0;
			my $len_written    = 0;
			my $offset         = $start_offset;
			for (  $count = 0;
			       $count <= $end_count; 
			      ($count++, $offset += $wrlen, $len_written += $wrlen))
			 {
				$wrlen = ((($count == $end_count) && $rem_size) ? $rem_size : $buffer_size);

				my $msg1 =        "Buffer_size:     $buffer_size\n".
						  "File size        $size_bytes_aux\n".
						  "Count:           $count\n".
						  "End_count:       $end_count\n".
					          "Offset:          $offset\n".
						  "Start_offset:    $start_offset\n".
						  "Bytes written:   $len_written\n".
						  "Writelen:        $wrlen\n".
				                  "File Path:       \"$file_path\"";

				if ($wrlen <= 0) {
				        $msg = "write_file_aux  Line: ".__LINE__.
						  "\n\n$msg1 \n\n".
						  "Internal 'write_file' error:\n".
						  "    'Write Len'=$wrlen is <= 0";
					logcomment($msg);
					die($msg);
				}
				if ($offset != $start_offset + $len_written) {
				        $msg = "write_file_aux  Line: ".__LINE__.
						  "\n\n$msg1 \n\n".
						  "Internal 'write_file' error:\n\n".
						  "  Offest=$offset  !=\n".
						  "     ( Start_Offset=$start_offset  +  ".
						  "len_written=$len_written) =   ".
						  ($start_offset + $len_written);
					logcomment($msg);
					die($msg);
				}
				if ( ($offset +  $wrlen)
				    >   ($start_offset + $size_bytes_aux) ) {
				        $msg = "write_file_aux  Line: ".__LINE__.
						  "\n\n$msg1 \n\n".
						  "Internal 'write_file' error:\n\n".
						  "  Quantity_A =  ( 'Offset'=$offset\n".
						  "                    +  'WriteLen'=$wrlen )  = ".
						  ($offset + $len_written + $wrlen)."\n\n".
						  "  ... is greater than ...\n\n".
						  "  Quantity_B =  ( 'Start Offset'=$start_offset\n".
						  "                    +  'File Size'=$size_bytes_aux )  = ".
						  ($start_offset + $size_bytes_aux)."\n";
					logcomment($msg);
					die($msg);
				}

				$write_data_fcn_args{write_len}         = $wrlen;
				$write_data_fcn_args{offset}            = $offset;
				$write_data_fcn_args{count}   	   	= $count;

			        my $write_data_fcn_args_copy = { %write_data_fcn_args };
			        $write_data_fcn_args_copy->{write_data_args} = "{ .... }";
				if ($argsref->{debug_verbose}) {
				      logDbg("write_file_aux  Line: ".__LINE__.
						  "\nCalling write_data_fcn ....\n\n".
						  Dumper($write_data_fcn_args_copy));
				}
				$$data_str_ref = &$write_data_fcn( %write_data_fcn_args );
				if ($argsref->{debug_verbose}) {
					logDbg("write_file_aux  Line: ".__LINE__.
						   "\n.... Returned from write_data_fcn");
				}

				# if (!$$buff_ref || ref($$buff_ref) || length($$buff_ref) < $wrlen)
				if (!defined($$data_str_ref) || !$$data_str_ref || ref($$data_str_ref)) {
				        $msg = "write_file_aux  Line: ".__LINE__.
						  "\n\n$msg1 \n\n".
						  "Error from 'write_data_fcn'\n".
						  "  -- did not return valid string: ".
						  "\"" . (defined($$data_str_ref) ? $$data_str_ref : 'UNDEF') .
						  "\"\n\nwrite_data_fcn_args:\n\n".
						  Dumper($write_data_fcn_args_copy);
					logcomment($msg);
					die($msg);
				}

				# my $data_len = length($$buff_ref);
				my $data_len = length($$data_str_ref);
				if ($data_len > $wrlen) {
				        $msg = "write_file_aux  Line: ".__LINE__.
						  "\n\n$msg1 \n".
						  "Data length:\t$data_len\n\n".
						  "Error in data returned from 'write_data_fcn' --\n".
						  "   Expected:  'data_length'=$data_len\n".
						  "                  ==  'write length'=$wrlen";
					logcomment($msg);
					die($msg);
				} elsif ($data_len > $wrlen) {
				        $msg = "write_file_aux  Line: ".__LINE__.
						  "\n\n$msg1 \n".
						  "Data length:\t$data_len\n\n".
						  "Unexpected length returned from 'write_data_fcn' --\n".
						  "   Expected:  'data_length'=$data_len\n".
						  "                  ==  'write length'=$wrlen";
					logcomment($msg);
				} else {
					if ($argsref->{debug_verbose}) {
					     $msg = "write_file_aux  Line: ".__LINE__.
						       "\n\n$msg1 \n".
						       "Data length:\t$data_len";
					     logDbg($msg);
					}
				}
				$write_file_subr_args_copy->{contents_length} = 
				    length($write_file_subr_args_copy->{contents});
				delete($write_file_subr_args_copy->{contents});

				$write_file_subr_args->{offset} = $offset;
				# $write_file_subr_args->{contents} = $data_str;
				if ($fstat && $count == $end_count) {
					$write_file_subr_args->{fstat} = 1;
					if ($argsref->{debug_verbose}) {
					     logDbg("write_file_aux  Line: ".__LINE__.
					                 "\nCalling FileManip::write_file:\n\n".
							  Dumper($write_file_subr_args_copy));
					}
					@attrs = FileManip::write_file( %$write_file_subr_args );
					if (!@attrs) {
						$msg = "write_file_aux  Line: ".__LINE__.
							    "FileManip::write_file failed: $@";
						logcomment($msg);
						die($msg);
					}

				} else {
					$write_file_subr_args->{fstat} = 0;
					if ($argsref->{debug_verbose}) {
					     logDbg("write_file_aux  Line: ".__LINE__.
					                "\nCalling FileManip::write_file:\n\n".
						        Dumper($write_file_subr_args_copy));
					}
					$rtn_status = FileManip::write_file( %$write_file_subr_args );
					if (!$rtn_status) {
						#last unless (@attrs);
						$msg = "write_file_aux  Line: ".__LINE__.
							    "\n\n'write_file' Args:\n\n".
							    Dumper($write_file_subr_args_copy)."\n\n".
							    "FileManip::write_file failed: $@";
						logcomment($msg);
						die($msg);
					}
					if ($argsref->{debug_verbose}) {
					     logDbg("Utilities::write_file_aux  Line: ".__LINE__.
						     "\nReturned $rtn_status from FileManip::write_file\n".
						     "File Path: \"$file_path\"");
					}
				}

				if ($truncate) {
					#  Truncate file to zero length using FileManip::setattr_file
					my %setattr_file_hash = (
						file     => $write_file_subr_args->{file},
						client   => $write_file_subr_args->{client},
						size     => 0,
					);
					if ($argsref->{debug_verbose}) {
					     logDbg("write_file_aux  Line: ".__LINE__.
					                "\nCalling FileManip::setattr_file:\n\n".
						        Dumper($write_file_subr_args_copy));
					}
					$rtn_status = FileManip::setattr_file( %setattr_file_hash );
					if (!$rtn_status) {
						$msg = "write_file_aux  Line: ".__LINE__.
							    "\n\n'write_file' Args:\n\n".
							    Dumper($write_file_subr_args_copy)."\n\n".
							    "FileManip::setattr_file failed: $@";
						logcomment($msg);
						die($msg);
					}
					if ($argsref->{debug_verbose}) {
					        logDbg("Utilities::write_file_aux  Line: ".__LINE__.
						         "\nReturned $rtn_status from FileManip::setattr_file\n".
						         "File Path: \"$file_path\"");
					}
				}

			}  # for ($count = 0 .... )

			if ($fstat && ref($rtn_fstat) eq 'HASH') {
				my %fst = ();
				($fst{dev}, $fst{ino}, $fst{mode}, $fst{nlink},
				     $fst{uid}, $fst{gid}, $fst{rdev}, $fst{size},
				     $fst{atime}, $fst{mtime}, $fst{ctime},
				     $fst{blksize}, $fst{blocks} ) = @attrs;
				if ($argsref->{debug_verbose}) {
					logDbg("Utilities::write_file_aux  Line: ".__LINE__.
						"\nReturned attrs from FileManip::write_file\n".
						"File Path: \"$file_path\"\n\n".
						Dumper(\%fst));
				}
				$rtn_fstat->{$file_path} = { %fst };
			}
		   }

		   if (ref($file_list) eq 'ARRAY') {
			push(@$file_list, $file_path);
		   }

		    my $errmsg = $@;
		    die($errmsg) if ($errmsg);
		    if ($argsref->{debug_verbose}) {
		         logDbg("Utilities::write_file_aux  Line: ".__LINE__.
			          "\nReturning\nErrmsg: \"$errmsg\"");
		    }
		    return($errmsg);
            };

	};  # End sub write_file_aux()
        #=================================================================================================


	return eval {

		my %orig_args = ( %args );
		my %args_copy = ( %args );
		%args = ( %orig_args );
		$args_copy{filer} = "Filer" if ($args{filer});
		$args_copy{Client} = "Client" if ($args{client});
		logDbg("Enter Utilities::write_file  Line: ".__LINE__."  Opts:\n\n".Dumper(\%args_copy)); 
	#        my %client_opts = ();
	#        foreach ( qw( contents mode random uid gid offset user ) ) {
	#		if (exists($args{$_})) {
	#			$client_opts{$_} = $args{$_};
	#		}
	#	}
		my $vol_path = $args{vol_path};
		my $mountpoint = $args{mountpoint};
		my $abs_path;
	 	if (exists($args{path})) {
	        	$abs_path = $args{path};
		} else {
			$abs_path = $args{file};
		}
	
		if ( $abs_path && ($args{vol_path} || $args{mountpoint}) ) {
			die("Utilities::write_file: If 'path' option is given then neither ".
				"'vol_path' or 'mountpoint' options should also be given");
		}
	
		my ($single_file, $numfiles);
		if ($args{numfiles}) {
			$numfiles = convert_to_bytes($args{numfiles});
			$single_file = ($numfiles <= 1);
		} else {
			$single_file = 1;
			$numfiles = undef;
		}
		my $step_size = convert_to_bytes($args{step_size});
		my $timeout = $args{timeout};
		my $client_agent;
		my $num_opts = 0;
		my $dont_overwrite = $args{dont_overwrite};
		my $proto;
		my $catch_errors = $args{catch_errors};
		my $msg;
		my $errmsg;
		my $agent = $args{agent};
		my $rtn_status;
		my @attrs;
		my $fstat = $args{fstat};
		my $verify = $args{verify};
	
		my $offset = $args{offset};
		my $size_bytes;
		my $contents;
	
		$num_opts = 0;
		$num_opts++ if ($args{size});
		$num_opts++ if ($args{contents});
		$num_opts++ if ($args{content});
		if ($num_opts != 1) {
			die("Utilities::write_file: Exactly 1 of 'size', 'contents', or \n".
			      "'content' options may be specified");
		}
		if ( defined($args{size}) ) {
			$size_bytes = convert_to_bytes($args{size});
			if ($size_bytes !~ /^\d+$/) {
				die("Utilities::write_file  Line: ".__LINE__.
				      "\nInvalid 'size' value: \"$args{size}\"");
			}
		} elsif (defined($args{contents}) || defined($args{content})) {
			$contents = $args{contents} || $args{content};
			if (ref($contents)) {
				die("Utilities::write_file  Line: ".__LINE__.
				      "\n'contents' must be a scalar");
			}
			if ($contents eq "") {
				die("Utilities::write_file  Line: ".__LINE__."\n".
				      "'contents' must not be empty string");
			}
			$size_bytes = length($contents);
		} else {
			die("Utilities::write_file  Line: ".__LINE__
			     ."\nNo 'size' or 'contents' options given");
		}
	
		if ($size_bytes == 0 && defined($offset) && $offset != 0) {
			die("Utilities::write_file  Line: ".__LINE__.
			     "\n'offset' must be zero if 'size' is zero");
		}
		$args{size_bytes} = $size_bytes;
		$args{contents} = $contents;
	
		my $write_data_fcn     = $args{write_data_fcn};
		my $write_data_args    = $args{write_data_args};;
	        if (!$write_data_fcn) {
			$write_data_fcn    = $default_write_data_fcn;
			$write_data_args   = { contents => $contents };
		}
		my %default_special_volume_options = (
			compression_enable    => $::COMPRESSION_ENABLE,
			asis_enable           => $::ASIS_ENABLE,
			encryption_enable     => $::ENCRYPTION_ENABLE,
			large_aggr	      => $::LARGE_AGGR,
		);
		$args{write_data_fcn} = $write_data_fcn;
		$args{write_data_args} = $write_data_args;
		my $special_volume_options = $args{special_volume_options};
		unless ($special_volume_options) {
			$special_volume_options = {};
		}
		# $special_volume_options = { %{$args{special_volume_options}} };
		$special_volume_options = $args{special_volume_options};
		$special_volume_options = { %$special_volume_options };
		while ( my ($k,$v) = each(%default_special_volume_options) ) {
			if ( defined($v) && !defined($special_volume_options->{$k}) ) {
				$special_volume_options->{$k} = $v
			}
		}
		$args{special_volume_options} = $special_volume_options;
	
		my $rtn_fstat = {};
		$args{rtn_fstat} = $rtn_fstat;
		if ($dont_overwrite) {
			logWarn("Utilities::write_file:  'dont_overwrite' is set");
		}
		my ($filer, $client, $cifs_client);
	
		if ($agent) {
			if ($agent !~ /localhost|client|cifs_client|filer/) {
				die("Utilities::write_file: Unexpected value for 'agent' option: \"$agent\"");
			}
			if ($args{debug_verbose}) {
				logDbg("Utilities::write_file  Line: ".__LINE__.
				          "   Agent option given: \"$agent\"");
			}
			if ($agent eq 'filer') {
				$filer = ($args{filer} || $::FILER);
			} elsif ($agent eq 'client') {
				$client = ($args{client} || $::CLIENT);
			} elsif ($agent eq 'cifs_client') {
				$client = ($args{cifs_client} || $::CIFS_CLIENT);
			}
		} else {
			$num_opts = 0;
			$num_opts++ if ($args{filer});
			$num_opts++ if ($args{client});
			$num_opts++ if ($args{cifs_client});
			if ($num_opts > 1) {
				die("Utilities::write_file: At most one of 'filer' or\n".
				    "'client', or 'cifs_client'\noptions may be specified\n".
				    "when 'agent' is not specified.");
			}
			if (!$agent && $args{filer}) {
				$agent = 'filer';
				if ($args{debug_verbose}) {
					logDbg("Utilities::write_file  Line: ".__LINE__."\n".
					       "No 'agent' specified, using '$agent'");
				}
			}
		}
	
		if ($agent eq 'filer') {
			if ($contents && length($contents) > 524) {
				$agent = undef;
			} else {
				$filer = $args{filer} || $::FILER;
			}
		}
		if (!$agent) {
			if ($args{client}) {
				$client = $args{client};
				$agent = 'client';
				if ($args{debug_verbose}) {
					logDbg("Utilities::write_file  Line: ".__LINE__."\n".
					       "No 'agent' specified, using '$agent' ($args{client})");
				}
			} elsif ($args{cifs_client}) {
				$cifs_client = $args{cifs_client};
				$agent = 'cifs_client';
				if ($args{debug_verbose}) {
					logDbg("Utilities::write_file  Line: ".__LINE__."\n".
					       "No 'agent' specified, using '$agent' ($args{client})");
				}
			} else {
				$agent = 'localhost';
				$client = 'localhost';
				if ($args{debug_verbose}) {
					logDbg("Utilities::write_file  Line: ".__LINE__."\n".
						"No 'agent', 'filer', 'client' or 'cifs_client' specified,\n".
					        "Using agent: '$agent'");
				}
			}
		}
		logDbg("Utilities::write_file  Line: ".__LINE__."   Using agent: \"$agent\": ");
	
		if ($agent eq 'filer') {
			unless ($abs_path || $vol_path) {
				die("Utilities::write_file: Either 'path' or 'vol_path' options must be given");
			}
			if (!$abs_path) {
				$abs_path = $vol_path;
			}
			my $fc = eval {make_conn($filer)};
			die("Utilities::write_file: 'make_conn' failed: $@") if ($@ || !$fc);
			my $fname = $fc->hostp()->hostname();
			my $size = $args{size};
			#my ($vol, $rel_path) = ($abs_path =~ m%/vol/([^/]+)/(.*)%);
			my (undef, undef, $vol, $rel_path) = split(m%/%, $abs_path, 4);
			if ($args{debug_verbose}) {
				logDbg("Utilities::write_file  Line: ".__LINE__."\n\n".
					"Filer: $fname\nPath: $abs_path\nSize: $args{size}\n".
					"Single File: $single_file\nNum Files: $numfiles\n".
					"Path: $abs_path\nVolume: $vol\nRel. Path: $rel_path\n" .
					"Timeout: $timeout");
			}
			if ($dont_overwrite && filer_vol_file_exists($fc, $vol, $rel_path)) {
				logDbg("Utilities::write_file  Line: ".__LINE__.
					"   'dont_overwrite' is set and file exists, returning...\n".
				        "File: $abs_path");
				return(1);
			}
			my $base_cmd;
			my $cmd_type;
			if ($contents) {
				$base_cmd = "wrfile -a $abs_path \"$contents\"";
				$cmd_type = 'wrfile';
			} else {
				$base_cmd = "mkfile $size $abs_path";
				$cmd_type = 'mkfile';
				if (defined($numfiles)) {
					$base_cmd .= " $numfiles";
					if ($step_size) {
						$base_cmd .= " $step_size";
					}
				} elsif (defined($step_size)) {
					#die("Utilities::write_file: Step-size option allowed only if numfiles option is also given");
					logWarn("Utilities::write_file: Step-size option given, \n".
					        "but not numfiles option");
					$numfiles = 1;
				}
			}
			my $txt;
			# my $txt = filer_cli_diag_say($fc, $base_cmd);
			my $priv_cmd = "priv set -q diag; $base_cmd";
			if ($timeout) {
				logDbg("Utilities::write_file  Line: ".__LINE__.
					"    Calling 'sayEOF' on filer --\nCmmd: \"$base_cmd\"\n".
					"Timeout: \"$timeout\"");
				$txt = sayEOF($fc, $priv_cmd, timeout=>$timeout);
			} else {
				logDbg("Utilities::write_file  Line: ".__LINE__.
					"\n\nCalling 'sayEOF' on filer --\nCmmd: \"$base_cmd\"\n(no timeout given)");
				$txt = sayEOF($fc, $priv_cmd);
			}
			if ($args{debug_verbose}) {
				logDbg("Utilities::write_file  Line: ".__LINE__."\n\nReply:\n$txt");
			}
			$rtn_status = ($txt eq "")?1:0;
			logDbg("Utilities::write_file  Line: ".__LINE__.
			          "  Returned from filer_cli_diag_say\n\nStatus: $rtn_status  Text: \"$txt\"");
			logDbg("Exit Utilities::write_file  Line: ".__LINE__."   Return status: $rtn_status");
			return($rtn_status);
		}
	
		#  'agent' is NOT filer...
	
		my ($vol, $rel_path);
		unless ($abs_path || ($vol_path && $mountpoint)) {
			die("Utilities::write_file: Either 'path' or both of 'vol_path' ".
				"and 'mountpoint' options must be given");
		}
		if ($rel_path && $rel_path =~ m%\\%) {
			$rel_path =~ s%\\+%\/%g;
		}
		if ($abs_path && $abs_path =~ m%\\%) {
			$abs_path =~ s%\\+%\/%g;
		}
		if (!$abs_path) {
			(undef, undef, $vol, $rel_path) = split(m%/%, $vol_path, 4);
			$abs_path = $mountpoint . '/' . $rel_path;
		}
		$abs_path =~ s%\/+%\/%g;
		unless ($abs_path =~ m%^([^\/]+:)?(\/[^\/]+)+%) {
			die("Utilities::write_file: Bad format for 'path': \"$abs_path\"");
		}
	
		if ($args{debug_verbose}) {
			logDbg("Utilities::write_file  Line: ".__LINE__."\n\nAgent: \"$agent\"\n".
				    "vol_path: \"$vol_path\"\nvolume: \"$vol\"\nrel_path: \"$rel_path\"\n".
				     "mountpoint: \"$mountpoint\"\nabs_path: \"$abs_path\"");
		}
	
		if ($agent eq 'localhost') {
			$client_agent = 'localhost';
		} elsif ($agent eq 'cifs_client') {
			$client_agent = 'cifs_client';
		} else {
			$client_agent = $client;
		}
		my @write_file_opts = qw( contents oflags fstat mode uid gid offset user );
		my %write_file_subr_args = ( client => $client_agent );
		for my $x (@write_file_opts) {
			if ( exists($args{$x}) && defined($args{$x}) ) {
				$write_file_subr_args{$x} = $args{$x};
			}
		}
		$args{write_file_subr_args} = \%write_file_subr_args;
	
	#	$abs_path =~ m%^(([^\/]+)?(\/[^\/]+)*)/[^\/]+$%;
		my $dirpath = $abs_path;
		$dirpath =~ s%\/[^\/]+$%%;
		my %stat_file_args = ( file => $dirpath, client => $client_agent, rtn_hash => 1 );
		my %dir_stat_info = FileManip::stat_file(%stat_file_args);
		logDbg("Utilities::write_file  Line: ".__LINE__."\n\nDir Path: \"$dirpath\"".
		        "\n\nStats:\n\n".Dumper(\%dir_stat_info));
		my $io_size = $dir_stat_info{io_size};
		$args{io_size} = $io_size;
	
		if ($verify && $args{offset}) {
			logWarn("Utilities::write_file:  Line: ".__LINE__.
			          "\n'verify' is set, and 'offset' is non-zero for file \"$abs_path\"\n".
				  "Will verify existance only");
		}
	
		my $verify_okay;
		my $file_size;
		my $msg1;
		if ($single_file) {
			$write_file_subr_args{file} = $abs_path;
	
			logDbg("Utilities::write_file  Line: ".__LINE__.
				"\nWriting single file --  Path: $abs_path");
			$errmsg = &$write_file_aux(\%args);
			if ($errmsg) {
				if ($catch_errors) {
			        	return eval { die($errmsg) };
				} else {
			        	die($errmsg);
				}
			}
			logDbg("Utilities::write_file  Line: ".__LINE__.
				"\n  -- Finished writing single file");
			if ($verify) {
				$msg1 = "";
				logDbg("Utilities::write_file  Line: ".__LINE__.
					"\nverifying file \"$abs_path\"");
				if ($args{offset}) {
					$verify_okay = test_file(file=>$abs_path, test=>'e', client=>$client);
					if ($verify_okay) {
						$msg1 = "";
					} else {
						$msg1 = "Error creating file: \"$abs_path\" -- file does not exist";
					}
				} else {
					my $actual_size = get_file_size(file=>$abs_path, client=>$client);
					$verify_okay = ($actual_size == $size_bytes);
					if ($verify_okay) {
						$msg1 = "";
					} else {
						$msg1 = "Bad size for created file \"$abs_path\"\n".
						           "Expected Size: \"$size_bytes\" bytes\n".
						           "Actual size:   \"$actual_size\" bytes";
					}
				}
				unless ($verify_okay) {
					$msg = "write_file  Line: ".__LINE__.
						    "\nVerify failed for file: \"$abs_path\"\n$msg1";
					logcomment($msg);
					if ($catch_errors) {
				        	return eval { die($msg) };
					} else {
				        	die($msg);
					}
				}
			}
		} else {
			logDbg("Utilities::write_file  Line: ".__LINE__.
				"\nWriting multiple files:\nNum Files: $numfiles\nBase Path: $abs_path");
			for (my $cnt=1; $cnt<=$numfiles; $cnt++) {
				my $cnt_path = $abs_path . "_" . $cnt;
				$write_file_subr_args{file} = $cnt_path;
	
				if ($args{debug_verbose}) {
					logDbg("Utilities::write_file  Line: ".__LINE__.
					         "    Calling 'write_file_aux'  Count: $cnt ....");
				}
				$errmsg = &$write_file_aux(\%args);
				if ($errmsg) {
					if ($catch_errors) {
				        	return eval { die($errmsg) };
					} else {
				        	die($errmsg);
					}
				}
				if ($args{debug_verbose}) {
					logDbg("Utilities::write_file  Line: ".__LINE__.
					          "    .... Returned from 'write_file_aux'  Count: $cnt");
				}
				if ($verify) {
					if ($args{debug_verbose}) {
						logDbg("Utilities::write_file  Line: ".__LINE__.
							"\nverifying file \"$cnt_path\"");
					}
					$msg1 = "";
					if ($args{soffset}) {
						$verify_okay = test_file(file=>$cnt_path, test=>'e', client=>$client);
						if ($verify_okay) {
							$msg1 = "";
						} else {
							$msg1 = "Error creating file: \"$cnt_path\"\n".
							        "    -- file does not exist";
						}
					} else {
						my $actual_size = get_file_size(file=>$cnt_path, client=>$client);
						$verify_okay = ($actual_size == $size_bytes);
						if ($verify_okay) {
							$msg1 = "";
						} else {
							$msg1 = "Bad size for created file \"$cnt_path\"\n".
							           "Expected Size: \"$size_bytes\" bytes\n".
							           "Actual size:   \"$actual_size\" bytes";
						}
					}
					unless ($verify_okay) {
						$msg = "write_file  Line: ".__LINE__.
							    "\nVerify failed for file: \"$cnt_path\"\n$msg1";
						logcomment($msg);
						if ($catch_errors) {
					        	return eval { die($msg) };
						} else {
					        	die($msg);
						}
					}
			}
		}
		logDbg("Utilities::write_file  Line: ".__LINE__.
			"\n  -- Finished writing multiple files");
		}

		if ($fstat) {
			logDbg("Exit Utilities::write_file  Line: ".__LINE__);
	#		logDbg("Exit Utilities::write_file  Line: ".__LINE__.
	#		       "\nReturn attrs:\n\n".Dumper($rtn_fstat));
			return($rtn_fstat);
		} else {
			logDbg("Exit Utilities::write_file  Line: ".__LINE__."\nReturn status: $rtn_status");
			return(1);
		}

	};
}


1;

