# Let emacs know that we are editing -*-perl-*-
#
# $Id$
##
## @summary Library of common fs setup routines
## @description
## Functions for wafl testing infrastructure that let users perform quick setup
## of common resources.
## @usage
##  use FS::Util::Setup;
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public

package FS::Util::Setup;

use strict;
no strict qw(refs);

use Exporter qw(import);

BEGIN {
    our @EXPORT = qw(
        &can_run_test
        &create_matching_container
        &get_matching_spare_disks
        &get_vol_identification
        &aggr_switch_root
        &vol_switch_root
        &create_vol_and_mount
        &create_and_mount_test_specific_vol
        &create_and_map_cifs_share
        &create_aggr
        &create_vol
        &create_test_specific_vol
        &mount_vol0
        &mount_root_vol
        &mount_root_vol_if_needed
        &mount_vol
        &umount_vol
        &common_cleanup
        &copy_system_files
        &calc_root_size
        &cold_start_filer
        &cross_populate_hosts_files
    );
} ## end BEGIN

use Carp qw(croak);
use Data::Dumper;
use Tharn;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate_pos BOOLEAN SCALAR HASHREF OBJECT UNDEF);
use Volume qw(vol_exists vol_root vol_container vol_create_flex vol_disk_free
    vol_create vol_purge vol_status vol_options tricky_vol_create_zapi
    vol_determine_flex_root_size volume_special_options_hook vol_online);
use Aggregate
    qw(aggr_exists aggr_create aggr_purge aggr_root aggr_online aggr_add aggr_status);
use Admin::License qw(license unlicense is_licensed);
use SPARSE::Util::Logger;
use SPARSE::Util::Filer_Utils qw(setup_cifs);
use QtreeCmds qw(qtreeSecurity qtreePresent);
use TestEnv::MinWin qw(minwin_compatible);
use MapShare qw(map_cifs_share);
use Utils qw(make_conn make_host make_conn_and_complain);
use VFILER::Util qw(get_vfiler_root_path);
use ExportFS qw(:nfs share_add export_path share_exists share_delete);
use mount qw(mount umount);
use Utils::Units
    qw(guess_unit convert_units strip_unit_spec convert_to_bytes UNIT_MB);
use RaidInfo;
use boot qw(reboot_to transit_to);
use SayWrap qw(sayEOF);
use snaplock qw(set_default_retention_period compliance_clock_initialize);
use FileManip qw(test_file make_dir);
use Filer::Version qw(
    are_we_ontapng
    get_wafl_version
    WV_SCRIMSHAW_MIN   WV_SCRIMSHAW_MAX
    WV_ANCHORSTEAM_MIN WV_ANCHORSTEAM_MAX
    WV_TSINGTAO_MIN  WV_TSINGTAO_MAX
    WV_GORDONBIERSCH_MIN WV_BOILERMAKER_MAX
    UNSUPPORTED_FEATURE_EXCEPTION
);
use disk qw(verify_spare_disks);
use File::Find qw(find);
use Data::Dumper;
use CIFS::Setup qw(is_cifs_setup cifs_setup);
use filer qw(filer_cifs_shares_get filer_cifs_start filer_cifs_stop);
use SnapCmds qw(snapsched_volume);
use HybridBuftree qw(aggr_is_convertible aggr_convert);
use install qw(newfs basic_setup);
use CLI::FileIO qw(rdfile wrfile);
use FS::Util::Utilities qw(set_wafl_scan_speed);
use Wafliron qw(wafliron_status wafliron_commit);
use constant DEFAULT_WAFL_SCAN_SPEED => 2000;

my $sudo_cmd
    = defined($main::SUDO_PATH) ? $main::SUDO_PATH : "/usr/local/bin/sudo";

my @MOUNT_POINTS;

# Used by mount_root_vol, mount_root_vol_if_needed  to 'remember' root vol mountpoints
#  Hash of  'client-name'  -->  'mountpoint'
my %_root_vol_mounts = ();

# Used by mount_vol, mount_vol_if_needed  to 'remember' non-root vol mountpoints
#  Hash of  'client-name'  -->  {  'vol-name'  -- > 'mountpoint' }
my %_non_root_vol_mounts = ();

###################################
#  strip_filer($);
#  Strips 'CONNECTREC' and 'HOSTREC' objects from hash (at top-level only)
sub _strip_filer($) {
    my ($opts) = @_;
    if ( ref($opts) !~ /hash/i ) {
        $@ = "";
        return (undef);
    }
    my $opts_copy = {%$opts};
    while ( my ( $k, $v ) = each(%$opts_copy) ) {
        if ( ref($v) =~ /(connectrec|hostrec)/i ) {
            $opts_copy->{$k} = "$v";
        }
    }
    $@ = "";
    return ($opts_copy);
} ## end sub _strip_filer($)

## @name ::create_vol_and_mount
## @summary connects to filer, creates a common volume, exports it,
##          licenses nfs, mounts the volume on localhost (or CLIENT).
##          by default, returns an array of connection object, volume name,
##          mount point.
##          if the 'return_created' option is set, then the array returned will
##          include two additional values: newly-created volumes list ref and
##          newly-created aggregates list ref.
## @arg %args. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER, or HOSTREC, or CONNECTREC object
##		   that represents the filer or an existing connection to the filer.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         filer_vol - name of the filer volume to create. defaults to
##                     $main::WORM_VOL if the 'snaplock' flag is set and
##                     $main::FILER_VOL otherwise.
##
##         filer_vol_mount - filer mount target. defaults to
##                           $main::FILER_VOL_MOUNT
##
##         mountpoint - local mount point. defaults to $main::MOUNTPOINT
##
##         vol_type - trad | flex. defaults to $main::VOL_TYPE
##
##         vol_size - volume size. defaults to $main::VOL_SIZE
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
## 	   snaplock - flag requesting a snaplock compliant volume. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
## 	   snaplock_enterprise - flag requesting a snaplock compliant volume
##                    with an enterprise-class license. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##
##

sub create_vol_and_mount(%) {
    my (%raw_opts) = @_;
    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    $raw_opts{client} = ref( $raw_opts{client} )
        if ( ref( $raw_opts{client} ) );
    logDbg(   "Enter create_vol_and_mount   Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my (%opts) = validate_mixed(
        @_,
        [],
        {   filer =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },

    #
    # Several existing scripts are using 'filer_conn' option to specify a
    # 'CONNECTREC' object, rather than a 'Connspec' string, as was intended.
    # To allow these scripts to work without modification, if 'filer_conn'
    # is given a 'CONNECTREC' object, we will treat that as if it was supplied
    # via the 'filer' option.
    #
    # filer_conn 	=> { type => SCALAR|UNDEF, optional => 1},
            filer_conn =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },
            filer_vol      => { type => SCALAR | UNDEF, optional => 1 },
            vol_type       => { type => SCALAR | UNDEF, optional => 1 },
            vol_size       => { type => SCALAR | UNDEF, optional => 1 },
            filer_aggr     => { type => SCALAR | UNDEF, optional => 1 },
            aggr_size      => { type => SCALAR | UNDEF, optional => 1 },
            timeout        => { type => SCALAR | UNDEF, optional => 1 },
            extern_setup   => { type => SCALAR | UNDEF, optional => 1 },
            cmdtimeout     => { type => SCALAR | UNDEF, optional => 1 },
            return_created => { type => SCALAR | UNDEF, optional => 1 },
            nfs_enable     => { type => SCALAR | UNDEF, optional => 1 },

            # "Mount" related options
            client        => { type => SCALAR | UNDEF, optional => 1 },
            filer_ip      => { type => SCALAR | UNDEF, optional => 1 },
            filer_vol_mnt => { type => SCALAR | UNDEF, optional => 1 },
            mountpoint    => { type => SCALAR | UNDEF, optional => 1 },
            mountopts     => { type => SCALAR | UNDEF, optional => 1 },

            #  "Special" Volume Options
            snaplock            => { type => SCALAR | UNDEF, optional => 1 },
            snaplock_enterprise => { type => SCALAR | UNDEF, optional => 1 },
            source              => { type => SCALAR | UNDEF, optional => 1 },
            large_aggr          => { type => SCALAR | UNDEF, optional => 1 },
            asis                => { type => SCALAR | UNDEF, optional => 1 },
            compression         => { type => SCALAR | UNDEF, optional => 1 },
            guarantee           => { type => SCALAR | UNDEF, optional => 1 },
            encryption          => { type => SCALAR | UNDEF, optional => 1 },
            raid_type           => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    return eval {
        logDbg(
            "create_vol_and_mount   Line: " . __LINE__ . "    args okay" );
        my @keys = keys(%opts);
        foreach my $k (@keys) {
            delete( $opts{$k} ) unless ( defined( $opts{$k} ) );
        }

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "create_vol_and_mount   Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $opts{filer}      = $opts{filer_conn};
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "create_vol_and_mount   Line: " 
                . __LINE__
                . "   Calling _create_vol_and_mount_aux" );
        return _create_vol_and_mount_aux(%opts);
    };
} ## end sub create_vol_and_mount(%)

## @name ::create_and_mount_test_specific_vol
## @summary connects to filer, creates a common volume, exports it,
##          licenses nfs, mounts the volume on localhost (or CLIENT).
##          by default, returns an array of connection object, volume name,
##          mount point.
##          if the 'return_created' option is set, then the array returned will
##          include two additional values: newly-created volumes list ref and
##          newly-created aggregates list ref.
##
## This routine differs from 'create_vol_and_mount' in that it is specifically
## intended to create volumes which are specific to the test script, and not
## to be used by any other script. The parameter 'vol_name' is required, and
## it must be different from the global var. $::FILER_VOL (if it is defined).
##
## In contrast, 'create_vol_and_mount' is to be used ONLY for the "common volume" $FILER_VOL.
##
## @arg %args. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         vol_name - name of the 'test specific' volume to create on the filer.
##		      This is a required arg. There is no default. An error is returned if
##                    filer_name matches $main::FILER_VOL. This option corresponds to the
##		      'filer_vol' option for 'create_vol', which is used for creating a
##		      "common resource" volume.
##
##         filer - hostspec of filer. defaults to $main::FILER, or HOSTREC, or CONNECTREC object
##		   that represents the filer or an existing connection to the filer.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         filer_vol_mount - filer mount target. defaults to
##                           $main::FILER_VOL_MOUNT
##
##         mountpoint - local mount point. defaults to $main::MOUNTPOINT
##
##         vol_type - trad | flex. defaults to $main::VOL_TYPE
##
##         vol_size - volume size. defaults to $main::VOL_SIZE
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
## 	   snaplock - flag requesting a snaplock compliant volume. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
## 	   snaplock_enterprise - flag requesting a snaplock compliant volume
##                    with an enterprise-class license. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##
##

sub create_and_mount_test_specific_vol(%) {
    my (%raw_opts) = @_;
    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    $raw_opts{client} = ref( $raw_opts{client} )
        if ( ref( $raw_opts{client} ) );
    logDbg(   "Enter create_and_mount_test_specific_vol   Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my (%opts) = validate_mixed(
        @_,
        [],
        {   filer =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },

    #
    # Several existing scripts are using 'filer_conn' option to specify a
    # 'CONNECTREC' object, rather than a 'Connspec' string, as was intended.
    # To allow these scripts to work without modification, if 'filer_conn'
    # is given a 'CONNECTREC' object, we will treat that as if it was supplied
    # via the 'filer' option.
    #
    # filer_conn 	=> { type => SCALAR|UNDEF, optional => 1},
            filer_conn =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },
            vol_name       => { type => SCALAR | UNDEF },
            vol_type       => { type => SCALAR | UNDEF, optional => 1 },
            vol_size       => { type => SCALAR | UNDEF, optional => 1 },
            filer_aggr     => { type => SCALAR | UNDEF, optional => 1 },
            aggr_size      => { type => SCALAR | UNDEF, optional => 1 },
            timeout        => { type => SCALAR | UNDEF, optional => 1 },
            extern_setup   => { type => SCALAR | UNDEF, optional => 1 },
            cmdtimeout     => { type => SCALAR | UNDEF, optional => 1 },
            return_created => { type => SCALAR | UNDEF, optional => 1 },
            nfs_enable     => { type => SCALAR | UNDEF, optional => 1 },

            # "Mount" related options
            client        => { type => SCALAR | UNDEF, optional => 1 },
            filer_ip      => { type => SCALAR | UNDEF, optional => 1 },
            filer_vol_mnt => { type => SCALAR | UNDEF, optional => 1 },
            mountpoint    => { type => SCALAR | UNDEF, optional => 1 },
            mountopts     => { type => SCALAR | UNDEF, optional => 1 },

            #  "Special" Volume Options
            snaplock            => { type => SCALAR | UNDEF, optional => 1 },
            snaplock_enterprise => { type => SCALAR | UNDEF, optional => 1 },
            source              => { type => SCALAR | UNDEF, optional => 1 },
            large_aggr          => { type => SCALAR | UNDEF, optional => 1 },
            asis                => { type => SCALAR | UNDEF, optional => 1 },
            compression         => { type => SCALAR | UNDEF, optional => 1 },
            guarantee           => { type => SCALAR | UNDEF, optional => 1 },
            encryption          => { type => SCALAR | UNDEF, optional => 1 },
            raid_type           => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    return eval {
        logDbg(   "create_and_mount_test_specific_vol   Line: " 
                . __LINE__
                . "    args okay" );
        $opts{test_specific_volume} = 1;
        if ( $::FILER_VOL && $opts{vol_name} eq $::FILER_VOL ) {
            die("create_and_mount_test_specific_vol: New volume name \"$opts{vol_name}\" cannot match 'FILER_VOL'"
            );
        }
        if ( !$opts{vol_name} ) {
            die("create_and_mount_test_specific_vol: No volume name \"$opts{vol_name}\" option specified"
            );
        }
        $opts{filer_vol} = $opts{vol_name};
        delete( $opts{vol_name} );
        my @keys = keys(%opts);
        foreach my $k (@keys) {
            delete( $opts{$k} ) unless ( defined( $opts{$k} ) );
        }

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "create_and_mount_test_specific_vol   Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $opts{filer}      = $opts{filer_conn};
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "create_vol_and_mount_test_specific_vol   Line: " 
                . __LINE__
                . "   Calling _create_vol_and_mount_aux" );
        return _create_vol_and_mount_aux(%opts);
    };
} ## end sub create_and_mount_test_specific_vol(%)

## @name ::_create_vol_and_mount_aux
## @summary connects to filer, creates a common volume, exports it,
##          licenses nfs, mounts the volume on localhost (or CLIENT).
##          by default, returns an array of connection object, volume name,
##          mount point.
##          if the 'return_created' option is set, then the array returned will
##          include two additional values: newly-created volumes list ref and
##          newly-created aggregates list ref.
## @arg %args. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         filer_vol - name of the filer volume to create. defaults to
##                     $main::WORM_VOL if the 'snaplock' flag is set and
##                     $main::FILER_VOL otherwise.
##
##         filer_vol_mount - filer mount target. defaults to
##                           $main::FILER_VOL_MOUNT
##
##         mountpoint - local mount point. defaults to $main::MOUNTPOINT
##
##         mountopts - String, comma-separated list of valid NFS mount options.
##
##         vol_type - trad | flex. defaults to $main::VOL_TYPE
##
##         vol_size - volume size. defaults to $main::VOL_SIZE
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
## 	   snaplock - flag requesting a snaplock compliant volume. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
## 	   snaplock_enterprise - flag requesting a snaplock compliant volume
##                    with an enterprise-class license. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##

sub _create_vol_and_mount_aux(%) {
    my (%opts) = @_;
    if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i ) {
        my $logmsg
            = "_create_vol_andmount_aux   Line: " 
            . __LINE__
            . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
            . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
            . "Will set 'filer' option to this object, and delete 'filer_conn' option";

        #	logDbg($logmsg);
        $opts{filer}      = $opts{filer_conn};
        $opts{filer_conn} = $main::FILER_CONN;
    } ## end if ( $opts{filer_conn}...

    my $FILER = $opts{filer}
        || $main::FILER;    # Can be HostSpec, Hostrec, or Connectrec.
    my $CLIENT     = $opts{client}     || $main::CLIENT;
    my $FILER_CONN = $opts{filer_conn} || $main::FILER_CONN;
    my $FILER_VOL  = $opts{filer_vol} ||

        # if snaplock is specified, then use a different
        # default name for the filer volume
        ( ( $opts{snaplock} || $opts{snaplock_enterprise} )
        ? $main::WORM_VOL
        : $main::FILER_VOL );
    my $FILER_IP      = $opts{filer_ip}      || $main::FILER_IP;
    my $FILER_VOL_MNT = $opts{filer_vol_mnt} || $main::FILER_VOL_MNT;
    my $MOUNTPOINT    = $opts{mountpoint}    || $main::MOUNTPOINT;
    my $MOUNTOPTS 
        = $opts{mountopts}
        || $main::MOUNTOPTS
        || 'actimeo=0,noac,hard,intr,noquota';
    my $VOL_TYPE = $opts{vol_type} || $main::VOL_TYPE;
    my $VOL_SIZE = $opts{vol_size} || $main::VOL_SIZE;
    my $FILER_AGGR = $opts{filer_aggr} ||

        # if snaplock is specified, then use a different
        # default name for the filer aggregate
        ( ( $opts{snaplock} || $opts{snaplock_enterprise} )
        ? $main::WORM_AGGR
        : $main::FILER_AGGR );
    my $AGGR_SIZE  = $opts{aggr_size}  || $main::AGGR_SIZE;
    my $LARGE_AGGR = $opts{large_aggr} || $main::LARGE_AGGR;
    my $EXTERN_SETUP
        = defined( $opts{extern_setup} )
        ? $opts{extern_setup}
        : $main::EXTERN_SETUP;
    my $TIMEOUT        = $opts{timeout}        || $main::TIMEOUT;
    my $CMDTIMEOUT     = $opts{cmdtimeout}     || $main::CMDTIMEOUT;
    my $RETURN_CREATED = $opts{return_created} || 0;

    #  -- Options not used explicitly, but passed to _create_vol_or_aggr:
    #     test_specific_volume, snaplock, snaplock_enterprise

    my %opts_copy = (%opts);
    if ( ref( $opts_copy{filer} ) ) {
        $opts_copy{filer} = sprintf( "%s", $opts_copy{filer} );
    }
    if ( ref( $opts_copy{client} ) ) {
        $opts_copy{client} = sprintf( "%s", $opts_copy{client} );
    }
    logDbg(   "_create_vol_and_mount_aux  Line: " 
            . __LINE__
            . "  Enter _create_vol_and_mount_aux\n\n"
            . Dumper( \%opts_copy ) );

    my $vol      = $FILER_VOL;
    my $vol_path = $FILER_VOL_MNT
        || ( length($FILER_VOL) ? "/vol/$FILER_VOL" : "/" );

    logDbg(   "_create_vol_and_mount_aux  Line: " 
            . __LINE__
            . "  Establish filer connection\n"
            . "  --  FILER: \"$FILER\"" );

    # filer connection
    my ( $fc, $fh );
    my $conn_timeout = $TIMEOUT;
    if ( ref($FILER) =~ /connectrec/i ) {
        $fc = $FILER;
    } else {
        $fc = make_conn( $FILER, $FILER_CONN );
        if ( !$fc ) {
            logDbg(   "_create_vol_and_mount_aux:  Line: " 
                    . __LINE__
                    . "   Could not establish filer connection: $@" );
            die("_create_vol_and_mount_aux:  Could not establish filer connection: $@"
            );
        } ## end if ( !$fc )
    } ## end else [ if ( ref($FILER) =~ /connectrec/i)
    $fh = $fc->hostp();
    if ( defined($CMDTIMEOUT) ) {
        $fc->timeout($CMDTIMEOUT);
    }
    $opts{filer} = $fc;

    my $hostname = $FILER_IP   || $fh->default_ip();
    my $mntpt    = $MOUNTPOINT || "/t/$hostname/$vol";
    my $pfiler   = $fh->{PFILER};

    $opts{return_created} = 1;
    $opts{nfs_enable}     = 1;
    my ( $new_vols, $new_aggrs );
    if ( !$pfiler ) {
        my ( $fc, $vol );

        logDbg(   "_create_vol_and_mount_aux  Line: " 
                . __LINE__
                . "   Calling _create_vol_or_aggr ..." );
        ( $fc, $vol, $new_vols, $new_aggrs ) = _create_vol_or_aggr(%opts)
            or croak $@;
        logDbg(   "_create_vol_and_mount_aux  Line: " 
                . __LINE__
                . "   ... Returned from _create_vol_or_aggr\n"
                . "fc:($fc), vol:($vol)" );
    } else {
        logTrc("vfiler $FILER running on $pfiler , NO volume setup done");
    }

    if ( $fc->hostp()->hosttype() eq "spinfs" ) {
        logDbg(   "_create_vol_and_mount_aux  Line: " 
                . __LINE__
                . "   'spinfs' hosttype" );
        $hostname = $fc->hostp()->{DATA_IP1};
        logcomment(
            "FS::Util::Setup::_create_vol_and_mount_aux - Detected NBLADE Settings, mounting $hostname\:$vol_path with $MOUNTOPTS"
        );
    } elsif ( my $server = are_we_ontapng($fc) ) {
        logDbg(   "_create_vol_and_mount_aux  Line: " 
                . __LINE__
                . "   ONTAP-NG (GX)" );
        $hostname = $fc->hostp()->{DATA_IP1};
        $vol_path = $server->volume_mount_point( volume => $vol );
        logcomment(
            "FS::Util::Setup::_create_vol_and_mount_aux - Detected NBLADE Settings, mounting $hostname\:$vol_path with $MOUNTOPTS"
        );
    } ## end elsif ( my $server = are_we_ontapng...

    my %mount_subr_opts = (
        hostname   => $hostname,
        vol_path   => $vol_path,
        client     => $CLIENT,
        mountpoint => $mntpt,
        mountopts  => $MOUNTOPTS,
        vers       => $main::MOUNT_VERS,
        chown      => 1,
        chmod      => '0777'
    );
    logDbg(   "_create_vol_and_mount_aux  Line: " 
            . __LINE__
            . "   Calling mount::mount\n\n"
            . Dumper( _strip_filer( \%mount_subr_opts ) ) );
    my $vol_mount_dir = mount(
        $hostname, $vol_path,
        client     => $CLIENT,
        mountpoint => $mntpt,
        mountopts  => $MOUNTOPTS,
        vers       => $main::MOUNT_VERS,
        chown      => 1,
        chmod      => '0777'
    ) or croak("mount::mount error: $@");
    logDbg(   "_create_vol_and_mount_aux  Line: " 
            . __LINE__
            . "    ... Returned from mount" );
    if ( my $server = are_we_ontapng($fc) ) {
        $vol = $server->get_volumes( volume => $vol );
    }

    logDbg(   "_create_vol_and_mount_aux  Line: " 
            . __LINE__
            . "\nReturning $fc, $vol, $vol_mount_dir" );
    push( @MOUNT_POINTS, $vol_mount_dir );
    if ($RETURN_CREATED) {
        return ( $fc, $vol, $vol_mount_dir, $new_vols, $new_aggrs );
    } else {
        return ( $fc, $vol, $vol_mount_dir );
    }
} ## end sub _create_vol_and_mount_aux(%)

## @name ::create_vol
## @summary creates a common volume on a filer. Connects to filer,
## by default, returns an array of filer connection and volume_name.
## if the 'return_created' option is set, then the array returned will include
## two additional values: newly-created volumes list ref and newly-created
## aggregates list ref.
##
## @arg %args. Optional. If not provided, values will be those of
##       NATE command line params. These can take values of:
##
##         filer - hostspec of filer. defaults to $main::FILER, or HOSTREC, or CONNECTREC object
##		   that represents the filer or an existing connection to the filer.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         filer_vol - name of the filer volume to create. defaults to
##                     $main::WORM_VOL if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_VOL otherwise.
##
##         filer_vol_mount - filer mount target. defaults to $main::FILER_VOL_MOUNT
##
##         vol_type - trad | flex. defaults to $main::VOL_TYPE
##
##         vol_size - volume size. defaults to $main::VOL_SIZE
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
## 	   snaplock - flag requesting a snaplock compliant volume. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
## 	   snaplock_enterprise - flag requesting a snaplock compliant volume
##                    license with an enterprise-class license. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
##         nfs_enable - Enable NFS (yes or no). Default is yes.
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##

sub create_vol(%) {
    my (%raw_opts) = @_;
    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    logDbg(   "Enter create_vol   Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my (%opts) = validate_mixed(
        @_,
        [],
        {   filer =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },

    #
    # Several existing scripts are using 'filer_conn' option to specify a
    # 'CONNECTREC' object, rather than a 'Connspec' string, as was intended.
    # To allow these scripts to work without modification, if 'filer_conn'
    # is given a 'CONNECTREC' object, we will treat that as if it was supplied
    # via the 'filer' option.
    #
    # filer_conn 	=> { type => SCALAR|UNDEF, optional => 1},
            filer_conn =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },
            filer_vol      => { type => SCALAR | UNDEF, optional => 1 },
            vol_type       => { type => SCALAR | UNDEF, optional => 1 },
            vol_size       => { type => SCALAR | UNDEF, optional => 1 },
            filer_aggr     => { type => SCALAR | UNDEF, optional => 1 },
            aggr_size      => { type => SCALAR | UNDEF, optional => 1 },
            extern_setup   => { type => SCALAR | UNDEF, optional => 1 },
            timeout        => { type => SCALAR | UNDEF, optional => 1 },
            cmdtimeout     => { type => SCALAR | UNDEF, optional => 1 },
            return_created => { type => SCALAR | UNDEF, optional => 1 },
            nfs_enable     => { type => SCALAR | UNDEF, optional => 1 },

            #  "Special" Volume Options
            snaplock            => { type => SCALAR | UNDEF, optional => 1 },
            snaplock_enterprise => { type => SCALAR | UNDEF, optional => 1 },
            source              => { type => SCALAR | UNDEF, optional => 1 },
            large_aggr          => { type => SCALAR | UNDEF, optional => 1 },
            asis                => { type => SCALAR | UNDEF, optional => 1 },
            compression         => { type => SCALAR | UNDEF, optional => 1 },
            guarantee           => { type => SCALAR | UNDEF, optional => 1 },
            encryption          => { type => SCALAR | UNDEF, optional => 1 },
            raid_type           => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    return eval {
        logDbg( "create_vol   Line: " . __LINE__ . "    args okay" );
        my @keys = keys(%opts);
        foreach my $k (@keys) {
            delete( $opts{$k} ) unless ( defined( $opts{$k} ) );
        }

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "create_vol   Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $opts{filer}      = $opts{filer_conn};
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "create_vol   Line: " 
                . __LINE__
                . "   Calling _create_vol_or_aggr" );
        return _create_vol_or_aggr(%opts);
    };
} ## end sub create_vol(%)

## @name ::create_test_specific_vol
## @summary creates a common volume on a filer. Connects to filer,
## by default, returns an array of filer connection and volume_name.
## if the 'return_created' option is set, then the array returned will include
## two additional values: newly-created volumes list ref and newly-created
## aggregates list ref.
##
## This routine differs from 'create_vol' in that it is specifically intended
## to create volumes which are specific to the test script, and not intended
## to be used by any other script. The parameter 'vol_name' is required, and
## it must be different from the global var. $::FILER_VOL (if it is defined).
##
## In contrast, 'create_vol' is to be used ONLY for the "common volume" $FILER_VOL.
##
## @arg %args. Optional. If not provided, values will be those of
##       NATE command line params. These can take values of:
##
##         filer - hostspec of filer. defaults to $main::FILER, or HOSTREC, or CONNECTREC object
##		   that represents the filer or an existing connection to the filer.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         vol_name - name of the filer volume to create. This is a required arg,
##                    there is no default. An error is returned if filer_name
##                    matches $main::FILER_VOL.
##
##         filer_vol_mount - filer mount target. defaults to $main::FILER_VOL_MOUNT
##
##         vol_type - trad | flex. defaults to $main::VOL_TYPE
##
##         vol_size - volume size. defaults to $main::VOL_SIZE
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
## 	   snaplock - flag requesting a snaplock compliant volume. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
## 	   snaplock_enterprise - flag requesting a snaplock compliant volume
##                    license with an enterprise-class license. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
##         nfs_enable - Enable NFS (yes or no). Default is yes.
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##

sub create_test_specific_vol(%) {
    my (%raw_opts) = @_;
    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    logDbg(   "Enter create_test_specific_vol   Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my (%opts) = validate_mixed(
        @_,
        [],
        {   filer =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },

    #
    # Several existing scripts are using 'filer_conn' option to specify a
    # 'CONNECTREC' object, rather than a 'Connspec' string, as was intended.
    # To allow these scripts to work without modification, if 'filer_conn'
    # is given a 'CONNECTREC' object, we will treat that as if it was supplied
    # via the 'filer' option.
    #
    # filer_conn 	=> { type => SCALAR|UNDEF, optional => 1},
            filer_conn =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },
            vol_name       => { type => SCALAR | UNDEF },
            vol_type       => { type => SCALAR | UNDEF, optional => 1 },
            vol_size       => { type => SCALAR | UNDEF, optional => 1 },
            filer_aggr     => { type => SCALAR | UNDEF, optional => 1 },
            aggr_size      => { type => SCALAR | UNDEF, optional => 1 },
            extern_setup   => { type => SCALAR | UNDEF, optional => 1 },
            timeout        => { type => SCALAR | UNDEF, optional => 1 },
            cmdtimeout     => { type => SCALAR | UNDEF, optional => 1 },
            return_created => { type => SCALAR | UNDEF, optional => 1 },
            nfs_enable     => { type => SCALAR | UNDEF, optional => 1 },

            #  "Special" Volume Options
            snaplock            => { type => SCALAR | UNDEF, optional => 1 },
            snaplock_enterprise => { type => SCALAR | UNDEF, optional => 1 },
            source              => { type => SCALAR | UNDEF, optional => 1 },
            large_aggr          => { type => SCALAR | UNDEF, optional => 1 },
            asis                => { type => SCALAR | UNDEF, optional => 1 },
            compression         => { type => SCALAR | UNDEF, optional => 1 },
            guarantee           => { type => SCALAR | UNDEF, optional => 1 },
            encryption          => { type => SCALAR | UNDEF, optional => 1 },
            raid_type           => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    return eval {
        logDbg(   "create_test_specific_vol   Line: " 
                . __LINE__
                . "    args okay" );
        $opts{test_specific_volume} = 1;
        if ( $::FILER_VOL && $opts{vol_name} eq $::FILER_VOL ) {
            die("create_test_specific_vol: New volume name \"$opts{vol_name}\" cannot match 'FILER_VOL'"
            );
        }
        $opts{filer_vol} = $opts{vol_name};
        delete( $opts{vol_name} );
        my @keys = keys(%opts);
        foreach my $k (@keys) {
            delete( $opts{$k} ) unless ( defined( $opts{$k} ) );
        }

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "create_test_specific_vol   Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $opts{filer}      = $opts{filer_conn};
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "create_test_specific_vol   Line: " 
                . __LINE__
                . "   Calling _create_vol_or_aggr" );
        return _create_vol_or_aggr(%opts);
    };
} ## end sub create_test_specific_vol(%)

## @name ::create_aggr
## @summary creates a common aggregate on a filer. Connects to filer,
## by default, returns an array of filer connection and aggregate_name.
## if the 'return_created' option is set, then the array returned will include
## one additional values: and newly-created aggregates list ref.
##
## @arg %args. Optional. If not provided, values will be those of
##       NATE command line params. These can take values of:
##
##         filer - hostspec of filer. defaults to $main::FILER, or HOSTREC, or CONNECTREC object
##		   that represents the filer or an existing connection to the filer.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
## 	   raid_type - The type of raid. Default is decided by the
##                     conr_type. If conr_type is "trad", it is "raid4".
##                     Where as when conr_type is "aggr", it is "raid_dp"
##
##	   large_aggr  -  If '0', then "regular" (32bit indirect) aggregate is created
##		          If '1', then "large" (64bit indirect) aggregate is created.
##			  Default is deternmined by global var LARGE_AGGR if it exists,
##			  otherwise '0'.
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##			    This routine never creates mew volumes, so the
##			    "volume list ref" is always an empty array.
##

sub create_aggr(%) {
    my (%raw_opts) = @_;
    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    logDbg(   "Enter create_aggr   Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my (%opts) = validate_mixed(
        @_,
        [],
        {   filer =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },

    #
    # Several existing scripts are using 'filer_conn' option to specify a
    # 'CONNECTREC' object, rather than a 'Connspec' string, as was intended.
    # To allow these scripts to work without modification, if 'filer_conn'
    # is given a 'CONNECTREC' object, we will treat that as if it was supplied
    # via the 'filer' option.
    #
    # filer_conn 	=> { type => SCALAR|UNDEF, optional => 1},
            filer_conn =>
                { type => SCALAR | UNDEF | HASHREF | OBJECT, optional => 1 },
            filer_aggr     => { type => SCALAR | UNDEF, optional => 1 },
            aggr_size      => { type => SCALAR | UNDEF, optional => 1 },
            extern_setup   => { type => SCALAR | UNDEF, optional => 1 },
            large_aggr     => { type => SCALAR | UNDEF, optional => 1 },
            timeout        => { type => SCALAR | UNDEF, optional => 1 },
            cmdtimeout     => { type => SCALAR | UNDEF, optional => 1 },
            raid_type      => { type => SCALAR | UNDEF, optional => 1 },
            return_created => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    return eval {
        logDbg( "create_aggr   Line: " . __LINE__ . "    args okay" );
        $opts{vol_type}         = 'flex';
        $opts{create_aggr_only} = 1;
        $opts{filer_vol}        = undef;
        my @keys = keys(%opts);
        foreach my $k (@keys) {
            delete( $opts{$k} ) unless ( defined( $opts{$k} ) );
        }

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "create_aggr   Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $opts{filer}      = $opts{filer_conn};
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "create_aggr   Line: " 
                . __LINE__
                . "   Calling _create_vol_or_aggr" );
        return _create_vol_or_aggr(%opts);
    };
} ## end sub create_aggr(%)

## @name ::_create_vol_or_aggr
## @summary creates a common volume or aggregate on a filer. Connects to filer,
## by default, returns an array of filer connection and volume_name.
## if the 'return_created' option is set, then the array returned will include
## two additional values: newly-created volumes list ref and newly-created
## aggregates list ref.
##
## @arg %args. Optional. If not provided, values will be those of
##       NATE command line params. These can take values of:
##
##         filer - hostspec of filer. defaults to $main::FILER
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN
##
##         filer_vol - name of the filer volume to create. defaults to
##                     $main::WORM_VOL if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_VOL otherwise.
##
##         filer_vol_mount - filer mount target. defaults to $main::FILER_VOL_MOUNT
##
##         vol_type - trad | flex. defaults to $main::VOL_TYPE
##
##         vol_size - volume size. defaults to $main::VOL_SIZE
##
##         filer_aggr - parent aggr (valid with flex vols). defaults to
##                     $main::WORM_AGGR if the 'snaplock_enterprise' or
##                     'snaplock' flag is set, and $main::FILER_AGGR otherwise.
##
##         aggr_size - aggregate size. defaults to $main::AGGR_SIZE
##
##	   large_aggr  -  If '0', then "regular" (32bit indirect) aggregate is created
##		          If '1', then "large" (64bit indirect) aggregate is created.
##			  Default is deternmined by global var LARGE_AGGR if it exists,
##			  otherwise '0'.
##
##         extern_setup - 0|1|preclean|reuse.
##
## 	   raid_type - The type of raid. Default is decided by the
##                     conr_type. If conr_type is "trad", it is "raid4".
##                     Where as when conr_type is "aggr", it is "raid_dp"
##
##         extern_setup - 0|1|preclean|reuse.
##                        1: vol should be there already; otherwise croak
##                        0: vol should not exist, create it; croak if exists
##                        preclean: destroy vol if exists, create new vol
##                        reuse: use vol if exists, otherwise create;
##                 for 1|reuse, croaks if vol is of wrong type;
##                 defaults to $main::EXTERN_SETUP from param SETUP_DONE
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
## 	   snaplock - flag requesting a snaplock compliant volume. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
## 	   snaplock_enterprise - flag requesting a snaplock compliant volume
##                    license with an enterprise-class license. if set,
##                    then filer_vol defaults to $main::WORM_VOL rather
##                    than $main::FILER_VOL, and filer_aggr defaults to
##                    $main::WORM_AGGR rather than $main::FILER_AGGR
##
##         nfs_enable - Enable NFS (yes or no). Default is yes.
##
##         return_created - return two additional values: newly-created volumes
##                          list ref and newly-created aggregates list ref.
##
##         test_specific_volume - Used only when called by 'create_test_specific_vol'
##

sub _create_vol_or_aggr(%) {
    my (%opts) = @_;

    my $FILER = $opts{filer}
        || $main::FILER;    # Can be HostSpec, Hostrec, or Connectrec.
    my $FILER_CONN = $opts{filer_conn} || $main::FILER_CONN;
    my $FILER_VOL = $opts{filer_vol} ||

        # if snaplock is specified, then use a different
        # default name for the filer volume
        ( ( $opts{snaplock} || $opts{snaplock_enterprise} )
        ? $main::WORM_VOL
        : $main::FILER_VOL );

    my $FILER_VOL_MNT = $opts{filer_vol_mnt} || $main::FILER_VOL_MNT;
    my $VOL_TYPE      = $opts{vol_type}      || $main::VOL_TYPE;
    my $VOL_SIZE      = $opts{vol_size}      || $main::VOL_SIZE;
    my $FILER_AGGR    = $opts{filer_aggr} ||

        # if snaplock is specified, then use a different
        # default name for the filer aggregate
        ( ( $opts{snaplock} || $opts{snaplock_enterprise} )
        ? $main::WORM_AGGR
        : $main::FILER_AGGR );

    my $AGGR_SIZE = $opts{aggr_size} || $main::AGGR_SIZE;
    my $EXTERN_SETUP
        = defined( $opts{extern_setup} )
        ? $opts{extern_setup}
        : $main::EXTERN_SETUP;

    my $LARGE_AGGR       = $opts{large_aggr}       || $main::LARGE_AGGR;
    my $TIMEOUT          = $opts{timeout}          || $main::TIMEOUT;
    my $CMDTIMEOUT       = $opts{cmdtimeout}       || $main::CMDTIMEOUT;
    my $RAID_TYPE        = $opts{raid_type}        || $main::RAID_TYPE;
    my $RETURN_CREATED   = $opts{return_created}   || 0;
    my $CREATE_AGGR_ONLY = $opts{create_aggr_only} || 0;
    my $TEST_SPECIFIC_VOLUME = $opts{test_specific_volume} || 0;

    logDbg( "Enter _create_vol_or_aggr  Line: " . __LINE__ );

    # these options are mutually exclusive - enforce it here
    ( $opts{snaplock} && $opts{snaplock_enterprise} )
        and croak
        "usage: 'snaplock' and 'snaplock_enterprise' are mutually exclusive options";

    my $aggr = $FILER_AGGR;
    my $vol;
    my $vol_path;
    if ($CREATE_AGGR_ONLY) {
        $vol       = undef;
        $vol_path  = undef;
        $FILER_VOL = undef;
        $VOL_TYPE  = undef;
        $VOL_SIZE  = undef;
    } else {
        $vol      = $FILER_VOL;
        $vol_path = $FILER_VOL_MNT
            || ( length($FILER_VOL) ? "/vol/$FILER_VOL" : "/" );
    }

    my $nfs_enable;
    if ($CREATE_AGGR_ONLY) {
        $nfs_enable = "no";
    } elsif ( !defined( $opts{nfs_enable} ) ) {
        $nfs_enable = "yes";
    } elsif ( $opts{nfs_enable} eq "1" ) {
        $nfs_enable = "yes";
    } elsif ( !$opts{nfs_enable} ) {
        $nfs_enable = "no";
    } else {
        $nfs_enable = $opts{nfs_enable};
    }
    logDbg(   "_create_vol_or_aggr  Line: " 
            . __LINE__
            . "\n\nopts{nfs_enable}: \"$opts{nfs_enable}\"\nnfs_enable: \"$nfs_enable\""
    );

    if ($TEST_SPECIFIC_VOLUME) {
        $EXTERN_SETUP = 'reuse';
        logDbg(   "_create_vol_or_aggr  Line: " 
                . __LINE__
                . "   TEST_SPECIFIC_VOLUME is true, so setting EXTERN_SETUP to 'reuse'"
        );
        $opts{extern_setup} = 'reuse';
    } ## end if ($TEST_SPECIFIC_VOLUME)

    logDbg(   "_create_vol_or_aggr  Line: " 
            . __LINE__
            . "  Establish filer connection\n"
            . "  --  FILER: \"$FILER\"" );

    # filer connection
    my ( $fc, $fh );
    my $conn_timeout = $TIMEOUT;
    if ( ref($FILER) =~ /connectrec/i ) {
        $fc = $FILER;
    } else {
        $fc = make_conn( $FILER, $FILER_CONN );
        if ( !$fc ) {
            logDbg(   "_create_vol_or_aggr  Line: " 
                    . __LINE__
                    . "   Could not establish filer connection: $@" );
            die("_create_vol_or_aggr:  Could not establish filer connection: $@"
            );
        } ## end if ( !$fc )
    } ## end else [ if ( ref($FILER) =~ /connectrec/i)

    $fh = $fc->hostp();
    if ( defined($CMDTIMEOUT) ) {
        $fc->timeout($CMDTIMEOUT);
    }

    logDbg(   "_create_vol_or_aggr  Line: " 
            . __LINE__
            . "   Check for 'snaplock', 'snaplock_enterprise' ...." );
    if ( $opts{snaplock} ) {
        logTrc("license snaplock on filer");
        license( $fc, "snaplock" ) or croak $@;
    } elsif ( $opts{snaplock_enterprise} ) {
        logTrc("license snaplock enterprise on filer");
        license( $fc, "snaplock_enterprise" ) or croak $@;
    }
    my $ver;
    if ( $opts{snaplock} || $opts{snaplock_enterprise} ) {
        logDbg( "_create_vol_or_aggr  Line: " . __LINE__ );
        $ver = get_wafl_version( $fc->{hostp} );
        if ( $ver > WV_SCRIMSHAW_MAX ) {
            logTrc(
                "initialize the compliance clock (if it is not initialized)");
            compliance_clock_initialize( $fc, reinit_ok => 1 );
        }
    } ## end if ( $opts{snaplock} ||...

    my %params = (
        fc                  => $fc,
        aggr                => $FILER_AGGR,
        aggr_size           => $AGGR_SIZE,
        raid_type           => $RAID_TYPE,
        vol                 => $FILER_VOL,
        vol_type            => $VOL_TYPE,
        vol_size            => $VOL_SIZE,
        setup               => $EXTERN_SETUP,
        large_aggr          => $opts{large_aggr},
        guarantee           => $opts{guarantee},
        compression         => $opts{compression},
        encryption          => $opts{encryption},
        asis                => $opts{asis},
        source              => $opts{source},
        snaplock            => $opts{snaplock},
        create_aggr_only    => $CREATE_AGGR_ONLY,
        snaplock_enterprise => $opts{snaplock_enterprise},
    );
    if ($RAID_TYPE) {
        $params{raid_type} = $RAID_TYPE;
    }

    my %params_copy = %params;
    $params_copy{fc} = ref( $params{fc} );
    logDbg(   "_create_vol_or_aggr  Line: " 
            . __LINE__
            . "   Calling _create_appropriate_vol ... \n\n"
            . "SETUP_DONE: \"$EXTERN_SETUP\"\nCREATE_AGGR_ONLY: $CREATE_AGGR_ONLY\nFILER_VOL: \"$FILER_VOL\"\n"
            . "VOL_TYPE: \"$VOL_TYPE\"\n\nParams to 'create_appropriate_vol':\n\n"
            . Dumper( \%params_copy ) );
    my ( $new_vols, $new_aggrs ) = _create_appropriate_vol(%params);
    logDbg(   "_create_vol_or_aggr  Line: " 
            . __LINE__
            . "   ... Returned from _create_appropriate_vol\n\n"
            . "New Vols: @$new_vols\nNew Aggrs: @$new_aggrs" );

    logDbg(   "Setup  Line: " 
            . __LINE__
            . "   _create_vol_or_aggr:  Enable NFS? \"$nfs_enable\"" );
    if ( $nfs_enable =~ /yes/i ) {
        logDbg(   "Setup  Line: " 
                . __LINE__
                . "   _create_vol_or_aggr:  license NFS" );
        logTrc(   "_create_vol_or_aggr  Line: " 
                . __LINE__
                . "   license nfs on filer" );
        license( $fc, "nfs" ) or croak $@;
        logDbg(   "Setup  Line: " 
                . __LINE__
                . "   _create_vol_or_aggr:  Export path \"$vol_path\"" );
        export_path( $fc, "$vol_path" )
            or croak($@);
    } ## end if ( $nfs_enable =~ /yes/i)

    if ( $opts{snaplock} || $opts{snaplock_enterprise} ) {
        logTrc(   "_create_vol_or_aggr  Line: " 
                . __LINE__
                . "   Set default retention period" );
        my $retention = eval {
            return set_default_retention_period( $fc, $main::WORM_RETENTION,
                $FILER_VOL );
        };
        if ($@) {
            if ( index( $@, UNSUPPORTED_FEATURE_EXCEPTION ) != -1 ) {
                if ( $opts{strict_retention} ) {
                    die $@;
                }
                logWarn(UNSUPPORTED_FEATURE_EXCEPTION);
            } else {
                die $@;
            }
        } ## end if ($@)
        logTrc(   "_create_vol_or_aggr  Line: " 
                . __LINE__
                . "   now, retention is '$retention'" );
    } ## end if ( $opts{snaplock} ||...

    logDbg(   "_create_vol_or_aggr  Line: " 
            . __LINE__ . "\n\n"
            . "Returning fc:($fc) vol:($vol)\nNew_vols: @{$new_vols}\nNew_aggrs: @{$new_aggrs}"
    );
    $Log->trace(
        "create_vol: Returning fc:($fc) vol:($vol) new_vols: @{$new_vols} new_aggrs: @{$new_aggrs}"
    );
    if ($RETURN_CREATED) {
        if ($CREATE_AGGR_ONLY) {
            return ( $fc, $aggr, $new_vols, $new_aggrs );
        } else {
            return ( $fc, $vol, $new_vols, $new_aggrs );
        }
    } else {
        if ($CREATE_AGGR_ONLY) {
            return ( $fc, $aggr );
        } else {
            return ( $fc, $vol );
        }
    } ## end else [ if ($RETURN_CREATED)
} ## end sub _create_vol_or_aggr(%)

# internal function
sub _create_appropriate_vol(%) {
    my (%opts) = @_;

    my $fc                  = $opts{fc};
    my $aggr                = $opts{aggr};
    my $aggr_size           = $opts{aggr_size};
    my $vol                 = $opts{vol};
    my $vol_type            = $opts{vol_type};
    my $large_aggr          = $opts{large_aggr};
    my $vol_size            = $opts{vol_size};
    my $vol_disks           = $opts{vol_disks};
    my $setup               = $opts{setup};
    my $guarantee           = $opts{guarantee};
    my $compression         = $opts{compression};
    my $encryption          = $opts{encryption};
    my $asis                = $opts{asis};
    my $snaplock            = $opts{snaplock};
    my $raid_type           = $opts{raid_type};
    my $create_aggr_only    = $opts{create_aggr_only};
    my $snaplock_enterprise = $opts{snaplock_enterprise};

    my %flex_args = ();
    $flex_args{compression} = $compression  if ( defined($compression) );
    $flex_args{encryption}  = $encryption   if ( defined($encryption) );
    $flex_args{asis}        = $asis         if ( defined($asis) );
    $flex_args{guarantee}   = $guarantee    if ( defined($guarantee) );
    $flex_args{source}      = $opts{source} if ( defined( $opts{source} ) );

    my %trad_or_aggr_args = ();
    $trad_or_aggr_args{snaplock} = $snaplock if ( defined($snaplock) );
    $trad_or_aggr_args{snaplock_enterprise} = $snaplock_enterprise
        if ( defined($snaplock_enterprise) );

    die("\"vol\" is undefined or NULL") unless ( $create_aggr_only || $vol );
    die("\"vol_type\" is undefined or NULL")
        unless ( $create_aggr_only || $vol_type );
    die("\"vol_size\" is undefined or NULL")
        unless ( $create_aggr_only || $vol_size );
    my $vol_initially_exists = undef;
    if ( $vol && !$create_aggr_only ) {
        $vol_initially_exists = vol_exists( $fc, $vol );
    }
    my $aggr_initially_exists = undef;
    if ( $create_aggr_only || $vol_type ne 'trad' ) {
        if ( !$aggr ) {
            die("\"aggr\" is undefined or NULL, and vol_type is \"flex\"");
        }
        $aggr_initially_exists = aggr_exists( $fc, $aggr );
    } ## end if ( $create_aggr_only...
    logDbg(   "_create_appropriate_vol  Line: " 
            . __LINE__ . "\n"
            . "SETUP: \t\t\"$setup\"\nVOL: \t\t\"$vol\"\nAGGR:\t\t\"$aggr\"\n"
            . "VOL_TYPE: \t\"$vol_type\"\n"
            . "Vol exists: \t\""
            . ( $vol_initially_exists ? "Yes" : "No" ) . "\"\n"
            . "Aggr exists: \t\""
            . ( $aggr_initially_exists ? "Yes" : "No" ) . "\"\n"
            . "Create aggr only:  \""
            . ( $create_aggr_only ? "Yes" : "No" ) . "\"\n"
            . "" );

    my ( @new_vols, @new_aggrs );

    #
    # Original behavior:
    # 0 -- volume missing, create it; croak if exists
    # 1 -- volume exists, use it; croak if missing
    #
    # Extensions to the original behavior:
    # preclean -- if the volume already exists, destroy it;
    #			 unconditionally create a new volume.
    # reuse -- If the volume exists, use it; otherwise create it.
    #		  Check that the volume is of the correct type.
    #

    if ( $setup =~ /^preclean$/i ) {
        if ($vol_initially_exists) {
            logTrc("purging any existing $vol");
            vol_purge( $fc, $vol ) or logWarn($@);
            $vol_initially_exists = 0;
        }
    } elsif ( $setup =~ /^reuse$/i ) {
        if ($vol_initially_exists) {
            logTrc("reusing existing volume $vol");
        }
    } elsif ( $setup eq "1" ) {
        if ($create_aggr_only) {
            if ( !$aggr_initially_exists ) {
                croak(
                    "_create_appropriate_vol:  SETUP_DONE == \"$setup\"\n\nExpected aggr $aggr "
                        . "to have previously been created but vol does not exist: \n $@"
                );
            } ## end if ( !$aggr_initially_exists)
        } elsif ( !$vol_initially_exists ) {
            croak(
                "_create_appropriate_vol:  SETUP_DONE == \"$setup\"\n\nExpected volume $vol "
                    . "to have previously been created but vol does not exist: \n $@"
            );
        } ## end elsif ( !$vol_initially_exists)
    } elsif ( $setup eq "0" ) {
        if ($vol_initially_exists) {
            croak(
                "_create_appropriate_vol:  SETUP_DONE == \"$setup\"\n\nDid not expect volume $vol "
                    . "to have previously been created but vol already exists: \n $@"
            );
        } ## end if ($vol_initially_exists)
    } ## end elsif ( $setup eq "0" )

    if ( $create_aggr_only || $vol_type ne 'trad' ) {

        # if we are not dealing with traditional vol
        # need to check for presence of aggregate
        my %aggr_args = %trad_or_aggr_args;
        logDbg(   "_create_appropriate_vol  Line: " 
                . __LINE__
                . "   Either have 'create_aggr_only' or 'non-trad'\n"
                . "type volume -- need to verify or create aggregate $aggr" );
        $aggr_args{raidtype} = $raid_type if ( defined($raid_type) );
        if ($aggr_initially_exists) {
            logDbg(   "_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   Aggr already exists" );
        } else {
            logDbg(   "_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   Aggr does not exist -- must create it" );
            if ( defined($large_aggr) ) {
                $aggr_args{indirect_type} = ( $large_aggr ? '64' : '32' );
                if ($large_aggr) {
                    logDbg(
                        "_create_appropriate_vol:  Creating \'LA\' aggregate"
                    );
                }
            } ## end if ( defined($large_aggr...
            logDbg(   "_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   Calling aggr_create ..." );
            aggr_create( $fc, $aggr, $aggr_size, %aggr_args )
                or croak("aggr_create: $@");
            logDbg(   "_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   ... Returned from aggr_create ..." );
            if ( !aggr_exists( $fc, $aggr ) ) {
                croak(    "_create_appropriate_vol  Line: " 
                        . __LINE__
                        . "   'aggr_create' failed: $@" );
            }
            push( @new_aggrs, $aggr );
        } ## end else [ if ($aggr_initially_exists)

        # make sure aggr is online
        my $retry = 1;
        my $agg_state;
        my $sleep_time = 300;
        my $max_loop   = 13;

        logStep("Verify that $aggr is online");
        ## @step verify that aggr is online
        while ( $retry < $max_loop ) {
            $agg_state = aggr_status( $fc, $aggr )->state();
            if ( $agg_state =~ /online/i ) {
                logcomment("$aggr is online");
                last;
            }

            # sleep for sometime and check aggr state
            logStep(
                "$aggr is still not online, sleep for $sleep_time secs and retry $retry"
            );
            sleep($sleep_time);
            $retry++;
        } ## end while ( $retry < $max_loop)

        if ( $retry == $max_loop ) {
            logStep(
                "TimedOut: $aggr didn't become online after aggregate creation"
            );
            croak($@);
        } ## end if ( $retry == $max_loop)

        if ($create_aggr_only) {
            return ( [], \@new_aggrs );
        }
    } ## end if ( $create_aggr_only...

    if ($vol_initially_exists) {
        logDbg(   "_create_appropriate_vol  Line: " 
                . __LINE__
                . "   Volume already exists -- see if it is suitible" );
        if ( $vol_type eq 'flex' ) {

            # Make sure that the volume being reused is the size
            # the caller requested.

            # If it is a flex vol then set the size to the requested
            # volume size. This may grow, shrink, or not change the
            # size of the volume. If the volume size cannot change
            # then ignore the error and use the existing volume.
            eval { sayEOF( $fc, "vol size $vol $vol_size" ) };
        } ## end if ( $vol_type eq 'flex')

        # check if the existing volume matches the correct type
        my $obj = vol_status( $fc, $vol ) or croak($@);
        my $type = $obj->type();
        if ( $type ne $vol_type ) {
            croak(    "Existing volume '$vol' is of the wrong type: "
                    . "$type. Expected $vol_type." );
        }
        my %vopts = $obj->options();
        if ( $snaplock and !$vopts{snaplock_compliance} ) {
            croak("volume $vol not a SnapLock volume");
        }
        if ( $snaplock_enterprise and !$vopts{snaplock_enterprise} ) {
            croak("volume $vol not a SnapLock Enterprise volume");
        }

        # Check if aggregate should be converted to 64-bit. We are
        # converting the aggregate during volume creation because
        # the goal is to test with as many hybrid buftrees as we
        # can. If the aggregate was converted as part of aggregate
        # creation, all its volumes would have pure 64-bit buftrees.
        if ( $vol_type ne 'trad' && $main::CONVERT_TO64 ) {

            # XXX This check should be changed to
            # >= WV_ROLLINGROCK_MIN when that value is defined
            if ( get_wafl_version( $fc->hostp() ) > WV_BOILERMAKER_MAX ) {
                if ( aggr_is_convertible( $fc, $aggr ) ) {
                    logcomment(
                        "FS::Util::Setup::_create_appropriate_vol - Converting aggregate $aggr to a mixed-format (hybrid) buftree"
                    );
                    aggr_convert( $fc, $aggr );
                } else {
                    logcomment(
                        "FS::Util::Setup::_create_appropriate_vol - Skipping conversion: aggregate $aggr is root or already 64-bit"
                    );
                }
            } else {
                croak(
                    "FS::Util::Setup::_create_appropriate_vol - filer version does not support converting to 64-bit aggregates"
                );
            }
        } ## end if ( $vol_type ne 'trad'...

        # XXXX need to check if vol options are correct

        # no need to create volume.
        return ( \@new_vols, \@new_aggrs );

    } else {    # Vol dous not exist, must create it
        logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: " 
                . __LINE__
                . "   Volume $vol does not already exist -- create it" );
        logTrc(
            "FS::Util::Setup::_create_appropriate_vol  Volume $vol does not already exist -- create it"
        );
        if ( $vol_type ne 'trad' ) {
            ## Burt 357348: Need to check the status of aggregate
            ## and make sure that the aggregate is online only
            ## in case of flexible volume creation.
            aggr_online( $fc, $aggr ) or croak($@);

            logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   Use 'vol_create_flex' for all vol types except 'trad'"
            );
            logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   Calling vol_create_flex ..." );
            vol_create_flex( $fc, $vol, $aggr, $vol_size, %flex_args )
                or die($@);
            logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   ... Returned from vol_create_flex ..." );
            if ( !vol_exists( $fc, $vol ) ) {
                croak(    "FS::Util::Setup::_create_appropriate_vol  Line: "
                        . __LINE__
                        . "   'vol_create_flex' failed: $@" );
            }
            push( @new_vols, $vol );
        } else {    # Vol type is 'trad'
            logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: " 
                    . __LINE__
                    . "   Use 'vol_create' for 'trad' vol type" );
            my %trad_args = %trad_or_aggr_args;
            if ( get_wafl_version( $fc->hostp() ) >= WV_SCRIMSHAW_MIN ) {
                logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: "
                        . __LINE__
                        . "   Calling vol_create ..." );
                vol_create( $fc, $vol, $vol_size, %trad_args )
                    or croak($@);
                logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: "
                        . __LINE__
                        . "   ... Returned from vol_create ..." );
                if ( !vol_exists( $fc, $vol ) ) {
                    croak( "FS::Util::Setup::_create_appropriate_vol  Line: "
                            . __LINE__
                            . "   'vol_create_flex' failed: $@" );
                }
                push( @new_vols, $vol );
            } else {
                logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: "
                        . __LINE__
                        . "   Calling vol_create ..." );
                vol_create( $fc, $vol, $vol_size, %trad_args )
                    or croak($@);
                logDbg(   "FS::Util::Setup::_create_appropriate_vol  Line: "
                        . __LINE__
                        . "   ... Returned from vol_create ..." );
                if ( !vol_exists( $fc, $vol ) ) {
                    croak( "FS::Util::Setup::_create_appropriate_vol  Line: "
                            . __LINE__
                            . "   'vol_create_flex' failed: $@" );
                }
                push( @new_vols, $vol );
            } ## end else [ if ( get_wafl_version(...
        } ## end else [ if ( $vol_type ne 'trad')
    } ## end else [ if ($vol_initially_exists)

    logDbg( "FS::Util::Setup::_create_appropriate_vol  Line: " . __LINE__ );

#if ($vol_type eq 'trad' or $compression eq 'no' or $encryption eq 'no' or $asis eq 'no') {
#	logDbg("logDbg("_create_appropriate_vol()  Not Calling ".
#			 "Volume::volume_special_options_hook  ....");
#} else {
#	 logDbg("_create_appropriate_vol()  Calling ".
#		  "Volume::volume_special_options_hook  ....");
#		   Volume::volume_special_options_hook(filer=>$fc, volume=>$vol);
#}

    logDbg(
        "FS::Util::Setup::_create_appropriate_vol - Exit Line: " . __LINE__ );
    return ( \@new_vols, \@new_aggrs );
} ## end sub _create_appropriate_vol(%)

sub mount_vol0 {
    logWarn(
        "Setup::mount_vol0() is DEPRECATED!. Please call Setup::mount_root_vol() instead"
            . "\n After sleeping for a minute, mount_root_vol() will be called. That is almost "
            . "enough time to change the test script to call Setup::mount_root_vol() ;-)"
    );
    sleep(60);
    mount_root_vol(@_);

} ## end sub mount_vol0

## @name ::mount_root_vol
##
## @summary this is useful for mucking with /etc config files
##         so common, that a wrapper is useful. returns mountpoint
##         if successful.
##
## @arg filer- hostspec of filer
##
## @arg mountopts - Optional. string which represents mount opts.
##      defaults to "actimeo=0,noac,hard,intr,noquota"
##
## @arg local_mnt_pt - Ooptional. where to mount. defaults to /t/<filer name>/root
##
## @arg %args. Optional. If not provided, values will be those of NATE command line params.
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##	   localhost  -  Boolean, default is '0'. If '1', will use 'localhost' as client.
##                       Overrides 'client' option. The motivation for this option is
##			 that many tests which use a remote NFS client would still
##			 do tasks such as read or modify "/etc" and other setup, config,
##			 log, etc. files on the 'local' host, not the NFS client.
##
## 	   timeout - timeout value for the connection object and mount
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##
##         chmod         - If given, then the octal code which will be used
##                          to 'chmod' the mountpoint

sub mount_root_vol {
    return eval {
        logDbg(
            "FS::Util::Setup::mount_root_vol - Enter -- Trying 'new_mount_root_vol' ...."
        );
        my ( $err, $mntpt ) = new_mount_root_vol(@_);
        if ($@) {

#  'new_mount_root_vol' encountered error not related to arglist format, so exit with error.
            logDbg(
                "FS::Util::Setup::mount_root_vol -  'new_mount_root_vol' encountered fatal error -- exiting: $@"
            );
            croak($@);
        } elsif ( $err == 0 ) {

            # 'new_mount_root_vol' succeeded.
            logDbg(
                "FS::Util::Setup::mount_root_vol -  'new_mount_root_vol'  succeeded -- returning: $mntpt"
            );
            return ($mntpt);
        } elsif ( $err == -1 ) {

  #  Continue processing request -- 'new_mount_root_vol' encounted error
  #  while checking arglist. The arglist is probably okay for 'mount_root_vol'
        } else {

#  'new_mount_root_vol' encountered error not related to arglist format, so exit with error.
            logDbg(
                "FS::Util::Setup::mount_root_vol -  'new_mount_root_vol' encountered fatal error -- exiting"
            );
            croak("FS::Util::Setup::mount_root_vol - $@");
        } ## end else [ if ($@)

        my ( $filer, $mountopts, $local_mnt_pt, %opts ) = @_;
        my %raw_opts = %opts;
        $raw_opts{filer}     = ( ref($filer) ) ? ref($filer)     : $filer;
        $raw_opts{mountopts} = ( ref($filer) ) ? ref($mountopts) : $mountopts;
        $raw_opts{local_mnt_pt}
            = ( ref($filer) ) ? ref($local_mnt_pt) : $local_mnt_pt;
        $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
            if ( ref( $raw_opts{filer_conn} ) );
        $raw_opts{client} = ref( $raw_opts{client} )
            if ( ref( $raw_opts{client} ) );
        logDbg(   "FS::Util::Setup::mount_root_vol -   Line: " 
                . __LINE__
                . "    Opts:\n\n"
                . Dumper( \%raw_opts ) );

        my $FILER_CONN = $opts{filer_conn} || $main::FILER_CONN;
        my $CLIENT     = $opts{client}     || $main::CLIENT;
        my $TIMEOUT    = $opts{timeout}    || $main::TIMEOUT;
        my $CMDTIMEOUT = $opts{cmdtimeout} || $main::CMDTIMEOUT;

        my ( $mnt_vol, $rootvol, $fc, $fh );
        if ( !$filer ) {
            $filer = $main::FILER;
        }
        $fh = host($filer);
        my $pfiler = $fh->{PFILER};

        my $uid = $>;

        if ( $opts{localhost} ) {
            logDbg(
                "FS::Util::Setup::mount_root_vol -- 'localhost' option set, setting CLIENT='localhost'"
            );
            $CLIENT = 'localhost';
        } ## end if ( $opts{localhost} )
        $_root_vol_mounts{$CLIENT} = undef;

        if ($pfiler) {
            logTrc(
                "FS::Util::Setup::mount_root_vol - mount_root_vol: $filer running on $pfiler "
            );
            $fc = connect( $FILER_CONN, -timeout, $TIMEOUT );
            $fc->timeout($CMDTIMEOUT);
            $mnt_vol = get_vfiler_root_path( $fc, $filer );

        } else {

            # then pfiler client is calling
            $fc = connect( $FILER_CONN, -timeout, $TIMEOUT );
            $fc->timeout($CMDTIMEOUT);
            $rootvol = vol_root($fc)
                or croak($@);
            logTrc(
                "FS::Util::Setup::mount_root_vol - root volume is $rootvol");
            $mnt_vol = "/vol/$rootvol";
        } ## end else [ if ($pfiler)

        $fh = host($filer);
        my $hostname = $fh->default_ip();

        $local_mnt_pt = "/t/$hostname/$rootvol"
            unless defined($local_mnt_pt);
        $mountopts = "actimeo=0,noac,hard,intr,noquota"
            unless defined($mountopts);

        if ( $fc && ref($fc) =~ /connectrec/i ) {
            logTrc(
                "FS::Util::Setup::mount_root_vol -  Filer connection established"
            );
        } else {
            logWarn(
                "FS::Util::Setup::mount_root_vol -  Could not establish filer connection"
            );
        }

        if ( !is_licensed( $fc, "nfs" ) ) {
            logTrc("FS::Util::Setup::mount_root_vol - license nfs on filer");
            license( $fc, "nfs" ) or croak $@;
        }

        export_path( $fc, "$mnt_vol" )
            or croak($@);

        my $sec = eval { qtreeSecurity( $fc, $mnt_vol ) };
        if ( ( $sec ne 'unix' ) && ( $sec ne 'mixed' ) ) {
            my $rtn = eval { qtreeSecurity( $fc, $mnt_vol, 'mixed' ) };
            logWarn(
                "FS::Util::Setup::mount_root_vol - can't change $mnt_vol access to 'mixed: $@"
            ) unless ($rtn);
        } ## end if ( ( $sec ne 'unix' ...

        logTrc(
            "FS::Util::Setup::mount_root_vol -  mnt_vol: \"$mnt_vol\"  local_mnt_pt: \"$local_mnt_pt\""
        );
        my %mount_subr_opts = (
            hostname   => $hostname,
            vol_path   => $mnt_vol,
            mountpoint => $local_mnt_pt,
            mountopts  => $mountopts,
            client     => $CLIENT,
            timeout    => $TIMEOUT,
            chmod      => '0777',
            vers       => $main::MOUNT_VERS,
        );
        logDbg(   "FS::Util::Setup::mount_root_vol -  Line: " 
                . __LINE__
                . "   Calling mount::mount\n\n"
                . Dumper( _strip_filer( \%mount_subr_opts ) ) );
        $mntpt = mount::mount(
            $hostname, $mnt_vol,
            mountpoint => $local_mnt_pt,
            mountopts  => $mountopts,
            client     => $CLIENT,
            timeout    => $TIMEOUT,
            chmod      => '0777',
            vers       => $main::MOUNT_VERS,
        ) or croak("mount::mount error: $@");
        logTrc(
            "FS::Util::Setup::mount_root_vol - Exiting mounted: \"$mntpt\"");
        push( @MOUNT_POINTS, $mntpt );

        $_root_vol_mounts{$CLIENT} = $mntpt;
        logDbg(
            "FS::Util::Setup::mount_root_vol - Exit _root_vol_mounts{$CLIENT}: \"$_root_vol_mounts{$CLIENT}\"\n"
                . " -- Returning mountpoint: \"$mntpt\"" );
        return $mntpt;
    };
} ## end sub mount_root_vol

## @name ::new_mount_root_vol
##
## @summary Mounts the root volume of the filer on localhost (or CLIENT). Similar to 'mount_root_vol',
##	    except arglist is expected to be a valid hashlist (even # of args, etc.)
##
##          This fcn. expects to be called from the existing 'mount_root_vol', and to allow that fcn.
##	    to accept both the current "old-style" (non-hashlist) arglist, or a "new-style" (valid hashlist),
##	    with hashlist optionslisted below. It is expected that 'mount_root_vol' would first call this fcn.
##	    before checking or processing the given args, and then this fcn. would check for a "old-style" vs.
##	    "new-style" arglist.
##
##	    for "new-style" args:    This fcn. proceeds to completion, and returns the new mountpoint to the caller
##
##	    For "old-style" args:    This fcn. returns an error code to the caller which indicates an "old-style"
##	                             arglist was given, and the caller continues on its own to get the mountpoint.
##
## @arg %opts. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER
##		May also be either a Hostrec or Connectrec object.
##
##	   fc, fh  -- 'Alias options' for 'filer' option
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         client - hostspec of client. defaults to $main::CLIENT
##		May also be either a Hostrec or Connectrec object.
##
##	   localhost  -  Boolean, default is '0'. If '1', will use 'localhost' as client.
##                       Overrides 'client' option.
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         mountpoint - local mount point. defaults to /t/FILER_IP/ROOT_VOL
##
##         timeout   - Timeout in secs. for mount to complete. Defaults to 300 (5 min).
##
##         cmdtimeout   - CMDTIMEOUT. Defaults to $main::CMDTIMEOUT
##
##         chmod        - If given, then the octal code which will be used
##                         to 'chmod' the mountpoint
##
##	Returns
##		We expect to be called in an array context from 'mount_root_vol', and in a scalar context
##		from everywhere else. 'mount_root_vol' expects a 2-elt. list, ($error_code, $mountpoint),
##		with '$error_code' values as follows:
##
##			0	=>	We succeeded in mounting the root volume, and '$mountpoint'
##					contains a valid mountpoint for the root vol.
##
##			-1	=>	The arglist is not valid for 'new_mount_root_vol' so we gave up.
##					However, it might be valid for 'mount_root_vol', so that function
##					should continue processing the mount request.
##					The value of '$moupoint' is undefined.
##
##			other   =>	Some other error occured not related to arglist format. $@ should
##					contain error message, and 'mount_root_vol' should exit with this error.
##					The value of '$moupoint' is undefined.
##
##		For scalar context (everywhere except from 'mount_root_vol') a scalar is returned, which is
##		either a valid mountpoint for the root vol, or is the empty striing, or UNDEF if an error
##		occured during the mount attempt. In this case $@ should contain an appropriate error message.
##
sub new_mount_root_vol(@) {
    my $msg;
    my @allowed_opts = qw(filer client localhost filer_conn filer_ip
        mountpoint mountopts timeout cmdtimeout chmod );

    my $nargs = scalar(@_);
    if ( ( scalar(@_) % 2 ) != 0 ) {
        $msg
            = "new_mount_root_vol:  Arg list is not valid hash list -- odd number of args: $nargs.";

        #logDbg($msg);
        if (wantarray) {
            return ( -1, "" );
        } else {
            croak($msg);
        }
    } ## end if ( ( scalar(@_) % 2 ...

    #    for (my ($m,$n)=(0,0); ($n<=$#_); $m++,$n=2*$m)
    for ( my $n = 0; ( $n < $nargs ); $n += 2 ) {
        my $k = $_[$n];
        if ( !$k || ref($k) ) {
            $msg
                = "new_mount_root_vol:  Arg[$n] \"$k\" is undefined, NULL or a reference\n"
                . " -- not a valid option (keyword) for this fcn.";

            #logDbg($msg);
            if (wantarray) {
                return ( -1, "" );
            } else {
                croak($msg);
            }
        } ## end if ( !$k || ref($k) )
        if ( !grep( /^$k$/, @allowed_opts ) ) {
            $msg
                = "new_mount_root_vol:  Arg[$n] \"$k\" is not a valid option (keyword) for this fcn.";

            #logDbg($msg);
            if (wantarray) {
                return ( -1, "" );
            } else {
                croak($msg);
            }
        } ## end if ( !grep( /^$k$/, @allowed_opts...
    } ## end for ( my $n = 0; ( $n <...

    my (%opts) = @_;
    my %raw_opts = %opts;

    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    $raw_opts{client} = ref( $raw_opts{client} )
        if ( ref( $raw_opts{client} ) );
    logDbg(   "FS::Util::Setup::new_mount_root_vol - Enter Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my $FILER 
        = $opts{fc}
        || $opts{fh}
        || $opts{filer}
        || $main::FILER;    # Can be HostSpec, Hostrec, or Connectrec.
    my $FILER_CONN = $opts{filer_conn} || $main::FILER_CONN;
    my $CLIENT     = $opts{client}     || $main::CLIENT || 'localhost';
    my $FILER_IP   = $opts{filer_ip}   || $main::FILER_IP;
    my $MOUNTPOINT = $opts{mountpoint};
    my $MOUNTOPTS  = $opts{mountopts}
        || 'actimeo=0,noac,hard,intr,noquota';

    my $TIMEOUT    = $opts{timeout}    || $main::TIMEOUT;
    my $CMDTIMEOUT = $opts{cmdtimeout} || $main::CMDTIMEOUT;
    my $CHMOD = defined( $opts{chmod} ) ? $opts{chmod} : '0777';

    my $save_uid = undef;

    return eval {

        if ( $opts{localhost} ) {
            logDbg(
                "FS::Util::Setup::new_mount_root_vol -- 'localhost' option set, setting CLIENT='localhost'"
            );
            $opts{client} = 'localhost';
            $CLIENT = 'localhost';
        } ## end if ( $opts{localhost} )
        my %opts_copy = (%opts);
        if ( ref( $opts_copy{filer} ) ) {
            $opts_copy{filer} = sprintf( "%s", $opts_copy{filer} );
        }
        if ( ref( $opts_copy{client} ) ) {
            $opts_copy{client} = sprintf( "%s", $opts_copy{client} );
        }
        logDbg(   "FS::Util::Setup::new_mount_root_vol - Line: " 
                . __LINE__
                . "\n\nOpts:\n\n"
                . Dumper( \%opts_copy ) );

        $_root_vol_mounts{$CLIENT} = undef;

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "new_mount_root_vol   Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $FILER            = $opts{filer_conn};
            $opts{filer}      = $opts{filer_conn};
            $FILER_CONN       = $main::FILER_CONN;
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "FS::Util::Setup::new_mount_root_vol - Line: " 
                . __LINE__
                . "  Establish filer connection\n"
                . "  --  FILER: \"$FILER\"" );

        # filer connection
        my ( $fc, $fh );
        my $conn_timeout = $TIMEOUT;
        if ( ref($FILER) =~ /connectrec/i ) {
            $fc = $FILER;
        } else {
            $fc = make_conn( $FILER, $FILER_CONN );
            if ( !$fc ) {
                $msg
                    = "FS::Util::Setup::new_mount_root_vol - Could not establish filer connection: $@";
                logDbg($msg);
                croak($msg);
            } ## end if ( !$fc )
            if ( defined($CMDTIMEOUT) ) {
                $fc->timeout($CMDTIMEOUT);
            }
        } ## end else [ if ( ref($FILER) =~ /connectrec/i)
        $fh = $fc->hostp();
        my $hostname = $fh->hostname();
        my $filer_ip = $FILER_IP || $fh->default_ip();

        my ( $vol_path, $rootvol );
        my $pfiler = $fh->{PFILER};
        $save_uid = $>;
        if ($pfiler) {
            logDbg(
                "FS::Util::Setup::new_mount_root_vol - vfiler $hostname running on host $pfiler "
            );
            $vol_path = get_vfiler_root_path( $fc, $hostname );
        } else {

            # then pfiler client is calling
            $rootvol = vol_root($fc)
                or croak($@);
            logDbg(
                "FS::Util::Setup::new_mount_root_vol - root volume is $rootvol"
            );
            $vol_path = "/vol/$rootvol";
        } ## end else [ if ($pfiler)

        my $mountpoint = $MOUNTPOINT || "/t/$filer_ip/$rootvol";

        my $sec = qtreeSecurity( $fc, $vol_path );
        if ( ( $sec ne 'unix' ) && ( $sec ne 'mixed' ) ) {
            my $rtn = eval { qtreeSecurity( $fc, $vol_path, 'mixed' ) };
            logWarn("can't change $vol_path access to 'mixed: $@")
                unless ($rtn);
        }

        logDbg("FS::Util::Setup::new_mount_root_vol - Calling mount_vol");
        my $rtn_mntpt = mount_vol(
            filer      => $fc,
            client     => $CLIENT,
            vol_path   => $vol_path,
            mountpoint => $mountpoint,
            mountopts  => $MOUNTOPTS,
            timeout    => $TIMEOUT,
            chmod      => $CHMOD,
        );

        croak("FS::Util::Setup::new_mount_root_vol - $@")
            if ( $@ || !$rtn_mntpt );
        logDbg(
            "FS::Util::Setup::new_mount_root_vol -  Returned from mount_vol");
        push( @MOUNT_POINTS, $rtn_mntpt );

        $_root_vol_mounts{$CLIENT} = $rtn_mntpt;
        logDbg(
            "FS::Util::Setup::new_mount_root_vol - Exit _root_vol_mounts{$CLIENT}: \"$_root_vol_mounts{$CLIENT}\"\n"
                . " -- Returning mountpoint: \"$rtn_mntpt\"" );
        if (wantarray) {
            return ( 0, $rtn_mntpt );
        } else {
            return ($rtn_mntpt);
        }
    };
} ## end sub new_mount_root_vol(@)

## @name ::mount_root_vol_if_needed
##
## @summary Checks the global var (in this module) "$_root_vol_mountpoint".
##	    If set, then this mountpoint is returned. Otherwise, the fcn.
##	    'mount_root_vol_if_needed' is called to mount the root volume,
##	    and that mountpoint is returned.
##
##	    The options accepted by this fcn are similar to those of
##	    'new_mount_root_vol', except that this fcn does not take
##	    a 'mountpoint' option.
##
##	    All args are 'keyword' type (hashlist).
##          Returns: mount point
##
## @arg %opts. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER
##		May also be either a Hostrec or Connectrec object.
##
##	   fc, fh  -- 'Alias options' for 'filer' option
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         client - hostspec of client. defaults to $main::CLIENT
##		May also be either a Hostrec or Connectrec object.
##
##	   localhost  -  Boolean, default is '0'. If '1', will use 'localhost' as client.
##                       Overrides 'client' option.
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         timeout   - Timeout in secs. for mount to complete. Defaults to 300 (5 min).
##
##         cmdtimeout   - CMDTIMEOUT. Defaults to $main::CMDTIMEOUT
##
##         chmod        - If given, then the octal code which will be used
##                         to 'chmod' the mountpoint

sub mount_root_vol_if_needed(%) {
    my $msg;
    my @allowed_opts = qw(filer client localhost filer_conn filer_ip
        mountopts timeout cmdtimeout chmod );

    my $nargs = scalar(@_);
    if ( ( scalar(@_) % 2 ) != 0 ) {
        $msg
            = "FS::Util::Setup::mount_root_vol_if_needed:  Arg list is not valid hash list -- odd number of args: $nargs.";
        logDbg($msg);
        croak($msg);
    } ## end if ( ( scalar(@_) % 2 ...

    #    for (my ($m,$n)=(0,0); ($n<=$#_); $m++,$n=2*$m)
    for ( my $n = 0; ( $n < $nargs ); $n += 2 ) {
        my $k = $_[$n];
        if ( !$k || ref($k) ) {
            $msg
                = "FS::Util::Setup::mount_root_vol_if_needed - Arg[$n] \"$k\" is undefined, NULL or a reference\n"
                . " -- not a valid option (keyword) for this fcn.";
            logDbg($msg);
            croak($msg);
        } ## end if ( !$k || ref($k) )
        if ( !grep( /^$k$/, @allowed_opts ) ) {
            $msg
                = "FS::Util::Setup::mount_root_vol_if_needed - Arg[$n] \"$k\" is not a valid option (keyword) for this fcn.";
            logDbg($msg);
            croak($msg);
        } ## end if ( !grep( /^$k$/, @allowed_opts...
    } ## end for ( my $n = 0; ( $n <...

    my (%opts) = @_;
    my %raw_opts = %opts;

    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    $raw_opts{client} = ref( $raw_opts{client} )
        if ( ref( $raw_opts{client} ) );
    logDbg(   "FS::Util::Setup::mount_root_vol_if_needed -  Enter Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    $opts{filer} 
        = $opts{fc}
        || $opts{fh}
        || $opts{filer}
        || $main::FILER;    # Can be HostSpec, Hostrec, or Connectrec.

    $opts{filer_conn} = $opts{filer_conn} || $main::FILER_CONN;
    $opts{client}     = $opts{client}     || $main::CLIENT || 'localhost';
    $opts{filer_ip}   = $opts{filer_ip}   || $main::FILER_IP;

    $opts{mountopts} = $opts{mountopts}
        || 'actimeo=0,noac,hard,intr,noquota';

    $opts{timeout}    = $opts{timeout}    || $main::TIMEOUT;
    $opts{cmdtimeout} = $opts{cmdtimeout} || $main::CMDTIMEOUT;
    $opts{chmod} = defined( $opts{chmod} ) ? $opts{chmod} : '0777';

    eval {

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "FS::Util::Setup::mount_root_vol_if_needed -   Line: "
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $opts{filer}      = $opts{filer_conn};
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        if ( $opts{localhost} ) {
            logDbg(
                "FS::Util::Setup::mount_root_vol_if_needed -  'localhost' option set, setting CLIENT='localhost'"
            );
            $opts{client} = 'localhost';
        } ## end if ( $opts{localhost} )

        if ( $_root_vol_mounts{ $opts{client} } ) {
            logDbg(   "FS::Util::Setup::mount_root_vol_if_needed - Line: "
                    . __LINE__
                    . "   Returning existing mount: \"$_root_vol_mounts{$opts{client}}\""
            );
            return ( $_root_vol_mounts{ $opts{client} } );
        } else {
            logDbg(   "FS::Util::Setup::mount_root_vol_if_needed - Line: "
                    . __LINE__
                    . "   Calling new_mount_root_vol" );
            my $mntpt = new_mount_root_vol(%opts);
            if ( !$mntpt || $mntpt eq "-1" || $@ ) {
                my $msg
                    = "FS::Util::Setup::mount_root_vol_if_needed - 'new_mount_root_vol' failed: $@";
                logDbg($msg);
                croak($msg);
            } ## end if ( !$mntpt || $mntpt...
            logDbg(   "FS::Util::Setup::mount_root_vol_if_needed - Line: "
                    . __LINE__
                    . "   Returning new mount: \"$mntpt\"" );
            return ($mntpt);
        } ## end else [ if ( $_root_vol_mounts...
    };
} ## end sub mount_root_vol_if_needed(%)

## @name ::mount_vol
##
## @summary Mounts the volume on localhost (or CLIENT), when 'dir' option is omittted.
##	    If 'dir' option is given, then mounts the given directory for the given volume.
##          Returns: mount point
##
## @arg %opts. Optional. If not provided, values will be those of NATE
##             command line params.
##
##         filer - hostspec of filer. defaults to $main::FILER
##		May also be either a Hostrec or Connectrec object.
##
##         client - hostspec of client. defaults to $main::CLIENT
##		May also be either a Hostrec or Connectrec object.
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##         filer_ip - ip of filer. defaults to $main::FILER_IP
##
##         vol          - Required. Name of the filer volume to mount, or else path (on filer)
##                         to the volume (or volume/dir) to be mounted. Defaults to $main::FILER_VOL.
##                        Examples:
##                                     vol1   /vol/vol1    /vol/vol1/qtree1
##
##         vol_path, filer_vol  -- 'alias' options for the 'vol' option.
##
##	   dir - name of directory which is a direct decsendent of the parent volume.
##               If given, this directory will be mounted rather than the entire volume.
##		 default is undef, so entire volume is mounted. The resulting 'vol_path'
##	         will then be the concatenation:  "/" . "vol" . <VOLUME> . "/" <DIR>
##
##         filer_vol_mount - filer mount target. defaults to
##                           $main::FILER_VOL_MOUNT
##
##         mountpoint - local mount point. defaults to $main::MOUNTPOINT
##
##         timeout   - Timeout in secs. for mount to complete. Defaults to 300 (5 min).
##
##

sub mount_vol(%) {
    my (%opts) = @_;
    my %raw_opts = %opts;

    if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i ) {
        my $logmsg
            = "FS::Util::Setup::mount_vol - Line: " 
            . __LINE__
            . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
            . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
            . "Will set 'filer' option to this object, and delete 'filer_conn' option";

        #		logDbg($logmsg);
        $opts{filer}      = $opts{filer_conn};
        $opts{filer_conn} = $main::FILER_CONN;
    } ## end if ( $opts{filer_conn}...
    $raw_opts{filer} = ref( $raw_opts{filer} ) if ( ref( $raw_opts{filer} ) );
    $raw_opts{filer_conn} = ref( $raw_opts{filer_conn} )
        if ( ref( $raw_opts{filer_conn} ) );
    $raw_opts{client} = ref( $raw_opts{client} )
        if ( ref( $raw_opts{client} ) );
    logDbg(   "FS::Util::Setup::mount_vol - Enter Line: " 
            . __LINE__
            . "    Opts:\n\n"
            . Dumper( \%raw_opts ) );

    my $FILER 
        = $opts{fc}
        || $opts{fh}
        || $opts{filer}
        || $main::FILER;    # Can be HostSpec, Hostrec, or Connectrec.
    my $FILER_CONN = $opts{filer_conn} || $main::FILER_CONN;
    my $VOL_PATH 
        = $opts{vol_path}
        || $opts{vol}
        || $opts{filer_vol}
        || $main::FILER_VOL;
    my $MOUNTPOINT = $opts{mountpoint};
    my $CLIENT     = $opts{client} || $main::CLIENT || 'localhost';
    my $FILER_IP   = $opts{filer_ip} || $main::FILER_IP;
    my $MOUNTOPTS 
        = $opts{mountopts}
        || $main::MOUNTOPTS
        || 'actimeo=0,noac,hard,intr,noquota';

    my $TIMEOUT    = $opts{timeout}    || $main::TIMEOUT;
    my $CMDTIMEOUT = $opts{cmdtimeout} || $main::CMDTIMEOUT;
    my $CHMOD = defined( $opts{chmod} ) ? $opts{CHMOD} : '0777';
    my $DIR = $opts{dir};

    return eval {
        my %opts_copy = (%opts);
        if ( ref( $opts_copy{filer} ) ) {
            $opts_copy{filer} = sprintf( "%s", $opts_copy{filer} );
        }
        if ( ref( $opts_copy{client} ) ) {
            $opts_copy{client} = sprintf( "%s", $opts_copy{client} );
        }
        logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                . __LINE__
                . "\n\nOpts:\n\n"
                . Dumper( \%opts_copy ) );

        if ( $opts{filer_conn} && ref( $opts{filer_conn} ) =~ /connectrec/i )
        {
            my $logmsg
                = "FS::Util::Setup::mount_vol - Line: " 
                . __LINE__
                . "\nA 'CONNECTREC' was passed via the 'filer_conn' option.\n"
                . "Expected a 'Connspec' string ('#rsh', '#telnet', etc)\n"
                . "Will set 'filer' option to this object, and delete 'filer_conn' option";

            #		logDbg($logmsg);
            $FILER            = $opts{filer_conn};
            $opts{filer}      = $opts{filer_conn};
            $FILER_CONN       = $main::FILER_CONN;
            $opts{filer_conn} = $main::FILER_CONN;
        } ## end if ( $opts{filer_conn}...

        logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                . __LINE__
                . "  Establish filer connection\n"
                . "  --  FILER: \"$FILER\"" );

        # filer connection
        my ( $fc, $fh );
        my $conn_timeout = $TIMEOUT;
        if ( ref($FILER) =~ /connectrec/i ) {
            $fc = $FILER;
        } else {
            $fc = make_conn( $FILER, $FILER_CONN );
            if ( !$fc ) {
                logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                        . __LINE__
                        . "   Could not establish filer connection: $@" );
                die("mount_vol:  Could not establish filer connection: $@");
            } ## end if ( !$fc )
        } ## end else [ if ( ref($FILER) =~ /connectrec/i)
        $fh = $fc->hostp();
        my $hostname = $fh->hostname();
        my $filer_ip = $FILER_IP || $fh->default_ip();
        if ( defined($CMDTIMEOUT) ) {
            $fc->timeout($CMDTIMEOUT);
        }

    # client connection/host
    # XXX/tapl Arvind advocate to always use #rsh for clients
    # -- we never use '$cc', so comment out next few lines
    #my $cc = make_conn($CLIENT, '#rsh');
    #$cc->timeout($CMDTIMEOUT);
    # We can use host object to pass to mount::mount() (but not a Connectrec).
    #my $ch = $cc->hostp();

        logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                . __LINE__
                . "  Establish client connection\n"
                . "  --  CLIENT: \"$CLIENT\"" );

    # client connection
    #
    # Use either 'localhost' or $ch (HOSTREC) as arg to mount( client => ... )
    #
        my ( $cc, $ch );
        if ( $CLIENT eq 'localhost' ) {
            $ch = 'localhost';
        } else {
            if ( ref($CLIENT) =~ /connectrec/i ) {
                $cc = $CLIENT;
            } else {
                $cc = make_conn( $CLIENT, '#rsh' );
                if ( !$cc ) {
                    logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                            . __LINE__
                            . "   Could not establish client connection: $@"
                    );
                    die("mount_vol:  Could not establish client connection: $@"
                    );
                } ## end if ( !$cc )
            } ## end else [ if ( ref($CLIENT) =~ /connectrec/i)
            $ch = $cc->hostp();
            if ( defined($CMDTIMEOUT) ) {
                $cc->timeout($CMDTIMEOUT);
            }
        } ## end else [ if ( $CLIENT eq 'localhost')

        if ( !is_licensed( $fc, "nfs" ) ) {
            logTrc("license nfs on filer");
            license( $fc, "nfs" ) or croak $@;
        }

        my ( $mntpt, $vol, $vol_path, $dir );
        if ( !$VOL_PATH ) {
            croak(
                "FS::Util::Setup::mount_vol:  One of 'vol_path', or 'vol' options must be defined\n"
                    . "with name or path of volume to be mounted" );
        }

        if ( $VOL_PATH =~ m%^([^/]+)$% ) {

            # 'vol name' format
            $vol      = $1;
            $vol_path = "/vol/$vol";
            $vol_path .= "/$DIR" if ($DIR);
            $dir = $DIR;
        } elsif ( $VOL_PATH =~ m%^(/vol/([^/]+))$% ) {

            # 'vol path' format, no subdir
            $vol_path = $1;
            $vol      = $2;
            $vol_path .= "/$DIR" if ($DIR);
            $dir = $DIR;
        } elsif ( $VOL_PATH =~ m%^(/vol/([^/]+)/([^/]+))$% ) {

            # 'vol path' format, with subdir
            $vol_path = $1;
            $vol      = $2;
            $dir      = $3;
        } else {
            croak(
                "new_mount_root_vol:  Invalid format for vol/vol_path option: \"$VOL_PATH\""
            );
        }

        my $vol_path_tail = $vol_path;
        $vol_path_tail =~ s%^/vol/%%;
        $mntpt = $MOUNTPOINT || "/t/$filer_ip/$vol_path_tail";

        logDbg(   "FS::Util::Setup::mount_vol -  Line: " 
                . __LINE__
                . "\nVOL_PATH: \"$VOL_PATH\"\nvol_path: \"$vol_path\"\n"
                . "vol: \"$vol\"\nDIR: \"$DIR\"\ndir: \"$dir\"\n"
                . "vol_path_tail:\"$vol_path_tail\"\nmntpt: \"$mntpt\"" );

        logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                . __LINE__
                . "\n\nvol_path: \"$vol_path\"" );
        my $qtpath;
        if ($dir) {
            $qtpath = $vol_path;
            if ( !eval { qtreePresent( $fc, $vol_path ) } ) {
                $qtpath = "/vol/$vol";
            }
        } else {
            $qtpath = "/vol/$vol";
        }
        my $sec = eval { qtreeSecurity( $fc, $qtpath ) };
        if ( ( $sec ne 'unix' ) && ( $sec ne 'mixed' ) ) {
            my $rtn = eval { qtreeSecurity( $fc, $qtpath, 'mixed' ) };
            logWarn("can't change $qtpath access from '$sec' to 'mixed': $@")
                unless ($rtn);
        }
        logDbg(   "FS::Util::Setup::mount_vol - Line: " 
                . __LINE__
                . "\n\nvol_path: \"$vol_path\"\nqtree path: \"$qtpath\"\n"
                . "qtree sec: \""
                . eval { qtreeSecurity( $fc, $qtpath ) }
                . "\"" );

        export_path( $fc, "$vol_path" )
            or croak("FS::Util::Setup::mount_vol - $@");

        logTrc("mount_vol: \"$vol\"  mountpoint: \"$mntpt\"");
        my %mount_subr_opts = (
            filer_ip   => $filer_ip,
            vol_path   => $vol_path,
            client     => $ch,
            mountpoint => $mntpt,
            mountopts  => $MOUNTOPTS,
            vers       => $main::MOUNT_VERS,
            chown      => 1,
            chmod      => $CHMOD,
        );
        logDbg( "FS::Util::Setup::mount_vol - Calling mount::mount\n\n"
                . Dumper( _strip_filer( \%mount_subr_opts ) ) );
        my $mntpt_returned = mount::mount(
            $filer_ip, $vol_path,
            client     => $ch,
            mountpoint => $mntpt,
            mountopts  => $MOUNTOPTS,
            vers       => $main::MOUNT_VERS,
            chown      => 1,
            chmod      => $CHMOD,
        );

        logDbg(
            "FS::Util::Setup::mount_vol - \"$vol_path\" returned error: $@")
            if ( $@ || !$mntpt_returned );
        croak("mount_vol \"$vol_path\" returned error: $@")
            if ( $@ || !$mntpt_returned );
        logDbg("FS::Util::Setup::mount_vol -   Returned from mount::mount");
        logTrc("FS::Util::Setup::mount_vol - returning $mntpt_returned");
        return ($mntpt_returned);
    };
} ## end sub mount_vol(%)

## @name ::umount_vol
## @summary unmount volume. returns 1 on success.
## @arg local_mnt_pt - what to unmount
## @arg %args. Optional. If not provided, values will be those of NATE command
##             line params.
##
##         client - hostspec of client. defaults to $main::CLIENT
##
##	NOT_MOUNTED_OKAY:	If '1', then don't complain if the volume is not
##				mounted when this function is called. If '0', then
##				issue a warning if the volume is not currently mounted.
##				and return (0).
##				Default is '1', unless NOT_MOUNTED_OKAY is set.
##				-- return (1) silently if the volume is not mounted.
##
##	NOT_MOUNTED_TEST:	Mutually exclusive with NOT_MOUNTED_OKAY. If '1' and if volume
##				is not mounted at start of test, return (0). Unlike NOT_MOUNTED_OKAY,
##				no warning is issued in this case. Default is '0'.
##
sub umount_vol($;%) {
    my ( $local_mnt_pt, %opts ) = @_;
    my $CLIENT = $opts{client} || $main::CLIENT || 'localhost';
    my $FORCE = defined( $opts{force} ) ? $opts{force} : 1;
    my $NOT_MOUNTED_TEST
        = defined( $opts{not_mounted_test} ) ? $opts{not_mounted_test} : 0;
    my $NOT_MOUNTED_OKAY
        = defined( $opts{not_mounted_okay} )
        ? $opts{not_mounted_okay}
        : !$NOT_MOUNTED_TEST;

    return eval {
        umount( $local_mnt_pt, client => $CLIENT, force => $FORCE );
        my $msg = $@;
        if ( !$msg ) {
            return 1;
        } elsif ( $msg =~ /not mounted|not in mnt/ ) {
            if ($NOT_MOUNTED_OKAY) {
                return 1;
            } elsif ($NOT_MOUNTED_TEST) {
                return 0;
            } else {
                logWarn(
                    "FS::Util::Setup::umount_vol - $local_mnt_pt is not mounted"
                );
                return 0;
            } ## end else [ if ($NOT_MOUNTED_OKAY)
        } else {
            croak("FS::Util::Setup::umount_vol - $@");
        }
    };
} ## end sub umount_vol($;%)

## @name ::create_and_map_cifs_share
## @summary creates cifs share on filer, and maps it as drive from cifs client
##          returns letter of mapped drive. for ex. X:
## @arg %args. Optional. If not provided, values will be those of NATE command line params.
##           These can take values of:
##
##         cifs_client - hostspec of client. defaults to $main::CIFS_CLIENT
##
##         cifs_drive - letter of the drive that should be mapped. defaults to $main:CIFS_DRIVE
##
##         user - name of cifs user which will be used for drive mapping.
##                     subsequently, all the files created on that drive will belong
##                     cifs_user. defaults to $main::CIFS_USER
##
##         passwd - password that should be provided when drive is mapped.
##                  defaults to $main::PASSWD
##
##         cifs_share - name of cifs share to create. defaults to $main::CIFS_SHARE
##
##         vol_name - volume name. defaults to $main::FILER_VOL
##
##         security - security type for cifs_share. mixed|nfs|cifs.
##		      defaults to 'mixed'.
##
## 	   timeout - timeout value for the connection object
##		     defaults to $main::TIMEOUT
##
## 	   cmdtimeout - timeout value for waiting for a response from a
##                      connection. defaults to $main::CMDTIMEOUT
##         purge_existing - is introduced to create_and_map_cifs_share() in
##                          order to handle the case when the share is
##                          existing. If peruge_existing=1,
##                          create_and_map_cifs_share() will purge and
##                          recreate the cifs shares.
sub create_and_map_cifs_share(%) {
    my (%args) = @_;

    my $cifs_client = $args{cifs_client}    || $main::CIFS_CLIENT;
    my $cifs_drive  = $args{drive}          || $main::CIFS_DRIVE;
    my $user        = $args{user}           || $main::CIFS_USER;
    my $passwd      = $args{passwd}         || $main::CIFS_PASSWD;
    my $filer       = $args{filer}          || $main::FILER;
    my $cifs_share  = $args{cifs_share}     || $main::CIFS_SHARE;
    my $vol_name    = $args{vol_name}       || $main::FILER_VOL;
    my $security    = $args{qtree_security} || 'mixed';
    my $TIMEOUT     = $args{timeout}        || $main::TIMEOUT;
    my $CMDTIMEOUT  = $args{cmdtimeout}     || $main::CMDTIMEOUT;
    my $filerconn   = $args{filerconn}      || $main::FILER_CONN;

    my $purge_existing = $args{purge_existing} ? $args{purge_existing} : 0;

    return eval {

        logTrc("checking if client is minwin compatible");
        minwin_compatible($cifs_client)
            or croak($@);

        logTrc("attempting to do cifs setup.. ");
        setup_cifs( $filer, $passwd );

        logTrc("cifs setup complete .. ");

        my $fh = host($filer);
        my $fc = connect( $filerconn, -timeout, $TIMEOUT );
        $fc->timeout($CMDTIMEOUT);

        logTrc("add new share $cifs_share on $vol_name");
        share_add(
            $fc, $cifs_share, "/vol/$vol_name",
            verify         => 1,
            umask          => 0,
            purge_existing => $purge_existing
        ) or croak("failed to export share $cifs_share");

        logDbg(
            "FS::Util::Setup::create_and_map_cifs_share - added share /vol/$vol_name"
        );

        qtreeSecurity( $fc, "/vol/$vol_name", $security )
            or croak("can't change /vol/$vol_name access to $security: $@");

        logDbg(
            "FS::Util::Setup::create_and_map_cifs_share - set qtree security to $security"
        );
        my $host = $fc->connhost();
        $cifs_drive = map_cifs_share(
            server   => $host,
            share    => $cifs_share,
            client   => $cifs_client,
            drive    => $cifs_drive,
            username => $user,
            password => $passwd
        ) or croak($@);
        logDbg(
            "FS::Util::Setup::create_and_map_cifs_share - mapped drive $cifs_drive"
        );

        if ( !$cifs_drive ) {
            croak(
                "FS::Util::Setup::create_and_map_cifs_share - could not map the drive"
            );
        } ## end if ( !$cifs_drive )

        $cifs_drive .= ":" unless $cifs_drive =~ /:$/;
        logDbg(
            "FS::Util::Setup::create_and_map_cifs_share - Share $cifs_share of $host mounted as $cifs_drive"
        );

        return $cifs_drive;
    };
} ## end sub create_and_map_cifs_share(%)

## @name ::common_cleanup
## @summary Performs common cleanup functions depending on $CLEANUP (or passed in value in %args)
## @arg $conn Required: The filer connection object
## @arg %args - optional. If not provided, values will be those of NATE command line params.
##          These can have the following values:
##
##  cleanup   - none|test|volumes|all. defaults to $main::CLEANUP
##
##  vol       - volume name. defaults to undef. $main::FILER_VOL will still
##              be destroyed when $main::CLEANUP is 'volumes' or 'all.
##
##  vol_list  - a reference to a list of volume names to be destroyed.
##                - defaults to undef.
##                - the list should be sorted by volume age, earliest first, because the volumes
##                will be destroyed in reverse order.
##                - when $main::CLEANUP is 'test', $main::FILER_VOL will be skipped even if it is
##                explicitly specified in the given list.
##                - when $main::CLEANUP is 'volumes' or 'all', $main::FILER_VOL will always be
##                destroyed, regardless of whether it appears in the given list.
##
##  aggr      - aggregate name. defaults to undef. $main::FILER_AGGR will
##              still be destroyed when $main::CLEANUP is 'volumes' or 'all.
##
##  aggr_list  - a reference to a list of aggregate names to be destroyed.
##                - defaults to undef.
##                - when $main::CLEANUP is 'test', $main::FILER_AGGR will be skipped even if it is
##                explicitly specified in the given list.
##                - when $main::CLEANUP is 'aggregates' or 'all', $main::FILER_AGGR will always be
##                destroyed, regardless of whether it appears in the given list.
##
##  mnt       - mount point.
##
##  client    - hosspec of client. defaults to $main::CLIENT
##
##  timeout   - Timeout in secs. for vol_offline/vol_purge to complete. Defaults to 300 (5 min).
##
##  vol_type  - flex|trad. defaults to $main::VOL_TYPE
##
##  snaplock  - flag requesting cleanup of snaplock stuff:
##                - removes snaplock compliance licensing
##                - destroys $main::WORM_VOL and/or $main::WORM_AGGR
##
##  snaplock_enterprise
##            - flag requesting cleanup of snaplock stuff:
##                - removes snaplock enterprise licensing
##                - destroys $main::WORM_VOL and/or $main::WORM_AGGR
##

## XXX/tapl Still to be decided on whether the parameters 'vol' or
##          'aggr' is of any use... 10/13/2004

sub common_cleanup($;%) {
    my $connp = shift;
    unless ( defined($connp) ) {
        logresult( "SCRIPT", "common_cleanup: undefined \$connp" );
    }
    my $conn = make_conn_and_complain($connp);
    my %args = @_;

    my $cleanup   = $args{cleanup}   || $main::CLEANUP;
    my $vol_name  = $args{vol}       || undef;
    my $aggr_name = $args{aggr}      || undef;
    my $vol_list  = $args{vol_list}  || undef;
    my $aggr_list = $args{aggr_list} || undef;
    my $mnt       = $args{mnt};
    my $timeout   = $args{timeout}   || 300;
    my $client    = $args{client}    || $main::CLIENT;
    my $vol_type  = $args{vol_type}  || $main::VOL_TYPE;
    my ( $ret_val, $err_msg );

    if ( defined($vol_name) && !defined($vol_list) ) {
        $vol_list = [$vol_name];
        $vol_name = undef;
    } else {

        # purge volumes in reverse creation order
        @{$vol_list} = reverse @{$vol_list};
    }
    if ( defined($aggr_name) && !defined($aggr_list) ) {
        $aggr_list = [$aggr_name];
        $aggr_name = undef;
    } else {

# no need to destroy aggregates in reverse order - is there??? (do it anyways,
# for consistency with vol_list behavior)
        @{$aggr_list} = reverse @{$aggr_list};
    } ## end else [ if ( defined($aggr_name...

    # Reset wafl scan speed to default value
    ( $ret_val, $err_msg ) = set_wafl_scan_speed(
        conn  => $conn,
        speed => DEFAULT_WAFL_SCAN_SPEED
    );
    if ( $ret_val ne 1 ) {
        logWarn($err_msg);
    }


    return eval {
        if ( $cleanup eq 'none' )
        {
            return 1;

        } elsif ( $cleanup eq 'test' ) {
            foreach my $vol_name (@$vol_list) {
                my $vol_obj = vol_status($conn, $vol_name);
                my $vol_status = $vol_obj->state();
                if ($vol_status =~ /iron_restricted/i) {
                    my $aggr_name;
                    if ($vol_obj->type() eq 'flex') {
                        $aggr_name = $vol_obj->parent();
                    } else {
                        $aggr_name = $vol_name;
                    }
                    my $status = wafliron_status($conn, $aggr_name);
                    while ($status eq '1') {
                        $status = wafliron_status($conn, $aggr_name);
                    }
                    if ($status eq '2') {
                        logWarn("Aggregate $aggr_name is in iron_restricted state"
                               ." Committing the changes to cleanup agregate for next run");
                        wafliron_commit($conn, $aggr_name);
                        aggr_online($conn, $aggr_name);
                    }
                } ## end if ($vol_status =~ /iron_restricted/i)

            } ## end foreach my $vol_name (@$vol_list)
            foreach my $aggr_name (@$aggr_list) {

                my $agg_state = aggr_status($conn, $aggr_name)->state();
                if ($agg_state =~ /iron_restricted/i) {
                    my $status;
                    do {
                        $status = wafliron_status($conn, $aggr_name);
                    } until ($status ne '1');
                    if ($status eq '2') {
                        logWarn("Aggregate $aggr_name is in iron_restricted state"
                               ." Committing the changes to cleanup agregate for next run");

                        wafliron_commit($conn, $aggr_name);
                        aggr_online($conn, $aggr_name);
                    }
                } ## end if ($agg_state =~ /iron_restricted/i)

            } ## end foreach my $aggr_name (@$aggr_list)

            # purge volumes as specified by caller
            _purge_vol_list( $conn, $vol_list );
            _purge_aggr_list( $conn, $aggr_list );
        } elsif ( ( $cleanup =~ /^volume/ ) or ( $cleanup eq 'all' ) ) {
            foreach my $vol_name (@$vol_list) {
                my $vol_obj = vol_status($conn, $vol_name);
                my $vol_status = $vol_obj->state();
                if ($vol_status =~ /iron_restricted/i) {
                    my $aggr_name;
                    if ($vol_obj->type() eq 'flex') {
                        $aggr_name = $vol_obj->parent();
                    } else {
                        $aggr_name = $vol_name;
                    }
                    my $status = wafliron_status($conn, $aggr_name);
                    while ($status eq '1') {
                        $status = wafliron_status($conn, $aggr_name);
                    }
                    if ($status eq '2') {
                        logWarn("Aggregate $aggr_name is in iron_restricted state"
                               ." Committing the changes to cleanup agregate for next run");
                        wafliron_commit($conn, $aggr_name);
                        aggr_online($conn, $aggr_name);
                    }
                } ## end if ($vol_status =~ /iron_restricted/i)

            } ## end foreach my $vol_name (@$vol_list)
            foreach my $aggr_name (@$aggr_list) {

                my $agg_state = aggr_status($conn, $aggr_name)->state();
                if ($agg_state =~ /iron_restricted/i) {
                    my $status;
                    do {
                        $status = wafliron_status($conn, $aggr_name);
                    } until ($status ne '1');
                    if ($status eq '2') {
                        logWarn("Aggregate $aggr_name is in iron_restricted state."
                               ." Committing the changes to cleanup agregate for next run");

                        wafliron_commit($conn, $aggr_name);
                        aggr_online($conn, $aggr_name);
                    }
                } ## end if ($agg_state =~ /iron_restricted/i)

            } ## end foreach my $aggr_name (@$aggr_list)

            # XXX/tapl there's some confusion here in which
            # whether 'volume' or 'volumes' is the correct
            # value. anyway, i changed the condition to just
            # grep for the word 'volume' to handle both case.
            # this WILL be resolved once the CLEANUP defintion
            # has been cleared up.
            #
            if ($main::FILER_VOL) {

                # add FILER_VOL unless its already there
                my $already_in_vol_list = grep $_ eq $main::FILER_VOL,
                    @{$vol_list};
                if ( $main::FILER_VOL && ( !$already_in_vol_list ) ) {

                    # add FILER_VOL to the vol_list
                    push( @{$vol_list}, $main::FILER_VOL );
                }
            } ## end if ($main::FILER_VOL)

            my $snaplock_destroy = $args{snaplock}
                || $args{snaplock_enterprise};

            if ( $snaplock_destroy && vol_exists( $conn, $main::WORM_VOL ) ) {

                # add WORM_VOL unless its already there
                my $already_in_vol_list = grep $_ eq $main::WORM_VOL,
                    @{$vol_list};
                if ( $main::WORM_VOL && ( !$already_in_vol_list ) ) {

                    # add WORM_VOL to the vol_list
                    push( @{$vol_list}, $main::WORM_VOL );
                }
            } ## end if ( $snaplock_destroy...

            _purge_vol_list( $conn, $vol_list );

            # some more stuff to do for 'all' case with flex volumes
            if ( ( $cleanup eq 'all' ) && ( $vol_type eq 'flex' ) ) {

                # if VOL_TYPE is flex, then we have to destroy $FILER_AGGR too
                my $already_in_aggr_list = grep $_ eq $main::FILER_AGGR,
                    @{$aggr_list};
                if ( $main::FILER_AGGR && ( !$already_in_aggr_list ) ) {

                    # add FILER_AGGR
                    push( @{$aggr_list}, $main::FILER_AGGR );
                }

                if ( $snaplock_destroy
                    && aggr_exists( $conn, $main::WORM_AGGR ) )
                {

                    # add WORM_AGGR (unless it is already there)
                    $already_in_aggr_list = grep $_ eq $main::WORM_AGGR,
                        @{$aggr_list};
                    if ( !$already_in_aggr_list ) {
                        push( @{$aggr_list}, $main::WORM_AGGR );
                    }

                } ## end if ( $snaplock_destroy...
            } ## end if ( ( $cleanup eq 'all'...

            _purge_aggr_list( $conn, $aggr_list );

            # delete snaplock licensing?
            unlicense( $conn, 'snaplock' ) if ( $args{snaplock} );
            unlicense( $conn, 'snaplock_enterprise' )
                if ( $args{snaplock_enterprise} );

        } else {
            die "invalid cleanup value $cleanup";
        }

        _delete_cifs_shares($conn);

        # BURT 427608 - change of default aggr snap resv from 5% to 0%
        SnapCmds::restore_saved_aggr_snap_reserve($conn);

        if ($mnt) {
            umount_vol(
                $mnt,
                client           => $client,
                force            => 1,
                not_mounted_okay => 1
            );
        } ## end if ($mnt)
        foreach my $leftover_mount (@MOUNT_POINTS) {
            if (umount_vol(
                    $leftover_mount,
                    client           => $client,
                    force            => 1,
                    not_mounted_test => 1
                )
                )
            {

# Warn about leftover mounts except for vol0
# vol0 is usually mounted by a library routine such as Wafliron.pm and is not really
# an issue with the script itself
                logWarn("cleaned up extraneous mountpoint $leftover_mount")
                    if ( $leftover_mount !~ /vol0/ );
            } ## end if ( umount_vol( $leftover_mount...
        } ## end foreach my $leftover_mount ...

        return 1;
    };
} ## end sub common_cleanup($;%)

sub _delete_cifs_shares($) {
    my ($fc) = @_;
    my $fh = $fc->hostp();
    logDbg("FS::Util::Setup::_delete_cifs_shares - Enter");

    my @well_known_shares = qw(ETC$ C$ HOME);
    my $terminate_cifs    = 0;
    my $status;
    if ( !is_cifs_setup( host => $fh ) ) {

        #		$status = eval {filer_cifs_start($fc) };
        logDbg(
            "FS::Util::Setup::_delete_cifs_shares - CIFS not running on filer -- exiting"
        );
        return;
    } ## end if ( !is_cifs_setup( host...
    my %shares      = filer_cifs_shares_get($fc);
    my @share_names = keys(%shares);
    foreach my $share_name (@share_names) {
        if ( grep( { $share_name eq $_ } @well_known_shares ) ) {
            logDbg(
                "FS::Util::Setup::_delete_cifs_shares - Not deleting 'well known share' \"$share_name\""
            );
            next;
        } ## end if ( grep( { $share_name...
        logDbg(
            "FS::Util::Setup::_delete_cifs_shares - Deleting share \"$share_name\""
        );
        eval { share_delete( $fc, $share_name ) };
        logWarn($@) if ($@);
    } ## end foreach my $share_name (@share_names)
    if ($terminate_cifs) {
        eval { filer_cifs_stop($fc) };
    }
    logDbg("FS::Util::Setup::_delete_cifs_shares - Exit");
} ## end sub _delete_cifs_shares($)

sub _delete_volume_snap_schedule($$) {
    my ( $fc, $vol ) = @_;
    snapsched_volume( $fc, $vol, 0, 0 );
}

sub _purge_aggr_list($$) {
    my ( $conn, $aggr_list ) = @_;

    return if ( $main::CLEANUP eq 'volume' || $main::CLEANUP eq 'volumes' );
    if ( defined($aggr_list) ) {
        $Log->trace( "\n\n*****************\n\n********* aggr_list is \n"
                . Dumper($aggr_list)
                . "\n\n******************\n\n" );
        foreach my $aggr_name (@$aggr_list) {
            my $cleanup = $main::CLEANUP;
            if (   ( ( $cleanup eq 'test' ) || ( $cleanup =~ /^volume/ ) )
                && ( $aggr_name eq $main::FILER_AGGR ) )
            {
                $Log->comment(
                    "FS::Util::Setup::_purge_aggr_list - skipping FILER_AGGR ($aggr_name) because CLEANUP=$cleanup"
                );
            } else {
                aggr_purge( $conn, $aggr_name )
                    or croak $@;
            }
        } ## end foreach my $aggr_name (@$aggr_list)
    } ## end if ( defined($aggr_list...
} ## end sub _purge_aggr_list($$)

sub _purge_vol_list($$) {
    my ( $conn, $vol_list ) = @_;

    die "empty vol_list" if ( !defined($vol_list) );

    $Log->trace( "\n\n*****************\n\n********* vol_list is \n"
            . Dumper($vol_list)
            . "\n\n******************\n\n" );
    foreach my $vol_name (@$vol_list) {
        if (   ( $main::CLEANUP eq 'test' )
            && ( $vol_name eq $main::FILER_VOL ) )
        {
            $Log->comment(
                "FS::Util::Setup::_purge_vol_list - skipping FILER_VOL ($vol_name) because CLEANUP=test"
            );
        } else {

       # 11/24/05 bobj -- It turns out that cleanup is failing in some tests
       # with vol_purge giving this message:
       #     vol offline: Volume is in use by other operations
       # We are seeing this in quota tests, where quotas were turned off on
       # a volume before common_cleanup() is run. "quota status" says that
       # quotas are off for the volume, but actually the "quota off" operation
       # hasn't finished, quota metadata files are still being truncated.
            my $MAX_RETRIES = 5;
            my $RETRY_PAUSE = 60;
            my $last_exception;
            for ( my $try = 0; $try <= $MAX_RETRIES; $try++ ) {

                # clear this before each try
                $last_exception = undef;

                logTrc(
                    "FS::Util::Setup::_purge_vol_list - Attempt to delete volume '$vol_name'"
                );
                if ( !vol_purge( $conn, $vol_name ) ) {

                   # capture the most recent exception and log a debug message
                    $last_exception = $@;
                    logDbg(
                        "FS::Util::Setup::_purge_vol_list - $last_exception");
                } ## end if ( !vol_purge( $conn...
                last
                    unless ( vol_exists( $conn, $vol_name ) );
                sleep($RETRY_PAUSE);
            } ## end for ( my $try = 0; $try...
            if ( vol_exists( $conn, $vol_name ) ) {
                die
                    "FS::Util::Setup::_purge_vol_list - Can't delete volume '$vol_name': $last_exception"
                    if ( $last_exception ne UNSUPPORTED_FEATURE_EXCEPTION );
                logWarn(UNSUPPORTED_FEATURE_EXCEPTION);
            } elsif ( defined($last_exception) ) {
                die
                    "FS::Util::Setup::_purge_vol_list - caught exception while deleting volume '$vol_name': $last_exception"
                    if ( $last_exception ne UNSUPPORTED_FEATURE_EXCEPTION );
                logWarn(UNSUPPORTED_FEATURE_EXCEPTION);
            } else {
                logTrc(
                    "FS::Util::Setup::_purge_vol_list - Deleted volume '$vol_name'"
                );
            }
        } ## end else [ if ( ( $main::CLEANUP ...
    } ## end foreach my $vol_name (@$vol_list)
} ## end sub _purge_vol_list($$)

## @name ::create_matching_container ($conn, $container_name, $container_size, %opts)
##
## @summary Create a container of type trad or aggr
##
## @description
##          Create a container with the specified size by
##          choosing the disks appropriately.
##
## Steps
##
## 1. Find the appropriate disks for the container to be created with
##   the requested space
##
## 2. Create the disk by specifying the disks
##
## 3. Return the result
##
## Assumptions
##
## See get_matching_spare_disks
##
## @arg $conn Required. The connection to the filer host,
##            where container to be created.
##
## @arg $conr_name Required. Name of the container to be created.
##
## @arg $conr_size Required. The size of the container to be created.
##
## @arg %opts This is hash with the following options
##
## @arg conr_type The type of the container. Default 'trad'
##
## @arg raid_type The type of raid. Default is decided by the
##                conr_type. If conr_type is "trad", it is "raid4".
##                Where as when conr_type is "aggr", it is "raid_dp"
##

sub create_matching_container ($$$;@) {
    my $conn = make_conn_and_complain(shift);
    my ( $conr_name, $conr_size, %opts ) = validate_mixed(
        @_,
        [ { type => SCALAR }, { type => SCALAR }, ],
        {   conr_type => {
                type     => SCALAR,
                optional => 1,
                default  => "trad",
                regex    => qr/trad|aggr/i
            },
            raid_type => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/raid4|raid_dp/i
            },
            raidtype => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/raid4|raid_dp/i
            },
        }
    );

    return eval {

        my $conr_type = $opts{conr_type};

        # The default raidtype is decided by conr_type, when it is
        # not given in the argument
        my $raid_type = ( $opts{raid_type} || $opts{raidtype} );
        if ( !defined($raid_type) ) {
            if ( $opts{conr_type} eq "trad" ) {
                $raid_type = "raid4";
            } else {
                $raid_type = "raid_dp";
            }
        } ## end if ( !defined($raid_type...

        my %vol_create_args = ();
        $vol_create_args{raidtype} = $raid_type;

        logDbg(
            "FS::Util::Setup::create_matching_container -FH=$conn, NAME=$conr_name, SIZE=$conr_size,"
                . " Con Type=$conr_type, Raid=$raid_type" );

        # Grap correct size of disks
        my $list_disk
            = get_matching_spare_disks( $conn, $conr_size, $raid_type )
            or croak "Not enough space: $@";

        # Issue the command on the FILER to create the Container
        # ( Trad or Aggr )
        if ( $conr_type eq "aggr" ) {
            aggr_create( $conn, $conr_name, $list_disk, %vol_create_args )
                or croak(
                "FS::Util::Setup::create_matching_container - aggr_create failed $@"
                );
        } else {
            vol_create( $conn, $conr_name, $list_disk, %vol_create_args )
                or croak(
                "FS::Util::Setup::create_matching_container -  vol_create failed $@"
                );
        } ## end else [ if ( $conr_type eq "aggr")

        # We are here, hence succeeded with everything
        logDbg("FS::Util::Setup::create_matching_container - Succeeded");
        return 1;
    };
} ## end sub create_matching_container ($$$;@)

## @name ::get_matching_spare_disks($conn, $size, $raid_type)
##
## @summary Get appropriate list of spare disks for the given size
##
## @description
##          Returns the list of appropriate spare disk
##          names for the size and raid_type.
##
##          This disk list can be used to create any type of
##          containers (trad|aggr).
##
## Assumptions
##
## Temporary assumptions to move forward quickly. These definitely
## need to be revisited (???)
##
## 1. some amount (???) is adjusted by the filer to have compatibility
##   across different hard disk vendors, hence 'used_sz' value is
##   used as opposed to 'avl_mb' value.
##
## 2. Around 50% of the space can not be used by the user for the
##   following reasons
##   (i)  20% space is alloted for snapshot storage.
##   (ii) 10% of space is reserved for file system performance
##   (iii) Some unknown % of space is taken for unknown reasons (???) :-)
##
## @arg $conn Required. The connection to the filer host
##
## @arg $size Required. The total usable size required
##
## @arg $raid_type Required. The type of the raid_type. This
##                 can be used to decide the number of parity disks.
##

sub get_matching_spare_disks ( $$$ ) {
    my $conn = make_conn_and_complain(shift);

    my ( $size, $raid_type )
        = validate_pos( @_, { type => SCALAR }, { type => SCALAR } );

    my ( %sdisks, @match_disks );
    my $parity_disks;

    # Decide the number of parity disks
    if ( $raid_type eq "raid4" ) {
        $parity_disks = 1;
    } else {
        $parity_disks = 2;
    }

    # Calculate(Resize) the required '$size' in MB
    my $cunit = guess_unit($size);
    my $sizen = strip_unit_spec($size);

    $size = convert_units( $sizen, $cunit, UNIT_MB );
    logDbg(
        "FS::Util::Setup::get_matching_spare_disks - SIZE=$size $cunit ==> SIZE=$size MB"
    );

    # Increase the required size by 2 (50% usable size)
    $size = $size * 2;

    return eval {

        # Get Spare disk name and the space
        my $rf = new RaidInfo($conn);
        my %sd = $rf->spare_disks();
        foreach my $disk ( values %sd ) {
            my $diskname = $disk->name();
            my $value    = $disk->parameter("used_sz");
            $sdisks{$diskname} = $value;
        }

        # Check the number of disks
        #      # should be > $parity_disks
        if ( scalar( keys %sdisks ) <= $parity_disks ) {
            croak("Not enough disks found");
        }

        logDbg( "FS::Util::Setup::get_matching_spare_disks - "
                . Dumper( \%sdisks ) );

        my $first_disk;
        logDbg(
            "FS::Util::Setup::get_matching_spare_disks - Find whether the required size of disk is available directly"
        );
        my $cnt = 0;

        # Sort the values (size) in the ascending order
        foreach my $diskname (
            sort { $sdisks{$a} <=> $sdisks{$b} }
            keys %sdisks
            )
        {
            my $this_disk = $sd{$diskname};
            if ( $size <= $sdisks{$diskname} ) {
                if ($first_disk) {
                    next unless ( $first_disk->{rpm} == $this_disk->{rpm} );
                } else {
                    $first_disk = $this_disk;
                }
                push( @match_disks, $diskname );
                logDbg(
                    "FS::Util::Setup::get_matching_spare_disks - Adding disk = $diskname"
                );
                $cnt += 1;
                if ( $cnt > $parity_disks ) {

                    # We got the required disks
                    logDbg(
                        "FS::Util::Setup::get_matching_spare_disks - Got the required disks"
                    );
                    return \@match_disks;
                } ## end if ( $cnt > $parity_disks)
            } ## end if ( $size <= $sdisks{...
        } ## end foreach my $diskname ( sort...

        # Initialize again to make sure, we don't have any
        # prior stored info.
        @match_disks = ();

        # We don't have enough disks yet, start getting the disks from
        # big size to smaller size until we get the required space.
        # Sort the values (size) in the descending order

        logDbg(
            "FS::Util::Setup::get_matching_spare_disks - required parity disks: $parity_disks, cnt = $cnt"
        );
        logDbg(
            "FS::Util::Setup::get_matching_spare_disks - Start getting the disks from bigger size"
        );
        my $got_space = 0;    # We have got 0 MB
        $cnt = 0;             # Number of disks
        foreach my $diskname (
            sort { $sdisks{$b} <=> $sdisks{$a} }
            keys %sdisks
            )
        {
            my $this_disk = $sd{$diskname};
            if ($first_disk) {
                next unless ( $first_disk->{rpm} == $this_disk->{rpm} );
            } else {
                $first_disk = $this_disk;
            }
            if ( $cnt < $parity_disks ) {

                # we have to leave the biggest disks for PARITY.
                push( @match_disks, $diskname );
                $cnt += 1;
                logDbg(
                    "FS::Util::Setup::get_matching_spare_disks - Adding parity disk = $diskname"
                );
            } else {
                push( @match_disks, $diskname );
                $cnt += 1;
                logDbg(
                    "FS::Util::Setup::get_matching_spare_disks - Adding data disk = $diskname"
                );
                $got_space += $sdisks{$diskname};
            } ## end else [ if ( $cnt < $parity_disks)

            # Check if the required space is fulfilled
            if ( $got_space > $size ) {

                # Wow!!! We have got the space, we want
                logDbg(
                    "FS::Util::Setup::get_matching_spare_disks - Managed to get enough disks"
                );
                return \@match_disks;
            } ## end if ( $got_space > $size)
        } ## end foreach my $diskname ( sort...

        # We are here, because we can't find enough disks
        croak "Not enough space found";
    };
} ## end sub get_matching_spare_disks ( $$$ )

## @name
## ::aggr_switch_root($conn, $vol_src, $vol_dst, %opts)
##
## @summary Switch the root aggr and volume from one to another via
##          the MAINT and BOOT_MENU options.
##
## @description
##          Sets up the $vol_dst as a root volume by optionally copying
##          the system files from $vol_src, then boots into MAINT mode
##          and sets the container of $vol_dst as the root aggr. Then
##          the filer is rebooted into BOOT_MENU where the $vol_dst is
##          set as the root volume via the vol_pick_root command.
##
##          By default, the filer is rebooted to the UP state, but that can
##          be controlled by the reboot option
##
## @example
##          aggr_switch_root($conn, "vol0", "vol1");
##          aggr_switch_root($conn, "vol0", "vol1", copy_sys_files=>0);
##          aggr_switch_root($conn, "vol0", "vol1", reboot=>0);
##
## @arg $conn Required. The connection to the filer host,
##            where the old root volume to be switched.
##
## @arg $vol_src Required. The name of the existing root volume.
##
## @arg $vol_dst Required. The name of the volume to be switched
##               as root volume.
##
## @arg %opts
##      This is hash with the following options.
##
##      copy_sys_files [0|1] - Whether or not $vol_src should be copied
##                   to $vol_dst before $vol_dst is set to root.
##                   By default, $vol_src is copied to $vol_dst.
##
##      reboot [0|1] - Whether or not system should be rebooted
##                     for the changes to take effect. By default, rebooted.
##
##      wait_time [120] - How long to wait for system to boot
##                     By default, 120.

sub aggr_switch_root ($$$;@) {
    my $conn     = make_conn_and_complain(shift);
    my $console  = make_conn( $conn->{hostp}, "#console" );
    my $filer_ip = $conn->{hostp}->{default_ip};

    my ( $vol_src, $vol_dst, %opts ) = validate_mixed(
        @_,
        [ { type => SCALAR }, { type => SCALAR }, ],
        {   copy_sys_files =>
                { type => BOOLEAN, optional => 1, default => 1 },
            reboot    => { type => BOOLEAN, optional => 1, default => 1 },
            wait_time => { type => SCALAR,  optional => 1, default => 120 },
            cmdtimeout => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    my $CMDTIMEOUT = $opts{cmdtimeout} || $main::CMDTIMEOUT;

    my ( $aggr_dst, $output, $prompt );
    my $P_MAINT = "\\*\>";

    return eval {

        logDbg(
            "FS::Util::Setup::aggr_switch_root - SRC=$vol_src, DST=$vol_dst, copy_sys_files=$opts{copy_sys_files}, "
                . "delete=$opts{delete}, reboot=$opts{reboot}" );

        # Make sure $vol_src and $vol_dst exists
        # (Just to make sure :-) )
        logDbg(   "FS::Util::Setup::aggr_switch_root - Check whether or not"
                . " $vol_src and $vol_dst exist" );
        if (   !vol_exists( $conn, $vol_src )
            || !vol_exists( $conn, $vol_dst ) )
        {
            croak
                "FS::Util::Setup::aggr_switch_root - Specified volume (either $vol_src or $vol_dst)"
                . " doesn't exist";
        } ## end if ( !vol_exists( $conn...

        # Get the container of vol_dst
        my $aggr_dst = vol_container( $conn, $vol_dst );

        if ( $opts{copy_sys_files} ) {

            # Copy all system files from $vol_src to $vol_dst
            # give a huge timeout just to be safe
            copy_system_files(
                $conn, $vol_src, $vol_dst,
                filer_ip => $filer_ip,
                timeout  => 3600
                )
                or croak
                "FS::Util::Setup::aggr_switch_root - Copying system files failed: $@";
        } ## end if ( $opts{copy_sys_files...

        logDbg("FS::Util::Setup::aggr_switch_root - Get to MAINT menu");
        transit_to( $console, 'MAINT', timeout => $CMDTIMEOUT );

        # Make $aggr_dst new root aggr
        ( undef, undef, $output ) = say( $console,
            "aggr options $aggr_dst root",
            '-wait',
            "Aggregate $aggr_dst will become root at the next boot|Could not copy",
            '-prompt',
            $P_MAINT
        );

        # Reboot to the boot menu
        transit_to( $console, 'BOOT_MENU' ) or croak($@);

        # Pick the new root volume as root
        ( undef, undef, $output, $prompt )
            = say( $console, "vol_pick_root $vol_dst $aggr_dst",
            '-noprompt', '-prompt', "Selection", '-timeout', $CMDTIMEOUT );

        if ( $opts{reboot} ) {
            logDbg(
                "FS::Util::Setup::aggr_switch_root - Reboot for the change to take effect"
            );
            if ( $prompt =~ /Selection/ ) {

                # Need to re-boot to activate new root volume
                say( $console,
                    "1",
                    '-wait',
                    "[mgr.boot.reason_ok:notice]|.*System rebooted.*",
                    '-noprompt',
                    '-timeout',
                    $CMDTIMEOUT
                );
            } ## end if ( $prompt =~ /Selection/)

            # Wait for system to completely boot up
            snooze( $opts{wait_time} );

# Make sure the system is in UP state - At this point it should only require a login
            transit_to( $console, 'UP' )
                or croak("FS::Util::Setup::aggr_switch_root - $@");

            # Make sure $vol_dst is root volume now
            my $root_vol = vol_root($conn);
            if ( $vol_dst ne $root_vol ) {
                croak
                    "FS::Util::Setup::aggr_switch_root - Setting $vol_dst to root failed $@";
            }
        } ## end if ( $opts{reboot} )

        return 1;
    };
} ## end sub aggr_switch_root ($$$;@)

## @name
## ::vol_switch_root($conn, $vol_src, $vol_dst, %opts)
##
## @summary Switch the root volume from one to another
##
## @description
##          Sets the $vol_dst as root volume from $vol_src.
##          If $opts{copy_sys_files} option is set (which is default),
##          $vol_src is copied to $vol_dst before $vol_dst
##          is set to root.
##
##          By default, the filer is rebooted for the changes
##          to take effect.
##
## @example
##          vol_switch_root($conn, "vol0", "vol1");
##          vol_switch_root($conn, "vol0", "vol1", delete=>1);
##          vol_switch_root($conn, "vol0", "vol1", copy_sys_files=>0);
##          vol_switch_root($conn, "vol0", "vol1", reboot=>0);
##
## @arg $conn Required. The connection to the filer host,
##            where the old root volume to be switched.
##
## @arg $vol_src Required. The name of the existing root volume.
##
## @arg $vol_dst Required. The name of the volume to be switched
##               as root volume.
##
## @arg %opts
##      This is hash with the following options.
##
##      delete [0|1] - Flag to delete $vol_src volume.
##                     By default, $vol_src is not deleted.
##
##      copy_sys_files [0|1] - Whether or not $vol_src should be copied
##                   to $vol_dst before $vol_dst is set to root.
##                   By default, $vol_src is copied to $vol_dst.
##
##      reboot [0|1] - Whether or not system should be rebooted
##                     for the changes to take effect. By default, rebooted.
##

sub vol_switch_root ($$$;@) {
    my $conn = make_conn_and_complain(shift);
    my ( $vol_src, $vol_dst, %opts ) = validate_mixed(
        @_,
        [ { type => SCALAR }, { type => SCALAR }, ],
        {   copy_sys_files =>
                { type => BOOLEAN, optional => 1, default => 1 },
            delete => { type => BOOLEAN, optional => 1, default => 0 },
            reboot => { type => BOOLEAN, optional => 1, default => 1 },
        }
    );

    return eval {

        logDbg(
            "FS::Util::Setup::vol_switch_root - SRC=$vol_src, DST=$vol_dst, copy_sys_files=$opts{copy_sys_files}, "
                . "delete=$opts{delete}, reboot=$opts{reboot}" );

        # Make sure $vol_src and $vol_dst exists
        # (Just to make sure :-) )
        logDbg(   "FS::Util::Setup::vol_switch_root -  Check whether or not"
                . " $vol_src and $vol_dst exist" );
        if (   !vol_exists( $conn, $vol_src )
            || !vol_exists( $conn, $vol_dst ) )
        {
            croak
                "FS::Util::Setup::vol_switch_root - Specified volume (either $vol_src or $vol_dst)"
                . " doesn't exist";
        } ## end if ( !vol_exists( $conn...

        if ( $opts{copy_sys_files} ) {

            # Copy all system files from $vol_src to $vol_dst
            # give a huge timeout just to be safe
            copy_system_files( $conn, $vol_src, $vol_dst, timeout => 3600 )
                or croak
                "FS::Util::Setup::vol_switch_root - Copying system files failed: $@";
        } ## end if ( $opts{copy_sys_files...

        # Set root option for $vol_dst
        vol_options( $conn, $vol_dst, root => '' ) or croak $@;

        if ( $opts{reboot} ) {
            logDbg(
                "FS::Util::Setup::vol_switch_root - Reboot for the change to take effect"
            );

            # Reboot the system for the change to take effect
            reboot_to( $conn->{hostp}, 'UP' )
                or croak("FS::Util::Setup::vol_switch_root - $@");

            # Make sure $vol_dst is root volume now
            my $root_vol = vol_root($conn);
            if ( $vol_dst ne $root_vol ) {
                croak
                    "FS::Util::Setup::vol_switch_root - Setting $vol_dst to root failed $@";
            }
        } ## end if ( $opts{reboot} )

        # If the $opts{delete} option is set, remove the $vol_src
        if ( $opts{delete} ) {
            logDbg(
                "FS::Util::Setup::vol_switch_root - Deleting volume $vol_src"
            );
            vol_purge( $conn, $vol_src ) or logWarn($@);
        } ## end if ( $opts{delete} )
        return 1;
    };
} ## end sub vol_switch_root ($$$;@)

sub get_vol_identification(%) {
    my %opt = (
        filer_connection => undef(),
        volume_path      => undef(),
        @_
    );

    if ( !defined( $opt{filer_connection} ) || !defined( $opt{volume_path} ) )
    {
        die(script_result(
                "FS::Util::Setup::get_vol_identification - A filer_connection ("
                    . $opt{filer_connection}
                    . ") and volume_path ("
                    . $opt{volume_path}
                    . ") are requried."
            )
        );
    } ## end if ( !defined( $opt{filer_connection...

    my $dsid;
    my $fileid;
    my $gen;

    my ( $before, $match, $output )
        = say( $opt{filer_connection},
        "priv set test; showfh $opt{volume_path}" );

    if ( $output
        =~ m/flags=(\S+)\ssnapid=(\S+)\sfileid=(\S+)\sgen=(\S+)\sfsid=(\S+)\sdsid=(\S+)/
        )
    {
        $fileid = hex($3);
        $gen    = hex($4);
        $dsid   = hex($6);
    } else {
        die("FS::Util::Setup::get_vol_identification - Cannot determine the file handle for $opt{volume_path}.\n"
        );
    }

    return ( $dsid, $fileid, $gen );
} ## end sub get_vol_identification(%)

##
## @name ::copy_system_files Copy system files from one volume to another
## @arg conn Required. Connection object to Filer
## @arg $svol Required. Source volume, where the system files copied from
## @arg $dvol Required. Destination volume, where the system files copied to.
## @arg client Optional. Named parameter referencing client.
## @arg filer_ip Optional. IP addr of filer, sometimes required when two filers are defined.
## @arg timeout Optional. Timeout to complete entire copy
##

sub copy_system_files($$$;%) {
    my ( $conn, $svol, $dvol, %opts ) = @_;
    my $timeout
        = $opts{timeout}
        ? $opts{timeout}
        : 900;    # timeout to complete entire copy for big root volumes
    my $client = $opts{client} ? $opts{client} : "localhost";
    my $filer_ip
        = $opts{filer_ip} ? $opts{filer_ip} : $conn->{hostp}->{default_ip};

    my $ch    = host($client);
    my $cc    = connect( '#rsh', -timeout, $main::TIMEOUT );
    my $filer = $conn->hostp()->id();
    my ( $text, $cmd, $out );
    my $errmsg;

    # Make sure both the volumes are exported with 'anon=0'
    # not to get into permission denied errors later.
    export_path(
        $conn, "/vol/" . $svol,
        OPTIONS   => "anon=0",
        WRITE_ETC => 1
    ) or croak "export_path failed for $svol: $@";

    export_path(
        $conn, "/vol/" . $dvol,
        OPTIONS   => "anon=0",
        WRITE_ETC => 1
    ) or croak "export_path failed for $dvol: $@";

    # Mount source and destination volumes
    logDbg(
        "FS::Util::Setup::copy_system_files - Mounting source volume $svol of $filer"
    );
    my $src_vol_path = mount_vol(
        filer_vol => $svol,
        filer     => $filer,
        filer_ip  => $filer_ip,
        client    => $client,
        timeout   => $timeout
    );
    if ( $@ || !$src_vol_path ) {
        my $msg = $@;
        logDbg(
            "FS::Util::Setup::copy_system_files - Could not mount src vol $filer:$svol on client $client: $msg"
        );
        $@ = $msg;
        goto clean;
    } ## end if ( $@ || !$src_vol_path)

    logDbg(
        "FS::Util::Setup::copy_system_files - Mounting dest volume $dvol of $filer"
    );

    my $dst_vol_path = mount_vol(
        filer_vol => $dvol,
        filer     => $filer,
        filer_ip  => $filer_ip,
        client    => $client,
        timeout   => $timeout
    );
    if ( $@ || !$dst_vol_path ) {
        my $msg = $@;
        logDbg(
            "FS::Util::Setup::copy_system_files - Could not mount dst vol $filer:$dvol on client $client: $msg"
        );
        $@ = $msg;
        goto clean;
    } ## end if ( $@ || !$dst_vol_path)

# here we will copy all the system files from   $src_vol_path to $dst_vol_path
# but will ignore 'cores' and 'logs' directories since this will take forever
# if those directories are full
    logDbg(
        "FS::Util::Setup::copy_system_files - Copying system files from $src_vol_path to $dst_vol_path"
    );
    $cmd
        = "(mkdir $dst_vol_path/etc; $sudo_cmd chmod 0777 $dst_vol_path/etc; cd $src_vol_path/etc;$sudo_cmd cp $src_vol_path/etc/* $dst_vol_path/etc/)";

# Copy the files and make sure there is space for a swapfile on the destination
    eval {
        my $out = sayEOF( $cc, $cmd, timeout => $timeout );
        sayEOF( $conn,
            "priv set -q diag; mkfile 1g /vol/$dvol/etc/swapfile" );
    };
    goto clean if ( $out !~ /^\d+ blocks/ );

clean:
    $errmsg = $@ ? $@ : $text;
    if ($src_vol_path) {
        umount_vol( $src_vol_path, client => $client, force => 1 )
            or logWarn($@);
    }

    if ($dst_vol_path) {
        umount_vol( $dst_vol_path, client => $client, force => 1 )
            or logWarn($@);
    }

    croak $errmsg if ($errmsg);
    return 1;
} ## end sub copy_system_files($$$;%)

#
## @name ::cold_start_filer Do a Filer/newfs.thpl
## Cold start a filer, basically does a newfs on the filer
## args are the same as for the Filer/newfs.thpl script
##
sub cold_start_filer(%) {

    my (%args) = @_;

    my $newfs = defined( $args{newfs} ) ? $args{newfs} : 1;

    my $fh       = host($main::FILER);
    my $fconsole = connect("#console");
    my $frsh     = connect("#rsh");

    if ($newfs) {
        logDbg("FS::Util::Setup::cold_start_filer - Cold starting the filer");
    } else {
        logDbg("FS::Util::Setup::cold_start_filer - Warm starting the filer");
    }

    my ( $cmd, $vol_type, $filer_args );
    my @opts;
    my %fields = $fh->param_fields();

    #logDbg("fields are: ".Dumper(%fields));

    push( @opts, "FILER_GATEWAY=$fields{GATEWAY}" );
    push( @opts, "FILER_DNS=Y" );
    push( @opts,
              "FILER_DNS_SERVER_LIST="
            . $fields{MGMT_DNS} . ","
            . $fields{MGMT_DNS2} );
    push( @opts, "FILER_NEWFS_COMMAND=$cmd" );
    defined($main::FILER_ROOT_VOL_TYPE)
        ? push( @opts, "FILER_ROOT_VOL_TYPE=$main::FILER_ROOT_VOL_TYPE" )
        : push( @opts, "FILER_ROOT_VOL_TYPE=f" );

    defined( $fields{ "IFC_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "IFC_" . $fields{DATA_PORT1} } )
        : push( @opts,
        "FILER_IFC_" . $fields{DATA_PORT1} . "=" . $fields{DATA_PORT1} );
    defined( $fields{ "IP_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "IP_" . $fields{DATA_PORT1} } )
        : push( @opts,
        "FILER_IP_" . $fields{DATA_PORT1} . "=" . $fields{DATA_IP1} );
    defined( $fields{ "MASK_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "MASK_" . $fields{DATA_PORT1} } )
        : push( @opts,
        "FILER_MASK_" . $fields{DATA_PORT1} . "=" . $fields{DATA_MASK1} );
    defined( $fields{ "PARTNER_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_PARTNER_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "PARTNER_" . $fields{DATA_PORT1} } )
        : undef;

    defined( $fields{ "IFC_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "IFC_" . $fields{DATA_PORT2} } )
        : push( @opts,
        "FILER_IFC_" . $fields{DATA_PORT2} . "=" . $fields{DATA_PORT2} );
    defined( $fields{ "IP_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "IP_" . $fields{DATA_PORT2} } )
        : push( @opts,
        "FILER_IP_" . $fields{DATA_PORT2} . "=" . $fields{DATA_IP2} );
    defined( $fields{ "MASK_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "MASK_" . $fields{DATA_PORT2} } )
        : push( @opts,
        "FILER_MASK_" . $fields{DATA_PORT2} . "=" . $fields{DATA_MASK2} );
    defined( $fields{ "PARTNER_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_PARTNER_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "PARTNER_" . $fields{DATA_PORT2} } )
        : undef;

    defined( $fields{ "IFC_" . $fields{CLUSTER_PORT1} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{ "IFC_" . $fields{CLUSTER_PORT1} } )
        : push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{CLUSTER_PORT1} );
    defined( $fields{ "IP_" . $fields{CLUSTER_PORT1} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{ "IP_" . $fields{CLUSTER_PORT1} } )
        : push( @opts,
        "FILER_IP_" . $fields{CLUSTER_PORT1} . "=" . $fields{CLUSTER_IP1} );
    defined( $fields{ "MASK_" . $fields{CLUSTER_PORT1} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{ "MASK_" . $fields{CLUSTER_PORT1} } )
        : push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{CLUSTER_MASK1} );

    defined( $fields{ "IFC_" . $fields{CLUSTER_PORT2} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{ "IFC_" . $fields{CLUSTER_PORT2} } )
        : push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{CLUSTER_PORT2} );
    defined( $fields{ "IP_" . $fields{CLUSTER_PORT2} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{ "IP_" . $fields{CLUSTER_PORT2} } )
        : push( @opts,
        "FILER_IP_" . $fields{CLUSTER_PORT2} . "=" . $fields{CLUSTER_IP2} );
    defined( $fields{ "MASK_" . $fields{CLUSTER_PORT2} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{ "MASK_" . $fields{CLUSTER_PORT2} } )
        : push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{CLUSTER_MASK2} );

    defined( $fields{ "IFC_" . $fields{MGMT_PORT} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{MGMT_PORT} . "="
            . $fields{ "IFC_" . $fields{MGMT_PORT} } )
        : push( @opts,
        "FILER_IFC_" . $fields{MGMT_PORT} . "=" . $fields{MGMT_PORT} );
    defined( $fields{ "IP_" . $fields{MGMT_PORT} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{MGMT_PORT} . "="
            . $fields{ "IP_" . $fields{MGMT_PORT} } )
        : push( @opts,
        "FILER_IP_" . $fields{MGMT_PORT} . "=" . $fields{MGMT_IP} );
    defined( $fields{ "MASK_" . $fields{MGMT_PORT} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{MGMT_PORT} . "="
            . $fields{ "MASK_" . $fields{MGMT_PORT} } )
        : push( @opts,
        "FILER_MASK_" . $fields{MGMT_PORT} . "=" . $fields{MGMT_MASK} );

    if ($newfs) {

        # Default to 24/7 and a flex root vol
        $cmd
            = ($main::FILER_NEWFS_COMMAND)
            ? $main::FILER_NEWFS_COMMAND
            : '24/7';
        $vol_type
            = ($main::FILER_ROOT_VOL_TYPE) ? $main::FILER_ROOT_VOL_TYPE : 'f';

        newfs(
            $fh,
            prefix        => "FILER",
            root_vol_type => $vol_type,
            command       => $cmd,
            timeout       => 360
        ) or die $@;
    } ## end if ($newfs)

    eval {
        logDbg("FS::Util::Setup::cold_start_filer - Calling basic_setup");
        basic_setup(
            $fconsole,
            prefix          => "FILER",
            issue_setup_cmd => 0,
            timeout         => 560
        );
    };
    if ($@) {
        if ( $@ =~ /Err=TIMEOUT/ ) {

            # Poke the console and try again
            #say($fconsole, "\n", -blind);
            logDbg(
                "FS::Util::Setup::cold_start_filer - Calling basic_setup again after poking the console"
            );
            basic_setup(
                $fconsole,
                prefix          => "FILER",
                issue_setup_cmd => 0,
                timeout         => 60
            ) or die("basic setup failed for a second time. $@");
        } else {
            die("basic setup failed for something other than a timeout. $@");
        }
    } ## end if ($@)

    boot::transit_to( $fconsole, 'UP' );

    subtest( "admin/security_defaults.thpl", "SAVE=1" );

    my $filer_pw = $fh->default_password();
    if ( $filer_pw ne '' ) {
        say( $fconsole, "passwd", -blind );
        while (1) {
            my $prompt;
            ( undef, undef, undef, $prompt )
                = say( $fconsole, "", -prompt,
                [ 'New password:', 'Retype new password:' ] );
            if ( $prompt =~ 'New password' ) {
                say( $fconsole, "$filer_pw", -blind );
            } elsif ( $prompt =~ 'Retype new password:' ) {
                say( $fconsole, "$filer_pw", -blind );
                last;
            }
        } ## end while (1)

    } ## end if ( $filer_pw ne '' )

    license( $frsh, [ 'nfs', 'cifs', 'http' ] )
        or die(
        "FS::Util::Setup::cold_start_filer - Unable to create required licenses $@"
        );

    if ($main::IRON_ON_ROOT) {
        aggr_add( $frsh, 'aggr0', 3 );
    }
} ## end sub cold_start_filer(%)

# NA_REVISIT: GV Nov14 2004
# These _ndmpxxxx functions should never be called by any other
# tests. These functions are temporarily used by vol_switch_root.
# This usage should be replaced by test/NDMP/NdmpCmds.pm eventually.
# test/NDMP is not part of the PERLLIB path, hence including
# this library is deferred for now.

# Copies the source volume to destination volume

sub _ndmpcopy ($$$;@) {
    my ( $conn, $svol, $dvol, %opts ) = @_;
    $conn = make_conn_and_complain($conn);
    my $cmd;
    my $text;

    if ( $opts{force} == 1 ) {
        $cmd = "ndmpcopy -f /vol/$svol /vol/$dvol";
    } else {
        $cmd = "ndmpcopy /vol/$svol /vol/$dvol";
    }

    # Make sure ndmpd is on
    _ndmpd_on($conn) or die $@;

    $text = sayEOF( $conn, $cmd );
    if ( $text =~ /Transfer successful/ ) {
        return 1;
    } else {
        die "FS::Util::Setup::_ndmpcopy - $text";
    }
} ## end sub _ndmpcopy ($$$;@)

# Returns the status of ndmpd
sub _ndmpd_status ($) {
    my $conn = make_conn_and_complain(shift);
    my $cmd  = "ndmpd status";
    my %opts;

    my $text = sayEOF( $conn, $cmd );

    if ( $text =~ /not found/ ) {
        die $text;
    }

    if ( $text =~ /ndmpd OFF/ ) {
        $opts{status} = "off";
    } elsif ( $text =~ /ndmpd ON/ ) {
        $opts{status} = "on";
    } else {
        die "FS::Util::Setup::_ndmpd_status - Unknown status $text";
    }

    return %opts;
} ## end sub _ndmpd_status ($)

# Returns true if it succeeded 'ndmpd on'

sub _ndmpd_on ($) {
    my $conn = make_conn_and_complain(shift);
    my $cmd  = "ndmpd on";

    my $text = sayEOF( $conn, $cmd );

    my %opts = _ndmpd_status($conn);

    if ( $opts{status} eq "on" ) {
        return 1;
    } else {
        die "FS::Util::Setup::_ndmpd_on failed. $text";
    }
} ## end sub _ndmpd_on ($)

# Returns true if it succeeded 'ndmpd off'

sub _ndmpd_off ($) {
    my $conn = make_conn_and_complain(shift);
    my $cmd  = "ndmpd off";

    my $text = sayEOF( $conn, $cmd );

    my %opts = _ndmpd_status($conn);

    if ( $opts{status} eq "off" ) {
        return 1;
    } else {
        die "FS::Util::Setup::_ndmpd_off failed. $text";
    }
} ## end sub _ndmpd_off ($)

# returns root minimum size as a bare number in MB
sub calc_root_size ($$) {
    my ( $conn, $vol_name ) = @_;

    # Decide the size of the volume we want to create
    my ( undef, $used, undef, undef ) = vol_disk_free(
        $conn,
        path        => "$vol_name",
        space_type  => 'space',
        report_type => UNIT_MB
    );

    # Have 30% more for snapshot and incase
    my $vol_size = int( $used + ( $used * 0.3 ) + 1 );

    # NA_REVISIT: GV
    # For some reason, the volume size is not enough to 'ndmpcopy'
    # Hence just double the size for now. It has to be fixed in a
    # clean way, once it is found.
    $vol_size = $vol_size * 2;

    my $vol_min = vol_determine_flex_root_size($conn)
        or die $@;
    $vol_min = ( int( convert_to_bytes($vol_min) + 1 ) ) / ( 1024 * 1024 )
        or die $@;
    $vol_min = int($vol_min);

    return ( $vol_size > $vol_min ) ? $vol_size : $vol_min;
} ## end sub calc_root_size ($$)

## @name ::can_run_test
## @summary Compares the values passed in to those defined in the current
##          environment variables, and decides whether the calling script
##          can continue. Relevant failures such as NA,CONFIG reported
##          back to the test with die(config_result('')).
##
## @arg %args. Optional. If not provided, No checks will be performned.
##
##         vol_type - trad | flex.
##
##         host_type - sim | hardware
##

sub can_run_test($;%) {
    my ( $fc, %opts ) = @_;

    logDbg( "FS::Util::Setup::can_run_test - "
            . Dumper( _strip_filer( \%opts ) ) );

    if ( $opts{host_type} eq "sim" ) {
        if ( $fc->{hostp}->{hosttype} !~ /filer-simulator/ ) {

            # This test can not be run as hosttype is not a sim.
            die(na_result(
                    "FS::Util::Setup::can_run_test - This test can only run on simulator"
                )
            );
        } else {

            # This test can run.
        }
    } elsif ( $opts{host_type} eq "hardware" ) {
        if ( $fc->{hostp}->{hosttype} =~ /filer-simulator/ ) {

            # This test can not be run as hosttype is sim.
            die(na_result(
                    "FS::Util::Setup::can_run_test - This test can only run on hardware filers"
                )
            );
        } else {

            # This test can run.
        }
    } elsif ( $opts{host_type} ) {
        die("FS::Util::Setup::can_run_test - Incorrect host_type passed $opts{host_type}"
        );
    }

    if ( $main::SETUP_DONE =~ /1|reuse/i ) {
        if ( $opts{vol_type} ) {
            if ( $opts{vol_type} ne $main::VOL_TYPE ) {
                die(config_result(
                        "FS::Util::Setup::can_run_test -Test requires vol_type $opts{vol_type}"
                    )
                );
            } ## end if ( $opts{vol_type} ne...
        } ## end if ( $opts{vol_type} )
    } else {

        # As SETUP_DONE=0, we don't have to have the check for now.
        # This require more re-thinking of how we want to execute the
        # tests. Let's keep this simple for now.
    } ## end else [ if ( $main::SETUP_DONE...
    return 1;
} ## end sub can_run_test($;%)

## @name ::cross_populate_hosts_files
## @summary use hosts files to make sure source and dest filers can resolve each
##          other's names (required for snapmirror, sparse/flexcache, etc).
##          returns a list of two entries, containing the text of the original
##          (unmodified) hosts files for the target filers.
##          NOTE: the new entry is written to the head of the file and will therefore
##          supercede any existing entry for the given hostname (a warning is logged
##          when this happens).
##
## @example
##          my ($src_hosts, $dest_hosts) = cross_populate_hosts_files($src_fc, $dest_fc);
##
## @arg %args. Required. If not provided, No checks will be performned.
##
##         fc_a - ConnectRec reference for first target filer.
##
##         fc_b - ConnectRec reference for seond target filer.
##
sub cross_populate_hosts_files {
    my ( $fc_a, $fc_b ) = @_;

    my $a_hostname = $fc_a->{hostp}->{hostname};
    my $b_hostname = $fc_b->{hostp}->{hostname};

    my $a_default_ip = $fc_a->{hostp}->{default_ip};
    my $b_default_ip = $fc_b->{hostp}->{default_ip};

    my $a_data = rdfile( $fc_a, "/etc/hosts" )
        or die("Could not read /etc/hosts from '$a_hostname': $@");
    my $b_data = rdfile( $fc_b, "/etc/hosts" )
        or die("Could not read /etc/hosts from '$b_hostname': $@");

# does the hostname for A appear in the host data from B? (outside of comments)
    my $b_new_data = "$a_default_ip $a_hostname\n" . $b_data;
    if ( $b_data =~ /^\s*([\.\d]+)(?:\s|\s.*\s)$a_hostname\s.*$/m ) {
        my $ip = $1;
        if ( $ip ne $a_default_ip ) {
            $Log->warn(
                "cross_populate_hosts_files: host '$a_hostname' already exists in hosts file for '$b_hostname' but with a different ip address ($ip), the new entry will supercede this pre-existing definition."
            );

            # add an entry for A to the data from B
            wrfile( $fc_b, "/etc/hosts", $b_new_data )
                or die("Could not write /etc/hosts to '$b_hostname': $@");
        } ## end if ( $ip ne $a_default_ip)
    } else {

        # add an entry for A to the data from B
        wrfile( $fc_b, "/etc/hosts", $b_new_data )
            or die("Could not write /etc/hosts to '$b_hostname': $@");
    } ## end else [ if ( $b_data =~ /^\s*([\.\d]+)(?:\s|\s.*\s)$a_hostname\s.*$/m)

    # is B defined in host data from A?
    my $a_new_data = "$b_default_ip $b_hostname\n" . $a_data;
    if ( $a_data =~ /^\s*([\.\d]+)(?:\s|\s.*\s)$b_hostname\s.*$/m ) {
        my $ip = $1;
        if ( $ip ne $b_default_ip ) {
            $Log->warn(
                "cross_populate_hosts_files: host '$b_hostname' already exists in hosts file for '$a_hostname' but with a different ip address ($ip), the new entry will supercede this pre-existing definition."
            );

            # add an entry for B to the data from A
            wrfile( $fc_a, "/etc/hosts", $a_new_data )
                or die("Could not write /etc/hosts to '$a_hostname': $@");
        } ## end if ( $ip ne $b_default_ip)
    } else {

        # add an entry for B to the data from A
        wrfile( $fc_a, "/etc/hosts", $a_new_data )
            or die("Could not write /etc/hosts to '$a_hostname': $@");
    } ## end else [ if ( $a_data =~ /^\s*([\.\d]+)(?:\s|\s.*\s)$b_hostname\s.*$/m)

    return ( $a_data, $b_data );
} ## end sub cross_populate_hosts_files

# ------------------------------------------------------------------------

1;

