#$Id: //depot/prod/test/Rsierranevada/lib/FS/Util/Storage_Ems.pm#1 $
##
## @summary Ems - Library module for ems command
## @author sajan@netapp.com
## @status private
## @change 10-31-2006 anishv: Added ems_error_monitor function
## @change 08-01-2007 anishv: Added ignore_error_values option to ems_error_monitor function
## @change 09-25-2007 anishv: Added key_value_errors option to ems_error_monitor function
## @change 02-20-2008 sreedhar: Removed unwanted console messages from the command ems log dump 1s command output
## @change 11-25-2008 anishv: Added ems.autosuppress.enable on
## @change 09-25-2009 deepak1: Changed the say command in ems_log_get_mark function to fix the burt 374241
## @change 10-27-2009 deepak1: Modified the ems_log_get_mark function to fix the burt 374241
## @note THIS MODULE IS INTENDED AS A TEMPORARY ADDITION, PENDING A FIX FOR lib/Storage/Ems.pm.
## @pod here

=head1 NAME

Storage::Ems

=head1 DESCRIPTION

This library defines a set of interfaces to interrogate ems subsystem 
and return (pointer to hash) ems log data

=head1 SYNOPSIS

use Storage::Ems qw(ems_log_get_mark ems_log_dump_by_type ems_log_dump_by_id ems_event_status ems_get_value_by_key);
   

my $host = host($FILER);
my $mark = ems_log_get_mark(host=>$host);
$event = $event_by_type->{"dfu.firmwareUpToDate"};
if ($event eq "") {
    croak "Disk firmware download didn't complete in given time";
}

OR 
my $host = host($FILER);
my $mark = ems_log_get_mark(host=>$host);
# Check if adpater 6a has been offlined or not
my $ret = ems_log_check(host=>$host, eventtype=>"fci.adapter.offline", key=>"adapterName" value=>"6a", timestamp=>$mark);
if ($ret == 0) {
    logcomment("The specified event is not found in the ems log");
} else {
    logcomment("Found the specified ems event $ret times in the ems log");
}

=head1 ROUTINES

=cut

package FS::Util::Storage_Ems;

use strict;
use warnings;

use Carp qw(confess croak);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF );
use Tharn qw(say script_result host logcomment logresult param);
use Filer::Version qw(:releases get_wafl_version);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

use Utils qw(make_conn);
use Data::Dumper;

use base qw(Exporter);
our @EXPORT_OK = qw(
                    ems_log_get_mark 
                    ems_log_dump_by_type 
                    ems_log_dump_by_id 
                    ems_event_status
                    ems_log_check
		    ems_error_monitor
		    ems_get_value_by_key
                   );

#----------------------------------------------------------------------
# The exported rotines are coded in the following section
#----------------------------------------------------------------------

=head2 ems_get_merk

This routine returns the current mark(timestamp in format <n>t) of the event 
log

my $host = host($FILER);
my $ret = ems_get_mark(host=>$host);
my $ret = ems_get_mark(host=>$host, timeout=>60);

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub ems_log_get_mark {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
                 });

    my $conn = make_conn($opts{host});
	# To Disable ems duplicate events

	# ems.log.dupsuppress.enable off options is deprecated in RR
	# See http://wikid.netapp.com/w/EMS/RollingRock/Convergence_Functional_Spec 
	my $wafl_ver = get_wafl_version($opts{host});
    if ( $wafl_ver < WV_ROLLINGROCK_MIN ) {
    	say($conn, "options ems.log.dupsuppress.enable off", "-timeout", $opts{timeout});
	}

    say($conn, "options ems.autosuppress.enable off", "-timeout", $opts{timeout});

    my $cmd = "ems log dump 1s";
    my $str = qr/^<NR\s[^<]+\/>/;
    my ($output, undef, undef) =  say($conn, "$cmd", $str, "-timeout", $opts{timeout});
    #Use header (NR) only to get the timestamp
    my %emslog_nr = emslog_parseNR($output);
    return ($emslog_nr{t} . "t");
}

=head2 ems_log_dump_by_type

This routine returns pointer to hash table having event name as key and 
array of event ids as value.

my $host = host($FILER);
my $ret = ems_log_dump_by_type(host=>$host);
OR
my $mark = ems_log_get_mark(host=>$host);
my $event_by_type = ems_log_dump_by_type(host=>$host, timestamp=>$mark);
$event = $event_by_type->{"dfu.firmwareDownloading"};
if ($event eq "") {
    croak "Disk firmware download is not started yet";
}

$event = $event_by_type->{"dfu.firmwareUpToDate"};
if ($event eq "") {
    croak "Disk firmware download didn't complete in given time";
}

The output looks like:
$VAR1 = {
          'fci.adapter.offlining' => [
                                       '1133838353/531',
                                       '1133838353/539'
                                     ],
          'fci.adapter.offline' => [
                                     '1133838353/532',
                                     '1133838353/540'
                                   ],
          'fci.adapter.onlining' => [
                                      '1133838353/536',
                                      '1133838353/543'
                                    ],
          'unowned.disk.reminder' => [
                                       '1133838353/529'
                                     ],
          'monitor.shelf.fault' => [
                                     '1133838353/527',
                                     '1133838353/528'
                                   ],
          'fci.adapter.online' => [
                                    '1133838353/537',
                                    '1133838353/544'
                                  ],
          'kern.syslog.msg' => [
                                 '1133838353/530',
                                 '1133838353/533',
                                 '1133838353/534',
                                 '1133838353/535',
                                 '1133838353/538',
                                 '1133838353/542'
                               ],
          'config.NotMultiPath' => [
                                     '1133838353/541'
                                   ],
          'kern.uptime.filer' => [
                                   '1133838353/526'
                                 ]
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item timestamp

(Optional) : Get the event from timestamp Default : ""(means only header)

Timestamp semantic can be : 

<n>s    Last n seconds
<n>m    Last n minutes
<n>h    Last n hours
<n>d    Last n days
<n>t    Since timestamp <n>

=item log

(Optional) : The logfile to get events from. Default : "" (means current)

=item timeout

(Optional) : Timeout value for the say. Default : 1200

=back

=cut

sub ems_log_dump_by_type {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec',default=>"",optional=>1 },
                     timestamp => { type=>SCALAR,default=>"", optional=>1 },
                     log => { type=>SCALAR,default=>"", optional=>1 },
                     verbose => { type=>SCALAR,default=>"yes", optional=>1 },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
		     ems_file => {type=>SCALAR,default=>"none",optional=>1},
                 });

    my $emslog_entry;
    my @emslog_events;
    my @emslog_event_lines;
    my %emslog_event_by_id;
    my %emslog_event_by_type;
    my %emslog_nr;
    my @lines;
    my $verbose = " ";
    if($opts{verbose} eq "no"){
    	 $verbose = "-noverbose";
    }
    if ($opts{ems_file} eq "none"){
    	my $conn = make_conn($opts{host});
    	my $cmd = "ems log dump $opts{log} $opts{timestamp}";
    	logcomment("Issuing $cmd  $verbose");
    	my (undef, undef, $output) = say($conn, "$cmd","$verbose", "-timeout", $opts{timeout});
    	@lines = split(/\r*\n/, $output);
    }
    else{
	open(FILE,"< $opts{ems_file}") or die "Could not open file $opts{ems_file} for reading: $!\n";
	while(<FILE>){
		push @lines,$_;
	    }
    }

    foreach my $line (@lines) {
        if (!emslog_parse($line, \%emslog_nr, \$emslog_entry)) {
            next;
	}
	emslog_process_line($emslog_entry, \@emslog_events, 
                            \@emslog_event_lines, 
                            \%emslog_event_by_id, 
                            \%emslog_event_by_type);
	$emslog_entry = "";
    }
    return(\%emslog_event_by_type);
}

=head2 ems_log_dump_by_id

This routine returns pointer to hash table having event info by id

my $ret = ems_log_dump_by_id(host=>$host);
my $ret = ems_log_dump_by_id(host=>$host, timeout=>300, timestamp=>23456788t);

The output looks like:
$VAR1 = {
          '1133838353/530' => {
                                'vfilerName' => '',
                                'header' => {
                                              'n' => 'raid-u23',
                                              'o' => 'rc',
                                              'p' => 6,
                                              's' => 'OK',
                                              't' => '1134007568',
                                              'd' => '08Dec2005 02:06:08',
                                              'vf' => '',
                                              'id' => '1133838353/530'
                                            },
                                'msg' => 'Taking adapter 6a offline at administrator&apos;s request.&nl;',
                                'type' => 'kern.syslog.msg',
                                'attr' => 'msg,vfilerName'
                              },
          '1133838353/531' => {
                                'header' => {
                                              'n' => 'raid-u23',
                                              'o' => 'rc',
                                              'p' => 6,
                                              's' => 'OK',
                                              't' => '1134007568',
                                              'd' => '08Dec2005 02:06:08',
                                              'vf' => '',
                                              'id' => '1133838353/531'
                                            },
                                'adapterName' => '6a',
                                'type' => 'fci.adapter.offlining',
                                'attr' => 'adapterName'
                              },
          .
          .
          .
}


The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item timestamp

(Optional) : Get the event from this timestamp Default : ""(means only header)

Timestamp semantic can be : 

<n>s    Last n seconds
<n>m    Last n minutes
<n>h    Last n hours
<n>d    Last n days
<n>t    Since timestamp <n>

=item log

(Optional) : The logfile to get events from. Default : "" (means current)

=item timeout

(Optional) : Timeout value for the say. Default : 1200: Increasing the default
value as the log file might be quite big and can take longer to dump the file 

=back

=cut

sub ems_log_dump_by_id {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec',default=>"",optional=>1 },
                     timestamp => { type=>SCALAR,default=>"", optional=>1 },
                     log => { type=>SCALAR,default=>"", optional=>1 },
                     verbose => { type=>SCALAR,default=>"yes", optional=>1 },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
		     ems_file => {type=>SCALAR,default=>"none",optional=>1},
                 });

    my $emslog_entry;
    my @emslog_events;
    my @emslog_event_lines;
    my %emslog_event_by_id;
    my %emslog_event_by_type;
    my %emslog_nr;
    my @lines;
    my $verbose = " ";
    if($opts{verbose} eq "no"){
    	$verbose = "-noverbose";
    }
if ($opts{ems_file} eq "none"){
    my $conn = make_conn($opts{host});
    my $cmd = "ems log dump $opts{log} $opts{timestamp}";
    logcomment("Issuing $cmd  $verbose");
    my (undef, undef, $output) = say($conn, "$cmd","$verbose", "-timeout", $opts{timeout});
    @lines = split(/\r*\n/, $output);
}
else{
	open(FILE,"< $opts{ems_file}") or die "Could not open file $opts{ems_file} for reading: $!\n";
	while(<FILE>){
		push @lines,$_;
	    }
}
    foreach my $line (@lines) {
        if (!emslog_parse($line, \%emslog_nr, \$emslog_entry)) {
            next;
	}
	emslog_process_line($emslog_entry, \@emslog_events, 
                            \@emslog_event_lines, 
                            \%emslog_event_by_id, 
                            \%emslog_event_by_type);
	$emslog_entry = "";
    }
    return(\%emslog_event_by_id);
}

=head2 ems_event_status

This routine returns pointer to hash table having event summery info return 
by the command "ems event status"

my $ret = ems_event_status(host=>$host);

The output looks like:
$VAR1 = {
          'dfu.firmwareDownloading' => {
                                         'LastTime' => '30Nov2005 07:02:00',
                                         'Drop' => '0',
                                         'Indication' => 12
                                       },
          'raid.cksum.replay.nvram' => {
                                         'LastTime' => '29Nov2005 06:32:53',
                                         'Drop' => '0',
                                         'Indication' => 6
                                       },
          'sfu.downloadStarted' => {
                                     'LastTime' => '29Nov2005 06:33:05',
                                     'Drop' => '0',
                                     'Indication' => 1
                                   },
          'dfu.firmwareUpToDate' => {
                                      'LastTime' => '30Nov2005 07:04:21',
                                      'Drop' => '0',
                                      'Indication' => 13
                                    },
          'sfu.downloadSummary' => {
                                     'LastTime' => '29Nov2005 06:41:28',
                                     'Drop' => '0',
                                     'Indication' => 1
                                   },
          'sfu.downloadSuccess' => {
                                     'LastTime' => '29Nov2005 06:41:28',
                                     'Drop' => '0',
                                     'Indication' => 1
                                   },
          .
          .
          .
        };

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item timeout

(Optional) : Timeout value for the say. Default : 300

=back

=cut

sub ems_event_status {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     timeout => { type=>SCALAR, default=>300, optional=>1 },
                 });

    my %ems_event_status;

    my $conn = make_conn($opts{host});
    my $cmd = "ems event status";

    my (undef, undef, $output) = say($conn, "$cmd", "-timeout", $opts{timeout});
    my (@lines) = split(/\r*\n/, $output);
    foreach my $line (@lines) {
        if ($line =~ /(\w+)\./) {    
            my @event_info = split(/\s+/, $line); 
            $ems_event_status{$event_info[0]}{'Indication'} = $event_info[1];
            $ems_event_status{$event_info[0]}{'Drop'} = $event_info[2];
            $ems_event_status{$event_info[0]}{'LastTime'} = $event_info[3] . " $event_info[4]";
        }
    }
    return(\%ems_event_status);
}

=head2 ems_log_check

This routine returns 0 if a given event on given objectname is not found from a specified timestamp otherwise returns <n> for the n number of instances found in the log.

my $host = host($FILER);
my $mark = ems_log_get_mark(host=>$host);
my $ret = ems_log_check(host=>$host, eventtype=>"fci.adapter.offline", key=>"adapterName" value=>"6a", timestamp=>$mark);
if ($ret == 0) {
    logcomment("The specified event is not found in the ems log");
} else {
    logcomment("Found the specified ems event $ret times in the ems log");
}

The following named arguments are accepted,

=over

=item host

(Required) : Hostspec to filer

=item eventtype

(Required) : The type of events to look into, like "fci_adapter_offline"...

=item key

(Required) : What is the key (the objectname for a given event type) to look into the event 
message. For example, in the following ems message key is "adapterName":

<LR d="07Dec2005 18:45:31" n="raid-u23" t="1133981131" id="1133838353/404" p="6" s="OK" o="ispfc_init_2" vf="">
<fci_adapter_offline_1
        adapterName="6a"/>
</LR>

=item value

(Required) : What is the value (the objectname for a given event type) to look into for the 
given event type and key. To check for the above ems message the value would be "6a".

=item timestamp

(Required) : Get the event from this timestamp.

Timestamp semantic can be : 

<n>s    Last n seconds
<n>m    Last n minutes
<n>h    Last n hours
<n>d    Last n days
<n>t    Since timestamp <n>

=item log

(Optional) : The logfile to get events from. Default : "" (means current)

=item timeout

(Optional) : Timeout value for the say. Default : 1200: Increasing the default
value as the log file might be quite big and can take longer to dump the file 

=back

=cut

sub ems_log_check {
    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     eventtype => { type=>SCALAR },
                     key => { type=>SCALAR },
                     value => { type=>SCALAR },
                     timestamp => { type=>SCALAR },
                     log => { type=>SCALAR,default=>"", optional=>1 },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
                 });

    my $event_by_type;
    my $event_by_id;
 
    if (defined($opts{log})) {
        $event_by_type = ems_log_dump_by_type(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout}, log=>$opts{log});
    } else {
        $event_by_type = ems_log_dump_by_type(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout});
    }

    if (defined($opts{log})) {
        $event_by_id = ems_log_dump_by_id(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout}, log=>$opts{log});
    } else {
        $event_by_id = ems_log_dump_by_id(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout});
    }

    my $event = $event_by_type->{"$opts{eventtype}"};
    if(!defined($event)) {
        return (0);
    } else {
        my $found;
        $found = 0;
        foreach my $id (@{$event}) {
            if ($event_by_id->{$id}{$opts{key}} eq "$opts{value}") {
                ++$found; 
            }
        }
        return($found);
    }
}



=head2 ems_error_monitor
	my @ret = ems_error_monitor(host=>$host, mark=>$mark, filermode=>$filermode,monitor_time=>$monitor_time,sleep_time=>$sleep_time,error_values=>\@errors,fatal_on_failure=>"yes");
	
This subrutine checks for ems errors 

The following arguments are accepted:

=over

=item host
(Required): Hostspec to filer

=item mark

(Required) : Time marker from where to start checking the errors

=item monitor_time 

(Optional): Monitor time. Default is 1 sec.

=item sleep_time

(Optional): Sleep time between error checks. Default is 1 sec.

=item error_values

(Optional): Reference to an Array with a list of errors to be checked. By default it will use a list defined in the library

=item additional_error_values

(Optional): Reference to an Array with a list of additional errors to be checked. The list of additional_error_values will be appended to the list of errors already defined in the function. By default it will use the  list defined in the library without additional errors.

=item key_value_errors

(Optional): Reference to a hash with a list of additional errors with key value pairs to be checked. The list of additional_error_values will be appended to the list of errors already defined in the function. By default it will use the  list defined in the library without additional errors. For example if we want to check for the following message:

<LR d="26Sep2007 02:40:44" n="stor-u104" pn="stor-u105" t="1190774444" id="1190710933/2316" p="5" s="C=1U" o="cf_main" vf="">
<cf_fsm_takeoverOfPartnerDisabled_1
        partner="stor-u105"
        reason="status of backup mailbox is uncertain"/>

we can use :
%err = ("cf.fsm.takeoverofPartnerDisabled" => "reason=status of backup mailbox is uncertain");
ems_error_monitor(host=>$host,key_value_errors=>\%err);

=item ignore_error_values

(Optional): Reference to an Array with a list of  errors to be ignored. The list of additional_error_values will be removed to the list of errors already defined in the function. By default it will use the  list defined in the library without removing errors.

=item fatal_on_failure

(Optional): To stop test on seeing errors? Default is yes.

=item timeout

(Optional) : Timeout value for the say. Default : 1200: Increasing the default
value as the log file might be quite big and can take longer to dump the file 

=item logresults

(Optional) : Allows users of this routine to prevent the routine from
logging test results. If set to 'no' then an error string is returned.
Default : yes.

=back

=cut

sub ems_error_monitor {
	my (%opts) = validate(@_,{
				host => {type=>OBJECT, isa=>'Hostrec'},
				mark => {type=>SCALAR},
				monitor_time => { type=>SCALAR,default=>1,optional=>1 },
				sleep_time => {type=>SCALAR,default=>1,optional=>1},
				error_values => {type=>ARRAYREF, optional =>1 },
				additional_error_values => {type=>ARRAYREF, optional =>1 },
				ignore_error_values => {type=>ARRAYREF, optional =>1 },
				key_value_errors => {type=>HASHREF, optional =>1 },
				fatal_on_failure => {type=>SCALAR,default=>"yes",optional=>1},
           			timeout => { type=>SCALAR, default=>1200, optional=>1 },
           			logresults => { type=>SCALAR, default=>'yes', optional=>1 },
	}
	);
        my $errmsg = '';
	my @error_values;
	my $mark;
	my $k;
	my $v;
	my $filer = $opts{host}->{'hostname'};
	print "FILER IS: $filer \n";
	#sleep 20;
	

	if(!defined($opts{error_values})){
		my %ems_error_messages=(#"01"=>"monitor.shelf.fault",
                    "02"=>"raid.disk.missing",
                    "03"=>"raid.config.spare.disk.missing",
                    "04"=>"cf.fsm.backupMailboxError",
		    "05"=>"fci.adapter.parity",
		    "06"=>"fci.adapter.bad",
		    "07"=>"sas.adapter.bad",
		    "08"=>"fci.adapter.failed",
		    "09"=>"sas.adapter.failed",
		    "10"=>"fci.adapter.nvram.invalid",
		    "11"=>"sas.adapter.nvram.invalid",
		    "12"=>"fci.device.invalid.wwn",
		    "13"=>"sas.device.invalid.wwn",
		    "14"=>"fci.device.invalidate.soft.address",
		    "15"=>"sas.device.invalidate.soft.address",
		    "16"=>"fci.adapter.firmware.fault",
		    "17"=>"sas.adapter.firmware.fault",
		    "18"=>"fci.adapter.firmware.pause",
		    "19"=>"sas.adapter.firmware.update.failed",
		    "20"=>"fci.adapter.not.ready",
		    "21"=>"sas.adapter.not.ready",
		    "22"=>"sas.adapter.offline",
		    "23"=>"fci.device.login.failure",
		    "24"=>"fci.adapter.unexpected.status",
		    "25"=>"sas.adapter.unexpected.status",
		    "26"=>"disk.write.failure",
		    "27"=>"diskown.errorReadingOwnership",
		    "28"=>"disk.senseError",
		    "29"=>"disk.readReservationFailed",
		    "30"=>"scsi.cmd.adapterFatal",
		    "31"=>"scsi.cmd.adapterHardwareError",
		    "32"=>"scsi.cmd.adapterResources",
		    "33"=>"scsi.cmd.contingentAllegiance",
		    "33"=>"scsi.cmd.noMorePaths",
		    "34"=>"raid.disk.unload.done",
		    "35"=>"raid.rg.recons.missing",
		    "35"=>"raid.disk.maint.start",
		    

                   );
		 @error_values=values(%ems_error_messages);
	}
	else{
		 @error_values = @{$opts{error_values}};
	}

	if(defined($opts{additional_error_values})){
		 push @error_values,@{$opts{additional_error_values}};
	 }

	if(defined($opts{ignore_error_values})){
		@error_values = _get_unique_elements($opts{ignore_error_values},\@error_values);
	 }


  my $loop_count = int($opts{monitor_time} / $opts{sleep_time});

   $mark = $opts{mark};

  my $i = 1;
  my $fail = "no";

	while ($i <= $loop_count) {
		my $mark1;
		if($loop_count != 1){
      		$mark1 = ems_log_get_mark(host=>$opts{host},timeout=>$opts{timeout});
		}
      		my $event_by_type = ems_log_dump_by_type(host=>$opts{host}, timestamp=>$mark,timeout=>$opts{timeout});
      		foreach my $msg (@error_values) {
          		my $event = $event_by_type->{"$msg"};
          		if (defined($event)) {
		  		#if($opts{fatal_on_failure} eq "yes"){
              			#	logresult('FATAL',"Failed because of: $msg");
		  		#}
		  		#else{
              				logcomment("FAILED because of: $msg");
					$fail = "yes";
		  		#}

          		}
      		}
			if(defined($opts{key_value_errors})){
				#print "Defined \n";
      				my $event_by_id = ems_log_dump_by_id(host=>$opts{host}, timestamp=>$mark,timeout=>$opts{timeout});
				#print "-----------------------------------\n";
				#print Dumper $opts{key_value_errors};
				#print Dumper $event_by_id;
				#print Dumper $event_by_type;
				#print "-----------------------------------\n";
				while(($k,$v) = each %{$opts{key_value_errors}}){
					#print "$k => $v \n";
    					my $event = $event_by_type->{"$k"};
					#print Dumper ($event);
					if(defined($event)) {
        					foreach my $id (@{$event}) {
							my @key_val = split(/=/,$v);
								#print "KEY is $key_val[0] VALUE TO CHECK: $key_val[1] \n";
            						if(defined($event_by_id->{$id}{$key_val[0]})){
            							my $val = $event_by_id->{$id}{$key_val[0]} ;
								my $value_to_check = $key_val[1];
								if($val =~ /$value_to_check/){
              									logcomment("FAILED: Saw unexpected ems event on filer $filer : $k => $v : Error: $val");
										$fail = "yes";
								}
							}
						}
					}
				}
			}
							


		if($loop_count != 1){
      			$mark = $mark1;
		}
      		sleep $opts{sleep_time};
      		++$i;
  	}

	if ($opts{logresults} eq 'yes') {
		# Didn't find any error in the given time
		if($fail eq "no"){
			logcomment("No recognizable errors found.");
		} else {
			if($opts{fatal_on_failure} eq "yes"){
				logresult("fatal","Unexpected EMS errors found. Please check the logs above for details");
			} else {
				logresult("fail","Unexpected EMS errors found. Please check the logs above for details");
			}
		}
	} else {
		# Didn't find any error in the given time
		if($fail eq "yes"){
			$errmsg = "Unexpected EMS errors found. Please check the logs above for details";
		}
		return ($errmsg);
	}
}


=head2 ems_get_value_by_key
	my @ret = ems_get_value_by_key(host=>$host, eventtype=>$eventtype, key=>$key,timestamp=>$timestamp,log=>$log,timeout=>$timeout);
	
This subrutine returns the value of a given key.

The following arguments are accepted:

=over

=item host
(Required): Hostspec to filer

=item eventtype

(Required) : Ems type. ex: "raid.rg.recons.aborted"

=item key 

(Required): The key which the requested value for. ex:dbn 

=item timestamp
(Required) : the event from this timestamp.
Timestamp semantic can be :
<n>s    Last n seconds
<n>m    Last n minutes
<n>h    Last n hours
<n>d    Last n days
<n>t    Since timestamp <n>

=item log

(Optional): 

=item timeout

(Optional) : Timeout value for the say. Default : 1200: Increasing the default
value as the log file might be quite big and can take longer to dump the file 

=back

=cut

sub ems_get_value_by_key {

    my (%opts) = validate(@_, { host => { type=>OBJECT, isa=>'Hostrec' },
                     eventtype => { type=>SCALAR },
                     key => { type=>SCALAR },
                     timestamp => { type=>SCALAR },
                     log => { type=>SCALAR,default=>"", optional=>1 },
                     timeout => { type=>SCALAR, default=>1200, optional=>1 },
                 });

    my $event_by_type;
    my $event_by_id;

    if (defined($opts{log})) {
        $event_by_type = ems_log_dump_by_type(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout}, log=>$opts{log});
    } else {
        $event_by_type = ems_log_dump_by_type(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout});
    }

    if (defined($opts{log})) {
        $event_by_id = ems_log_dump_by_id(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout}, log=>$opts{log});
    } else {
        $event_by_id = ems_log_dump_by_id(host=>$opts{host}, timestamp=>$opts{timestamp}, timeout=>$opts{timeout});
    }

    my $event = $event_by_type->{"$opts{eventtype}"};
    if(!defined($event)) {
        return (0);
    } else {
        foreach my $id (@{$event}) {
            return $event_by_id->{$id}{$opts{key}} ;
        }
        return(undef);
    }
}

  
#----------------------------------------------------------------------
# The local rotines are coded in the following section
#----------------------------------------------------------------------

sub emslog_parse {
    my($line, $emslog_nr, $emslog_entry) = @_;

    if ($line =~ /^<NR/) {
        %{$emslog_nr} = emslog_parseNR($line);
        return 0;
    }

    if ($line =~ /^<LR/) {
        ${$emslog_entry} = $line;
        return 0;
    }
	
    if ($line =~ /^<\/LR/) {
        ${$emslog_entry} = ${$emslog_entry} . $line;
        return 1;
    }

    if (!defined(${$emslog_entry})) {
        ${$emslog_entry} = " $line";
    } else {
        ${$emslog_entry} = ${$emslog_entry} . " $line";
    }
    return 0;
}

sub emslog_process_line {
    my($line, $emslog_events, $emslog_event_lines, $emslog_event_by_id, 
       $emslog_event_by_type) = @_;

    my($header, $event, $id, $i);
    my(@headerlist, @eventlist);
    my($type, $version);
	
    if ($line =~ /^\s*<LR (.*)>\s*<(.*)\/>\s*<\/LR>\s*$/) {
        $header = "$1";
        $event = $2;
    } else {
        logcomment("emslog_process_line: mismatch\n");
        return;
    }
	
    #
    # Get the id.
    #
    if ($header =~ /\s*id=\"([^\"]+)\"/) {
        $id = $1;
    }
	
    push @{$emslog_events}, $id;
    push @{$emslog_event_lines}, $line;
	

    #
    # Parse the header (LR) section of the event record
    #	
    while ($header ne "") {
        if ($header =~ /^\s*([^=]+)=\"([^\"]*)\"\s*(.*)$/) {
            $emslog_event_by_id->{$id}{header}{$1} = $2;
            $header = $3;
        } else {
            $header = "";
        }
    }
	
    if ($event =~ /^\s*(\S+)\s*(.*)/) {
        $type = $1;
        $event = $2;
    }
	
    if ($type =~ /^(\S+)_([0-9]+)$/) {
        $type = $1;
        $type =~ s/_/\./g;
    }
	
    emslog_add_by_type($type, $id, $emslog_event_by_type);
    $emslog_event_by_id->{$id}{__event_def} = $type;

    #
    # Parse the per-event specific part of the event record
    #
    while ($event ne "") {
        if ($event =~ /^\s*([^=]+)=\"([^\"]*)\"\s*(.*)$/) {
            $emslog_event_by_id->{$id}{$1} = $2;
            if (!exists($emslog_event_by_id->{$id}{attr})) {
                $emslog_event_by_id->{$id}{attr} = $1;
            } else {
                $emslog_event_by_id->{$id}{attr} .= ",$1";
            }
            $event = $3;
        } else {
            $event = "";
        }
    }
}

sub emslog_parseNR {
    my($line) = @_;
    my %emslog_nr = ();
    my($header);
    if ($line =~ /^\s*<NR (.*)\/>\s*$/) {
        $header = "$1";
    } else {
        logcomment("parseNR mismatch: '$line'\n");
        croak $line;
    }
	
    #
    # Parse the header (NR) section of the event record
    #	
    while ($header ne "") {
        if ($header =~ /^\s*([^=]+)=\"([^\"]*)\"\s*(.*)$/) {
            $emslog_nr{$1} = $2;
            $header = $3;
        } else {
            $header = "";
        }
    }
    return (%emslog_nr);
}

sub emslog_add_by_type {
    my($type, $id, $emslog_event_by_type) = @_;
	
    #if ($emslog_event_by_type->{$type} eq "") {
    #    $emslog_event_by_type->{$type} = $id;
    #} else {
    #    $emslog_event_by_type->{$type} .= ",$id";
    #}
    push(@{$emslog_event_by_type->{$type}}, $id);
}

#----------------------------------------------------------------------
# The testme routine to test the exported functions of this library
#
# This routine can be invoked in the following manner
#
# ntest -hw FILER=raid-u24 PKGNAME=Storage::Ems util/test_pkg (test all functions)
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_get_mark PKGNAME=Storage::Ems util/test_pkg
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_dump_by_type PKGNAME=Storage::Ems util/test_pkg
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_dump_by_type TIMESTAMP=1133355221t PKGNAME=Storage::Ems util/test_pkg
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_dump_by_id TIMESTAMP=2s PKGNAME=Storage::Ems util/test_pkg
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_dump_by_id PKGNAME=Storage::Ems util/test_pkg
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_dump_by_id TIMESTAMP=1133355221t PKGNAME=Storage::Ems util/test_pkg
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_event_status TIMESTAMP=1133355221t PKGNAME=Storage::Ems util/test_pkg
#
#
# ntest -hw FILER=raid-u24 FUNCTION_NAME=ems_log_check TIMESTAMP=30s PKGNAME=Storage::Ems util/test_pkg
#
#----------------------------------------------------------------------

sub testme {

    package main;
    our ($FILER);
    our ($FUNCTION_NAME);
    our ($TIMESTAMP);
    package Storage::Ems;
    param('FILER',       '-mesg',    "Filer to test");
    param('FUNCTION_NAME',       '-default',    "all");
    param('TIMESTAMP',       '-default',    "1s");

    logcomment("Testing on FILER $FILER");
    logcomment("Testing $FUNCTION_NAME");
    logcomment("Testing $TIMESTAMP");

    my $host = host($FILER);
    my $ret;

    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "ems_log_get_mark") {
        eval {
            $ret = ems_log_get_mark(host=>$host);
        };

        if ($@) {
            logresult('fail', "function ems_log_get_mark $@");
        } else { 
            logcomment("The marker is: $ret\n");
        }

        logresult('pass', "Got expected result from function ems_log_get_mark");
    }

    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "ems_log_dump_by_type") {
        eval {
            $ret = ems_log_dump_by_type(host=>$host, timestamp=>$TIMESTAMP);
        };

        if ($@) {
            logresult('fail', "function ems_log_dump_by_type $@");
        } else { 
            print Dumper($ret);
        }

        logresult('pass', "Got expected result from function ems_log_dump_by_type");
    }

    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "ems_log_dump_by_id") {
        eval {
            $ret = ems_log_dump_by_id(host=>$host, timestamp=>$TIMESTAMP);
        };

        if ($@) {
            logresult('fail', " function ems_log_dump_by_id $@");
        } else { 
            print Dumper($ret);
        }

        logresult('pass', "Got expected result from function ems_log_dump_by_id");
    }

    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "ems_event_status") {
        eval {
            $ret = ems_event_status(host=>$host);
        };

        if ($@) {
            logresult('fail', " function ems_event_status $@");
        } else { 
            print Dumper($ret);
        }

        logresult('pass', "Got expected result from function ems_event_status");
    }

    if ($FUNCTION_NAME eq "all" || $FUNCTION_NAME eq "ems_log_check") {
        eval {
            $ret = ems_log_check(host=>$host, eventtype=>"fci.adapter.offline", key=>"adapterName", value=>"6a", timestamp=>$TIMESTAMP);
        };

        if ($ret != 0) {
            logresult('fail', " function ems_log_check: $ret");
        } else { 
            logresult('pass', "Got expected result from function ems_log_check");
        }
    }
}

#-------------------------------------------------
# Local subroutine to get unique elements which are in array2 but not in array1
#-------------------------------------------------
sub _get_unique_elements {
	my ($array1,$array2) = @_;
	my @disks = ();
	my $item;
	my %seen = ();
	foreach $item (@$array1) {
		$seen{$item} = 1;
	}
	foreach $item (@$array2) {
		unless ($seen{$item}){
			push (@disks, $item) ;
		}
	}
	return @disks;
}

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Testme

=head1 ATTRIBUTES

    ONTAP_COMMANDS = ems

    ONTAP_ZAPI_APIS = none 

=head1 AUTHOR

    Sajan Agrawal (sajan@netapp.com)

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;

