#-*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/VFILER_Vfiler.pm#1 $

## @summary Routines for vfiler
##
## @description
##	functions for the vfiler test automation
##
## @keywords vfiler
##
## @status public
##
## @dependencies
##	use telnet to connect your filer, not rsh
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
##
## @pod here
##

package FS::Util::VFILER_Vfiler;

use strict;
use warnings;
use Tharn;
use base qw(Exporter);
use SayWrap qw(sayEOF);
use Admin::License qw(is_licensed);
use SPARSE::Util::Logger;
use Carp qw(croak);
use Params::Validate qw(:all);

BEGIN
{
	our @EXPORT = qw(
		vfiler_add
		vfiler_allow
		vfiler_context
		vfiler_create
		vfiler_create_non_default_ipspace
                vfiler_create_default_ipspace
		vfiler_create_r
		vfiler_destroy
		vfiler_disallow
		vfiler_dr_activate
		vfiler_dr_configure
		vfiler_dr_delete
		vfiler_dr_dst_clear
		vfiler_dr_dst_set
		vfiler_dr_dst_show
		vfiler_dr_restart
		vfiler_dr_resync
		vfiler_dr_status
		vfiler_limit
		vfiler_limit_in_use
		vfiler_limit_hard_limit
		vfiler_limit_set
		vfiler_migrate
		vfiler_migrate_cancel
		vfiler_migrate_complete
		vfiler_migrate_start
		vfiler_migrate_status
		vfiler_move
		vfiler_remove
		vfiler_rename
		vfiler_start
		vfiler_stop
		vfiler_stop_destroy
		vfiler_status
		vfiler_run
		vfiler_run_vfiler_add_default_route
		vfiler_run_vfiler_cifs_setup
		vfiler_run_vfiler_ftp_enable
		vfiler_run_vfiler_exportfs
		vfiler_run_vfiler_http_enable
		vfiler_run_vfiler_setup
		run_cmd_on_partner
		vfiler_change_ip_binding
		vfiler_options_get
		vfiler_options_set
                vfiler_ssh_setup
                vfiler_ssh_cmd
	);
}

use Filer::Version qw(
	get_wafl_version
	get_ontap_version
	WV_ANCHORSTEAM_MIN
	WV_ANCHORSTEAM_MAX
	WV_TSINGTAO_MIN
	WV_GORDONBIERSCH_MIN
	UNSUPPORTED_FEATURE_EXCEPTION
	);

=head1 NAME

Vfiler - Vfiler routines

=head1 SYNOPSIS

use VFILER::Vfiler;

$ret = vfiler_add($fc, $vfname, %args);

$ret = vfiler_allow($fc, $vfilertemplate, [%args]);

$ret = vfiler_context($fc, $vfname);

$ret = vfiler_create($fc, $vfname, %args);

$ret = vfiler_create_non_default_ipspace($fc, $vfname, $ipsname, %args);

$ret = vfiler_create_r($fc, $vfname, $origin_vfiler_path, [%args]);

$ret = vfiler_destroy($fc, $vfname);

$ret = vfiler_disallow($fc, $vfilertemplate, [%args]);

$ret = vfiler_dr_activate($fc, $remote_vfiler, $remote_filer);

$ret = vfiler_dr_configure($fc, $remote_vfiler, $remote_filer, %args);

$ret = vfiler_dr_delete($fc, $remote_vfiler, $remote_filer, [$option]);

$ret = vfiler_dr_dst_clear($fc, $vfname, $dest_filer);

$ret = vfiler_dr_dst_set($fc, $vfname, $dest_filer, %args);

$ret = vfiler_dr_dst_show($fc, $vfname);

$ret = vfiler_dr_restart($fc, $vfname);

$ret = vfiler_dr_resync($fc, $vfname, $dest_filer, %args);

$ret = vfiler_dr_status($fc, $remote_vfiler, $remote_filer, $volname);

$ret = vfiler_limit($fc);

$ret = vfiler_limit_in_use($fc);

$ret = vfiler_limit_hard_limit($fc);

$ret = vfiler_limit_set($fc, $maxvfilers);

$ret = vfiler_migrate($fc, $remote_vfiler, $remote_filer, %args);

$ret = vfiler_migrate_cancel($fc, $remote_vfiler, $remote_filer);

$ret = vfiler_migrate_complete($fc, $remote_vfiler, $remote_filer);

$ret = vfiler_migrate_start($fc, $remote_vfiler, $remote_filer, %args);

$ret = vfiler_migrate_status($fc, $remote_vfiler, $remote_filer, $volname);

$ret = vfiler_move($fc, $vfname_from, $vfname_to, %args);

$ret = vfiler_remove($fc, $vfname, %args);

$ret = vfiler_rename($fc, $old_vfname, $new_vfname);

$ret = vfiler_start($fc, $vfilertemplate);

$ret = vfiler_stop($fc, $vfilertemplate);

$ret = vfiler_stop_destroy($fc, $vfname, [$remote_filer]);

$ret = vfiler_status($fc, [$vfilertemplate], [%args]);

$ret = vfiler_run($fc, $vfilertemplate, $cmd, [%args]);

$ret = vfiler_run_vfiler_add_default_route($fc, $vfname, [$gateway]);

$ret = vfiler_run_vfiler_cifs_setup($fc, $vfname, %args);

$ret = vfiler_run_vfiler_ftp_enable($fc, $vfname, $home_dir);

$ret = vfiler_run_vfiler_exportfs($fc, $vfname, $vfFS);

$ret = vfiler_run_vfiler_http_enable($fc, $vfname, $http_rootdir);

$ret = vfiler_run_vfiler_setup($fc, $vfname, %args);

$ret = run_cmd_on_partner($fc, $cmd);

$ret = vfiler_change_ip_binding(%args);

=head1 AUTHOR

Jae Bae

=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success,
and false on failure. On failure C<$@> is set to the appropriate error
value via C<die()>.

=head2 run_cmd_on_partner($fc, $cmd)

invoke commands on partner

=over 4

=item fc

Required: hostspec of the filer

=item cmd

Required: commands

=back

Examples:

	run_cmd_on_partner($fc, 'vfiler run vfilername exportfs -a');
	run_cmd_on_partner($fc,
		'vfiler run vfilername route add default 172.16.64.1 1');

=cut

sub run_cmd_on_partner($$)
{
	my ($fc, $cmd) = @_;

	return eval {
		sleep(4);

		my (undef, undef, $after) = say($fc, "cf status", -swallow);
        	if ($after !~ /has\s+taken\s+over/gm) {
			die "Takeover failed\n";
        	}

		sleep(4);
		(undef, undef, $after) = say($fc, "partner cf status", -swallow);
        	if ($after !~ /has\s+been\s+taken\s+over/gm) {
			die "Hasn't been taken over\n";
        	}
		sayEOF($fc, "partner $cmd");

		return 1;
	};
}

=head2 vfiler_run($fc, $vfilertemplate, $cmd, [%args])

runs the command on the vfiler(s) specified in the vfilertemplate.
If more than one vfiler matches, the command will be run separately
for each vfiler.

=over 4

=item fc

Required: hostspec of the filer

=item vfilertemplate

Required: name of vfiler or *

=item cmd

Required: commands available for vfiler

=item args

Optional: option, arguments

=back

Examples:

	vfiler_run($fc, 'vFiler100', 'netstat', arguments => "-rn");
	vfiler_run($fc, 'vFiler100', 'netstat', option => "-q", arguments => "-rn");
	vfiler_run($fc, '*', 'netstat', arguments => "-rn");
	vfiler_run($fc, '*', 'netstat', option => "-q", arguments => "-rn");

=cut

sub vfiler_run($$$;%)
{
	my ($fc, $vfilertemplate, $cmd, %args) = @_;

	my $option = $args{option} || "";
	my $arguments = $args{arguments} || "";

	if ($vfilertemplate eq "*") {
		$vfilertemplate = "*";
	}

	return eval {
		sleep(4);

                my (undef, undef, $after) =
                 say($fc, "vfiler run $option $vfilertemplate $cmd $arguments");
		     return (output => $after) if wantarray;
		return 1;
	};
}

=head2 vfiler_dr_dst_clear($fc, $vfname, $dest_filer)

removes the dr destination information

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of a vfiler

=item dest_filer

Required: name of the destination filer

=back

Examples:

	vfiler_dr_dst_clear($fc, 'vFiler100', 'dest_filername');

=cut

sub vfiler_dr_dst_clear
{
	my ($fc, $vfname, $dest_filer) = @_;

	my $fs_version = get_wafl_version($fc->{hostp});
	if ($fs_version < WV_GORDONBIERSCH_MIN) {
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION));
	}

	return eval {
		sleep(4);

		my $text = sayEOF($fc, "vfiler dr dst clear $vfname\@$dest_filer")
			or die $@;
		if ($text !~ /DR destination deleted/) {
			die("vfiler dr dst clear: $text");
		}

		return 1;
	};
}

=head2 vfiler_dr_dst_show($fc, $vfname)

removes the dr destination information

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of a vfiler

=back

Examples:

	vfiler_dr_dst_show($fc, 'vFiler100');

=cut

sub vfiler_dr_dst_show
{
	my ($fc, $vfname) = @_;

	my $fs_version = get_wafl_version($fc->{hostp});
	if ($fs_version < WV_GORDONBIERSCH_MIN) {
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION));
	}

	return eval {
		sleep(4);

		my $text = sayEOF($fc, "vfiler dr dst show $vfname")
			or die $@;
		if ($text !~ /Remote Filer/) {
			die("vfiler dr dst show: $text");
		}

		return 1;
	};
}

=head2 vfiler_dr_restart($fc, $vfname)

bring up a vfiler on the source that has been stopped due to a dr
activate on the destination

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of a vfiler

=back

Examples:

	vfiler_dr_restart($fc, 'vFiler100');

=cut

sub vfiler_dr_restart
{
	my ($fc, $vfname) = @_;

	my $fs_version = get_wafl_version($fc->{hostp});
	if ($fs_version < WV_TSINGTAO_MIN) {
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION));
	}

	return eval {
		sleep(4);

		my $text = sayEOF($fc, "vfiler status $vfname");
		if ($text =~ /stopped/) {
			sayEOF($fc, "vfiler dr restart $vfname");
		} else {
			die("can't do dr restart because $vfname is not ".
				"stopped: $text");
		}

		$text = sayEOF($fc, "vfiler status $vfname");
		if ($text =~ /stopped|not found/) {
			die("vfiler dr restart failed: $text");
		}

		return 1;
	};
}

=head2 vfiler_dr_dst_set($fc, $vfname, $dest_filer, %args)

set the remote vfiler characteristics on the destination filer in a dr
relationship to check for dr activations at every reboot as well as
for resync of the source with an activated dr destination

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of a vfiler

=item dest_filer

Required: name of the destination filer

=item args

Required: arguments

=back

Examples:

	vfiler_dr_dst_set($fc, 'vFiler100', 'dest_filername');

	vfiler_dr_dst_set($fc, 'vFiler100', 'dest_filername'
			alt_remote=>"111.111.111.112",
			remote_login=>"root"
			roption=>"-r"
			);

=cut

sub vfiler_dr_dst_set
{
	my ($fc, $vfname, $dest_filer, %args) = @_;

	my $remote_login = $args{remote_login} || "";
	my $remote_passwd = $args{remote_passwd} || "";
	my $alt_remote = $args{alt_remote} || "";
	my $loption = $args{loption} || "";
	my $aoption = $args{aoption} || "";
	my $roption = $args{roption} || "";
	my $timeout = $args{timeout} || 240;
	my $use_secure_channel=$args{use_secure_channel}||"false";
	my ($l_option, $a_option);

	if ($loption) {
		$l_option = "$loption $remote_login:$remote_passwd";
	}
	if ($aoption) {
		$a_option = "$aoption $alt_remote";
	}
	
	my $coption="";
	if($use_secure_channel eq "true"){
		$coption='-c secure';
	}
	my $fs_version = get_wafl_version($fc->{hostp});
	if ($fs_version < WV_GORDONBIERSCH_MIN) {
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION));
	}

	return eval {
		my (undef, undef, $after);

	   	$fc->timeout($timeout);
		sleep(4);

		if ($loption && $aoption) {
			(undef, undef, $after) = say($fc,
			"vfiler dr dst set $roption $l_option $a_option $coption $vfname\@$dest_filer",
			'-swallow');
		} elsif ($loption && !$aoption) {
			say($fc,
			"vfiler dr dst set $roption $l_option $coption $vfname\@$dest_filer",
			'-noprompt');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Alternate IP address\/Hostname for remote filer.*:");
			if ($alt_remote) {
				(undef, undef, $after) =
					say($fc, "$alt_remote", '-swallow');
			} else {
				(undef, undef, $after) =
					say($fc, "\r", '-nocr', '-swallow');
			}
		} elsif ($aoption && !$loption) {
			say($fc,
			"vfiler dr dst set $roption $a_option $coption $vfname\@$dest_filer",
			'-noprompt');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$dest_filer\'s Administrative login.*:");
			say($fc, "$remote_login", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$dest_filer\'s Administrative password.*:");
			(undef, undef, $after) =
				say($fc, "\r", '-nocr', '-swallow');
		} else {
			say($fc,
			"vfiler dr dst set $roption $coption $vfname\@$dest_filer", '-noprompt');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Alternate IP address\/Hostname for remote filer.*:");
			if ($alt_remote) {
				say($fc, "$alt_remote", '-blind');
			} else {
				say($fc, "\r", '-nocr', '-blind');
			}
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$dest_filer\'s Administrative login.*:");
			say($fc, "$remote_login", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$dest_filer\'s Administrative password.*:");
			(undef, undef, $after) =
				say($fc, "\r", '-nocr', '-swallow');
		}

		if ($after !~ /DR destination set/gm) {
			die("DR destination can be set");
		}

		return 1;
	};
}

=head2 vfiler_dr_resync($fc, $vfname, $dest_filer, %args)

resync a source with an activated destination or resync a
destination with a source

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of a vfiler

=item dest_filer

Required: name of the destination filer

=item args

Required: arguments

=back

Examples:

	vfiler_dr_resync($fc, 'vFiler100', 'dest_filername');

	vfiler_dr_resync($fc, 'vFiler100', 'dest_filername'
			alt_local_ip=>"111.111.111.111",
			alt_remote_ip=>"111.111.111.112",
			remote_login=>"root"
			soption=>"-s"
			);
	vfiler_dr_resync($fc,'vfiler100','dest_filername',
			remote_login=>"root",
			use_secure_channel=>"true");
=cut

sub vfiler_dr_resync
{
	my ($fc, $vfname, $dest_filer, %args) = @_;

	my $remote_login = $args{remote_login} || "root";
	my $remote_passwd = $args{remote_passwd} || "";
	my $alt_local_ip = $args{alt_local_ip} || "";
	my $alt_remote_ip = $args{alt_remote_ip} || "";
	my $soption = $args{soption} || "";
	my $loption = $args{loption} || "";
	my $aoption = $args{aoption} || "";
	my $timeout = $args{timeout} || 240;
	my $use_secure_channel=$args{use_secure_channel}||"false";
	
	my $coption ="";
	if($use_secure_channel eq "true"){
		$coption='-c secure';
	}
	
	my ($l_option, $a_option);

	if ($loption) {
		$l_option = "$loption $remote_login:$remote_passwd";
	} else {
                $l_option = "";
        }
	if ($aoption) {
		$a_option = "$aoption $alt_local_ip,$alt_remote_ip";
	} else {
                $a_option = "";
        }

	my $fs_version = get_wafl_version($fc->{hostp});
	if ($fs_version < WV_TSINGTAO_MIN) {
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION));
	}

	return eval {
		sleep(4);

	   	$fc->timeout($timeout);
		say($fc, "vfiler dr resync $l_option $soption $a_option $coption $vfname\@$dest_filer", '-noprompt');
		if ($soption && !$loption && !$aoption) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Alternate IP address\/Hostname for remote filer.*:");
			say($fc, "$alt_remote_ip\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Alternate IP address\/Hostname for local filer.*:");
			say($fc, "$alt_local_ip\r", '-nocr', '-blind');
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"$dest_filer\'s Administrative login.*:");
		say($fc, "$remote_login\r", '-nocr', '-blind');
		say($fc, "", '-noprompt', '-nocr', '-wait',
			"$dest_filer\'s Administrative password.*:");
		my (undef, undef, $after) =
				say($fc, "$remote_passwd\r", '-nocr', '-swallow');
		if ($after !~ /Revert to resync|Starting snapmirror/) {
			die("vfiler dr resync failed");
		}

		sleep(4);

		my $text = sayEOF($fc, "vfiler status $vfname");
		if ($text !~ /stopped, DR backup/) {
			die("didn't return an expected vfiler status after vfiler dr resync");
		}

		return 1;
	};
}

=head2 vfiler_dr_configure($fc, $remote_vfiler, $remote_filer, %args)

configures the specified remote vfiler from the specified remote filer for
disaster recovery on the local filer

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the the remote vfiler

=item remote_filer

Required: name of the remote filer

=item args

Required: arguments

=back

Examples:

	vfiler_dr_configure($fc, 'vFiler100', 'nwk-u31',
		remote_login => "root",
		if_to_use => "$VFILER_IF_TO_USE",
		ifnames_ipaddrs_netmasks =>
			"$VFILER_IF_TO_USE:$VFILER_IPADDR_1:$VFILER_NETMASK",
		DNSserveripaddrs => "$VFILER_DNS_SERVER",
		NISserveripaddrs => "$VFILER_NIS_SERVER");

	vfiler_dr_configure($fc, 'vFiler100', 'nwk-u31',
		soption => "-s",
		remote_login => "root",
		remotefiler_ipaddr => "$filerA_ip",
		localfiler_ipaddr => "$filerB_ip",
		if_to_use => "$VFILER_IF_TO_USE");

	vfiler_dr_configure($fc, 'vFiler100', 'nwk-u31',
		remote_login => "root",
		remotefiler_ipaddr => "$filerA_ip",
		localfiler_ipaddr => "$filerB_ip",
		if_to_use => "$VFILER_IF_TO_USE");
	
	vfiler_dr_configure($fc,'vfiler100','nwk-u31',
	   		  remote_login=>"root",
	  		  if_to_use => "$VFILER_IF_TO_USE",
          		  ifnames_ipaddrs_netmasks =>
                    "$VFILER_IF_TO_USE:$VFILER_IPADDR_1:$VFILER_NETMASK",
                    DNSserveripaddrs => "$VFILER_DNS_SERVER",
                    NISserveripaddrs => "$VFILER_NIS_SERVER",
			  use_secure_channel=>"true");

=cut

sub vfiler_dr_configure($$$%)
{
	my ($fc, $remote_vfiler, $remote_filer, %args) = @_;

	my $soption = $args{soption} || "";
	my $remotefiler_ipaddr = $args{remotefiler_ipaddr} || "";
	my $localfiler_ipaddr = $args{localfiler_ipaddr} || "";
	my $if_to_use = $args{if_to_use} || "";
	my $use_secure_channel=$args{use_secure_channel}||"false";

	my $remote_login = $args{remote_login} || "";
	my $remote_passwd = $args{remote_passwd} || "";
	my $ifnames_ipaddrs_netmasks = $args{ifnames_ipaddrs_netmasks} || "";
	my $DNSserveripaddrs = $args{DNSserveripaddrs} || "";
	my $NISserveripaddrs = $args{NISserveripaddrs} || "";
	my $alt_src = $args{alt_src} || "";
	my $alt_dst = $args{alt_dst} || "";
	my $timeout = $args{timeout} || 240;

	my $loption = "-l $remote_login:$remote_passwd";
	my $eoption = "-e $ifnames_ipaddrs_netmasks";
	my $doption = "-d $DNSserveripaddrs";
	my $noption = "-n $NISserveripaddrs";

	my $aoption = "";
	if ($alt_src && $alt_dst) {
		 $aoption = "-a $alt_src,$alt_dst";
	}
	my $coption="";
	if($use_secure_channel eq "true"){
		$coption='-c secure';
	}

	return eval {
		sleep(4);

	   	$fc->timeout($timeout);
		my (undef, undef, $after) = say($fc, "vfiler status $remote_vfiler",
						-swallow);
		if ($after =~ /stopped/) {
			die("can't continue because $remote_vfiler is already "
				."existed, but it is stopped");
		} elsif ($after =~ /not found/) {
			;
		} else {
			die("$remote_vfiler is already running");
		}

		my $fs_version = get_wafl_version($fc->{hostp});

		if ($remote_login && $ifnames_ipaddrs_netmasks
			&& $DNSserveripaddrs && $NISserveripaddrs) {
			if ($soption) {
				if ($aoption) {
					my (undef, undef, $after) = say($fc,
					"vfiler dr configure $loption $eoption $doption $noption $soption $aoption $coption  $remote_vfiler\@$remote_filer",
					-swallow);

					die "Failed to configure DR\n" if
					(($after =~ /$remote_vfiler\s+is\s+offline/gm) ||
					($after =~ /$remote_vfiler\s+does\s+not\s+exist/gm));

					die "vfiler dr configure failed\n" unless
					($after =~ /SnapMirror transfer initiated for vfiler/gm);
				} else {
					say($fc,
					"vfiler dr configure $loption $eoption $doption $noption $soption $coption $remote_vfiler\@$remote_filer",
					'-noprompt');
					if ($fs_version > WV_ANCHORSTEAM_MAX) {
						say($fc, "", '-noprompt', '-nocr', '-wait',
						"Alternate IP address\/Hostname for remote filer $remote_filer.*:");
					} else {
						say($fc, "", '-noprompt', '-nocr', '-wait',
						"$remote_filer\'s alternate IP address\/hostname for Sync SnapMirror connection.*:");
					}
					say($fc, "$remotefiler_ipaddr", '-blind');
					if ($fs_version > WV_ANCHORSTEAM_MAX) {
						say($fc, "", '-noprompt', '-nocr', '-wait',
						"Alternate IP address\/Hostname for local filer.*:");
					} else {
						say($fc, "", '-noprompt', '-nocr', '-wait',
						"Local alternate IP address\/hostname for Sync SnapMirror connection.*:");
					}
					my (undef, undef, $after) =
						say($fc, "$localfiler_ipaddr", '-swallow');
					die "vfiler dr configure failed\n" unless
					($after =~ /SnapMirror transfer initiated for vfiler/gm);
				}
			} else {
				my (undef, undef, $after) = say($fc,
				"vfiler dr configure $loption $eoption $doption $noption $aoption $coption $remote_vfiler\@$remote_filer",
				-swallow);

				die "Failed to configure DR\n" if
				(($after =~ /$remote_vfiler\s+is\s+offline/gm) ||
				($after =~ /$remote_vfiler\s+does\s+not\s+exist/gm));

				die "vfiler dr configure failed\n" unless
				($after =~ /SnapMirror transfer initiated for vfiler/gm);
			}
		} else {
			if ($soption) {
				say($fc,
				"vfiler dr configure $soption $aoption $coption $remote_vfiler\@$remote_filer",
				'-noprompt');

				if (!$aoption) {
					if ($fs_version > WV_ANCHORSTEAM_MAX) {
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"Alternate IP address\/Hostname for remote filer $remote_filer.*:");
					say($fc, "$remotefiler_ipaddr\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"Alternate IP address\/Hostname for local filer.*:");
					say($fc, "$localfiler_ipaddr\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s Administrative login.*:");
					say($fc, "$remote_login\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s Administrative password.*:");
					say($fc, "$remote_passwd\r", '-nocr', '-blind');
					} else {
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s Administrative login.*:");
					say($fc, "$remote_login\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s Administrative password.*:");
					say($fc, "$remote_passwd\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s alternate IP address\/hostname for Sync SnapMirror connection.*:");
					say($fc, "$remotefiler_ipaddr\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"Local alternate IP address\/hostname for Sync SnapMirror connection.*:");
					say($fc, "$localfiler_ipaddr\r", '-nocr', '-blind');
					}
				} else {
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s Administrative login.*:");
					say($fc, "$remote_login\r", '-nocr', '-blind');
					say($fc, "", '-noprompt', '-nocr', '-wait',
					"$remote_filer\'s Administrative password.*:");
					say($fc, "$remote_passwd\r", '-nocr', '-blind');
                                }

				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Upon Activation, configure vfiler IP address.*?");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Interface to assign this address to.*:");
				say($fc, "$if_to_use", '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
						"Netmask to use:.*:");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"set of DNS servers from the production vfiler.*?");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"set of preferred NIS servers from the production vfiler.*?");
				my (undef, undef, $after) =
					say($fc, "\r", '-nocr', '-swallow');
				die "vfiler dr configure failed\n" unless
				($after =~ /SnapMirror transfer initiated for vfiler/gm);
			} else {
				say($fc,
				"vfiler dr configure $remote_vfiler\@$remote_filer",
				'-noprompt');

				say($fc,
				"", '-noprompt', '-nocr', '-wait',
				"$remote_filer\'s Administrative login.*:");
                                say($fc, "$remote_login\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"$remote_filer\'s Administrative password.*:");
				say($fc, "$remote_passwd\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Upon Activation, configure vfiler IP address.*?");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Interface to assign this address to.*:");
				say($fc, "$if_to_use\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
						"Netmask to use:.*:");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"set of DNS servers from the production vfiler.*?");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"set of preferred NIS servers from the production vfiler.*?");
				my (undef, undef, $after) =
					say($fc, "\r", '-nocr', '-swallow');
				die "vfiler dr configure failed\n" unless
				($after =~ /SnapMirror transfer initiated for vfiler/gm);
			}
		}

		sleep(30);

		return 1;
	};
}

=head2 vfiler_dr_status($fc, $remote_vfiler, $remote_filer, $volname)

monitor the status of this mirror

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the remote vfiler

=item remote_filer

Required: name of the remote filer

=item volname

Required: name of volume

=back

Examples:

	vfiler_dr_status($fc, 'vFiler100', 'nwk-u31', 'vvolVFT0');

=cut

sub vfiler_dr_status($$$$)
{
	my ($fc, $remote_vfiler, $remote_filer, $volname) = @_;
        my ($line, %return);

	return eval {
		sleep(4);

		my (undef, undef, $after) = say($fc,
			"vfiler dr status $remote_vfiler\@$remote_filer", -swallow);
                if ($volname) { # We are looking for a specific volume's status
                    if (wantarray) { # We want to know the status of this volume
                        $after =~ s/\r//g;
                        unless ($after =~ /^($volname)\s+(.*)$/m) {
                            die "Volume $volname not found in vfiler dr status";
                        }
                        return ( $1 => $2 );
                    } else { # We only want to know if it is snapmirrored or not
                        die "$volname is not snapmirrored\n"
                          unless ($after =~ /$volname\s+Snapmirrored|$volname\s+snapmirrored/gm);

                        return 1;
                    }
                } else { # We are not looking for a specific volume, return all volumes' status
                    if (wantarray) { # We are looking for status of all volumes
                        $after =~ s/\r//g;
                        my @after = split('\n', $after);
                        while ($line = shift @after) {
                            next if $line !~ /^(\w+)\s+(snapmirrored|initializing)$/i;
                            $return{$1} = $2;
                        }
                        if (!keys(%return)) {
                            die "Unable to parse vfiler dr status output";
                        }
                        return %return;
                    } else { # We didn't ask for a specific volume, and aren't looking
                             # for a scalar, indicate if any volume is snapmirrored.
                        die "No volumes are snapmirrored"
                          unless ($after =~ /$volname\s+Snapmirrored|$volname\s+snapmirrored/gm);
                        return 1;
                    }
                }
	};
}


=head2 vfiler_dr_activate($fc, $remote_vfiler, $remote_filer)

activate the remote vfiler on the local machine

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the remote vfiler

=item remote_filer

Required: name of the remote filer

=back

Examples:

	vfiler_dr_activate($fc, 'vFiler100', 'nwk-u31');

=cut

sub vfiler_dr_activate($$$)
{
	my ($fc, $remote_vfiler, $remote_filer) = @_;

	return eval {
		sleep(4);

		my (undef, undef, $after) = say($fc,
			"vfiler dr activate $remote_vfiler\@$remote_filer", -swallow);
		die "$remote_vfiler is not activated\n"
			unless ($after =~ /Vfiler\s+$remote_vfiler\s+activated/gm);
		die "$remote_vfiler is not found\n"
			if ($after =~ /Cannot\s+find/);

		sleep(4);

		return 1;
	};
}

=head2 vfiler_dr_delete($fc, $remote_vfiler, $remote_filer, [$option])

remove the DR configuration

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the remote vfiler

=item remote_filer

Required: name of the remote filer

=item option

Optional: -f option

=back

Examples:

	vfiler_dr_delete($fc, 'vFiler100', 'nwk-u31');
	vfiler_dr_delete($fc, 'vFiler100', 'nwk-u31', '-f');

=cut


sub vfiler_dr_delete($$$;$)
{
	my ($fc, $remote_vfiler, $remote_filer, $option) = @_;

	return eval {
		sleep(4);

		my $text;
		if ($option) {
			if ($option eq "-f") {
				$text = sayEOF($fc,
				"vfiler dr delete $option $remote_vfiler\@$remote_filer");
			} else {
				die "$option is not valid option\n";
			}
		} else {
			$text = sayEOF($fc,
				"vfiler dr status $remote_vfiler\@$remote_filer");
		}
		die "DR configuration is not deleted for vfiler\n"
		unless ($text =~ /DR\s+configuration\s+deleted\s+for\s+vfiler/);

		sleep(4);

		return 1;
	};
}

=head2 vfiler_migrate($fc, $remote_vfiler, $remote_filer, %args)

moves the specified remote vfiler from the specified remote filer
to the local filer

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the source vfiler

=item remote_filer

Required: name of the source filer

=item args

Required: arguments

=back

Examples:

	vfiler_migrate($fc, 'vFiler100', 'nwk-u31',
		remote_login => "root",
		if_to_use => "$VFILER_IF_TO_USE",
		ifnames_ipaddrs_netmasks =>
			"$VFILER_IF_TO_USE:$VFILER_IPADDR_1:$VFILER_NETMASK");

=cut

sub vfiler_migrate($$$%)
{
	my ($fc, $remote_vfiler, $remote_filer, %args) = @_;

	my $method = $args{method} || "";
	my $if_to_use = $args{if_to_use} || "";
	my $remote_login = $args{remote_login} || "";
	my $remote_passwd = $args{remote_passwd} || "";
	my $ifnames_ipaddrs_netmasks = $args{ifnames_ipaddrs_netmasks} || "";
	my $timeout = $args{timeout} || 240;
	my $use_secure_channel=$args{use_secure_channel} ||"false";


	my $moption = "";
	my $coption="";
	if ($method eq "nocopy") {
		$moption = "-m $method";
	}

	my $loption = "-l $remote_login:$remote_passwd";
	my $eoption = "-e $ifnames_ipaddrs_netmasks";
	if ($use_secure_channel eq "true") {
		$coption='-c secure';
	}

	return eval {
		sleep(4);

	   	$fc->timeout($timeout);
		my (undef, undef, $after) = say($fc, "vfiler status $remote_vfiler",
						-swallow);
		if ($after =~ /stopped/) {
			(undef, undef, $after) =
				say($fc, "vfiler status $remote_vfiler", -swallow);
			if ($after =~ /stopped/) {
				die("can't continue because $remote_vfiler is ".
					"already existed, but it is stopped");
			}
		}
		if ($after =~ /not found/) {
			;
		} else {
			die("$remote_vfiler is already running");
		}

		if ($remote_login && $ifnames_ipaddrs_netmasks) {

			my (undef, undef, $after) = say($fc,
			"vfiler migrate $loption $eoption $moption $coption $remote_vfiler\@$remote_filer",
				-swallow);
			if ($after =~ /busy destination. retrying/gm) {
				sleep(60);
			} elsif ($after =~
				/$remote_vfiler\s+is\s+already\s+being\s+migrated/gm) {
				die("vfiler migrate failed because $remote_vfiler".
					" is already being migrated");
			} else {
				if ($after !~ /Vfiler\s+$remote_vfiler\s+migrated/gm) {
					die("vfiler migrate failed: $after");
				}
			}

		} else {
			say($fc,
			"vfiler migrate $moption $coption $remote_vfiler\@$remote_filer",
			'-noprompt');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$remote_filer\'s Administrative login.*:");
			say($fc, "$remote_login", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$remote_filer\'s Administrative password.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"After migrate, configure vfiler IP address.*?");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Interface to assign this address to.*:");
			say($fc, "$if_to_use", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
					"Netmask to use:.*:");
			my (undef, undef, $after) =
				say($fc, "\r", '-nocr', '-swallow');
			if ($after =~ /busy destination. retrying/gm) {
				sleep(60);
			} elsif ($after =~
				/$remote_vfiler\s+is\s+already\s+being\s+migrated/gm) {
				die("vfiler migrate failed because $remote_vfiler".
					" is already being migrated");
			} else {
				if ($after !~ /Vfiler\s+$remote_vfiler\s+migrated/gm) {
					die("vfiler migrate failed: $after");
				}
			}
		}

		sleep(60);

		return 1;
	};
}

=head2 vfiler_migrate_start($fc, $remote_vfiler, $remote_filer, %args)

initiates the movement of the remote vfiler's storage to the local filer
using snapMirror

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the source vfiler

=item remote_filer

Required: name of the source filer

=item args

Required: arguments

=back

Examples:

	vfiler_migrate_start($fc, 'vFiler100', 'nwk-u31',
		remote_login => "root",
		ifnames_ipaddrs_netmasks =>
			"$VFILER_IF_TO_USE:$VFILER_IPADDR_1:$VFILER_NETMASK");

	vfiler_migrate_start($fc, 'vFiler100', 'nwk-u31',
		remote_login => "root",
		if_to_use => "$VFILER_IF_TO_USE",
		ifnames_ipaddrs_netmasks =>
			"$VFILER_IF_TO_USE:$VFILER_IPADDR_1:$VFILER_NETMASK");

=cut

sub vfiler_migrate_start($$$%)
{
	my ($fc, $remote_vfiler, $remote_filer, %args) = @_;

	my $method = $args{method} || "";
	my $if_to_use = $args{if_to_use} || "";
	my $remote_login = $args{remote_login} || "";
	my $remote_passwd = $args{remote_passwd} || "";
	my $ifnames_ipaddrs_netmasks = $args{ifnames_ipaddrs_netmasks} || "";
	my $timeout = $args{timeout} || 240;
	my $use_secure_channel=$args{use_secure_channel} ||"false";
	my $moption = "";
	my $coption="";
	if ($method eq "nocopy") {
		$moption = "-m $method";
	}
	if ($use_secure_channel eq "true") {
		$coption='-c secure';
	}
	my $loption = "-l $remote_login:$remote_passwd";
	my $eoption = "-e $ifnames_ipaddrs_netmasks";

	return eval {
		sleep(4);

	   	$fc->timeout($timeout);
		my (undef, undef, $after) = say($fc, "vfiler status $remote_vfiler",
						-swallow);
		if ($after =~ /stopped/) {
			die("can't continue because $remote_vfiler is already "
				."existed, but it is stopped");
		} elsif ($after =~ /not found/) {
			;
		} else {
			die("$remote_vfiler is already running");
		}

		if ($remote_login && $ifnames_ipaddrs_netmasks) {

			my (undef, undef, $after) = say($fc,
			"vfiler migrate start $loption $eoption $moption $coption $remote_vfiler\@$remote_filer",
				-swallow);
			
			die "vfiler migrate failed\n" unless
			($after =~ /Data\s+transfer\s+for\s+vfiler\s+storage\s+units\s+initiated/gm);
		} else {
			say($fc,
			"vfiler migrate start $moption $coption $remote_vfiler\@$remote_filer",
			'-noprompt');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$remote_filer\'s Administrative login.*:");
			say($fc, "$remote_login", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"$remote_filer\'s Administrative password.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"After migrate, configure vfiler IP address.*?");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Interface to assign this address to.*:");
			say($fc, "$if_to_use", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
					"Netmask to use:.*:");
			my (undef, undef, $after) =
				say($fc, "\r", '-nocr', '-swallow');
			die "vfiler migrate failed\n" unless
			($after =~ /Data\s+transfer\s+for\s+vfiler\s+storage\s+units\s+initiated/gm);
		}

		return 1;
	};
}

=head2 vfiler_migrate_status($fc, $remote_vfiler, $remote_filer, $volname)

monitor the status of the data movement

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the remote vfiler

=item remote_filer

Required: name of the remote filer

=item volname

Required: name of volume

=back

Examples:

	vfiler_migrate_status($fc, 'vFiler100', 'nwk-u31', 'vvolVFT0');

=cut

sub vfiler_migrate_status($$$$)
{
	my ($fc, $remote_vfiler, $remote_filer, $volname) = @_;

	return eval {
		sleep(4);

		my (undef, undef, $after) = say($fc,
			"vfiler migrate status $remote_vfiler\@$remote_filer",
			-swallow);
		die "$volname is not snapmirrored\n"
			unless ($after =~ /$volname\s+Snapmirrored|$volname\s+snapmirrored/gm);

		return 1;
	};
}

=head2 vfiler_migrate_cancel($fc, $remote_vfiler, $remote_filer)

cancel the data movement

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the remote vfiler

=item remote_filer

Required: name of the remote filer

=back

Examples:

	vfiler_migrate_cancel($fc, 'vFiler100', 'nwk-u31');

=cut

sub vfiler_migrate_cancel($$$%)
{
	my ($fc, $remote_vfiler, $remote_filer,%args) = @_;
	my $use_secure_channel=$args{use_secure_channel} || 'false';
	my $coption="";
	if ($use_secure_channel eq 'true') {
		$coption='-c secure';
	}
	return eval {
		sleep(4);

		my (undef, undef, $after) = say($fc,
			"vfiler migrate cancel $coption $remote_vfiler\@$remote_filer",
			-swallow);
		die "Vfiler migration can't be cancelled\n"
			unless ($after =~ /Vfiler\s+migration\s+cancelled/gm);

		return 1;
	};
}

=head2 vfiler_migrate_complete($fc, $remote_vfiler, $remote_filer)

complete the data movement

=over 4

=item fc

Required: hostspec of the filer

=item remote_vfiler

Required: name of the remote vfiler

=item remote_filer

Required: name of the remote filer

=back

Examples:

	vfiler_migrate_complete($fc, 'vFiler100', 'nwk-u31');

=cut

sub vfiler_migrate_complete($$$%)
{
	my ($fc, $remote_vfiler, $remote_filer,%args) = @_;
	my $use_secure_channel=$args{use_secure_channel} || 'false';
	my $coption="";
	if ($use_secure_channel eq 'true') {
		$coption='-c secure';
	}

	return eval {
		sleep(4);

		my (undef, undef, $after) = say($fc,
			"vfiler migrate complete $coption $remote_vfiler\@$remote_filer",
			-swallow);

		sleep(4);

		die "Vfiler migration can't be cancelled\n"
			unless ($after =~ /Vfiler\s+$remote_vfiler\s+migrated/gm);

		return 1;
	};
}

=head2 vfiler_run_vfiler_ftp_enable($fc, $vfname, $homedir)

enable ftpd.enable and ftpd.anonymous.enable and set ftpd.anonymous.home_dir

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item home_dir

Required: directory name for ftpd.anonymous.home_dir

=back

Examples:

	vfiler_run_vfiler_ftp_enable($fc, 'vFiler100', '/vol/$FILER_VVOL/home');

=cut

sub vfiler_run_vfiler_ftp_enable($$$)
{
	my ($fc, $vfname, $home_dir) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /stopped|not found/) {
			die("$vfname is not running");
		}
		say($fc, "vfiler run $vfname options ftpd.enable on");
		say($fc, "vfiler run $vfname options ftpd.anonymous.enable on");
		(undef, undef, $after) =
		say($fc, "vfiler run $vfname options ftpd.anonymous.enable", -swallow);

		if ($after =~ /ftpd.anonymous.enable\s+on/) {
			say($fc,
			"vfiler run $vfname options ftpd.anonymous.home_dir $home_dir");

			(undef, undef, $after) =
			say($fc, "vfiler run $vfname options ftpd.anonymous.home_dir",
					-swallow);
			if ($after =~ /ftpd.anonymous.home_dir\s+$home_dir/) {
				return(1);
			} else {
				die "ftpd.anonymous.home_dir is not set with $home_dir\n";
			}
		} else {
			die "ftpd.anonymous.enable is off\n";
		}
	};
}

=head2 vfiler_run_vfiler_http_enable($fc, $vfname, $http_rootdir)

enable httpd.enable and set httpd.rootdir

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item http_rootdir

Required: directory name of httpd.rootdir

=back

=cut

sub vfiler_run_vfiler_http_enable($$$)
{
	my ($fc, $vfname, $http_rootdir) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /stopped|not found/) {
			die("$vfname is not running");
		}
		say($fc, "vfiler run $vfname options httpd.enable on");
		(undef, undef, $after) =
			say($fc, "vfiler run $vfname options httpd.enable", -swallow);
		if ($after =~ /httpd.enable\s+on/) {
			say($fc,
			"vfiler run $vfname options httpd.rootdir $http_rootdir");

			(undef, undef, $after) =
			say($fc, "vfiler run $vfname options httpd.rootdir",
					-swallow);
			if ($after =~ /httpd.rootdir\s+$http_rootdir/) {
				return(1);
			} else {
				die "httpd.rootdir is not set with $http_rootdir\n";
			}
		} else {
			die "httpd.enable is off\n";
		}
	};
}

=head2 vfiler_run_vfiler_exportfs($fc, $vfname, $vfFS)

export vfiler FS

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item vfFS

Required: FS of vfiler

=item persistent

Optional: boolean flag, default value is false. when true, the exports are
written into the vfiler's /etc/rc file, to make them persistent. this is
required for CFO testing.

=back

Examples:

	vfiler_run_vfiler_exportfs($fc, 'vFiler100', "/vol/volname");

=cut

sub vfiler_run_vfiler_exportfs($$$;$)
{
	my ($fc, $vfname, $vfFS, $persistent) = @_;

	return eval {
	        my $flags = "-i -o";
		if (defined($persistent) && $persistent) {
		    # persistent exports are required for vfiler CFO operations
		    $flags = "-p ";
		}

		sleep(4);
		say($fc, "vfiler run $vfname exportfs $flags anon=0 $vfFS");
		my (undef, undef, $after) =
			say($fc, "vfiler run $vfname exportfs", -swallow);
		if ($after =~ /$vfFS/) {
			return(1);
		} else {
			die "$vfFS is not exported\n";
		}
	};
}

=head2 vfiler_run_vfiler_add_default_route($fc, $vfname, [$gateway])

add default route for vFiler

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item gateway

Optional: gateway

=back

Examples:

	vfiler_run_vfiler_add_default_route($fc, 'vFiler100');

=cut

sub vfiler_run_vfiler_add_default_route($$;$)
{
	my ($fc, $vfname, $gateway) = @_;

	return eval {
		sleep(4);

		if (!$gateway) {
			my $fh = $fc->{'hostp'};
			my $filer_ip = $fh->default_ip();
			$filer_ip =~ /(\d+).(\d+).(\d+).(\d+)/s;
			$gateway = "$1."."$2."."$3."."1";
		}
		sleep(4);

		my (undef, undef, $after) =
			say($fc, "vfiler run $vfname route add default $gateway 1",
			-swallow);
		die "network unreachable\n" if ($after =~ /network\s+unreachable/gm);

		return 1;
	};
}

=head2 vfiler_run_vfiler_setup($fc, $vfname, %args)

run vFiler setup for an non-default IPspace

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item args

Required: arguments

if_to_use=>"$VFILER_IF_TO_USE"

netmask=>"$VFILER_NETMASK"

dnsdomain=>"$VFILER_DNS_DOMAIN"

dnsserver=>"$VFILER_DNS_SERVER"

nisdomain=>"$VFILER_NIS_DOMAIN"

nisserver=>"$VFILER_NIS_SERVER"

count=>1

=back

Examples:

	vfiler_run_vfiler_setup($fc, 'vFiler100',
		if_to_use=>"$VFILER_IF_TO_USE",
		netmask=>"$VFILER_NETMASK",
		dnsdomain=>"$VFILER_DNS_DOMAIN",
		dnsserver=>"$VFILER_DNS_SERVER",
		nisdomain=>"$VFILER_NIS_DOMAIN",
		nisserver=>"$VFILER_NIS_SERVER",
		count=>1
		);
	vfiler_run_vfiler_setup($fc, 'vFiler100',
		if_to_use=>"$VFILER_IF_TO_USE",
		netmask=>"$VFILER_NETMASK",
		dnsdomain=>"$VFILER_DNS_DOMAIN",
		dnsserver=>"$VFILER_DNS_SERVER",
		nisdomain=>"$VFILER_NIS_DOMAIN",
		nisserver=>"$VFILER_NIS_SERVER",
		after_recreate=>"y",
		count=>1
		);

=cut

sub vfiler_run_vfiler_setup($$%)
{
	my ($fc, $vfname, %args) = @_;

	my ($if_to_use) = $args{if_to_use} || "";
	my ($netmask) = $args{netmask} || "";
	my ($dnsdomain) = $args{dnsdomain} || "";
	my ($dnsserver) = $args{dnsserver} || "";
	my $vfiler_password=$args{vfiler_password}||"pa55w0rd";
	my ($dnsserver1) = $args{dnsserver1} || "";
	my ($dnsserver2) = $args{dnsserver2} || "";
	my ($nisdomain) = $args{nisdomain} || "";
	my ($nisserver) = $args{nisserver} || "";
	my ($after_recreate) = $args{after_recreate} || "";
	my ($ipaddr_count) = $args{count} || 1;
	my ($timeout) = $args{timeout} || 240;

	my %prompts = (
		1	=> "Configure vfiler IP address.*:",
		2	=> "Interface to assign this address to.*:",
		3	=> "Netmask to use:.*:"
	);

	my %inputs = (
		1	=> "y",
		2	=> "$if_to_use",
		3	=> "$netmask"
	);

	return eval {
		if ($if_to_use) {
			my $text = sayEOF($fc, "ifconfig $if_to_use");
			if ($text =~ /no\s+such\s+interface/) {
				die("$if_to_use: no such interface");
			}
		} else {
			die("network interface to use is not set");
		}

		$fc->timeout($timeout);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /stopped|not found/) {
			die("$vfname is not running");
		}

		sleep(4);

		say($fc, "vfiler run $vfname setup",
			'-noprompt', '-wait', "setup command will rewrite");

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Are you sure you want to continue.*?");
		say($fc, "y", '-blind');

                for (my $cnt = 1; $cnt <= $ipaddr_count; $cnt++) {
                        foreach my $key (qw (1 2 3)) {
                                say($fc, "", '-noprompt',
                                        '-nocr', '-wait', "$prompts{$key}");
                                say($fc, "$inputs{$key}", '-blind');
                        }
                }

		say($fc, "", '-noprompt', '-nocr', '-wait',
		"^Please enter the name or IP address of the administration host:.*");

		say($fc, "\r", '-nocr', '-blind');

		say($fc, "", '-noprompt', '-nocr',
			'-wait', "Do you want to run DNS resolver.*:");
                if ( $dnsdomain ne "" ) {
		    say($fc, "y", '-blind');

		    say($fc, "", '-noprompt', '-nocr',
			'-wait', "Please enter DNS domain name.*:");
		    say($fc, "$dnsdomain", '-blind');

		    say($fc, "", '-noprompt', '-nocr', '-wait',
			"Please enter the IP address for first nameserver.*:");
		    say($fc, "$dnsserver", '-blind');

		    if ($dnsserver1) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want another nameserver.*:");
			say($fc, "y", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Please enter the IP address for alternate nameserver.*:");
			say($fc, "$dnsserver1", '-blind');
			if ($dnsserver2) {
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Do you want another nameserver.*:");
				say($fc, "y", '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Please enter the IP address for alternate nameserver.*:");
				say($fc, "$dnsserver2", '-blind');
			} else {
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Do you want another nameserver.*:");
				say($fc, "n", '-blind');
			}
		    } else {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want another nameserver.*:");
		        say($fc, "n", '-blind');
                    }
		} else {
                    # If nisserver is not passed ( ie "" ), then the test
                    # should send <cr> or "n" for the NIS questions.
                    # Typing "n" is better than <cr>, since <cr> will imply
                    # default values which could be "y"
                    say($fc, "n", '-blind');
		}

		say($fc, "", '-noprompt', '-nocr',
			'-wait', "Do you want to run NIS client.*:");
                if ( $nisdomain ne "" ) {
		    say($fc, "y", '-blind');

		    say($fc, "", '-noprompt', '-nocr',
			'-wait', "Please enter NIS domain name.*:");
		    say($fc, "$nisdomain", '-blind');

		    say($fc, "", '-noprompt', '-nocr',
			'-wait', "Please enter list of preferred NIS servers.*:");
		    say($fc, "$nisserver", '-blind');

		} else {
                    # If dnsserver is not passed ( ie "" ), then the test
                    # should send <cr> or "n" for the DNS questions.
                    # Typing "n" is better than <cr>, since <cr> will imply
                    # default values which could be "y
                    say($fc, "n", '-blind');
                }

		if (!$after_recreate) {

			say($fc, "", '-noprompt', '-nocr', '-wait', "^New password:");
   		        say($fc, "$vfiler_password", '-blind');
   
   		        say($fc, "", '-noprompt', '-nocr', '-wait', "^Retype new password:");
   		        say($fc, "$vfiler_password", '-blind');

                        say($fc, "", '-noprompt', '-nocr', '-wait', "passwd for user \'root\' changed");
                        say($fc, "\r", '-nocr', '-blind');

		}

		sleep(3);
		return(1);
	};
}

=head2 vfiler_run_vfiler_cifs_setup($fc, $vfname, %args)

run vFiler cifs setup for an non-default IPspace

use Windows Domain authentication as a default type of user authentication

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item args

Required: arguments

wins_ipaddr1=>"$VFILER_WINS_IPADDR_1"

wins_ipaddr2=>"$VFILER_WINS_IPADDR_2"

cifs_name=>"$FILER"."cf"

wins_domain=>"$VFILER_WINS_DOMAIN"

wins_user=>"$VFILER_WINS_USER"

wins_user_passwd=>"$VFILER_WINS_USER_PASSWD"

=back

Examples:

	vfiler_run_vfiler_cifs_setup($fc, 'vFiler100',
			cifs_name=>"$FILERB"."cf",
			wins_domain=>"$VFILER_WINS_DOMAIN",
			wins_user=>"$VFILER_WINS_USER",
			wins_user_passwd=>"$VFILER_WINS_USER_PASSWD");

=cut

sub vfiler_run_vfiler_cifs_setup($$%)
{
	my ($fc, $vfname, %args) = @_;

	my $wins_ipaddr1 = $args{wins_ipaddr1} || "";
	my $wins_ipaddr2 = $args{wins_ipaddr2} || "";
	my $cifs_name = $args{cifs_name} || "";
	my $wins_domain = $args{wins_domain} || "";
	my $wins_user = $args{wins_user} || "";
	my $wins_user_passwd = $args{wins_user_passwd} || "";
	my $wins_ipaddr_count = $args{wins_ipaddr_count} || 0;
	my $workgroup_name = $args{workgroup_name} || "";
	my $user_auth_type = $args{user_auth_type} || "";
	my $after_migration = $args{after_migration} || "";
	my $timeout = $args{timeout} || 240;

	if (length($cifs_name) > 15) {
		$cifs_name = substr($cifs_name, 0, 15);
	}

	return eval {
	   sleep(4);

	   $fc->timeout($timeout);
	   my (undef, undef, $after) =
		say($fc, "vfiler status $vfname", -swallow);
	   if ($after =~ /stopped|not found/) {
		die("$vfname is not running");
	   }

	   my $old_cifs_setup = 0;
	   my $os_version = get_ontap_version($fc->{hostp}) or die $@;
	   my $fs_version = get_wafl_version($fc->{hostp}) or die $@;
	   if ($os_version =~ /7.0/) {
		my $text = sayEOF($fc, "version");
		$old_cifs_setup = ($text =~ /R7.0xN|7.0X|7.0RC|7.0P/) ? 1 : 0;
	   }

	   if (($os_version =~ /6.5|6.4/) ||
			(($os_version =~ /7.0/ && $old_cifs_setup))) {
	   	say($fc, "vfiler run $vfname cifs setup",
			'-noprompt', '-wait', "Enable CIFS access to the filer");
		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you wish to make the system visible via WINS.*:");
		say($fc, "y", '-blind');

		if ($wins_ipaddr_count > 0) {
			say($fc, "", '-noprompt', '-nocr',
				'-wait', "Please enter next WINS IP address.*:");
			say($fc, "$wins_ipaddr1", '-blind');
			if ($wins_ipaddr_count == 2) {
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"Please enter next WINS IP address.*:");
				say($fc, "$wins_ipaddr2", '-blind');
			}
		}

		say($fc, "", '-noprompt', '-nocr',
			'-wait', "Please enter next WINS IP address.*:");
		say($fc, "\r", '-nocr', '-blind');

		if ($after_migration) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to configure this filer as a NTFS-only.*?");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to delete the existing filer account.*:");
			say($fc, "y", '-blind');
		}

		say($fc, "", '-noprompt', '-nocr',
			'-wait', "Do you want to modify this name.*:");
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr',
			'-wait', "Enter the CIFS name for the filer:");
		(undef, undef, $after) = say($fc, "$cifs_name", '-blind');
		if ($after =~ /Enter\s+the\s+CIFS\s+name/) {
			if (length($cifs_name) > 15) {
				$cifs_name = substr($cifs_name, 0, 14);
			}
			$cifs_name = "$cifs_name"."1";
			say($fc, "$cifs_name", '-blind');
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"What type of authentication will this filer use.*:");

		if (($user_auth_type && $workgroup_name) ||
			(ref($fc->hostp) =~ /Simulator/)) {
			say($fc, "2\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Do you want to modify this name.*:");
			say($fc, "y", '-blind');
			if (ref($fc->hostp) =~ /Simulator/) {
				my $filer_name = $fc->hostp()->hostname();
				$workgroup_name = "$filer_name"."w";
			}
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the Workgroup for the filer.*:");
			(undef, undef, $after) =
				say($fc, "$workgroup_name", '-blind');
			if ($after =~ /Enter\s+the\s+Workgroup/) {
				$workgroup_name = "$workgroup_name"."1";
				say($fc, "$workgroup_name", '-blind');
			}

#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "Create the.*:");
			say($fc, "\r", '-nocr', '-blind');
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "^Set the password for.*:");
			say($fc, "\r", '-nocr', '-blind');
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "^Retype the password:");
			say($fc, "\r", '-nocr', '-blind');
		} else {
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the Windows Domain for the filer.*:");
			say($fc, "$wins_domain", '-blind');

			if (!$after_migration) {
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"Create the.*:");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"^Set the password for.*:");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"^Retype the password:");
				say($fc, "\r", '-nocr', '-blind');
			}

			say($fc, "", '-noprompt', '-nocr',
				'-wait', "Please enter the Windows 2000 user.*:");
			say($fc, "$wins_user", '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"^Password for $wins_user:");
			say($fc, "$wins_user_passwd", '-blind');

		}
	   } else {
	   	say($fc, "vfiler run $vfname cifs setup",
			'-noprompt', '-wait', "This process will enable CIFS access");
		if ($after_migration) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to continue and change the current filer account information.*:");
			say($fc, "y", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to make the system visible via WINS.*:");
			say($fc, "n", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Would you like to reconfigure this filer to be an NTFS-only filer.*:");
			say($fc, "n", '-blind');
		} else {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to make the system visible via WINS.*:");
			say($fc, "n", '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Selection.*:");
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Should CIFS create default.*:");
			say($fc, "y", '-blind');
			say($fc, "", '-noprompt', '-nocr',
				'-wait', "^Enter the password for the root user.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr',
				'-wait', "^Retype the password:");
			say($fc, "\r", '-nocr', '-blind');

	   		if ($fs_version > WV_ANCHORSTEAM_MAX) {
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Would you like to enable NIS group caching.*:");
				say($fc, "y", '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"Enter the hour.*:");
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "", '-noprompt', '-nocr', '-wait',
				"Would you like to specify additional hours.*:");
				say($fc, "\r", '-nocr', '-blind');
			}
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Would you like to change this name.*:");
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Enter the CIFS server name for the filer.*:");
		my (undef, undef, $after) = say($fc, "$cifs_name", '-blind');
		if ($after =~ /Enter\s+the\s+CIFS\s+server/) {
			if (length($cifs_name) > 15) {
				$cifs_name = substr($cifs_name, 0, 14);
			}
			$cifs_name = "$cifs_name"."1";
			say($fc, "$cifs_name", '-blind');
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Selection.*:");

		if (($user_auth_type && $workgroup_name) ||
			(ref($fc->hostp) =~ /Simulator/)) {
			say($fc, "3\r", '-nocr', '-blind');

			if (ref($fc->hostp) =~ /Simulator/) {
				my $filer_name = $fc->hostp()->hostname();
				$workgroup_name = "$filer_name"."w";
			}
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"What is the name of the Workgroup.*:");
			(undef, undef, $after) =
				say($fc, "$workgroup_name", '-blind');
			if ($after =~ /What\s+is\s+the\s+name/) {
				$workgroup_name = "$workgroup_name"."1";
				say($fc, "$workgroup_name", '-blind');
			}
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "Do you want to create.*:");
			say($fc, "\r", '-nocr', '-blind');
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "^Enter the new password for.*:");
			say($fc, "\r", '-nocr', '-blind');
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "^Retype the password:");
			say($fc, "\r", '-nocr', '-blind');
		} else {
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"What is the name of the Active Directory domain.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Enter the name of the Windows user.*:");
			say($fc, "$wins_user", '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"^Password for $wins_user:");
			say($fc, "$wins_user_passwd", '-blind');

			say($fc, "\r", '-nocr', '-blind');

			if (!$after_migration) {
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "\r", '-nocr', '-blind');
				say($fc, "\r", '-nocr', '-blind');
			}
		}
	   }
	   say($fc, "\r", '-nocr', '-blind');
	   sleep(3);
	   return(1);
	};
}

=head2 vfiler_allow($fc, $vfilertemplate, [%args])

allows the use of the specified protocols on the vfiler(s) specified
in the vfilertemplate

=over 4

=item fc

Required: hostspec of the filer

=item vfilertemplate

Required: name of vfiler or *

=item args

Optional: arguments

	cifs => "proto=cifs"
	nfs => "proto=nfs"
	rsh => "proto=rsh"
	iscsi => "proto=iscsi"
	ftp => "proto=ftp"
	http => "proto=http"

=back

Examples:

	vfiler_allow($fc, 'vFiler100');
	vfiler_allow($fc, 'vFiler100', cifs => "proto=cifs");
	vfiler_allow($fc, '*');
	vfiler_allow($fc, '*', cifs => "proto=cifs", nfs => "proto=nfs");

=cut

sub vfiler_allow($$;%)
{
	my ($fc, $vfilertemplate, %args) = @_;

	my $cifs = $args{cifs} || "";
	my $nfs = $args{nfs} || "";
	my $rsh = $args{rsh} || "";
	my $iscsi = $args{iscsi} || "";
	my $ftp = $args{ftp} || "";
	my $http = $args{http} || "";

	if ($vfilertemplate eq "*") {
		$vfilertemplate = "*";
	}

	return eval {
		sleep(4);
		my $text = sayEOF($fc,
			"vfiler allow $vfilertemplate $cifs $nfs $rsh $iscsi $ftp $http");
		die "vfiler allow: incorrect usage\n"
			if ($text =~ /vfiler: incorrect usage/);

		return 1;
	};
}

=head2 vfiler_disallow($fc, $vfilertemplate, [%args])

disallows the use of the specified protocols on the vfiler(s) specified
in the vfilertemplate

=over 4

=item fc

Required: hostspec of the filer

=item vfilertemplate

Required: name of vfiler or *

=item args

Optional: arguments

	cifs => "proto=cifs"
	nfs => "proto=nfs"
	rsh => "proto=rsh"
	iscsi => "proto=iscsi"
	ftp => "proto=ftp"
	http => "proto=http"

=back

Examples:

	vfiler_disallow($fc, 'vFiler100');
	vfiler_disallow($fc, 'vFiler100', cifs => "proto=cifs");
	vfiler_disallow($fc, '*');
	vfiler_disallow($fc, '*', cifs => "proto=cifs", nfs => "proto=nfs");

=cut

sub vfiler_disallow($$;%)
{
	my ($fc, $vfilertemplate, %args) = @_;

	my $cifs = $args{cifs} || "";
	my $nfs = $args{nfs} || "";
	my $rsh = $args{rsh} || "";
	my $iscsi = $args{iscsi} || "";
	my $ftp = $args{ftp} || "";
	my $http = $args{http} || "";

	if ($vfilertemplate eq "*") {
		$vfilertemplate = "*";
	}

	return eval {
		sleep(4);
		my $text = sayEOF($fc,
		"vfiler disallow $vfilertemplate $cifs $nfs $rsh $iscsi $ftp $http");
		die "vfiler disallow: incorrect usage\n"
			if ($text =~ /vfiler: incorrect usage/);

		return 1;
	};
}

=head2 vfiler_stop_destroy($fc, $vfname, [$remote_filer])

causes the matching vfilers to stop receiving network traffic and
releases all resources and removes the configuration information
associated with the named vfiler

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=back

Examples:

	vfiler_stop_destroy($fc, 'vFiler100');
	vfiler_stop_destroy($fc, 'vFiler100', 'filername');

=cut

sub vfiler_stop_destroy($$;$)
{
	my ($fc, $vfname, $remote_filer) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /running/) {
			say($fc, "vfiler stop $vfname");
			say($fc, "vfiler destroy $vfname -f");
			return(1);
		} elsif ($after =~ /$vfname\s+stopped/s) {
			my $text = sayEOF($fc, "vfiler destroy $vfname -f");
			if (($text =~ /being migrated/) && $remote_filer) {
				sayEOF($fc,
				"vfiler migrate cancel $vfname\@$remote_filer");
			}
			if (($text =~ /DR configuration/) && $remote_filer) {
				sayEOF($fc,
				"vfiler dr delete $vfname\@$remote_filer");
			}
			return(1);
		} elsif ($after =~ /$vfname\s+stopped,\s+inconsistent/s) {
			my $text = sayEOF($fc, "vfiler destroy $vfname -f");
			if (($text =~ /being migrated/) && $remote_filer) {
				sayEOF($fc,
				"vfiler migrate cancel $vfname\@$remote_filer");
			}
			if (($text =~ /DR configuration/) && $remote_filer) {
				sayEOF($fc,
				"vfiler dr delete $vfname\@$remote_filer");
			}
			return(1);
		} else {
			die "$vfname is not running\n";
		}
	};
}

=head2 vfiler_stop($fc, $vfilertemplate)

causes the matching vfilers to stop receiving network traffic

=over 4

=item fc

Required: hostspec of the filer

=item vfilertemplate

Required: name of vfiler or *

=back

Examples:

	vfiler_stop($fc, 'vFiler100');
	vfiler_stop($fc, '*');

=cut

sub vfiler_stop($$)
{
	my ($fc, $vfilertemplate) = @_;

	my $all;

	if ($vfilertemplate eq "*") {
		$all = $vfilertemplate;
	}

	return eval {
		sleep(4);
		if ($all) {
			my $text = sayEOF($fc, "vfiler stop $all");
			if (($text !~ /already stopped/) ||
				($text !~ /stopped/)) {
				die "vfiler stop $all: failed\n";
			}

			return 1;
		} else {
			my (undef, undef, $after) =
				say($fc, "vfiler status $vfilertemplate", -swallow);
			die "$vfilertemplate is not running\n"
				unless ($after =~ /running/);

			say($fc, "vfiler stop $vfilertemplate");
			return(1);
		}
	};
}

=head2 vfiler_destroy($fc, $vfname)

releases all resources and removes the configuration information
associated with the named vfiler

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=back

Examples:

	vfiler_destroy($fc, 'vFiler100');

=cut

sub vfiler_destroy($$)
{
	my ($fc, $vfname) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		die "$vfname is not stopped\n" unless
			($after =~ /stopped/);

		say($fc, "vfiler destroy $vfname -f");
		return(1);
	};
}

=head2 vfiler_start($fc, $vfilertemplate)

start the stopped vfiler

=over 4

=item fc

Required: hostspec of the filer

=item vfilertemplate

Required: name of vfiler or *

=back

Examples:

	vfiler_start($fc, 'vFiler100');
	vfiler_start($fc, '*');

=cut

sub vfiler_start($$)
{
	my ($fc, $vfilertemplate) = @_;

	my $all;

	if ($vfilertemplate eq "*") {
		$all = $vfilertemplate;
	}

	return eval {
		sleep(4);
		if ($all) {
			my $text = sayEOF($fc, "vfiler start $all");
			if (($text !~ /already running/) ||
				($text !~ /running/)) {
				die "vfiler start $all: failed\n";
			}

			return 1;
		} else {
			my (undef, undef, $after) =
				say($fc, "vfiler status $vfilertemplate", -swallow);
			if ($after =~ /stopped/) {
				(undef, undef, $after) =
				say($fc, "vfiler start $vfilertemplate", -swallow);
				die "$vfilertemplate is not running\n" unless
					($after =~ /running/);

				return 1;
			} else {
				die "$vfilertemplate is running\n";
			}
		}
	};
}

=head2 vfiler_limit($fc)

get the current vfiler limit

=over 4

=item fc

Required: hostspec of the filer

=back

Examples:

	vfiler_limit($fc);

=cut

sub vfiler_limit($)
{
	my ($fc) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) = say($fc, "vfiler limit", -swallow);
		die "Current limit is not found\n"
			unless ($after =~ /Current\s+limit:\s+(\d+)/);

		return $1;
	};
}

=head2 vfiler_limit_in_use($fc)

get the current vfiler in use

=over 4

=item fc

Required: hostspec of the filer

=back

Examples:

	vfiler_limit_in_use($fc);

=cut

sub vfiler_limit_in_use($)
{
	my ($fc) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) = say($fc, "vfiler limit", -swallow);
		die "Current in use is not found\n"
			unless ($after =~ /Current\s+in\s+use:\s+(\d+)/);

		return $1;
	};
}


=head2 vfiler_limit_hard_limit($fc)

get the platform hard limit

=over 4

=item fc

Required: hostspec of the filer

=back

Examples:

	vfiler_limit_hard_limit($fc);

=cut

sub vfiler_limit_hard_limit($)
{
	my ($fc) = @_;

	return eval {
		sleep(4);
		my (undef, undef, $after) = say($fc, "vfiler limit", -swallow);
		die "Platform hard limit is not found\n"
			unless ($after =~ /Platform\s+hard\s+limit:\s+(\d+)/);

		return $1;
	};
}

=head2 vfiler_limit_set($fc, $maxvfilers)

set the vfiler limit

=over 4

=item fc

Required: hostspec of the filer

=item maxvfilers

Required: maximum number of vfilers

=back

Examples:

	vfiler_limit_set($fc, 65);
	vfiler_limit_set($fc, 25);

=cut

sub vfiler_limit_set($$)
{
	my ($fc, $maxvfilers) = @_;

	return eval {
		my (undef, undef, $after) = say($fc, "vfiler limit", -swallow);
		$after =~ /Platform\s+hard\s+limit:\s+(\d+)/;
		my $platform_limit = $1;
		$after =~ /Current\s+limit:\s+(\d+)/;
		my $current_limit = $1;

		say($fc, "vfiler limit $maxvfilers");
		sleep(4);

		if ($platform_limit != $current_limit) {
			(undef, undef, $after) = say($fc, "vfiler limit", -swallow);
			$after =~ /Current\s+limit:\s+\d+\s+\((\d+)/;
			if ($maxvfilers != $1) {
				$after =~ /Current\s+limit:\s+(\d+)/;
				if ($maxvfilers != $1) {
					die "vfiler limit was set to $1, not $maxvfilers\n";
				}
			}
		}

		return 1;
	};
}

=head2 vfiler_add($fc, $vfname, %args)

add the specified IP addresses and/or paths to an existing vfiler

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item args

Required: arguments

option  => "-f"
ipaddr  => "-i $VFILER_IPADDR_2"
path    => "/vol/$FILER_VOL/qt2"

=back

=cut

sub vfiler_add($$%)
{
	my ($fc, $vfname, %args) = @_;

	my $option = $args{option} || "";
	my $ipaddr = $args{ipaddr} || "";
	my $path = $args{path} || "";

	return eval {
		sleep(4);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /($vfname\s+stopped)|(Vfiler not found: $vfname)/) {
			die("$vfname is not running");
	   	}

		say($fc, "vfiler add $vfname $option $ipaddr $path");

		sleep(4);

		if ($path) {
			(undef, undef, $after) =
			say($fc, "vfiler status -r $vfname", -swallow);
			if ($after =~ /$path/gm) {
				return(1);
			} else {
				die "$path is not found\n";
			}
		}

		return(1);
	};
}

=head2 vfiler_context($fc, $vfname)

set the vfiler context

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=back

Examples:

	vfiler_context($fc, 'vFiler100');

=cut

sub vfiler_context($$)
{
	my ($fc, $vfname) = @_;


	return eval {
		sleep(4);

		if ($vfname eq "vfiler0") {
			my (undef, undef, $after) =
				say($fc, "vfiler context $vfname", -swallow);
			die "Not vfiler context $vfname\n"
				unless ($after =~ /vfiler\s+context\s+$vfname/gm);

			return 1;

		} else {
			my (undef, undef, $after) =
				say($fc, "vfiler status $vfname", -swallow);
			die "Vfiler not found: $vfname\n"
				if ($after =~ /Vfiler\s+not\s+found:\s+$vfname/);

			(undef, undef, $after) =
				say($fc, "vfiler status $vfname", -swallow);
	   		if ($after =~ /stopped|not found/) {
				die("$vfname is not running");
	   		}

			(undef, undef, $after) =
				say($fc, "vfiler context $vfname", -swallow);
			die "Not vfiler context $vfname\n"
				unless ($after =~ /vfiler\s+context\s+$vfname/gm);

			return 1;
		}
	};
}

=head2 vfiler_status($fc, [$vfilertemplate], [%args])

check a status of vfiler

=over 4

=item fc

Required: hostspec of the filer

=item vfilertemplate

Optional: name of vfiler or *

=item args

Optional: arguments

=back

Examples:

	vfiler_status($fc);
	vfiler_status($fc, 'vFiler100');
	vfiler_status($fc, '*');
	vfiler_status($fc, 'vFiler100', option=>"-r");
	vfiler_status($fc, 'vFiler100', option=>"-a");

=cut

sub vfiler_status($;$%)
{
	my ($fc, $vfilertemplate, %args) = @_;

	my $option = $args{option} || "";

	my $all;
        my ($after, $cmd);

	if (defined $vfilertemplate && $vfilertemplate eq "*") {
		$all = $vfilertemplate;
	}

	return eval {
		sleep(4);
		if ($option) {
                        $cmd = "vfiler status $option $vfilertemplate";
			(undef, undef, $after) =
				say($fc, $cmd, -swallow);
			if (($after =~ /Protocols\s+allowed:\s+\d+/gm) ||
				($after =~ /running/gm)) {
				return 1 unless wantarray;
			} else {
				die "$after\n";
			}
		} elsif ($vfilertemplate) {
			if ($all) {
                                $cmd = "vfiler status $all";
				(undef, undef, $after) =
					say($fc, $cmd, -swallow);
				die "vfiler status $all: failed\n"
					unless ($after =~ /running/s);
				return 1 unless wantarray;
			} else {
                                $cmd = "vfiler status $vfilertemplate";
				(undef, undef, $after) =
					say($fc, $cmd, -swallow);
				die "$vfilertemplate is not running\n" unless
					($after =~ /running/);
				return 1 unless wantarray;
			}
		} else {
                        $cmd = "vfiler status";
			(undef, undef, $after) =
				say($fc, $cmd, -swallow);
			die "vfiler status: failed\n"
				unless ($after =~ /running/gm);
                        return 1 unless wantarray;
		}

                if (wantarray) {
                    my (@res, $line, $vfiler_name, %status,
                        $key, $value);

                    # Remove \r characters from result string
                    $after =~ s/\r//g;
                    # Split result into a hash of values
                    @res = split('\n', $after);
                    # Remove preceding lines until the first status
                    # output line is found
                    shift @res while ($res[0] !~ /running|stopped/);


                    while (defined ($line = shift @res) && $line =~ /running|stopped/) {{
                        # First line should be vfiler name and status
                        if ($line !~ /^(\w+)\s+(.*)$/) {
                            die "Error parsing vfiler status output: $line";
                        } else {
                            $vfiler_name = $1;
                            $status{$vfiler_name}{status} = $2;
                        }

                        # No option specifiec, only get vfilers and status,
                        next if !$option;

                        # Parse the option output
                        while (defined ($line = shift @res) && $line ne '') {
                            if ($line !~ /^\s*(.*): (.*)$/) {
                                die "Error parsing vfiler status output: $line";
                            } else {
                                $key = $1;
                                $value = $2;
                                $key =~ s/ /_/g;
                                if ($key eq 'Path') {
                                    push (@{$status{$vfiler_name}{$key}}, $value);
                                } elsif ($key =~ /^(Allowed|Disallowed)$/) {
                                    $value =~ s/proto=//;
                                    push (@{$status{$vfiler_name}{$key}}, $value);
                                } else {
                                    $status{$vfiler_name}{$key} = $value;
                                }
                            }
                        }
                    }}
                    if (@res) {
                        # @res should be empty at this point
                        die "Error parsing vfiler status output: " .
                          "data still in result array";
                    } elsif (!keys(%status)) {
                        die "No status found for vfiler(s)";
                    }
                    return %status;
                }

	};
}

=head2 vfiler_remove($fc, $vfname, %args)

remove the specified IP addresses and/or paths from an existing vfiler

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item args

Required: arguments

	option => "-f"
	ipaddrs => "-i ipaddr1 [-i ipaddr] ..."
	paths => "path [path] ..."

=back

Examples:

	vfiler_remove($fc, 'vFiler100', ipaddrs => "-i ipaddr");
	vfiler_remove($fc, 'vFiler100', ipaddrs => "-i ipaddr", paths => "path");
	vfiler_remove($fc, 'vFiler100', "-f",
			ipaddrs => "-i ipaddr [-i ipaddr] ...",
			paths => " path [path] ...");

=cut

sub vfiler_remove($$%)
{
	my ($fc, $vfname, %args) = @_;

	my $option = $args{option} || "";
	my $ipaddrs = $args{ipaddrs} || "";
	my $paths = $args{paths} || "";
	my @apaths;
	my @aipaddrs;

	if ($paths) {
		@apaths = split (" ", $paths);
	}
	if ($ipaddrs) {
		@aipaddrs = split (" ", $ipaddrs);
	}

	return eval {
		sleep(4);
		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
	   	if ($after =~ /stopped|not found/) {
			die("$vfname is not running");
	   	}

		if (($option) && ($option ne "-f")) {
			die "$option is not valid\n";
		}
		sayEOF($fc, "vfiler remove $vfname $option $ipaddrs $paths");

		sleep(4);

		my $text = sayEOF($fc, "vfiler status -r $vfname");

		if ($paths) {
			foreach my $path (@apaths) {
				die "$path shouldn't be existed\n"
					if ($text =~ /$path/);
			}
		}
		if ($ipaddrs) {
			foreach my $ipaddr (@aipaddrs) {
				if ($ipaddr ne "-i") {
					die "$ipaddr shouldn't be existed\n"
						if ($text =~ /$ipaddr/);
				}
			}
		}

		return 1;
	};
}

=head2 vfiler_move($fc, $vfname_from, $vfname_to, %args)

removes the named IP addresses and/or paths from vfiler_from and
adds them to vfiler_to

=over 4

=item fc

Required: hostspec of the filer

=item vfname_from

Required: name of vfiler

=item vfname_to

Required: name of vfiler

=item args

Required: arguments

	option => "-f"
	ipaddrs => "-i ipaddr1 [-i ipaddr] ..."
	paths => "path [path] ..."

=back

Examples:

	vfiler_move($fc, 'vFiler100', 'vFiler200', ipaddrs => "-i ipaddr");
	vfiler_move($fc, 'vFiler100', 'vFiler200', ipaddrs => "-i ipaddr",
			paths => "path");
	vfiler_move($fc, 'vFiler100', 'vFiler200', "-f",
			ipaddrs => "-i ipaddr [-i ipaddr] ...",
			paths => "path [path] ...");

=cut

sub vfiler_move($$$%)
{
	my ($fc, $vfname_from, $vfname_to, %args) = @_;

	my $option = $args{option} || "";
	my $ipaddrs = $args{ipaddrs} || "";
	my $paths = $args{paths} || "";
	my @apaths;
	my @aipaddrs;

	if ($paths) {
		@apaths = split (" ", $paths);
	}
	if ($ipaddrs) {
		@aipaddrs = split (" ", $ipaddrs);
	}

	return eval {
		sleep(4);
		my (undef, undef, $after) = say($fc,
			"vfiler status $vfname_from", -swallow);
		if ($after =~ /($vfname_from\s+stopped)|(Vfiler not found: $vfname_from)/) {
			die("$vfname_from is not running");
	   	}

		sleep(4);
		(undef, undef, $after) = say($fc,
			"vfiler status $vfname_to", -swallow);
		if ($after =~ /($vfname_to\s+stopped)|(Vfiler not found: $vfname_to)/) {
			die("$vfname_to is not running");
	   	}

		if (($option) && ($option ne "-f")) {
			die "$option is not valid\n";
		}
		sayEOF($fc,
		"vfiler move $vfname_from $vfname_to $option $ipaddrs $paths");

		sleep(4);

		my $text = sayEOF($fc, "vfiler status -r $vfname_to");

		if ($paths && $vfname_to ne "vfiler0") {
			foreach my $path (@apaths) {
				die "$path should be existed\n"
					unless ($text =~ /$path/);
			}
		}
		if ($ipaddrs && $vfname_to ne "vfiler0") {
			foreach my $ipaddr (@aipaddrs) {
				if ($ipaddr ne "-i") {
					die "$ipaddr should be existed\n"
						unless ($text =~ /$ipaddr/);
				}
			}
		}

		sleep(4);

		$text = sayEOF($fc, "vfiler status -r $vfname_from");

		if ($paths) {
			foreach my $path (@apaths) {
				die "$path shouldn't be existed\n"
					if ($text =~ /$path/);
			}
		}
		if ($ipaddrs) {
			foreach my $ipaddr (@aipaddrs) {
				if ($ipaddr ne "-i") {
					die "$ipaddr shouldn't be existed\n"
						if ($text =~ /$ipaddr/);
				}
			}
		}

		return 1;
	};
}

=head2 vfiler_rename($fc, $old_vfname, $new_vfname)

rename an existing vfiler from oldname to newname

=over 4

=item fc

Required: hostspec of the filer

=item old_vfname

Required: name of old vfiler

=item new_vfname

Required: name of new vfiler

=back

Examples:

	vfiler_rename($fc, 'oldvfilername', 'newvfilername');

=cut

sub vfiler_rename($$$)
{
	my ($fc, $old_vfname, $new_vfname) = @_;

	my $fs_version = get_wafl_version($fc->{hostp});
	if ($fs_version < WV_ANCHORSTEAM_MIN) {
		croak(na_result(UNSUPPORTED_FEATURE_EXCEPTION));
	}

	return eval {
		sleep(4);
		my (undef, undef, $after) = say($fc, "vfiler status $old_vfname",
			-swallow);
		if ($after =~ /running/) {
			(undef, undef, $after) =
				say($fc,
				"vfiler rename $old_vfname $new_vfname", -swallow);
			if ($after =~ /vfiler\s+rename\s+complete/) {
				return(1);
			} else {
				die "vfiler rename is not complete\n";
			}
		} else {
			die "$old_vfname is not running\n";
		}
	};
}

=head2 vfiler_create_non_default_ipspace($fc, $vfname, $ipsname, %args)

create a vfiler with the given name using an non-default ipspace

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item ipsname

Required: name of ipspace

=item args

Required: arguments

=back

Examples:

	vfiler_create_non_default_ipspace($fc, 'vFiler100', 'ips1',
				ipaddr_paths=>"-i $VFILER_IPADDR_1 /vol/$cv1");

=cut

sub vfiler_create_non_default_ipspace
{
	my ($fc, $vfname, $ipsname, %args) = @_;

	my ($if_to_use) = $args{if_to_use} || "";
	my ($ipaddr_paths) = "$args{ipaddr_paths}" || "";

	return eval {
		if ($if_to_use) {
			my $text = sayEOF($fc, "ifconfig $if_to_use");
			if ($text =~ /no\s+such\s+interface/) {
				die("$if_to_use: no such interface");
			}
			say($fc, "ifconfig $if_to_use 0.0.0.0 down");
		}

		sleep(4);

		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /stopped/) {
			die("can't continue because $vfname is already ".
				"existed, but it is stopped");
		} elsif ($after =~ /not found/) {
			;
		} else {
			vfiler_stop_destroy($fc, $vfname);
			sleep(3);
		}

		say($fc,
		"vfiler create $vfname -n -s $ipsname $ipaddr_paths");
		sleep(4);
		(undef, undef, $after) =
		say($fc, "vfiler status $vfname", -swallow);

		if ($after =~ /running/) {
			return(1);
		} elsif ($after =~ /already configured/) {
			die $after;
		} else {
			die "$vfname is not created\n";
		}
	};
}

=head2 vfiler_create_default_ipspace($fc, $vfname, %args)

create a vfiler with the given name using default ipspace

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item args

Required: arguments

=back

Examples:

	vfiler_create_default_ipspace($fc, 'vFiler100',
				ipaddr_paths=>"-i $VFILER_IPADDR_1 /vol/$cv1");

=cut

sub vfiler_create_default_ipspace
{
	my ($fc, $vfname, %args) = @_;

	my ($if_to_use) = $args{if_to_use} || "";
	my ($ipaddr_paths) = "$args{ipaddr_paths}" || "";

	return eval {
		if ($if_to_use) {
			my $text = sayEOF($fc, "ifconfig $if_to_use");
			if ($text =~ /no\s+such\s+interface/) {
				die("$if_to_use: no such interface");
			}
			say($fc, "ifconfig $if_to_use 0.0.0.0 down");
		}

		sleep(3);

		my (undef, undef, $after) =
			say($fc, "vfiler status $vfname", -swallow);
		if ($after =~ /stopped/) {
			die("can't continue because $vfname is already ".
				"existed, but it is stopped");
		} elsif ($after =~ /not found/) {
			;
		} else {
			die("$vfname is already running");
		}

		say($fc,
		"vfiler create $vfname -n -i $ipaddr_paths");
		sleep(3);
		(undef, undef, $after) =
		say($fc, "vfiler status $vfname", -swallow);

		if ($after =~ /running/) {
			return(1);
		} elsif ($after =~ /already configured/) {
			die $after;
		} else {
			die "$vfname is not created\n";
		}
	};
}

=head2 vfiler_create_r($fc, $vfname, $origin_vfiler_path, [%args])

recreates the original vfiler that you destroyed

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item origin_vfiler_path

Required: original storage path of the vfiler

=item args

Optional: arguments

	old_vfname => "vFiler100"
	srcdst_path => "/vol/src1:/vol/dst1,/vol/src2:/vol/dst2,..."

=back

Examples:

	vfiler_create_r($fc, 'vFiler100', '/vol/vvol1000');
	vfiler_create_r($fc, 'vFiler200', '/vol/vvol1000',
		old_vfname => "vFiler100");
	vfiler_create_r($fc, 'vFiler100', '/vol/vvol1000',
		srcdst_path => "/vol/vvol1000;/vol/vvol1001,...");
	vfiler_create_r($fc, 'vFiler200', '/vol/vvol1000',
		old_vfname => "vFiler100",
		srcdst_path => "/vol/vvol1000;/vol/vvol1001,...");
=cut

sub vfiler_create_r($$$;%)
{
	my ($fc, $vfname, $origin_vfiler_path, %args) = @_;

	my $old_vfname = $args{old_vfname} || "";
	my $srcdst_path = $args{srcdst_path} || "";

	my $b_option = "";
	my $p_option = "";

	if ($old_vfname) {
		$b_option = "-b "."$old_vfname";
	}

	if ($srcdst_path) {
		$p_option = "-p "."$srcdst_path";
	}

	return eval {
		sleep(4);
		my (undef, undef, $after) = say($fc,
			"vfiler status $vfname", -swallow);
		die "$vfname is running\n"
			if ($after =~ /running/);

		if ($old_vfname) {
			sayEOF($fc,
			"vfiler create $vfname -r $origin_vfiler_path $b_option $p_option");
		} else {
			sayEOF($fc,
			"vfiler create $vfname -r $origin_vfiler_path $p_option");
		}

		sleep(4);
		(undef, undef, $after) = say($fc, "vfiler status $vfname", -swallow);
		die "$vfname is not running\n"
			unless ($after =~ /running/);

		return 1;
	};
}

=head2 vfiler_create($fc, $vfname, %args)

create a vfiler with the given name with the default ipspace

-n option is not used to create a vfiler

use Windows Domain authentication as a default type of user authentication

=over 4

=item fc

Required: hostspec of the filer

=item vfname

Required: name of vfiler

=item args

Required: arguments

if_to_use=>"$VFILER_IF_TO_USE"

netmask=>"$VFILER_NETMASK"

dnsdomain=>"$VFILER_DNS_DOMAIN"

dnsserver=>"$VFILER_DNS_SERVER"

nisdomain=>"$VFILER_NIS_DOMAIN"

nisserver=>"$VFILER_NIS_SERVER"

wins_ipaddr1=>"$VFILER_WINS_IPADDR_1"

wins_ipaddr2=>"$VFILER_WINS_IPADDR_2"

cifs_name=>"$FILER"."vf"

wins_domain=>"$VFILER_WINS_DOMAIN"

wins_user=>"$VFILER_WINS_USER"

wins_user_passwd=>"$VFILER_WINS_USER_PASSWD"

ipaddr_paths=>"-i $VFILER_IPADDR_1 /vol/$FILER_VOL/qt1"

count=>1

=back

Examples:

	vfiler_create($fc, 'vFiler100',
			if_to_use=>"$VFILER_IF_TO_USE",
			netmask=>"$VFILER_NETMASK",
			dnsdomain=>"$VFILER_DNS_DOMAIN",
			dnsserver=>"$VFILER_DNS_SERVER",
			nisdomain=>"$VFILER_NIS_DOMAIN",
			nisserver=>"$VFILER_NIS_SERVER",
			wins_ipaddr1=>"$VFILER_WINS_IPADDR_1",
			wins_ipaddr2=>"$VFILER_WINS_IPADDR_2",
			cifs_name=>"$FILER"."cf",
			wins_domain=>"$VFILER_WINS_DOMAIN",
			wins_user=>"$VFILER_WINS_USER",
			wins_user_passwd=>"$VFILER_WINS_USER_PASSWD",
			ipaddr_paths=>"-i $VFILER_IPADDR_1 /vol/$FILER_VVOL",
			count=>1
			);

=cut

sub vfiler_create
{
	my ($fc, $vfname, %args) = @_;

	my $if_to_use = $args{if_to_use} || "";
	my $netmask = $args{netmask} || "";
	my $vfiler_password=$args{vfiler_password}||"pa55w0rd";
	my $dnsdomain = $args{dnsdomain} || "";
	my $dnsserver = $args{dnsserver} || "";
	my $dnsserver1 = $args{dnsserver1} || "";
	my $dnsserver2 = $args{dnsserver2} || "";
	my $nisdomain = $args{nisdomain} || "";
	my $nisserver = $args{nisserver} || "";
	my $wins_ipaddr1 = $args{wins_ipaddr1} || "";
	my $wins_ipaddr2 = $args{wins_ipaddr2} || "";
	my $cifs_name = $args{cifs_name} || "";
	my $wins_domain = $args{wins_domain} || "";
	my $wins_user = $args{wins_user} || "";
	my $wins_user_passwd = $args{wins_user_passwd} || "";
	my $ipaddr_paths = "$args{ipaddr_paths}" || "";
	my $ipaddr_count = $args{count} || 1;
	my $wins_ipaddr_count = $args{wins_ipaddr_count} || 0;
	my $workgroup_name = $args{workgroup_name} || "";
	my $user_auth_type = $args{user_auth_type} || "";
	my $no_cifs = $args{no_cifs} || "";
	my $timeout = $args{timeout} || 240;

	my %prompts = (
		1	=> "Configure vfiler IP address.*:",
		2	=> "Interface to assign this address to.*:",
		3	=> "Netmask to use:.*:"
	);

	my %inputs = (
		1	=> "y",
		2	=> "$if_to_use",
		3	=> "$netmask"
	);

	if (length($cifs_name) > 15) {
		$cifs_name = substr($cifs_name, 0, 15);
	}

	return eval {
	sleep(4);

	$fc->timeout($timeout);
	my (undef, undef, $after) = say($fc, "ipspace list", -swallow);
	$after =~ /Number of ipspaces configured: (\d+)/;
	if ($1 > 1) {
		my @out1 = split(" ", $after);
		foreach my $ips (@out1) {
			if ($ips =~ /\((\w+)/s) {
				if ($1 ne 'no') {
					my $retstr = sayEOF($fc, "ifconfig $1");
					my @out2 = split(" ", $retstr);
					my $found = 0;
					foreach my $ip_addr (@out2) {
						if ($ip_addr =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/s) {
							$found = 1;
						}
					}
					if (!$found) {
						sayEOF($fc,
						"ipspace assign default-ipspace $if_to_use");
					}
				}
			}
		}
	}


	vfiler_stop_destroy($fc, $vfname);

	sleep(4);

	my $cifs_not_licensed = 0;
	if (!is_licensed($fc, 'cifs')) {
		$cifs_not_licensed = 1;
	}

	my $old_cifs_setup = 0;
	my $os_version = get_ontap_version($fc->{hostp}) or die $@;
	my $fs_version = get_wafl_version($fc->{hostp}) or die $@;
	if ($os_version =~ /7.0/) {
		my $text = sayEOF($fc, "version");
		$old_cifs_setup = ($text =~ /R7.0xN|7.0X|7.0RC|7.0P/) ? 1 : 0;
	}

	sleep(2);
	say($fc, "vfiler create $vfname $ipaddr_paths",
		'-noprompt', '-wait', "Setting up vfiler $vfname");

	for (my $cnt = 1; $cnt <= $ipaddr_count; $cnt++) {
		foreach my $key (qw (1 2 3)) {
			say($fc, "", '-noprompt', '-nocr', '-wait', "$prompts{$key}");
			say($fc, "$inputs{$key}", '-blind');
		}
	}

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"^Please enter the name or IP address of the administration host:.*");
	say($fc, "\r", '-nocr', '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"Do you want to run DNS resolver.*:");
	say($fc, "y", '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"Please enter DNS domain name.*:");
	say($fc, "$dnsdomain", '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"Please enter the IP address for first nameserver.*:");
	say($fc, "$dnsserver", '-blind');

	if ($dnsserver1) {
		say($fc, "", '-noprompt', '-nocr', '-wait',
		"Do you want another nameserver.*:");
		say($fc, "y", '-blind');
		say($fc, "", '-noprompt', '-nocr', '-wait',
		"Please enter the IP address for alternate nameserver.*:");
		say($fc, "$dnsserver1", '-blind');
		if ($dnsserver2) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want another nameserver.*:");
			say($fc, "y", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Please enter the IP address for alternate nameserver.*:");
			say($fc, "$dnsserver2", '-blind');
		} else {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want another nameserver.*:");
			say($fc, "n", '-blind');
		}
	} else {
		say($fc, "", '-noprompt', '-nocr', '-wait',
		"Do you want another nameserver.*:");
		say($fc, "n", '-blind');
	}

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"Do you want to run NIS client.*:");
	say($fc, "y", '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"Please enter NIS domain name.*:");
	say($fc, "$nisdomain", '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait',
		"Please enter list of preferred NIS servers.*:");
	say($fc, "$nisserver", '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait', "^New password:");
	say($fc, "$vfiler_password", '-blind');

	say($fc, "", '-noprompt', '-nocr', '-wait', "^Retype new password:");
	say($fc, "$vfiler_password", '-blind');

	if ($cifs_not_licensed) {
		return(1);
	}

	say($fc, "", '-noprompt', '-nocr', '-wait', "Do you want to setup CIFS.*:");
	if ($args{no_cifs}) {
		say($fc, "n");
	} else {
	   if (($os_version =~ /6.5|6.4/) ||
			(($os_version =~ /7.0/) && ($old_cifs_setup))) {
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you wish to make the system visible via WINS.*:");
		say($fc, "y", '-blind');

		if ($wins_ipaddr_count > 0) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Please enter next WINS IP address.*:");
			say($fc, "$wins_ipaddr1", '-blind');
			if ($wins_ipaddr_count == 2) {
				say($fc, "", '-noprompt', '-nocr', '-wait',
					"Please enter next WINS IP address.*:");
				say($fc, "$wins_ipaddr2", '-blind');
			}
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Please enter next WINS IP address.*:");
		say($fc, "\r", '-nocr', '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to modify this name.*:");
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Enter the CIFS name for the filer:");
		(undef, undef, $after) = say($fc, "$cifs_name", '-blind');
		if ($after =~ /Enter\s+the\s+CIFS\s+name/) {
			if (length($cifs_name) > 15) {
				$cifs_name = substr($cifs_name, 0, 14);
			}
			$cifs_name = "$cifs_name"."1";
			say($fc, "$cifs_name", '-blind');
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"What type of authentication will this filer use.*:");

		if (($user_auth_type && $workgroup_name) ||
			(ref($fc->hostp) =~ /Simulator/)) {
			say($fc, "2\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Do you want to modify this name.*:");
			say($fc, "y", '-blind');
			if (ref($fc->hostp) =~ /Simulator/) {
				my $filer_name = $fc->hostp()->hostname();
				$workgroup_name = "$filer_name"."w";
			}
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the Workgroup for the filer.*:");
			(undef, undef, $after) =
				say($fc, "$workgroup_name", '-blind');
			if ($after =~ /Enter\s+the\s+Workgroup/) {
				$workgroup_name = "$workgroup_name"."1";
				say($fc, "$workgroup_name", '-blind');
			}
			say($fc, "", '-noprompt', '-nocr',
				'-wait', "Create the.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr',
				'-wait', "^Set the password for.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr',
				'-wait', "^Retype the password:");
			say($fc, "\r", '-nocr', '-blind');
		} else {
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the Windows Domain for the filer.*:");
			say($fc, "$wins_domain", '-blind');

			say($fc, "", '-noprompt', '-nocr',
				'-wait', "Create the.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"^Set the password for.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"^Retype the password:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Please enter the Windows 2000 user.*:");
			say($fc, "$wins_user", '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"^Password for $wins_user:");
			say($fc, "$wins_user_passwd", '-blind');

		}
		say($fc, "\r", '-nocr', '-blind');
	   } else {
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Do you want to make the system visible via WINS.*:");
		say($fc, "n", '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Selection.*:");
		say($fc, "\r", '-nocr', '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Should CIFS create default.*:");
		say($fc, "y", '-blind');
		say($fc, "", '-noprompt', '-nocr',
			'-wait', "^Enter the password for the root user.*:");
		say($fc, "\r", '-nocr', '-blind');
		say($fc, "", '-noprompt', '-nocr',
			'-wait', "^Retype the password:");
		say($fc, "\r", '-nocr', '-blind');

	   	if ($fs_version > WV_ANCHORSTEAM_MAX) {
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Would you like to enable NIS group caching.*:");
			say($fc, "y", '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the hour.*:");
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "", '-noprompt', '-nocr', '-wait',
			"Would you like to specify additional hours.*:");
			say($fc, "\r", '-nocr', '-blind');
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Would you like to change this name.*:");
		say($fc, "y", '-blind');

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Enter the CIFS server name for the filer.*:");
		my (undef, undef, $after) = say($fc, "$cifs_name", '-blind');
		if ($after =~ /Enter the CIFS server/) {
			if (length($cifs_name) > 15) {
				$cifs_name = substr($cifs_name, 0, 14);
			}
			$cifs_name = "$cifs_name"."1";
			say($fc, "$cifs_name", '-blind');
		}

		say($fc, "", '-noprompt', '-nocr', '-wait',
			"Selection.*:");

		if (($user_auth_type && $workgroup_name) ||
			(ref($fc->hostp) =~ /Simulator/)) {
			say($fc, "3\r", '-nocr', '-blind');

			if (ref($fc->hostp) =~ /Simulator/) {
				my $filer_name = $fc->hostp()->hostname();
				$workgroup_name = "$filer_name"."w";
			}
			say($fc, "", '-noprompt', '-nocr', '-wait',
				"What is the name of the Workgroup.*:");
			(undef, undef, $after) =
				say($fc, "$workgroup_name", '-blind');
			if ($after =~ /What\s+is\s+the\s+name/) {
				$workgroup_name = "$workgroup_name"."1";
				say($fc, "$workgroup_name", '-blind');
			}
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "Do you want to create.*:");
			say($fc, "\r", '-nocr', '-blind');
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "^Enter the new password for.*:");
			say($fc, "\r", '-nocr', '-blind');
#			say($fc, "", '-noprompt', '-nocr',
#				'-wait', "^Retype the password:");
			say($fc, "\r", '-nocr', '-blind');
		} else {
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
			"What is the name of the Active Directory domain.*:");
			say($fc, "\r", '-nocr', '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"Enter the name of the Windows user.*:");
			say($fc, "$wins_user", '-blind');

			say($fc, "", '-noprompt', '-nocr', '-wait',
				"^Password for $wins_user:");
			say($fc, "$wins_user_passwd", '-blind');

			say($fc, "\r", '-nocr', '-blind');
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "\r", '-nocr', '-blind');
			say($fc, "\r", '-nocr', '-blind');
		}
		say($fc, "\r", '-nocr', '-blind');
	   }
	}

	return (1);
	};
}


=head2 vfiler_change_ip_binding(%args)

bind a newly added ip address resource to the vfiler or remove the binding of
an ip address from the vfiler

=over 4

=item args

Required: arguments

conn => Filer connection object

vfiler => vfiler name

action => add or delete

ipaddr => ip address for which to add or delete binding

interface => interface to bind ip address to

netmask => netmask for ip address


=back

Examples:

vfiler_change_ip_binding(conn => $fc,
                         vfiler => $vfiler_name,
                         action => "add",
                         ipaddr => $VFILER_IPADDR_1,
                         interface => $VFILER_IF_TO_USE,
                         netmask => $VFILER_NETMASK,
			 );

vfiler_change_ip_binding(conn => $fc,
                         vfiler => $vfiler_name,
                         action => "delete",
                         ipaddr => $VFILER_IPADDR_1,
                         interface => $VFILER_IF_TO_USE,
                         netmask => $VFILER_NETMASK,
			 );

=cut



sub vfiler_change_ip_binding(%) {

    my %args =  Params::Validate::validate(@_, {conn 	  => { type => OBJECT },
						vfiler 	  => { type => SCALAR },
						action 	  => { type => SCALAR },
						ipaddr 	  => { type => SCALAR },
						interface => { type => SCALAR },
						netmask	  => {type => SCALAR, optional => 1,}
					       });

    my $conn = $args{'conn'};
    my $vfiler = $args{'vfiler'};
    my $tries;
    my $max_tries = 50;
    my ($after,$prompt);

    # build hash of prompts and possible answers
    my %prompts_vfiler_setup =
	("Are you sure you want to continue\?" 	  => "y",
	 "Change binding for vfiler IP address.*" => sub { _get_vfiler_reply($prompt,$args{'ipaddr'},$args{'action'}) },
	 "Configure vfiler IP address.*" 	  => sub { _get_vfiler_reply($prompt,$args{'ipaddr'},$args{'action'}) },
	 "Interface to assign this address to" 	  => $args{'interface'},
#	 "Netmask to use" 			  => "",
	 "Netmask to use" 			  => $args{'netmask'},
	 "Please enter the name or IP address of the administration host" => "",
	 "Do you want to run DNS resolver\?" 	  => "",
	 "Please enter DNS domain name" 	  => "",
	 "Please enter the IP address for first nameserver" => "",
	 "Do you want another nameserver\?" 	  => "",
	 "Do you want to run NIS client" 	  => "",
	 "Please enter NIS domain name" 	  => "",
	 "Please enter list of preferred NIS servers" => "",
	);

    my @prompts = keys(%prompts_vfiler_setup);

    (undef, undef, $after) =
	say($conn, "vfiler status $vfiler", -swallow);
    if ($after =~ /stopped|not found/) {
	croak ("$vfiler is not running");
    }

    sleep(4);

    say($conn, "vfiler run $vfiler setup",-blind);
    LOOP: for ($tries=0;$tries<$max_tries;$tries++) {
	    (undef,undef,$after,$prompt) =
		say ($conn, "",
		     '-noprompt',
		     '-prompt', [@prompts, $conn->prompts()],
		     '-timeout',60,
		     '-eval_error');
	    logcomment("Current PROMPT : $prompt AFTER: $after");

	    croak $@ if ($@);
	    foreach my $key (@prompts) {
		my $value = $prompts_vfiler_setup{$key};
		if ($prompt =~ /$key/) {
		    my $send;
		    if (ref($value) eq "CODE") {
			# The routine will return what we should send
			# (or undef if we should send nothing)
			$send = &$value();
		    } else {
			$send=$value;
		    }
		    if (defined($send)) {
			say ($conn, "$send\r",
			     '-nocr',
			     '-blind',
			     '-eval_error');
			die $@ if $@;
		    }
		    next LOOP;
		}
	    }

	    last LOOP;
	}

    if ($tries==$max_tries) {
	croak "Too many prompts ($max_tries), probably a loop\n";
    }
    return 1;
}

sub _get_vfiler_reply() {

    my ($prompt,$ipaddr,$action) = @_;

    logcomment("$prompt - $ipaddr - $action \n");

    if ($action =~ /add/) {
	if ($prompt =~ /Configure vfiler/) {
	    return ($prompt =~ /$ipaddr/ ? "y" : "n");
	} elsif ($prompt =~ /Change binding/) {
	    return ($prompt =~ /$ipaddr/ ? "y" : "n");
#	    return "";
	}
    } elsif ($action =~ /delete/) {
	if ($prompt =~ /Configure vfiler/) {
	    return ($prompt =~ /$ipaddr/ ?  "n" : "n");
	} elsif ($prompt =~ /Change binding/) {
	    return ($prompt =~ /$ipaddr/ ?  "y" : "n");
	}
    }
}

=head2 vfiler_options_get($fc, $vfname, $option)

Get the value for option given on the specified vfiler

=over 4

=item fc

Required: The filer connection name where this function is executed

=item vfname

Required: The name of the vfiler to run the command on

=item option

Required: The name of the option

=back

Example:

vfiler_options_get($fc, 'vfiler0', 'snapmirror.enable');

=cut

#NOTE: Do not want to put this in filer.pm because then everybody that
# used filer_options_get or filer_options_set is impacted with a vfiler
# check as it is currently designed in filer.pm. A better way would have
# been to add this information into the connection object (i.e. quick
# check without talking to the filer to see if the command needs to
# run on the vfiler).
# NOTE: Did not use named parameters because I wanted the interface
# to closely match filer_options_get().
sub vfiler_options_get ($$$;) {
    my ($fc, $vfname, $option) = @_;

    my $cmd = "vfiler run $vfname options $option";
    my $text = sayEOF($fc, $cmd);

    my $value = "";
    my @info = split(/\r*\n/,$text);
    foreach my $line (@info) {
        if ($line =~ /$option /) {
            (undef, $value) = split(/\s+/, $line);
        }
    }
    return $value;
}

=head2 vfiler_options_set($fc, $vfname, $option, $value)

Set the option to the given value on the specified vfiler

=over 4

=item fc

Required: The filer connection name where this function is executed

=item vfname

Required: The name of the vfiler to run the command on

=item option

Required: The name of the option

=item value

Required: The value to set the option to

=back

Example:

vfiler_options_set($fc, 'vfiler0', 'snapmirror.enable', 'on');

=cut

sub vfiler_options_set ($$$$;) {
    my ($fc, $vfname, $option, $value) = @_;

    my $cmd = "vfiler run $vfname options $option $value";
    sayEOF($fc, $cmd);

    # Verify the option got set
    my $new_value = vfiler_options_get($fc, $vfname, $option);
    croak "Could not set option $option exp: $value got: $new_value"
        if ($new_value ne $value);
}

=head2 vfiler_ssh_setup($filer, $vfname, $passwd)

Setup root ssh for the vfiler running on the physical filer.
Uses a telnet connection to the filer to set vfiler context.

=over 4

=item filer

Required: The physical filer hostspec

=item vfname

Required: The name of the vfiler to run the command on

=item passwd

Required: The password to set for root on the vfiler

=back

Example:

vfiler_ssh_setup($FILERB, $vfname, $root_passwd);

=cut

sub vfiler_ssh_setup ($$$;) {
    my ($filer, $vfname, $passwd) = @_;
    my ($cmdtimeout, $fc_telnet, $text, $timeout);

    # Set the timeouts
    $cmdtimeout = (defined($main::CMDTIMEOUT) ? $main::CMDTIMEOUT : 120);
    $timeout = (defined($main::TIMEOUT) ? $main::TIMEOUT : 120);

    # Create telnet connection to vfiler and set context
    host($filer);
    $fc_telnet = connect('#telnet', -timeout, $timeout);
    $fc_telnet->timeout($cmdtimeout);
    vfiler_context($fc_telnet, $vfname);
    if ($@) {
        disconnect('#telnet');
        croak($@);
    }

    # Enable ssh access on vfilers on FILERB.
    sayEOF($fc_telnet, 'secureadmin setup -q ssh 768 512 768');

    # SSH doesn't allow login to accounts without passwords. Make sure
    # root has password $passwd.
    say($fc_telnet, "passwd", '-noprompt', '-wait', "^New password:");
    say($fc_telnet, "$passwd", '-noprompt', '-wait', "^Retype new password:");
    say($fc_telnet, "$passwd", '-blind');
    vfiler_context($fc_telnet, 'vfiler0');
    $text = $@;
    disconnect('#telnet');
    croak($text) if ($text);

    # Give the sshd a chance to start running
    sleep(5);
}

=head2 vfiler_ssh_cmd($ip, $passwd, $cmd)

Execute a command over ssh to the vfiler.

It would be nice to use a #ssh NATE connection but vfilers don't allow
shells so there has to be a command at the end of the ssh command and
you don't get that with #ssh-persist and if you use #ssh-persay then
you have to deal with prompts.

Returns the output from the command.

NOTE:
Running this routine on a Linux RHEL4.2 repeatedly will lock up
the kernel.

=over 4

=item ip

Required: IP address of vfiler

=item passwd

Required: Password to connect to vfiler

=item cmd

Required: Command to be run on the vfiler

=back

Example:

$text = vfiler_ssh_cmd($VFILER_IPADDR_2, $root_passwd, 'help');

=cut

sub vfiler_ssh_cmd ($$$;) {
    my ($ip, $passwd, $cmd) = @_;
    my ($cc, $cmdtimeout, $text, $timeout);
    my $khosts = "/tmp/known_hosts" . time();

    # Set the timeouts
    $cmdtimeout = (defined($main::CMDTIMEOUT) ? $main::CMDTIMEOUT : 120);
    $timeout = (defined($main::TIMEOUT) ? $main::TIMEOUT : 120);

    # Setup the rsh connection as root, otherwise it will be the uid
    # of whoever started the test. The problem with that is they may
    # have ssh setup with settings that will not work for this test.
    # Most root accounts have not been configured for automatic ssh root login.
    $cc = connect('localhost#rsh-persist;username=root', -timeout, $timeout);
    $cc->timeout($cmdtimeout);
    unlink $khosts if (-e $khosts);
    sayEOF($cc, "id");
    say($cc, "ssh -o BatchMode=no -o PasswordAuthentication=yes " .
        "-o CheckHostIP=no -o UserKnownHostsFile=$khosts root\@$ip $cmd",
        '-noprompt', '-wait', 'password');
    $text = sayEOF($cc, "yes");
    $text = sayEOF($cc, "$passwd");
    unlink $khosts if (-e $khosts);
    return $text;
}

1;

