# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/FS/Util/file_folding.pm#1 $

# Copyright (c) 2004 Network Appliance, Inc.
# All rights reserved.

## @summary        Library that defines general purpose file folding routines
## @description
##        File folding start, stop, do_fold, wafliron

##
## @author dl-waflqa-art@netapp.com 

package FS::Util::file_folding;

use Exporter qw(import);
BEGIN {
our @EXPORT = qw(&do_fold
		 &file_fold_start
		 &wafl_scan_start
		 &wafl_scan_status				    
		 &wafl_scan_abort
	);
}

use strict;
no strict "subs";
use Tharn;
use Carp;
use FileHandle;
use mount qw(mount);
use Admin::License qw(license);
use Utils qw(make_conn_and_complain);
use SnapCmds qw(snapshot_list snapshot_create snapshot_delete snaprestore_file);
use FS::Util::Utilities qw(uSystem);
use FS::Util::Setup qw(mount_root_vol);
use mount qw(mount umount);
use Wafliron qw(get_container);
use Aggregate qw(aggr_offline);
use Filer::Version qw(get_wafl_version
                      WV_ANCHORSTEAM_MIN
                      WV_IRONCITY_MIN
		      WV_BOILERMAKER_MIN
		      );
use Module::Common qw(validate_mixed);
use Params::Validate
	qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF OBJECT);
use SPARSE::Util::Logger;
use Data::Dumper;

use constant ON => "on";
use constant OFF => "off";
use constant RSH_TIMEOUT => 1000;
use constant LOCAL_WAFLIRON_TIMEOUT => 2400;
use constant SLEEPTIME => 45;

# -----------------
# Sub routines
# -----------------

## @name ::do_fold
## @summary initiates a file_fold on a single file then waits
##
## @arg 
##
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##	   vol - Name of the volume containing the file
##
##	   filename - The fully qualified name of the file to be folded
##
##
##

sub do_fold($$$)
{
        my ($fc, $vol, $filename) = @_;
	my (undef, undef, $output) = say($fc, "priv set -q advanced; wafl scan fold /vol/$vol/$filename");
	croak("Error in say(wafl scan status: $output") if ($output =~ /error/i);
	do {
	    snooze(5);		# check every 5 seconds
	    (undef, undef, $output) = say($fc, "priv set -q advanced; wafl scan status -V $vol");
	    croak("Error in say(wafl scan status: $output") if ($output =~ /error/i);
	} while ($output =~ /file folding/);
}

## @name ::file_fold_start
## @summary initiates a file_fold on a single file - Does not wait for it to complete
##
## @arg 
##
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##	   vol - Name of the volume containing the file
##
##	   filename - The fully qualified name of the file to be folded
##
##	   timeout - optional timeout for file folding scan start
##
## Returns:	id of the scan
##

sub file_fold_start($$$@)
{
    my ($fc, $vol, $filename, @all) = @_;

    my $timeout = (defined(@all) ? $all[0] : 200);
    my $endtime = time() + $timeout;

    # issue the fold command
    say($fc,"priv set -q advanced; wafl scan fold /vol/$vol/$filename");
    
    # Gather output from status command to return in hash table
    my $id;
    while (time() < $endtime) {
        my (undef, undef, $scan_status) = say($fc,"priv set -q advanced; wafl scan status -V $vol");
    
        # strip white spaces, newlines, and carriage returns
        if ($scan_status !~ /No WAFL scans in progress/) {
	        my @lines = split("\n", $scan_status);
	        foreach (@lines) {
	            # if we have a file folding status line
	            # break it up and send back the info in the hash table
	            if ($_ =~ /file folding/) {
		            # break out the scan id
		            if ($_ !~ /^\s*(\w+)\s+file folding\s+(\w+)/m) {
		                croak("Unable to parse scanner output: $_");
		            }
		            $id = $1;
		            last;
	            }
	        }
        } else {
	        croak("fold did not start");
        }
        last if defined $id;
        sleep(2);
    }
    croak("fold did not start") unless defined $id;
    return $id;
}

## @name ::wafl_scan_start
##
## @summary Starts a scan of the specified type against the specified object
##
## @arg 
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##	   type -> Type of scan to start
##
##	   object -> Depending upon "type" this is the rest of the wafl scan command
##
## Returns: Result string from wafl scan start 
##
sub wafl_scan_start {
    my $fc = shift(@_);
    
    my (%args) = validate(
         @_,
         {
	    type => { type => SCALAR},
	    object => { type => SCALAR, optional => 1,
                            default => '' },
         }
    );
    my $type = $args{type};
    my $object = $args{object};
    
    my $cmd = "priv set -q advanced; wafl scan ";
    
    if (defined($type)) {
	    $cmd .= $type;
    } else {
	die "type is a required parameter for wafl_scan_start";
    }
    
    if (defined($object)) {
	    $cmd .= " ".$object;
    }
    
    # Issue the command
    my $result = '';
    (undef, undef, $result) = say($fc, $cmd);
    
    return $result;
}

## @name ::wafl_scan_status
##
## @summary Returns the status of a scan
##
## @arg 
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##	   object -> vol / aggr - Name of the volume or aggregate whose scan status is to be returned [optional]
##
## Returns: hash of hashs by Scan Id -> type[Volume|Aggregate], name, status and progress
##

sub wafl_scan_status
{
    my $fc = shift;
    
    my (%args) = validate(
         @_,
         {
           object => { type => SCALAR, optional => 1,
                            default => '' },
         }
    );
    
    my $object = $args{object};
    my (%info, $scan_status, $opts, $cmd);
    
    $cmd = "priv set -q advanced; wafl scan status ";
    
    if (defined($object)) {
	# See if the object passed is a volume or an agggr
	(undef, undef, $scan_status) = say($fc, "vol status $object");
	# issue appropriate command
	if ($scan_status =~ /is an aggregate/i) {
	    $cmd .= "-A $object";
	} else {
	    $cmd .= "-V $object";
	}
    }

    (undef, undef, $scan_status) = say($fc, $cmd);
    
    
    # strip white spaces, newlines, and carriage returns
    my $name;
    my $type;
    my $id;
    my @junk;
    
    if ($scan_status !~ /No WAFL scans in progress/) {
	my @lines = split("\n", $scan_status);
	foreach (@lines) {
	    my %info = ();
	    # For now return volume name in hash
	    if ($_ =~ /Volume (.*):/m) {
		$name = $1;
		$type = "Volume";
		next;
	    }
	    if ($_ =~ /Aggregate (.*):/m) {
		$name = $1;
		$type = "Aggregate";
		next;
	    }	    
	    # Skip any header lines
	    if ($_ =~ /Scan id/) {
		next;
	    }
	    # we have a status line
	    # remove leading blanks
	    $_ =~ s/^\s+//;
	    # Extract the numeric id
	    $_ =~ s/^(\d+)//;
	    $id = $1;
	    # remove leading blanks
	    $_ =~ s/^\s+//;
	    # break up the two strings
	    @junk = split("    ", $_);
	    $info{$id}{name} = $name;
	    $info{$id}{type} = $type;
	    $info{$id}{state} = shift(@junk);
	    $info{$id}{state} =~ s/^\s+//;
	    $info{$id}{progress} = shift(@junk);
	    $info{$id}{progress} =~ s/^\s+//;
	    logDbg("info is: ".Dumper(%info));
	}
	
    } else {
	$info{0}{name} = "Unknown";
	$info{0}{state} = "No WAFL scans in progress";
    }
     return(%info);
}

## @name ::wafl_scan_abort
##
## @summary Terminate a scan by id
##
## @arg 
##
##         filer_conn - connspec of filer conn. defaults to $main::FILER_CONN. E.g. "#rsh", "#telnet", etc.
##		-- To allow some existing scripts to work 'as is', this fcn. will accept 'CONNECTREC' objects
##		   to be passed via the 'filer_conn' option, as if they were passed via the 'filer' option.
##
##	   scan_id - Id of the scan to be terminated
##
## Returns:	result string from the scan abort
##

sub wafl_scan_abort($$)
{
    my ($fc, $scan_id) = @_;
    
    my (undef, undef, $results) = say($fc,"priv set -q advanced; wafl scan abort $scan_id");
    if ($results =~ /usage:/) {
	$results = "scan $scan_id not active";
    }
    
    return $results

}

# ------------------------------------------------------------------------
1;
# ------------------------------------------------------------------------
