# Hey emacs, this is basically -*- perl -*-
# $Id: filer_messages.pm,v 1.10 2004/08/18 18:12:05 sidlo Exp sidlo $
#
# Copyright (c) 2004 Network Appliance, Inc.
# All rights reserved.
#

package filer_messages;

## @summary
## This object performs a non-blocking tail on any
## file within a filer's root volume.
##
## @description
## Use this package to look for messages
## within the filer's /etc/messages (or any other file
## on the filer) to find indications of
## system state changes or events, within the main
## event loop of your test. It is possible to do this
## because this object performs non-blocking reads
## of the log file. When we are ready to look for
## data from the log file, we invoke the ->get_lines()
## method to read log file messages since the
## last call to ->get_lines().
##
## An array of regexp patterns can also be passed
## into this object. All regexp are processed on
## each log message when ->get_lines() in called.
## If any patters match the latest batch of
## log messages, an internal {trigger} is set, which
## is examined with the ->triggered() method.
##
## Lines that trigger Pattern matches are stored
## to an array of arrays that can be accessed by
## calling the method ->get_matched_lines().
##
## This package inherits the nb_tail packages which
## performs a non-blocking tailed read of the
## log file.
##
## @example
##   ----------------------------------------------------------
##   #!/bin/perl
##
##   use strict;
##   use FS::Util::filer_messages;
##   use SPARSE::Util::Logger qw(logLevel logError logStep logInfo logWarn logTrc);
##
##   logLevel(500);
##
##   # Turn trigger on when we wee CIFS or WAFL messages.
##   my @pats = ( 'CIFS',
##                'WAFL',
##                'WAFL layout ratio for volume' );
##
##   my $tail = filer_messages->new('sidlo1', \@pats);
##
##   my $done = 0;
##   do {
##
##     # Let latest lines from the log file.
##     my $lines = $tail->get_lines();
##
##     # Check for a pattern match
##     if ($tail->triggered()) {
##
##       logTrc "*** TRIGGERED! ***\n";
##
##       # Dump lines that match our patterns
##       my @matches = $tail->get_matched_lines();
##       my ($i, $j);
##       for ($i=0; $i<@matches; $i++) {
##         if (defined $matches[$i]) {
##           for ($j=0; $j<@{$matches[$i]}; $j++) {
##   	      logTrc "[(".$pats[$i].") $i:$j] ".${$matches[$i]}[$j]."\n";
##           }
##         }
##       }
##
##       # Alternative dump of matching lines:
##       # my $txt = $tail->dumpMatches(@matches);
##       # logTrc("Dump of the match array:\n$txt");
##
##       # Done if we match to pattern-3 'WAFL layout ratio for volume'
##       $done = 1 "Done." if (defined $matches[2]);
##
##     } else {
##
##       # Output the log-messages since last call to $tail->get_lines()
##       logTrc "$lines";
##
##       # Pretend to do more processing.
##       sleep 5;
##
##     }
##
##   } until ($done);
##   ----------------------------------------------------------
##
## @note Please review nb_tail.pm for inherited methods and options.
##
## @keywords filer tail syslog messages etc
##
## @status public
##
## @author dl-johny@netapp.com (FSS WAFL Quality)

use strict;
use Date::Manip;

use SPARSE::Util::Logger qw(logLevel logError logStep logInfo logWarn logTrc);
use FS::Util::nb_tail;
use FS::Util::Setup;
use Tharn;
use SayWrap;

# Inherit from the nb_tail class
@filer_messages::ISA = qw ( nb_tail );

sub new ($;$%) {

  ## @name filer_messages::new()
  ## @usage
  ##   my @patterns = ('wafl\.', 'Error:')
  ##   my $fmObj = filer_messages->new(<filePath>,
  ##                                  \@patterns,
  ##                                  ('optionName' => <optValue>));
  ##
  ##   Options:
  ##
  ##      - logFilePath => "pathToLogFile"
  ##        This is the path to the file on the filer's root volume.
  ##        1st the root volume is mounted.
  ##        2nd the full file path to the log file is created
  ##        using <mountPoint>/<{logFilePath}>.
  ##
  ##      - forceFilePath => "pathToLogFile"
  ##        If this options is set. The file path that uses the
  ##        <mountPoint>/<{logFilePath}> is overridden by
  ##        <{forceFilePath}>. Usefull if your log files does not
  ##        actually exist on the filer's volume, or exists at a
  ##        non-standard mountpoint.
  ##
  ## @synopsis
  ## Create a new instance of a filer_messages Object.
  ## This method will mount (as needed) the FILER's
  ## root volume (vol0 via NFS) so that perl has access
  ## to the /etc/messages logfile.

  my $class = shift;
  my $self = {};
  bless $self, $class;

  logLevel(500);

  $self->_init(@_);
  $self->SUPER::_init($self->{filePath},
		      $self->{patterns},
		      %{$self->{opts}}
		    );
  return $self;

}

sub _init ($;$%) {

  #
  # Initialize instance variables.
  #
  # Open the file.
  #

  my $self = shift;
  my ($filer, $patterns, %opts) = @_;

  # Param: filer
  $self->{filer} = $filer;
  # Make sure that the filer is mounted
  logTrc("filer_messages:_init(): mounting rootVol of filer $filer");
  my $mountPoint = mount_root_vol($filer);
  logTrc("filer_messages:_init(): mountPoint: $mountPoint");

  # Determine path to /etc/messages
  # May override with " logFilePath => '/<path' "
  my $filePath = $mountPoint .
    ($opts{logFilePath} ? $opts{logFilePath} : "/etc/messages");

  if ($opts{forceFilePath}) {
    logTrc("filePath will be forced from ($filePath) to (".$opts{forceFilePath}.")");
  }

  # Option: forceFilePath
  # - If defined, forceFilePath will totaly ignore the filer's mount path
  #   and enforces an explicit path. Usefull for unit testing a test.
  $filePath = ($opts{forceFilePath} ? $opts{forceFilePath} : $filePath);
  $self->{filePath} = $filePath;

  if (! -f $filePath) {
    die "filer_messages: _init(), the filePath option ($filePath) must point to an existing file.";
  } else {
    logTrc("Valid FilePath: $filePath ->".$opts{forceFilePath});
  }

  # Save out patterns
  $self->{patterns} = $patterns;

  # Save out all options
  $self->{opts} = \%opts;


}

sub avg_time_between_messages (@) {

  ## @name filer_messages::avg_time_between_messages ()
  ## @usage
  ##   my @matchedLines = $fmObj->get_matched_lines()
  ##   my @interestingLines = @{$matchedLines[3]}
  ##   my $eventInterval = $fmObj->avg_time_between_messages (@interestingLines)
  ## @synopsis
  ## Given an array of log messages (from ->get_matched_lines()),
  ## determine the average time between these messages.
  ## Can be used to detect process-thrashing via /etc/messages.

  my $self = shift;

  my (@logMsgs) = @_;

  my ($line, $currDate, $prevDate, $err, $diffSecs, $avgDiffSecs, $lDate);
  my ($numDates);
  if (@logMsgs >= 2) {

    my $numDates = 0;
    $prevDate = 0;
    foreach $line (@logMsgs) {

      $line =~ m/^(\D\D\D \D\D\D \d\d \d\d\:\d\d\:\d\d \D+) (\[.*)/;
      $lDate = $1;
      if ($lDate eq "") {
	logWarn ("avg_time_between_messages (): Failed to get a dateTime from a log entry.\n$line");
      }
      $currDate = ParseDate($lDate);
      # logTrc ("currDate from ParseDate($lDate): [$currDate]");
      $numDates++;
      if ($prevDate) {
	$diffSecs += _DateToInt(DateCalc($prevDate, $currDate, \$err));
	# logTrc("diffSecs: $diffSecs");
      }
      $prevDate = $currDate;
    }
    my $avgDiffSecs = $diffSecs / $numDates;

    return $avgDiffSecs;

  } else {

    return 0;

  }

}


sub _DateToInt ($) {

  #
  # Given a dateStr YY:MM:WW:DD:HH:Mn:Sc
  # return an integer representing seconds,
  # begining with the number of Days.
  #
  # Used to convert log message timestamps to integers.
  #

  my ($dateStr) = @_;

  $dateStr =~ m/([\+\-]?)(\d+)\:(\d+)\:(\d+)\:(\d+)\:(\d+)\:(\d+)\:(\d+)/;
  my $Sn = $1;
  my $YY = $2;
  my $MM = $3;
  my $WK = $4;
  my $DD = $5;
  my $HH = $6;
  my $Mn = $7;
  my $Sc = $8;
  if (($Sc eq "")|($Mn eq "")|($HH eq "")|($DD eq "")) {
    logWarn "DateToInt(): Failed to match parse dateString input ($dateStr).";
  }
  my $tVal = $Sc + ($Mn * 60) + ($HH * 60 * 60) + ($DD * 24 * 60 * 60);
  # logTrc ("DateToInt(): $tVal = $Sc + ($Mn * 60) + ($HH * 60 * 60) + ($DD * 24 * 60 * 60)");
  $tVal *= -1 if ($Sn eq '-');
  # logTrc("dateStr: ($dateStr)->($tVal)");

  return $tVal;

}

sub _test_DateToInt () {

  #
  # This unit test Validates the DateToInt() function
  #

  logLevel(500);

  my @lines = ('Fri Jun 11 18:58:57 PDT [wafl.snap.autoDelete:info]: Deleting snapshot',
	       'Fri Jun 11 18:59:04 PDT [wafl.snap.autoDelete.createStateSnap:info]:',
	       'Fri Jun 11 18:59:22 PDT [wafl.snap.autoDelete:info]: Deleting snapshot',
	       'Fri Jun 11 18:59:25 PDT [wafl.snap.autoDelete.createStateSnap:info]:',
	       '+0:0:0:0:0:0:1',
	       '+0:0:0:0:0:1:0',
	       '+0:0:0:0:1:0:0',
	       '+0:0:0:0:0:0:15',
	       '+0:0:0:0:0:1:15',
	       '+0:0:0:0:1:0:15'
	      );

  my ($l, $lDate);
  foreach $l (@lines) {
    my ($pDate, $tDate);
    if ($l =~ m/^(\D\D\D \D\D\D \d\d \d\d\:\d\d\:\d\d \D+) \[/) {
      $lDate = $1;
      $pDate = ParseDate($lDate);
      $pDate =~ m/(\d\d(\d\d))(\d\d)(\d\d)(\d\d):(\d\d):(\d\d)/;
      $tDate = "+$2\:$3\:00\:$4\:$5\:$6\:$7";
    } else {
      $tDate = $l;
    }
    my $intDate = _DateToInt($tDate);
    logTrc ("$l\nDate ($lDate -> $pDate -> $tDate -> ".$intDate.")");
  }

  exit;

}


sub _test_avg_time_between_messages() {

  #
  # Validates the avg_time_between_messages() function.
  #

  logLevel(500);

  my $logTxt = "
Fri Jun 11 19:01:06 PDT [wafl.snap.autoDelete.createStateSnap:info]: Creating snapshot 'autodelete_base' in aggregate 'aggr1' to preserve state after auto delete of all snapshots.
Fri Jun 11 19:01:20 PDT [wafl.snap.autoDelete:info]: Deleting snapshot 'autodelete_base' in aggregate 'aggr1' to recover storage
Fri Jun 11 19:01:34 PDT [wafl.snap.autoDelete.createStateSnap:info]: Creating snapshot 'autodelete_base' in aggregate 'aggr1' to preserve state after auto delete of all snapshots.
Fri Jun 11 19:01:54 PDT [wafl.snap.autoDelete:info]: Deleting snapshot 'autodelete_base' in aggregate 'aggr1' to recover storage
Fri Jun 11 19:02:04 PDT [wafl.snap.autoDelete.createStateSnap:info]: Creating snapshot 'autodelete_base' in aggregate 'aggr1' to preserve state after auto delete of all snapshots.
Fri Jun 11 19:02:22 PDT [wafl.snap.autoDelete:info]: Deleting snapshot 'autodelete_base' in aggregate 'aggr1' to recover storage
Fri Jun 11 19:02:24 PDT [wafl.snap.autoDelete.createStateSnap:info]: Creating snapshot 'autodelete_base' in aggregate 'aggr1' to preserve state after auto delete of all snapshots.
";
  my @logLines = split("\n", $logTxt);


  my $testLogFile = '/tmp/_test_filer_messages.log';

  open(FH_tlog, "> $testLogFile");

  # Patterns look for adb thrashing
  my @trigPats = ( 'wafl.snap.autoDelete:info',
		   'wafl.snap.autoDelete.createStateSnap:info' );
  # Create an instance of a filer_messages object.
  my $FILER = 'sidlo1';
  my $fmObj = filer_messages->new($FILER,
				 \@trigPats,
				 ('forceFilePath' => $testLogFile) );

  my ($l, $thrashingFound, $avgDiffSecs);
  foreach $l (@logLines) {
    # logTrc("===> LogOutput: $l");
    print FH_tlog "$l\n";
    # Flush the FH_tlog file handle.
    select FH_tlog; $|=1;

    logTrc ("==> $l\n");
    sleep 1;
  }

  $fmObj->get_lines();

  if ($fmObj->triggered()) {

    my @matches = $fmObj->get_matched_lines();
    $avgDiffSecs = $fmObj->avg_time_between_messages(@{$matches[0]});

    logTrc ("Success! \$avgDiffSecs = $avgDiffSecs\n");

  } else {

    logTrc ("Failure! failed to find pattern matches.");

  }

  exit;

}

sub _test_filer_messages ($$) {

  my ($filer, $vol) = @_;

  logLevel(500);
  logTrc("_test_filer_messages(): Starting Test.");

  my $fh = Tharn::host($filer);
  my $fc = Tharn::connect("#rsh-socket");

  my @trigPats = ('\[wafl.scan.layout.advise:info\]: WAFL layout ratio for volume');

  # Create an instance of a filer_messages object.
  my $fmObj = filer_messages->new($filer, \@trigPats);

  my $rtxt = sayEOF($fc, "priv set -q advanced; wafl scan measure_layout $vol");

  my $done = 0;
  do {

    my $lines = $fmObj->get_lines();
    logTrc("Got Lines: \n$lines");
    if ($fmObj->triggered()) {
      logTrc("Cool: Test passed!");
      $done = 1;
    }

    sleep 1;

  } until ($done);

  my @matches = $fmObj->get_matched_lines();
  my $lines = $matches[0]->[0];
  $lines =~ m/\:info\]\: WAFL layout .* is (\d+\.\d+)\./;
  my $ratio = $1;
  logTrc("WAFL Layout ratio is: $ratio");

  logTrc("_test_filer_messages(): Test Completed.");

}

1;

# Uncomment the one of the following functions to run
# unit tests on key routines.
# - cd to the test dir
# - create a link to rm the .pm file as a .thpl
#   ln -s lib/FS/Util/filer_messages.pm filer_messages.thpl
# - run the module as a thpl script
# - nate ./filerMessage.thpl
# - when you are done testing remember to delete the link.
#
# filer_messages::_test_DateToInt ();
#
# filer_messages::_test_avg_time_between_messages()
#
# filer_messages::_test_filer_messages ('sidlo1', 'vol0');
