# Hey emacs, this is basically -*- perl -*-
# $Id: fragMeister.pm,v 1.11 2004/08/09 21:09:39 sidlo Exp sidlo $
#
# Copyright (c) 2004 Network Appliance, Inc.
# All rights reserved.
#

package fragMeister;

## @summary
##   Use this library to create a fragmented Trad or Flex Volume.
##
## @description
##
## This package creates a fragmented volume.
##
## The files generated are created by the filer's internal
## command "mkfile", and are placed on the root dir of the
## volume.
##
## Useful methods:
##
##   fragMeister->new(<filer>, <filerAggr>, <%options>);
##
##     Creates a new instance of the fragMeister object
##     and initializes key instance variables.
##
##   fragMeister->createFraggedVolume(<volName>);
##
##     Creates a fragged volume with the volume name <volName>.
##     The volume must not exist on the $FILER_AGGR specifed
##     upon this object initialization (new()).
##     The aggr must have enough room to accomodate 1 additional
##     flexVol (20m).
##
##   fragMeister->fragOneVol(<volName>,<%options>);
##
##     Generates fragmentation within the existing flexVol or
##     tradVol <volName>.
##
##   fragMeister->fillVol(<volName>, <numFiles>, <pcntFull>);
##
##     Quickly Fills a volume with <numFiles> to <pcntFull>,
##     evenly disbursing available space across those files.
##     The files are generated by the filer's internal
##     command "mkfile".
##
##   fragMeister->fragVol(<volName>, <fragIterate>);
##
##     Deletes a random set of files from the volumes root dir.
##     The number of files will be atleast 20% of the files
##     found on the root dir.
##     Generates missing files that are determined using the
##     file pattern "file_nnnnn" where nnnnn is a count from
##     0 to {fragFileCount}. If the file does not exist
##     on the root dir, it is created.
##     All available space on <volName> is randomly disbursed
##     across the missing files.
##
## Create a Fragged Volume Method 1: (works only for flexvols):
##
## The following step are performed to create
## a fragmented flexVol given an aggr with atleast
## 20m available:
##
##  fragMeister->createFraggedVolume(<volName>);
##
##  1. Create 3 drives on the aggr (atleast 20m each)
##
##  2. Vol option guarantee=volume is set for the 3 vols.
##
##  3. Fill vol1 with files (to 100%).
##
##  4. Fill vol2 with files (to 100%).
##
##  5. Delete vol3
##
##  6. Loop n times
##
##    6.1 Randomly Delete some files from vol1
##
##    6.2 Fill vol 1 with files of random size.
##
##  7. Delete vol2
##
##  8. Loop n times
##
##    8.1 Randomly Delete some files from vol1
##
##    8.2 Fill vol 1 with files of random size.
##
##
## Create a Fragged Volume Method 2: (works for tradVols and flexVols)
##
## The following method will create a fragmented
## volume given a existing empty volume (trad or flex).
##
##  fragMeister->fragOneVol(<volName>);
##
##  1. Create n (3 or more) files evenly distributing
##     available space across these files. We will
##     call this first set of files {firstFiles}.
##
##  2. Delete odd files.
##
##  3. Run fragVol():
##
##     3.1 Loop (fragIterate} times.
##
##         3.1.1. Randomly delete some files.
##
##         3.1.2. Randomly distribute available space
##                accross {fragFileCount} files and
##                create them.
##
##  4. Delete the all the existing {firstFiles}.
##
##  5. Run FragVol()
##
##
## @notes
##
##  - Creates a persistent connection to the filer using
##    rsh-socket to reduce execution time.
##
##  - Measure fragmentation with "wafl scan measure_layout".
##
## @param FILER Required. Filer on which to perform actions.
##
## @param FILER_AGGR Required. Name of aggr on which to
##        create flexVols.
##
## @param 'firstFileCount' Optional. Number of files to
##         create on initial pass of fragOneVol(). This
##         initial pass distributes the space available
##         on the volume evenly across these files.
##         Default: 3 (files)
##
## @param 'fragFileCount' Optional. Number of files to
##         create on the volume during the fragmentation
##         process. The files are created on the volumes
##         root dir using the filer's "mkfile" command.
##         The file names look like file_nnnnnnn "file_0000000",
##         "file_0000001", "file_0000301".
##         "maxfiles" are checked and changed on the volume
##         as necessary to ensure that enough inodes exist
##         to generate the files required.
##         A higher number should result in more fragmentation
##         on larger drives.
##         Default: 1000 (files)
##
## @param 'fragIterate' Optional. Number of times to iterate
##         through the fragmentation process. A higher
##         number means greater fragmentation and a longer
##         process time.
##         Default: 3 (iterations)
##
## @param 'fragFileDir' Optional. Location of existing dir in
##         which fragmenation files will be placed.
##         This dir has to be created by mounting the volume
##         as there is no 'mkdir' command on the filer 8-(.
##         Default: '' (existing dir begining at root '/')
##
## @param 'dfWaitTimeout' Optional. Time allowance before
##         a timeout occurs when waiting for df stats to quiesce.
##         df stats may not quiesce if there is other activity
##         occuring on the volume being analyzed.
##         (see: FS::Utils::space_calc wait_vol_df() and wait_aggr_df)
##         Default: 50 (seconds)
##
## @synopsis
##
##  my $fm = fragMeister->new(
##              $FILER, $FILER_AGGR,
##              # Only used by the method fragOneVol()
##              'firstFileCount' => <3 or Greater>,
##              # Number of files to create during file fragmentation
##              'fragFileCount'  => $FRAG_FILE_COUNT,
##              # Place fragFiles into a given dir (who's origin is root '/')
##              # This dir must have been created earlier.
##              'fragFileDir'    => 'milford/testProc1',
##              # Number of time to loop through the frag process
##              'fragIterate'    => $FRAG_ITERATE,
##              # Max to to wait for df to quiesce
##              'dfWaitTimeout'  => $DF_WAIT_TIMEOUT );
##
## @example
##
##   use fragMeister;
##   my $fm = fragMeister->new($FILER, $FILER_AGGR,
##                             'firstFileCount' => 5,
##                             'fragFileCount'  => 2000,
##                             'fragIterate'    => 6,
##                             'dfWaitTimeout'  => $DF_WAIT_TIMEOUT );
##    # Create a fragmented volume on $FILER_AGGR
##    $FraggedVol = $fragger->createFraggedVolume($FILER_VOL);
##    # Create a fragmented volume on $FILER_AGGR
##    $FraggedVol = $fragger->fragOneVol($TRAD_VOL);
##
##
##
## @duration
##
##    Depends on:
##      - Aggr and Volume Sizes
##      - {fragFileCount} setting
##      - {fragIterate}   setting
##
## @keywords  trad flex vol status fragmentation measure layout
##
## @status    public
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
##


use strict;
use Volume qw(vol_exists vol_purge vol_offline vol_options vol_rename);
use SayWrap qw(sayEOF);
use SPARSE::Util::Logger qw(logLevel logError logStep logInfo logWarn logTrc);
use FS::Util::space_calc;
use FS::Util::filer_messages;
use filer;
use Tharn;

sub new ($$%) {

  #
  # Create a new instance of a fragMiester object.
  #

  my $class = shift;
  my $self = {};
  bless $self, $class;
  $self->init(@_);
  logTrc("Created a fragMeister");
  return $self;

}

sub init ($$%) {

  #
  # Initialize instance variables.
  #

  my $self = shift;
  my ($filer, $aggrName, %opts) = @_;

  my $fh = Tharn::host($filer);
  my $fc = Tharn::connect("#rsh-socket");
  logTrc ("fh->hostname() => ".$fh->hostname());

  # Filer Handle Object.
  $self->{fh}              = $fh;
  # Filer Connection Object.
  $self->{fc}              = $fc;
  # Name of Aggregate
  $self->{aggrName}        = $aggrName;
  # Number of files created on initial pass.
  # Available space is evenly distributed over these files.
  $self->{firstFileCount}  = $opts{firstFileCount}
    ? $opts{firstFileCount} : 3;
  logWarn ("fragMeister->{firstFileCount} s/be 3 or Greater.")
    if ($self->{firstFileCount} < 3);
  # Number of files to generate during fragmentation process.
  # Available space is randomly distributed over these files.
  $self->{fragFileCount}   = $opts{fragFileCount}
    ? $opts{fragFileCount}  : 1000;
  # Number of iterations to run the fragmentation process, consisting of
  # deleteSomeFiles() then createSomeFiles().
  $self->{fragIterate}     = $opts{fragIterate}
    ? $opts{fragIterate}    : 3;
  # Amount of time allowed before we timeout waiting for df stats to quiesce.
  $self->{dfWaitTimeout}   = $opts{dfWaitTimeout}
    ? $opts{dfWaitTimeout}  : 50;
  # Place files into a specific dir
  $self->{fragFileDir}     = $opts{fragFileDir}
    ? $opts{fragFileDir}    : undef;
  # FileName Template
  $self->{fnTemplate}     = $opts{fnTemplate}
    ? $opts{fnTemplate}    : "file_%7.7d";
}

sub createFraggedVolumeOld ($) {

  #
  # Given an aggr, create a single fragmented volume.
  #
  # Replaced by a new version that does not create
  # intermediate flexVols 1 2 and 3, just creates
  # a single volume.
  #
  # Keeping this version around in case the new version
  # does not created fragmentation as expected.
  #

  my $self = shift;
  my ($volName) = @_;

  my $fc       = $self->{fc};

  # - Ceate 3 volumes
  $self->createVols($volName);

  # - Fill 2 volumes with data
  #   Vol3 does not need to be fill as it will be deleted soon.
  # +++ Could be done in subtests to parallelize for speed.
  logTrc("Filling Volume (".$volName."_001".")");
  $self->fillVol($volName."_001", $self->{fragFileCount});
  logTrc("Filling Volume (".$volName."_002".")");
  $self->fillVol($volName."_002", 20);

  # Fragment Volume 1
  logTrc("Fragment Volume (".$volName."_001".")");
  $self->fragVol($volName."_001", $self->{fragIterate});

  # - Destroy the 3rd volume
  logTrc("Destroying Volume (".$volName."_003".")");
  $self->destroyVol($volName."_003");

  # - Resize Vol-1 by 100%
  my $aggrAvail = $self->getAggrAvail();
  logTrc("Resize Volume (".$volName."_001".") to +$aggrAvail\KB");
  $self->resizeVol($volName."_001", "+".($aggrAvail)."KB");

  # Fragment Volume 1
  logTrc("Fragment Volume (".$volName."_001".")");
  $self->fragVol($volName."_001", $self->{fragIterate});

  # - Destroy vol-2
  logTrc("Destroying Volume (".$volName."_002".")");
  $self->destroyVol($volName."_002");

  # - Resize Vol-1 by 100%
  $aggrAvail = $self->getAggrAvail();
  logTrc("Resize Volume (".$volName."_001".") to +$aggrAvail\KB");
  $self->resizeVol($volName."_001", "+".($aggrAvail)."KB");

  # Fragment Volume 1
  logTrc("Fragment Volume (".$volName."_001".")");
  $self->fragVol($volName."_001", $self->{fragIterate});

  # Rename vol
  logTrc("Rename Volume (".$volName."_001".") to (.$volName)");
  if (!Volume::vol_rename($fc, $volName."_001", $volName)) {
    die "Failed to rename volume  ($volName\_001) to ($volName)."
  }

  return $volName;

}

sub createFraggedVolume ($) {

  #
  # Given an aggr, create a single fragmented volume.
  #

  my $self = shift;
  my ($volName) = @_;

  my $fc       = $self->{fc};
  my $aggrName = $self->{aggrName};

  # Create 1 FlexVols
  my $aggrAvail = (get_aggr_df($fc, $aggrName, 'avail'))/1024;
  # We only use 90% of the aggr (.90). w/out it we run out of space for vol_003.
  my $volSize = int (($aggrAvail * 0.90))."m";
  my $aggrName = $self->{aggrName};
  logTrc("Creating Volume(".$volName.")");
  Volume::vol_create_flex($fc, $volName, $aggrName, $volSize)
      or die "Failed to create FlexVol ($volName)";

  # Create a snapCmds object to handle volumes
  my $snapCmds = snapCmds->new($fc, "Vol");

  # Turn off snap schedules.
  $snapCmds->snapsched_volume($fc, $volName, "0 0 0");

  # Ensure that maxfiles for each volume can handle the number of files.
  # - If maxfiles allowed < fragFileCount, increase maxfiles by fragFileCount
  $self->setMaxfiles($volName, $self->{fragFileCount});

  # Fragment Volume 1
 logTrc("Fragment Volume (".$volName.")");
  $self->fragOneVol($volName, $self->{fragIterate});

  return $volName;

}

sub fragOneVol ($%) {

  #
  # Create fragmented files on a volume.
  #

  my $self = shift;
  my ($vol, %opts) = @_;

  my $fc = $self->{fc};
  $self->{firstFileCount}  =    $opts{firstFileCount}
    ? $opts{firstFileCount} : $self->{firstFileCount};
  $self->{fragFileCount}   =    $opts{fragFileCount}
    ? $opts{fragFileCount}  : $self->{fragFileCount};
  $self->{fragIterate}     =    $opts{fragIterate}
    ? $opts{fragIterate}    : $self->{fragIterate};

  logTrc("fragOneVol(): Fragmenting a single volume ($vol)");

  # - Fill the volume with 3 files.
  $self->fillVol($vol, $self->{firstFileCount});
  my @fileList = $self->getFileList($vol);
  # - Delete files 1 and 3, Leaving file 2
  my @filesLeft = $self->deleteOddFiles($vol, (sort(@fileList)));
  # - run fragVol()
  $self->fragVol($vol, $self->{fragIterate});
  # - Delete file 2
  $self->deleteTheseFiles($vol, @filesLeft);
  # - run fragVol()
  $self->fragVol($vol, $self->{fragIterate});

  logTrc("fragOneVol(): Completed volume fragmentation ($vol)");

  # Report the amount of fragmentation on the volume.
  # You'll have to look at the filer's log for the return data.
  my $cmd = "priv set -q test; wafl scan measure_layout $vol";
  my (undef, undef, $fragStat) = say($fc, $cmd);

}

sub deleteOddFiles ($@) {

  my $self = shift;
  my ($vol, @fileList) = @_;

  my ($f, $i, @filesLeft, $fp);
  foreach $f (sort(@fileList)) {
    $i++;
    if ((int($i/2) - ($i/2)) != 0) {
      push (@filesLeft, $f);
    } else {
      # Delete the file;
      $fp = $self->makeFilePath($vol, $f);
      filer_file_destroy($self->{fc}, $fp);
    }
  }

  return @filesLeft;
}

sub deleteTheseFiles ($@) {

  my $self = shift;
  my ($vol, @fileList) = @_;

  my ($f, $i, @filesLeft, $fp);
  foreach $f (sort(@fileList)) {
    # Delete the file;
    $fp = $self->makeFilePath($vol, $f);
    filer_file_destroy($self->{fc}, $fp);
  }

}

sub createVols () {

  #
  # Ceate an Aggr
  # Create 3 vols
  #

  my $self = shift;
  my ($volName) = @_;

  my $fc       = $self->{fc};
  my $aggrName = $self->{aggrName};

  # Create 3 FlexVols
  my $aggrAvail = (get_aggr_df($fc, $aggrName, 'avail'))/1024;
  # We only use 90% of the aggr (.90). w/out it we run out of space for vol_003.
  my $volSize = int (($aggrAvail * 0.90) / 3)."m";
  my $aggrName = $self->{aggrName};
  logTrc("Creating Volume(".$volName."_001".")");
  Volume::vol_create_flex($fc, $volName."_001", $aggrName, $volSize)
      or die "Failed to create FlexVol ($volName\_001)";
  logTrc("Creating Volume (".$volName."_002".")");
  Volume::vol_create_flex($fc, $volName."_002", $aggrName, $volSize)
      or die "Failed to create FlexVol ($volName\_002)";
  logTrc("Creating Volume (".$volName."_003".")");
  Volume::vol_create_flex($fc, $volName."_003", $aggrName, $volSize)
      or die "Failed to create FlexVol ($volName\_003)";

  # Create a snapCmds object to handle volumes
  my $snapCmds = snapCmds->new($fc, "Vol");

  # Turn off snap schedules.
  $snapCmds->snapsched_volume($fc, $volName."_001", "0 0 0");
  $snapCmds->snapsched_volume($fc, $volName."_002", "0 0 0");
  $snapCmds->snapsched_volume($fc, $volName."_003", "0 0 0");

  # Ensure that maxfiles for each volume can handle the number of files.
  # - If maxfiles allowed < fragFileCount, increase maxfiles by fragFileCount
  $self->setMaxfiles($volName."_001", $self->{fragFileCount});
  $self->setMaxfiles($volName."_002", $self->{fragFileCount});
  $self->setMaxfiles($volName."_003", $self->{fragFileCount});

}

sub setMaxfiles ($$) {

  my $self = shift;
  my ($vol, $fileCount) = @_;

  my $fc = $self->{fc};

  my ($currMaxfiles, $currUsedfiles);

  my $cmd = "maxfiles $vol";
  my $mftxt = sayEOF($fc, $cmd);
  if ($mftxt =~ 
      m/maximum number of files is currently\s+(\d+)\s+\((\d+)\s+used\)/) {
    $currMaxfiles  = $1;
    $currUsedfiles = $2;
  }

  if ($currMaxfiles < $fileCount) {
    my $newMaxfiles = $currUsedfiles + $fileCount + 200;
    my $cmd = "maxfiles -f $vol $newMaxfiles";
    my (undef, undef, $mftxt, $mfPrompt) = say($fc, $cmd, '-prompt' => '.*increase .*files\? ');
    logTrc("mftxt: ($mftxt)");
    if ($mfPrompt =~ m/.*increase .*files\? /) {
      # Anwser the question: 
      #    Are you sure you want to increase the maximum number of files?
      $fc->send("yes\n");
    } elsif ($mftxt ne "") {
      logWarn("Unexpected response from maxfiles-set command: ($mftxt)");
    }
  }
}

sub destroyVol () {

  #
  # Destroy a given volume.
  #

  my $self = shift;
  my ($volName) = @_;

  my $fc = $self->{fc};

  logTrc("destroyVol(): Destroying $volName");
  Volume::vol_offline($fc, $volName)
      or die "Failed to destroy volume ($volName).";
  Volume::vol_destroy($fc, $volName)
      or die "Failed to destroy volume ($volName).";

}

sub fillVol($$;$) {

  #
  # Fill a volume with to 100%, or to $pcntFull using
  #  the filer's "mkfile" command (w/out NFS).
  #

  my $self = shift;
  my ($vol, $numFiles, $pcntFull) = @_;

  $pcntFull = 100 if ($pcntFull <= 0);

  logTrc ("fillVol(): Filling Volume ($vol) to ($pcntFull)% full.");

  my $fc = $self->{fc};

  # Ensure that maxfiles for each volume can handle the number of files.
  # - If maxfiles allowed < fragFileCount, increase maxfiles by (fragFileCount+10%)
  $self->setMaxfiles($vol, int($numFiles * 1.10));

  # Fill the volume with n files, using mkfile.

  # - Determine space available
  my $tv_avail = int(get_vol_df($fc, $vol, 'avail', 0) * ($pcntFull / 100));

  # - Calculate size of files to create
  my $fileSize = int($tv_avail / $numFiles);

  # - Get existing files
  my @existingFiles = $self->getFileList($vol);

  my $i;
  my $fName;
  my $fileName;
  my @fl = undef;
  my @missingFiles = undef;
  my $fnTemplate = $self->{fnTemplate};

  if (@existingFiles) {
    #
    # If we have existing files, Allocate space across missing
    # file randomly.
    #
    for ($i=0; $i<=$numFiles; $i++) {
      $fName = sprintf ($fnTemplate, $i);
      if (!grep (/$fName/, @existingFiles)) {
	push(@missingFiles, $fName);
      }
    }
    # Allocate existing randomly evenly across missing files.
    wait_vol_df($fc, $vol, $self->{dfWaitTimeout});
    my $va = $self->getVolAvail($vol);
    my %newFiles = $self->allocateSpace($va, @missingFiles);
    my $newFilesCount = (keys (%newFiles));
    my $fName;

    # - Create files
    $i = 0;
  MakeMissingFiles:
    foreach $fName (keys(%newFiles)) {
      $fileName = $self->makeFilePath($vol, $fName);
      $i++;
      $fileSize = $newFiles{$fName};
      logTrc("Creating Missing file ($fileName) ".$i." of $newFilesCount");
      if (!filer_file_make ($fc, $fileName, "$fileSize")) {
	# Wait for filer to accumulate stats for df.
	wait_vol_df($fc, $vol, $self->{dfWaitTimeout});
	my $tva = get_vol_df($fc, $vol, 'capacity', 0);
	if ($tva == 100) {
	  logTrc("Expected mkfile error (vol full).");
	  push(@fl, $fileName);
	  last MakeMissingFiles;
	} else {
	  logWarn ("fillVol(): Error occured creating Missing file with mkfile".
		   " ($fileName)");
	}
      } else {
	push(@fl, $fileName);
      }
    }
  } else {
    #
    # If we have no existing files, Allocate space across missing
    # file evenly.
    #

    my $fnTemplate = $self->{fnTemplate};

    # - Create files
  MakeFiles:
    for ($i=0; $i<=$numFiles; $i++) {
      $fName = sprintf ($fnTemplate, $i);
      $fileName = $self->makeFilePath($vol, $fName);
      if (grep (/$fName/, @existingFiles)) {
	logTrc("File exists ($fileName), will not create.");
	next MakeFiles;
      }
      logTrc("Filling with file ($fileName) ".($i+1)." of $numFiles");
      my $sz = $fileSize * 1024;
      if (!filer_file_make ($fc, $fileName, $sz)) {
	# Wait for filer to accumulate stats for df.
	wait_vol_df($fc, $vol, $self->{dfWaitTimeout});
	my $tva = get_vol_df($fc, $vol, 'capacity', 0);
	if ($tva == 100) {
	  logTrc("Expected mkfile error (vol full).");
	  push(@fl, $fileName);
	  last MakeFiles;
	} else {
	  logWarn ("fillVol(): Error occured creating file with mkfile".
		   " ($fileName)");
	}
      } else {
	push(@fl, $fileName);
      }
    }
  }

  logTrc ("fillVol(): Filled Volume ($vol) to ($pcntFull)% full, with (".@fl.") files.");

  return @fl;
}

sub getFileList ($) {

  #
  # Return a shuffled list of files on the given volume.
  #

  my $self = shift;
  my ($vol) = @_;

  logTrc ("getFileList(): Getting file list ($vol)");

  my $fc = $self->{fc};

  my $dirPath = $self->makeFilePath($vol, '');
  my $cmd = "priv set -q test; ls $dirPath";
  my (undef, undef, $text) = say($fc, $cmd);
  my @lines = split(/\s+\r?\n/,$text);
  my @files = grep (/file_[0-9]+/, @lines);

  # logTrc("getFileList(): (".@files.") Files \n".join("\n-->", @files));

  $self->shuffle(\@files);

  return @files;

}

sub deleteSomeFiles () {

  #
  # Randomly Delete some files.
  # Returns a shuffled list of files that have been deleted.
  #

  my $self = shift;
  my ($vol, $maxDels) = @_;

  logTrc ("deleteSomeFiles(): on ($vol)");

  my @files = $self->getFileList($vol);

  my $f;
  my $rndVal;
  my $fp;
  my @deletedFiles;
  my $i = 0;

 FileDels:
  foreach $f (@files) {
    $rndVal = int(rand(3)); # Random value 0-2
    if ($rndVal == 1) {
      # Delete the file;
      $fp = $self->makeFilePath($vol, $f);
      filer_file_destroy($self->{fc}, $fp);

      # Add the file to the list of deleted files.
      push(@deletedFiles, $f);

      if ($maxDels) {
	$i++;
	if ($i >= $maxDels) {
	  last FileDels;
	}
      }
	
    }
  }

  logTrc ("deleteSomeFiles(): Deleted ".@deletedFiles." files");

  $self->shuffle(\@deletedFiles);

  return @deletedFiles;

}

sub createSomeFiles () {

  #
  # Create some files of random size.
  #

  my $self = shift;
  my ($vol, @fileNames) = @_;

  logTrc ("createSomeFiles(): on ($vol)");

  my $fc = $self->{fc};

  my $volAvail = $self->getVolAvail($vol);
  my $lastElmnt = $#fileNames;
  my $fileSize;
  my $fileName;
  my @fl;
  my $i = 0;
  my $f;

  logTrc ("Available Space on Vol ($vol): ($volAvail)");
  my %newFiles = $self->allocateSpace($volAvail, @fileNames);
  my $newFilesCount = (keys (%newFiles));

 CreateFiles:
  foreach $f (@fileNames) {
    $i++;
    $fileSize = $newFiles{$f};
    $fileName = $self->makeFilePath($vol,$f);
    logTrc("Creating file ($fileName) ".($i+1)." of ".$newFilesCount);
    if (!filer_file_make ($fc, $fileName, "$fileSize")) {
      # Wait for filer to accumulate stats for df.
      wait_vol_df($fc, $vol, $self->{dfWaitTimeout});
      my $tva = get_vol_df($fc, $vol, 'capacity', 0);
      if (($tva == 100) and ($i == @fileNames)) {
	last CreateFiles;
      } else {
	logTrc ("createSomeFiles(): Could not create file with mkfile".
		" ($fileName) no more room.");
      }
    }
  }

  logTrc ("createSomeFiles(): Created ".($i)." files");

}

sub makeFilePath ($$) {

  #
  # Create the filePath for a file to be created by mkfile
  # on the filer.
  #

  my $self = shift;
  my ($vol, $fileName) = @_;

  my $filePath;
  if ($self->{fragFileDir}) {
    $filePath = "/vol/$vol/".$self->{fragFileDir}."/$fileName";
  } else {
    $filePath = "/vol/$vol/$fileName";
  }

  return $filePath;

}

sub fragVol ($$) {

  #
  # Create fragmented files on a volume.
  #

  my $self = shift;
  my ($vol, $iterate) = @_;

  my $fc = $self->{fc};

  logTrc("fragVol(): Fragmenting volume ($vol)");

  my $i;

  for ($i=0; $i<$iterate; $i++) {

    logTrc("fragVol(): fragIteration (".($i + 1)." of $iterate)");

    # Delete some files.
    my @dFiles = $self->deleteSomeFiles($vol);

    # Create some files.
    $self->createSomeFiles($vol, @dFiles);

    # Fill in missing filenames
    $self->fillVol($vol, $self->{fragFileCount});

  }

  logTrc("fragVol(): Completed volume ($vol)");

  # Report the amount of fragmentation on the volume.
  # You'll have to look at the filer's log for the return data.
  my $cmd = "priv set -q test; wafl scan measure_layout $vol";
  my (undef, undef, $fragStat) = say($fc, $cmd);

}

sub getAggrAvail ($) {

  #
  # Return an aggr's space available.
  #

  my $self = shift;
  my ($vol) = @_;

  my $fc = $self->{fc};

  FS::Util::space_calc::wait_aggr_df($fc, $vol, $self->{dfWaitTimeout});
  return get_aggr_df($fc, $self->{aggrName}, 'avail');

}

sub getVolAvail ($) {

  #
  # Return a volume's space available.
  #

  my $self = shift;
  my ($vol) = @_;

  my $fc = $self->{fc};

  FS::Util::space_calc::wait_vol_df($fc, $vol, $self->{dfWaitTimeout});
  get_vol_df($fc, $vol, 'avail', 0);

}

sub resizeVol ($$) {

  #
  # Resize a volume.
  #

  my $self = shift;
  my ($vol, $volResize) = @_;

  my $fc = $self->{fc};

  logTrc("resizeVol(): Started...");

  Volume::vol_size($fc, $vol, $volResize);

  logTrc("resizeVol(): Completed.");

}

sub shuffle ($) {

  #
  # Performs and inplace shuffle of an array.
  #
  # Input: a reference to an array.
  #

  my $self = shift;

  my $array = shift;

  # No sense is suffling arrays of less that 2 elements.
  return if ($#{$array} < 1);

  for (my $i = @$array; -- $i;) {
    my $r = int rand ($i + 1);
    ($array -> [$i], $array -> [$r]) = ($array -> [$r], $array -> [$i]);
  }

}

sub allocateSpace ($$) {

  #
  # Allocate space on a volume accross a set of
  # files randomly.
  #
  # inputs: <volAvail>KB
  #         <arrayOfFiles>
  #
  # Returns an assoc array keyed by filename
  # containing the size of the file (to be created)
  # in bytes.
  #

  my $self = shift;
  my ($volAvail, @files) = @_;

  my $fcnt = @files;

  # Allocate file space by blocks (4096)
  # $volAvail is in KB
  my $numBlocks = ($volAvail * 1024)/4096;
  my (%fileAlloc, $iFile, $blocks);

  # Disburse available blocks across files
  logTrc("Distributing ($numBlocks) Blocks over ".@files." files...");
  if ($numBlocks > 0) {
    do {
      if ($numBlocks > 1000000) {
	# For over 1,000,000 blocks, make block allocation faster.
	$blocks = 10;
      } else {
	$blocks = 1;
      }
      $iFile = int(rand($fcnt));
      $fileAlloc{$files[$iFile]} += $blocks;
      $numBlocks -= $blocks;
      logTrc("Blocks to go: ($numBlocks)") 
	if ((($numBlocks / 10000) - int($numBlocks / 10000)) == 0);
    } until ($numBlocks <= 0);
  }

  logTrc("Recalc FileSizes from blocks. Make 0 size files 1 byte size...");
  my $f;
  foreach $f (@files) {
    if ($fileAlloc{$f} == 0) {
      # Create a single byte file.
      $fileAlloc{$f} = 1;
    } else {
      # Remove last block
      $fileAlloc{$f}--;
      # Convert blocks to bytes
      $fileAlloc{$f} *= 4096;
      # Randomize last block
      $fileAlloc{$f} += int(rand(4096));
    }
  }

  return %fileAlloc;

}

sub measure_layout_wait($) {

  my $self = shift;
  my ($vol, %opts) = @_;

  my $fh = $self->{fh};
  my $fc = $self->{fc};

  my @pats = ( "WAFL layout ratio for volume $vol" );

  my $tail = filer_messages->new($fh->hostname, \@pats);

  logStep ("Start: wafl scanmeasure_layout");
  my $cmd = "priv set -q advanced; wafl scan measure_layout $vol";
  my $wsRt = sayEOF($fc, $cmd);
  die "Failed to run command:\n  -> $cmd\n ($wsRt)" if ($wsRt ne '');

  my $timeout       = 0;
  my $startTime     = time();
  my $timeOutDfWait = $opts{timeout} ?
    ($opts{timeout} * 60) : (60 * 60); # 1hr timeout.

  my $done = 0;
  do {
    $done = $tail->get_lines_triggered();
    # Perform a timeout check.

    $timeout = 1 if ((time() - $startTime) > $timeOutDfWait);
    logWarn("measure_layout_wait(): Timed-out while waiting for".
	    " a console message.") if ($timeout);

    sleep 1;

  } until ($done or $timeout);

  # Thu Jul 29 23:08:14 GMT [wafl.scan.layout.advise:info]:
  #   WAFL layout ratio for volume adbt_vol_src1 is 1.02.
  #   A ratio of 1 is optimal.
  #   Based on your free space, 1.79 is expected.

  my @matches = $tail->get_matched_lines();
  my $msgLine = ${$matches[0]}[0];

  my %rVals;

  $msgLine =~ m/^.*volume.*is (\d+\.\d+)\. A ratio of (\d+) is optimal.* space,\s+(\d+.\d+) is expected/;
  $rVals{ratio}    = $1;
  $rVals{optimal}  = $2;
  $rVals{expected} = $3;
  $rVals{message}  = $msgLine;

  return %rVals;

}

1;
