# Hey emacs, this is basically -*- perl -*-
# $Id: nb_tail.pm,v 1.7 2004/08/18 18:10:48 sidlo Exp sidlo $
#
# Copyright (c) 2004 Network Appliance, Inc.
# All rights reserved.
#

package nb_tail;

## @summary
## Creates an object that can perform a non-blocking
## tail-read of a file.
##
## @description
## Creates an object that can perform a non-blocking
## tail-read of a file.
##
## As the object does not block on read, it can be used
## do determine the state of the system from log files.
##
## An array of regular expression patterns can be
## used to look for logged messages of interest.
## If any of the patterns match a line from the
## tailed-file, a call to the method ->trigger()
## returns true.
##
## This function is specifically not entwined with
## Tharn or other test libraries so that it continues
## to be a useful primative.
##
##
## @example
##  ----------------------------------------------------------------
##  #!/bin/perl
##
##  use strict;
##  use FS::Util::nb_tail;
##
##  # Patterns that watch for
##  my @pats = ( '10', '[257]', '131' );
##
##  my $tail = nb_tail->new("./test.out", @pats);
##
##  do {
##
##    my $lines = $tail->get_lines();
##    if ($tail->triggered()) {
##      print "*** TRIGGERED! ***\n";
##      # Dump lines that match our patterns
##      my @matches = $tail->get_matched_lines();
##      my ($i, $j);
##      for ($i=0; $i<@matches; $i++) {
##        if (defined $matches[$i]) {
##  	for ($j=0; $j<@{$matches[$i]}; $j++) {
##  	  print "[(".$pats[$i].") $i:$j] ".${$matches[$i]}[$j]."\n";
##          }
##        }
##      }
##      # Die if we match to pattern-3 '131'
##      die "Done." if (defined $matches[2]);
##    } else {
##      # Print lines since last call to $tail->get_lines()
##      print "$lines";
##      sleep 5;
##    }
##
##  } until (0);
##  ---------------------------------------------------------------
##
## @notes
##    Added a check for file truncation (using stat()) if nothing
##    new appears in our read buffer. If file truncation has occured,
##    we reopen the file being tailed, returning data written to the
##    file so far, and then set the file handle to EOF and non-blocking
##    read. This allows us to interoperate with legacy routines
##    that perform truncation of /etc/messages upon initialization.
##    Without this check a file truncation can leave us reading a file
##    handle which seems okay, but is no-longer attached to the
##    file of interest.
##
## @keywords tail syslog messages etc
##
## @status public
##
## @author dl-johny@netapp.com (FSS WAFL Quality)

use strict;
use Fcntl;
use POSIX;
use IO::Handle;
use Data::Dumper;

sub new ($$;%) {


  ## @name nb_tail::new()
  ## @usage
  ##   my @patterns = ('wafl\.', 'Error:')
  ##   my $tailObj = nb_tail->new(<filePath>,
  ##                             \@patterns);
  ## @synopsis
  ## Create a new instance of a nb_tail object.
  ## nb_tail means (non-blocking tail).
  ## This object is usefull for keeping an eye
  ## on a log files.
  ##   - Specify a file to tail.
  ##   - Specify an array of patterns to look for.

  my $class = shift;
  my $self = {};
  bless $self, $class;
  $self->_init(@_);
  return $self;

}

sub _init ($;$%) {

  #
  # Initialize instance variables.
  #
  # Open the file.
  #

  my $self = shift;
  my ($filePath, $patterns, %opts) = @_;

  $self->{filePath} = $filePath;

  # Save out patterns
  $self->{patterns} = $patterns;

  # Option: bufferSize
  $self->{bufferSize} = $opts{bufferSize} ?
    $opts{bufferSize} : (256*1024);

  $self->{FH} = new IO::Handle;

  @{$self->{origStat}}  = stat($self->{filePath});

  # Open the file and tail it.
  $self->_nb_open_file(0);

}

sub get_lines () {

  ## @name nb_tail::get_lines()
  ## @usage
  ##   my $tailedLines = $tailObj->get_lines()
  ## @synopsis
  ##   - Get tailed lines since the last call get_lines()
  ##
  ##   - A call to this method will also perform
  ##     pattern matching and turn on the instance's
  ##     trigger variable if any matches occur.
  ##     - Call _nb_read() to pull queued data.
  ##     - Call _check_lines() to look for triggers
  ##
  ##   - Returns the lines that have just been read
  ##     as a text string.
  ##     - Return the string from _nb_read()
  ##
  ##   - The lines last read are also available from
  ##     the instance variable nb_tail->{lastLines}

  my $self = shift;

  my $lines = $self->_nb_read();
  $self->_check_lines($lines);
  # Save out the last lines read into {lastLines}
  $self->{lastLines} = $lines;

  return $lines;

}

sub triggered () {

  ## @name nb_tail::triggered()
  ## @usage
  ##    if ($tailObj->triggered()) { ... }
  ## @synopsis
  ##   Returns true if any pattern matches have occured.
  ##
  ##   - See if any patterns have been triggered
  ##     Returns true (1) if any patterns have
  ##     matched to any lines from the tailed file.
  ##     Once called, the trigger is reset to zero
  ##     such that an immediate subsequent calls
  ##     to this same function will return 0.

  my $self = shift;

  my $triggered = $self->{trigger};
  $self->{trigger} = 0;

  return $triggered;

}

sub get_lines_triggered () {

  ## @name nb_tail::get_lines_triggered()
  ## @usage
  ##   if ($tailObj->get_lines_triggered()) {
  ##     my $lastLinesRead = $tailObj->{lastLinesRead};
  ##    ...
  ##   }
  ## @synopsis
  ##   - Read currently queued lines.
  ##   - Check for pattern matches.
  ##   - Return true if pattern match occurs.
  ##   - The lines last read in are available from
  ##     nb_tail->{lastLines}

  my $self = shift;

  $self->get_lines();

  my $triggered = $self->{trigger};
  $self->{trigger} = 0;

  return $triggered;

}

sub get_matched_lines () {

  ## @name nb_tail::get_matched_lines()
  ## @usage
  ##  my @matchedMsgs = $tailObj->get_matched_lines()
  ##  my @pat0matches = @{$matchedMsgs[0]}
  ##  my @pat1matches = @{$matchedMsgs[1]}
  ##  my @pat2matches = @{$matchedMsgs[2]}
  ##  my $pat2Line0   = ${$matchedMsgs[2]}[0]
  ## @synopsis
  ## Returns an array of array-refs where the 1st dimension
  ## corresponds with the pattern and the 2nd dimension
  ## is a list of the patterns matching lines.
  ##
  ## This method clears the array that accumulates the
  ## matching lines.
  ##
  ##   - Get an array of matching line patterns
  ##     since the last call to getmatchingLines()
  ##   - Returns a 2d array
  ##     - 1d corresponds to the patterns in the
  ##       pattern array.
  ##     - 2d is the list of lines that matched
  ##       the a corresponding pattern.
  ##     - 1d is undefined if no matches to the
  ##       corresponding pattern occured.
  ##       You must check for this with
  ##       "if (defined $matches[$i])"
  ##     - Once called, the array that holds on
  ##       the pattern matches is cleared of any
  ##       values. Thus, immediate subsequent calls
  ##       to this same function will return
  ##       an undefined array. Thus ->get_lines()
  ##       should be the next call.


  my $self = shift;

  my @retArray = @{$self->{matches}};

  # Clear the instance variable that accumulates
  # lines that match patterns.
  @{$self->{matches}} = ();

  return @retArray;


}

sub dumpMatches (@2) {

  ## @name nb_tail::dumpMatches()
  ## @usage
  ##   my $matchArray = $tailObj->dumpMatches()
  ## @synopsis
  ##   Returns a text string that displays
  ##   the array of lines that matched a corresponding
  ##   pattern.

  my $self = shift;
  my (@matches) = @_;

  my @pats = @{$self->{patterns}};
  my $rv = "";

  my ($i, $j);
  for ($i=0; $i<@matches; $i++) {
    if (defined $matches[$i]) {
      for ($j=0; $j<@{$matches[$i]}; $j++) {
	$rv .= "[(".$pats[$i].") $i:$j] ".${$matches[$i]}[$j]."\n";
      }
    }
  }


  return $rv;

}

sub _nb_open_file ($) {

  my $self = shift;
  my ($newFile) = @_;

  # Param: filePath
  my $filePath = $self->{filePath};
  if (! -f $filePath) {
    die "nb_tail: _init(), the filePath option ($filePath) must point to an existing file.";
  }

  # Tail the file of interest.
  open($self->{FH}, "< ".$filePath)
    or die "Failed to open file for reading ($filePath).";

  my ($buffer, $retBuffer);
  if ($newFile) {
    my $BUFSIZ=$self->{bufferSize};
    $retBuffer = '';
    my $rv;
    $buffer = '';
    $rv = sysread($self->{FH}, $buffer, $BUFSIZ);
    if (!defined($rv) && $! == EAGAIN) {
      # nothing available
      # would have blocked
      # print "_nb_read(): Got no data\n";
    } else {
      # got some data
      # successfully read $rv bytes from HANDLE
      # print "_nb_read(): Got some data\n$buffer\n";
      $retBuffer = $buffer;
    }
  }

  # Seek to the end of the file.
  sysseek($self->{FH}, 0, SEEK_END);

  # Get the file handle's current properties
  my $flags = '';
  fcntl($self->{FH}, F_GETFL, $flags)
    or die "Couldn't GET flags for HANDLE : $!\n";
  # Set the file handle on the tail to non-blocking.
  $flags |= O_NONBLOCK;
  fcntl($self->{FH}, F_SETFL, $flags)
    or die "Couldn't SET flags for HANDLE: $!\n";

  if ($newFile) {
    return $retBuffer;
  }

}

sub _nb_read (@) {

  #
  # Get messages in the input que.
  #
  # This method performs non-blocking
  # reads and removed all messages from the
  # input buffer.
  #
  # Return arrays associated to the input
  # match patterns.
  #

  my $self = shift;

  my ($buffer, $retBuffer);
  my $BUFSIZ=$self->{bufferSize};
  $retBuffer = '';
  my $rv;

  # This loop will allow us to read all the queued
  # data if the queued data is greater than our BUFSIZ.
  do {
    $buffer = '';
    $rv = sysread($self->{FH}, $buffer, $BUFSIZ);
    if (!defined($rv) && $! == EAGAIN) {
      # nothing available
      # would have blocked
      # print "_nb_read(): Got no data\n";
    } else {
      # got some data
      # successfully read $rv bytes from HANDLE
      # print "_nb_read(): Got some data\n$buffer\n";
      $retBuffer .= $buffer;
    }

  } until (defined($rv) || $! != EAGAIN);

  # Use fstat to see if our file was recently recreated.
  # - Reopen the file.
  # - Load up the retBuffer w/current data.
  # - Tail the file.
  #
  # Required if file is replaced from under us.
  if (length($retBuffer) == 0) {
    if ($self->_new_file_detected()) {
      # Close the current file handle.
      $self->{FH}->close();
      $self->{FH} = new IO::Handle;
      # Reopen the file tailing it, but also
      # read the file to the end.
      $retBuffer = $self->_nb_open_file(1);
    }
  }

  return $retBuffer;

}

sub _new_file_detected () {

  my $self = shift;

  my @newStat = stat($self->{filePath});

  my $origCtime = $self->_get_fstat($self->{origStat}, 'ctime');
  my $origSize  = $self->_get_fstat($self->{origStat}, 'size');

  my $newCtime  = $self->_get_fstat(\@newStat, 'ctime');
  my $newSize   = $self->_get_fstat(\@newStat, 'size');

  my $rv = 0;

    if  ($origSize > $newSize) {
    # A new file exists.
    # Hold on to the new file stat data.
    $self->{origStat} = \@newStat;
    $rv = 1;
    # print STDOUT "*** New File Detected ***\n";
  }

  return $rv;

}

my %fsItems = ('dev'    =>0,  'ino'   =>1,  'mode' =>2, 'nlink'=>3,
	       'uid'    =>4,  'gid'   =>5,  'rdev' =>6, 'size' =>7,
	       'atime'  =>8,  'mtime' =>9,  'ctime'=>10,
	       'blksize'=>11, 'blocks'=>12 );

sub _get_fstat () {

  my $self = shift;
  my ($fsArray, $fsi) = @_;

  return @{$fsArray}[$fsItems{$fsi}];

}

sub _check_lines () {

  #
  # Check out input buffer for any matching
  # patterns.
  #
  # - Turn on the {trigger} if any patterns match.
  # - Save matching patterns to instance array var.
  #

  my $self = shift;

  my ($linesIn) = @_;

  my @pats = @{$self->{patterns}};

  # print "_check_lines(): lines -> \n$linesIn\n";

  my @lines = split(/\r?\n/, $linesIn);
  my ($l, $i, $pat);
  foreach $l (@lines) {
    foreach ($i=0; $i<@pats; $i++) {
      $pat = @pats[$i];
      # print "_check_lines(): matching($l =~ m/$pat/)\n";
      if ($l =~ m/$pat/) {
	$self->{trigger} = 1;
	# Push the matching line into an array that
	# corresponds with the index of it's pattern.
	push (@{$self->{matches}[$i]}, $l);
      }
    }
  }

}

sub _test_nb_tail () {

  my $i;

  my @pats = ( '10', '[257]', '131' );

  my $filePath = "./test.out";

  my $fh = new IO::Handle;

  open($fh, "> $filePath");

  my $tail = nb_tail->new($filePath, \@pats);

  my @matches;
  my $done = 0;

  do {

    $i++;

    # Close and reopen the file to test
    # nb_tail's ability to continue to tail
    # the new file.
    # - DOING THIS CAUSES ONE LINE TO GET LOST.
    if ($i == 10) {
      print STDOUT "\n===Deleting the file!===\n";
      close $fh;
      unlink $filePath;
      open($fh, "> $filePath");
    }

    print $fh "Line ----> $i <-----\n";
    # Flush the $fh file handle.
    my $oldFH = select($fh); $| = 1; select($oldFH);

    my $lines = $tail->get_lines();
    if ($tail->triggered()) {
      print STDOUT "*** TRIGGERED! ***\n";
      # Dump lines that match our patterns
      @matches = $tail->get_matched_lines();
      my ($i, $j);
      for ($i=0; $i<@matches; $i++) {
	if (defined $matches[$i]) {
	  for ($j=0; $j<@{$matches[$i]}; $j++) {
	    print STDOUT "[(".$pats[$i].") $i:$j] ".${$matches[$i]}[$j]."\n";
          }
        }
      }
      # Die if we match to pattern-3 '131'
      $done = 1 if (defined $matches[2]);
    } else {
      # Print lines since last call to $tail->get_lines()
      print STDOUT "--->\n";
      print STDOUT "$lines";
      print STDOUT "<---\n";
      sleep 1;
    }

    $| = 1;

  } until ($done);

  $fh->close();

  if (${$matches[2]}[0] eq "Line ----> 131 <-----") {
    print STDOUT "\n*** Test Successful!\n\n";
  }
}

# Unit test this package
# - uncomment the Line: _test_nb_tail()
# - run: perl ./lib/FS/Util/nb_tail.pm
#
# _test_nb_tail();

 
## @name nb_tail::DESTROY()
## @usage
##   my $tailObj = nb_tail->new(<filePath>,
##                             \@patterns);
##   ...
##   $tailObj->DESTROY();
## @description
##   destructor of the object
##
sub DESTROY {
    my $self = shift;

    # interesting corner case - if we die during construction, then the
    # object goes out of scope immediately and DESTROY is called before
    # the exception is propagated. this means we need to preserve any eval
    # error before proceeding.
    my $eval_error = $@ if ($@);

    print STDOUT __PACKAGE__ . "::DESTROY: starting...\n";

    if (defined($self->{FH})) {
	if (!$self->{FH}->close()) {
	    if ($!) {
		warn "nb_tail filehandle close: $!";
	    } else {
		warn "exit status $? from nb_tail filehandle close";
	    }
	}
    }

    # now, reset the eval error
    $@ = $eval_error;
}

1;
