# -*- perl -*-
#  $Id: snapCmds.pm,v 1.5 2004/06/25 18:38:56 sidlo Exp sidlo $

## @summary SnapCmds library -- modification subroutines
##
## @description
##    This package defines routines which attempt to perform snap*
##    operations that actually modify filer state, instead of
##    just gathering information from the filer and returning it.
##
##    This library is cloned from remi's SnamCmds.pm and modified
##    to handle Aggregates "-A".
##
##    SnapCmds have been encapsulated into an object so that multiple
##    objects within a process may use the same library and choose
##    different modes of execution "-V" volumes or "-A" aggrs.
##
## @usage
##
##    use FS::Util::snapCmds;
##
##    $SnapAggr = new snapCmds (<AggrOrVolMode>);
##
##    Parameters to new():
##        'A|a|Aggr' Aggregate mode
##        'V|v|Vol'  FlexVol   mode
##
##    $SnapAggr->{TEST_MODE} = 1; # Prefix Commands with "priv set test;"
##
##    # Turn off scheduled snaps down to the min.
##
##    my $sched = $SnapAggr->snapsched_volume($Fc,$FILER_AGGR, "0 0 0 0");
##
##    # Create a snapshot of an Aggr
##
##    $SnapAggr->snapshot_create ($filer, $FILER_AGGR, $snapname);
##
##    Important Instance Variables:
##      VOL_AGGR [-A {Operate on Aggregates} -V {Operate on Volumes}]
##      TESt_MODE [ prefix some commands with "priv set test"]
##
## @dependencies
##    Time::Local
##    lib/SayWrap.pm
##    FSR/FSRutils.pm
##    Tharn
##    SPARSE::Util::Logger
##
## @bugs
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public

package snapCmds;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw();

use Time::Local;
use SayWrap;
use FSR::FSRutils;
use Tharn;
use strict;
use SPARSE::Util::Logger qw(logLevel logError logStep logInfo logWarn logTrc);
use Filer::Version qw(get_wafl_version 
		      WV_SCRIMSHAW_MIN   WV_SCRIMSHAW_MAX
		      WV_ANCHORSTEAM_MIN WV_ANCHORSTEAM_MAX);

my $ERR_NO_SNAP = "No snapshots exist";
use Carp qw(croak);

sub new {

  ## @name ::new
  ## @usage new snapCmds | snapCmds->new
  ## @synopsis
  ##    Create a new instance of a snapCmds object.
  ##    Returns a reference to a snapCmds object.

  my $class = shift;
  my ($fc, $mode) = @_;
  my $self = {};
  bless $self, $class;
  $self->initialize($fc, $mode);
  return $self;
}

sub initialize() {

  ## @name ::initialize
  ## @usage $self->initialize();
  ## @synopsis
  ##    Called from new() to initialize a new snapCmds object.
  ##    Nothing returned.

  my $self = shift;
  my ($fc, $mode) = @_;
  # -V operate on FlexVols   (Default)
  # -A operate on Aggregates
  my $vers = get_wafl_version($fc->{hostp}) 
	  or croak $@;
  if ($mode =~ /Aggr|A|a/) {
	  if ($vers <  WV_ANCHORSTEAM_MIN) {
		  croak "can not call snap cmds on aggregates for version $vers \n";
	  }
	  $self->{VOL_AGGR} = "-A";
  } elsif (($mode =~ /Vol|V|v/) || (!$mode)) {
	  if ($vers < WV_ANCHORSTEAM_MIN) {
		  $self->{VOL_AGGR} = "";
	  } else {
		  $self->{VOL_AGGR} = "-V";
	  }
  } else {
	  croak "Unknown mode sent to snapCmds->new()";
  }
  # Required to test a "snap schedule" with a resolution of mins.
  $self->{TEST_MODE} = 0;
  # Flag that we are running for the fist time.
  $self->{firstTime} = 1;
  $self->{currTime}  = 0;
  $self->{priorTime} = 0;
  undef @{$self->{sl_prev}};
  undef @{$self->{sl_curr}};
}

sub snapshot_create($$$) {

  ## @name ::snapshot_create
  ## @usage $snapCmds->snapshot_create($filer, $volume, $snapname)
  ## @synopsis
  ##    Attempts to create a snapshot on the given volume.
  ##    Returns success or failure, with an error message in $@ on failure.

  my $self = shift;
  my ($filer, $volume, $snapname) = @_;

  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my $text = sayEOF($filer, "snap create $vol_aggr $volume $snapname");
    $text =~ s/^creating snapshot\.\.\.(\r\n|$)//;
    croak "$text\n" if ($text);
    return 1;
  }
}

sub snapshot_delete($$$) {

  ## @name ::snapshot_delete
  ## @usage $snapCmds->snapshot_delete($filer, $volume, $snapname)
  ## @synopsis
  ##    Attempts to delete a snapshot on the given volume.
  ##    Returns success or failure, with an error message in $@ on failure.

  my $self = shift;
  my ($filer, $volume, $snapname) = @_;

  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my $text = sayEOF($filer, "snap delete $vol_aggr $volume $snapname");
    $text =~ s/^deleting snapshot\.\.\.(\n|$)//;
    croak "$text\n" if ($text);
    return 1;
  }
}

sub snapshot_delete_all($$) {

  ## @name ::snapshot_delete_all
  ## @usage $snapCmds->snapshot_delete_all($filer, $volume)
  ## @synopsis
  ##    Attempts to delete all snapshots on the given volume.
  ##    Returns success or failure, with an error message in $@ on failure.

  my $self = shift;
  my ($filer, $volume) = @_;

  return eval {
    my @snapshots = $self->snapshot_list($filer, $volume);
    my $ss;
    foreach $ss (@snapshots) {
      $self->snapshot_delete($filer, $volume, $ss);
    }
  }
}

sub snapshot_rename($$$$) {

  ## @name ::snapshot_rename
  ## @usage $snapCmds->snapshot_rename($filer, $vol, $snap_from, $snap_to)
  ## @synopsis
  ##    Attempts to rename a snapshot on the give volume.
  ##    Returns success or failure, with error message in $@ on failure.

  my $self = shift;
  my ($filer, $volume, $snap_from, $snap_to) = @_;

  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my $cmd = "snap rename $vol_aggr $volume $snap_from $snap_to";
    logTrc("snapshot_rename() - cmd:($cmd)");
    my $text = sayEOF($filer, $cmd);
    croak "$text\n" if ($text);
    return 1;
  }
}

sub snapshot_exists($$$) {

  ## @name ::snapshot_exists
  ## @usage $snapCmds->snapshot_exists($filer, $volume, $snap_name)
  ## @synopsis
  ##    Return whether a snapshot of specific name exists.
  ##    On error, an error message is placed in $@

  my $self = shift;
  my ($filer, $volume, $snap_name) = @_;

  return eval {
    my @dst_snap_list = $self->snapshot_list($filer, $volume);
    if ($@ =~ /volume $volume is offline/) {
      # Due to a bug in slideruler
      sleep 10;
      @dst_snap_list = $self->snapshot_list($filer, $volume);
    }
    croak $@ if ($@);

    my $snap;
    while (defined ($snap = pop(@dst_snap_list))) {
      last if ($snap eq $snap_name)
    }

    if (defined $snap) {
      return 1;
    } else {
      return 0;
    }
  };
}

sub snapshot_list($$) {

  ## @name ::snapshot_list
  ## @usage $snapCmds->snapshot_list($filer, $volume)
  ## @synopsis
  ##    Returns a list of all the snapshots in the specified volume

  my $self = shift;
  my ($filer, $volume) = @_;

  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my @snapshots = ();
    my $text = sayEOF($filer, "snap list $vol_aggr $volume");

    # Return if list is empty
    return @snapshots
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "----------  ----------  ------------  --------")
    croak "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s+\-+\s+\-+\s*(.*)$/$1/s);

    chomp $text;

    my @lines = split(/\r?\n/,$text);
    @snapshots = map{
      croak "Error parsing snapshot list: $_\n"
      unless (/^(\s*\S+\s*)*?(\S+)\s*(\(\S+\))?$/);
      $2
    } @lines;
    return @snapshots;
  }
}

sub snapshot_list_wtime($$) {

  ## @name ::snapshot_list_wtime
  ## @usage $snapCmds->snapshot_list_wtime($filer, $volume)
  ## @synopsis
  ##    Returns a list of all the snapshots in the specified volume
  ##    Along with time data.
  ##     "05 25 20:00|hourly.0"
  ##     "05 25 16:00|hourly.1"
  ##     "05 25 12:00|hourly.2"
  ##     "05 25 08:00|hourly.3"
  ##     "05 25 00:00|nightly.0"
  ##     "05 24 20:00|hourly.4"
  ##     "05 24 16:00|hourly.5"
  ##     "05 24 00:00|nightly.1"
  ##

  my $self = shift;
  my ($filer, $volume) = @_;

  my %months = ('Jan' => '01', 'Feb' => '02', 'Mar' => '03',
		 'Apr' => '04', 'May' => '05', 'Jun' => '06',
		 'Jul' => '07', 'Aug' => '08', 'Sep' => '09',
		 'Oct' => '10', 'Nov' => '11', 'Dec' => '12'
	         );

  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my @snapshots = ();
    my $text = sayEOF($filer, "snap list $vol_aggr $volume");

    # Return if list is empty
    return @snapshots
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "----------  ----------  ------------  --------")
    croak "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s+\-+\s+\-+\s*(.*)$/$1/s);

    chomp $text;

    my @lines = split(/\r?\n/,$text);
    # logTrc("lines: ".join("\n ===>", @lines)."\n");
    my $sf;
    foreach $sf (@lines) {
      # logTrc("=@=> $sf");
      $sf =~ s/^.*\)\s+((\D{3,3} \d\d \d\d\:\d\d)\s+(.*))$/\2|\3/;
      $sf =~ s/^(\D{3,3}) / /;
      $sf = $months{$1}.$sf;
      $sf =~ s/\s+$//;
      # logTrc("=x=> $sf");
      push (@snapshots, $sf)
    }
    return @snapshots;
  }
}

sub snapshot_schedualize($$$) {

  ## @name ::snapshot_schedualize
  ## @usage $snapCmds->snapshot_schedualize($filer, $volume, $timestr)
  ## @synopsis
  ##    Renames all existing snapshots such that they look like
  ##    they were all scheduled.
  ##
  ##      newest file <$timestr>.0
  ##
  ##      oldest file <$timestr>.nn
  ##
  ##    Caveat: snapshots taken within less that 1min of each other
  ##            may not fall into proper time sequence. This is
  ##            due to the 1min resolution of the "snap list"'s
  ##            output.
  ##

  #=========================================
  # Rename snapshots so that they look like
  # min to min scheduled snaps.
  #
  # Valid timestr values:
  #   weekly, daily, hourly, min
  #=========================================

  my $self = shift;
  my ($filer, $volume, $timestr) = @_;
  my @validTimestrs = ( 'weekly', 'daily', 'hourly', 'min' );

  # Get list of aggr shapshots
  # Get this list along with time sequence
  $self->{VOL_AGGR} = "-A";
  my @snapList = $self->snapshot_list_wtime($filer, $volume);
  logTrc("snapList: ".join("\n", @snapList)."\n");


  # Go through list, renaming files from
  #     min.0(latest) to min._n_(oldest)
  my $i = @snapList;
  my $sf;
  foreach $sf ((sort(@snapList))) {
    $i--;
    logTrc("sf: $sf");
    my ($snapTime, $snapName) = split('\|', $sf);
    my $newName = "$timestr.".$i;
    logTrc("snapTime:($snapTime)   snapName:($snapName)   renameTo:($newName)");
    $self->{VOL_AGGR} = "-A";
    $self->snapshot_rename($filer, $volume, $snapName, $newName);
  }

  # Output the new snapshot list.
  $self->{VOL_AGGR} = "-A";
  my @newSnapList = $self->snapshot_list_wtime($filer, $volume);
  logTrc("newSnapList: ".join("\n", @newSnapList)."\n");

}

sub snapshot_flags($$$) {

  ## @name ::snapshot_flags
  ## @usage snapshot_flags($filer, $volume, $snapshot)
  ## @synopsis
  ##    For a given snapshot name, returns the list of all the active flags on it
  ##    e.g., for remi2(0007462702)_rum0_dreg0-dst.48 returns ("busy","snapmirror")
  ##    Does 'croak' if the snapshot does not exist.

  my $self = shift;
  my ($filer, $volume, $snapshot) = @_;
  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my @flags = ();
    my $text = sayEOF($filer, "snap list $vol_aggr $volume");

    # Return if list is empty
    croak "No snapshots on volume $volume\n"
      if ($text =~ /$ERR_NO_SNAP/i);

    # Strip everything up to the beginning of the list
    # (Keep everything after the "----------  ----------  ------------  --------")
    croak "Error stripping output header: $text\n"
      unless ($text =~ s/.*?\s*\-+\s+\-+\s+\-+\s+\-+\s*(.*)$/$1/s);

    my $snapexp = regx_escape($snapshot);

    if ($text =~ /$snapexp\s*(\(\S+\))?/s) {
      my $flagstr = ($1 ? $1 : undef);
      if ($flagstr && ($flagstr =~ /^\((\S+)\)$/)) {
        # Return the list of flags
        @flags = split(/,/,$1);
        return @flags;
      }
      else {
        # No flags, return empty list
        return @flags;
      }
    }
    else {
      # Return if snapshot isn't there
      croak "No such snapshot \"$snapshot\"\n";
    }
  }
}

sub snaplist_identical($$$$) {

  ## @name ::snaplist_identical
  ## @usage $snapCmds->snaplist_identical($filer1, $vol1, $filer2, $vol2)
  ## @synopsis
  ##	Return whether the list of snapshots on $vol1 and $vol2
  ##	are identical.

  my $self = shift;
  my ($filer1, $vol1, $filer2, $vol2) = @_;

  my (@snap_list1, @snap_list2);
  my ($snap1, $snap2);

  @snap_list1 = $self->snapshot_list($filer1, $vol1);
  @snap_list2 = $self->snapshot_list($filer2, $vol2);

  while (1) {
    $snap1 = pop(@snap_list1);
    $snap2 = pop(@snap_list2);

    return 1 if (!$snap1 && !$snap2);

    return 0 if ($snap1 ne $snap2);
  }
}

sub snaprestore_volume($$$) {

  ## @name ::snaprestore_volume
  ## @usage snaprestore_volume($filer, $volume, $snapshot)
  ## @synopsis
  ##    Does a snaprestore of the given volume to the given snapshot.
  ##    Returns success or failure, with an error message in $@ on failure.

  my $self = shift;
  my ($filer, $volume, $snapname) = @_;

  return eval {
    my $vol_aggr = $self->{VOL_AGGR};
    my $text = sayEOF($filer, "snap restore $vol_aggr -f -s $snapname $volume");
    $text =~ s/(Volume $volume\: revert successful\.)|(Aggregate $volume\: revert successful\.)//;
    croak "$text\n" if ($text);
    return 1;
  }
}

sub snaprestore_file($$$%) {

  ## @name ::snaprestore_file
  ## @summary Restore a file from a snapshot
  ## @usage $snapCmds->snaprestore_file($filer, $file_path, $snapshot [, restore_as_path=><path> ])
  ## @args filer Required: hostspec of filer
  ## @args lun_path Required: path to the lun to be restored
  ## @args snapshot Required: name of snapshot from where to be restored
  ## @args restore_as_path Optional: File will be restored to this path if specfied
  ## @synopsis
  ##    Does a snaprestore of the given file to the given snapshot.
  ##    The file is restored to its original path if restore_as_path is not specified.
  ##    If restore_as_path is  specified, the path must be a full path to a filename, 
  ##    and must be in the same volume as the volume used for the restore.
  ##    Returns success or failure, with an error message in $@ on failure.
  ##
  ## @notes Dependencies for sfsr
  ##    snaprestore must be licensed
  ##    volume must be online
  ##    volume must not be full
  ##    restore path if specified must be in same volume
  ##    cannot restore directories, files with NT Streams
  ##    exisiting files will be overwritten

  # Possible error messages (for reference)
  # 1. SNAPRESTORE is not licensed.
  #    (Use the "license" command to license it.)
  # 2. snap restore: File does not exist in snapshot.
  # 3. snap restore: Must specify file name to be restored as.
  # 4. snap restore: Volume <volname> does not have a snapshot named <snapname>.
  # 5. snap restore: volume name or path have to be specified.
  # 6. snap restore: Snapshot name is required for -s option.
  # 7. snap restore error: No space left on device.
  # 8. snap restore: volname or pathname <path> does not exist.
  # 9. snap restore: restore_as_path and restore_from_path have to be specified in the same volume.

  # Usage: snap restore [-A | -V] [-f] [-t vol | file ] [-s name]
  #                     [-r <restore_as_path>] <vol-name> | <restore_from_path>

  my $self = shift;
  my ($filer, $path, $snapname, %args) = @_;

  my $restore_as_path = defined($args{restore_as_path}) ? $args{restore_as_path} : '';

  return eval {

    my $cmd;
    my $vol_aggr = $self->{VOL_AGGR};
    if ($restore_as_path) {
        $cmd = "snap restore $vol_aggr -f -s $snapname -r $restore_as_path $path";
    } else {
        $cmd = "snap restore $vol_aggr -f -s $snapname $path";
    }
    my $text = sayEOF($filer, $cmd);
    # this command does not put out any output via rsh - never has
    # so success if there is no output
    # burt 97022
    croak "$text\n" if ($text);
    return 1;
  }
}

sub snaprestore_lun($$$%) {

  ## @name ::snaprestore_lun
  ## @summary Restore a lun from a snapshot
  ## @usage $snapCmds->snaprestore_lun($filer, $lun_path, $snapshot [, restore_as_path=><path> ])
  ## @args filer Required: hostspec of filer
  ## @args lun_path Required: path to the lun to be restored
  ## @args snapshot Required: name of snapshot from where to be restored
  ## @args restore_as_path Optional: Lun will be restored to this path if specfied
  ## @synopsis
  ##    Does a snaprestore of the given lun to the given snapshot.
  ##    The lun is restored to its original path if restore_as_path is not specified.
  ##    If restore_as_path is  specified, the path must be a full lun path, 
  ##    and must be in the same volume as the volume used for the restore.
  ##    Returns success or failure, with an error message in $@ on failure.

  my $self = shift;
  $self->snaprestore_file(@_);
}

sub snapreserve_volume($$;$) {

  ## @name ::snapreserve_volume
  ## @usage $snapCmds->snapreserve_volume($filer, $volume, $reserve_percent)
  ## @synopsis
  ##   Gets or sets the snap reserve for the specified volume.
  ##   The third argument is optional.  If given, it specifies the
  ##   percentage to set the snap reserve to, and the function returns
  ##   success or failure with an error message in $@ on failure. If the
  ##   third argument is omitted, and the function returns a tuple of
  ##   ($reserve_percent, $kilobytes), with an error message in $@ on failure.

  my $self = shift;
  my ($filer, $volume, $reserve) = @_;

  return eval {

    my $vol_aggr = $self->{VOL_AGGR};

    if (defined $reserve) {
      # Set the snap reserve
      my $text = sayEOF($filer, "snap reserve $vol_aggr $volume $reserve");
      croak "$text\n" if ($text);
      return 1;

    } else {
      # Determine the snap reserve
      my $text = sayEOF($filer, "snap reserve $vol_aggr $volume");
      croak "$text\n"
	unless ($text =~ s/Volume $volume: current snapshot reserve is (\d+)% or (\d+) k-bytes\.//);
      my ($percent, $kb) = ($1, $2);
      return ($percent, $kb);
    }

  }
}

sub snapsched_volume($$;@) {

  ## @name ::snapsched_volume
  ## @usage $snapCmds->snapsched_volume($filer, $volume, @settings)
  ## @synopsis
  ##   Gets or sets the snap schedule for the specified volume.
  ##   If given settings specify the snap schedul for the volume.
  ##   Return success or failure with an error message in $@ on failure.
  ##   Success is either 1 or a list of the current settings if no @settings
  ##   are given.
  ##   Failure sets $@.

  my $self = shift;
  my ($filer, $volume, @settings) = @_;

  my $vol_aggr = $self->{VOL_AGGR};
  my $priv_set_test = "";
  if ($self->{TEST_MODE}) {
    $priv_set_test = "priv set -q test; ";
  }

  return eval {

    if (defined $settings[0]) {
      # Set the snap schedule
      my $text = sayEOF($filer, $priv_set_test . "snap sched $vol_aggr $volume @settings");
      croak "$text\n" if ($text);
      return 1;

    } else {
      # Determine the snap schedule
      my $text = sayEOF($filer, "snap sched $vol_aggr $volume");
      croak "$text\n"
	unless ($text =~ s/Volume $volume: //);
      return (split(' ', $text));
    }

  }
}

sub snapsched_save($$) {

  ## @name ::snapsched_save
  ## @usage $snapCmds->snapsched_save($filer, $volume)
  ## @synopsis
  ##   Saves the current snapsched settings for the volume.

  my $self = shift;
  my ($fc, $filer_vol) = @_;

  my ($ss_wly, $ss_dly, $ss_hly) = $self->snapsched_volume($fc, $filer_vol);
  my $snapSched = "$ss_wly $ss_dly $ss_hly";
  logTrc("Saved Snap Schedule: $snapSched");
  $self->{'snapSchedSave_'.$filer_vol} = $snapSched;

  return $snapSched;

}

sub snapsched_restore($$) {

  ## @name ::snapsched_restore
  ## @usage $snapCmds->snapsched_volume($filer, $volume)
  ## @synopsis
  ##   Restores the snapsched for the given volume
  ##   As saved by the prior call snapsched_save().

  my $self = shift;
  my ($fc, $filer_vol) = @_;

  my $snapSched = $self->{'snapSchedSave_'.$filer_vol};
  if ($snapSched) {
    logTrc("Restoring the original snap sched: ".$snapSched);
    # Turn off snapshot schedules
    my @snapSched = split(/\s+/, $snapSched);
    $self->snapsched_volume($fc, $filer_vol, @snapSched);
  } else {
    croak "snapsched_save() was not called previously!. " .
      "We don't have a saved snap-schedule";
  }

}

sub snapsched_review () {

  ## @name ::snapsched_review
  ## @usage $snapCmds->snapsched_review($filer, $volume, %opts)
  ## @synopsis
  ##     Call on this method multiple times to validate
  ##     that autodelete "snap sched" is occuring properly.
  ##     The first call to this method captures the current
  ##     snapshot list for the aggr/volume.
  ##     Subsequent calls will compare the prior snapshot list
  ##     against the current list.
  ## @notes
  ##     It is possible for some or all snapshots to be deleted
  ##     if there is not enough room to accomodate the next
  ##     snapshot. If some snapshots are deleted, we have a
  ##     shorter list of snapshots. If all snapshots are deleted,
  ##     the filer creates a file called 'autodelete_base'
  ##     to preserve state after autodeletion of all snapshots.

  my $self = shift;
  my ($filer, $volume, %opts) = @_;

  logTrc ("snapsched_review (): Started");

  my (@sl_curr, @sl_prev, $firstTime);

  if ($self->{firstTime} == 1 || (exists $opts{first_time})) {

    @sl_curr = $self->snapshot_list_wtime($filer, $volume);
    @{$self->{sl_curr}} = @sl_curr;
    $self->{firstTime} = 0;
    # Just return if the first time called.
    logTrc ("snapsched_review(): First time called , returning Success.");

    $self->{currTime}  = time();
    $self->{priorTime} = 0;
    undef @{$self->{sl_prev}};
    undef @{$self->{sl_curr}};

    return 1;

  } else {

    $self->{currTime} = time();
    if ($self->{priorTime} == 0) {
      $self->{priorTime} = $self->{currTime};
    }

    logTrc("PriorTime: ".$self->{priorTime});
    logTrc(" CurrTime: ".$self->{currTime});
    logTrc("CurrTime - PriorTime: ".($self->{currTime} - $self->{priorTime}));

    if (($self->{currTime} - $self->{priorTime}) < 60) {
      # Return with Success if time is less than 1min
      # Between calls.
      logTrc ("snapsched_review(): Less than 1min. between calls, returning Success.");
      return 1;
    }

    $self->{priorTime} = $self->{currTime};

    @sl_curr = $self->snapshot_list_wtime($filer, $volume);
    if (defined(@{$self->{sl_prev}})) {
      @sl_prev = @{$self->{sl_prev}};
    }
    @{$self->{sl_curr}} = @sl_curr;

    if ($#sl_prev < 0) {
      logTrc("snapsched_review(): No data in sl_prev. Returning.");
      @{$self->{sl_prev}} = @{$self->{sl_curr}};
      return 1;
    }

    # Compare currSnapList against prevSnapList
    # - What indicates an error?
    #   - ss times NOT changed by atleast 1 min.
    #   - Snapshot space USED should stay less than AVAIL

    # Place SnapLists into Assoc Arrays
    my (%asl_prev, %asl_curr, $slf);
    %asl_prev = undef;
    %asl_curr = undef;
    foreach $slf (@sl_prev) {
      my($slTime, $slFile) = split('\|', $slf);
      logTrc("snapsched_review(): \$asl_prev{$slFile} = $slTime");
      $asl_prev{$slFile} = $slTime;
    }
    foreach $slf (@sl_curr) {
      my($slTime, $slFile) = split('\|', $slf);
      logTrc("snapsched_review(): \$asl_curr{$slFile} = $slTime");
      $asl_curr{$slFile} = $slTime;
    }

    # Check Current SL file against Prior SL file
    my $slFile;
    foreach $slFile (sort(keys(%asl_curr))) {
      logTrc("\$slFile: |".$slFile."|");
      next if ($slFile eq "");
      # It is possible that the lists don't match exactly
      # as space requirements for snapshots depend on the
      # amount of file activity on the volumes of the aggr.
      # It is possible that "snap list" returns nothing as it may find
      # that it needs to delete all snapshots to accomodate the next
      # snapshot, and it has not yet completed that snapshot.
      if (exists($asl_prev{$slFile})) {
	logTrc("\$asl_prev{$slFile}:".$asl_prev{$slFile});
	my $prevTime = timeVal($asl_prev{$slFile});
	logTrc("\$asl_curr{$slFile}:".$asl_curr{$slFile});
	my $currTime = timeVal($asl_curr{$slFile});
	logTrc ("snapsched_review (): File:($slFile)  PrevTime:($prevTime)   CurrTime:($currTime)");
	if ($currTime <= $prevTime) {
	  # This is a definate error
	  croak "Looks like Snapshot autodeletion failed.";
	}
      }
    }

  }

  logTrc ("snapsched_review (): Returning.");

  @{$self->{sl_prev}} = @{$self->{sl_curr}};

  return 1;

}

sub snapsched_review_test () {

  ## @name ::snapsched_review_test
  ## @usage $self->snapsched_review_test($FilerConn, $FILER_VOLUME);
  ## @synopsis
  ##     This function will test the snapsched_review() method.

  my $self = shift;
  my ($fc, $aggr) = @_;

  my $i;
  for ($i=0; $i<30; $i++) {
    $self->{VOL_AGGR} = "-A";
    $self->{TEST_MODE} = 1;
    $self->snapsched_review($fc, $aggr);
    if ($i == 5) {
    }
    sleep 15;
  }
  exit;

}

sub timeVal ($) {

  ## @name ::timeVal (not a  method, call as subroutine)
  ## @usage timeVal("<m:1-12> <mday:1-31> <hh>:<mm>"
  ## @synopsis
  ##     This is a utility function that converts a time string
  ##     from "snap list" to a time() value.

  my ($timeStr) = @_;

  logTrc("timeStr: $timeStr");
  $timeStr =~ m/(\d+) +(\d+) +(\d+)\:(\d+)/;
  my $sec   = 0;
  my $min   = $4;
  my $hours = $3;
  my $mday  = $2;
  my $month = $1;
  my $year  = `date +%Y`;

  logTrc("timeVal() - Sec:$sec, Min:$min, Hrs:$hours, Day:$mday, Mnth:$month, Year:$year");
  my $timeVal = timelocal($sec,$min,$hours,$mday,$month-1,$year);

  return $timeVal;

}

sub snapshot_reserve () {

  ## @name ::snapshot_reserve
  ## @usage $self->snapshot_reserve($FilerConn, [$FILER_VOL, [$SNAP_RESRV_PCNT]]);
  ## @synopsis
  ##     Modify the "snap reserve" percentage for a given volume.
  ##     Returns the new snapReservePercent for the volume modified.
  ##     Croaks on failure of 'snap reserve'.
  ##
  ##     If $FILER_VOL is not given, just returns an assoc array containing
  ##     the snapReservePercent for all Volumes of the same type as that of
  ##     this objects instantiation (See: new('[Aggr|Vol]').
  ##
  ##     If SNAP_RESRV_PCNT is not given, just returns an assoc array containing
  ##     the snapReservePercent for $FILER_VOL.
  ##
  ## Return Values:
  ##
  ## $x = $snapCmds->snapshot_reserve($fc);
  ## $volType     = $x->{'[volName]'}->{'type'}    ... [Aggregate|Volume] Volume is {flex|trad}
  ## $volSnapRsrv = $x->{'[volName]'}->{'reserve'} ... [Snap Reserve Percent]
  ## $volSnapRsrv = $x->{'[volName]'}->{'k-bytes'} ... [Snap Reserve in k-bytes]
  ##
  ## @example
  ##
  ##    $SnapCmds = snapCmds->new('Aggr');
  ##    my $x = $SnapCmds->snapshot_reserve($Fc);
  ##    my ($k1, $k2);
  ##    foreach $k1 (keys %$x) {
  ##       logTrc("Volume: ".$k1);
  ##       foreach $k2 (keys %{$x->{$k1}}) {
  ##          logTrc("      $k2 -> ".$x->{$k1}->{$k2});
  ##       }
  ##    }
  ##


  my $self = shift;
  my ($fc, $vol, $snap_rsrv) = @_;
  my $vol_aggr = $self->{VOL_AGGR};
  my $sr;

  if (!$vol and !$snap_rsrv) {
    my $text = sayEOF($fc, "snap reserve $vol_aggr");
    if ($text) {
      $sr = parse_snaprsrv ($text);
    }
  } elsif ($vol and !$snap_rsrv) {
    my $text = sayEOF($fc, "snap reserve $vol_aggr $vol");
    if ($text) {
      $sr = parse_snaprsrv ($text);
    }
  } elsif ($vol and $snap_rsrv) {
    my $text = sayEOF($fc, "snap reserve $vol_aggr $vol $snap_rsrv");
    croak "$text" if ($text);
    # Get the new reserve values for the volume that was changed.
    $sr = $self->snapshot_reserve($fc, $vol);
  } else {
    croak "Invalid Parameters to $self->snapshot_reserve($fc, $vol, $snap_rsrv)";
  }

  return $sr;

}

sub parse_snaprsrv ($%) {

  ## @name ::parse_snaprsrv (not a  method, call as subroutine)
  ## @usage parse_snaprsrv(<textReturnedFromSnapReserveCommand>)
  ## @synopsis
  ##     This is a utility function parses command output from
  ##     'snap reserve' into assoc array structure.

  my ($text) = @_;
  my @lines = split('\n', $text);
  my ($ln, $volName, $volType, $ssRsrv, $kbytes, $sr);
  foreach $ln (@lines) {
    next if ($ln eq "");
    if ($ln =~ m/(\w+)\s+(\w+)\:\s+current snapshot reserve is\s+(\d+)\%\s+or\s+(\d+)\s+k-bytes/) {
      $volType = $1;
      $volName = $2;
      $ssRsrv  = $3;
      $kbytes  = $4;
      $sr->{$volName}->{'type'}    = $volType;
      $sr->{$volName}->{'reserve'} = $ssRsrv;
      $sr->{$volName}->{'k-bytes'} = $kbytes;
    }
  }

  return $sr;
}

1;
