# -*- perl -*-

## @summary Library for managing commonly-used iron parameters
## @status public
## @pod here

=head1 NAME

FS::commonIronParams.pm - provide support for managing commonly-used iron parameters

=head1 SYNOPSIS

  use Tharn;

  BEGIN {
     # set param values here, before loading FS::commonIronParams
     param('EXAMPLE_PARAM', -default, 'example_value');
  }

  # load the module
  use FS::commonIronParams;

  # this declares the common parameters in the 'main' package, and sets default
  # values for any of them which are undefined.
  my $cip = FS::commonIronParams->new();

=head1 DESCRIPTION

commonIronParams - this module provides support for managing commonly-used iron parameters.

=head1 STATUS

  Shared

=head1 AUTHOR

  pavans@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/FS/commonIronParams.pm#2 $
  Last revised $Date: 2013/01/31 $

=head1 COPYRIGHT

  Copyright 2013, NetApp Inc., All Rights Reserved.

=cut

package main;

use strict;
use Data::Dumper;

# declare PERL vars in 'main' - please keep the list in alphabetical order
use vars qw(
	    $IRON_CONTAINER_NAME
	    $IRON_CONTAINER_SIZE
	    $IRON_CONTAINER_RAID_TYPE
	    $IRON_CONTAINER_OPTIONS
	    $IRON_CONTAINER_SEGMENT_CLEANING
	    $IRON_CONTAINER_NUMOF_SNAPS
	    $IRON_CONTAINER_TYPE
	    $IRON_CONTAINER_LARGE_AGGR
	    $IRON_CONTAINER_STATUS_BEFORE_IRONING
	    $IRON_CONTAINER_SNAP_RESTORE
	    $IRON_CONTAINER_VSERVER
	    $IRON_MAXVVBN_DISABLE
	    $IRON_MDLC_DISABLE
	    $IRON_LDC_ENABLE
	    $IRON_SNAPIRON_ENABLE

	    $IRON_FILERSIO_RUNTIME
	    $IRON_NUMOF_VVOL_SETS
	    $IRON_NUMOF_VVOLS_LIST
	    $IRON_NUMOF_FILES_LIST
	    $IRON_NUMOF_HOLEYFILES_LIST
	    $IRON_NUMOF_SPACERESFILES_LIST
	    $IRON_NUMOF_CIFSFILES_LIST
	    $IRON_SNAP_NAME_PREFIX

	    $IRON_VVOL_NAME
	    $IRON_VVOL_NAME_PREFIX
	    $IRON_VVOL_SIZE_LIST
	    $IRON_VVOL_A_SIS_LIST
	    $IRON_VVOL_COMPRESSION_LIST
	    $IRON_VVOL_UNIXLUN_LIST
	    $IRON_VVOL_UNIXLUN_SIZE_LIST
	    $IRON_VVOL_TYPE_LIST
	    $IRON_VVOL_SPACE_GUARANTEE_LIST
	    $IRON_VVOL_PFSVOL_LIST
	    $IRON_VVOL_CLONES_LIST
	    $IRON_VVOL_CLONES_CLONE_LIST
	    $IRON_VVOL_SIS_CLONES_LIST
	    $IRON_VVOL_ATTRIB_LIST
	    $IRON_VVOL_VSM_LIST
	    $IRON_VVOL_QUOTA_LIST
	    $IRON_VVOL_OPTIONS_LIST
	    $IRON_VVOL_RESIZE_LIST
	    $IRON_VVOL_AACACHE_REVALIDATION
	    $IRON_VVOL_NUMOF_SNAPS_LIST
	    $IRON_SNAP_RESTORE_LIST
	    $IRON_SNAP_CREATE_PREIRON
	    $IRON_SNAP_DELETE_PREIRON
	    $IRON_SNAP_CREATE_POSTIRON
	    $IRON_SNAP_DELETE_POSTIRON
	    $IRON_SNAP_CREATE
	    $IRON_SNAP_DELETE
	    $IRON_VVOL_CREATION
	    $IRON_VVOL_DELETION
	    $IRON_VVOL_STOP
	    $IRON_VVOL_PCT_USAGE
	    $IRON_VVOL_TESTFILE_REUSE
	    $IRON_VVOL_VSERVER_LIST
	    $IRON_USE_VBN_ZERO

	    $IRON_FILE_NAME_PREFIX
	    $IRON_FILE_CREATE_METHOD_LIST
	    $IRON_FILE_CREATE_TIMEOUT
	    $IRON_FILE_RESERV_LIST
	    $IRON_FILE_SNAP_RESTORE_LIST

	    $IRON_IOD_ENABLED
	    $IRON_DAC_ENABLED
	    $IRON_COD_ENABLED
	    $IRON_DSC_ENABLED
	    $IRON_OPTIONAL_COMMIT
	    $IRON_SKIP_REVIEW_CMD
	    $IRON_COMMIT_CHANGES
	    $IRON_ON_ROOT
	    $IRON_TIMEOUT
	    $IRON_WAIT
	    $IRON_RECOVERY_RUN
	    $IRON_RECOVERY_WAIT
	    $IRON_IGNORE_START_MSG
        $IRON_PREDICTABILITY_TESTING

	    $IRON_SCANNER_SUSPEND_DURATION
	    $IRON_SCANNER_SPEED
	    $IRON_FORCEFUL
	    $IRON_DSC_SCANNER_SUSPEND_DURATION
	    $IRON_DSC_SCANNER_ABORT_LIST
	    $IRON_DSC_CONTAINER_ABORT
	    $IRON_DSC_VVOL_ABORT

	    $IRON_LOCAL_STATUS_FILES
	    $IRON_SECOND_RUN_RECOVERY_TOOL
	    $IRON_PANIC_ON_CORRUPTION
	    $IRON_RECOVERY_PANIC_ON_CORRUPTION
	    $IRON_RECOVERY_CLONE_SPLIT
	    $IRON_RECOVERY_CLONE_SPLIT_ALL
	    $IRON_RECOVERY_SNAPSHOT_DELETE
	    $IRON_RECOVERY_SNAPSHOT_DELETE_ALL
	    $IRON_RECOVERY_LOSTVOL_DELETE
	    $IRON_RECOVERY_IGNORE_SNAPSHOT_ERROR

	    $IRON_START_FILE_FOLDING_LIST
	    $IRON_START_KIREETI_LIST
	    $IRON_START_DESWIZZLER_LIST
	    $IRON_START_SIS_SCANNER_LIST
	    $IRON_SIS_INCLUDE_TRAPPED_BLOCKS_LIST
	    $IRON_START_SISUNDO_SCANNER_LIST
	    $IRON_START_COMPRESSION_SCANNER_LIST
	    $IRON_START_DECOMPRESSION_SCANNER_LIST

	    $IRON_VVOL_CLONE_SPLIT
	    $IRON_VVOL_CLONE_CREATE
	    $IRON_NEW_CLONE_TESTFILES
	    $IRON_POSTMOUNT_HAMMER
	    $IRON_USE_FILER_IO_TOOL
	    $IRON_CLIENT_IO_TOOL
	    $IRON_NFS_IO_LOAD
	    $IRON_CIFS_IO_LOAD
	    $IRON_NFS_HAMMER
	    $IRON_CIFS_HAMMER
	    $IRON_FILE_ACCESS_THRU_SNAPSHOTS
	    $IRON_VVOL_NEW_SPACE_GUARANTEE_LIST
	    $IRON_NEW_MAXFILES_LIST
	    $IRON_CONTAINER_STOP
	    $IRON_PVOL_AACACHE_REVALID
	    $IRON_BUF_CACHE_REVALID
	    $IRON_BUF_CACHE_REVALID_SLEEPTIME
	    $IRON_POLL_WAFLIRON_STATUS
	    $IRON_POLL_WAFLIRON_SLEEPTIME
	    $IRON_START_KIREETI_SCAN
	    $IRON_FORCE_CP
	    $IRON_CLIENT_PRIORITY_ENABLED
	    $IRON_CLIENT_LOAD_ROOT_VOL
	    $IRON_RUN_SIO_ALL_FILES
	    $IRON_ACCESS_STATUS_FILES
	    $IRON_ONLINE_CORRUPTED_VOL
	    $IRON_PSEUDOBAD
	    $IRON_FAKE_RUN
	   );

package FS::commonIronParams;

use strict;
use warnings;

use Carp qw(croak);

use Tharn;

use Params::Validate qw(validate SCALAR);

sub new {
    my $class = shift;
    $class = ref($class) || $class;

    my $self = bless {}, $class;
    $self->_set_defaults();

    #
    $self->{IRON_CONTAINER_PARAMS} = {};
    $self->{IRON_VVOL_PARAMS} = {};
    $self->{IRON_FLAGS_AND_METHODS} = {};
    $self->{IRON_OPERATIONS} = {};
    $self->_set_iron_vars();

    return $self;
}

# this method sets default values
sub _set_defaults {
    my $self = shift;

    # set params in package 'main', so the param refs below can be resolved properly
    package main;

    param('IRON_CONTAINER_NAME',                  '-default', 'itestaggr');
    param('IRON_CONTAINER_SIZE',                  '-default', '5');
    param('IRON_CONTAINER_RAID_TYPE',             '-default', 'raid_dp');
    param('IRON_CONTAINER_OPTIONS',               '-default', 'nosnap'); # comma separated
    param('IRON_CONTAINER_SEGMENT_CLEANING',      '-default', '1');
    param('IRON_CONTAINER_NUMOF_SNAPS',           '-default', '2');
    param('IRON_CONTAINER_TYPE',                  '-default', 'aggr');
    param('IRON_CONTAINER_VSERVER',               '-default', 'vs0'); # Assign aggr to vserver
    param('IRON_CONTAINER_LARGE_AGGR',            '-default', '1'); # either 1 or 0
    param('IRON_CONTAINER_STATUS_BEFORE_IRONING', '-default', undef);
    param('IRON_CONTAINER_SNAP_RESTORE',          '-default', undef);
    param('SETUP_DONE',                           '-default', '1'); # Always try to re-use
    param('CLEANUP',                              '-default', 'volume'); # vol cleanup
    param('IRON_MAXVVBN_DISABLE',                 '-default', '0');
    param('IRON_MDLC_DISABLE',                    '-default', '0'); # MDLC feature
    param('IRON_LDC_ENABLE',                      '-default', '1'); # LDC feature
    param('IRON_SNAPIRON_ENABLE',                 '-default', '1'); # SNAPIRON feature

    param('IRON_NUMOF_FILES_LIST',                '-default', '8 8 8');
    param('IRON_NUMOF_HOLEYFILES_LIST',           '-default', '0 0 0');
    param('IRON_NUMOF_SPACERESFILES_LIST',        '-default', '0 0 0');
    param('IRON_NUMOF_CIFSFILES_LIST',            '-default', '0 0 0');
    param('IRON_SNAP_NAME_PREFIX',                '-default', 'snap');
    param('IRON_FILERSIO_RUNTIME',                '-default', '1800');
    param('IRON_SNAP_CREATE',                     '-default', '1'); # Create snapshot while iron
    param('IRON_SNAP_DELETE',                     '-default', '1'); # Delete created snapshot while iron
    param('IRON_SNAP_RESTORE_LIST',               '-default', '0 0 0'); # Perform snap restore

    param('IRON_SNAP_CREATE_PREIRON',             '-default', '1'); # Create snapshot pre iron
    param('IRON_SNAP_DELETE_PREIRON',             '-default', '1'); # Delete created snapshot pre iron
    param('IRON_SNAP_CREATE_POSTIRON',            '-default', '1'); # Create snapshot post iron
    param('IRON_SNAP_DELETE_POSTIRON',            '-default', '1'); # Delete created snapshot post iron

    param('IRON_NUMOF_VVOL_SETS',                 '-default', undef);
    param('IRON_NUMOF_VVOLS_LIST',                '-default', '1 0 0'); # Number of VVOLs in each set
    param('IRON_VVOL_A_SIS_LIST',                 '-default', '0 0 0'); # 1 to enable
    param('IRON_VVOL_COMPRESSION_LIST',           '-default', '0 0 0'); # 1 to enable
    param('IRON_VVOL_UNIXLUN_LIST',               '-default', '0 0 0'); # 1 to enable
    param('IRON_VVOL_UNIXLUN_SIZE_LIST',          '-default', '5g 5g 5g'); # 5g is default size
    param('IRON_VVOL_NAME',                       '-default', undef); # vol name to use existing vol
    param('IRON_VVOL_NAME_PREFIX',                '-default', 'itestvol');
    param('IRON_VVOL_SIZE_LIST',                  '-default', '80g 50g 60g');
    param('IRON_VVOL_TYPE_LIST',                  '-default', 'flex flex flex');
    param('IRON_VVOL_VSERVER_LIST',               '-default', 'vs0 vs0 vs0');
    param('IRON_VVOL_SPACE_GUARANTEE_LIST',       '-default', 'volume file volume');
    param('IRON_VVOL_NUMOF_SNAPS_LIST',           '-default', '2 1 2'); # Number of snapshots
    param('IRON_VVOL_PFSVOL_LIST',                '-default', '0 0 0'); # Number of pfs vols
    param('IRON_VVOL_CLONES_LIST',                '-default', '0 0 0'); # Number of clones
    param('IRON_VVOL_CLONES_CLONE_LIST',          '-default', '0 0 0'); # Number of clones to each clone vol
    param('IRON_VVOL_SIS_CLONES_LIST',            '-default', '0 0 0'); # Number of SIS clones
    param('IRON_VVOL_ATTRIB_LIST',                '-default', 'regular regular regular'); # regular/compression/a_sis
    param('IRON_VVOL_VSM_LIST',                   '-default', '0 0 0'); # Number of volume snapmirror to flex vol
    param('IRON_VVOL_QUOTA_LIST',                 '-default', '0 0 0');
    # 557689,523371 - with nvfail we might see this burt for corruption cases where recovery run will find corruption
    #param('IRON_VVOL_OPTIONS_LIST',               '-default', 'nosnap,nvfail,no_atime_update nosnap nosnap');
    param('IRON_VVOL_OPTIONS_LIST',               '-default', 'nosnap,no_atime_update nosnap nosnap');
    param('IRON_VVOL_RESIZE_LIST',                '-default', '+5g -10g +30g');
    param('IRON_VVOL_AACACHE_REVALIDATION',       '-default', '0'); # TODO
    param('IRON_VVOL_CREATION',                   '-default', '0');
    param('IRON_VVOL_DELETION',                   '-default', '0');
    param('IRON_VVOL_STOP',                       '-default', '0');
    param('IRON_VVOL_PCT_USAGE',                  '-default', '20');
    param('IRON_VVOL_TESTFILE_REUSE',             '-default', '1'); # reuse setup files
    param('IRON_NEW_CLONE_TESTFILES',             '-default', '1'); # Create new test files for clone snap shots

    param('IRON_USE_VBN_ZERO',                    '-default', '0'); # Works with mkfile only
    param('IRON_FILE_NAME_PREFIX',                '-default', 'file');
    param('IRON_FILE_CREATE_METHOD_LIST',         '-default', 'mkfile mkfile mkfile'); # filersio/mkfile
    param('IRON_FILE_RESERV_LIST',                '-default', '0 0 0');
    param('IRON_FILE_SNAP_RESTORE_LIST',          '-default', '0 0 0');
    param('IRON_FILE_CREATE_TIMEOUT',             '-default', '6000'); # filersio/mkfile timeout

    param('IRON_DSC_ENABLED',                     '-default', '1');
    param('IRON_IOD_ENABLED',                     '-default', undef);
    param('IRON_DAC_ENABLED',                     '-default', undef);
    param('IRON_COD_ENABLED',                     '-default', undef);
    param('IRON_WAIT',                            '-default', undef);
    param('IRON_PREDICTABILITY_TESTING',          '-default', '1');
    param('IRON_IGNORE_START_MSG',                '-default', '0');
    param('IRON_OPTIONAL_COMMIT',                 '-default', '0');
    param('IRON_SKIP_REVIEW_CMD',                 '-default', '0');
    param('IRON_COMMIT_CHANGES',                  '-default', '1'); # default commit changes, 0 - reject
    param('IRON_ON_ROOT',                         '-default', '0');
    param('IRON_TIMEOUT',                         '-default', '3600'); # wafliron timeout
    param('IRON_RECOVERY_RUN',                    '-default', '1');
    # Wait time before we start recovery iron/tool
    param('IRON_RECOVERY_WAIT',                   '-default', '8000');

    param('IRON_SCANNER_SUSPEND_DURATION',        '-default', '0');
    param('IRON_SCANNER_SPEED',                   '-default', '10000');
    param('IRON_FORCEFUL',                        '-default', undef);
    param('IRON_DSC_SCANNER_SUSPEND_DURATION',    '-default', undef);
    param('IRON_DSC_SCANNER_ABORT_LIST',          '-default', 'undef undef undef');
    param('IRON_DSC_CONTAINER_ABORT',             '-default', undef);
    param('IRON_DSC_VVOL_ABORT',                  '-default', undef);

    param('IRON_LOCAL_STATUS_FILES',              '-default', '0');
    param('IRON_SECOND_RUN_RECOVERY_TOOL',        '-default', 'iron');
    param('IRON_PANIC_ON_CORRUPTION',             '-default', '0');
    # To avoid filer panic, if there is a corruption on snapmap.
    # In RR we are not deleting corrupted snapshots by default.
    param('IRON_RECOVERY_PANIC_ON_CORRUPTION',    '-default', '0');
    param('IRON_RECOVERY_CLONE_SPLIT',            '-default', '0');
    param('IRON_RECOVERY_CLONE_SPLIT_ALL',        '-default', '0');
    param('IRON_RECOVERY_SNAPSHOT_DELETE',        '-default', '0');
    param('IRON_RECOVERY_SNAPSHOT_DELETE_ALL',    '-default', '0');
    param('IRON_RECOVERY_LOSTVOL_DELETE',         '-default', '0');
    param('IRON_RECOVERY_IGNORE_SNAPSHOT_ERROR',  '-default', '1');

    param('IRON_START_FILE_FOLDING_LIST',          '-default', '0 0 0'); # TODO
    param('IRON_START_KIREETI_LIST',               '-default', '0 0 0'); # TODO
    param('IRON_START_DESWIZZLER_LIST',            '-default', '0 0 0'); # TODO
    param('IRON_START_SIS_SCANNER_LIST',           '-default', '0 0 0');
    param('IRON_START_SISUNDO_SCANNER_LIST',       '-default', '0 0 0');
    param('IRON_START_COMPRESSION_SCANNER_LIST',   '-default', '0 0 0');
    param('IRON_START_DECOMPRESSION_SCANNER_LIST', '-default', '0 0 0');
    param('IRON_SIS_INCLUDE_TRAPPED_BLOCKS_LIST',  '-default', '0 0 0');

    param('IRON_VVOL_CLONE_SPLIT',                '-default', '0');
    param('IRON_VVOL_CLONE_CREATE',               '-default', '0');
    param('IRON_POSTMOUNT_HAMMER',                '-default', '1');
    param('IRON_USE_FILER_IO_TOOL',               '-default', '1'); # use filer internal io tool
    param('IRON_CLIENT_IO_TOOL',                  '-default', 'hammer'); # hammer/filersio
    param('IRON_NFS_IO_LOAD',                     '-default', '1');
    param('IRON_CIFS_IO_LOAD',                    '-default', '0');
    param('IRON_NFS_HAMMER',                      '-default', '1');
    param('IRON_CIFS_HAMMER',                     '-default', '0');
    param('IRON_FILE_ACCESS_THRU_SNAPSHOTS',      '-default', '0');
    param('IRON_VVOL_NEW_SPACE_GUARANTEE_LIST',   '-default', 'file file file');
    param('IRON_NEW_MAXFILES_LIST',               '-default', undef);
    param('IRON_CONTAINER_STOP',                  '-default', '0');
    param('IRON_PVOL_AACACHE_REVALID',            '-default', '0');
    param('IRON_BUF_CACHE_REVALID',               '-default', '1'); # flush cache
    param('IRON_BUF_CACHE_REVALID_SLEEPTIME',     '-default', '60'); # flush cache every 60sec
    param('IRON_POLL_WAFLIRON_STATUS',            '-default', '1');
    param('IRON_POLL_WAFLIRON_SLEEPTIME',         '-default', '30'); # poll every 30sec
    param('IRON_START_KIREETI_SCAN',              '-default', '1'); # try start every 60sec
    param('IRON_FORCE_CP',                        '-default', '1');
    param('IRON_CLIENT_PRIORITY_ENABLED',         '-default', '0');
    param('IRON_CLIENT_LOAD_ROOT_VOL',            '-default', '0');
    param('IRON_RUN_SIO_ALL_FILES',               '-default', '1');
    param('IRON_ACCESS_STATUS_FILES',             '-default', '1');
    param('IRON_ONLINE_CORRUPTED_VOL',            '-default', '0'); # with MDLC
    param('IRON_PSEUDOBAD',                       '-default', '0');
    param('IRON_FAKE_RUN',                        '-default', '0'); # Don't run iron

}

sub _set_iron_vars {
    my $self = shift;
    my ($key, $value);

    # set params in package 'main', so the param refs below can be resolved properly
    package main;

    my ( %iron_container_params, %iron_vvol_params,
         %iron_flags_and_methods, %iron_operations, $tmp );

    # CONTAINER opts
    $iron_container_params{IRON_CONTAINER_NAME} = $IRON_CONTAINER_NAME;
    $iron_container_params{IRON_CONTAINER_SIZE} = $IRON_CONTAINER_SIZE;
    $iron_container_params{IRON_CONTAINER_TYPE} = $IRON_CONTAINER_TYPE;
    $iron_container_params{IRON_CONTAINER_VSERVER}     = $IRON_CONTAINER_VSERVER;
    $iron_container_params{IRON_CONTAINER_RAID_TYPE}   = $IRON_CONTAINER_RAID_TYPE;
    $iron_container_params{IRON_CONTAINER_NUMOF_SNAPS} = $IRON_CONTAINER_NUMOF_SNAPS;
    $iron_container_params{IRON_CONTAINER_LARGE_AGGR}  = $IRON_CONTAINER_LARGE_AGGR;
    $iron_container_params{IRON_CONTAINER_OPTIONS}     = $IRON_CONTAINER_OPTIONS;
    $iron_container_params{IRON_CONTAINER_SEGMENT_CLEANING} = $IRON_CONTAINER_SEGMENT_CLEANING;
    $iron_container_params{IRON_CONTAINER_FILE_NAME_PREFIX} = $IRON_FILE_NAME_PREFIX;
    $iron_container_params{IRON_CONTAINER_SNAP_NAME_PREFIX} = $IRON_SNAP_NAME_PREFIX;
    $iron_container_params{IRON_USE_VBN_ZERO} = $IRON_USE_VBN_ZERO;

    $iron_container_params{IRON_CONTAINER_SNAP_RESTORE}         = $IRON_CONTAINER_SNAP_RESTORE;
    $iron_container_params{IRON_CONTAINER_SNAP_CREATE_PREIRON}  = $IRON_SNAP_CREATE_PREIRON;
    $iron_container_params{IRON_CONTAINER_SNAP_DELETE_PREIRON}  = $IRON_SNAP_DELETE_PREIRON;
    $iron_container_params{IRON_CONTAINER_SNAP_CREATE_POSTIRON} = $IRON_SNAP_CREATE_POSTIRON;
    $iron_container_params{IRON_CONTAINER_SNAP_DELETE_POSTIRON} = $IRON_SNAP_CREATE_POSTIRON;

    $iron_container_params{IRON_MAXVVBN_DISABLE} = $IRON_MAXVVBN_DISABLE;
    $iron_container_params{IRON_MDLC_DISABLE}     = $IRON_MDLC_DISABLE;
    $iron_container_params{IRON_LDC_ENABLE}      = $IRON_LDC_ENABLE;
    $iron_container_params{IRON_SNAPIRON_ENABLE} = $IRON_SNAPIRON_ENABLE;

    while ( ($key, $value) = each(%iron_container_params) ) {
        $iron_container_params{$key} = undef
            if ( !defined $iron_container_params{$key}
                || $iron_container_params{$key} =~ /undef/i );
    }
    $self->{IRON_CONTAINER_PARAMS} = \%iron_container_params;


    # VVOL opts
    $iron_vvol_params{IRON_MAXVVBN_DISABLE} = $IRON_MAXVVBN_DISABLE;
    $IRON_NUMOF_VVOLS_LIST = '1 0 0' if defined $IRON_VVOL_NAME;
    my @vlist = split(/\s+/, $IRON_NUMOF_VVOLS_LIST);

    if (defined $IRON_NUMOF_VVOL_SETS && $IRON_NUMOF_VVOL_SETS <= @vlist) {
        $iron_vvol_params{IRON_NUMOF_VVOL_SETS} = $IRON_NUMOF_VVOL_SETS;
    } else {
        my $vlist_no = 0;
        foreach my $i ( @vlist ) {
            $vlist_no++ unless $i eq 0;
        }
        $iron_vvol_params{IRON_NUMOF_VVOL_SETS} = $vlist_no;
    }

    my @tmparray;
    for (my $i=0; $i < $iron_vvol_params{IRON_NUMOF_VVOL_SETS}; $i++) {
        @tmparray = split(/\s+/, $IRON_NUMOF_VVOLS_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NUMOF_VVOLS} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_SIZE_LIST);
        $tmparray[$i] = $tmparray[$i-1] unless defined $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SIZE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_TYPE_LIST);
        $tmparray[$i] = $tmparray[$i-1] unless defined $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_TYPE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_VSERVER_LIST);
        $tmparray[$i] = $tmparray[$i-1] unless defined $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_VSERVER} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_ATTRIB_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_ATTRIB} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_SPACE_GUARANTEE_LIST);
        $tmparray[$i] = $tmparray[$i-1] unless defined $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SPACE_GUARANTEE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_NUMOF_SNAPS_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NUMOF_SNAPS} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_PFSVOL_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_PFSVOL} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_CLONES_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_CLONES} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_CLONES_CLONE_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_CLONES_CLONE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_SIS_CLONES_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SIS_CLONES} = $tmparray[$i];

        @tmparray = split(/\s+/, $IRON_VVOL_VSM_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_VSM} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_QUOTA_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_QUOTA} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_OPTIONS_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_OPTIONS} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_SNAP_RESTORE_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SNAP_RESTORE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_FILE_SNAP_RESTORE_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_FILE_SNAP_RESTORE} = $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_FILE_CREATE_TIMEOUT} = $IRON_FILE_CREATE_TIMEOUT;
        @tmparray = split(/\s+/, $IRON_FILE_RESERV_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_FILE_RESERV} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_NUMOF_FILES_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NUMOF_FILES} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_NUMOF_HOLEYFILES_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_NUMOF_HOLEYFILES} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_NUMOF_SPACERESFILES_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_NUMOF_SPACERESFILES} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_NUMOF_CIFSFILES_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_NUMOF_CIFSFILES} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_NEW_SPACE_GUARANTEE_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NEW_SPACE_GUARANTEE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_A_SIS_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_A_SIS} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_COMPRESSION_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_UNIXLUN_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_UNIXLUN} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_VVOL_UNIXLUN_SIZE_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_UNIXLUN_SIZE} = $tmparray[$i];

        # Compression can only be enabled on a 64-bit volume contained in a 64-bit aggregate
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} = 0
           if $iron_container_params{IRON_CONTAINER_LARGE_AGGR} eq 0;

        @tmparray = split(/\s+/, $IRON_START_COMPRESSION_SCANNER_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_START_COMPRESSION_SCANNER} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_START_DECOMPRESSION_SCANNER_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_START_DECOMPRESSION_SCANNER} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_FILE_CREATE_METHOD_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_FILE_CREATE_METHOD} = $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_FILE_CREATE_METHOD} = $tmparray[$i];

        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NAME}        = $IRON_VVOL_NAME;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NAME_PREFIX} = $IRON_VVOL_NAME_PREFIX . "_SET$i";
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_CONTAINER_NAME} = $IRON_CONTAINER_NAME;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_FILE_NAME_PREFIX} = $IRON_FILE_NAME_PREFIX;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SNAP_NAME_PREFIX} = $IRON_SNAP_NAME_PREFIX;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_PCT_USAGE} = $IRON_VVOL_PCT_USAGE;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_TESTFILE_REUSE} = $IRON_VVOL_TESTFILE_REUSE;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_USE_VBN_ZERO} = $IRON_USE_VBN_ZERO;

        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SNAP_CREATE_PREIRON} = $IRON_SNAP_CREATE_PREIRON;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SNAP_DELETE_PREIRON} = $IRON_SNAP_DELETE_PREIRON;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SNAP_CREATE_POSTIRON}  = $IRON_SNAP_CREATE_POSTIRON;
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_SNAP_DELETE_POSTIRON}  = $IRON_SNAP_DELETE_POSTIRON;

        @tmparray = split(/\s+/, $IRON_FILE_SNAP_RESTORE_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NUMOF_FILE_SNAP_RESTORE} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_START_SIS_SCANNER_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_START_SIS_SCANNER} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_SIS_INCLUDE_TRAPPED_BLOCKS_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_SIS_INCLUDE_TRAPPED_BLOCKS} = $tmparray[$i];
        @tmparray = split(/\s+/, $IRON_START_SISUNDO_SCANNER_LIST);
        $iron_vvol_params{"VVOL_SET$i"}{IRON_START_SISUNDO_SCANNER} = $tmparray[$i];
        $iron_vvol_params{"VVOL_SET$i"}{IRON_NEW_CLONE_TESTFILES} = $IRON_NEW_CLONE_TESTFILES;

        if ( defined $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_ATTRIB}
              && $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_ATTRIB} =~ /^a_sis$/i ) {
            $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_A_SIS} = 1;
        }

        if ( defined $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_ATTRIB}
              && $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_ATTRIB} =~ /^compression$/i ) {
            $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} = 1;
        }

        if ( $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_TYPE} !~ /flex/i ) {
            # SIS Can not be enabled on trad vol
            $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_A_SIS} = 0;
            $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} = 0;
        } else {
            # Enable SIS on compression enable
            $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_A_SIS} = 1
              if $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} eq 1;
            $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} = 0
              if ( ($iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_COMPRESSION} =~ /undef/i)
                  && ($iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_A_SIS} !~ /undef/i) );
        }
    }

    # Change undef string to undef
    for (my $i=0; $i < $iron_vvol_params{IRON_NUMOF_VVOL_SETS}; $i++) {
        $tmp = $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NUMOF_VVOLS};
        $iron_vvol_params{"VVOL_SET$i"}{IRON_VVOL_NUMOF_VVOLS} = undef
             if $tmp =~ /undef/;
    }

    # Assign to return variable
    $self->{IRON_VVOL_PARAMS} = \%iron_vvol_params;


    # IRON flag & methods
    $iron_flags_and_methods{IRON_WAIT} = $IRON_WAIT;
    $iron_flags_and_methods{IRON_PANIC_ON_CORRUPTION}  = $IRON_PANIC_ON_CORRUPTION;
    $iron_flags_and_methods{IRON_IGNORE_START_MSG}     = $IRON_IGNORE_START_MSG;
    $iron_flags_and_methods{IRON_PREDICTABILITY_TESTING} = $IRON_PREDICTABILITY_TESTING;
    $iron_flags_and_methods{IRON_RECOVERY_CLONE_SPLIT}   = $IRON_RECOVERY_CLONE_SPLIT;
    $iron_flags_and_methods{IRON_RECOVERY_CLONE_SPLIT_ALL}        = $IRON_RECOVERY_CLONE_SPLIT_ALL;
    $iron_flags_and_methods{IRON_RECOVERY_PANIC_ON_CORRUPTION}    = $IRON_RECOVERY_PANIC_ON_CORRUPTION;
    $iron_flags_and_methods{IRON_RECOVERY_SNAPSHOT_DELETE}        = $IRON_RECOVERY_SNAPSHOT_DELETE;
    $iron_flags_and_methods{IRON_RECOVERY_SNAPSHOT_DELETE_ALL}    = $IRON_RECOVERY_SNAPSHOT_DELETE_ALL;
    $iron_flags_and_methods{IRON_RECOVERY_LOSTVOL_DELETE}         = $IRON_RECOVERY_LOSTVOL_DELETE;
    $iron_flags_and_methods{IRON_RECOVERY_IGNORE_SNAPSHOT_ERROR}  = $IRON_RECOVERY_IGNORE_SNAPSHOT_ERROR;
    $iron_flags_and_methods{IRON_CONTAINER_STATUS_BEFORE_IRONING} = $IRON_CONTAINER_STATUS_BEFORE_IRONING;
    $iron_flags_and_methods{IRON_IOD_ENABLED} = $IRON_IOD_ENABLED;
    $iron_flags_and_methods{IRON_DSC_ENABLED} = $IRON_DSC_ENABLED;
    $iron_flags_and_methods{IRON_DAC_ENABLED} = $IRON_DAC_ENABLED;
    $iron_flags_and_methods{IRON_COD_ENABLED} = $IRON_COD_ENABLED;
    $iron_flags_and_methods{IRON_TIMEOUT}     = $IRON_TIMEOUT;
    $iron_flags_and_methods{IRON_FORCEFUL}    = $IRON_FORCEFUL;
    $iron_flags_and_methods{IRON_OPTIONAL_COMMIT}    = $IRON_OPTIONAL_COMMIT;
    $iron_flags_and_methods{IRON_SKIP_REVIEW_CMD}    = $IRON_SKIP_REVIEW_CMD;
    $iron_flags_and_methods{IRON_COMMIT_CHANGES}     = $IRON_COMMIT_CHANGES;
    $iron_flags_and_methods{IRON_ON_ROOT}            = $IRON_ON_ROOT;
    $iron_flags_and_methods{IRON_RECOVERY_RUN}       = $IRON_RECOVERY_RUN;
    $iron_flags_and_methods{IRON_RECOVERY_WAIT}      = $IRON_RECOVERY_WAIT;
    $iron_flags_and_methods{IRON_SCANNER_SPEED}      = $IRON_SCANNER_SPEED;
    $iron_flags_and_methods{IRON_CONTAINER_TYPE}     = $IRON_CONTAINER_TYPE;
    $iron_flags_and_methods{IRON_CONTAINER_VSERVER}  = $IRON_CONTAINER_VSERVER;
    $iron_flags_and_methods{IRON_LOCAL_STATUS_FILES} = $IRON_LOCAL_STATUS_FILES;
    $iron_flags_and_methods{IRON_SCANNER_SUSPEND_DURATION}     = $IRON_SCANNER_SUSPEND_DURATION;
    $iron_flags_and_methods{IRON_DSC_SCANNER_SUSPEND_DURATION} = $IRON_DSC_SCANNER_SUSPEND_DURATION;
    $iron_flags_and_methods{IRON_SECOND_RUN_RECOVERY_TOOL}     = $IRON_SECOND_RUN_RECOVERY_TOOL;
    $iron_flags_and_methods{IRON_CLIENT_PRIORITY_ENABLED} = $IRON_CLIENT_PRIORITY_ENABLED;
    $iron_flags_and_methods{IRON_ONLINE_CORRUPTED_VOL}    = $IRON_ONLINE_CORRUPTED_VOL;
    $iron_flags_and_methods{IRON_PSEUDOBAD}               = $IRON_PSEUDOBAD;
    $iron_flags_and_methods{IRON_FAKE_RUN}                = $IRON_FAKE_RUN;

    $iron_flags_and_methods{IRON_MAXVVBN_DISABLE} = $IRON_MAXVVBN_DISABLE;
    $iron_flags_and_methods{IRON_MDLC_DISABLE}    = $IRON_MDLC_DISABLE;
    $iron_flags_and_methods{IRON_LDC_ENABLE}      = $IRON_LDC_ENABLE;
    $iron_flags_and_methods{IRON_SNAPIRON_ENABLE} = $IRON_SNAPIRON_ENABLE;

    $iron_flags_and_methods{IRON_LDC_ENABLE} = 0 unless $IRON_DSC_ENABLED;
    $iron_flags_and_methods{IRON_ONLINE_CORRUPTED_VOL} = 0
                   unless $iron_flags_and_methods{IRON_LDC_ENABLE};

    while ( ($key, $value) = each(%iron_flags_and_methods) ) {
        $iron_flags_and_methods{$key} = undef
            if ( !defined $iron_flags_and_methods{$key}
                || $iron_flags_and_methods{$key} =~ /undef/i );
    }
    $self->{IRON_FLAGS_AND_METHODS} = \%iron_flags_and_methods;

    # IRON operations
    $iron_operations{IRON_FORCE_CP}          = $IRON_FORCE_CP;
    $iron_operations{IRON_VVOL_STOP}         = $IRON_FORCE_CP;
    $iron_operations{IRON_NFS_HAMMER}        = $IRON_NFS_HAMMER;
    $iron_operations{IRON_CIFS_HAMMER}       = $IRON_CIFS_HAMMER;
    $iron_operations{IRON_USE_FILER_IO_TOOL} = $IRON_USE_FILER_IO_TOOL;
    $iron_operations{IRON_CLIENT_IO_TOOL}    = $IRON_CLIENT_IO_TOOL;
    $iron_operations{IRON_NFS_IO_LOAD}       = $IRON_NFS_IO_LOAD;
    $iron_operations{IRON_CIFS_IO_LOAD}      = $IRON_CIFS_IO_LOAD;
    $iron_operations{IRON_POSTMOUNT_HAMMER}  = $IRON_POSTMOUNT_HAMMER;
    $iron_operations{IRON_CONTAINER_STOP}    = $IRON_CONTAINER_STOP;
    $iron_operations{IRON_VVOL_SNAP_CREATE}  = $IRON_SNAP_CREATE;
    $iron_operations{IRON_VVOL_SNAP_DELETE}  = $IRON_SNAP_DELETE;
    $iron_operations{IRON_VVOL_CREATION}     = $IRON_VVOL_CREATION;
    $iron_operations{IRON_VVOL_DELETION}     = $IRON_VVOL_DELETION;
    $iron_operations{IRON_VVOL_CLONE_SPLIT}  = $IRON_VVOL_CLONE_SPLIT;
    $iron_operations{IRON_VVOL_CLONE_CREATE} = $IRON_VVOL_CLONE_CREATE;
    $iron_operations{IRON_NEW_MAXFILES_LIST} = $IRON_NEW_MAXFILES_LIST;
    $iron_operations{IRON_BUF_CACHE_REVALID} = $IRON_BUF_CACHE_REVALID;
    $iron_operations{IRON_POLL_WAFLIRON_STATUS}  = $IRON_POLL_WAFLIRON_STATUS;
    $iron_operations{IRON_START_KIREETI_SCAN}    = $IRON_START_KIREETI_SCAN;
    $iron_operations{IRON_SNAP_CREATE_POSTIRON}  = $IRON_SNAP_CREATE_POSTIRON;
    $iron_operations{IRON_SNAP_DELETE_POSTIRON}  = $IRON_SNAP_DELETE_POSTIRON;
    $iron_operations{IRON_PVOL_AACACHE_REVALID}  = $IRON_PVOL_AACACHE_REVALID;
    $iron_operations{IRON_CONTAINER_SNAP_CREATE} = $IRON_SNAP_CREATE;
    $iron_operations{IRON_CONTAINER_SNAP_DELETE} = $IRON_SNAP_DELETE;
    $iron_operations{IRON_CLIENT_LOAD_ROOT_VOL}  = $IRON_CLIENT_LOAD_ROOT_VOL;
    $iron_operations{IRON_RUN_SIO_ALL_FILES}     = $IRON_RUN_SIO_ALL_FILES;
    $iron_operations{IRON_ACCESS_STATUS_FILES}   = $IRON_ACCESS_STATUS_FILES;
    $iron_operations{IRON_POLL_WAFLIRON_SLEEPTIME}     = $IRON_POLL_WAFLIRON_SLEEPTIME;
    $iron_operations{IRON_CONTAINER_SNAP_RESTORE}      = $IRON_CONTAINER_SNAP_RESTORE;
    $iron_operations{IRON_FILE_ACCESS_THRU_SNAPSHOTS}  = $IRON_FILE_ACCESS_THRU_SNAPSHOTS;
    $iron_operations{IRON_CONTAINER_SEGMENT_CLEANING}  = $IRON_CONTAINER_SEGMENT_CLEANING;
    $iron_operations{IRON_BUF_CACHE_REVALID_SLEEPTIME} = $IRON_BUF_CACHE_REVALID_SLEEPTIME;

    while ( ($key, $value) = each(%iron_operations) ) {
        $iron_operations{$key} = undef
            if ( !defined $iron_operations{$key}
                || $iron_operations{$key} =~ /undef/i );
    }
    $self->{IRON_OPERATIONS} = \%iron_operations;
}

1;
