# -*- perl -*-

## @summary Library for managing commonly-used parameters
## @author mcrae@netapp.com, dl-waflqa@netapp.com
## @status public
## @pod here

=head1 NAME

FS::commonParams.pm - provide support for managing commonly-used parameters

=head1 SYNOPSIS

  use Tharn;

  BEGIN {
     # set param values here, before loading FS::commonParams
     param('EXAMPLE_PARAM', -default, 'example_value');
  }

  # load the module
  use FS::commonParams;

  # this declares the common parameters in the 'main' package, and sets default
  # values for any of them which are undefined.
  my $cp = FS::commonParams->new(set_legacy_defaults => 1);

  # reset the VOL_TYPE parameters for the given volume type. this is useful for
  # multi-testcase scripts, where the vol_type may change over the course of a
  # run.
  my $value_hash = $cp->sync_vol_type_params(
      vol_type => 'trad',
  );

=head1 DESCRIPTION

commonParams - this module provides support for managing commonly-used parameters.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-waflqa-automation@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/FS/commonParams.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=cut

package main;

use strict;

# declare PERL vars in 'main' - please keep the list in alphabetical order
use vars qw(
	    $AGGR_SIZE
	    $ASIS_ENABLE
	    $CIFS_CLIENT
	    $CIFS_DRIVE
	    $CIFS_PASSWD
	    $CIFS_SHARE
	    $CIFS_USER
	    $CIFS_WORKGROUP
	    $CLEANUP
	    $CLIENT
	    $CLIENT_CONN
	    $CMDTIMEOUT
	    $COMPRESSION_ENABLE
	    $CONVERT_TO64
	    $COPY_PREDEF
	    $DF_SYNC_DELAY
	    $EXTERN_SETUP
	    $FILER_MODE
	    $FILER_LIST
	    $FILER
	    $FILERA
	    $FILERB
	    $CREATE_COMMON_RESOURCES
	    $FILER_AGGR
	    $FILER_CONN
	    $FILER_IP
	    $FILER_QTREE
	    $FILER_VOL
	    $FILER_VOL_FLEX
	    $FILER_VOL_TRAD
	    $FILER_VOL_MNT
	    $LARGE_AGGR
	    $MODIFY_VOL
	    $MOUNTOPTS
	    $MOUNTPOINT
            $NBLADE_START
            $NBLADE_HOST
            $NBLADE_DIR
	    $NODES
	    $ONTAPDIR
	    $POPULATE_VOL
	    $PFILER
	    $TIMEOUT
	    $VOL_SIZE
	    $TRAD_SIZE
	    $FLEX_SIZE
	    $VOL_TYPE
	    $RAID_TYPE
	    $MOUNT_VERS
	    $CONTAINER
	    $SETUP_DONE
	    $WORM_VOL
	    $WORM_AGGR
	    $WORM_RETENTION
	    $WAFLIRON_TIMEOUT
	    $SKIP_SNAPLOCK_RETENTION_SETUP
	    $SLEEP_TIMEOUT_MULTIPLIER
	    $SLCC_STRICT
	    $SNAPLOCK_TYPE
	    $SUDO_PATH
	    $FS_PROTO
	    $UPGRADE_AGGR
	    $USE_ZAPI_IF_AVAILABLE
	    $BOOT_TIMEOUT
	    $IRON_ON_ROOT
	    $VSERVER
            $HYBRID_AGGR
            $HYBRID_SSD_SIZE
	    );

package FS::commonParams;

use strict;
use warnings;

use Carp qw(croak);

use Tharn;

use Params::Validate qw(validate SCALAR);

sub new {
    my $class = shift;
    $class = ref($class) || $class;

    # validate args
    my (%args) = validate (
	@_,
	{
	    set_legacy_defaults => {
		type    => SCALAR,
		default => 0,
	    },
	}
    );

    my $self = bless {}, $class;

    # provided for backwards-compatibility with pre-existing test code,
    # see lib/FS/commonParams.pl
    if ($args{set_legacy_defaults}) {
        $self->_set_legacy_defaults();
    }

    return $self;
}

# this method provides support for pre-existing test code, see lib/FS/commonParams.pl
sub _set_legacy_defaults {
    my $self = shift;

    # set params in package 'main', so the param refs below can be resolved properly
    package main;

    param("AGGR_SIZE",      "-default", 5);
    param("ASIS_ENABLE",    "-default", undef);   # 0==>disable, 1==>enable, undef==>no_action
    param("CIFS_CLIENT",      "-default", undef());
    param("CIFS_DRIVE",     "-default", '*');
    param("CIFS_PASSWD", 	"-default", "CIFS_pass-123");
    param("CIFS_SHARE",     "-default", "testshare");
    param("CIFS_USER", 	"-default", "administrator");
    param("CIFS_WORKGROUP",	"-default", "WORKGROUP");
    param("CLEANUP",        "-default",  'test');	 # 'none', 'test', 'volume', 'all'
    param("CLIENT",		"-default", "localhost");
    param("CLIENT_CONN",    "-default", "#rsh");
    param("CMDTIMEOUT",     "-default", 120); # 2 mins
    param("CONVERT_TO64",   "-default", undef); # if true, change aggregate to 64-bit (hybrid) after flexvol setup
    param("COPY_PREDEF",    "-default", 0);
    param("DF_SYNC_DELAY",  "-default", 20);   # Delay to use before 'df' in vol_disk_free(), etc. 
    param("COMPRESSION_ENABLE", "-default", undef);   # 0==>disable, 1==>enable, undef==>no_action
    param("SETUP_DONE",     "-default", 'reuse'); #this will replace EXTERN_SETUP in the future -- 1, 0, "preclean", "reuse"
    param("EXTERN_SETUP",   "-default", $SETUP_DONE);
    param("CREATE_COMMON_RESOURCES",  "-default", 1);  # See tools/utils.pm
    param("FILER_AGGR",     "-default", 'testaggr');
    param("FILER_VOL_FLEX",     "-default", 'testvol_flex');
    param("FILER_VOL_TRAD",     "-default", 'testvol_trad');
    param("UPGRADE_AGGR", '-default', 'upgrade_aggr');
    param("FILER_CONN",     "-default", "#rsh-socket");
    param("FILER_IP",       "-default", undef);
    param('FILER_QTREE',    "-default", "filerqtree");
    param('FILER_VOL_MNT', 	'-default', undef);
    param("LARGE_AGGR",     '-default', undef);   # 0==>disable, 1==>enable, undef==>no_action
    param('MODIFY_VOL',	'-default', 0);
    param('MOUNTOPTS',	'-default', 'actimeo=0,noac,hard,intr,noquota');
    param('MOUNTPOINT',	'-default', undef);
    param("NBLADE_START",   "-default", undef());
    param("NBLADE_HOST",    "-default", undef());
    param("NBLADE_DIR",     "-default", undef());
    param("ONTAPDIR",	"-default", "/rlse/mainN");
    param('POPULATE_VOL',	'-default', 0);		# 'clear', 'preclean', 'reuse', 0
    param("PFILER",      	"-default", "");
    param("TIMEOUT", 	"-default", 600); # 10 mins
    param("MOUNT_VERS",     "-default", undef); #could be 4
    param("FLEX_SIZE",      "-default", '30m');
    param("TRAD_SIZE",      "-default", 3);
    param("WORM_RETENTION", "-default", '30s');
    param("WAFLIRON_TIMEOUT", "-default", 2400);
    param("SKIP_SNAPLOCK_RETENTION_SETUP",      "-default", 0);
    param("SLEEP_TIMEOUT_MULTIPLIER",    "-default", 1);  # Used to increase 'sleep' and/or 'timeout' values for slow systems.
    param("SLCC_STRICT",    "-default", 1);
    param("SNAPLOCK_TYPE", "-default", 'snaplock');
    param("FS_PROTO", "-default", 'nfs');
    param("FILER_MODE",     "-default", 'mono');
    param("RAID_TYPE", "-default", "raid_dp");   #  raid4 or raid_dp
    param("USE_ZAPI_IF_AVAILABLE", '-default', undef);
    param("BOOT_TIMEOUT", '-default', 2400);
    param("IRON_ON_ROOT", "-default", undef());      
    param("VSERVER",        "-default", "test_vserver");      
    param("HYBRID_AGGR", "-default", 0);
    param("HYBRID_SSD_SIZE", "-default", 3);

    # NA_REVISIT: Temporary fix. For now, pick the first element of FILER_LIST
    # to comply with the sitations where FILER is not defined.

    param('FILER_LIST','-mesg','Comma separated list of filer prefixes');
    # NATE magically defaults FILER_LIST, see ntest help (run ntest -h)
    if (!$FILER_LIST) {
       Carp::croak "FILER_LIST is empty! (run 'ntest -h' to see how this value is defined)";
    }

    my @prefixes = split(/,\s*/,$FILER_LIST);
    my $prefix = $prefixes[0];

    # default FILER to the first filer name provided
    if ($prefix eq 'NODES') {
	# NODES is the GX/10-mode standard parameter
	param('NODES', '-mesg', "List of nodes");
	my @nodes = split(/,\s*/, $NODES);
	param("FILER", '-default', $nodes[0]);
    } else {
	# Get the first element of FILER_LIST and assign into FILER
	param($prefix, '-mesg', "Filer");
	no strict 'refs';
	param("FILER", '-default', ${$prefix});
	use strict 'refs';
    }

	# set VOL_TYPE defaults
	$self->sync_vol_type_params(
	    action   => '-default',
	    vol_type => 'flex',
	);

    if ($SNAPLOCK_TYPE eq 'snaplock_enterprise') {
	    param("WORM_AGGR", "-default", 'wormaggr_enterprise');
    } else {
	    param("WORM_AGGR", "-default", 'wormaggr_compliance');
    }

    # This is now done in the lib fcn. which uses 'DF_SYNC_DELAY'.
    #if ($SLEEP_TIMEOUT_MULTIPLIER) {
    #	$DF_SYNC_DELAY *= $SLEEP_TIMEOUT_MULTIPLIER;
    #}
}

# the parameters below relate to the VOL_TYPE and must be kept 
# in sync. so, we handle them as a group in this method.
sub sync_vol_type_params {
    my $self = shift;

    # validate args
    my (%args) = validate (
	@_,
	{
	    # allow for either defaulting or re-setting the VOL_TYPE params
	    action => {
		type    => SCALAR,
		regex   => qr/-default|-set/,
		default => '-set',
	    },
	    vol_type => {
		type    => SCALAR,
	    },
	}
    );
    my $action   = $args{action};
    my $vol_type = $args{vol_type};

    # do param assignments in package 'main'
    package main;

    param("VOL_TYPE", $action,  $vol_type);

    if ($::VOL_TYPE ne 'trad') {
	param("VOL_SIZE", $action, $::FLEX_SIZE);
	param("FILER_VOL",$action, 'testvol_flex');
	param("CONTAINER", $action, $::FILER_AGGR);
	if ($SNAPLOCK_TYPE eq 'snaplock_enterprise') {
	    param("WORM_VOL", $action, 'wormvol_enterprise_flex');
	} else {
	    param("WORM_VOL", $action, 'wormvol_compliance_flex');
	}
    } else {
	param("VOL_SIZE", $action, $::TRAD_SIZE);
	param("FILER_VOL",$action, 'testvol_trad');
	param("CONTAINER", $action, $::FILER_VOL);
	if ($SNAPLOCK_TYPE eq 'snaplock_enterprise') {
	    param("WORM_VOL", $action, 'wormvol_enterprise_trad');
	} else {
	    param("WORM_VOL", $action, 'wormvol_compliance_trad');
	}
    }

    # return a hash of the values set by this method
    return {
        VOL_TYPE  => $::VOL_TYPE,
        FILER_VOL => $::FILER_VOL,
        CONTAINER => $::CONTAINER,
        WORM_VOL  => $::WORM_VOL,
    };
}

1;

