# Let emacs know that we are editing -*-perl-*-
	
#
# $Id: //depot/prod/test/Rsierranevada/lib/FS/i2p/utils.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library utility for fs/i2p tests.
##
## @description
##     This library contains functions which perform tasks common to the
##     fs/i2p tests:
##
##     create_tree - create a tree consisting of directories, files, and
##                   hard links
##
##     create_inode_table - create a table mapping paths to an inode
##
##     verify_path_list - for an inode, compare the path list reported by
##                        inodepath against a given path table
##
##     compare_i2p_data - compare the i2p data stored in inodes
##
## @usage
##     use FS::i2p::lutils;
##
## @author sager@netapp.com
## @author gibbons@netapp.com Created a new library in the agnostic branch
## @status public
##
## Modified as a replica of BURT 299178 by sudarsh1@netapp.com, The create_tree sub routinue definition has a new hash element being defined $args{client}.
## It is defaulted to localhost unless defined. This value is retained in $client scalar value. This value is being passed to _create_directories,
## _create_files and _create_links. In these subroutinues a connection parameter to client is being defined. This is used to create directories,
## create files and create links. The parameter passed is conn => $cc. 

package FS::i2p::utils;

use strict;
use Carp qw(croak);
use Data::Dumper;
use Exporter;
use Tharn qw(say);
use Filer::Version qw(get_wafl_version);
use FS::Util::Utilities qw(util_mkdir util_chmod util_mkfile util_link util_stat);
use Utils qw(make_conn);
use SayWrap qw(sayEOF);

# Result codes for verify_path_list()
use constant I2P_MATCH => 0;           # the tables match
use constant I2P_MISSING_PATH => 1;    # a path in the specified table was not
                                       # found in the inodepath list
use constant I2P_EXTRANEOUS_PATH => 2; # the inodepath list contained a path
                                       # that does not exist in the specified
                                       # table

my @result_codes = qw(I2P_MATCH
                      I2P_MISSING_PATH
                      I2P_EXTRANEOUS_PATH);

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(&create_tree
                    &create_inode_table
                    &verify_path_list
                    &compare_i2p_data);
push(@EXPORT_OK, @result_codes);
our %EXPORT_TAGS = ( result_codes => [@result_codes] );

## @name ::create_tree
##
## @summary
##     Create a tree consisting of directories, files, and hard links.
##
## @description
##     This function takes a base directory and creates a tree of directories,
##     files, and hard links.  The entities are created randomly within the
##     tree.  The number of entities created is dictated by named parameters.
##     Hard links that are created will point to other files within the tree
##     (and not necessarily to files in the same directory as the link).  At
##     least one of ndirs or nfiles must be specified.  If nlinks is specified,
##     nfiles must be specified as well.
##
##     The function returns a reference to a list of relative path names.
##
##     For example:
##
##     my $paths_ref = create_tree(client_conn => $hc, basedir => 'mytree', ndirs => 5,
##                                 nfiles => 10, nlinks => 20);
##
##     This will create a directory tree under 'mytree' and return a list
##     of 30 relative path names (10 files + 20 links).
##
## @arg %options Optional: tree info name/value pairs
##
## @arg client_conn  client connection object. Defaults to $main::CLIENT_CONN
##
## @arg basedir Optional: base directory.  Defaults to . (current directory)
##
## @arg ndirs Optional: number of directories to create.  Defaults to 0.
##
## @arg nfiles Optional: number of files to create.  Defaults to 0.
##
## @arg nlinks Optional: number of hard links to create.  Defaults to 0.
##
## @arg client Optional: client parameter for NFS mount paths.  Defaults to localhost.

sub create_tree
{
    my (%args) = @_;
    my $client_conn = defined($args{conn})? $args{conn}: $main::CLIENT_CONN;
    my $basedir = $args{basedir} || '.';
    my $ndirs = defined($args{ndirs}) ? $args{ndirs} : 0;
    my $nfiles = defined($args{nfiles}) ? $args{nfiles} : 0;
    my $nlinks = defined($args{nlinks}) ? $args{nlinks} : 0;
    my $client = defined($args{client}) ? $args{client} : 'localhost';

    if ($ndirs == 0 && $nfiles == 0) {
         croak "At least one of ndirs and nfiles must be non-zero";
    }

    if ($nlinks > 0 && $nfiles == 0) {
         croak "nfiles is required when nlinks is greater than 0";
    }

    my(@dirs, @files, @links);
    my @paths;

    _create_dirs($basedir, \@dirs, $ndirs, $client, $client_conn);
    _create_files($basedir, \@dirs, \@files, $nfiles, $client, $client_conn);
    _create_links($basedir, \@dirs, \@files, \@links, $nlinks, $client, $client_conn);

    return([@files, @links]);
}

## @name ::_create_dirs
##
## @summary
##     Create directories within a tree (internal function)
##
## @description
##     Create directories randomly within a tree and store the relative
##     paths in $dirs_ref.
##
## @arg $basedir Required: base directory path
##
## @arg $dirs_ref Required: reference to a list of directories (will
##                          be populated)
##
## @arg $ndirs Required: number of directories to create.
##
## @arg $client Required: client for NFS mount paths.
##
## @arg $c_conn Required: Client connection object

sub _create_dirs
{
    my ($basedir, $dirs_ref, $ndirs, $client, $c_conn) = @_;

    my $conn = make_conn($client, $c_conn);

    for (my $cnt = 1, @$dirs_ref = (""); $cnt <= $ndirs; ++$cnt) {
        my $idx = int rand($cnt);
        my $dir = $dirs_ref->[$idx] . "dir$cnt/";

        util_mkdir("$basedir/$dir", conn => $conn) or croak $!;
        util_chmod(0777, "$basedir/$dir", conn => $conn) or croak $!;

        push(@$dirs_ref, $dir);
    }
}

## @name ::_create_files
##
## @summary
##     Create files within a tree (internal function)
##
## @description
##     Create files randomly within a tree and store the relative paths
##     in $files_ref.
##
## @arg $basedir Required: base directory path
##
## @arg $dirs_ref Required: reference to list of directories (read only).
##
## @arg $files_ref Required: reference to list of files (will be populated).
##
## @arg $nfiles Required: number of files to create.
##
## @arg $client Required: client for NFS mount paths.
##
## @arg $c_conn Required: Client connection object

sub _create_files
{
    my ($basedir, $dirs_ref, $files_ref, $nfiles, $client, $c_conn) = @_;

    my $conn = make_conn($client, $c_conn);

    for (my $cnt = 1; $cnt <= $nfiles; ++$cnt) {
        my $idx = int rand(@$dirs_ref);
        my $file = $dirs_ref->[$idx] . "file$cnt";

	my $cmd = "touch $basedir/$file";
	sayEOF($conn, $cmd);
        util_mkfile("$basedir/$file", conn => $conn) or croak "Unable to create $file";

        push(@$files_ref, $file);
    }
}

## @name ::_create_links
##
## @summary
##     Create hard links within a tree (internal function)
##
## @description
##     Create hard links randomly within a tree and store the relative paths
##     in $links_ref.
##
## @arg $basedir Required: base directory path
##
## @arg $dirs_ref Required: reference to list of directories (read only).
##
## @arg $files_ref Required: reference to list of files (read only).
##
## @arg $links_ref Required: reference to list of links (will be populated).
##
## @arg $nfiles Required: number of hard links to create.
##
## @arg $client Required: client for NFS mount paths.
##
## @arg $c_conn Required: Client connection object

sub _create_links
{
    my ($basedir, $dirs_ref, $files_ref, $links_ref, $nlinks, $client, $c_conn) = @_;

    my $conn = make_conn($client, $c_conn);

    for (my $cnt = 1; $cnt <= $nlinks; ++$cnt) {
        my $idx = int rand(@$dirs_ref);
        my $target_idx = int rand(@$files_ref);
	my $link = $dirs_ref->[$idx] . "link$cnt";

        util_link("$basedir/$$files_ref[$target_idx]", "$basedir/$link", conn => $conn) or croak $!;

        push(@$links_ref, $link);
    }
}

## @name ::create_inode_table
##
## @summary
##     Retrieve inode numbers and create a table mapping paths to inodes
##
## @description
##     This function retrieves inode numbers for each path and creates a
##     table of inodes.  Each entry in the table is an inode number that
##     points to its list of paths.  A hash is used to represent the path
##     list in order to simplify lookups.
##
## @arg $basedir Required: base directory.
##
## @arg $path_list Required: list of paths

sub create_inode_table
{
    my ($basedir, $path_list, %args) = @_;

    my %inode_table;
    my $client = defined($args{client}) ? $args{client} : 'localhost';
    my $c_conn = defined($args{client_conn})? $args{client_conn} : $main::CLIENT_CONN; 
    my $conn = make_conn($client, $c_conn);

    foreach my $path (@$path_list) {
        my $inode = (util_stat("$basedir/$path", conn => $conn))[1] or croak $!;
        $inode_table{$inode}{$path} = "";
    }

    return(\%inode_table);
}

## @name ::verify_path_list
##
## @summary
##     Verify a list of paths against a table of expected paths.
##
## @description
##     This function compares a list of paths as reported by inodepath
##     against a specified table of expected paths.  The function generates
##     error codes from the perspective of the expected path table.
##
##     A reference to a result hash is returned.  Within each hash,
##     $result{code} will contain a mnemonic identifying the result.
##
##     If the paths match, the result code will be I2P_MATCH.  If the
##     paths do not match, additional hash entries will be defined and
##     populated accordingly so that a calling function can report error
##     messages in whatever way is appropriate for its task on hand.
##
##     Valid result codes are defined in @result_codes above.  The only
##     additional hash entry currently defined for reporting purposes is
##     'path'.
##
##
## @arg $expected_path_table Required: reference to a hash containing the
##                                     expected paths
##
## @arg $path_list Required: reference to a list of paths in the format
##                           returned by i2p_inode_to_path()

sub verify_path_list
{
    my ($expected_path_table, $path_list) = @_;

    my %result;

    $result{code} = I2P_MATCH;

    # Loop through path list and create a path table for easier comparison

    my %path_table;
    foreach (@$path_list) {
        my $path = $_->{path};
        $path_table{$path} = "";
    }

    # Check for missing paths

    foreach my $path (keys(%$expected_path_table)) {
        if (!defined($path_table{$path})) {
            $result{code} = I2P_MISSING_PATH;
            $result{path} = $path;
            return(\%result);
        }
    }

    # Check for extraneous paths.

    if (keys(%$expected_path_table) < keys(%path_table)) {
        foreach my $path (keys(%path_table)) {
            if (!defined($expected_path_table->{$path})) {
                $result{code} = I2P_EXTRANEOUS_PATH;
                $result{path} = $path;
                return(\%result);
            }
        }
    }

    return(\%result);
}

## @name ::compare_i2p_data
##
## @summary
##     Compare i2p data stored in inodes
##
## @description
##     This function compares the i2p data stored in an inode vs that
##     stored in another inode.  The function's arguments are expected
##     to be in the format returned by i2p_inode_to_path(), i.e., a list
##     reference in which each element in the list corresponds to path
##     data.  The path data elements are hash references.  The hash's keys
##     are name, path, par_inode, and par_cookie.
##
##     If the entries match, the function returns 1; otherwise, it returns
##     0.
##
## @arg $list1 Required: path data as defined above
##
## @arg $list2 Required: path data as defined above

sub compare_i2p_data($$;$%)
{
   my($list1, $list2, $fh, %opts) = @_;
   my $snap_file = 0;  # change 506054 (burt185033)

   # The number of entries in each list corresponds to the number of
   # paths for a particular inode.  For each entry, there are four fields:
   # file name, file path, parent inode number, and parent cookie (directory
   # index).  Compare each entry and return accordingly.

   return 0 if (@$list1 != @$list2);

   if (defined($opts{snap1}) || defined($opts{snap2})) {
       if (!defined($fh) ) {
          croak("compare_i2p_data need fh argument ".
                "if snap option(s) is set");
       }
       my $version = get_wafl_version($fh);
 #         if ($version >= WV_GORDONBIERSCH_MIN) {
           $snap_file = 1;
#       }
    }

   for (0..@$list1-1) {
       my $name1 = $list1->[$_]->{name};
       my $name2 = $list2->[$_]->{name};
       my $path1 = $list1->[$_]->{path};
       my $path2 = $list2->[$_]->{path};
       if ($snap_file && defined ($opts{snap1} )) {
          my $snap = $opts{snap1};
          $name1 =~ s/\.snapshot\/$snap\///; 
          $path1 =~ s/\.snapshot\/$snap\///;
       }
       if ($snap_file && defined ($opts{snap2} )){
          my $snap = $opts{snap2};
          $name2 =~ s/\.snapshot\/$snap\///;
          $path2 =~ s/\.snapshot\/$snap\///;
       }
       return 0 if ($name1 ne $name2 ||
                    $path1 ne $path2 ||
                    $list1->[$_]->{par_inode} != $list2->[$_]->{par_inode} ||
                    $list1->[$_]->{par_cookie} != $list2->[$_]->{par_cookie});
        
          
   }

   return 1;
}

1;
