#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/FSH/CmdServer.pm#1 $
#
#Copyright (c) 2005 Network Appliance, Inc.
#All rights reserved.
#
## @summary
##     CmdServer - CmdServer class to invoke rpc server commands on the
##     remote host.  This is the parent class for NfsServer and CifsServer
##     classes
##
## @description
##     This library contains functions which perform tasks common to the
##     fs/write_correctness/bad_disk tests.
##
## @usage
##     use FSH::CmdServer;
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public
use strict;
package CmdServer;
use Carp;
use Tharn;
use MapShare;
use FS::Util::Setup qw(create_and_map_cifs_share);
use SPARSE::Util::Logger qw( logTrc logWarn );
use Data::Dumper;

# The literal code in $CMD are the commands that the rpc server can
# invoke on the remote host.  These commands are written in portable
# perl so that they may be identically use on either *nix or windows.

my $CMD =<<'EOF';
{
    use strict;
    use Data::Dumper;
    use Fcntl qw(:DEFAULT);
    use File::Copy;
    use IO::Handle;
    Subtest::ipc()->serve();

    sub my_create_file( $$ );
    sub my_copy_file( $$ );
    sub my_mkdir( $ );

    sub my_create_file( $$ ) {
        my ($path, $line_count) = @_;
        my $line = 'abcdefghijklmnopqrstuvwxyz0123456789`~!@#%^&*()_+{}|\][:;<.,>?ABCDEFGHIJKLMNOPQUSTUVWXYZ';
        open(FILE, ">$path") or return 0;
        for (my $i=0; $i<$line_count; $i++) {
            print FILE "$line\n" or return 0;
        }
        close(FILE);
        return 1;
    }

    sub my_truncate_file( $ ){
        my ($path) = @_;
        my_create_file($path, 0);
        return 1;
    }

    sub my_copy_file( $$ ) {
        my ($source_file, $target_file) = @_;
        copy ($source_file, $target_file) or return 0;
        return 1;
    }

    sub my_mkdir( $ ){
        my ($path) = @_;
        mkdir($path) or return 0;
        return 1;
    }

    sub my_rmdir( $ ){
        my ($path) = @_;
        rmdir($path) or return 0;
        return 1;
    }
    sub my_continue_writing_files( $$$ ) {
        my ($source, $target, $count) = @_;
        my $line_count = 1000;
        my $skip = 10;
        # by design, a count of -1 will make this loop a *long* time.
        while ($count != 0) {
            my $ts = time();
            my $source_file = sprintf("%s_%s_%s", $source, $ts, $count);
            my $target_file = sprintf("%s_%s_%s", $target, $ts, $count);
            open(FILE, ">$source_file");
            for (my $i=0; $i<$line_count; $i++) {
                my $line = sprintf("%s__", $ts) x 3;
                print FILE "$line\n";
                $i += 1;
            }
            close(FILE);
            copy ($source_file, $target_file);
            $skip -= 1;
            if ($skip == 0) {
                print "$source_file --> $target_file\n";
                sleep(1);
                $skip = 10;
            }
            unlink ($source_file, $target_file);
            $count -= 1;
        }
        return 1;
    }
}
EOF

# XFS_Client object knows what host to run on, and can create begin
# and finish any number of server instances, but only in strictly
# sequential order.  In typical use, only one server would be created
# that would be used to handle all of the commands issued.
#
# All of the commands issued run in the background. Only the
# rpc_finish method waits for command to finish before returning.  To
# run command in strict sequence requires that they be bracketed with
# calls to setup_rpc_server and rpc_finish.
sub new( $$ ) {
    my ($class, $client) = @_;
    my $self = {};
    $self->{CLIENT} = $client;
    $self->{CMD_BLOCK} = $CMD;
    $self->{SUBTEST} = undef;
    bless $self, $class;
    return $self;
}

# Should be called as the first step on the new instance.
sub setup_rpc_server( $ ) {
    my $self = shift;
    my $client = $self->my_client();
    my $subtest_obj = subtest($self->{CMD_BLOCK},
                              '-bg',
                              '-ipc',
                              '-runid', "subtest on $client",
                              '-client', $client );
    my $timer_on = time();
    while ( ! $subtest_obj->ipc()->eval('return 1;') ) {
        sleep 1;
        if ( (time() - $timer_on) > 10 ) {
            die " rpc server is not running ";
        }
    }
    $self->{SUBTEST} = $subtest_obj;
}

sub send_copy_file ( $$$ ) {
    my ($self, $source, $target) = @_;
    my @frozen_args = map { NATE::RPC::Freeze::freeze($_) } ($source, $target);
    my $client = $self->subtestx()->ipc();
    my $function_call = "my_copy_file" . "(" . join(",", @frozen_args) . ")";
    $client->eval_request($function_call) or
        die "eval_request failed to invoke the function: $function_call on ".
            "client = $client";
}

sub send_create_file ( $$$ ) {
    my ($self, $path, $line_count) = @_;
    my @frozen_args = map { NATE::RPC::Freeze::freeze($_) } ($path, $line_count);
    my $client = $self->subtestx()->ipc();
    my $function_call = "my_create_file" . "(" . join(",", @frozen_args) . ")";
    $client->eval_request($function_call) or
        die "eval_request failed to invoke the function: $function_call on ".
            "client = $client";
}

sub send_mkdir ( $$$ ) {
    my ($self, $path) = @_;
    my @frozen_args = map { NATE::RPC::Freeze::freeze($_) } ($path);
    my $client = $self->subtestx()->ipc();
    my $function_call = "my_mkdir" . "(" . join(",", @frozen_args) . ")";
    $client->eval_request($function_call) or
        die "eval_request failed to invoke the function: $function_call on ".
            "client = $client";
}

sub send_rmdir ( $$$ ) {
    my ($self, $path) = @_;
    my @frozen_args = map { NATE::RPC::Freeze::freeze($_) } ($path);
    my $client = $self->subtestx()->ipc();
    my $function_call = "my_rmdir" . "(" . join(",", @frozen_args) . ")";
    $client->eval_request($function_call) or
        die "eval_request failed to invoke the function: $function_call on ".
            "client = $client";
}

sub send_truncate_file ( $$$ ) {
    my ($self, $path) = @_;
    my @frozen_args = map { NATE::RPC::Freeze::freeze($_) } ($path);
    my $client = $self->subtestx()->ipc();
    my $function_call = "my_truncate_file" . "(" . join(",", @frozen_args) . ")";
    $client->eval_request($function_call) or
        die "eval_request failed to invoke the function: $function_call on ".
            "client = $client";
}

sub send_continue_writing_files ( $$$ ) {
    my ($self, $source, $target, $count) = @_;
    my @frozen_args = map { NATE::RPC::Freeze::freeze($_) } ($source, $target, $count);
    my $client = $self->subtestx()->ipc();
    my $function_call = "my_continue_writing_files" . "(" . join(",", @frozen_args) . ")";
    #print "F>$function_call\n";
    $client->eval_request($function_call) or
        die "eval_request failed to invoke the function: $function_call on ".
            "client = $client";
}

sub rpc_finish ( $ ) {
    my $self = shift;
    my $begin = time();
    if ( defined($self->{SUBTEST}) ) {
        $self->{SUBTEST}->ipc()->quit();
        $self->{SUBTEST}->wait_finish();
    }
    logcomment(sprintf("Waited %d secs for subtests to finish.", time()-$begin));
}

sub kill ( $ ) {
    my $self = shift;
    if ( defined($self->{SUBTEST}) ) {
        $self->{SUBTEST}->kill();
    }
}

sub my_client( $ ) {
    my $self = shift;
    return $self->{CLIENT};
}

sub subtestx( $ ) {
    my $self = shift;
    return $self->{SUBTEST};
}

1;
