#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/FSH/LogTail.pm#1 $
#
#Copyright (c) 2005 Network Appliance, Inc.
#All rights reserved.
#
## @summary
##     LogTail - Class log to tail the /etc/messages file on the filer
##
## @description
##     This library contains functions which perform tasks common to the
##     fs/write_correctness/bad_disk tests.
##
## @usage
##     use FSH::LogTail;
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public
use strict;

use IO::File;
use IO::Seekable;
use POSIX qw(:errno_h);
use Fcntl qw(:mode);


package LogTail;
use Data::Dumper;
# Create a logtail reader.  Once created, all new lines at the end of
# the file are returned each time get_new() method is called.  The
# lines in the log that were presented before class is created will
# never be seen.  The required argument is the path to the logfile.
# Normally, this will be the path to the etc/messages of a filer's
# root vol.  It is up to the called to mount the volume.  The second
# argument is the guaranteed pause between reads to the log file. The
# default is 5 seconds.  This allows the caller write a simple loop of
# get_new() calls and not have to worry about hammering the log file
# with reads.
sub new( $$;% ) {
    my $class = shift;
    my $self = {};
    my $path = shift;
    $self->{PATH} = $path;
    $self->{DELAY} = 5; # seconds
    if (@_) {
        my %kw = @_;
        if ($kw{delay}) {
            $self->{DELAY} = $kw{delay};
        }
    }
    $self->{LOG} = IO::File->new($path, "r") or die($!);
    # advance to the end now to avoid race condition.
    $self->{OFFSET} = sysseek($self->{LOG}, 0, 2);
    $self->{TIME} = time();
    bless $self, $class;
    my $r_s = $self->get_stat();
    $self->{INO} = $r_s->{INO};
    $self->{NLINK} = $r_s->{NLINK};
    $self->{SIZE} = $r_s->{SIZE};
    return $self;
}

# Wait for the delay to run out, and then return any new data at the
# end of the file. Complain if it has become a different file.
sub get_new( $ ) {
    my $self = shift;
    my $buf;
    $self->remaining_delay();
    $self->update_stat();
    $self->{LOG}->clearerr();
    my $count = 0;
    my $data = '';
    while (1) {
        my $rc = sysread($self->{LOG}, $buf, 1);
        if (! defined($rc)) {
            die $!;
        }
        if ($rc == 0) {
            last;
        }
        $count += $rc;
        $data .= $buf;
    }
    $self->{OFFSET} += $count;
    return $data;
}

sub to_string ( $ ) {
    my $self = shift;
    return sprintf "LogTail->new('%s',delay=>%s)", $self->path(), $self->delay();
}

sub delay( $ ) {
    my $self = shift;
    return $self->{DELAY};
}

sub path( $ ) {
    my $self = shift;
    return $self->{PATH};
}

sub finish( $ ) {
    my $self = shift;
    close($self->{LOG});
}

# Write /etc/messages to a tty window. To use, create a new xterm
# window to be used as output.  Determine its device name with the
# 'tty' command, and pass that string to this method.  A script that
# call this method should run in some window other that the output
# window. This methods loops forever; kill the script to stop.
sub monitor( $$ ) {
    my ($self, $tty) = @_;
    open(MYTTY, ">$tty") or die $!;
    while(1) {
        my @msgs = $self->get_new();
        foreach my $m ( @msgs ) {
            print MYTTY $m;
        }
    }
}

### ---- private ------- ####

sub update_stat( $ ) {
    my $self = shift;
    my $r_s = $self->get_stat();
    if ($r_s->{INO} != $self->{INO}) {
        die "different_file(INO)";
    }
    if ($r_s->{SIZE} < $self->{SIZE}) {
        die "different_file(SIZE)";
    }
#    if ($r_s->{NLINK} == 0) {
#        die "unlinked_file";
#    }

    $self->{INO} = $r_s->{INO};
    $self->{NLINK} = $r_s->{NLINK};
    $self->{SIZE} = $r_s->{SIZE};
}

sub get_stat( $ ) {
    my $self = shift;
    my @stats = stat($self->{PATH})
        or die "different_file($!)";
    my %s = ();
    $s{INO} = $stats[1];
    $s{NLINK} = $stats[3];
    $s{SIZE} = $stats[7];
    return \%s;
}

sub remaining_delay( $ ) {
    my $self = shift;
    my $diff = time() - $self->{TIME};
    if ($diff < $self->delay()) {
        sleep($self->delay - $diff);
    }
    $self->{TIME} = time();
}


package LogMatcher;
use Data::Dumper;

sub new( $$ ) {
    my $class = shift;
    my $self = {};
    $self->{LOGTAIL_OBJ} = shift;
    $self->{BUF} = [];
    bless $self, $class;
    return $self;
}

sub accumulate ( $ ) {
    my $self = shift;
    my $data = $self->log_obj()->get_new();
    my @temp_array = split("\n", $data);

    foreach my $entry (@temp_array) {
            push(@{$self->{BUF}}, $entry);
    }
    #push ( @{$self->{BUF}}, (split("\n", $data)) );
}

sub get_matches ( $$ ) {
    my $self = shift;
    my $pattern = shift;
    my @matched = ();

    my $temp = $self->get_buf();
#    my @ta = @{$blah};
#    print (Dumper(\@ta));

    foreach my $q ( @{$temp} ) {
        if ($q =~ /$pattern/) {
            push @matched, $q;
        }
    }
    return @matched;
}

sub get_buf ( $ ) {
    my $self = shift;
    return $self->{BUF};
}

sub buf_length ( $ ) {
    my $self = shift;
    return scalar @ {$self->get_buf() };
}

sub flush ( $ ) {
    my $self = shift;
    $self->{BUF} = ();
}

# flush_to_match(pat) removes all lines that match pat and everything
# that preceeds them.
sub flush_to_pattern ( $$ ) {
    my $self = shift;
    my $pat = shift;
    my $final = -1;
    for (my $i=0; $i < $self->buf_length(); $i++) {
        if ( @{$self->get_buf()}[$i] =~ /$pat/) {
            $final = $i;
        }

    }
    if ($final == -1) {
        die "arg_error('pattern not found')";
    }

    my @temp = @{$self->get_buf()};
    @{$self->{BUF}} = @temp[ $final .. $#temp];
    # @{$self->{BUF}} =  @{$self->get_buf()}[$final..$self->buf_length() -1];

}

sub log_obj ( $ ) {
    my $self = shift;
    return $self->{LOGTAIL_OBJ};
}

sub finish( $ ) {
    my $self = shift;
    $self->log_obj->finish();
}

1;

# TODO
# /etc/messages are release-specific, in general.
