#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/FSH/baddisk.pm#1 $
#
#Copyright (c) 2005 Network Appliance, Inc.
#All rights reserved.
#
## @summary
##     Shared bad_disk routine.
##
## @description
##     This library contains functions which perform tasks common to the
##     fs/write_correctness tests:
##
##     create_tree - create a tree consisting of directories, files, and
##                   hard links
##
## @usage
##     use baddisk;
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public

use strict;

package baddisk;

use Tharn;
use SPARSE::Util::Logger qw( logError logStep logWarn logTrc);
use SPARSE::Resource::Filer;
use Carp;
use Data::Dumper;
use Volume qw(vol_disks disk_blocks_written);
use SayWrap qw(sayEOF);
use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( bs  );

sub bs {
    return;
}
## @name ::disk_fail
##
## @summary
##     Fail a disk.
##
## @description
##     Implements the 'disk fail -i -f' command.
##
## @arg $fc Required: filer connection.
##
## @arg $disk_name Required: name of disk (v0.20).
##
sub disk_fail ( $$ ) {
    # TDB: belongs in a library, but which one?
    my ($fc, $fail_disk) = @_;
    my $cmd = "disk fail -i -f $fail_disk";
    my $ret = sayEOF($fc, $cmd);
}

## @name ::unfail_disk
##
## @summary
##     Convert a previously failed disk to a spare.
##
## @description
##     Implements the 'disk unfail' command.
##
## @arg $fc Required: filer connection.
##
## @arg $disk_name Required: name of disk (v0.20).
##
sub unfail_disk ( $$ ) {
    # TDB: belongs in a library, but which one?
    my ($fc, $fail_disk) = @_;
    if (!defined($fail_disk)) {
        die 'arg_error(fail_disk is argument undefined)';
    }
    my $cmd = "priv set -q advanced; disk unfail $fail_disk";
    my $ret = sayEOF($fc, $cmd);
    if ($ret =~ /Disk is not currently failed/) {
        die("error($ret)");
    }
}


## @name ::data_disks
##
## @summary
##     Get the data disks used by the aggregate that contains the
##     named volume.  Parity disk are specially omitted.
##
## @description
##     Return an array of disk names.  These name are data disks of
##     the test volume.
##
## @arg $fc Required: filer connection.
##
## @arg $vol_name Required: name of volume.
##
sub data_disks ( $$ ) {
    my ($fc, $filer_vol) = @_;
    return grep {$_->{type} eq 'data'} vol_disks($fc, $filer_vol);
}


## @name ::disks_to_fail
##
## @summary
##     Find to disk to fail.
##
## @description
##      Return a pair of disks from the volume.
##
## @arg $fc Required: filer connection.
##
## @arg $vol_name Required: name of volume.
##
sub disks_to_fail ( $$ ) {
    my ($fc, $filer_vol) = @_;
    my ($disk1, $disk2);
    my @ds = data_disks($fc, $filer_vol);
    ($disk1, $disk2) = ($ds[$#ds-1], $ds[$#ds]);
    if ($disk1->{phys_mb} > $disk1->{phys_mb}) {
        ($disk1, $disk2) = ($disk2, $disk1)
    }
    foreach my $d (@ds) {
        if ($d->{phys_mb} < $disk1->{phys_mb}) {
            $disk1 = $d;
        } elsif ($d->{phys_mb} < $disk1->{phys_mb}) {
            $disk2 = $d;
        }
    }
    return ($disk1, $disk2);
}

## @name ::volume_percentage_free
##
## @summary
##     Determine volume free space
##
## @description
##      Return the remaining free space of vol_name as a percent.
##
## @arg $fc Required: filer connection.
##
## @arg $vol_name Required: name of volume.
##


sub volume_percentage_free ( $$ ) {
    my ($fc, $vol_name) = @_;
    my $r = Volume::vol_disk_free($fc, path=>"$vol_name");
    return 100 - $r->{capacity}
}

## @name ::had_dparity_disk
##
## @summary
##     Insure that volume is of an aggregate that uses raid_dp.
##
## @description
##     Insure that volume is of an aggregate that uses raid_dp.
##
## @arg $fc Required: filer connection.
##
## @arg $vol_name Required: name of volume.
##
sub has_dparity_disk ( $$ ) {
    my ($fc, $filer_vol) = @_;
    my @d = grep {$_->{type} eq 'dparity'} vol_disks($fc, $filer_vol);
    return scalar @d;
}

## @name ::setflag
##
## @summary
##     Set a flag on the filer.
##
## @description
##     Set a flag on the filer, or raise and exception. There is no
##     checking of either the name of the flag or the value.  An
##     exception is raised only if the filer complains.
##
## @arg $fc Required: filer connection.
##
## @arg $flag Required: Flag name
##
## @arg $value Required: New setting for flag.
##
sub setflag ( $$$ ) {
    my ($fc, $flag, $value) = @_;
    my $sf = SPARSE::Resource::Filer::sparse_filer($::FILER);
    eval {
        $sf->setflag({$flag => $value});
    };
    die $@ if $@;
}

## @name ::etc_messages_path
##
## @summary
##     Return /etc/messages location.
##
## @description
##     Compose the path to the log file on the filer. The path is
##     though a mounted root volume. No filer or NFS operations are
##     performed
##
## @arg $flag Required: Filer root mount point
##
sub etc_messages_path( $ ) {
    my $root_vol = shift;
    my $mpath = $root_vol . '/etc/messages';
    return $mpath;
}

## @name ::disks_io_stats
##
## @summary Return /etc/messages location.
##
## @description
##     Watch the disk I/O for a time, and return a summary of the
##     statistics for each data disk.
##
## @arg $fc Required: filer connection.
##
## @arg $vol_name Required: name of volume.
##
## @arg $count Required: Number of samples to take.
##
## @arg @disks Required: List of disk to sample/
##
sub disks_io_stats( $$$@ ) {
    my ($fc, $vol_name, $count, @disks) = @_;
    my @prof = accum_all_disk_writes($fc, $vol_name, $count, @disks);
    my @flat = flatten(@prof);
    my @norm = normalize(@flat);
    my %sum = sum(@norm);
    return %sum;
}

# --------------------- The remaining functions are private ---------- #

# Sum the total written and queued for each disk.
sub sum( @ ) {
    my @data = @_;
    my %out = ();
    my %disk_set = ();
    my @disk_list = ();
    my %total_queued = ();
    my %total_written = ();
    foreach my $rec ( @data ) {
        my ($q2, $w2);
        my @new_rec;
        my ($group, $disk, $hot, $queued, $written) = @$rec;
        if (defined($disk_set{$disk})) {
            $total_queued{$disk} += $queued;
            $total_written{$disk} += $written;
        } else {
            $disk_set{$disk} = 1;
            push @disk_list, ($disk);
            $total_queued{$disk} = $queued;
            $total_written{$disk} = $written;
        }
    }
    foreach my $d (@disk_list) {
        $out{$d} = {queued => $total_queued{$d}, written => $total_written{$d}};
        #push @out, (\@new_rec);
    }
    return %out;
}

# Subtract the first entries from all that follow, for each disk
sub normalize( @ ) {
    my @data = @_;
    my @out = ();
    my %first_queued = ();
    my %first_written = ();
    foreach my $rec ( @data ) {
        my ($q2, $w2);
        my @new_rec;
        my ($group, $disk, $hot, $queued, $written) = @$rec;
        if (defined($first_queued{$disk})) {
            $q2 = $queued - $first_queued{$disk};
            $w2 = $written - $first_written{$disk};
        } else {
            $first_queued{$disk} = $queued;
            $first_written{$disk} = $written;
            $q2 = $queued;
            $w2 = $written;
        }
        @new_rec = ($group, $disk, $hot, $q2, $w2);
        push @out, (\@new_rec);
    }
    return @out;
}

# Return a list of disk records, numbered according to their original
# order.  The input is a list of set of disk records.
sub flatten ( @ ) {
    my @accum = @_;
    my @flat = ();
    my $group = 0;
    foreach my $state ( @accum ) {
        foreach my $disk_set ( @{ $state } ) {
            my @rec = ($group);
            $group++;
            push @rec, (@{$disk_set});
            push @flat, (\@rec);
        }
    }
    return @flat;
}

# Poll the writing activity of all the @disks $count times.
sub accum_all_disk_writes( $$$@ ) {
    my $fc = shift;
    my $vol_name = shift;
    my $count = shift;
    my @disks = @_;
    my @accum = ();

    for (my $i=0;$i<$count;$i++) {
        my @stats = disk_writes($fc, $vol_name, @disks);
        push @accum, (\@stats);
        #sleep(1);
    }
    return @accum;
}

# Get the write activity for listed disks.
sub disk_writes( $$@) {
    my $fc = shift;
    my $vol_name = shift;
    my @disks = @_;
    my @stats;

        foreach my $d ( @disks ) {
            my @rec;
            next if ($d->{type} =~ /parity/);
            my $h = disk_blocks_written($fc, $vol_name, $d->{name});
            @rec = ($h->{disk},$h->{hot},$h->{queued},$h->{written});
            push @stats, (\@rec);
        }
    return @stats;
}

# ---------------------- begin TBD ------------------------#

# Verify named disk is a data disk.
sub check_for_disk ( $$$ ) {
    my ($fc, $filer_vol, $disk_name) = @_;
    foreach my $d ( data_disks($fc,$filer_vol) ) {
        if ( $disk_name eq $d->{name} ) {
            return 1; # good, disk is here.
        }
    }
    return 0; # never saw it.
}

# Count the raid groups used by the aggregate containing vol_name.
sub vol_raid_groups( $$ ) {
    my ($fc, $vol_name) = @_;
    return 1;
}

# Spares for aggregate. Return a pair of lists, on for each plex.  The
# entry in each pair is a list of disk descriptions. The diak
# description is a hash, like that returned by vol_disk and
# additionally tells if the disk has been zeroed.
sub aggr_spares( $$ ) {
    my ($fc, $vol_name) = @_;
}

# Return true if there is a zeroed spare ready to replace $disk_name,
# should it fail.
sub disk_spare_ready( $$ ) {
    my ($fc, $disk_name) = @_;
}

# ---------------------- end TBD   ------------------------#

1;
