#Hey emacs, this is basically -*- perl -*-
# $Id: FileManip.pm,v 1.2 2008/07/28 21:09:35 grybak Exp $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package FileManip;

## @summary Library routines for basic file I/O
##
## @description
##  This module contains library routines for doing basic I/O to
##  files and includes functions to read, write, remove, getattr
##  on files and named attributes (streams). All functions use the
##  same interface for both unix and Windows clients.
##  The manipulation of the file could be either locally or remotely
##  according to the 'client' option passed as a parameter.
##  Also, all functions accept a 'user' option to specify the user
##  to perform the operation on the file/stream.
##
##  The 'client' option could be either a hostspec, Hostrec object
##  or a Connectrec object, but since subtest is used to connect
##  to the client it only takes the Hostrec object out of the
##  Connectrec object so the connection type of the Connectrec
##  object is not used. This option defaults to localhost.
##
##  The first function to be called for a given client and user will
##  start an rpc server on that client as the user, therefore subsequent
##  calls to any function to the same client and as the same user are
##  served by the rpc server with no need to subtest the request again.
##
## @usage
##  use FileManip;
##
## @status	review
##
## @author	dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @notes
##  The user parameter will not work on Window clients because the subtest
##  '-username' option does not work properly on a Windows causing the test
##  to fail with little information about why it failed. A croak was added
##  to catch this case.
##
##  The stream subroutines are currently only supported on Solaris 10
##  and Windows unless otherwise noted.
##
##  Not all possible file I/O routines have been implemented yet so
##  feel free to add routines to this library.

use strict;
use warnings;
use base qw(Exporter);

use Fcntl;
use File::Find;
use File::Basename qw(fileparse);
use Math::BigInt;
use Carp qw(croak);
use Cwd qw(getcwd);
use Storable qw(thaw);
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT ARRAYREF);
use Data::Dumper;

# Check if library is running in NATE environment or standalone
eval { require Tharn };
my $_use_nate = $@ ? 0 : 1;
my $_hosttype;

if ( $_use_nate ) {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
    use vars qw($RPC_LIB_PREPEND);
    Tharn::param('RPC_LIB_PREPEND','-default', undef);
} else {
    if ( $^O eq 'MSWin32' ) {
        $_hosttype = 'windows';
    } else {
        $_hosttype = "unix-$^O";
    }
}

# These functions are for internal use only
# and are exported to be used by the rpc
# service to be run remotely
our @internals = qw(
  _read_file
  _write_file
  _stat_file
  _get_file_size
  _test_file
  _get_dir_entries
  _make_dir
  _remove_dir
  _link_file
  _rename_file
  _remove_file
  _getattr_file
  _setattr_file
  _read_stream
  _write_stream
  _stat_stream
  _rename_stream
  _remove_stream
  _setattr_stream
  _write_sparse
  _vol_info
  _vol_free
);

our @EXPORT_OK = ( @internals, qw(
  read_file
  write_file
  stat_file
  get_file_size
  test_file
  get_dir_entries
  make_dir
  remove_dir
  link_file
  rename_file
  remove_file
  getattr_file
  setattr_file
  read_stream
  write_stream
  stat_stream
  rename_stream
  remove_stream
  setattr_stream
  write_sparse
  vol_info
  vol_free
) );

our %EXPORT_TAGS = (
    # This tag is for internal use only
    # to import all the internal functions
    # on the remote rpc service
    internal => \@internals
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '0.01';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub do_call($%);
sub read_file(%);
sub _read_file(%);
sub write_file(%);
sub _write_file(%);
sub stat_file(%);
sub _stat_file(%);
sub get_file_size(%);
sub _get_file_size(%);
sub test_file(%);
sub _test_file(%);
sub get_dir_entries(%);
sub _get_dir_entries(%);
sub make_dir(%);
sub _make_dir(%);
sub remove_dir(%);
sub _remove_dir(%);
sub link_file(%);
sub _link_file(%);
sub rename_file(%);
sub _rename_file(%);
sub remove_file(%);
sub _remove_file(%);
sub getattr_file(%);
sub _getattr_file(%);
sub setattr_file(%);
sub _setattr_file(%);
sub read_stream(%);
sub _read_stream(%);
sub write_stream(%);
sub _write_stream(%);
sub stat_stream(%);
sub _stat_stream(%);
sub rename_stream(%);
sub _rename_stream(%);
sub remove_stream(%);
sub _remove_stream(%);
sub setattr_stream(%);
sub _setattr_stream(%);
sub write_sparse(%);
sub _write_sparse(%);
sub vol_info(%);
sub _vol_info(%);
sub vol_free(%);
sub _vol_free(%);
sub _win32api_call($@);

#---------------------------------------------------------------------
# Package local constants
#---------------------------------------------------------------------
my %common_parameters = (
    # The user parameter is optional
    user     => { type => SCALAR,         optional => 1 },

    # The client parameter could be either a hostspec, Hostrec object
    # or a Connectrec object. Defaults to 'localhost'.
    client   => { type => SCALAR|OBJECT,   optional => 1,   default => 'localhost' },
);

my @common_params = qw( user  client );

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my %sobj_h;
my %win32api;
my $use_w32api = 1;
my $log_index  = 0;
our $main_pid  = 0;
my $VLOFFSET   = 0xFFFFFFFF;

#---------------------------------------------------------------------
# Internal function.
# Run the specified function locally or remotely.
# When running remotely, it starts an rpc server on
# the remote client if it is not running. Multiple calls to the
# same client and as the same user are served by the rpc server
# with no need to subtest the function again.
#---------------------------------------------------------------------
sub do_call($%) {
    my ($function, %opts) = @_;
    my ($out, @results, $err);

    if ( $_use_nate ) {
        # Get the hostrec object from the client option
        if ( UNIVERSAL::isa($opts{client}, "Connectrec") ) {
            # Get hostrec object from a connectrec object
            $opts{client} = $opts{client}->hostp();
        } elsif ( ! UNIVERSAL::isa($opts{client}, "Hostrec") ) {
            # Get hostrec object from a hostspec
            $opts{client} = Tharn::host($opts{client});
        }
    } else {
        $opts{client} = {
            hostname => 'localhost',
            hosttype => $_hosttype,
        };
    }

    # Check for correct use of the 'user' option
    if ( defined($opts{user}) and $opts{client}{hosttype} =~ /^windows/ ) {
        croak "_do_call: Option 'user' is not supported on Windows";
    }

    # Check if function need to run locally or remotely
    if ( $opts{client}{hostname} eq 'localhost' and !$opts{user} ) {
        if ( $opts{client}{hosttype} =~ /^windows/ ) {
            require Win32API::File;
        }
        # Run function locally
        return &{$function}(%opts);
    }

    # Running function remotely so get name of function to call
    my $func_name = (caller 1)[3];
    $func_name =~ s/.*::/_/;

    if ( $main_pid != $$ ) {
        # This must be a child or the first time
        # calling this function
        %sobj_h = ();
        $log_index = 0;
    }

    # Get hostname and user from the arguments,
    # this uniquely identifies the subtest to send request to
    my $sindex = (defined($opts{user}) ? $opts{user} : '') . '@' .
                 $opts{client}{hostname};

    if ( ! $sobj_h{$sindex} ) {
        # Subtest is not running on client as the specified user
        my $win_use = '';
        my @extra_options;
        push @extra_options, '-username' => $opts{user} if $opts{user};

        # Create different log for subtest instance
        ++$log_index;

        if ( $opts{client}{hosttype} =~ /^windows/ ) {
            $win_use = q{use Win32API::File qw();};
        }

        # Script to run this module remotely.
        # First, it includes this module and the Storable module to
        # freeze the return array if the function returns an array.
        # Then, the rpc server starts listening for requests.

        # For the remote host to load the libraries, we will be defaulting
        # to only X: path for Windows and /usr/software/test for Unix
        # We are avoiding using web.netapp.com, as this server becomes slower
        # from time to time and this impacts the smokes runs.
        # If the use of this server is needed for any reason, the prepend path
        # can be modified by using the command-line argument RPC_LIB_PREPEND
        my $prepend =    "\"X:/agnostic\"," .
                         "\"X:/agnostic/lib\"," .
                         "\"/usr/software/test/agnostic/test\"," .
                         "\"/usr/software/test/agnostic/test/lib\"";

        if (defined($main::RPC_LIB_PREPEND)) {
            my $prepend_str = $main::RPC_LIB_PREPEND;
            # Need to process the cmd-line arg, to make it of the form
            # "path1","path2" (each individual path needs quotation added)
            my @paths = split(/,/, $prepend_str);
            $prepend = '';
            foreach my $path (@paths) {
                $prepend .= "\"$path\",";
            }
        }

        my $cmd = qq{ \{

	########################################################################
	#  Workaround for: "SCRIPT ERROR: Can't locate FileManip.pm in @INC (@INC contains: ..."
	#  when used with Windows clients
	BEGIN {
	   use NATE::Inc;
	   NATE::Inc::prepend(
                $prepend
	   );
	};
	########################################################################
            $win_use
            use Storable qw(nfreeze);
            use FileManip qw(:internal);
            Subtest::ipc()->serve();
        \} };
        $sobj_h{$sindex} = Tharn::subtest( $cmd, '-bg', '-ipc',
                                           '-client', $opts{client},
                                           '-runid', "0${log_index}_FileManip",
                                           '-onexit', sub {},
                                           @extra_options );
        # Save the process id of the process starting subtest
        # so subtest can be properly shutdown
        $main_pid = $$;
    }

    # Call the remote function and get the output
    if ( wantarray ) {
        # Function is expected to return an array
        # so freeze the arguments to call eval() instead of call()
        # to be able to freeze the output of the function
        my $args = join(',', map { freeze($_) } (%opts));
        my $eval_cmd = qq{my \@out = $func_name($args); nfreeze(\\\@out);};
        $out = $sobj_h{$sindex}->ipc()->eval($eval_cmd);
    } else {
        # Function is expected to return a scalar
        $out = $sobj_h{$sindex}->ipc()->call($func_name, %opts);
    }
    if ( $@ ) {
        # Remove 'at (eval ??) line 1' from error message
        # since error will always be at line 1
        $@ =~ s/ at \(eval \d+\) line.*\n//;
        # Check for any errors
        if ( $@ =~ /socket read failed/ ) {
            @results = $sobj_h{$sindex}->own_results;
            $err = NATE::Result::worst(@results);
            croak join("\n", $err->messages) if $err && $err->severity > 2;
        }
        croak $@;
    }

    if ( wantarray ) {
       # The output array is frozen
       my $output = thaw($out);
       return @{$output};
    } else {
       # Plain scalar
       return $out;
    }
}

#---------------------------------------------------------------------
## @name ::read_file
## @description
##  Reads the contents of a file from a client as the user
##  (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = read_file( %options );
##
## @arg file
##  Required: File to be read.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  Contents of file.
#---------------------------------------------------------------------
sub read_file(%) {
    my (%opts) = validate (
        @_,
        {
            file => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_read_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to read a file.
#---------------------------------------------------------------------
sub _read_file(%) {
    my %opts = @_;
    my $oflags = O_RDONLY|O_BINARY;

    # Open the file and read the contents
    sysopen(FH, $opts{file}, $oflags)
      or croak "_read_file: Could not open file for reading $opts{file}: $!";
    my @cont = <FH>;
    close(FH);

    return join('', @cont);
}

#---------------------------------------------------------------------
## @name ::write_file
## @description
##  Writes a file on the client as the user
##  (or any other user if specified).
##  Get the attributes of the file if the option 'fstat' is true,
##  of if any of the setattr options are specified (i.e, mode, uid, gid).
##  Arguments are passed in a "key => value" style.
##
## @synopsis	@attrs = write_file( %options );
##
## @arg file
##  Required: File to be created.
##
## @arg contents
##  Optional: Contents of file.
##
## @arg oflags
##  Optional: Additional open flags (default is 'O_CREAT').
##  The O_WRONLY and O_BINARY open flags are always set.
##
## @arg fstat
##  Optional: Get file attributes after writing (default is 0).
##
## @arg mode
##  Optional: Permission mode to set for file.
##  Octal number or string, e.g., 0644, '644'
##  (default is 0600).
##  Unix clients only.
##
## @arg uid
##  Optional: Change ownership of file to this uid
##  (default is the effective user id of process).
##  This option implies <<user=>'root'>>.
##  Unix clients only.
##
## @arg gid
##  Optional: Change group of file to this gid
##  (default is the effective group id of process).
##  Unix clients only.
##
## @arg offset
##  Optional: Offset where to begin writing (default is 0).
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @arg rtn_hash
##  Optional: If set, then return hashtable instead of array.
##  Default: 0.
##
## @return
##  List of attributes for file as returned by 'stat' if 'fstat' is true,
##  or 1 if 'fstat' is false.
#---------------------------------------------------------------------
sub write_file(%) {
    my (%opts) = validate (
        @_,
        {
            file       => { type => SCALAR },
            contents   => { type => SCALAR,   optional   => 1 },
            oflags     => { type => SCALAR,   default    => 'O_CREAT' },
            fstat      => { type => BOOLEAN,  default    => 0 },
            mode       => { type => SCALAR,   optional   => 1,    regex => qr/^\d+$/ },
            uid        => { type => SCALAR,   optional   => 1,    regex => qr/^\d+$/ },
            gid        => { type => SCALAR,   optional   => 1,    regex => qr/^\d+$/ },
            offset     => { type => SCALAR,   default    => 0,    regex => qr/^\d+$/ },
            rtn_hash   => { type => SCALAR,   optional   => 1 },
            %common_parameters
        }
    );
    # Option uid implies root
    $opts{user} = 'root' if defined($opts{uid});

    do_call(\&_write_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to write a file.
#---------------------------------------------------------------------
sub _write_file(%) {
    my %opts = @_;
    my (@attrs, %attr_set, $err, $ok);
    my $file   = $opts{file};
    my $oflags = O_WRONLY|O_BINARY;

    # Check conflicting options
    if ( defined($opts{uid}) and $opts{user} and $opts{user} ne 'root' ) {
        croak "_write_file: Options 'uid' and 'user' are mutually exclusive " .
              "since specifying option 'uid' requires " .
              "running as root: user=>'root'";
    }

    # Windows does not support chown and chmod
    my $ugmode = ( defined($opts{uid}) or defined($opts{gid})
                   or defined($opts{mode}) );
    my $no_win_opts = ($ugmode || $opts{user});
    my $hosttype = $opts{client}{hosttype};
    if ( $hosttype =~ /^windows/ and $no_win_opts ) {
        croak "_write_file: Options 'user', 'uid', 'gid' or 'mode' are not supported on windows";
    }

    # Append any open flags passed as arguments
    $oflags |= eval "$opts{oflags}" if $opts{oflags};

    # Open the file for writing
    sysopen(FH, $file, $oflags)
      or croak "_write_file: Could not open file for writing $file: $!";

    # Catch any errors to properly close the file
    eval {
        # Set file pointer to correct offset
        seek(FH, $opts{offset}, 0)
          or croak "_write_file: Could not seek on file $file: $!";

        if ( defined($opts{contents}) ) {
            print FH $opts{contents}
              or croak "_write_file: Could not write to file $file: $!";
        }
    };
    $err = $@;

    # Close the file
    $ok = close(FH);
    # Die if an error on the seek or print
    croak "_write_file: $err" if $err;
    # Die if an error on the close
    croak "write_file: Could not close file $file: $!" if !$ok;

    if ( !($opts{fstat} or defined($opts{mode})
      or defined($opts{uid}) or defined($opts{gid})) ) {
        # Only write is requested, and it is fine
        return 1;
    }

    # Set file attributes, if requested.
    $attr_set{mode} = $opts{mode} if defined($opts{mode});
    $attr_set{uid}  = $opts{uid}  if defined($opts{uid});
    $attr_set{gid}  = $opts{gid}  if defined($opts{gid});

    if ( %attr_set ) {
	    # Include 'common_parameters' (user, client, ...)
	    foreach my $parm (@common_params) {
		if (exists($opts{$parm})) {
			$attr_set{$parm} = $opts{$parm};
		}
	    }

	    _setattr_file(file=>$opts{file}, %attr_set);
    }

    @attrs = stat $file or croak "_write_file: Could not get attributes of file $file: $!";
    return(@attrs) unless ($opts{rtn_hash});

    my @keys = qw( dev ino mode nlink uid gid rdev size atime mtime ctime io_size blocks );
    my %ret = ();
    for (0 .. $#keys) {
        $ret{$keys[$_]} = $attrs[$_];
    }
    return( %ret );
}

#---------------------------------------------------------------------
## @name ::stat_file
## @description
##  Get the attributes of a file from a client as the user
##  (or any other user if specified). This is a wrapper for
##  the stat function, which runs both locally or remotely.
##  Arguments are passed in a "key => value" style.
##
## @synopsis	@attr = stat_file( %options );
##
## @arg file
##  Required: File to get attributes.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg rtn_hash
##  Optional: If set, then return hashtable instead of array.
##  Default: 0.
##
## @arg okay_if_not_exists
##   Optional. If '0', then an error is signaled if
##   the file does not exist. If '1', then
##   no error is signaled, and an empty list (hash)
##   is returned.
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  List of attributes for file as returned by 'stat'.
##  If 'rtn_hash' option is set, then return hash instead.
#---------------------------------------------------------------------
sub stat_file(%) {
    my (%opts) = validate (
        @_,
        {
            file                 => { type => SCALAR },
            rtn_hash             => { type => SCALAR,    optional => 1 },
            okay_if_not_exists   => { type => BOOLEAN,   optional => 1 },
            %common_parameters
        }
    );
    do_call(\&_stat_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to get the attributes of a file.
#---------------------------------------------------------------------
sub _stat_file(%) {
    my %opts = @_;
    my $file = $opts{file};
    if (!defined($file) || $file eq "") {
        croak("_stat_file: File is empty string or undefined");
    }

    # Get rid off multiple '/'s
    $file =~ s{/+}{/};
    # Get rid off trailing '/'
    $file =~ s{/$}{};

    # Get the attributes of the file
    if ($opts{okay_if_not_exists} && !-e $file) {
	return ();
    }
    my @attrs = stat($file)
      or croak "_stat_file: Could not get attributes of file $opts{file}: $!";
    return(@attrs) unless ($opts{rtn_hash});
    my @keys = qw( dev ino mode nlink uid gid rdev size atime mtime ctime io_size blocks );
    my %ret = ();
    for (0 .. $#keys) {
        $ret{$keys[$_]} = $attrs[$_];
    }
    return( %ret );
}

#---------------------------------------------------------------------
## @name ::get_file_size
## @description
##  Get the size of a file, as measured by number of bytes
##  from start of file to file to EOF. File 'holes', or 
##  data 'past-EOF' are not accounted for.
##  The current user (or other specified user) must have
##  "read" permission for the file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis	@attr = get_file_size( %options );
##
## @arg file
##  Required: File to get size.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  Size of file (scalar).
#---------------------------------------------------------------------
sub get_file_size(%) {
    my (%opts) = validate (
        @_,
        {
            file       => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_get_file_size, %opts);
}

#---------------------------------------------------------------------
# Internal function to get the size of a file.
#---------------------------------------------------------------------
sub _get_file_size(%) {
    my %opts = @_;
    my $file = $opts{file};
    if (!defined($file) || $file eq "") {
        croak("_get_file_size: File is empty string or undefined");
    }

    # Get rid off multiple '/'s
    $file =~ s{/+}{/};
    # Get rid off trailing '/'
    $file =~ s{/$}{};

    # Get 'read' filehandle for file.
    open(F, "<$file") || die("get_file_size: could not open \"$file\" for reading");
    my $start = tell(F);
    seek(F, 0, 2) || die("get_file_size: could not seek in \"$file\"");
    my $end = tell(F);
    close(F);
    my $size = $end - $start;
    return($size);
}

#---------------------------------------------------------------------
## @name ::test_file
## @description
##
##  Perform '-X' file-tests on file, from a client, where
##  '-X' stands for any of '-r', '-w', '-x', '-d', etc. test flags
##  available thru Perl.
##
##  The 'dash' (-) preceding the test flag letter is not required,
##  but is allowed.
##
##  -- See PerlDoc "-X" entry under catagory:
##     "Functions for filehandles files, or directories"
##     for full list of tests
##
## @synopsis
##  $result = test_file( %options );
##
## @example
##  $is_directory = test_file( file=>"/tmp/foo", test=>'d' );
##  $is_executable = test_file( file=>"/tmp/foo", test=>'x' );
##
## @arg file
##  Required: File on which test will be performed
##
## @arg test
##  Required: Test to perform. Either a single char
##  (e.g. 'x') which represents one of the Perl built-in
##  file tests, or a two-char string, with '-' as the first
##  char (e.g. '-x')
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  1 if test result is 'true', 0 otherwise
#---------------------------------------------------------------------
sub test_file(%) {
    my (%opts) = validate (
        @_,
        {
            file     => { type => SCALAR },
            test     => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_test_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to execute '-X' type tests on a file
#---------------------------------------------------------------------
sub _test_file(%) {
    my %opts = @_;

    # Make sure 'test' has correct format.
    my $file = $opts{file};
    if (!defined($file) || $file eq "") {
        croak("_test_file: File is empty string or undefined");
    }
    my $test = $opts{test};
    if ($test =~ m/^-[A-Za-z]/) {
    } elsif ($test =~ m/^[A-Za-z]/) {
        $test = "-" . $test;
    } else {
      croak("_test_file: Invalid file test name: \"$test\"");
    }

    # Get rid off multiple '/'s
    $file =~ s{/+}{/};
    # Get rid off trailing '/'
    $file =~ s{/$}{};
    # Build $test_expr, then 'eval' it for the test.
    my $test_expr = "($test \"$file\") ?1:0";
    my $rslt = eval($test_expr);
      croak "_test_file: $@" if ($@);
    return $rslt;
}

#---------------------------------------------------------------------
## @name ::get_dir_entries
## @description
##
##  Returns entries (filenames) contained in the specified directory.
##  ('..' and '.' are omitted if present)
##
## @synopsis
##  @entries = get_dir_entries( %options );
##
## @example
##  @entries = get_dir_entries( dir=>"/tmp/MyDir" );
##
## @arg dir
##  Required: Directory to be checked.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  List of directory entries ('..' and '.' are omitted).
#---------------------------------------------------------------------
sub get_dir_entries(%) {
    my (%opts) = validate (
        @_,
        {
            dir     => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_get_dir_entries, %opts);
}

#---------------------------------------------------------------------
# Internal function to return directory entries
#---------------------------------------------------------------------
sub _get_dir_entries(%) {
    my %opts = @_;

    my $dir = $opts{dir};
    if (!defined($dir) || $dir eq "") {
        croak("_get_dir_entries: Dir is empty string or undefined");
    }

    # Get rid off multiple '/'s
    $dir =~ s{/+}{/};
    # Get rid off trailing '/'
    $dir =~ s{/$}{};
    if (! -d $dir) {
        croak("_get_dir_entries: $dir is not a directory");
    }

    my @entries = ();
    opendir(DX, $dir) or
        croak("_get_dir_entries: opendir failed: $!");
    while ( my $fname = readdir(DX) ) {
        next if ($fname eq ".." || $fname eq ".");
        push(@entries, $fname);
    }
    closedir(DX) or
       croak("_get_dir_entries: closedir: $!");
    return @entries;
}

#---------------------------------------------------------------------
## @name ::make_dir
## @description
##
##  Create a directory at the specified location
##
## @synopsis
##  $status = make_dir( %options );
##
## @example
##  make_dir( dir=>"/tmp/MyDir")
##     or die("mkdir);
##  make_dir( dir=>"/tmp/MyDir/MySubdir1/MySubdir2", recursive=>1 )
##     or die("mkdir);
##
## @arg dir
##  Required: Directory to be created.
##
## @arg mode
##   Optional. Permissions 'mode' to use for the
##   new directory. Default is '0777'.
##
## @arg recursive
##  Optional. If '0', then an error is signaled unless
##  the  parent directory to the new directory already
##  exists. If '1', then any 'missing' ancestor 
##  directories of the new directory will be created
##  before creating the new directory. Default is '0'.
##
## @arg okay_if_exists
##   Optional. If '0', then an error is signaled if
##   the directory already exists. If '1', then
##   no error is signaled, and '1' is returned.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  1 if successful.
#---------------------------------------------------------------------
sub make_dir(%) {
    my (%opts) = validate (
        @_,
        {
            dir               => { type => SCALAR },
            mode              => { type => BOOLEAN,   optional => 1,   default => '0777' },
            recursive         => { type => BOOLEAN,   optional => 1 },
            okay_if_exists    => { type => BOOLEAN,   optional => 1 },
            %common_parameters
        }
    );
    do_call(\&_make_dir, %opts);
}

#---------------------------------------------------------------------
# Internal function to create directory
#---------------------------------------------------------------------
sub _make_dir(%) {
    my %opts = @_;
    my $dir = $opts{dir};
    if (!defined($dir) || $dir eq "") {
        croak("_make_dir: Dir is empty string or undefined");
    }
    my $mode = $opts{mode};

    # Get rid off multiple '/'s
    $dir =~ s{/+}{/};
    # Get rid off trailing '/'
    $dir =~ s{/$}{};
    if ( $dir =~ /^.:\\?\*?$/ or $dir =~ m{^/\*?$} ) {
        croak "_make_dir: Not creating root directory";
    }
    my (undef, $parent, undef) = fileparse($dir);
    # Get rid off trailing '/'
    $parent =~ s{/$}{};

    my @entries = ();
    if ( -e $dir) {
        if (-d $dir && $opts{okay_if_exists}) {
            return 1;
        } else {
            croak("_make_dir: $dir alrewady exists");
        }
    }
    if (-e $parent && -d $parent) {
        mkdir($dir)
            or croak("_make_dir: mkdir: $!");
    } elsif (-e $parent && !-d $parent) {
        croak("_make_dir: $parent is not a directory");
    } else {
        if ($opts{recursive}) {
            my %opts_copy = %opts;
            $opts_copy{dir} = $parent;
            _make_dir(%opts_copy) or
               croak("_make_dir recursive call failed: \"$parent\"");
            if (!-d $parent) {
               croak("_make_dir recursive call failed: \"$parent\"");
            }
            mkdir($dir)
               or croak("_make_dir: mkdir: $!");
        } else {
            croak("_make_dir: $parent does not exist");
        }
    }
    return 1;
}

#---------------------------------------------------------------------
## @name ::remove_dir
## @description
##
##  Remove specified directory
##
## @synopsis
##  $status = remove_dir( %options );
##
## @example
##  remove_dir( dir=>"/tmp/MyEmptyDir" )
##      or die("remove_dir");
##  remove_dir( dir=>"/tmp/MyDir", recursive=>1 )
##     or die("remove_dir");
##
## @arg dir
##  Required: Directory to be removed.
##
## @arg recursive
##  Optional. If '0', then only the specified
##  directory is removed. If '1' then the directory,
##  and all its descendents are also deleted.
##  Default is '0'.
##
## @arg contents_only
##  Optional: Boolean. Ignored unless 'recursive' is also set.
##  If '1', then only delete (recursively) the contents of the
##  directory, not the directory itself.
##  (default is 0).
##
## @arg okay_if_not_exists
##   Optional. If '0', then an error is signaled if
##   the directory does not exist. If '1', then
##   no error is signaled, and '1' is returned.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  1 if successful.
#---------------------------------------------------------------------
sub remove_dir(%) {
    my (%opts) = validate (
        @_,
        {
            dir                  => { type => SCALAR },
            recursive            => { type => BOOLEAN,   optional => 1 },
            contents_only        => { type => BOOLEAN,   optional => 1 },
            okay_if_not_exists   => { type => BOOLEAN,   optional => 1 },
            %common_parameters
        }
    );
    do_call(\&_remove_dir, %opts);
}

#---------------------------------------------------------------------
# Internal function to remove directory
#---------------------------------------------------------------------
sub _remove_dir(%) {
    my %opts = @_;

    my $dir = $opts{dir};
    if (!defined($dir) || $dir eq "") {
        croak("_remove_dir: Dir is empty string or undefined");
    }

    # Get rid off multiple '/'s
    $dir =~ s{/+}{/};
    # Get rid off trailing '/'
    $dir =~ s{/$}{};
    # Do not attempt to remove root directory
    if ( $dir =~ /^.:\\?\*?$/ or $dir =~ m{^/\*?$} ) {
        croak "_remove_dir: Not removing root directory";
    }

    if (! -e $dir) {
        if ($opts{okay_if_not_exists}) {
            return 1;
        } else {
            croak("_remove_dir: $dir does not exist");
        }
    } elsif (! -d $dir) {
        croak("_remove_dir: $dir is not a directory");
    } elsif ($opts{recursive}) {
        my %opts_copy = %opts;
        delete($opts_copy{dir});
        delete($opts_copy{okay_if_not_exists});
        $opts_copy{file} = $dir;
        return _remove_file(%opts_copy);
    } else {
        rmdir($dir)
            or croak("_remove_dir: rmdir: $!");
    }
    return 1;
}

#---------------------------------------------------------------------
## @name ::link_file
## @description
##  Create a hard/symbolic link to a file from a client as the user
##  (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = link_file( %options );
##
## @arg file
##  Required: File to be linked to.
##
## @arg flink
##  Required: Hard/symbolic link to create.
##
## @arg symbolic
##  Optional: Boolean, true to create a symbolic link (default is 0).
##  Symbolic links are not supported on Windows.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub link_file(%) {
    my (%opts) = validate (
        @_,
        {
            file     => { type => SCALAR },
            flink    => { type => SCALAR },
            symbolic => { type => BOOLEAN, default => 0 },
            %common_parameters
        }
    );
    do_call(\&_link_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to create a link to a file.
#---------------------------------------------------------------------
sub _link_file(%) {
    my %opts = @_;
    my $file  = $opts{file};
    my $flink = $opts{flink};

    # Windows does not support symbolic links
    my $hosttype = $opts{client}{hosttype};
    if ( $opts{symbolic} and $hosttype =~ /^windows/ ) {
        croak "_link_file: Symbolic links are not supported on Windows";
    }

    # Create hard/symbolic link
    if ( $opts{symbolic} ) {
        symlink $file, $flink
          or croak "_link_file: Could not create symbolic link $flink -> $file: $!";
    } else {
        link $file, $flink
          or croak "_link_file: Could not create hard link $flink -> $file: $!";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::rename_file
## @description
##  Rename the specified file from a client as the user
##  (or as any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = rename_file( %options );
##
## @arg file
##  Required: Original file.
##
## @arg nfile
##  Required: New file.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub rename_file(%) {
    my (%opts) = validate (
        @_,
        {
            file  => { type => SCALAR },
            nfile => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_rename_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to rename a file.
#---------------------------------------------------------------------
sub _rename_file(%) {
    my %opts = @_;
    my $ofile = $opts{file};
    my $nfile = $opts{nfile};

    # Do not attempt to rename root directory
    if ( $ofile =~ /^.:\\?$/ or $ofile =~ m{^/$} ) {
        croak "_rename_file: Not renaming root directory";
    }

    # Get rid off multiple '/'s
    $ofile =~ s{/+}{/};
    # Get rid off trailing '/'
    $ofile =~ s{/$}{};
    # Do not rename file if it is in the root directory on unix,
    # i.e., contains only one '/'.
    my $nslash = ($ofile =~ tr{/}{/});
    if ( $ofile !~ /^.:/ and (($ofile =~ m{^/} and $nslash == 1)
      or getcwd() eq '/' and $nslash == 0) ) {
        croak "_rename_file: Not renaming file located in root directory";
    }

    # Rename file
    rename $ofile, $nfile or croak "_rename_file: Could not rename file $ofile: $!";

    return 1;
}

#---------------------------------------------------------------------
## @name ::remove_file
## @description
##  Removes a file on the client as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = remove_file( %options );
##
## @arg file
##  Required: File to remove.
##
## @arg recursive
##  Optional: Boolean, recursively removes directories and subdirectories.
##  This option allows the use of wild cards (e.g., '*', '?') on the file name
##  (default is 0).
##
## @arg contents_only
##  Optional: Boolean. Ignored unless 'recursive' is also set.
##  If '1', then only delete (recursively) the contents of the
##  directory, not the directory itself.
##  (default is 0).
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @arg okay_if_not_exists
##   Optional. If '0', then an error is signaled if
##   the directory does not exist. If '1', then
##   no error is signaled, and '1' is returned.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub remove_file(%) {
    my (%opts) = validate (
        @_,
        {
            file            => { type => SCALAR },
            recursive       => { type => BOOLEAN,   default => 0 },
            contents_only   => { type => BOOLEAN,   default => 0 },
            okay_if_not_exists   => { type => BOOLEAN,   optional => 1 },
            %common_parameters
        }
    );
    do_call(\&_remove_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to remove a file.
#---------------------------------------------------------------------
sub _remove_file(%) {
    my %opts = @_;
    my $file = $opts{file};


    # Do not attempt to remove root directory
    if ( $file =~ /^.:\\?\*?$/ or $file =~ m{^/\*?$} ) {
        croak "_remove_file: Not removing root directory";
    }

    # Get rid off multiple '/'s
    $file =~ s{/+}{/};
    # Get rid off trailing '/'
    $file =~ s{/$}{};
    # Do not delete file if it is in the root directory on unix,
    # i.e., contains only one '/'.
    my $nslash = ($file =~ tr{/}{/});
    if ( $file !~ /^.:/ and (($file =~ m{^/} and $nslash == 1)
      or getcwd() eq '/' and $nslash == 0) ) {
        croak "_remove_file: Not removing file located in root directory";
    }

    if (! -e $file) {
        if ($opts{okay_if_not_exists}) {
            return 1;
        } else {
            croak("_remove_file: $file does not exist");
        }
    }

    if ( $opts{recursive} ) {
        # Remove a directory tree.
        # It also accepts wild cards ('*', '?', ...)
        # to get a list of top level files and
        # directories and removes them recursively.

        sub rmelem() {
            return if /^\.$/;
	    ((-d $_ && ! -l $_) ? rmdir : unlink)
              or croak "_remove_file: Could not remove file or directory $_: $!";
        }

	sub make_rmelem_no_topdir(@) {
	    my (@topdirs) = @_;
	    my $topdir = $topdirs[0];
	    my $is_list = scalar(@topdirs) > 1;
	        return sub () {
                    return if (/^\.$/);
                    if ($is_list) {
                         return if (grep( {$_ eq $File::Find::name}  @topdirs));
                    } else {
                         return if ($topdir eq $File::Find::name);
                    }
	            ((-d $_ && ! -l $_) ? rmdir : unlink)
                         or croak "remove_file: Could not remove file or directory $File::Find::name: $!";
                 }
        }

        my @list = glob($file);
        if ( @list < 2 and not -e $list[0] ) {
            croak "_remove_file: No such file or directory: $file";
        }

        if ($opts{contents_only}) {
            my $rmelem_sub = make_rmelem_no_topdir(@list);
            finddepth({wanted => \&$rmelem_sub, no_chdir => 1}, @list);
        } else {
            finddepth({wanted => \&rmelem, no_chdir => 1}, @list);
            (rmdir or croak "_remove_file: Could not remove directory $_: $!")
               foreach grep {-d} @list;
        }
    } else {
        # Remove a single file
        unlink $file or croak "_remove_file: Could not remove file $file: $!";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::getattr_file
## @description
##  Get the windows attributes of a file from a client
##  as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = getattr_file( %options );
##
## @arg file
##  Required: File to get attributes for.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  Hash of windows attributes.
##
## @notes
##  Only supported on windows clients.
#---------------------------------------------------------------------
sub getattr_file(%) {
    my (%opts) = validate (
        @_,
        {
            file =>   { type => SCALAR },
            %common_parameters
        }
    );

    do_call(\&_getattr_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to get the attributes of a file.
#---------------------------------------------------------------------
sub _getattr_file(%) {
    my %opts = @_;
    my $file = $opts{file};
    my ($w, %ret);

    # Only windows is supported
    my $hosttype = $opts{client}{hosttype};
    if ( $hosttype !~ /^windows/ ) {
        croak "_getattr_file: Hosttype '$hosttype' is not supported";
    }

    # Get file attributes
    $w = _win32api_call('GetFileAttributes', $file);

    # Constants for the encrypted and sparse file bits are not defined
    # in library so used the number
    %ret = (
        archive  => ($w & Win32API::File::FILE_ATTRIBUTE_ARCHIVE())    ? 1 : 0,
        compress => ($w & Win32API::File::FILE_ATTRIBUTE_COMPRESSED()) ? 1 : 0,
        encrypt  => ($w & 0x00004000)                                  ? 1 : 0,
        hidden   => ($w & Win32API::File::FILE_ATTRIBUTE_HIDDEN())     ? 1 : 0,
        sparse   => ($w & 0x00000200)                                  ? 1 : 0,
    );

    return %ret;
}

#---------------------------------------------------------------------
## @name ::setattr_file
## @description
##  Set the attributes of a file from a client
##  as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = setattr_file( %options );
##
## @arg file
##  Required: File to set attributes for.
##
## @arg size
##  Optional: Set the size (truncate) of file.
##
## @arg uid
##  Optional: Change ownership of file to this uid.
##  This option implies <<user=>'root'>>.
##  Only supported on unix.
##
## @arg gid
##  Optional: Change group of file to this gid.
##  Only supported on unix.
##
## @arg mode
##  Optional: Permission mode to set for file.
##  Octal number or string, e.g., 0644, '644'.
##  Only supported on unix.
##
## @arg compress
##  Optional: Set or clear the compress attribute.
##  Only supported on windows.
##
## @arg encrypt
##  Optional: Set or clear the encrypt attribute.
##  Only supported on windows.
##
## @arg archive
##  Optional: Set or clear the archive attribute.
##  Only supported on windows.
##
## @arg hidden
##  Optional: Set or clear the hidden attribute.
##  Only supported on windows.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user. (Unix clients only)
##
## @return
##  1 if succeeds.
##
## @notes
##  At least one of the following options must be specified:
##  size, uid, gid, mode, encrypt, archive or hidden.
#---------------------------------------------------------------------
sub setattr_file(%) {
    my (%opts) = validate (
        @_,
        {
            file       => { type => SCALAR },
            size       => { type => SCALAR,    optional => 1, regex => qr/^\d+$/ },
            uid        => { type => SCALAR,    optional => 1, regex => qr/^\d+$/ },
            gid        => { type => SCALAR,    optional => 1, regex => qr/^\d+$/ },
            mode       => { type => SCALAR,    optional => 1, regex => qr/^\d+$/ },
            compress   => { type => BOOLEAN,   optional => 1 },
            encrypt    => { type => BOOLEAN,   optional => 1 },
            archive    => { type => BOOLEAN,   optional => 1 },
            hidden     => { type => BOOLEAN,   optional => 1 },
            %common_parameters
        }
    );
    # Option uid implies root
    $opts{user} = 'root' if defined($opts{uid});

    do_call(\&_setattr_file, %opts);
}

#---------------------------------------------------------------------
# Internal function to set the attributes of a file.
#---------------------------------------------------------------------
sub _setattr_file(%) {
    my %opts = @_;
    my $file = $opts{file};
    my $ugmode = ( defined($opts{uid}) or defined($opts{gid})
                   or defined($opts{mode}) );
    my $no_win_opts = ($ugmode || $opts{user});
    my $wattrs = ( defined($opts{encrypt}) or defined($opts{archive})
                   or defined($opts{hidden}) or defined($opts{compress}) );

    # Windows does not support chown and chmod, or 'user' option.
    my $hosttype = $opts{client}{hosttype};
    if ( $hosttype =~ /^windows/ and $no_win_opts ) {
        croak "_setattr_file: Options 'user', 'uid', 'gid' or 'mode' are not supported on windows";
    }

    # Only windows supports the encrypt option
    if ( $hosttype !~ /^windows/ and $wattrs ) {
        croak "_setattr_file: Options 'compress', 'encrypt', 'archive' or 'hidden'" .
              "are only supported on windows";
    }

    # Validate that at least one option is specified
    if ( ! ( defined($opts{size}) or $ugmode or $wattrs ) ) {
        croak "_setattr_file: At least one of these options must be specified:" .
              "size, uid, gid, mode, encrypt, archive or hidden";
    }

    # Check conflicting options
    if ( defined($opts{uid}) and $opts{user} and $opts{user} ne 'root' ) {
        croak "_setattr_file: Options 'uid' and 'user' are mutually exclusive " .
              "since specifying option 'uid' requires " .
              "running as root: user=>'root'";
    }

    # Set the size (truncate) of file
    if ( defined($opts{size}) ) {
        truncate $file, $opts{size}
          or croak "_setattr_file: Could not truncate file $file: $!";
    }

    # Change the mode (permissions) of file
    if ( defined($opts{mode}) ) {
        if ( (~$opts{mode} & $opts{mode}) ne '0' ) {
            # Mode was passed as an octal string,
            # convert it to an octal number.
            $opts{mode} = oct($opts{mode});
        }
        chmod($opts{mode}, $file)
          or croak "_setattr_file: Could not change mode of file $file: $!";
    }

    # Change owner and group on file
    if ( defined($opts{uid}) or defined($opts{gid}) ) {
        # Do not change it if not specified
        my $euid = defined($opts{uid}) ? $opts{uid} : -1;
        my $egid = defined($opts{gid}) ? $opts{gid} : -1;
        # Chown command
        chown($euid, $egid, $file)
          or croak "_setattr_file: Could not change ownership of file $file: $!";
    }

    # Set or clear the encrypt attribute
    if ( defined($opts{encrypt}) ) {
        if ( $opts{encrypt} ) {
            # Set the encrypt attribute on the file
            _win32api_call('EncryptFile', $opts{file});
        } else {
            # Clear the encrypt attribute on the file
            _win32api_call('DecryptFile', $opts{file}, 0);
        }
    }

    # Set or clear the archive or hidden attributes
    if ( defined($opts{archive}) or defined($opts{hidden}) ) {
        # Get file attributes
        $wattrs = _win32api_call('GetFileAttributes', $opts{file});

        if ( defined($opts{archive}) ) {
            # Set or clear the archive bit
            $wattrs = $opts{archive} ?
                      $wattrs | Win32API::File::FILE_ATTRIBUTE_ARCHIVE() :
                      $wattrs & ~Win32API::File::FILE_ATTRIBUTE_ARCHIVE();
        }

        if ( defined($opts{hidden}) ) {
            # Set or clear the hidden bit
            $wattrs = $opts{hidden} ?
                      $wattrs | Win32API::File::FILE_ATTRIBUTE_HIDDEN() :
                      $wattrs & ~Win32API::File::FILE_ATTRIBUTE_HIDDEN();
        }

        # Set file attributes
        _win32api_call('SetFileAttributes', $opts{file}, $wattrs);
    }

    # Set or clear the compress attribute
    if ( defined($opts{compress}) ) {
        # Get file handle of file or directory
        # The FILE_FLAG_BACKUP_SEMANTICS is required to get
        # a handle for a directory
        my $hObject = Win32API::File::createFile($file, 'rwke',
          { Flags => Win32API::File::FILE_FLAG_BACKUP_SEMANTICS() })
          or croak "_setattr_file: Could not get file handle of $file: $^E";
        # Set file to be compressed
        my @ret = (pack('S', $opts{compress}?1:0), 0, [], [], [], []);
        Win32API::File::DeviceIoControl($hObject, 0x9C040, @ret)
          or croak "_setattr_file: Could not set compress attribute on file $file: $^E";
        Win32API::File::CloseHandle($hObject);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::read_stream
## @description
##  Read a stream (named attribute) on file from a client
##  as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = read_stream( %options );
##
## @arg stream
##  Required: Named attribute to read.
##
## @arg file
##  Required: File where named attribute is located.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  Contents of stream.
#---------------------------------------------------------------------
sub read_stream(%) {
    my (%opts) = validate (
        @_,
        {
            stream => { type => SCALAR },
            file   => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_read_stream, %opts);
}

#---------------------------------------------------------------------
# Internal function to read a stream.
#---------------------------------------------------------------------
sub _read_stream(%) {
    my %opts = @_;
    my $stream;
    my $stream_name = $opts{file} . ':' . $opts{stream};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    # Streams or named attributes are only supported
    # on solaris 10 and windows
    if ( $hosttype eq 'unix-solaris' ) {
        # Command to read a named attribute
        $stream = qq(runat $opts{file} "cat $opts{stream}" |);
    } elsif ( $hosttype =~ /^windows/ ) {
        # Command to read a stream
        $stream = qq(<$opts{file}:$opts{stream});
    } else {
        croak "Hosttype '$hosttype' is not supported";
    }

    # Open a stream for reading
    open(FH, $stream)
      or croak "Could not open stream for reading $stream_name: $!";
    my @cont = <FH>;
    close(FH);

    return join('', @cont);
}

#---------------------------------------------------------------------
## @name ::write_stream
## @description
##  Write a stream (named attribute) on file from a client
##  as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = write_stream( %options );
##
## @arg stream
##  Required: Named attribute to write.
##
## @arg file
##  Required: File where named attribute is located.
##
## @arg contents
##  Optional: Contents of stream.
##
## @arg sstat
##  Optional: Get stream attributes after writing (default is 0).
##  Only supported on Solaris 10.
##
## @arg offset
##  Optional: Offset where to begin writing (default is 0).
##  If value is non-zero, the stream must exists.
##  Only supported on Windows.
##
## @arg append
##  Optional: Append the contents to the stream (default is 0).
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  List of attributes for stream if 'sstat' is true,
##  or 1 if 'sstat' is false.
#---------------------------------------------------------------------
sub write_stream(%) {
    my (%opts) = validate (
        @_,
        {
            stream   => { type => SCALAR },
            file     => { type => SCALAR },
            contents => { type => SCALAR, optional => 1 },
            sstat    => { type => BOOLEAN, default => 0 },
            offset   => { type => SCALAR, optional => 1 },
            append   => { type => BOOLEAN, default => 0 },
            %common_parameters
        }
    );
    do_call(\&_write_stream, %opts);
}

#---------------------------------------------------------------------
# Internal function to write a stream.
#---------------------------------------------------------------------
sub _write_stream(%) {
    my %opts = @_;
    my ($open_stream, $err);
    my $file   = $opts{file};
    my $stream = $opts{stream};
    my $append = $opts{append} ? '>>' : '>';
    my $stream_name = $opts{file} . ':' . $opts{stream};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    # Check for valid flags
    if ( $opts{append} and defined($opts{offset}) ) {
        croak "Both append and offset cannot be set";
    }
    if ( $opts{sstat} and $hosttype ne 'unix-solaris' ) {
        croak "Option 'sstat' is only supported on Solaris";
    }

    # Streams or named attributes are only supported
    # on solaris 10 and windows
    if ( $hosttype eq 'unix-solaris' ) {
        if ( defined($opts{offset}) ) {
            croak "Option 'offset' is only supported on Windows";
        }

        # Command to write a named attribute
        $open_stream = qq(| runat $file "cat $append $stream");

        # Open a stream for writing
        open(FH, $open_stream)
          or croak "Could not open stream for writing $stream_name: $!";

        # Catch any errors to properly close the stream
        eval {
            if ( defined($opts{contents}) ) {
                print FH $opts{contents}
                  or croak "Could not write to stream $stream_name: $!";
            }
        };
        $err = $@;

        # Close the pipe
        close(FH) or croak "Could not create stream $stream_name: $!";
        croak $err if $err;
    } elsif ( $hosttype =~ /^windows/ ) {
        # Using Win32API to create a stream since
        # perl open or sysopen cannot open a
        # stream on the current directory '.'
        my $count;
        my $offset = 0;
        my $nbytes = length($opts{contents});
        my $foffset = $opts{offset};
        # Open flags wke: write keep existing
        #            wtc: write truncate create
        my $access = ($opts{append} or defined($opts{offset})) ? 'wke' : 'wtc';
        $file = '' if $file eq '.';
        $open_stream = qq($file:$stream);

        # Open stream for writing
        my $hObject = Win32API::File::createFile($open_stream, $access)
          or croak "Could not open stream for writing $stream_name: $^E";

        # Catch any errors to properly close the stream
        eval {
            # Request for append or set to specific offset
            my @seek_args = $opts{append} ? (0, 2) : $foffset ?
                            ($foffset > $VLOFFSET ?
                             Math::BigInt->new($foffset) : $foffset, 0) : ();
            if ( @seek_args ) {
                # Position file pointer
                Win32API::File::setFilePointer($hObject, @seek_args )
                  or croak "Could not position filer pointer $stream_name: $^E";
            }

            while ( $offset < $nbytes ) {
                # Try to write whole buffer
                Win32API::File::WriteFile($hObject, $opts{contents},
                                          0, $count, [])
                  or croak "Could not write to stream $stream_name: $^E";
                $offset += $count;
                $opts{contents} = substr($opts{contents}, $count);
            }
        };
        $err = $@;

        # Close the stream
        Win32API::File::CloseHandle($hObject);
        croak $err if $err;
    } else {
        croak "Hosttype '$hosttype' is not supported";
    }

    # Get stream attributes
    if ( $opts{sstat} ) {
        my @attr = _stat_stream(stream => $stream,
                                file   => $file,
                                client => $opts{client});
        return @attr;
    } else {
        # Only write is requested
        return 1;
    }
}

#---------------------------------------------------------------------
## @name ::stat_stream
## @description
##  Get the attributes of a stream (named attribute) from a
##  client as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = stat_stream( %options );
##
## @arg stream
##  Required: Named attribute.
##
## @arg file
##  Required: File where named attribute is located.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  List of attributes for stream.
##
## @notes
##  Currently only supported on Solaris 10.
#---------------------------------------------------------------------
sub stat_stream(%) {
    my (%opts) = validate (
        @_,
        {
            stream => { type => SCALAR },
            file   => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_stat_stream, %opts);
}

#---------------------------------------------------------------------
# Internal function to get the attributes of a stream.
#---------------------------------------------------------------------
sub _stat_stream(%) {
    my %opts = @_;
    my $file   = $opts{file};
    my $stream = $opts{stream};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    # Streams or named attributes are only
    # supported on solaris 10 for getattr
    if ( $hosttype ne 'unix-solaris' ) {
        croak "Hosttype '$hosttype' is not supported";
    }

    # Command to getattr a named attribute
    my $stat = qq{my \@attr = stat q($stream) or die \$!; } .
               qq{print q(--sstat_ok--) . join q(--), \@attr;};
    # Run command to get the attributes
    my $out = qx(runat $file "perl -e '$stat'" 2>&1);

    # Remove the 'sstat_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($out =~ s/--sstat_ok--//) ) {
        croak "Could not get attributes for stream '$stream' " .
              "on file '$file': $out";
    }

    return split(/--/, $out);
}

#---------------------------------------------------------------------
## @name ::rename_stream
## @description
##  Rename the specified stream (named attribute) from a
##  client as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = rename_stream( %options );
##
## @arg stream
##  Required: Original named attribute.
##
## @arg nstream
##  Required: New named attribute.
##
## @arg file
##  Required: File where named attribute is located.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  1 if succeeds.
##
## @notes
##  Currently only supported on Solaris 10.
#---------------------------------------------------------------------
sub rename_stream(%) {
    my (%opts) = validate (
        @_,
        {
            stream  => { type => SCALAR },
            nstream => { type => SCALAR },
            file    => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_rename_stream, %opts);
}

#---------------------------------------------------------------------
# Internal function to rename a stream.
#---------------------------------------------------------------------
sub _rename_stream(%) {
    my %opts = @_;
    my $file    = $opts{file};
    my $stream  = $opts{stream};
    my $nstream = $opts{nstream};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    # Streams or named attributes are only
    # supported on solaris 10 for rename
    if ( $hosttype ne 'unix-solaris' ) {
        croak "Hosttype '$hosttype' is not supported";
    }

    # Command to rename a named attribute
    my $rename = qq{rename q($stream), q($nstream) or die \$!; } .
                 qq{print q(--srename_ok--);};
    # Run command to rename the named attribute
    my $out = qx(runat $file "perl -e '$rename'" 2>&1);

    # Remove the 'srename_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($out =~ s/--srename_ok--//) ) {
        croak "Could not rename stream '$stream' on file '$file': $out";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::remove_stream
## @description
##  Remove the specified stream (named attribute) from a
##  client as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = remove_stream( %options );
##
## @arg stream
##  Required: Named attribute.
##
## @arg file
##  Required: File where named attribute is located.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub remove_stream(%) {
    my (%opts) = validate (
        @_,
        {
            stream => { type => SCALAR },
            file   => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_remove_stream, %opts);
}

#---------------------------------------------------------------------
# Internal function to remove a stream.
#---------------------------------------------------------------------
sub _remove_stream(%) {
    my %opts = @_;
    my $out;
    my $file   = $opts{file};
    my $stream = $opts{stream};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    if ( $hosttype eq 'unix-solaris' ) {
        # Command to remove a named attribute on solaris
        my $remove = qq{unlink q($stream) or die \$!; } .
                     qq{print q(--sremove_ok--);};
        # Run command to remove the named attribute
        $out = qx(runat $file "perl -e '$remove'" 2>&1);
    } elsif ( $hosttype =~ /^windows/ ) {
        # Command to remove a stream on windows
        $file = '' if $file eq '.';
        $out = Win32API::File::DeleteFile("$file:$stream")
               ? '--sremove_ok--' : $^E;
    } else {
        croak "Hosttype '$hosttype' is not supported";
    }

    # Remove the 'sremove_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($out =~ s/--sremove_ok--//) ) {
        croak "Could not remove stream '$stream' on file '$file': $out";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::setattr_stream
## @description
##  Set the attributes of a stream (named attribute) from a client
##  as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = setattr_stream( %options );
##
## @arg stream
##  Required: Named attribute.
##
## @arg file
##  Required: File where named attribute is located.
##
## @arg size
##  Required: Set the size (truncate) of stream.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub setattr_stream(%) {
    my (%opts) = validate (
        @_,
        {
            stream => { type => SCALAR },
            file   => { type => SCALAR },
            size   => { type => SCALAR, regex => qr/^\d+$/ },
            %common_parameters
        }
    );
    do_call(\&_setattr_stream, %opts);
}

#---------------------------------------------------------------------
# Internal function to set the attributes of a stream.
#---------------------------------------------------------------------
sub _setattr_stream(%) {
    my %opts = @_;
    my $stream = $opts{stream};
    my $file   = $opts{file};
    my $size   = $opts{size};

    # Streams or named attributes are only supported
    # on solaris 10 and windows
    my $hosttype = $opts{client}{hosttype};
    if ( $hosttype eq 'unix-solaris' ) {
        # Command to set attributes on a named attribute
        my $setattr = qq{truncate q($stream), $size or die \$!; } .
                     qq{print q(--ssetattr_ok--);};
        my $out = qx(runat $file "perl -e '$setattr'" 2>&1);
        # Remove the 'ssetattr_ok' flag from the output
        # and if it does not exist croak the error
        if ( ! ($out =~ s/^--ssetattr_ok--//) ) {
            croak "Could not set attributes for stream " .
                  "'$stream' on file '$file': $out";
        }
    } elsif ( $hosttype =~ /^windows/ ) {
        # Command to set attributes on a stream
        $file = '' if $file eq '.';
        truncate qq($file:$stream), $size
          or croak "Could not truncate stream '$stream' on file '$file': $!";
    } else {
        croak "Hosttype '$hosttype' is not supported";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::write_sparse
## @description
##  Write a sparse (holey) file from a client
##  as the user (or any other user if specified).
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = write_sparse( %options );
##
## @arg file
##  Required: File to be created.
##
## @arg data
##  Optional: Reference to an array of hashes having the following:
##  { offset => Offset of data, data => Actual data }.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @arg user
##  Optional: Perform operation as this user.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub write_sparse(%) {
    my (%opts) = validate (
        @_,
        {
            file => { type => SCALAR },
            data => { type => ARRAYREF },
            %common_parameters
        }
    );
    do_call(\&_write_sparse, %opts);
}

#---------------------------------------------------------------------
# Internal function to write a sparse file.
#---------------------------------------------------------------------
sub _write_sparse(%) {
    my %opts = @_;
    my ($item, $err, $foffset);
    my $file = $opts{file};
    my $data = $opts{data};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    if ( $hosttype =~ /^unix/ ) {
        # Open file
        sysopen(FH, $file, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY)
          or croak "Could not open file for writing $file: $!";

        # Catch any errors to properly close the file
        eval {
            # Write the data to the file
            foreach $item ( @$data ) {
                # Set file pointer to correct offset for this data block
                seek(FH, $item->{offset}, 0)
                  or croak "Could not seek on file $file: $!";
                print FH $item->{data}
                  or croak "Could not write to file $file: $!";
            }
        };
        $err = $@;

        # Close the file
        close(FH);
        croak $err if $err;
    } elsif ( $hosttype =~ /^windows/ ) {
        my ($zinfo, $zlen);
        my $offset = 0;
        my ($off_l, $off_h, $foff_l, $foff_h);
        my @ret = ([], [], [], [], [], []);

        # Open file for writing
        my $hObject = Win32API::File::createFile($file, 'wtc')
          or croak "Could not open file for writing $file: $^E";
        # Catch any errors to properly close the file
        eval {
            # Set file to be sparse
            Win32API::File::DeviceIoControl($hObject, 0x900C4, @ret)
              or croak "Could not set sparse attribute on file $file: $^E";

            # Set all zero regions and write all data
            foreach $item ( @$data ) {
                $foffset = $item->{offset};
                $foffset = Math::BigInt->new($foffset) if $foffset > $VLOFFSET;
                $zlen = $foffset - $offset;
               # XXX ONTAP does not support NTFS sparse file IOCTLs
               # See Burt 20305
               # if ( $zlen ) {
               #     ($off_h, $foff_h) = (0, 0);
               #     $off_l  = $offset;
               #     $foff_l = $foffset;
               #     # Convert cummulative offset to a BitInt
               #     # if larger than 32 bits
               #     $offset = Math::BigInt->new($offset)
               #       if (!ref $offset && $offset > $VLOFFSET);
               #     if ( ref $offset ) {
               #         # Start of zero region is larger than 32 bits,
               #         # Separate into 32 bit low and high parts
               #         $off_l = ($offset & $VLOFFSET)->numify;
               #         $off_h = ($offset >> 32)->numify;
               #     }
               #     if ( ref $foffset ) {
               #         # End of zero region is larger than 32 bits,
               #         # Separate into 32 bit low and high parts
               #         $foff_l = ($foffset & $VLOFFSET)->numify;
               #         $foff_h = ($foffset >> 32)->numify;
               #     }
               #     # Pack the FILE_ZERO_DATA_INFORMATION structure
               #     $zinfo = pack('LLLL', $off_l, $off_h, $foff_l, $foff_h);
               #     @ret[0,1] = ($zinfo, 16);
               #     # Set zero region
               #     Win32API::File::DeviceIoControl($hObject, 0x980C8, @ret)
               #       or croak "Could not set zero region on $file: $^E";
               # }
                # Move file pointer where actual data is located
                Win32API::File::setFilePointer($hObject, $foffset, 0)
                  or croak "Could not seek on file $file: $^E";
                # Write the actual data
                Win32API::File::WriteFile($hObject, $item->{data}, 0, [], [])
                  or croak "Could not write to file $file: $^E";

                $offset = $foffset + length($item->{data});
            }
        };
        $err = $@;

        # Close the file
        Win32API::File::CloseHandle($hObject);
        croak $err if $err;
    } else {
        croak "Hosttype '$hosttype' is not supported";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::vol_info
## @description
##  Get volume information.
##
## @synopsis
##  $op = vol_info( %options );
##
## @arg vol
##  Required: Root directory of volume.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @return
##  Hash of volume attributes.
##
## @notes
##  Only supported on Windows.
#---------------------------------------------------------------------
sub vol_info(%) {
    my (%opts) = validate (
        @_,
        {
            vol => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_vol_info, %opts);
}

#---------------------------------------------------------------------
# Internal function to get volume information
#---------------------------------------------------------------------
sub _vol_info(%) {
    my %opts = @_;
    my ($out, $flags, %ret);
    my $nsize    = 256;
    my $volname  = ' ' x $nsize;
    my $serial   = ' ' x 4;
    my $maxlen   = ' ' x 4;
    my $sysflags = ' ' x 4;
    my $sysname  = ' ' x $nsize;

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    if ( $hosttype !~ /^windows/ ) {
        croak "Hosttype '$hosttype' is not supported";
    }

    # Append '\' to the volume, it is required
    $opts{vol} .= '\\' if $opts{vol} !~ /\\$/;

    # Call the windows api
    $out = _win32api_call(
        'GetVolumeInformation',
        $opts{vol},
        \$volname, $nsize,
        \$serial,
        \$maxlen,
        \$sysflags,
        \$sysname, $nsize,
    );

    croak "Could not get volume information for $opts{vol}" unless $out;

    # Process output strings
    $volname =~ s/\0.*//;
    $sysname =~ s/\0.*//;
    # Convert sysflags to an integer for further processing
    $flags = unpack('I', $sysflags);

    # Create return hash
    %ret = (
        volname  => $volname,
        sysname  => $sysname,
        serial   => unpack('I', $serial),
        maxlen   => unpack('I', $maxlen),
        sysflags => {
            sparse   => ($flags & 0x000040) ? 1 : 0,
            encrypt  => ($flags & 0x020000) ? 1 : 0,
            compress => ($flags & 0x000010) ? 1 : 0,
            acls     => ($flags & 0x000008) ? 1 : 0,
            quotas   => ($flags & 0x000020) ? 1 : 0,
            case_sensitive_search => ($flags & 0x000001) ? 1 : 0,
            case_preserved_names  => ($flags & 0x000002) ? 1 : 0,
        },
        sysflags_raw => $flags,
    );

    return %ret;
}

#---------------------------------------------------------------------
## @name ::vol_free
## @description
##  Get free space of volume.
##
## @synopsis
##  $op = vol_free( %options );
##
## @arg vol
##  Required: Root directory of volume, e.g., 'C:', 'Z:', etc.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type
##  (default is localhost).
##
## @return
##  Hash of information about the amount of free, used and total
##  space available on the volume
##
## @notes
##  Only supported on Windows.
#---------------------------------------------------------------------
sub vol_free(%) {
    my (%opts) = validate (
        @_,
        {
            vol => { type => SCALAR },
            %common_parameters
        }
    );
    do_call(\&_vol_free, %opts);
}

#---------------------------------------------------------------------
# Internal function to get free space of volume
#---------------------------------------------------------------------
sub _vol_free(%) {
    my %opts = @_;
    my ($hObject, $drive, %ret, $out, $err);
    my $free_avail  = ' ' x 8;
    my $total_avail = ' ' x 8;
    my $free_total  = ' ' x 8;
    my $vol = $opts{vol};

    # Get hosttype of client
    my $hosttype = $opts{client}{hosttype};

    if ( $hosttype !~ /^windows/ ) {
        croak "Hosttype '$hosttype' is not supported";
    }

    if ( $vol !~ /^([a-zA-Z]):/ ) {
        croak "Invalid volume name $vol";
    }
    $drive = $1;

    # Call the windows api to get the free space
    $out = _win32api_call(
        'GetDiskFreeSpaceEx',
        "$drive:\\",
        \$free_avail,
        \$total_avail,
        \$free_total,
    );

    croak "Could not get volume free space for $vol" unless $out;

    # Calculate free, used and total space from 64 bit outputs
    %ret = (
        free_avail  => unpack('I', substr($free_avail,  0, 4)) +
                       unpack('I', substr($free_avail,  4, 4)) * 4294967296,
        total_avail => unpack('I', substr($total_avail, 0, 4)) +
                       unpack('I', substr($total_avail, 4, 4)) * 4294967296,
        free        => unpack('I', substr($free_total,  0, 4)) +
                       unpack('I', substr($free_total,  4, 4)) * 4294967296,
    );

    # Set the total to the total space available for the current user
    $ret{total} = $ret{total_avail};
    # Calculate used space
    $ret{used} = $ret{total} - $ret{free};

    return %ret;
}

#---------------------------------------------------------------------
# Internal function to load the specified Win32::API function and run it
#---------------------------------------------------------------------
sub _win32api_call($@) {
    my ($name, @args) = @_;
    my ($arg, $out, $index, @args_ref);

    # Function prototypes
    my %api = (
        EncryptFile       => ['advapi32', 'EncryptFile',       'P',  'N'],
        DecryptFile       => ['advapi32', 'DecryptFile',       'PN', 'N'],
        GetFileAttributes => ['kernel32', 'GetFileAttributes', 'P',  'N'],
        SetFileAttributes => ['kernel32', 'SetFileAttributes', 'PN', 'N'],
        GetDiskFreeSpaceEx   => ['kernel32', 'GetDiskFreeSpaceEx', 'PPPP', 'N'],
        GetVolumeInformation => ['kernel32', 'GetVolumeInformation',
                                 'PPNPPPPN', 'N'],
    );

    # Find which arguments are references
    $index = 0;
    foreach $arg ( @args ) {
        if ( ref $arg ) {
            # Argument is a reference, save it and its position
            push @args_ref, { arg => $arg, idx => $index };

            # Replace argument in list with its contents
            $arg = $$arg;
        }
        $index++;
    }

    if ( $use_w32api ) {
        # Load module only once
        require Win32::API;
        $use_w32api = 0;
    }

    if ( ! defined($win32api{$name}) ) {
        # Load the DLL and import the Windows API function
        $win32api{$name} = Win32::API->new(@{$api{$name}});
        if ( ! defined($win32api{$name}) ) {
            croak "Can not import API $api{$name}: $!";
        }
    }

    # Call the API
    $out = $win32api{$name}->Call(@args) or croak "$name: $^E";

    foreach $arg ( @args_ref ) {
        # Replace contents of reference with new value of argument after
        # calling the API so the calling function can see this change
        ${$arg->{arg}} = $args[$arg->{idx}];
    }

    return $out;
}

#---------------------------------------------------------------------
# Internal function to freeze a variable
#---------------------------------------------------------------------
sub freeze {
    my ($thing) = @_;
    my $dumper = new Data::Dumper([$thing]);
    $dumper->Purity(1)->Deepcopy(1)->Indent(0)->Terse(1);
    return $dumper->Dump();
}

# Shutdown the rpc server(s) after module
# is out of scope
END {
    # Shutdown all subtests started by this process
    if ( $_use_nate && $main_pid == $$ ) {
        # Shutdown rpc server(s)
        $sobj_h{$_}->ipc()->quit() foreach keys %sobj_h;
        # Wait for subtest(s) to finish
        eval {
            Subtest::wait_finish(subtests => [ values %sobj_h ], timeout => 15);
        };
    }
    # Make sure subtest does not report an error to parent process
    exit 0;
};

1;

