# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Filer/Decore.pm#1 $

## @summary Operations on filer cores
## @author dl-central-libs
## @status Shared
## @pod here


=head1 NAME

Filer::Decore - routines to help with moving core files from filers

=head1 SYNOPSIS

   use Filer::Decore qw(wait_for_savecore core_dir find_cores core_info_nz gdb_on_core root_chmod);

   # filer and simulator cores:

   wait_for_savecore(filer => $filer);
   $core_dir = core_dir(filer => $filer);
   @cores = find_cores(core_dir => $core_dir);
   root_chmod(mode => "0444", filelist => [@cores]);
   $output = gdb_on_core(corefile => $core, commands => ["bt"]);

   # filer cores only:

   my $info = core_info_nz(corefile => $core);

=head1 DESCRIPTION

This library contains routines which assist in moving core files from
filers and in using gdb to examine them.  It supports ONTAP-C in both
filer and netcache flavors, in both simulator and hardware form.

Use Filer/decore.thpl to test this library.

=head1 METHODS

=cut


package Filer::Decore;

use vars ('@ISA','@EXPORT');

use base qw(Exporter);
our @EXPORT_OK = qw(wait_for_savecore core_dir find_cores core_info_nz gdb_on_core root_chmod);


use strict;
use warnings;

use Tharn qw(host connect say script_result);
use IPC::Open3 qw(open3);
use mount qw(mount);
use Carp qw(confess);
use Params::Validate qw(validate SCALAR ARRAYREF);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}
use Volume qw(vol_root);
use File::Temp qw(tempfile);
use Admin::License qw(license);

# for wait_for_savecore, below

my @Re_saving = (
                 "Preparing to save (local |partner |)core",
                 "Waiting for partner dump to complete\.",
                 # (AnchorSteam and later)
                 "Saving core\\.\\S+\\.nz:",
                 "Generating mini-core",
                 # (Early AnchorSteam)
                 "Saving core: ",
                 # (Krakus, Fullsail)
                 "Saving (local |partner |)core file",
                 "Generating minicore",
                 );

my @Re_done = (
               "Previous attempts to save core failed",
               "No core needs to be saved",
               "No savecore in progress",
               # For simulators, AnchorSteam and later
               "savecore is not supported for this platform",
               # For simulators, Scrimshaw
               "savecore is not supported in simulator",
               );

=head2 wait_for_savecore

  wait_for_savecore(filer => $filer)

Wait for the appliance to finish saving core (as well as any extra
steps in that process such as generating and sending mini cores).

Arguments:

=over

=item filer

(Required): A host record for the appliance (as returned by C<host>)

=back

=cut

sub wait_for_savecore {
    my (%opts) = validate(@_, {
        filer => {isa => 'Hostrec'},
    });

    my $host = host($opts{filer});
    my $conn = connect($host->id() . "#rsh");

    while(1) {
        my $match = say($conn, "priv set -q advanced; savecore -s",
                        @Re_saving, @Re_done, '-eval_error');
        die $@ if $@;
        return 1 if (grep { $match =~ $_ } @Re_done);
        sleep(30);
    }
}

=head2 core_dir

   core_dir(filer => $filer);

Arrange for the directory where the given appliance stores cores to be
accessible, and return a path to it.

(For simulators, this just means returning simdir---the directory
where the simulator is run---since unix coredumps will appear there.
For filer or cache hardware, this means mounting over NFS the
directory on the root volume where it stores cores)

The routine takes the following named arguments

=over

=item filer

(Required): A host record for the appliance (as returned by C<host>)

=back

=cut

sub core_dir {
    my (%opts) = validate(@_, {
        filer => {isa => 'Hostrec'},
    });

    my $host = host($opts{filer});

    my $simdir = $host->simdir();
    if ($simdir) {
        return $simdir;
    }

    my $conn = connect($host->id() . "#rsh");

    my $export_point;
    if ($host->hosttype() =~ /^cache/) {
        _netcache_enable_nfs($conn);
        $export_point = "/crash";
    } else {
        # XXX - dont we have to enable NFS here as well ?
        $export_point = "/vol/" . vol_root($conn) . "/etc/crash";
    }

    my $mountpoint = mount ($host->default_ip(),
                            $export_point)
        or die $@;

    return $mountpoint;
}


=head2 find_cores

   find_cores(core_dir => $core_dir);

Scan the given core directory and return the names of files which seem
to be core dumps (based on filenames).

The routine takes the following named arguments

=over

=item core_dir

(Required): A directory containing cores, such as that returned by C<core_dir>, above.

=back

=cut

sub find_cores {
    my (%opts) = validate(@_, {
        core_dir => {type => SCALAR},
    });
    my $core_dir = $opts{core_dir};

    opendir(DIR,$core_dir)
        or die "Could not open directory $core_dir\: $!\n";
    my @filenames = readdir(DIR);
    closedir(DIR)
        or die "Could not close directory $core_dir\: $!\n";

    my @cores = grep { $_ eq "core" || $_ =~ /^core\./ } @filenames;

    # Skip "core.tmp.nz", which is a temporary file the filer uses
    # when dealing with mini-cores
    @cores = grep { $_ ne "core.tmp.nz" } @cores;

    return map { "$core_dir/$_" } @cores;
}

=head2 core_info_nz

  $info = core_info_nz(corefile => $corefile)

return a hash reference containing information about the given core
file.

Unlike most routines in this library, this cannot be used with
simulator (native unix) core dumps, only netapp core dumps.  (only
netapp core dumps have this kind of metadata readily available without
the need of a debugger like gdb)

This relies on the program "coretool", which must be in
/usr/software/rats/bin.  See "//depot/tools/main/coretool/..." in perforce.

Fields in $info:

=over

=item $info->{arch}

The processor type.  X86, ALPHA, MIPS, or AMD64.

=item $info->{full_panic}

The panic string for the core

=item $info->{panic}

The panic string from the core, with any suffix like "on release X"
removed from the message.

=item $info->{time}

The panic time for the core.

=item $info->{version}

The version string from ontap.  (The version is usually listed twice
by coretool.  If it matters, this is taken from "version=X" in the
IDENT_INFO section, rather than "on release X" in the panic message)

=item $info->{cc_flags}

The flags passed to the build command, such as "3d".  (Some other NATE
libraries in a different context call this value the "maytag suffix")

=item $info->{serialnumber}

The system serial number (such as "228663")

=item $info->{nvram_systemid}

The system id, traditionally stored in the nvram card.  (such as
"0016794837")

=item $info->{model}

The appliance model name (such as "F720")

=item $info->{appliance}

The appliance type (such as "filer")

=back

=cut

my %core_info_nz_re = (
        arch => qr/^processor type\s+CPU_(\S+)/m,
        full_panic => qr/^PANIC MESSAGE:\s+(.*?)\s+PANIC TIME:/ms,
        time => qr/^PANIC TIME:\s+(.*?)\s+REGISTERS:/ms,
        version => qr/^IDENT INFO:\s+version=(.*?)\n/ms,
        cc_flags => qr/^cc_flags=(\S+)/m,
        serialnumber => qr/^serialnumber=(\w+)/m,
        nvram_systemid => qr/^nvram_systemid=(\w+)/m,
        model => qr/^model=(\w+)/m,
        appliance_type => qr/^appliance=(\w+)/m,
);

sub core_info_nz {
    my (%opts) = validate(@_, {
        corefile => {type => SCALAR},
    });
    my $corefile = $opts{corefile};
    my $info = {};
    my @cmd = ("/usr/software/rats/bin/coretool",$corefile);
    my $data = _qx_list(@cmd);
    if ($?) {
        die(join(" ",@cmd) . " failed with exit status $?\n".
            "output follows:\n".
            "$data\n");
    }
    while (my ($key, $regex) = each %core_info_nz_re) {
        my ($match) = ($data =~ $regex);
        if (defined($match)) {
            $info->{$key} = $match;
        }
    }

    if (defined($info->{full_panic})) {
        $info->{panic} = $info->{full_panic};
        $info->{panic} =~ s/on release .*//;
    }

    return $info;
}

# given the simulator core file, return the full path to the
# corresponding maytag image.

# the argv[] of the program is apparently accessible to gdb even
# without a symbol table, and we know already for simulators that gdb
# will be native (assumed to be /usr/software/bin/gdb), so use gdb to
# extract argv[0] of the dead simulator, which is the path to the
# maytag they ran.

# one nice side effect is that attempts to get simulator core dumps
# from native, non-simulator cores will probably work (to the extent
# that other executables on the system actually have symbols, etc).

sub _guess_image_from_sim_core {
    my ($gdb,$corefile) = @_;
    my $output = _run_gdb(gdb => "/usr/software/bin/gdb",
                          commands => ["core-file $corefile"]);
    my ($path) = ($output =~ /Core was generated by \`(\S+)[^\']*\'\./);
    if (!$path) {
        die("Could not find 'Core was generated by' in output of gdb.\n".
            "/usr/software/bin/gdb output follows:\n".
            "$output\n");
    }
    return $path;
}

# given core_info_nz() values for a netapp core, return the full path
# to the corresponding maytag image.

# for private builds (determined by the version beginning with a "/"),
# it's easy: the ontap directory is part of the version string.

# for public builds, we guess that the build is found under
# /x/eng/rlse.

sub _guess_image_from_info_nz {
    my ($corefile, $info) = @_;
    foreach my $field qw(version cc_flags) {
        if (!$info->{$field}) {
            die "Cannot guess ontap image path: could not find \"$field\" field in coretool output for core file $corefile";
        }
    }
    my ($short_version) = ($info->{version} =~ /NetApp Release ([^:]+):/);
    if (!$short_version) {
        die "Cannot guess short version of version \"" . $info->{version} . "\": version did not match /NetApp Release (.*):/";
    }
    my $ontapdir;
    if ($short_version =~ m@^/@) {
        # version seems to be private build, build daemon build, etc.
        $ontapdir=$short_version;
    } else {
        $ontapdir="/x/eng/rlse/" . $short_version;
    }
    return "$ontapdir/prod/build/maytag." . $info->{cc_flags};
}



# for _guess_gdb_from_info_nz, below

my %nz_gdb_map = (
                "X86" => "i386",
                "ALPHA" => "alpha",
                "MIPS" => "mips",
                "AMD64" => "amd64",
                );


# given core_info_nz() values for a netapp core, return the
# conventional path to the gdb that handles it.

sub _guess_gdb_from_info_nz {
    my ($info) = @_;
    my $arch = $info->{arch};
    my $prefix = $nz_gdb_map{$arch};
    if (!$prefix) {
        die "Could not guess gdb pathname for core architecture \"$arch\"\n";
    }
    return "/usr/local/bin/${prefix}-netapp-gdb";
}


# run the given gdb with the given extra arguments, and return its
# output.  (The commands are put into a temporary file, and the -batch
# and -x options are passed to the gdb to have it perform those commands)

sub _run_gdb {
    my (%opts) = validate(@_, {
        gdb => {type => SCALAR},
        args => {type => ARRAYREF, default => []},
        commands => {type => ARRAYREF},
    });
    my $gdb = $opts{gdb};
    my @args = @{$opts{args}};
    my @commands = @{$opts{commands}};
    my ($tempfh, $tempfilename) = tempfile(UNLINK => 0);
    my $retval = eval {
        my $cmdtext = join("\n",@commands) . "\n";
        syswrite($tempfh,$cmdtext)
            or die "Could not write to tempfile \"$tempfilename\": $!";
        close($tempfh)
            or die "Could not close tempfile \"$tempfilename\": $!";
        my $output = _qx_list($gdb,"-batch","-x",$tempfilename,@args);
        if ($? != 0) {
            die ("gdb exited with status $?\n".
                 "gdb batch file was:\n".
                 "    " . join("\n    ",@commands) . "\n".
                 "$gdb output was:\n".
                 "$output\n");
        }
        return $output;
    };
    unlink($tempfilename);
    die $@ if $@;
    return $retval;
}

=head2 gdb_on_core

  $output = gdb_on_core(corefile => $corefile, commands => \@commands);

  ($output,$image) = gdb_on_core(corefile => $corefile, commands => \@commands);

Start gdb with core file C<$corefile>, and have gdb run the given
C<@commands> in C<-batch> mode.  In a scalar context, return the output of
gdb.  In an array context, the second return value will be a path to
the executable (maytag image) which was used.

This will guess the executable and gdb paths to use based on values
obtained from the core file.  For netapp (non-simulator) cores, the
path to gdb and path to the ontap image are guessed from C<version> and
C<cc_flags> fields determined by F</usr/software/rats/bin/coretool>.  For
simulator cores, the gdb image is known to be F</usr/software/bin/gdb>
and that is used to retrieve the executable path used when the
simulator was invoked.  Netapp and simulator cores are distinguished
by assuming the former end in ".nz".

gdb combines the output of all commands in C<@commands> into a single
output stream, with no explicit separation between the output of
different commands, and there is some output emitted by gdb before the
output of the first user-provided command.  If you wish to extract the
exact output of a particular command, consider adding appropriate
"print" or "printf" calls to add markers around that command.

This will die with an error if gdb exits with nonzero status (such as
if gdb could not be executed).  It will not die if commands fail.

Requirements for hardware filers:

=over

=item *

The program "coretool" must be in /usr/software/rats/bin.  See
"//depot/tools/main/coretool/..." in perforce.

=item *

The appropriate cross-debugging gdb for this architecture
(i386-netapp-gdb, alpha-netapp-gdb, mips-netapp-gdb, amd64-netapp-gdb)
must be in /usr/software/bin.

=back

Requirements for simulators:

=over

=item *

The host running these routines must be native to the simulator
(for example, it must run on linux for a maytag.L simulator)

=item *

gdb for debugging simulators must be in /usr/software/bin/gdb.

=item *

The path used to run the simulator will be the path used by gdb to get
symbol information.  Thus, avoid symlinks in the path whose value is
expected to change over time (such as /rlse/mainN).

=back

=cut

sub gdb_on_core {
    my (%opts) = validate(@_, {
        corefile => {type => SCALAR},
        commands => {type => ARRAYREF},
    });
    my $corefile = $opts{corefile};

    my ($gdb, $image);
    if ($corefile =~ /\.nz/) {
        # assume this is a non-simulator core
        my $info = core_info_nz(corefile => $corefile);
        $gdb = _guess_gdb_from_info_nz($info);
        $image = _guess_image_from_info_nz($corefile, $info);
    } else {
        $gdb = "/usr/software/bin/gdb";
        $image = _guess_image_from_sim_core($gdb,$corefile);
    }
    my $output = _run_gdb(gdb => $gdb,
                          args => [$image, $corefile],
                          commands => $opts{commands});
    if (wantarray) {
        return ($output, $image);
    } else {
        return $output;
    }
}


=head2 root_chmod

  root_chmod( mode => $mode, filelist => \@files )

As the root user, perform C<chmod> with the given mode on the given
files.  Return true if this succeeds, or die with a message describing
the failure if it fails.

The intended usage is to let one chmod core files with a mode of
"0444" so that programs running as non-root users can read them, or a
mode of "0777" on the directory containing the core files so that
non-root users can move or delete the cores.

(permission changes are assumed to be more reliable and secure than
trying to do moving, copying, analysis, etc., in programs running as
root)

The routines takes the following named arguments

=over

=item mode

(Required): What mode to use for chmod. Default: "0444"

=item filelist

(Required): An array reference containing a list of files to chmod.

=back

=cut

sub root_chmod {
    my (%opts) = validate(@_, {
        mode => { type => SCALAR, default => "0444" },
        filelist => {type => ARRAYREF},
    });
    my $mode = $opts{mode};
    my @filelist = @{$opts{filelist}};
    if (!@filelist) {
        return;
    }
    my @cmd = ("/usr/software/bin/sudo","/bin/chmod",$mode,@filelist);
    my $data = _qx_list(@cmd);
    if ($?) {
        die(join(" ",@cmd) . " failed with exit status $?\n".
            "output follows:\n".
            "$data\n");
    }
    if ($data) {
        die(join(" ",@cmd) . " failed (printed extra output)\n".
            "output follows:\n".
            "$data\n");
    }
    return 1;
}


# Cause a netcache box to be willing to serve its root directory over
# NFS
# XXX - replace this with setup routines from setup.pm after its cleaned

sub _netcache_enable_nfs {
    my ($conn) = @_;
    license($conn, "nfs") or die $@;
    say($conn, "priv set -q advanced; nfs on");
    say($conn, "priv set -q advanced; exportfs -i -o anon=0 /");
}


# It sure would be nice if qx() could take list and treat it as the
# exact argv[] to use, just as system and exec can, rather than
# necessarily running a shell, so that we don't have to quote
# filenames, etc.  This uses IPC::Open3 in order to have that effect.

# It captures stderr at the same output as stdout (like 2>&1).

sub _qx_list {
    my (@cmd) = @_;
    local *DUMMY;
    local *IO;
    open3(\*DUMMY,\*IO,\*IO,@cmd)
        or die "Could not invoke " . join(" ",@cmd) . ": $!\n";
    local $/;
    my $data = <IO>;
    close(IO);
    close(DUMMY);
    wait();
    return $data;
}


# Local Variables:
# indent-tabs-mode: nil
# End:

1;

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = savecore, nfs, exportfs

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-central-libs
	$Id: //depot/prod/test/Rsierranevada/lib/Filer/Decore.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

