# $Id: //depot/prod/test/Rsierranevada/lib/Filer/Foreach.pm#1 $

## @summary Helper to handle all filers in parallel (in prejobs)
## @author lincoln@netapp.com, dl-test-infra@netapp.com
## @status Shared
## @pod here

package Filer::Foreach;

use strict;
use warnings;

use Tharn qw(param host hosts subtest logresult script_result);
use Params::Validate qw(validate SCALAR CODEREF ARRAYREF);
use Carp qw(confess);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}


require Exporter;
our @ISA=qw(Exporter);
our @EXPORT_OK=qw(foreach_filer);

=head1 NAME

Filer::Foreach - Call a function in parallel for each filer in $FILER_LIST

=head1 SYNOPSIS

    use Filer::Foreach qw(foreach_filer);

    foreach_filer(function => \&function,
		  runid_prefix => 'mytest',
		  subtest_args => []);

    sub function {
        param('FILER','-mesg','...');
        param('FILER_SHOESIZE','-default','12');
	# Now do things to filer $FILER
    }

=head1 DESCRIPTION

B<foreach_filer> calls a function for every filer prefix listed in the
parameter C<FILER_LIST>.  The functions will execute in parallel
(often in separate subtests).  This is expected to be most useful and
tasteful at the beginning of a utility test meant to run as a prejob,
but less so elsewhere.

By convention (and by behavior of B<ntest>) C<FILER_LIST> 
contains a comma separated list of parameter prefixes.  The hostspec
of the filer is given in a parameter by the same name as the prefix,
but other parameters beginning with that prefix are also presumed to
apply to that particular filer.  For example, C<FILER_SRC_IP> would be
the conventional parameter passed in to use a different IP address
than the default for the filer specified in parameter C<FILER_SRC>.

When B<foreach_filer> calls the function to operate on a particular
filer prefix, all parameters beginning with that prefix will be
renamed to begin with the prefix C<FILER>.  Other parameters which
already begin with "FILER" (those which are not directly replaced by
this renaming) will be left alone, so users can tend to pass in values
intended for all filers by beginning them with the prefix "FILER".

As a special case, it is suspected that a filer with prefix C<FILERB>
may be the CFO/SFO partner of a filer with prefix C<FILERA> and vice
versa.  Whenever B<foreach_filer> renames one of those prefixes to
C<FILER> it renames the other to C<PARTNER>.

If two filers indirectly referred to by C<FILER_LIST> are a CFO/SFO
pair, where one host's part_host field is set to another host's
hostname, then the other host's parameters are renamed to begin with
C<PARTNER> whenever the first host's parameters are renamed to begin
with C<FILER>.

In the degenerate case where C<FILER_LIST> is set to C<FILER>,
B<foreach_filer> just calls the given function without creating any
subtests.  (Note that in this case 'subtest_args' will be ignored.)

The following text is recommended when documenting a test which uses
B<forach_filer> to call a function that does the entirety of its work
(replace "mangled" with something that describes what the test is
actually supposed to do to those filers):

	@param FILER_LIST Required: A comma separated list of
	parameter names, such as "FILER" or "FILERA,FILERB" or "NODES".
	All of the filers listed in parameters by those names will be
	mangled in parallel.  Also, parameters shown below as beginning
	with "FILER" may be begun with one of these parameter names
	instead, in which case it will apply only to that filer.

=head1 EXAMPLE

This:

    foreach_filer( function => \&f, 
                   runid_prefix => "example",
		   subtest_args => ['-timeout',3600]);

is equivalent to the following if FILER_LIST is "FILERA,FILERB":

    subtest(\&f, '-bg',
	    '-runid', "example-$FILERA",
	    '-mapparam', 's/^FILERA/FILER/',
	    '-mapparam', 's/^FILERB/PARTNER/',
	    '-timeout',3600);
    subtest(\&f, '-bg',
	    '-runid', "example-$FILERB",
	    '-mapparam', 's/^FILERB/FILER/',
	    '-mapparam', 's/^FILERA/PARTNER/',
	    '-timeout',3600);
    Subtest::wait_finish();

and equivalent to the following if FILER_LIST is "FILER":

    f();

=cut

# switch package name since $FILER_LIST will be in main package
package main;
our ($FILER_LIST);
package Filer::Foreach;

sub foreach_filer {
    my (%opts) = validate(@_, {
                    function => { type => CODEREF },
                    runid_prefix => { type => SCALAR },
                    subtest_args => { type => ARRAYREF, default=>[] },
                    });
    my $scriptspec = $opts{function};
    my $runid_prefix = $opts{runid_prefix};
    my @subtest_args = @{$opts{subtest_args}};

    param('FILER_LIST','-mesg','Comma separated list of filer prefixes');
    # NATE magically defaults FILER_LIST to FILERA,FILERB... or FILER or NODES
   

    if ($FILER_LIST eq "FILER") {
        # only one filer so execute the function directly instead of subtesting
        # XXX - dont we need to param here also ?
        &{$opts{function}};
    } else {
        my @subtests;
        my $preserve_hosttype = 0;
        my @prefixes = split(/,\s*/,$FILER_LIST);
        # create a list of hostids and the prefix from where they came
        # do this first or so that we can search for the SFO partner
        my @host_list;
        foreach my $prefix (@prefixes) {
            my $hostspec = param($prefix,'-mesg',undef);
            # if someone passed in a hosttype on the hostspec, we want to keep it
            if ( $hostspec =~ /hosttype/ ) {
                $preserve_hosttype = 1;
            }
            my @temp = hosts($hostspec);
            foreach my $hostobj (@temp) {
                if ( $preserve_hosttype ) {
                    push(@host_list, [$hostobj->id().";hosttype=".$hostobj->hosttype(), $prefix]);
                } else {
                    push(@host_list, [$hostobj->id(), $prefix]);
                }
            }
        }
        # Launch the scriptspec in background for all the hosts
        foreach my $href (@host_list) {
            my ($host_id, $prefix) = @$href;
            my $hostobj = host($host_id);
            my $host_name = $hostobj->hostname();
            my @subtest_opts = ('-mapparam', sub { s/^$prefix\B/FILER/; });
            push @subtest_opts, ("FILER=$host_id");
            my $partner_host = $hostobj->part_host();
            if ($partner_host) {
                my $found;
                # Search for partner host prefix in $FILER_LIST
                # Partner could be present the same prefix or other prefixes
                foreach my $pref (@host_list) {
                    my ($partner_id, $partner_prefix) = @$pref;
                    if ($partner_id eq $partner_host) { 
                        $found=1;
                        push @subtest_opts, ("PARTNER=$partner_id");
                        push @subtest_opts, ('-mapparam', 
                                    sub { s/^$partner_prefix\B/PARTNER/; });
                        last;
                    }
                }
                logresult('warn',"PARTNER $partner_host for $prefix is not in the param FILER_LIST")
                    if (!$found);
            }
            push(@subtests,
                 &subtest($scriptspec,
                        '-runid', $runid_prefix . "-" . $host_name,
                        @subtest_args,
                        '-bg',
                        @subtest_opts,
                        'FILER_LIST=FILER'));
        }
        Subtest::waitFinish(subtests=>\@subtests);
    }
}

1;

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    lincoln@netapp.com, dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Filer/Foreach.pm#1 $
    Last revised $Date: 2013/01/22 $ 

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut
