# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Filer/Revert.pm#1 $
##
## @summary Filer::Revert is derived from Filer::Transform. 
## The only real differences between an upgrade and a revert are that you
## cannot move back more than two versions in a revert, whereas for an upgrade,
## the rules are more complicated. In addition, you have to issue the revert_to
## command to the filer. This library handles these additional tasks.
##
## @example
##
## my $revert = Filer::Revert->new($conn, to_version => $TO_VERSION,
##                                 FILER => $FILER, CLIENT => $CLIENT);
##
## @status public
##
## @author  shenry@netapp.com (originally amandag@netapp.com)
##
## @burt 377973 IOC: revert test cases failing with error message
##              "Error: Invalid respond from revert_to: "
## @change 2009-10-07 author:bhoomika
## Modified the code to return valid response on running revert command with
## revert response as 'no'


package Filer::Revert;

use strict;
use Filer::Transform;
use Carp;
use Data::Dumper;
use Filer::Version qw(version_contained version_compare);
use SPARSE::Util::Logger qw(logDbg logTrc logWarn);
use Tharn;
use SnapCmds qw(snapshot_list snapshot_delete);

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $LIB_REVERT_TRIALS);
@ISA = qw(Filer::Transform);
$VERSION = '0.5';
param("LIB_REVERT_TRIALS", "-default", 8);

# --------------------
# sub new
# the constructor for this class basically just adds an array ref as an
# object member; the array contains the two previous versions, or more
# (THOUGH this is not yet supported -- old version like 6.3 or 6.2 sometimes
# lets you revert back 3 versions, not just 2)
# --------------------
sub Filer::Revert::new {
  my ($class, @args) = @_;
  print "class is $class";
  my $self = $class->SUPER::new(@args);

  # add the revert versions
  $self->{_prev_versions} = [ $self->_prev_versions() ];

  return $self;
}

# --------------------
# sub _version_allows
# _version_allows checks whether the "to_version" that's been
# requested is compatible with the acceptable previous versions
# --------------------
sub _version_allows {
    my $self = shift;

    return undef unless $self->{_to_version_short};

    logDbg("_version_allows: to_version=" . $self->{_to_version});
    logDbg("_version_allows: pre_versions=" . join " ", @{$self->prev_versions()});
    my $short_vers = $self->{_to_version_short};
    logDbg("_version_allows: short_vers=$short_vers");

    logDbg("_version_contained: ". version_contained($short_vers, @{$self->prev_versions()}));
    return version_contained($short_vers, @{$self->prev_versions()});
}

# --------------------
# sub _running_status_allows
# checks whether status of online volumes or other run time items
# would prevent a revert
# This is incomplete
# --------------------
sub _running_status_allows {
    my $self = shift;
    # flex vols in anchorsteam cannot be reverted to scrimshaw 
    if ($self->v_compare($self->from_version, "7.0") >= 0 &&
        $self->v_compare($self->to_version, "6.5") <= 0) {
        # FIXME check for flexible volumes
    }
    return 1;
}

# --------------------
# sub _min_max_allows
# inherited
# --------------------

# --------------------
# sub _hardware_allows
# inherited
# --------------------

# --------------------
# sub _config_allows
# inherited
# --------------------

# --------------------
# sub do
# Execute the steps necessary to revert the filer
# after all other steps have been taken to initialize
# and information about the versions and paths
# have been gathered
# All optional arguments
#  download_kernel => [0|1]
#  waitstrings => $array_ref #unsupported 
#  revert_response => [yes|no]
#  goforit => [-1|1|2|etc.]
#  releases => [releases|releases_pbo|releases_dbg] # defaults to => "releases"
# --------------------
sub do {
    my $self = shift;
    my %parms = @_;
    
    my $download_kernel = defined $parms{download_kernel} ? 
                            $parms{download_kernel} : 1;
    logWarn("argument 'download_kernel' should be either 0 or 1 in ".
            "Filer::Revert->do(); value was $download_kernel")
        unless ($download_kernel == 1 || $download_kernel == 0);

    # THESE HAVE GONE UNSUPPORTED IN NEW CODE
    my $waitstrings = $parms{waitstring} || [ ];
    $waitstrings = [ map { quotemeta($_) } @$waitstrings ];

    # STILL SUPPORTED
    my $rvt_response = $parms{revert_response} || "yes";

    # <0 == result only, say "no" if would succeed
    # 1 == try once and return
    # 2+ number of trys, fix blockers
    logDbg("param(LIB_REVERT_TRIALS) set to '$LIB_REVERT_TRIALS'");
    my $goforit = defined $parms{goforit} ? $parms{goforit} : $LIB_REVERT_TRIALS;
    if ( $goforit !~ /^-?\d+$/) {
        logWarn("variable goforit has invalid value: '$goforit', set to 8"); 
        $goforit = 8;
    }

    my $trials = 0;
    my $response = '';
    my $abbrev_version = $self->{_to_version_short};

    print "abbrv version is $abbrev_version";


    say($self->conn, "nfs off", -eval_error); # or croak "Could not turn off NFS";
    say($self->conn, "cifs terminate", -eval_error); # or croak "Could not turn off CIFS";

    # Get the kernel
    if ( $download_kernel ) {
	my $download_status = $self->download_kernel();
	logDbg("download_kernel() returned $download_status");
	die "download returned failure code '$download_status'" if (!$download_status);
    }

    $self->online_vols();
    my $Revert_Retry = qr/Please address the above conditions?, then try "revert_to" again/;
    my $Revert_Query1 = '\(Yes\/No\)\?';
    my $Revert_Query2 = '\[yes\/no\]\?';
    my $Revert_Starting = 'Starting revert to';
    my ($match, $before, $after);
    my $perform_revert_flag = 1;
    my $endtime = time() + "$::TIMEOUT";

    do {
        logDbg("revert->do: trials=$trials, goforit=>$goforit");
	if($perform_revert_flag == 1)
	{
        ($match, $before, $after)
          = say($self->conn, "revert_to " . $abbrev_version,
		qr/$Revert_Query1|$Revert_Query2|$Revert_Retry/,
                -swallow,  -eval_error, -noprompt);
        croak $@ if $@;
	$perform_revert_flag = 0;
	}
        if ($match =~ /$Revert_Query1|$Revert_Query2/i) {
            logDbg("revert->do: got a yes/no query, goforit=$goforit");
            if ($goforit <= 0) {
                say($self->conn, "no", -eval_error, -noprompt);
                return ($before, $after);
            } else {
	    ($match,$before,$after) = say($self->conn, $rvt_response,
	      qr/$Revert_Query1|$Revert_Query2|$Revert_Retry|$Revert_Starting/,
	      -swallow, -eval_error, -noprompt, '-timeout', $::CMDTIMEOUT*10,
	      '-maxidle', $::TIMEOUT);
            }
        } elsif ($match =~ /$Revert_Retry/) {
            logDbg("revert->do: got a revert retry, goforit=$goforit");
            # parse $before for stuff to do...
            logDbg("BEFORE IS: $before");
            return $before if $goforit <= 0;
	    $self->process_conditions($before);
	    $perform_revert_flag = 1;
        } else {
            croak "unexpected match string from revert_to: $match";
        }
        # this code needs to be last to allow $goforit < 0 to work...
      	$trials++;
      	croak ("Tried $goforit times and still failed") if ($trials > $goforit);
        logDbg("Starting try $trials of $goforit to revert to $abbrev_version");
    } while (($match !~ /$Revert_Starting/i) && (time() < $endtime));
    return $before;
}

# --------------------
# sub process_conditions
# this routine takes a string - specifically the "before" string for a
# say() command - and iterates through it looking at various errors
# that the filer mentions
# If it determines that it can do something about the error because it
# "knows" about it, then it handles that in the condition, or dies if it
# is an error that can't be handled and is basically fatal to the test
# --------------------
sub process_conditions {
    my $self = shift;
    my $before = shift;

    my @lines = split(/\r+\n/, $before);
    for my $i (0..$#lines) {
        my $line = $lines[$i];
        chomp($line);
        next if $line =~ /^\s*$/;
        if ($line =~ /All volumes must be online/) {
            return $before;
        } elsif (
    $line =~ /^New.*snapshot.*(volume|aggregate) (\S+) .*deleted prior to reverting:/ ||
    $line =~ /^Snapshots.*(volume|aggregate) (\S+) .*deleted prior to reverting:/ ) {
	    my $type = $1;
            my $vol = $2;
            while ($lines[++$i] !~ /^\s*$/) {
                my @snaps = split(/\s+/, $lines[$i]);
                for my $snap (@snaps) {
		    if ( $type eq "aggregate" ) {
                    	snapshot_delete($self->conn, "-A " . $vol, $snap) if $snap;
		    } else { 
                    	snapshot_delete($self->conn, $vol, $snap) if $snap;
		    }
                }
            }
        } elsif ($line =~ /prepared for revert/) {
            croak "The install must have failed: $line";

        } elsif ($line =~ /Inofile.*volume.*is still in progress/) {
            croak $line;

	} elsif ($line =~ /SnapMirror must be turned off before doing \"revert_to\"/ ) {
    	    say($self->conn, "snapmirror off", -eval_error); # or croak "Could not turn off SnapMirror";

	} elsif ($line =~ /FCP must be turned off before doing \"revert_to\"./ ) {
    	    say($self->conn, "fcp stop", -eval_error); # or croak "Could not turn off FCP";

	} elsif ($line =~ /Please revert SIS metafiles on Volume (.*) to (.*)/) {
	    say($self->conn, "sis stop /vol/$1",
		qr/Operation is currently idle/,
                -swallow,  -eval_error, -noprompt);
	    say($self->conn, "sis revert_to " . $2 . " -delete",
		     qr/command has succeeded/,
		     -swallow,  -eval_error, -noprompt);

	} elsif ($line =~ /Inode (.*) in volume (.*) is used by file (.*) Delete it .*/) {
	    my $file = $3;
	    chop($file);
	    say($self->conn, "priv set -q advanced;rm $file", -eval_error);

	} elsif ($line =~ /Summary file cleanup on volume.*still in progress/) {
	    croak $line;

        } else {
            logDbg("NEW CONDITION FOUND: $line");
        }
        # add additional processing conditions above the else...
    }
    sleep 15;                   # give things time to quiesce
}

1;
