# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Filer/Transform.pm#1 $
##
## @summary Transform.pm is the base class for upgrading and reverting a filer
##
## @example
##
## @status public
##
## @author  shenry@netapp.com (originally amandag@netapp.com)
##
## @author  bhoomika@netapp.com
## @burt 365553 DI : Upgrade/Revert scripts do not support the latest
##	 nomenclature of BR builds i.e. RboilermakerN_090613_0100
## @change 2009-07-23 author:bhoomika
## Added support for the new release path for the latest BR builds
##
## @author  vanik@netapp.com
## @burt 374935 DI: Upgrade/Revert scripts (Library) support
## for new version string (R8.0xN)
## @change 2009-09-17 author:vanik
## Added support for latest BR Nomenclature





package Filer::Transform;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK 
            $ONLINE_VOLS $DONOTREALLYUPGRADE $DONOTREALLYREVERT);

use Sysconfig;
use SayWrap qw(sayEOF sayWAIT);
use Carp;
use filer qw(filer_version_get);
use SPARSE::Util::Logger;
use FS::Util::Setup qw(mount_root_vol umount_vol);
use FS::Util::Utilities qw(uSystem);
use Data::Dumper;
use Volume qw(vol_root vol_status vol_list vol_online);
use Tharn qw(subtest connect host say param config_result);
use Utils qw(make_host make_conn);
use install;
BEGIN {
    use TestUtils qw(get_site_info);

    # these lines are here only to make TestUtils work...but TestUtils should
    # just be fixed instead - it needs the '$Options->process()' bit.
    use Options;
    my $Options = new Options();
    $Options->process();
}
use SnapCmds qw(snapshot_list snapshot_delete);
use Filer::Version qw(get_ontap_version version_compare codename_to_version version_allowed get_wafl_version ontap_to_wafl_ver WV_BOILERMAKER_MIN);
use boot qw(transit_to reboot_to boot_state);


$VERSION = '0.5';

# ---------------------
# sub new()
# Constructor requires a filer connection record as the first argument
# All other parameters are supplied in a named fashion. If not
# supplied, the "to_version" argument will default to undef
# Some of the records for the constructor include:
#  the filer connection record
#  the filer IP address
#  the to_version that the test runner has chosen
#  the from_version, retrieved from the filer
#  the previous versions, also retrieved from the filer -- "previous"
#    refers to the acceptable versions for revert, not the revert history
#  the filer hostspec
#  the Sysconfig object
# ---------------------
sub new
{
    my $class = shift;
    my $conn = shift;
    my %parms = @_;
    $ENV{TIMEOUT} = 1200 if $ENV{TIMEOUT} < 1200;
    $ENV{CMDTIMEOUT} = 2400 if $ENV{CMDTIMEOUT} < 2400;

    my $self = {
		_conn               => $conn,
		_filer_ip           => $conn->hostp->default_ip,
		_to_version         => $parms{to_version} || undef,
		_to_version_short   => $parms{to_version} || undef,
		_from_version       => $parms{from_version} || 'RUNNING',
		_from_version_short => $parms{from_version} || 'RUNNING',
		_f                  => $parms{FILER} || $::FILER,
		_sysconfig          => undef,
		_prev_versions      => [ _prev_versions($conn) ],
		_mount              => $parms{rootvol_mount} || undef,
		_min_to_version     => $parms{min_to_version} || undef,
		_max_to_version     => $parms{max_to_version} || undef,
		_min_from_version   => $parms{min_from_version} || undef,
		_max_from_version   => $parms{max_from_version} || undef,
	       };
    # first check if we need to find out the version we are upgrading from
    # normally FROM_VERSION is set to 'RUNNING'
    if ($self->{_from_version} eq "RUNNING") {
	$self->{_from_version} = _current_version($conn); #get the version string from filer
	$self->{_from_version_short} = get_ontap_version($conn->hostp()); #get the version from the wafl version
	logDbg("Transform::new -- from_version = $self->{_from_version}");
	if ($self->{_from_version} =~ /.*\/(.*)/) {
	    $self->{_from_version} = $1;
	}
	logDbg("RUNNING version is $self->{_from_version}");
    }
    # else the version has been supplied as a param
    return 0 unless $self->{_to_version} && $self->{_from_version};
    $self->{_to_version_short} = codename_to_version($self->{_to_version});
    if ($self->{_to_version_short} =~ /^(\d+\.\d+).*/) {
	$self->{_to_version_short} = $1;
    }

    # at this stage from_version_short should be something like x.x,
    # from_version can be formatted in a number (e.g 6.3.2RN) of ways
    # but must have an x.x substr. to_version_short must be x.x.x or x.x.
    # To specify a release which cannot be deduced from the the TO_VERSION
    # set the env var RELEASEVERSION (still supply TO_VERSION)
    logDbg("from_version is $self->{_from_version} to_version is $self->{_to_version}.");

    bless $self, $class;
}

# --------------------
# Read-only Accessors
# These convenience methods do not prevent the caller from writing code like
#  $self->{_conn}
# but they do protect against potential mistakes like
#  $self{_conn}
# or a situation that assigns undef to $self->{_conn} accidentally, which
# would be disastrous. $self->conn would simply ignore the undef and all
# would be well.
#
# usage example:
#  my $from_version = $self->from_version();
#
# The read-only accessors are for each member of the object that were
# blessed in the constructor that shouldn't be changed.
#
# It would be appropriate to upgrade these hand-coded methods with a
# caching AUTOLOAD approach, but there is some disagreement on what is
# an acceptable level of Perl education required to for library code.
# --------------------
sub conn          { $_[0]->{_conn}          }
sub filer         { $_[0]->{_f}             }
sub filer_ip      { $_[0]->{_filer_ip}      }
sub from_version  { $_[0]->{_from_version}  }
sub prev_versions { $_[0]->{_prev_versions} }
sub sysconfig {
  my $self = shift;
  if (! $self->{_sysconfig}) {
    my $cfg = Sysconfig->new($self->filer) || croak $@;
    $self->{_sysconfig} = $cfg;
  }
  return $self->{_sysconfig};
}

# --------------------
# Read-Write Accessors
# Again, these are convenience methods that are called in object fashion.
# If an argument is supplied, the corresponding object member is updated
# to reflect this new value, e.g.
#
#  $self->to_version("6.5");
#
# Thus the user can either set the value as above, or retrieve the value
# as below
#  my $to_version = $self->to_version();
#
# 0 is not a valid version.
#
# --------------------
sub to_version {
  my $self = shift;
  if (my $version = shift) {
    $self->{_to_version} = $version;
  }
  return $self->{_to_version};
}

sub mount {
  my $self = shift;
  $self->{_mount} = shift || $self->{_mount} || $self->make_mount_point();
}


# --------------------
# sub allowed
# The upgrade or revert is only allowed if four specific "tests" pass:
# (1) version
# (2) hardware
# (3) config
# (4) run status
# --------------------
sub allowed {
  my $self = shift;
  if ($self->_min_max_allows()   &&
      $self->_version_allows()   &&
      $self->_hardware_allows()  &&
      $self->_config_allows()    &&
      $self->_running_status_allows() ) {
    return 1
  } else {
    return 0
  }
}

# --------------------
# sub _min_max_allows
# sub _version_allows
# sub _hardware_allows
# sub _config_allows
# sub _running_status_allows
# --------------------
sub _version_allows        { return 1 }
sub _hardware_allows       { return 1 }
sub _config_allows         { return 1 }
sub _running_status_allows { return 1 }

sub _min_max_allows
{
    my $self = shift;

    # version_allowed defaults to die(na_result()) on failure
    # check the from (running) version limits
    version_allowed($self->conn->hostp(),
		    min_version => $self->{_min_from_version},
		    current_version => $self->{_from_version},
		    max_version => $self->{_max_from_version});

    # check the to version limits
    version_allowed($self->conn->hostp(),
		    min_version => $self->{_min_to_version},
		    current_version => $self->{_to_version},
		    max_version => $self->{_max_to_version});

    return 1;
}

# --------------------
# sub v_compare
#  now implemented in Filer/Version.pm
# --------------------
sub v_compare
{
    my $self = shift if ref $_[0];
    return version_compare($_[0], $_[1], $_[2]);
}


# --------------------
# sub _current_version
# This retrieves and returns what the filer thinks its current version is
# --------------------
sub _current_version
{
  my $self = shift;
  print "SELF IS $self\n\n\n";
  my $conn = $self =~ 'Connectrec' ? $self : $self->conn;

  my $version = filer_version_get($conn) or croak $@;
  logDbg("The version on the filer is: $version");
  return $version;
}

# --------------------
# sub _prev_versions
#
# Here's a slightly more experimental method for you, called _prev_versions;
# the title refers to the two versions that a filer can revert to from its
# current version, e.g. 6.5 and 6.4 if the filer's at 7.0.
#
# May be called either as a class method or as an object method.
# If called as a class method, it must be supplied with a filer connection
# record as the argument. If called as an object method, the method already
# has a copy of the filer connection record, so no arguments are needed.
#
# This method can be called in list or scalar context as well;
# In list context, this returns the previous two versions that the filer
# could potentially revert to. In scalar context, this returns the version
# immediately previous to the filer's current version.
#
# usage examples:
# my $last_version = Revert::_prev_versions($conn);   # scalar
# my $last_version = $self->_prev_versions();         # scalar
# my @last_2_vrsns = _prev_versions($conn);           # list
# my @last_2_vrsns = $self->_prev_versions();         # list
# my @last_1_vrsn  = scalar $self->_prev_versions();  # force scalar context
#
# ASSUMPTION: The filer command `revert_to` always lists acceptable versions in
# ascending order, e.g. 6.4 followed by 6.5
# --------------------
sub _prev_versions
{
  my $self = shift;
  my $conn = $self =~ 'Connectrec' ? $self : $self->conn;

  my $out = (say($conn, 'priv set -q diag; revert_to;',-eval_error))[2];
  my @lines = split /\n/, $out;
  my @versions;

  logDbg("\n\nThe raw result of _prev_versions was: ". join "\t", @lines);
  foreach (@lines) {
    if ($_ =~ /(\d+\.\d+)/) {
      push @versions, $1;
    }
  }
  if (wantarray) {
    return @versions
  } else {
    return $versions[1] || undef
  }
}


# --------------------
# sub make_mount_point
# Issue a call to mount_root_vol to create and mount a mountpoint
# for the root volume on the filer
# --------------------
sub make_mount_point
{
  my $self = shift;

  print "GOT HERE!!!\n\n";
  my $filer_ip  = $self->filer_ip;
  my $root_vol   = Volume::vol_root($self->conn) || croak $@;
  my $mountpoint = "/t/$filer_ip/$root_vol";
  $mountpoint = mount_root_vol($self->filer, undef, $mountpoint,
                               (filer_conn => $self->conn,
                                client     => 'localhost') ) ||
    croak $@;

  $self->{_mount} = $mountpoint;

  return $mountpoint;
}

# --------------------
# sub umount
# Issue a call to umount_vol to unmount the root volume
# --------------------
sub umount
{
  my $self = shift;

  if (defined $self->{_mount}) {
    umount_vol($self->{_mount}, not_mounted_okay => 1) || croak $@;
  }

  return;
}

# --------------------
# sub _determine_ontap_url
# Based on the hardware config and to_version, determine which
# image to use.
# on a per-test basis, you can send $self->install(kernel => 'debug')
# or you can override on the command line with $RELEASESTYPE.
# --------------------
sub _determine_ontap_url {
  my $self = shift;
  my %parms = @_;
  my $from_version = $self->{_from_version};
  my $debug;
  my $version = $self->{_to_version};
  my ($processor, $dist) = ('','');
  # Get the kernel from the kernel => ... param passed in through install()
  # or simply set to releases (works for all) by default.
  # kernel path postfix options are "releases", "releases_debug", or
  # "releases_pbo";
  my $kernel = defined $ENV{RELEASESTYPE} ? $ENV{RELEASESTYPE} :
               defined $::RELEASESTYPE ? $::RELEASESTYPE :
               defined $parms{kernel} ? $parms{kernel} : "releases";

  # FIXME this does not work for 6.4
    my $version_string = (say($self->conn, "version -v"), -eval_error)[2];
    if ($kernel eq "releases_debug")
    {
        $debug = ".debug";
    }

   my $current_ver_is_ng = 0; #this flag is set if _from_version is BR
   my $target_ver_is_ng = 0; #this flag is set if _to_version is BR
   
   # checking if the version to be upgraded to is BR.
   # If it is then the value of $target_ver_is_ng is set to 1
   ## burt 436762: to support BR and later
   my $convert_ver =  ontap_to_wafl_ver ( $version );
   if ( $convert_ver >= WV_BOILERMAKER_MIN ) {
	$target_ver_is_ng = 1;
   }

   # If it is then the value of $current_ver_is_ng is set to 1
   ## burt 436762: to support BR and later
   my $wafl_ver = get_wafl_version( $self->conn->hostp() );
   if ( $wafl_ver >= WV_BOILERMAKER_MIN )
   {
        $current_ver_is_ng = 1; 
	# Get the processor so we know which dist
        if($version_string =~ /\<(\d).?\>/)
	{
        	$processor = $1;
        	if ($processor == 1) {
            		$processor = '2';
        	} elsif ($processor == 5) {
            		$processor = '8';
        	}
	}
	else
	{
	    logWarn("Warning: Could not determine platform, defaulting to PC (1)");
            $processor = '3';
        }
   } else {
        $version_string = (say($self->conn, "version -v"), -eval_error)[2];
        if ($version_string =~ /\<(\d).?\>/) {
            $processor = $1;
        } else {
            logWarn("Warning: Could not determine platform, defaulting to PC (2)");
            $processor = '3';
        }
  }
  if ($processor == 1) {
    $dist = "alpha";
  } elsif ($processor == 2) {
    if((($target_ver_is_ng == 1) && ($current_ver_is_ng == 1)) || ($target_ver_is_ng == 1))
    { 
		$dist = "x86_64" . "$debug";
    } 
    elsif ($current_ver_is_ng == 1) 
    {
	$dist = "x86-64";
    } 
    else
    {
	$dist = "x86-64";
    }
  } elsif ($processor == 3) {
    $dist = "pc";
  } elsif ($processor == 6) {
    $dist = "mips";
  } elsif ($processor == 8) {    
   if((($target_ver_is_ng == 1) && ($current_ver_is_ng == 1)) || ($target_ver_is_ng == 1))
   { 
		$dist = "x86" . "$debug";
   }
   elsif ($current_ver_is_ng == 1) 
   {
		$dist = "pc_elf";
   } 
   else
   {
	$dist = "pc_elf";
   } 
  } else {
    croak "This library only supports alpha (1), x66-64 (2), pc (3), mips ".
          "(6), pc_elf (8) and the processor version detected was ($processor)";         
    }

  # the image path is different from boilermaker on
  my $image_path;
  if ($target_ver_is_ng) {
    ## bhoomika : added the support for the latest release path in BR (BURT 365553)
    ## Added support for the latest release in BR (Burt 374935)
    ## burt 436762 (fixed) to support BR and later
    if ( ontap_to_wafl_ver ( $version ) >= WV_BOILERMAKER_MIN ) {
	$version = "/x/eng/rlse/DOT/$version" if $version !~ m{^/};
        if ($current_ver_is_ng == 1) {
            # if BR to BR use image.tgz
            # refer http://wikid.netapp.com/w/NGS_NPI/Data_ONTAP/8.0.0_TOI/Upgrade
            $image_path = "$version/final/bedrock/export/$dist/tarball/image.tgz";
        } else {
            $image_path = "$version/final/bedrock/export/$dist/tgz2zip/image.zip";
        }
    	croak "Path $image_path was not found" unless -e $image_path;
    } else {
    	$version = "/rlse/ontapng/$version" if $version !~ m{^/};
    	$image_path = "$version/final/bedrock/export/$dist/tgz2zip/image.zip";
    	croak "Path $image_path was not found" unless -e $image_path;
    }
  } else {
    $version = "/rlse/$version" if $version !~ m{^/};
    $image_path = "$version/$kernel/$dist/dist/setup.exe";
    croak "Path $image_path was not found" unless -e $image_path;
  }

  my $site = $parms{SITE} || $self->conn->hostp->{SITE} || $::SITE;
  croak "no site defined!" if (!$site);

  # build the install url using site-specific values for various components
  my %siteinfo = get_site_info();
  croak "unknown site '$site', can't continue" if (!grep $site, keys %siteinfo);
  # we just need data for the current site
  %siteinfo = %{$siteinfo{$site}};
  my $ontap_url = $siteinfo{install_proto} . '://' . $siteinfo{install_user} . ':' . $siteinfo{install_pass} . '@' . $siteinfo{install_server} . $image_path;

  # these overrides are handy for debugging
  #$ontap_url = 'http://10.57.83.10/engineering/rlse/RgordonbierschN/releases_debug/pc_elf/dist/PC_ELFdist.zip',
  #$ontap_url = 'http://10.57.83.10/engineering/rlse/ontapng/mainngN_tested/final/bedrock/export/x86.debug/tgz2zip/image.zip',

  logDbg("ONTAP_URL IS '$ontap_url'\n");

  return $ontap_url;
}

# --------------------
# sub install_release
# This method sets up the mount on the client, downloads the data files,
# and runs the install program
# --------------------
sub install_release {
  my $self = shift;
  my %parms = @_;

  # DEBUGGING ONLY!!!
  return 1 if ($::DONT_REALLYINSTALL or $ENV{"DONT_REALLYINSTALL"});
  
  my $client = $parms{client} || $::CLIENT || 'localhost';

  my $ontap_url = $self->_determine_ontap_url(%parms);

  logDbg "PROCEEDING TO INSTALL $ontap_url";

  my $result = install::copy_system_files(
      $self->conn->hostp(),
      $client,
      ONTAP_URL => $ontap_url,
  );
  die "copy_system_files() failed: $@" if ($@);

  logDbg "FINISHED INSTALL $result\n";

  return 1;
}


# --------------------
# sub download_kernel
# Issue the download command
# Return the response from the filer
# --------------------
sub download_kernel {
    my $self = shift;
    my ($match, $after) = ('','');

    return 1 if $self->conn->hostp->hosttype eq 'filer-simulator';

    logDbg "PROCEEDING TO DOWNLOAD THE KERNEL: ";

    my $client = $::CLIENT || 'localhost';

    # initiate the download
    install::do_download(
	$self->conn->hostp,
	$client,
	@_
    );

    # check for error
    return 1 unless $@;
    croak "do_download failed: $@";
}

# --------------------
# sub zero_spares
# Test utility function
# Check environment for $ZERO_SPARES (enabled by default)
# Can disable from command line or in a test with
# $ZERO_SPARES=0 or $ZERO_SPARES=[n|no|N|No]
# Sends a background message to the filer to zero the spare disks
# Which can reduce the amount of time it takes to run a test
# --------------------
sub zero_spares
{
  my $self = shift;
  # Does not apply below Scrimshaw
  return if (v_compare($self->current_version(), '6.5') == -1);

  if (! defined $ENV{ZERO_SPARES} || $ENV{ZERO_SPARES} =~ /y|1/i) {
    logDbg "Zeroing spares now\n";
    sayEOF(shift, "disk zero spares;result;");
    sleep(5);
  }
}


# --------------------
# sub online_vols
# Check environment for $ONLINE_VOLS (enabled by default)
# Can disable from command line or in a test with
# $ONLINE_VOLS=0 or $ONLINE_VOLS=[n|N|no|NO]
# Bring all vols online
# croak if they can't be brought online
# --------------------
sub online_vols
{
  my $self = shift;
  my @vols = Volume::vol_list($self->conn);

  if ($::ONLINE_VOLS =~ /y|1/i) {
    map { vol_online($self->conn, $_) or croak $@ } @vols;
  }
}

# --------------------
# sub filer_snaps
# Returns a list of all snaps on the filer
# --------------------
sub filer_snaps
{
  my $self = shift;
  my @snaps = map { snapshot_list($self->conn, $_) } vol_list($self->conn);
  return @snaps;
}


# --------------------
# sub do_reboot
# Issues a reboot to the filer, unless the test runner wants to disable
# --------------------
sub do_reboot
{
    my $self = shift;
    my %parms = shift;
    return 1 if ($ENV{DONOTREALLYUPGRADE} =~ /y|1/i ||
		 $ENV{DONOTREALLYREVERT}  =~ /y|1/i ||
		 $DONOTREALLYUPGRADE      =~ /y|1/i ||
		 $DONOTREALLYREVERT       =~ /y|1/i ||
		 ($parms{REBOOT} && $parms{REBOOT} !~ /y|1/i)
		);
    logDbg "PROCEEDING TO REBOOT: $@";
    my $filer = $self->filer;
    my $fhost = host($filer);
    my $conn = $self->conn;
    reboot_to($fhost, 'UP', timeout=>'2400') ;
    my $state = boot_state($fhost, timeout => '900') or die $@;
    if ($state  ne 'UP') {
        warn "Unexpected boot state: $state, Rebooting the filer";
        transit_to($fhost, 'UP') or die "Failed to reboot $@";
     }
    logStep("brought the system back up");
    return 1;
}

# --------------------
# sub validate
# --------------------
sub validate
{
  my $self = shift;
  return (version_compare($self->to_version, $self->_current_version) == 0);
}

# --------------------
# sub matches_target
# Compare the _to_version value to the current release of the filer
# Return 1 (true) if they "match" in some regexpy manner
# --------------------
sub matches_target
{
    my $self = shift;

    my $current = $self->_current_version();
    my $target = $self->{_to_version};
    if (version_compare($current, $target) == 0) {
	return;
    } else {
	return "current version $current not equal to target version $target";
    }
}


1;
