# $Id: //depot/prod/test/Rsierranevada/lib/Filer/Version.pm#3 $

## @summary Get filer internal version numbers
## @author dl-central-libs
## @status Shared
## @pod here

package Filer::Version;

use strict;
use warnings;

use Zapi;
use ATI::ati_ontapi qw(ATI_SUCCESS ATI_ONTAPI_TRANSPORT_HTTP ATI_ONTAPI_VERSION_DEFAULT
		       ATI_ONTAPI_TRANSPORT_D ATI_ONTAPI_D_VERSION_DEFAULT
		       ATI_ONTAPI_PASSWORD_DEFAULT ati_ontapi_create_context ati_ontapi_call_api);
use Tharn qw(connect host say na_result config_result);
use Carp qw(croak);
use Utils qw(make_host make_conn);
use config_base qw(host_is_alive);
use SPARSE::Util::Logger qw(logWarn logDbg);
use SayWrap qw(sayEOF);
use Tie::IxHash;

# these two constants are used by supported_features(). the values themselves
# don't matter, as long as WV_NEG_INFINITY is less than the smallest value in
# Wafl_Releases and WV_POS_INFINITY is greater than the largest value.
use constant WV_POS_INFINITY => 999999;
use constant WV_NEG_INFINITY => -(WV_POS_INFINITY);

# these values are taken from prod/common/wafl/version.h
# NOTE that ironcity/7.3 and later use a different version numbering scheme!
my %Wafl_Releases = (
		     WV_GUINNESS_MIN => 10,
		     WV_GUINNESS_MAX => 12,
		     WV_VB_MIN => 13,
		     WV_VB_MAX => 13,
		     WV_REDHOOK_MIN => 14,
		     WV_REDHOOK_MAX => 27,
		     WV_KRAKUS_MIN => 28,
		     WV_KRAKUS_MAX => 28,
		     WV_FULLSAIL_MIN => 29,
		     WV_FULLSAIL_MAX => 35,
		     WV_SCRIMSHAW_MIN => 36,
		     WV_SCRIMSHAW_MAX => 38,
		     WV_ANCHORSTEAM_MIN => 39,
		     WV_ANCHORSTEAM_MAX => 55,
		     WV_TSINGTAO_MIN => 56,
		     WV_TSINGTAO_MAX => 57,
		     WV_GORDONBIERSCH_MIN => 64,
		     WV_GORDONBIERSCH_MAX => 73,
		     WV_TRICKY_MIN => 74,
		     WV_TRICKY_MAX => 75,
		     WV_KESTREL_MIN => 76,
		     WV_KESTREL_MAX => 77,
                     WV_IRONCITY_MIN => (77 << 8), # new version format
                     WV_IRONCITY_MAX => (78 << 8) | 255,
                     WV_BOILERMAKER_MIN => (82 << 8), 
                     WV_BOILERMAKER_MAX => (82 << 8) | 255,
                     WV_BOILERQAN_MIN => (82 << 8), 
                     WV_BOILERQAN_MAX => (82 << 8) | 255,
                     WV_ROLLINGROCK_MIN => (87 << 8),
                     WV_ROLLINGROCK_MAX => (87 << 8) | 255,
                     # Changes for BURT 492060
                     WV_SIERRANEVADA_MIN => (92 << 8),
                     WV_SIERRANEVADA_MAX => (92 << 8) | 255,
                     WV_FULLSTEAM_MIN => (97 << 8),
                     WV_FULLSTEAM_MAX => (97 << 8) | 255,
		     );

# index is major version. value is minimum WAFL version for that version.
my %Wafl_Versions;
#Storing the hash in the ordered way using Perl's Tie::IxHash module
#This is done to match exactly the key to the Version string
#Refer Burt 412905
#Any additions being added in this hash, should be such that it should not be
#a substring of another string which is following it. Else there will be a
#false match, and wrong wafl version would be returned
tie %Wafl_Versions, 'Tie::IxHash';
%Wafl_Versions = (
		     "pre-6"   => 1,	# rather arbitrary
		     "6.0" => 10,
		     "6.1" => 13,
		     "6.2" => 14,
		     "6.3" => 28,
		     "6.4" => 29,
		     "6.5" => 36,
		     "7.0" => 39,
		     "7.1" => 56,
		     "7.2" => 64,
		     "7.3" => (77 << 8),
                     # Added the wafl version 8.0 (Burt 374935)
		     "8.0" => (82 << 8),
                     "8.1" => (87 << 8),
                     "8.2" => (92 << 8),

		     # tricky & kestrel do not have a release number yet
		     # check prod/common/fsinfo.h for updates
		     "guinness" => 10,
		     "vb" => 13,
		     "redhook" => 14,
		     "krakus" => 28,
		     "fullsail" => 29,
		     "scrimshaw" => 36,
		     "anchorsteam" => 39,
		     # (Bidir Slide Ruler, branched from Ranchorsteam)
		     "bsr" => 54,
		     "tsingtao" => 56,
		     "tt" => 56,
		     "gordonbiersch" => 64,
		     "tricky" => 74,
		     "kestrel" => 76,
		     "ironcity" => (77 << 8),
		     #storagemotion
		     "storagemotionn" => (77 << 8),
		     #  "REP" versions are based on ironcity  #
		     "REP" =>  (77 << 8),
		     "REP2" => (77 << 8),
		     "boilermaker" => (82 << 8),
		     "compress" => (82 << 8),
		     "boilerqan" => (82 << 8),
		     # Added the wafl version R8.0xN (Burt 374935)
		     "R8.0"    => (82 << 8),
 		     "Rbr"     => (82 << 8),			 
		     # Added wafl versions for different RR branches (Burt 412905)					 
		     "rollingrock" => (87 << 8),
  		     "wmdb" => (87 << 8),
		     "palomarr" => (87 << 8),
		     "fsb" => (87 << 8),
		     "waflmainrr" => (87 << 8),
		     "semainrr" => (87 << 8),
		     "whitealligator" => (87 << 8),
		     "SAN" => (87 << 8),
		     "nascon" => (87 << 8),
		     "scon" => (87 << 8),
		     "rr-platstor" => (87 << 8),
		     "pie" => (87 << 8),
		     "R8.1"    => (87 << 8),	 
		     # Changes for BURT 492060
		     "hya" => (87 << 8),
		     "rr0_fixes"  => (87 << 8),
		     "rr1_fixes"  => (87 << 8),
		     "main" => (92 << 8),
		     "sierranevada" => (92 << 8),
		     "warm" => (92 << 8),
	         "sn_" => (92 << 8),  # for all the SN branches like sn_platstor, sn_protocols, sn_warm_*, sn_ib_*
		     "R8.2"    => (92 << 8),
		     "sn-" => (92 << 8),
             "dev" => (97 << 8),
             "fullsteam" => (97 << 8),
 
		);

# Define constants for each of the above
while (my ($key, $value) = each %Wafl_Releases) {
    eval "sub $key { $value }";
}

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = (
    'get_wafl_version', 'get_ontap_version', 'ontap_to_wafl_ver',
    'version_compare', 'codename_to_version',
    'version_contained', 'version_allowed', keys %Wafl_Releases,
    'supported_feature', 'are_we_ontapng', 'is_ontapng_version'
);
push @EXPORT_OK, qw(UNSUPPORTED_FEATURE_EXCEPTION);
push @EXPORT_OK, qw(REVERT_TOO_FAR_BACK);
push @EXPORT_OK, qw(REVERT_NO_RAIDDP);
push @EXPORT_OK, qw(REVERT_NO_FLEX);
our %EXPORT_TAGS =         # Tag groups
    ("releases" => [keys %Wafl_Releases] );

use constant PRIVATE_VERSION => 'private_version';
use constant UNSUPPORTED_FEATURE_EXCEPTION => 'This version of OnTap does not support the requested operation';
use constant REVERT_TOO_FAR_BACK => 'revert_to: You can only revert to Data ONTAP ';
use constant REVERT_NO_RAIDDP => 'There are raid_dp volumes present which cannot be reverted';
use constant REVERT_NO_FLEX => 'Aggregates and flexible volumes cannot be reverted';


# zapi can fail under so many circumstances, so shut it up
# since we dont want FAIL results showing up from libs
sub _fail_default {
      my $msg = shift;
      logDbg("zapi failed: $msg ($@)");
}

sub get_wafl_version
{
    my ($host) = @_;
    # for most of the normal cases, we can grab fsinfo via zapi
    my $version;
    my $shortvers;

    # if filer is down, assume we are on the ok prompt
    if (!host_is_alive($host->default_ip(), 10)) {
	logWarn("host ".$host->default_ip()." is not pingable");
	# we are at OK prompt, get version string via SAY
	$version = _get_version_string($host, 30); #probably dead anyways
	$version = _convert_string_version_to_num($version);
    } else {
	#let's try to get wafl version via zapi
	eval {
	    $version = _get_wafl_version($host); #does only zapi
	};
	if ($@) {
            logDbg("_get_wafl_version() returned $@");
	    $@ = undef;
	    # get version string via zapi
	    eval {
		$version = _system_get_version($host);
		$version = _convert_string_version_to_num($version);
	    };
	    if ($@) {
                logDbg("_system_get_version() returned $@");
		# get version string via SAY
		$@ = undef;
		$version = _get_version_string($host);
		$version = _convert_string_version_to_num($version);
	    }
	}
    }
    if (!$version) {
	croak "could not get version: $@";
    }
    $@ = undef;
    return $version;
}

#   Determines if test is using filer in 10-mode.
sub are_we_ontapng
{
    my ($conn) = @_;
    if ($conn->hostp()->hosttype() =~ m/filer-ng/ ) {
	eval "use Server";
	croak($@) if ($@);
	return new Server();
    } else {
	return undef;
    }
}

# Determines if filer is running Clustered ONTAP
#   (test may be using either 7-mode or 10-mode "personality")
#   See BURT 270865
sub is_ontapng_version
{
	my ($host, %opts) = @_;
	my $conn = make_conn($host);
	$host = make_host($host);
	my $timeout = $opts{timeout};
	$timeout = 60 if (!defined($timeout));
	my $version;
	# $version = _get_version_string($conn);
	# eval { (undef, undef, $version) = say($conn, "version", -eval_error, '-timeout', $timeout); };
	# $version = get_ontap_version($host);
	$version = _system_get_version($host);
	croak("Could not get filer version: $@") if $@;
	#  BURT 279959
	my $ng = ($version =~ /mainng/) ||
		($version =~ /boiler/i) ||
		($version =~ /^RBR/) ||
		($version =~ /rolling/i) ||
		($version =~ /^R?RRR/) ||
		($version =~ /tricky/i) ||
		($version =~ /GX/) ||
		($version =~ /^R?8/) ||
		($version =~ /^R?10/) ||
		($version =~ /ontapng/) ||
		($version =~ /NG/) ||
		undef;
	my $classic = ($version =~ /ironcity/i) ||
		  ($version =~ /^R?IC/) ||
		  ($version =~ /gordonbiersch/i) ||
		  ($version =~ /^R?GB/) ||
		  ($version =~ /tsingtao/i) ||
		  ($version =~ /^R?TT/) ||
		  ($version =~ /anchorsteam/i) ||
		  ($version =~ /^R?AS/) ||
		  ($version =~ /^R?7/) ||
		  undef;
	if (defined($classic)) {
		$ng = $classic?0:1;
	}
	if (!defined($ng)) {
		logDbg("Filer::Version::IS_ONTAPNG_VERSION   Could not determine 'NG' version -- assuming 'NG'");
		$ng = 1;
	}

	my $ng_str = (defined($ng))?$ng:"UNDEF";
	logDbg("Filer::Version::IS_ONTAPNG_VERSION   Line: ".__LINE__."\n    VERSION: \"$version\"   NG?: $ng");
	return($ng);
}

sub ontap_to_wafl_ver {
    my $version = shift;
    my $wafl_ver;
    eval {
        $wafl_ver = _convert_string_version_to_num($version);
	};
    if ($@) {
        croak "Error from calling _convert_string_version_to_num(): $@";
    }
    return $wafl_ver;
}

sub _convert_string_version_to_num
{
    my $version = shift || croak "no version provided";
    return PRIVATE_VERSION if ($version eq PRIVATE_VERSION);
    if ($version =~ /^(\d+\.\d+)\S*/) {
	my $shortvers = $1;
	if ($shortvers =~ /^[0-5]/) {
	    $version =  $Wafl_Versions{"pre-6"};
	}
	## fix burt436762: to support BR and beyond
	elsif ($shortvers > 5) {
	    $version = $Wafl_Versions{$shortvers};
	} else {
	    croak "unknown version $version";
	}
		
	return $version;
		
    } else {
	foreach my $key (keys %Wafl_Versions) {
	    if ($version =~ /\Q$key\E/i) {
		$version =  $Wafl_Versions{$key};
		return $version;
	    }
	}
    }
    croak "unknown version $version";
}

sub _system_get_version
{
    my ($host) = @_;
    my $ver;
    eval {
	$ver = _ati_system_get_version($host, ATI_ONTAPI_TRANSPORT_HTTP, ATI_ONTAPI_VERSION_DEFAULT);
    };
    if ($@) {
	logDbg("transport_http ontapi failed: $@");
	eval {
	    $ver = _ati_system_get_version($host, ATI_ONTAPI_TRANSPORT_D, ATI_ONTAPI_D_VERSION_DEFAULT);
	};
	if ($@) {
	    logDbg("transport_d ontapi failed: $@");
	    croak ("could not get version : $@");
	}
    }
    # make the return version string be consistent with
    # _get_version_string() by removing the 'NetApp Release'
    # prefix string.
    if ($ver =~ /(?:NetApp|Data ONTAP) Release (\d+\.\d+\S*)\s*\S*:/) {
	$ver = $1;
    } elsif ($ver =~ /(?:NetApp|Data ONTAP) Release.+D-Blade/) {
        #temporary hack until more ZAPI's are implemented
	$ver = "tricky";
    } elsif ($ver =~ /(?:NetApp|Data ONTAP) Release (\w+\S+)\s*\S*:/) {
	$ver = $1;
    } elsif ($ver =~ /(?:NetApp|Data ONTAP) Release/) {
	# we are running against someone's private build
	# lets assume that zapi works on it
	$ver = PRIVATE_VERSION;
    }
    return $ver;
}

sub _ati_system_get_version
{
    my ($host, $transport, $version) = @_;
    # ati_ontapi_create_context will die if passwd is undef
    my $passwd = $host->default_password();
    $passwd = ATI_ONTAPI_PASSWORD_DEFAULT if (!$passwd);
    my $context = ati_ontapi_create_context(filer     => $host->default_ip(),
					    transport => $transport,
					    version   => $version,
					    user      => "root",
					    password  => $passwd);
    my @out = ati_ontapi_call_api(context => $context,
				  api => 'system-get-version');
    ($out[0] == ATI_SUCCESS) || croak "Error $out[2]: $out[3]\n";
    return $out[1]->{'version'};
}



sub _get_wafl_version
{
    my ($host) = @_;

    # this only works from 6.5.1 and later
    my $zapi = Zapi->new($host->default_ip(),
			 "root",
			 $host->default_password());
    $zapi->{fail} = \&_fail_default;

    my $reply = $zapi->invoke("volume-get-root-name");
    if ($reply->{attrivals}->[0] =~ /failed/) {
        croak "Could not invoke volume-get-root-name";
    }

    my $volname = $reply->child_get_string("volume");
    if (!$volname) {
	  croak ("Could not get root volume name" . "Zapi reply was: " . $reply->results_reason());
    }

    # perf-object-get-instances has two levels of elements in its
    # input, so it's too complex for "invoke", so we build the request
    # here and use "invoke_elem".

    # <perf-object-get-instances>
    #    <objectname>volume</objectname>
    #    <instances> <instance>vol0</instance> </instances>
    #    <counters> <counter>wv_fsinfo_fs_version</counter> </counters>
    # </perf-object-get-instances>

    my $objectname = NaElement->new("objectname","volume");
    my $instances = NaElement->new("instances","");
    $instances->child_add_string("instance",$volname);
    my $counters = NaElement->new("counters","");
    $counters->child_add_string("counter","wv_fsinfo_fs_version");
    my $perf_object_get_instances = NaElement->new("perf-object-get-instances","");
    $perf_object_get_instances->child_add($objectname);
    $perf_object_get_instances->child_add($instances);
    $perf_object_get_instances->child_add($counters);

    $reply = $zapi->invoke_elem($perf_object_get_instances,"passed");

    # <results status="passed">
    #    <timestamp>1101814790</timestamp>
    #    <instances>
    #      <instance-data>
    #        <name>vol0</name>
    #        <counters>
    #          <counter-data>
    #            <name>wv_fsinfo_fs_version</name>
    #            <value>65</value>
    #          </counter-data>
    #        </counters>
    #      </instance-data>
    #    </instances>
    # </results>

    my $elem=$reply;
    foreach my $name (qw(instances instance-data counters counter-data)) {
	$elem=$elem->child_get($name);
	if (!$elem) {
	    if (defined $reply->results_reason()) {
	        croak( "Could not get root volume data. " ,
		       "Element $name missing. " ,
		       "Zapi reply was: " . $reply->results_reason());
	  }
	    else {
	         croak ( "Could not get root volume data. " ,
			 "Element $name missing");
	       }
	}
    }
    my $version = $elem->child_get_int("value");
    return $version;
}

# return a X.X version from the running system
sub get_ontap_version
{
    my ($host) = @_;
    # for most of the normal cases, we can grab fsinfo via zapi
    my $version;

    # the order of what to call first to get the version is almost
    # identical to get_wafl_version().

    # if filer is down, assume we are on the ok prompt
    if (!host_is_alive($host->default_ip(), 10)) {
	logWarn("host ".$host->default_ip()." is not pingable");
	# we are at OK prompt, get version string via SAY
	$version = _get_version_string($host, 10);

    } else {
	# let's try to get wafl version via zapi
	my $wv;
	eval {
	    $wv = _get_wafl_version($host);
	};

	if (!$@) {
	    return _map_wafl_to_release($wv);

	} else {
            logDbg("_get_wafl_version() returned $@");
	    $@ = undef;
	    # get version string via zapi
	    eval {
		$version = _system_get_version($host);
	    };

	    if ($@) {
                logDbg("_system_get_version() returned $@");
		# get version string via SAY
		$@ = undef;
		$version = _get_version_string($host);
	    }
	}
    }

    logDbg("get_ontap_version: got $version from $host");
    $@ = undef;

    if ($version =~ /^(\d+\.\d+)/) {
	# return numeric versions immediately
	return $1;
    } elsif ($version =~ /^R?\w+N?_\d+/) {
        my $vers = codename_to_version($version);
        $vers =~ /^(\d+\.\d+).*/;
        return $1;
    } elsif ($version =~ /^R(\d+\.\d+).*N_\d+/) { 
        return $1;
    } else {
	croak "can't determine numeric release: filer down/innacessible?";
    }
}

sub _map_wafl_to_release
{
    my ($wv) = @_;
    # gag -- brute force method...
    # XXX add release number for Tricky and Kestrel when it's available
    if ($wv >= WV_FULLSTEAM_MIN()) {
     return '8.3';
    }elsif ($wv >= WV_SIERRANEVADA_MIN() and $wv <= WV_SIERRANEVADA_MAX()) {
     return '8.2';
    }elsif ($wv >= WV_ROLLINGROCK_MIN() and $wv <= WV_ROLLINGROCK_MAX()) {
     #Take if from http://wikid.netapp.com/w/RollingRock
     return '8.1';
    }elsif ($wv >= WV_BOILERMAKER_MIN() and $wv <= WV_BOILERMAKER_MAX() ) {
	# What should this be?
	return '8.0';
    } elsif ($wv >= WV_IRONCITY_MIN() and $wv <= WV_IRONCITY_MAX()) {
	return '7.3';
    } elsif ($wv >= WV_GORDONBIERSCH_MIN() and $wv <= WV_GORDONBIERSCH_MAX()) {
	return '7.2';
    } elsif ($wv >= WV_TSINGTAO_MIN() and $wv <= WV_TSINGTAO_MAX()) {
	return '7.1';
    } elsif ($wv >= WV_ANCHORSTEAM_MIN() and $wv <= WV_ANCHORSTEAM_MAX()) {
	return '7.0';
    } elsif ($wv >= WV_SCRIMSHAW_MIN() and $wv <= WV_SCRIMSHAW_MAX()) {
	return '6.5';
    } elsif ($wv >= WV_FULLSAIL_MIN() and $wv <= WV_FULLSAIL_MAX()) {
	return '6.4';
    } elsif ($wv >= WV_KRAKUS_MIN() and $wv <= WV_KRAKUS_MAX()) {
	return '6.3';
    } elsif ($wv >= WV_REDHOOK_MIN() and $wv <= WV_REDHOOK_MAX()) {
	return '6.2';
    } elsif ($wv >= WV_VB_MIN() and $wv <= WV_VB_MAX()) {
	return '6.1';
    } elsif ($wv >= WV_GUINNESS_MIN() and $wv <= WV_GUINNESS_MAX()) {
	return '6.0';
    } else {
	return '5.0';		# rather arbitrary
    }
}

sub _get_version_string
{
    my ($host) = shift || croak "host is not provided";
    my $timeout = shift || 60;

    # get the output of the version command...
    my ($version, $conn);
    if (ref($host) =~ /Hostrec|Connectrec/) {
	$conn = make_conn($host, '#rsh');
	$host = make_host($host);
    } else {
	croak("host arg is invalid type: " . ref($host) .
	      ". Expecting Hostrec or Connectrec");
    }
    my ($match, undef, $text);
    eval { ($match, undef, $text) = say($conn, "version", -eval_error,
					'-timeout', $timeout); };
    if ($text and $text =~ /(?:NetApp|Data ONTAP) Release (\d+\.\d+\S*)\s*\S*:/) {
	$version = $1;
    } elsif ($text and $text =~ /(?:NetApp|Data ONTAP) Release (\w+\S+)\s*\S*:/) {
	$version = $1;
    } elsif ($text and $text =~ /(?:NetApp|Data ONTAP) Release/) {
	# we are running against someone's private build
	# lets assume that zapi works on it
	$version = PRIVATE_VERSION;
    } else {
	# perhaps in firmware?
	# (only works with #console connection)
	# makes no sense for sims
	if ($host->hosttype() =~ /simulator/ ) {
	    croak "can not get version of ".$host->hostname;
	}
	$conn = make_conn($host, '#console');
	($match, undef, $text) = say($conn, "printenv last-OS-booted-ver",
				     "printenv", -prompt, $host->prompt_fw(),
				     -eval_error, -timeout, $timeout);
	croak $@ if $@;
	if ($text =~ /(?:NetApp|Data ONTAP) Release (\S+)\s*/) {
	    $version = $1;
	} else {
	    croak("unknown return from version command: $text");
	}
    }
    logDbg("_get_version_string() => $version");
    return $version;
}

# sub codename_to_version
#
# Takes a codename string (eg: "*anchorsteam*") and returns an appropriate
# 3-digit version number. Since the named streams are expected to be the
# latest version of that code-line, a ".90" sub-version number is appended

sub codename_to_version
{
    my ($version) = @_;
    for my $name (grep(/^[a-z]/i, keys %Wafl_Versions)) {
	if ($version =~ quotemeta($name)) {
	    return _map_wafl_to_release($Wafl_Versions{$name}).".90";
	}
    }
    return $version;
}


# --------------------
# sub version_compare

# version_compare is a function that takes two netapp version strings
# and converts them to a representative integer. If the first version
# is higher than the second version, version_compare returns 1. If
# they are equal, it returns 0. If the first version is lower than the
# second version, it returns -1 See the documentation for
# _integer_version for more information about acceptable string
# formats Currently, it croaks if the format is wrong. This needs to
# be fixed.

# --------------------

sub version_compare
{
    my ($v1, $v2, $pos) = @_;
    $pos = 2 unless defined($pos);
    logDbg("version_compare($v1, $v2, $pos)");
    my $result = _integer_version($v1, $pos) - _integer_version($v2, $pos);
    if ($result > 0 ) {
	return 1
    } elsif ($result < 0 ) {
	return -1
    } elsif ($result == 0) {
	return 0
    } else {
	return undef
    }
}

# --------------------
# sub _integer_version

# Versioning follows very strict rules here
# If a word is supplied, then we attempt to look up the version
# (1) Versions are split on a single period
# (2) No letters or other characters are allowed
# (3) Only the major and minor digits are significant
# (4) A partial version of .11 is greater than a partial version of .2
# but a partial version of .20 is greater than a partial version of .11
# (i.e. right-padded oeroes are significant)

# --------------------

sub _integer_version
{
    my $version = shift;
    my $positions = shift || 2;
    $version =~ s/(X|RC|P|D|H)(\d+).*$/.$2/;
    my @partials = (split /\./, $version)[0..$positions];
    if (grep /\D/, @partials) {
	# attempt to convert the string to a num
	@partials = (split /\./, codename_to_version($version))[0..$positions];
	# let anything else fall through?
    }
    my $intvers = join "", map { sprintf("%06d", $_) } @partials;
    logDbg("_integer_version($version) => $intvers (partials=@partials)");
    return ( $intvers ) ||
      croak "Could not split version '$version' into partials";
}

# ------------------
# test if arg1 is contained within the list of the remaining args,
# when all are mapped by _integer_version()

sub version_contained
{
    my ($vers, @verslist) = @_;
    if (grep { _integer_version($vers) == $_ }
	map { _integer_version($_) } @verslist ) {
	return 1
    } else {
	return 0
    }
}

# ------------------------
# Test if the current filer version is within the range supplied.
#

sub version_allowed
{
    my ($host, %args) = @_;
    my $current_version = $args{current_version} || _get_version_string($host);
    my $min_version = $args{min_version} || undef;
    my $max_version = $args{max_version} || undef;
    my $on_error = $args{on_error} || 'na_result';
    my $min_v_or_more = 0;
    my $max_v_or_less = 0;
    my $errstr = "running version $current_version is";

    if (defined($min_version)) {
	if (version_compare($min_version, $current_version) <= 0) {
	    $min_v_or_more = 1;
	} else {
	    $errstr .= " older than minimum version $min_version";
	}
    } else {
	$min_v_or_more = 1;
    }
    if (defined($max_version)) {
	if (version_compare($max_version, $current_version) > 0) {
	    $max_v_or_less = 1;
	} else {
	    $errstr .= "newer than maximum version $max_version";
	}
    } else {
	$max_v_or_less = 1;
    }

    if ($max_v_or_less != 0 && $min_v_or_more != 0) {
	return 1;
    } elsif ($on_error eq 'ignore' or $on_error eq 'return') {
	return 0;
    } elsif ($on_error eq 'config_result') {
	die(config_result($errstr));
    } else {
	die(na_result($errstr));
    }
    # NOTREACHED
}

my %feature_hash = (
    AGGR => {
	MIN_VERSION => WV_ANCHORSTEAM_MIN(),
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 0,
    },
    LARGER_AGGR => {
	MIN_VERSION => WV_BOILERMAKER_MIN(),
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 0,			# should this be 1 ?
    },
    FLEXVOL => {
	MIN_VERSION => WV_ANCHORSTEAM_MIN(),
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 0,
    },
    FAKE_LARGER_AGGR => {
	MIN_VERSION => WV_BOILERMAKER_MIN(),
	MAX_VERSION => WV_BOILERMAKER_MAX(),
	HARDWARE_ONLY => 0,
	callback  => sub {
		      my ($host) = @_;
		      my $fc = make_conn($host) || die $@;
		      my $out = sayEOF($fc, 'priv set -q test; bootargs print raidtop_fake_rg_max_count');
		      return 1 if ($out=~/(\d+)/)[0] > 2; # Must be no less then 3, but its recommended to set
		      return 0;				  # to 10 or filer will go to panic loop if vol > 50Tb 
	}
    },
    SET_SNAPLOCK_RETENTION => {
	MIN_VERSION => WV_ANCHORSTEAM_MIN(),
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 0,
    },
    SET_CLOCK => {
	MIN_VERSION => WV_NEG_INFINITY,
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 1,
    },
    SET_COMPLIANCE_CLOCK => {
	MIN_VERSION => WV_ANCHORSTEAM_MIN(),
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 0,
    },
    SCHEDSNAPNAME=> {
	MIN_VERSION => WV_GORDONBIERSCH_MIN(),
	MAX_VERSION => WV_POS_INFINITY,
	HARDWARE_ONLY => 0,
    },
   HOTDISK=> {
	MIN_VERSION =>WV_ANCHORSTEAM_MIN(), #available in AS.2
	      MAX_VERSION => WV_POS_INFINITY,
	      HARDWARE_ONLY => 0,
	      callback  => sub {
		      # XXX this assumes that the filer is up,
		      # but will have to live with this for now
		      my ($host) = @_;
		      my $fc = make_conn($host)
			      or die $@;
		      my $out = sayEOF($fc, 'priv set test; vol disk -help');
		      if ($out =~ 'vol disk hot') {
			      return 1;
		      }
		      return 0;
	      }
   },

);

sub supported_feature {
    my ($host, $feature, $msgref) = @_;

    if (!exists $feature_hash{$feature}) {
	croak "unknown feature: $feature";
    }

    my $min_version = $feature_hash{$feature}{MIN_VERSION};
    my $max_version = $feature_hash{$feature}{MAX_VERSION};

    my $version = get_wafl_version($host);

    # If PRIVATE_VERSION found always return true.
    if($version eq PRIVATE_VERSION) {
	logWarn("Private version found. supported_feature couldn't".
	       "figure out if this feature ($feature) is supported".
	       "in this private build ($version). Returning true.");
	return 1;
    }

    #my $later_than_min = 1;
    #$later_than_min = ($version >= $min_version) if (defined($min_version));
    my $later_than_min = ($version >= $min_version);

    #my $earlier_than_max = 1;
    #$earlier_than_max = ($version <= $max_version) if (defined($max_version));
    my $earlier_than_max = ($version <= $max_version);

    my $needs_hw = $feature_hash{$feature}{HARDWARE_ONLY};
    my $supported_by_environment = ! ($needs_hw && ($host->hosttype() =~ /simulator/));

    logDbg("later_than_min is '$later_than_min'");
    logDbg("earlier_than_max is '$earlier_than_max'");
    logDbg("needs_hw is '$needs_hw'");
    logDbg("supported_by_environment is '$supported_by_environment'");

    my $result = ($later_than_min && $earlier_than_max && $supported_by_environment);

    if ($result && defined($feature_hash{$feature}{callback})) {
	    # we match in version numbers but still need to check
	    # with callback (since the feature might be supported in .2
	    # release for example)
	    $result =  $feature_hash{$feature}{callback}($host);
    }

    if(!$result) {
	${$msgref} = UNSUPPORTED_FEATURE_EXCEPTION;
	${$msgref} .= ". The feature \"$feature\" is not supported in WAFL ";
	${$msgref} .= "Version $version. It is supported only between ";
	${$msgref} .= "WAFL version $min_version and $max_version. ";
	if(!$supported_by_environment) {
	    ${$msgref} .= "Environment is not supported. Can run only on hardware";
	}
    }

    return $result;
}

1;

=head1 NAME

Filer::Version - Get filer internal version numbers

=head1 SYNOPSIS

    use Filer::Version qw(get_wafl_version WV_ANCHORSTEAM_MIN);
    my $fh = host($FILER);

    if (get_wafl_version($fh) >= WV_ANCHORSTEAM_MIN) {
        ...
    }

=head1 DESCRIPTION

This library defines methods to read and manipulate internal version
numbers from filers, suitable to help a test tailor what it expects
based on which release happens to be running.

Internal versions are used because the usual release or version
strings (as printed by 'version' or found in prod/build/vers-*.c) are
based on the current directory name, current time, etc., which depend
entirely on the environment where the build was performed.  Automated
and promoted builds follow a pattern, naturally, but developer-built
(privately-built) trees do not,

So far, this only includes getting the WAFL version number.

=head1 FUNCTIONS

=over

=item get_wafl_version

    my $version = get_wafl_version($host);

B<get_wafl_version> returns the filesystem version native to the
software currently running.  C<$host> is a host record object (created
using C<Tharn::host>).  get_wafl_version dies if it cannot get the
version.

The WAFL version used is the same value as WAFL_FSINFO_VERSION in
prod/common/wafl/fsinfo.h.  Developers only increment this when making
non-backward-compatible filesystem changes.  However, because it has
so far been incremented at least once (and usually many more times)
per major release, it is usually sufficient to distinguish major
releases in general.

=item get_ontap_version

    my $version = get_ontap_version($host);

B<get_ontap_version> returns a string containing the ONTAP software
N.M major version of the software currently running.  C<$host> is a
host record object (created using C<Tharn::host>).  get_ontap_version
dies if it cannot get the version.

If the filer's version command returns a string that does not look
like a version number, it falls back to calling get_wafl_version to
try to deduce the ONTAP major version based on the WAFL version.

=item version_compare

    my $comp;
    eval { $comp = version_compare($vers1, $vers2, 3); };
    if ($@) {
        # do something with unknown version string
    } elsif ($comp > 0) {
        # version string 1 is later than version string 2
    }

B<version_compare> takes 2 ontap version strings and returns 1, 0, -1
if $vers1 is greater than, equal to, or less than $vers2 respectively.
The results are predictably correct if the versions are of the form
B<N.M> or B<N.M.O>, and will be best try accurate using
B<codename_to_version>. Currently croak()s if it can't figure out both
version numbers.

=item codename_to_version

    my $version = codename_to_version("RtsingtaoN");

B<codename_to_version> takes a codename string (eg: "RanchorsteamN")
and returns an appropriate 3-part numeric version number (eg:
"7.0.90"). Since the named streams are expected to be the latest
version of that code-line, a ".90" sub-version number is appended. If
the argument doesn't contain a codeline name, the argument will be
returned unchanged.

=item version_contained

    if (version_contained($vers1, @vers_list)) {
        # match found
    }

B<version_contained> tests if $vers1 is contained within @vers_list,
the list of the remaining arguments.

=item version_allowed

    if (version_allowed($host, min_version=>"7.0", max_version=>"7.2.1")) {
        # we're fine to go!
    }

B<version_allowed> tests if the running version (or optional argument
current version) lies between the versions that the script knows that
the feature exists in.

=item WV_ constants

   WV_GUINNESS_MIN WV_GUINNESS_MAX
   WV_VB_MIN WV_VB_MAX
   WV_REDHOOK_MIN WV_REDHOOK_MAX
   WV_KRAKUS_MIN WV_KRAKUS_MAX
   WV_FULLSAIL_MIN WV_FULLSAIL_MAX
   WV_SCRIMSHAW_MIN WV_SCRIMSHAW_MAX
   WV_ANCHORSTEAM_MIN WV_ANCHORSTEAM_MAX
   WV_TSINGTAO_MIN WV_TSINGTAO_MAX
   WV_GORDONBIERSCH_MIN WV_GORDONBIERSCH_MAX
   WV_TRICKY_MIN WV_TRICKY_MAX
   WV_KESTREL_MIN WV_KESTREL_MAX

These constants represent the minimum and maximum wafl fsinfo versions
that a given major release has ever had or ever could have.  (There is
a range because the releases may have had multiple different versions
as they were developed, or may have more as development continues)

Generally one should test the filer version against the _MIN
constants, not the _MAX constants.  The _MAX values for the very
latest releases may not be defined yet and are more open to change.
The maximum values may be the maximum ever reserved for a release, not
the specific value that was shipped, so you generally should not
attempt to skip the range concept and test for equality.

=item Other constants

    UNSUPPORTED_FEATURE_EXCEPTION

This string constant is thrown when an request cannot be fulfilled because
the operation is not supported by the version of OnTap currently running.

=back

=head1 AUTHOR

 dl-central-libs
 $Id: //depot/prod/test/Rsierranevada/lib/Filer/Version.pm#3 $
 Last revised $Date: 2013/11/01 $

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = version

    ONTAP_ZAPI_APIS = system-get-version, volume-get-root-name,
    perf-object-get-instances

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

