#!perl -w

package Finisar::Usage;

# Author:      Daniel Coupal
# Date:        October 1999

# Description: Module that isolate the work around
#              processing options
#
#              In the Main module, you define
#                 $description
#                 $implementation
#                 $usage
#                 $help
#                 @options
#
#              Execute your script with the option "-checkusage"
#              to validate the use of that module

# TODO - Parse OR/AND/... options based on usage

use Exporter;
use Getopt::Long;

use strict;
use warnings;

use vars qw(@ISA @EXPORT %OPT); 

@ISA = qw(Exporter);
@EXPORT = qw(newGetOptions %OPT printUsage);

# Options that the script designer does not have to define 
my $defaultOptions = "doc help usage";
my $hiddenOptions  = "checkusage";

# This subroutine checks if the usage is coherent
# with the help and examples
sub checkUsage() {

    my %definedOptions;
    my $oneOption;
    my $oneOptionName;

    # Check if the variables exist.
    unless ( defined($main::description) ) {
        print("ERROR - \$description is not defined\n");
    }
    unless ( defined($main::help) ) {
        print("ERROR - \$help is not defined\n");
    }
    unless ( defined($main::implementation) ) {
        print("ERROR - \$implementation is not defined\n");
    }
    unless ( defined($main::usage) ) {
        print("ERROR - \$usage is not defined\n");
    }
    unless ( defined(@main::options) ) {
        print("ERROR - \@optionsList is not defined\n");
    }

    # Check if all @options defined are in the "usage" and "help" documentation
    foreach $oneOption ( @main::options ) {
        $oneOptionName = $oneOption;
        $oneOptionName =~ s/[:=].*$//;
        $definedOptions{$oneOptionName}++;

        if ( $main::usage !~ /$oneOptionName/ ) {
            print "ERROR - \"$oneOptionName\" does not appear in the usage\n";
        }
        if ( $main::help  !~ /\n\s*-$oneOptionName/ ) {
            print "ERROR - \"$oneOptionName\" does not appear in the help\n";
        }
    } 

    # Check if all the options referenced in $usage exist
    while ( $main::usage =~ /-(\w+)/g ) {
        $oneOptionName = $1;
        if ( ! exists $definedOptions{$oneOptionName} &&
             " $defaultOptions " !~ / $oneOptionName /   ) {
            print "ERROR - \"$oneOptionName\" referenced in \$usage does not exist\n";
        }
    }

    # Check if all the options referenced in $help exist
    while ( $main::help =~ /\s-(\w+)/g ) {
        $oneOptionName = $1;
        if ( ! exists $definedOptions{$oneOptionName} &&
             " $defaultOptions " !~ / $oneOptionName /   ) {
            print "ERROR - \"$oneOptionName\" referenced in \$help does not exist\n";
        }
    }
}


# Subroutine that is a wrapper around the "GetOptions"
# that comes with Perl. 
# Returns an associative array "%OPT" with the values 
# for all switches 
sub newGetOptions() {

    my $optionList;
    my $oneOption;
    my $oneOptionName;

    $optionList = "\"usage\"      => \\\$OPT{'usage'}," .
                  "\"help\"       => \\\$OPT{'help'},"  .
                  "\"doc\"        => \\\$OPT{'doc'},"   .
                  "\"checkusage\" => \\\$OPT{'checkusage'},";

    foreach $oneOption ( @main::options ) {
        $oneOptionName = $oneOption;
        $oneOptionName =~ s/[:=].*$//;
        $optionList .= "\"$oneOption\" => \\\$OPT{'$oneOptionName'},";
    }

    $optionList =~ s/,$//;

    if ( ! GetOptions(eval($optionList)) ) {

        print($main::usage);
        open(STDERR,">/dev/null");
        exit(2);
    } elsif ( $OPT{'usage'} ) {
        print($main::usage);
        open(STDERR,">/dev/null");
        exit(1);
    } elsif ( $OPT{'help'} ) {
        print($main::usage);
        print($main::description);
        print($main::help);
        open(STDERR,">/dev/null");
        exit(1);
    } elsif ( $OPT{'doc'} ) {
        print($main::usage);
        print($main::description);
        print($main::help);
        print($main::implementation);
        open(STDERR,">/dev/null");
        exit(1);
    } elsif ( $OPT{'checkusage'} ) {
        checkUsage();
        open(STDERR,">/dev/null");
        exit(1);
    }
}

sub printUsage() {

    print("$main::usage");
    return();
}


1;
