#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/7Mode/Vserver.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Vfiler-level functionality for 7Mode.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::7Mode::Vserver

=head1 DESCRIPTION

This module provides release agnostic vfiler-level functionality for 7Mode.
This module should only be used internally by the FlexCache library.
It is NACL-compliant. To be noted, though, that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

=head1 SYNOPSIS

N/A.

=cut

package FlexCache::7Mode::Vserver;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use NATE::BaseException qw(:try);
use NACL::STask::Node;
use NACL::CS::Options;
use NACL::C::Options;
use NACL::STask::SystemLicense;
use NACL::C::Vfiler;
use NACL::C::Ifconfig;

# Constructor for the FlexCache::7Mode::Vserver class.
# It is only intended to be used within the FlexCache library.
# The constructor will obtain the dlif information from the tharnhost file,
# to be used for NFS setup.
# If the running in a multistore environment,
# it will create/reuse the vfiler and configure the vfiler's interface.
# Otherwise, no filer configuration is changed.
# Unless the attach_only parameter is true, it will also execute the NFS setup
# (CIFS is not supported for 7-mode).
# Arguments: They should always be kept identical (+- optionality)
#            to the allowed arguments for the constructor of the
#            FlexCache::Vserver class,
#            even if some of them may be ignored for 7Mode.
#            These arguments should be documented in the
#            FlexCache::Vserver module (externally-available)
# Return: FlexCache::7Mode::Vserver object.
#
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name => {type => SCALAR | UNDEF, optional => 1,},
            node           => {type => OBJECT, isa => "NACL::STask::Node",},
            aggregate_name => {type => SCALAR,},
            dlif_name => {type => SCALAR | UNDEF, optional => 1,},
            dlif_id   => {type => SCALAR | UNDEF, optional => 1,},
            client_type => {type => SCALAR, regex => qr/^linux|windows/,},
            recycle =>
                {type => BOOLEAN, default => FCACHE_DFLT_VFILER_RECYCLE,},
            perform_cleanup => {type => BOOLEAN, default => TRUE,},
            debug           => {type => BOOLEAN, default => FALSE,},
            attach_only     => {type => BOOLEAN, default => FALSE,},
            ui_messages =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
            config_type => {type => SCALAR, regex => qr/^origin|cache$/,},
        }
    );

    my $self = {};
    $self->{_is_multistore}   = FALSE;
    $self->{_name}            = _FCACHE_VSERVER_NO_VFILER;
    $self->{_debug}           = $args{debug};
    $self->{_nacl_node}       = $args{node};
    $self->{_node_name}       = $self->{_nacl_node}->name();
    $self->{_recycle}         = $args{recycle};
    $self->{_attach_only}     = $args{attach_only};
    $self->{_perform_cleanup} = $args{perform_cleanup};
    $self->{_config_type}     = $args{config_type};
    $self->{_creator_pid}     = $$;
    $self->{_ui_messages}     = $args{ui_messages};
    $self->{_cleanup_options} = {};

    bless($self, $class);

    my $hostrec = $self->{_nacl_node}->hostrec();

    if (   (defined($args{name}))
        && ($args{name} ne _FCACHE_VSERVER_NO_VFILER)) {

        # Script needs to run in a multistore environment
        # In this case, we need to license multistore,
        # create/reuse the vfiler, etc
        $self->{_is_multistore} = TRUE;
        $self->{_name}          = $args{name};
        my $name = $self->{_name};

        fcache_log_comment(
            message => "SETUP: Multistore environment is required:\n"
                . "Vfiler name: $name");

        if ($self->{_attach_only}) {

            # Set the recycle flag, so we do not destroy the vfiler at cleanup
            $self->{_recycle} = TRUE;
            fcache_log_comment(
                message => "SETUP: Attempting to attach to vfiler $name");
            $self->{__nacl_vfiler} = NACL::C::Vfiler->find(
                command_interface => $self->{_nacl_node},
                filter            => {vfiler => $name,}
            );

            # Detect the IP address. Only look at the first interface
            my $state            = $self->{__nacl_vfiler}->state();
            my $ip_configuration = $state->ip_address_configuration()->[0];
            $self->{__dlif_port} = $ip_configuration->{interface};
            $self->{_dlif_ip}    = $ip_configuration->{ip_address};
            if (!defined($self->{_dlif_ip})) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_CONFIG,
                    message =>
                        "Could not detect IP address for vfiler $name:\n"
                        . Dumper($ip_configuration)
                );
            }
        } else {
            my $dlif_tharnhost_id = $args{dlif_id};
            if (   (!defined($dlif_tharnhost_id))
                || ($dlif_tharnhost_id !~ /^\d+$/)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Multistore environment was required:\n"
                        . "Vfiler name: $name\n"
                        . "The dlif_id argument MUST be provided\n"
                        . "And it must be an integer."
                );
            }

            my $hostrec_vfiler_name
                = $hostrec->{'VFILER_NAME' . $dlif_tharnhost_id};
            $self->{_dlif_ip} = $hostrec->{'VFILER_IP' . $dlif_tharnhost_id};
            $self->{__dlif_port}
                = $hostrec->{'VFILER_PORT' . $dlif_tharnhost_id};
            $self->{__dlif_mask}
                = $hostrec->{'VFILER_MASK' . $dlif_tharnhost_id};
            $self->{__hosting_node_dlif_ip} = $hostrec->{default_ip_7g};
            if (   (!defined($hostrec_vfiler_name))
                || ($hostrec_vfiler_name ne $name)
                || (!defined($self->{_dlif_ip}))
                || (!defined($self->{__dlif_port}))
                || (!defined($self->{__dlif_mask}))
                || (!defined($self->{__hosting_node_dlif_ip}))) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Error while extracting the VFILER"
                        . " information for $name\n"
                        . "Please make sure that the provided dlif_id"
                        . " is correct: $dlif_tharnhost_id\n"
                        . "and that the tharnhost file contains entries"
                        . " for all of the following:\n"
                        . "VFILER_NAME$dlif_tharnhost_id"
                        . " (MUST match with the vfiler name provided: $name)\n"
                        . "VFILER_IP$dlif_tharnhost_id\n"
                        . "VFILER_PORT$dlif_tharnhost_id\n"
                        . "VFILER_MASK$dlif_tharnhost_id\n"
                        . "default_ip_7g\n"
                        . "Current hostrec information:\n"
                        . Dumper($hostrec)
                );
            }
            fcache_log_comment(
                message => "SETUP: Obtained vfiler from the tharnhost file:\n"
                    . "Vfiler IP: "
                    . $self->{_dlif_ip} . "\n"
                    . "Vfiler port: "
                    . $self->{__dlif_port})
                if $self->{_debug};

            fcache_log_comment(
                message => "SETUP: Adding required multistore license");
            NACL::STask::SystemLicense->add(
                command_interface => $self->{_nacl_node},
                feature           => 'multistore',
            );

            my $nacltask_if_exists = ($self->{_recycle}) ? 'reuse' : 'purge';
            my $root_vol_name = $name . '_root';
            $root_vol_name =~ s/\-/_/g;

            fcache_log_comment(
                message => "SETUP: Creating/reusing the vfiler root volume:"
                    . " $root_vol_name");
            $self->{__vfiler_root_vol} = NACL::STask::Volume->create(
                command_interface  => $self->{_nacl_node},
                volume             => $root_vol_name,
                aggregate          => $args{aggregate_name},
                size               => FCACHE_DFLT_VFILER_ROOT_SIZE,
                nacltask_if_exists => $nacltask_if_exists
            );

            $self->_create_or_recycle_vfiler();
        }

        fcache_log_comment(message => "SETUP: Verifying vfiler interface for "
                . $self->{_name});
        my $api_set = $self->{_nacl_node}->get_7m_or_nodescope_apiset();
        $api_set->ping(destination => $self->{_dlif_ip},);
    } else {
        fcache_log_comment(
            message => "SETUP: Multistore environment not required");

        # In this case, the dlif_ip will map to the IP of the filer
        $self->{_dlif_ip} = $hostrec->{default_ip_7g};

        if (!defined($self->{_dlif_ip})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message =>
                    "Error while extracting the IP information for node "
                    . $self->{_node_name} . "\n"
                    . "Please make sure that the nate-host file"
                    . " contains entry for:\n"
                    . "default_ip_7g\n"
                    . "Current hostrec information:\n"
                    . Dumper($hostrec)
            );
        }

        fcache_log_comment(message =>
                  "SETUP: Obtained IP information from the tharnhost file:\n"
                . "Data lif IP: "
                . $self->{_dlif_ip})
            if $self->{_debug};
    }

    if ($self->{_attach_only}) {
        return $self;
    }

    if ($args{client_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        $self->_setup_cifs();
    } else {
        $self->_setup_nfs();
    }

    return $self;

} ## end sub new

=head1 AVAILABLE METHODS

=head2 get_name

Gets the name of this vfiler (only in a multistore environment).
If not in a multistore environment,
then the script execution is terminated with N/A.

=head3 ARGUMENTS

None.

=head3 RETURN

Name of the vfiler (Scalar).

EXAMPLE

	my $vfiler_name = $obj->get_name();

=cut

sub get_name {
    my $class = shift;
    my %args  = validate(@_,
        {_internal_use => {type => BOOLEAN, default => FALSE,},});

    if (($args{_internal_use}) || ($class->{_is_multistore})) {
        return $class->{_name};
    } else {

# Finish with NA, in the idea that get_name should be called
# by the script with a very good reason (to do something with the vserver/vfiler)
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message =>
                "This test should only be run in a multistore environment"
        );
    }
} ## end sub get_name

=head2 get_dlif_ip

Gets the IP of the data interface for this vfiler.
If the script is run in a multistore environment,
this will be the IP of the vfiler.
Otherwise, it will be the filer's IP.

=head3 ARGUMENTS

None.

=head3 RETURN

IP of the data interface for this vfiler (Scalar).

EXAMPLE

	my $mount_ip = $obj->get_dlif_ip();

=cut

sub get_dlif_ip {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_dlif_ip};
} ## end sub get_dlif_ip

=head2 set_flexcache_delegations_state

Enable/disable delegations for this node/vfiler.
Please note that for 7-mode there is no way to disable/enable only one
of the read/write delegations.

=head3 ARGUMENTS

=over

=item B<read_delegations> I<Optional> Scalar:

	State to set for the read-delegations.
	Legal values: FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED.
	Default value: FCACHE_STATE_ENABLED.

=item B<write_delegations> I<Optional> Scalar:

	State to set for the write-delegations.
	Legal values: FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED.
	Default value: FCACHE_STATE_ENABLED.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->set_flexcache_delegations_state(read_delegations => FCACHE_STATE_DISABLED,
	                                      write_delegations => FCACHE_STATE_DISABLED);
	$obj->set_flexcache_delegations_state(read_delegations => FCACHE_STATE_ENABLED,
	                                      write_delegations => FCACHE_STATE_ENABLED);

=cut

sub set_flexcache_delegations_state {
    my $class = shift;
    my %args  = validate(
        @_,
        {   read_delegations =>
                {type => SCALAR, default => FCACHE_STATE_ENABLED,},
            write_delegations =>
                {type => SCALAR, default => FCACHE_STATE_ENABLED,},
        }
    );

    my $read_delegations  = $args{read_delegations};
    my $write_delegations = $args{write_delegations};

    # TODO: setflag flexcache_delegations 0 (for both read-write at once)

    fcache_log_result(
        result_type => FCACHE_LOGRESULT_SCRIPT,
        message     => "This method is not yet available for 7-mode"
    );

    return FCACHE_RC_SUCCESS;
} ## end sub set_flexcache_delegations_state

########################################################
# Internal methods
########################################################

# Any method that is not specifically implemented in this module
# will terminate the script execution with an N/A result.
# We do this in the idea that the script called it
# with a very good reason (to do something with the vserver/vfiler)
sub AUTOLOAD {
    our $AUTOLOAD;
    my $class  = shift;
    my $method = $AUTOLOAD;

    fcache_log_result(
        result_type => FCACHE_LOGRESULT_NA,
        message     => "This specific method is not available for"
            . " the current configuration: $method\n"
            . "The reason for this may be:\n"
            . "- The required action is only available for C-mode"
            . " (and this is a 7-mode filer)\n"
            . "- The required action is available for 7-mode,"
            . " but only in a multistore setting\n"
            . "- The VSERVER* command-line arguments are not set correctly\n"
            . "  (please see fcache_common_params.pl documentation)"
    );
} ## end sub AUTOLOAD

# Perform NFS setup for this vserver:
# - create/reuse a data interface,
#   using the information obtained from the tharnhost file
# - create/reuse a route for this data interface
# - create/reuse an nfs configuration for the vserver
# - create/reuse a default export-policy rule for the vserver
# Arguments: none.
# Return: FCACHE_RC_SUCCESS.
sub _setup_nfs {
    my $class = shift;
    my %args = validate(@_, {});

    my $setup_header = 'Vserver 7Mode NFS setup for ' . $class->{_name};

    fcache_log_comment(message => "$setup_header: Setting up NFS")
        if ($class->{_debug});

    my $default_options;

    # TODO: need to keep track of these in a better way -
    # revisit after "nfs modify"
    if ($class->{_config_type} eq 'cache') {
        $default_options = {
            'licensed_feature.flexcache_nfs.enable' => {
                value                => 'on',
                allow_no_entry_error => TRUE,
            },
            'flexcache.access' => {value => 'all',},
            'flexcache.enable' => {value => 'on',},
        };
    } else {
        $default_options = {
            'licensed_feature.flexcache_nfs.enable' => {
                value                => 'on',
                allow_no_entry_error => TRUE,
            },
            'flexcache.access' => {value => 'all',},
            'flexcache.enable' => {value => 'on',},
        };
    }

    foreach my $option (keys(%{$default_options})) {

        # TODO: for this specific options,
        # it may be better to not clean up in the destructor
        if ($class->{_is_multistore}) {
            $class->_set_vfiler_option(
                name => $option,
                %{$default_options->{$option}}
            );
        } else {
            $class->_check_and_set_option_value(
                name => $option,
                %{$default_options->{$option}}
            );
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub _setup_nfs

# Set the value of a specific option on this filer.
# This method is for internal use only.
# It internally keeps track of any modified option value,
# to restore it at cleanup time
# Arguments:
# - name: Required: Scalar: Option name to look for.
# - value: Required: Scalar: New option value.
# Return: FCACHE_RC_SUCCESS.
sub _set_option_value {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name  => {type => SCALAR,},
            value => {type => SCALAR,},
        }
    );

    my $name  = $args{name};
    my $value = $args{value};

    NACL::C::Options->option(
        command_interface => $class->{_nacl_node},
        'option-name'     => $name,
        'option-value'    => $value,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub _set_option_value

# Set the value of a specific option on this filer,
# only if the value is not already set accordingly.
# Throws an error if the option is not found.
# Arguments:
# - name: Required: Scalar: Option name to look for.
# - value: Required: Scalar: Value to set.
# - allow_no_entry_error: Optional: Boolean: Flag to overwrite the default
#   behavior in case the option is not found. If this flag is set to true, then
#   an "no entry error" will only cause the method to return the empty string.
# Return: FCACHE_RC_SUCCESS.
sub _check_and_set_option_value {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name                 => {type => SCALAR,},
            value                => {type => SCALAR,},
            allow_no_entry_error => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $name              = $args{name};
    my $value             = $args{value};
    my $need_to_set_value = TRUE;

    # In case the _cleanup_options hash does not already contain
    # an entry for this option,
    # we need to obtain the current value of the option.
    # We remember the old value only the first time we ever change it
    # This is the value we will need to restore to at cleanup time
    if (!exists($class->{_cleanup_options}->{$name})) {
        my $old_value = '';

        try {
            my $option_state = NACL::CS::Options->fetch(
                command_interface => $class->{_nacl_node},
                filter            => {name => $name},
            );
            if (defined($option_state)) {
                $old_value = $option_state->value();
            }
        }

        catch NATE::BaseException with {
            my $exception = shift;
            my $output    = $exception->text();

            # This is a NACL-hard-coded error message
            if (   ($output =~ /No matching options.+ found/)
                && ($args{allow_no_entry_error})) {

                # Do nothing, old_value stays ''
            } else {
                $exception->throw();
            }
        };

        if ($old_value eq '') {
            fcache_log_comment(
                message => "Option is not supported in this release:\n"
                    . "Name: $name")
                if ($class->{_debug});
            $need_to_set_value = FALSE;
        } elsif ($old_value eq $value) {
            fcache_log_comment(
                      message => "Option already set to the required value:\n"
                    . "Name: $name\n"
                    . "Value: $old_value")
                if ($class->{_debug});
            $need_to_set_value = FALSE;
        } else {
            $class->{_cleanup_options}->{$name} = $old_value;
        }
    }

    if ($need_to_set_value) {
        $class->_set_option_value(
            name  => $name,
            value => $value
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub _check_and_set_option_value

# Uses "vfiler run" to set a specific option on the vfiler.
# It does not check if the option is already set to the desired value.
# Arguments:
# - name: Required: Scalar: Option name to look for.
# - value: Required: Scalar: Value to set.
# - allow_no_entry_error: Optional: Boolean: This flag is ignored.
#    Allowed only for compatibility with _check_and_set_option_value.
# Return: FCACHE_RC_SUCCESS.
sub _set_vfiler_option {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name                 => {type => SCALAR,},
            value                => {type => SCALAR,},
            allow_no_entry_error => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $name  = $args{name};
    my $value = $args{value};

    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
    $api_set->vfiler_run(
        vfiler  => $class->{_name},
        command => "options $name $value",
    );

    return FCACHE_RC_SUCCESS;
} ## end sub _set_vfiler_option

# Not supported.
# Terminates script execution with N/A
# Arguments: none.
sub _setup_cifs {
    my $class = shift;

    fcache_log_result(
        result_type => FCACHE_LOGRESULT_NA,
        message     => "FlexCache does not support CIFS in 7-mode"
    );
} ## end sub _setup_cifs

# Create or recycle a vfiler.
# Class constructor utility method.
# Creates or recycles a vfiler, depending on the _recycle flag value.
# Assumptions: the vfiler's root volume is already created.
# Arguments: none.
# Return: FCACHE_RC_SUCCESS.
sub _create_or_recycle_vfiler {
    my $class = shift;
    my %args = validate(@_, {});

    my $root_vol_state = $class->{__vfiler_root_vol}->state();
    my $root_vol_path  = $root_vol_state->junction_path();
    my $name           = $class->{_name};
    my $need_to_create = TRUE;
    my $output         = '';
    my $exception;

    try {
        $class->{__nacl_vfiler} = NACL::C::Vfiler->find(
            command_interface => $class->{_nacl_node},
            filter            => {vfiler => $name,}
        );
    }
    catch NATE::BaseException with {
        $exception = shift;
        $output    = $exception->text();
    };

    my $rc = $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VFILER_FIND,
        output      => $output,
        expected_rc => FCACHE_RC_FAILURE_NON_FATAL,
    );

    if ($rc == FCACHE_RC_SUCCESS) {
        if ($class->{_recycle}) {
            fcache_log_comment(message => "SETUP: Reusing the vfiler: $name");
            $need_to_create = FALSE;

            # Check that the vfiler is using the right IP
            my $ip_config_ok      = FALSE;
            my $state             = $class->{__nacl_vfiler}->state();
            my $ip_configurations = $state->ip_address_configuration();
            foreach my $configuration (@{$ip_configurations}) {
                if (   ($configuration->{interface} eq $class->{__dlif_port})
                    && ($configuration->{ip_address} eq $class->{_dlif_ip})) {
                    $ip_config_ok = TRUE;
                    last;
                }
            }
            if (!($ip_config_ok)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_CONFIG,
                    message     => "Reusing the vfiler $name\n"
                        . "Some of the requirements for this vfiler are not met. "
                        . "Please address them before running the script:\n"
                        . "Expected IP address for interface "
                        . $class->{__dlif_port} . ": "
                        . $class->{_dlif_ip}
                );
            }
        } else {
            fcache_log_comment(message =>
                    "SETUP: Attempting to stop the existent vfiler: $name");

           # Do this as a best-effort try, do not deal with any error here
           # (e.g. vfiler already stopped)
           # If stop fails for other reasons, then the destroy will fail later
            try { $class->{__nacl_vfiler}->stop(); }
            catch NATE::BaseException with {};
            fcache_log_comment(
                message => "SETUP: Destroying the existent vfiler: $name");
            $class->{__nacl_vfiler}->destroy();
        }
    } elsif ($rc != FCACHE_RC_NO_ENTRY_FOUND) {
        $exception->throw();
    }

    if ($need_to_create) {
        fcache_log_comment(message => "SETUP: Creating the vfiler: $name");
        $class->{__nacl_vfiler} = NACL::C::Vfiler->create(
            command_interface => $class->{_nacl_node},
            vfiler            => $name,
            'storage-units'   => [$root_vol_path,],
            'ip-addresses'    => [$class->{_dlif_ip},],
            'bypass-setup'    => TRUE
        );
        fcache_log_comment(
            message => "SETUP: Configuring vfiler's IP: $name");
        NACL::C::Ifconfig->ifconfig(
            command_interface => $class->{_nacl_node},
            interface         => $class->{__dlif_port},
            alias             => $class->{_dlif_ip},
            netmask           => $class->{__dlif_mask},
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub _create_or_recycle_vfiler

# Class destructor
# Unless the constructor was called with perform_cleanup set to FALSE,
# it takes care of any cleanup that needs to be done on the vfiler:
# - Resets any filer option that was set by this object.
# - Besides this, if in a multistore environment:
#   the vfiler is destroyed (unless recycling was requested).
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if (($class->{_creator_pid} == $$) && ($class->{_perform_cleanup})) {
        fcache_log_comment(
            message => "CLEANUP: Starting 7Mode Vserver cleanup for "
                . $class->{_name})
            if $class->{_debug};

        foreach my $option_name (keys(%{$class->{_cleanup_options}})) {
            $class->_set_option_value(
                name  => $option_name,
                value => $class->{_cleanup_options}->{$option_name},
            );
        }

        if ($class->{_is_multistore}) {
            if ($class->{_recycle}) {
                fcache_log_comment(
                    message => "CLEANUP: Recycling requested. Leaving behind"
                        . " the vfiler "
                        . $class->{_name});
            } else {
                if (defined($class->{__nacl_vfiler})) {
                    fcache_log_comment(message => "CLEANUP: Purging vfiler "
                            . $class->{_name});
                    try { $class->{__nacl_vfiler}->stop(); }
                    catch NATE::BaseException with {};
                    $class->{__nacl_vfiler}->destroy();
                }
                if (defined($class->{__vfiler_root_vol})) {
                    fcache_log_comment(
                        message => "CLEANUP: Purging vfiler root volume of "
                            . $class->{_name});
                    $class->{__vfiler_root_vol}
                        ->purge(nacltask_if_purged => 'pass');
                }
            }
        }
    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

