#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Aggregate.pm#3 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Mode and release agnostic module for aggregate-level functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Aggregate

=head1 DESCRIPTION

This module provides aggregate-level functionality for all the flexcache scripts.
Its functionality is made available through the 'aggregate' member of the FlexCache::Origin object.
The module is release agnostic and provides functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though, that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

=head1 SYNOPSIS

This module is intended to be used only through the FlexCache::Origin objects,
and/or through the FlexCache::Cache objects, as follows:

	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Cache;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
	$origin->{aggregate}->call_any_method_but_the_constructor();
        
        my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
                                          aggregate => {name => $AGGR_DST,
                                                        block_type => $AGGR_DST_BLOCK_TYPE,},
                                          volume => {name => $VOL_DST,},
                                          vserver => {name => $VSERVER,
                                                      dlif_name => $DLIF_DST,
                                                      dlif_id => $DLIF_DST_ID,},
                                          client => {name => $CLIENT_DST,},
                                          origin => $origin);
        $cache->{aggregate}->call_any_method_but_the_constructor();
	
Using this module independently is possible, with the risk that the constructor's arguments may change at any time.

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl

=cut

package FlexCache::Aggregate;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use NATE::BaseException qw(:try);
use NACL::OptionSet;
use NACL::STask::Aggregate;
use NACL::STask::Node;
use NACL::STask::Wreck;
use NACL::MTask::Wafliron;

=head1 CONSTRUCTOR

=head2 new

Constructor for the FlexCache::Aggregate class.
It creates or reuses an aggregate with the required attributes.

=head3 ARGUMENTS

=over

=item B<name> I<Required> Scalar:

        Name of the aggregate.

=item B<diskcount> I<Optional> Scalar:

        Disk count for this aggregate.
        This value is not currently enforced in the case of attach_only and recycle
        (i.e. if the aggregate already exists, the number of disks will not be checked).
        Default value: constant FCACHE_DFLT_AGGR_DISKS (defined in FlexCache constants).

=item B<block_type> I<Optional> Scalar:

        Block type for this aggregate (64-bit or 32-bit).
        This value is enforced in the case of attach_only and recycle
        (i.e. if the aggregate already exists, it will be required to be of the specified block type).
        Legal values: constants FCACHE_AGGR_BLOCK_TYPE_32 and FCACHE_AGGR_BLOCK_TYPE_64 (defined in FlexCache::Constants).
        Default value: constant FCACHE_DFLT_AGGR_BLOCK_TYPE 

=item B<recycle> I<Optional> Boolean:

        Flag to determine if this aggregate should be reused.
        If this flag is set to TRUE, then:
        - at setup the aggregate will be either created or reused (if it already exists)
        - at cleanup it will not be destroyed.
        If the flag is set to FALSE, then:
        - at setup the aggregate will be created or purged+recreated (if it already exists)
        - at cleanup it will be purged.
        Default value: constant FCACHE_DFLT_AGGR_RECYCLE (defined in FlexCache constants).

=back

=head3 RETURN

FlexCache::Aggregate object.

=head3 EXAMPLE

	# Create a flexcache origin with the default settings for the aggregate member:
	my $origin = FlexCache::Origin->new(aggregate => {name => $AGGR_SRC,},
                                            ....);
	# Create a flexcache origin, requiring a 32-bit aggregate filer:
	my $origin = FlexCache::Origin->new(aggregate => {name => $AGGR_SRC,
	                                                 block_type => FCACHE_AGGR_BLOCK_TYPE_32,},
                                            ....);
	# Create a flexcache origin, requiring a non-recycled aggregate:
	my $origin = FlexCache::Origin->new(aggregate => {name => $AGGR_SRC,
	                                                 recycle => FALSE,},
                                            ....);
	# All of the above examples apply to the FlexCache::Cache constructor as well.

=cut

# "Somehow internal" arguments (to limit the use of this constructor to the FlexCache library.
#                               These arguments are not documented with POD, to not confuse the script developers):
# - node: Required: Object: NACL::STask::Node object for the hosting node/filer.
# - ui_messages: Required: Object: FlexCache::UI::OntapMessages, to be used for any ONTAP output validation.
# - debug: Optional: Boolean: Flag to increase the verbosity of this module, useful for debugging purposes.
#          Default value: FALSE.
# - perform_cleanup: Optional: Boolean: Flag to control the cleanup execution.
#                    Default value: TRUE.
# - attach_only: Optional: Boolean: Flag to determine if the constructor should only attach to the required aggregate,
#                and not attempt to create/recreate. If this flag is set to TRUE, then the recycle flag is ignored.
#                Default value: FALSE.
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name => {type => SCALAR,},
            node => {type => OBJECT, isa => "NACL::STask::Node"},
            diskcount => {
                type    => SCALAR,
                regex   => qr/^\d+$/,
                default => FCACHE_DFLT_AGGR_DISKS,
            },
            block_type =>
                {type => SCALAR, default => FCACHE_DFLT_AGGR_BLOCK_TYPE},
            recycle => {type => BOOLEAN | UNDEF, optional => 1,},
            ui_messages =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
            debug           => {type => BOOLEAN, default => FALSE,},
            perform_cleanup => {type => BOOLEAN, default => TRUE,},
            attach_only     => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $self                = {};
    my $required_block_type = $args{block_type};

    $self->{_name}            = $args{name};
    $self->{_nacl_node}       = $args{node};
    $self->{_attach_only}     = $args{attach_only};
    $self->{_debug}           = $args{debug};
    $self->{_perform_cleanup} = $args{perform_cleanup};
    $self->{_ui_messages}     = $args{ui_messages};
    $self->{_creator_pid}     = $$;
    $self->{_recycle}         = $args{recycle};
    if (!defined($self->{_recycle})) {
        my $options = NACL::OptionSet->new();
        $self->{_recycle} = $options->get('FCACHE_AGGR_RECYCLE');
    }
    if (!defined($self->{_recycle})) {
        $self->{_recycle} = FCACHE_DFLT_AGGR_RECYCLE;
    }

    if ($self->{_attach_only}) {

        # Declare the object recycled, so we do not destroy the aggregate
        # at setup/cleanup
        $self->{_recycle} = TRUE;
    }

    fcache_log_comment(message => "SETUP: Creating/reusing the aggregate "
            . $self->{_name} . "\n"
            . "Recycling requested: "
            . $self->{_recycle});

    my $need_to_create = FALSE;
    my $node_name      = $self->{_nacl_node}->name();

    try {
        $self->{_nacl_aggr} = NACL::STask::Aggregate->find(
            command_interface => $self->{_nacl_node},
            filter            => {
                nodes     => [$node_name],
                aggregate => $self->{_name},
            }
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching .* aggregate.* found/) {
            if ($self->{_attach_only}) {
                $exception->throw();
            }
            $need_to_create = TRUE;
        } else {
            $exception->throw();
        }
    };

    if (($need_to_create == FALSE) && (!($self->{_recycle}))) {
        $self->{_nacl_aggr}->purge();
        $need_to_create = TRUE;
    }

    if ($need_to_create == TRUE) {
        $self->{_nacl_aggr} = NACL::STask::Aggregate->create(
            command_interface => $self->{_nacl_node},
            aggregate         => $self->{_name},
            diskcount         => $args{diskcount},
            'block-type'      => $required_block_type,
        );
    }

    my $aggr_state         = $self->{_nacl_aggr}->state();
    my $current_block_type = $aggr_state->block_type();
    $self->{_block_type} = $current_block_type;
    if ($current_block_type ne $required_block_type) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_CONFIG,
            message     => 'Reusing aggregate '
                . $self->{_name} . "\n"
                . "Aggregate's current block type is $current_block_type.\n"
                . "Script requires block type: $required_block_type"
        );
    }

    bless($self, $class);

    fcache_log_comment(message =>
            "SETUP: Completed creation of new Aggregate object for aggregate "
            . $self->{_name})
        if $self->{_debug};
    return $self;

} ## end sub new

=head1 OTHER METHODS

=head2 get_name

Method to obtain the name of this aggregate;

=head3 ARGUMENTS

None.

=head3 RETURN

Name of the aggregate (Scalar).

EXAMPLE

	my $aggr_name = $obj->get_name();

=cut

sub get_name {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_name};
} ## end sub get_name

=head2 get_block_type

Method to obtain the block type of this aggregate;

=head3 ARGUMENTS

None.

=head3 RETURN

Block type of the aggregate (Scalar).
It will be one of the FCACHE_AGGR_BLOCK_TYPE_* constants.

EXAMPLE

	my $block_type = $obj->get_block_type();

=cut

sub get_block_type {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_block_type};
} ## end sub get_block_type

=head2 get_size

Obtain the size of this aggregate.

=head3 ARGUMENTS

None.

=head3 RETURN

Size in bytes of the aggregate (Scalar).

EXAMPLE

	my $size = $obj->get_size();

=cut

sub get_size {
    my $class = shift;
    my %args = validate(@_, {});

    my $aggr_state = $class->{_nacl_aggr}->state();
    my $size       = $aggr_state->size();
    if (!defined($size)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Unable to retrieve aggregate size"
        );
    }

    fcache_log_comment(
        message => "Aggregate: " . $class->{_name} . "\n" . "Size: $size");

    return $size;
} ## end sub get_size

=head2 get_available_size

Obtain the size of the available space of this aggregate.

=head3 ARGUMENTS

None.

=head3 RETURN

Size in bytes of the available space (Scalar).

EXAMPLE

	my $avail = $obj->get_available_size();

=cut

sub get_available_size {
    my $class = shift;
    my %args = validate(@_, {});

    my $aggr_state = $class->{_nacl_aggr}->state();
    my $size       = $aggr_state->availsize();
    if (!defined($size)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Unable to retrieve aggregate available size"
        );
    }

    fcache_log_comment(message => "Aggregate: "
            . $class->{_name} . "\n"
            . "Available space size: $size");

    return $size;
} ## end sub get_available_size

=head2 wreck_run

Inject corruption to this aggregate or to one of the volumes
by running wreck. Please note that a call to wreck_run should be followed
(at some point) in the script by a call to wreck_cleanup.

=head3 ARGUMENTS

=over

=item B<inject_commands> I<Required> Arrayref:

        Arraref of wreck commands to be run. Its elements are hash references,
	representing the command names and all the pertinent options.
	E.g. [{command=>'block_pattern',
	       inum => 'priv:86',
	       bno => 0,
	       start => 0,
	       end => 255,
	       fill => 0,}],

=item B<volumes> I<Optional> Arrayref:

	In case the wreck command should be run on specific volume(s),
	this argument should be used to specify the volume name(s).
	If this argument is not present,then the wreck command is run for the
	whole aggregate.

=item B<modify_flags> I<Optional> Arrayref:

	Flags to modify after running the wreck command.
          
=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->wreck_run(volumes => [$origin_name],
	                inject_commands => [{command=>'block_pattern',
			                     inum => 'priv:86',
					     bno => 0,
					     start => 0,
					     end => 255,
					     fill => 0,}],
                        modify_flags => {wafl_iron_panic_on_corruption => 0,},);
	# run wafliron and/or other steps
	$obj->wreck_cleanup();

=cut

sub wreck_run {
    my $class = shift;
    my %args  = validate(
        @_,
        {   volumes         => {type => ARRAYREF, optional => 1,},
            inject_commands => {type => ARRAYREF,},
            modify_flags    => {type => HASHREF,  optional => 1,},
        }
    );

    if (defined($class->{_wreck_nacl_obj})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "wreck_run should not be called the second time "
                . "without calling wreck_cleanup in between calls",
        );
    }

    my %nacl_wreck_args = (
        command_interface        => $class->{_nacl_node},
        nacltask_inject_commands => $args{inject_commands},
    );
    if (exists($args{volumes})) {
        $nacl_wreck_args{nacltask_volumes} = $args{volumes};
    } else {
        $nacl_wreck_args{nacltask_aggregates} = [$class->{_name}];
    }
    if (exists($args{modify_flags})) {
        $nacl_wreck_args{nacltask_modify_flags} = $args{modify_flags};
    }

    $class->{_wreck_nacl_obj} = NACL::STask::Wreck->run(%nacl_wreck_args);

    return FCACHE_RC_SUCCESS;
} ## end sub wreck_run

=head2 wreck_cleanup

Perform cleanup after running a wreck command: restore any flags back to their
original value and remove the files that were created for the wreck command.

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->wreck_cleanup();

=cut

sub wreck_cleanup {
    my $class = shift;
    my %args = validate(@_, {});

    if (!defined($class->{_wreck_nacl_obj})) {
        return FCACHE_RC_SUCCESS;
    }

    $class->{_wreck_nacl_obj}->wreck_cleanup();
    delete($class->{_wreck_nacl_obj});

    return FCACHE_RC_SUCCESS;
} ## end sub wreck_cleanup

=head2 wafliron_run

Run wafliron on this aggregate.

=head3 ARGUMENTS

=over

=item B<expected_corruptions> I<Optional> Arrayref:

        Arraref of "expected corruption codes", for which a correspondent
	error message is expected to be seen in the wafliron output.
	If this argument is not present, then wafliron is expected to complete
	without any corruption being detected. In the case when wafliron is
	expected to find and fix one or more corruptions, then this argument
	should be set. Its elements can be any of the FCACHE_CORRUPTION_*
	constants (defined in FlexCache::Constants).
          
=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	# No corruption expected
	$obj->wafliron_run();
	# Corruption expected
	$obj->wafliron_run(expected_corruptions => [FCACHE_CORRUPTION_QTREE_METADATA,]);

=cut

sub wafliron_run {
    my $class = shift;
    my %args  = validate(@_,
        {expected_corruptions => {type => ARRAYREF, optional => 1,},});

    my $expected_corruptions = $args{expected_corruptions};

    my $wi = NACL::MTask::Wafliron->new(
        command_interface => $class->{_nacl_node},
        aggregate         => $class->{_name},
    );
    $wi->start();
    my $corruption_found = $wi->corruption_found();
    if ($corruption_found) {
        my $wi_messages = $wi->get_wafliron_messages();
        fcache_log_comment(
            message => "\n\n\nWafliron messages:\n\n$wi_messages");
        if (!defined($expected_corruptions)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Wafliron detected unexpected corruption\n"
                    . "\n\n$wi_messages"
            );
        }
        $class->{_ui_messages}->verify_wafliron_detected_corruption(
            output               => $wi_messages,
            expected_corruptions => $expected_corruptions
        );
    } else {
        if (defined($expected_corruptions)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Wafliron did not detect any corruption\n"
                    . "Expected corruptions:\n"
                    . Dumper($expected_corruptions)
            );
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub wafliron_run

########################################################
# Internal methods
########################################################

# Class destructor
# Unless the constructor was called with perform_cleanup set to FALSE,
# it takes care of any cleanup that needs to be done on the aggregate:
# - purges the aggregate (unless recycling was requested)
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if (($class->{_creator_pid} == $$) && ($class->{_perform_cleanup})) {
        $class->wreck_cleanup();
        if ($class->{_recycle}) {
            fcache_log_comment(message =>
                    "CLEANUP: Recycling requested. Leaving behind the aggregate "
                    . $class->{_name});
        } else {
            fcache_log_comment(message => "CLEANUP: Purging the aggregate "
                    . $class->{_name});
            $class->{_nacl_aggr}->purge();
        }
    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

