#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/CMode/Vserver.pm#2 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Vserver-level functionality for CMode.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::CMode::Vserver

=head1 DESCRIPTION

This module provides release agnostic vserver-level functionality for CMode.
This module should only be used internally by the FlexCache library.

=head1 SYNOPSIS

N/A.

=cut

package FlexCache::CMode::Vserver;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use NACL::OptionSet;
use NATE::BaseException qw(:try);
use NACL::STask::Node;
use NACL::STask::Vserver;
use NACL::STask::NetworkInterface;
use NACL::STask::NetworkRoutingGroupsRoute;
use NACL::STask::VserverNFS;
use NACL::C::ExportPolicy;
use NACL::C::VserverExportPolicyRule;
use NACL::C::UnixUser;
use NACL::C::UnixGroup;
use NACL::C::VserverServicesDns;
use NACL::C::VserverNameMapping;
use NACL::C::VserverCifs;
use NACL::C::AntivirusOnAccessPolicy;
use NACL::C::VolumeFlexcacheCachePolicy;

# Constructor for the FlexCache::CMode::Vserver class.
# It is only intended to be used within the FlexCache library.
# The constructor will obtain the dlif information from the tharnhost file,
# to be used for NFS/CIFS setup.
# It will create/reuse the vserver and
# verify some minimal requirements in case of reuse.
# Unless the attach_only parameter is true,
# it will also execute the NFS/CIFS setup, depending on the client_type.
# Arguments: They should always be kept identical (+- optionality)
#            to the allowed arguments for the constructor of
#            the FlexCache::Vserver class,
#            even if some of them may be ignored for CMode.
#            These arguments should be documented in the FlexCache::Vserver
#            module (externally-available)
# Return: FlexCache::CMode::Vserver object.
#
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name           => {type => SCALAR | UNDEF,},
            node           => {type => OBJECT, isa => "NACL::STask::Node",},
            aggregate_name => {type => SCALAR,},
            dlif_name      => {type => SCALAR | UNDEF, optional => 1,},
            dlif_id        => {type => SCALAR | UNDEF, optional => 1,},
            client_type    => {type => SCALAR,},
            recycle =>
                {type => BOOLEAN, default => FCACHE_DFLT_VSERVER_RECYCLE,},
            perform_cleanup => {type => BOOLEAN,        default  => TRUE,},
            debug           => {type => BOOLEAN,        default  => FALSE,},
            config_type     => {type => SCALAR | UNDEF, optional => 1,}
            ,    # Ignored for C-mode
            ui_messages =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
            attach_only => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $self           = {};
    my $aggregate_name = $args{aggregate_name};

    $self->{_name} = (defined($args{name})) ? $args{name} : FCACHE_VSERVER;
    $self->{_debug} = $args{debug};
    $self->{_nacl_node}               = $args{node};
    $self->{_node_name}               = $self->{_nacl_node}->name();
    $self->{_recycle}                 = $args{recycle};
    $self->{_attach_only}             = $args{attach_only};
    $self->{_perform_cleanup}         = $args{perform_cleanup};
    $self->{_cleanup_av_policies}     = {};
    $self->{_cleanup_export_policies} = {};
    $self->{_cleanup_default_flexcache_cache_policy} = {};
    $self->{_cleanup_flexcache_cache_policy}         = {};
    $self->{_cleanup_dlif_status_admin}              = {};
    $self->{_ui_messages}                            = $args{ui_messages};
    $self->{_creator_pid}                            = $$;
    $self->{_dlifs}                                  = [];
    fcache_log_comment(
        message => "Creating new CMode Vserver object for vserver "
            . $self->{_name})
        if $self->{_debug};

    # Create/reuse the vserver
    if ($self->{_attach_only}) {
        fcache_log_comment(
            message => "Attaching to vserver " . $self->{_name})
            if $self->{_debug};
        $self->{_nacl_vserver} = NACL::STask::Vserver->find(
            command_interface => $self->{_nacl_node},
            filter            => {vserver => $self->{_name},}
        );

        my $data_interface = NACL::STask::NetworkInterface->find(
            command_interface => $self->{_nacl_node},
            filter            => {
                vserver           => $self->{_name},
                role              => FCACHE_DLIF_ROLE,
                'home-node'       => $self->{_node_name},
                'is-home'         => 'true',
                'status-admin'    => 'up',
                'status-oper'     => 'up',
                'failover-policy' => FCACHE_REQUIRED_DLIF_FAILOVER_POLICY,
            }
        );
        my $data_interface_state = $data_interface->state();
        $self->{_dlif_name} = $data_interface_state->lif();
        $self->{_dlif_ip}   = $data_interface_state->address();

        # Declare the object recycled,
        # so we do not destroy the vserver at cleanup
        $self->{_recycle} = TRUE;
    } else {
        fcache_log_comment(
            message => "Creating/reusing vserver " . $self->{_name})
            if $self->{_debug};
        my $dlif_tharnhost_id
            = (defined($args{dlif_id}))
            ? $args{dlif_id}
            : FCACHE_DFLT_DLIF_ID;
        if ($dlif_tharnhost_id !~ /^\d+$/) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message => "dlif_id must be provided as a positive integer"
            );
        }

        # Obtain all the needed information from the
        # tharnhost file for this dlif (ip, mask, gateway, port)
        my $hostrec = $self->{_nacl_node}->hostrec();
        $self->{_dlif_ip} = $hostrec->{'DATA_IP' . $dlif_tharnhost_id};

        # Note: the __xxx internal fields should be used with extra care:
        # they may be undefined in some cases
        $self->{__dlif_mask} = $hostrec->{'DATA_MASK' . $dlif_tharnhost_id};
        $self->{__dlif_gateway}
            = $hostrec->{'DATA_GATEWAY' . $dlif_tharnhost_id};
        $self->{__dlif_port} = $hostrec->{'DATA_PORT' . $dlif_tharnhost_id};

        if (   (!defined($self->{_dlif_ip}))
            || (!defined($self->{__dlif_mask}))
            || (!defined($self->{__dlif_gateway}))
            || (!defined($self->{__dlif_port}))) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message => "Error while extracting the dlif information for "
                    . "Data lif with ID $dlif_tharnhost_id\n"
                    . "Please make sure that the provided dlif_id"
                    . " is correct: $dlif_tharnhost_id\n"
                    . "and that the nate-host file contains"
                    . " entries for all of the following:\n"
                    . "DATA_IP$dlif_tharnhost_id\n"
                    . "DATA_MASK$dlif_tharnhost_id\n"
                    . "DATA_GATEWAY$dlif_tharnhost_id\n"
                    . "DATA_PORT$dlif_tharnhost_id\n\n"
                    . "Current hostrec information:\n"
                    . Dumper($hostrec)
            );
        }
        fcache_log_comment(message =>
                "SETUP: Obtained dlif information from the nate-host file:\n"
                . "Data lif IP: "
                . $self->{_dlif_ip} . "\n"
                . "Data lif mask: "
                . $self->{__dlif_mask} . "\n"
                . "Data lif gateway: "
                . $self->{__dlif_gateway} . "\n"
                . "Data lif port: "
                . $self->{__dlif_port})
            if $self->{_debug};

        $self->{_dlif_name} = $args{dlif_name};
        if (!defined($self->{_dlif_name})) {

            # Auto-detect the name of the data lif
            my $data_lif = NACL::STask::NetworkInterface->find(
                command_interface => $self->{_nacl_node},
                filter            => {
                    vserver   => $self->{_name},
                    role      => FCACHE_DLIF_ROLE,
                    'address' => $self->{_dlif_ip},
                }
            );
            my $data_lif_state = $data_lif->state();
            $self->{_dlif_name} = $data_lif_state->lif();
            fcache_log_comment(
                      message => "Data lif name was not provided by user.\n"
                    . "Auto-detected data lif: "
                    . $self->{_dlif_name});
        }

        my $nacltask_if_exists     = ($self->{_recycle}) ? 'reuse' : 'purge';
        my $options                = NACL::OptionSet->new();
        my $nacltask_cifs_username = $options->get('AD_SERVER_USER');
        my $nacltask_cifs_password = $options->get('AD_SERVER_PASSWORD');
        $self->{_nacl_vserver} = NACL::STask::Vserver->create(
            command_interface           => $self->{_nacl_node},
            vserver                     => $self->{_name},
            rootvolume                  => $self->{_name} . '_root',
            aggregate                   => $aggregate_name,
            'ns-switch'                 => [FCACHE_DFLT_VSERVER_NS_SWITCH],
            'nm-switch'                 => [FCACHE_DFLT_VSERVER_NM_SWITCH],
            'rootvolume-security-style' => FCACHE_DFLT_VSERVER_ROOTVOL_SEC,

          # TODO: the following are needed for purging: NACL is failing if set
          #nacltask_cifs_username => $nacltask_cifs_username,
          #nacltask_cifs_password => $nacltask_cifs_password,
            nacltask_if_exists => $nacltask_if_exists,
        );
    }

    bless($self, $class);

    if ($self->{_attach_only}) {
        return $self;
    }

    $self->_setup_common();
    if ($args{client_type} eq FCACHE_CLIENT_TYPE_UNIX) {
        $self->_setup_nfs();
    } elsif ($args{client_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        $self->_setup_cifs();
    } else {

        # This is just in case Protocol::ClientFileOperations
        # changes the type strings
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_CONFIG,
            message => 'Cannot configure the vserver: unknown client type: '
                . $args{client_type} . "\n"
                . 'Supported client types: '
                . FCACHE_CLIENT_TYPE_UNIX . ', '
                . FCACHE_CLIENT_TYPE_WINDOWS
        );
    }

    return $self;

} ## end sub new

=head1 AVAILABLE METHODS

=head2 get_name

Gets the name of this vserver.

=head3 ARGUMENTS

None.

=head3 RETURN

Name of the vserver (Scalar).

EXAMPLE

	my $vserver_name = $obj->get_name();

=cut

sub get_name {
    my $class = shift;
    my %args  = validate(@_,
        {_internal_use => {type => BOOLEAN, default => FALSE,},});

    # For C-mode, the _internal_use is ignored,
    # it is only needed for 7-mode
    return $class->{_name};
} ## end sub get_name

=head2 get_dlif_ip

Gets the IP of the data interface for this vserver.

=head3 ARGUMENTS

None.

=head3 RETURN

IP of the data interface for this vserver (Scalar).

EXAMPLE

	my $mount_ip = $obj->get_dlif_ip();

=cut

sub get_dlif_ip {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_dlif_ip};
} ## end sub get_dlif_ip

=head2 set_flexcache_delegations_state

Enable/disable delegations for this vserver.

=head3 ARGUMENTS

=over

=item B<read_delegations> I<Optional> Scalar:

	State to set for the read-delegations.
	Legal values: FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED.
	Default value: FCACHE_STATE_ENABLED.

=item B<write_delegations> I<Optional> Scalar:

	State to set for the write-delegations.
	Legal values: FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED.
	Default value: FCACHE_STATE_ENABLED.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->set_flexcache_delegations_state(read_delegations => FCACHE_STATE_DISABLED,
	                                      write_delegations => FCACHE_STATE_DISABLED);
	$obj->set_flexcache_delegations_state(read_delegations => FCACHE_STATE_ENABLED,
	                                      write_delegations => FCACHE_STATE_ENABLED);

=cut

sub set_flexcache_delegations_state {
    my $class = shift;
    my %args  = validate(
        @_,
        {   read_delegations =>
                {type => SCALAR, default => FCACHE_STATE_ENABLED,},
            write_delegations =>
                {type => SCALAR, default => FCACHE_STATE_ENABLED,},
        }
    );

    my $read_delegations  = $args{read_delegations};
    my $write_delegations = $args{write_delegations};

    NACL::STask::VserverNFS->modify(
        command_interface       => $class->{_nacl_node},
        vserver                 => $class->{_name},
        'v4.0-read-delegation'  => $read_delegations,
        'v4.0-write-delegation' => $write_delegations,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub set_flexcache_delegations_state

=head2 dlif_create

Creates a data interface for this vserver.
Throws an exception if the creation fails (including the case when the dlif already exists).

=head3 ARGUMENTS

=over

=item B<name> I<Required> Scalar:

    Name of the interface to create.

=item B<home_port> I<Required> Scalar:

    Home-port for this dlif.

=item B<address> I<Required> Scalar:

    IP address for this dlif.

=item B<netmask> I<Required> Scalar:

    Netmask for this dlif.

=item B<failover_policy> I<Optional> Scalar:

    Failover policy for this dlif.
    Legal values: FCACHE_DLIF_FAILOVER_POLICY_DISABLED,
    FCACHE_DLIF_FAILOVER_POLICY_NEXTAVAIL,
    FCACHE_DLIF_FAILOVER_POLICY_PRIORITY, FCACHE_REQUIRED_DLIF_FAILOVER_POLICY
    (defined in FlexCache::Constants).
    Default value: FCACHE_REQUIRED_DLIF_FAILOVER_POLICY.

=item B<use_failover_group> I<Optional> Scalar:
    use-failure-group setting for this dlif.
    Legal values: FCACHE_DFLT_DLIF_USE_FAILOVER_GROUP,
                  FCACHE_STATE_ENABLED,
                  FCACHE_STATE_DISABLED (defined in FlexCache::Constants).
    Default value: FCACHE_DFLT_DLIF_USE_FAILOVER_GROUP.

=back

=head3 RETURN

The routing group of the new data interface (Scalar).

EXAMPLE

	# Create with the default settings
	my $dlif_routing_group = $obj->dlif_create(name => 'vs1_lif',
	                                           home_port => 'e1c',
						   address => '1.2.3.4',
						   netmask => '255.255.254.0',);
	# Create custom
	my $dlif_routing_group = $obj->dlif_create(name => 'vs1_lif',
	                                           home_port => 'e1c',
						   address => '1.2.3.4',
						   netmask => '255.255.254.0',
						   failover_policy => FCACHE_DLIF_FAILOVER_POLICY_NEXTAVAIL,);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent dlif.
#           In the case of reusing the dlif, some basic requirements
#           for flexcache testing are still performed,
#           to ensure that the dlif settings are correct for proper testing
#           (especially needed to ensure that the reads/writes are performed
#           through the origin/cache).
#           Default value: FALSE.
sub dlif_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name            => {type => SCALAR,},
            home_port       => {type => SCALAR,},
            address         => {type => SCALAR,},
            netmask         => {type => SCALAR,},
            failover_policy => {
                type     => SCALAR,
                optional => 1,
            },
            use_failover_group => {
                type    => SCALAR,
                default => FCACHE_DFLT_DLIF_USE_FAILOVER_GROUP,
            },
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $dlif_name = $args{name};
    my $nacltask_if_exists = ($args{_reuse}) ? 'reuse' : 'die';
    my $was_created;
    my %nacl_args = (
        command_interface    => $class->{_nacl_node},
        lif                  => $dlif_name,
        vserver              => $class->{_name},
        role                 => FCACHE_DLIF_ROLE,
        'home-node'          => $class->{_node_name},
        'home-port'          => $args{home_port},
        address              => $args{address},
        netmask              => $args{netmask},
        'use-failover-group' => $args{use_failover_group},
        nacltask_if_exists   => $nacltask_if_exists,
        _was_created         => \$was_created,
    );
    if (defined($args{failover_policy})) {
        $nacl_args{failover_policy} = $args{failover_policy};
    }
    my $interface = NACL::STask::NetworkInterface->create(%nacl_args);

    my $interface_state         = $interface->state();
    my $interface_routing_group = $interface_state->routing_group();

    if (!$was_created) {

        # This should only be the case of _reuse
        # NACL is relying on the "already exists" error message
        # In case of reuse, we need to ensure that the already-existent
        # dlif has the attributes that we care about.
        # For correct flexcache testing, we need to ensure at least:
        # dlif "is home" and has the correct home-node,
        # it belongs to our vserver, it has the correct:
        # address, role, status-admin, status-oper.
        my $current_home_node       = $interface_state->home_node();
        my $current_is_home         = $interface_state->is_home();
        my $current_vserver         = $interface_state->vserver();
        my $current_failover_policy = $interface_state->failover_policy();
        my $current_address         = $interface_state->address();
        my $current_role            = $interface_state->role();
        my $current_status_admin    = $interface_state->status_admin();
        my $current_status_oper     = $interface_state->status_oper();

        if (   ($current_home_node ne $class->{_node_name})
            || ($current_vserver ne $class->{_name})
            || ($current_address ne $args{address})
            || ($current_role    ne FCACHE_DLIF_ROLE)
            || ((   defined($args{failover_policy})
                    && ($current_failover_policy ne $args{failover_policy})
                )
            )
            || ($current_status_admin ne 'up')
            || ($current_status_oper  ne 'up')
            ) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_CONFIG,
                message     => "Reusing the data lif $dlif_name\n"
                    . "Some of the requirements for this data lif are not met."
                    . " Please address them before running the script:\n"
                    . "Dlif should be home and the home node should be "
                    . $class->{_node_name} . "\n"
                    . "  Current home node is $current_home_node,"
                    . " is-home: $current_is_home\n"
                    . "Dlif's vserver should be "
                    . $class->{_name} . "\n"
                    . "  Current vserver is $current_vserver\n"
                    . "Dlif's role should be "
                    . FCACHE_DLIF_ROLE . "\n"
                    . "  Current role is $current_role\n"
                    . "Dlif's address should be "
                    . $args{address} . "\n"
                    . "  Current address is $current_address\n"
                    . "Dlif's status-admin and status-oper should be up\n"
                    . "  Current status-admin is $current_status_admin,"
                    . " status-oper is $current_status_oper"
            );
        }

        if ($current_is_home ne 'true') {
            fcache_log_comment(message => "Reusing the data lif $dlif_name\n"
                    . "Bringing the data lif back home");
            my $home_node = $interface_state->home_node();
            my $home_port = $interface_state->home_port();
            $interface->migrate(
                'dest-node' => $home_node,
                'dest-port' => $home_port,
            );
        }
    }

    push(@{$class->{_dlifs}}, $dlif_name);

    return $interface_routing_group;
} ## end sub dlif_create

=head2 dlif_route_create

Creates a route for a network interface for this vserver.
Throws an exception if the creation fails
(including the case when the route already exists).

=head3 ARGUMENTS

=over

=item B<routing_group> I<Required> Scalar:

    Routing group for which to create the route.

=item B<gateway> I<Required> Scalar:

    Gateway for this new route.

=item B<destination> I<Optional> Scalar:

    Destination for this new route.
    Default value: constant FCACHE_DLIF_IP_MATCH_ALL
    (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	# Create with the default settings
	$obj->dlif_route_create(routing_group => 'd10.53.0.0/20',
	                        gateway => '10.53.0.1',);
	# Create custom
	$obj->dlif_route_create(routing_group => $my_routing_group,
	                        gateway => $my_gateway,
				destination => $my_custom_destination,);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           Default value: FALSE.
sub dlif_route_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   routing_group => {type => SCALAR,},
            gateway       => {type => SCALAR,},
            destination =>
                {type => SCALAR, default => FCACHE_DLIF_IP_MATCH_ALL,},
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $nacltask_if_exists = ($args{_reuse}) ? 'reuse' : 'die';
    NACL::STask::NetworkRoutingGroupsRoute->create(
        command_interface  => $class->{_nacl_node},
        vserver            => $class->{_name},
        'routing-group'    => $args{routing_group},
        gateway            => $args{gateway},
        destination        => $args{destination},
        nacltask_if_exists => $nacltask_if_exists,
    );

    # No extra checks in case of reuse are needed here. NACL does it for us.
    return FCACHE_RC_SUCCESS;
} ## end sub dlif_route_create

=head2 dlif_revert

Revert a dlif to its home port.

=head3 ARGUMENTS

=over

=item B<dlif_name> I<Required> Scalar:

	Name of the dlif to revert.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->dlif_revert(dlif_name => 'my_dlif');

=cut

sub dlif_revert {
    my $class = shift;
    my %args = validate(@_, {dlif_name => {type => SCALAR,},});

    NACL::STask::NetworkInterface->revert(
        command_interface => $class->{_nacl_node},
        lif               => $args{dlif_name},
        vserver           => $class->{_name},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub dlif_revert

=head2 dlif_migrate

Migrate a dlif.

=head3 ARGUMENTS

=over

=item B<dlif_name> I<Required> Scalar:

	Name of the dlif to migrate.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->dlif_migrate(dlif_name => 'my_dlif',
	                   destination_node => 'my_node');

=cut

sub dlif_migrate {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dlif_name        => {type => SCALAR,},
            destination_node => {type => SCALAR,},
        }
    );

    NACL::STask::NetworkInterface->migrate(
        command_interface => $class->{_nacl_node},
        lif               => $args{dlif_name},
        vserver           => $class->{_name},
        'dest-node'       => $args{destination_node},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub dlif_migrate

=head2 dlif_status_modify

Modify a dlif's admin status (up or down).

=head3 ARGUMENTS

=over

=item B<dlif_name> I<Required> Scalar:

	Name of the dlif to modify.
	
=item B<status> I<Required> Scalar:

	New status-admin. Can be one of the FCACHE_DLIF_STATUS_* constants.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->dlif_status_modify(dlif_name => 'my_dlif',
	                         status => FCACHE_DLIF_STATUS_UP);

=cut

sub dlif_status_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dlif_name => {type => SCALAR,},
            status    => {type => SCALAR,},
        }
    );

    my $lif_name = $args{dlif_name};
    my $status   = $args{status};

    my $lif = NACL::STask::NetworkInterface->find(
        command_interface => $class->{_nacl_node},
        filter            => {
            lif     => $lif_name,
            vserver => $class->{_name},
        },
    );
    my $lif_state      = $lif->state();
    my $current_status = $lif_state->status_admin();
    if ($current_status eq $status) {
        fcache_log_comment(
            message => "Dlif $lif_name already has status-admin $status");
        return FCACHE_RC_SUCCESS;
    }

    NACL::STask::NetworkInterface->modify(
        command_interface => $class->{_nacl_node},
        lif               => $lif_name,
        vserver           => $class->{_name},
        'status-admin'    => $status,
    );

    if (!exists($class->{_cleanup_dlif_status_admin}->{$lif_name})) {
        $class->{_cleanup_dlif_status_admin}->{$lif_name} = $current_status;
    }

    return FCACHE_RC_SUCCESS;
} ## end sub dlif_status_modify

=head2 dlifs_migrate_back_home

Migrate the date lifs back home.

=head3 ARGUMENTS

=over

=item B<dlifs> I<Optional> Arrayref:

	Arrayref of dlif names to be migrated back home.
	If this argument is not passed in, then all the dlifs that were
	created by this object will be processed.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->dlifs_migrate_back_home();

=cut

sub dlifs_migrate_back_home {
    my $class = shift;
    my %args  = validate(@_,
        {dlifs => {type => ARRAYREF, default => $class->{_dlifs},},});

    my $dlifs = $args{dlifs};
    fcache_log_comment(message => "Migrating dlifs home:\n" . Dumper($dlifs))
        if ($class->{_debug});

    foreach my $dlif_name (@{$dlifs}) {
        my $interface = NACL::STask::NetworkInterface->find(
            command_interface => $class->{_nacl_node},
            filter            => {
                lif     => $dlif_name,
                vserver => $class->{_name},
            },
        );
        my $interface_state = $interface->state();
        my $home_node       = $interface_state->home_node();
        my $current_node    = $interface_state->curr_node();

        if ($home_node ne $current_node) {
            fcache_log_comment(message => "Migrating dlif back home\n"
                    . "Dlif name: $dlif_name\n"
                    . "Home node: $home_node\n"
                    . "Current node: $current_node");
            $class->dlif_migrate(
                dlif_name        => $dlif_name,
                destination_node => $home_node,
            );
        } else {
            fcache_log_comment(message => "Dlif is already home:\n"
                    . "Dlif name: $dlif_name\n"
                    . "Home node: $home_node");
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub dlifs_migrate_back_home

=head2 nfs_create

Creates an NFS configuration for this vserver.
Throws an exception if the creation fails
(including the case when the configuration already exists).

=head3 ARGUMENTS

=over

=item B<v3> I<Optional> Scalar:

    Required state for NFS version 3 (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED,
                  FCACHE_STATE_DISABLED
                  (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_0> I<Optional> Scalar:

    Required state for NFS version 4.0 (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED,
    FCACHE_STATE_DISABLED (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_0_acl> I<Optional> Scalar:

    Required state for NFS4.0-acl (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED,
    FCACHE_STATE_DISABLED (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_0_read_delegation> I<Optional> Scalar:

    Required state for NFS4.0-read-delegation (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED
    (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_0_write_delegation> I<Optional> Scalar:

    Required state for NFS4.0-write-delegation (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED
    (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_1> I<Optional> Scalar:

    Required state for NFS version 4.1 (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED,
    FCACHE_STATE_DISABLED (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_1_pnfs> I<Optional> Scalar:

    Required state for NFS4.1-parallel-nfs (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED,
    FCACHE_STATE_DISABLED (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<v4_1_acl> I<Optional> Scalar:

    Required state for NFS4.1-acl (enabled/disabled).
    Legal values: FCACHE_STATE_ENABLED,
    FCACHE_STATE_DISABLED (defined in FlexCache::Constants).
    Default value: FCACHE_STATE_ENABLED.

=item B<default_win_user> I<Optional> Scalar:

    Name of the default windows user.
    Default value: constant FCACHE_DFLT_WIN_USER
    (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	# Create with the default settings
	$obj->nfs_create();
	# Create custom
	$obj->nfs_create(v4_0 => FCACHE_STATE_DISABLED);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           Default value: FALSE.
sub nfs_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   v3       => {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_0     => {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_0_acl => {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_0_read_delegation =>
                {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_0_write_delegation =>
                {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_1      => {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_1_pnfs => {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            v4_1_acl  => {type => BOOLEAN, default => FCACHE_STATE_ENABLED,},
            default_win_user =>
                {type => SCALAR, default => FCACHE_DFLT_WIN_USER,},
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $nacltask_if_exists = ($args{_reuse}) ? 'reuse' : 'die';
    my $was_created;

    my $nfs = NACL::STask::VserverNFS->create(
        command_interface       => $class->{_nacl_node},
        vserver                 => $class->{_name},
        v3                      => $args{v3},
        'v4.0'                  => $args{v4_0},
        'v4.0-acl'              => $args{v4_0_acl},
        'v4.0-read-delegation'  => $args{v4_0_read_delegation},
        'v4.0-write-delegation' => $args{v4_0_write_delegation},
        'v4.1'                  => $args{v4_1},
        'v4.1-pnfs'             => $args{v4_1_pnfs},
        'v4.1-acl'              => $args{v4_1_acl},
        'default-win-user'      => $args{default_win_user},
        if_exists               => $nacltask_if_exists,
        _was_created            => \$was_created,
    );

    if (!$was_created) {

        # This should only be the case of _reuse
        # NACL is not taking care of ensuring the asked-for
        # settings are the same with the ones of the reused entry
        my $state             = $nfs->state();
        my $current_v3        = $state->v3();
        my $current_v4_0      = $state->v4_0();
        my $current_v4_1      = $state->v4_1();
        my $current_v4_0_acl  = $state->v4_0_acl();
        my $current_v4_0_rdel = $state->v4_0_read_delegation();
        my $current_v4_0_wdel = $state->v4_0_write_delegation();
        my $current_v4_1_pnfs = $state->v4_1_pnfs();
        my $current_v4_1_acl  = $state->v4_1_acl();

        if (   ($current_v3 ne $args{v3})
            || ($current_v4_0      ne $args{v4_0})
            || ($current_v4_1      ne $args{v4_1})
            || ($current_v4_0_acl  ne $args{v4_0_acl})
            || ($current_v4_0_rdel ne $args{v4_0_read_delegation})
            || ($current_v4_0_wdel ne $args{v4_0_write_delegation})
            || ($current_v4_1_pnfs ne $args{v4_1_pnfs})
            || ($current_v4_1_acl  ne $args{v4_1_acl})) {
            fcache_log_comment(
                      message => 'Reusing the nfs configuration for vserver '
                    . $class->{_name} . "\n"
                    . "Some of the requirements for this nfs configuration"
                    . " are not met.\n"
                    . "Addressing these requirements...");
            $nfs->modify(
                v3                      => $args{v3},
                'v4.0'                  => $args{v4_0},
                'v4.0-acl'              => $args{v4_0_acl},
                'v4.0-read-delegation'  => $args{v4_0_read_delegation},
                'v4.0-write-delegation' => $args{v4_0_write_delegation},
                'v4.1'                  => $args{v4_1},
                'v4.1-pnfs'             => $args{v4_1_pnfs},
                'v4.1-acl'              => $args{v4_1_acl},
            );
        }
    }
} ## end sub nfs_create

=head2 export_policy_create

Create an export policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policyname> I<Required> Scalar:

	Name of the export-policy to be created.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->export_policy_create();

=cut

sub export_policy_create {
    my $class = shift;
    my %args = validate(@_, {policyname => {type => SCALAR,},});

    my $policy_name = $args{policyname};

    NACL::C::ExportPolicy->create(
        command_interface => $class->{_nacl_node},
        vserver           => $class->{_name},
        policyname        => $policy_name
    );

    $class->{_cleanup_export_policies}->{$policy_name} = 1;

    return FCACHE_RC_SUCCESS;
} ## end sub export_policy_create

=head2 export_policy_delete

Delete an export policy of this vserver.

=head3 ARGUMENTS

=over

=item B<policyname> I<Required> Scalar:

	Name of the export-policy to be created.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->export_policy_delete(policyname => 'my_policy');

=cut

sub export_policy_delete {
    my $class = shift;
    my %args = validate(@_, {policyname => {type => SCALAR,},});

    my $policy_name = $args{policyname};

    NACL::C::ExportPolicy->delete(
        command_interface => $class->{_nacl_node},
        vserver           => $class->{_name},
        policyname        => $policy_name
    );

    delete($class->{_cleanup_export_policies}->{$policy_name});

    return FCACHE_RC_SUCCESS;
} ## end sub export_policy_delete

=head2 export_policy_rule_create

Creates an export policy rule for this vserver.
Terminates the script execution if the creation fails
(including the case when the rule (with the specified ruleindex) already exists).

=head3 ARGUMENTS

=over

=item B<policyname> I<Optional> Scalar:

    Name of the export-policy rule to be created.
    Default value: constant FCACHE_DFLT_EXPORT_POLICY_RULE
    (defined in FlexCache::Constants).

=item B<ruleindex> I<Optional> Scalar:

    Rule index of the export-policy rule to be created.
    Default value: constant FCACHE_DFLT_EXPORT_POLICY_RULEINDEX
    (defined in FlexCache::Constants).

=item B<clientmatch> I<Optional> Scalar:

    Clientmatch filter to be set for the export-policy rule to be created.
    Default value: constant FCACHE_DLIF_IP_MATCH_ALL
    (defined in FlexCache::Constants).

=item B<rorule> I<Optional> Arrayref:

    Arrayref of rorules for the export-policy rule to be created.
    Default value: [FCACHE_DFLT_EXPORT_RORULE]
    (as defined in FlexCache::Constants).

=item B<rwrule> I<Optional> Arrayref:

    Arrayref of rwrules for the export-policy rule to be created.
    Default value: [FCACHE_DFLT_EXPORT_RWRULE]
    (as defined in FlexCache::Constants).

=item B<allow_suid> I<Optional> Scalar:

    Allow-suid setting for the export-policy rule to be created.
    Default value: constant FCACHE_DFLT_EXPORT_ALLOW_SUID
    (defined in FlexCache::Constants).

=item B<allow_dev> I<Optional> Scalar:

    Allow-dev setting for the export-policy rule to be created.
    Default value: constant FCACHE_DFLT_EXPORT_ALLOW_DEV
    (defined in FlexCache::Constants).

=item B<protocol> I<Optional> Arrayref:

    Arrayref of protocols for the rule to be created.
    Default value: [FCACHE_DFLT_EXPORT_PROTOCOL]
    (as defined in FlexCache::Constants).

=item B<superuser> I<Optional> Arrayref:

    Arrayref of superusers for the rule to be created.
    Default value: [FCACHE_DFLT_EXPORT_SUPERUSER]
    (as defined in FlexCache::Constants).

=item B<anon> I<Optional> Scalar:

    Anon setting for the export-policy rule to be created.
    Default value: constant FCACHE_DFLT_EXPORT_ANON
    (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	# Create with the default settings
	$obj->export_policy_rule_create();
	# Create custom
	$obj->export_policy_rule_create(policyname => 'custom',
	                                ruleindex => 2,
					protocol => $my_protocol,
					rorule => [$rorule],);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           In case an entry with the required name and ruleindex already
#           exists, it is reused (no other checks are executed).
#           Default value: FALSE.
sub export_policy_rule_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policyname =>
                {type => SCALAR, default => FCACHE_DFLT_EXPORT_POLICY_RULE,},
            ruleindex => {
                type    => SCALAR,
                default => FCACHE_DFLT_EXPORT_POLICY_RULEINDEX,
            },
            clientmatch =>
                {type => SCALAR, default => FCACHE_DLIF_IP_MATCH_ALL,},
            rorule =>
                {type => ARRAYREF, default => [FCACHE_DFLT_EXPORT_RORULE],},
            rwrule =>
                {type => ARRAYREF, default => [FCACHE_DFLT_EXPORT_RWRULE],},
            allow_suid =>
                {type => SCALAR, default => FCACHE_DFLT_EXPORT_ALLOW_SUID,},
            allow_dev =>
                {type => SCALAR, default => FCACHE_DFLT_EXPORT_ALLOW_DEV,},
            protocol =>
                {type => ARRAYREF, default => [FCACHE_DFLT_EXPORT_PROTOCOL],},
            superuser => {
                type    => ARRAYREF,
                default => [FCACHE_DFLT_EXPORT_SUPERUSER],
            },
            anon   => {type => SCALAR,  default => FCACHE_DFLT_EXPORT_ANON,},
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $reuse            = delete($args{_reuse});
    my $policy_name      = $args{policyname};
    my $ruleindex        = $args{ruleindex};
    my $vserver_name     = $class->{_name};
    my $ruleindex_exists = TRUE;
    my $nacl_obj;

    # The current SW behavior is the following
    # (TODO: follow up on behavior: burt 585650):
    # If rule with ruleindex x already exists,
    # then the create command will not error out, but:
    # - my "new" rule will be created, with ruleindex x and
    # - the already-existent rule will now have ruleindex x+1
    # (also applies to any x+y ruleindex).
    # We need to specifically take care of this case,
    # to fail/reuse if the rule already exists.
    # First see if the rule exists (look only by the ruleindex),
    # otherwise we may end up with the added extra entries
    try {
        $nacl_obj = NACL::C::VserverExportPolicyRule->find(
            command_interface => $class->{_nacl_node},
            filter            => {
                vserver    => $vserver_name,
                policyname => $policy_name,
                ruleindex  => $ruleindex,
            }
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching vserver export policy rule.+ found/) {
            $ruleindex_exists = FALSE;
        }
    };

    if ($ruleindex_exists == TRUE) {
        if ($reuse == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message =>
                    "An export-policy rule \"$policy_name\" with rule index"
                    . " $ruleindex already exists"
            );
        }

        # We will only get here in the internal-use case, for setup
        fcache_log_comment(
            message => "Reusing the export-policy rule \"$policy_name\" with"
                . " rule index $ruleindex for vserver $vserver_name");
        return FCACHE_RC_SUCCESS;
    }

    # Rule does not exists, create it
    NACL::C::VserverExportPolicyRule->create(
        command_interface => $class->{_nacl_node},
        vserver           => $vserver_name,
        'allow-suid'      => delete($args{allow_suid}),
        'allow-dev'       => delete($args{allow_dev}),
        %args,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub export_policy_rule_create

=head2 unix_user_create

Creates an unix user for this vserver.
Terminates the script execution if the creation fails
(including the case when the user already exists).

=head3 ARGUMENTS

=over

=item B<name> I<Optional> Scalar: Name of the unix user to be created.
Default value: FCACHE_DFLT_UNIX_USER (defined in FlexCache::Constants).

=item B<uid> I<Optional> Scalar: Uid for the unix user to be created.
Default value: FCACHE_DFLT_UNIX_UID (defined in FlexCache::Constants).

=item B<gid> I<Optional> Scalar: Gid for the unix user to be created.
Default value: FCACHE_DFLT_UNIX_GID (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	# Create with the default settings
	$obj->unix_user_create();
	# Create custom
	$obj->unix_user_create(name => $my_user,
	                       uid => $my_uid,
			       gid => $my_gid,);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           In case an entry with the required name, it is reused
#           (no other checks are executed).
#           Default value: FALSE.
sub unix_user_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name   => {type => SCALAR,  default => FCACHE_DFLT_UNIX_USER,},
            uid    => {type => SCALAR,  default => FCACHE_DFLT_UNIX_UID,},
            gid    => {type => SCALAR,  default => FCACHE_DFLT_UNIX_GID,},
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $reuse        = $args{_reuse};
    my $user         = $args{name};
    my $vserver_name = $class->{_name};
    my $user_exists  = TRUE;

    try {
        NACL::C::UnixUser->find(
            'command_interface' => $class->{_nacl_node},
            filter              => {
                vserver => $vserver_name,
                user    => $user,
            },
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching unix user.+ found/) {
            $user_exists = FALSE;
        }
    };

    if ($user_exists == TRUE) {
        if ($reuse == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "An unix user \"$user\" already exists"
                    . " for vserver $vserver_name"
            );
        }

        # We will only get here in the internal-use case, for setup
        fcache_log_comment(message =>
                "Reusing the unix user \"$user\" vserver $vserver_name")
            if $class->{_debug};
        return FCACHE_RC_SUCCESS;
    }

    # User does not exists, create it
    NACL::C::UnixUser->create(
        command_interface => $class->{_nacl_node},
        vserver           => $vserver_name,
        user              => $user,
        id                => $args{uid},
        'primary-gid'     => $args{gid},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub unix_user_create

=head2 unix_group_create

Creates an unix group for this vserver.
Terminates the script execution if the creation fails
(including the case when the group already exists).

=head3 ARGUMENTS

=over

=item B<name> I<Optional> Scalar: Name of the unix group to be created.
Default value: FCACHE_DFLT_UNIX_GROUP (defined in FlexCache::Constants).

=item B<gid> I<Optional> Scalar: Gid for the unix group to be created.
Default value: FCACHE_DFLT_UNIX_GID (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	# Create with the default settings
	$obj->unix_group_create();
	# Create custom
	$obj->unix_group_create(name => $my_group,
	                        gid => $my_gid,);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           In case an entry with the required name, it is reused
#           (no other checks are executed).
#           Default value: FALSE.
sub unix_group_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name   => {type => SCALAR,  default => FCACHE_DFLT_UNIX_GROUP,},
            gid    => {type => SCALAR,  default => FCACHE_DFLT_UNIX_GID,},
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $reuse        = $args{_reuse};
    my $group        = $args{name};
    my $vserver_name = $class->{_name};
    my $entry_exists = TRUE;

    try {
        NACL::C::UnixGroup->find(
            'command_interface' => $class->{_nacl_node},
            filter              => {
                vserver => $vserver_name,
                name    => $group,
            },
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching unix group.+ found/) {
            $entry_exists = FALSE;
        }
    };

    if ($entry_exists == TRUE) {
        if ($reuse == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "An unix group \"$group\" already exists"
                    . " for vserver $vserver_name"
            );
        }

        # We will only get here in the internal-use case, for setup
        fcache_log_comment(message =>
                "Reusing the unix group \"$group\" vserver $vserver_name")
            if $class->{_debug};
        return FCACHE_RC_SUCCESS;
    }

    # Group does not exists, create it
    NACL::C::UnixGroup->create(
        command_interface => $class->{_nacl_node},
        vserver           => $vserver_name,
        name              => $group,
        id                => $args{gid},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub unix_group_create

=head2 dns_services_create

Creates a dns service for this vserver.
Terminates the script execution if the creation fails
(including the case when an entry already exists).

=head3 ARGUMENTS

=over

=item B<domains> I<Required> Arrayref:

    Domains for which this dns service to be configured.

=item B<name_servers> I<Required> Arrayref:

    Name-servers to be used for this dns service.

=item B<state> I<Optional> Scalar:
    State of this new DNS service.
    Legal values: constants FCACHE_STATE_ENABLED, FCACHE_STATE_DISABLED.
    Default value: FCACHE_STATE_ENABLED.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->dns_services_create(domains => [$dns_domain1, $dns_domain2],
	                          name_servers => [$name_server1, $name_server2],);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           In case an entry exists for the required domain,
#           it is reused (the name-servers are not checked).
#           Default value: FALSE.
sub dns_services_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   domains      => {type => ARRAYREF,},
            name_servers => {type => ARRAYREF,},
            state => {type => SCALAR, default => FCACHE_STATE_ENABLED,},
            _reuse => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $reuse        = $args{_reuse};
    my $domains      = $args{domains};
    my $vserver_name = $class->{_name};
    my $entry_exists = TRUE;

    try {
        NACL::C::VserverServicesDns->find(
            'command_interface' => $class->{_nacl_node},
            filter              => {
                vserver => $vserver_name,
                state   => $args{state},
                domains => $domains,
            },
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching vserver services dns.+ found/) {
            $entry_exists = FALSE;
        }
    };

    if ($entry_exists == TRUE) {
        if ($reuse == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Dns services already exist for"
                    . " vserver $vserver_name, domains:\n"
                    . Dumper($domains)
            );
        }

        # We will only get here in the internal-use case, for setup
        fcache_log_comment(
            message => "Reusing the already configured dns services"
                . " for vserver $vserver_name")
            if $class->{_debug};
        return FCACHE_RC_SUCCESS;
    }

    # Entry does not exists, create it
    NACL::C::VserverServicesDns->create(
        command_interface => $class->{_nacl_node},
        vserver           => $vserver_name,
        domains           => $domains,
        'name-servers'    => $args{name_servers},
        state             => $args{state}
    );

    return FCACHE_RC_SUCCESS;
} ## end sub dns_services_create

=head2 name_mapping_create

Creates a name-mapping for this vserver.
Terminates the script execution if the creation fails
(including the case when an entry already exists).

=head3 ARGUMENTS

=over

=item B<direction> I<Required> Scalar:

    Direction for the name-mapping.
    Legal values: constants FCACHE_NAME_MAPPING_UNIX_WIN,
    FCACHE_NAME_MAPPING_WIN_UNIX.

=item B<pattern> I<Required> Scalar:

    Pattern to be mapped by this name-mapping.

=item B<replacement> I<Required> Scalar:

    Pattern to be replaced-with by this name-mapping.

=item B<position> I<Required> Scalar:

    Position for this name-mapping.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->name_mapping_create(direction => FCACHE_NAME_MAPPING_WIN_UNIX,
	                          pattern => $win_user,
				  replacement => $unix_user,
				  position => 1,);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           In case an entry exists, it is reused.
#           Default value: FALSE.
sub name_mapping_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   direction   => {type => SCALAR,},
            pattern     => {type => SCALAR,},
            position    => {type => SCALAR,},
            replacement => {type => SCALAR,},
            _reuse      => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $reuse        = delete($args{_reuse});
    my $direction    = $args{direction};
    my $vserver_name = $class->{_name};
    my $entry_exists = TRUE;

    try {
        NACL::C::VserverNameMapping->find(
            'command_interface' => $class->{_nacl_node},
            filter              => {
                vserver => $vserver_name,
                %args
            },
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching vserver name mapping.+ found/) {
            $entry_exists = FALSE;
        }
    };

    if ($entry_exists == TRUE) {
        if ($reuse == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message => "Name-mapping already exists for $vserver_name,\n"
                    . Dumper(%args)
            );
        }

        # We will only get here in the internal-use case, for setup
        fcache_log_comment(message =>
                "Reusing the $direction name-mapping for vserver $vserver_name"
        ) if $class->{_debug};
        return FCACHE_RC_SUCCESS;
    }

    # Entry does not exists, create it
    NACL::C::VserverNameMapping->create(
        command_interface => $class->{_nacl_node},
        vserver           => $vserver_name,
        %args
    );

    return FCACHE_RC_SUCCESS;
} ## end sub name_mapping_create

=head2 cifs_create

Creates a CIFS configuration on this vserver.
Terminates the script execution if the creation fails
(including the case when an entry already exists).

=head3 ARGUMENTS

=over

=item B<name> I<Required> Scalar:

	Name of the cifs configuration.

=item B<domain> I<Required> Scalar:

	Name of the domain.

=item B<username> I<Required> Scalar:

	Username.

=item B<password> I<Required> Scalar:

	Password.

=item B<ou> I<Required> Scalar:

	Ou for this configuration.
	Default value: FCACHE_DFLT_CIFS_OU
	(defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->cifs_create(name => $cifs_server,
	                 domain => $cifs_domain_long,
			 username => $ad_server_user,
			 password => $ad_server_pass,);

=cut

# Internal arguments:
# - _reuse: Optional: Boolean: Flag to overwrite the documented
#           behavior and allow the reuse of an already-existent entry.
#           In case an entry exists, it is reused.
#           Default value: FALSE.
sub cifs_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name     => {type => SCALAR,},
            domain   => {type => SCALAR,},
            username => {type => SCALAR,},
            password => {type => SCALAR,},
            ou       => {type => SCALAR, default => FCACHE_DFLT_CIFS_OU,},
            _reuse   => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $reuse            = delete($args{_reuse});
    my $cifs_server_name = delete($args{name});
    my $vserver_name     = $class->{_name};
    my $entry_exists     = TRUE;

    try {
        NACL::C::VserverCifs->find(
            'command_interface' => $class->{_nacl_node},
            filter              => {
                vserver       => $vserver_name,
                'cifs-server' => $cifs_server_name,
            },
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching vserver cifs.+ found/) {
            $entry_exists = FALSE;
        }
    };

    if ($entry_exists == TRUE) {
        if ($reuse == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "CIFS server $cifs_server_name is already"
                    . " configured for $vserver_name"
            );
        }

        # We will only get here in the internal-use case, for setup
        fcache_log_comment(
            message => "Reusing the CIFS server $cifs_server_name"
                . " for vserver $vserver_name")
            if $class->{_debug};
        return FCACHE_RC_SUCCESS;
    }

    # Entry does not exists, create it
    NACL::C::VserverCifs->create(
        command_interface => $class->{_nacl_node},
        vserver           => $vserver_name,
        'cifs-server'     => $cifs_server_name,
        %args
    );

    return FCACHE_RC_SUCCESS;
} ## end sub cifs_create

=head2 antivirus_policy_create

Create an antivirus on-demand policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Required> Scalar:

	Name of the antivirus policy.

=item B<scan> I<Optional> Scalar:

	Scan status: FCACHE_STATE_ON/FCACHE_STATE_OFF.
	Default value: FCACHE_STATE_ON.

=item B<scan_mandatory> I<Optional> Scalar:

	Scan-mandatory status: FCACHE_STATE_ON/FCACHE_STATE_OFF.
	Default value: FCACHE_STATE_ON.

=item B<protocols> I<Optional> Arrayref:

	Arrayref of protocols for which to apply this policy.
	Default value: [FCACHE_AV_PROTOCOL_CIFS,
	                FCACHE_AV_PROTOCOL_NFS4,
			FCACHE_AV_PROTOCOL_NFS3].

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->antivirus_policy_create(policy_name => 'my_av_policy');

=cut

sub antivirus_policy_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policy_name    => {type => SCALAR,},
            scan           => {type => SCALAR, default => FCACHE_STATE_ON,},
            scan_mandatory => {type => SCALAR, default => FCACHE_STATE_ON,},
            protocols => {
                type    => ARRAYREF,
                default => [
                    FCACHE_AV_PROTOCOL_CIFS, FCACHE_AV_PROTOCOL_NFS4,
                    FCACHE_AV_PROTOCOL_NFS3
                ],
            },
        }
    );

    my $policy_name = $args{policy_name};

    NACL::C::AntivirusOnAccessPolicy->create(
        command_interface => $class->{_nacl_node},
        name              => $policy_name,
        scan              => $args{scan},
        'scan-mandatory'  => $args{scan_mandatory},
        protocols         => $args{protocols},
        vserver           => $class->{_name},
    );
    $class->{_cleanup_av_policies}->{$policy_name} = 1;

    return FCACHE_RC_SUCCESS;
} ## end sub antivirus_policy_create

=head2 antivirus_policy_modify

Modify an antivirus on-demand policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Required> Scalar:

	Name of the antivirus policy.

=item B<scan> I<Optional> Scalar:

	Scan status: FCACHE_STATE_ON/FCACHE_STATE_OFF.
	Default value: FCACHE_STATE_ON.

=item B<scan_mandatory> I<Optional> Scalar:

	Scan-mandatory status: FCACHE_STATE_ON/FCACHE_STATE_OFF.
	Default value: FCACHE_STATE_ON.

=item B<protocols> I<Optional> Arrayref:

	Arrayref of protocols for which to apply this policy.
	Default value: [FCACHE_AV_PROTOCOL_CIFS, FCACHE_AV_PROTOCOL_NFS4].

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->antivirus_policy_create(policy_name => 'my_av_policy');

=cut

sub antivirus_policy_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policy_name    => {type => SCALAR,},
            scan           => {type => SCALAR, default => FCACHE_STATE_ON,},
            scan_mandatory => {type => SCALAR, default => FCACHE_STATE_ON,},
            protocols => {
                type    => ARRAYREF,
                default => [FCACHE_AV_PROTOCOL_CIFS, FCACHE_AV_PROTOCOL_NFS4],
            },
        }
    );

    my $policy_name = $args{policy_name};

    NACL::C::AntivirusOnAccessPolicy->modify(
        command_interface => $class->{_nacl_node},
        name              => $policy_name,
        scan              => $args{scan},
        'scan-mandatory'  => $args{scan_mandatory},
        protocols         => $args{protocols},
        vserver           => $class->{_name},
    );
    $class->{_cleanup_av_policies}->{$policy_name} = 1;

    return FCACHE_RC_SUCCESS;
} ## end sub antivirus_policy_modify

=head2 antivirus_policy_delete

Delete an antivirus on-demand policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Required> Scalar:

	Name of the antivirus policy.

=item B<scan> I<Optional> Scalar:

	Scan status: FCACHE_STATE_ON/FCACHE_STATE_OFF.
	Default value: FCACHE_STATE_ON.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->antivirus_policy_delete(policy_name => 'my_av_policy');

=cut

sub antivirus_policy_delete {
    my $class = shift;
    my %args = validate(@_, {policy_name => {type => SCALAR,},});

    my $policy_name = $args{policy_name};

    NACL::C::AntivirusOnAccessPolicy->delete(
        command_interface => $class->{_nacl_node},
        name              => $policy_name,
        vserver           => $class->{_name},
    );
    delete($class->{_cleanup_av_policies}->{$policy_name});

    return FCACHE_RC_SUCCESS;
} ## end sub antivirus_policy_delete

=head2 flexcache_cache_policy_modify

Modify a flexcache cache policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Optional> Scalar:

	Name of the antivirus policy.
	It defaults to the default policy.

=item B<deleg_lru_timeout> I<Optional> Scalar:

	New value for the deleg_lru_timeout field.

=item B<dir_ttl> I<Optional> Scalar:

	New value for the dir_ttl field.

=item B<reg_ttl> I<Optional> Scalar:

	New value for the reg_ttl field.

=item B<meta_ttl> I<Optional> Scalar:

	New value for the meta_ttl field.

=item B<sym_ttl> I<Optional> Scalar:

	New value for the sym_ttl field.

=item B<other_ttl> I<Optional> Scalar:

	New value for the other_ttl field.

=item B<prefer_local_cache> I<Optional> Scalar:

	New value for the prefer_local_cache field.
	
=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->flexcache_cache_policy_modify(deleg_lru_timeout => 200);

=cut

sub flexcache_cache_policy_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policy_name        => {type => SCALAR, default  => 'default',},
            deleg_lru_timeout  => {type => SCALAR, optional => 1,},
            dir_ttl            => {type => SCALAR, optional => 1,},
            reg_ttl            => {type => SCALAR, optional => 1,},
            meta_ttl           => {type => SCALAR, optional => 1,},
            sym_ttl            => {type => SCALAR, optional => 1,},
            other_ttl          => {type => SCALAR, optional => 1,},
            prefer_local_cache => {type => SCALAR, optional => 1,},
            ui_type            => {type => SCALAR, optional => 1,},
        }
    );

    my $policy_name    = delete($args{policy_name});
    my $need_to_modify = FALSE;
    my $ui_type        = delete($args{ui_type});
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    #Obtain the current settings, to be able to reset them at cleanup
    my $policy = NACL::C::VolumeFlexcacheCachePolicy->find(
        command_interface => $class->{_nacl_node},
        filter            => {
            policy  => $policy_name,
            vserver => $class->{_name},
        }
    );
    my $current_settings = $policy->state();

    my %nacl_args = (
        vserver => $class->{_name},
        policy  => $policy_name
    );
    foreach my $field (keys(%args)) {
        my $new_value     = $args{$field};
        my $current_value = $current_settings->$field();

        if ("$new_value" eq "$current_value") {
            fcache_log_comment(message => "Flexcache cache-policy modify:\n"
                    . "Policy: $policy_name\n"
                    . "Field $field is already set to $new_value");
            next;
        }

        # Only for the default policy (other policies shold be deleted
        # as part of cleanup):
        # Only remember the value from the first time we ever change
        # a setting. This is what we need to reset to at cleanup.
        if (($policy_name eq 'default')
            && (!exists(
                    $class->{_cleanup_default_flexcache_cache_policy}
                        ->{$field}
                )
            )
            ) {
            $class->{_cleanup_default_flexcache_cache_policy}->{$field}
                = $current_value;
        }

        $field =~ s/_/-/g;
        $nacl_args{$field} = $new_value;
        $need_to_modify = TRUE;
    }

    if ($need_to_modify) {
        NACL::C::VolumeFlexcacheCachePolicy->modify(
            command_interface => $class->{_nacl_node},
            apiset_must       => {interface => $ui_type},
            %nacl_args
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_cache_policy_modify

=head2 flexcache_cache_policy_verify

Verify the settings of a flexcache cache policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Optional> Scalar:

	Name of the flexcache cache policy.
	It defaults to the default policy.

=item B<deleg_lru_timeout> I<Optional> Scalar:

	Expected value for the deleg_lru_timeout field.

=item B<dir_ttl> I<Optional> Scalar:

	Expected value for the dir_ttl field.

=item B<reg_ttl> I<Optional> Scalar:

	Expected value for the reg_ttl field.

=item B<meta_ttl> I<Optional> Scalar:

	Expected value for the meta_ttl field.

=item B<sym_ttl> I<Optional> Scalar:

	Expected value for the sym_ttl field.

=item B<other_ttl> I<Optional> Scalar:

	Expected value for the other_ttl field.

=item B<prefer_local_cache> I<Optional> Scalar:

	Expected value for the prefer_local_cache field.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=item B<zapi_cmd_type> I<Optional> Scalar:

	ZAPI command type to be used to verify the cache policy (get or
	get-iter). This argument is only used if ui_type is ZAPI. Legal values:
	FCACHE_ZAPI_CMD_TYPE_GET, FCACHE_ZAPI_CMD_TYPE_GET_ITER. Default value:
	FCACHE_ZAPI_CMD_TYPE_GET_ITER.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->flexcache_cache_policy_verify(deleg_lru_timeout => 200);
	$obj->flexcache_cache_policy_verify(
            deleg_lru_timeout => 200,
	    ui_type => FCACHE_UI_TYPE_ZAPI,
	    zapi_cmd_type => FCACHE_ZAPI_CMD_TYPE_GET
	);

=cut

sub flexcache_cache_policy_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policy_name        => {type => SCALAR, default  => 'default',},
            deleg_lru_timeout  => {type => SCALAR, optional => 1,},
            dir_ttl            => {type => SCALAR, optional => 1,},
            reg_ttl            => {type => SCALAR, optional => 1,},
            meta_ttl           => {type => SCALAR, optional => 1,},
            sym_ttl            => {type => SCALAR, optional => 1,},
            other_ttl          => {type => SCALAR, optional => 1,},
            prefer_local_cache => {type => SCALAR, optional => 1,},
            ui_type            => {type => SCALAR, optional => 1,},
            zapi_cmd_type =>
                {type => SCALAR, default => FCACHE_ZAPI_CMD_TYPE_GET_ITER,},
        }
    );

    my $policy_name   = delete($args{policy_name});
    my $zapi_cmd_type = delete($args{zapi_cmd_type});
    my $ui_type       = delete($args{ui_type});
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    if (   ($zapi_cmd_type ne FCACHE_ZAPI_CMD_TYPE_GET_ITER)
        && ($zapi_cmd_type ne FCACHE_ZAPI_CMD_TYPE_GET)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Illegal zapi_cmd_type passed in: $zapi_cmd_type"
        );
    }

    my $nacl_filter = {
        vserver => $class->{_name},
        policy  => $policy_name,
    };
    foreach my $key (keys(%args)) {
        my $value = $args{$key};
        $key =~ s/_/-/g;
        $nacl_filter->{$key} = $value;
    }

    if (   ($ui_type eq FCACHE_UI_TYPE_ZAPI)
        && ($zapi_cmd_type eq FCACHE_ZAPI_CMD_TYPE_GET)) {
        my $apiset = $class->{_nacl_node}->apiset(
            category  => "Node",
            interface => "ZAPI",
        );
        my $res = $apiset->flexcache_cache_policy_get(
            'cache-policy-name' => $policy_name,
            vserver             => $class->{_name},
        );

        my $parsed_output = $res->get_parsed_output();
        my $attributes    = $parsed_output->[0]->{attributes}->[0]
            ->{'flexcache-cache-policy-info'}->[0];
        if (!$attributes) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Failed to detect cache policy information:\n"
                    . "Vserver: "
                    . $class->{_name} . "\n"
                    . "Flexcache cache policy: $policy_name\n"
                    . Dumper($parsed_output)
            );
        }

        my $attributes_mapping = {
            deleg_lru_timeout  => 'deleg-lru-timeout',
            dir_ttl            => 'time-to-live-directories',
            reg_ttl            => 'time-to-live-files',
            meta_ttl           => 'time-to-live-metafiles',
            sym_ttl            => 'time-to-live-symbolic',
            other_ttl          => 'time-to-live-other',
            prefer_local_cache => 'prefer-local-cache',
        };

        foreach my $arg (keys(%args)) {
            my $arg_mapping = $attributes_mapping->{$arg};
            if (!defined($arg_mapping)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Internal library problem found:\n"
                        . "Attribute mapping not defined for $arg"
                );
            }

            my $expected_value = $args{$arg};
            my $obtained_value = $attributes->{$arg_mapping};

            if ("$expected_value" ne "$obtained_value") {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message => "Mismatch in flexcache cache policy field:\n"
                        . "Vserver: "
                        . $class->{_name} . "\n"
                        . "Flexcache cache policy: $policy_name\n"
                        . "Field: $arg\n"
                        . "Expected value: $expected_value\n"
                        . "Obtained value: $obtained_value"
                );
            }

            fcache_log_comment(
                      message => "Verified flexcache cache policy field:\n"
                    . "Vserver: "
                    . $class->{_name} . "\n"
                    . "Flexcache cache policy: $policy_name\n"
                    . "Field: $arg\n"
                    . "Expected value: $expected_value\n"
                    . "Obtained value: $obtained_value");
        }
    } else {
        NACL::C::VolumeFlexcacheCachePolicy->find(
            command_interface => $class->{_nacl_node},
            apiset_must       => {interface => $ui_type},
            filter            => $nacl_filter,
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_cache_policy_verify

=head2 flexcache_cache_policy_create

Create a flexcache cache policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Required> Scalar:

	Name of the flexcache cache policy.

=item B<deleg_lru_timeout> I<Optional> Scalar:

	Value for the deleg_lru_timeout field.

=item B<dir_ttl> I<Optional> Scalar:

	Value for the dir_ttl field.

=item B<reg_ttl> I<Optional> Scalar:

	Value for the reg_ttl field.

=item B<meta_ttl> I<Optional> Scalar:

	Value for the meta_ttl field.

=item B<sym_ttl> I<Optional> Scalar:

	Value for the sym_ttl field.

=item B<other_ttl> I<Optional> Scalar:

	Value for the other_ttl field.

=item B<prefer_local_cache> I<Optional> Scalar:

	Value for the prefer_local_cache field.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->flexcache_cache_policy_create(policy_name => 'custom_policy',
	                                    dir_ttl => 60,
					    reg_ttl => 60);

=cut

sub flexcache_cache_policy_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policy_name        => {type => SCALAR,},
            deleg_lru_timeout  => {type => SCALAR, optional => 1,},
            dir_ttl            => {type => SCALAR, optional => 1,},
            reg_ttl            => {type => SCALAR, optional => 1,},
            meta_ttl           => {type => SCALAR, optional => 1,},
            sym_ttl            => {type => SCALAR, optional => 1,},
            other_ttl          => {type => SCALAR, optional => 1,},
            prefer_local_cache => {type => SCALAR, optional => 1,},
            ui_type            => {type => SCALAR, optional => 1,},
        }
    );

    my $policy_name = delete($args{policy_name});
    my $ui_type     = delete($args{ui_type});
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    my %nacl_args = (
        command_interface => $class->{_nacl_node},
        apiset_must       => {interface => $ui_type},
        vserver           => $class->{_name},
        policy            => $policy_name,
    );
    foreach my $field (keys(%args)) {
        my $value = $args{$field};
        $field =~ s/_/-/g;
        $nacl_args{$field} = $value;
    }

    NACL::C::VolumeFlexcacheCachePolicy->create(%nacl_args);
    $class->{_cleanup_flexcache_cache_policy}->{$policy_name} = 1;

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_cache_policy_create

=head2 flexcache_cache_policy_delete

Delete a flexcache cache policy for this vserver.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Required> Scalar:

	Name of the flexcache cache policy.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->flexcache_cache_policy_delete(policy_name => 'custom_policy');

=cut

sub flexcache_cache_policy_delete {
    my $class = shift;
    my %args  = validate(
        @_,
        {   policy_name => {type => SCALAR,},
            ui_type     => {type => SCALAR, optional => 1,},
        }
    );

    my $policy_name = $args{policy_name};
    my $ui_type     = delete($args{ui_type});
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    NACL::C::VolumeFlexcacheCachePolicy->delete(
        command_interface => $class->{_nacl_node},
        apiset_must       => {interface => $ui_type},
        vserver           => $class->{_name},
        policy            => $policy_name,
    );
    delete($class->{_cleanup_flexcache_cache_policy}->{$policy_name});

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_cache_policy_delete

########################################################
# Internal methods
########################################################

# Any method that is not specifically implemented in this module
# will terminate the script execution with an N/A result.
# We do this in the idea that the script called it
# with a very good reason (to do something with the vserver/vfiler)
sub AUTOLOAD {
    our $AUTOLOAD;
    my $class  = shift;
    my $method = $AUTOLOAD;

    fcache_log_result(
        result_type => FCACHE_LOGRESULT_NA,
        message => "This specific method is not available for C-mode: $method"
    );
} ## end sub AUTOLOAD

# Perform common setup for this vserver (NFS/CIFS)

sub _setup_common {
    my $class = shift;
    my %args = validate(@_, {});

    my $vserver = $class->{_name};
    fcache_log_comment(
        message => "SETUP: Creating/reusing data lif for vserver $vserver");
    my $routing_group = $class->dlif_create(
        name      => $class->{_dlif_name},
        home_port => $class->{__dlif_port},
        address   => $class->{_dlif_ip},
        netmask   => $class->{__dlif_mask},
        _reuse    => $class->{_recycle},
    );

    fcache_log_comment(
        message => "SETUP: Creating/reusing the route to the data lif");
    $class->dlif_route_create(
        routing_group => $routing_group,
        gateway       => $class->{__dlif_gateway},
        _reuse        => $class->{_recycle},
    );

    fcache_log_comment(
        message => "SETUP: Creating/reusing the default export policy rule");
    $class->export_policy_rule_create(_reuse => TRUE);
} ## end sub _setup_common

# Perform NFS setup for this vserver:
# - create/reuse a data interface, using the information obtained from the tharnhost file
# - create/reuse a route for this data interface
# - create/reuse an nfs configuration for the vserver
# - create/reuse a default export-policy rule for the vserver
# Arguments: none.
# Return: FCACHE_RC_SUCCESS.
sub _setup_nfs {
    my $class = shift;
    my %args = validate(@_, {});

    my $vserver = $class->{_name};
    fcache_log_comment(message =>
            "SETUP: Creating/reusing the NSF configuration for vserver $vserver"
    );
    $class->nfs_create(_reuse => $class->{_recycle},);

    fcache_log_comment(message =>
            "SETUP: Creating/reusing the default user for vserver $vserver");
    $class->unix_user_create(_reuse => TRUE);

    fcache_log_comment(message =>
            "SETUP: Creating/reusing the default group for vserver $vserver");
    $class->unix_group_create(_reuse => TRUE);

    return FCACHE_RC_SUCCESS;
} ## end sub _setup_nfs

# Perform CIFS setup for this vserver
# - ensure that the dns services are configured
# - create/reuse name-mappings: unix-win, win-unix
# - configure the cifs server
# Arguments: none.
# Return: FCACHE_RC_SUCCESS.
sub _setup_cifs {
    my $class = shift;
    my %args = validate(@_, {});

    my $options          = NACL::OptionSet->new();
    my $dns_domain       = $options->get('DNS_DOMAIN');
    my $dns_server       = $options->get('DNS_SERVER');
    my $ad_server        = $options->get('AD_SERVER');
    my $ad_server_user   = $options->get('AD_SERVER_USER');
    my $ad_server_pass   = $options->get('AD_SERVER_PASSWORD');
    my $cifs_domain      = $options->get('CIFS_DOMAIN');
    my $cifs_domain_long = $options->get('CIFS_DOMAIN_LONG');
    my $vserver          = $class->{_name};

    if (   (!defined($dns_domain))
        || (!defined($dns_server))
        || (!defined($ad_server))
        || (!defined($ad_server_user))
        || (!defined($ad_server_pass))
        || (!defined($cifs_domain))
        || (!defined($cifs_domain_long))) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "For complete CIFS setup, please provide all of the command-line arguments:\n"
                . 'DNS_DOMAIN, DNS_SERVER, AD_SERVER, AD_SERVER_USER, AD_SERVER_PASSWORD, '
                . 'CIFS_DOMAIN, CIFS_DOMAIN_LONG'
        );
    }

    my $win_user = $cifs_domain . '\\\\' . $ad_server_user;

    fcache_log_comment(message => "SETUP: Configuring dns services");
    $class->dns_services_create(
        domains      => [$dns_domain],
        name_servers => [$dns_server, $ad_server],
        _reuse       => TRUE,
    );

    fcache_log_comment(message => "SETUP: Configuring name-mappings");
    $class->name_mapping_create(
        direction   => FCACHE_NAME_MAPPING_UNIX_WIN,
        pattern     => FCACHE_DFLT_UNIX_USER,
        replacement => $win_user,
        position    => FCACHE_DFLT_NAME_MAPPING_POSITION,
        _reuse      => TRUE
    );
    $class->name_mapping_create(
        direction   => FCACHE_NAME_MAPPING_WIN_UNIX,
        pattern     => $win_user,
        replacement => FCACHE_DFLT_UNIX_USER,
        position    => FCACHE_DFLT_NAME_MAPPING_POSITION,
        _reuse      => TRUE
    );

    fcache_log_comment(
        message => "SETUP: Configuring cifs server for $vserver");
    my $option_cifs_server_name = $options->get('CIFS_SERVER_NAME');
    my $option_cifs_ou          = $options->get('CIFS_OU');
    my $cifs_server
        = (defined($option_cifs_server_name))
        ? ($option_cifs_server_name)
        : $vserver;
    my $cifs_ou
        = (defined($option_cifs_ou))
        ? ($option_cifs_ou)
        : FCACHE_DFLT_CIFS_OU;
    $class->cifs_create(
        name     => $cifs_server,
        domain   => $cifs_domain_long,
        ou       => $cifs_ou,
        username => $ad_server_user,
        password => $ad_server_pass,
        _reuse   => TRUE
    );

    return FCACHE_RC_SUCCESS;
} ## end sub _setup_cifs

# Class destructor
# Unless the constructor was called with perform_cleanup set to FALSE,
# it takes care of any cleanup that needs to be done on the vserver:
# - If recycling is not requested, then the vserver is deleted
# - If recycling is requested, then:
#   - any antivirus policies created by this object are deleted
#   - any export policies created by this object are deleted
#   - any modifications to the default flexcache cache policies are reverted
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if (($class->{_creator_pid} == $$) && ($class->{_perform_cleanup})) {
        fcache_log_comment(
            message => "CLEANUP: Starting CMode Vserver cleanup for "
                . $class->{_name})
            if $class->{_debug};

        if ($class->{_recycle}) {
            fcache_log_comment(message =>
                    "CLEANUP: Recycling requested. Leaving behind the vserver "
                    . $class->{_name});

            fcache_log_comment(
                message => "CLEANUP: Making sure all data lifs are home");
            $class->dlifs_migrate_back_home();

            my $cleanup_dlif_status = $class->{_cleanup_dlif_status_admin};
            foreach my $lif (keys(%{$cleanup_dlif_status})) {
                fcache_log_comment(
                    message => "CLEANUP: Modifying status-admin of lif $lif");
                $class->dlif_status_modify(
                    dlif_name => $lif,
                    status    => $cleanup_dlif_status->{$lif}
                );
            }

            my $cleanup_av_policies = $class->{_cleanup_av_policies};
            foreach my $policy (keys(%{$cleanup_av_policies})) {
                fcache_log_comment(message =>
                        "CLEANUP: Deleting antivirus policy $policy for vserver "
                        . $class->{_name});
                $class->antivirus_policy_delete(policy_name => $policy);
            }

            my $export_policies = $class->{_cleanup_export_policies};
            foreach my $exp_policy (keys(%{$export_policies})) {
                fcache_log_comment(message =>
                        "CLEANUP: Deleting export policy $exp_policy for vserver "
                        . $class->{_name});
                $class->export_policy_delete(policyname => $exp_policy);
            }

            my $cache_policies = $class->{_cleanup_flexcache_cache_policy};
            foreach my $policy (keys(%{$cache_policies})) {
                fcache_log_comment(message =>
                        "CLEANUP: Deleting cache policy $policy for vserver "
                        . $class->{_name});
                $class->flexcache_cache_policy_delete(policy_name => $policy);
            }

            my $cleanup_default_fcache_policy
                = $class->{_cleanup_default_flexcache_cache_policy};
            if (scalar(keys(%{$cleanup_default_fcache_policy})) != 0) {
                fcache_log_comment(
                    message => "CLEANUP: Resetting the default flexcache"
                        . " cache-policy for vserver "
                        . $class->{_name});
                $class->flexcache_cache_policy_modify(
                    %{$cleanup_default_fcache_policy});
            }
        } else {
            fcache_log_comment(
                message => "CLEANUP: Purging the vserver " . $class->{_name});
            $class->{_nacl_vserver}->purge(nacltask_if_purged => 'pass');
        }
    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

