#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Cache.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Module providing complete functionality for a flex-cache.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Cache

=head1 DESCRIPTION

This module provides release and mode agnostic functionality for all the 
flexcache scripts. It is responsible for performing all the steps that are 
needed to configure a cache volume. The module is release agnostic and provides 
functionality for both 7-mode and C-mode. It is NACL-compliant. To be noted, 
though, that no direct access to any NACL object is given. This is to ensure 
the encapsulation that an object-oriented design must follow and to not break
the cleanup mechanism provided by the FlexCache library. A FlexCache::Cache 
object will have the following members, to be used by scripts:

=over

=item node

This member is responsible for the node-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Node.

=item aggregate

This member is responsible for the aggregate-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Aggregate.

=item volume

This member is responsible for the volume-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::CacheVolume.

=item vserver

This member is responsible for the vserver/vfiler level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Vserver.

=item client

This member is responsible for the client-side functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Client.

=back

The test cleanup will be executed by the library, within the destructor of each
object that was instantiated by a script. A destructor will be responsible for
 cleaning up after the actions that its respective object executed.

=head1 SYNOPSIS
	
	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Cache;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
        
        my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
                                          aggregate => {name => $AGGR_DST,
                                                        block_type => $AGGR_DST_BLOCK_TYPE,},
                                          volume => {name => $VOL_DST,},
                                          vserver => {name => $VSERVER,
                                                      dlif_name => $DLIF_DST,
                                                      dlif_id => $DLIF_DST_ID,},
                                          client => {name => $CLIENT_DST,},
                                          origin => $origin);

=head1 UNIT TESTING

Unit tests for this module: All the scripts under:

        flexcache/library_unit_tests/

=cut

package FlexCache::Cache;

# Standard libs
use strict;
use warnings;
use Data::Dumper;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

# NetApp libs
use base qw(FlexCache::Origin);
use NACL::STask::Node;
use FlexCache::Constants qw(:all);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result
    fcache_get_random_name);
use FlexCache::Node;
use FlexCache::Aggregate;
use FlexCache::ClientOperations;
use FlexCache::Vserver;
use FlexCache::CacheVolume;
use FlexCache::UI::OntapMessages;

=head1 CONSTRUCTOR

=head2 new

Returns a FlexCache::Cache object, representing a cache volume, with members
that can be used for basic testing. The constructor also takes care (unless
otherwise required by the caller) of the basic setup steps needed for testing.
Please see the constructor documentation for each of the class members for
detailed descriptions of the setup steps.

=head3 ARGUMENTS

=over

=item B<node> I<Required> Hashref:

        Constructor options for the node member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Node.

=item B<aggregate> I<Required> Hashref:

        Constructor options for the aggregate member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Aggregate.

=item B<volume> I<Required> Hashref:

        Constructor options for the volume member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::CacheVolume.

=item B<client> I<Required> Hashref:

        Constructor options for the client member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::ClientOperations.

=item B<vserver> I<Required> Hashref:

        Constructor options for the vserver member.
        The vserver member represents a vserver for C-mode.
        If running on 7-mode in a multistore environment, then
        this member represents a vfiler.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Vserver.

=item B<origin> I<Required> Object:

        FlexCache::Origin object corresponding to the origin for this cache.

=item B<debug> I<Optional> Boolean:

        Flag to increase the verbosity of the library, useful when debugging.
        Default value: The "debug" settings of the origin object.

=item B<perform_cleanup> I<Optional> Boolean:

        Flag to control if the cleanup should be executed.
        Default value: The "perform_cleanup" settings of the origin object.

=item B<attach> I<Optional> Boolean:

        Setting this flag to true means that all the resources
	(aggregate, vserver, volume) are required to already exist.
	In this case, the library will not attempt to create any of them,
        but it will terminate script execution if they do not already exist.
        Default value: FALSE.

=item B<verify_vldb> I<Optional> Boolean:

        Flag to control if a vldb-check should be executed by the constructor.
        If this flag is set to true, then the cache vldb is checked against
	the origin vldb. An exception is thrown in case the expected results
	are not seen (MSIDs should match, DSIDs should not).
        This argument is ignored in 7-mode.
        Default value: TRUE.

=back

=head3 EXAMPLE

        my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
        
	
	my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
					  aggregate => {name => $AGGR_DST,
							block_type => $AGGR_DST_BLOCK_TYPE,},
					  volume => {name => $VOL_DST,},
					  vserver => {name => $VSERVER,
						      dlif_name => $DLIF_DST,
						      dlif_id => $DLIF_DST_ID,},
					  client => {name => $CLIENT_DST,},
                                          origin => $origin,);

=cut

sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   node            => {type => HASHREF,},
            aggregate       => {type => HASHREF,},
            volume          => {type => HASHREF,},
            client          => {type => HASHREF,},
            vserver         => {type => HASHREF, optional => 1,},
            origin          => {type => OBJECT, isa => "FlexCache::Origin"},
            attach          => {type => BOOLEAN, default => FALSE,},
            debug           => {type => BOOLEAN, optional => 1,},
            perform_cleanup => {type => BOOLEAN, optional => 1,},
            verify_vldb     => {type => BOOLEAN, default => TRUE,},
        }
    );

    my $self = {};

    my $origin = $args{origin};
    $self->{_debug}
        = (defined($args{debug})) ? $args{debug} : $origin->{_debug};
    $self->{_perform_cleanup}
        = (defined($args{perform_cleanup}))
        ? $args{perform_cleanup}
        : $origin->{_perform_cleanup};
    $self->{_attach} = $args{attach};

    bless($self, $class);

    # WARNING!!!!
    # A FlexCache::Cache object is inheriting from FlexCache::Origin
    # For this to work correctly,
    # make sure that all the class members of the parent class
    # are set correctly here as well

   # Connect to the node. Create the NACL node object here
   # and pass it in to each constructor for the members.
   # Because of this, we need to process here the name entry in the node hash.
    $self->{_node_name} = delete($args{node}->{name});
    if (!defined($self->{_node_name})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Missing or undefined \"name\" field in the cache node hash:\n"
                . Dumper(%args),
        );
    }
    $self->{_nacl_node} = NACL::STask::Node->new(node => $self->{_node_name});

    # Connect to the client (for now, just to find out the client type)
    # We need to do this as a two-step process such that the vserver setup
    # is done correctly
    # before mounting/mapping the volume/share
    $self->{client} = FlexCache::ClientOperations->new(
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        _postpone_setup => TRUE,
        %{$args{client}}
    );
    my $client_type = $self->{client}->get_type();

    # Create the node member and process any needed information
    $self->{node} = FlexCache::Node->new(
        node            => $self->{_nacl_node},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        %{$args{node}}
    );
    $self->{_boot_mode} = $self->{node}->get_boot_mode();
    $self->{_version}   = $self->{node}->_get_version();

    if (   ($self->{_boot_mode} eq FCACHE_ONTAP_7MODE)
        && ($client_type eq FCACHE_CLIENT_TYPE_WINDOWS)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "FlexCache does not support CIFS in 7-mode\n"
                . "This test-case should not be run on a Windows client"
        );
    }

    $self->{_ui_messages} = FlexCache::UI::OntapMessages->new(
        boot_mode => $self->{_boot_mode},
        release   => $self->{_version},
        debug     => $self->{_debug},
    );
    $self->{node}->_attach_ui_member(ui_member => $self->{_ui_messages});

    # Create/reuse the aggregate
    $self->{aggregate} = FlexCache::Aggregate->new(
        node            => $self->{_nacl_node},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        attach_only     => $self->{_attach},
        ui_messages     => $self->{_ui_messages},
        %{$args{aggregate}}
    );
    my $aggr_name = $self->{aggregate}->get_name;

    # Create and configure the vserver

   # In the case when the cache is C-mode, the origin is 7-mode, we need to
   # make sure we are not interpreting the origin's vserver name as a required
   # vfiler for the 7-mode cache
    if (   (defined($args{vserver}->{name}))
        && ($origin->{_boot_mode}        eq FCACHE_ONTAP_CMODE)
        && ($self->{_boot_mode}          eq FCACHE_ONTAP_7MODE)
        && ($origin->{vserver}->get_name eq $args{vserver}->{name})) {
        fcache_log_comment(
                  message => "C-mode origin, 7-mode cache combination.\n"
                . "Ignoring cache vserver name: "
                . $args{vserver}->{name})
            if ($self->{_debug});
        $args{vserver}->{name} = undef;
    }

    $self->{vserver} = FlexCache::Vserver->new(
        node            => $self->{_nacl_node},
        aggregate_name  => $aggr_name,
        client_type     => $client_type,
        debug           => $self->{_debug},
        config_type     => 'cache',
        perform_cleanup => $self->{_perform_cleanup},
        attach_only     => $self->{_attach},
        ui_messages     => $self->{_ui_messages},
        %{$args{vserver}},
    );
    my $vserver_name = $self->{vserver}->get_name(_internal_use => TRUE,);

    # Create/reuse the cache volume
    my $origin_name          = $origin->{volume}->get_name();
    my $origin_junction_path = $origin->{volume}->get_junction_path();
    my $origin_ip            = $origin->{vserver}->get_dlif_ip();
    $self->{volume} = FlexCache::CacheVolume->new(
        %{$args{volume}},
        node                 => $self->{_nacl_node},
        aggregate_name       => $aggr_name,
        vserver_name         => $vserver_name,
        client_type          => $client_type,
        origin_name          => $origin_name,
        origin_junction_path => $origin_junction_path,
        origin_7mode_ip      => $origin_ip,
        attach_only          => $self->{_attach},
        debug                => $self->{_debug},
        ui_messages          => $self->{_ui_messages},
        perform_cleanup      => $self->{_perform_cleanup},
    );

    # If the client is windows, we need to ensure that we have a share
    # available on the origin (may be the case if
    # the origin was configured for NFS only)
    if ($client_type eq FCACHE_CLIENT_TYPE_WINDOWS) {
        my $share_name = $origin->{volume}
            ->_ensure_volume_has_share(reuse_action => 'reuse');
        $self->{volume}->_ensure_volume_has_share(share_name => $share_name);
    }

    # Now we can mount/map the volume/share
    my $dlif_ip        = $self->{vserver}->get_dlif_ip();
    my $mount_junction = $self->{volume}->get_client_mount_junction();
    $self->{client}->_execute_setup(
        mount_from_ip       => $dlif_ip,
        mount_from_junction => $mount_junction,
    );

    # Check the origin-cache vldb (this is only available in C-mode)
    if (($args{verify_vldb}) && ($self->{_boot_mode} eq FCACHE_ONTAP_CMODE)) {
        my %cache_vldb = $self->{volume}->get_vldb_info();
        $origin->{volume}->get_vldb_info(cache_vldb_info => \%cache_vldb);
    }

    return $self;

} ## end sub new

=head1 OTHER METHODS

This class inherits from FlexCache::Origin.
All the methods documented there are also available for FlexCache::Cache object.

=head2 ttl_modify

Modify ttl values for this cache.

=head3 ARGUMENTS

=over

=item B<dir_ttl> I<Optional> Scalar:

	New value for dir_ttl.

=item B<reg_ttl> I<Optional> Scalar:

	New value for reg_ttl.

=item B<sym_ttl> I<Optional> Scalar:

	New value for sym_ttl.

=item B<deleg_lru_timeout> I<Optional> Scalar:

	New value for deleg_lru_timeout. Not supported for 7-mode.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $cache->ttl_modify(dir_ttl => 1);

=cut

sub ttl_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dir_ttl           => {type => SCALAR, optional => 1,},
            reg_ttl           => {type => SCALAR, optional => 1,},
            sym_ttl           => {type => SCALAR, optional => 1,},
            deleg_lru_timeout => {type => SCALAR, optional => 1,},
        }
    );

    if (scalar(keys(%args)) == 0) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "At least one argument must be passed in.\n"
                . "None was provided:\n"
                . Dumper(%args)
        );
    }

    if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
        my $policy
            = fcache_get_random_name(type => 'fcache_policy_lib_created',);
        $class->{vserver}->flexcache_cache_policy_create(
            policy_name => $policy,
            %args
        );
        $class->{volume}->set_flexcache_policy(new_policy => $policy);
    } else {
        my $ttl_mapping = {
            dir_ttl => 'acdirmax',
            reg_ttl => 'acregmax',
            sym_ttl => 'acsymmax',
        };

        my $vol_name = $class->{volume}->get_name();
        my $api_set  = $class->{_nacl_node}->apiset();
        foreach my $ttl (keys(%args)) {
            my $ttl_7mode = $ttl_mapping->{$ttl};
            if (!defined($ttl_7mode)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Library problem found:\n"
                        . "No 7-mode mapping defined for ttl: $ttl"
                );
            }
            $api_set->vol_options(
                volume         => $vol_name,
                'option-name'  => $ttl_7mode,
                'option-value' => $args{$ttl},
            );
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub ttl_modify

########################################################
# Internal methods
########################################################

# Class destructor
# This top-level destructor is only to ensure the correct order of cleanup
# steps. For details on each member cleanup, please see the appropriate
# module documentation.
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    fcache_log_comment(
        message => "CLEANUP: Entering FlexCache cleanup for cache...");

    # Enforce the correct order of member cleanup
    $class->{client}    = undef;
    $class->{volume}    = undef;
    $class->{vserver}   = undef;
    $class->{aggregate} = undef;
    $class->{node}      = undef;

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

