#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/CacheVolume.pm#2 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Agnostic module for cache-level functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::CacheVolume

=head1 DESCRIPTION

This module provides release and mode agnostic volume-level functionality for all the flexcache scripts.
Its functionality is made available through the 'volume' member of the FlexCache::Cache object.
The module is release agnostic and provides functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though, that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

=head1 SYNOPSIS

This module is intended to be used only through the FlexCache::Cache object, as follows:

	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Cache;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
	$origin->{volume}->call_any_method_but_the_constructor();
        
        my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
                                          aggregate => {name => $AGGR_DST,
                                                        block_type => $AGGR_DST_BLOCK_TYPE,},
                                          volume => {name => $VOL_DST,},
                                          vserver => {name => $VSERVER,
                                                      dlif_name => $DLIF_DST,
                                                      dlif_id => $DLIF_DST_ID,},
                                          client => {name => $CLIENT_DST,},
                                          origin => $origin);
        $cache->{volume}->call_any_method_but_the_constructor();
	
Using this module independently is possible, with the risk that the constructor's arguments may change at any time.

=cut

package FlexCache::CacheVolume;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

# NetApp libs
use base qw(FlexCache::Volume);
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use NACL::STask::Volume;
use NACL::STask::Node;

=head1 CONSTRUCTOR

=head2 new

Constructor for the FlexCache::CacheVolume class.
It creates or reuses a flexcache cache volume with the required attributes.
This class is inheriting from FlexCache::Volume.

=head3 ARGUMENTS 

=over

=item B<name> I<Optional> Scalar|Undef:

        Name of the volume. 
        Default value: <name_of_origin>_cache
        (e.g. If the origin name is 'ovol', the cache name will be 'ovol_cache').

=item B<size> I<Optional> Scalar:

        Size of the volume.
        Default value: FCACHE_DFLT_VOL_SIZE (defined in FlexCache constants).

=item B<recycle> I<Optional> Boolean:

        Flag to determine if this volume should be reused.
        If this flag is set to TRUE, then:
        - at setup, the volume will be either created or reused (if it already exists)
        - at cleanup it will not be destroyed.
        If the flag is set to FALSE, then:
        - at setup, the volume will be created or purged+recreated (if it already exists)
        - at cleanup it will be purged.
        Default value: constant FCACHE_DFLT_VOL_RECYCLE (defined in FlexCache constants).

=back

=head3 RETURN

FlexCache::Volume object.

=head3 EXAMPLE

	# Create a flexcache cache with the default settings for the volume member:
	my $cache = FlexCache::Cache->new(volume => {name => $VOL_DST,},
	                                  ....);
	# Create a flexcache cache, requiring a size of 200m:
	my $cache = FlexCache::Cache->new(volume => {name => $VOL_SRC,
	                                             size => '200m',},
					   ....);

=cut

# "Somehow internal" arguments (to limit the use of this constructor to the FlexCache library.
#                               These arguments are not documented with POD, to not confuse the script developers):
# - node: Required: Object: NACL::STask::Node object for this node.
# - aggregate_name: Required: Scalar: Name of the aggregate that will contain the volume.
# - vserver_name: Required: Scalar: Name of the vserver that will export the volume.
# - origin_name: Required: Scalar: Name of the origin volume for this cache.
# - origin_junction_path: Required: Scalar: Junction path of the origin.
# - origin_7mode_ip: Required: Scalar: IP of the hosting filer, used for 7-mode setup (ignored for C-mode).
# - client_type: Optional: Scalar: Type of client for which this volume should be ready-to-use.
#                Legal values:
#                  FCACHE_CLIENT_TYPE_UNIX (volume is created/reused
#                    and the client-side junction path of the object is mapped to the origin NFS junction-path),
#                  FCACHE_CLIENT_TYPE_WINDOWS (volume is created/reused, a cifs share is created/reused,
#                    and the client-side junction path of the object is mapped to the CIFS share name of the origin volume).
#                Default value: FCACHE_CLIENT_TYPE_UNIX.
# - debug: Optional: Boolean: Flag to increase the verbosity of this module, useful for debugging purposes.
#   Default value: FALSE.
# - perform_cleanup: Optional: Boolean: Flag to control if the cleanup should be executed.
#   Default value: TRUE.
# - ui_messages: Required: Object: FlexCache::UI::OntapMessages, to be used for any ONTAP output validation.
# - attach_only: Optional: Boolean: Flag to determine if the constructor should only attach to the required volume,
#                and not attempt to create/recreate. If this flag is set to TRUE, then the recycle flag is ignored.
#                Default value: FALSE.
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name => {type => SCALAR | UNDEF, default => FCACHE_VOL_CACHE,},
            node => {type => OBJECT,         isa     => "NACL::STask::Node"},
            aggregate_name       => {type => SCALAR,},
            vserver_name         => {type => SCALAR,},
            origin_name          => {type => SCALAR,},
            origin_junction_path => {type => SCALAR,},
            size => {type => SCALAR, default => FCACHE_DFLT_VOL_SIZE,},
            client_type =>
                {type => SCALAR, default => FCACHE_CLIENT_TYPE_UNIX,},
            origin_7mode_ip => {type => SCALAR},
            recycle => {type => BOOLEAN, default => FCACHE_DFLT_VOL_RECYCLE,},
            debug           => {type => BOOLEAN, default => FALSE,},
            perform_cleanup => {type => BOOLEAN, default => TRUE,},
            ui_messages =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
            attach_only => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $self = {};

    # WARNING!!!!!
    # This class is inheriting from FlexCache::Volume.
    # For this to work correctly,
    # make sure that all the class members of the parent class
    # are set correctly here as well
    $self->{_creator_pid}          = $$;
    $self->{_is_purged}            = FALSE;
    $self->{_ui_messages}          = $args{ui_messages};
    $self->{_name}                 = $args{name};
    $self->{_vserver_name}         = $args{vserver_name};
    $self->{_nacl_node}            = $args{node};
    $self->{_boot_mode}            = $self->{_nacl_node}->mode();
    $self->{_client_type}          = $args{client_type};
    $self->{_attach_only}          = $args{attach_only};
    $self->{_recycle}              = $args{recycle};
    $self->{_debug}                = $args{debug};
    $self->{_perform_cleanup}      = $args{perform_cleanup};
    $self->{_origin_name}          = $args{origin_name};
    $self->{_origin_junction_path} = $args{origin_junction_path};

    if (   ($self->{_client_type} ne FCACHE_CLIENT_TYPE_UNIX)
        && ($self->{_client_type} ne FCACHE_CLIENT_TYPE_WINDOWS)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Unknown client-type-specific configuration required for:\n"
                . "Cache volume: "
                . $self->{_name} . "\n"
                . "Client type: "
                . $self->{_client_type}
        );
    }
    my $share_reuse_action = 'die';

    if (!defined($self->{_name})) {
        $self->{_name} = $self->{_origin_name} . '_cache';
    }

    my $was_created = FALSE;
    if ($self->{_attach_only}) {
        fcache_log_comment(
            message => "Attempting to attach to volume " . $self->{_name})
            if $self->{_debug};
        my $node_name = $self->{_nacl_node}->name();
        $self->{_nacl_vol} = NACL::STask::Volume->find(
            command_interface => $self->{_nacl_node},
            filter            => {
                volume    => $self->{_name},
                aggregate => $args{aggregate_name},
                vserver   => $args{vserver_name},
                type      => FCACHE_VOL_TYPE_DC,
            }
        );

        # Declare the object recycled,
        # so we do not destroy the volume at cleanup
        $self->{_recycle} = TRUE;
        $share_reuse_action = 'attach';
    } else {
        my $nacltask_if_exists = ($self->{_recycle}) ? 'reuse' : 'purge';
        $share_reuse_action = $nacltask_if_exists;
        $self->{_nacl_vol} = NACL::STask::Volume->create(
            command_interface         => $self->{_nacl_node},
            volume                    => $self->{_name},
            aggregate                 => $args{aggregate_name},
            vserver                   => $args{vserver_name},
            size                      => $args{size},
            type                      => FCACHE_VOL_TYPE_DC,
            'flexcache-origin-volume' => $self->{_origin_name},
            nacltask_if_exists        => $nacltask_if_exists,
            'remote-location'         => $args{origin_7mode_ip} . ':'
                . $self->{_origin_name},
            _was_created => \$was_created,
        );

        # NACL is relying on the "already exists" error message
        # for the create command. In case of reuse, we need to ensure that
        # the already-existent volume has the attributes that we care about
        if (!$was_created) {
            my $recycled_vol_state = $self->{_nacl_vol}->state();

            # Verify the hosting aggregate
            my $aggregate = $recycled_vol_state->aggregate();
            if ($aggregate ne $args{aggregate_name}) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_CONFIG,
                    message     => 'Reusing volume '
                        . $self->{_name} . "\n"
                        . "Volume's current aggregate is $aggregate. Script requires aggregate: "
                        . $args{aggregate_name}
                );
            }

            my $boot_mode = $self->{_nacl_node}->mode();
            if ($boot_mode eq FCACHE_ONTAP_CMODE) {
                my $vserver = $recycled_vol_state->vserver();
                if ($vserver ne $args{vserver_name}) {
                    fcache_log_result(
                        result_type => FCACHE_LOGRESULT_CONFIG,
                        message     => 'Reusing volume '
                            . $self->{_name} . "\n"
                            . "Volume's current vserver is $vserver. Script requires vserver: "
                            . $args{vserver_name}
                    );
                }
            }
        }
    }

    bless($self, $class);

    if ($was_created) {
        $self->_exportfs_if_7mode(options => FCACHE_DFLT_EXPORTFS_CACHE);
    }

    fcache_log_comment(
        message => "Completed creation of new CacheVolume object for "
            . $self->{_name})
        if $self->{_debug};
    return $self;

} ## end sub new

=head1 OTHER METHODS

This class inherits from FlexCache::Volume.
All the methods documented there are also available for a cache volume.

=head2 origin_has_been_renamed

Method to be used after the origin volume has been renamed. It updates all the
internally-needed information pertaining to the origin volume.

=head3 ARGUMENTS

=over

=item B<new_name> I<Required> Scalar:

        New name of the origin volume.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->origin_has_been_renamed(new_name => $origin_new_name);

=cut

sub origin_has_been_renamed {
    my $class = shift;
    my %args = validate(@_, {new_name => {type => SCALAR,},});

    $class->{_origin_name} = $args{new_name};

    return FCACHE_RC_SUCCESS;
} ## end sub origin_has_been_renamed

=head2 flexcache_eject

Eject contents of this cache volume.

=head3 ARGUMENTS

=over

=item B<force> I<Optional> Boolean:

        Forcefully eject the cache contents.
	Default value: TRUE.

=item B<snapshots> I<Optional> Boolean:

        Eject the cached snapshots directory.
	Default value: FALSE.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->flexcache_eject();
	$obj->flexcache_eject(snapshots => TRUE);

=cut

sub flexcache_eject {
    my $class = shift;
    my %args  = validate(
        @_,
        {   force     => {type => BOOLEAN, default => TRUE,},
            snapshots => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $path = '/vol/' . $class->{_name};
    if ($args{snapshots} == TRUE) {
        $path .= '/.snapshot';
    }
    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
    $api_set->flexcache_eject(
        path  => $path,
        force => $args{force},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_eject

# Overwrite parent's method
# In the cache case, return the origin's junction path
sub get_junction_path {
    my $class = shift;
    my %args = validate(@_, {});

    my $boot_mode = $class->{_nacl_node}->mode();
    if ($boot_mode eq FCACHE_ONTAP_CMODE) {
        return $class->{_origin_junction_path};
    } else {
        my $vol_state = $class->{_nacl_vol}->state();
        return $vol_state->junction_path();
    }

} ## end sub get_junction_path

# Overwrite parent's method
# In the cache case:
# - if share_name is passed in, then set this to be the object-associated share
# - in both cases, check that this object has an associated share.
# Arguments:
# - share_name: Optional: Scalar: The name of the share to be associated to this object.
# Return: Name of the object-associated share (Scalar).
sub _ensure_volume_has_share {
    my $class = shift;
    my %args = validate(@_, {share_name => {type => SCALAR, optional => 1,}});

    if (defined($args{share_name})) {
        $class->{__share_name} = $args{share_name};
    }

    if (!defined($class->{__share_name})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_CONFIG,
            message     => 'Cannot configure the cache volume for CIFS. '
                . 'No origin share has been detected'
        );
    }

    return $class->{__share_name};
} ## end sub _ensure_volume_has_share

1;

