#
# $Id: //depot/prod/test/main/lib/FlexCache/ClientOperations.pm#2 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Agnostic module for client-side functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::ClientOperations

=head1 DESCRIPTION

This module provides client-side functionality for all the flexcache scripts (for both nfs and cifs).
This includes mounting/umounting a test volume and performing file operations on this volume.
Its functionality is made available through the 'client' member of the FlexCache::Origin object.

=head1 SYNOPSIS

This module is intended to be used only through the FlexCache::Origin objects,
and/or through the FlexCache::Cache objects, as follows:

	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Cache;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
	$origin->{client}->call_any_method_but_the_constructor();
        
        my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
                                          aggregate => {name => $AGGR_DST,
                                                        block_type => $AGGR_DST_BLOCK_TYPE,},
                                          volume => {name => $VOL_DST,},
                                          vserver => {name => $VSERVER,
                                                      dlif_name => $DLIF_DST,
                                                      dlif_id => $DLIF_DST_ID,},
                                          client => {name => $CLIENT_DST,},
                                          origin => $origin);
        $cache->{client}->call_any_method_but_the_constructor();
	
Using this module independently is possible, with the risk that the constructor's arguments may change at any time.

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl
	flexcache/library_unit_tests/Dataset.thpl

=cut

package FlexCache::ClientOperations;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;
use POSIX qw(floor);

# NetApp libs
use Tharn;
use NACL::OptionSet;
use NATE::BaseException qw(:try);
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::UI::ClientMessages;
use FlexCache::Utils qw(fcache_log_comment
    fcache_log_result
    fcache_verify_expected_rc
    fcache_compare_hashes);
use NACL::C::Client;
use NACL::STask::Client::Mount;
use Protocol::ClientFileOperations;
use RichData;

=head1 CONSTRUCTOR

=head2 new

Constructor for the FlexCache::ClientOperations class.

=head3 ARGUMENTS

=over

=item B<name> I<Required> Scalar:

        Name of the client.

=item B<auto_mount> I<Optional> Boolean:

        Flag to control if the volume/share should be mounted by the constructor.
        Default value: FALSE.

=item B<nfs_mount_version> I<Optional> Scalafr:

        NFS version to use when mounting the volume.
        This argument is ignored for CIFS.
        Legal values: constants FCACHE_CLIENT_PROTOCOL_NFSv3 and FCACHE_CLIENT_PROTOCOL_NFSv4.
        Default value: FCACHE_CLIENT_PROTOCOL_NFSv3.

=item B<required_type> I<Optional> Scalar:

        Required type for this client.
        If this flag is passed in and the client is not of the required type,
        then execution is terminated, with CONFIG result.
        Legal values: any of FCACHE_CLIENT_TYPE_* (defined in FlexCache::Constants).

=item B<require_delegations_compliance> I<Optional> Boolean:

        Flag to require that this client is delegations-compliant
	(it asks for delegations from the filer).
	Default value: FALSE.
	If this flag is set to FALSE, then no check is done and script execution
	continues unconditionally.
	If this flag is set to TRUE, then the nate-host file of the client is
	looked-up for the following line:
	FCACHE_DELEGATIONS=<0 or 1>
	  - If the nate-host information is not found, script execution continues,
	  but the log will contain a warning (as testing failures are possible
	  because of an not-compliant client).
	  - If the nate-host file specifies that the client does not ask
	  for delegations (FCACHE_DELEGATIONS=0), then the script execution is
	  terminated, with a "config" result.
	
=item B<required_protocols> I<Optional> Arrayref:

        Argument to require that this client is supports one or more protocols.
	If this argument is passed in, the nate-host file of the client is
	looked-up for the following line:
	FCACHE_PROTOCOLS=<comma-separated list of supported protocols>
	where: the supported protocols must be the VALUES of any of the
	FCACHE_CLIENT_PROTOCOL_* constants('NFS_3', 'NFS_4', 'SMB_1', 'SMB_2.0', 'SMB_2.1').
	  E.g. FCACHE_PROTOCOLS='NFS_3,NFS_4'
	       FCACHE_PROTOCOLS='SMB_2.0'
	  - If the nate-host information is not found, script execution continues,
	  but the log will contain a warning (as testing failures are possible
	  because of an not-compliant client).
	  - If the nate-host information is found, but not all of the required protocols
	  are listed as supported, then the script execution is
	  terminated, with a "config" result.

=back

=head3 RETURN

        FlexCache::ClientOperations object.

=head3 EXAMPLES

        # Create a flexcache origin with the default settings for the client member:
	my $origin = FlexCache::Origin->new(client => {name => $CLIENT_SRC,},
                                            ....);
	# Create a flexcache origin, requiring to not mount the volume:
	my $origin = FlexCache::Origin->new(client => {name => $CLIENT_SRC,
                                                       auto_mount => FALSE,},
                                            ....);
	# Create a flexcache origin, requiring a windows client
	my $origin = FlexCache::Origin->new(client => {name => $CLIENT_SRC,
                                                       required_type => FCACHE_CLIENT_TYPE_WINDOWS,},
                                            ....);
	# Create a flexcache origin, requiring that the client asks for delegations
	my $origin = FlexCache::Origin->new(client => {name => $CLIENT_SRC,
                                                       require_delegations_compliance => TRUE,},
                                            ....);
	# Create a flexcache origin, requiring that the client supports SMB 2.0
	my $origin = FlexCache::Origin->new(client => {name => $CLIENT_SRC,
                                                       required_protocols => [FCACHE_CLIENT_PROTOCOL_SMBv2_0],},
                                            ....);
					    # Create a flexcache origin, requiring that the client supports NFS v3 and v4
	my $origin = FlexCache::Origin->new(client => {name => $CLIENT_SRC,
                                                       required_protocols => [FCACHE_CLIENT_PROTOCOL_NFSv3,
						                              FCACHE_CLIENT_PROTOCOL_NFSv4],},
                                            ....);
	# All of the above examples apply to the FlexCache::Cache constructor as well.

=cut

# "Somehow internal" arguments (to limit the use of this constructor to the FlexCache library.
#                               These arguments are not documented with POD, to not confuse the script developers):
# - mount_from_ip: Required: Scalar: IP of the node/filer that hosts the volume/share to mount/map from.
# - mount_from_junction: Required: Scalar: Name of the junction (volume/share) to mount from.
# - debug: Optional: Boolean: Flag to increase the verbosity of this module, useful for debugging purposes.
#   Default value: FALSE.
# - perform_cleanup: Optional: Boolean: Flag to control if the cleanup should be executed.
#   Default value: TRUE.
# Internal arguments (should only be used by the FlexCache library):
# - _postpone_setup: Boolean: Flag to be used to postpone the setup part provided by the constructor.
#                    In this case, any passed-in arguments are saved internally, to be used at the time when
#                    _execute_setup is called. Warning:  this flag should be only used from a higher-level constructor.
#                    _execute_setup NEEDS to be called before this higher-level constructor exits.
#                    The flag is useful in the FlexCache::Origin constructor because of all the dependencies:
#                    we first connect to the client to determine its type, then we need to create and configure the vserver accordingly
#                    (nfs or cifs), to be able to finally mount/map.
#                    Using this flag will make the mount_from_ip and mount_from_junction arguments optional. They will need to be passed in
#                    later on, to the _execute_setup method.
#
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name => {type => SCALAR,},
            auto_mount =>
                {type => BOOLEAN, default => FCACHE_DFLT_AUTO_MOUNT},
            nfs_mount_version =>
                {type => SCALAR, default => FCACHE_CLIENT_PROTOCOL_NFSv3,},
            mount_from_ip       => {type => SCALAR, optional => 1,},
            mount_from_junction => {type => SCALAR, optional => 1,},
            required_type       => {type => SCALAR, optional => 1,},
            require_delegations_compliance =>
                {type => BOOLEAN, default => FALSE,},
            required_protocols => {type => ARRAYREF, optional => 1,},
            debug              => {type => BOOLEAN,  default  => FALSE,},
            perform_cleanup    => {type => BOOLEAN,  default  => TRUE,},
            _postpone_setup    => {type => BOOLEAN,  default  => FALSE},
        }
    );

    my $self = {};
    $self->{_creator_pid}     = $$;
    $self->{_perform_cleanup} = $args{perform_cleanup};

    if ((!$args{_postpone_setup})
        && (   (!defined($args{mount_from_ip}))
            || (!defined($args{mount_from_junction})))
        ) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Arguments \"mount_from_ip\" and \"mount_from_junction\""
                . " must be defined even if \"auto_mount\" is TRUE.\n"
                . "Constructor was called with the following arguments:\n"
                . Dumper(%args)
        );
    }

    $self->{_name}              = $args{name};
    $self->{_auto_mount}        = $args{auto_mount};
    $self->{_nfs_mount_version} = $args{nfs_mount_version};
    $self->{_debug}             = $args{debug};
    $self->{_opened_files}      = {};

    fcache_log_comment(message =>
            "SETUP: Starting to construct the ClientOperations object from:\n"
            . Dumper(%args))
        if $self->{_debug};
    $self->{_nacl_client} = NACL::C::Client->new(client => $self->{_name});
    $self->{_file_operations} = Protocol::ClientFileOperations->new(
        client => $self->{_nacl_client});
    $self->{_type}                = $self->{_file_operations}->{client_type};
    $self->{_agnostic_rpc_module} = 'FlexCache::ClientRPC::Agnostic';
    $self->{_platform_rpc_module}
        = ($self->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS)
        ? 'FlexCache::ClientRPC::Windows'
        : 'FlexCache::ClientRPC::Linux';

    if (   (defined($args{required_type}))
        && ($args{required_type} ne $self->{_type})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_CONFIG,
            message     => 'Current client type is '
                . $self->{_type}
                . '. Script requires client of type: '
                . $args{required_type}
        );
    }

    if ($args{require_delegations_compliance}) {
        my $is_delegations_compliant
            = $self->{_nacl_client}->hostrec()->{FCACHE_DELEGATIONS};
        if (!defined($is_delegations_compliant)) {
            fcache_log_comment(message => "WARNING! WARNING! WARNING!\n"
                    . "The client nate-host file is missing"
                    . " delegations information.\n"
                    . "Proceeding with the script execution anyway.\n"
                    . "This may result in a failure if this client is"
                    . " not asking for delegations:\n"
                    . "Client name: "
                    . $self->{_name} . "\n"
                    . "Delegations info should be defined as: "
                    . "FCACHE_DELEGATIONS=0/1");
        } elsif (!($is_delegations_compliant)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_CONFIG,
                message     => "This client is not asking for delegations.\n"
                    . "Script requires delegations-compliant client.\n"
                    . "Client name: "
                    . $self->{_name}
            );
        }
    }

    if (defined($args{required_protocols})) {
        my $required_protocols = $args{required_protocols};
        my $supported_protocols
            = $self->{_nacl_client}->hostrec()->{FCACHE_PROTOCOLS};
        if (!defined($supported_protocols)) {
            fcache_log_comment(message => "WARNING! WARNING! WARNING!\n"
                    . "The client nate-host file is missing protocol"
                    . " information.\n"
                    . "Proceeding with the script execution anyway.\n"
                    . "This may result in a failure if the client"
                    . " does not meet the requirements:\n"
                    . "Client name: "
                    . $self->{_name} . "\n"
                    . "Delegations info should be defined as:"
                    . " FCACHE_PROTOCOLS=<supported_protocols>");
        } else {
            my @client_protocols = split(/,/, $supported_protocols);
            foreach my $required_protocol (@{$required_protocols}) {
                if (!grep(/^$required_protocol/, @client_protocols)) {
                    fcache_log_result(
                        result_type => FCACHE_LOGRESULT_CONFIG,
                        message =>
                            "This client is not supporting a required protocol:\n"
                            . "Protocol: $required_protocol\n"
                            . "Client name: "
                            . $self->{_name} . "\n"
                            . "Supported protocols: $supported_protocols"
                    );
                }
            }
        }
    }

  # Protocol::ClientFileOperations categorizes all the *nix clients as 'linux'
  # Log the actual host type, for a more detailed tracking
    my $host_type = $self->{_nacl_client}->hostrec()->{hosttype};
    fcache_log_comment(
        message => "SETUP: Using $host_type client: " . $self->{_name});

    $self->{_ui_messages} = FlexCache::UI::ClientMessages->new(
        client_type => $self->{_type},
        debug       => $self->{_debug}
    );

    bless($self, $class);

    if (!($args{_postpone_setup})) {
        $self->_execute_setup(
            mount_from_ip       => $args{mount_from_ip},
            mount_from_junction => $args{mount_from_junction},
        );
    }

    return $self;

} ## end sub new

=head1 OTHER METHODS

=head2 mount

Mount/map the volume/share on this client (depending on the type of client).

=head3 ARGUMENTS

=over

=item B<chmod> I<Optional> Scalar:

        Permissions to set on the NFS mount-point.
        This argument is ignored for CIFS.
        Default value: constant FCACHE_DFLT_CHMOD (defined in FlexCache::Constants).


=item B<version> I<Optional> Scalar:

        NFS version to use for mounting.
        This argument is ignored for CIFS.
        Legal values: constants FCACHE_CLIENT_PROTOCOL_NFSv3 and FCACHE_CLIENT_PROTOCOL_NFSv4.
        Default value: the constructor-defined nfs mount version.
        
=item B<domain> I<Optional> Scalar:

        CIFS domain to be used when mapping the share.
        This argument is ignored for NFS.
        Default value: The CIFS_DOMAIN command-line argument value.
        
=item B<user> I<Optional> Scalar:

        User name to be used when mapping the share.
        This argument is ignored for NFS.
        Default value: The AD_SERVER_USER command-line argument value.

=item B<password> I<Optional> Scalar:

        Password to be used when mapping the share.
        This argument is ignored for NFS.
        Default value: The AD_SERVER_PASSWORD command-line argument value.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

=head3 EXAMPLE
	
	# Mount/map with the default settings (NFS/CIFS):
	$obj->mount();
	# Mount with NFSv4:
	$obj->mount(version => FCACHE_CLIENT_PROTOCOL_NFSv4);

=cut

sub mount {
    my $class = shift;
    my %args  = validate(
        @_,
        {   chmod => {type => SCALAR | UNDEF, default => FCACHE_DFLT_CHMOD,},
            version =>
                {type => SCALAR, default => $class->{_nfs_mount_version},},
            domain   => {type => SCALAR, optional => 1,},
            user     => {type => SCALAR, optional => 1,},
            password => {type => SCALAR, optional => 1,},
        }
    );

    if (defined($class->{_mount_point})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                'Volume/share was previously mounted/mapped on this client: '
                . $class->{_name}
        );
    }

    my %mount_args = (
        command_interface => $class->{_nacl_client},
        server            => $class->{_mount_src_ip},
    );

    if ($class->{_type} eq FCACHE_CLIENT_TYPE_UNIX) {
        fcache_log_comment(
            message => "Performing NFS-mount on client " . $class->{_name})
            if $class->{_debug};
        $mount_args{path} = $class->{_mount_src_junction};
        if (defined($args{chmod})) {
            $mount_args{chmod} = $args{chmod};
        }
        my $nfs_version = $args{version};
        if ($nfs_version eq FCACHE_CLIENT_PROTOCOL_NFSv3) {
            $mount_args{vers} = 3;
        } elsif ($nfs_version eq FCACHE_CLIENT_PROTOCOL_NFSv4) {
            $mount_args{vers} = 4;
        } else {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Unsupported NFS type: $nfs_version"
            );
        }
    } elsif ($class->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        my $options = NACL::OptionSet->new();
        my $user
            = (defined($args{user}))
            ? ($args{user})
            : ($options->get('AD_SERVER_USER'));
        my $password
            = (defined($args{password}))
            ? ($args{password})
            : ($options->get('AD_SERVER_PASSWORD'));
        my $domain
            = (defined($args{domain}))
            ? ($args{domain})
            : ($options->get('CIFS_DOMAIN'));

        if (   (!defined($user))
            || (!defined($password))
            || (!defined($domain))) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message =>
                    "For mapping a CIFS share, please provide either:\n"
                    . "the method parameters: user, password, domain,\n"
                    . "or the command-line arguments:"
                    . " AD_SERVER_USER, AD_SERVER_PASSWORD, CIFS_DOMAIN"
            );
        }

        $mount_args{share}    = $class->{_mount_src_junction};
        $mount_args{user}     = "$domain\\$user";
        $mount_args{password} = $password;
    } else {

# This is just in case Protocol::ClientFileOperations changes the type strings
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_CONFIG,
            message =>
                'Cannot mount/map the volume/share: unknown client type: '
                . $class->{_type} . "\n"
                . 'Supported client types: '
                . FCACHE_CLIENT_TYPE_UNIX . ', '
                . FCACHE_CLIENT_TYPE_WINDOWS
        );
    }

    my $mount = NACL::STask::Client::Mount->create(%mount_args);
    $class->{_mount_point} = $mount->mount_point();
    fcache_log_comment(message => "Mount point is " . $class->{_mount_point})
        if $class->{_debug};

    return FCACHE_RC_SUCCESS;
} ## end sub mount

=head2 umount

Umount/unmap the volume/share on this client (depending on the type of client).

=head3 ARGUMENTS

None.

=head3 RETURN

        FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->umount();

=cut

sub umount {
    my $class = shift;
    my %args = validate(@_, {});

    if (!defined($class->{_mount_point})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                'Volume/share was not previously mounted/mapped on this client: '
                . $class->{_name}
        );
    }

    NACL::STask::Client::Mount->purge(
        command_interface => $class->{_nacl_client},
        mount_point       => $class->{_mount_point}
    );
    delete($class->{_mount_point});

    return FCACHE_RC_SUCCESS;
} ## end sub umount

=head2 get_mount_point

Obtain the mount-point of the volume/share on this client.

=head3 ARGUMENTS

None.

=head3 RETURN

        The mount-point of the volume/share on this client (Scalar).
        If the volume/share is not currently mounted, the method returns undef.

=head3 EXAMPLE

	my $mount_point = $obj->get_mount_point();

=cut

sub get_mount_point {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_mount_point};
} ## end sub get_mount_point

=head2 get_name

Obtain the name of this client.

=head3 ARGUMENTS

None.

=head3 RETURN

        The hostname of this client (Scalar).

=head3 EXAMPLE

	my $client_name = $obj->get_name();

=cut

sub get_name {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_name};
} ## end sub get_name

=head2 get_type

Get the OS type of this client.

=head3 ARGUMENTS

None.

=head3 RETURN

        Client  type (Scalar).
	Possible values: constants
	FCACHE_CLIENT_TYPE_UNIX or FCACHE_CLIENT_TYPE_WINDOWS.

=head3 EXAMPLE

	my $client_type = $obj->get_type();

=cut

sub get_type {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_type};
} ## end sub get_type

=head2 set_mount_junction

Set the mount junction for this object.
This method is useful for objects that were initially created for DP volumes,
but which became RW after the snapmirror relationship was broken.
The mount junction path is needed for future mounts.

=head3 ARGUMENTS

=over

=item B<junction> I<Required> Scalar:

        New mount junction path to set for this object.

=back

=head3 RETURN

        Client  type (Scalar).
	Possible values: constants
	FCACHE_CLIENT_TYPE_UNIX or FCACHE_CLIENT_TYPE_WINDOWS.

=head3 EXAMPLE

	$obj->set_mount_junction(junction => $mount_junction);

=cut

sub set_mount_junction {
    my $class = shift;
    my %args = validate(@_, {junction => {type => SCALAR,},});

    $class->{_mount_src_junction} = $args{junction};

    return FCACHE_RC_SUCCESS;
} ## end sub set_mount_junction

=head2 mount_point_list

List the contents of the mountpoint from this client.

=head3 ARGUMENTS

=over

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=item B<snapshots> I<Optional> Boolean: 

        If this flag is set to TRUE, then the snapshots directory will be listed.
        Default value: FALSE.

=back

=head3 RETURN

        Array of content names.

=head3 EXAMPLE

	# List the mountpoint:
	my $contents = $obj->mount_point_list();
	# List the snapshots directory:
	my $contents = $obj->mount_point_list(snapshots => TRUE);

=cut

sub mount_point_list {
    my $class = shift;
    my %args  = validate(
        @_,
        {   expected  => {type => SCALAR,  default => FCACHE_RC_SUCCESS,},
            snapshots => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $path          = $class->get_mount_point();
    my $snapshots_dir = $args{snapshots};
    if ($snapshots_dir) {
        $path .= '/.snapshot';
    }

    return $class->directory_list(
        directory_name => $path,
        expected       => $args{expected},
    );
} ## end sub mount_point_list

=head2 file_exists

Check if a file exists on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to check.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<snapshot_name> I<Optional> Scalar:

        Name of snapshot in which to look for the file.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Legal values: FCACHE_RC_SUCCESS, FCACHE_RC_NO_ENTRY_FOUND.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	# Check that my_file exists
        $obj->file_exists(file_name => 'my_file');
        # Check that my_file does not exist
        $obj->file_exists(file_name => 'my_file',
                          expected => FCACHE_RC_NO_ENTRY_FOUND);

=cut

sub file_exists {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name     => {type => SCALAR,},
            snapshot_name => {type => SCALAR, optional => 1,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name = $class->_ensure_absolute_path(
        name          => $args{file_name},
        snapshot_name => $args{snapshot_name},
    );
    my $expected_rc = $args{expected};

    my $file_exists
        = $class->{_file_operations}->exists(file_name => $file_name);
    my $rc = ($file_exists) ? FCACHE_RC_SUCCESS : FCACHE_RC_NO_ENTRY_FOUND;

    fcache_log_comment(message => "Checked if file exists:\n"
            . "File: $file_name\n"
            . "File exists: $file_exists");

    return fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $expected_rc,
        action      => "Verify if file exists: $file_name",
        output      => "File exists: $file_exists",
    );
} ## end sub file_exists

=head2 file_open

Open a file on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to open.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1
	
=item B<snapshot_name> I<Optional> Scalar:

        Name of snapshot from which to open the file.
        
=item B<open_mode> I<Required> Scalar:

        How should the file be opened.
        Legal values: any of the FCACHE_FILE_OPEN_MODE_* constants
        (defined in FlexCache::Constants).
        
=item B<access_mode> I<Optional> Scalar:

        Access mode with which the file should be opened.
        Legal values: any of the FCACHE_FILE_ACCESS_MODE_* constants
        (defined in FlexCache::Constants).
        Default value: FCACHE_FILE_ACCESS_MODE_RW.
        
=item B<share_mode> I<Optional> Scalar:

        Share mode with which the file should be opened. Ignored for Unix
        Legal values: any of the FCACHE_FILE_SHARE_MODE_* constants
        (defined in FlexCache::Constants).
        Default value: FCACHE_FILE_SHARE_MODE_RWD.
        
=item B<permissions> I<Optional> Scalar:

        The permissions (4 bits) for the file. Ignored for Windows.
        
=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        Instance number under which this particular file has been opened by this object (Scalar).
        This is usually not needed by the script unless it needs to work
        with multiple instances of the file being opened at the same time.


EXAMPLE

        # Open a single instance of the file.
        # In this case, no need to remember the instance number.
        # The library will take care of it.
	$obj->file_open(file_name => $file1);
        $obj->file_open(file_name => $file2,
                        access_mode => FCACHE_FILE_ACCESS_MODE_R);
                        
        # Open multiple instances of the file.
        # In this case, need to remember the instance number
        # to be able to execute specific actions on a particular opened instance.
        my $instance1 = $obj->file_open(file_name => 'file'); # $instance1 will be 1
        my $instance2 = $obj->file_open(file_name => 'file'); # $instance2 will be 2

=cut

sub file_open {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name     => {type => SCALAR,},
            snapshot_name => {type => SCALAR, optional => 1,},
            open_mode     => {type => SCALAR,},
            access_mode =>
                {type => SCALAR, default => FCACHE_FILE_ACCESS_MODE_RW,},
            share_mode =>
                {type => SCALAR, default => FCACHE_FILE_SHARE_MODE_RWD,},
            permissions => {type => SCALAR, optional => 1,},
            expected    => {type => SCALAR, default  => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name     = delete($args{file_name});
    my $snapshot_name = delete($args{snapshot_name});
    my $expected_rc   = delete($args{expected});
    my $output        = '';
    $file_name = $class->_ensure_absolute_path(
        name          => $file_name,
        snapshot_name => $snapshot_name
    );

    $args{file_name} = $file_name;

    my $file_instance
        = $class->_get_file_largest_instance_number(name => $file_name);
    $file_instance++;

    fcache_log_comment(message => "Opening file:\n"
            . "Path: $file_name\n"
            . "Client: "
            . $class->{_name})
        if $class->{_debug};

    try {
        $class->{_opened_files}->{$file_name}->{$file_instance}
            = $class->{_file_operations}->open(%args);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
        delete($class->{_opened_files}->{$file_name}->{$file_instance});
        $file_instance = 0;
    };

    my $rc = $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_OPEN,
        output      => $output,
        expected_rc => FCACHE_RC_FAILURE_NON_FATAL,
    );

    if ($rc != $expected_rc) {
        if ($rc == FCACHE_RC_SUCCESS) {

            # For the main test, we would be closing all the files anyway,
            # as part of cleanup.  Still, have this here so we can clean up
            # any "unwanted" files opened by subtests, since in the
            # subtests we will not be making use of the destructors.
            $class->file_close(
                file_name       => $file_name,
                instance_number => $file_instance
            );
        }
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Failure opening file $file_name\n"
                . "Expected: $expected_rc\n"
                . "Obtained: $rc\n"
                . "Output: $output"
        );
    }

    fcache_log_comment(message => "Executed file open for:\n"
            . "File: $file_name\n"
            . "Obtained result: $expected_rc");

    return $file_instance;
} ## end sub file_open

=head2 file_close

Close a file on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to close.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly
	mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<snapshot_name> I<Optional> Scalar:

        Name of snapshot from which to open the file.
          
=item B<instance_number> I<Optional> Scalar:

        This argument should be used in the case where multiple instances
	of this file have been opened by this object and only a particular
	instance needs to be closed.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then all the opened instances
	of this file will be closed.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

	# Close any opened instances of the file.
	$obj->file_close(file_name => $file1);
        
        # Close specific instances of the file.
        my $instance1 = $obj->file_open(file_name => 'file'); 
        my $instance2 = $obj->file_open(file_name => 'file');
        ...
        $obj->file_close(file_name => 'file',
                         instance_number => $instance1);
        ......
        $obj->file_close(file_name => 'file',
                         instance_number => $instance2); 

=cut

sub file_close {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            snapshot_name   => {type => SCALAR, optional => 1,},
            instance_number => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name = $class->_ensure_absolute_path(
        name          => $args{file_name},
        snapshot_name => $args{snapshot_name}
    );

    my $output = '';
    if (defined($args{instance_number})) {
        my $file_instance = $args{instance_number};
        fcache_log_comment(message => "Closing file:\n"
                . "Path: $file_name\n"
                . "Instance: $file_instance\n"
                . "Client: "
                . $class->{_name})
            if $class->{_debug};
        try {
            $class->{_file_operations}->close(file_handle =>
                    $class->{_opened_files}->{$file_name}->{$file_instance});
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };

        # Delete the internal tracking entry anyway, otherwise we will hit
        # failure during cleanup
        delete($class->{_opened_files}->{$file_name}->{$file_instance});
    } else {
        fcache_log_comment(message => "Closing file (all instances):\n"
                . "Path: $file_name\n"
                . "Client: "
                . $class->{_name})
            if $class->{_debug};
        try {
            $class->{_file_operations}->close(file_name => $file_name);
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };

        # Delete the internal tracking entry anyway, otherwise we will hit
        # failure during cleanup
        delete($class->{_opened_files}->{$file_name});
    }

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_CLOSE,
        output      => $output,
        expected_rc => $args{expected},
    );
} ## end sub file_close

=head2 file_write

Write to a file on this client.
Note: The file needs to be previously opened.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to write to.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1
          
=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple instances of this file
        have been opened by this object.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then it will default to the latest-opened instance.
        
=item B<offset> I<Optional> Scalar:

        Offset at which to start writing.
        Default value: 0.
        
=item B<num_bytes> I<Optional> Scalar:

        The number of bytes to write to the file.
        Either the num_bytes or contents parameter must be specified.
        
=item B<contents> I<Optional> Scalar:

        The contents string to write to the file.
        Either the num_bytes or contents parameter must be specified.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.
        
=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

	# Open a single instance of the file and write.
	$obj->file_open(file_name => $file1);
        $obj->file_write(file_name => $file1,
                         num_bytes => MB);
                        
        # Open multiple instances of the file.
        # Write to a specific instance.
        my $instance1 = $obj->file_open(file_name => 'file'); # $instance1 will be 1
        my $instance2 = $obj->file_open(file_name => 'file'); # $instance2 will be 2
        $obj->file_write(file_name => $file1,
                         num_bytes => MB,
                         instance_number => $instance2);

=cut

sub file_write {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
            offset => {type => SCALAR, regex => qr/^\d+$/, default => 0,},
            num_bytes => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            contents => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default  => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name     = delete($args{file_name});
    my $file_instance = delete($args{instance_number});
    my $expected_rc   = delete($args{expected});
    my $output        = '';
    my $rc            = FCACHE_RC_SUCCESS;
    my $fh            = $class->_get_file_handle(
        name     => $file_name,
        instance => $file_instance
    );

  # TODO: As of 2012/11/02, the Protocol library does not handle large writes.
  # This is because the contents to be written are stored into  variable
    if (   (exists($args{num_bytes}))
        && ($args{num_bytes} > FCACHE_MAX_FILE_SIZE)) {
        my $num_bytes = $args{num_bytes};
        $args{num_bytes} = FCACHE_MAX_FILE_SIZE;
        my $num_chunks = floor($num_bytes / FCACHE_MAX_FILE_SIZE);

        for (my $i = 0; $i < $num_chunks; $i++) {
            $output = '';

            fcache_log_comment(message => "Writing to file:\n"
                    . "Path: $file_name\n"
                    . "Client: "
                    . $class->{_name} . "\n"
                    . "Write parameters:\n"
                    . Dumper(%args));
            try {
                $fh->write(%args);
            }
            catch NATE::BaseException with {
                my $exception = shift;
                $output = $exception->text();
            };

            # For "no more space available", we may not hit the error from the
            # beginning
            $rc = $class->{_ui_messages}->verify_expected_rc(
                command     => FCACHE_CLIENT_CMD_FILE_WRITE,
                output      => $output,
                expected_rc => FCACHE_RC_FAILURE_NON_FATAL,
            );

            if ($rc != FCACHE_RC_SUCCESS) {
                last;
            }

            $args{offset} = $args{offset} + FCACHE_MAX_FILE_SIZE;
        }

        fcache_verify_expected_rc(
            obtained_rc => $rc,
            expected_rc => $expected_rc,
            action      => "Write to file $file_name",
            output      => $output,
        );

        $args{num_bytes} = $num_bytes % FCACHE_MAX_FILE_SIZE;
        if (($args{num_bytes} == 0) || ($rc != FCACHE_RC_SUCCESS)) {
            return FCACHE_RC_SUCCESS;
        }
    }

    fcache_log_comment(message => "Writing to file:\n"
            . "Path: $file_name\n"
            . "Client: "
            . $class->{_name} . "\n"
            . "Write parameters:\n"
            . Dumper(%args));

    try {
        $fh->write(%args);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_WRITE,
        output      => $output,
        expected_rc => $expected_rc,
    );

    return $rc

} ## end sub file_write

=head2 file_read

Read a file from this client.
Note: The file needs to be previously opened.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to delete.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly
	mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<snapshot_name> I<Optional> Scalar:

        Name of snapshot from which to open the file.
	
=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple instances 
        of this file have been opened by this object.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then it will default to
	the latest-opened instance.
        
=item B<offset> I<Optional> Scalar:

        Offset at which to start reading.
        Default value: 0.
        
=item B<num_bytes> I<Required> Scalar:

        The number of bytes to read from the file.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, defined in FlexCache::Constants.
        Execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.
        
=back

=head3 RETURN

        The contents of the filer (Scalar). 

=head3 EXAMPLE

        # Read 100 bytes, from the beginning of the file
	my $contents = $obj->file_read(file_name => 'my_file',
                                       num_bytes => 100);
        # Read 5 bytes, starting at offset 10
	my $contents = $obj->file_read(file_name => 'my_file',
                                       offset => 10,
                                       num_bytes => 5);
	# Read 100 bytes, from the beginning of the file in snapshot 'snap1'
	my $contents = $obj->file_read(file_name => 'my_file',
                                       num_bytes => 100,
				       snapshot_name => 'snap1');

=cut

sub file_read {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            snapshot_name   => {type => SCALAR, optional => 1,},
            instance_number => {type => SCALAR, optional => 1,},
            offset => {type => SCALAR, regex => qr/^\d+$/, default => 0,},
            num_bytes => {type => SCALAR, regex   => qr/^\d+$/,},
            expected  => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name     = delete($args{file_name});
    my $file_instance = delete($args{instance_number});
    my $expected_rc   = delete($args{expected});
    my $snapshot_name = delete($args{snapshot_name});
    my $error_message = '';
    my $content       = '';
    my $fh            = $class->_get_file_handle(
        name          => $file_name,
        snapshot_name => $snapshot_name,
        instance      => $file_instance
    );

    my $message = "Reading from file:\n" . "Path: $file_name\n";
    if (defined($snapshot_name)) {
        $message .= "Snapshot: $snapshot_name\n";
    }
    $message
        .= "Client: "
        . $class->{_name} . "\n"
        . "Read parameters:\n"
        . Dumper(%args);
    fcache_log_comment(message => $message);
    try {
        $content = $fh->read(%args);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $error_message = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_READ,
        output      => $error_message,
        expected_rc => $expected_rc,
    );
    return $content;
} ## end sub file_read

=head2 file_delete

Delete a file from this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to delete.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, defined in FlexCache::Constants.
        Execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.
	
=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	$obj->file_delete(file_name => 'my_file');

=cut

sub file_delete {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name => {type => SCALAR,},
            expected  => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name = $class->_ensure_absolute_path(name => $args{file_name});
    my $expected_rc = $args{expected};
    my $error_message = '';

    try {
        $class->{_file_operations}->delete(file_name => $file_name);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $error_message = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_DELETE,
        output      => $error_message,
        expected_rc => $expected_rc,
    );
} ## end sub file_delete

=head2 file_rename

Rename a file on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to rename.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1
          
=item B<new_name> I<Required> Scalar:

        New name of the file.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file2
          - for file2 in directory dir1 under <vol>: dir1/file3
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file2

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	$obj->file_rename(file_name => 'my_file',
                          new_name => 'my_renamed_file');

=cut

sub file_rename {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name => {type => SCALAR,},
            new_name  => {type => SCALAR,},
        }
    );

    my $file_name = $class->_ensure_absolute_path(name => $args{file_name});
    my $new_file_name
        = $class->_ensure_absolute_path(name => $args{new_name});

    $class->{_file_operations}->rename(
        file_name     => $file_name,
        new_file_name => $new_file_name
    );

    return FCACHE_RC_SUCCESS;
} ## end sub file_rename

=head2 file_truncate

Truncate a file on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to truncate.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1
          
=item B<size> I<Required> Scalar:

        Size, in bytes, to truncate the file to.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	$obj->file_truncate(file_name => 'my_file',
                            size => KB);

=cut

sub file_truncate {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name => {type => SCALAR,},
            size      => {type => SCALAR,},
        }
    );

    my $file_name = $class->_ensure_absolute_path(name => $args{file_name});
    my $size = $args{size};

    if ($size < 0) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Invalid truncate size: $size"
        );
    }

    $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'file_truncate',
        args    => {
            file_name => $file_name,
            size      => $size,
        },
    );

    return FCACHE_RC_SUCCESS;
} ## end sub file_truncate

=head2 file_copy

Make a copy of a file on this client.

=head3 ARGUMENTS

=over

=item B<file_name_src> I<Required> Scalar:

        Name of the file to make a copy of.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<file_name_dst> I<Required> Scalar:

        Name of the new copy file.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

	$obj->file_copy(file_name_src => 'my_file',
                        file_name_dst => 'my_copy');

=cut

sub file_copy {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name_src => {type => SCALAR,},
            file_name_dst => {type => SCALAR,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $src_name
        = $class->_ensure_absolute_path(name => $args{file_name_src});
    my $dst_name
        = $class->_ensure_absolute_path(name => $args{file_name_dst});

    my $output = $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'file_copy',
        args    => {
            file_src => $src_name,
            file_dst => $dst_name,
        },
    );

    fcache_log_comment(message => "Executed file copy:\n"
            . "Source: $src_name\n"
            . "Destination: $dst_name\n"
            . "Output: $output");

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_COPY,
        output      => $output,
        expected_rc => $args{expected},
    );
} ## end sub file_copy

=head2 file_lock

Acquire a byte-range lock on a file.
Note: the file needs to be previously opened.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to lock.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1


=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple instances of this file
        have been opened by this object.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then it will default to the latest-opened instance.

=item B<lock_type> I<Optional> Scalar:

        Type of lock to acquire.
        Legal values: any of the FCACHE_LOCK_TYPE_* constants
        (defined in FlexCache::Constants).
        Default value: FCACHE_LOCK_TYPE_EXCLUSIVE.

=item B<lock_wait> I<Optional> Scalar:

        Type of request for the lock: waiting or non waiting. 
        Legal values: FCACHE_LOCK_NON_WAITING, FCACHE_LOCK_WAITING.
        (defined in FlexCache::Constants).
        Default value: FCACHE_LOCK_NON_WAITING.

=item B<offset> I<Optional> Scalar:

        Offset at which to acquire the lock.
        Default value: 0.
        
=item B<num_bytes> I<Required> Scalar:

        The number of bytes to lock on the file.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        The lock id of the lock request (Scalar).
        This is an integer id number, representing this request.
        For non-waiting requests, it will always be 0.
        For waiting requests, it will be a positive number
        which can be used later on as the lock_id argument to
        file_lock_request_status_verify, to verify the status of the request.

=head3 NOTES

        When acquiring an exclusive byte lock from NFS
        and then request a second exclusive byte lock
        over the same byte range from the same client,
        it is treated as a duplicate request in the eyes of the server.
        Such, the lock is considered granted.
        Different clients should be used to test this scenario.
        All NFS versions support waiting locks.
        
        CIFS waiting locks are only supported from SMB2.0 onwards:
        Windows XP : CIFS (or SMB1) - No waiting locks support
        Windows Vista/2008 : SMB2.0
        Windows 7/2008R2 : SMB2.1
        Windows 8 : SMB3


=head3 EXAMPLE

        # Lock first 10 bytes of the file, using the defaults
        # (non-waiting request, exclusive lock)
	$obj->file_lock(file_name => 'my_file',
                        num_bytes => 10);
        # Lock the second byte of the file, shared lock
	$obj->file_lock(file_name => 'my_file',
                        lock_type => FCACHE_LOCK_TYPE_SHARED,
                        offset => 1,
                        num_bytes => 1);
        my $waiting_lock_id = $origin->{client}->file_lock(file_name => $file1,
                                                           instance_number => $file_instance2,
                                                           lock_wait => FCACHE_LOCK_WAITING,
                                                           num_bytes => 5,);

=cut

sub file_lock {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
            lock_type =>
                {type => SCALAR, default => FCACHE_LOCK_TYPE_EXCLUSIVE,},
            lock_wait =>
                {type => SCALAR, default => FCACHE_LOCK_NON_WAITING,},
            offset    => {type => SCALAR, regex => qr/^\d+$/, default => 0,},
            num_bytes => {type => SCALAR, regex => qr/^\d+$/,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name     = delete($args{file_name});
    my $file_instance = delete($args{instance_number});
    my $expected_rc   = delete($args{expected});
    my $lock_id       = 0;
    my $output        = '';

    my $fh = $class->_get_file_handle(
        name     => $file_name,
        instance => $file_instance
    );

    if ($args{lock_wait} eq FCACHE_LOCK_WAITING) {
        $args{waiting_id} = \$lock_id;
    }
    try {
        $fh->lock(%args);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_LOCK,
        output      => $output,
        expected_rc => $expected_rc,
    );

    return $lock_id;
} ## end sub file_lock

=head2 file_unlock

Release a byte-range lock on a file.
The unlock request is non-waiting.
Note: the file needs to be previously opened.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to unlock.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple instances of this file
        have been opened by this object.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then it will default to the latest-opened instance.

=item B<offset> I<Optional> Scalar:

        Offset at which to release the lock.
        Default value: 0.
        
=item B<num_bytes> I<Required> Scalar:

        The number of bytes to unlock from the file.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $origin->{client}->file_unlock(file_name => $file1,
                                       offset => 1,
                                       num_bytes => 5,);

=cut

sub file_unlock {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
            offset => {type => SCALAR, regex => qr/^\d+$/, default => 0,},
            num_bytes => {type => SCALAR, regex => qr/^\d+$/,},
        }
    );

    my $file_name     = delete($args{file_name});
    my $file_instance = delete($args{instance_number});

    my $fh = $class->_get_file_handle(
        name     => $file_name,
        instance => $file_instance
    );

    $fh->unlock(%args);

    return FCACHE_RC_SUCCESS;
} ## end sub file_unlock

=head2 file_lock_request_status_verify

Verifies the status of a waiting lock request.
The waiting request must be initiated by this object.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file for which to check the lock request status.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple instances of this file
        have been opened by this object.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then it will default to the latest-opened instance.

=item B<lock_id> I<Required> Scalar:

        The lock request id, as returned by the file_lock method.
        
=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: FCACHE_RC_FILE_LOCK_REQUEST_WAITING, FCACHE_RC_FILE_LOCK_REQUEST_GRANTED.
        Default value: FCACHE_RC_FILE_LOCK_REQUEST_WAITING.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        # Request is expected to be still waiting
        $obj->file_lock_request_status_verify(file_name => $file1,
                                              lock_id => $waiting_lock_id,);
        # Request is expected to be granted
        $obj->file_lock_request_status_verify(file_name => $file1,
                                              lock_id => $waiting_lock_id,
                                              expected => FCACHE_RC_FILE_LOCK_REQUEST_GRANTED);

=cut

sub file_lock_request_status_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
            lock_id         => {type => SCALAR,},
            expected => {
                type    => SCALAR,
                default => FCACHE_RC_FILE_LOCK_REQUEST_WAITING
            },
        }
    );

    my $lock_id     = $args{lock_id};
    my $expected_rc = $args{expected};
    my $file_name   = $args{file_name};

    my $fh = $class->_get_file_handle(
        name     => $file_name,
        instance => $args{instance_number}
    );

    my $status = $fh->get_waiting_lock_status(waiting_id => $lock_id);
    fcache_log_comment(message => "Lock request status:\n"
            . "File: $file_name,\n"
            . "Lock ID: $lock_id\n"
            . "Current status: $status")
        if ($class->{_debug});

    # We only have two posibilities for $status at this point.
    # Any other failures will be caught by the Protocol library.
    my $rc
        = ($status eq 'waiting')
        ? FCACHE_RC_FILE_LOCK_REQUEST_WAITING
        : FCACHE_RC_FILE_LOCK_REQUEST_GRANTED;

    return fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $expected_rc,
        action      => "Check waiting lock status for file: $file_name",
        output      => "Status: $status",
    );
} ## end sub file_lock_request_status_verify

=head2 file_acls_modify

	Modify the acls for this file.
	This method is using the following tools:
	- For Windows: cacls.exe
	- For Unix: v4t_setacl

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to truncate.
        It should be:
        - a relative path: the path for the file name as seen in the volume.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
          
=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=item B<acl_pattern_unix> I<Optional> Scalar: 

        Absolute path to the sample acl file to be used when changing
	file acls from Unix clients.
        Default value: the value of the command-line argument
	FILE_SETACL_PATTERN_UNIX.

=item B<acl_pattern_win> I<Optional> Scalar: 

        The pattern to be used when changing
	file acls from Windows clients.
        Default value: the value of the command-line argument
	FILE_SETACL_PATTERN_WIN.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	$obj->file_acls_modify(file_name => 'my_file',);
	$obj->file_acls_modify(file_name => 'my_file',
	                       expected => FCACHE_RC_PERMISSION_DENIED);

=cut

sub file_acls_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name        => {type => SCALAR,},
            acl_pattern_unix => {type => SCALAR, optional => 1,},
            acl_pattern_win  => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name   = $args{file_name};
    my $client_type = $class->get_type();
    my $tool;
    my $acl_pattern;
    my $command = '';

    # We need access to command-line arguments, to get the tool name/path.
    my $options = NACL::OptionSet->new();

    if ($client_type eq FCACHE_CLIENT_TYPE_WINDOWS) {

        $tool = $options->get('FILE_SETACL_CMD_WIN');
        if (!defined($tool)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Command-line argument FILE_SETACL_CMD_WIN"
                    . " must be defined for file acls modification"
            );
        }

        $acl_pattern = $args{acl_pattern_win};
        if (!defined($acl_pattern)) {
            $acl_pattern = $options->get('FILE_SETACL_PATTERN_WIN');
        }
        if (!defined($acl_pattern)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Command-line argument FILE_SETACL_PATTERN_WIN"
                    . " must be defined for file acls modification\n"
                    . "Or the acl_pattern_win argument must be passed in"
            );
        }

        my $path = $class->_ensure_absolute_path(name => $file_name);
        $command = "$tool $path $acl_pattern";
    } else {

        $tool = $options->get('FILE_SETACL_CMD_UNIX');
        if (!defined($tool)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Command-line argument FILE_SETACL_CMD_UNIX"
                    . " must be defined for file acls modification"
            );
        }

        my $mount_junction = $class->{_mount_src_junction};
        my $ip             = $class->{_mount_src_ip};
        $acl_pattern = $args{acl_pattern_unix};
        if (!defined($acl_pattern)) {
            $acl_pattern = $options->get('FILE_SETACL_PATTERN_UNIX');
        }
        if (!defined($acl_pattern)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message => "Command-line argument FILE_SETACL_PATTERN_UNIX"
                    . " must be defined for file acls modification\n"
                    . "Or the acl_pattern_unix argument must be passed in"
            );
        }

        $command
            = "$tool -I $ip -F $mount_junction/$file_name -f $acl_pattern";
    }

    fcache_log_comment(message => "Setting acls for file $file_name\n"
            . "Command:\n $command")
        if ($class->{_debug});

    my $output = $class->{_nacl_client}->execute(command => $command);

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_FILE_SET_ACL,
        output      => $output,
        expected_rc => $args{expected},
    );
} ## end sub file_acls_modify

=head2 symlink_create

Create a symbolic link to a file.

=head3 ARGUMENTS

=over

=item B<path_name> I<Required> Scalar:

        Name of the file/directory that the link will point to.
        It can be:
        - a relative path: the relative path as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped
	on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<link_name> I<Required> Scalar:

        Name of the link to be created.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped
	on this client.
        E.g.
          - for link1 right under <vol>: link1
          - for link2 in directory dir1 under <vol>: dir1/link2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/link1

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $obj->symlink_create(path_name => 'my_file',
                             link_name => 'my_symlink',);

=head3 NOTES

        As documented in the Protocol library:
        - Perl's symlink() should be supported for SMB2.0+ (Windows Vista/2008 onwards)
        - If not, we need the CPAN module, Win32::Symlink installed
          [http://search.cpan.org/~autrijus/Win32-Symlink-0.04/Symlink.pm]

=cut

sub symlink_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   path_name => {type => SCALAR,},
            link_name => {type => SCALAR,},
        }
    );

    my $path_name = $class->_ensure_absolute_path(name => $args{path_name});
    my $link_name = $class->_ensure_absolute_path(name => $args{link_name});

    fcache_log_comment(message => "Creating symlink:\n"
            . "From: $path_name\n"
            . "Link name: $link_name");

    $class->{_file_operations}->symlink(
        file_name => $path_name,
        link_name => $link_name
    );

    return FCACHE_RC_SUCCESS;
} ## end sub symlink_create

=head2 hardlink_create

Create a hard link to a file.

=head3 ARGUMENTS

=over

=item B<path_name> I<Required> Scalar:

        Name of the file/directory that the link will point to.
        It can be:
        - a relative path: the path as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped
	on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<link_name> I<Required> Scalar:

        Name of the link to be created.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped
	on this client.
        E.g.
          - for link1 right under <vol>: link1
          - for link2 in directory dir1 under <vol>: dir1/link2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/link1

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $obj->hardlink_create(path_name => 'my_file',
                              link_name => 'my_link',);

=cut

sub hardlink_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   path_name => {type => SCALAR,},
            link_name => {type => SCALAR,},
        }
    );

    my $path_name = $class->_ensure_absolute_path(name => $args{path_name});
    my $link_name = $class->_ensure_absolute_path(name => $args{link_name});

    $class->{_file_operations}->link(
        file_name => $path_name,
        link_name => $link_name
    );

    return FCACHE_RC_SUCCESS;
} ## end sub hardlink_create

=head2 symlink_read

Obtain the pathname of the file pointed to by the symlink.

=head3 ARGUMENTS

=over

=item B<link_name> I<Required> Scalar:

        Name of the link to be checked.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for link1 right under <vol>: link1
          - for link2 in directory dir1 under <vol>: dir1/link2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/link1
    
=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Legal values: FCACHE_RC_PATH_IS_NOT_A_SYMLINK, FCACHE_RC_SUCCESS.
        Default value: FCACHE_RC_SUCCESS.
        
=back

=head3 RETURN

        The symlink target, if link_name is a symlink,
        undef otherwise.
        
=head3 NOTES

        Support:
        - CIFS SMB2.0+ (Windows Vista/2008 onwards)
        - NFS

=head3 EXAMPLE

        $obj->symlink_read(link_name => 'my_symlink',);
        $obj->symlink_read(link_name => 'this_is_not_a_symlink',
                           expected => FCACHE_RC_PATH_IS_NOT_A_SYMLINK);

=cut

sub symlink_read {
    my $class = shift;
    my %args  = validate(
        @_,
        {   link_name => {type => SCALAR,},
            expected  => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $link_name = $class->_ensure_absolute_path(name => $args{link_name});
    my $expected_rc = $args{expected};

    my $link_target = $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'symlink_read',
        args    => {link_name => $link_name,},
    );

    my $rc
        = (defined($link_target))
        ? FCACHE_RC_SUCCESS
        : FCACHE_RC_PATH_IS_NOT_A_SYMLINK;
    fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $expected_rc,
        action      => "Verify if path is a link: $link_name\n",
        output      => "Link target: " . Dumper($link_target)
    );

    return $link_target;
} ## end sub symlink_read

=head2 path_unlink

Unlinks/removes a file/directory.

=head3 ARGUMENTS

=over

=item B<link_name> I<Required> Scalar:

        Name of the link to be removed.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for link1 right under <vol>: link1
          - for link2 in directory dir1 under <vol>: dir1/link2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/link1
        
=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $obj->path_unlink(link_name => 'my_link');

=cut

sub path_unlink {
    my $class = shift;
    my %args = validate(@_, {link_name => {type => SCALAR,},});

    my $link_name = $class->_ensure_absolute_path(name => $args{link_name});

    $class->{_file_operations}->unlink(link_name => $link_name);

    return FCACHE_RC_SUCCESS;
} ## end sub path_unlink

=head2 files_compare

Compare two files.

=head3 ARGUMENTS

=over

=item B<file_name1> I<Required> Scalar:

        Name of the first file.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for link1 right under <vol>: link1
          - for link2 in directory dir1 under <vol>: dir1/link2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/link1

=item B<file_name2> I<Required> Scalar:

        Name of the second file.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for link1 right under <vol>: link1
          - for link2 in directory dir1 under <vol>: dir1/link2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/link1

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $obj->files_compare(file1 => 'my_file1',
	                    file2 => 'my_file2');
	$obj->files_compare(file1 => 'my_file1',
	                    file2 => 'my_file2',
			    expected => FCACHE_RC_FILES_NOT_EQUAL);

=cut

sub files_compare {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name1 => {type => SCALAR,},
            file_name2 => {type => SCALAR,},
            expected   => {type => SCALAR, default => FCACHE_RC_SUCCESS},
        }
    );

    my $file1 = $class->_ensure_absolute_path(name => $args{file_name1});
    my $file2 = $class->_ensure_absolute_path(name => $args{file_name2});

    my $result = $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'files_compare',
        args    => {
            file1 => $file1,
            file2 => $file2,
        },
    );

    fcache_log_comment(message => "Compared files:\n"
            . "$file1\n"
            . "$file2\n"
            . "Compare result (0 = equal, 1 = not equal): $result");

    my $rc = ($result == 0) ? FCACHE_RC_SUCCESS : FCACHE_RC_FILES_NOT_EQUAL;

    return fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $args{expected},
        action      => "Compare $file1 and $file2",
        output      => $result,
    );
} ## end sub files_compare

=head2 pipe_create

Create a named pipe (FIFO).
This method is not available on Windows.

=head3 ARGUMENTS

=over

=item B<pipe_name> I<Required> Scalar:

        Name of the pipe to be created.
        It can be:
        - a relative path: the path for the link as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for pipe1 right under <vol>: pipe1
          - for pipe2 in directory dir1 under <vol>: dir1/pipe2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/pipe1

=item B<pipe_name> I<Required> Scalar:

        Name of the pipe to create.

=item B<pipe_mode> I<Optional> Scalar:

        Pipe mode.
        Default value: FCACHE_DFLT_FIFO_MODE (defined in FlexCache::Constants).
        
=item B<pipe_umask> I<Optional> Scalar:

        Umask to set for the process when creating the named pipe.
        Default value: FCACHE_DFLT_FIFO_UMASK (defined in FlexCache::Constants).

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.
        
=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $obj->pipe_create(pipe_name => 'fifo1');
        $obj->pipe_create(pipe_name => 'fifo1',
                          expected => FCACHE_RC_FILE_ALREADY_EXISTS);

=cut

sub pipe_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   pipe_name => {type => SCALAR,},
            pipe_mode => {type => SCALAR, default => FCACHE_DFLT_FIFO_MODE,},
            pipe_umask =>
                {type => SCALAR, default => FCACHE_DFLT_FIFO_UMASK,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $pipe_name = $class->_ensure_absolute_path(name => $args{pipe_name});
    my $output = '';

    if ($class->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Named pipes are not available on Windows.\n"
                . "Script cannot be run on on this client: "
                . $class->{_name}
        );
    }

    try {
        $class->{_file_operations}->{executor}->execute(
            client  => $class->{_nacl_client},
            package => $class->{_platform_rpc_module},
            method  => 'pipe_create',
            args    => {
                pipe_name  => $pipe_name,
                pipe_mode  => $args{pipe_mode},
                pipe_umask => $args{pipe_umask},
            },
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_PIPE_OPEN,
        output      => $output,
        expected_rc => $args{expected},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub pipe_create

=head2 directory_create

Create a directory.

=head3 ARGUMENTS

=over

=item B<directory_name> I<Required> Scalar:

        Name of the directory to be created.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for dir1 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1
          
=item B<recursive> I<Optional> Boolean: 

        Flag to control if the creation should be recursive.
        If this flag is set to TRUE, then the library will handle
        the creation of any missing parent directories.
        Any error (except for "already existent")
        in this "parent" creation will result in script termination.
        Default value: TRUE.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.
        
=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        $obj->directory_create(directory_name => 'my_dir');
        $obj->directory_create(directory_name => 'my_dir',
                               expected => FCACHE_RC_DIR_ALREADY_EXISTS);

=cut

sub directory_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   directory_name => {type => SCALAR,},
            recursive      => {type => BOOLEAN, default => TRUE},
            expected       => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $full_dir_path
        = $class->_ensure_absolute_path(name => $args{directory_name});
    my $recursive = $args{recursive};
    my $output    = '';

    if ($recursive) {
        my $directory_name      = $full_dir_path;
        my $created_parent_dirs = '';
        fcache_log_comment(
            message => "Recursive directory creation requested:\n"
                . "Full directory path: $directory_name")
            if ($class->{_debug});

        # So we don't spend our time "re-creating" the mountpoint...
        my $mount_point = $class->{_mount_point};
        if ($directory_name =~ /^$mount_point/) {
            $directory_name =~ s/^$mount_point\///;
            $created_parent_dirs = "$mount_point";
            fcache_log_comment(message =>
                    "Ignoring the mountpoint from the parent directory chain:\n"
                    . "Mountpoint: $mount_point\n"
                    . "New subdirectory chain: $directory_name")
                if ($class->{_debug});
        }

        my @parent_directories = split(/\//, $directory_name);

        # Just remove the deepest directory (our directory).
        # It will be handled separately after the loop exits.
        # For all the other parent directories in the chain,
        # we will allow the "already exists" error.
        pop(@parent_directories);
        fcache_log_comment(message => "Recursive directory creation:\n"
                . "Parent directory chain:\n"
                . Dumper(@parent_directories))
            if ($class->{_debug});

        foreach my $parent_dir (@parent_directories) {
            my $parent_dir_output = '';
            fcache_log_comment(
                message => "Creating parent directory: $parent_dir")
                if ($class->{_debug});
            $created_parent_dirs .= "/$parent_dir";

            try {
                $class->{_file_operations}
                    ->mkdir(directory_name => $created_parent_dirs);
            }
            catch NATE::BaseException with {
                my $exception = shift;
                $parent_dir_output = $exception->text();
            };

            my $rc = $class->{_ui_messages}->verify_expected_rc(
                command     => FCACHE_CLIENT_CMD_DIR_CREATE,
                output      => $parent_dir_output,
                expected_rc => FCACHE_RC_FAILURE_NON_FATAL,
            );

            if (   ($rc != FCACHE_RC_SUCCESS)
                && ($rc != FCACHE_RC_DIR_ALREADY_EXISTS)) {
                NATE::BaseException->throw(
                          "Failed to create parent directory chain:\n"
                        . "$parent_dir_output");
            }

        }
    }

    try {
        $class->{_file_operations}->mkdir(directory_name => $full_dir_path);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_DIR_CREATE,
        output      => $output,
        expected_rc => $args{expected},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub directory_create

=head2 directory_list

List the contents of a directory.

=head3 ARGUMENTS

=over

=item B<directory_name> I<Required> Scalar:

        Name of the directory.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<expected_contents_count> I<Optional> Scalar: 

        Expected count for the directory contents (excluding .. and .).
        If this argument is passed in, then the directory is expected to
        contain this many files or directories. If the two numbers do not match,
        then script execution is terminated with a fatal failure.
        
=back

=head3 RETURN

        Array of directory contents.

=head3 EXAMPLE

        $obj->directory_list(directory_name => 'my_dir',
                             expected_contents_count => 7);
        my @dir_contents = $obj->directory_list(directory_name => 'my_dir');

=cut

sub directory_list {
    my $class = shift;
    my %args  = validate(
        @_,
        {   directory_name => {type => SCALAR,},
            expected       => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            expected_contents_count =>
                {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
        }
    );

    my $expected = $args{expected};
    my $full_dir_path
        = $class->_ensure_absolute_path(name => $args{directory_name});
    my $dir_contents = '';
    my $error        = '';

    try {
        $dir_contents = $class->{_file_operations}->{executor}->execute(
            client  => $class->{_nacl_client},
            package => $class->{_agnostic_rpc_module},
            method  => 'directory_list',
            args    => {directory_name => $full_dir_path,},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $error        = $exception->text();
        $dir_contents = '';
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_DIR_LIST,
        output      => $error,
        expected_rc => $expected,
    );

    my @dir_contents = split(/,/, $dir_contents);

    if (   ($expected == FCACHE_RC_SUCCESS)
        && (defined($args{expected_contents_count}))) {
        my $expected_count = $args{expected_contents_count};
        my $actual_count   = scalar(@dir_contents);

        if ($expected_count != $actual_count) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Directory list count mismatch:\n"
                    . "Directory name: $full_dir_path\n"
                    . "Expected contents count: $expected_count\n"
                    . "Obtained contents count: $actual_count\n"
                    . "Directory contents: $dir_contents"
            );
        }
    }

    fcache_log_comment(message => "Directory list:\n"
            . "Directory name: $full_dir_path\n"
            . "Contents: $dir_contents");

    return @dir_contents;
} ## end sub directory_list

=head2 directory_exists

Check if a directory exists on this client.

=head3 ARGUMENTS

=over

=item B<directory_name> I<Required> Scalar:

        Name of the directory to be checked.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Legal values: FCACHE_RC_SUCCESS, FCACHE_RC_NO_ENTRY_FOUND.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	$obj->directory_exists(directory_name => $dir1);
        $obj->directory_exists(directory_name => $dir1,
                               expected => FCACHE_RC_NO_ENTRY_FOUND);

=cut

sub directory_exists {
    my $class = shift;
    my %args  = validate(
        @_,
        {   directory_name => {type => SCALAR,},
            expected       => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $full_dir_path
        = $class->_ensure_absolute_path(name => $args{directory_name});

    my $dir_exists = $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'directory_exists',
        args    => {directory_name => $full_dir_path,},
    );

    my $rc = ($dir_exists) ? FCACHE_RC_SUCCESS : FCACHE_RC_NO_ENTRY_FOUND;

    return fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $args{expected},
        action      => "Verify if directory exists: $full_dir_path",
        output      => "Directory exists: $dir_exists"
    );
} ## end sub directory_exists

=head2 directory_delete

Delete a directory.

=head3 ARGUMENTS

=over

=item B<directory_name> I<Required> Scalar:

        Name of the directory to be deleted.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<recursive> I<Optional> Boolean: 

        Flag to control if the deletion should be recursive.
        Any error will result in script termination.
        Default value: TRUE.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	$obj->directory_delete(directory_name => $dir1);
        $obj->directory_delete(directory_name => $dir1,
                               expected => FCACHE_RC_NO_ENTRY_FOUND);

=cut

sub directory_delete {
    my $class = shift;
    my %args  = validate(
        @_,
        {   directory_name => {type => SCALAR,},
            recursive      => {type => BOOLEAN, default => TRUE,},
            expected       => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $full_dir_path
        = $class->_ensure_absolute_path(name => $args{directory_name});
    my $output = '';

    try {
        if ($args{recursive}) {
            $class->{_file_operations}->{executor}->execute(
                client  => $class->{_nacl_client},
                package => $class->{_agnostic_rpc_module},
                method  => 'directory_tree_remove',
                args    => {directory_name => $full_dir_path,},
            );
        } else {
            $class->{_file_operations}
                ->rmdir(directory_name => $full_dir_path,);
        }
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_DIR_DELETE,
        output      => $output,
        expected_rc => $args{expected},
    );

} ## end sub directory_delete

=head2 dataset_create

Create a dataset tree.

=head3 ARGUMENTS

=over

=item B<dataset_name> I<Optional> Scalar:

        Name of the dataset directory to be created.
        If this argument is not passed in, then the dataset will be created
        directly under the volume.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<dataset_depth> I<Required> Scalar:

        Depth of the dataset tree (number of subdirectories to contain).
        It can be 0 if no subdirectories should be created or
        any positive number of subdirectories to create.
        Default value: 0.

=item B<file_count_per_level> I<Required> Scalar:

        How many files to create per each directory/subdirectory level.

=item B<file_size> I<Required> Scalar:

        Size for the files to create per each directory/subdirectory level.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

        # Directly under the mountpoint, 5 files of size 16b
	$obj->dataset_create(file_count_per_level => 5,
                                  file_size => 16);
        # In <mount_point>/dataset1/level1, 3 subdir levels,
        # 4 files of size 4b per level
        $obj->dataset_create(dataset_name => 'dataset1/level1',
                             dataset_depth => 3,
                             file_count_per_level => 4,
                             file_size => 4);

=cut

sub dataset_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dataset_name => {type => SCALAR, optional => 1,},
            dataset_depth =>
                {type => SCALAR, regex => qr/^\d+$/, default => 0},
            file_count_per_level => {type => SCALAR, regex => qr/^\d+$/,},
            file_size            => {type => SCALAR, regex => qr/^\d+$/,},
        }
    );

    my $dataset_name         = $args{dataset_name};
    my $directory_depth      = $args{dataset_depth};
    my $file_count_per_level = $args{file_count_per_level};
    my $file_size            = $args{file_size};
    my $full_dir_path        = $class->{_mount_point};

    if (   ($directory_depth < 0)
        || ($file_count_per_level <= 0)
        || ($file_size <= 0)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Invalid arguments: They should be\n"
                . "directory_depth: positive (was $directory_depth)\n"
                . "file_count_per_level: strictly positive (was $file_count_per_level)\n"
                . "file_size: strictly positive (was $file_size)"
        );
    }

    if (defined($dataset_name)) {
        $full_dir_path = $class->_ensure_absolute_path(name => $dataset_name);
        $class->directory_create(directory_name => $full_dir_path);
    }

    for (my $level = 0; $level <= $directory_depth; $level++) {
        if ($level > 0) {
            $full_dir_path .= '/subdir_' . $level;
            $class->directory_create(
                directory_name => $full_dir_path,
                recursive      => FALSE
            );
        }

        for (
            my $file_count = 1;
            $file_count <= $file_count_per_level;
            $file_count++
            ) {
            my $file_name = $full_dir_path . '/file_' . $file_count;
            $class->file_open(
                file_name => $file_name,
                open_mode => FCACHE_FILE_OPEN_MODE_CREATE_NEW,
            );
            $class->file_write(
                file_name => $file_name,
                num_bytes => $file_size
            );
            $class->file_close(file_name => $file_name);
        }
    }
} ## end sub dataset_create

=head2 dataset_list_files

List the files in a dataset tree.

=head3 ARGUMENTS

=over

=item B<dataset_name> I<Optional> Scalar:

        Name of the dataset directory to be listed.
        If this argument is not passed in, then the volume contents will be listed.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<relative_paths> I<Optional> Boolean: 

        Flag to control if the content paths are returned as absolute paths or
	as relative (to the volume) paths. Setting this flag to true will cause
	the library to return all the paths for the dataset contents as paths
	relative to the volume.
	Default value: FALSE.
        
=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Legal values: FCACHE_RC_SUCCESS, FCACHE_RC_NO_CONTENTS_FOUND;
        Default value: FCACHE_RC_SUCCESS.

=item B<expected_file_count> I<Optional> Scalar: 

        Expected number of files to be contained by this dataset.

=back

=head3 RETURN

        An array of files (absolute or relative paths) contained by this dataset.

=head3 EXAMPLE

        # List the volume tree (get the absolute file paths)
	my @files = $obj->dataset_list_files();
	# List the volume tree (get the relative file paths)
	my @files = $obj->dataset_list_files(relative_paths => TRUE);
        # List a specific tree
        my @files = $obj->dataset_list_files(dataset_name => 'dataset1/level1',);

=cut

sub dataset_list_files {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dataset_name   => {type => SCALAR,  optional => 1,},
            relative_paths => {type => BOOLEAN, default  => FALSE,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            expected_file_count =>
                {type => SCALAR, optional => 1, regex => qr/^\d+$/},
        }
    );

    my $mount_point = $class->{_mount_point};
    my $dataset_name;
    if (defined($args{dataset_name})) {
        $dataset_name
            = $class->_ensure_absolute_path(name => $args{dataset_name});
    } else {
        $dataset_name = $mount_point;
    }

    my $files = $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'dataset_list_files',
        args    => {dataset_name => $dataset_name,},
    );

    if ($args{relative_paths}) {
        fcache_log_comment(message => "Relative paths required:\n"
                . "Starting to process:\n"
                . "$files")
            if ($class->{_debug});
        $files =~ s/^$mount_point\///;
        $files =~ s/,$mount_point\//,/g;
        fcache_log_comment(message => "Relative paths required:\n"
                . "Completed processing:\n"
                . "$files")
            if ($class->{_debug});
    }

    my @dataset_files = split(/,/, $files);
    my $rc
        = (scalar(@dataset_files) > 0)
        ? FCACHE_RC_SUCCESS
        : FCACHE_RC_NO_CONTENTS_FOUND;
    fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $args{expected},
        action      => "List files in dataset $dataset_name",
        output      => $files,
    );

    fcache_log_comment(message => "Listing dataset files:\n"
            . "Dataset name: $dataset_name\n"
            . "Dataset files:\n"
            . Dumper(\@dataset_files));

    if (defined($args{expected_file_count})) {
        my $expected_file_count = $args{expected_file_count};
        my $obtained_file_count = scalar(@dataset_files);
        if ($obtained_file_count != $expected_file_count) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Mismatch in dataset expected file count:\n"
                    . "Dataset: $dataset_name\n"
                    . "Expected file count: $expected_file_count\n"
                    . "Obtained file count: $obtained_file_count"
            );
        }

        fcache_log_comment(message => "Dataset files count match:\n"
                . "Dataset name: $dataset_name\n"
                . "Expected file count: $expected_file_count\n"
                . "Obtained file count: $obtained_file_count");
    }

    return @dataset_files;
} ## end sub dataset_list_files

=head2 dataset_read_files

Read all the files in a dataset tree.

=head3 ARGUMENTS

=over

=item B<dataset_name> I<Optional> Scalar:

        Name of the dataset directory.
        If this argument is not passed in, then the whole volume contents will be processed.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<num_bytes> I<Optional> Scalar:

        The number of bytes to read from each file.
        Default value: 1.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

        # Read 5 bytes from each of the files in this volume
	$obj->dataset_read_files(num_bytes => 5);
        # Read the files in a specific dataset tree (first byte of each file)
        $obj->dataset_read_files(dataset_name => 'dataset1/level1',);

=cut

sub dataset_read_files {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dataset_name => {type => SCALAR, optional => 1,},
            num_bytes => {type => SCALAR, regex => qr/^\d+$/, default => 1,},
        }
    );

    my $num_bytes = delete($args{num_bytes});

    my @dataset_files = $class->dataset_list_files(%args);

    foreach my $file (@dataset_files) {
        $class->file_open(
            file_name   => $file,
            access_mode => FCACHE_FILE_ACCESS_MODE_R,
            open_mode   => FCACHE_FILE_OPEN_MODE_OPEN_EXIST,
        );
        $class->file_read(
            file_name => $file,
            num_bytes => $num_bytes
        );
        $class->file_close(file_name => $file);
    }

    return FCACHE_RC_SUCCESS;
} ## end sub dataset_read_files

=head2 dataset_rich_data_create

Create a dataset and populate it with rich data.

=head3 ARGUMENTS

=over

=item B<dataset_name> I<Optional> Scalar:

        Name of the dataset directory.
        If this argument is not passed in, then the whole volume contents will be populated.
        It can be:
        - a relative path: the path for the directory as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for di11 right under <vol>: dir1
          - for dir2 in directory dir1 under <vol>: dir1/dir2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/dir1

=item B<file_size_avg> I<Optional> Scalar:

        Average size of files.
        Default value: KB.

=item B<dir_avg> I<Optional> Scalar:

        Average number of subdirectories to create under each directory.
        default value: 2.

=item B<file_avg> I<Optional> Scalar:

        Average number of files to create under each directory.
        Default value: 2.

=item B<sparse_num> I<Optional> Scalar:

        Number of sparse files.

=item B<link_num> I<Optional> Scalar:

        Number of links.

=item B<slink_num> I<Optional> Scalar:

        Number of symbolic links.

=item B<compress_num> I<Optional> Scalar:

        Number of compressed files.

=item B<encrypt_num> I<Optional> Scalar:

        Number of encrypted files.

=item B<archive_num> I<Optional> Scalar:

        Number of archive files.

=item B<hidden_num> I<Optional> Scalar:

        Number of hidden files.
        
=back

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

        $obj->dataset_rich_data_create(dataset_name => 'my_rich_dataset',
                                       file_size_avg => MB,
                                       dir_avg => 3,
                                       compress_num => 2,
                                       slink_num => 1);

=cut

sub dataset_rich_data_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   dataset_name  => {type => SCALAR, optional => 1,},
            file_size_avg => {type => SCALAR, default  => KB,},
            dir_avg       => {type => SCALAR, default  => 2,},
            file_avg      => {type => SCALAR, default  => 2,},
            sparse_num    => {type => SCALAR, optional => 1,},
            link_num      => {type => SCALAR, optional => 1,},
            slink_num     => {type => SCALAR, optional => 1,},
            compress_num  => {type => SCALAR, optional => 1,},
            encrypt_num   => {type => SCALAR, optional => 1,},
            archive_num   => {type => SCALAR, optional => 1,},
            hidden_num    => {type => SCALAR, optional => 1,},
        }
    );

    my $dataset_name  = delete($args{dataset_name});
    my $full_dir_path = $class->{_mount_point};

    if (defined($dataset_name)) {
        $full_dir_path = $class->_ensure_absolute_path(name => $dataset_name);
        $class->directory_create(directory_name => $full_dir_path);
    }

    RichData::populate(
        path    => $full_dir_path,
        client  => $class->{_name},
        timeout => FCACHE_DFLT_DATASET_CREATE_TIMEOUT,
        %args
    );

    return FCACHE_RC_SUCCESS;
} ## end sub dataset_rich_data_create

=head2 path_attributes_verify

Verify status information fields for a file or directory.
Please note that not all the fields are supported by all filesystem types.
For more details, please see http://perldoc.perl.org/perlport.html#Files-and-Filesystems.
Fields that cannot be obtained for a specific filesystem will 
not be included in the method's return hash.

=head3 ARGUMENTS

=over

=item B<path_name> I<Required> Scalar:

        Name of the file or directory.
        It can be:
        - a relative path: the path for the file/dir name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<device_number> I<Optional> Scalar:

        Device number of filesystem.

=item B<inode> I<Optional> Scalar:

        Inode number for the file.

=item B<mode> I<Optional> Scalar:

        Mode (type and permissions).
        WARNING: This needs to be OR-ed with the type of file/dir for a correct comparison:
        S_IFREG S_IFDIR S_IFLNK S_IFBLK S_IFCHR S_IFIFO S_IFSOCK S_IFWHT S_ENFMT,
        exported by Fcntl ':mode' (not all available on all filesystems).
        More information at: http://perldoc.perl.org/functions/stat.html
        
        For example, a script that first changes the file mode, then verifies the change, should look like:
        
        ...
        use Fcntl ':mode';
        
        ...
        $origin->{client}->path_attributes_modify(file_name => $file2,
                                                  mode => 0777,);
        $origin->{client}->path_attributes_verify(file_name => $file2,
                                                  mode => (S_IFREG|0777),);

=item B<links> I<Optional> Scalar:

        Number of (hard) links to the file/dir.

=item B<uid> I<Optional> Scalar:

        Numeric user ID of owner.

=item B<gid> I<Optional> Scalar:

        Numeric group ID of owner.

=item B<device_identifier> I<Optional> Scalar:

        The device identifier (special files only).

=item B<size> I<Optional> Scalar:

        Total size of file/dir, in bytes

=item B<atime> I<Optional> Scalar:

        Last access time in seconds since the epoch.

=item B<mtime> I<Optional> Scalar:

        Last modify time in seconds since the epoch.

=item B<ctime> I<Optional> Scalar:

        Inode change time in seconds since the epoch.

=item B<block_size> I<Optional> Scalar:

        Preferred block size for file system I/O.

=item B<archive> I<Optional> Boolean:

        Archive bit. Available only for Windows.
        
=item B<compress> I<Optional> Boolean:

        Compressed bit. Available only for Windows.
        
=item B<encrypt> I<Optional> Boolean:

        Encrypted bit. Available only for Windows.
        
=item B<hidden> I<Optional> Boolean:

        Hidden bit. Available only for Windows.
        
=item B<sparse> I<Optional> Boolean:

        Sparse bit. Available only for Windows.

=back

=head3 RETURN

        Hash containing all the fields that could be obtained
        from the stats command. The hash key names are identical
        with the method's arguments, as documented above,
        except for the path_name.

=head3 EXAMPLE

        # Verify that the file size is 1k
	$obj->path_attributes_verify(path_name => 'my_file',
                                     size => 1024);
        # Verify that the file size is 1k
        # and that there is 1 hard link to the file
        $obj->path_attributes_verify(path_name => 'my_file',
                                     size => 1024,
                                     links => 1);

=cut

sub path_attributes_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   path_name         => {type => SCALAR,},
            inode             => {type => SCALAR, optional => 1,},
            mode              => {type => SCALAR, optional => 1,},
            links             => {type => SCALAR, optional => 1,},
            uid               => {type => SCALAR, optional => 1,},
            gid               => {type => SCALAR, optional => 1,},
            size              => {type => SCALAR, optional => 1,},
            atime             => {type => SCALAR, optional => 1,},
            mtime             => {type => SCALAR, optional => 1,},
            ctime             => {type => SCALAR, optional => 1,},
            block_size        => {type => SCALAR, optional => 1,},
            blocks            => {type => SCALAR, optional => 1,},
            device_number     => {type => SCALAR, optional => 1,},
            device_identifier => {type => SCALAR, optional => 1,},
            archive           => {type => SCALAR, optional => 1,},
            compress          => {type => SCALAR, optional => 1,},
            encrypt           => {type => SCALAR, optional => 1,},
            hidden            => {type => SCALAR, optional => 1,},
            sparse            => {type => SCALAR, optional => 1,},
        }
    );

    my $path_name = $class->_ensure_absolute_path(name => $args{path_name});
    delete($args{path_name});

    my $skip_checks = FALSE;
    if (scalar(keys(%args)) == 0) {
        $skip_checks = TRUE;
    }

    my $path_stats_str = $class->{_file_operations}->{executor}->execute(
        client  => $class->{_nacl_client},
        package => $class->{_agnostic_rpc_module},
        method  => 'path_stat',
        args    => {path_name => $path_name,},
    );

    my %stats_hash = ();
    $class->_process_RPC_output(
        output       => $path_stats_str,
        mapping_keys => [
            'device_number',     'inode',
            'mode',              'links',
            'uid',               'gid',
            'device_identifier', 'size',
            'atime',             'mtime',
            'ctime',             'block_size',
            'blocks',
        ],
        description            => "Stats for path: $path_name",
        allow_undefined_fields => TRUE,
        results_hash           => \%stats_hash
    );

    if ($class->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {

        # Also get the attributes: encrypt, archive, hidden, compress, sparse
        my $win_stats_str = $class->{_file_operations}->{executor}->execute(
            client  => $class->{_nacl_client},
            package => $class->{_platform_rpc_module},
            method  => 'get_windows_path_attributes',
            args    => {path_name => $path_name,},
        );
        $class->_process_RPC_output(
            output => $win_stats_str,
            mapping_keys =>
                ['archive', 'compress', 'encrypt', 'hidden', 'sparse',],
            description  => "Windows attributes for: $path_name",
            results_hash => \%stats_hash
        );
    }

    if ($class->{_debug}) {
        my $mode      = $stats_hash{mode} & 07777;
        my $user_rwx  = ($mode & 00700) >> 6;
        my $group_rwx = ($mode & 00070) >> 3;
        my $other_rwx = ($mode & 00007);
        fcache_log_comment(message => "Attributes for:\n"
                . "Path: $path_name\n"
                . "Permissions:\n"
                . "- user: $user_rwx\n"
                . "- group: $group_rwx\n"
                . "- other: $other_rwx\n"
                . Dumper(\%stats_hash));
    }

    if ($skip_checks) {
        fcache_log_comment(message => "Skipping attribute checks")
            if ($class->{_debug});
        return %stats_hash;
    }

    fcache_compare_hashes(
        obtained         => \%stats_hash,
        expected         => \%args,
        hash_description => "Stats for $path_name"
    );

    return %stats_hash;
} ## end sub path_attributes_verify

=head2 path_attributes_compare

Compare the attributes of two files or directories.
If the fields attribute is passed in, then
all the specified fields are compared for the two files/dirs.
Otherwise, this method will compare all the fields
that the client's filesystem is supporting,
except for the device_number field.
If the two files/dirs have different values for any of the
stat fields, then script execution is terminated
with a fatal failure.

=head3 ARGUMENTS

=over

=item B<path_name1> I<Required> Scalar:

        Name of the file or directory.
        It can be:
        - a relative path: the path for the file/dir name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<path_name2> I<Required> Scalar:

        Name of the file or directory.
        It can be:
        - a relative path: the path for the file/dir name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<fields> I<Optional> Arrayref:

        Arrayref of specific field names to check.
        The field names can be any of the keys
        of the hash returned by path_attributes_verify.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

        # Compare all the supported fields
        $origin->{client}->path_attributes_compare(path_name1 => $file2,
                                                   path_name2 => $file2,);
        # Compare only size and atime
        $origin->{client}->path_attributes_compare(path_name1 => $file2,
                                                   path_name2 => $file2,
                                                   fields => ['size', 'atime']);

=cut

sub path_attributes_compare {
    my $class = shift;
    my %args  = validate(
        @_,
        {   path_name1 => {type => SCALAR,},
            path_name2 => {type => SCALAR,},
            fields     => {type => ARRAYREF, optional => 1,},
        }
    );

    my $path_name1 = $class->_ensure_absolute_path(name => $args{path_name1});
    my $path_name2 = $class->_ensure_absolute_path(name => $args{path_name2});
    my $compare_fields        = $args{fields};
    my %path1_fields_to_check = ();

    my %path1_stats
        = $class->path_attributes_verify(path_name => $path_name1,);

    if ($args{fields}) {

        # Save in %path1_fields_to_check only (and all)
        # the fields required by the caller
        foreach my $field (@{$compare_fields}) {
            if (!exists($path1_stats{$field})) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_NA,
                    message     => "Stat field $field  is not supported by:\n"
                        . "Client type: "
                        . $class->{_type}
                );
            }
            $path1_fields_to_check{$field} = $path1_stats{$field};
        }
    } else {

     # Check all the fields that are available for path1,
     # but ignoring device_number (may not match for origin/cache comparisons)
        delete($path1_stats{device_number});
        %path1_fields_to_check = %path1_stats;
    }

    $class->path_attributes_verify(
        path_name => $path_name2,
        %path1_fields_to_check
    );

    return FCACHE_RC_SUCCESS;
} ## end sub path_attributes_compare

=head2 path_attributes_modify

Modify one or more attributes of a file or directory.
Please note that not all the fields are supported by all filesystem types.

=head3 ARGUMENTS

=over

=item B<path_name> I<Required> Scalar:

        Name of the file or directory.
        It can be:
        - a relative path: the path for the file/dir name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<mode> I<Optional> Scalar:

        File mode (type and permissions).

=item B<uid> I<Optional> Scalar:

        Numeric user ID of file's owner.

=item B<gid> I<Optional> Scalar:

        Numeric group ID of file's owner.

=item B<archive> I<Optional> Boolean:

        Archive bit. Available only for Windows.
        
=item B<hidden> I<Optional> Boolean:

        Hidden bit. Available only for Windows.

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if expected return code is not obtained.
        Legal values: FCACHE_RC_SUCCESS, FCACHE_RC_NO_CONTENTS_FOUND;
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

=head3 EXAMPLE

        $obj->path_attributes_modify(path_name => 'my_file',
                                     mode => 0777,);
        $obj->path_attributes_modify(path_name => $file2,
                                     archive => FALSE,
                                     hidden => TRUE,

=cut

sub path_attributes_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   path_name => {type => SCALAR,},
            mode      => {type => SCALAR, optional => 1,},
            uid       => {type => SCALAR, optional => 1,},
            gid       => {type => SCALAR, optional => 1,},
            archive   => {type => BOOLEAN, optional => 1,},
            hidden    => {type => BOOLEAN, optional => 1,},
            expected  => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $expected_rc = delete($args{expected});
    my $output      = '';
    my $path_name   = $class->_ensure_absolute_path(name => $args{path_name});
    delete($args{path_name});
    if (scalar(keys(%args)) == 0) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "At least one attribute must be specified for modification"
        );
    }

    if ($class->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        if (   defined($args{mode})
            || defined($args{uid})
            || defined($args{gid})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_NA,
                message =>
                    "The following attributes are not supported on Windows:\n"
                    . "mode, uid, gid"
            );
        }
    } elsif (defined($args{archive}) || defined($args{hidden})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message =>
                "The following attributes are only supported on Windows:\n"
                . "archive, hidden"
        );
    }

    if (defined($args{mode})) {
        try {
            $class->{_file_operations}->{executor}->execute(
                client  => $class->{_nacl_client},
                package => $class->{_platform_rpc_module},
                method  => 'set_path_mode',
                args    => {
                    path_name => $path_name,
                    mode      => $args{mode},
                },
            );
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };
    }

    if (defined($args{uid}) || defined($args{gid})) {

        # Even if the caller only specifies one,
        # chown requires both.
        # Set them accordingly.
        my $uid = defined($args{uid}) ? $args{uid} : -1;
        my $gid = defined($args{gid}) ? $args{gid} : -1;

        try {
            $class->{_file_operations}->{executor}->execute(
                client  => $class->{_nacl_client},
                package => $class->{_platform_rpc_module},
                method  => 'set_path_owner',
                args    => {
                    path_name => $path_name,
                    uid       => $uid,
                    gid       => $gid,
                },
            );
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };
    }

    if ($class->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {

        # If we got here, the %args hash only contains the windows attributes
        # that can be simply set/unset
        # Let the RPC subroutine handle them individually
        try {
            $class->{_file_operations}->{executor}->execute(
                client  => $class->{_nacl_client},
                package => $class->{_platform_rpc_module},
                method  => 'set_windows_path_attributes',
                args    => {
                    path_name => $path_name,
                    %args,
                },
            );
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };
    }

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CLIENT_CMD_PATH_ATTRIBUTES_MODIFY,
        output      => $output,
        expected_rc => $expected_rc,
    );
} ## end sub path_attributes_modify

=head2 stream_open

Open a file stream on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped
	on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<stream_name> I<Required> Scalar:

        Name of the stream.

=item B<open_mode> I<Required> Scalar:

        How should the file stream be opened.
        Legal values: any of the FCACHE_FILE_OPEN_MODE_* constants
        (defined in FlexCache::Constants).

=item B<access_mode> I<Optional> Scalar:

        Access mode with which the file stream should be opened.
        Legal values: any of the FCACHE_FILE_ACCESS_MODE_* constants
        (defined in FlexCache::Constants).
        Default value: FCACHE_FILE_ACCESS_MODE_RW.

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if expected return code is not obtained.
        Legal values: FCACHE_RC_SUCCESS, FCACHE_RC_NO_CONTENTS_FOUND;
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        Instance number for this instance (Scalar).
        This is usually not needed by the script unless it needs to work
        with multiple instances of the streams being opened at the same time.


EXAMPLE

        $obj->stream_open(file_name => $file,
	                  stream_name => 'stream',
                          open_mode => FCACHE_FILE_OPEN_MODE_CREATE_NEW,);

=cut

sub stream_open {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name   => {type => SCALAR,},
            stream_name => {type => SCALAR,},
            open_mode   => {type => SCALAR,},
            access_mode =>
                {type => SCALAR, default => FCACHE_FILE_ACCESS_MODE_RW,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name   = delete($args{file_name});
    my $stream_name = delete($args{stream_name});

    if ($class->{_type} ne FCACHE_CLIENT_TYPE_WINDOWS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Streams are currently only supported for Windows"
        );
    }

    return $class->file_open(
        file_name => "$file_name:$stream_name",
        %args
    );
} ## end sub stream_open

=head2 stream_write

Write to a file stream on this client.
Note: The file stream needs to be previously opened.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file stream to write to.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly mounted/mapped.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<stream_name> I<Required> Scalar:

        Name of the stream.

=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple stream instances
        have been opened by this object.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then it will default to the
	latest-opened instance.

=item B<offset> I<Optional> Scalar:

        Offset at which to start writing.
        Default value: 0.

=item B<num_bytes> I<Optional> Scalar:

        The number of bytes to write to the file.
        Either the num_bytes or contents parameter must be specified.

=item B<contents> I<Optional> Scalar:

        The contents string to write to the file.
        Either the num_bytes or contents parameter must be specified.

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

	# Open a single instance of the file and write.
	$obj->stream_write(file_name => $file1, stream_name => 'stream1');
        $obj->stream_write(file_name => $file1,
	                   stream_name => 'stream1',
                           num_bytes => MB);

=cut

sub stream_write {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            stream_name     => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
            offset => {type => SCALAR, regex => qr/^\d+$/, default => 0,},
            num_bytes => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            contents => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default  => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name   = delete($args{file_name});
    my $stream_name = delete($args{stream_name});

    if ($class->{_type} ne FCACHE_CLIENT_TYPE_WINDOWS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Streams are currently only supported for Windows"
        );
    }

    return $class->file_write(
        file_name => "$file_name:$stream_name",
        %args
    );

} ## end sub stream_write

=head2 stream_read

Read a file stream from this client.
Note: The file needs to be previously opened.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly
	mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<stream_name> I<Required> Scalar:

        Name of the stream.

=item B<instance_number> I<Optional> Scalar:

        This argument could be used in the case where multiple instances
        of this stream have been opened by this object.
        In this case, the instance number is the number returned by stream_open.
        If this argument is not specified, then it will default to
	the latest-opened instance.

=item B<offset> I<Optional> Scalar:

        Offset at which to start reading.
        Default value: 0.

=item B<num_bytes> I<Required> Scalar:

        The number of bytes to read from the file.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, defined in FlexCache::Constants.
        Execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        The contents of the file (Scalar).

=head3 EXAMPLE

        # Read 100 bytes, from the beginning of the file
	my $contents = $obj->stream_read(file_name => 'my_file',
	                               stream_name => 'stream1',
                                       num_bytes => 100);
        # Read 5 bytes, starting at offset 10
	my $contents = $obj->stream_read(file_name => 'my_file',
	                                 stream_name => 'stream1',
                                         offset => 10,
                                         num_bytes => 5);

=cut

sub stream_read {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            stream_name     => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
            offset => {type => SCALAR, regex => qr/^\d+$/, default => 0,},
            num_bytes => {type => SCALAR, regex   => qr/^\d+$/,},
            expected  => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name   = delete($args{file_name});
    my $stream_name = delete($args{stream_name});

    if ($class->{_type} ne FCACHE_CLIENT_TYPE_WINDOWS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Streams are currently only supported for Windows"
        );
    }

    return $class->file_read(
        file_name => "$file_name:$stream_name",
        %args
    );
} ## end sub stream_read

=head2 stream_close

Close a file stream on this client.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file.
        It can be:
        - a relative path: the path for the file name as seen in the volume.
        In this case the volume/share needs to be previosly
	mounted/mapped on this client.
        E.g.
          - for file1 right under <vol>: file1
          - for file2 in directory dir1 under <vol>: dir1/file2
        - an absolute path:
        E.g.
          /client/side/absolute/path/to/file1

=item B<stream_name> I<Required> Scalar:

        Name of the stream.

=item B<instance_number> I<Optional> Scalar:

        This argument should be used in the case where multiple instances
	of this file have been opened by this object and only a particular
	instance needs to be closed.
        In this case, the instance number is the number returned by file_open.
        If this argument is not specified, then all the opened instances
	of this file will be closed.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

=head3 EXAMPLE

	$obj->stream_close(file_name => $file1, stream_name => 'stream1');
=cut

sub stream_close {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,},
            stream_name     => {type => SCALAR,},
            instance_number => {type => SCALAR, optional => 1,},
        }
    );

    my $file_name   = delete($args{file_name});
    my $stream_name = delete($args{stream_name});

    if ($class->{_type} ne FCACHE_CLIENT_TYPE_WINDOWS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Streams are currently only supported for Windows"
        );
    }

    return $class->file_close(
        file_name => "$file_name:$stream_name",
        %args
    );
} ## end sub stream_close

=head2 rpc_session_refresh

Open a new RPC session for this client. This method should be used if and only
if any of the client-side operations need to be executed within a subtest. In
this case, this method should be called at the very start of the subtest code
block (otherwise any client-side action within the subtest will hung). The
subtest code block should also contain, at the very end a call to the
rpc_session_close method, to close this subtest-owned RPC session (otherwise
the subtest execution will result in a test warning).

=head3 ARGUMENTS

None.

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	my $subtest = subtest(
	    sub {
		$cache->{client}->rpc_session_refresh();
		
		# Custom subtest actions...
		
		$cache->{client}->rpc_session_close();
	    },
	    '-runid'  => "action_description",
	    '-bg',
	);
	
	# Other script actions...
	
	$subtest->wait_finish();

=cut

sub rpc_session_refresh {
    my $class = shift;
    my %args = validate(@_, {});

    $class->{_file_operations} = Protocol::ClientFileOperations->new(
        client => $class->{_nacl_client});

    return FCACHE_RC_SUCCESS;
} ## end sub rpc_session_refresh

=head2 rpc_session_close

Close the RPC session for this client. This method should be used if and only
if any of the client-side operations need to be executed within a subtest,
along with the rpc_session_refresh method. In this case, this method should be
called at the very end of the subtest code block, to close the subtest-owned
RPC session that was opened by rpc_session_refresh (otherwise the subtest 
execution will result in a test warning).

=head3 ARGUMENTS

None.

=head3 RETURN

        FCACHE_RC_SUCCESS 

=head3 EXAMPLE

	my $subtest = subtest(
	    sub {
		$cache->{client}->rpc_session_refresh();
		
		# Custom subtest actions...
		
		$cache->{client}->rpc_session_close();
	    },
	    '-runid'  => "action_description",
	    '-bg',
	);
	
	# Other script actions...
	
	$subtest->wait_finish();

=cut

sub rpc_session_close {
    my $class = shift;

    $class->{_file_operations}->destroy();

    return FCACHE_RC_SUCCESS;
} ## end sub rpc_session_close

########################################################
# Internal methods
########################################################

# Get the file handle of file that has been opened on this client.
sub _get_file_handle {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name          => {type => SCALAR,},
            snapshot_name => {type => SCALAR | UNDEF, optional => 1,},
            instance      => {type => SCALAR | UNDEF, optional => 1,},
        }
    );

    my $file_name = $class->_ensure_absolute_path(
        name          => $args{name},
        snapshot_name => $args{snapshot_name}
    );
    my $file_instance = $args{instance};

    if (!defined($file_instance)) {
        $file_instance
            = $class->_get_file_largest_instance_number(name => $file_name);
    }

    if (!exists($class->{_opened_files}->{$file_name}->{$file_instance})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This file is not currently opened on this client:\n"
                . "File: $file_name\n"
                . "Instance: $file_instance\n"
                . "Client: "
                . $class->{_name}
        );
    }

    fcache_log_comment(message => "Returning file handle for:\n"
            . "File: $file_name\n"
            . "Instance: $file_instance\n"
            . "Client: "
            . $class->{_name})
        if ($class->{_debug});
    return $class->{_opened_files}->{$file_name}->{$file_instance};
} ## end sub _get_file_handle

# Get the largest instance number for a file that has been opened on this client.
# In case of relative path:
# Arguments: Absolute file path (Scalar).
# Return: The largest number with which this file has been opened on this client.
#         0 if this file was never opened.
sub _get_file_largest_instance_number {
    my $class = shift;
    my %args = validate(@_, {name => {type => SCALAR,},});

    my $file_name        = $args{name};
    my $largest_instance = 0;

    if (exists($class->{_opened_files}->{$file_name})) {
        fcache_log_comment(
                  message => "Looking for largest instance of opened files:\n"
                . "File: $file_name\n"
                . "Instances:\n"
                . Dumper(keys(%{$class->{_opened_files}->{$file_name}})))
            if $class->{_debug};
        foreach my $instance (keys(%{$class->{_opened_files}->{$file_name}}))
        {
            if ($instance > $largest_instance) {
                $largest_instance = $instance;
            }
        }
    }

    return $largest_instance;
} ## end sub _get_file_largest_instance_number

# Verify that a given file/dir path is an absolute path.
# In case of relative path:
# - prepend the mountpoint, if the volume is mounted on this client.
# - terminate the script execution if the volume is not mounted.
# Arguments:
# - name (Required, Scalar): file name.
# - snapshot_name (Optional, Scalar): Name of snapshot.
# Return: Absolute file/dir path (Scalar).
sub _ensure_absolute_path {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name          => {type => SCALAR,},
            snapshot_name => {type => SCALAR | UNDEF, optional => 1,},
        }
    );

    my $file_name     = $args{name};
    my $snapshot_name = $args{snapshot_name};

    my $is_absolute_path = FALSE;
    if ((      ($class->{_type} eq FCACHE_CLIENT_TYPE_WINDOWS)
            && ($file_name =~ /^\w\:/)
        )
        || (   ($class->{_type} eq FCACHE_CLIENT_TYPE_UNIX)
            && ($file_name =~ /^\//))
        ) {
        $is_absolute_path = TRUE;
    }

    if (!($is_absolute_path)) {
        if (!defined($class->{_mount_point})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message =>
                    "Cannot determine absolute path for file $file_name\n"
                    . 'Volume/share was not mounted/mapped on this client: '
                    . $class->{_name}
            );
        }
        if (defined($snapshot_name)) {
            $file_name
                = $class->{_mount_point}
                . "/.snapshot/$snapshot_name"
                . "/$file_name";
        } else {
            $file_name = $class->{_mount_point} . "/$file_name";
        }
    }

    return $file_name;
} ## end sub _ensure_absolute_path

# Executes setup on this client.
# Arguments: none.
# Return: Client  type (Scalar). Possible values: constants FCACHE_CLIENT_TYPE_UNIX or FCACHE_CLIENT_TYPE_WINDOWS.
sub _execute_setup {
    my $class = shift;
    my %args  = validate(
        @_,
        {   mount_from_ip       => {type => SCALAR,},
            mount_from_junction => {type => SCALAR,},
        }
    );

    $class->{_mount_src_ip}       = $args{mount_from_ip};
    $class->{_mount_src_junction} = $args{mount_from_junction};

    if ($class->{_auto_mount}) {
        $class->mount();
    }

    return FCACHE_RC_SUCCESS;
} ## end sub _execute_setup

# Process the output returned by the RPC executor.
# Because the executor can only pass scalar return to us,
# we are doing the following: the RPC subroutine will
# return to us output in the form of comma-separated information fields.
# In here, we then split this information and populate to a hash.
# Arguments:
# - output: Required: Scalar: The comma-separated information returned by the RPC subroutine.
# - mapping_keys: Required: Arrayref: Arrayref of key names
#                 (name for each information contained by the output, in the order in which is expected).
# - description: Required: Scalar: Short description of the output to be processed,
#                used for logging purposes.
# - allow_undefined_fields: Optional: Boolean: Flag to control
#                           if the output to be processed is allowed
#                           to miss information for any fields.
#                           Default value: FALSE.
# - results_hash: Required: Hashref: Reference to the hash to which the results should be added
#                 (keys will be the key names as specified by the mapping_keys argument).
# Return: FCACHE_RC_SUCCESS.
sub _process_RPC_output {
    my $class = shift;
    my %args  = validate(
        @_,
        {   output                 => {type => SCALAR,},
            mapping_keys           => {type => ARRAYREF,},
            description            => {type => SCALAR,},
            allow_undefined_fields => {type => SCALAR, default => FALSE,},
            results_hash           => {type => HASHREF,},
        }
    );

    my $output                 = $args{output};
    my $description            = $args{description};
    my $mapping_keys           = $args{mapping_keys};
    my $allow_undefined_fields = $args{allow_undefined_fields};
    my $results_hashref        = $args{results_hash};
    my $client_type            = $class->{_type};

    my @obtained_values = split(/,/, $output);

    my $index = 0;
    foreach my $key (@{$mapping_keys}) {
        my $value = $obtained_values[$index];
        if (!defined($value)) {
            if ($allow_undefined_fields) {
                fcache_log_comment(message => "$description\n"
                        . "Client of type: $client_type\n"
                        . "Field not available: $key");
            } else {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message     => "$description\n"
                        . "Missing information for: $key"
                );
            }
        } else {
            $results_hashref->{$key} = $value;
        }
        $index++;
    }

    return FCACHE_RC_SUCCESS;
} ## end sub _process_RPC_output

# Class destructor.
# Unless the constructor was called with perform_cleanup set to FALSE,
# it takes care of any cleanup that needs to be done on the client:
# - umounts/unmaps the volume/share
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if ($class->{_creator_pid} == $$) {
        fcache_log_comment(
            message => "CLEANUP: Executing ClientOperations cleanup for "
                . $class->{_name})
            if $class->{_debug};

        foreach my $opened_file (keys(%{$class->{_opened_files}})) {
            if (scalar(keys(%{$class->{_opened_files}->{$opened_file}}))) {
                fcache_log_comment(message =>
                        "CLEANUP: Closing all the instances of opened files for:\n"
                        . "File: $opened_file\n"
                        . "Client: $class->{_name}");
                $class->file_close(file_name => $opened_file);
            }
        }

        if (defined($class->{_mount_point})) {
            if ($class->{_perform_cleanup}) {
                fcache_log_comment(message => "CLEANUP: Umounting "
                        . $class->{_mount_point}
                        . ' from client '
                        . $class->{_name});
                $class->umount();
            } else {
                fcache_log_comment(message =>
                        "CLEANUP: Cleanup is disabled. Leaving behind the mountpoint "
                        . $class->{_mount_point}
                        . ' on client '
                        . $class->{_name});
            }
        }

        $class->{_file_operations}->destroy();
    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

