
#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Utility module for platform-independent RPC.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::ClientRPC::Agnostic

=head1 DESCRIPTION

This module provides functionality for platform-independent  RPC.
It should only be used within the FlexCache library,
as RPC functions needed by the FlexCache::ClientOperations methods.

=head1 SYNOPSIS

N/A

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl

=cut

package FlexCache::ClientRPC::Agnostic;

# Standard libs
use strict;
use warnings;
use Params::Validate qw(validate SCALAR GLOB OBJECT ARRAYREF);
use threads 'exit' => 'threads_only';
use Data::Dumper;
use File::Path qw(remove_tree);
use File::Find;
use File::Copy;
use File::Compare;

# NetApp libs
use NATE::BaseException;

# Subroutine to be remotely executed on the client
# to check a symlink.
# Some Windows restrictions apply.
# Arguments:
# - link_name: Required: Scalar: Absolute path for the link.
#(please see the FlexCache::ClientOperations::symlink_create/symlink_read documentation).
sub symlink_read {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {link_name => {type => SCALAR,},});

    my $link_name = $args{link_name};

    return readlink($link_name);
} ## end sub symlink_read

# Subroutine to be remotely executed on the client
# to truncate a file.
# Uses Perl's truncate.
# Returns truncate's return status.
# Throws an exception if truncate failed.
# Arguments:
# - file_name: Required: Scalar: Absolute path for the file.
# - size: Required: Scalar: File size.
#(please see the FlexCache::ClientOperations::file_truncate documentation).
sub file_truncate {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   file_name => {type => SCALAR,},
            size      => {type => SCALAR,},
        }
    );

    my $file_name = $args{file_name};
    my $size      = $args{size};

    my $success = truncate($file_name, $size);
    if (!$success) {
        NATE::BaseException->throw("File truncate failed: " . $!);
    }

    return $success;
} ## end sub file_truncate

# Subroutine to be remotely executed on the client
# to make a copy of a file.
# Uses File::Copy.
# Returns the empty string if success, or the error message if failure.
# Arguments:
# - file_src: Required: Scalar: Absolute path for the src file.
# - file_dst: Required: Scalar: Absolute path for the dst file.
sub file_copy {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   file_src => {type => SCALAR,},
            file_dst => {type => SCALAR,},
        }
    );

    my $src_name = $args{file_src};
    my $dst_name = $args{file_dst};

    my $success = copy($src_name, $dst_name);
    my $output = '';
    if (!$success) {
        $output = $!;
    }

    return $output;
} ## end sub file_copy

# Subroutine to be remotely executed on the client to compare two files.
# Uses File::Compare.
# Returns compare's return status.
# Arguments:
# - file1: Required: Scalar: Absolute path for the first file.
# - file2: Required: Scalar: Absolute path for the second file.
sub files_compare {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   file1 => {type => SCALAR,},
            file2 => {type => SCALAR,},
        }
    );

    my $file1 = $args{file1};
    my $file2 = $args{file2};

    my $result = compare($file1, $file2);
    if (($result != 1) && ($result != 0)) {
        NATE::BaseException->throw("Files compare failed: " . $!);
    }

    return $result;
} ## end sub files_compare

# Subroutine to be remotely executed on the client
# to get the stats of a file or directory.
# Uses Perl's stat.
# Arguments:
# - path_name: Required: Scalar: Absolute path for the file/dir.
# Returns a scalar with comma-separated fields of the stat output.
sub path_stat {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {path_name => {type => SCALAR,},});

    my $path_name = $args{path_name};

    my @stats = stat($path_name);

    # The Protocol library is using ClientRPC.pm.
    # This only passes scalar returns back to the parent.
    my $return_string = join(',', @stats);
    if (!$return_string) {
        NATE::BaseException->throw("Stat failed: " . $!);
    }

    return $return_string;
} ## end sub path_stat

# Subroutine to be remotely executed on the client
# to list the contents of a directory.
# Arguments:
# - directory_name: Required: Scalar: Absolute path for the directory.
# Returns a scalar with comma-separated directory contents.
sub directory_list {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {directory_name => {type => SCALAR,},});

    my $directory_name = $args{directory_name};

    # The Protocol library is using ClientRPC.pm.
    # This only passes scalar returns back to the parent.
    my $return_string = '';

    my $success = opendir(my $DIR, $directory_name);
    if (!$success) {
        NATE::BaseException->throw("Directory open failed: " . $!);
    }

    while (my $content = readdir($DIR)) {
        next if ($content eq ".." || $content eq ".");
        $return_string .= "$content,";
    }

    closedir($DIR);

    $return_string =~ s/,$//;

    return $return_string;
} ## end sub directory_list

# Subroutine to be remotely executed on the client
# to check if a directory exists.
# Arguments:
# - directory_name: Required: Scalar: Absolute path for the directory.
# Returns 0 or 1.
sub directory_exists {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {directory_name => {type => SCALAR,},});

    my $directory_name = $args{directory_name};

    if ((-e $directory_name) && (-d $directory_name)) {
        return 1;
    } else {
        return 0;
    }
} ## end sub directory_exists

# Subroutine to be remotely executed on the client
# to remove a directory tree.
# Arguments:
# - directory_name: Required: Scalar: Absolute path for the directory.
# Returns the success status.
sub directory_tree_remove {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {directory_name => {type => SCALAR,},});

    my $directory_name = $args{directory_name};

    my $success = remove_tree($directory_name, {safe => 1},);
    if (!$success) {
        NATE::BaseException->throw("Directory delete failed: " . $!);
    }

    return $success;
} ## end sub directory_tree_remove

# Subroutine to be remotely executed on the client
# to recursively list a dataset tree.
# Arguments:
# - dataset_name: Required: Scalar: Absolute path for the dataset tree.
# Returns a comma-separated list of absolute-path files.
sub dataset_list_files {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {dataset_name => {type => SCALAR,},});

    my $dataset_name = $args{dataset_name};
    my @entries      = ();

    if (!(-d $dataset_name)) {
        NATE::BaseException->throw(
            "Dataset is not a directory:\n" . "$dataset_name\n" . $!);
    }

    find(sub { push(@entries, $File::Find::name) if -f; }, "$dataset_name");

    # The Protocol library is using ClientRPC.pm.
    # This only passes scalar returns back to the parent.
    my $file_list = join(',', @entries);

    return $file_list;
} ## end sub dataset_list_files

1;

