#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/ClientRPC/Linux.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Utility module for Linux clients RPC.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::ClientRPC::Linux

=head1 DESCRIPTION

This module provides functionality for Linux clients RPC.
It should only be used within the FlexCache library,
as RPC functions needed by the FlexCache::ClientOperations methods.

=head1 SYNOPSIS

N/A

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl

=cut

package FlexCache::ClientRPC::Linux;

# Standard libs
use strict;
use warnings;
use Params::Validate qw(validate SCALAR GLOB OBJECT ARRAYREF);
use threads 'exit' => 'threads_only';
use POSIX qw(mkfifo);

# NetApp libs
use NATE::BaseException;

# Subroutine to be remotely executed on the client
# to open a named pipe.
# Uses Perl's mkfifo.
# Returns mkfifo's return status.
# Throws an error if mkfifo failed.
# (please see the FlexCache::ClientOperations::pipe_create documentation
# for input information).
sub pipe_create {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   pipe_name  => {type => SCALAR,},
            pipe_mode  => {type => SCALAR,},
            pipe_umask => {type => SCALAR,},
        }
    );

    my $pipe_name  = $args{pipe_name};
    my $pipe_mode  = $args{pipe_mode};
    my $pipe_umask = $args{pipe_umask};
    my $original_umask;

    if (defined($pipe_umask)) {
        $original_umask = umask($pipe_umask);
    }

    my $success = mkfifo($pipe_name, $pipe_mode);

    if (defined($pipe_umask)) {
        umask($original_umask);
    }

    if (!$success) {
        NATE::BaseException->throw("Fifo create failed: " . $!);
    }

    return $success;
} ## end sub pipe_create

# Subroutine to be remotely executed on the client
# to change the mode (type and permissions) for a file or directory.
# Uses Perl's chmod.
# Returns chmod's return status.
# Throws an error if chmod failed.
# Arguments:
# - path_name: Scalar: Required: Fully-qualified path.
# - mode: Scalar: Required: New file mode to be set.
sub set_path_mode {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   path_name => {type => SCALAR,},
            mode      => {type => SCALAR,},
        }
    );

    my $path_name = $args{path_name};
    my $mode      = $args{mode};

    if ((~$mode & $mode) ne '0') {

        # Mode was passed in as an octal string,
        # convert it to an octal number.
        $mode = oct($mode);
    }

    my $count_changed = chmod($mode, $path_name);

    if (!$count_changed) {
        NATE::BaseException->throw("Chmod failed: " . $!);
    }

    return $count_changed;
} ## end sub set_path_mode

# Subroutine to be remotely executed on the client
# to change the owner (uid/gid) of a file or directory.
# Uses Perl's chown.
# Returns chown's return status.
# Throws an error if chown failed.
# Arguments:
# - path_name: Scalar: Required: Fully-qualified path.
# - uid: Scalar: Required: New file uid to be set.
# - gid: Scalar: Required: New file gid to be set.
sub set_path_owner {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   path_name => {type => SCALAR,},
            uid       => {type => SCALAR,},
            gid       => {type => SCALAR,},
        }
    );

    my $path_name = $args{path_name};
    my $uid       = $args{uid};
    my $gid       = $args{gid};

    my $count_changed = chown($uid, $gid, $path_name);

    if (!$count_changed) {
        NATE::BaseException->throw("Chown failed: " . $!);
    }

    return $count_changed;

} ## end sub set_path_owner

1;

