#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/ClientRPC/Windows.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Utility module for Windows clients RPC.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::ClientRPC::Windows

=head1 DESCRIPTION

This module provides functionality for Windows clients RPC.
It should only be used within the FlexCache library,
as RPC functions needed by the FlexCache::ClientOperations methods.

=head1 SYNOPSIS

N/A

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl

=cut

package FlexCache::ClientRPC::Windows;

# Standard libs
use strict;
use warnings;
use Params::Validate qw(validate SCALAR GLOB OBJECT ARRAYREF);
use threads 'exit' => 'threads_only';

# NetApp libs
use NATE::BaseException;
use Win32::API;
use Win32::File;
use Win32API::File qw(
    :Func :Misc :FILE_ :FILE_FLAG_ :FILE_SHARE_ :FILE_ATTRIBUTE_ :GENERIC_ );
use constant KERNEL32 => 'C:\WINDOWS\system32\kernel32.dll';

# Subroutine to be remotely executed on the client
# to obtain file/dir attributes.
# Returns: Scalar: Comma-separated list of attributes,
# in the following order: archive, compressed,
# encrypted, hidden, sparse.
# Terminates script execution if obtaining the attributes fails.
# Arguments:
# - path_name: Scalar: Required: Fully-qualified path.
sub get_windows_path_attributes {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(@_, {path_name => {type => SCALAR,},});

    my $path_name = $args{path_name};

    my $attributes = Win32API::File::GetFileAttributes($path_name);
    if (!defined($attributes)) {
        NATE::BaseException->throw(
            "Failed to obtain attributes " . "for $path_name: " . $!);
    }

    #Now start to construct the comma-separated return string
    my $attributes_str = '';

    my $info_bit
        = ($attributes & Win32API::File::FILE_ATTRIBUTE_ARCHIVE()) ? 1 : 0;
    $attributes_str .= "$info_bit,";
    $info_bit
        = ($attributes & Win32API::File::FILE_ATTRIBUTE_COMPRESSED()) ? 1 : 0;
    $attributes_str .= "$info_bit,";
    $info_bit
        = ($attributes & Win32API::File::FILE_ATTRIBUTE_ENCRYPTED()) ? 1 : 0;
    $attributes_str .= "$info_bit,";
    $info_bit
        = ($attributes & Win32API::File::FILE_ATTRIBUTE_HIDDEN()) ? 1 : 0;
    $attributes_str .= "$info_bit,";
    $info_bit
        = ($attributes & Win32API::File::FILE_ATTRIBUTE_SPARSE_FILE())
        ? 1
        : 0;
    $attributes_str .= "$info_bit,";

    return $attributes_str;
} ## end sub get_windows_path_attributes

# Subroutine to be remotely executed on the client
# to modify attributes of a file or directory.
# Returns: Scalar: Success status.
# Throws an error if changing the attributes fails.
# Arguments:
# - path_name: Scalar: Required: Fully-qualified path.
# - archive: New value to set for the archive bit.
# - hidden: New value to set for the hidden bit.
sub set_windows_path_attributes {
    shift;    # Skip the package name (see p4 change 1771691)
    my %args = validate(
        @_,
        {   path_name => {type => SCALAR,},
            archive   => {type => SCALAR, optional => 1,},
            hidden    => {type => SCALAR, optional => 1,},
        }
    );

    my $path_name = $args{path_name};

    # Get the current attributes, to only modify the pertinent bit
    my $attributes = Win32API::File::GetFileAttributes($path_name);

    if (defined($args{archive})) {
        my $archive_bit_to_set = $args{archive};
        if ($archive_bit_to_set) {
            $attributes
                = $attributes | Win32API::File::FILE_ATTRIBUTE_ARCHIVE();
        } else {
            $attributes
                = $attributes & ~Win32API::File::FILE_ATTRIBUTE_ARCHIVE();
        }
    }

    if (defined($args{hidden})) {
        my $hidden_bit_to_set = $args{hidden};
        if ($hidden_bit_to_set) {
            $attributes
                = $attributes | Win32API::File::FILE_ATTRIBUTE_HIDDEN();
        } else {
            $attributes
                = $attributes & ~Win32API::File::FILE_ATTRIBUTE_HIDDEN();
        }
    }

    my $success = Win32::File::SetAttributes($path_name, $attributes);

    if (!$success) {
        NATE::BaseException->throw(
            "Failed to modify attributes " . "for $path_name: " . $!);
    }

    return $success;
} ## end sub set_windows_path_attributes

1;

