#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Clone.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Module providing complete functionality for flex-clones.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Cache

=head1 DESCRIPTION

This module provides release and mode agnostic clone functionality for all the 
flexcache scripts. It is responsible for performing all the steps that are 
needed to configure a clone volume. The module is release agnostic and provides 
functionality for both 7-mode and C-mode. It is NACL-compliant. To be noted, 
though, that no direct access to any NACL object is given. This is to ensure 
the encapsulation that an object-oriented design must follow and to not break
the cleanup mechanism provided by the FlexCache library. A FlexCache::Clone 
object will have the following members, to be used by scripts:

=over

=item node

This member is responsible for the node-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Node.

=item aggregate

This member is responsible for the aggregate-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Aggregate.

=item volume

This member is responsible for the volume-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::CacheVolume.

=item vserver

This member is responsible for the vserver/vfiler level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Vserver.

=item client

This member is responsible for the client-side functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Client.

=back

The test cleanup will be executed by the library, within the destructor of each
object that was instantiated by a script. A destructor will be responsible for
 cleaning up after the actions that its respective object executed.

=head1 SYNOPSIS
	
	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Clone;
	
	my $parent = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
        
        my $clone = FlexCache::Cache->new(name => $VOL_DST,
                                          client => {name => $CLIENT_DST,},
                                          parent => $parent,);

=head1 UNIT TESTING

Unit tests for this module can be found under:

        flexcache/library_unit_tests/

=cut

package FlexCache::Clone;

# Standard libs
use strict;
use warnings;
use Data::Dumper;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

# NetApp libs
use base qw(FlexCache::Origin);
use NACL::STask::Node;
use FlexCache::Constants qw(:all);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use FlexCache::Node;
use FlexCache::Aggregate;
use FlexCache::ClientOperations;
use FlexCache::Vserver;
use FlexCache::Volume;
use FlexCache::UI::OntapMessages;
use NACL::STask::VolumeClone;
use NACL::STask::Volume;

=head1 CONSTRUCTOR

=head2 new

Returns a FlexCache::Clone object, representing a clone volume, with members
that can be used for basic testing. The constructor also takes care (unless
otherwise required by the caller) of the basic setup steps needed for testing.
Please see the constructor documentation for each of the class members for
detailed descriptions of the setup steps.

=head3 ARGUMENTS

=over

=item B<name> I<Optional> Scalar:

        Name for the new clone. In case this is not defined, the new clone
        will have the form <parent_name>_clone.

=item B<client> I<Required> Hashref:

        Constructor options for the client member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::ClientOperations.

=item B<parent> I<Required> Object:

        FlexCache::Origin object corresponding to the parent of this clone.

=item B<debug> I<Optional> Boolean:

        Flag to increase the verbosity of the library, useful when debugging.
        Default value: The "debug" settings of the origin object.

=item B<perform_cleanup> I<Optional> Boolean:

        Flag to control if the cleanup should be executed.
        Default value: The "perform_cleanup" settings of the origin object.

=back

=head3 EXAMPLE

        my $parent = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
        
	my $clone = FlexCache::Cache->new(name => $VOL_DST,
                                          client => {name => $CLIENT_DST,},
                                          parent => $parent,);

=cut

sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name => {type => SCALAR | UNDEF, optional => 1,},
            parent => {type => OBJECT, isa => "FlexCache::Origin"},
            snapshot => {type => SCALAR,  optional => 1,},    # TODO: document
            client   => {type => HASHREF,},
            debug    => {type => BOOLEAN, optional => 1,},
            perform_cleanup => {type => BOOLEAN, optional => 1,},
        }
    );

    my $self = {};

    my $parent = $args{parent};
    $self->{_debug}
        = (defined($args{debug})) ? $args{debug} : $parent->{_debug};
    $self->{_perform_cleanup}
        = (defined($args{perform_cleanup}))
        ? $args{perform_cleanup}
        : $parent->{_perform_cleanup};
    $self->{_attach} = FALSE;

    bless($self, $class);

    # WARNING!!!!
    # A FlexCache::Clone object is inheriting from FlexCache::Origin
    # For this to work correctly,
    # make sure that all the class members of the parent class
    # are set correctly here as well

    # Connect to the node. Create the NACL node object here .
    $self->{_node_name} = $parent->{_node_name};
    $self->{_nacl_node} = NACL::STask::Node->new(node => $self->{_node_name});

    # Connect to the client (for now, just to find out the client type)
    # We need to do this as a two-step process such that the vserver setup
    # is done correctly
    # before mounting/mapping the volume/share
    $self->{client} = FlexCache::ClientOperations->new(
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        _postpone_setup => TRUE,
        %{$args{client}}
    );
    my $client_type = $self->{client}->get_type();

    # Create the node member and process any needed information
    $self->{node} = FlexCache::Node->new(
        node            => $self->{_nacl_node},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
    );
    $self->{_boot_mode} = $self->{node}->get_boot_mode();
    $self->{_version}   = $self->{node}->_get_version();

    $self->{_ui_messages} = FlexCache::UI::OntapMessages->new(
        boot_mode => $self->{_boot_mode},
        release   => $self->{_version},
        debug     => $self->{_debug},
    );
    $self->{node}->_attach_ui_member(ui_member => $self->{_ui_messages});

    # Reuse the aggregate
    my $aggr_name = $parent->{aggregate}->get_name();
    $self->{aggregate} = FlexCache::Aggregate->new(
        node            => $self->{_nacl_node},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        attach_only     => TRUE,
        ui_messages     => $self->{_ui_messages},
        name            => $aggr_name
    );

    # Reuse the vserver
    my $vserver_name = $parent->{vserver}->get_name(_internal_use => TRUE,);
    $self->{vserver} = FlexCache::Vserver->new(
        node            => $self->{_nacl_node},
        aggregate_name  => $aggr_name,
        client_type     => $client_type,
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        attach_only     => TRUE,
        ui_messages     => $self->{_ui_messages},
        name            => $vserver_name,
    );

    # Create the clone volume
    my $parent_name = $parent->{volume}->get_name();
    my $clone_name
        = (defined($args{name})) ? ($args{name}) : ($parent_name . '_clone');
    my %create_args = (
        command_interface => $self->{_nacl_node},
        flexclone         => $clone_name,
        vserver           => $vserver_name,
        'parent-volume'   => $parent_name,
        nacltask_wait     => 1,
        'junction-path'   => '/' . $clone_name,
    );
    if (defined($args{snapshot})) {
        $create_args{'parent-snapshot'} = $args{snapshot};
    }

    # 7-mode behavior not investigated yet
    elsif ($self->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
        my $parent_type = $parent->{volume}->get_type();
        if ($parent_type eq FCACHE_VOL_TYPE_DP) {

            # For DP volumes, we need a snapshot to create the clone from.
            # Best effort-try to find a snapmirror-generated snapshot
            fcache_log_comment(
                      message => "Creating a clone for a DP volume.\n"
                    . "No user-provided snapshot to use when cloning.\n"
                    . "Attempting to locate a snapmirror snapshot");
            my @snaps = $parent->{volume}->snapshot_list();
            foreach my $snap (@snaps) {
                if ($snap =~ /^snapmirror\./) {
                    $create_args{'parent-snapshot'} = $snap;
                    last;
                }
            }
        }
    }

    NACL::STask::VolumeClone->create(%create_args);

    $self->{volume} = FlexCache::Volume->new(
        name            => $clone_name,
        node            => $self->{_nacl_node},
        aggregate_name  => $aggr_name,
        vserver_name    => $vserver_name,
        client_type     => $client_type,
        attach_only     => TRUE,
        ui_messages     => $self->{_ui_messages},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
    );

    # If the client is windows, we need to ensure that we have a share
    # available on the origin (may be the case if
    # the origin was configured for NFS only)
    if ($client_type eq FCACHE_CLIENT_TYPE_WINDOWS) {
        my $share_name = $parent->{volume}
            ->_ensure_volume_has_share(reuse_action => 'reuse');
        $self->{volume}->_ensure_volume_has_share(share_name => $share_name);
    }

    # Now we can mount/map the volume/share
    my $dlif_ip        = $self->{vserver}->get_dlif_ip();
    my $mount_junction = $self->{volume}->get_client_mount_junction();
    $self->{client}->_execute_setup(
        mount_from_ip       => $dlif_ip,
        mount_from_junction => $mount_junction,
    );

    return $self;

} ## end sub new

=head1 OTHER METHODS

This class inherits from FlexCache::Origin.
All the methods documented there are also available for FlexCache::Clone object.

=cut

########################################################
# Internal methods
########################################################

# Class destructor
# This top-level destructor is only to ensure the correct order of cleanup
# steps. For details on each member cleanup, please see the appropriate
# module documentation.
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    fcache_log_comment(
        message => "CLEANUP: Entering FlexCache cleanup for clone...");

    # Enforce the correct order of member cleanup
    $class->{client} = undef;
    $class->{volume}->purge();
    $class->{volume}    = undef;
    $class->{vserver}   = undef;
    $class->{aggregate} = undef;
    $class->{node}      = undef;

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

