#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Node.pm#3 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Agnostic module for node-level functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Node

=head1 DESCRIPTION

This module provides release and mode agnostic node-level functionality for all the flexcache scripts.
Its functionality is made available through the 'node' member of the FlexCache::Origin object.
The module is release agnostic and provides functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though, that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

=head1 SYNOPSIS

This module is intended to be used only through the FlexCache::Origin objects,
and/or through the FlexCache::Cache objects, as follows:

	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Cache;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
	$origin->{node}->call_any_method_but_the_constructor();
        
        my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
                                          aggregate => {name => $AGGR_DST,
                                                        block_type => $AGGR_DST_BLOCK_TYPE,},
                                          volume => {name => $VOL_DST,},
                                          vserver => {name => $VSERVER,
                                                      dlif_name => $DLIF_DST,
                                                      dlif_id => $DLIF_DST_ID,},
                                          client => {name => $CLIENT_DST,},
                                          origin => $origin);
        $cache->{node}->call_any_method_but_the_constructor();
	
Using this module independently is possible, with the risk that the constructor's arguments may change at any time.

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl

=cut

package FlexCache::Node;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate validate_with SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;
use Net::Ping;

# NetApp libs
use NATE::BaseException qw(:try);
use FlexCache::Constants qw(:all);
use FlexCache::Utils
    qw(fcache_log_comment fcache_log_result fcache_compare_hashes);
use FlexCache::UI::License
    qw(fcache_get_licensing_info _fcache_get_feature_mapping);
use NACL::STask::Node;
use NACL::STask::SystemLicense;

# We are only using STask::StorageFailover modules for enable/disable/wait_for_state
# (because they are only available at the task layer).
# Its methods require objects for both local and partner nodes (too costly).
# For the rest of SFO actions, we are using the component module.
use NACL::STask::StorageFailover;
use NACL::C::StorageFailover;
use NACL::CS::ClusterIdentity;
use NACL::C::Statistics;
use NACL::STask::Fiji;
use NACL::CS::AntivirusEngine;
use NACL::C::QosPolicyGroup;
use NACL::C::JobScheduleInterval;
use NACL::C::SystemCapability;
use NACL::Transit;
use dblade::wafl::tools::panamax::Version qw(get_version_number);

=head1 CONSTRUCTOR

=head2 new

Constructor for the FlexCache::Node class.
It connects to the node and gathers build information (which will be printed in the script logs).
Depending on the values of its arguments, may perform extra configuration steps.

=head3 ARGUMENTS

=over

=item B<required_boot_mode> I<Optional> Scalar:

        Required ONTAP boot mode for this node.
        If this argument is passed in and the node is running in a different mode,
        then execution is terminated, with NA result.
        Legal values: FCACHE_ONTAP_7MODE, FCACHE_ONTAP_CMODE (defined in FlexCache::Constants).
        Default value: none.

=item B<require_debug_build> I<Optional> Boolen:

        Flag to require that the node is running a debug build.
        If this argument is TRUE but the current build is not debug,
        then execution is terminated, with NA result.
        If the argument is FALSE, then no check is performed.
        Default value: FALSE.

=item B<license_ready> I<Optional> Boolen:

        Flag to determine if any required flexcache licenses should be installed on this node.
        Please note that any licenses added as part of setup will not be removed at cleanup time.
        This is to avoid any issues created to recycled resources.
        Default value: TRUE.

=item B<require_sfo_enabled> I<Optional> Boolen:

        Flag to determine Storage Failover should be enabled.
        For 7-mode, it is assumed that CF is already licensed.
        This flag is required to be set to TRUE by any script
        that needs to perform any failover actions.
        Default value: FALSE.

=back

=head3 RETURN

FlexCache::Node object.

=head3 EXAMPLES

	# Create a flexcache origin with the default settings for the node member
        # (the same arguments are also accepted by the FlexCache::Cache constructor):
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
	                                   ....);
	# Create a flexcache origin, requiring a C-mode filer:
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,
	                                             required_boot_mode => FCACHE_ONTAP_CMODE,},
					   ....);
	# Create a flexcache origin, requiring a debug build:
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,
	                                             require_debug_build => TRUE,},
					   ....);

=cut

# "Somehow internal" arguments (to limit the use of this constructor to the FlexCache library.
#                               These arguments are not documented with POD, to not confuse the script developers):
# - node: Required: Object: NACL::STask::Node object for this node.
# - debug: Optional: Boolean: Flag to increase the verbosity of this module, useful for debugging purposes.
#          Default value: FALSE.
# - perform_cleanup: Optional: Boolean: Flag to control if the cleanup should be executed.
#                    Default value: TRUE.
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   node => {type => OBJECT, isa => "NACL::STask::Node",},
            required_boot_mode  => {type => SCALAR,  optional => 1,},
            require_debug_build => {type => BOOLEAN, default  => FALSE,},
            license_ready       => {type => BOOLEAN, default  => TRUE,},
            require_sfo_enabled => {type => BOOLEAN, default  => FALSE,},
            require_antivirus_engine_on =>
                {type => BOOLEAN, default => FALSE,},
            debug           => {type => BOOLEAN, default => FALSE,},
            perform_cleanup => {type => BOOLEAN, default => TRUE,},
        }
    );

    my $self = {};

    $self->{_nacl_node} = $args{node};
    my $boot_mode = $self->{_nacl_node}->mode();
    $self->{_boot_mode}                  = $boot_mode;
    $self->{_debug}                      = $args{debug};
    $self->{_perform_cleanup}            = $args{perform_cleanup};
    $self->{_creator_pid}                = $$;
    $self->{_cleanup_licenses}           = {};
    $self->{_cleanup_fiji_points}        = {};
    $self->{_cleanup_qos_policy_groups}  = {};
    $self->{_cleanup_schedule_intervals} = {};
    $self->{_cleanup_capabilities}       = {};
    $self->{_cleanup_flags}              = {};
    $self->{_cleanup_sfo}                = FALSE;
    $self->{_cleanup_sfo_takeover}       = FALSE;

    bless($self, $class);

    $self->{_name}               = $self->get_name();
    $self->{_version}            = get_version_number($self->{_nacl_node});
    $self->{_sfo_was_configured} = FALSE;

    # Get some build information and print it out
    my $api_set    = $self->{_nacl_node}->get_7m_or_nodescope_apiset();
    my $vm         = $api_set->get_version_manager();
    my $build_type = $vm->get_version_attribute(attribute => 'build_options');
    my $build_path = $vm->get_version_attribute(attribute => 'build_path');

    fcache_log_comment(message => 'SETUP: node '
            . $self->{_name} . ":\n"
            . "ONTAP boot mode: $boot_mode\n"
            . 'ONTAP version: '
            . $self->{_version} . "\n"
            . "Build type: $build_type\n"
            . "Build path: $build_path");

    if (   (defined($args{required_boot_mode}))
        && ($boot_mode ne $args{required_boot_mode})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_CONFIG,
            message     => "Script requires boot mode: "
                . $args{required_boot_mode}
                . ". Node's current ONTAP boot mode is $boot_mode."
        );
    }

    if (($args{require_debug_build}) && !($vm->is_debug)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Script requires DEBUG build. "
                . "Node's current build type is $build_type."
        );
    }

    if ($args{license_ready}) {
        my @required_licenses = fcache_get_licensing_info(
            version   => $self->{_version},
            boot_mode => $boot_mode,
        );

        $self->license_add(_features => \@required_licenses);
    }

    if ($args{require_antivirus_engine_on}) {
        my @engines;
        try {
            @engines = NACL::CS::AntivirusEngine->fetch(
                command_interface => $self->{_nacl_node},
                requested_fields  => ['state', 'vendor'],
            );
        }
        catch NATE::BaseException with {
            my $exception = shift;
            my $output    = $exception->text();
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_CONFIG,
                message => "Script requires an antivirus engine to be on.\n"
                    . "No antivirus engine could be detected:\n"
                    . $output
            );
        };

        my $found_engine_on = FALSE;
        foreach my $engine (@engines) {
            if ($engine->state() eq 'on') {
                my $vendor = $engine->vendor();
                fcache_log_comment(message => "Detected antivirus engine:\n"
                        . "Vendor: $vendor");
                $found_engine_on = TRUE;
                last;
            }
        }

        if ($found_engine_on == FALSE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_CONFIG,
                message => "Script requires an antivirus engine to be on.\n"
                    . "No running antivirus engine could be detected:\n"
                    . Dumper(@engines)
            );
        }
    }

    if ($args{require_sfo_enabled}) {
        my $sfo = NACL::STask::StorageFailover->new(
            command_interface => $self->{_nacl_node},
            node              => $self->{_name}
        );
        my $sfo_state = $sfo->state();
        my $enabled   = $sfo_state->enabled();

        # Aparently enabled returns undef in the case when cf is not licensed
        if (!defined($enabled)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_CONFIG,
                message     => "Script requires Storage Failover.\n"
                    . "Could not obtain the failover state.\n"
                    . "Possible reasons (please address them manually):\n"
                    . "- This is a 7-mode node and cf is not licensed"
            );
        }

        if ($enabled ne 'true') {
            fcache_log_comment(
                message => 'SETUP: Enabling Storage Failover for '
                    . $self->{_name});
            $sfo->enable(command_interface => $self->{_nacl_node},);
            $self->{_cleanup_sfo} = TRUE;
        }

        $sfo_state = $sfo->state();
        $self->{_sfo_partner_name} = $sfo_state->partner_name();

        #Note: this was not tested with a more-than-2-nodes cluster
        if (!defined($self->{_sfo_partner_name})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_CONFIG,
                message =>
                    "Script requires Storage Failover. Could not obtain any partner name"
            );
        }

        fcache_log_comment(
                  message => 'SETUP: Storage Failover is configured for '
                . $self->{_name} . "\n"
                . "Partner name: "
                . $self->{_sfo_partner_name});
        $self->{_sfo_was_configured} = TRUE;
    }

    fcache_log_comment(
        message => "SETUP: Completed creation of new Node object for system "
            . $self->{_name})
        if ($self->{_debug});
    return $self;
} ## end sub new

=head1 OTHER METHODS

=head2 get_name

Method to obtain the name of this node.

=head3 ARGUMENTS

None.

=head3 RETURN

Name of the node/filer (Scalar).

EXAMPLE

	my $node_name = $obj->get_name();

=cut

sub get_name {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_nacl_node}->name();
} ## end sub get_name

=head2 get_boot_mode

Method to obtain the boot mode of this node.

=head3 ARGUMENTS

None.

=head3 RETURN

Boot mode of the node/filer (Scalar).
Possible values: constants FCACHE_ONTAP_CMODE or FCACHE_ONTAP_7MODE.

EXAMPLE

	my $boot_mode = $obj->get_boot_mode();

=cut

sub get_boot_mode {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_boot_mode};
} ## end sub get_boot_mode

=head2 license_add

Add a license (if not already added).

=head3 ARGUMENTS

=over

=item B<feature> I<Required> Scalar:

        Feature to be licensed.
        Legal values: Any FCACHE_LICENSE_* constant (defined in FlexCache::Constants).

=item B<license_prerequisites> I<Optional> Boolean:

        Flag to control if any pre-requisite license for this feature should also be added.
        Default value: TRUE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 NOTES

Please note that any licenses added by a script will not be removed at cleanup time.
This is to avoid any issues created to recycled resources.

=head3 EXAMPLE

	$obj->license_add(feature => FCACHE_LICENSE_CIFS);

=cut

# Internal arguments:
# - _features: Optional: Arrayref: Feature list, already mapped into the NACL strings.
#              If this argument is present, then any other arguments are ignored
#              and only the specified features are licensed, directly using the provided feature names.
sub license_add {
    my $class = shift;
    my %args  = validate(
        @_,
        {   feature               => {type => SCALAR,   optional => 1,},
            license_prerequisites => {type => BOOLEAN,  default  => TRUE,},
            _features             => {type => ARRAYREF, optional => 1,},
        }
    );

    my @features_to_license = ();

    if (defined($args{_features})) {
        @features_to_license = @{$args{_features}};
    } else {
        if (!defined($args{feature})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Required argument is missing: feature"
            );
        }

        my $boot_mode = $class->get_boot_mode();
        @features_to_license = _fcache_get_feature_mapping(
            feature               => $args{feature},
            boot_mode             => $boot_mode,
            include_prerequisites => $args{license_prerequisites},
        );
    }

    foreach my $feature (@features_to_license) {
        my $was_added = FALSE;
        fcache_log_comment(message => "Adding license for $feature")
            if ($class->{_debug});
        NACL::STask::SystemLicense->add(
            command_interface => $class->{_nacl_node},
            feature           => $feature,
            _was_added        => \$was_added,
        );

        if ($was_added) {
            $class->{_cleanup_licenses}->{$feature} = 1;
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub license_add

=head2 license_remove

Remove a license (if not already removed).

=head3 ARGUMENTS

=over

=item B<feature> I<Required> Scalar:

        Feature to be un-licensed.
        Legal values: Any FCACHE_LICENSE_* constant (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->license_remove(feature => FCACHE_LICENSE_CIFS);

=cut

# Internal arguments:
# - _cleanup: Optional: Boolean: If this flag is set to true,
#             then the method will remove all the previously-installed licenses.
#             Default value: FALSE.
sub license_remove {
    my $class = shift;
    my %args  = validate(
        @_,
        {   feature  => {type => SCALAR,  optional => 1,},
            _cleanup => {type => BOOLEAN, default  => FALSE,},
        }
    );

    my $boot_mode = $class->get_boot_mode();
    my @features  = ();

    if ($args{_cleanup}) {
        foreach my $cleanup_features (keys(%{$class->{_cleanup_licenses}})) {
            if ($class->{_cleanup_licenses}->{$cleanup_features} == 1) {
                push(@features, $cleanup_features);
            }
        }
    } else {
        if (!defined($args{feature})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Required argument is missing: feature"
            );
        }

        @features = _fcache_get_feature_mapping(
            feature               => $args{feature},
            boot_mode             => $boot_mode,
            include_prerequisites => FALSE,
        );
    }

    foreach my $feature (@features) {
        fcache_log_comment(message => "Removing license for $feature")
            if ($class->{_debug});
        NACL::STask::SystemLicense->purge(
            command_interface  => $class->{_nacl_node},
            feature            => $feature,
            nacltask_if_purged => 'pass',
        );
        $class->{_cleanup_licenses}->{$feature} = 0;
    }

    return FCACHE_RC_SUCCESS;
} ## end sub license_remove

=head2 get_sfo_partner_name

Get the name of the SFO partner.

=head3 ARGUMENTS

None.

=head3 RETURN

The name of the SFO partner (Scalar).

EXAMPLE

	my $partner_name = $obj->get_sfo_partner_name();

=cut

sub get_sfo_partner_name {
    my $class = shift;
    my %args = validate(@_, {});

    if (!$class->{_sfo_was_configured}) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "The sfo methods should only be used for  node objects "
                . "created with the require_sfo_enabled argument set to TRUE"
        );
    }

    return $class->{_sfo_partner_name};
} ## end sub get_sfo_partner_name

=head2 sfo_takeover

Perform Storage Failover takeover.

=head3 ARGUMENTS

=over

=item B<partner_name> I<Optional> Scalar:

        Name of the partner to take over.
        Default value: partner name, as detected by the constructor.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->sfo_takeover();

=cut

sub sfo_takeover {
    my $class = shift;
    my %args  = validate(
        @_,
        {   partner_name =>
                {type => SCALAR, default => $class->{_sfo_partner_name}},
        }
    );

    if (!$class->{_sfo_was_configured}) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "The sfo methods should only be used for  node objects "
                . "created with the require_sfo_enabled argument set to TRUE"
        );
    }

    fcache_log_comment(message => "Performing SFO takeover:\n"
            . "By node: "
            . $class->{_name} . "\n"
            . "Of node: "
            . $args{partner_name})
        if ($class->{_debug});
    NACL::C::StorageFailover->takeover(
        command_interface => $class->{_nacl_node},
        ofnode            => $args{partner_name},
    );
    $class->{_cleanup_sfo_takeover} = TRUE;

    return FCACHE_RC_SUCCESS;
} ## end sub sfo_takeover

=head2 sfo_wait_for_state

Wait for specific Storage Failover state.

=head3 ARGUMENTS

=over

=item B<partner_name> I<Optional> Scalar:

        Name of the partner to take over.
        Default value: partner name, as detected in the constructor.

=item B<local_state> I<Required> Scalar:

        State in which the local node is expected to be in.
        Legal values: any FCACHE_SFO_STATE_* constant, defined in FlexCache::Constants.

=item B<partner_state> I<Required> Scalar:

        State in which the partner node is expected to be in.
        Legal values: any FCACHE_SFO_STATE_* constant, defined in FlexCache::Constants.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->sfo_wait_for_state(local_state => FCACHE_SFO_STATE_TAKEOVER,
                                 partner_state => FCACHE_SFO_STATE_WAIT_FOR_GIVEBACK);
                                 
        $obj->sfo_wait_for_state(local_state => FCACHE_SFO_STATE_UP,
                                 partner_state => FCACHE_SFO_STATE_UP);

=cut

sub sfo_wait_for_state {
    my $class = shift;
    my %args  = validate(
        @_,
        {   partner_name =>
                {type => SCALAR, default => $class->{_sfo_partner_name},},
            local_state   => {type => SCALAR,},
            partner_state => {type => SCALAR,},
        }
    );

    my $local_state   = $args{local_state};
    my $partner_state = $args{partner_state};

    if (!$class->{_sfo_was_configured}) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "The sfo methods should only be used for  node objects "
                . "created with the require_sfo_enabled argument set to TRUE"
        );
    }

    fcache_log_comment(message => "Waiting for SFO state:\n"
            . "Local node ("
            . $class->{_name}
            . "): $local_state\n"
            . "Partner node ("
            . $args{partner_name}
            . "): $partner_state")
        if ($class->{_debug});
    my $partner_node_obj
        = NACL::STask::Node->new(node => $args{partner_name});
    NACL::STask::StorageFailover->wait_for_state(
        command_interface => $class->{_nacl_node},
        node              => $class->{_nacl_node},
        partner           => $partner_node_obj,
        node_state        => $local_state,
        partner_state     => $partner_state,
        'method-timeout'  => FCACHE_TOGB_TIMEOUT,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub sfo_wait_for_state

=head2 sfo_giveback

Perform Storage Failover giveback.

=head3 ARGUMENTS

=over

=item B<partner_name> I<Optional> Scalar:

        Name of the partner to give back.
        Default value: partner name, as detected in the constructor.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->sfo_giveback();

=cut

sub sfo_giveback {
    my $class = shift;
    my %args  = validate(
        @_,
        {   partner_name =>
                {type => SCALAR, default => $class->{_sfo_partner_name}},
        }
    );

    if (!$class->{_sfo_was_configured}) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "The sfo methods should only be used for  node objects "
                . "created with the require_sfo_enabled argument set to TRUE"
        );
    }

    fcache_log_comment(message => "Performing SFO giveback:\n"
            . "From node: "
            . $class->{_name} . "\n"
            . "Of node: "
            . $args{partner_name})
        if ($class->{_debug});
    NACL::C::StorageFailover->giveback(
        command_interface => $class->{_nacl_node},
        ofnode            => $args{partner_name},
        'override-vetoes' => 'true',
    );
    $class->{_cleanup_sfo_takeover} = FALSE;

    return FCACHE_RC_SUCCESS;
} ## end sub sfo_giveback

=head2 statistics_verify

Verify performance statistics for this node.

=head3 ARGUMENTS

=over

=item B<object> I<Required> Scalar:

        Object for which to retrieve the statistics.
        Legal values: any of the FCACHE_NODE_STATISTICS_OBJ_*
        constants, defined in FlexCache::Constants.

=item B<instance> I<Optional> Scalar:

        Object instance for which to retrieve the statistics.
        Except for the case when the instance is a resource
        (e.g. node name in the case of (FCACHE_NODE_STATISTICS_OBJ_LMGR_NG),
        this argument should only be passed in as a FCACHE_NODE_STATISTICS_INST_*
        constant, defined in FlexCache::Constants.
        Default value: the node name.
        
=item B<counters> I<Optional> Hashref: 

        Hashref of pairs "counter name" => "expected value",
        to be passed in if specific counter values need to be verified.
        Unless this argument is passed in, the method does not execute any value
        check against the performance counters, but it just test that
        the statistics can be obtained for the specifc object/instance.

=back

=head3 RETURN

        A hash of pairs "counter name" => "expected value"
        obtained for this specific object/instance.

EXAMPLE

        # Obtain the lmgr_ng statistics for this node
        $obj->statistics_verify(object => FCACHE_NODE_STATISTICS_OBJ_LMGR_NG,);
        # Verify that the inavssuccess counter for the avoa object is 1
        $obj->statistics_verify(object => FCACHE_NODE_STATISTICS_OBJ_AVOA,
                                instance => FCACHE_NODE_STATISTICS_INST_AVOA,
                                counters => {inavssuccess => 1},);

=cut

sub statistics_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   object   => {type => SCALAR,},
            instance => {type => SCALAR, default => $class->{_name},},
            counters => {type => HASHREF, optional => 1,},
        }
    );

    my $counters         = delete($args{counters});
    my $object           = $args{object};
    my $is_statistics_v2 = FALSE;
    my %return_hash      = ();

    # Note: as of now, the keys of the return hash are not processed
    # (left just as returned by NACL). If this will create problems
    # in the future (caused by differences between releases),
    # then we will have to map at least the keys used by the flex-cache
    # scripts to some release-independent names.
    # For now, we are not doing this, due to the multitude of fields.
    if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
        $is_statistics_v2 = $class->{_nacl_node}
            ->is_feature_supported(feature => 'statistics_v2');
    }

    if ($is_statistics_v2) {
        fcache_log_comment(message => "Running the statistics-v1 command")
            if ($class->{_debug});

        # TODO: Revisit this once burt 605768 is fixed
        # (NACL does not have v1 SN support other than at APISet level)
        my $api_set = $class->{_nacl_node}->apiset();
        my $res     = $api_set->statistics_v1_show(
            node   => $class->{_name},
            object => $args{object}
        );
        my $parsed_output = $res->get_parsed_output();
        foreach my $counter_instance (@{$parsed_output}) {
            fcache_log_comment(message => "Processing counter instance:\n"
                    . Dumper($counter_instance))
                if ($class->{_debug});
            my $instance_name = $counter_instance->{counter};
            $return_hash{$instance_name} = $counter_instance->{value};

        }
    } else {

        # Add the required arguments for the NACL method filter
        $args{node} = $class->{_name};
        my @statistics = NACL::C::Statistics->find(
            command_interface => $class->{_nacl_node},
            filter            => \%args,
        );

        foreach my $statistic (@statistics) {
            my $state        = $statistic->state();
            my $counter_name = $state->{counter};
            $return_hash{$counter_name} = $state->{value};
        }
    }

    if (defined($counters)) {
        fcache_compare_hashes(
            obtained         => \%return_hash,
            expected         => $counters,
            hash_description => "Node statistics ($object)",
        );
    }

    fcache_log_comment(
        message => "statistics_verify returning:\n" . Dumper(\%return_hash))
        if ($class->{_debug});

    return %return_hash;
} ## end sub statistics_verify

=head2 fiji_set

Set a fiji point on this node.

=head3 ARGUMENTS

=over

=item B<tag> I<Required> Scalar:

	Tag representing the fiji point to set.
	Legal values: any of the FCACHE_FIJI_TAG_* constants,
	as defined in FlexCache::Constants.

=item B<Extra arguments>:

	Depending on which fiji point is required to be set, there will
	be extra arguments required, depending on the SW requirements for the
	specific fiji. Some of these arguments will have a default value
	and they will be considered optional.
	If any of the required arguments are missing when the method is called,
	the script execution will be terminated with ea "script" result.
	These arguments are as follows:
	FCACHE_FIJI_TAG_WI_INDIRECT_BLK_CNT_SET:
	  - volume_name: required
	  - value: optional, defaults to 0.

=back

=head3 RETURN

        The slot number under which the fiji point was set.

EXAMPLE

        my $slot_no = $obj->fiji_set(tag => FCACHE_FIJI_TAG_WI_INDIRECT_BLK_CNT_SET,
	                             volume_name => $cache_vol,);

=cut

sub fiji_set {
    my $class = shift;
    my (%args) = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {tag => {type => SCALAR,},}
    );

    my $tag          = $args{tag};
    my $fiji_info    = $class->_get_fiji_info(tag => $tag);
    my $fiji_program = $fiji_info->{program};
    my $fiji_tag     = $fiji_info->{description};
    my $fiji_params  = $fiji_info->{parameters};

    if (   (!defined($fiji_program))
        || (!defined($fiji_tag))
        || (!defined($fiji_params))) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Internal library problem:\n"
                . "Incomplete information defined for fiji $tag\n"
                . Dumper($fiji_info)
        );
    }

    my @nacltask_params = ();
    foreach my $fiji_param_info (@{$fiji_params}) {
        my $param_name = $fiji_param_info->{name};
        if (!defined($param_name)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Internal library problem:\n"
                    . "Incomplete information defined for fiji $tag\n"
                    . "Missing parameter name:\n"
                    . Dumper($fiji_param_info)
            );
        }

        #Now figure out what the value of thie fiji param should be
        my $value;
        if (defined($args{$param_name})) {
            $value = $args{$param_name};
        } else {
            if (defined($fiji_param_info->{default})) {
                $value = $fiji_param_info->{default};
            } else {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Fiji point $tag requires the following "
                        . "argument to be passed in: $param_name"
                );
            }
        }

        if (defined($fiji_param_info->{regex})) {
            my $regex = $fiji_param_info->{regex};
            if ($value !~ /$regex/) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Fiji point argument regex mismatch:\n"
                        . "Fiji tag: $tag\n"
                        . "Argument: $param_name\n"
                        . "Must match regex: $regex"
                );
            }
        }
        push(@nacltask_params, $value);
    }

    my $fiji = NACL::STask::Fiji->new(
        command_interface => $class->{_nacl_node},
        tag               => $fiji_tag,
        program           => $fiji_program,
    );
    $fiji->set(nacltask_param => \@nacltask_params,);
    my $fiji_slot = $fiji->{slot};
    $class->{_cleanup_fiji_points}->{$fiji_slot} = 1;

    fcache_log_comment(message => "Fiji point set:\n"
            . "Fiji: $fiji_tag\n"
            . "Slot: $fiji_slot");

    return $fiji_slot;
} ## end sub fiji_set

=head2 fiji_clear

Clear a fiji point on this node.

=head3 ARGUMENTS

=over

=item B<slot> I<Required> Scalar:

	Slot number for the fiji to be cleared.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->fiji_clear(slot => 5);

=cut

sub fiji_clear {
    my $class = shift;
    my %args = validate(@_, {slot => {type => SCALAR,},});

    my $fiji_slot = $args{slot};

    NACL::STask::Fiji->clear(
        command_interface => $class->{_nacl_node},
        slot              => $fiji_slot,
    );
    delete($class->{_cleanup_fiji_points}->{$fiji_slot});

    return FCACHE_RC_SUCCESS;
} ## end sub fiji_clear

=head2 qos_policy_group_create

Create a Qos policy group.
This method is not applicable for 7-mode.

=head3 ARGUMENTS

=over

=item B<policy_group> I<Required> Scalar:

	Name of the Qos policy group to create.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->qos_policy_group_create(policy_group => 'my_qos_policy_group');

=cut

sub qos_policy_group_create {
    my $class = shift;
    my %args = validate(@_, {policy_group => {type => SCALAR,},});

    my $policy_group = $args{policy_group};

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Qos policy groups are not available in 7-mode",
        );
    }

    NACL::C::QosPolicyGroup->create(
        command_interface => $class->{_nacl_node},
        'policy-group'    => $policy_group,
    );
    $class->{_cleanup_qos_policy_groups}->{$policy_group} = 1;

    return FCACHE_RC_SUCCESS;
} ## end sub qos_policy_group_create

=head2 qos_policy_group_delete

Delete a Qos policy group.
This method is not applicabke for 7-mode.

=head3 ARGUMENTS

=over

=item B<policy_group> I<Required> Scalar:

	Name of the Qos policy group to delete.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->qos_policy_group_delete(policy_group => 'my_qos_policy_group');

=cut

sub qos_policy_group_delete {
    my $class = shift;
    my %args = validate(@_, {policy_group => {type => SCALAR,},});

    my $policy_group = $args{policy_group};

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Qos policy groups are not available in 7-mode",
        );
    }

    NACL::C::QosPolicyGroup->delete(
        command_interface => $class->{_nacl_node},
        'policy-group'    => $policy_group,
    );
    delete($class->{_cleanup_qos_policy_groups}->{$policy_group});

    return FCACHE_RC_SUCCESS;
} ## end sub qos_policy_group_delete

=head2 schedule_interval_create

Create a job schedule interval.
This method is not applicable for 7-mode.

=head3 ARGUMENTS

=over

=item B<interval_name> I<Required> Scalar:

	Name of the schedule interval.

=item B<seconds> I<Required> Scalar:

	Number of seconds for this schedule interval.

=item B<minutes> I<Required> Scalar:

	Number of minutes for this schedule interval.

=item B<hours> I<Required> Scalar:

	Number of hours for this schedule interval.

=item B<days> I<Required> Scalar:

	Number of days for this schedule interval.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->schedule_interval_create(interval_name => 'interval1',
	                               seconds => 0,
				       minutes => 5,
				       hours => 0,
				       days => 0);

=cut

sub schedule_interval_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   interval_name => {type => SCALAR,},
            seconds => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            minutes => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            hours   => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            days    => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
        }
    );

    my $interval_name = delete($args{interval_name});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Schedule intervals are not available in 7-mode",
        );
    }

    NACL::C::JobScheduleInterval->create(
        command_interface => $class->{_nacl_node},
        name              => $interval_name,
        %args
    );
    $class->{_cleanup_schedule_intervals}->{$interval_name} = 1;

    return FCACHE_RC_SUCCESS;
} ## end sub schedule_interval_create

=head2 schedule_interval_delete

Delete a job schedule interval.
This method is not applicable to 7-mode.

=head3 ARGUMENTS

=over

=item B<interval_name> I<Required> Scalar:

	Name of the schedule interval.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->schedule_interval_delete(interval_name => 'interval1');

=cut

sub schedule_interval_delete {
    my $class = shift;
    my %args = validate(@_, {interval_name => {type => SCALAR,},});

    my $interval_name = $args{interval_name};

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Schedule intervals are not available in 7-mode",
        );
    }

    NACL::C::JobScheduleInterval->delete(
        command_interface => $class->{_nacl_node},
        name              => $interval_name,
    );
    delete($class->{_cleanup_schedule_intervals}->{$interval_name});

    return FCACHE_RC_SUCCESS;
} ## end sub schedule_interval_delete

=head2 reboot

Reboot the filer.
This method is also waiting for the filer to boot up and is resetting the
node's connections.

=head3 ARGUMENTS

None.

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->reboot();

=cut

sub reboot {
    my $class = shift;
    my %args = validate(@_, {});

    my $transit = NACL::Transit->new(name => $class->{_name});
    $transit->reboot();
    $class->{_nacl_node}->refresh_command_interface(no_version_change => 1);

    return FCACHE_RC_SUCCESS;
} ## end sub reboot

=head2 is_pingable

Check if the node responds to ping.

=head3 ARGUMENTS

None.

=head3 RETURN

        TRUE/FALSE.

EXAMPLE

        my $is_pingable = $obj->is_pingable();
	if ($is_pingable == TRUE) { ... }

=cut

sub is_pingable {
    my $class = shift;
    my %args = validate(@_, {});

    my $hostrec = $class->{_nacl_node}->hostrec();
    my $host    = $hostrec->{default_ip};
    if (!defined($host)) {

        # Try to ping by name (will not work if dns is not set up correctly)
        $host = $class->{_name};
    }
    my $net_ping    = Net::Ping->new();
    my $is_pingable = $net_ping->ping($host);
    $net_ping->close();

    if ($is_pingable) {
        fcache_log_comment(
            message => "Node " . $class->{_name} . " responds to ping");
        return TRUE;
    } else {
        fcache_log_comment(message => "Node "
                . $class->{_name}
                . " does not respond to ping");
        return FALSE;
    }
} ## end sub is_pingable

=head2 refresh_connection

Refresh connection(s) to this node.
Useful after SFO operations.

=head3 ARGUMENTS

None.

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->refresh_connection();

=cut

sub refresh_connection {
    my $class = shift;
    my %args = validate(@_, {});

    $class->{_nacl_node}->refresh_command_interface(no_version_change => 1);

    return FCACHE_RC_SUCCESS;
} ## end sub refresh_connection

=head2 system_capability_verify

Verifies the status of a system capability for this node.
This method is only available for C-mode.

=head3 ARGUMENTS

=over

=item B<capability_name> I<Required> Scalar:

	Name of capability to verify. It can be any of the FCACHE_CAPABILITY_*
	constants (defined in FlexCache::Constants).

=item B<expected_enabled> I<Optional> Boolean:

	If this argument is passed in, then the library will also check if
	the required capability is enabled or disabled.

=back

=head3 RETURN

        Enabled state of the capability, as a boolean value.

EXAMPLE

        $obj->system_capability_verify(
	    capability_name => FCACHE_CAPABILITY_FLEXCACHE_DELEGATIONS,
	    );
	$obj->system_capability_verify(
	    capability_name => FCACHE_CAPABILITY_FLEXCACHE_DELEGATIONS,
	    expected_enabled => TRUE,
	    );

=cut

sub system_capability_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   capability_name  => {type => SCALAR,},
            expected_enabled => {type => BOOLEAN, optional => 1,},
        }
    );

    my $capability_name  = $args{capability_name};
    my $expected_enabled = $args{expected_enabled};
    my $node_name        = $class->{_name};
    if ($class->{_boot_mode} ne FCACHE_ONTAP_CMODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is only supported for C-mode"
        );
    }

    my $capability = NACL::C::SystemCapability->find(
        command_interface => $class->{_nacl_node},
        filter            => {
            node       => $node_name,
            capability => $capability_name
        }
    );
    my $state         = $capability->state();
    my $enabled_state = $state->enabled();
    my $enabled = ($enabled_state eq FCACHE_SETTING_TRUE) ? TRUE : FALSE;

    if (   (defined($expected_enabled))
        && ($expected_enabled != $enabled)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Mismatch in capability setting:\n"
                . "Node: $node_name\n"
                . "Capability: $capability_name\n"
                . "Expected enabled: $expected_enabled\n"
                . "Currently enabled: $enabled"
        );
    }

    fcache_log_comment(message => "System capability check for:\n"
            . "Node: $node_name\n"
            . "Capability: $capability_name\n"
            . "Currently enabled: $enabled");

    return $enabled;
} ## end sub system_capability_verify

=head2 system_capability_disable

Disable a system capability.
WARNING: the capability will be disabled for all the nodes in the cluster.
This method is only supported for C-mode.

=head3 ARGUMENTS

=over

=item B<capability_name> I<Required> Scalar:

	Name of capability to disable. It can be any of the FCACHE_CAPABILITY_*
	constants (defined in FlexCache::Constants).

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

EXAMPLE

        $obj->system_capability_disable(
	    capability_name => FCACHE_CAPABILITY_FLEXCACHE_DELEGATIONS);

=cut

sub system_capability_disable {
    my $class = shift;
    my %args = validate(@_, {capability_name => {type => SCALAR,},});

    my $capability_name = $args{capability_name};
    my $need_to_disable = FALSE;
    if ($class->{_boot_mode} ne FCACHE_ONTAP_CMODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is only supported for C-mode"
        );
    }

    my @enabled_capabilities = NACL::C::SystemCapability->is_enabled(
        command_interface => $class->{_nacl_node},
        'cap-in'          => [$capability_name]
    );
    fcache_log_comment(message => "Enabled system capabilities:\n"
            . Dumper(@enabled_capabilities))
        if ($class->{_debug});

    if (grep(/^$capability_name$/, @enabled_capabilities)) {
        $need_to_disable = TRUE;

        # We will need to re-enable as part of cleanup.
        if (!(exists($class->{_cleanup_capabilities}->{$capability_name}))) {
            $class->{_cleanup_capabilities}->{$capability_name} = TRUE;
        }
    }

    if ($need_to_disable) {
        NACL::C::SystemCapability->disable(
            command_interface => $class->{_nacl_node},
            capabilities      => [$capability_name]
        );
    } else {
        fcache_log_comment(message => "System capability is not enabled:\n"
                . "Capability: $capability_name");
    }

    return FCACHE_RC_SUCCESS;
} ## end sub system_capability_disable

=head2 system_capability_enable

Enable a system capability.
WARNING: the capability will be enabled for all the nodes in the cluster.
This method is only supported for C-mode.

=head3 ARGUMENTS

=over

=item B<capability_name> I<Required> Scalar:

	Name of capability to enable. It can be any of the FCACHE_CAPABILITY_*
	constants (defined in FlexCache::Constants).

=back

=head3 RETURN

        FCACHE_RC_SUCCESS

EXAMPLE

        $obj->system_capability_enable(
	    capability_name => FCACHE_CAPABILITY_FLEXCACHE_DELEGATIONS);

=cut

sub system_capability_enable {
    my $class = shift;
    my %args = validate(@_, {capability_name => {type => SCALAR,},});

    my $capability_name = $args{capability_name};
    my $need_to_enable  = FALSE;
    if ($class->{_boot_mode} ne FCACHE_ONTAP_CMODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is only supported for C-mode"
        );
    }

    my @enabled_capabilities = NACL::C::SystemCapability->is_enabled(
        command_interface => $class->{_nacl_node},
        'cap-in'          => [$capability_name]
    );
    fcache_log_comment(message => "Enabled system capabilities:\n"
            . Dumper(@enabled_capabilities))
        if ($class->{_debug});

    if (!grep(/^$capability_name$/, @enabled_capabilities)) {
        $need_to_enable = TRUE;

        # We will need to re-disable as part of cleanup.
        if (!(exists($class->{_cleanup_capabilities}->{$capability_name}))) {
            $class->{_cleanup_capabilities}->{$capability_name} = FALSE;
        }
    }

    if ($need_to_enable) {
        NACL::C::SystemCapability->enable(
            command_interface => $class->{_nacl_node},
            'cap-in'          => [$capability_name],
            'all-or-none'     => 'true'
        );
    } else {
        fcache_log_comment(message => "System capability is enabled:\n"
                . "Capability: $capability_name");
    }

    return FCACHE_RC_SUCCESS;
} ## end sub system_capability_enable

=head2 wafl_flush_cache

Flush the WAFL cache on this node.

=head3 ARGUMENTS

None.

=head3 RETURN

        FCACHE_RC_SUCCESS

EXAMPLE

        $obj->wafl_flush_cache();

=cut

sub wafl_flush_cache {
    my $class = shift;
    my %args = validate(@_, {});

    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
    $api_set->wafl_flush_cache();

    return FCACHE_RC_SUCCESS;
} ## end sub wafl_flush_cache

=head2 printflag

Get the current value of a flag on this mode.

=head3 ARGUMENTS

=over

=item B<flag> I<Required> Scalar:

	Name of flag to set. It can be any of the FCACHE_FLAG_*
	constants (defined in FlexCache::Constants).

=back

=head3 RETURN

        The flag value (Scalar).

EXAMPLE

        my $flag_value = $obj->printflag(flag => FCACHE_FLAG_FLEXCACHE_FORWARDING);

=cut

sub printflag {
    my $class = shift;
    my %args = validate(@_, {flag => {type => SCALAR,},});

    my $flag = $args{flag};
    my $value;

    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();

    # TODO: revisit after NACL burt 659672 is fixed
    #my $res = $api_set->printflag(flag => $args{flag});
    #my $parsed_output = $res->get_parsed_output();
    my $output = $api_set->execute_command(command => "printflag $flag");
    if ($output =~ /$flag\s*=\s*(\w+)/) {
        $value = $1;
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Failed to obtain value for flag $flag: $output"
        );
    }

    return $value;
} ## end sub printflag

=head2 setflag

Set the value of a flag on this mode.

=head3 ARGUMENTS

=over

=item B<flag> I<Required> Scalar:

	Name of flag to set. It can be any of the FCACHE_FLAG_*
	constants (defined in FlexCache::Constants).

=item B<value> I<Required> Scalar:

	Value to set for this flag.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->setflag(flag => FCACHE_FLAG_FLEXCACHE_FORWARDING,
	              value => 2);

=cut

sub setflag {
    my $class = shift;
    my %args  = validate(
        @_,
        {   flag  => {type => SCALAR,},
            value => {type => SCALAR,},
        }
    );

    my $flag  = $args{flag};
    my $value = $args{value};

    my $current_value = $class->printflag(flag => $flag);
    if ("$current_value" eq $value) {
        fcache_log_comment(message => "Flag $flag is already set to $value");
        return FCACHE_RC_SUCCESS;
    }

    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
    $api_set->setflag(
        flag  => $flag,
        value => $value,
    );

    # We will need to re-enable as part of cleanup.
    if (!(exists($class->{_cleanup_flags}->{$flag}))) {
        $class->{_cleanup_flags}->{$flag} = $current_value;
    }

    return FCACHE_RC_SUCCESS;
} ## end sub setflag

########################################################
# Internal methods
########################################################

# Obtain the three-digit version of this node.
# Arguments: none.
# Return: Three digit version number (Scalar). E.g.: 8.2.0
sub _get_version {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_version};
} ## end sub _get_version

# Obtain the name of the cluster containing this node.
# Arguments: none.
# Return: C-mode: cluster name (Scalar).
#         7-mode: ''.
sub _get_cluster_name {
    my $class = shift;
    my %args = validate(@_, {});

    my $cluster_name = '';

    if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
        my $cluster_cs_obj = NACL::CS::ClusterIdentity->fetch(
            command_interface => $class->{_nacl_node},
            requested_fields  => ['name'],
        );
        $cluster_name = $cluster_cs_obj->name();
    }

    return $cluster_name;
} ## end sub _get_cluster_name

# Attach the _ui_messages member to this object.
# Needed for any ONTAP output validation.
# Arguments:
# - ui_member: Required: Object: FlexCache::UI::OntapMessages instance.
# Return: None.
sub _attach_ui_member {
    my $class = shift;
    my %args  = validate(
        @_,
        {   ui_member =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
        }
    );

    $class->{_ui_messages} = $args{ui_member};
} ## end sub _attach_ui_member

# Obtain the fiji information for this tag: program, required_params.
# Arguments:
# - tag: Required: Scalar: Tag that was defined for this fiji point.
# Return: Hashref with all the pertinent information for this fiji point.
sub _get_fiji_info {
    my $class = shift;
    my %args = validate(@_, {tag => {type => SCALAR,},});

    my $tag = $args{tag};

    # Hash containing all the needed information for the fiji points
    # The keys should be the VALUES of the FCACHE_FIJI_TAG_* constants,
    # as defined in FlexCache::Constants
    my $fiji_lookup_table = {
        WI_INDIRECT_BLK_CNT_SET => {
            program =>
                "\'(wafl.iicu.set.ind.blk.cnt (eq \@vol %s) (return %d))\'",
            description => "Set the wi_indirect_blk_cnt value",
            parameters  => [
                {name => 'volume_name',},
                {   name    => 'value',
                    regex   => '^\d+$',
                    default => 0
                },
            ],
        },

        IICU_SCAN_RESTART => {
            program =>
                "\'(repeat %d wafl.iicu.scan.restart (eq \@vol %s) (return 1))\'",
            description => "Keep restarting the iicu scan",
            parameters  => [
                {   name    => 'restarts_count',
                    regex   => '^\d+$',
                    default => 50000000
                },
                {name => 'volume_name',},
            ],
        },

        RAID_RETURN_PSEUDO_BAD_ERROR => {
            program =>
                "\'(repeat %d raid.return.pseudo.bad.error (&& (&& (&& "
                . "(eq \@vol %s) (eq \@fileid %d)) (eq \@is_priv 0)) "
                . "(eq \@level %d)) (return true))\'",
            description => "RAID return pseudo_bad error",
            parameters  => [
                {   name    => 'returns_count',
                    regex   => '^\d+$',
                    default => 9999
                },

                {name => 'volume_name',},

                {name => 'file_id',},

                {   name    => 'level',
                    regex   => '^\d+$',
                    default => 0
                },
            ],
        },

        PFS_FILL_DELEG_ERROR => {
            program =>
                "\'(repeat %d (pfs.fill.deleg.error (eq 1 1) (return 1)))\'",
            description => "Keep restarting the iicu scan",
            parameters  => [
                {   name    => 'repeat_count',
                    regex   => '^\d+$',
                    default => 1
                },
            ],
        },
    };

    if (!(defined($fiji_lookup_table->{$tag}))) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Fiji tag is not supported:\n" . "Tag: $tag"
        );
    }

    return $fiji_lookup_table->{$tag};
} ## end sub _get_fiji_info

# Class destructor.
# Unless the constructor was called with perform_cleanup set to FALSE,
# it takes care of any cleanup that needs to be done on the filer:
# - removes any licenses that were installed by this object
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if (($class->{_creator_pid} == $$) && ($class->{_perform_cleanup})) {
        fcache_log_comment(message => "CLEANUP: Executing Node cleanup for "
                . $class->{_name})
            if ($class->{_debug});

        if ($class->{_cleanup_sfo_takeover}) {

            # Best effort try.....
            fcache_log_comment(message => "CLEANUP: Attempting giveback... ");
            $class->sfo_giveback();
            $class->sfo_wait_for_state(
                local_state   => FCACHE_SFO_STATE_UP,
                partner_state => FCACHE_SFO_STATE_UP,
            );
        }

        if ($class->{_cleanup_sfo}) {
            fcache_log_comment(
                message => "CLEANUP: Disabling Storage Failover for "
                    . $class->{_name});
            NACL::STask::StorageFailover->disable(
                command_interface => $class->{_nacl_node},
                node              => $class->{_nacl_node},
            );
        }

        my $fiji_points = $class->{_cleanup_fiji_points};
        foreach my $slot (keys(%{$fiji_points})) {
            fcache_log_comment(
                message => "CLEANUP: Clearing fiji slot $slot for "
                    . $class->{_name});
            $class->fiji_clear(slot => $slot);
        }

        my $qos_policy_groups = $class->{_cleanup_qos_policy_groups};
        foreach my $qos_policy_group (keys(%{$qos_policy_groups})) {
            fcache_log_comment(message =>
                    "CLEANUP: Deleting Qos policy group $qos_policy_group");
            $class->qos_policy_group_delete(
                policy_group => $qos_policy_group);
        }

        my $cleanup_schedule_intervals
            = $class->{_cleanup_schedule_intervals};
        foreach my $interval (keys(%{$cleanup_schedule_intervals})) {
            fcache_log_comment(
                message => "CLEANUP: Deleting schedule interval $interval");
            $class->schedule_interval_delete(interval_name => $interval);
        }

        my $cleanup_capabilities = $class->{_cleanup_capabilities};
        foreach my $capability (keys(%{$cleanup_capabilities})) {
            my $enable = $cleanup_capabilities->{$capability};
            if ($enable) {
                fcache_log_comment(
                    message => "CLEANUP: Enabling capability: $capability");
                $class->system_capability_enable(
                    capability_name => $capability);
            } else {
                fcache_log_comment(
                    message => "CLEANUP: Disabling capability: $capability");
                $class->system_capability_disable(
                    capability_name => $capability);
            }
        }

        my $cleanup_flags = $class->{_cleanup_flags};
        foreach my $flag (keys(%{$cleanup_flags})) {
            my $value = $cleanup_flags->{$flag};
            fcache_log_comment(
                message => "CLEANUP: Setting $flag back to $value");
            $class->setflag(
                flag  => $flag,
                value => $value,
            );
        }

    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

