#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Origin.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Agnostic setup utility for flexcache testing.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Origin

=head1 DESCRIPTION

This module provides functionality for all the flexcache scripts.
The module is release agnostic and provides
functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though,
that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.
It is responsible for performing all the steps that are needed to create
a flexcache origin volume and to mount/map the volume.
A FlexCache::Origin object will have the following members,
to be used by scripts:

=over

=item node

This member is responsible for the node-level functionality
(filer-level for 7-mode).
For a complete list of available actions on this member,
please see the documentation for FlexCache::Node.

=item aggregate

This member is responsible for the aggregate-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Aggregate.

=item volume

This member is responsible for the volume-level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Volume.

=item vserver

This member is responsible for the vserver/vfiler level functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Vserver.

=item client

This member is responsible for the client-side functionality.
For a complete list of available actions on this member,
please see the documentation for FlexCache::Client.

=back

The test cleanup will be executed by the library,
within the destructor of each object that was instantiated by a script.
A destructor will be responsible for cleaning up
after the actions that its respective object executed.

=head1 SYNOPSIS

        BEGIN {require "flexcache/fcache_common_params.pl";}
        
	use FlexCache::Origin;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
                                            client => {name => $CLIENT_SRC,},
                                            aggregate => {name => $AGGR_SRC,
                                                          block_type => $AGGR_SRC_BLOCK_TYPE,},
                                            volume => {name => $VOL_SRC,},
                                            vserver => {name => $VSERVER,
                                                        dlif_name => $DLIF_SRC,
                                                        dlif_id => $DLIF_SRC_ID,},);

=head1 UNIT TESTING

Unit tests for this module: All the scripts under:

        flexcache/library_unit_tests/

=cut

package FlexCache::Origin;

# Standard libs
use strict;
use warnings;
use Data::Dumper;
$Data::Dumper::Maxdepth = 3;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

# NetApp libs
use Tharn;
use NACL::OptionSet;
use NATE::BaseException qw(:try);
use NACL::STask::Node;
use NACL::STask::VolumeMove;
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Node;
use FlexCache::Aggregate;
use FlexCache::ClientOperations;
use FlexCache::Vserver;
use FlexCache::Volume;
use FlexCache::UI::OntapMessages;
use FlexCache::Utils qw(fcache_log_comment
    fcache_log_result);

=head1 CONSTRUCTOR

=head2 new

Returns a FlexCache::Origin object, representing an origin volume,
with members that can be used for basic testing.
The constructor also takes care (unless otherwise required by the caller)
of the basic setup steps needed for testing.
Please see the constructor documentation for each of the class members
for detailed descriptions of the setup steps.

=head3 ARGUMENTS

=over

=item B<node> I<Required> Hashref:

        Constructor options for the node member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Node.

=item B<aggregate> I<Required> Hashref:

        Constructor options for the aggregate member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Aggregate.

=item B<volume> I<Required> Hashref:

        Constructor options for the volume member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Volume.

=item B<client> I<Required> Hashref:

        Constructor options for the client member.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::ClientOperations.

=item B<vserver> I<Required> Hashref:

        Constructor options for the vserver member.
        The vserver member represents a vserver for C-mode.
        If running on 7-mode in a multistore environment, then
        this member represents a vfiler.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Vserver.

=item B<debug> I<Optional> Boolean:

        Flag to increase the verbosity of the library,
	useful for debugging purposes.
        This parameter is reading the command-line argument FCACHE_DEBUG.
        Default value: FALSE.

=item B<perform_cleanup> I<Optional> Boolean:

        Flag to control if the cleanup should be executed.
        This parameter is reading the command-line argument FCACHE_CLEANUP.
        Default value: TRUE.

=item B<attach> I<Optional> Boolean:

        Setting this flag to true means that all the resources
	(aggregate, vserver, volume) are required to already exist.
	In this case, the library will not attempt to create any of them,
        but it will terminate script execution if they do not already exist.
        Default value: FALSE.

=back

=head3 EXAMPLE

        my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);

=cut

sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   node            => {type => HASHREF,},
            aggregate       => {type => HASHREF,},
            volume          => {type => HASHREF,},
            client          => {type => HASHREF,},
            vserver         => {type => HASHREF, optional => 1,},
            attach          => {type => BOOLEAN, default => FALSE,},
            debug           => {type => BOOLEAN, optional => 1,},
            perform_cleanup => {type => BOOLEAN, optional => 1},
        }
    );

    my $self = {};
    bless($self, $class);

    # For the params which recognize command-line arguments...
    my $options = NACL::OptionSet->new();

    # WARNING!!!!
    # The FlexCache::Cache and the FlexCache::Clone objects are inheriting
    # from FlexCache::Origin. For this to work correctly,
    # make sure that all the class members of the parent class
    # are set correctly there as well

    $self->{_debug}
        = (defined($args{debug}))
        ? ($args{debug})
        : ($options->get('FCACHE_DEBUG'));
    $self->{_perform_cleanup}
        = (defined($args{perform_cleanup}))
        ? ($args{perform_cleanup})
        : ($options->get('FCACHE_CLEANUP'));
    $self->{_attach} = $args{attach};

   # Connect to the node.
   # Create the NACL node object here and pass it in to each
   # constructor for the members.
   # Because of this, we need to process here the name entry in the node hash.
    $self->{_node_name} = delete($args{node}->{name});
    if (!defined($self->{_node_name})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Missing or undefined \"name\" field in the node hash:\n"
                . Dumper(%args),
        );
    }
    $self->{_nacl_node} = NACL::STask::Node->new(node => $self->{_node_name});

    # Connect to the client (for now, just to find out the client type)
    # We need to do this as a two-step process such that the vserver setup is
    # done correctly before mounting/mapping the volume/share
    $self->{client} = FlexCache::ClientOperations->new(
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        _postpone_setup => TRUE,
        %{$args{client}}
    );
    my $client_type = $self->{client}->get_type();

    $self->{node} = FlexCache::Node->new(
        node            => $self->{_nacl_node},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        %{$args{node}}
    );
    $self->{_boot_mode} = $self->{node}->get_boot_mode();
    $self->{_version}   = $self->{node}->_get_version();

    if (   ($self->{_boot_mode} eq FCACHE_ONTAP_7MODE)
        && ($client_type eq FCACHE_CLIENT_TYPE_WINDOWS)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "FlexCache does not support CIFS in 7-mode\n"
                . "This test-case should not be run on a Windows client"
        );
    }

    $self->{_ui_messages} = FlexCache::UI::OntapMessages->new(
        boot_mode => $self->{_boot_mode},
        release   => $self->{_version},
        debug     => $self->{_debug},
    );
    $self->{node}->_attach_ui_member(ui_member => $self->{_ui_messages});

    # Create/reuse the aggregate
    $self->{aggregate} = FlexCache::Aggregate->new(
        node            => $self->{_nacl_node},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        attach_only     => $self->{_attach},
        ui_messages     => $self->{_ui_messages},
        %{$args{aggregate}}
    );
    my $aggr_name = $self->{aggregate}->get_name;

    # Create and configure the vserver
    $self->{vserver} = FlexCache::Vserver->new(
        node            => $self->{_nacl_node},
        aggregate_name  => $aggr_name,
        client_type     => $client_type,
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
        config_type     => 'origin',
        attach_only     => $self->{_attach},
        ui_messages     => $self->{_ui_messages},
        %{$args{vserver}},
    );
    my $vserver_name = $self->{vserver}->get_name(_internal_use => TRUE,);

    # Create/reuse the volume
    $self->{volume} = FlexCache::Volume->new(
        %{$args{volume}},
        node            => $self->{_nacl_node},
        aggregate_name  => $aggr_name,
        vserver_name    => $vserver_name,
        client_type     => $client_type,
        attach_only     => $self->{_attach},
        ui_messages     => $self->{_ui_messages},
        debug           => $self->{_debug},
        perform_cleanup => $self->{_perform_cleanup},
    );

    # Now we can mount/map the volume/share
    my $dlif_ip        = $self->{vserver}->get_dlif_ip();
    my $mount_junction = $self->{volume}->get_client_mount_junction();
    $self->{client}->_execute_setup(
        mount_from_ip       => $dlif_ip,
        mount_from_junction => $mount_junction,
    );

    return $self;

} ## end sub new

=head1 OTHER METHODS

=head2 volume_move

Move the volume from one aggregate to another.

=head3 ARGUMENTS

=over

=item B<node_dst> I<Optional> SCALAR:

        Name of the node hosting the destination aggregate.
        Default value: local node.

=item B<aggregate_dst> I<Required> HASHREF:

        Options for the destination aggregate.
        For a description of the key-value pairs for this hash,
        please see the constructor arguments for FlexCache::Aggregate.

=item B<expected> I<Optional> SCALAR: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 EXAMPLE

        $origin->volume_move(aggregate_dst => {name => $MOVE_TO_AGGR});
        $cache->volume_move(aggregate_dst => {name => $MOVE_TO_AGGR},
                            expected => FCACHE_OPERATION_NOT_ALLOWED_ON_A_CACHE);

=cut

sub volume_move {
    my $class = shift;
    my %args  = validate(
        @_,
        {   node_dst => {type => SCALAR, default => $class->{_node_name},},
            aggregate_dst => {type => HASHREF,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $expected_rc = $args{expected};

    # Do this based on the expected result
    # (depending on this, we'll have to do different house-keeping)
    if ($expected_rc == FCACHE_RC_SUCCESS) {
        $class->_volume_move_expect_success(%args);
    } else {
        $class->_volume_move_expect_failure(%args);
    }

    return FCACHE_RC_SUCCESS;
} ## end sub volume_move

=head2 delay_execution

Add a delay in script execution.

=head3 ARGUMENTS

=over

=item B<duration> I<Optional> SCALAR:

        Duration, in seconds, for the delay to be added.
        Default value: FCACHE_CP_DURATION + 1.

=back

=head3 EXAMPLE

        $origin->delay_execution();
	$origin->delay_execution(duration => (2 * FCACHE_CP_DURATION));
        $origin->delay_execution(duration => 30);

=cut

sub delay_execution {
    my $class = shift;
    my %args  = validate(
        @_,
        {   duration => {
                type    => SCALAR,
                regex   => qr/^\d+$/,
                default => (FCACHE_CP_DURATION + 1),
            },
        }
    );

    my $duration = $args{duration};
    if ($duration <= 0) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Delay duration must have a positive value"
        );
    }

    fcache_log_comment(
        message => "Delaying script execution with $duration seconds...");

    Tharn::snooze($duration);

    return FCACHE_RC_SUCCESS;
} ## end sub delay_execution

=head1 MEMBER METHODS

Any FlexCache::Origin object can be used by calling the appropriate method
on any of its members:

        # 'method' = any available method of FlexCache::Node
        $origin->{node}->method(...);
        # 'method' = any available method of FlexCache::Aggregate
        $origin->{aggregate}->method(...);
        # 'method' = any available method of FlexCache::Vserver
        $origin->{vserver}->method(...);
        # 'method' = any available method of FlexCache::Volume
        $origin->{volume}->method(...);
        # 'method' = any available method of FlexCache::ClientOperations
        $origin->{client}->method(...);

=cut

########################################################
# Internal methods
########################################################

# Volume move, when the operation is expected to succeed.
# Please see volume_move documentation for input/output.
sub _volume_move_expect_success {
    my $class = shift;
    my %args  = validate(
        @_,
        {   node_dst => {type => SCALAR, default => $class->{_node_name},},
            aggregate_dst => {type => HASHREF,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $node_dst = $args{node_dst};

    my $nacl_node_src = $class->{_nacl_node};
    my $node_src      = $class->{node};
    if ($node_dst ne $class->{_node_name}) {
        $class->{_nacl_node} = NACL::STask::Node->new(node => $node_dst);
        $class->{node} = FlexCache::Node->new(
            node            => $class->{_nacl_node},
            debug           => $class->{_debug},
            perform_cleanup => $class->{_perform_cleanup},
        );
        $class->{_boot_mode}   = $class->{node}->get_boot_mode();
        $class->{_version}     = $class->{node}->_get_version();
        $class->{_ui_messages} = FlexCache::UI::OntapMessages->new(
            boot_mode => $class->{_boot_mode},
            release   => $class->{_version},
            debug     => $class->{_debug},
        );
    }

    fcache_log_comment(
        message => "Volume move: Creating/reusing the destination aggregate");

    # Remember the source aggregate, to clean up after the move in done
    my $src_aggr = $class->{aggregate};
    $class->{aggregate} = FlexCache::Aggregate->new(
        node            => $class->{_nacl_node},
        debug           => $class->{_debug},
        perform_cleanup => $class->{_perform_cleanup},
        attach_only     => $class->{_attach},
        ui_messages     => $class->{_ui_messages},
        %{$args{aggregate_dst}}
    );

    fcache_log_comment(message => "Volume move: Starting the move operation");
    my $vol_name      = $class->{volume}->get_name();
    my $dst_aggr_name = $class->{aggregate}->get_name();
    my $vserver_name  = $class->{vserver}->get_name(_internal_use => TRUE,);
    NACL::STask::VolumeMove->start(
        command_interface       => $nacl_node_src,
        volume                  => $vol_name,
        'destination-aggregate' => $dst_aggr_name,
        vserver                 => $vserver_name,
        nacltask_wait           => 1,
    );

    fcache_log_comment(
        message => "Volume move: Cleaning up the source aggregate");
    undef($src_aggr);

    # Re-attach to the volume
    # (has a new container aggr now and may have a new hosting node)
    my $client_type = $class->{client}->get_type();

    # Disable the destructor of the "old" volume
    my $vol_recycle = $class->{volume}->_get_recycle();
    $class->{volume}->_set_perform_cleanup(FALSE);
    $class->{volume} = FlexCache::Volume->new(
        name            => $vol_name,
        node            => $class->{_nacl_node},
        aggregate_name  => $dst_aggr_name,
        vserver_name    => $vserver_name,
        client_type     => $client_type,
        attach_only     => TRUE,
        ui_messages     => $class->{_ui_messages},
        debug           => $class->{_debug},
        perform_cleanup => $class->{_perform_cleanup},
    );

    # The "new" volume should now remember the "old" recycle settings
    $class->{volume}->_set_recycle($vol_recycle);
    fcache_log_comment(message => "Volume move completed successfully");

    return FCACHE_RC_SUCCESS;
} ## end sub _volume_move_expect_success

# Volume move, when the operation is expected to fail.
# Please see volume_move documentation for input/output.
sub _volume_move_expect_failure {
    my $class = shift;
    my %args  = validate(
        @_,
        {   node_dst => {type => SCALAR, default => $class->{_node_name},},
            aggregate_dst => {type => HASHREF,},
            expected      => {type => SCALAR,},
        }
    );

    my $node_dst_name = $args{node_dst};

    my $nacl_node_dst = $class->{_nacl_node};
    my $node_dst      = $class->{node};
    my $node_dst_ui   = $class->{_ui_messages};
    if ($node_dst_name ne $class->{_node_name}) {
        $nacl_node_dst = NACL::STask::Node->new(node => $node_dst_name);
        $node_dst = FlexCache::Node->new(
            node            => $nacl_node_dst,
            debug           => $class->{_debug},
            perform_cleanup => $class->{_perform_cleanup},
        );
        my $node_dst_boot_mode = $node_dst->get_boot_mode();
        my $node_dst_version   = $node_dst->_get_version();
        $node_dst_ui = FlexCache::UI::OntapMessages->new(
            boot_mode => $node_dst_boot_mode,
            release   => $node_dst_version,
            debug     => $class->{_debug},
        );
    }

    fcache_log_comment(
        message => "Volume move: Creating/reusing the destination aggregate");
    my $dst_aggr = FlexCache::Aggregate->new(
        node            => $nacl_node_dst,
        debug           => $class->{_debug},
        perform_cleanup => $class->{_perform_cleanup},
        ui_messages     => $node_dst_ui,
        %{$args{aggregate_dst}}
    );

    fcache_log_comment(
        message => "Volume move: Attempting the move operation");
    my $vol_name      = $class->{volume}->get_name();
    my $dst_aggr_name = $dst_aggr->get_name();
    my $vserver_name  = $class->{vserver}->get_name(_internal_use => TRUE,);
    my $output        = '';

    try {
        NACL::STask::VolumeMove->start(
            command_interface       => $class->{_nacl_node},
            volume                  => $vol_name,
            'destination-aggregate' => $dst_aggr_name,
            vserver                 => $vserver_name,
            nacltask_wait           => 0,
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
        fcache_log_comment(message => "\n\n\nCaught the error: $output")
            if ($class->{_debug});
    };

    undef($dst_aggr);

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_MOVE,
        output      => $output,
        expected_rc => $args{expected}
    );

    fcache_log_comment(
        message => "Volume move attempt completed successfully");

    return FCACHE_RC_SUCCESS;
} ## end sub _volume_move_expect_failure

# Class destructor
# This top-level destructor is only to ensure the correct order of cleanup steps.
# For details on each member cleanup, please see the appropriate module documentation.
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    fcache_log_comment(
        message => "CLEANUP: Entering FlexCache cleanup for origin...");

    # Enforce the correct order of member cleanup
    $class->{client}    = undef;
    $class->{volume}    = undef;
    $class->{vserver}   = undef;
    $class->{aggregate} = undef;
    $class->{node}      = undef;

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

