#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Snapmirror.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Mode and release agnostic module for snapmirror functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Snapmirror

=head1 DESCRIPTION

This module provides snapmirror functionality for all the flexcache scripts.
The module is release agnostic and provides
functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though,
that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

The snapmirror cleanup will be executed by the library,
within the destructor of each object that was instantiated by a script.

=head1 SYNOPSIS

        use FlexCache::Snapmirror;
	
	BEGIN {require "flexcache/fcache_common_params.pl";}
	
	my $snapmirror = FlexCache::Snapmirror->new(source => $vol1,
                                                    destination => $vol2,
                                                    type => FCACHE_SNAPMIRROR_TYPE_DP,);

=head1 LIMITATIONS

C-mode LS mirror types are supported,
only with destination volumes within the same cluster.

=head1 UNIT TESTING

Unit test for this module:

        flexcache/library_unit_tests/Snapmirror.thpl

=cut

package FlexCache::Snapmirror;

# Standard libs
use strict;
use warnings;
use Data::Dumper;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use NATE::BaseException qw(:try);

# NetApp libs
use Tharn;
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use NACL::STask::Node;
use NACL::STask::Snapmirror;
use NACL::C::Snapmirror;
use NACL::C::SnapmirrorPolicy;

=head1 CONSTRUCTOR

=head2 create

Returns a FlexCache::Snapmirror object, representing a snapmirror relationship.
The relationship is created/initialized at this time.
For C-mode, this means: create+initialize; for 7-mode: initialize.
LS-type relationships are also supported, with the assumption that
the destination volumes are all in the same cluster.
In case of an LS relationship, the constructor will:
perform create on each source-destination pair,
then it will perform initialize-ls-set.

=head3 ARGUMENTS

=over

=item B<source> I<Required> Object:

        FlexCache::Origin or FlexCache::Cache object,
	representing the source volume.

=item B<destination> I<Required> Arrayref:

        Arrayref of FlexCache::Origin or FlexCache::Cache object(s),
	representing the destination(s) volume.
        This array can have more than one element for a C-mode LS relation,
        and it should have only one element otherwise (in this case,
	if more than one element is passed in,
        only the first one will be considered as the destination;
	the rest will be ignored).

=item B<type> I<Optional> Scalar:

        Type of the snapmirror relationship to be created.
        This argument is ignored for 7-mode.
        Legal values: any of the FCACHE_SNAPMIRROR_TYPE_* constants,
	defined in FlexCache::Constants.
        Default value: FCACHE_SNAPMIRROR_TYPE_DP.

=item B<recycle> I<Optional> Boolean:

        Flag to determine if the snapmirror relationship should be recycled.
        If set to false, then any already existent relationship
	for the specified volumes will be purged and recreated.
        At cleanup, the relationship will be destroyed.
        If set to true, then an already existent relationship will be reused,
	and at cleanup it will not be purged.
        Default value: FALSE.
        
=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        If this argument is set to anything other than FCACHE_RC_SUCCESS,
        this means that the create/initialize command is expected to fail.
        Such, the constructor will return undef in this case
	(since there will be no relationship created).
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=item B<perform_cleanup> I<Optional> Boolean:

        Flag to control if the cleanup should be executed.
        Default value: the "perform_cleanup" setting of the origin object.

=back

=head3 EXAMPLES

        # Create a "default-type (DP) relationship
        my $snapmirror = FlexCache::Snapmirror->new(source => $vol1,
                                                    destination => [$vol2],);
        # Create an LS relationship
        my $snapmirror = FlexCache::Snapmirror->new(source => $vol1,
                                                    destination => [$vol2, $vol3],
                                                    type => FCACHE_SNAPMIRROR_TYPE_LS,);
        # Negative testing: a cache cannot be a snapmirror destination
        FlexCache::Snapmirror->new(source => $origin,
                                   destination => [$cache],
                                   expected => FCACHE_RC_DESTINATION_CANNOT_BE_A_CACHE,,);

=cut

sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   source      => {type => OBJECT,},
            destination => {type => ARRAYREF,},
            type => {type => SCALAR, default => FCACHE_SNAPMIRROR_TYPE_DP},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            recycle  => {type => SCALAR, default => FALSE,},
            perform_cleanup => {type => BOOLEAN, optional => 1,},
        }
    );

    my $source       = $args{source};
    my $destinations = $args{destination};
    my $type         = $args{type};
    my $expected     = $args{expected};
    my $recycle      = $args{recycle};
    my $if_exists    = ($recycle) ? 'reuse' : 'purge';
    my $ui_type      = FCACHE_UI_TYPE_CLI
        ;    # As of 05/12, nacl does not support ZAPI for create...

    my $self = {};
    $self->{_creator_pid} = $$;
    $self->{_recycle}     = $recycle;
    $self->{_perform_cleanup}
        = (defined($args{perform_cleanup}))
        ? ($args{perform_cleanup})
        : ($source->{_perform_cleanup});
    $self->{_type} = $type;

    bless($self, $class);

    # Obtain all the needed information for the source
    if (   !($source->isa("FlexCache::Origin"))
        && !($source->isa("FlexCache::Cache"))) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "The source parameter to create should be either an "
                . "FlexCache::Origin or an FlexCache::Cache object"
        );
    }
    my $source_path      = $source->{volume}->get_fullname();
    my $source_vol_name  = $source->{volume}->get_name();
    my $source_boot_mode = $source->{node}->get_boot_mode();
    if (   ($source_boot_mode eq FCACHE_ONTAP_7MODE)
        && ($self->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "LS mirrors are only available in C-mode"
        );
    }
    $self->{_source_path}         = $source_path;
    $self->{_source_vol_name}     = $source_vol_name;
    $self->{_source_cluster_name} = $source->{node}->_get_cluster_name();
    $self->{_source_vserver_name} = $source->{volume}->get_vserver_name();

    $self->{_nacl_sm_relations} = {};

    # Process the destinations and creating the relationship(s)
    my $do_initialize = ($self->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS) ? 0 : 1;
    my $rc            = FCACHE_RC_FAILURE;
    my $output        = '';
    foreach my $destination (@{$destinations}) {
        if (   !($destination->isa("FlexCache::Origin"))
            && !($destination->isa("FlexCache::Cache"))) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message =>
                    "The elements of the destination parameter to create"
                    . " should be either:\n"
                    . "FlexCache::Origin or FlexCache::Cache objects"
            );
        }

        my $destination_node_name = $destination->{node}->get_name();
        my $destination_path      = $destination->{volume}->get_fullname();
        my $destination_vol_name  = $destination->{volume}->get_name();

        fcache_log_comment(
                  message => "Creating snapmirror relationship for:\n"
                . "Source: $source_path\n"
                . "Destination: $destination_path\n"
                . "Type: $type");

        if (!defined($self->{_nacl_dest_node})) {
            $self->{_nacl_dest_node}
                = NACL::STask::Node->new(node => $destination_node_name);
            $self->{_boot_mode} = $destination->{node}->get_boot_mode();
        }

      # Because of NACL burt 594705, have the following temporary work-around:
      # remove once the burt is fixed:
        ### start work-around
        if (!($self->{_recycle})) {
            try {
                NACL::STask::Snapmirror->purge(
                    command_interface  => $self->{_nacl_dest_node},
                    'destination-path' => $destination_path,
                    'source-path'      => $source_path,
                    nacltask_if_purged => 'pass',
                    sync_type          => FCACHE_SNAPMIRROR_DFLT_SYNC_TYPE
                );
            }
            catch NATE::BaseException with {

                # Do nothing here - leave this a best-effort try
                # (vsims do not really behave here)
                # If anything bad happens in the purge,
                # then the create will fail anyway...
            };
        }
        ### end work-around

        try {
            $self->{_nacl_sm_relations}->{$destination_path}
                = NACL::STask::Snapmirror->create(
                command_interface  => $self->{_nacl_dest_node},
                'destination-path' => $destination_path,
                'source-path'      => $source_path,
                if_exists =>
                    'reuse',    #$if_exists, #TODO: after burt 594705 is fixed
                type          => $type,
                sync_type     => FCACHE_SNAPMIRROR_DFLT_SYNC_TYPE,
                do_initialize => $do_initialize,
                apiset_must   => {interface => $ui_type}
                );
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };

        # Since for flexcache, the only negative testing is for 'create'
        # use the destination's ui member to check for the expected result.
        # If this changes, then the Snapmirror class should
        # have its own ui member, to use in any of the class methods.
        $destination->{_ui_messages}->verify_expected_rc(
            command     => FCACHE_CMD_SNAPMIRROR_NEW,
            output      => $output,
            expected_rc => $expected,
            ui_type     => $ui_type
        );

        # For any type other than LS,
        # we will only consider the first destination that was passed in
        if ($self->{_type} ne FCACHE_SNAPMIRROR_TYPE_LS) {
            $self->{__destination_path} = $destination_path;
            last;
        }
    }

    if ($expected ne FCACHE_RC_SUCCESS) {
        return undef;
    }

    # If LS:
    # We are done creating all the individual LS relationships.
    # Now we need to do the ls-initialize-set.
    if ($self->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS) {
        fcache_log_comment(message => "Initializing the snapmirror LS set");
        NACL::C::Snapmirror->initialize_ls_set(
            command_interface => $self->{_nacl_dest_node},
            'source-cluster'  => $self->{_source_cluster_name},
            'source-vserver'  => $self->{_source_vserver_name},
            'source-volume'   => $self->{_source_vol_name},
            foreground        => 'true',
        );
    }

    return $self;
} ## end sub new

=head1 OTHER METHODS

=head2 get_policy_name

Obtain the snapmirror policy name associated to this relationship.
This method only applies to C-mode.

=head3 ARGUMENTS

None.

=head3 RETURN

Policy name (Scalar).

=head3 EXAMPLE

	my $policy_name = $obj->get_policy_name();

=cut

sub get_policy_name {
    my $class = shift;
    my %args = validate(@_, {});

    if (   ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS)
        || ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This method is not yet supported for LS mirrors or for 7-mode"
        );
    }

    my $destination_path = $class->{__destination_path};
    my $state = $class->{_nacl_sm_relations}->{$destination_path}->state();
    my $policy_name = $state->policy();

    return $policy_name;
} ## end sub get_policy_name

=head2 get_label

Obtain the snapmirror label of the policy associated to this relationship.
This method is useful for use with Vault relationships, where incremental
transfer will not capture current image of the AFS of source volume, instead it
will pick up snapshots that are newer than the common snapshot and match them
against the snapmirror policy configured for this relationship. Only snapshots
that match the policy will be transferred by incremental transfer

=head3 ARGUMENTS

None.

=head3 RETURN

Snapmirror label(s) to be used for this relationship (Scalar/Array).

=head3 EXAMPLE

	my $snapmirro_label = $obj->get_label();

=cut

sub get_label {
    my $class = shift;
    my %args = validate(@_, {});

    if (   ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS)
        || ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This method is not yet supported for LS mirrors or for 7-mode"
        );
    }

    my $policy_name = $class->get_policy_name();

    my $policy = NACL::C::SnapmirrorPolicy->find(
        command_interface => $class->{_nacl_dest_node},
        filter            => {policy => $policy_name},
    );

    my $state = $policy->state();
    my $label = $state->snapmirror_label();

    if (wantarray) {
        return (@{$label});
    } else {
        return $label->[0];
    }
} ## end sub get_label

=head2 update

Perform an update on the snapmirror relationship
(for LS relationships, the action performed is update-ls-set).
It waits for the update to complete.

=head3 ARGUMENTS

=over

=item B<background> I<Optional> Boolean:

        Flag to control if the update should happen in the background.
	This argument is not yet supported for LS mirrors.
	Default value: FALSE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->update();

=cut

sub update {
    my $class = shift;
    my %args
        = validate(@_, {background => {type => BOOLEAN, default => FALSE,},});

    my $background = $args{background};

    if ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS) {
        if ($background == TRUE) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message =>
                    "Background option is not yet supported for LS mirrors"
            );
        }
        NACL::C::Snapmirror->update_ls_set(
            command_interface => $class->{_nacl_dest_node},
            'source-cluster'  => $class->{_source_cluster_name},
            'source-vserver'  => $class->{_source_vserver_name},
            'source-volume'   => $class->{_source_vol_name},
            foreground        => 'true',
        );
    } else {
        my $destination_path = $class->{__destination_path};
        if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
            my $nacltask_wait = ($background == TRUE) ? 0 : 1;
            $class->{_nacl_sm_relations}->{$destination_path}
                ->update(nacltask_wait => $nacltask_wait);
        } else {

            # NACL::STask is not currently supporting update for 7-mode.
            NACL::C::Snapmirror->update(
                command_interface  => $class->{_nacl_dest_node},
                'source-path'      => $class->{_source_path},
                'destination-path' => $class->{__destination_path}
            );
            if ($background == FALSE) {
                $class->{_nacl_sm_relations}->{$destination_path}
                    ->wait_for_transfer_to_complete();
            }
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub update

=head2 wait_until_transfer_completes

Wait for a snapmirror transfer to complete.

=head3 ARGUMENTS

None

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

    $obj->wait_until_transfer_completes();

=cut

sub wait_until_transfer_completes {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is not yet supported for LS mirrors"
        );
    }

    my $destination_path = $class->{__destination_path};
    $class->{_nacl_sm_relations}->{$destination_path}
        ->wait_for_transfer_to_complete();

    return FCACHE_RC_SUCCESS;
} ## end sub wait_until_transfer_completes

=head2 abort

Abort a snapmirror transfer.

=head3 ARGUMENTS

None

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

    $obj->abort();

=cut

sub abort {
    my $class = shift;
    my %args = validate(@_, {});

    if (   ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS)
        || ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This method is not yet supported for LS mirrors or for 7-mode"
        );
    }

    my $destination_path = $class->{__destination_path};
    $class->{_nacl_sm_relations}->{$destination_path}
        ->abort(nacltask_verify => 1);

    return FCACHE_RC_SUCCESS;
} ## end sub abort

=head2 quiesce

Quiesce a snapmirror transfer.

=head3 ARGUMENTS

None

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

    $obj->quiesce();

=cut

sub quiesce {
    my $class = shift;
    my %args = validate(@_, {});

    if (   ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS)
        || ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This method is not yet supported for LS mirrors or for 7-mode"
        );
    }

    my $destination_path = $class->{__destination_path};
    $class->{_nacl_sm_relations}->{$destination_path}
        ->quiesce(nacltask_verify => 1);

    return FCACHE_RC_SUCCESS;
} ## end sub quiesce

=head2 resume

Resume a snapmirror transfer.

=head3 ARGUMENTS

None

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

    $obj->resume();

=cut

sub resume {
    my $class = shift;
    my %args = validate(@_, {});

    if (   ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS)
        || ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This method is not yet supported for LS mirrors or for 7-mode"
        );
    }

    my $destination_path = $class->{__destination_path};
    $class->{_nacl_sm_relations}->{$destination_path}
        ->resume(nacltask_verify => 0);

    return FCACHE_RC_SUCCESS;
} ## end sub resume

=head2 break

Break the snapmirror relationship.

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->break();

=cut

sub break {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Snapmirror break is not supported for LS relationships"
        );
    } else {
        my $destination_path = $class->{__destination_path};
        $class->{_nacl_sm_relations}->{$destination_path}->break();
    }

    return FCACHE_RC_SUCCESS;
} ## end sub break

=head2 resync

Resync the snapmirror relationship.

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->resync();

=cut

sub resync {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_type} eq FCACHE_SNAPMIRROR_TYPE_LS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Snapmirror resync is not supported for LS relationships"
        );
    } else {
        my $destination_path = $class->{__destination_path};
        if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
            $class->{_nacl_sm_relations}->{$destination_path}->resync();
        } else {

            # NACL::STask is not currently supporting resync for 7-mode.
            NACL::C::Snapmirror->resync(
                command_interface  => $class->{_nacl_dest_node},
                'source-path'      => $class->{_source_path},
                'destination-path' => $class->{__destination_path}
            );
            $class->{_nacl_sm_relations}->{$destination_path}
                ->wait_for_transfer_to_complete();
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub resync

=head2 delete

Delete (purge) the snapmirror relationship.

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->delete();

=cut

sub delete {
    my $class = shift;
    my %args
        = validate(@_, {_if_purged => {type => SCALAR, default => 'die'}});

    my $action_if_purged = $args{_if_purged};
    if (   ($action_if_purged eq 'die')
        && (scalar(keys(%{$class->{_nacl_sm_relations}})) == 0)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Snapmirror relationship was already deleted within this script"
        );
    }

    foreach my $destination_path (keys(%{$class->{_nacl_sm_relations}})) {
        $class->{_nacl_sm_relations}->{$destination_path}
            ->purge(nacltask_if_purged => $action_if_purged);
        delete($class->{_nacl_sm_relations}->{$destination_path});
    }

    return FCACHE_RC_SUCCESS;
} ## end sub delete

=head2 modify_settings

Modify setting(s) for this relationship.
At least one of the arguments must be provided.
This method is only supported for C-mode
(in 7-mode, script execution will be terminated with N/A). 

=head3 ARGUMENTS

=over

=item B<schedule> I<Optional> Scalar:

        Schedule to be set for this relationship.
        Legal values: any of the FCACHE_SNAPMIRROR_SCHEDULE_* constants,
	defined in FlexCache::Constants.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->modify_settings(schedule => FCACHE_SNAPMIRROR_SCHEDULE_HOURLY);

=cut

sub modify_settings {
    my $class = shift;
    my %args = validate(@_, {schedule => {type => SCALAR, optional => 1,},});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message =>
                "Snapmirror verify_settings is not available for 7-mode"
        );
    }

    if (scalar(keys(%args)) == 0) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "At least one of the method arguments MUST be provided"
        );
    }

    foreach my $destination_path (keys(%{$class->{_nacl_sm_relations}})) {
        $class->{_nacl_sm_relations}->{$destination_path}->modify(
            foreground => 'true',
            %args
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub modify_settings

=head2 verify_settings

Verify specific setting(s) for this relationship.
If any of the passed-in argument-value pairs do not match with the current
setting, script execution is terminated with a fatal failure.
Calling this method with no arguments will result with only checking that
the snapmirror relationship exists.
This method is only supported for C-mode
(in 7-mode, script execution will be terminated with N/A). 

=head3 ARGUMENTS

=over

=item B<schedule> I<Optional> Scalar:

        Expected schedule for this relationship.
        Legal values: any of the FCACHE_SNAPMIRROR_SCHEDULE_* constants,
	defined in FlexCache::Constants.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->verify_settings(schedule => FCACHE_SNAPMIRROR_SCHEDULE_HOURLY);

=cut

sub verify_settings {
    my $class = shift;
    my %args = validate(@_, {schedule => {type => SCALAR, optional => 1,},});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message =>
                "Snapmirror verify_settings is not available for 7-mode"
        );
    }

    foreach my $destination_path (keys(%{$class->{_nacl_sm_relations}})) {
        NACL::C::Snapmirror->find(
            command_interface => $class->{_nacl_dest_node},
            filter            => {
                'destination-path' => $destination_path,
                'source-path'      => $class->{_source_path},
                %args
            }
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub verify_settings

=head2 ls_set_add_destination

Add a destination volume to this LS relationship.

=head3 ARGUMENTS

=over

=item B<destination> I<Required> Object:

        FlexCache::Origin or FlexCache::Cache object,
        representing the destination volume that will be added
	to this LS relationship.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 WARNING!

When using this method, please create the Origin/Cache object for the
destination to be added BEFORE creating the Snapmirror object.
This is because Perl's garbage collection happens in reverse order of
object creation. Such, the snapmirror relationship needs to be purged before
the volumes are destroyed (otherwise both this destination and the source,
as well as the corresponding snapmirror relationship will be left behind).

        DO THIS:
        my $source = ...;
        my $dest1 = ...;
        my $dest2 = ...;
        my $dest3 = ...;
        my $ls_mirror = FlexCache::Snapmirror->new(source => $source,
                                                   destination => [$dest1, $dest2],
                                                   type => FCACHE_SNAPMIRROR_TYPE_LS);
        ....
        $ls_mirror->ls_set_add_destination(destination => $dest3);
        
        DON'T DO THIS:
        my $source = ...;
        my $dest1 = ...;
        my $dest2 = ...;
        my $ls_mirror = FlexCache::Snapmirror->new(source => $source,
                                                   destination => [$dest1, $dest2],
                                                   type => FCACHE_SNAPMIRROR_TYPE_LS);
        ....
        my $dest3 = ...;
        $ls_mirror->ls_set_add_destination(destination => $dest3);

=head3 EXAMPLE

	$obj->ls_set_add_destination(destination => $dest_obj);

=cut

sub ls_set_add_destination {
    my $class = shift;
    my %args = validate(@_, {destination => {type => OBJECT,},});

    if ($class->{_type} ne FCACHE_SNAPMIRROR_TYPE_LS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This method only applies for C-mode LS relationships"
        );
    }

    my $destination = $args{destination};
    if (   !($destination->isa("FlexCache::Origin"))
        && !($destination->isa("FlexCache::Cache"))) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "The destination parameter should be either"
                . "FlexCache::Origin or FlexCache::Cache objects"
        );
    }

    my $destination_node_name = $destination->{node}->get_name();
    my $destination_path      = $destination->{volume}->get_fullname();

    if (exists($class->{_nacl_sm_relations}->{$destination_path})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This destination is already part of the LS relationship:\n"
                . "$destination_path"
        );
    }

    $class->{_nacl_sm_relations}->{$destination_path}
        = NACL::STask::Snapmirror->create(
        command_interface  => $class->{_nacl_dest_node},
        'destination-path' => $destination_path,
        'source-path'      => $class->{_source_path},
        if_exists          => 'die',
        type               => FCACHE_SNAPMIRROR_TYPE_LS,
        do_initialize      => 1
        );

    return FCACHE_RC_SUCCESS;
} ## end sub ls_set_add_destination

=head2 ls_set_remove_destination

Remove a destination from this LS relationship.
The destination must be previously added by a method of this object
(constructor or ls_set_add_destination).

=head3 ARGUMENTS

=over

=item B<destination> I<Required> Object:

        FlexCache::Origin or FlexCache::Cache object,
        representing the destination volume that will be added to this LS relationship.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	$obj->ls_set_remove_destination(destination => $dest_obj);

=cut

sub ls_set_remove_destination {
    my $class = shift;
    my %args = validate(@_, {destination => {type => OBJECT,},});

    if ($class->{_type} ne FCACHE_SNAPMIRROR_TYPE_LS) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This method only applies for C-mode LS relationships"
        );
    }

    my $destination = $args{destination};
    if (   !($destination->isa("FlexCache::Origin"))
        && !($destination->isa("FlexCache::Cache"))) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "The destination parameter should be either"
                . "FlexCache::Origin or FlexCache::Cache objects"
        );
    }

    my $destination_node_name = $destination->{node}->get_name();
    my $destination_path      = $destination->{volume}->get_fullname();

    if (!exists($class->{_nacl_sm_relations}->{$destination_path})) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "This destination was not previously added to LS relationship:\n"
                . "$destination_path"
        );
    }

    $class->{_nacl_sm_relations}->{$destination_path}->purge();
    delete($class->{_nacl_sm_relations}->{$destination_path});

    return FCACHE_RC_SUCCESS;
} ## end sub ls_set_remove_destination

########################################################
# Internal methods
########################################################

# Class destructor
# It takes care of purging the relationship
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if (($class->{_creator_pid} == $$) && ($class->{_perform_cleanup})) {
        if ($class->{_recycle}) {
            fcache_log_comment(message => "CLEANUP: Recycling requested."
                    . " Leaving behind the snapmirror relationship:\n"
                    . "Source: "
                    . $class->{_source_path} . "\n"
                    . "Type: "
                    . $class->{_type});
        } else {
            fcache_log_comment(
                message => "CLEANUP: Purging the snapmirror relationship:\n"
                    . "Source: "
                    . $class->{_source_path} . "\n"
                    . "Type: "
                    . $class->{_type});
            $class->delete(_if_purged => 'pass');
        }
    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

