#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/UI/ClientMessages.pm#3 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Utility library for flexcache client-side output processing.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::UI::ClientMessages

=head1 DESCRIPTION

This is an utility module, to be used by the FlexCache library
whenever client output needs to be validated (negative testing).
The module provides functionality for multiple types of clients.
It is intended to be only used internally by the FlexCache library.

=head1 SYNOPSIS

        use FlexCache::Constants qw(:all);
        use FlexCache::Constants qw(:_internal);
        use FlexCache::UI::ClientMessages;
        
        my $ui_messages = FlexCache::UI::ClientMessages->new(client_type => FCACHE_CLIENT_TYPE_UNIX);

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl
        
=cut

package FlexCache::UI::ClientMessages;

# Standard libs
use strict;
use warnings;
use Data::Dumper;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Utils qw(fcache_log_comment
    fcache_log_result
    fcache_verify_expected_rc
    _map_message_to_rc);

=head1 CONSTRUCTOR

=head2 new

Returns a FlexCache::UI::ClientMessages object, to be used 
whenever client-side output needs to be validated (negative testing).

=head3 ARGUMENTS

=over

=item B<client_type> I<Required> Scalar:

        Client type (*nix, windows).
	Legal values: any FCACHE_CLIENT_TYPE_* constants
	(defined in FlexCache::Constants).
        
=item B<debug> I<Optional> Scalar:

        Flag to increase the verbosity of this module, useful for debugging purposes.
        Default value: FALSE.

=back

=head3 EXAMPLE

        my $ui_messages = FlexCache::UI::ClientMessages->new(client_type => FCACHE_CLIENT_TYPE_UNIX);

=cut

sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   client_type => {type => SCALAR,},
            debug       => {type => SCALAR, default => FALSE,},
        }
    );

    my $self = {};
    $self->{_client_type} = $args{client_type};
    $self->{_debug}       = $args{debug};

    bless($self, $class);

    return $self;

} ## end sub new

=head1 OTHER METHODS

=head2 verify_expected_rc

For a given client action, verify that the output message that was obtained 
corresponds to the expected return code.
Terminates the script execution if the two do not match.
This behavior can be disabled by passing 'expected_rc' as FCACHE_RC_FAILURE_NON_FATAL.
In this case, a failure to match the two return codes will not result in failure.
Instead, the obtained rc is passed back to the caller, for further processing.

=head3 ARGUMENTS

=over

=item B<command> I<Required> Scalar:

        Code for the executed command/action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_CLIENT_CMD_* constants.
        
=item B<output> I<Required> Scalar:

        Obtained output, to be matched to the expected return code.
        
=item B<expected_rc> I<Required> Scalar:

        Expected return code, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.

=item B<resource_name> I<Optional> Scalar:

        Resource name to look for in the output message
        (unless the message is empty, in which case the resource_name is ignored).
        E.g. name of a volume.

=back

=head3 RETURN

Return code (Scalar).
If expected_rc is not FCACHE_RC_FAILURE_NON_FATAL, the return will be FCACHE_RC_SUCCESS.
Otherwise, the obtained_rc will be returned, for any further processing.

EXAMPLE

	$obj->verify_expected_rc(command => FCACHE_CLIENT_CMD_FILE_OPEN,
                                 output => $output,
                                 expected_rc => $expected,);

=cut

sub verify_expected_rc {
    my $class = shift;
    my %args  = validate(
        @_,
        {   command       => {type => SCALAR,},
            output        => {type => SCALAR,},
            expected_rc   => {type => SCALAR,},
            resource_name => {type => SCALAR, optional => 1,},
        }
    );

    my $expected_rc = delete($args{expected_rc});
    my $command     = $args{command};

    my $obtained_rc = $class->_obtained_rc_lookup(%args);
    fcache_log_comment(
        message => "Client RC lookup returned: $obtained_rc for:\n"
            . "Command: $command\n")
        if ($class->{_debug});

    return fcache_verify_expected_rc(
        obtained_rc => $obtained_rc,
        expected_rc => $expected_rc,
        action      => "Client action: $command",
        output      => $args{output},
    );
} ## end sub verify_expected_rc

########################################################
# Internal methods
########################################################

# Lookup the corresponding return code for the obtained output string for a specific command.
# If match is found, return the matched return code.
# Otherwise, return FCACHE_RC_FAILURE.
# The lookup is done based on an internally-maintained list of possible messages for this command.
# Arguments: none.
# - command: Required: Scalar: Command code, as defined in FlexCache::Constants.
# - output: Required: Scalar: Obtained message, for which to find the correspondent rc.
# - resource_name: Optional: Scalar|Undef: resource name to look for in the output message
#                  (unless the message is empty, in which case the resource_name is ignored).
# Return:
# - Matched rc, if found
# - FCACHE_RC_FAILURE otherwise.
sub _obtained_rc_lookup {
    my $class = shift;
    my %args  = validate(
        @_,
        {   command       => {type => SCALAR,},
            output        => {type => SCALAR,},
            resource_name => {type => SCALAR, optional => 1,},
        }
    );

    my $command       = $args{command};
    my $output        = $args{output};
    my $resource_name = $args{resource_name};
    my $matched_rc    = FCACHE_RC_FAILURE;

    # Only print the output message once, in the beginning
    # It is too long and makes the debugging harder
    fcache_log_comment(message => "RC lookup for client output:\n"
            . "Command: $command\n"
            . "Output message: $output")
        if ($class->{_debug});

# Hash layout explained:
# For each: client type (*nix, windows) - any of the FCACHE_CLIENT_TYPE_* constants
#   command-specific information.
#   Keys are: command codes - constants FCACHE_CLIENT_CMD_*
#                  WARNING: the keys need to match the value of the constant!
#   Values: hashrefs of key => value pairs: 'error message as seen on the client' => return code
# The search for the return code that corresponds to the obtained output
# is done in the following order:
# - specific command
# - all the other commands, with the restriction that if a
#   FCACHE_RC_SUCCESS rc is found within those, it will be ignored.
#
# WARNING!
# All the entries in this hash should have as matching return code
# either FCACHE_RC_SUCCESS or a specific error code.
# Using FCACHE_RC_FAILURE or FCACHE_RC_FAILURE_NON_FATAL in this hash
# is NOT ALLOWED!!!.

    my $possible_messages;
    if ($class->{_client_type} eq FCACHE_CLIENT_TYPE_UNIX) {
        $possible_messages = {
            FILE_OPEN => {
                ''                      => FCACHE_RC_SUCCESS,
                'File exists'           => FCACHE_RC_FILE_ALREADY_EXISTS,
                'Permission denied'     => FCACHE_RC_PERMISSION_DENIED,
                'Input/output error'    => FCACHE_RC_INPUT_OUTPUT_ERROR,
                'Read-only file system' => FCACHE_RC_READ_ONLY_FILE_SYSTEM,
            },
            FILE_CLOSE => {
                ''                    => FCACHE_RC_SUCCESS,
                'Disk quota exceeded' => FCACHE_RC_DISK_QUOTA_EXCEEDED,
            },
            FILE_LOCK => {
                '' => FCACHE_RC_SUCCESS,
                'File or segment already locked by other process' =>
                    FCACHE_RC_FILE_ALREADY_LOCKED,
            },
            PIPE_OPEN => {
                ''            => FCACHE_RC_SUCCESS,
                'File exists' => FCACHE_RC_FILE_ALREADY_EXISTS,
            },
            DIR_CREATE => {
                ''            => FCACHE_RC_SUCCESS,
                'File exists' => FCACHE_RC_DIR_ALREADY_EXISTS,
            },
            DIR_DELETE => {
                ''                          => FCACHE_RC_SUCCESS,
                'No such file or directory' => FCACHE_RC_NO_ENTRY_FOUND,
            },
            DIR_LIST => {
                ''                  => FCACHE_RC_SUCCESS,
                'Permission denied' => FCACHE_RC_PERMISSION_DENIED,
            },
            FILE_WRITE => {
                ''                        => FCACHE_RC_SUCCESS,
                'No space left on device' => FCACHE_RC_NOT_ENOUGH_SPACE,
            },
            FILE_READ    => {'' => FCACHE_RC_SUCCESS,},
            FILE_SET_ACL => {
                'set attr -- went OK' => FCACHE_RC_SUCCESS,
                'access denied'       => FCACHE_RC_PERMISSION_DENIED,
            },
            FILE_DELETE => {
                ''                          => FCACHE_RC_SUCCESS,
                'No such file or directory' => FCACHE_RC_NO_ENTRY_FOUND
            },
            FILE_COPY => {
                ''                        => FCACHE_RC_SUCCESS,
                'No space left on device' => FCACHE_RC_NOT_ENOUGH_SPACE,
            },
            PATH_ATTRIBUTES_MODIFY => {
                ''                       => FCACHE_RC_SUCCESS,
                ' Read-only file system' => FCACHE_RC_READ_ONLY_FILE_SYSTEM,
            },
        };
    } elsif ($class->{_client_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        $possible_messages = {
            FILE_OPEN => {
                ''                 => FCACHE_RC_SUCCESS,
                'The file exists'  => FCACHE_RC_FILE_ALREADY_EXISTS,
                'Access is denied' => FCACHE_RC_PERMISSION_DENIED,
            },
            FILE_CLOSE => {'' => FCACHE_RC_SUCCESS,},
            FILE_LOCK  => {
                '' => FCACHE_RC_SUCCESS,
                'another process has locked a portion of the file' =>
                    FCACHE_RC_FILE_ALREADY_LOCKED,
            },
            DIR_CREATE => {
                ''            => FCACHE_RC_SUCCESS,
                'File exists' => FCACHE_RC_DIR_ALREADY_EXISTS,
            },
            DIR_DELETE => {
                ''                          => FCACHE_RC_SUCCESS,
                'No such file or directory' => FCACHE_RC_NO_ENTRY_FOUND,
            },
            DIR_LIST => {
                ''                  => FCACHE_RC_SUCCESS,
                'Permission denied' => FCACHE_RC_PERMISSION_DENIED,
            },
            FILE_WRITE => {
                '' => FCACHE_RC_SUCCESS,
                'another process has locked a portion of the file' =>
                    FCACHE_RC_FILE_ALREADY_LOCKED,
            },
            FILE_READ => {
                '' => FCACHE_RC_SUCCESS,
                'another process has locked a portion of the file' =>
                    FCACHE_RC_FILE_ALREADY_LOCKED,
            },
            FILE_SET_ACL => {'processed file: .+' => FCACHE_RC_SUCCESS,},
            FILE_DELETE  => {
                ''                          => FCACHE_RC_SUCCESS,
                'No such file or directory' => FCACHE_RC_NO_ENTRY_FOUND
            },
            FILE_COPY => {
                ''                        => FCACHE_RC_SUCCESS,
                'No space left on device' => FCACHE_RC_NOT_ENOUGH_SPACE,
            },
            PATH_ATTRIBUTES_MODIFY => {'' => FCACHE_RC_SUCCESS,},
        };
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => 'Client type is currently not supported: '
                . $class->{_client_type}
        );
    }

# First, try to find a match within the command-specific messages, if they are defined
    if (exists($possible_messages->{$command})) {
        $matched_rc = _map_message_to_rc(
            possible_messages => $possible_messages->{$command},
            obtained_message  => $output,
            resource_name     => $resource_name,
        );
        if ($matched_rc == FCACHE_RC_FAILURE) {
            fcache_log_comment(
                message => "No client command-specific messages match for:\n"
                    . "Command: $command")
                if ($class->{_debug});
        } else {
            return $matched_rc;
        }
    }

    fcache_log_comment(
              message => "Command-specific messages are not defined.\n"
            . "Trying to find a match within all the existent defined messages:\n"
            . "Command: $command")
        if ($class->{_debug});
    foreach my $attempt_code (keys(%{$possible_messages})) {
        fcache_log_comment(message =>
                "Looking for match within messages defined for $attempt_code.\n"
                . "Command: $command\n"
                . "Trying to look in:\n"
                . Dumper($possible_messages->{$attempt_code}))
            if ($class->{_debug});
        $matched_rc = _map_message_to_rc(
            possible_messages => $possible_messages->{$attempt_code},
            obtained_message  => $output,
            resource_name     => $resource_name,
        );
        if ($matched_rc == FCACHE_RC_FAILURE) {
            fcache_log_comment(message =>
                    "Looked for match within messages defined for $attempt_code:\n"
                    . "Command: $command\n"
                    . "Nothing found in:\n"
                    . Dumper($possible_messages->{$attempt_code}))
                if ($class->{_debug});
        } elsif ($matched_rc == FCACHE_RC_SUCCESS) {

            # This would be a 'success' case, but for a different command
            fcache_log_comment(
                message => "Ignoring FCACHE_RC_SUCCESS match within messages"
                    . " defined for $attempt_code:\n"
                    . "Command: $command")
                if ($class->{_debug});

            # Change it back to generic failure and keep on looking
            $matched_rc = FCACHE_RC_FAILURE;
        } else {
            return $matched_rc;
        }
    }

    # Still nothing found...
    # Give up and return generic failure
    return $matched_rc;
} ## end sub _obtained_rc_lookup

1;

