#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/UI/License.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Utility module for flexcache licensing.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::UI::License

=head1 DESCRIPTION

This is an utility module, providing licensing information for the FlexCache library.

=head1 SYNOPSIS

	TODO

=cut

package FlexCache::UI::License;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Utils qw(fcache_log_result);

BEGIN {

    use vars ('@ISA', '@EXPORT');
    require Exporter;
    @ISA    = ('Exporter');
    @EXPORT = qw(fcache_get_licensing_info
        _fcache_get_feature_mapping);
}

=head1 SUBROUTINES

=head2 fcache_get_licensing_info

Get the list of required flexcache licenses for this release and boot mode.

=head3 ARGUMENTS

=over

=item B<boot_mode> I<Required> Scalar: Boot mode, as defined in FlexCache::Constants.
Legal values: FCACHE_ONTAP_CMODE, FCACHE_ONTAP_7MODE.

=item B<version> I<Required> Scalar: Three-digit release number.

=back

=head3 RETURN

Array of required licenses (Array).
The array will contain the features that need to be licensed
(feature names, as accepted by NACL::STask::SystemLicense).

=head3 EXAMPLE

	my @required_licenses = fcache_get_licensing_info(version => '8.2.0',
                                                          boot_mode => FCACHE_ONTAP_CMODE,);
	
=cut

sub fcache_get_licensing_info {
    my (%args) = validate(
        @_,
        {   boot_mode => {type => SCALAR,},
            version   => {type => SCALAR,},
        }
    );

    my $boot_mode     = $args{boot_mode};
    my $version       = $args{version};
    my $major_release = '';
    my @licenses      = ();

    if ($version =~ /^(\d+\.\d+)\.\d+$/) {
        $major_release = $1;
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Invalid version string: $version. Cannot retrieve licensing information.\n"
                . "The version string should be the three-digit release number"
        );
    }

    my $required_licenses = {
        'CMode' => {default => ['NFS', 'CIFS',],},
        '7Mode' => {
            default => ['nfs',],
            '7.2'   => ['nfs', 'flexcache_nfs',],
            '7.2'   => ['nfs', 'flex_cache',],
            '7.1'   => ['nfs', 'flex_cache',],
            '7.0'   => ['nfs', 'flex_cache',],
        },
    };

    # Licensing info for this boot mode
    my $license_info = $required_licenses->{$boot_mode};
    if (!defined($license_info)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Unrecognized boot mode: $boot_mode. Cannot retrieve licensing information.\n"
                . "Allowed boot modes:\n"
                . Dumper(keys(%$required_licenses))
        );
    }

    # See if any minor-release specific information exists
    if (exists($license_info->{$version})) {
        @licenses = @{$license_info->{$version}};
    }

    # Else, see if any major-release information exists
    elsif (exists($license_info->{$major_release})) {
        @licenses = @{$license_info->{$major_release}};
    }

# Else, look for the default (this should be defined. Otherwise return the empty array)
    elsif (exists($license_info->{default})) {
        @licenses = @{$license_info->{default}};
    }

    return @licenses;
} ## end sub fcache_get_licensing_info

# This method should only be used within the FlexCache library
# It provides "licensing-feature" mapping between the FlexCache::Constants features
# and the feature strings accepted by NACL::STask::SystemLicense.
# This will take care of any boot-mode differences and of any licensing pre-requisites.
# Arguments:
# - feature: Required: Scalar: Feature name (constant defined in FlexCache::Constants).
# - boot_mode: Required: Scalar: Boot mode, as defined in FlexCache::Constants.
#              Legal values: FCACHE_ONTAP_CMODE, FCACHE_ONTAP_7MODE.
# - include_prerequisites: Optional: Boolean: Flag to decide if any prerequired features should be included in the returned list.
#                          Default value: FALSE.
# Return: Array of feature strings, as accepted by NACL::STask::SystemLicense.
sub _fcache_get_feature_mapping {
    my (%args) = validate(
        @_,
        {   feature               => {type => SCALAR,},
            boot_mode             => {type => SCALAR,},
            include_prerequisites => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $feature               = $args{feature};
    my $boot_mode             = $args{boot_mode};
    my $include_prerequisites = $args{include_prerequisites};
    my @return_features       = ();

# Licensing info hashref
# Its fields should be as follows:
# For each boot-mode/feature constant (as defined in FlexCache::Constants):
# - prerequisites = Arrayref of any prerequisites for this feature.
#   If this is defined and the include_prerequisites flag is true,
#   then this will be the base of the return array.
# - feature_string = feature string, as accepted by NACL::STask::SystemLicense.
#   If none is defined, it is considered that the feature does not have a corresponding license for this mode,
#   so no feature string is added to the return array.
#   If defined, then this string is added to the return array.
# If, for a specific boot-mode/feature constant, there is no entry in the $license_mapping hash,
# Then the return array will simply contain the feature string, as passed in by the caller.
    my $license_mapping = {
        'CMode' => {
            'NFS'  => {feature_string => 'NFS',},
            'CIFS' => {feature_string => 'CIFS',},
        },
        '7Mode' => {
            'NFS'  => {feature_string => 'nfs',},
            'CIFS' => {feature_string => 'cifs',},
        },
    };

    if (exists($license_mapping->{$boot_mode}->{$feature})) {
        my $feature_info = $license_mapping->{$boot_mode}->{$feature};
        if (   ($include_prerequisites)
            && (exists($feature_info->{prerequisites}))) {
            @return_features = @{$feature_info->{prerequisites}};
        }
        if (defined($feature_info->{feature_string})) {
            push(@return_features, $feature_info->{feature_string});
        }
    } else {

        # No extra mapping is needed, simply return the feature as is
        push(@return_features, $feature);
    }

    return @return_features;
} ## end sub _fcache_get_feature_mapping

1;
