#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/UI/OntapMessages.pm#2 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Utility library for flexcache UI output processing.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::UI::OntapMessages

=head1 DESCRIPTION

This is an utility module, to be used by the FlexCache library
whenever ONTAP output needs to be validated (negative testing).
The module is release agnostic and provides functionality for both 7-mode and C-mode.
It is intended to be only used internally by the FlexCache library.

=head1 SYNOPSIS

        use FlexCache::Constants qw(:all);
        use FlexCache::Constants qw(:_internal);
        use FlexCache::UI::OntapMessages;
        
        my $ui_messages = FlexCache::UI::OntapMessages->new(boot_mode => FCACHE_ONTAP_CMODE,
                                                            release => '8.2.0',);

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl
        flexcache/library_unit_tests/Snapmirror.thpl
        
=cut

package FlexCache::UI::OntapMessages;

# Standard libs
use strict;
use warnings;
use Data::Dumper;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Utils qw(fcache_log_comment
    fcache_log_result
    fcache_verify_expected_rc
    _map_message_to_rc);

=head1 CONSTRUCTOR

=head2 new

Returns a FlexCache::UI::OntapMessages object, to be used 
whenever ONTAP output needs to be validated (negative testing).

=head3 ARGUMENTS

=over

=item B<boot_mode> I<Required> Scalar:

        ONTAP boot mode, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_ONTAP_*MODE constants.
        
=item B<release> I<Required> Scalar:

        Three-digit release number.
        
=item B<debug> I<Optional> Scalar:

        Flag to increase the verbosity of this module, useful for debugging purposes.
        Default value: FALSE.

=back

=head3 EXAMPLE

        my $ui_messages = FlexCache::UI::OntapMessages->new(boot_mode => FCACHE_ONTAP_CMODE,
                                                            release => '8.2.0',
                                                            debug => TRUE,);

=cut

sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   boot_mode => {type => SCALAR,},
            release   => {type => SCALAR,},
            debug     => {type => SCALAR, default => FALSE,},
        }
    );

    my $self = {};
    $self->{_boot_mode} = $args{boot_mode};
    $self->{_debug}     = $args{debug};

    if (   ($self->{_boot_mode} ne FCACHE_ONTAP_7MODE)
        && ($self->{_boot_mode} ne FCACHE_ONTAP_CMODE)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "FlexCache library internal error:\n"
                . "Illegal ONTAP boot mode: "
                . $self->{_boot_mode}
        );
    }

    my $minor_release = $args{release};
    if ($minor_release =~ /^(\d+\.\d+)\.\d+$/) {
        $self->{_major_release} = $1;
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "FlexCache library internal error:\n"
                . "Illegal minor release format: $minor_release"
        );
    }
    $self->{_minor_release} = $minor_release;

    bless($self, $class);

    return $self;

} ## end sub new

=head1 OTHER METHODS

=head2 verify_expected_rc

For a given command, verify that the output message that was obtained 
corresponds to the expected return code.
Terminates the script execution if the two do not match.
This behavior can be disabled by passing 'expected_rc' as FCACHE_RC_FAILURE_NON_FATAL.
In this case, a failure to match the two return codes will not result in failure.
Instead, the obtained rc is passed back to the caller, for further processing.

=head3 ARGUMENTS

=over

=item B<command> I<Required> Scalar:

        Command code, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_CMD_* constants.
        
=item B<output> I<Required> Scalar:

        Obtained output, to be matched to the expected return code.
        
=item B<expected_rc> I<Required> Scalar:

        Expected return code, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.

=item B<resource_name> I<Optional> Scalar:

        Resource name to look for in the output message
        (unless the message is empty, in which case the resource_name is ignored).
        E.g. name of a volume.

=item B<ui_type> I<Required> Scalar:

        UI type, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_UI_TYPE_* constants.
	Default value: FCACHE_UI_TYPE_CLI.

=back

=head3 RETURN

Return code (Scalar).
If expected_rc is not FCACHE_RC_FAILURE_NON_FATAL, the return will be FCACHE_RC_SUCCESS.
Otherwise, the obtained_rc will be returned, for any further processing.

EXAMPLE

	$obj->verify_expected_rc(command => FCACHE_CMD_SNAPMIRROR_NEW,
                                 output => $output,
                                 expected_rc => $expected,
                                 resource_name => $destination_vol_name,
                                 ui_type => FCACHE_UI_TYPE_CLI);

=cut

sub verify_expected_rc {
    my $class = shift;
    my %args  = validate(
        @_,
        {   command       => {type => SCALAR,},
            output        => {type => SCALAR,},
            expected_rc   => {type => SCALAR,},
            resource_name => {type => SCALAR, optional => 1,},
            ui_type       => {type => SCALAR, default => FCACHE_UI_TYPE_CLI},
        }
    );

    my $expected_rc = delete($args{expected_rc});
    my $command     = $args{command};
    my $ui_type     = $args{ui_type};

    my $obtained_rc = $class->_obtained_rc_lookup(%args);
    fcache_log_comment(message => "RC lookup returned: $obtained_rc for:\n"
            . "Command: $command\n"
            . "UI type: $ui_type")
        if ($class->{_debug});

    return fcache_verify_expected_rc(
        obtained_rc => $obtained_rc,
        expected_rc => $expected_rc,
        action      => "$ui_type command: $command",
        output      => $args{output},
    );
} ## end sub verify_expected_rc

=head2 get_constant_value

This method is to be used for obtaining values for
release or boot mode specific constants.
Depending on the current release and boot mode, the
requested constant is looked-up internally and returned to the caller
for further use.

=head3 ARGUMENTS

=over

=item B<constant> I<Required> Scalar:

        Constant name/code.

=back

=head3 RETURN

Constant value (Scalar).

EXAMPLE

	$obj->get_constant_value(constant => 'FCACHE_DELEGATIONS_LOCK_PROTOCOL');

=cut

sub get_constant_value {
    my $class = shift;
    my %args = validate(@_, {constant => {type => SCALAR,},});

    my $constant      = $args{constant};
    my $major_release = $class->{_major_release};
    my $minor_release = $class->{_minor_release};
    my $value;

    # Possible hash keys:
    # - minor release (e.g. '8.2.0'): will be considered first
    # - major release (e.g. '8.2'): will be considered second
    # - default: will be considered last.
    my $constants_table = {
        default => {
            FCACHE_DELEGATIONS_LOCK_PROTOCOL     => 'flexcache_deleg',
            FCACHE_VOL_DELEGATION_MODE_READ      => 'Read',
            FCACHE_VOL_DELEGATION_MODE_WRITE     => 'Write',
            FCACHE_VOL_DELEGATION_STATE_LOCKED   => 'LOCKED',
            FCACHE_VOL_DELEGATION_STATE_UNLOCKED => 'UNLOCKED',
            FCACHE_VOL_DELEGATION_STATE_GRANTED  => 'granted',
            FCACHE_VOL_OPLOCK_LEVEL_READ_BATCH   => 'read-batch',
            FCACHE_VOL_OPLOCK_LEVEL_LEVEL2       => 'level2',
            FCACHE_VOL_OPLOCK_LEVEL_NONE         => '-',
            FCACHE_VOL_OPLOCK_TYPE_BYTE_RANGE    => 'byte-range',
            FCACHE_VOL_OPLOCK_TYPE_SHARE_LEVEL   => 'share-level',
        },
    };

    if (exists($constants_table->{$minor_release}->{$constant})) {
        $value = $constants_table->{$minor_release}->{$constant};
    } elsif (exists($constants_table->{$major_release}->{$constant})) {
        $value = $constants_table->{$major_release}->{$constant};
    } elsif (exists($constants_table->{default}->{$constant})) {
        $value = $constants_table->{default}->{$constant};
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "FlexCache library or script problem found\n"
                . "Please make sure that the script is only using the "
                . "documented legal values for all the arguments.\n"
                . "If this problem is eliminated, then:\n"
                . "Please file a TESTLIB burt:\n"
                . "No UI constant value defined for \"$constant\""
        );
    }

    fcache_log_comment(message => "Looked up UI-dependent constant value:\n"
            . "Constant name: $constant\n"
            . "Constant value: $value")
        if ($class->{_debug});
    return $value;
} ## end sub get_constant_value

=head2 verify_wafliron_detected_corruption

Verify in the wafliron output that a specific set of corruptions
was detected.

=head3 ARGUMENTS

=over
        
=item B<output> I<Required> Scalar:

        Obtained wafliron output, in which to look for specific corruption
	messages.
        
=item B<expected_corruptions> I<Required> Arrayref:

        Arayref of expected corruption codes.
        Legal values for the corruption codes: any of the FCACHE_CORRUPTION_*
	constants (defined in FlexCache::Constants).

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

	$obj->verify_wafliron_detected_corruption(
	        output => $wafliron_output,
		expected_corruptions => [FCACHE_CORRUPTION_QTREE_METADATA]);

=cut

sub verify_wafliron_detected_corruption {
    my $class = shift;
    my %args  = validate(
        @_,
        {   output               => {type => SCALAR,},
            expected_corruptions => {type => ARRAYREF,},
        }
    );

    my $expected_corruptions = $args{expected_corruptions};
    my $wafliron_output      = $args{output};
    my $boot_mode            = $class->{_boot_mode};
    my $major_release        = $class->{_major_release};
    my $minor_release        = $class->{_minor_release};
    my $corruption_messages;

    if ($boot_mode eq FCACHE_ONTAP_7MODE) {
        $corruption_messages = {
            QTREE_METADATA => {default => ['Correcting .+ in qtree file',],},
            WI_INDIRECT_BLK_CNT =>
                {default => ['indirect cnt is .+ Correcting it to',],},
            FSINFO_BLKS_USED =>
                {default => ['Total blocks used corrected',],},
        };
    } else {    # FCACHE_ONTAP_CMODE
        $corruption_messages = {
            QTREE_METADATA => {default => ['Correcting .+ in qtree file',],},
            WI_INDIRECT_BLK_CNT =>
                {default => ['indirect cnt is .+ Correcting it to',],},
            VOL_SNAPDIR =>
                {default => ['Directory name index inconsistencies',],},
            FSINFO_BLKS_USED =>
                {default => ['Total blocks used corrected',],},
        };
    }

    foreach my $corruption (@{$expected_corruptions}) {

        # Look-up the expected messages for this corruption.
        # First, try the minor release
        my $expected_messages
            = $corruption_messages->{$corruption}->{$minor_release};
        if (!defined($expected_messages)) {

            # Try the major release
            $expected_messages
                = $corruption_messages->{$corruption}->{$major_release};
        }
        if (!defined($expected_messages)) {

            # Try the default
            $expected_messages
                = $corruption_messages->{$corruption}->{default};
        }
        if (!defined($expected_messages)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "FlexCache library internal error:\n"
                    . "Expected wafliron messages not defined for:\n"
                    . "Corruption code: $corruption"
            );
        }

        foreach my $expected_string (@{$expected_messages}) {
            if ($wafliron_output !~ /$expected_string/) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message =>
                        "Wafliron did not detect expected corruption:\n"
                        . "Corruption code: $corruption\n"
                        . "Could not detect wafliron message: \"$expected_string\""
                );
            }
        }

        fcache_log_comment(message => "Wafliron detected the corruption:\n"
                . "Corruption code: $corruption");
    }

    return FCACHE_RC_SUCCESS;
} ## end sub verify_wafliron_detected_corruption

########################################################
# Internal methods
########################################################

# Lookup the corresponding return code for the obtained
# output string for a specific command.
# If match is found, return the matched return code.
# Otherwise, return FCACHE_RC_FAILURE.
# The lookup is done based on an internally-maintained list
# of possible messages for this command.
# Arguments: none.
# - command: Required: Scalar: Command code, as defined in FlexCache::Constants.
# - output: Required: Scalar: Obtained message, for which to find the correspondent rc.
# - resource_name: Optional: Scalar|Undef: resource name to look for in the output message
#                  (unless the message is empty, in which case the resource_name is ignored).
# - ui_type: Optional: Scalar: UI type.
#            Legal values: any of the FCACHE_UI_TYPE_* constants.
#            Default value: FCACHE_UI_TYPE_CLI.
# Return:
# - Matched rc, if found
# - FCACHE_RC_FAILURE otherwise.
sub _obtained_rc_lookup {
    my $class = shift;
    my %args  = validate(
        @_,
        {   command       => {type => SCALAR,},
            output        => {type => SCALAR,},
            resource_name => {type => SCALAR, optional => 1,},
            ui_type       => {type => SCALAR, default => FCACHE_UI_TYPE_CLI},
        }
    );

    my $command       = $args{command};
    my $output        = $args{output};
    my $ui_type       = $args{ui_type};
    my $resource_name = $args{resource_name};
    my $boot_mode     = $class->{_boot_mode};
    my $major_release = $class->{_major_release};
    my $minor_release = $class->{_minor_release};
    my $matched_rc    = FCACHE_RC_FAILURE;

# Hash layout explained:
# For each: ui type (CLI, ZAPI) - any of the FCACHE_ONTAP_*MODE constants
#  For each: boot mode (CMode, 7Mode) - any of the FCACHE_UI_TYPE_* constants
#   For each: command names - any of the FCACHE_CMD_* constants:
#    default or release-specific information for a specific command:
#    - keys can be : default, major-release number (e.g. '8.2'), minor-release number (e.g. '8.2.0')
#    - values should be hashrefs to keys <-> values: 'error to match' <-> return code
# For each command, the search for the return code that corresponds to the obtained output
# is done in the following order:
# - minor-release specific messages, if defined
# - major-release specific messages, if defined
# - default messages, if defined
#
# WARNING!
# All the entries in this hash should have as matching return code
# either FCACHE_RC_SUCCESS or a specific error code.
# Using FCACHE_RC_FAILURE or FCACHE_RC_FAILURE_NON_FATAL in this hash
# is NOT ALLOWED!!!.

# Go through a series of conditionals here (instead of defining a huge hash for everything),
# so we only load a smaller hash each time we check for an rc.
    my $possible_messages;
    if ($ui_type eq FCACHE_UI_TYPE_CLI) {
        if ($boot_mode eq FCACHE_ONTAP_7MODE) {
            if ($command eq FCACHE_CMD_SNAPMIRROR_NEW) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Transfer aborted: destination volume is sparse' =>
                            FCACHE_BRE_NOT_POSSIBLE_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VFILER_FIND) {
                $possible_messages = {
                    default => {
                        ''                 => FCACHE_RC_SUCCESS,
                        'Vfiler not found' => FCACHE_RC_NO_ENTRY_FOUND,
                    },
                };
            } elsif ($command eq FCACHE_CMD_DELEGATIONS_FIND) {
                $possible_messages = {default => {'' => FCACHE_RC_SUCCESS,},};
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Snapshot operation not allowed while sparse' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_DELETE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Snapshot operation not allowed while sparse' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_RESTORE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        '.+ is a sparse volume. It cannot be restored' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_MOVE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Specified source volume is a flexcache volume' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_QTREE_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'remote volume .+ does not support local qtrees' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_SE_ENABLE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Volume type has to be Flexvol' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_FILE_CLONE_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'SIS clone not supported on this kind of volume' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_QUOTA_ENABLE) {
                $possible_messages = {default => {'' => FCACHE_RC_SUCCESS,},};
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_RESTORE_FILE) {
                $possible_messages = {default => {'' => FCACHE_RC_SUCCESS,},};
            }
        } else {    # FCACHE_ONTAP_CMODE
            if ($command eq FCACHE_CMD_SNAPMIRROR_NEW) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Volume .+ is of type "CACHE_RW", which is not supported by SnapMirror'
                            => FCACHE_BRE_NOT_POSSIBLE_ON_A_CACHE,
                        'Source of a load-sharing relationship cannot be the origin of a cache'
                            => FCACHE_RC_LS_SOURCE_CANNOT_BE_AN_ORIGIN
                    },
                };
            } elsif ($command eq FCACHE_CMD_DELEGATIONS_FIND) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'No matching volume flexcache delegations.* found' =>
                            FCACHE_RC_NO_ENTRY_FOUND,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_FLEXCACHE_POLICY_SET) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Modification of .+ flexcache-cache-policy not allowed for .+ Read-Write volume'
                            => FCACHE_RC_NOT_ALLOWED_ON_RW_VOL,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VSERVER_LOCKS_FIND) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'No matching vserver locks.* found' =>
                            FCACHE_RC_NO_ENTRY_FOUND,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Snapshot copies can only be created on Read_Write .* volumes'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_DELETE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Snapshot copies can only be deleted on Read_Write .* volumes'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_RESTORE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Only a Snapshot copy of a read-write volume can be promoted'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_MOVE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        ' This operation is not supported on a volume of type "DC"'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_OPLOCKS_MODE_SET) {
                $possible_messages = {
                    default => {
                        '\d+ entr.+ acted on' => FCACHE_RC_SUCCESS,
                        'Volume is FlexCache' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_QTREE_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Qtree creation not supported on FlexCache volume' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_QTREE_STATISTICS) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'No matching volume qtree.* found' =>
                            FCACHE_RC_NO_ENTRY_FOUND
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_FLEXCACHE_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        ' Origin can only be a regular .* or snapmirror .* volume'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_MODIFY) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'invalid argument .+' =>
                            FCACHE_RC_INVALID_CMD_ARGUMENT,
                        'not allowed for volumes of the type .*Cache Read-Write'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                        'not allowed for volumes of the type .+Read-Write volume'
                            => FCACHE_RC_NOT_ALLOWED_ON_RW_VOL,
                    },
                };
            } elsif ($command eq FCACHE_CMD_SE_ENABLE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'Volume type has to be Flexvol' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_FILE_CLONE_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        'SIS clone not supported on this kind of volume' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_FLEXCACHE_DELETE) {
                $possible_messages = {default => {'' => FCACHE_RC_SUCCESS,},};
            } elsif ($command eq FCACHE_CMD_VOL_QUOTA_ENABLE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        ' Quota Rules can be created only for RW Volumes' =>
                            FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_SNAPSHOT_RESTORE_FILE) {
                $possible_messages = {default => {'' => FCACHE_RC_SUCCESS,},};
            }
        }
    } elsif ($ui_type eq FCACHE_UI_TYPE_ZAPI) {
        if ($boot_mode eq FCACHE_ONTAP_7MODE) {

        } else {    # FCACHE_ONTAP_CMODE
            if ($command eq FCACHE_CMD_VOL_FLEXCACHE_CREATE) {
                $possible_messages = {
                    default => {
                        '' => FCACHE_RC_SUCCESS,
                        ' Origin can only be a regular .* or snapmirror .* volume'
                            => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE
                    },
                };
            } elsif ($command eq FCACHE_CMD_VOL_FLEXCACHE_DELETE) {
                $possible_messages = {default => {'' => FCACHE_RC_SUCCESS,},};
            }
        }
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Illegal UI type for command:\n"
                . "Command: \"$command\"\n"
                . "UI type: \"$ui_type\"\n"
        );
    }

    if (!defined($possible_messages)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "FlexCache library internal error:\n"
                . "Unspecified possible messages for:\n"
                . "Command: \"$command\"\n"
                . "UI type: \"$ui_type\"\n"
                . "Boot mode: \"$boot_mode\""
        );
    }

    fcache_log_comment(message =>
            "Looking for matching output message for command $command:\n"
            . Dumper($possible_messages))
        if ($class->{_debug});

    # If minor-release specific messages are defined, look through those first
    if (defined($possible_messages->{$minor_release})) {
        $matched_rc = _map_message_to_rc(
            possible_messages => $possible_messages->{$minor_release},
            obtained_message  => $output,
            resource_name     => $resource_name,
        );
        if ($matched_rc == FCACHE_RC_FAILURE) {
            fcache_log_comment(
                      message => "No minor-release messages match for:\n"
                    . "Command: $command\n"
                    . "Output message: $output")
                if ($class->{_debug});
        } else {
            return $matched_rc;
        }
    } elsif ($class->{_debug}) {
        fcache_log_comment(
            message => "No minor-release specific messages are defined for:\n"
                . "Command: $command\n"
                . "Looking for major-release messages...")
            if ($class->{_debug});
    }

    # There was no match within the minor-release messages
    # Try the major release
    if (defined($possible_messages->{$major_release})) {
        $matched_rc = _map_message_to_rc(
            possible_messages => $possible_messages->{$major_release},
            obtained_message  => $output,
            resource_name     => $resource_name,
        );
        if ($matched_rc == FCACHE_RC_FAILURE) {
            fcache_log_comment(
                      message => "No minor-release messages match for:\n"
                    . "Command: $command\n"
                    . "Output message: $output")
                if ($class->{_debug});
        } else {
            return $matched_rc;
        }
    } elsif ($class->{_debug}) {
        fcache_log_comment(
            message => "No major-release specific messages are defined for:\n"
                . "Command: $command\n"
                . "Looking for default messages...")
            if ($class->{_debug});
    }

    # There was no match within the major-release messages either
    # Try the defaults
    if (defined($possible_messages->{default})) {
        $matched_rc = _map_message_to_rc(
            possible_messages => $possible_messages->{default},
            obtained_message  => $output,
            resource_name     => $resource_name,
        );
        if ($matched_rc == FCACHE_RC_FAILURE) {
            fcache_log_comment(
                      message => "No minor-release messages match for:\n"
                    . "Command: $command\n"
                    . "Output message: $output")
                if ($class->{_debug});
        } else {
            return $matched_rc;
        }
    } elsif ($class->{_debug}) {
        fcache_log_comment(message => "No default messages are defined for:\n"
                . "Command: $command\n"
                . "Giving up...")
            if ($class->{_debug});
    }

    # No match within the defaults either.
    # Give up and return generic failure
    return $matched_rc;
} ## end sub _obtained_rc_lookup

1;

