#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Utils.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Module providing multiple utilitiy functions for the FlexCache library.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Utils

=head1 DESCRIPTION

This module provides multiple utilitiy functions for
the FlexCache library and scripts.

=head1 SYNOPSIS

use Flexcache::Utils qw(<subroutines to use>);

=cut

package FlexCache::Utils;

use strict;
use warnings;

use Tharn;
use Data::Dumper;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);

use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);

BEGIN {

    use vars ('@ISA', '@EXPORT');
    require Exporter;
    @ISA    = ('Exporter');
    @EXPORT = qw(fcache_log_comment
        fcache_log_result
        fcache_get_random_name
        fcache_verify_expected_rc
        fcache_compare_hashes
        fcache_convert_size_string_to_value
        _map_message_to_rc);
}

=head1 SUBROUTINES

=head2 fcache_log_comment

Prints a comment to the script log.

=head3 ARGUMENTS

=over

=item B<message> I<Required> Scalar:

        Text message to be logged.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	fcache_log_comment(message => "This is a comment");
	
=cut

sub fcache_log_comment {
    my (%args) = validate(@_, {message => {type => SCALAR,},});

    my $delimitator = "-" x 80;
    my $log_message = "$delimitator\n" . $args{message} . "\n$delimitator";
    $Tharn::Log->comment($log_message);

    return FCACHE_RC_SUCCESS;
} ## end sub fcache_log_comment

=head2 fcache_log_result

Terminates the script execution, with the required result type.

=head3 ARGUMENTS

=over

=item B<message> I<Required> Scalar:

        Text message to be printed out.

=item B<result_type> I<Required> Scalar:

        Type of result to mark the execution with.
        Legal values: Any of the FCACHE_LOGRESULT_* constants
	(defined in FlexCache::Constants).

=back

=head3 RETURN

None (Script execution is terminated).

=head3 EXAMPLE

	fcache_log_result(result_type => FCACHE_LOGRESULT_PASS,
                          message => "Test-case .... passed");
	
=cut

sub fcache_log_result {
    my (%args) = validate(
        @_,
        {   message     => {type => SCALAR,},
            result_type => {type => SCALAR,},
        }
    );

    my $result_type = $args{result_type};
    my $message     = $args{message};

    if ("$result_type" eq FCACHE_LOGRESULT_PASS) {
        $Log->log(pass_result($message));
    } elsif ("$result_type" eq FCACHE_LOGRESULT_NA) {
        $Log->log(na_result('Not Applicable: ' . $message));
    } elsif ("$result_type" eq FCACHE_LOGRESULT_FATAL) {
        $Log->log(fatal_result($message));
    } elsif ("$result_type" eq FCACHE_LOGRESULT_SCRIPT) {
        $Log->log(script_result($message));
    } elsif ("$result_type" eq FCACHE_LOGRESULT_CONFIG) {
        $Log->log(config_result('Configuration: ' . $message));
    } else {
        $Log->log(
            script_result(
                      "Illegal result type used within the FlexCache scripts:"
                    . " $result_type\n"
                    . "encountered while attempting to log:\n$message"
            )
        );
    }

    # Exit here. This is because:
    # In the non-pass case, we want the script execution to stop
    # ($Log->log(*) only adds the correspondent result to this execution,
    # but does not halt it)
    # Even in the pass case, exiting will enforce the rule of
    # "only one pass per script"
    exit;
} ## end sub fcache_log_result

=head2 fcache_get_random_name

Generates a random name, based on the current time.
The generated name will be of form: type<random_integer>_yyyymmdd_hhmmss.
E.g. vol150567_20120326_115905.

=head3 ARGUMENTS

=over

=item B<type> I<Required> Scalar:
        
        Type, to be prepended to the randomly-generated name.
        E.g. vol, aggr.

=back

=head3 RETURN

Newly-generated name (Scalar).

=head3 EXAMPLE

	my $random_vol_name = fcache_get_random_name(type => 'vol');
	
=cut

sub fcache_get_random_name {
    my (%args) = validate(@_, {type => {type => SCALAR,},});

    my $type = $args{type};
    my $time = time();
    my ($sec, $min, $hour, $mday, $mon, $year,) = localtime($time);
    $mon++;
    $year += 1900;
    my $seed  = int(scalar($time));
    my $range = $year * $mon * $mday;

    srand($seed);
    my $rand_num = int(rand($range));

    my $new_name = sprintf("%s%d_%04d%02d%02d_%02d%02d%02d",
        $type, $rand_num, $year, $mon, $mday, $hour, $min, $sec);

    return $new_name;
} ## end sub fcache_get_random_name

=head2 fcache_verify_expected_rc

Check if obtained/expected return codes match.
Output processing tool. Provides standardized subroutine output and
pass/fail behavior for the FlexCache library.
Terminates the script execution if 'obtained_rc' is different than 'expected_rc'.
This behavior can be disabled by passing 'expected_rc' as
FCACHE_RC_FAILURE_NON_FATAL.
In this case, a failure to match the two return codes will not result in failure.
Instead, the obtained rc is passed back to the caller, for further processing.

=head3 ARGUMENTS

=over

=item B<obtained_rc> I<Required> Scalar:

        Obtained return code.
        Constants for return codes supported by the FlexCache library
        are defined in FlexCache::Constants.

=item B<expected_rc> I<Required> Scalar:

        Expected return code.
        Constants for return codes supported by the FlexCache library
        are defined in FlexCache::Constants.

=item B<action> I<Required> Scalar:

        Few-words description of the action being processed.
        This is used for logging purposes only.

=item B<output> I<Optional> Undef|Scalar|Arrayref|Hashref:

        Output, to be logged along with the failure message.

=back

=head3 RETURN

Return code (Scalar).
If expected_rc is not FCACHE_RC_FAILURE_NON_FATAL,
the return will be FCACHE_RC_SUCCESS.
Otherwise, the obtained_rc will be returned, for any further processing.

=head3 EXAMPLE

	fcache_verify_expected_rc(obtained_rc => $obtained_rc,
                                  expected_rc => FCACHE_RC_SUCCESS,
                                  action => "Detect cache volumes having $my_vol as origin",
                                  output => $output_obtained_from_filer,);
	
=cut

sub fcache_verify_expected_rc {
    my (%args) = validate(
        @_,
        {   obtained_rc => {type => SCALAR,},
            expected_rc => {type => SCALAR,},
            action      => {type => SCALAR,},
            output =>
                {type => UNDEF | SCALAR | ARRAYREF | HASHREF, optional => 1,},
        }
    );

    my $rc       = $args{obtained_rc};
    my $expected = $args{expected_rc};

    if ($expected == FCACHE_RC_FAILURE_NON_FATAL) {

        # Non-fatal failures, needed for some internal handling
        # Do nothing here. Rc stays unchanged for the caller to handle.
    } elsif ($rc == $expected) {
        $rc = FCACHE_RC_SUCCESS;
    } else {
        my $fail_comment
            = "Failure while executing \""
            . $args{action} . "\"\n"
            . "Expected: $expected,\n"
            . "Obtained: $rc";

        if (defined($args{output})) {
            $fail_comment
                .= "\n\nCommand output:\n\n" . Dumper($args{output});
        }

        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => $fail_comment
        );
    }

    return $rc;
} ## end sub fcache_verify_expected_rc

=head2 fcache_compare_hashes

Verify that all the key-value pairs in the expected hash
are present in the obtained hash and that their values match.
Unless the 'identical_match' flag is set to true,
the obtained hash can be a super-set of the expected hash.

=head3 ARGUMENTS

=over

=item B<obtained> I<Required> Hashref:

        Hash reference, representing the fields obtained
        after executing a specific command/action.

=item B<expected> I<Required> Hashref:

        Hash reference, representing the fields
        that are expected to be obtained.

=item B<hash_description> I<Required> Scalar:

        Few-words description of the hash.
        This is used for logging purposes only.

=item B<identical_match> I<Optional> Boolean:

        Flag to require that the two hashes should be identical.
        If this is set to true, then the two hashes are required to have
        the same number of elements and that all the key-value pairs match.
        Default value: FALSE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

=head3 EXAMPLE

	fcache_compare_hashes(obtained => $ref_to_obtained_hash,
                              expected => $ref_to_expected_hash,
                              hash_description => "File statistics",);
	
=cut

sub fcache_compare_hashes {
    my (%args) = validate(
        @_,
        {   obtained         => {type => HASHREF,},
            expected         => {type => HASHREF,},
            hash_description => {type => SCALAR,},
            identical_match  => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $obtained         = $args{obtained};
    my $expected         = $args{expected};
    my $hash_description = $args{hash_description};

    if ($args{identical_match}) {
        my $nr_keys_obtained = scalar(keys(%{$obtained}));
        my $nr_keys_expected = scalar(keys(%{$expected}));

        if ($nr_keys_obtained != $nr_keys_expected) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Hash comparison failed for:\n"
                    . "$hash_description.\n"
                    . "Reason:\n"
                    . "The two hashes have different elements counts:\n"
                    . "Expected: $nr_keys_expected\n"
                    . "Obtained: $nr_keys_obtained"
            );
        }
    }

    foreach my $key (keys(%{$expected})) {
        if (!exists($obtained->{$key})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Hash comparison failed for:\n"
                    . "$hash_description.\n"
                    . "Reason:\n"
                    . "Obtained hash is missing the entry for: $key\n"
                    . Dumper($obtained)
            );
        }

        my $obtained_value = $obtained->{$key};
        my $expected_value = $expected->{$key};
        fcache_log_comment(message => "Comparing values for:\n"
                . "$hash_description\n"
                . "Field: $key\n"
                . "Expected: $expected_value\n"
                . "Obtained: $obtained_value");

        # String-ify... to be able to compare both strings and integers
        if ("$obtained_value" ne "$expected_value") {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Hash comparison failed for:\n"
                    . "$hash_description.\n"
                    . "Reason:\n"
                    . "Value mismatch for field: $key\n"
                    . "Expected: $expected_value\n"
                    . "Obtained: $obtained_value"
            );
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub fcache_compare_hashes

=head2 fcache_convert_size_string_to_value

Convert a data size string to the integer value.
E.g. converts 100m to 104857600.

=head3 ARGUMENTS

=over

=item B<size_string> I<Required> Scalar:

        Data size string, given in the following form:
	<integer><modifier>, where modifier can be one of: k,m,g.
	E.g. 100m, 32k, 2g.

=back

=head3 RETURN

Converted data size (in bytes), as an integer value.

=head3 EXAMPLE

	my $size_in_bytes = fcache_convert_size_string_to_value(size_string => '150m',);
	
=cut

sub fcache_convert_size_string_to_value {
    my (%args) = validate(@_, {size_string => {type => SCALAR,},});

    my $size_string = $args{size_string};
    my ($modifier, $size_in_bytes);

    # Extract modifier
    if ($size_string =~ /^(\d+)(\w+)$/) {
        $size_in_bytes = $1;
        $modifier      = $2;
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "Unsupported data size string: $size_string"
        );
    }

    if ($modifier eq 'k') {
        $size_in_bytes = $size_in_bytes * KB;
    } elsif ($modifier eq 'm') {
        $size_in_bytes = $size_in_bytes * MB;
    } elsif ($modifier eq 'g') {
        $size_in_bytes = $size_in_bytes * GB;
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "Unsupported modifier for data size ($size_string):"
                . " $modifier"
        );
    }

    return $size_in_bytes;
} ## end sub fcache_convert_size_string_to_value

# Search through a 'message' => rc hash for a possible match.
# If match is found, return the matched return code.
# Otherwise, return FCACHE_RC_FAILURE.
# Arguments: none.
# - possible_messages: Required: Hashref: Reference to hash of key-value: message-rc.
# - obtained_message: Required: Scalar: Obtained message, for which to find the correspondent rc.
# - resource_name: Optional: Scalar|Undef: resource name to look for in the error message.
# Return:
# - Matched rc, if found
# - FCACHE_RC_FAILURE otherwise.
sub _map_message_to_rc {
    my %args = validate(
        @_,
        {   possible_messages => {type => HASHREF,},
            obtained_message  => {type => SCALAR,},
            resource_name     => {type => SCALAR | UNDEF, optional => 1,},
        }
    );

    my $possible_messages = $args{possible_messages};
    my $obtained_message  = $args{obtained_message};
    my $resource_name     = $args{resource_name};
    my $matched_rc        = FCACHE_RC_FAILURE;

    # Need to treat the empty string differently (match, not regex)
    if ($obtained_message eq '') {
        if (exists($possible_messages->{''})) {
            $matched_rc = $possible_messages->{''};
        }
    } else {
        foreach my $possible_message (keys(%{$possible_messages})) {
            if (   ($possible_message ne '')
                && ($obtained_message =~ /$possible_message/)) {
                if (   (defined($resource_name))
                    && ($obtained_message !~ /$resource_name/)) {
                    fcache_log_comment(message => "WARNING:\n"
                            . "Obtained error message was matched to"
                            . " return code $matched_rc\n"
                            . "Expected resource name was not matched"
                            . " in this error message.\n"
                            . "For this reason, the return code match"
                            . " is being ignored....\n"
                            . "Resource name: $resource_name\n"
                            . "Error message: $obtained_message");
                } else {
                    $matched_rc = $possible_messages->{$possible_message};
                    last;
                }
            }
        }
    }

    return $matched_rc;
} ## end sub _map_message_to_rc

1;
