#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Volume.pm#2 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Agnostic module for volume-level functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Volume

=head1 DESCRIPTION

This module provides release and mode agnostic volume-level functionality for all the flexcache scripts.
Its functionality is made available through the 'volume' member of the FlexCache::Origin object.
The module is release agnostic and provides functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though, that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

=head1 SYNOPSIS

This module is intended to be used only through the FlexCache::Origin object, as follows:

	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
	$origin->{volume}->call_any_method_but_the_constructor();
	
Using this module independently is possible, with the risk that the constructor's arguments may change at any time.

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl

=cut

package FlexCache::Volume;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate validate_with SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Constants qw(:_internal);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result
    fcache_verify_expected_rc fcache_get_random_name);
use NATE::BaseException qw(:try);
use NACL::STask::Volume;
use NACL::STask::Node;
use NACL::STask::VserverCifsShare;
use NACL::STask::VolumeSnapshot;
use NACL::C::VolumeQtree;
use NACL::C::VolumeFlexcache;
use NACL::C::Vfiler;
use NACL::C::VolumeFlexcacheDelegations;
use NACL::C::VolumeFile;
use NACL::C::VserverLocks;
use NACL::C::VolumeSnapshot;
use NACL::C::QosWorkload;
use NACL::C::VolumeFlexcacheOptions;
use NACL::CS::VolumeFlexcacheOptions;
use NACL::C::VolumeFlexcache;
use NACL::CS::VolumeFlexcache;
use NACL::C::VolumeFile;
use NACL::STask::VolumeEfficiency;
use NACL::STask::AntivirusOnDemandCommandScanDir;
use NACL::STask::AntivirusOnDemand;
use NACL::C::VolumeFileClone;
use NACL::C::VolumeQuotaPolicyRule;
use NACL::C::VolumeQuota;

=head1 CONSTRUCTOR

=head2 new

Constructor for the FlexCache::Volume class.
It creates or reuses a flexcache origin volume with the required attributes.
If the origin object is created with an associated windows client,
then a CISF share will also be created/reused for the volume. 
For C-mode, the volume is created with the following defaults
(all values are constants defined in FlexCache::Constants):
unix-permissions => FCACHE_DFLT_VOL_PERMS (for RW volumes),
user => FCACHE_DFLT_UNIX_USER,
group => FCACHE_DFLT_UNIX_GROUP,
policy => FCACHE_DFLT_VOL_POLICY,
security-style => FCACHE_DFLT_VOL_SEC_STYLE.

=head3 ARGUMENTS

=over

=item B<name> I<Optional> Scalar:

        Name of the volume. In case this is not defined,
        a name will be generated by the library.
        In case the constructor is called with the option
        to either attach or recycle, then this argument must be defined.

=item B<size> I<Optional> Scalar:

        Size of the volume.
        Default value: constant FCACHE_DFLT_VOL_SIZE (defined in FlexCache constants).

=item B<type> I<Optional> Scalar:

        Type of the volume. Ignored for 7-mode.
        Legal values: any of the FCACHE_VOL_TYPE_* (defined in FlexCache constants).
        Default value: constant FCACHE_VOL_TYPE_RW.

=item B<recycle> I<Optional> Boolean:

        Flag to determine if this volume should be reused.
        If this flag is set to TRUE, then:
        - at setup, the volume will be either created or reused (if it already exists)
        - at cleanup it will not be destroyed.
        If the flag is set to FALSE, then:
        - at setup the volume will be created or purged+recreated (if it already exists)
        - at cleanup it will be purged.
        Default value: constant FCACHE_DFLT_VOL_RECYCLE (defined in FlexCache constants).

=back

=head3 RETURN

FlexCache::Volume object.

=head3 EXAMPLE

	# Create a flexcache origin with the default settings for the volume member:
	my $origin = FlexCache::Origin->new(volume => {name => $VOL_SRC,},
	                                   ....);
	# Create a flexcache origin, requiring a size of 200m:
	my $origin = FlexCache::Origin->new(volume => {name => $VOL_SRC,
	                                              size => '200m',},
					   ....);

=cut

# "Somehow internal" arguments (to limit the use of this constructor to the FlexCache library.
#                               These arguments are not documented with POD, to not confuse the script developers):
# - node: Required: Object: NACL::STask::Node object for this node.
# - aggregate_name: Required: Scalar: Name of the aggregate that will contain the volume.
# - vserver_name: Required: Scalar: Name of the vserver that will export the volume.
# - client_type: Optional: Scalar: Type of client for which this volume should be ready-to-use.
#                Legal values:
#                  FCACHE_CLIENT_TYPE_UNIX (volume is created/reused
#                    and the client-side junction path of the object is mapped to the NFS junction-path),
#                  FCACHE_CLIENT_TYPE_WINDOWS (volume is created/reused, a cifs share is created/reused,
#                    and the client-side junction path of the object is mapped to the CIFS share name).
#                Default value: FCACHE_CLIENT_TYPE_UNIX.
# - debug: Optional: Boolean: Flag to increase the verbosity of this module, useful for debugging purposes.
#   Default value: FALSE.
# - perform_cleanup: Optional: Boolean: Flag to control if the cleanup should be executed.
#   Default value: TRUE.
# - ui_messages: Required: Object: FlexCache::UI::OntapMessages, to be used for any ONTAP output validation.
# - attach_only: Optional: Boolean: Flag to determine if the constructor should only attach to the required volume,
#                and not attempt to create/recreate. If this flag is set to TRUE, then the recycle flag is ignored.
#                Default value: FALSE.
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name => {type => SCALAR | UNDEF, default => FCACHE_VOL_ORIGIN,},
            node => {type => OBJECT,         isa     => "NACL::STask::Node"},
            aggregate_name => {type => SCALAR,},
            vserver_name   => {type => SCALAR,},
            type           => {type => SCALAR, default => FCACHE_VOL_TYPE_RW},
            size => {type => SCALAR, default => FCACHE_DFLT_VOL_SIZE,},
            client_type =>
                {type => SCALAR, default => FCACHE_CLIENT_TYPE_UNIX,},
            recycle => {type => BOOLEAN, default => FCACHE_DFLT_VOL_RECYCLE,},
            debug   => {type => BOOLEAN, default => FALSE,},
            perform_cleanup => {type => BOOLEAN, default => TRUE,},
            ui_messages =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
            attach_only => {type => BOOLEAN, default => FALSE,},
        }
    );

    my $self = {};

    $self->{_creator_pid}  = $$;
    $self->{_name}         = $args{name};
    $self->{_vserver_name} = $args{vserver_name};
    $self->{_nacl_node}    = $args{node};
    $self->{_boot_mode}    = $self->{_nacl_node}->mode();
    $self->{_client_type}  = $args{client_type};
    $self->{_attach_only}  = $args{attach_only};
    $self->{_recycle}      = ($self->{_attach_only}) ? TRUE : $args{recycle};
    $self->{_debug}        = $args{debug};
    $self->{_cleanup_flexcache_options} = {};
    $self->{_perform_cleanup}           = $args{perform_cleanup};
    $self->{_ui_messages}               = $args{ui_messages};
    $self->{_creator_pid}               = $$;
    $self->{_is_purged}                 = FALSE;

    if (   ($self->{_client_type} ne FCACHE_CLIENT_TYPE_UNIX)
        && ($self->{_client_type} ne FCACHE_CLIENT_TYPE_WINDOWS)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Unknown client-type-specific configuration required for:\n"
                . "Volume: "
                . $self->{_name} . "\n"
                . "Client type: "
                . $self->{_client_type}
        );
    }

    if (!defined($self->{_name})) {
        if (($self->{_recycle}) || ($self->{_attach_only})) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Volume name MUST be provided"
            );
        } else {
            $self->{_name}
                = fcache_get_random_name(type => _FCACHE_NAME_TYPE_VOL);
        }
    }

    bless($self, $class);

    fcache_log_comment(message => "Starting volume configuration for "
            . $self->{_name} . "\n"
            . "Attach: "
            . $self->{_attach_only} . "\n"
            . "Recycle: "
            . $self->{_recycle})
        if $self->{_debug};

    # Need to do this in a more complicated fashion because:
    # in case recycling is false but the volume is already there,
    # we must delete first all its caches to be able to purge it
    # (NACL does not do this for us)
    my $volume_exists = TRUE;

    # Leave the $share_reuse_action as 'reuse',
    # even in the case when $self->{_attach_only} is true.
    # This is because it may happen that the first origin that we ever create
    # is only configured for Unix. In this case, a second origin, attach-only,
    # will still need to have a way of creating the cifs share.
    my $share_reuse_action = 'reuse';
    my $aggr_name          = $args{aggregate_name};
    my $type               = $args{type};

    try {
        $self->{_nacl_vol} = NACL::STask::Volume->find(
            command_interface => $self->{_nacl_node},
            filter            => {
                volume    => $self->{_name},
                aggregate => $aggr_name,
                vserver   => $self->{_vserver_name},
                type      => $type,
            }
        );
    }

    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();

        # This is a NACL-hard-coded error message
        if ($output =~ /No matching volume.+ found/) {
            if ($self->{_attach_only}) {
                $exception->throw();
            }
            $volume_exists = FALSE;
        } else {
            $exception->throw();
        }
    };

    if (   ($self->{_boot_mode} eq FCACHE_ONTAP_CMODE)
        && (($volume_exists) && !($self->{_recycle}))) {
        fcache_log_comment(message => "Not recycling the volume. "
                . $self->{_name}
                . " Purging any associated caches")
            if $self->{_debug};
        $self->get_cache_volumes(_purge_if_found => TRUE);
        $self->{_nacl_vol}->purge();
        $volume_exists      = FALSE;
        $share_reuse_action = 'purge';
    }

    my %create_args = (
        command_interface => $self->{_nacl_node},
        volume            => $self->{_name},
        aggregate         => $aggr_name,
        vserver           => $self->{_vserver_name},
        size              => $args{size},
        type              => $type,
        user              => FCACHE_DFLT_UNIX_USER,
        group             => FCACHE_DFLT_UNIX_GROUP,
        policy            => FCACHE_DFLT_VOL_POLICY,
        'security-style'  => FCACHE_DFLT_VOL_SEC_STYLE,
    );

    if ($type eq FCACHE_VOL_TYPE_RW) {
        $self->{__junction_path}         = '/' . $self->{_name};
        $create_args{'junction-path'}    = $self->{__junction_path};
        $create_args{'unix-permissions'} = FCACHE_DFLT_VOL_PERMS;
    }

    if ($volume_exists == FALSE) {
        $self->{_nacl_vol} = NACL::STask::Volume->create(%create_args);
        $self->_exportfs_if_7mode();
    }

    if ($self->{_client_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        $self->_ensure_volume_has_share(reuse_action => $share_reuse_action,);
    }

    fcache_log_comment(
        message => "Completed creation of new Volume object for "
            . $self->{_name})
        if $self->{_debug};
    return $self;

} ## end sub new

=head1 OTHER METHODS

=head2 get_name

Method to obtain the name of this volume.

=head3 ARGUMENTS

None.

=head3 RETURN

Name of the volume (Scalar).

EXAMPLE

	my $vol_name = $obj->get_name();

=cut

sub get_name {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_name};
} ## end sub get_name

=head2 get_fullname

Method to obtain the full path of this volume.
For C-mode, this will be of the form: cluster_name://vserver_name/volume_name.
For 7-mode, this will be of the form: node_name:volume_name.

=head3 ARGUMENTS

None.

=head3 RETURN

Full path of the volume (Scalar).

EXAMPLE

	my $fullname = $obj->get_fullname();

=cut

sub get_fullname {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_nacl_node}->mode() eq FCACHE_ONTAP_7MODE) {
        my $node_name = $class->{_nacl_node}->node();
        my $vol_name  = $class->get_name();
        my $fullname  = $node_name . ':' . $vol_name;

        return $fullname;
    } else {
        return $class->{_nacl_vol}->fullname();
    }
} ## end sub get_fullname

=head2 get_vserver_name

Method to obtain the vserver of this volume.

=head3 ARGUMENTS

None.

=head3 RETURN

Vserver name (Scalar).

EXAMPLE

	my $vserver_name = $obj->get_vserver_name();

=cut

sub get_vserver_name {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{_nacl_vol}->vserver();
} ## end sub get_vserver_name

=head2 get_junction_path

Obtain the junction path of this volume.
For an origin volume, this is the junction-path on which the volume is mounted on the filer.
For a C-mode cache volume, this is the junction-path on which its origin is mounted on the filer.

=head3 ARGUMENTS

None.

=head3 RETURN

The junction path of the volume (Scalar).

EXAMPLE

	my $junction_path = $obj->get_junction_path();

=cut

sub get_junction_path {
    my $class = shift;
    my %args = validate(@_, {});

    my $vol_state = $class->{_nacl_vol}->state();
    return $vol_state->junction_path();
} ## end sub get_junction_path

=head2 get_client_mount_junction

Obtain the junction name on which this volume can be mounted on a client.
For NFS, this is the junction-path.
For CIFS, this will be the share name that was automatically created for this volume at setup.

=head3 ARGUMENTS

None.

=head3 RETURN

The junction name/path for the volume (Scalar).

EXAMPLE

	my $mount_junction = $obj->get_client_mount_junction();

=cut

sub get_client_mount_junction {
    my $class = shift;
    my %args = validate(@_, {});

    my $junction;
    if ($class->{_client_type} eq FCACHE_CLIENT_TYPE_WINDOWS) {
        $junction = $class->{__share_name};
    } else {
        $junction = $class->get_junction_path();
    }

    if (!defined($junction)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Mount junction is undefined for volume "
                . $class->{_name}
        );
    }

    return $junction;
} ## end sub get_client_mount_junction

=head2 get_uuid

Get the UUID of this volume.
This method is currently only supported for C-mode.

=head3 ARGUMENTS

None.

=head3 RETURN

The UUID of the volume (Scalar).

EXAMPLE

	my $uuid = $obj->get_uuid();

=cut

sub get_uuid {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "This method is only available for C-mode",
        );
    }

    my $vol_state = $class->{_nacl_vol}->state();
    my $uuid      = $vol_state->uuid();

    if (!defined($uuid)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Could not determine volume uuid:\n"
                . "Volume: "
                . $class->{_name} . "\n"
                . "Volume state:\n"
                . Dumper($vol_state)
        );
    }

    return $uuid;
} ## end sub get_uuid

=head2 get_type

Get the type of this volume (RW, DP, DC).
This method is currently only supported for C-mode.

=head3 ARGUMENTS

None.

=head3 RETURN

The type of the volume (Scalar). It will be one of the constants
FCACHE_VOL_TYPE_*, as defined in FlexCache::Constants.

EXAMPLE

	my $type = $obj->get_type();

=cut

sub get_type {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is only available for C-mode",
        );
    }

    my $vol_state = $class->{_nacl_vol}->state();
    my $type      = $vol_state->type();
    if (!defined($type)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Could not determine volume type:\n"
                . "Volume: "
                . $class->{_name} . "\n"
                . "Volume state:\n"
                . Dumper($vol_state)
        );
    }

    return $type;
} ## end sub get_type

=head2 get_max_file_count

Get the maximum number of files allowed on this volume.
This method is only supported for C-mode for now.

=head3 ARGUMENTS

None.

=head3 RETURN

The maximum number of files allowed on this volume (Scalar).

EXAMPLE

	my $max_files = $obj->get_max_file_count();

=cut

sub get_max_file_count {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is only available for C-mode",
        );
    }

    my $vol_state = $class->{_nacl_vol}->state();
    my $max_files = $vol_state->files();
    if (!defined($max_files)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Could not determine max-files:\n"
                . "Volume: "
                . $class->{_name} . "\n"
                . "Volume state:\n"
                . Dumper($vol_state)
        );
    }

    return $max_files;
} ## end sub get_max_file_count

=head2 cifs_share_create

Create a CIFS share for this volume.

=head3 ARGUMENTS

=over

=item B<name> I<Optional> Scalar:

        Name of the share to be created.
        Default value: name of the volume.

=item B<properties> I<Optional> Arrayref:

        Properties to be set for this share.
        Individual properties that can be set are defined as constants in FlexCache::Constants.
        By default, the following properties are set:
        FCACHE_CIFS_SHARE_PROP_OPLOCKS, FCACHE_CIFS_SHARE_PROP_BROWSABLE,
        FCACHE_CIFS_SHARE_PROP_CHANGENOTIFY, FCACHE_CIFS_SHARE_PROP_SHOWSNAPSHOT.

=back

=head3 RETURN

The name of the share (Scalar).

EXAMPLE

	$obj->cifs_share_create(name => 'my_share',
	                        properties => [FCACHE_CIFS_SHARE_PROP_OPLOCKS,
				               FCACHE_CIFS_SHARE_PROP_BROWSABLE,],);

=cut

# Internal arguments:
# - _reuse_action: Optional: Scalar: Overwrite the documented behavior in case of an already-existent entry.
#                  Legal values: die (die if an entry already exists),
#                                reuse (reuse an already-existent entry, otherwise, create a new one)
#                                attach (expect that an entry already exists).
#                  Default value: FALSE.
sub cifs_share_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name       => {type => SCALAR, default => $class->{_name},},
            properties => {
                type    => ARRAYREF,
                default => [
                    FCACHE_CIFS_SHARE_PROP_OPLOCKS,
                    FCACHE_CIFS_SHARE_PROP_BROWSABLE,
                    FCACHE_CIFS_SHARE_PROP_CHANGENOTIFY,
                    FCACHE_CIFS_SHARE_PROP_SHOWSNAPSHOT,
                ],
            },
            _reuse_action => {type => SCALAR, default => 'die',},
        }
    );

    my $share_name       = $args{name};
    my $share_properties = $args{properties};
    my $reuse_action     = $args{_reuse_action};
    my $share;

    if ($reuse_action eq 'attach') {
        $share = NACL::STask::VserverCifsShare->find(
            command_interface => $class->{_nacl_node},
            filter            => {
                volume       => $class->{_name},
                vserver      => $class->{_vserver_name},
                'share-name' => $share_name,
            }
        );
    } else {
        $share = NACL::STask::VserverCifsShare->create(
            command_interface  => $class->{_nacl_node},
            volume             => $class->{_name},
            vserver            => $class->{_vserver_name},
            'share-name'       => $share_name,
            nacltask_if_exists => $reuse_action,
        );
    }

    return $share_name;
} ## end sub cifs_share_create

=head2 cifs_share_delete

Delete a CIFS share from this volume.

=head3 ARGUMENTS

=over

=item B<name> I<Optional> Scalar:

        Name of the share to be deleted.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->cifs_share_delete(name => 'my_share',);

=cut

sub cifs_share_delete {
    my $class = shift;
    my %args = validate(@_, {name => {type => SCALAR,},});

    NACL::STask::VserverCifsShare->delete(
        command_interface => $class->{_nacl_node},
        vserver           => $class->{_vserver_name},
        'share-name'      => $args{name},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub cifs_share_delete

=head2 get_cache_volumes

Detect the caches associated to this volume.
This method is only available for C-mode.
Called in 7-mode, it will terminate the script execution with N/A.

=head3 ARGUMENTS

=over

=item B<expected> I<Optional> Scalar:

        Expected return code.
        Legal values:
        FCACHE_RC_SUCCESS: at least one cache is expected to be found for this volume.
        FCACHE_RC_NO_ENTRY_FOUND: no cache is expected to be found for this volume.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

        Hash with key <-> values:
        cache_name <-> hash with : vserver (name of the vserver for this cache),
                                   aggregate (name of the aggregate containing the cache)
				   node (name of the node hosting the cache).

EXAMPLE

        my %caches = $obj->get_cache_volumes();
	my %caches = $obj->get_cache_volumes(expected => FCACHE_RC_NO_ENTRY_FOUND);

=cut

# Internal arguments:
# - _purge_if_found: Optional: Boolean: If this flag is true, then any detected cache will be deleted.
#                    In this case, expected is ignored and no failure is caused in either of the cases
#                    (volume has or does not have caches).
#                    Default value: FALSE.
sub get_cache_volumes {
    my $class = shift;
    my %args  = validate(
        @_,
        {   expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            _purge_if_found => {type => SCALAR, default => FALSE,},
        }
    );

    my $purge = $args{_purge_if_found};
    my $expected_rc
        = ($purge) ? FCACHE_RC_FAILURE_NON_FATAL : $args{expected};
    my $obtained_rc = FCACHE_RC_NO_ENTRY_FOUND;
    my @caches      = ();
    my %return_hash = ();
    my $output      = '';
    my $vol_name    = $class->{_name};

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "This method is only available for C-mode",
        );
    }

    try {
        @caches = NACL::CS::VolumeFlexcache->fetch(
            command_interface => $class->{_nacl_node},
            requested_fields  => [
                'origin-volume', 'vserver',
                'cache-volume',  'cache-aggregate',
                'cache-node'
            ],
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();

      # Hard-coded NACL message, no need to process through UI at this moment.
        if ($output =~ /No matching .+ found/) {
            $obtained_rc = FCACHE_RC_NO_ENTRY_FOUND;

        } else {
            $obtained_rc = FCACHE_RC_FAILURE;
        }
        @caches = ();
    };

    foreach my $cache (@caches) {
        my $origin_vol = $cache->origin_volume();
        if ($origin_vol ne $vol_name) {
            fcache_log_comment(
                message => "Skipping cache volume with origin $origin_vol")
                if ($class->{_debug});
            next;
        }
        $obtained_rc = FCACHE_RC_SUCCESS;
        my $cache_name = $cache->cache_volume();
        fcache_log_comment(message => "Found cache: $cache_name")
            if ($class->{_debug});
        $return_hash{$cache_name}{vserver}   = $cache->vserver();
        $return_hash{$cache_name}{aggregate} = $cache->cache_aggregate();
        $return_hash{$cache_name}{node}      = $cache->cache_node();
        if ($purge) {
            fcache_log_comment(message => "Purging cache: $cache_name")
                if ($class->{_debug});
            NACL::STask::Volume->purge(
                command_interface => $class->{_nacl_node},
                volume            => $cache_name,
                vserver           => $class->{_vserver_name}
            );
        }
    }

    fcache_verify_expected_rc(
        obtained_rc => $obtained_rc,
        expected_rc => $expected_rc,
        action      => "Detect cache volumes having $vol_name as origin",
        output      => $output . Dumper(\%return_hash),
    );

    return %return_hash;
} ## end sub get_cache_volumes

=head2 get_vldb_info

Get the vldb (MSID and DSID) info for this volume.
This method is only applicable for C-mode.
Called in 7-mode, it will terminate the script execution with N/A.

=head3 ARGUMENTS

=over

=item B<cache_vldb_info> I<Optional> Hashref:

        Vldb info of a cache of this volume
        If passed in, then the origin's vldb information is verified against the cache's
        (MSIDs should be identical, DSIDs should not).
        If the verification fails, then the script execution is terminated.

=back

=head3 RETURN

Hash with: msid, dsid, vol_name.

EXAMPLE

	my %cache_vldb = $cache->{volume}->get_vldb_info();
        $origin->{volume}->get_vldb_info(cache_vldb_info => \%cache_vldb);

=cut

sub get_vldb_info {
    my $class = shift;
    my %args  = validate(@_,
        {cache_vldb_info => {type => HASHREF, optional => 1,},});

    my %return_hash = ();

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message     => "Vldb information is only available for C-mode",
        );
    }

    my $vol_state = $class->{_nacl_vol}->state();
    my $msid      = $vol_state->msid();
    my $dsid      = $vol_state->dsid();
    my $vol_name  = $class->{_name};

    if (!defined($msid) || !defined($dsid)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Failed to obtain the vldb information for volume "
                . $class->{_name}
        );
    }

    if (defined($args{cache_vldb_info})) {
        my $cache_vldb_info = $args{cache_vldb_info};
        my $cache_msid      = $cache_vldb_info->{msid};
        my $cache_dsid      = $cache_vldb_info->{dsid};
        my $cache_name      = $cache_vldb_info->{vol_name};

        if (   !defined($cache_msid)
            || !defined($cache_dsid)
            || !defined($cache_name)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_SCRIPT,
                message     => "Invalid cache_vldb_info format:\n"
                    . Dumper($cache_vldb_info)
            );
        }

        if ($msid != $cache_msid) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Vldb check failure for:\n"
                    . "Origin volume: $vol_name\n"
                    . "Cache volume: $cache_name\n"
                    . "MSIDs should be identical, but they are not:\n"
                    . "Origin MSID: $msid\n"
                    . "Cache MSID: $cache_msid"
            );
        }

        if ($dsid == $cache_dsid) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Vldb check failure for:\n"
                    . "Origin volume: $vol_name\n"
                    . "Cache volume: $cache_name\n"
                    . "DSIDs should be not identical, but they are:\n"
                    . "Origin DSID: $dsid\n"
                    . "Cache DSID: $cache_dsid"
            );
        }

        fcache_log_comment(
                  message => "Vldb check completed successfully for:\n"
                . "Origin volume: $vol_name\n"
                . "Cache volume: $cache_name\n"
                . "Origin: MSID = $msid, DSID = $dsid\n"
                . "Cache: MSID = $cache_msid, DSID = $cache_dsid");
    }

    $return_hash{msid}     = $msid;
    $return_hash{dsid}     = $dsid;
    $return_hash{vol_name} = $vol_name;

    return %return_hash;
} ## end sub get_vldb_info

=head2 offline

Offline the volume.

=head3 ARGUMENTS

=over

=item B<address_dependencies> I<Optional> Boolean:

        Flag to control if any dependencies should be addressed by the library.
        If this flag is set to FALSE, then no dependency will be addressed.
        If the flag is set to TRUE, then dependencies will be addressed in a
        best-effort attempt (any dependency failure will be ignored).
        Dependencies:
        - 7-mode: none.
        - C-mode: unmount.
        Default value: TRUE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->offline();
        $obj->offline(address_dependencies => FALSE);

=cut

sub offline {
    my $class = shift;
    my %args  = validate(@_,
        {address_dependencies => {type => BOOLEAN, default => TRUE,},});

    if (   ($args{address_dependencies})
        && ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE)) {
        try {
            $class->{_nacl_vol}->unmount();
        }
        catch NATE::BaseException with {
            my $exception = shift;
            my $output    = $exception->text();
            fcache_log_comment(
                message => "Volume unmount failed:\n" . $output);
        };
    }

    $class->{_nacl_vol}->offline();

    return FCACHE_RC_SUCCESS;
} ## end sub offline

=head2 online

Online the volume.

=head3 ARGUMENTS

=over

=item B<restore_dependencies> I<Optional> Boolean:

        Flag to control if any dependencies should be restored by the library 
        (dependencies that needed to be addressed when offlining).
        If this flag is set to FALSE, then no dependency will be restored.
        If the flag is set to TRUE, then dependencies will be restored.
        Dependencies:
        - 7-mode: none.
        - C-mode: re-mount.
        Default value: TRUE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->online();
        $obj->online(restore_dependencies => FALSE);

=cut

sub online {
    my $class = shift;
    my %args  = validate(@_,
        {restore_dependencies => {type => BOOLEAN, default => TRUE,},});

    $class->{_nacl_vol}->online();

    if (   ($args{restore_dependencies})
        && ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE)
        && (defined($class->{__junction_path}))) {
        $class->{_nacl_vol}
            ->mount('junction-path' => $class->{__junction_path});
    }

    return FCACHE_RC_SUCCESS;
} ## end sub online

=head2 restrict

Restrict the volume.

=head3 ARGUMENTS

=over

=item B<address_dependencies> I<Optional> Boolean:

        Flag to control if any dependencies should be addressed by the library.
        If this flag is set to FALSE, then no dependency will be addressed.
        If the flag is set to TRUE, then dependencies will be addressed in a
        best-effort attempt (any dependency failure will be ignored).
        Dependencies:
        - 7-mode: none.
        - C-mode: unmount.
        Default value: TRUE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->restrict();

=cut

sub restrict {
    my $class = shift;
    my %args  = validate(@_,
        {address_dependencies => {type => BOOLEAN, default => TRUE,},});

    if (   ($args{address_dependencies})
        && ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE)) {
        try {
            $class->{_nacl_vol}->unmount();
        }
        catch NATE::BaseException with {
            my $exception = shift;
            my $output    = $exception->text();
            fcache_log_comment(
                message => "Volume unmount failed:\n" . $output);
        };
    }
    
    $class->{_nacl_vol}->restrict();

    return FCACHE_RC_SUCCESS;
}

=head2 mount

Mount the volume (on the filer).
For 7-mode, this method is a no-op.

=head3 ARGUMENTS

None

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->mount();

=cut

sub mount {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_comment(message => "Skipping filer mount for 7-mode");
        return FCACHE_RC_SUCCESS;
    }

    $class->{_nacl_vol}->mount();
    $class->{__junction_path} = $class->get_junction_path();

    return FCACHE_RC_SUCCESS;
} ## end sub mount

=head2 rename

Rename the volume.

=head3 ARGUMENTS

=over

=item B<new_name> I<Optional> Scalar:

        New name for the volume.
        If this argument is not passed in, then the new name of the volume
        will be <current_name>_renamed

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->rename();
        $obj->rename(new_name => 'my_new_vol');

=cut

sub rename {
    my $class = shift;
    my %args = validate(@_, {new_name => {type => SCALAR, optional => 1,},});

    my $new_name = $args{new_name};
    if (!defined($new_name)) {
        $new_name = $class->{_name} . '_renamed';
    }

    $class->{_nacl_vol}->rename(newname => $new_name);
    $class->{_name} = $new_name;

    return FCACHE_RC_SUCCESS;
} ## end sub rename

=head2 get_size

Get the size of this volume.

=head3 ARGUMENTS

None.

=head3 RETURN

Volume size (Scalar).

EXAMPLE

        my $size = $origin->{volume}->get_size();

=cut

sub get_size {
    my $class = shift;
    my %args = validate(@_, {});

    my $vol_state = $class->{_nacl_vol}->state();
    my $size      = $vol_state->size();
    if (!defined($size)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Failed to get volume size:\n" . Dumper($vol_state)
        );
    }

    return $size;
} ## end sub get_size

=head2 set_size

Change the size of the volume.

=head3 ARGUMENTS

=over

=item B<new_size> I<Required> Scalar:

        New size for the volume.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->set_size(new_size => '10g');

=cut

sub set_size {
    my $class = shift;
    my %args = validate(@_, {new_size => {type => SCALAR,},});

    $class->{_nacl_vol}->modify(size => $args{new_size});

    return FCACHE_RC_SUCCESS;
} ## end sub set_size

=head2 set_flexcache_policy

Set a flexcache cache policy for the volume.
This method is only applicable for C-mode.

=head3 ARGUMENTS

=over

=item B<new_policy> I<Required> Scalar:

        New policy name to set for the volume.

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->set_flexcache_policy(new_policy => 'default');

=cut

sub set_flexcache_policy {
    my $class = shift;
    my %args  = validate(
        @_,
        {   new_policy => {type => SCALAR,},
            expected   => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output = '';

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message =>
                "Volume flexcache policy support is not available for 7-mode"
        );
    }

    try {
        $class->{_nacl_vol}
            ->modify('flexcache-cache-policy' => $args{new_policy});
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_FLEXCACHE_POLICY_SET,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub set_flexcache_policy

=head2 flexcache_policy_verify

Verify the flexcache cache policy for the volume.
This method is only applicable for C-mode.

=head3 ARGUMENTS

=over

=item B<policy> I<Required> Scalar:

        Expected flexcache cache policy for the volume.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->flexcache_policy_verify(policy => 'default');

=cut

sub flexcache_policy_verify {
    my $class = shift;
    my %args = validate(@_, {policy => {type => SCALAR,},});

    my $expected_policy = $args{policy};

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_NA,
            message =>
                "Volume flexcache policy support is not available for 7-mode"
        );
    }

    my $vol_state      = $class->{_nacl_vol}->state();
    my $current_policy = $vol_state->flexcache_cache_policy();

    if ($current_policy ne $expected_policy) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Flexcache cache policy mismatch:\n"
                . "Volume: "
                . $class->{_name} . "\n"
                . "Expected policy: $expected_policy\n"
                . "Obtained policy: $current_policy"
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_policy_verify

=head2 flexcache_fstat_verify

Verify the flexcache fstat information for this volume.

=head3 ARGUMENTS

=over

=item B<file_name> I<Optional> Scalar:

        To check the fstat of a specific file in the volume:
	file name or path, as it appears within the volume.

=item B<inode_file> I<Optional> Boolean:

        Flag to control if the flexcache stats should be obtained for the inode file.
        Default value: FALSE.
        
=item B<expected_values> I<Optional> Hashref:

        Hashref of pairs "field name" => "expected value",
        to be passed in for the specific fields that need to be verified.
        Unless this argument is passed in, the method does not execute any value
        check against the fstat fields, but it just test that
        the fstats can be obtained for the specifc file.
        
=back

=head3 RETURN

Hash with parsed flexcache fstat fields and the obtained values.

EXAMPLE

        $obj->flexcache_fstat_verify(file_name => $file1,
                                     expected_values => {local_blocks => 1},);

=cut

sub flexcache_fstat_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name       => {type => SCALAR,  optional => 1,},
            inode_file      => {type => BOOLEAN, default  => FALSE,},
            expected_values => {type => HASHREF, optional => 1,},
        }
    );

    my $path = '/vol/' . $class->{_name};
    if (defined($args{file_name})) {
        $path .= '/' . $args{file_name};
    }
    my %return_hash = ();

    # TODO: revisit this once burt 608750 is fixed
    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
    my $command = "flexcache fstat $path";
    if ($args{inode_file}) {
        $command .= ' -inode-file';
    }
    my $output = $api_set->execute_command(command => $command);

    # This should not be in here,
    # they are just a work-around until NACL support is added
    # (we may need to re-map some keys for NACL compatibility)
    # This is the "parsing info" for the fields that we are interested in:
    #  key (name of the field, as it will be returned by this method) =>
    #  value (identification string, as it appears on the filer)
    my %fields_of_interest = (
        blocks          => 'Blocks: ',
        indirect_blocks => 'Indirect Blocks: ',
        local_blocks    => 'FlexCache Local Blocks: ',
        local_flags     => 'FlexCache Local Flags: ',
    );

    foreach my $field (keys(%fields_of_interest)) {
        my $identifier = $fields_of_interest{$field};
        if ($output =~ /$identifier(\S*)/) {
            $return_hash{$field} = $1;
        } elsif ($output =~ /$identifier/) {

            # If we got here, it means that the field is there in the output,
            # but there is no value
            $return_hash{$field} = '';
        }
    }

    fcache_log_comment(message => "Flexcache fstat:\n"
            . "Path: $path\n"
            . "Parsed fields:\n"
            . Dumper(\%return_hash))
        if ($class->{_debug});

    if (defined($args{expected_values})) {
        my $expected_values = $args{expected_values};
        foreach my $field_to_check (keys(%{$expected_values})) {
            my $expected_value = $expected_values->{$field_to_check};
            my $obtained_value = $return_hash{$field_to_check};
            if (!defined($obtained_value)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message     => "Flexcache fstat:\n"
                        . "Path: $path\n"
                        . "Failed to detect field $field_to_check\n"
                        . "Fields detected:\n"
                        . Dumper(\%return_hash)
                );
            }

            if ("$expected_value" ne "$obtained_value") {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message     => "Flexcache fstat mismatch:\n"
                        . "Path: $path\n"
                        . "Field: $field_to_check\n"
                        . "Expected value: $expected_value\n"
                        . "Obtained value: $obtained_value"
                );
            }

            fcache_log_comment(message => "Verified flexcache fstat:\n"
                    . "Path: $path\n"
                    . "Field: $field_to_check\n"
                    . "Expected value: $expected_value\n"
                    . "Obtained value: $obtained_value");
        }
    }

    return %return_hash;
} ## end sub flexcache_fstat_verify

=head2 flexcache_delegation_verify

Verify the type and state of a flexcache delegation for a specific file.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file (as seen in the volume) for which to verify the delegation.

=item B<file_id> I<Optional> Scalar:

        File id, to be used when looking for a delegation. This argument is
        useful when looking for revoked delegations afler a file was deleted.

=item B<delegation_mode> I<Optional> Scalar:

        Delegation mode (read or write).
        This argument needs to be passed in if the caller wants to verify
        the delegation mode. Otherwise, no mode check is performed.
        Legal values: any of the FCACHE_VOL_DELEGATION_MODE_* constants
        (defined in FlexCache::Constants).
	This argument is not supported for 7-mode (ignored).

=item B<delegation_state> I<Optional> Scalar:

        State of the delegation.
        This argument needs to be passed in if the caller wants to verify
        the delegation state. Otherwise, no state check is performed.
        Legal values: any of the FCACHE_VOL_DELEGATION_STATE_* constants
        (defined in FlexCache::Constants).
	This argument is not supported for 7-mode (ignored).

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

File id for this file (can be used for future calls to flexcache_delegation_verify,
especially for deleted files).

EXAMPLE

        # Verify that there are no delegations for my file
	$obj->flexcache_delegation_verify(file_name => 'my_file',
                                          expected => FCACHE_RC_NO_ENTRY_FOUND);
        # Verify that there are read delegations granted to my file
        $obj->flexcache_delegation_verify(file_name => 'my_file',
                                          delegation_mode => FCACHE_VOL_DELEGATION_MODE_READ,
                                          delegation_state => FCACHE_VOL_DELEGATION_STATE_GRANTED);

=cut

sub flexcache_delegation_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name => {type => SCALAR,},
            file_id   => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            delegation_mode  => {type => SCALAR, optional => 1,},
            delegation_state => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        return $class->_7Mode_flexcache_delegation_verify(%args);
    } else {
        return $class->_CMode_flexcache_delegation_verify(%args);
    }
} ## end sub flexcache_delegation_verify

=head2 locks_verify

Verify the vserver locks for this volume.
This method is not yet supported in 7-mode.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file (as seen in the volume) for which to verify the lock.

=item B<type> I<Optional> Scalar:

        Type of lock to look for.
	Legal values: any of the FCACHE_VOL_OPLOCK_TYPE_* constants,
	as defined in FlexCache::Constants.

=item B<oplock_level> I<Optional> Scalar:

        Required oplock-type for the lock.
	Legal values: any of the FCACHE_VOL_OPLOCK_LEVEL_* constants,
	as defined in FlexCache::Constants.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=item B<cached_by> I<Optional> Scalar: 

	Name of the cache for which to check that the lock is cached.
	This argument is ignored if expected is other than FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->locks_verify(file_name => $file1);
	$obj->locks_verify(file_name => $file1,
	                   type => FCACHE_VOL_OPLOCK_TYPE_BYTE_RANGE);
        $obj->locks_verify(file_name => $file1,
                           expected => FCACHE_RC_NO_ENTRY_FOUND);
	$obj->locks_verify(file_name => $file1,
	                   cached_by => $cache_name);

=cut

sub locks_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name    => {type => SCALAR,},
            oplock_level => {type => SCALAR, optional => 1,},
            type         => {type => SCALAR, optional => 1,},
            expected     => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            cached_by    => {type => SCALAR, optional => 1,},
        }
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Applicability for 7-mode is not yet investigated for this method"
        );
    }

    my $file_name = $args{file_name};
    my $expected  = $args{expected};
    my $vol_name  = $class->{_name};
    my $output    = '';
    my $lock;
    my $filter = {
        volume  => $class->{_name},
        vserver => $class->{_vserver_name},
        path    => "/$vol_name/$file_name"
    };

    if (defined($args{type})) {
        $filter->{'type'} = $class->{_ui_messages}
            ->get_constant_value(constant => $args{type});
    }

    if (defined($args{oplock_level})) {
        $filter->{'oplock-level'} = $class->{_ui_messages}
            ->get_constant_value(constant => $args{oplock_level});
    }

    try {
        $lock = NACL::C::VserverLocks->find(
            command_interface => $class->{_nacl_node},
            filter            => $filter,
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VSERVER_LOCKS_FIND,
        output      => $output,
        expected_rc => $expected
    );

    if (   ($expected == FCACHE_RC_SUCCESS)
        && (defined($args{cached_by}))) {

        my $cache_name = $args{cached_by};

        # Obtain all the pertinent information from the origin lock,
        # to be used when checking for the cached lock
        my $lock_state = $lock->state();

        my $cache_filter = {volume => $cache_name};
        my @fields = qw(client_address lif vserver sharelock_mode);
        foreach my $field (@fields) {
            my $value = $lock_state->$field();
            $field =~ s/_/-/g;
            $cache_filter->{$field} = $value;
        }

        fcache_log_comment(message => "Looking for cached lock for:\n"
                . "Origin volume: "
                . $class->{_name} . "\n"
                . "Cache volume: $cache_name\n"
                . "Extra attributes:\n"
                . Dumper($cache_filter));

        NACL::C::VserverLocks->find(
            command_interface => $class->{_nacl_node},
            filter            => $cache_filter,
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub locks_verify

=head2 locks_break

Break the vserver locks for this volume.
This method is not yet supported in 7-mode.

=head3 ARGUMENTS

=over

=item B<file_name> I<Optional> Scalar:

        Name of the file (as seen in the volume) for which to break the lock.
	If this is not specified, then the locks for all the paths will be broken.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->locks_break();
        $obj->locks_break(file_name => $file1);

=cut

sub locks_break {
    my $class = shift;
    my %args = validate(@_, {file_name => {type => SCALAR, optional => 1,},});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message =>
                "Applicability for 7-mode is not yet investigated for this method"
        );
    }

    my $vol_name = $class->{_name};
    my $path
        = (defined($args{file_name}))
        ? ("/$vol_name/" . $args{file_name})
        : FCACHE_ENTITY_ANY;

    NACL::C::VserverLocks->break(
        command_interface => $class->{_nacl_node},
        volume            => $vol_name,
        vserver           => $class->{_vserver_name},
        path              => $path
    );

    return FCACHE_RC_SUCCESS;
} ## end sub locks_break

=head2 snapshot_create

Take a snapshot of this volume.

=head3 ARGUMENTS

=over

=item B<snapshot_name> I<Required> Scalar:

        Name of the snapshot.

=item B<snapmirror_label> I<Optional> Scalar:

        Snapmirror label to associate to this snapshot.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->snapshot_create(snapshot_name => 'snap1');
        $obj->snapshot_create(snapshot_name => 'snap1',
                              expected => FCACHE_RC_ALLOWED_ONLY_ON_RW_VOL);

=cut

sub snapshot_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   snapshot_name    => {type => SCALAR,},
            snapmirror_label => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output      = '';
    my %create_args = (
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        snapshot          => $args{snapshot_name},
    );
    if (defined($args{snapmirror_label})) {
        $create_args{'snapmirror-label'} = $args{snapmirror_label};
    }

    try {
        NACL::STask::VolumeSnapshot->create(%create_args);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_SNAPSHOT_CREATE,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub snapshot_create

=head2 snapshot_delete

Delete a snapshot of this volume.

=head3 ARGUMENTS

=over

=item B<snapshot_name> I<Required> Scalar:

        Name of the snapshot.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->snapshot_delete(snapshot_name => 'snap1');
        $obj->snapshot_delete(snapshot_name => 'snap1',
                              expected => FCACHE_RC_ALLOWED_ONLY_ON_RW_VOL);

=cut

sub snapshot_delete {
    my $class = shift;
    my %args  = validate(
        @_,
        {   snapshot_name => {type => SCALAR,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output = '';

    try {
        NACL::STask::VolumeSnapshot->purge(
            command_interface => $class->{_nacl_node},
            volume            => $class->{_name},
            vserver           => $class->{_vserver_name},
            snapshot          => $args{snapshot_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_SNAPSHOT_DELETE,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub snapshot_delete

=head2 snapshot_restore

Perform a snapshot restore of this volume.

=head3 ARGUMENTS

=over

=item B<snapshot_name> I<Required> Scalar:

        Name of the snapshot.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->snapshot_restore(snapshot_name => 'snap1');
        $obj->snapshot_restore(snapshot_name => 'snap1',
	                       expected => FCACHE_RC_ALLOWED_ONLY_ON_RW_VOL);

=cut

sub snapshot_restore {
    my $class = shift;
    my %args  = validate(
        @_,
        {   snapshot_name => {type => SCALAR,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output = '';

    try {
        NACL::C::VolumeSnapshot->restore(
            command_interface => $class->{_nacl_node},
            volume            => $class->{_name},
            vserver           => $class->{_vserver_name},
            snapshot          => $args{snapshot_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_SNAPSHOT_RESTORE,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub snapshot_restore

=head2 snapshot_list

List the snapshots of this volume.

=head3 ARGUMENTS

=over

=item B<expected_snapshots> I<Optional> Arrayref:

        Arrayref of expected snapshot names. If this argument is passed in,
	the library will also check if all the elements of expected_snapshots
	are present in the list of snapshots for this volume.

=back

=head3 RETURN

Array with names of all detected snapshots for this volume.

EXAMPLE

        my @snap_list = $obj->snapshot_list();
	$obj->snapshot_list(expected_snapshots => ['snap1', 'snap2']);

=cut

sub snapshot_list {
    my $class = shift;
    my %args  = validate(@_,
        {expected_snapshots => {type => ARRAYREF, optional => 1},});

    my @snap_list = ();

    try {
        my @snapshots = NACL::C::VolumeSnapshot->find(
            command_interface => $class->{_nacl_node},
            filter            => {
                volume  => $class->{_name},
                vserver => $class->{_vserver_name},
            }
        );
        foreach my $snap (@snapshots) {
            my $snap_name = $snap->snapshot();
            push(@snap_list, $snap_name);
        }
    }
    catch NATE::BaseException with {
        my $exception = shift;
        my $output    = $exception->text();
        if ($output !~ /No matching volume snapshot.* found/) {
            $exception->throw();
        }
    };

    if (defined($args{expected_snapshots})) {
        my $expected_snapshots = $args{expected_snapshots};
        foreach my $expected_snap (@{$expected_snapshots}) {
            if (!grep(/^$expected_snap/, @snap_list)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message => "Failed to detect expected volume snapshot:\n"
                        . "Volume: "
                        . $class->{_name} . "\n"
                        . "Expected snapshot: $expected_snap\n"
                        . "Detected snapshots:\n"
                        . Dumper(@snap_list)
                );
            } else {
                fcache_log_comment(message =>
                          "Successfully detected expected volume snapshot:\n"
                        . "Volume: "
                        . $class->{_name} . "\n"
                        . "Snapshot: $expected_snap");
            }
        }
    }

    return @snap_list;
} ## end sub snapshot_list

=head2 snapshot_schedule_disable

Disable snapshot schedule for this volume.

=head3 ARGUMENTS

None

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->snapshot_schedule_disable();

=cut

sub snapshot_schedule_disable {
    my $class = shift;
    my %args = validate(@_, {});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is not yet supported for 7-mode"
        );
    }

    $class->{_nacl_vol}->modify('snapshot-policy' => 'none');

    return FCACHE_RC_SUCCESS;
} ## end sub snapshot_schedule_disable

=head2 snapshot_restore

Perform a snapshot restore of a file of this volume.

=head3 ARGUMENTS

=over

=item B<snapshot_name> I<Required> Scalar:

        Name of the snapshot.

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->snapshot_restore_file(
            file_name     => $file1,
            snapshot_name => $snapshot,
        );

=cut

sub snapshot_restore_file {
    my $class = shift;
    my %args  = validate(
        @_,
        {   snapshot_name => {type => SCALAR,},
            file_name     => {type => SCALAR,},
            expected      => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output   = '';
    my $vol_name = $class->{_name};

    try {
        NACL::C::VolumeSnapshot->restore_file(
            command_interface => $class->{_nacl_node},
            volume            => $class->{_name},
            vserver           => $class->{_vserver_name},
            snapshot          => $args{snapshot_name},
            path              => "/vol/$vol_name/" . $args{file_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_SNAPSHOT_RESTORE_FILE,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub snapshot_restore_file

=head2 set_antivirus_policy

Set the antivirus policy for this volume.

=head3 ARGUMENTS

=over

=item B<policy_name> I<Required> Scalar:

        Policy name to set.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->set_antivirus_policy(policy_name => 'default');

=cut

sub set_antivirus_policy {
    my $class = shift;
    my %args = validate(@_, {policy_name => {type => SCALAR,},});

    if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
        $class->{_nacl_vol}
            ->modify('antivirus-on-access-policy' => $args{policy_name},);
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is not yet supported for 7-mode"
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub set_antivirus_policy

=head2 set_oplocks_mode

Enable/disable oplock-mode for this volume,
for one or all qtrees.

=head3 ARGUMENTS

=over

=item B<mode> I<Required> Scalar:

        Mode to set for oplocks (enable/disable).

=item B<qtree_name> I<Optional> Scalar:

        Name of the qtree for which to set oplocks mode.
	If this argument is not passed in, then oplocks mode will be set
	for all the qtrees of this volume.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->set_oplocks_mode(mode => FCACHE_MODE_ENABLE);

=cut

sub set_oplocks_mode {
    my $class = shift;
    my %args  = validate(
        @_,
        {   mode       => {type => SCALAR,},
            qtree_name => {type => SCALAR, optional => 1,},
            expected   => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {

        my $qtree = (defined($args{qtree_name})) ? ($args{qtree_name}) : '*';

        # TODO: Revisit this after burt 613397 is fixed
        my $command
            = "oplocks -volume "
            . $class->{_name}
            . " -vserver "
            . $class->{_vserver_name}
            . " -qtree $qtree "
            . " -oplock-mode "
            . $args{mode};
        my $api_set = $class->{_nacl_node}->apiset();
        my $output = $api_set->execute_command(command => $command);

        $class->{_ui_messages}->verify_expected_rc(
            command     => FCACHE_CMD_VOL_OPLOCKS_MODE_SET,
            output      => $output,
            expected_rc => $args{expected},
        );
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is not yet supported for 7-mode"
        );
    }

    return FCACHE_RC_SUCCESS;
} ## end sub set_oplocks_mode

=head2 modify

Modify settings of this volume.

=head3 ARGUMENTS

=over

=item B<privilege_level> I<Optional> Scalar:

        Privilege level to use when attempting the modify command.
	If this argument is not passed in, then the privilege level is not
	changed as part of testing (being left at the default level set by NACL).
	If this argument is passed in, then the library is also taking care of
	resetting the privilege level to the value it was set before calling
	this method.
	Legal values are: any of the FCACHE_PRIVILEGE_* constants.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=item B<unix_permissions> I<Optional> Scalar:

        Unix permissions to set.

=item B<security_style> I<Optional> Scalar:

        Security style to set.

=item B<qos_policy_group> I<Optional> Scalar:

        Qos policy group to set.

=item B<fc_delegs_enabled> I<Optional> Scalar:

        Flexcache-delegations-enabled value to set
	(FCACHE_SETTING_TRUE or FCACHE_SETTING_FALSE).
	
=item B<antivirus_on_access_policy> I<Optional> Scalar:

        Name of antivirus on-access policy to set
	(FCACHE_SETTING_TRUE or FCACHE_SETTING_FALSE).

=item B<flexcache_min_reserve> I<Optional> Scalar:

        flexcache-min-reserve value to set

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->modify(unix_permissions => FCACHE_DFLT_VOL_PERMS);
	$obj->modify(fc_delegs_enabled => FCACHE_SETTING_FALSE,
	             privilege_level => FCACHE_PRIVILEGE_ADMIN,
		     expected => FCACHE_RC_INVALID_CMD_ARGUMENT);


=cut

sub modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   privilege_level => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            unix_permissions           => {type => SCALAR, optional => 1,},
            security_style             => {type => SCALAR, optional => 1,},
            qos_policy_group           => {type => SCALAR, optional => 1,},
            fc_delegs_enabled          => {type => SCALAR, optional => 1,},
            antivirus_on_access_policy => {type => SCALAR, optional => 1,},
            flexcache_min_reserve      => {type => SCALAR, optional => 1,},
        }
    );

    my $privilege = delete($args{privilege_level});
    my $expected  = delete($args{expected});
    my ($apiset, $current_priv);
    my $error = '';

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is not yet supported for 7-mode"
        );
    }

    if (defined($privilege)) {

        # Remember the current privilege level, to reset it afterwards
        $apiset       = $class->{_nacl_node}->apiset();
        $current_priv = $apiset->get_privilege();
        $apiset->set_privilege('privilege-level' => $privilege);
    }

    my %args_to_set = ();
    foreach my $key (keys(%args)) {
        my $value = $args{$key};
        $key =~ s/_/-/g;
        $args_to_set{$key} = $value;
    }

    try {
        $class->{_nacl_vol}->modify(%args_to_set);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $error = $exception->text();
    };

    if (defined($privilege)) {

        # Reset the privilege
        $apiset->set_privilege('privilege-level' => $current_priv);
    }

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_MODIFY,
        output      => $error,
        expected_rc => $expected
    );
} ## end sub modify

=head2 get_qos_workload

Get Qos workload for this volume.

=head3 ARGUMENTS

None.

=head3 RETURN

Qos workload (Scalar).

EXAMPLE

        my $qos_workload = $obj->get_qos_workload();

=cut

sub get_qos_workload {
    my $class = shift;
    my %args = validate(@_, {});
    my $qos_workload;

    if ($class->{_boot_mode} eq FCACHE_ONTAP_CMODE) {
        my $result = NACL::C::QosWorkload->find(
            'command_interface' => $class->{_nacl_node},
            filter              => {volume => $class->{_name},},
        );
        $qos_workload = $result->workload();
    } else {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is not yet supported for 7-mode"
        );
    }

    return $qos_workload;
} ## end sub get_qos_workload

=head2 qtree_create

Create a qtree in this volume.

=head3 ARGUMENTS

=over

=item B<qtree_name> I<Required> Scalar:

        Name of the qtree.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->qtree_create(qtree_name => 'my_qtree',
	                   expected => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE);
	$obj->qtree_create(qtree_name => 'my_qtree');

=cut

sub qtree_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   qtree_name => {type => SCALAR,},
            expected   => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output = '';

    try {
        NACL::C::VolumeQtree->create(
            command_interface => $class->{_nacl_node},
            volume            => $class->{_name},
            vserver           => $class->{_vserver_name},
            qtree             => $args{qtree_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_QTREE_CREATE,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub qtree_create

=head2 qtree_modify

Modify qtree settings.

=head3 ARGUMENTS

=over

=item B<qtree_name> I<Required> Scalar:

        Name of the qtree.

=item B<security_style> I<Optional> Scalar: 

        Security style to set for the qtree.

=item B<oplock_mode> I<Optional> Scalar: 

        Oplock mode to set for the qtree.

=item B<unix_permissions> I<Optional> Scalar: 

        Unix permissions to set for the qtree.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->qtree_modify(security_style => FCACHE_DFLT_VOL_SEC_STYLE);

=cut

sub qtree_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   qtree_name       => {type => SCALAR,},
            security_style   => {type => SCALAR, optional => 1,},
            oplock_mode      => {type => SCALAR, optional => 1,},
            unix_permissions => {type => SCALAR, optional => 1,},
        }
    );

    my $qtree     = delete($args{qtree_name});
    my %nacl_args = (
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        qtree             => $qtree,
    );

    foreach my $field (keys(%args)) {
        my $value = $args{$field};
        $field =~ s/_/-/g;
        $nacl_args{$field} = $value;
    }

    NACL::C::VolumeQtree->modify(%nacl_args);

    return FCACHE_RC_SUCCESS;
} ## end sub qtree_modify

=head2 qtree_statistics_verify

Verify qtree statistics.

=head3 ARGUMENTS

=over

=item B<qtree_name> I<Required> Scalar:

        Name of the qtree.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=item B<expected_values> I<Optional> Hashref:

        Hashref of pairs "field name" => "expected value",
        to be passed in for the specific fields that need to be verified.
        Unless this argument is passed in, the method does not execute any value
        check against the statistics fields, but it just test that
        the qtree statistics can be obtained for the specifc file.
	This argument is only considered if 'expected' is FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

	$obj->qtree_statistics_verify(qtree_name => $test_qtree,
	                              expected => FCACHE_RC_NO_ENTRY_FOUND);
	$obj->qtree_statistics_verify(qtree_name => $test_qtree,);
	$obj->qtree_statistics_verify(qtree_name => $test_qtree,
	                              expected_values => {cifs_ops => 0,
				                          nfs_ops => 0,});

=cut

sub qtree_statistics_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   qtree_name => {type => SCALAR,},
            expected   => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
            expected_values => {type => HASHREF, optional => 1,},
        }
    );

    my $expected        = $args{expected};
    my $qtree           = $args{qtree_name};
    my $expected_values = $args{expected_values};
    my $error           = '';
    my $statistics;
    my %return = ();

    try {
        $statistics = NACL::C::VolumeQtree->statistics(
            command_interface => $class->{_nacl_node},
            filter            => {
                volume  => $class->{_name},
                vserver => $class->{_vserver_name},
                qtree   => $qtree,
            }
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $error = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_QTREE_STATISTICS,
        output      => $error,
        expected_rc => $expected,
    );

    if ($expected == FCACHE_RC_SUCCESS) {
        fcache_log_comment(
            message => "Statistics for qtree $qtree:\n" . Dumper($statistics))
            if ($class->{_debug});

        my @stats_fields = qw(cifs_ops no_reset_cifs_ops
            nfs_ops no_reset_nfs_ops
            internal_ops no_reset_internal_ops);

        foreach my $field (@stats_fields) {
            $return{$field} = $statistics->$field();
        }

        if (defined($expected_values)) {
            foreach my $expected_field (keys(%{$expected_values})) {
                my $value          = $return{$expected_field};
                my $expected_value = $expected_values->{$expected_field};
                if (   (!defined($value))
                    || ("$value" ne "$expected_value")) {

                    fcache_log_result(
                        result_type => FCACHE_LOGRESULT_FATAL,
                        message     => "Mismatch in qtree statistics:\n"
                            . "Qtree: $qtree\n"
                            . "Field: $expected_field\n"
                            . "Obtained value: $value\n"
                            . "Expected value: $expected_value"
                    );
                }

                fcache_log_comment(message => "Verified qtree statistics:\n"
                        . "Qtree: $qtree\n"
                        . "Field: $expected_field\n"
                        . "Obtained value: $value\n"
                        . "Expected value: $expected_value");
            }
        }
    }

    return %return;
} ## end sub qtree_statistics_verify

=head2 qtree_delete

Delete a qtree of this volume.

=head3 ARGUMENTS

=over

=item B<qtree_name> I<Required> Scalar:

        Name of the qtree.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->qtree_delete(qtree_name => 'my_qtree');

=cut

sub qtree_delete {
    my $class = shift;
    my %args = validate(@_, {qtree_name => {type => SCALAR,},});

    NACL::C::VolumeQtree->delete(
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        qtree             => $args{qtree_name},
    );

    return FCACHE_RC_SUCCESS;
} ## end sub qtree_delete

=head2 flexcache_options_verify

Verify volume flexcache options.
If this method is called without any arguments, then no value verification is
done, the only check will be that the command is successful.
WARNING: Even though the "volume flexcache options" is a volume-level command,
there is no volume-level granularity provided by the SW at this moment 
(the options apply to ALL the volumes).

=head3 ARGUMENTS

=over

=item B<delegations_high_watermark> I<Optional> Scalar:

	Expected value for the delegations_high_watermark field.

=item B<delegations_low_watermark> I<Optional> Scalar:

	Expected value for the delegations_low_watermark field.

=item B<capability> I<Optional> Scalar:

	Expected value for the capability field.

=item B<per_cache_statistics> I<Optional> Scalar:

	Expected value for the per_cache_statistics field.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

Hash with pairs field-value of obtained results.

EXAMPLE

        $obj->flexcache_options_verify(delegations_high_watermark => 75,
	                               delegations_low_watermark => 25,
				       per_cache_statistics => FCACHE_SETTING_FALSE);

=cut

sub flexcache_options_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   delegations_high_watermark =>
                {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            delegations_low_watermark =>
                {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            capability           => {type => SCALAR, optional => 1,},
            per_cache_statistics => {type => SCALAR, optional => 1,},
            ui_type              => {type => SCALAR, optional => 1,},
        }
    );

    my $ui_type = delete($args{ui_type});
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    my %return = (
        delegations_high_watermark => 0,
        delegations_low_watermark  => 0,
        capability                 => 0,
        per_cache_statistics       => 0
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This method is currently not supported for 7-mode",
        );
    }

    my $options = NACL::CS::VolumeFlexcacheOptions->fetch(
        command_interface => $class->{_nacl_node},
        apiset_must       => {interface => $ui_type},
        requested_fields  => [
            'delegations-high-watermark', 'delegations-low-watermark',
            'per-cache-statistics',       'capability',
        ],
    );

    foreach my $field (keys(%return)) {
        my $expected_value = $args{$field};
        my $obtained_value = $options->$field();
        $return{$field} = $obtained_value;

        if (defined($expected_value)) {
            fcache_log_comment(
                      message => "Verifying volume flexcache options:\n"
                    . "Field: $field\n"
                    . "Expected value: $expected_value\n"
                    . "Obtained value: $obtained_value");

            if ("$obtained_value" ne "$expected_value") {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_FATAL,
                    message     => "Mismatch in volume flexcache options:\n"
                        . "Field: $field\n"
                        . "Expected value: $expected_value\n"
                        . "Obtained value: $obtained_value"
                );
            }
        }
    }

    return %return;
} ## end sub flexcache_options_verify

=head2 flexcache_options_modify

Modify volume flexcache options.
WARNING: Even though the "volume flexcache options" is a volume-level command,
there is no volume-level granularity provided by the SW at this moment 
(the options apply to ALL the volumes).

=head3 ARGUMENTS

=over

=item B<delegations_high_watermark> I<Optional> Scalar:

	New value for the delegations_high_watermark field.

=item B<delegations_low_watermark> I<Optional> Scalar:

	New value for the delegations_low_watermark field.

=item B<capability> I<Optional> Scalar:

	New value for the capability field.

=item B<per_cache_statistics> I<Optional> Scalar:

	New value for the per_cache_statistics field.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->flexcache_options_modify(delegations_high_watermark => 75,
	                               delegations_low_watermark => 25,
				       per_cache_statistics => FCACHE_SETTING_FALSE);

=cut

sub flexcache_options_modify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   delegations_high_watermark =>
                {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            delegations_low_watermark =>
                {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            capability           => {type => SCALAR, optional => 1,},
            per_cache_statistics => {type => SCALAR, optional => 1,},
            ui_type              => {type => SCALAR, optional => 1,},
        }
    );

    my $ui_type = delete($args{ui_type});
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This method is currently not supported for 7-mode",
        );
    }

    #Obtain the current settings, to be able to reset them at cleanup
    my %current_settings = $class->flexcache_options_verify();

    my %nacl_args = (apiset_must => {interface => $ui_type},);

    foreach my $param (keys(%args)) {
        my $new_value     = $args{$param};
        my $current_value = $current_settings{$param};

        if ("$current_value" eq "$new_value") {
            fcache_log_comment(message => "Volume flexcache options:\n"
                    . "$param already set to $new_value");
            next;
        }

        # Only remember the value from the first time we ever change
        # an option. This is what we need to reset to at cleanup.
        if (!exists($class->{_cleanup_flexcache_options}->{$param})) {
            $class->{_cleanup_flexcache_options}->{$param} = $current_value;
        }

        $param =~ s/_/-/g;
        $nacl_args{$param} = $new_value;
    }

    NACL::C::VolumeFlexcacheOptions->modify(
        command_interface => $class->{_nacl_node},
        %nacl_args
    );

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_options_modify

=head2 flexcache_create

Create cache volumes for this volume by using the flexcache create command.

=head3 ARGUMENTS

=over

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

        Hash with key <-> values of newly created caches:
        cache_name <-> hash with : vserver (name of the vserver for this cache),
                                   aggregate (name of the aggregate containing the cache)
				   node (name of the node hosting the cache).

EXAMPLE

        my %caches = $obj->flexcache_create();

=cut

sub flexcache_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   expected => {type => SCALAR, default  => FCACHE_RC_SUCCESS,},
            ui_type  => {type => SCALAR, optional => 1,},
        }
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This method is currently not supported for 7-mode",
        );
    }

    my $vol_name    = $class->{_name};
    my $expected_rc = $args{expected};
    my $ui_type     = $args{ui_type};
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }
    my $output = '';
    my %existent_caches;
    my %created_caches;

    if ($expected_rc == FCACHE_RC_SUCCESS) {

        # First see if the volume already has caches,
        # they should not be returned as created by this method
        %existent_caches = $class->get_cache_volumes(
            expected => FCACHE_RC_FAILURE_NON_FATAL);
        fcache_log_comment(message => "Before executing flexcache create:\n"
                . "Volume $vol_name has the following caches:\n"
                . Dumper(\%existent_caches))
            if ($class->{_debug});
    }

    try {
        NACL::C::VolumeFlexcache->create(
            command_interface => $class->{_nacl_node},
            apiset_must       => {interface => $ui_type},
            vserver           => $class->{_vserver_name},
            'origin-volume'   => $class->{_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_FLEXCACHE_CREATE,
        output      => $output,
        expected_rc => $expected_rc
    );

    if ($expected_rc == FCACHE_RC_SUCCESS) {
        %created_caches = $class->get_cache_volumes();
        fcache_log_comment(message => "After executing flexcache create:\n"
                . "Volume $vol_name has the following caches:\n"
                . Dumper(\%created_caches))
            if ($class->{_debug});

        foreach my $existent_cache (keys(%existent_caches)) {
            delete($created_caches{$existent_cache});
        }

        fcache_log_comment(
            message => "Executed flexcache create for volume $vol_name:\n"
                . "Newly created caches:\n"
                . Dumper(\%created_caches));
    }

    return %created_caches;
} ## end sub flexcache_create

=head2 flexcache_verify

Volume flexcache verify.
This method is only available in C-mode.
(CLI: "volume flexcache show" , ZAPI: "flexcache-get-iter").
If this method is called without any arguments, then no value verification is
done, the only check will be that the command is successful.

=head3 ARGUMENTS

=over

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=item B<expected_values> I<Optional> Hashref:

	Hashref of values to be checked for one or more caches.
	The key => values pairs for this hashref should be:
	<name of cache to look for> => <hashref of values to check>
	If this argument is not passed in, then no value verification is
	done, the only check will be that the command is successful.

=back

=head3 RETURN

Hash with pairs field-value of obtained results.

EXAMPLE

        $obj->flexcache_verify(cache1 => {origin_volume => 'origin1'},
	                       cache2 => {origin_volume => 'origin2',
			                  cache_state => FCACHE_VOL_STATE_ONLINE,
					  cache_aggregate => 'aggr2'},);

=cut

sub flexcache_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   ui_type         => {type => SCALAR,  optional => 1,},
            expected_values => {type => HASHREF, optional => 1,},
        }
    );

    my $expected_values = $args{expected_values};
    my $ui_type         = $args{ui_type};
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }

    # CLI output contains 'cache_node' as well. ZAPI doesn't return this field
    my @return_fields = (
        'vserver',            'origin_volume',
        'origin_aggregate',   'origin_state',
        'cache_volume',       'cache_aggregate',
        'cache_size',         'cache_state',
        'cache_percent_used', 'cache_available',
    );
    my %return = ();

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "This method is only supported for C-mode",
        );
    }

    my @caches = NACL::CS::VolumeFlexcache->fetch(
        command_interface => $class->{_nacl_node},
        apiset_must       => {interface => $ui_type},
    );
    fcache_log_comment(message => "Processing \"volume flexcache\" output:\n"
            . Dumper(@caches))
        if ($class->{_debug});

    foreach my $cache (@caches) {

        my $cache_name = $cache->cache_volume();
        if (!defined($cache_name)) {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Incomplete volume flexcache output:\n"
                    . "Missing: cache_volume\n"
                    . Dumper($cache)
            );
        }

        foreach my $field (@return_fields) {
            my $obtained_value = $cache->$field();
            if (defined($obtained_value)) {
                $return{$cache_name}{$field} = $obtained_value;
            }

            my $expected_value = $expected_values->{$cache_name}->{$field};
            if (defined($expected_value)) {
                if (   (!defined($obtained_value))
                    || ("$obtained_value" ne "$expected_value")) {

                    fcache_log_result(
                        result_type => FCACHE_LOGRESULT_FATAL,
                        message => "Mismatch in volume flexcache fields:\n"
                            . "Field: $field\n"
                            . "Expected value: $expected_value\n"
                            . "Obtained value: $obtained_value"
                    );
                }

                fcache_log_comment(
                          message => "Verified volume flexcache field:\n"
                        . "Cache: $cache_name\n"
                        . "Field: $field\n"
                        . "Expected value: $expected_value\n"
                        . "Obtained value: $obtained_value");
            }
        }
    }

    return %return;
} ## end sub flexcache_verify

=head2 flexcache_delete

Delete caches of this volume by using the flexcache delete command.

=head3 ARGUMENTS

=over

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=item B<ui_type> I<Optional> Scalar:

	UI to be used when executing the command (ZAPI or CLI).
	The value of the command-line argument FCACHE_UI.

=back

=head3 RETURN

        FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->flexcache_delete();

=cut

sub flexcache_delete {
    my $class = shift;
    my %args  = validate(
        @_,
        {   expected => {type => SCALAR, default  => FCACHE_RC_SUCCESS,},
            ui_type  => {type => SCALAR, optional => 1,},
        }
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message => "This method is currently not supported for 7-mode",
        );
    }

    my $vol_name    = $class->{_name};
    my $expected_rc = $args{expected};
    my $ui_type     = $args{ui_type};
    if (!defined($ui_type)) {
        my $options = NACL::OptionSet->new();
        $ui_type = $options->get('FCACHE_UI');
    }
    my $output = '';

    try {
        NACL::C::VolumeFlexcache->delete(
            command_interface => $class->{_nacl_node},
            apiset_must       => {interface => $ui_type},
            vserver           => $class->{_vserver_name},
            'origin-volume'   => $class->{_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_FLEXCACHE_DELETE,
        output      => $output,
        expected_rc => $expected_rc
    );

    return FCACHE_RC_SUCCESS;
} ## end sub flexcache_delete

=head2 get_file_id

Get the ID of a file within this volume.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

	Name of the file (relative path, as seen in the volume).

=back

=head3 RETURN

File ID (Scalar).

EXAMPLE

        $obj->get_file_id(file_name => 'my_file',);

=cut

# Internal-use argument: _file_path: Optional, scalar: file path to be used,
# useful only in the case when the path is not in this volume (for example,
# if we need to consider the origin volume, instead of the cache).
sub get_file_id {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name  => {type => SCALAR,},
            _file_path => {type => SCALAR | UNDEF, optional => 1,},
        }
    );

    my $file_path;
    if (defined($args{_file_path})) {
        $file_path = $args{_file_path};
    } else {
        $file_path = '/vol/' . $class->{_name} . '/' . $args{file_name};
    }

    my $fh = NACL::C::VolumeFile->showfh(
        command_interface                    => $class->{_nacl_node},
        path                                 => $file_path,
        vserver                              => $class->{_vserver_name},
        determine_appropriate_nodescope_node => 1
    );
    my $file_id = $fh->fileid();
    if (!defined($file_id)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Unable to determine the file id:\n"
                . "File path: $file_path\n"
                . Dumper($fh)
        );
    }

    return hex($file_id);
} ## end sub get_file_id

=head2 purge

Purge this volume (offline and destroy it).

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $obj->purge();

=cut

sub purge {
    my $class = shift;
    my %args = validate(@_, {});

    my $volume = $class->{_name};

    $class->{_nacl_vol}->purge();
    $class->{_is_purged} = TRUE;

    fcache_log_comment(
        message => "Volume has been purged:\n" . "Volume name: $volume");

    return FCACHE_RC_SUCCESS;
} ## end sub purge

=head2 efficiency_enable

Enable storage efficiency (dedupe) on this volume. It will not complain if SE is
already turned on for the volume. This method only turns SE on, it does not 
start the SE scanner to gain any savings. Unless otherwise specified, after 
turning SE on, it also configures the SE scheduler to not start the SE scanner
automatically.

=head3 ARGUMENTS

=over

=item B<disable_scheduler> I<Optional> Boolean:

        Flag to control if the SE schedule should be configured to none after
	SE has been turned on. Default value: TRUE.

=item B<expected> I<Optional> Scalar:

        Expected return code for this action, as defined in FlexCache::Constants.
        Script execution is terminated if the expected return code is not obtained.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

=head3 WARNING

	None of the SE operations will be "undone" at the time of cleanup. In
	other words, when a script exits, no SE cleanup will be executed. This
	is because in most of the cases, the test volumes will be destroyed at
	the time of cleanup.

EXAMPLE

        $obj->efficiency_enable();

=cut

sub efficiency_enable {
    my $class = shift;
    my %args  = validate(
        @_,
        {   disable_scheduler => {type => BOOLEAN, default => TRUE,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $volume          = $class->{_name};
    my $configure_sched = $args{disable_scheduler};
    my $output          = '';
    my $se;

    try {
        $se = NACL::STask::VolumeEfficiency->on(
            command_interface => $class->{_nacl_node},
            volume            => $volume,
            vserver           => $class->{_vserver_name},
            ignore_enabled    => 1,
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    if ((defined($se)) && ($configure_sched)) {
        try {
            $se->config(schedule => '-');
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $output = $exception->text();
        };
    }

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_SE_ENABLE,
        output      => $output,
        expected_rc => $args{expected},
    );
} ## end sub efficiency_enable

=head2 efficiency_disable

Disable storage efficiency (dedupe) on this volume. It will not complain if SE 
is already turned off for the volume. 

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS

=head3 WARNING

	None of the SE operations will be "undone" at the time of cleanup. In
	other words, when a script exits, no SE cleanup will be executed. This
	is because in most of the cases, the test volumes will be destroyed at
	the time of cleanup.

EXAMPLE

        $obj->efficiency_disable();

=cut

sub efficiency_disable {
    my $class = shift;
    my %args = validate(@_, {});

    NACL::STask::VolumeEfficiency->off(
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        ignore_disabled   => 1,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub efficiency_disable

=head2 compression_configure

Configure SE compression settings for this volume. Please note that SE needs to
be already turned on in order to be able to configure compression setings.

=head3 ARGUMENTS

=over

=item B<background_compression> I<Optional> Boolean:

        Flag to control if background compression should be enabled or disabled
	for the volume. Default value: TRUE.
	
=item B<inline_compression> I<Optional> Boolean:

        Flag to control if inline compression should be enabled or disabled
	for the volume. Default value: TRUE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

=head3 WARNING

	None of the SE operations will be "undone" at the time of cleanup. In
	other words, when a script exits, no SE cleanup will be executed. This
	is because in most of the cases, the test volumes will be destroyed at
	the time of cleanup.

EXAMPLE

        # Turn on both background and inline compression
	$obj->compression_configure();
	# Turn on only background compression
	$obj->compression_configure(inline_compression => FALSE);
	# Turn off both background and inline compression
	$obj->compression_configure(background_compression => FALSE,
	                            inline_compression => FALSE);

=cut

sub compression_configure {
    my $class = shift;
    my %args  = validate(
        @_,
        {   background_compression => {type => BOOLEAN, default => TRUE,},
            inline_compression     => {type => BOOLEAN, default => TRUE,},
        }
    );

    my $background
        = ($args{background_compression})
        ? FCACHE_SETTING_TRUE
        : FCACHE_SETTING_FALSE;
    my $inline
        = ($args{inline_compression})
        ? FCACHE_SETTING_TRUE
        : FCACHE_SETTING_FALSE;

    NACL::STask::VolumeEfficiency->config(
        command_interface        => $class->{_nacl_node},
        volume                   => $class->{_name},
        vserver                  => $class->{_vserver_name},
        'background-compression' => $background,
        'inline-compression'     => $inline,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub compression_configure

=head2 efficiency_scanner_start

Start the SE scanner on this volume. Please note that SE needs to be already 
turned on. Unless otherwise specified, it will wait until the scanner
finishes processing all the data. The savings gained after the scanner
completes depends on the type of SE settings (dedupe/compression) are
configured for this volume.

=head3 ARGUMENTS

=over

=item B<background> I<Optional> Boolean:

        Flag to control if the method should wait until the scanner becomes
	idle. Default value: FALSE.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

=head3 WARNING

	None of the SE operations will be "undone" at the time of cleanup. In
	other words, when a script exits, no SE cleanup will be executed. This
	is because in most of the cases, the test volumes will be destroyed at
	the time of cleanup.

EXAMPLE

        $obj->efficiency_scanner_start();

=cut

sub efficiency_scanner_start {
    my $class = shift;
    my %args
        = validate(@_, {background => {type => BOOLEAN, default => FALSE,},});

    my $wait = ($args{background}) ? 0 : 1;

    # For now, not supporting extra options
    NACL::STask::VolumeEfficiency->start(
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        nacltask_wait     => $wait,
    );

    return FCACHE_RC_SUCCESS;
} ## end sub efficiency_scanner_start

=head2 efficiency_scanner_wait

Wait until the SE scanner on this volume becomes idle. Useful if the
efficiency_scanner_start was called with the background option.

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS

=head3 WARNING

	None of the SE operations will be "undone" at the time of cleanup. In
	other words, when a script exits, no SE cleanup will be executed. This
	is because in most of the cases, the test volumes will be destroyed at
	the time of cleanup.

EXAMPLE

        $obj->efficiency_scanner_wait();

=cut

sub efficiency_scanner_wait {
    my $class = shift;
    my %args = validate(@_, {});

    NACL::STask::VolumeEfficiency->wait_for_status(
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        status            => 'Idle',
    );

    return FCACHE_RC_SUCCESS;
} ## end sub efficiency_scanner_wait

=head2 get_efficiency_savings

Obtain SE (dedupe/compression) space savings for this volume.

=head3 ARGUMENTS

=over

=item B<expected_values> I<Optional> Hashref:

	Hashref of expected values for the SE savings. If this argument is not
	passed in, then no actual SE savings value check is executed by the
	library, the obtained savings are simply returned to the caller in the
	form of a hash. If this argument is passed in, then the values obtained
	for the SE savings are checked against the user-expected values.
	The keys of this hash may be any of the keys of the hash that this
	method would otherwise return. Currently, those are: compression_saved,
	compression_saved_percent, dedupe_saved, dedupe_saved_percent.
	The value that the caller can pass in for any of these keys can be:
	- a scalar, in which case the obtained SE savings value for that key
	  should match exactly the value of this scalar.
	- a hashref with 'min' and/or 'max' values, in which case the obtained 
	  SE savings value for that key should be at least equal to 'min' and/or
	  at most equal to 'max'.

=back

=head3 RETURN

	Hash containing the obtained values for SE savings.

EXAMPLE

        # Only get the SE savings, no actual value check is done.
        $obj->get_efficiency_savings();
	# Check that there are no compression savings.
        $obj->get_efficiency_savings(
	    expected_values => {compression_saved => 0,});
	# Check that there are some compression savings.
        $obj->get_efficiency_savings(
	    expected_values => {compression_saved => {min => 1},});
	# Check that dedupe savings are not bigger than 100.
        $obj->get_efficiency_savings(
	    expected_values => {dedupe_saved => {max => 100},});
	# Check that compression savings are in between 200 and 500 and
	# there are no dedupe savings.
        $obj->get_efficiency_savings(
	    expected_values => {
	        compression_saved => {min => 200, max => 500},
		dedupe_saved => 0
	    });

=cut

sub get_efficiency_savings {
    my $class = shift;
    my %args  = validate(@_,
        {expected_values => {type => HASHREF, optional => 1,},});

    my $vol_name = $class->{_name};
    my $state    = $class->{_nacl_vol}->state();
    my %return   = ();

    $return{compression_saved} = $state->compression_space_saved();
    $return{compression_saved_percent}
        = $state->compression_space_saved_percent();
    $return{dedupe_saved}         = $state->dedupe_space_saved();
    $return{dedupe_saved_percent} = $state->dedupe_space_saved_percent();

    if (defined($args{expected_values})) {
        my $expected_values = $args{expected_values};
        foreach my $field (keys(%{$expected_values})) {
            my $obtained_value = $return{$field};
            if (!defined($obtained_value)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message     => "Invalid SE savings field: $field\n"
                        . "Available SE savings fields are:\n"
                        . Dumper(keys(%return))
                );
            }

            my $expected_value = $expected_values->{$field};
            if (!defined($expected_value)) {
                fcache_log_result(
                    result_type => FCACHE_LOGRESULT_SCRIPT,
                    message =>
                        "Undefined expected value for SE savings field: $field"
                );
            } elsif (ref($expected_value) eq 'HASH') {
                my %margins = validate_with(
                    params => $expected_value,
                    spec   => {
                        min => {type => SCALAR, optional => 1,},
                        max => {type => SCALAR, optional => 1,},
                    },
                    allow_extra => 0,
                );

                if (scalar(keys(%margins)) == 0) {
                    fcache_log_result(
                        result_type => FCACHE_LOGRESULT_SCRIPT,
                        message     => "At least one of the min or max value "
                            . "for field $field must be provided when passing "
                            . "the expected value as a hash reference"
                    );
                }

                if (defined($margins{min})) {
                    my $min = $margins{min};
                    if ($obtained_value < $min) {
                        fcache_log_result(
                            result_type => FCACHE_LOGRESULT_FATAL,
                            message =>
                                "Mismatch in expected SE savings values:\n"
                                . "Field: $field\n"
                                . "Volume: $vol_name\n"
                                . "Minimum allowed value: $min\n"
                                . "Obtained value: $obtained_value"
                        );
                    }

                    fcache_log_comment(message =>
                              "Verified SE savings for volume $vol_name:\n"
                            . "Field: $field\n"
                            . "Minimum allowed value: $min\n"
                            . "Obtained value: $obtained_value");
                }

                if (defined($margins{max})) {
                    my $max = $margins{max};
                    if ($obtained_value > $max) {
                        fcache_log_result(
                            result_type => FCACHE_LOGRESULT_FATAL,
                            message =>
                                "Mismatch in expected SE savings values:\n"
                                . "Field: $field\n"
                                . "Volume: $vol_name\n"
                                . "Maximum allowed value: $max\n"
                                . "Obtained value: $obtained_value"
                        );
                    }

                    fcache_log_comment(message =>
                              "Verified SE savings for volume $vol_name:\n"
                            . "Field: $field\n"
                            . "Maximum allowed value: $max\n"
                            . "Obtained value: $obtained_value");
                }
            } else {

                # Scalar value
                if ($expected_value != $obtained_value) {
                    fcache_log_result(
                        result_type => FCACHE_LOGRESULT_FATAL,
                        message => "Mismatch in expected SE savings values:\n"
                            . "Field: $field\n"
                            . "Volume: $vol_name\n"
                            . "Expected value: $expected_value\n"
                            . "Obtained value: $obtained_value"
                    );
                }

                fcache_log_comment(
                    message => "Verified SE savings for volume $vol_name:\n"
                        . "Field: $field\n"
                        . "Expected value: $expected_value\n"
                        . "Obtained value: $obtained_value");
            }
        }
    }

    return %return;
} ## end sub get_efficiency_savings

=head2 antivirus_on_demand_scan_run

Run the antivirus on-demand scan on this volume.

=head3 ARGUMENTS

None.

=head3 RETURN

    FCACHE_RC_SUCCESS.

EXAMPLE

        $obj->antivirus_on_demand_scan_dir_run();

=cut

sub antivirus_on_demand_scan_dir_run {
    my $class = shift;
    my %args = validate(@_, {});

    my $vol_name     = $class->{_name};
    my $command_name = 'scan_dir_' . $vol_name;
    my $command_was_created;

    NACL::STask::AntivirusOnDemandCommandScanDir->create(
        command_interface  => $class->{_nacl_node},
        name               => $command_name,
        vserver            => $class->{_vserver_name},
        'scan-path'        => "/$vol_name/",
        nacltask_if_exists => 'reuse',
        _was_created       => \$command_was_created,
    );

    if ($command_was_created) {
        $class->{_cleanup_av_scan_dir_command} = $command_name;
    }

    NACL::STask::AntivirusOnDemand->run(
        command_interface => $class->{_nacl_node},
        command           => $command_name,
        vserver           => $class->{_vserver_name},
        verbose           => 'true',
        print             => 'true',
    );

    return FCACHE_RC_SUCCESS;
} ## end sub antivirus_on_demand_scan_dir_run

=head2 file_clone_create

Create a file clone for a file in this volume.

=head3 ARGUMENTS

=over

=item B<file_name> I<Required> Scalar:

        Name of the file to be cloned.

=item B<clone_name> I<Optional> Scalar:

        Name of the new clone, to be created.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $origin->{volume}->file_clone_create(
            file_name  => $file,
	    clone_name => $clone_file,
        );
        $cache->{volume}->file_clone_create(
	    file_name  => 'file',
	    clone_name => 'clone',
	    expected   => FCACHE_RC_OPERATION_NOT_ALLOWED_ON_A_CACHE,
        );

=cut

sub file_clone_create {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name  => {type => SCALAR,},
            clone_name => {type => SCALAR,},
            expected   => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $output      = '';
    my $vol_name    = $class->{_name};
    my %create_args = (
        command_interface  => $class->{_nacl_node},
        volume             => $vol_name,
        vserver            => $class->{_vserver_name},
        'source-path'      => $args{file_name},
        'destination-path' => $args{clone_name},
        clone_type         => 'file',
    );

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        $create_args{'source-path'} = "/vol/$vol_name/" . $args{file_name};
        $create_args{'destination-path'}
            = "/vol/$vol_name/" . $args{clone_name};
    }

    try {
        NACL::C::VolumeFileClone->create(%create_args);
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_FILE_CLONE_CREATE,
        output      => $output,
        expected_rc => $args{expected}
    );
} ## end sub file_clone_create

=head2 autogrow_enable

Enable autogrow for this volume.

=head3 ARGUMENTS

None.

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $origin->{volume}->autogrow_enable();

=cut

sub autogrow_enable {
    my $class = shift;
    my %args = validate(@_, {});

    # For 7-mode caches, autogrow is a volume option
    if (   ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE)
        && (defined($class->{_origin_name}))) {
        my $api_set = $class->{_nacl_node}->apiset();
        $api_set->vol_options(
            volume         => $class->{_name},
            'option-name'  => 'flexcache_autogrow',
            'option-value' => FCACHE_STATE_ON,
        );
    } else {
        $class->{_nacl_vol}->modify('autosize' => FCACHE_SETTING_TRUE);
    }

    return FCACHE_RC_SUCCESS;
} ## end sub autogrow_enable

=head2 quota_enable

Enable quota for this volume.

=head3 ARGUMENTS

=over

=item B<disk_limit> I<Optional> Scalar:

        Size for the quota disk limit.

=item B<file_limit> I<Optional> Scalar:

        Quota file limit.

=item B<expected> I<Optional> Scalar: 

        Expected return code for this action, as defined in FlexCache::Constants.
        Legal values: any of the FCACHE_RC_* constants.
        Default value: FCACHE_RC_SUCCESS.

=back

=head3 RETURN

FCACHE_RC_SUCCESS

EXAMPLE

        $origin->{volume}->quota_enable(disk_limit => '25m');

=cut

sub quota_enable {
    my $class = shift;
    my %args  = validate(
        @_,
        {   disk_limit => {type => SCALAR, optional => 1,},
            file_limit => {type => SCALAR, optional => 1,},
            expected   => {type => SCALAR, default  => FCACHE_RC_SUCCESS,},
        }
    );

    my $expected_rc = delete($args{expected});
    my $output      = '';

    if (scalar(keys(%args)) == 0) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_SCRIPT,
            message     => "At least one of the arguments must be provided:\n"
                . "disk_limit, file_limit"
        );
    }

    my %rule_args = (
        command_interface => $class->{_nacl_node},
        volume            => $class->{_name},
        vserver           => $class->{_vserver_name},
        'policy-name'     => 'default',
        type              => 'user',
    );

    if (defined($args{disk_limit})) {
        $rule_args{'disk-limit'} = $args{disk_limit};
    }

    if (defined($args{file_limit})) {
        $rule_args{'file-limit'} = $args{file_limit};
    }

    if ($class->{_boot_mode} eq FCACHE_ONTAP_7MODE) {
        $rule_args{target} = '*';
        $rule_args{qtree}  = '*';
    } else {
        $rule_args{target} = '""';
        $rule_args{qtree}  = '""';
    }

    try {
        NACL::C::VolumeQuotaPolicyRule->create(%rule_args);

        NACL::C::VolumeQuota->on(
            command_interface => $class->{_nacl_node},
            volume            => $class->{_name},
            vserver           => $class->{_vserver_name},
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    return $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_VOL_QUOTA_ENABLE,
        output      => $output,
        expected_rc => $expected_rc
    );
} ## end sub quota_enable

=head2 get_flexcache_min_reserve

Get the flexcache-min-reserve value.

=head3 ARGUMENTS

=over

=item B<expected_value> I<Optional> Scalar:

        Expected value for flexcache-min-reserve. If this argument is not
        passed in, then no value check is done, but it is just verified that
        the current value can be retrieved.

=back

=head3 RETURN

Flexcache-min-reserve value (Scalar).

EXAMPLE

        $origin->{volume}->get_flexcache_min_reserve();
        $origin->{volume}->get_flexcache_min_reserve(
            expected_value => 100 * MB,
        );

=cut

sub get_flexcache_min_reserve {
    my $class = shift;
    my %args
        = validate(@_, {expected_value => {type => SCALAR, optional => 1,},});

    my $expected_value = $args{expected_value};

    my $vol_state     = $class->{_nacl_vol}->state();
    my $current_value = $vol_state->flexcache_min_reserve();

    if (!defined($current_value)) {
        fcache_log_result(
            result_type => FCACHE_LOGRESULT_FATAL,
            message     => "Unable to retrieve flexcache-min-reserve value:\n"
                . Dumper($vol_state)
        );
    }

    if (defined($expected_value)) {
        fcache_log_comment(
                  message => "Verifying flexcache-min-reserve value:\n"
                . "Volume: "
                . $class->{_name} . "\n"
                . "Expected value: $expected_value\n"
                . "Obtained value: $current_value");

        if ("$current_value" ne "$expected_value") {
            fcache_log_result(
                result_type => FCACHE_LOGRESULT_FATAL,
                message     => "Mismatch in flexcache-min-reserve value:\n"
                    . "Volume: "
                    . $class->{_name} . "\n"
                    . "Expected value: $expected_value\n"
                    . "Obtained value: $current_value"
            );
        }
    }

    return $current_value;
} ## end sub get_flexcache_min_reserve

########################################################
# Internal methods
########################################################

# flexcache_delegation_verify
# implementation for C-mode.
# Please see parent method for documentation.
sub _CMode_flexcache_delegation_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name => {type => SCALAR,},
            file_id   => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            delegation_mode  => {type => SCALAR, optional => 1,},
            delegation_state => {type => SCALAR, optional => 1,},
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name              = $args{file_name};
    my $file_id                = $args{file_id};
    my $expected               = $args{expected};
    my $required_lock_protocol = $class->{_ui_messages}
        ->get_constant_value(constant => 'FCACHE_DELEGATIONS_LOCK_PROTOCOL');
    my $output           = '';
    my $found_delegation = FALSE;
    my @delegations      = ();
    my $vol_name         = $class->{_name};

    # Paths are listed as origin-paths (<origin_vol>/<file>)
    # _origin_name is only defined for cache volumes, so:
    my $origin_vol
        = (defined($class->{_origin_name}))
        ? ($class->{_origin_name})
        : ($vol_name);

    # We are going to keep a brief listing with all the delegations
    # found for this volume, useful in case of failure.
    # Dumping the complete state is not useful in this case,
    # as it is a huge hash that makes finding the pertinent info
    # extremely hard.
    my $current_delegations_listing = '';

    # We need to search for the delegation by file id, and not by file path.
    # This is because in the case of a revoked delegation, the path
    # is not available in the show command, but the file id always is.
    if (!defined($file_id)) {
        $file_id = $class->get_file_id(
            file_name  => $file_name,
            _file_path => "/vol/$origin_vol/$file_name",
        );
    }

    my %find_args = (
        volume   => $vol_name,
        vserver  => $class->{_vserver_name},
        fileid   => $file_id,
        protocol => $required_lock_protocol,
    );

    if (defined($args{delegation_state})) {
        my $required_state = $class->{_ui_messages}
            ->get_constant_value(constant => $args{delegation_state});
        $find_args{'lock-state'} = $required_state;
    }

    if (defined($args{delegation_mode})) {
        my $required_mode = $class->{_ui_messages}
            ->get_constant_value(constant => $args{delegation_mode});
        $find_args{mode} = $required_mode;
    }
    my $delegation;
    try {
        $delegation = NACL::C::VolumeFlexcacheDelegations->find(
            command_interface => $class->{_nacl_node},
            filter            => \%find_args,
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $output = $exception->text();
    };

    # Handle the timing problem when we expect the delegation to be revoked.
    # In this case, depending on timing, we may either not find any delegation
    # at all, but we may also find an "UNLOCKED" one (which after a while no
    # longer appears). In case the user expects to not find a delegation,
    # also consider ok to find an "UNLOCKED" one.
    if (($expected == FCACHE_RC_NO_ENTRY_FOUND) && (defined($delegation))) {
        fcache_log_comment(
            message => "Delegations not expected, but found for $file_name.\n"
                . "Verifying that they are in unlocked state...");
        my $state;
        try {
            $state = $delegation->state();
        }
        catch NATE::BaseException with {

            # Delegation must have disappeared in the meantime. Return ok.
            return $file_id;
        };
        my $delegation_state = $state->lock_state();
        my $unlocked_state   = $class->{_ui_messages}->get_constant_value(
            constant => FCACHE_VOL_DELEGATION_STATE_UNLOCKED);
        if ($delegation_state eq $unlocked_state) {
            fcache_log_comment(message =>
                    "Delegations for file $file_name is in unlocked state");
            return $file_id;
        }
    }

    $class->{_ui_messages}->verify_expected_rc(
        command     => FCACHE_CMD_DELEGATIONS_FIND,
        output      => $output,
        expected_rc => $expected
    );

    return $file_id;
} ## end sub _CMode_flexcache_delegation_verify

# flexcache_delegation_verify
# implementation for 7-mode.
# Please see parent method for documentation.
sub _7Mode_flexcache_delegation_verify {
    my $class = shift;
    my %args  = validate(
        @_,
        {   file_name => {type => SCALAR,},
            file_id   => {type => SCALAR, regex => qr/^\d+$/, optional => 1,},
            delegation_mode  => {type => SCALAR, optional => 1,},    # ignored
            delegation_state => {type => SCALAR, optional => 1,},    # ignored
            expected => {type => SCALAR, default => FCACHE_RC_SUCCESS,},
        }
    );

    my $file_name = $args{file_name};
    my $file_id   = $args{file_id};
    my $expected  = $args{expected};
    my $vol_name  = $class->{_name};
    my $rc        = FCACHE_RC_NO_ENTRY_FOUND;

    # For 7-mode, delegations are only listed by file id. Need to find it,
    # if not passed in.
    if (!defined($file_id)) {
        $file_id = $class->get_file_id(file_name => $file_name);
    }

    # For 7-mode, on the origin we will not see the delegations.
    # To enable the scripts to work on both C-mode and 7-mode,
    # make this method return success for 7-mode origins, such that C-mode
    # checks can be still performed on origins as well.
    # _origin_name is only defined for cache volumes
    if (!defined($class->{_origin_name})) {
        return $file_id;
    }

    # No support for 7-mode at the component layer yet. No cdefs
    my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
    my $command = "flexcache delegations $vol_name";
    my $output  = $api_set->execute_command(command => $command);

    if ($output =~ /\s$file_id($|\n|\s)/) {
        $rc = FCACHE_RC_SUCCESS;
    }

    fcache_verify_expected_rc(
        obtained_rc => $rc,
        expected_rc => $args{expected},
        action      => "Find flexcache delegation for file $file_name",
        output      => $output,
    );

    return $file_id;
} ## end sub _7Mode_flexcache_delegation_verify

# If this is a 7-mode volume, set the exportfs options.
# Arguments:
# - options: Optional: Scalar: Exportfs options to set.
#            Default: FCACHE_DFLT_EXPORTFS_ORIGIN.
# Return: FCACHE_RC_SUCCESS.
sub _exportfs_if_7mode {
    my $class = shift;
    my %args  = validate(
        @_,
        {   options =>
                {type => SCALAR, default => FCACHE_DFLT_EXPORTFS_ORIGIN,},
        }
    );

    my $options = $args{options};
    my $name    = $class->{_name};
    my $vfiler  = $class->{_vserver_name};

    if ($class->{_nacl_node}->mode() eq FCACHE_ONTAP_7MODE) {

# The NACL::C::Volume is also executing the exportfs (only if apiset_should is set to CLI interface),
# but, unfortunatelly, with big limitations: it does not allow for any options
# (the only option that it may blindly set is -p)
# So use APISet directly
        my $api_set = $class->{_nacl_node}->get_7m_or_nodescope_apiset();
        my $junction_path = $class->get_junction_path();

        if ($vfiler eq _FCACHE_VSERVER_NO_VFILER) {

            # Not in a multistore environment, execute regular exportfs
            fcache_log_comment(
                message => "SETUP: Configuring exportfs options for $name");
            $api_set->exportfs(
                path                 => $junction_path,
                persistent           => 1,
                'persistent-options' => $options,
            );
        } else {
            fcache_log_comment(message =>
                    "SETUP: Adding volume $name as a resource to vfiler $vfiler"
            );
            NACL::C::Vfiler->add(
                command_interface => $class->{_nacl_node},
                vfiler            => $vfiler,
                'storage-units'   => [$junction_path,],
            );
            fcache_log_comment(message =>
                    "SETUP: Configuring vfiler exportfs options for $name");
            my $exportfs_cmd = "exportfs -p $options $junction_path";
            $api_set->vfiler_run(
                vfiler  => $vfiler,
                command => $exportfs_cmd,
            );
        }
    }

    return FCACHE_RC_SUCCESS;
} ## end sub _exportfs_if_7mode

# Obtain the name of the share that was created automatically at setup.
# Note that, if the volume was configured for NFS,
# then this method will return undef (no share was created).
# Arguments: none.
# Return: Name of the share (Scalar) or undef.
sub _get_share_name {
    my $class = shift;
    my %args = validate(@_, {});

    return $class->{__share_name};
} ## end sub _get_share_name

# Method to check if an associated share was created for this object.
# To be used in the class constructor.
# It may also be used while creating a cache: if the origin was configured for NFS
# (no share was created at setup), but the cache must be configured for CIFS,
# then the cache constructor will call this method (on the origin), to create one.
# Arguments:
# - reuse_action: Optional: Scalar: What action to take for this share
# Return: Name of the object-associated share (Scalar) or undef.
sub _ensure_volume_has_share {
    my $class = shift;
    my %args  = validate(
        @_,
        {   reuse_action => {
                type    => SCALAR,
                regex   => qr/attach|die|reuse|purge/,
                default => 'die',
            }
        }
    );

    if (!defined($class->{__share_name})) {
        $class->{__share_name}
            = $class->cifs_share_create(_reuse_action => $args{reuse_action},
            );
    }

    return $class->{__share_name};
} ## end sub _ensure_volume_has_share

# Get the _recycle member value
sub _get_recycle {
    my $class = shift;

    return $class->{_recycle};
}

# Overwrite the _recycle member value
# Useful for actions that imply big changes in the volume's state,
# when we need to re-attach to the volume during the script execution.
# E.g. vol move
sub _set_recycle {
    my $class       = shift;
    my $new_recycle = shift;
    $class->{_recycle} = $new_recycle;
}

# Overwrite the _perform_cleanup member value
# Useful for actions that imply big changes in the volume's state,
# when we need to re-attach to the volume during the script execution.
# E.g. vol move
sub _set_perform_cleanup {
    my $class     = shift;
    my $new_value = shift;
    $class->{_perform_cleanup} = $new_value;

    return FCACHE_RC_SUCCESS;
} ## end sub _set_perform_cleanup

# Class destructor
# Unless the constructor was called with perform_cleanup set to FALSE,
# it takes care of any cleanup that needs to be done on the volume:
# - the volume is purged, unless recycling was requested to the constructor
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    if (   ($class->{_creator_pid} == $$)
        && ($class->{_perform_cleanup})
        && (defined($class->{_nacl_vol}))) {

        # Flexcache options still need to be reset, even if the volume
        # was/will be purged
        my $cleanup_flexcache_options = $class->{_cleanup_flexcache_options};
        if (scalar(keys(%{$cleanup_flexcache_options})) != 0) {
            fcache_log_comment(
                message => "CLEANUP: Resetting the volume flexcache options");
            $class->flexcache_options_modify(%{$cleanup_flexcache_options});
        }

        if ($class->{_cleanup_av_scan_dir_command}) {
            NACL::STask::AntivirusOnDemandCommandScanDir->purge(
                command_interface  => $class->{_nacl_node},
                name               => $class->{_cleanup_av_scan_dir_command},
                vserver            => $class->{_vserver_name},
                nacltask_if_exists => 'pass',
            );
        }

        if ($class->{_is_purged}) {
            fcache_log_comment(message =>
                    "CLEANUP: Volume was already purged by the script: "
                    . $class->{_name});
        } elsif ($class->{_recycle}) {
            fcache_log_comment(message =>
                    "CLEANUP: Recycling requested. Leaving behind the volume "
                    . $class->{_name});
        } else {
            fcache_log_comment(
                message => "CLEANUP: Purging the volume " . $class->{_name});
            if (($class->{_boot_mode} eq FCACHE_ONTAP_CMODE)
                && !(defined($class->{_origin_name}))) {
                $class->get_cache_volumes(_purge_if_found => TRUE);
            }
            $class->{_nacl_vol}->purge(nacltask_if_purged => 'pass');
        }
    }

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

