#
# $Id: //depot/prod/test/Rsierranevada/lib/FlexCache/Vserver.pm#1 $
#

#
# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Agnostic module for vserver-level functionality.
## @author voicu@netapp.com, dl-flexcache-qa
## @status Private
## @pod here

=head1 NAME

FlexCache::Vserver

=head1 DESCRIPTION

This module provides release and mode agnostic vserver/vfiler
level functionality for all the flexcache scripts.
Its functionality is made available through the 'vserver'
member of the FlexCache::Origin object.
The module is release agnostic and provides
functionality for both 7-mode and C-mode.
It is NACL-compliant. To be noted, though,
that no direct access to any NACL object is given.
This is to ensure the encapsulation that an object-oriented design must follow
and to not break the cleanup mechanism provided by the FlexCache library.

=head1 SYNOPSIS

This module is intended to be used only through the FlexCache::Origin objects,
and/or through the FlexCache::Cache objects, as follows:

	BEGIN { require "flexcache/fcache_common_params.pl"; }
	
	use strict;
	use warnings;
	
	use FlexCache::Origin;
        use FlexCache::Cache;
	
	my $origin = FlexCache::Origin->new(node => {name => $FILER_SRC,},
					    aggregate => {name => $AGGR_SRC,
							  block_type => $AGGR_SRC_BLOCK_TYPE,},
					    volume => {name => $VOL_SRC,},
					    vserver => {name => $VSERVER,
							dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
					    client => {name => $CLIENT_SRC,},);
	$origin->{vserver}->call_any_method_but_the_constructor();
        
        my $cache = FlexCache::Cache->new(node => {name => $FILER_DST,},
                                          aggregate => {name => $AGGR_DST,
                                                        block_type => $AGGR_DST_BLOCK_TYPE,},
                                          volume => {name => $VOL_DST,},
                                          vserver => {name => $VSERVER,
                                                      dlif_name => $DLIF_DST,
                                                      dlif_id => $DLIF_DST_ID,},
                                          client => {name => $CLIENT_DST,},
                                          origin => $origin);
        $cache->{vserver}->call_any_method_but_the_constructor();
	
Using this module independently is possible,
with the risk that the constructor's arguments may change at any time.

=head1 UNIT TESTING

Unit tests for this module:

        flexcache/library_unit_tests/Origin.thpl
        flexcache/library_unit_tests/Cache.thpl

=cut

package FlexCache::Vserver;

# Standard libs
use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR OBJECT HASHREF ARRAYREF UNDEF BOOLEAN);
use Data::Dumper;

# NetApp libs
use FlexCache::Constants qw(:all);
use FlexCache::Utils qw(fcache_log_comment fcache_log_result);
use FlexCache::CMode::Vserver;
use FlexCache::7Mode::Vserver;

=head1 CONSTRUCTOR

=head2 new

	Constructor for the FlexCache::Vserver class.
	For C-mode, an object represents a vserver.
	For 7-mode in a multistore environment, an object represents a vfiler.
	The constructor will also take care of the creation of the vserver and
	of all the steps needed for flexcache testing, as follows:
	- C-mode: vserver is created/reused, nfs/cifs configuration is done.
	- 7-mode, multistore environment: create the vfiler (including its root volume),
	          set flex-cache specific options on the vfiler, configure the vfiler's IP.
	- 7-mode, not in a multistore environment: no filer configuration is changed.

=head3 ARGUMENTS

=over

=item B<name> I<Required> Scalar|Undef:

	Name of the vserver.
	If the name is undefined for C-mode, it will default to
	FCACHE_VSERVER (defined in FlexCache::Constants).
	For 7-mode: if the name is undefined, the script will be run in a
	"non multistore" environment (default vfiler); if the name is defined,
	then the script will be run in a multistore environment
	(the vfiler <name> will be created and the test volume
	will be added as a resource to this vfiler).

=item B<dlif_name> I<Optional> Scalar: 

	Name of the data lif.
	This argument is required for C-mode and it is ignored in 7-mode.
	
=item B<dlif_id> I<Optional> Scalar: 

	Identifier for the vserver/vfiler information as found in the tharnhost file.
	For C-mode:
	- it is the count of the data interface as listed in the tharnhost file
	(e.g. setting this attribute to 1 means that the data interface that is listed as 1
	in the tharnhost file of the node is to be used
	(DATA_IP1/DATA_PORT1/DATA_MASK1/etc)).
	For 7-mode:
	- if script is not required to be run in a multistore environment,
	then this argument is ignored.
	- if script is required to be run in a multistore envirnment, then
	this atribute represents the count of the vfiler as listed in the tharnhost file
	(e.g. setting this attribute to 1 means that the vfiler used for testing is listed as 1
	in the tharnhost file of the node and it is expected that this tharnhost file contains the following information: 
	(VFILER1 (should be the same as the name argument), VFILER_IP1, VFILER_MASK1, VFILER_PORT1).
	
=item B<recycle> I<Optional> Boolean:

        Flag to determine if this aggregate should be reused.
        If this flag is set to TRUE, then:
        - at setup the vserver/vfiler will be either created or reused (if it already exists)
        - at cleanup it will not be destroyed.
        If the flag is set to FALSE, then:
        - at setup the vserver/vfiler will be created or purged+recreated (if it already exists)
        - at cleanup it will be purged.
        Default value:
	- C-mode: constant FCACHE_DFLT_VSERVER_RECYCLE (defined in FlexCache constants).
	- 7-mode: constant FCACHE_DFLT_VFILER_RECYCLE (defined in FlexCache constants).

=back

=head3 RETURN

FlexCache::Vserver object.

=head3 EXAMPLE

	# Create a flexcache origin with the default settings for the vserver member:
	my $origin = FlexCache::Origin->new(vserver => {name => $VSERVER,
	                                                dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,},
                                            ....);
	# Create a flexcache origin, requiring a non-recycled vserver/vfiler:
	my $origin = FlexCache::Origin->new(vserver => {name => $VSERVER,
	                                                dlif_name => $DLIF_SRC,
							dlif_id => $DLIF_SRC_ID,
							recycle => FALSE,},
                                            ....);
	# All of the above examples apply to the FlexCache::Cache constructor as well.

=cut

# "Somehow internal" arguments (to limit the use of this constructor to the FlexCache library.
#                               These arguments are not documented with POD, to not confuse the script developers):
# - node: Required: Object: NACL::STask::Node object for this node.
# - aggregate_name: Required: Scalar: Name of the aggregate to contain the root volume of the vserver/vfiler.
# - client_type: Required: Scalar: Type of the client for which the vserver needs to be ready.
#   This will determine if nfs or cifs setup needs to be executed.
# - debug: Optional: Boolean: Flag to increase the verbosity of this module, useful for debugging purposes.
#   Default value: FALSE.
# - perform_cleanup: Optional: Boolean: Flag to control if the cleanup should be executed.
#   Default value: TRUE.
# - ui_messages: Required: Object: FlexCache::UI::OntapMessages, to be used for any ONTAP output validation.
# - config_type: Optional: Scalar: Type of configuration for which to execute any setup (origin/cache).
#   This argument is ignored for C-mode.
#   Default: origin.
# - attach_only: Optional: Boolean: Flag to determine if the constructor should only attach to the required vserver/vfiler,
#                and not attempt to create/recreate. If this flag is set to TRUE, then the recycle flag is ignored.
#                Default value: FALSE.
sub new {
    my $class = shift;
    my %args  = validate(
        @_,
        {   name            => {type => SCALAR | UNDEF,},
            node            => {type => OBJECT, isa => "NACL::STask::Node",},
            dlif_name       => {type => SCALAR | UNDEF, optional => 1,},
            dlif_id         => {type => SCALAR | UNDEF, optional => 1,},
            recycle         => {type => BOOLEAN, optional => 1,},
            aggregate_name  => {type => SCALAR,},
            client_type     => {type => SCALAR,},
            debug           => {type => BOOLEAN, default => FALSE,},
            perform_cleanup => {type => BOOLEAN, default => TRUE,},
            attach_only     => {type => BOOLEAN, default => FALSE,},
            ui_messages =>
                {type => OBJECT, isa => "FlexCache::UI::OntapMessages"},
            config_type => {type => SCALAR | UNDEF, optional => 1,},
        }
    );

    my $self = {};
    my $mode = $args{node}->mode();

    if ($mode eq FCACHE_ONTAP_CMODE) {
        $self->{_vserver} = FlexCache::CMode::Vserver->new(%args);
    } else {
        $self->{_vserver} = FlexCache::7Mode::Vserver->new(%args);
    }

    bless($self, $class);

    return $self;
} ## end sub new

=head1 OTHER METHODS

	Please see the detailed documentation for each mode-specific Vserver module:
	FlexCache::CMode::Vserver
	FlexCache::7Mode::Vserver

=cut

########################################################
# Internal methods
########################################################

# For all the other methods except for constuctor/destructor,
# call the mode-specific method in XMode::Vserver
sub AUTOLOAD {
    our $AUTOLOAD;
    my $class  = shift;
    my $method = $AUTOLOAD;

    # Obtain the name of the method to call (remove package name)
    $method =~ s/.*\:\://;

    return $class->{_vserver}->$method(@_);
} ## end sub AUTOLOAD

# Class destructor
# Please see the documentation for the XMode::Vserver destructor.
sub DESTROY {
    my $class = shift;

    # Need to remember any script error, due to a Perl bug
    # where the destructors lose the value of $@ (fixed only in perl 5.14).
    # More details at:
    # https://metacpan.org/module/perl5140delta#Exception-Handling
    my $script_error = $@;

    $class->{_vserver} = undef;

    if ($script_error) {
        $@ = $script_error;
    }
} ## end sub DESTROY

1;

