# $Id: //depot/prod/test/Rsierranevada/lib/ForeachHost.pm#1 $

## @summary Helpers to handle multiple hosts in parallel (in prejobs)
## @author dl-setupcleanup-libs@netapp.com
## @status Shared
## @pod here

package ForeachHost;

use strict;
use warnings;
use Tharn qw(param hosts subtest logresult script_result config_result);
use Params::Validate qw(validate SCALAR CODEREF ARRAYREF BOOLEAN);
use Carp qw(confess);
BEGIN
{
    Params::Validate::validation_options(
                                         on_fail => sub {
                                             my @msgs = @_; chomp(@msgs);
                                             die script_result([@msgs]);
                                         },
                                        );
}

require Exporter;
our @ISA=qw(Exporter);
our @EXPORT_OK=qw(foreach_filer foreach_client foreach_host);

=head1 NAME

ForeachHost - Call a function in parallel for each host (filer, client, etc)

=head1 SYNOPSIS

    use ForeachHost qw(foreach_filer foreach_client foreach_host);

    foreach_filer(scriptspec => $coderef_or_script,
                  runid_prefix => 'mytest',
                  subtest_args => [],
                  wait => 1,
                  allow_direct_call => 1,
                  );

    foreach_client(scriptspec => $coderef_or_script,
                   runid_prefix => 'mytest',
                   subtest_args => [],
                   wait => 1,
                   allow_direct_call => 1,
                   );

    foreach_host(scriptspec => $coderef_or_script,
                 runid_prefix => 'mytest',
                 host_params => ["FILERA","FILERB","FILERC",...],
                 child_prefix => "FILER",
                 subtest_args => [],
                 wait => 1,
                 allow_direct_call => 1,
                 );

=head1 DESCRIPTION

C<foreach_filer> and C<foreach_client> (and a generalized subset,
C<foreach_host>) create subtests (child processes) to invoke the given
scriptspec (test script filename or perl subroutine) in parallel, with
one subtest for every filer/client/host.

Because of the global way that C<foreach_filer> and C<foreach_client>
take as input which filers are involved, this library is more useful
for construction of test scripts that from the very top need to split
off in parallel, rather than as a general wrapper around running
background subtests (for which one can just call subtest(-bg)
directly).

For C<foreach_filer> and C<foreach_client> it is assumed that there is
a parameter that is passed to the tests which contains a list of all
of the parameters which contain hosts or lists of hosts.  For filers,
this parameter is C<FILER_LIST> and for clients this parameter is
C<CLIENT_PARAM_LIST>.  (not C<CLIENT_LIST> because too many existing
tests think of that as a list of hosts rather than of parameters of
hosts; not C<FILER_PARAM_LIST>, because C<FILER_LIST> is already too
well established for this purpose).

Each of the parameters in C<FILER_LIST> (C<foreach_filer>) or
C<CLIENT_PARAM_LIST> (C<foreach_client>) or passed in the
C<host_params> option (C<foreach_host>) can each contain a single
hostspec (as is standard for parameters like C<FILER>, C<FILERA>,
C<FILERB>, C<CLIENT1>, C<CLIENT2>, etc.) or can each contain a
comma-separated or colon-separated list of hostspecs (as is standard
for parameters like C<NODES> or C<FILERS> or C<CLIENTS>).

For each such hostspec, the foreach routine will start a subtest,
passing the host id of the host in a particular parameter which we
will call the "child prefix" (for C<foreach_filer> this is C<FILER> and
for C<foreach_client> this is C<CLIENT>).  Also, every parameter
beginning with the name of the parameter that contained the hostspec
is passed to the subtest but renamed to begin with this child prefix.

=over

For example, if C<FILER_LIST> contains C<FILERA,FILERB,FILERC> then
C<foreach_filer> will start a subtest for the host in C<FILERA>,
passing the id of that host in C<FILER> and passing every parameter
that begins with C<FILERA_something> as C<FILER_something>, and will
start a subtest for the host in C<FILERB>, passing the id of that host
in C<FILER> and passing every parameter that begins with
C<FILERB_something> as C<FILER_something>, and so on.

Thus, the user invoking the subtest that uses C<foreach_filer> can
pass parameters meant just for the host in C<FILERA> under the name
C<FILERA_>I<something>, parameters meant just for the host in
C<FILERB> using the name C<FILERB_>I<something>, or parameters meant
for all hosts under the name C<FILER_>I<something>.  The C<scriptspec>
itself can just always expect the host id to be in test parameter
C<FILER> and related parameters will be in test parameters named
C<FILER_>I<something>.

As another example, if C<FILER_LIST> contains C<NODES>, then
C<foreach_filer> will start a subtest for each of the hosts in
C<NODES>, passing the id of each host in C<FILER> and passing every
parameter that begins with C<NODES_something> as C<FILER_something>.
(If one uses parameters like NODES that contain multiple hosts, then
there is no way to pass additional parameters with separate values for
each host)

=back

If one of the hostspecs has a C<part_host> field (presumably a filer,
not any other kind of host), meaning it is a member of a CFO/SFO pair,
and if another hostspec in the list has the same name as C<part_host>,
then its hostid is passed in the parameter C<PARTNER>, and other
parameters that begin with the name of the parameter that contained
the partner hostspec are renamed to begin with C<PARTNER>.  If no
other hostspec is found corresponding to C<part_host>, a warning is
logged.

=over

Thus, the child scriptspec that operates on a CFO pair would be
written to take parameters beginning with C<FILER> and C<PARTNER>.
The scriptspec has to expect that it is also being run in parallel
with the mirror image situation, being called in parallel with the
partner as C<FILER> and this filer as C<PARTNER>.

=back

The following text is recommended when documenting a test which uses
C<forach_filer> at the very top, where the function passed to
C<filer_foreach> does all of that test's "real" work in parallel
(replace "mangled" with something that describes what the test is
actually supposed to do to those filers):

        @param FILER_LIST Required: A comma separated list of
        parameter names, such as "FILER" or "FILERA,FILERB" or "NODES".
        All of the filers listed in parameters by those names will be
        mangled in parallel.  Also, parameters shown below as beginning
        with "FILER" may instead begin with one of these parameter names,
        in which case it will apply only to that filer.

Or similarly for a test that is basically a call to C<foreach_client>:

        @param CLIENT_PARAM_LIST Optional: A comma separated list of
        parameter names, such as "CLIENT" or "CLIENT1,CLIENT2" or
        "CLIENTS".  All of the clients listed in parameters by those
        names will be mangled in parallel.  Also, parameters shown
        below as beginning with "CLIENT" may instead begin with one of
        these parameter names, in which case it will apply only to
        that client.  Defaults to "CLIENT1" (if "CLIENT1" has a value
        but "CLIENT2" does not), or "CLIENT1,CLIENT2" (if "CLIENT1"
        and "CLIENT2" have values, but "CLIENT3" does not) and so
        on, or else paramter "CLIENTS" if that has a value, or else
        parameter "CLIENT" if that has a value.

=head1 OPTIONS

The same general options apply to all of C<foreach_filer>,
C<foreach_client>, and C<foreach_host>:

=over

=item scriptspec => $coderef_or_script

A code reference to invoke once for each filer and in
parallel (anything that can be passed as the first argument to
C<Tharn::subtest>).

=item runid_prefix => $prefix

A string to use in the "runid" (and logfile name) of the child test
process.  The runid of the subtests will be this value, plus "-", plus
the id of the host that this subtest is for.

=item subtest_args => $arrayref

Additional arguments to pass to the subtest call.

=item wait => $boolean

If true (the default) then wait for the completion of the subtests
that have been launched.

=item allow_direct_call => $boolean

If true (the default), and if the circumstances are right (if the
scriptspec is a function, and there is only one host to work on, and
no renaming of parameters for the child would need to be done, and
"wait" is 1) then just call the function directly rather than starting
and waiting for a single subtest which would do the same thing.  This
assumes that the arguments in C<subtest_args> aren't needed in this
case.

=item child_prefix => $string

The name of the parameter as which to pass the host id to each
subtest, and the prefix to which to rename the other parameters for
that host.  Defaults to "FILER" for C<foreach_filer>, "CLIENT" for
C<foreach_client>, and has no default for C<foreach_host>.

=item host_params => $arrayref

The names of all of the parameters containing hostspecs.  Only an
option for C<foreach_host>.

For C<foreach_filer>, the values are the parameters listed in
C<FILER_LIST>, which must have a value.

For C<foreach_client>, the values are the names in parameter
C<CLIENT_PARAM_LIST>, which defaults to "CLIENT1" (if C<CLIENT1> has a
value but C<CLIENT2> does not), or "CLIENT1,CLIENT2" (if C<CLIENT1>
and C<CLIENT2> have values, but C<CLIENT3> does not) and so on, or
else paramter C<CLIENTS> if that has a value, or else parameter
C<CLIENT> if that has a value.


=head1 EXAMPLE

This:

    foreach_filer(scriptspec => \&f, 
                  runid_prefix => "example",
                  subtest_args => ['-timeout',3600]);

is equivalent to the following if FILER_LIST is "FILERA,FILERB", and
if FILERA and FILERB have each other's names listed in each others'
"part_host" field:

    subtest(\&f, '-bg',
            '-runid', "example-$FILERA",
            '-mapparam', 's/^FILERA/FILER/',
            '-mapparam', 's/^FILERB/PARTNER/',
            '-timeout',3600);
    subtest(\&f, '-bg',
            '-runid', "example-$FILERB",
            '-mapparam', 's/^FILERB/FILER/',
            '-mapparam', 's/^FILERA/PARTNER/',
            '-timeout',3600);
    Subtest::wait_finish();

and equivalent to the following if FILER_LIST is "FILER":

    f();

The sample unit testscripts are available in Examples directory,namely

   Example/setup_host.thpl
   Example/setup_filer.thpl
   Example/setup_client.thpl

=cut

sub foreach_filer {
   my (%opts) = validate(@_, {
                               scriptspec => { type => CODEREF|SCALAR },
                               runid_prefix => { type => SCALAR },
                               wait => { type => BOOLEAN, optional=>1 },
                               subtest_args => { type => ARRAYREF, default=>[] }
                         });
    # ntest should provide a value for FILER_LIST, so require it rather than construct a default
    my $filer_list = param("FILER_LIST",'-mesg','Comma separated list of host prefixes');
    my @host_params = split(/\s*,\s*/,$filer_list);
    push @{$opts{subtest_args}}, "FILER_LIST=FILER";
    $opts{scriptspec} ||= delete $opts{scriptspec};
    $opts{child_prefix} ||= "FILER";
    $opts{host_params} ||= [@host_params];
    &foreach_host(%opts);
}

sub foreach_client {
    my (%opts) = validate(@_, {
                               scriptspec => { type => CODEREF|SCALAR },
                               runid_prefix => { type => SCALAR },
                               wait => { type => BOOLEAN, optional=>1 },
                               subtest_args => { type => ARRAYREF, default=>[] }
                          });
    my @host_params;
    if (param("CLIENT_PARAM_LIST")) {
        @host_params = split(/\s*,\s*/,param("CLIENT_PARAM_LIST"));
    } elsif (param("CLIENT1")) {
        for (my $suffix=1; param("CLIENT$suffix"); $suffix++) {
            push @host_params,"CLIENT$suffix";
        }
    } elsif (param("CLIENTS")) {
        @host_params=("CLIENTS");
    } elsif (param("CLIENT")) {
        @host_params=("CLIENT");
    } else {
        die config_result("Cannot guess list of clients for this test (CLIENT_PARAM_LIST, CLIENT1, CLIENTS, CLIENT not defined)\n");
    }
    $opts{subtest_args} ||= [];
    push @{$opts{subtest_args}}, "CLIENT_PARAM_LIST=CLIENT";
    $opts{child_prefix} ||= "CLIENT";
    $opts{host_params} ||= [@host_params];
    &foreach_host(%opts);
}

sub foreach_host {
    my (%opts) = validate(@_, {
                               scriptspec => { type => CODEREF|SCALAR },
                               runid_prefix => { type => SCALAR },
                               subtest_args => { type => ARRAYREF, default=>[] },
                               child_prefix => { type => SCALAR },
                               host_params => { type => ARRAYREF },
                               wait => { type => BOOLEAN, optional=>1 },
                               allow_direct_call => { type => BOOLEAN, default=>undef },
                              });
    my $scriptspec = $opts{scriptspec};
    my $runid_prefix = $opts{runid_prefix};
    my @subtest_args = @{$opts{subtest_args}};
    my $child_prefix = $opts{child_prefix};
    my $wait = exists($opts{wait}) ? $opts{wait} : 1;
    my $allow_direct_call = exists($opts{allow_direct_call}) ? $opts{allow_direct_call} : 1;
    my $host_params = $opts{host_params};
    my @subtests;

    # build up a list of hostids and the prefix from where they came
    my @host_list;
    foreach my $prefix (@$host_params) {
        my $hostspec = param($prefix,'-mesg',undef);
        foreach my $host_obj (hosts($hostspec)) {
            push(@host_list,[$host_obj,$prefix]);
        }
    }

    if (
        ($allow_direct_call) &&
        (@host_list == 1) &&
        (ref($scriptspec) && ref($scriptspec) eq "CODE") && 
        ($host_list[0][1] eq $child_prefix) &&
        ($wait)
       ) {
        # only one host, and the subtest to call is a function, and
        # the parameter prefix that the subtest expects is the
        # parameter prefix used to pass this host to us-the-parent (no
        # mapparam needed), and caller has us wait for the function to
        # complete, and so we can just call the
        # code-that-would-be-subtested directly.
        $scriptspec->();
        return;
    }

    # Launch the scriptspec in background for all the hosts
    foreach my $href (@host_list) {
        my ($host_obj, $prefix) = @$href;
        my $host_id = $host_obj->id();
        my $host_type = $host_obj->hosttype();
        my @subtest_opts = ('-mapparam',sub { s/^${prefix}\B/${child_prefix}/ });
        if ($host_type =~ /filer/ || $host_type =~ /sim/) {
           push @subtest_opts, ("${child_prefix}=$host_id;hosttype=$host_type");
        } else {
           push @subtest_opts, ("${child_prefix}=$host_id");
        }

        my $partner_host = $host_obj->part_host();
        if ($partner_host) {
            my $found;
            # Search for partner in our list of hosts, and pass it in
            # under the prefix "PARTNER" if we can
            foreach my $pref (@host_list) {
                my ($partner_obj, $partner_prefix) = @$pref;
                my $partner_id = $partner_obj->id();
                if ($partner_id eq $partner_host) { 
                    $found=1;
                    push @subtest_opts, ("PARTNER=$partner_id");
                    push @subtest_opts, ('-mapparam', sub { s/^${partner_prefix}\B/PARTNER/ });
                    last;
                }
            }
            if (!$found) {
                logresult('WARN',"Host $host_id ($prefix) has partner '$partner_host', but '$partner_host' not found in host parameters (".join(",",@$host_params).")");
            }
        }
        push(@subtests,&subtest($scriptspec,
                                '-bg',
                                '-runid',$runid_prefix . "-" . $host_id,
                                @subtest_opts,
                                @subtest_args
             ));

    }
    if ($wait) {
        Subtest::wait_finish(subtests=>\@subtests);
    }
}

1;

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-setupcleanup-libs@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/ForeachHost.pm#1 $
    Last revised $Date: 2013/01/22 $ 

=head1 COPYRIGHT

    Copyright 2005,2008, Network Appliance Inc., All Rights Reserved.

=cut
