# -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE.pm#1 $
#
## @summary NOTE Netapp Objects for Test Engineering
##
## @author dl-dpgqa-automation@netapp.com
## @status Private (for now)
## @pod here

=head1 NAME

NOTE - A thin object wrapper around the test library

=head1 SYNOPSIS

If you find writing test scripts tedious, if you think they
require much too much effort to code and debug, if you don't
actually, really, like to code very much and prefer to
get the job done with as little effort as possible, then this
package is for you.

This library package contains object classes that models the various objects
used to test ONTAP such as volume, aggregate, qtree, snapmirror etc. Each object class
for an ONTAP object, can be instantiated in one of two modes, 7-mode or C(luster)-mode.
A
L<C<7G>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-7G.html>
object represents an object in 7-mode and a
L<C<GX>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX.html>
object in C-mode. If the underlying
object exists only in 7-mode than only a 7G object can be instantiated.
In addition, modules modeling objects outside of ONTAP can be found in the main
L<C<NOTE>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE.html>
directory.

The hierarchy of the library consists of the base classes located in the
L<C<NOTE>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE.html>
directory together with two sub-directories,
L<C<7G>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-7G.html>
and
L<C<GX>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX.html>
, which contains the corresponding 7-mode and C-mode methods. So for a C-mode volume object,
the methods for the C-mode specific behavior can be found in
L<C<NOTE::GX::Volume>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html>
with methods that are common to both 7-mode and C-mode in 
L<C<NOTE::Volume>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html>

For objects that are independent of 7-mode or C-mode, such as
L<C<NOTE::Client>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Client-pm.html>
, a single module exists in the main
L<C<NOTE>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE.html>
directory.

=head1 DESCRIPTION

This is the base class for the NOTE package, which is intended
to be a very thin object layer around the test library. Its purpose
is to make it possible to write test code using objects without
have a full-blown object model for the filer.

=head1 METHODS

=cut

# Base class for Netapp Objects for Test Engineering
package NOTE;

use Tharn;
use Carp qw(croak confess);

BEGIN {
use GNOTE::Log;
    $NOTE::TRACELOG            = NOTE::Log->new(); # sets up trace log file
    $NOTE::TRACELOG->init_log();

}
use GNOTE::Client;
use GNOTE::Qtree;
use GNOTE::FileSystem;
use GNOTE::FileSystem::NFS;
use GNOTE::FileSystem::NFS::File;
use GNOTE::Printf;
use GNOTE::GX;
use GNOTE::7G;

use GNOTE::Client::RPC;

use Utils::Units qw();
use Getopt::Std;
use IO::File;
use Data::Random qw( rand_words );

#our $TRACELOG;            # used for internal logging

our $_DONT_SPARE_VOLUMES = 0;								      # destroy volume all the time if NOTE_DONT_SPARE_VOLUMES is set
our $_SPARE_ALL          = 0;                                                                 # If set to 1, destructors do nothing to filer
our $_SPARE_ALL_ON_ERROR = 0;                                                                 # If set to 1, spare all if error encountered
our $_CONSUMED_RESOURCES = 0;
our $_PANICKED           = 0;
our $USAGE_ANALYSIS      = defined $ENV{'NOTE_STATS'} ? $ENV{'NOTE_STATS'} : 0;
our $_MEM_STATS          = defined $ENV{'NOTE_MEMSTAT'} ? $ENV{'NOTE_MEMSTAT'} : 0;
our $_SAVE_LOGS          = defined $ENV{'NOTE_SAVE_LOGS'} ? $ENV{'NOTE_SAVE_LOGS'} : undef;
our $LOG                 = defined $ENV{'NOTE_LOG'} ? $ENV{'NOTE_LOG'} : undef;
our $LOGLEV              = undef;
our $LOGFILE             = undef;
our $Trace               = TracePkg->new();
our $_RPC_CLIENTS;									     # hash reference of RPC servers initiated by GNOTE::RPC::Handler;

# This creates an object representing the tied handle which
# is now the default file handle for print, printf, and write.
# This the handlers for this tied hash redirect the output
# data to NOTE::logcomment. When the object is destroyed,
# prints are restored to STDOUT.

if ( defined $LOG ) {
    if ( $LOG =~ /^(\d+)$/ ) {
        $LOGLEV  = $1;
        $LOGFILE = "/dev/null";
    }
    elsif ( $LOG =~ /^(\d+),(.+)$/ ) {
        $LOGLEV  = $1;
        $LOGFILE = $2;
    }
    elsif ( $LOG =~ /^(.+)$/ ) {
        $LOGLEV  = '0';
        $LOGFILE = $1;
    }
    else {
        die "NOTE_LOG ($LOG) is incomprehensible\n";
    }
}
our $stdout = NOTE::Printf->new->start;
our $STDOUT = $stdout->raw;

my $init            = 0;
my $teardown_called = 0;

sub init {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $NOTE::init = $$;
    $teardown_called = 0;
}

# This BEGIN and END block make it possible to write
# a NOTE script without explicitly calling NOTE::teardown.
# Care is taken to make sure that teardown is called only
# once and that it is not executed in the context of a
# subtest (where $$ != $NOTE::init).
BEGIN {
    NOTE::init;
}

END {
    return unless ( $$ == $NOTE::init );
    eval { NOTE::teardown( exitcode => $? ) };
    NOTE::logcomment("Errors during teardown: $@") if ($@);
    $NOTE::TRACELOG->log_call_count();
    $NOTE::TRACELOG->close_log();
    $NOTE::TRACELOG->write_csv_log();
    $NOTE::TRACELOG->write_central_logs();
}

use strict;

=head2 ::is_GX

Checks the arguments for clues to determine if this
is a GX style call or a 7G style call.
 Returns TRUE if:
   1) the Node argument exists or
   2) the Vserver argument exists or
   3) the Filer argument exists and is in a NOTE::GX class or
   4) the VSERVER option is defined
 Otherwise, returns FALSE.

=over

=item Arguments

B<args> => I<hashref of arguments>

HASHREF to list of arguments to determine if a GX object is
being requested.

=item Examples

     my %args = { Vserver => "vso" };
     my $gx_flag = GNOTE::is_GX(\%args);

=back

=cut

sub is_GX {
    my (%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return 1
      if exists $args{Filer}
          && ref( $args{Filer} ) =~ /^NOTE::GX/;
    return 1
      if NOTE::option(q/$VSERVER/)
          || NOTE::option(q/$VSERVER_SRC/)
          || NOTE::option(q/$VSERVER_DST/);
    return 1 if exists $args{Cluster};
    return 1 if exists $args{Vserver};
    if ( exists $args{Aggregate} && ref $args{Aggregate} ) {

        # Check whether this is (or will be) a GX Aggregate
        return 1 if ( ref( $args{Aggregate} ) =~ /^NOTE::GX/ );
        return 1 if exists $args{Aggregate}->{type};            # 'striped' or 'flex'
        return 1 if exists $args{Aggregate}->{volume_style};    # 'striped' or 'flex'
    }
    return 1 if exists $args{Node};
    return 1 if exists $args{Nodes};
    return 1 if exists $args{nodes};

    return 0;
}

our %_wordHits = ();

sub _newWord {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my @r;
    my $len = shift || 8;

    my $retry = 0;
    while (1) {
        @r = rand_words( size => 32 );
        for (@r) {
            next unless ( length($_) == $len );
            next if ( $_ =~ /(-|')/ );
            next if ( $_ =~ /shared/i );    # Reserved word for GenericMount
            next if ( $_ =~ /assert/i );    # smokes complains if it sees this
            next if ( $_wordHits{$_} );
            $_wordHits{$_} = 1;
            return $_;
        }
        if ( ++$retry > 29 ) {
            NOTE::logcomment("Resetting newName after $retry failed attempts");
            %_wordHits = undef;
        }
    }
}

# Reset 'used' words if unable to find new ones after too many retries

sub newName {
    my ( $self, $descriptor ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name;
    $name = sprintf( "%05d_%s_%s", $$, _newWord(4), _newWord(8) );
    $name = $descriptor . $name if ($descriptor);
    return $name;
}

# Given a volume or aggregate name, returns the decoded dataset string
# associated with the container, if any.
sub datasetDecode {
    my ($name) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $name =~ /..(.....)_(....)_(........)_(.*)/ ) {
        return _datasetDecode($4);
    }
    return "";
}

# Given a container name and a dataset string, returns a new name
# with the dataset string encoded into it.
sub datasetEncode {
    my ( $name, $dataset ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $name . "_" . _datasetEncode($dataset);
}

my @base64 = split //, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_+";

sub _datasetEncode {
    my $code = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # unpack into array of 8 bit bytes
    my @bytes = split //, $code;

    # Pad with whitespace to a multiple of 6 bytes
    push @bytes, ' ' for ( 1 .. ( 6 - ( ( ( scalar @bytes ) + 6 ) % 6 ) ) );

    my $bitstr = "";
    for my $byte (@bytes) {
        my $bstr = unpack( "b7", $byte );
        $bitstr .= $bstr;
    }
    my $index = 0;
    my $nextChunk;
    my $bytestr = "";
    my $b6;
    while ( $index < length $bitstr ) {
        $nextChunk = substr( $bitstr, $index, 6 );
        $index += 6;
        $b6 = unpack( "C", pack( "b6", $nextChunk ) ) & 0x3f;
        my $c = $base64[$b6];
        if ( $c eq "_" ) {
            $c = "__";
        }
        elsif ( $c eq "+" ) {
            $c = "_P";
        }
        $bytestr .= $c;
    }
    return $bytestr;
}

sub _datasetDecode {
    my $bytestr = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $bitstr  = "";
    my $b7;
    $bytestr =~ s/__/=/g;
    $bytestr =~ s/_P/+/g;
    $bytestr =~ s/=/_/g;
    my @bytes = split //, $bytestr;
    for my $byte (@bytes) {
        my $pos;
        for ( $pos = 0 ; $pos < @base64 ; $pos++ ) {
            last if ( $byte eq $base64[$pos] );
        }
        my $bstr = unpack( "b6", pack( "C", $pos & 0x3f ) );
        $bitstr .= $bstr;
    }
    my $index = 0;
    my $nextChunk;
    $bytestr = "";
    while ( $index < length $bitstr ) {
        $nextChunk = substr( $bitstr, $index, 7 );
        $index += 7;
        $b7 = unpack( "C", pack( "b7", $nextChunk ) ) & 0x7f;
        $bytestr .= chr($b7);
    }
    $bytestr =~ s/\s*$//;    # Remove trailing whitespace
    return $bytestr;
}

sub _hashOverlay {
    my %args      = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $source    = $args{source};
    my $target    = $args{target} || {};
    my $overwrite = defined $args{overwrite} ? $args{overwrite} : 1;

    if ( $source && ref $target eq 'HASH' ) {
        for my $key ( keys %{$source} ) {
            if ( ref $source->{$key} eq 'HASH' ) {
                $target->{$key} = _hashOverlay(
                    target    => $target->{$key},
                    source    => $source->{$key},
                    overwrite => $overwrite,
                );
            }
            else {
                if ( !defined( $target->{$key} ) || $overwrite ) {
                    $target->{$key} = $source->{$key};
                }
            }
        }
    }
    return $target;
}

my @line2step = ();
my $_initted  = 0;

sub _getSteps {
    my $file = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $curstep;
    my $stepno = undef;
    my @steps;
    my $offset;

    return if $file =~ /^\(/;    # Ignore invalid name; inline NOTE script invoked in subtest mode
    open( F, $file ) or die "Can't open $file";
    while (<F>) {
        if (/^\s*##\s*\@step\s+(.*)/) {
            $curstep = $1;
        }
        elsif ( $curstep && /^\s+##\s*(.*)/ ) {
            $curstep .= " " . $1;
        }
        else {
            if ($curstep) {
                push @steps, $curstep;
                $stepno = $#steps;
                $offset = 0;
            }
            $curstep = undef;
        }
        if ( defined $stepno ) {
            my $n = $.;
            $line2step[$n] = {
                step   => "[" . $steps[$stepno] . "]",
                offset => $offset
            };
            $offset++;
        }
    }
}

# Given a string of the form
#    x,y=1,z=abcde
# will return a reference to a hash:
#    x => undef,  # i.e. exists, but undefined
#    y => 1,
#    z => abcde
sub opts {
    my $string = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts   = ();

    for my $p ( split( /,/, $string ) ) {
        if ( $p =~ /^([^=]+)=(.*)$/ ) {
            $opts{$1} = $2;
        }
        else {
            $opts{$p} = 1;
        }
    }
    return \%opts;
}

=head2 ::option

Uses L<C<Tharn::param>|engineering::nate::releases::latest::html::Tharn/param>
to process command line parameters.

If the argument is of the form
     $param1:$param2:...:$paramN[:default]
returns the first parameter that resolves as defined using Tharn::param
or the last element in the list (if it is not preceeded by a $).
Otherwise returns undef.

=over

=item Arguments

B<param1:...:paramN>

Returns the first parameter that resolves to a value.

=item Example:

    my $filer = NOTE::option('$FILER_DST:$FILER_SRC:$FILER');
    my $type = NOTE::option('$VOL_TYPE:flex');

=back

=cut

sub option {
    my $pstring = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $pstring unless ( defined($pstring) && $pstring =~ /^\$/ );
    my @plist = split( /:/, $pstring );
    my ( $pkey, $pval );
    while ( $pkey = shift @plist ) {
        return $pkey unless ( $pkey =~ /^\$/ );
        $pkey =~ s/^\$//;
        $pval = Tharn::param( $pkey, '-default', undef );
        return $pval if ( defined $pval );
    }
    return undef;
}

my $_lastHeader    = undef;
my $_lastLine      = undef;
my @_lastFrame     = ();
my $_lastStep      = "";
my $_logInitted    = 0;
my $_trace_enabled = 0;
my $_curphase      = "";
my ( $_odev, $_oino );
my ( $_edev, $_eino );

our (
    $opt_b,    # print backtrace
    $opt_s,    # print steps
    $opt_p,    # print test phase
    $opt_r,    # trace resource usage
);

=head2 ::logcomment

Uses L<C<Tharn::logcomment>|engineering::nate::releases::latest::html::Tharn/logcomment>
with additional NOTE information.

=over

=item Example

    NOTE::logcomment("Create Volume");

=back

=cut

sub logcomment (@) {
    my @args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( !$_logInitted ) {
        ( $_odev, $_oino ) = stat(STDOUT);
        ( $_edev, $_eino ) = stat(STDERR);
        $_logInitted    = 1;
        $_trace_enabled = defined $ENV{NOTE_TRACE};
        if ($_trace_enabled) {
            local @ARGV = split / /, $ENV{NOTE_TRACE};
            getopts('bspr');
        }
    }
    my $s;
    if ( $args[0] =~ /%/ && ( scalar @args > 1 ) ) {
        my $f = shift(@args);
        $s = sprintf( $f, @args );
    }
    else {
        $s = join '', @args;
    }
    my $newStep;
    if ($_trace_enabled) {
        my $prefix = "";
        my $level  = 0;
        my $step;
        my @frames      = ();
        my $lastLine    = "";
        my $lastFile    = "";
        my $nearestStep = undef;
        my $inMain      = 0;
        my $inTeardown  = 0;
        my ( $package, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require );
        my $i = 0;

        while (1) {
            ( $package, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require ) = caller($i);
            $i++;
            last unless $package;
            if ($inMain) {
                if ( $package ne 'main' ) {
                    if ($opt_p) {    # Display test phase, if new
                        my $subname = $subroutine;
                        $subname =~ s/^main:://;    # Phase is top-level subroutine name in ::main
                        $subname = "Unknown" if ( $subname eq '(eval)' );
                        if ( $subname ne $_curphase ) {

                            # Entering new phase
                            my $out = sprintf( "\n\n%s %s %s\n", '%%%% Phase: ', $subname, '%%%%' );
                            if ( ( $_odev != $_edev ) || ( $_oino != $_eino ) ) {
                                if ( $NOTE::LOGLEV eq '1' ) {
                                    print $NOTE::STDOUT $out, "\n";
                                    STDERR->print( $out, "\n" );
                                }
                            }
                            Tharn::logcomment($out);
                            $_curphase = $subname;
                        }
                    }
                    last;
                }
            }
            next if ( $subroutine eq "(eval)" );
            next if ($is_require);
            $subroutine =~ s/^NOTE::/::/;

            if ( $subroutine ne "::logcomment" ) {
                unshift @frames, $subroutine . " [" . $lastLine . "]";
                $level++;
            }
            if ( $package eq 'main' ) {
                $inMain++;
                $inTeardown++ if ( $subroutine eq '::teardown' );
                if ($opt_s) {
                    if ( !$_initted ) {
                        $_initted++;
                        _getSteps($filename);
                    }
                    if ( !defined $nearestStep ) {
                        if ( $line2step[$line]->{offset} < 10 ) {    # ADJUST this
                            $nearestStep = $line;
                        }
                    }
                }
            }
            $lastLine = $line;
            $lastFile = $filename;
        }
        my $ind = 0;
        while (( defined $_lastFrame[$ind] && defined $frames[$ind] )
            && ( $_lastFrame[$ind] eq $frames[$ind] ) )
        {
            $ind++;
        }

        my $_nextHeader = "";
        if ($opt_b) {
            for my $tlev ( $ind .. ( $level - 1 ) ) {
                $_nextHeader .= "    " for ( 0 .. $tlev );
                $_nextHeader .= $frames[$tlev];
                $_nextHeader .= "\n";
            }
            chomp($_nextHeader);
        }

        my $_nextLine;
        if ( $opt_s && defined( $step = $line2step[$nearestStep]->{step} ) ) {

            # Hack because teardown usually doesn't have @step:
            $step = "[Teardown]" if ($inTeardown);
            $_nextLine = $step;
            my $offset = $line2step[$nearestStep]->{offset};
            $_nextLine .= "+$offset" if ($offset);
            $_nextLine .= " [$nearestStep]";
        }
        else {
            $_nextLine = "$lastFile [$lastLine]";
        }
        $newStep = 0;
        if ( $_lastLine ne $_nextLine ) {
            $prefix .= "\n$_nextLine\n";
            $newStep   = 1;
            $_lastLine = $_nextLine;
        }
        if ( $_lastHeader ne $_nextHeader ) {
            $prefix .= "$_nextHeader\n" if ($_nextHeader);
            $_lastHeader = $_nextHeader;
        }
        if ($opt_b) {
            my $pad = "    " x ( $level + 1 );
            $s =~ s/\n/\n$pad/g;
            $s = $prefix . $pad . $s;
        }
        else {
            my $pad = "    ";
            $s =~ s/\n/\n$pad/g;
            $s = $prefix . $pad . $s;
        }
        @_lastFrame = @frames;
    }
    if ( ( $_odev != $_edev ) || ( $_oino != $_eino ) ) {
        if ( $NOTE::LOGLEV eq '1' ) {
            print $NOTE::STDOUT $s, "\n";

            #	    STDERR->print( $s, "\n" );
        }
    }
    if ($_trace_enabled) {
        $s = "\n" . $s if ( $s !~ /^\n/ );
        if ($newStep) {
            $s = "%%%" . $s;
        }
        else {
            $s = "%%" . $s;
        }
    }
    Tharn::logcomment($s);
}

=head2 ::units

If passed a value with a unit specifier, (i.e. t = terabytes, g = gigabytes, m = megabytes, etc.. ),
returns the value in kilobytes (without a unit specifier). If no unit specifier is present,
the value is assumed to already be in kilobytes.

=over

=item Argument

I<value>

=item Example

     my $kb = NOTE::units("200m");

=back

=cut

sub units {
    my $size = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $size if ( $size =~ /^\d+$/ );
    my $bytes = Utils::Units::convert_to_bytes($size);
    my $kb = Utils::Units::convert_units( $bytes, Utils::Units::UNIT_B, Utils::Units::UNIT_KB );
    return $kb;
}

my @_objList = ();

sub mark {
    my $mark = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::reference( \$mark );
    return $mark;
}

sub reference {
    my $ref = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    push @_objList, $ref if ($ref);
    if (wantarray) {
        return @_objList;
    }
    else {
        return scalar @_objList;
    }
}

sub unreference {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    for my $i ( 0 .. $#_objList ) {
        if ( $_objList[$i] == $self ) {
            delete $_objList[$i];
        }
    }
}

my $_resourceSerial = 0x0000;

sub resource {
    my ( $self, $mode ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return unless ($opt_r);

    my $props = "";

    # Trace resource usage
    if ( $mode eq 'allocate' ) {
        $self->{_serialNumber} = $_resourceSerial++;
    }
    if ( ref $self eq 'NOTE::Aggregate' ) {
        $props .= "Aggregate $mode: ";
        $props .= sprintf( "filer=%s ", $self->Filer->host );
        $props .= sprintf( "serialNumber=%d ", $self->{_serialNumber} );
        $props .= sprintf( "disks=%d ", scalar $self->disks );
        $props .= sprintf( "name=%s ", $self->name );
        if ( $mode eq 'allocate' ) {
            my %properties = $self->Properties;
            $props .= sprintf( "share=%d ",     $self->is_shareable );
            $props .= sprintf( "recycle=%d ",   $self->is_recyclable );
            $props .= sprintf( "anonymous=%d ", $self->is_anonymous );
            $props .= sprintf( "snaplock=%s ",  $properties{snaplock} );
            $props .= sprintf( "raidtype=%s ",  $properties{raidtype} );
            $props .= sprintf( "raidsize=%s ",  $properties{rgsize} );
        }
    }
    elsif ( ref $self eq 'NOTE::Volume' && $self->type eq 'trad' ) {
        $props .= "Volume $mode: ";
        $props .= sprintf( "filer=%s ", $self->Filer->host );
        $props .= sprintf( "serialNumber=%d ", $self->{_serialNumber} );
        $props .= sprintf( "disks=%d ", scalar $self->disks );
        $props .= sprintf( "name=%s ", $self->name );
        if ( $mode eq 'allocate' ) {
            my %properties = $self->Properties;
            $props .= sprintf( "type=%s ",      $self->type );
            $props .= sprintf( "recycle=%d ",   $self->is_recyclable );
            $props .= sprintf( "anonymous=%d ", $self->is_anonymous );
            $props .= sprintf( "snaplock=%s ",  $properties{snaplock} );
            $props .= sprintf( "raidtype=%s ",  $properties{raidtype} );
            $props .= sprintf( "raidsize=%s ",  $properties{rgsize} );
        }
    }
    else {
        return;
    }
    NOTE::logcomment("Resource: $props");
}

sub RECYCLE { 
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    NOTE::option(q/$NOTE_RECYCLE:1/) 
}
sub DATAMANIP { 
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    NOTE::option(q/$NOTE_DATAMANIP:1/) 
}

=head2 ::DONT_SPARE_VOLUMES

Returns value of $NOTE::_DONT_SPARE_VOLUMES global variable, which can be set by
NOTE_DONT_SPARE_VOLUMES CLI parameter. By default NOTE doesn't destroy volumes
with datasets. This parameter will force volumes to be destroyed even if dataset
was created.

NOTE_SPARE has higher priority over DONT_SPARE_VOLUMES, and volumes wont get
destroyed if both NOTE_SPARE and DONT_SPARE_VOLUMES are present

=cut

sub DONT_SPARE_VOLUMES { 
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $NOTE::_DONT_SPARE_VOLUMES || NOTE::option(q/$NOTE_DONT_SPARE_VOLUMES/) 
}

sub ALL_SPARED { 
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $NOTE::_SPARE_ALL || NOTE::option(q/$NOTE_SPARE/) 
}

sub SPARE_ALL_ON_ERROR {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return 1 if ($NOTE::_SPARE_ALL_ON_ERROR);
    return 1 if ( NOTE::option(q/$NOTE_SPARE_ALL_ON_ERROR/) );
    return 0;
}

sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    NOTE::logcomment("NOTE::SPARE: Disabling object destruction on teardown");
    $NOTE::_SPARE_ALL = 1;
}

sub consume {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $NOTE::_CONSUMED_RESOURCES = 1;
}

sub panicked {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $NOTE::_PANICKED++;
}

=head2 ::teardown

Cleanup all NOTE objects by calling their DESTROY methods in reverse order of creation.
Attempts to restart any non-pingable filers before cleanup of their objects.

Cleanup will be skipped if CLI parameter NOTE_SPARE is set to 1, which can be overridden
within a script by calling NOTE::SPARE;

If NOTE_SPARE_ALL_ON_ERROR is set to 1, teardown will set NOTE_SPARE to 1 if
the exit code passed to it is non-zero (i.e. a failure code) or if any object
destructor croaks.

=over

=item Argument

B<exitcode> => I<value>

I<Optional> Optional value to be use as NOTE exit code

=item Example

      NOTE::teardown(exitcode=>255);

=back

=cut

my $tear_errors = "";

# Takes an optional argument, that if specified, will cause the stack
# to be torn down only until a matching reference is encountered.
# This can be used to teardown all the objects created by a
# particular code block:
#
#    NOTE::mark( "here" );
#
#    $sm = NOTE::SnapMirror->initialize;
#    ...
#
#    NOTE::down( "here" );
#
sub down {
    my $mark = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    while (@_objList) {
        my $errmsg = "";
        my $o      = pop @_objList;
        next unless defined $o;
        if ( ref $o eq 'CODE' ) {
            eval { &$o() };
            $errmsg .= "    - " . $@ if ($@);
        }
        elsif ( ref $o eq 'SCALAR' ) {
            return if ( $$o eq $mark );
        }
        else {
            if ( $o->{_teardown} ) {
                if ( ref $o->{_teardown} eq 'HASH' && $o->{_teardown}->{pre} ) {
                    eval { &{ $o->{_teardown}->{pre} }($o) };
                    $errmsg .= "    - " . $@ if ($@);
                }
            }
            eval { $o->DESTROY };
            $errmsg .= $@;
            if ( $o->{_teardown} ) {
                if ( ref $o->{_teardown} eq 'HASH' && $o->{_teardown}->{post} ) {
                    eval { &{ $o->{_teardown}->{post} }( $o, $errmsg ) };
                    $errmsg .= "    - " . $@ if ($@);
                }
                elsif ( ref $o->{_teardown} eq 'CODE' ) {
                    eval { &{ $o->{_teardown} }( $o, $errmsg ) };
                    $errmsg .= "    - " . $@ if ($@);
                }
            }
        }
        if ($errmsg) {
            $tear_errors .= $errmsg;
            NOTE::logcomment($errmsg);
            if ( NOTE->SPARE_ALL_ON_ERROR ) {
                NOTE::logcomment("NOTE::SPARE_ALL_ON_ERROR is set");
                NOTE->SPARE;
            }
        }
    }
}

sub teardown {
    local $?;    # Don't allow exit code to be changed during teardown
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ($teardown_called);
    $teardown_called = 1;

    $stdout->flush if ($stdout);

    #
    # 20090413		tyu@netapp.com
    #
    # The following block handle the situation where the user (test executor)
    # indicate he/she wants to preserve the configured entities (i.e. bypass
    # teardown) if there's any failure occurred during the test.
    #

    if ( NOTE->SPARE_ALL_ON_ERROR() ) {
        my @result_objects = own_results();

        if (@result_objects) {
            my $flag_need_spare = 0;
            my $num_of_results  = @result_objects;
            my $idx             = 0;

            NOTE::logcomment( "(GNOTE::teardown): num_of_results is " . $num_of_results );

            #
            # 20090413		tyu@netapp.com
            #
            # The following block handles the condition where croak, confess,
            # or die was called during the course of the test. This is working
            # in connection with TestWrapper.pm, which is responsible to detect
            # the calling of croak, confess, and die, and posted a special INFO
            # result object into the Test_DB. This enable the block here to
            # determine whether the test is in the failure condition or not.
            #

            while ( ( $flag_need_spare != 1 ) && ( $idx < $num_of_results ) ) {
                my @the_messages = $result_objects[$idx]->message();
                if (@the_messages) {
                    if ( $the_messages[0] eq "Error detected for NOTE_SPARE_ALL_ON_ERROR" ) {
                        $Trace->debug( msg => "(GNOTE::teardown): Need Spare, A" );
                        $flag_need_spare = 1;
                    }
                }
                $idx++;
            }

            #
            # 20090413		tyu@netapp.com
            #
            # The following block handles the condition where the test script
            # specifically log a result into the TestDB with severity higher
            # than Warn.
            #

            if ( $flag_need_spare != 1 ) {
                my $worst_result_obj = NATE::Result::worst(@result_objects);

                $Trace->debug( msg => "(GNOTE::teardown): worst severity is " . $worst_result_obj->severity() );
                if ( $worst_result_obj->severity() > NATE::Result::Warn::severity() ) {
                    $Trace->debug( msg => "(GNOTE::teardown): Need Spare, B" );
                    $flag_need_spare = 1;
                }
            }

            if ( $flag_need_spare == 1 ) {
                $Trace->debug( msg => "(GNOTE::teardown): Before NOTE::SPARE" );
                NOTE->SPARE();
            }
        }
    }

    if ( defined $args{exitcode} ) {

        # probably called from an END block
        if ( $args{exitcode} == 0 ) {
            NOTE::logcomment("NOTE: (Success) Teardown started");
        }
        else {
            NOTE::logcomment("NOTE: (Failure[$args{exitcode}]) Teardown started");
            if ( NOTE->SPARE_ALL_ON_ERROR ) {
                NOTE::logcomment("NOTE::SPARE_ALL_ON_ERROR is set");
                NOTE->SPARE;
            }
        }
    }
    else {
        NOTE::logcomment("NOTE: Teardown started");
    }

    if ( NOTE->ALL_SPARED ) {
        NOTE::logcomment("NOTE::SPARE is set: no objects will be destroyed");
    }

    # Check that all filers are up. If not, bring them up
    my @SaveLogsPid = ();
    for my $obj (@_objList) {
        next unless defined $obj;    # This might happen because of NOTE::unreference
        next if ( ref $obj eq 'CODE' );
        next if ( ref $obj eq 'SCALAR' );
        if ( $obj->isa("NOTE::Filer") ) {
            if ( !$obj->ping ) {
                if ( NOTE->ALL_SPARED || $obj->{_spare_me} ) {
                    NOTE::logcomment( "NOTE: %s seems to have panicked.", $obj->hostspec );
                    next;
                }
                NOTE::logcomment( "NOTE: Bringing up %s because it seems to have panicked.", $obj->hostspec );
                $obj->panicked;
                eval { $obj->transit_to("UP") };
                if ($@) {
                    NOTE::logcomment( "NOTE: Bringup of %s failed; $@", $obj->hostspec );
                }
                else {
                    NOTE::logcomment( "NOTE: Bringup of %s successful.", $obj->hostspec );
                }
            }

            if ( $obj->ping && $_SAVE_LOGS ) {    # try to save logs...
                my $logsto = $obj->host . '_filerlogs';    # this will effect directory name
                push @SaveLogsPid, subtest( \&NOTE::7G::save_logs, '-runid', $logsto, '-bg', '-onexit', sub { }, '--', $obj );
            }
        }
    }
    foreach (@SaveLogsPid) { $_->wait_finish if ( defined $_ ); }

    NOTE::down();

    if ($NOTE::USAGE_ANALYSIS) {
        NOTE::logcomment( "NOTE: %s resources were consumed.", $NOTE::_CONSUMED_RESOURCE ? "Some" : "No" );
        if ($NOTE::_PANICKED) {
            NOTE::logcomment( "NOTE: %s filer%s may have panicked.", $NOTE::_PANICKED, $NOTE::_PANICKED == 1 ? "" : "s" );
        }
    }
    if ($tear_errors) {
        if ( defined $args{exitcode} && $args{exitcode} == 0 ) {
            die $tear_errors;
        }
        else {
            NOTE::logcomment( "Errors during teardown:\n%s", $tear_errors );
        }
    }
}

=head1 STATUS

Public.

=head1 AUTHOR

dl-dpgqa-automation@netapp.com

=cut

1;
