#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/7G/Qtree.pm#1 $
#
## @summary  NOTE Qtree class
## @status   Public
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::Qtree

=head1 DESCRIPTION

This package provides methods for creating and manipulating Qtrees.
There are three constructor methods for Qtree objects;
C<::create>,
C<::attach>,
and
C<::new>.
The appropriate one to use depends on
the circumstances as explained for each method.

Qtree objects support many of the same methods that Volume objects
do.

=head1 METHODS

=cut

package NOTE::7G::Qtree;
use GNOTE::7G;
use GNOTE::Qtree;
our @ISA = qw(NOTE::Qtree NOTE::7G);

use Carp qw(croak);
use strict;

# NOTE: Do *NOT* import any external functions into this module!
# This is to avoid namespace collisions namely prototype mismatches and
# all that. Call the external function directly by prepending the package
# name.
use QtreeCmds qw();

=head2 B<Creation, Naming, and Deletion>

=head2 ::create

Creates a Qtree and returns
an object representing the created Qtree.

=over

=item Arguments

B<Volume> => I<object> | I<spec>

I<Optional.>
The volume in which to create the Qtree.
May be a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
object or a volume spec.
If absent, a B<Filer> or B<Aggregate> argument must be given
and an anonymous volume will be created or attached
before the call returns.

B<name> => I<string>

I<Optional.>
The name of the Qtree.
If absent, an anonymous Qtree name will be generated.

B<Filer> => I<host spec>

I<Optional.>
The filer on which to create the Qtree.

B<Aggregate> => I<object> | I<spec>

I<Optional.>
The aggregate in which to create a volume
in which to create the Qtree.
May be a
L<C<NOTE::Aggregate>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Aggregate-pm/>
object or an aggregate spec.
If absent, a B<Volume> or B<Filer> argument must be given.

B<tag> => I<string>

I<Optional.>
A text description of the object.
See L<C<::tag>|/::tag> for details.

One of B<Volume>, B<Aggregate>, or B<Filer> must be given.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip for
L<C<::populate>|/::populate> and L<C<::modify>|/::modify>.
The default is 0.

=item Examples

=back

     $qt = NOTE::Qtree->create(
         Volume => $vol,
         tag    => 'destination'
     );

     $qt = NOTE::Qtree->create(
         Filer     => { host => $FILER_SRC },
         Volume    => { size => "20g" },
         Aggregate => { disks => 5, raidtype => 'raid4' },
         tag       => 'really big Qtree'
     );

=cut

sub create {
    my ( $self, %args ) = @_;

    $self = __PACKAGE__->new(%args);
    $self->SUPER::create(%args);
    $self->{_destroyed} = 0;
    return $self;
}

sub _qtree_create {
    my ($self) = @_;

    NOTE::logcomment( "Creating Qtree " . $self->Volume->Filer->host . ":" . $self->mount_path );
    QtreeCmds::qtreeCreate( $self->Volume->Filer->conn, $self->mount_path );
    croak $@ if $@;
    return $self;
}

=head2 ::attach

Creates an object representing an existing Qtree.

=over

=item Arguments

B<name> => I<string>

I<Required.>
The name of an existing Qtree.

B<Volume> => I<object> | I<spec>

I<Required.>
The volume in which to find the named Qtree.
Specified as a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/> object
or a volume spec.

B<Filer> => I<host spec>

I<Dependent.>
Required only when the B<Volume> argument is a volume spec.
Specifies the filer where the volume exists.

B<mntpoint> => I<string>

I<Optional.>
The full path name on which to mount the Qtree.

B<exportfs> => I<export options>

I<Optional.>
Export options as described in L<C<::exportfs>|/::exportfs>.

B<dataset> => I<dataset specifier>

I<Optional.>
A description of some data to put in the Qtree.
The dataset will be fully created
before the call to C<::create> returns.
See L<C<NOTE::Volume::fill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::fill>
for a description of dataset specifiers.

B<force> => C<1>

I<Optional.>
A flag to force the creation of a Qtree object
even if the named Qtree does not exist.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip for
L<C<::populate>|/::populate> and L<C<::modify>|/::modify>.
The default is 0.

=item Examples

=back
     $qt = NOTE::Qtree->attach(
         Volume => $vol,
         name   => 'qtree42'
     );

=cut

sub attach {
    my ( $class, %args ) = @_;
    my $self          = {};
    my $force         = $args{force};
    my $exportfs_opts = $args{exportfs} || "rw,anon=0";
    delete $args{exportfs};

    bless $self, $class;

    $self->SUPER::attach(%args);

    unless ( $force || QtreeCmds::qtreePresent( $self->{_Volume}->Filer->conn, $self->mount_path ) ) {

        # $@ is not always set above when it should be, so here's a back up.
        croak "Qtree " . $self->hostpath . " does not exist";
    }
    eval { $self->exportfs($exportfs_opts) };    # Might fail if qtree doesn't exist
    croak $@ if ( $@ && !$force );

    return $self;
}

=head2 ::dup

Create a qtree with attributes similar to an existing qtree.

=over

=item Arguments

Takes the same arguments as NOTE::Qtree::create.

=item Examples

  for my $i ( 0..100 ) {
     $qt[$i] = $qt_main->dup;
  }

=back

=cut

sub dup {
    my ( $origin, %args ) = @_;
    my %properties = $origin->Properties;

    %properties = %{ NOTE::_hashOverlay( target => \%properties, source => \%args, overwrite => 1 ) };
    $properties{Filer} ||= $origin->Filer;    ## change Filer to Volume if you find a dup script
    return NOTE::Qtree->create(%properties);
}

=head2 ::new

Creates an object representing a Qtree,
but does B<not> cause the creation of a Qtree on the filer.
Qtree objects created in this manner
may be used as destinations with
L<C<NOTE::SnapMirror::initialize>|engineering::ontap_tests::agnostic::docs::lib-NOTE-SnapMirror-pm/::initialize>.

=over

=item Arguments

All of the arguments to L<C<::create>|/::create> are also
valid arguments to C<::new> and have the same effect, except
that C<::new> does not accept the B<dataset> argument.


=item Examples

=back

     $qt_dst = NOTE::Qtree->new(
         Volume => $vol,
         tag    => 'destination'
     );
     NOTE::SnapMirror::initialize(
         src => $qt_src,
     dst => $qt_dst
     );

=cut

sub new {
    my ( $class, %args ) = @_;
    my $self = {};

    exists $args{snaplock} and $args{snaplock} = NOTE::option( $args{snaplock} );

    # if the $args{Volume} is a volume object, use it,
    # otherwise try to create one by calling the NOTE::Volume constructor.
    if ( !defined( $args{Volume} ) || ( ref $args{Volume} eq "HASH" ) ) {
        $args{Volume}->{snaplock} ||= $args{snaplock};

        ## XXX Should I croak here?:::
        delete $args{Volume}->{Filer} if ( $args{Filer} );

        if ( $args{Aggregate} ) {
            $self->{_Volume} = NOTE::7G::Volume->create(
                Filer     => $args{Filer},
                Aggregate => $args{Aggregate},
                %{ $args{Volume} },
            );
        }
        else {
            $self->{_Volume} = NOTE::7G::Volume->create(
                Filer => $args{Filer},
                %{ $args{Volume} },
            );
        }
    }
    else {
        $self->{_Volume} = $args{Volume};
    }
    delete $args{Filer};
    delete $args{Aggregate};
    delete $args{Volume};

    bless $self, $class;

    $self->SUPER::new(%args);

    return $self;
}

=head2 ::delete

Removes the Qtree.  Does not affect the Qtree object.
(The resultant state of the object will be like that of a newly created object,
and as such the object becomes suitable for use as a destination in
L<C<NOTE::SnapMirror::initialize>|engineering::ontap_tests::agnostic::docs::lib-NOTE-SnapMirror-pm/::initialize>.) 

=over

=item Arguments

None.

=item Example

=back

     $qt->delete;

=cut

sub destroy {
    my $self = shift;

    $self->cd('/');
    $self->delete(@_);
}

sub delete {
    my $self = shift;

    if ( $self->pwd eq '/' ) {
        NOTE::logcomment( "Deleting " . $self->description );
        $self->umount;
        my $container_mountpoint = $self->Volume->mountpoint;
        if ($container_mountpoint) {

            # Take advantage of the fact that the container is
            # already mounted.
            QtreeCmds::qtreeDelete( $self->Filer->conn, $self->mount_path, VOL_MOUNTPATH => $container_mountpoint, );
        }
        else {
            QtreeCmds::qtreeDelete( $self->Filer->conn, $self->mount_path );
        }
        croak $@ if $@;
    }
    else {

        # Delete a subdirectory
        NOTE::7G::Volume::delete($self);
    }
}

# Changes the name of the Qtree.

# The new name for the Qtree.
# If absent a new anonymous name is generated
# and the Qtree becomes anonymous.
# If present the Qtree will no longer be anonymous,
# causing its volume (and aggregate) to be spared
# (not destroyed) when the Qtree object is destroyed.

# $qt->rename;
# $qt->rename( 'newQtreeName' );

sub _rename {
    my ( $self, $oldpath ) = @_;
    my $olddesc = $self->description;

    # Need to get the current exportfs options, unexportfs
    # the old name and re-exportfs them using the new name.
    eval {
        NOTE::logcomment( "Renaming " . $olddesc . " to " . $self->mount_path );
        QtreeCmds::qtreeRename(
            conn   => $self->Filer->conn,
            qt_src => $oldpath,
            qt_dst => $self->mount_path
        );
    };
    croak $@ if $@;
}

=head2 B<Waiting>

=head2 ::wait

Waits for all background processes to complete.
These are the processes created by using the B<background> option of
L<C<::populate>|/::populate>, L<C<::modify>|/::modify>
or L<C<::compare>|/::compare>

=over

=item Arguments

None.

=item Example

=back

     $vol->wait;

=cut

sub wait {
    return NOTE::7G::Volume::wait(shift);
}

=head2 B<Related Objects>

=head2 ::Filer

Returns an object representing the filer containing the Qtree.

=over

=item Arguments

None.

=item Example

=back

     $filer = $qt->Filer;

=cut

sub Filer { return shift->{_Volume}->Filer }

=head2 B<Disk Space>

=head2 ::disk_free

Returns a list of items about the disk usage
of the volume containing the Qtree.
The items returned,
and the order in which they are returned are:

B<total> -- the total disk space extant in the volume

B<used> -- the total disk space currently used in the volume

B<avail> -- the total disk space currently available in the volume

B<capacity> -- the percentage of the total disk space
in the volume that is currently being used

There are also methods by each of the item names
above that return only the value for that particular item.

=over

=item Arguments

None.

=item Example

=back

     ($total, $used, $avail, $capacity) = $qt->disk_free;

=cut

sub disk_free {
    my ( $self, %args ) = @_;
    return NOTE::7G::Volume::disk_free( $self, %args );
}

=head2 B<Mounting>

=head2 ::exportfs

Sets NFS export options
on the volume containing the Qtree.

=over

=item Arguments

A list of export options as defined in
L<C<NOTE::Volume::exportfs>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::exportfs>.

=item Example

=back

     $qt->exportfs( 'rw,nosuid,anon=0' );   # No spaces allowed within the list

=cut

sub exportfs {
    my ( $self, %opts ) = @_;
    return NOTE::7G::Volume::exportfs( $self, %opts );
}

=head2 ::remount

Remounts the NFS volume containing the Qtree.

=over

=item Arguments

I<string>

I<Optional.>
A list of mount options as defined in
L<C<NOTE::Volume::mount>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::remount>.

=item Example

=back

     $qt->remount;

=cut

sub remount {
    my ( $self, %opts ) = @_;
    return NOTE::7G::Volume::remount( $self, %opts );
}

=head2 ::mount_path

Returns the full path name of the Qtree,
i.e. C</vol/><I<containing volume name>>C</>I<<Qtree nameE<gt>>

=over

=item Arguments

None.

=item Example

=back

     $mntpt = $qt->mount_path;

=cut

sub mount_path {
    my $self = shift;
    return "/vol/" . $self->Volume->name . "/" . $self->{_name};
}

=head2 ::mount

NFS mounts the 7G volume containing the Qtree.

=over

=item Arguments

I<string>

I<Optional.>
A list of mount options as defined in
L<C<NOTE::7G::Volume::mount>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-7G-Volume-pm.html#%3a%3amount>

=item Example

=back

     $qt->mount( chmod => '777' );

=cut

sub mount {
    my ( $self, %opts ) = @_;
    return NOTE::7G::Volume::mount( $self, %opts );
}

=head2 ::umount

Unmounts the 7G volume containing the Qtree.

=over

=item Arguments

I<None>

=item Example

=back

     $qt->umount;

=cut

sub umount {
    my ( $self, %opts ) = @_;
    return NOTE::7G::Volume::umount( $self, %opts );
}

=head2 ::hostpath

Returns the full path name of the Qtree
(as returned by L<C<::mount_path>|/::mount_path>)
prepended by its filer's name.

=over

=item Arguments

None.

=item Example

=back

     $hp = $qt->hostpath;

=cut

sub hostpath {
    my $self = shift;
    return $self->Filer->host . ":" . $self->mount_path;
}

=head2 B<Access Control and Status>

=head2 ::oplocks

Get or set the oplocks state of the Qtree.

=over

=item Arguments

I<boolean>

I<Optional.>
If true, enables oplocks on the Qtree.
If false, disables oplocks on the Qtree.
If absent, returns the current oplock setting for the Qtree.
(Where true means oplocks are enabled,
and false means oplocks are disabled.)

=item Examples

=back

     $oplocks_on = $qt->oplocks;
     $ret = $qt->oplocks( 1 );   # Enable oplocks
     $ret = $qt->oplocks( 0 );   # Disable oplocks

=cut

sub oplocks {
    my ( $self, $val ) = @_;
    my $ret;
    if ( defined $val ) {
        $ret = QtreeCmds::qtreeOplocks( $self->Filer->conn, $self->mount_path, $val );
    }
    else {
        $ret = QtreeCmds::qtreeOplocks( $self->Filer->conn, $self->mount_path );
    }
    croak $@ if $@;
    return $ret;
}

=head2 ::security

Get or set the security style of the Qtree.

=over

=item Arguments

C<ntfs> | C<unix> | C<mixed>

I<Optional.>
If present, sets the security style to the specified type.
If absent, returns the current security setting for the Qtree.

=item Examples

=back

     $type = $qt->security;
     $ret = $qt->security( 'unix' );

=cut

sub security {
    my ( $self, $val ) = @_;
    my $ret;
    if ( defined $val ) {
        $val = NOTE::option($val);
        $ret = QtreeCmds::qtreeSecurity( $self->Filer->conn, $self->mount_path, $val );
    }
    else {
        $ret = QtreeCmds::qtreeSecurity( $self->Filer->conn, $self->mount_path );
    }
    croak $@ if $@;
    return $ret;
}

=head2 ::status

Get the status of the Qtree.
The return value is a hash reference returned by
L<C<lib::QtreeCmds::qtreeStatus>|engineering::ontap_tests::agnostic::docs::lib-QtreeCmds-pm/qtreeStatus($connobj,$qtreepath)>


=over

=item Arguments

None.

=item Examples

=back

     $status = $qt->status;

=cut

sub status {
    my $self = shift;
    my $ret = QtreeCmds::qtreeStatus( $self->Filer->conn, $self->mount_path );
    croak $@ if $@;
    return $ret;
}

=head2 ::snapmirrored

Returned true if the qtree is a snapmirror destination.

=over

=item Arguments

None.

=item Examples

=back

     if ( $qtree->snapmirrored ) {
    ...
     }

=cut

sub snapmirrored { return shift->status->{status} =~ /snapmirrored/ }

=head2 ::writeable

Returns a boolean indicating whether the qtree is writeable.

=over

=item Arguments

None.

=item Examples

=back

      if ( $qtree->writeable ) {
    ...
      }

=cut

sub writeable {
    my $self = shift;

    return ( $self->Volume->writeable && !$self->snapmirrored );
}

=head2 ::readable

Returns a boolean indicating whether a qtree is readable.
(It is readable if the containing volume is readable)

=over

=item Arguments

None.

=item Examples

=back

      if ( $qtree->readable ) {
    ...
      }

=cut

sub readable { return shift->Volume->readable }

=head2 ::is_recycled

Returns a true if a qtree was recycled
and false if it was physically created.

=over

=item Arguments

None.

=item Examples

=back

      my $qtree = NOTE::Qtree->create( recycle => 1, name => 'myqtree' );
      if ( $qtree->is_recycled ) {
          printf( "This Qtree was recycled\n" );
      } else {
          printf( "This Qtree was created\n" );
      }

=cut

sub is_recycled { return shift->{_was_recycled} }

=head2 ::is_recyclable

Returns 1 if the qtree is a recyclable qtree.
This is also assignable, if it is desired to change the
status.


=over

=item Arguments

None.

=item Examples

    if ( $qtree->is_recyclable ) {
    printf( "Qtree %s is recyclable\n", $qtree->name );
    }

    ...

    $qtree->is_recycleable = 1;

=back

=cut

sub is_recyclable : lvalue {
    shift->{_recycle};
}

=head2 ::is_free

Returns a boolean to indicate whether this qtree is in the free pool.

=over

=item Arguments

None.

=item Eample

=back

    for my $qtree ( $filer->Qtrees ) {
    next unless ( $qtree->is_free );
    printf( "Qtree %s is available for reuse\n", $qtree->name );
    }

=cut

sub is_free { return shift->{_name} =~ /^_vl/ }

=head2 ::exists

Returns true if the Qtree exists.

=over

=item Arguments

None.

=item Examples

=back

     $qt->destroy;
     if ( $qt->exists ) {
    die "Qtree wasn't destroyed";
     }

=cut

sub exists {
    return eval { shift->status };
}

=head2 B<Creating and Using Files>

=head2 ::fill

Identical to L<C<NOTE::Volume::fill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::fill>
except that the scope of the operation is the Qtree.

=head2 ::refill

Identical to L<C<NOTE::Volume::refill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::refill>
except that the scope of the operation is the Qtree.

=head2 ::checkfill

Identical to L<C<NOTE::Volume::checkfill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::checkfill>
except that the scope of the operation is the Qtree.


Identical to L<C<NOTE::Volume::unfill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::unfill>
except that the scope of the operation is the Qtree.

=head2 ::fillfile

Identical to L<C<NOTE::Volume::fillfile>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::fillfile>
except that the scope of the operation is the Qtree.

=head2 ::populate

Identical to L<C<NOTE::Volume::populate>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::populate>
except that the scope of the operation is the Qtree.

=head2 ::modify

Identical to L<C<NOTE::Volume::modify>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::modify>
except that the scope of the operation is the Qtree.

=head2 ::compare

Identical to L<C<NOTE::Volume::compare>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::compare>
except that the scope of the operation is the Qtree.

=cut

sub fill {
    my ( $self, %args ) = @_;
    return NOTE::7G::Volume::fill( $self, %args );
}

sub refill {
    my ( $self, %args ) = @_;
    return NOTE::7G::Volume::refill( $self, %args );
}

sub checkfill {
    my ( $self, %args ) = @_;
    return NOTE::7G::Volume::checkfill( $self, %args );
}

sub unfill {
    my ( $self, %args ) = @_;
    return NOTE::7G::Volume::unfill( $self, %args );
}

sub fillfile {
    my ( $self, %args ) = @_;
    return NOTE::7G::Volume::fillfile( $self, %args );
}

sub sio {
    my ( $self, %args ) = @_;

    return NOTE::7G::Volume::sio( $self, %args );
}

use FSR::FSRutils;

sub compare {
    my ( $self, $other, %args ) = @_;

    # I don't particularly like this, but it's necessary
    # because otherwise DataManip::compare will fail due
    # to permission problems:
    $args{IGNORE_LIST} ||= FSR::FSRutils::serialize( ".snapshot", );
    return NOTE::7G::Volume::compare( $self, $other, %args );
}

=head2 B<Misc>

sub DESTROY {
    my $self = shift;

    return if ( $$ != $self->{_owner} );
    $self->sio(stop=>1) if exists $self->{_sio}; # Terminate any on-going SIP process
    return unless ( $self->{_anonymous} );
    return if ( $self->{_destroyed} );
    $self->{_destroyed} = 1;
    return if ( NOTE->ALL_SPARED );
    
    $self->umount;
    if ( $self->{_spare_me} ) {
        NOTE::logcomment( "Qtree ".$self->description." has been spared");
    return;
    }
    # Don't bother deleting qtree since it will be cleaned up when
    # the containing volume is restored using the __clean__ snapshot.
}

=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::agnostic::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::FileSystem::NFS>|engineering::ontap_tests::agnostic::docs::lib-NOTE-FileSystem-NFS-pm/>

L<C<NOTE::SnapMirror>|engineering::ontap_tests::agnostic::docs::lib-NOTE-SnapMirror-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>

=head1 STATUS

Public.

=head1 AUTHOR

dl-dpgqa-automation@nwtapp.com

=cut

1;
