#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/7G/SnapVault.pm#1 $
#
## @summary  NOTE SnapVault class
## @status   Review
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::SnapVault

=head1 DESCRIPTION

This is the NOTE class for working with SnapVault relationships.

=head1 METHODS

=cut

use strict;

package NOTE::SnapVault;

use Carp qw(croak);
use GNOTE::7G;
our @ISA = qw(NOTE::7G);

use POSIX qw();
use SnapVault qw();
use FSR::FSRutils qw();

# Expose the following registry keys on the object
my $regSlice = {
    'snapvault.access' => sub {
        'options.snapvault.access'
    },
    'snapvault.lockvault_log_volume' => sub {
        'options.snapvault.lockvault_log_volume'
    },
    'snapvault.nbu.trace_category' => sub {
        'options.snapvault.nbu.trace_category'
    },
    'snapvault.nbu.trace_level' => sub {
        'options.snapvault.nbu.trace_level'
    },
    'snapvault.nbu.verify_backup' => sub {
        'options.snapvault.nbu.verify_backup'
    },
    'snapvault.nbu.verify_duplicate' => sub {
        'options.snapvault.nbu.verify_duplicate'
    },
    'snapvault.offline_reboot.enable' => sub {
        'options.snapvault.offline_reboot.enable'
    },
    'snapvault.repair_bug137685' => sub {
        'options.snapvault.repair_bug137685'
    },
    'snapvault.repair_bug137685_stats' => sub {
        'options.snapvault.repair_bug137685_stats'
    },
    'snapvault.suppress_bli_warnings' => sub {
        'options.snapvault.suppress_bli_warnings'
    },
    'snapvault.conf.qtree' => sub {
        'options.snapvaultconf.qtree.'.shift->dst_name
    },
};

=head2 B<Relationship Creation>

=head2 ::start

Creates a SnapVault relationship
and returns an object representing the relationship.

=over

=item Arguments

B<src> => I<Qtree object> | I<string>

I<Dependent.>
A Qtree object or a host path.

B<dst> => I<Qtree object> | I<string>

I<Dependent.>
A Qtree object or a host path.

Both arguments are required when C<::start> is called as a class method
and neither may be used when called via an object.

A host path is a full path name
prepended with the filer name.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<tries_count> => I<number>

I<Optional.>
Snapvault tries count for this relationship.

B<background> => C<0> | C<1>

I<Optional.>
Run the initial transfer in the background. The default is 0.

B<parallel> => C<0> | C<1>

I<Optional.>
Start the transfer and return. The background option is ignored
and the licenses are not checked to speed up start of transfer.
Use only when starting a lot of transfers concurrently. Default is 0.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Examples

=back

     $sv = NOTE::SnapVault->start(
         src => $src_qtree,
         dst => $dst_qtree,
     );

     $sv = NOTE::SnapVault->start(
         src        => 'filer_src:/vol/vol_src/qtree_src',
         dst        => 'filer_dst:/vol/vol_dst/qtree_dst',
         throttle   => 100,
         background => 1
     );

     $sv->start;

=cut

sub start {
    my ($class, %args) = @_;
    my $self = {};
    my (%sv_args, $ret, $op);

    # Most likely an object. note that 'ref $class' is not compared
    # to NOTE::SnapVault because this method can be inherited by a
    # subclass and 'ref $class' will not reflect 'NOTE::SnapVault'
    # in the subclass.
    if ( ref $class ne '' ) {
        $self = $class;

        if ( exists $args{src} || exists $args{dst} ) {
            croak "Cannot override existing 'src' or 'dst' attributes ".
                  "in existing object.";
        }
    } else {
        # Instantiate this object
        $self = NOTE::SnapVault->new( %args );
    }
    NOTE::reference( $self );

    delete $args{src};
    delete $args{dst};

    # Set flags for normal transfer
    $self->{_rxfer}    = 0;
    $self->{_released} = 0;
    $self->{_stopped}  = 0;

    $op = "Snapvault Start" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Override default timeout
    $self->timeout(%args);
    $args{timeout} = $self->{_timeout};

    if ( ! $args{parallel} ) {
        # Enable SnapVault on both filers: license, open access and enable
        $self->_license_and_enable;
    }

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Start snapvault relationship
    $ret = SnapVault::sv_start(
        $self->{_dst}->Filer->conn,
        $self->dst_path,
        SOURCE => $self->src_path,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $self;
}

=head2 ::attach

Creates an object representing a SnapVault relationship,
but does B<not> affect or cause the creation
of a SnapVault relationship on the filer.
If the source and destination given
are already in a SnapVault relationship with each other
then other C<NOTE::SnapVault> methods may be used
to control and manipulate the relationship
just as if L<C<::start>|/::start>
was used to create the relationship.
If the source and destination given
are not already in a SnapVault relationship with each other
the L<C<::start>|/::start> method
of the returned object may be used to create a relationship.

=over

=item Arguments

B<src> => I<Qtree object> | I<string>

I<Dependent.>
A Qtree object or a host path.

B<dst> => I<Qtree object> | I<string>

I<Dependent.>
A Qtree object or a host path.

Both arguments are required when C<::attach> is called as a class method
and neither may be used when called via an object.

A host path is a full path name
prepended with the filer name.

B<license> = C<0> | C<1>

I<Optional.>
License and enable both filers. The default is 0.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Example

=back

     $sv = NOTE::SnapVault->attach(
         src => $src_qtree,
         dst => $dst_qtree,
     );

     $sv = NOTE::SnapVault->attach(
         src => 'filer_src:/vol/vol_src/qtree_src',
         dst => 'filer_dst:/vol/vol_dst/qtree_dst',
         license => 1
     );

     $sv->attach(
         timeout => 600,
         license => 1
     );

=cut

sub attach {
    my ($class, %args) = @_;
    my $self = {};

    # Most likely an object. note that 'ref $class' is not compared
    # to NOTE::SnapVault because this method can be inherited by a
    # subclass and 'ref $class' will not reflect 'NOTE::SnapVault'
    # in the subclass.
    if ( ref $class ne '' ) {
        $self = $class;

        if ( exists $args{src} || exists $args{dst} ) {
            croak "Cannot override existing 'src' or 'dst' attributes ".
                  "in existing object.";
        }

        # Override default timeout
        $self->timeout(%args);
    } else {
        # Instantiate this object
        $self = NOTE::SnapVault->new( %args );
    }

    if ( $args{license} ) {
        # Enable SnapVault on both filers: license, open access and enable
        $self->_license_and_enable;
    }

    return $self;
}

=head2 ::convert

Converts the named replica destination into a SnapVault destination
and returns an object representing the relationship.

=over

=item Arguments

B<replica> => I<replica object>

I<Dependent.>
Named replica destination object
(e.g., L<C<NOTE::SnapMirror>|engineering::ontap_tests::main::docs::lib-NOTE-SnapMirror-pm/>).

Argument is required when C<::convert> is called as a class method
and should not be used when called via an object.

B<resync> => C<0> | C<1>

I<Optional.>
Resync the relationship after the conversion.
The default is 1.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Examples

=back

     $sv = NOTE::SnapVault->convert(
         replica => $sm_obj
     );

     $sv->convert;

=cut

sub convert {
    my ($class, %args) = @_;
    my $self = {};

    # Most likely an object. note that 'ref $class' is not compared
    # to NOTE::SnapVault because this method can be inherited by a
    # subclass and 'ref $class' will not reflect 'NOTE::SnapVault'
    # in the subclass.
    if ( ref $class ne '' ) {
        $self = $class;

        if ( exists $args{replica} ) {
            croak "Cannot override existing attributes in existing object.";
        }
    } else {
        if ( ! $args{replica}->dst->isa('NOTE::Qtree') ) {
            croak "SnapMirror object must represent a QSM relationship";
        }

        # Instantiate this object
        $self = NOTE::SnapVault->new(
            src => $args{replica}->src,
            dst => $args{replica}->dst,
        );
    }
    NOTE::reference( $self );

    # Set flags for normal transfer
    $self->{_rxfer}    = 0;
    $self->{_released} = 0;
    $self->{_stopped}  = 0;

    NOTE::logcomment( "Snapvault Convert" . $self->description );

    # Enable SnapVault on both filers: license, open access and enable
    $self->_license_and_enable(dst_enable=>0);

    # Disable SnapVault and SnapMirror
    $self->{_dst}->Filer->snapvault_disable;
    $self->{_dst}->Filer->snapmirror_disable;

    # Override default timeout
    $self->timeout(%args);

    # Convert named replica destination to a SnapVault relationship
    SnapVault::snapvault_convert(
        $self->{_dst}->Filer->conn,
        $self->{_dst}->mount_path
    );

    # Check for errors
    croak $@ if $@;

    # Enable SnapVault and SnapMirror
    $self->{_dst}->Filer->snapvault_enable;
    $self->{_dst}->Filer->snapmirror_enable;

    if ( !exists $args{resync} || $args{resync} ) {
        # Resync SnapVault relationship
        $self->resync;
    }

    return $self;
}

=head2 ::new

Creates an object representing a SnapVault relationship,
but does B<not> cause the creation
of a SnapVault relationship on the filer.
The L<C<::start>|/::start> method
of the returned object may be used to create the relationship.

This method should not normally be used;
use L<C<::start>|/::start> directly instead.

=over

=item Arguments

B<src> => I<Qtree object> | I<string>

I<Required.>
A Qtree object or a host path.

B<dst> => I<Qtree object> | I<string>

I<Required.>
A Qtree object or a host path.

A host path is a full path name
prepended with the filer name.

B<preprocess> => C<0> | C<1>

I<Optional.>
If true, verifies licenses on both filers.
Use only in conjunction with parallel option on L<C<::start>|/::start>
to create a lot of snapvault objects and start all transfers faster.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Example

=back

     $sv = NOTE::SnapVault->new(
         src => $src_qtree,
         dst => $dst_qtree,
     );

     $sv = NOTE::SnapVault->new(
         src => 'filer_src:/vol/vol_src/qtree_src',
         dst => 'filer_dst:/vol/vol_dst/qtree_dst',
     );

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    # Check arguments
    if ( !exists $args{src} || !exists $args{dst} ) {
        croak "Src or dst argument is missing";
    }

    # Set private variables for object
    $self->{_owner}   = $$;
    $self->{_timeout} = exists $args{timeout} ? $args{timeout} : 600;
    $self->{_src}     = _qtree_path_attach($args{src});
    $self->{_dst}     = _qtree_path_attach($args{dst});
    $self->{_Filer}   = $self->{_dst}->Filer; # needed by NOTE::Registry.pm

    # Bless this object
    bless $self, $class;

    # Make this object a tie variable so registry options can be accessed
    # via a hash element of the object
    tie %{$self}, 'NOTE::Registry::Slice', $self, $regSlice;

    # Add the default schedule to cleanup list
    push @{$self->{_schedules}}, {};
    $self->{_schedules}[-1]{':'.$self->dst->hostpath} = {
        volume => $self->dst->Volume,
        source => 0,
    };

    if ( $args{preprocess} ) {
        # Enable SnapVault on both filers: license, open access and enable
        $self->_license_and_enable;
    }

    return $self;
}

=head2 B<SnapVault Commands>

The methods names in this section correspond to
ONTAP's C<snapvault> subcommands.
Only administrator privilege level subcommands are listed in this section.
These methods eventually call the self-named subcommands.
Only the briefest mention of what these subcommands do is given here.
See the SnapVault documentation for more information on them.

SnapVault commands that apply to the filer as a whole are part of
L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>
and are described in the
L<C<NOTE::Filer::SnapVault>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/snapvault>
section.

=head2 ::abort

Aborts an active SnapVault transfer on the destination.
If this function is called after L<C<::restore>|/::restore>, it aborts the
restore transfer on the source.

=over

=item Arguments

B<hard> => C<0> | C<1>

I<Optional.>
If set to 1, does a hard abort, such that the transfer is not restartable.
The default is 0.

B<source> => C<0> | C<1>

I<Optional.>
If set to 1, aborts from the source filer. If set to 0, aborts from the destination
filer. If not specified, the default is 0 if called before L<C<::restore>|/::restore>,
and 1 if called after L<C<::restore>|/::restore>.

=item Examples

=back

     $sv->abort;

     $sv->abort(
         hard => 1
     );

     $sv->abort( source => 1 );

=cut

sub abort {
    my ($self, %args) = @_;
    my (%sv_args, $ret);
    my $qtree = $self->{_dst};
    my $source = delete $args{source};

    NOTE::logcomment( "Snapvault Abort" . $self->description );

    # Check if coming from a restore
    if ( $self->{_rxfer} && exists($self->{_restore}) ) {
        # Abort restore transfer on the source
        $qtree = $self->{_restore};
        $source = 1 unless defined($source);

        # Save qtree for restore restart
        $self->{_restore_restart} = $self->{_restore};
        delete $self->{_restore};
    }

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Abort transfer
    $ret = SnapVault::snapvault_abort(
        { conn => $qtree->Filer->conn,
	  source => defined $source ? $source : 0
	},
        $qtree->hostpath,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    return $ret;
}

=head2 ::stop

Stop SnapVault relationship and unconfigure the destination.
Returns a 'subtest' object, which may be ignored if the stop completes
before the call to C<::stop> returns.

=over

=item Arguments

B<background> => C<0> | C<1>

I<Optional.>
Run the stop command in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Example

=back

     $sv->stop;

     $subtest = $sv->stop(
         background => 1,
     );

=cut

sub stop {
    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);

    $op = "Snapvault Stop" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Unconfigure destination
    $ret = SnapVault::sv_stop(
        $self->{_dst}->Filer->conn,
        $self->dst_path,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    # Set flag that this relationship has already been stopped
    $self->{_stopped} = 1;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $ret;
}

=head2 ::release

Releases the source-side resources being held
for a particular SnapVault destination.

=over

=item Arguments

B<restore> => C<0> | C<1>

I<Optional.>
Release the destination-side resources of a restore relationship.
The default is 0.

=item Example

=back

     $sv->release;

     $sv->release(
         restore => 1
     );

=cut

sub release {
    my ($self, %args) = @_;
    my $ret;
    my $filer = $self->{_src}->Filer;
    my $src_name = $self->src_name;
    my $dst_path = $self->dst_path;

    NOTE::logcomment( "Snapvault Release" . $self->description );

    if ( $args{restore} ) {
        croak "Option 'restore' is only valid after calling restore"
          if ! $self->{_restore};

        $filer = $self->{_dst}->Filer;
        $src_name = $self->dst_name;
        $dst_path = $self->{_restore}->hostpath;
    }

    # Release the source-side resources for snapvault
    $ret = SnapVault::snapvault_release(
        { conn => $filer->conn, source => $args{restore} ? 0 : 1 },
        $src_name,
        $dst_path,
    );

    # Check for errors
    croak $@ if $@;

    # Set flag that this relationship has already been released
    $self->{_released} = 1;

    return $ret;
}

=head2 ::status

Returns a hash reference to the status of the relationship.

=over

=item Arguments

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the relationship from the source filer instead.
The default is 0 (destination filer).

B<restore> => C<0> | C<1>

I<Optional.>
Get the status of the restore transfer.
The default is 0.

=item Examples

=back

     $status = $sv->status;   # Get status from destination filer

     $status = $sv->status(
         source => 1          # Get status from source filer
     );

     $status = $sv->status(
         restore => 1         # Get restore status from destination filer
     );

     $status = $sv->status(
         source => 1,
         restore => 1         # Get restore status from source filer
     );

=cut

sub status {
    my ($self, %args) = @_;
    my $status;
    my $filer = $args{source} ? $self->{_src}->Filer : $self->{_dst}->Filer;
    my $dst_path = $self->dst_path;
    my $src_path = $self->src_path;

    if ( $args{restore} ) {
        croak "Option 'restore' is only valid after calling restore"
          if ! $self->{_restore};

        $dst_path = $self->{_restore}->hostpath;
        $src_path = $self->dst_path;
    }

    # Get status of snapvault relationship
    $status = SnapVault::snapvault_status_detailed(
        { conn => $filer->conn, source => $args{source} },
        $dst_path,
        $src_path
    );

    # The above function may return undef with no error
    # so croak on error if it exists
    croak $@ if $@;

    return $status;
}

=head2 ::configuration

Returns a hash reference to the configuration of the relationship.

=over

=item Arguments

None.

=item Examples

=back

     $conf = $sv->configuration;

=cut

sub configuration {
    my ($self, %args) = @_;
    my %config;

    # Get configuration of snapvault relationship
    %config = SnapVault::snapvault_status_config(
        $self->{_dst}->Filer->conn,
        dstpath => $self->dst_name,
        source  => $self->src_path
    );

    # Check for errors
    croak $@ if $@;

    return %config ? \%config : undef;
}

=head2 ::modify

Modifies the existing SnapVault configuration for the relationship.

=over

=item Arguments

B<src> => I<Qtree object> | I<string>

I<Optional.>
New source qtree or host path.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<tries_count> => I<number>

I<Optional.>
Snapvault tries count for this relationship.

B<back_up_open_files> => C<0> | C<1>

I<Optional.>
Allow the inclusion of open files on the primary system at the time
of the transfer.

=item Examples

=back

     $sv->modify(
         throttle => 100,
     );

=cut

sub modify {
    my ($self, %args) = @_;
    my (%sv_args, $ret);

    # Translate options to proper arguments for function
    # This function is not consistent with the rest of the
    # functions in SnapVault.pm
    $sv_args{kbs}     = $args{throttle}    if exists $args{throttle};
    $sv_args{tries}   = $args{tries_count} if exists $args{tries_count};
    $sv_args{back_up_open_files} = $args{back_up_open_files}
      if exists $args{back_up_open_files};

    # Change the source qtree if option is given
    if ( exists($args{src}) ) {
        $self->{_src} = _qtree_path_attach($args{src});
    }
    $sv_args{source} = $self->src_path;

    # Modify configuration of snapvault relationship
    $ret = SnapVault::snapvault_modify(
        $self->{_dst}->Filer->conn,
        $self->dst_name,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    return $ret;
}

=head2 ::update

Starts an incremental transfer for a SnapVault relationship.
Returns a 'subtest' object, which may be ignored if the transfer completes
before the call to C<::update> returns.

=over

=item Arguments

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the source to synchronize from.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<background> => C<0> | C<1>

I<Optional.>
Run the update transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->update;

     $subtest = $sv->update(
         background => 1,
         throttle   => 100,
     );

=cut

sub update {
    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);

    # Set flag for normal transfer
    $self->{_rxfer} = 0;

    $op = "Snapvault Update" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Start snapvault update
    $ret = SnapVault::sv_update(
        $self->{_dst}->Filer->conn,
        $self->dst_path,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $ret;
}

=head2 ::restore

Starts a baseline restore transfer for a SnapVault relationship.
Returns a 'subtest' object, which may be ignored if the transfer completes
before the call to C<::restore> returns.

=over

=item Arguments

B<dst> => I<Qtree object> | I<string>

I<Dependent.>
A Qtree object or host path. This is where the data will be restored to
so this will become the new primary (src) for the SnapVault relationship
after calling L<C<::resync>|/::resync>.

Argument is optional when restarting a previously aborted restore,
otherwise it is required.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the destination to synchronize from.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<background> => C<0> | C<1>

I<Optional.>
Run the restore transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->restore(
         dst => new_qtree_src,
     );

     $subtest = $sv->restore(
         dst        => new_qtree_src,
         throttle   => 100,
         background => 1,
     );

     # Restart previously aborted restore
     $sv->restore;

=cut

sub restore {
    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);

    # Set flag for normal transfer
    $self->{_rxfer} = 1;

    if ( exists($self->{_restore_restart}) ) {
        # Restart restore on same qtree
        $args{dst} = $self->{_restore_restart} unless exists $args{dst};
        delete $self->{_restore_restart};
    }

    croak "Argument dst must be specified" unless exists $args{dst};

    # The dst argument will become the new primary when resync is called
    $self->{_restore} = _qtree_path_attach($args{dst});
    delete $args{dst};

    $op = "Snapvault Restore" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Start snapvault restore
    $ret = SnapVault::sv_restore(
        $self->{_restore}->Filer->conn,
        $self->{_restore}->hostpath,
        SOURCE => $self->dst_path,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $ret;
}

=head2 ::resync

Starts a resync-rollback transfer for a SnapVault relationship.
Returns a 'subtest' object, which may be ignored if the transfer completes
before the call to C<::resync> returns.

=over

=item Arguments

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<tries_count> => I<number>

I<Optional.>
Snapvault tries count for this relationship.

B<background> => C<0> | C<1>

I<Optional.>
Run the resync transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->resync;

     $subtest = $sv->resync(
         background => 1,
         throttle   => 100,
     );

=cut

sub resync {
    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);

    # Check if coming from a restore
    if ( exists($self->{_restore}) ) {
        # The qtree where data was restored to becomes the new primary
        $self->{_src} = $self->{_restore};
        delete $self->{_restore};
    }

    $op = "Snapvault Resync" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Start snapvault resync
    $ret = SnapVault::sv_resync(
        $self->{_dst}->Filer->conn,
        $self->dst_path,
        SOURCE => $self->src_path,
        %sv_args
    );

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    # Create another entry on cleanup list
    push @{$self->{_schedules}}, {};

    return $ret;
}

=head2 ::dst_snapshot

Returns the name of the snapshot being retained for the given
source/destination pair on the source filer.

=over

=item Arguments

B<restore> => C<0> | C<1>

I<Optional.>
Returns the snapshot for the restore relationship on the destination filer
instead. The default is 0.

=item Examples

=back

     $snapshot = $sv->dst_snapshot;

     $snapshot = $sv->dst_snapshot(
         restore => 1
     );

=cut

sub dst_snapshot {
    my ($self, %args) = @_;
    my $snapshot;
    my $filer = $self->{_src}->Filer;
    my $src_path = $self->src_path;
    my $dst_path = $self->dst_path;

    if ( $args{restore} ) {
        croak "Option 'restore' is only valid after calling restore"
          if ! $self->{_restore};

        $filer = $self->{_dst}->Filer;
        $src_path = $self->dst_path;
        $dst_path = $self->{_restore}->hostpath;
    }

    # Get the snapshot for relationship on the source filer
    $snapshot = SnapVault::snapvault_dst_snap(
        { conn => $filer->conn, source => $args{restore} ? 0 : 1 },
        $src_path,
        $dst_path,
    );

    # Check for errors
    croak $@ if $@;

    return $snapshot;
}

=head2 ::snap_sched

Configures or lists the SnapVault backup schedule for either the destination
or the source volume.

=over

=item Arguments

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot to configure or list.

B<sched> => I<string> | I<schedule_info>

I<Optional.>
Schedule to set for the given snapshot.

B<xfer> => C<0> | C<1>

I<Optional.>
Configure a 'xfer' snapshot instead of a 'create' snapshot. The default is 0.

B<source> => C<0> | C<1>

I<Optional.>
Configure the source volume. The default is 0 (destination volume).

=item Examples

=back

     # List all backup schedules for the destination volume
     @schedules = $sv->snap_sched;

     # Get the schedule entry for the 'snapshot1' snapshot on the source volume
     $sched_h = $sv->snap_sched(
         source   => 1,
         snapshot => 'snapshot1',
     );

     # Set the schedule for the 'snapshot1' snapshot on the destination volume
     $sv->snap_sched(
         snapshot => 'snapshot1',
         sched    => '2@1,2'
     );

=cut

sub snap_sched {
    my ($self, %args) = @_;
    my (@opts, @schedules, $sched_h);
    my $qtree = $args{source} ? $self->{_src} : $self->{_dst};

    # Check for proper options
    if ( exists($args{sched}) && ! exists($args{snapshot}) ) {
        croak "Option 'sched' requires 'snapshot' to be defined";
    }

    if ( exists($args{xfer}) && ! exists($args{sched}) ) {
        croak "Option 'xfer' requires 'sched' to be defined";
    }

    # Add fixed parameters
    @opts = (
        { conn => $qtree->Filer->conn, source => $args{source} },
        $qtree->Volume->name,
    );

    # Add optional parameters
    push @opts, $args{snapshot}       if exists $args{snapshot};
    push @opts, $args{sched}          if exists $args{sched};
    push @opts, ($args{xfer} ? 1 : 0) if exists $args{xfer};

    # Configure or list the backup schedule
    # Use the '&' to avoid prototype of function so the
    # optional array '@opts' is not evaluated in a scalar context
    if ( exists($args{snapshot}) ) {
        # Must be evaluated in scalar context to get a hash reference
        $sched_h = &SnapVault::snapvault_snap_sched(@opts);
    } else {
        @schedules = &SnapVault::snapvault_snap_sched(@opts);
    }

    # Check for errors
    croak $@ if $@;

    if ( exists $args{sched} ) {
        # Add schedule to cleanup hash
        $self->{_schedules}[-1]{$args{snapshot}.':'.$qtree->hostpath} = {
            volume => $qtree->Volume,
            source => $args{source},
        };
    }

    if ( exists($args{snapshot}) ) {
        return $sched_h;
    } else {
        return @schedules;
    }
}

=head2 ::snap_unsched

Unconfigures the SnapVault backup schedule for the given snapshot on either
the destination or the source volume.

=over

=item Arguments

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot to unconfigure.

B<source> => C<0> | C<1>

I<Optional.>
Unconfigure the source volume snapshot. The default is 0 (destination volume).

=item Examples

=back

     # Unconfigure snapshot on destination
     $sv->snap_unsched(
         snapshot => 'snapshot1',
     );

     # Unconfigure snapshot on source
     $sv->snap_unsched(
         source   => 1,
         snapshot => 'snapshot1',
     );

     # Unconfigure all snapshots created by this object
     $sv->snap_unsched;

=cut

sub snap_unsched {
    my ($self, %args) = @_;
    my ($item, $key, $snapname, %removed, $ret);
    my $qtree = $args{source} ? $self->{_src} : $self->{_dst};

    if ( exists $args{snapshot} ) {
        # Unconfigure the backup schedule
        $ret = SnapVault::snapvault_snap_unsched(
            { conn => $qtree->Filer->conn, source => $args{source} },
            $qtree->Volume->name,
            $args{snapshot},
        );

        # Check for errors
        croak $@ if $@;

        # Remove schedule from cleanup hash
        delete $self->{_schedules}[-1]{$args{snapshot}.':'.$qtree->hostpath};
    } else {
        # Delete all schedules created by this object
        while ( @{$self->{_schedules}} ) {
            $item = $self->{_schedules}[0];
            while ( $key = each %$item ) {
                if ( ! $removed{$key} ) {
                    ($snapname) = ($key =~ /([^:]*):/);
                    # Delete schedule and ignore error
                    SnapVault::snapvault_snap_unsched(
                        { conn   => $item->{$key}->{volume}->Filer->conn,
                          source => $item->{$key}->{source} },
                        $item->{$key}->{volume}->name,
                        $snapname
                    );
                    croak $@ if ($@ && !$args{ignore_errors});

                    # Keep track of schedules already removed
                    $removed{$key} = 1;
                }
                delete $item->{$key};
            }
            splice(@{$self->{_schedules}}, 0, 1);
        }
        $ret = 1;
    }

    return $ret;
}

=head2 ::snap_create

Creates a configured SnapVault snapshot schedule on either
the destination or the source volume.
Returns a 'subtest' object, which may be ignored if the command completes
before the call to C<::snap_create> returns.

=over

=item Arguments

B<snapshot> => I<string>

I<Required.>
The name of the configured snapshot to create.

B<source> => C<0> | C<1>

I<Optional.>
Create the snapshot on the source volume.
The default is 0 (destination volume).

B<background> => C<0> | C<1>

I<Optional.>
Run the command in the background.
Only used when snapshot is created on the destination. The default is 0.

B<timeout> => I<number>

I<Optional.>
Timeout period waiting for the command to complete.
Only used when snapshot is created on the destination.
The default is object's default timeout.

=item Examples

=back

     # Create snapshot on the destination
     $sv->snap_create(
         snapshot => 'snapshot1',
     );

     # Create snapshot on the destination (background)
     $subtest = $sv->snap_create(
         snapshot   => 'snapshot1',
         background => 1
     );

     # Create snapshot on the source
     $sv->snap_create(
         snapshot => 'snapshot1',
         source   => 1,
     );

=cut

sub snap_create {
    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);
    my $snapshot = $args{snapshot};

    # Set flag for normal transfer
    $self->{_rxfer} = 0;

    $op = "Snapvault Snapshot Create\n    +---" .  $args{snapshot} . ' on ' .
             ($args{source} ? $self->{_src}->Volume->description
                            : $self->{_dst}->Volume->description);
    NOTE::logcomment( $op );

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    if ( $args{source} ) {
        # Create snapshot on the source volume
        $ret = SnapVault::snapvault_snap_create(
            $self->{_src}->Filer->conn,
            $self->{_src}->Volume->name,
            $snapshot,
        );

        # Check for errors
        croak $@ if $@;
    } else {
        delete $args{snapshot};
        delete $args{source};

        # Process background option
        if ( exists $args{background} ) {
            $args{complete} = $args{background} ? 0 : 1;
            delete $args{background};
        } else {
            $args{complete} = 1;
        }

        # Convert all arguments to upper case
        %sv_args = map { uc($_) => $args{$_} } keys %args;

        # Create snapshot on the destination volume
        $ret = SnapVault::sv_snap_create(
            $self->{_dst}->Filer->conn,
            $self->{_dst}->Volume->name,
            $snapshot,
            %sv_args,
        );

        # Check for errors
        croak $@ if $@;
    }

    # Save destination snapshot for use with ::snap_abort
    $self->{_snapshot} = $snapshot;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $ret;
}

=head2 ::snap_status

Returns a list of configured SnapVault snapshot targets for the destination
or source volume of the relationship or a hash reference of the schedule if
the 'snapshot' argument is specified.

=over

=item Arguments

B<snapshot> => I<string>

I<Optional.>
The name of the configured snapshot.

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the snapshot from the source filer instead.
The default is 0 (destination filer).

=item Examples

=back

     # Get list of snapshot targets for destination volume
     @schedules = $sv->snap_status;

     # Get the status info of the specified snapshot target on the source
     $status = $sv->snap_status(
         snapshot => 'snap_name',
         source => 1
     );

=cut

sub snap_status {
    my ($self, %args) = @_;
    my (%sv_args, @ret);
    my $qtree = $args{source} ? $self->{_src} : $self->{_dst};

    $sv_args{SNAPSHOT} = $args{snapshot} if exists $args{snapshot};

    # Get status of snapvault relationship
    @ret = SnapVault::snapvault_status_snapshot(
        { conn => $qtree->Filer->conn, source => $args{source} },
        VOLUME => $qtree->Volume->name,
        %sv_args
    );

    # Check for error
    croak $@ if $@;

    return (exists $args{snapshot} ? $ret[0] : @ret);
}

=head2 ::snap_abort

Aborts the creation of an active snapshot target.

=over

=item Arguments

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot to abort.
The default is the snapshot specified in the last call to
L<C<::snap_create>|/::snap_create>.

B<source> => C<0> | C<1>

I<Optional.>
Unconfigure the source volume snapshot. The default is 0 (destination volume).

=item Examples

=back

     # Abort the snapshot create previously created by ::snap_create
     $sv->snap_abort;

     # Abort the snapshot create on the source
     $sv->snap_abort(
         snapshot => 'snapshot1',
         source   => 1
     );

=cut

sub snap_abort {
    my ($self, %args) = @_;
    my $ret;
    my $snapshot = $args{snapshot} || $self->{_snapshot};
    my $qtree = $args{source} ? $self->src : $self->dst;

    croak "Snapshot name must be specified" unless $snapshot;

    # Abort the snapshot create
    $ret = SnapVault::snapvault_abort_snapshot(
        { conn => $qtree->Filer->conn, source => $args{source} },
        $qtree->Volume->name,
        $snapshot,
    );

    # Check for errors
    croak $@ if $@;

    return $ret;
}

=head2 B<Checker/Fixer>

=head2 ::check

Checks the consistency of the SnapVault relationship.

The destination must exist and be in a 'snapvaulted' or 'quiesced' state.
The checker process runs on the destination filer.

Returns a 'subtest' object if background => 1.

=over

=item Arguments

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<check_level> => C<data> | C<checksum>

I<Optional.>
Specify data comparison level as either data or checksum.
The default is C<data>.

B<quick_mode> => C<0> | C<1>

I<Optional.>
If 1, Starts C<snapvault check> in quick mode.
The default is 0.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the source to use.
By default the base snapshot for the SnapVault relationship will be used.

B<ignore> => C<ctime>

I<Optional.>
Ignores some type of comparisons during the check session. At present only
C<ctime> can be ignored.

B<background> => C<0> | C<1>

I<Optional.>
If 1, returns while checking is still running in the background.
The result of the check must be checked later using
L<C<::wait>|/::wait> or the C<wait> method of the returned object.
If the status is not checked using one of the wait methods
any errors in the check will go unnoticed.
The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->check;

     $subtest = $sv->check(
         background => 1
     );

=cut

sub check {
    my ($self, %args) = @_;
    my (%sv_args, $item, $ret, $op);
    my %tr_args = (
        fix         => 'VERIFIER_FIX',
        quick_mode  => 'VERIFIER_QUICK',
        check_level => 'VERIFIER_CKSUM',
        snapshot    => 'VERIFIER_SNAPSHOT',
        ignore      => 'VERIFIER_IGNORE',
    );

    $op = "Snapvault Check" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{complete};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Process the check_level option
    $args{check_level} = 'data' if ! exists $args{check_level};

    # Translate arguments
    foreach $item ( keys %args ) {
        if ( exists $tr_args{$item} ) {
            $args{$tr_args{$item}} = $args{$item};
            delete $args{$item};
        }
    }

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Run the snapvault check
    $ret = SnapVault::snapvault_check(
        $self->dst->Filer->conn,
        $self->dst_path,
        SOURCE => $self->src_path,
        %sv_args,
    );

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $ret;
}

=head2 ::fix

Checks the consistency of the SnapVault relationship
while repairing any inconsistencies.

The destination must exist and be in a 'snapvaulted' or 'quiesced' state.
The fixer process runs on the destination filer.

Returns a 'subtest' object if background => 1.

=over

=item Arguments

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<check_level> => C<data> | C<checksum>

I<Optional.>
Specify data comparison level as either data or checksum.
The default is C<data>.

B<quick_mode> => C<0> | C<1>

I<Optional.>
If 1, Starts C<snapvault check> in quick mode.
The default is 0.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the source to use.
By default the base snapshot for the SnapVault relationship will be used.

B<ignore> => C<ctime>

I<Optional.>
Ignores some type of comparisons during the check session. At present only
C<ctime> can be ignored.

B<background> => C<0> | C<1>

I<Optional.>
If 1, returns while checking is still running in the background.
The result of the check must be checked later using
L<C<::wait>|/::wait> or the C<wait> method of the returned object.
If the status is not checked using one of the wait methods
any errors in the check will go unnoticed.
The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->fix;

     $subtest = $sv->fix(
         background => 1
     );

=cut

sub fix {
    my ($self, %args) = @_;
    return $self->check(fix => 1, %args);
}

=head2 ::check_abort

Aborts a currently running snapvault check process.
Does not abort currently running SnapVault transfers.

=over

=item Arguments

B<source> => C<0> | C<1>

I<Optional.>
Abort the check from the source filer instead.
The default is 0 (destination filer).

=item Example

=back

     $sm->check_abort;

     $sm->check_abort(
         source => 1,
     );

=cut

sub check_abort {
    my ($self, %args) = @_;
    my $ret;
    my $qtree = $args{source} ? $self->src : $self->dst;

    # Run the snapvault check abort
    $ret = SnapVault::snapvault_check_abort(
        $qtree->Filer->conn,
        $self->dst_path,
    );

    # Check for errors
    croak $@ if $@;

    return $ret;
}

=head2 B<Waiting>

=head2 ::wait

Waits for all background processes to complete.
These are the processes created by using the B<background> option of
L<C<::start>|/::start>, L<C<::stop>|/::stop>, L<C<::update>|/::update>,
L<C<::restore>|/::restore>, L<C<::resync>|/::resync>
or L<C<::snap_create>|/::snap_create>.

This method is completely unrelated to L<C<::status_wait>|/::status_wait>.

=over

=item Arguments

None.

=item Example

=back

     $sv->wait;

=cut

sub wait {
    my $self = shift;
    my ($obj, $test, $op);

    while ( $test = pop @{$self->{_subtests}} ) {
        $op  = $test->{operation};
        $obj = $test->{subtest};
        if ( $obj->is_alive() ) {
            NOTE::logcomment( "Waiting for: ".$op );
        }
        $obj->wait_finish();
        croak $@ unless FSR::FSRutils::subtestResults($obj);
    }

    return;
}

=head2 ::status_wait

Wait for the SnapVault relationship to reach specific states.
Returns when all of the specified status elements
have their respective values I<at the same time>.
(Hint: don't use element/value combinations
that are mutually exclusive.)

Note that there are many SnapVault status elements,
one of which is thoughtfully named 'status'.
Another one is named 'state'.
There are none however named 'element' (not yet, anyway).

The status elements checked are the same ones
that may appear in the hash generated by L<C<::status>|/::status>.
(Internally C<::status> is used; this is a convenience method.)

Because you are a valued NOTE user,
the values for the status elements may be given
as regular expressions.

This method is completely unrelated to L<C<::wait>|/::wait>.

=over

=item Arguments

I<status element name> => I<status element value> | I<regex>

I<Required.>
The status elements to check are given as named arguments.
Their values may be given as literal strings
or regular expressions that (hopefully) match valid values.

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the relationship from the source filer instead.
The default is 0 (destination filer).

B<restore> => C<0> | C<1>

I<Optional.>
Get the status of the restore transfer.
The default is 0.

B<ignore_transfer_error> => C<0> | C<1>

I<Optional.>
If 1, don't fail if the status contains the C<current-transfer-error> element.
The default is 0.

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 5 seconds.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->status_wait(
         status => 'idle'
     );

     $sv->status_wait(
         status => 'transferring'
     );

     $sv->status_wait(
         'base-snapshot' => '\w'
     );

=cut

sub status_wait {
    my ($self, %args) = @_;
    my (%opts, $wait_for, $matchall, $havekeys, $status, $etime, $delta);
    my $sleepis = 5;    # sleep number of seconds between status read
    my $noerror = 0;    # croak by default if 'current-transfer-error'

    $opts{source} = exists $args{source} ? $args{source} : 0;
    delete $args{source};
    $opts{restore} = exists $args{restore} ? $args{restore} : 0;
    delete $args{restore};

    if( exists $args{'timeout'} ) {
        $etime = $args{'timeout'};
        delete $args{'timeout'};
    } else {
        $etime = $self->{_timeout};
    }

    if( exists $args{'sleep'} ) {
        $sleepis = $args{'sleep'};
        delete $args{'sleep'};
    }

    if( exists $args{'ignore_transfer_error'} ) {
        $noerror++;
        delete $args{'ignore_transfer_error'};
    }

    $wait_for = ''; # Show what we are waiting for
    foreach( keys %args ) { $wait_for .= "$_ =~ /$args{$_}/\n" }

    $matchall = 0;
    $havekeys = scalar( keys %args );

    # Set expiration time
    $etime += time;

    while( $matchall < $havekeys ) {
        NOTE::logcomment("status_wait:\n$wait_for");

        croak "Status Timeout waiting for:\n$wait_for" if time > $etime;

        $status = $self->status(%opts);
        $matchall = 0;

        if( $status->{'current-transfer-error'} && $noerror < 1 ) {
            croak $status->{'current-transfer-error'};
        }

        # check keys one by one and match all of them
        foreach( keys %args ) {
            $matchall++ if( $status->{$_} =~ qr/$args{$_}/i );
        }

        # Do not wait anymore if everything has been found
        last unless ( $matchall < $havekeys );

        # Wait for next iteration
        $delta = $etime - time;
        $delta = 0 if $delta < 0;
        $sleepis = $delta if $delta < $sleepis;
        sleep $sleepis if( $sleepis > 0 );
    }

    return;
}

=head2 ::dst_idle_wait

Waits for the destination and all snapshot targets in its volume to
become idle.

=over

=item Arguments

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 5 seconds.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->dst_idle_wait;

=cut

sub dst_idle_wait {
    my ($self, %args) = @_;
    my (%sv_args, $ret);
    $args{delay}    = 5;
    $args{complete} = 1;

    # Process the sleep option
    if( exists $args{'sleep'} ) {
        $args{delay} = $args{'sleep'};
        delete $args{'sleep'};
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    $ret = SnapVault::sv_wait_idle(
        { conn => $self->dst->Filer->conn, source => 0 },
        $self->dst_path,
        %sv_args,
    );

    # Check for error
    croak $@ if $@;

    return $ret;
}

=head2 B<Component Objects>

=head2 ::src

Returns the Qtree object representing the current source.

=over

=item Arguments

None.

=item Examples

=back

     $src = $sv->src;

=cut

sub src {
    my $self = shift;
    return $self->{_src};
}

=head2 ::dst

Returns the Qtree object representing the current destination.

=over

=item Arguments

None.

=item Examples

=back

     $dst = $sv->dst;

=cut

sub dst {
    my $self = shift;
    return $self->{_dst};
}

=head2 B<Names and Paths>

=head2 ::src_name

Returns the name of the source Qtree.

=over

=item Arguments

None.

=item Example

=back

     $srcname = $sv->src_name;

=cut

sub src_name {
    my $self = shift;
    return $self->{_src}->mount_path;
}

=head2 ::dst_name

Returns the name of the destination Qtree.

=over

=item Arguments

None.

=item Example

=back

     $dstname = $sv->dst_name;

=cut

sub dst_name {
    my $self = shift;
    return $self->{_dst}->mount_path;
}

=head2 ::src_path

Returns the host-qualified path name
of the source Qtree.

=over

=item Arguments

None.

=item Example

=back

     $srcpath = $sv->src_path;

=cut

sub src_path {
    my $self = shift;
    return $self->{_src}->hostpath;
}

=head2 ::dst_path

Returns the host-qualified path name
of the destination Qtree.

=over

=item Arguments

None.

=item Example

=back

     $dstpath = $sv->dst_path;

=cut

sub dst_path {
    my $self = shift;
    return $self->{_dst}->hostpath;
}

=head2 B<Object Identification>

=head2 ::description

Returns a description of the SnapVault relationship
consisting of the tag for the source,
a rightward pointing arrow,
and the tag for the destination.

=over

=item Arguments

None.

=item Example

=back

     $desc = $sv->description;

=cut

sub description {
    my $self = shift;
    my $desc;

    if ( $self->{_rxfer} && exists($self->{_restore}) ) {
        $desc = "\n    +---" . $self->dst->description .
                "\n    +-> " . $self->{_restore}->description;
    } else {
        $desc = "\n    +---" . $self->src->description .
                "\n    +-> " . $self->dst->description;
    }

    return $desc;
}

=head2 B<Object Configuration>

=head2 ::timeout

Sets or returns the current default timeout for the object before being set.
A value of -1 is to wait forever.

=over

=item Arguments

B<timeout> => I<number>

I<Optional.>
Set default timeout for the object.

=item Example

=back

     $curr_timeout = $sv->timeout;    # Get current default timeout

     $orig_timeout = $sv->timeout(
         timeout => 300               # Set default timeout
     );

=cut

sub timeout {
    my ($self, %args) = @_;

    $self->{_timeout} = $args{timeout} if exists $args{timeout};
    $self->{_timeout} = -1 if ! defined $self->{_timeout};
    return $self->{_timeout};
}

=head2 B<Object Destruction>

=head2 ::destroy

Destroy the SnapVault relationship by unconfiguring both the source
and destination and removing the registry entries for the relationship.
It also deletes all snapshot schedules created by this object.

=over

=item Arguments

None.

=item Example

=back

     $sv->destroy;

=cut

sub destroy {
    my $self = shift;
    my $conf;
    $self->{_destroyed} = 1;

    NOTE::logcomment( "Destroying Snapvault " . $self->description );

    # Check if this is a valid relationship
    $conf = $self->configuration;

    # Abort any transfer currently active, ignore error
    eval {
        if ( $conf && $self->abort(hard => 1) ) {
            $self->status_wait(
                status => 'idle',
                ignore_transfer_error => 1,
            );
        }
    };

    # Delete all schedules created by this object
    eval { $self->snap_unsched(ignore_errors => 1) };

    # Unconfigure destination
    $self->stop unless $self->{_stopped};

    # Release the source-side resources for snapvault
    $self->release unless $self->{_released};

    # Delete registry entry for this relationship
    delete $self->{'snapvault.conf.qtree'};

    NOTE::logcomment("Snapvault relationship destroyed");

    return;
}

# Internal function to license, enable and open access for SnapVault
sub _license_and_enable {
    my ($self, %args) = @_;
    my $dst_enable = exists($args{dst_enable}) ? $args{dst_enable} : 1;

    # Enable SnapVault licenses on both source and destination
    $self->{_src}->Filer->check_license( "sv_ontap_pri" );
    $self->{_dst}->Filer->check_license( "sv_ontap_sec" );

    # Open SnapVault access to all on both filers
    $self->{_Filer} = $self->{_src}->Filer;
    $self->{'snapvault.access'} = 'all';
    $self->{_Filer} = $self->{_dst}->Filer;
    $self->{'snapvault.access'} = 'all';

    # Enable SnapVault on both filers
    $self->{_src}->Filer->snapvault_enable;
    $self->{_dst}->Filer->snapvault_enable if $dst_enable;

    return;
}

# Internal function which takes as argument a Qtree object or a qtree
# host path string and returns the Qtree object attached to the qtree
sub _qtree_path_attach {
    my $qtree = shift;
    my ($qtree_obj, $filer_name, $vol_name, $qtree_name);

    if ( ref($qtree) ) {
        if ( ! $qtree->isa('NOTE::Qtree') ) {
            croak "Argument must be a Qtree object: <" . ref($qtree) . ">";
        }
        # It is already a Qtree object
        $qtree_obj = $qtree;
    } elsif ( $qtree =~ /([^:]+):\/vol\/([^\/]+)\/(.*)/ ) {
        # This is a host path so create the Qtree object
        ($filer_name, $vol_name, $qtree_name) = ($1, $2, $3);
        # Attach the qtree
        $qtree_obj = NOTE::Qtree->attach(
            Filer  => { host => $filer_name },
            Volume => { name => $vol_name },
            name   => $qtree_name,
            force  => 1, # don't croak if qtree not present
        );
    } else {
        croak "Invalid qtree path: $qtree";
    }

    return $qtree_obj;
}

# Object destructor
sub DESTROY {
    my $self = shift;

    # Only the creator can destroy the object
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_spare_me} );    # Means don't destroy
    return if ( $self->{_destroyed} );   # Already dead
    return if ( NOTE->ALL_SPARED );

    # Destroy object
    $self->destroy;
}

=head1 SEE ALSO

L<C<NOTE::Filer::SnapVault>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/snapvault>

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::main::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>

=head1 STATUS

Review

$Id: //depot/prod/test/Rsierranevada/lib/GNOTE/7G/SnapVault.pm#1 $

$Date: 2013/01/22 $

=head1 AUTHOR

Author information is in the
L<NOTE overview|engineering::ontap_tests::main::docs::lib-NOTE-pm/::author>.

=cut

1;

