#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/7G/Volume/Clone.pm#1 $
#

## @summary  NOTE Volume::Clone module
## @author   dl-dpgqa-automation@netapp.com
## @status   Public
## @pod      here

=head1 DESCRIPTION

This is the NOTE class for working with volume clones

=head1 METHODS

=cut

package NOTE::7G::Volume::Clone;
use GNOTE;
use GNOTE::7G::Registry::Slice;
our @ISA = (NOTE::7G::Volume NOTE::7G NOTE::Volume);
use strict;
use Carp qw(croak);

use Volume::Clone qw();

=head2 ::new

Clone constructor. It's a derivative of volume class instance.

=over

=item Syntax

    NOTE::Volume::Clone->new($parent, %arguments_hash);

=item Arguments

B<parent> instance

I<Required> Instance of NOTE's volume class

B<debug> = 0|1

I<Optional> Additional debugging during run time, default is 0.

B<recycle> = 0|1

I<Optional> Set/Clear recycle status. Default is inherited from parent

B<name> string

I<Optional> Set name of the clone to given name. Default is anonymous

B<anonymous> = 0|1

I<Optional> Force/Clear anonymous flag regardless of name

B<exportfs> string

I<Optional> Parameters for exportfs

=item Examples

    use NOTE;

    my $parent = NOTE::Volume->create( Filer => { host => $FILER_SRC }, tag => 'Parent' );
    my $clone  = NOTE::Volume::Clone->new( $parent, name => 'clone0' );

=back

=cut

sub new {
    my ($class,$parent,%args) = @_;
    my $self = {};

    croak "Can't clone traditional volumes" if ( $parent->type eq 'trad' );

    $parent->Filer->check_license( "flex_clone" );
    $self->{_debug}    = $args{debug};
    $self->{_owner}    = $$;
    $self->{_dataset}  = $parent->{_dataset};
    $self->{_tag}      = "Clone of ".$parent->description;
    $self->{_recycle}  = $args{recycle};

    $args{name} = NOTE::option( $args{name} ) if exists $args{name};

    if ( $args{name} ) {
	$self->{_name} = $args{name};
	$self->{_anonymous} = $args{anonymous};
    } else {			# Create a unique name
	$self->{_name} = NOTE::Volume->newName;
	$self->{_anonymous} = 1;
    }
    $self->{_type}      = $parent->type;
    $self->{_Aggregate} = $parent->Aggregate;
    $self->{_Filer}	= $self->{_Aggregate}->Filer;
    $self->{_is_clone}  = 1;
    $self->{_parent}    = $parent;

    bless $self, $class;
    tie %{$self}, 'NOTE::Registry::Slice', $self, $NOTE::Volume::regSlice;

    return $self;
}

=head2 ::attach

Attach to existing clone and parent using clone volume name

=over

=item Arguments

B<Filer> Object or string
I<Required> Object of NOTE::Filer class or host name of the filer

B<name> string
I<Required> Volume name of clone.

B<force> = 0|1
I<Optional> Force operation

=back

=cut

sub attach {
    my ($class,%args) = @_;
    my $self;
    my $filer;
    my $exportfs_opts = $args{exportfs} || "rw,anon=0";
    my $force = $args{force};
    delete $args{force};

    # if the $args{Filer} is a filer object, use it,
    # otherwise try to create one by calling the NOTE::Filer constructor.
    if ( !defined($args{Filer}) ) {
        # In this case, there are no arguments (i.e. no hostname) for the
        # filer. I call the filer method anyway and let it decide to
        # handle it or croak.
        $filer = NOTE::Filer->connect();
    } elsif ( ref $args{Filer} eq "HASH" ) {
	$filer = NOTE::Filer->connect( %{$args{Filer}} );
    } elsif ( ref $args{Filer} eq "NOTE::Filer" ) {
	$filer = $args{Filer}; # This is already an object
    } elsif ( ref \$args{Filer} eq "SCALAR" ) {
	# Is just a hostspec
	$filer = NOTE::Filer->connect( host => $args{Filer} );
    } else {
	croak "Invalid filer specification: ".ref \$args{Filer}."\n";
    }
    delete $args{Filer};

    croak "Clone attach requires a name" unless ( $args{name} );

    $args{name} = NOTE::option( $args{name} ) if exists $args{name};
    
    my ($parent_name,$base_snapshot) = Volume::Clone::get_clone_info(
	conn  => $filer->conn,
	clone => $args{name}
    );
    croak $@ if $@;
    croak "Can't determine parent of clone $args{name}" unless( $parent_name );
    my $parent = NOTE::Volume->attach( Filer => $filer, name => $parent_name );

    $self = __PACKAGE__->new( $parent, %args );
    $self->exportfs( $exportfs_opts );

    return $self;
}

=head2 ::create

Create clone for existing volume

=over

=item Arguments

B<Filer> Object or string

I<Required> Object of NOTE::Filer class or host name of the filer

B<name> string

I<Required> Volume name of clone.

B<force> = 0|1

I<Optional> Force operation

B<debug> = 0|1

I<Optional> Additional debugging during run time, default is 0.

B<recycle> = 0|1

I<Optional> Set/Clear recycle status. Default is inherited from parent

B<anonymous> = 0|1

I<Optional> Force/Clear anonymous flag regardless of name

B<exportfs> string

I<Optional> Parameters for exportfs

=item Examples

    my $clone = NOTE::Volume::Clone->attach(
        Filer => $filer,
	name  => $volname,
	exportfs => 'ro,anon=0',
	recycle => 1
    );


=back

=cut

sub create {
    my ($class,$parent,%args) = @_;
    my $exportfs_opts = $args{exportfs} || "rw,anon=0";

    my $self = __PACKAGE__->new( $parent, %args );
    NOTE::reference( $self );

    delete $args{recycle};
    delete $args{name};
    delete $args{anonymous};
    delete $args{exportfs};

    NOTE::logcomment( "Cloning ".$parent->description );

    Volume::Clone::vol_clone_create(
	conn   => $parent->Filer->conn,
	parent => $parent->name,
	clone  => $self->{_name},
	%args,
    );
    croak $@ if $@;
    $self->exportfs( $exportfs_opts );
    return $self;
}

=head2 ::Volume

Return parent volume as NOTE::Volume object

=cut

sub Volume { return shift->{_parent} }

=head2 ::split

Split clone. Clone becomes regular NOTE::Volume object

=over

=item Arguments

B<background> = 0|1

I<Optional> Wait for split to complete. Default behaior is to run in foreground

=item Examples

    $clone->split( background => 1 );

=back

=cut

sub split {
    my ($self,%args) = @_;
    my $background = $args{background};
    delete $args{background};

    NOTE::logcomment( "Splitting clone ".$self->description );
    $self->Filer->check_license( "flex_clone" );
    Volume::Clone::vol_clone_split_start(
	conn => $self->Filer->conn,
	clone => $self->name,
	wait_to_complete => $background ? 0 : 1,
	%args,
    );
    croak $@ if ($@);
}

=head2 ::split_stop

Abort clone split process.

=cut

sub split_stop {
    my ($self,%args) = @_;

    NOTE::logcomment( "Stop splitting clone ".$self->description );
    Volume::Clone::vol_clone_split_stop(
	conn => $self->Filer->conn,
	clone => $self->name,
	%args,
    );
    croak $@ if ($@);
}

=head2 ::split_status

Return split progress status

=cut

sub split_status {
    my ($self,%args) = @_;
    my %status;

    eval { # split_status dies if no status available
	%status = Volume::Clone::vol_clone_split_status(
	    conn => $self->Filer->conn,
	    clone => $self->name,
	    %args,
	);
    };
    return undef if $@; # Just means no clones are being split
    return $status{$self->name};
}

=head2 ::verify

Verify clone

=cut

sub verify {
    my ($self,%args) = @_;

    NOTE::logcomment( "Verifying clone ".$self->description );
    Volume::Clone::verify_clone(
	conn => $self->Filer->conn,
	clone => $self->name,
	%args,
    );
    croak $@ if ($@);
}
1;
