#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Aggregate.pm#2 $
#
## @summary  GNOTE Aggregate module
## @author   tmichael@netapp.com
## @status   Private
## @pod      here

# NOTE: Do *NOT* import any external functions into this module!
# This is to avoid namespace collisions namely prototype mismatches and
# all that. Call the external function directly by prepending the package
# name.

package NOTE::Aggregate;
use Data::Dumper;
use GNOTE;
our @ISA = qw(NOTE);
use strict;
use Carp;

=head1 NAME

NOTE::Aggregate

=head1 DESCRIPTION

This is the NOTE class for working with Aggregates. It contains the common
logic between 7G and GX (C-mode) aggregate classes, and manages the
construction of objects based on whether the call is made in a 7G or GX
environment.

This package should only contain methods that are completely common between 7G
and GX.

=head1 METHODS

=head2 create

This constructor creates an object representing an aggregate.  It can be passed
either a filer object or a filer descriptor.

The method determines if a B<NOTE::7G::Aggregate> or B<NOTE::GX::Aggregate>
object is required (see C<L<NOTE::is_GX|lib-GNOTE-pm.html#is_GX>> for details),
and calls the corresponding C<create> method to do the initial creation/setup
of the object. This method then completes the common construction of the
aggregate object.

=over

=item Arguments

As with all GNOTE methods, arguments will be in the form of a hash table. The
arguments will be passed to B<NOTE::7G::Aggregate::create()> or
B<NOTE::GX::Aggregate::create()> for processing, as appropriate.

Please see
C<L<NOTE::7G::Aggregate::create|lib-GNOTE-7G-Aggregate-pm.html#create>> or
C<L<NOTE::GX::Aggregate::create|lib-GNOTE-GX-Aggregate-pm.html#create>> for
details.

=item Examples

To create a GX (C-mode) aggregate:

    my $aggr_gx = NOTE::Aggregate->create(
        Vserver => "vs0",
        Node    => "dpg-gx4"
    );

To create a 7G aggregate:

    my $aggr_7g = NOTE::Aggregate->create(Filer => "dpg-gx4");

Note that if you have set the NATE parameter C<VSERVER> (or C<VSERVER_SRC> or
C<VSERVER_DST>) you will still get a C-mode aggregate.

=back

=cut

sub create {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # If ref $self returns false, $self is a string representing
    # the class. This will be NOTE::Aggregate, which means that
    # it is not known yet what the actual class should be. Look
    # at the parameters for clues to figure out whether we should
    # call the GX or 7G constructor.
    if ( !ref $self ) {
        if ( NOTE::is_GX(%args) ) {
            return NOTE::GX::Aggregate->create(%args);
        }
        else {
            return NOTE::7G::Aggregate->create(%args);
        }
    }

    # If we get here, then $self is already an object.
    # It will be either type NOTE::GX::Aggregate or NOTE::7G::Aggregate
    # and will have been partially initialized by the constructor of
    # that subclass. The code below must work for either class.

    $self->{_tag} = delete $args{tag};

    my $share = $args{share};

    # default to 1 for share flag, unless $WAFLIRON is true, since
    # snapmirror activity in a shared aggregate may cause wafliron to fail.
    $self->{_share} = ( ( !defined $share ) && !NOTE::option(q/$WAFLIRON/) )
      || $share;
    $self->{_teardown} = delete $args{teardown};

    if ( $args{dataset} ) {
        unless ( $self->can('snapshot_create') ) {
            croak("Datasets are not supported by this class");
        }
        $args{dataset} =~ s/\s+/ /g;    # Compress multiple whitespace to one space
        $args{dataset} =~ s/^\s+//g;    # Remove leading whitespace
        $args{dataset} =~ s/\s+$//g;    # Remove trailing whitespace
    }

    if ( $args{name} ) {
        $self->{_name}      = $args{name};
        $self->{_share}     = 0;                  # Share makes no sense with named aggregates
                                                  # Allow named aggregates that behave like anonymous aggregates:
        $self->{_anonymous} = $args{anonymous};
        if ( $args{dataset} && !$self->{_anonymous} ) {

            # Don't allow non-anonymous aggregates with datasets.
            croak("Recycled datasets can only exist in anonymous aggegates");
        }
    }
    else {                                        # Create a unique name
        if ( $args{dataset} ) {

            # Don't allow shared aggregates with datasets.
            $self->{_share} = 0;
        }
        $self->{_name} = $self->newName( share => $self->{_share} );
        $self->{_anonymous} = 1;
    }
    delete $args{share};
    delete $args{name};
    delete $args{anonymous};

    # Default to recycle=>1 for aggregates
    $args{recycle} = 1 unless ( defined $args{recycle} );

    # Don't print name yet; it's not necessarily set in stone
    NOTE::logcomment( "Require ( $self->{_type} ) aggregate " .
                      ($self->{_tag} ? "( $self->{_tag} )" : q{}) );

    $args{raidtype} ||= NOTE::option(q/$RAIDTYPE:raid_dp/);    # Default

    my $node_cnt = scalar( split /,/, $args{nodes} );
    if ( !( $args{disks} ) && !( $args{disklist} ) ) {
        $args{disks} = $self->_disk_cnt( $args{raidtype} );
        if ( $args{nodes} ) {

            # Number of disks must be a multiple of the number of nodes:
            $args{disks} *= $node_cnt;
        }
    }

    my $numdisks;
    $numdisks = scalar @{ $args{disklist} } if ( $args{disklist} );
    $numdisks ||= $args{disks};
    NOTE::logcomment( "Require a %s %s aggregate with %d disks", $self->{_share} ? "shareable" : "non-shareable", $args{raidtype}, $numdisks, );

    my $diskspec = $args{disklist} ? $args{disklist} : $args{disks};
    $args{disks} = split " ", $args{disklist}
      if exists $args{disklist};

    delete $args{disks};
    delete $args{disklist};

    $self->{_recycle} = delete $args{recycle};
    $self->{_dataset} = delete $args{dataset};
    if ( !defined $self->{_type} ) {
        if ( $self->isa('NOTE::GX') ) {
            if ( $node_cnt > 1 ) {
                $self->{_type} = 'striped';
            }
            else {
                $self->{_type} = 'flex';
            }
        }
    }

    if ( $self->{_recycle} || $self->{_share} ) {
        if ( !$self->{_anonymous} ) {
            if ( $self->exists ) {
                if ($self->verify_aggregate_attributes(%{$self->{_args}},
                                                       disks => $diskspec)) {
                    confess('CONFIG ERROR: named aggregate does not match prerequisites');
                }

                NOTE::logcomment( "Re-using Aggregate " . $self->description );
                NOTE::reference($self) if ( $self->{_teardown} );
                $self->resource('allocate');
                return $self;
            }
        }
        if ( $self->_recycle($diskspec) ) {
            NOTE::logcomment( "Using aggregate " . $self->description );
            NOTE::reference($self);
            $self->resource('allocate');
            return $self;
        }

        # This is a bit tricky: If I wasn't able to recycle
        # an aggregate that was populated with a dataset, try to
        # recycle an empty one and then populate it as if
        # it was just created.
        if ( my $dataset = $self->{_dataset} ) {
            $self->{_dataset} = undef;
            my $gotit = $self->_recycle($diskspec);
            $self->{_dataset} = $dataset;
            if ($gotit) {
                NOTE::reference($self);
                $self->resource('allocate');
                return $self;
            }
        }
    }
    if ( $self->{_anonymous} && $self->{_share} && !NOTE->RECYCLE ) {

        # If we're here, we are about to create a shared aggregate
        # when recycling has been disabled. In this case, we
        # should create a regular aggregate instead, so it will
        # be destroyed when the script exits.  $self->{_share} = 0;
        $self->{_name} = $self->newName( share => $self->{_share} );
    }
    NOTE::logcomment( "Creating Aggregate " . $self->description );

    my ( $success, $err );
    do {
        $success = $self->_aggr_create( $diskspec, %args );
        $err = $@;
    } while ( ( !$success ) && ( _releaseSpareAggr( $self->{_Filer} ) ) );

    croak("Create of aggregate failed ($err)") unless $success;
    NOTE::reference($self);
    $self->resource('allocate');
    return $self;
}

=head2 attach

Attach to an existing aggregate instead of creating new one. Uses the
C<L<aggr_exists|lib-Aggregate-pm/aggr_exists($conn,$aggr)>> function from
F<lib/Aggregate.pm>.

=over

=item Arguments

=over

=item B<Filer> = I<filer object> | I<filer spec>

I<Required> - the filer on which the aggregate to attach to exists

=item B<name> = string

I<Required> - name of the aggregate to attach to, aggregate must be already
present on B<Filer>

=item B<tag> = string

I<Optional> - comments, will be printed in the log file along with the
aggregate name

=back

=item Examples

Connect to aggregate C<aggr0> on the filer C<spectre-l15>:

    my $filer = NOTE::GX::Node->connect(name => 'spectre-l15');
    my $aggr = NOTE::Aggregate->attach(
        Filer => $filer,
        name  => 'aggr0'
    );

=back

=cut

sub attach {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # If ref $self returns false, $self is a string representing
    # the class. This will be NOTE::Aggregate, which means that
    # it is not known yet what the actual class should be. Look
    # at the parameters for clues to figure out whether we should
    # call the GX or 7G constructor.
    if ( !ref $self ) {
        if ( NOTE::is_GX(%args) ) {
            return NOTE::GX::Aggregate->attach(%args);
        }
        else {
            return NOTE::7G::Aggregate->attach(%args);
        }
    }
    my $args = $self->{_args};

    # If we get here, then $self is already an object.
    # It will be either type NOTE::GX::Aggregate or NOTE::7G::Aggregate
    # and will have been partially initialized by the constructor of
    # that subclass. The code below must work for either class.

    $self->{_owner} = $$;
    $self->{_debug} = delete $args->{debug};
    $self->{_tag}   = delete $args->{tag};

    croak "Aggregate attach requires a name" unless ( $args->{name} );
    $self->{_name}    = $args->{name};
    $self->{_dataset} = NOTE::datasetDecode( $self->{_name} );

    unless ( $self->exists ) {
        croak "Aggregate " . $self->{_name} . " does not exist.";
    }
    return $self;
}

=head2 dup

Create (or reuse) an aggregate with attributes similar to an existing
aggregate.

=over

=item Arguments

Takes the same arguments as L<create|/create>

=item Examples

    # Make a destination aggregate similar to the source
    $aggr_dst = $aggr_src->dup;

    # The same, but using a different filer
    $aggr_dst = $aggr_src->dup(Filer => $filer_dst);

=back

=cut

sub dup {
    my ( $origin, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %properties = $origin->Properties;
    my $class      = ref $origin;

    %properties = %{
        NOTE::_hashOverlay(
            target    => \%properties,
            source    => \%args,
            overwrite => 1
        )
      };
    $properties{Filer} ||= $origin->Filer;
    return $class->create(%properties);
}

=head2 Properties

Returns all the properties of this aggregate in a hash-reference.

=over

=item Arguments

I<None>

=item Examples

    my $properties = $aggr->Properties;
    NOTE::logcomment(
        'Properties of aggregate ' . $aggr->name . ':'
    );
    for my $key (sort keys %{$properties}) {
        NOTE::logcomment(
            sprintf("\t%10s => \%s", $key, $properties->{$key})
        );
    }

=back

=cut

sub Properties {
    my $self       = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $opts       = $self->options;
    my %properties = (
        recycle   => $self->{_recycle},
        anonymous => $self->{_anonymous},
        snaplock  => $self->{_snaplock},
        disks     => scalar $self->disks,
        raidtype  => $opts->{raidtype},
        rgsize    => $opts->{raidsize},
    );
    return %properties;
}

=head2 add

Add disks to aggregate. Adds one disk only if called without arguments.  This
method is a wrapper around
C<L<aggr_add|lib-Aggregate-pm/aggr_add($conn,$aggr,$disklist[,%opts])>> from
F<lib/Aggregrate.pm>. All arguments are passed as a hash to it.

=over

=item Arguments

=over

=item B<disks> = integer

I<Optional> - Number of disks to add. This argument has no effect if
B<disklist> is also present.

=item B<disklist> = array-reference

I<Optional> - Array of disk IDs. Uses the C<-d> option to add each disk via its
disk name taken from each element of the array.

=item B<force> = boolean

I<Optional> - Force operation

=item B<timeout> = integer (seconds)

I<Optional> - Wait for validation timeout. By default it waits for all disks to
be zeroed if needed.

=back

=item Examples

Add one disk (any available disk) to the aggregate:

    $aggr->add();

Add two specific disks to the aggregate:

    $aggr->add(disklist => [ qw(15.3 12.6) ]);

=back

=cut

sub add {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $ret;

    my $diskspec = $args{disks} || $args{disklist};
    delete $args{disks};
    delete $args{disklist};
    $diskspec = "1" unless ($diskspec);

    if ( $diskspec =~ /^\d+$/ ) {
        my $what = $diskspec == 1 ? "disk" : "disks";
        NOTE::logcomment( "Adding $diskspec $what to " . $self->description );
    }
    else {
        my $what = $diskspec == 1 ? "a disk" : "disks";
        NOTE::logcomment( "Adding $what to " . $self->description );
    }
    $ret = $self->_aggr_add( $diskspec, %args, );
    croak $@ if $@ && $@ !~ /aggregate to be quiesced/;
    $self->{_disks} = scalar $self->disks;
    if ( !$self->is_shareable && ( $self->{_anonymous} && $self->{_recycle} ) ) {
        NOTE::logcomment( "Added disk to anonymous, non-shareable, recyclable aggregate.\n"
              . "Turning recycle flag off so aggregate will be destroyed at teardown.\n"
              . "Name the aggregate if you want to get around this." );
        $self->{_recycle} = 0;
    }
    return $ret;
}

=head2 online

Change the state of the aggregate to "online". Returns success (if the
aggregate was changed or is already in the desired state) or failure
otherwise (if the aggregate couldn't change or didn't exist).

=over

=item Arguments

=over

=item B<force> = boolean

Attempt to force the operation

=back

=item Examples

Make an aggregate be online:

    $aggr->online;

Same, with an attempt to force the operation if there would be a problem:

    $aggr->online(force => 1);

=back

=cut

sub online {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->_aggr_online(@_);
    croak $@ if $@;
    return $self;
}

=head2 offline

Change the state of the aggregate to "offline". Returns success (if the
aggregate was changed or is already in the desired state) or failure
otherwise (if the aggregate couldn't change or didn't exist).

=over

=item Arguments

=over

=item B<force> = boolean

Attempt to force the operation

=back

=item Examples

Move the aggregate to an offline state:

    $aggr->offline;

Same operation, attempting to force the move if there is a problem:

    $aggr->offline(force => 1);

=back

=cut

sub offline {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Cannot offline a sharable aggregate" if $self->is_shareable;
    $self->_aggr_offline(@_);
    croak $@ if $@;
    return $self;
}

=head2 restrict

Change the state of the aggregate to "restricted". Returns success (if the
aggregate was changed or is already in the desired state) or failure
otherwise (if the aggregate couldn't change or didn't exist).

=over

=item Arguments

None.

=item Example

Change the aggregate to a restricted state:

     $aggr->restrict;

=back

=cut

sub restrict {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Cannot restrict a sharable aggregate" if $self->is_shareable;
    $self->_aggr_restrict(@_);
    croak $@ if $@;
    return $self;
}

=head2 rename

Renames the aggregate to the given name, or to another anonymous name if no new
name is given.

Anonymous aggregates retain the anonymous flag after being renamed if the new
name looks like that of an anonymous aggregate. This is needed for tests that
"swap" names between two anonymous aggregates.

Renaming an aggregate to an anonymous name by calling this method with no
argument will also set the anonymous flag on the aggregate.

=over

=item Argument

B<name> = string

I<Optional> - renames the aggregate to the given value

=item Examples

Rename an aggregate:

    $aggr->rename('Some_New_Name');

Rename an aggregate to a new anonymous name (sets the anonymous flag if it
wasn't already set):

    $aggr->rename();

=back

=cut

sub _rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->_aggr_rename(@_)
}

sub rename {
    my ( $self, $newname ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $newname ) {
        $self->_rename($newname);

        # Allow renaming an aggregate to a name that appears
        # to be anonymous without making it named. This is
        # needed for tests that 'swap' names between two anonymous
        # aggregates.
        if ( $self->{_anonymous} && !$self->_looksAnonymous ) {
            $self->{_anonymous} = 0;
        }
    }
    else {
        $newname = $self->newName;
        $self->{_anonymous} = 1;
        $self->_rename($newname);
    }
}

sub _spareAggrs {
    my $filer      = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @spareAggrs = ();
    my ( $aggrname, $aggr );

    # Get a list of spare aggregates (or tradvols) and their associated disks
    for $aggrname ( $filer->aggregates ) {
        next unless ( $filer->looks_free($aggrname) );
        eval { $aggr = NOTE::Aggregate->attach( Filer => $filer, name => $aggrname, ); };
        next if $@;
        next unless ( $aggr->writeable );
        $aggr->{_disks} = $aggr->disks;
        push @spareAggrs, $aggr;
    }
    return sort { $a->{_disks} <=> $b->{_disks} } @spareAggrs;
}

sub _shareableAggrs {
    my $filer          = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @shareableAggrs = ();
    my ( $aggrname, $aggr );

    # Get a list of shareable aggregates and their associated disks
    for $aggrname ( $filer->aggregates ) {
        next unless ( $filer->looks_shareable($aggrname) );
        eval { $aggr = NOTE::Aggregate->attach( Filer => $filer, name => $aggrname, ); };
        if ($@) {
            NOTE::logcomment( "$aggrname attach failed: %s\n", $@ );
            next;
        }
        next unless ( $aggr->writeable );
        $aggr->{_disks} = $aggr->disks;
        push @shareableAggrs, $aggr;
    }

    # Probably should sort according to capacity (most free space first)
    return sort { $a->{_disks} <=> $b->{_disks} } @shareableAggrs;
}

# Release the smallest spare aggregate, if one exists
sub _releaseSpareAggr {
    my $filer      = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @spareAggrs = _spareAggrs($filer);
    return 0 unless (@spareAggrs);
    my $aggr = pop @spareAggrs;
    NOTE::logcomment( "Releasing spare container " . $aggr->name );
    $aggr->purge;
    return 1;
}

# This internal function is used to log the reason that
# a resource was rejected as a recyclable candidate
sub _reject {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $msg  = "Rejecting " . $args{name} . ": " . $args{reason};

    if ( $args{wanted} ) {
        $msg .= "\n    ";
        $msg .= $args{wanted};
    }
    if ( $args{got} ) {
        $msg .= "\n    ";
        $msg .= $args{got};
    }
    NOTE::logcomment($msg);
}

#
# Choose the closest fitting free aggregate.
# If $diskspec is defined, match must be perfect.
# Match against raidtype and raidgroup size.
#
sub _recycle {
    my ( $self, $disks ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $args        = $self->{_args};
    my $node_to_use = $self->{_Filer}->name;
    if ( $self->{_share} ) {
        NOTE::logcomment("Looking for existing shared aggregate");
        foreach my $aggr ( _shareableAggrs( $self->Filer ) ) {
            NOTE::logcomment( "Consider " . $aggr->name );
            if ( $aggr->snaplock || $args->{snaplock} ) {
                unless ( $aggr->snaplock eq $args->{snaplock} ) {
                    _reject(
                        name   => $aggr->name,
                        reason => "aggregate snaplock qualities don't match",
                        wanted => "requested {snaplock} [" . $args->{snaplock} . "]",
                        got    => "candidate {snaplock} [" . $aggr->snaplock . "]",
                    );
                    next;
                }
            }
            my $aggropts = $aggr->options;    # So we only have to do this once

            if ( $args->{indirect_type} ) {
                unless ( $aggropts->{'block-type'} eq $args->{indirect_type} ) {
                    _reject(
                        name   => $aggr->name,
                        reason => "aggregate block-type doesn't match",
                        wanted => "requested block-type [" . $args->{indirect_type} . "]",
                        got    => "candidate block-type [" . $aggropts->{'block-type'} . "]",
                    );
                    next;
                }
            }
            if ( $args->{raidtype} ) {
                unless ( $aggropts->{raidtype} eq $args->{raidtype} ) {
                    _reject(
                        name   => $aggr->name,
                        reason => "aggregate raidtype doesn't match",
                        wanted => "requested raidtype [" . $args->{raidtype} . "]",
                        got    => "candidate raidtype [" . $aggropts->{raidtype} . "]",
                    );
                    next;
                }
            }
            if ( $args->{rgsize} ) {
                unless ( $aggropts->{raidsize} eq $args->{rgsize} ) {
                    _reject(
                        name   => $aggr->name,
                        reason => "aggregate raidsize doesn't match",
                        wanted => "requested raidsize [" . $args->{rgsize} . "]",
                        got    => "candidate raidsize [" . $aggropts->{raidsize} . "]",
                    );
                    next;
                }
            }
            if ( $self->isa('NOTE::GX') ) {
                if ( $args->{type} ) {
                    my $aggr_type = $aggropts->{voltype}
                      || $aggropts->{'volume-style'};
                    unless ( $aggr_type eq $args->{type} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate volume type doesn't match",
                            wanted => "requested type [" . $args->{type} . "]",
                            got    => "candidate type [" . $aggr_type . "]",
                        );
                        next;
                    }
                }
                my @nodes = sort $aggr->nodes;
                my $node_list = join ",", @nodes;
                if ( $args->{local_storage} && !( $node_list =~ /$node_to_use/ ) ) {
                    _reject(
                        name   => $aggr->name,
                        reason => "shareable aggregate does not reside in local node(s): $node_to_use",
                        wanted => "aggregates from node(s): " . $node_list
                    );
                    next;
                }
                if ( !$args->{nodes} ) {

                    # Default to any healthy aggr with one node
                    unless ( scalar @{ $aggropts->{nodes} } == 1 ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate doesn't have the desired number of nodes",
                            wanted => "requested default number of nodes (i.e. 1)",
                            got    => "candidate has [" . scalar @{ $aggropts->{nodes} } . "] nodes",
                        );
                        next;
                    }
                }
                else {

                    # Make sure list of nodes matches ($args->{nodes} is already a sorted list)
                    my $nodestr = join( ',', sort @{ $aggropts->{nodes} } );
                    unless ( $nodestr eq $args->{nodes} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate doesn't comprise the desired nodes",
                            wanted => "requested     [" . $args->{nodes} . "]",
                            got    => "candidate has [" . $nodestr . "]",
                        );
                        next;
                    }
                }

                if ( $aggropts->{state} ne 'online' ) {
                    _reject(
                        name   => $aggr->name,
                        reason => 'aggregate state not online',
                        wanted => 'requested     [ online ]',
                        got    => "candidate has [ $aggropts->{state} ]",
                    );
                    next;
                }
            }
            if ( $disks && ( $aggr->{_disks} < $disks ) ) {

                # At this point, the only thing wrong with this
                # shared aggregate is that it doesn't contain enough disks.
                # Try to add enough disks to accomodate the request.
                my $err;
                do {
                    eval { $aggr->add( disks => $disks - $aggr->{_disks} ) };
                    $err = $@;
                } while ( $err && ( _releaseSpareAggr( $aggr->Filer ) ) );
                croak $err if $err;
            }

            # Don't rename the aggregate, rename the object
            $self->{_name} = $aggr->{_name};
            NOTE::logcomment( "Found shared aggregate %s", $aggr->description );
            return 1;
        }
    }
    if ( $self->{_recycle} ) {
        NOTE::logcomment( "Looking for recyclable aggregate (share = %d)", $self->{_share} );
        my @spares =
          grep { $self->looks_free( $_->name ) } _spareAggrs( $self->{_Filer} );
        my $desperate = 0;
      again:
        for my $aggr (@spares) {
            eval {    # If any error occurs, simply try the next candidate
                if ( !$desperate ) {

                    # Don't bother with this aggr at all, if the datasets aren't
                    # the same.
                    if ( $self->{_dataset} ) {
                        unless ( $aggr->dataset eq $self->{_dataset} ) {
                            _reject(
                                name   => $aggr->name,
                                reason => "aggregate dataset doesn't match",
                                wanted => "requested dataset [" . $self->{_dataset} . "]",
                                got    => "candidate dataset [" . $aggr->dataset . "]",
                            );
                            next;
                        }
                    }
                    else {
                        if ( $aggr->dataset ) {
                            _reject(
                                name   => $aggr->name,
                                reason => "aggregate dataset doesn't match",
                                wanted => "requested no dataset",
                                got    => "candidate has dataset [" . $aggr->dataset . "]",
                            );
                            next;
                        }
                    }
                }

                if ($disks) {
                    unless ( $aggr->{_disks} >= $disks ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate doesn't comprise the desired number of disks",
                            wanted => "requested $disks disks",
                            got    => "candidate has [" . $aggr->{_disks} . "]",
                        );
                        next;
                    }
                }
                if ( $self->isa('NOTE::GX') ) {
                    my @nodes = sort $aggr->nodes;
                    my $node_list = join ",", @nodes;
                    if ( $args->{local_storage} && !( $node_list =~ /$node_to_use/ ) ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "recycle aggregate does not reside in local node(s): $node_to_use",
                            wanted => "aggregates from node(s): " . $node_list
                        );
                        next;
                    }
                    if ( $args->{nodes} ) {

                        # So here, nodes could be a number, a (comma separated) string,
                        # or an array.
                        if ( $args->{nodes} =~ /^\d+/ ) {

                            # Handling a number is easy:
                            unless ( scalar @nodes eq $args->{nodes} ) {
                                _reject(
                                    name   => $aggr->name,
                                    reason => "aggregate doesn't comprise the desired number of nodes",
                                    wanted => "requested " . $args->{nodes} . " node(s)",
                                    got    => "candidate has [" . scalar @nodes . "]",
                                );
                                next;
                            }
                        }
                        else {

                            # Otherwise, create two strings, one containing the
                            # sorted list of nodes for the candidate aggregate
                            # and the other the sorted list of required nodes.
                            # If the strings don't match, then reject this candidate.
                            # Sorting and joining makes it easy to tell whether
                            # the arrays contain the same elements, even if they
                            # are not in the same order.
                            if ( ref $args->{nodes} eq 'ARRAY' ) {
                                $args->{nodes} = [ sort @{ $args->{nodes} } ];
                            }
                            else {
                                $args->{nodes} = [ sort split( /,/, $args->{nodes} ) ];
                            }
                            my $nodestr1 = join( ',', @nodes );
                            my $nodestr2 = join( ',', @{ $args->{nodes} } );
                            unless ( $nodestr1 eq $nodestr2 ) {
                                _reject(
                                    name   => $aggr->name,
                                    reason => "aggregate doesn't comprise the desired list of nodes",
                                    wanted => "requested     [" . $nodestr2 . "]",
                                    got    => "candidate has [" . $nodestr1 . "]",
                                );
                                next;
                            }
                        }
                    }
                }

                if ( $args->{snaplock} ) {
                    unless ( $aggr->snaplock eq $args->{snaplock} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate snaplock qualities don't match",
                            wanted => "requested {snaplock} [" . $args->{snaplock} . "]",
                            got    => "candidate {snaplock} [" . $aggr->snaplock . "]",
                        );
                        next;
                    }
                }
                my $aggropts = $aggr->options;    # So we only have to do this once
                if ( $args->{raidtype} ) {
                    unless ( $aggropts->{raidtype} eq $args->{raidtype} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate raidtype doesn't match",
                            wanted => "requested raidtype [" . $args->{raidtype} . "]",
                            got    => "candidate raidtype [" . $aggropts->{raidtype} . "]",
                        );
                        next;
                    }
                }
                if ( $args->{rgsize} ) {
                    unless ( $aggropts->{raidsize} eq $args->{rgsize} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate raidsize doesn't match",
                            wanted => "requested raidsize [" . $args->{rgsize} . "]",
                            got    => "candidate raidsize [" . $aggropts->{raidsize} . "]",
                        );
                        next;
                    }
                }
                if ( $args->{fake_large} ) {
                    unless ( $aggropts->{'block-type'} eq $args->{indirect_type} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate block-type doesn't match for fake large aggr",
                            wanted => "requested block-type [" . $args->{indirect_type} . "]",
                            got    => "candidate block-type [" . $aggropts->{'block-type'} . "]",
                        );
                        next;
                    }
                }
                if ( $args->{indirect_type} ) {
                    unless ( $aggropts->{'block-type'} eq $args->{indirect_type} ) {
                        _reject(
                            name   => $aggr->name,
                            reason => "aggregate block-type doesn't match",
                            wanted => "requested block-type [" . $args->{indirect_type} . "]",
                            got    => "candidate block-type [" . $aggropts->{'block-type'} . "]",
                        );
                        next;
                    }
                }


                # Everything looks OK..

                if ($desperate) {
                    $aggr->rename;    # Grab it
                    if ( $aggr->dataset ) {    # Datasets won't match
                        $aggr->empty;
                    }
                    if ( $self->{_dataset} ) {
                        $aggr->{_dataset} = $self->{_dataset};
                        NOTE::logcomment( "Initializing dataset of " . $aggr->description );
                        my $func = eval 'sub {' . $aggr->{_dataset} . '}';
                        croak $@ if $@;
                        eval { $aggr->$func };
                        croak $@ if $@;
                        if ( $aggr->_looksAnonymous ) {
                            $aggr->_rename( NOTE::datasetEncode( $aggr->{_name}, $aggr->{_dataset} ) );
                            eval { $aggr->snapshot_delete("__clean__") };    # Don't care if this fails
                            $aggr->snapshot_create("__clean__");
                        }
                    }
                }

                if ( $self->{_dataset} ) {
                    $self->{_name} = NOTE::datasetEncode( $self->{_name}, $self->{_dataset} );
                }
                NOTE::logcomment( "Recycling Aggregate; " . $aggr->name, " --> " . $self->{_name} );
                $aggr->_rename( $self->{_name} );

                if ( $aggr->can('snapshot_create') ) {
                    if ( $aggr->is_shareable ) {

                        # Delete the __clean__ and __empty__ snapshots if we
                        # are recycling an aggregate in order to make a shareable one.
                        eval { $self->snapshot_delete("__clean__") };    # Don't care if this fails
                        eval { $self->snapshot_delete("__empty__") };    # Don't care if this fails
                    }
                    else {
                        if ( !$aggr->snapshot_exists("__empty__") ) {
                            $aggr->purge;
                            NOTE::logcomment( "Rejecting aggregate " . $aggr->name . " because it had no __empty__ snapshot" );
                            next;
                        }
                        if ( !$aggr->snapshot_exists("__clean__") ) {
                            $aggr->empty;

                            # If a dataset is required, this aggregate is no longer appropriate
                            if ( $self->{_dataset} ) {
                                NOTE::logcomment( "Rejecting aggregate " . $aggr->name . " because it had no __clean__ snapshot" );
                                next;
                            }
                        }
                    }
                }
                $self->{_was_recycled} = 1;
                goto gotOne;    # got one
            };
            NOTE::logcomment("Attempted recycle failed: $@") if ($@);
        }

        # If I get here, no candidates were found.
        # Try once more, but this time don't reject aggregates with
        # datasets we're not looking for. If we find one that meets all
        # the other requirements, just change the dataset to be what
        # we want it to be.
        if ( $self->can('snapshot_create') && !$desperate ) {
            NOTE::logcomment("No candidate found; trying harder");
            $desperate = 1;

            # Re-sort @spares according to creation difficulty:
            # Lowest number of disks and smallest population of data first.
            for (@spares) {

                # Cache 'used', so the sort is efficient.
                $_->{_used} = $_->used;
            }
            @spares =
              sort { $a->{_disks} <=> $b->{_disks} || $a->{_used} <=> $b->{_used} } @spares;
            goto again;
        }
    }
    NOTE::logcomment("No candidate found");

    # XXX If I want an aggregate with a dataset, I should be
    # able to convert an empty aggregate to a populated aggregate
    # if no appropriate populated aggregate exists.
    # And vice-versa - If I want an empty aggregate,
    # and none exist, I might want to consider removing
    # data from a populated aggregate, if there are no more
    # disks available.
    return 0;
  gotOne:
    return 1;
}

=head2 Filer

Return the filer or node object that this aggregate object is connected
to. This is a read-only accessor, you cannot change this value. The return
value is either a B<NOTE::7G::Filer> object or a B<NOTE::GX::Node> object.

=over

=item Arguments

None

=item Examples

Use this method to retrieve the filer object so that we can ping the underlying
filer to ensure it is still reachable:

    if (! $aggr->Filer->ping()) {
        NOTE::logcomment('Warning: Filer not reachable');
    }

=back

=cut

sub Filer {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Filer}
}

=head2 dataset

Return the string (of code) representing the dataset of this aggregate. This is
a read-only accessor, the dataset cannot be changed.

=over

=item Arguments

None

=item Example

A simple example of displaying the content of C<dataset>:

    NOTE::logcomment(
        'Dataset of this aggregate: ' . $aggr->dataset()
    );

=back

=cut

sub dataset {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_dataset}
}

=head2 tag

Gets or sets the tag of the aggregate. If called with no argument, the current
tag is returned. If B<tag> is given, sets the tag to the value of the argument
and returns the new value.

=over

=item Arguments

B<tag> = string

I<Optional> Set the tag to the given value. Some points in the NOTE/GNOTE code
display the tag along with the aggregate name in logs, to distinguish otherwise
anonymous aggregates.

=item Examples

Set a tag value on C<$aggr>:

    $aggr->tag('Source Aggregate');

Get the current tag for C<$aggr>:

    my $atag = $aggr->tag();

=back

=cut

sub tag {
    my ( $self, $newtag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $newtag ) {
        $self->{_tag} = $newtag;
    }
    return $self->{_tag};
}

=head2 description

Returns a string describing the aggregate, consisting of the filer's name, the
aggregate name and the tag, if any.

=over

=item Arguments

None.

=item Examples

Use B<description> in a log comment:

    NOTE::logcomment(
        'Operating on aggregate ' . $aggr->description()
    );

=back

=cut

sub description {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $txt;

    $txt = $self->hostpath;
    if ( $self->{_tag} ) {
        $txt .= " (" . $self->{_tag} . ")";
    }
    return $txt;
}

=head2 newName

Generates a new anonymous name for an aggregate. Names start with C<AG> for
shared aggregates or with C<ag> for non-shared aggregates. This method doesn't
execute any filer commands.

This method may be called as a static method.

This method does not change the name of the aggregate if/when called on a
B<NOTE::Aggregate> object. It only generates names and returns them.

=over

=item Arguments

B<share> = boolean

I<Optional> Define the aggregate type. The default is false (0).

=item Examples

Get a name for a shared aggregate:

    $aggr->newName(share => 1);

Get 5 new names for non-shared aggregates:

    my @names;
    for (1 .. 5) {
        # Could also say "newName(share => 0)":
        push(@name, NOTE::Aggregate->newName());
    }

=back

=cut

sub newName {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $args{share} ) {
        return shift->SUPER::newName("AG");
    }
    else {
        return shift->SUPER::newName("ag");
    }
}

=head2 name

Return the name of this aggregate. This is a read-only accessor, and cannot
change the name of the aggregate. See L<the rename() method|/rename> for
changing an aggregate's name.

=over

=item Arguments

None.

=item Examples

Log an aggregate's name:

    NOTE::logcomment('Aggregate name is: ' . $aggr->name());

=back

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name}
}

=head2 is_recyclable

Returns 1 (true) if the aggregate is a recyclable aggregate, 0 (false)
otherwise. This is also assignable, if it is desired to change the status.

=over

=item Arguments

None

=item Examples

Show the recyclable status of C<$aggr>:

    if ($aggr->is_recyclable) {
        printf("Aggregate %s is recyclable\n", $aggr->name());
    } else {
        printf("Aggregate %s is not recyclable\n", $aggr->name());
    }

Explicitly set C<$aggr> to not be recycleable:

    $aggr->is_recyclable = 0;

=back

=cut

sub is_recyclable : lvalue {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # [See comments in is_anonymous]
    my $r = \$self->{_recycle};
    $$r;
}

=head2 is_anonymous

Returns 1 (true) if aggregate is anonymous, 0 (false) otherwise. This is also
assignable, if it is desired to change the status. Anonymous aggregates are
destroyed or recycled when the script exits.

=over

=item Arguments

None

=item Examples

Notify the user whether this aggregate will be destroyed at script exit:

    if ($aggr->is_anonymous) {
        NOTE::logcomment(
              'Note that this aggregate ('
            . $aggr->name()
            . ') will be destroyed/recycled at end.'
        );
    }

Set C<$aggr> as anonymous:

    $aggr->is_anonymous = 1;

=back

=cut

sub is_anonymous : lvalue {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Must do this trickery due to a bug in perl regarding
    # tied hashes and lvalues.
    # Actually, for NOTE::Aggregate.pm, this isn't strictly necessary
    # because while NOTE::Volume objects are tied hashes, NOTE::Aggregate
    # objects are not (nothing interesting in the registry about them).
    # But I put this workaround here anyway, in case they ever become
    # tied hashes.
    my $r = \$self->{_anonymous};
    $$r;
}

=head2 is_free

Returns 1 (true) if the aggregate is available for reuse, 0 (false)
otherwise. This is not an attribute, and cannot be manually set.

=over

=item Arguments

None

=item Examples

Identify all the aggregates on C<$filer> that are available for reuse:

    for my $aggr ($filer->Aggregates) {
        next unless ($aggr->is_free());
        printf("Aggregate %s is available for reuse\n", $aggr->name());
    }

=back

=cut

sub is_free {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return ( $self->isa('NOTE::Aggregate') && $self->looks_free );
}

=head2 is_recycled

Returns 1 (true) if the aggregate was recycled, i.e. obtained from the pool of
available aggregates. Returns 0 otherwise. A shared aggregate is not considered
recycled.

=over

=item Arguments

None

=item Examples

Identify all the aggregates on C<$filer> that are recycled:

    for my $aggr ($filer->Aggregates) {
        next unless ($aggr->is_recycled());
        printf("Aggregate %s is recycled\n", $aggr->name());
    }

=back

=cut

sub is_recycled {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_was_recycled}
}

=head2 is_shareable

Returns 1 (true) if aggregate is a shareable aggregate. Returns 0 (false) if
it is not.

=over

=item Arguments

None

=item Examples

Identify all the aggregates on C<$filer> that are shareable:

    for my $aggr ($filer->Aggregates) {
        next if (! $aggr->is_shareable());
        printf("Aggregate %s is shareable\n", $aggr->name());
    }

=back

=cut

sub is_shareable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->looks_shareable
}

=head2 exists

Returns a true value if the aggregate physically exists, C<undef> otherwise.

=over

=item Arguments

None

=item Examples

Verify that an aggregate was destroyed:

    eval { $aggr->destroy };
    if ($aggr->exists) {
        croak 'Unable to destroy aggregate ' . $aggr->name();
    }

=back

=cut

sub exists {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->_aggr_exists
}

=head2 disks

In a list context, returns a list of the names of disks that this aggregate
comprises. In a scalar context, returns a count of the disks.

=over

=item Arguments

None

=item Examples

Show the list of disks:

    NOTE::logcomment(
          'Aggregate ' . $aggr->name() . ' contains disks '
        . join(q{, } => $aggr->disks())
    );

Get the count of disks:

    my $num_disks = $aggr->disks();

=back

=cut

sub disks {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if (wantarray) {
        return shift->_aggr_disks;
    }
    else {
        return scalar( shift->_aggr_disks );
    }
}

=head2 SPARE

Don't physically destroy (or recycle) this aggregate, even if it is
anonymous. Clears the recycle flag and sets the spare flag.

=over

=item Arguments

None

=item Examples

Spare the aggregate if the NOTE parameter C<SPARE_ALL> was set:

    if (NOTE::param('SPARE_ALL')) {
        $aggr->SPARE();
    }

=back

=cut

# Don't physically destroy (or recycle), even if anonymous
sub SPARE {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_recycle}  = 0;
    $self->{_spare_me} = 1;
}

=head2 release

Unless this aggregate has been marked to be spared (see L<SPARE()|/SPARE>),
returns the aggregate to free pool (if recycleable) or destroys the aggregate
(if not).

=over

=item Arguments

None

=item Example

Release all the non-spared aggregates on a filer:

    for my $aggr ($filer->Aggregates) {
        $aggr->release;
    }

=back

=cut

sub release {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->_release;
    $self->unreference;    # Remove from teardown stack
    return $self;
}

sub _release {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return unless ( $self->{_anonymous} );
    return if ( $self->{_share} );
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    $self->{_destroyed} = 1;
    return if exists $self->{_parent};

    if ( NOTE->RECYCLE && $self->{_recycle} ) {
        eval { $self->online };
        if ( $@ && $@ !~ /is already online/ ) {
            NOTE::logcomment( "Online of " . $self->{_name} . " failed: $@" );
        }
        else {
            eval { $self->clean if ( $self->can('clean') ) };
            unless ($@) {
                NOTE::logcomment( "Returning Aggregate " . $self->{_name} . " to free pool" );
                if ( !$self->_looksAnonymous ) {
                    $self->_rename( $self->_make_free( $self->newName ) );
                }
                else {
                    $self->_rename( $self->_make_free( $self->name ) );
                }
                $self->resource('free');
                return;
            }
            NOTE::logcomment( "Restore to clean state of " . $self->{_name} . " failed: $@" );

            # restore using __clean__ failed, so just destroy the aggregate
        }
    }
    eval { $self->purge };
}

=head2 total

Returns the overall aggregate size in KB. Output is taken from a C<df -A -k>
command.

=over

=item Arguments

None

=item Examples

Report the total size of the aggregate in the desired unit:

    my $size = $aggr->total();
    if ($unit eq 'G') {
        # Gigabytes
        $size = int($size / (1024 * 1024));
        printf("Aggregate %s size is %dG\n", $aggr->name(), $size);
    } elsif ($unit eq 'M') {
        # Megabytes
        $size = int($size / 1024);
        printf("Aggregate %s size is %dM\n", $aggr->name(), $size);
    } else {
        # Kilobytes
        printf("Aggregate %s size is %dK\n", $aggr->name(), $size);
    }

=back

=cut

sub total {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free )[0]
}

=head2 used

Returns the used size in KB. Output is taken from a C<df -A -k> command.

=over

=item Arguments

None

=item Examples

Report the used size of the aggregate in the desired unit:

    my $size = $aggr->used();
    if ($unit eq 'G') {
        # Gigabytes
        $size = int($size / (1024 * 1024));
        printf("Aggregate %s used is %dG\n", $aggr->name(), $size);
    } elsif ($unit eq 'M') {
        # Megabytes
        $size = int($size / 1024);
        printf("Aggregate %s used is %dM\n", $aggr->name(), $size);
    } else {
        # Kilobytes
        printf("Aggregate %s used is %dK\n", $aggr->name(), $size);
    }

=back

=cut

sub used {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free )[1]
}

=head2 avail

Returns aggregate available space in KB. Output is taken from a C<df -A -k>
command.

=over

=item Arguments

None

=item Examples

Report the available space of the aggregate in the desired unit:

    my $size = $aggr->avail();
    if ($unit eq 'G') {
        # Gigabytes
        $size = int($size / (1024 * 1024));
        printf("Aggregate %s avail is %dG\n", $aggr->name(), $size);
    } elsif ($unit eq 'M') {
        # Megabytes
        $size = int($size / 1024);
        printf("Aggregate %s avail is %dM\n", $aggr->name(), $size);
    } else {
        # Kilobytes
        printf("Aggregate %s avail is %dK\n", $aggr->name(), $size);
    }

=back

=cut

sub avail {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free )[2]
}

=head2 capacity

Returns capacity as a percentage. Output is taken from a C<df -A -k> command.

=over

=item Arguments

None

=item Examples

Report the capacity of the aggregate in the desired unit:

    my $cap = $aggr->capacity();
    printf("Aggregate %s capacity is %d%%\n", $aggr->name(), $cap);

=back

=cut

sub capacity {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free )[3]
}

###############################################################################
#
#   Sub Name: _disk_cnt
#
#   Description:
#       Determines the mininum number of disks to use when creating an
#       aggregate or adding disks to an existing striped aggregate.
#
#   Arguments:
#       The arguments are passed as an ordered list not a hash.
#
#               NAME        IN/OUT  TYPE    DESCRIPTION
#               raidtype    in      scalar  raid4|raid_dp
#
#   Globals:
#
#   Environment:
#
#   Returns:
#       an integer value
#
#   Examples:
#
#   use the aggregate show output to get the raid type
#       $aggrStatus = $aggr->status;
#       $diskcnt = $aggr->_disk_cnt( $aggrStatus[0]->{raidtype} );
#
#   use the original aggregate arguments to get the raid type
#       $diskcnt = $self->_disk_cnt( $self->{_args}->{raidtype} );
#
###############################################################################

sub _disk_cnt {
    my ( $self, $raidtype ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $cnt = NOTE::option(q/$DISK_CNT/);
    if ($cnt) {
        return $cnt;
    }
    else {
        confess("Invalid raid type $raidtype")
          unless $raidtype =~ /raid(4|_dp)/;
        return ( $raidtype =~ /raid4/ ? 4 : 5 );
    }
}

=head2 DESTROY

Object destructor for aggregate objects. Users do not call this method
directly.

=cut

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return unless ( $self->{_anonymous} );
    return if ( $self->{_share} );
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    return if ( NOTE->ALL_SPARED );
    if ( $self->{_spare_me} ) {
        NOTE::logcomment( "Aggregate $self->{_name} has been spared" );
        return;
    }
    $self->release;
}

1;
