# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Client.pm#1 $
##
## @summary note Client class
##
## @status public
##
## @pod    here
##
## @author sydneyw@netapp.com, dl-dpg-auto@netapp.com
##

=head1 NAME

NOTE::Client

=head1 DESCRIPTION

Provides APIs for various actions on a client machine, such as Unix or Windows.
Provides support for streams, ACLs and LUNs.

A CLI parameter "UTIL_BIN" is used to locate external executable/tools such as Hammer, Otherwise the standard location will be used. (nfs: /usr/software/test/bin  cifs: X:\bin

If you are using a Windows client for CIFS testing you may need to specify
additional CLI parameters if you are testing in a CIFS domain that's not
in the default SVL lab domain.

The CLI parameters are "CIFS_DOMAIN", "CIFS_USER", "CIFS_PASSWD", "CIFS_OU".

The default values for these parameters are: CIFS_DOMAIN=LAB.NETAPP.COM,
CIFS_USER=dpgadmin and CIFS_PASSWD=sundance

Here is an example of the usage of these CLI parameters for a Windows client
located in the RTP lab CIFS domain.

ntest
FILER_SRC=abbott
VSERVER_SRC=abbott_vs0
DATASET=DS-small
CLIENT_LIST=craik
CIFS_DOMAIN=CIFSQA.LAB.NETAPP.COM
CIFS_USER=labclient
CIFS_PASSWD=sundance
AD_SERVER_USER=adadmin
AD_SERVER_PASSWORD=adadmin
AD_SERVER=CIFSQA-DC-1,CIFSQA-DC-2,CIFSQA-DC-3
AD_SERVER_IP=172.19.2.30,172.19.3.32,172.19.3.34
my_cifs_test_script.thpl


Fully qualified domain should be used instead of short domain name. For example CIFSQA.LAB.NETAPP.COM instead of just CIFSQA. Library will convert fully
qualified domain name into short and long versions when talking to domain controller

=head1 METHODS

=cut

## no critic(RequireFilenameMatchesPackage)
package NOTE::Client;

use strict;
use warnings;
no warnings 'redefine';

use Tharn;
use POSIX qw();
use GNOTE;

our @ISA = qw(NOTE);

use Digest::MD5 qw(md5_hex);
use File::Basename;
use Data::Dumper;
use Carp qw(confess);
use dp::snapmirror::SetupCifs;
use nfs::ACL::AclNFSV4 qw( :ACE4_TYPES :ACE4_FLAGS :ACE4_MASK );
use blocks_filer;
use SayWrap;
use TracePkg;
use Options;

my $Trace = TracePkg->new();

# 7-mode CIFS parameter - not settable from CLI
my %param = (
    "CIFS_DOMAIN"              => "lab.netapp.com",
    "CIFS_DOMAIN_ADMIN"        => "labclient",
    "CIFS_DOMAIN_ADMIN_PASSWD" => "sundance",
    "CIFS_PASSWD"              => "sundance"

    ,
    "CIFS_USER"  => "Administrator",
    "FILER_CONN" => "#rsh-socket",
    "CMDTIMEOUT" => 120,
    "TIMEOUT"    => 600
);

# 10-mode CIFS parameters - CLI driven
my @cifs_data;

my $options = new Options;

# ADMIN NOTES:
# dpgqa-dc1.lab.netapp.com is the pri. DC and DNS server dpgqa-dc2.lab.netapp.com is the sec. DC
#
# Domain name is dpgqa.lab.netapp.com, so the two servers above will be dpgqa-dc1.lab.netapp.com,
# and dpgqa-dc2.lab.netapp.com within the domain.
#
#   dpgqa-dc1.lab.netapp.com -> 172.17.130.93
#   dpgqa-dc2.lab.netapp.com -> 172.17.128.43
#
# username: administrator
# password: sundance
#
# connect to the two machines via remote desktop to dpgqa-dc1.lab.netapp.com or dpgqa-dc2.lab.netapp.com
#
# To access the machines within the domain, two things need to be setup on your clients:
# 1. dns server 172.17.130.93 needs to be added to this list of dns servers, via
#    control panel-> network connections -> nic properties -> internet protocol -> advanced -> dns
#
# 2. append the dpgqa.lab.netapp.com suffix in the same settings page as above.

$options->parameters(
    definitions => [
        {
            name        => 'CIFS_DOMAIN',
            description => 'CIFS domainname. Default: LAB',
            default     => "DPGQA.LAB.NETAPP.COM",
        },
        {
            name        => 'CIFS_USER',
            description => 'CIFS username. Default: administrator',
            default     => "administrator",
        },
        {
            name        => 'CIFS_PASSWD',
            description => 'CIFS password. Default: sundance',
            default     => "sundance",
        },
#	{
#	    name        => 'CIFS_OU',
#	    description => 'Organization Unit. e.g. dpgqa',
#	    default     => 'dpgqa'
#        },

    ],
);

$options->process();

my $params = $options->get_hash();

# Transfer all CLI parameters to an array
foreach my $name ( "DOMAIN", "USER", "PASSWD" ) {
    my $tmp = $params->{ "CIFS_" . $name };
    if ( ref($tmp) =~ /ARRAY/ ) {
        push @cifs_data, join ",", @{$tmp};
    }
    else {
        push @cifs_data, $tmp;
    }
}
# use CIFS_OU instead OU, this is because CIFS.pm will set
# it anyway if OU will be defined
my $tmp = $params->{CIFS_OU} || q{};
push @cifs_data, $tmp unless ($tmp =~ /undef/);


# autoload is really private. Using it as a bridge to GNOTE::Client::Windows or GNOTE::Client::Unix via RPC
our $AUTOLOAD;
sub AUTOLOAD {
    my ($self,@args) = @_;

    ( my $name = $AUTOLOAD ) =~ s/.*://g;
    my $trace = $NOTE::TRACELOG->start_timer(message=>"rpc:$name");  # log internal metrics
    $name = $self->{_rpc}->{_lib} . '::' . $name;
    $self->_rpc_exec($name,@args);
}


# private function
sub _rpc_exec {
    my ($self,$name,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer(message=>"rpc:$name");  # log internal metrics


    unless( $name =~ /::/ ) {						    # tecnically there is no limitation what code to execute
	$name = $self->{_rpc}->{_lib} . '::' . $name;			    # accept it always suppose to be an absolute path to
    }									    # library function. Use one stored in RPC if nothing was given

    my @perl_eval_args;							    # separate local and remote arguments if separator
    if( grep{ /--/ } @args ) {						    # present, separator is first accurance of '--'
	while( my $tmp = shift @args ) {				    # basically same principals as in subtest()
	    last if( $tmp eq '--' );					    # loop moves parameters from one array to another.
	    push @perl_eval_args, $tmp;					    # If separator is not present, then all parameters
	}								    # assumed to be remote
    }

    confess("Unable to construct hash due to odd number of parameters for perl_eval request") if( scalar(@perl_eval_args) & 1 );

    ( my $params = Dumper(\@args) ) =~ s/(^\$VAR1 = \[|\];$)//gs;	    # convert array of parameters into single string, and only then
    my %perl_eval_args = ( @perl_eval_args, code => "$name($params)" );	    # build complete hash for perl_eval using both local and remote arguments

    NOTE::logcomment(Dumper(\%perl_eval_args));

    return $self->{_rpc}->{_handle}->perl_eval( %perl_eval_args );
}


=head2 ::new

Given a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
or a
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object, returns a client object that uses it as to root for all operations.

=over

=item arguments

B<name> => I<string>

Host name of client. Default: localhost

B<storage> => I<filer storage object>

A NOTE::Volume or NOTE::Qtree object when the client will create 'objects' on.

=item Examples

=back

    my $client = NOTE::Client->new(storage=>$qt_obj);

    my $cifs = NOTE::Client->new(name=>"win_client", storage=>$qt_obj);

=cut

use TestEnv::MinWin qw(:all);

# Returns an array consisting of CIFS_DOMAIN, CIFS_USER and CIFS_PASSWD CLI parameter
sub _cifs_params {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return @cifs_data;
}

sub _gx_storage {
    my ( $self, $storage ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $storage ||= $self->{_storage};
    return ( ref $storage ) =~ /^NOTE::GX::/;
}

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    $self->{_storage} = $args{storage};
    $self->{_args}{client} = $args{name} || "localhost";

    confess "storage must be a NOTE::Volume or NOTE:Qtree object, Object is a " . ( ref $args{storage} )
      unless ( ref $args{storage} ) =~ /^NOTE::.*(Volume|Qtree)/;

    ( $self->{_CIFS_DOMAIN}, $self->{_CIFS_USER}, $self->{_CIFS_PASSWD} ) = _cifs_params();

    $self->{_rpc} = NOTE::Client::RPC->new( client => $self->{_args}{client} );	    # create RPC STAF handle

    NOTE::logcomment( "Using Client: " . $self->{_args}{client} );
    my $is_gx = _gx_storage( $class, $args{storage} );

    if ( host( $self->{_args}{client} )->hosttype =~ /windows/i ) {
        NOTE::logcomment("Client: CIFS start setup");

	Tharn::host($self->{_args}{client});						# make sure STAF and RSH are running under same user name by running 'set' command
	my ( $envcmp, $enverror, $conn ) = ( undef, undef, Tharn::connect('#rsh') );	# via RSH and STAF's RPC, then convert output to hash and compare hashes
	%{$$envcmp{rsh}} = map {split('=', $_)} grep(/^(USER|HOMEPATH|LOGONSERVER)/, split( /[\n\r]/, join('',Tharn::say($conn, "set"))));
	%{$$envcmp{rpc}} = map {split('=', $_)} grep(/^(USER|HOMEPATH|LOGONSERVER)/, split( /[\n\r]/, $$self{_rpc}{_handle}->perl_eval(code=>'GNOTE::Client::Windows::showenv();')));
	map { $enverror++ if $$envcmp{rsh}{$_} ne $$envcmp{rpc}{$_} } ( keys %{$$envcmp{rsh}}, keys %{$$envcmp{rpc}} );
	confess "STAF/RSH environment mismatch\n". Dumper( $envcmp ) if $enverror;

        if ($is_gx) {
            $self->{_cifs} = "have_gx_cifs";    # Dummy cifs entry

            $self->{_storage}->Vserver->cifs_setup( storage => $self->{_storage} );
            $self->{_drive} = $self->{_storage}->map_share(
                client      => $self->{_args}{client},
                CIFS_DOMAIN => $self->{_CIFS_DOMAIN},
                CIFS_USER   => $self->{_CIFS_USER},
                CIFS_PASSWD => $self->{_CIFS_PASSWD}
            );
        } else {
            $self->{_storage}->NOTE::Qtree::security("mixed");
            $self->{_cifs} = dp::snapmirror::SetupCifs->new(
                filer  => $self->{_storage}->Filer->hostspec,
                client => $self->{_args}{client},
                params => \%param
            );
            $self->{_drive} = $self->{_cifs}->create_and_map_share( qtree => $self->{_storage} );
        }
    } else {
        unless ( $args{skip_remount} ) {
            NOTE::logcomment( "Client: Remount storage on client " . $self->{_args}{client} );
            my %opts;

            # Filer / Node options
            if ($is_gx) {
                $opts{'security-style'}   = 'mixed';
                $opts{'unix-permissions'} = '---rwxrwxrwx';
            } else {
                $opts{exportfs} = "rw,nosuid,anon=0";
            }

            # nfs side options
            $opts{chmod} = "777";

            $self->{_storage}->umount( client => $self->{_args}{client} ) if $self->{_storage}->mounted( client => $self->{_args}{client} );
            $self->{_storage}->mount( client => $self->{_args}{client}, %opts );
        }
    }


    $self->{_owner} = $$;
    bless $self, $class;
    NOTE::reference($self);

    return $self;
}

=head2 ::description

Returns a text description of this object consisting of the client name and storage name

=over

=item Arguments

I<None>

=item Example

     my $name = $client->description;

=back

=cut

sub description {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    return "Client " . $self->{_args}{client} . ", Storage " . $self->storage->description;
}

=head2 ::mountpoint

If Unix client then returns the mountpoint. For a windows client returns the drive letter

=over

=item Arguments

None.

=item Example

      $client->mountpoint;

=back

=cut

sub mountpoint {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    return (
          $self->{_cifs}
        ? $self->{_drive}
        : $self->{_storage}->mountpoint( client => $self->{_args}{client} )
    );
}

=head2 ::fullpath

Given a path will return the full path from the root of the 'storage' tree.

=over

=item Arguments

B<rel_path> = I<path>

Relative path

=item Example

      my $fpath = $client->fullpath("/abc/efg");

=back

=cut

sub fullpath {
    my ( $self, $rel_path ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    $rel_path = "/" . $rel_path if $rel_path !~ /^\//;
    my $fp = $self->mountpoint . $rel_path;
    $fp =~ s/\//\\/g
      if $self->{_cifs};
    NOTE::logcomment("fullpath: $fp");
    return $fp;
}

=head2 ::storage

Returns the storage object being used by this object. It is either a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
or a
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object.

=over

=item Argument

None.

=item Example

      my $note_obj = $client->storage

=back

=cut

sub storage {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_storage};
}

=head2 ::get_handle

Returns an object to be used by this object as a handle to the 'path' name
supplied much like 'open' returning a file handle. This handle will be used
by other methods to identify the object represented by pathname 'path'.

'path' is relative to the root of the 'storage' object used by this Client
object

This method uses diag level commands so can be used in testing promoted
builds (burt 490236).

=over

=item Argument

B<path> = I<string>

Relative pathname to root of storage object in Client.

=item Example

      my $handle = $client->get_handle("/a_file");

=back

=cut

sub get_handle {
    my ( $self, $path ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $handle = {};
    $handle->{path} = $path;
    bless $handle, 'NOTE::Client::handle';

    my $txt;
    my $priv_cmd = 'priv set -q diag';
    if ( $self->_gx_storage() ) {
        eval {
            # Get inode info for path
            my @inode_info = $self->storage->wex_info( path => $path );
            # Get inode list for path's streams
            my %stream_info = $self->storage->_stream_list( \@inode_info );
            map { $handle->{stream}{$_} = 1; } keys %stream_info;
        };

        # Ignore error if file does not exists
        if ( $@ && ($@ !~ /unable to find name/i) ) {
            confess "Failed to get stream info for $path. Error: $@";
        }
    }
    else {
        # this shoud probably use the Volume::slist method
        $path = $self->storage->mount_path . "/" . $path;
        $txt = SayWrap::sayEOF( $self->storage->Filer->conn,
                                "$priv_cmd; slist $path" );
        unless ( $txt =~ /invalid path/i ) {
            my @line = split /\n/, $txt;
            my $i = 0;

            # The format of 'slist' is as follows:
            #   Inode#        File size      Name
            # -----------  ---------------   ----
            # <fileInode>  <file size>       <fullpath of file>
            # <streamInode><stream size>     <stream name>
            # <next stream entry, etc>
            while ( $i < @line ) {
                # Skip header and file entry
                if ( $i > 2 ) {
                    # obtain just the stream name
                    $line[$i] =~ s/\s+\d+\s+\d+\s+(\S+)/$1/;
                    $handle->{stream}{ $line[$i] } = 1;
                }
                $i++;
            }
        }
    }
    return $handle;
}

=head2 ::populate

Depending on the type of the client it will generate NFS or CIFS data using RichData or DataManip. Default
behavior is to use RichData. Use NOTE_DATAMANIP environment variable to force DataManip. DataManip will
ignore CIFS clients.

=over

=item Arguments

B<%args> - I<RichData or DataManip argument Hash>

I<Optional.>

All parameters, in the form of a hash is passed to RichData::populate for
processing

I<background> => I<boolean>

Indicates whether to run data population in the background or not.
Default: 0 (foreground)

=item Examples

=back

    $client->populate;
    $client->populate(path=>"/subdir"); # Populate sub directory 'subdir' within storage

=cut

sub populate {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $background = delete $args{background} || 0;
    my $eval_success = eval {
        if ( host( $self->{_args}{client} )->hosttype =~ /windows/i )
        {
            $args{path} = $self->mountpoint . '\\\\' . $args{path};

            # XXX Workaround for Burt 268195 - Do not generate any
            # hard link creations from windows.
            $args{link_pct} = 0;

            # Per BURT 359759 - streams not supported for coral volumes
            if ( $self->storage->is_coral ) {
                $args{'dir_stream_pct'}  = 0;
                $args{'file_stream_pct'} = 0;
            }
        }
        else {
            $args{path} = $self->mountpoint . "/$args{path}";
        }

        $self->{_subtest} = Tharn::subtest(
            sub {
                if (NOTE->DATAMANIP == 0 ||
                    host($self->{_args}{client})->hosttype =~ /windows/i) {
                    # Launder %args so that only names that RichData knows are
                    # in it when it is passed to populate(). The set of args
                    # we are passed are meant for either RichData or DataManip,
                    # without knowing which one will be called at the time. So
                    # there are duplications in some cases, and in other cases
                    # there are options for one that are not valid for the
                    # other.
                    #
                    # Only the arg-names are laundered. The args content is
                    # not examined at this level.

                    # Start by defining the list of known args. Any changes to
                    # this list should be made just in the qw() list here:
                    my @valid_args =
                        qw(path template name_avg name_dev dir_avg dir_dev
                           file_avg file_dev file_size_avg file_size_dev
                           depth_avg depth_dev stream_avg stream_dev
                           stream_size_avg stream_size_dev dir_acl_num
                           dir_acl_pct file_acl_num file_acl_pct
                           dir_stream_num dir_stream_pct file_stream_num
                           file_stream_pct sparse_num sparse_pct link_num
                           link_pct slink_num slink_pct compress_pct
                           encrypt_num encrypt_pct archive_num archive_pct
                           hidden_num hidden_pct acl_opts sparse_opts fill_vol
                           timeout dbfile);
                    # Turn the list into a hash table for easy testing:
                    my %valid_args = map { $_ => 1 } @valid_args;
                    # Launder %args:
                    my %clean_args;
                    for my $key (keys %args) {
                        $clean_args{$key} = $args{$key} if $valid_args{$key};
                    }

                    NOTE::logcomment(
                        "Using RichData to populate data at $args{path}"
                    );
                    RichData::populate(
                        client => $self->{_args}{client},
                        %clean_args
                    );
                    Tharn::logresult(FATAL => "RichData Populate Error: $@")
                        if $@;
                } else {
                    # Launder %args so that only names that DataManip knows are
                    # in it when it is passed to populate_tree(). The set we
                    # are passed are meant for either RichData or DataManip,
                    # without knowing which one will be called at the time. So
                    # there are duplications in some cases, and in other cases
                    # there are options for one that are not valid for the
                    # other.
                    #
                    # Only the arg-names are laundered. The args content is
                    # not examined at this level.
                    #
                    # Note that DataManip wants args that are all upper-case.
                    # We do that during the laundering process, but since they
                    # come in to here lower-case, we compare them as lower.

                    # Start by defining the list of known args. Any changes to
                    # this list should be made just in the qw() list here:
                    my @valid_args =
			qw( breadth_avg breadth_dev create_dir_avg
			    create_dir_dev create_file_avg create_file_dev
			    create_file_size_avg create_file_size_dev depth_avg
			    depth_base depth_dev dir_avg dir_dev 
			    environment_type file_avg file_dev flood fork_count 
			    fragment_avg fragment_dev fragment_raw grow_avg 
			    grow_dev grow_raw log_file min_dir_size mod_avg 
			    mod_dev mod_raw persevere scramble seed shrink_avg 
			    shrink_dev shrink_raw subtree_depth_avg 
			    subtree_depth_dev total_del total_grow total_mod 
			    total_shrink total_subtree_del verbose );
                    # Turn the list into a hash table for easy testing:
                    my %valid_args = map { $_ => 1 } @valid_args;
                    # Launder %args:
                    my %clean_args;
                    for my $key (keys %args) {
                        $clean_args{uc $key} = $args{$key} if $valid_args{$key};
                    }

                    my $path = delete $args{path};
                    NOTE::logcomment(
                        "Using DataManip to populate data at $args{path}"
                    );
                    DataManip::populate_tree(
                        $self->{_args}{client},
                        $path,
                        %clean_args
                    );
                    Tharn::logresult(FATAL => "DataManip error: $@" ) if $@;
                }
            },
            -bg
        );

        if ($background) {
            NOTE::logcomment('Population running in background...');
        }
        else {
            $self->wait;
        }

        1;
    };
    confess $@ if (! $eval_success);

    return;
}

=head2 ::modify

Depending on the type of the client it will use RishData or DataManip to modify data generated by
the previous call to 'populate'

=over

=item Arguments

B<%args> - I<RichData or DataManip argument Hash>

I<Optional.>

All parameters, in the form of a hash is passed to RichData::modify for
processing

I<background> => I<boolean>

Indicates whether to run data population in the background or not.
Default: 0 (foreground)

=item Examples

=back

    $client->modify;
    $client->modify(path=>"/subdir"); # Modify sub directory 'subdir' within storage

=cut

sub modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $background = delete $args{background} || 0;
    eval {
        if ( host( $self->{_args}{client} )->hosttype =~ /windows/i )
        {
            $args{path} = $self->mountpoint . '\\\\' . $args{path};

            # XXX Workaround for Burt 268195 - Do not generate any
            # hard link creations from windows.
            $args{link_pct} = 0;

            # Per BURT 359759 - streams not supported for coral volumes
            if ( $self->storage->is_coral ) {
                $args{'dir_stream_pct'}    = 0;
                $args{'file_stream_pct'}   = 0;
                $args{'modify_stream_pct'} = 0;
            }
        }
        else {
            $args{path} = $self->mountpoint . "/" . $args{path};
        }

        $self->{_subtest} = Tharn::subtest(
            sub {
		if( NOTE->DATAMANIP == 0 || host( $self->{_args}{client} )->hosttype =~ /windows/i ) {
		    NOTE::logcomment( "Use RichData to modify data at " . $args{path} );
		    RichData::modify( client => $self->{_args}{client}, %args );
		    Tharn::logresult( "FATAL", "RichData Modify Error: $@" ) if $@;
		} else {
		    my %ARGS = map{ uc($_), $args{$_} } keys %args;
		    my $path = delete $args{path};
		    DataManip::modify_tree( $self->{_args}{client}, $path,%ARGS );
		    Tharn::logresult( "FATAL", "DataManip error: $@" ) if $@;
		}
            },
            -bg,
            -onexit,
            sub { }
        );
        if ($background) {
            if( NOTE->DATAMANIP == 0 || host( $self->{_args}{client} )->hosttype =~ /windows/i ) {
                NOTE::logcomment("Data modification by RichData running in background...");
            }
            else {
                NOTE::logcomment("Data modification by DataManip running in background...");
            }
        }
        else {
            $self->wait;
        }
    };
    confess $@ if $@;
}

=head2 ::wait

Wait for client subtest to complete

=over

=item Arguments

I<None>

=item Example

    $self->populate(background=>1);
	$self->wait;

=back

=cut

sub wait {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    confess "No background process running..."
      unless exists $self->{_subtest};
    my $ret = delete $self->{_subtest};
    $ret->wait_finish();
    FSR::FSRutils::subtestResults($ret) or confess $@;
    NOTE::logcomment("Client subtest completed");
}

=head2 ::write_file

Writes data to file given by '$handle'. Wrapper to
L<C<FileManip::write_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_file>

Croaks if failed.

=over

=item Arguments

B<handle> = I<file handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>
method.

B<args> = I<args>

Hash of arguments to determine what and how it is written. See
L<C<FileManip::write_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_file>
for details. Please make sure to provide oflags=>'O_CREAT' when creating new file.

=item Example

     $client->write_file($handle, contents => "some data");

=back

=cut

sub write_file {
    my ( $self, $handle, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "<client handle> is required"
      unless $handle;

    $args{file}   = $self->fullpath( $handle->{path} );
    $args{client} = $self->{_args}{client};

    if ( $self->_rpc_exec("write_file", %args) ) {
        return $handle;
    }
    else {
        confess "Failed to write data to file $handle->{path}";
    }
    confess "Error detected while writing file $handle->{path}: $@" if $@;
}

=head2 ::read_file

Read data from file given by '$handle'. Wrapper to
L<C<FileManip::read_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/read_file>

Croaks if failed.

=over

=item Arguments

B<handle> = I<file handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>
method

B<args> = I<args>

Hash of arguments to determine what and how content is being read. See
L<C<FileManip::read_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/read_file>
for details.

=item Example

     $client->read_file($handle);

=back

=cut

sub read_file {
    my ( $self, $handle, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "<client handle> is required"
      unless $handle;

    $args{file}   = $self->fullpath( $handle->{path} );
    $args{client} = $self->{_args}{client};
    return $self->_rpc_exec("read_file", %args); # Will execute FileManip::read_file(%args) on remote client via RPC
}

=head2 ::add_stream

Add a stream file to an object indicated by '$handle'. Wrapper to
L<C<FileManip::add_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/add_stream>

Returns a modified handle for the object. Croaks if failed.

=over

=item Argument

B<handle> = I<file/directory handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stream_name> = <string>

Name of the stream file to create

=item Example

      my $handle = $client->get_handle("/a_file_with_stream");
      $handle = $client->add_stream($handle, "stream_file");

=back

=cut

sub add_stream {
    my ( $self, $handle, $stream_name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %info;
    confess "stream $stream_name already exists for file " . $handle->{path}
      if exists $handle->{stream}{$stream_name};

    return $self->_write_stream( $handle, $stream_name, 0, undef, 0 );
}

=head2 ::write_stream

Writes to the named stream. Returns handle or confess if detected error. Wrapper to
L<C<FileManip::write_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_stream>

=over

=item Arguements

B<handle> = <object handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stram_name = <string>

Name of the stream to be written to.

B<offset> = <integer>

Offset to start writing, starting from 0. Default 0

B<size> = <integer>

Number of bytes to write. Default 1

=item Example

      $handle = $client->write_stream ($handle, "stream_file", 4, 5);

=back

=cut

sub write_stream {
    my ( $self, $handle, $stream_name, $offset, $size ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $offset = $offset || 0;
    $size   = $size   || 1;
    return $self->_write_stream( $handle, $stream_name, 0, $offset, $size );
}

sub _write_stream {
    my ( $self, $handle, $stream_name, $append, $offset, $size ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $data = "a" x $size;
    my %opts;
    $opts{file}     = $self->fullpath( $handle->{path} );
    $opts{stream}   = $stream_name;
    $opts{append}   = $append;
    $opts{offset}   = $offset if defined $offset;
    $opts{contents} = $data;
    $opts{client}   = $self->{_args}{client};

    FileManip::write_stream(%opts);
    $handle->{stream}{$stream_name} = 1;

    return $handle;
}

=head2 ::delete_stream

Removed stream(s) from '$handle' object and returns the modified '$handle' object. Wrapper to
L<C<FileManip::remove_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/remove_stream>

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stream_name> = <string>

If supplied named the stream to be remove, otherwise removes all streams from $handle

=item Example

     $handle = $client->delete_stream($handle, 'a_stream_to_be_remove');
     $handle = $client->delete_stream($handle);

=back

=cut

sub delete_stream {
    my ( $self, $handle, $stream_name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list;
    if ( defined $stream_name ) {
        confess "No stream $stream_name found for file " . $handle->{path}
          unless defined $handle->{stream}{$stream_name};
        push @list, $stream_name;
    }
    else {
        push @list, keys %{ $handle->{stream} };
    }

    foreach my $stream (@list) {
        eval { FileManip::remove_stream( file => $self->fullpath( $handle->{path} ), stream => $stream, client => $self->{_args}{client} ); };
        confess $@ if $@;

        delete $handle->{stream}{$stream};
    }
    return $handle;
}

=head2 :rename_stream

Renames a named stream. Wrapper for
L<C<FileManip::rename_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/rename_stream>
Returns the modified '$handle' object or confess.

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<old_stream> = <string>

stream to rename

B<new_stream> = <string>

new name of stream

=item Example

      $handle = $cleint->rename_stream($handle, 'this_stream', 'new_stream_name');

=back

=cut

sub rename_stream {
    my ( $self, $handle, $old_stream, $new_stream ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # XXX This supposedly only works for unix-solaris m/c, but it does not work
    confess "stream $old_stream does not exists for file " . $handle->{path}
      unless exists $handle->{stream}{$old_stream};
    confess "new stream $new_stream already exists for file " . $handle->{path}
      if exists $handle->{steam}{$new_stream};

    eval { FileManip::rename_stream( file => $self->fullpath( $handle->{path} ), stream => $old_stream, nstream => $new_stream ); };
    confess $@ if $@;

    $handle->{stream}{$new_stream} = 1;
    return $handle;
}

=head2 ::set_file_attr

Sets the attribute of a file. Uses
L<C<FileManip::setattr_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/setattr_file>
to set the attribute of file

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<args> = I<args>

Hash of arguments to determine what and how to set attribute. See
L<C<FileManip::setattr_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_file>
for details.

=item Example

      $client->set_file_attr( $handle, hidden => 1 );

=back

=cut

sub set_file_attr {
    my ( $self, $handle, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "<client handle> is required"
      unless $handle;

    $args{file}   = $self->fullpath( $handle->{path} );
    $args{client} = $self->{_args}{client};
    if ( FileManip::setattr_file(%args) ) {
        return $handle;
    }
    else {
        confess "Failed to write data to file $handle->{path}";
    }
    confess "Error detected while writing file $handle->{path}: $@" if $@;
}

=head2 ::set_stream_size

Sets the size of a stream. Uses
L<C<FileManip::setattr_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/setattr_stream>
to set the stream size

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stream_name> = I<string>

Stream to set the size to

B<size> = <integer>

Size to set the stream to.

=item Example

      $handle = $client->set_stream_size($handle, "stream_to_change", 4096);

=back

=cut

sub set_stream_size {
    my ( $self, $handle, $stream_name, $size ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "stream $stream_name does not exists for file " . $handle->{path}
      unless exists $handle->{stream}{$stream_name};

    eval { FileManip::setattr_stream( file => $self->fullpath( $handle->{path} ), stream => $stream_name, size => $size, %{ $self->{_args} } ); };
    confess $@ if $@;

    return $handle;
}

=head2 ::create_with_acl

=cut

sub create_with_acl {
    my ( $self, $handle, %acl ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %args;
    foreach( qw( mode contents oflags fstat uid gid offset rtn_hash ) ) { $args{$_} = delete $acl{$_} if $acl{$_} } # split acl and write_file args
    $args{oflags} = 'O_CREAT' unless( $args{oflags} =~ /\w+/ );

    $self->write_file($handle, %args);
    $self->set_acl($handle, %acl);

    return $handle;
}

=head2 ::get_acl

Request ACL hash via RPC and store it in handle structure $handle->{acl} as hash reference to all ACL's ACEs. CIFS ONLY, UNIX is pending.

=over

=item Arguments

B<handle> = I<file/directory handle>

L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

=item Example

 my $cifs = NOTE::Client->new(name=>$WINDOWS_CLIENT, storage=>$vol);
 my $handle = $cifs->write_file( $cifs->get_handle("/whats_up_new_file") );
 $cifs->get_acl($handle);
 logcomment( "Existing ACL:\n", Dumper( $handle->{acl} ));

Output from Windows should be something like this:

 20090913 093537 [0001_gnote_test] .         $VAR1 = {
 20090913 093537 [0001_gnote_test] .                   'Everyone' => [
 20090913 093537 [0001_gnote_test] .                                   'GENERIC_READ'
 20090913 093537 [0001_gnote_test] .                                 ],
 20090913 093537 [0001_gnote_test] .                   'NETAPP\\cpeter' => [
 20090913 093537 [0001_gnote_test] .                                         'DELETE',
 20090913 093537 [0001_gnote_test] .                                         'READ_CONTROL',
 20090913 093537 [0001_gnote_test] .                                         'SYNCHRONIZE',
 20090913 093537 [0001_gnote_test] .                                         'STANDARD_RIGHTS_READ',
 20090913 093537 [0001_gnote_test] .                                         'STANDARD_RIGHTS_WRITE',
 20090913 093537 [0001_gnote_test] .                                         'STANDARD_RIGHTS_EXECUTE',
 20090913 093537 [0001_gnote_test] .                                         'GENERIC_READ',
 20090913 093537 [0001_gnote_test] .                                         'GENERIC_WRITE',
 20090913 093537 [0001_gnote_test] .                                         'GENERIC_EXECUTE',
 20090913 093537 [0001_gnote_test] .                                         'READ',
 20090913 093537 [0001_gnote_test] .                                         'CHANGE',
 20090913 093537 [0001_gnote_test] .                                         'ADD'
 20090913 093537 [0001_gnote_test] .                                       ]
 20090913 093537 [0001_gnote_test] .                 };

=back

=cut

sub get_acl {
    my ( $self, $handle ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $path = $self->fullpath( $handle->{path} );
    $handle->{acl} = $self->Get_ACL_Mask($path);
    return $handle;
}

=head2 ::set_acl

Sends RPC call to client requesting to set ACL and then reads it again. However no
verification is done between set/received ACEs. Windows may ignore some ACEs if there
not enough privileges and will only set permitted ACEs without throwing an error.
Returned ACLE is stored in handle structure $handle->{acl} as hash reference to all
ACL's ACEs. CIFS ONLY, UNIX is pending.

=over

=item Arguments

B<handle> = I<file/directory handle>

L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<acl> = I<hash of arrays> where each key is a user name and value is an array of permission
constants. Allowed valuse for Windows:

    DELETE, READ_CONTROL, WRITE_DAC, WRITE_OWNER,
    SYNCHRONIZE, STANDARD_RIGHTS_REQUIRED,
    STANDARD_RIGHTS_READ, STANDARD_RIGHTS_WRITE,
    STANDARD_RIGHTS_EXECUTE, STANDARD_RIGHTS_ALL,
    SPECIFIC_RIGHTS_ALL, ACCESS_SYSTEM_SECURITY,
    MAXIMUM_ALLOWED, GENERIC_READ, GENERIC_WRITE,
    GENERIC_EXECUTE, GENERIC_ALL, F, FULL, R, READ,
    C, CHANGE

=item Example

 $cifs = NOTE::Client->new(name=>$WINDOWS_CLIENT, storage=>$vol);
 my $handle = $cifs->write_file( $cifs->get_handle("/whats_up") );
 $cifs->get_acl($handle);
 logcomment( "Existing ACL:\n", Dumper( $handle->{acl} ));
 $cifs->set_acl($handle,( 'NETAPP\cpeter' => [ qw( CHANGE GENERIC_WRITE GENERIC_READ GENERIC_EXECUTE ) ],
                          'Everyone' => [ qw( GENERIC_READ ) ] ) );
 logcomment( "New ACL:\n", Dumper( $handle->{acl} ));

=back

=cut

sub set_acl {
    my ( $self, $handle, %ace ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $path = $self->fullpath( $handle->{path} );
    $handle->{acl} = $self->Set_ACL_Mask($path,\%ace);
    return $handle;
}

sub _Volume {
    my ($obj) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return ( ( ref $obj ) =~ "NOTE::.*Volume" ? $obj : $obj->Volume );
}

sub _run_ipc {
    my ( $self, $cmd, @args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    unless ( $self->{_ipc} ) {
        my @extra_options;
        my $script = qq{ \{
		use Tharn;
		use nfs::ACL::AclNFSV4 qw( :ACE4_TYPES :ACE4_FLAGS :ACE4_MASK );
		use nfs::ACL::Acl qw( get_ace_array );
		use nfs::GenericAcl qw( get_acl get_acl_array create_with_acl
					set_acl set_acl_ace clear_acl );

		Subtest::ipc()->serve();

		sub result \{
		    my (\$data) = \@_;

		    my \$rc = \{\};
		    \$rc->\{error\} = \$\@;
		    \$rc->\{data\} = \$data;
		    bless \$rc, "ipc_rc";
		    return \$rc;
		\}

		sub get_fconn \{
		    my (\$conn_spec) = \@_;

		    return connect(\$conn_spec, -timeout, 60);
		\}

		sub remote_get_acl \{
			my (\$conn_spec, %args) = \@_;

			\$args\{fc\} = get_fconn(\$conn_spec);
			my \$acl;
			eval \{ \$acl = get_acl(%args); \};
			delete \$acl->\{fc\} if \$acl;
			return result(\$acl);
		\}

		sub remote_create_with_acl \{
			my (\$conn_spec, %args) = \@_;

			\$args\{fc\} = get_fconn(\$conn_spec);
			my \@file_aces = get_ace_array(type=>\$args\{filetype\}."_aces");
			\$args\{ace\} = \\\@file_aces;
			my \$acl;
			eval \{ \$acl = create_with_acl(%args); \};
			delete \$acl->\{fc\} if \$acl;
			return result(\$acl);
		\}

		sub remote_set_acl \{
			my (\$conn_spec, %args) = \@_;

			my \%ace = %\{\$args\{ace\}\};
			my \$acl = \$args\{acl\};
			\$acl->\{fc\} = get_fconn(\$conn_spec);
			eval \{ set_acl_ace (acl=>\$acl, %ace); \};
			return result() if \$\@;
			my \$rc;
			eval \{ \$rc = set_acl (acl=>\$acl); \};
			return result() if \$\@;
			\$\@ = \$rc unless \$rc =~ /NFS4_OK/;
			delete \$acl->\{fc\};
			return result(\$acl);
		\}

		sub remote_clear_acl \{
			my (\$conn_spec, \$del_acl) = \@_;

			\$del_acl->\{fc\} = get_fconn(\$conn_spec);
			eval \{ clear_acl(acl=>\$del_acl); \};
			return result();
		\}

	\} };
        NOTE::logcomment("remote script: $script");
        $self->{_ipc} = Tharn::subtest( $script, '-bg', '-ipc', '-client', $self->{_args}{client}, '-runid', "01_Client", '-onexit', sub { }, @extra_options );
    }

    #    my $conn_spec = $self->{_storage}->Filer->hostspec.
    #			"#".$self->{_storage}->Filer->conn->conntype();
    NOTE::logcomment( "Calling $cmd, with: ", Dumper(@args) );

    #    my $rc = $self->{_ipc}->ipc->call("remote_".$cmd, $conn_spec, @args);
    my $rc = $self->{_ipc}->ipc->call( "remote_" . $cmd, $self->{_storage_address}, @args );
    confess "$cmd failed. Error: " . $rc->{error} if $rc->{error};
    return $rc->{data};
}

###############################################################################
#
#   Sub Name:       _create_lock_mgr
#
#   Description:    Create the lock-manager closure that lock_mgr will use to
#                   marshall requests from the user.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub _create_lock_mgr {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    require ClientRPC::FileOperations;

    # These are needed to track the locks between calls to the manager:
    my %locks = (
        anonymous => {
            exclusive => {
                read  => [],
                write => [],
            },
            shared    => {
                read  => [],
                write => [],
            },
        },
        'byte-range'  => {},
        'share-level' => {},
    );
    # This is used several times, so might as well only look it up once:
    my $clientname = $self->{_args}{client};
    # The filename pattern used for anonymous locks:
    my $filebase = "lock_mgr_${clientname}_$$";
    # A counter used to create filename along with $filebase, above:
    my $filecount = 0;
    # This is derived outside the closure so that it need not be re-computed
    # over and over:
    my $MNTPT = $self->mountpoint;
    # We'll need to know whether this is a Windows client or not:
    my $IS_WIN = host($clientname)->hosttype =~ /windows/i;
    # Set the file-separator on $MNTPT depending on $IS_WIN:
    if ($IS_WIN) {
        $MNTPT .= q{\\};
    } else {
        $MNTPT .= q{/};
    }
    # Valid values for the 'lock' parameter:
    my %LOCKS = (
        'byte-range'   => 1,
        'share-level' => 1,
    );
    # Table mapping valid 'lock' values and their synonyms:
    my %TYPES = (
        exclusive => 'exclusive',
        write     => 'exclusive',
        shared    => 'shared',
        read      => 'shared',
    );
    # Valid values for the 'access' parameter:
    my %ACCESS = (
        'read'  => 1,
        'write' => 1,
    );
    # And we'll need a ClientRPC object to encapsulate the remote calls
    my $clientobj = ClientRPC::FileOperations->new(
        client => $clientname,
        on_err => 'rtn'
    );

    $self->{_lock_mgr} = sub {
        my %args = @_;

        my $state = {
            SELF   => $self, # Two of the actions need access to $self
            IS_WIN => $IS_WIN,
            MNTPT  => $MNTPT,
            CLIENT => $clientobj,
            LOCKS  => \%locks,
        };

        # "action" defaults to "acquire"
        $args{action} ||= 'acquire';

        # First, see if this is just a status request:
        if ($args{action} eq 'status') {
            return _lock_mgr_status(
                state => $state,
                ($args{debug} ? (debug => 1) : ())
            );
        }

        # Is it a "show" request?
        if ($args{action} eq 'show') {
            my $keep_null = (exists $args{keep_null}) ? $args{keep_null} : 0;
            return _lock_mgr_show(
                state     => $state,
                keep_null => $keep_null
            );
        }

        # Is it a shutdown request?
        if ($args{action} eq 'shutdown') {
            # It is necessary to shut down the lock manager before the client
            # can unmount the storage volume it is associated with. Else, we
            # get a "device busy" error because of the open filehandles.
            return _lock_mgr_shutdown(state => $state);
        }

        # Are we initializing some files?
        if ($args{action} eq 'initialize') {
            # First make sure we have one or more files. This cannot be done
            # on anon files.
            if (! $args{file}) {
                confess 'Within lock_mgr closure: \'initialize\' requires ' .
                    'one or more files to be specified';
            }
            if (ref($args{file}) ne 'ARRAY') {
                $args{file} = [ $args{file} ];
            }

            return _lock_mgr_initialize(
                state  => $state,
                file   => $args{file},
                data   => $args{data},
                offset => $args{offset}
            );
        }

        # Allow for defaults for the "access", "type" and "lock" arguments:
        $args{access} ||= 'read';
        $args{type}   ||= 'shared';
        $args{lock}   ||= 'share-level';

        # Validate the lock type, throwing an exception if it doesn't match
        # our known list. We quietly allow some aliases, for backwards
        # compatibility, hence the assignment against %TYPES:
        $args{type} = $TYPES{$args{type}};
        if (! $args{type}) {
            confess "Within lock_mgr closure: unknown type '$args{type}'";
        }

        # Validate the "lock" argument, throwing an exception if it isn't in
        # the table of values:
        if (! $LOCKS{$args{lock}}) {
            confess "Within lock_mgr closure: unknown lock '$args{lock}'";
        }

        # Validate the 'access' argument:
        if (! $ACCESS{$args{access}}) {
            confess "Within lock_mgr closure: unknown access '$args{lock}'";
        }

        # Whether the locks we are creating/releasing are named or anonymous:
        my $is_anon;
        # Return value after we've done the locking/unlocking
        my $retval = q{};

        if ($args{file}) {
            # These will be "named" locks, not anon files. This will matter
            # if/when the locks are released, as anon files are cleaned up.
            $is_anon = 0;

            # Normalize $args{file} into an arrayref if it isn't already:
            if (ref($args{file}) ne 'ARRAY') {
                $args{file} = [ $args{file} ];
            }

            # Make sure offset and length are at reasonable defaults:
            $args{offset} ||= 0;
            $args{length} ||= 0; # Should cause length to be max length of file
        } else {
            $is_anon = 1;

            # Firstly, throw an exception if the user requested a share-level
            # anon lock, as we don't support that. Later we'll warn the user
            # that anon locks are not recommended.
            if ($args{lock} eq 'share-level') {
                confess 'Anonymous locks must be byte-range locks';
            }

            # To make the rest of this process as vanilla as possible, make
            # the rest of %args look like it would for a named-locks call.

            # Default the number of locks to create at 1
            $args{count} ||= 1;

            # If we are acquiring locks, create $count anon filenames and push
            # them on the listref.
            if ($args{action} eq 'acquire') {
                NOTE::logresult(WARN => 'Anonymous locks are deprecated');
                for (1 .. $args{count}) {
                    my $name = sprintf '%s_%04d', $filebase, $filecount++;
                    push(@{$args{file}}, $name);
                }
            }

            # Force offset and length to defaults
            $args{offset} = 0;
            $args{length} = 0; # Should cause length to be max length of file
        }

        # Process the request, keyed by $args{action}
        if ($args{action} eq 'acquire') {
            # Acquire one or more locks, anon or named

            $retval = _lock_mgr_acquire(
                state     => $state,
                anonymous => $is_anon,
                args      => \%args
            );
        } elsif ($args{action} eq 'release') {
            # Release one or more locks, anon or named

            $retval = _lock_mgr_release(
                state     => $state,
                anonymous => $is_anon,
                args      => \%args
            );
        } else {
            # We got an 'action' that we don't know about, so throw an error
            confess "Within lock_mgr closure: unknown action '$args{action}'";
        }

        return $retval;
    };

    return $self;
}

###############################################################################
#
#   Sub Name:       _lock_mgr_status
#
#   Description:    Perform the "status" command for the lock manager.
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   The main element in %args is called "state", and represents
#                   the state of the lock_mgr. It is a hash reference with the
#                   following keys:
#
#                     SELF   => a ref to the NOTE::Client object
#                     IS_WIN => a boolean, whether this client is Windows
#                     MNTPT  => the local mount-point/device of the volume
#                     CLIENT => the RPC client object for remote execution
#                     LOCKS  => the current internal locks-table
#
#                   All the private _lock_mgr_* subroutines get this argument.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of parameters
#
#   Returns:        hashref
#
###############################################################################
sub _lock_mgr_status {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %stats = (
        share_level_shared     => 0,
        share_level_exclusive  => 0,
        share_level_total      => 0,
        byte_range_shared      => 0,
        byte_range_exclusive   => 0,
        byte_range_total       => 0,
        shared_total           => 0,
        exclusive_total        => 0,
        total                  => 0,
    );

    # For each of share-level-locks and byte-range-locks, total the count for
    # shared and exclusive lock types, and keep a running total of all:
    for my $lock (qw(share-level byte-range)) {
        (my $lockkey = $lock) =~ s/-/_/g;
        my $locks = $args{state}->{LOCKS}->{$lock};
        for my $tuple (values %{$locks}) {
            my $type = $tuple->[0];
            $stats{"${lockkey}_$type"}++;
        }
        $stats{"${lockkey}_total"} =
            $stats{"${lockkey}_shared"} + $stats{"${lockkey}_exclusive"};
        $stats{shared_total} += $stats{"${lockkey}_shared"};
        $stats{exclusive_total} += $stats{"${lockkey}_exclusive"};
        $stats{total} += $stats{"${lockkey}_total"};
    }

    if ($args{debug}) {
        NOTE::logcomment(
            'lock_mgr internal locks table: ' . Dumper($args{state}->{LOCKS})
        );
    }

    return \%stats;
}

###############################################################################
#
#   Sub Name:       _lock_mgr_show
#
#   Description:    Do a "vserver locks show" by going through the client's
#                   storage object to the underlying vserver object. Massage
#                   the returned data into a series of lists and counts.
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   The main element in %args is called "state", and represents
#                   the state of the lock_mgr. It is a hash reference with the
#                   following keys:
#
#                     SELF   => a ref to the NOTE::Client object
#                     IS_WIN => a boolean, whether this client is Windows
#                     MNTPT  => the local mount-point/device of the volume
#                     CLIENT => the RPC client object for remote execution
#                     LOCKS  => the current internal locks-table
#
#                   All the private _lock_mgr_* subroutines get this argument.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of parameters
#
#   Returns:        Success:    hash reference
#                   Failure:    dies
#
###############################################################################
sub _lock_mgr_show {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $keep_null = $args{keep_null};
    my $state     = $args{state};
    my $self      = $state->{SELF};
    my $volume    = $self->{_storage};

    my $vserver_locks = $volume->Vserver->locks_show(
        volume    => $volume,
        keep_null => $keep_null
    );
    my $locks = {
        'share-level' => [],
        'byte-range'  => [],
        'op-lock'     => [],
        delegation    => [],
        all           => [],
    };

    # Sort the locks into buckets based on the "type" field
    for my $lock (@{$vserver_locks}) {
        push(@{$locks->{$lock->{type}}}, $lock);
        push(@{$locks->{all}},           $lock);
    }
    # Count the contents of each bucket and create an MD5 hash of the lock data
    my @all_lines;
    for my $key (qw(share-level byte-range op-lock delegation)) {
        $locks->{"$key-count"} = scalar(@{$locks->{$key}});
        if ($locks->{"$key-count"}) {
            my @lines;
            for my $lock (@{$locks->{$key}}) {
                push(
                    @lines,
                    join(q{} => map { $lock->{$_} } (sort keys %{$lock}))
                );
            }
            $locks->{"$key-hash"} = md5_hex(sort(@lines));
            push(@all_lines, @lines);
        } else {
            $locks->{"$key-hash"} = q{};
        }
    }

    # Set all-count and all-hash
    $locks->{'all-count'} = scalar(@{$locks->{all}});
    if (@all_lines) {
        $locks->{'all-hash'} = md5_hex(sort(@all_lines));
    } else {
        $locks->{'all-hash'} = q{};
    }

    return $locks;
}

###############################################################################
#
#   Sub Name:       _lock_mgr_shutdown
#
#   Description:    Shutdown the lock manager. In this case, this means
#                   releasing all locks currently held (both named-file and
#                   anon) and calling the "destroy" method on the ClientRPC
#                   object.
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   The main element in %args is called "state", and represents
#                   the state of the lock_mgr. It is a hash reference with the
#                   following keys:
#
#                     SELF   => a ref to the NOTE::Client object
#                     IS_WIN => a boolean, whether this client is Windows
#                     MNTPT  => the local mount-point/device of the volume
#                     CLIENT => the RPC client object for remote execution
#                     LOCKS  => the current internal locks-table
#
#                   All the private _lock_mgr_* subroutines get this argument.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of parameters
#
#   Returns:        Success:    message string
#                   Failure:    dies
#
###############################################################################
sub _lock_mgr_shutdown {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $state     = $args{state};
    my $self      = $state->{SELF};
    my $clientobj = $state->{CLIENT};
    my $locks     = $state->{LOCKS};

    # Release the anonymous locks first, since we have to pass the 'anonymous'
    # flag to the call so that the file is removed when done.
    for my $type (qw(shared exclusive)) {
        for my $access (qw(read write)) {
            # Releasing all the anon locks for ($type,$access) is pretty easy,
            # since it just needs a "count" argument to the call.
            if (@{$locks->{anonymous}{$type}{$access}}) {
                $self->lock_mgr(
                    action => 'release',
                    lock   => 'byte-range',
                    type   => $type,
                    access => $access,
                    count  => scalar(@{$locks->{anonymous}{$type}{$access}})
                );
            }
        }
    }

    # Releasing all the named locks is completely different, though. Since
    # every file will have a share-level lock whether it has any byte-range
    # locks or not, and since closing that file will remove all locks on the
    # file handle, we only have to unlock the share-level lock and all
    # byte-range locks will go away.
    for my $name (keys %{$locks->{'share-level'}}) {
        my ($type, $access) = @{$locks->{'share-level'}->{$name}};
        $self->lock_mgr(
            action => 'release',
            file   => $name,
            lock   => 'share-level',
            type   => $type,
            access => $access
        );
    }

    # Terminates the RPC connection
    $clientobj->destroy;

    return 'shutdown successful';
}

###############################################################################
#
#   Sub Name:       _lock_mgr_initialize
#
#   Description:    Initialize one or more files with some simple content
#                   before they get locked by the lock_mgr().
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   The main element in %args is called "state", and represents
#                   the state of the lock_mgr. It is a hash reference with the
#                   following keys:
#
#                     SELF   => a ref to the NOTE::Client object
#                     IS_WIN => a boolean, whether this client is Windows
#                     MNTPT  => the local mount-point/device of the volume
#                     CLIENT => the RPC client object for remote execution
#                     LOCKS  => the current internal locks-table
#
#                   All the private _lock_mgr_* subroutines get this argument.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of parameters
#
#   Returns:        Success:    message string
#                   Failure:    dies
#
###############################################################################
sub _lock_mgr_initialize {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Pull some basic things out of %args and $args{state}:
    my $state     = $args{state};
    my $clientobj = $state->{CLIENT};
    my $MNTPT     = $state->{MNTPT}; # To make it easier to use in strings.
    my $IS_WIN    = $state->{IS_WIN};
    my $offset    = $args{offset} || 0;
    my $data      = $args{data}   || 1024;

    if (! ($args{file} && @{$args{file}})) {
        # Can't do anything without files to do it to
        return 'no files specified, no initialization done';
    }

    # If $data is a number, convert it to a string of that length
    if ($data =~ /^\d+$/) {
        $data = 'a' x $data;
    }

    # Set up %open_args with keys for cflags and oflags. Even if under
    # Windows, we won't need sflags for this. We also don't need to both
    # faking a call to _lock_mgr_file_open_args since this is a much simpler
    # situation.
    my %open_args;
    if ($IS_WIN) {
        $open_args{cflags} = 'OPEN_ALWAYS';
        $open_args{oflags} = 'GENERIC_READ + GENERIC_WRITE';
    } else {
        $open_args{cflags} = 'O_CREAT';
        $open_args{oflags} = 'O_RDWR';
    }

    for my $file (@{$args{file}}) {
        # Open the file
        $clientobj->open_file(file => "$MNTPT$file", %open_args);
        if ($clientobj->error) {
            confess '_lock_mgr_initialize: failed to open ' .
                "$MNTPT$file: " . $clientobj->errmsg;
        }
        # Write $data to the file
        $clientobj->write_file(
            file     => "$MNTPT$file",
            contents => $data,
            offset   => $offset
        );
        if ($clientobj->error) {
            confess '_lock_mgr_initialize: failed to write to ' .
                "$MNTPT$file: " . $clientobj->errmsg;
        }
        # Close the file
        $clientobj->close_file(file => "$MNTPT$file");
        if ($clientobj->error) {
            confess '_lock_mgr_initialize: failed to close ' .
                "$MNTPT$file: " . $clientobj->errmsg;
        }
    }

    return scalar(@{$args{file}}) . ' files initialized';
}

###############################################################################
#
#   Sub Name:       _lock_mgr_acquire
#
#   Description:    Perform the acquisition of locks for the lock_mgr.
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   The main element in %args is called "state", and represents
#                   the state of the lock_mgr. It is a hash reference with the
#                   following keys:
#
#                     SELF   => a ref to the NOTE::Client object
#                     IS_WIN => a boolean, whether this client is Windows
#                     MNTPT  => the local mount-point/device of the volume
#                     CLIENT => the RPC client object for remote execution
#                     LOCKS  => the current internal locks-table
#
#                   All the private _lock_mgr_* subroutines get this argument.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of parameters
#
#   Returns:        Success:    message string
#                   Failure:    dies
#
###############################################################################
sub _lock_mgr_acquire {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Pull some basic things out of %args and $args{state}:
    my $state     = $args{state};
    my $clientobj = $state->{CLIENT};
    my $argsin    = $args{args};     # Redundancy, what redundancy?
    my $locks     = $state->{LOCKS};
    my $MNTPT     = $state->{MNTPT}; # To make it easier to use in strings.
    my $IS_WIN    = $state->{IS_WIN};
    my $is_anon   = $args{anonymous};

    # These three arguments from the original lock_mgr call make up the tuple
    # that defines the flags for opening and locking the file:
    my $lock      = $argsin->{lock};
    my $type      = $argsin->{type};
    my $access    = $argsin->{access};

    # Set up the arguments that will be passed to $clientobj->open_file
    # for each file we open. Note that this call will throw an exception for
    # the NFS cases where the combination of ($type, $access) is not
    # allowed. Also note that we are now passing in the reference to $self's
    # storage, as we need to check to see if it is a DP or LS volume when
    # we set flags, due to a NFSv4 issue.
    my %open_args = _lock_mgr_file_open_args(
        is_win => $IS_WIN,
        type   => $type,
        access => $access,
        volume => $state->{SELF}->{_storage}
    );
    # Set up the locking constant that will be used to create the lock,
    # based on $type and $IS_WIN. Note that this constant has no meaning
    # here, it is eval'd on the remote server. This won't actually be used if
    # we only create share-level locks.
    my $ltype;
    if ($IS_WIN) {
        $ltype = 'LOCKFILE_FAIL_IMMEDIATELY';
        if ($type eq 'exclusive') {
            $ltype .= '|LOCKFILE_EXCLUSIVE_LOCK';
        }
    } else {
        $ltype = ($type eq 'exclusive') ? 'F_WRLCK' : 'F_RDLCK';
    }

    for my $file (@{$argsin->{file}}) {
        # The remote server has to have the file opened first, before
        # it can get a byte-range lock. Opening the file itself creates the
        # share-level lock. Start by checking to see if we already have a
        # share-level lock for this file. This flag will be used to note when
        # a share-level lock is created solely to support a byte-range lock
        # request:
        my $forced_share_level_lock = 0;
        if ($locks->{'share-level'}->{$file}) {
            # If this is a request for a share-level lock, we have an error as
            # we cannot share-level-lock the same file more than once.
            if ($lock eq 'share-level') {
                confess "_lock_mgr_acquire: File $file is already " .
                    'share-level-locked, cannot lock again';
            }
        } else {
            # We don't have a lock yet. Whether this request is for a byte-range
            # lock or a share-level lock, we have to have the file open.
            $clientobj->open_file(file => "$MNTPT$file", %open_args);
            if ($clientobj->error) {
                confess '_lock_mgr_acquire: failed to open ' .
                    "$MNTPT$file: " . $clientobj->errmsg;
            }

            # Note it in the share-level locks table:
            $locks->{'share-level'}->{$file} = [ $type, $access ];

            # If this is in fact a byte-range lock request, note that this
            # share-level lock was only created for the sake of the byte-range
            # lock. It will mean that the share-level lock is quietly removed
            # along with the byte-range lock if/when the user releases the
            # byte-range lock.
            if ($lock eq 'byte-range') {
                $forced_share_level_lock = 1;
            }
        }

        # Now that we know for certain the file is open on the remote end, we
        # can byte-range-lock it if this is a byte-range-lock request:
        if ($lock eq 'byte-range') {
            $clientobj->lock_file(
                file   => "$MNTPT$file",
                ltype  => $ltype,
                offset => $argsin->{offset},
                length => $argsin->{length}
            );
            if ($clientobj->error) {
                confess '_lock_mgr_acquire: failed to lock ' .
                    "$MNTPT$file: " . $clientobj->errmsg;
            }

            # Note it in the byte-range locks table:
            $locks->{'byte-range'}{"$file,$argsin->{offset},$argsin->{length}"}
                = [ $type, $access, $forced_share_level_lock ];
        }

        # If the lock was an anonymous lock, also record it in the apropos
        # part of the anon locks table, so that it can also be anonymously
        # released.
        if ($is_anon) {
            # Anon locks just keep a series of queues of filenames for each
            # (type, access) combination, since we don't care which one we
            # free when the release request is made.
            push(@{$locks->{anonymous}->{$type}->{$access}}, $file);
        }
    }

    return sprintf(
        "Created %d $lock lock%s of type $type-$access",
        scalar(@{$argsin->{file}}),
        (@{$argsin->{file}} == 1) ? q{} : 's'
    );
}

###############################################################################
#
#   Sub Name:       _lock_mgr_file_open_args
#
#   Description:    Calculate the arguments needed for the remote open of a
#                   file, based on whether the file-system is NFS or CIFS
#                   ($args{is_win}), and based on the tuple of
#                   @args{qw(type access)}.
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   This routine does not actually get the same "state"
#                   value that the other private routines get, as it does
#                   not need those values.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of arguments
#
#   Returns:        Success:    hash table of args for the open_file RPC call
#                   Failure:    confesses
#
###############################################################################
sub _lock_mgr_file_open_args {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Just some short-hand variables so I don't have to keep indexing the hash:
    my ($is_win, $type, $access, $volume) =
        @args{qw(is_win type access volume)};
    # This will be the return value:
    my %open_args = ();

    if ($is_win) {
        # Set up arguments for the Win32::API calls, using Win32API::File
        # constants. These have no meaning here (which is why they're done as
        # strings), but will be eval'd by the remote server.

        # OPEN_ALWAYS is the equivalent of O_CREAT, it causes an existing file
        # to be opened or a non-existing file to be initially created. But we
        # have the same issue with CIFS as with NFSv4, regarding using this
        # flag on a read-only volume.
        $open_args{cflags} =
            ($volume->{_mirror_type} && $volume->{_mirror_type} =~ /dp|ls/i) ?
                'OPEN_EXISTING' : 'OPEN_ALWAYS';

        if ($access eq 'read') {
            # For read-only access
            $open_args{oflags} = 'GENERIC_READ';
            if ($type eq 'shared') {
                $open_args{sflags} = 'FILE_SHARE_READ';
            }
        } else {
            # For read-write access
            $open_args{oflags} = 'GENERIC_READ + GENERIC_WRITE';
            if ($type eq 'shared') {
                $open_args{sflags} = 'FILE_SHARE_READ + FILE_SHARE_WRITE';
            }
        }
    } else {
        # Set up arguments for the UNIX/Linux/POSIX calls. These are basic
        # open(2) constants. Again, these are not evaluated here, they are
        # eval'd on the remote server. Which is why they are strings.

        # First, check for the combinations that are not permitted:
        if ((($type eq 'shared')    && ($access eq 'write')) ||
            (($type eq 'exclusive') && ($access eq 'read'))) {
            confess "Combination of '$type' and '$access' not supported in NFS";
        }

        # Start with O_CREAT to ensure the file is created if it doesn't exist.
        # We don't do this if the volume is a DP or LS mirror, because we
        # couldn't create files on them anyway, and NFSv4 fails with the flag
        # on a read-only volume. This is likely a bug in the NFSv4 stack, and
        # if it gets fixed/changed we can do away with this conditional and
        # just assign $open_args{cflags} as O_CREAT, as it was previously.
        #
        # Note that the way I'm detecting DP/LS status on $volume is a hack.
        # The volume class should provide this information via a method, but
        # it doesn't.
        $open_args{cflags} =
            ($volume->{_mirror_type} && $volume->{_mirror_type} =~ /dp|ls/i) ?
                q{} : 'O_CREAT';

        # Since we know that we have a valid pair of ($type, $access), this
        # test is basically just on $type:
        $open_args{oflags} = ($type eq 'exclusive') ? 'O_RDWR' : 'O_RDONLY';
    }

    return %open_args;
}

###############################################################################
#
#   Sub Name:       _lock_mgr_release
#
#   Description:    Release locks for the lock_mgr.
#
#                   Note: This is not a method, there is no $self in @_.
#
#                   The main element in %args is called "state", and represents
#                   the state of the lock_mgr. It is a hash reference with the
#                   following keys:
#
#                     SELF   => a ref to the NOTE::Client object
#                     IS_WIN => a boolean, whether this client is Windows
#                     MNTPT  => the local mount-point/device of the volume
#                     CLIENT => the RPC client object for remote execution
#                     LOCKS  => the current internal locks-table
#
#                   All the private _lock_mgr_* subroutines get this argument.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   %args     in      hash      Hash table of parameters
#
#   Returns:        Success:    message string
#                   Failure:    dies
#
###############################################################################
sub _lock_mgr_release {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # This part seems to be thankfully free of UNIX/Windows specifics.
    # It only cares whether we're dealing with anon or named locks.
    my $state     = $args{state};
    my $clientobj = $state->{CLIENT};
    my $argsin    = $args{args}; # Yes, my naming is very creative.
    my $locks     = $state->{LOCKS};
    my $MNTPT     = $state->{MNTPT}; # To make it easier to use in strings
    my $is_anon   = $args{anonymous};

    # These three arguments from the original lock_mgr call make up the tuple
    # that identifies what lock(s) are being released.
    my $lock      = $argsin->{lock};
    my $type      = $argsin->{type};
    my $access    = $argsin->{access};

    my $retval;

    if ($is_anon) {
        for my $index (1 .. $argsin->{count}) {
            # Unlock temp files in the same order they were locked. Pull them
            # off the head of the queue. If we run out of files before $index
            # reaches the count, that's an exception.
            my $file = shift(@{$locks->{anonymous}->{$type}->{$access}});
            if (! $file) {
                confess '_lock_mgr_release: Not enough anon locks of type ' .
                    "'$type-$access', ran out at $index";
            }

            # We need to close then delete. Closing the file will release the
            # lock(s) on it, so we don't need the extra RPC call to explicitly
            # unlock it.
            $clientobj->close_file(file => "$MNTPT$file");
            if ($clientobj->error) {
                confess '_lock_mgr_release: failed to close anon file ' .
                    "$MNTPT$file: " . $clientobj->errmsg;
            }
            $clientobj->delete_file(file => "$MNTPT$file");
            if ($clientobj->error) {
                confess '_lock_mgr_release: failed to delete anon file ' .
                    "$MNTPT$file: " . $clientobj->errmsg;
            }

            # Clear out this file entry from the byte-range and share-level
            # locks tables:
            delete $locks->{'byte-range'}->{"$file,0,0"};
            delete $locks->{'share-level'}->{$file};
        }

        $retval = sprintf(
            "Released %d lock%s of type $type-$access",
            $argsin->{count},
            ($argsin->{count} == 1) ? q{} : 's'
        );
    } else {
        for my $file (@{$argsin->{file}}) {
            # The behavior here depends on whether the lock is byte-range or
            # share-level. A byte-range lock is merely released. A share-level
            # lock is released by closing the file, which also releases any and
            # all byte-range locks on the same file.

            if ($lock eq 'byte-range') {
                # First make sure we have a lock of this type, for this
                # file, at the offset/length pair given. Then test that the
                # tuple saves for that lock matches our values for $type and
                # $access.
                my $key = "$file,$argsin->{offset},$argsin->{length}";
                my $tuple = $locks->{'byte-range'}->{$key};
                if (! $tuple) {
                    confess '_lock_mgr_release: No existing byte-range lock ' .
                        "found for $file with given offset/length";
                }
                if (($tuple->[0] ne $type) || ($tuple->[1] ne $access)) {
                    confess '_lock_mgr_release: Lock for given offset/length ' .
                        "does not match the type $type-$access";
                }
                # If this flag is true, the share-level lock associated with
                # this file was only created to prop up the byte-range lock.
                # We'll want to quietly release it, as well, if there are no
                # other byte-range locks still on this file.
                my $forced_share_level_lock = $tuple->[2];

                # Everything matches, we're OK to release this lock. Start by
                # deleting the entry in the hash table.
                delete $locks->{'byte-range'}->{$key};
                # Here we just unlock, we don't close the file.
                $clientobj->unlock_file(
                    file   => "$MNTPT$file",
                    offset => $argsin->{offset},
                    length => $argsin->{length}
                );
                if ($clientobj->error) {
                    confess '_lock_mgr_release: failed to unlock ' .
                        "$MNTPT$file: " . $clientobj->errmsg;
                }

                # Now release the share-level lock if it was only created for
                # the sake of propping up the byte-range lock. But don't release
                # it if there are other byte-range locks still, as that would
                # immediately nuke them, as well.
                if ($forced_share_level_lock) {
                    my @existing =
                        grep { /^$file,/ } (keys %{$locks->{'byte-range'}});
                    if (! @existing) {
                        $clientobj->close_file(file => "$MNTPT$file");
                        if ($clientobj->error) {
                            confess '_lock_mgr_release: failed to close ' .
                                "$MNTPT$file: " . $clientobj->errmsg;
                        }
                        delete $locks->{'share-level'}->{$file};
                    }
                }
            } else {
                # For share-level locks we just close the file. That deletes
                # the share-level lock and any byte-range locks. Then we delete
                # the hash table entry for the share-level lock and all table
                # entries for byte-range locks under this file name.
                $clientobj->close_file(file => "$MNTPT$file");
                if ($clientobj->error) {
                    confess '_lock_mgr_release: failed to close ' .
                        "$MNTPT$file: " . $clientobj->errmsg;
                }
                delete $locks->{'share-level'}->{$file};
                my @byte_range_keys =
                    grep { /^$file,/ } (keys %{$locks->{'byte-range'}});
                for my $key (@byte_range_keys) {
                    delete $locks->{'byte-range'}->{$key};
                }
            }
        }

        $retval = sprintf(
            "Released %d $lock lock%s of type $type-$access",
            scalar(@{$argsin->{file}}),
            (@{$argsin->{file}} == 1) ? q{} : 's'
        );
    }

    return $retval;
}

=head2 lock_mgr

The lock manager abstracts the creation and deletion of file locks between
Windows and UNIX/Linux clients. The call sends commands to the manager to
acquire or release locks, to fetch status, or to shut down the lock manager.

Parameters are passed as a hash table of key/value pairs, detailed below. Locks
can be obtained on specific files, or may be created on anonymous files simply
so that locks of the desired type are created and present. Locks on specific
files can specify the byte-range that is to be locked (if so desired). Locks on
anonymous files can only be created over the whole file. Locks created may be
byte-range or share-level, exclusive or shared locks, for read or write
access. When releasing a lock, the type and access have to match the
type/access given when the lock was created.

The return value of the call is a string message detailing what was done,
except for the C<status> call, in which case the return value is a hash
reference.

If an error occurs, an exception is thrown.

You may also want to read the
L<http://wikid.netapp.com/w/DPG_QA_Automation/Library/GNOTE_Client_Lock_Mgr>
wiki page, which has more information on the lock manager.

=over

=item Arguments

The arguments to the lock manager are a hash of parameter/value pairs. The
allowed parameters are:

=over

=item B<action> = string

The specific action being taken in this call to the manager. Must be one of
the following values (the default value is C<acquire>):

=over

=item I<acquire>

Acquire one or more locks. If the C<file> parameter is provided, the locks will
be "named" locks: locks on specific files. Otherwise they will be "anonymous"
locks: locks on anonymous files created solely for the duration of being
locked. Anonymous files are deleted as soon as the lock is released, where
named locks do I<not> remove the file after the lock is released.

=item I<release>

Release one or more locks. If the locks being released are anonymous locks,
then the underlying files are removed, as well. Otherwise, the files are
left in place.

When releasing locks for specific files ("named locks"), any values for
C<type>, C<access>, C<offset> and C<length> must match the values that were
given when the locks were created. Otherwise, the lock-manager might not be
able to distinguish between two locks on the same file (but covering different
regions).

=item I<initialize>

Initializes one or more files by creating the file on the volume associated
with this client, then writing some amount of data to it. This is only useful
for cases where you specifically do not want a zero-length file. There are
also other methods available within GNOTE for creating and populating files.
This is meant to be minimal in nature, and as such is not the most-efficient
implementation.

I<initialize> requires the C<file> parameter. Additionally, C<offset> may
be passed to tell where in the file to start writing data, and C<data> may
be passed to provide specific data (see the documentation of C<data>, below).

=item I<status>

Return a simple hash-reference that gives some small amount of detail on the
current state of the lock manager. The keys of the hash reference will be:

=over

=item share_level_shared

The count of shared share-level locks.

=item share_level_exclusive

The count of exclusive share-level locks.

=item byte_range_shared

The count of shared byte-range locks.

=item byte_range_exclusive

The count of exclusive byte-range locks.

=item share_level_total

The total count of share-level locks.

=item byte_range_total

The total count of byte-range locks.

=item shared_total

The total count of shared locks.

=item exclusive_total

The total count of exclusive locks.

=item total

The total of all locks of all types.

=back

All keys will be present, even if the value of the key is 0.

=item I<show>

This action is a front-end to the C<locks_show> method in B<NOTE::GX::Vserver>.
It reads just the locks for the volume this client object is associated with,
it does not read locks for the whole Vserver or for other volumes mounted by
the same client. The locks data returned by the Vserver is then sorted by
lock type (C<share-level>, C<byte-range>, C<op-lock> and C<delegation>), and
some other information is derived from the locks themselves.

The return value is a hash-reference with the following keys:

=over

=item share-level

=item byte-range

=item op-lock

=item delegation

These four keys are each array references that hold all the lock records of the
given type. Any of them may be empty if no locks of that type are currently in
place.

=item share-level-count

=item byte-range-count

=item op-lock-count

=item delegation-count

These four keys are the integer counts of the number of locks of the given
type. If there are no locks of a type, the value will be 0.

=item share-level-hash

=item byte-range-hash

=item op-lock-hash

=item delegation-hash

These four keys contain MD5 checksums of the lock data for each type. If there
are no locks of a type, the hash will be a null string. The hash is calculated
in a way that should ensure that the same set of locks always yields the same
checksum, even if their order in the list is changed.

=item all

=item all-count

=item all-hash

These three keys represent the cumulative data of all the locks on the
Vserver for this client's volume. The key C<all> will be an array reference
with all the lock records. The key C<all-count> will be the count of all
locks. The C<all-hash> will be the MD5 checksum of all the lock data. If there
are no locks on the Vserver for this volume, then C<all> will be an empty
array, C<all-count> will be 0 and C<all-hash> will be a null string.

=back

Note that the data returned by C<show> may not match exactly the data returned
by C<status>. The results from this action will include all locks on the
volume, including those created by other clients mounting the same
volume. Also, if the volume is mounted as NFSv3 the lock manager keeps some
internal counters on share-level locks that are there just for internal
consistency and don't actually exist in the Vserver's locks table.

=item I<shutdown>

This action shuts down the lock manager. All existing locks are closed, and
the remote-connection to the client is terminated. Any future calls to the
lock manager will create new connections. This should rarely be needed by
the end-user, and is generally only called during the NOTE/GNOTE teardown
phase, where it is called automatically.

=back

=item B<lock> = string

The kind of lock to create, one of C<byte-range> or C<share-level>. The
default is C<share-level>. Note that share-level locks are not provided under
NFS version 3, so requests for a share-level lock on those file systems will
not create any entries in the Vserver's locks table. Under NFSv3, you will want
to make sure you explicitly request C<byte-range> locks.

=item B<type> = string

The type of lock to create. The allowed values are either C<shared> or
C<exclusive>. For sake of backwards-compatibility, you may specify C<read>
to mean C<shared>, or specify C<write> to mean C<exclusive>. When more than
one file is being locked, the same type is applied to all locks created. The
default is C<shared>.

=item B<access> = string

The type of file access for the lock, one of C<read> for read-only or C<write>
for read-write access. The default is C<read>. Note that if you create a
share-level lock with C<read> access (meaning read-only), you will not be able
to create any byte-range locks with C<write> access on the same file
afterwards, due to limitations in the RPC layer used to interface with the
client host.

=item B<file> = string or array-reference

The file name (or array-reference of multiple file names) to use in creating
locks. If this parameter is set, the B<count> parameter (detailed below) is
ignored. Each name in the list will be locked with a lock of type B<type>,
using the same offset and length as given in the next two parameters. If you
need to lock multiple files with different types or different byte-ranges,
you will need to make multiple calls to B<lock_mgr>.

Any file that exists already must be at least readable by the user trying to
acquire the lock, as the process will first attempt to open it before the lock
is applied. If the lock type is C<exclusive>, it will also need to have write
permission.

=item B<offset> = integer

If the action is I<acquire>, this represents the offset (from 0) within the
file being locked, where the lock will begin. In this case, the option is
ignored for anonymous locks (it is ignored unless C<file> is also set).

If the action is I<initialize>, then this is used to determine where within the
file the content specified by C<data> is written.

Defaults to 0.

=item B<length> = integer

The length of the region (in bytes) to lock, within the file. This is measured
from the point defined by B<offset>, above. This option is ignored unless
B<file> is set. Defaults to 0, which is interpreted as locking to the maximum
size of the file.

=item B<count> = integer

The number of anonymous locks of type B<type> to create. This option is
ignored if B<file> is set (since setting B<file> means that named-file locks
are being created, not anonymous locks). Defaults to 1.

=item B<data> = integer or string

For the I<initialize> action, this specifies the data to be written to the
file or files being initialized. The value of this parameter should be either
a string or an integer number. If the value is a string, that data is written
to all the files initialized. If the value is a number, that many bytes are
written to the file, with the content of the bytes being arbitrary.

Defaults to 1024, which means 1024 bytes of arbitrary data are written to each
file.

=back

=item Examples

    # Acquire 2 byte-range locks on anonymous files,
    # exclusive with write access
    $result = $client->lock_mgr(
        action => 'acquire',
        lock   => 'byte-range',
        type   => 'exclusive',
        access => 'write',
        count  => 2
    );

    # Acquire 1 share-level lock on an anonymous file,
    # shared with read access
    $result = $client->lock_mgr(
        action => 'acquire',
        lock   => 'share-level',
        type   => 'shared',
        access => 'read'
    );

    # Acquire a lock on a specific part of a specific file
    $result = $client->lock_mgr(
        action => 'acquire',
        lock   => 'byte-range',
        type   => 'exclusive',
        access => 'write',
        file   => 'large_file.dat',
        offset => 0,
        length => 1024
    );

    # Release the two anonymous write-locks
    $result = $client->lock_mgr(
        action => 'release',
        type   => 'exclusive',
        count  => 2
    );

    # Release the lock on a specific file. Note that the offset,
    # length and other parameters have to match what was passed when
    # the lock was created:
    $result = $client->lock_mgr(
        action => 'release',
        lock   => 'byte-range',
        type   => 'exclusive',
        access => 'write',
        file   => 'large_file.dat',
        offset => 0,
        length => 1024
    );

    # Create multiple locks on specific files, defaulting offset
    # to zero and length to the whole file. Uses the defaults for
    # lock, type and access:
    $result = $client->lock_mgr(
        action => 'acquire',
        file   => [ 'shared_file_1', 'shared_file_2' ],
    );

=back

=head3 Notes on Locks

The lock manager supports the creation of different types of locks, but there
are also locks that are created by ONTAP on the filer itself, that the lock
manager does not have control over. The locks that the lock manager does create
are sometimes referred to as I<hard> locks, where the ones created by ONTAP
are sometimes referred to as I<soft> locks.

Soft locks are created (both in NFSv4 and CIFS) by merely opening a file. Soft
locks are not supported in NFSv3. In NFSv4 the soft locks are represented by
"delegations", and in CIFS they are represented by "opportunistic locks".

In order to create a byte-range lock, the file must first be opened. Thus, on
those file systems which do provide/support share-level locks, the act of
creating a byte-range lock will also create a share-level lock if one does not
already exist for that file. If you explicitly created a share-level lock
before the byte-range lock, you will not get an extra one, the existing one is
all you will have. This can be confusing if you create the byte-range locks
directly and are not expecting the "bonus" of an extra (share-level) lock for
each byte-range lock you created.

If you create a share-level lock with read-only access (C<access> set to
C<read>), you will not be able to create byte-range locks with C<write>
access. This is due to a limitation in the RPC layer use to handle the file
operations, in which a given file can only have one open file-handle on it at a
time. If the file-handle is open for read-only access, you cannot create
write-access locks on it, and you can't get a second file-handle to use for the
write-access locks.

If you create a byte-range lock directly without first creating a share-level
lock on the file, the lock is flagged so that removing the byte-range lock also
removes the share-level lock. This saves you the trouble of having to remove
two locks for every one lock you create. However, if you created additional
byte-range locks on the same file, then the share-level lock will be left in
place. Otherwise, removing the share-level lock would destroy all the existing
byte-range locks.

=head3 Notes on NFS

Requesting a share-level lock on a NFSv3 file system will open the file for
future lock calls, but will not produce a record in the vserver locks
table. The lock manager will still think it created a share-level lock, and it
will be reflected in the structure returned by the C<status> command.

The type/access combinations available to NFS are more limited than those
available to CIFS. When the volume you are creating locks on is mounted via
NFS, you are limited to only the following combinations: (C<type>=shared,
C<access>=read) and (C<type>=exclusive, C<access>=write). That is, you can only
create shared locks with read-only access, or exclusive locks with read-write
access. If you try to create an exclusive/read or shared/write lock, an
exception will be thrown.

NFSv4 file systems also support delegations, if enabled on the filer. If
delegations are enabled on your NFSv4 source, you will get additional records
in the C<vserver locks show> output, for each delegation. Coupled with the
distinction between share-level and byte-range locks, you could get as many as
three entries in the vserver locks table from just one file being locked by the
lock manager.

=head3 Notes on CIFS

CIFS supports both share-level and byte-range locks. CIFS also creates a
third type of lock, called an I<opportunistic lock> (or C<op-lock>). Like the
delegations of NFSv4, these are created automatically by the server, and
cannot be explicitly created by a user. They can be disabled by turning them
off in the CIFS configuration for the server.

If a file has an op-lock of type C<batch>, then byte-range locks created on
that same file-handle will I<not> show up on the server (they will not be
visible in the output of C<vserver locks show>). This is because the batch
op-lock allows the client to cache the byte-range locks. If the op-lock is
broken, the byte-range locks should then appear on the server.

=head3 Notes on Anonymous vs. Named Locks

Anonymous locks (calls to B<lock_mgr> without the C<file> option) I<always>
create files, then lock them. Because of this, you cannot create anonymous
locks on read-only file-systems such as DP or LS SnapMirror volumes. Also, you
cannot retrieve the names used for anonymous lock files. This means that you
cannot create anonymous locks on a SnapMirror source volume, then try to lock
those same files on the destination volume. You will not know the file names,
and because of this you will not be able to specify the existing files to lock.

In order to create locks on read-only volumes, you must have the name of a file
that already exists on the volume. There are many ways to create files on a
mounted read-write (source) volume, this will not be covered here. Once the
source volume is mirrored to the read-only destination volume you can then
create a lock on the file, since you know the name.

One easy way to create a file on the source volume is to lock it by means of a
client object's B<lock_mgr>. This creates the file if it doesn't exist. You can
then do a SnapMirror update, and lock the same file name on the destination
volume. Take the following code as an example:

    # Assume that $client_src is a NOTE::Client object with the
    # source vol mounted, and $client_dst is a client object with
    # the destination vol mounted. $snapm is a NOTE::SnapMirror
    # object that represents the SnapMirror relationship between
    # the two.
    $client_src->lock_mgr(
        action => 'acquire',
        type   => 'exclusive',
        file   => 'locked_file'
    );
    # This update should propagate "locked_file" to dst
    $snapm->update();
    # Now this should succeed, as we are asking for a read-only
    # lock on a file that already exists on the volume:
    $client_dst->lock_mgr(
        action => 'acquire',
        type   => 'shared',
        file   => 'locked_file'
    );

If you want to create multiple locks, this is easy to do by passing the C<file>
parameter an array-reference instead of a single value. For example, if you
want to create 10 locks, the above code can be modified to this:

    # Assume the same variables.
    my @names = map { sprintf('locked_file_%02d', $_) } (1 .. 10);
    $client_src->lock_mgr(
        action => 'acquire',
        type   => 'exclusive',
        file   => \@names
    );
    # This update will propagate all the files to dst
    $snapm->update();
    # Now this should succeed, as we are asking for read-only
    # locks on the files that already exist on the volume:
    $client_dst->lock_mgr(
        action => 'acquire',
        type   => 'shared',
        file   => \@names
    );

See also the
L<http://wikid.netapp.com/w/DPG_QA_Automation/Library/GNOTE_Client_Lock_Mgr>
web page, which has more information on how GNOTE client locking works.

=cut

###############################################################################
#
#   Sub Name:       lock_mgr
#
#   Description:    Provide a generic interface to the user for the mass
#                   creation and deletion of file-level locks.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %args     in      hash      Args that are passed on to
#                                                 the underlying manager. These
#                                                 are passed along unchanged.
#
#   Returns:        Success:    return value from internal manager
#                   Failure:    dies
#
###############################################################################
sub lock_mgr {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Create the underlying manager subroutine if it doesn't already exist.
    # It will be a closure that traps a number of stateful lexical values,
    # including tables of the existing locks, the ClientRPC object, etc.
    $self->_create_lock_mgr() if (! $self->{_lock_mgr});

    my $response = $self->{_lock_mgr}->(%args);
    if ($args{debug}) {
        NOTE::logcomment(
            'Reply from lock_mgr: ' .
            (ref $response ? Dumper($response) : $response)
        );
    }

    # If this was a shutdown command, finish it by deleting the lock manager
    # closure, which should release its resources and close the RPC
    # connection.
    if ($args{action} eq 'shutdown') {
        delete $self->{_lock_mgr};
    }

    return $response;
}

=head2 ::lun_path

Returns the LUN path on the filer given a LUN name.

=over

=item Argument

B<lun_name> = <string>

Lun name

=item Example

     my $filer_path = $client->lun_path("a_lun");

=back

=cut

sub lun_path {
    my ( $self, $lun_name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->{_storage}->mount_path . "/" . $lun_name;
}

=head2 ::create_lun

Creates a LUN on the Filer. The LUN is create on the root directory of the volume or Qtree.
Wrapper for
L<C<blocks_filer::filer_lun_create>|engineering::ontap_tests::agnostic::docs::lib-blocks_filer-pm/filer_lun_create>.
Croaks if error.

=over

=item Argument

B<name> => I<string>

Required. Lun name

B<size> => I<lun size>

Optional. Lun size. Default "32M' for a windows LUN and "4m" otherwise

B<ostype> => I<lun type>

Optional. OS type of lun. Default linux

=item Example

    $client->create_lun("a_lun", "32m", "windows");

=back

=cut

sub create_lun {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing name" unless defined $args{name};
    $args{ostype} = $args{ostype} || "linux";
    $args{size} = $args{size} || ( $args{ostype} =~ /windows/i ? "32m" : "4m" );
    eval { blocks_filer::filer_lun_create( $self->{_storage}->Filer->conn, $self->lun_path( $args{name} ), $args{size}, $args{ostype} ); };
    confess "Lun create failed: $@" if $@;
    NOTE::logcomment( "Lun created. Parameter: " . Dumper(%args) );
}

=head2 ::resize_lun

Resizes the specified LUN on the filer. Wrapper for
L<C<blocks_filer::filer_lun_size_change>|engineering::ontap_tests::agnostic::docs::lib-blocks_filer-pm/filer_lun_size_change>.
Croaks if failed.

=over

=item Arguments

B<name> => I<string>

Required. Name of Lun

B<size> => I<value>

Required. New size of the LUN to change to

=item Examples

      $client->resize_lun("a_lun", "8m");

=back

=cut

sub resize_lun {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing name" unless defined $args{name};
    confess "Missing size" unless defined $args{size};

    eval { blocks_filer::filer_lun_size_change( $self->{_storage}->Filer->conn, $self->lun_path( $args{name} ), $args{size} ); };
    confess "Lun resize failed: $@" if $@;
    NOTE::logcomment("Lun set to $args{size}");
}

=head2 ::modify_lun

Given a LUN, with a start block number and number of blocks, will modify specific location on the LUN

=over

=item Arguments

B<name> => I<string>

Name of the LUN to modify

B<start> => I<integer>

Optional. Block to start modification. Default 0

B<count> => I<integer>

Optional. Number of blocks to modify. Default 1

=item Examples

    $client->modify_lun(name=>"a_lun"); # Modify block 0

    $client=>modify_lun(name=>"a_lun", start=>20, count=> 5); # modify 5 blocks starting from block 20

=back

=cut

sub modify_lun {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{start} = 0 unless defined $args{start};
    $args{count} = 1 unless $args{count};

    confess "modify_lun not supported on Windows client" if $self->{_cifs};
    confess "Missing name"       unless defined $args{name};
    confess "start must be >= 0" unless $args{start} >= 0;
    confess "count must be +ve"  unless $args{count} > 0;

    my $blk_size = $self->{_storage}->Filer->block_size;

    my $rsh_prefix   = "rsh " . $self->{_args}{client} . " ";
    my $cln_lun_path = $self->{_storage}->mountpoint . "/" . $args{name};

    # Create a file of 'count' blocks in size to modify LUN with
    my $in_file = $self->{_storage}->mountpoint . "/data_for_lun_" . $args{name};

    my $script = qq { \{
    			my \$string = 1 x ($blk_size*$args{count});
			open F, ">", '$in_file' || die "open failed for input file to LUN";
			syswrite(F, \$string) || die "syswrite failed while creating input file to LUN";
			close F;
			\} };
    Tharn::subtest( $script, -client, $self->{_args}{client} );
    NOTE::logcomment("Modify LUN $args{name} from block $args{start} for $args{count} block(s)");
    eval {

        # Enable read / write access to LUN
        blocks_filer::filer_lun_share_all( $self->{_storage}->Filer->conn, $self->lun_path( $args{name} ) );

        # Use 'dd' to modify LUN with it
        my $cmd = "/usr/local/bin/sudo dd conv=notrunc " . "seek=$args{start} count=$args{count} bs=$blk_size " . "if=$in_file of=$cln_lun_path";
        NOTE::logcomment("LUN modify cmd: $cmd");
        system( $rsh_prefix. $cmd );
        confess "dd failed, rc = $?" if $?;
    };
    system( $rsh_prefix. "rm $in_file" );
    confess "LUN modify failed: $@" if $@;

    return 1;
}

=head2 ::start_hammer

Start hammer run from the client on the storage. reture the subtest obj if it's nfs hammer or connection if it's cifs hammer.

CLI parameter "UTIL_BIN" is used to locate "hammer". Default it's use the standard location. (nfs: /usr/software/test/bin  cifs: X:\bin

=over

=item Arguments

B<fill> => I<1|0>

if it's defined will fill disk and then keep it full.
Optional. Default is 0.

B<blocksize> => I<integer>

Optional. Use the blocksize as NBYTES.

B<minbsize> => I<integer>

Optional. Set the minimum block size to the given NBYTES.

B<maxbsize> => I<integer>

Optional. Set the maximum block size to the given NBYTES.

B<minfsize> => I<integer>

Optional. Set the minimum file size to the given NBYTES.

B<maxfsize> => I<integer>

Optional. Set the maximum file size to the given NBYTES.

=item Examples

    $client->start_hammer();

=back

=cut

sub start_hammer {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess("Hammer is already running")
      if exists $self->{_hammer};

    $Trace->entry( msg => "start_hammer" );

    my $hammer_cmd = " -nofilercore -runtime -1 -bg ";
    $hammer_cmd = $hammer_cmd . "-fill "
      if $args{fill};
    $hammer_cmd = $hammer_cmd . "-blocksize $args{blocksize} "
      if $args{blocksize};
    $hammer_cmd = $hammer_cmd . "-minbsize $args{minbsize} "
      if $args{minbsize};
    $hammer_cmd = $hammer_cmd . "-maxbsize $args{maxbsize} "
      if $args{maxbsize};
    $hammer_cmd = $hammer_cmd . "-minfsize $args{minfsize} "
      if $args{minfsize};
    $hammer_cmd = $hammer_cmd . "-maxfsize $args{maxfsize} "
      if $args{maxfsize};

    my $mnt_point = $self->mountpoint();

    my ( $hammer_bin, $hammer_path );
    my $id = NOTE::newName();

    my $bin = NOTE::option(q/$UTIL_BIN/)
      || ( $self->{_cifs} ? "X:/bin" : "/usr/software/test/bin" );

    if ( $self->{_cifs} ) {    # cifs hammer cmd here
        $hammer_bin  = $bin . "/hammer.exe ";
        $hammer_path = "$mnt_point\\hammer_cifs_$id";
        $hammer_cmd  = $hammer_bin . $hammer_path . $hammer_cmd;

        my $fh = Tharn::host( $self->{_args}{client} ) or die $@;
        my $fc = Tharn::connect("#rsh") or die $@;
        say( $fc, "mkdir $hammer_path" );
        confess "mkdir failed: $@"
          if $@;
        say( $fc, $hammer_cmd, -noprompt, -nocr, -blind );
        confess "Start hammer failed: $@"
          if $@;
        $self->{_hammer} = $fc;
        return $fc;

    }
    else {    # nfs hammer cmd here
        $hammer_bin  = $bin . "/hammer ";
        $hammer_path = "$mnt_point/hammer_nfs_$id";

        mkdir( $hammer_path, 0777 ) || die "mkdir failed: $!";
        $hammer_cmd = $hammer_bin . $hammer_path . $hammer_cmd;

        my $hammer_obj;

        my $hammer_id = "hammer" . $id;

        #NOTE::logcomment("HAMMER ID IS $hammer_id");
        eval {
            $hammer_obj = subtest(
                $hammer_cmd,
                -bg,
                -client => $self->{_args}{client},
                -runid  => $hammer_id,
                -onexit => sub { }
            );
        };

        confess "Could not start hammer: $@"
          if $@;

        $self->{_hammer} = $hammer_obj;
        return $hammer_obj;
    }

}

=head2 ::stop_hammer

Stop hammer run

=over

=item Examples

    $client->stop_hammer();

=back

=cut

sub stop_hammer {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $Trace->entry( msg => "stop_hammer" );

    NOTE::logcomment("HAMMER OBJ IS $self->{_hammer}");
    if ( $self->{_cifs} ) {    # cifs hammer cmd here
        $self->{_hammer}->send_intr();
    }
    else {
        $self->{_hammer}->kill();
    }

    delete $self->{_hammer};
}

sub SPARE {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_spare_me} = 1;
}

sub destroy {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_destroy} = 1;

    if ( $self->{_hammer} ) {
        $self->stop_hammer();
    }

    $self->{_ipc}->ipc()->quit() if defined $self->{_ipc};

    # Shutdown the lock manager if it exists:
    if (exists $self->{_lock_mgr}) {
        $self->lock_mgr(action => 'shutdown');
    }

    if ( defined $self->{_cifs} ) {
        if ( $self->_gx_storage() ) {
            $self->{_storage}->delete_share();
        }
        else {
            $self->{_cifs}->teardown();
        }
    }
}

sub DESTROY {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return
      if $$ != $self->{_owner}
          or $self->{_spare_me}
          or $self->{_destroy}
          or NOTE->ALL_SPARED;

    $self->destroy;
}

1;
