# -*- perl -*-

# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Client/RPC.pm#1 $
##
## @summary NOTE Client RPC class
##
## @status private
##
## @pod    here
##
## @author cpeter@netapp.com, dl-dpg-auto@netapp.com
##

=head1 NAME

NOTE::Client::RPC

=head1 DESCRIPTION

Provides transparent RPC support for client operations. Starts RPC servers on Windows
or Unix and then nicely cleans them up when script exits or dies without leaving zombies
on remote clients.

This subclass is initialized by GNOTE::Client class. It's not recommended to use this
subclass directly.

=cut

package NOTE::Client::RPC;

use strict;
use Carp qw(confess);
use Tharn;
use GNOTE;


=head2 ::new

Checks if RPC is already registered on remote client, if not then
starts RPC service on remote client and stores handle in GNOTE
global stack which is later used to perform RPC calls to given client.


Remote code is one of following depending on client type:

 use GNOTE::Client::Windows;
 while(1) {
    NATE::Events::eventloop_once();
 }

or

 use GNOTE::Client::Unix;
 while(1) {
    NATE::Events::eventloop_once();
 }

=over

=item Arguments

B<client> - I<Hostname of client>. Host must have proper NATEHOST file.

I<Optional.> Default is 'localhost'

=back

Check NATE::STAF::Peer and nateipc for more information on NATE's RPC/IPC
functionality.

=cut

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $$self{client} = $args{client} ? $args{client} : 'localhost';

# commented out per BURT 392617, if anonymous object gets destroyed, then reference to RPC gets lost, so we can not save it
#    if( defined $NOTE::_RPC_CLIENTS->{ $$self{client} } ) {
#	NOTE::logcomment("RPC server is already running");
#	return $NOTE::_RPC_CLIENTS->{ $$self{client} };
#    }

    my $childcode = undef;
    $$self{_lib} = 'GNOTE::Client::Windows' if ( host( $$self{client} )->hosttype =~ /windows/i );  # this piece of code will get executed
    $$self{_lib} = 'GNOTE::Client::Unix' if ( host( $$self{client} )->hosttype =~ /unix/i );	    # on remote, so call it as "GNOTE::
    unless( $$self{_lib} ) {
	confess "Unable to create RCP client for host ".
	         $$self{client} .", unknown host type ".
		 "or missing tharnhost file";
    }

    $$self{_child_code} = "{use $$self{_lib};while(1){NATE::Events::eventloop_once()}}";
    NOTE::logcomment("STAF/RPC config:\nClient: ".$$self{client}."\nCode:".$$self{_child_code});

    $$self{_handle} = subtest($$self{_child_code}, '-client' => $$self{client}, '-bg');
    my $remote_environment = "PerlLib:\n\t". $$self{_handle}->perl_eval(code=>'join("\n\t",@INC)');
    $remote_environment .= "\n\nUser Environment:\n" . $$self{_handle}->perl_eval(code=>'`set`');
    NOTE::logcomment("Remote environment:\n$remote_environment");

    bless $self, $class;

# commented out per BURT 392617, if anonymous object gets destroyed, then reference to RPC gets lost, so we can not save it
#    $NOTE::_RPC_CLIENTS->{ $$self{client} } = $self;

    NOTE::reference($self);

    return $self;
}

=head2 ::DESTROY

Destroys remote RPC service instance. This method is called by NOTE and/or Perl automatically on exit.

=cut

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logomment("Waiting for ".$$self{_tag}." to exit") if( $$self{_tag} );
    $$self{_handle}->quit();
    $$self{_handle}->wait_finish();
    # NOTE::unreference($self);		# this is done by NOTE itself...
};

1;
