# -*- perl -*-

# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Client/Windows.pm#1 $
##
## @summary Set of Windows methods executed via RPC calls
##
## @status private
##
## @pod    here
##
## @author cpeter@netapp.com, dl-dpg-auto@netapp.com
##

=head1 NAME

NOTE::Client::RPC::Windwows

=head1 DESCRIPTION

Provides set of Windows specific functions invoked via RPC by NOTE::Client class.

=cut

package GNOTE::Client::Windows;

use strict;
use Tharn qw($Staf);
use Carp qw(confess);
use Data::Dumper;
use FileManip;
use Win32::FileSecurity qw( Get Set MakeMask EnumerateRights );                 # This library is not available on UNIX platform.

sub _rpchandle_info { return $Staf->handle_query( handle => $Staf->handle()); }
sub _namespace_info { return Dumper(\%::); }

=head2 ::write_file(%args)

Calls FileManip::_write_file(%args) as is with a difference that
actual call is done from UNIX client from GNOTE::Client class, as

 $client = NOTE::Client->new( client=>'some_windows_xp' );
 $client->write_file( ... parameters for file_manip ... );

See NOTE::Client::write_file and FileManip::write_file for more info and examples

=cut

sub write_file {
    my %args = @_;
    $args{client} = Tharn::host($args{client});	    # file manip uses it for OS type detection
    FileManip::_write_file(%args) || confess($@);   # call low level library directly
};

=head2 ::read_file(%args)

Calls FileManip::_read_file(%args) as is with a difference that
actual call is done from UNIX client from GNOTE::Client class, as

 $client = NOTE::Client->new( client=>'some_windows_xp' );
 $client->read_file( ... parameters for file_manip ... );

See NOTE::Client::write_file and FileManip::write_file for more info and examples

=cut

sub read_file {
    my %args = @_;
    $args{client} = Tharn::host($args{client});	    # file manip uses it for OS type detection
    FileManip::_read_file(%args) || confess($@);   # call low level library directly
}

=head2 ::Get_ACL_Mask($path)

Calls Win32::FileSecurity::Get function for given path and returns results
back as hash of arrays translated from ACL by Win32::FileSecurity::EnumerateRights
as

 { user1 => [ permissions,..], user2 => [ perm...] } 

Function is normally invoked by NOTE::Client->get_acl, see NOTE::Client::get_acl
for more details and examples

=over

=item Arguments

B<path> - I<path to file> for which ACL info will be returned.

=back

=cut

sub Get_ACL_Mask {
    my $path = shift;
    my $return = undef;

    $path =~ s/\//\\/g;

    confess "Error: path \"$path\" not found" unless( -e $path );

    my %full_acl = ();
    if( Get( $path, \%full_acl ) ) {
	foreach my $user ( keys %full_acl ) {
	    my @ace = ();
	    EnumerateRights( $full_acl{$user}, \@ace );
	    $full_acl{$user} = \@ace;
	}									
	return \%full_acl;
    }
    confess "Can not read ACL: $!";
}										

=head2 ::Set_ACL_Mask($path,$acl)

Calls Win32::FileSecurity::MakeMask and Win32::FileSecurity::Set to set ACL for
given path in responce to RPC request. Usually called by NOTE::Client->set_acl.

See NOTE::Client::set_acl for more details and examples.

=over

=item Arguments

B<path> - I<path to file> for which ACL info will be returned.

B<acl> - ACL hash reference in format { user1 => [ permissions,..], user2 => [ perm...] } 

Permissions are strings from following list:

  DELETE, READ_CONTROL, WRITE_DAC, WRITE_OWNER,
  SYNCHRONIZE, STANDARD_RIGHTS_REQUIRED, 
  STANDARD_RIGHTS_READ, STANDARD_RIGHTS_WRITE,
  STANDARD_RIGHTS_EXECUTE, STANDARD_RIGHTS_ALL,
  SPECIFIC_RIGHTS_ALL, ACCESS_SYSTEM_SECURITY, 
  MAXIMUM_ALLOWED, GENERIC_READ, GENERIC_WRITE,
  GENERIC_EXECUTE, GENERIC_ALL, F, FULL, R, READ,
  C, CHANGE

=back

=cut

sub Set_ACL_Mask {
    my ($path,$aces) = @_;
    $path =~ s/\//\\/g;

    confess "Error: path \"$path\" not found" unless( -e $path );

    my %set_acl = ();
    my $skip_zero = 0;
    foreach my $user ( keys %{$aces} ) {
	$set_acl{$user} = MakeMask( @{ $$aces{$user} } );
    }
    Win32::FileSecurity::Set($path,\%set_acl);
    return Get_ACL_Mask($path);
}

=head2 ::showenv

Returns remote environment. Output from "set" command;

=cut

sub showenv {
    return `set`;
}

1;
