# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/FileSystem/NFS.pm#1 $
## @summary  NOTE FileSystem NFS class
## @status   Public
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NOTE::FileSystem::NFS

=head2 DESCRIPTION

This module implements NFS operations on Volumes and Qtrees via a
simple interface where filesystem elements are represented by objects
and filesystem operations are implemented as methods. Filesystem
elements can be either named or anonymous. 

An object representing the root directory of a Qtree or Volume
can be obtained by calling the C<nfs> method:

   $qt_root  = $qt->nfs;

   $vol_root = $vol->nfs;

Directory objects can be used to create subdirectories and
files within themselves:

   # Create file called 'somefile' in root directory
   $file = $qt->nfs->create( "somefile" );

   # Create anonymous directory in the root directory
   $dir = $qt->nfs->mkdir;

   # Create anonymous file 3 levels deep
   $newfile = $qt_root->mkdir->mkdir->create;

Files are empty when created, but can be filled:

   # Write up to 1 gigabytes of data into file (or until filesystem is full)
   $file->fill( max => '1g' );

The C<::fill> method returns the file object, so this is possible:

   # Create a '100m' file in $dir
   $file = $dir->create->fill( max => '100m' );

Unix special files (a.k.a. pipes and devices) are supported with C<::mknod>:

   # make anonymous character device
   $dir->mknod( type => 'c', major => 10, minor => 2 );

   # make a fifo
   $dir->mknod( name => 'mypipe', type => 'p' );

Symbolic links can be made of any filesystem object using the C<::symlink> method:

   # Make anonymous symbolic link to file (in the same directory as file):
   $symlink = $file->symlink;

   # Make anonymous symbolic link to $file in $dir
   $symlink = $file->symlink( $dir->path );

The contents of a symlink can be obtained using C<::readlink>:

   print "symlink points to ", $symlink->readlink, "\n";

Hardlinks can be made of any non-directory filesystem object using C<::link>

   # Anonymous link to $file
   $samefile = $file->link;

   # Create an anonymous pipe and 3 anonymous links to it:
   ($mypipe = $root->mknod( type => 'p' )->link->link->link;

You can get an object representing any existing filesystem element
using C<::attach>:

   $file = $vol->nfs->attach( "/home/somedatafile" );

You can move filesystem elements around using C<::mv>:

   # Move a file into the directory containing its parent
   $file->mv( ".." );

   # Move directory beneath "/a/b/c"
   $dir->mv( "/a/b/c" );

Note that, as expected, moving a directory will change the paths
of all elements beneath it:

   $file = $qt->mkdir("/a/b/c")->create( "somefile" );
   print $file->path, "\n"; # prints /a/b/c/somefile

   $dir  = $qt->mkdir("/x/y/z");

   # Move directory 'c' beneath 'z':
   $file->parent->mv( $dir->path );

   print $file->path, "\n"; # prints /x/y/z/c/somefile

Deleting files can be accomplished using C<::unlink>.

   $file->unlink;

Deleting directories can be accomplished using C<::rmdir>

   $dir->rmdir;
   croak $dir->path." still exists!"
	if ( $dir->stat );

The status of any filesystem element can be obtained with C<::stat>:

   ($dev,$inode,$mode,$nlink,$uid,$gid,$rdev,$size,
    $atime,$mtime,$ctime,$blksize,$blocks)
    = $file->stat;

Note that C<::stat> actually does an C<lstat>, so a stat on a
symlink returns the status of the link, not what the link points to.

There are convenience methods to get the individual status elements:

    croak $file->path." is empty"
	if ( $file->size == 0 );

    croak $link->path." is not a hardlink to ".$file->path
	if ( $link->inode != $file->inode );

There are also methods to get the type of any filesystem element and
major and minor numbers of devices:

    $type = $element->type;

    $major = $device->major;
    $minor = $device->minor;

You can get the path (relative to the root of the volume or qtree)
of any filesystem element using the C<::path> method:

    $path = $file->path;

If you want the fullpath, including the mountpoint use C<::fullpath>:

    my $filepath = $file->fullpath;
    open( F, '<', $filepath )
	or die "Can't open $filepath: $!";

If you want just the name:

    $name = $file->name;

or just the containing directory name:

    $dir_path = $file->dirname;

You can get an object representing the parent of the filesystem
element with C<::parent>:

   $dir = $file->parent;

   # create a sibling
   $brother = $file->parent->create;

and you can affect the access and modification times with C<::touch>:

   $file->touch; # Change both access and modification times to 'now'

   # just change access time
   $file->touch( atime => time );

   # change both access and modification time to sometime in the past
   $file->touch( atime => (time - 1800), mtime => (time - 3600) );

And last, but not least, you can get the output of C<ls> on a
filesystem element using C<::ls>:

   # log an 'ls -l' of $file
   NOTE::logcomment( $file->ls( '-l' ) );


=head2 METHODS

=cut

use GNOTE;

package NOTE::FileSystem::NFS;
#use strict;

use bigint;
use constant SNAPLOCK_EPOCH => 1044057600; # 00:00:00 Jan 1 2003
use constant BIG_INFINITY   => 0x7fffffffffffffff;
use constant SMALL_INFINITY => 0x7fffffff;

use GNOTE::FileSystem;
use Carp;
use Config;
use snaplock;
use POSIX;
use Date::Calc;

our @ISA = qw(NOTE);


=head3 ::mount

Mounts container and returns NFS object.

=over

=item Syntax

$obj->NFS::mount($container,%args);

=item Arguments

B<container> => I<NOTE::Volume | NOTE::Qtree>

I<Required>
Container is NOTE::Volume or NOTE::Qtree

B<$args{chmod}> => I<number>

I<Optional> Mount point permission. Default is '0777'.
If the chmod param exists, but is set to undef, don't
do a chmod. This is useful for mounting read-only filesystems.

=back

=cut

sub mount {
    my ($class,$container,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fsid; 
    my $self = {};

    # BURT 506501:
    # If the volume is already mounted for this client (defaulting to
    # localhost), don't try mounting it a second time unless there are args
    # in %args (other than "client" itself).
    my $args_not_client = scalar(grep { $_ ne 'client' } (keys %args));
    if ((! $container->mounted(client => $args{client})) || $args_not_client) {
        if ((exists $args{'chmod'} && !defined $args{'chmod'}) ||
            ! $container->writeable) {
            # If the chmod param exists, but is set to undef,
            # or if the container isn't writeable, don't do a chmod.
            delete $args{'chmod'};
        } else {
            $args{'chmod'} ||= '0777';  # Otherwise, default to 'read/write'
            $args{'rw'} ||= 1;      # must have for c-mode
        }

        $container->mount( %args );
    }

    if ( $container->isa( "NOTE::Qtree" ) ) {
	$fsid = $container->Volume->fsid . ":" . $container->name;
    } else {
	$fsid = $container->fsid;
    }
    $self->{_model}  = NOTE::FileSystem->init( $fsid );
    $self->{_wdir}   = "/"; # working directory
    $self->{_container}  = $container;
    bless $self, $class;
    return $self;
}


=head3 ::lookup

Checks if object is mounted and then returns
C<NOTE::FileSystem::lookup> element if lstat is
successful on given path.

=over

=item Syntax

NOTE::FileSystem::NFS->lookup($path,@args);

=item Arguments

B<$path> => I<string>

I<State unknown>
Full path. Calls $obj->path if $path is not present, however
it's impossible to skip $path while leaving @args because
@args is an array, not a pointer to an array. In other words
call it with all arguments or completely without arguments.

B<@args> => I<array>

I<Optional>
Array is passed to NOTE::FileSystem->lookup as is. See NOTE::FileSystem->lookup for more information.

=back

=cut

sub lookup {
    my ($self,$path,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix path with working directory if necessary
    $path = $self->path . $path unless ( $path =~ /^\// );

    my $fullpath = $mountpoint.$path;
    return undef unless lstat( $fullpath );
    my $is_dir = -d _ ? 1 : 0;

    my $element = $self->{_model}->lookup(
	$path, @args
    );
    return undef unless $element;
    $element->is_dir( $is_dir );
    return $element;
}


=head3 ::attach

Attach to existing mount point and return results as NFS object.

=over

=item Syntax

NOTE::FileSystem::NFS->attach($path,@args);

=item Arguments

B<$path> => I<string>

I<State unknown>
Full path. Calls $obj->path if $path is not present, however
it's impossible to skip $path while leaving @args because
@args is an array, not a pointer to an array. In other words
call it with all arguments or completely without arguments.

B<@args> => I<array>

I<Optional>
Array is passed to NOTE::FileSystem->lookup as is.

=back

=cut

sub attach {
    my ($self,$path,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $element;

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }
    # Prefix path with working directory if necessary
    $path = $self->path. $path unless ( $path =~ /^\// );

    my $fullpath = $mountpoint.$path;
    return undef unless stat( $fullpath );

    my $elem = $self->lookup( $path );
    croak "Didn't find $path" unless( $elem );
    return $elem;
}


=head3 ::create

Creates file by given path.

=over

=item Syntax

$obj->NFS::create($path,@args);

=item Arguments

B<path> => I<string>

I<Optional>
Path to file within object.

=item Examples

 # Create file called 'somefile' in root directory
 $file = $qt->nfs->create( "somefile" );

=back

=cut

sub create {
    my ($self,$path,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fullpath = $self->fullpath( $path );

    # already exists?
    if ( stat( $fullpath ) ) {
        NOTE::logcomment( "Can't create: $fullpath exists" );
	return undef;
    }

    local *F;
    open( F, '>', $fullpath ) or croak "Can't create $fullpath: $!";
    close F;

    my $elem = $self->lookup( $path );
    croak "Didn't find $path" unless( $elem );
    return $elem;
}


=head3 ::utime

Changes the access and modification times. Returns the number of files successfully changed.

=over

=item Syntax

NOTE::FileSystem::NFS->utime($path, $atime, $mtime);

=item Arguments

B<path> => I<string>

I<Required>
Path to file or directory

B<atime> => I<number>

I<Required>
NUMERICAL access times

B<mtime> => I<number>

I<Required>
NUMERICAL modification times

=back

=cut

sub utime {
    my ($self,$path,$atime,$mtime) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fullpath = $self->fullpath( $path );

    return utime( $atime, $mtime, $fullpath );
}


=head3 ::chmod

Changes the permissions of a path.
Returns the number of files successfully changed.

=over

=item Syntax

NOTE::FileSystem::NFS->chmod($path,$mode);

=item Arguments

B<path> => I<string>

I<Required>
Path to file or directory

B<mode> => I<number>

I<Required>
Must be the numerical mode, which should probably be an octal number, and which
definitely should not be a string of octal digits: 0644 is okay, '0644' is not.

=back

=cut

sub chmod {
    my ($self,$path,$mode) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fullpath = $self->fullpath( $path );
    if( $^O =~ /win/i ) {			 # $^O on windows contains: MSWin, win32, cygwin.
	return chmod( $mode, $fullpath );
    } else {
	if( -x "/usr/software/bin/sudo" ) {
	    my $octal_mode = sprintf "%o", $mode;
	    my $out = `/usr/software/bin/sudo -u root chmod $octal_mode "$fullpath" 2>&1 && echo \$?`;	# capture error and exit code
	    confess $out if $out =~ /not permitted/i; # some linux flavors do not grant permissions even for root if changing on NFS share.
	    $out =~ s/\D//g; # extract return code if no error message.
	    return $out == 0 ? 1 : 0;		 # Perl returns 1 if success, shell returns 0 if success...
	} else {
	    return chmod( $mode, $fullpath );
	}
    }
}


=head3 ::mknod

Creates device file using UNIX mknod command

    # Unfortunately, there is no built-in support for mknod.
    # There is an add-on package available (Unix::mknod)
    # that has good support, but it isn't installed on any of
    # our machines. So, I just wing it here.

=over

=item Syntax

NOTE::FileSystem::NFS->mknod(%args);

=item Arguments

B<name> => I<string>

I<Optional>
Name for device file. Will be generated if not given.

B<type> => I<(c|u|b|p)>

I<Required>
Device type

 b      Create a block-type special file.
 c, u   Create a character (unbuffered) special file
 p      Create a FIFO (named pipe).

B<major> => I<number>

I<Required>
The major device number.

B<minor> => I<number>

I<Required>
The minor device number; can be either decimal or octal.
The assignment of major device numbers is specific to each system.

=back

=cut

sub mknod {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $opts;
    my $path;

    if ( $args{'name'} ) {
	$path = $args{'name'};
    } else {
	$path = $self->newName( "dv" );
    }
    my $fullpath = $self->fullpath( $path );

    # already exists?
    return undef if stat( $fullpath );
    croak "device type needs type"
	unless( exists $args{type} );
    if ( $args{type} eq 'p' ) { # fifo
	$opts = $args{type};
    } elsif ( $args{type} =~ /^(c|u|b)$/ ) {
        croak "device type $args{type} needs major and minor numbers"
	    unless( exists $args{major} && exists $args{minor} );
	$opts = "$args{type} $args{major} $args{minor}";
    } else {
	croak "Illegal device type $args{type}";
    }
    NOTE::logcomment( "mknod $fullpath $opts" );
    if ( $Config{'osname'} eq 'solaris' ) {
	system( "sudo /etc/mknod $fullpath $opts" );
    } else {
	system( "sudo /bin/mknod $fullpath $opts" );
    }
    my $elem = $self->lookup( $path );
    croak "Didn't find $path" unless( $elem );
    return $elem;
}


=head3 ::path

Return working directory

=over

=item Syntax

NOTE::FileSystem::NFS->path();

=back

=cut

sub path {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_wdir}
}

=head3 ::fullpath

Returns full path for given location including mount point.

=over

=item Syntax

NOTE::FileSystem::NFS->fullpath($path);

=item Arguments

B<$path> => I<string>

I<Required>
Relative (filer's) path

=item Examples

=back

=cut

sub fullpath {
    my ($self,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix path with working directory if necessary
    $path = $self->path . $path unless ( $path =~ /^\// );

    return $mountpoint.$path;
}


=head3 ::chdir

Change working directory to $path.

=over

=item Syntax

NOTE::FileSystem::NFS->chdir($path);

=item Arguments

B<$path> => I<string>

I<Required>
New location

=back

=cut

sub chdir {
    my ($self,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    unless ( $self->lookup( $path ) ) {
	croak "$path: no such directory";
    }
    $self->{_wdir} = $path;
}


=head3 ::mkdir

description

=over

=item Syntax

NOTE::FileSystem::NFS->mkdir($path,$mode);

=item Arguments

B<$path> => I<string>

I<Required>
Directory path

B<$mode> => I<number>

I<Optional>
Permission settings. I believe this is not working properly.


=back

=cut

sub mkdir {
    my ($self,$path,$mode) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    my $mkpath = $mountpoint . $path;
    stat( $mkpath );
    unless ( -d _ ) {
	#mkdir( $mkpath, $mode ) or croak "$mkpath: $!";
	mkdir( $mkpath ) or croak "$mkpath: $!";
    }
    my $elem = $self->lookup( $path );
    croak "Didn't find $path" unless( $elem );
    return $elem;
}


=head3 ::link

Creates a new filename linked to the old filename. Returns true for success, false otherwise.

=over

=item Syntax

NOTE::FileSystem::NFS->link($old,$new);

=item Arguments

B<$old> => I<String>

I<Required>
Path to file

B<$new> => I<String>

I<Required>
Path to file

=back

=cut

sub link {
    my ($self,$old,$new) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix paths with working directory if necessary
    $old = $self->path . $old unless ( $old =~ /^\// );
    $new = $self->path . $new unless ( $new =~ /^\// );
    $new =~ s/\/$//;

    link( $mountpoint.$old, $mountpoint.$new) or croak "link $old $new: $!";

    my $elem = $self->lookup( $new );
    croak "Didn't find $new" unless( $elem );
    return $elem;
}


=head3 ::symlink

Creates a new filename symbolically linked to the old filename.
Returns 1 for success, 0 otherwise. On systems that don't support
symbolic links, produces a fatal error at run time. To check for that, use eval.

=over

=item Syntax

NOTE::FileSystem::NFS->symlink($old,$new);

=item Arguments

B<$old> => I<String>

I<Required>
Path to file

B<$new> => I<String>

I<Required>
Path to file

=back

=cut

sub symlink {
    my ($self,$old,$new) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix new link path with working directory if necessary
    $new = $self->path . $new unless ( $new =~ /^\// );

    symlink( $old, $mountpoint.$new ) or croak "symlink $old $new: $!";

    my $elem = $self->lookup( $new );
    croak "Didn't find $new" unless( $elem );
    return $elem;
}


=head3 ::unlink

Deletes a list of files. Returns the number of files successfully deleted.
Note: unlink will not attempt to delete directories unless you are superuser
and the -U flag is supplied to Perl. Even if these conditions are met, be
warned that unlinking a directory can inflict damage on your filesystem.
Finally, using unlink on directories is not supported on many operating
systems. Use rmdir instead.

=over

=item Syntax

NOTE::FileSystem::NFS->unlink($path);

=item Arguments

B<$path> => I<string>

I<Required>
Path to file or directory

=item Examples

=back

=cut

sub unlink {
    my ($self,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix path with working directory if necessary
    $path = $self->path . $path unless ( $path =~ /^\// );

    my $fullpath = $mountpoint.$path;
    unlink( $fullpath ) or croak "unlink $path: $!";
}

=head3 ::rmdir

Deletes the directory specified by PATH if that directory is empty.
If it succeeds it returns true, otherwise it returns false and
sets $! (errno). If PATH is omitted,

=over

=item Syntax

NOTE::FileSystem::NFS->rmdir($PATH);

=item Arguments

B<$PATH> => I<string>

I<Optional>
Path to directory

=back

=cut

sub rmdir {
    my ($self,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix path with working directory if necessary
    $path = $self->path . $path unless ( $path =~ /^\// );

    my $fullpath = $mountpoint.$path;
    rmdir( $fullpath ) or croak "rmdir $path: $!";
}


=head3 ::stat

Returns a 13-element list giving the status info for a file.

    Not all fields are supported on all filesystem types. Here are the meanings of the fields:

      0 dev      device number of filesystem
      1 ino      inode number
      2 mode     file mode  (type and permissions)
      3 nlink    number of (hard) links to the file
      4 uid      numeric user ID of file's owner
      5 gid      numeric group ID of file's owner
      6 rdev     the device identifier (special files only)
      7 size     total size of file, in bytes
      8 atime    last access time in seconds since the epoch
      9 mtime    last modify time in seconds since the epoch
     10 ctime    inode change time in seconds since the epoch (*)
     11 blksize  preferred block size for file system I/O
     12 blocks   actual number of blocks allocated

    (The epoch was at 00:00 January 1, 1970 GMT.)

    (*) Not all fields are supported on all filesystem types. Notably,
    the ctime field is non-portable. In particular, you cannot expect
    it to be a ``creation time'', see Files and Filesystems in the
    perlport manpage for details.


=over

=item Syntax

    ($dev,$ino,$mode,$nlink,
     $uid,$gid,$rdev,$size,
     $atime,$mtime,$ctime,
     $blksize,$blocks
    ) = NOTE::FileSystem::NFS->stat($path);

=item Arguments

B<$path> => I<string>

I<Required>
Path to file or directory

=back

=cut

sub stat {
    my ($self,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix path with working directory if necessary
    $path = $self->path . $path unless ( $path =~ /^\// );
    my $fullpath = $mountpoint.$path;
    lstat( $fullpath );
}


=head3 ::fill

Writes data to file

=over

=item Syntax

NOTE::FileSystem::NFS->fill($path,%args);

=item Arguments

B<$path> => I<string>

I<Required>
Path to file

B<$args{timeout}> => I<seconds>

I<Optional>
Default is 60 seconds

B<$args{max}> => I<[0-9]tgmbk>

I<Optional>
Do not fill over maxmum number of Tb, Gb, Mb, Blocks, or Kb

B<$args{proto}> => I<string>

I<Optional>
Protocol, default is 'nfs'

B<$args{type}> => I<data|sparse>

I<Optional>
Fill up type, default is 'data'

B<$args{percent}> => I<number>

I<Optional>
Fill up only so much, default is 100.

=back

=cut

sub fill {
    my ($self,$path,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $percent;
    my $start_time;
    my $total;
    my $log_interval = 5; # seconds
    my $timeout = $args{timeout} || 60;
    my $fname;
    my $maxmegs = $args{max};
    my $proto = $args{proto} || "nfs";
    my $type = $args{type} || "data";
    my $total_secs;
    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix path with working directory if necessary
    $path = $self->path . $path unless ( $path =~ /^\// );
    my $fullpath = $mountpoint.$path;
    NOTE::logcomment( "Filling $fullpath" );

    my $bytes = Utils::Units::convert_to_bytes( $args{max} );
    if ( $bytes < 1024*1024 ) {
	my $buf = pack( "C", rand(0xff) ) x ($bytes);
	$self->Print( $path, $buf );
	NOTE::logcomment( "Wrote $bytes [".$args{max}."] bytes" );
	return $self;
    }

    if ( $args{max} =~ /[tgmbk]$/i ) { # unit specifier?
	# NOTE::units returns KB; convert to MB
	$maxmegs = POSIX::ceil(NOTE::units( $args{max} ) / 1024);
	$maxmegs = 1 if ( $maxmegs == 0 );
    } else {
	$maxmegs = $args{max};
    }
    
    $percent = $args{percent} || 100;

    $total = ((($percent/100) * $self->{_container}->total) - $self->{_container}->used) / 1024;
    if ( defined $maxmegs ) {
	$total = $total > $maxmegs ? $maxmegs : $total;
    }

    $fname = $fullpath;
    if ( open( F, "+<", $fname ) ) {
	sysseek( F, 0, 2 ); # seek to end
	NOTE::logcomment( "Open existing $fname" );
    } else {
	open( F, ">", $fname ) or croak "Can't open $fname: $!";
	NOTE::logcomment( "Open new $fname" );
    }

    if ( $total < 0 ) {
	# Need to shrink filefile
	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	    $atime,$mtime,$ctime,$blksize,$blocks)
	    = CORE::stat(F);
	
	my $truncpos = (($size / (1024*1024)) - abs($total)) * (1024*1024);
	if ( $truncpos < 0 ) {
	    NOTE::logcomment( "Can't reduce ".$self->{_container}->description );
	} else {
	    NOTE::logcomment( "Truncating $fname by %d MB", abs($total) );
	    truncate( F, $truncpos );
	}
	close F;
    } else {
	my $megs_to_write = $total;
	my $total_secs;

	if ( $type eq 'data' ) {
	    my $buf = pack( "C", rand(0xff) ) x (1024*1024);
	    NOTE::logcomment( "Writing %d MB to $fname", $megs_to_write );
	    $start_time = time;
	    my $next_log_time = $start_time + $log_interval;
	    my $last_megs = $megs_to_write;
	    while( $megs_to_write > 0 ) {
		if ( time >= $next_log_time ) {
		    my $megs_written = $last_megs - $megs_to_write;
		    $next_log_time = time + $log_interval;
		    NOTE::logcomment( "Wrote %d MB, %.3f MB/s, %.3f percent complete",
			$total - $megs_to_write, $megs_written / $log_interval,
			(($total - $megs_to_write) / $total) * 100 );
		    $last_megs = $megs_to_write;
		}
		eval {
		   local $SIG{ALRM} = sub { die "Timeout writing to $fname\n" };
		   alarm $timeout;
		   syswrite( F, $buf );
		   alarm 0;
		};
		croak $@ if ($@);
		$megs_to_write--;
	    }
	} elsif ( $type eq 'sparse' ) {
	    NOTE::logcomment( "Creating $total MB hole in $fname" );
	    $start_time = time;
	    $megs_to_write = 0;
	    sysseek( F, $total * (1024*1024)-1, 2 ); # seek to end + total
	    syswrite( F, '\0', 1 ); # Write one byte
	}
	close F;
	$total_secs = time - $start_time;
	$total_secs = 1 if ( $total_secs == 0 ); # Prevent divide-by-zero
	NOTE::logcomment( "Wrote %d MB in %d seconds, %.3f MB/s",
	    $total - $megs_to_write,
	    $total_secs,
	    $total / $total_secs
	);
    }
    return $self;
}

=head3 ::rename

Renames files or directory

=over

=item Syntax

NOTE::FileSystem::NFS->rename($from,$to);

=item Arguments

B<$from> => I<string>

I<Required>
Path to file or directory

B<$to> => I<string>

I<Required>
Path to file or directory

=back

=cut

sub rename {
    my ($self,$from,$to) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mountpoint = $self->{_container}->mountpoint;
    unless ( $mountpoint ) {
	croak "Filesystem not mounted: ".$self->{_container}->description;
    }

    # Prefix paths with working directory if necessary
    $from = $self->path . $from unless ( $from =~ /^\// );
    $to   = $self->path . $to   unless ( $to   =~ /^\// );

    my $full_from = $mountpoint . $from;
    my $full_to   = $mountpoint . $to;
   
    rename( $full_from, $full_to ) or croak "rename $full_from $full_to: $!";
}

sub is_open {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_fh};
}

sub Open {
    my ($self, $path, $mode) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $mode ||= '+<';
    $self->Close;
    $self->{_fh} = new IO::File($self->fullpath($path), $mode) or
	croak "Couldn't open @{[$self->fullpath]}: $!";
    return $self;
}

sub Close {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $self->{_fh} ) {
	$self->{_fh}->close;
        delete $self->{_fh};
    }
    return $self;
}

sub Print {
    my ($self, $path, @args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $was_open = $self->is_open;
    
    $self->Open( $path, "+<" ) unless $was_open;
    $self->{_fh}->print( @args );
    $self->Close() unless $was_open;
    return $self;
}

sub Read {
    my ($self,$path,$buf,$size,$offset) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $was_open = $self->is_open;

    $self->Open( $path, "<" ) unless $was_open;
    read( $self->{_fh}, $buf, $size, $offset || 0 ) or
	croak "read failed: @{[$self->fullpath]}: $!";
    $self->Close() unless $was_open;
    return $self;
}

sub Printf {
    my ($self,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->Print( sprintf( @args ) );
    return $self;
}

sub Seek {
    my ($self, $path, $position, $whence) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->Open( $path, "+>" ) unless $self->is_open;
    Seek( $self->{_fh}, $position, $whence ) or
	croak "seek failed: @{[$self->fullpath]}: $!";
    return $self;
}

sub Syswrite {
    my ($self,$path, $buf,$size,$offset) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $was_open = $self->is_open;

    $self->Open( $path, ">" ) unless $was_open;
    Syswrite( $self->{_fh}, $buf, $size, $offset || 0 ) or
	croak "syswrite failed: @{[$self->fullpath]}: $!";
    $self->Close unless $was_open;
    return $self;
}

sub Sysread {
    my ($self,$path, $buf,$size,$offset) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $was_open = $self->is_open;

    $self->Open( $path, "<" ) unless $was_open;
    Sysread( $self->{_fh}, $buf, $size, $offset || 0 ) or
	croak "sysread failed: @{[$self->fullpath]}: $!";
    $self->Close unless $was_open;
    return $self;
}


1;
