#$Id: //depot/prod/test/Rsierranevada/lib/GNOTE/FileSystem/NFS/File.pm#1 $
## @summary NOTE FileSystem NFS File Class
## @status Public
## @pod here
## @author dl-dpgqa-automation@netapp.com

package NOTE::FileSystem::NFS::File;

use GNOTE;

use bigint;
use constant SNAPLOCK_EPOCH => 1044057600; # 00:00:00 Jan 1 2003
use constant BIG_INFINITY   => 0x7fffffffffffffff;
use constant SMALL_INFINITY => 0x7fffffff;

use FileManip qw( setattr_file );
use fileSystem qw( write_file );
use FS::Util::Utilities qw( file_reserve );
use Carp;
use Config;
use snaplock;
use POSIX;
use Date::Calc;

our @ISA = qw(NOTE);
use Carp;
use Config;

=head2 METHODS

=head3 ::init

Object constractor. Normally called by NOTE::Volume->nfs or NOTE::Qtree->nfs.

=over

=item Syntax

NOTE::FileSystem::NFS::File->init($container,$path,%args);

=item Arguments

B<$container> => I<obj>

I<Required>
Volume or Qtree object

B<$path> => I<string>

I<Required>
Path to file or directory

B<%args> => I<hash>

I<Required>
Options for NOTE::FileSystem::NFS->mount.

=item Examples

=back

=cut

sub init {
    my ($class,$container,$path,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    my $model;

    $self->{_fs} = NOTE::FileSystem::NFS->mount(
	$container,
	%args,
    );
    unless ( $model = $self->{_fs}->lookup( $path ) ) {
	croak "Can't locate: $path";
    }
    $self->{_model} = $model;
    bless $self, $class;
    return $self;
}


=head3 ::lookup

Alias to NOTE::FileSystem->lookup. Returns "NOTE::FileSystem::NFS::File" object.

=over

=item Syntax

NOTE::FileSystem::NFS::File->lookup($path);

=item Arguments

B<$path> => I<string>

I<Required>
Path to file or directory

=back

=cut

sub lookup {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_fs} = $this->{_fs};
    $self->{_model} = $self->{_fs}->lookup(
	$path
    );
    return undef unless $self->{_model};

    bless $self, "NOTE::FileSystem::NFS::File";
    return $self;
}

=head3 ::create

Creates an empty file. If no path is specified, an anonymous
file is created.

=over

=item Examples

 $root = $qt->nfs;
 my $file = $root->create;
 $file->fill( max => '100m' );

=back

=cut

sub create {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $model;
    my ($self, $path) = _get_new_self($this, "Create", "fl", $path);
    unless ( $model = $this->{_fs}->create( $path ) ) {
	croak "Failed to create: $path";
    }
    return _bless_new_self($self, $model);
}

sub _get_new_self {
    my ($old_self, $action, $prefix, $path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    $self->{_fs} = $old_self->{_fs};

    unless ( defined $path ) {
	$path = $old_self->path . $old_self->newName( $prefix );
    }
    # make sure path is rooted
    $path = $old_self->path . $path unless ( $path =~ /^\// );

    NOTE::logcomment(
	"$action $path\n".
	"    on ".$self->{_fs}->{_container}->description
    );
    return ($self, $path);
}

sub _bless_new_self {
    my ($self, $model) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_model} = $model;
    $self->{_owner} = $$;
    
    bless $self, "NOTE::FileSystem::NFS::File";
    NOTE::reference( $self );
    return $self;
}

=head3 ::create_fifo

Create a fifo. If no fifopath is given, create fifo with
an anonymous name

=over

=item Arguments

B<pathname> => I<pathname>

Path for the FIFO.

B<mode> => I<file mode>

Defaults 0777

B<client> => I<client>

Default: localhost

B<timeout> => I<timeout in seconds>

Default: 3600

=item Example

    $root = $qt->nfs;
    my $fifo = $root->create_fifo(

=back

=cut

sub create_fifo {
    my ($this, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, $path) = _get_new_self($this, "Create fifo", "fi", delete $args{pathname} );
    my $fifopath = $this->{_fs}->fullpath($path);
    croak "FIFO $fifopath already exists..."
	if -e $fifopath;
    waflobjCreator::create_fifo($fifopath, %args);
    croak "File to create FIFO $fifopath. Error $@" if $@;

    return _bless_new_self($self, $this->{_fs}->lookup($path));
}

=head3 ::create_socket

Create a socket. If no socketpath given, create socket with
an anonymous name

=over

=item Example

 $root = $qt->nfs;
 my $socket = $root->create_socket;

=back

=cut

sub create_socket {
    my ($this, $socketpath) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, $path) = _get_new_self($this, "Create socket", "so", $socketpath );
    
    my $fullpath = $this->{_fs}->fullpath( $path );
    croak "Socket $fullpath already exists..."
	if -e $fullpath;
    my $rc = waflobjCreator::create_socket($fullpath);
    croak "Fail to create socket $fullpath. Error $@" if $@;

    return _bless_new_self($self, $this->{_fs}->lookup( $path ));
}

=head3 ::attach

Attaches an existing file.

=over

=item Examples

 $conf = $qt->nfs->attach( "/etc/snapmirror.conf" );

=back

=cut

sub attach {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $model;
    my $self = {};
    $self->{_fs} = $this->{_fs};

    # make sure path is rooted
    $path = $this->path . $path unless ( $path =~ /^\// );
    unless ( $model = $this->{_fs}->lookup( $path ) ) {
	croak "Can't locate: $path";
    }
    $self->{_model}     = $model;
    bless $self, "NOTE::FileSystem::NFS::File";
    return $self;
}

=head3 ::mknod

Alias to NOTE::FileSystem::NFS->mknod().

=over

=back

=cut

sub mknod {
    my ($this,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $model;
    my $self = $this->_get_new_self("mknod", "dv");
    my $self = {};
    $self->{_fs} = $this->{_fs};

    my $path = $args{name};
    unless ( defined $path ) {
	$path = $this->newName( "dv" );
    }

    # make sure path is rooted
    $path = $this->path . $path unless ( $path =~ /^\// );

    delete $args{name};
    NOTE::logcomment(
	"mknod $path '".$args{type}."'\n".
	"    on ".$self->{_fs}->{_container}->description
    );
    unless ( $model = $this->{_fs}->mknod( name => $path, %args ) ) {
	croak "Can't mknod: $path";
    }
    $self->{_model} = $model;
    $self->{_owner} = $$;
    
    bless $self, "NOTE::FileSystem::NFS::File";
    NOTE::reference( $self );
    return $self;
}

=head3 ::stat

Returns the status information for a file. The effect is the
same as the perl \C<stat> function. There are also several
convenience methods for getting the individual status elements:

=over

=item Convenience methods

::dev
Returns the device number of the filesystem.

::inode
Returns the inode number of the file

::mode
Returns the file permissions

::type
Returns a string indicating the type of the file:
Possible values are C<'fifo'>, C<'chardev'>,
C<'directory'>,  C<'blockdev'>, C<'file'>, C<'symlink'>, C<'socket'>

::nlink
Returns the number of hardlinks.

::uid
Returns the user id of the file's owner.

::gid
Returns the group id of the file's group.

::rdev
Returns the device information for this file, if this
is a special file.

::size
Returns the size of the file in bytes.

::atime
Returns the time of last access.

::mtime
Returns the time of last modification.

::ctime
Returns the time that the inode was last changed.

::blksize
Returns the preferred block size for file system I/O

::blocks
Returns the number of blocks allocated for this file.

=item Examples

 $qt->nfs->stat( "/home" );
 if ( -d _ ) {
    print "/home exists and is a directory\n";
 }

 $f = $qt->nfs->create( "/home/myfile" );
 $f->fill( max => "300m" );
 print( "After fill, %s is %d bytes\m", $f->name, $f->size );

=back

=cut

sub stat {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $path ) {
	$this->{_fs}->stat( $this->path . $path );
    } else {
	$this->{_fs}->stat( $this->path );
    }
}

sub dev    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[0];
}

sub inode  {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[1];
}

sub mode   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[2] & 07777;
}

sub type   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my @types = (
	'fifo','chardev', 'directory', 'blockdev',
	'file', 'symlink', 'socket'
    );
    return $types[((((shift->stat)[2] >> 13) & 0x7))];
}

sub nlink  {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[3];
}

sub uid    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[4];
}

sub gid    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[5];
}

sub rdev   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[6];
}

# The contents of the following methods (major and
# minor) could/should be replaced by calls to equivalent
# functions in Unix::mknod, if that package is ever installed
# globally. In the mean-time, these should be adequate for
# linux and both 32 and 64 bit versions of solaris.

=head3 ::major

Returns major number of device file

=over

=item Syntax

$obj->NOTE::FileSystem::NFS::File->major();

=item Arguments

None

=back

=cut

sub major  {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $rdev = shift->rdev;
    if ( $Config{'osname'} eq 'solaris' ) {
	if ( $Config{'archname'} =~ /64int/ ) {
	    return ( ($rdev >> 32) & 0xFFFFFFFF ); # 32 bit major
	} else {
	    return ( ($rdev >> 18) & 0x3FFF ); # 14 bit major
	}
    } elsif ( $Config{'osname'} eq 'linux' ) {
	#return ((($rdev >> 8) & 0xfff) | (($rdev >> 32) & ~0xfff));
	return (($rdev >> 8) & 0xff);
    } else {
	croak "Don't know how to get major number";
    }
}

=head3 ::minor

Returns minor number of device file

=over

=item Syntax

$obj->NOTE::FileSystem::NFS::File->minor();

=item Arguments

None

=back

=cut

sub minor  {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $rdev = shift->rdev;
    if ( $Config{'osname'} eq 'solaris' ) {
	if ( $Config{'archname'} =~ /64int/ ) {
	    return ( $rdev & 0xFFFFFFFF ); # 32 bit minor
	} else {
	    return ( $rdev & 0x3FFFF ); # 18 bit minor
	}
    } elsif ( $Config{'osname'} eq 'linux' ) {
	#return (($rdev & 0xff) | (($rdev >> 12) & ~0xff));
	return ($rdev & 0xff);
    } else {
	croak "Don't know how to get minor number";
    }
}

sub size   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[7];
}

sub atime  {
    my ($this,$format) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $timestamp = ($this->stat)[8];
    if ( $this->Volume->snaplock ) {
	if ( $timestamp == SMALL_INFINITY ) {
	    $timestamp = BIG_INFINITY;
	} elsif ( $timestamp < SNAPLOCK_EPOCH ) {
	    # Timestamps earlier than snaplock epoch represent dates
	    # after 2038.
	    $timestamp += 0x7fffffff;
	}
    }
    if ( defined $format ) {
	# Return a string representing a date
	if ( $timestamp == BIG_INFINITY ) {
	    return "the day Satan skates to work";
	}
	my $t = Date::Manip::ParseDateString( "epoch $timestamp" );
	return Date::Manip::UnixDate( $t, $format );
    } else {
	return $timestamp;
    }
}

sub mtime  {
    my ($this,$format) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $timestamp = ($this->stat)[9];
    if ( defined $format ) {
	my $t = Date::Manip::ParseDateString( "epoch $timestamp" );
	return Date::Manip::UnixDate( $t, $format );
    } else {
	return $timestamp;
    }
}

sub ctime  {
    my ($this,$format) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $timestamp = ($this->stat)[10];
    if ( defined $format ) {
	my $t = Date::Manip::ParseDateString( "epoch $timestamp" );
	return Date::Manip::UnixDate( $t, $format );
    } else {
	return $timestamp;
    }
}

sub blksize{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[11];
}

sub blocks {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    (shift->stat)[12];
}

sub _mkdir {
    my ($this,$name,$mode) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $model;
    my $self = {};
    $self->{_fs} = $this->{_fs};

    unless ( defined $name ) {
	$name = $this->newName( "dr" );
    }
    my $path = $this->path . $name;
    if ( $model = $this->{_fs}->lookup( $path ) ) {
	#NOTE::logcomment( "exists: path is %s", $model->path );
    } elsif ( $model = $this->{_fs}->mkdir( $path, $mode ) ) {
	#NOTE::logcomment( "made: path is %s", $model->path );
    } else {
	croak "Can't mkdir $path: $!";
    }
    $self->{_model} = $model;
    $self->{_owner} = $$;
    $self->{_is_dir} = 1; 
    
    bless $self, "NOTE::FileSystem::NFS::File";
    NOTE::reference( $self );
    return $self;
}

=head3 ::mkdir

Creates a directory and returns an object referring to it.

=over

=item Arguments

I<path>

If path is not specified, an anonymous directory will be created.
If path is specified, all elements in the path will be created,
unless they exist already.

=item Examples

 $home = $qt->nfs->mkdir( "/home" );
 $dir = $home->mkdir->mkdir; # anonymous directory two levels deep

=back

=cut

sub mkdir {
    my ($this,$path,$mode) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $nextelem;
    my $tmp;
    my $fs = $this->{_fs};

    if ( !defined $path ) {
	$this = $this->_mkdir( undef, $mode );
    } else {
	NOTE::logcomment(
	    "mkdir $path\n".
	    "    on ".$this->{_fs}->{_container}->description
	);
	my @elems = split /\//, $path;
	while ( @elems ) {
	    $nextelem = shift(@elems);
	    next unless ( $nextelem );
	    next if ( $nextelem eq "." );
	    if ( $nextelem eq ".." ) {
		$this = $this->parent;
		next;
	    }
	    $this = $this->_mkdir( $nextelem, $mode );
	}
    }
    return $this;
}


=head3 ::rename

Alias for C<mv>

=cut

sub rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    mv( @_ )
}


=head3 ::mv

Move file or directory to new location

=over

=item Syntax

NOTE::FileSystem::NFS::File->mv($from_path,$to_path);

=item Arguments

B<$from_path> => I<string>

I<Required>
A 'FROM' path.

B<$to_path> => I<string>

I<Optional>
A new location path.  No defined destination means just rename the current
node, anonymously.

If destination exists: If it is a directory,
the destination must be transformed to have the basename
of the source.

=back

=cut

sub mv {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($from_path,$to_path);
    my ($from,$to,$abs);
    my $argc = scalar @_;
    my ($arg0,$arg1) = @_;

    # Set flag for leading '/' and remove it
    $abs = ($arg0 =~ s/^\///) if ( $arg0 );
    $from_path = $self->path;
    if ( $argc > 1 ) {
	$from_path .= $arg0 if ( $arg0 );
        $abs = ($arg1 =~ s/^\///) if ( $arg1 );
	$to_path   = $arg1;
    } else {
	$to_path   = $arg0;
    }
    $from = $self->attach( $from_path );

    if ( ! defined $to_path ) {
	# No defined destination means just rename the current
	# node, anonymously
	$to_path = $from->newName( $from->is_dir ? "dl" : "fl" );
    }
    # if destination exists: If it is a directory,
    # the destination must be transformed to have the basename
    # of the source.
    if ( ($to = $self->lookup( $to_path )) && $to->is_dir ) {
	$to_path = $to->path . $from->basename;
    }
    unless ( $abs || $to_path =~ /^\// ) {
	$to_path = $from->parent->path . $to_path;
    }
    $to_path =~ s/\/$//;

    NOTE::logcomment(
        "mv $from_path --> $to_path\n".
	"    on ".$from->{_fs}->{_container}->description
    );
    $from->{_fs}->rename( $from_path, $to_path );
    $from->{_model}->move( $to_path );
    return $self;
}

=head3 ::rmdir

Removes directory

=over

=item Syntax

NOTE::FileSystem::NFS::File->rmdir($path);

=item Arguments

B<$path> => I<string>

I<Required>
Directory path

=back

=cut

sub rmdir {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $path ) {
	$path = $this->path . $path;
    } else {
	$path = $this->path;
    }
    $this->{_lastpath} = $path;
    return unless ( $this->{_fs}->stat( $path ) ); # already gone

    my $what = $this->{_fs}->lookup( $path );
    croak "cannot rmdir $path" unless ( $what->is_dir );
    $this->{_fs}->rmdir( $path );
    croak "Couldn't rmdir $path" if $this->{_fs}->stat( $path );
}


=head3 ::unlink

Calls Perl's unlink command for given path.

=over

=item Syntax

NOTE::FileSystem::NFS::File->unlink($path);

=item Arguments

B<$path> => I<string>

I<Required>
Directory path

=back

=cut

sub unlink {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $path ) {
	$path = $this->path . $path;
    } else {
	$path = $this->path;
    }
    $this->{_lastpath} = $path;
    my $what = $this->{_fs}->lookup( $path );
    if ( $what ) {
	croak "Unlinking directories not allowed: $path" if ( $what->is_dir );
	$this->{_fs}->unlink( $path );
	croak "Couldn't unlink $path" if $this->{_fs}->stat( $path );
    }
}


=head3 ::rm

Remove directories and their contents recursively

=over

=item Syntax

NOTE::FileSystem::NFS::File->rm($path);

=item Arguments

B<$path> => I<string>

I<Required>
Directory path

=back

=cut

sub rm {
    my ($this,$path) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $path ) {
	$path = $this->path . $path;
    } else {
	$path = $this->path;
    }
    my $what = $this->{_fs}->lookup( $path );
    if ( $what->is_dir ) {
	next if ( $path eq '/.snapshot/' );
	for my $child ( $this->children ) {
	    $child->rm;
	}
	$this->rmdir unless ( $path eq '/' ); # Don't remove root
    } else {
	$this->unlink;
    }
}


=head3 ::touch

Touch NOTE::FileSystem::NFS::File object. Sets atime and mtime to the current time unless
atime and mtime arguments are provided.

=over

=item Syntax

$obj->touch(%args);

=item Arguments

B<$args{atime}> => I<number>

I<Optional>
Reset access time to given number. Time from 0 to 0x7fffffff.

B<$args{mtime}> => I<number>

I<Optional>
Reset modification time to given number.

=item Examples

=back

=cut

sub touch {
    my ($this,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Default is to set both atime and mtime to the current time.
    $args{atime} = $args{mtime} = time
	unless ( defined $args{atime} || defined $args{mtime} );

    if ( defined $args{atime} ) {
	if ( $args{atime} > 0x7fffffff ) {
	    # Wrap around to support dates between 2038 and 2071
	    $args{atime} -= 0x7fffffff;
	}
    } else {
	$args{atime} = $this->atime;
    }
    $args{mtime} = $this->mtime unless ( defined $args{mtime} );

    my $path = $this->path;
    croak "Couldn't touch $path: $!"
	unless $this->{_fs}->utime( $path, $args{atime}, $args{mtime} );
    return $this;
}


=head3 ::chmod

Change permission

=over

=item Syntax

$obj->chmod($mode);

=item Arguments

B<$mode> => I<number>

I<Optional> Permission mode. Default is '0777'.
If the chmod param exists, but is set to undef, don't
do a chmod. This is useful for mounting read-only filesystems.

=back

=cut

sub chmod {
    my ($this,$mode) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $path = $this->path;
    croak "Couldn't chmod $path"
	unless $this->{_fs}->chmod( $path, $mode );
    return $this;
}


=head3 ::link

Object wrapper to NOTE::FileSystem::NFS->link.
Creates a new filename linked to the old filename. Returns true for success, false otherwise.

=over

=item Syntax

$obj->link($new);

=item Arguments

B<$new> => I<efg>

I<Required>
New filename

=back

=cut

sub link {
    my ($this,$new) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    my $old = $this->path;

    $self->{_fs} = $this->{_fs};

    unless ( defined $new ) {
	$new = $this->newName( "ln" );
    }
    $new = $this->dirname . $new unless ( $new =~ /^\// );

    NOTE::logcomment(
        "link $old <-- $new\n".
	"    on ".$this->{_fs}->{_container}->description
    );

    $self->{_model}  = $self->{_fs}->link( $old, $new );
    $self->{_owner}  = $$;
    $self->{_is_dir} = $this->is_dir; 
    
    bless $self, "NOTE::FileSystem::NFS::File";
    NOTE::reference( $self );
    return $self;
}

sub pathto {
    ($source,$target) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @s = split( /\//, $source->path );
    my @t = split( /\//, $target->path );

    shift @s while ( @s && $s[0] eq '' );
    shift @t while ( @t && $t[0] eq '' );

    while ( @s && @t && ($s[0] eq $t[0]) ) {
        shift @s;
        shift @t;
    }
    return (('../' x scalar @s) . join( '/', @t )) || '.';
}



=head3 ::symlink

Creates a new filename symbolically linked to the old filename.
Returns 1 for success, 0 otherwise. On systems that don't support
symbolic links, produces a fatal error at run time. To check for that, use eval.

=over

=item Syntax

$fileObj->symlink( [$name] );

$dirObj->symlink( [$obj[, $name]] );

=item Arguments

B<$name> => I<new link name>

I<Optional>

B<$obj> => I<target of the link> If the calling object is
a directory, the first argument may be an object referring to
a file or directory. In that case, the symlink created will
contain a relative link to that file or directory. For example:

 $file = $vol->nfs->mkdir( 'top/A1/A2' )->create( 'somefile' );
 $dir  = $vol->nfs->mkdir( 'top/B1/B2' );
 $syml = $dir->symlink( $file, "mysymlink" );

This will create a symlink called top/B1/B2/mysymlink whose value is
"../../A1/A2/somefile".
 

=item Examples

 $file = $vol->nfs->create;
 $syml = $vol->nfs->mkdir( "L1/L2/L3" )->symlink( $file );

=back

=cut

sub symlink {
    my $self = {};
    my $this = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($source,$target);
    if ( @_ > 1 ) {
	$target = $this->pathto( shift );
    } else {
	$target = $this->basename;
    }
    my $linkname = shift;

    $self->{_fs} = $this->{_fs};

    unless ( defined $linkname ) {
	$linkname = $this->newName( "sl" );
    }
    unless ( $linkname =~ /^\// ) {
	if ( $this->is_dir ) {
            $linkname = $this->path . $linkname;
	} else {
            $linkname = $this->dirname . $linkname;
	}
    }

    NOTE::logcomment(
        "symlink $target <-- $linkname\n".
	"    on ".$this->{_fs}->{_container}->description
    );

    $self->{_model}  = $self->{_fs}->symlink( $target, $linkname )
	or croak "Couldn't find new symlink: $linkname";
    $self->{_owner}  = $$;
    $self->{_is_dir} = $this->is_dir; 
    
    bless $self, "NOTE::FileSystem::NFS::File";
    NOTE::reference( $self );
    return $self;
}

sub open {
    my ($this,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "open ".$this->fullpath );

    $self->{_fs}->open( $this->path, @args )
	or croak "Couldn't open: ".$this->fullpath.": $?";
    return $this;
}

sub print {
    my ($this,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "print ".$this->fullpath );

    $this->{_fs}->print( $this->path, @args )
	or croak "Couldn't print: ".$this->fullpath.": $?";
    return $this;
}


=head3 ::readlink

Returns the value of a symbolic link, if symbolic links are implemented.
If not, gives a fatal error. If there is some system error, returns the
undefined value and sets $! (errno).

=over

=item Syntax

$obj->readlink();

=item Arguments

None

=back

=cut

sub readlink {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return readlink shift->fullpath;
}


=head3 ::fill

Writes data to file

=over

=item Syntax

$obj->fill(%args);

=item Arguments

B<$args{timeout}> => I<seconds>

I<Optional>
Default is 60 seconds

B<$args{max}> => I<[0-9]tgmbk>

I<Optional>
Do not fill over maxmum number of Tb, Gb, Mb, Blocks, or Kb

B<$args{proto}> => I<string>

I<Optional>
Protocol, default is 'nfs'

B<$args{type}> => I<data|sparse>

I<Optional>
Fill up type, default is 'data'

B<$args{percent}> => I<number>

I<Optional>
Fill up only so much, default is 100.

=back

=cut

sub fill {
    my ($this,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $this->{_fs}->fill( $this->path, @args );
    return $this;
}

=head3 ::modify

Randomly modify data in file
Croaks if fails.
Wrapper for
L<C<fileSystem::write_file>|engineering::ontap_tests::agnostic::docs::lib-fileSystem-pm/write_file()>.

=over

=item Arguments

B<offset> => <integer>

Offet in file to start modification. Default: 0

B<size> => <integer>

Number of bytes to modify. Default: 1

=item Example

     $file->modify(offset=>512, size=>4);

=back

=cut

sub modify {
    my ($this, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $offset = $args{offset} ? $args{offset} : 0; delete $args{offset};
    my $size = $args{size} ? $args{size} : 1; delete $args{size};

    croak "Failed to modify ".$args{size}." bytes at offset ".$args{offset}
	unless write_file($this->fullpath, $offset, $size, %args);
}

=head3 :truncate

Truncate a file to a particular size

=over

=item Arguments

B<size> => <integer>

Size, in bytes, to truncate a file size to

=item Example

    $file->fille(max=>2);
    $file->truncate(size=>63);

=back

=cut

sub truncate {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{size} = 0 unless defined $args{size};
    croak "Invalid size of $args{size} to truncate. Must be >= 0"
	unless $args{size} >= 0;

    truncate $self->fullpath, $args{size};
}

=head3 ::setattr

Set file attributes such as uid, gid, mode, size etc
Wrapper for
L<C<FileManip::setattr_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/setattr_file()>.

Croaks if object is not a file or failed to set the attribute

=over

=item Arguments

B<client> => I<client name>

Performs action on client, default localhost.

B<uid> => I<new UID>

Set file's uid

B<gid> => I<new GID>

Set file's gid

=item Example

    $file->setattr(uid=>1234);

=back

=cut

sub setattr {
    my ($this, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak $this->fullpath." is not a file"
	unless $this->type() =~ /file/i;

    my $rc = 0;
    $args{client} = "localhost" unless $client;
    eval {
	$rc = setattr_file(file=>$this->fullpath, %args);
    };
    croak "Error detected setting file attr: $@" if $@;
    croak "Failed to set file attr" unless $rc;

    return $this;
}

=head3 ::reservation

Set hole reservation to 'on' or 'off'
Wrapper for 
L<C<FS::Utils::Utilities::file_reserve>|engineering::ontap_tests::agnostic::docs::lib-FS-Util-Utilities-pm/file_reserve()>.

=over

=item Argument

B<flag> = I<on|off>

=item Example

     $file->reservation("on");

=back

=cut

sub reservation {
    my ($this, $flag) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return file_reserve ($this->Volume->Filer->conn, 
			$this->{_fs}->{_container}->mount_path."/".$this->$this->path,
			reservation => $flag);
}

=head3 ::fullpath

Returns full path of given object.

=over

=item Syntax

$obj->fullpath();

=item Arguments

None

=back

=cut

sub fullpath {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->{_fs}->fullpath( $self->path );
}

=head3 ::path

Returns relative path of given object

=over

=item Syntax

$obj->path();

=item Arguments

None

=back

=cut

sub path {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $self->{_destroyed} ) {
	return $self->{_lastpath};
    }
    return $self->{_model}->path;
}

=head3 ::name

Alias to C<basename>

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->basename;
}

=head3 ::basename

Returns basename

=over

=item Syntax

$obj->basename();

=item Arguments

None

=back

=cut

sub basename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $path = shift->path;
    $path =~ s/\/$//; # Remove any trailing /
    $path =~ /\/([^\/]*$)/;
    return $1 || $path;
}

=head3 ::dirname

Returns directory name of the object.

=over

=item Syntax

$obj->dirname();

=item Arguments

None

=back

=cut

sub dirname {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $path = shift->path;
    my $dir = '';
    my @elems = split( /\//, $path );
    pop @elems;
    if ( @elems ) {
        $dir = join( '/', @elems );
    }
    return $dir . '/';
}

=head3 ::is_dir

Returns true if object is a directory

=over

=item Syntax

$obj->is_dir();

=item Arguments

None

=back

=cut

sub is_dir {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_model}->is_dir;
}

=head3 ::parent

Returns parent as "NOTE::FileSystem::NFS::File" object.

=over

=item Syntax

$parent = $obj->parent();

=item Arguments

None

=back

=cut

sub parent {
    my $this = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    $self->{_spare_me} = 1;
    $self->{_fs}    = $this->{_fs};
    $self->{_model} = $this->{_fs}->lookup( $this->{_model}->parent->path );
    $self->{_owner} = $$;
    bless $self, "NOTE::FileSystem::NFS::File";
    return $self;
}

# Return a list of objects that represent directory entries

=head3 ::children

Returns all children files and directories as
array of objects of "NOTE::FileSystem::NFS::File" type

=over

=item Syntax

$obj->children();

=item Arguments

None

=back

=cut

sub children {
    my $this = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @kids = ();

    if ( $this->is_dir ) {
        my $dirpath = $this->fullpath;
	opendir( DIR, $dirpath ) or die "Can't opendir $dirpath: $!";
	for my $entry ( readdir(DIR) ) {
	    next if $entry eq '.' || $entry eq '..';
	    push @kids, $this->attach( $entry );
	}
	closedir DIR;
    }
    return @kids;
}

=head3 ::create_files

Create the specified number of files in a directory
Croaks if object is not a directory

Returns list of File objects created

=over

=item num_entries

The number of files to create

=item Example

    my @files = $dir->create_files(5);

=back

=cut

sub create_files {
    my ($this, $num_entries) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak $this->fullpath." is not a directory"
	unless $this->is_dir;

    my @list;
    while ($num_entries--) {
	push @list, $this->create;
    };

    return @list;
}

=head3 ::Volume

Returns volume as NOTE::Volume object

=over

=item Syntax

$obj->Volume();

=item Arguments

=back

=cut

sub Volume {
    my $this = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $this->{_fs}->{_container}->isa( 'NOTE::Qtree' ) ) {
	return $this->{_fs}->{_container}->Volume;
    } else {
	# Must be a volume
	return $this->{_fs}->{_container};
    }
}


=head3 ::ls

Returns output of ls

=over

=item Syntax

$obj->ls(@args);

=item Arguments

B<@args> => I<strings>

I<Optional>
Arguments are passed to ls command as

 ls @args $self->fullpath

=back

=cut

sub ls {
    my ($self,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $cmd = "ls @args ".$self->fullpath;
    if ( defined wantarray ) {
	return `$cmd 2>&1`;
    } else {
	# When called in void context, do a logcomment:
	# $q->nfs->ls( '-lR' );
	NOTE::logcomment( "$cmd\n".`$cmd 2>&1` );
    }
}

=head3 ::tree

Returns output of tree

=over

=item Syntax

$obj->tree(@args);

=item Arguments

B<@args> => I<strings>

I<Optional>
Arguments are passed to tree command as

 tree @args $self->fullpath

=back

=cut

sub tree {
    my ($self,@args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $cmd = "tree @args ".$self->fullpath;
    if ( defined wantarray ) {
	return `$cmd 2>&1`;
    } else {
	# When called in void context, do a logcomment:
	# $q->nfs->tree( '-F' );
	NOTE::logcomment( "$cmd\n".`$cmd 2>&1` );
    }
}

# Snaplock support

=head3 ::commit

Commits a file on a SnapLock volume to WORM state.

With this function, it is possible to commit files with
infinite retention, however, as of 7.2, it must do this by temporarily
setting the filer's default retention period to infinite and
allowing the file to acquire the default retention.
Note that the actual retention is bounded by the filer's
maximum retention period, which is NOT changed by this function.

=over

=item Arguments

B<retention> => I<retention period>

I<Optional.>
If this option is specified, the specified retention period
will be set before the commit.

B<force> => C<0> | C<1>

I<Optional.>
Normally, this commit function will refuse to commit files to
compliance volumes (but not enterprise volumes) when the retention
period is more the 24 hours. This flag will override that safety feature.

=item Examples

 $file = $vol->nfs->create->fill( max => '1m' );
 $file->commit( retention => '120s' );

=back

=cut

sub commit {
    my ($this,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $vol        = $this->Volume;
    my $lock_model = $this->Volume->snaplock;
    croak "Tried to commit file residing on non-snaplock volume"
	unless ( $lock_model =~ /(enterprise|compliance)/ );

    my $comp_time = $vol->Filer->time( clock => 'compliance' );
    my $sys_time  = $vol->Filer->time( clock => 'system' );
    my $skew      = $sys_time - $comp_time;
    my $retention;

    if ( defined $args{retention} ) { # Set retention first
	$retention = $this->retention( $args{retention} );
    } else {
	$retention = $this->retention; # Get retention
    }

    NOTE::logcomment( "Commiting %s with %d seconds retention", $this->path, $retention );
    if ( ($retention > 24*60*60) && ($lock_model eq 'compliance') ) {
	croak "For compliance volumes, cannot commit if retention > 1 day unless force => 1."
	    unless ( $args{force} == 1 );
    }
    my $def_period_save = undef;
    if ( $retention == BIG_INFINITY ) {
	$def_period_save = $vol->{'options.snaplock_default_period'};
	$vol->{'options.snaplock_default_period'} = 'infinite';
	$this->touch( atime => $this->ctime ); # After commit, atime will become 0x7ffffff
    } else {
	my $time_dilation = ($retention - _einterval( $skew, $retention ));
	if ( $retention <= $skew ) {
	    # Unfortunately, in order to set a retention period that is
	    # smaller than the skew, it is necessary to set the default
	    # period to the desired period and then allow the file to
	    # 'acquire' the default period.
	    $def_period_save = $vol->{'options.snaplock_default_period'};
	    $vol->{'options.snaplock_default_period'} = sprintf( "%ds", $retention + $time_dilation );
	    $this->touch( atime => $comp_time ); # After commit, atime will become ctime+period
	} elsif ( $skew > 0 ) { 
	    # If the retention is greater than the skew, we can adjust:
	    $this->touch( atime => ($this->atime - $skew) + $time_dilation );
	} elsif ( $skew < 0 ) {
	    # Negative skew has already been accounted for
	    $this->touch( atime => $this->atime + $time_dilation );
	}
    }
    my $mode = $this->mode;
    unless ( $mode & 0222 ) {
	# If no write bit is set, set one so we can make a transition from on to off.
        eval { $this->chmod( $this->mode | 0002 ) };
    }
    $this->chmod( $this->mode & ~0222 );

    # Restore the default period if we mucked with it above
    if ( $def_period_save ) {
	$vol->{'options.snaplock_default_period'} = $def_period_save;
    }
    return $this;
}


=head3 ::is_worm

Returns true if object is W.O.R.M.

=over

=item Syntax

$obj->is_worm();

=back

=cut

sub is_worm {
    my $this = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $atime = $this->atime;
    my $mtime = $this->mtime;
    # Can't do this if worm
    eval{ $this->touch( mtime => $mtime, atime => ($atime - 60) ) };
    return 1 if ( $@ =~ /read-only/i );
    # Set it back
    eval{ $this->touch( mtime => $mtime, atime => $atime ) };
    return 0;
}

=head3 ::expire

Will not return until file can be deleted

=over

=item Syntax

$obj->expire(%args);

=item Arguments

There are no arguments for this method itself.
Everything is passed to C<retention>.

=back

=cut

sub expire {
    my ($this,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return unless $this->is_worm; # don't wait unless committed worm file
    my $retention = $this->retention;
    if ( $retention > 0 ) {
	NOTE::logcomment(
	    "Waiting for snaplock expiration (in %d secs) of %s",
	    $this->retention, $this->fullpath
	);
	sleep $retention;
	sleep 1 while ( $this->retention ); # Make sure retention is zero
	NOTE::logcomment(
	    "File has expired; proceeding..."
	);
    }
    return $this;
}


# Recomputes the interval, taking the compliance clock's time
# dilation into account. If there is a skew, the compliance clock runs
# slightly fast or slow in order to catch up with the system clock.
# This function converts compliance clock seconds to system clock seconds
# taking this drift into account.
use constant CCLOCK_DRIFT   => ((86400.0*7)/31556926.0); # Rate compliance clock drifts toward system clock

sub _einterval {
    my ($skew, $compliance_secs) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $system_secs = 0;

    # skewtime is the number of system clock seconds that the compliance clock will be drifting.
    # In other words, it's the amount of real time it will take for the compliance clock to
    # catch up with the system clock.
    my $skewtime = POSIX::ceil(abs($skew) / &CCLOCK_DRIFT);

    if ( $skew > 0 ) { # compliance clock runs fast, but only until skew is zero
	$system_secs = POSIX::ceil(($compliance_secs > $skewtime ? $skewtime : $compliance_secs) * (1-&CCLOCK_DRIFT));
    } elsif ( $skew < 0 ) { # compliance clock runs slow, but only until skew is zero
	$system_secs = POSIX::ceil(($compliance_secs > $skewtime ? $skewtime : $compliance_secs) * (1+&CCLOCK_DRIFT));
    }
    if ( ($compliance_secs -= $skewtime) > 0 ) {
	# Any remaining compliance clock seconds are synchronous with the system clock
	$system_secs += $compliance_secs;
    }
    return $system_secs;
}

# Takes an epoch time and adds a period to it,
# where period is defined as what is allowed as a snaplock period.
# Returns an epoch time.
# Note: avoids infinity value unless period is 'infinity'.

sub _add_period {
    my ($t,$period) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $delta;

    return SMALL_INFINITY if ( $period eq 'infinite' );

    my $timestamp = Date::Manip::ParseDateString( "epoch $t" );

    $period .= 's' if ( $period =~ /^(\d+)$/ ); # Default to seconds
    if ( $period =~ /^(\d+)([sdmy])$/ ) {
        my $increment = $1;
	my $unit = $2;
        if ( $unit eq 's' ) {
	    $delta = "+ $increment seconds";
	} elsif ( $unit eq 'd' ) {
	    $delta = "+ $increment days";
	} elsif ( $unit eq 'm' ) {
	    $delta = "+ $increment months";
	} elsif ( $unit eq 'y' ) {
	    $delta = "+ $increment years";
	}
        my $newdate = Date::Manip::DateCalc( $timestamp, $delta );
	my $newtime = Date::Manip::UnixDate( $newdate, "%s" );
	$newtime -= 1 if ( $newtime == SMALL_INFINITY );
	return $newtime;
    } else {
	croak "Invalid period specified: $period";
    }
}

# Can be used to get or set the retention date.
#
# $file->retention( '1y' );  # Extend the retention date by a year
# $file->retention( 'default' ); # Set to the default retention period
# $file->retention( 'minimum' );
# $file->retention( 'maximum' );
# $file->retention( '2m' ); # Extend retention date 2 months
#
# $seconds_to_expire = $file->retention;
#
# Always returns the number of system clock seconds remaining until the file can
# be deleted.
#

=head3 ::retention

Sets and reads retention

 $file->retention( '1y' );  # Extend the retention date by a year
 $file->retention( 'default' ); # Set to the default retention period
 $file->retention( 'minimum' );
 $file->retention( 'maximum' );
 $file->retention( '2m' ); # Extend retention date 2 months

 $seconds_to_expire = $file->retention;

 Always returns the number of system clock seconds remaining until the file can
 be deleted.


=over

=item Syntax

$obj->retention($period);

=item Arguments

B<$period> => I<time>

I<Optional>
Retention time.

=back

=cut

sub retention {
    my ($this,$period) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $vol       = $this->Volume;
    my $comp_time = $vol->Filer->time( clock => 'compliance' );
    my $sys_time  = $vol->Filer->time( clock => 'system' );
    my $skew      = $sys_time - $comp_time;
    my $is_worm   = $this->is_worm;
    my $secs	  = 0;

    if ( defined $period ) {
        $period = NOTE::option( $period );
	if ( $period eq 'default' ) {
	    $period = $vol->{'options.snaplock_default_period'};
	    if ( $period eq 'min' ) {
		$period = $vol->{'options.snaplock_minimum_period'};
	    } elsif ( $period eq 'max' ) {
		$period = $vol->{'options.snaplock_maximum_period'};
	    }
	} elsif ( $period eq 'minimum' ) {
	    $period = $vol->{'options.snaplock_minimum_period'};
	} elsif ( $period eq 'maximum' ) {
	    $period = $vol->{'options.snaplock_maximum_period'};
	}
	if ( $skew < 0 ) {
	    # With negative skew, adjust ctime because it is already
	    # in the past (by skew seconds). Otherwise, retention period
	    # will be too short (by skew seconds).
	    $this->touch( atime => _add_period( $this->ctime+abs($skew), $period ) );
	} else {
	    # Positive skew is dealt with in commit because if the
	    # retention period is smaller than the skew, the 
	    # only way to set it is to allow the file to 'acquire' it
	    # via the filer's default period.
	    $this->touch( atime => _add_period( $this->ctime, $period ) );
	}
    }
    my $atime = $this->atime;
    my $ctime = $this->ctime;
    if ( $atime > $ctime ) { # Retention set
	if ( $atime == BIG_INFINITY ) {
	    $secs = BIG_INFINITY;
	} else {
	    if ( $is_worm ) {
		$secs = _einterval( $skew, ($atime - $comp_time) );
	    } else {
		$secs = ($atime - $ctime);
		$secs -= abs($skew) if ( $skew < 0 );
	    }
	    $secs = $secs < 0 ? 0 : $secs; # already expired
        }
    } else {
	# Retention hasn't been set. Predict what it would be
	# if the file were committed as is.
	my $min = $vol->{'options.snaplock_minimum_period'};
	my $max = $vol->{'options.snaplock_maximum_period'};
	my $def = $vol->{'options.snaplock_default_period'};
	if ( $def eq 'min' ) {
	    $def = $min;
	} elsif ( $def =~ /^(max|infinite)$/ ) {
	    $def = $max;
	}
	croak "$def is incomprehensible."
	    unless ( $def =~ /^(\d+[sdmy]|infinite)$/ );
	my $min_secs = $min eq 'infinite' ?  BIG_INFINITY : (_add_period( $comp_time, $min ) - $comp_time);
	my $max_secs = $max eq 'infinite' ?  BIG_INFINITY : (_add_period( $comp_time, $max ) - $comp_time);
	my $def_secs = $def eq 'infinite' ?  BIG_INFINITY : (_add_period( $comp_time, $def ) - $comp_time);

	$def_secs = $min_secs if ( $def_secs < $min_secs );
	$def_secs = $max_secs if ( $def_secs > $max_secs );
	$secs = $def_secs;
    }
    return $secs;
}

sub SPARE {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_spare_me} = 1;
    $self->{_container}->SPARE;

    # TO DO: implement this:
    # If $self is root, also spare everything in the
    # filesystem, including all future files created in it.
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    return if ( NOTE->ALL_SPARED );
    if ( $self->{_spare_me} ) {
	return;
    }
    # Don't bother unless container is mounted
    return unless ( $self->{_fs}->{_container}->mounted );

    NOTE::logcomment( "Destroying ".$self->path );
    if ( $self->is_dir ) {
	$self->rmdir;
    } elsif ( $self->stat ) {
	if ( $self->is_worm ) {
	    if ( $self->{_fs}->{_container}->snaplock eq 'compliance' ) {
		if ( $self->retention < (15*60) ) {
		    $self->expire;
		} else {
		    NOTE::logcomment( "Long retention time on compliant module!! : ".$self->fullpath );
		}
	    }
	}
	$self->unlink;
    }
    $self->{_destroyed} = 1;
}

1;
