#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX.pm#1 $
#
## @summary  GNOTE GX module
## @author   dl-dpgqa-automation@netapp.com
## @status   Private
## @pod      here
package NOTE::GX;

#use GNOTE::Log;
use GNOTE::GX::Cluster;
use GNOTE::GX::VIF;
use GNOTE::GX::Node;
use GNOTE::GX::Port;
use GNOTE::GX::Disk;
use GNOTE::GX::Volume;
use GNOTE::GX::Qtree;
use GNOTE::GX::Aggregate;
use GNOTE::GX::Vserver;
use GNOTE::GX::Mirror;
use GNOTE::GX::SnapMirror;
use GNOTE::GX::SnapMirrorSet;
use GNOTE::GX::SnapVault;
use GNOTE::GX::SnapVaultSet;
use GNOTE::GX::SnapMirrorUtil;
use GNOTE::GX::Job;
use GNOTE::GX::DatasetTransform;
use Carp qw(croak confess);

use Server;
our @ISA = qw(NOTE);

our $AUTOLOAD;

sub AUTOLOAD {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $type = ref $self;
    my $name = $AUTOLOAD;
    $name =~ s/.*://;    # strip fully-qualified portion

    return if ( $name eq 'DESTROY' );

    if ( $type && $self->can('status') ) {
        my $status = $self->status or croak "Status failed: $type ($name)";

        $name =~ s/_/-/g;    # convention is dashes, not underscores
        unless ( exists $status->{$name} ) {
            NOTE::logcomment("Can't access '$name' field in class $type");
            croak "Can't access '$name' field in class $type";
        }
        return ref $status->{$name} eq 'ARRAY'
          ? @{ $status->{$name} }
          : $status->{$name};
    }
    else {
        $type ||= $self;
        NOTE::logcomment(
            "Can't locate object method \"$name\" via package \"$type\"\n");
        croak "Can't locate object method \"$name\" via package \"$type\"";
    }
}

sub _processError {
    my ( $self, $error ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $str;

    if ($error) {
        if ( $error =~ /Reason: ([^\n]*)/ ) {
            $str = $1;
        }
        else {
            $str = $error;
        }
        if ( $str !~ /no entries matching your query/ ) {
            croak $str;
        }
    }
}

=head2 ::background

Convenience method to transition foreground option to background and make both
available in the interim.

=over

=item Arguments

B<args> => I<HASHREF of parameters>

Hashref on parameters consisting of 'foreground', 'w' and/or 'background'
parameter to translate to use 'background', but sets the 'foreground'
parameter according to 'background' setting for backward compatibilities.

=item Example

    my $background = NOTE::GX::background(\%args);

=cut

sub background {
    my ($args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $bg = 0;
    if ( exists $args->{background} ) {
        $bg = delete $args->{background};
        croak "Invalid 'background' value. Must be 1 or 0"
          unless ( $bg == 0 || $bg == 1 );
    }
    elsif ( exists $args->{foreground} ) {
        NOTE::logcomment("WARNING: Please use 'background' instead of "
                         . "deprecated parameter 'foreground'");
        my $fg = delete $args->{foreground} || 'true';
        croak "Invalid 'foreground' value. Must be 'true' or 'false'..."
              . "Recommended to use 'background' => <1|0>."
          unless $fg =~ /(true|false)/;
        $bg = ( $fg =~ /false/ );
    }
    elsif ( exists $args->{w} ) {
        NOTE::logcomment("WARNING: Please use 'background' instead of "
                         . "deprecated parameter 'w'");
        my $w = delete $args->{w} || 'true';
        croak "Invalid 'foreground' value. Must be 'true' or 'false'..."
              . "Recommended to use 'background' => <1|0>."
          unless $w =~ /(true|false)/;
        $bg = ( $w =~ /false/ );
    }
    $args->{foreground} = ( $bg ? "false" : "true" );
    return $bg;
}

# Returns TRUE if an object *looks* anonymous according to
# its name. This is used by 'rename' methods to determine
# whether a renamed object should be spared or not.
sub _looksAnonymous {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $id;

    if ( $self->isa("NOTE::Aggregate") ) {
        $id = "ag";
    }
    elsif ( $self->isa("NOTE::Volume") ) {
        $id = "vl";
    }
    return ( $self->name =~ /^(_|)$id(.....)_(....)_(........)/i );
}

# Returns TRUE if an object *looks* reusable according to
# its name.
sub looksReusable {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->_looksAnonymous && ( $self->name =~ /^(_|AG)/ );
}

# If string is passed, it is tested for whether it
# looks like something shareable. If string is NOT passed
# the name of the object is tested.
#
# The only currently shareable things are aggregates.
sub looks_shareable {
    my ( $self, $string ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name;

    if ( scalar @_ > 1 ) {
        $name = $string;
    }
    else {    # Test the name of the object
        $name = $self->name;
    }
    return $name =~ /^AG/;
}

# If string is passed, it is tested for whether it
# looks like something this is free. If string is NOT passed
# the name of the object is tested.
#
# GX aggregate and volume names cannot start with an '_', so
# the convention is that they start with 'N_' instead.
sub looks_free {
    my ( $self, $string ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name;

    if ( scalar @_ > 1 ) {
        $name = $string;
    }
    else {    # Test the name of the object
        $name = $self->name;
    }
    return $name =~ /^_(ag|vl)/;
}

# If string is passed, it is returned with an a prefix that makes
# it look like it's free.
# If string is NOT passed the name of the object is returned
# with the free prefix.
#
sub _make_free {
    my ( $self, $string ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name;

    if ( scalar @_ > 1 ) {
        $name = $string;
    }
    else {    # Test the name of the object
        $name = $self->name;
    }
    return '_' . $name;
}

#
# All GX objects *could* have a job associated with them
#
sub job : lvalue {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Must do this trickery due to a bug in perl regarding
    # tied hashes and lvalues.
    my $r = \$self->{_job};
    $$r;
}
1;
