#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Aggregate.pm#1 $
#
## @summary  GNOTE GX Aggregate module
## @author   dl-dpgqa-automation@netapp.com
## @status   Private
## @pod      here
#

=head1 NAME

NOTE::GX::Aggregate

=head1 DESCRIPTION

The purpose of this module is to provide methods for a 10 mode aggregate

When creating a new aggregate, the default number of disk used is 2 for raid4
and 3 for raid_dp. It is possible to specify the exact number of disks to use
from parameters in the call to the 'create' method or via the CLI parameter
DISK_CNT.

=cut

package NOTE::GX::Aggregate;

use strict;
use warnings;
no warnings qw/ redefine /;

use GNOTE::GX;
use GNOTE::Aggregate;
use LogReader::MlogMessages;
use LogReader::Messages;
use Carp qw(croak confess);
use Params::Validate qw(validate ARRAYREF HASHREF OBJECT SCALAR UNDEF);
use Data::Dumper;

our @ISA = qw(NOTE::Aggregate NOTE::GX);

sub _get_node {
    my ($node) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( !defined($node) ) {

        # In this case, there are no arguments (i.e. no hostname) for the
        # filer. I call the filer method anyway and let it decide to
        # handle it or croak.
        return NOTE::GX::Node->connect();
    }
    elsif ( ref $node eq "HASH" ) {
        return NOTE::GX::Node->connect( %{$node} );
    }
    elsif ( ref $node eq "NOTE::GX::Cluster" ) {
        return $node->Node;
    }
    elsif ( ref $node eq "NOTE::GX::Node" ) {
        return $node;
    }
    elsif ( ref \$node eq "SCALAR" ) {

        # Is just a hostspec
        return NOTE::GX::Node->connect( name => $node );
    }
    else {
        croak( "Invalid node specification: " . ref \$node . "\n" );
    }
}

=head2 ::create

Constructor for a GX::Aggregate object. 

=over

=item Arguments

=item Examples

=back

=cut

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;
    $self->{_args}  = \%args;
    my $args = \%args;

    exists $args->{name} and $args->{name} = NOTE::option( $args->{name} );
    exists $args->{snaplock}
      and $args->{snaplock} = NOTE::option( $args->{snaplock} );
    exists $args->{raidtype}
      and $args->{raidtype} = NOTE::option( $args->{raidtype} );
    exists $args->{rgsize}
      and $args->{rgsize} = NOTE::option( $args->{rgsize} );
    exists $args->{disks} and $args->{disks} = NOTE::option( $args->{disks} );
    exists $args->{anonymous}
      and $args->{anonymous} = NOTE::option( $args->{anonymous} );
    exists $args->{dataset}
      and $args->{dataset} = NOTE::option( $args->{dataset} );
    exists $args->{recycle}
      and $args->{recycle} = NOTE::option( $args->{recycle} );
    exists $args->{indirect_type}
      and $args->{indirect_type} = NOTE::option( $args->{indirect_type} );
    exists $args->{fake_large}
      and $args->{fake_large} = NOTE::option( $args->{fake_large} );
    exists $args->{hole_size}
      and $args->{hole_size} = NOTE::option( $args->{hole_size} );

    # Default for fake large aggregate is 0
    $args{fake_large} = 0 if (! defined $args{fake_large});

    # The hole size (in pvbns) has to be a multiple of 4k
    if (defined $args{hole_size}) {
        if ((NOTE::units($args{hole_size})) % 4) {
            confess "Specified hole size ($args{hole_size}) is not a multiple of 4K blocks.";
        }
    }

    # GX only:
    exists $args->{type} and $args->{type} = NOTE::option( $args->{type} );

    # if the $args{Node} is a node object, use it,
    # otherwise try to create one by calling the NOTE::GX::Cluster constructor.
    $args->{Node} ||= $args->{Filer};
    delete $args->{Filer};

    $self->{_Node} = _get_node( $args->{Node} );
    delete $args->{Node};

    if ( defined $args{indirect_type} ) {
       $args{'block-type'} = $args{indirect_type};
    } 

    $self->{_Filer}   = $self->{_Node};
    $self->{_Cluster} = $self->{_Node}->Cluster;

    $self->{_destroyed} = 1;    # Do this because the aggr doesn't physically exist yet.
    bless $self, $class;

    if ($args{fake_large}) {
        #TODO: Find out whether the node is running above BR.0.
        if (!defined $args{indirect_type}) {
            NOTE::logcomment("Setting aggregate indirect_type to 64 to create fake large aggregate.");
            $args{indirect_type} = '64-bit';
        } elsif ($args{indirect_type} ne '64-bit') {
            confess "Fake large aggregate can only be created with block-type 64-bit.\n";
        }
        $args{share} = 0;                            # Do not share fake larger aggrs
    } 

    if ( defined $args->{nodes} ) {
        my $nodestr = "";
        if ( $args->{nodes} eq '*' ) {

            # Pick all healthy nodes from the cluster.
            my @nodes = sort $self->Cluster->nodes(
                health      => 'true',    # Pick only good ones
                eligibility => 'true',
            );
            for my $n ( 1 .. scalar(@nodes) ) {
                $nodestr .=
                  $nodes[ $n - 1 ] . ( $n eq scalar @nodes ? "" : "," ),
                  ;
            }
        }
        elsif ( ref $args->{nodes} eq "ARRAY" ) {
            $nodestr = join( ',', sort @{ $args->{nodes} } );
        }
        elsif ( $args->{nodes} =~ /^\d+$/ ) {

            # Here, only the number of nodes was specified..
            # So pick some nodes from the cluster. Wrap-around
            # if more nodes were asked for than are available.
            my @nodes = sort $self->Cluster->nodes(
                health      => 'true',    # Pick only good ones
                eligibility => 'true',
            );
            for my $n ( 1 .. $args->{nodes} ) {
                $nodestr .= $nodes[ ( ( $n - 1 ) + @nodes ) % @nodes ] . ( $n eq $args->{nodes} ? "" : "," ),;
            }
        }
        else {
            $nodestr = $args->{nodes};
        }
        $args->{nodes} = $nodestr;
    }
    $self->SUPER::create(%args);

    # Must wait for aggregate to come online
    if ($self->state ne 'online') {
        $self->_wait_for_creating_state_to_proceed_to_online_state();
    }

    $self->{_destroyed} = 0;    # Now it exists

    if ($self->state ne 'online') {
        croak(q{Aggregate } . $self->name .
              q{ should be 'online', but instead is '} . $self->state . q{'});
    }

    if ($args{fake_large}) { 
        my $aggr_status = $self->status;
        my $diskcount;                                    # No of disks to add after pvbn hole
        my $name = $self->name; 

        if ( !defined ($args{rgsize}) ) {
             $args{rgsize} = 5;
        } else {
             if ( $args{rgsize} < 4 ) {
                 confess "Raid group size can not be less than 4 for fake large aggr.";
             }
        }

        $self->convert_to_fake_la($args{holesize});
    }
    return $self;
}

sub attach {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;
    $self->{_args}  = \%args;

    if ( exists $args{parent} ) {
        croak unless ref( $args{parent} ) =~ /^NOTE::.*Aggregate/;
        $self->{_Node}    = _get_node( delete $args{Node} );
        $self->{_parent}  = delete $args{parent};
        $self->{_Cluster} = $self->{_parent}->Cluster;
    }
    else {
        if ( ref $args{Filer} eq "NOTE::GX::Cluster" ) {
            $self->{_Cluster} = $args{Filer};
            $self->{_Node}    = $self->{_Cluster}->Node;
        }
        elsif ( ref $args{Filer} eq "NOTE::GX::Node" ) {
            $self->{_Node}    = $args{Filer};
            $self->{_Cluster} = $self->{_Node}->Cluster;
        }
        else {
            $self->{_Node} = _get_node( $args{Node} );
            delete $args{Node};
            $self->{_Cluster} = $self->{_Node}->Cluster;
        }
    }

    $self->{_Filer} = $self->{_Cluster};

    bless $self, $class;
    NOTE::reference($self);

    $self->{_type} = (
        exists $self->{_parent}
        ? $self->{_parent}->{_type}
        : $self->_aggr_status( name => $args{name} )->[0]->{'volume-style'}
    );

    $self->SUPER::attach(%args);

    # Determine node that the aggregate resides on
    my $s = $self->status();
    my $node_list = join ",", @{ $s->{nodes} };
    unless ( $self->{_Node}->name =~ /$node_list/ ) {
        $self->{_Node} = NOTE::GX::Node->connect( name => $s->{nodes}[0] );
    }

    return $self;
}

=head2 ::parent

Returns the parent aggregate of a striped aggregate or otherwise this aggregate.
Croaks if not a striped aggregate.

=over

=item Arguments

B<None>

=item Example

    my $parent = $striped_aggr_member->parent;

=back

=cut

sub parent {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak $self->description . " is not a striped aggregate"
      unless $self->{_type} =~ /striped/;
    return ( exists $self->{_parent} ? $self->{_parent} : $self );
}

=head2 ::options

Sets or gets the options of an aggregate.

=over

=item Examples

Reading example:

    my $crtime = $aggr->options('create-time');

Setting example:

    $aggr->options( raidtype => 'raid4' );

=back

=cut

sub options {
    my ( $self, $which_one, $value ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $options, $ret );
    if ( @_ == 3 ) {

        # Set aggroption $which_one to $value
        $self->server->storage_aggregate_modify(
            aggregate  => $self->name,
            $which_one => $value,
        );
        return 1;
    }
    else {
        $options = $self->status;
        if ($which_one) {
            return $options->{$which_one};
        }
        else {
            return $options;
        }
    }
}

=head2 ::convert_to_fake_la

Converts a regular 64-bit aggregate into fake large aggregate by creating a hole and then adding disks.

WARNING: Please be careful about using fake large aggregates. If the user writes into the area where the hole is, WAFL panics. So it is advisable not to do much IO on the volumes lying on these aggregates. 

=over

=item Arguments

B<holesize> => I<integer>

I<Optional> 

An integer specifying how big the hole after the first set of disks.
Default 20TB(5*1024*1024*1024)

B<diskcount> => I<integer>

I<Optional> 

An integer specifying how many disks to add after creating a hole.
Default For raid_dp it is 5, for raid4 it is 4.

=item Examples

Example:

   $aggr->convert_to_fake_la((4*1024*1024*1024), 3);

=back

=cut

sub convert_to_fake_la {
    my ($self, $holesize, $diskcount) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $aggr_status = $self->status;
    my $name = $self->name;

    eval {
        $self->{_Node}->server->dblade( cmd=>"priv set test;bootargs set bootarg.dblade.fake_large_aggr true" );
        $self->{_Node}->server->dblade( cmd=>"priv set test;bootargs set raidtop_fake_rg_max_count 3" );
    };
    if ($@) {
        confess "Error while setting the bootarg params for fake large aggregate creation.";
    }

    # Determine no of disks to add after the pvbn hole
    if ( $aggr_status->{'raidtype'} eq 'raid_dp' ) {
        $diskcount = 5;
    } else {
        $diskcount = 4;
    }
    # Holesize has to be more than 16TB unless passed.
    if ( !defined $holesize ) {
        $holesize = 5 * 1024 * 1024 * 1024;           # Default holesize
    }
    my $cmd = "priv set test;aggr add $name -s $holesize $diskcount";
    NOTE::logcomment("Adding %d disks after a hole of size %d", $diskcount, $holesize);
    $self->{_Node}->server->dblade( cmd => $cmd );
    $self->{_recycle} = 0;                            # Do not recycle a fake la
}
    
# Returns comma separated spare disks for each comma separated nodes specified

sub _disklist_from_nodes {
    my ( $self, $nodes, $cnt ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my @disk_list;
    foreach my $n ( split ",", $nodes ) {
        my $node = NOTE::GX::Node->connect( name => $n );
        my @disks = $node->disks( state => 'spare' );
        if ( @disks > $cnt ) {
            my $i = 0;
            while ( $i++ < $cnt ) {
                push @disk_list, shift @disks;
            }
        }
        else {
            confess "Node $n only have " . @disks . " spare disks. $cnt is required...";
        }
    }
    return join ",", @disk_list;
}

=head2 ::add_members

Adds member(s)/stripe(s) to this striped aggregate. Parameters are available to specify the node(s) and / or the disks to use
to create new member aggregates.

It is possible to specify either a list of nodes to add members for or a disk spec to be used to create member
aggregates from.

If both nodes and disk spec is specified then this will be used to create the aggregate member(s).

If only nodes are specified then the number of disks to use for each new member will be determined from an existing
member aggregate.

If a disk count is specified then each member aggregate will contain minimal number of disks based on raidtype.

If an array of disk names are specified then member aggregates will be created composing of the disks from that node.

=over

=item Arguments

B<nodes> => I<string>

I<Optional>

Comma separated list of node names to create the member aggregate(s) on (1 per mode), or . Default: Use node from an existing member

B<disk_spec> => I<integer or disklist>

I<Optional> 

An integer specifying how many disks to use to create each member aggregate(s) or a comma separated list of disk names.
Default Same count as an existing member

B<add-vol> => I<boolean>

Indicates whether member volumes should be add to the new aggregate members. Default: 0

B<background> => I<boolean>

Boolean flag to indicate background excution of command. Default: 0

=item Examples

    $aggr->add_members(); # Add a member aggregate to a node in the cluster consisting of mimimal number of disks based on raidtype
    $aggr->add_members(nodes=>"dpg-gx4,dpg-gx5"); # add 2 members to this striped aggregate on nodes dpg-gx4 and dpg-gx5
    $aggr->add_members(disk_spec=>3); # add a 3-disk member aggregate to a node in the cluster
    $aggr->add_members(disk_spec=>"dpg-gx4:0a.26,dpg-gx4:0a.27,dpg-gx4:0a.28"); # Add a aggregate member on node dpg-gx4 using the 3 specified disks

=back

=cut

sub add_members {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    confess "Aggr " . $self->name . ", Type " . $self->{_type} . ": only striped aggregate can add member(s)"
      unless $self->is_stripe;

    confess "Cannot add member to member aggregate " . $self->name
      if exists $self->{_parent};

    my $background = delete $args{background} || 0;

    unless ( exists $args{nodes} ) {

        # Determine node list
        if ( $args{disk_spec} =~ /,/ ) {

            # Determine nodes from disk list
            my %nodes;
            foreach my $d ( split ",", $args{disk_spec} ) {

                # Disk name format: <node>:<disk id> e.g. dpg-gx5:0a.28
                my ( $node, $disk_id ) = split ":", $d;
                $nodes{$node} = 1;
            }
            $args{nodes} = join ",", ( keys %nodes );
        }
        else {

            # Pick a node from the cluster
            ( $args{nodes} ) = $self->Cluster->nodes();
        }
    }

    # Determine disk list from disk spec and node list
    if ( defined $args{disk_spec} ) {

        # Determine disks to use from node(s) list if given a disk count
        $args{disk_spec} = $self->_disklist_from_nodes( $args{nodes}, $args{disk_spec} )
          unless $args{disk_spec} =~ /,/;
    }
    else {

        # No disk spec: Determine disk count from a member aggregate
        #        my ($aggr) = $self->Members;
        #        my $disk_cnt = $aggr->disks;
        my $disk_cnt = $self->_disk_cnt( $self->{_args}->{raidtype} );

        # Determine disk list from node(s) and count
        $args{disk_spec} = $self->_disklist_from_nodes( $args{nodes}, $disk_cnt );
    }

    my $add_vol = ( $args{'add-vol'} ? "true" : "false" );

    NOTE::logcomment( $self->description . ": Add members using" . "\n  Nodes: " . $args{nodes} . "\n   Disks: " . $args{disk_spec} . "\n   Add member volumes " . $add_vol );
    my $text;
    eval {
        $text = $self->server->cli(
            cmd => "storage aggregate add-members -aggregate " . $self->name . " -nodes " . $args{nodes} . " -disklist " . $args{disk_spec} . " -add-member-volumes $add_vol" . " -foreground false" );
    };
    $self->_processError($@);
    $self->{_job} = NOTE::GX::Job->new( Cluster => $self->Cluster )
      unless exists $self->{_job};
    $self->{_job_id} = $self->{_job}->get_id($text);

    $self->job_wait
      unless $background;
}

=head2 ::job_wait

Waits for aggregate job to reach a certain state, with ability to specify a specific column
to return from job status.

=over

=item Arguments

B<state> => I<string>

Expected state to wait for, can be one of success or running. Default: success

B<column> => I<string>

A column to return from job status. Default: state

B<timeout> => I<integer>

Timeout in seconds. Default: 7200 (1 hour in case disk zeroing is needed)

B<interval> => I<integer>

Interval, in secs, between polling job status. Default: 30

=item Examples

     $aggr->job_wait();

=back

=cut

sub job_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    confess "No job id"
      unless exists $self->{_job_id};

    $args{state}    ||= "success";
    $args{column}   ||= "state";
    $args{timeout}  ||= 60 * 60;     # 1 hour in case disks needs zeroing
    $args{interval} ||= 30;

    my $id = delete $self->{_job_id};    # Remove id in case job fails

    my $column;
    eval { $column = $self->{_job}->wait( %args, id => $id ); };
    confess $self->description . " job failed. Error: " . $@
      if $@;

    if ( $args{state} =~ /success/ ) {
        delete $self->{_job_id};
        NOTE::logcomment( $self->description . " Job completed" );
    }
    else {
        $self->{_job_id} = $id;          # Return id if job not completed
    }

    return $column;
}

=head2 ::Members

Returns an ARRAY of NOTE::Aggregate objects representing the members of a striped aggregate.
If node parameter is specified will return member aggregates residing on a certain node.

=over

=item Arguments

B<node> => I<string>

I<Optional>

Return member aggregates residing on a particular node

=item Examples

    my @member_list = $striped_aggr->Members(node=>"dpg-gx4"); # Returns member aggregate on node dpg-gx4
    my @all = $striped_aggr->Members; # Returns all member aggregates for this striped aggregate

=back

=cut

sub Members {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    croak $self->description . " is not a striped aggregate. Type " . $self->{_type}
      unless $self->{_type} =~ /striped/;

    if ( exists $self->{_parent} ) {
        return $self->parent->Members(%args);
    }
    else {
        my $node   = delete $args{node} || ".*";
        my $prefix = $self->name;
        my $status = $self->server->storage_aggregate_member_show();
        my ( @names, @list );
        foreach my $s ( @{$status} ) {
            next
              unless $s->{aggregate} =~ /^$prefix/;    # discard member for other striped aggregates
            if ( $s->{node} =~ /^$node$/i ) {
                push @names, $s->{aggregate};
                push @list,
                  NOTE::GX::Aggregate->attach(
                    tag    => "member of " . $self->description,
                    name   => $s->{aggregate},
                    Node   => $s->{node},
                    parent => $self
                  );
            }
        }
        NOTE::logcomment( "Members for " . $self->description . ":\n    " . ( join ", ", @names ) );
        return @list;
    }
}

=head2 ::status

Returns a HASHREF of the status of this aggregate.

=over

=item Arguements

I<None>

=item Example

     my $status1 = $flex_aggr->status(); 

=back

=cut

sub status {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $status = $self->_aggr_status()->[0];
    croak $@ if $@;

    # Fix this up so common code can use rgsize:
    $status->{rgsize} = $status->{maxraidsize};

    # Server.pm sometimes returns strings if there is
    # one element and arrays if there are multiple elements..
    # This is tedious.. So just make it consistent..
    if ( !ref $status->{nodes} ) {
        $status->{nodes} = [ $status->{nodes} ];
    }
    if ( !ref $status->{plexes} ) {
        $status->{plexes} = [ $status->{plexes} ];
    }
    if ( !ref $status->{raidgroups} ) {
        if ( $status->{raidgroups} eq '-' ) {
            $status->{raidgroups} = [];    # still initializing..
        }
        else {
            $status->{raidgroups} = [ $status->{raidgroups} ];
        }
    }
    if ( !ref $status->{disklist} ) {
        if ( $status->{disklist} eq '-' ) {
            $status->{disklist} = [];      # still initializing..
        }
        else {
            $status->{disklist} = [ $status->{disklist} ];
        }
    }

    # But on the other hand, raidstatus is sometimes
    # an array (while initializing, for example), but it is easier
    # to use as a string.
    if ( ref $status->{raidstatus} eq 'ARRAY' ) {
        $status->{raidstatus} = join( ',', @{ $status->{raidstatus} } );
    }

    return $status;
}

sub Cluster {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->{_Cluster}
}

sub Node {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->{_Node}
}

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->Node->server;
}

sub Nodes {
    my $self  = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @nodes = ();

    for my $node ( $self->nodes ) {
        push @nodes, NOTE::GX::Node->connect( name => $node );
    }
    return @nodes;
}

=head2 ::restricted

Returns a boolean indicating whether an aggregate is restricted

=over

=item Arguments

B<None>

=item Examples

      $aggr->restricted;

=back

=cut

sub restricted {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->status->{'state'} =~ /restricted/i );
}

=head2 ::dblade_node

Returns the name of the node the Aggregate resides on
At present does not support coral aggregates.

=over

=item Arguments

I<None>

=item Examples

    my $dnode = $aggr->dblade_node;

=back

=cut

sub dblade_node {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @nodes   = $self->nodes;
    my $my_node = $nodes[0];
    return $my_node;
}

=head2 ::volumes

Returns array of volume names found in aggregate. Array is simply array of strings.

=cut

sub volumes {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $r;
    eval { $r = $self->server->volume_show( aggregate => $self->name, %args, ) };
    if ($@) {
        croak $@ unless ( $@ =~ /There are no entries matching your query/ );
        return ();
    }
    my @names = ();
    for ( @{$r} ) {
        push @names, $_->{volume};
    }
    return @names;
}

=head2 ::Volumes

Returns array of
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm>
objects found in aggrgate.

=over

=item Arguments

B<spare_only> = 0|1

I<Optional> If set to 1, then it will return list of
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm>
objects of recycled type which have not yet been reused. By default all volumes are retured.

=item Examples

    my @vol = $aggr->Volumes;

=back

=cut

sub Volumes {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $vserver;
    my $spare_only = delete $args{spare_only};
    my @objList    = ();

    NOTE::logcomment( "Obtain list of volumes from " . $self->description );
    if ( exists $args{Vserver} ) {
        $args{vserver} = $args{Vserver}->name;
        delete $args{Vserver};
    }
    $vserver = $args{vserver};

    for my $vol ( $self->volumes(%args) ) {
        next if ( $spare_only && !$self->looks_free($vol) );
        push @objList,
          NOTE::GX::Volume->attach(
            Aggregate => $self,
            name      => $vol,
            Node      => $self->Node,
            Vserver   => $vserver,
          );
    }
    return @objList;
}

=head2 ::scrub_status

Returns a HASHREF of aggregate's scrub status. If a specific column is requested, only that column is returned.

=over

=item Argument

B<column> => I<string>

I<Optional>

Name of column to request.

=item Example

    my $scrub_state = $aggr->scrub_status();
    my $complete = $aggr->scrub_status("complet-percentage"); # Returns the complete-percentage column

=back

=cut

sub scrub_status {
    my ( $self, $column ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $out;
    eval { $out = $self->server->cli( cmd => "aggregate show-scrub-status -aggregate " . $self->name ); };
    confess "Failed to get scrub statue. Error: $out"
      unless $out =~ /storage aggregate show-scrub-status/;

    # Example Output format:
    #              (storage aggregate show-scrub-status)
    #            aggregateX;XraidgroupX;XsuspendedX;Xcomplete-percentageX;Xlast-scrub-timeX;X
    #            AggregateX;XRAID GroupX;XIs SuspendedX;XPercentage CompletedX;XLast Scrub TimeX;X
    #            AG31631_0528_54c1d339X;X/AG31631_0528_54c1d339/plex0/rg0X;XfalseX;X63%X;X-X;X
    my @line   = split /\n/,  $out;
    my @values = split /X;X/, $line[3];
    my $stat;
    foreach $_ ( split /X;X/, $line[1] ) {
        $stat->{$_} = shift @values;
    }
    if ( defined $column ) {
        return $stat->{$column};
    }
    else {
        return $stat;
    }
}

=head2 ::scrub

Invokes the specified scrub action on this aggregate

=over

=item Argument

B<action> => I<start|stop|suspend|resume>

I<Optional>

Indicate which scrub action to invoke. Default: start


=item Examples

    $aggr->scrub(); # Starts an aggregate scrub
    .
    $aggr->scrub("suspend"); # Some time later, suspends it

=back

=cut

sub scrub {
    my ( $self, $action ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $action ||= "start";

    NOTE::logcomment( $self->description . " invoking scrub action: $action" );
    my $output;
    eval { $output = $self->server->cli( cmd => "aggregate scrub -aggregate " . $self->name . " -action $action" ); };
    confess "Failed scrub action: $action. Error: $output"
      unless ( $output =~ /storage aggregate scrub/ );
}

=head2 ::scrub_wait

Wait for aggregate scrub to complete.

=over

=item Arguments

B<timeout> => I<integer>

Number of seconds to wait for completion. The timeout clock is only counted down when there are no progress
between polling period. Default: 300

B<interval> => I<integer>

Wait time in between polling intervals. Default: 60

=item Example

    $aggr->scrub_wait();

=back

=cut

sub scrub_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $timeout  = $args{timeout}  || 300;
    my $interval = $args{interval} || 60;

    NOTE::logcomment( $self->description . ": Wait for aggregate scrub to complete:" );
    my $remain = $timeout;
    my $last_perc;
    do {
        my $curr_perc = $self->scrub_status("complete-percentage");
        if ( $curr_perc =~ /-/ ) {
            NOTE::logcomment( $self->description . ": aggregate scrub completed" );
            return;
        }
        else {
            NOTE::logcomment( $self->description . ": aggregate scrub $curr_perc completed" );
            $remain -= $interval
              if $last_perc eq $curr_perc;
        }
        $last_perc = $curr_perc;
        NOTE::logcomment("$remain sec(s) to timeout 'clock' remaining...");
        Tharn::snooze($interval);
    } while $remain > 0;
    confess "Timeout waiting for aggregate scrub to complete...";
}

=head2 wreck

The wreck command is used to inject persistent corruption into the file system.This is useful for testing WAFL_check and wafliron, the file system consistency checkers.
This command is only available at the priv set test command level, and hence, is not
available in promoted builds.
For more information on usage: http://web.nane.netapp.com/mainN/files/man/man1t/na_wreck.1t.html

=over

=item Arguments

B<opts> = hash table

Hash table of arguments:

=item B<inject_comand> = string

The inject_command argument specifies a single, injection command used to corrupt the aggregate. This is a required parameter and the other hash entries passed to this routine will be used as options for constructing the wreck command.

=item Examples
    $sm->src->Aggregate->wreck(inject_command =>"pseudo_bad", level=>0,inum=>96,bno=>5 );
    $aggr->wreck(inject_command =>"inode_tid_equal",inum=>96,inum2=>57);

=back

=cut

sub wreck {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $cmd;
    my $wreck_string;
    my $output;

    confess ('Inject command not specified for wreck')
        if (not defined $args{inject_command});

    NOTE::logcomment('Name '.$self->name.", inject_command:\n".$args{inject_command});

    # constructing the wreck string with its options
    $wreck_string = delete $args{inject_command};
    while(my($key,$value) = each %args){
        $wreck_string .= " $key=$value";
    }

    # Constructing the cli command
    $cmd = "wreck -A $self->{_name} ";
    $cmd = qq( "priv set test;$cmd\'$wreck_string\'");
    $cmd = 'node run -node '.$self->{_Node}->name.$cmd;

    # Setting the wafl flags before wrecking the Aggreagte to avoid filer panic
    $self->{_Node}->wafl_flag('wafl_inconsistent_panic_testing',1);
    $self->{_Node}->wafl_flag('wafl_iron_panic_on_corruption',0);

    # Issuing the command on the filer
    $output = $self->server->cli(cmd => $cmd);

    confess("Wreck failed to corrupt filesystem: $output")
        if ($output !~ /Finished/s);

    # Bringing the volume back online after the wreck command is successfully issued
    $self->online
        if $self->restricted;

    return $self;
}

=head2 ::Disks

Returns array of object references to disks found in the aggregate.

=over

=item Arguments

B<raid_type> => I<dparity|parity|data|all>

I<Optional>

Specifies what disk raid_type to return. Default: all

=item Example

    my @disks = $aggr->Disks(); # Returns all disks in this aggregate
    my @data_disks = $aggr->Disks(raid_type=>'data'); # Returns all 'data' disks in this aggregate

=back

=cut

sub Disks {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $raid_type = delete $args{raid_type} || "all";

    confess "Invalid raid_type $raid_type"
      unless $raid_type =~ /^(parity|dparity|data|all)$/;

    confess "Invalid raid_type $raid_type for raid_dp aggregate"
      if $raid_type =~ /dparity/ && $self->{_args}->{raidtype} =~ /raid_dp/;

    my @objList = ();
    for my $disk ( $self->disks ) {
        my $disk = NOTE::GX::Disk->attach(
            Aggregate => $self,
            name      => $disk,
        );
        push @objList, $disk
          if ( $disk->status->{raidtype} =~ /^$raid_type$/ || $raid_type =~ /all/ );
    }
    return @objList;
}

=head2 ::hostpath

Returns the name of the aggregate
prepended by its Cluster's name.

=over

=item Arguments

None.

=item Example

=back

     $hp = $aggr->hostpath;

=cut

sub hostpath {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Cluster->name . "(" . $self->Cluster->serialnumber . "):" . $self->{_name};
}

=head2 ::writeable

Returns 1 if aggregate is writable. Aggregate is 
writable if state is 'online' and it's not 'read-only'
and it's not 'restricted'

=cut

sub writeable {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # TODO: figure out how to do this for GX
    return 1;
}

=head2 ::blocks

Returns total nummber of blocks in aggregate.

=cut

sub blocks {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $blocks;

    # TODO: figure out how to do this for GX
    return $blocks;
}

=head2 ::disk_free

Returns values of 'total', 'used', 'avail', 'capacity' as hash for self.
Uses 'aggregate show' to provide the above information.

=over

=item Arguments

I<None>

=item Example

    $aggr_size = $aggr->disk_free()[0];

=back

=cut

sub disk_free {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $s = $self->status;
    return ( $s->{size}, $s->{usedsize}, $s->{availsize}, $s->{'percent-used'} );
}

=head2 ::block_size

Returns number of bytes per block.

=cut

sub block_size {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

=head2 ::fs_size

Returns aggregate size in blocks.

=cut

sub fs_size {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

=head2 ::purge

Delete all volumes within aggregate and then destroy aggregate itself.

=over

=item Arguments

B<vol_only> = 1|0

I<Optional> Delete all volumes within aggregate but do not destroy aggregate itself. Default is 0.

=item Examples

    $aggr->purge( vol_only => 1 );

=back

=cut

sub purge {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $args{vol_only} == 1 ) {
        NOTE::logcomment( "Deleting contents of aggregate " . $self->description );
        for my $vol ( $self->Volumes ) {
            $vol->purge;
        }
    }
    else {
        NOTE::logcomment( "Purging Aggregate " . $self->description );
        for my $vol ( $self->Volumes ) {
            $vol->purge;
        }
        $self->destroy;
    }
}

=head2 ::destroy

Destroy aggregate.

=cut

sub destroy {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Destroying Aggregate " . $self->description );
    my @Nodes = $self->Nodes;
    my $r;
    eval { $r = $self->server->storage_aggregate_delete( aggregate => $self->name ); };
    croak $@ if ($@);

    $self->{_destroyed} = 1;
    $self->resource('destroy');

    for (@Nodes) {
        $_->do_zerospares;    # Zero spares on teardown
    }
    return $self;
}

=head2 ::is_stripe

Return 1 if this is a Striped Aggregate. Otherwise, return 0.

=over

=item Arguments

B<None>

=item Example

    if ($aggr->is_stripe) {
        #
    # the $aggr instance is a stripe aggregate
    #
    }

=back

=cut

sub is_stripe {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $self->{_type} =~ /striped/ ) {
        return 1;
    }

    return 0;
}

=head2 ::is_member

Return 1 if this is a Member Aggregate. Otherwise, return 0.

=over

=item Arguments

B<None>

=item Example

    if ($aggr->is_member) {
        #
    # the $aggr instance is a member aggregate
    #
    }

=back

=cut

sub is_member {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $self->is_stripe ) {
        if ( exists $self->{_parent} ) {
            return 1;
        }
    }

    return 0;
}

# ::_start_logging_message
#
# Start to monitor the messages logged by the Wafliron session on the corresponding node/s of the current Aggregate. This function should be called when a new Wafliron session is initiated to remember the current mark in the message log file. Used together with _stop_logging_message (), and _get_wafliron_result () to eventually extract messages related to the Wafliron session operation in between.
#
#
#
# Examples
#
#    $aggr->_start_logging_message ();
#
#
#
#
#
sub _start_logging_message {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    #
    # Start LogReader for all members if this is a parent Aggregate
    #
    if ( ( $self->is_stripe ) && ( !$self->is_member ) ) {
        my @all_members = $self->Members;

        foreach my $mem_aggr (@all_members) {
            my ( $logrdr, $mlogrdr ) = $mem_aggr->_start_logging_message();

            $self->{_member_log_reader}{ $mem_aggr->name }  = $logrdr;
            $self->{_member_mlog_reader}{ $mem_aggr->name } = $mlogrdr;
        }
    }

    if ( !exists $self->{_log_reader} ) {
        NOTE::logcomment( "(Aggregate::_start_logging_message): Create _log_reader for Aggregate " . $self->description );

        $self->{_log_reader} = LogReader::Messages->new( $self->Node->server );
    }

    if ( !exists $self->{_mlog_reader} ) {
        NOTE::logcomment( "(Aggregate::_start_logging_message): Create _mlog_reader for Aggregate " . $self->description );
        $self->{_mlog_reader} = LogReader::MlogMessages->new( $self->Node->server );
    }

    my $logrdr = $self->{_log_reader};
    NOTE::logcomment( "(Aggregate::_start_logging_message): start log_reader for Aggregate " . $self->description );
    $logrdr->start();

    my $mlogrdr = $self->{_mlog_reader};
    NOTE::logcomment( "(Aggregate::_start_logging_message): start mlog_reader for Aggregate " . $self->description );
    $mlogrdr->start();

    return ( $logrdr, $mlogrdr );
}

# _stop_logging_message
#
# This function is used to mark the message log after the Wafliron session is finished. This shall be used together with _start_logging_message (), and _get_wafliron_result () to eventually extract messages related to the Wafliron session operation in between.
#
#
#
# Examples
#
#    $aggr->_stop_logging_message ();
#
#
#
#
#
sub _stop_logging_message {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( ( $self->is_stripe ) && ( !$self->is_member ) ) {
        my @all_members = $self->Members;

        foreach my $mem_aggr (@all_members) {
            if ( exists $self->{_member_log_reader}{ $mem_aggr->name } ) {
                my $logrdr = $self->{_member_log_reader}{ $mem_aggr->name };
                NOTE::logcomment( "(Aggregate::_stop_logging_message): stop log_reader for Member Aggregate " . $mem_aggr->description );
                $logrdr->restart();

                my $mlogrdr = $self->{_member_mlog_reader}{ $mem_aggr->name };
                NOTE::logcomment( "(Aggregate::_stop_logging_message): stop mlog_reader for Member Aggregate " . $mem_aggr->description );
                $mlogrdr->restart();
            }
            else {
                NOTE::logcomment( "(Aggregate::_stop_logging_message): logrdr not exists for Aggregate " . $mem_aggr->description );
            }
        }
    }

    if ( exists $self->{_log_reader} ) {
        my $logrdr = $self->{_log_reader};
        NOTE::logcomment( "(Aggregate::_stop_logging_message): stop log_reader for Aggregate " . $self->description );
        $logrdr->restart();
    }
    else {
        NOTE::logcomment( "(Aggregate::_stop_logging_message): log_reader not exists for Aggregate " . $self->description );
    }

    if ( exists $self->{_mlog_reader} ) {
        my $mlogrdr = $self->{_mlog_reader};
        NOTE::logcomment( "(Aggregate::_stop_logging_message): stop mlog_reader for Aggregate " . $self->description );
        $mlogrdr->restart();
    }
    else {
        NOTE::logcomment( "(Aggregate::_stop_logging_message): log_reader not exists for Aggregate " . $self->description );
    }
}

# ::_get_wafliron_result
#
# This function returns the result of either an outstanding or the last finished Wafliron session. This shall be used together with _start_logging_message (), and _stop_logging_message () to eventually extract messages related to the Wafliron session operation in between. However, the function does not stop an outstanding wafliron session. It just go to the corresponding log files, and extract the result since the beginning mark of the last invoked wafliron session.
#
#
# Examples
#
#    my $lines = $aggr->_get_wafliron_result ();
#    if (@$lines) {
#        my $result = join "\n", @$lines;
#
#   NOTE::logcomment ("Wafliron result :\n$result");
#    }
#
#
#
#

sub _get_wafliron_result {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @result  = ();
    my $pattern = ".*";

    $self->_stop_logging_message();

    if ( ( $self->is_stripe ) && ( !$self->is_member ) ) {
        my @all_members = $self->Members;

        foreach my $mem_aggr (@all_members) {
            my $logrdr = $self->{_member_log_reader}{ $mem_aggr->name };
            my $lines  = $logrdr->matches($pattern);

            if (@$lines) {
                push @result, @$lines;

                my $found = join "\n", @$lines;
                NOTE::logcomment("(Aggregate::_get_wafliron_result): M1, lines : $found");
                $found = join "\n", @result;
                NOTE::logcomment("(Aggregate::_get_wafliron_result): M2, result : $found");
            }

            my $mlogrdr = $self->{_member_mlog_reader}{ $mem_aggr->name };
            $lines = $mlogrdr->matches($pattern);

            if (@$lines) {
                push @result, @$lines;

                my $found = join "\n", @$lines;
                NOTE::logcomment("(Aggregate::_get_wafliron_result): M3, lines : $found");
                $found = join "\n", @result;
                NOTE::logcomment("(Aggregate::_get_wafliron_result): M4, result : $found");
            }
        }
    }

    if ( exists $self->{_log_reader} ) {
        my $logrdr = $self->{_log_reader};

        my $lines_b = $logrdr->matches($pattern);
        if (@$lines_b) {
            push @result, @$lines_b;

            my $found = join "\n", @$lines_b;
            NOTE::logcomment("(Aggregate::_get_wafliron_result): P1, lines_b : $found");
            $found = join "\n", @result;
            NOTE::logcomment("(Aggregate::_get_wafliron_result): P2, result : $found");
        }
    }

    if ( exists $self->{_mlog_reader} ) {
        my $mlogrdr = $self->{_mlog_reader};

        my $lines_b = $mlogrdr->matches($pattern);
        if (@$lines_b) {
            push @result, @$lines_b;

            my $found = join "\n", @$lines_b;
            NOTE::logcomment("(Aggregate::_get_wafliron_result): P1, lines_b : $found");
            $found = join "\n", @result;
            NOTE::logcomment("(Aggregate::_get_wafliron_result): P2, result : $found");
        }
    }

    my $flag_corruption = $self->_wafliron_is_corrupted( messages => \@result );

    return ( $flag_corruption, \@result );
}

=head2 ::wafliron_start

Start Wafliron on aggregate. By default it runs in the background. Set I<wait> argument to 1 if foreground execution is required.

=over

=item Arguments

B<wait> = 1|0

I<Optional>

Wait for wafliron to finish and exit. By default it starts wafliron as
background process. If I<wait> is set to 1, this function will block and wait
for the wafliron operation to finish, extract the reports from
/mroot/etc/messages and /mroot/etc/mlog/messages.log files from all nodes
exercised by the wafliron operation.


B<force> = 1|0

I<Optional>

Indicate whether this function shall continue to start Wafliron operation even
though if it detect that there is/are outstanding wafliron operation/s running
in the targeted aggregate. Setting force to 1 instruct the function to ignore
any outstanding running wafliron sessions, and continue its attempt to start
the operation.  Default is 0.

B<include_mirror> = 1|0

I<Optional>

Indicate whether the wafliron operation will run on aggregates with existing
mirror relationship. Default is 1, which means that it will include aggregate
with mirror relationship.

=item Return Values

if (wait == 0), return status of the wafliron_start

if (wait == 1), return (corruption_found, result) if wafliron completed;

                corruption_found == 1 means "found corruption"

        result is an array containing the report

=item Examples

    Example 1 :

    (1) Start Wafliron on the aggregate
    (2) Block while waiting for the operation to complete
    (3) Get the result from the return value.

    #
    # Code fragment start here .....
    #

    my ($flag_corruption, $result) = $aggr->walfiron_start (wait=>1);
    if (@$result) {
        my $report = join "\n", @$result;

    NOTE::logcomment ("Wafliron report :\n$report");
    }

    Example 2 :

    (1) Start Wafliron on the aggregate without waiting
    (2) Call wafliron_wait later to wait for the completion
    (3) Get the result from wafliron_wait's return value

    #
    # Code fragment start here .....
    #

    my $status = $vol->walfiron_start (wait=>0);

    #
    # Do something else in between
    #

    my ($flag_corruption, $result) = $aggr->walfiron_wait ();
    if (@$result) {
        my $report = join "\n", @$result;

    NOTE::logcomment ("Wafliron report :\n$report");
    }

    Example 3 :

    (1) Start Wafliron on the aggregate without waiting
    (2) Call wafliron_stop later to stop the operation, possibly at
        a time prior to the natural completion of the operation.
    (3) Get the result from wafliron_stop's return value

    #
    # Code fragment start here .....
    #

    my $status = $aggr->walfiron_start (wait=>0);

    #
    # Do something else in between
    #

    my ($flag_corruption, $result) = $aggr->walfiron_stop ();
    if (@$result) {
        my $report = join "\n", @$result;

    NOTE::logcomment ("Wafliron report :\n$report");
    }

=back

=cut

sub wafliron_start {
    my $self = shift;
    my %args = validate(
        @_,
        {
            force          => { type => SCALAR, optional => 1, regex => qr/0|1/i, default => 1 },
            wait           => { type => SCALAR, optional => 1, regex => qr/0|1/i, default => 1 },
            include_mirror => { type => SCALAR, optional => 1, regex => qr/0|1/i, default => 1 }
        }
    );
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if( Tharn::param('WAFLIRON_OFF') ) {
	$self->{_wafliron_forced_off} = 1;
	NOTE::logcomment("WAFLIRON_OFF is pressent! Won't iron anything!");
	return undef;
    }

    return if Tharn::param('WAFLIRON_OFF');

    my $flag_force          = delete $args{force};
    my $flag_wait           = delete $args{wait};
    my $flag_include_mirror = delete $args{include_mirror};

    #
    # Initialize all_aggregates to include parent aggregate, and all
    # member aggregates if any
    my @all_aggregates = ($self);
    my @all_members    = ();

    if ( ( $self->is_stripe ) && ( !$self->is_member ) ) {
        @all_members = $self->Members;
        push( @all_aggregates, @all_members );
    }

    #
    # If it is not a forced run, the function will check to see if there
    # is/are any outstanding wafliron session/s running. Confess if there exist
    # any outstanding session. If this is a striped aggregate, the function
    # will move to check also the member aggregate for the outstanding sessions.
    #

    if ( !$flag_force ) {
        foreach my $aggr (@all_aggregates) {
            my $status;
            eval { $status = $aggr->wafliron_status(); };
            if ($@) {
                confess $@ unless ( $@ =~ /There are no entries matching your query/ );
            }

            if ( ( defined $status ) && ( $status->{'iron-status'} ne 'not_running' ) ) {
                confess "Wafliron is already running on " . $aggr->description;
            }
        }
    }

    my $val_ignore_member_start_error = "true";
    my $val_include_mirror            = "false";

    if ( !$flag_force ) {
        $val_ignore_member_start_error = "false";
    }

    if ($flag_include_mirror) {
        $val_include_mirror = "true";
    }

    NOTE::logcomment( "(Aggregate::wafliron_start): Starting Wafliron on " . $self->description );

    my $cmd_response;

    eval {
        $self->_start_logging_message();

        $self->server->set( privilege => "diag", confirmations => "off" );

        if ( $self->is_member ) {
            $cmd_response = $self->server->storage_aggregate_wafliron_member_start(
                aggregate         => $self->name,
                'include-mirrors' => $val_include_mirror,
                %args
            );
        }
        else {
            if ( $self->is_stripe ) {
                $cmd_response = $self->server->storage_aggregate_wafliron_start(
                    aggregate                    => $self->name,
                    'ignore-member-start-errors' => $val_ignore_member_start_error,
                    'include-mirrors'            => $val_include_mirror,
                    %args
                );
            }
            else {
                $cmd_response = $self->server->storage_aggregate_wafliron_start(
                    aggregate         => $self->name,
                    'include-mirrors' => $val_include_mirror,
                    %args
                );
            }
        }

        NOTE::logcomment("Response from wafliron start command : $cmd_response");
    };

    if ($@) {
        confess "Error messages while in wafliron command $@";
    }

    if ($flag_wait) {
        return $self->wafliron_wait;
    }
    else {
        return $cmd_response;
    }
}

=head2 ::wafliron_stop

Terminate outstanding Wafliron session on this Aggregate.

=over

=item Arguments

B<wait> = 1|0

I<Optional>

Wait for wafliron to stop and exit. If I<wait> is set to 1, this function will
block and wait for the wafliron stop operation to complete, extract information
from various sources, such as /mroot/etc/messages and /mroot/etc/mlog/messages.log
files,  from all nodes exercised by the wafliron operation. Default is 1.

=item Return Values

if (wait == 0), return status of the wafliron_stop

if (wait == 1), return (corruption_found, result) if wafliron completed;

                corruption_found == 1 means "found corruption"

        result is an array containing the report

=item Examples

    $aggr->wafliron_stop (wait => 0);

=back

=cut

sub wafliron_stop {
    my $self = shift;
    my %args = validate( @_, { wait => { type => SCALAR, optional => 1, regex => qr/0|1/i, default => 1 }, } );
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if( $self->{_wafliron_forced_off} ) {
	NOTE::logcomment("WAFLIRON_OFF is pressent! Nothing to stop");
	return undef;
    }

    my $flag_wait = delete $args{wait};

    my $cmd_response;

    if ( $self->is_member ) {
        $cmd_response = $self->server->storage_aggregate_wafliron_member_stop( aggregate => $self->name, %args );
    }
    else {
        $cmd_response = $self->server->storage_aggregate_wafliron_stop( aggregate => $self->name, %args );
    }

    NOTE::logcomment("Response from wafliron stop command : $cmd_response");

    if ($flag_wait) {
        return $self->wafliron_wait;
    }
    else {
        return $cmd_response;
    }
}

=head2 ::wafliron_status

Return status of the outstanding or previous Wafliron session 

=over

=item Arguments

B<None>

=item Example

    eval {
        $status = $self->wafliron_status ();
    };
    confess $@ if $@;
    if ($status->{'iron-status'} eq 'not_running') {
        #
    # Some code here
    #
    }

=back

=cut

sub wafliron_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if( $self->{_wafliron_forced_off} ) {
	NOTE::logcomment("WAFLIRON_OFF is pressent! Nothing to show");
	return undef;
    }

    my $cmd_response;

    if ( $self->is_member() ) {
        $cmd_response = $self->server->storage_aggregate_wafliron_member_show( aggregate => $self->name, %args );
    }
    else {
        $cmd_response = $self->server->storage_aggregate_wafliron_show( aggregate => $self->name, %args );
    }

    return $cmd_response->[0];
}

=head2 ::wafliron_wait

Wait for the outstanding Wafliron session (if exist) to finish, indicated by the "not_running" status of the session.

=over

=item Arguments

B<timeout> = I<number of seconds>

I<Optional>The elapsed time limit in seconds after which this wait operation will be terminated and return as error  if the corresponding wafliron session is still not finished. Default is 3600 seconds (i.e. 1 hour).

B<interval> = I<number of seconds>

I<Optional>

The number of seconds to sleep between iterations of status check on the
wafliron session. This value should be smaller than the <timeout> value above.
Default is 30 seconds.

=item Return Values

(corruption_found, result) if wafliron completed;

corruption_found == 1 means "found corruption"

result is an array containing the report

=item Examples

    $aggr->wafliron_wait ();
    $aggr->wafliron_wait (timeout => 300);
    $aggr->wafliron_wait (timeout => 600, interval => 30);

=back

=cut

sub wafliron_wait {
    my $self = shift;
    my %args = validate(
        @_,
        {
            timeout  => { type => SCALAR | UNDEF, optional => 1, default => 3600 },
            interval => { type => SCALAR | UNDEF, optional => 1, default => 30 },
        }
    );
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if( $self->{_wafliron_forced_off} ) {
	NOTE::logcomment("WAFLIRON_OFF is pressent! Nothing to wait for");
	return undef;
    }

    my $timeout  = delete $args{timeout};
    my $interval = delete $args{interval};

    my $status;
    my $end_time = time() + $timeout;

    while ( time() < $end_time ) {
        eval { $status = $self->wafliron_status(); };
        if ($@) {
            confess "Error from wafliron_status () : $@";
        }

        if ( $status->{'iron-status'} eq 'not_running' ) {
            return $self->_get_wafliron_result();
        }
        else {
            sleep($interval);
        }
    }
    confess "Timed out waiting for wafliron to complete on aggregate " . $self->description;
}

sub _wafliron_is_corrupted {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $found = 0;

    MESSAGES:
    foreach my $m (@{$args{messages}}) {
        # i suspect the check for /CORRUPT DETECTED/ is a hold over from
        # wafl_check (7mode only) and is no longer needed. - stuw
        if (($m =~ /CORRUPT DETECTED/) ||
            (($m =~ /WAFLIRON/) && ($m =~ /\[wafl[.]check[.]info:error\]/))) {
            $found = 1;
            last MESSAGES;
        }
    }

    return ($found);
}

=head2 ::wafl_check_from_bootmenu

This function invokes the wafl_check on the aggregate from the boot menu. Note
that the wafl_check from bootmenu can only perform the operation on the per
node basis, thus, for Stripe Aggregate, this only apply to member aggregate.

It performs the following steps :

    (1) It tries to bring the system into the IBM (Interactive Boot Menu)
        stage by means of rebooting the filer, and send control-C at the
    specific time.

    (2) It then invokes the wafl_check command, on the aggregate itself. 

    (3) The filer will again reboot, and then run wafl_check when it boot up.

    (4) The function will collect the output from the console during the
        wafl_check process, as well as in between reboot operations of filer,
    until it get into the login prompt again.

    (5) It then move on to try to get the filer back to the CLI stage

    (6) Return the accummulated outputs captured during in step (4) above.

=over

=item Arguments

B<None>

=item Return Values

returns an array of text containing the output captured during the course
of wafl_check, reboot and startup in between, until wafl_check completed.  

=item Examples

    #
    # Code fragment start here .....
    #

    my $result = $aggr->walf_check_from_bootmenu ();
    if (@$result) {
        my $report = join "\n", @$result;

    NOTE::logcomment ("wafl_check report :\n$report");
    }

=back

=cut

#
# Various responses to look for during communication with the FILER
#
use constant {
    RESP_REBOOT     => "to reboot",
    RESP_CONTINUE   => "Continue",
    RESP_WAFL_CHECK => "boot state: wafl-check",
    RESP_LOGIN      => "login:",
    RESP_CHECK      => "Check",
};

#
# Number of seconds to sleep in between some commands so as to let the
# FILER to have some ample time to react
#
use constant DELAY_IN_SECONDS => 15;

#
# Number of seconds to wait for the response from FILER after sending
# the command when using Tharn::say ()
#
use constant CMD_TIMEOUT_IN_SECONDS => 600;

#
# Number of retries to communicate with the FILER in case when it's not
# reacting in the expected sequence.
#
use constant NUM_OF_RETRIES => 5;

sub wafl_check_from_bootmenu {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Collect and store the result, basically capture the outputs
    # from the console, during the course of wafl_check operation.
    my @result;

    #
    # Return values from various Tharn::say calls.
    #
    my ( $str_match, $str_before, $str_after, $str_prompt );

    my $the_console;
    my ( $cmd, $cmd_response );

    my $aggr_name = $self->name();

    if ( ( $self->is_stripe ) && ( !$self->is_member ) ) {
        confess( "Aggregate $aggr_name is a stripe aggregate but not a member aggregate.\n" . "For Stripe Aggregate, wafl_check from the bootmenu can only work on the member aggregate." );
    }

    my $node_name  = $self->Node->name();
    my $the_server = $self->server;

    $the_server->set_execution_server( -Server => $node_name );

    NOTE::logcomment("Trying to transit to IBM (Interactive Boot Menu)...");

    eval { $cmd_response = $the_server->transit_to( state => $the_server->IBM() ); };

    if ($@) {
        confess("ERROR when transit to IBM: $@");
    }

    NOTE::logcomment("After transit to IBM, resp is $cmd_response");

    $the_console = $the_server->get_connection( -Server => $node_name, type => 'console_connection' );

    sleep(DELAY_IN_SECONDS);

    #
    # Note :
    #
    # Date 20090320, By tyu@netapp.com
    #
    # The purpose of the following code block is to communicate with the
    # Filer during the process of trying to run wafl_check from the IBM
    # (Interactive Boot Menu) state. The Filer will reboot in between.
    #
    # A typical communication sequence is as follow :
    #
    #   1. Test Program ==> FILER
    #      - "wafl_check the_aggr_name"
    #
    #   2. FILER ==> Test Program
    #      - Some warning messages, and then ask for "Continue?"
    #
    #   3. Test Program ==> FILER
    #      - answer "yes"
    #
    #   4. FILER ==> Test Program
    #      - The FILER eventually get to the state where it is ready to reboot,
    #        and ask for "Enter to reboot"
    #
    #   5. Test Program ==> FILER
    #      - answer " "
    #      - will get the FILER to reboot, and it will run wafl_check during
    #        startup
    #
    #   6. FILER ==> Test Program
    #      - Eventually will get back to the login prompt, i.e. "login:"
    #
    #   The challenge is that there are occasional situations where the FILER
    #   might not act in the exact sequence as expected. This can be caused
    #   by many different reasons. One reason could be that the answer from
    #   the test program got dropped, e.g. the answer "yes". Thus, a do-while
    #   loop is being used here to give the test program some opportunities to
    #   react to certain unforseeable situations and try its best to complete
    #   the process.
    #
    #   Another possible problem is that the aggregate might be not found by
    #   the time when the FILER tried to perform the wafl_check. In such case,
    #   the filer will prompt the user to see if he/she wants to start
    #   wafl_check on other aggregates subsequently. The issue here is that
    #   the filer is effectively stucked in this mode until the user enter
    #   the answer. Thus, the function cannot "confess" right away, since the
    #   filer will stop in this forever, and subsequent automated tests, e.g.
    #   regression tests, will be unable to proceed until some human
    #   intervention involved. The approach taken in this function to alleviate
    #   the problem is that, when such problem is detected, it will continue to
    #   answer "no" to the "Check aggregate" question/s from the filer until
    #   it get the login prompt. That's where the responses, RESP_CHECK, and
    #   $resp_aggr_not_found, come into the picture.
    #

    my $flag_done   = 0;
    my $num_retries = 0;

    $cmd = "wafl_check " . $aggr_name;

    my $flag_not_found      = 0;
    my $resp_aggr_not_found = "aggregate $aggr_name not found";

    my @possible_responses = ( RESP_REBOOT, RESP_CONTINUE, RESP_WAFL_CHECK, $resp_aggr_not_found );

    do {
        NOTE::logcomment( "Sending command, <$cmd>\nLooking for one of the following responses :\n[" . join( "]\n[", @possible_responses ) . "]" );
        eval { ( $str_match, $str_before, $str_after, $str_prompt ) = say( $the_console, $cmd, '-noprompt', '-wait', \@possible_responses, -timeout, CMD_TIMEOUT_IN_SECONDS ); };

        my $err_msg;

        if ($@) {
            $err_msg = $@;
        }

        NOTE::logcomment("Returns from FILER for command <$cmd> :\nmatch=<$str_match>\nbefore=<$str_before>\nafter=<$str_after>\nprompt=<$str_prompt>\n");

        if ($err_msg) {
            if (   ( $err_msg =~ /Err=TIMEOUT/ )
                && ( $num_retries++ < NUM_OF_RETRIES ) )
            {
                NOTE::logcomment("ERROR when sending <$cmd>: $err_msg");
                NOTE::logcomment("TIMEOUT ERROR: Retry by sending <$cmd> to see what i might get back and hopefully get out of it! This is retry number $num_retries");
            }
            else {
                confess "Failed to communicate with the FILER, $node_name. Error Msg : $@";
            }
        }
        else {

            #
            # Capture results throughout the course of wafl_check command,
            # and the reboot process.
            #
            push( @result, split( "\n", $str_before ) );

            if ( $str_match =~ RESP_CONTINUE ) {
                NOTE::logcomment("I_GOT: CONTINUE, GOING TO SEND <yes>");
                $cmd = "yes";
            }
            elsif ( $str_match =~ RESP_REBOOT ) {
                NOTE::logcomment("I_GOT: TO_REBOOT, GOING TO SEND < >, and LOOKING_FOR <login:>");
                $cmd = " ";
                @possible_responses = ( RESP_REBOOT, RESP_CONTINUE, RESP_WAFL_CHECK, RESP_LOGIN, $resp_aggr_not_found );
            }
            elsif ( $str_match =~ RESP_WAFL_CHECK ) {
                NOTE::logcomment("I_GOT: WAFL_CHECK, GOING TO SEND <yes>");
                $cmd = "yes";
            }
            elsif ( $str_match =~ RESP_LOGIN ) {
                NOTE::logcomment("I_GOT: LOGIN_PROMPT, DONE!");
                $flag_done = 1;
            }
            elsif ( $str_match =~ /$resp_aggr_not_found/ ) {

                #
                # Keep answering no to "Check aggr" questions,  until we get
                # to login prompt
                #
                NOTE::logcomment("I_GOT: AGGR_NOT_FOUND, GOING TO SEND <no> thereafter until it get back login prompt!");
                $cmd                = "no";
                $flag_not_found     = 1;
                @possible_responses = ( RESP_REBOOT, RESP_LOGIN, RESP_CHECK );
            }
            elsif ( $str_match =~ RESP_CHECK ) {
                NOTE::logcomment("I_GOT: CHECK for aggregate to start wafl_check, GOING TO SEND <no> until it get back login prompt!");
                $cmd = "no";
                @possible_responses = ( RESP_REBOOT, RESP_LOGIN, RESP_CHECK );
            }
            else {
                $cmd = " ";
                $num_retries++;
                NOTE::logcomment("I_GOT: <$str_match> which is UNKNOWN and UNEXPECTED, Just send < > to see what i might get back and hopefully get out of it! This is retry number $num_retries");
                sleep(DELAY_IN_SECONDS);
            }
        }
    } while ( ( $flag_done != 1 ) && ( $num_retries < NUM_OF_RETRIES ) );

    if ( $flag_not_found == 1 ) {
        confess "Unable to start wafl_check on aggregate $aggr_name, in FILER $node_name. Reason : aggregate not found!";
    }

    if ( $flag_done != 1 ) {
        confess "Unable to bring FILER, $node_name, back to the login prompt";
    }

    NOTE::logcomment("wafl_check command completed! Trying to transit to CLI stage.");

    eval { $the_server->transit_to( state => $the_server->CLI() ); };

    if ($@) {
        confess "Unable to bring FILER, $node_name, back to the CLI stage. Error: $@";
    }

    NOTE::logcomment("FILER, $node_name, is back to CLI stage.");
    return ( \@result );
}
sub snaplock {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return 0
}    # Not supported in GX

sub _aggr_create {
    my ( $self, $diskspec, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $args{type} ) {
        $args{'volume-style'} = delete $args{type};
    }
    $args{"-Timeout"} = 2 * 60 * 60 unless exists $args{"-Timeout"};    # Default to 2 hours
    my $server = ( $args{local_storage} ? $self->Node->server : $self->server );
    my $ret = eval {
        $server->storage_aggregate_new(
            aggregate => $self->name,

            #disks => $diskspec,
            diskcount     => $diskspec,
            -Verification => undef,                                     # Needed to prevent failure if aggr create
            %args,                                                      # couldn't complete due to disks being
        );                                                              # zeroed
    };
    $self->_processError($@);
    return $ret;
}

sub _aggr_rename {
    my ( $self, $name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Cannot rename member aggregate " . $self->name
      if exists $self->{_parent};

    $self->server->storage_aggregate_rename(
        aggregate => $self->name,
        newname   => $name,
    );
    $self->{_name} = $name;
}

sub _aggr_offline {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Cannot offline member aggregate " . $self->name
      if exists $self->{_parent};

    return eval { $self->server->storage_aggregate_modify( aggregate => $self->name, state => 'offline', @_, ); };
}

sub _aggr_online {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Cannot online member aggregate " . $self->name
      if exists $self->{_parent};

    return eval { $self->server->storage_aggregate_modify( aggregate => $self->name, state => 'online', @_, ); };
}

sub _aggr_restrict {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Cannot restrict member aggregate " . $self->name
      if exists $self->{_parent};

    return eval { $self->server->storage_aggregate_modify( aggregate => $self->name, state => 'restricted', @_, ); };
}

sub _aggr_add {
    my ( $self, $diskspec, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @args;
    if ( $diskspec =~ /^\d+$/ ) {
        push @args, diskcount => $diskspec;
    }
    else {
        push @args, disklist => $diskspec;
    }

    my $curstate = $self->_aggr_status->[0];
    my $maxcount = 5; # wait for 10 minutes ( 5 * 120 seconds ) after zeroing is completed and then fail
    $curstate->{'raidstatus'} = join(',', @{$curstate->{'raidstatus'}} ) if ref($curstate->{'raidstatus'}) eq 'ARRAY';
    while( $curstate->{'raidstatus'} =~ /growing/ && $maxcount ) {
	NOTE::logcomment( "Waiting for disks being added to aggregate" );
	sleep 120;
	$curstate = $self->_aggr_status->[0];
	$curstate->{'raidstatus'} = join(',', @{$curstate->{'raidstatus'}} ) if ref($curstate->{'raidstatus'}) eq 'ARRAY';
	$maxcount = scalar(grep { /\d+/ } map{ $_->{'zeroing-percent'} } @{$self->server->storage_disk_show(aggregate=>$self->name)}) ? 5 : $maxcount--;
    }

    return eval { $self->server->storage_aggregate_add_disks( aggregate => $self->name, @args, %args, ); };
}

sub _aggr_exists {
    my ( $self, $name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $name ||= $self->name;

    return $self->_aggr_status( name => $name );
}

sub _aggr_status {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name = delete $args{name} || $self->{_name};
    if (! $name) {
        croak 'Missing "name" parameter';
    }

    my $r = undef;
    eval {
              $r = ( exists $self->{_parent} ?
                      $self->server->storage_aggregate_member_show(
                                                         aggregate => $name) :
                      $self->server->storage_aggregate_show(
                                                         aggregate => $name) );
              1;
         } or do {
             $self->_processError($@);
         };

    return $r;
}

# check if aggregate is online yet, and if not monitor initializing disks
# until aggregate is ready.
#
# this is supposed to be a private method, but someone has used it in
# GNOTE/GX/San/Lun.pm.  Amazing.  Do not remove this even though perl critic
# complains it is an unused private method.  (BS 01-24-11)
sub _aggr_disks {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $r = undef;

    # check if the aggregate is available first
    # (tried some version of Job waiting on this at first, but it seems there is
    #  no persistent Job associated with aggregate initialization. BS 12/14/10)
    if ($self->state eq 'creating') {
        $self->_wait_for_creating_state_to_proceed_to_online_state();
    }

    $r = $self->_aggr_status();

    # watch for misbehaviors
    if (! defined($r)) {
        confess '_aggr_status result undefined.';
    }
    elsif (! defined($r->[0])) {
        confess 'No elements in _aggr_status result.';
    }
    elsif ((! $r->[0]->{disklist}) || ($r->[0]->{disklist} !~ /\w/)) {
        confess 'Aggregate ' . $self->name . ' disks not defined.';
    }

    return @{$r->[0]->{disklist}};
}

# this method is mostly a wrapper around _wait_for_creation_disk_zeroing() but
# also checks that the aggregate has proceeded to 'online' state from 'creating'
# state in a reasonable amount of time (in case any is needed at all.)
sub _wait_for_creating_state_to_proceed_to_online_state {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ($self->state ne 'online') {
        if ($self->state eq 'creating') {
            $self->_wait_for_creation_disk_zeroing();

            # while is used here in case there is any delay in state proceeding
            # to online once the disks are all zeroed.
            my $max_wait_secs = 30;
            while (($self->state eq 'creating') && ($max_wait_secs > 0)) {
                sleep 1;
                $max_wait_secs--;
            }
        }
    }

    # should be online by now
    if ($self->state ne 'online') {
        croak('Aggregate ' . $self->name . ' state = ' . $self->state .
              ' is invalid.  Expected to be \'online\'.');
    }

    return 1;    # croaks previously if unsuccessful
}


# if disks at creation time must be zeroed before aggregate comes online,
# this method will wait for and monitor such disk zeroing until they are
# all zeroed.
sub _wait_for_creation_disk_zeroing {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $TIMEOUT__NO_PROGRESS = 20 * 60;
    my $DELTA_T_RECHECK = 30;
    my $REPORT_PROGRESS_DELTA = 5;

    my $no_progress_timer = 0;
    my $percent_created = $self->_get_percent_created();
    my $previous_reported_progress = 0;
    while ($percent_created < 100) {
        # report progress occasionally only
        if ((! $previous_reported_progress) ||
                ($percent_created >
                    ($previous_reported_progress + $REPORT_PROGRESS_DELTA))) {
            NOTE::logcomment('Creating aggregate ' . $self->name .
                         ": $percent_created% ready...");
            $previous_reported_progress += $REPORT_PROGRESS_DELTA;
        }

        sleep $DELTA_T_RECHECK;
        my $old_percent_created = $percent_created;
        $percent_created = $self->_get_percent_created();
        if ($percent_created == $old_percent_created) {
            $no_progress_timer += $DELTA_T_RECHECK;

            if ($no_progress_timer > $TIMEOUT__NO_PROGRESS) {
                croak('Aggregate ' . $self->name . ' creation has ' .
                      'timed out due to stalled progress.');
            }
        } else {
            $no_progress_timer = 0;
        }
    }
    return 1;    # croaks inside while loop if unsuccessful
}

# returns the average (rounded down) percent zeroed of all still zeroing disks
# if no disks are still being zeroed, returns 100.
sub _get_percent_created {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $r = undef;
    eval {
             $r = $self->server->storage_disk_show(aggregate => $self->name);
             1;
         } or do {
             $self->_processError($@);
         };

    my $still_zeroing_disk_count = 0;
    my $zeroing_disks_percent_zeroed__total = 0;
    foreach my $rec (@{$r}) {
        if ($rec->{zeroingpercent} =~ /^\d+$/) {
            $still_zeroing_disk_count++;
            $zeroing_disks_percent_zeroed__total += $rec->{zeroingpercent};
        }
    }

    my $zeroing_disks_percent_zeroed = $still_zeroing_disk_count ?
            int($zeroing_disks_percent_zeroed__total /
                $still_zeroing_disk_count)
            : 100;

    return $zeroing_disks_percent_zeroed;
}

=head2 ::upgrade_start

Start a 32 to 64-bit upgrade on the calling aggregate. It does not wait for
the upgrade to complete, that is left to the caller.

The 64bit-upgrade command is a diag level command so the privileges will be
set to diag before issuing the command and then reset to its previous value
before returning.

The method confesses for the following reason,
1) the command cannot be issued, represented by the following error messages
   a) One or more volumes are not large enough
   b) Invalid operation on 64-bit format volume
2) the upgrade is not started

Returns:
SUCCESS: If mode != check then 0 is returned
         If mode == check then the command output is returned as a string
FAILURE: confesses on all failures

If mode == check then an array of hash references is returned. each hash
represents the statistics of the aggregate or one of its volumes.
Example hash,
    $href->{'aggregate'}        = Aggregate name
    $href->{'volume'}           = Volume name, or
                                  if describing an aggregate ""
    $href->{'used'}             = Bytes of volume used
    $href->{'available'}        = Bytes of volume available
    $href->{'capacity'}         = Percent of volume used
    $href->{'additional_space'} = Number of bytes to increase the
                                  volume by so it can be upgraded
    $href->{'fixed_size'}       = 1|0 (TRUE|FALSE)

=over

=item Arguments

I<Optional>

B<mode> => I<check|normal|grow-reserved|grow-all>

B<NOTE: Currently mode = check is not fully supported. The output is returned
as a string. This will be corrected in the future.>

=item Examples

    # start an upgrade on the aggregate
    my $status = $aggr->upgrade_start();
    if ($status == 0) {
        # upgrade started successfully
    } else {
        # upgrade start failed
    }

    # check if the aggregate has sufficient space to perform an upgrade
    my $href = $aggr->upgrade_start( mode => check );
    if (exists($href->{'additional_space'})) {
        # allocate additional space for volume
    }

=back

=cut

sub upgrade_start {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name, $mode, $rtn, $output, $result);
    my (@aggrs);

    $mode = $args{mode};
    $name = $self->{_name};

    $args{aggregate} = $name;
    $args{output} = \$output;

    # start the upgrade
    # for upgrade start Server returns a string instead of an array of
    # hash references.
    #   if mode != check then the string is the started message
    #       Started 64-bit upgrade on aggregate <name>
    #   if mode == check then the string is the upgrade check message
    #       Checking for additional space required to upgrade aggregate <name>..
    eval {
        $rtn = $self->server->storage_aggregate_64bit_upgrade_start( %args );
    };

    # check for errors
    # currently there are two known error messages,
    #   One or more volumes are not large enough
    #   Invalid operation on 64-bit format volume
    if ($@) {
        confess "ERROR: 'storage aggregate 64bit-upgrade start $name' failed: $@";
    } else {
        # check the command output for errors
        if ((!defined($mode)) || ($mode !~ /check/i)) {
            if ($rtn =~ /Start(ed|ing) 64-bit upgrade on aggregate $name/i) {
                $result = 0;
            } else {
                confess("ERROR: unable to start 64-bit upgrade of aggregate $name\n$rtn");
            }
        } else {
# XXX: need to deal with different output from '-mode check' option - stuw
            $result = $rtn;
        }
    }

    return $result;
}

=head2 ::upgrade_scan_start

Restart a 64-bit upgrade that has previously been stopped (via
64bit-upgrade scan-stop) on the calling aggregate. It does not wait for
the upgrade to complete, that is left to the caller.

The 64bit-upgrade command is a diag level command so the privileges will be
set to diag before issuing the command and then reset to its previous value
before returning.

The method confesses for the following reason,
1) the command cannot be issued
2) the upgrade was not restarted

Returns:
SUCCESS: 0 - restarted the upgrade
FAILURE: confesses on all failures

=over

=item Arguments


=item Examples

    # restart an upgrade on the aggregate
    my $result = $aggr->upgrade_scan_start();
    if ($result == 0) {
        # successfully restarted upgrade
    } else {
        # upgrade restart failed
    }

=back

=cut

sub upgrade_scan_start {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name, $server, $rtn, $result);

    $name = $self->{_name};
    $server = $self->server;

    $args{aggregate} = $name;

    # restart the upgrade
    # Server returns a string representing the success or failure of the
    # command.
    #   SUCCESS = Started the 64-bit upgrade scanner on aggregate <name>
    #   FAILURE = anything else, currently no known error messages
    eval { $rtn = $server->storage_aggregate_64bit_upgrade_scan_start(%args); };

    # verify the command succeeded
    if ($@) {
        confess "ERROR: 'storage aggregate 64bit-upgrade scan-start' failed\n$@";
    }

    # check the command output for errors
    if ($rtn =~
        /Start(ed|ing) the 64-bit upgrade scanner on aggregate $name/i) {
        $result = 0;
    } else {
        confess("ERROR: scan-start failed on aggregate $name\n$rtn");
    }

    return $result;
}

=head2 ::upgrade_scan_stop

Stop a 64-bit upgrade that is in progress on the calling aggregate. the
upgrade may be an aggregate or volume upgrade.

The 64bit-upgrade command is a diag level command so the privileges will be
set to diag before issuing the command and then reset to its previous value
before returning.

The method confesses for the following reasons,
1) the command cannot be issued
2) the upgrade could not be stopped

Returns:
SUCCESS: 0 - the upgrade was stopped
FAILURE: confesses on all failures

B<NOTE: The scan-stop command returns success even if no upgrade has been
started on the aggregate or its volumes. Therefore this method will return
success in that situation also.>

=over

=item Arguments


=item Examples

    # stop an upgrade on the aggregate
    my $result = $aggr->upgrade_scan_stop();
    if ($result == 0) {
        # upgrade was successfully stopped
    } else {
        # upgrade could not be stopped
        # maybe already completed?
    }

=back

=cut

sub upgrade_scan_stop {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name, $server, $rtn, $result);
    my $output;

    $name = $self->{_name};
    $server = $self->server;

    $args{aggregate} = $name;
    $args{output} = \$output;

    # stop the upgrade
    # Server returns a string representing the success or failure of the
    # command.
    #   SUCCESS = Stopped the 64-bit upgrade scanner on aggregate <name>
    #   FAILURE = anything else, currently no known error messages
    eval { $rtn = $server->storage_aggregate_64bit_upgrade_scan_stop(%args); };

    # did the command succeed?
    if ($@) {
        confess "ERROR: 'storage aggregate 64bit-upgrade scan-stop' failed\n$@";
    }

    # check the command output for errors
    if ($rtn =~ /Stopp(ed|ing) the 64-bit upgrade scanner on aggregate $name/) {
        $result = 0;
    } else {
        confess("WARNING: Failed to stop upgrade on aggregate $name\n$rtn");
    }

    return $result;
}

=head2 ::upgrade_status

Get the status for a 64-bit upgrade that has previously been started (via
64bit-upgrade start or 64bit-upgrade scan-start) on the calling aggregate.

The 64bit-upgrade command is a diag level command so the privileges will be
set to diag before issuing the command and then reset to its previous value
before returning.

The method confesses for the following reasons,
1) the command cannot be issued
2) the aggregate is not being upgraded, represented by the following error
   messages.
   a) There are no entries matching your query
   b) 64-bit upgrade is not currently in progress
3) no known errors were reported but no status was returned, the array of
   hash references is empty

Returns:
SUCCESS: A reference to an array of hash references is returned.
FAILURE: confesses on all failures

Each hash in the array describes the status for either an aggregate or volume.
The hashes are defined by storage_aggregate_64bit_upgrade_status.

Example hash,
    $href->{'vserver'}               = Vserver name
    $href->{'aggregate'}             = Aggregate name
    $href->{'volume'}                = Volume name, or if describing an
                                       aggregate ""
    $href->{'format'}                = Volume format (32-bit, 64-bit, upgrading)
    $href->{'scanner-status'}        = State of scanner (running, stopped)
    $href->{'scan-percent-complete'} = Percent of scan completed (42%)
    $href->{'scan-time-to-complete'} = Estimated time to complete scan (hh:mm)
    $href->{'scan-progress'}         = String stating progress based on inodes
                                       and fbns (fbn 0, inode 1 of 321, Public)

=over

=item Arguments


=item Examples

    # get the status of an upgrade on the aggregate
    eval { $aref = $aggr->upgrade_status(); };
    if (defined($@)) {
        # status command failed for some reason
        if ($@ =~ /64-bit upgrade is not currently in progress/) {
            # aggregate is not being upgraded
        }
    } else {
        for my $href (@{$aref}) {
            # process the status of each aggregate/volume
        }
    }

=back

=cut

sub upgrade_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name, $server);
    my $output;
    my @rtn;

    $name = $self->{_name};
    $server = $self->server;
    $args{aggregate} = $name;
    $args{output} = \$output;

    # get the status
    # Server returns an array of hash references if any status is returned.
    # known error messages are returned via the output parameter. it is
    # unclear what is returned in $@ on failure.
    eval { @rtn = $server->storage_aggregate_64bit_upgrade_status( %args ); };

    # verify the command did not fail
    if ($@) {
        confess "ERROR: 'storage aggregate 64bit-upgrade status' failed\n$@";
    } elsif (($output =~ /There are no entries matching your query/i) ||
             ($output =~ /64-bit upgrade is not currently in progress/i)) {
        confess("ERROR: no upgrade in progress on aggregate $name\n$output");
    } elsif (scalar(@rtn) > 0) {
        return \@rtn;
    } else {
        confess "ERROR: \'storage aggregate 64bit-upgrade status\' unexpected output\n$output";
    }
}


=head2 ::upgrade_wait

Wait for the aggregate to finish upgrading


Returns:
0  : upgrade completed before timeout
>1 : upgrade did not complete before timeout

=over

=item Arguments

I<Optional>

B<timeout> => I<seconds to wait for upgrade to complete, default = 300>
B<delay> => I<seconds to wait before polling status, default = 5>


=item Examples

    # wait up to 1 hour for the upgrade to complete polling every 5 seconds
    $status = $aggr->upgrade_wait( timeout => 3600 );
    if ($status != 0) {
        confess "Upgrade took too long";
    }

    # wait for 10 minutes for the upgrade to complete polling every 10 seconds
    $status = $aggr->upgrade_wait( timeout => 600, delay => 10 );
    if ($status != 0) {
        confess "Upgrade took too long";
    }

=back

=cut

sub upgrade_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $timeout = $args{timeout} || 300;
    my $delay = $args{delay} || 5;
    my $name = $self->{_name};

    my ($time, $status);

    $time = 0;
    do {
        Tharn::snooze($delay);
        $time += $delay;
        eval { $status = $self->upgrade_status(); };
    } until ((defined($@)) || ($time >= $timeout));

    $status = ((defined($@)) ? 0 : 1);

    return $status;
}


=head2 ::is_32bit

Check if the aggregate block-type (format) is 32-bit and return 1 (TRUE)
if it is


Returns:
1 if the aggregate block-type == '32-bit'
0 otherwise

=over

=item Arguments

=item Examples

    # check if the aggregate is of the correct format
    if ($aggr->is_32bit == 0) {
        confess "aggregate is not 32-bit format";
    }

=back

=cut

sub is_32bit {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->get_block_type();
    if ($fmt =~ /32-bit/i) {
        return 1;
    } else {
        return 0;
    }
}


=head2 ::is_64bit

Check if the aggregate block-type (format) is 64-bit and return 1 (TRUE)
if it is


Returns:
1 if the aggregate block-type == '64-bit'
0 otherwise

=over

=item Arguments

=item Examples

    # check if the aggregate is of the correct format
    if ($aggr->is_64bit == 0) {
        confess "aggregate is not 64-bit format";
    }

=back

=cut

sub is_64bit {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->get_block_type();
    if ($fmt =~ /64-bit/i) {
        return 1;
    } else {
        return 0;
    }
}


=head2 ::is_upgrading

Check if the aggregate block-type (format) is upgrading and return 1 (TRUE)
if it is


Returns:
1 if the volume block-type == 'upgrading'
0 otherwise

=over

=item Arguments

=item Examples

    # check if the aggregate is being upgraded, mixed format
    if ($aggr->is_upgrading == 0) {
        confess "aggregate is not being upgraded";
    }

=back

=cut

sub is_upgrading {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->get_block_type();
    if ($fmt =~ /upgrading/i) {
        return 1;
    } else {
        return 0;
    }
}


=head2 ::get_block_type

Get the block-type of the aggregate


Returns:
The string representing the block-type of the aggregate,
    '32-bit',
    '64-bit',
    'upgrading'

=over

=item Arguments

=item Examples

    # check if this aggregate is being upgraded
    if ($aggr->get_block_type eq 'upgading') {
        # do whatever while the aggregate is being upgraded
    }

=back

=cut

sub get_block_type {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $href = $self->status();
    return $href->{'block-type'};
}


=head2 verify_aggregate_attributes

Verify that the aggregate's attributes match the caller's request. Returns a
value of 0 (zero) if the attributes match, or a non-zero error code. The error
codes are:

    1 - aggregate is offline
    2 - fake_large is not correct
    3 - local_storage is not correct
    4 - the aggregate is on the wrong nodes
    5 - the aggregate uses fewer disks than requested
    6 - dataset is not correct
    7 - snaplock is not correct
    8 - one of the other attributes is not as requested

=over

=item Arguments

=over

=item B<fake_large> = boolean

Aggregate should be a fake large aggregate

=item B<local_storage> = boolean

Aggregate should be on local storage

=item B<nodes> = string

Sorted, comma separated list of nodes the aggregate should be on

=item B<disk> = integer

The number of disks required for the aggregate

=item B<snaplock> = boolean

Aggregate should be a snaplock aggregate

=item B<name> = string

Should match the aggregate name

=item B<indirect_type> = string

Aggregate block-type, one of C<32-bit> or C<64-bit>

=item B<raidtype> = string

Aggregate raid type, one of C<raid_dp> or C<raid4>

=item B<rgsize> = integer

Aggregate raid group size

=item B<type> = string

Aggregate type, one of C<flex> or C<striped>

=back

=item Examples

Check that C<$aggr> is a 64-bit aggregate of 3 disks on local storage:

    my $code = $aggr->verify_aggregate_attributes(
        local_storage => 1,
        disk          => 3,
        indirect_type => '64-bit'
    );
    if ($code) {
        # Remember, any value of $code > 0 is an error!
        ...
    }

=back

=cut

###############################################################################
#
#   Sub Name: verify_aggregate_attributes
#
#   Description:
#       Called with $server->{_args} and diskspec created by
#       GNOTE::Aggregate::create and GNOTE::GX::Aggregate:create to verify
#       the aggregate attributes match the callers request.
#
#       NOTE: this method needs to support both NOTE and GNOTE argument
#       values (the compression group using scripts that run in both).
#       so far this has only been a problem for indirect_type but others
#       may eventually show up.
#
#       Finally, this method exists in both 7G and GX. The GX version is
#       fully functional. The 7G version is a stub. The method used to
#       reside in GNOTE::Aggregate.pm but the differences between the 7G
#       and GX versions of the status() method could not be rectified.
#       More information about the problem can be found in burt 492987.
#
#   Arguments:
#       This is the current list of arguments. Any additional arguments
#       are noted in the log and ignored. If other arguments are needed
#       a burt should be opened to have them added.
#
#               NAME        IN/OUT  TYPE    DESCRIPTION
#               fake_large  in      boolean should be a fake large aggregate
#               local_storage
#                           in      boolean should be on local storage
#               nodes       in      scalar  sorted comma separated list
#               disk        in      scalar  the number of disks required for
#                                           this aggregate
#               snaplock    in      boolean should be a snaplock aggregate
#               name        in      scalar  aggregate name
#               indirect_type
#                           in      scalar  block-type, '32-bit'|'64-bit'
#               raidtype    in      scalar  raid type, 'raid_dp'|'raid4'
#               rgsize      in      scalar  raid group size, integer
#               type        in      scalar  aggregate type, 'flex'|'striped'
#
#   Globals:
#
#   Environment:
#
#   Returns:    Success:    0 - aggregate matches requirements
#               Failure:    1 - aggregate is offline
#                           2 - fake_large is not correct
#                           3 - local_storage is not correct
#                           4 - the aggregate is on the wrong nodes
#                           5 - the aggregate uses fewer disks than requested
#                           6 - dataset is not correct
#                           7 - snaplock is not correct
#                           8 - one of the other attributes is not as requested
#                           9 - indirect_type/block-type do not match
#
###############################################################################
sub verify_aggregate_attributes {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($opts, $rtn, $attr, $msg);

    # get the aggregate attributes
    $opts = $self->status();

    $rtn = 0;

    # check the basics first, aggregate is online and healthy on at least
    # one node
## is this check assuming too much? nodes and state can be arguments. - stuw
    if (($opts->{state} eq 'online') &&
        (scalar( @{$opts->{nodes}} ) > 0)) {
        # check the aggregate attributes match each argument
        PROCESS_ARGUMENTS:
        for my $arg (keys(%args)) {
            if ($arg eq 'fake_large') {
                # check if requested fake large aggregate
                if ( ($args{$arg} == 1) &&
                     ($opts->{'block-type'} ne '64-bit') ) {
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           'aggregate cannot be a fake large aggregate' .
                           'wanted block-type = 64-bit' .
                           "got block-type = $opts->{'block-type'}";
                    NOTE::logcomment($msg);

                    $rtn = 2;
                    last PROCESS_ARGUMENTS;
                }
            } elsif ($arg eq 'local_storage') {
                # if using local storage verify aggregate is on our node
## what does local_storage really mean? should aggregate be only on our node?
## at least on our node? - stuw
                my $node = $self->{_Node}->name;

                if ( ($args{$arg} == 1) &&
                     (grep({ /$node/ } @{$opts->{nodes}}) == 0) ) {
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           'aggregate is not on local node',
                           "\n    wanted node $node",
                           "\n    got node(s) $opts->{filers}",
                    NOTE::logcomment($msg);

                    $rtn = 3;
                    last PROCESS_ARGUMENTS;
                }
            } elsif ($arg eq 'nodes') {
                # if we were given a list of node names make sure it
                # matches the list for the aggregate
                # args{nodes} is already a sorted comma separated list (thanks
                # to GX::Aggregate::create), do the same with the attributes
                my $optsNodes = join(q{,}, sort(@{$opts->{nodes}}));
                if ($optsNodes ne $args{nodes}) {
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           'aggregate is not on local node' .
                           "\n    wanted node(s) $args{nodes}" .
                           "\n    got node(s) $optsNodes";
                    NOTE::logcomment($msg);

                    $rtn = 4;
                    last PROCESS_ARGUMENTS;
                }
            } elsif ($arg eq 'disks') {
                # verify the correct number of disks being used
                if ($args{$arg} > $opts->{diskcount}) {
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           'aggregate does not have enough disks' .
                           "\n    wanted $args{$arg} disks" .
                           "\n    got $opts->{diskcount} disks";
                    NOTE::logcomment($msg);

                    $rtn = 5;
                    last PROCESS_ARGUMENTS;
                }
            } elsif ($arg eq 'datasets') {
## how to handle datasets???? - stuw
                NOTE::logcomment('WARNING: ignoring dataset argument');
            } elsif ($arg eq 'snaplock') {
                if ( $self->snaplock() ne $args{snaplock} ) {
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           'aggregate snaplock state not as expected' .
                           "\n    wanted [$args{snaplock}]" .
                           "\n    got [" . $self->snaplock . ']';
                    NOTE::logcomment($msg);

                    $rtn = 7;
                    last PROCESS_ARGUMENTS;
                }
            } elsif ($arg eq 'indirect_type') {
                # the indirect_type can be the cmode values (32-bit, 64-bit)
                # or the 7g values (32, 64) so cannot do a simple compare
                if ($opts->{'block-type'} !~ /$args{$arg}/i) {
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           'aggregate block-type not as expected' .
                           "\n    wanted [$args{$arg}]" .
                           "\n    got [$opts->{'block-type'}]";
                    NOTE::logcomment($msg);

                    $rtn = 9;
                    last PROCESS_ARGUMENTS;
                }
            } elsif ($attr = $self->_map_create_argument_to_attribute(
                                                            argument => $arg)) {
                # this argument is directly associated with an aggregate
                # attribute so can be compared directly
                #   raidtype/raidtype
                #   rgsize/raidsize|rgsize (which to use???)
                #   type/voltype|volume-style (which to use???)
## maybe convert both to lowercase for comparison? - stuw
                if ($args{$arg} ne $opts->{$attr}) {
                    # reject: expected value (argument) does not match
                    # existing value (attribute)
                    $msg = 'Rejecting ' . $self->name . ': ' .
                           "aggregate attribute $attr not as expected" .
                           "\n    wanted [$args{$arg}]" .
                           "\n    got [$opts->{$attr}]";
                    NOTE::logcomment($msg);

                    $rtn = 8;
                    last PROCESS_ARGUMENTS;
                }
            } else {
                NOTE::logcomment("WARNING: ignoring unexpected argument: $arg");
            }
        }
    } else {
        # reject: aggregate not online or no active nodes
        $msg = 'Rejecting ' . $self->name . ': ' .
               'aggregate not online or has no active nodes' .
               "\n    aggregate state: $opts->{state}" .
               "\n    aggregate nodes:\n".Dumper($opts->{nodes});
        NOTE::logcomment($msg);

        $rtn = 1;
    }

NOTE::logcomment("DEBUG: verify_aggregate_attributes: result: $rtn");
    return $rtn;
}


###############################################################################
#   Sub Name:   _map_create_agrument_to_attribute
#
#   Description:
#       Map the given create aggregate argument to its cooresponding aggregate
#       attribute. If there is no cooresponding attribute return undef.
#
#   Agruments:
#               NAME        IN/OUT  TYPE    DESCRIPTION
#               argument    in      scalar  argument name, see below for
#                                           supported values
#
#       I believe these are the only arguments that coorespond to a specific
#       attribute. If necessary more can be easily added.
#               ARGUMENT        ATTRIBUTE
#               name            aggregate
#               raidtype        raidtype
#               rgsize          rgsize  (_recycle uses raidsize for this)
#               type            volume-style (voltype is also used for this)
#
#   Globals:
#
#   Environment:
#
#   Returns:    Success: the aggregate attribute, as used in the hash returned
#                        by aggregate_show, for the given argument (see above
#                        for mapping)
#               Failure: undef - there is no attribute for the given argument
#
###############################################################################
sub _map_create_argument_to_attribute {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
NOTE::logcomment("DEBUG: _map_create_argument_to_attribute:\n".Dumper(\%args));

    my $arg = $args{argument};

    my %attributes = (
            'name' => 'aggregate',
            'raidtype' => 'raidtype',
            'rgsize' => 'rgsize',             # _recycle uses raidsize - stuw
            'type' => 'volume-style',         # should this use voltype? -stuw
        );

    return ( exists($attributes{$arg}) ) ? $attributes{$arg} : undef;
}

1;
