# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Cluster.pm#1 $
#
## @summary  GNOTE GX Cluster module
## @author   dl-dpg-auto@netapp.com
## @status   public
## @pod      here

=head1 NAME

NOTE::GX::Cluster

=head1 DESCRIPTION

The purpose of this module is to provide 10-mode specific methods for the cluster object

=head1 METHODS

=cut


package NOTE::GX::Cluster;

use strict;
use warnings;
no warnings 'redefine';

use GNOTE;
use GNOTE::GX;
our @ISA = qw(NOTE::GX);
use Server;
use Carp qw( confess );
use Data::Dumper;

my %_CLUSTER_CONN;

my %_NAME_TO_IP;    # Conversion table for cluster name to get Cluster object

=head2 ::connect

Returns an object reference that represents the cluster.

=over

=item Arguments

B<Node> => I<string>

Optional

If present, the connection to the cluster will be made through
the specified node.

=item Example

=back

     $cluster = NOTE::GX::Cluster->connect( Node => q/$FILER:$FILER_SRC/ );

=cut

sub attach {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->connect(@_)
}


sub connect {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    $self->{_tag}      = delete $args{tag};
    $self->{_teardown} = delete $args{teardown};

    eval {
        if ( defined $args{Node} )
        {
            my $node = delete $args{Node};
            if ( ref $node eq 'NOTE::GX::Node' ) {
                $self->{_node} = $node;
            }
            else {
                $self->{_node} = NOTE::GX::Node->connect( name => $node );
            }
        }
        else {    # node list will be obtained from $FILER
            $self->{_node} = NOTE::GX::Node->connect( name =>  NOTE::option( '$FILER' ));
        }
    };
    confess $@ if ($@);

    $self->{_license_policy} = {};
    bless $self, $class;
    my $ip;
    eval { ($ip) = $self->ip_list( role => "cluster_mgmt" ); };    # Ignore failure accessing cluster_mgmt ip as some builds does not have it.
                                                                   # Pick the first one of a sorted list in the case when there are no cluster mgmt ip.
    ($ip) = sort $self->ip_list( role => "cluster" )
      unless $ip;
    if ( exists $_CLUSTER_CONN{$ip} ) {
        return $_CLUSTER_CONN{$ip};
    }
    else {
        NOTE::reference($self);
        $self->{_cluster_mgmt_ip}   = $ip;
        $_CLUSTER_CONN{$ip}         = $self;
        $_NAME_TO_IP{ $self->name } = $ip;
    }

    return $self;
}

=head2 ::Job

Returns the Job object for this Cluster

=over

=item Arguments

I<None>

=item Examples

=back

    my $job_obj = $cluster->Job;

=cut

sub Job {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_job} = NOTE::GX::Job->new( Cluster => $self )
      unless exists $self->{_job};
    return $self->{_job};
}

=head2 license

Gets and/or sets a license for a particular feature. If called with just the
name of a license, returns a boolean value indicating whether the license is
currently installed. If the flag is passed as well, the value of the flag
indicates whether to remove or install the license.

If the flag requests that a license is installed when the license is already
present, or that a license is removed when it is not present, then nothing is
actually done.

The return value from the method is I<original> state of the license, before
any installation or removal (if requested) was done. See the examples, below.

Valid licenses are:

    Base
    CIFS
    FlexClone
    FlexVol_HPO
    iSCSI
    Mirror
    NFS
    SnapMirror_DP
    SnapRestore
    Striped_Volume

NOTE: SnapMirror_LS was a valid license but is not valid as of RR.0(?).
    Because many scripts already try to install the SnapMirror_LS license
    this method will not fail if SnapMirror_LS is passed. Instead it will
    trap the request, log a message that it is being ignored and return
    success to the caller.

NOTE: Should GX_Base, GX_Mirror, GX_FlexVol_HPO, GX_CIFS and GX_SnapRestore
    be valid licenses? That is the way Base, Mirror, FlexVol_HPO, CIFS,
    SnapRestore are listed in real life.

=over

=item Arguments

B<name> = I<string>

I<Required>. A valid feature name.

B<flag> = I<boolean>

I<Optional>. Boolean flag to indicate whether to set or clear the feature given
in 'name'. If passed, this value I<must> be one of C<0> (zero) or C<1> (one).

=item Examples


    # Determine current license setting for SnapMirror_DP license:
    my $current_state = $cluster->license('SnapMirror_DP');

    # Install the license for iSCSI (if not already installed),
    # and returns the old state of the license:
    my $old_state = $cluster->license('iSCSI', 1);

    # A longer example: Set the license for FlexClone, and later
    # turn it off if it was off previously.
    my $status = $cluster->license('FlexClone', 1);
    # Do some tests that involve FlexClone capabilities...
    # If the license was originally off, restore that setting
    if (! $status) {
        $cluster->license('FlexClone', 0);
    }

=back

=cut

sub license {
    my ($self, $name, $flag) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($eval_return, $output, $old_flag);
    my (%clusters, %cluster_licenses, %code, %cluster_licenses_v2, %code_v2, $featurekey);

    # unique cluster to serial number mapping
    # to allow the caller to provide which cluster license they want to
    # install. not fully supported yet..
    %clusters = (
        cluster0 => '1-80-123456',
        cluster1 => '1-80-000011',
        cluster2 => '1-80-000013',
        cluster3 => '1-80-000017',
        vsim0    => '1-80-000008',
    );

    # All the licenses we know about here. This list is kept (more or less)
    # alphabetical so that it is easier to check if a specific license is
    # already listed here. The holes in the list are for licenses that once
    # were valid but no longer are. It is used for backward compatibility
    # with older scripts.
    %cluster_licenses = (
        '1-80-123456' => {                        # soon to be no longer used
            Base           => 'NWUZZJPJYBFDAA',
            CIFS           => 'JAMHCKPJYBFDAA',
            FlexClone      => 'HPHLDKPJYBFDAA',
            FlexVol_HPO    => 'XFOSBKPJYBFDAA',
            iSCSI          => 'RYAEFKPJYBFDAA',
            Mirror         => 'LLQDBKPJYBFDAA',
            NFS            => 'TJFAEKPJYBFDAA',
            SnapMirror_DP  => 'FEDPEKPJYBFDAA',
	    SnapMirror_XDP => 'FEDPEKPJYBFDAA',   # shares license with DP
            SnapMirror_LS  => undef,              # originally RYAEFKPJYBFDAA
            SnapRestore    => 'VUJWCKPJYBFDAA',
            Striped_Volume => 'DTYSFKPJYBFDAA',
        },
        '1-80-000011' => {
            Base           => 'VMZIVZGKWVMADF',
            CIFS           => 'RQQQXZGKWVMADF',
            FCP            => 'XDBRBAHKWVMADF',
            FlexClone      => 'PFMUYZGKWVMADF',
            FlexVol_HPO    => 'FWSBXZGKWVMADF',
            iSCSI          => 'ZOFNAAHKWVMADF',
            Mirror         => 'TBVMWZGKWVMADF',
            NFS            => 'BAKJZZGKWVMADF',
            SnapMirror_DP  => 'NUHYZZGKWVMADF',
	    SnapMirror_XDP => 'NUHYZZGKWVMADF',   # shares license with DP
            SnapMirror_LS  => undef,
            SnapRestore    => 'DLOFYZGKWVMADF',
            Striped_Volume => 'LJDCBAHKWVMADF',
        },
        '1-80-000013' => {
            Base           => 'XBOLYLDPMZSEBG',
            CIFS           => 'TFFTAMDPMZSEBG',
            FCP            => 'ZSPTEMDPMZSEBG',
            FlexClone      => 'RUAXBMDPMZSEBG',
            FlexVol_HPO    => 'HLHEAMDPMZSEBG',
            iSCSI          => 'BEUPDMDPMZSEBG',
            Mirror         => 'VQJPZLDPMZSEBG',
            NFS            => 'DPYLCMDPMZSEBG',
            SnapMirror_DP  => 'PJWADMDPMZSEBG',
	    SnapMirror_XDP => 'PJWADMDPMZSEBG',   # shares license with DP
            SnapMirror_LS  => undef,
            SnapRestore    => 'FADIBMDPMZSEBG',
            Striped_Volume => 'NYREEMDPMZSEBG',
        },
        '1-80-000017' => {
            Base           => 'ZYSOQSYLTDIFMA',
            CIFS           => 'VCKWSSYLTDIFMA',
            FCP            => 'BQUWWSYLTDIFMA',
            FlexClone      => 'TRFAUSYLTDIFMA',
            FlexVol_HPO    => 'JIMHSSYLTDIFMA',
            iSCSI          => 'DBZSVSYLTDIFMA',
            Mirror         => 'XNOSRSYLTDIFMA',
            NFS            => 'FMDPUSYLTDIFMA',
            SnapMirror_DP  => 'RGBEVSYLTDIFMA',
	    SnapMirror_XDP => 'RGBEVSYLTDIFMA',   # shares license with DP
            SnapMirror_LS  => undef,
            SnapRestore    => 'HXHLTSYLTDIFMA',
            Striped_Volume => 'PVWHWSYLTDIFMA',
        },
        '1-80-000008' => {                        # vsim licenses (only 2 nodes)
            Base           => 'JWFJEXMWZWYQSD',
            CIFS           => 'FAXQGXMWZWYQSD',
            FCP            => 'LNHRKXMWZWYQSD',
            FlexClone      => 'DPSUHXMWZWYQSD',
            FlexVol_HPO    => 'TFZBGXMWZWYQSD',
            iSCSI          => 'NYLNJXMWZWYQSD',
            Mirror         => undef,
            NFS            => 'PJQJIXMWZWYQSD',
            SnapMirror_DP  => 'BEOYIXMWZWYQSD',
	    SnapMirror_XDP => 'BEOYIXMWZWYQSD',   # shares license with DP
            SnapMirror_LS  => undef,
            SnapRestore    => 'RUUFHXMWZWYQSD',
            Striped_Volume => 'ZSJCKXMWZWYQSD',
        },
    );

    # v2 licences: http://wikid.netapp.com/w/Licensing/Internal_Licenses#V2.0_Cluster-locked_Licenses
    %cluster_licenses_v2 = (
	'1-80-000011' => {
	    Base	     => 'ETYYFLXUQUMADFAAAAAAAAAAAAAA',
	    CDMI             => 'WSQFRWDUSAFDAAAAAAAAAAAAAAAA',
	    CIFS	     => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA',
	    FCP		     => 'KYSQQWDUSAFDAAAAAAAAAAAAAAAA',
	    FlexClone	     => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA',
	    iSCSI	     => 'YDVBQWDUSAFDAAAAAAAAAAAAAAAA',
	    NFS		     => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapLock	     => 'ERFCUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror       => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror_DP    => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapMirror_XDP   => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapRestore	     => 'INOURWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapVault	     => 'SWHNTWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapManagerSuite => 'QLDRUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapProtectApps  => 'CGBGVWDUSAFDAAAAAAAAAAAAAAAA',
	},

	'1-80-000013' => {
	    Base	     => 'GINBJXTZGYSEBGAAAAAAAAAAAAAA',
	    CDMI             => 'WSQFRWDUSAFDAAAAAAAAAAAAAAAA',
	    CIFS	     => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA',
	    FCP		     => 'KYSQQWDUSAFDAAAAAAAAAAAAAAAA',
	    FlexClone	     => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA',
	    iSCSI	     => 'YDVBQWDUSAFDAAAAAAAAAAAAAAAA',
	    NFS		     => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapLock	     => 'ERFCUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror       => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror_DP    => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapMirror_XDP   => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapRestore	     => 'INOURWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapVault	     => 'SWHNTWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapManagerSuite => 'QLDRUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapProtectApps  => 'CGBGVWDUSAFDAAAAAAAAAAAAAAAA',
	},

	'1-80-000017' => {
	    Base	     => 'IFSEBEPWNCIFMAAAAAAAAAAAAAAA',
	    CDMI             => 'WSQFRWDUSAFDAAAAAAAAAAAAAAAA',
	    CIFS	     => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA',
	    FCP		     => 'KYSQQWDUSAFDAAAAAAAAAAAAAAAA',
	    FlexClone	     => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA',
	    iSCSI	     => 'YDVBQWDUSAFDAAAAAAAAAAAAAAAA',
	    NFS		     => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapLock	     => 'ERFCUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror       => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror_DP    => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapMirror_XDP   => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapRestore	     => 'INOURWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapVault	     => 'SWHNTWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapManagerSuite => 'QLDRUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapProtectApps  => 'CGBGVWDUSAFDAAAAAAAAAAAAAAAA',
	},

	'1-81-000000' => {
	    CDMI             => 'WSQFRWDUSAFDAAAAAAAAAAAAAAAA',
	    CIFS	     => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA',
	    FCP		     => 'KYSQQWDUSAFDAAAAAAAAAAAAAAAA',
	    FlexClone	     => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA',
	    iSCSI	     => 'YDVBQWDUSAFDAAAAAAAAAAAAAAAA',
	    NFS		     => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapLock	     => 'ERFCUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror       => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapMirror_DP    => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapMirror_XDP   => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',	# no such key really, its here for compatibility
	    SnapRestore	     => 'INOURWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapVault	     => 'SWHNTWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapManagerSuite => 'QLDRUWDUSAFDAAAAAAAAAAAAAAAA',
	    SnapProtectApps  => 'CGBGVWDUSAFDAAAAAAAAAAAAAAAA',
	}
    );

    # get the cluster serial number through the Base license
    $featurekey = 'feature';	# see burt
    $eval_return = eval {
        $output = $self->server->system_license_show($featurekey => 'Base');
        1;
    };

    if( (! $eval_return) && $@ =~ /invalid argument/si ) {
	$eval_return = eval {
	    $featurekey = 'package';
	    $output = $self->server->system_license_show($featurekey => 'Base');
	};
    }


    # if there is no Base license installed fail? pick a cluster?
    # install a Base license?
    if (! $eval_return) {
        confess "Cluster has no Base license installed:\n$@";
    }

    NOTE::logcomment('Using license codes associated with serial number: '
                     . $output->[0]->{'serial-number'} );

    %code = %{$cluster_licenses{$output->[0]->{'serial-number'}}};
    %code_v2 = %{$cluster_licenses_v2{$output->[0]->{'serial-number'}}} if defined $cluster_licenses_v2{$output->[0]->{'serial-number'}};

    if ( (! exists($code{$name}) ) && (! exists($code_v2{$name})) ) {
        confess "Unknown License name: '$name'";
    }

    # this license is no longer needed, return that the feature is enabled
    if ((! defined($code{$name})) && (! defined($code_v2{$name})) ) {
        NOTE::logcomment("License no longer required for feature: $name");
        return 1;
    }

    $eval_return = eval {
        $output = $self->server->system_license_show($featurekey => $name);
        1;
    };

    # There is only an error if the eval failed AND $@ does not contain
    # the string 'no entries matching your query'.
    # Snapmirror_LS is a feature that does not require a license. So
    # the "license show" command fails in a different way for Snapmirror_LS
    # saying Snapmirror_LS is an invalid value for "-feature" key.
    if ((! $eval_return) && (index($@, 'an invalid value')== -1 ) &&
        (index($@, 'no entries matching your query') == -1)) {
        confess "Error checking status of license for '$name': $@";
    }

    $old_flag = $output->[0]->{$featurekey} =~ /$name/ ? 1 : 0;

    if (defined $flag && $flag != $old_flag) {
        $eval_return = eval {
            if ($flag)
            {
		NOTE::logcomment("Trying old style license first: $name => $code{$name}");
		eval{
		    $self->server->system_license_add( 'license-code' => $code{$name});
		};
		if( $@ ) {
		    NOTE::logcomment("Old style failed with $@, retrying with: $name => $code_v2{$name}");
		    $self->server->system_license_add( 'license-code' => $code_v2{$name});
		}
            } else {
                $self->server->system_license_delete($featurekey => $name);
            }

            1;
        };
        if (! $eval_return) {
            confess "Failed to change license $name. Error: $@";
        }
    }

    return $old_flag;
}

sub host { return shift->name }

sub identity {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    unless ( exists $self->{_identity} ) {
        $self->{_identity} = shift->server->cluster_identity_show->[0];
    }
    return $self->{_identity};
}

=head2 ::description

Returns a description of this Cluster object

=over

=item Arguments

I<None>

=item Example

=back

    my $title = $cluster->description;

=cut

sub description {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->name . " (Mgmt IP " . $self->cluster_mgmt_ip . ")";
}

=head2 ::name

Gets or sets the name of the cluster

=over

=item Arguments

B<name> - I<string>

Optional

If present, sets the name of the cluster.

Returns the previous name.

=item Example

=back

     $myoldname = $cluster->name( "mycluster" );

=cut

sub name {
    my ( $self, $name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $oldname = $self->identity->{name};
    if ( defined $name ) {
        $self->server->cluster_identity_modify( name => $name );
        $self->{_identity} = $name;
    }
    return $oldname;
}

=head2 ::ip_list

Returns all IPs of a certain type for this cluster. The default list is for
'cluster' and 'intercluster' IPs.

=over

=item Arguments

Arguments are passed by a hash table.

Any options accepted by 'network interface show' cli command is a valid
argument with the option being the hash key and its value being the hash
key value.

I<optional>

=over

=item B<-Timeout> => Seconds to wait for command to complete.
                     Default: 180 seconds

=back

=item Returns

=over

=item B<Success:> an array of ip addresses

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 Method will confess if command I<network interface show> fails
        for any reason

=back

=item Notes

=item Example

=back

    my @list = $cluster->ip_list(role=>'cluster');

    my ($dataip) = $cluster->ip_list(lif=>'dpg-drwho9_data2');

    # get the list of ip's performing the cluster and intercluster roles
    # waiting up to 4 minutes for the command to complete
    @a_ips = $cluster->ip_list( role => 'cluster,intercluster',
                                '-Timeout' => 240 );

=cut

sub ip_list {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($net, @a_ips);

# i have not found any calls to ip_list that rely on role being set to
# its default value. since this is a confusing way to define a default
# and it prohibits callers from getting all the lifs i am removing this
# functionality. in case i missed a reference somewhere i will only
# comment out the code so it can be uncommented if it causes a problem.
# this code has been tested reverting back will be easy. - stuw 2011-02-23
    # because the role command line option restricts what is displayed
    # by other options only set the default value if no other command
    # options are provided. -Timeout is a Server.pm argument not a cli
    # option so it is exempt.
    #if ( !(%args) || (join(q{,}, keys(%args)) eq '-Timeout') ) {
        #$args{role} = 'cluster,intercluster';
    #}

    # on larger clusters this can take a while so allow the caller to set
    # a timeout value for the command (Server.pm). default should be good
    # for a 4 node cluster.
    $args{'-Timeout'} = $args{'-Timeout'} || 180;

    $net = $self->server->network_interface_show( %args );

    if ($ENV{THARN_VERBOSE}) {
        NOTE::logcomment( "ip_list:\n".Dumper( $net ) );
    }

    @a_ips = map{ $_->{address} } @{$net};

    return @a_ips;
}

=head2 :: lifs

Returns an array of LIF (Logical Interface) names in this cluster
Qualifying entries can be used to restrict entries in the list.

=over

=item Arguments

Please see GX::Cluster::lifs methods below for the list
of qualifying parameters.

=item Examples

=back

    my @node_lifs = $cluster->lifs(vserver=>"vs0", role=>"data", link=>"up"); # all active data LIFs for vserver vs0
    my @data_lifs = $cluster->lifs(role=>"data"); # all data LIFs in the cluster

=cut

sub lifs {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @lifnames = ();
    for ( $self->lif_list(%args) ) {
        push @lifnames, $_->{vif};
    }
    return @lifnames;
}

=head2 ::lif_modify

Modifies the up/down status of a lif in the cluster. Lif name must be provided.
In addition either a vserver name or a node name must be provided. But both should not be provided.
Returns a boolean to indicate whether the modify was successful or not

=over

=item Arguments

B<name> => I<string>

I<Required>

A specific named LIF.

B<vserver> => I<vserver name>

I<Optional>

The name of the vserver to which the lif belongs.

B<node> => I<node name>

I<Optional>

The name of the node to which the lif belongs.

B<link> => I<up|down>

I<Required>

Indicates to change the status(admin) of the lif to be up or down.

=item Examples

=back

    $cluster->lif_modify(name=>'vs0_data1',vserver=>$vserver->name,link =>'up');
    $cluster->lif_modify(name=>'clus1', node=>$node->name, link => 'down');

=cut

sub lif_modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess("Must specify lif name") if not exists $args{name};

    confess("Must specify either vserver or node name")
      if not exists $args{vserver} and not exists $args{node};

    confess("Cannot specify both vserver and node name")
      if exists $args{vserver} and exists $args{node};

    confess("Must specify the link status as up or down")
      if not exists $args{link}
          or $args{link} !~ /(up|down)/i;

    $args{lif} = delete $args{name};

    $args{'status-admin'} = delete $args{link};

    confess "server name cannot be directly specified" if exists $args{server};

    # This code causes the command to be invoked on the node from where this
    # method is called
    if ( defined $args{node} ) {
        NOTE::logcomment( 'Setting execution node to : ' . $args{node} );
        $self->{_node} = NOTE::GX::Node->connect( name => $args{node} );
    }

    $args{server} =
      exists $args{vserver}
      ? delete $args{vserver}
      : delete $args{node};

    eval { $self->server->network_interface_modify(%args); };
    if ($@) {
        confess $@ unless ( $@ =~ /There are no entries matching your query/ );
        return 0;
    }
    return 1;
}

=head2 ::lif_list

Returns an array of HASHREFs each containing information for a LIF in this
cluster.  Qualifying parameters can be use to restrict entries in the list.

=over

=item Arguments

B<name> => I<string>

I<Optional>

Query a specific named LIF. Default: all LIFs in the current node

B<vserver> => I<vserver name>

I<Optional>

Restricts the list of LIFs from a specific vserver.

B<node> => I<node name>

I<Optional>

Restricts the list of LIFs from a specific node.

B<role> => I<string>

I<Optional>

One of data, node-mgmt, intercluster or cluster-mgmt. All roles are returned if
not specified.

B<link> => I<up|down>

I<Optional>

Indicates to return LIF that is in a certain state. Default: both up and down

=item Examples

=back

    # Obtain IP address for the first LIF for a vserver
    my @lif_list = $cluster->lif_list(vserver=>$vserver->name);
    my $lif_ip = $lif_list[0]->{address};

=cut

sub lif_list {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list;

    $args{lif} = delete $args{name}
      if exists $args{name};
    $args{server} = delete $args{vserver}
      if exists $args{vserver};
    $args{'curr-node'} = delete $args{node}
      if exists $args{node};
    $args{'status-oper'} = delete $args{link}
      if exists $args{link};

    eval { @list = $self->server->network_interface_show(%args); };
    if ($@) {
        confess $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    return @list;
}

=head2 ::cluster_mgmt_ip

Returns the cluster mgmt ip that represents this cluster.

=over

=item Arguments

B<None>.

=item Examples

=back

    my $ip = $cluster->cluster_mgmt_ip;

=cut

sub cluster_mgmt_ip {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_cluster_mgmt_ip};
}

=head2 ::peer_create

Given a NOTE::Cluster Object, it will create a cluster peer connection with it.
The names of the two clusters involved must be unique.

If it finds a 'cluster peer' entry for the 'dst' cluster, it will verify that the
entry is valid, and provide a valid entry if it is not.

This function can be run on both promoted and non-promoted builds.

=over

=item Arguments

B<cluster> => I<NOTE::Cluster>

A NOTE::Cluster object to create peer connection with.

B<username> => I<string>

I<Optional>

An optional username used to create the connection with. Default: admin

B<password> => I<string>

I<Optional>

Password to used for username.

B<lif> => I<string>

I<Optional>

An optional lif name. Use address associated with cpecific LIF

B<srclif> => I<string>

I<Optional>

An optional lif name. Use address associated with cpecific LIF on source cluster. Default behavior is:
Use B<srclif> first, if not exist then check SRCLIF cli parameter, if cli parameter is not present
then try B<lif> argument. If B<lif> is not present then try to use LIF cli parameter. If cli LIF is
not present then use filer's default.

B<dstlif> => I<string>

I<Optional>

An optional lif name. Use address associated with cpecific LIF on destination cluster Default behavior is:
Use B<dstlif> first, if not exist then check DSTLIF cli parameter, if cli parameter is not present
then try B<lif> argument. If B<lif> is not present then try to use LIF cli parameter. If cli LIF is
not present then use filer's default.

=item Example

=back

    $cluster->peer_create(cluster=>$another_cluster_obj);   # create using cluster interfaces

    $cluster->peer_create(lif=>'mgt1', cluster=>$another_cluster_obj);  # create using management interfece mtg1 on both ends

    $cluster->peer_create(srclif=>'mgt1', dstlif=>'mgt2', cluster=>$another_cluster_obj);   # force specific LIFs on both sides

=cut

sub peer_create {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing cluster parameter."
      unless exists $args{cluster};

    my $dst_cluster = delete $args{cluster};
    my $dst_name    = $dst_cluster->name;

    return
      if exists $self->{_peer}{$dst_name}
          || $self->cluster_mgmt_ip eq $dst_cluster->cluster_mgmt_ip;

    confess "Unique cluster name required."
      if $self->name eq $dst_name;

    my $good_src = $self->_good_peer_entry($dst_cluster);
    my $good_dst = $dst_cluster->_good_peer_entry($self);
    if ( $good_src && $good_dst ) {
        NOTE::logcomment( "Found existing cluster peer entries for clusters " . $self->name . " and " . $dst_name );
    }
    else {
        NOTE::logcomment( "No good peer found\n");
        my %cli = map{ $_, Tharn::param($_, '-default', undef ) } qw( LIF SRCLIF DSTLIF );      # pull some defaults from CLI/.ntest
        my (%my_args, %dst_args);                                   # place holder for ip_list arguments

        delete $my_args{lif} unless $my_args{lif} = ( $args{srclif} || $cli{SRCLIF} || $args{lif} || $cli{LIF} );
        delete $dst_args{lif} unless $dst_args{lif} = ( $args{dstlif} || $cli{DSTLIF} || $args{lif} || $cli{LIF} );

        $my_args{role} = 'intercluster' unless $my_args{lif};                   # use cluster/intercluster role by default
        $dst_args{role} = 'intercluster' unless $dst_args{lif};                 # unless overtaken by LIFs parameters

        my ($my_ip) = $self->ip_list( %my_args );
        my ($dst_ip) = sort $dst_cluster->ip_list( %dst_args );


        my $msg = $self->name . " (IP $my_ip) to $dst_name (IP $dst_ip)";
        NOTE::logcomment("Attempting cluster peer create for $msg");

        $args{username} ||= "admin";
        $args{password} ||= $self->Node->default_password();
        $args{password} = 'netapp1!' unless( $args{password} =~ /\w+/ );

        $args{"-Verification"} = undef;

        $self->Node->exec( level        =>'admin',
                           cmd          =>'cluster_peer_create',
                           'peer-addrs' => $dst_ip,
                           %args );

        NOTE::logcomment("Cluster peer create successful from $msg") if $self->peer_entry( cluster => $dst_cluster );

        $self->{_peer}{$dst_name} = 1;
        $dst_cluster->{_peer}{ $self->name } = 1;    # Work-around to BURT 336067

    }
}

# Returns boolean indicating existance of cluster peer entry

sub _good_peer_entry {
    my ( $self, $dst_cluster ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    eval { $self->peer_entry( cluster => $dst_cluster ); };
    if ($@) {
        NOTE::logcomment("cluster peer show error: $@");
        eval {

            # Remove invalid or unavailable peer entry
            $self->peer_delete( name => $dst_cluster->name )
              unless $@ =~ /no entry/i;
        };
        return 0;
    }
    else {
        return 1;
    }
}

=head2 ::peer_delete

Removes the cluster peer connection entries from the clusters, effectively disconnecting
the clusters from each other.

This function can be run on both promoted and non-promoted builds.

=over

=item Arguments

B<name> => I<string>

Cluster name of to remove cluster peer connection.

=item Example

=back

    $cluster->peer_delete(name=>"other_cluster");

=cut

sub peer_delete {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing name parameter."
      unless exists $args{name};

    my $out = $self->Node->exec( level=>'admin',
                                 cmd=>'cluster_peer_delete',
                                 cluster => $args{name}
                               );
}

=head2 ::peer_entry

Returns the 'cluster peer' entry for the given cluster

This function can be run on both promoted and non-promoted builds.

=over

=item Arguments

B<cluster> => I<NOTE::GX::Cluster>

A NOTE::GX::Cluster object to return the 'cluster peer' entry for.

=item Examples

=back

    my $status = $cluster->peer_entry(cluster=>$dst_cluster);

=cut

sub peer_entry {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing 'cluster' parameter" unless exists $args{cluster};

    my $dst_cluster = delete $args{cluster};
    my $dst_name    = $dst_cluster->name;

    my ($my_ip) = $self->ip_list( role => "intercluster" );

    my $msg = $self->name . " (IP $my_ip) to $dst_name ";

    my $entry = ( $self->Node->exec( level=>'admin',
                                     cmd=>'cluster_peer_show',
                                     cluster => $dst_name ) )->[0];
    NOTE::logcomment("Peer availability:" . Dumper($entry));

    if ( $entry->{availability} =~ /^available$/i ) {
        $entry->{'peer-addrs'} = shift @{$entry->{'peer-addrs'}} if( ref( $entry->{'peer-addrs'} ) eq 'ARRAY' );

        if( $entry->{'peer-addrs'} && $ENV{BURT395211} ) {              # following code is work around BURT 395211
            NOTE::logcomment('Returning existing peer without IP check. BURT395211');   # it should be removed as soon as BURT is fixed
            return $entry;                              # $ENV{BURT395211} is same as Tharn::param('BURT395211'...)
        }                                           # changes have been made by cpeter (dl-dpg-auto)

        $msg .= "(IP " . $entry->{'stable-addrs'} . ")";
        my @dst_ip = sort $dst_cluster->ip_list( role => "intercluster" );

        return @dst_ip if( @dst_ip );                           # return if peer is found ( also see BURT 422997 )
#
# BURT 422997 - UI change. There is no 'stable-addrs' anymore. Leaving original code
# within for testing period.g
#
#        foreach (@dst_ip) {
#            if ( $entry->{'stable-addrs'} =~ /$_/ ) {
#                NOTE::logcomment( "Found cluster peer entry: " . $msg );
#                return $entry;
#            }
#            else {
#                NOTE::logcomment( "IP $_ not in stable-addrs: " . $entry->{'stable-addrs'} );
#            }
#        }

        confess "Found invalid cluster peer entry: " . $msg . ". Valid dst IP addresses: " . ( join ", " . @dst_ip );
    }
    else {
        confess $msg. "(" . $entry->{availability} . ")";
    }
}

=head2 ::snapmirror_status

Returns a pointer to an array of hashes representing the statuses of all
all snapmirror relationships in the cluster.

=over

=item Arguments

B<None>

=item Example

=back

    my $statuses = $cluster_obj->snapmirror_list ();
    my $id_idle = $statuses->[0]->{status} eq "idle";

=cut

sub snapmirror_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{"-Timeout"} = 600
      unless exists $args{"-Timeout"};

    my $result = $self->Node->snapmirror_status(%args);
    return $result;
}

=head2 ::SnapMirrorSets

Returns array of existing SnapMirrorSet objects to represent the existing snapmirror
relationsips on this cluster.

For each src and dst volumes found in a snapmirror relationship, this method composes a
NOTE::GX::Volume objects out of them, and sets up the appropriate parameters so that it can
calling the SnapMirrorSet::attach method to create the SnapMirrorSet objects to be returned
by this method.

NOTE: Use SnapMirrorSet() method to create set in a first place.

=over

=item Arguments

B<src> => I<src spec>

I<None>

A NOTE::GX::Volume object or full specification of the volume name consisting of <cluster>://<vserver>/<volume>.
If not specified than all snapmirror relationships will be returned.

=item Example

=back

     my @sms = $cluster->SnapMirrorSet();

=cut

sub SnapMirrorSets {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $source_param;
    $source_param = "-source-path " . ( ref( $args{src} ) =~ /^NOTE::.*Volume/ ? $args{src}->hostpath : $args{src} )
      if exists $args{src};

# Sample of snapmirror show output:
# snapmirror show LIB_CLUSTER://vs0/vl20126_bake_ballroom -fields source-path,source-cluster,source-vserver,source-volume,type,destination-path,destination-cluster,destination-vserver,destination-volume\r\n
# source-pathX;Xsource-clusterX;Xsource-vserverX;Xsource-volumeX;Xdestination-pathX;Xdestination-clusterX;Xdestination-vserverX;Xdestination-volumeX;XtypeX;X\r\n
# LIB_CLUSTER://vs0/vl20126_bake_ballroomX;XLIB_CLUSTERX;Xvs0X;Xvl20126_bake_ballroomX;XLIB_CLUSTER://vs0/vl20126_core_subslotsX;XLIB_CLUSTERX;Xvs0X;Xvl20126_core_subslotsX;XDPX;X\r\n
# LIB_CLUSTER://vs0/vl20126_bake_ballroomX;XLIB_CLUSTERX;Xvs0X;Xvl20126_bake_ballroomX;XLIB_CLUSTER://vs0/vl20126_Hume_electorsX;XLIB_CLUSTERX;Xvs0X;Xvl20126_Hume_electorsX;XLSX;X\r\n
# LIB_CLUSTER://vs0/vl20126_bake_ballroomX;XLIB_CLUSTERX;Xvs0X;Xvl20126_bake_ballroomX;XLIB_CLUSTER://vs0/vl20126_Dyke_RussiansX;XLIB_CLUSTERX;Xvs0X;Xvl20126_Dyke_RussiansX;XLSX;X\r\n
# LIB_CLUSTER://vs0/vl20126_tone_broilingX;XLIB_CLUSTERX;Xvs0X;Xvl20126_tone_broilingX;XLIB_CLUSTER://vs0/vl20126_Polk_leathersX;XLIB_CLUSTERX;Xvs0X;Xvl20126_Polk_leathersX;XDPX;X\r\n
# LIB_CLUSTER://vs0/vl20126_tone_broilingX;XLIB_CLUSTERX;Xvs0X;Xvl20126_tone_broilingX;XLIB_CLUSTER://vs0/vl20126_huts_portendsX;XLIB_CLUSTERX;Xvs0X;Xvl20126_huts_portendsX;XDPX;X\r\n
# LIB_CLUSTER://vs0/vl20126_tone_broilingX;XLIB_CLUSTERX;Xvs0X;Xvl20126_tone_broilingX;XLIB_CLUSTER://vs0/vl20126_zoom_escapeesX;XLIB_CLUSTERX;Xvs0X;Xvl20126_zoom_escapeesX;XLSX;X\r\n
# \r\n

    $self->server->cli( cmd => "set -showallfields false" );
    my $text =
      $self->server->cli( cmd => "snapmirror show $source_param -fields "
          . "source-path,source-cluster,source-vserver,source-volume," . "type,"
          . "destination-path,destination-cluster,destination-vserver,destination-volume" );
    $self->server->cli( cmd => "set -showallfields true" );
    my @lines    = split /\n/,  $text;
    my @key_list = split /X;X/, shift @lines;
    shift @lines;

    my ( @sms, %src );

    # Build the volume objects
    while ( my $line = shift @lines ) {
        my @v = split /X;X/, $line;
        my %status;
        foreach my $k (@key_list) {
            $status{$k} = shift @v;
        }

        my $src_cluster = $status{"source-cluster"};
        my $src_path    = $status{"source-path"};
        my $src_vserver = $status{"source-vserver"};
        my $src_volume  = $status{"source-volume"};
        $src{$src_path}{src} = $self->_attach_volume( $src_volume, $src_vserver, $src_cluster )
          unless exists $src{$src_path};

        my $dst_cluster = $status{"destination-cluster"};
        my $dst_path    = $status{"destination-path"};
        my $dst_vserver = $status{"destination-vserver"};
        my $dst_volume  = $status{"destination-volume"};
        my $dst         = $self->_attach_volume( $dst_volume, $dst_vserver, $dst_cluster );

        my $type = lc( $status{"type"} );
        push @{ $src{$src_path}{$type} }, $dst;
    }

    # Build SnapMirrorSet objects from volume list
    foreach my $src_path ( keys %src ) {
        push @sms, NOTE::SnapMirrorSet->attach( %{ $src{$src_path} } );
    }

    return @sms;
}

sub _attach_volume {
    my ( $self, $volume, $vserver, $cluster ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $vol = NOTE::GX::Volume->attach( name => $volume, vserver => $vserver, Node => $self->_node_from_cluster($cluster) );
    return $vol;
}

# Obtain a Node from a cluster name

sub _node_from_cluster {
    my ( $self, $cluster ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $cluster_obj = $_CLUSTER_CONN{ $_NAME_TO_IP{$cluster} };
    if ($cluster_obj) {
        return $cluster_obj->Node;
    }
    else {
        confess "Unable to get to Cluster $cluster";
    }
}

=head2 ::serialnumber

Returns the serial number of the cluster.

=over

=item Arguments

None.

=item Example

=back

     print "Cluster serial number is ", $cluster->serialnumber, "\n";

=cut

sub serialnumber {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->identity->{serialnumber}
}

=head2 ::uuid

Returns the uuid of the cluster.

=over

=item Arguments

None.

=item Example

=back

     print "Cluster uuid is ", $cluster->uuid, "\n";

=cut

sub uuid {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->identity->{uuid}
}

=head2 ::vservers

Returns the list of vservers that make up this cluster.

=over

=item Arguments

None.

=item Example

=back

    for my $vserver ( $cluster->vservers ) {
        print "Vserver: $vserver\n";
    }

=cut

sub vservers {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @names = ();

    for ( @{ $self->server->vserver_show(%args) } ) {
        push @names, $_->{vserver};
    }
    return @names;
}

sub Vservers {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @vservers = ();

    for ( $self->vservers ) {
        push @vservers, NOTE::GX::Vserver->connect( name => $_ );
    }
    return @vservers;
}

=head2 ::nodes

Returns the list of nodes that make up this cluster.

=over

=item Arguments

None.

=item Example

=back

    for my $node ( $cluster->nodes ) {
        print "Node: $node\n";
    }

=cut

sub nodes {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @names = ();

    for ( @{ $self->server->cluster_show(%args) } ) {
        push @names, $_->{node};
    }
    return @names;
}

=head2 ::Nodes

Returns an array of NOTE::GX::Node objects representing the
Nodes in the current cluster.

=over

=item Arguments

I<None.>

=item Examples

=back

    my @nodes = $cluster->Nodes();

=cut

sub Nodes {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @nodes = ();

    for ( $self->nodes(%args) ) {
        push @nodes, NOTE::GX::Node->connect($_);
    }
    return @nodes;
}

=head2 ::volumes

Returns the list of volumes associated with this cluster.

=over

=item Arguments

The following key/value pairs can be used to filter the list:

=item Example

=back

    printf( "Online volumes in this cluster:\n" );
    for my $vol ( $cluster->volumes( state => 'online' ) ) {
        print "$vol\n";
    }
    printf( "Offline volumes in this cluster:\n" );
    for my $vol ( $cluster->volumes( state => 'offline' ) ) {
        print "$vol\n";
    }

=cut

sub volumes {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @vols     = $self->_get_vol_list(%args);
    my @volnames = ();
    for (@vols) {
        push @volnames, $_->{volume};
    }
    return @volnames;
}

sub Volumes {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @vols       = ();
    my $spare_only = delete $args{spare_only};

    my @volumes = $self->_get_vol_list(%args);
    for my $v (@volumes) {
        next if ( $spare_only && !$self->looks_free( $v->{volume} ) );
        push @vols,
          NOTE::GX::Volume->attach(
            Node    => $self->Node,
            vserver => $v->{vserver},
            name    => $v->{volume},
          );
    }
    return @vols;
}

sub _get_vol_list {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @vols;

    eval { @vols = $self->server->volume_show(%args); };
    if ($@) {
        confess $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    return @vols;
}

=head2 ::aggregates

Returns the list of aggregates associated with this cluster.

=over

=item Arguments

The following key/value pairs can be used to filter the list:

    aggregate        <name>
    availsize        <integer>
    chksumenabled    <true|false>
    chksumstatus     <true|false>
    chksumstyle      <block|?>
    diskcount        <integer>
    inconsistent     <true|false>
    maxraidsize      <integer>
    nodes            <string>
    percent-used     <string: e.g. "10%">
    plexes           <string>
    raidgroups       <string>
    raidstatus       <raid_dp|raid4|pending>
    raidtype         <raid_dp|raid4>
    size             <integer>
    state            <online|offline|restricted>
    usedsize         <integer>
    volcount         <integer>
    volume-style     <flex|striped>

=item Example

=back

    printf( "Online aggregates in this cluster:\n" );
    for my $aggr ( $cluster->aggregates( state => 'online' ) ) {
        print "$aggr\n";
    }
    printf( "Offline aggregates in this cluster:\n" );
    for my $aggr ( $cluster->aggregates( state => 'offline' ) ) {
        print "$aggr\n";
    }

=cut

sub aggregates {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @aggrs;
    my @aggrnames = ();

    eval { @aggrs = $self->server->storage_aggregate_show(%args); };
    if ($@) {
        confess $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    for (@aggrs) {
        push @aggrnames, $_->{aggregate};
    }
    return @aggrnames;
}

sub Aggregates {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @aggrs = ();

    my @aggrnames = $self->aggregates(%args);

    for (@aggrnames) {
        push @aggrs,
          NOTE::GX::Aggregate->attach(
            Node => $self->Node,
            name => $_,
          );
    }
    return @aggrs;
}

=head2 ::ha

Gets or sets the high-availibility configuration status of the cluster.

=over

=item Arguments

B<status> - I<0 | 1>

Optional

If present, sets the high-availibility status of the cluster.

Returns the configuration status ( 0 = false, 1 = true ).

=item Example

=back

    $cluster->ha( 1 );
    print "High-availibility configuration status is ", $cluster->ha, "\n";

=cut

sub ha {
    my ( $self, $status ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $status ) {
        $self->server->cluster_ha_modify( configured => $status ? "true" : "false" );
    }
    return $self->server->cluster_ha_show->{configured} =~ /true/;
}

=head2 ::statistics

Returns a reference to a hash containing statistics pertaining to the cluster.

=over

=item Arguments

None.

=item Example

=back

    my $stats = $cluster->statistics;
    for my $key ( keys %{$stats} ) {
        printf( "%15s -- %s\n\0", $key, $stats->{$key} );
    }
    printf( "\n" );

=cut

sub statistics {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    # Note: the values in this hash are in bytes.
    # It would be better if they could be returned in MB.
    return $self->server->cluster_statistics_show->[0];
}

=head2 ::alarms

Returns an array of references to hashes that contain information about
alarms pending in the cluster. Each hash has the following fields:

  type      Type of alarm
  node      Node the alarm is associated with
  name      Name of the object with the alarm
  state     Alarm state (ok|warning|critical)
  ems-state The EMS state (rising|falling)
  value     The most recent value
  high-value    The highest value

=over

=item Arguments

Each of the keys described above can be passed along with
a value. This will act as a filter. For example,

=back

    @alarms = $cluster->alarms( type => 'aggregate-used' );

    @alarms = $cluster->alarms(
        type  => 'aggregate-used',
        state => 'critical',
    );

=cut

sub alarms {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @alarms;

    eval { @alarms = $self->server->dashboard_alarm_show( %args, ) };
    if ($@) {
        confess $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    return @alarms;
}

sub Node {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->{_node}
}

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->{_node}->server
}

sub zerospares {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    for my $node ( $self->Nodes ) {
        $node->zerospares;
    }
}

sub is_a_simulator {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    # Unfortunately, I can't tell from the Server.pm
    # object, whether this is a simulator or not..
    # It would be nice if it could give me the
    # NATE connection rec...

    return 0;
}

=head2 set_mixed_geometry

Set mixed geometry mode for all nodes in this cluster

WARNING: This method requires privilege level test to run.

=over

=item Arguments

B<value> => I<Boolean>

Boolean flag to indicate mixed geometry mode. Default: 1

=item Returns

=over

=item B<Success:>

nothing is returned

=item B<Failure:>

confesses:
1) if test mode not available

=back

=item Examples

=back

    # disable mixed mode geometry for the cluster
    $cluster->set_mixed_geometry( value => 0);

    # enable mixed mode geometry for the cluster, using defaults
    $cluster->set_mixed_geometry();

=cut

sub set_mixed_geometry {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # default is to enable mixed mode geometry, value can be 0 so have to
    # check if argument exists not if it is defined
    my $value = exists($args{value}) ? $args{value} : 1;

    foreach my $node ( $self->Nodes ) {
        # use setflag to get the correct privilege level for this flag
        # and so the flag is reset to its original value during teardown
        # (burt 490236)
        $node->setflag(name => 'paloma_allow_unlike_geometry', value => $value);
    }
}

=head2 ::job_list

Returns a list of hash references in the cluster where each hash contains job information.

=over

=item Arguments
B<timeout> => I<number>
<Optional>
Timeout for Server.pm to wait for filer output.

=item Examples

=back

    my @jobs = $cluster->job_list();
    my @jobs = $cluster->job_list(timeout => 30);

=cut

sub job_list {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts;
    my @job_list;

    # setting the timeout to be passed to Server.pm
    if (defined $args{timeout}) {
        $opts{'-Timeout'} = delete $args{timeout};
    }

    my $job_output = $self->server->job_show(%opts);
    # Remove an extra level of reference.
    foreach my $job_ref (@{$job_output}) {
        push @job_list, $job_ref;
    }

    return @job_list;
}

=head2 ::ip_to_lif

Map an IP address in the cluster to the LIF name it represents. If passed
with no IP argument, returns a hash-reference of all IP addresses mapped to
their LIFs.

=over

=item Arguments

B<ip_address> = string

I<Optional> IP address to map. If not given, then a hashref of all IP-to-LIF
mappings is returned. If an IP address is given but is not in the map of LIFs,
an exception is thrown using B<confess>.

=item Examples

For looking up a single LIF:

    @addrs = $src_cluster->peer_entry(cluster => $dst_cluster);
    $lif = $dst_cluster->ip_to_lif($addrs[0]);

Getting all LIFs at once:

    @addrs = $src_cluster->peer_entry(cluster => $dst_cluster);
    $all_lifs = $dst_cluster->ip_to_lif();
    print "Address-to-LIF map:\n";
    foreach my $addr (@addrs) {
        printf "\t%s:\t%s\n", $addr, $all_lifs->{$addr} || '(none)';
    }

=back

=cut

sub ip_to_lif {
    my ($self, $ip_address) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    # Get all the LIFs for this cluster
    my @lifs = $self->lif_list();

    # Since the IP address is unique, even if the LIF name isn't (cluster LIFs
    # names are the same on each machine), this should still yield a map with
    # no collisions on the keys.
    my %map = map { ($_->{address}, $_->{lif}) } @lifs;

    if ($ip_address && ! $map{$ip_address}) {
        confess "Address '$ip_address' does not match any LIFs";
    }

    return $ip_address ? $map{$ip_address} : \%map;
}

=head2 ::transfer_engine

Sets the Paloma physical flag to true or false for the whole cluster.

=over

=item Arguments

B<flag> = I<Boolean>

=item Examples

=back

Cross-cluster friendly example:

      1 # -*- perl -*-
      2 # $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Cluster.pm#1 $
      3
      4 use strict;
      5 use GNOTE;
      6 use dp::TestWrapper qw( run_test );
      7
      8 my ($util, $src, $dst, $sm);
      9 run_test();
     10
     11 sub do_setup {
     12     $util = NOTE::SnapMirrorUtil->new();
     13     ($src) = $util->create_volume(tag=>"src");
     14     ($dst) = $util->create_volume(tag=>"dst", 'mirror-type'=>"DP", count => 1);
     15
     16 }
     17
     18 sub do_test {
     19     $src->Cluster->transfer_engine('BRE');
     20     $dst->Cluster->transfer_engine('BRE');
     21     $sm = NOTE::SnapMirror->initialize(src => $src, dst => $dst);
     22     $src->Cluster->transfer_engine('LRE');
     23     $dst->Cluster->transfer_engine('LRE');
     24     $sm->update;
     25 }
     26
     27 sub do_teardown { NOTE::teardown }

=cut

sub transfer_engine {
    my ($self, $engine ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my @nodes = $self->Nodes;
    foreach my $Node ( @nodes ) {
        $Node->transfer_engine( $engine );  # Set the transfer engine per node
    }
}

sub DESTROY {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    return if ( $$ != $self->{_owner} );    # Only the creator can destroy an object
    return if ( $self->{_spare_me} );       # Means don't destroy
    return if ( $self->{_destroyed} );      # Already dead
    return if ( NOTE->ALL_SPARED );

    $self->{_destroyed} = 1;
}

1;
