# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/DatasetTransform.pm#1 $
#
## @summary NOTE DatasetTransform Class
##
## @status public
##
## @author tyu@netapp.com, dl-dpgqa-automation@netapp.com
##
## @pod here

=head1 NAME

NOTE::DatasetTransform

=head1 DESCRIPTION

The purpose of this module is to provide some convenient functions to construct
and modify dataset/s that is/are specifically designed for testing the
baseline and incremental snapmirror operations.

=cut

package NOTE::DatasetTransform;
use strict;
use GNOTE;
use TracePkg;
use Data::Dumper;
use Carp qw(confess);
use Params::Validate qw(validate ARRAYREF HASHREF OBJECT SCALAR UNDEF);

our @ISA = qw(NOTE);

my $Trace = TracePkg->new ();

=head2 ::new

Constructor for the DatasetTransform class.

=over

=item Arguments

B<src> => I<Volume Object>

I<Required>

A NOTE::Volume object that represent the volume where the Qtrees and Files are being manipulated.

B<prefix> => I<string>

I<Optional.>

Prefix dataset name. Default: 'dsx'

=item Examples

=back

    my $dataset1 = NOTE::DatasetTransform->new (vol=>$volume, prefix=>'dsx1');

    $dataset1->tranform_qtree_to_filetype (num_of_nfsfile=>10, num_of_nfsdir=>2, num_of_nfsfifo=>5);
    $dataset1->tranform_filetype_to_qtree (num_of_nfsfile=>10, num_of_nfsdir=>2, num_of_nfsfifo=>5);
    $dataset1->tranform_qtree_delete_and_add (num_of_delete=>20, num_of_add=>5);
    $dataset1->tranform_qtree_rename (num_of_rename=>10, to_name=>'new');

    my $snapshot_name = $dataset1->get_snapshot_name (iteration=>1, stata=>'post_xform');
    NOTE::logcomment ("Snapshot for iteration 1 is : $snapshot_name");

=cut

sub new () {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    bless $self, $class;
    NOTE::reference($self);

    $self->{_owner} = $$;

    if ( ref( $args{vol} ) =~ /NOTE::GX::Volume/ ) {
        $self->{_vol} = ( $args{vol} );
    } else {
        confess "Invalid vol specification : ".ref $args{vol};
    }

    $Trace->debug (msg=>"(DatasetTransform::new): _vol is :".$self->{_vol});

    if (exists $args{prefix}) {
	$self->{_str_prefix} = $args{prefix};
    } else {
	$self->{_str_prefix} = "dsx";
    }
    $self->{_iteration} = 0;

    $Trace->debug (msg=>"(DatasetTransform::new): _prefix is :".$self->{_str_prefix});
    $Trace->debug (msg=>"(DatasetTransform::new): _iteration is :".$self->{_iteration});

    $self->{_snapshot_names} = ();
    return $self;
}

#
# Various Transformation Types provided by this class
#
use constant {
    XFORM_TYPE_QTREE_TO_FILE      => "q2f",
    XFORM_TYPE_FILE_TO_QTREE      => "f2q",
    XFORM_TYPE_QTREE_DELETE_ADD   => "delete_add",
    XFORM_TYPE_QTREE_RENAME       => "rename",
};

#
# Various Transformation States which are being used in naming of snapshots
#
use constant {
    XFORM_STATE_PRE_XFORM      => "pre_xform",
    XFORM_STATE_IN_PROGRESS    => "in_progress",
    XFORM_STATE_POST_XFORM     => "post_xform",
};

=head2 ::transform_qtree_to_filetype

This function perform the testing of the expected behavior of inode reuse when
Qtree is transformed to other file types. Transformation means that when a
Qtree is being removed, and immedidately followed by the creation of a new
file (in various file types). The newly created file should reuse the inode
number of the just-deleted Qtree.

This function will start from creating a list of Qtrees, based on the user's
specification of files that he/she wants to test for the transformation. It
will then create a snapshot to capture the situation before the transformation/s
happen. Then it move on the perform the various transformation/s as requested
by the user. At the end, it will create another snapshot to record the
post transformations situation.

=over

=item Arguments

B<num_of_nfsfile> = I<integer>

I<Optional>

Number of NFS File/s to transform towards. Default: 0

B<num_of_nfsdir> = I<integer>

I<Optional>

Number of NFS Directories to transform towards. Default: 0

B<num_of_nfsfifo> = I<integer>

I<Optional>

Number of NFS Fifo/s to transform towards. Default: 0

B<num_of_additional_qtree> = I<integer>

I<Optional>

Number of Qtree/s to create initially in addition to the sum of the files. Default: 0

=item Return Values

Return transform iteration ID, which identify the specific iteration
of transformation called by the user. This information can be use
later to retrieve the snapshot name.

=item Examples

    my $xform_iter_id = $dataset1->tranform_qtree_to_filetype (num_of_nfsfile=>10, num_of_nfsdir=>2, num_of_nfsfifo=>5);

=back

=cut

sub transform_qtree_to_filetype {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    num_of_nfsfile => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_nfsdir => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_nfsfifo => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_cifsfile => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_cifsfile_with_acl => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_cifsfile_with_stream => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_additional_qtree => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $num_of_nfsfile = delete $args {num_of_nfsfile};
    my $num_of_nfsdir = delete $args {num_of_nfsdir};
    my $num_of_nfsfifo = delete $args {num_of_nfsfifo};
    my $num_of_cifsfile = delete $args {num_of_cifsfile};
    my $num_of_cifsfile_with_acl = delete $args {num_of_cifsfile_with_acl};
    my $num_of_cifsfile_with_stream = delete $args {num_of_cifsfile_with_stream};
    my $num_of_additional_qtree = delete $args {num_of_additional_qtree};

    NOTE::logcomment("In transform_qtree_to_filetype:".
        "\nnum_of_nfsfile=$num_of_nfsfile".
	"\nnum_of_nfsdir=$num_of_nfsdir".
	"\nnum_of_nfsfifo=$num_of_nfsfifo".
	"\nnum_of_cifsfile=$num_of_cifsfile".
	"\nnum_of_cifsfile_with_acl=$num_of_cifsfile_with_acl".
	"\nnum_of_cifsfile_with_stream=$num_of_cifsfile_with_stream".
	"\nnum_of_additional_qtree=$num_of_additional_qtree"
	);

    my $total_num_of_files = $num_of_nfsfile + $num_of_nfsdir +
        $num_of_nfsfifo + $num_of_cifsfile +
	$num_of_cifsfile_with_acl + $num_of_cifsfile_with_stream;

    if ($total_num_of_files <= 0) {
        confess ("Invalid input of the number of files in various filetypes. The total must be > 0!");
    }

    my $total_num_of_qtrees = $total_num_of_files + $num_of_additional_qtree;

    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_to_filetype): total_num_of_files is :".$total_num_of_files);
    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_to_filetype): total_num_of_qtrees is :".$total_num_of_qtrees);

    my ($count, $qt);

    #
    # This is to keep track of the number of times that the various
    # transform functions are being called.
    #
    $self->{_iteration}++;

    my $iteration = $self->{_iteration};
    my ($vol) = $self->{_vol};

    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_to_filetype): iteration is :".$iteration);
    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_to_filetype): vol is :".$vol);

    #
    # 1. Create Qtrees for later manipulation
    #
    for ($count = 0; $count < $total_num_of_qtrees; $count++) {
	my $tag = $self->_get_qt_tag_prefix ()."_$count";
	$Trace->debug (msg=>"(DatasetTransform::transform_qtree_to_filetype): tag is :".$tag);

	($qt) = NOTE::GX::Qtree->create (Volume=>$vol, tag=>$tag);

	$self->_set_qtree_array (index=>$count, qt=>$qt);
    }

    my $index=0;

    $self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_TO_FILE, state=>XFORM_STATE_PRE_XFORM);

    #
    # 2. Testing Qtree to NFS File Transformation
    #    - Delete Qtree
    #	 - Add NFS File
    #    - Compare inode
    #

    for ($count=0; $count<$num_of_nfsfile; $count++, $index++) {
	$Trace->debug (msg=>"(DatasetTransform::xform): Before _get_qtree_array iteration is :".$self->{_iteration});
        $qt = $self->_get_qtree_array (index=>$index);

	# Get the qtree root directory object and its inode number
	my $qt_root  = $qt->nfs;
	my $old_inode = $self->_get_file_inode (file=>$qt_root);
	NOTE::logcomment ("Qtree root directory inode number before transform :",
			  $old_inode);

#	$self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_TO_FILE, state=>XFORM_STATE_IN_PROGRESS);

	# Delete the qtree after the snapshot is created
	$qt->purge;

        my $nfsfile_name = $self->_get_nfsfile_prefix ()."_".join ("", $count);

	my $nfs_file = $vol->nfs->create ($nfsfile_name);
	$nfs_file->fill (max => '200m');
	my $new_inode = $self->_get_file_inode (file=>$nfs_file);

	NOTE::logcomment ("New file created, old inode=",$old_inode, ", new inode=", $new_inode);

	if ($old_inode == $new_inode) {
	    NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree root directory inode number ",
	                      $old_inode, " was recycled and ",
			      "now represents a NFS file!");
	} else {
	    NOTE::logcomment ("Fail to transform Qtree to NFS File, inode numbers are different. Qtree's inode = ".$old_inode." NFS File's inode = ".$new_inode);
	}
    }

    #
    # 3. Testing Qtree to NFS Directory Transformation
    #

    for ($count=0; $count<$num_of_nfsdir; $count++, $index++) {
        $qt = $self->_get_qtree_array (index=>$index);

	# Get the qtree root directory object and its inode number
	my $qt_root  = $qt->nfs;
	my $old_inode = $self->_get_file_inode (file=>$qt_root);
	NOTE::logcomment ("Qtree root directory inode number before ",
			  "transformation: ", $old_inode);

#	$self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_TO_FILE, state=>XFORM_STATE_IN_PROGRESS);

	# Delete the qtree
	$qt->purge;

        my $nfsdir_name = $self->_get_nfsdir_tag_prefix ()."_".join ("", $count);
        my $nfsfile_name = $nfsdir_name."_file";

	my $nfs_dir = $vol->nfs->mkdir ($nfsdir_name);
	my $nfs_file = $nfs_dir->create ($nfsfile_name);
	$nfs_file->fill (max => '200m');

	my $new_inode = $self->_get_file_inode (file=>$nfs_dir);

	NOTE::logcomment("Created a NFS directory with a file after ",
			 "deleting the qtree");
	NOTE::logcomment ("Old inode=",$old_inode, ", New inode=", $new_inode);

	if ($old_inode == $new_inode) {
	    NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree root directory inode number ",
	                      $old_inode, " was recycled and ",
			      "now represents a NFS directory!");
	} else {
	    NOTE::logcomment ("Fail to transform Qtree to NFS Directory, inode numbers are different. Qtree's inode = ".$old_inode." NFS File's inode = ".$new_inode);
	}
    }

    #
    # 4. Testing Qtree to NFS FIFO Transformation
    #

    for ($count=0; $count<$num_of_nfsfifo; $count++, $index++) {
        $qt = $self->_get_qtree_array (index=>$index);

	# Get the qtree root directory object and its inode number
	my $qt_root  = $qt->nfs;
	my $old_inode = $self->_get_file_inode (file=>$qt_root);
	NOTE::logcomment ("Qtree root directory inode number before ",
			  "transform: ", $old_inode);

#	$self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_TO_FILE, state=>XFORM_STATE_IN_PROGRESS);

	# Delete the qtree after the snapshot is created
	$qt->purge;

        my $nfsfifo_name = $self->_get_nfsfifo_tag_prefix ()."_".join ("", $count);

	# Create a FIFO (a named pipe)
	my $nfs_fifo = $vol->nfs->create_fifo (pathname=>$nfsfifo_name);

	my $new_inode = $self->_get_file_inode (file=>$nfs_fifo);

	NOTE::logcomment ("Old inode=",$old_inode,", New inode=", $new_inode);

	if ($old_inode == $new_inode) {
	    NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree root directory inode number ", $old_inode,
	        " was recycled and now represents a NFS fifo file!");
	} else {
	    NOTE::logcomment ("Fail to transform Qtree to NFS Fifo, inode numbers are different. Qtree's inode = ".$old_inode." NFS File's inode = ".$new_inode);
	}
    }

    $self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_TO_FILE, state=>XFORM_STATE_POST_XFORM);
    return $iteration;
}

=head2 ::transform_filetype_to_qtree

This function perform the testing of the expected behavior of inode reuse when
file in various types are transformed to Qtree. Transformation means that the file will be removed, and then immediately followed by the creation of a Qtree.
The newly created Qtree should reuse the inode number of the just-deleted file.

This function will start from creating a list of files, based on the user's
specification to test for the transformations. It will then create a snapshot
to capture the situation before the transformation/s happen. Then it move on
to perform the various transformation/s as requested by the user (i.e. deleting
the file and creating the corresponding Qtree). At the end, it will create
another snapshot to record the post transformations situation.

=over

=item Arguments

B<num_of_nfsfile> = I<integer>

I<Optional>

Number of NFS File/s to transform towards. Default: 0

B<num_of_nfsdir> = I<integer>

I<Optional>

Number of NFS Directories to transform towards. Default: 0

B<num_of_nfsfifo> = I<integer>

I<Optional>

Number of NFS Fifo/s to transform towards. Default: 0

=item Return Values

Return transform iteration ID, which identify the specific iteration
of transformation called by the user. This information can be use
later to retrieve the snapshot name.

=item Examples

    my $status = $dataset1->tranform_filetype_to_qtree (num_of_nfsfile=>10, num_of_nfsdir=>2, num_of_nfsfifo=>5);

=back

=cut

sub transform_filetype_to_qtree {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    num_of_nfsfile => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_nfsdir => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_nfsfifo => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_cifsfile => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_cifsfile_with_acl => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	    num_of_cifsfile_with_stream => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $num_of_nfsfile = delete $args {num_of_nfsfile};
    my $num_of_nfsdir = delete $args {num_of_nfsdir};
    my $num_of_nfsfifo = delete $args {num_of_nfsfifo};
    my $num_of_cifsfile = delete $args {num_of_cifsfile};
    my $num_of_cifsfile_with_acl = delete $args {num_of_cifsfile_with_acl};
    my $num_of_cifsfile_with_stream = delete $args {num_of_cifsfile_with_stream};

    NOTE::logcomment("In transform_filetype_to_qtree:".
        "\nnum_of_nfsfile=$num_of_nfsfile".
	"\nnum_of_nfsdir=$num_of_nfsdir".
	"\nnum_of_nfsfifo=$num_of_nfsfifo".
	"\nnum_of_cifsfile=$num_of_cifsfile".
	"\nnum_of_cifsfile_with_acl=$num_of_cifsfile_with_acl".
	"\nnum_of_cifsfile_with_stream=$num_of_cifsfile_with_stream");

    my $total_num_of_files = $num_of_nfsfile + $num_of_nfsdir +
        $num_of_nfsfifo + $num_of_cifsfile +
	$num_of_cifsfile_with_acl + $num_of_cifsfile_with_stream;

    if ($total_num_of_files <= 0) {
        confess ("Invalid input of the number of files in various filetypes. The total must be > 0!");
    }

    my $total_num_of_qtrees = $total_num_of_files;

    NOTE::logcomment ("(DatasetTransform::transform_filetype_to_qtree): total_num_of_files is :".$total_num_of_files);
    NOTE::logcomment ("(DatasetTransform::transform_filetype_to_qtree): total_num_of_qtrees is :".$total_num_of_qtrees);

    my ($count, $qt);

    #
    # This is to keep track of the number of times that the various
    # transform functions are being called.
    #
    $self->{_iteration}++;

    my $iteration = $self->{_iteration};
    my ($vol) = $self->{_vol};

    $Trace->debug (msg=>"(DatasetTransform::new): iteration is :".$iteration);
    $Trace->debug (msg=>"(DatasetTransform::new): vol is :".$vol);

    #
    # 1. Create NFS Files for later transformation
    #    - Create NFS files and store the handles
    #

    my $index = 0;
    for ($count=0; $count<$num_of_nfsfile; $count++, $index++) {
        my $nfsfile_name = $self->_get_nfsfile_prefix ()."_".join ("", $count);

	my $nfs_file = $vol->nfs->create ($nfsfile_name);
	$nfs_file->fill (max => '200m');
	$self->_set_file_array (index=>$index, file=>$nfs_file);
    }

    #
    # 2. Create NFS Directories, and store the handles for later transformation
    #

    for ($count=0; $count<$num_of_nfsdir; $count++, $index++) {
        my $nfsdir_name = $self->_get_nfsdir_tag_prefix ()."_".join ("", $count);
        my $nfsfile_name = $nfsdir_name."_file";

	my $nfs_dir = $vol->nfs->mkdir ($nfsdir_name);
	$self->_set_file_array (index=>$index, file=>$nfs_dir);
    }

    #
    # 3. Create NFS FIFO and store the handles
    #

    for ($count=0; $count<$num_of_nfsfifo; $count++, $index++) {
        my $nfsfifo_name = $self->_get_nfsfifo_tag_prefix ()."_".join ("", $count);

	# Create a FIFO (a named pipe)
	my $nfs_fifo = $vol->nfs->create_fifo (pathname=>$nfsfifo_name);
	
	$self->_set_file_array (index=>$index, file=>$nfs_fifo);
    }

    $self->_take_snapshot (transform_type=>XFORM_TYPE_FILE_TO_QTREE, state=>XFORM_STATE_PRE_XFORM);

    my $num_of_files_pass1 = $index;

    #
    # Take a snapshot
    #
    # In Pass 2, Go through all the files that had been created, for each file
    #   delete the file
    #   create Qtree
    #   verify inode and Qtree ID reuse
    #
    # Take a snapshot

    for ($count = 0; $count < $num_of_files_pass1; $count++) {

        my $the_file = $self->_get_file_array (index=>$count);

	NOTE::logcomment ("(DatasetTransform::transform_filetype_to_qtree): file[$count] is :".$the_file);

	my $old_inode = $self->_get_file_inode (file=>$the_file);
	my $the_file_type = $the_file->type ();

	if ($the_file_type =~ /file/) {
	    $the_file->unlink ();
	} elsif ($the_file_type =~ /directory/) {
	    $the_file->rm ();
	} elsif ($the_file_type =~ /fifo/) {
	    $the_file->unlink ();
	} else {
	    confess "Unknown file type, $the_file_type, for transformation to Qtree";
	}

	my $tag = $self->_get_qt_tag_prefix ()."_$count";
	NOTE::logcomment ("(DatasetTransform::transform_filetype_to_qtree): Before create Qtree with tag ".$tag);

	($qt) = NOTE::GX::Qtree->create (Volume=>$vol, tag=>$tag);

	$self->_set_qtree_array (index=>$count, qt=>$qt);

	# Get the qtree root directory object and its inode number
	my $qt_root  = $qt->nfs;
	my $new_inode = $self->_get_file_inode (file=>$qt_root);

	NOTE::logcomment ("Old inode=",$old_inode, ", New inode=", $new_inode);

	if ($old_inode == $new_inode) {
	    NOTE::logcomment ("(TRANSFORM_SUCCEED): File object with file type $the_file_type with inode $old_inode had transformed to Qtree root directory");
	} else {
	    NOTE::logcomment ("Fail to transform File Type, $the_file_type, to Qtree, inode numbers are different. Qtree's inode = ".$old_inode." File's inode = ".$new_inode);
	}
    }

    $self->_take_snapshot (transform_type=>XFORM_TYPE_FILE_TO_QTREE, state=>XFORM_STATE_POST_XFORM);
    return $iteration;
}

=head2 ::transform_qtree_delete_and_add

This function perform the testing of the expected behavior of inode, and
Qtree ID reuse when Qtree is deleted and immediately follow by the creation
of another Qtree.

This function will start from creating a list of Qtrees, based on the user's
specification of number of Qtrees to delete. It will then create a snapshot to
capture the situation before the transformation happen. Then it move on to
delete the Qtree one by one, and for each Qtree deletion, it will immediately
create another Qtree following it, and then validate the resue of the Qtree ID
as well as the inode.  At the end, it will create another snapshot to record
the post transformations situation.

=over

=item Arguments

B<num_of_delete> = I<integer>

I<Optional>

Number of Qtree/s to delete. Default: 10

B<num_of_add> = I<integer>

I<Optional>

Number of Qtree/s to add. Default: 5

B<num_of_additional_qtree> = I<integer>

I<Optional>

Number of Qtree/s to create initially in addition to the sum of the files. Default: 0

=item Return Values

Return transform iteration ID, which identify the specific iteration
of transformation called by the user. This information can be use
later to retrieve the snapshot name.

=item Examples

    my $status = $dataset1->tranform_qtree_delete_and_add (num_of_delete=>10, num_of_add=>2);

=back

=cut

sub transform_qtree_delete_and_add {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    num_of_delete => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>10},
	    num_of_add => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>5},
	    num_of_additional_qtree => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $num_of_delete = delete $args {num_of_delete};
    my $num_of_add = delete $args {num_of_add};
    my $num_of_additional_qtree = delete $args {num_of_additional_qtree};

    NOTE::logcomment ("In transform_qtree_delete_and_add:".
        "\nnum_of_delete=$num_of_delete".
	"\nnum_of_add=$num_of_add".
	"\nnum_of_additional_qtree=$num_of_additional_qtree"
	);

    my $total_num_of_files = $num_of_delete + $num_of_add;

    if ($total_num_of_files <= 0) {
        confess ("Invalid input of the number of files in various filetypes. The total must be > 0!");
    }

    my $total_num_of_qtrees = $num_of_delete + $num_of_additional_qtree;

    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_delete_and_add): total_num_of_files is :".$total_num_of_files);
    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_delete_and_add): total_num_of_qtrees is :".$total_num_of_qtrees);

    my ($count, $qt);

    #
    # This is to keep track of the number of times that the various
    # transform functions are being called.
    #
    $self->{_iteration}++;

    my $iteration = $self->{_iteration};
    my ($vol) = $self->{_vol};

    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_delete_and_add): iteration is :".$iteration);
    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_delete_and_add): vol is :".$vol);

    #
    # 1. Create Qtrees for later manipulation
    #
    for ($count = 0; $count < $total_num_of_qtrees; $count++) {
	my $tag = $self->_get_qt_tag_prefix ()."_$count";
	NOTE::logcomment ("(DatasetTransform::new): tag is :".$tag);

	($qt) = NOTE::GX::Qtree->create (Volume=>$vol, tag=>$tag);

	$self->_set_qtree_array (index=>$count, qt=>$qt);
    }

    my $index=0;
    my $hash_ref;
    my $old_qtree_id;

    $self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_DELETE_ADD, state=>XFORM_STATE_PRE_XFORM);

    #
    # 2. Delete and Add Qtree
    #    - Delete Qtree
    #	 - Add it back
    #    - Compare tree_id
    #

    my $curr_num_of_add = 0;

    for ($count=0; $count<$num_of_delete; $count++, $index++) {

        $qt = $self->_get_qtree_array (index=>$index);

	$hash_ref = $qt->status;
	$old_qtree_id = $hash_ref->{'qtree-id'};

	NOTE::logcomment ("Qtree ID of qtree before transform to file: ", $old_qtree_id);

	# Get the qtree root directory object and its inode number
	my $qt_root  = $qt->nfs;
	my $old_inode = $self->_get_file_inode (file=>$qt_root);
	NOTE::logcomment ("Qtree root directory inode number before transform: ", $old_inode);

	# Delete the qtree
	$qt->purge;

	if ($curr_num_of_add < $num_of_add) {
	    my $tag = $self->_get_qt_tag_prefix ()."_$count";
	    NOTE::logcomment ("(DatasetTransform::transform_qtree_delete_and_add): tag is :".$tag);

	    ($qt) = NOTE::GX::Qtree->create (Volume=>$vol, tag=>$tag);

	    $self->_set_qtree_array (index=>$count, qt=>$qt);

	    # Extract the qtree ID of newly created qtree
	    $hash_ref = $qt->status;
	    my $new_qtree_id = $hash_ref->{'qtree-id'};
	    NOTE::logcomment ("Qtree ID of newly added Qtree : ", $new_qtree_id);

	    # Get the qtree root directory object and its inode number
	    $qt_root  = $qt->nfs;
	    my $new_inode = $self->_get_file_inode (file=>$qt_root);
	    NOTE::logcomment ("Inode of newly added Qtree : ", $new_inode);

	    if ($old_qtree_id eq $new_qtree_id) {
		NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree ID ", $new_qtree_id, " was reused after delete and add!");
	    } else {
		confess ("Fail to transform Qtree in delete and add operation, the Tree ID/s are different. Old Qtree's ID = ".$old_qtree_id." New Qtree's ID = ".$new_qtree_id);
	    }

	    if ($old_inode eq $new_inode) {
		NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree root directory inode number ", $new_inode, " was reused after delete and add!");
	    } else {
		NOTE::logcomment ("Fail to transform Qtree in delete and add operation, inode numbers are different. Qtree's inode = ".$old_inode." NFS File's inode = ".$new_inode);
	    }

	    $curr_num_of_add ++;
	}
    }

    $self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_DELETE_ADD, state=>XFORM_STATE_POST_XFORM);

    return $iteration;
}

=head2 ::transform_qtree_rename

This function perform the testing of renaming Qtree, and verify that there's
no impact on the underneath inode and Tree ID of the Qtree.

This function will start from creating a list of Qtrees, based on the user's
specification of files that he/she wants to test for the transformation. It
will then create a snapshot to capture the situation before the transformation/s
happen. Then it move on the rename the Qtree one by one, and verify that the
inode and Tree ID stay the same. At the end, it will create another snapshot
to record the post transformations situation.

=over

=item Arguments

B<num_of_qtree> = I<integer>

I<Optional>

Number of Qtree/s to rename. Default: 10

B<to_name> = I<string>

I<Optional>

Prefix used as name of the new Qtree. Default: 'newqt'

B<num_of_additional_qtree> = I<integer>

I<Optional>

Number of Qtree/s to create initially in addition to the sum of the files. Default: 0

=item Return Values

Return transform iteration ID, which identify the specific iteration
of transformation called by the user. This information can be use
later to retrieve the snapshot name.

=item Examples

    my $status = $dataset1->tranform_qtree_rename (num_of_qtree=>10, to_name=>'newqt');

=back

=cut

sub transform_qtree_rename {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    num_of_rename => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>10},
	    to_name => {type => SCALAR, optional=>1, default=>'newqt'},
	    num_of_additional_qtree => {type => SCALAR, optional=>1, regex=>qr/^\d+$/, default=>0},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $num_of_rename = delete $args {num_of_rename};
    my $to_name = delete $args {to_name};
    my $num_of_additional_qtree = delete $args {num_of_additional_qtree};

    NOTE::logcomment ("In transform_qtree_rename:".
        "\nnum_of_rename=$num_of_rename".
	"\nto_name=$to_name".
	"\nnum_of_additional_qtree=$num_of_additional_qtree"
	);

    my $total_num_of_qtrees = $num_of_rename + $num_of_additional_qtree;

    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_rename): total_num_of_qtrees is :".$total_num_of_qtrees);

    my ($count, $qt);

    #
    # This is to keep track of the number of times that the various
    # transform functions are being called.
    #
    $self->{_iteration}++;

    my $iteration = $self->{_iteration};
    my ($vol) = $self->{_vol};

    $Trace->debug (msg=>"(DatasetTransform::transform_qtree_rename): iteration is :".$iteration);

    #
    # 1. Create Qtrees for later manipulation
    #
    for ($count = 0; $count < $total_num_of_qtrees; $count++) {
	my $tag = $self->_get_qt_tag_prefix ()."_$count";
	NOTE::logcomment ("(DatasetTransform::transform_qtree_rename): tag is :".$tag);

	($qt) = NOTE::GX::Qtree->create (Volume=>$vol, tag=>$tag);

	$self->_set_qtree_array (index=>$count, qt=>$qt);
    }

    my $index=0;
    my $hash_ref;
    my $old_qtree_id;

    $self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_RENAME, state=>XFORM_STATE_PRE_XFORM);

    #
    # 2. Testing Qtree transformation by renaming
    #    - Rename the Qtree
    #    - Validate that the inode and tree id stay the same
    #

    for ($count=0; $count<$num_of_rename; $count++, $index++) {

        $qt = $self->_get_qtree_array (index=>$index);

	$hash_ref = $qt->status;
	$old_qtree_id = $hash_ref->{'qtree-id'};

	NOTE::logcomment ("Qtree ID of qtree before transform to file: ",
			  $old_qtree_id);

	# Get the qtree root directory object and its inode number
	my $qt_root  = $qt->nfs;
	my $old_inode = $self->_get_file_inode (file=>$qt_root);
	NOTE::logcomment ("Qtree root directory inode number before ",
			  "transform: ", $old_inode);

	my $new_name = $self->_get_qt_tag_prefix ().$to_name."_$count";

	# Delete the qtree
	$qt->rename (newname=>$new_name);

	# Extract the qtree ID of newly created qtree
	$hash_ref = $qt->status;
	my $new_qtree_id = $hash_ref->{'qtree-id'};
	NOTE::logcomment ("Qtree ID of qtree after transform: ",
			  $new_qtree_id);

	# Get the qtree root directory object and its inode number
	$qt_root  = $qt->nfs;
	my $new_inode = $self->_get_file_inode (file=>$qt_root);

	if ($old_qtree_id eq $new_qtree_id) {
	    NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree ID ", $new_qtree_id, " was reused in Qtree rename!");
	} else {
	    confess ("Fail to transform Qtree in rename operation, the Tree ID/s are different. Old Qtree's ID = ".$old_qtree_id." New Qtree's ID = ".$new_qtree_id);
	}

	if ($old_inode eq $new_inode) {
	    NOTE::logcomment ("(TRANSFORM_SUCCEED): Qtree root directory inode number $new_inode was reused in Qtree rename!");
	} else {
	    NOTE::logcomment ("Fail to transform Qtree in rename operation, inode numbers are different. Qtree's inode = ".$old_inode." NFS File's inode = ".$new_inode);
	}
    }

    $self->_take_snapshot (transform_type=>XFORM_TYPE_QTREE_RENAME, state=>XFORM_STATE_POST_XFORM);

    return $iteration;
}

#
# Take a snapshot, with naming convention by means of iteration, transform type
# and state.
#

sub _take_snapshot () {
    my $self = shift;
    my %args = validate (
	@_,
	{
	    transform_type => {type => SCALAR, optional=>0, regex=>qr/(q2f|f2q|delete_add|rename)/},
	    state => {type => SCALAR, optional=>0, regex=>qr/(pre_xform|in_progress|post_xform)/},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $str_transform_type = delete $args {transform_type};
    my $str_state = delete $args {state};

    my $ss_prefix = $self->_get_snapshot_prefix ()."_".$str_transform_type."_".$str_state;

    my $vol = $self->{_vol};

    my $snapshot_name = "";
    
    eval {
	$snapshot_name = $vol->snapshot_create ($ss_prefix);
    };

    if ($@) {
        confess "Failed to create snapshot, name = $ss_prefix";
    }

    my $iteration = $self->{_iteration};
    $self->{_snapshot_names}[$iteration]{$str_state} = $snapshot_name;
    $self->{_transform_types}[$iteration] = $str_transform_type;
}

=head2 ::get_snapshot_name

This function return the name of a snapshot that was taken in the
particular iteration of tranform call, as well as the state (pre_xform
vs post_xform).

=over

=item Arguments

B<None>

=item Return Values

Return the name of the snapshot created during a particular iteration, and the user specification of the state (pre_xform Vs post_xform).

=item Examples

    $snapshot_name = $dataset1->get_snapshot_name (iteration=>1, state=>'post_xform');

=back

=cut

sub get_snapshot_name () {
    my $self = shift;
    my %args = validate (
	@_,
	{
	    iteration => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    state => {type => SCALAR, optional=>0, regex=>qr/(pre_xform|in_progress|post_xform)/},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = delete $args {iteration};
    my $str_state = delete $args {state};

    if ($iteration > $self->{_iteration}) {
	confess ("Invalid input: iteration = $iteration, which is larger than the number of transformations called so far.");
    }

    return $self->{_snapshot_names}[$iteration]{$str_state};
}

#
# Store the qtree object, by means of current iteration, and index
#

sub _set_qtree_array () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    qt => {type => OBJECT, optional=>0, isa=>'NOTE::GX::Qtree'},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = $self->{_iteration};

    $self->_set_qtree_array_by_iteration (iteration=>$iteration, %args);
}

#
# Get the stored qtree object, by means of current iteration, and index
#
sub _get_qtree_array () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = $self->{_iteration};

    return $self->_get_qtree_array_by_iteration (iteration=>$iteration, %args);
}

#
# Get the stored qtree object, by means of input iteration, and index
#
sub _get_qtree_array_by_iteration () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    iteration => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = $args{iteration};
    my $index = $args{index};

    if (exists $self->{_my_qtree}[$iteration][$index]) {
        return ($self->{_my_qtree}[$iteration][$index]);
    }
    
    return undef;
}

#
# Stored the qtree object, by means of input iteration, and index
#
sub _set_qtree_array_by_iteration () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    iteration => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    qt => {type => OBJECT, optional=>0, isa=>'NOTE::GX::Qtree'},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_my_qtree}[$args{iteration}][$args{index}] = $args{qt};
}

#
# Get the stored file object, by means of current iteration, and index
#
sub _get_file_array () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = $self->{_iteration};

    return $self->_get_file_array_by_iteration (iteration=>$iteration, %args);
}

#
# Get the stored file object, by means of input iteration, and index
#
sub _get_file_array_by_iteration () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    iteration => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = $args{iteration};
    my $index = $args{index};

    if (exists $self->{_my_file}[$iteration][$index]) {
        return ($self->{_my_file}[$iteration][$index]);
    }
    
    return undef;
}

#
# Store the file object, by means of current iteration, and index
#
sub _set_file_array () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    file => {type => OBJECT, optional=>0, isa=>'NOTE::FileSystem::NFS::File'},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iteration = $self->{_iteration};

    $self->_set_file_array_by_iteration (iteration=>$iteration, %args);
}

#
# Store the file object, by means of input iteration, and index
#
sub _set_file_array_by_iteration () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    index => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    iteration => {type => SCALAR, optional=>0, regex=>qr/^\d+$/},
	    file => {type => OBJECT, optional=>0, isa=>'NOTE::FileSystem::NFS::File'},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_my_file}[$args{iteration}][$args{index}] = $args{file};
}

#
# Get the inode of the input file object.
#
sub _get_file_inode () {
    my $self = shift;
    my %args = validate(
	@_,
	{
	    file => {type => OBJECT, optional=>0, isa=>'NOTE::FileSystem::NFS::File'},
	});
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $the_file = $args{file};
    my $file_inode = $the_file->inode;	# Fall back strategy

#    my ($vol) = $self->{_vol};
#    my ($file_inode) = $vol->wex_info (path=>$the_file->path, column=>"inum");

    return $file_inode;
}

#
# Get the tag prefix to be used to name the qtree
#
sub _get_qt_tag_prefix () {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iter = $self->{_iteration};

    return $self->{_str_prefix}."_qt_".$iter;
}

#
# Get the tag prefix to be used to name the NFS Directory
#
sub _get_nfsdir_tag_prefix () {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iter = $self->{_iteration};

    return $self->{_str_prefix}."_nfsdir_".$iter;
}

#
# Get the tag prefix to be used to name the NFS File
#
sub _get_nfsfile_prefix () {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iter = $self->{_iteration};

    return $self->{_str_prefix}."_nfsfile_iter".join ("", $iter)."_";
}

#
# Get the tag prefix to be used to name the NFS Fifo
#
sub _get_nfsfifo_tag_prefix () {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iter = $self->{_iteration};

    return $self->{_str_prefix}."_nfsfifo_".$iter;
}

#
# Get the tag prefix to be used to name the snapshot
#
sub _get_snapshot_prefix () {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $iter = $self->{_iteration};

    return $self->{_str_prefix}."_ssn_".$iter;
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    $self->{_destroyed} = 1;
}

1;
