#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Disk.pm#1 $
#
## @summary  GNOTE GX Disk module
## @author   dl-dpgqa-automation@netapp.com
## @status   Public
## @pod      here

=head1 NAME

NOTE::GX::Disk

=head1 DESCRIPTION

This module provides the functionality for a disk in ONTAP

=head1 METHODS

=cut

package NOTE::GX::Disk;
use GNOTE::GX;
use Carp qw( croak confess );
our @ISA = qw(NOTE::GX);

my $_DISK_LIST = ();

=head2 ::new

Returns an object reference for a disk.

=over

=item Arguments

B<name> => I<string>

=item Example

=back

    $disk = NOTE::GX::Disk->new( name => 'dpg-gx1:0a.16' );

=cut

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    return $_DISK_LIST{ $args{name} } if exists $_DISK_LIST{ $args{name} };

    $self->{_name} = $args{name};
    my $node = ( split( ':', $args{name} ) )[0];
    $self->{_container} = NOTE::GX::Node->connect( name => $node );

    bless $self, $class;
    NOTE::reference($self);

    $_DISK_LIST{ $args{name} } = $self;

    $self->{_init_state} = $self->status->{state};

    return $self;
}

=head2 ::attach

Returns an object reference for a disk.
Croaks if the disk doesn't exist.

=over

=item Arguments

B<name> => I<string>

=item Example

=back

    $disk = NOTE::GX::Disk->attach( name => 'dpg-gx1:0a.16' );

=cut

sub attach {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = $class->new(%args);
    return $self;
}

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_container}->server
}

=head2 ::name

Returns the name of this disk.

=over

=item Arguments

I<None>

=item Example

    $disk_name = $disk->name();

=back

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name}
}

=head2 ::state

Get or set the state of a disk (present, broken, spare, zeroing)

=over

=item Arguments

B<newstate> = I<spare|broken|removed|unfail>

State to set the disk to. If not set, just returns the current state of the disk.

=item Example

    printf( "Disk %s is %s\n", $disk->name, $disk->state );

=back

=cut

sub state {
    my ( $self, $newstate ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Invalid new state $newstate"
      if defined $newstte && ( $newstate !~ /(spare|broken|removed|unfail)/ );

    my $old_state = $self->status->{state};
    if ( defined $newstate && $old_state !~ /$new_state/ ) {

        $self->modify( state => $newstate );
    }
    return $old_state;
}

=head2 ::status

Get the status of a disk

=over

=item Arguments

=item Example

=back

=cut

sub status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $r = $self->server->storage_disk_show(
        disk => $self->name,
        %args,
    )->[0];
    return $r;
}

=head2 ::modify

Modify the properties of a disk.

=over

=item Arguments

=item Example

=back

=cut

sub modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->cli( cmd => "set -confirmation off" );
    $self->server->storage_disk_modify(
        disk => $self->name,
        %args,
    );
}

sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_spare_me} = 1;
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    return if $self->{_spare_me};
    return if ( NOTE->ALL_SPARED );

    $self->{_destroyed} = 1;

    # Return disk to initial state
    if ( $self->{_init_state} ne "broken" && $self->status->{state} eq "broken" ) {

        # Attempt to restore disk only if it is 'broken' now and was not at the begining
        NOTE::logcomment( "Attempt to restore disk " . $self->name );
        eval {
            $self->state("unfail");    # Ignore any errors
        };
    }
}

1;
