#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Job.pm#1 $
#
## @summary  GNOTE GX job module
## @author   dl-dpgqa-auto@netapp.com
## @status   Public
## @pod      here

=head1 NAME

NOTE::GX::Job

=head1 DESCRIPTION

This module provides job related support to Cluster ONTAP

=head1 METHODS

=cut

package NOTE::GX::Job;

use warnings;
no warnings 'redefine';

use GNOTE::GX;
use Time::Local;
use Data::Dumper;

our @ISA = qw(NOTE::GX);
use Server;
use Carp qw( croak confess );
use re 'eval';

my %_JOB_LIST = ();

=head2 ::new

Creates the Job object to administer and interact with ONTAP for job relation 
issues.
It is important to understand that the job object does not automagically
track the current job id. The internal _id and _job_id are populated by
specific calls. Of special note for SnapMirror users is the fact that some,
but not all, snapmirror command calls populate the _job_id with the current
snapmirror job.

=over

=item Arguments

B<Cluster> => I<GX::Cluster>

A GX::Cluster object to indicate where the jobs will be located. (Jobs are 
cluster scoped.)

B<id> => I<integer>

I<Optional>

A job id. If not supplied implies that the returned Job object will be a 
generic Job object for the supplied Cluster.

=item Examples

    my $job = NOTE::GX::Job(Cluster=>$cluster);

=back

=cut

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $self = {};

    $self->{_cluster} = delete $args{Cluster};
    croak "cluster not specified" unless ( $self->{_cluster} );

    $self->{_id} = delete $args{id}
      if exists $args{id};

    return $_JOB_LIST{ _job_key($self) }
      if exists $_JOB_LIST{ _job_key($self) };

    $self->{_owner} = $$;

    bless $self, $class;
    NOTE::reference($self);

    $_JOB_LIST{ $self->_job_key } = $self;

    return $self;
}
###############################################################################
#
#   Sub Name:       _job_key
#
#   Description:    Used by the new method to keep track of all the job
#                   objects created. Jobs are cluster scoped.
#                   
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#
#   Globals:        -
#
#   Environment:    
#
#   Returns:        Success: key    
#                   Failure: -   
#
###############################################################################
sub _job_key {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $key = $self->{_cluster};
    $key .= ":" . $self->{_id} if exists $self->{_id};
    return $key;
}

=head2 ::schedule_list

Returns the schedule(s) present in the system, based on the arguments supplied.
 Commonly, is called with the I<name> parameter.

The ONTAP command executed is I<job schedule cron show>.

Returns a standard 'cron' like columns: minute, hour, month, dayofweek etc.

=over

=item Arguments

=item I<Optional>

B<name> => I<string, schedule name>

I<See other options for job schedule cron.>

       "month" => 
       "dayofweek" => 
       "day" => 
       "hour" => 
       "minute" => 
       "description" =>

=back

=over

=item Returns

=back

=over

=item B<Success:> ARRAY of HASHREFs corresponding to the schedule requested.

=item B<Failure:> Confess when command execution fails.

=back

=over 

=item Examples

=back

    # Returns job schedule details for schedule 'test'
    my $test = $job->schedule_list(name => "test"); 
    # Returns an array of job schedules in the cluster
    my $list = $job->schedule_list(); 
    # Return array of job schedules that is set to every 5 mins past the hour
    my $list = $job->schedule_list(minute=>5); 

=cut

sub schedule_list {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $sch;
    eval { $sch = $self->server->job_schedule_cron_show(%args); };
    confess "Failed to obtain schedule information. Error: $@"
      if $@;
    return $sch;
}

=head2 ::schedule_add

Add a job schedule. If I<name> is given a unique name will
be created. Once the schedule is added, the name of the schedule is returned.

The other parameters of this method are used to setup a cron job, such
as 'month', 'dayofweek', 'day' 'hour' and 'minute' can be set.

Command executed is I<job schedule cron create>.

=over

=item Arguments

B<minute> => I<string>

I<Required>

A string containing a comma separated minutes that the schedule applies.

B<name> => I<string>

I<Optional>

Name of the schedule. If not set, a unique name will be given for the schedule.

B<month> => I<string>

I<Optional>

A string containing a comma separated list of months that the schedule applies.

B<dayofweek> => I<string>

I<Optional>

A string containing a comma separated list of days of the week that the schedule applies.

B<day> => I<string>

I<Optional>

A string containing a comma separated list of days of the month that the schedule applies.

B<hour> => I<string>

I<Optional>

A string containing a comma sepatated list of hours that the schedule applies.

=back

=over

=item Returns

=back

=over

=item B<Success:> The name of the created schedule is returned.

=item B<Failure:> Confess when command execution fails.

=back

=over

=item Examples


   my $sch_name = $job->schedule_add(month=>"march, june, september, december", minute=>"5");

=back

=cut

sub schedule_add {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{name} = NOTE->newName("sc") unless defined $args{name};
    $args{"-Verification"} = undef;    # Turn off Server.pm verfication as it is case sensitive, e.g. ONTAP translates 'march' to 'March' and verification fails. Hence the need to disable it.
    eval { $self->server->job_schedule_cron_create(%args); };
    if ($@) {
        confess "Failed to add job schedule " . $args{name} . ". Error: $@";
    }
    else {
        $self->{_schedule}{ $args{name} } = 1;
    }
    return $args{name};
}

=head2 ::schedule_delete

Delete a cron schedule.

=over

=item Arguments

B<name> = I<string>

I<Required>

The name of the job schedule to remove.

=back

=over

=item Returns

=back

=over

=item B<Success:> 1

=item B<Failure:> Confess when command execution fails.

=back

=over

=item Notes

You cannot delete any cron schedule that is associated with a job.

=back

=over

=item Examples

    $job->schedule_delete("test");

=back

=cut

sub schedule_delete {
    my ( $self, $name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Missing name parameter..."
      unless defined $name;
    NOTE::logcomment("Deleting job schedule $name:");
    eval { $self->server->job_schedule_cron_delete( name => $name ); };
    if ($@) {
        croak "Failed to remove job schedule $name. Error: $@";
    }
    else {
        delete $self->{_schedule}{$name}
          if exists $self->{_schedule}{$name};
    }
    return 1;
}

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_cluster}->server
}

=head2 ::get_id

Extracts the job id from text returned by ONTAP when it accepts a job.

=over

=item Arguments

B<text> = I<string>

The text returned by ONTAP after starting a job.

Example of the usual output:

I<B<[Job 809]> Job is queued: snapmirror create the relationship...>

=back

=over

=item B<Returns>

=back

=over

=item B<Success:> Job ID 

=item B<Failure:> Confesses if cannot parse the supplied output.

=back

=over

=item Example

    my $id = $job->get_id($ontap_text);

=back

=cut

sub get_id {
    my ( $self, $text ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment("Determining job id from: $text");
    if ( $text =~ /.*\[job (\d+)/i ) {
        my $id = $1;
        NOTE::logcomment("Job id: $id");
        return $id;
    }
    elsif ( $text =~ /^\d+$/ ) {

        # If it is only a number then assume it is just the job id
        NOTE::logcomment("Job Id: $text");
        return $text;
    }
    else {
        confess "Unable to determine job id from: $text";
    }
}

=head2 ::column_wait

Given a job id, will wait until the indicated column/field returns the specified string.
Since the state of a job cannot be obtained after it is complete
the typical use of this method is to verify the 'running' state.

Croaks if it cannot find job id.

=over

=item Arguments

B<id> => I<integer>

Job id of job to wait for.

B<check> => I<string>

Column name to match B<expected> string to. Possible interesting columns
of interest are: state, progress, description.

Default: I<state>

B<expected> => I<string>

Value to match the B<check> column output against. This is regular expression. See examples section for RegEx samples.

Default: I<success>

B<timeout> => I<integer>

Timeout in seconds. Default: 300

B<interval> => I<integer>

Polling interval in seconds. Default: 1

B<return> => I<string>

Default: I<state>

=back

=over

=item Returns

=back

=over

=item B<Success:> Returns the value of the <return> column. Returns a hashref of status columns if 'all' is specified.

=item B<Failure:> Confesses if job id not supplied, or does not see desired value or value is failure 
or times out before seeing desired value.

=back

=over 

=item Examples

=back

    $job->wait_column(id => 5, check => "state", timeout => 30, interval => 5); # Returns 'state' column
    my $tstaus_hashrec = $job->wait_column(id => 6, return => 'progress'); # Returns status progress column when job has completed

    $job->wait_column(id => 5, check => "state", expected => '(?i:(running|success))'); # Matching case insensitive  'running' or 'success'

=cut

sub column_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $id      = delete $args{id} || $self->{_id};
    my $exp_col = delete $args{check} || "state";
    my $exp_val = delete $args{expected} || "success";
    my $ret_col = delete $args{return} || "state";

    confess "No Job id supplied" unless defined $id;

    # Set low timeout if wating for 'running' state otherwise use longer timeout
    my $timeout = delete $args{timeout} || ( $exp_col =~ /state/i && $exp_val =~ /running/i ? 60 : 900 );
    my $interval = delete $args{interval} || 1;

    my $status;
    do {
        $timeout -= $interval;
        Tharn::snooze($interval);
        my @s = $self->server->job_show( id => $id );
        $status = $s[0];
        if ( $status->{state} =~ /failure/i ) {
            confess "Job " . $status->{id} . " failed. State: " . $status->{state} . ", Execution Progress: " . $status->{'progress'};
        }
        else {
            NOTE::logcomment( "Job "
                  . $status->{id}
                  . ", State "
                  . $status->{state}
                  . ( $exp_col =~ /state/            ? "" : ", $exp_col: " . $status->{$exp_col} )
                  . ( $ret_col =~ qr/(state|$exp_col)/i ? "" : ", Returning column $ret_col: " . $status->{$ret_col} ) );
        }
        return ( $ret_col =~ /all/i ? $status : $status->{$ret_col} )
          if $status->{$exp_col} =~ qr/$exp_val/i;

        confess "Job $id completed without setting column \"$exp_col\" to: $exp_val. Current state is $status->{'state'}"
          if $status->{'state'} =~ /success/i;
    } while $timeout > 0;
    confess "Timeout waiting for: Job $id, Column $exp_col, Value $exp_val";
}

=head2 ::wait

Given a job id, will wait until the job reaches a certain state.
Since the state of a job cannot be obtained after it is complete
the typical use of this method is to verify the 'running' state.

=over

=item Arguments

B<id> => I<integer>

Job id of job to wait for.

B<state> => I<string>]

State to wait for. Default: success

B<timeout> => I<integer>

Timeout in seconds. Default: 300

B<interval> => I<integer>

Polling interval in seconds. Default: 1

B<column> => I<string>

Job status column to return. Default: state

=item Returns

Calls column_wait without eval. See column_wait.

=back

=over

=item Examples

=back

    $job->wait(id => 5, state => "running", timeout => 30, interval => 5); # Returns 'state' column
    $my $tstaus_hashrec = $job->wait(id => 6, column => 'all'); # Returns status hashref

=cut

sub wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts = %args;
    $opts{check}    = "state";
    $opts{expected} = delete $args{state} || "success";
    $opts{return}   = delete $args{column} || "all";
    return NOTE::GX::Job::column_wait( $self, %opts );
}

=head2 ::status

Returns a single status for job show or job history show. If a job id
was passed to the object constructor or is passed into this method, then
the status of the indicated job will be returned otherwise the call
will error out.

=over

=item Arguments

I<Optional>

B<id> => I<integer>

Job id of the job that status is required.

B<history> => I<boolean>

Returns result from 'job history show' rather then 'job
show'. Default: 0

=item B<Returns>

=back

=over

=item B<Success:> Hash reference to the show fields for job id. 

I<Updates internal data for status.>

=item B<Failure:> 

B<Confess> on command failure;

B<0> if no entry found for the job id used/provided. Internal data for
status not changed.

=back

=over

=item B<Notes>

If you are using job history show, you potentially will get numberous
entries for your id, but only the most recent event for that ID will
be returned to you.

You need to check your return value, especially if filtering with
fields and values.

=back

=over

=item Examples

    # Return status for job id 2
    my $jobx = $job->status(id=>2);

    # Create a job object and instantiate with job id 2.
    my $job = NOTE::GX::Job(Cluster=>$cluster, id=> 2);
    # Return job show status for job id 2
    my $jobx = $job->status();

    # returns last entry for job history show id 2
    my $jobx = $job->status(history=>1);

    # returns job show for job id 9
    my $jobx = $job->status(id=> 9);

=back

=cut

sub status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer(); # log internal metrics

    my @jobs;

    my $history = delete $args{history};
    my $cmd = ($history ? 'job_history_show' : 'job_show');
    $args{id} = $self->{_id} unless exists $args{id};
    eval { @jobs = $self->server->$cmd(%args); };
    if ($@) {
        if ( $@ =~ /no entries matching your query/ ) {
            return 0;
        } else {
            confess $@;
        }
    }
    $self->{_status} = $jobs[0];
    return $self->{_status};
}
=head2 ::show

Returns a pointer to an array of hashes for the entries of a job show
or job history show, filtered based on the user supplied arguments. No
internal data structures are updated by this method. 

=over

=item Arguments

I<Optional>

B<history> => I<boolean>

Returns result of 'job history show' rather then 'job show'. Default: 0

=item Returns

=back

=over

=item B<Success:>  Reference to an array of hashes each detailing a job entry
(see the cdef for job_show/job_history_show).

=item B<Failure:> 

B<Confess> on command failure;

B<0> if no entries found matching the arguments provided.

=item B<Notes>

Be aware that asking for job history show, even if filtered can
results and 1000's of entries.

Always check your return value.

Be aware that supplying invalid arguments results in no arguments
passed, thus getting all the entries for job [history] show.


=item Examples

$jobs_array_ref = $job->show();

$jobs_array_ref = $job->show(history=>1);

$jobs_array_ref = $job->show(name=> '"CLUSTER BACKUP AUTO weekly"');

# Sample using some arbitrary filtering.

$jobs_array_ref = $job->show(history=>1, 
                             name=>'"SnapMirrorDaemon*"',
                             'code'=>'1258185');

=back

=cut 

# This method was added since the implementation of the method
# status failed to allow for a bulk show and rather than risk breaking
# scripts, added this method to serve that purpose.

sub show {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer(); # log internal metrics

    my @jobs;

    my $history = delete $args{history};
    my $cmd = ($history ? 'job_history_show' : 'job_show');

    my $ret_val = eval { @jobs = $self->server->$cmd(%args); 1;};

    if (!$ret_val) {
        if ( $@ =~ /no entries matching your query/ ) {
            return 0;
        } else {
            confess $@;
        }
    }
    return \@jobs;
}

=head2 ::action

Invoke certain actions on a NOTE Job object. If no id is given then the internal id is used.

=over

=item Arguments

B<action> = I<string>

I<Required>

Valid actions are: pause, resume, stop or delete

I<object> => I<NOTE object>

A NOTE object with a defined running job.

=back

=over

=item Returns

=back

=over

=item B<Success:> No explicit return.

=item B<Failure:> Confesses if called with invalid or inadequate parameters; if job action
call fails, if the action waited for does not result in reaching desired condition. 

=back

=over

=item Notes:

If action is delete or stop, removes the job id datum from the calling object.

=back

=over

=item Example

      $job->action("stop", object=>$vol);

=back

=cut

sub action {
    my ( $self, $action, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Invalid action: $action. Must be one of: pause, resume, stop or delete"
      unless $action =~ /^(pause|resume|stop|delete)$/;

    confess "Missing object..."
      unless exists $args{object};

    confess $args{object}->description . ": No job id"
      unless exists $args{object}->{_job_id};

    my $cmd = "job_" . $action;

    my $id = $args{object}->{_job_id};
    my $output;
    my $title = $args{object}->description . ", Job $id" . ", Action $action";
    eval { $output = $self->server->$cmd( id => $id ); };
    confess "$title, failed to start. Error: $@"
      if $@;

    eval { $self->wait( id => $id, state => "failure" ); };
    if ($@) {
        if ( $action =~ /stop/ ) {
            confess "$title, failed to stop job. Error: $@"
              unless $@ =~ /(stopped|aborted)/;
        }
        elsif ( $action =~ /delete/ ) {
            confess "$title, failed to delete job. Error: $@"
              unless $@ =~ /deleted/;
        }
        else {
            confess "$title. Unexpected failed. Error: $@";
        }
    }
    delete $args{object}->{_job_id}
      if $action =~ /(stop|delete)/;
}

=head2 ::next_schedule_time

Returns next UTC epoch time for schedule identified by name. (Number of seconds to next schedule since January 1, 1970 UTC time zone)

=over

=item Arguments

B<name> = I<string>

I<Required>

Schedule name

=back

=over

=item Returns

=back

=over

=item B<Success:> Epoch time of the next occurance of I<schedule_name>.

=item B<Failure:> undef

=back

=over

=item Note:

Uses B<schedule_list>.

=back

=over

=item Example

      $epoch_time = $job->next_schedule_time(name=>$snapmirror_schedule_name);
      $seconds_to_update = $epoch_time - $Node_Obj->time;
      logcomment( "Next update will take place at:", POSIX::strftime("%C\n", $epoch_time);
      sleep( $seconds_to_update );  # wait for update to happen

=back

=cut

sub next_schedule_time {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    croak "Missing required schedule name parameter: must call as next_schedule_time( name=>'schedule_tag' )" unless( $args{name} );

    my $node_epoch_time = $self->{_cluster}->Node->time;				    # get filer's UTC (GMT) timestamp in seconds
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($node_epoch_time); # and convert it to something more human

    my $entry = ($self->schedule_list( %args ))->[0];					    # schedule list returns pointer to an array of
											    # hash references even if there is only one entry

    my %str2num = ( 'Sunday' => 0, 'Monday' => 1, 'Tuesday' => 2, 'Wednesday' => 3,         # Months and weekdays can be defined both as
                    'Thursday' => 4, 'Friday' => 5, 'Saturday' => 6, 'January' =>0,         # numerical value or as string, since all
                    'February' =>1, 'March' =>2, 'April' =>3, 'May' =>4,                    # time functions work only with numerical values
                    'June' =>5, 'July' =>6, 'August' =>7, 'September' =>8,                  # string values must be converted to numbers
                    'October' =>9, 'November' =>10, 'December' =>11 );

    foreach( keys %$entry ) {                                                               # each elemen of cron job is SCALAR or ARRAY
        $$entry{$_} = ref( \$$entry{$_} ) eq 'SCALAR' ? [ $$entry{$_} ] : $$entry{$_};      # force it to be an ARRAY for foreach loop to work
        @{$$entry{$_}} = map{defined $str2num{$_} ? $str2num{$_} : $_ } @{$$entry{$_}};     # and replace string names by numbers for
    }                                                                                       # time functions to work properly

    foreach my $yr_loop ( $year, $year + 1 ) {                                                          # try to leap forward to find next available time. In cases
        foreach my $cron_month ( @{$$entry{'month'}} ) {                                                # where we run every 10 minutes we have to account Dec 31st
            $mon = $cron_month if $cron_month =~ /\d+/;                                                 # and include next year in calculation.
            foreach my $cron_day ( @{$$entry{'day'}} ) {                                                # loop through each element (month, day, hour etc...)
                $mday = $cron_day if $cron_day =~ /\d+/;                                                # and if there is valid type and value within cron
                foreach( @{$$entry{'hour'}} ) {                                                         # then replace epoch entry with one found in cron job
                    $hour = $_ if /\d+/;
                    foreach( @{$$entry{'minute'}} ) {
                        $min = $_ if /\d+/;
                        my $newepoch = timelocal(0,$min,$hour,$mday,$mon,$yr_loop);                     # calculate new epoch when all replacements are done

                        foreach( @{$$entry{'dayofweek'}} ) {                                            # if day of week is also present, then find next
                            if( /\d/i ) {                                                               # appropriate week day after newly calculated
                                my $days = $_ - ( localtime($newepoch) )[6];                            # epoch time. Calculation of weekday difference
                                $newepoch += ( $days < 0 ? $days + 7 : $days ) * 86400;                 # is rounded to zero hours in order to preserve just
                                my( $new_month,$new_day ) = (localtime($newepoch))[4,3];                # just calculated hours and minutes in new epoch then
                                $newepoch=undef if($cron_month =~ /\d/ && $new_month ne $cron_month);   # then weekdays difference is added to new epoch time
                                $newepoch=undef if($cron_day =~ /\d/ && $new_day ne $cron_day);         # however if month and day was also specified explicitly
                            }                                                                           # then date may roll out of original day/month boundaries
                                                                                                        # so appropriate check must be present to prevent it

                            return $newepoch if( $newepoch > $node_epoch_time );                        # check if calculated epoch is greater then original since
                        }                                                                               # it's possible for cron to point to past tense
                    }
                }
            }
        }
    }
    return undef;
}

=head2 ::completion

Returns to completion field from job. Will check 'job show' or 'job history show'
to get this value. Will abort if job is not found.

=over

=item Arguments

B<id> => I<integer>

Job id to query. Default: job id created at instaniation

=item Examples

=back

    my $txt = $job->completion(id=>5);

=cut

sub completion {
	my ($self, %args) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	return $self->_get_column(%args);
}

# Internal method to return a job column by trying 'job show' followed by 'job history show'

sub _get_column {
	my ($self, %args) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $id = (exists $args{id} ? delete $args{id} : $self->{_job_id});
	confess "No job id found"
		unless $id;
	confess "Missing column name"
		unless exists $args{column};

	my $result;
	foreach my $history (0, 1) {
		eval { $result = $self->status(id=>$id, history=>$history); };
		return $result->{$args{column}}
			if $result && exists $result->{$args{column}};
	}
	confess "Unable to obtain column ".$args{column}." for job $id";
}

=head2 ::show_node

Returns the node name on which the job is currently running 

=over

=item Arguments

B<object> => I<object>

snapmirror or volume object where job has been triggered 

B<id> => I<integer>

Job id to query. 

=item Examples

=back

    my $node_name = $job->show_node(id=>5);

    $sm->initialize(background => 1);
    my $node_name = $job->show_node(object => $sm);

=cut

sub show_node{
	my ($self, %args) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my @s;
	confess " object and job id passed, please pass either object or job id "
	    if ( exists $args{object} && $args{id} );

 	if ( exists $args{object} ) {	# check to see if the user passed a object
   	 confess $args{object}->description . ": No job id"
      		unless exists $args{object}->{_job_id};	# if the object passed does not have an id confess
	   $self->{_job_id} = $args{object}->{_job_id};	# else set the internal job id from object
	} else {
	   $self->{_job_id} = $args{id};	# else set the internal job id from object
	}

        eval{@s = $self->server->job_show( id => $self->{_job_id});};
	   confess " job id not found : $@ " if $@;
        my $status = $s[0];
        return $status->{node} ;
}
=head2 ::affinity

Get the value for I<affinity> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<affinity> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub affinity     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{affinity}
}

=head2 ::category

Get the value for I<category> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<category> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut


sub category     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{category}
}

=head2 ::code

Get the value for I<code> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<code> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut


sub code         {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{code}
}

=head2 ::description

Get the value for I<description> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<description> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut


sub description  {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{description}
}

=head2 ::dropdeadtime

Get the value for I<dropdeadtime> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<dropdeadtime> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut


sub dropdeadtime {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{dropdeadtime}
}

=head2 ::endtime

Get the value for I<endtime> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<endtime> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut


sub endtime      {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{endtime}
}

=head2 ::filer

Get the value for I<filer> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<filer> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut


sub filer        {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{filer}
}

=head2 ::id

Get the value for I<id> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<id> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub id           {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{id}
}

=head2 ::jobtype

Get the value for I<jobtype> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<jobtype> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub jobtype      {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{jobtype}
}

=head2 ::name

Get the value for I<name> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<name> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub name         {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{name}
}

=head2 ::node

Get the value for I<node> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<node> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub node         {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{node}
}

=head2 ::priority

Get the value for I<priority> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<priority> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub priority     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{priority}
}

=head2 ::progress

Get the value for I<progress> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<progress> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub progress     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{progress};
}

=head2 ::queuetime

Get the value for I<queuetime> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<queuetime> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub queuetime    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{queuetime}
}

=head2 ::restarted

Get the value for I<restarted> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<restarted> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub restarted    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{restarted}
}

=head2 ::schedule

Get the value for I<schedule> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<schedule> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub schedule     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{schedule}
}

=head2 ::starttime

Get the value for I<starttime> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<starttime> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub starttime    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{starttime}
}

=head2 ::state

Get the value for I<state> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<state> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub state        {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{state}
}

=head2 ::uuid

Get the value for I<uuid> from the job show for calling object's job id.

=over

=item Arguments

I<None>

=item Returns

=back

=over

=item B<Success:> Returns job show value for I<uuid> for the caller's job id.

=item B<Failure:> Fails if Object does not know its job id. See sub I<status>.

=back

=cut

sub uuid         {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->status->{uuid}
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    $self->{_destroyed} = 1;
    return if ( NOTE->ALL_SPARED );

    foreach my $name ( keys %{ $self->{_schedule} } ) {
        eval {

            # Ignore errors
            $self->schedule_delete($name);
        };
    }
}

1;

