#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Mirror.pm#1 $
#
## @summary  GNOTE GX Mirror module
## @author   dl-dpgqa-automation@netapp.com
## @status   Private
## @pod      here
package NOTE::GX::Mirror;
use GNOTE;
use GNOTE::GX;
our @ISA = qw(NOTE::GX);
use Carp qw(croak);

sub new {
    my ($class, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = NOTE::GX::SnapMirror::_set_src_dst (%args);

    croak "Destination object must not exist already"
	if ( $self->{_dst}->substant );

    $self->{_owner} = $$;
    $self->{_destroyed} = 1;

    $self->{_args}->{type}     = $args{type}     || NOTE::option( $args->{type} );
    $self->{_args}->{protocol} = $args{protocol} || NOTE::option( $args->{protocol} );
    $self->{_args}->{timeout}  = $args{timeout}  || NOTE::option( $args->{timeout} );

    bless $self, $class;
    NOTE::reference( $self );

    return $self;
}

sub create {
    my ($class, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = __PACKAGE__->new( %args );

    NOTE::logcomment( "Creating mirror: ".$self->description );
    eval {
	$self->{_src}->server->volume_mirror_new(
            vserver   => $self->{_src}->Vserver->name,
	    source    => $self->{_src}->name,
	    mirror    => $self->{_dst}->name,
	    aggregate => $self->{_dst}->Aggregate->name,
	    type      => delete $self->{_args}->{type} || 'DR',
	    protocol  => delete $self->{_args}->{protocol} || "LSM",
	    Timeout   => delete $self->{_args}->{timeout} || 60,
	    %{$self->{_args}},
	); 
    };
    croak $@ if $@;

    $self->is_destroyed = 0; # substant now

    unless ( $self->dst->exists ) {
	croak( "Mirror created but Destination volume doesn't exists\n" );
    }
    $self->dst->is_destroyed = 0; # Make destination volume substant.
    # Don't auto-destroy, because this volume is destroyed when the mirror is deleted.
    $self->dst->is_anonymous = 0;
    return $self;
}

sub replicate {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $background   = delete $args{background};
    my $timeout      = delete $args{timeout};

    NOTE::logcomment( "Replicating ".$self->{_src}->description );
    eval {
	$self->{_src}->server->volume_mirror_replicate(
            vserver    => $self->{_src}->Vserver->name,
	    source     => $self->{_src}->name,
	    mirror     => $self->{_dst}->name,
	    foreground => $background ? 'false' : 'true',
	    Timeout    => $timeout || 7200,
	    %{$self->{_args}}, # remaining args specified during create
	    %args,             # args specified directly
	)
    };
    croak $@ if $@;
}

# For convenience; creates a mirror and replicates it..
# Note that both create and replicate arguments are specified
# here. The call to replicate doesn't need any arguments because
# they have been stored in the object. I don't particularly like
# this; hopefully it can go away when Paloma converts to a
# SnapMirror interface..
sub initialize {
    my ($class, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = __PACKAGE__->create( %args );
    $self->replicate;
    return $self;
}

sub description {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $desc = "";

    $desc .= "\n    +---";
    $desc .= $self->src->description;
    $desc .= "\n    +-> ";
    $desc .= $self->dst->description;
    return $desc;
}

sub delete {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->destroy
}

sub src {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_src}
}
sub dst {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_dst}
}

=head2 ::exists

Returns true if this object represents a mirror that
physically exists.

=over

=item Argument

None.

=item Example

=back

     die unless ( $mirror->exists );

=cut

sub exists {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    eval {
	$self->src->server->volume_mirror_show(
            vserver => $self->src->Vserver->name,
	    source  => $self->src->name,
	    mirror  => $self->dst->name,
	); 
    };
    return $@ ? 0 : 1;
}

=head2 ::substant

Returns true if this object represents a volume that
might exist. This is different than the ::exists method
because it doesn't actually query the server, but rather
returns the state of an internal flag that indicates
whether the volume has been created yet.

=over

=item Argument

None.

=item Example

=back

     $vol = NOTE::Volume->new;
     die if ( $vol->substant );
     printf( "Volume size will be %dk\n", $vol->size );

=cut

sub substant {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return !shift->is_destroyed
}

=head2 ::is_destroyed

Returns true if this mirror is destroyed.  This can also be
assigned to, in order to make a mirror 'un-destroyed',
but this feature is intended for internal use only.

=over

=item Argument

None.

=item Example

=back

     if ( $mirror->is_destroyed ) {
         ...
     }

=cut

sub is_destroyed :lvalue {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Must do this trickery due to a bug in perl regarding
    # tied hashes and lvalues.
    my $r = \$self->{_destroyed};
    $$r;
};

sub destroy {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Destroying Mirror " . $self->description );
    $self->is_destroyed = 1;
    eval {
	$self->{_src}->server->volume_mirror_delete(
            vserver => $self->src->Vserver->name,
	    source  => $self->src->name,
	    mirror  => $self->dst->name,
	); 
    };
    croak $@ if $@;
    unless ( $self->dst->exists ) {
	# Ugly, but necessary, since deleting mirror destroys destination:
	$self->dst->is_destroyed = 1;
    }
    return $self;
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    return if ( NOTE->ALL_SPARED );
    if ( $self->{_spare_me} ) {
    	NOTE::logcomment( "Mirror ".$self->description." has been spared" );
	return;
    }
    $self->destroy;
}

1;
