#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Node.pm#1 $
#
## @summary  GNOTE GX node module
## @author   dl-dpgqa-automation@netapp.com
## @status   Public
## @pod      here

=head1 NAME

NOTE::GX::Node

=head1 DESCRIPTION

This module represents a GX Node within a cluster.
This module has a constructor "connect" which returns a GX::Node 
object after connecting to the Node.
This module contains methods performing various operations on node. 

=head1 METHODS

=cut

package NOTE::GX::Node;

use strict;
use warnings;
no warnings 'redefine';

use GNOTE::GX;
our @ISA = qw(NOTE::GX);

use GNOTE::GX::Cluster;
use Options;
use GNOTE::GX::VIF;
use GNOTE::GX::Port;
use NACL::APISet::Version;
use Net::Ping;
use Time::Local;
use Tharn;
use Carp qw( carp croak confess );
use Data::Dumper;
use Time::ParseDate;

my %_FILER_CONNS = ();
my $current_node = ".*";

=head2 attach

This is a constructor for connecting to the node and return the 
node object. This function does not do anything other than calling 
the other constructor 'connect'. See the documentation of 'connect'.
It is better calling the constructor 'connect' than calling 
this 'attach'.

=over

=back

    #Connecting to the node "dpg-gx5", by passing the node name as a scalar
    my $node_obj = NOTE::GX::Node->attach("dpg-gx5");

    #Connecting to the node "dpg-gx5", by passing the arguments as hash
    my $node_obj = NOTE::GX::Node->attach(name => "dpg-gx5");

=cut


sub attach {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->connect(@_)
}

=head2 connect

This is the constructor method for the NOTE::GX::Node module.
This constructor upon being called, connects and returns
a NOTE::GX::Node object representing the ONTAP node.
The returned node object can be used to call different methods
that are implemented in this module which perform vaious
operations on the node represented by this object.
After connecting this constructor also prints the 'version'
odf node and 'sysconfig -v' output of node.

=over

=item Arguments

The arguments can be passed as an hash or as a scalar.
The filer name can alone be passed, as single argument.
Or the node parameters can be passed as an hash.
Check the below examples section, for both these versions.

In case of passing the node/filer name as a scalar:

=over

=item B<nodename>

nodename as a string

=back

In case of passing as an hash:
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<name> => I<string> nodename

If no this param is not provided, the environmental variable $FILER value
is considered as the node name.
If $FILER is not defined, then the first node name from the value of
$NODES is taken as the name of the node and connect to that.

=item B<direct> => I<0|1>

Parameter 'direct' is for backward compatibility
Which enables scripts from using SnapMirrorUtil
from using direct execution of commands of this node

=item B<teardown> => I<teardown>

code reference or hash reference. the hash reference can contain
'pre' or 'post' keys with values being code reference.
this code gets executed before the DESTROY function call.
If both 'pre' and 'post' code references are provided, 'pre' code
reference will be executed before 'DESTROY' and 'post' code reference
gets executed after DESTROY function call. All these will be
done as part of NOTE::teardown section.

=back

=item Returns

=over

=item B<Success:>

NOTE::GX::Node object reference for the node

=item B<Failure:>

confesses, if name of node is provided as argument and
$FILER and $NODES are not defined.

=back


=item Notes

=over

=item 1

This function prints the version of ONTAP running on the node
and 'sysconfig -v' output, which would be helpful for debugging
later on.

=back

=item Examples

=back

	#Connecting to the node "dpg-gx5", by passing the node name as a scalar
    my $node_obj = NOTE::GX::Node->connect("dpg-gx5");

    #Connecting to the node "dpg-gx5", by passing the arguments as hash
    my $node_obj = NOTE::GX::Node->connect(name => "dpg-gx5");

=cut

sub connect {
    my $class = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my %args;
    my $self = {};

    # Allow for single argument to specify node name:
    if ( scalar @_ == 1 ) {
        $args{name} = shift;
    }
    else {
        %args = @_;
    }

    # Parameter 'direct' is for backward compatibility
    # Which enables scripts from using SnapMirrorUtil
    # from using direct execution of commands of this node
    if ( ( delete $args{direct} ) && ( $current_node eq ".*" ) ) {
        NOTE::logcomment("Enabled direct node command execution");
        $current_node = "Not a valid node";
    }

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    $self->{_tag}      = delete $args{tag};
    $self->{_teardown} = delete $args{teardown};

    $args{name} ||= NOTE::option( '$FILER' );
    my $nodes_param = NOTE::option( '$NODES' ) ;

    if ( defined($nodes_param) and ( !$args{name} or $nodes_param =~ /$args{name}/) ) {
        my @nodeslist = split(/,/,$nodes_param);
        $self->{_server} = Server->new(%args);
        if ($args{name}) {
            $self->{_server}->add_nodes( nodes=> [$args{name}]);
            $self->{_name} = $args{name};
            $self->{_server}->set_execution_server(-Server => $args{name});
        } else {
            $self->{_name} = $nodeslist[0];
        }
    } else {
        $self->{_name} = $args{name}
            or croak "Host specification '" . $args{name} . "' doesn't resolve to anything";
        $self->{_server} = Server->new( nodes => $self->{_name}, %args, );
    }

    #as these might croak if the version can't be determined, we'll eval
    my $hostrec = Tharn::host($self->{_name});
    my $version_manager = NACL::APISet::Version->new(hostobj    => $hostrec,
                                                     category   => 'Node',
                                                     shell_type => 'CMode',
                                                 );
    my $filer_eval = eval {
        $self->{_wafl_version} = $version_manager->get_version_attribute(
                                                   attribute => 'wafl_version',
                                               );
        1;
    };
    if (!$filer_eval) {
        carp "Couldn't determine wafl version for $self->{_name}";
    } else {
        NOTE::logcomment(
            "wafl version of node $self->{_name}:$self->{_wafl_version}");
    }

    if ($args{get_version}) {
        $filer_eval = eval {
            $self->{_ontap_version} = $version_manager->get_version_attribute(
                                                       attribute => 'release',
                                                   );
            1;
        };
        if (!$filer_eval) {
            carp "Couldn't determine ontap version for $self->{_name}";
        } else {
            NOTE::logcomment(
                "ontap version of node $self->{_name}:$self->{_ontap_version}");
        }
    }

    # If there already was an object with this same connection object, return
    # a reference to it, rather than creating a new one. This is useful
    # because when this object gets destroyed, we know it is the last reference,
    # so we can print out status befre actually disconnecting.
    return $_FILER_CONNS{ $self->{_name} }
      if ( exists $_FILER_CONNS{ $self->{_name} } );

    NOTE::logcomment( "Connected to " . $self->{_name} );

    $self->{_stats}          = $args{stats};
    $self->{_do_zero_spares} = undef;
    $self->{_stampfile}      = undef;		# create empty file on FreeBSD, then find all files changed after _stampfile creation date

    bless $self, $class;
    NOTE::reference($self);

    $_FILER_CONNS{ $self->{_name} } = $self;

    $self->{_Cluster} = NOTE::GX::Cluster->connect( Node => $self );

    $self->{_original_privilege} = $self->privilege();	    # safe privelege level before executing any commands, restore in DESTROY
    eval { NOTE::logcomment( $self->version ) };	    # try to print version, but don't fail if ONTAP is changed or node is down

    # print the sysconfig of node, which would be of help in debugging
    my $sys_config_op;
    my $return_eval = eval { $sys_config_op = $self->sysconfig; 1;};
    if(!$return_eval){
        confess 'Error in getting output of sysconfig of node ' . $self->{_name} . ' :' . $@;
    }
    else {
        NOTE::logcomment( 'sysconfig -v output of node ' . $self->{_name} . ' :' . $sys_config_op );
    }

    eval {
	( my $path = $self->bsdshell(cmd=>"pwd") ) =~ s/\s+//g;							# get home directory path on FreeBSD box
	my $stampname = "$path/" . NOTE::newName();			                                        # create unique file name for stampfile
	$self->bsdshell(cmd=>"date > $stampname");		                                                # dump date info into stampfile
	( my $confirmed = $self->bsdshell(cmd=>"test -f $stampname && echo $stampname") ) =~ s/\s+//g;		# verify file creation, and store filename
	if( $stampname eq $confirmed ) {									# as part of object if file was actually
	    $self->{_stampfile} = $stampname;									# on FreeBSD. $stampname eq $confirmed
	}													# means file is created.
    };
    return $self;
}

=head2 reconnect

Re-establishes the connections to the node by first closing all 
connections and then opening them again.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<list> => I<string>

list of connections (as comma seperated) to this node, 
that are to be reconnected.
valid values are: console, shell or cli.
I<Default> : "shell,cli"

=back

=item Returns

=over

=item B<Success:> does not return anything

=item B<Failure:> no failure scenario (i.e. expected to be passed always)

=back


=item Notes

=over

=item 1 Also sets back the transfer engine, to the value prior to reconnect

=back

=item Examples

=back

    #reconnect
	$node->reconnect;

=cut

sub reconnect {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $list = $args{list} || "shell,cli";

    foreach my $type ( split ",", $list ) {
        my $conn_type = $type . "_connection";
        NOTE::logcomment( "Close $conn_type for " . $self->name );
        $self->server->handle_connections( close => [$conn_type] );
        NOTE::logcomment( "Open $conn_type for " . $self->name );
        $self->server->handle_connections( open => [$conn_type] );
    }

    # If we were using BRE as the transfer engine, we should set the
    # relation back to using BRE.
    if ( $self->{_BRE} ) {
       $self->transfer_engine( 'BRE' );
    }
}

=head2 ::tricky_parity

Determine from CLI parameter TRICKY_PARITY to determine if
Node is running in TRICKY_PARITY mode. In this mode certain
commands, such as creating a striped DP volume, snapmirror create
of striped volumes, snapmirror break and restore, needed to run
in test mode.

=over

=item Arguments

I<None>

=item Example

=back

    my $mode = $node->tricky_parity;

=cut

sub tricky_parity {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    unless ( exists $self->{_tricky_parity} ) {
        my $option = new Options;
        $option->parameters(
            definitions => [
                {
                    name        => 'TRICKY_PARITY',
                    description => 'Run in Tricky Parity mode',
                    default     => 1
                }
            ]
        );
        $option->process();
        my $hash = $option->get_hash();
        $self->{_tricky_parity} = $hash->{'TRICKY_PARITY'};
        NOTE::logcomment( "GX::Node " . $self->name . " runs in TRICKY_PARITY MODE" )
          if $self->{_tricky_parity};
    }
    return $self->{_tricky_parity};
}

=head2 default_password

Returns the default password for the node as set in the NATE host file.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> password as string

=item B<Failure:> does not fail

=back

=item Examples

=back

	my $node_pwd = $node->default_password;

=cut

sub default_password {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    unless ( exists $self->{_pwd} ) {
        my $hostrec = Tharn::host( $self->{_name} );
        $self->{_pwd} = $hostrec->default_password();
        NOTE::logcomment( "default password for " . $self->{_name} . " is " . $self->{_pwd} );
    }
    return $self->{_pwd};
}

=head2 reboot

Reboot the node and wait for the reboot to complete.
And reconnects to the node again.

=over

=item Arguments

B<None>

=item Example

=back

    $node->reboot;

=cut

sub reboot {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $srvr = $self->server;
    NOTE::logcomment( "Rebooting Node " . $self->name . ":" );
    $srvr->transit_to( state => $srvr->REBOOT );

    # This is needed in case AUTOBOOT is not set. If set, this is a no-op
    NOTE::logcomment( "Bring Node " . $self->name . " up to login prompt" );
    $srvr->transit_to( state => $srvr->LOGIN );
    NOTE::logcomment( "Node " . $self->name . " reboot completed" );

    # Reconnect after the reboot
    $self->reconnect();
}

=head2 panic

This function will cause the Node to panic and wait for the node 
to come back to normal prompt and reconnect. This functions waits for 
NUM_OF_RETRY_FOR_RECONNECT_AFTER_PANIC number of retries, after succesfully
panicked. The NUM_OF_RETRY_FOR_RECONNECT_AFTER_PANIC is equal to 5.
On failure to panic, confesses.

Caveat: Current implementation doesn't support tests running with GDB

=over

=item Arguments

B<None>

=item Returns

=over

=item B<Success:> Nothing

=item B<Failure:> confesses

=back

=item Examples

=back

    $node->panic();

=cut

use constant NUM_OF_RETRY_FOR_RECONNECT_AFTER_PANIC => 5;

sub panic {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $the_server    = $self->server;
    my $str_node_name = $self->name;

    eval {
        $the_server->shell( -Server => $str_node_name, cmd => "sudo sysctl debug.panic=1" );

        #	$the_server->set( privilege => "test", confirmations => "off" );
        #       $the_server->cli( cmd => "run -node $str_node_name -command panic" );
    };

    if ($@) {
        confess $@ unless ( $@ =~ /TIMEOUT/ );
    }

    my $flag_done    = 0;
    my $num_of_retry = 0;
    do {

        # This is needed in case AUTOBOOT is not set. If set, this is a no-op
        eval { $the_server->transit_to( state => $the_server->LOGIN ); };

        if ($@) {

            #
            # Panic in this situation is actually "normal", since it is
            # intentionally triggered by the method.
            #
            if ( $@ !~ /panicked/ ) {
                confess($@);
            }
        }
        else {
            $flag_done = 1;
        }
    } while ( ( $flag_done != 1 ) && ( $num_of_retry++ < NUM_OF_RETRY_FOR_RECONNECT_AFTER_PANIC ) );

    NOTE::logcomment( "Node " . $str_node_name . " panic completed" );

    # Reconnect after the reboot
    $self->reconnect();
}

=head2 core_dump

Enable or disable core dump for the node

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<state> => I<enable|disable>
Default: enable

=back

=item Returns

=over

=item B<Success:> Nothing

=item B<Failure:> confesses

=back

=item Examples

=back

    # Disable core dump before panic-ing the filer
    $node->core_dump(state=>"disable");
    $node->panic();

=cut

sub core_dump {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{state} = "enable" unless exists $args{state};
    confess "Invalid state. Must be enable or disable..."
      unless $args{state} =~ /(en|dis)able/;

    my @names;
    push @names, $self->name;
    eval { $self->server->core_dump_state( nodes => \@names, %args ); };
    confess "Failed to " . $args{state} . " core dump for node " . $self->name . ". Error: $@"
      if $@;
}

=head2 power_cycle

This function will cause the Node to perform a power cycle,
and then reconnect to the node 

=over

=item Arguments

B<None>

=item Returns

B<None>

=item Examples

=back

    $node->power_cycle();

=cut

sub power_cycle {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $the_server    = $self->server;
    my $str_node_name = $self->name;

    $the_server->rlm( -Server => $str_node_name, cmd => "system power cycle", -Timeout => 1000 );

    # This is needed in case AUTOBOOT is not set. If set, this is a no-op
    $the_server->transit_to( state => $the_server->LOGIN );
    NOTE::logcomment( "Node " . $str_node_name . " panic completed" );

    # Reconnect after the reboot
    $self->reconnect();
}

=head2 ::redirect

set, get or clear node to which commands will be redirected to execute from
instead of the current node. 

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<node> => I<GX::Node>

Node object: to set the redirection to that node.
undef      : to remove the redirection.
none       : to get the current redirect node

=back

=item Returns

Returns the node reference to which the redirection is set to.

=item Examples

=back

    # Redirects command for this node to be executed on dpg-gx4 instead
    $node->redirect(node=>$dpg_gx4_node);
    # Returns redirect node if set or undef if not
    my $redirect_node = $node->redirect();
    # Causes subsequent commands to be run on this node again
    $node->redirect(node=>undef);

=cut

sub redirect {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( exists $args{node} ) {
        if ( defined $args{node} ) {
            confess "Invalid redirect node parameter. NOTE::GX::Node object is needed..."
              unless ref( $args{node} ) =~ /NOTE::GX::Node/;
            $self->{_redirect_node} = $args{node};
        }
        else {
            delete $self->{_redirect_node};
        }
    }
    else {
        return $self->{_redirect_node};
    }
}

=head2 ::ha_configure

Gets or sets High Availability configuration for the Node.

=over

=item Arguments

I<Optional Arguments>

=over

=item B<flag> : I<boolean>

true/false: it sets to the given state
none      : does not do anything

=back

=item Returns

=over

=item B<Success:> 
Returns the old state, if set to the other value
Returns the existing state, if no value is provided

=item B<Failure:> croaks in case error in setting

=back

=item Examples

=back

    # Sets HA configuration for node
    $node->ha_configure(1);

    # Obtain the current HA mode for node
    my $is_ha = $nde->ha_configure();

=cut

sub ha_configure {
    my ( $self, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $output = $self->server->cluster_ha_show();
    my $old_state = ( $output->[0]->{configured} =~ /true/i );

    if ( defined $flag && $flag != $old_state ) {
        my $configured_flag = ( $flag ? "true" : "false" );
        NOTE::logcomment("Set HA configuration to $configured_flag");
        eval { $self->server->cli( cmd => "ha modify -configured $configured_flag" ); };
        croak "Failed to change HA configuration. Error: $@"
          if $@;
    }

    return $old_state;
}

=head2 sfo_enable

Gets or sets SFO enable flag.

=over

=item Arguments

I<Optional Arguments>

=over

=item B<flag> : boolean 

true/false: it sets to the given state
none      : does not do anything

=back

=item Returns

=over

=item B<Success:>
Returns the old state, if set to the other value
Returns the existing state, if no value is provided

=item B<Failure:> croaks in case error in getting the current state

=back

=item Examples

=back

    # Enable SFO for node
    $node->sfo_enable(1);

    # To get current SFO mode for node
    my $sfo_state = $node->sfo_enable();

=cut

sub sfo_enable {
    my ( $self, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Obtain current SFO state
    my ( $old_state, $current );
    eval { $current = $self->sfo_status("enabled"); };
    croak "Error determining current SFO state for " . $self->name . ". Error: $@"
      if $@;
    $old_state = ( $current =~ /true/i );

    # Set new state
    $self->sfo_modify( enabled => ( $flag ? "true" : "false" ) )
      if ( defined $flag && $flag != $old_state );

    # Return old SFO status
    return $old_state;
}

# Verify SFO enabled and set auto-giveback and enable HA if needed

sub _verify_sfo_ha {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $status = $self->sfo_status();
    croak "SFO not enabled for " . $self->name
      unless $status->{"enabled"} =~ /true/i;

    # Enable HA for Clusters of 2 or less nodes
    my @nodes = @{ $self->server->cluster_show() };
    $self->ha_configure(1)
      if @nodes < 3;

    return $status->{"partner-name"};
}

=head2 takeover

Performs takeover of partner node by this node. It first determines 
if the node is located on a cluster consisting of 3 or more nodes.
If not, it will enable High Availability prior to invoking
the takeover action.

The nodes typically are set up with a 5 mins auto-giveback period. 
If you have test that needs a longer period of takeover then you should 
either disable auto-giveback or set an appropriate interval.

This method will accept pass-through parameters to be send 
to sfo_monitor, such as timeout, interval etc

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<background> => I<boolean>
If set, method will return immediately after starting 'takeover' of it's partner. Default: 0

=item B<timeout> => I<timeout>
pass through parameter for server cdef function call

=item B<interval> => I<interval>
pass through parameter for server cdef function call

=back

=item Returns

=over

=item B<Success:> 1

=item B<Failure:> croaks

=back

=item Examples

=back

    # Take over partner of this node
    $node->takeover();

=cut

sub takeover {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $background = delete $args{background};
    my $partner    = $self->_verify_sfo_ha();

    my $title = "takeover of partner $partner by " . $self->name;
    NOTE::logcomment("Start $title");
    $self->{_partner} = NOTE::GX::Node->connect( name => $partner )
      unless exists $self->{_partner};
    eval { $self->server->storage_failover_takeover( "bynode" => $self->name, %args ); };
    croak "$title failed. Error: $@"
      if $@;

    $self->{_sfo_action} = "takeover";
    $self->{_partner}->{_redirect_node} = $self;
    $self->sfo_monitor(%args)
      unless $background;

    return 1;
}

=head2 auto_giveback

Gets and sets the auto-giveback flag within node.

=over

=item Arguments

I<Optional Arguments>

=over

=item B<flag> : boolean

true/false: it sets to the given state
none      : does not do anything

=back

=item Returns

=over

=item B<Success:>
Returns the old state, if set to the other value
Returns the existing state, if no value is provided

=item B<Failure:> confess in case of error in changing current state

=back

=item Examples

    # Get the current auto-gvieback state
    my $curr_state = $node->auto_giveback();

    # Turns off auto-giveback and returns the old state
    my $old_state = $node->autogiveback(0);

=back

=cut

sub auto_giveback {
    my ( $self, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $old_flag = $self->sfo_status->{"auto-giveback"};
    $old_flag = ( $old_flag eq 'true' ? 1 : 0 );

    if ( defined $flag && $flag != $old_flag ) {
        my $new_auto = ( $flag ? "on" : "off" );
        my $output;
        eval { $output = $self->server->dblade( cmd => "options cf.giveback.auto.enable $new_auto" ); };
        confess "Error changing auto-giveback. Error: $@"
          if $@;
        confess "Unexpected output when setting auto-giveback. Output: $output"
          if $output;
    }
    return $old_flag;
}

=head2 giveback

Perform a giveback to partner node by this node. It first determines if the 
node is located on a cluster consisting of 3 or more nodes.
If not, it will enable High Avaialbility prior to invoking the giveback action.

This method will accept pass-through parameters to be send to sfo_monitor,
such as timeout, interval etc

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<background> => I<boolean>

If set method will return immediately after starting a 'giveback' to it's parter. Default: 0

=item B<retry> => I<integer>

Number of retrys before aborting. After a node have been taken over, 
there is a approximately 2.5 minutes before it is ready for a 'giveback',
during which time a 'giveback' will fail. The retry mechanism is designed
to provide a mechanism to handle this transition period. 
Default: 5

=item B<interval> => I<integer>

Number of seconds to pause in between retries. Default: 60

=item B<force> => I<boolean>

force to override-vetoes. if not provided, NOTE::option $SFO_FORCE is taken. 
Default is false.

=back

=item Returns

=over

=item B<Success:> 1

=item B<Failure:> croaks

=back

=item Examples

=back

    #To give back with default values
    $node->giveback();

    $node->giveback(retry => 3,background=> 0, interval => 500);

=cut

sub giveback {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $background = delete $args{background};
    my $partner    = $self->_verify_sfo_ha();

    my $retry    = delete $args{retry}    || 5;
    my $interval = delete $args{interval} || 60;

    $args{"override-vetoes"} = ( exists $args{force} ? delete $args{force} : NOTE::option('$SFO_FORCE:0') ) ? "true" : "false";

    my $title = "giveback of $partner from " . $self->name;
    NOTE::logcomment("Start $title");
    do {
        eval { $self->server->storage_failover_giveback( "fromnode" => $self->name, %args ); };
        if ($@) {
            if ($retry) {
                NOTE::logcomment("Pause a little before retry...");
                Tharn::snooze($interval);    # Pause a little and retry
            }
        }
        else {
            $self->{_sfo_action} = "giveback";
            $self->sfo_monitor(%args)
              unless $background;
            return 1;
        }
    } while $retry--;
    croak "Too many retries for $title. Last error: $@";
}

=head2 sfo_monitor

Waits for the node for SFO state to reach the expected state and reconnect to the partner,
if the _sfo_action parameter value is giveback
If the SFO state of the node does not reach the expected state within the timeout,
reconnects to the partner if the _sfo_action parameter value is takeover and then croak

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<_sfo_action> => I<takeover|giveback>

SFO action

=item B<state> => I<string>

State to wait for. 
Default : 'connected' for giveback action and 'takeover' from takeover action

=item B<timeout> => I<integer>

Amount of time to wait in secs . Default : 900

=item B<interval> => I<interval>

Wait time in between polling in secs . Default : 30

=back

=item Returns

=over

=item B<Success:> 1

=item B<Failure:> croaks

=back


=item Examples

=back

   $node->sfo_monitor();

=cut

sub sfo_monitor {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $exp_state = delete $args{state}    || ( $self->{_sfo_action} =~ /takeover/i ? "takeover" : "connected" );
    my $timeout   = delete $args{timeout}  || 900;
    my $interval  = delete $args{interval} || 30;

    my $end_time = time + $timeout;

    do {
        my $status = $self->sfo_status();
        my $sfo_status = $status->{'state'};
        # Due to duplication of partial text must search for eact state
        if ( $sfo_status =~ /^$exp_state$/i ) {
            NOTE::logcomment( 'SFO status: ' . $sfo_status );
            # Pausing a little before reconnecting partner
            sleep 15;
            $self->_reconnect_partner() if ( $self->{_sfo_action} =~ /giveback/ );
            return 1;
        }
        else {
            NOTE::logcomment( 'SFO Current status: ' . $sfo_status .
                              ", Waiting for $exp_state state" );
            Tharn::snooze($interval);
        }
    } while time < $end_time;
    NOTE::logcomment("Takeover timed out. Trying reconnect to previous node before croaking");
    $self->_reconnect_partner() if ( $self->{_sfo_action} =~ /takeover/ );
    croak "SFO Time out after $timeout sec(s) waiting for $exp_state state";
}

#Internal method to reconnect to the partner node
sub _reconnect_partner {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_partner} = NOTE::GX::Node->connect( name => $self->sfo_status->{'partner-name'} )
      unless exists $self->{_partner};
    delete $self->{_partner}->{_redirect_node};
    $self->{_partner}->reconnect();
}

=head2 sfo_status

Returns a HASHREF of the SFO status for this node. 
A single column will be returned if specified.

=over

=item Arguments

I<Optional Arguments>

=over 

=item B<coumn_name>

Significant column names are: 
'partner-name', 'enabled', 'possible', 'interconnect-up' and 'statue'

=back

=item Returns

=over

=item B<Success:> 

value of the specified column, in the sfo status object if column_name is specified
if not specified, complete hashref of sfo status object

=item B<Failure:> 

croaks if problem in getting sfo status

=back

=item Examples

=back

    #to just get the partname colum value
	my $partner = $node->sfo_status("partnername");

    #to get the complete hashref to the sfo status 
    my $sfo = $node->sfo_status;
    if ($sfo->{"interconnect-up"} =~ /true/ && $sfo->{"enabled"} =~ /true/ ) {
        $node->takeover();
    }

=cut

sub sfo_status {
    my ( $self, $column ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $s;
    eval { $s = $self->server->storage_failover_show( node => $self->name ); };
    croak "Error determining SFO status for " . $self->name . ". Error: $@" if $@;


    #the cdef outout is changed to have the attribute 'state-description'
    # in place of 'state'. so have both the attributes
    # sudheerg - 20120409
    if((defined $s->[0]->{'state-description'}) and
       (!defined $s->[0]->{'state'}) ){
        $s->[0]->{'state'} = $s->[0]->{'state-description'};
    }
    if((defined $s->[0]->{'state'}) and
       (!defined $s->[0]->{'state-description'}) ){
        $s->[0]->{'state-description'} = $s->[0]->{'state'};
    }

    #strip off the forward and trailing double quotes, if present
    $s->[0]->{'state'} =~ s/^"|"$//g;
    $s->[0]->{'state-description'} =~ s/^"|"$//g;

    if ( defined $column ) {
        return $s->[0]->{$column};
    }
    else {
        return $s->[0];
    }
}

=head2 ::sfo_modify

Changes SFO settings for a node using 'storage failover modify' command.

=over

=item Arguments

B<column> => I<new-value>

=item Examples

=back

    $node->sfo_modify("auto-giveback"=>"false");

=cut

sub sfo_modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{node} = $self->name;
    eval { $self->server->storage_failover_modify(%args); };
    croak "Error modifying SFO settings for " . $self->name . ". Error: $@" if $@;
}

=head2 :: lifs

Returns an array of LIF names on this node.

=over

=item Arguments

B<name> => I<string>

I<Optional>

Returns a specific LIF. Can be used to check for existence of a specific LIF on a node

B<role> => I<string>

I<Optional>

One of data, node-mgmt, intercluster or cluster-mgmt. If not specify then all roles are returned.

B<link> => I<up|down>

I<Optional>

Indicates to return LIF names that is in a certain state. Default: both up and down

=item Examples

    my $node_lifs = $node->lifs(); # all LIF names on the current node
	my $data_lifs = $node->lifs(role=>"data", link=>"up");

=back

=cut

sub lifs {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Cluster->lifs( %args, node => $self->name );
}

=head2 ::lif_modify

Modifies the up/down status of a lif on the node.
Lif name must be specified.
Returns a boolean to indicate whether the modify was successful or not

=over

B<name> => I<string>

I<Required>

The name of the lif whose status change is required.

B<link> => I<up|down>

I<Required>

Indicates the desired new status(admin) of the lif. The only allowed values are 'up' or 'down'.

=item Examples

=back

    $node->lif_modify(name=>'clus1', link => 'down');

=cut

sub lif_modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Vserver name should not be provided" if exists $args{vserver};

    confess "Node name if provided should be the same as this node"
      if exists $args{node} and $args{node} ne $self->name;

    delete $args{node} if exists $args{node};

    return $self->Cluster->lif_modify( %args, node => $self->name );
}

=head2 ports

Returns array of network port names of the Node

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<name> => I<name of the port>

name of the port. if this port exist, returns this port name as part of array

=back

=item Returns

=over

=item B<Success:> array of port names

=item B<Failure:> croaks if error in getting network port show output

=back

=item Examples

=back

	#to get all the port names
	my @node_ports = $node->ports;

=cut

sub ports {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list;

    if ( exists $args{name} ) {
        $args{port} = delete $args{name};
    }

    eval { @list = $self->server->network_port_show( node => $self->name, %args, ); };
    if ($@) {
        croak $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    my @portnames = ();
    for (@list) {
        push @portnames, $_->{port};
    }
    return @portnames;
}

=head2 Ports

Returns array of objects created for network ports of the Node, if wantarray
if not wantarray, returns the first port object

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<name> => I<name of the port>

name of the port. if this port exist, returns the object of this port alone as part of array

=back

=item Returns

=over

=item B<Success:> 

array of port objects if wantarray
if not wantarray, returns the first object of the array

=item B<Failure:> croaks if error in getting network port show output

=back

=item Examples

=back

    #to get all the port objects
    my @node_ports_obj = $node->Ports;

=cut

sub Ports {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @Ports = ();
    for ( $self->ports(%args) ) {
        push @Ports,
          NOTE::GX::Port->new(
            container => $self,
            name      => $_,
          );
    }
    return wantarray ? @Ports : $Ports[0];
}

=head2 ::disks

Returns an array of names of disks belonging to this node.

=over

=item Arguments

B<state> => I<string>

I<optional>

Possible states are: spare, broken, reconstructing, zeroing

=item Examples

     my @spares = $node->disks(state=>"spare");

=back

=cut

sub disks {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list = ();

    eval { @list = $self->server->storage_disk_show( owner => $self->name, %args, ); };
    if ($@) {
        croak $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    my @disknames = ();
    for (@list) {
        push @disknames, $_->{disk};
    }
    return @disknames;
}

=head2 ::Disks

Returns an arrary of GX::Disk objects belonging to this node.

=over

=item Arguments

B<state> => I<string>

I<optional>

Possible states are: spare, broken, reconstructing, zeroing

=item Examples

     my @spare_objs = $node->Disks(state=>"spare");

=back

=cut

sub Disks {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @Disks = ();
    for ( $self->disks(%args) ) {
        push @Disks, NOTE::GX::Disk->new( name => $_ );
    }
    return @Disks;
}

=head2 status

Returns the status of the node from the output of cluster show

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:
None.

=item Returns

=over

=item B<Success:> status string

=item B<Failure:> croaks if error getting the status

=back

=item Examples

=back

	my $status = $node->status;

=cut

sub status {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $status;

    eval { $status = $self->server->cluster_show( node => $self->name, )->[0]; };
    croak $@ if $@;
    return $status;
}

=head2 name

Returns the name of the node this object is referring to.

=over

=item Arguments

None.

=item Returns

=over

=item B<Success:> name of the node, as string

=item B<Failure:> no failure scenario

=back

=item Examples

=back

    my $node_name = $node->name;

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_name}
}

=head2 server

Sets the execution server to this node before returning the Server object.
The need to set the execution server is a work-around for the Server object
which does not allow for execution of command on specific node. This is
specifically needed for inter-cluster testing, which the nodes represent
filers in different clusters.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:
None

=item Returns

=over

=item B<Success:> Returns the server object

=item B<Failure:> no failure scenario for this method. does not croak/confess

=back

=item Examples

=back

	my $srv_obj = $node->server;

=cut

sub server {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $self->{_redirect_node} ) {
        NOTE::logcomment( "Redirect to takeover node: " . $self->{_redirect_node}->name );
        return $self->{_redirect_node}->server;
    }

    $current_node = $self->{_server}->get_execution_server;
    unless ( $self->{_name} =~ /^$current_node$/ ) {
        NOTE::logcomment( "Node switch: from $current_node, to " . $self->{_name} );
        $current_node = $self->{_name};
        $self->{_server}->set_execution_server( -Server => $current_node );
    }

    return $self->{_server};
}

=head2 printflag

Gets the value of the Dblade flag specified

Filer flags are visible in either diag or test mode. To ensure scripts can
be run on promoted builds, and to support an interface similar to setflag,
the flag is attempted to be gotten first in diag mode. if that fails it will
be retried in test mode. This is not the most efficient but it does
maintain the right of the script to be ignorant of what mode it needs to
be run in without causing additional overhead of keeping a table of which
flags require what mode. This is also the way NOTE handled this type of
problem.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<name> => I<string>

name of the flag.

=back

=item Returns

=over

=item B<Success:> the value of the flag

=item B<Failure:> confesses, if flag does not exist

=back

=item Examples

=back

    # To get the value of the paloma_allow_striped_replication
    $value = $node->printflag( name => 'paloma_allow_striped_replication' );

    # To get the value of the paloma_allow_unlike_geometry
    $value = $node->printflag( name => 'paloma_allow_unlike_geometry' );

=cut

sub printflag {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $cmd, $output, $value );

    # also accept flag as an argument to support consistency between
    # NOTE and GNOTE
    if (exists($args{flag})) {
        if (! exists($args{name})) {
            $args{name} = $args{flag};
        }

        delete $args{flag};
    }

    if (! exists $args{name}) {
        confess q(Missing 'name' parameter...);
    }

    $cmd = "printflag $args{name}";

    # changes for burt 490236.
    # some flags are visible in diag mode and others require test mode
    # first try diag mode
    $output = $self->server->dblade( cmd => "priv set diag; $cmd" );

    # if the flag does not exist try again in test mode, it's brute
    # force but simple
    if ($output && ($output =~ /no such flag/i)) {
        NOTE::logcomment("running command in test mode: '$cmd'");

        $output = $self->server->dblade( cmd => "priv set test; $cmd" );

        if ($output && ($output =~ /unrecognized privilege/i)) {
            confess "build does not support privilege level test, error:\n"
                    . $output;
        }

        if ($output && ($output =~ /no such flag/i)) {
            confess "printflag command failed: '$cmd', error: $output";
        }
    } elsif ($output && ($output =~ /unrecognized privilege/i)) {
        confess "build does not support privilege level diag, error:\n$output";
    }

    if (!defined($output)) {
        confess "printflag failed for unknown reason: '$cmd'";
    }

    # get the value of the flag so it can be returned
    # NOTE: this assumes the format '<flag> = <value>'
    LINE:
    for my $l (split(/\n/, $output)) {
        if ($l =~ /$args{name}/) {
            $value = (split(/ /, $l))[2];
            last LINE;
        }
    }

    return $value;
}

=head2 setflag

Sets a Dblade flag to a specific value.

This method will record the original value of a flag (by storing off the
value of the flag the first time it is called for that flag) so the flag
can be rest to its original value during teardown. NOTE, if a flag is set
outside the GNOTE mechanism before setflag is called teardown will reset
the flag to an incorrect value.

Filer flags are visible in either diag or test mode. To ensure scripts can
be run on promoted builds, and to support the original setflag interface,
the flag is attempted to be set first in diag mode. if that fails it will
be retried in test mode. This is not the most efficient but it does
maintain the right of the script to be ignorant of what mode it needs to
be run in without causing additional overhead of keeping a table of which
flags require what mode. This is also the way NOTE handled this type of
problem.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<name> => I<string>

name of the dblade param

=back

I<Optional Arguments>

=over 

=item B<value> => I<1|0>

Value to set the flag to.
Default: 1

=back

=item Returns

=over

=item B<Success:> does not return anything

=item B<Failure:> confesses, if error in setting the flag

=back

=item Examples

=back

    #To disbale paloma_allow_striped_replication
	$node->setflag(name=>"paloma_allow_striped_replication", value=>0);

    #To enable paloma_allow_unlike_geometry
    $node->setflag(name=>"paloma_allow_unlike_geometry");

=cut

sub setflag {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $flag, $value, $cmd, $org_value, $output );

    # also accept flag as an argument to support consistency between
    # NOTE and GNOTE
    if (exists($args{flag})) {
        if (! exists($args{name})) {
            $args{name} = $args{flag};
        }

        delete $args{flag};
    }

    if (! exists $args{name}) {
        confess q(Missing 'name' parameter...);
    }

    $flag = $args{name};
    $value  = ( exists $args{value} ? delete $args{value} : 1 );
    $cmd    = "setflag $flag $value";

    # changes for burt 490236.
    # get the original value so the flag can be reset later
    if ( !exists($self->{_wafl_flag}{$flag}) ) {
        $org_value = $self->printflag( name => $flag );
        if(not defined $org_value) {
            confess "Could not get original value for flag: $flag";
        }

        $self->{_wafl_flag}{$flag} = $org_value;
    }

    # some flags are visible in diag mode and others require test mode
    # first try diag mode
    $output = $self->server->dblade( cmd => "priv set diag; $cmd" );

    # if the flag does not exist try again in test mode, it's brute
    # force but simple
    if ($output && ($output =~ /no such flag/i)) {
        NOTE::logcomment("running command in test mode: '$cmd'");

        $output = $self->server->dblade( cmd => "priv set test; $cmd" );

        if ($output && ($output =~ /unrecognized privilege/i)) {
            confess "build does not support privilege level test, error:\n"
                    . $output;
        }

        if ($output && ($output =~ /no such flag/i)) {
            confess "setflag command failed: '$cmd', error: $output";
        }
    } elsif ($output && ($output =~ /unrecognized privilege/i)) {
        confess "build does not support privilege level diag, error:\n$output";
    }

    NOTE::logcomment("'$cmd' successful");
    return;
}

=head2 Cluster

Returns the GX::Cluster object representing the cluster that this node belongs to.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:
None

=item Returns

=over

=item B<Success:> returns the GX:Cluster object

=item B<Failure:> no failure scenario

=back

=item Examples

=back

	#get the cluster object
	my $cluster_obj = $node->Cluster;

=cut

sub Cluster {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_Cluster}
}

=head2 ::alarms

Returns an array of references to hashes that contain information about
alarms pending for the node. Each hash has the following fields:

  type 		Type of alarm
  node 		Node the alarm is associated with
  name 		Name of the object with the alarm
  state		Alarm state (ok|warning|critical)
  ems-state	The EMS state (rising|falling)
  value		The most recent value
  high-value	The highest value

=over

=item Arguments

Each of the keys described above can be passed along with
a value. This will act as a filter. For example,

=item Examples

=back

    @alarms = $node->alarms( type => 'aggregate-used' );

    @alarms = $node->alarms(
	type  => 'aggregate-used',
	state => 'critical',
    );

=cut

sub alarms {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @alarms;

    eval { @alarms = $self->server->dashboard_alarm_show( node => $self->{_name}, %args, ); };
    if ($@) {
        croak $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    return @alarms;
}

=head2 aggregates

Returns the list of names of aggregates associated with this node.
If a criteria is specified by providing the following mentioned arguments, 
aggregate names matching that criteria only will be returned as an array.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<aggregate> => I<name> Name of the aggregate

=item B<availsize> => I<integer> available size

=item B<chksumenabled> => I<true|false> 

=item B<chksumstatus> => I<true|false>

=item B<chksumstyle> => I<block|?>

=item B<diskcount> => I<integer>

=item B<inconsistent> => I<true|false>

=item B<maxraidsize> => I<integer>

=item B<percent-used> => I<string: e.g. "10%">

=item B<plexes> => I<string> 

=item B<raidgroups> => I<string> raid groups name

=item B<raidstatus> => I<raid_dp|raid4|pending>

=item B<raidtype> => I<raid_dp|raid4>

=item B<size> => I<integer>

=item B<state> => I<online|offline|restricted>

=item B<usedsize> => I<integer>

=item B<volcount> => I<integer>

=item B<volume-style> => I<flex|striped>

=back

=item Returns

=over

=item B<Success:> array of aggregate names matching the criteria

=item B<Failure:> croaks if there is error in getting the output of aggregate show

=back

=item Examples

=back

    #To get all the aggregate names
    my @all_aggregate_names = $node->aggregates;

    #To print each aggregate name that is online
    for my $aggr ( $node->aggregates( state => 'online' ) ) {
       print "$aggr\n";
    }

    #To get all aggregates for whcih raidstatus is raid_dp and status is offline
    my @raiddp_offline_aggrs = $node->aggregates( state => 'offline', raidstatus=> 'raid_dp' );	


=cut

sub aggregates {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @aggrs;
    my @aggrnames = ();

    eval { @aggrs = $self->server->storage_aggregate_show( nodes => $self->name, %args ); };
    if ($@) {
        croak $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    for (@aggrs) {
        push @aggrnames, $_->{aggregate};
    }
    return @aggrnames;
}

=head2 Aggregates

Returns an array of GX::Aggregate objects created from the results
of the list returned by the 'aggregates' method in this module.
If a criteria is specified by providing the following mentioned arguments,
objects of aggregates matching that criteria only will be returned as an array.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<aggregate> => I<name> Name of the aggregate

=item B<availsize> => I<integer> available size

=item B<chksumenabled> => I<true|false>

=item B<chksumstatus> => I<true|false>

=item B<chksumstyle> => I<block|?>

=item B<diskcount> => I<integer>

=item B<inconsistent> => I<true|false>

=item B<maxraidsize> => I<integer>

=item B<percent-used> => I<string: e.g. "10%">

=item B<plexes> => I<string>

=item B<raidgroups> => I<string> raid groups name

=item B<raidstatus> => I<raid_dp|raid4|pending>

=item B<raidtype> => I<raid_dp|raid4>

=item B<size> => I<integer>

=item B<state> => I<online|offline|restricted>

=item B<usedsize> => I<integer>

=item B<volcount> => I<integer>

=item B<volume-style> => I<flex|striped>

=back

=item Returns

=over

=item B<Success:> array of aggregate objects matching the criteria

=item B<Failure:> croaks if error in getting the output of aggregate show

=back

=item Examples

=back

	<example>

=cut


=head2 ::Aggregates

Returns an array of GX::Aggregate objects created from the results
of the list returned by the 'aggregates' method above.

=over

=item Arguments

The arguments are passed to 'aggregates' method above for processing.

=item Examples

=back

    #To get objects of all the aggregates
    my @aggregate_obj_array = $node->Aggregates;

    #To get objects of aggregates for which raidstatus is raid_dp and status is offline
    my @raiddp_offline_aggrs_obj = $node->Aggregates( state => 'offline', raidstatus=> 'raid_dp' );


=cut

sub Aggregates {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @aggrs = ();

    for my $aggr ( $self->aggregates(%args) ) {
        push @aggrs,
          NOTE::GX::Aggregate->attach(
            Node => $self,
            name => $aggr,
          );
    }
    return @aggrs;
}

=head2 ::do_zerospares

Enables or disables the scheduling of spare disk zeroing
when the cluster object is destroyed.
(either explicitly or upon test completion).
If called with no arguments zeroing is enabled unless it
was previously explicitly disabled.

=over

=item Arguments

C<0> | C<1>

I<Optional.>
A flag to turn on (C<1>) or off (C<0>)
the scheduling of spare disk zeroing.

=item Examples

=back

     $node->do_zerospares;      # Enables zeroing
     $node->do_zerospares( 0 ); # Disables zeroing
     $node->do_zerospares;      # Won't re-enable zeroing
     $node->do_zerospares( 1 ); # Will re-enable zeroing

=cut

sub do_zerospares {
    my ( $self, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $flag ) {
        $self->{_do_zero_spares} = $flag;
    }
    else {
        $self->{_do_zero_spares} = 1
          unless ( defined $self->{_do_zero_spares} );
    }
}

=head2 <name>

Does zerospares of disks of the current node.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:
None

=item Returns

=over

=item B<Success:> None
 
=item B<Failure:> no failure scenario

=back


=item Examples

=back

    #To zerospare the disks of the node
	$node->zerospares;

=cut


sub zerospares {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->storage_disk_zerospares( owner => $self->name );
}

=head2 ::snapmirror_status

Returns a pointer to an array of hashes representing the statuses of all
all snapmirror relationships in the cluster.

=over

=item Arguments

B<None>

=item Example

=back

      my $status = $node->snapmirror_status();

=cut

sub snapmirror_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $s;
    eval { $s = $self->server->snapmirror_show(%args); };
    croak $@ if $@;
    return $s;
}

=head2 ::job_id

This method have been deprecated. Please call JOB::get_id instead.

Given the text returns from ONTAP after starting a command, parses the text to obtain and return
the job id.

=over

=item Arguments

B<text> = I<string>

The text returned by ONTAP after starting a job.
The supplied text must be of the format '[Job 123]', where '123' is the job id

=item Example

=back

    my $id = $node->job_id($ontap_text);

=cut

sub job_id {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return NOTE::GX::Job::get_id(@_);
}

=head2 ::job_wait

This method has been deprecated. Please call Job::wait instead.

Given a job id, will wait until the job reaches a certain state.
Since the state of a job cannot be obtained after it is complete
the typical use of this method is to verify the 'running' state.

Croaks if it cannot find job id.

=over

=item Arguments

B<id> => I<integer>

Job id of job to wait for.

B<state> => I<string>]

State to wait for. Default: success

B<timeout> => I<integer>

Timeout in seconds. Default: 300

B<interval> => I<integer>

Polling interval in seconds. Default: 1

B<column> => I<string>

Job status column to return. Default: state

=item Examples

=back

    $node->job_wait(id => 5, state => "running", timeout => 30, interval => 5); # Returns 'state' column
    $my $tstaus_hashrec = node->job_wait(id => 6, column => 'all'); # Returns status hashref

=cut

sub job_wait {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return NOTE::GX::Job::wait(@_);
}

=head2 addresses

Returns array of network port ip addresses of the Node

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<name> => I<name of the port>

name of the port. if this port exist, returns this port ip address as part of array

=back

=item Returns

=over

=item B<Success:> array of ip addresses

=item B<Failure:> croaks if error in getting network port show output

=back

=item Examples

=back

    #to get all the port names
    my @node_ip_addresses = $node->addresses;

=cut

sub addresses {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @addresses;
    my @interfaces = ();

    eval { @interfaces = $self->{_server}->network_interface_show( server => $self->name, %args ); };
    if ($@) {
        croak $@ unless ( $@ =~ /There are no entries matching your query/ );
    }
    for (@interfaces) {
        push @addresses, $_->{address};
    }
    return @addresses;
}

=head2 epsilon

Returns boolean indicating whether this node has epsilon or not.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:
None

=item Returns

=over

=item B<Success:> boolean

=item B<Failure:> croaks, if error determining the epsilon status

=back

=item Examples

=back

    #To get the status of epsilon on this node
	my $epsilon_exists = $node->epsilon();

=cut

sub epsilon {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->set( privilege => "diag", confirmations => "off" );
    my $result;
    eval { $result = $self->server->cluster_show( node => $self->name ); };
    confess "Failed to determine epsilon status. Error $@"
      if $@;
    return $result->[0]->{epsilon} =~ /true/i;
}

=head2 set_epsilon

Sets the current node to be the epsilon node.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> 

Returns node object in one case, but not in other case. (I.e., Do not use returned value.) 

=item B<Failure:>

No failure scenario.

=back

=item Examples

=back

	#make the curent node to be the epsilon node
	$node->set_epsilon();

=cut

sub set_epsilon {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $current_epsilon;
    my $ha_status;
    my $result;
    my $disable = 0;

    $self->server->set( privilege => "diag", confirmations => "off" );
    $result = $self->server->cluster_show();
    $ha_status = $self->server->cluster_ha_show();

    # Disabling Cluster HA for a two node cluster
    if ($self->{_Cluster}->Nodes == 2){
        NOTE::logcomment("Need to disable Cluster HA to assign epsilon");
        if ($ha_status->[0]->{configured} =~ /true/i) {
            NOTE::logcomment("Disabling Cluster HA");
            $self->server->cluster_ha_modify( configured => "false");
            $self->{_disabled_ha} = 1;
        }
    }

    # Getting the Cluster information after disabling HA
    $result = $self->server->cluster_show();
    foreach my $node (@{$result}) {
        if ($node->{epsilon} =~ /true/i) {
            $current_epsilon = $node->{node};
            last;
        }
    }
    if (defined ($current_epsilon)) {
        if ($current_epsilon eq $self->name) {
            NOTE::logcomment("Node ".$self->name." is already the Epsilon node");
            return;
        } else {
            # Setting the current epsilon to false
            $self->server->cluster_modify(node => $current_epsilon,
                                          epsilon => "false");
            # Setting the new epsilon
            $self->server->cluster_modify(node => $self->name,
                                          epsilon => "true");
        }
    } else {
        # Since there is no current epsilon, setting this node to be the epsilon
        NOTE::logcomment("NO EPSILON NO HA");
        $self->server->cluster_modify(node => $self->name,
                                      epsilon => "true");
    }
    # Doing a cluster show so the new epsilon is visible in the logs
    $self->server->cluster_show();

    return $self;
}

=head2 wafl_flag

This function sets or retrieves the wafl flags on the filer
To get: pass just the flag, it retrieves the value of the flag.
To set: pass both flag and value are passed, sets the flag to the provided
value.
Returns hash ref contatining the old value in
get mode and the new value in set mode.

=over

=item Arguments

=over

=item flag 

wafl flag as string

=back

I<Optional Arguments>

=over 

=item value

value to be set to the provided flag

=back

=item Returns

=over

=item B<Success:> 

hash ref contatining the old value in
get mode and the new value in set mode.

$href->{flag} = flag provided by caller
$href->{value} = current value of flag

=item B<Failure:>

no failure scenario 

=back

=item Examples

=back

    #To get the value of wafl flag wafl_inconsistent_panic_testing
    # and verify it is as expected
	my $hr_flag_val = $node->wafl_flag("wafl_inconsistent_panic_testing");
    if ($hr_flag_val->{value} != $expected_value) {
        confess "flag value not as expected ($hr_flag_val->{flag}):\n" .
                "got: $hr_flag_val->{value}\nexpected: $expected_value";
    }

    #To set the value of wafl flag wafl_iron_panic_on_corruption to 0
    # and verify it was set correctly
    my $hr_flag_val = $node->wafl_flag("wafl_iron_panic_on_corruption","0");
    if ($hr_flag_val->{value} != 0) {
        confess "flag value not as expected ($hr_flag_val->{flag}):\n" .
                "got: $hr_flag_val->{value}\nexpected: 0";
    }

=cut

sub wafl_flag{
    my ($self, $flag, $value) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $current_value, $info );

    if (not defined $flag) {
        confess 'No wafl flag provided to set or retreive';
    }

    if (defined $value) {
        $self->setflag( name => $flag, value => $value );
    }

    # get the current value of the flag to return to the caller
    $current_value = $self->printflag( name => $flag );
    if(not defined $current_value) {
        confess "Could not get current value for flag: $flag";
    }

    $info->{flag} = $flag;
    $info->{value} = $current_value;

    return $info;
}

=head2 crash_mgwd

Gets the process id of the mgwd process and kills it, 
and checks whether mgid is started with new process id or not.
Returns boolean indicating mgwd is crashed succesfully or not.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> 1 if mgwd is succesfully crashed

=item B<Failure:> 0, if mgwd is not succesfully crashed

=back

=item Examples

=back

    #crash mgwd process and get the status of succesfulness
	my $mgwd_crashed = $node->crash_mgwd();

=cut

sub crash_mgwd {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $pid  = $self->_get_pid("mgwd");
    return 0 if not defined $pid;

    my $kill_cmd = "sudo kill $pid";

    # eval is required because server will close the connection after crash
    eval { $self->server->shell( cmd => $kill_cmd ); };

    # reconnect is required because server will close the connection after crash
    $self->reconnect;

    my $new_pid = $self->_get_pid("mgwd");

    if ( $new_pid != $pid ) {
        NOTE::logcomment("mgwd pid changed from $pid to $new_pid");
        return 1;
    }

    return 0;
}

=head2 ::ping

Ping filer on default ip. Uses Perl 'Net::Ping->new->ping' method. All arguments
are passed to Net::Ping as is.

=cut

sub ping {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $default_ip_gx = host($self->name)->default_ip_gx;
    Net::Ping->new('external')->ping( $default_ip_gx, @_ );
}

=head2 ::transit_to

Transition the filer to a specific state. If the filer is already in the requested state
no action is taken. The arguments control the selection of states to put the filer in
and the process by which states are attained.

=over

=item Arguments

I<boot state constant>

I<Required, positional.>
The final state the filer should be left in. The valid states are described in Node::NG::Transit.
Additionally, for mode-agnostic convenience, the states FIRMWARE, BOOT_MENU and UP are accepted
and translated to the appropriate NG state. Must be given before any named options.

I<named options>

I<Optional.>
There are many named options. These options are described in Node::NG::Transit

=item Examples

=back

	 $filer->transit_to( 'FIRMWARE' );
	 $filer->transit_to( 'CLI', timeout => 300 );

=cut


sub transit_to {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($to, %args) = @_;

    # Turn 7-mode states into appropriate 10-mode state
    # Based on reverse map in lib/Node.pm::agnostic_transit_to:
    # my %state_map = (
    #    CFE => "FIRMWARE",
    #    IBM => "BOOT_MENU",
    #    LOGIN => "UP",
    #    REBOOT => "REBOOT",
    #    CLI => "UP",
    #    MAINT => "MAINT",
    #    SHELL => "UP",
    # );
    my %state_map = (
        FIRMWARE  => 'CFE',
        BOOT_MENU => 'IBM',
        UP        => 'CLI',
    );

    my $new_to = $state_map{$to};
    $new_to = $to if (!$new_to);
    my %params = ( state=>$new_to, %args );
    $self->server->transit_to(%params);
}

###############################################################################
##
##   Sub Name:       get_wafl_version
##
##   Description:    returns wafl version (wafl id) for node (accessor)
##
##   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
##                   $self     IN      ref       object of this class
##
##   Globals:        
##
##   Environment:    
##
##   Returns:        Success:    scalar of wafl id
##                   Failure:    
##
################################################################################
sub get_wafl_version {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return $self->{_wafl_version};
}

###############################################################################
##
##   Sub Name:       get_ontap_version
##
##   Description:    returns ontap release version for node (accessor)
##
##   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
##                   $self     IN      ref       object of this class
##
##   Globals:        
##
##   Environment:    
##
##   Returns:        Success:    scalar of ontap release string
##                   Failure:    
##
################################################################################
sub get_ontap_version {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return $self->{_ontap_version};
}

sub _get_pid {
    my $self      = shift;
    my $proc_name = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $exit_code;
    my $output;
    my $cmd = 'ps -axc';

    NOTE::logcomment("The command to be executed is $cmd");
    $self->server->shell(
        cmd       => $cmd,
        output    => \$output,
        exit_code => \$exit_code
    );
    return undef if $exit_code;

    # sample of 'ps -axc' output
    # <<hornby#shell|  PID  TT  STAT      TIME COMMAND\r\r\n
    # <<hornby#shell|    0  ??  WLs    0:34.76 swapper\r\r\n
    # <<hornby#shell|    1  ??  ILs    0:00.02 init\r\r\n
    # <<hornby#shell|   22  ??  WL     0:00.00 irq10:\r\r\n
    # <<hornby#shell| 1086  ??  Is     0:00.00 mlogd\r\r\n
    # <<hornby#shell|12976  ??  Ss     0:01.90 mgwd\r\r\n
    # <<hornby#shell|12977  ??  Is     0:00.02 sshd\r\r\n

    my @lines       = split /\n/, $output;
    my @header_list = split ' ',  shift @lines;
    my $pid         = undef;
    while ( my $line = shift @lines ) {
        my @v = split ' ', $line;

        my %fields;
        foreach my $field (@header_list) {
            $fields{$field} = shift @v;
        }

        if ( $fields{'COMMAND'} =~ /($proc_name)/ ) {
            $pid = $fields{'PID'};
            last;
        }
    }
    return $pid;
}

=head2 datetime

Returns the date / time for this node as a string in the below mentioned format.
Return format: <month>/<day>/<year> <hour>:<minute>:<second> <timezone>
Example of return: 8/10/2009 13:59:09 US/Pacific

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<utc> => I<boolean> 

To get the UTC (or Greenwich Mean Time, GMT) time instead of local time.
Default: 0 (i.e. local time)

=back

=item Returns

=over

=item B<Success:> Returns date as a string 

=item B<Failure:> confess, in case of error getting datetime of node

=back

=item Examples

=back

	#To get local datetime
    my $local_dt_str = $node->datetime();

    #To get gmt time of node
    my $gmt_dt_str = $node->datetime(utc=>1);

=cut

sub datetime {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $output;
    eval { $output = $self->server->system_node_date_show( node => $self->name ); };
    confess( $self->name . ": failed to get date. Error: $@" )
      if $@;
    my $dt;
    if ( $args{utc} ) {
        $dt = $output->[0]->{'utc-date'} . " UTC";
    }
    else {
        $dt = $output->[0]->{date} . " " . $output->[0]->{timezone};
    }
    $dt =~ s/"//g;
    return $dt;
}

use Time::Local;

=head2 time

Translate the Node's datetime to an integer representing
the number of non-leap seconds since 'Jan 1, 1970, UTC'.
This is the equivalent as calling Perl's 'time' function
except using this Node's datetime.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> returns seconds as integer since 'Jan 1, 1970, UTC'

=item B<Failure:> confesses, if getting error in getting datetime of node

=back

=item Examples

=back

	#To get the time in seconds since 'Jan 1, 1970, UTC'
    my $node_time_since_jan11970 = $node->time;

    #the return value can also be used to get time greenwitch mean time
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = gmtime(node->time());


=cut

sub time {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $time_str = $self->datetime( utc => 1 );
    $time_str =~ /^(\d+)\/(\d+)\/(\d+) (\d+):(\d+):(\d+) .+$/;
    my ( $mn, $dy, $yr, $hr, $mi, $se ) = ( $1, $2, $3, $4, $5, $6 );
    return timegm( $se, $mi, $hr, $dy, $mn - 1, $yr - 1900 );
}

=head2 version

Return node version as a string

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> version as a string

=item B<Failure:> no failure scenario

=back

=item Examples

=back

	my $node_version = $node->version;

=cut

sub version {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->server->dblade( cmd => 'version' )
}

=head2 sysconfig

Returns the output of "sysconfig -v" of node

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> output of "sysconfig -v" as a string

=item B<Failure:> no failure scenario

=back

=item Examples

=back

    my $node_sysconfig = $node->sysconfig;

=cut

sub sysconfig {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return 'skipped by request' if Tharn::param('GNOTE_NOSYSCONFIG') =~ /\w+/;
    return shift->server->dblade( cmd => 'sysconfig -v')
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Restoring the wafl flags
    if (defined $self->{_wafl_flag}) {
        foreach my $flag (keys(%{$self->{_wafl_flag}})) {
            $self->setflag(name => $flag, value => $self->{_wafl_flag}{$flag} );
        }
    }

    # Enabling the Cluster HA config
    $self->server->cluster_ha_modify( configured => "true") if $self->{_disabled_ha};
    # Enabling the ntp services after modifying the date with ntp_override
    $self->_ntp_server_modify("enable") if $self->{_ntp_flag};
    $self->zerospares if $self->{_do_zero_spares};
    map{ $$self{_fiji}{$_}{keepfiji} ? undef : $self->fiji_disable('fault-id'=>$_ ) } keys %{$$self{_fiji}} if $$self{_fiji};	# remove FIJI if registered to be destroyed
    $self->server->shell(cmd=>"rm ".$self->{_stampfile} ) if $self->{_stampfile};						# remove timestamp file from FreeBSD
    $self->privilege( level => $self->{_original_privilege} );									# restore privilege level so negative tests will fail where suppose to
}

=head2 privilege

Sets or gets the privilege level for this node.
In the case of setting a new level, the previous
level is returned.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<level > => I<admin|advance|diagnostic|test> 

=back

=item Returns

=over

=item B<Success:> 

In the case getting, returns the existing level
In the case of setting to a new level, the previous level

=item B<Failure:> 

No failure scenario

=back

=item Examples

=back

     #To get the current privilege level
     my $curr_priv = $node->privilege;

     #To set it to new privilege level of diag
     #This returns earlier level, in case earlier privilege level is not diagnostic
     #If the earlier level is also diagnostics, it just returns tis level, without setting
     my $curr_priv = $node->privilege(level => 'diagnostic');

=cut

sub privilege {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $out        = $self->server->set();
    my %info       = map { split /:/, $_ } split /\n/, $out;      # if it's plain text, then split by ':', or see BURT 394938
    if( $out =~ /X;X/ ) {                                         # on how to make sence of 'X;X' see lib/Node/NG.pm _cli_output_handler()
        my @all;                                                  # sub array of names (or values)
        push @all, [ split( /X;X/,$_ ) ] foreach(split /\n/,$out);
        $info{$_} = shift @{$all[2]} foreach( @{$all[1]} );
    }

    my $curr_level = $info{'Privilege Level'};
    $curr_level =~ s/(^\s+|\s+$)//g;                                # Remove leading white space
    if ( exists $args{level} && $args{level} ne $curr_level ) {
        $self->server->set( privilege => $args{level}, confirmations => "off" );
        NOTE::logcomment( $self->name . ": privilege level changed from $curr_level to " . $args{level} );
    }
    else {
        NOTE::logcomment( $self->name . ": current privilege level: $curr_level" . ( exists $args{level} ? " (unchanged)" : "" ) );
    }
    return $curr_level;
}

=head2 exec

Execute a single ONTAP command at a specific privilege level, using server object.
This makes it equivanet to calling the provided cdef.
Privilege level is set back to previous level, after the execution of the command.
This method takes two arguments, cmd and level, and other parameters are
pass-thru to ONTAP for processing.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<cmd> => I<string> 

command to be executed.
Ex: vserver_show (i.e not "vserver show")
This is some thing like a cdef function name.

=item B<level> => I<admin|advance|diagnostic|test>

privilege level with which command is to be executed

All the arguments required for the command(cdef) to be executed are 
to be passed, as part of the hash. The has of arguments 
(excluding level and cmd) are passed to the command.
Please refer each command cdef parameters, before providing.

=back

=item Returns

=over

=item B<Success:>

Reference to an array of hash references. sometimes raw output(most of server calls return this).
However some return the output as string.
Better checking for the type returned, and parse appropriately.

=item B<Failure:>

confess, in case of error executing the command using server object

=back


=item Notes

=over

=item 1 

In either case of success/failure, the privilege is always set back to previous level.

=back

=item Examples

=back

    # Enable NFSv4 for vserver in test mode
    $node->exec(level=>"test", cmd=>"vserver_nfs_create", vserver=>name, vs=>enabled);	

=cut

sub exec {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing level parameter."
      unless exists $args{level};
    my $new_level = delete $args{level};

    confess "Missing cmd parameter"
      unless exists $args{cmd};
    my $cmd = delete $args{cmd};

    my $old_level = $self->privilege( level => $new_level );
    my $result;
    eval { $result = $self->server->$cmd(%args); };
    my $error = $@;    # Save error for return to called
    $self->privilege( level => $old_level )
      if $old_level ne $new_level;

    if ($error) {
        confess $error;
    }
    else {
        return $result;
    }
}

=head2 fiji_show

Returns array reference of debug fault-injection show hashes
i.e FIJI (FIPs) hashes.
If no arguments are provided, it returns all the fiji ids.
If one or more below mentioned arguments are passed. fiji ids
matching the criteria will be returned.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<fault-id>  => I<string>

=item B<fault-mode>  => I<string>

=item B<probability>  => I<string>

=item B<period>  => I<string>

=item B<num-calls>  => I<string>

=item B<start-call>  => I<string>

=item B<end-call>  => I<string>

=item B<enabled>  => I<string>

=item B<total-calls>  => I<string>

=item B<action>  => I<string>

=item B<action-arg>  => I<string>

=item B<node>  => I<string>

=back

=item Returns

=over

=item B<Success:> Refernce to an array of hash references

=item B<Failure:> No failure scenario

=back

=item Notes

=over

=item 1 Refer the cli "debug fault-injection show" 

=back

=item Examples

=back

    #To get the fiji show hashes and print all fiji flags fault-id status
	my $node = NOTE::GX::Node->connect(name=>$FILER);
    my $fips = $node->fiji_show;
    foreach( @$fips ) {
        print "Fiji: " . $_->{'fault-id'} . " set to " . $_->{'enabled'} . "\n";
    }

    #Return single FIJI for specific name of fault-id
    my $node = NOTE::GX::Node->connect(name=>$FILER);
    my $fiji = ( $node->fiji_show( 'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion' ) )[0];

=cut

sub fiji_show {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->privilege( level => 'test' );
    $self->server->debug_fault_injection_show( %args );
}

=head2 fiji_enabled

Returns array reference of hash references, containing the fiji 
ids which are enabled.
If no arguments are provided, it checks all the fiji ids.
If one or more below mentioned arguments are passed. fiji ids 
matching the criteria and enabled will be returned.


=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<fault-id>  => I<string>

=item B<fault-mode>  => I<string>

=item B<probability>  => I<string>

=item B<period>  => I<string>

=item B<num-calls>  => I<string>

=item B<start-call>  => I<string>

=item B<end-call>  => I<string>

=item B<enabled>  => I<string>

=item B<total-calls>  => I<string>

=item B<action>  => I<string>

=item B<action-arg>  => I<string>

=item B<node>  => I<string>

=back

=item Returns

=over

=item B<Success:> Refernce to an array of hash references

=item B<Failure:> No failure scenario

=back

=item Examples

=back

    #Return single FIJI for specific name of fault-id and which is enabled
    my $node = NOTE::GX::Node->connect(name=>$FILER);
    my $fiji = ( $node->fiji_enabled( 'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion' ) )[0];	

=cut

sub fiji_enabled {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->privilege( level => 'test' );
    my @out = grep { $_->{'enabled'} eq 'true' }  $self->server->debug_fault_injection_show( %args );
    return \@out;
}

=head2 ::fiji_enable

Enables FIJI and returns hash reference of enabled FIJI (FIPs)

=over

=item Arguments

B<keepfiji> => I<boolean>

I<Optional> NOTE will not disable fiji during teardown if keepfiji parameter is true (other then undef or zero)
Default is zero, by default FIJI is removed when script exits


Other parameters are same as for B<debug fault-injection disable-fault> filer's command. See ONTAP manual for details. One of them
is required.

B<'fault-id'> => I<string>

I<Required> FIJI name. See examples below.

B<'node'> => I<string>

I<Optional> => Node to set FIJI on. Default is node object points to.

=item Examples

 my $node = NOTE::GX::Node->connect(name=>$FILER);
 my $fiji = $node->fiji_enable( keepfiji => 1, 'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion' );

Output example:

 {
   'num-calls' => '0',
   'period' => '0',
   'end-call' => '10000',
   'probability' => '0',
   'fault-mode' => 'GX_FAULT_MODE_ALWAYS',
   'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterDstWritable',
   'start-call' => '1',
   'enabled' => 'true'
 }

=item Design notes

fiji_enable method creates hash entry for each FIJI as follows:

 $node->{_fiji} = {
		    'GX_FAULT_SM_...' => {
				 	  timestamp => output from time(), # time when FIJI was enabled
				 	  keepfiji => 0,		   # disable FIJI during teardown, 1 - means keep it
				 	  status => {
				 	              'num-calls' => '0',
				                       'period' => '0',
		                                       'end-call' => '10000',
                                                       'probability' => '0',
                                                       'fault-mode' => 'GX_FAULT_MODE_ALWAYS',
                                                       'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterDstWritable',
                                                       'start-call' => '1',
                                                       'enabled' => 'true'
                                                     },
				 	 }
		};

When fiji_enable gets called it checks if it's enabled and if timestamp is already set, if it's already set
then timestamp doesn't get reset. If timestamp is not defined, then local time is used to set timestamp.

Localtime is used as timestmap because mgwd log stores time in localtime, not UTC.

=back

=cut

sub fiji_enable {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $keepfiji = delete $args{keepfiji};

    $self->privilege( level => 'test' );
    $args{node} ||= $self->name;

    $$self{_fiji}{ $args{'fault-id'} }{timestamp} = time() unless( $$self{_fiji}{ $args{'fault-id'} }{timestamp} );
    $$self{_fiji}{ $args{'fault-id'} }{keepfiji} = $keepfiji;
    $self->server->debug_fault_injection_enable_fault( %args );
    $$self{_fiji}{ $args{'fault-id'} }{status} = ($self->server->debug_fault_injection_show(node=>$args{node}, 'fault-id' => $args{'fault-id'}))[0];
    unless( $$self{_fiji}{ $args{'fault-id'} }{status}{enabled} eq 'true' ) {
	my $error = "Unable to set fiji on node ". $self->name . ", current state:\n". Dumper($$self{_fiji}{ $args{'fault-id'}});
	delete $$self{_fiji}{ $args{'fault-id'} };
	confess $error;
    }
    return $$self{_fiji}{ $args{'fault-id'} }{status};
}

=head2 ::fiji_wait

Waits for FIJI to be triggered.

=over

=item Arguments

B<-timeout> => I<number of seconds>

I<Optional> Number of seconds to wait for FIJI to be triggered. Default is 900 seconds (15 minutes). Please do not forget it's C<-timeout>,
not a C<timeout> and it's not a B<-Timeout>. The B<-Timeout> will be passed to server and it indicates cli command timeout,
not a wait for FIJI timeout, our is all lower case with '-' in front of it.

B<-sleep> => I<number of seconds>
I<Optional> Number of seconds to wait between status and log checks. Default is 10 seconds.

B<-usemgwd> => I<boolean>
I<Optional> Use mgwd log file for detection. Default is true.

B<-useshow> => I<boolean>
I<Optional> Use debug fault-injection show' command during check. Default is true. This option will become true if
both -useshow and -usemgwd are negative because you must have at least one of them and filer's command is safest

Other parameters are same as for B<debug fault-injection disable-fault> filer's command. See ONTAP manual for details. One of them
is required.

B<'fault-id'> => I<string>

I<Required> FIJI name. For example 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion'.

=back

=cut

sub fiji_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $usemgwd = delete $args{'-usemgwd'};
    my $useshow = delete $args{'-useshow'};

    $useshow = 1 unless( $usemgwd | $useshow );										    # force safest 'debug fault-injection show' command
															    # in case both of check methods are 0

    my $timeout = ( $args{'-timeout'} ? delete $args{'-timeout'} : 900 ) + time();					    # time to wait for FIJI or fail
    my $sleep   = $args{'-sleep'} ? delete $args{'-sleep'} : 10;							    # how many seconds to wait between checks

    confess "Fiji ".$args{'fault-id'}." was not set" unless( $self->fiji_enabled( %args ) );				    # fail if FIJI is not actually set on a filer
    my $start_time = $$self{_fiji}{ $args{'fault-id'} }{timestamp} ? $$self{_fiji}{ $args{'fault-id'} }{timestamp} : time();  # get timestamp starting point for log file
    my $faultid = ( split(/_/, $args{'fault-id'}) )[-1];								    # log files has shorterned version of FIJI name
    my $called = ($self->server->debug_fault_injection_show( %args ))[0]->{'num-calls'};				    # get original FIJI count, in case if it was already triggered

    do {
	if( $usemgwd ) {
	    my $found = $self->mgwd_log( epoch => $start_time, regex => "Fault Injection Point:.*$faultid.*invoked" );	    # return if log file indicates FIJI was triggered
	    return $found if( $found =~ /Fault Injection Point/ );							    # will return actual FIJI string from log
	}

	if( $useshow ) {
	    my $current_count = ($self->server->debug_fault_injection_show(%args))[0]->{'num-calls'};			    # another way to check is to watch for counter from
	    return $current_count if( $current_count & ( $called ^ $current_count ) );					    # show command and return 1 if counter was incremented
	}

	sleep $sleep;													    # sleep for about 10 seconds between checks
    } while( time() < $timeout );
    confess "Timeout waiting for ".$args{'fault-id'}." to be triggeregd on node ". $self->name;
}

=head2 ::fiji_disable

Disables FIJI and returns hash reference of disabled FIJI (FIPs)

=over

=item Arguments

Parameters are same as for B<debug fault-injection disable-fault> filer's command. See ONTAP manual for details. One of them
is required.

B<'fault-id'> => I<string>

I<Required> FIJI name. For example 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion'.

B<'node'> => I<string>

I<Optional> => Node to set FIJI on. Default is node object points to.

=item Examples

 my $node = NOTE::GX::Node->connect(name=>$FILER);
 my $fiji = $node->fiji_disable( keepfiji => 1, 'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion' );

Output example:

 {
   'num-calls' => '0',
   'period' => '0',
   'end-call' => '10000',
   'probability' => '0',
   'fault-mode' => 'GX_FAULT_MODE_ALWAYS',
   'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterDstWritable',
   'start-call' => '1',
   'enabled' => 'false'
 }

=back

=cut

sub fiji_disable {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->privilege( level => 'test' );
    $args{node} ||= $self->name;
    $self->server->debug_fault_injection_disable_fault( %args );

    my $status = ( $self->server->debug_fault_injection_show( node=>$args{node}, 'fault-id' => $args{'fault-id'} ) )[0];
    unless( $$status{ 'enabled' } ne 'true' ) {
	confess( "Unable to clear fiji on node ". $self->name . ", current state:\n". Dumper( $$self{_fiji}{ $args{'fault-id'} } ));
    }
    delete $$self{_fiji}{ $args{'fault-id'} } if( $$self{_fiji}{ $args{'fault-id'} } );

    return $status;
}

=head2 ::fiji_setmode

Wrapper around 'debug fault-injection set-fault-mode' command

=over

=item Arguments

Parameters are same as for B<debug fault-injection set-fault-mode> filer's command.

=back

=cut

sub fiji_setmode {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->privilege( level => 'test' );
    $self->server->debug_fault_injection_set_fault_mode( %args );
}

=head2 ::bsdshell

Executes command on FreeBSD. It's "AS IS" wrapper around shell() method of Server.pm library.
See server.pm library for details.

=over

=item Arguments

B<-Timeout>

How long to wait for output from specified command.  Default value is 20
seconds.  Supplied value must be in seconds.

=back

=cut

sub bsdshell {
    my ($self, @args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->server->shell( @args );
}

=head2 find_changed_logs

Returns array of names of log files changed/modified since script start

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> array of log file names

=item B<Failure:> no failure scenario

=back

=item Examples

=back

    #Getting the changed log files since testscript start
	my @changed_logfiles = $node->find_changed_logs;

=cut

sub find_changed_logs {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Timestamp file was not created, unable to search for changes" unless( $self->{_stampfile} =~ /\w+/ );
    my @changed = grep {!/No such file or directory/} split(/[\n\r]/, $self->bsdshell(cmd=>"find /mroot -type f -newer ". $self->{_stampfile}) );
    push @changed, grep {!/No such file or directory/} split(/[\n\r]/, $self->bsdshell(cmd=>"find /var/log -type f -newer ".$self->{_stampfile}) );
    return @changed;
}

=head2 dump_bsd_file

Return content of the provided file from FreeBSD space.
However no check is done, whether file exists or not.
i.e. if the file does not exist, the content can be empty or contain error string

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<filename> => I<abosulte path of file>

=back

=item Returns

=over

=item B<Success:> file content as string

=item B<Failure:> retruned content can have no content or error string.

=back

=item Examples

=back

	my $mgwd_log_content = $node->(filename => '/mroot/etc/log/mlog/mgwd.log');

=cut

sub dump_bsd_file {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $filename = delete $args{filename};
    return $self->bsdshell( cmd => "cat $filename", %args );
}

=head2 mgwd_log

Returns content of mgwd.log(s) generated since script start.
Also, by providing the time and/or regular expression, content matching 
the provided criteria is returned as a string.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<epoch> => I<number>

Start output from specific time instead of beginning of the file. 
Number is number of seconds since epoch.

=item B<regex> => I<string>

Grep mgwd content by given RegEx

=back

=item Returns

=over

=item B<Success:> 

mgwd log file content as a string, matcing the given criteria.
if no criteria is provided, starting from the beginning of the file, 
since start of the test script

=item B<Failure:>

No failed scenario

=back

=item Examples

=back

    #To get the mgwd log from a specific time
	my $node = NOTE::GX::Node->connect(name=>$FILER);
    my $startfrom = time();
    # do something here
    my $latest_mgwd = $node->mgwd_log( epoch => $startfrom );

    #Check whether a particluar fiji id when set, is appeared in mgwd log file or not
    my $node = NOTE::GX::Node->connect(name=>$FILER);
    my $startfrom = time();
    $node->fiji_enable( 'fault-id' => 'GX_FAULT_SM_HIT_PromoteAfterResetDataVersion' );
    unless( $node->mgwd_log( epoch => $startfrom, regex => 'PromoteAfterResetDataVersion' ) ) {
        die "FIJI is not set!";
    }

=cut

sub mgwd_log {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %abbr = ( Jan=>0, Feb=>1, Mar=>2, Apr=>3, May=>4, Jun=>5, Jul=>6, Aug=>7, Sep=>8, Oct=>9, Nov=>10, Dec=>11 );
    my @mgwds = sort grep {/\/mroot\/etc\/log\/mlog\/mgwd\.log(\.previous\.\d)*$/} $self->find_changed_logs();
    push @mgwds, '/mroot/etc/log/mlog' unless( @mgwds );
    my $out = undef;
    while( @mgwds ) {
	my $filedata = $self->dump_bsd_file( filename => pop @mgwds, -Timeout => 300 );
	if( $args{epoch} ) {
	    foreach my $line ( split(/[\n\r]/, $filedata) ) {
		# extracting date like Nov 23 2009 00:44:40 from
		# 00000001.00e4e0f6 060b7d53 Mon Nov 23 2009 00:44:40 -08:00 [kern_mgwd:info:1842] telnet :: localhost :: diag :: run -node dpg-drwho9 -command " version" :: Pending
		if( $line =~ /^[0-9a-f]+\.[0-9a-f]+\s+[0-9a-f]+\w+\s+\w+\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+):(\d+):(\d+)/) {
		    my ($mn, $dy,$yr,$hr,$mm,$ss) = ( $abbr{$1},$2,$3 - 1900,$4,$5,$6 );
		    my $timestamp = Time::Local::timelocal( $ss,$mm, $hr, $dy, $mn, $yr );
		    $out .= "$line\n" if( $timestamp > $args{epoch} );
		}
	    }
	} else {
	    $out .= $filedata;
	}
    }

    $out = join("\n", grep(/$args{regex}/, split(/\n/, $out))) if( $args{regex} );
    return $out;
}

=head2 dump_bsd_file

Returns a hash containing build information contained in the 
BUILD file on the filer.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> Hash containing the build info

=item B<Failure:> No failure scenario

=back

=item Notes

=over

=item 1  The output hash definition. 

=over

        $build_info->{BRANCH} => build release
        $build_info->{CLIENT} => client
        $build_info->{DELTAS} => deltas
        $build_info->{TIME} => build date
        $build_info->{CHANGE} => change number
        $build_info->{TYPE} => ontap build type, e.g. DEBUG
        $build_info->{VERSION} => build version

=back

=back

=item Examples

=back

    #to gte the build info hash
    my %build_info = $node->build_info();

=cut

sub build_info {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %build_info;
    my $build_contents;

    ## The format of the build file is:
    ## CLIENT: R8.0xN_091027_2100
    ## CHANGE: 984151
    ## DELTAS: unknown
    ## TIME: 091028001624PDT
    ## BRANCH: R8.0x
    ## TYPE: OPTIMIZE
    $build_contents = $self->server->shell( cmd=> "cat /BUILD" );
    %build_info = map { (split/:\s+/) } grep /\w+/, split /[\n\r]/, $build_contents;

    ## Note important pieces of build information and store it
    ## in the object, so when queried we can return that information
    ## instead of going to the filer
    $self->{_build_release} = $build_info{BRANCH};
    $self->{_build_type} = $build_info{TYPE};
    $self->{_build_date} = $build_info{TIME};

    ## Get the software version
    $self->{_software_version} = $self->server->shell( cmd=> "cat /VERSION");
    $build_info{VERSION} = $self->{_software_version};

    return %build_info;
}

=head2 build_release

Returns the release of the build e.g. R8.0, Rboilermaker, as a string

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> build release as a string

=item B<Failure:> No failure scenario

=back

=item Examples

=back

    #to get the build release
    my $build_release = $node->build_release;

=cut

sub build_release {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->build_info() unless defined($self->{_build_release});
    return $self->{_build_release};
}

=head2 build_type

Returns the type of the build e.g. OPTIMIZE, DEBUG, Regular, as a string

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> build type as a string

=item B<Failure:> No failure scenario

=back

=item Examples

=back

    #to get the build type
    my $build_type = $node->build_type;

=cut

sub build_type {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->build_info() unless defined($self->{_build_type});
    return $self->{_build_type};
}

=head2 build_date

Returns the date of the build as a string.
The format is 091028001624PDT.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> build date as a string

=item B<Failure:> No failure scenario

=back

=item Examples

=back

    #to get the build date
    my $build_date = $node->build_date;

=cut

sub build_date{
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->build_info() unless defined($self->{_build_date});
    return $self->{_build_date};
}

=head2 build_version

Returns the software version of the build as a string.
Eg: OPTIMIZE, DEBUG, Regular

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> software version of build as a string

=item B<Failure:> No failure scenario

=back

=item Examples

=back

    #to get the build software version
    my $build_date = $node->build_version;

=cut

sub build_version {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->build_info() unless defined($self->{_software_version});
    return $self->{_software_version};
}

=head2 ::autoboot_value

Returns a string containing the AUTOBOOT value for a Node. If a value is passed "true/false"
value is set and old value is returned

=over

=item Arguments

B<true|false> => I<string>

=item Examples

my $current_value= $vol_src->Node->autoboot_value();
my $old_value= $vol_src->Node->autoboot_value("false");

=back

=cut

sub autoboot_value{
    my ($self,$value) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $output =
         $self->server->dblade(cmd => 'priv set -q diag;bootargs get AUTOBOOT');

    # truncating the output string to remove extra info
    $output =~ s/AUTOBOOT == //;

    if($value =~ /^(true|false)$/i) { # setting the autoboot value
        $value = lc($1);
        $self->server->dblade(
                       cmd => "priv set -q diag;bootargs set AUTOBOOT $value");
        return $output;
    }elsif($value =~ /^$/){	#retrieving the autoboot value
        return $output;
    }else{	# croak if invalid option
        croak "value of AUTOBOOT can only be true|false not $value";
    }
}

=head2 ::get_date

Returns a array of hash ref which contains the date,time,timezoner,utc-date

=over

=item Arguments

B<all_nodes> => I<string>
default is local node

=item Examples

my ($node_date) = $vol_src->Node->get_date;
$date = $node_date->{date};

my @cluster_node_dates = $vol_src->Node->get_date("all_nodes");
node_1_date = $cluster_node_dates[0]->{date};
node_1_name = $cluster_node_dates[0]->{node};
node_2_name = $cluster_node_dates[1]->{node};
node_2_date = $cluster_node_dates[1]->{date};

=back

=cut

sub get_date{
    my ($self,$choice) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $node_name = $self->name;
    $node_name = "\*"
        if( $choice =~ /all_nodes/i);
    my @output = $self->server->system_node_date_show(node => $node_name);
    return @output;
}


=head2 ::set_date

set the date,time, timezone etc on the node
confesses, if error in setting the date.
returns nothing.

=over

=item Arguments

B<all_nodes> => I<string>
Default is "false" runs on local node
To run command on all nodes set to "true"

B<date> => I<string>
please see documentation on filer to set this value

B<time_zone> => I<string>
please see documentation on filer to set this value

B<ntp_override> => I<string>
use true|false to modify ntp server on the filer

=back

=over

=item Examples

$vol_src->Node->set_date(all_nodes => 'true',
			             date => "3/2/2011 11:07:15",
                         ntp_override => "true",
                         time_zone => "US/Pacific");

$vol_src->Node->set_date( date => "3/2/2011 11:07:15",
                         time_zone => "US/Pacific");

$vol_src->Node->set_date( date => "3/2/2011 11:07:15");

=back

=cut

sub set_date{
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts;
    my $old_ntp_status;
    my $flag = 0;
    my $over_ride = 0;

    # get the ntp server status
    $old_ntp_status = $self->server->system_services_ntp_config_show();
    $flag = 1 if $old_ntp_status->[0]->{'enabled'} eq "true";
    $over_ride = 1 if  $args{ntp_override} eq "true";
    # Disabling ntp service before modifying date if ntp_override is true and ntp service is on
    if ($over_ride && $flag) {
        $self->{_ntp_flag} = 1;
        $self->_ntp_server_modify("disable");
    }
    my $node_name = ($args{all_nodes} =~ /true/i) ? "\*" : $self->name;
    $opts{date}= $args{date} if (defined ($args{date}));
    $opts{timezone}= $args{time_zone} if (defined ($args{time_zone}));

    #Need to make this server cdef call in eval, as this call might fail, 
    #due to the UI change: Burt#456876
    my $eval_ret = eval {
                       $self->server->system_node_date_modify(node => $node_name,%opts);
                       1;
                   };
    NOTE::logcomment("system node date modify eval return code: $eval_ret");
    if(!$eval_ret) {
        confess "Error changing the system node date modifying. Error: $@";
    }
    return;
}

# helper function to enable/disable the ntp service on the node
sub _ntp_server_modify{
    my ($self, $action) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment("$action the NTP Service on the node: ".$self->name);
    if ($action eq "enable"){
        $action = "true";
    } elsif ($action eq "disable") {
        $action = "false";
    } else {
        confess "incorrect value: only disable/enable allowed as acceptable actions";
    }
    $self->server->system_services_ntp_config_modify(enabled => $action);
    $self->server->system_services_ntp_config_show();
}

=head2 transfer_engine

Sets the Paloma physical flag to true or false for this node.
Paloma physical flag is set to true, means transfer engine is BRE.
Paloma physical flag is set to false, means transfer engine is not BRE.

=over

=item Arguments

I<Optional Arguments>

=over 

=item B<engine> 

To set the transfer engine to bre, pass value as 'bre'.
any other value or not providing this parameter means, setting it to non bre.

=back

=item Returns

=over

=item B<Success:> 

does not return anything

=item B<Failure:> 

confess, if not able to set or not able to bring the privilege back to original. 

=back

=item Notes

=over

=item 1 

The system is set to privilege level of 'test' for setting 
this Paloma physical flag. After setting the flag, the privilege
level is set back to original level.

=back

=item Examples

=back

	#To set the transfer engine to bre
    $node->transfer_engine('BRE');

    #To set to non-bre
    $node->transfer_engine();
            or
    $node->transfer_engine('LRE');

=cut

sub transfer_engine {
    my ($self, $engine) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my  $cmd = "set test; debug smdb table node_set_kenv run -node " . $self->name . " -var paloma.physical -val ";

    if ( $engine =~ /bre/i ) {
        $self->{_BRE} = 1;
        $cmd = $cmd . "true";
    } else {
        $cmd = $cmd . "false";
    }
    my $old_level = $self->privilege(level=>"test");
    NOTE::logcomment("Setting BRE flag on " . $self->name );
    eval {
        $self->server->cli( cmd => $cmd );
    };
    $self->privilege(level=>$old_level);
    if ( $@ ) {
        confess "Could not set BRE as transfer engine: $@";
    }
}

=head2 eventlog_msg_get_aref

Get the event log messages as an array ref, that contains hash 
refernces to each event
   - if messagename is provided as arg, the array contains events 
     matching the messagename
   - if node_name is provided, array contains the event log 
     messages generated for that node
   - if start_timestamp, end_timestamp are provided, array 
     contains messages fall into the given time range.
And it returns reference to an empty array, if there are no matching events

=over

=item Arguments

Arguments passed as an hash reference

Hash table of arguments:

I<Optional Arguments>

=over 

=item B<messagename > => I<string> 

specific eventlog messagename to search for

=item B<node> => I<string> 

eventlog messages of this node are considered

=item B<start_timestamp> => I<timestamp> 

this makes function to look for eventlog messages generated after this timestamp.
timestamp can be time stamp returned by the function Node::datetime.

=item B<end_timestamp> => I<timestamp>

this makes function to look for eventlog messages generated before this timestamp.
timestamp can be time stamp returned by the function Node::datetime.

=back

=item Returns

=over

=item B<Success:> array ref containing references to hashes. 

=item B<Failure:> confesses, if error in getting the eventlog messages

=back


=item Notes

=over

=item 1 

The array reference that contains references to hashes looks like,
as mentioned below. 
Note that the number of hash reference are equal to the number of 
event log messages getting returned.

=over

        ( {"severity" => "-severity",
           "ems-severity" => "-ems-severity",
           "source" => "-source",
           "messagename" => "-messagename",
           "event" => "-event",
           "kernelgen" => "-kernelgen",
           "kernelseqnum" => "-kernelseqnum",
           "event-xml-len" => "-event-xml-len",
           "num-suppressed-since-last" => "-num-suppressed-since-last"},)

=back

=back

=item Examples

=back

    #To get the event log messages generated on node $node,
    # and generated in between the timestamps of $start_timestamp, 
    # and $end_tiemstamp
	my $msg_aref = $node1->eventlog_msg_get_aref({
                                   messagename     => $message_name,
                                   node            => $node,
                                   start_timestamp => $start_timestamp,
                                   end_timestamp   => $end_tiemstamp,
                                  });

    #To get all the event log messages generated on the node $node
    my $msg_aref = $node1->eventlog_msg_get_aref({
                                   messagename     => $message_name,
                                   node            => $node,
                                  });

=cut

sub eventlog_msg_get_aref {
    my ($self, $args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($start_timestamp, $end_timestamp, $messagename, $node);

    if(defined $args->{messagename}){
        $messagename = $args->{messagename};
    }
    if(defined $args->{node}){
        $node = $args->{node};
    }
    if(defined $args->{start_timestamp}){
        $start_timestamp = $args->{start_timestamp};
    }
    if(defined $args->{end_timestamp}){
        $end_timestamp = $args->{end_timestamp};
    }

    my %cdef_args;
    $cdef_args{node} = $node               if(defined $node);
    if(defined $messagename){
        # Message format changed: see burt # 507167
        # Necessary to return message with asterisk terminator
        if($messagename !~ /[*]$/){
            $messagename .= q{*};
        }
        $cdef_args{messagename} = $messagename;
    }

    my $total_eventlog_msg_aref;
    my $event_msg_aref = [];
    my $eval_return = eval {
                         $total_eventlog_msg_aref = $self->server->event_log_show(%cdef_args);
                         1;
                      };
    if(!$eval_return){
        #means, $@ is set some value
        confess "Could not get event log messages: error: $@";
    }

    ## Obtain timezone in order to fix burt609510
    my $date = $self->datetime();
    my $timezone = ${[split(' ',$date)]}[2];

    EVENT_LOG_PARSE:
    foreach my $temp_event_hash_ref (@{$total_eventlog_msg_aref}){

        ## The key "time" from hash does not contain timezone information.
        ## So calling parsedate on this would be interpreted as a time 
        ## in the timezone of the client. Hence it may lead to issues like burt609510
        ## Hence the timezone needs to be obtained from the filer and
        ## appended to event_log_timestamp
        my $event_log_timestamp = $temp_event_hash_ref->{time};
        $event_log_timestamp =~ s/"//g;
        $event_log_timestamp = $event_log_timestamp." ".$timezone if(scalar(split(' ',$event_log_timestamp)) == 2);
        my $this_event_considered = 0; #False
        if(defined $start_timestamp){
            if(parsedate($event_log_timestamp) >= parsedate($start_timestamp) ){
                if(defined $end_timestamp){
                    if( parsedate($event_log_timestamp) <= parsedate($end_timestamp) ) {
                        $this_event_considered = 1; #True
                    }
                }
                else {
                    $this_event_considered = 1; #True
                }
            }
        }
        else {
            if(defined $end_timestamp){
                if(parsedate($event_log_timestamp) <= parsedate($end_timestamp) ) {
                    $this_event_considered = 1; #True
                }
            }
            else {
                $this_event_considered = 1; #True
            }
        }

        if($this_event_considered) {
            push @{$event_msg_aref}, $temp_event_hash_ref;
        }
    }
    return $event_msg_aref;
}

=head2 snapmirror_log_get_job_href

Get the snapmirror log messages as a hash ref(empty hash in case of no messages) 
 for the given snapmirror job id (By default it
checks present day snapmiror log - see parameters below to change this).

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<job_id> => I<job_id> job id for which snapmirror log generated

=back

I<Optional Arguments>

=over 

=item B<day> => I<integer> 

snapmirror log to search for , based on the day.
0  => present day log
-1 => previous day log
-2 => 2 days before log, and so on.
I<Default>: 0 (means present day)

=back

=item Returns

=over

=item B<Success:> 

reference to an hash. see below for the hash structure.
reference to an empt hash, in case of no snapmirror log messages

=item B<Failure:> 

confesses, in case of error getting snapmirror log content

=back


=item Notes

=over

=item 1 The hash that this method returns looks like...

       {$msg_num}{job_id}
       {$msg_num}{job_type}
       {$msg_num}{job_action}
       {$msg_num}{job_action_status}
       {$msg_num}{job_action_time_stamp}
       {$msg_num}{job_action_message_type}
       {$msg_num}{job_action_message}

$msg_num ranges from 1 to total number messages generated for this job id.

=back

=item Examples

=back


    #To get for the present/current day for $job_id
	my $snapmirror_log_msg_href = $node1->snapmirror_log_get_job_href(
                                      job_id => $job_id,
                                      day    => 0,
                                  );

=cut

sub snapmirror_log_get_job_href {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($job_id, $day, $job_type, $job_action, $job_action_status, $job_action_time_stamp);
    my ($job_action_message_type, $job_action_message);

    $job_id = $args{job_id};
    $day    = $args{day} || 0;

    if( (!defined $job_id) or ($job_id !~ /^\d+$/) ){
        croak 'valid job_id needs to be provided';
    }
    if( ($day !~ /^[-]?\d+$/) or ($day > 0 ) ){
        croak 'Provide 0 or -ve value for parameter day';
    }
    my $snapmirror_log_filename;
    $day =~ s/-//;
    if($day == 0){
        $snapmirror_log_filename = 'snapmirror.log';
    }
    else {
        $snapmirror_log_filename = 'snapmirror.log.previous.' . $day;
    }

    my $snapmirror_log;
    my $cmd = 'cat /mroot/etc/log/mlog/' . $snapmirror_log_filename . " | grep \":$job_id \"";
    my $eval_return = eval {
                         $snapmirror_log = $self->server->shell(cmd => $cmd);
                         1;
                      };
    if(!$eval_return){
        #means, $@ is set some value
        confess "Could not get snapmirror.log content: error: $@";
    }

    my @lines = split /\n/, $snapmirror_log;

    my $snapmirror_log_href = {};
    my $msg_num = 1;
    foreach my $line (@lines){
        #The message line should follow the pattern mentioned in the wiki:
        # http://wikid.netapp.com/w/DPG/UCM/Snapmirror_Logging_DS
        # If not, discard the message

        if ( $line =~ m{^
                         #Checking for the Boot sequence number and message sequence number in hex
                         [\da-fA-F]+[.][\da-fA-F]+\s+
                         #Checking for the Time delta representation
                         [\da-fA-F]+\s+
                         #Checking for the Time string
                         (\w+\s+\w+\s+\w+\s+\w+\s+\d\d:\d\d:\d\d\s+[+-]\d\d:\d\d)\s+
                         #Checking for the Source, log level, pid of message
                         \[kern_snapmirror:(\w+):\d+\]\s+
                         #Checking for the "thread ID in hex", snapmirror operation,Job number
                         0x[\da-fA-F]+:(\w+):$job_id\s+
                         #Checking for the action
                         action=(\w+)\s+
                         #Getting the rest of the message
                         (.*)
                         }xi ) {
            $job_action_time_stamp= $1;
            $job_action_message_type = $2;
            $job_type = $3;
            $job_action = $4;
            $job_action_message = $5;

            #The remainder of the message which is job_action_message might contain the status, for specific types of actions
            if($job_action_message =~ /status=(\w+)\s+(.*)/i ) {
                $job_action_status = $1;
                $job_action_message = $2;
            }
            else {
                $job_action_status = q{};
            }

            $snapmirror_log_href->{$msg_num}{job_id}                 = $job_id;
            $snapmirror_log_href->{$msg_num}{job_type}               = $job_type;
            $snapmirror_log_href->{$msg_num}{job_action}             = $job_action;
            $snapmirror_log_href->{$msg_num}{job_action_status}      = $job_action_status;
            $snapmirror_log_href->{$msg_num}{job_action_time_stamp}  = $job_action_time_stamp;
            $snapmirror_log_href->{$msg_num}{job_action_message_type} = $job_action_message_type;
            chomp $job_action_message;
            $snapmirror_log_href->{$msg_num}{job_action_message}     = $job_action_message;

            $msg_num++;
        }
    }
    return $snapmirror_log_href;
}

=head2 mgwd_log_check_msg

Check the existence of the given message in the mgwd log
 - Returns 1 for true
 - Returns 0 for false

- By default it checks present day mgwd log 
(see parameters below to change this).

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<message> => I<string> message to search for

=back

I<Optional Arguments>

=over 

=item B<day> => I<integer>

mgwd log to search for , based on the day.
0  => present day log
-1 => previous day log
-2 => 2 days before log, and so on.
I<Default>: 0 (means present day)

=back

=item Returns

=over

=item B<Success:> 

1 if the message is presnt in the mgwd log.
0 if the message is NOT presnt in the mgwd log.

=item B<Failure:>

confesses, in case of error getting mgwd log content.

=back

=item Examples

=back

	#Code snippet using this function:
    if($node1->mgwd_log_check_msg(msg => $msg, day => 0) {
        # code
    }

=cut

sub mgwd_log_check_msg {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $msg = $args{msg};
    my $day    = $args{day} || 0;
    if( !defined $msg ){
        croak 'msg needs to be provided';
    }
    if( ($day !~ /^[-]?\d+$/) or ($day > 0 ) ){
        croak 'Provide 0 or -ve value for parameter day';
    }

    my $mgwd_log_filename;
    $day =~ s/-//;
    if($day == 0) {
        $mgwd_log_filename = 'mgwd.log';
    }
    else {
        $mgwd_log_filename = 'mgwd.log.previous.' . $day;
    }

    my $cmd = 'grep \'' . $msg . '\' /mroot/etc/log/mlog/' . $mgwd_log_filename;
    my $msg_matching_content = q{};
    my $eval_return = eval {
                         $msg_matching_content = $self->server->shell(cmd => $cmd);
                         1;
                      };
    if(!$eval_return){
        #means, $@ is set some value
        confess "Could not search mgwd.log content for the gievn msg: error: $@";
    }

    if( $msg_matching_content ne q{} ) {
        NOTE::logcomment('The given msg appear in mgwd.log. msg_matching_content: ' .$msg_matching_content);
        return 1;
    }
    else {
        return 0;
    }
}

=head2 dump

dump the whole volume or a specific snapshot to another given volume.
confesses if error in running dump.

The dump command must be run in diag mode to allow for it to back up
a c-mode volume.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<dump_to_volume> => I<string> 

name of volume dumped to (i.e destination volume)

=item B<volume_to_dump> => I<string> i

name of volume to be dumped (i.e. source volume)

=back

I<Optional Arguments>

=over 

=item B<snapshot_to_dump> => I<string> 

in case of dumping snapshot, name of snapshot that is to be dumped
Default: whole volume

=item B<dump_to_file> => I<string>

in case of need of dumping to specific file, name of the file to be dumped to
Default: Function picks up some random name

=item B<timeout> => I<timeout in seconds>

timeout for dump. default 1200 seconds

=item B<background> => I<1|0>

determines whether dump is to be run in background/foreground. 
I<Default>: 0 (i.e foreground)

=back

=item Returns

=over

=item B<Success:> nothing (i.e does not return anything)

=item B<Failure:> confesses, in case of dump failure/ error doing dump

=back

=item Notes

=over

=item 1 The dump is run as a subtest.

=back

=item Examples

=back

	#Code snippet using this function:
    $node->dump(dump_to_volume => $dst_vol_dump->name,
                volume_to_dump => $dst_vol->name,
                snapshot_to_dump => $snapshot_name,
                timeout          => 1500,
                background => 1,
              );

=cut

sub dump {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $dump_to_volume   = $args{dump_to_volume};
    my $volume_to_dump   = $args{volume_to_dump};
    my $snapshot_to_dump = $args{snapshot_to_dump};
    my $dump_to_file     = $args{dump_to_file} || NOTE::newName('dump');
    my $background       = $args{background} || 0; #dafault false
    my $timeout          = $args{timeout} || 1200; #dafault value of 20 min

    my ($ret, $eval_return, $dump_cmd, $dump_job, $dump_output);

    if(!exists $args{dump_to_volume}) {
        confess 'dump_to_volume needs to be provided';
    }

    if(!exists $args{volume_to_dump}) {
        confess 'volume_to_dump needs to be provided';
    }

    #get the names of volumes, if user pass the volume objects
    if( ref($dump_to_volume) eq 'NOTE::GX::Volume' ) {
        $dump_to_volume = $dump_to_volume->name;
    }
    if( ref($volume_to_dump) eq 'NOTE::GX::Volume' ) {
        $volume_to_dump = $volume_to_dump->name;
    }

    # the dump command is available in admin mode but requires diag mode
    # to dump a c-mode volume from a 7-mode cli, which is what we do in
    # GNOTE::GX
    $dump_cmd = 'priv set -q diag;';

    # build the dump command, order is important because some dump options
    # require values. dump requires that the values be listed in the same
    # order as their options. see na_dump for more details.
    $dump_cmd .= "dump 0f /vol/$dump_to_volume/$dump_to_file "
                 . "/vol/$volume_to_dump";
    if(defined $snapshot_to_dump) {
        $dump_cmd .= "/.snapshot/$snapshot_to_dump";
    }

    # running dump as a subtest, as cli dump cannot be run in background
    # using subtest, dump can be run either as foreground or background,
    # according to user choice
    $dump_job = 'dump_of_' . $dump_to_volume;
    if($snapshot_to_dump) {
        $dump_job .= '_snapshot_of_' . $snapshot_to_dump;
    }

    $ret = Tharn::subtest(
                sub {
                    $eval_return = eval {
                            $dump_output = $self->server->dblade(
                                                        cmd => "$dump_cmd",
                                                        '-Timeout' => $timeout);
                            1;
                        };
                    if (!$eval_return) {
                        #means, $@ is set some value
                        confess "Error executing dump command: $@";
                    }

                    if ($dump_output !~ /DUMP:\s+DUMP\s+IS\s+DONE/i) {
                        confess 'dump is NOT succesful. dump_output: ' .
                                $dump_output;
                    }
                },
                -runid => $dump_job,
                $background ? '-bg' : '-nobg',
                q{--},
                $timeout,
            );

    return;
}

=head2 is_dump_running

Returns 1 (true) if dump is running
Returns 0 (false) if dump is not running

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<dump_start_timestamp> => I<dump_start_timestamp>

dump start time stamp. based on this, this function looks 
for messages generated in /mroot/etc/log/backup file)
for the dump started after this time stamp. The timestamp 
can be the time stamp returned, by the NODE::datetime function.

=back

=item Returns

=over

=item B<Success:> 

Returns 1 (true) if dump is running
Returns 0 (false) if dump is not running

=item B<Failure:> 

confesses, if error getting the /mroot/etc/log/backup file content.

=back


=item Notes

=over

=item 1 

Based on the messages logged to /mroot/etc/log/backup file, this function 
determines, whether dump is still in progress or not.

=back

=item Examples

=back

	#Code snippet using this function:
    #get the timstamp first
    my $dump_start_timestamp = $node->datetime;
    #start the dump
    $node->dump(dump_to_volume => $dst_vol_dump->name,
                volume_to_dump => $dst_vol->name,
                snapshot_to_dump => $snapshot_name,
                timeout          => 1500,
                background => 1,
              );
    #Get the status of dump
    my $dump_running = $node->is_dump_running(dump_start_timestamp => $dump_start_timestamp );

=cut

sub is_dump_running {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess 'dump_start_timestamp is to be provided' if(!exists $args{dump_start_timestamp});

    my $start_timestamp = $args{dump_start_timestamp};

    my $backup_file_content;
    my $cmd = 'cat /mroot/etc/log/backup';
    my $eval_return = eval {
                         $backup_file_content = $self->server->shell(cmd => $cmd);
                         1;
                      };
    if(!$eval_return){
        #means, $@ is set some value
        confess "Could not get backup file content: error: $@";
    }

    #look for messages generated after the given start_timestamp
    my $dump_running = 1; #true
    my @backup_op_lines = split /\n/, $backup_file_content;
    foreach my $line (@backup_op_lines) {
        #get the timestamp of this message line
        #checking the line matches the message printed for a dump
        #Sample dump msg line is below:
        #dmp Sun Oct 17 20:23:14 EDT /vol/ssop1_src(0) Log_msg (Tape changes: 1)
        if($line =~ /^dmp\s+(\w+\s+\w+\s+\d+\s+\d\d:\d\d:\d\d\s+\w+)\s+/x){
            my $this_msg_ts = $1;  #this message time stamp
            #consider only if this message is generated after the given start_timestamp
            if(parsedate($this_msg_ts) >= parsedate($start_timestamp) ) {
                #If dump is finished, the following type of line appear in 'backup' file
                #dmp Sun Oct 17 20:23:14 EDT /vol/ssop1_src(0) Log_msg (Tape changes: 1)
                if($line =~ /Log_msg\s+[(]Tape\s+changes:\s+1[)]/i ){
                    #means dump is over
                    $dump_running = 0; #false
                    last;
                }
                elsif ($line =~ /Abort\s+[(]\d+\s+[MKG]?B[)]/ix){
                    #As the purpose is to know whether dump is running or not,
                    # we are not going to die, just printing a warning message
                    NOTE::logcomment('WARNING: dump aborted');
                    $dump_running = 0; #false
                    last;
                }
            }
        }
    }

    return $dump_running;
}

=head2 wait_for_dump_finish

Waits for dump to finish.
Returns if dump is finished within the timeout period
Dies if dump is not completed within the timeout period

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

=over

=item B<dump_start_timestamp> => I<dump_start_timestamp>

dump start time stamp. based on this, this function looks
for messages generated in /mroot/etc/log/backup file)
for the dump started after this time stamp. The timestamp
can be the time stamp returned, by the NODE::datetime function.

=back

I<Optional Arguments>

=over 

=item B<timeout> => I<timeout in seconds>

timeout in seconds, within which dump is checked for its completion.
I<Default>: 1200

=back

=item Returns

=over

=item B<Success:> does not return anything

=item B<Failure:> confesses, if dump is not finished within the timeout period.

=back

=item Examples

=back

	#Code snippet using this function:
    $node->wait_for_dump_finish(dump_start_timestamp => $dump_start_timestamp,
                                                   timeout => 1500,
                                                 );

=cut

sub wait_for_dump_finish {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess 'dump_start_timestamp is to be provided' if(!exists $args{dump_start_timestamp});

    my $start_timestamp = $args{dump_start_timestamp};
    my $timeout         = $args{timeout} || 1200; #not sure 1200 is ok or not

    my $interval = 5; # 5 seconds
    my $wait_time = 0;
    my $dump_running;
    while(  ($dump_running = $self->is_dump_running(%args)) and ($wait_time <= $timeout) ) {
        sleep $interval;
        $wait_time += $interval;
    }

    if($dump_running){
        confess "ERROR: dump is not finished within timeout $timeout seconds";
    }
    return;
}

=head2 release

Returns the software release running on the filer, as a string.
Croaks if error getting the information.

=over

=item Arguments

None

=item Returns

=over

=item B<Success:> software release running on the filer, as a string

=item B<Failure:> confesses, if error getting the information.

=back

=item Examples

=back

    #getting the release of ontap software running on node
	$release = $node->release;
    if ($release =~ /RollingRock/i){
        NOTE::logcomment("Cannot do snapmirror restore in RollingRock");
    }

=cut

sub release {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($ret,$out);
    my %opts;

    $opts{'node'} = $self->name;
    $opts{'iscurrent'} = 'true';

    $ret = eval {
        $out = $self->server->system_node_image_show(%opts);
        1;
    };
    confess "Could not retrieve the version nformation: $@"
        if(!$ret);

    return $out->[0]->{'version'};
}

=head2 ::lifs_up

Method takes care to bring up the lifs

=over

=item Arguments

B<role> = type of lifs to bring up cluster-mgmt,inter-cluster,cluster

=item Examples

=back

    $cluster->lifs_up(role => 'intercluster');

=cut

sub lifs_up {

    my ( $self, %args ) = @_;

    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @down_lifs = $self->lifs(%args , link => 'down');


    foreach (@down_lifs){
                            $self->lif_modify(name => $_ , link => 'up');
                        }

    return;
}

=head2 ::prom_get

Return list of PROM variables as hash. Returns listed variables only if "arguements" are provided.

=over

=item Examples

      1 use strict;
      2 use GNOTE;
      3 use Data::Dumper;
      4 eval {
      5     my $ut = NOTE::SnapMirrorUtil->new();
      6     my $node = NOTE::GX::Node->connect( name => $ut->param('FILER') );
      7     my %prom = $node->prom_get;
      8     print "Dump everything:\n". Dumper(\%prom);
      9 
     10     %prom = $node->prom_get('last-OS-booted-raid-ver'=>1, 'last-OS-booted-wafl-ver'=>1);
     11     print "Check last-OS-booted-raid-ver and last-OS-booted-wafl-ver:".Dumper(\%prom);
     13 };
     14 logresult('FAIL',$@) if $@;
     15 NOTE::teardown;

=back

=cut

sub prom_get {
    # using hash instead of list make prom_get a bit ugly to call with params, but simplifies
    # overall flow when common %args hash is used in scripts or other methods where %args are
    # being passed as is. See prom_set() implementation for an example
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();
    my $raw = $self->server->system_node_run( node=>$self->name, command=>"\"priv set -q diag; bootargs dump\"" );
    my %out = ( $raw =~ /(\S+)\s+==\s+(\S+)/gs ); # parse output into hash
    map{ delete $out{$_} unless exists $args{$_} } keys %out if keys %args; # if $args, then only return requested args
    return %out;
}

=head2 ::prom_set

Sets PROM variable(s), optionally reboots node and returns requested variables as a hash. All arguments except
B<node-reboot> are treated as PROM variables to set.

=over

=item Arguments

B<node-reboot> I<optional> Reboot node after setting or changing of variable's value.

=item Examples

      1 use strict;
      2 use GNOTE;
      1 use strict;
      2 use GNOTE;
      3 eval {
      4     my $ut = NOTE::SnapMirrorUtil->new();
      5     my $node = NOTE::GX::Node->connect( name => $ut->param('FILER') );
      6     $node->prom_set('snapmirror.sn_enabled'=>'true', 'node-reboot'=>1);
      7 };
      8 logresult('FAIL',$@) if $@;
      9 NOTE::teardown;

=back

=cut

sub prom_set {
    my ( $self, %args ) = @_;
    my $reboot = delete $args{'node-reboot'};
    $self->server->system_node_run( node=>$self->name, command=>"\"priv set -q diag; bootargs set $_ $args{$_}\"" ) foreach( keys %args );
    $self->reboot if $reboot;
    my %chk = $self->prom_get( %args );
    my $err = grep { defined $_ } map{ $args{$_} eq $chk{$_} ? undef : 1 } keys %args;	# get mismatch count
    confess "Unable to set PROM $err parameter(s). Requested:".Dumper(\%args)."\nReceived:".Dumper(\%chk) if $err;
    return %chk;
}

1;
