# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Qtree.pm#1 $
#
## @summary NOTE GX Qtree library
##
## @description This library will provide qtree support for Paloma testing.
## @dependencies
##
## @author corina@netapp.com, dl-dpgqa-automation@netapp.com
##
## @pod here

=head1 NAME

NOTE::GX::Qtree module

=head1 DESCRIPTION

This module provides 10-mode specific methods for the Qtree object

=cut

package NOTE::GX::Qtree;

use strict;
use warnings;
no warnings 'redefine';

use GNOTE;
use Tharn;
use RichData;
use GNOTE::GX::Quota;
use Data::Dumper;
use Carp qw(croak confess);
our @ISA = qw(NOTE::GX::Quota NOTE::GX::San::Lun NOTE::Qtree NOTE::GX);
use Params::Validate qw(validate ARRAYREF HASHREF OBJECT SCALAR UNDEF);

=head2 ::new

Creates a Qtree object and returns it.

=over

=item Arguments

B<Volume> => I<object> 
B<name> => I<string>

=item Examples

=back

     $qt = NOTE::Qtree->new(
         Volume => $vol,
         tag    => 'destination',
         name    => 'qtree1'
     );
     
=cut

sub new {

    # Args: Volume,name,tag
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {%args};

    $self->{_owner} = $$;

    if ( ref( $args{Volume} ) =~ /NOTE::.*Volume/ ) {
        $self->{_Volume} = $args{Volume};
    }
    else {
        if ( !defined( $args{Volume} ) || ( ref( $args{Volume} ) eq "HASH" ) ) {
            if ( $args{Aggregate} ) {
                $self->{_Volume} = NOTE::GX::Volume->create(
                    Aggregate => $args{Aggregate},
                    %{ $args{Volume} }
                );
            }
            else {
                $self->{_Volume} = NOTE::GX::Volume->create( %{ $args{Volume} } );
            }
        }
    }
    $self->{_Filer}   = $self->{_Volume}->Filer;
    $self->{_Cluster} = $self->{_Volume}->Cluster;
    $self->{_Node}    = $self->{_Volume}->Node;

    delete $args{Aggregate};
    delete $args{Volume};

    bless( $self, $class );
    NOTE::reference($self);
    $self->is_destroyed = 1;

    $self->SUPER::new(%args);

    return $self;
}

=head2 ::create

Creates a Qtree and returns
an object representing the created Qtree.

=over

=item Arguments

B<Volume> => I<object> 
B<name> => I<string>

=item Examples

=back

     $qt = NOTE::Qtree->create(
         Volume => $vol,
         tag    => 'destination',
         name    => 'qtree1'
     );
     
=cut

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = __PACKAGE__->new(%args);
    $self->SUPER::create(%args);
    $self->is_destroyed = 0;

    return $self;
}

sub _qtree_create {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %qtree_create_hash = ();

    %qtree_create_hash = (
        vserver            => $self->{_Volume}->{_Vserver}->name,
        volume             => $self->{_Volume}->name,
        qtree              => $self->{_name},
        'security-style'   => $args{'security-style'} || 'mixed',
        'oplock-mode'      => $args{'oplock-mode'} || 'enable',
        'unix-permissions' => $args{'unix-permissions'} || '---rwxrwxrwx'
    );
    eval { $self->server->volume_qtree_create(%qtree_create_hash); };
    croak $@ if ($@);

    return $self;
}

=head2 ::attach

Creates an object representing an existing Qtree.

=over

=item Arguments

B<name> => I<string>

I<Required.>
The name of an existing Qtree.

B<Volume> => I<object> | I<spec>

I<Required.>
The volume in which to find the named Qtree.
Specified as a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/> object
or a volume spec.

B<Filer> => I<host spec>

I<Dependent.>
Required only when the B<Volume> argument is a volume spec.
Specifies the filer where the volume exists.

B<mntpoint> => I<string>

I<Optional.>
The full path name on which to mount the Qtree.

B<exportfs> => I<export options>

I<Optional.>
Export options as described in L<C<::exportfs>|/::exportfs>.

B<dataset> => I<dataset specifier>

I<Optional.>
A description of some data to put in the Qtree.
The dataset will be fully created
before the call to C<::create> returns.
See L<C<NOTE::Volume::fill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::fill>
for a description of dataset specifiers.

B<force> => C<1>

I<Optional.>
A flag to force the creation of a Qtree object
even if the named Qtree does not exist.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip for
L<C<::populate>|/::populate> and L<C<::modify>|/::modify>.
The default is 0.

=item Examples


=back

     $qt = NOTE::Qtree->attach(
         Volume => $vol,
         name   => 'qtree42'
     );

=cut

sub attach {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self  = {};
    my $force = $args{force};

    $self->{_owner} = $$;

    bless $self, $class;
    NOTE::reference($self);

    $self->SUPER::attach(%args);

    eval {
        unless ($force)
        {
            NOTE::logcomment( "Checking for Qtree " . $self->{_name} );
            $self->_exists( $self->{_name} );
        }
    };
    croak $@ if ( $@ && !$force );

    return $self;
}

=head2 ::delete

Deletes a qtree.
Croaks on failure. 

=over

=item Arguments

B<Volume> => I<object> 
B<name> => I<string>

=item Examples

=back

     qtree->delete();
     
=cut

sub delete {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Delete Qtree " . $self->description );
    $self->umount;
    eval { $self->server->volume_qtree_delete( vserver => $self->{_Volume}->{_Vserver}->name, volume => $self->{_Volume}->name, qtree => $self->name, foreground => 'true' ); };
    if ($@) {
        croak "\n volume qtree delete FAILED: $@";
    }
}

=head2 ::purge

Deletes the qtree after removing all it's contents.

=over

=item Arguments

I<None.>

=item Example

=back

    $qtree->purge;

=cut

sub purge {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Purge Qtree " . $self->description );
    $self->mount;
    my $cmd = "rm -rf " . $self->mountpoint() . "/*";
    system($cmd);
    system( "ls -l " . $self->mountpoint() );
    $self->delete;
}

=head2 ::rename

This function rename the Qtree to the newname specified by the user.

=over

=item Arguments

B<newname> => I<string>

I<Optional>

New name for the Qtree.

=item Return Values

b<None>

=item Examples

    $qt->rename ('newqt');

=back

=cut

sub rename {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %args = validate( @_, { newname => { type => SCALAR, optional => 1, default => 'newqt' }, } );

    my $newname = delete $args{newname};

    my @list;

    eval { @list = $self->server->volume_qtree_rename( vserver => $self->Vserver->name, volume => $self->Volume->name, qtree => $self->name, newname => $newname ); };
    if ($@) {
        confess("\n volume qtree rename FAILED: $@ ");
    }

    $self->{_name} = $newname;
}

## Checks to see if a qtree with the same name already exists or not
## Croaks on failure.

## $self->_exists($self->{_name});

sub _exists {
    my $self = shift;
    my ($qtree_name) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    eval { $self->server->volume_qtree_show( volume => $self->Volume->name, qtree => $qtree_name ); };
    if ($@) {
        croak "\n\n Qtree does not exist: $@ \n";
    }
}

=head2 ::status

Returns the qtree status as a hash ref or a specified column. 

=over

=item Arguments

B<element> = I<string>

I<Optional>

Column name from qtree show. If not specified a HASHREF of all columns will be returned
 
=item Examples

=back

    $qtree->status;
    $qtree->status("qtree-id");
=cut

sub status {
    my ( $self, $element ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $status = "";
    eval { $status = $self->server->volume_qtree_show( volume => $self->Volume->name, qtree => $self->name, fields => 'status' ); };
    croak $@ if $@;
    if ( defined $element ) {
        return $status->[0]{$element};
    }
    else {
        return $status->[0];
    }
}

=head2 ::writeable

Returns a boolean indicating whether the qtree is writeable.

=over

=item Arguments

None.

=item Examples

=back

    if ( $qtree->writeable ) {
        ...
    }

=cut

sub writeable {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $s;
    eval { $s = $self->status("status"); };
    if ($@) {

        # When the containing volume is offlined, qtree status gets a 'no entries' error
        # which for the purposes of this method will be treated as not writable
        if ( $@ =~ /no entries/ && $self->Volume->offline ) {
            return 0;
        }
        else {
            confess "Status check for " . $self->description . "failed. Error: $@";
        }
    }
    return ( $s =~ /normal/ );
}

=head2 ::modify_attr

Croaks on failure. 

=over

=item Arguments

my %modattr_options = ( 'security-style' => 'ntfs',
                        'oplock-mode'    => 'disable',
                        'unix-permissions' => '444');

=item Examples

$qtree->modify_attr( %modattr_options );

=back

     if ( $qtree->modify_attr() ) {
        ...
     }
     
  
=cut

sub modify_attr {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $path = "";

    $args{'vserver'} = $self->{_Volume}->{_Vserver}->name;
    $args{'volume'}  = $self->{_Volume}->name;
    $args{'qtree'}   = $self->name;

    eval { $self->server->volume_qtree_modify(%args); };
    if ($@) {
        croak "\n Qtree attributes could not be modified";
    }
}

=head2 ::show

Deprecated routine. Please call method 'status' instead.

=over

=item Examples

=back

    my $hash_ref = $qtree->show;
    
=cut

sub show {
    my ( $self, $element ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment("WARNING: Deprecated method Qtree::show, please call Qtree::status instead");
    return $self->status($element);
}

=head2 ::is_acro

Return 1 if type of the volume of this qtree is Acro. Otherwise, return 0.

=over

=item Arguments

B<None>

=back

=item Example

    if ($qtree->is_acro) {
        #
	# the $vol instance is a of Acro stripe type
	#
    }

=cut

sub is_acro {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->Volume->is_acro();
}

=head2 ::is_coral

Return 1 if type of the volume of this qtree is Coral. Otherwise, return 0.

=over

=item Arguments

B<None>

=back

=item Example

    if ($qtree->is_coral) {
        #
	# the $vol instance is a of Coral stripe type
	#
    }

=cut

sub is_coral {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->Volume->is_coral();
}

=head2 ::mount

Mounts qtree on client or localhost (default) and returns the mount point. 

=over

=item Examples

=back

    $qtree->mount;
    
=cut

sub mount {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $data_ip;
    my %nfs_args = ();
    $nfs_args{chmod} = $args{chmod} if exists $args{chmod};
    $nfs_args{client} = delete $args{client} || "localhost";

    $self->{_mntpoint} = {} unless $self->{_mntpoint};
    unless ( $self->{_mntpoint}->{ $nfs_args{client} } ) {
        # First try with RR.x specific command
        my $eval_ret = eval {
                             $data_ip = $self->{_Volume}->{_Vserver}->lif_address('data-protocol' => 'nfs');
                             1;
                            };
        # backward compatibility with BR.x builds
        if ( !$eval_ret ){
            $data_ip = $self->{_Volume}->{_Vserver}->lif_address();
        }
        # For RW mount add '/.admin'
        my $path = ( $args{RW} ? "/.admin" : "" ) . $self->{_Volume}->mount_path() . '/' . $self->name;
        NOTE::logcomment( "Mounting " . $self->description );
        $self->{_mntpoint}->{ $nfs_args{client} } = GenericMount::mount( $data_ip, $path, %nfs_args );
        croak $@ if $@;
    }
    return $self->{_mntpoint}->{ $nfs_args{client} };
}

=head2 ::umount

Unmounts qtree on client or localhost (default).
Croaks on failure. 

Refer to L<C<Volume::mountpoint>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3umount>.

=over

=item Examples

=back

    $qtree->umount ();
    $qtree->umount (client=>'dpg-vml6');
    
=cut

sub umount {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::GX::Volume::umount( $self, %args );
}

=head2 ::mount_path

Returns the full path name of the Qtree,
i.e. C</vol/><I<containing volume name>>C</>I<<Qtree nameE<gt>>

=over

=item Arguments

None.

=item Example

=back

     $mntpt = $qt->mount_path;

=cut

sub mount_path {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->{_Volume}->mount_path() . '/' . $self->name;
}

=head2 ::create_share

Refer to L<C<Volume::create_share>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3acreate_share>.

=over

=item Arguments

None.

=item Example

=back

     $qt->create_share

=cut

sub create_share {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Construct the path for this qtree's CIFS share by concatenating
    # the mount path of the containing volume and the this qtree's
    # name.
    $args{path} = ( $self->{_Volume} )->mount_path . "/" . $self->{_name};
    return NOTE::GX::Volume::create_share( $self, %args );
}

=head2 ::delete_share

Refer to L<C<Volume::create_share>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3adelete_share>.

=over

=item Arguments

None.

=item Example

=back

     $qt->delete_share

=cut

sub delete_share {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::GX::Volume::delete_share( $self, %args );
}

=head2 ::map_share

Refer to L<C<Volume::create_share>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3amap_share>.

=over

=item Arguments

None.

=item Example

=back

     $qt->map_share

=cut

sub map_share {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::GX::Volume::map_share( $self, %args );
}

=head2 ::unmap_share

Refer to L<C<Volume::create_share>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3aunmap_share>.

=over

=item Arguments

None.

=item Example

=back

     $qt->unmap_share

=cut

sub unmap_share {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::GX::Volume::unmap_share( $self, %args );
}

=head2 ::Cluster

Return the qtree's cluster object 

=over

=item Arguments

None.

=item Example

=back

     $qt->Cluster

=cut

sub Cluster {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Volume}->Cluster
}

=head2 ::Node

Return the qtree's node object

=over

=item Arguments

None.

=item Example

=back

     $qt->Node

=cut

sub Node {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Volume}->Node
}

=head2 ::server

Return the qtree's server object

=over

=item Arguments

None.

=item Example

=back

     $qt->server

=cut

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Volume}->server
}

=head2 ::Vserver

Returns the vserver object.

=over

=item Examples

=back

	$qtree->vserver;

=cut

sub Vserver {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Volume}->Vserver
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    $self->sio( stop => 1 ) if exists $self->{_sio};    # Terminate any on-going SIO process
    return if ( $self->{_destroyed} );
    return if ( !$self->{_anonymous} );
    $self->{_destroyed} = 1;
    return if ( NOTE->ALL_SPARED );

    $self->_san_cleanup();				# blind clean up SAN if present. This will never fail.

    foreach ( keys %{ $self->{_mntpoint} } ) {
        $self->umount( client => $_ );
    }
    if ( $self->{_spare_me} ) {
        NOTE::logcomment( "Qtree " . $self->description . " has been spared" );
        return;
    }
    eval { $self->purge(); };
    croak "$@"
      if $@ and $@ !~ /no such file or directory/i;    # ignore if qtree does not exists as it may have already been deleted
}

# not implemented in BR
sub exportfs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

# not implemented in BR
sub unexportfs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

# not supported in GX

sub snaplock {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return 0;
}

=head2 ::disk_free

Returns the disk free in the volume.

=over

=item Examples

=back

        $qtree->disk_free;

=cut

sub disk_free {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Returns " . $self->description . " Volume's disk_free information" );
    return NOTE::GX::Volume::disk_free( $self->Volume, %args );
}

1;
