#
## $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Quota.pm#1 $
##
## @summary  GNOTE GX Quota module
##
## @author   dl-dpgqa-automation@netapp.com
##
## @status   Public
##
## @pod      here

=head1 NAME

NOTE::GX::Quota

=head1 DESCRIPTION

The purpose of this module is to provide quota support in GNOTE. There is no
quota class so one cannot create a quota object. It is its own module so it
can be inherited by any class needing quota support. Initially the Volume and
Qtree classes have quota support.

There are four things that must be done to add quota support to a class.
1)  Include the Quota module in the class.
    use GNOTE::GX::Quota;
2)  Allow the quota package to be part of the class by including it in the
    class ISA statement.
    our @ISA = qw(NOTE::GX::Quota NOTE::GX::MyClass ...);
3)  Create a method in the class called _get_volume_name() that returns the
    volume associated with the object. See GNOTE/Qtree.pm for an example
4)  Create a method in the class called _get_qtree_name() that returns the
    string "" (two double quotes). This is the string the filer expects when
    a rule is not being created against a qtree. See GNOTE/Volume.pm for an
    example.

=head1 METHODS

=cut

package NOTE::GX::Quota;

use strict;
use warnings;
no warnings 'redefine';

use Carp qw(croak confess);
use GNOTE;

use Data::Dumper;


=head2 quota_on

Enables quotas on a volume.

=over

=item Arguments

Arguments are passed by hash.

Hash table of arguments:

I<Optional Arguments>

=over

=item B<foreground|w> => I<true|false> Execute the command in the foreground
    (B<w>ait) or background. for the command to complete. Default = true.

=back

=item Returns

=over

=item B<Success:> 0

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1 This method will always provide the vserver and volume options to the
    command.

=back

=item Examples

=back

    # enable quotas on a volume, run in the foreground by default
    $vol->quota_on();

    # enable quotas on a volume in the background
    $vol->quota_on( foreground => 'false' );
    # wait for quotas to be turned on, note this example will not timeout
    do {
        Tharn::snooze(5);
        $status = $vol->quota_show( state => 'on');
    } until (defined($status));

    # enable quotas on a qtree's volume in the foreground
    $qtree->quota_on (foreground => 'true' );

=cut

sub quota_on {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $results;

    $results = eval {
            $self->server->volume_quota_on( vserver => $self->Vserver->name,
                                            volume => $self->_get_volume_name,
                                            %args );
            1;
        };

    if (! $results) {
        confess "volume quota on failed: $@";
    }

    return 0;
}


=head2 quota_policy_create

Creates a quota policy for a vserver

=over

=item Arguments

Arguments are passed by hash table

Hash table of arguments:

=over

=item B<policy-name> => I<string, 1-32 characters long> The name of the policy

=back

=item Returns:

=over

=item B<Success:> 0

=item B<Failure:> confesses

=back

=item Confessions:

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1 This method will always provide the vserver option to the command.

=back

=item Examples

=back

    # create a quota policy from a volume object
    $vol->quota_policy_create( 'policy-name' => 'policy_vs0' );

    # create a quota policy from a qtree object
    $qtree->quota_policy_create( 'policy-name' => 'policy_vs0' );

=cut

sub quota_policy_create {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $results;

    $results = eval {
            $self->server->volume_quota_policy_create(
                                                vserver => $self->Vserver->name,
                                                %args);
            1;
        };

    if (! $results) {
        confess "volume quota policy create failed: $@";
    }

    ## to maintain a hash of policies created for cleanup purpose
    if(! defined ($self->{'_quota_policy_cleanup'})) {
        $self->{'_quota_policy_cleanup'} = {};
    }

    $self->{'_quota_policy_cleanup'}{$args{'policy-name'}} = 1;
    return 0;
}


=head2 quota_policy_delete

Deletes a quota policy for a vserver

=over

=item Arguments

Arguments are passed by hash.

Hash table of arguments:

=over

=item B<policy-name> => I<string, 1-32 characters long> The name of the policy

=back

=item Returns

=over

=item B<Success:> 0

=item B<Failure:> confesses

=back

=item Confessions:

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1  This method will always provide the vserver option to the command.

=back

=item Examples

=back

    # delete a quota policy from a volume object
    $vol->quota_policy_delete( 'policy-name' => 'policy_vs0' );

=cut

sub quota_policy_delete {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $results;

    $results = eval {
            $self->server->volume_quota_policy_delete(
                                                vserver => $self->Vserver->name,
                                                %args);
            1;
        };

    if (! $results) {
        confess "volume quota policy delete failed: $@";
    }

    #deleting the policy entry from hash once it is deleted
    delete $self->{'_quota_policy_cleanup'}->{$args{'policy-name'}};

    return 0;
}


=head2 quota_policy_rule_create

Creates a rule for an existing quota policy.

=over

=item Arguments

Arguments are passed by hash table

Hash table of arguments:

=over

=item B<policy-name> => I<string, 1-32 characters long> The name of the policy

=item B<type> => I<tree|user|group> The resource type acted upon by the rule

=item B<target> => I<string> Tree|user|group name based on type

=back

I<Optional Arguments>

=over

=item B<user-mapping> => I<on|off> Enable/disable nfs/cifs user mapping

=item B<disk-limit> => I<size|-> Set a hard disk usage limit

=item B<file-limit> => I<integer|-> Set a hard file count limit

=item B<threshold> => I<size|-> Threshold for disk limit

=item B<soft-disk-limit> => I<size|-> Set a soft disk limit

=item B<soft-file-limit> => I<integer|-> Set a soft file limit

=back

=item Returns

=over

=item B<Success:> 0

=item B<Failure:> confesses

=back

=item Confessions:

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1  This method will always provide the vserver and volume options to the
    command.

=item 2  The rule create command does not allow the qtree option with
    type == tree. Therefore the qtree option is set as follows.

=over

=item * if this method is called via a volume object, qtree = ""

=item * if this method is called via a qtree object and type == tree, qtree = ""

=item * if this method is called via a qtree object and type != tree,
      qtree = <qtree name>

=back

=item 3 By default do not have Node.pm verify the command results. Node does not
    convert disk size values correctly, it expects the filer so display a disk
    size of 500m as 500m not 524288000, which affects which affects disk-limit,
    threshold and soft-disk-limit. The caller can override it if they want.

=back

=item Examples

=back

    # create a qtree rule governing a user, setting the soft disk limit and
    # threshold, and enabling user mapping
    $qtree->quota_policy_rule_create( 'policy-name' => 'superhero_limits',
                                      type => 'user',
                                      target => 'superman',
                                      'user-mapping' => 'on',
                                      'soft-disk-limit' => '100g',
                                      threshold => '95g' );

    # create a qtree rule governing a user, setting the disk limit, soft disk
    # limit and threshold
    $qtree->quota_policy_rule_create( 'policy-name' => 'superhero_limits',
                                      type => 'user',
                                      target => 'wonderwoman',
                                      'disk-limit' => '120g',
                                      'soft-disk-limit' => '100g',
                                      threshold => '95g' );

    # create a volume rule governing a group, setting the disk limit, soft
    # disk limit and threshold
    $vol->quota_policy_rule_create( 'policy-name' => 'superhero_limits',
                                    type => 'group',
                                    target => 'superheroes',
                                    'disk-limit' => '200g',
                                    'soft-disk-limit' => '175g',
                                    threshold => '150g' );

    # create a volume rule governing a tree, setting soft file limits and
    # file limits
    $vol->quota_policy_rule_create( 'policy-name' => $PolicyName,
                                    type => 'tree',
                                    target => $QtreeName,
                                    'soft-file-limits' => 1024,
                                    'file-limits' => $MaxFiles );

    # create a qtree rule governing a tree, setting disk limits, file limits
    # and threshold
    $qtree->quota_policy_rule_create( 'policy-name' => $PolicyName,
                                      type => 'tree',
                                      target => $qtree->name,
                                      'disk-limit' => $MaxStorage,
                                      'file-limit' => $MaxFiles,
                                      threshold => $threshold );

=cut

sub quota_policy_rule_create {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $results;

    # a rule of type = tree does not allow the qtree option to be set.
    # therefore check the type and set the qtree value appropriately.
    if ($args{type} eq 'tree') {
        $args{qtree} = q{""};
    } else {
        $args{qtree} = $self->_get_qtree_name();
    }

    # disable Node.pm verification be default. Node does not convert disk size
    # values correctly, it expects the filer so display a disk size of 500m as
    # 500m not 524288000, which affects which affects disk-limit, threshold
    # and soft-disk-limit. allow the caller to override it if they want.
    $args{'-Verification'} = $args{'-Verification'} || 0;

    $results = eval {
            $self->server->volume_quota_policy_rule_create(
                                            vserver => $self->Vserver->name,
                                            volume => $self->_get_volume_name,
                                            %args );
            1;
        };

    if (! $results) {
        confess "volume quota policy rule create failed: $@";
    }

    return 0;
}


=head2 quota_policy_rule_modify

Modifies an existing quota policy rule.

=over

=item Arguments

Arguments are passed by hash table

Hash table of arguments:

=over

=item B<policy-name> => I<string, 1-32 characters long> The name of the policy

=item B<type> => I<tree|user|group> The resource type acted upon by the rule

=item B<target> => I<string> Tree|user|group name based on type

=back

I<Optional Arguments>

=over

=item B<user-mapping> => I<on|off> Enable/disable nfs/cifs user mapping

=item B<disk-limit> => I<size|-> Set a hard disk usage limit

=item B<file-limit> => I<integer|-> Set a hard file count limit

=item B<threshold> => I<size|-> Threshold for disk limit

=item B<soft-disk-limit> => I<size|-> Set a soft disk limit

=item B<soft-file-limit> => I<integer|-> Set a soft file limit

=back

=item Returns

=over

=item B<Success:> 0

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1 This method will always provide the vserver and volume options to the
    command.

=item 2 The rule modify command does not allow the qtree option with
    type == tree. Therefore the qtree option is set as follows.

=over

=item * if this method is called via a volume object, qtree = ""

=item * if this method is called via a qtree object and type == tree, qtree = ""

=item * if this method is called via a qtree object and type != tree,
      qtree = <qtree name>

=back

=item 3 The following rule attributes form the name of the rule and cannot be
    modified.

=over

=item * vserver

=item * volume

=item * policy-name

=item * type 

=item * target

=back

=item 4 The following rule attributes define the rule and can be modified.

=over

=item * user-mapping (only valid for type == user)

=item * disk-limit

=item * file-limit

=item * threshold

=item * soft-disk-limit

=item * soft-file-limit

=back

=item 5 By default do not have Node.pm verify the command results. Node does not
    convert disk size values correctly, it expects the filer so display a disk
    size of 500m as 500m not 524288000, which affects which affects disk-limit,
    threshold and soft-disk-limit. The caller can override it if they want.

=back

=item Examples

=back

    # modify a qtree rule governing a user, changing the disk limit and
    # threshold, and enabling user mapping
    $qtree->quota_policy_rule_modify( 'policy-name' => 'superhero_limits',
                                      type => 'user',
                                      target => 'batman',
                                      'user-mapping' => 'on',
                                      'disk-limit' => '120g',
                                      threshold => '100g' );

    # modify a qtree rule governing a user, changing the disk limit, soft
    # disk limit and threshold
    $qtree->quota_policy_rule_modify( 'policy-name' => 'superhero_limits',
                                      type => 'user',
                                      target => 'robin',
                                      'disk-limit' => '120g',
                                      'soft-disk-limit' => '100g',
                                      threshold => '100g' );

    # modify a volume rule governing a group, changing the disk limit and
    # soft disk limit
    $vol->quota_policy_rule_modify( 'policy-name' => 'superhero_limits',
                                    type => 'group',
                                    target => 'superheroes',
                                    'disk-limit' => '400g',
                                    'soft-disk-limit' => '375g' );

    # modify a volume rule governing a tree, changing the file limits
    $vol->quota_policy_rule_modify( 'policy-name' => $PolicyName,
                                    type => 'tree',
                                    target => $QtreeName,
                                    'file-limits' => $NewMaxFiles );

    # modify a qtree rule governing a tree, changing the disk limits, file
    # limits and threshold
    $qtree->quota_policy_rule_modify( 'policy-name' => $PolicyName,
                                      type => 'tree',
                                      target => $qtree->name,
                                      'disk-limit' => $NewMaxStorage,
                                      'file-limit' => $NewMaxFiles,
                                      threshold => $NewThreshold );

=cut

sub quota_policy_rule_modify {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $results;

    # a rule of type = tree does not allow the qtree option to be set.
    # therefore check the type and set the qtree value appropriately.
    if ($args{type} eq 'tree') {
        $args{qtree} = q{""};
    } else {
        $args{qtree} = $self->_get_qtree_name();
    }

    # disable Node.pm verification be default. Node does not convert disk size
    # values correctly, it expects the filer so display a disk size of 500m as
    # 500m not 524288000, which affects which affects disk-limit, threshold
    # and soft-disk-limit. allow the caller to override it if they want.
    $args{'-Verification'} = $args{'-Verification'} || 0;

    $results = eval {
            $self->server->volume_quota_policy_rule_modify(
                                            vserver => $self->Vserver->name,
                                            volume => $self->_get_volume_name,
                                            qtree => $self->_get_qtree_name,
                                            %args );
            1;
        };

    if (! $results) {
        confess "volume quota policy rule modify failed: $@";
    }

    return 0;
}


=head2 quota_policy_rule_show

Show the quota policy rules for the volume/qtree.

=over

=item Arguments

Arguments are passed by hash table

Hash table of arguments:

I<Optional Arguments>

The default for all options is to not provide the option.

=over

=item B<fields> => I<fieldname, ...> The list of fields to be displayed

=item B<policy-name> => I<string, 1-32 characters long> The name of the policy

=item B<type> => I<tree|user|group> The resource type acted upon by the rule

=item B<target> => I<string> Tree|user|group name based on type

=item B<qtree> => I<qtree name> The name of the qtree

=item B<user-mapping> => I<on|off> Enable/disable nfs/cifs user mapping

=item B<disk-limit> => I<size|-> Set a hard disk usage limit

=item B<file-limit> => I<integer|-> Set a hard file count limit

=item B<threshold> => I<size|-> Threshold for disk limit

=item B<soft-disk-limit> => I<size|-> Set a soft disk limit

=item B<soft-file-limit> => I<integer|-> Set a soft file limit

=back

=item Returns

=over

=item B<Success:> an array of hash references, or undef.
         the returned hash is defined in the notes section below.

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1 This method will always provide the vserver, volume and qtree options to
    the command.

=item 2 This command does not support the following options:
    -instance

=item 3 The output hash definition. All undefined attributes are set to '-'
    except where specifically stated otherwise.

=over

=item B<$href->{vserver}> => vserver name to which policy is associated

=item B<$href->{volume}> => volume name to which rule is applied

=item B<$href->{'policy-name'}> => the policy name the rule is associated
                                  with, string

=item B<$href->{type}> => the rule type, user|group|tree

=item B<$href->{target}> => the target value, string

=item B<$href->{qtree}> => qtree name, undefined value = ""

=item B<$href->{'user-mapping'}> => user mapping, on|off

=item B<$href->{'disk-limit'}> => disk limit, disk size (ex. 200m, 1g,..)

=item B<$href->{'file-limit'}> => file limit, integer

=item B<$href->{threshold}> => disk limit threshold, disk size (ex. 200m, 1g,..)

=item B<$href->{'soft-disk-limit'}> => the soft limit for disk storage,
                                      disk size (ex. 200m, 1g,..)

=item B<$href->{'soft-file-limit'}> => the soft limt for number of files,
                                      integer

=back

=back

=item Examples

=back

    # show all the rules associated with this volume and vserver
    $vol->quota_policy_rule_show();

    # show all the rules associated with a specific volume and policy
    $vol->quota_policy_rule_show( 'policy-name' => 'superhero_limits' );

    # show all the rules associated with a specific qtree and policy
    $qtree->quota_policy_rule_show( 'policy-name' => $MyPolicy,
                                     qtree => $qtree->name );

    # show all the rules associated with a specific qtree on a volume using
    # a volume object
    $vol->quote_policy_rule_show( qtree => $vol->Qtrees->[0]->name );

    # show all the rules associated with a specific qtree on a volume using
    # a qtree object
    $qtree->quota_policy_rule_show( qtree => $qtree->name );

    # show the rules of a specific type for a policy using a qtree object
    $qtree->quota_policy_rule_show( 'policy-name' => $MyPolicy, type => $type );

    # show all the rules for a specific group on this volume using a
    # volume object
    $vol->quota_policy_rule_show( type => 'group', target => $group );

    # show all the rules for a specific user on this volume using a qtree
    # object
    $qtree->quota_policy_rule_show( type => 'user', target => $user );

=cut

sub quota_policy_rule_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($results, $out);

    # the filer will find bad arguments and if command options change
    # we won't have to.
    $results = eval {
                    $out = $self->server->volume_quota_policy_rule_show(
                                            vserver => $self->Vserver->name,
                                            volume => $self->_get_volume_name,
                                            %args );
                    1;
                };

    if (! $results) {
        # turn any expected errors into returning undef (no data)
        if ($@ =~ /There are no entries matching your query/i) {
            $out = undef;
        } else {
            confess "quota policy rule show failed: $@";
        }
    }

    return $out;
}


=head2 quota_policy_show

Show the quota policies for the vserver.

=over

=item Arguments

Arguments are passed by hash table

Hash table of arguments:

I<Optional Arguments>

The default for all options is to not provide the option.

=over

=item B<fields> => I<fieldname, ...> The list of fields to be displayed

=item B<policy-name> => I<string, 1-32 characters long> The name of the policy

=item B<policyid> => I<integer> The unique id assigned to this policy name

=item B<last-modified> => I<"MM/DD/YYYY HH:MM:SS"> Show only those policies
    with this last modified date, the hour field is 24 hour format and must
    match exactly for the policy to be displayed

=back

=item Returns

=over

=item B<Success:> an array of hash references, or undef.
         hash defined in the notes section below.

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 If command fails for any reason

=back


=item Notes

=over

=item 1 This method will always provide the vserver option to the command.

=item 2 This command does not support the following options:

=over

=item * instance

=back

=item 3 The output hash definition.

=over

=item B<$href->{vserver}> => vserver name to which policy is associated

=item B<$href->{'policy-name'}> => the policy name provided by the caller
    at create time

=item B<$href->{policyid}> => a numerical string generated by the filer
    and unique to this policy

=item B<$href->{last-modified}> => the date the policy was last modified,
    '"MM/DD/YYYY HH:MM:SS"', the double quotes around the time stamp are
    part of the string and the hour field is 24 hour format

=back

=back

=item Examples

=back

    # display all the policies for a vserver of a given volume and get the
    # policy id for a specific policy
    $ar_policies = $vol->quota_policy_show();
    for my $p (@{$ar_policies}) {
        if ($p->{'policy-name'} eq $MyPolicy) {
            $MyPoilcyId = $p->{policyid};
        }
    }

    # an easier way to get the policy id, this time using a qtree object
    $ar_policies = $qtree->quota_policy_show( 'policy-name' => $MyPolicy );
    # quota_policy_show still returns an array reference but with only our entry
    $MyPolicyId = $p->[0]{policyid};

    # get only the policy id for my policy using a volume object
    $ar_policies = $vol->quota_policy_show( 'policy-name' => $MyPolicy,
                                            fields => 'policyid' );
    # quota_policy_show still returns an array reference but with only our entry
    $MyPolicyId = $p->[0]{policyid};

=cut

sub quota_policy_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($results, $out);

    # the filer will find bad arguments and if command options change
    # we won't have to.
    $results = eval {
                    $out = $self->server->volume_quota_policy_show(
                                        vserver => $self->Vserver->name,
                                        %args );
                    1;
                };

    if (! $results) {
        # turn any expected errors into returning undef (no data)
        if ($@ =~ /There are no entries matching your query/i) {
            $out = undef;
        } else {
            confess "quota policy show failed: $@";
        }
    }

    return $out;
}


=head2 quota_show

Show the quota policies for the volume.

=over

=item Arguments

Arguments are passed by hash table

Hash table arguments:

I<Optional Arguments>

The default for all options is to not provide the option.

=over

=item B<fields> => I<fieldname, ...> The list of fields to be displayed

=item B<state> => I<on|off> Display only quotas of this state

=item B<scan-status> => I<percent> Display quotas with this percentage

=item B<logging> => I<on|off> Display only quotas with logging on/off

=item B<logging-interval> => I<string> Display quotas wtih this logging interval

=item B<sub-status> => I<string> Display quotas with this sub quota status

=item B<last-error> => I<string> Display quotas whose last quota error was this

=item B<errors> => I<string> Display quotas who have reported this collection
    of quota errors

=item B<is-user-quota-enforced> => I<true|false> Display quotas that are/are
    not enforcing user quotas

=item B<is-group-quota-enforced> => I<true|false> Display quotas that are/are
    not enforcing group quotas

=item B<is-tree-quota-enforced> => I<true|false> Display quotas that are/are
    not enforcing tree quotas

=back

=item Returns

=over

=item B<Success:> an array of hash references, or undef.
         hash defined in the notes section below.

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 If command fails for any reason

=back

=item Notes

=over

=item 1 This method will always provide the vserver and volume options to
    the command.

=item 2 This command does not support the following options.

=over

=item * -instance

=item * -logmsg

=back

=item 3 The output hash definition. All undefined attributes are set to '-'
    except where specifically stated otherwise.

=over

=item B<$href->{vserver}> => vserver name

=item B<$href->{volume}> => volume name

=item B<$href->{state}> => quota state for the volume, on|off

=item B<$href->{'scan-status'}> => scan status, percentage

=item B<$href->{logging}> => quota logging for this volume, on|off

=item B<$href->{'logging-interval'}> => logging interval for this volume, string

=item B<$href->{'sub-status'}> => sub quota status, string

=item B<$href->{'last-error'}> => last error message for this quota, string

=item B<$href->{errors}> => collection of quota errors, string

=item B<$href->{'is-user-quota-enforced'}> => user quota enforced, true|false

=item B<$href->{'is-group-quota-enforced'}> => group quota enforced, true|false

=item B<$href->{'is-tree-quota-enforced'}> => tree quota enforced, true|false

=back

=back

=item Examples

=back

    # display the comoplete quota status for this volume
    $vol->quota_show();

    # display the comoplete quota status for this qtree's volume
    $qtree->quota_show();

    # check if quotas are enabled for this volume
    $ar_status = $vol->quota_show( field => 'state' );
    # quota_show still returns an array reference but with only our entry
    if ($ar_status->[0]{state} eq 'on') {
        # quota state is enabled
    } else {
        # quota state is disabled
    }

    # determine the user quotas are being enforced on this qtree's volume
    $ar_status = $qtree->quota_show();
    # quota_show still returns an array reference but with only our entry
    if ($ar_status->[0]{'is-user-quota-enforced'} eq 'true') {
        # user quotas are being enforced
    } else {
        # user quotas are not being enforced
    }

=cut

sub quota_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($results, $out);

    # the filer will find bad arguments and if command options change
    # we won't have to.
    $results = eval {
                    $out = $self->server->volume_quota_show(
                                        vserver => $self->Vserver->name,
                                        volume => $self->_get_volume_name,
                                        %args );
                    1;
                };

    if (! $results) {
        # turn any expected errors into returning undef (no data)
        if ($@ =~ /There are no entries matching your query/i) {
            $out = undef;
        } else {
            confess "quota show failed: $@";
        }
    }

    return $out;
}

1;
