#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/SnapMirror.pm#1 $
#
## @summary  GNOTE GX SnapMirror module
## @author   dl-dpgqa-automation@netapp.com
## @status   public
## @pod      here

=head1 NAME

NOTE::GX::SnapMirror

=head1 DESCRIPTION

This is SnapMirror object for the 10 mode SnapMirror relationship in GX

Commands by default will run in 'diag' mode. CLI parameter TRICKY_PARITY can
be used to determine if the object runs in tricky parity mode.  In tricky parity
mode certain commands runs in test mode.  The default is TRICKY_PARITY=1.

=head1 METHODS

=cut

package NOTE::GX::SnapMirror;

use strict;
use warnings;
no warnings 'redefine';
use re 'eval';

use GNOTE;
use GNOTE::GX;
use GNOTE::SnapMirror;
use GNOTE::GX::Mirror;
use Carp qw(confess);
use Data::Dumper;
use TracePkg;

our @ISA = qw(NOTE::SnapMirror NOTE::GX);

my $Trace = TracePkg->new();

sub _set_src_dst {
    my (%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    my $args;
    my $opts;
    $args{dst} ||= {};
    $args{src} ||= {};

    $self->{_args} = $args = \%args;

    my $src_opts = {
        Node      => q/$NODES_SRC:$NODES:$FILER_SRC:$FILER/,
        Vserver   => q/$VSERVER_SRC:$VSERVER/,
        name      => q/$SRC_VOL:$FILER_SRC_VOL/,
        tag       => 'Source volume',
        type      => q/$VOL_TYPE/,
        recycle   => 1,
        size      => q/$VOL_SIZE/,
        Aggregate => {
            name  => q/$SRC_AGGR:$FILER_SRC_AGGR/,
            tag   => 'Source aggregate',
            share => 1,
        },

        # '*' means use all nodes in the cluster
        striped => { Aggregate => { nodes => '*' }, },
    };

    if ( ref $args->{src} eq "HASH" ) {
        if ( ref \$args->{src} eq 'SCALAR' && $args->{src} ne 'NOTE::Volume' ) {
            confess "Invalid src specification: " . $args->{src};
        }
        $opts = $src_opts;
        if ( ref $args->{src} eq 'HASH' ) {
            $opts = NOTE::_hashOverlay(
                target    => $opts,
                source    => $args->{src},
                overwrite => 1
            );
        }
        $self->{_src} = NOTE::Volume->create( %{$opts} );
    }
    elsif ( ( ref $args->{src} eq "NOTE::GX::Volume" ) ) {
        $self->{_src} = $args->{src};    # already an object
    }
    else {
        confess "Invalid src specification: " . ref $args->{src};
    }
    $self->{_src_cluster} = $args->{src_cluster};    # inter-cluster workaround, to be removed

    # Destination
    if ( ref $args->{dst} eq "HASH" ) {
        $opts = {
            Filer => q/$NODES_DST:$NODES:$FILER_DST:$FILER_SRC:$FILER/,
            name  => q/$DST_VOL:$FILER_DST_VOL/,
            tag   => 'Destination volume',
            $self->{_src}->Properties,
        };
        delete $opts->{dataset};                     # Don't inherit dataset property from source
        $opts->{Aggregate} = {
            name    => q/$DST_AGGR:$FILER_DST_AGGR/,
            tag     => 'Destination aggregate',
            recycle => 1,
            nodes   => scalar $self->{_src}->Aggregate->nodes,
            $self->{_src}->Aggregate->Properties,

            # No shared destination aggregate if wafliron will run
            share => !NOTE::option(q/$WAFLIRON/),
        };
        $opts = NOTE::_hashOverlay(
            target    => $opts,
            source    => $args->{dst},
            overwrite => 1
        );
        $self->{_dst} = NOTE::GX::Volume->new( %{$opts} );
    }
    elsif ( ref $args->{dst} eq "NOTE::GX::Volume" ) {
        $self->{_dst} = $args->{dst};
    }
    else {
        confess "Invalid dst specification: " . ref \$args->{dst};
    }
    $self->{_dst_cluster} = $args->{dst_cluster};    # inter-cluster workaround, to be removed

    return $self;
}

# Work around method to provide backward compatibility to src_cluster and dst_cluster

sub _cluster_name {
    my ( $self, $type ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $key     = "_" . $type;
    my $tmp_key = $key . "_cluster";
    return (
        defined $self->{$tmp_key}
        ? $self->{$tmp_key}
        : $self->{$key}->Cluster->name
    );
}

=head2 new

Creates an object representing a SnapMirror relationship but does B<not> cause
the creation of a snapmirror relationship on the GX Vserver.  The create() or
initialize() methods may be executed on the new object to create the
SnapMirror relationship on the filer.

This method should be directly used only in exceptional cases.  create() or
initialize() should be used ordinarily.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item B<src> => I<NOTE::Volume spec>

A NOTE::Volume Object, or a hash containing the specifications to create a
NOTE::Volume, to be used as the source of the snapmirror relationship.

=item B<dst> => I<NOTE::Volume spec>

A NOTE::Volume object, or a hash containing the specifications to create a
NOTE::Volume, to be used as the destination of the snapmirror relationship.

=back

=item Returns

=over

=item B<Success:> $self (SnapMirror object)

=item B<Failure:> confesses

=back

=item Examples

=back

    my $sm = NOTE::SnapMirror->new(
                 src => $src_vol,
                 dst => { tag  => 'sm dst', Node => $FILER_DST,
                          type => 'flex',   size => '20g'
                        });

=cut

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing src parameter"
      unless exists $args{src};
    confess "Missing dst parameter"
      unless exists $args{dst};

    # Use Mirror internal method since the conditions are the same
    my $self = _set_src_dst(%args);

    $self->{_owner}     = $$;
    $self->{_destroyed} = 1;

    my $util = NOTE::SnapMirrorUtil->new();
    $self->{_timeout} = $util->param("JOB_TIMEOUT") || 60 * 60;    # Default to 1 hour timeout
    my $throttle;
    eval {
        $throttle = $util->param("THROTTLE");
    };
    $self->{_throttle} = $throttle unless $@;

    bless $self, $class;
    NOTE::reference($self);

    # Run in tricky_parity mode if either src or dst Node is in tricky_parity mode
    $self->{_tricky_parity} = $self->src->Node->tricky_parity || $self->dst->Node->tricky_parity;

    $self->{_type} = $args{type} || $util->param('MIRROR_TYPE') || "dp";

    return $self;
}

=head2 create

Creates a snapmirror relationship and returns an object representing the
relationship.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item B<src> => I<NOTE::Volume spec>

A NOTE::Volume Object, or a hash containing the specifications to create a
NOTE::Volume, to be used as the source of the snapmirror relationship.

=item B<dst> => I<NOTE::Volume spec>

A NOTE::Volume object, or a hash containing the specifications to create a
NOTE::Volume, to be used as the destination of the snapmirror relationship.

=back

I<Optional Arguments>

=over

=item B<type> => I<XDP|DP|LS>

Type of Mirror: DP, LS or XDP. Default: DP

=item B<throttle> => I<integer>

Sets max transfer speed in KB per sec.

=item B<tries> => I<integer>

Number of retries if encounter error.

=item B<schedule> => I<string>

Name of the schedule to set for this mirror.

=item B<policy> => I<string>

Name of the snapmirror policy to use for this mirror.

=back

I<Optional Environment Arguments>

=over

=item B<MIRROR_TYPE> => I<XDP|DP|LS>

If create is called without explicit mirror type then
CLI and shell environment is being checked for B<MIRROR_TYPE> variable.
There is no default for global B<MIRROR_TYPE> variable.

=back

=item Examples

=back

    my $sm = NOTE::SnapMirror->create(src  => $src_vol,
                                      dst  => $dst_vol,
                                      type => 'LS');

    my $sm = NOTE::SnapMirror->create(src  => $src_vol,
                                      dst  => $dst_vol,
                                      type => 'XDP',
                                      policy => 'xdp_policy');

=cut

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $self;
    if ( ref($class) =~ /NOTE::.*SnapMirror/ ) {
        $self = $class;
    }
    else {
        $self = __PACKAGE__->new(%args);
    }

    my %opts;
    $opts{'destination-volume'}  = $self->{_dst}->name;
    $opts{'destination-vserver'} = $self->{_dst}->Vserver->name;
    $opts{'destination-cluster'} = $self->_cluster_name("dst");
    $opts{'source-volume'}       = $self->{_src}->name;
    $opts{'source-vserver'}      = $self->{_src}->Vserver->name;
    $opts{'source-cluster'}      = $self->_cluster_name("src");
    $opts{'type'}		 = $self->{_type};
    $opts{'throttle'}            = $self->{_throttle} if defined $self->{_throttle};

    $self->_hash_xfer(
        label => [ 'comment', 'schedule', 'throttle',
                   'tries', 'type', 'policy' ],
        from  => \%args,
        to    => \%opts
    );

    $opts{'-Timeout'} = 120;

    if ( $self->{_src}->Cluster ne $self->{_dst}->Cluster ) {
        $self->{_dst}->Cluster->peer_create( 'cluster' => $self->{_src}->Cluster );
    }
    unless ( delete $args{cmd_only} ) {
        $self->{_src}->Cluster->license( "SnapMirror_" . uc( $opts{type} || "DP" ), 1 );
        $self->{_dst}->Cluster->license( "SnapMirror_" . uc( $opts{type} || "DP" ), 1 );
    }

    NOTE::logcomment( "Creating snapmirror: " . $self->description );

    my $diff_geometry = $self->_diff_geometry;
    my $stripe_sm = $self->src->is_stripe || $self->dst->is_stripe;

    if ( $stripe_sm && $self->{_tricky_parity} ) {

        # Set paloma_allow_striped_replication flag for striped volumes in TRICKY_PARITY mode
        eval { $self->_setflag("paloma_allow_striped_replication"); };    # Ignore error
    }
    if ($diff_geometry) {

        # Set paloma_allow_unlike_geometry in src
        $self->_setflag("paloma_allow_unlike_geometry");
    }

    # test mode required for different geometry or striped volumes in TRICKY_PARITY mode
    my $test_mode = $stripe_sm || $diff_geometry;

    $test_mode = delete $args{test}
      if exists $args{test};                                              # To override set test => 0

    $self->{_dst}->server->set( privilege => "test", confirmations => "off" )
      if $test_mode;

    $opts{"-Verification"} = undef;
    eval { $self->{_dst}->server->snapmirror_create(%opts); };
    confess "$@" if $@;

    $self->{'relationship-id'} = $self->status->{'relationship-id'} =~ /^\S+\d\S+$/ ? $& : undef; # set undef if '-'

    $self->{_dst}->server->set( privilege => "diag", confirmations => "off" )
      if $test_mode;

    $self->{_created} = 1;

    $self->{_destroyed} = 0;

    return $self;
}

sub _setflag {
    my ( $self, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %done;
    foreach my $node ( $self->src->Aggregate->Nodes, $self->dst->Aggregate->Nodes ) {
        unless ( exists $done{ $node->name } ) {
            $node->setflag( name => $flag, value => 1 );
            $done{ $node->name } = 1;
        }
    }
}

=head2 initialize

Performs a 'snapmirror initialize' to perform the baseline transfer of data.
If the relationship have not been created, it will call the C<::create>
method to create the relationship.


Returns:
SUCCESS: SnapMirror object
FAILURE: confesses

Confessions:
1) if the snapmirror create command fails for any reason
2) if the snapmirror initialize command fails for any reason

=over

=item Arguments
Arguments are passed by hash.

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
source-cluster, source-vserver and source-volume options.

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.

B<target-snapshot> => I<snapshot name (string)>

The name of a specific snapshot to be used for the initialize.
The default is to not provide a target snapshot

B<throttle|k> => I<integer|'unlimited'>

Max transfer rate in KB / secs. Default: no option provided = unlimited.

[ B<foreground|w> => I<true|false>
  | B<background> => I<1|0> ]

Whether the filer should wait for the command to complete before returning
(foreground|w) or start the command in the background. Default see notes.

B<timeout> => I<seconds to wait for initialize to complete>

Used when the initialize is done in the foreground so Server won't time it out.
The default is the GNOTE default, currently 3600 seconds

B<-Timeout> => I<seconds for Server to wait for the command to return>

Used when the initialize is done in the background but the command might take
longer than 40 seconds to return. The default is to use the Server timeout,
currently 40 seconds

=item Notes

1)  The ntest parameters SNAPMIRROR_CLUSTER_PATH and SNAPMIRROR_7G_PATH
    override all method arguments causing GNOTE to generate commands using
    either 7g or cluster style options.
2)  The default functionality of the method is to describe the source volume
    in terms of the options 'source-cluster', 'source-vserver' and
    'source-volume'. If either 'source-path' or 'S' argument is provided then
    the source volume will be described as a path using the given option.
    * Argument 'source-path' produces a command using the '-source-path' option
    * Argument 'S' produces a command using the '-S' option
3)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.
4)  The default action if none of foreground, w, or background arguments are
    provided is to run the command in the background and monitor the command
    until it completes. Effectively making it look to the caller as if the
    command was run in the foreground.
5)  This method does not provide a way to access the '-type' command line
    option. If 'type' is passed as an argument it is ignored.
6)  This method provides an additional check if run in the background and is
    type XDP. A check is made that the correct snapshot is transfered.

=item Examples

=back

    # initialize a snapmirror relationship using test defaults
    my $sm = NOTE::SnapMirror->initialize();

    # initialize a snapmirror relationship using existing volumes
    my $vol_src = NOTE::Volume->create();
    my $vol_dst = NOTE::Volume->create();
    my $sm = NOTE::SnapMirror->initialize( src => $vol_src,
                                           dst => $vol_dst );

    # initialize a snapmirror relationship that has already been created
    $sm = $sm->initialize();

    # initialize a snapmirror relationship that has already been created in
    # the background using a source path and letting the destination volume
    # be defined by the defaults
    $sm = $sm->initialize( S => '//vserver/volume', foreground => 'false' );

    # initialize a snapmirror relationship using a source and destination path
    # and run in the foreground
    $sm = NOTE::SnapMirror->initialize( src => $vol_src,
                                        dst => $vol_dst,
                                        'source-path' => 'volume',
                                        'destination-path' => 'volume',
                                         w => 'true' );

    # initialize a snapmirror relationship letting the volumes be described by
    # cluster/vserver/volume options and run in the foreground
    my $sm = NOTE::SnapMirror->initialize( background => 1 );

=cut

sub initialize {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Option used only with Vault type to avoid creation of a ss prior to
    # issuing the update command. Used as a means to NOT create ss.
    # Expect to be undefined or 0.
    my $skip_ss = defined $args{'skip_ss'} ? delete $args{'skip_ss'} : 0;

    my ( $self, $background, $text, $opts, $rtn );

    # Create the mirror first
    if ( ref($class) =~ /NOTE::.*SnapMirror/ ) {
        $self = $class;
        $self->create(%args)
          unless $self->{_created};
    }
    else {
        $self = __PACKAGE__->create(%args);
    }

    # build the appropriate command options
    $opts = $self->_adjust_command_options( %args, 'source-needed' => 1 );

    # deal with the conflick of the background, foreground and w arguments.
    # note that foreground and w are no longer valid so should never be
    # copied to opts.
    $background = NOTE::GX::background( \%args );

    # handle any additional arguments
    if (exists($args{'target-snapshot'})) {
        $opts->{'target-snapshot'} = $args{'target-snapshot'};
    }

    $opts->{'-Verification'} = undef;

    NOTE::logcomment( 'Initialize snapmirror: ' . $self->description );
    $rtn = eval {
                $text = $self->{_dst}->server->snapmirror_initialize(%{$opts});
                1;
            };
    confess "snapmirror initialize failed: $@" if (! $rtn);

    # No actual snapshot to check against passing regex to match not dash
    $self->_check_job( $background, $text, $args{timeout}, '^(?!(^\-$))' );
    return $self;
}

=head2 update

Performs a 'snapmirror update' to perform the incremental transfer of data.

Returns:
SUCCESS: 0
FAILURE: confesses

Confessions:
1) if the snapmirror update command fails for any reason

=over

=item Arguments

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
source-cluster, source-vserver and source-volume options.

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.

B<source-snapshot> => I<snapshot name (string)>

The name of a specific snapshot to be used for the update.
The default is to not provide a source snapshot

B<throttle|k> => I<integer|'unlimited'>

Max transfer rate in KB / secs. Default: no option provided = unlimited.

[ B<foreground|w> => I<true|false>
  | B<background> => I<1|0> ]

Whether the filer should wait for the command to complete before returning
(foreground|w) or start the command in the background. Default see notes.

B<timeout> => I<seconds to wait for update to complete>

Used when the update is done in the foreground so Server won't time it out.
The default is the GNOTE default, currently 3600 seconds

B<-Timeout> => I<seconds for Server to wait for the command to return>

Used when the update is done in the background but the command might take
longer than 40 seconds to return. The default is to use the Server timeout,
currently 40 seconds

B<skip_ss> => I<0|1>

Used for XDP relationships only.
Value of 1 results in skipping the generation of a snapshot for update to
transfer.
Value of 0 or even UNDEF for this parameter results in generation of a
snapshot for the update to transfer.

=item Notes

1)  The ntest parameters SNAPMIRROR_CLUSTER_PATH and SNAPMIRROR_7G_PATH
    override all method arguments causing GNOTE to generate commands using
    either 7g or cluster style options.

2)  The default functionality of the method is to describe the source volume
    in terms of the options 'source-cluster', 'source-vserver' and
    'source-volume'. If either 'source-path' or 'S' argument is provided then
    the source volume will be described as a path using the given option.
    * Argument 'source-path' produces a command using the '-source-path' option
    * Argument 'S' produces a command using the '-S' option

3)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.

4)  The default action if none of foreground, w, or background arguments are
    provided is to run the command in the background and monitor the command
    until it completes. Effectively making it look to the caller as if the
    command was run in the foreground.

5)  This method provides an additional check if run in the background and is
    type XDP. A check is made that the correct snapshot is transfered.


=item Examples

=back

    # update using all defaults, source and destination volumes described by
    # cluster/vserver/volume tuple
    $sm = NOTE::SnapMirror->initialize(src => $vol_src, dst => $vol_dst);
    $sm->update();

    # update in the background using volume paths, then wait for the update
    # to complete
    $sm->update( S => '//vserver/volume',
                 'destination-path' => '//vserver/volume',
                 w => 'false' );
    $sm->monitor();

    # update in the foreground using all the bandwidth possible and giving
    # the update 10 minutes to complte. allowing the volumes to be described
    # by cluster/vserver/volume tuples
    $sm->update( background => 0, k => 'unlimited', timeout => 600 );

    # update in the background using the source path and giving the command
    # 60 seconds to return
    $sm->update( foreground => 'false',
                 'source-path' => 'volume',
                 '-Timeout' => 60 );

    # update in the foreground giving the update 5 minutes to complete
    $sm->update( w => 'true', timeout => 300 );

=cut

sub update {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Option used only with Vault type to avoid creation of a ss prior to
    # issuing the update command. Used as a means to NOT create ss.
    # Expect to be undefined or 0.
    my $skip_ss = defined $args{'skip_ss'} ? delete $args{'skip_ss'} : 0;

    my ( $opts, $rtn, $background, $text, $util, $ss_name );

    # needed for later hack
    $util = NOTE::SnapMirrorUtil->new();

    # build the appropriate command options
    $opts = $self->_adjust_command_options( %args, 'source-needed' => 0 );

    # deal with the conflick of the background, foreground and w arguments.
    # note that foreground and w are no longer valid so should never be
    # copied to opts.
    $background = NOTE::GX::background( \%args );

    # qa requested that a source volume description not be provided if gnote
    # is running in 7g test mode (SNAPMIRROR_7G_PATH is set) but the default
    # functionality of gnote was to provide a source volume description so
    # add it back if necessary. - stuw
    if ((!exists($util->{_SNAPMIRROR_7G_PATH})) &&
        (!exists($opts->{S})) && (!exists($opts->{'source-path'}))) {
        $opts->{'source-volume'} = $args{'source-volume'}
                                   || $self->{_src}->name;
        $opts->{'source-vserver'} = $args{'source-vserver'}
                                    || $self->{_src}->Vserver->name;
        $opts->{'source-cluster'} = $args{'source-cluster'}
                                    || $self->_cluster_name('src');
    }

    # handle any additional arguments
    if (exists($args{'source-snapshot'})) {
        $opts->{'source-snapshot'} = $args{'source-snapshot'};
	confess "Snapshot $args{'source-snapshot'} doesn't exists on volume " . $self->src->name unless $self->src->snapshot_exists($args{'source-snapshot'});
    }

    $opts->{'-Verification'} = undef;

    #Determine current exported snapshot
    my $snapshot = $self->status('exported-snapshot');

    # Provide a snapshot for vault transfer.
    # (Primarily for legacy tests ported from DP.)
    # If XDP type & update -source-snapshot is not configured
    # && user has not asked to skip this.
    if ( !defined $args{'source-snapshot'}
             && ($self->{_type} =~ /XDP/i) && !$skip_ss) {
        $ss_name  = $self->_xdp_create_ss();
	confess "Snapshot $ss_name was not created on volume " . $self->src->name unless $self->src->snapshot_exists($ss_name);
	NOTE::logcomment( "XPD Source Snapshot: $ss_name" );
    }


    my (@src_ss,@dst_ss) = ( $self->src->snapshot_list, $self->dst->snapshot_list );
    NOTE::logcomment( 'Update snapmirror: ' . $self->description . 
			'Source snapshot list: ' . Dumper(\@src_ss) ."\n" .
			'Destination snapshot list: ' . Dumper(\@dst_ss)
                    );
    $rtn = eval {
                $text = $self->{_dst}->server->snapmirror_update( %{$opts} );
                1;
            };
    confess "snapmirror update failed: $@" if (! $rtn);

    #Regex is to not match current exported snapshot for DP, but match for xdp
    NOTE::logcomment( "DEBUG - exported snapshot $snapshot");
    
    if( $self->{_type} =~ /(xdp|vault)/i ) {
	if( ! $background ) {
	    sleep 10;	# make sure we didn't catch state of older back to back updates
	    $self->snapmirror_wait(
				    'state' => 'Snapmirrored',
				    'last-transfer-type' => 'update',
				    'status' => 'Idle',
				    'timeout' => $args{timeout} ? $args{timeout} : 900,	# 15 minutes at least
			      );
	    sleep 30;	# sleep for exported snapshot to be remounted if needed, this not always a case
	}
    } else {
	$self->_check_job( $background, $text, $args{timeout}, '^(?!(^' . $snapshot . '$))' );
    }

    unless( $args{background} || $args{foreground} =~ /false/ ) {
	if (defined $ss_name && !$self->_xdp_ss_check($ss_name)) {
	    confess 'XDP snapshot generated for update failed transfer.';
	}
    }

    return 0;
}

=head2 logical_dir_xfer

Sets or clears paloma_force_logical_dirs flag.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<flag> = I<boolean>

Boolean flag to indicate to set or clear paloma_force_logical_dirs flag.
Default: 1

=back

=item Examples

=back

    $sm->logical_dir_xfer();     # Sets paloma_force_logical_dirs flag
    $sm->logical_dir_xfer(0);    # Clears the flag

=cut

sub logical_dir_xfer {
    my ( $self, $value ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $value = 1 unless defined $value;

    my $flag = "paloma_force_logical_dirs";
    $self->src->Node->setflag( name => $flag, value => $value );
    $self->dst->Node->setflag( name => $flag, value => $value );
}

=head2 wait_for

Wait for the snapmirror relationship to reach a certain conidition
by specifying values for columns in the 'snapmirror show' output.

This method provides for the ability to wait for a match or
mismatch of the values given, and also to specify a value
or use the current snapmirror show value.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item B<match> => I<boolean>

Determine if waiting for condition to match or not. Default: 1

=item B<timeout> => I<integer>

Timeout in seconds to wait for specified condition. Default: 3600 (1 hour)

=item B<interval> => I<integer>

Polling interval to check for specified condition. Default: 15

=item B<conditions> => I<column=>expected>

A list of conditions of the form <column>=><expected>,
where <column> is name of 'snapmirror show' column name
and <expected> is the expected value for the column

=item B<current> => I<comma separated columns>

A comma separated list of 'snapmirror show' column names.
This allows for the specification of using the current
values of these columns instead of giving a value.

Columns listed by this parameter and the <conditions> parameter above
are mutually exclusive.

=back

=item Returns

=over

=item B<Success:> Nothing specific.  (I.e., Do not use returned value.)

=item B<Failure:> confesses

=back

=item Examples

=back

    # wait for mirror to be idled
    $sm->wait_for(state  => "snapmirrored", status => "idle");

    # wait specific snapshot, polling every 5 secs
    $sm->wait_for('transfer-snapshot'  =>  $snapshot_name, interval => 5);

    # Wait until the current state is not snapmirrored and
    # change in snapshot-checkpoint
    $sm->wait_for(match   => 0,
                  state   => 'snapmirrored',
                  current => 'snapshot-checkpoint');

    # To wait for a snapshot-checkpoint change
    $sm->wait_for(match => 0, current => "snapshot-checkpoint');

=cut

sub wait_for {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $match    = ( exists $args{match}    ? delete $args{match}    : 1 );
    my $timeout  = ( exists $args{timeout}  ? delete $args{timeout}  : $self->{_timeout} );
    my $interval = ( exists $args{interval} ? delete $args{interval} : 15 );
    my $end_time = time + $timeout;

    my %opts = $self->_get_wait_args(%args);
    my $action = ( $match ? "" : "mis" ) . "match";
    my $criteria;
    my $cnt = 0;
    foreach my $key ( keys %opts ) {
        $criteria .= "\n    $key => " . $opts{$key};
        $cnt++;
    }
    confess( $self->description . "No criteria to wait for.." )
      unless $cnt;
    NOTE::logcomment("Wait for $action of: $criteria");
    do {
        my $found = 1;
        my $s     = $self->status;
        my ( $good, $bad );
        foreach my $key ( keys %opts ) {
            my $actual;

            # Verify condition/key is valid
            if ( exists $s->{$key} ) {
                $actual = $s->{$key};
            }
            else {
                confess "No such column $key";
            }

            # check condition
            my $expected       = $opts{$key};
            my $compare_result = ( $actual =~ /$expected/i );
            my $msg            = "\n    $key actual($actual), expected($expected)";
            if ( $match == $compare_result ) {
                $good .= $msg;
            }
            else {
                $bad .= $msg;
                $found = 0;
            }
        }

        # Process result of check
        if ($found) {
            NOTE::logcomment( $self->description . "- Found $action:" . $good );
            return $found;
        }
        else {
            NOTE::logcomment( $self->description . "- Failed to $action:" . $bad );
        }
    } while $self->_pause( $interval, $end_time );
    confess "Timeout waiting for specified condition(s): $criteria";
}

sub _pause {
    my ( $self, $interval, $end_time ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $remain = $end_time - time;
    if ( $remain > 0 ) {
        NOTE::logcomment("Pause a little...");
        Tharn::snooze( $remain < $interval ? $remain : $interval );
    }
    else {
        return 0;
    }
}

=head2 snapmirror_wait

Pull snapmirror status in given interval and compare return hash against
submitted arguments until they match or timeout takes place.

Method can be used to fait for snapmirror to reach requested state. Each value of
the arguments is treated as regular expression.

=over

=item Arguments
Arguments passed by hash table and must contain at least one of keys returned by
SnapMirror.

=over

=item B<timeout> = I<number> of seconds to wait for all elements to match, default is 15 minutes

=item B<interval> = I<number> of seconds to wait between polling, default is 10 seconds

=item other arguments are keys from snapmirror show

 {
      'destination-volume' => 'vl00824_boss_ascribes',
      'source-volume' => 'cpeter',
      'lag-time' => '-',
      'total-progress' => '55644160',
      'source-vserver' => 'vs0',
      'state' => 'Uninitialized',
      'destination-path' => 'dpg-v-spectre9a1304395978://vs0/vl00824_boss_ascribes',
      'policy' => '-',
      'newest-snapshot-timestamp' => '-',
      'source-cluster' => 'dpg-v-spectre9a1304395978',
      'tries' => '8',
      'last-transfer-from' => '-',
      'schedule' => '-',
      'last-transfer-error' => '-',
      'last-transfer-type' => '-',
      'physical-replica' => '-',
      'snapshot-progress' => '55644160',
      'snapshot-checkpoint' => '4096',
      'throttle' => 'unlimited',
      'exported-snapshot-timestamp' => '-',
      'type' => 'DP',
      'destination-vserver' => 'vs0',
      'status' => 'Transferring',
      'newest-snapshot' => '-',
      'healthy' => '-',
      'current-transfer-type' => '-',
      'last-transfer-size' => '-',
      'destination-cluster' => 'dpg-v-spectre9a1304395978',
      'in-vserver-relationship' => '-',
      'last-transfer-duration' => '-',
      'QOS-policy' => '-',
      'percent-complete' => '-',
      'vserver' => 'vs0',
      'transfer-snapshot' => 'snapmirror.7edfeafb-7540-11e0-bcf6-123478563412_3_2147484715.2011-05-11_165841',
      'relationship-capability' => '-',
      'exported-snapshot' => '-',
      'identity-preserving' => '-',
      'progress-last-updated' => '-',
      'current-transfer-error' => '-',
      'source-path' => 'dpg-v-spectre9a1304395978://vs0/cpeter'
    };

=back

=item Examples

Waiting for transfer to be started:

      1 use strict;
      2 use GNOTE;
      3 eval {
      4     my $ut = NOTE::SnapMirrorUtil->new();
      5     my $sm = NOTE::GX::SnapMirror->initialize(
      6                                                 src=>$ut->create_volume(tag=>'src'),
      7                                                 dst=>$ut->create_volume(tag=>'dst'),
      8                                                 background=>1
      9                                              );
     10     $sm->snapmirror_wait( 'total-progress'=>'^(\d+)', 'status' => 'Transferring', 'state' => 'Uninitialized', timeout => 600 );
     11     # do something here...
     12	    # while snapmirror runs...
     13 };
     14 logresult('FAIL',$@) if $@;
     15 NOTE::teardown;

Wait for first Gb of data to be transferred using 40 seconds polling interval, then continue after 1G is moved

      1 use strict;
      2 use GNOTE;
      3 eval {
      4     my $ut = NOTE::SnapMirrorUtil->new();
      5     my $sm = NOTE::GX::SnapMirror->initialize(
      6                                                 src=>$ut->create_volume(tag=>'src'),
      7                                                 dst=>$ut->create_volume(tag=>'dst'),
      8                                                 background=>1
      9                                              );
     10     $sm->snapmirror_wait( 'total-progress'=>'^(\d+)(?(?{$1>1073741824})(?!))', 'status' => 'Transferring', interval=>40 );
     11     # do something
     12 };
     13 logresult('FAIL',$@) if $@;
     14 NOTE::teardown;

=back

=cut

sub snapmirror_wait {
    my ($self,%args) = @_;

    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $maxtime    = time() + ( delete $args{timeout} || 900 ); # wait {timeout} seconds or 15 minutes
    my $interval   = delete $args{interval} || 10;              # check state every 10 seconds
    my ($matchcount,$foundmatch,$status) = ( scalar( keys %args ), undef, undef );

    confess "No parameters are supplied to match against snapmirror show output." if $matchcount < 1;

    do {
	$status = $self->status;
	print Dumper($status);
        $foundmatch = scalar( grep( $status->{$_} =~ qr/$args{$_}/, keys %args ) );
	sleep  $interval if( $foundmatch < $matchcount );
    } while ( $foundmatch < $matchcount && $maxtime > time() );
    confess "Timeout waiting for:".Dumper(\%args)."\nReceived so far: ".Dumper($status) if $foundmatch < $matchcount;
}

# Use current snapmirror show values

sub _get_wait_args {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( exists $args{current} ) {
        NOTE::logcomment( $self->description . "- set current show value" );
        my $s = $self->status;
        foreach my $key ( split ",", $args{current} ) {
            confess "Multiple specification for column $key not allowed"
              if exists $args{$key};
            confess "Invalid column $key"
              unless exists $s->{$key};
            $args{$key} = $s->{$key};
        }
        delete $args{current};
    }
    else {
        NOTE::logcomment( $self->description . "- no current parameter" );
    }
    return %args;
}

=head2 common_snapshots

Return a list of common snapshot names between src and destination volume

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item I<None>

=back

=item Returns

=over

=item B<Success:> Array of snapshots

=item B<Failure:> May return an empty list, but will never fail.

=back

=item Examples

=back

    my $ssn_list = $sm->common_snapshots();

=cut

sub common_snapshots {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @src_ssn = $self->src->snapshot_list();
    my %dst_ssn = map { $_ => 1 } $self->dst->snapshot_list();
    my @common_list;
    foreach my $s_ssn (@src_ssn) {
        push @common_list, $s_ssn
          if exists $dst_ssn{$s_ssn};
    }
    return @common_list;
}

=head2 monitor

Monitors the snapmirror action by checking until the snapmirror reaches a
specific state/status combination.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<state> => I<expected state>

Wait until mirror is at this state. Default: snapmirrored

=item B<status> => I<expected status>

Wait until mirror is at this status. Default: idle

=item B<timeout> => I<timeout>

Timeout in seconds. Default: 3600 (1 hour)

=item B<interval> => I<interval>

Pause interval in between check, in seconds. Default: 5

=back

=item Returns

=over

=item B<Success:> Nothing specific.  (I.e., Do not use returned value.)

=item B<Failure:> confesses on error or timeout

=back

=item Examples

=back

    $sm->monotor();
    $sm->monitor(timeout => 60*60,          interval => 15,
                 state   => 'snapmirrored', status   => 'idle');

=cut

sub monitor {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $exp_state  = $args{state}    || 'snapmirrored';
    my $exp_status = $args{status}   || 'idle';
    my $timeout    = $args{timeout}  || $self->{_timeout};
    my $interval   = $args{interval} || 5;

    NOTE::logcomment( 'Checking ' . $self->description . " for\n    ($exp_state, $exp_status):" );

    # GX does not provide progress such as amount of data transferred.
    # So for now just check state-status combination to determine completion state
    my ( $progress, $stat );

    my $remain    = $timeout;
    my $sm_exists = 0;          # flag to indicate whether existing snapmirror status have been previously retrieved by this call
    do {
        Tharn::snooze($interval);

        my $active = $self->_job_active();
        eval { $stat = $self->status; };
        if ($@) {

            # Ignore 'no entries' error as some jobs creates the mirror relationship
            # which may not exists at the time of the first status query
            if ( $sm_exists || !( $@ =~ /no entries/i ) ) {
                confess "snapmirror status error $@";
            }
            else {
                NOTE::logcomment("Ignoring snapmirror status error: $@");
            }
        }
        if (   $stat
            && $stat->{state}  =~ /$exp_state/i
            && $stat->{status} =~ /$exp_status/i )
        {
            $sm_exists = 1;
            my $ignore = 0;
            if ( $exp_status =~ /idle/i ) {

                # Verify job completion
                if ( $self->_job_active() ) {
                    NOTE::logcomment( 'Job ' . $self->{_job_id} . ' still running' );
                    $ignore = 1;
                }
                else {
                    delete $self->{_job_id};
                }
            }
            unless ($ignore) {
                NOTE::logcomment('Monitor completed.');

                # if a background snapmirror check just completed, collect the
                # result and save it for later.
                if ( $self->{_bg_check_in_progress} ) {
                    $self->{_bg_check_result} = _get_sm_check_result( $self->{_progress}, $self->{_bg_check_abort_on_mismatch} );
                    $self->{_bg_check_in_progress} = 0;
                    # We need to release the lock on the target snapshot in case
                    # of checker.
                    if (exists $self->{_bg_locked_snapshot}) {
                        _snapshot_checker_op(
                                    op => 'unlock',
                                    snapshot => $self->{_bg_locked_snapshot},
                                    volume => $self->dst->name,
                                    node   => $self->{_dst},
                        );
                        delete $self->{_bg_locked_snapshot};
                    }
                }
                return;
            }
        }

        if ( $progress eq $self->{_progress} ) {

            # Countdown the clock if no progress was made
            $remain -= $interval;
            NOTE::logcomment("$remain sec(s) to timeout 'clock' remaining...");
        }
        $progress = $self->{_progress};

        # gradually increase interval of polling to 1 mins unless explicitly set by call
        $interval += 5 unless ( $interval > 44 || defined $args{interval} );
    } while $remain > 0;

    confess $self->description . ": timed-out after $timeout.";
}

=head2 status

Returns a hash reference to the status of the relationship.

=over

=item Arguments

=over

=item B<element> = I<status element>

If defined returns just the named status element.

=item B<timeout> => I<integer>

Time out in secs. Default: 300

=back

I<Optional Arguments>

=over

=item B<exec_vol> => I<src|dst>

Cross cluster configuration permits calling show from Source or Destination
node. Default is destination.

Parameters for status are positional, so be sure to add I<undef> for the element
if you are not supplying an element argument. See examples.

=back

=item Returns

=over

=item B<Success:> A hashref to the status of the relationship, or just the named
status element.

=item B<Failure:> confesses

=back

=item Examples

=back

    # Setting timeout for 1 single column query
    my $state = $sm->status('state', timeout => 600);

    # Setting timeout when obtaining entire hash
    my $status = $sm->status(undef, timeout => 900);

    # Call show from Source node.
    my $status = $sm->status(undef, exec_vol => 'src');

=cut

sub status {
    my ( $self, $element, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $stat;

    NOTE::logcomment( 'Snapmirror status: ' . $self->description );

    my $exec_vol = _set_execution_srv(\%args);

    # GX does not have 'sm status' command, so using 'sm show' instead
    my $timeout = delete $args{timeout} || 300;
    eval {
        $stat = $self->{$exec_vol}->server->snapmirror_show(
            # original requirement for SN was to provide 'destination-path' in SN format. Keep for reference
	    # 'destination-path' => $self->dst->Cluster->name . '://' . $self->dst->Vserver->name . '/' . $self->dst->name,
            'destination-volume'  => $self->dst->name,
            'destination-vserver' => $self->dst->Vserver->name,
            'destination-cluster' => $self->dst->Cluster->name,
            '-Timeout'            => $timeout
        );
    };
    confess "$@" if $@;
    NOTE::logcomment( $self->description . ":\n " . $stat->[0]{state} . ', ' . $stat->[0]{status} );


    if ( defined $element ) {
        return $stat->[0]{$element};
    }
    else {
        return $stat->[0];
    }
}

=head2 wait_for_job

Wait for an active job for this mirror. Returns boolean to indicate
if a job was found. If a job is found, the 'monitor' method can be
used to monitor its progress.

For LS mirror scheduled jobs, the job will apply to the LS set as a
whole, and so if this mirror is part of a GX::SnapMirrorSet object
the job will be recorded in the set rather than in this object.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item B<timeout> => I<integer>

Timeout in seconds to wait for an active job. Default: 300

=item B<interval> => I<integer>

Polling interval in seconds. Default: 15

=back

=item Returns

=over

=item B<boolean> Returns true if an active job was found before the timeout,
false if the timeout is reached instead.

=back

=item Examples

=back

     # Monitor progress of any active job found within a minute.
     if ($sm->wait_for_job(timeout => 60)) {
         $sm->monitor();
    }

=cut

sub wait_for_job {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Return update-ls-set job id if exists
    my $sms = $self->dst->SnapMirrorSet();
    return $sms->{_job_id}
      if $self->{_type} =~ /ls/i && $sms && exists $sms->{_job_id};

# _job_id is removed in monitor when the job completes so if we have
# one we should have an active job
return $self->{_job_id} if (exists($self->{_job_id}));

# this code does not serve much purpose. it also is broken. during
# initialization the destination volume is restricted not online - stuw
    # Return false if src or dst volume is not online
    #return 0
      #unless ( $self->src->writeable() && $self->dst->readable() );

    my $end_time = time + ( exists $args{timeout} ? $args{timeout} : 300 );
    my $interval = exists $args{interval} ? $args{interval} : 15;

    # Check for snapmirror/daemon job for src or dst
    my $exp_dmn = "Snapmirror Daemon for \\d_" . $self->dst->msid();
    my $exp_src = "snapmirror update-ls-set for source " . $self->src->hostpath;
    my $exp_dst = "snapmirror update of destination " . $self->dst->hostpath;
    do {
        my @jobs = $self->_job_list();
        foreach my $j (@jobs) {

            # If this mirror is part of a SnapMirrorSet object search for the 'set'
            if ( $self->{_type} =~ /ls/i && $sms && $j->{description} =~ /($exp_src|$exp_dmn)/ ) {
                $sms->{_job_id} = $j->{id};
                return 1;
            }

            # Search for job specifiecally for this mirror
            if ( $j->{description} =~ /($exp_dst|$exp_dmn)/ ) {
                $self->{_job_id} = $j->{id};
                return 1;
            }
        }

        NOTE::logcomment( "No snapmirror job for " . $self->description );
    } while $self->_pause( $interval, $end_time );
    NOTE::logcomment( "No snapmirror job found for " . $self->description );
    return 0;
}

sub _job_list {
    my ( $self, $desc ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @job;
    eval { @job = $self->{_src}->server->job_show( state => "running" ); };
    if ($@) {
        confess "Job show error. Error $@"
          unless $@ =~ /no entries/;
        return undef;
    }
    else {
        return @job;
    }
}

=head2 is_catching_up

Returns boolean flag to indicate if this mirror is currently in catchup mode.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item I<None>

=back

=item Returns

=over

=item B<boolean> flag to indicate if this mirror is currently in catchup mode

=back

=item Examples

=back

     my $catchup_flag = $sm->is_catching_up();

=cut

sub is_catching_up {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $sms = $self->dst->SnapMirrorSet();
    if ($sms) {
        return ( $sms->catch_up( dst => $self->dst, timeout => 0 ) > 0 );
    }
    else {
        NOTE::logcomment( $self->description . " need to be part of a GNOTE::GX::SnapMirrorSet object to determine catup status..." );
        return 0;
    }
}

=head2 is_lagging

Returns boolean to indicate if this mirror is lagging behind others in the set.
This methods is meant to used for 'ls' mirrors.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

=over

=item I<None>

=back

=item Returns

=over

=item B<boolean> flag to indicate if this mirror is lagging behind others in the
set

=back

=item Examples

=back

    my $lag_flag = $sm->is_lagging();

=cut

sub is_lagging {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->status("exported-snapshot") eq "-" );
}

=head2 src_name

Returns the name of the source volume.

=over

=item Arguments

=over

=item None.

=back

=item Returns

=over

=item B<string> name of the source volume

=back

=item Example

=back

     my $srcname = $sm->src_name();

=cut

sub src_name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_src}->name;
}

=head2 dst_name

Returns the name of the destination volume.

=over

=item Arguments

=over

=item None.

=back

=item Returns

=over

=item B<string> name of the destination volume

=back

=item Example

=back

     $dstname = $sm->dst_name();

=cut

sub dst_name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_dst}->name;

}

=head2 src_path

Returns the vserver-qualified path name of the source volume.

=over

=item Arguments

=over

=item None.

=back

=item Returns

=over

=item B<string> the vserver-qualified path name of the source volume

=back

=item Example

=back

     $srcpath = $sm->src_path();

=cut

sub src_path {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return $self->{_src}->Vserver->name . ":" . $self->src_name;
}

=head2 dst_path

Returns the vserver-qualified path name of the destination volume.

=over

=item Arguments

=over

=item None.

=back

=item Returns

=over

=item B<string> the vserver-qualified path name of the destination volume

=back

=item Example

=back

     $dstpath = $sm->dst_path();

=cut

sub dst_path {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->{_dst}->Vserver->name . ":" . $self->dst_name;
}

=head2 destroy

Aborts (if needed) any update currently running on this SM, and then deletes the
SM relationship from the filer.

=over

=item Arguments

=over

=item None.

=back

=item Returns

=over

=item $self

=back

=item Example

=back

     $sm->destroy();

=cut

sub destroy {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    
    NOTE::logcomment( "Destroying snapmirror " . $self->description );

    if( Tharn::param('SMCHECKOVERRIDE') =~ /on/i ) {	# this is BURT 565351
	NOTE::logcomment('SMCHECKOVERRIDE is ON, forcing sm check first');
	eval { $self->check };
	logresult('FAIL', $@) if $@ && $@ =~ /\w+/;	# we need this for loggin test errors from teardown
    }

    $self->{_destroyed} = 1;

    # Remove the checker lock if the user forgot to unlock it.
    if (exists $self->{_bg_locked_snapshot}) {
        _snapshot_checker_op(
                    op => 'unlock',
                    snapshot => $self->{_bu_locked_snapshot},
                    volume => $self->dst->name,
                    node   => $self->{_dst},
        );
        NOTE::logcomment('Deleting checker lock in destroy');
        delete $self->{_bg_locked_snapshot};
    }

    # Perform sequence of events to remove the mirror - ignore error since
    # these steps may not be needed
    foreach my $action ( 'abort', 'delete', 'release' ) {
        eval { $self->$action; };
    }
    return $self;
}

=head2 check

Performs a "snapmirror check" to verify the integrity of
transferred data.  If the checker detects an inconsistent
mirror, anomalies are recorded in a consolidated log file
that is specified as a node and absolute pathname. Possible
return values:

    undef - The checker is running in background.  See ::get_check_result.

    0 - Checker ran successfully and found consistent mirrors

    $hashref - Checker ran successfully and found inconsistent mirrors where:

        node => $nodename
        path => $logfile_abs_path

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<target-snapshot> => I<string>

target-snapshot name for a specific snapshot to be
used to the snapmirror check.

=item B<background> => I<1|0>

Boolean flag to indicate background execution. Default: 0

=item B<timeout> => I<timeout>

Timeout in seconds. Default: 30 mins

=item B<abort_on_mismatch> => I<boolean>

Flag to indicate 'confessing' if mismatch found as opposed to returning hash of
differences.

=back

=item Returns

=over

=item B<Success:> See method description above.

=item B<Failure:> confesses

=back

=item Examples

=back

        $sm->check('target-snapshot' => "snapshot");

        $chk_log = $sm->check();
        if ( $chk_log ) { ... }

        $sm->check (background=>1);
        { ... }
        $sm->monitor();
        $chk_log = $sm->get_check_result();
        if ( $chk_log ) { ... }

=cut

sub check {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my %opts = ();
    my $snapshot_to_be_checked;

    my %old_checker_errlist = undef;   # checker generates files on dblade under /etc/log/.snapmirrorChecker/
    my %new_checker_errlist = undef;   # BURT 559234 requires to check for new checker files.

    %old_checker_errlist = map{$_,1} grep(/[0-9]/, split(/[\r\n]/, $self->dst->Node->server->dblade( cmd=>"priv set -q test; ls /etc/log/.snapmirrorChecker/" )));

    if( Tharn::param('SMCHECKOVERRIDE') =~ /off/i ) {	# this is BURT 565351
	NOTE::logcomment('SMCHECKOVERRIDE is set to OFF, not checks will be performed');
	return;
    }

    # throw a warning if a check is already running in background.
    # this can happen if the previous check failed and the _bg_check_in_progress flag
    # didn't get cleaned up.  if a check is already running in background, the new check
    # will fail.
    if ( $self->{_bg_check_in_progress} ) {
        NOTE::logcomment("WARNING: snapmirror check called when it is running in background.");
    }

    my $abort_on_mismatch = delete $args{abort_on_mismatch};

    exists $args{'destination-volume'}
      ? $opts{'destination-volume'} = $args{'destination-volume'}
      : $opts{'destination-volume'} = $self->dst->name;
    exists $args{'destination-vserver'}
      ? $opts{'destination-vserver'} = $args{'destination-vserver'}
      : $opts{'destination-vserver'} = $self->dst->Vserver->name;
    exists $args{'destination-cluster'}
      ? $opts{'destination-cluster'} = $args{'destination-cluster'}
      : $opts{'destination-cluster'} = $self->_cluster_name("dst");
    exists $args{'source-volume'}
      ? $opts{'source-volume'} = $args{'source-volume'}
      : $opts{'source-volume'} = $self->{_src}->name;
    exists $args{'source-vserver'}
      ? $opts{'source-vserver'} = $args{'source-vserver'}
      : $opts{'source-vserver'} = $self->{_src}->Vserver->name;
    exists $args{'source-cluster'}
      ? $opts{'source-cluster'} = $args{'source-cluster'}
      : $opts{'source-cluster'} = $self->_cluster_name("src");
    exists $args{'target-snapshot'}
      ? $opts{'target-snapshot'} = $args{'target-snapshot'}
      : delete $opts{'target-snapshot'};
    exists $args{'timeout'}
      ? $opts{'-Timeout'} = $args{'timeout'}
      : $opts{'-Timeout'} = 36000;    # 10 hours [BURT 434384]

    # BURT 633351
    # We have to lock down the snapshot that is going to be verified
    # by the checker. If the user mentioned a target snapshot, then
    # that needs to be locked down. Otherwise, if no target snapshot
    # has been passed by the user, then last snapshot has to be locked
    # down, so that it won't get rotated while checker is in progress.
    if ($self->{_type} =~ /(xdp|vault)/i) {
        if (exists $opts{'target-snapshot'}) {
            $snapshot_to_be_checked = $opts{'target-snapshot'};
        } else {
            # Get the last entry in the list as that should have the latest
            # snapshot.
            $snapshot_to_be_checked = $self->status( 'newest-snapshot');
        }
        _snapshot_checker_op(
                    op => 'lock',
                    snapshot => $snapshot_to_be_checked,
                    volume => $self->dst->name,
                    node   => $self->{_dst},
        );
    }

    # pick up the background argument.  the check command will be done in
    # background in any case (cf, $opts{foreground} is always false).  the
    # background flag just determines whether we hang around and wait for the
    # result here (background == false), or whether we just exit after starting
    # the check and let the caller get the result later (background == true).
    my $background = NOTE::GX::background( \%args );

    # this stuff gets used only if the check is being done in background.
    # for now, make all the background result info invalid.
    $self->{_bg_check_in_progress}       = 0;
    $self->{_bg_check_abort_on_mismatch} = 0;
    $self->{_bg_check_result}            = undef;
    if ($snapshot_to_be_checked) {
        $self->{_bg_locked_snapshot} = $snapshot_to_be_checked;
    }

    # do the check.  if it fails, confess here.
    my $text;
    eval { $text = $self->{_dst}->server->snapmirror_check(%opts); };
    if ($@) {
        NOTE::logcomment("error during snapmirror check");
        confess "$@";
    }

    $self->{_chk_log_content} = undef; # BURT 559234
    unless($background) { # BURT 559234 requires to check for new checker files.
	    my $log_content = undef;
	    %new_checker_errlist = map{$_,1} grep(/[0-9]/, split(/[\r\n]/, $self->dst->Node->server->dblade( cmd=>"priv set -q test; ls /etc/log/.snapmirrorChecker/" )));	# compare new list of
	    delete $new_checker_errlist{$_} foreach( keys %old_checker_errlist );												# files with old list
	    if( %new_checker_errlist ) {																	# and then
	        foreach( keys %new_checker_errlist ) {
		        $log_content .= "Checker files found: $_\n";
		        $log_content .= $self->dst->Node->server->dblade( cmd=>"priv set -q test; rdfile /etc/log/.snapmirrorChecker/$_" );
		        $log_content .= "\n\n\n";
	        }
	        NOTE::logcomment("$log_content");
	        $self->{_chk_log_content} = $log_content;
	    }
    }

    # if background, set up to monitor this job.  else wait for the
    # job to complete.
    $self->_check_job( $background, $text, $args{timeout} );

    $self->{_chk_job_id} = $self->{_job_id};    # Save check job id for 'checking' method to query state of job

    # if background, exit now.  the caller is responsible for picking
    # up the result later.
    if ($background) {
        $self->{_bg_check_in_progress}       = 1;
        $self->{_bg_check_abort_on_mismatch} = $abort_on_mismatch;
        return undef;
    } else { # if foreground, the sm check completed.  return the result
        if (exists $self->{_bg_locked_snapshot}) {
            _snapshot_checker_op(
                        op => 'unlock',
                        snapshot => $self->{_bg_locked_snapshot},
                        volume => $self->dst->name,
                        node   => $self->{_dst},
            );
            delete $self->{_bg_locked_snapshot};
        }
        return _get_sm_check_result( $self->{_progress}, $abort_on_mismatch );
    }
}

    # for backward compatibility only.  deprecated usage.
sub compare {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    &check;
}

=head2 checking

Returns a boolan to indicate if the 'snapmirror check' command is running.
Aborts if the 'snapmirror check' job fails or found inconsistencies.

With the above definition, it is possible to deduce that the 'snapmirror check'
completed successfully without inconsistencies if this method returns 0.

=over

=item Arguments

=over

=item B<None>

=back

=item Returns

=over

=item B<boolean> See method description above.

=item B<Failure:> confesses

=back

=item Examples

=back

    my $chk_running = $sm->checking();

=cut

sub checking {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "No background checker job found"
      unless exists $self->{_bg_check_in_progress};
    confess "No check job id found"
      unless exists $self->{_chk_job_id};

    my $id     = $self->{_chk_job_id};
    my $label  = "'snapmirror check' job $id";
    my $job    = NOTE::GX::Job->new( Cluster => $self->dst->Cluster, id => $id );
    my $status = $self->_get_check_info( $label, $job, $id, 0 ) || $self->_get_check_info( $label, $job, $id, 1 );
    confess $self->description . ": Unable to determine status for $label."
      unless $status;
    if ( exists $status->{state} && $status->{state} =~ /running/i ) {

        # If state is running than return
        NOTE::logcomment( $self->description . ": $label is running" );
        return 1;
    }
    else {

        # If state if not running or no state column, i.e. job history result, then check the completion column
        my $completion = $status->{completion};
        if ( $completion =~ /(abort|error|fail|inconsistent)/i ) {
            confess $self->description . " found error for $label: $completion";
        }
        elsif ( $completion =~ /(success|done)/i ) {
            NOTE::logcomment( $self->description . ": $label completed" );
            return 0;
        }
        else {
            confess $self->description . "Unexpected completion text for $label: $completion";
        }
    }
}

sub _get_check_info {
    my ( $self, $label, $job, $id, $history ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $status;
    eval { $status = $job->status( id => $id, history => $history ); };
    confess $self->description . ": encountered error while querying status for $label. Error: $@"
      if $@ && !( $@ =~ /no entries/ );
    return $status;
}

sub _snapshot_checker_op {
    my %opts = @_;
    my $op;

    if ($opts{'op'} eq 'lock') {
        $op = '-s';
    } else {
        $op = '-d';
    }

    # The output of the following dblade command looks like this:
    # VOPL_owner_snapmirror.checkInProgress:: succeeded
    # SUCCESS
    my $cmd = sprintf("priv set -q test; snaptags_test %s ", $op) .
                     sprintf("-v %s ", $opts{volume}) .
                     sprintf("-S %s ", $opts{snapshot}) .
                     'VOPL_owner_snapmirror.checkInProgress:,snapmirror,5';
    NOTE::logcomment("DEBUG: snaptags_test command: $cmd");
    my $response = $opts{node}->server->dblade(cmd=> $cmd);
    if (grep 'SUCCESS', $response) {
        NOTE::logcomment('Succeeded in locking/unlocking snapshot '.
                                                        $opts{'snapshot'});
    } else {
        # There is no reason to stop the checker if we can not lock down the
        # snapshot. Just put a warning saying that we could not lock the
        # snapshot.
        NOTE::logcomment('WARNING: Could not lock the snapshot for checker');
    }
}


=head2 get_check_result

Retrieves the result of a background snapmirror check.  Return value is as
described in check(), above.  If the abort_on_mismatch option is specified
when check() is called, the script will confess if an inconsistent snapmirror
is found by monitor().

=over

=item Arguments

=over

=item None

=back

=item Returns

=over

=item B<Success:> See return value of check() method.

=item B<Failure:> confesses

=back

=item Example

=back

        $sm->check (background=>1);
        { ... }
        $sm->monitor();
        $chk_log = $sm->get_check_result();
        if ( $chk_log ) { ... }

=cut

sub get_check_result {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return $self->{_bg_check_result};
}

# build expected snapmirror check result.
# takes the raw snapmirror check result text and abort flag as inputs.
# if the mirror is ok, returns zero.
# if the mirror is inconsistent, confess if the abort flag is set.
# else return a hash containing the node name and a path to the error data.
sub _get_sm_check_result {
    my $chk_result = shift;
    my $abort_flag = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    if ( $chk_result =~ /NOTE: Found inconsistent mirror/ ) {
        NOTE::logcomment("snapmirror check - Found inconsistent mirror");

        if ($abort_flag) {
            confess "snapmirror check found inconsistency: " . $chk_result;
        }
        else {
            $chk_result =~ /Please find details on node (\S+) in file (\S+)/;
	    my ($node,$path) = ($1,$2);		# due to UI change
	    $path =~ s/\.$//;			# we have to remove last dot
            return { node => $node, path => $path };
        }
    }
    else {
        NOTE::logcomment("snapmirror check - mirror OK");
        return 0;
    }
}

sub _diff_geometry {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $src_status = $self->{_src}->status;
    my $dst_status = $self->{_dst}->status;
    if ( $src_status->{'volume-style'} eq $dst_status->{'volume-style'} ) {
        if ( $src_status->{'volume-style'} =~ /striped/ ) {
            return !( $src_status->{'stripe-count'} == $dst_status->{'stripe-count'}
                && $src_status->{'stripe-width'} == $dst_status->{'stripe-width'}
                && $src_status->{'stripe-format'} eq $dst_status->{'stripe-format'} );
        }
        else {
            return 0;
        }
    }
    else {
        return 1;
    }
}

=head2 abort

Performs a 'snapmirror abort' to abort an active transfer.


Returns:
SUCCESS: 0
FAILURE: confesses

Confessions:
1) if the snapmirror abort command fails for any reason
2) if the transaction was not stopped after the abort
3) if the snapmirror state is not idle after the abort

=over

=item Arguments
Arguments are passed by hash.

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
source-cluster, source-vserver and source-volume options.

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.

B<hard|h> => I<true|false>

Hard causes the checkpoint present to be deleted when the transfer aborts. Can
be issued even when no transfers are active to remove previous checkpoint. The
default is to not use the option.

B<check-only> => I<true|false>

Causes the checker to abort. Default is to use '-check-only false'

B<-Timeout|timeout> => I<seconds for Server to wait for the command to return>

The default is to use the Server timeout, currently 40 seconds.

=item Notes

1)  The ntest parameters SNAPMIRROR_CLUSTER_PATH and SNAPMIRROR_7G_PATH
    override all method arguments causing GNOTE to generate commands using
    either 7g or cluster style options.
2)  The default functionality of the method is to not provide a description
    for the source volume in the command.
3)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.

B<exec_vol> => I<src|dst>

I<Optional>  Cross cluster configuration permits calling abort from Source or Destination node. Default is destination.

=item Example

=back

    # abort taking GNOTE defaults, no source volume definition, destination
    # volume described using cluster/vserver/volume tuple, check-only = false,
    # and using the default Server timeout
    $sm->abort();

    # abort using a source and destination volume path and a Server timeout
    # of 60 seconds
    $sm->abort( S => '//vserver/volume',
                'destination-path' => '//vserver/volume',
                '-Timeout' => 60 );

    # hard abort using a destination volume path only
    $sm->abort( 'destination-path' => 'volume', h => 'true' );

    # abort using cluster/vserver/volume tuple to describe the destination
    # volume and a Server timeout of 60 seconds
    $sm->abort( hard => 'false', '-Timeout' => 60 );

    # abort executed on the source volume node.
    $sm->abort(exec_vol=> 'src');

=cut

sub abort {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my ( $opts, $rtn, $status, $text );

    # extract the user selected execution server
    my $exec_vol = _set_execution_srv(\%args);

    my $capability = $self->status('relationship-capability');

    # build the appropriate command options
    $opts = $self->_adjust_command_options( %args, 'source-needed' => 0 );

    # handle any additional arguments
    $opts->{'check-only'} = ( exists $args{'check-only'} ? $args{'check-only'}
                                                         : 'false' );

    # In case of aborting a checker transfer, as there is no abort
    # for checker, we will be doing a "job stop" on the checker job id.
    if (($capability =~ /8.2 and above/) and
                    ($opts->{'check-only'} =~ /true/i)) {
        if (! $self->{_chk_job_id}) {
            confess "Check job id not found";
        }
        # Stop the job and return in case of checker as there is on
        # abort for checker.
        $rtn = eval {
            $text  = $self->{$exec_vol}->server->job_stop(id =>
                                                       $self->{_chk_job_id});
            1;
        };
        if (!$rtn) {
            confess "job stop failed: $@";
        }
        if (exists $self->{_bg_locked_snapshot}) {
            _snapshot_checker_op(
                     op => 'unlock',
                     snapshot => $self->{_bg_locked_snapshot},
                     volume   => $self->dst->name,
                     node     => $self->{_dst}
            );
            delete $self->{_bg_locked_snapshot};
        }
        return 0;
    }

    # We need to remove the 'check-only' option as that is not 
    # supported in 8.2 or above.
    delete $opts->{'check-only'} if $capability =~ /8.2 and above/;

    # Setting default abort cmd timeout to 120 seconds
    $opts->{'-Timeout'} = 120;

    # Use the user defined timeout value if defined
    if (exists($args{'-Timeout'}) || exists($args{timeout})) {
        $opts->{'-Timeout'} = $args{'-Timeout'} || $args{timeout};
    }

    $rtn = eval {
                $text  = $self->{$exec_vol}->server->snapmirror_abort( %{$opts} );
                1;
            };
    if (!$rtn) {
        confess "snapmirror abort failed: $@";
    }

    if( $capability =~ /8.2 and above/ ) {
	$self->snapmirror_wait( state => '(?i:uninitialized|snapmirrored|\-)', status => '(?i:idle|quiesced)' );  # case insensitive
    } else {
	$self->_check_job( 1, $text, $args{timeout} );

	# Verify (state, status) is (uninitialized or snapmirrored, idle) after
	# the abort
	$rtn = eval {
		    $self->monitor( state => '(uninitialized|snapmirrored)',
				    status => 'idle' );
		    1;
		};
	if ((! $rtn) && ($@ !~ /aborted/i)) {
	    confess $self->description . " Abort failed. Error: $@";
	}
    }

    $status = $self->status;
    if ( $status->{status} !~ /idle|quiesced/i ) {
        confess $self->description . " unexpected (status) after abort: ($status->{status})";
    }

    return 0;
}

=head2 break

Performs a 'snapmirror break' to break a source/destination relationship and
makes destination writable. Verifies that destination is writable after break.

Returns:
SUCCESS: 0
FAILURE: confesses

Confessions:
1) if the snapmirror break command fails for any reason
2) if the destination volume is not writable after break

=over

=item Arguments
Arguments are passed by hash.

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
source-cluster, source-vserver and source-volume options.

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.

B<force|f> => I<true|false>

Forces the relationship to be broken without confirmation. Default is to not
provide a force option.

B<test> => I<1|0>

Run the command in test mode, overrides ntest paramter TRICKY_PARITY. Default: 0

=item Notes

1)  The ntest parameters SNAPMIRROR_CLUSTER_PATH and SNAPMIRROR_7G_PATH
    override all method arguments causing GNOTE to generate commands using
    either 7g or cluster style options.
2)  The default functionality of the method is to not provide a description
    for the source volume in the command.
3)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.
4)  Setting CLI parameter TRICKTY_PARITY will cause this command to run in
    test mode. This can be overridden by setting the test method argument.
    See the Arguments section for details on the test argument.


=item Examples

=back

    # break taking GNOTE defaults, no source volume definition, destination
    # volume described using cluster/vserver/volume tuple and using the
    # default Server timeout
    $sm->break();

    # force a break allowing 90 seconds for the command to complete
    $sm->break( force => 'true', '-Timeout' => 90 );

    # break using a source and destination path
    $sm->break( S => 'volume', 'destination-path' => 'volume', f => 'false' );

=cut

sub break {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( $opts, $test_mode, $rtn );

    #BURT 407766
    # check wafl version.
    # If BR, use test mode.  If >= RR don't. (for tricky parity)
    my $wafl_version = $self->src->Node->get_wafl_version();

    #my $rolling_rock_wafl_id = Filer::Version::WV_ROLLINGROCK_MIN(); #or ...
    if ($wafl_version >= 22272) { #rolling rock min wafl id == 87 << 8 == 22272
        $test_mode = ( exists $args{test} ? delete $args{test} : 0 );
    } else { #before RR, use test mode
        # test mode is needed for tricky parity which can be set explicitly via
        # 'test' parameter or TRICKY_PARITY CLI parameter
        $test_mode = ( exists $args{test} ? delete $args{test}
                                          : $self->{_tricky_parity} );
    }

    # Test mode is required for breaking a vault relationship and
    # making it an RW.
    $test_mode = 1 if $$self{_type} =~ /xdp/i;

    # build the appropriate command options
    $opts = $self->_adjust_command_options( %args, 'source-needed' => 0 );

    # handle any additional arguments
    if (exists($args{'-Timeout'}) || exists($args{timeout})) {
        $opts->{'-Timeout'} = $args{'-Timeout'} || $args{timeout};
    }

    $self->{_dst}->server->set( privilege => 'test', confirmations => 'off' )
      if $test_mode;

    $rtn = eval { $self->{_dst}->server->snapmirror_break( %{$opts} ); 1; };
    confess "snapmirror break failed: $@" if (! $rtn);

    $self->{_dst}->server->set( privilege => 'diag', confirmations => 'off' )
      if $test_mode;

    # Show that the volume is RW after the break.
    # sleep a while to allow DST vol to become RW (burt328012)
    Tharn::snooze(10);
    my $status = $self->dst->status();

    # debug statement until burt328012 gets fixed
    NOTE::logcomment("status = $status->{'type'} ");
    NOTE::logcomment( Dumper( \$status ) );
    if ( $status->{'type'} eq "RW" ) {
        NOTE::logcomment( "snapmirror break succeeded: " . $self->dst->name . " is writable now" );
    }
    else {
        confess "destination" . $self->dst->name . " is not RW after break";
    }

    return 0;
}

=head2 restore

Performs a 'snapmirror restore'.

Setting CLI parameter TRICKTY_PARITY will cause this command to run in test
mode.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<src> => I<NOTE::Volume spec>

Supply this parameter is the restore is to a new src volume.  The parameter can
be a NOTE::Volume object or spec to allow a NOTE::Volume object to be created.

=item B<background> => I<1|0>

Boolean flag to indicate background execution. Default: 0

=item B<timeout> => I<number>

Maximum time allowed for commands to complete, in seconds.

=item B<throttle> => I<integer>

Max transfer rate in KB / sec.

=item B<test> => I<1|0>

Issue command in test mode. Default: 0

=back

=item Returns

=over

=item B<Success:> Returns the existing dst volume object.  If 'src' parameter
is specified, returns the old 'src' volume object.

=item B<Failure:> confesses

=back

=item Examples

=back

    # Restore to existing dst vol
    $sm->restore();

    # Restore to a new dst volume
    $sm->restore(src => $new_dst_vol);

=cut

sub restore {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $old_src  = $self->{_src};
    my $old_desc = $self->description;
    if ( exists $args{src} ) {
        if ( ref( $args{src} ) =~ /NOTE::.*Volume/ ) {
            $self->{_src} = delete $args{src};
        }
        else {
            $self->{_src} = NOTE::Volume->create( delete $args{src} );
        }
        NOTE::logcomment( "$old_desc, restore to new src: " . $self->{_src}->description );
    }

    my %opts = ();
    $opts{'destination-volume'}  = $self->dst->name;
    $opts{'destination-vserver'} = $self->dst->Vserver->name;
    $opts{'destination-cluster'} = $self->_cluster_name("dst");
    $opts{'source-volume'}       = $self->{_src}->name;
    $opts{'source-vserver'}      = $self->{_src}->Vserver->name;
    $opts{'source-cluster'}      = $self->_cluster_name("src");

    $self->_hash_xfer(
        label => [ "force", "throttle" ],
        from  => \%args,
        to    => \%opts
    );
    $opts{force} = "true" unless exists $opts{force};

    my $background = NOTE::GX::background( \%args );

    $opts{'destination-path'} = $args{'destination-path'}
      if exists $args{'destination-path'};
    $opts{'source-path'} = $args{'source-path'} if exists $args{'source-path'};

    # test mode is needed for tricky parity which can be set explicitly via 'test' parameter or TRICKY_PARITY CLI parameter
    my $test_mode = ( exists $args{test} ? delete $args{test} : $self->{_tricky_parity} );

    $self->{_dst}->server->set( privilege => "test", confirmations => "off" )
      if $test_mode;

    NOTE::logcomment( "Snapmirror Restore: " . $self->description );
    my $text;
    eval { $text = $self->{_dst}->server->snapmirror_restore(%opts); };
    confess "$@" if $@;

    $self->{_dst}->server->set( privilege => "diag", confirmations => "off" )
      if $test_mode;

    $self->_check_job( $background, $text, $args{timeout} );

    return $old_src;
}

=head2 resync

Performs a 'snapmirror resync' on the source and destination associated with
the SnapMirror object.


Returns:
SUCCESS: its SnapMirror object
FAILURE: confesses

Confessions:
1) if the snapmirror resync command fails for any reason

=over

=item Arguments
Arguments are passed by hash.

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
source-cluster, source-vserver and source-volume options.

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.

B<source-snapshot> => I<snapshot name (string)>

The name of a specific snapshot to be used for the resync.
The default is to not provide a source snapshot

B<force|f> => I<true|false>

Forces the relationship to be broken without confirmation. Default is to not
provide a force option.

B<throttle|k> => I<integer|'unlimited'>

Max transfer rate in KB / secs. Default: no option provided = unlimited.

[ B<foreground|w> => I<true|false>
  | B<background> => I<1|0> ]

Whether the filer should wait for the command to complete before returning
(foreground|w) or start the command in the background. Default see notes.

B<timeout> => I<seconds to wait for resync to complete>

Used when the resync is done in the foreground so Server won't time it out.
The default is the GNOTE default, currently 3600 seconds

B<-Timeout> => I<seconds for Server to wait for the command to return>

Used when the resync is done in the background but the command might take
longer than 40 seconds to return. The default is to use the Server timeout,
currently 40 seconds

=item Notes

1)  The ntest parameters SNAPMIRROR_CLUSTER_PATH and SNAPMIRROR_7G_PATH
    override all method arguments causing GNOTE to generate commands using
    either 7g or cluster style options.
2)  The default functionality of the method is to describe the source volume
    in terms of the options 'source-cluster', 'source-vserver' and
    'source-volume'. If either 'source-path' or 'S' argument is provided then
    the source volume will be described as a path using the given option.
    * Argument 'source-path' produces a command using the '-source-path' option
    * Argument 'S' produces a command using the '-S' option
3)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.
4)  The default action if none of foreground, w, or background arguments are
    provided is to run the command in the background and monitor the command
    until it completes. Effectively making it look to the caller as if the
    command was run in the foreground.
5)

=item Examples

=back

    # resync taking GNOTE defaults, the source and destination volume will
    # be described using cluster/vserver/volume tuple and using the default
    # Server timeout, then wait for the resync to complete
    $sm->resync();
    $sm->monitor();

    # resync the volumes in the background and wait for the resync to complete
    $sm->resync( foreground => 'false' );
    $sm->monitor();

    # resync the volumes in the foreground giving the resync up to 10 minutes
    # to complete
    $sm->resync( w => 'true', '-Timeout' => 600 );

    # resync the volumes in the background using a source and destination
    # path and allowing the command to take up to 60 seconds to return
    $sm->resync( S => 'volume', 'destination-path' => 'volume',
                 w => 'false', '-Timeout' => 60 );
    $sm->monitor();

    # force the resync to happen in the background using 1024k of bandwidth
    # and wait for the resync to complete
    $sm->resync( force => 'true', k => 1024, background => 1 );
    $sm->monitor();

=cut

sub resync {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $opts, $background, $text, $rtn );

    my $skip_ss = defined $args{'skip_ss'} ? delete $args{'skip_ss'} : 0;

    # build the appropriate command options
    $opts = $self->_adjust_command_options( %args, 'source-needed' => 1 );

    if ( !defined $args{'source-snapshot'}
             && ($self->{_type} =~ /XDP/i) && !$skip_ss) {
	$$opts{'source-snapshot'} = $self->_xdp_create_ss();
	confess "Snapshot ". $$opts{'source-snapshot'} ." was not created on volume " . $self->src->name unless $self->src->snapshot_exists($$opts{'source-snapshot'});
	NOTE::logcomment( "XPD Source Snapshot: ".$$opts{'source-snapshot'} );
    }


    # deal with the conflict of the background, foreground and w arguments.
    # note that foreground and w are no longer valid so should never be
    # copied to opts.
    $background = NOTE::GX::background( \%args );

    # handle any additional arguments
    $$opts{'type'} = $self->{_type};
#    $$opts{'policy'} = $self->{_policy} if $self->{_policy};
    foreach( qw/ source-snapshot type policy /) {
	$$opts{$_} = $args{$_} if exists $args{$_};
    }

    NOTE::logcomment( 'Snapmirror Resync: ' . $self->description . Dumper($opts) );
    $rtn = eval {
                $text = $self->{_dst}->server->snapmirror_resync( %{$opts} );
                1;
            };
    if( $@ =~ /invalid argument "-type"/si ) {
	delete $$opts{'type'};
	$rtn = eval{ $text = $self->{_dst}->server->snapmirror_resync( %{$opts} );1;};
    }

    confess "snapmirror resync failed: $@" if (! $rtn);

    $self->_check_job( $background, $text, $args{timeout} );

    return $self;
}

=head2 flip_resync

Performs flip/resync...

Swaps SRC and DST and performs a 'snapmirror resync'. See B<resync> call for all arguments
and usage.

NOTE: B<No error checks, abort, quiesce, break and delete operations are performed!!!> See BURT 620725
for details.

=cut

sub flip_resync {
    my ( $self, %args ) = @_;
    my $tmp = $self->{_src};
    $self->{_src} = $self->{_dst};
    $self->{_dst} = $tmp;
    return $self->resync(%args);
}

=head2 delete

Deletes the snapmirror relationship.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<source-cluster> => I<string>

Source cluster identity/name.

=item B<source-vserver> => I<string>

Source vserver name.

=item B<source-volume> => I<string>

Source volume name.

=item B<destination-cluster> => I<string>

Destination cluster identity/name.

=item B<destination-vserver> => I<string>

Destination vserver name.

=item B<destination-volume> => I<string>

Destination volume name.

=item B<exec_vol> => I<src|dst>

Cross cluster configuration permits calling delete from Source or Destination
node. Default is destination.

B<timeout> => I<seconds to wait for delete to complete>

Used when the delete is done in the foreground so Server won't time it out.
The default is the GNOTE default, currently 3600 seconds

=back

=item Returns

=over

=item B<Success:> $self

=item B<Failure:> confesses

=back

=item Example

=back

     $sm->delete( 'source-cluster'      => $src_cluster->name,
                  'source-vserver'      => $src_vserver->name,
                  'source-volume'       => $src_volume->name,
                  'destination-vserver' => $dst_vserver->name,
                  'destination-cluster' => $dst_cluster->name,
                  'destination-volume'  => $dst_volume->name,
                  'timeout'             => $timeout,
                 );
    $sm->delete('exec_vol'=> 'src'); # Call delete from source node.

=cut

sub delete {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my %opts = ();
    my $exec_vol = _set_execution_srv(\%args);

    if( defined $args{'source-cluster'} ){
        $opts{'source-cluster'} = delete $args{'source-cluster'};
    }
    if( defined $args{'source-vserver'} ){
        $opts{'source-vserver'} = delete $args{'source-vserver'};
    }
    if( defined $args{'source-volume'} ){
        $opts{'source-volume'} = delete $args{'source-volume'};
    }

    $opts{'destination-cluster'} =
      ( exists $args{'destination-cluster'} )
      ? $args{'destination-cluster'}
      : $self->_cluster_name("dst");
    $opts{'destination-vserver'} =
      ( exists $args{'destination-vserver'} )
      ? $args{'destination-vserver'}
      : $self->dst->Vserver->name;
    $opts{'destination-volume'} =
      ( exists $args{'destination-volume'} )
      ? $args{'destination-volume'}
      : $self->dst->name;
    $opts{'-Timeout'} = $args{'timeout'}
      if exists $args{'timeout'};

    ## Workaround for burt 376536
    $opts{'-Verification'} = delete $args{'-Verification'} || 0;

    my $result = eval { $self->{$exec_vol}->server->snapmirror_delete(%opts); 1};
    if (!$result) {
        confess $@;
    }

    return $self;
}

=head2 release

Releases the snapmirror relationship, always executed on the source.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<source-vserver> => I<string>

Source vserver name.

=item B<source-volume> => I<string>

Source volume name.

=item B<destination-vserver> => I<string>

Destination vserver name.

=item B<destination-volume> => I<string>

Destination volume name.

B<timeout> => I<seconds to wait for delete to complete>

Used when the delete is done in the foreground so Server won't time it out.
The default is the GNOTE default, currently 3600 seconds

=back

=item Returns

=over

=item B<Success:> $self

=item B<Failure:> confesses

=back

=item Example

=back

    # release the snapmirror relationship overriding the default timeout
    $sm->release( 'source-vserver'      => $src_vserver->name,
                  'source-volume'       => $src_volume->name,
                  'destination-vserver' => $dst_vserver->name,
                  'destination-volume'  => $dst_volume->name,
                  'timeout'             => $timeout,
                 );

    # release the relationship taking all the defaults
    $sm->release();

=cut


sub release {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $text, $timeout, $job_id, $Job, $status );
    my %opts = ();

    # note: the source-cluster and destination-cluster command options where
    # at one time supported by release. then they were not and using them
    # caused the command to fail. due to this the options are no longer
    # passed to the command line. at the time of these changes not providing
    # the options worked for versions of ontap that allowed the options and
    # versions that did not. this was tested using cluster admin for both
    # inter and intra cluster configurations. testing using vserver admin is
    # irrelevant because vserver admin has no knowledge or access to the
    # cluster information.

    if( defined $args{'source-vserver'} ){
        $opts{'source-vserver'} = delete $args{'source-vserver'};
    }
    if( defined $args{'source-volume'} ){
        $opts{'source-volume'} = delete $args{'source-volume'};
    }

    $opts{'destination-vserver'} =
      ( exists $args{'destination-vserver'} )
      ? $args{'destination-vserver'}
      : $self->dst->Vserver->name;
    $opts{'destination-volume'} =
      ( exists $args{'destination-volume'} )
      ? $args{'destination-volume'}
      : $self->dst->name;
    $opts{'-Timeout'} = $args{'timeout'}
      if exists $args{'timeout'};

    ## Workaround for burt 376536
    $opts{'-Verification'} = delete $args{'-Verification'} || 0;

    $text = $self->src->server->snapmirror_release(%opts);

    # check the job for the status of the command
    $job_id = NOTE::GX::Job->get_id($text);

    $Job = NOTE::GX::Job->new( Cluster => $self->src->Cluster, id => $job_id );

    # a few branches has an asynchronous version of release, give them a
    # few seconds to complete. the final command will be synchronous so
    # once all branches have collapsed this can be removed.
    Tharn::sleep(5);
    $status = $Job->status();

    # verify state, status, completion (anything else?)
    if ( ( lc($status->{state}) ne 'success' ) || ( $status->{code} != 0 ) ||
         ( $status->{completion} !~ /SnapMirror Release Succeeded/i ) ) {
        confess "$status->{description} failed:\n"
                . "State:              $status->{state}\n"
                . "Status Code:        $status->{code}\n"
                . "Completion String:  $status->{completion}";
    }

    return $self;
}

=head2 list_destinations

Returns the raw output from the list-destination commnd for the relationship.
The command is issued using the destination-vserver and destination-volume
options. This command is always executed on the source.

Note: The snapmirror list-destinations command is not supported by Server.pm
      which is why the raw output of the command is returned instead of an
      array of hash references.

=over

=item Arguments

I<Optional Arguments>

=over

=item B<timeout> => I<integer>

Time out in secs. Default: 300

=back

=item Returns

=over

=item B<Success:> A scalar containing the raw ouput of the command.

=item B<Failure:> Confesses

=back

=item Examples

=back

    # Verify the relationship exists on the source
    $sm->list_destinations();

    # Get the source view of the status of the relationship
    # overriding the default timeout
    my $output = $sm->list_destinations( timeout => 900 );

    # Get the source view of the status of the relationship
    my $output = $sm->list_destinations();

=cut

sub list_destinations {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $timeout, $cmd, $output, $result );

    NOTE::logcomment( 'Snapmirror list-destinations: ' . $self->description );

    $timeout = delete $args{timeout} || 300;

    # use the destination-vserver and destination-volume which is the
    # standard for gnote
    $cmd = 'snapmirror list-destinations';
    $cmd .= ' -destination-volume ' . $self->dst->name;
    $cmd .= ' -destination-vserver ' . $self->dst->Vserver->name;

    # this should be a call to server->snapmirror_list_destinations but
    # it doesn't seem to be supported by Server.pm so replacing it with
    # the raw command sent via cli - the down side is the raw output is
    # returned
    $result = eval {
            $output = $self->{_src}->server->cli( cmd        => $cmd,
                                                  '-Timeout' => $timeout );
            1;
        };

    if (!$result) {
        confess $@;
    }

    # if someone knows of a standard method that will parse the raw output
    # and return its hash it could be used here to make this method much
    # more useful
    return $output;
}

=head2 promote

Performs a 'snapmirror promote'.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<timeout> => I<number>

Maximum time allowed for commands to complete, in seconds.

=item B<force> => I<String>

Set the force value to true|false. Default:false

=back

=item Returns

=over

=item B<Success:> Returns the promoted volume object.

=item B<Failure:> confesses

=back

=item Examples

=back

        $sm->promote();

        $sm->promote('force' => 'true');

=cut

sub promote {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Promoting dst vol " . $self->dst->description );

    my %opts = ();
    $opts{'destination-volume'}  = $self->dst->name;
    $opts{'destination-vserver'} = $self->dst->Vserver->name;
    $opts{'destination-cluster'} = $self->_cluster_name("dst");
    $opts{'force'} = exists $args{'force'}  ? $args{'force'} : 'false';
    $opts{'-Timeout'}            = $args{'timeout'} if exists $args{'timeout'};
    $opts{'source-volume'}       = $self->src->name;
    $opts{'source-vserver'}      = $self->src->Vserver->name;
    $opts{'source-cluster'}      = $self->_cluster_name("src");

    eval { $self->{_dst}->server->snapmirror_promote(%opts); };
    confess "$@" if $@;

    my $new_src = $self->{_dst};

    NOTE::logcomment( $new_src->description . " successfully promoted" );

    # Refresh dst's junction and share and mountpoints from src
    $self->dst->_replace_mountpoint( $self->src );

    # Mirror and src volume nolonger exists after the promote
    $self->{_destroyed} = 1;
    $self->src->SPARE;

    return $new_src;
}

=head2 modify

Modify the snapmirror relationship by altering the SM relationship or its
properties.  If Changing the destination or source please refer to the SM man
page for restrictions of use
http://web.netapp.com/engineering/design-depot/edm/docs/specs.cgi?proj=Paloma&doc=Paloma_UI&targ=Paloma_CLI&ext=txt

At least one of the following parameters is required.  Furthermore, it is
invalid to call this method specifying both a new source (src) and a new
destination (dst) volume.  It is valid only to specify neither or to
specify one or the other, not both.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<src> => I<NOTE::Volume object>

It is invalid to pass both an src argument and a dst argument.

=item B<dst> => I<NOTE::Volume object>

It is invalid to pass both an src argument and a dst argument.

=item B<schedule> => I<string>

Name of the schedule on which the relationship operates

=item B<tries> => I<number>

The maximum number of times to attempt a transfer or restart a transfer
for the relationship at each scheduled time. range [0..1000000000].

=item B<throttle> => I<integer>

Maximum transfer rate in kilobytes per second (an integer)

=back

=item Returns

=over

=item B<Success:> Nothing specific.  (I.e., Do not use returned value.)

=item B<Failure:> confesses

=back

=item Examples

=back

    my %opts = (
        src      => $src_obj,
        throttle => 500,
               );
    $sm->modify(%opts);

    $sm->modify(src => $new_src_obj);
    $sm->modify(dst => $new_dst_obj, tries => 3);
    $sm->modify(tries => 400);

=cut

sub modify {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    my %opts = ();
    my $change_type;

    if ( $args{dst} && $args{src} ) {
        confess " cannot modify both source and destination in modify";
    }

    if ( ref( $args{src} ) =~ /NOTE::.*Volume/ ) {
        $opts{'source-volume'}  = $args{src}->name;
        $opts{'source-vserver'} = $args{src}->Vserver->name;
        $opts{'source-cluster'} = $args{src}->Cluster->name;
        $change_type            = 'src';
    }
    elsif ( $args{src} ) {
        confess "Invalid src specification: " . $args{src};
    }

    if ( ref( $args{dst} ) =~ /NOTE::.*Volume/ ) {
        $opts{'destination-volume'}  = $args{dst}->name;
        $opts{'destination-vserver'} = $args{dst}->Vserver->name;
        $opts{'destination-cluster'} = $args{dst}->Cluster->name;
        $change_type                 = 'dst';
    }
    elsif ( $args{dst} ) {
        confess "Invalid dst specification: " . $args{dst};
    }

    # if either src or dst is not passed get values from self
    unless ( $args{dst} ) {
        $opts{'destination-volume'}  = $self->{_dst}->name;
        $opts{'destination-vserver'} = $self->{_dst}->Vserver->name;
        $opts{'destination-cluster'} = $self->{_dst}->Cluster->name;
    }
    unless ( $args{src} ) {
        $opts{'source-volume'}  = $self->{_src}->name;
        $opts{'source-vserver'} = $self->{_src}->Vserver->name;
        $opts{'source-cluster'} = $self->{_src}->Cluster->name;
    }

    if ( exists $args{'schedule'} ) {
        $opts{'schedule'} = $args{'schedule'};
    }

    if ( exists $args{'tries'} ) {
        $opts{'tries'} = $args{'tries'};
    }

    if ( exists $args{'throttle'} ) {
        $opts{'throttle'} = $args{'throttle'};
    }

	if ( exists $args{'policy'} ) {
        $opts{'policy'} = $args{'policy'};
    }

    # if dst is passed execute on dst and set dst passed value
    if ( $change_type =~ /dst/ ) {
        eval {
            $self->{_src}->server->snapmirror_modify(%opts);
            $self->{_dst} = $args{dst};
        };
        confess "$@" if $@;
    }
    else {    # execute on dst and set the src if src is passed.
        eval {
            $self->{_dst}->server->snapmirror_modify(%opts);
            $self->{_src} = $args{src} if exists $args{src};
        };
        confess "$@" if $@;
    }

}

=head2 schedule

Get, Clear or Restore the mirror schedule that was set during mirror creation
or with the 'modify' method.

The 'clear' option also records the schedule locally for later 'restore'.  The
'clear stack' is only one deep.

=over

=item Arguments

=over

=item B<cmd> => I<get|clear|restore>

Action:

=over

=item I<get> - Returns the current schedule

=item I<clear> - Remove schedule from filer and record it in this object.
Returns the schedule that was cleared

=item I<restore> - Restore the previous 'clear' schedule. Returns the restore
schedule

Default: get

=back

=back

=item Returns

=over

=item B<Success:> A SnapMirror schedule.  See Action results above.

=item B<Failure:> confesses

=back

=item Examples

=back

    my $sm = NOTE::SnapMirror->create(src => $src, dst => $dst);

    $sm->modify(schedule => "5min");    # set 5 minute schedule

    $sm->schedule(cmd => "clear");      # remove the schedule

    $sm->schedule(cmd => "restore");    # restore the schedule

=cut

sub schedule {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{cmd} = "get" unless exists $args{cmd};

    my $no_sch   = '""';
    my $curr_sch = $self->status("schedule");
    if ( $args{cmd} =~ /get/ ) {

        # Just return the current schedule
        NOTE::logcomment( $self->description . ": " . ( $curr_sch =~ $no_sch ? "No schedule" : "current schedule is $curr_sch" ) );
        return $curr_sch;
    }
    elsif ( $args{cmd} =~ /clear/ ) {
        if ( $curr_sch =~ $no_sch ) {
            confess $self->description . ": no schedule to clear";
        }
        else {

            # Clears the schedule and return the cleared schedule
            $self->{_save_schedule} = $curr_sch;
            $self->modify( schedule => $no_sch );
            NOTE::logcomment( $self->description . ": $curr_sch schedule removed and saved" );
        }
    }
    elsif ( $args{cmd} =~ /restore/ ) {

        # Sets the saved schedule and return it
        confess $self->description . ": No saved schedule to restore"
          unless exists $self->{_save_schedule};
        $curr_sch = delete $self->{_save_schedule};
        $self->modify( schedule => $curr_sch );
        NOTE::logcomment( $self->description . ": $curr_sch schedule restored" );
    }
    else {
        confess $self->description . ": Invalid cmd parameter " . $args{cmd};
    }
    return $curr_sch;
}

# Waits for job to start running or completion based on background flag

sub _check_job {
    my ( $self, $background, $text, $timeout, $snapshot ) = @_;
    my $rtn;
    my $SNflavor;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Assign job id if one is present. SN flavor does not have one
    # Confess if get_id fails and error is not unable to determine job id
    $rtn = eval {
        NOTE::logcomment("_check_job job id text: $text");
        $self->{_job_id} = $self->{_dst}->Node->job_id($text);
        1;
    };
    if ((! $rtn) && ($@ !~ /Unable to determine job id/i)) {
        confess "Unknown error: $@";
    } elsif (! $rtn ) {
        NOTE::logcomment('DEBUG - _check_job setting SN flavor');
        $SNflavor = 1; #Failed to get a job id but not an unknown error
    }

    #SN flavor has no job ids - for some but not all jobs
    if ($background) {
        if (! $SNflavor) {
            $self->_job_active();
        }
        return;
    }

    $timeout = $self->{_timeout} unless $timeout;
    #SN flavor uses snapmirror_wait RR flavor jobs use monitor
    if ($SNflavor) {
        NOTE::logcomment('DEBUG - using SN flavor');
        # If snapshot information was passed, use it
        if (defined $snapshot) {
            NOTE::logcomment("DEBUG - snapshot information pass $snapshot");
	    $self->snapmirror_wait( 'status' => 'Idle',
				'state' => 'Snapmirrored',
				'exported-snapshot-timestamp' => '\d+',
				'exported-snapshot' => $snapshot,
				timeout => $timeout );
        } else {
            NOTE::logcomment('DEBUG - no snapshot information');
	    $self->snapmirror_wait( 'status' => 'Idle',
				'state' => 'Snapmirrored',
				timeout => $timeout );
        }
    } else {
        NOTE::logcomment('DEBUG - using RR flavor');
        $self->monitor( timeout => $timeout );
    }
}

=head2 progress

Returns the progress column from current (if any) snapmirror job.

=over

=item Arguments

=over

=item I<None>

=back

=item Returns

=over

=item B<Success:> String description of joba(if any) progress.

=item B<Failure:> confesses

=back

=item Examples

    my $progress = $sm->progress;

=back

=cut

sub progress {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "No snapmirror job id found"
      unless $self->{_job_id};

    my $progress;
    eval { $progress = $self->{_dst}->Node->job_wait( id => $self->{_job_id}, state => $self->_sm_job_states(), column => "progress", timeout => 5 ); };
    confess "snapmirror job " . $self->{_job_id} . " failed. Error: $@"
      if $@ && $@ !~ /no entries/;
    return $progress;
}

# Determine if a job is currently running or not

sub _job_active {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "No snapmirror job id found"
      unless $self->{_job_id};

    my $status;
    eval { $status = $self->{_dst}->Node->job_wait( id => $self->{_job_id}, state => $self->_sm_job_states(), column => "all", timeout => 30 ); };
    confess "snapmirror job " . $self->{_job_id} . " failed. Error: $@"
      if $@ && $@ !~ /no entries/;
    $self->{_progress} = $status->{'progress'};
    return ( $status->{'state'} =~ /running/i );
}

# Returns a regular expression used to determine list of snapmirror job states of interest

sub _sm_job_states {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return "(dead|dormant|error|quit|running|success)";
}

# Transfer hash entries based on supplied list of keys

sub _hash_xfer {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    foreach ( @{ $args{label} } ) {
        $args{to}->{$_} = delete $args{from}->{$_} if exists $args{from}->{$_};
    }
}

# Set the execution node to either user choice or default dst.

sub _set_execution_srv {

    my ($args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $exec_vol;

    # In XCDR support for executing show from either source node or
    # destination node is supported functionality RR.0.

    if ($args->{exec_vol} && $args->{exec_vol} =~ /dst|src/i)  {
        $exec_vol = '_' . $args->{exec_vol};
        delete $args->{exec_vol};
        NOTE::logcomment("User request execution on $exec_vol volume.");
    } else {
        $exec_vol = '_dst';
    }
    return $exec_vol;
}

=head2 get_daemon_info

Retrive the information about a daemon job.

The routine returns a hash ref which contains the following info:
filer, node, state, starttime, category, endtime, id, code, queuetime,
dropdeadtime, uuid progress, name, description, completion, schedule,
jobtype, priority, restarted, affinity

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<timeout> => I<number>
Timeout for Server.pm to wait for filer output.

=back

=item Returns

=over

=item B<Success:> hashref

=item B<Failure:> confesses

=back

=item Examples

    my $info = $sm->get_daemon_info();

    my $info = $sm->get_daemon_info(timeout => 30);
    $node_name = $info->{name};
    $daemon_state = $info->{state};

=back

=cut

sub get_daemon_info {
   my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   my ($msid) = $self->dst->msid();

   my @job_list = $self->dst->Cluster->job_list(%args);
   foreach my $job (@job_list) {
     if ( $job->{'name'} =~ /$msid/ ) {
          return $job;
     }
   }
}

=head2 Job

Returns the GX::Job object associated with this Snapmirror operation

=over

=item Arguments

=over

=item B<None>

=back

=item Returns

=over

=item B<Success:> GX::Job object

=item B<Failure:> confesses

=back

=item Example

=back

    my $job = $sm->Job();

=cut

sub Job{
 my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 confess " no snapmirror job available "
   unless  $self->{_job_id} ;
 return NOTE::GX::Job->new( Cluster => $self->dst->Cluster, id => $self->{_job_id});
}

=head2 lun_compare

Compare luns between source and destination and return reference to mismatches if
found. Returns undef if no mismatches found.

=over

=item Arguments

B<skip> => I<array> I<Optional>

Array of fields to ignore during compare. For example you may skip lun state if it's not important.
Defaults are: [ 'volume', 'lun-path', 'vserver' ] since they can be different.

B<mismatch> => I<array> I<Optional>

Array of required desimilar fields. Method will return entry unless they are desimilar. For example lun's
serial number must be different for same lun on source and destination. Default faules are: [ 'serial', 'uuid' ]

=item Example of fields

 {
   'volume' => 'vl03752_dead_senators',
   'serial' => 'idWqJ4VcFQHb',
   'vserver' => 'vs0',
   'lun-path' => '/vol/vl03752_dead_senators/lun_windows',
   'uuid' => '23f096cf-36fb-11df-9e85-123478563412',
   'space-resv' => '-',
   'state' => 'online',
   'size' => '49351680',
   'lun' => 'lun_windows',
   'ostype' => 'windows'
 }

=item Examples

 my $lun_check = $sm->lun_compare( skip => [ 'ostype', 'state' ], mismatch => ['serial', 'uuid', 'volume', 'lun-path'] );
 if( $lun_check ) {
    die "Unexpected dissimilarity between source and destination:" . Dumper( $lun_check );
 }

example of output:

 {
  'missing' => {
                 'src' => [
                            'lun_windows_hey'
                          ]
               },
  'lun_windows_hey' => {
                         'mismatch' => {
                                 'state' => {
                                              'dst' => 'online_hey',
                                              'src' => undef
                                            },
                                 'serial' => {
                                               'dst' => 'idWqJ4VcFQHb_hey',
                                               'src' => undef
                                             },
                                 'volume' => {
                                               'dst' => 'vl03752_dead_senators_hey',
                                               'src' => undef
                                             },
                                 'vserver' => {
                                                'dst' => 'vs0_hey',
                                                'src' => undef
                                              },
                                 'lun-path' => {
                                                 'dst' => '/vol/vl03752_dead_senators/lun_windows_hey',
                                                 'src' => undef
                                               },
                                 'lun' => {
                                            'dst' => 'lun_windows_hey',
                                            'src' => undef
                                          },
                                 'uuid' => {
                                             'dst' => '23f096cf-36fb-11df-9e85-123478563412_hey',
                                             'src' => undef
                                           },
                                 'space-resv' => {
                                                   'dst' => '-_hey',
                                                   'src' => undef
                                                 },
                                 'size' => {
                                             'dst' => '49351680_hey',
                                             'src' => undef
                                           },
                                 'ostype' => {
                                               'dst' => 'windows',
                                               'src' => undef
                                             }
                               }
               },
 }

=back

=cut

sub lun_compare {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    #adding 'isMapped' and 'read-only' in $args{skip} since we don't care about them
    $args{skip}     ||= [ 'volume', 'path', 'vserver', 'space-resv', 'rawName', 'mount', 'fsType','isMapped','read-only' ]; # default values for arguments
    $args{mismatch} ||= [ 'serial', 'uuid', 'vdiskId' ];                                    # since everything is optional

    my $rslt = undef;           # return mismatches or falsematches or nothing

    my $slun = $self->src->lun_show();
    my $dlun = $self->dst->lun_show();

    # Due to an UI change the output of "lun show" have a parameter name
    # changed. The parameter name was "alignmnet", now chnaged to "alignment"
    # Check the output of lun_show to see whether the current build has
    # parameter "aligned", and if so, replace the "skip" and "mismatch"
    # arrays to have the updated name.
    if(defined $slun){
        if(defined $slun->[0]->{alignment}){
            foreach ( @{$args{skip}} ) {
                s/^\s*alignmnet\s*$/alignment/;
            }
            foreach ( @{$args{mismatch}} ) {
                s/^\s*alignmnet\s*$/alignment/;
            }
        }
    }

    my %srcs = map { $_->{'lun'} => $_ } @{$slun};  # luns names from source
    my %dsts = map { $_->{'lun'} => $_ } @{$dlun};  # luns names from dest
    my %nams = map { $_ => 1 } ( keys %srcs, keys %dsts );             # all unique lun names from source and destination combined

    foreach my $lun ( keys %nams ) {
        # check missing luns
        unless( $srcs{$lun} && $dsts{$lun} ) {
            my $missing = $srcs{$lun} ? 'dst' : 'src';
            push @{ $rslt->{missing}{$missing} }, $lun;
        }

        # field by field compare
        my %flds = map { ( $_, $_ ) } keys %{ $srcs{$lun} }, keys %{ $dsts{$lun} }; # get all unique field names
        delete $flds{$_} foreach( @{$args{skip}} );                 # remove what we must ignore
        delete $flds{$_} foreach( @{$args{mismatch}} );                 # fields which must be mismatched (serial numbers, uuids etc)

        foreach( keys %flds ) {             # compare what must be matched
            if( $srcs{$lun}->{$_} ne $dsts{$lun}->{$_} ) {              # store mismatches to be printed
                $rslt->{$lun}{mismatch}{$_}{src} = $srcs{$lun}->{$_};    # later
                $rslt->{$lun}{mismatch}{$_}{dst} = $dsts{$lun}->{$_};
            }
        }

        foreach( @{$args{mismatch}} ) {     # compare what surely must mismatch
            if( $srcs{$lun}->{$_} eq $dsts{$lun}->{$_} ) {              # store false matches to be printed
                $rslt->{$lun}{falsematch}{$_}{src} = $srcs{$lun}->{$_}; # later
                $rslt->{$lun}{falsematch}{$_}{dst} = $dsts{$lun}->{$_};
            }
        }
    }
    $Data::Dumper::Maxdepth = 10;                         # make Dumper happy.
    NOTE::logcomment( 'WARNING! Lun mismatches found:', Dumper($rslt) ) if $rslt;   # dump mismatches/false matches into log if they exist
    return $rslt;                                   # but dont fail? Hm. this needs to be reworked.
}


# this method provides support for the testing of the continuum project
# (434894). it adjusts the arguments passed to a snapmirror method to reflect
# the GNOTE mode, 7g style command options, c-mode style command options, or
# use the arguments passed to the method directly. it also provides support
# for the following command line options, S, source-path, destination-path, k,
# f and h.

sub _adjust_command_options {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $util, $mode, $path, $opts );

    # determine the mode of the snapmirror commands and the path description
    $util = NOTE::SnapMirrorUtil->new();
    if (exists($util->{'_SNAPMIRROR_7G_PATH'})) {
        $mode = '7g';
        $path = $util->{'_SNAPMIRROR_7G_PATH'};
    } elsif (exists($util->{'_SNAPMIRROR_CLUSTER_PATH'})) {
        $mode = 'cluster';
        $path = $util->{'_SNAPMIRROR_CLUSTER_PATH'};
    } else {
        $mode = 'disabled';
    }

    # SNAPMIRROR_7G_PATH or SNAPMIRROR_CLUSTER_PATH override method
    # arguments and are mutualy exclusive
    if ($mode eq '7g') {
        # ignore snapmirror arguments and set up a path definition with
        # 7g style options
        # certain commands do not want/need source volume descriptions so
        # unless specified don't give one
        if ($args{'source-needed'} == 1) {
            $opts->{'S'} = $self->{_src}->_build_volume_path(
                                                         description => $path );
        }
        $opts->{'destination-path'} = $self->{_dst}->_build_volume_path(
                                                         description => $path );

        # if any other options are provided translate them to 7g style
        # f/force options
        if (exists($args{'f'})) {
            $opts->{'f'} = $args{'f'};
        } elsif (exists($args{'force'})) {
            $opts->{'f'} = $args{'force'};
        }
        # h/hard options
        if (exists($args{'h'})) {
            $opts->{'h'} = $args{'h'};
        } elsif (exists($args{'hard'})) {
            $opts->{'h'} = $args{'hard'};
        }
        # k/throttle options
        if (exists($args{'k'})) {
            $opts->{'k'} = $args{'k'};
        } elsif (exists($args{'throttle'})) {
            $opts->{'k'} = $args{'throttle'};
        }
    } elsif ($mode eq 'cluster') {
        # ignore arguments and set up a path definition with c-mode options
        $opts->{'source-path'} = $self->{_src}->_build_volume_path(
                                                         description => $path );
        $opts->{'destination-path'} = $self->{_dst}->_build_volume_path(
                                                         description => $path );

        # if any other options are provided translate them to c-mode style
        # f/force options
        if (exists($args{'force'})) {
            $opts->{'force'} = $args{'force'};
        } elsif (exists($args{'f'})) {
            $opts->{'force'} = $args{'f'};
        }
        # h/hard options
        if (exists($args{'hard'})) {
            $opts->{'hard'} = $args{'hard'};
        } elsif (exists($args{'h'})) {
            $opts->{'hard'} = $args{'h'};
        }
        # k/throttle options
        if (exists($args{'throttle'})) {
            $opts->{'throttle'} = $args{'throttle'};
        } elsif (exists($args{'k'})) {
            $opts->{'throttle'} = $args{'k'};
        }
    } else {
        $opts = $self->_set_command_options( %args );
    }

    # copy other common library options to the new options hash
    foreach ( qw/ -Timeout timeout / ) {
        $opts->{$_} = $args{$_} if exists $args{$_};
    }

    return $opts;
}


# set the command options based on the arguments provided and the defaults
# gnote has implemented

sub _set_command_options {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $opts;

    # if provided use method argument to define path
    # -S and -source-path are mutually exclusive
    if (exists($args{'S'})) {
        $opts->{'S'} = $self->{_src}->_build_volume_path(
                                                    description => $args{'S'} );
    } elsif (exists($args{'source-path'})) {
        $opts->{'source-path'} = $self->{_src}->_build_volume_path(
                                          description => $args{'source-path'} );
    # certain commands do not want/need source volume descriptions so
    # unless specified don't give one
    } elsif ($args{'source-needed'} == 1) {
        $opts->{'source-volume'} = $args{'source-volume'}
                                    || $self->{_src}->name;
        $opts->{'source-vserver'} = $args{'source-vserver'}
                                    || $self->{_src}->Vserver->name;
        $opts->{'source-cluster'} = $args{'source-cluster'}
                                    || $self->{_src}->Cluster->name;
    }

    if (exists($args{'destination-path'})) {
        $opts->{'destination-path'} = $self->{_dst}->_build_volume_path(
                                     description => $args{'destination-path'} );
    } else {
        $opts->{'destination-volume'} = $args{'destination-volume'}
                                         ||$self->{_dst}->name;
        $opts->{'destination-vserver'} = $args{'destination-vserver'}
                                         ||$self->{_dst}->Vserver->name;
        $opts->{'destination-cluster'} = $args{'destination-cluster'}
                                         ||$self->{_dst}->Cluster->name;
    }

    # if any other options are provided translate them to 7g style
    # f/force options
    if (exists($args{'f'})) {
        $opts->{'f'} = $args{'f'};
    } elsif (exists($args{'force'})) {
        $opts->{'force'} = $args{'force'};
    }
    # h/hard options
    if (exists($args{'h'})) {
        $opts->{'h'} = $args{'h'};
    } elsif (exists($args{'hard'})) {
        $opts->{'hard'} = $args{'hard'};
    }
    # k/throttle options
    if (exists($args{'k'})) {
        $opts->{'k'} = $args{'k'};
    } elsif (exists($args{'throttle'})) {
        $opts->{'throttle'} = $args{'throttle'};
    }

    return $opts;
}


# this method replaces GX::background for most of the snapmirror methods.
# this is needed to support the testing of the continuum project (434894).
# it adjusts the arguments passed to a snapmirror method to reflect the GNOTE
# mode, 7g style command options, c-mode style command options, or the default
# behavior. it also provides support for the w command line option.

sub _adjust_foreground_option {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($util, $option, $opts, $bg);
    $opts = $args{options};

    # determine the mode of the snapmirror commands to know what option to
    # use. undefined means use the option provided in the method arguments.
    $util = NOTE::SnapMirrorUtil->new();
    if (exists($util->{'_SNAPMIRROR_7G_PATH'})) {
        $option = 'w';
    } elsif (exists($util->{'_SNAPMIRROR_CLUSTER_PATH'})) {
        $option = 'foreground';
    }

    # SNAPMIRROR_7G_PATH or SNAPMIRROR_CLUSTER_PATH override method
    # arguments and are mutualy exclusive
    if (defined($option)) {
        # if any of background, foreground or w arguments are given
        # convert it to the w option
        if (exists($args{'w'})) {
            $opts->{$option} = $args{'w'};
            $bg = (($opts->{$option} eq 'true') ? 0 : 1);
        } elsif (exists($args{'foreground'})) {
            $opts->{$option} = $args{'foreground'};
            $bg = (($opts->{$option} eq 'true') ? 0 : 1);
        } elsif (exists($args{'background'})) {
            $opts->{$option} = (($args{'background'} == 1) ? 'false' : 'true');
            $bg = (($opts->{$option} eq 'true') ? 0 : 1);
        } else {
            # the legacy default is to have the job run in the background
            # and have gnote wait for it to complete
            $opts->{$option} = 'false';
            $bg = 0;
        }

        #$bg = (($opts->{$option} eq 'true') ? 0 : 1);
    } else {
        # if provided use method argument to define determine
        # whether to run the command in the foreground or not
        if (exists($args{'w'})) {
            $opts->{'w'} = $args{'w'};
            $bg = (($opts->{'w'} eq 'true') ? 0 : 1);
        } elsif (exists($args{'foreground'})) {
            $opts->{'foreground'} = $args{'foreground'};
            $bg = (($opts->{'foreground'} eq 'true') ? 0 : 1);
        } elsif (exists($args{'background'})) {
            $opts->{'foreground'} = (($args{'background'} == 1)
                                     ? 'false' : 'true');
            $bg = $args{'background'};
        } else {
            # the legacy default is to have the job run in the background
            # and have gnote wait for it to complete
            $opts->{'foreground'} = 'false';
            $bg = 0;
        }
    }

    # if command will be run in the foreground the timeout argument
    # must be passed to Server.pm. if not SnapMirror.pm uses it to
    # timeout waiting for the command to complete.
    if ( $bg == 0 ) {
        $opts->{'-Timeout'} = (exists($args{'timeout'})
                               ? $args{'timeout'} : $self->{_timeout});
    } elsif (exists($args{'-Timeout'})) {
        # if the command is being run in the background, check if there is a
        # timeout value specific for Server.pm (-Timeout)
        $opts->{'-Timeout'} = $args{'-Timeout'};
    }

    return $bg;
}


=head2 quiesce

Performs a 'snapmirror quiesce' to suspend snapmirror transfers.

Returns:
SUCCESS: 0
FAILURE: confesses

Confessions:
1) if the command is not supported by this build
2) if the command fails for any reason

=over

=item Arguments
Arguments are passed by hash.

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided.
Default: not to provide a source volume description

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.


=item Notes

1)  The default functionality of the method is to not provide a description
    for the source volume in the command.
2)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.
3)  The 'snapmirror quiesce' command is issued from the destination filer.


=item Examples

=back

    # quiesce snapmirror transfers on the snapmirror pair defined by the $sm
    # snapmirror object
    $sm->quiesce();

    # quiesce snapmirror transfers using the ''destination-path'' command option
    $sm->quiesce( 'destination-path' => '//vserver/volume' );

    # quiesce snapmirror transfers forcing the description of the source volume
    $sm->quiesce( S => 'cluster://vserver/volume' );

    # halt snapmirror transfers after an update completes, waiting for the
    # snapmirror relationship is quiesced
    # start an update
    $sm->update();
    # quiesce snapmirror transfers
    $sm->quiesce();
    # wait for the snapmirror to become quiesced
    if ($sm->wait_for_job(timeout => 0) != 0) {
        # wait for update to complete...
        $sm->monitor( status => 'Quiesced' );
    } else {
        # snapmirror update has completed, must have missed it
    }

    # quiesce snapmirror transfers catching errors to verify the command is
    # supported
    $rtn = eval { $sm->quiesce(); 1; };
    # check for command failures
    if (! $rtn) {
        # command failed due to something other than not being supported,
        # fail test
        if ($@ !~ /quiesce command not supported by this build/i) {
            confess('FAIL: quiesce command failed');
        }
    }

    # command succeeded, or is not supported so is not needed (remember,
    # resume not supported either)

=cut

sub quiesce {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $opts, $rtn, $ar_status );

    # verify the quiesce command is supported by this build
    if (!$self->{_dst}->server->has_change(change => 'snapmirror-quiesce-resume')) {
        confess('quiesce command not supported by this build');
    }

    # build the appropriate command options
    $opts = $self->_set_command_options( %args, 'source-needed' => 0 );

    # handle any additional arguments
    if (exists($args{'-Timeout'}) || exists($args{timeout})) {
        $opts->{'-Timeout'} = $args{'-Timeout'} || $args{timeout};
    }

    $rtn = eval { $self->{_dst}->server->snapmirror_quiesce( %{$opts} ); 1; };
    if (! $rtn) {
        confess("snapmirror quiesce failed: $@");
    }

    # verify the relationship is either quiesced or quiescing
    # /Quiesced|Quiescing/i
    $ar_status = $self->status();
    if ($ar_status->{status} !~ /Quiesced|Quiescing/i) {
        confess("snapmirror quiesce did not succeed\n".Dumper($ar_status));
    }

    # return 0 (false) so if at some point errors are returned there can be
    # multiple return codes. currently this method confesses on any error.
    # it is somewhat a gnote standard which i doubt will change so i expect
    # this return value will be ignored anyway.
    return 0;
}



=head2 resume

Performs a 'snapmirror resume' to enable snapmirror transfers that had been
supsended by the 'snapmirror quiesce' command.

Returns:
SUCCESS: 0
FAILURE: confesses

Confessions:
1) if the command is not supported by this build
2) if the command fails for any reason

=over

=item Arguments
Arguments are passed by hash.

I<Optional>

[ B<source-path | S> => I<The description of the volume path. Not the actual
                        source volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<source-cluster> => I<The source cluster name>
     B<source-vserver> => I<The source vserver name>
     B<source-volume> => I<The source volume name> ]

Source volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided.
Default: not to provide a source volume description

[ B<destination-path> => I<The description of the volume path. Not the actual
                        destination volume path. It is defined as:
                        cluster[cluster:][//vserver/]volume>
   | B<destination-cluster> => I<The destination cluster name>
     B<destination-vserver> => I<The destination vserver name>
     B<destination-volume> => I<The destination volume name> ]

Destination volume information. Either one of the path options or the
cluster/vserver/volume tuple can be provided. Default: provide the
Destination-cluster, Destination-vserver and Destination-volume options.


=item Notes

1)  The default functionality of the method is to not provide a description
    for the source volume in the command.
2)  The default functionality of the method is to describe the destination
    volume in terms of the options 'destination-cluster', 'destination-vserver'
    and 'destination-volume'. If the 'destination-path' argument is provided
    then the destination volume will be described as a path using the
    '-destination-path' option.
3)  The 'snapmirror quiesce' command is issued from the destination filer.


=item Examples

=back

    # resume snapmirror transfers on the snapmirror pair defined by the $sm
    # snapmirror object and update the destination with any changes that
    # happened while transfers were stopped
    $sm->resume();
    $sm->update();

    # resume snapmirror transfers using the ''destination-path'' command option
    $sm->resume( 'destination-path' => '//vserver/volume' );

    # resume snapmirror transfers forcing the description of the source volume
    $sm->resume( S => 'cluster://vserver/volume' );

    # resume snapmirror transfers catching errors to verify the command is
    # supported
    $rtn = eval { $sm->resume(); 1; };
    # check for command failures
    if (! $rtn) {
        # command failed due to something other than not being supported,
        # fail test
        if ($@ !~ /resume command not supported by this build/i) {
            confess('FAIL: resume command failed');
        }
    }

    # command succeeded, or is not supported so is not needed (remember,
    # quiesce not supported either)

=cut

sub resume {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $opts, $rtn, $ar_status );

    # verify the resume command is supported by this build
    if (!$self->{_dst}->server->has_change(change => 'snapmirror-quiesce-resume')) {
        confess('resume command not supported by this build');
    }


    # build the appropriate command options
    $opts = $self->_set_command_options( %args, 'source-needed' => 0 );

    # handle any additional arguments
    if (exists($args{'-Timeout'}) || exists($args{timeout})) {
        $opts->{'-Timeout'} = $args{'-Timeout'} || $args{timeout};
    }

    $rtn = eval { $self->{_dst}->server->snapmirror_resume( %{$opts} ); 1; };
    if (! $rtn) {
        confess("snapmirror quiesce failed: $@");
    }

    # verify the relationship is in the correct state...
    # if it is still in the quiesced|quiescing states not good, all other
    # states are ok.
    $ar_status = $self->status();
    if ($ar_status->{status} =~ /Quiesced|Quiescing/i) {
        confess("snapmirror quiesce did not succeed\n".Dumper($ar_status));
    }

    # return 0 (false) so if at some point errors are returned there can be
    # multiple return codes. currently this method confesses on any error.
    # it is somewhat a gnote standard which i doubt will change so i expect
    # this return value will be ignored anyway.
    return 0;
}

# Local method to validate the snapshot is transferred correctly on an XDP
# type operation.
# SN MS3 functionality limited to matching the name and version uuid.
sub _xdp_ss_check {

    my ($self, $ss_name) = @_;
    my $result = 0;

    # These calls will croak if something goes wrong but will return undef
    # if there are no entries found.
    my $s_ss = $self->src->snapshot_status(snapshot=> $ss_name);
    my $d_ss = $self->dst->snapshot_status(snapshot=>$ss_name);

    # If there were no entries or more than one entry we are failed.
    if ( (!defined $s_ss) || (!defined $d_ss)) {
        confess 'Check for XDP ss transfer failed.';
    } elsif ((@{$s_ss} != 1) || (@{$d_ss} != 1) ) {
        NOTE::logcomment($ss_name);
        NOTE::logcomment(Dumper($s_ss));
        NOTE::logcomment(Dumper($d_ss));
        confess 'Check for XDP ss transfer failed. Name match failed.';
    } elsif ( $s_ss->[0]->{'snapshot'} ne $d_ss->[0]->{'snapshot'}
             || $s_ss->[0]->{'version-uuid'} ne $d_ss->[0]->{'version-uuid'}) {
            confess 'Snapshot name and/or version uuid did not match src-dst';
    } else {
        return 1;
    }
}

# Local method to construct arguments for the call to snapshot create
# needed by xdp update.
sub _xdp_create_ss {

    my ($self) = @_;

    my $ss_name;

    # Generate a unique name for the snapshot
    my @stamp = split m/ UTC/,$self->dst->Node->datetime(utc=> 1);
    $stamp[0] =~ s/[:\/ ]/_/g;
    $ss_name = 'XDP_UPDATE_'.$$.'_'.$stamp[0];

    # create the snapshot
    # Query the object to get the snapmirror policy and a label
    my $output = $self->{_dst}->server->
        snapmirror_policy_show( 'policy'=> $self->status('policy'));
    my $label= $output->[0]{'snapmirror-label'}[0];
    if (!defined $label) {
        confess 'Unable to retrieve snapmirror policy label.';
    }
    $ss_name = $self->src->snapshot_create($ss_name,
                                           'snapmirror-label'=>$label);
    return $ss_name;
}

1;
