#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/SnapMirrorSet.pm#1 $
#
## @summary  NOTE SnapMirrorSet class
## @status   Public
## @pod      here
## @author   sydneyw@netapp.com dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::SnapMirrorSet

=head1 DESCRIPTION

This is a "container" class for a set of 1xN SnapMirror relationships. It
consists of 1 source (src) volume and any combination of DP and LS destination
(dst) volumes that are associated with the src volume.

This class contains a few public methods directly corresponding to methods
which can be called on SnapMirror objects, e.g.,

$sms->update(); 

It also facilitates most public calls on SnapMirror methods, with these
additional parameters to indicate which specific SnapMirror relationships in
the set to apply the method to (if only a subset of those in the set are
desired):

=over

=item B<dst> => I<NOTE::Volume>

The mirror destination NOTE::Volume object in this set to update.

=item B<list> => I<ARAYREF to array of NOTE::Volume>

A reference to an array of mirror destination NOTE::Volume objects in this set
to update.

=item B<type> => I<dp|ls>

Update all dp or ls destination volumes in this set.

=back

SnapMirrorSet has 2 public constructors: 'create' and 'initialize'.  Either of
these can be called to create a SnapMirrorSet object.  Each of the
mirror relationships in this object is represented by a NOTE::GX:SnapMirror
object, and methods in that class can be accessed via this class.

In addition an 'update' method exists to allow for the execution of the
'snapmirror update/update-ls-set' commands. This 'update' method may be invoked
for all the mirrors in the SnapMirrorSet, for just the mirrors of a particular
type (dp or ls), or for a single mirror in the SnapMirrorSet.

For SnapMirror methods like 'status', which returns detailed information as a
result of the call, the returned value will be a hash keyed by each destination
GX::Volume object's reference.

For SnapMirror methods that return a single boolean value, such as 'wait_for'
and 'checking', this class will combine (via user choice of AND or OR mechanics)
the results together to give a single return value.  To obtain individual
boolean results, invoke the single mirror's method on it's own.

Examples:

my $sms = NOTE::GX::SnapMirror->initialize(src => $src_vol,
                                           dp  => \@dp_vols,
                                           ls  => { count => 3 });

    # Break a single mirror
$sms->break(dst => $dp_vol);

    # Perform update on all LS mirrors in the set
$sms->update(type => "ls");

    # obtain a hashref keyed by dst GX::Volumes paired to values of status
    # hashref data structures
$HR_status_vs_dst_vol = $sms->status();

    # perform a SnapMirror->check() on all dp mirrors in the set
$sms->check(type => "dp");


All methods support a boolean 'parallel' option which allows for execution on
multiple mirrors concurrently, and is true (1) by default.

=head1 METHODS

=cut

# General Developer's Note:
# An internal 'run' flag is associated with each mirror in the set, and is used
# to identify upon which mirrors in the set the current command should be
# executed.  These flags are also used to determine which mirrors need
# subsequent monitoring.

# This package name seems odd compared to the package name in SnapMirror.pm.
# (I.e., package NOTE::GX::SnapMirror) (BS 10/27/10)
package NOTE::SnapMirrorSet;

use strict;
use GNOTE;
our @ISA = qw(NOTE::GX::SnapMirrorSet GNOTE::GX);
use Carp qw(confess);
use Tharn qw();
use Data::Dumper;

# SMSet contains a list of SMs, and AUTOLOAD here is set up to "allow" calls
# to SM methods on the SMSet itself.  AUTOLOAD then (via SMSet::execute())
# redirects this call to some or all (see "run" flag) of the list of SMs.
# (BS 10/27/10)
our $AUTOLOAD;
sub AUTOLOAD {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $name = $AUTOLOAD;
    $name =~ s/.*://;

    return $self->execute( action => $name, @_ );
}

# Creates the specified number of volumes to be used for snapmirror
# relationships
#
# $tag must =~ /^(src|dp|ls)$/i
# $args{$tag} must be a NOTE::.*Volume object, a ref to an array of
# NOTE::.*Volume objects, or a ref to a hash of parameters as expected
# by NOTE::GX::SnapMirrorUtil->create_volume() with the exception of the
# parameter 'tag', which is set in this hash according to the $tag parameter
# passed to this method.

sub _get_vols {
    my ( $self, $tag, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list;
    if ( ref( $args{$tag} ) =~ /NOTE::.*Volume/ ) {
        @list = ( $args{$tag} );
    }
    elsif ( ref( $args{$tag} ) =~ /ARRAY/ ) {
        @list = @{ $args{$tag} };
    }
    else {

        # Create volume(s) from spec using SnapMirrorUtil::create_volume
        my $vol_tag = ( $tag =~ /dp/i ? "dst" : $tag );
        $args{src}{count} = 1
          unless exists $args{src}{count};
        if ( $tag =~ /src/i ) {
            confess "count parameter for 'src' volume spec should not be > 1"
              if $args{src}{count} > 1;
        }
        else {
            $args{$tag}{'mirror-type'} = "DP";
        }
        @list = $self->{_util}->create_volume( tag => $vol_tag, %{ $args{$tag} } );
    }

    # Reset volume tags to reflect DP or LS volume type
    if ( $tag =~ /(dp|ls)/i ) {
        $tag = uc($tag);
        foreach (@list) {
            $_->tag( $_->tag() . " $tag" );
        }
    }

    # Add back reference to self
    foreach my $vol (@list) {
        $vol->SnapMirrorSet($self);
    }
    return @list;
}

# new() is almost never used as a public constructor for SnapMirrorSet,
# and end users never seem to need access to it, because in the few times it
# is called in thpls, it is always followed by a call to create().  create() in
# and of itself makes the end-user call to new() superfluous, as create() calls
# new() in exactly the same manner itself if new() has not already been called.
# Therefore, I am removing new() from the "public domain" by not creating
# public docs for it.  There is no need for multiple ways to do exactly the same
# thing, then requiring extra docs, etc.  (BS 01/26/2011)

# see ##argstory: tags for details of %args manipulation and use
sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_owner} = $$;

    bless $self, $class;
    NOTE::reference($self);

    $self->{_util} = NOTE::SnapMirrorUtil->new();

    ##argstory: src is the only arg that matters in new() because %args is only
    # used in new() when passed to _get_vols("src",...), and %args is filtered
    # on the src argument therein.
    my @src = $self->_get_vols( "src", %args );
    if ( @src > 1 ) {
        confess "Only single snapmirror src volume is allowed...";
    }
    else {
        $self->{_src} = $src[0];
    }

    $self->{_job} = NOTE::GX::Job->new( Cluster => $src[0]->Cluster );

    return $self;
}


=head2 create

One of three public utility SnapMirrorSet constructors (See also initialize(),
and attach()).  Creates a 1 (src) x N (dst) snapmirror relationship between a
single source (src) volume and a combination of DP and/or LS destination (dst)
mirrors.

create() does not initialize the associated SnapMirrors.  In other words, it
does not trigger the initial snapshot transfer from src to any dst volumes.
See also initialize().

create() can also be called on an already existing SnapMirrorSet object to
place more dp and/or ls SnapMirror relationships in the existing SnapMirrorSet.

When called as a SnapMirrorSet constructor, a src parameter may be passed.
If a src parameter is not passed in such a case, a default src volume will
be created for the new SnapMirrorSet object.

When called on an existing SnapMirrorSet object, the presence of a src
parameter is a fatal exception.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<src> => I<NOTE::Volume spec>

A NOTE::Volume object or a hash containing parameters to create a NOTE::Volume
object to be used as the source of the 1 x N relationships.

When create() is called as a SnapMirrorSet constructor, a src parameter may be
passed.  If a src parameter is not passed in such a case, a default src volume
will be created for the new SnapMirrorSet object.

When create() is called on an existing SnapMirrorSet object, the presence of a
src parameter is a fatal exception.

=item B<dp> => I<NOTE::Volume spec>

An ARRAYREF to an array of NOTE::Volume objects or a hash containing parameters
to create NOTE::Volume objects to be used as the destinations of DP snapmirror
relationships.

=item B<ls> => I<NOTE::Volume spec>

An ARRAYREF to an array of NOTE::Volume objects or a hash containing parameters
to create NOTE::Volume objects to be used as the destionation of LS snapmirror
relationships.

=item B<skip_license> => I<boolean>

Flag to indicate to skip snapmirror licenses or not. Default: 0

=back

=item Returns

=over

=item B<Success:> $self (SnapMirrorSet object)

=item B<Failure:> confesses

=back

=item Examples

=back

    # called as a SnapMirrorSet constructor
    my $sms_using_hash = NOTE::SnapMirrorSet->create(src => {Vserver => $vsrc},
                                                     dp  => {count => 1});

    # called as a SnapMirrorSet constructor
    my $sms_multi_dst = NOTE::SnapMirrorSet->create(src => $vol_src,
                                                    dp  => \@dp_list,
                                                    ls  => \@ls_list);

    # called on an existing SnapMirrorSet object
	$sms_multi_dst->create(dp=>{count=>1}); # Add another 'DP' mirror to the set

=cut

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;
    ##argstory: skip_license optional.  defaults FALSE.  deleted from %args.
    my $skip_license = delete $args{skip_license} || 0;
    if ( ref($class) =~ /NOTE::.*SnapMirrorSet/ ) {
        $self = $class;

        confess q{Existing SnapMirrorSet object missing src volume.}
          unless exists $self->{_src};

        ##argstory: src not allowed.  fatal.
        confess q{Cannot override existing 'src' attribute: },
                $self->{_src}->description()
          if exists $args{src};
    }
    else {
        ##argstory: current %args passed into new().
        $self = NOTE::SnapMirrorSet->new(%args);
    }
    ##argstory: dst not allowed.  fatal.
    confess "Invalid 'dst' parameter. Should use 'dp' or 'ls'..."
      if exists $args{dst};

    my $title = "snapmirror create for set";
    NOTE::logcomment("Start $title:");
    ##argstory: dp exists?  then pass current %args into _get_vols.
    $self->_create( "dp", 0, $skip_license, $self->_get_vols( "dp", %args ) )
      if exists $args{dp};
    ##argstory: ls exists?  then pass current %args into _get_vols.
    $self->_create( "ls", 0, $skip_license, $self->_get_vols( "ls", %args ) )
      if exists $args{ls};
    NOTE::logcomment("Completed $title");

    return $self;
}

# Internal method called by both 'create' and 'initialize' method to actually
# create the mirror(s) of one type (DP or LS)
sub _create {
    my ( $self, $category, $run, $skip_license, @vol_list ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if (@vol_list) {
        my $title = "'snapmirror create' for " . @vol_list . " " . uc($category) . " dst(s)";
        NOTE::logcomment("Start $title:");
        my @list;
        my $type = "_" . $category;
        foreach my $dst (@vol_list) {
            my $name = $dst->name;
            my $key  = $category . ":" . $dst->Cluster->name;
            $self->{$type}{$name}{mirror} = NOTE::GX::SnapMirror->create(
                src      => $self->{_src},
                dst      => $dst,
                type     => $category,
                cmd_only => $skip_license || $self->{_licensed}{$key}    # Run licensed check /set once only
            );
            $self->{_licensed}{$key} = 1;

            # set run flag to indicate whether initilize is needed
            $self->{$type}{$name}{run} = $run;
        }
        NOTE::logcomment("Completed $title");
    }
}


=head2 initialize

One of three public utility SnapMirrorSet constructors (See also create() and
attach()).  Creates a 1 (src) x N (dst) snapmirror relationship between a single
source (src) volume and a combination of DP and/or LS destination (dst) mirrors and initializes the mirrors on the given dst volumes. Can be used to initialize
existing mirrors previously created with the create() method or to create and
initialize new mirrors. This method can handle a mixture of both DP and LS
mirrors.  Individual 'initialize' commands will be issued for DP mirrors. For LS
mirrors, 'initialize-ls-set' command will be issued for the first initialize
request involving multiple LS mirrors, otherwise individual 'initialize'
commands will be issued for each 'ls' mirror.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<dp> => I<NOTE::Volume spec>

An ARRAYREF to an array of NOTE::Volume objects or a hash containing parameters
to create NOTE::Volume objects to be used as the destinations of DP snapmirror
relationships.

=item B<ls> => I<NOTE::Volume spec>

An ARRAYREF to an array of NOTE::Volume objects or a hash containing parameters
to create NOTE::Volume objects to be used as the destionation of LS snapmirror
relationships.

=item B<background> => I<1|0>

Indicates whether control should return at the start (1) or end (0) of
'initialize' commands. Default: 0

=item B<parallel> => I<1|0>

Execute the 'initialize' commands in parallel (1) or not (0). Default: 1

=item B<run_set> => I<1|0>

Always (1) run the 'ls' set command irrespective of number of LS mirrors
involved.  Default: 0

=item B<skip_license> => I<boolean>

Flag to indicate to skip (1) snapmirror licenses or not. Default: 0

=back

=item Returns

=over

=item B<Success:> $self (SnapMirrorSet object)

=item B<Failure:> confesses

=back

=item Examples

=back

    # Used as constructor to create and initialize the mirrors
    my $sms = NOTE::SnapMirrorSet->initialize(src => $src_vol,
                                              dp  => { count => 1 },
                                              ls  => \@dst_vol_list);

    # Adding additional mirrors to an existing object
    $sms->initialize(ls => { count => 2 });

    # Using pre-created mirrors
    {
        # use create() to construct un-initialized SnapMirrorSet
        my $sms = NOTE::SnapMirrorSet->create(src => $src_vol,
                                              dp  => { count => 1 },
                                              ls  => \@dst_vol_list);
        # Initialize all mirrors in the set
        $sms->initialize();
        # Add to existing mirror set
        $sms->initialize(ls => { count => 1 });
    }

=cut

sub initialize {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;

    NOTE::logcomment("Start snapmirror initialize:");

    ##argstory: run_set value saved and deleted from %args.
    my $force_set = delete $args{run_set};

    if ( ref($class) =~ /NOTE::.*SnapMirrorSet/ ) {
        $self = $class;

        # an existing SnapMirrorSet cannot take a new source volume as an
        # argument
        ##argstory: src not allowed here.  fatal.
        if (exists $args{src}) {
            confess q{Invalid 'src' parameter for existing },
                    q{NOTE::GX::SnapMirrorSet object};
        }

        ##argstory: dp and ls optionals.
        if ( exists $args{'dp'} || exists $args{'ls'} ) {

            # Clear previous run flags
            $self->_clear_run();

            # Create new mirror(s)
            ##argstory: skip_license optional.  accessed here and deleted.
            my $skip_license = delete $args{skip_license} || 0;
            foreach my $type ( 'dp', 'ls' ) {
                if (exists $args{$type}) {
                    ##argstory: %args filtered on ls|dp, passed to _get_vols
                    $self->_create($type, 1, $skip_license,
                                   $self->_get_vols($type, %args));
                }
            }
        }
        else {

            # Run everthing so far
            $self->_mark();
        }
    }
    else {
        ##argstory: %args passed to create(), then src arg deleted.
        $self = NOTE::SnapMirrorSet->create(%args);
        delete $args{src};
        $self->_mark();
    }


    # Determine whether to use 'set' command or not
    # Prevents subsequent 'ls-set' commands
    eval {
        $self->src->server->snapmirror_show(
            'source-path'    => $self->{_src}->fullname,
            type             => "ls",
            state            => "snapmirrored"
        );
    };
    my $run_ls_set = $self->_ls_run_cnt() && ( $@ =~ /no entries/ );

    NOTE::logcomment("Initialize the mirror(s):");
    ##argstory: %args passed to _execute(), filtered on dp
    $self->_execute( "initialize", "dp", %args );
    if ( $run_ls_set || $force_set ) {
        my $text;
        eval {
            $text = $self->{_src}->server->snapmirror_initialize_ls_set(
                'source-volume'  => $self->{_src}->name,
                'source-vserver' => $self->{_src}->Vserver->name,
                'source-cluster' => $self->{_src}->Cluster->name,
                %args
            );
        };
        confess $@ if $@;
        $self->{_job_id} = $self->{_src}->Node->job_id($text);
        $self->{_src}->Node->job_wait( id => $self->{_job_id}, state => $self->_sm_job_states, timeout => 15 );
    }
    else {
        ##argstory: %args passed to _execute(), filtered on ls
        $self->_execute( "initialize", "ls", %args );
    }

    ##argstory: background optional.
    $self->monitor(%args)
      unless $args{background};

    NOTE::logcomment("Completed snapmirror initialize");

    return $self;
}

=head2 attach

One of three public constructors for SnapMirrorSet objects.  See also create()
and initialize().

Creates a SnapMirrorSet object out of an existing 1 x N snapmirror
relationsips representing one src and one of more DP and/or LS destinations.

The parameters for this method are the same ones passed to 'create' and
'initialize' method. This method will create the object without actually
executing any snapmirror commands.

This method will create an object based on it's parameters so that it is
possible to create an object that is a subset of an existing 1 x N
relationships. Please use GX::Cluster::SnapMirrorSet method to obtain an
GX::SnapMirrorSet object that represents an existing 1 x N relationship.

This method is a constructor to this class and returns a NOTE::SnapMirrorSet
object.

=over

=item Arguements
Arguments passed by hash table

Hash table of arguments:

=over

B<src> => I<NOTE::Volume spec>

A NOTE::Volume object object  to be used as the source of the 1 x N
relationships.

=back

I<Optional Arguments>

=over

B<dp> => I<NOTE::Volume spec>

An ARRAY to an array of NOTE::Volume objects to be used as the destinations of
DP snapmirror relationships.

B<ls> => I<NOTE::Volume spec>

An ARRAY to an array of NOTE::Volume objects to be used as the destionation of
LS snapmirror relationships.

=back

=item Examples

=back

    my $dp_set = NOTE::SnapMirrorSet->attach(src => $src_vol, dp => \@dp_vols);
    my $ls_set = NOTE::SnapMirrorSet->attach(src => $src_vol, ls => \@ls_vols);
    my $sms = NOTE::SnapMirrorSet->attach(src => $src_vol,
                                          dp =>  \@dp_vols,
                                          ls =>  \@ls_vols);

=cut

sub attach {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing 'src' parameter"
      unless exists $args{src};
    confess "src must be a NOTE::GX::Volume object"
      unless ref( $args{src} ) =~ /NOTE::GX::Volume/;
    confess "dp must be an ARRAY of NOTE::GX::Volume objects: " . ref( $args{dp} )
      if exists $args{dp} && !( ref( $args{dp} ) =~ /ARRAY/ );
    confess "ls must be an ARRAY of NOTE::GX::Volume objects"
      if exists $args{ls} && !( ref( $args{ls} ) =~ /ARRAY/ );

    my $self = NOTE::SnapMirrorSet->new( src => $args{src} );
    NOTE::logcomment( "Attaching SnapMirrorSet for src " . $self->src->description );
    my $msg = "	src: " . $self->src->fullname . "\n";

    foreach my $dst_type ( "dp", "ls" ) {
        if ( exists $args{$dst_type} ) {
            my $type = "_" . lc($dst_type);

            foreach my $vol ( $self->_get_vols( $dst_type, %args ) ) {
                $self->{$type}{ $vol->name }{mirror} = NOTE::GX::SnapMirror->new( src => $self->src, dst => $vol, type => $dst_type );
                $self->{$type}{ $vol->name }{mirror}->{_created} = 1;
                $msg .= "		$dst_type dst " . $vol->fullname . "\n";
            }
        }
    }
    NOTE::logcomment("Attached to SnapMirrorSet:\n$msg");

    return $self;
}

=head2 update

Perform snapmirror update on the mirrors in this set. Single 'update' commands
will be executed for 'DP' mirrors. For 'LS' mirrors the 'update-ls-set' command
will be executed if there are more then one 'LS' mirrors, otherwise the regular
'update' command will be used.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

Note that if none of the following 3 optional arguments are specified
(dst|list|type), this method will trigger an update of all destination volumes
in this SnapMirrorSet.

=over

=item B<dst> => I<NOTE::Volume>

The mirror destination NOTE::Volume object in this set to update.

=item B<list> => I<ARAYREF to array of NOTE::Volume>

A reference to an array of mirror destination NOTE::Volume objects in this set
to update.

=item B<type> => I<dp|ls>

Update all dp or ls destination volumes in this set.

=item B<background> => I<1|0>

Indicates whether control should return at the start or end of 'update'
commands. Default: 0

=item B<parallel> => I<1|0>

Execute the 'update' commands in parallel or not. Default: 1

=item B<run_set> => I<1|0>

If TRUE, force execution of 'update-ls-set' command for any ls mirrors in the
update list, even if there is only a single ls mirror to update. Default: 0

=back

=item Returns

=over

=item B<Success:> Nothing specific.  (I.e., Do not use returned value.)

=item B<Failure:> confesses

=back

=item Examples

=back

    # Update all destinations in this 1 x N snapmirror relationships set
    $sms->update();

    # Update only the mirror with $dst_vol destination
    $sms->update(dst => $dst_vol);

    # Update all LS mirrors, using 'update-ls-set' cmnd for multiple LS mirrors
    $sms->update(type => "ls");

=cut

sub update {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment("Start snapmirror update:");

    $self->_clear_run();

    # mark mirror(s) to include in the update and determine which 'ls-set'
    # command to use
    $self->_mark(%args);

    # run update-ls-set if there are at least one ls dst
    # and any of the following is true:
    #	1. run_set option is set
    #	2. type is 'ls'
    #	3. Not a single mirror update
    my $run_set = $self->_ls_run_cnt() &&
        (delete $args{run_set} || $args{type} =~ /ls/ || !exists $args{dst});

    delete $args{type};
    delete $args{dst};

    $self->_execute( "update", "dp", %args );

    if ($run_set) {
        my $text;
        eval {
            $text = $self->{_src}->server->snapmirror_update_ls_set(
                'source-volume'  => $self->{_src}->name,
                'source-vserver' => $self->{_src}->Vserver->name,
                'source-cluster' => $self->{_src}->Cluster->name,
            );
        };
        confess $@ if $@;
        $self->{_job_id} = $self->{_src}->Node->job_id($text);
        my %poll_opts;

        # Default timeout to 5 mins as on a busy system a queued job may take a
        # while to become active
        $poll_opts{timeout}  = $args{timeout}  || 300;
        $poll_opts{interval} = $args{interval} || 5;
        $self->{_src}->Node->job_wait(id    => $self->{_job_id},
                                      state => $self->_sm_job_states,
                                      %poll_opts );
    }
    else {
        $self->_execute( "update", "ls", %args );
    }
    $self->monitor(%args)
      unless $args{background};

    NOTE::logcomment("Completed snapmirror update");
}

sub _sm_job_states {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return "(dead|dormant|error|quit|running|success)";
}

# execute() should not be a public interface method (despite the fact that this
# class also contains a _execute() method as well).  execute should only be used
# indirectly by end-users via AUTOLOAD().  For this reason, I am commenting out
# the old pod for execute().  The old pod has not been validated at this time.

#=head2 ::execute
#
#Execute specific snapmirror command against specified mirror(s) within the set.
#
#This method is meant as the main interface between this object and the
#underlying SnapMirror object. It is encouraged that calls are made to the
#underlying SnapMirror methods instead of calling this method, e.g. call
#$sms->update() instead of $sms->execute(action=>"update");, which are
#equivalent.

#This method returns the result of the underlying SnapMirror objects as an
#HASHREF with the dst GX::Volume objects as key.
#
#=over
#
#=item Arguments
#
#B<action> => I<snapmirror command>
#
#Specify what snapmirror command to run, e.g. break, resync, etc.
#
#B<dst> => I<NOTE::Volume>
#
#A dst volumne name, NOTE::Volume object for a dst volume, or a NOTE::SnapMirror
#object, in the set, to execute snapmirror command on.
#
#B<list> => I<array of dst>
#
#A ARRARYREF of 'dst' entries to provide flexible mixing and matching of the
#mirrors in the set.
#
#B<type> => I<dp|ls>
#
#I<Optional>
#
#Checks state/status for 'DP' or 'LS' mirrors.
#
#Additional parameters will be passed to GNOTE::GX::SnapMirror for processing.
#
#B<background> => I<t1|0>
#
#I<Optional>
#
#Indicates whether control should return at the start or end of command
#execution. Default: 0
#
#B<parallel> => I<1|0>
#
#I<Optional>
#
#Determine to execute the commands in parallel or not. Default: 1
#
#=item Examples
#
#=back
#
#    $sms->execute(action=>"break", dst=>$dst_vol);
#
#	# returns hash, keyed by dst volumes, of mirror status
#	my %status = $sms->execute(action=>"status");
#
#=cut

sub execute {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing 'action' parameter"
      unless exists $args{action};
    my $action = delete $args{action};

    confess "Cannot call execute method for 'create' or 'initialize"
      if $action =~ /(create|initialize)/;

    return $self->update(%args)
      if $action =~ /update/;

    return $self->_promote(%args)
      if $action =~ /promote/;

    NOTE::logcomment("Start action $action:");
    $self->_clear_run();

    if ( exists $args{dst} ) {
        my $dst = delete $args{dst};
        if ( my $type = $self->_mark_object($dst) ) {
            $self->_prepare_args( $action, $type, %args );
            $self->_execute( $action, $type, %args );
        }
        else {
            confess "Unable to find active mirror to run $action on.";
        }
    }
    else {
        $self->_mark(%args);
        delete $args{type};
        foreach ( "dp", "ls" ) {
            $self->_prepare_args( $action, $_, %args );
        }
        foreach ( "dp", "ls" ) {
            $self->_execute( $action, $_, %args );
        }
    }

    if ( $args{background} ) {
        NOTE::logcomment("Action $action running in background");
    }
    else {

        # Check status only for break, check, restore or resync
        # No status check for initialize, update and promote, as they are
        # handled by their respective methods
        if ( $action =~ /^(break|check|restore|resync)$/ ) {
            if ( $action =~ /break/ ) {
                $args{state}  = "broken-off";
                $args{status} = "idle";
            }
            $self->monitor(%args);
        }
        NOTE::logcomment("Completed action $action");
    }

    if ( $action =~ /(wait_for_job|checking)/ ) {
        my %op = (
            "wait_for_job" => "or"    # mirrors returns when condition are met, otherwise it times out
            , "checking" => "or"      # Any +ve result means that check is still running
        );

        # Return combined result
        return $self->_rc( op => $op{$action}, list => $self->{_result} );
    }
    else {
        return $self->{_result};
    }
}

# Pre-process the arguments

sub _prepare_args {
    my ( $self, $action, $type, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $action =~ /wait_for/ ) {
        if ( exists $args{current} ) {
            $self->_execute( "_get_wait_args", $type, %args );
            $self->{_wait_args} = $self->{_result};
            delete $args{current};
        }
        else {
            delete $self->{_wait_args};
        }
    }
}

sub _rc {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{op} = "and" unless exists $args{op};
    confess "Invalid 'op'. Must be 'and' or 'or'"
      unless $args{op} =~ /(and|or)/;
    confess "Invalid or missing 'list'. Must be a HASH"
      unless ref( $args{list} ) =~ /HASH/;

    my $result = ( $args{op} =~ /and/ ? 1 : 0 );
    foreach my $dst ( keys %{ $args{list} } ) {
        if ( $args{op} =~ /and/ ) {
            $result &&= $args{list}->{$dst};
        }
        else {
            $result ||= $args{list}->{$dst};
        }
    }
    NOTE::logcomment("RC: $result");
    return $result;
}

# Perform 'promote' operation

sub _promote {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "'promote' action can only be applied to one destination mirror. Missing 'dst' parameter"
      unless exists $args{dst};
    confess "'promote' runs only in 'foreground' mode."
      if $args{background};

    # Remove mirror from set only after successful 'promote' as the 'promote' will abort if it fails
    my $mirror = $self->mirror( name => $args{dst}->name );
    $self->{_src} = $mirror->promote(%args);
    $self->_remove_from_set( $args{dst}->name );

    NOTE::logcomment( "Modify set's src volume to " . $self->{_src}->description . ":" );
    $self->_set_src("_dp");
    $self->_set_src("_ls");
    NOTE::logcomment("Promote completed for snapmirror set");
}

# Remove a single mirror from set

sub _remove_from_set {
    my ( $self, $name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    foreach my $type ( "_dp", "_ls" ) {
        if ( exists $self->{$type}{$name} ) {
            my $mirror = $self->{$type}{$name}{mirror};
            delete $self->{$type}{$name};
            $mirror->dst->_sm_clr();    # Clear the internal SmSet object from dst volume
            return $mirror;
        }
    }

    confess "Unable to find specified dst " . $name;
}

# Set src

sub _set_src {
    my ( $self, $type ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Change src for all mirrors of type
    foreach ( keys %{ $self->{$type} } ) {
        my $old_src = $self->{$type}{$_}{mirror}->src( $self->{_src} );
        NOTE::logcomment( "Replaced src " . $old_src->description . " with " . $self->{$type}{$_}{mirror}->src->description . " for dst " . $self->{$type}{$_}{mirror}->dst->description );

    }
}

=head2 monitor

After any method is called on a SnapMirrorSet, a subsequent call to monitor()
may be used to wait for all of the SnapMirrors involved (which may be a
subset of all the mirrors in the set) to reach a given state and/or status.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<state> => I<string>

State to monitor mirror for. Default: snapmirrored

=item B<status> => I<string>

Status to monitor mirror for. Default: idle

=item B<timeout> => I<int>

Monitor wait timeout.  Default: SnapMirrorUtil->param("JOB_TIMEOUT")

=item B<interval> => I<int>

Monitor sleep time between calls to 'snapmirror show's for status updates.

=back

=item Returns

=over

=item B<Success:> Nothing specific.  (I.e., Do not use returned value.)

=item B<Failure:> confesses

=back

=item Examples

=back

    # Wait until all mirrors in this set is in snapmirrored-idled state
    $sms->initialize();
    $sms->monitor();

    # Wait until all ls mirrors are actively updating
    $sms->update(type => 'ls');
    $sms->monitor(status => 'transfering');

=cut

sub monitor {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    ##argstory: state optional.  read and deleted.
    my $state  = delete $args{state}  || "snapmirrored";
    ##argstory: status optional.  read and deleted.
    my $status = delete $args{status} || "idle";
    ##argstory: timeout optional.  read and deleted.
    my $timeout = delete $args{timeout}
      || $self->{_util}->param("JOB_TIMEOUT");    # default to JOB_TIMEOUT CLI
    ##argstory: interval optional.  read and deleted.
    my $interval = delete $args{interval} || 10;

    # Build hash of mirror(s) to obtain results for
    my $total = 0;
    my %tag;
    foreach my $type ( "_dp", "_ls" ) {
        foreach my $name ( keys %{ $self->{$type} } ) {
            if ( $self->{$type}{$name}{run} ) {
                my $mirror = $self->{$type}{$name}{mirror};
                my $path   = $mirror->dst->hostpath;
                $tag{$path} = $mirror->dst->tag;
                $total++;
            }
        }
    }
    my $title = "Monitor $total mirror(s). Wait for ($state, $status):";
    if ($total) {
        NOTE::logcomment("Start $title");
    }
    else {
        NOTE::logcomment("No specified mirror(s) to monitor");
        return;
    }

    my $end_time = time + $timeout;
    my $done_cnt;
    do {
        confess "Timeout: $title"
          if time > $end_time;
        Tharn::snooze($interval);

        # Burt 443369: log associated job show information (BS 11/8/10)
        my $job_msg = q{};
        if ($self->{_job_id}) {
            $job_msg .= "Job id: $self->{_job_id}, ";

            eval {
                     my @jobs =  $self->{_src}->server->job_show(
                         'id' => $self->{_job_id},
                                                                );
                     $job_msg .= "State: $jobs[0]->{state}, " .
                                 $jobs[0]->{description};
                     1;
                 } or do {
                     if ($@ =~ /no entries matching your query/) {
                         $job_msg .= 'Job not found.';
                     } else {
                         confess $@;
                     }
                 };
        }

        # count completed and compose status message
        my $msg = "Monitor result:";
        $done_cnt = 0;

        foreach my $type ("_dp","_ls") {
            foreach my $name ( keys %{ $self->{$type} } ) {
                if ( $self->{$type}{$name}{run} ) {
                    my $mirror = $self->{$type}{$name}{mirror};
                    my $sm1 = $mirror->status;

                    my $path = $sm1->{'destination-path'};
                    $msg .= "\n $path, state: $sm1->{state}, " .
                                "status: $sm1->{status}";
                    # Count as done if expected state:status have been reached or
                    # error 'uninitialized:idled'

                    # If error state then check on job status below will show the
                    # error
                    if (        # Expected state:status
                        ($sm1->{state} =~ /$state/i &&
                         $sm1->{status} =~ /$status/i)
                        ||
                        # Terminal Error state:status
                        ($sm1->{state} =~ /uninitialized/i &&
                         $sm1->{status} =~ /idle/i)) {
                         $done_cnt++;
                    }
                 }
             }
        }

        # Verify all mirrors are idled by calling each mirror's monitor
        # This in turn verifies that corresponding job are completed
        # fix for burt341399
        # This is only needed for 'snapmirrored:idle' state:status
        if (        ( $done_cnt == $total ) &&
                    ( $state =~ /snapmirrored/i ) && ( $status =~ /idle/i ) ) {
            if ( $self->{_job_id} ) {

                # State dormant is the termination job for the schedule Daemon
                eval {
                         $self->{_src}->Node->job_wait(id => $self->{_job_id},
                                 state => '(success|dormant)', timeout => 300);
                         1;
                     } or do {
                         ($@ !~ /no entries/) && confess $@;
                     };

                delete $self->{_job_id};
            }
            else {
                $self->_verify_idle( "_ls", $state, $status );
            }
            $self->_verify_idle( "_dp", $state, $status );
        }

        # log SnapMirrorSet state
        NOTE::logcomment("\n$job_msg");
        NOTE::logcomment("\n$msg\n$done_cnt out of $total done");

    } while $done_cnt < $total;

    NOTE::logcomment("Completed: $title");
}

=head2 catch_up

Returns an array of NOTE::SnapMirror (LS only) objects that are currently in
catch-up mode.  An optional parameter allows for the query to be restricted to a
single LS mirror destination.

=over

=item Arguments 
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<dst> => I<NOTE::Volume>

Restricts the test to the supplied LS destination volume.  Default: All LS
destination volumes.

=item B<timeout> => I<int>

Timeout for obtaining the result of this query.  Default:
SnapMirrorUtil->param("JOB_TIMEOUT")

=item B<interval> => I<int> 

Sleep time between calls to 'snapmirror show's for status updates.  Default: 15
seconds.

=back

=item Returns

=over

=item B<Success:> Array of NOTE::SnapMirror objects still in catch-up mode.

=item B<Failure:> confesses

=back

=item Examples

=back

    # Obtain a list of all mirrors in catch-up mode
    my @mirrors_currently_catching_up = $sms->catch_up();

    # Used to indicate if a particular mirror is in catch-up mode
    my ($an_ls_mirror_dst, $another) = $sms->dst(type => 'ls');
    my $a_catch_up_flag = scalar($sms->catch_up(dst => $an_ls_mirror_dst));

=cut

sub catch_up {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $title = "SnapMirrorSet Src: " . $self->src->description;
    confess "Invalid 'dst' parameter: " . ref( $args{dst} ) . ". Must be <NOTE::Volume> object"
      if exists $args{dst} && ref( $args{dst} ) !~ /^NOTE::.*Volume$/;
    confess "No job id found for $title"
      unless exists $self->{_job_id};

    $args{timeout} = ( $self->{_util}->param("JOB_TIMEOUT") || 60 * 60 ) unless exists $args{timeout};
    $args{interval} = 15 unless exists $args{interval};

    # Wait for 'ls-set' job to be in catch-up mode
    my $progress;
    eval { $progress = $self->{_job}->column_wait( id => $self->{_job_id}, %args, check => "progress", expected => "waiting for catch-up", return => "progress" ); };
    confess $@
      if $@ && !( $@ =~ /completed/i );

    my @catchup_list = ();
    my $msg;

    # The list of mirror dsts in catch-up mode is listed in the 'description' field, so parse it to obtain this list

    #The progress string would be in one of the below mentioned formats
    # ....update for destination(s) dpgwahoo12://vs0/vl24077_bane_polishes"
    # ....update for destinations cluster-h5-h6://vs0/vl11969_volt_trembles,cluster-h5-h6://vs0/vl24077_bane_polishes."
    # ....update for destination cluster-h5-h6://vs0/vl11969_volt_trembles."
    if ( $progress =~ /destination[(s)]*\s+(.*)$/i ) {
        my $list = $1;
        $list =~ s/"//;
        $list =~ s/\.$//;
        foreach my $n ( split ",", $list ) {
            my $name = $n;
            if ( $name =~ /\/*(\w+)$/ ) {
                $name = $1;
            }
            next if exists $args{dst} && !( $args{dst}->name eq $name );
            if ( my @m = $self->mirror( name => $name ) ) {
                push @catchup_list, $m[0];
                $msg .= "\n	" . $m[0]->dst->description;
            }
        }
    }
    if ($msg) {
        NOTE::logcomment( "Catchup dst for $title:" . $msg );
    }
    else {
        NOTE::logcomment("No catchup mirror found $title");
    }
    return @catchup_list;
}

=head2 laggers

Returns an array of NOTE::SnapMirror objects in the set currently lagging.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<dst> => I<NOTE::Volume>

Restricts the test to the supplied NOTE::Volume LS destination. Default: All LS destinations.

=back

=item Returns

=over

=item B<Success:> Array of NOTE::SnapMirror objects

=item B<Failure:> confesses

=back

=item Examples

=back

     my @lagger_list = $sms->laggers();

=cut

sub laggers {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Invalid 'dst' parameter: " . ref( $args{dst} ) . ". Must be <NOTE::GX::Volume> object"
      if exists $args{dst} && !( ref( $args{dst} ) =~ /^NOTE::.*Volume$/ );

    my @lag_list;
    my $msg;
    foreach my $m ( $self->mirror() ) {
        next if defined $args{dst} && $args{dst} ne $m->dst;
        if ( $m->is_lagging() ) {
            push @lag_list, $m;
            $msg .= "\n	" . $m->dst->description;
        }
    }
    if ($msg) {
        NOTE::logcomment( "Found following lagging dst:" . $msg );
    }
    else {
        NOTE::logcomment("No lagging dst found");
    }
    return @lag_list;
}

# Internal method to verify the mirrors are truely in 'idle' state

sub _verify_idle {
    my ( $self, $type, $state, $status ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    foreach my $name ( keys %{ $self->{$type} } ) {
        $self->{$type}{$name}{mirror}->monitor( interval => 1, state => $state, status => $status )
          if $self->{$type}{$name}{run}
              && exists $self->{$type}{$name}{mirror}->{_job_id};    # Monitor only if there is a job
    }
}

=head2 src

Returns the Source NOTE::Volume object.

=over

=item Arguments 
Arguments passed by hash table

Hash table of arguments:

=over

=item I<None>

=back

=item Returns

=over

=item B<Success:> Src NOTE::Volume object.

=item B<Failure:> undef

=back

=item Examples

=back

    my $vol = $sms->src();

=cut

sub src {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->{_src};
}

=head2 dst

Returns destination NOTE::Volume(s) that matches the specify criteria. Criteria
to determine volumes to return can be name or type.  All destination
NOTE::Volume destination objects will be returned if no criteria is supplied.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

item= B<name> => I<string>

Returns only the destination NOTE::Volume name matching this name.

item= B<type> => I<dp|ls>

Returns DP or LS destination NOTE::Volume(s)

=back

=item Returns

=over

=item B<Success:> Array of NOTE::Volume objects

=item B<Failure:> confesses

=back

=item Examples

=back

    my ($dst_vol) = $sms->dst(name => 'a_dp_vol');
    my @dp_list = $sms->dst(type => 'dp');

=cut

sub dst {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->_set_object( %args, object => "dst" );
}

=head2 mirror

Returns NOTE::GX::SnapMirror object(s) that matches the specified criteria.
Criteria to determine volumes to return can be name of dst volume or type.  All
SnapMirror objects objects will be returned if no criteria is supplied.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

item= B<name> => I<string>

Returns only the NOTE::GX::SnapMirror with dst name matching this name.

item= B<type> => I<dp|ls>

Returns DP or LS NOTE::GX::SnapMirror object(s).

=back

=item Returns

=over

=item B<Success:> Array of NOTE::GX::SnapMirror objects

=item B<Failure:> confesses

=back

=item Examples

=back

    my ($mirror) = $sms->mirror(name => 'a_dp_vol');
    my @dp_mirror = $sms->mirror(type => 'dp');

=cut

sub mirror {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->_set_object( %args, object => "mirror" );
}

sub _set_object {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Unexpected 'dst' parameter."
      if exists $args{dst};

    my @list;
    my $exp_name = $args{name} || ".*";
    my $exp_type = $args{type} || "(dp|ls)";
    foreach my $t ( "dp", "ls" ) {
        my $type = "_" . $t;
        foreach my $name ( keys %{ $self->{$type} } ) {
            push @list,
              (
                  $args{object} =~ /mirror/
                ? $self->{$type}{$name}{mirror}
                : $self->{$type}{$name}{mirror}->dst
              ) if ( $name =~ /$exp_name/ && $type =~ /$exp_type/ );
        }
    }
    return ( wantarray ? @list : $list[0] );
}

# Clears 'run' flag for all mirrors

sub _clear_run {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->_mark( run => 0 );
    $self->{_result} = {};
}

# Determine and set 'run' flag from 'dst', 'list' or 'type' parameters

sub _mark {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $run_flag = ( exists $args{run} ? delete $args{run} : 1 );
    if ( exists $args{dst} ) {
        $self->_mark_object( $args{dst}, $run_flag );
    }
    elsif ( exists $args{list} ) {
        confess "Invalid list " . $args{list} . ". Must be an ARRAYREF"
          unless ref( $args{list} ) =~ /ARRAY/;
        $self->_mark_list( $args{list}, $run_flag );
    }
    elsif ( exists $args{type} ) {
        confess "Invalid type value " . $args{type} . ". Must be either 'dp' or 'ls'"
          unless $args{type} =~ /^(dp|ls)$/i;
        $self->_mark_type( $args{type}, $run_flag );
    }
    else {
        $self->_mark_type( "dp", $run_flag );
        $self->_mark_type( "ls", $run_flag );
    }
}

# Determine type of object and set the corresponding mirror's 'run' flag

sub _mark_object {
    my ( $self, $item, $run_flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name;
    my $obj_type = ref($item);
    if ( $obj_type =~ /NOTE::.*Volume/ ) {
        $name = $item->name;
    }
    elsif ( $obj_type =~ /NOTE::.*SnapMirror/ ) {
        $name = $item->dst->name;
    }
    elsif ( ref( \$item ) =~ /SCALAR/ ) {
        $name = $item;
    }
    else {
        confess "Unexpected value ($obj_type). Must be a dst volume name, NOTE::GX::Volume or NOTE::GX::SnapMirror object";
    }

    $run_flag = 1 unless defined $run_flag;

    return undef if $run_flag && $self->{_suspend}{$name};    # Ignore suspended mirrors

    foreach my $type ( "dp", "ls" ) {
        if ( exists $self->{ "_" . $type }{$name} ) {
            $self->{ "_" . $type }{$name}{run} = $run_flag;
            NOTE::logcomment("$type mirror dst $name: run flag set to $run_flag");
            return $type;
        }
    }
    NOTE::logcomment("Unable to find $name in set. Ignoring...");
    return undef;
}

# Determine and set 'run' flag for a list of mirrors from the 'list' parameter

sub _mark_list {
    my ( $self, $list, $run_flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $total = 0;
    my $title = "run flag set to $run_flag";
    foreach my $entry ( @{$list} ) {
        $self->_mark_object( $entry, $run_flag );
        $total++;
    }
    NOTE::logcomment("$total $title");
}

# Determin and set 'run' flag for all mirrors from 'type' parameter

sub _mark_type {
    my ( $self, $type, $run_flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $run_flag = 1 unless defined $run_flag;
    my $total = 0;
    my $title = "$type mirror(s): run flag set to $run_flag";
    $type = "_" . $type;
    foreach my $name ( keys %{ $self->{$type} } ) {
        next if $run_flag && $self->{_suspend}{$name};    # Ignore suspended mirrors
        $self->{$type}{$name}{run} = $run_flag;
        $total++;
    }
    NOTE::logcomment("$total $title");
}

# Internal method to be used by NOTE::GX::Volume to suspend the mirror

sub _suspend {
    my ( $self, $vol, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $vol->name eq $self->src->name ) {

        # No need to set suspend for volume source.
        NOTE::logcomment( "Ignore " . ( $flag ? "" : "un" ) . "suspend request for SMSet src vol " . $vol->description );
        return;
    }

    confess $vol->description . " not a dst in this SnapMirrorSet"
      unless $self->dst( $vol->name );
    $self->{_suspend}{ $vol->name } = $flag;
    NOTE::logcomment( "SMSet dst " . $vol->description . ( $flag ? " " : " un" ) . "suspended" );
}

# Determine number of LS mirrors to execute command

sub _ls_run_cnt {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $cnt = 0;
    foreach my $name ( keys %{ $self->{_ls} } ) {
        $cnt++ if $self->{_ls}{$name}{run};
    }
    return $cnt;
}

# Execute the specified snapmirror action on the mirrors with 'run' flag set
#
# When simulating abort-ls-set, the first LS dst is used to mimic the ONTAP
# functionality of aborting a set with a single LS dst. A possible alternative
# implementation is to determine active LS dst with a 'sm show' and pick abort
# one of the active LS dst.

##argstory: herein we see that %args may contain key/values for any one of
##          the actions that can be called on a SnapMirror object.  Backing
##          up to the higher level calls, this makes it difficult to
##          document the possible arguments that may be called on them.  But
##          not impossible in a case such as initialize() which always calls
##          _execute() with $action specifically set to 'initialize', so
##          we know that the arguments passed into SnapMirror->initialize() will
##          be the ones that matter in that case.  But I am not sure if all
##          available SnapMirror->initialize() parameters make sense in the
##          context of a call to SnapMirrorSet->initialize().
sub _execute {
    my ( $self, $action, $type, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    ##argstory: parallel optional. read and deleted from %args.
    my $parallel = delete $args{parallel} || 1;
    if ( $parallel && $action =~ /(initialize|update|check|restore|resync)/ ) {

        # force background run if need to run in parallel for actions with background option
        ##argstory: background set in %args.
        $args{background} = 1;
    }

    my ( $total, $cnt ) = ( 0, 0 );
    my $title = ( $parallel ? "Parallel" : "Sequential" ) . " Execution: Action $action, $type Mirror(s)";
    $type = "_" . $type;
    my $abort_ls_set_cnt = 0;
    my $have_ls_set      = exists $self->{_job_id};
    ##argstory: %args copied into %opts (for no particular reason (?)).
    my %opts             = %args;
    foreach my $name ( keys %{ $self->{$type} } ) {
        $total++;
        if ( $self->{$type}{$name}{run} ) {
            my $mirror = $self->{$type}{$name}{mirror};
            if ( $action =~ /abort/ && $type =~ /ls/i && $have_ls_set ) {
                if ($abort_ls_set_cnt) {
                    NOTE::logcomment( "Skipping abort for " . $mirror->description . ", as abort-ls-set already executed" );
                    $cnt++;
                    next;
                }
                else {

                    # Simulate abort-ls-set by trasnferring job id into the mirror object before calling 'abort'
                    $mirror->{_job_id} = delete $self->{_job_id}
                      unless exists $mirror->{_job_id};
                    NOTE::logcomment( "Simulate abort-ls-set by aborting " . $mirror->description );
                    $abort_ls_set_cnt++;
                }
            }

            # Use 'current' value from 'wait_for' call
            ##argstory: %opts nee %args may be written over (?)
            %opts = @{ $self->{_wait_args}{ $mirror->dst } }
              if $action =~ /wait_for/ && exists $self->{_wait_args}{ $mirror->dst };

            # execute the action
            eval {
                my @result = $mirror->$action(%opts);
                $self->{_result}{ $mirror->dst } = ( @result > 1 ? \@result : $result[0] );
            };
            confess $mirror->description . ": Action $action, Error: $@"
              if $@;

            $self->_remove_from_set($name)
              if $action =~ /delete/;

            $cnt++;
        }
    }
    NOTE::logcomment( $title . " $cnt out of $total" )
      if $cnt;
}

sub DESTROY {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_destroyed} = 1;
}

1

