# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/SnapMirrorUtil.pm#1 $
#
## @summary NOTE SnapMirrorUtil Class
##
## @status public
##
## @author sydneyw@netapp.com, dl-dpgqa-automation@netapp.com
##
## @pod here

=head1 NAME

NOTE::SnapMirrorUtil

=head1 DESCRIPTION

The purpose of this module is to provide a central processing of CLI parameters
required by SnapMirror test scripts and to provide a set of useful methods to
aid in the writing of snapmirror test scripts.

The constructor of this module processes the CLI parameters and make a copy
available via the 'param' method.  In addition the CLI parameters are
subsequently used to provide useful methods such as 'create_volume'.

The collection of CLI parameters can be broken down into two categories, those
that are a related set of CLI parameters and those that exist stand alone.
SnapMirror tests typically consist of source and destination DP volumes and/or
LS volumes, and each volume type shares similar characteristics, which may be
different from volumes of other types.

Since LS destination is a new type of mirror destination volumes,
the term "DST" will be used for DP desintations, and new term "LS" introduced
to represent LS desintations.

Each 'set' of CLI parameter is used to specify one particular aspect of the
volume for the volume type.  An example of volume parameters is Vserver the
volume resides/created in, which may or may not be different for source and
destination volumes.  So the Vserver specification for source volumes are given
by the CLI parameter VSERVER_SRC, and for destination volumes, VSERVER_DST.
In addition, a single CLI parameter VSERVER can be used where the source and
destination share the same value. These CLI parameters exist in a 'chain'
property. The following is an example of the chaining algorithm:

VSERVER => VSERVER_SRC => (VSERVER_DST, VSERVER_LS)

The first CLI parameter VSERVER is a single value parameter, the others are
comma separated list of entries.

A special relationship exists between the VSERVER and the first entry of
VSERVER_SRC.  Only one needs to be specified, and if one is not specified
it will take the value of the other.

And if VSERVER_DST or VSERVER_LS are not set than it takes its value from the
corresponding entries of VSERVER_SRC. And if there is no corresponding entry in
VSERVER_SRC, it will take it's value from VSERVER.

This 'chaining' logic is provided once at CLI processing, and is not carried out
if their values are later explicitly changed within a script using the 'param'
method, e.g. changing VOL_SIZE within the script will not propagate the change
to VOL_SRC_SIZE or VOL_DST_SIZE.

The following are the CLI parameter 'sets' that are provided in this manner:

B<FILER> => B<FILER_SRC> => (B<FILER_DST>, B<FILER_LS>)

Filer or Node name of a filer in the Cluster where the volume will be
created/located.

B<VSERVER> => B<VSERVER_SRC> => (B<VSERVER_DST>, B<VSERVER_LS>)

Vserver name of the volume.

B<FILER_AGGR> => B<FILER_SRC_AGGR> => (B<FILER_DST_AGGR>, B<FILER_LS_AGGR>)

Specifies the name of an aggregate to create/locate the volume on.

B<VOL_NAME> => B<VOL_SRC_NAME>
B<VOL_DST_NAME>
B<VOL_LS_NAME>

Specifies the name of the volume if it exists.  Since it does not make sense
that the same volume name be given to source, destination and ls volumes,
so for this particular parameter the 'chain' logic describe above does not
apply between VOL_SRC_NAME, VOL_DST_NAME and VOL_LS_NAME.

B<VOL_SIZE> => B<VOL_SRC_SIZE> => (B<VOL_DST_SIZE>, B<VOL_LS_SIZE>)

The size of the volume to be created.  A unit, m for MB, g for GB etc can be
used to specify the size required. It is possible to specify a number followed
by any combination of k, m or g followed with an optional b, e.g. 10g, 10GB
or 10G, are all allowed. This will then be converted into a numberic
value in units of KB.

Default: 20g

B<VOL_TYPE> => B<VOL_SRC_TYPE> => (B<VOL_DST_TYPE>, B<VOL_LS_NAME>)

The type of the volume to create.  Valid values are 'flex', 'coral' or 'acro'.
Default: flex

B<STRIPE_NODE> => B<STRIPE_SRC_NODE> => (B<STRIPE_DST_NODE>, B<STRIPE_LS_NODE>)

An integer or a comma separated list of nodes on the cluster used to create
member aggregates of a stripe aggregate.  Default: No. of nodes in cluster.

B<STRIPE_COUNT> => B<STRIPE_SRC_COUNT> => (B<STRIPE_DST_COUNT>, B<STRIPE_LS_COUNT>)

The number of volume member when creating a striped volumed.  Default: 1
volume member per aggregate member

B<STRIPE_WIDTH> => B<STRIPE_SRC_WIDTH> => (B<STRIPE_DST_WIDTH>, B<STRIPE_LS_WIDTH>)

The width of each stripe of a stripe volume.  Valid range is between 256K and
1G(inclusive).  Default: 64M

B<Stand alone CLI parameters:>

B<CLIENT_LIST>

Optional.  A comma separated list of client name to be used to generate data.
NFS data will be generated for UNIX/LINUX clients, and CIFS data for Windows
clients.  Default: localhost

If a Windows client is used to generate CIFS data, it is also possible to set the
domain name, username and password using the CLI parameters CIFS_DOMAIN, CIFS_USER and CIFS_PASSWORD.

B<DATASET>

Optional. Indicate how data will be generated for a volume using the
B<populate> method. Valid values are C<DS-tiny>, C<DS-small>, C<DS-medium> or
C<DS-populate>, C<DS-large> and C<< DS-<n>G >>:

=over 4

=item DS-tiny

Generates a dataset roughly 20MB to 27MB in size, for cases such as simulators
where less disk space is available or situations where only a very small
amount of data is needed.

=item DS-small

Generates a dataset in the range of 99MB to 145MB, for cases where a small
dataset is desired.

=item DS-medium

=item DS-populate

Generates a dataset in the range of 484MB to 585MB, for general test
cases. This is the default dataset size, and is used when no C<DATASET>
argument is given.

The C<DS-populate> label is an alias for C<DS-medium>, to maintain
backwards-compatibility.

=item DS-large

Generates a dataset in the range of 3.98GB to 4.37GB, for large-scale dataset
tests.

=item DS-<n>G

Generates an arbitrary dataset of roughly I<n> GB. The same number of files
and directories are created for this dataset as are for C<DS-large> and
C<DS-medium>, the difference in dataset size is managed by adjusting the size
of the files themselves. Thus the deviance in dataset size is a very small
percentage of the overall average size.

=back

Default: DS-populate

B<DATASET_SEED>

Optional. Specifies an explicit seed for the random number generator used by
DataManip or RichData. This allows the user to exactly replicate a dataset
from a previous run, as all "randomness" will be recreated. The value must be
a positive integer

Note: the RichData package does not have a parameter to control the seed, the
way DataManip does. Instead, it looks for a global variable called
C<$main::DATA_SEED> Thus, in order for this parameter to affect RichData it has
to set this value. This means that setting this parameter has the direct effect
of setting a variable in B<main::>.

B<BRE>

Optional. Indicates whether to use the Block Replication Engine or the Logical Replication
Engine.

1 forces the cluster to use the Block Replication Engine.
0 forces the cluster to use the Logical Replication Engine.

B<THROTTLE>

Optional. Sets the throttle speed of a snapmirror relationship in a cross-cluster configuration. Takes the parameter as an integer and sets the throttle to that integer value.

B<USE_SE>

Optional. Controls whether volumes created by B<create_volume> will have
Storage Efficiency (dedupe) activated on them as part of the creation/setup
of the volume. The value must be one of:

=over

=item C<dedupe>

Enables deduplication only.

=item C<compression>

Enables deduplication and background compression.

=item C<inline-compression>

Enables deduplication and background compression, and also switches on
inline compression.

=back

Compression cannot be enabled unless deduplication is, as well. Likewise,
inline compression cannot be enabled unless compression (and therefore
dedupe) is as well. Thus there is no need to combine options, as each one
implies all the ones before it.

B<START_SE>

Optional. Controls whether an initial Storage Efficiency scan is triggered
for volumes that have SE activated on them (due to the B<USE_SE> parameter,
previous). Triggers a call to the B<efficiency_start> method of
B<GNOTE::GX::Volume>. The value must be one of the following:

=over

=item C<dedupe>

The initial scan does deduplication checking.

=item C<compression>

The initial scan does compression.

=item C<all>

The initial scan does both compression and deduplication.

=back

There is no distinction between background compression and inline compression
from the perspective of a SE scan.

Note that B<START_SE> is meaningless if B<USE_SE> is not also given. Thus,
it is ignored when B<USE_SE> is not set.

B<DEDUPE_DATASET>

If set to a true value, tells the B<volume_create> method to pass a C<SCRAMBLE>
parameter to the code that generates datasets for pre-populating volumes. This
creates datasets that are better set up for testing Storage Efficiency. See
also the documentation for B<volume_create> to see how to enable this on a
per-volume basis. Note that creating dedupe datasets requires using the
B<DataManip> (see L<the DataManip documentation|DataManip>) data-population
tool, which may require passing C<DATAMANIP=1> to B<ntest>. Dedupe-style
datasets cannot be created on CIFS shares.

Default: 0

B<SCRAMBLE>

When creating dedupe-style datasets, the C<SCRAMBLE> parameter determines the
randomness of the data. A value less than 1 means that percentage chance of a
single byte in a filesystem block being randomized (i.e., 0.20 means a 20%
chance of one random byte). A value greater than 1 means that many bytes in the
filesystem block always being randomized. This parameter is only recognized
when the dataset is created by the B<DataManip> tool, and is only passed along
when B<DEDUPE_DATASET> is set to a true value, or when a volume is explicitly
created with dedupe enabled via parameters (see the section for
B<volume_create>, below). B<volume_create> also has a parameter that can be
used to override this value, if there is a need for multiple volumes with
different levels of randomness.

Default: 0.20

=cut

## no critic(RequireFilenameMatchesPackage)
package NOTE::SnapMirrorUtil;

use strict;
use warnings;

use POSIX qw();
use Carp qw(confess);
use Data::Dumper;
use Time::HiRes qw(tv_interval gettimeofday);

use Tharn;
use Hostrec;
use GNOTE;
use Options;

my $util = undef;

our @ISA = qw(NOTE);
my $OBJ     = "";
my $options = "";

=head2 ::new

Creates a NOTE::SnapMirrorUtil object and returns it. One of the purpose of
this object is to process CLI parameters for SnapMirror tests.

=over

=item Arguments

direct> => I<1|0>

Invoke in 'direct' command execution mode, i.e. call 'set_execution_server' per node. Default: 1

=item Examples

=back

    my $util = new SnapMirrorUtil();

=cut

sub new () {

    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    return $util
      if $util;

    my $self = {};
    bless $self, $class;

    $util = $self;    # Force all to use the same 'util' object

    $options = new Options;

    $options->parameters(
        definitions => [
            {
                name        => 'FILER',
                description => 'Name of the Filer or Node to which the connection will be made initially',
            },
            {
                name        => 'FILER_SRC',
                description => 'Comma separated list of the SRC filers. Default: FILER'
            },
            {
                name        => 'FILER_DST',
                description => 'Comma separared list of the DST filers. Defaults: FILER'
            },
            {
                name        => 'FILER_LS',
                description => 'Comma separated list of LS filers. Default: FILER'
            },
            {
                name        => 'VSERVER',
                description => 'Name of the VSERVER used for creating volumes and SnapMirror relationships'
            },
            {
                name        => 'VSERVER_SRC',
                description => 'Comma separated list of source vservers. Default: VSERVER'
            },
            {
                name        => 'VSERVER_DST',
                description => 'Comma separated list of destination vservers. Default: VSERVER_SRC'
            },
            {
                name        => 'VSERVER_LS',
                description => 'Comma separated list of LS vservers. Default: VSERVER'
            },
            {
                name        => 'FILER_AGGR',
                description => 'Name of the aggregate to be used'
            },
            {
                name        => 'FILER_SRC_AGGR',
                description => 'Name of the source filer aggregate. Default: FILER_AGGR'
            },
            {
                name        => 'FILER_DST_AGGR',
                description => 'Name of the destination filer aggregate. Default: FILER_SRC_AGGR'
            },
            {
                name        => 'FILER_LS_AGGR',
                description => 'Name of the LS filer aggregate. Default: FILER_SRC_AGGR'
            },
            {
                name        => 'VOL_NAME',
                description => 'Name of the volume that NOTE object creates or attaches to'
            },
            {
                name        => 'VOL_SRC_NAME',
                description => 'Name of the source volume name. Default: VOL_NAME'
            },
            {
                name        => 'VOL_DST_NAME',
                description => 'Name of the destination volume name.'
            },
            {
                name        => 'VOL_LS_NAME',
                description => 'Name of the LS  volume name.'
            },
            {
                name        => 'VOL_SRC_SIZE',
                description => 'The size of source volume to be created. Default: VOL_SIZE'
            },
            {
                name        => 'VOL_DST_SIZE',
                description => 'The size of destination volume to be created. Default: VOL_SRC_SIZE'
            },
            {
                name        => 'VOL_LS_SIZE',
                description => 'The size of destination volume to be created. Default: VOL_SRC_SIZE'
            },
            {
                name        => 'VOL_TYPE',
                description => 'Type of volume that needs to be created. Default: flex',
                default     => 'flex'
            },
            {
                name        => 'VOL_SRC_TYPE',
                description => 'Type of the source volume. Default: VOL_TYPE'
            },
            {
                name        => 'VOL_DST_TYPE',
                description => 'Type of the destination volume Default: VOL_SRC_TYPE'
            },
            {
                name        => 'VOL_LS_TYPE',
                description => 'Type of the LS volume Default: VOL_SRC_TYPE'
            },
            {
                name        => 'STRIPE_NODE',
                description => 'integer count or Comma separated list of Node(s) to be use for creating striped aggregates. Default: # nodes in cluster'
            },
            {
                name        => 'STRIPE_SRC_NODE',
                description => 'Comma separated list of Name(s) of Node(s), on SRC Cluster, to be use for creating striped aggregates. Default: STRIPE_NODE'
            },
            {
                name        => 'STRIPE_DST_NODE',
                description => 'Comma separated list of Name(s) of Node(s), on DST Cluster, to be use for creating striped aggregates. Default: STRIPE_SRC_NODE'
            },
            {
                name        => 'STRIPE_LS_NODE',
                description => 'Comma separated list of Name(s) of Node(s), on LS Cluster, to be use for creating striped aggregates. Default: STRIPE_SRC_NODE'
            },
            {
                name        => 'STRIPE_COUNT',
                description => 'Sets the number of volume member when creating a coral or acro source volume. Default: 1 per Aggregate member'
            },
            {
                name        => 'STRIPE_SRC_COUNT',
                description => 'Sets the number of volume member when creating a coral or acro source volume. Default: STRIPE_COUNT'
            },
            {
                name        => 'STRIPE_DST_COUNT',
                description => 'Sets the number of volume member when creating a coral or acro destination volume. Default: STRIPE_SRC_COUNT'
            },
            {
                name        => 'STRIPE_LS_COUNT',
                description => 'Sets the number of volume member when creating a coral or acro LS volume. Default: STRIPE_SRC_COUNT'
            },
            {
                name        => 'STRIPE_WIDTH',
                description => 'Sets STRIPE_WIDTH stripes for a coral or acro source volume. The valid range is between 256K and 1G(inclusive). Default: 64MB',
                default     => '64MB',
            },
            {
                name        => 'STRIPE_SRC_WIDTH',
                description => 'Sets STRIPE_WIDTH stripes for a coral or acro source volume. The valid range is between 256K and 1G(inclusive). Default: STRIPE_WIDTH'
            },
            {
                name        => 'STRIPE_DST_WIDTH',
                description => 'Sets STRIPE_WIDTH stripes for a coral or acro destination volume. The valid range is between 256K and 1G(inclusive). Default: STRIPE_'
            },
            {
                name        => 'STRIPE_LS_WIDTH',
                description => 'Sets STRIPE_WIDTH stripes for a coral or acro LS volume. The valid range is between 256K and 1G(inclusive). Default: STRIPE_'
            },
            {
                name        => 'CLIENT_LIST',
                description => 'List of clients to use for population. Default: localhost',
                default     => "localhost",
            },
            {
                name        => 'DATASET',
                description => 'Indicate the size of the dataset to create when populating a volume. Valid values are DS-tiny, DS-small, DS-medium (a.k.a. DS-populate) and DS-large. Default: DS-populate',
                default     => 'DS-populate',
            },
            {
                name        => 'DATASET_SEED',
                description => 'An explicit random-number seed for DataManip and/or RichData to use, instead of their defaults.',
            },
            {
                name        => 'JOB_TIMEOUT',
                description => 'Timeout in seconds for jobs. Default: 3600 (i.e. 1 hour)',
                default     => 3600
            },
            {
                name        => 'LOCAL_STORAGE',
                description => 'Use storage local to node when creating or considering new flex storage (volume or aggregate). Default: 1',
                default     => 1
            },
            {
                name        => 'RUN_CHECK',
                description => 'Determine whether consistency check functions should be invoked or not, prior to finishing the test. Default: 0 (i.e. not running)',
                default     => 0
            },
            {
                name        => 'THROTTLE',
                description => 'Apply mentioned throttle speed to the snapmirror relationships created by this test in cross-cluster configuration',
            },
            {
                name        => 'BRE',
                description => 'Determine whether the Block Replication Engine should be used as the transfer engine instead of the Logical Replication Engine',
            },
            {
                name        => 'USE_SE',
                description => 'Use Storage Efficiency engine on new volumes created. Value should be one of dedupe, compression or inline-compression.',
            },
            {
                name        => 'START_SE',
                description => 'Start an initial Storage Efficiency scan after activating SE on a new volume. Value should be one of dedupe, compression or all.',
            },
            {
                name        => 'DEDUPE_DATASET',
                description => 'Define whether new volumes being pre-populated should be populated with data tuned for Storage Efficiency testing. If set to a non-zero value, new volumes that populate will be passed the value of the SCRAMBLE parameter when data is generated by DataManip.',
                default     => 0,
            },
            {
                name        => 'SCRAMBLE',
                description => 'Scramble-factor (randomness) used when creating populate data-sets with DataManip. Value defaults to 0.20. Only applied when populating for dedupe data-sets (when DEDUPE_DATASET is set, or when a volume is specifically created to have dedupe data).',
                default     => '0.20',
            },
            {
                name        => 'SNAPMIRROR_7G_PATH',
                description => 'Enable 7g style option testing of snapmirror commands and provide the volume path description ([cluster:][//vserver/]volume)',
            },
            {
                name        => 'SNAPMIRROR_CLUSTER_PATH',
                description => 'Enable cluster style option testing of snapmirror commands and provide the volume path description ([cluster:][//vserver/]volume)',
            },
	    {
		name	    => 'MIRROR_TYPE',
		description => 'Default snapmirror type, one of DP/XDP/LS',
		default	    => undef,
	    },
       ],
        require_one_of => [ [ "FILER", "FILER_SRC" ], [ "VSERVER", "VSERVER_SRC" ] ]
    );

    $options->set_default( "FIND_ALL_NODES"  => 1 );
    $options->set_default( "NODES_ON_DEMAND" => 1 );

    # set the defaults we want for some options that are pulled in from other libraries
    $options->set_default( "VOL_SIZE" => "20g" );
    $options->process();

    $self->{_params} = $options->get_hash();

    # Transfer all CLI parameters to internal table
    foreach my $name ( keys %{ $self->{_params} } ) {
        $self->{ "_" . $name } = $self->_get_scalar($name);
    }

    $args{direct} = 1 unless defined $args{direct};

    $self->_set_group( "FILER",        "FILER_SRC",        "FILER_DST",        "FILER_LS" );
    $self->_set_group( "VSERVER",      "VSERVER_SRC",      "VSERVER_DST",      "VSERVER_LS" );
    $self->_set_group( "FILER_AGGR",   "FILER_SRC_AGGR",   "FILER_DST_AGGR",   "FILER_LS_AGGR" );
    $self->_set_group( "VOL_SIZE",     "VOL_SRC_SIZE",     "VOL_DST_SIZE",     "VOL_LS_SIZE" );
    $self->_set_group( "VOL_TYPE",     "VOL_SRC_TYPE",     "VOL_DST_TYPE",     "VOL_LS_TYPE" );
    $self->_set_group( "STRIPE_NODE",  "STRIPE_SRC_NODE",  "STRIPE_DST_NODE",  "STRIPE_LS_NODE" );
    $self->_set_group( "STRIPE_COUNT", "STRIPE_SRC_COUNT", "STRIPE_DST_COUNT", "STRIPE_LS_COUNT" );
    $self->_set_group( "STRIPE_WIDTH", "STRIPE_SRC_WIDTH", "STRIPE_DST_WIDTH", "STRIPE_LS_WIDTH" );

    $self->_set_group( "VOL_NAME", "VOL_SRC_NAME" );
    $self->{"_VOL_DST_NAME"} = $self->_get_scalar("VOL_DST_NAME");
    $self->{"_VOL_LS_NAME"}  = $self->_get_scalar("VOL_LS_NAME");

    $self->{_MIRROR_TYPE} = $self->_get_scalar("MIRROR_TYPE");
    $self->{_CLIENT_LIST} = $self->_get_scalar("CLIENT_LIST");

    # In case FILER_LS is provided and FILER_LS_AGGR is not provided,
    # FILER_LS_AGGR gets defaulted to FILER_SRC_AGGR/FILER_AGGR, in which case
    # the LS volumes will be created on FILER_SRC_AGGR/FILER_AGGR,
    # which means ls volumes being created on src node.
    # To avoid this, make FILER_LS_AGGR to '',
    # if FILER_LS is not same as FILER_SRC/FILER. so that when ls volumes are created,
    # an aggregate on the node specified by FILER_LS will be created,
    # which makes ls volumes being created on correct node specified by FILER_LS
    if( ($self->{_params}->{'FILER_LS'}) and (! $self->{_params}->{'FILER_LS_AGGR'}) ) {
        if( ($self->{_params}->{'FILER_SRC'}) and ($self->{_params}->{'FILER_LS'} ne $self->{_params}->{'FILER_SRC'}) ) {
            $self->{'_FILER_LS_AGGR'} = q{};
        }
        elsif( ($self->{_params}->{'FILER'}) and ($self->{_params}->{'FILER_LS'} ne $self->{_params}->{'FILER'}) ){
            $self->{'_FILER_LS_AGGR'} = q{};
        }
    }

    # Dummy connect to grap all nodes in case it is a multiple cluster test
    # A work-around for Server
    my %filer_list;
    map { $filer_list{$_} = 1 } split ",", $self->{_FILER_SRC} . "," . $self->{_FILER_DST} . "," . $self->{_FILER_LS};
    my $node_list = ( join ",", keys %filer_list );
    NOTE::logcomment("Creating Server object with nodes: $node_list");
    Server->new( nodes => $node_list );

    NOTE::logcomment("Instantiate a dummmy NOTE::GX::Node object to enabled 'set_execution_server'");
    NOTE::GX::Node->connect( name => $self->param("FILER_SRC:0"), direct => $args{direct} );    # Dummy connect to enable direct node execution
                                                                                                # Check for unique cluster name
    $self->_check_clusters(%filer_list);

    $self->_set_default_stripe_info( "FILER:0",     "VSERVER:0",     "_STRIPE_NODE",     "_STRIPE_COUNT" );
    $self->_set_default_stripe_info( "FILER_SRC:0", "VSERVER_SRC:0", "_STRIPE_SRC_NODE", "_STRIPE_SRC_COUNT" );
    $self->_set_default_stripe_info( "FILER_DST:0", "VSERVER_DST:0", "_STRIPE_DST_NODE", "_STRIPE_DST_COUNT" );
    $self->_set_default_stripe_info( "FILER_LS:0",  "VSERVER_LS:0",  "_STRIPE_LS_NODE",  "_STRIPE_LS_COUNT" );

    foreach my $key ( "_VOL_SIZE", "_VOL_SRC_SIZE", "_VOL_DST_SIZE", "_VOL_LS_SIZE" ) {
        my $size = $self->{$key};
        if ( $size =~ /(\d+)([kmg])(b)*/i ) {
            $self->{$key} = NOTE::units( $1 . lc($2) );
        }
        else {
            $self->{$key} = int( $self->{$key} / 1024 );
        }
    }

    if ( defined $self->{_BRE} ) {
        $self->_set_BRE;
        $options->print;
    }

    if (exists($util->{_SNAPMIRROR_7G_PATH}) &&
        exists($util->{_SNAPMIRROR_CLUSTER_PATH})) {
        confess 'Invalid parameter combination: SNAPMIRROR_7G_PATH and SNAPMIRROR_CLUSTER_PATH cannot be set at the same time';
    }

    $self->{_SRC_COUNT} = 0;
    $self->{_DST_COUNT} = 0;

    return $self;
}

# Make sure all unique cluster entries have a unique cluster names

sub _check_clusters {
    my ( $self, %nodes ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list = keys %nodes;
    return unless @list > 1;    # must be unique if only 1 node

    NOTE::logcomment("Verify unique cluster names:");
    my ( %clist, %cname );
    foreach (@list) {
        my $n = NOTE::GX::Node->connect( name => $_ );
        my $c = NOTE::GX::Cluster->connect( Node => $n );
        NOTE::logcomment( 'Checking Node ' . $n->name . ' in cluster ' . $c->description );
        unless ( exists $clist{$c} ) {
            confess "\nDuplicate Cluster Name found:\n" . ' Node ' . $cname{ $c->name } . ': ' . $c->description . "\n" . ' Node ' . $n->name . ': ' . $c->description
              if exists $cname{ $c->name };

            $clist{$c} = $c->name;
            $cname{ $c->name } = $n->name;
        }
    }
}

# Sets the default aggregate member node(s) from a stripe aggregate

sub _set_default_stripe_info {
    my ( $self, $filer, $vserver, $node_key, $cnt_key ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @nodes;
    unless ( $self->{$node_key} ) {
        my $vsrv = NOTE::GX::Vserver->connect(
            name => $self->param($vserver),
            Node => $self->param($filer)
        );

        # Pick all healthy nodes from the cluster.
        @nodes = sort $vsrv->Cluster->nodes(
            health      => 'true',    # Pick only good ones
            eligibility => 'true',
        );
        $self->{$node_key} = join ",", @nodes;
    }
    unless ( $self->{$cnt_key} ) {
        if ( $self->{$node_key} =~ /^\d+$/ ) {
            $self->{$cnt_key} = $self->{$node_key};
        }
        else {
            $self->{$cnt_key} = @nodes;
        }
    }
}

# Provide the processing of the set of 4 parameters, such as FILER, FILER_SRC and FILER_DST, FILER_LS etc

sub _set_group {
    my ( $self, $first, $second, $third, $fourth ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $first_key  = "_" . $first;
    my $second_key = "_" . $second;

    my $first_value = $self->_get_scalar($first);
    my $sec_value   = $self->_get_scalar($second);

    # First value must be a single entry, either from CLI or first entry of second value
    if ($first_value) {
        if ( $first_value =~ /,/ ) {
            NOTE::logcomment("WARNING: Comma not allowed for $first: $first_value. Using first entry only");
            ($first_value) = split ",", $first_value;
        }
    } elsif ($sec_value) {
        # use first entry of sec_value
        ( $self->{$first_key} ) = split ",", $sec_value;
    }
    confess "One of $first or $second must be specified"
      unless $first;

    $self->{$second_key} = $sec_value ? $sec_value : $self->{$first_key};
    $self->_fill_missing( $first_key, $second_key );

    $self->_set_final( $first, $second, $third )
      if defined $third;
    $self->_set_final( $first, $second, $fourth )
      if defined $fourth;
}

# Expand last parameter using first 2 parameter

sub _set_final {
    my ( $self, $first, $second, $third ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $first_key  = "_" . $first;
    my $second_key = "_" . $second;
    my $third_key  = "_" . $third;

    my $third_value = $self->_get_scalar($third);
    $self->{$third_key} = $third_value ? $third_value : $self->{$second_key};
    $self->_fill_missing( $second_key, $third_key );
}

# Routine to add default to missing entries in comma separated list from a src list

sub _fill_missing {
    my ( $self, $src_key, $dst_key ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @src = split /,/ => ($self->{$src_key} || q{});
    my @dst = split /,/ => ($self->{$dst_key} || q{});
    my $idx = 0;
    while ( $idx < @dst ) {
        $dst[$idx] = ( $src[$idx] ? $src[$idx] : $src[0] )
            unless $dst[$idx];
        $idx++;
    }
    $self->{$dst_key} = join q{,} => @dst;
}

# Translate option parameters into a string

sub _get_scalar {
    my ( $self, $name ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $tmp = $self->{_params}->{$name};
    if ( ref($tmp) =~ /ARRAY/ ) {
        return join ",", @{$tmp};
    }
    else {
        return $tmp;
    }
}

=head2 ::param

Method to set or return a CLI parameter value.
In the event where a new value is passed, this method will return the old
value of the specified CLI paramter.

=over

=item Arguments

B<name> => I<string[:index]>

Required: The CLI parameter current value to request for.
'string' is the CLI parameter, with an optional [:index].

In the event that the CLI parameter is an array of comma separated
list of values, the option 'index' is an integer used to specify the
ith entry in the array, with the first entry designated by zero.
The value of 'index' defaults to zero.

To obtain the entire command separated list, set 'index' to asterisk, '*'.

If 'index' is greater then the list of entries in the array, then a 'round robin'
algorithm will be use to find the corresponding entry.

B<value> => I<scalar>

Optional. The new value to set the CLI parameter to.

=item Examples

=back

    # Returning the current value of DATASET parameter:
    my $value = $util->param('DATASET');

    # Returning the current, or old, value of VOL_SIZE parameter and setting
    # a new one:
    my $old_value = $util->param('VOL_SRC_SIZE','30g');

    # Return the 5th entry of a comma separated list
    my $entry = $util->param("FILER_DST:4");

    # Return all entries of the list
    my $list = $util->param("FILER_LS:*");

=cut

sub param {
    my ( $self, $name, $value ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $name ) {
        my ( $cli, $index ) = split ":", $name;
        $index = 0 unless defined $index;

        my $key       = "_" . $cli;
        my $old_value = "";

        if ( $index =~ /\*/ ) {

            # Return entire list
            $old_value = $self->{$key};
        }
        else {

            # Return one entry
            my @list = split ",", $self->{$key};

            if (@list) {
                $index     = $index % @list;
                $old_value = $list[$index];
            }
            else {
                confess "Unknown CLI parameter $name"
                  unless exists $self->{$key};
            }
            if ( defined $value ) {

                # Compose new value(s)
                $list[$index] = $value;
                $value = join ",", @list;
            }
        }

        # Replace value if passed and return old value
        $self->{$key} = $value if defined $value;
        return $old_value;
    }
    else {
        confess "name parameter missing";
    }
}

=head2 ::set_mode

Sets privilege mode for a Node or Nodes.
If Node is not specified, sets privilege for all FILER_SRC, FILER_DST and FILER_LS.

=over

=item Argument

B<mode> => I<string>

Mode to set filer(s) to.

B<node_list> => I<comma separated list of node names>

I<Optional>

Indicates which node's privilege should be changed. Default: FILER_SRC, FILER_DST and FILER_LS

=item Example

=back

    $util->set_mode("test");                    # Sets test mode all known nodes
    $util->set_mode("diag", "dpg_gx4,dpg-gx5"); # Sets diag mode for dpg-gx4 and dpg-gx5

=cut

sub set_mode {
    my ( $self, $mode, $node_list ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $node_list = $self->{_FILER_SRC} . "," . $self->{_FILER_DST} . "," . $self->{_FILER_LS}
      unless defined $node_list;
    my %list = map { $_ => 1 } split ",", $node_list;
    for ( keys %list ) {
        my $filer = NOTE::GX::Node->connect($_);
        $filer->server->set( privilege => $mode, confirmations => "off" );
    }
}

=head2 ::create_volume

Method to create and return an array of standard volume(s). The 'tag' parameter
determines category of volume. The value of 'tag' will determine which CLI
parameter set to use to determine volume characteristics, e.g. 'src' means
to use FILER_SRC, VOL_SRC_TYPE, VOL_SRC_SIZE etc to be used to specify volume
attributes. All these parameters can be overridden by passing additional
parameters using key-value pair to this method.

The filer CLI parameters will, by default, determine node the volume will be
created on using a round robin algorithm. So with the right combination
specification of CLI parameters, any volume configurations should
be possible. See examples below.

=over

=item Arguments

B<tag> => I<src|dst|ls>

Used to indicate which type of snapmirror volumes to create.
src means a snapmirror source volume with type of 'RW".
dst means a snapmirror DP destination volume with a type of 'DP'.
ls means a snapmirror LS destination volume with a type of 'DP'.

%args contains override parameters to 'volume create'

B<populate> => I<boolean|dataset>

Optional. If set to 1, indicates that the volume will be pre-populated with data
using specifications from DATASET and CLIENT_LIST CLI parameters.

The parameter can also be set to a dataset value to cause pre-population using
the dataset size specified.

B<scramble>

Optional. If set to 1, indicates that the dataset created when populating the
volume will have some random "scrambling" done to the blocks of data, in order
to tune it for testing with Storage Efficiency functionality.  The
B<DEDUPE_DATASET> CLI parameter can force this, if it is set to a true value.

B<scramble_value>

If set, the value of this parameter is used as the "scramble" setting when
the B<scramble> parameter (above) is set to 1. If not passed, the default
value is 0.20. See L<the DataManip documentation|DataManip> for explanation of
the C<SCRAMBLE> parameter to the B<populate_tree()> function.

The B<SCRAMBLE> CLI parameter provides an alternate default to use instead of
0.20 when creating a storage-efficiency-enabled dataset. However, this
parameter will override the value of the CLI parameter, if present.

=item Examples

=back

    # Uses only CLI parameter values to set volume attributes
    my ($vol_obj) = $util->create_volume(tag=>'src');

    # Uses CLI parameters and override the size explicitly to 500G
    my @big_vol = $paloma->create_volume(tag=>'dst', size=>'500G');

    my %vol_opts = (
             type => 'acro',
             count => 3,
             Aggregate => {
                    nodes => 'dpg-gx23,dpg-gx24',
                    raidtype => 'raid4'
                    }
             );
    # Overrides CLI parameters values with then set by %vol_opts above
    my @vol_obj_2 = $util->create_volume(tag=>'src', %vol_opts);

    # The following will create 1 src on each of 2 nodes
    # when CLI parameter is set to: FILER_SRC=dpg-gx4,dpg-gx5
    my @src_list = $utl->create_volume(tag=>"src", count=>2);

    # Similiarly with dst volumes use: FILER_DST=dpg-gx4,dpg-gx5
    my @dst_list = $utl->create_volume(tag=>"dst", count=>2);

=cut

sub create_volume {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $tag = uc( $args{tag} );
    confess "Invalid tag: " . $args{tag} . ". Possible values: src or dst"
      unless $tag =~ /^(SRC|DST|LS)$/;
    my $type = $args{type} || $self->{"_VOL_${tag}_TYPE"};

    # see BURT 462630, change id 1217215, back out after real fix!
    if ( $type !~ /^(flex|coral|acro)$/ ) {
        $type = 'flex';
        logresult('WARN', 'BURT 462630, is not fixed. See BURT for details' .
                  ' and backout change 1217215 when it gets fixed');
    }

    confess "Invalid count: " . $args{count}
      if exists $args{count} && $args{count} < 0;

    my $have_vserver = exists $args{Vserver};
    my $node         = $args{node} || $args{Filer};
    my $count        = ( exists $args{count} ? delete $args{count} : 1 );

    # Set Volume attributes
    $args{name} = $self->{"_VOL_${tag}_NAME"} unless exists $args{name};
    $args{type} = $type;
    $args{'mirror-type'} = $args{'mirror-type'} || ( $tag =~ /SRC/ ? "RW" : "DP" );
    $args{size} = $self->{"_VOL_${tag}_SIZE"} unless exists $args{size};
    $args{local_storage} = $self->{"_LOCAL_STORAGE"}
      unless exists $args{local_storage};
    if ( $args{type} =~ /flex/ ) {
        $args{Aggregate}->{type} = "flex"
          unless ref( $args{Aggregate} ) =~ /NOTE::.*Aggregate/;
    }
    else {
        if ( $args{type} =~ /coral/ ) {
            $args{'stripe-format'} = 'data';
        }
        elsif ( $args{type} =~ /acro/ ) {
            $args{'stripe-format'} = 'full';
        }
        $args{type} = 'striped';
    }

    # Turn off hourly snapshot generation
    $args{'snapshot-policy'} = 'none' unless exists $args{'snapshot-policy'};

    unless ( ref $args{Aggregate} =~ /NOTE::.*Aggregate/ ) {

        # Set Aggregate attributes
        $args{Aggregate}->{name} = $self->{"_FILER_${tag}_AGGR"}
          unless exists $args{Aggregate}->{name};
        $args{Aggregate}->{type} = $args{type}
          unless exists $args{Aggregate}->{type};
        $args{Aggregate}->{nodes} = $self->{"_STRIPE_${tag}_NODE"}
          if $args{Aggregate}->{type} =~ /striped/
              && !exists $args{Aggregate}->{nodes};
        $args{Aggregate}->{local_storage} = $self->{"_LOCAL_STORAGE"}
          unless exists $args{Aggregate}->{local_storage};
    }

    if ( $args{type} =~ /striped/ ) {

        # Complete stripe volume spec
        $args{'stripe-count'} = $self->{"_STRIPE_${tag}_COUNT"}
          unless exists $args{'stripe-count'};
        $args{'stripe-width'} = $self->{"_STRIPE_${tag}_WIDTH"}
          unless exists $args{'stripe-width'};
    }

    if ( $args{populate} ) {
        # If $args{populate} looks like a dataset spec, move it to $dataset
        # and change $args{populate} to 1 for the next if-test:
        my $dataset = $self->{_DATASET}; # Start with the default value
        if ($args{populate} =~ /^DS-(tiny|small|medium|populate|large|\d+G)/x) {
            $dataset = $args{populate};
            $args{populate} = 1;
        }

        # Note that this is a string-compare, to avoid "non-numeric" warning
        if ($args{populate} eq '1') {
            # When $args{populate} is "1", we're setting up for an internal
            # eval of a call to _populate(). This gets eventually eval'd in
            # the _init() method of GNOTE::Volume.pm (package NOTE::Volume).
            # We have to pass any values gleaned from CLI or $self in the call
            # we set up, as when _populate() is eventually called it won't
            # have access to $self or any of the CLI, as it will be on the
            # client (which may or may not be "localhost").

            # Initialize the _populate() args-list to the contents of
            # CLIENT_LIST and the $dataset value we derived above (either
            # from populate arg, DATASET CLI param or from default):
            my $p_args = qq(client_list => "$self->{_CLIENT_LIST}", ) .
                qq(dataset => "$dataset");

            # Add a "scramble" parameter if dedupe/scrambling was requested
            if ($args{scramble} || $self->{_DEDUPE_DATASET}) {
                # The value for the scramble parameter is first "scramble_value"
                # parameter passed here, then any command-line value, then
                # defaults to 0.20.
                my $scramble_value =
                    $args{scramble_value} || $self->{_SCRAMBLE} || '0.20';
                $p_args .= qq(, scramble => "$scramble_value");
            }

            # Add the "seed" parameter if DATASET_SEED was given
            if (my $seed = $self->{_DATASET_SEED}) {
                confess "DATASET_SEED value must be an integer ($seed)"
                    if ($seed !~ /^\d+$/);
                $p_args .= ", seed => $seed";
            }

            $args{dataset} = "NOTE::SnapMirrorUtil::_populate(shift, $p_args)";
        } else {
            # This assumes that whatever was passed in $args{populate} can be
            # eval'd later in NOTE::Volume::_init(). If the user passed in
            # something that fails eval, they'll have to figure it out...
            $args{dataset} = $args{populate};
        }

        # We no longer need this and don't want it to propagate:
        delete $args{populate};
    }
    # Also delete these two parameters if they exist:
    delete @args{qw(dedupe scramble)};

    # create count number of volumes
    NOTE::logcomment("$count $tag $type volume(s) will be created ");
    my @vol_obj;
    foreach ( 1 .. $count ) {
        my $idx = $_ - 1;
        $args{Filer} = $args{Node} = $node || $self->param("FILER_${tag}:${idx}");
        $args{Vserver}{name} = $self->param("VSERVER_${tag}:${idx}")
          unless $have_vserver;
        $self->{"_${tag}_COUNT"} += 1;
        $args{tag} = lc($tag) . "_" . $self->{"_${tag}_COUNT"};
        my $vol = NOTE::Volume->create(%args);
        push( @vol_obj, $vol );

        # Apply Storage Efficiency if so requested:
        if ($self->{_USE_SE}) {
            # Arguments to be passed to efficiency_on() based on the actual
            # value of $self->{_USE_SE}:
            my %se_args = (
                dedupe               => [],
                compression          => [ compression => 1 ],
                'inline-compression' => [ inline => 1 ],
                # "inline" is offered as an alias/shortcut for the previous
                inline               => [ inline => 1 ],
            );
            # Is the value of $self->{_USE_SE} valid?
            confess "Invalid value for USE_SE: $self->{_USE_SE}"
                if (! $se_args{$self->{_USE_SE}});

            # If it is valid, call efficiency_on(), using the table of arguments
            # constructed above.
            $vol->efficiency_on(@{$se_args{$self->{_USE_SE}}});

            # Now that efficiency is enabled, did they also ask to start a
            # scan?
            if ($self->{_START_SE}) {
                my %start_args = (
                    dedupe      => [ dedupe => 1 ],
                    compression => [ compression => 1 ],
                    all         => [],
                );
                # Is the value of $self->{_START_SE} valid?
                confess "Invalid value for START_SE: $self->{_START_SE}"
                    if (! $start_args{$self->{_START_SE}});

                # If it is valid, call efficiency_start() with arguments from
                # the table constructed above. Also add "background" so that
                # the scan is backgrounded.
                $vol->efficiency_start(
                    background => 1,
                    @{$se_args{$self->{_USE_SE}}}
                );
            }
        }
    }

    return wantarray ? @vol_obj : $vol_obj[0];
}

=head2 ::populate

Populates the supplied volume using DataManip or RichData. The list of clients
to use for performing the population is the value of the 'client_list' key in
the args hash. The data will be created under a directory using each client's
name.  The host type of each client, either Unix or Windows, will be used to
determine whether NFS data for Unix/Linux client, or CIFS data for Windows
client, will be generated.

If no client_list key is specified the CLI parameter CLIENT_LIST will be used
to determine the client list to use to generate data.

To keep data generated for each client apart, a directory using the client
name is created to hold the generated data by that client.

In addition this method also create snapshot(s) along with the data generation.
It creates 3 snapshot for a volume object and 1 for a qtree object.

Another function of this method is to create the files required by traffic()
to run RO on a DP volume. If populate is not run, traffic must be run on
the source volume and its data updated to the DP volume.

=over

=item Arguments

B<storage> => I<NOTE storage object>

Required. A NOTE::Volume or a NOTE::Qtree object to generate data on.

B<client_list> => I<comma separated list of clients>

Optional. If not supplied then the CLI parameter CLIENT_LIST will be used.

B<dataset> => I<DS-name>

Indicates which pre-defined dataset to generate data: C<DS-populate> (default),
C<DS-tiny>, C<DS-small> or C<DS-large>. C<DS-medium> is an alias for
C<DS-populate>. You may also specify C<< DS-<n>G >> to get a dataset of I<n>
gigabytes.

Optional. Defaults to the DATASET CLI parameter.

B<%args> => I<DataManip|RichData argument Hash>

Optional. Additional data generation parameters passed directly to B<DataManip>
or B<RichData>. These parameters are not acted upon by B<populate()> itself.
The parameters can be used to fine-tune the characteristics of the dataset that
gets generated. Any parameters passed here will augment the defaults for the
given dataset.

=over

=item seed

This specifies a random-number seed to use for the random number generator
used by either B<DataManip> or B<RichData>. This allows the user to recreate
a specific dataset by giving this value explicitly. If not passed, a seed is
derived internally from the system clock and process ID value.

=item file_size_avg

The average size of created files, in bytes.

=item file_size_dev

The deviation in file size, in bytes. File size will be the value of
C<file_size_avg> plus or minus a random percentage of this value. If this value
is 0, then there will be no deviation in the size of files.

=item depth_avg

The average depth of the directory structure from the root of the tree.

=item depth_dev

The deviation in the depth, from C<depth_avg>. If this value is zero, then
the depth will always be C<depth_avg>.

=item dir_avg

The average number of subdirectories that will be created in each directory,
up until the depth (calculated via C<depth_avg> and C<depth_dev>) is reached.

=item dir_dev

The deviation in the number of subdirectories per directory. The number of
subdirectories created in a given directory will be C<dir_avg> plus or minus
a random percentage of this value, rounded down to the nearest whole
integer. If this value is 0 there will be no deviation in the number of
subdirectories.

=item file_avg

The average number of files created in each directory within the tree that
B<populate()> creates. Files are created at each level, from the root to the
last leaf of the tree.

=item file_dev

The deviation in the number of files created per directory. The number of files
created in a given directory will be C<file_avg> plus or minus a random
percentage of this value, rounded down to the nearest whole integer. If this
value is 0 there will be no deviation in the number of files.

=back

Note that parameters to B<DataManip> are all upper-case, and parameters to
B<RichData> are all lower-case. This is handled internally, you should
I<always> specify them as lower-case here.

B<RichData> supports a range of additional parameters, to control things such
as ACLs, streams, sparse files, etc. that are not supported by B<DataManip>.
For a list and explanation of these extra arguments, see the documentation for
B<RichData>. These additional parameters will be ignored by B<DataManip>.

Note: The USEDATAGEN cli parameter is supported by this call as well.
If USEDATAGEN is used, then populate() will subtest datagen.
See dp/snapmirror/paloma/datagen10.thpl for details.

=item CLI parameters

B<DATASET_SNAPSHOT> => I<name>

Name of snapshot to use for predefined dataset. If snapshot is present, then snaprestore will be called instead of actually
running populate. If snapshot is not present, then populate() will run first, then snapshot will be taken for future use.

B<USEDATAGEN> => I<1/0>

Use datagen with populate, ignores all other types of populate. Datagen will need its own parameters as well.
Check dp/snapmirror/paloma/datagen10.thpl for info on datagen's parameters.

=item Examples

=back

    $util->populate( storage => $vol);
    $util->populate(
        storage     => $vol_obj,
        client_list => 'dpg-sun1,nam21',
        %DataManipOptions,
    );

    $util->populate(storage=>$src_vol, dataset=>'DS-small');

=cut

sub populate {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $dataset_snapshot = Tharn::param('DATASET_SNAPSHOT');

    my $obj;
    if ( defined $args{storage} ) {
        $obj = delete $args{storage};
	if( $dataset_snapshot && $obj->snapshot_exists($dataset_snapshot)) {
	    NOTE::logcomment( "Found snapshot with dataset, restoring to snapshot '$dataset_snapshot'");
	    $obj->snapshot_restore($dataset_snapshot);
	    return;
	}
	if ( $args{datagen} || Tharn::param('USEDATAGEN') ) {
	    $obj->populate();	# calling populate from Volume.pm directly.
	    return;		# this call will take DATASET_SNAPSHOT if needed
	}
    } else {
        confess 'Populate: Missing NOTE storage object';
    }
    NOTE::logcomment('Populate ' . $obj->description);

    $args{client_list} ||= $self->{_CLIENT_LIST};
    $args{dataset}     ||= $self->{_DATASET};
    _populate( $obj, %args );

    if( ! ( $args{BACKGROUND} || $args{background} ) && $dataset_snapshot ) {
	NOTE::logcomment("Taking dataset snapshot '$dataset_snapshot'");
	$obj->snapshot_create($dataset_snapshot);
    }
}

# Start sio and return a flag to indicate wthether sio was started or already running, or confess for other errors
# Modified routine to just create sio data and not run sio

sub _start_sio {
    my ( $obj, %opts ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $opts{cycle} = 0 unless exists $opts{cycle};    # By default do not kick off the sio process in the Volume::sio method
    eval { $obj->sio(%opts); };
    if ($@) {
        if ( $@ =~ /sio already running/i ) {
            NOTE::logcomment("Ignore sio error: $@");
            return 0;                               # It is ok to not start sio if it is already running. Just return flag to indicate state
        }
        else {
            confess "Failed to start sio. Error: $@";    # Unexpected error
        }
    }
    return $opts{cycle};                                 # sio run state determined by 'cycle' parameter
}

# Internal function to perform the population method for each client in the
# list with the specified dataset.
#
# The data generated by this method is slightly more than what is shown for
# each data set. That is due to sio being run to create the files required
# by traffic running in RO mode. The additional space can be determined by
# looking at the _set_sio_param method.

sub _populate {
    my ( $obj, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # These are the settings for RichData/DataManip to use for each of the
    # different sizes of datasets. Any values passed in %args will be allowed
    # to override values here, but this is the set that we start with:
    my %DATASET_PARAMS = (
        # Tiny dataset, yields 20-27MB
        'DS-tiny'   => {
            file_size_avg => 16384,
            file_size_dev => 1024,
            depth_avg     => 4,
            depth_dev     => 0,
            dir_avg       => 4,
            dir_dev       => 0,
            file_avg      => 4,
            file_dev      => 0,
	    verbose       => 1
        },
        # Small dataset, yields 107-137MB
        'DS-small'  => {
            file_size_avg => 32768,
            file_size_dev => 2048,
            depth_avg     => 4,
            depth_dev     => 0,
            dir_avg       => 5,
            dir_dev       => 0,
            file_avg      => 4,
            file_dev      => 0,
	    verbose       => 1
        },
        # Medium dataset, yields 468-601MB
        'DS-medium' => {
            file_size_avg => 32768,
            file_size_dev => 1024,
            depth_avg     => 4,
            depth_dev     => 0,
            dir_avg       => 6,
            dir_dev       => 0,
            file_avg      => 10,
            file_dev      => 0,
	    verbose       => 1
        },
        # Large dataset, yields 4.11-4.24GB
        'DS-large'  => {
            file_size_avg => 262144,
            file_size_dev => 4096,
            depth_avg     => 4,
            depth_dev     => 0,
            dir_avg       => 6,
            dir_dev       => 0,
            file_avg      => 11,
            file_dev      => 0,
	    verbose       => 1
        },
        # Basis for arbitrary-sized dataset
        'DS-nG'     => {
            file_size_avg => 65536,
            file_size_dev => 4096,
            depth_avg     => 4,
            depth_dev     => 0,
            dir_avg       => 6,
            dir_dev       => 0,
            file_avg      => 11,
            file_dev      => 0,
	    scramble      => 0,
	    flood         => 1,
	    fork_count    => 4,
	    verbose       => 1
        },
    );
    # DS-populate is an alias for DS-medium:
    $DATASET_PARAMS{'DS-populate'} = $DATASET_PARAMS{'DS-medium'};

    # Record hi-res time when we start, so we can report how long it took:
    my $time_start = [ gettimeofday ];

    my @client_list = split /,/ => delete $args{client_list};
    my $dataset = delete $args{dataset};
    if( $DATASET_PARAMS{$dataset} ) {
	foreach( keys %{ $DATASET_PARAMS{$dataset} } ) {
	    $args{$_} = $args{$_} ? $args{$_} : $DATASET_PARAMS{$dataset}->{$_};
	}
    }

    my %sio_param;

    NOTE::logcomment('Start data generation of ' . $obj->description);

    # must run sio as part of populate to create the files needed
    # by traffic(). this allows RO traffic to work on a DP volume.
    my $sio_size = _set_sio_param( $dataset, \%sio_param );
    # Since we are running r/w traffic, mount RW-Tree
    my $sio_started = _start_sio( $obj, %sio_param, read => 90, RW => 1 );

    # During data generation create upto 3 snapshots for Volume object or 1 for
    # Qtree object
    _make_pop_ssn($obj) if $sio_started;

    # Set up %args with the defaults from %DATASET_PARAMS, and also set the
    # seed and a few other things:
    my $defaults;
    if ($dataset =~ /^DS-(\d+)G$/) {
        # We're going to modify it, so make a copy instead of using it directly:
        $defaults = { %{$DATASET_PARAMS{'DS-nG'}} };
        # The base value in file_size_avg is for 1GB, so multiply it by the
        # number of GB requested, which we matched in the regex above:
        $defaults->{file_size_avg} *= $1;
    } elsif (! ($defaults = $DATASET_PARAMS{$dataset})) {
        confess "Unknown dataset value: '$dataset'";
    }
    # For each of these 8 keys, apply them to %args only if there is not
    # already a key present with a non-undef value (note that I can't just
    # check for false, as 0 is a valid option for some of these keys).
    for my $dskey (qw(file_size_avg file_size_dev depth_avg depth_dev
                      dir_avg dir_dev file_avg file_dev)) {
        $args{$dskey} = $defaults->{$dskey} if (! defined $args{$dskey});
    }
    # DataManip and RichData actually overlap their arguments a lot, except for
    # the pair that specify average and deviation of file size for file
    # creation. Copy the one set (which is for RichData, and has other meaning
    # for DataManip) onto the names that DataManip would want.
    $args{breadth_avg} = $args{file_size_avg};
    $args{breadth_dev} = $args{file_size_dev};
    # To be safe, set the environment_type value for DataManip. I think we're
    # passing all the values, but in case I missed one keep the internal
    # defaults.
    $args{environment_type} = 'filer';
    # Set the value for "seed" if it hasn't been passed in with %args:
    if (! $args{seed}) {
        # This calculation for a seed is present in both DataManip and RichData,
        # I imagine one of them copied it from the other. Either way, I'm
        # reusing it here to keep a degree of consistency in place, since we
        # are now *always* giving an explicit seed, whether the user provides
        # it or we do. This allows us to report the seed in the log-comments
        # at the end, so a user can re-use it if the need to recreate a
        # specific run.
        while (! $args{seed}) {
            $args{seed} = time ^ ($$ + ($$ << 15));
        }
        # For now, RichData doesn't take "seed" as a parameter, it expects a
        # global value in the main:: namespace. So we have to accomodate that
        # to ensure that RichData gets the right seed value:
        $main::DATA_SEED = $args{seed};
    }

    my @clients;
    $obj->mount( RW => 1 );    # Mount volume on localhost
    foreach my $cname (@client_list) {
        my $root = $obj->mountpoint . "/$cname";
        if ( -e $root ) {
            NOTE::logcomment('Existing data in dir:');
            system "ls -l $root";

            NOTE::logcomment("Removing any existing data from dir $root");

            # Remove previous populated data if any
            my $cmd = "sudo rm -rf $root";
            my $rc  = system "$cmd 2>\&1";
            confess "Removal of existing data failed. RC = $rc"
                if $rc;
        }

        NOTE::logcomment( "Create dir $root to generate data in" );

        # Create directory to storage data
        mkdir $root;

        # Pre-create share / mount to RW-Tree as it is for R/W access
        if ( host($cname) =~ /windows/i ) {
            # Pre-create a RW-Tree share
            $obj->create_share( RW => 1 );
        }
        else {
            # Mount on the client if not already mounted
            $obj->mount( client => $cname, RW => 1 )
                unless $obj->mounted( client => $cname );
        }
        my $client = NOTE::Client->new(
            storage      => $obj,
            name         => $cname,
            skip_remount => 1
        );
        $args{path} = "/$cname";
        $client->populate( %args, background => 1 );
        push @clients, $client;
    }
    if ( ref($obj) =~ /Qtree/ ) {
        _make_pop_ssn( $obj->Volume );
    }
    else {
        _make_pop_ssn($obj);
    }
    NOTE::logcomment('Wait for data generation to complete:');
    while ( my $clt = shift @clients ) {
        $clt->wait;
    }
    if ($sio_started) {
        _make_pop_ssn( $obj, 3 );
        $obj->sio( stop => 1 );
    }

    # We're done, record the hi-res elapsed time:
    my $elapsed = tv_interval($time_start);

    # Derive the amount of space we've just used up, and what's left:
    my ( $total, $used, $avail, $capacity ) = $obj->disk_free;

    NOTE::logcomment('Data generation completed for ' . $obj->description);
    NOTE::logcomment(
        sprintf 'Usage %s (%s SIO, %s structured), Available %s',
        _amount($used), _amount($sio_size), _amount($used - $sio_size),
        _amount($avail)
    );
    NOTE::logcomment(sprintf 'Time spent: %.5f seconds', $elapsed);
    NOTE::logcomment("Random seed used: $args{seed}");

    return;
}

sub _amount {
    my ($amt) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $unit = "MB";
    if ( $amt > 1024 * 1024 ) {
        $amt = $amt / ( 1024.0 * 1024.0 );
        $unit = "GB";
    }
    else {
        $amt = $amt / 1024.0;
    }

    return sprintf( "%.2lf$unit", $amt );
}

sub _make_pop_ssn {
    my ($storage) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Only make snapshot for volume object
    $storage->snapshot_create( NOTE->newName("sn") ) if ref($storage) =~ /Volume/;
}

=head2 ::modify

Modifies the data generated using 'populate' method above.  A client list may
be specified to determine NFS or CIFS data modification.  This client list may
be a subset of the ones used by 'populate'. If no client is specified then CLI
parameter CLIENT_LIST used be used to determine the clients used.

=over

=item Arguments

B<storage> => I<NOTE storage object>

Required.  A NOTE::Volume or a NOTE::Qtree object where data modification is
to take place. It is assumed that a previous call to 'populate' with the given
client list has already been performed. The data modification operates on the
data previously generated using 'populate'.

B<client_list> => I<comma separated list of clients>

Optional.  This list maybe a subset of the list supplied to 'populate'. If not
supplied then CLI parameter CLIENT_LIST used be used.

B<%args> => I<dataManip|RichData argument Hash>

I<Optional.>

Addditional data generation parameters.

Note: The USEDATAGEN cli parameter is supported by this call as well.
If USEDATAGEN is used, then modify() will subtest datagen.
See dp/snapmirror/paloma/datagen10.thpl for details.

=item Examples

=back

    $util->modify( storage => $vol );
    $util->modify(
        storage     => $vol_obj,
        client_list => 'dpg-sun1,nam21',
        %DataManipOptions
    );

=cut

sub modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $obj;
    if ( defined $args{storage} ) {
        $obj = delete $args{storage};
	if ( $args{datagen} || Tharn::param('USEDATAGEN') ) {
	    $obj->modify();	# calling modify() from Volume.pm which will call DATAGEN
	    return;
	}
    }
    else {
        confess 'Modify: Missing NOTE storage object...';
    }
    NOTE::logcomment( 'Modify ' . $obj->description );

    my @client_list = split ',', ( $args{client_list} || $self->{_CLIENT_LIST} );
    NOTE::logcomment( 'Start data modification of ' . $obj->description );
    my $dataset = delete $args{dataset} || $self->param('DATASET');
    my %sio_param;
    _set_sio_param( $dataset, \%sio_param );
    my $sio_started = _start_sio( $obj, %sio_param, read => 90, RW => 1 );    # Since we are perform r/w traffic, mount RW-Tree
    my @clients;
    foreach my $cname (@client_list) {

        # Pre-create share / mount to RW-Tree as it is for R/W access
        if ( host($cname) =~ /windows/i ) {

            # Pre-create RW-Tree share
            $obj->create_share( RW => 1 );
        }
        else {

            # Mount on the client if not already mounted
            # Check if mountpoint already exist or not
            # The function Volume::mountpoint confesses, if not mounted,
            # so, call in eval
            my $eval_return = eval {
                                    $obj->mountpoint( client => $cname );
                                    1;
                              };
            if(!$eval_return) {
                $obj->mount( client => $cname, RW => 1 );
            }
        }
        my $client = NOTE::Client->new(
            storage      => $obj,
            name         => $cname,
            skip_remount => 1
        );
        $args{path} = '/' . $cname;
        $client->modify( %args, background => 1 );
        push @clients, $client;
    }
    NOTE::logcomment('Wait for data modification to complete:');
    while ( my $clt = shift @clients ) {
        $clt->wait();
    }
    $obj->sio( stop => 1 ) if $sio_started;
    my ( $total, $used, $avail, $capacity ) = $obj->disk_free;
    NOTE::logcomment( 'Data modification completed for ' . $obj->description . "\n  Usage " . _amount($used) . ', Available ' . _amount($avail) );
}

=head2 ::traffic

Provide continous read/write traffic to supplied volume or qtree. The volume
or qtree must have had it's data generated using the 'populate' method from
above.

In order to run RO traffic on a DP volume either populate or traffic must
first be run on the source volume and the data synchronized to the DP volume.

=over

=item Arguments

B<storage> = I<NOTE storage>

A NOTE::Volume or NOTE::Qtree object

B<type> => I<r|w|b>

I<Optional.>

Indicates whether to provide (r)ead, (w)rite, or (b)oth (read and write) traffic. Default: r

B<stop> => I<boolean>

I<Optional.>

Indicates to stop traffic. If not set or set to zero, then starts traffic. Default: 0

B<RW> => I<1|0|undef>

I<Optional.>

Indicates whether to mount RW-Tree or R-Tree in read/write mode.
If not defined, a remount test will be applied to the R-Tree.
Default: undef

=item Example

=back

    $util->populate(storage=>$vol);
    $util->traffic($vol); # Starts nfs read traffic on $vol
    $util->traffic($vol, stop=>1); # Stops read traffic on $vol

=cut

sub traffic {
    my ( $self, $obj, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Invalid 'storage' parameter: $obj"
      unless ref($obj) =~ /NOTE::.*(Volume|Qtree)/;

    # Accepts both rw or RW parameter
    $args{RW} = delete $args{rw}
      if exists $args{rw} && !exists $args{RW};

    my $label;
    if ( $args{stop} ) {
        $obj->sio( stop => 1 );
        NOTE::logcomment( $obj->description . ": traffic stopped" );
    }
    else {
        my %sio_param;
        $args{dataset} ||= $self->param("DATASET");
        _set_sio_param( $args{dataset}, \%sio_param );
        $sio_param{RW} = ( exists $args{RW} ? $args{RW} : 1 );
        if ( $args{type} =~ /b/i ) {
            $label = "read/write";
            $sio_param{read} = 50;
        }
        elsif ( $args{type} =~ /w/ ) {
            $label = "write";
            $sio_param{read} = 0;
        }
        else {
            $label = "read";
            $sio_param{read} = 100;
            delete $sio_param{RW} unless exists $args{rw};
        }

        NOTE::logcomment( $obj->description . ": Starting $label traffic" );
        $obj->sio(%sio_param);
    }
}

# Set sio param from dataset

sub _set_sio_param {
    my ( $dataset, $sio_param ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $sio_param->{client} = 'localhost';
    $sio_param->{suffix} = 'smUtil';

    # the fixes for burts 478343 and 506218 the sizes of the files created
    # by sio have been reduced. this allows populate() to create the files
    # needed by traffic() without substancially increasing the run time of
    # populate()
    if ($dataset eq 'DS-tiny') {
        # Use for very small tests, unit tests, etc., 20MB
        $sio_param->{count}     = 2;
        $sio_param->{file_size} = 10;
    } elsif ($dataset eq 'DS-small') {
        # Use for script development, 40MB
        $sio_param->{count}     = 2;
        $sio_param->{file_size} = 20;
    } elsif (($dataset eq 'DS-populate') || ($dataset eq 'DS-medium')) {
        # General usage, 60MB
        $sio_param->{count}     = 3;
        $sio_param->{file_size} = 20;
    } elsif ($dataset =~ 'DS-large') {
        # Create 160MB (40MB of sio data / 1GB of structured data)
        $sio_param->{count}     = 8;
        $sio_param->{file_size} = 20;
    } elsif ($dataset =~ /DS-(\d+)G/) {
        # Create 40MB of sio data / 1GB of structured data
        $sio_param->{count}     = ($1 * 2);
        $sio_param->{file_size} = 20;
    } else {
        confess "Unknown dataset specification '$dataset'";
    }

    return $sio_param->{count} * $sio_param->{file_size} * 1024;
}

=head2 ::snapshot

Create or delete 'count' snapshots from a volume.

=over

=item Arguments

B<vol> = I<NOTE::Volume>

A NOTE::Volume object to operate on.

B<action> => I<create|delete|delete_all>

I<Optional.>

Specify creation or deletion of snapshot(s). Default: create

B<count> => I<integer>

I<Optional.>

The nunber of snapshot(s) to delete. Default: 1.  Ignored if action is 'delete_all'.

B<prefix> => I<string>

I<Optional.>

Prefix snapshot name. Default: 'ssn_' for 'create' action and undef for 'delete' action.
Ignored for 'delete_all' action.

B<offset> => I<integer>

I<Optional.>

The offset to start naming the snapshot(s), used for 'create' action only. Default: 1

=item Example

=back

    $util->snapshot($vol, action=>"create"); # Creates 1 snapshot in $vol volume
    $util->snapshot($vol, action=>"create", prefix=>"new_ssn_", count=>2); # Creates 2 snapshots with 'new_ssn_' prefix
    $util->snapshot($vol, action=>"delete", count=>5);  # Delete 5 snapshots from $vol volume
    $util->snapshot($vol, action=>"delete", prefix=>"new_ssn_");  # Deletes a snapshot containing 'new_ssn_' from $vol volume
    $util->snapshot($vol, action=>"create", count=>2, offset=>2); # Creates snapshot ssn_2 and ssn_3
    $util->snapshot($vol, action=>"delete_all"); # Deletes all snapshots in $vol volume

=cut

sub snapshot {
    my ( $self, $vol, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{count}  = 1 unless $args{count};
    $args{offset} = 1 unless $args{offset};

    confess "Missing NOTE::Volume object"
      unless $vol;
    confess "Invalid action ", $args{action} . ", must be 'create', 'delete', or 'delete_all'"
      unless $args{action} =~ /(create|delete|delete_all)/i;
    confess "Invalid count " . $args{count}
      unless $args{count} > 0;
    confess "Invalid offset " . $args{offset}
      unless $args{offset} > 0;

    my $label = "Snapshot: Action " . $args{action};
    if ( $args{action} !~ /delete_all/i ) {
        $label .= ", Count ";
        $label .= $args{count};
    }

    NOTE::logcomment( $vol->description . ", Start $label" );
    my $cnt    = 0;
    my $prefix = delete $args{prefix};
    if ( $args{action} =~ /create/i ) {
        $prefix = "ssn_" unless $prefix;
        while ( $cnt < $args{count} ) {
            $vol->snapshot_create( $prefix . ( $args{offset} + $cnt++ ) );
        }
    }
    elsif ( $args{action} =~ /delete_all/i ) {
        my @ssn_list = $vol->snapshot_list();
        foreach my $del_ssn (@ssn_list) {
            $vol->snapshot_delete($del_ssn);
        }
    }

    # if here, must be delete
    else {
        my @ssn_list = $vol->snapshot_list();
        my $del_ssn;
        while ( $cnt < $args{count} && ( $del_ssn = pop @ssn_list ) ) {
            if ($prefix) {

                # skip snapshot that does not include 'prefix', if defined
                next unless $del_ssn =~ /$prefix/i;
            }
            $vol->snapshot_delete($del_ssn);
            $cnt++;
        }
    }

    # confess if 'count' argument was not satisfied.  does not apply to 'delete_all' action.
    if ( ( $args{action} !~ /delete_all/i ) && ( $cnt != $args{count} ) ) {
        confess "Failed to " . $args{action} . " " . $args{count} . " snapshot(s)...Only $cnt affected";
    }

    NOTE::logcomment( $vol->description . ", Completed $label" );
}

=head2 ::fill_inode

Use up 'count' inodes in the given volume by creating 'count' files.
If 'count' is not specified all inodes will be used up.

=over

=item Arguments

B<volume> = I<NOTE::Volume object>

A NOTE::Volume object to fill up all inodes for.

B<count> => I<integer>

I<Optional.>

Create <count> inodes. Default: use all inodes

=item Examples

=back

     $util->fill_inode($vol); # use all inodes
     $util->fill_inode($vol, count=>50); # create 50 files/inodes

=cut

sub fill_inode {
    my ( $self, $vol, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $vol_info = $vol->status();
    my $make_cnt = $args{count} || ( $vol_info->{'files'} - $vol_info->{'files-used'} );
    my $cnt      = 0;
    $vol->mount();
    my $root = $vol->mountpoint();
    NOTE::logcomment( "Generate $make_cnt files to fill up all inodes in volume " . $vol->description );
    while ( ++$cnt <= $make_cnt ) {
        my $cmd;
        if ( $cnt % 120000 ) {
            $cmd = "touch " . $root . "/fill_inode_" . $cnt;
        }
        else {
            $root .= "/dir_" . $cnt;
            $cmd = "mkdir " . $root;
        }
        my $rc = system($cmd);
        confess "Failed running cmd: $cmd, RC: $rc"
          if $rc;
    }
    $vol_info = $vol->status();
    if ( $vol_info->{'files'} - $vol_info->{'files-used'} > 0 ) {
        confess "failed to use up all inodes within the volume. " . "Max " . $vol_info->{'files'} . ", Used " . $vol_info->{'files-used'};
    }
    else {
        NOTE::logcomment( "All inodes used volume " . $vol->description );
    }
}

=head2 ::unique_clusters

Given an array of NOTE::GX::Node objects, will return a true if all the supplied nodes are located

on different clusters.

The routine confess if the nodes supplied are not in different clusters.

=over

=item Arguments

B<nodes> = I<Array of NOTE::GX::Node>

An array of NOTE::GX::Node objects representing the nodes to check.

=item Examples

=back

    my @nodes = ($src_node, $dst_vol->Node);
    my $unique = $util->unique_clusters( @nodes );
    my $unique = $util->unique_clusters($src_node, $dst_vol->Node);

=cut

sub unique_clusters {
    my ( $self, @node_list ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my %cluster;

    # Check for missing input params
    confess 'Missing input parameters. See help documentation.'
        unless @node_list;

    foreach my $node (@node_list) {
        # Check to see if $node is a NOTE::GX::Node object
        if( ref($node) ne 'NOTE::GX::Node' ) {
            confess('One of the input params is not a "NOTE::GX::Node" object');
        }

        # Create a hash with keys as Cluster names
        if ( exists $cluster{ $node->Cluster->name } ) {
            confess('Nodes: ' . $node->name . ',' . $cluster{ $node->Cluster->name }
                  . ' found in same cluster: ' . $node->Cluster->name);
        }
        else {
            $cluster{ $node->Cluster->name } = $node->name;
        }
    }

    NOTE::logcomment('Given nodes have unique clusters');
    return 1;
}

=head2 ::make_special_files

Routine to create Unix special files.
The type of special files are: block device, character device, named pipe and socket.
A directory is created to locate the created special files.

=over

=item Arguments

B<storage> => I<Storage Object>

A NOTE::Volume or NOTE::Qtree object to create the special files in.

B<count> => I<integer>

I<Optional.>

Number of each special file types to create. Default: 1

B<path> => I<string>

I<Optional>

A path relative to the root of the volume where the special files will be created.
Default: /

=item Example

=back

    $util->make_special_files(storage=>$src_vol, count=>2, path=>"/special_dir");

=cut

sub make_special_files {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing 'storage' parameter"
      unless exists $args{storage};
    confess "'storage' must be a 'NOTE::Volume' or 'NOTE::Qtree' object"
      unless ref( $args{storage} ) =~ /NOTE::.*(Qtree|Volume)/;

    $args{path} = "/" unless exists $args{path};

    my $max = $args{count} || 1;

    my $root = $args{storage}->nfs( RW => 1 );
    my %types = (
        "c" => "character_device",
        "b" => "block_device",
        "p" => "named_pipe"
    );
    my $msg;
    for my $idx ( 1 .. $max ) {
        my $pathname;
        for my $key ( keys %types ) {
            my $major = $idx % 255;                 # XXX Add variation in this parameter (0 - 255)
            my $minor = abs( 255 - $idx ) % 255;    # XXX Add variation in this parameter (0 - 255)
            $pathname = $args{path} . "/special_" . $types{$key} . "_" . $idx . "." . time();
            $msg .= "\n $pathname";
            $root->mknod( name => $pathname, type => $key, major => $major, minor => $minor );
        }
        $pathname = $args{path} . "/special_socket_" . $idx . "." . time();
        $msg .= "\n $pathname";
        $root->create_socket($pathname);
    }
    NOTE::logcomment("Created the following special files:$msg");
}

=head2 :: modify_special_files

Modifies the special files created by make_special_files method.

=over

=item Arguments

B<storage> => I<Storage Object>

A NOTE::Volume or NOTE::Qtree object to create the special files in.

B<path> => I<string>

I<Optional>

A path relative to the root of the volume where the special files are located.
Default: /

=item Example

=back

    $util->modify_special_files(storage=>$src_vol, path=>"/special_dir");

=cut

sub modify_special_files {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing 'storage' parameter"
      unless exists $args{storage};
    confess "'storage' must be a 'NOTE::Volume' or 'NOTE::Qtree' object"
      unless ref( $args{storage} ) =~ /NOTE::.*(Qtree|Volume)/;

    $args{path} = "/" unless exists $args{path};

    my $root = $args{storage}->nfs( RW => 1 );
    opendir DIR, $root->fullpath . $args{path};
    my @list = readdir DIR;
    closedir DIR;
    my $msg;
    while ( my $old_name = pop @list ) {
        if ( $old_name =~ /special_/ ) {
            my ( $name, $extension ) = split /\./, $old_name;
            my $new_name = $name . "." . time();
            $msg .= "\n $old_name -> $new_name";
            $root->mv( $args{path} . "/" . $old_name, $args{path} . "/" . $new_name );
        }
    }
    NOTE::logcomment( "Dir " . $args{path} . ", Modify (rename) special files:$msg" );
}

=head2 ::consistency_check

This function is a place holder for any functions to be invoked prior to the end of any test, for the maintenance purpose.

At present, this function will invoke wafliron operation on the corresponding aggregate/s of the input volumes.

=over

=item Arguments

B<force> => B<wafliron | all>

I<Optional>

Specify which of the test/s must be execute by the test. If this is not specified, none of the functions will be invoked.
wafliron means to invoke wafliron on the input volume/s.
all means to invoke all functions on the input volume/s.

B<vol> = I<Array of NOTE::GX::Volume>

I<Required>

An array of NOTE::GX::Volume objects representing the volumes to check.

=item Examples

=back

    $util->consistency_check (\@vol_list);
    $util->consistency_check (force => 'wafliron', vol => \@vol_list);
    $util->consistency_check (force => 'all', vol => \@vol_list);

=cut

sub consistency_check {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $tag = "vol";

    confess "Invalid 'force' value of " . $args{force} . ". Must be one of 'wafliorn or all'"
      if exists $args{force} && !$args{force} =~ /(wafliron|all)/i;

    my $flag_iron = 0;

    if ( exists $args{force} && $args{force} =~ /(wafliron|all)/i ) {
        $flag_iron = 1;
    }

    my $flag_run_check = 0;
    my $run_check      = $self->param('RUN_CHECK');
    if ( $run_check == 1 ) {
        $flag_run_check = 1;
    }

    if ( $flag_iron || $flag_run_check ) {
        my @vol_list = ();

        if ( ref( $args{$tag} ) =~ /NOTE::.*Volume/ ) {
            @vol_list = ( $args{$tag} );
        }
        elsif ( ref( $args{$tag} ) =~ /ARRAY/ ) {
            @vol_list = @{ $args{$tag} };
        }

        #
        # Combine the vol_list, in case if the list contains both the parent
        # and member volume, only keep the parent volume
        #
        #   Pass 1: Extract all the parent (non-member) volume
        #
        #   Pass 2: On the remaining list, for each member volume, if
        #           its parent is not on the final_vol_list, add it in.
        #

        my $vol;

        my @member_vol_list = ();
        my %final_vol_list;

        my $aggr_name = "";

        foreach $vol (@vol_list) {
            if ( !$vol->is_member() ) {
                $aggr_name = $vol->Aggregate()->name();
                $final_vol_list{$aggr_name} = $vol;
            }
            else {
                $aggr_name = $vol->parent->Aggregate()->name();

                if ( !exists $final_vol_list{$aggr_name} ) {
                    push( @member_vol_list, $vol );
                }
            }
        }

        foreach $vol (@member_vol_list) {
            $aggr_name = $vol->parent->Aggregate()->name();
            if ( !exists $final_vol_list{$aggr_name} ) {
                $final_vol_list{$aggr_name} = $vol;
            }
        }

        my $report = "";

        #
        # Start Wafliron on all volume first without waiting for completion
        #
        foreach $vol ( values(%final_vol_list) ) {
            $vol->wafliron_start( wait => 0, force => 1, include_mirror => 1 );
        }

        my $flag_failed = 0;

        #
        # For each volume, wait for the Wafliron completion, and then display
        # the report
        #
        foreach $vol ( values(%final_vol_list) ) {
            my $result = "";
            my ( $flag_corruption, $report_lines ) = $vol->wafliron_wait();

            if ($flag_corruption) {
                $vol->SPARE;
                $flag_failed = 1;
            }

            if (@$report_lines) {
                $result = join "\n", @$report_lines;
            }
            NOTE::logcomment( "Wafliron result for Volume, " . $vol->name . " :\n$result" );
        }

        if ( $flag_failed == 1 ) {
            confess "Call to consistency_check () failed!";
        }
    }
}

=head2 ::is_read_accessible

Validate whether the input file can be accessible by the input window's clients

=over

=item Arguments

B<storage> = I<NOTE::Volume object>

Required. A NOTE::Volume object to fill up all inodes for.

B<file_path> = I<path of the file to test in the client>

Required. The path of the file for testing the accessible of the client

B<client_list> = I<list of clients>

Optional. List of clients that access the Filer, by means of the input file. Use default CLIENT_LIST if not provided.

=item Example

=back

     $util->is_read_accessible (storage=>$vol, file_path=>"/dr08430_lads_anodizes/file.c");
     $util->is_read_accessible (storage=>$vol, file_path=>"/dr08430_lads_andizes/file.c", client_list="dpg-vm10");

=cut

sub is_read_accessible {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $in_storage = $args{storage};
    if ( defined $args{storage} ) {
        $in_storage = $args{storage};
    }
    else {
        confess "is_read_accessible: Missing NOTE storage object...";
    }

    my $in_file_path;
    if ( defined $args{file_path} ) {
        $in_file_path = $args{file_path};
    }
    else {
        confess "is_read_accessible: Missing file_path...";
    }

    my @client_list = split ",", ( $args{client_list} || $self->{_CLIENT_LIST} );

    foreach my $cname (@client_list) {
        my $client = NOTE::Client->new( name => $cname, storage => $in_storage );
        my $handle = $client->get_handle($in_file_path);

        #
        # 20090418  tyu@netapp.com
        #
        # Add a block to retry in case if read_file failed.
        # This is basically trying to alleviate the problem faced in
        # reading file in Windows (CIFS). We got "permission denied" error
        # particularly when accessing files in volumes with CORAL striping.
        #
        my $retry        = 0;
        my $flag_read_ok = 0;
        my $err_msg      = "";
        do {
            eval {
                my $content = $client->read_file($handle);

                #       NOTE::logcomment("content of $in_file_path is ".Dumper(\$content));
            };
            if ($@) {
                $err_msg = $@;
                NOTE::logcomment("Unable to read (attempt# $retry) $in_file_path: $@");
                sleep(30);
            }
            else {
                $flag_read_ok = 1;
            }
        } while ( ( $flag_read_ok != 1 ) && ( $retry++ < 10 ) );

        if ( $flag_read_ok != 1 ) {
            confess "Unable to read file $in_file_path, error is $@.";
        }
    }

    return 1;
}

=head2 ::has_client_type

Validate whether the client list contains at least one client of the seeked type/s

=over

=item Arguments

B<client_type> = I<windows|unix|both>

Required. Define what type of the client it is seeking

B<client_list> = I<list of clients>

Optional. List of clients. Use default CLIENT_LIST if not provided.

=item Example

=back

     $util->has_client_type (client_type=>'windows', client_list="dpg-vm10");
     $util->has_client_type (client_type=>'unix');

=cut

sub has_client_type {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $self->client_list(%args) > 0 ) {
        return 1;
    }
    else {
        NOTE::logcomment("No client of requested type found...");
        return 0;
    }
}

=head2 ::clist_list

Returns list containing client(s) of the seeked type/s
'Linux' clients will be included in the category of 'Unix'
clients. Essentially there are 2 type: Windows and non-Windows.

=over

=item Arguments

B<client_type> = I<windows|unix|both>

Required. Define what type of the client it is seeking

B<client_list> = I<list of clients>

Optional. List of clients. Use default CLIENT_LIST if not provided.

=item Example

=back

     my @win_list = $util->client_list (client_type=>'windows', client_list="dpg-vm10");
     my @unix_list = $util->client_list (client_type=>'unix');

=cut

sub client_list {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @client_list = split ",", ( $args{client_list} || $self->{_CLIENT_LIST} );

    my $need_check_window = 0;
    my $need_check_unix   = 0;

    my @list;

    if ( $args{client_type} =~ /both/i ) {
        $need_check_window = 1;
        $need_check_unix   = 1;
    }
    else {
        if ( $args{client_type} =~ /windows/i ) {
            $need_check_window = 1;
        }
        if ( $args{client_type} =~ /(unix|linux)/i ) {
            $need_check_unix = 1;
        }
    }

    foreach my $client_name (@client_list) {
        if ( host($client_name)->hosttype =~ /windows/i ) {
            push @list, $client_name
              if $need_check_window;
        }
        elsif ( host($client_name)->hosttype =~ /(unix|linux)/i ) {
            push @list, $client_name
              if $need_check_unix;
        }
    }

    return @list;
}

sub _set_BRE {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ($c1, $c2);                  #Cluster object references
    my $bre_flag;

    $bre_flag = $self->param("BRE");
    $c1 = NOTE::GX::Cluster->connect( Node =>  $self->param("FILER_SRC:0"),
                                      direct => 1 );
    logcomment("Setting BRE flag in ". $c1->description );
    $c1->transfer_engine( $bre_flag ? "BRE" : "LRE" );

    my @nodes = $c1->nodes;

    # Determine whether cross-cluster test or not. If it is, set the
    # BRE engine on the other cluster too.
    my $destination_node = $self->param("FILER_DST:0");
    foreach my $node (@nodes) {
        if ( $node eq $destination_node ) {
            return;
        }
    }

    $c2 = NOTE::GX::Cluster->connect( Node => $self->param("FILER_DST:0"),
                                      direct => 1 );
    if ( $c1->description ne $c2->description ) {
       logcomment("Setting BRE flag in " . $c2->description );
       $c2->transfer_engine( $bre_flag ? "BRE" : "LRE" );
    }
}

=head2 ::get_gnote_continuum_mode

Determines what continuum mode GNOTE is in and returns the appropriate string
representing that mode.

The mode determines the type of options used in the snapmirror commands, 7g
style (single character), cluster style (multi-character), or mixed (as
defined by the method arguments). The mode is determined by the two ntest
paramters SNAPMIRROR_7G_PATH and SNAPMIRROR_CLUSTER_PATH.

if SNAPMIRROR_7G_PATH is set mode = 7g
if SNAPMIRROR_CLUSTER_PATH is set mode = cluster
if neither is set mode = mixed


Returns:
SUCCESS: 7g|cluster|mixed
FAILURE:

confessions:
None

=over

=item Arguments
None

=item Notes

=item Examples

=back

    # get the current gnote continuum mode
    $mode = $util->get_gnote_continuum_mode();

=cut

sub get_gnote_continuum_mode {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mode;

    if (exists($util->{_SNAPMIRROR_7G_PATH})) {
        $mode = '7g';
    } elsif (exists($util->{_SNAPMIRROR_CLUSTER_PATH})) {
        $mode = 'cluster';
    } else {
        $mode = 'mixed';
    }

    return $mode;
}

1;

