# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/SnapVault.pm#1 $
#
## @summary  GNOTE GX SnapVault module
## @author   dl-dpgqa-automation@netapp.com, cpeter@netapp.com
## @status   private
## @pod      here

=head1 NAME

NOTE::GX::SnapVault

=head1 DESCRIPTION

SnapVault is subset of SnapMirror class.

STATE IS STILL PRIVATE!!! Use at your own risk


=head1 METHODS

Same as NOTE::GX::SnapMirror. Differences or expensions are listed here. See NOTE::GX::SnapMirror for rest of the methids.

=cut

package NOTE::GX::SnapVault;

use strict;
use warnings;
no warnings 'redefine';
use re 'eval';

use GNOTE;
use GNOTE::GX;
use GNOTE::SnapMirror;
use GNOTE::GX::Mirror;
use GNOTE::GX::SnapMirror;
use Carp qw(confess);
use Data::Dumper;
use TracePkg;

our @ISA = qw(NOTE::GX::SnapMirror NOTE::GX);

my $Trace = TracePkg->new();

=head2 create

Creates a snapvault object only. Returns object representing the
relationship.

No restore related actions are performed on a filer. The difference
between NOTE::GX::SnapMirror->create() and NOTE::GX::SnapVault->create()
is that NOTE::GX::SnapMirror->create() executes 'snapmirror create' command
on a filer and NOTE::GX::SnapVault->create() does not. Everything else
remains the same.

=over

=item Arguments

See NOTE::GX::SnapMirror::create for arguments and examples

=back

=cut

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $self;
    if ( ref($class) =~ /NOTE::GX::SnapVault/ ) {
        $self = $class;
    }
    else {				    # constructor is inherited from GNOTE::GX::SnapMirror
        $self = __PACKAGE__->new(%args);    # GNOTE::GX::SnapMirror->new() will get called,
    }					    # but 'GX::SnapVault' object is returned.

    my %opts;
    $opts{'destination-volume'}  = $self->{_dst}->name;
    $opts{'destination-vserver'} = $self->{_dst}->Vserver->name;
    $opts{'destination-cluster'} = $self->_cluster_name("dst");
    $opts{'source-volume'}       = $self->{_src}->name;
    $opts{'source-vserver'}      = $self->{_src}->Vserver->name;
    $opts{'source-cluster'}      = $self->_cluster_name("src");
    $opts{'throttle'}            = $self->{_throttle} if defined $self->{_throttle};

    $self->_hash_xfer(
        label => [ "comment", "schedule", "throttle", "tries", "type" ],
        from  => \%args,
        to    => \%opts
    );

    $opts{'-Timeout'} = 120;

    if ( $self->{_src}->Cluster ne $self->{_dst}->Cluster ) {
        $self->{_dst}->Cluster->peer_create( 'cluster' => $self->{_src}->Cluster );
    }
    unless ( delete $args{cmd_only} ) {
        $self->{_src}->Cluster->license( "SnapMirror_" . uc( $opts{type} || "DP" ), 1 );
        $self->{_dst}->Cluster->license( "SnapMirror_" . uc( $opts{type} || "DP" ), 1 );
    }

    NOTE::logcomment( "Creating snapmirror: " . $self->description );

    my $diff_geometry = $self->_diff_geometry;
    my $stripe_sm = $self->src->is_stripe || $self->dst->is_stripe;

    if ( $stripe_sm && $self->{_tricky_parity} ) {

        # Set paloma_allow_striped_replication flag for striped volumes in TRICKY_PARITY mode
        eval { $self->_setflag("paloma_allow_striped_replication"); };    # Ignore error
    }
    if ($diff_geometry) {

        # Set paloma_allow_unlike_geometry in src
        $self->_setflag("paloma_allow_unlike_geometry");
    }

    # test mode required for different geometry or striped volumes in TRICKY_PARITY mode
    my $test_mode = $stripe_sm || $diff_geometry;

    $test_mode = delete $args{test}
      if exists $args{test};                                              # To override set test => 0

    $self->{_dst}->server->set( privilege => "test", confirmations => "off" )
      if $test_mode;

    $self->{_created} = 1;

    $self->{_destroyed} = 0;

    return $self;
}

=head2 restore

Performs a 'snapvault restore'.

Setting CLI parameter TRICKTY_PARITY will cause this command to run in test
mode.

Returns job id if run in background, or returns undef. Croaks on failure.

=over

=item Arguments
Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<snapshot> => I<name>

Name of snapshot.

=back

=item Examples

    $sv->restore( background => 1, snapshot => $snapname );

      1 use strict;
      2 use GNOTE;
      3 use dp::TestWrapper qw( run_test );
      4
      5 my ($util, $src, $dst, $rst, $sm, $sv, $snapname);
      6
      7 run_test();
      8
      9 sub do_setup {
     10     ## @step Create Source, Destination and Restore volumes
     11     $util = NOTE::SnapMirrorUtil->new();
     12     ($src) = $util->create_volume(tag=>"src");                                      # source volume
     13     ($dst) = $util->create_volume(tag=>"dst", 'mirror-type'=>"DP", count => 1);     # destination volume
     14     ($rst) = $util->create_volume(tag=>"dst", 'mirror-type'=>"DP", count => 1);     # restore volume
     15
     16     ## @step Create XDP relationship
     17     $sm = NOTE::GX::SnapMirror->create(src => $src, dst => $dst, type=>'XDP');
     18
     19     ## @step Fill XDP source with data and take snashot after its being created
     20     $sm->src->fill( proto => 'local', type => 'filersio', max => '100M' );
     21     $snapname = $sm->src->snapshot_create();
     22
     23     ## @step Initialize XDP snapmirror - transfer everything to destination
     24     $sm->initialize();
     25 }
     26
     27 sub do_test {
     28     ## @step Create SnapVault object
     29     $sv = NOTE::GX::SnapVault->create( src=>$sm->dst, dst=>$rst );
     30
     31     ## @step Initiate background restore of named snapshot
     32     $sv->restore( background => 1, snapshot => $snapname );
     33
     34     ## @step Wait for restore to complete
     35     $sv->wait_for_restore;
     36 }
     37
     38 sub do_teardown { NOTE::teardown }

=back

=cut

sub restore {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( $self, %args ) = @_;

    my $background = NOTE::GX::background( \%args );
    my $timeout = delete $args{timeout} || 60 * 45; # 45 minute for restore

    my %opts = ();

    $opts{'destination-volume'}  = $self->dst->name;
    $opts{'destination-vserver'} = $self->dst->Vserver->name;
    $opts{'destination-cluster'} = $self->_cluster_name("dst");
    $opts{'source-volume'}       = $self->{_src}->name;
    $opts{'source-vserver'}      = $self->{_src}->Vserver->name;
    $opts{'source-cluster'}      = $self->_cluster_name("src");
    $opts{'destination-path'}	 = $args{'destination-path'} if exists $args{'destination-path'};
    $opts{'source-path'}         = $args{'source-path'} if exists $args{'source-path'};
    $opts{'source-snapshot'}	 = $args{'snapshot'} if exists $args{'snapshot'};
    $opts{'source-snapshot'}	 = $args{'source-snapshot'} if exists $args{'source-snapshot'};

    NOTE::logcomment( "Snapmirror Restore: " . $self->description . "\n" . Dumper(\%opts) );
    my $text;

    my $cmd = "snapmirror restore " . join(' ', map{ '-'.$_, $opts{$_} } keys %opts);
    eval { $text = $self->{_dst}->server->cli( cmd => $cmd, -Timeout => $timeout ) };
    confess $@ if $@;
    $self->{_job_id} = $self->{_dst}->Node->job_id($text);

    return $self->_job_active() if $background;
    $self->wait_for_restore();
}

=head2 wait_for_restore

Wait for background restore job if one exists. No arguments required. See B<restore> method for script examples.

=cut

sub wait_for_restore {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $self = shift;
    return unless $self->{_job_id};
    eval {  # this eval will always fail by design. The only "Good error" is SM no found.
	while( $self->_job_active() ) {	# Dump job and SM status while waiting for completion
	    my @snaplist = $self->dst->snapshot_list;
	    NOTE::logcomment("Waiting for job to complete\nJob:".Dumper($self->{_last_job_show}) .
			     "\nSm Show: " . Dumper($self->{_last_sm_show}) . "\nDest Snaps: " .
			     Dumper(\@snaplist) );
	    sleep 4;	# polling every 4 seconds. Used to be 5, but its too long, 3 is too short
	}		# trying to get balance between too much output and not miss any required info
    };
    confess $@ if $@ && $@ !~ /Relationship with destination \S+ not found/s;
    $self->{_job_id} = undef;
    #confess "No transfer snapshot found in SnapMirror show during trasfer" unless $self->{_transfer_snapshot};
    #confess "No transfer snapshot found on destination volume" unless $self->dst->snapshot_list( snapshot => $self->{_transfer_snapshot} );
}

# modified version of SUPER::_job_active
sub _job_active {
    my $trace = $NOTE::TRACELOG->start_timer();	# log internal metrics
    my ($self) = @_;						    # many of fields are short lived. We need them
                                                                    # for future references and analisys. Storing
    confess "No snapmirror job id found" unless $self->{_job_id};   # them as private variables of the object. They are
                                                                    # _last_job_show, _last_sm_show, _transfer_snapshot
    my $jobstatus;                                                  # and _progress.
    eval { $jobstatus = $self->{_dst}->Node->job_wait( id => $self->{_job_id}, state => $self->_sm_job_states(), column => "all", timeout => 30 ); };
    confess "snapmirror job " . $self->{_job_id} . " failed. Error: $@" if $@ && $@ !~ /(no entries|Error sending the command "job show -id \d+" to the node)/si;
    $self->{_last_job_show} = $jobstatus;

    my $smshow;
    eval { $smshow = $self->status };
    $self->{_last_sm_show} = $smshow unless $@;
    $self->{_transfer_snapshot} = $self->{_last_sm_show}->{'transfer-snapshot'} if $self->{_last_sm_show}->{'transfer-snapshot'} =~ /\w+/;

    $self->{_progress} = $self->{_last_job_show}->{'progress'};
    return  $jobstatus->{'state'} =~ /running/i ? $self->{_job_id} : undef;
}

sub destroy {	# need this to prevent NOTE::GX::SnapMirror to be called by teardown
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    print "Calling SnapVault DESTROY PLACEHOLDER!!!\n";
}

1;
