#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/UserspaceFiji.pm#1 $
#
## @summary  NOTE Userspace Fiji module
## @status   Public
## @pod      here

=head1 NAME

NOTE::GX::UserspaceFiji

=head1 DESCRIPTION

This module contains methods for working with NetApp Userspace FIJIs.

In order to be used, a FIJI must first be registered.  Registering a FIJI
makes it possible to create FIJI objects (using C<::new>)

Userspace FIJI objects are associated with a particular filer, and support
methods such as C<::set> and C<::clear>.

During the teardown process, FIJI objects will be destroyed like any other
object.

=head1 METHODS

=cut

package NOTE::GX::UserspaceFiji;
use GNOTE::GX;
our @ISA = qw(NOTE::GX);

use strict;
use Carp qw(croak);
use Server;
my %_FIJI_table = ();

#-------------------------

=head2 ::register

Register a Userspace FIJI

=over

=item Arguments

B<fault-id> => I<string>

I<Required.>
The fault-id for the FIJI.

=item Examples

=back

     NOTE::UserspaceFiji->register(
         'fault-id' => 'mgwd.voltrans.voltransitionjob.onlinefail',
     );

=cut

sub register {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $id = $args{'fault-id'};

    croak 'Registering Userspace FIJI requires fault-id' unless $id;

    if (exists $_FIJI_table{$id}) {
        NOTE::logcomment("Overriding previously registered Userspace FIJI: \"$id\"");
    } else {
        NOTE::logcomment("Registering Userspace FIJI: \"$id\"")
            unless($args{quiet});
    }
    delete $args{quiet};
    $_FIJI_table{$id} = \%args;

    return $self;
}

=head2 ::new

Create a Userspace FIJI object

=over

=item Arguments

B<Filer> => I<host spec> | <filer object>

I<Required.>
The filer on which this FIJI will be associated

B<fault-id> => I<string>

I<Required.>
Userspace FIJI fault-id.

Note that this does NOT set the FIJI on the filer; it only creates the object.

=item Examples

=back

     my $fiji = NOTE::UserspaceFiji->new(
        Filer => { host => $FILER_SRC },
        'fault-id' => 'mgwd.voltrans.voltransitionjob.onlinefail',
     );

=cut

sub new {
    my ($class,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};
    bless $self, $class;

    $self->{_owner} = $$;

    # Use the server object they specify, or just create a default one
    $self->{_Filer} = $args{Filer} || new Server();

    $self->{_set} = 0;
    delete $args{Filer};
    $self->{_id} = $args{'fault-id'};
    delete $args{'fault-id'};
    croak 'Cannot create a Userspace FIJI object without a fault-id' unless ($self->{_id});

    NOTE::reference($self);
    return $self;
}

sub _set {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment("Setting Userspace FIJI '%s'", $self->{_id});

    my $text = $self->_fijiEnable(@_);
    croak "Error setting Userspace FIJI: $text" if ($text =~ /ERROR:/);
    $self->{_set} = 1;
}

=head2 ::set

Set a Userspace FIJI.

This will activate the FIJI on filer with which it was associated when the
object was created (see ::new).

=over

=item Arguments

B<host> => I<hostname>

I<Optional.>
Specifies the host that should have the fiji set. If omitted, sets on all
hosts.

=item Examples

=back

     $fiji->set;

=cut

sub set {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    _set(@_);
}

=head2 ::clear

Clear a Userspace FIJI.

This will clear the FIJI.

=over

=item Arguments

B<host> => I<hostname>

I<Optional.>
Specifies the host that should have the fiji cleared. If omitted, clears on all
hosts.

=item Examples

=back

     $fiji->clear;

=cut

sub clear {
    my ($self) = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Clearing a fiji which wasn't set is benign, and this complicates calling
    # set & clear with different sets of nodes, so just skip this test...
    #return unless $self->{_set};

    NOTE::logcomment("Clearing Userspace FIJI '%s'", $self->{_id});

    my $text = $self->_fijiDisable(@_);
    croak "Error clearing Userspace FIJI: $text" if ($text =~ /ERROR:/);
    $self->{_set} = 0;
}

=head2 ::modify

Modifies a Userspace FIJI.

=over

=item Arguments

B<host> => I<hostname>

I<Optional.>
Specifies the host that should have the fiji modified. If omitted, modifies on
all hosts.

Any arguments allowed by set-fault-mode.

=item Examples

=back

     $fiji->modify('fault-mode' => 'GX_FAULT_MODE_ALWAYS');

=cut

sub modify {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment("Modifying Userspace FIJI '%s'", $self->{_id});

    my $text = $self->_fijiModify(@_);
    croak "Error modifying Userspace FIJI: $text" if ($text =~ /ERROR:/);
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ($$ != $self->{_owner});
    return if ($self->{_destroyed});
    $self->{_destroyed} = 1;
    $self->clear();
}

sub _fijiEnable {
    my $self = shift;
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $id = $self->{_id};
    my $Server = $self->{_Filer};
    my $host = $args{host};
    my $current_host = $Server->get_execution_server();
    my @servers;
    if (defined $host) {
        @servers = ($host);
    } else {
        @servers = ((map { $_ ne $current_host ? $_ : () } $Server->servers()),
                    $current_host);
    }

    my $result = '';
    foreach my $server (@servers) {
        $Server->set_execution_server('-Server'=>$server)
            if $current_host ne $server;
        $current_host = $server;
        $Server->cli(cmd => "set -privilege test");
        $result .= $Server->debug_fault_injection_enable_fault('fault-id'=>$id) . '\n';
    }
    return $result;
}

sub _fijiDisable {
    my $self = shift;
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $id = $self->{_id};
    my $Server = $self->{_Filer};
    my $host = $args{host};
    my $current_host = $Server->get_execution_server();
    my @servers;
    if (defined $host) {
        @servers = ($host);
    } else {
        @servers = ((map { $_ ne $current_host ? $_ : () } $Server->servers()),
                    $current_host);
    }

    my $result = '';
    foreach my $server (@servers) {
        $Server->set_execution_server('-Server'=>$server)
            if $current_host ne $server;
        $current_host = $server;
        $Server->cli(cmd => "set -privilege test");
        $result .= $Server->debug_fault_injection_disable_fault('fault-id'=>$id) . '\n';
    }
    return $result;
}

sub _fijiModify {
    my $self = shift;
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $id = $self->{_id};
    my $Server = $self->{_Filer};
    my $host = $args{host};
    my $current_host = $Server->get_execution_server();
    my @servers;
    if (defined $host) {
        @servers = ($host);
    } else {
        @servers = ((map { $_ ne $current_host ? $_ : () } $Server->servers()),
                    $current_host);
    }

    my $result = '';
    foreach my $server (@servers) {
        $Server->set_execution_server('-Server'=>$server)
            if $current_host ne $server;
        $current_host = $server;
        $Server->cli(cmd => "set -privilege test");
        $result .= $Server->debug_fault_injection_set_fault_mode(@_, 'fault-id'=>$id) . '\n';
    }
    return $result;
}

=head1 SEE ALSO

L<C<NOTE::Fiji>|engineering::ontap_tests::main::docs::lib-NOTE-Fiji-pm/>

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

=head1 STATUS

Public.

$Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/UserspaceFiji.pm#1 $

$Date: 2013/01/22 $

=head1 AUTHOR

Author information is in the
L<NOTE overview|engineering::ontap_tests::main::docs::lib-NOTE-pm/::author>.

=cut

1;
