#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/VIF.pm#1 $
#
## @summary  GNOTE GX VIF module
## @author   tmichael@netapp.com
## @status   Private
## @pod      here
package NOTE::GX::VIF;
use Carp qw( croak );

our @ISA = qw(NOTE::GX);

sub new {
    my ($class,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_debug}   = $args{debug};
    $self->{_owner}   = $$;

    unless( $args{server} ) {
	if ( ref $args{Vserver} eq 'HASH' ) {
	    $args{server} = NOTE::GX::Vserver->connect( %{$args{Vserver}} );
	} elsif ( ref $args{Vserver} eq 'NOTE::GX::Vserver' ) {
	    $args{server} = $args{Vserver};
	} elsif ( ref $args{Node} eq 'HASH' ) {
	    $args{server} = NOTE::GX::Node->connect( %{$args{Node}} );
	} elsif ( ref $args{Vserver} eq 'NOTE::GX::Node' ) {
	    $args{server} = $args{Node};
	} else {
	    $args{server} = NOTE::GX::Vserver->connect;
	}
    }

    $self->{_container} = $args{server};

    if ( $args{name} ) {
	$self->{_name} = $args{name};
	# Allow named interfaces that behave like anonymous interfaces:
	$self->{_anonymous} = $args{anonymous};
    } else { # Create a unique name
	$self->{_name} = $self->newName;
	$self->{_anonymous} = 1;
    }
    bless $self, $class;

    return $self;
}

sub attach {
    my ($this,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;

    if ( ref $this ) { # Already an object
	$self = $this;
    } else {
	croak "Virtual Interface attach requires a name" unless( $args{name} );
	$self = $this->new( %args );
    }
    $self->status; # make sure this vif exists

    return $self;
}

sub create {
    my ($this,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;

    if ( ref $this ) { # Already an object
	$self = $this;
    } else {
	$self = $this->new( %args );
    }

    delete $args{debug};
    delete $args{name};
    delete $args{anonymous};
    delete $args{container};

    $self->server->network_interface_create(
	server => $self->container->name,
	vif    => $self->name,
	%args,
    );
    NOTE::reference( $self );
    return $self;
}

sub delete {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->network_interface_delete(
	server => $self->container->name,
	vif    => $self->name
    );
}


sub container {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_container}
}

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_container}->server
}

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name}
}

sub status {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->server->network_interface_show(
	server => $self->container->name,
	vif    => $self->name,
	%args,
    )->[0];
}

sub migrate {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->network_interface_migrate(
	server => $self->container->name,
	vif    => $self->name,
	%args,
    );
}

sub revert {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->network_interface_revert(
	server => $self->container->name,
	vif    => $self->name,
    );
}


=head2 ::modify

Modify the properties of a virtual interface.

=over

=item Arguments

B<home-node> => I<nodename>

Home Node

B<home-port> => I<(netport|trunk)>

Home Port

B<address> => I<IP Address>

Network Address

B<netmask> => I<netmask>

Netmask

B<status-admin>  => I<(up|down)>

Administrative Status

B<failover-policy> => I<(nextavail|priority|disabled)>

Failover Policy

B<firewall-policy> => I<policy>

Firewall Policy

B<auto-revert> => I<(true|false)>

Auto Revert

B<use-failover-group> => I<(enabled|disabled)>

Use Failover Group

=item Example

=back

     print $vserver->vif( role => 'data' )->[0]->address, "\n";

     $node->vif( name => 'data1' )->modify(
         address => '192.168.0.122',
         netmask => '255.255.255.0',
     );

=cut

sub modify {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->network_interface_modify(
	server => $self->container->name,
	vif    => $self->name,
	%args,
    );
}

sub destroy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->delete
}

# Don't destroy during teardown
sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_spare_me} = 1
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    return if ( NOTE->ALL_SPARED );
    return if ( !$self->{_anonymous} );
    $self->{_destroyed} = 1;
    if ( $self->{_spare_me} ) {
    	NOTE::logcomment( "VIF ".$self->name." has been spared" );
	return;
    }
    $self->destroy;
}

1;
