#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Vserver.pm#1 $
#
## @summary  GNOTE Vserver module
## @author   dl-dpgqa-automation@netapp.com
## @status   Public
## @pod      here

=head1 NAME

NOTE::GX::Vserver

=head1 DESCRIPTION

This module represents the GX Vserver within a Cluster.

=head1 METHODS

=cut

package NOTE::GX::Vserver;

use strict;
use warnings;
no warnings 'redefine';

our @ISA = qw(NOTE::GX);
use GNOTE::GX::Cluster;
use CIFS::Setup_GX;
use Utils;
use Data::Dumper;
use Options;

use Carp;

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    $self->{_name} = $args{name};
    bless $self, $class;
    NOTE::reference($self);  # Need to be called by GNOTE for igroup clean up

    return $self;
}

sub attach {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->connect(@_);
}


sub connect {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{name} ||= q/$VSERVER/;
    my $self = $class->new( name => NOTE::option( $args{name} ), );

    if ( !defined( $args{Node} ) ) {

        # In this case, there are no arguments (i.e. no hostname) for the
        # filer. I call the filer method anyway and let it decide to
        # handle it or confess.
        $self->{_Cluster} = NOTE::GX::Cluster->connect();
    }
    elsif ( ref $args{Node} eq "HASH" ) {
        $self->{_Cluster} = NOTE::GX::Cluster->connect( %{ $args{Node} } );
    }
    elsif ( ref $args{Node} eq "NOTE::GX::Cluster" ) {
        $self->{_Cluster} = $args{Node};    # This is already an object
    }
    elsif ( ref $args{Node} eq "NOTE::GX::Node" ) {
        $self->{_Cluster} = $args{Node}->Cluster;    # This is already an object
    }
    elsif ( ref \$args{Node} eq "SCALAR" ) {

        # Is just a hostspec
        $self->{_Cluster} = NOTE::GX::Cluster->connect( Node => $args{Node} );
    }
    else {
        confess "Invalid filer specification: " . ref \$args{Node} . "\n";
    }

    # Check that the vserver actually exists
    eval { $self->status };
    confess $@ if ($@);

    return $self;
}

=head2 ::Root

Returns the GX::Volume object representing the root volume of this Vserver object.

=over

=item Arguments

I<None>

=item Example

    my $sroot = $vserver->Root();

=back

=cut

sub Root {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name = $self->status()->{'rootvolume'};
    my $root = NOTE::GX::Volume->attach( name => $name, Vserver => $self, Node => $self->Node, tag => "root volume of Vserver " . $self->name );
    $root->{_path} = $root->_junction_path("/");
    return $root;
}

=head2 ::status

Returns a HASHREF containing status of this Vserver

=over

=item Arguments

I<None>

=item Example

    my $status = $vserver->status;

=back

=cut

sub status {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($package, $filename, $line, $subroutine) = caller(1);
    print "package: $package\nfilename: $filename\nline: $line\nsubroutine: $subroutine\n";
    return $self->server->vserver_show( vserver => $self->name )->[0];
}

sub modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->server->vserver_modify(
        vserver => $self->name,
        %args,
    );
}

=head2 ::Cluster

Returns the GX::Cluster object that this Vserver object is associated with.

=over

=item Arguments

I<None>

=item Examples

=back

    my $clu = $vserver->Cluster();

=cut

sub Cluster { shift->{_Cluster} }

=head2 ::Node

Returns a GX::Node object within the cluster where this vserver resides.

=over

=item Arguments

I<None>

=item Examples

=back

    my $node = $vserver->Node();

=cut

sub Node    { shift->Cluster->Node }
sub server  { shift->Cluster->server }

=head2 ::name

Returns the name of this Vserver.

=over

=item Arguments

I<None>

=item Examples

=back

    my $name = $vserver->name

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->{_name};
}

=head2 ::lifs

Returns a list of LIF entries for this Vserver.

=over

=item Arguments

B<name> => I<string>

I<Optional>

Returns information for a specific named LIF. Default: all LIFs in the current vserver

B<role> => I<string>

I<Optional>

One of data, node-mgmt, intercluster or cluster-mgmt. If not specify then all roles are returned.

B<link> => I<up|down>

I<Optional>

Indicates to return LIF information that is in a certain state. Default: both up and down

=item Examples

    my @lif_info = $vserver->lifs(role=>"data"); # Return all data role vifs

=back

=cut

sub lifs {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @list;

    return $self->Cluster->lifs( %args, vserver => $self->name );
}

=head2 ::lif_address

Return a single working LIF IP address for this Vserver.

=over

=item Arguments

B<role> => I<string>

I<Optional>

One of data, node-mgmt, intercluster or cluster-mgmt. Default: data

B<link> => I<up|down>

I<Optional>

Indicates to return a LIF IP address in a certain state. Default: up

=item Example

    my $data_vif = $vserver->lif_address();

=back

=cut

sub lif_address {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{role} = "data" unless exists $args{role};
    $args{link} = "up"   unless exists $args{link};

    my @lifs = $self->Cluster->lif_list( %args, vserver => $self->name );
    if (@lifs) {
        return $lifs[0]->{address};
    }
    else {
        confess 'No data logical interfaces for vserver ' . $self->name;
    }
}

=head2 ::lif_modify

Modifies the up/down status of lifs on the vserver.
Lif name must be specified.
Returns a boolean to indicate whether the modify was successful or not

=over

B<name> => I<string>

I<Required>

The name of the lif whose status change is required.

B<link> => I<up|down>

I<Required>

Indicates the desired new status(admin) of the lif. The only allowed values are 'up' or 'down'.

=item Examples

=back

    $vserver->lif_modify(name=>'dpg-gx40-data1', link => 'down');

=cut

sub lif_modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Node name should not be provided" if exists $args{node};

    confess "Vserver name if provided should be the same as this vserver"
      if exists $args{vserver} and $args{vserver} ne $self->name;

    delete $args{vserver} if exists $args{vserver};

    return $self->Cluster->lif_modify( %args, vserver => $self->name );
}

sub export_policy {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $r = $self->server->vserver_export_policy_show( vserver => $self->name, );
    return wantarray ? %{ $r->[0] } : $r->[0]->{policyname};
}

=head2 ::nfs

Creates or modifies an NFS configuration for a vserver.
Arguments are passthru to ONTAP directly.

=over

=item Arguments

I<ONTAP Arguments:>

for create:

  [ access {true|false} ]                      General NFS access (default: true)
  [ v3 {enabled|disabled} ]                    NFS v3 (default: enabled)
  [ v4.0 {enabled|disabled} ]                  NFS v4.0 (default: disabled)
  [ udp {enabled|disabled} ]                   UDP Protocol (default: enabled)
  [ tcp {enabled|disabled} ]                   TCP Protocol (default: enabled)
  [ spinauth {enabled|disabled} ]              Spin Authentication (default: disabled)
  [ default-win-user <text> ]                  Default Windows User
  [ v4.0-acl {enabled|disabled} ]              NFSv4.0 ACL Support (default: disabled)
  [ v4.0-read-delegation {enabled|disabled} ]  NFSv4.0 Read Delegation Support (default: disabled)
  [ v4.0-write-delegation {enabled|disabled} ] NFSv4.0 Write Delegation Support (default: disabled)
  [ v4-id-domain <nis domain> ]                NFSv4 ID Mapping Domain (default: defaultv4iddomain.com)
  [ v4.1 {enabled|disabled} ]                  NFSv4.1 Minor Version Support (default: disabled)
  [[rquota] {enabled|disabled}]                Rquota Enable (default: disabled)
  [ v4.1-pnfs {enabled|disabled} ]             NFSv4.1 Parallel NFS Support   (default: enabled)
  [ v4.1-acl {enabled|disabled} ]              NFSv4.1 ACL Support (default: disabled)
  [ vstorage {enabled|disabled} ]              NFS vStorage Support (default: disabled)

for modify:

  [[access] {true|false}]                      General NFS access
  [ v3 {enabled|disabled} ]                    NFS v3
  [ v4.0 {enabled|disabled} ]                  NFS v4.0
  [ udp {enabled|disabled} ]                   UDP Protocol
  [ tcp {enabled|disabled} ]                   TCP Protocol
  [ spinauth {enabled|disabled} ]              Spin Authentication
  [ default-win-user <text> ]                  Default Windows User
  [ v4.0-acl {enabled|disabled} ]              NFSv4.0 ACL Support
  [ v4.0-read-delegation {enabled|disabled} ]  NFSv4.0 Read Delegation Support
  [ v4.0-write-delegation {enabled|disabled} ] NFSv4.0 Write Delegation Support
  [ v4-id-domain <nis domain> ]                NFSv4 ID Mapping Domain
  [ v4.1 {enabled|disabled} ]                  NFSv4.1 Minor Version Support
  [ rquota {enabled|disabled} ]                Rquota Enable
  [ v4.1-pnfs {enabled|disabled} ]             NFSv4.1 Parallel NFS Support
  [ v4.1-acl {enabled|disabled} ]              NFSv4.1 ACL Support
  [ vstorage {enabled|disabled} ]              NFS vStorage Support

=item Examples

=back

    $vol->Vserver->nfs(v3=>"enabled"); # Enables NFSv3
    $vsrv->nfs(access=>"false"); # Disables NFS access for this vserver

Working example:

      1 use strict;
      2 use Data::Dumper;
      3 use GNOTE;
      4 eval {
      5     my $ut = NOTE::SnapMirrorUtil->new();
      6     my $vl = $ut->create_volume(tag=>'src');
      7     my $nfsout = $vl->Vserver->nfs();
      8     print "NFS settings:\n".Dumper($nfsout);
      9 };
     10 logresult('FAIL',$@) if $@;
     11 NOTE::teardown;

Output from example above:

   NFS settings:
   $VAR1 = {
             'v4.0-migration' => 'disabled',
             'v4' => 'disabled',
             'v4-id-domain' => 'defaultv4iddomain.com',
             'udp-max-xfer-size' => '32768',
             'v3-connection-drop' => 'enabled',
             'v2-enabled' => 'disabled',
             'v4.1-implementation-date' => '"Wed Dec 31 16:00:00 1969"',
             'force-spinnp-readdir' => 'false',
             'v4.1-implementation-domain' => 'defaultv41impliddomain.com',
             'udp-enabled' => 'enabled',
             'v4.1-acl' => 'disabled',
             'spinauth' => 'disabled',
             'v4-write-delegation' => 'disabled',
             'v4-fsid-change' => 'enabled',
             'chown-mode' => 'use_export_policy',
             'v3-enabled' => 'enabled',
             'v4-acl-preserve' => 'enabled',
             'rpcsec-ctx-high' => '0',
             'v3' => 'enabled',
             'rquota' => 'disabled',
             'ntfs-unix-security-ops' => 'use_export_policy',
             'tcp-enabled' => 'enabled',
             'v4.1-referrals' => 'disabled',
             'trigger' => '60',
             'rpcsec-ctx-idle' => '0',
             'v4.0-referrals' => 'disabled',
             'v4-lease-seconds' => '30',
             'v4.0-acl' => 'disabled',
             'v4-validate-symlinkdata' => 'disabled',
             'tcp-max-xfer-size' => '65536',
             'v4.1' => 'disabled',
             'v4.1-pnfs' => 'enabled',
             'default-win-user' => '-',
             'v4.0-read-delegation' => 'disabled',
             'udp' => 'enabled',
             'trace-enabled' => 'false',
             'v4-read-delegation' => 'disabled',
             'enable-ejukebox' => 'false',
             'v4.1-implementation-name' => 'defaultv41implidname',
             'access' => 'true',
             'vserver' => 'vs0',
             'v3-require-read-attributes' => 'false',
             'v4.0' => 'disabled',
             'tcp' => 'enabled',
             'vstorage' => 'disabled',
             'v4-grace-seconds' => '45',
             'v3-fsid-change' => 'enabled',
             'v4.0-write-delegation' => 'disabled',
             'v2' => 'disabled',
             'v4-acl' => 'disabled',
             'v4.1-migration' => 'disabled'
           };

=cut

sub nfs {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $r = undef;
    my $mode = ((join ",", keys %args) =~ /v4/ && $self->Node->tricky_parity()) ? "test" : "diagnostic"; # NFSv4 required test mode for tricky parity node
    eval { ($r) = $self->server->vserver_nfs_show( vserver => $self->name ) };	# pull NFS info from Vserver and if error then
    confess $@ if( $@ && $@ !~ /There are no entries matching your query/ );	# die unless error is 'no matching entries'
    return $r unless %args;							# return result or undef if no arguments submitted
    my $cmd = 'vserver_nfs_create';						# command must be create NFS if nothing found
    $cmd = 'vserver_nfs_modify' if $r;						# or run 'modify' for existing NFS
    $self->Node->exec(level=>$mode, cmd=>$cmd, vserver => $self->name, %args );	# configuration of Vserver
    ($r) = $self->server->vserver_nfs_show( vserver => $self->name);		# pull and return new configuration, convert array of hashes to hash
    return $r;									# and final return is a hash reference
}

=head2 ::cifs_setup

Setups a cifs server for this Vserver.
Calls CIFS::Setup_GX::cifs_setup to setup a cifs server.

Most of the parameters are passed to CIFS::Setup_GX::cifs_setup as is. Check CIFS::Setup_GX::cifs_setup for complete list.
Parameters below are the most important ones.

If CIFS server is already configured, then passed domain name is compared to configured domain name. If domain names match
then existing configuration will be used as is.

This method also performs some minor checks like - is active directory sever alive, can we ping it using all data interfaces.

=over

=item Arguments


B<cifs_server> => I<string>

I<Optional>

CIFS server name for vserver. Default is to use concatenation of node name and vserver name with bare word 'CIFS'.
If parameter exist but its empty then C<CIFS_SERVER> cli parameter is used. Whole string should be no more then
15 characters long or it will be cut off after 15th character.

B<cifs_domain_fullname> => I<string>

I<Required>

Full domain name, such as DPGQA.LAB.NETAPP.COM. If parameter is not set, then library will attempt to read CIFS_DOMAIN_LONG cli parameter, and will set to
CIFSQA.NETAPP.COM. So make sure you set it in a script or in your environment.

B<cifs_domain_shortname> => I<string>

I<Optional>

Short name for domain, used as workgroup parameter by filer. Default is taken from fully qualified name. For example DPGQA.LAB.NETAPP.COM will become
DPGQA. If parameter exist but its empty then C<CIFS_DOMAIN> cli parameter is used before trying to create one from long name.

B<ad_server> => I<string>

I<Required>

Active directory host name(s). Can be a comman separated list of multiple names with no spaces between.
If parameter exist but its empty then C<AD_SERVER> cli parameter is used.

B<ad_server_ip> => I<string>

I<Required>

Address of active directory server(s). Can be comma separated list of addresses. List must exactly correspond to names from
B<ad_server> parameters. For example if AD1 has ip of 10.0.0.11 and AD2 has ip of 10.0.0.12 then set should be given as
as ad_erver=>"AD1,AD2", ad_server_ip=>"10.0.0.11,10.0.0.12". Make sure to keep correct order of parameters. If order will be
incorrect, like ad_erver=>"AD1,AD2", ad_server_ip=>"IP2,IP1", then library will try to use first name with second ip and setup
will fail. If parameter exist but its empty then C<AD_SERVER_IP> cli parameter is used

B<ad_server_user> => I<string>

I<Required>

Administrator's user name used to login to active directory to register new CIFS machine (vserver) as member of CIFS domain.
If parameter exist but its empty then C<AD_SERVER_USER> cli parameter is used. Please do not mix it with Client user name.
Client's user name used only to map a share. This one is used to mess with domain controller.

B<ad_server_password> => I<string>

I<Required>

Administrator's password used to login to active directory to register new CIFS machine (vserver) as member of CIFS domain.
If parameter exist but its empty then C<AD_SERVER_PASSWORD> cli parameter is used.

B<nis_server> => I<string>

I<Optional>

Address of NIS server. If parameter exist but its empty then C<NIS_SERVER> cli parameter is used

B<nis_domain> => I<string>

I<Optional>

NIS domain name. If parameter exist but its empty then C<NIS_DOMAIN> cli parameter is used

B<timezone> => I<string>

I<Optional>

Time zone string. If parameter exist but its empty then C<TIMEZONE> cli parameter is used

B<nsswitch> => I<string>

I<Optional>

Name resolution order. Default is C<file,nis>. If parameter exist but its empty then C<NS_SWITCH> cli parameter is used

B<ou> => I<string>

I<Optional>

Organizational unit. Library does not require this parameter, however domain itself can be configured to use OU to
separate different engineering groups within same domain. Please consult engineering support if setup has failed
due to OU parameter. If parameter exist but its empty then C<OU> cli parameter is used.

B<use_force> => I<boolean>

I<Optional>

Use force, skip precheck, go directly to CIFS::Setup_GX::cifs_setup.

=back

=cut

sub cifs_setup {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $force = delete $args{use_force};    # should we use force?
    delete $args{storage};		    # backward compatibility with old code

    my $options = new Options;		    # CLI parameters reader
    $options->process();		    # CIFS::Setup_GX::cifs_setup creates it anyway
    my $params = $options->get_hash();	    # but we have to do minor pre-check, so get your own object

    $args{cifs_domain_fullname} = $args{cifs_domain_fullname} =~ /\w+/ ? $args{cifs_domain_fullname} : $params->{CIFS_DOMAIN_LONG}; # arbitrate between args & CLI
    confess "Missing fully qualified domain name." unless $args{cifs_domain_fullname} =~ /\w+\./;				    # or fail if missing

    ( $args{cifs_server} = $args{cifs_server} ? $args{cifs_server} : uc($self->Node->name . $self->name . "CIFS") ) =~ s/[-\._]//g; # come up with CIFS server name, clean up,
    $args{cifs_server} = uc(substr( $args{cifs_server}, 0, 15));								    # upper case it and shrink to 15 characters

    my $info; # placeholder for domain info - hash reference with data from filer itself
    eval { $info = ( $self->server->vserver_cifs_show( vserver => $self->name ) )[0] };     # pull info from node and die
    confess $@ if $@ && !( $@ =~ /no entries/i );					    # if error other then "no entries"

    if( !$force && $$info{'domain'} =~ /\w+/ ) {									# if $force, then call setup without check
	if( uc($args{cifs_domain_fullname}) eq $$info{'domain'} ) {							# or compare existing domain with requested
	    NOTE::logcomment( "Vserver " . $self->name . " already have a cifs server\n" . Dumper( $info ) );		# name.
	    return 1;
	} else {													# it's better to fail if domains don't match
	    confess "Domain is already configured but it doesn't match required domain name.\n".			# because script may not have username and
		    "Required name is [".$args{cifs_domain_fullname}."], preconfigured name is [".$$info{'domain'}."]";	# password for original domain in order to
	}														# delete wrong entry
    }

    CIFS::Setup_GX::cifs_setup(node => $self->Node->name, %args);

    eval { $info = ( $self->server->vserver_cifs_show( vserver => $self->name ) )[0] };     # Don't really need eval, but if if I skip it, then traces will point
    confess $@ if $@;									    # to absolutely unknown location instead of Vserver.pm. So eval then

    NOTE::logcomment( "CIFS Server Config:\n".Dumper($info) );							# Dump info, but do not trust it just yet
    confess "Domain mismatch\nRequired name is [".$args{cifs_domain_fullname}."], preconfigured name is [".	# compare requested domain with
	     $$info{'domain'}."]"										# actually returned from filer and
				 if($$info{'domain'} ne uc($args{cifs_domain_fullname}));			# fail if they mismatch

    return 1;
}

=head2 locks_show

Run C<vserver locks show> on the vserver this object is connected to. Returns
a list-reference of zero or more hash-references containing data on all
active locks on the Vserver. If there are no locks currently, returns an empty
list-reference. If an error occurs, an exception is thrown.

Each hash-reference has the following keys:

    bytelock-exclusive      bytelock-length     bytelock-mandatory
    bytelock-offset         bytelock-soft       bytelock-super
    client-id               delegation-type     filer
    lif                     lock-state          lockid
    node                    oplock-level        owner-id
    path                    protocol            sharelock-mode
    sharelock-soft          sharelock-super     type
    vif                     volume              vserver

The various arguments that are passable in the B<opts> hash table (see
below) are used to restrict the list of locks returned by volume, protocol,
type, etc.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments. The arguments are used to narrow down the list of
locks returned. Fields that are passed in the arguments are matched
exactly:

=over

=item B<vserver> = string|object

The v-server to query for locks; defaults to the v-server whose object
called the method. May be the string name or a B<NOTE::GX::Vserver> object.

=item B<volume> = string|object

The volume the lock is on; may be the string name or a B<NOTE::GX::Volume>
object

=item B<path> = string

Path of the object that the lock is on

=item B<lif> = string

Logical interface

=item B<lockid> = string

Lock UUID

=item B<protocol> = string

Lock protocol, one of C<nlm>, C<cifs>, C<nfsv4> or C<fcache>

=item B<type> = string

Lock type, one of C<byte-range>, C<share-level>, C<op-lock> or C<delegation>

=item B<node> = string

Node holding the lock state

=item B<lock-state> = string

Lock state

=item B<bytelock-offset> = integer

Offset into the file where the byte-lock starts

=item B<bytelock-length> = integer

Length of range of the byte-lock

=item B<bytelock-mandatory> = boolean

Whether the byte-lock is mandatory

=item B<bytelock-exclusive> = boolean

Whether the byte-lock is exclusive

=item B<bytelock-super> = boolean

Whether the byte-lock is a super-lock

=item B<bytelock-soft> = boolean

Whether the byte-lock is soft

=item B<oplock-level> = string

One of C<exclusive>, C<read> or C<batch>, match the op-lock level field

=item B<sharelock-mode> = string

Match the shared lock access mode

=item B<sharelock-super> = boolean

Whether the shared lock is a super-lock

=item B<sharelock-soft> = boolean

Whether the shared lock is soft

=item B<delegation-type> = string

One of C<read> or C<write>, match locks with this delegation type

=back

=item Examples

    # Get all the locks on a v-server
    my $locks = $vserver->locks_show();

    # Get the locks for a specific volume
    my $locks = $volume->Vserver->locks_show(volume => $volume);

=back

=head3 Lock types in the Vserver

ONTAP manages different types of locks within a Vserver, depending on the
protocol with which each client mounts a given volume. Some of these locks
are known as C<hard> locks, and others as C<soft> locks. The distinction can
sometimes be unclear.

I<Hard> locks are locks created by the client, either by opening a file
(which creates a I<shared-level> lock under NFSv4 and CIFS) or by explicitly
creating a I<byte-range> lock (with B<fcntl> under NFSv3 and NFSv4, or with
B<LockEx> under CIFS). These locks are created by the client, and can be
removed by the client.

I<Soft> locks are locks created by ONTAP itself, usually in conjunction with
operations by clients. For NFSv4, the soft locks are called "delegations".
For CIFS, they are called "opportunistic locks", or "oplocks". These are not
explicitly created by clients, and cannot be explicitly removed by clients.

Each record returned by this method has a B<type> field, which identifies the
lock as one of C<shared-level>, C<byte-range>, C<delegation> or C<oplock>.

See also the B<lock_mgr> method in F<lib/GNOTE/Client.pm>.

=cut

###############################################################################
#
#   Sub Name:       locks_show
#
#   Description:    Call the vserver_locks_show CDEF and return the resulting
#                   data.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %args     in      hash      Additional args, see POD
#
#   Returns:        Success:    listref of hashrefs
#                   Failure:    dies
#
###############################################################################
sub locks_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $data;

    # They can use the vserver argument to specify a different VS to query.
    # If they don't specify it, use $self. Either way, make sure it's
    # the name-string, and not an object:
    $args{vserver} ||= $self;
    if (ref $args{vserver}) {
        $args{vserver} = $args{vserver}->name;
    }

    # Check to see if volume argument was passed, and if so make sure it's
    # also a string and not an object:
    if ($args{volume} && ref $args{volume}) {
        $args{volume} = $args{volume}->name;
    }

    # This option tells us whether to filter out the CIFS locks that show up
    # as "NULL". They have actually been deleted, but still linger in the
    # vserver for some reason. See BURT503961. Delete it from %args so that
    # it doesn't filter through to the CDEF call.
    my $keep_null = delete $args{keep_null};

    # These options are boolean, and may have been passed as "true"/"false"
    # or Perl-ish boolean values. Normalize them to "true"/"false":
    for (qw(bytelock-mandatory bytelock-exclusive bytelock-super
            bytelock-soft sharelock-super sharelock-soft)) {
        if (exists $args{$_}) {
            $args{$_} = ($args{$_} && $args{$_} !~ /false/i) ? 'true' : 'false';
        }
    }

    # Because vserver_locks_show() throws an exception when there are no locks
    # (rather than just returning undef or an empty listref), we have to
    # catch the error-string and check if it is just the "no locks" error or
    # a "real" error.
    my $result = eval {
        $data = $self->server->vserver_locks_show(%args);
        1;
    };
    if (! $result) {
        if ($@ =~ /no entries matching your query/) {
            # This just means that there are no locks currently. It's not a
            # "real" error...
            $data = [];
        } else {
            confess $@;
        }
    }

    if (! $keep_null) {
        # Filter out the records that have NULL in the oplock-level field
        @{$data} = grep { lc $_->{'oplock-level'} ne 'null' } @{$data};
    }

    return $data;
}

# lif_list is used by SAN

=head2 ::lif_list

Returns an array of HASHREFs each containing information for a LIF in this vserver
Qualifying parameters can be use to restrict entries in the list.

=over

=item Arguments

Additional filtering of output can be done by passing following arguments:

B<lif> I<Optional> Logical Interface

B<role> I<Optional> Role

B<data-protocol> I<Optional> Data Protocol

B<home-node> I<Optional> Home Node

B<home-port> I<Optional> Home Port

B<curr-node> I<Optional> Current Node

B<curr-port> I<Optional> Current Port

B<status-oper> I<Optional> Operational Status

B<numeric-id> I<Optional> *Numeric ID

B<is-home> I<Optional> Is Home

B<address> I<Optional> Network Address

B<netmask> I<Optional> Netmask

B<netmask-length> I<Optional> Bits in the Netmask

B<routing-group> I<Optional> Routing Group Name

B<status-admin> I<Optional> Administrative Status

B<failover-policy> I<Optional> Failover Policy

B<firewall-policy> I<Optional> Firewall Policy

B<auto-revert> I<Optional> Auto Revert

B<sticky> I<Optional> *Sticky Flag

B<use-failover-group> I<Optional> Use Failover Group

B<dns-zone> I<Optional> DNS Zone

B<allow-lb-migrate> I<Optional> *Load Balancing Migrate Allowed

B<lb-weight> I<Optional> *Load Balanced Weight of a Lif

B<failover-group> I<Optional> Failover Group Name

=item Examples

Initial config:

 cpeter-vsim1::*> net int show
   (network interface show)
             Logical    Status     Network            Current       Current Is
 Server      Interface  Admin/Oper Address/Mask       Node          Port    Home
 ----------- ---------- ---------- ------------------ ------------- ------- ----
 cpeter-vsim1
             cluster      up/up    172.31.5.245/18    cpeter-vsim1  e0a     true
             mgmt1        up/up    10.53.5.245/18     cpeter-vsim1  e0c     true
 vs0
             data1        up/up    10.53.5.246/18     cpeter-vsim1  e0d     true
             data2        up/up    10.53.5.247/18     cpeter-vsim1  e0d     true
 4 entries were displayed.

Print routing group of first lif:

 my $routing_group = ($vol->Vserver->lif_list)[0]->{'routing-group'};

Print address of last data interface:

 my $address = ($vol->Vserver->lif_list(role=>'data'))[-1]->{'address'};

Print all interfaces:

     27 sub do_setup {
     28     my $vol = NOTE::Volume->create( 'mirror-type' => "RW", Node=>$FILER, Vserver => $VSERVER, size=>$VOL_SIZE, 'snapshot-policy' => 'none'  );
     29     my @lifs = $vol->Vserver->lif_list;
     30     print Dumper(\@lifs);

Output:

 $VAR1 = [
           {
             'curr-filer' => 'cpeter-vsim1',
             'lb-weight' => 'load',
             'curr-node' => 'cpeter-vsim1',
             'failover-group' => '""',
             'home-port' => 'e0d',
             'sticky' => 'false',
             'firewall-policy' => 'data',
             'failover-policy' => 'nextavail',
             'netmask' => '255.255.192.0',
             'address' => '10.53.5.246',
             'server' => 'vs0',
             'home-node' => 'cpeter-vsim1',
             'auto-revert' => 'false',
             'home-filer' => 'cpeter-vsim1',
             'allow-lb-migrate' => 'false',
             'status-admin' => 'up',
             'use-failover-group' => 'disabled',
             'data-protocol' => 'iscsi',
             'dns-zone' => 'none',
             'curr-port' => 'e0d',
             'status-oper' => 'up',
             'is-home' => 'true',
             'netmask-length' => '18',
             'vif' => 'data1',
             'numeric-id' => '1025',
             'role' => 'data',
             'lif' => 'data1',
             'routing-group' => 'd10.53.0.0/18'
           },
           {
             'curr-filer' => 'cpeter-vsim1',
             'lb-weight' => 'load',
             'curr-node' => 'cpeter-vsim1',
             'failover-group' => '""',
             'home-port' => 'e0d',
             'sticky' => 'false',
             'firewall-policy' => 'data',
             'failover-policy' => 'nextavail',
             'netmask' => '255.255.192.0',
             'address' => '10.53.5.247',
             'server' => 'vs0',
             'home-node' => 'cpeter-vsim1',
             'auto-revert' => 'true',
             'home-filer' => 'cpeter-vsim1',
             'allow-lb-migrate' => 'false',
             'status-admin' => 'up',
             'use-failover-group' => 'system-defined',
             'data-protocol' => [
                                  'nfs',
                                  'cifs'
                                ],
             'dns-zone' => 'none',
             'curr-port' => 'e0d',
             'status-oper' => 'up',
             'is-home' => 'true',
             'netmask-length' => '18',
             'vif' => 'data2',
             'numeric-id' => '1026',
             'role' => 'data',
             'lif' => 'data2',
             'routing-group' => 'd10.53.0.0/18'
           }
         ];

=back

=cut

sub lif_list {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $args{server} = $self->name;		# set or replace vserver name with one of the object
    return $self->Cluster->lif_list( %args );	# then pass everything to Cluster parent
}

=head2 ::get_iscsi_lif

Returns an array of HASHREFs each containing information for iSCSI LIFs in this vserver. See B<lif_list> method
for details on output and examples.

=over

=item Arguments

none

=back

=cut

sub get_iscsi_lif {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    shift->lif_list( 'data-protocol' => 'iscsi', role => 'data' );
}

=head2 set_iscsi_lif

Creates new iSCSI LIF.
If user provides the required arguments, creates with user provided values,
If not, search for a free data ip and create with the available free data ip.
If not available, converts an existing non-iscsi lif.
And returns hash reference for the created iscsi lif.
hash references for created interface.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<no_convert_existing_lif> => I<1|0>

Default: 0, means false
If user not providing any arguments required to create the iscsi lif,
and do not want to convert the existing non-iscsi lif, but want the iscsi
lif to be created out of free data ip, provide 1.

=item B<lif> => I<string>

Logical Interface name for the iscsi lif to be created

=item B<address> => I<ip_address>

ip address for the iscsi lif to be created

=item B<netmask> => I<netmask>

netmask for the iscsi lif to be created

=item B<home-port> => I<port name>

port name for the iscsi lif to be created

=item B<home-node> => I<home node name>

home node name for the iscsi lif to be created

=item B<role> => I<data>

Role, should be set to 'data' unless doing negative testing

See B<lif_list> method for several other optional arguments.

=back

=item Returns

=over

=item B<Success:>

Returns hash reference to the created iscsi lif
See below Notes section, for the returned hash structure.

=item B<Failure:>

scenario 1: confesses, If user does not provide the arguments, and free data ips
            are not found, and user provide the argument no_convert_existing_lif=1
scenario 2: if error creating the iscsi lif

=back


=item Notes

=over

=item 1

The creation of iscsi lif takes several optional arguments. the required
arguments are mentioned in this function pod. for all other options, refer
the cli or B<lif_list> function.

=item 2

The returned hash structure would be as follows:
my $iscsi_lif = $src_vol->Vserver->set_iscsi_lif;

Returned $iscsi_lif:

 {
   'curr-filer' => 'dpg-reg-3170-f1',
   'lb-weight' => 'load',
   'curr-node' => 'dpg-reg-3170-f1',
   'failover-group' => '""',
   'home-port' => 'e0d',
   'sticky' => 'false',
   'firewall-policy' => 'data',
   'failover-policy' => 'nextavail',
   'netmask' => '255.255.192.0',
   'address' => '10.53.5.246',
   'server' => 'vs0',
   'home-node' => 'dpg-reg-3170-f1',
   'auto-revert' => 'false',
   'home-filer' => 'dpg-reg-3170-f1',
   'allow-lb-migrate' => 'false',
   'status-admin' => 'up',
   'use-failover-group' => 'disabled',
   'data-protocol' => 'iscsi',
   'dns-zone' => 'none',
   'curr-port' => 'e0d',
   'status-oper' => 'up',
   'is-home' => 'true',
   'netmask-length' => '18',
   'vif' => 'data1',
   'numeric-id' => '1025',
   'role' => 'data',
   'lif' => 'data1',
   'routing-group' => 'd10.53.0.0/18'
 };

=back

=item Examples

=back

	#For creating iscsi lif, without providing any values
    my $iscsi_lif = $src_vol->Vserver->set_iscsi_lif;

    #creating an iscsi lif, providing the ip and port
    my $iscsi_lif = $src_vol->Vserver->set_iscsi_lif('address' => '10.10.10.22',
                                                     'home-port' => 'e4a',
                                                    );

    #creating the iscsi lif, without providing values, but do not want to convert
    #existing data lif
    my $iscsi_lif = $src_vol->Vserver->set_iscsi_lif(no_convert_existing_lif => 1);

=cut

sub set_iscsi_lif {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $no_convert_existing_lif = delete $args{no_convert_existing_lif};
    if(!defined $no_convert_existing_lif) {
        $no_convert_existing_lif = 0; #false, means convert
    }

    my @reqd_params = ('lif','role', 'address', 'netmask', 'home-port',
                       'home-node', 'data-protocol', 'server',);

    #Check whether user has provided ip address and port
    #If provided, create the iscsi lif with user provided values
    if( (defined $args{'address'} ) and (defined $args{'home-port'}) ) {
        NOTE::logcomment('Creating the iscsi lif with user provided ip and home-port');

        #for all other required parameters not provided by user, get them from natehost file
        my $node1 = Utils::make_host($self->Node->name);

        #The netmask can be the same as other data lifs,
        #in case user do not provide this parameter
        $args{'netmask'} = $args{'netmask'} ? $args{'netmask'} : $node1->{'DATA_MASK1'};
        $args{'home-node'} = $args{'home-node'} ? $args{'home-node'} : $self->Node->name;
    }
    else { #user did not provide both the ip and home-port
        #Search all the free data_ips which are not used for any other lif
        #DATA_IP values would be provided in nate_host file of the cluster nodes.
        my @cluster_nodes = $self->Cluster->Nodes;
        my @lifs = $self->Cluster->lif_list();
        NOTE::logcomment('Existing lifs in the cluster: ' . Dumper(@lifs));

        FREE_DATAIP_SEARCH_ALL_NODES:
        foreach my $node_temp ($self->Cluster->Nodes) {
            my $node1 = Utils::make_host($node_temp->name);
            #generally in natehost file any number of data_ips would be given
            #so, search till we run out of data_ips given in natehosts file
            my $n = 0;

            FREE_DATAIP_SEARCH_THIS_NODE:
            while(1){
                $n++;
                my $data_ip_tmp = $node1->{'DATA_IP'. $n};
                if( (!defined $data_ip_tmp) or ($data_ip_tmp eq q{} ) ) {
                    last FREE_DATAIP_SEARCH_THIS_NODE;
                }
                my $data_ip_tmp_inuse = 0;
                foreach my $lif_temp (@lifs) {
                    if($lif_temp->{'address'} eq $data_ip_tmp) {
                        $data_ip_tmp_inuse = 1;
                    }
                }
                if(!$data_ip_tmp_inuse) {
                    $args{'address'}   = $args{'address'} ? $args{'address'} : $node1->{'DATA_IP' . $n};
                    $args{'netmask'}   = $args{'netmask'} ? $args{'netmask'} : $node1->{'DATA_MASK' . $n};
                    $args{'home-port'} = $args{'home-port'} ? $args{'home-port'} : $node1->{'DATA_PORT' . $n};
                    $args{'home-node'} = $args{'home-node'} ? $args{'home-node'} : $node_temp->name;
                    last FREE_DATAIP_SEARCH_ALL_NODES;
                }
            }
        }
    }
    #if no free data ips found, check the option of converting an existing lif
    if(! $args{'address'} ) {
        NOTE::logcomment('No free data ips found.');

        #convert an existing lif, if arg no_convert_existing_lif is set to false
        if($no_convert_existing_lif){
            confess 'There are no free data ips to create a iscsi lif';
        }
        else {
            #convert an existing lif
            #get the existing lif list
            my @lifs = $self->Cluster->lif_list();
            #the data lif cannot be converted in single step, it has to be deleted
            #and then iscsi lif is to be created
            #however convert only a non-iscsi lif,
            #otherwise the user cannot create more than one iscsi lif
            DELETE_EXIST_LIF:
            foreach my $lif_temp (@lifs) {
                if( ($lif_temp->{'role'} eq 'data') and ( $lif_temp->{'data-protocol'} ne 'iscsi') and
                     $lif_temp->{'vserver'} eq $self->name()){
                    $self->server->network_interface_delete( server=>$self->name, lif=>$lif_temp->{'lif'} );
                    push @{$self->{_deleted_data_lifs_aref}},$lif_temp;
                    #now get all the deleted lif properties to create iscsi lif out of it
                    foreach my $param (@reqd_params) {
                        $args{$param} = $args{$param} ? $args{$param} : $lif_temp->{$param};
                    }
                    last DELETE_EXIST_LIF;
                }
            }

        }

    }

    if(! $args{'address'} ) {
        confess 'There are no free data ips, nor lifs to convert';
    }

    #set the remaining attribute values required to create the iscsi lif
    #create an unique lif name
    $args{'lif'} = $args{'lif'} ? $args{'lif'} : $self->Node->name . time . '_data';
    #snooze for a second, in case script tries to create multiple iscsi lifs
    #in the same second time
    Tharn::snooze(1);

    $args{'role'} = $args{'role'} ? $args{'role'} : 'data';
    $args{'data-protocol'} = 'iscsi';
    $args{'server'} = $self->name;

    NOTE::logcomment('args passed to creation of iscsi lif: ' . Dumper(%args));

    #Create the iscsi interface
    my $eval_ret = eval {
                       $self->server->network_interface_create(%args);
                       push @{$self->{_created_data_lif_names_aref}},$args{'lif'};
                       1;
                   };
    if( !$eval_ret){
        confess "Error in creating the iscsi lif. Error: $@";
    }

    #Now return the iscsi lif created
    return ($self->get_iscsi_lif(lif=>$args{lif}))[0];
}

=head2 iscsi_lif_delete_all

This function will delete all the iscsi lifs in the cluster.
Does not confess in the case failure, just report the error to logfile.
This is because this function mostly will be called by teardown or DESTROY.

=over

=item Arguments

B<None>

=item Returns

=over

=item B<Success:> Returns 1 on deleting all, 0 on failure

=item B<Failure:> Does not confess. Report the error in logfile.

=back

=item Examples

=back

    #To delete all the existing iscsi lifs
    $myvserver->iscsi_lif_delete_all();

=cut

sub iscsi_lif_delete_all {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @iscsi_lifs = $self->get_iscsi_lif;
    NOTE::logcomment('iscsi_lifs: ' . Dumper(@iscsi_lifs));
    my $all_deleted = 1; #true
    foreach my $lif_temp (@iscsi_lifs) {
        my $ret_eval = eval {
                          $self->server->network_interface_delete(
                             server => $self->name,
                             vif    => $lif_temp->{lif},
                          );
                          1;
                       };
        if(!$ret_eval) {
            #Do not confess, as this function is called mostly as part of teardown or DESTROY
            NOTE::logcomment('WARNING: The iscsi delete of ' . $lif_temp->{lif} . " is failing. Error: $@");
            $all_deleted = 0; #fail
        }

    }
    return $all_deleted;
}

=head2 ::iscsi_show

Returns array of hashrefs iscsi settings if configured or returns undef.

=over

=item Arguments

none

=item Examples

Code:

     27 sub do_setup {
     28     my $vol = NOTE::Volume->create( 'mirror-type' => "RW", Node=>$FILER, Vserver => $VSERVER, size=>$VOL_SIZE, 'snapshot-policy' => 'none'  );
     29     my $iscsi = $vol->Vserver->iscsi_show;
     30     print Dumper( $iscsi );

Output:

 $out = [
          {
            'filername' => 'iqn.1992-08.com.netapp:sn.65abf0f85de511dfa7dc123478563412:vs.2',
            'nodename' => 'iqn.1992-08.com.netapp:sn.65abf0f85de511dfa7dc123478563412:vs.2',
            'status' => 'online',
            'vserver' => 'vs0',
            'alias' => 'vs0'
          }
        ];

=back

=cut

sub iscsi_show {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $show = undef;
    eval{ $show = $self->server->vserver_iscsi_show( vserver=>$self->name ) };

    confess $@ if $@ && $@ !~ /(						# check error message (burt447561), if error
			        There\sare\sno\sentries\smatching\syour\squery|	# is something like service is not found
                                This\stable\sis\scurrently\sempty|		# then don't treat it as error and
			        Cannot\sfind\siSCSI\sservice			# for everything else should fail as is
			       )/x;
    return $show;
}

=head2 ::iscsi_create

Create iscsi service if not present. Method checks for available iSCSI LIFs first and tries to convert
one of available data interfaces if no operational iSCSI LIFs found. Method won't create service
without operation iSCSI LIF.

Methods B<GNOTE::GX::Vserver::iscsi_start> calls this method automatically to make sure service
is created before making any attempt to start it.

Returns array of hashes from B<iscsi_show>.

=over

=item Arguments

B<status-admin> => I<up|down> Default is B<up>.

B<alias> => I<string>. Name of iscsi alias, default is name of vserver

Other arguments are same as for B<GNOTE::GX::Vserver::set_iscsi_lif>

=back

=cut

sub iscsi_create {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $status = $args{'status-admin'} ? delete $args{'status-admin'} : 'up';		# parameter for iSCSI LIF and default service state
    my $alias  = $args{alias} ? delete $args{alias} : $self->name;			# Use vserver name for alias.

    eval{ $self->server->system_license_show( feature=>'iSCSI' ) };			# check iSCSI license
    if( $@ =~ /There are no entries matching your query./ ) {				# and add one if required
	# $self->server->system_license_add( "license-code" => 'RYAEFKPJYBFDAA' );
	$self->Cluster->license('iSCSI',1); # if this code fails, then check GX::Cluster->license, make sure key exists for each type
    }

    #Need to make sure iscsi lif is present
    if(! $self->get_iscsi_lif) {
        $self->set_iscsi_lif(%args);
    }

    if(! $self->iscsi_show ) {
        #Check for iSCSI LIF dependency first
        confess('No iSCSI data interfaces found') if(! $self->get_iscsi_lif(%args));
        #Create vserver iscsi now
        my $eval_ret = eval {
                           $self->server->vserver_iscsi_create( 'alias'       =>$alias,
                                                                'status-admin'=>$status,
                                                                'vserver'     =>$self->name
                                                              );
            1;
        };
        if(!$eval_ret) {
            if($@ !~ /invalid argument/i) {
              confess "Could not create vserver iscsi. Error: $@";
            }
            else {
                #Try providing the target-alias
                #adding target-alias instead of alias per burt#461728
                my $eval_ret2 = eval {
                              $self->server->vserver_iscsi_create('target-alias'=>$alias,
                                                                  'status-admin'=>$status,
                                                                  vserver=>$self->name);
                              1;
                           };
                confess "Could not create vserver iscsi. Error: $@" if (!$eval_ret2);
            }
        }
    }
    return $self->iscsi_show();
}

=head2 ::iscsi_start

Attempts to start iscsi service if it's not already running or creates it if service was not ever created.

Methods like B<GNOTE::GX::Vserver::igroup_create> and B<GNOTE::GX::Volume::lun_create>
call this method automatically to make sure iSCSI service is running.

Method returns same output as B<GNOTE::GX::Vserver::iscsi_show>

=over

=item Arguments

All arguments are optional and same as for B<GNOTE::GX::Vserver::iscsi_create>

=back

=cut

sub iscsi_start {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $show = $self->iscsi_create(%args);
    return $show if $$show[0]->{'status-admin'} =~ /(up|online)/ || $$show[0]->{'status'} =~ /(up|online)/;
    $self->server->vserver_iscsi_start( vserver=>$self->name );
    $show = $self->iscsi_show();
    confess( "iscsi is not started" ) unless $$show[0]->{'status-admin'} =~ /(up|online)/ || $$show[0]->{'status'} =~ /(up|online)/;
    return $show;
}

=head2 ::iscsi_stop

Attempts to stop iscsi service if it's running.

Method returns same output as B<GNOTE::GX::Vserver::iscsi_show>

=over

=item Arguments

none

=back

=cut

sub iscsi_stop {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $show = $self->iscsi_show();
    $self->server->vserver_iscsi_stop( vserver=>$self->name ) if $$show[0]->{'status-admin'} =~ /(up|online)/ || $$show[0]->{'status'} =~ /(up|online)/;
    return $self->iscsi_show();
}

=head2 ::igroup_create

Creates an igroup and returns a hash with its information. If you provide igroup name and it exists it will try re-use it.
If you provide an initiator name it will will try to add the initiator name and if it exists it will try re-use it.

=over

=item Arguments

B<none>

I<Optional>

B<ostype> => I <windows|linux|solaris|hpux|aiz|netware|vmware|openvms >

I<Optional>

you can enter more then one initiator at once - comma seperated e.g initiator=>'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2,iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-3'

B<initiator> => I<string>

I<Optional>

name of the vserver.

B<vserver> => I<string>

I<Optional>

name of the igroup.

B<igroup> => I<string>

I<Optional>

name of the protocol.

B<'protocol-type'> => I<iscsi|fcp|mixed>

I<Optional>

name of the port.

B<portset> => I<string>

=item Examples

    my $util = NOTE::SnapMirrorUtil->new();
    my $src = $util->create_volume(tag=>'src',count=>1);
    my $igroup_create = $src->Vserver->igroup_create(initiator=>'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2');

Following is the hash returned by igroup_create:

 $VAR1 = {
    'throttle-borrow' => 'false',
    'vserver' => 'vs0',
    'throttle-resv' => '0',
    'uuid' => '2104d632-312f-11df-92cf-123478563412',
    'portset' => '-',
    'igroup' => 'test',
    'ostype' => 'windows',
    'initiator' => [
		    'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2',
		    'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-3'
		    ],
     'alua' => 'false',
     'protocol-type' => 'iscsi'
     };

=back

=cut

sub igroup_create {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $server = $self->server;

    $self->iscsi_start if $args{'protocol-type'} eq 'iscsi';

    $args{vserver} = $self->name;
    $args{ostype} = 'windows' unless ($args{ostype}); #default ostype is set to windows
    my $output = undef;
    #following block is for already existing igroups
    #if group name is given
    my $igroup_show = $self->igroup_show(igroup=>$args{igroup}, vserver=>$args{vserver}) if($args{igroup}); #adding vserver args after snapmirror script failed
    $output = $igroup_show->[0];  # returning a hash ref to the output instead of an array, because this output
    #is generated from igroup show, since its in a re-use condition we would like to be consistent with our output

    if(defined($output)) {  #if output is defined
	if($output->{igroup} eq $args{igroup}){    #comparing if the provided igroup equals the given igroup
	    NOTE::logcomment("igroup: $output->{igroup} already exists, re-using $args{igroup}\n");
	    if($args{initiator}) {  #Now we have two cases if initiator name is provided
		if(ref($output->{initiator}) eq 'ARRAY') { # if there are more then one initiator then we need to search through an array
		    foreach (@{$output->{initiator}}) {
			if($args{initiator} eq $_) {
			    NOTE::logcomment("Initiator: $args{initiator} is already in igroup: $output->{igroup}, therefore keeping the initiator.\n");
			    return $output;
			}
		    }
		}
		if($args{initiator} eq $output->{initiator}) {   # to check whether provided intiator exists in the igroup
		    NOTE::logcomment("Initiator: $args{initiator} is already in igroup: $output->{igroup}\n");
		} else {
		    $self->igroup_add(igroup=>$args{igroup},initiator=>$args{initiator});
		    NOTE::logcomment("Initiator: $args{initiator} is added to $args{igroup}\n");
		}
	    }
	}
    } else {								    # this block is if igroup doesnt already exists
   	$args{igroup} = $self->newName("igroup_") unless( $args{igroup} );  # generate igroup name if not given

	$args{'protocol-type'} = 'iscsi' unless($args{'protocol-type'} || $args{'protocol'});    # default prototype is set to iscsi. Also see
        eval { $output = ($self->server->lun_igroup_create(%args))->[0]; };			# BURT 461112. Leaving protocol type in place
	if( $@ =~ /invalid argument.*protocol-type/s ) {		    			# for backwards compatibility
	    $args{'protocol'} = delete $args{'protocol-type'};					# See also BURT 469152
	    $output = ($self->server->lun_igroup_create(%args))->[0];
	} elsif( $@ ) {
	    confess $@;
	}
	NOTE::logcomment("Igroup $args{igroup} was created successfully!\n");
    }
    return $output;
}

=head2 ::igroup_show

Runs igroup show command and returns the populated hash with its attributes.

=over

=item Arguments

B<none>

I<Optional>

B<ostype> => I <windows|linux|solaris|hpux|aiz|netware|vmware|openvms >

I<Optional>

you can enter more then one initiator at once - comma seperated e.g initiator=>'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2,iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-3'

B<initiator> => I<string>

I<Optional>

name of the vserver.

B<vserver> => I<string>

I<Optional>

name of the igroup.

B<igroup> => I<string>

I<Optional>

name of the protocol.

B<'protocol-type'> => I<iscsi|fcp|mixed>

I<Optional>

name of the port.

B<portset> => I<string>

I<Optional>

B<uuid> => I<string>

I<Optional>

B<'throttle-resv'> => I<string>

I<Optional>

B<'throttle-borrow'> => I<true|false>

I<Optional>

B<alua> => I<string>

I<Optional>

B<'igroup-ostype'> => I<string>

=item Examples

    my $util = NOTE::SnapMirrorUtil->new();
    my $src = $util->create_volume(tag=>'src',count=>1);
    my $igroup_show = $src->Vserver->igroup_show(igroup=>'test');

Following is the hash returned by igroup_show:

  $VAR1 = [
   {
     'throttle-borrow' => 'false',
     'vserver' => 'vs0',
     'throttle-resv' => '0',
     'uuid' => '420b136a-312e-11df-92cf-123478563412',
     'portset' => '-',
     'igroup' => 'group1',
     'ostype' => 'windows',
     'initiator' => '-',
     'alua' => 'false',
     'protocol-type' => 'iscsi'
   },
   {
     'throttle-borrow' => 'false',
     'vserver' => 'vs0',
     'throttle-resv' => '0',
     'uuid' => 'e1450c6b-312f-11df-92cf-123478563412',
     'portset' => '-',
     'igroup' => 'igroup15807_Pict_carefree',
     'ostype' => 'windows',
     'initiator' => [
                      'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2',
                      'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-3'
                    ],
     'alua' => 'false',
     'protocol-type' => 'iscsi'
   },

=back

=cut

sub igroup_show {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $output = undef;
    eval { $output = $self->server->lun_igroup_show(%args) };
    confess $@ if( $@ && $@ !~ /(This table is currently empty.|There are no entries matching your query.)/ );
    return $output;
}

=head2 ::igroup_add

Runs igroup add command, it will add the provided initiator to the provided igroup

=over

=item Arguments


B<initiator> => I<string>B<none>

B<igroup> => I<string>

I<Optional>

B<force> => I<true|false>

=item Examples

    my $util = NOTE::SnapMirrorUtil->new();
    my $src = $util->create_volume(tag=>'src',count=>1);
    $src->Vserver->igroup_remove(igroup=>'test',initiator=>'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2');

=back

=cut

#requred args : igroup and initiator

sub igroup_add {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $server = $self->server;
    $args{vserver} = $self->name;
    confess "initiator is not provided\n" unless($args{initiator});
    confess "igroup is not provided\n" unless($args{igroup});
    my $output = undef;
    eval { $output = $server->lun_igroup_add(%args) };
    confess $@ if $@;
    NOTE::logcomment("initiator $args{initiator} is successfully added to $args{igroup}\n");
    return $output;
}

=head2 ::igroup_remove

Runs igroup remove command, it will remove the provided initiator from the provided igroup.

=over

=item Arguments

B<initiator> => I<string>B<none>

B<igroup> => I<string>

I<Optional>

B<force> => I<true|false>

=item Examples

    my $util = NOTE::SnapMirrorUtil->new();
    my $src = $util->create_volume(tag=>'src',count=>1);
    $src->Vserver->igroup_add(igroup=>'test',initiator=>'iqn.1991-05.com.microsoft:dpg-vm-w2k8-iscsi-2');

=back

=cut

#requred args : igroup and initiator
sub igroup_remove {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "initiator is not provided\n" unless($args{initiator});
    confess "igroup is not provided\n" unless($args{igroup});
    my $output = undef;
    $args{vserver} = $self->name;
    eval { $output = $self->server->lun_igroup_remove(%args) };
    confess $@ if $@;
    NOTE::logcomment("initiator $args{initiator} is successfully removed from $args{igroup}\n");
    return $output;
}

=head2 ::igroup_destroy

Runs igroup destroy command, it will destroy the provided igroup. Method igroup_destroy expects
luns to be unmapped first.

=over

=item Arguments

B<igroup> => I<string>

I<Optional>

B<force> => I<true|false>

=item Examples

    my $util = NOTE::SnapMirrorUtil->new();
    my $src = $util->create_volume(tag=>'src',count=>1);
    $src->Vserver->igroup_destroy(igroup=>'test');

=back

=cut

sub igroup_destroy {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $output = undef;
    $args{vserver} = $self->name;
    confess "igroup name is not provided\n" unless($args{igroup});
    my $result = eval {$self->server->lun_igroup_delete(%args);1};  #changing from lun_igroup_destroy to lun_igroup_delete (UI change)burt#451020
    confess $@ if (!$result);
    NOTE::logcomment("igroup: $args{igroup} was successfully destroyed\n");
    return $result;
}

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    eval {
        if(defined($self->igroup_show)) { # igroup destroy should be in Vserver DESTROY it seems like a hack at the moment
            if(! NOTE->ALL_SPARED) {
                foreach my $igrp_temp ( @{$self->igroup_show()} ) {
                    $self->igroup_destroy(igroup=>$igrp_temp->{'igroup'});
                }
            }
        }
    };

    NOTE::logcomment('DESTROY: _created_data_lif_names_aref: ' . Dumper($self->{_created_data_lif_names_aref}));
    NOTE::logcomment('DESTROY: _deleted_data_lifs_aref: ' . Dumper($self->{_deleted_data_lifs_aref}));
    #Delete all the created iscsi lifs
    if(! NOTE->ALL_SPARED) {
        if( defined $self->{_created_data_lif_names_aref} ) {
            my $iscsi_lif;
            while ($iscsi_lif = shift @{$self->{_created_data_lif_names_aref}} ) {
                NOTE::logcomment('DESTROY: deleting the created iscsi lif: ' . $iscsi_lif);
                my $eval_ret = eval {
                                   $self->server->network_interface_delete(
                                                       server => $self->name,
                                                       lif => $iscsi_lif,);
                                   1;
                               };
                if(!$eval_ret) {
                    #do not confess, as this is DESTROY function
                    NOTE::logcomment("Error in deleting the created iscsli lif $iscsi_lif. Error: $@");
                }
            }
        }
    }

    #Create back all the deleted data lifs
    if(! NOTE->ALL_SPARED) {
        if( defined $self->{_deleted_data_lifs_aref} ) {
            my $lif_temp;
            while ( $lif_temp = shift @{$self->{_deleted_data_lifs_aref}} ) {
                NOTE::logcomment('DESTROY: creating back the deleted data lif: ' . $lif_temp->{'lif'});
                my %args;
                foreach my $param ('lif','role', 'address', 'netmask', 'home-port',
                                   'home-node', 'data-protocol', 'server',) {
                    $args{$param} = $lif_temp->{$param};
                }
                my $eval_ret = eval {
                                  $self->server->network_interface_create(%args);
                                  1;
                               };
                if(!$eval_ret) {
                    #do not confess, as this is DESTROY function
                    NOTE::logcomment('Error in creating back the deleted data lif: ' . $lif_temp->{'lif'} . " Error: $@");
                }
            }
        }
    }
}

1;
