#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Printf.pm#1 $
#
## @summary  NOTE class for support for print(f) within NOTE scripts.
## @status   Public
## @pod      here
## @author   tmichael@netapp.com

=head1 NAME

NOTE::Printf

=head1 DESCRIPTION

This class makes it possible for NOTE scripts to use print or printf
instead of logcomment or NOTE::logcomment. This is desirable for 
several reasons:

1) Since print and printf are standard Perl functions, they are natural
choices for a script writer.

2) printf has built-in support for formatted prints: 

     printf "Average throughput: %.3f MB/second\n", $MB/$t;

is more readable than these alternatives:

     logcomment(sprintf("Average throughput: %.3f MB/second", $MB/$t));
     NOTE::logcomment("Average throughput: %.3f MB/second", $MB/$t);

3) print and printf can be used to build lines of output:

     printf "Timed options: ";
     for my $k ( keys %{$timed} ) {
	 printf "%s=%s ", $k, $timed->{$k};
     }
     printf "\n";

would put everything on one line, where logcomment generates a new
line on each call.


To support the NATE logfile format, multiple lines can be output
as a single 'comment' in a NATE log by putting a '\0' after newlines
to tell NOTE to continue buffering:

     printf "Timed options:\n\0";
     for my $k ( keys %{$timed} ) {
	 printf "    %s = %s\n\0", $k, $timed->{$k};
     }
     printf "\n";

generates this in the log:

    20060920 155508 [01_timed] ........................................
    20060920 155508 [01_timed] . Comment:
    20060920 155508 [01_timed] . Timed options:
    20060920 155508 [01_timed] .     max_skew = 30m
    20060920 155508 [01_timed] .     proto = ntp
    20060920 155508 [01_timed] .     window = 15s
    20060920 155508 [01_timed] .     min_skew =
    20060920 155508 [01_timed] .     sched = 1h
    20060920 155508 [01_timed] .     log = on
    20060920 155508 [01_timed] .     servers = time1.lab.netapp.com time2.lab.netapp.com
    20060920 155508 [01_timed] .
    20060920 155508 [01_timed] ........................................

Only newlines that are the *last* character of a string cause the print
buffer to be flushed:

    printf "%s:\n    is %d KB\n", $vol->name, $vol->size;

generates:

    20060920 161133 [01_vol] ........................................
    20060920 161133 [01_vol] . Comment:
    20060920 161133 [01_vol] . vl20272_ca9b_b359e5d3:
    20060920 161133 [01_vol] .     is 20480 KB
    20060920 161133 [01_vol] ........................................

It is possible to use the NOTE_LOG environment variable to filter
the log output. The usage is:

    NOTE_LOG = [<loglevel>][,<logfile>]

where

    <loglevel> is 0 or 1.
        0 means *ONLY* prints or printfs are to be printed on the screen.
	  All other output will be thrown away. Since print(f)s are restricted
	  (by convention) to the scripts themselves (please don't use them in
	  libraries!), this is a way to filter out everything except output that
	  the script itself generates. NOTE: tracing is not available when
	  this log level is selected.
        1 means that prints, printfs, or NOTE::logcomments are to be printed
          on the screen. This would include output from scripts as well as
	  NOTE libraries.
    <,logfile>
        If this is specifed, it will also receive the entire log, without filtering.
        NOTE: this file, if it already exists, is *overwritten*.

For example, this complete program:

    use GNOTE;
    printf "\nVolumes on %s:\n\0", NOTE::option( q/$FILER/ );
    for my $vol ( NOTE::Filer->connect(q/$FILER/)->Volumes ) {
	printf( "%9d KB [%s] %s\n\0", $vol->size, $vol->type, $vol->name );
	printf( "+---------------->  \"%s\"\n\0", $vol->dataset ) if ( $vol->dataset );
    }

when run like this:

    #> ntest FILER=fsr-u34 vvls.thpl NOTE_LOG=0,vvls.log

will generate this output on stdout:

    NATE 2.1.1 (change #544584)
    20060920 165331 [00_ntest] !! Subtest 01_vvls started (pid 4731)
    20060920 165331 [00_ntest]    Script /u/tmichael/e3/vvls.thpl

    Volumes on fsr-u34:
     50119928 KB [trad] vol0
	20480 KB [flex] vl12920_6879_ad1cd663
       512000 KB [flex] _vl01894_cf63_0957f584_zRn2M9W__MnmNbUggXW__4yupDqnPQ0gBKFIQ
    +---------------->  "shift->fill( percent => 40 )"
       512000 KB [flex] _vl03719_d3bb_700dfbe5_tDn2tOdtDvucU6yLXGFo5oTbmOLQnanPKd5lLfCUqW__g_PGed2hpXGEIQ
    +---------------->  "main::makeQtree( shift, 'myQtree' )->populate"

    20060920 165341 [01_vvls] END OF TEST / RESULTS SUMMARY
    20060920 165341 [01_vvls]     No results reported
    20060920 165341 [01_vvls]
    20060920 165341 [00_ntest] END OF TEST / RESULTS SUMMARY
    20060920 165341 [00_ntest]     No results reported
    20060920 165341 [00_ntest]
    ntest: Results Summary:
    ntest: Type    Executions  Results(incl. subtests)
    ntest: Normal      1       1 PASS

and vvls.log will contain the entire, unfiltered NATE log.

=cut

use GNOTE;
package NOTE::Printf;
use strict;

# NOTE_LOG = [0|1][,<logfile>]
# NOTE_LOG = 0 means only print(f)s are printed on the screen.
# NOTE_LOG = 1 means print(f)s or NOTE::logcomments are printed on
#              the screen.
#     To accomplish this,
#         If STDOUT and STDERR are the same,
#             (i.e. STDOUT has not been re-directed), the STDOUT
#             file handle is reopened on <logfile> (default is /dev/null)
#             and NOTE::STDOUT is opened on fd 1.
#         If STDOUT and STDERR are different (STDOUT has been redirected),
#             the NOTE::STDOUT file handle is opened on fd 2.

my ($_odev,$_oino) = stat(STDOUT);
my ($_edev,$_eino) = stat(STDERR);

sub new {
    my $class = shift;

    my ($self,$handle,$raw);

    $handle = IO::File->new;
    if ( defined $NOTE::LOGFILE ) {
	$raw = IO::File->new( ">&1" ) or die "Couldn't open NOTE::STDOUT on fd 1: $!";
	$raw->autoflush;
	open( LOGFILE, '>', $NOTE::LOGFILE ) or die "Can't open $NOTE::LOGFILE: $!";
	STDOUT->fdopen( fileno(LOGFILE), 'w' ) or die "Reopen of stdout failed: $!";
	($_odev,$_oino) = stat(STDOUT);
	($_edev,$_eino) = stat(STDERR);
    } else {
	# no logfile specified.
	$raw = IO::File->new( ">&1" ) or die "Couldn't open NOTE::STDOUT on fd 2: $!";
    }

    $self = tie $$handle, 'NOTE::Printf';
    $self->{_STDOUT} = $handle;
    $self->{_RAW}    = $raw;
    return $self;
}

sub start {
    my $self = shift;
    $self->{_OLDFH} = select( $self->{_STDOUT} );
    return $self;
}

sub stop {
    my $self = shift;
    select( $self->{_OLDFH} );
    return $self;
}

sub flush {
    my $self = shift;
    if ( defined $self->{_linebuf} ) {
	$self->PRINT( "\n" );
    }
    return $self;
}

sub fh  { return shift->{_STDOUT} }
sub raw { return shift->{_RAW} }

sub TIEHANDLE {
    my $class = shift;
    my $self = {};
    $self->{_linebuf} = "";
    bless $self, $class;
    return $self;
}

my $count = 0;

sub PRINT {
    my $self = shift;
    my $string = join( '', @_ );
    if ( $NOTE::LOGLEV eq '0' ) {
	my $s = $string;
	$s =~ s/\0$//;
	$self->{_RAW}->print( $s );
	return unless ( $NOTE::LOGFILE );
    }
    if ( $string =~ /\n$/ ) {
	if ( defined $self->{_linebuf} ) {
	    $string = $self->{_linebuf} . $string;
	    $self->{_linebuf} = "";
	}
	$string =~ s/\n$//; # logcomment adds a \n, so remove trailing one here.
	if ( $string =~ /\n/ ) { # Prepend a new line if the string contains newlines
	    $string = "\n".$string;
	}
	NOTE::logcomment( $string );
	return;
    }
    $string =~ s/\0$//;
    $self->{_linebuf} .= $string;
}

sub PRINTF {
    my $self = shift;
    my $fmt = shift;
    $self->PRINT( sprintf( $fmt, @_ ) );
}

sub WRITE {
    my $self = shift;
    my ($buf,$len,$offset) = @_;
    $self->PRINT( $buf );
}

sub DESTROY {
    my $self = shift;
    if ( defined $self->{_linebuf} ) {
	NOTE::logcomment( $self->{_linebuf} );
	$self->{_linebuf} = "";
    }
    $self->stop;
    untie $self;
}
1;
