#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Qtree.pm#1 $
#
## @summary  NOTE Qtree class
##
## @status   Public
##
## @author   dl-dpgqa-automation@netapp.com
##
## @pod      here

=head1 NAME

NOTE::Qtree

=head1 DESCRIPTION

This is the GNOTE base class for Qtrees.

=head1 METHODS

=cut

package NOTE::Qtree;

use GNOTE;
our @ISA = qw(NOTE);
use Carp;
use Tharn qw();

=head2 ::new

Creates a GNOTE::Qtree object and returns it.

=over

=item Arguments

B<Volume> => I<object> 
B<name> => I<string>

=item Examples

=back

     $qt = NOTE::Qtree->new(
         tag    => 'destination',
         name    => 'qtree1',
     %args
     );
     
=cut

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;
    if ( ref($class) =~ /NOTE::.*Qtree/ ) {
        $self = $class;
    }
    else {
        return _do_method( "new", @_ );
    }

    exists $args{name} and $args{name} = NOTE::option( $args{name} );

    $self->{_tag}      = delete $args{tag};
    $self->{_richdata} = delete $args{richdata};
    $self->{_teardown} = delete $args{teardown};

    if ( defined( $args{name} ) ) {
        $self->{_name} = $args{name};

        # Allow named qtrees that behave like anonymous qtrees:
        $self->{_anonymous} = ( exists $args{anonymous} ? $args{anonymous} : $self->_looksAnonymous );
        delete $args{name};
    }
    else {

        # Create a unique name
        $self->{_name}      = NOTE::Qtree->newName;
        $self->{_anonymous} = 1;
    }
    delete $args{anonymous};

    my $desc = $self->{_name};
    $desc .= " (" . $self->{_tag} . ")" if ( $self->{_tag} );

    # If the qtree isn't anonymous, make certain the parent
    # container is spared.
    if ( !$self->{_anonymous} ) {
        $self->{_Volume}->SPARE;
    }

    return $self;
}

sub _do_method {
    my ( $method, $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( NOTE::is_GX(%args) ) {
        return NOTE::GX::Qtree->$method(%args);
    }
    else {
        return NOTE::7G::Qtree->$method(%args);
    }
}

=head2 ::attach

Creates an object representing an existing Qtree.

=over

=item Arguments

B<name> => I<string>

I<Required.>
The name of an existing Qtree.

B<Volume> => I<object> | I<spec>

I<Required.>
The volume in which to find the named Qtree.
Specified as a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/> object
or a volume spec.

B<Filer> => I<host spec>

I<Dependent.>
Required only when the B<Volume> argument is a volume spec.
Specifies the filer where the volume exists.

B<mntpoint> => I<string>

I<Optional.>
The full path name on which to mount the Qtree.

B<exportfs> => I<export options>

I<Optional.>
Export options as described in L<C<::exportfs>|/::exportfs>.

B<dataset> => I<dataset specifier>

I<Optional.>
A description of some data to put in the Qtree.
The dataset will be fully created
before the call to C<::create> returns.
See L<C<NOTE::Volume::fill>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::fill>
for a description of dataset specifiers.

B<force> => C<1>

I<Optional.>
A flag to force the creation of a Qtree object
even if the named Qtree does not exist.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip for
L<C<::populate>|/::populate> and L<C<::modify>|/::modify>.
The default is 0.

=item Examples

=back

     $qt = NOTE::Qtree->attach(
         Volume => $vol,
         name   => 'qtree42'
     );

=cut

sub attach {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;

    if ( ref($class) =~ /NOTE::.*Qtree/ ) {
        $self = $class;
    }
    else {
        return _do_method( "attach", @_ );
    }

    my $force = $args{force};

    $args{name} = NOTE::option( $args{name} ) if exists $args{name};

    if ( ref( $args{Volume} ) =~ /NOTE::.*Volume/ ) {
        $self->{_Volume} = $args{Volume};
    }
    elsif ( ref $args{Volume} eq "HASH" ) {
        if ( NOTE::is_GX(%args) ) {
            $self->{_Volume} = NOTE::GX::Volume->attach( %{ $args{Volume} } );
        }
        else {
            $self->{_Volume} = NOTE::7G::Volume->attach( Filer => $args{Filer}, %{ $args{Volume} }, force => $force );
        }
    }
    else {
        croak "Attaching a qtree requires a volume name or object";
    }

    unless ( defined( $args{name} ) ) {
        croak "Attaching a qtree requires a qtree name";
    }
    $self->{_name} = $args{name};

    $self->{_tag} = $args{tag};
    delete $args{tag};

    $self->{_richdata} = $args{richdata};
    delete $args{richdata};

    if ( exists $args{mntpoint} ) {
        $self->{_mntpoint} = $args{mntpoint};
    }

    return $self;
}

sub _exists {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( !ref $self ) {
        if ( NOTE::is_GX(%args) ) {
            return NOTE::GX::Qtree->_exists(@_);
        }
        else {
            return NOTE::7G::Qtree->exists(@_);
        }
    }
}

=head2 ::create

Creates a Qtree and returns
an object representing the created Qtree.

=over

=item Arguments

B<Volume> => I<object> | I<spec>

I<Optional.>
The volume in which to create the Qtree.
May be a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
object or a volume spec.
If absent, a B<Filer> or B<Aggregate> argument must be given
and an anonymous volume will be created or attached
before the call returns.

B<name> => I<string>

I<Optional.>
The name of the Qtree.
If absent, an anonymous Qtree name will be generated.

B<Filer> => I<host spec>

I<Optional.>
The filer on which to create the Qtree.

B<Aggregate> => I<object> | I<spec>

I<Optional.>
The aggregate in which to create a volume
in which to create the Qtree.
May be a
L<C<NOTE::Aggregate>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Aggregate-pm/>
object or an aggregate spec.
If absent, a B<Volume> or B<Filer> argument must be given.

B<tag> => I<string>

I<Optional.>

See L<C<::tag>|/::tag> for details.

One of B<Volume>, B<Aggregate>, or B<Filer> must be given.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip for
L<C<::populate>|/::populate> and L<C<::modify>|/::modify>.
The default is 0.

=item Examples

=back

     $qt = NOTE::Qtree->create(
         Volume => $vol,
         tag    => 'destination'
     );

     $qt = NOTE::Qtree->create(
         Filer     => { host => $FILER_SRC },
         Volume    => { size => "20g" },
         Aggregate => { disks => 5, raidtype => 'raid4' },
         tag       => 'really big Qtree'
     );

=cut

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self;
    if ( ref($class) =~ /NOTE::.*Qtree/ ) {
        $self = $class;
    }
    else {
        return _do_method( "create", @_ );
    }

    my $is_gx = NOTE::is_GX(%args);

    unless ($is_gx) {
        my $exportfs_opts = $args{exportfs} || "rw,anon=0";
        delete $args{exportfs};
    }

    my $need_to_create = 1;
    if ( $args{name} ) {

        # Dummy check for existing Qtree
        my $found_qtree = eval { NOTE::Qtree->attach(%args) };
        unless ($@) {
            NOTE::logcomment( "Re-using existing Qtree " . $self->description );
            $need_to_create = 0;
            $found_qtree->is_destroyed;    # No need to use this object as the current object is sufficient
        }
    }

    $self->_qtree_create(%args)
      if $need_to_create;

    $self->{_dataset} = $args{dataset};
    delete $args{dataset};
    $self->{_recycle} = $args{recycle};
    delete $args{recycle};

    unless ($is_gx) {
        $self->exportfs($exportfs_opts) if ($exportfs_opts);
    }
    if ( $self->{_dataset} ) {
        NOTE::logcomment( "Populating " . $self->description );
        my $func = eval 'sub {' . $self->{_dataset} . '}';
        croak $@ if $@;
        eval { $self->$func };
        croak $@ if $@;
    }

    return $self;
}

=head2 ::name

Returns the name of the Qtree.

=over

=item Arguments

None.

=item Example

=back

     $name = $qt->name;

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name};
}

=head2 ::Volume

Returns an object representing the volume containing the Qtree.

=over

=item Arguments

None.

=item Example

=back

     $vol = $qt->Volume;

=cut

sub Volume {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Volume};
}

=head2 ::newName

Returns a name for qtree.

=over

=item Arguments

None.

=item Example

=back

     $name = $qt->newName;

=cut

sub newName {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->SUPER::newName("qt");
}

=head2 ::Properties

Duplicates the Properties of a qtree and
returns a hash with its attributes. 

=over

=item Examples

=back

     my %prop = qtree->Properties();
     
=cut

sub Properties {
    my $self       = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %properties = (
        anonymous => $self->{_anonymous},
        richdata  => $self->{_richdata},
        recycle   => $self->{_recycle},
        exportfs  => $self->exportfs,
        Volume    => { $self->Volume->Properties },
    );
    return %properties;
}

=head2 ::Aggregate

Returns an object representing the aggregate
containing the volume containing the Qtree.

Croaks if the containing volume does not have an Aggregate object, i.e. a trad volume.

=over

=item Arguments

None.

=item Example

=back

     $aggr = $qt->Aggregate;

=cut

sub Aggregate {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $aggr;

    eval { $aggr = $self->{_Volume}->Aggregate; };
    confess $self->description . " Parent error: $@"
      if $@;
    return $aggr;
}

=head2 ::tag

A tag is a text description for a filer entity.
It is only stored in the object referring to the entity.
The purpose of the tag is to improve readability of the log.
If an object is tagged,
NOTE will include the tag in comments it logs about the object.

Returns the tag for the object.

=over

=item Arguments

I<string>

I<Optional.>
A tag to store in the object.

=item Examples

=back

     $qt->tag( 'source' );
     $which = $qt->tag;

=cut

sub tag {
    my ( $self, $newtag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $newtag ) {
        $self->{_tag} = $newtag;
    }
    return $self->{_tag};
}

=head2 ::description

Returns a description of the Qtree consisting of
its Volume's hostpath, its name, and its tag, if any.

=over

=item Arguments

None.

=item Example

=back

     $desc = $qt->description;

=cut

sub description {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Volume->hostpath() . "/" . $self->name . ( $self->{_tag} ? "( " . $self->{_tag} . " )" : "" );
}

=head2 ::SPARE

Don't destroy / recycling the Qtree and it's containing Volume.

=over

=item Arguments

None.

=item Example

=back

     $qt->SPARE;

=cut

sub SPARE {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_spare_me} = 1;
    $self->Volume->SPARE;
}

=head2 ::container


Refer to L<C<Volume::container>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3acontainer>.

=cut

sub container {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Volume}
}

=head2 ::mountpoint

Refer to L<C<Volume::mountpoint>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3amountpoint>.

=cut

sub mountpoint {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::mountpoint( $self, %args );
}

=head2 ::mounted

Refer to L<C<Volume::mounted>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3amounted>.

=over

=item Arguments

None.

=item Example

=back

     $mntpt = $qt->mounted;

=cut

sub mounted {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::mounted( $self, %args );
}

=head2 ::populate

Populate the qtree with data, using localhost, by invoke the parent
L<C<Volume::populate>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3apopulate>
method.

=over

=item Examples

=back

        $qtree->populate;

=cut

sub populate {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::populate( $self, %args );
}

=head2 ::modify

Modifies the qtree's data, using localhost, by invoke the parent
L<C<Volume::modify>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3amodify>
method.

=over

=item Examples

=back

        $qtree->modify;

=cut

sub modify {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::modify( $self, %args );
}

=head2 ::is_anonymous

Returns true if this volume is anonymous.  This can also be
assigned to, in order to make a volume anonymous.

=over

=item Argument

None.

=item Example

=back

     if ( $qtree->is_anonymous ) {
         ...
     }

     ...

     $qtree->is_anonymous = 1;

=cut

sub is_anonymous : lvalue {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Must do this trickery due to a bug in perl regarding
    # tied hashes and lvalues.
    my $r = \$self->{_anonymous};
    $$r;
}

=head2 ::is_destroyed

Returns true if this qtree is destroyed.  This can also be
assigned to, in order to make a qtree 'un-destroyed',
but this feature is intended for internal use only.

=over

=item Argument

None.

=item Example

=back

     if ( $qtree->is_destroyed ) {
         ...
     }

     ...

     $qtree->is_destroyed = 0;

=cut

sub is_destroyed : lvalue {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Must do this trickery due to a bug in perl regarding
    # tied hashes and lvalues.
    my $r = \$self->{_destroyed};
    $$r;
}

=head2 ::rename

Changes the name of the Qtree.

=over

=item Arguments

I<string>

I<Optional.>
The new name for the Qtree.
If absent a new anonymous name is generated
and the Qtree becomes anonymous.
If present the Qtree will no longer be anonymous,
causing its volume (and aggregate) to be spared
(not destroyed) when the Qtree object is destroyed.

=item Examples

=back

     $qt->rename;
     $qt->rename( 'newQtreeName' );

=cut

sub rename {
    my ( $self, $newname, %opts ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $was_mounted = $self->mounted;
    my $oldpath     = $self->mount_path;
    my $oldname     = $self->{_name};

    if ( !$opts{anonymous} && $newname ) {
        $self->{_name} = $newname;

        # Not anonymous anymore, so make sure containers aren't destroyed
        $self->Volume->SPARE;
    }
    else {
        $self->{_name} = $newname ? $newname : $self->newName;
        $self->{_anonymous} = 1;
    }

    if ($was_mounted) {
        $self->umount;
    }
    if ( NOTE::is_GX(%args) ) {
        return NOTE::GX::Qtree->_rename($oldname);
    }
    else {
        return NOTE::7G::Qtree->_rename($oldpath);
    }
    if ($was_mounted) {
        $self->mount;
    }
}

=head2 ::mount

NFS mounts the volume containing the Qtree.

=over

=item Arguments

I<string>

I<Optional.>
A list of mount options as defined in
L<C<NOTE::Volume::mount>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3amount>

=item Example

=back

     $qt->mount( chmod => '777' );

=cut

sub mount {
    my ( $self, %opts ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::mount( $self, %opts );
}

=head2 ::umount

Unmounts the volume containing the Qtree.

=over

=item Arguments

I<None>

=item Example

=back

     $qt->umount;

=cut

sub umount {
    my ( $self, %opts ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::umount( $self, %opts );
}

=head2 ::cd

Sets the current working directory for the calling process.

=over

=item Arguments

I<string>

I<Required.>
A path name.

=item Example

=back

     $qt->cd( '../testdir2' );

=cut

sub cd {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::cd( $self, %args );
}

=head2 ::pwd

Returns the current working directory for the calling process.

=over

=item Arguments

None.

=item Example

=back

     $curdir = $qt->pwd;

=cut

sub pwd {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::pwd($self);
}

=head2 ::nfs

Returns a handle that can be used to perform file system operations
via NFS on the Qtree.

=over

=item Arguments

none

=item Example

=back

     my $qt_root = $qt->nfs;
     $qt_root->create( '/somedir' );

=cut

sub nfs {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::nfs( $self, %args );
}

=head2 B<Waiting>

=over

=head2 ::wait

Refer to L<C<Volume::wait>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#Waiting>.

=over

=item Arguments

None.

=item Example

=back

     $qt->wait;

=cut

sub wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::wait( $self, %args );
}

=head2 B<fill>

=head2 ::fill

Refer to L<C<Volume::fill>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3afill>.

=over

=item Arguments

None.

=item Example

=back

     $qt->fill;

=cut

sub fill {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::fill( $self, %args );
}

=head2 B<refill>

=head2 ::refill

Refer to L<C<Volume::refill>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3arefill>.

=over

=item Arguments

None.

=item Example

=back

     $qt->refill;

=cut

sub refill {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::refill( $self, %args );
}

=head2 B<checkfill>

=head2 ::checkfill

Refer to L<C<Volume::checkfill>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3acheckfill>.

=over

=item Arguments

None.

=item Example

=back

     $qt->checkfill;

=cut

sub checkfill {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::checkfill( $self, %args );
}

=head2 B<unfill>

=head2 ::unfill

Refer to L<C<Volume::unfill>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3aunfill>.

=over

=item Arguments

None.

=item Example

=back

     $qt->unfill;

=cut

sub unfill {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::unfill( $self, %args );
}

=head2 B<fillfile>

=head2 ::fillfile

Refer to L<C<Volume::fillfile>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3afillfile>.

=over

=item Arguments

None.

=item Example

=back

     $qt->fillfile;

=cut

sub fillfile {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::fillfile( $self, %args );
}

=head2 B<sio_proc>

=head2 ::sio_proc

Refer to L<C<Volume::sio_proc>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3asio_proc>.

=over

=item Arguments

None.

=item Example

=back

     $qt->sio_proc;

=cut

sub sio_proc {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::sio_proc($self);
}

=head2 B<sio>

=head2 ::sio

Refer to L<C<Volume::sio>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-Volume-pm.html#%3a%3asio>.

=over

=item Arguments

None.

=item Example

=back

     $qt->sio;

=cut

sub sio {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return NOTE::Volume::sio( $self, %args );
}

=head2 ::total

Returns the total disk space extant in the volume.

=over

=item Arguments

None.

=item Example

=back

     $total = $qt->total;

=head2 ::used

Returns the total disk space currently used in the volume.

=over

=item Arguments

None.

=item Example

=back

     $used = $qt->used;

=head2 ::avail

Returns the total disk space currently available in the volume.

=over

=item Arguments

None.

=item Example

=back

     $total = $qt->avail;

=head2 ::capacity

Returns the percentage of the total disk space
in the volume that is currently being used.

=over

=item Arguments

None.

=item Example

=back

     $capacity = $qt->capacity;

=cut

sub total    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[0]
}
sub used     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[1]
}
sub avail    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[2]
}
sub capacity {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[3]
}

=head2 ::Filer

Returns an object representing the filer containing the volume.

=over

=item Arguments

None.

=item Example

=back

     $filer = $vol->Filer;

=cut

sub Filer {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->Volume->Filer;
}

=head2 ::DESTROY

Destroys the current qtree object

=over

=item Arguments

None.

=item Example

=back

     $self->DESTROY;
=cut

sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( !$self->substant );
    return if ( NOTE->ALL_SPARED );

    if ( $self->{_spare_me} ) {
        NOTE::logcomment( "Qtree " . $self->{_name} . " has been spared" );
        return;
    }
    return NOTE::Volume::_umount($self);
}

###############################################################################
#
#   Sub Name:       _get_volume_name
#
#   Description:    get the volume name for this qtree. this method is needed
#                   to support quotas in both volumes and qtrees. a method of
#                   the same name exists in Volume.pm. this allows Quota.pm
#                   to get its volume name without making any special checks.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    volume name for this qtree (string)
#                   Failure:    not an option
#
###############################################################################

sub _get_volume_name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->Volume->name();
}

###############################################################################
#
#   Sub Name:       _get_qtree_name
#
#   Description:    get the qtree name for this qtree. this method is needed
#                   to support quotas in both volumes and qtrees. a method of
#                   the same name exists in Volume.pm. this allows Quota.pm
#                   to get its qtree name without making any special checks.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    qtree name for this qtree (string)
#                   Failure:    not an option
#
###############################################################################

sub _get_qtree_name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->name();
}

1;
