#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/Volume.pm#1 $
#
## @summary  NOTE Volume class
## @status   Public
## @pod      here
## @author   tmichael@netapp.com

=head1 NAME

NOTE::Volume

=head1 DESCRIPTION

This is the NOTE class for working with volumes.

=head1 METHODS

=cut

use strict;

package NOTE::Volume;
use GNOTE;
our @ISA = qw(NOTE);
use Carp qw(confess);
use Data::Dumper;
use File::Basename;
use Tharn qw();
use Params::Validate qw( validate validation_options SCALAR BOOLEAN HASHREF );

sub _spareVols {
    my ($filer) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @spareVols = $filer->Volumes( spare_only => 1 );
    return sort { $a->size <=> $b->size } @spareVols;
}

# This internal function is used to log the reason that
# a resource was rejected as a recyclable candidate
sub _reject {
    my %args = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $msg  = "Rejecting " . $args{name} . ": " . $args{reason};

    if ( $args{wanted} ) {
        $msg .= "\n    ";
        $msg .= $args{wanted};
    }
    if ( $args{got} ) {
        $msg .= "\n    ";
        $msg .= $args{got};
    }
    NOTE::logcomment($msg);
}

sub _recycle {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $type = $self->{_type};
    my ( $vol, $aggr, $rtn );
    my $aggrQualities;
    my @volList = ();

    my $args = $self->{_args};

    my ( %opts, $aggrOpts );

    # build the options needed to define the volume being requested
    if ( exists $self->{_Vserver} ) {
        $opts{vserver} = $self->{_Vserver}->name;
    }

    if ( $args{local_storage}
         && ($self->{_type} =~ /flex/)
         && ($args{Node} =~ /NOTE::.*Node/) ) {
        $opts{node} = $args{Node}->name;
    }

    $opts{'volume-style'} = $self->{_type};

    if ( $opts{'volume-style'} =~ /striped/ ) {
        if ( exists $self->{_args}->{'stripe-format'} ) {
            $opts{'stripe-format'} = $self->{_args}->{'stripe-format'};
        }
        if ( exists $self->{_args}->{'stripe-count'} ) {
            $opts{'stripe-count'} = $self->{_args}->{'stripe-count'};
        }
        if ( exists $self->{_args}->{'stripe-width'} ) {
            $opts{'stripe-width'} = $self->{_args}->{'stripe-width'};
        }
    }

    NOTE::logcomment('Looking for recyclable volume');

    # get the list of possible volumes based on the restrictions we have,
    # order is important.
    #
    # if the user provided an Aggregate argument use it as guidence.
    # else if there is an aggregate associated with this object use it.
    # else get the list of volumes on the filer/cluster.
    #
    # the Aggregate argument is given preference over an aggregate
    # associated with this object because it allows the search to include
    # all the recyclable volumes in all the aggregates in the recycle pool.
    # otherwise only the recyclable volumes in the aggregate associated
    # with this object could be searched. if a recycled volume is used its
    # aggregate will replace the aggregate currently associated with this
    # object.
    if ( $args->{Aggregate} ) {
        NOTE::logcomment("Volume::_recycle: using Aggregate arguments:\n"
                         . Dumper($args));
        if ( ref $args->{Aggregate} eq 'HASH' ) {
            if ( defined( $args->{Aggregate}->{name} ) ) {

                # Resolve this now, because it might end up being resolved to
                # undef
                $args->{Aggregate}->{name} =
                                     NOTE::option( $args->{Aggregate}->{name} );
            }
            if ( defined( $args->{Aggregate}->{name} ) ) {

                # named aggregate
                # Only look for volumes within this aggregate
                $rtn = eval {
                            $aggr = NOTE::Aggregate->attach(
                                                    Filer => $self->{_Filer},
                                                    %{ $args->{Aggregate} }, );
                            1;
                        };

                if (! $rtn) {
                    return 0;
                }

                @volList = $aggr->Volumes( spare_only => 1, %opts );
            }
            else {

                # anonymous aggregate
                # Here, I just want to note the characteristics
                # of the aggregate so I can match candidates
                # using them.
                $aggrQualities = $args->{Aggregate};
                @volList = $self->{_Filer}->Volumes( spare_only => 1, %opts );
            }
        }
        elsif ( ref $args->{Aggregate} eq 'NOTE::Aggregate' ) {
            @volList = $args->{Aggregate}->Volumes( spare_only => 1, %opts );
        }
    } elsif ( defined($self->{_Aggregate}) ) {
        @volList = $self->{_Aggregate}->Volumes( spare_only => 1, %opts );
    } else {
        @volList = $self->{_Filer}->Volumes( spare_only => 1, %opts );
    }

    for $vol (@volList) { $vol->size }    # Build 'size' cache
    @volList = sort { $a->{_size} <=> $b->{_size} } @volList;

    for $vol (@volList) {
        eval {             # If any error occurs, simply try the next candidate
            next unless ( $vol->is_free );
            unless ( $vol->type eq $type ) {
                _reject(
                    name   => $vol->name,
                    reason => 'Volume types do not match',
                    wanted => 'Requested type is [' . $type . ']',
                    got    => 'Candidate type is [' . $vol->type . ']',
                );
                next;
            }
            NOTE::logcomment("volume type = $type ");

            if ( $type =~ /flex|striped/ ) {
                if ( defined $args->{size} ) {
                    unless ( $vol->{_size} == NOTE::units( $args->{size} ) ) {
                        _reject(
                            name   => $vol->name,
                            reason => 'Volume sizes do not match',
                            wanted => 'Requested size is ['
                                      . NOTE::units( $args->{size} ) . 'k]',
                            got    => "Candidate size is [$vol->{_size}k]",
                        );

                        # Since volume list is sorted from smallest to largest,
                        # we can stop looking if a candidate is larger than the
                        # requested size..
                        if ( $vol->{_size} > NOTE::units( $args->{size} ) ) {
                            last;
                        }

                        next;    # Just move on to the next candidate
                    }
                }

                # If flex reject volumes that aren't in aggregates
                # with characteristics that match specified disks,
                # raid type, etc..
                if ($aggrQualities) {
                    $aggr = $vol->Aggregate;
                    $aggrOpts = $aggr->options;

                    # Make sure we get a volume in a shareable aggregate if
                    # we say we want one or a volume in an unshareable
                    # aggregate if we say we don't want one.
                    if ( defined $aggrQualities->{share} ) {
                        if ( $aggrQualities->{share} ne $aggr->is_shareable ) {
                            _reject(
                                name   => $vol->name,
                                reason => 'containing aggregate [' . $aggr->name
                                          . '] qualities do not match',
                                wanted => 'requested {share} ['
                                          . $aggrQualities->{share} . ']',
                                got    => 'candidate {share} ['
                                          . $aggr->is_shareable . ']',
                            );
                            next;
                        }
                    }
                    if ( $self->isa('NOTE::GX') ) {
                        if ( $aggrQualities->{voltype} ) {
                            my $aggrtype = scalar( $aggr->voltype );
                            unless ( $aggrQualities->{voltype} != $aggrtype ) {
                                _reject(
                                    name   => $vol->name,
                                    reason => 'containing aggregate is the '
                                              . 'wrong type',
                                    wanted => 'requested type ['
                                              . $aggrQualities->{type} . ']',
                                    got    => 'candidate type ['
                                              . $aggrtype . ']',
                                );
                                next;
                            }
                        }
                        if ( $aggrQualities->{nodes} ) {
                            my $nodestr = join( ',', sort @{ $aggr->nodes } );
                            unless ( $aggrQualities->{nodes} eq $nodestr ) {
                                _reject(
                                    name   => $vol->name,
                                    reason => 'containing aggregate comprises '
                                              . 'incompatable nodes',
                                    wanted => 'requested nodes ['
                                              . $aggrQualities->{nodes} . ']',
                                    got    => 'candidate nodes ['
                                              . $nodestr . ']',
                                );
                                next;
                            }
                        }
                    }
                    if ( $aggrQualities->{disks} ) {
                        if ( $aggrQualities->{disks} > scalar($aggr->disks) ) {
                            _reject(
                                name   => $vol->name,
                                reason => 'containing aggregate has too few '
                                          . 'disks',
                                wanted => 'requested disks ['
                                          . $aggrQualities->{disks} . ']',
                                got    => 'candidate disks ['
                                          . $aggr->disks . ']',
                            );
                            next;
                        }
                    }
                    if ( $aggrQualities->{raidtype} ) {
                        if ( $aggrQualities->{raidtype}
                             ne $aggrOpts->{raidtype} ) {
                            _reject(
                                name   => $vol->name,
                                reason => 'containing aggregate raidtype '
                                          . 'does not match',
                                wanted => 'requested raidtype ['
                                          . $aggrQualities->{raidtype} . ']',
                                got    => 'candidate raidtype ['
                                          . $aggrOpts->{raidtype} . ']',
                            );
                            next;
                        }
                    }
                    if ( $aggrQualities->{raidsize} ) {
                        if ($aggrQualities->{raidsize} != $aggrOpts->{rgsize}) {
                            _reject(
                                name   => $vol->name,
                                reason => 'containing aggregate raidsize '
                                          . 'does not match',
                                wanted => 'requested raidsize ['
                                          . $aggrQualities->{raidsize} . ']',
                                got    => 'candidate raidsize ['
                                          . $aggrOpts->{raidsize} . ']',
                            );
                            next;
                        }
                    }
                    if ( $aggrQualities->{indirect_type} ) {
                        # the indirect_type can be the cmode values (32-bit,
                        # 64-bit) or the 7g values (32, 64) so cannot do a
                        # simple compare
                        if ( $aggrOpts->{'block-type'}
                             !~ /$aggrQualities->{indirect_type}/i ) {
                            _reject(
                                name   => $vol->name,
                                reason => 'containing aggregate block-type '
                                          . 'does not match',
                                wanted => 'requested block-type ['
                                          . "$aggrQualities->{indirect_type}]",
                                got    => 'candidate block-type ['
                                          . $aggrOpts->{'block-type'} . ']',
                            );
                            next;
                        }
                    }
                    if ( $self->isa('NOTE::7G') ) {

                        # If we want a snaplock aggregate, make sure they're
                        # the same type
                        if ( $aggrQualities->{snaplock} ) {
                            if ($aggrQualities->{snaplock} ne $aggr->snaplock) {
                                _reject(
                                    name   => $vol->name,
                                    reason => 'containing aggregate snaplock '
                                              . 'qualities do not match',
                                    wanted => 'requested {snaplock} ['
                                              . "$aggrQualities->{snaplock}]",
                                    got    => 'candidate {snaplock} ['
                                              . $aggr->snaplock . ']',
                                );
                                next;
                            }
                        }
                        else {

                            # If we don't want one, reject any that are.
                            if ( $aggr->snaplock ) {
                                _reject(
                                    name   => $vol->name,
                                    reason => 'containing aggregate snaplock '
                                              . 'qualities do not match',
                                    wanted => 'requested no snaplock ',
                                    got    => 'candidate {snaplock} ['
                                              . $aggr->snaplock . ']',
                                );
                                next;
                            }
                        }
                    }
                }
            }
            elsif ( $type eq 'trad' ) {
                if ( defined $args->{disks} ) {
                    unless ( scalar( $vol->disks ) == $args->{disks} ) {
                        _reject(
                            name   => $vol->name,
                            reason => 'number of disks do not match',
                            wanted => "requested disks [$args->{disks}]",
                            got    => 'candidate disks [' . $vol->disks . ']',
                        );
                        next;
                    }
                }
                if ( defined $args->{snaplock} ) {
                    unless ( $vol->snaplock eq $args->{snaplock} ) {
                        _reject(
                            name   => $vol->name,
                            reason => 'volume snaplock type does not match',
                            wanted => "requested snaplock [$args->{snaplock}]",
                            got    => 'candidate snaplock ['
                                      . $vol->snaplock . ']',
                        );
                        next;
                    }
                }
                else {

                    # If we don't want a snaplock volume, reject any that are.
                    if ( $vol->snaplock ) {
                        _reject(
                            name   => $vol->name,
                            reason => 'volume snaplock qualities do not match',
                            wanted => 'requested no snaplock ',
                            got    => 'candidate {snaplock} ['
                                      . $vol->snaplock . ']',
                        );
                        next;
                    }
                }
            }
            else {
                NOTE::logcomment( "Unknown volume type: %s\n", $vol->type );
                next;
            }
            unless ( $vol->writeable ) {
                _reject(
                    name   => $vol->name,
                    reason => 'Volume is not writeable',
                );
                next;
            }
            my ( $prev_name, $new_name );
            $prev_name = $vol->name;
            if ( $self->{_dataset} ) {
                next unless ( $vol->dataset eq $self->{_dataset} );
            }
            else {

                # If no dataset required, reject volumes with datasets
                next if ( $vol->dataset );
            }
            $new_name = $self->{_name};
            $self->{_Filer} = $vol->Filer;
            if ( $self->{_type} ne 'trad' ) {
                $self->{_Aggregate} = $vol->Aggregate;
            }
            NOTE::logcomment( 'Recycling %s volume:  %s --> %s ',
                              $self->{_type}, $prev_name, $new_name );
            $vol->_rename($new_name);
            if ( $self->isa('NOTE::7G') && ( $type eq 'flex' ) ) {
                if ( $aggrQualities->{snaplock} eq 'compliance' ) {
                    $self->{_Aggregate}->Filer->check_license('snaplock');
                }
                elsif ( $aggrQualities->{snaplock} eq 'enterprise' ) {
                    $self->{_Aggregate}->Filer->check_license(
                                                         'snaplock_enterprise');
                }
            }
            $self->{_was_recycled} = 1;
            goto gotOne;    # got one
        };

        if ($@) {
            NOTE::logcomment("Attempted recycle failed: $@");
        }
    }
    NOTE::logcomment('No candidate found');

    # No candidate found

    # XXX If I want a volume with a dataset, I should be
    # able to convert an empty volume to a populated volume
    # if no appropriate populated volumes exists.
    # And vice-versa - If I want an empty volume,
    # and none exist, I might want to consider removing
    # data from a populated volume, if there are no more
    # disks available.
    return 0;
  gotOne:
    $self->is_destroyed = 0;
    return 1;
}

=head2 ::create

This constructor creates an object representing a volume.
The method determines if a NOTE::7G::Volume or NOTE::GX::Volume
object is required,
see L<C<NOTE::is_GX>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-pm.html#%3a%3ais_GX>
for details, and calls the corresponding 'create' method to complete the common construction
of this object.

=over

=item Arguments

Arguments will be in the form of a hash which will be passed to NOTE::7G::Volume
or NOTE::GX::Volume for processing.

Please see
L<C<NOTE::7G::Volume::create>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-7G-Volume-pm.html#%3a%3acreate>
or
L<C<NOTE::GX::Volume::create>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3acreate>
for details

=item Examples

=back

    my $vol_7g = NOTE::Volume::create(Filer=>"dpg-gx4");
    my $vol_gx = NOTE::Volume::create(Vserver=>"vs0", Node=>"dpg-gx4");

=cut

sub create {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($args, $attempt, $error, $max_attempts, $rtn, $num_disks);

    if ( !ref $self ) {
        if ( NOTE::is_GX(%args) ) {
            return NOTE::GX::Volume->create(%args);
        }
        else {
            return NOTE::7G::Volume->create(%args);
        }
    }

    # At this point, $self is an object
    if ( $self->exists ) {
        return $self;
    }

    $args = $self->args;
    if ( $self->{_type} =~ /^(flex|striped)$/i ) {
        # make 4 attempts to create the volume (0-3), adding up to
        # 3 disks in the process
        $max_attempts = 3;

        $attempt = 0;
        $rtn = eval { $self->_vol_create( %{$args} ); 1; };
        $error = $@;
        while ( ($attempt < $max_attempts) &&
                ( ($error =~ /Not enough space/i) ||      # flex aggr error
                  ($error =~ /Insufficient space/i) ||    # 7g aggr error?
                  ($error =~ /no more space/i) ) ) {      # striped aggr error
            NOTE::logcomment( "Insufficient capacity in %s; trying to grow "
                              . 'aggregate', $self->Aggregate->name, );

            # Try adding a disk(s)
            # Look at the type of the aggr
            $num_disks = 1;
            if ( $self->{_type} eq 'striped' ) {
                $num_disks = 2 * scalar( split( /,/,
                                         $self->Aggregate->{_args}->{nodes}) );
            }

            # If this fails, we could also try deleting free volumes
            # with datasets in this aggregate
            $self->Aggregate->add( disks => $num_disks );
# assume this is to allow the new disk to come online,
# should use something better for this - stuw
            sleep(10);

            $attempt += 1;
            $rtn = eval { $self->_vol_create( %{$args} ); 1; };
            $error = $@;
        }

        if ( $error && $attempt ) {
            NOTE::logcomment('Volume create failed even after growing aggregate');
        }

        if ($error) {
            confess $error;
        }
    } else {
        delete $args->{type};
        my $diskspec = $args->{disklist} ? $args->{disklist} : $args->{disks};
        delete $args->{disks};
        delete $args->{disklist};

        if ( $self->{_snaplock} eq 'compliance' ) {
            $self->Filer->check_license('snaplock');
        }
        elsif ( $self->{_snaplock} eq 'enterprise' ) {
            $self->Filer->check_license('snaplock_enterprise');
        }
        while (1) {
            eval { $self->_vol_create_trad( disks => $diskspec, %{$args} ) };
            last unless ($@);
            if ( $@ =~ /Not enough disks/ ) {

                # Try again if we can free some disks
                next if NOTE::Aggregate::_releaseSpareAggr( $self->{_Filer} );
            }
            confess "Trad volume create error: $@";
        }
    }
    return $self->_init;
}

=head2 ::args

Returns reference to arguments used in "new", "create" and "attach" methods.

=cut

sub args {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_args}
}

=head2 ::new

Constructor to that creates the volume object without
creating or attaching to an ONTAP volume.

Upon determining the type of volume objexct required, whether NOTE::7G::Voleume
or NOTE::GX::Volume, this method will call the corresponding 'new' methods
to complete the construction of the object.

=over

=item Arguments

Arguments will be in the form of a hash which will be passed to NOTE::7G::Volume
or NOTE::GX::Volume for processing.

Please see
L<C<NOTE::7G::Volume::new>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-7G-Volume-pm.html#%3a%3anew>
or
L<C<NOTE::GX::Volume::new>|http://web.netapp.com/engineering/ontap_tests/agnostic/docs/lib-GNOTE-GX-Volume-pm.html#%3a%3anew>
for details

=item Examples

=back

    my $vol_7g = NOTE::Volume::new(Filer=>"dpg-gx4");
    my $vol_gx = NOTE::Volume::new(Vserver=>"vs0", Node=>"dpg-gx4");

=cut

sub new {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $exportfs_opts;
    my $diskspec;
    my $size;
    my $rtn;

    if ( !ref $self ) {
        if ( NOTE::is_GX(%args) ) {
            return NOTE::GX::Volume->new(%args);
        }
        else {
            return NOTE::7G::Volume->new(%args);
        }
    }
    my $args = $self->{_args};
    $self->{_destroyed} = 1;    # Doesn't physically exist yet

    $self->{_type} = delete $args->{type};
    if ( $args->{name} ) {
        $self->{_name} = $args->{name};

        # Allow named volumes that behave like anonymous volumes:
        $self->{_anonymous} = $args->{anonymous};
    }
    else {                      # Create a unique name
        $self->{_name}      = NOTE::Volume->newName;
        $self->{_anonymous} = 1;
    }
    delete $args->{name};
    delete $args->{anonymous};

    $self->{_recycle}  = delete $args->{recycle};
    $self->{_tag}      = delete $args->{tag};
    $self->{_teardown} = delete $args->{teardown};
    $self->{_snaplock} = delete $args->{snaplock};
    $self->{_richdata} = delete $args->{richdata};
    $self->{_dataset}  = delete $args->{dataset};

    if ( $args->{'mirror-type'} =~ /dp/i ) {
        confess "DP type and recycle are mutually exclusive"
          if $self->{_recycle};
    }
    else {

        # RW volumes are recycled by default
        $self->{_recycle} = 1 unless defined $self->{_recycle};
    }

    # At this point, $self is an object
    if ( $self->exists ) {
        $self->{_destroyed} = 0;
        if ( $self->{_type} ne 'trad' ) {
            $self->{_Aggregate} = NOTE::Aggregate->attach(
                Filer => $self->{_Filer},
                name  => $self->_vol_aggr,
                tag   => "Container of " . $self->description,
            );
        }
    }
    elsif ( !$self->{_type} || $self->{_type} =~ /^(flex|striped)$/ ) {
        if ( $self->{_type} ) {
            NOTE::logcomment( "Require %s volume, requested size is %s", $self->{_type}, defined $args->{size} ? $args->{size} : "'default'", );
        }

        # if the $args->{Aggregate} is an aggregate object, use it,
        # otherwise try to create one by calling the NOTE::Aggregate
        # constructor.
        my %extra = ();
        if ( $args->{local_storage} ) {
            $extra{local_storage} = $args->{local_storage};
            $extra{Node}          = $self->{_Node};
            NOTE::logcomment( $args->{local_storage} ? "Use local storage on node " . $extra{Node}->name : "" );
            $extra{Filer} = $self->{_Node};
        }

        if ( !defined( $args->{Aggregate} ) ) {

            # If the volume is a snaplock volume, propagate this information
            # to the aggregate create method.
            if ( $self->{_snaplock} ) {
                $extra{'snaplock'} = $self->{_snaplock};
            }
            $self->{_Aggregate} = NOTE::Aggregate->create(
                Filer => $self->{_Filer},
                tag   => "Container of " . $self->description,
                %extra
            );

        }
        elsif ( ref $args->{Aggregate} eq "HASH" ) {

            unless ( $args->{Aggregate}->{tag} ) {
                $args->{Aggregate}->{tag} = "Container of " . $self->description;
            }

            # If the volume is a snaplock volume, propagate this information
            # to the aggregate create/attach method.
            if ( $self->{_snaplock} ) {
                $args->{Aggregate}->{'snaplock'} = $self->{_snaplock};
            }

            # If it's named, try to attach it.
            if ( $args->{Aggregate}->{name} ) {
                $rtn = eval {
                        $self->{_Aggregate} = NOTE::Aggregate->attach(
                                                    Filer => $self->{_Filer},
                                                    %{ $args->{Aggregate} },
                                                    %extra );
                        1;
                    };

                if (1 == $rtn) {
                    if ( $self->{_Aggregate}->verify_aggregate_attributes(
                                                    Filer => $self->{_Filer},
                                                    %{ $args->{Aggregate} },
                                                    %extra) ) {
                        confess('CONFIG ERROR: named aggregate does not match prerequisites');
                    }
                }
            }

            # If it's unnamed or the attach fails, try
            # to create it.
            if ( ( !$args->{Aggregate}->{name} ) || !$rtn ) {
                $self->{_Aggregate} = NOTE::Aggregate->create(
                    Filer => $self->{_Filer},
                    %{ $args->{Aggregate} },
                    %extra
                );
            }
        }
        else {
            $self->{_Aggregate} = $args->{Aggregate};
        }
        delete $args->{Aggregate};
        unless ( $self->{_type} ) {
            if ( $self->isa('NOTE::GX') ) {
                $self->{_type} = $self->Aggregate->voltype;
            }
            else {
                $self->{_type} = 'flex';
            }
        }
        if ( $self->{_recycle} ) {
            if ( $self->{_anonymous} ) {

                # Only bother recycling anonymous flex/striped vols with datasets
                if ( $self->{_dataset} && $self->_recycle(%extra) ) {
                    return $self;
                }
            }
            else {
                if ( $self->exists ) {
                    my $type = $self->_vol_type;
                    if ( $self->_vol_type !~ /(flex|striped)/ ) {
                        confess "Volume named " . $self->{_name} . " is a '" . $type . "'";
                    }
                }
            }
        }
    }
    elsif ( $self->{_type} eq 'trad' ) {
        $args->{raidtype} ||= NOTE::option(q/$RAIDTYPE:raid_dp/);    # Default
        if ( !( $args->{disks} ) && !( $args->{disklist} ) ) {
            if ( $args->{raidtype} eq "raid4" ) {
                $args->{disks} = "2";
            }
            elsif ( $args->{raidtype} eq "raid_dp" ) {
                $args->{disks} = "3";                                # Need at least 3 disks for raid_dp
            }
            else {
                confess "Invalid raidtype: '" . $args->{raidtype} . "'";
            }
        }

        my $numdisks = scalar @{ $args->{disklist} } if ( $args->{disklist} );
        $numdisks ||= $args->{disks};
        NOTE::logcomment( "Require %s trad volume with %d disks: " . $args->{raidtype} . " " . $numdisks );

        if ( $self->{_recycle} ) {
            if ( $self->{_anonymous} ) {
                unless ( $self->_recycle ) {

                    # This is a bit tricky: If I wasn't able to recycle
                    # a volume that was populated with a dataset, try to
                    # recycle an empty one and then populate it as if
                    # it was just created.
                    if ( my $dataset = $self->{_dataset} ) {
                        $self->{_dataset} = undef;
                        my $gotit = $self->_recycle;
                        $self->{_dataset} = $dataset;
                        return $self->_init if ($gotit);
                    }
                }
            }
            else {    # Named recyclable volume
                if ( $self->_vol_type ne "trad" ) {
                    confess "Volume named " . $self->{_name} . " is not a tradvol";
                }
            }
        }
    }
    else {
        confess "Unknown volume type " . $self->{type};
    }
    return $self;
}

sub _match_aggr_type {
    my ( $self, $aggr, $vol_type ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    # Only check type if it is an Aggregate Object
    if ( ref($aggr) =~ /NOTE::.*Aggregate/ ) {
        my $aggr_type = $aggr->voltype;
        confess( "Unable to create $vol_type volume on $aggr_type aggregate " . $aggr->name )
          unless $aggr_type eq $vol_type;
    }
}

sub _init {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    $self->is_destroyed = 0;
    NOTE::reference($self);
    if ( $self->{_dataset} ) {
        NOTE::logcomment( "Initializing dataset of " . $self->description );
        my $func = eval 'sub {' . $self->{_dataset} . '}';
        confess $@ if $@;
        eval { $self->$func };
        confess $@ if $@;

        #        if ( $self->_looksAnonymous ) {
        #            $self->_rename( NOTE::datasetEncode( $self->{_name}, $self->{_dataset} ) );
        #        }

        # Make sure container of volume is spared if dataset is recyclable
        if ( $self->{_recycle} && $self->type ne 'trad' ) {
            $self->Aggregate->SPARE;
        }
    }
    if ( $self->is_anonymous ) {
        eval { $self->snapshot_delete("__clean__") };    # Don't care if this fails
        if ( $self->{_recycle} ) {
            $self->snapshot_create("__clean__");
        }
    }
    return $self;
}

=head2 ::substant

Returns true if this object represents a volume that
might exist. This is different than the ::exists method
because it doesn't actually query the server, but rather
returns the state of an internal flag that indicates
whether the volume has been created yet.

=over

=item Argument

None.

=item Example

=back

     $vol = NOTE::Volume->new;
     die if ( $vol->substant );
     printf( "Volume size will be %dk\n", $vol->size );

=cut

sub substant {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->is_destroyed ) != 1
}

=head2 ::exists

Returns true if this object represents a volume that
physically exists.

=over

=item Argument

None.

=item Example

=back

     $vol = NOTE::Volume->create;
     die unless ( $vol->exists );
     printf( "Volume size is be %dk\n", $vol->size );

=cut

sub exists {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->_vol_exists
}

=head2 ::attach

Creates an object representing an existing ONTAP volume without creating it's
ONTAP counterpart.

=over

=item Arguments

B<name> => I<string>

I<Required.>
The name of an existing volume.

B<type> => C<flex> | C<trad>

I<Optional.>
The type of volume given by B<name>.
If given the attach fails if the volume is not of that type.
If absent the actual type is recorded in the volume object.

B<Filer> => I<host spec>

I<Required.>
Specifies the filer where the volume exists.

B<mntpoint> => I<string>

I<Optional.>
The full path name on which to mount the volume.

B<exportfs> => I<export options>

I<Optional.>
Export options as described in L<C<::exportfs>|/::exportfs>.
You will probably be sorry if you don't use at least 'C<anon=0>'.

B<force> => C<1>

I<Optional.>
A flag to force the creation of a volume object
even if the named volume does not exist.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip for
L<C<::populate>|/::populate> and L<C<::modify>|/::modify>.
The default is 0.

=item Examples

=back

     $vol = NOTE::Volume->attach(
         Filer => $FILER,
         name  => 'vold',
     );

=cut

sub attach {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( !ref $self ) {
        if ( NOTE::is_GX(%args) ) {
            return NOTE::GX::Volume->attach(%args);
        }
        else {
            return NOTE::7G::Volume->attach(%args);
        }
    }

    my $force = delete $args{force};

    $self->_connect;

    my $args = $self->{_args};

    confess "Volume attach requires a name" unless ( $args->{name} );
    $self->{_name}    = NOTE::option( delete $args->{name} );
    $self->{_dataset} = NOTE::datasetDecode( $self->{_name} );

    $self->{_tag}      = delete $args->{tag};
    $self->{_richdata} = NOTE::option( delete $args->{'richdata'} );

    if ( exists $self->{_parent} ) {
        return $self;
    }
    my $desc = $self->{_name};
    $desc .= " (" . $self->{_tag} . ")" if ( $self->{_tag} );

    my $type = $self->_vol_type;
    unless ( $type || $force ) {
        confess "Volume $desc does not exist";
    }

    if ( $type eq "trad" ) {
        if ( exists $args->{type} && $args->{type} ne "trad" ) {
            confess "Volume $desc is of type trad, not " . $args->{type};
        }
        $self->{_type} = "trad";
    }
    elsif ( $type =~ /(flex|striped)/ ) {
        if ( exists $args->{type} && $args->{type} ne $type ) {
            confess "Volume $desc is of type $type, not " . $args->{type};
        }
        $self->{_type}      = $type;
        $self->{_Aggregate} = NOTE::Aggregate->attach(
            Filer => $self->{_Filer},
            name  => $self->_vol_aggr
        );
    }
    delete $args->{type};
    return $self;
}

=head2 ::dup

Create (or reuse) a volume with attributes similar to an existing volume.

=over

=item Arguments

Takes the same arguments as NOTE::Volume::create

=item Examples

=back

  $vol_dst = $vol_src->dup;  # Make a destination volume similar to the source

  $vol_dst = $vol_src->dup( Filer => q/$FILER_DST/ ); # The same, but different filer

  $vol_dst = $vol_src->dup( Aggregate => $aggr_dst ); # The same, but different aggregate

=cut

sub dup {
    my ( $origin, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %properties = $origin->Properties;

    %properties = %{
        NOTE::_hashOverlay(
            target    => \%properties,
            source    => \%args,
            overwrite => 1
        )
      };
    $properties{Filer} ||= $origin->Filer;
    return NOTE::Volume->create(%properties);
}

=head2 ::Properties

Deterimnes a volume's properties via 'volume show' and returns them
in a hash.

=over

=item Arguments

I<None>

=item Example

=back

    my $vol_info = $vol->Properties();

=cut

sub Properties {
    my $self       = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %properties = (
        anonymous => $self->{_anonymous},
        richdata  => $self->{_richdata},
        recycle   => $self->{_recycle},
        type      => $self->{_type},
        snaplock  => $self->{_snaplock},
        exportfs  => $self->{_exportfs},
    );
    if ( $properties{type} eq 'trad' ) {
        my $opts = $self->options;
        $properties{disks}    = scalar $self->disks;
        $properties{raidtype} = $opts->{raidtype};
        $properties{rgsize}   = $opts->{raidsize};
    }
    else {
        $properties{Aggregate} = { $self->Aggregate->Properties };
        $properties{size}      = $self->size;
    }
    return %properties;
}

=head2 ::rename

Changes the name of the volume.

=over

=item Arguments

I<string>

I<Optional.>
The new name for the volume.
If absent a new anonymous name is generated
and the volume becomes anonymous.
If present the volume will no longer be anonymous,
causing its aggregate to be spared
(not destroyed) when the volume object is destroyed.

=item Examples

=back

     $vol->rename;
     $vol->rename( 'newName' );

=cut

sub rename {
    my ( $self, $newname ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $oldname = $self->{_name};
    if ( defined $newname ) {
        $self->_rename($newname);
        if ( $self->{_anonymous} && !$self->_looksAnonymous ) {

            # Converting from anonymous to non-anonymous
            $self->SPARE;
            $self->{_anonymous} = 0;
        }
    }
    else {
        $newname = $self->newName;
        $self->{_anonymous} = 1;
        $self->_rename($newname);
    }
    return $oldname;
}

sub _rename {
    my ( $self, $newname ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->_vol_rename($newname);
    $self->{_name} = $newname;
}

=head2 ::mounted

Returns the NFS mount point of the volume.
Returns C<undef> if the volume is not mounted.

=over

=item Arguments

B<client> => I<hostname>

I<optional>

Name of client to check if volume have been mounted on. Default: localhost

=item Example

=back

     $mntpt = $vol->mounted; # checks localhost mount
     $mntpt = $vol->mounted(client=>'siml8'); # checks mount for client 'siml8

=cut

sub mounted {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $client = $args{client} || "localhost";
    if ( $self->{_mntpoint} ) {
        return $self->{_mntpoint}->{$client};
    }
    else {
        return;
    }
}

=head2 ::mountpoint

A synonym for L<C<::mounted>|/::mounted>.

=over

=item Arguments

B<client> => I<hostname>

I<Optional>

Name of a client to obtain mountpoint for.

=item Example

=back

    my $mp = $vol->mountpoint(); # obtains mountpoint of 'localhost'
    my $cm = $vol->mountpoint(client=>"siml8"); # returns mountpoint of host 'siml8'

=cut

sub mountpoint {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $client = $args{client} || "localhost";
    if ( exists $self->{_mntpoint} ) {
        if ( exists $self->{_mntpoint}->{$client} ) {
            return $self->{_mntpoint}->{$client};
        }
        else {
            confess $self->description . ": no mountpoint for client $client";
        }
    }
    else {
          confess $self->description . ": no mountpoints";
    }
}

=head2 B<Options and Attributes>

=cut

=head2 ::type

Returns the type of the volume.
The return value is one of
the strings C<'flex'> or C<'trad'>.

=over

=item Arguments

None.

=item Examples

=back

     $type = $vol->type;

=cut

sub type {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_type}
}

use DataManip qw();
use FSR::FSRutils;

=head2 ::populate

Populate volume with data using DataManip::populate_tree or RichData::populate.

=over

=item Arguments

B<datagen> => C<0 | 1>

I<Optional.>
Use dp/snapmirror/paloma/datagen10.thpl as engine. This can also be turned on by setting up
cli parameter: USEDATAGEN=1

All CLI parameters normally used with datagen can be used now with any GNOTE script.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip.
The default is the value set on L<C<::create>|/::create>
or L<C<::attach>|/::attach>.
Subsequent calls to either L<C<::populate>|/::populate>
or L<C<::modify>|/::modify> will use this setting as default.

B<%args> - I<dataManip|RichData argument Hash>

Optional.

All parameters, in the form of a hash is passed to DataManip::populate_tree
or RichData::populate for processing.

=item CLI parameters

B<DATASET_SNAPSHOT> => I<name>

Name of snapshot to use for predefined dataset. If snapshot is present, then snaprestore will be called instead of actually
running populate. If snapshot is not present, then populate() will run first, then snapshot will be taken for future use.

B<USEDATAGEN> => I<1/0>

Use datagen with populate, ignores all other types of populate. Datagen will need its own parameters as well.
Check dp/snapmirror/paloma/datagen10.thpl for info on datagen's parameters.

=item Examples

=back

      $vol->populate;

      $vol->populate("FILE_DEV" => 100, "DIR_DEPTH" => 3);

      # Use RichData::populate to create streams, sparse files, links, etc.
      # in addition to files and directories
      $vol->populate(richdata => 1);

=cut

sub populate {
      my ( $self, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      # Allow populate/modify/compare/delete_tree to be bypassed:
      return $self if ( NOTE->DATAMANIP == 0 );

      my $dataset_snapshot = Tharn::param('DATASET_SNAPSHOT');
      if( $dataset_snapshot && $self->snapshot_exists($dataset_snapshot)) {
          NOTE::logcomment( "Found snapshot with dataset, restoring to snapshot '$dataset_snapshot'");
	  $self->snapshot_restore($dataset_snapshot);
	  return $self;
      }

      if ( $args{datagen} || Tharn::param('USEDATAGEN') ) {
          if( $$self{_datagenused} ) {
              Tharn::subtest("dp/snapmirror/paloma/datagen10.thpl", "FILER_VOL=".$self->name, "FILER=".$self->Node->name, "DATAGEN_OP=DELETE", "VSERVER=".$self->Vserver->name );
              delete $$self{_datagenused};
          }
          Tharn::subtest("dp/snapmirror/paloma/datagen10.thpl", "FILER_VOL=".$self->name, "FILER=".$self->Node->name, "DATAGEN_OP=POPULATE", "VSERVER=".$self->Vserver->name );
          $$self{_datagenused} = 1;

	  if( $dataset_snapshot ) {
	      NOTE::logcomment( "Creating DATASET snapshot '$dataset_snapshot'");
	      $self->snapshot_create($dataset_snapshot);
	  }
          return $self;
      }

      my $client = $args{client} || $self->{_client} || "localhost";
      delete $args{client};

      my $was_mounted = $self->mounted( client => $client );
      my ( $percent, $ret, $fullpath );
      my $path = $self->pwd;
      my %ARGS = ();

      if ( exists $args{richdata} ) {

          # Overwrite the richdata option for the object
          $self->{_richdata} = $args{richdata};
          delete $args{richdata};
      }

      $ARGS{ uc $_ } = $args{$_} for ( keys %args );

      # Communicate filer type to DataManip
      if ( ref($self) =~ /^NOTE::GX/ ) {
          $ARGS{ENVIRONMENT_TYPE} ||= "filer";
      }
      else {
          $ARGS{ENVIRONMENT_TYPE} ||= $self->Filer->is_a_simulator ? "simulator" : "filer";
          $self->exportfs("rw,nosuid,anon=0");
      }
      unless ($was_mounted) {
          $self->mount( 'chmod' => '777', client => $client );
      }

      my $op = "Populating $path of " . $self->description . " from client $client";
      NOTE::logcomment($op);
      if( ref($self->{_mntpoint}) ) {
          $fullpath = $self->{_mntpoint}->{$client} . $path;
      } else {
          $fullpath = $self->{_mntpoint} . $path;
      }
      if ( $self->{_richdata} ) {
          delete $args{background};
          $args{client} = $client;
          $ret = Tharn::subtest(
              sub {
                  NOTE::logcomment("Using RichData::populate");
                  eval { RichData::populate( path => $fullpath, %args ) };
                  Tharn::logresult( 'FATAL', "EVAL_ERROR=$@" ) if ($@);
              },
              '-bg',
              '-onexit',
              sub { }
          );
          $ret->wait_finish unless $ARGS{BACKGROUND};
          FSR::FSRutils::subtestResults($ret) or confess $@;
      }
      else {
          NOTE::logcomment("Using DataManip::populate_tree");
          $ret = DataManip::populate_tree( $client, $fullpath, %ARGS );
          confess $@ if $@;
      }

      if ( ref $ret eq 'Subtest' ) {
          push @{ $self->{_subtests} }, { operation => $op, subtest => $ret };
      }
      else {
          $ret = $self;
      }

      unless ($was_mounted) {
          if ( $ARGS{BACKGROUND} ) {

              # Don't unmount if in background because it might
              # still be running. Warn unless anonymous, because named volume
              # mounts don't get cleaned up on teardown
              NOTE::logcomment( "Warning: " . $self->name . " was not unmounted because populate was run in the background" ) unless ( $self->{_anonymous} );
          }
          else {
              $self->umount( client => $client );
          }
      }

      if( ! $ARGS{BACKGROUND} && $dataset_snapshot ) {			    # if dataset snapshot is requested, and not yet created, and not a background process,
	  NOTE::logcomment("Taking dataset snapshot '$dataset_snapshot'");  # then create requested snapshot. There is no need to check if snapshot already exists
	  $self->snapshot_create($dataset_snapshot);                        # because if it would exists then routine would not even get down here
      }

      return $ret;
}

=head2 ::modify

Modify data using DataManip::modify_tree or RichData::modify.

=over

=item Arguments

B<datagen> => C<0 | 1>

I<Optional.>
Use dp/snapmirror/paloma/datagen10.thpl as engine. This can also be turned on by setting up
cli parameter: USEDATAGEN=1

All CLI parameters normally used with datagen can be used now with any GNOTE script.

B<richdata> => C<0 | 1>

I<Optional.>
If set, use RichData library instead of DataManip.
The default is the value set on L<C<::create>|/::create>
or L<C<::attach>|/::attach>.
Subsequent calls to either L<C<::populate>|/::populate>
or L<C<::modify>|/::modify> will use this setting as default.

B<%args> - I<dataManip|RichData argument Hash>

Optional.

All parameters, in the form of a hash is passed to DataManip::modify_tree
or RichData::modify for processing.

=item Examples

=back

      # Modify volume using either DataManip or RichData according to
      # the setting of the richdata flag on populate, create or attach
      $vol->modify;

=cut

sub modify {
      my ( $self, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      # Allow populate/modify/compare/delete_tree to be bypassed:
      return $self if ( NOTE->DATAMANIP == 0 );

      if ( $$self{_datagenused} && ( $args{datagen} || Tharn::param('USEDATAGEN') ) ) {
        Tharn::subtest("dp/snapmirror/paloma/datagen10.thpl", "FILER_VOL=".$self->name, "FILER=".$self->Node->name, "DATAGEN_OP=MODIFY", "VSERVER=".$self->Vserver->name );
        delete $$self{_datagenused};
        return $self;
      }

      my $client = $args{client} || $self->{_client} || "localhost";
      delete $args{client};

      my $was_mounted = $self->mounted( client => $client );
      my ( $percent, $ret, $fullpath );
      my $path = $self->pwd;
      my %ARGS = ();

      if ( exists $args{richdata} ) {

          # Overwrite the richdata option for the object
          $self->{_richdata} = $args{richdata};
          delete $args{richdata};
      }

      $ARGS{ uc $_ } = $args{$_} for ( keys %args );

      # Communicate filer type to DataManip
      if ( ref($self) =~ /^NOTE::GX/ ) {
          $ARGS{ENVIRONMENT_TYPE} ||= "filer";
      }
      else {
          $ARGS{ENVIRONMENT_TYPE} ||= $self->Filer->is_a_simulator ? "simulator" : "filer";
          $self->exportfs("rw,nosuid,anon=0");
      }
      unless ($was_mounted) {
          $self->mount( 'chmod' => '777', client => $client );
      }

      my $op = "Modifying $path of " . $self->description . " from client $client";
      NOTE::logcomment($op);
      $fullpath = $self->{_mntpoint}->{$client} . $path;
      if ( $self->{_richdata} ) {
          delete $args{background};
          $args{client} = $client;
          $ret = Tharn::subtest(
              sub {
                  NOTE::logcomment("Using RichData::modify");
                  eval { RichData::modify( path => $fullpath, %args ) };
                  Tharn::logresult( 'FATAL', "EVAL_ERROR=$@" ) if ($@);
              },
              '-bg',
              '-onexit',
              sub { }
          );
          $ret->wait_finish unless $ARGS{BACKGROUND};
          FSR::FSRutils::subtestResults($ret) or confess $@;
      }
      else {
          NOTE::logcomment("Using DataManip::modify_tree");
          $ret = DataManip::modify_tree( $client, $fullpath, %ARGS );
          confess $@ if $@;
      }

      if ( ref $ret eq 'Subtest' ) {
          push @{ $self->{_subtests} }, { operation => $op, subtest => $ret };
      }
      else {
          $ret = $self;
      }

      unless ($was_mounted) {
          if ( $ARGS{BACKGROUND} ) {

              # Don't unmount if in background because it might
              # still be running. Warn unless anonymous, because named volume
              # mounts don't get cleaned up on teardown
              NOTE::logcomment( "Warning: " . $self->name . " was not unmounted because modify was run in the background" ) unless ( $self->{_anonymous} );
          }
          else {
              $self->umount( client => $client );
          }
      }
      return $ret;
}

=head2 ::compare

Calls dataManip::compare_tree to compare this object with the 'other' object

Returns results of DataManip::compare_tree

=over

=item Arguments

B<other> - I<NOTE object>

The other NOTE object to compare with

B<%args> - I<argument hash>

Optional

Hash of arguments passed to dataManip::compare_tree for processing.

=item Examples

=back

     $vol_src->compare($vol_dst);

=cut

sub compare {
      my ( $self, $other, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      # Allow populate/modify/compare/delete_tree to be bypassed:
      if ( NOTE->DATAMANIP == 0 ) {
          NOTE::logcomment("WARNING: Skipping compare as requested via CLI parameter DATAMANIP...");
          return $self;
      }

      my $self_mounted  = $self->mounted;
      my $other_mounted = $other->mounted;
      my $path_self     = $self->pwd;
      my $path_other    = $other->pwd;
      my $client        = $args{client} || 'localhost';
      delete $args{client};

      my %ARGS = ();

      $ARGS{ uc $_ } = $args{$_} for ( keys %args );

      $self->exportfs("rw,nosuid,anon=0");
      $self->mount  unless $self_mounted;
      $other->mount unless $other_mounted;

      my $op = "Comparing\n" . "         $path_self of " . $self->description . "\n" . "    with $path_other of " . $other->description;
      NOTE::logcomment($op);
      my (@ret) = DataManip::compare_trees( $client, $self->{_mntpoint}->{localhost} . $path_self, $other->{_mntpoint}->{localhost} . $path_other, %ARGS );
      confess $@ if $@;

      if ( ref $ret[0] eq 'Subtest' ) {
          push @{ $self->{_subtests} }, { operation => $op, subtest => $ret[0] };
      }

      for my $volinfo ( [ $self, $self_mounted ], [ $other, $other_mounted ] ) {
          my ( $vol, $was_mounted ) = @{$volinfo};
          next if $was_mounted;
          if ( $ARGS{BACKGROUND} ) {
              NOTE::logcomment( "Warning: " . $vol->name . " was not unmounted because compare was run in the background" ) unless ( $vol->{_anonymous} );
          }
          else {
              $vol->umount;
          }
      }
      return @ret;
}

=head2 ::delete

Calls DataManip::delete_tree to delete data from the 'pwd' of this object

=over

=item Arguments

None.

=item Examples

=back

     $vol->delete;

=cut

sub delete {
      my ( $self, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      # Allow populate/modify/compare/delete_tree to be bypassed:
      return $self if ( NOTE->DATAMANIP == 0 );

      my $was_mounted = $self->mounted;
      my $path        = $self->pwd;
      my %ARGS        = ();

      $ARGS{ uc $_ } = $args{$_} for ( keys %args );

      $self->exportfs("rw,nosuid,anon=0");
      unless ($was_mounted) {
          $self->mount( 'chmod' => '777' );
      }
      my $op = "Deleting $path of " . $self->description;
      NOTE::logcomment($op);
      my $ret = DataManip::delete_tree( "localhost", $self->{_mntpoint}->{localhost} . $path, %ARGS );
      confess $@ if $@;

      if ( ref $ret eq 'Subtest' ) {
          push @{ $self->{_subtests} }, { operation => $op, subtest => $ret };
      }
      else {
          $ret = $self;
      }

      unless ($was_mounted) {
          if ( $ARGS{BACKGROUND} ) {

              # Don't unmount if in background because it might
              # still be running. Warn unless anonymous, because named volume
              # mounts don't get cleaned up on teardown
              NOTE::logcomment( "Warning: " . $self->name . " was not unmounted because delete was run in the background" ) unless ( $self->{_anonymous} );
          }
          else {
              $self->umount;
          }
      }
      return $ret;
}

=head2 B<Creating and Using Files>

=cut

=head2 ::fill

Fills a volume to a specified amount of data.

=over

=item Arguments

B<max> => I<number>

Optional

Number of MB to write to Volume. Default 1

B<percent> => I<number>

Optional

Percent of volume to fill rounded to the nearest MB, between 1 - 100.
If supplied supersedes the I<max> parameter above

B<proto> => I<nfs> | I<local>

Optional.

Method to create the file, from client (nfs) or mkfile (local). Default nfs

B<type> => I<spares> | I<data> | <filersio>

Optional.

Create spare file or a regular file (data) or regular file using filersio, Default data. Filersio will be only triggered for
local protocol and is not available for NFS. Filersio is using synchronous method.

B<fsio_fill_pattern> => integer

Optional.

Pattern, that the blocks should be filled up with. Pattern is a 4 byte integer between 0 and 2^31 - 1. -2 and -3
are the only acceptable negative values for the pattern -2 implies that the block will be filled with the block number
-3 implies that the block will be filled with junk - It provides a faster way to fill up a file. Default is 0xAA

B<fsio_blk_size> => IO size.

Optional.

Block size in bytes or Kb. Can not be larger then 64k. Default is 64k.

B<fsio_rand_percentage> => integer

Optional.

Percentage of I/Os with file offsets determined using uniformly random distribution. Rest of them will be sequential. Default is 0.

B<fsio_num_active> => integer

Optional.

Number of IOs that will be in flight at any given point of time. Default is 10. Please do not use more than 50 or you'll kill
filer's performance.

B<timeout> => integer

Optional

Fill up timeout. Method will fail if write is not completed within given time. Default is 600 seconds (10 minutes). It takes
about 3 hrs per terabyte of data using local protocol and possibly 10 times longer using NFS. It took 2 minutes to fill up
10Gb of data using filersio with default parameters on 21 disks aggregate.

=item CLI parameters

B<DATASET_SNAPSHOT> => I<name>

Name of snapshot to use for predefined dataset. If snapshot is present, then snaprestore will be called instead of actually
running fill. If snapshot is not present, then fill() will run first, then snapshot will be taken for future use.

=item Examples

=back

     $vol->fill(max=>10, proto=>"local");

     $vol->fill(percent=>50, type=>data);

     $vol->fill( proto => 'local', type => 'filersio', max => '10g' );

=cut

sub fill {
      my ( $self, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      my $dataset_snapshot = Tharn::param('DATASET_SNAPSHOT');
      if( $dataset_snapshot && $self->snapshot_exists($dataset_snapshot)) {
          NOTE::logcomment( "Found snapshot with dataset, restoring to snapshot '$dataset_snapshot'");
	  $self->snapshot_restore($dataset_snapshot);
	  return $self;
      }

      my $was_mounted = $self->mounted;
      my $percent;
      my $start_time;
      my $total;
      my $log_interval = 5;                 # seconds
      my $timeout = $args{timeout} || 600;
      my $fname;
      my $maxmegs = $args{max};
      my $proto   = $args{proto} || "nfs";
      my $type    = $args{type} || "data";

      my $fsio_fill = delete $args{fsio_fill_pattern} || 0xAA;  # 10101010
      my $fsio_blsz = delete $args{fsio_blk_size} || '64k';
      my $fsio_rndm = delete $args{fsio_rand_percentage} || 0;
      my $fsio_thrd = delete $args{fsio_num_active} || 10;

      my $total_secs;

      if ( $args{max} =~ /[tgmbk]$/i ) {    # unit specifier?
                                            # NOTE::units returns KB; convert to MB
          $maxmegs = POSIX::ceil( NOTE::units( $args{max} ) / 1024 );
          $maxmegs = 1 if ( $maxmegs == 0 );
      }
      else {
          $maxmegs = $args{max};
      }

      $percent = $args{percent} || 100;

      # The volume will be mounted and permissions changed regardless of the
      # proto. This is done so that any subsequent calls to fill() or refill
      # do not run into "permission denied" issues.
      $self->exportfs("rw,anon=0");
      unless ($was_mounted) {
          my %opts;
          $opts{chmod} = '777';

          # Mount RW-Tree if it is a GX Volume
          $opts{RW} = 1
            if ref($self) =~ /^NOTE::GX/;
          $self->mount(%opts);
      }
      if ( $proto eq 'local' ) {
          my $fname = $self->fillfile( proto => 'local' );

          VFILER::Util::rm_file( $self->Filer->conn, $fname ) unless ref($self) =~ /^NOTE::GX/;

          $total = POSIX::ceil( ( ( ( $percent / 100 ) * $self->total ) - $self->used ) / 1024 );
          if ( defined $maxmegs ) {
              $total = $total > $maxmegs ? $maxmegs : $total;
          }

          if ( $type =~ /data|filersio/ ) {
              NOTE::logcomment( "Writing %d MB to $fname", $total );
              $start_time = time;
              if( ref($self) =~ /^NOTE::GX/ ) {
                  my $cmd = qq("priv set -q diag;filersio asyncio_active 0 -r $fsio_rndm $fsio_blsz 0 ${total}m 60 $fsio_thrd /vol$fname -fillonce $fsio_fill -synchronous -print_stats 30 -create");
                  $cmd = qq("priv set -q diag;mkfile ${total}m /vol$fname") unless $type eq 'filersio';
                  my $out = $self->server->system_node_run( -Timeout => $timeout, node=>$self->Aggregate->Node->name, command=>$cmd);
                  confess $out if $out =~ /filersio: failed to start/s;
              } else {
                  VFILER::Util::mk_file( $self->Filer->conn, $total . 'm', $fname );
              }
              my $filename = File::Basename::basename($fname);
              my $fpath    = $self->mountpoint;
              my $cmd      = "/usr/local/bin/sudo chmod 0666 $fpath/$filename";
              NOTE::logcomment("cmd = $cmd");
              system($cmd);
          }
          elsif ( $type eq 'sparse' ) {
              NOTE::logcomment("Creating $total MB file $fname");
              my ( $path, $name ) = $fname =~ /^(.*)\/([^\/]*$)/;
              my $kilos = $total * 1024;
              $start_time = time;
              eval { Volume::vol_unsparse( $self->Filer->conn, $path, $name, $kilos . 'k' ); };
          }

          # If there was an error, make sure the job is finished by
          # resorting to NFS.
          if ($@) {
              $fname = $self->fillfile( proto => 'nfs' );
              system("/usr/local/bin/sudo chmod 0666 $fname");

              # OK, this is just plain ugly:
              # Unfortunately, if unsparse determines there is not enough room in
              # the volume, it gives up early (like up to 4g early!). So
              # here, I compensate by finishing the job with NFS.
              # When/If unsparse is fixed, this should be removed.
              $args{'proto'} = 'nfs';
              $args{'type'}  = 'data';
              $self->fill(%args);
          }
          $total_secs = time - $start_time;
          $total_secs = 1 if ( $total_secs == 0 );    # Prevent divide-by-zero
          NOTE::logcomment( "Wrote %d MB in %d seconds, %.3f MB/s", $total, $total_secs, $total / $total_secs );
      }
      elsif ( $proto eq 'nfs' ) {
          $total = ( ( ( $percent / 100 ) * $self->total ) - $self->used ) / 1024;
          if ( defined $maxmegs ) {
              $total = $total > $maxmegs ? $maxmegs : $total;
          }

          $fname = $self->fillfile;
          if ( open( F, "+<", $fname ) ) {
              sysseek( F, 0, 2 );    # seek to end
              NOTE::logcomment("Open existing $fname");
          }
          else {
              open( F, ">", $fname ) or confess "Can't open $fname: $!";
              NOTE::logcomment("Open new $fname");
          }

          if ( $total < 0 ) {

              # Need to shrink filefile
              my ( $dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks ) = stat(F);

              my $truncpos = ( ( $size / ( 1024 * 1024 ) ) - abs($total) ) * ( 1024 * 1024 );
              if ( $truncpos < 0 ) {
                  NOTE::logcomment( "Can't reduce " . $self->description );
              }
              else {
                  NOTE::logcomment( "Truncating $fname by %d MB", abs($total) );
                  truncate( F, $truncpos );
              }
              close F;
          }
          else {
              my $megs_to_write = $total;
              my $total_secs;

              if ( $type eq 'data' ) {
                  my $buf = pack( "C", rand(0xff) ) x ( 1024 * 1024 );
                  NOTE::logcomment( "Writing %d MB to $fname", $megs_to_write );
                  $start_time = time;
                  my $next_log_time = $start_time + $log_interval;
                  my $last_megs     = $megs_to_write;
                  while ( $megs_to_write > 0 ) {
                      if ( time >= $next_log_time ) {
                          my $megs_written = $last_megs - $megs_to_write;
                          $next_log_time = time + $log_interval;
                          NOTE::logcomment( "Wrote %d MB, %.3f MB/s, %.3f percent complete", $total - $megs_to_write, $megs_written / $log_interval, ( ( $total - $megs_to_write ) / $total ) * 100 );
                          $last_megs = $megs_to_write;
                      }
                      eval {
                          local $SIG{ALRM} = sub { die "Timeout writing to $fname\n" };
                          alarm $timeout;
                          syswrite( F, $buf );
                          alarm 0;
                      };
                      confess $@ if ($@);
                      $megs_to_write--;
                  }
              }
              elsif ( $type eq 'sparse' ) {
                  NOTE::logcomment("Creating $total MB hole in $fname");
                  $start_time    = time;
                  $megs_to_write = 0;
                  sysseek( F, $total * ( 1024 * 1024 ) - 1, 2 );    # seek to end + total
                  syswrite( F, '\0', 1 );                           # Write one byte
              }
              close F;
              $total_secs = time - $start_time;
              $total_secs = 1 if ( $total_secs == 0 );              # Prevent divide-by-zero
              NOTE::logcomment( "Wrote %d MB in %d seconds, %.3f MB/s", $total - $megs_to_write, $total_secs, $total / $total_secs );
          }
      }

      unless ($was_mounted) {
          $self->umount;
      }

      if( $dataset_snapshot ) {						    # if dataset snapshot is requested, and not yet created
	  NOTE::logcomment("Taking dataset snapshot '$dataset_snapshot'");  # then create requested snapshot. There is no need to check if snapshot already exists
	  $self->snapshot_create($dataset_snapshot);                        # because if it would exists then routine would not even get down here
      }
      return $self;
}

=head2 ::checkfill

Verify that the given file has the amount of data as given by stats
by reading the contents.

=over

=item Arguments

B<fillfh> - I<file handle>

Optional

File handle of the fillfile. Defaults to the internal fillfile

=item Examples

=back

     $vol->checkfill;

=cut

sub checkfill {
      my ( $self, $fillfh ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      my $was_mounted;
      my $fname;
      my $fh;
      my $total;
      my $buf;
      my $timeout = 60;

      if ($fillfh) {
          $fh    = $fillfh;
          $fname = "<fillfile>";
      }
      else {
          $was_mounted = $self->mounted;

          $self->exportfs("rw,nosuid,anon=0");
          unless ($was_mounted) {
              $self->mount( 'chmod' => '777' );
          }
          confess $@ if $@;
          $fname = $self->fillfile;
          open( F, "<$fname" ) or confess "Can't open $fname";
          $fh = *F;
      }

      my $megs = POSIX::ceil( ( stat($fh) )[7] / ( 1024 * 1024 ) );
      NOTE::logcomment( "Verifying %d MB in $fname", $megs );
      my $log_interval    = 5;                             # seconds
      my $total_megs_read = 0;
      my $start_time      = time;
      my $next_log_time   = $start_time + $log_interval;
      my $last_megs       = 0;

      eval {
          local $SIG{ALRM} = sub { die "Timeout reading from $fname\n" };
          alarm $timeout;
          while ( sysread( $fh, $buf, 1024 * 1024 ) ) {
              alarm 0;
              if ( time >= $next_log_time ) {
                  my $megs_read = $total_megs_read - $last_megs;
                  $next_log_time = time + $log_interval;
                  NOTE::logcomment( "Read %d MB; %.3f MB/s, %.3f complete", $total_megs_read, $megs_read / $log_interval, 100 * ( $total_megs_read / $megs ) );
                  $last_megs = $total_megs_read;
              }
              $total_megs_read++;
              alarm $timeout;
          }
          alarm 0;
        };
      confess $@ if ($@);
      my $total_secs = time - $start_time;
      $total_secs = 1 if ( $total_secs == 0 );
      NOTE::logcomment( "Read a total of %d MB at %.3f MB/s in %d seconds from %s", $total_megs_read, $total_megs_read / $total_secs, $total_secs, $fname );

      unless ($fillfh) {
          close F;
          unless ($was_mounted) {
              $self->umount;
          }
      }
      if ( $megs != $total_megs_read ) {
          confess sprintf( "Expected %d MB, but read only %d MB", $megs, $total_megs_read );
      }
      return $self;
}

=head2 ::refill

Recreate the fillfile

=over

=item Arguments

B<fillfh> - I<file handle>

Optional.

File handle of file to recreate. If not supplied the default fillfile is recreated.

=item Examples

=back

     $vol->refill;

     $vol->refill(FH);

=cut

sub refill {
      my ( $self, $fillfh ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      my $was_mounted;
      my $fname;
      my $fh;
      my $total;
      my $timeout = 60;

      if ($fillfh) {
          $total = POSIX::ceil( ( stat($fillfh) )[7] / ( 1024 * 1024 ) );
          my $fd = fileno($fillfh);
          $fname = "<fillfile>";
          open( F, ">", "&" . $fd ) or confess "Can't reopen $fname: $!";
          close($fillfh);
      }
      else {
          $was_mounted = $self->mounted;

          $self->exportfs("rw,nosuid,anon=0");
          unless ($was_mounted) {
              $self->mount( 'chmod' => '777' );
          }
          confess $@ if $@;
          $fname = $self->fillfile;
          open( F, "+<$fname" ) or confess "Can't reopen $fname: $!";
          $total = POSIX::ceil( ( stat(F) )[7] / ( 1024 * 1024 ) );
      }
      $fh = *F;

      my $log_interval  = 5;                                           # seconds
      my $buf           = pack( "C", rand(0xff) ) x ( 1024 * 1024 );
      my $megs_to_write = $total;
      NOTE::logcomment( "Rewriting %d MB to $fname", $megs_to_write );
      my $start_time    = time;
      my $next_log_time = $start_time + $log_interval;
      my $last_megs     = $megs_to_write;
      while ( $megs_to_write > 0 ) {

          if ( time >= $next_log_time ) {
              my $megs_written = $last_megs - $megs_to_write;
              $next_log_time = time + $log_interval;
              NOTE::logcomment( "Rewrote %d MB, %.3f MB/s, %.3f percent complete", $total - $megs_to_write, $megs_written / $log_interval, ( ( $total - $megs_to_write ) / $total ) * 100 );
              $last_megs = $megs_to_write;
          }
          eval {
              local $SIG{ALRM} = sub { die "Timeout writing to $fname\n" };
              alarm $timeout;
              syswrite( F, $buf );
              alarm 0;
          };
          confess $@ if ($@);
          $megs_to_write--;
      }
      close F;
      my $total_secs = time - $start_time;
      $total_secs = 1 if ( $total_secs == 0 );    # Prevent divide-by-zero
      NOTE::logcomment( "Rewrote %d MB in %d seconds, %.3f MB/s", $total - $megs_to_write, $total_secs, $total / $total_secs );
      unless ($was_mounted) {
          $self->umount;
      }

      return $self;
}

=head2 ::unfill

Remove the fillfile from the volume

=over

=item Arguments

None.

=item Example

=back

     $vol->unfill

=cut

sub unfill {
      my ( $self, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      my $avail;

      my $fname = $self->fillfile( proto => 'local' );
      NOTE::logcomment("Deleting $fname");
      VFILER::Util::rm_file( $self->Filer->conn, $fname );
      return $self;
}

=head2 ::fillfile

Sets or gets the fillfile name

=over

=item Argument

B<name> => I<string>

Optional.

Name to set the fillfile to. Default I<__FILL__>

=item Example

=back

     $vol->fillfile;

     $vol->fillfile(name=>"testfile");

=cut

sub fillfile {
      my ( $self, %args ) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      my $newname = $args{'name'};
      my $proto = $args{'proto'} || 'nfs';
      my $fname;

      if ( defined $newname ) {
          $self->{_fillfile} = $newname;
      }
      if ( !$self->{_fillfile} ) {
          $self->{_fillfile} = "__FILL__";
      }
      if ( $proto eq 'nfs' ) {

          # Target must be mounted because otherwise mountpoint is undefined
          if ( $self->mounted ) {
              $fname = $self->{_mntpoint}->{localhost};
              $fname .= "/" . $self->{_fillfile};
          }
          else {
              confess "Tried to access fillfile of " . $self->description . " when not mounted";
          }
      }
      elsif ( $proto eq 'local' ) {
          $fname = $self->mount_path . "/" . $self->{_fillfile},;
      }
      return $fname;
}

=head2 ::sio_proc

Returns the SIO process if one is running for this object else returns undef

=over

=item Example

=back

    my $sio = $vol->sio_proc;

=cut

sub sio_proc {
      my ($self) = @_;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      if ( $self->{_sio} ) {
          return $self->{_sio};
      }
      else {
          return undef;
      }
}

=head2 ::sio

Provide SIO support. See:
http://perf-web.corp.netapp.com/twiki/bin/view/Perfweb/SIO
for full details of SIO functionality.

The SIO process is continuously 'churned' until terminated by call with 'stop' set
or automatically at Teardown.

The SIO program is expected to be located at /usr/local/test/bin/sio

Returns the SIO object is successful, or 1 if stopping SIO

=over

=item Arguments

B<client> => I<hostspec>

Client to run SIO from. Default localhost or set by a previous call.
Client must be a non-Windows client.

B<read> => I<percent>

Percent of read access vs write access. Default 100

B<rand> => I<percent>

Percent of random access vs sequential access. Default 100

B<blksize> => I<IO block size>

Size of IO. Default 5k

B<start_byte> => I<integer>


Lower bound for access. Default 0

B<file_size> => I<size>

File size of each file used by sio. Default 20m

B<numthreads> => I<integer>

Number of IO generators. Default 1

B<count> => I<integer>

If specified determines number of files to create for sio to use. Default 5

B<file_list> => I<file list>

List of space separated files for SIO to access.
If not specified, <count> files of <file_size> size will be created for sio to use.
This override <count> option.

B<cycle> => I<integer>

How long to run sio before repeating, in seconds. A special value of zero indicates
only create data files and not to start sio process. Default 60

B<stop> => I<1|0>

Stop the SIO process if set. Default 0

B<restart> => I<integer>

Flag to indicate to start sio with parameters from previous invokation.

B<RW} => I<1|0|undef>

If set indicates that this is a read/write volume
and the boolean indicate whether to mount RW-Tree or R-Tree.
If not set remount test will be carried out to determine R/W or R/O
volume, with the mount to the R-Tree
Default: undef

B<suffix> => I<string>

I<Optional>

Suffix to SIO file names. Default: <PID> of current process

=item Examples

=back

     # Default:  run 'read' sio on localhost against 5 files of 20m each
     $vol->sio();

     # Create file for SIO to use then return without starting sio
     $vol->sio(count=>2, cycle=>0);

     # create 3 files of 10m each for sio to use with 2 threads
     $vol->sio(numthreads=>2, count=>3, file_size=>"10m");

     # Tell sio to use listed files on client nam9
     $vol->sio(client=>"nam9", file_list=>"f1 f2");

     # Perform sio on src (25% read, 75% write) and dst (100% read) of a snapmirror relationship
     $src_vol->sio(read=>25);
     $sm->initialize(src=>$src_vol, dst=>$dst_vol);
     $dst_vol->sio();

     # Stop sio
     $vol->sio(stop=>1);

=cut

sub sio {
      my $self = shift;
      my (%args) = validate( @_, {
              suffix => { type => SCALAR, optional => 1 },
              cycle  => { type => SCALAR, optional => 1 },
              client => { type => SCALAR, optional => 1 },
              read   => {
                  type     => SCALAR,
                  optional => 1,
                  default  => 100,
                  regex    => qr/\d/
              },
              rand => {
                  type     => SCALAR,
                  optional => 1,
                  default  => 100,
                  regex    => qr/\d/
              },
              blksize    => { type => SCALAR, optional => 1, default => "50k" },
              start_byte => { type => SCALAR, optional => 1, default => 0, regex => qr/\d/ },
              count      => { type => SCALAR, optional => 1, default => 5, regex => qr/\d/ },
              file_size  => { type => SCALAR, optional => 1 },
              numthreads => { type => SCALAR, optional => 1, default => 1, regex => qr/\d/ },
              file_list  => { type => SCALAR, optional => 1 },
              RW      => { type => SCALAR, optional => 1, default => undef },
              stop    => { type => SCALAR, optional => 1, default => 0 },
              restart => { type => SCALAR, optional => 1, default => 0 }

      } );
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      confess "sio 'stop' and 'restart' options are mutually exclusive"
        if $args{stop} && $args{restart};

      $args{client} = $args{client} || $self->{_client} || "localhost";
      confess "SIO on Windows not supported..."
        if Tharn::host( $args{client} )->hosttype =~ /windows/i;

      if ( $args{stop} ) {
          if ( $self->{_sio} ) {
              NOTE::logcomment( "Stopping SIO for " . $self->description );

              # Create stop file to indicate to sio process to terminate immediately rather then wait till end of cycle
              _sio_stop_file( $self, "create", $args{client} );

              # Give SIO subtest a chance to exit
              my $retry = 0;
              my $pause = 5;
              do {
                  $retry++;
                  Tharn::snooze($pause);
              } while $self->{_sio}->is_alive() && ( $retry * $pause ) < 300;    # Wait for upto 5mins for SIO to exit
              if ( $self->{_sio}->is_alive() ) {

                  # Blindly kill process if stop file does not work
                  NOTE::logcomment("Killing SIO...");
                  $self->{_sio}->raise("TERM");

          Tharn::snooze($pause);                    # this is fix for
          $self->{_sio}->raise('KILL') if $self->{_sio}->is_alive();    # BURT367205

              }
              $self->{_sio}->wait_finish( timeout => 900 );
              FSR::FSRutils::subtestResults( $self->{_sio} ) or confess $@;
              delete $self->{_sio};
              _sio_stop_file( $self, "remove", $args{client} );
              NOTE::logcomment("SIO process terminated");
          }
          else {
              confess "No SIO process running";
          }
          return 1;
      }

      confess("sio already running...") if $self->{_sio};

      $args{suffix} ||= $$;
      $args{file_size} = ( $self->Filer->is_a_simulator ? "2" : "20" ) . "m"
        unless defined $args{file_size};
      confess "numthreads must be > 0"
        unless $args{numthreads} > 0;

      if ( $args{restart} && exists $self->{_sio_args} ) {
          NOTE::logcomment("Restore previous sio parameters");
          delete $args{restart};
          foreach my $key ( keys %{ $self->{_sio_args} } ) {
              $args{$key} = $self->{_sio_args}{$key};
          }
      }
      %{ $self->{_sio_args} } = %args;

      $args{file_size} .= "m" unless $args{file_size} =~ /m$/i;

      my $rw_mnt = $args{RW} || 0;
      my $fname = $args{file_list};
      unless ($fname) {

          # XXX 'fill' method works on localhost only
          $rw_mnt = _rw_mount_check( $self, "localhost", $rw_mnt );
          my $idx = 1;
          do {
              my $sio_file = "sio_" . $idx . "_" . $args{suffix};
              my $fullpath = $self->mountpoint . "/" . $sio_file;
              if ( $rw_mnt && !-e $fullpath ) {

                  # Create file if R/W mount and file does not exists
                  $self->fillfile( name => $sio_file );
                  $self->fill( proto => 'nfs', max => $args{file_size} );
              }
              confess "Unable to locate sio data file $sio_file"
                unless -e $fullpath;
              $fname .= " " if $fname;
              $fname .= $sio_file;    # Add to sio 'dev' list
          } while $idx++ < $args{count};
          $self->umount( client => "localhost" )
            if defined $self->{_client} && $self->{_client} !~ $args{client};
      }
      $self->{_sio_args}->{file_list} = $fname;

      return 1
        if defined $args{cycle}
            && $args{cycle} <= 0;     # Create and return data files only, do not start sio

      # Mount for SIO access
      $rw_mnt = _rw_mount_check( $self, $args{client}, $rw_mnt );
      confess "Invalid write traffic request for readonly volume " . $self->description
        if $args{read} == 0 && !$rw_mnt;

      # XXX SIO -stop file option only works for linux host, so use short 60 cycle
      #     unless this is the case
      my $stop_file = _sio_stop_file( $self, "set", $args{client} );
      my $stop_cmd = "-stop -" . $stop_file;
      $args{cycle} = 900
        unless defined $args{cycle};

      my $cmd = "rsh "
        . $args{client}
        . " \"cd "
        . $self->mountpoint
        . "; /usr/software/test/bin/sio"
        . " $args{read} $args{rand} $args{blksize} $args{start_byte}"
        . " $args{file_size} $args{cycle} $args{numthreads} $fname $stop_cmd" . "\"";
      my $ret = Tharn::subtest( sub {
              my ( $rc, $run ) = ( 0, 1 );
              $SIG{TERM} = sub {
                  NOTE::logcomment( "Stopping " . $self->description . " SIO" );
                  $run = 0;
              };

              close STDIN;

              # continuous SIO loop until stopped by parent
              NOTE::logcomment( "Starting " . $self->description . " SIO" );
              my $error_file = "/var/tmp/sio_" . $self->name . ".$$";
              my $cycle      = 1;
              do {
                  NOTE::logcomment( "Running SIO Cycle " . $cycle++ . ": $cmd" . " (" . $self->tag . ")" );
                  $rc = system( $cmd. " 2>$error_file" );
                  if ($rc) {
                      my $text;
                      open F, "<$error_file";
                      while ( $text .= <F> ) { }
                      close F;
                      unlink $error_file;
                      NOTE::logcomment( "SIO for " . $self->description . ", RC: $rc. Error: $text" );
                  }
              } while $run && !( -e $stop_file || $rc );
              unlink $error_file;
              Tharn::logresult( "FATAL", "Attempt at " . $self->description . " SIO failed: rc = $rc" )
                if $rc > 0;
              NOTE::logcomment( "SIO subtest for " . $self->description . " exiting" );
      }, '-bg', '-onexit', sub { } );
      confess "Fail to start SIO: $@"
        if $@;
      $self->{_sio} = $ret;

      # Schedule SIO to be stopped during teardown:
      NOTE::reference( sub { $self->sio( stop => 1 ) if exists $self->{_sio} } );
      return $ret;
  }

  sub _rw_mount_check {
        my ( $self, $client, $rw_flag ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        if ($rw_flag) {
            $self->mount( client => $client, exportfs => "rw,nosuid,anon=0", chmod => "777", RW => $rw_flag );
            return $rw_flag;
        }
        NOTE::logcomment("Remount to determine r/w access to Volume/Qtree");
        $self->umount( client => $client ) if $self->mounted( client => $client );
        my $exp = ",nosuid,anon=0";
        $self->{_client} = $client;
        eval {

            # Attempt r/w mount
            $self->mount(
                client   => $client,
                exportfs => "rw" . $exp,
                chmod    => "777"
            );
          };
        if ( $self->mounted( client => $client ) ) {
            return 1;    # to indicate r/w mount
        }
        else {
            $self->mount( client => $client, exportfs => "ro" . $exp );
            return 0;    # to indicate r/o mount
        }
  }

  sub _sio_stop_file {
        my ( $self, $mode, $client ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        if ( $mode =~ /set/i ) {
            $self->{_sio_stop_file} = "/var/tmp/" . $self->tag . "_sio_stop_" . $$;
            $self->{_sio_stop_file} =~ s/ /_/g;
        }
        else {
            return undef unless defined $self->{_sio_stop_file};

            my $fullpath = $self->{_sio_stop_file};

            my $cmd = "rsh " . $client . " " . ( $mode =~ /create/i ? "touch" : "rm" ) . " $fullpath";
            NOTE::logcomment("SIO stop file cmd: $cmd");
            if ( $mode =~ /create/i ) {

                # Don't abort if stop file creation failed
                system($cmd);
                NOTE::logcomment( "SIO stop file " . $self->{_sio_stop_file} . " created" )
                  if -e $fullpath;
            }

            if ( $mode =~ /remove/i && -e $fullpath ) {

                # Attempt to remove stop file if exists
                if ( -e $fullpath ) {
                    system($cmd);
                    confess "Failed to removed stop file " . $self->{_sio_stop_file}
                      if -e $fullpath;
                }
                delete $self->{_sio_stop_file};    # No need to save stop after
            }
        }

        return $self->{_sio_stop_file};
  }

=head2 ::dataset

Name of the dataset of this object

=over

=item Arguments

None.

=item Example

=back

     $vol->dataset;

=cut

  sub dataset {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_dataset}
}

=head2 ::pwd

Returns the current working directory for the calling process.

=over

=item Arguments

None.

=item Example

=back

     $curdir = $vol->pwd;

=cut

  sub pwd {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $path = shift->{_wdir};
    return $path ? $path : '/'
}

=head2 ::cd

Sets the current working directory for the calling process.

=over

=item Arguments

I<string>

I<Required.>
A path name.

=item Example

=back

     $vol->cd( '../testdir2' );

=cut

  sub cd {
        my ( $self, $newdir ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        # Turn newdir into absolute path
        $newdir = '/' unless ($newdir);
        unless ( $newdir =~ /^\// ) {
            $newdir = $self->pwd . '/' . $newdir;
        }
        unless ( $newdir =~ /\/$/ ) {
            $newdir .= '/';
        }

        # Collapse multiple '/'s to one '/'
        $newdir =~ s/[\/]{2,}//g;

        NOTE::logcomment( "Changing working directory to $newdir for " . $self->description );
        $newdir = undef if ( $newdir eq '/' );
        $self->{_wdir} = $newdir;
        return $self;
  }

=head2 B<Waiting>

=head2 ::wait

Waits for all background processes to complete.
These are the processes created by using the B<background> option of
L<C<::populate>|/::populate>, L<C<::modify>|/::modify>,
L<C<::compare>|/::compare>, or L<C<::delete>|/::delete>.

=over

=item Arguments

None.

=item Example

=back

     $vol->wait;

=cut

  sub wait {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        my ( $obj, $test, $op );

        while ( $test = pop @{ $self->{_subtests} } ) {
            $op  = $test->{operation};
            $obj = $test->{subtest};
            if ( $obj->is_alive() ) {
                NOTE::logcomment( "Waiting for: " . $op );
            }
            $obj->wait_finish();
            FSR::FSRutils::subtestResults($obj) or confess $@;
        }
        return $self;
  }

=head2 B<Related Objects>

=cut

=head2 ::Filer

Returns an object representing the filer containing the volume.

=over

=item Arguments

None.

=item Example

=back

     $filer = $vol->Filer;

=cut

  sub Filer {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        return $self->{_Filer} if ( $self->{_Filer} );
        return $self->{_Aggregate}->Filer if ( $self->{_Aggregate} );
        return undef;
  }

=head2 ::Aggregate

Returns an object representing the aggregate
containing the volume.

=over

=item Arguments

None.

=item Example

=back

     $aggr = $vol->Aggregate;

=cut

  sub Aggregate {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        if ( $self->type ne 'trad' ) {
            return $self->{_Aggregate};
        }
        else {
            confess "This traditional volume does not have an aggregate";
        }
  }

=head2 ::container

A synonym for L<C<::Aggregate>|/::Aggregate>.

=cut

  sub container {
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
        return shift->Aggregate;
  }

=head2 B<Object Identification>

=cut

=head2 ::tag

A tag is a text description for a filer entity.
It is only stored in the object referring to the entity.
The purpose of the tag is to improve readability of the log.
If an object is tagged,
NOTE will include the tag in comments it logs about the object.

Returns the tag for the object.

=over

=item Arguments

I<string>

I<Optional.>
A tag to store in the object.

=item Examples

=back

     $vol->tag( 'source' );
     $which = $vol->tag;

=cut

  sub tag {
        my ( $self, $newtag ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        if ( defined $newtag ) {
            $self->{_tag} = $newtag;
        }
        return $self->{_tag};
  }

=head2 ::description

Returns a description of the volume consisting of
its filer's name, its mount path, and its tag, if any.

=over

=item Arguments

None.

=item Example

=back

     $desc = $vol->description;

=cut

  sub description {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        my $txt;

        $txt = $self->hostpath;
        if ( $self->{_tag} ) {
            $txt .= " (" . $self->{_tag} . ")";
        }
        return $txt;
  }

=head2 ::is_destroyed

Returns true if this volume is destroyed.  This can also be
assigned to, in order to make a volume 'un-destroyed',
but this feature is intended for internal use only.

=over

=item Argument

None.

=item Example

=back

     if ( $vol->is_destroyed ) {
         ...
     }

     ...

     $vol->is_destroyed = 0;

=cut

  sub is_destroyed : lvalue {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        # Must do this trickery due to a bug in perl regarding
        # tied hashes and lvalues.
        my $r = \$self->{_destroyed};
        $$r;
  }

=head2 ::is_recycled

Returns a true if a volume was recycled
and false if it was physically created.

=over

=item Arguments

None.

=item Examples

=back

      my $vol = NOTE::Volume->create( recycle => 1 );
      if ( $vol->is_recycled ) {
          printf( "This volume was recycled\n" );
      } else {
          printf( "This volume was created\n" );
      }

=cut

  sub is_recycled {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_was_recycled}
}

=head2 ::is_recyclable

Returns 1 if the volume is a recyclable volume.
This is also assignable, if it is desired to change the
status.


=over

=item Arguments

None.

=item Examples

    if ( $vol->is_recyclable ) {
    printf( "Volume %s is recyclable\n", $vol->name );
    }

    ...

    $vol->is_recycleable = 1;

=back

=cut

sub is_recyclable : lvalue {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->{_recycle};
}

=head2 ::is_free

Returns a boolean to indicate whether this volume is in the free pool

=over

=item Arguments

None.

=item Eample

=back

    for my $vol ( $filer->Volume ) {
    next unless ( $vol->is_free );
    printf( "Volume %s is available for reuse\n", $vol->name );
    }

=cut

sub is_free {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name} =~ /^_vl/
}

=head2 ::is_anonymous

Returns true if this volume is anonymous.  This can also be
assigned to, in order to make a volume anonymous.

=over

=item Argument

None.

=item Example

=back

     if ( $vol->is_anonymous ) {
         ...
     }

     ...

     $vol->is_anonymous = 1;

=cut

sub is_anonymous : lvalue {
   my $self = shift;
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

   # Must do this trickery due to a bug in perl regarding
   # tied hashes and lvalues.
   my $r = \$self->{_anonymous};
   $$r;
}

=head2 ::name

Returns name of the volume as string.

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name}
}

=head2 ::newName

Returns a new name

=over

=item Argument

None.

=item Example

=back

      $vol->newName;

=cut

sub newName {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->SUPER::newName( $_[0] || "vl" )
}

=head2 ::SPARE

Spare Volume from recycling

=over

=item Argument

None.

=item Example

=back

     $vol->SPARE

=cut

  sub SPARE {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        $self->{_recycle}  = 0;
        $self->{_spare_me} = 1;
        if ( $self->{_type} ne "trad" ) {
            $self->Aggregate->SPARE;
        }
        return $self;
  }

=head2 ::nfs

Returns an NFS object for this volume

=over

=item Arguments

None.

=item Example

=back

     $nfs = $vol->nfs;

=cut

  sub nfs {
        my ( $self, %args ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        confess "Invalid parameter 'client'"
          if exists $args{client};
        $self->{_NFS} ||= NOTE::FileSystem::NFS::File->init( $self, '/', %args );

        # The following makes it possible to unmount the filesystem..
        # Subsequent calls to 'nfs' cause it to be remounted again..
        # $qt->nfs->ls( '-lR' );
        # $qt->umount;
        # $qt->nfs->ls( '-lR' );

        $self->mount(
            client => "localhost",
            %args
          )
          unless ( $self->mounted( client => "localhost" ) );
        return $self->{_NFS};
  }

=head2 ::clean

If a __clean__ snapshot exists, the volume will be restored
using this snapshot.

Croaks on error.
Returns $self.

=over

=item Arguments

=item Examples

=back

     $vol->clean

=cut

  sub clean {
        my ( $self, %args ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        unless ( $self->is_anonymous ) {
            confess "Cannot clean a non-anonymous volume: " . $self->description;
        }
        $self->snapshot_restore("__clean__");
        return $self;
  }

=head2 ::release

Release a volume when it is no longer needed by a script.
This means put it in the recyclable pool or destroy
it, as appropriate.

=over

=item Argument

None.

=item Example

=back

  # The following will recycle a new volume each time
  # through the loop, although it is likely to get the
  # same volume each time.
  for ( 0..10000 ) {
      my $trad = NOTE::Volume->create( type => 'trad' );
      $trad->fill( max => '500m' );
      $trad->release;
  }

=cut

  sub release {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        $self->_release;
        $self->unreference;    # Remove the volume's reference from teardown stack
        return $self;
  }

=head2 ::destroy

Destroys the volume and renders the volume object unuseable.
If the volume is a traditional volume
L<C<NOTE::Filer::zero_spares>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/::zero_spares>
is called.

=over

=item Arguments

B<force> => C<1>

I<Optional.>
A flag to force the destruction of a volume
even if it is corrupt.

=item Example

=back

     $vol->destroy;

     $vol->destroy( force => 1 );

=cut

  sub destroy {
        my ( $self, %args ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        $args{force} = ( $args{force} ? "true" : "false" )
          if exists $args{force};

        NOTE::logcomment( "Destroying Volume " . $self->description );
        $self->_vol_destroy(%args);
        $self->is_destroyed = 1;
        return $self;
  }

=head2 ::purge

Explicitly offline and destroy a volume

=over

=item Arguments

Same arguments as ::destroy.

=item Example

    $vol->purge;

=back

=cut

  sub purge {
        my ( $self, %args ) = @_;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        NOTE::logcomment( "Purging Volume " . $self->description );
        eval { $self->offline };
        $self->destroy(%args);
        return $self;
  }

=head2 ::uuid

Get the UUID of this volume

Returns:
The string representing the UUID of the volume extracted from 'volume show'.

=over

=item Arguments
None

=item Examples

    if ($srcvol->uuid ne $srcvol->get_instance_UUID) {
        # do whatever if the uuid and instance uuid of src
        # volume are not the same
    }

=back

=cut

  sub uuid {
      my $self = shift;
      my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

      return $self->status('uuid');
  }

  #
  # This is an internal function that does all that is necessary
  # to free a volume when it is no longer needed by a script.
  #
  sub _release {
        my $self = shift;
        my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        return if ( $$ != $self->{_owner} );
        return if ( !$self->{_anonymous} );
        return unless ( $self->substant );
        $self->is_destroyed = 1;
        return if ( !$self->{_anonymous} );
        return if exists $self->{_parent};

        # Code below only applies to anonymous volumes:
        foreach ( keys %{ $self->{_mntpoint} } ) {
            $self->umount( client => $_ );
        }

        # Remove it's junction for 10 mode volume
        $self->remove_junction
          if ref($self) =~ /^NOTE::GX/;

        # Abandon recycling since volume could not be use for future SM tests
        my $epoch = $self->status("striping-epoch");
        $self->{_recycle} = 0 if $epoch > 1;

        if ( NOTE->RECYCLE && $self->{_recycle} ) {

            # Don't bother recycling anonymous flex/striped volumes that are empty
            unless ( $self->type =~ /flex|striped/ && !$self->dataset ) {
                eval { $self->unexportfs };    # Don't leave recyclable volumes in exports list

                # Bring it online, in case it's offline.
                eval { $self->online };
                if ( $@ && $@ =~ /was left in an inconsistent/ ) {
                    NOTE::logcomment( "Volume " . $self->{_name} . " cannot be recycled" );
                    eval { $self->offline };    # Might already be offline
                    $self->destroy;             # No choice but to destroy the volume.
                    return;
                }

                # Only non-anonymous volumes are preserved and recycled.
                # If the restore fails, let the volume be destroyed.
                eval { $self->clean };
                unless ($@) {
                    NOTE::logcomment( "Returning Volume " . $self->{_name} . " to free pool" );
                    if ( $self->_looksAnonymous ) {
                        my $name = (
                            exists $self->{_dataset}
                            ? NOTE::datasetEncode( $self->{_name}, $self->{_dataset} )
                            : $self->{_name}
                        );
                        $self->_rename( $self->_make_free($name) );
                    }
                    else {
                        $self->_rename( $self->_make_free( $self->newName ) );
                    }
                    return;
                }

                # restore using __clean__ failed, so just destroy the volume
            }
        }
        eval { $self->offline };    # Don't care if this fails
        $self->destroy;
  }

  sub DESTROY {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return if ( $$ != $self->{_owner} );
    return if ( !$self->substant );

    unless( NOTE->DONT_SPARE_VOLUMES ) {
        return if ( NOTE->ALL_SPARED );

        if ( $self->{_spare_me} ) {
        NOTE::logcomment( 'Volume ' . $self->{_name} . ' has been spared' );
        return;
        }
    } else {
        NOTE::logcomment( 'Volume ' . $self->{_name} . ' will be destroyed due to NOTE_DONT_SPARE_VOLUMES flag' );
    }

        $self->release;
  }

###############################################################################
#
#   Sub Name:       _get_volume_name
#
#   Description:    get the volume name for this volume. this method is needed
#                   to support quotas in both volumes and qtrees. a method of
#                   the same name exists in Qtree.pm. this allows Quota.pm
#                   to get its volume name without making any special checks.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    volume name for this qtree (string)
#                   Failure:    not an option
#
###############################################################################

sub _get_volume_name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->name();
}

###############################################################################
#
#   Sub Name:       _get_qtree_name
#
#   Description:    get the qtree name for this volume. this method is needed
#                   to support quotas in both volumes and qtrees. a method of
#                   the same name exists in Qtree.pm. this allows Quota.pm
#                   to get its qtree name without making any special checks.
#                   for the purpose of quotas if a volume object is being used
#                   no qtree is needed. in that case the quota command expects
#                   a null value. the filer has defined the null qtree name
#                   to be "" (two double quotes).
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    qtree name for this qtree (string)
#                   Failure:    not an option
#
###############################################################################

sub _get_qtree_name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    # return the string expected by the filer for a null qtree name
    return q{""};
}

1;
