#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/ontap/main/test/lib/GenericMount.pm#4 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
package GenericMount;

## @summary
##   Provide an OS independent interface to mount and umount. The mounts
##   may be local or remote, and are idempotent.
##
## @description
##   The library "mount" subroutine attempts to perform a mount using
##   an OS specific mount command on the given mountpoint regardless of
##   the previous state of the mountpoint. In particular, it will create
##   the mount point if it does not already exist, will succeed without
##   doing the actual mount if the mountpoint is already mounted. It will
##   intelligently process mount failures to determine if the mount should
##   be retried or not.
##
##   The library "umount" subroutine invokes the OS umount and interprets
##   its results.
##
##   Sharing of Mountpoints
##
##   The sharing option allows callers to specify if they want the mount
##   to be exclusive access where another mount call for the exact same
##   mount will fail, or shared access (default) where independent tests can
##   share mountpoints. Mount chooses a mountpoint based on $server, $path,
##   $sharing, and OS specific mount parameters. The sharing type of the
##   mount can immediately be determined from the mount path. Callers
##   should generally allow mount to choose the mountpoint value.
##
##   There is no protection mechanism for shared mountpoints. One test
##   could successfully umount the mountpoint while another test
##   is using it. One test could try to umount a mountpoint that
##   another test has as its working directory so that the umount
##   would return an error that the mountpoint is busy.
##
##   The best way to share mountpoints between tests is to have one test
##   script do the mount and then run the test thpl scripts and finally
##   run a thpl script that would umount the mountpoint.
##
##   Detecting Existing Mounts
##
##   The mount routine succeeds iff the chosen mountpoint is mounted,
##   regardless of mount command output.  This mount function detects
##   mountpoints by examining the C<st_dev> fields in stat of the
##   mountpoint and its parent.
##
##   Retrying Mounts
##
##   The OS specific mount subroutine will look at the error returned
##   from the mount command and decide if the mount should be retried.
##
##   Mount Command Specific Options
##
##   The generic mount options documented for the mount library routine
##   defined below are derived from the Solaris 10 mount command.
##   It is up to the OS specific mount class to map
##   the Solaris 10 option to the OS dependent value of the flag.
##   Only the generic defaults and the options specified by the
##   caller are passed to the OS dependent mount method.
##
##   For the defaults, the OS dependent mount method must set or unset
##   parameters so that the default options are correctly set on the
##   mountpoint. This is done to ensure that default mount library
##   routine behavior is consistent irregardless of the underlying OS.
##
##   For caller specified options, the OS dependent method will attempt to
##   map the options to OS dependent options if the option is known to
##   it. An error is returned if an option is specified that is not
##   supported by the underlying operating system.
##
##   Note that option validation is left to the OS mount command.
##   Otherwise a bunch of code would have to be written to validate
##   options and this code would have to be modified every time the OS
##   mount command changed option processing. Instead a minimalistic
##   approach is taken where only the OS dependent mount classes have to
##   be updated and that is only if options have to be mapped.
##
##   A mount that consistently times out usually indicates that there
##   is some type of error occurring during the mount and that the
##   output may be going to the system log. Try the mount manually
##   with the exact same parameters to determine what is going on.
##
## @usage
##   use GenericMount;
##
## @status  review
##
## @author  dl-maryv (FSS Protocols QA)
## @notes   Created by EP Network Storage Performance Lab, Inc.
##
## @notes   Differences between GenericMount.pm and mount.pm
##
##  1) The "type" parameter is no longer used,
##  2) Changed default timeout for mount from 60 secs to 30 secs.
##  3) Instead of blindly retrying the mount, look at the error and
##     figure out if the mount could succeed, if so retry it.
##  4) Written for Unix-type OSs.

use strict;
use warnings;
use Scalar::Util 'blessed';

use NACL::C::Client;
use NACL::C::Client::Mount;
use NACL::STask::Client::Mount;
use Carp qw{croak};
use Module::Common qw(validate_mixed);
use Params::Validate qw(OBJECT SCALAR UNDEF);
use Tharn qw{$Log};

use Net::IP qw{ip_is_ipv6 ip_is_ipv4};
use Data::Dumper qw{Dumper};

#-------------------------------------------------------
# Global variables
#-------------------------------------------------------
use vars qw(
  $VERSION
  @ISA
  @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  &mount
  &umount
);

#-------------------------------------------------------
# Package global variables
#-------------------------------------------------------
my $VERSION = '1.00';

## Internal function to fix the client type.
## Input is a client name/Hostrec/Connrec.
## Output is a key-value pair.
sub _fix_client($) {
    # What exactly is the client ? a simple hostname or some hostrec or connrec ?
    my $client = shift;
    if (!$client) {
        return ('name', 'localhost');
    }
    if ( blessed($client) ) {
        if ( $client->isa("Hostrec") ) {
            return ('hostrec', $client);
        } elsif ( $client->isa("Connectrec") ) {
            return ('hostrec', $client->hostp());
        } else {
            croak(qq{Unrecognized client type given:$client});
        }
    } else {
        return ('name', $client);
    }

}

## @name ::mount
## @description
##  Contains the generic front end to the OS dependent mount command.
##  It handles the following:
##   1) Processes options that are not passed to the OS dependent mount.
##   2) Add default options to the list of options but do not override user
##      settings of these parameters.
##   3) Construct mount class that has the OS specific info.
##   4) Call the OS specific mount with generic options.
##
##  If you want to do a mount at the very least a path parameter is required.
##  If it is called with path set to undef then it just returns the
##  output from running the OS specific mount command.
##
##  All the arguments are options, passed in "key => $value" style:
##
## @synopsis
##  $mountpoint = mount(server, path, chmod => "0755", mntopts => "noac,proto=udp");
##
## @arg
##  server  Required: The name or IP address of a file server to
##          mount from when doing a remote mount. Set to undef if
##          not doing a remote mount.
##
##  path    Required: For remote mounts $path is the export point
##          on $server to mount. For local mounts $path could be the
##          name of the block device to mount or a directory if
##          doing a loopback mount (assuming the client OS supports
##          loopback mounts). Set to undef if you just want the
##          output of the mount command.
##
##  chmod   Optional: The mode (permissions) to set on the mountpoint
##      after mounting.  The chmod is only performed if the
##      mountpoint was not already mounted. The mode should be an
##      octal number inside a string.
##
##  client  Optional: Hostspec or Connectrec of a client that
##      should perform the mount.  A hostspec (such as "$CLIENT")
##      is preferred, in which case the connection "#rsh" will be
##      used.  The default is "localhost", which means to perform
##      the mount locally.  If a Connectrec is given it must be for
##      a "per-say" type of connection i.e. one which effectively
##      connects/disconnects for every command.  (Thus, for example,
##      the Connectrec returned by connect("#rlogin") or
##      connect("#telnet") could not be used).
##
##  fstype  Optional: The file system type to pass to the "mount"
##      executable. The default is not to pass any file system name.
##
##  mntopts Optional: A string that contains a comma separated list of
##      options for the OS dependent mount command.
##
##  mountpoint  Optional: The directory to mount this file system onto. The
##      default is "/t/server/path/{exclusive|shared}/extra"
##      where "server" is the given "$server" value, and "path" is a
##      string generated from the "$path", followed by the sharing
##      type of the mount, and then OS dependent options used to
##      ensure that the mountpoint is unique. As described in
##      "Sharing of Mountpoints", above, one should generally let
##      "mount" choose the value of this option or else do at least
##      as good a job of avoiding ever having separate tests choose
##      identical mountpoint paths but non-identical mount options.
##      If a mountpoint is specified it should not contain any
##      of the sharing options {shared|exclusive} and the
##      sharing option must be set to exclusive.
##
##  mountretry  Optional: Specifies the number of times to retry the
##      mount operation. The default is 1. Set this value to 0
##      if you do not want to retry the mount.
##
##  sharing Optional: The sharing option allows you to specify if you
##      want the mount to be "exclusive" access where another mount
##      call for the exact same mount will fail, or "shared" access
##      (default) where tests can share mountpoints. If the
##      mountpoint is specified then the sharing option will default
##      to exclusive. The best way to share mountpoints between
##      tests is to have one thpl script do the mount and then run
##      the test thpl scripts and finally run a thpl script that
##      would umount the mountpoint.
##
##  timeout Optional: How many seconds to wait for a mount or umount
##      attempt to finish.  The default is 30 seconds.  "mount" may
##      take up to mountretry of these periods before it completely
##              fails. This options is used to set the timeout on the say
##      command.
##
##  verbose Optional: Be more verbose in the mount execution. This is a
##      pass-through option to NACL. The default is 0(not verbose)
##
##  NFS_MOUNT_OPTIONS: The following are NFS options. These options
##      should be set in the mntopts parameter. The defaults
##      are 'hard,intr,noquota,vers=3'.
##
##  acdirmax=n
##      Hold cached attributes for no more  than  n  seconds
##      after directory update. The default value is 60.
##
##  acdirmin=n
##      Hold cached attributes for at least n seconds  after
##      directory update. The default value is 30.
##
##  acregmax=n
##      Hold cached attributes for no more  than  n  seconds
##      after file modification. The default value is 60.
##
##  acregmin=n
##      Hold cached attributes for at least n seconds  after
##      file modification. The default value is 3.
##
##  actimeo=n
##      Set min and max times for regular files  and  directories
##      to n seconds.
##
##  bg|fg
##      If the first attempt fails, retry in the background,
##      or, in the foreground. The default is fg.
##
##  forcedirectio|noforcedirectio
##      If forcedirectio is specified, then for the duration
##      of  the  mount,  forced  direct  I/O is used. If the
##      file system is mounted using forcedirectio,  data  is
##      transferred directly between client and server, with
##      no buffering on the client.  If  the  file system  is
##      mounted  using  noforcedirectio, data is buffered on
##      the client. forcedirectio is  a  performance  option
##      that  is  of  benefit  only in large sequential data
##      transfers. The default behavior is noforcedirectio.
##
##  grpid
##      By default, the GID associated with a newly  created
##      file  obeys the System V semantics; that is, the GID
##      is set to the effective GID of the calling  process.
##      This  behavior  can be overridden on a per-directory
##      basis by setting  the  set-GID  bit  of  the  parent
##      directory;  in this case, the GID of a newly created
##      file is set to the GID of the parent directory  (see
##      open(2) and mkdir(2)). Files created on file systems
##      that are mounted with the  grpid  option  obeys  BSD
##      semantics  independent of whether the set-GID bit of
##      the parent directory is set; that  is,  the  GID  is
##      unconditionally  inherited  from  that of the parent
##      directory.
##
##  hard|soft
##      Continue to retry requests until the server responds
##      (hard)  or  give  up and return an error (soft). The
##      default value is hard.
##
##  intr|nointr
##      Allow (do not allow) keyboard interrupts to  kill  a
##      process that is hung while waiting for a response on
##      a hard-mounted file system.  The  default  is  intr,
##      which  makes  it  possible  for clients to interrupt
##      applications that can be waiting for a remote mount.
##
##  noac
##      Suppress data and attribute caching. The data  caching
##      that  is  suppressed  is  the write-behind. The
##      local page  cache  is  still  maintained,  but  data
##      copied into it is immediately written to the server.
##
##  nocto
##      Do not perform the normal close-to-open consistency.
##      When  a file is closed, all modified data associated
##      with the file is flushed to the server and not  held
##      on  the  client.  When  a  file is opened the client
##      sends a  request  to  the  server  to  validate  the
##      client's  local  caches.  This  behavior  ensures  a
##      file's consistency across multiple NFS clients. When
##      nocto is in effect, the client does not perform the
##      flush on  close  and  the  request  for  validation,
##      allowing  the possibility of differences among copies
##      of the same file as stored on multiple clients.
##      This option can be used where it can  be  guaranteed
##      that  accesses  to  a specified file system are made
##      from only one client and  only  that  client.  Under
##      such  a  condition,  the  effect  of nocto can be a
##      slight performance gain.
##
##  port=n
##      The server IP port number.
##
##  posix
##      Request  POSIX.1  semantics  for  the  file  system.
##      Requires a mount Version 2 mountd(1M) on the server.
##      See standards(5) for information regarding POSIX.
##
##  proto=tcp|udp
##      The default is to use tcp. When TCP is specified and a
##      TCP connection cannot be setup then a UDP connection will
##      be attempted. If a UDP fails then an error is returned.
##      The UDP protocol is not supported for NFS Version 4.
##      If you specify a UDP protocol with the proto option,
##      NFS version 4 is not used.
##      NOTE: The behavior above is not the same as Solaris
##      10 where the default protocol used is RDMA.
##
##  quota|noquota
##      Enable or prevent quota(1M)  to  check  whether  the
##      user  is over quota on this file system; if the file
##      system has quotas enabled on the server, quotas  are
##      still checked for operations on this file system.
##
##  retrans=n
##      Set the number of  NFS  retransmissions  to  n.  The
##      default  value  is  5. For connection-oriented
##      transports, this option has no effect because it  is
##      assumed  that the transport performs retransmissions
##      on behalf of NFS.
##
##  retry=n
##      The number of times to retry  the  mount  operation.
##      The default for the mount command is 10000.
##
##  rsize=n
##      Set the read buffer size to  n  bytes.
##
##  sec=mode
##      Set the security mode for NFS transactions. If  sec=
##      is  not specified, then the default action is to use
##      sys over NFS Version 2 mounts. Some OS mount commands
##      will negotiate a mode over NFS Version 3 while others
##      will default  to sys.  No validation  is performed on
##      this  parameter,  it is just  passed down  to the  OS
##      dependent  mount  command.   Only  one  mode  can  be
##      specified with the sec= option.
##      It is  best to not  use this option  unless your test
##      requires specific NFS authentication testing. If that
##      is  the  case  then  you  should  make sure  that the
##      authentication modes you specify are supported on the
##      client OSs that will be used by your test.
##      NFS Version 4 mounts negotiate a security mode  when
##      the  server  returns an array of security modes. The
##      client attempts the mount with each  security  mode,
##      in order, until one is successful.
##
##  timeo=n
##      Set the NFS timeout to n tenths  of  a  second.  The
##      default  value  is 11 tenths of a second for
##      connectionless transports, and 600 tenths of a second
##      for connection-oriented transports.
##
##  vers=NFS_version_number
##      By default, the version of NFS protocol used between
##      the  client and the server is version 3.
##
##  wsize=n
##      Set the write buffer size to n  bytes.
##
##  xattr|noxattr
##      Allow or disallow the creation and  manipulation  of
##      extended  attributes.  The  default  is  xattr.  See
##      fsattr(5) for a description of extended attributes.
##
## @return
##  returns the mountpoint on success and on error returns undef with
##  the error in $@.
##
## @notes
##  Although the $server and $path parameters are positional, yet
##  optional because they can be undef, they are left that way instead
##  of using named parameters for backwards compatibility with the
##  lib/mount.thpl which this file is replacing.

sub mount ($$%) {

    my ( $server, $path, %opts ) = validate_mixed(
        @_,
        [ { type => SCALAR | UNDEF }, { type => SCALAR | UNDEF }, ],
        {
            chmod => { type => SCALAR, optional => 1, regex => qr{^\d+$} },
            client     => { type => SCALAR | OBJECT, default  => "localhost" },
            fstype     => { type => SCALAR,          optional => 1 },
            mntopts    => { type => SCALAR,          optional => 1 },
            mountpoint => { type => SCALAR,          optional => 1 },
            mountretry => { type => SCALAR,          default  => 1 },
            sharing    => {
                type    => SCALAR,
                default => "shared",
                regex   => qr/shared|exclusive/
            },
            timeout => { type => SCALAR, default => 30 },

            verbose => {
                type     => SCALAR,
                default  => 0,
                optional => 1,
                regex    => qr/[01]/
            },
        }
    );

    my ( $h, $mountobj, $errstr );

    # The meaning of exclusive is different from GenericMount and NACL.
    # In GenericMount it means not to share the same directory for different mounts.
    # In NACL mount, it means create a seperate directory, like "mountpoint-<some randome hex number>".
    # This causes problems in scripts which assume that the mountpoint they passed is the only mountpoint to be
    # used, not something else like NACL gives. So what I have done is to use this option in GenericMount to set
    # nacltask_if_exists. This way we will abort if the mount already exists.
    if ( $opts{mountpoint} ) {
        if ( $opts{mountpoint} =~ /shared|exclusive/ ) {
            $errstr =
              "mountpoint cannot contain the words shared or exclusive\n";
        }
        if ( $opts{sharing} && $opts{sharing} ne "exclusive" ) {
            $errstr .= "if mountpoint specified sharing must be exclusive\n";
        }
        if ($errstr) {
            croak $errstr;
        }
        $opts{sharing} = "exclusive";
    }

    # Connect to client.
    my %client_opt = _fix_client($opts{client});
    delete $opts{client};

    # This should be moved to NACL:
    # http://web.netapp.com/~burt/views/577/577616.html
    # ----- Cut -----
    if ( $opts{mntopts} =~ s/(proto=)?(tcp|udp)(6)?// ) {
        my $keyword       = $1 || "";
        my $proto         = $2 || "";
        my $proto_version = $3 || "";
        my $req_proto     = "proto=$proto$proto_version";

        my $brackets_present = 0;
        if ( $server =~ /\[.+\]/ ) {
            $brackets_present = 1;
            $server =~ s/\[(.+)\]/$1/;
        }

        #$Log->comment(qq{server = $server});
        if ( ip_is_ipv4($server) ) {
            $req_proto = "proto=$proto";
        } elsif ( ip_is_ipv6($server) ) {
            $req_proto = "proto=${proto}6";
        } else {
            # We don't know what was given. it is a string. We will use the proto
            # and the server addresses as they were given, without modifications to them.
            if ($brackets_present) {
                $server = "[" . $server . "]";
                if ( $req_proto !~ /6$/ ) {
                    $req_proto .= "6";
                }
            }
        }

        $opts{mntopts} .= ",$req_proto";
    }
    if (ip_is_ipv6($server) && $server !~ /\[/) {
        $server = qq{[$server]};
    }

    if ( $opts{mntopts} =~ /,,/ ) {
        $opts{mntopts} =~ s/,,/,/g;
    }
    if ( $opts{mntopts} =~ /^,/ ) {
        $opts{mntopts} =~ s/^,//g;
    }
    if ( $opts{mntopts} =~ /,$/ ) {
        $opts{mntopts} =~ s/,$//g;
    }

    # ----- Paste -----

    # Rename the options to the correct entry for NACL.
    # Needed due to:612376 
    $opts{server}      = qq{'$server'};
    $opts{path}        = $path;
    $opts{mount_point} = $opts{mountpoint} if $opts{mountpoint};
    $opts{mountopts}   = $opts{mntopts} if $opts{mntopts};
    $opts{retries}     = $opts{mountretry} if $opts{mountretry};
    $opts{type}        = "nfs";

    # Convert sharing option to nacltask_if_exists option.
    if ( $opts{sharing} eq "shared" ) {
        $opts{nacltask_if_exists} = "reuse";
    } elsif ( $opts{sharing} eq "exclusive" ) {
        $opts{nacltask_if_exists} = "die";
    }

    # Delete the wrong entries for NACL.
    delete $opts{sharing};
    delete $opts{mntopts};
    delete $opts{mountretry};
    delete $opts{mountpoint};

    eval {
        my $client_obj = NACL::C::Client->new(%client_opt);
        #$Log->comment( qq{arguments for mount:} . Dumper( \%opts ) );
        $mountobj = NACL::STask::Client::Mount->create(
            command_interface => $client_obj,
            %opts
        );
    };

    if ($mountobj) {
        eval { $mountobj->mount_point(); };
        if ($@) {
            croak(qq{There seems to be a nacl problem, please talk to NACL team. The error:$@});
        }
        return $mountobj->mount_point();
    }
    else {
        croak(  qq{Couldnot mount, mount failed for }
              . Dumper( \%opts )
              . qq{Reason:$@} );
    }
}

## @name ::umount
## @description
##  Contains the generic front end to the OS dependent umount command.
##  It handles the following:
##   1) Processes options that are not passed to the OS dependent mount.
##   2) Construct mount class that has the OS specific info.
##   3) Call the OS specific mount with generic options.
##
## @synopsis
##  $status = umount($mntpoint, force => "1");
##
## @arg
##  mountpoint  Required: The mountpoint to unmount.
##
##  client  Optional: Hostspec or Connectrec of a client that
##          should perform the mount.  A hostspec (such as "$CLIENT")
##          is preferred, in which case the connection "#rsh" will be
##          used.  The default is "localhost", which means to perform
##          the mount locally.  If a Connectrec is given it must be for
##          a "per-say" type of connection i.e. one which effectively
##          connects/disconnects for every command.  (Thus, for example,
##          the Connectrec returned by connect("#rlogin") or
##          connect("#telnet") could not be used).
##
##  force   Optional: Without this option, umount does not allow a file
##          system to  be  unmounted  if a file on the file system is busy.
##          Using this option can cause data loss  for  open  files;
##          programs  which  access  files after the file system has
##          been unmounted will get an error (EIO).
##
##  timeout Optional: How many seconds to wait for a mount or umount
##          attempt to finish. Default is 60 seconds.
##
## @return
##  1 if succeeds else an error in $@

sub umount ($%) {
    my ( $mountpoint, %opts ) = validate_mixed(
        @_,
        [ { type => SCALAR }, ],
        {
            client  => { type => SCALAR | OBJECT | UNDEF, default  => undef },
            force   => { type => SCALAR,          optional => 1 },
            timeout => { type => SCALAR,          default  => 60 },
        }
    );
    my ( $h, $mountobj );

    # Add mountpoint to the opts hash
    $opts{mount_point} = $mountpoint;
    my %client_opt = _fix_client($opts{client});
    delete($opts{client});

    eval {

        # Get the hostrec entry and create a connection for the client
        my $client_obj = NACL::C::Client->new(%client_opt);

        # Delete using the MountClass object
        $mountobj = NACL::C::Client::Mount->delete(
            command_interface => $client_obj,
            mount_point       => $mountpoint,
            %opts
        );

        #$client_obj->destroy();
    };

    if ($@) {
        croak $@;
    }
    else {
        return 1;
    }
}

1;
