#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/HTTP_File_Server.pm#1

#***********************************************************************
#
# NAME
#   HTTP_File_Server.pm - 
#
# SYNOPSIS
#   An HTTP server implemented via PERL.
#
# DESCRIPTION
#   This library provides a basic HTTP server under Perl.  It's currently
#       catered to host files. Allows put, post, and gets.
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
# 7/13/2009 Initial revision
# 8/19/2009 Added retry logic when creating server/binding to port
# 6/1/2010 Support for put/resume 
# 5/1/2011 Support for files
#***********************************************************************

## @summary     HTTP_File_Server.pm

package HTTP_File_Server;

use strict;

#use lib '/x/eng/kc-users/weiss/perl';
#required for Archive::Extract, which also needs IPC::Cmd and Module::Load::Conditional

use Options;
use Tharn qw(host snooze subtest $Parent $Log fatal_result);

use NATE::Result;
use Data::Dumper;
use IO::Socket;
use IO::Select;  #for http server
use Sys::Hostname;   #for http server
use HTTP_Daemon; #for http server
use HTTP::Status; #for http server
use HTTP::Response;
use HTTP::Headers;
use HTTP::Request::Common;
use Carp;
use MIME::Parser;
use Data::Dumper;

my $opts = new Options;

$opts->parameters(definitions=>
    [
       {
        name => "HTTP_PUT_SUPPORT",
        description => "Does this Server support PUTs. Default is yes.",
        default => 1,
       },
       {
        name => "HTTP_POST_SUPPORT",
        description => "Does this Server support POSTs. Default is yes.",
        default => 1,
       },
        {
        name => "HTTP_GET_SUPPORT",
        description => "Does this Server support GETs. Default is yes.",
        default => 1,
       },
       {
        name => "HTTP_METHOD_NOT_SUPPORTED",
        description => "What HTTP equate to return if method is not supported. Default is 405.",
        default => 405,
       },
    ]);

my $opt_href = $opts->get_hash();

# new method - Create a new HTTP_File_Server object
#  Parameter hash (* means required):
#   *client               - Client on which to run the server
#    port                 - Port on which to run the server (default:80)
#   *dir             - Directory in which to store the files 
#                           (default: undef - don't store)
#    put_supported        - Should server support PUTs?  (default: 1)
#    get_supported        - Should server support GETs?  (default: 1)
#    post_supported       - Should server support POSTs? (default: 1)
#    method_not_supported - Return code to use when method is not supported? (default: 405)
#    read_bytes           - How many bytes to read of request before terminating server. (default: all)
#
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my %options = (
        dir=>undef,
        port=>80,
        put_supported=>$opt_href->{HTTP_PUT_SUPPORT},
        post_supported=>$opt_href->{HTTP_POST_SUPPORT},
        get_supported=>$opt_href->{HTTP_GET_SUPPORT},
        method_not_supported=>$opt_href->{HTTP_METHOD_NOT_SUPPORTED},
        ipv6=>"",
        read_bytes=>"all",
        @_
    );

    my @subArgs;
    if (!$options{client}) {
        $Log->error("HTTP File Server requires a client parameter \n");
        return 0;
    }

    if (!$options{dir}) {
        $Log->error("HTTP File Server requires a directory \n");
        return 0;
    }

    if ($options{put_supported} < 0) {
        $options{put_supported}=$opt_href->{HTTP_PUT_SUPPORT};
    }

    if ($options{post_supported} < 0) {
        $options{post_supported}=$opt_href->{HTTP_POST_SUPPORT};
    }

    if ($options{get_supported} < 0) {
        $options{get_supported}=$opt_href->{HTTP_GET_SUPPORT};
    }

    if ($options{method_not_supported} < 0) {
        $options{method_not_supported}=$opt_href->{HTTP_METHOD_NOT_SUPPORTED};
    }

    @subArgs = ('{use HTTP_File_Server; HTTP_File_Server::run(port=>"'.$options{port}.'",dir=>"'.$options{dir}.'"'.
                                                 ',client=>"'.$options{client}.
                                                 '",put_supported=>'.$options{put_supported}.
                                                 ',post_supported=>'.$options{post_supported}.
                                                 ',get_supported=>'.$options{get_supported}.
                                                 ',method_not_supported=>'.$options{method_not_supported}.
                                                 ',ipv6=>"'.$options{ipv6}.
                                                 '",read_bytes=>"'.$options{read_bytes}.'"); }',
                 '-bg', '-username', 'root', '-client', "$options{client}");

    my $subtest = subtest @subArgs;
    my $self = {
        "put_supported"        => $options{put_supported},
        "get_supported"        => $options{get_supported},
        "post_supported"       => $options{post_supported},
        "method_not_supported" => $options{method_not_supported},
        "client"               => $options{client},
        "port"                 => $options{port},
        "ipv6"                 => $options{ipv6},
        "read_bytes"           => $options{read_bytes},
        "subtest"              => $subtest,
        @_,
    };
    bless($self, $class);
    return $self;
}

sub restart {
    my $self = shift;
    my %options = (
        dir=>undef,
        client=>$self->{client},
        port=>$self->{port},
        put_supported=>$self->{put_supported},
        get_supported=>$self->{get_supported},
        post_supported=>$self->{post_supported},
        method_not_supported=>$self->{method_not_supported},
        ipv6=>$self->{ipv6},
        read_bytes=>$self->{read_bytes},
        @_
    );
    my @subArgs = ('{use HTTP_File_Server; HTTP_File_Server::run(port=>"'.$options{port}.'",dir=>"'.$options{dir}.'"'.
                                                 ',client=>"'.$options{client}.
                                                 '",put_supported=>'.$options{put_supported}.
                                                 ',post_supported=>'.$options{post_supported}.
                                                 ',get_supported=>'.$options{get_supported}.
                                                 ',method_not_supported=>'.$options{method_not_supported}.
                                                 ',ipv6=>"'.$options{ipv6}.
                                                 '",read_bytes=>"'.$options{read_bytes}.'"); }',
                 '-bg', '-username', 'root', '-client', "$options{client}");
    my $subtest = subtest @subArgs;
    $self->{"subtest"} = $subtest;
    $self->{client}=$options{client};
    $self->{dir}=$options{dir};
    $self->{port}=$options{port};
    $self->{get_supported}=$options{get_supported};
    $self->{put_supported}=$options{put_supported};
    $self->{post_supported}=$options{post_supported};
    $self->{method_not_supported}=$options{method_not_supported};
    $self->{ipv6}=$options{ipv6};
    $self->{read_bytes}=$options{read_bytes};
    sleep(30);
    return;
}

sub stop {
    my $self = shift;
    $Log->trace("Stopping the server...\n");
    $self->{"subtest"}->kill();
    sleep(10); #wait a few seconds to ensure the process dies and the port is free
    return;
}

sub get_entry {
    my $self = shift;
    my $message=undef;

    #get the single message from the queue
    #$message=$self->{"subtest"}->message_get(filter => sub{$_[0]->{message} =~ qr/HTTP/}, type => "HTTP", timeout => 3);
    $message=$self->{"subtest"}->message_get(type => "HTTP", timeout => 3);
    $Log->debug("Found raw message:  ".Dumper($message)."\n");
    return $message;
}

sub get_all {
    my @msgConcat = ();
    my $msgTmp = undef;

    while (1) {
        $msgTmp=undef;
        #get an entry that is on the queue, a message reported from the server
        eval {
            $msgTmp = get_entry(@_);
        };
        #if no message available, then exit
        if (!defined $msgTmp) {
            last;
        }
        #else add the message to the array
        else {
            $Log->debug("Adding message to array: ".Dumper($msgTmp));
            push @msgConcat, $msgTmp;
        }
    }
    #return the array of messages
    return \@msgConcat;
}

sub run {
    my %params = (
            client=>undef,
            dir=>undef,
            port=>80,
            put_supported=>1,
            post_supported=>1,
            method_not_supported=>405,
            read_bytes=>"all",
            ipv6=>"",
            @_,
    );

    if (!$params{client}) {
        return fatal_result("HTTP Server run requires a client parameter \n");
    }

    my $counter=0;

    # per message type counter
    my %counter= ( PUT => 0, POST => 0 );
    my $max_clients=5;
    
    if ($params{port} <= 0) {
        $params{port}=80;
    }

    if ($params{put_supported} < 0) {
        $params{put_supported}=1;
    }

    if ($params{post_supported} < 0) {
        $params{post_supported}=1;
    }

    if ($params{method_not_supported} < 0) {
        $params{method_not_supported}=405;
    }

    my $addr;
    if ( $params{ipv6} ) {
        $addr = $params{ipv6};
    } else {
        my $clientobj = host($params{client});
        $addr = $clientobj->default_ip();
    }

    $Log->comment("Creating HTTP Server with Address: http:\\\\$addr:$params{port} \n");
    $Log->comment("Protocol Support - PUT:$params{put_supported}   POST:$params{post_supported}  \n");
    $Log->comment("Bytes to read for request: $params{read_bytes}  \n");
    $Log->comment("HTTP File Directory = $params{dir} \n");
    my $d = HTTP_Daemon->new(LocalAddr=>$addr, LocalPort=>$params{port}) || undef;
    if (!defined($d)) {
        for (my $retries = 1; (!defined ($d) && $retries < 11); $retries++) {
            $Log->comment("Connection is ".Dumper($d)."\n");
            $d = HTTP_Daemon->new(LocalAddr=>$addr, LocalPort=>$params{port}) || undef;
            $Log->comment("Try $retries: Waiting 6 seconds for HTTP server to start on port $params{port}...Error: $!\n");
            snooze(6);
        }

        if (!defined($d)) {
            $Log->error("Could not bind to port or server failed to start...\n");
            return fatal_result("No server, unable to handle connetions");
        }
    }
    $Log->comment("HTTP Server now ready for connections \n");
    while (my $connection = $d->accept) { #$connection is ClientConn returned by the accept call.  Accept waits until a connection is opened.
        $counter++;
        
        while (my $request = $connection->get_request) { #get the request from the client connection, returns an http request object
    
                my $method = $request->method();
    
                if ( exists $counter{$method} ) {
                    $counter{$method}++;
                } else {
                    $counter{$method} = 0;
                }
    
                $Log->comment("HTTP $method request received...\n");
    
                if ( ($params{put_supported}  && ($method eq "PUT") ) ||
                     ($params{post_supported} && ($method eq "POST") ) ||
                     ($params{put_supported}  && ($method eq "HEAD") ) ) {
    
                    $Log->comment("HTTP $method request is supported...\n");
                    $Log->comment("HTTP Header: \n".$request->headers_as_string()."\n");
                    
                    if ($request->method() eq 'HEAD') { #if this is a resume, check for file and return appropriate codes
                        my $dir=$params{dir}.$request->uri();
                        my @file_info=split(" ",`du -s -b $dir/body.*`);
                        my $rec_bytes=$file_info[0];
                        if (!defined($rec_bytes)) {
                            $rec_bytes=0;
                        }
                        my $res = HTTP::Response->new(200,"OK");
                        $res->header("Content-Length"=>$rec_bytes);
                        $connection->send_response($res);          
                    }
                    else {
                        my $dir=undef;
                        if ($params{dir}) {
                            system ("mkdir $params{dir}");
                            $dir= $params{dir};
                            my $filename = $params{dir}."/".$request->uri();
                            $Log->comment("SAVING HTTP REQUEST TO: $filename\n");
                            open (datafile,">$filename");
                            #$Log->comment("Incoming content:".$request->content);
                            print datafile $request->content();
                            close (datafile);
                        }
                        $Parent->message_put(type => "HTTP", 
                                             message => "HTTP: HTTP $method of ".$request->header('X-Netapp-Asup-Subject').
                                             " received from node ".$request->header('X-Netapp-Asup-Hostname').
                                             ". Saved to directory $dir");
                        $Log->comment("***I've received $counter{$method} total HTTP $method requests***\n");
                        $Log->comment("***Returning code 200 in response to HTTP $method request***\n");

                        my $r = HTTP::Response->new(200);
                        $r->push_header('Content-type', 'text/html');
                        $r->content( "OK");
                        $connection->send_response($r);
                        $Log->comment("***Sent response back***");
                    }
                } elsif ($method eq "GET" && $params{get_supported}) {

                    my $filename = $request->uri();

                    if ($params{dir}) {
                        $filename = $params{dir} . $filename;
                    }
                    
                    my $rc = 200;
                    open (FILE, "$filename") or $rc = $params{method_not_supported};
                    if($rc == $params{method_not_supported})
                    {
                        my $output = `/bin/ls -lart $filename`;
                        $Log->comment("Not able to read $filename\n");
                        $Log->comment("Output from ls:$output\n");
                        $connection->send_status_line($params{method_not_supported});
                    } else {
                        $Log->comment("Reading file: $filename\n");
                        binmode FILE;
                        my ($buf, $n, $data);
                        while(($n = read FILE, $data, 4) != 0) {
                            $buf .= $data;
                        }
                        my $r = HTTP::Response->new(200);
                        $r->push_header('Content-type', 'Content-Disposition');
                        $r->content($buf);
                        $connection->send_response($r);
            
                        $Log->comment("Sent response back.");
                    }
                    
                } else {
                    $Log->comment("***HTTP $method request is NOT supported***\n");
                    $Log->comment("***Returning code $params{method_not_supported} in response to HTTP $method request***\n");
                    $connection->send_status_line($params{method_not_supported}); #send status code back to the client
                }
            }
        $connection->close;
    }
}


1;

=pod

=head1 NAME

HTTP_File_Server.pm - Basic HTTP server.


=head1 SYNOPSIS

    use HTTP_File_Server;
    my $http_server = new HTTP_File_Server(client => $params{client}, dir=>$params{save_dir});
    my $single_post = $http_server->get_entry();
    my $posts = $http_server->get_all();
    $http_server->stop();

=head1 DESCRIPTION

This library provides a basic HTTP server in perl.
  
    
=head1 METHODS

new(client=>CLIENT);
        Description:
            Creates a new instance of the HTTP_File_Server object.  It will bind to
                port 80 on the specified client.
        Inputs:
            client => the client to run this HTTP_File_Server on.
            dir => directory to save files to
            port => the port to use for the server
        Outputs:
            HTTP_File_Server object instance
        Examples:
            <see synopsis above>
            
stop();
        Description:
            Stops an HTTP_File_Server instance
        Inputs:
            <none>
        Outputs:
            <none>
        Examples:
            <see synopsis above>
            
get_entry();
        Description:
            Gets a single entry that was posted to the HTTP Server
        Inputs:
            <none>
        Outputs:
            $post => A single entry that was posted to the server.
        Examples:
            <see synopsis above>

get_all();
        Description:
            Get all entries that were posted to the HTTP Server
        Inputs:
            <none>
        Outputs:
            $posts => All entries that were posted to the server.
        Examples:
            <see synopsis above>
            
run();
        Description:
            The main HTTP server method.  This method will bind to a specific
                port on the specified client and listen for incoming request,
                like an HTTP server.  When a post is received, it will parse
                the post for header information (subject, from, etc.) and will
                place certain post information on the HTTP server queue to be
                read by a client application.  Also if requested, the server
                method will parse off the post request and save the full
                request to a location (decoding mime attachments, etc.)
        Inputs:
            $logdir => the directory to save off the post request
        Outputs:
            <none>
        Examples:
            <see synopsis above>
=cut
