
# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/lib/Hammer.pm#5 $
#
## @summary	Hammer.pm - Routines for starting and stopping hammer
## @author	dl-platform-qa@netapp.com
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

Hammer - Routines for starting and stopping hammer (7 & 10-mode)

=head1 SYNOPSIS

  # The server parameter can be used instead of the fc parameter if
  # your tests are using the Server.pm object. The CIFS side may need
  # additional parameters specified for 10-mode.

  # Running hammer from an CIFS client
  use Hammer qw(cifs_hammer_start cifs_hammer_stop);

  $cifs_hamid = cifs_hammer_start(
      fc             => $fc,
      cifs_client    => $CIFS_CLIENT,
      cifs_drive     => $CIFS_DRIVE,
      cifs_passwd    => $CIFS_PASSWD,
      cifs_user      => $CIFS_USER,
      cifs_workgroup => $CIFS_WORKGROUP,
      verify_traffic => 'yes',
      vol_name       => $vol,
  );

  cifs_hammer_stop(id => $cifs_hamid);

  # Running hammer locally from a UNIX client
  use Hammer qw(local_hammer_start local_hammer_stop);

  $local_hamid = local_hammer_start(
      drive       => '/dev/sdb',
      fc          => $fc,
      client      => 'localhost',
  );

  local_hammer_stop(id => $local_hamid);

  # Running hammer from an NFS client
  use Hammer qw(nfs_hammer_start nfs_hammer_stop);

  $nfs_hamid = nfs_hammer_start(
      fc             => $fc,
      client         => $CLIENT,
      verify_traffic => 'yes',
      vol_name       => $vol,
  );

  nfs_hammer_stop(id => $nfs_hamid);

  # Running hammer locally on the filer
  use Hammer qw(ontap_hammer_start ontap_hammer_stop);
  ontap_hammer_start(
      fc          => $fc,
      vol_name    => $vol,
  );

  ontap_hammer_abort(fc => $fc);

=head1 DESCRIPTION

Library routines for starting and stopping hammer from an NFS
client(s), a CIFS client(s), and the filer. Hammer is run in
the background for all cases. The stopping routines
are designed to be called from a testcase cleanup routine and
will log warnings instead of croaking on errors. More than one
instance of hammer can be started up from one or more clients.
Be careful not to run too many instances of hammer for a client
or filer as it may wedge the OS or make the filer so slow that
it becomes unresponsive.

The routines in this library are designed to be standalone in
that they will make sure that the filer services are licensed,
the filer exports/shares the volume and that the clients
mount/map them. These routines are ideal for creating background
load to a filer.

It is a good idea to not run the test case that starts and stops
hammer on the same client that is running hammer as it can lead
to timeouts due to an overloaded system.

CLIENT PLATFORM NOTES:

=over

- Seems to work on RHES4.2.

- RHES 5.0 see a warning on the umount but the umount is performed.

- Works fine on Solaris 10.

- Works fine on XP_1

- Works fine on W2003

=back

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-platform-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Hammer.pm#5 $
  Last revised $Date: 2013/12/09 $

head1 COPYRIGHT

  Copyright 2007 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Hammer;

use strict;
use warnings;

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

use base qw(Exporter);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );

    our @EXPORT_OK = qw(
        cifs_hammer_start
        cifs_hammer_stop
        local_hammer_start
        local_hammer_stop
        nfs_hammer_start
        nfs_hammer_stop
        ontap_hammer_start
        ontap_hammer_abort
        Gxontap_hammer_start
        Gxontap_hammer_abort
    );
};

use Admin::License qw(is_licensed license);
use Carp qw(croak);
use ExportFS qw(share_add share_delete);
use Fcntl;
use File::Temp qw(:mktemp);
use FileManip qw(read_file);
use GenericMount qw(mount umount);
use LogReader;
use MapShare qw(map_cifs_share unmap_cifs_share);
use Params::Validate qw(validate OBJECT SCALAR HASHREF UNDEF);
use Tharn;
use filer qw(filer_sysstat_get);
use nfs::GenericExport qw(export_path unexport_path);
use Server;

# Use this until a library routine that can dynamically determine
# the mode of a filer is available.
use platqa::user_interfaces::shell_commands::SysCmdUtils qw(filer_mode);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub _server_to_fc($;);
sub _is_7_mode($;);
sub _verify_hammer($$$$$;);
sub cifs_hammer_start(%);
sub cifs_hammer_stop(%);
sub local_hammer_start(%);
sub local_hammer_stop(%);
sub nfs_hammer_start(%);
sub nfs_hammer_stop(%);
sub ontap_hammer_start(%);
sub ontap_hammer_abort(%);
sub Gxontap_hammer_start(%);
sub Gxontap_hammer_abort(%);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $Cifs_hammer_runid = 0;
my $Local_hammer_runid = 0;
my $Nfs_hammer_runid = 0;
my $Ontap_hammer_runid = 0;

#---------------------------------------------------------------------
# Internal function to map a server object to a filer connection.
#---------------------------------------------------------------------

sub _server_to_fc($;) {
    my ($server) = shift;
    my ($fc);

    # Get  connection object
    $fc = $server->get_connection(type => 'cli_connection');
    if (!defined($fc)) {
        croak('_server_to_fc: cannot get cli connection from Server object');
    }
    return ($fc);
}

#---------------------------------------------------------------------
# Internal function to see if the filer is running 7-mode.
# Takes a Server.pm object as a parameter and will return 1 if
# the filer is running 7-mode or 0 otherwise.
#---------------------------------------------------------------------

sub _is_7_mode($;) {
    my ($server) = shift;
    my ($mode);

    $mode = filer_mode(server => $server);
    if ($mode eq '7') {
        return (1);
    } else {
        return (0);
    }
}

#---------------------------------------------------------------------
# Internal function to verify that hammer started up on NFS or CIFS.
# If the parmeter $verify_traffic is set to 'no' then verify that the
# subtest started. If $verify_traffic is set to 'yes' then verify the
# traffic on the filer. The parmeter $type should be either 'NFS' or 'CIFS'
#---------------------------------------------------------------------

sub _verify_hammer($$$$$;) {
    my ($obj, $mode7, $verify_traffic, $hammer_obj, $type) = @_;
    my ($avg, $end_time, $line, @lines, $msg, $next, $out, %stats);

    $msg = '';
    $end_time = time() + 60;
    if ($verify_traffic eq 'no') {
        $Log->comment("Wait 60 secs for the hammer execution to start");
        while (!$hammer_obj->is_alive() && time() < $end_time) {
            sleep(5);
        }

        # After waiting 60 seconds make sure subtest is still running
        # and did not exit because of hammer error
        if (!$hammer_obj->is_alive()) {
            $msg = 'Hammer subtest did not continue running';
        }
    } else {
        $Log->comment("Verify $type traffic starts up before 60 seconds");
        do {
            if ($mode7) {
                # Collect 2 samples of stats every 5 secs
                %stats = filer_sysstat_get($obj,
                                           {'FILER_SYSSTAT_COUNT'    => 2,
                                            'FILER_SYSSTAT_INTERVAL' => 5});
                $avg = $stats{$type}{Avg};
            } else {
                $out = $obj->statistics_periodic(interval   => 5,
                                                 iterations => 2,
                                                 summary    => 'true',
                                                 '-Timeout' => 120);
                @lines = split(/\n/, $out);
                $avg = $next = 0;
                foreach $line (@lines) {
                    if ($line =~ /^Averages/) {
                        $next = 1;
                        next;
                    }
                    if ($next) {
                        # Remove whitespace from front of line
                        $line =~ s/^\s+//;
                        @lines = split(/\s+/, $line);
                        if ($type eq 'NFS') {
                            $avg = $lines[3];
                        } else {
                            $avg = $lines[4];
                        }
                        last;
                    }
                }
            }
        } while (($avg == 0) && (time() < $end_time));

        if ($avg == 0) {
            $msg = "Initial $type traffic did not start up";
        }
    }
    return ($msg);
}

#---------------------------------------------------------------------

=head2 cifs_hammer_start

    $cifs_hamid = cifs_hammer_start(
        fc             => $fc,
        cifs_client    => $CIFS_CLIENT,
        cifs_drive     => $CIFS_DRIVE,
        cifs_passwd    => $CIFS_PASSWD,
        cifs_user      => $CIFS_USER,
        cifs_workgroup => $CIFS_WORKGROUP,
        hammer_exe     => $HAMMER_EXE,
        hammer_time    => -1,
        vol_name       => $vol,
    );

Invokes a CIFS version of hammer on the specified cifs_client. The
library routine will share the volume on the filer, map the volume
to a drive letter on the cifs_client, then startup hammer on the
CIFS client. Running on 7-mode will use workgroup authentication, running
on 10-mode will use domain authentication. If hammer was started up
successfully then it will
return a hash ref to the state required to stop this invocation
of hammer. More than one invocation of hammer can be performed
per client but be careful because running too many on Windows clients
has been known to wedge them.

This routine will cleanup before croaking.

=over

=item Arguments

=over

=item cifs_client

(Required) : CIFS Client to execute cifs hammer subtest on.

=item cifs_drive

(Optional) : Drive letter to use when mapping the CIFS share on
cifs_client. Defaults to '*' (let the OS assign the drive letter).

=item cifs_passwd

(Required) : Password for the cifs_user.

=item cifs_user

(Optional) : Username to map cifs share as. Defaults to 'administrator'.

=item cifs_workgroup

(Optional) : CIFS workgroup name to use when setting up CIFS on the filer.
Not used for 10-mode. Defaults to 'WORKGROUP'.

=item cifs_domain

(Optional) : CIFS active directory domain name to use when mapping the share
on the client name.
Not used for 7-mode.

=item fc

(Optional) : Connection object to a filer. Either the fc or server
parameter must be defined.

=item filer_ip

(Optional) : Filer IP address to send hammer traffic to. Defaults to
filer's hostspec default IP address.

=item hammer_exe

(Optional) : Path to CIFS hammer executable. Defaults to 'X:/bin/hammer.exe'.

=item hammer_time

(Optional) : Amount of time to run hammer. Defaults to -1 (infinite).

=item panic

(Optional) : Panic the filer if corruption is detected by hammer.
Values are 'yes' or 'no'. Default is 'yes'.

=item server

(Optional) : Server object for a filer. Either the fc or server
parameter must be defined.

=item verify_traffic

(Optional) : Use systat on the filer to verify that CIFS traffic
is occurring on the filer. Default is 'no'.

=item vol_name

(Required) : Name of volume to mount and execute hammer stress

=item vserver

(Optional) : Name of vserver. Required for 10-mode.

=back

=item Returns

Hash reference that contains the state associated with this hammer
invocation. Pass it in to the cifs_hammer_stop() routine to stop
this hammer invocation. Croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub cifs_hammer_start(%) {
    my (%opts) = validate (
        @_,
        {
            cifs_client    => { type => SCALAR },
            cifs_drive     => { type => SCALAR, default => '*' },
            cifs_passwd    => { type => SCALAR },
            cifs_user      => { type => SCALAR, default => 'administrator' },
            cifs_domain    => { type => SCALAR|UNDEF, optional => 1 },
            cifs_workgroup => { type => SCALAR, default => 'WORKGROUP' },
            fc             => { type => OBJECT, optional => 1 },
            filer_ip       => { type => SCALAR, optional => 1 },
            hammer_exe     => { type => SCALAR,
                                default => 'X:/bin/hammer.exe' },
            hammer_time    => { type => SCALAR, default => -1 },
            panic          => { type => SCALAR, default => 'yes',
                                regex => qr/^(no|yes)$/ },
            server         => { type => OBJECT, optional => 1 },
            verify_traffic => { type => SCALAR, default => 'no',
                                regex => qr/^(yes|no)$/ },
            vol_name       => { type => SCALAR },
            vserver        => { type => SCALAR|UNDEF, optional => 1 },
        }
    );
    my ($drive, $end_time, $err, %fields, $hammer_args, $hammer_cmd,
        $hammer_obj, $hammer_path, %id, $logfile, $mode7, $server);

    if (!defined($opts{fc}) && !defined($opts{server})) {
        croak('cifs_hammer_start: Either the fc or server parameter must ' .
              'be specified.');
    }

    $Log->comment("cifs_hammer_start: It's CIFS Hammer Time!");

    # If a Server.pm object was specified then see if filer is 7 mode. If
    # so then use the filer connection from the Server.pm object. If a
    # Server.pm object was not passed in then filer is 7-mode.
    if (defined($opts{server})) {
        $opts{fc} = _server_to_fc($opts{server});
        $mode7 = _is_7_mode($opts{server});
    } else {
        $mode7 = 1;
    }
    $id{cifs_client} = $opts{cifs_client};
    $id{licensed} = 0;
    $id{mode7} = $mode7;

    if ($mode7) {
        # 7 mode
        $id{fc} = $opts{fc};
        $id{server} = $opts{server};

        if (!is_licensed($opts{fc}, 'cifs')) {
            $id{licensed} = 1;
            license($opts{fc}, 'cifs');
            croak $@ if $@;
        }

        # Default the IP address to send traffic to the filer to be the
        # filer default IP address if not specified as an option.
        if (!defined($opts{filer_ip})) {
            $opts{filer_ip} = $opts{fc}->hostp->default_ip;
        }

        share_add(
            $opts{fc},
            $opts{vol_name},
            "/vol/$opts{vol_name}",
            verify         => 1,
            umask          => 0,
            purge_existing => 1,
            caching        => 'no',
        );
    } else {
        # 10 mode
        if (!defined($opts{vserver})) {
            croak("Running CIFS hammer on 10-mode filer requires the " .
                  "vserver parameter to be specified");
        }
        if (!defined($opts{cifs_domain})) {
            croak("Running CIFS hammer on 10-mode filer requires the " .
                  "cifs_domain parameter to be specified");
        }

        $server = $id{server} = $opts{server};
        $id{vserver} = $opts{vserver};

        # Grab the DATA_IP1 address for the CIFS servers IP address
        %fields = $opts{fc}->hostp()->param_fields();
        if (!defined($fields{DATA_IP1})) {
            croak("Running hammer on 10-mode filer requires the DATA_IP1 " .
                  "parameter to be defined in the filers hostspec");
        }
        if (!defined($opts{filer_ip})) {
            $opts{filer_ip} = $fields{DATA_IP1};
        }

        # See if the volume is already shared. If not share it.
        eval {
            $server->vserver_cifs_share_show(
                            vserver               => $opts{vserver},
                            'share-name'          => $opts{vol_name});
        };
        if ($@) {
            $server->vserver_cifs_share_create(
                            vserver               => $opts{vserver},
                            path                  => "/$opts{vol_name}",
                            'dir-umask'           => 0,
                            'file-umask'          => 0,
                            'share-name'          => $opts{vol_name});
        }
        $opts{cifs_workgroup} = $opts{cifs_domain};
    }

    $id{share} = $opts{vol_name};
    $id{runid} = ++$Cifs_hammer_runid;

    # Wait a few seconds to make sure that the filer has really shared
    # the volume. A fast client can try to do the map too quickly.
    sleep(5);
    $Log->comment("Map the drive from the windows client");

    # Try up to 10 times to map the drive. Sometimes the
    # CIFS client gives an invalid error that goes away after
    # a few tries.
    for (1..10) {
        $drive = map_cifs_share(
                          server    => $opts{filer_ip},
                          client    => $opts{cifs_client},
                          drive     => $opts{cifs_drive},
                          share     => $opts{vol_name},
                          username  => $opts{cifs_user},
                          password  => $opts{cifs_passwd},
                          workgroup => $opts{cifs_workgroup},
        );
        last if ($drive);
        sleep(10);
    }
    if (!$drive) {
        $err = $@;
        cifs_hammer_stop(id => \%id);
        croak ("Mapping CIFS share failed: $err");
    }
    $id{drive} = $drive;

    $hammer_path = "$drive/hamcifs$id{runid}_$$";
    # The CIFS version of hammer started creating a random directory name
    # to hammer in therefore the mktemp call is not used here. There were
    # issues with using it on a windows client.
    $logfile = 'hammer.log';
    $hammer_args = "-bg -logfile $logfile";
    if ($opts{panic} eq 'yes') {
        $hammer_args .= " -filercore " . $opts{fc}->hostp()->hostname();
    } else {
        $hammer_args .= " -nofilercore ";
    }
    $hammer_args .= " -runtime $opts{hammer_time}";

    $hammer_cmd = qq{\{
                      mkdir \'$hammer_path\' or die "mkdir $hammer_path failed: \$!";
                      chmod 0777, \'$hammer_path\' or die "chmod $hammer_path failed: \$!";
                      chdir \'$hammer_path\' or die "chdir $hammer_path failed: \$!";
                      qx($opts{hammer_exe} $hammer_args);
                  \}};

    eval {
        $hammer_obj =
            subtest(
                $hammer_cmd,
                -bg,
                -client, $opts{cifs_client},
                -onexit, sub {},
            );
    };
    if ($@) {
        $err = $@;
        cifs_hammer_stop(id => \%id);
        croak ("Could not start hammer execution: $err");
    }
    $id{subtest} = $hammer_obj;
    $id{hammer_path} = $hammer_path;
    $id{logfile} = $logfile;

    if ($mode7) {
        $err = _verify_hammer($opts{fc}, $mode7, $opts{verify_traffic},
                              $hammer_obj, 'CIFS');
    } else {
        $err = _verify_hammer($opts{server}, $mode7, $opts{verify_traffic},
                              $hammer_obj, 'CIFS');
    }
    if ($err) {
        cifs_hammer_stop(id => \%id);
        croak ($err);
    }

    return (\%id);
}

#---------------------------------------------------------------------

=head2 cifs_hammer_stop

    cifs_hammer_stop(id => $cifs_hamid);

Call this routine to stop hammer running on a cifs_client that
was started up with cifs_hammer_start(). Pass in the id that
was returned from the cifs_hammer_start() call. This routine
will kill the hammer process running on the cifs client, it
will unmap the share on the client, and delete the share on the filer.
Typically this routine is called from a test cases cleanup routine.

=over

=item Arguments

=over

=item id

(Required) : Hash reference that contains the state associated with the hammer
invocation by cifs_hammer_start.

=item timeout

(Optional) : Number of seconds to wait for events to occur.
Defaults to 300 secs.

=back

=item Returns

Nothing. LogWarns on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub cifs_hammer_stop(%) {
    my (%opts) = validate (
        @_,
        {
            id      => { type => HASHREF },
            timeout => { type => SCALAR, default => 300 },
        }
    );
    my ($cc, $end_time, $file, $line, @lines, $logfile, $not_found, $out,
        $pid, $runid, $sig_cmd, $sobj, $status, $path, $rdir, $server);

    $Log->comment('cifs_hammer_stop: Stopping hammer');

    if (defined($opts{id}{subtest}) && $opts{id}{subtest}->is_alive()) {
        # Need to send a signal to the hammer process. Sending a signal
        # via subtest just kills the subtested process but the shell
        # that spawned hammer and hammer continue to run.
        $Log->comment('cifs_hammer_stop: Get hammers pid from the log file');

        # If hammer was just started, it may not have had a chance to
        # log its pid yet so we just poll for it.
        $logfile = "$opts{id}{hammer_path}/$opts{id}{logfile}";
        $not_found = $pid = 0;
        $end_time = time() + $opts{timeout};
        while ($pid == 0 && time() < $end_time) {
            eval {
                $out = read_file(client => $opts{id}{cifs_client},
                                 file => $logfile);
                @lines = split(/\n/, $out);
                foreach $line (@lines) {
                    if ($line =~ /^path=/i) {
                        ($pid) = ($line =~ /pid=0x(\S+)/);
                        $pid = hex($pid);
                        last;
                    }
                }
                sleep(10) if ($pid == 0);
            };
            if ($@) {
                if ($not_found++ > 0) {
                    $not_found = 0;
                    # Try 2 times to read the file, first look in the
                    # expected location. If not there then look in the
                    # subdirectories under the expected location because
                    # hammer started creating a random directory.
                    if (!defined($cc)) {
                        host($opts{id}{cifs_client});
                        $cc = connect('#rsh', -timeout, 300);
                        $cc->timeout(120);
                    }

                    # Map path to DOS format
                    $path = $opts{id}{hammer_path};
                    $path =~ s/\//\\/g;

                    # Look for the randomly created directory
                    (undef, undef, $out) = say($cc, "cmd /c dir $path /w");
                    ($rdir) = ($out =~ /\[([^\]]{3,})\]/);

                    if (defined($rdir)) {
                        $logfile="$opts{id}{hammer_path}/$rdir/$opts{id}{logfile}";
                    }
                }
                sleep(10);
            }
        }

        if ($pid != 0) {
            # Use subtest to send a signal to hammer. Using subtest because
            # if client is specified then the kill has to run remotely.
            # Using the PERL kill command
            #
            # On windows it appears that hammer is not catching the
            # signal and is dieing without printing a message about
            # stopping. Therefore, cannot validate through the log file
            # that hammer stopped. Just validate it via the subtest.
            $sig_cmd = qq{\{
                           kill 15, $pid or die "Kill failed: $!";
                       \}};

            eval {
                $sobj =
                    subtest(
                        $sig_cmd,
                        -client, $opts{id}{cifs_client},
                        -onexit, sub {},
                    );
            };
            $Log->warn("Could not send signal to hammer: $@") if ($@);
        } else {
            $Log->warn("Could not get hammers pid from logfile: $logfile: $@");
        }

        $Log->comment("Wait up to 60 secs for the hammer execution to stop");
        $end_time = time() + 60;
        do {
            sleep(5);
        } while ($opts{id}{subtest}->is_alive() && time() < $end_time);

        # Make sure hammer is dead
        if ($opts{id}{subtest}->is_alive()) {
            $Log->warn("Hammer subtest did not die");
        }

        # Save the hammer log file to the NATE log directory
        eval {
            $out = read_file(client => $opts{id}{cifs_client},
                             file => $logfile);
            $runid = $opts{id}{runid};
            $file = "$opts{id}{subtest}->{logdir}/cifs_hammer_$runid.log";
            sysopen(OUT, $file, O_WRONLY|O_CREAT) or
                die("open failed: $file: $!");
            print OUT "$out\n";
            close(OUT);
        };
        $Log->warn("Could not save hammer log file: $@") if ($@);
    }

    if (defined($opts{id}{drive})) {
        for (1..10) {
            $out = unmap_cifs_share(
                mappath => $opts{id}{drive},
                client  => $opts{id}{cifs_client},
                force   => 1,
            );
            $status = $@;
            last if ($out);
            sleep(10);
        }
        if (!$out) {
            $Log->warn("Unmapping $opts{id}{drive} on " .
                       "$opts{id}{cifs_client} failed: $status");
        }
    }

    if (defined($opts{id}{share})) {
        if ($opts{id}{mode7}) {
            # 7-mode
            eval {
                share_delete($opts{id}{fc}, $opts{id}{share}, force => 1);
            };
            $Log->warn($@) if ($@);
        } else {
            # 10-mode
            $server = $opts{id}{server};
            # Make sure the volume is shared
            eval {
                $server->vserver_cifs_share_show(
                                vserver               => $opts{id}{vserver},
                                'share-name'          => $opts{id}{share});
            };
            if (!$@) {
                eval {
                    $server->vserver_cifs_share_delete(
                                vserver               => $opts{id}{vserver},
                                'share-name'          => $opts{id}{share});
                };
                $Log->warn($@) if ($@);
            }
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 local_hammer_start

    $local_hamid = local_hammer_start(
        drive       => $drive,
        fc          => $fc,
        client      => $CLIENT,
        hammer_time => -1,
    );

Invokes a hammer on a local mount on the specified client. Useful
for running hammer on filer iscsi drives. The
library routine will mount the volume on the client, and then
startup hammer on the client. If hammer was started up successfully
then it will return a hash ref to the state required to stop this
invocation of hammer. More than one invocation of hammer can be
performed per client but be careful because running too many on Linux
clients has been known to wedge them.

This routine will cleanup before croaking.

=over

=item Arguments

=over

=item drive

(Required) : Drive to mount and run hammer against on client.

=item fc

(Optional) : Connection object to a filer. Either the fc or hostname
parameter must be defined.

=item client

(Optional) : Client to execute subtest on

=item hammer_bin

(Optional) : Path to Solaris/Linux hammer executable.
Defaults to '/usr/software/test/bin/hammer'.

=item hammer_time

(Optional) : Timeout for the hammer executable. Default is infinite
run time.

=item hostname

(Optional) : Hostname of the filer. Either the fc or hostname
parameter must be defined.

=item panic

(Optional) : Panic the filer if corruption is detected by hammer.
Values are 'yes' or 'no'. Default is 'yes'.

=back

=item Returns

Hash reference that contains the state associated with this hammer
invocation. Pass it in to the local_hammer_stop() routine to stop
this hammer invocation. Croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub local_hammer_start(%) {
    my (%opts) = validate (
        @_,
        {
            drive       => { type => SCALAR },
            fc          => { type => OBJECT, optional => 1 },
            client      => { type => SCALAR, default => 'localhost' },
            hammer_bin  => { type => SCALAR,
                             default => '/usr/software/test/bin/hammer' },
            hammer_time => { type => SCALAR, default => -1 },
            hostname    => { type => OBJECT, optional => 1 },
            panic       => { type => SCALAR, default => 'yes',
                             regex => qr/^(no|yes)$/ },
        }
    );
    my ($end_time, $err, $hammer_args, $hammer_cmd, $hammer_obj, $hammer_path,
        $hostname, %id, $line, $logdir, $logfile, $mount_dir);

    $Log->comment("local_hammer_start: It's NFS Hammer Time!");

    if (defined($opts{hostname})) {
        $hostname = $opts{hostname};
    } elsif (defined($opts{fc})) {
        $hostname = $opts{fc}->hostp()->hostname();
    } else {
        croak('local_hammer_start: Either the fc or hostname parameter must ' .
              'be specified.');
    }

    if (defined($ENV{LOGDIR})) {
        $logdir = $ENV{LOGDIR};
    } else {
        croak "Problem with NATE - LOGDIR not set in env";
    }

    $id{client} = $opts{client};
    $id{runid} = ++$Local_hammer_runid;

    $mount_dir = mount(
        undef,
        $opts{drive},
        client => $opts{client},
        chmod  => '0777',
    );
    if (!$mount_dir) {
        $err = $@;
        local_hammer_stop(id => \%id);
        croak $err;
    }
    $id{mountdir} = $mount_dir;

    # This will run hammer with its output going to $LOGDIR/local_hammer.*.txt
    # Stderr is not redirected, so if there is a problem, it is captured
    # and printed at the end of the subtest, and should appear in the
    # subtest log.

    $hammer_path = "$mount_dir/hammer_local_$id{runid}";
    $logfile = mktemp("$logdir/hammer_local.XXXXXX");
    $hammer_args = "$hammer_path";
    if ($opts{panic} eq 'yes') {
        $hammer_args .= " -filercore " . $hostname;
    } else {
        # Hammer on Linux needs the nofilercore option
        if (host($opts{client})->hosttype() =~ /linux/) {
            $hammer_args .= " -nofilercore ";
        }
    }
    $hammer_args .= " -runtime $opts{hammer_time} -blocksize random";

    $hammer_cmd = qq{\{
                      if (!-e \'$hammer_path\') {
                          mkdir \'$hammer_path\' or die 'mkdir failed';
                      }
                      qx($opts{hammer_bin} $hammer_args > $logfile);
                  \}};

    eval {
        $hammer_obj =
            subtest(
                $hammer_cmd,
                -bg,
                -client, $opts{client},
                -onexit, sub {},
            );
    };
    if ($@) {
        $err = $@;
        local_hammer_stop(id => \%id);
        croak ("Could not start hammer execution: $err");
    }
    $id{subtest} = $hammer_obj;
    $id{logfile} = $logfile;

    $Log->comment("Wait up to 60 secs for the hammer execution to start");
    $end_time = time() + 60;
    while (!$hammer_obj->is_alive() && time() < $end_time) {
        sleep(5);
    }

    # After waiting 60 seconds make sure subtest is still running
    # and did not exit because of hammer error
    if (!$hammer_obj->is_alive()) {
        local_hammer_stop(id => \%id);
        croak ("Hammer subtest did not continue running");
    }

    return (\%id);
}

#---------------------------------------------------------------------

=head2 local_hammer_stop

    local_hammer_stop(id => $local_hamid);

Call this routine to stop hammer running on a local_client that
was started up with local_hammer_start(). Pass in the id that
was returned from the local_hammer_start() call. This routine
will kill the hammer process running on the nfs client, it
will umount the volume, and unexport the volume on the filer.
Typically this routine is called from a test cases cleanup routine.

=over

=item Arguments

=over

=item id

(Required) : Hash reference that contains the state associated with the hammer
invocation by local_hammer_start.

=item timeout

(Optional) : Number of seconds to wait for events to occur.
Defaults to 300 secs.

=back

=item Returns

Nothing. LogWarns on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub local_hammer_stop(%){
    my (%opts) = validate (
        @_,
        {
            id      => { type => HASHREF },
            timeout => { type => SCALAR, default => 300 },
        }
    );
    my ($end_time, $file, $line, $pid, $runid, $sig_cmd, $sobj, $status);

    $Log->comment('local_hammer_stop: Stopping hammer');

    if (defined($opts{id}{subtest}) && $opts{id}{subtest}->is_alive()) {
        # Need to send a signal to the hammer process. Sending a signal
        # via subtest just kills the subtested process but the shell
        # that spawned hammer and hammer continue to run.
        $Log->comment('local_hammer_stop: Get hammers pid from the log file');

        # If hammer was just started, it may not have had a chance to
        # log its pid yet so we just poll for it.
        $pid = 0;
        $end_time = time() + $opts{timeout};
        while ($pid == 0 && time() < $end_time) {
            sysopen(IN, $opts{id}{logfile}, O_RDONLY) or
                $Log->warn("open failed: $opts{id}{logfile}: $!");
            while (($pid == 0) && ($line = <IN>)) {
                if ($line =~ /^path=/i) {
                    ($pid) = ($line =~ /pid=0x(\S+)/);
                    $pid = hex($pid);
                }
            }
            close IN;
            sleep(10) if ($pid == 0);
        }

        if ($pid != 0) {
            # Use subtest to send a signal to hammer. Using subtest because
            # if client is specified then the kill has to run remotely.
            # Using the UNIX kill command
            $sig_cmd = qq{\{
                           kill 15, $pid or die "Kill failed: $!";
                       \}};

            eval {
                $sobj =
                    subtest(
                        $sig_cmd,
                        -client, $opts{id}{client},
                        -onexit, sub {},
                    );
            };
            if (!$@) {
                $Log->comment("Verifying hammer stopped");
                $status = 0;
                $end_time = time() + $opts{timeout};
                while ($status == 0 && time() < $end_time) {
                    sysopen(IN, $opts{id}{logfile}, O_RDONLY) or
                        $Log->warn("open failed: $opts{id}{logfile}: $!");
                    while (($status == 0) && ($line = <IN>)) {
                        $status = 1 if ($line =~ /hammer stopped/i);
                    }
                    close IN;
                    # Sleep regardless of whether hammer stopped or not
                    # to give the OS a chance to cleanup the process.
                    sleep(10);
                }
                if ($status == 0) {
                    $Log->warn("Timed out waited $opts{timeout} secs for " .
                           "hammer to stop");
                }
            } else {
                $Log->warn("Could not send signal to hammer: $@");
            }
        } else {
            $Log->warn("Could not get hammers pid from logfile: $@");
        }

        # Save the hammer logfile to the subtest log directory
        if (-e $opts{id}{logfile}) {
            $runid = $opts{id}{runid};
            $file = "$opts{id}{subtest}->{logdir}/local_hammer_$runid.log";
            rename $opts{id}{logfile}, $file or
              $Log->warn("Rename $opts{id}{logfile} to $file failed");
        }
    }

    if (defined($opts{id}{mountdir})) {
        umount($opts{id}{mountdir}, client => $opts{id}{client}, force => 1) or
          $Log->warn("Unable to unmount dir: $opts{id}{mountdir}: $@");
        $opts{id}{mountdir} = undef;
    }

    $Log->warn($@) if ($@);

    return;
}

#---------------------------------------------------------------------

=head2 nfs_hammer_start

    $nfs_hamid = nfs_hammer_start(
        fc          => $fc,
        client      => $CLIENT,
        hammer_time => -1,
        vol_name    => $vol,
    );

Invokes a NFS version of hammer on the specified client. The
library routine will export the volume on the filer, mount the volume
on the client, and then startup hammer on the client. If hammer was
started up successfully then it will return a hash ref to the state
required to stop this invocation of hammer. More than one invocation
of hammer can be performed per client but be careful because running
too many on Linux clients has been known to wedge them.

This routine will cleanup before croaking.

=over

=item Arguments

=over

=item fc

(Optional) : Connection object to a filer. Either the fc or server
parameter must be defined.

=item client

(Optional) : Client to execute subtest on

=item filer_ip

(Optional) : Filer IP address to send hammer traffic to. Defaults to
filer's hostspec default IP address.

=item hammer_bin

(Optional) : Path to Solaris/Linux hammer executable.
Defaults to '/usr/software/test/bin/hammer'.

=item hammer_time

(Optional) : Timeout for the hammer executable. Default is infinite
run time.

=item panic

(Optional) : Panic the filer if corruption is detected by hammer.
Values are 'yes' or 'no'. Default is 'yes'.

=item proto

(Optional) : The protocol to use for the NFS mount. The options are
'tcp' or 'udp'. The default is 'tcp'.

=item rsize

(Optional) : Set the read buffer size to n bytes.
The default is 32768.

=item server

(Optional) : Server object for a filer. Either the fc or server
parameter must be defined.

=item version

(Optional) : NFS version to use. The default is 3. Using version 4
is just looking for trouble, you will either break the NFS client
or the Filer or both. Solaris works best, with Linux you get what
you paid for.

=item verify_traffic

(Optional) : Use systat on the filer to verify that NFS traffic
is occurring on the filer. Default is 'no'.

=item vol_name

(Required) : Name of volume to mount and execute hammer stress

=item mode

(Optional) : Mode in which hammer should be invoked.
mode => mixed | creates | overwrites. If nothing is specifed,
then hammer will be invoked in its default mode, which is mixed mode.

=item wsize

(Optional) : Set the write buffer size to n bytes.
The default is 32768.

=back

=item Returns

Hash reference that contains the state associated with this hammer
invocation. Pass it in to the nfs_hammer_stop() routine to stop
this hammer invocation. Croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub nfs_hammer_start(%) {
    my (%opts) = validate (
        @_,
        {
            fc             => { type => OBJECT, optional => 1 },
            client         => { type => SCALAR, default => 'localhost' },
            filer_ip       => { type => SCALAR, optional => 1 },
            hammer_bin     => { type => SCALAR,
                                default => '/usr/software/test/bin/hammer' },
            hammer_time    => { type => SCALAR, default => -1 },
            panic          => { type => SCALAR, default => 'yes',
                                regex => qr/^(no|yes)$/ },
            proto          => { type => SCALAR, default => 'tcp',
                                regex => qr/^(udp|tcp)$/ },
            rsize          => { type => SCALAR, default => 32768,
                                regex => qr/^\d+$/ },
            server         => { type => OBJECT, optional => 1 },
            version        => { type => SCALAR, default => 3 },
            verify_traffic => { type => SCALAR, default => 'no',
                                regex => qr/^(yes|no)$/ },
            vol_name       => { type => SCALAR },
            mode           => { type => SCALAR, optional => 1 },
            wsize          => { type => SCALAR, default => 32768,
                                regex => qr/^\d+$/ },
        }
    );
    my ($end_time, $err, %fields, $hammer_args, $hammer_cmd, $hammer_obj,
        $hammer_path, $hostname, %id, $line, $logdir, $logfile, $mode7,
        $mount_dir, $server, $srv_vol);

    if ( defined($main::MOUNT_VERS) ) {
        $opts{version} = $main::MOUNT_VERS;
    }

    if (!defined($opts{fc}) && !defined($opts{server})) {
        croak('nfs_hammer_start: Either the fc or server parameter must ' .
              'be specified.');
    }

    $Log->comment("nfs_hammer_start: It's NFS Hammer Time!");

    if (defined($ENV{LOGDIR})) {
        $logdir = $ENV{LOGDIR};
    } else {
        croak "Problem with NATE - LOGDIR not set in env";
    }

    # If a Server.pm object was specified then see if filer is 7 mode. If
    # so then use the filer connection from the Server.pm object. If a
    # Server.pm object was not passed in then filer is 7-mode.
    if (defined($opts{server})) {
        $opts{fc} = _server_to_fc($opts{server});
        $mode7 = _is_7_mode($opts{server});
    } else {
        $mode7 = 1;
    }

    $id{licensed} = 0;
    $id{client} = $opts{client};
    if ($mode7) {
        # 7 mode
        $id{fc} = $opts{fc};

        if (!is_licensed($opts{fc}, 'nfs')) {
            $id{licensed} = 1;
            license($opts{fc}, 'nfs');
            croak $@ if $@;
        }

        export_path(
                $opts{fc},
                "/vol/$opts{vol_name}",
                OPTIONS => 'rw,anon=0',
        ) or croak $@;
        $id{export} = "/vol/$opts{vol_name}";

        # Default the IP address to send traffic to the filer to be the
        # filer default IP address if not specified as an option.
        if (!defined($opts{filer_ip})) {
            $opts{filer_ip} = $opts{fc}->hostp->default_ip;
        }

        $srv_vol = "/vol/$opts{vol_name}";
    } else {
        # 10 mode
        $server = $id{server} = $opts{server};
        if (!defined($opts{filer_ip})) {
            # Grab the DATA_IP1 address for the NFS servers IP address
            %fields = $opts{fc}->hostp()->param_fields();
            if (!defined($fields{DATA_IP1})) {
                croak("Running hammer on 10-mode filer requires the DATA_IP1 " .
                 "parameter to be defined in the filers hostspec");
            }
            $opts{filer_ip} = $fields{DATA_IP1};
        }
        $srv_vol = "/$opts{vol_name}";
    }

    $hostname = $opts{fc}->hostp()->hostname();

    $id{runid} = ++$Nfs_hammer_runid;

    $mount_dir = mount(
        $opts{filer_ip},
        $srv_vol,
        client => $opts{client},
        chmod  => '0777',
        mntopts => "vers=$opts{version},proto=$opts{proto},noac,timeo=600," .
                   "rsize=$opts{rsize},wsize=$opts{wsize}",
    );

    if (!$mount_dir) {
        $err = $@;
        nfs_hammer_stop(id => \%id);
        croak $err;
    }
    $id{mountdir} = $mount_dir;

    # This will run hammer with its output going to $LOGDIR/nfs_hammer.*.txt
    # Stderr is not redirected, so if there is a problem, it is captured
    # and printed at the end of the subtest, and should appear in the
    # subtest log.

    $hammer_path = "$mount_dir/hammer_nfs_$id{runid}";
    $logfile = mktemp("$logdir/hammer_nfs.XXXXXX");
    $hammer_args = "$hammer_path";
    if ($opts{panic} eq 'yes') {
        $hammer_args .= " -filercore " . $hostname;
    } else {
        # Hammer on Linux needs the nofilercore option
        if (host($opts{client})->hosttype() =~ /linux/) {
            $hammer_args .= " -nofilercore ";
        }
    }

    $hammer_args .= " -mode $opts{mode} " if defined $opts{mode};
    $hammer_args .= " -runtime $opts{hammer_time} -blocksize random";

    $hammer_cmd = qq{\{
                      if (!-e \'$hammer_path\') {
                          mkdir \'$hammer_path\' or die 'mkdir failed';
                      }
                      qx($opts{hammer_bin} $hammer_args > $logfile);
                  \}};

    eval {
        $hammer_obj =
            subtest(
                $hammer_cmd,
                -bg,
                -client, $opts{client},
                -onexit, sub {},
            );
    };
    if ($@) {
        $err = $@;
        nfs_hammer_stop(id => \%id);
        croak ("Could not start hammer execution: $err");
    }
    $id{subtest} = $hammer_obj;
    $id{logfile} = $logfile;

    if ($mode7) {
        $err = _verify_hammer($opts{fc}, $mode7, $opts{verify_traffic},
                              $hammer_obj, 'NFS');
    } else {
        $err = _verify_hammer($opts{server}, $mode7, $opts{verify_traffic},
                              $hammer_obj, 'NFS');
    }
    if ($err) {
        nfs_hammer_stop(id => \%id);
        croak ($err);
    }

    return (\%id);
}

#---------------------------------------------------------------------

=head2 nfs_hammer_stop

    nfs_hammer_stop(id => $nfs_hamid);

Call this routine to stop hammer running on a nfs_client that
was started up with nfs_hammer_start(). Pass in the id that
was returned from the nfs_hammer_start() call. This routine
will kill the hammer process running on the nfs client, it
will umount the volume, and unexport the volume on the filer.
Typically this routine is called from a test cases cleanup routine.

=over

=item Arguments

=over

=item id

(Required) : Hash reference that contains the state associated with the hammer
invocation by nfs_hammer_start.

=item timeout

(Optional) : Number of seconds to wait for events to occur.
Defaults to 900 secs.

=item do_umount

(Optional) : Whether to perform umount after stopping hammer. Set to 1
to perform umount. Default is 1.

=back

=item Returns

Nothing. LogWarns on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub nfs_hammer_stop(%){
    my (%opts) = validate (
        @_,
        {
            id        => { type => HASHREF },
            timeout   => { type => SCALAR, default => 900 },
            do_umount => { type => SCALAR, default => 1 },
        }
    );
    my ($end_time, $file, $line, $pid, $runid, $sig_cmd, $sobj, $status);

    $Log->comment('nfs_hammer_stop: Stopping hammer');

    if (defined($opts{id}{subtest}) && $opts{id}{subtest}->is_alive()) {
        # Need to send a signal to the hammer process. Sending a signal
        # via subtest just kills the subtested process but the shell
        # that spawned hammer and hammer continue to run.
        $Log->comment('nfs_hammer_stop: Get hammers pid from the log file');

        # If hammer was just started, it may not have had a chance to
        # log its pid yet so we just poll for it.
        $pid = 0;
        $end_time = time() + $opts{timeout};
        while ($pid == 0 && time() < $end_time) {
            sysopen(IN, $opts{id}{logfile}, O_RDONLY) or
                $Log->warn("open failed: $opts{id}{logfile}: $!");
            while (($pid == 0) && ($line = <IN>)) {
                if ($line =~ /^path=/i) {
                    ($pid) = ($line =~ /pid=0x(\S+)/);
                    $pid = hex($pid);
                }
            }
            close IN;
            sleep(10) if ($pid == 0);
        }
        if ($pid != 0) {
            # Use subtest to send a signal to hammer. Using subtest because
            # if client is specified then the kill has to run remotely.
            # Using the UNIX kill command
            $sig_cmd = qq{\{
                           kill 15, $pid or die "Kill failed: $!";
                       \}};

            eval {
                $sobj =
                    subtest(
                        $sig_cmd,
                        -client, $opts{id}{client},
                        -onexit, sub {},
                    );
            };
            if (!$@) {
                $Log->comment("Verifying hammer stopped");
                $status = 0;
                $end_time = time() + $opts{timeout};
                while ($status == 0 && time() < $end_time) {
                    sysopen(IN, $opts{id}{logfile}, O_RDONLY) or
                        $Log->warn("open failed: $opts{id}{logfile}: $!");
                    while (($status == 0) && ($line = <IN>)) {
                        $status = 1 if ($line =~ /hammer stopped/i);
                    }
                    close IN;
                    # Sleep regardless of whether hammer stopped or not
                    # to give the OS a chance to cleanup the process.
                    sleep(10);
                }
                if ($status == 0) {
                    $Log->warn("Timed out waited $opts{timeout} secs for " .
                           "hammer to stop");
                }
            } else {
                $Log->warn("Could not send signal to hammer: $@");
            }
        } else {
            $Log->warn("Could not get hammers pid from logfile: $@");
        }

        # Save the hammer logfile to the subtest log directory
        if (-e $opts{id}{logfile}) {
            $runid = $opts{id}{runid};
            $file = "$opts{id}{subtest}->{logdir}/nfs_hammer_$runid.log";
            rename $opts{id}{logfile}, $file or
              $Log->warn("Rename $opts{id}{logfile} to $file failed");
        }
    }

    if ($opts{do_umount}) {
        if (defined($opts{id}{mountdir})) {
            umount($opts{id}{mountdir}, client => $opts{id}{client}, force => 1) or
              $Log->warn("Unable to unmount dir: $opts{id}{mountdir}: $@");
            $opts{id}{mountdir} = undef;
        }

        if (defined($opts{id}{export})) {
            unexport_path($opts{id}{fc}, $opts{id}{export}) or
              $Log->warn("Unable to unexport path: $opts{id}{export}: $@");
            $opts{id}{export} = undef;
        }
        return 1;
    } else {
        return 1;
    }
}

#---------------------------------------------------------------------

=head2 ontap_hammer_start

    ontap_hammer_start(
        fc          => $fc,
        vol_name    => $vol,
    );

Invokes a ONTAP version of hammer that runs locally on the filer.
Does not return any state because the filer command 'hammer abort'
doesn't allow specifying an instance of hammer to abort, it will
just abort all hammers running on the filer.

NOTE: The ONTAP version of hammer is only supported in 7 mode.

=over

=item Arguments

=over

=item fc

(Optional) : Connection object to a filer. Either the fc or server
parameter must be defined.

=item iterations

(Optional) : Number of read/write iterations to perform. Default
is '-1' which is infinite.

=item server

(Optional) : Server object for a filer. Either the fc or server
parameter must be defined.

=item vol_name

(Required) : Name of volume to hammer.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub ontap_hammer_start(%) {
    my (%opts) = validate (
        @_,
        {
            fc          => { type => OBJECT, optional => 1 },
            iterations  => { type => SCALAR, default => '-1'},
            server      => { type => OBJECT, optional => 1 },
            vol_name    => { type => SCALAR },
        }
    );
    my ($done, $end_time, $file, $runid, $text);

    if (!defined($opts{fc}) && !defined($opts{server})) {
        croak('ontap_hammer_start: Either the fc or server parameter must ' .
              'be specified.');
    }

    # If a Server.pm object was specified then see if we are 7 mode. If
    # so then use the filer connection from the Server.pm object.
    if (defined($opts{server})) {
        $opts{fc} = _server_to_fc($opts{server});
    }
    if (!defined($opts{fc})) {
        $Log->comment('ontap_hammer_start: Not started, only supported on 7-mode');
        return;
    }

    $Log->comment("ontap_hammer_start: It's ONTAP Hammer Time!");

    $runid = ++$Ontap_hammer_runid;
    $file = "/vol/$opts{vol_name}/hammer_file_$runid";

    # Use say instead of sayEOF since sayEOF chop_wouldbe chops off output
    # from the blacksmith (hammer).

    # hammer -f <# of iterations> <path of file> <# of BLOCK_SIZE blks in file>
    (undef, undef, $text) = say($opts{fc}, "priv set -q diag;".
                  "hammer -f $opts{iterations} $file 4096");
    if ($text =~ /already working on file/i) {
        croak ("Hammer already running on $file");
    }

    $Log->comment("Wait 60 secs for the hammer execution to start");
    $done = 0;
    $end_time = time() + 60;
    while (!$done && (time() < $end_time)) {
        (undef, undef, $text) = say($opts{fc},
                                    'priv set -q diag;hammer status');
        if ($text =~ /$file/) {
            $done = 1;
        } else {
            sleep(5);
        }
    }
    if (!$done) {
        ontap_hammer_abort(fc => $opts{fc});
        croak ("Hammer subtest did not continue running");
    }
    return;
}

#---------------------------------------------------------------------

=head2 ontap_hammer_abort

    ontap_hammer_abort(fc => $fc);

Call this routine to stop all hammers running on the filer.
Typically this routine is called from a test cases cleanup routine.

=over

=item Arguments

=over

=item fc

(Optional) : Connection object to a filer. Either the fc or server
parameter must be defined.

=item server

(Optional) : Server object for a filer. Either the fc or server
parameter must be defined.

=back

=item Returns

Nothing. LogWarns on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub ontap_hammer_abort(%) {
    my (%opts) = validate (
        @_,
        {
            fc     => { type => OBJECT, optional => 1 },
            server => { type => OBJECT, optional => 1 },
        }
    );
    if (!defined($opts{fc}) && !defined($opts{server})) {
        croak('ontap_hammer_abort: Either the fc or server parameter must ' .
              'be specified.');
    }

    # If a Server.pm object was specified then see if we are 7 mode. If
    # so then use the filer connection from the Server.pm object.
    if (defined($opts{server})) {
        $opts{fc} = _server_to_fc($opts{server});
    }
    if (!defined($opts{fc})) {
        $Log->comment('ontap_hammer_abort: Not aborted, only supported on 7-mode');
        return;
    }

    $Log->comment('ontap_hammer_abort: Abort all hammer jobs on the filer');
    eval {
        # No output is returned from a hammer abort
        say($opts{fc}, 'priv set -q diag;hammer abort');
    };
    $Log->warn($@) if ($@);
    return;
}
#---------------------------------------------------------------------

=head2 Gxontap_hammer_start

    Gxontap_hammer_start(
        server      => $server,
        vol_name    => $vol,
    );

Invokes a ONTAP version of hammer that runs locally on the filer.
Does not return any state because the filer command 'hammer abort'
doesn't allow specifying an instance of hammer to abort, it will
just abort all hammers running on the filer.

=over

=item Arguments

=over

=item server

(Optional) : Connection object to a filer. Server parameter must be defined.

=item iterations

(Optional) : Number of read/write iterations to perform. Default
is '-1' which is infinite.

=item vol_name

(Required) : Name of volume to hammer.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub Gxontap_hammer_start(%) {
    my (%opts) = validate (
        @_,
        {
            server      => { type => OBJECT, optional => 1 },
            iterations  => { type => SCALAR, default => '-1'},
            vol_name    => { type => SCALAR },
        }
    );
    my ($done, $end_time, $file, $runid, $text);

    if (!defined($opts{server})) {
        croak('ontap_hammer_start: Server parameter must ' .
              'be specified.');
    }

    $Log->comment("ontap_hammer_start: It's ONTAP Hammer Time!");

    $runid = ++$Ontap_hammer_runid;
    $file = "/vol/$opts{vol_name}/hammer_file_$runid";


    # run local hammer -f <# of iterations> <path of file> <# of BLOCK_SIZE blks in file>
    my $cmd = "set -priv diag;"."run local hammer -f $opts{iterations} $file 4096";
    my %prompt = (
            'Do you want to continue\? \{y\|n\}\:'  => 'y',
            "\\*>"                 => 'FINAL_PROMPT',
    );
    my $Server = $opts{server};
    $text = $Server->cli(cmd => $cmd,
                     additional_prompts_timeout => "60",
                          additional_prompts => \%prompt);
    if ($text =~ /already working on file/i) {
        croak ("Hammer already running on $file");
    }

    $Log->comment("Wait 60 secs for the hammer execution to start");
    $done = 0;
    $end_time = time() + 60;
    my $statuscmd = "set -priv diag;"."run local hammer status";
    while (!$done && (time() < $end_time)) {
        $text = $Server->cli(cmd => $statuscmd,
                          additional_prompts_timeout => "60",
                          additional_prompts => \%prompt);
        if ($text =~ /$file/) {
            $done = 1;
        } else {
            sleep(5);
        }
    }
    if (!$done) {
        croak ("Hammer subtest did not continue running");
    }
    return;
}

#---------------------------------------------------------------------

=head2 Gxontap_hammer_abort

    Gxontap_hammer_abort(server => $server);

Call this routine to stop all hammers running on the filer.
Typically this routine is called from a test cases cleanup routine.

=over

=item Arguments

=over

=item server

(Optional) : Server object for a filer. Server
parameter must be defined.

=back

=item Returns

Nothing. LogWarns on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub Gxontap_hammer_abort(%) {
    my (%opts) = validate (
        @_,
        {
            server => { type => OBJECT, optional => 1 },
        }
    );
    if (!defined($opts{server})) {
        croak('ontap_hammer_abort: Server parameter must ' .
              'be specified.');
    }

    my $Server = $opts{server};
    $Log->comment('ontap_hammer_abort: Abort all hammer jobs on the filer');
    my $abortcmd = "set -priv diag;"."run local hammer abort";
    eval {
        # No output is returned from a hammer abort
        my $out = $Server->cli(cmd => $abortcmd);
    };
    $Log->warn($@) if ($@);
    return;
}
#---------------------------------------------------------------------

1;
