# -*- perl -*-
#$Id:$
#
# Copyright (c) 2009 Network Appliance, Inc.
# All rights reserved.

## @summary 	Hybrid Buftree Library
## @description Provides functionality necessary to test Large Aggregate In
##		Place Upgrade.  This library contains the following functions:
##
##		aggr_convert - convert a 32-bit aggregate into a hybrid aggr
##
##		volume_[un]populate - [Un]Populate a volume with a dataset which
##				      is interesting to In Place Upgrade.
##
##		volume_is_hybrid - Is volume in a hybrid state?
##
##		volume_is_64bit - Is volume in 64bit?
##
##		file_hybridize - Touch portions of a file's buftree such that
##				 the buftree becomes hybrid.
##
##		hybridize_default_dataset - Hybridize default dataset.
##
##		file_buftree_check - Run the buftree checker on a file
## 
##		container_buftree_check - Run the buftree checker on a container file
##
##		buftree_check - Generic interface to the buftree checker
##		
## @author	dl-quercus-dev@netapp.com (The Larger Aggregates Project's In 
##					   Place Upgrade Team)
## @usage	use HybridBuftree;
##		

package HybridBuftree;

# Global Variables
# XXX Still need to define a proper default dataset.
# Default dataset file format
#	{ file      => "filename",
#	  size      => "#[b|B|k|K|m|M|g|G]",
#	  numfiles  => "#",
#	  step_size => "#[b|B|k|K|m|M|g|G]",
#	  timeout   => "#" }
#
my $Default_dataset_dir = "testdir";
my $Default_dataset = [
	{file => $Default_dataset_dir."/smallfile", size => "64"},
	{file => $Default_dataset_dir."/full_l1_file", size => "32768"},
	{file => $Default_dataset_dir."/full_l2_file", size => "16711680"}
];

use Exporter qw(import);
BEGIN {
    our @EXPORT = qw(aggr_convert
		hybridize_default_dataset
		volume_populate
		volume_unpopulate
		volume_is_hybrid
		volume_is_64bit
		file_hybridize
		aggr_is_convertible
		file_buftree_check
		container_buftree_check
		buftree_check
		buftree_stats_dump
		testme);
};

use strict;
use warnings;
use Tharn;
use Aggregate		qw(aggr_root aggr_exists aggr_status);
use Filer::Version	qw(get_wafl_version WV_BOILERMAKER_MAX);
use FS::Util::Setup	qw(mount_vol umount_vol);
use FS::Util::Utilities	qw(deltree util_exists util_mkdir util_mkfile
			   util_stat write_file);
use Module::Common	qw(validate_mixed);
use mount		qw(mounts);
use Params::Validate	qw(validate_pos ARRAYREF HASHREF SCALAR OBJECT);
use SPARSE::Util::Logger;
use Utils::Units	qw(convert_to_bytes);
use vol_db;
use Volume		qw(vol_exists);
use SnapCmds	qw(snapshot_create snapshot_status snapshot_delete);
use Utils qw(make_conn);
##
## @name        ::aggr_convert
## @arg 	$fc   - required: Filer connection
## @arg 	$aggr - required: aggregate to be converted
## @arg		$start_64bit_scan - optional: start 64bit upgrade scanner
## @description Takes an existing aggregate and makes it hybrid.  Aggregate
##		must be 32-bit before calling this function.
##
sub aggr_convert ( $@ ){
	my ($fc, $aggr, $start_64bit_scan) = validate_pos (
		@_,
		{ type => OBJECT },
		{ type => SCALAR },
		{ type => SCALAR, default => 0, optional => 1 }
	);

	# Verify that $aggr is in a state capable of being converted.
	if (volume_is_hybrid($fc, $aggr)) {
		die ("aggr '$aggr' is already hybrid");
	}

	if (!$start_64bit_scan) {
		(undef, undef, undef) = say($fc, 
		"priv set -q test; setflag wafl_64bit_upgrade_scan_disable 1");
	} else {
		(undef, undef, undef) = say($fc,
                "priv set -q test; setflag wafl_64bit_upgrade_scan_disable 0");
	}

	# Convert aggregate
	my $output = (say($fc, "priv set -q test; aggr 64bit-upgrade start ".
				"$aggr"))[2];
	if ($output =~ m/aggr: No such command/) {
        	logcomment("Couldn't find new aggr 64bit-upgrade command, trying the old aggr upgrade_format command");
		my $output = (say($fc, "priv set -q test; aggr upgrade_format ".
				"$aggr"))[2];
		if (!volume_is_hybrid($fc, $aggr) || $output !~ 
			m/Upgrade of aggregate format succeeded/) {
			die ("Aggr convert failed : '$output'");
		}
	}
	say($fc, "priv set -q test;aggr 64bit-upgrade status $aggr -all");
	if (!volume_is_hybrid($fc, $aggr)) {
		die ("Aggr 64bit-upgrade failed : '$output'");
	}
}

##
## @name        ::volume_populate
## @arg 	fc	   required -Filer connection
## @arg 	vol	   required - Volume 
## @arg 	file_list  optional - List of hashrefs containing the following
##				      fields:
##			file - required : filename to create
##			size - optional : default (2m) : size of file
##			numfiles - optional : default (1) 
##				   Number of files to create.  When numfiles is
##				   greater than one, files will be named
##				   <filename>.<#>
##			step_size - optional: default (1)
##				    The step_size specifies that mkfile should
##				    write one zero filled block for every 
##				    step_size blocks.  Thus, if you were to use
##				    a step size of '2', blocks 0,2,4,... would 
##				    be allocated and blocks 1,3,5,... would be
##				    holes.  For further documentation on
##				    step_size and other mkfile parameters,
##				    please see the mkfile man page.
##			timeout -   optional: default ($CMDTIMEOUT)
##				    When creating large files, it is
##				    recommended that this function be called
##				    with an appropriately large timeout value.
##				    The default timeout value is often too
##				    too small.
## @usage       volume_populate($fc, $vol, 
##			file_list => [
##				 {file=>'foo1', size=>20m},
##				 {file=>'dir1/dir2/foo2'},
##				 {file=>'dir1/foo3', size=>10g}
##			]);
## @usage	volume_populate($fc, $vol);
##
## @description Populates a 32-bit volume with datasets which are
## 		interesting for an upgrading volume/aggr.
## 		volume_populate should only be called on a 32-bit
## 		volume or aggregate.  The dataset definition is optional.  If
##		not given, a default dataset will be populated.
##
sub volume_populate( $$;% ) {
	my ($fc, $vol, %opts) = validate_mixed (
		@_,
		[ 
		  {type => OBJECT},
		  {type => SCALAR},
		],
		{
		  'file_list' => {type => ARRAYREF,
				  optional => 1 },
		}
	);

	my $die_string = 0;
	my $vol_mounted = &_vol_is_mounted($fc, $vol);
	my $mntpt = mount_vol(vol=>$vol) or die ("Mount of '$vol' failed:'$@'");
	my $list = (defined($opts{'file_list'})) ? $opts{'file_list'}
						 : $Default_dataset;

	foreach my $file_info (@$list) {
		my $filepath = $file_info->{'file'};

		# If size is not specified, a default size of '2m'
		# is used.
		my $size      = (defined($file_info->{'size'})) ?
				$file_info->{'size'} : '2m';
		my $step_size = (defined($file_info->{'step_size'})) ?
				$file_info->{'step_size'} : '1';
		my $numfiles  = (defined($file_info->{'numfiles'})) ?
				$file_info->{'numfiles'} : '1';
		my $timeout   = (defined($file_info->{'timeout'})) ?
				$file_info->{'timeout'} :
				$main::CMDTIMEOUT;
                my $client = ($::CLIENT || 'localhost') ;
                my $cc = make_conn($client);
		if ($filepath =~ m/(.*)\/.*/) {
			my $dir_path = $1;
			if (!util_mkdir($mntpt.'/'.$dir_path, conn => $cc)) {
				$die_string = "Unable to create ".
					      "'$mntpt/$dir_path'";
				goto END;
			}
		}
		say ($fc ,"priv set test;ls /vol/$vol/");
		if (!write_file(filer => $fc,
			file  => "/vol/$vol/".$filepath,
			size  => $size,
			numfiles => $numfiles,
			step_size => $step_size,
			timeout => $timeout)) {
			    $die_string = "Unable to create '/".
				          "vol/$vol/$filepath'";
			    goto END;
		}
	}

END:
	if (!$vol_mounted) {
		umount_vol($mntpt);
	}

	if ($die_string) {
		die ($die_string);
	}
}

##
## @name        ::volume_unpopulate
## @arg 	fc	  required - Filer connection
## @arg 	vol	  required - Volume 
## @arg 	file_list optional - List of hashrefs containing the files to
##				     be deleted.
## @usage       volume_unpopulate($fc, $vol, 
##			file_list => 
##				[{file=>'foo1'},
##				 {file=>'dir1/dir2/foo2'},
##				 {file=>'dir1/foo3}
##				]);
## @usage	volume_unpopulate($fc, $vol);
##
## @description Destroys files populated by volume_populate.
## 		The dataset definition is optional.  If
##		not given, the default dataset will be removed.
##
sub volume_unpopulate( $$;% ) {
	my ($fc, $vol, %opts) = validate_mixed (
		@_,
		[ 
		  {type => OBJECT},
		  {type => SCALAR},
		],
		{
		  'file_list' => {type => ARRAYREF,
				  optional => 1 },
		}
	);
	my $die_string = 0;
	my $vol_mounted = &_vol_is_mounted($fc, $vol);
        my $mntpt = mount_vol(vol=>$vol) or die("Mount of '$vol' failed: '$@'");

	my $list = (defined($opts{'file_list'})) ? $opts{'file_list'}
						 : $Default_dataset;

	# Cleaning up a specified dataset
	foreach my $file_info (@$list) {
		my $file = $file_info->{'file'};

		# Strip any preceding '/' characters
		$file =~ s/\/*(.*)/$1/;

		# Find the top component of the filepath
		$file =~ s/(.*?)\/.*/$1/;
                my $client = ($::CLIENT || 'localhost') ;
                my $cc = make_conn($client);

		if (util_exists($mntpt.'/'.$file, conn => $cc)) {
			if (!deltree($mntpt.'/'.$file, conn => $cc)) {
				$die_string = "Deltree failed for ".
					      "'$mntpt/$file' with error = $@";
				goto END;
			}
		}
	}

END:
	if (!$vol_mounted) {
		umount_vol($mntpt);
	}

	if ($die_string) {
		die ($die_string);
	}
}

##
## @name	::file_hybridize
##
## @arg		fc required - Filer connection
## @arg		vol required - Volume
## @arg		file required - File to hybridize
## @arg		level optional
##		Level of the buftree where this function will operate.  If 
##		level is not specified, we operate on level zero buffers.
##		The default "level zero" behavior will result in an upgrade of
##		the L0 buffer's ancestry.
## @arg		fbn-start optional
##		When hybridizing the buftree, we use 'wafl touch' to 'touch' or
## 		'fake dirty' a particular buffer in the buftree. fbn-start 
##		designates which fbn to start hybridizing.  If fbn-start is
##		not specified, hybridization will begin at the beginning of the
##		file.
## @arg		fbn-end optional
##		The fbn where hybridization should stop.  end-fbn must be
##		greater than or equal to start-fbn.  If end-fbn is not given,
##		hybridization will continue to the end of the file.
## @arg		fbn-skip optional
##		The number of fbns to skip between each hybridization.  For 
##		example: if fbn-start is 1000 and fbn-skip is 5, the following
##		fbns will by hybridized {1000, 1005, 1010, 1015, ... end_fbn}
##		If fbn-skip is not specified, the following default 
##		skips values will be used.
##			Level	Default_Skip
##		 	  0         2048
##		 	  1	    2097152
##		 	  2	    2147483648
##		 	  3+	    4286529242
##		***These numbers come from 2 x span(level+1) for a traditional
##		   32-bit format aggr/volume and should result in a
##		   partial upgrade of the buftree. The skip tops out at
##		   '4286529242' blocks which is equivalent to '16TB'. 
## @usage	file_hybridize($fc, $vol, "dir/dir2/file2",
##			       level=>"2", start-fbn=>"10", end-fbn=>"10000",
##			       fbn-skip=>255);
## @description	This routine will take a file and touch various portions of the
##		buftree (as specified by the input parameters) in order to
##		create a hybrid buftree (a buftree with both 32-bit and 64-bit
##		indirect blocks).  This function has no hybridization effects on
##		pure 32-bit/64-bit volumes.
##
sub file_hybridize ($$$;%){
	my ($fc, $vol, $file, %opts) = validate_mixed (
		@_,
		[ 
		  {type => OBJECT},
		  {type => SCALAR},
		  {type => SCALAR},
		],
		{
		  'level' => {type => SCALAR, default => 0, optional => 1 },
		  'fbn-start' => {type => SCALAR, optional => 1},
		  'fbn-end' => {type => SCALAR, optional => 1},
		  'fbn-skip' => {type => SCALAR, optional => 1},
		}
	);

	my $vol_mounted = &_vol_is_mounted($fc, $vol);
	my $die_string = 0; 

	# Verify that we are operating on a hybrid volume.
	if (!volume_is_hybrid($fc, $vol)) {
		$die_string = "HybridBuftree::file_hybridize is not intended ".
			      "for non-hybrid volumes.";
		goto END;
	}

	# Verify that $file exists
	my $mntpt = mount_vol(vol=>$vol) or die("Mount of '$vol' failed: '$@'");
	my $file_name = $mntpt."/".$file;
        my $client = ($::CLIENT || 'localhost') ;
        my $cc = make_conn($client);
	if (!util_exists($file_name, conn=>$cc)) {
		$die_string = "File '$file' does not exist.";
		goto END;
	}
	my @stats = util_stat($file_name, conn => $cc);

	# Setup fbn_end
	my $fbn_end;
	if (defined($opts{'fbn-end'})) {
		$fbn_end = $opts{'fbn-end'};
	} else {
		#
		# fbn-end was not passed in as a param.  Compute it based on
		# the size of the file.
		#
		my $file_size = $stats[7];
		$fbn_end = POSIX::floor($file_size / 4096);
	}

	# Setup fbn_start
	my $fbn_start   = defined($opts{'fbn-start'})? $opts{'fbn-start'}:  0;

	# Setup level
	my $level       = defined($opts{level})    ? $opts{level}    :  0;
	if ($level > 7) {
		$die_string = "Level is not valid: current value '$level'.";
		goto END;
	}

	# Setup fbn_skip
	my $fbn_skip;
	if (defined($opts{'fbn-skip'})) {
		$fbn_skip = $opts{'fbn-skip'};
	} else {
		# Compute the approriate skip level based on level.
		my @skips = (2048, 2097152, 2147483648, 4286529242, 4286529242,
		             4286529242, 4286529242, 4286529242);
		$fbn_skip = $skips[$level];
	}

	# Map '$file' to inode number.
	my $inodenum = $stats[1];

	# Issue wafl touch command.
	my $output = (say($fc, "priv set -q test; wafl touch $vol $inodenum ".
			       "$level $fbn_start $fbn_end $fbn_skip"))[2];
	if ($output =~ m/wafl touch:/) {
		 $die_string = $output;
		 goto END;
	}

END:

	if (!$vol_mounted) {
		umount_vol($mntpt);
	}

	if ($die_string) {
		die ($die_string);
	}
}

##
## @name 	 hybridize_default_dataset
## @description  Issues the necessary commands to hybridize the default dataset
##		 as defined in $Default_dataset
## @arg		 fc - required : Filer connection
## @arg		 vol - required : Volume
## @usage	 hybridize_default_dataset($fc, $vol);
sub hybridize_default_dataset( $$ ) {
	my ($fc, $vol) = validate_pos(
		@_,
		{ type => OBJECT },
		{ type => SCALAR }
	);

	foreach my $file_descr (@$Default_dataset) {
		file_hybridize($fc, $vol, $file_descr->{'file'});
	}
}

##
## @name	::volume_is_hybrid
## @description	Returns true if volume is hybrid, false otherwise.
## @arg		fc - required : Filer connection
## @arg		vol - required : Volume
## @usage	volume_is_hybrid($fc, $vol);
## @returns	0: Volume is not hybrid
##		1: Volume is hybrid
##
sub volume_is_hybrid( $$ ) {

	my ($fc, $vol) = validate_pos (
		@_,
		{ type => OBJECT },
		{ type => SCALAR }
	);

        my $output = (say($fc, "priv set -q diag; vol_db iter"))[2];

        my @vols_hash = split(/\n/, $output);

        foreach my $vol_hash (@vols_hash) {
                my $name = (say($fc, "priv set -q diag; vol_db get name ".
                                     "$vol_hash"))[2];

                if ($name =~/^\tname: $vol\s+/) {
                        my $buftree_type = (say($fc, "priv set -q diag; ".
                                                "vol_db get buftree_type ".
                                                "$vol_hash"))[2];
                        if ($buftree_type =~ m/upgrading/) {
                                return 1;
                        }
                        return 0;
                }
        }
        return 0;


# XXX We are facing some problems with Server.pm which is used by vol_db.pm.
# Until these problems are fixed, leaving this code commented out.
#
# 	Create a new vol_db object
#	my $vdb = vol_db->new(host => $fc->hostp());

	# Grab the vol_db handle for '$vol'
#	my ($vdb_handle, $error) = $vdb->lookup(cmd_args => "name '$vol'");
#	if ($error) {
#		die ("vol_db lookup failed : '$error'");
#	}

	# Check to see if buftree_type is upgrading.
#	my ($buftree_type) = $vdb->get(attr => 'buftree_type',
#				       handle => $vdb_handle);
#	if ($buftree_type =~ m/upgrading/) {
#		return 1;
#	}
#	return 0;
}

##
## @name        ::volume_is_64bit
## @description Returns true if volume is 64bit (or hybrid), false otherwise.
## @arg         fc - required : Filer connection
## @arg         vol - required : Volume
## @usage       volume_is_64bit($fc, $vol);
## @returns     0: Volume is not 64bit 
##              1: Volume is 64bit (or hybrid) 
##
sub volume_is_64bit( $$ ) {

        my ($fc, $vol) = validate_pos (
                @_,
                { type => OBJECT },
                { type => SCALAR }
        );

        my $output = (say($fc, "priv set -q diag; vol_db iter"))[2];

        my @vols_hash = split(/\n/, $output);

        foreach my $vol_hash (@vols_hash) {
                my $name = (say($fc, "priv set -q diag; vol_db get name ".
                                     "$vol_hash"))[2];

                if ($name =~/^\tname: $vol\s+/) {
                        my $buftree_type = (say($fc, "priv set -q diag; ".
                                                "vol_db get buftree_type ".
                                                "$vol_hash"))[2];
                        if ($buftree_type =~ m/dual64/ ||
			    $buftree_type =~ m/trad64/) {
                                return 1;
                        }
                        return 0;
                }
        }
        return 0;
}

##
## @name	aggr_is_convertible
## @description	Returns true if aggregate could be converted to hybrid,
##               false otherwise.
## @arg 	fc - required : Filer connection
## @arg		aggr - required : Aggregate to check
## @usage	aggr_is_convertible($fc, $aggr);
## @returns	0: Aggregate cannot be converted
##		1: Aggregate can be converted to hybrid
sub aggr_is_convertible($$) {
	my ($fc, $aggr) = validate_pos (
		@_,
		{ type => OBJECT },
		{ type => SCALAR }
	);

	if (get_wafl_version($fc->hostp()) <= WV_BOILERMAKER_MAX) {
		# aggregate conversion not supported in BR or earlier
		return 0;
	}
	if (aggr_root($fc) eq $aggr) {
		# root aggregate currently cannot be converted
		return 0;
	}
	if (!aggr_exists($fc, $aggr)) {
		# can't convert what doesn't exist
		return 0;
	}
	my $statusobj = aggr_status($fc, $aggr);
	if ($statusobj->status() =~ /32-bit/) {
		return 1;
	} else {
		# already either pure or hybrid 64-bit
		return 0;
	}
}

#
# @name    ::helper_snapshot_snapid
# @arg     fc        required - Filer connection
# @arg     vol       required - Volume or Aggregate
# @arg     snapname  required - Snapshot name
# @returns                      $snapid : $snapname exists
#                               undef   : $snapname does not exist
# @usage   my $snapid = helper_snapshot_snapid($fc, $vol, $snapname)
# @description	Returns the $snapid of the specified snapshot
#
sub helper_snapshot_snapid($$$) {
	my ($fc, $vol, $snapname) = validate_pos (
		@_,
		{ type => OBJECT },
		{ type => SCALAR },
		{ type => SCALAR },
	);

	my @snapshots = snapshot_status($fc, $vol);

	foreach my $snapshot (@snapshots) {
		if ($snapname eq $snapshot->{name}) {
			return $snapshot->{snapid};
		}
	}

	return undef;
}

#
# @name    ::helper_snapshot_create
# @arg     fc        required - Filer connection
# @arg     vol       required - Volume or Aggregate
# @arg     snapname  required - Snapshot name
# @returns                      snapid of new snapshot
# @usage   my $snapid = helper_snapshot_create($fc, $vol, $snapname)
# @description	Creates a snapshot and returns its $snapid
#
sub helper_snapshot_create($$$) {
	my ($fc, $vol, $snapname) = validate_pos (
		@_,
		{ type => OBJECT },
		{ type => SCALAR },
		{ type => SCALAR },
	);

	snapshot_create($fc, $vol, $snapname) or
		die "Failed to create snapshot $snapname in $vol";

	my $snapid = helper_snapshot_snapid($fc, $vol, $snapname);
	if (!$snapid) {
		die "Can't get snapid for snapshot $snapname in $vol";
	}

	return $snapid;
}

##
## @name    ::buftree_check
##
## @arg     fc        required - Filer connection
## @arg     vol       required - Volume or Aggregate
## @arg     inodespec required - Inode specification; one of:
##                                    <inum>
##                                        public inode
##                                    priv:<inum>
##                                        private inode
##                                    vvol:<vvolname>/label
##                                        vvol raid label
##                                    vvol:<vvolname>/container
##                                        vvol container file
##                                    vvol:<vvolname>/uuid
##                                        vvol uuid dir
##                                    vvol:<vvolname>/uuid/<pathname>
##                                        inode under vvol uuid dir
##                                    path:<pathname>
##                                        inode under rootdir
##                                    meta:<pathname>
##                                        inode under metadir
## @arg     is_aggr   optional - Whether "vol" is an Aggregate;
##                               Defaults to 0 (false, not an aggr)
## @arg     snapid    optional - Snapshot ID; Defaults to 0 (AFS)
## @arg     force_afs optional - By default, if snapid is not specified,
##                               checking occurs on a temporary snapshot 
##                               created (and deleted) inside this function.
##                               If the buftree is modified during the check
##                               the buftree checker's wi_blk_cnt check
##                               would be skipped.  Doing the check on a
##                               snapshot guards against this.
##                               Setting force_afs to 1 prevents the 
##                               creation of this temporary snapshot
##                               and forces checking to occur on the
##                               active filesystem.
##                               Defaults to 0 (auto-create snapshot).
## @arg     dont_panic optional- By default, the filer will be panic if the
##                               buftree checker finds an inconsistency.
##                               This option turns off the panicing.
##                               Defaults to 0 (panic on inconsistency).
## @arg     timeout    optional - Timeout in seconds.
## @returns                      A buftree stats object, which looks like:
##    my $stats = {
##    	levels => [
##    		{ blks32 => 0, holes32 => 0, blks64 => 0, holes64 => 0, total => 0 },
##          ... additional levels as required ...
##    		{ blks32 => 0, holes32 => 0, blks64 => 0, holes64 => 0, total => 0 },
##    	],
##    	total => 
##    		{ blks32 => 0, holes32 => 0, blks64 => 0, holes64 => 0, total => 0 },
##    };
##    Note that the level with the largest index refers to the embedded data
##    or indirect within the inode itself.
## @usage   buftree_check($fc, $vol, "path:/my/path/to/boxmanager");
##          buftree_check($fc, $vol, "priv:-1");
##          my $stats = buftree_check($fc, $aggr, "vvol:$vvol/label");
##          my $wi_blk_count =   $stats->{total}->{blks32}
##                             + $stats->{total}->{blks64}
##                             - 1;   # minus the inode "block"
##          my $wi_level = scalar(@{$stats->{levels}}) - 1;
##          my $lev0_blks =   ($wi_level == 0) 
##                          ? 0
##                          : (  $stats->{levels}->[0]->{blks32}
##                             + $stats->{levels}->[0]->{blks64});
##          my $wi_indirect_blk_count = $wi_blk_count - $lev0_blks;
##          my $inode_is_32bit = ($stats->{levels}->[$wi_level]->{blks32} > 0);
## @description	Run the buftree checker on the specified inode.
##              Returns buftree stats.
##              Note that a check failure will panic the filer.
##              More documentation on the buftree checker at:
## http://wikid.netapp.com/w/WAFL/Larger_Aggregates/Mile3Tasks/TD/BuftreeCheck
##
sub buftree_check($$$;%) {
	my ($fc, $vol, $inodespec, %opts) = validate_mixed(
		@_,
		[
			{ type => OBJECT },
			{ type => SCALAR },
			{ type => SCALAR },
		],
		{
			'snapid' => { type => SCALAR, default => 0, optional => 1 },
			'is_aggr' => { type => SCALAR, default => 0, optional => 1 },
			'force_afs' => { type => SCALAR, default => 0, optional => 1 },
			'dont_panic' => { type => SCALAR, default => 0, optional => 1 },
			'timeout' => { type => SCALAR, default => 0, optional => 1 },
		}
	);

	my $arg_vol = ($opts{is_aggr} ? "-A $vol" : "$vol");

	my $snapid = (defined($opts{snapid}) ? $opts{snapid} : 0);
	my $tmp_snapname;

	if (!$snapid && !$opts{force_afs}) {
		$tmp_snapname = "buftree_check_tmp";
		$snapid = helper_snapshot_create($fc, $arg_vol, $tmp_snapname);
	}

	my $opt_snapid = ($snapid ? "-s $snapid" : "");

	my $opt_dont_panic = ($opts{dont_panic} ? "-n" : "");

	my $cmd_str = 
		"wafl buftree check $opt_snapid $opt_dont_panic $arg_vol $inodespec";

	my $full_cmd_str = "priv set -q test; $cmd_str";
	my @full_output;

	if ($opts{timeout}) {
		@full_output = say($fc, $full_cmd_str, -timeout, $opts{timeout});
	}
	else {
		@full_output = say($fc, $full_cmd_str);
	}

	my $output = $full_output[2];

	if (defined($tmp_snapname)) {
		snapshot_delete($fc, $arg_vol, $tmp_snapname) or
			die "Failed to delete snapshot $tmp_snapname";
	}

	if ($output =~ m/wafl buftree check: Error:/) {
		die ($output);
	}

	return buftree_stats_parse(split(/\n/, $output));
}

#
# @name    ::buftree_stats_empty
# @returns                         A buftree stats object.
#                                  See ::buftree_check for details.
# @usage   my $buftree_stats = buftree_stats_empty();
# @description	Return empty buftree stats object
#
sub buftree_stats_empty
{
	return { levels => [ ] };
}

#
# @name    ::buftree_stats_parse
# @arg     stats_lines  required - Array of lines of output from
#                                  "wafl buftree check"
# @returns                         A buftree stats object.
#                                  See ::buftree_check for details.
# @usage   my $buftree_stats = buftree_stats_parse(@stats_lines);
# @description	Parses output of "wafl buftree check"
#
sub buftree_stats_parse($@)
{
	my @stats_lines = @_;
	my $stats = buftree_stats_empty();

	# parse this:
	#
	# ----- ------------ ------------ ------------ ------------ ------------
	# level 32-bit blks  32-bit holes 64-bit blks  64-bit holes     total
	# ----- ------------ ------------ ------------ ------------ ------------
	#     3            0            0            1            0            1
	#     2            0            0            2            1            3
	#     1            0            0            4          509          513
	#     0            0            0           54       130515       130569
	# total            0            0           61       131025       131086
	# ----- ------------ ------------ ------------ ------------ ------------
	#
	# This is followed by an empty line, and then with extra tables
	# in the verbose (-v) output.

	foreach (@stats_lines) {
		last if /^$/;
		next if ! /^\s*(\d+|[^\s]+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)$/;

		my $level = $1;    # 0, 1, 2, ... or "total"

		my $statlev = {
			blks32 => $2,
			holes32 => $3,
			blks64 => $4,
			holes64 => $5,
			total => $6,
		};

		if ($level =~ /^\d+$/) {
			$stats->{levels}->[$level] = $statlev;
		}
		else {
			$stats->{$level} = $statlev;
		}
	}

	return $stats;
}

##
## @name    ::buftree_stats_dump
## @arg     stats        required - A buftree stats object.
##                                  See ::buftree_check for details.
## @usage   buftree_stats_dump($stats);
## @description	Dumps buftree stats object to standard output.
##
sub buftree_stats_dump($)
{
	my ($stats) = validate_pos(
		@_,
		{ type => HASHREF },
	);

	printf "----- ------------ ------------ ------------ ------------ ------------\n";
	printf "level 32-bit blks  32-bit holes 64-bit blks  64-bit holes     total   \n";
	printf "----- ------------ ------------ ------------ ------------ ------------\n";

	my $statlev;

	# Note that the rows are printed in decreasing level order,
	# with the "total" row at the end.
	for (my $level = scalar(@{$stats->{levels}}) - 1; 
		$level >= 0; 
		$level = $level - 1) {
		$statlev = $stats->{levels}->[$level];
		next if not $statlev;
		printf "%5u %12llu %12llu %12llu %12llu %12llu\n",
			$level,
			$statlev->{blks32}, $statlev->{holes32},
			$statlev->{blks64}, $statlev->{holes64},
			$statlev->{total};
	}

	$statlev = $stats->{total};
	if ($statlev) {
		printf "%5s %12llu %12llu %12llu %12llu %12llu\n",
			"total",
			$statlev->{blks32}, $statlev->{holes32},
			$statlev->{blks64}, $statlev->{holes64},
			$statlev->{total};
	}

	printf "----- ------------ ------------ ------------ ------------ ------------\n";
	printf "\n";
}

##
## @name    ::file_buftree_check
##
## @arg     fc         required - Filer connection
## @arg     vol        required - Volume
## @arg     file       required - File to check
## @arg     snapid     optional - Snapshot ID, defaults to AFS
## @arg     force_afs  optional - Force checking on AFS.
##                                See ::buftree_check for details.
## @arg     dont_panic optional - Don't panic on inconsistency.
##                                See ::buftree_check for details.
## @arg     timeout    optional - Timeout in seconds.
## @returns                       A buftree stats object.
##                                See ::buftree_check for details.
## @usage   file_buftree_check($fc, $vol, "dir/file", snapid => 1);
## @description Run the buftree checker on the specified file.
##      See ::buftree_check for more info.
##
sub file_buftree_check($$$;%) {
	my ($fc, $vol, $file, %opts) = validate_mixed(
		@_,
		[
			{ type => OBJECT },
			{ type => SCALAR },
			{ type => SCALAR },
		],
		{
			'snapid' => { type => SCALAR, default => 0, optional => 1 },
			'force_afs' => { type => SCALAR, default => 0, optional => 1 },
			'dont_panic' => { type => SCALAR, default => 0, optional => 1 },
			'timeout' => { type => SCALAR, default => 0, optional => 1 },
		}
	);
	return buftree_check($fc, $vol, "path:/$file", %opts);
}

##
## @name    ::container_buftree_check
##
## @arg     fc         required - Filer connection
## @arg     aggr       required - Aggregate
## @arg     vvol       required - Flexible Volume
## @arg     snapid     optional - Snapshot ID, defaults to AFS
## @arg     force_afs  optional - Force checking on AFS.
##                                See ::buftree_check for details.
## @arg     dont_panic optional - Don't panic on inconsistency.
##                                See ::buftree_check for details.
## @arg     timeout    optional - Timeout in seconds.
## @returns                       A buftree stats object.
##                                See ::buftree_check for details.
## @usage   container_buftree_check($fc, $aggr, $vvol, snapid => 1);
## @description Run the buftree checker on the specified container file.
##          See ::buftree_check for more info.
##
sub container_buftree_check($$$;%) {
	my ($fc, $aggr, $vvol, %opts) = validate_mixed(
		@_,
		[
			{ type => OBJECT },
			{ type => SCALAR },
			{ type => SCALAR },
		],
		{
			'snapid' => { type => SCALAR, default => 0, optional => 1 },
			'force_afs' => { type => SCALAR, default => 0, optional => 1 },
			'dont_panic' => { type => SCALAR, default => 0, optional => 1 },
			'timeout' => { type => SCALAR, default => 0, optional => 1 },
		}
	);
	return buftree_check($fc, $aggr, "vvol:$vvol/container", %opts, is_aggr => 1);
}

#
# @name 	testme
# @description	Test basic HybridBuftree Library functionality.
# @usage	ntest -hw PKGNAME=HybridBuftree SUBNAME=testme FILER=<filer>
#		   util/test_pkg.thpl
sub testme() {

    use Aggregate 	qw(aggr_exists aggr_purge aggr_status aggr_root);
    use FS::Util::Setup	qw(create_vol);
    use Volume		qw(vol_exists vol_purge);
    use vars qw($FILER $FILER_CONN $FILER_AGGR $FILER_VOL);
    require "FS/commonParams.pl";

    eval {
	logcomment("Performing self test of HybridBuftree Library");

	my $fh = host($main::FILER);
	my $fc = connect($main::FILER_CONN);

	# Remove existing volume and aggregate
	logcomment("Destroying '$main::FILER_VOL' and $main::FILER_AGGR'.");
	vol_purge($fc, $main::FILER_VOL) or logWarn("'$main::FILER_VOL not ".
						    "destroyed :'$@'");
	aggr_purge($fc, $main::FILER_AGGR) or logWarn ("'$main::FILER_AGGR' ".
						    " not destroyed :'$@'");

	# Create aggregate and volume in 32-bit mode.
	logcomment("Create aggregate and volume.");
	create_vol(vol_size => '100g')
		or die ("Unable to create aggregate/volume: $@");
	if (!aggr_exists($fc, $main::FILER_AGGR) ||
	    !vol_exists($fc, $main::FILER_VOL)) {
		die ("Aggr or Vol not created successfully.");
	}

	# Populate volume
	logcomment("Populate volume '$main::FILER_VOL'");
	volume_populate($fc, $main::FILER_VOL);

	# Populate volume with specific files.
	logcomment("Populate volume '$main::FILER_VOL' with a specific ".
		   "dataset.");
	volume_populate($fc, $main::FILER_VOL, 
		    file_list => [
			{file => 'filea'},
			{file => 'fileb', size => '1m', step_size => '4'},
			{file => 'dir1/filec'},
			{file => 'dir1/filee', size => '150g',
				step_size => '26214400', timeout => '1200'},
			{file => 'dir2/filed', numfiles => 5, size => '20'}
		    ]);

	# Create snapshots before conversion
	my $testaggr_snapname = "testaggr_snapname";
	my $testaggr_snapid =
			helper_snapshot_create($fc, "-A $main::FILER_AGGR", $testaggr_snapname);
	my $testvol_snapname = "testvol_snapname";
	my $testvol_snapid =
			helper_snapshot_create($fc, $main::FILER_VOL, $testvol_snapname);

	# Check aggr_is_convertible
	logcomment("Check aggr_is_convertible with a 32-bit aggr.");
	if (!aggr_is_convertible($fc, $main::FILER_AGGR)) {
		die ("A 32-bit aggr should be convertible.");
	}

	# Convert aggr
	logcomment("Convert aggr to hybrid mode.");
	aggr_convert($fc, $main::FILER_AGGR);

	# Check aggr_is_convertible again
	logcomment("Check aggr_is_convertible with a 64-bit aggr.");
	if (aggr_is_convertible($fc, $main::FILER_AGGR)) {
		die ("A 64-bit aggr should not be convertible.");
	}

	# Hybridize Default Files
	logcomment("Hybridizing default file list.");
	hybridize_default_dataset($fc, $main::FILER_VOL);

	# Hybridize Specific Files
	logcomment("Hybridizing specific file list.");
	file_hybridize($fc, $main::FILER_VOL, 'filea');
	file_hybridize($fc, $main::FILER_VOL, 'fileb', level => 2,
		       'fbn-start' => 3, 'fbn-end' => '300',
		       'fbn-skip' => 10);
	file_hybridize($fc, $main::FILER_VOL, 'dir1/filec',
	    	       'fbn-skip' => 5);
	file_hybridize($fc, $main::FILER_VOL, 'dir1/filee', level => 4,
		      'fbn-start' => 2048);
	eval {
		file_hybridize($fc, $main::FILER_VOL, 'dir2/filed.0', level=>8);
	};
	if ($@ !~ m/Level is not valid/) {
		die ("'$@'");
	}

	# Run the buftree on non-existent file
	eval {
		file_buftree_check($fc, $main::FILER_VOL, 'bad_filename');
	};
	if ($@ !~ m/wafl buftree check: Error:/) {
		die ("Expecting error on non-existent file");
	}

	# run the buftree checker on filea
	my $stats = file_buftree_check($fc, $main::FILER_VOL, 'filea');
	buftree_stats_dump($stats);
	my $wi_level = scalar(@{$stats->{levels}}) - 1;
	my $inode_is_64bit = ($stats->{levels}->[$wi_level]->{blks64} > 0);
	if (not $inode_is_64bit) {
		die ("filea inode should be 64-bit");
	}
	if ($stats->{total}->{blks32} == 0 or $stats->{total}->{blks64} == 0) {
		die ("filea should be hybrid");
	}
	# filea  =>  2 MB  =>  512 L0s  =>  3 L1s  =>  L2 inode
	if ($wi_level != 2) {
		die ("filea has wrong level $wi_level");
	}
	# assume no one is changing filea right now
	my $stats2 = file_buftree_check($fc, $main::FILER_VOL, 'filea',
					force_afs => 1);
	if ($stats->{total}->{blks32} + $stats->{total}->{blks64} !=
	    $stats2->{total}->{blks32} + $stats2->{total}->{blks64}) {
		die ("filea block count has changed");
	}
	$stats = file_buftree_check($fc, $main::FILER_VOL, 'filea', 
					snapid => $testvol_snapid);
	$inode_is_64bit = ($stats->{levels}->[$wi_level]->{blks64} > 0);
	if ($inode_is_64bit) {
		die ("filea inode in snapshot should not be 64-bit");
	}

	# run buftree checker on container file
	$stats = container_buftree_check($fc, $main::FILER_AGGR, $main::FILER_VOL, timeout => 300);
	$wi_level = scalar(@{$stats->{levels}}) - 1;
	$inode_is_64bit = ($stats->{levels}->[$wi_level]->{blks64} > 0);
	if (not $inode_is_64bit) {
		die ("container inode should be 64-bit");
	}
	$stats = container_buftree_check($fc, $main::FILER_AGGR, $main::FILER_VOL,
					snapid => $testaggr_snapid, timeout => 300);
	$inode_is_64bit = ($stats->{levels}->[$wi_level]->{blks64} > 0);
	if ($inode_is_64bit) {
		die ("container inode in snapshot should not be 64-bit");
	}

	# Unpopulate volume - default dataset
	logcomment("Unpopulate default dataset.\n");
	volume_unpopulate($fc, $main::FILER_VOL);

	# Unpopulate volume - specific dataset
	logcomment("Unpopulate specific dataset.\n");
	volume_unpopulate($fc, $main::FILER_VOL,
			  file_list => [
				{file => 'filea'},
				{file => 'fileb'},
				{file => '/dir1/filec'},
				{file => 'dir1/filee'},
				{file => 'dir2/filed'}
			  ]);

	# Remove hybrid volume and aggregate
	logcomment("Destroying '$main::FILER_VOL' and $main::FILER_AGGR'.");
	vol_purge($fc, $main::FILER_VOL) or logWarn("'$main::FILER_VOL not ".
						    "destroyed :'$@'");
	aggr_purge($fc, $main::FILER_AGGR) or logWarn ("'$main::FILER_AGGR' ".
						    " not destroyed :'$@'");

	# Check aggr_is_convertible with a nonexistent aggr, and the root aggr
	logcomment("Check aggr_is_convertible with a nonexistent aggr.");
	if (aggr_is_convertible($fc, $main::FILER_AGGR)) {
		die ("A nonexistent aggr should not be convertible.");
	}
	
	logcomment("Check aggr_is_convertible with the root aggr.");
	if (aggr_is_convertible($fc, aggr_root($fc))) {
		die ("The root aggr should not be convertible.");
	}

    };
    if ($@) {
    	logresult('FAIL', "$@");
    }
}

sub _vol_is_mounted( $@ ) {
	my ($fc, $vol) = @_;
	my $client = ($::CLIENT || 'localhost') ;
	my $mounts = mounts($client);
	foreach my $mount (keys %{$mounts}) {
		if ((($mounts->{$mount}{src_host} eq
			$fc->{hostp}{hostname}) or
		     ($mounts->{$mount}{src_host} eq
		        $fc->{hostp}{default_ip})) and
		     ($mounts->{$mount}{mountpoint} =~ m/$vol/)){
			return 1;
		}
	}
	return 0;
}

1;
