## Package to perform Hyper-V opertaions
## This package offers the following
## 1. Cloning
## 2. Deletion
## 3. Host Migration
## 4. Storage Migration
## @author
## 		dl-nst-virtual@netapp.com

package HyperV;

use strict;
use warnings;

use Data::Dumper;
use NACL::APISet;

# Define constants for VM state
use constant FRESH => "Non-existent";
use constant CREATE_TRANSIT => "Create-transit";
use constant OFF => "Off";
use constant RUNNING => "Running";
use constant CLONE_TRANSIT => "Clone-transit";
use constant DESTROY_TRANSIT => "Destroy-transit";

# Create a new Hyper-V VM object
sub new {
    my $class  = shift;
    my $params = shift;

    my $self = {};

    die "'apiset_obj' is a required NACL apiset object for VM object" if(!defined($params->{apiset_obj}));
    $self->{apiset_obj} = $params->{apiset_obj};

    if(defined($params->{name})) {
        $self->{name} = $params->{name};
    }

    if(defined($params->{datastore})) {
        $self->{datastore} = $params->{datastore};
    }

    if(defined($params->{state})) {
        $self->{state} = $params->{state};
    }

    if(defined($params->{host})) {
        $self->{host} = $params->{host};
    }

    if(defined($params->{ps_lib_path})) {
        $self->{ps_lib_path} = $params->{ps_lib_path};
    }

    if(defined($params->{disk_type})) {
        $self->{disk_type} = $params->{disk_type};
        if($params->{disk_type} =~  /fixed/i) {
            $self->{disk_format} = "vhd";
        }
        else {
            $self->{disk_format} = "vhdx";
        }
    }

    if(defined($params->{disk_size})) {
        $self->{disk_size} = $params->{disk_size};
    }

    if(defined($params->{iso_path})) {
        $self->{iso_path} = $params->{iso_path};
    }

    if(defined($params->{mem_size})) {
        $self->{mem_size} = $params->{mem_size};
    }

    if(defined($params->{cpu_count})) {
        $self->{cpu_count} = $params->{cpu_count};
    }

    bless( $self, $class );

    return $self;
}

# Power on VM
sub power_on {
    my $self = shift;
    my $params = shift;
    my $response;

    if($self->{state} =~ /Non\-existent/i){
        $self->{state} = FRESH;
        die "Cannot perform this action on this VM in the non-existent state";
    }
    
    my $name = (defined $params->{name})? $params->{name} : $self->{name};

    $params->{cmd} = "Start-VM -Name $name";

    $response = $self->_execute_ps_command($params);

    if($response->{execution_status} =~ /True/i) {
        $self->{state} = RUNNING;
    } elsif ($response->{output} =~ /already/i) {
    	$self->{state} = RUNNING;
    } else {
        $self->{state} = OFF;
        print "Failure while starting ". $self->{name} .
                        " on '$self->{host}' \n";
        return 1;
    }
    return 0;
}

# Power off VM
sub power_off {
    my $self = shift;
    my $params;
    my $MAX_RETRY = 3;

    my $retry_count = 0;
    while(1)  {
        my $state = $self->get_state();
        last	if( $state eq RUNNING || $state eq OFF );
        sleep(10);
        $retry_count++;
        if( $retry_count >= $MAX_RETRY ) {
            print("The current state is not ready for powering off VM '$params->{name} during the last 3 min'\n");
            return 1;
        }
    }

    $params->{cmd} = "Stop-VM -Name $self->{name} -Force";

    my $response = $self->_execute_ps_command($params);
    if($response->{execution_status} =~ /True/i) {
	$self->{state} = OFF;
    } elsif ($response->{output} =~ /already/i) {
     	$self->{state} = OFF;
    } else {
        print "Failure while stopping ". $self->{name} .
                " on '$self->{host}' \n";
        return 1;
    }
    return 0;
}

# Create a VM
sub create_vm {
    my $self = shift;
    my $params;

    $params->{cmd}  =  "$self->{ps_lib_path}\\Create-VM.ps1 ";
    $params->{cmd} .= "-Name $self->{name} ";
    $params->{cmd} .= "-Path $self->{datastore} ";
    $params->{cmd} .= "-Format $self->{disk_format} ";
    $params->{cmd} .= "-Size $self->{disk_size} ";
    $params->{cmd} .= "-Type $self->{disk_type} ";
    $params->{cmd} .= "-ControllerType SCSI ";
    $params->{cmd} .= "-MemStartupBytes $self->{mem_size} ";
    $params->{cmd} .= "-ISOPath '$self->{iso_path}' ";
    $params->{cmd} .= "-CPUCount $self->{cpu_count}";

    print "Creating " .
        "VM: $self->{name} on\n" .
        "Datastore: $self->{datastore} \n" .
        "host: $self->{host} \n"
        ;

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/i) {
        print "VM '$self->{name}' was created sucessfully.";
    } elsif ($response->{output} =~ /The file exists/) {
            print "The Disk already exists, skipping creating VM '$params->{vm_name}'";
    } else {
        print "Failure while creating VM '$self->{name}' on '$self->{host}' \n";
        return 1;
    }
    return 0;
}

# Clone a VM
sub clone_vm {
    my $self = shift;
    my $params = shift;

    $params->{cmd} = "$self->{ps_lib_path}\\Clone-VM.ps1 ";
    $params->{cmd} .= "-Name '$params->{template}' ";
    $params->{cmd} .= "-ExpPath '$params->{export_path}' ";
    $params->{cmd} .= "-Prefix '$self->{name}' ";
    $params->{cmd} .= "-CloneDir '$params->{clone_dir}' ";
    $params->{cmd} .= "-SkipRmExpPath 1 ";
    $params->{cmd} .= "-Copies 1 ";
    $params->{cmd} .= "-Debugger 1";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/i) {
        print "VM '$self->{name}' was cloned sucessfully.";
    } elsif ($response->{output} =~ /The file exists/) {
        print "The VM already exists, skipping cloning VM '$params->{vm_name}'";
    } elsif ($response->{output} =~ /Insufficient Storage/) {
        print "Insufficient Storage at '$params->{clone_dir}'";
	return 2;
    } elsif ($response->{output} =~ /file was not found/) {
        print "Export is being processed by another process at '$params->{export_path}'";
        return 3;
    } else {
        print "Failure while cloning VM '$self->{name}' on '$self->{host}' \n";
        print "Trace: " . Dumper($response) . "\n";
        return 1;
    }

    return 0;
}

# Clone a VM
sub concurrent_clone_vm {
    my $self = shift;
    my $params = shift;

    $params->{cmd} = "$self->{ps_lib_path}\\Clone-VM-v2.ps1 ";
    $params->{cmd} .= "-Name '$params->{template}' ";
    $params->{cmd} .= "-ExpPath '$params->{export_path}' ";
    $params->{cmd} .= "-Prefix '$self->{name}' ";
    $params->{cmd} .= "-CloneDir '$params->{clone_dir}' ";
    $params->{cmd} .= "-Copies $params->{copies} ";
    $params->{cmd} .= "-throttleLimit $params->{parallel} ";
    $params->{cmd} .= "-StartMac '$params->{startmac}' ";    
    $params->{cmd} .= "-Debugger 1";

    my $response = $self->_execute_ps_command($params);
    
    my @results;
    $response->{output} =~ s/(\n|\r)//g;
    if($response->{output} =~ /Failure/i) {
	return $params->{copies};
    }
    elsif( $response->{output} =~ /\{(.+)\}/ ) {
	my @rows = split (',', $1);
	for( my $i = 0; $i < scalar(@rows) ; $i++) {
	    $rows[$i] =~ /\[(.+)\]/;
	    ($results[$i]->{pass}, $results[$i]->{vm_name}, $results[$i]->{reason})
		= split('::', $1);
	}    
    }    
    #print "results :", Dumper(\@results), "\n";
    
    return \@results;
}

# Register VM
sub register_vm {
    my $self = shift;
    my $params;

    $params->{cmd} = "$self->{ps_lib_path}\\Register-VM.ps1 -VMName $self->{name} -ConfigLocation $self->{datastore}";
    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
        print "VM '$self->{name}' was registeded sucessfully.";
        return 0;
    }
    else {
        print "Failed at registering  a VM '$self->{name}'\n";
        return 1;
    }
}


# Destroy VM
sub destroy_vm {
    my $self = shift;
    my $params;

    # Setting state as DESTROY_TRANSIT
    $self->{state} = DESTROY_TRANSIT if($self->{state} ne DESTROY_TRANSIT);

    # Deleting snapshots associated to a vm
    foreach my $snapshot (@{$self->list_snapshots()}) {
        $params->{cmd} = "$self->{ps_lib_path}\\Delete-Snapshot.ps1 -VMName $self->{name} -SnapshotName '$snapshot'";
        my $response = $self->_execute_ps_command($params);
        if($response->{output} =~ /Success/) {
            print "Deleted a snapshot '$snapshot' on vm '$self->{name}'\n";
        }
        else {
            print "Failed at deleting a snapshot '$snapshot' on VM '$self->{name}'\n";
        }
    }

    # Constructing the destroy command
    $params->{cmd} = "$self->{ps_lib_path}\\Destroy-Clones.ps1 -Name $self->{name}";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
        print "Deleted vm '$self->{name}";
    }
    elsif($response->{output} =~ /unable to find a virtual/) {
        print "Unable to find a vm '$self->{name}";
    }
    else {
        print "Failure at deleting VMs on '$self->{host}' \n";
        return 1;
    }
    return 0;
}

sub get_used_percent{
    my $self = shift;
    my $params;

    # Get used percent of drive for San
    $params->{cmd} = "$self->{ps_lib_path}\\Get-UsedPercent.ps1 -Dir $self->{datastore}";

    my $response = $self->_execute_ps_command($params);

    if($response->{output}  =~ /Success/) {
        return $response->{output};
    }
    else {
        print("Failed at getting used percent on drive '$self->{datastore}'");
        return 1;
    }
}

# Routine to migrate host
sub host_migrate_vm {
    my $self = shift;
    my $params = shift;

    $params->{cmd} = "$self->{ps_lib_path}\\Move-VM.ps1 ";
    $params->{cmd} .= "-VMName $self->{name} ";
    $params->{cmd} .= "-SourceHost $self->{host} ";
    $params->{cmd} .= "-DestinationHost $params->{target_host}";

     my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/i) {
        print "VM '$self->{name}' was moved sucessfully to Host '$params->{target_host}'.";
    }
    elsif($response->{output} =~ /The process cannot access the file because another process has locked a portion of the file/i) {
        print "VM '$self->{name}' is being accessed by another process.";
    }
    else {
        print "VM '$self->{name}' was failed at migrating to Host '$params->{target_host}'.";
        return 1;
    }

    return 0;
}

# Routine to migrate storage
sub storage_migrate_vm {
    my $self = shift;
    my $params = shift;

    $params->{cmd} = "$self->{ps_lib_path}\\Move-VMStorage.ps1 ";
    $params->{cmd} .= "-VMName $self->{name} ";
    $params->{cmd} .= "-SourceHost $self->{host} ";
    $params->{cmd} .= "-DestinationStoragePath $params->{target_storage}";

     my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/i) {
        print "VM '$self->{name}' was moved sucessfully to storage '$params->{target_storage}'";
    } else {
        print "VM '$self->{name}' was failed at migrating to storage '$params->{target_storage}'";
        return 1;
    }
    return 0;
}

# Routine to migrate storage in parallel
sub parallel_storage_migrate_vm {
    my $self = shift;
    my $params = shift;

    $params->{cmd} = "$self->{ps_lib_path}\\Parallel-Move-VMStorage.ps1 ";
    $params->{cmd} .= "-SourceHost $self->{host} ";
    $params->{cmd} .= "-DestinationStoragePath $params->{destPath} ";
    $params->{cmd} .= "-ThrottleLimit $params->{throttleLimit} ";
    $params->{cmd} .= "-IncludeRegex $params->{includeRegex}";
    
    my $response = $self->_execute_ps_command($params);

    return 0;
}

sub destroy_all_vm {
    my $self = shift;
    my $params = shift;

    # Get used percent of drive for San
    $params->{cmd} = "$params->{ps_lib_path}\\CleanSlate.ps1 -Hostname $self->{host} -IncludeRegex $params->{regex}";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
         return 0;
    }
    else {
        print "Failed at wiping out HyperV Host '$self->{host}'\n";
        return 1;
    }
}

sub deregister_vm {
    my $self = shift;
    my $params = shift;

    # Get used percent of drive for San
    $params->{cmd} = "$params->{ps_lib_path}\\Deregister-VM.ps1 -IncludeRegex $params->{regex}";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
         return 0;
    }
    else {
        print "Failed at deregistering from HyperV Host '$params->{host}'\n";
        return 1;
    }
}

sub clean_datastore {
    my $self = shift;
    my $params = shift;

    unless( defined($params) ) {
	$params->{regex} = '*';
    }
    my $datastore = $params->{datastore} . "\\" . $params->{regex};

    $params->{cmd} = "Remove-Item $datastore -Recurse -Force";

    my $response = $self->_execute_ps_command($params);
    if($response->{execution_status} =~ /True/i) {
        return 0
    }
    else {
        print "Failed at removing datastore $datastore : ", Dumper($response);
        return 1;
    }
}

sub snapshot_vm {
    my $self = shift;
    my $params;

    my $snap_name = "Snap_" . $self->{name} . "_" . time();
    $params->{cmd} = "$self->{ps_lib_path}\\Take-Snapshot.ps1 -VMName $self->{name} -SnapshotName $snap_name";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
         return 0;
    }
    else {
        print "Failed at taking snapshot of VM ". $self->{name} .
                " on Host '$self->{host}' \n";
        return 1;
    }
}

sub list_snapshots {
    my $self = shift;
    my $params;
    my @list_snapshots;

    $params->{cmd} = "$self->{ps_lib_path}\\List-Snapshots.ps1 -VMName $self->{name}";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
        $response->{output} =~ s/\r|\n//g;
        if($response->{output} =~ /\(\)\s+(.+)Success$/) {
            @list_snapshots = split(/\(\d*\)\s+/, $1)    if(defined $1);
        }
        print "snapshot list of VM '$self->{name}': " . Dumper(\@list_snapshots);
        return \@list_snapshots;
    }
    else {
        print "Failure at listing snapshots of VM ". $self->{name} .
                " on Host '$self->{host}' \n";
        return \@list_snapshots;
    }
}

# Destroy snapshots
sub destroy_snapshot {
    my $self = shift;
    my $params;
    my $fail_op_count = 0;

    # Deleting snapshots associated to a vm
    foreach my $snapshot (@{$self->list_snapshots()}) {
        $params->{cmd} = "$self->{ps_lib_path}\\Delete-Snapshot.ps1 -VMName $self->{name} -SnapshotName '$snapshot'";
        my $response = $self->_execute_ps_command($params);
        if($response->{output} =~ /Success/) {
            print "Deleted a snapshot '$snapshot' on vm '$self->{name}'\n";
        }
        else {
            print "Failed at deleting a snapshot '$snapshot' on VM '$self->{name}'\n";
            $fail_op_count++;
        }
    }

    return ( $fail_op_count == 0 ) ?  0 : 1;
}

sub restore_vm {
    my $self = shift;
    my $params;

    # parsing output
    my @snapshots = @{$self->list_snapshots()};

    # Pick a random snapshot from the snapshot lists.
    my $num_snapshots = scalar(@snapshots);
    if($num_snapshots == 0) {
        print "There is no snapshot of VM ". $self->{name} .
                " on Host '$self->{host}' \n";
        return 0;
    }

    my $snapshot_to_restore = $snapshots[int(rand($num_snapshots))];
    $params->{cmd} = "$self->{ps_lib_path}\\Apply-Snapshot.ps1 -VMName $self->{name} -SnapshotName '$snapshot_to_restore'";

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Success/) {
        # Deleting snapshots associated to a vm
        foreach my $snapshot (@snapshots) {
            $params->{cmd} = "$self->{ps_lib_path}\\Delete-Snapshot.ps1 -VMName $self->{name} -SnapshotName $snapshot";
            $response = $self->_execute_ps_command($params);
            if($response->{output} =~ /Success/) {
                print "Deleted a snapshot '$snapshot' on vm '$self->{name}'\n"
            }
            else {
                print "Failed at deleting a snapshot '$snapshot' on VM '$self->{name}'\n"
            }
        }
        return 0;
    }
    else {
        print "Failure while storign a VM '$self->{name}'" .
                "from snapshot '$snapshot_to_restore'".
                " on Host '$self->{host}' \n";
        return 1;
    }
}

# Get the MAC address of the VM
sub get_mac {
    my $self = shift;
    my $params;

    $params->{cmd} = "$self->{ps_lib_path}\\Get-Mac.ps1 ";
    $params->{cmd} .= "-Name $self->{name} ";

    my $response = $self->_execute_ps_command($params);

    if($response->{output} !~ /Failure/i) {
        $response->{output} =~ s/\r?\n$//;
        $self->{mac} = $response->{output};

        return 0;
    }
    else {
        print "Could not get mac address of VM '". $self->{name} .
                "' on Host '$self->{host}' \n";
        return 1;
    }

}

# Get the state of the VM
sub get_path {
    my $self = shift;
    my $params;

    $params->{cmd} = "(Get-VM -Name $self->{name}).path";

    my $response = $self->_execute_ps_command($params);
    $response->{output} =~ s/\r|\n//g;
    print "VM '$self->{name}' path : $response->{output}\n";
    return $response->{output};
}

sub check_vm {
        my $self = shift;
        my $params = shift;
        my $rsh_params;

        $rsh_params->{cmd} = "Get-VM ";
        $rsh_params->{cmd} .= "-Name '$self->{name}' ";

        my $response = $self->_execute_ps_command($rsh_params);

        if( $response->{execution_status} =~ /True/i ) {
                #Destroy this template. Invoke the mighty destroy-clones.ps1
                my $ret = $self->destroy_vm();
                if($ret){
                        print "Destroying the vm $params->{template_name} failed.\n";
                        return 1;
                }
                else{
                        print "Destroying the vm $params->{template_name} succeeded.\n";
                }
        }
        else {
            print "vm $params->{template_name} does not exit.\n";
        }
        return 0;
}

# Exports a vm to a specified location
sub export_vm {
        my $self = shift;
        my $params = shift;
        my $rsh_params;

        $rsh_params->{cmd} = "$self->{ps_lib_path}\\Export.ps1 ";
        $rsh_params->{cmd} .= "-Name '$params->{template_name}' ";
        $rsh_params->{cmd} .= "-ExpPath '$params->{export_location}' ";
        $rsh_params->{cmd} .= "-isClusterResource 0 ";
        $rsh_params->{cmd} .= "-Debugger 1 ";

    my $response = $self->_execute_ps_command($rsh_params);
        if($response->{output} =~ /Success/i) {
                print "The VM : $params->{template_name} was exported successfully.\n";
        }
        else{
                print "Export failed.\n";
                return 1;
        }
        return 0;
}

sub import_vm {
        my $self = shift;
        my $params = shift;
        my $rsh_params;
        $self->check_vm($params);
        $rsh_params->{cmd} = "$self->{ps_lib_path}\\Import.ps1 ";
        $rsh_params->{cmd} .= "-CloneDir '$params->{import_location}' ";
        $rsh_params->{cmd} .= "-TemplateName '$params->{template_name}' ";
        $rsh_params->{cmd} .= "-Prefix '$self->{name}' ";
        $rsh_params->{cmd} .= "-Copies 1 ";
        $rsh_params->{cmd} .= "-ExpPath '$params->{export_location}' ";

    my $response = $self->_execute_ps_command($rsh_params);
        if($response->{output} =~ /Success/i) {
                print "The VM : $params->{template_name} was imported successfully.\n";
        }
        else{
                print "Import failed.\n";
                return 1;
        }
        return 0;
}

# Get the VMid of the VM
sub get_vmid {
    my $self = shift;
    my $params;

    $params->{cmd} = "(Get-VM -Name $self->{name}).vmid";

    my $response = $self->_execute_ps_command($params);

    $response->{output} =~ s/\r|\n//g;
    $response->{output} =~ s/\s//g;
    $response->{output} =~ s/Guid\-\-\-\-//g;

    if(!defined($response) || $response->{output} =~ /unable/i) {
        return "Failure. Could not find VM ID for VM '$self->{name}'";
    } else {
        print "VM '$self->{name}' id : $response->{output}\n";
        return $response->{output};
    }
}

# Get the state of the VM
sub get_state {
    my $self = shift;
    my $params = shift;

    if( defined $params->{host} ) {
        $params->{cmd} = "(Get-VM -ComputerName $params->{host} -Name $self->{name}).state";
    }
    else {
        $params->{cmd} = "(Get-VM -Name $self->{name}).state";
    }

    my $response = $self->_execute_ps_command($params);
    if($response->{output} =~ /Running/i) {
        $self->{state} = RUNNING;
    }
    elsif($response->{output} =~ /Off/i) {
        $self->{state} = OFF;
    }
    else {
        print "Failure while getting state of VM ". $self->{name} .
                " on '$self->{host}' \n";
        $self->{state} = FRESH;
    }

    return $self->{state};
}

# Set the state of the VM
sub set_state {
    my $self = shift;
    my $state = shift;
    $self->{state} = $state;
}


# Send NACL execution command
sub _execute_ps_command {
    my $self = shift;
    my $params = shift;

    my $host = $self->{apiset_obj};

    if($params->{cmd} =~ /ps1/i) {
    	$params->{cmd} = "powershell.exe -ExecutionPolicy Bypass -NoLogo -NonInteractive -Command \"$params->{cmd}\"";
    }
    #else {
    #	$params->{cmd} = "powershell.exe -ExecutionPolicy Bypass -NoLogo -NonInteractive $params->{cmd}";
    #}

    my $response;
    $host->set_timeout( timeout => 72000);
    eval {
        $response = $host->execute_command(
                    command => $params->{cmd},
                    return_response_object => 1
                        );
    };

    unless ( ref $response && $response->isa('NACL::APISet::Response::CLI::PowerShell') ) {
        print "Expected response object, but obtained:\n" . Dumper($response) ;
    } else {
        #print "Response object is obtained as expected.";
        #print Dumper($response);
    }

    return $response;
}

1;
