package IOGen;
use Clone qw(clone);
use Data::Dumper;
use Params::Validate qw(:all);
use Tharn qw ($Log);

###Required Imported libraries
use TCD;
use POSIX;

# NACL Component Libraries required
use NACL::C::Node;
use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::C::Client;

# NACL STask Libraries required
use NACL::STask::Volume;
use NACL::STask::VserverCifsShare;
use NACL::STask::Client::Mount;

# NACL MTask Libraries required
use NACL::STask::VserverNFS;
use NACL::MTask::Data;
use NACL::MTask::Traffic;

# NACL Exception handling Libraries required
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::TimeoutException;

our @ISA = qw(Utility);

my $Package = "IOGen";

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my %opts = validate(
        @_, {
            debugLevel       => { type => SCALAR,   optional => 1, default => 0 },
            volCount         => { type => SCALAR,   optional => 1, default => 1 },
            nodeName         => { type => SCALAR,   optional => 1, default => undef },
            vserver          => { type => SCALAR,   optional => 1, default => "vs0" },
            clientNameList   => { type => ARRAYREF, optional => 1, default => undef },
            loadBalance      => { type => SCALAR,   optional => 1, default => 0 },
            useExistingVols  => { type => SCALAR,   optional => 1, default => 0 },
            setVolEfficiency => { type => SCALAR,   optional => 1, default => 0 },
            localNodeOnly    => { type => SCALAR,   optional => 1, default => 0 },
        }
    );

    my $self = $class->SUPER::new( debugLevel => $opts{debugLevel} );
    %opts = ( %{$self}, %opts );
    %opts = (
        %opts, (
            node       => undef,
            client     => undef,
            nodeApiset => undef,
            trafficRef => undef,
            mountRef   => undef,
        )
    );
    $self = \%opts;
    bless $self, $class;

    $self->init();

    return $self;
}

sub init {
    my $self = shift;

    $self->_getNode();
    $self->_initClient();
    $self->_installLicense();
    foreach my $aggr ( $self->_getAggregates() ) {
        my $existingVolCount = $self->_getVolCount( aggregate => $aggr );
        my $nextIncrement    = $self->_getVolNextIncrement( aggregate => $aggr );

        my $freeSpace        = $self->_getFreeSpace( aggregate => $aggr );
        my $freeSpacePerVol  = $self->_computeFreeSpacePerVol(
                freespace=>$freeSpace,
                volCount => $self->{volCount}
            );
        $self->log( 5, "$Package: DEBUG: Freespace is $freespace" );

        if ( $freeSpace > 0 and $self->{volCount} > 0) {
            my @vols = $self->_createVolumes(
                aggregate     => $aggr,
                volCount      => $self->{volCount},
                nextIncrement => $nextIncrement,
                volSize       => $freeSpacePerVol,
            );

        }

        $volsPresent = $self->_findCurrentVolumes( aggregate => $aggr );
        $self->setVolumes( volumes => $volsPresent );
    }
}
####################################################################################################
# setVolumes
# Setter function for volumes
# I/P     : array reference of the volumes to be used
# return  : None
####################################################################################################

sub setVolumes {
    my $self = shift;
    my %opts = validate(
        @_, {
            volumes => { type => ARRAYREF },
        }
    );
    @volumesPresent = ( @{ $self->{volumes} }, @{ $opts{volumes} } );
    $self->{volumes} = \@volumesPresent;
}

####################################################################################################
# setVolumes
# Getter function for volumes
# I/P     : array reference of the volumes to be used
# return  : array of volumes stored in the class which are effectively the list of volumes
#           that the tool must use to run IO on
####################################################################################################

sub getVolumes {
    my $self = shift;
    return @{ $self->{volumes} };
}

####################################################################################################
# _getNode
# Creates a NACL::C::Node object  for the filer and creates the apiset object
# I/P     : None
# return  : 1
####################################################################################################

sub _getNode() {
    my $self = shift;
    $self->log( 3, "$Package: INFO: in getNode()" );

    if ( defined $self->{nodeName} ) {
        $self->{node} = NACL::C::Node->new( node => $self->{nodeName} );
    }
    else {
        $self->{node} = NACL::C::Node->find();
    }

    my $nodeName        = $self->{node}->node();
    $self->{nodeName}   = $nodeName;                                    #This line assumes that name of the node and  tharnhost file name matches
    $self->log( 2, "$Package: NOTICE: Connected to $nodeName" );

    my $nodeApiset  = $self->{node}->get_7m_or_nodescope_apiset();
    $self->{nodeApiset} = $nodeApiset;

    $self->log( 3, "$Package: INFO: exiting getNode()" );
    return 1;
}

####################################################################################################
# _installLicense
# Installs the NFS license
# I/P     : None
# return  : 1
####################################################################################################

sub _installLicense {
    my $self = shift;
    $self->log( 3, "$Package: INFO: starting _installLicense" );
    $self->log( 2, "$Package: NOTICE: Installing NFS license" );
    my $feature = "nfs";
    # Try installing v2 license , if it fails, try installing v1 license
    my $is_v2 = $self->{node}->is_feature_supported( feature => 'license_v2' );

    try {
        if($is_v2) {
            $self->log( 3, "$Package: INFO: v2 Licenses are supported " );
            my $license = NACL::C::SystemLicense->add(
                command_interface => $self->{node},
                'license-code'    => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
                package           => "NFS",
            );
        }
        else {
            $self->log( 3, "$Package: INFO: v1 Licenses are supported " );
            my $license = NACL::C::SystemLicense->add(
                command_interface => $self->{node},
                'license-code'    => 'UYTCNCG',
            );
        }

    } catch NATE::BaseException with {
        my $exception = shift;
        $self->log( 2, "$Package: ERROR: Could not install NFS license" );
        $self->log( 3, "$Package: ERROR: ". $exception->text());
    };
    $self->log( 3, "$Package: INFO: exiting _installLicense" );
    return 1;
}

####################################################################################################
# _initClient
# Creates a NACL::C::Client objects for input client list
# If no client list is provided, the default client will be used for all IOs
# I/P     : None
# return  : 1
################################################################################ ####################

sub _initClient() {
    my $self = shift;
    #get all clients
    my @usableClients           = ();
    my @clientNames             = @{$self->{clientNameList}};
    my @availableClients        = NACL::C::Client->find();
    my @availableClientNames    = map {$_->client()} @availableClients;
    $self->log( 2, "$Package: INFO: Available list of clients - @availableClientNames" );

    if(!scalar(@availableClients)) {
        die("$Package: Error: There are no clients available for this testbed");
    }

    if(@clientNames) {
        foreach my $clientObj (@availableClients) {
            my $client = $clientObj->client();
            push(@usableClients, $clientObj) if(grep(/$client/, @clientNames));
        }
    } 
    
    @usableClients = @availableClients if(!scalar(@usableClients));
    
    if(!scalar(@usableClients)) {
        die("$package: Error: There are no clients available for this testbed");
    }

    my @usableClientNames   = map {$_->client()} @usableClients;
    $self->{clientList}     = [@usableClients];
    @clientNames            = map {$_->client()} @{$self->{clientList}};
    $self->log( 3, "$Package: INFO: List of clients used to run IO - @clientNames" );

    $self->{nextClientIndex} = 0;
    $self->_getNextClient();
    my $clientName           = $self->{currentClient}->client();

    $self->log( 2, "$Package: NOTICE: Connected to $clientName" );
    $self->log( 3, "$Package: INFO: exiting initClient()" );
}

####################################################################################################
# getNextClient
# Updates the value of current client object with next client available in clientList.
# I/P     : None
# return  :
####################################################################################################

sub _getNextClient() {
    my $self = shift;
    $self->log( 3, "$Package: INFO: starting getNextClient");

    if(ref($self->{clientList}->[$self->{nextClientIndex}])) {
        $self->{currentClient}      = $self->{clientList}->[$self->{nextClientIndex}];
        $self->{nextClientIndex}   += 1;
    } else {
        $self->{currentClient}      = $self->{clientList}->[0];     #First usable client
        $self->{nextClientIndex}    = 1;
    }

    $self->log( 2, "$Package: NOTICE: Set currentClient - ". $self->{currentClient}->client());
    $self->log( 3, "$Package: INFO: exiting getNextClient()" );
}

####################################################################################################
# _getAggregates
# Identifies the set of aggregates that are online apart from the root aggregate
# I/P     : None
# return  : Array of aggregates on the filer that are online and are non-root
####################################################################################################

sub _getAggregates {
    my $self = shift;
    $self->log( 3, "$Package: INFO: starting _getAggregates" );

    my $nodes = ($self->{localNodeOnly} == 1) ? [$self->{nodeName}] : ['*'];
    my @aggrsOnFiler = NACL::C::Aggregate->find(command_interface => $self->{node}, filter => {nodes => $nodes});
    my @aggrsOnFilerToBeUsed;
    foreach $aggrOnFiler (@aggrsOnFiler) {
        my $aggrName = $aggrOnFiler->aggregate();
        if ( $aggrOnFiler->get_one_state_attribute('state') =~ /online/i) {
            if ($aggrOnFiler->state->root() =~ /true/i or $aggrOnFiler->state->has_mroot() =~ /true/i) {
                $self->log( 2, "$Package: NOTICE: $aggrName is a root aggregate and hence will not be used ");
                next;
            }
            push @aggrsOnFilerToBeUsed, $aggrOnFiler;
        } else {
            $self->log( 2, "$Package: NOTICE: $aggrName is not online hence will not be used ");
        }
    }
    @aggrsOnFilerToBeUsed = map { $_->aggregate() } @aggrsOnFilerToBeUsed;
    $self->log( 2, "$Package: INFO: Aggregates present on the filer are @aggrsOnFilerToBeUsed" );

    $self->log( 3, "$Package: INFO: exiting _getAggregates" );
    return @aggrsOnFilerToBeUsed;
}

####################################################################################################
# _getFreeSpace
# if volumes are to be created on an aggregate, gets the available space for creation of volumes.
# I/P     : Aggregate name
# return  : available size in bytes
####################################################################################################

sub _getFreeSpace {
    my $self = shift;
    my %opts = validate(
        @_, {
            aggregate => { type => SCALAR, optional => 0 }
        }
    );
    $self->log( 3, "$Package: INFO: starting getFreeSpace" );

    my $aggrName        = $opts{aggregate};
    my $volCountForAggr = $self->{volCount};
    if ( $volCountForAggr == 0 ) {
        $self->log( 2, "$Package: NOTICE: No volumes to be created hence not computing freespace" );
        return 0;
    }

    my $aggrObj = NACL::C::Aggregate->find(
        command_interface => $self->{node},
        filter            => { aggregate => $opts{aggregate} },
    );

    my $freeSpaceAvailable = $aggrObj->state->availsize();
    $self->log( 3, "$Package: INFO: exiting _getFreeSpace" );
    return $freeSpaceAvailable;
}

####################################################################################################
# _computeFreeSpacePerVol
# Computes the size of volumes to be created
# I/P     : freespace available on the aggregate
#           No of volumes to be created in aggregate
# return  : available size in bytes
####################################################################################################

sub _computeFreeSpacePerVol{
    my $self = shift;
    my %opts = validate(
        @_, {
            freespace => { type => SCALAR },
            volCount  => { type => SCALAR}
        }
    );

    $self->log( 3, "$Package: INFO: starting _computeFreeSpacePerVol" );
    my $freeSpaceAvailable = $opts{freespace};
    my $volCountForAggr = $opts{volCount};
    if ($freeSpaceAvailable == 0) {
        return 0;
    }
    my $spacePerVolume = int( ( $freeSpaceAvailable / $volCountForAggr ) * 0.85 );
    my $minSizeOfVolume = 20 * 1024 * 1024;
    $self->log( 2, "$Package: NOTICE: Space available on $aggrName is $freeSpaceAvailable" );

    if ( $spacePerVolume <= $minSizeOfVolume ) {
        $self->log( 2, "$Package: Warning: Cant create $volCountForAggr volume(s) on $aggrName" );
        $volCountForAggr = int( $freeSpaceAvailable * 0.85 / $minSizeOfVolume );
        $spacePerVolume  = $minSizeOfVolume;
    }

    $self->log( 2, "$Package: INFO: Will attempt to create $volCountForAggr volume(s) on $aggrName");
    $self->{volCount} = $volCountForAggr;
    $self->log( 2, "$Package: INFO: Space used per volume on $aggrName is " . $spacePerVolume );

    $self->log( 3, "$Package: INFO: exiting _computeFreeSpacePerVol" );
    return $spacePerVolume;
}

####################################################################################################
# _getVolCount
# Identifies  the number of volumes created by IOGen
# I/P     : Aggregate
# return  : Largest index + 1
####################################################################################################

sub _getVolCount {
    my $self = shift;
    my %opts = validate(
        @_, {
            aggregate => { type => SCALAR, optional => 0 }
        }
    );
    $self->log( 3, "$Package: INFO: starting _getVolCount" );

    my $aggrName = $opts{aggregate};
    my @volName;
    try {
        @volName = NACL::C::Aggregate->find(
            command_interface => $self->{node},
            filter            => { aggregate => $opts{aggregate} }
        )->get_volumes();

        @volName = map { $_->volume() } @volName if scalar(@volName);

        my $noOfVols = 0;
        $self->log( 2, "$Package: INFO: $aggrName has @volName" );
        @volName = grep( /iogen/, @volName );
        $noOfVols = scalar @volName;
        $self->log( 3, "$Package: INFO: exiting _getVolCount with returnValue : " . $noOfVols );
        return $noOfVols;
    } otherwise {
        $self->log( 2, "$Package: INFO: $aggrName has no volumes" );
        $self->log( 3, "$Package: INFO: exiting _getVolCount" );
        return 0;
    };
}

####################################################################################################
# _getVolNextIncrement
# Identifies the largest index (i) in the volumes created by IOGen which are named as aggrname_vol_i
# I/P     : Aggregate
# return  : Largest index + 1
####################################################################################################

sub _getVolNextIncrement {
    my $self = shift;
    my %opts =  validate(
        @_, {
            aggregate => { type => SCALAR, optional => 0 }
        }
    );
    $self->log( 3, "$Package: INFO: starting _getVolNextIncrement" );

    my $aggrName = $opts{aggregate};
    my @volName;
    my $volIndex = 0;
    try {
        @volName = NACL::C::Volume->find(
            command_interface => $self->{node},
            filter            => { aggregate => $opts{aggregate} },
        );

        @volName = map { $_->volume() } @volName if scalar(@volName);
        $self->log( 3, "$Package: NOTICE: $aggrName has @volName" );
        @volName = grep( /iogen/, @volName );

        if ( ( scalar @volName ) != 0 ) {
            my @volIndice;

            for $index (@volName) {
                my @arr = split /_/, $index;
                $index = $arr[-1];
                push @volIndice, $index;
            }
            $self->log( 2, "$Package: INFO: Indexes for $aggrName is @volIndice" );
            @volIndice = sort { int($a) <=> int($b) } @volIndice;
            $self->log( 3, "$Package: INFO: Ascending Indexes for $aggrName is @volIndice" );
            $volIndex = $volIndice[-1];
            $self->log( 2, "$Package: INFO: Max Index for $aggrName is $volIndex" );
        }
        else {
            $volIndex = 0;
            $self->log( 2, "$Package: NOTICE: No volumes created by IOGen exists yet" );
        }
    } otherwise {
        $volIndex = 0;
    };
    $self->log( 3, "$Package: INFO: exiting getVolNextIncrement with returnValue : $temp");
    return $volIndex + 1;
}

####################################################################################################
# _findCurrentVolumes
# Return the list of volumes created by IOGen
# I/P     : Aggregate
# return  : Reference of volumes created by IOGen
####################################################################################################

sub _findCurrentVolumes {
    my $self = shift;
    my %opts = validate(
        @_, {
            aggregate => { type => SCALAR, optional => 0 },
        }
    );
    $self->log( 3, "$Package: INFO: starting _findCurrentVolumes" );

    my $aggregate = $opts{aggregate};
    my @volumes;
    try {
        @volumes = NACL::C::Volume->find(
            command_interface => $self->{node},
            filter            => { aggregate => $aggregate },
        );
        @volumes = map { $_->volume() } @volumes if scalar(@volumes);
        $self->log( 2, "$Package: NOTICE : Existing volumes on $aggregate is @volumes" );
        if ( $self->{useExistingVols} == 0 ) {
            @volumes = grep /_iogen_/, @volumes;
            @volumes = grep !/^temp/, @volumes;
            $self->log( 2, "$Package: NOTICE : volumes Created by IOGen on $aggregate is @volumes , and hence will be used"
            );
        }
        else {
            $self->log( 2, "$Package: NOTICE : Volumes Used by IOGen on $aggregate is @volumes" );
        }
    }  catch NATE::BaseException with {
        my $exception = shift;
        $self->log( 2, "$Package: ERROR: Failed to list volumes ");
        $self->log( 3, "$Package: ERROR: ". $exception->text());
    };

    $self->log( 3, "$Package: INFO: exiting _findCurrentVolumes" );
    return \@volumes;
}

####################################################################################################
# _createVolumes
# Create volumes based on the nomenclature <aggrName>_iogen_<index>
# with index = maximum existant index on that volume + 1 for all online aggregates
# I/P     : Aggregate name
#           Number of volumes to be created
#           index for nuomenclature
#           Size of the volume to be created
# return  : Array of volumes created by IOGen
####################################################################################################

sub _createVolumes {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            aggregate     => { type => SCALAR, optional => 0 },
            volCount      => { type => SCALAR, optional => 0 },
            nextIncrement => { type => SCALAR, optional => 0 },
            volSize       => { type => SCALAR, optional => 0 },
        }
    );
    $self->log( 3, "$Package: INFO: in createVolumes" );

    my $aggrObj = NACL::C::Aggregate->find(
        command_interface => $self->{node},
        filter            => { aggregate => $opts{aggregate} },
    );
    if ( $aggrObj->state() eq 'false' ) {
        $self->log( 2, "$Package: ERROR: $opts{aggregate} is not online and hence no volumes can be created ");
        return ('');
    }
    my @vols = ( $opts{nextIncrement} .. $opts{volCount} + $opts{nextIncrement} - 1 );

    @vols = map { $opts{aggregate} . "_iogen_" . $_ } @vols;
    ## Create the volumes
    foreach my $volName (@vols) {
        try {
            my $temp_vol_obj = NACL::C::Volume->create(
                volume            => $volName,
                command_interface => $self->{node},
                aggregate         => $opts{aggregate},
                vserver           => $self->{vserver},
                size              => $opts{volSize},
                'junction-path'   => "/".$volName,
            );
            $self->log( 2, "$Package: INFO: Created Volume $volName in $opts{aggregate}" );
        } otherwise {
            $self->log( 2, "$Package: ERROR: Failed to create Volume $volName in $opts{aggregate}" );
        };
    }

    if($self->{setVolEfficiency} == 1) {
        foreach my $volName (@vols) {
            try {
                my $temp_vol_obj_eff = NACL::STask::VolumeEfficiency->on(
                                    command_interface => $self->{node},
                                    volume            => $volName,
                    vserver           => $self->{vserver},
                );
                                $temp_vol_obj_eff->modify(
                    "compression"      => "on",
                    "inline-compression" => "true",
                );
                $self->log( 2, "$Package: INFO: Set volume efficiency on  $volName in $opts{aggregate}" );
            }
            otherwise {
                $self->log( 2, "$Package: ERROR: Failed to set volume efficiency on $volName in $opts{aggregate}" );
            };
        }
    }

    $self->log( 3, "$Package: INFO: exiting createVolumes" );
    return @vols;
}

####################################################################################################
# deleteVol
# Delete volumes  created by IOGen
# with index = maximum existant index on that volume + 1 for all online aggregates
# I/P     : None
# return  : None
####################################################################################################

sub deleteVol {
    my $self = shift;
    my @volName = NACL::C::Volume->find( command_interface => $self->{node} );
    @volName = map { $_->volume() } @volName;
    @volName = grep( /_iogen_/, @volName );

    try {
    foreach $vol (@volName) {
        my $volObj = NACL::C::Volume->new(
                             command_interface => $self->{node},
                             volume            => $vol,
                             vserver           => $self->{vserver},
                         );
        if ( $volObj->get_one_state_attribute('state') !~ /offline/i) {
        $volObj->offline();
        }
           NACL::C::Volume->delete(
                command_interface => $self->{node},
                volume            => $vol,
                vserver           => $self->{vserver},
            );
        }
    }  catch NATE::BaseException with {
        my $exception = shift;
        $self->log( 2, "$Package: ERROR: Failed to list volumes during deleteVol ");
        $self->log( 3, "$Package: ERROR: ". $exception->text());
    };

}

####################################################################################################
# startIO
# Compute the size of the files, number of possible files that can be created , quantise the size
# and time and initiate calls to corresponding IO functions
# I/P     : type of IO  [filersio|clientsio|clienthammer|hammer|dir]
#           volume name
#           number of files
#           runtime
#           number of directories to be created
#           depth of the directory tree
# return  : None
####################################################################################################

sub startIO {
    my $self = shift;
    my %opts = validate(
        @_, {
            typeOfIO    => { type => SCALAR, optional => 0, default => 'filersio' },
            volume      => { type     => SCALAR, optional => 0 },
            noOfFiles   => { optional => 1 },
            sizeOfFiles => { optional => 1 },
            runTime     => { type => SCALAR, optional => 1, default => 1000 },
            noOfDir     => { type => SCALAR, optional => 1, default => 2 },
            depthOfDir  => { type => SCALAR, optional => 1, default => 2 },
            readPercentage  => { type => SCALAR, optional => 1, default => 50 },

        }
    );
    $self->log( 3, "$Package: INFO: in startIO" );

    my $volName = $opts{volume};
    my $residingOn = $self->_getResidingNode( volume => $volName );
    $self->log( 3, "$Package: INFO: $volName is residing on $residingOn" );

    my $volObj = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => { volume => $opts{volume} },
        allow_empty       => 1,
    );

    if(!defined($volObj)) {
        $Log->warn("$Package: Error: Volume not found, skipping this volume - $volName");
        return 0;
    }

    my $availableSizeOfVolume = $volObj->state->size();
    my $sizeOfFile            = $opts{sizeOfFiles};
    my $noOfFiles             = $opts{noOfFiles};
    my $runTime               = $opts{runTime};
    my $noOfDir               = $opts{noOfDir};
    my $depthOfDir            = $opts{depthOfDir};
    my $readPercentage        = $opts{readPercentage};
    my $op;
    ## Convert to seconds
    $runTime = $self->convertTime(runTime => $runTime);

    for(my $iter=0; $iter < 5; $iter++) {
        try {
            if ($volObj->get_one_state_attribute('state') !~ /online/i) {
                $self->log( 3, "$Package: INFO: exiting startIO for $volName as it is not online and hence IO cant be generated for it" );
            }
        } catch NATE::BaseException with {
            my $exception = shift;
            if($exception->text() =~ /temporarily busy/i) {
                   $self->log(2, "$Package: ERROR :  IO cant be started on this volume as it is temporarily busy. Will retry in 10 seconds attempt $iter of 5 ");
                sleep 10;
                next;
            } else {
                $self->log(2, "$Package: ERROR :  skipping IO on $volume as " . $exception->text());
            }
        };
    }

    ## convert to unit size
    if ( ( defined $opts{sizeOfFiles} ) && ( defined $opts{noOfFiles} ) ) {
        $sizeOfFile = $self->_convertSize( size => $sizeOfFile );
        if ( $sizeOfFile * $noOfFiles < $availableSizeOfVolume ) {
            $sizeOfFile = int( $availableSizeOfVolume / $noOfFiles );
            $self->log( 2,"$Package: NOTICE : Not enough space  in $volName hence creating $noOfFiles of size $sizeOfFile");
        }
        if ( $sizeOfFile < 4096 ) {
            $noOfFiles = int( $availableSizeOfVolume / 4096 );
            $self->log( 2,"$Package: NOTICE : Not enough space  in $volName hence creating $noOfFiles of size $sizeOfFile");
        }
    }   elsif ( ( defined $opts{sizeOfFiles} ) or ( defined $opts{noOfFiles} ) ) {
        $noOfFiles =
          int( $availableSizeOfVolume / $self->_convertSize($sizeOfFile) )
          if ( defined $opts{sizeOfFiles} );
        $sizeOfFile = int( $availableSizeOfVolume / $noOfFiles )
          if ( defined $opts{noOfFiles} );
    }   else {
        $noOfFiles  = 2;
        $sizeOfFile = int( $availableSizeOfVolume / $noOfFiles );
    }


    my %optsIO;
    $optsIO{volume}      = $volName;
    $optsIO{noOfFiles}   = $noOfFiles;
    $optsIO{sizeOfFiles} = $sizeOfFile;
    $optsIO{runTime}     = $runTime;
    $optsIO{residingOn}  = $residingOn;
    if ( $opts{typeOfIO} =~ m/dir/i ) {
       $optsIO{noOfDir}     = $noOfDir;
       $optsIO{depthOfDir}  = $depthOfDir;
    }

    if ( $opts{typeOfIO} =~ m/sio/i ) {
       $optsIO{readPercentage} = $readPercentage;
    }
    
    $self->log(3, "$Package: INFO: loadBalance - ".$self->{loadBalance});
    if($self->{loadBalance} == 0 && $opts{typeOfIO} =~ /(clienthammer|clientsio)/i) {
        my @clients = @{$self->{clientList}} ;
        $self->log(5, "$Package: DEBUG : total clients - ".scalar(@clients));
        foreach my $clientObj (@clients) {
            $optsIO{client} = $clientObj;
            $self->log(3, "$Package: INFO: volume - ".$optsIO{volume}." client - ". $optsIO{client}->client());
            $self->{&{$opts{typeOfIO}}($self, %optsIO)};
        }
    } else {
        if( $opts{typeOfIO} =~ /(clienthammer|clientsio)/i) {
            $optsIO{client} = $self->{currentClient};
            $self->_getNextClient();
            $self->log(3, "$Package: INFO: volume - ".$optsIO{volume}." client - ". $optsIO{client}->client());
        }
        $self->{&{$opts{typeOfIO}}($self, %optsIO)};
    }

    $self->log( 3, "$Package: INFO: exiting startIO" );
}

####################################################################################################
# filersio
# Initiate filersio on specified number of files of specified size residing on volume for 'runtime' s
# I/P     : volume name
#           number of files
#           size of files to be created
#           runtime for the IO to run
#           Node on which the volume is residing
# return  : None
####################################################################################################

sub filersio {
    my $self = shift;
    my %opts = validate(
        @_, {
            volume      => { type     => SCALAR, optional => 0 },
            noOfFiles   => { optional => 1 },
            sizeOfFiles => { optional => 1 },
            runTime     => { type => SCALAR, optional => 1, default => 1000 },
            residingOn  => { type => SCALAR | UNDEF, optional => 1, default => undef },
            readPercentage => { type => SCALAR, optional => 1, default => 50 },
            }
    );

    my $sizeOfFile  = $opts{sizeOfFiles};
    my $volName     = $opts{volume};
    my $noOfFiles   = $opts{noOfFiles};
    my $runTime     = $opts{runTime};
    my $residingOn  = $opts{residingOn};
    my $readPercentage  = $opts{readPercentage};
    my $residingNode = $self->_getResidingNode( 'volume' => $volName, );
    my $residingNodeApiset = NACL::C::Node->new( node => $residingNode )->get_7m_or_nodescope_apiset();
    my $output;

    if ( $sizeOfFile < 4096 ) {
        $noOfFiles = int(($noOfFiles * $sizeOfFile)/4096);
        $self->log( 2, "$Package: ERROR: Not enough space to create $opts{noOfFiles}, hence creating $noOfFiles ");
    }
    try {
        for $iter ( 1 .. $noOfFiles ) {

            my $response = $residingNodeApiset->filersio_asyncio_active(
            'read-percentage' => $readPercentage,
            'rand-percentage' => 50,
            'blk-size'        => '4k',
            'start-byte'      => 0,
            'end-byte'        => $sizeOfFile ,
            'run-time'        => $runTime,
            'num-active'      => 2,
            'filename'        => "/vol/$volName/file_iogen_$iter",
            'create'          => 1,
            );
            $self->log( 2, "$Package: INFO: Filersio started on file.$iter on volume $volName ". Dumper( $response->get_parsed_output()) );
        }

    } catch  NATE::BaseException with {
        my $exceptionObject = shift;
        $self->log( 2, "$Package: ERROR: FilerSIO failed to start with error ". $exceptionObject->text());
    };

    $self->log( 3, "$Package: INFO: Reached the end of the sub filerSIO" );
}

####################################################################################################
# hammer
# Initiate hammer on specified number of files of specified size residing on volume for 'runtime' s
# I/P     : volume name
#           number of files
#           size of files to be created
#           runtime for the IO to run
#           Node on which the volume is residing
# return  : None
####################################################################################################

sub hammer{
    my $self = shift;
    my %opts = validate(
        @_, {
            volume      => { type     => SCALAR, optional => 0 },
            noOfFiles   => { optional => 1 },
            sizeOfFiles => { optional => 1 },
            runTime     => { type => SCALAR, optional => 1, default => 1000 },
            residingOn  => { type => SCALAR | UNDEF, optional => 1, default => undef },
        }
    );

    my $sizeOfFile  = $opts{sizeOfFiles};
    my $volName     = $opts{volume};
    my $noOfFiles   = $opts{noOfFiles};
    my $runTime     = $opts{runTime};
    my $residingOn  = $opts{residingOn};
    my $output;

    $sizeOfFile = int( $sizeOfFile / 4096 );

    for $iter ( 1 .. $noOfFiles ) {
        my $command    = "\"priv set test; hammer -f $runTime /vol/$volName/file.$iter $sizeOfFile \"";
        if ( $self->{node}->mode() eq 'CMode' ) {
            $output = NACL::C::SystemNode->run(
                command_interface   => $self->{node},
                node                => $residingOn,
                command             => $command,
            );
        }
        else {
            $command    = "priv set test; hammer -f $runTime /vol/$volName/file.$iter $sizeOfFile ";
            $output = $self->{nodeApiset}->execute_command( command =>$command );
        }
    }

}

####################################################################################################
# hammer
# Mount volume on client at mountpath and Initiate hammer on specified number of files of specified
# size residing on volume for 'runtime' s from the client
# I/P     : volume name
#           number of files
#           size of files to be created
#           runtime for the IO to run
#           Node on which the volume is residing
# return  : None
####################################################################################################

sub clienthammer {
    my $self = shift;
    my %opts = validate(
        @_, {
            volume      => { type     => SCALAR, optional => 0 },
            noOfFiles   => { optional => 1 },
            sizeOfFiles => { optional => 1 },
            runTime     => { type => SCALAR, optional => 1, default => 1000 },
            client      => { type => OBJECT},
            residingOn  => { type => SCALAR | UNDEF, optional => 1, default => undef },
        }
    );

    my $sizeOfFile  = $opts{sizeOfFiles};
    my $volName     = $opts{volume};
    my $noOfFiles   = $opts{noOfFiles};
    my $runTime     = $opts{runTime};
    my $clientObj   = $opts{client};
    my $apiset      = $clientObj->apiset( interface => "CLI", );
    my $output;
    my $path = $self->_mountAndgetPath( volume => $volName, client => $clientObj);
    my @fileNameIterArray = ();

    for $iter ( 1 .. $noOfFiles ) {
        my $fileNameIter=$path;
        push(@fileNameIterArray, $fileNameIter);
    }

    #   defining exception Handler
    #   This function is expected to fail if there isn't enough space on the volume to start hammer
    my $on_exception = sub {
        my $future = shift;
        $self->log( 3, "$Package: INFO on_exception handler is invoked");
        if ( $future->exception() ) {
            $self->log( 2, "$Package: ERROR Error in running traffic");
            $self->log( 2, "$Package: ERROR Error is: \n" . $future->exception() );
        } else {
        $self->log( 2, "$Package: INFO Completed running traffic");
        $apiset->execute_command( command => "ls -al $path");
        }
    };

    try {
        my $traffic = NACL::MTask::Traffic->start(
            command_interface   => $self->{client},
            paths               => $path,
            tool                => 'hammer',
            nofilercore         => 1,
            fill                => 1,
            iterations          => $noOfFiles,
            runtime             => $runTime,
            nocleanup           => 1,
            maxfsize            => $sizeOfFile,
            on_exception        => $on_exception
        );
    } catch NATE::BaseException with {
        my $exception = shift;
        $self->log( 2, "$Package: ERROR: Failed start hammer on $path, moving on.... ");
        $self->log( 3, "$Package: ERROR: ". $exception->text());
    };
    $self->log( 2, "$Package: INFO Hammer Processes running on the client are as follows ");
    $apiset->execute_command( command => "ps -aef | grep hammer" );
    #for(my $iter =0; $iter < $runTime; $iter+=10 ) {
    if($self->{debugLevel} > 2) {
        $self->log( 3, "$Package: INFO Files mounted on $volName are as follows ");
        $apiset->execute_command( command => "ls -al $path");
    }
    #    sleep 10;
    #}
    my @traffic = @{$self->{trafficRef}};
    push @traffic,$trafficObj;
    $self->{trafficRef} = \@traffic;
}


####################################################################################################
# clientsio
# Mount volume on client at mountpath and Initiate sio on specified number of files of specified
# size residing on volume for 'runtime' s from the client
# I/P     : volume name
#           number of files
#           size of files to be created
#           runtime for the IO to run
#           Node on which the volume is residing
# return  : None
####################################################################################################

sub clientsio {
    my $self = shift;
    my %opts = validate(
        @_, {
            volume          => { type     => SCALAR, optional => 0 },
            noOfFiles       => { optional => 1 },
            sizeOfFiles     => { optional => 1 },
            runTime         => { type => SCALAR, optional => 1, default => 1000 },
            client          => { type => OBJECT},
            residingOn      => { type => SCALAR | UNDEF, optional => 1, default => undef },
            readPercentage  => { type => SCALAR, optional => 1, default => 50 },
        }
    );

    my $sizeOfFile  = $opts{sizeOfFiles};
    my $volName     = $opts{volume};
    my $noOfFiles   = $opts{noOfFiles};
    my $runTime     = $opts{runTime};
    my $clientObj   = $opts{client};
    my $readPercentage  = $opts{readPercentage};
    my $output;
    my $path        = $self->_mountAndgetPath( volume => $volName, client => $clientObj);
    my @fileNameIterArray = ();
    my $apiset      = $clientObj->apiset( interface => "CLI");

    #   This function is expected to fail if there isn't enough space on the volume to start sio
    my $on_exception = sub {
            my $future = shift;
            $self->log( 3, "$Package: INFO on_exception handler is invoked");
            if ( $future->exception() ) {
                $self->log( 2, "$Package: ERROR Error in running traffic");
                $self->log( 2, "$Package: ERROR Error is: \n" . $future->exception() );
            }
    };

    for $iter ( 1 .. $noOfFiles ) {
        my $fileNameIter=$path."/file_iogen_".$iter;
        push(@fileNameIterArray, $fileNameIter);
#       TO BE CHGED TO ABSOLUTE PATH
#       Create a sparse file
#       MTask Data.pm can be used to achieve the same
        my $cmd = "/u/sunilnk,btc/p4/test/test/raidqa/utilities/IOGen/mkfile-unix $fileNameIter $sizeOfFile" ;
        $self->log( 2, "$Package: NOTICE StartIO command: $cmd");
        $apiset->execute_command( command => $cmd, timeout => 100);
        $apiset->execute_command( command => "chmod 777 $fileNameIter" );
    }

    if($self->{debugLevel} > 2) {
        $self->log( 3, "$Package: INFO Files mounted on $volName are as follows ");
        $apiset->execute_command( command => "ls -al $path");
    }
    try {
    my $trafficObj = NACL::MTask::Traffic->start(
            command_interface   => $clientObj,
            paths               => \@fileNameIterArray,
            tool                => 'sio',
            read                => $readPercentage,
            rand                => '100',
            block_size          => '4k',
            start_byte          => '0',
            file_size           => $sizeOfFile,
            'num_of_threads'    => $$fileNameIterArray,
            on_exception        => $on_exception,
            duration_in_secs    => $runTime,
    );
    } catch NATE::BaseException with {
        my $exception = shift;
        $self->log( 2, "$Package: ERROR: Failed start sio on $path, moving on.... ");
        $self->log( 3, "$Package: ERROR: ". $exception->text());
    };
    #for(my $iter =0; $iter < $runTime; $iter+=10 ) {
    $self->log( 2, "$Package: INFO SIO Processes running on the client are as follows ");
    $apiset->execute_command( command => "ps -aef | grep sio" );

    if($self->{debugLevel} > 2) {
        $self->log( 3, "$Package: INFO Files  on $volName after the IO was started are as follows ");
        $apiset->execute_command( command => "ls -al $path");
    }

    #    sleep 10;
    #}

    my @traffic = @{$self->{trafficRef}};
    push @traffic,$trafficObj;
    $self->{trafficRef} = \@traffic;
}

####################################################################################################
# raidsio
# WIP
# I/P     : volume name
#           number of files
#           size of files to be created
#           runtime for the IO to run
#           Node on which the volume is residing
# return  : None
####################################################################################################

sub raidsio {
    my $self = shift;
    my %opts = validate(
        @_, {
            volume      => { type     => SCALAR, optional => 0 },
            noOfFiles   => { optional => 1 },
            sizeOfFiles => { optional => 1 },
            runTime     => { type => SCALAR, optional => 1, default => 1000 },
            residingOn  => { type => SCALAR | UNDEF, optional => 1, default => undef },
        }
    );

    my $sizeOfFile  = $opts{sizeOfFiles};
    my $volName     = $opts{volume};
    my $noOfFiles   = $opts{noOfFiles};
    my $runTime     = $opts{runTime};
    my $residingOn  = $opts{residingOn};
    my $output;

    if ( $sizeOfFile <4096 ) {
        $noOfFiles = int(($noOfFiles * $sizeOfFile)/4096);
        $self->log( 2, "$Package: ERROR: Not enough space to create $opts{noOfFiles}, hence creating $noOfFiles ");
    }
    my $volObj = NACL::C::Volume->new(
        command_interface   => $self->{node},
        volume              => $volName
        );
    my $residingAggr = $volObj->get_aggregate();
    for $iter ( 1 .. $noOfFiles ) {
        my $command    = "\"priv set test; raidsio asyncio_active 50 -r 0  $sizeOfFile $runTime 2  $residingAggr -synchronous -print_stats 10 \"";
        if ( $self->{node}->mode() eq 'CMode' ) {
            $output = NACL::C::SystemNode->run(
                command_interface   => $self->{node},
                node                => $residingOn,
                command             => $command,
            );
        }
        else {
            $command    = "priv set test; raidsio asyncio_active 50 -r 0  $sizeOfFile $runTime 2  $residingAggr -synchronous -print_stats 10";
            $output     = $self->{nodeApiset}->execute_command( command =>$command );
        }
        $self->log( 2, "$Package: NOTICE:Output of raidsio is $output" );
        if ( $output =~ m/failed to start/g ) {
            $self->log( 2, "$Package: ERROR: Failed to start raidsio on file.$iter on volume $volName ");
        }
        if ( $output =~ m/found/g ) {
            $self->log( 2, "$Package: ERROR: Failed to start raidsio on file.$iter on volume $volName . raidsio is only supported in SN+ releases");
        }
    }
}

####################################################################################################
# dir
# Creates a directory structure having noOfDir directories at each level and depthOfDir levels
# I/P     : volume name
#           number of files
#           size of files to be created
#           runtime for the IO to run
#           Number of directories to be created at each level
#           Depth of the directory tree
#           Node on which the volume is residing
# return  : None
####################################################################################################

sub dir {
    my $self = shift;
    my %opts = validate(
        @_, {
            volume      => { type     => SCALAR, optional => 0 },
            noOfFiles   => { optional => 1 },
            sizeOfFiles => { optional => 1 },
            runTime     => { type => SCALAR, optional => 1, default => 1000 },
            noOfDir     => { type => SCALAR, optional => 1, default => 2 },
            client      => { type => OBJECT},
            depthOfDir  => { type => SCALAR, optional => 1, default => 2 },
            residingOn  => { type => SCALAR | UNDEF, optional => 1, default => undef },
        }
    );

    my $sizeOfFile  = $opts{sizeOfFiles};
    my $volName     = $opts{volume};
    my $noOfFiles   = $opts{noOfFiles};
    my $runTime     = $opts{runTime};
    my $noOfDir     = $opts{noOfDir};
    my $depthOfDir  = $opts{depthOfDir};
    my $clientObj   = $opts{client};
    my $output;
    my @fileNameIterArray = ();
    my $apiset  = $clientObj->apiset( interface => "CLI" );
    my $path    = $self->_mountAndgetPath( volume => $volName, client => $clientObj);

    $sizeOfFile = int($sizeOfFile/ ($depthOfDir*$noOfDir));
    if($self->{debugLevel} > 2) {
        $self->log( 3, "$Package: INFO Files mounted on $volName are as follows ");
        $apiset->execute_command( command => "ls -al $path");
    }

    try {
    my $traffic = NACL::MTask::Data->populate(
            command_interface => $clientObj,
            path              => $path,
            file_avg          => $noOfFiles,
            file_dev          => 0,
            dir_avg           => $noOfDir,
            dir_dev           => 0,
            depth_avg         => $depthOfDir,
            depth_dev         => 0,
            stream_avg        => $noOfDir,
        );

    } catch NATE::BaseException with {
    my $exception = shift;
    $self->log(2, "$Package: ERROR : " . $exception->text());
    };

    #for(my $iter = 0; $iter < $runTime; $iter= $iter + 10 ) {
    if($self->{debugLevel} > 2) {
        $self->log( 3, "$Package: INFO Files mounted on $volName  after IO started are as follows ");
        $apiset->execute_command( command => "ls -al $path");
    }
}

####################################################################################################
# unmount
# Unmount volumes mounted by IOGen
# I/P     : None
# return  : None
####################################################################################################

sub unmount {
    my $self = shift;
    my @mountPoints = @{$self->{mountRef}};

    foreach $mount (@mountPoints) {
        $mount->purge();
    }

}

####################################################################################################
# _mountAndgetPath
# Mount the volume on the client and return the mount path at the client
# Using UNIX only
# Create Vserver if Cmode
# Mount Volume at Filer
# Create mount point at Client
# Return path to mount point at the client
# I/P     : Volume name
# return  : mount path
####################################################################################################

sub _mountAndgetPath {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            volume => { type => SCALAR, optional => 0 },
            client => { type => OBJECT},
        }
    );
    my $client     = $opts{client};
    my $clientName = $client->client();
    my $volName    = $opts{volume};
    my $nfs;
    $self->log( 3, "$Package: INFO: in _mountAndgetPath" );
    $self->log( 2, "$Package: INFO: Using client $clientName " );

    if ( $self->{node}->mode() eq 'CMode' ) {
        try {
            $nfs = NACL::STask::VserverNFS->create(
                command_interface => $self->{node},
                vserver           => $self->{vserver},
                access            => "true",
                'v3'              => "enabled",
                udp               => "enabled",
                tcp               => "enabled"
            );
        }
        otherwise {
                    #Ignore the exception
        };
    }

    my $volObj = NACL::C::Volume->find(
        command_interface   => $self->{node},
        filter              => { volume => $volName },
        allow_empty         => 1,
    );

    if(!ref($volObj)) {
        $Log->warn("$Package: Warning: Volume not found - $volName");
        return 0;
    }

    if ( $self->{node}->mode() eq 'CMode' ) {
        $volObj->modify(
            policy             => 'default',
            'unix-permissions' => 777
        );
        $volObj->mount(
            command_interface => $self->{node},
            volume            => $volName,
            vserver           => $self->{vserver},
            'junction-path'   => "/$volName",
        );
        sleep 10;
        my $volState = $volObj->state();
        $self->log(2, "$Package: INFO: Volume state of $volName after modify is ". $volState->{state});
     } else {
        $self->{nodeApiset}->execute_command( command => "nfs on;exportfs -io sec=sys,rw,anon=0,nosuid /vol/$volName" );
        $self->{nodeApiset}->execute_command( command => 'rdfile /etc/exports' );
     }

    ## Create mount at Client
    my %mountParams = ( volume => $volObj, nacltask_if_exists => 'purge' );
    my  $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        %mountParams,
    );

    my @mountPoints = @{$self->{mountRef}};
    push @mountPoints,$mount;
    $self->{mountRef} = \@mountPoints;

    my $path = $mount->mount_point();
    $client->apiset->chmod(
        paths             => $path,
        mode              => "a+rwx",
        'privilege-level' => "root"
    );

    $self->log( 2, "$Package: INFO: $volName is mounted at $path" );
    $self->log( 3, "$Package: INFO: exiting _mountAndgetPath" );
    return $path;
}
####################################################################################################
# stopIO
# Stop hammer and filersio if they have been initiated
# I/P     : type of IO
# return  : None
####################################################################################################


sub stopIO {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            typeOfIO => { type => SCALAR, optional => 0, default => 'filersio' },
        }
    );

    $self->log( 3, "$Package: INFO: Entering stopIO" );
    $self->log( 2, "$Package: INFO: Stopping $opts{typeOfIO} " );

    if ( $self->{node}->mode() eq 'CMode' ) {
        if ( $opts{volume} =~ m/filersio/i ) {
            NACL::C::SystemNode->run(
                command_interface => $self->{node},
                'node'            => $self->{nodeName},
                command           => "\"priv set test; filersio stop \""
                );
        }
        elsif ( $opts{volume} =~ m/hammer/i ) {
            NACL::C::SystemNode->run(
                command_interface => $self->{node},
                'node'            => $self->{nodeName},
                command           => "\"priv set test; hammer abort \""
            );
        }
    } else {
        if ( $opts{volume} =~ m/filersio/i ) {
            $self->{nodeApiset}->execute_command( command => "priv set test; filersio stop");
        } elsif ( $opts{volume} =~ m/hammer/i ) {
            $self->{nodeApiset}->execute_command( command => "priv set test; hammer abort");
        }
    }
    $self->log( 3, "$Package: INFO: exiting stopIO" );
}

####################################################################################################
# _getResidingNode
# Return the node on which the volume is residing on
# I/P     : Volume name
# return  : Node name on which the volume is residing
####################################################################################################

sub _getResidingNode {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            volume => { type => SCALAR, optional => 0 },
        }
    );
    $self->log( 3, "$Package: INFO: in getResidingNode" );
    if ( $self->{node}->mode() ne 'CMode' ) {
        return $self->{nodeName};
    }

    my $volObj = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter            => { volume => $opts{volume}},
        allow_empty       => 1,
    );

    if(!ref($volObj)) {
        $Log->warn("$Package: Warning: Volume not found, returning default node - ".$opts{volume});
        return $self->{nodeName};
    }

    my $residingOnAggr = $volObj->state->aggregate();
    $residingOnAggr = NACL::C::Aggregate->find(
        command_interface => $self->{node},
        filter            => { aggregate => $residingOn },
    );

    my @residingOnNodes = $residingOnAggr->state->nodes();
    ## NOTE: For now lets not consider striped aggrs
    $self->log( 3, "$Package: INFO: exiting getResidingNode" );
    return $residingOnNodes[0];
}

####################################################################################################
# _convertSize
# Convert the size into bytes
# I/P     : size
# return  : size in bytes
####################################################################################################

sub _convertSize() {
    my $self = shift;
    my %opts = validate(
        @_, {
            size => { type => SCALAR, optional => 0 },
        }
    );
    $self->log( 3, "$Package: INFO: in convertSize" );

    my $factor_size = 1;
    my $size        = $opts{size};
    $factor_size    = 1024 * 1024        if ( $size =~ m/m/i );
    $factor_size    = 1024               if ( $size =~ m/k/i );
    $factor_size    = 1024 * 1024 * 1024 if ( $size =~ m/g/i );
    $self->log( 3, "$Package: INFO: exiting convertSize" );
    return $size * $factor_size;
}

####################################################################################################
# convertTime
# Convert the run time into seconds. Returns a large number if runtime = -1
# I/P     : Run time
# return  : Run time in seconds
####################################################################################################

sub convertTime {
    my $self = shift;
    my %opts = validate(
        @_, {
            runTime => { type => SCALAR, optional => 0 },
        }
    );

    $self->log( 3, "$Package: INFO: in convertTime" );
    my $factorSize = 1;
    my $runTime     = $opts{runTime};

    if ( $runTime =~ m/h/i) {
        $factorSize    = 60*60;
    } elsif ( $runTime =~ m/m/i ) {
        $factorSize    = 60;
    } elsif ( $runTime =~ m/d/i ) {
        $factorSize    = 60*60*24;
    }
    $runTime  =~ s/[^0-9\-]+//g;
    $runTime  = $runTime * $factorSize;
    $self->log( 2, "$Package: INFO: IO will run for $runTime seconds" );
    $self->log( 3, "$Package: INFO: exiting convertTime" );
    if($runTime =~ m/-1/) {
    # returning a large number x tends to infinity, we cant use -1 uniformly across all IOs, hence using a very large time interval (317 years)
        return 10000000000;
    }
    return $runTime;
}
1;

package Utility;
use Data::Dumper;
use Params::Validate qw(:all);
use Tharn qw ($Log);

sub new {
    my $class = shift;
    my $self =
      validate( @_,
        { debugLevel => { type => SCALAR, optional => 1, default => 0 }, } );
    bless $self, $class;
    return $self;
}

# Print the info
sub print {
    my $self = shift;
    my (@output) = @_;
    print @output;
}

# Print the log information  based on the debug level
sub log {
    my $self = shift;
    my ( $level, @output ) = @_;
    my $op="";
    if ( $level <= $self->{debugLevel} ) {
        $op= $op . "    " x ( $level - 1 ) . "@output";
        $Log->comment("$op");
    }
}

# Return unique entries from input array
sub uniq {
    my $self = shift;
    my %hash;
    return grep { !$hash{$_}++ } @_;
}

# Trims preceeding and ending white spaces in the string
sub trim {
    my $self = shift;
    print "I am in utility";
    my $str = shift;
    $str =~ s/^\s*//;
    $str =~ s/\s*$//;
    return $str;
}

sub remove_array_item {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            array => { type => ARRAYREF },
            item  => { type => SCALAR },
        }
    );
    my $arrayref = $opts{array};
    my $item     = $opts{item};

    my $index = 0;
    while ( $index <= $#{$arrayref} ) {
        my $value = ${$arrayref}[$index];
        if ( $value eq $item ) {
            splice @{$arrayref}, $index, 1;
        }
        else {
            $index++;
        }
    }
    return $arrayref;
}

1;
