package IOMan;

# Compiler directives
use strict;

# Imported modules
use Data::Compare;
use Params::Validate qw ( :all );
use IOMan::ClientIO;
use IOMan::IOManUtil;

our @ISA = qw( IOMan::IOManUtil );

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:     new
# Objective:  Create new IOMan object
# Details:    See POD at the end of this file
###############################################################################
sub new {
    my ($class, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my ($debug_level);
    my $self = {
        clientio_objs => [],
        generic_result_keys => [qw (dev avg_transfer_rate io_per_sec
                                   total_elapsed_time start_time
                                   end_time error_count)],
    };

    bless $self, $class;

    $self->entry(-msg_level => $IOManUtil::VERBOSE_LOW);
    &ParamsMgr::process_args(%args);
    $debug_level = &ParamsMgr::get_arg(-arg => "debug_level", -required => 0,
                                       -default => $IOManUtil::VERBOSE_LOW);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    # Set debug level
    $self->set_debug_level(-debug_level => $debug_level);
    if ($self->error()){
        $self->_set_err("failed to create IOMan", $self->errmsg());
        return $self;
    }

    $self->sanlog(-msg => "Creating IOMan Object with args " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    # _validate_params validates all parameters passed to IOMan. Currently, it
    # its use is limited but this will be useful once IOMan->new() takes
    # multiple arguments
    (%args) = $self->_validate_params(%args);
    if ($self->error()) {
        # Error set, bubble up
        return $self;
    }

    $self->{debug_level} = $debug_level;
    $self->{tool_version} = $args{tool_version};
    $self->{failure_action} = $args{failure_action};
    $self->{failure_policy} = $args{failure_policy};
    $self->{testing_method} = $args{testing_method};

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $self;
}

###############################################################################
# Method:    create_io_records
# Objective: Create io recs which are user handles to control io runs
# Details:   See POD at the end of this file
###############################################################################
sub create_io_records {
    my ($self, %args) = @_;

    my ($client, $clientio_obj);
    my (@io_records);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Creating IO recs with args :" .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);
    $client = &ParamsMgr::get_arg(-arg => "client", -required => 1);
    $args{tool_version} = &ParamsMgr::get_arg(
                                -arg => "tool_version",
                                -optional => 1,
                                -default => $self->{tool_version});
    $args{failure_action} = &ParamsMgr::get_arg(
                                -arg => "failure_action",
                                -optional => 1,
                                -default => $self->{failure_action});
    $args{failure_policy} = &ParamsMgr::get_arg(
                                -arg => "failure_policy",
                                -optional => 1,
                                -default => $self->{failure_policy});
    $args{testing_method} = &ParamsMgr::get_arg(
                                -arg => "testing_method",
                                -optional => 1,
                                -default => $self->{testing_method});

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # For every unique io_run that user intends to do, we create io_objs. These
    # are created and maintained by ClientIO on per client basis. Since only
    # IOMan.pm methods are exposed to users, we need a wrapper around ClientIO
    # create_io_records. This method serves fulfills that wrapper role and just
    # requests ClientIO to create io_records.
    # On the other end, ClientIO, creates a Manage process (the creation happens
    # only the first time) that continously runs and takes requests for io_obj
    # actions (like create/start/stop etc)
    # For create_io_records, after it creates Manage process and sending a
    # create request to Manage process, it creates an io_record which is a hash
    # with all the details of an io run. Below is an example of how an io_record
    # looks:
    # {
    #      '-devices' => [
    #                       'C:\\ioman_test\\Pattern.txt',
    #                       'C:\\ioman_test\\Pattern1.txt'
    #                    ],
    #      '-runtime' => 10,
    #      '-iotype' => 'random',
    #      '-passes' => 1,
    #      '-write' => 1,
    #      '-bs' => '4k',
    #      '-client' => 'ssan-w2k8-3',
    #      '-limit' => '50k',
    #      '-pattern' => 'a',
    #      '-io_tool' => 'sio'
    # };
    #
    # This then gets returned to this method. After validation this io_record
    # then gets returned to the user. After user gets this io_record all
    # possible actions that can be done on an io run can be done with this
    # io_record. Most of the APIs in IOMan.pm take io_records as inputs and acts
    # on them by sending requests to ClientIO, which in turn sends those
    # requests to Manage.pm. The communication between user and IOMan is usually
    # in terms of io_record

    # As described above IOMan creates one clientio obj per client and makes use
    # of it to do all operations on io objs. If clientio_obj doesn't exist for a
    # client we create one and then use it to create io_records.
    if ($self->_clientio_obj_exist(-client => $args{client})) {
        $clientio_obj = $self->_get_clientio_obj(-client => $args{client});
    } else {
        # If create clientio_obj
        $clientio_obj = $self->_create_clientio_obj(
                            -client => $args{client},
                            -debug_level => $self->{debug_level}
                        );
        if ($clientio_obj->error()){
            $self->_set_err("Failed to create clientIO obj ",
                            $clientio_obj->errmsg());
            return (wantarray()) ? () : [];
        }
    }

    # delete client argument. The clientio_obj knows the client and its
    # create_io_objs method doesn't take client as an argument
    delete $args{client};

    # At this point, we have a valid clientio_obj for the client. Use it
    # to call create_io_records. create_io_records returns io_records, each
    # io_record represents one instance of a tool run. It contains all the
    # required params to run IO
    @io_records = $clientio_obj->create_io_records(%args);
    if ($clientio_obj->error()){
        $self->_set_err("Failed to create io recs ", $clientio_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return (wantarray()) ? @io_records : \@io_records;
}

###############################################################################
# Method:    create_io_objs
# Objective: This exists for backward compatibility.
# Details:   See POD at the end of this file
###############################################################################
sub create_io_objs {
    my ($self, %args) = @_;

    return $self->create_io_records(%args);
}

###############################################################################
# Method:    start
# Objective: start runs for ids passed
# Details:   See POD at the end of this file
###############################################################################
sub start {
    my ($self, %args) = @_;

    my ($clientio_obj, $ids, $io_records, $is_running_wait);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Starting IO objs with args :" .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # An id identifies a unique io run. It is a combination of client + tool +
    # device. io_record is a hash with all params needed for running io
    # including id. Users are allowed to pass either ids (regex is supported) or
    # io_records Eventually usage of ids will be deprecated.
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1, default => ".*");
    $is_running_wait = &ParamsMgr::get_arg(-arg     => "is_running_wait",
                                           -optional => 1,
                                           -default => 4);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # validate io_records if passed. Every io_record passed must match with
    # atleast one active io_record registered within ioman. Otherwise set_err
    # and return
    if (defined $io_records) {
        $self->_validate_io_records(-io_records => $io_records);
        if ($self->error()){
            $self->_set_err("Failed to validate io obj params", $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # Pass @io_records to each clientio_obj's start and let clientio_obj handle
    # the io start of matched io_records
    # If error is returned because ids doesn't match ignore it and proceed to
    # next clientio_obj
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        $self->sanlog(-msg => "Processing clientio obj :" .
                              $self->dump_output($clientio_obj),
                      -msg_level => $IOManUtil::DEBUG_HIGH);

        $clientio_obj->start(-io_records => $io_records,
                             -ids => $ids,
                             -is_running_wait => $is_running_wait);
        if ($clientio_obj->error()) {
            if ($clientio_obj->errmsg() =~ /No Matching ids/i) {
                $clientio_obj->_clear_err();
                next;
            } else {
                $self->_set_err("Failed to start io objs ",
                                $clientio_obj->errmsg());
            }
        }
    }

    # Note that we are not stopping the loop when we encounter a start error
    # All the failures will be concatenated and returned to the user. After
    # failure policy is implemented this will handled gracefully
    if ($self->error()) {
        $self->_set_err("Failed to start runs for all io objs ",
                        $self->errmsg());
        return $IOManUtil::FALSE
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    wait_finish
# Objective: wait until io runs to finish
# Details:   See POD at the end of this file
###############################################################################
sub wait_finish {
    my ($self, %args) = @_;

    my ($clientio_obj, $io_records, $ids);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Waiting for io runs to finish. Params passed :" .
                           $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Users are allowed to pass ids or io_records. Each element identifies an
    # io_obj
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1, default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # validate io_records if passed. Every io_record passed must match with
    # atleast one active io_record registered within ioman. Otherwise set_err
    # and return
    if (defined $io_records) {
        $self->_validate_io_records(-io_records => $io_records);
        if ($self->error()){
            $self->_set_err("Failed to validate io obj params", $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # Loop through all clientio_objs and call wait_finish()
    # $clientio_obj->wait_finish() waits until all io objs runs it started are
    # finished
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        $clientio_obj->wait_finish(-io_records => $io_records, -ids => $ids);
        if ($clientio_obj->error()) {
            $self->_set_err("Failed to wait until all io objs run is finished ",
                            $clientio_obj->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    stop
# Objective: stop io runs
# Details:   See POD at the end of this file
###############################################################################
sub stop {
    my ($self, %args) = @_;

    my ($clientio_obj, $io_records, $ids);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Request received to stop io. Params passed :" .
                           $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Users are allowed to pass ids or io_records. Each element identifies an
    # io_obj
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1, default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # validate io_records if passed. Every io_record passed must match with
    # atleast one active io_record registered within ioman. Otherwise set_err
    # and return
    if (defined $io_records) {
        $self->_validate_io_records(-io_records => $io_records);
        if ($self->error()){
            $self->_set_err("Failed to validate io obj params", $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # Loop through all clientio_objs and call stop()
    # $clientio_obj->stop() stops all io objs runs on a specific client
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        $clientio_obj->stop(-io_records => $io_records, -ids => $ids);
        if ($clientio_obj->error()) {
            $self->_set_err("Failed to stop all io obj runs ",
                            $clientio_obj->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    is_running
# Objective: whether the io_records are running or not
# Details:   See POD at the end of this file
###############################################################################
sub is_running {
    my ($self, %args) = @_;

    my ($clientio_obj, $io_records, $ids, $on_fail);
    my (@results) = ();

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "verifying whether io recs are running " .
                           $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Users are allowed to pass ids or io_records. Each element identifies an
    # io_obj
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $on_fail = &ParamsMgr::get_arg(-arg => "on_fail", -optional => 1,
                                   -default => undef);
    $ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1, default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    if (defined $on_fail and $on_fail !~ /ignore/ ) {
        $self->_set_err("on_error failedregex validation /ignore/ ");
        return $IOManUtil::FALSE;
    }

    # validate io_records if passed. Every io_record passed must match with
    # atleast one active io_record registered within ioman. Otherwise set_err
    # and return
    if (defined $io_records) {
        $self->_validate_io_records(-io_records => $io_records);
        if ($self->error()){
            $self->_set_err("Failed to validate io obj params", $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # Loop through all clientio_objs and call is_running()
    # if any of the $clientio_obj->is_running() returns false, it means that
    # we found atleast one io_obj that is not running. In such a case, return
    # false
    my $is_running = 0;
    my $not_running = 0;
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        my $status = $clientio_obj->is_running(-io_records => $io_records,
                                               -ids        => $ids,
                                               -on_fail    => $on_fail);
        if ($clientio_obj->error()) {
            $self->_set_err("Failed to verify if the io_records are running ",
                            $clientio_obj->errmsg());
            return $IOManUtil::FALSE;
        }
        #
        # If any test is still running, TRUE is returned.
        # If any test failed or nothing is running, FALSE is returned.
        #
        if ($status) {
            $is_running++;
        } else {
            $not_running++;
        }
    }

    #
    # If we are looking at results and something has completed, then check for failures.
    # Also, when *all* tests have completed, also check for failures (existing logic).
    #
    if ( (($on_fail ne "ignore") && ($not_running > 0)) || ($is_running == 0) ) {

        # If an error is set in any of the results, set error message
        @results = $self->get_result(-io_records  => $io_records,
                                     -ids         => $ids,
                                     -wait_finish => 0);
        if ($self->error()) {
            $self->_set_err("Failed to check if io run resulted in failure " .
                            $self->errmsg());
            return $IOManUtil::FALSE;
        }
        foreach my $result (@results){
            if ($result->{error_count}) {
                $self->_set_err("IO Failure detected for id " . $result->{id} .
                                "\n" . $result->{error})
            }
        }
        if ($self->error()){
            return $IOManUtil::FALSE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    if ($is_running > 0) {
        return $IOManUtil::TRUE;
    } else {
        return $IOManUtil::FALSE;
    }
}

###############################################################################
# Method:    get_result
# Objective: return result objs from all clientio objs
# Details:   See POD at the end of this file
###############################################################################
sub get_result {
    my ($self, %args) = @_;

    my ($clientio_obj, $wait_finish, $io_records, $ids, $results_ref);
    my (@results);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Getting results " . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Users are allowed to pass ids or io_records. Each element identifies an
    # io_obj. wait_finish arg controls whether we need to wait for all matched
    # io_obj runs to finsih or not. By default, its set to 1, which means we
    # wait for io_obj runs to finish. If set to 0, we return whatever the
    # current result is.
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1, default => ".*");
    $wait_finish = &ParamsMgr::get_arg(-arg => "wait_finish", -optional => 1,
                                       -default => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # validate io_records if passed. Every io_record passed must match with
    # atleast one active io_record registered within ioman. Otherwise set_err
    # and return
    if (defined $io_records) {
        $self->_validate_io_records(-io_records => $io_records);
        if ($self->error()){
            $self->_set_err("Failed to validate io obj params", $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # Loop through all clientio_objs and get result from each of them. Push
    # them onto a array and return
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        my $results_ref = $clientio_obj->get_result(-io_records => $io_records,
                                                    -ids => $ids,
                                                    -wait_finish => $wait_finish);
        if ($clientio_obj->error()) {
            $self->_set_err("Failed to get result obj ",
                            $clientio_obj->errmsg());
            return (wantarray()) ? () : [];
        }
        push (@results, @$results_ref);
    }

    $self->sanlog(-msg => "Results being returned is " .
                          $self->dump_output(\@results),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return (wantarray()) ? @results : \@results;
}

###############################################################################
# Method:    display_results
# Objective: displays results from all io objs
# Details:   See POD at the end of this file
###############################################################################
sub display_results {
    my ($self, %args) = @_;

    my ($clientio_obj, $msg, $result, $ids, $io_records);
    my (@results);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Users are allowed to pass ids or io_records. Each element identifies an
    # io_obj.
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1, default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # validate io_records if passed. Every io_record passed must match with
    # atleast one active io_record registered within ioman. Otherwise set_err
    # and return
    if (defined $io_records) {
        $self->_validate_io_records(-io_records => $io_records);
        if ($self->error()){
            $self->_set_err("Failed to validate io obj params", $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # This method is just a wrapper that gets the results and prints them in
    # a pretty format. Only the generic result keys are printed. Other gets lost
    # Users are encouraged to use get_result() and get array of results hash
    # if they want everything returned by the tool
    @results = $self->get_result(-io_records => $io_records,
                                 -ids => $ids);
    if ($self->error()) {
        $self->_set_err("Unable to get result " . $self->errmsg());
        return (wantarray()) ? () : [];
    }
    $self->_print_results(-results => \@results);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return (wantarray()) ? @results : \@results;
}

###############################################################################
# Method:    get_io_records
# Objective: Returns matching io recs
# Details:   See POD at the end of this file
###############################################################################
sub get_io_records {
    my ($self, %args) = @_;

    my ($io_record, $key, $device, $devices, $index, $clientio_obj);
    my (@io_records, @tmp_io_records);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # This method is a filter for all io_records present within ioman. All
    # io_records that match the arguments  will be returned.
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "IOMan::get_io_records called to find matched " .
                          "io_records with args" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    # Loop through all clientio_objs, pool all matched io_records and return
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        my @tmp_io_records = $clientio_obj->get_io_records(%args);
        if ($clientio_obj->error()) {
            $self->_set_err("Failed to get io_records ",
                            $clientio_obj->errmsg());
            return (wantarray()) ? () : [];
        }
        push (@io_records, @tmp_io_records);
    }

    $self->sanlog(-msg => "IOMan::get_io_records. Returning" .
                          $self->dump_output(@io_records),
                  -msg_level => $IOManUtil::DEBUG_LOW);
    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return (wantarray()) ? @io_records : \@io_records;
}

###############################################################################
# Method:    get_params
# Objective: returns param for an id
# Details:   See POD at the end of this file
###############################################################################
sub get_params {
    my ($self, %args) = @_;

    my ($id, $params_ref, $client, $client_obj, $io_record);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    %args = &SanUtil::strip_dash(%args);

    $self->_clear_err();
    &ParamsMgr::process_args(%args);

    # Either io_record or id needs to be passed. If an io_record is passed,
    # we get the id from it.
    $io_record = &ParamsMgr::get_arg(-arg => "io_record", default => undef);
    unless ($id = $io_record->{id}) {
        $id = &ParamsMgr::get_arg(-arg => "id", -required => 1);
    }

    if (&ParamsMgr::error()) {
            $self->_set_err(&ParamsMgr::get_error());
            return {};
    }

    # First get the clientio obj. The id is of the format:
    # <client>::<io_tool>::<device_id>
    ($client, undef, undef) = split (/::/, $id);

    $client_obj = $self->_get_clientio_obj(-client => $client);
    if ($self->error()) {
        # Error already set
        return {};
    }

    # get_params method of client_obj is called to get a ref to params
    $params_ref = $client_obj->get_params(-id => $id);
    if ($client_obj->error()) {
        $self->_set_err($client_obj->errmsg());
        return {};
    }
    $self->sanlog(-msg => "get params from clienio_obj returned" .
                           $self->dump_output($params_ref),
                  -msg_level=> $IOManUtil::DEBUG_LOW);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $params_ref;
}

###############################################################################
# Method:    cleanup
# Objective: calls cleanup for all clientio_objs
# Details: See POD at the end of this file
###############################################################################
sub cleanup {
    my ($self) = @_;

    my ($clientio_obj);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();

    # Loop through all clientio objs and call cleanup
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        $self->sanlog(-msg => "Processing clientio obj :" .
                              $self->dump_output($clientio_obj),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        $clientio_obj->cleanup();
        if ($clientio_obj->error()) {
            $self->_set_err("Failed to complete clientio cleanup ",
                            $clientio_obj->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    # Reset clientio_objs
    @{$self->{clientio_objs}} = ();

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}


###############################################################################
# Method:    _create_clientio_obj
# Objective: Create ClientIO obj
# Details:   See POD at the end of this file
###############################################################################
sub _create_clientio_obj {
    my ($self, %args) = @_;
    my ($client, $clientio_obj);

    %args = &SanUtil::strip_dash(%args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Creating ClientIO obj " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    &ParamsMgr::process_args(%args);
    $client = &ParamsMgr::get_arg(-arg => "client", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Validation of args is not done here. It will be done in ClientIO->new()
    $clientio_obj = IOMan::ClientIO->new(%args);
    if ($clientio_obj->error()){
        $self->_set_err("Failed to create clientio obj ",
                        $clientio_obj->errmsg());
        return $clientio_obj;
    }

    # Push clientio_obj that is just created to $self. A list of clientio_objs
    # is maintained so that subsequent requests for create, start of io_records
    # can be done using these clientio_objs
    push (@{$self->{clientio_objs}}, $clientio_obj);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $clientio_obj;
}

###############################################################################
# Method:    _get_clientio_obj
# Objective: Returns clientio_obj
# Details:   See POD at the end of this file
###############################################################################
sub _get_clientio_obj {
    my ($self, %args) = @_;

    my ($client, $clientio_obj);

    %args = &SanUtil::strip_dash(%args);
    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    &ParamsMgr::process_args(%args);
    $client = &ParamsMgr::get_arg(-arg => "client", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    $self->sanlog(-msg => "Getting Client io obj for " . $client .
                          $self->dump_output(%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Loop through all clientio_objs and if the client name matches, return
    # that clientio_obj
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        if ($clientio_obj->get_client(-client => $client) eq $client) {
            $self->sanlog(-msg => "Client io obj for $client found",
                          -msg_level => $IOManUtil::DEBUG_HIGH);
            $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
            return $clientio_obj;
        }
    }

    $self->sanlog(-msg => "Client io obj for $client not found",
                          -msg_level => $IOManUtil::DEBUG_HIGH);
    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return {};
}

###############################################################################
# Method:    _clientio_obj_exist
# Objective: Returns true if clientio_obj exists else false
# Details:   See POD at the end of this file
###############################################################################
sub _clientio_obj_exist {
    my ($self, %args) = @_;

    my ($client, $clientio_obj);

    %args = &SanUtil::strip_dash(%args);
    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    &ParamsMgr::process_args(%args);
    $client = &ParamsMgr::get_arg(-arg => "client", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Loop through all clientio_objs and return true if there is a clientio_obj
    # that matches with name of the client parameter
    foreach $clientio_obj (@{$self->{clientio_objs}}) {
        if ($clientio_obj->get_client(-client => $client) eq $client) {
            $self->sanlog(-msg => "Client io obj for $client found",
                          -msg_level => $IOManUtil::DEBUG_HIGH);
            $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
            return $IOManUtil::TRUE;
        }
    }

    $self->sanlog(-msg => "Client io obj for $client not found",
                          -msg_level => $IOManUtil::DEBUG_HIGH);
    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $IOManUtil::FALSE;
}

###############################################################################
# Method:    _print_results
# Objective: This method takes an array ref, loops through all keys and prints
#            the output in a nice format
# Details: See POD at the end of this file
###############################################################################
sub _print_results {
    my ($self, %args) = @_;

    my ($results_ref, $result, $key, $msg);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    $results_ref = &ParamsMgr::get_arg(-arg => "results", default => []);

    unless (scalar @$results_ref) {
        $self->_set_err("At least one io_record result needs to be passed");
        return $IOManUtil::FALSE;
    }

    foreach $result (@$results_ref) {
        $self->sanlog(-msg => "Results found " . $self->dump_output($result),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        $msg = "Client Used: ". $result->{client} . ", Tool Used: " .
               $result->{io_tool} . ", Device: " . $result->{dev};

        $msg = $self->_print_header($msg);
        foreach $key (@{$self->{generic_result_keys}}) {
            $msg = $msg . $key . " is " . $result->{$key} . "\n";
        }

        $self->sanlog(-msg => $msg);
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _validate_params
# Objective: Validate params
# Details: See POD at the end of this file
###############################################################################
sub _validate_params {
    my ($self, %args) = @_;

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    my @input_params = (%args);

    eval {
        (%args) = Params::Validate::validate( @input_params,
                      { debug_level    => { type => SCALAR,
                                            optional => 1,
                                            regex => qr /^\d+$/},
                        tool_version   => { type     => SCALAR,
                                            optional => 1,
                                            regex    => qr /(stable|current|staging|last|^\d[_\.]\d+$)/,
                                            default  => 'stable'},
                        failure_action => { type     => SCALAR,
                                            optional => 1,
                                            default  => "" },
                        failure_policy => { type     => SCALAR,
                                            optional => 1,
                                            regex    => qr /(ignore|stop_all|stop_all_client)/,
                                            default  => 'ignore'},
                        testing_method => { type     => SCALAR,
                                            optional => 1,
                                            regex    => qr /(positive|negative)/,
                                            default  => 'positive'},
                      });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: ". $@);
        return ();
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return %args;
}

###############################################################################
# Method:    _validate_io_records
# Objective: Returns matching io recs
# Details:   See POD at the end of this file
###############################################################################
sub _validate_io_records {
    my ($self, %args) = @_;

    my ($io_records, $key, $device, $devices, $index, $clientio_obj,
        $match_found);
    my (@tmp_io_records);

    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Takes an array ref to io_records and returns true if all the io_records
    # passed are registered within ioman. This allows us to catch cases where
    # users pass invalid io_records.
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    $self->sanlog(-msg => "IOMan::validate_io_records called to find matched " .
                          "io_records with args" .
                           $self->dump_output($io_records),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Loop through all io_records, and try to find a match with  get_io_records
    # returned from any of the clientio_objs
    foreach my $io_record (@$io_records) {
        $match_found = $IOManUtil::FALSE;
        foreach $clientio_obj (@{$self->{clientio_objs}}) {
            my @tmp_io_records = $clientio_obj->get_io_records(
                                     -client => $clientio_obj->get_client()
                                 );
            if ($clientio_obj->error()) {
                $self->_set_err("IOMan::validate_io_records: Failed to get " .
                                "io_records ", $clientio_obj->errmsg());
                return $IOManUtil::FALSE;
            }

            # If current io_record hash data matches with one of the io_records
            # already created then set match_found
            if (List::MoreUtils::any
                    { Compare($_, $io_record) } @tmp_io_records) {
                $match_found = $IOManUtil::TRUE;
            }
        }

        unless ($match_found) {
            $self->_set_err("IOMan::validate_io_records: No match found" .
                            $self->dump_output($io_record));
            return $IOManUtil::FALSE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $SanUtil::TRUE;
}

###############################################################################
# Method:    kill_io_tools
# Objective: ensure I/O tools are not running on test client
###############################################################################
sub kill_io_tools {
    my ($self, %args) = @_;
    my ($exe_to_kill, @exes_to_kill);
    my (@subtests);

    &ParamsMgr::process_args(%args);
    my $client = &ParamsMgr::get_arg(-arg => "client", -required => 1, -default => "");
    my $io_tool = &ParamsMgr::get_arg(-arg => "io_tool", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # determine the process application to kill (all instances)
    if ($io_tool =~ /dt/ or $io_tool =~ /all/i) {
        push (@exes_to_kill, "dt");
    }

    if ($io_tool =~ /sio/ or $io_tool =~ /all/i) {
        push (@exes_to_kill, "sio");
    }

    if ($io_tool !~ /sio/ and
        $io_tool !~ /dt/ and
        $io_tool !~ /all/i) {
        $self->_set_err("No valid io_tool passed in ($io_tool)");
        return $IOManUtil::FALSE;
    }

    foreach $exe_to_kill (@exes_to_kill) {
        # kill the I/O process still on the host (all platforms)
        my $subtest = &SubIPC::create(-id => "io_task_kill_" . $exe_to_kill . "_" . $client,
                                      -run => sub {
           &ClientUtil::kill_process_instances($client, $exe_to_kill);
           &SubIPC::pass(-msg => "I/O Tasks for tool '$exe_to_kill' have been cleaned up on $client");
           return $SanMgr::TRUE;
                                      },
                                      -parent => $self,
                                      -ipc => 1, -bg => 1);

        if (&SubIPC::error()) {
            $self->_set_err("Warning: Failed to kill I/O Tasks for tool '$exe_to_kill' on $client ",
                            &SubIPC::get_error());
            next;
        }
        push (@subtests, $subtest);
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish(-subtest => [@subtests]);
    if (&SubIPC::error()) {
        $self->_set_err("Failed while reaping volume io killing subtests",
                        &SubIPC::get_error());
        return $IOManUtil::FALSE;
    }
    return $IOManUtil::TRUE;
}

1;
