package IOMan::ClientIO;

# Compiler directives
use strict;

# Imported modules
use Tharn;
use NATE::Process;
use Params::Validate qw ( :all );
use IOMan::ClientIO::Manage;
use List::MoreUtils qw(all any first_index);
use NATE::Log qw(log_global);

our @ISA = qw(IOMan::IOManUtil);
###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new ClientIO object
# Details:   See POD at the end of this file
###############################################################################
sub new {
    my ($class, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my ($debug_level, $manage_id);
    my $self = {
        io_obj_ids     => [],
        manage         => {},
        client         => '',
        io_records     => []
    };

    bless $self, $class;

    &ParamsMgr::process_args(%args);
    $debug_level = &ParamsMgr::get_arg(-arg => "debug_level", -required => 0,
                                       -default => $IOManUtil::DEBUG_LOW);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    $self->set_debug_level(-debug_level => $debug_level);
    if ($self->error()) {
        $self->_set_err("Cannot set debug level", $self->errmsg());
        return $self;
    }

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Creating ClientIO object with args " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    (%args) = $self->_validate_params(%args);
    if ($self->error()) {
        # Error set, bubble up
        $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
        return $self;
    }

    $self->{client} = $args{client};
    $self->sanlog(-msg => "Starting Manage subtest for " . $self->{client},
                  -msg_level => $IOManUtil::DEBUG_LOW);
    $manage_id = $args{client} . '__Manage';

    # ClientIO new automatically starts Manage::manage() subtest which
    # continously runs and takes requests for create, start, get_result and
    # wait_finish of io_objs ClientIO and manage() runs in two different
    # processes. ClientIO runs in same process as IOMan and thpl. Its methods
    # are mostly wrappers that sends messages to manage() and gets work done
    $self->{manage} = NATE::Process->new(
        runid => $manage_id,
        codespec => '{#IOMAN ' . $args{client} . '
               use IOMan::ClientIO::Manage;
               IOMan::ClientIO::Manage->manage(-client => "'. $args{client} . '",
                                               -debug_level => "' . $debug_level . '");}',
        client => 'localhost',
        onexit => \&_manage_proc_onexit
    );
    $self->{manage}->start();

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $self;
}

sub _manage_proc_onexit {
    my ($st) = @_;
    my $res;
    if ($res = $st->worst_result() and $res->severity() > 0) {
        log_global()->warn("The IO Manage object for a client failed");
    }
}

###############################################################################
# Method:    create_io_records
# Objective: Create io recs
# Details:   See POD at the end of this file
###############################################################################
sub create_io_records {
    my ($self, %args) = @_;

    my (@io_objs, @devices, @io_tools, @io_records);
    my ($io_obj, $io_tool, $device, $io_obj_id, $device_id, $message);
    my ($device_count, $device_index, $devices_per_tool, $device_sep);

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # io_records are users handle over the io_objs. Programmatically, they are
    # just a hash. This method takes args required for Manage process to
    # create io_objs, validates the args, sends a request to Manage process
    # to create io_objs, waits for acknowledgement from Manage process to make
    # sure that the job handed over to manage process is done.
    # After the ack is returned, it in turn returns io_records which is a
    # handle to the objs created by Manage process to the users
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "ClientIO::create_io_records called for creating ".
                          " IO objs with params" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->sanlog(-msg => "Validating params for creating io objs ",
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # validate create_io_records arguments
    (%args) = $self->_validate_create_io_records_params(%args);
    if ($self->error()){
        $self->_set_err("Failed to validate io obj params", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Set the tool path based on the following rules:
    # 1) If user passes tool_path arg just return it
    # 2) If user passes tool_version, determine hosttype and return path using
    #    IOManUtil.pm
    # After tool_path is set, there is no need of tool_version. tool_version
    # is taken into account by _get_tool_path while determining tool_path
    $args{tool_path} = $self->_get_tool_path(%args);
    if ($self->error()){
        $self->_set_err("Failed to determine tool path", $self->errmsg());
        return (wantarray()) ? () : [];
    }
    delete $args{tool_version};

    # manage() requires one request to be sent for every io_obj. If there are
    # multiple devices we send that many requests
    #
    # We now support multiple devices per tool, based on 'devices-per-tool':
    #  - the default remains the same if count is 1 or less.
    #  - the keyword "all" will send all devices to one tool.
    #  - otherwise a value specified how many devices per tool.
    #
    # IOMan supports either dt or sio IO tools.
    #  - dt's multiple devices supports a comma separated device list.
    #  - sio's multiple devices accepts a space separated device list.
    #
    @devices = @{delete $args{devices}};
    $device_count = scalar(@devices);
    $devices_per_tool = delete $args{devices_per_tool};
    if ($devices_per_tool eq "all") {
        $devices_per_tool = $device_count;
    } elsif ($device_count < $devices_per_tool) {
        $devices_per_tool = $device_count;
    }
    if ($devices_per_tool > 1) {
        # Device separator for dt is comma, for sio it's space.
        $device_sep = ($args{io_tool} eq "dt") ? ',' : ' ';
        if ($device_count == $devices_per_tool) {
            @devices = join($device_sep, @devices);
        } else {
            my @devs;
            my @all_devs = @devices;
            @devices = ();
            foreach $device (@all_devs) {
                push(@devs, $device);
                if (scalar(@devs) == $devices_per_tool) {
                    push(@devices, join($device_sep, @devs));
                    @devs = ();
                }
            }
            if (scalar(@devs)) {
                push(@devices, join($device_sep, @devs));
            }
        }
        $device_index = 1;
    }

    # We let user pass write => 1 but convert this to read => 0 in
    # _validate_create_io_records_params(). write key should be deleted here
    # since Manage doesn't expect it
    delete $args{write} if exists $args{write};

    # Looping through all devices passed and sending one request to Manage
    # process per device. id for io_obj is of the format
    # <client>::<io_tool>::<device_id>device_id is the device path with
    # \ and / substituted by :
    foreach $device (@devices) {
        if ($devices_per_tool <= 1) {
            $device_id = $device;
        } else {
            if ($device_count eq $devices_per_tool) {
                $device_id = "::MultipleDevices-$device_count";
            } else {
                my $count = 1;
                while ($device =~ m/$device_sep/g) {
                    $count++;
                }
                $device_id = "::MultipleDevices-$device_index:$count";
                $device_index += $count;
            }
        }
        $device_id =~ s/(\\|\/)/:/g;
        $io_obj_id = $self->{client} . "::" . $args{io_tool} . "::" .
                     $device_id;

        # Note that %io_obj_args needs to be declared within the foreach loop
        # That way the hashref (\%io_obj_args) we are pushing to @io_records
        # will be different with each iteration
        my %io_obj_args = ( %args,
                            device => $device,
                            id     => $io_obj_id);

        $self->sanlog(-msg => "Writing to Manage process :" .
                              $self->dump_output(\%io_obj_args),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        $self->{manage}->message_put(message => \%io_obj_args,
                                     type    => "create_io_obj");

        $message = $self->{manage}->message_get(type  => "create_io_ack");
        unless ($message eq "creation done") {
            $self->_set_err("creation failed with error message", $message);
            return (wantarray()) ? () : [];
        }

        $self->sanlog(-msg => "Creation complete for " .
                              $self->dump_output(\%io_obj_args) .
                              " Ack received " .
                              $self->dump_output($message),
                      -msg_level=> $IOManUtil::DEBUG_HIGH);

        # Push io_obj_id to io_obj_ids internal array maintained by ClientIO.pm
        # only if its unique
        unless (List::MoreUtils::any
                    { $_ eq $io_obj_id } @{$self->{io_obj_ids}}) {
            push (@{$self->{io_obj_ids}}, $io_obj_id);
        }

        $io_obj_args{client} = $self->{client};

        # io_record is nothing but args passed to Manage for creating io_obj in
        # addition to client name. Push the io_record associated with newly
        # created io_obj onto temporary array
        push (@io_records, \%io_obj_args);
    }

    # Make sure none of these io_records are running. This is possible if the
    # user attempted a recreate on an io_recored that is already started.
    # Ideally, Manage.pm create_io_obj should return an error. This needs to
    # be implemented when the ACK from Manage.pm is modified to return hash
    # setting errors whenever its necessary. For now check if any of these
    # io_records are running. If so, return error
    if ($self->is_running(-io_records => \@io_records)) {
        $self->_set_err("Attempted to recreate io_record when io is still " .
                        "running. This is not allowed ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Save the io_records created to internal array maintained by ClientIO
    $self->_save_io_records(-io_records => \@io_records);
    if ($self->error()) {
        $self->_set_err("Failed to save io recs ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    $self->sanlog(-msg => "\$self->{io_records} is " .
                          $self->dump_output($self->{io_records}),
                  -msg_level=> $IOManUtil::DEBUG_LOW);


    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return (wantarray()) ? @io_records : \@io_records;
}

###############################################################################
# Method:    create_io_objs
# Objective: This exists for backward compatibility
# Details:   See POD at the end of this file
###############################################################################
sub create_io_objs {
    my ($self, %args) = @_;

    return $self->create_io_records(%args);
}

###############################################################################
# Method:    start
# Objective: start
# Details:   See POD at the end of this file
###############################################################################
sub start {
    my ($self, %args) = @_;

    my ($id, $ids_ref, $regex_ids, $message, $io_records, $is_running_wait,
        $results_ref);
    my (@io_records);
    my (%start_args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "ClientIO::start called to start IO objects " .
                          "with params" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Either io_records (user handle of io_objs) or ids can be passed to this
    # method.  If io_records are passed, all io_objs associated with the
    # io_record will have io started. Note that io_obj is a term associated with
    # the actual object that Manage process creates for each possible io run
    # that can happen.
    # ids can be regular expressions. All ids matched with the regex
    # will have io started. Note that only a request is sent to Manage process
    # to start io. Manage does the real work of starting io. This method only
    # waits for an acknowledgement from Manage
    #
    # is_running_duration is the time this method waits and makes sure io is
    # running before returning
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $is_running_wait = &ParamsMgr::get_arg(-arg     => "is_running_wait",
                                           -default => 4);
    $regex_ids = &ParamsMgr::get_arg(-arg => "ids",
                                     -optional => 1,
                                     -default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # io_records takes precedence over ids if both are passed. _get_ids handles
    # this for us. It returns an array of ids associated with io_records/regex.
    # The common language spoken by ClientIO and Manage is ids. This is why
    # we translate io_records to ids.
    $ids_ref = $self->_get_ids(-io_records => $io_records, -ids => $regex_ids);
    if ($self->error()){
        $self->_set_err("Failed to _get_ids by matching passed regex $regex_ids",
                        $self->errmsg());
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "Matched ids are :" .
                          $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Loop through all ids and send a request to Manage to start the run on
    # io obj. Manage can handle only one id at a time. We will do the extra
    # work here to send one request at a time. This will give use extra control
    # for handling errors while starting io runs
    foreach $id (@$ids_ref) {
        $start_args{id} = $id;
        $self->sanlog(-msg => "Starting IO objs with id :" .
                              $self->dump_output(\%start_args) .
                              "manage process" .
                              $self->dump_output($self->{manage}),
                     -msg_level => $IOManUtil::DEBUG_HIGH);
         $self->{manage}->message_put(message => \%start_args,
                                      type    => "start_io_obj");

        $message = $self->{manage}->message_get(type => "start_io_ack");
        $self->sanlog(-msg => "Ack received " . $self->dump_output($message),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        unless ($message eq "start done") {
            $self->_set_err("Start failed with error message", $message);
            return $IOManUtil::FALSE;
        }
    }

    # The below code verifies if the start actually resulted in the io run to
    # start as claimed by Manage process or whether it failed right away
    # Common cases where users ran into this are when tool is not available on
    # the client or when devices are not available. A call to is_running is
    # made to make sure that the io is running. If not, results are captured
    # and only if error count is set, we know for sure that the run resulted
    # in an error. If error count is not set and is_running failed it might
    # be because of a very short io run

    # Get io_records associated with ids. This could have been avoided by
    # passing both ids_ref and io_records to is_running but just to keep the
    # code consistent all future changes will only talk in terms of io_record.
    # The plan going forward is to deprecate the use of regex everywhere
    foreach my $id (@$ids_ref) {
        push(@io_records, $self->get_io_record (-id => $id));
    }

    unless ($self->_has_started(-wait_time => $is_running_wait,
                                -io_records => \@io_records)) {
        $self->_set_err("io_obj start failed immediately");
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "successfully started io for io_records" .
                          $self->dump_output(@io_records),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    wait_finish
# Objective: wait until all subtests are finished
# Details:   See POD at the end of this file
###############################################################################
sub wait_finish{
    my ($self, %args) = @_;

    my ($message, $ids_ref, $io_records, $regex_ids);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "ClientIO::wait_finish called to wait until finish " .
                          "with args" . $self->dump_output(\%args),
                          -msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Either io_records (user handle of io_objs) or ids can be passed to this
    # method. See documentation in start() to understand io_records/ids
    # This method waits until io on the Manage processes's io_objs associated
    # with io_records/ids passed is completed
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => []);
    $regex_ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1,
                                     -default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # _get_ids returns an array of ids associated with io_records/regex.
    # io_records takes precedence over ids
    $ids_ref = $self->_get_ids(-io_records => $io_records, -ids => $regex_ids);
    if ($self->error()){
        $self->_set_err("Failed to _get_ids by matching passed regex " .
                        $regex_ids, $self->errmsg());
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "Matched ids are" . $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    unless ($ids_ref) {
         $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
         return $IOManUtil::TRUE;
    }

    # send a message of type wait_finish to manage() and wait for wait_ack
    $self->{manage}->message_put( message  => {-ids_ref => $ids_ref},
                                  type     => "wait_finish");
    $message = $self->{manage}->message_get(type    => "wait_ack");

    $self->sanlog(-msg => "wait complete Ack received" .
                           $self->dump_output($message),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    $self->sanlog(-msg => "Waiting done" . $self->dump_output($self),
                          -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    get_result
# Objective: Returns array of results. Each result is a hash
# Details: See POD at the end of this file
###############################################################################
sub get_result {
    my ($self, %args) = @_;

    my ($ids_ref, $regex_ids, $results_ref, $wait_finish, $message,
        $io_records);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->sanlog(-msg => "ClientIO::result called to get results for IO " .
                          " objects with params" .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Either io_records (user handle of io_objs) or ids can be passed to this
    # method. See documentation in start() to understand io_records/ids
    # wait_finish controls whether this method needs to wait for io runs
    # associated with io_records/ids passed needs to complete or not
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $regex_ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1,
                                     -default => ".*");
    $wait_finish = &ParamsMgr::get_arg(-arg => "wait_finish", -optional => 1,
                                       -default => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # _get_ids returns an array of ids associated with io_records/regex.
    # io_records takes precedence over ids
    $ids_ref = $self->_get_ids(-io_records => $io_records, -ids => $regex_ids);
    if ($self->error()){
        $self->_set_err("Failed to _get_ids by matching passed regex " .
                        $regex_ids, $self->errmsg());
        return (wantarray()) ? () : [];
    }

    $self->sanlog(-msg => "Matched ids are:" .
                          $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # send a message to Manage.pm and get the result
    # We receive an array of result hashes one for each io_obj. Push all of
    # these using result_obj
    $self->{manage}->message_put(message => { ids_ref => $ids_ref,
                                              wait_finish => $wait_finish },
                                 type    => "get_result");
    $results_ref = $self->{manage}->message_get(type  => "result_ack");

    $self->sanlog(-msg => "Returned result is:" .
                          $self->dump_output($results_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return (wantarray()) ? @$results_ref : $results_ref;
}

###############################################################################
# Method:    cleanup
# Objective: kills the manage process
# Details: See POD at the end of this file
###############################################################################
sub cleanup {
    my ($self) = @_;

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # Cleanup stops the manage process which effectively means everything
    # (io_objs, results_objs etc) is lost. Much more graceful cleanup is
    # possible by stopping the start subtests of all io_objs but there is no
    # reason why a user needs that. Also doing so will take time
    $self->sanlog(-msg => "Quitting manage process" .
                           $self->dump_output($self->{manage}),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();

    eval {
        $self->{manage}->quit();
        $self->{manage}->wait();
    };
    if ($@) {
        $self->_set_err("Failed to quit manage process: ". $@);
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "After Quitting manage process" .
                           $self->dump_output($self->{manage}),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);
    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $SanUtil::TRUE;
}

###############################################################################
# Method:    is_running
# Objective: Returns $IOManUtil::TRUE if all the io_objs are running, otherwise
#            $IOManUtil::FALSE
# Details: See POD at the end of this file
###############################################################################
sub is_running {
    my ($self, %args) = @_;

    my ($status, $message, $ids_ref, $regex_ids, $io_records, $on_fail);
    my (%is_running_args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    # Either io_records (user handle of io_objs) or ids can be passed to this
    # method. See documentation in start() to understand io_records/ids
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $on_fail = &ParamsMgr::get_arg(-arg => "on_fail", -optional => 1,
                                   -default => undef);
    $regex_ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1,
                                     -default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    if (defined $on_fail) {
        $is_running_args{on_fail} = $on_fail;
    }

    $self->sanlog(-msg => "ClientIO::is_running to verify if io_objs are " .
                          " running" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # _get_ids returns an array of ids associated with io_records/regex.
    # io_records takes precedence over ids
    $ids_ref = $self->_get_ids(-io_records => $io_records, -ids => $regex_ids);
    if ($self->error()){
        $self->_set_err("Failed to _get_ids by matching passed regex $regex_ids",
                        $self->errmsg());
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "Matched ids are: " . $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    unless ($ids_ref) {
         $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
         return $IOManUtil::TRUE;
    }

    # send a message of type is_running to manage() and wait for ack
    $is_running_args{ids_ref} = $ids_ref;
    $self->{manage}->message_put( message => { %is_running_args },
                                  type     => "is_running");
    $message = $self->{manage}->message_get(type    => "is_running");

    $self->sanlog(-msg => "status received is " . $self->dump_output($message),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    # If status is empty something is wrong
    unless (%$message) {
        $self->_set_err("is_running returned nothing");
        return $IOManUtil::FALSE;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $message->{status};
}

###############################################################################
# Method:    stop
# Objective: Returns $IOManUtil::TRUE if all the io_objs are stopped, otherwise
#            $IOManUtil::FALSE
# Details:   See POD at the end of this file
###############################################################################
sub stop {
    my ($self, %args) = @_;

    my ($status, $message, $ids_ref, $regex_ids, $io_records, $on_fail);
    my (%stop_args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    # Either io_records (user handle of io_objs) or ids can be passed to this
    # method. See documentation in start() to understand io_records/ids
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    $regex_ids = &ParamsMgr::get_arg(-arg => "ids", -optional => 1,
                                     -default => ".*");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "ClientIO::stop called to stop io_objs with args " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # _get_ids returns an array of ids associated with io_records/regex.
    # io_records takes precedence over ids
    $ids_ref = $self->_get_ids(-io_records => $io_records, -ids => $regex_ids);
    if ($self->error()){
        $self->_set_err("Failed to _get_ids by matching passed regex $regex_ids",
                        $self->errmsg());
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "Matched ids are: " . $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    unless ($ids_ref) {
         $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
         return $IOManUtil::TRUE;
    }

    # send a message of type is_running to manage() and wait for ack
    $stop_args{ids_ref} = $ids_ref;
    $self->{manage}->message_put( message => { %stop_args },
                                  type     => "stop");
    $message = $self->{manage}->message_get(type    => "stop");

    $self->sanlog(-msg => "status received is " . $self->dump_output($message),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    # If status is empty something is wrong
    unless (%$message) {
        $self->_set_err("stop returned nothing");
        return $IOManUtil::FALSE;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $message->{status};
}

###############################################################################
# Method:    get_io_records
# Objective: Returns matching io recs
# Details:   See POD at the end of this file
###############################################################################
sub get_io_records {
    my ($self, %args) = @_;

    my ($io_record, $key, $device, $devices, $index);
    my (@matched_io_records);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # This method is a open ended, meaning it takes a hash and every io_record
    # maintained within ClientIO that matches this, will be returned
    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    $self->sanlog(-msg => "ClientIO::get_io_records called to find matched " .
                          "io_records with args " . $self->dump_output(\%args),
                          -msg_level => $IOManUtil::DEBUG_HIGH);

    # Validate get_io_records params
    $self->_validate_get_io_records_params(%args);
    if ($self->error()) {
        # Error set, bubble up
        return (wantarray()) ? () : [];
    }

    # Loop through all io_records maintained by ClientIO, if a match is found
    # for all the key => value pairs passed to this method that io_record is
    # pushed to matched_io_records array, which gets returned at the end
    foreach $io_record (@{$self->{io_records}}) {
        # verify if the io_record matches with %args
        if (List::MoreUtils::all
                { $io_record->{$_} eq $args{$_} } (keys %args)) {
            push (@matched_io_records, $io_record);
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);

    return (wantarray()) ? @matched_io_records: \@matched_io_records;
}

###############################################################################
# Method:    get_io_record
# Objective: Returns matching unique io rec. Either $id or Client + io_tool
#            + device needs to be passed
# Details:   See POD at the end of this file
###############################################################################
sub get_io_record {
    my ($self, %args) = @_;

    my ($io_record, $key, $device, $devices, $index);
    my (@matched_io_records);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # This method returns io_record that matches with id or client + io_tool
    # + device that user passes. If more params (as a hash) are passed, such
    # a hash will also be compared while trying to find a match with the
    # io_records stored in ClientIO
    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    $self->sanlog(-msg => "ClientIO::get_io_record called to find matched " .
                          "io_record with args " . $self->dump_output(\%args),
                          -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_validate_get_io_records_params(%args);
    if ($self->error()) {
        # Error set, bubble up
        return {};
    }

    # Required keys are: client, device, io_tool. Either these or id is a must.
    # Rest are optional
    unless ($args{id} or ($args{client} and $args{device} and $args{io_tool})) {
        $self->_set_err("Either id or client + device + io_tool needs to be " .
                         "passed");
        return  {};
    }

    # Loop through all io_records maintained by ClientIO, if a match is found
    # for all the key => value pairs passed to this method in addition to id or
    # client + device + io_tool, push i t to matched_io_records
    foreach $io_record (@{$self->{io_records}}) {
        # verify if the io_record matches with %args
        if (List::MoreUtils::all
                { $io_record->{$_} eq $args{$_} } (keys %args)) {
            push (@matched_io_records, $io_record);
        }
    }

    # Only 1 matched io_record needs to be found. If not, set error and return
    unless (scalar @matched_io_records == 1) {
        $self->_set_err("Only 1 match needs to be found when get_io_record " .
                         "is called");
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return  {};
    }

     $self->sanlog(-msg => "ClientIO::get_io_records Match found" .
                           $self->dump_output($matched_io_records[0]),
                   -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $matched_io_records[0];
}


###############################################################################
# Method:    get_params
# Objective: Return params ref
# Details: See POD at the end of this file
###############################################################################
sub get_params {
    my ($self, %args) = @_;

    my ($id, $message, $io_record, $ids_ref);
    my (@io_records);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->sanlog(-msg => "ClientIO::get_params called with args " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);

    # Either io_record (user handle of io_objs) or id can be passed to this
    # method. See documentation in start() to understand io_records/ids
    $id = &ParamsMgr::get_arg(-arg => "id", -optional => 1, -default => '');
    $io_record = &ParamsMgr::get_arg(-arg => "io_record", -default => {});
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # We need an array for _get_ids to work. Convert $io_record to @io_records
    if (scalar keys %$io_record) {
        push (@io_records, $io_record);
    }

    # _get_ids returns an array of ids associated with io_records/regex.
    # io_records takes precedence over ids
    $ids_ref = $self->_get_ids(-io_records => \@io_records, -ids => $id);

    $self->sanlog(-msg => "Matched id with is: " . $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);
    if ($self->error()){
        $self->_set_err("Failed to _get_ids by matching passed regex/io_record",
                        $self->errmsg());
        return {};
    }

    unless (scalar @$ids_ref == 1) {
         $self->_set_err("No matched ids found" . $self->errmsg(),
                        $self->errmsg());
        return {};
    }

    #_get_ids retunrs an array of ids. At this line ids_ref will only contain
    # one element and that will be the id needed
    $args{id} = $$ids_ref[0];

    # send a message of type wait_finish to manage() and wait for wait_ack
    $self->{manage}->message_put( message  => \%args,
                                  type     => "get_params");
    $message = $self->{manage}->message_get(type    => "get_params_ack");

    $self->sanlog(-msg => "get params complete Ack received " .
                           $self->dump_output($message),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    # If params is empty something is wrong
    unless (%$message) {
        $self->_set_err("get_params returned empty hash");
        return {};
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $message;
}

###############################################################################
# Method:    get_client
# Objective: Return client name
# Details: See POD at the end of this file
###############################################################################
sub get_client {
    my ($self) = @_;

    return $self->{client};
}

###############################################################################
# Method:    _get_tool_path
# Objective: get the tool path based on the parameters passed in
# Details: See POD at the end of this file
###############################################################################
sub _get_tool_path {
    my ($self, %args) = @_;
    my ($tool_path, $host);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();

    if (defined $args{tool_path}) {
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return $args{tool_path};
    };

    $host = &Tharn::host($self->get_client());

    # Retrieve the tool path from IOManUtil::TOOL_PATH for the hosttype
    # detected. To do this, loop through all hosttype of the
    # IOManUtil::TOOL_PATH and if a match of hosttype is found, then loop
    # through all entries of the array and get the tool_path by matching
    # io_tool and tool_version.
    foreach my $hosttype (keys %IOManUtil::TOOL_PATH) {
        if ($host->hosttype() =~ /$hosttype/) {
            foreach my $href ( @{$IOManUtil::TOOL_PATH{$hosttype}} ) {
                if (defined ($href->{$args{io_tool}}->{$args{tool_version}})) {
                    $tool_path = $href->{$args{io_tool}}->{$args{tool_version}};
                    last;
                }
            }
        }
    }

    unless (defined $tool_path){
        # dt versions are of this form: dt.17.58
        # sio versions are of this form: sio_6_29
        my $version_sep = ($args{io_tool} eq "dt") ? '.' : '_';
        if ($host->hosttype() =~ /windows/) {
            $tool_path = "x:\\bin\\$args{io_tool}${version_sep}$args{tool_version}.exe";
        } else {
            $tool_path = "/usr/software/test/bin/$args{io_tool}${version_sep}$args{tool_version}";
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $tool_path;
}

###############################################################################
# Method:    _has_started
# Objective: validate if runs associated with io_records are started
#            maintains
# Details:   See POD at the end of this file
###############################################################################
sub _has_started {
    my ($self, %args) = @_;

    my ($wait_time, $io_records, $results_ref);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    $self->sanlog(-msg => "_has_started called with following params" .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    # wait time controls how long we need to keep verifying whether
    # io_records are running
    $wait_time = &ParamsMgr::get_arg(-arg     => "wait_time",
                                     -default => 4);
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => undef);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Increment the iterator twice since the sleep time is 2 secs
    for (my $i = 0; $i < $wait_time; $i += 2) {
        unless ($self->is_running(-io_records => \@$io_records)) {
            $results_ref = $self->get_result(-io_records => \@$io_records,
                                             -wait_finish => 0);
            foreach my $result (@$results_ref) {
                if ($result->{error_count}) {
                    $self->_set_err("io_obj start failed immediately");
                    return $IOManUtil::FALSE;
                }
            }
        }
        &Tharn::snooze(2);
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $IOManUtil::TRUE;
}
###############################################################################
# Method:    _save_io_records
# Objective: saves io_records passed to internal io_records array that ClientIO
#            obj maintains
# Details:   See POD at the end of this file
###############################################################################
sub _save_io_records {
    my ($self, %args) = @_;

    my ($match_index, $io_records);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    # This method takes a reference to an array of io_records as a param and
    # saves that io_record to its internal io_records array
    # ($self->{io_records}).
    # While saving the io_record to $self->{io_records}, io_record id is matched
    # with ids of all $self->{io_records} and if a match is found, its
    # overwritten otherwise io_record is pushed onto $self->{io_records}
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # The below code compares id of each io_record passed with id of ClientIO
    # io_records. If a match is found, such io_record is overwritter otherwise,
    # the user passed io_record is pushed onto ClientIO io_records
    foreach my $io_record (@$io_records) {
        $match_index = List::MoreUtils::first_index {
                          $io_record->{id} eq $_->{id}
                       } @{$self->{io_records}};

        $self->sanlog(-msg => " _save_io_records: match index is" .
                              $match_index .
                              $self->dump_output(@{$self->{io_records}}),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        if ($match_index != -1) {
            ${$self->{io_records}}[$match_index] = $io_record;
        } else {
            push (@{$self->{io_records}}, $io_record);
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _get_ids
# Objective: Returns matching ids with passed regex
# Details:   See POD at the end of this file
###############################################################################
sub _get_ids {
    my ($self, %args) = @_;

    my ($id, $regex_ids, $io_records, $ids_ref);
    my (@io_records, @match_ids, @wildcards);

    # Regex wildcards * ? { } + are allowed. The following is how
    # wildcards are expanded
    # *   Match 0 or more times
    # +   Match 1 or more times
    # ?   Match 1 or 0 times
    # {n} Match exactly n times
    @wildcards = qw (\* \? \{ \} \+);
    @match_ids = ();

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    &ParamsMgr::process_args(%args);

    # If io_records is defined, use it to retrieve io_obj ids
    $io_records = &ParamsMgr::get_arg(-arg => "io_records", default => []);
    if (scalar @$io_records) {
        foreach my $io_record (@$io_records) {
            if ($io_record->{client} eq $self->{client}) {
                push (@{$ids_ref}, $io_record->{id});
            }
        }

        $self->sanlog(-msg => "ids found is" . $self->dump_output($ids_ref),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        return $ids_ref;
    }

    $regex_ids = &ParamsMgr::get_arg(-arg => "ids", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }
    $self->sanlog(-msg => "Matching ids with " . $self->dump_output($regex_ids),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # A simple regex matching won't be enough here. Consider a case where
    # regex_ids is substring of id and the user doesn't use any wild card
    # Using regex matching alone will return all strings that contain regex id
    # as substring. This case will be handled by using eq if none of the regex
    # wild cards are used
    unless (List::MoreUtils::any {$regex_ids =~ /($_)/} @wildcards) {
        if (List::MoreUtils::any {$_ eq $regex_ids} @{$self->{io_obj_ids}}) {
            push (@match_ids, $regex_ids);
        } else {
            $self->_set_err("No matching ids found ");
        }
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return (\@match_ids);
    }

    # Looping through all io_obj_ids and returning all matching ids
    foreach $id (@{$self->{io_obj_ids}}) {
        $self->sanlog(-msg => "id is " . $id,
                      -msg_level => $IOManUtil::DEBUG_LOW);
        if ($id =~ /($regex_ids)/) {
            push (@match_ids, $id);
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    # We expect atleast one matched id. If not set error and return
    if (scalar @match_ids) {
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return (\@match_ids);
    } else {
        $self->_set_err("No matching ids found ");
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return [];
    }
}

###############################################################################
# Method:    _validate_params
# Objective: Validate params
# Details: See POD at the end of this file
###############################################################################
sub _validate_params{
    my ($self, %args) = @_;

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    my @input_params = (%args);

    eval {
        (%args) = Params::Validate::validate( @input_params,
                      { client        => { type => SCALAR,
                                           required => 1},
                        debug_level   => { type => SCALAR,
                                           optional => 1,
                                           regex => qr /^\d+$/},
                      });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: ". $@);
        return ();
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return %args;
}

###############################################################################
# Method:    _validate_create_io_records_params
# Objective: Validate io_records creation params
# Details: See POD at the end of this file
###############################################################################
sub _validate_create_io_records_params{
    my ($self, %args) = @_;

    my @input_params = (%args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    # validation_options is set to allow extra params not defined in validate
    # param arg. These are expected to be tool specific params
    Params::Validate::validation_options( allow_extra => 1);

    eval {
        (%args) = Params::Validate::validate( @input_params,
                      { devices           => { type     => ARRAYREF,
                                               required => 1,
                                               callbacks => {'at least one device must be passed'
                                                               => sub{ scalar(@{$_[0]}) > 0 }}},
                        io_tool           => { type => SCALAR,
                                               required => 1 },
                        verify            => { optional => 1,
                                               default  => 1,
                                               regex    => qr /(1|0)/ },
                        iotype            => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /(random|sequential)/,
                                               default  => 'random'},
                        runtime           => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /^\d+$/ },
                        pattern           => { type     => SCALAR,
                                               optional => 1,},
                        limit             => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /^\d+[bdkgmtBDKGMT]?$/},
                        passes            => { optional => 1,
                                               default  => 1,
                                               regex    => qr /^\d+$/},
                        bs                => { optional => 1,
                                               type     => SCALAR },
                        dispose           => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /keep|delete|keeponerror/},
                        read              => { optional => 1,
                                               default  => 1,
                                               type     => SCALAR,
                                               regex    => qr /(1|0)/  },
                        write             => { optional => 1,
                                               regex    => qr /(1|0)/ },
                        debug_level       => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /^\d+$/},
                        tool_version      => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /(stable|current|staging|last|^\d[_\.]\d+$)/,
                                               default  => 'stable'},
                        devices_per_tool  => { optional => 1,
                                               default  => 1,
                                               regex    => qr /(all|^\d+$)/},
                        tool_path         => { type     => SCALAR,
                                               optional => 1},
                        testing_method    => { type     => SCALAR,
                                               optional => 1,
                                               regex    => qr /(positive|negative)/,
                                               default  => 'positive'},
                        failure_action    => { type     => SCALAR,
                                               optional => 1,
                                               default  => "" },
                      });
    };
    if ($@) {
        $self->_set_err("Failed to validate parameter: ". $@);
        return {};
    }

    if ($args{write}) {
        $args{read} = 0;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return %args;
}

###############################################################################
# Method:    _validate_get_io_records_params
# Objective: validate get_io_records params.
# Details: See POD at the end of this file
###############################################################################
sub _validate_get_io_records_params{
    my ($self, %args) = @_;

    my (%validate_params_arg);

    my @input_params = (%args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    # validate_params_arg is the argument passed while validating param. This
    # only has ioman_params
    %validate_params_arg = (
                      devices       => { type     => ARRAYREF,
                                         optional => 1 },
                      io_tool       => { type     => SCALAR,
                                         optional => 1,
                                         regex    => qr /(sio|dt)/ },
                      iotype        => { type     => SCALAR,
                                         optional => 1,
                                         regex    => qr /(random|sequential)/ },
                      runtime       => { type     => SCALAR,
                                         optional => 1,
                                         regex    => qr /^\d+$/ },
                      pattern       => { type     => SCALAR,
                                         optional => 1 },
                      limit         => { type     => SCALAR,
                                         optional => 1,
                                         regex    => qr /^\d+[kgmtKGMT]?$/ },
                      bs            => { optional => 1,
                                         type     => SCALAR },
                      dispose       => { optional => 1,
                                         type     => SCALAR,
                                         regex    => qr /(keep|delete)/ },
                      verify        => { optional => 1,
                                         type     => SCALAR,
                                         regex    => qr /(1|0)/ },
                      read          => { optional => 1,
                                         type     => SCALAR,
                                         regex    => qr /(1|0)/  },

                    );
    eval {
        (%args) = Params::Validate::validate( @input_params,
                                              \%validate_params_arg);
    };
    if ($@) {
        $self->_set_err("Failed to validate parameter: ". $@);
        return ();
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return %args;
}

1;
