package IOMan::ClientIO::IO;

# Compiler directives
use strict;

# imported modules
use Data::Dumper;
use Params::Validate qw( :all );
use IOMan::ClientIO::IO::Sio;
use IOMan::ClientIO::IO::Dt;
use IOMan::IOManUtil;
use Tharn;

our @ISA = qw( IOMan::IOManUtil );

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Creates new IO object and acts as  base class for <io_tool>.pm
# Details:   See POD at the end of this file
###############################################################################
sub new {
    my ($class, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # If $class is not package name, then the call came from derived class
    # (IO.pm, sio.pm) In this case we store all common io params to
    # $self->{ioman_params} bless $self to correct type and return control to
    # derived class
    if ($class ne __PACKAGE__) {

        # Store all common IOObjs params here, bless and return
        my $self = {};
        my ($debug_level, $key);

        bless $self, $class;

        &ParamsMgr::process_args(%args);
        $debug_level = &ParamsMgr::get_arg(-arg => "debug_level",
                                           -required => 0,
                                           -default => $IOManUtil::VERBOSE_LOW);
        if (&ParamsMgr::error()) {
            $self->_set_err(&ParamsMgr::get_error());
            return $self;
        }

        $self->set_debug_level(-debug_level => $debug_level);
        if ($self->error()) {
            $self->_set_err("Cannot set debug level", $self->errmsg());
            return $self;
        }

        $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
        $self->sanlog(-msg => "Creating IO Object" . $self->dump_output(\%args),
                      -msg_level => $IOManUtil::DEBUG_LOW);

        $self->sanlog(-msg => "Validating generic io obj params",
                      -msg_level => $IOManUtil::DEBUG_HIGH);

        # Input args passed (%args) will be organized into
        # %{$self->{ioman_params}} as follows:
        # Assume $ioman_params = $self->{ioman_params}
        # 1) All generic interchangeable params across all tools are stored
        #    in $ioman_params->{$param} where $param is a generic param
        # 2) All tool specific params will be stored in
        #    $ioman_params->{tool_params}. For instance, read_perc which is
        #    sio specific param will be stored in
        #    $ioman_params->{tool_params}->{read_perc}
        #
        # The below generate ioman_params does everthing stated above and also
        # apart from validating $ioman_params. While validating we let all the
        # params not identified as ioman_params to just pass through and store
        # them in $ioman_params->{tool_params}. It will be the responsibility of
        # <io_tool.pm> to validate these params
        $self->_generate_ioman_params(%args);
        if ($self->error()) {
            # Error set, bubble up
            return $self;
        }

        # Creating a new conn obj using ConnectionMgr
        if (defined $self->{conn_type}) {
            $self->{conn_mgr}->make_connect_for_host(-client => $self->{client},
                                                     -conntype => $self->{conn_type});
        } else {
            $self->{conn_mgr}->make_connect_for_host(-client => $self->{client});
        }
        $self->{conn_obj} = $self->{conn_mgr}->get_connection(
                                                    -filer => $self->{client});

        $self->sanlog( -msg => "Created conn obj" . $self->{conn_obj},
                       -msg_level => $IOManUtil::DEBUG_HIGH);

        # TO DO: create result obj.
        #$self->{result_obj} = Result->new();

        return $self;
    }

    # If $class is IO package then, based on the tool type we make call to
    # appropriate <io_tool>.pm

    my ($io_tool);
    my $self = {};
    $io_tool = $args{io_tool};

    if ($io_tool eq 'sio') {
        return IOMan::ClientIO::IO::Sio->new(%args);
    }  elsif ($io_tool eq 'dt') {
        return IOMan::ClientIO::IO::Dt->new(%args);
    }  else {
        # If io_tool is not supported, bless it as generic IO, set error and
        # return
        bless $self, $class;
        $self->_set_err("IO tool passed is not supported");
        return $self;
    }
}


###############################################################################
# Method:    get_id
# Objective: Return id attr
# Details: See POD at the end of this file
###############################################################################
sub get_id {
    my ($self) = @_;

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return $self->{id};
}

###############################################################################
# Method:    get_params
# Objective: Return generic params. If -all => true then we return tool params
#            also
# Details:   See POD at the end of this file
###############################################################################
sub get_params{
    my ($self, %args) = @_;

    my ($key);
    my (%params);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    %args = &SanUtil::strip_dash(%args);

    %params = %{$self->{ioman_params}};
    unless ($args{all}) {
        delete $params{tool_params};
        return \%params;
    }


    foreach $key (keys %{$self->{ioman_params}->{tool_params}}) {
        %params = (%params,
                   $key => $self->{ioman_params}->{tool_params}->{$key}
                   );
    }
    delete $params{tool_params};

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return \%params;
}

###############################################################################
# Method:    get_result
# Objective: returns result
# Details:   See POD at the end of this file
###############################################################################
sub get_result {
    my ($self) = @_;

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->sanlog(-msg => "Result is " . $self->dump_output($self->{result}),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return $self->{result};
}

###############################################################################
# Method:    print_result
# Objective: returns result as a scalar in a nice format
# Details:   See POD at the end of this file
###############################################################################
sub print_result {
    my ($self) = @_;

    my ($key, $msg);
    my (@generic_result_keys);

    @generic_result_keys = qw (dev avg_transfer_rate io_per_sec total_elapsed_time
                               start_time end_time error_count);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);


    $msg = "IO run statistics: \t" . "Tool Used: " . $self->{io_tool} .
           ", Device: " . $self->{device};
    $msg = $self->_print_header($msg);
    foreach $key (@generic_result_keys) {
        $msg = $msg . $key . " is " . $self->{result}->{$key} . "\n";
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return ($msg);
}

###############################################################################
# Method:    get_timeout
# Objective: Determine timeout to be used while running io_tool cmd
# Details: See POD at the end of this file
###############################################################################
sub get_timeout{
    my ($self) = @_;

    my ($size_in_mb);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # If runtime param is passed, we just add 1800 to it
    if ($self->{ioman_params}->{runtime}) {
        $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
        return ($self->{ioman_params}->{runtime} + $IOManUtil::RUN_TIMEOUT);
    }

    # We assume here that we can transfer 1Mb/s and determine timeout if runtime
    # is not passed
    if ($self->{ioman_params}->{limit}) {
        $size_in_mb = &ClientUtil::convert_to_bytes(
                          $self->{ioman_params}->{limit}, "m"
                       );
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return ($size_in_mb + $IOManUtil::RUN_TIMEOUT);
}

###############################################################################
# Method:    _generate_ioman_params
# Objective: validate and generate ioman params. Non-identified params will be
#            stored in tool_params hash for further processing by <io_tool.pm>
# Details: See POD at the end of this file
###############################################################################
sub _generate_ioman_params{
    my ($self, %args) = @_;

    my ($key);
    my (%validate_params_arg);

    my @input_params = (%args);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # validation_options is set to allow extra params not defined in validate
    # param arg. These are expected to be tool specific params
    Params::Validate::validation_options( allow_extra => 1);

    # validate_params_arg is the argument passed while validating param. This
    # only has ioman_params
    %validate_params_arg = (
                      iotype         => { type     => SCALAR,
                                          optional => 1,
                                          regex    => qr /(random|sequential)/,
                                          default  => 'random' },
                      io_tool        => { type     => SCALAR,
                                          optional => 1,
                                          regex    => qr /(sio|dt)/,
                                          default  => 'random' },
                      runtime        => { type     => SCALAR,
                                          optional => 1,
                                          regex    => qr /^\d+$/ },
                      pattern        => { type     => SCALAR,
                                          optional => 1 },
                      limit          => { type     => SCALAR,
                                          optional => 1,
                                          regex    => qr /^\d+[bdkgmtBDKGMT]?$/ },
                      bs             => { optional => 1,
                                          type     => SCALAR },
                      position       => { optional => 1,
                                          type     => SCALAR,
                                          regex    => qr /^\d+[bdkgmtBDKGMT]?$/ },
                      dispose        => { optional => 1,
                                          default  => "keep",
                                          type     => SCALAR,
                                          regex    => qr /(keep|delete|keeponerror)/ },
                      verify         => { optional => 1,
                                          default  => 1,
                                          type     => SCALAR,
                                          regex    => qr /(1|0)/ },
                      read           => { optional => 1,
                                          default  => 1,
                                          type     => SCALAR,
                                          regex    => qr /(1|0)/  },
                      id             => { type     => SCALAR,
                                          required => 1 },
                      client         => { type     => SCALAR,
                                          required => 1 },
                      device         => { type     => SCALAR,
                                          required => 1 },
                      debug_level    => { type     => SCALAR,
                                          optional => 1,
                                          regex    => qr /^\d+$/},
                      conn_mgr       => { isa      => 'ConnectionMgr',
                                          required => 1 },
                      conn_type      => { type     => SCALAR,
                                          optional => 1 },
                      ipc_timeout    => { type     => SCALAR,
                                          optional => 1,
                                          default  => 600 },
                      tool_path      => { type     => SCALAR,
                                          required => 1 },
                      failure_action => { type     => SCALAR,
                                          optional => 1,
                                          default  => "" },
                      failure_policy => { type     => SCALAR,
                                          optional => 1,
                                          default  => "ignore",
                                          regex    => qr /(ignore|stop_all|stop_all_client)/ },
                      testing_method => { type     => SCALAR,
                                          optional => 1,
                                          regex    => qr /(positive|negative)/,
                                          default  => 'positive'},
                    );
    eval {
        (%args) = Params::Validate::validate( @input_params,
                                              \%validate_params_arg);
    };
    if ($@) {
        $self->sanlog(-msg => "Failed to validate parameters " . $@,
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        $self->_set_err("Failed to validate parameter: ". $@);
        return $IOManUtil::FALSE;
    }

    # Now that the validation is done, generate $self->{ioman_params}
    # Loop through all params validate args and populate $self->{ioman_params}
    # These are the generic ioman_params and they will be stored in ioman_params
    foreach $key (keys %validate_params_arg) {
        if (defined $args{$key}) {
            $self->{ioman_params}->{$key} = $args{$key};
        }
    }

    # Rest of the params that doesn't go into $self->{ioman_params} are tool
    # specific params. These will be stored in
    # $self->{ioman_params}->{tool_params}
    foreach $key (keys %args) {
        unless (exists $self->{ioman_params}->{$key}) {
            $self->{ioman_params}->{tool_params}->{$key} = $args{$key};
        }
    }

    # In the process of validating all %args passed, we stored some IO.pm params
    # to $self->{ioman_params}. Below piece of code removes all of these params
    # copies them directly under $self and delete them from
    # $self->{ioman_params}. While doing so, we not leave $self->{ioman_params}
    # with only ioman_params.
    foreach $key ('id', 'client', 'device', 'conn_mgr', 'conn_type',
                   'debug_level', 'io_tool') {

            $self->{$key} = delete $self->{ioman_params}->{$key};
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE
}

###############################################################################
# Method:     execute_failure_action - Execute the failure action.
# Objective:  Prepare parameters to invoke an external script.
# Parameters: $cmd = The failing command.
#             $error = The associated error messsage.
# Returns:   The script results on failure or nothing on success.
###############################################################################
sub execute_failure_action {
    my ($self, $cmd, $error) = @_;
    my $client="localhost"; # This is NATE control server (at present)!
    my @script_args;
    my %args;

    $args{'args'} = \@script_args;
    $args{'runid'} = "failure_action";
    $args{'codespec'} = "$self->{failure_action}";

    # Information we're passing to the failure script.
    push(@script_args, "$self->{client}");
    push(@script_args, "$self->{device}");
    push(@script_args, "'$error'");
    push(@script_args, "'$cmd'");
    # BURT 570079 prevents params as environment variables!
    $ENV{'FAILURE_CMD'} = "$cmd";
    $ENV{'FAILURE_DEVICE'} = "$self->{device}";
    $ENV{'FAILURE_CLIENT'} = "$self->{client}";
    $ENV{'FAILURE_ERROR'} = "$error";

    my $output = $self->execute_script(%args);
    $self->sanlog(-msg => "execute_script returned: " . $output,
                  -msg_level => $IOManUtil::VERBOSE_LOW);
    return $output;
}

1;
