package IOMan::ClientIO::IO::Dt;

# Compiler directives
use strict;

use Data::Dumper;
use Params::Validate qw( :all );
use List::MoreUtils qw(any indexes first_index);

our @ISA = qw(IOMan::ClientIO::IO);
###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create Dt object
# Details:   See POD at the end of this file
###############################################################################
sub new {
    my ($class, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Call base class IO.pm. It does basic param validation and sets all ioman
    # params in $self->{ioman_params}
    my $self = $class->SUPER::new(%args);
    if ($self->error()) {
        # Error set by IO.pm, just bubble up
        return $self;
    }

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # run_params contains all params required to run dt. The params here will
    # no longer be generic IOMan params but params specific to dt

    $self->{run_params} = {};

    $self->sanlog(-msg => "Translating generic IOMan params to dt specific " .
                          "params and then validating:\n\t" .
                           $self->dump_output($self),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # _translate_params() translates ioman params to dt specific params and
    # stores it in $self->{run_params}. _validate_params() validates
    # $self->{run_params}. Effectively, it validates  params required for dt
    # command invocation
    $self->_translate_params();
    $self->_validate_params();
    if ($self->error()) {
        # Error set by translate params, just bubble up
        return $self;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $self;
}

###############################################################################
# Method:    _translate_params
# Objective: Translate generic IOMan params to dt specific params
# Details: See POD at the end of this file
###############################################################################
sub _translate_params{
    my ($self) = @_;

    my ($ioman_params, $run_params, $key);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $ioman_params = $self->{ioman_params};
    $self->sanlog(-msg => "Translating params:\n\t" . Dumper($ioman_params),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $run_params = $self->{run_params};

    # The syntax of IO.pm's ioman_params is different from params required for
    # dt command invocation. $run_params translates ioman_params to params
    # required for running dt command. Also,
    # $self->{ioman_params}->{tool_params} are appropriately converted to
    # run_params. Generating $run_params is also necessary to validate params
    # required for dt command invocation (done in new())
    $self->{translate} = 1;
    $self->{failure_action} = $ioman_params->{failure_action};
    $self->{failure_policy} = $ioman_params->{failure_policy};
    $self->{testing_method} = $ioman_params->{testing_method};
    if ($ioman_params->{read}) {
        $run_params->{if} = $self->{device};
    } else {
        $run_params->{of} = $self->{device};
    }

    if (defined $ioman_params->{iotype}) {
        $run_params->{iotype} = $ioman_params->{iotype};
    }
    if (defined $ioman_params->{runtime}) {
        $run_params->{runtime} = $ioman_params->{runtime};
    }
    if (defined $ioman_params->{pattern}) {
        $run_params->{pattern} = $ioman_params->{pattern};
    }
    if (defined $ioman_params->{bs}) {
        $run_params->{bs} = $ioman_params->{bs};
    }
    if (defined $ioman_params->{limit}) {
        $run_params->{limit} = $ioman_params->{limit};
    }
    if (defined $ioman_params->{dispose}) {
        $run_params->{dispose} = $ioman_params->{dispose};
    }
    if (defined $ioman_params->{position}) {
        $run_params->{position} = $ioman_params->{position};
    }
    if (defined $ioman_params->{tool_path}) {
        $run_params->{tool_path} = $ioman_params->{tool_path};
    }

    # If there is a conflict between tool params and generic params, tool params
    # takes precedence.
    # Allowed dt params: passes, procs
    # There is very minimal checking for pass_thru. We will just let it run and
    # and fail
    # Below code snippet takes all tool_params of $io_params and stores them
    # directly under $run_params
    if (exists $ioman_params->{tool_params}) {
        foreach $key (keys %{$ioman_params->{tool_params}}) {
            if (scalar $ioman_params->{tool_params}->{$key}) {
                $run_params->{$key} = $ioman_params->{tool_params}->{$key};
            }
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE
}

###############################################################################
# Method:    validate_params
# Objective: Validate dt params
# Details: See POD at the end of this file
###############################################################################
sub _validate_params{
    my ($self) = @_;

    my ($run_params);
    my (@input_params);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $run_params = $self->{run_params};
    $self->sanlog(-msg => "Validating IOMan::new() params:\n\t" .
                          Dumper($run_params),
                  -msg_level => $IOManUtil::DEBUG_HIGH);
    @input_params = ($self->{run_params});

    eval {
        (%$run_params) = Params::Validate::validate( @input_params,
                         {
                             aios          =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+$/ },
                             runtime       =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+$/ },
                             pattern       =>  { type     => SCALAR,
                                                 optional => 1 },
                             iodir         =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /(forward|reverse)/ },
                             iotype        =>  { type     => SCALAR,
                                                 default  => 0,
                                                 regex    => qr /(random|sequential)/ },
                             limit         =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+[bdkgmtBDKGMT]?$/ },
                             passes        =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+$/ },
                             bs            =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+[bdkgmBDKGM]?$|random/ },
                             position      =>  { optional => 1,
                                                 type     => SCALAR,
                                                 regex    => qr /^\d+[bdkgmtBDKGMT]?$/ },
                             dispose       =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /delete|keep|keeponerror/ },
                             if            =>  { type     => SCALAR,
                                                 optional => 1 },
                             of            =>  { type     => SCALAR,
                                                 optional => 1 },
                             pf            =>  { type     => SCALAR,
                                                 optional => 1 },
                             passes        =>  { optional => 1,
                                                 default  => 1,
                                                 regex    => qr /^\d+$/},
                             procs         =>  { optional => 1,
                                                 regex    => qr /^\d+$/},
                             oncerr        =>  { optional => 1,
                                                 regex    => qr /abort|continue/},
                             disable       =>  { type     => SCALAR,
                                                 optional => 1 },
                             enable        =>  { type     => SCALAR,
                                                 optional => 1 },
                             oflags        =>  { type     => SCALAR,
                                                 optional => 1 },
                             errors        =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+$/},
                             noprogt       =>  { type     => SCALAR,
                                                 optional => 1 },
                             noprogtt      =>  { type     => SCALAR,
                                                 optional => 1 },
                             keepalive     =>  { type     => SCALAR,
                                                 optional => 1 },
                             alarm         =>  { type     => SCALAR,
                                                 optional => 1 },
                             trigger       =>  { type     => SCALAR,
                                                 optional => 1 },
                             incr          =>  { type     => SCALAR,
                                                 optional => 1 },
                             min           =>  { type     => SCALAR,
                                                 optional => 1 },
                             max           =>  { type     => SCALAR,
                                                 optional => 1 },
                             slice         =>  { type     => SCALAR,
                                                 optional => 1 },
                             slices        =>  { type     => SCALAR,
                                                 optional => 1,
                                                 regex    => qr /^\d+$/},
                             step          =>  { type     => SCALAR,
                                                 optional => 1 },
                             pass_thru     =>  { type      => SCALAR,
                                                 optional  => 1},
                             tool_path     =>  { type      => SCALAR,
                                                 required  => 1,
                                                 callbacks => {'tool path not accessible'
                                                                => sub { $self->tool_path_exists(tool_path => $_[0]) } }},
                         });
    };
    if ($@) {
        $self->sanlog(-msg => "Failed to validate params for dt run " . $@,
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        $self->_set_err("Failed to validate parameters for dt run: ". $@);
        return $IOManUtil::FALSE;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    start
# Objective: start
# Details: See POD at the end of this file
###############################################################################
sub start {
    my ($self, %args) = @_;

    my ($run_params, $cmd, $key, $status, $result, $timeout);
    my (@output);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    %args = &SanUtil::strip_dash(%args);
    $run_params = $self->{run_params};

    $cmd  = $run_params->{tool_path} . " ";
    delete $run_params->{tool_path};

    foreach $key (keys (%{$run_params})) {
        if (scalar($run_params->{$key}) and $key ne "pass_thru") {
            $cmd = $cmd . $key . "=" . $run_params->{$key} . " ";
        }
    }
    if (scalar($run_params->{pass_thru})) {
        $cmd = $cmd . $run_params->{pass_thru};
    }

    $self->sanlog(-msg => "Running dt cmd" . $cmd .
                          "with params \n\t" . $self->dump_output($run_params),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->sanlog( -msg => "Using conn obj" . $self->{conn_obj},
                   -msg_level => $IOManUtil::DEBUG_HIGH);

    $timeout = $self->get_timeout();

    ($status, @output) = &ClientUtil::execute($self->{conn_obj}, $cmd, $timeout);
    $self->sanlog(-msg => "dt run complete, result is " . @output,
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # $self->{result} parses the output and stores as follows:
    # 1) raw_output is stored directly under it
    # 2) All generic result params (those that are fairly common across
    #    different io_tools will be stored directly under $self->{result}
    #    So far, these are identified to be the common params:
    #    dev, avg_transfer_rate, io_per_sec, total_elapsed_time, start_time,
    #    end_time and error_count. If the output does not return any of these
    #    common params, set them to "NA"
    # 3) All remaining tool specific result output is stored in
    #    $self->{result}->{tool_specific}
    # The idea of the result hash is to store the params in a way that is
    # consistent across all tools so that Result.pm can expose interfaces that
    # the upper layers (ClientIO.pm and IOMan.pm) can benefit from
    #
    $self->{result}->{raw_output} = \@output;
    $self->{result}->{io_tool} = $self->{io_tool};
    $self->{result}->{client} = $self->{client};
    $self->{result}->{id} = $self->{id};
    $self->{result}->{error_count} = 0;

    #
    # Note: We now support negative and positive testing. For positive testing,
    # the default, we log a failure when an error occurs, as we've always done.
    # But for negative testing, if success is returned, we'll log a failure.
    #

    # If dt command invocation returned an error, set error and return
    if ($status) {
        my $regexp = '(dt.*ERROR:.*)';
        my ($error) = grep(/$regexp/, @output);
        if ( !defined($error) ) {
            $error = "dt run failed";
        }
        if ($self->{testing_method} eq "positive") {
            $self->{result}->{error_count} = 1;
            $self->{result}->{error} = "dt run failed with output " . @output;
            $self->sanlog(-msg => "Failure: Result is " .
                                  $self->dump_output($self->{result}),
                          -msg_level => $IOManUtil::VERBOSE_LOW);
            $self->execute_failure_action($cmd, $error) if (length $self->{failure_action});
            if ($args{subtested}) {
                &Tharn::addresult(type => "FAIL", "run failed");
                $Tharn::Parent->message_put(type    => "result",
                                            message => $self->{result});
                $Tharn::Parent->message_put(
                    type    => "failure_policy",
                    message => { failure_policy => $self->{failure_policy} }
                    );
            }
            $self->_set_err($self->{result}->{error});
            return $IOManUtil::FALSE;
        } else { # Testing method is "negative"!
            # Expected failure occurred, so return success!
            $self->sanlog(-msg => "Failure occured as expected, result is " .
                                  $self->dump_output($self->{result}),
                          -msg_level => $IOManUtil::VERBOSE_LOW);
            $Tharn::Parent->message_put(type    => "result",
                                        message => $self->{result});
            return $IOManUtil::TRUE;
        }
    } elsif ($self->{testing_method} eq "negative") {
        my $error = "dt run succeeded but failure expected";
        # Expected failure, but success was returned!
        $self->{result}->{error_count} = 1;
        $self->{result}->{error} = $error;
        $self->sanlog(-msg => "Unexpected success, result is " .
                              $self->dump_output($self->{result}),
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        # Note: We must execute before sending parent result, or parent will exit!
        $self->execute_failure_action($cmd, $error) if (length $self->{failure_action});
        if ($args{subtested}) {
            &Tharn::addresult(type => "FAIL", "run succeeded, but failure expected");
            $Tharn::Parent->message_put(type    => "result",
                                        message => $self->{result});
            $Tharn::Parent->message_put(
                type    => "failure_policy",
                message => { failure_policy => $self->{failure_policy} }
                );
        }
        $self->_set_err($self->{result}->{error});
        return $IOManUtil::FALSE;
    }

    # parse the output and build dt specific result hash. If errors are found
    # just return failure
    $result = $self->_parse_output();

    # copy result to $self->result
    foreach my $key (keys %$result) {
        $self->{result}->{$key} = $result->{$key};
    }

    if ($self->error() or $result->{error_count}) {
        # log failure and return
        $self->{result}->{error} = "Failed to build result " . $self->errmsg();
        $self->sanlog(-msg => "Failure: Result returned is " .
                              $self->dump_output($result),
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        if ($args{subtested}) {
            $Tharn::Parent->message_put(type    => "result",
                                        message => $self->{result});
            $Tharn::Parent->message_put(
                type    => "failure_policy",
                message => { failure_policy => $self->{failure_policy} }
                );
        }
        return $IOManUtil::FALSE;
    }

    # The result hash that was generated using _parse_output() has one flat list
    # of key value pairs. Using translate_result() the result hash is converted
    # to a result hash that is common across different tools
    # Generating result hash is reverse of how common io_params are converted
    # to run_params to invoke the sio command
    # The dt output has dt specific result and it is coverted to generic
    # result hash by calling _translate_result
    $self->_translate_result($result);

    if ($args{subtested}) {
        $Tharn::Parent->message_put(type    => "result",
                                    message => $self->{result});
    }

    $self->sanlog(-msg => "Result returned is " . $self->dump_output($result),
                  -msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _parse_output
# Objective: parse dt output and return result hash ref
# Details:   See POD at the end of this file
###############################################################################
sub _parse_output {
    my ($self) = @_;

    my ($line, $result, $index, $i);
    my (@output, @output_for_proc, @indexes, $start_index, $end_index);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    @output = @{$self->{result}->{raw_output}};
    $result = {};

    # total statistics of dt is used to generated result hash. The output presents
    # total statistics as follows:
    #
    # Total Statistics: Output device/file name: /dev/rfd0c (Device: floppy, type=disk)
    # Type of I/O’s performed: random (seed 0xa775f81)
    # Data pattern read/written: 0x00ff00ff
    # Total records processed: 109 with min=512, max=65536, incr=3584
    # Total bytes transferred: 3011072 (2940.500 Kbytes, 2.872 Mbytes)
    # Average transfer rates: 16642 bytes/sec, 16.252 Kbytes/sec Number I/O’s per second: 0.602
    # Total passes completed: 1 Total errors detected: 0/1
    # Total elapsed time: 03m00.93s Total system time: 00m00.01s Total user time: 00m00.16s
    # Starting time: Wed Jan 12 16:38:38 2000 Ending time: Wed Jan 12 16:41:43 2000
    #
    # If procs params is passed to dt, then the total statistics output looks
    # like following:
    #
    # Total Statistics (29090):
    # Output device/file name: /dev/rrz2c (Device: BB01811C, type=disk)
    # Type of I/O’s performed: sequential
    # Current Process Reported: 1/2
    # Data pattern string used: ’IOT Pattern’
    # Total records processed: 32768 @ 65536 bytes/record (64.000 Kbytes)
    # Total bytes transferred: 2147483648 (2097152.000 Kbytes, 2048.000 Mbytes)
    # Average transfer rates: 3615597 bytes/sec, 3530.856 Kbytes/sec
    # Number I/O’s per second: 55.170
    # Total passes completed: 1/1 Total errors detected: 0/1
    # Total elapsed time: 09m53.95s
    # Total system time: 00m06.63s
    # Total user time: 02m32.51s
    # Starting time: Thu Nov 9
    # Ending time: Thu Nov 9
    #
    # Total Statistics (29105):
    # Output device/file name: /dev/rrz2c
    # Type of I/O’s performed: sequential
    # Current Process Reported: 2/2
    # Data pattern string used: ’IOT Pattern’
    # Total records processed: 32768 @ 65536 bytes/record (64.000 Kbytes)
    # Total bytes transferred: 2147483648 (2097152.000 Kbytes, 2048.000 Mbytes)
    # Average transfer rates: 3604370 bytes/sec, 3519.893 Kbytes/sec
    # Number I/O’s per second: 54.998
    # Total passes completed: 1/1
    # Total errors detected: 0/1
    # Total elapsed time: 09m55.80s
    # Total system time: 00m05.90s
    # Total user time: 02m38.21s
    # Starting time: Thu Nov 9 09:50:28 2000
    # Ending time: Thu Nov 9 10:00:24 2000
    #
    # Determine whether there are multiple processes involved. The parsing and
    # the way the result hash is generated is different if multiple processes
    # are involved

    $index = List::MoreUtils::first_index {$_ =~ /Current Process Reported: \d+\/\d+/} @output;
    if ($index) {
        # Multiple processes are run. In this case total statistics will be
        # reported for each process once. The number of processes is parsed
        # here
        $output[$index] =~ /Current Process Reported: \d+\/(\d+)/;
        $result->{procs} = $1;
    }

    unless ($result->{procs}) {
        # If only one procs is involved. Get the line index of where Total
        # statistics is found. Splice output till that point
        $index = List::MoreUtils::first_index {$_ =~ /Total Statistics/} @output;
        splice(@output, 0, $index+1);
        # _parse_total_statistics takes The lines under Total Statistics and
        # stores the values in a hash. If something unexpected happens
        # it sets error
        $result = $self->_parse_total_statistics(-output => \@output);
        if ($self->error()) {
            # return failure
            $self->_set_err("Failed to build result " . $self->errmsg());
            return {};
        }
        return $result;
    }

    # Multiple processes are involved. $result->{procs} is already set to
    # number of processes. The number of times /Total statistics/ found in the
    # output should match to the number of procs. Otherwise set error and return
    $i = 0;
    @indexes = List::MoreUtils::indexes {$_ =~ /Total Statistics/} @output;
    unless (@indexes eq $result->{procs}) {
        $result->{error_count} = 1;
        $self->_set_err("The output doesn't have a summarized result for " .
                        "every proc ");
        return {};
    }

    # Look for the first index where /Total Statistics/ is found.
    # Store the output till Ending time and pass it to _parse_total_statistics
    # to build the result hash
    # The result hash for each process is stored under $result as a hash ref.
    # Repeat this for each process
    # At the end of it, if no errors are found the result will look like
    # following:
    # $result->{procs} = 2;
    # $result->{0}->{result_hash_for_proc_1}
    # $result->{1}->{result_hash_for_proc_2}
    while ($i < $result->{procs}) {
        # Get the index of where total statistics is found
        $start_index = List::MoreUtils::first_index {$_ =~ /Total Statistics/} @output;

        # Splice the output to that point. We don't need that
        splice(@output, 0, $start_index+1);

        # Get the index of Ending time
        $end_index = List::MoreUtils::first_index {$_ =~ /Ending time: .*/} @output;

        # Splice the output from Total Statistics - Ending time and pass it to
        # _parse_total_statistics
        @output_for_proc = splice(@output, 0, $end_index+1);
        $result->{$i} = $self->_parse_total_statistics(-output => \@output_for_proc);
        if ($self->error()) {
            # return failure
            $self->_set_err("Failed to build result " .
                            $self->dump_output($result));
            return {};
        }

        # set error_count if an error is found for this proc
        $result->{error_count} = 1 if $result->{$i}->{error_count};
        $i++;
    }

    return $result;
}

###############################################################################
# Method:    _parse_total_statistics
# Objective: parses total statistics in output return hash ref
# Details:   See POD at the end of this file
###############################################################################
sub _parse_total_statistics {
    my ($self, %args) = @_;

    my ($output_ref, $line, $result);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);
    $output_ref = &ParamsMgr::get_arg(-arg => "output", -required => 0,
                                       -default => $IOManUtil::VERBOSE_LOW);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Parse each line and generate result hash ref
    foreach $line (@$output_ref) {
        if ($line =~ /.*device\/file name: (.*)/) {
            $result->{dev} = $1;
        } elsif ($line =~ /Type of I\/O\'s performed: (.*)/) {
            $result->{type} = $1;
        } elsif ($line =~ /Data pattern.*: (.*)/) {
            $result->{pattern} = $1;
        } elsif ($line =~ /Total records processed: (.*)/) {
            $result->{total_records} = $1;
        } elsif ($line =~ /Total bytes transferred: (.*)/) {
            $result->{total_bytes} = $1;
        } elsif ($line =~ /Average transfer rates: (.*)/) {
            $result->{avg_transfer_rate} = $1;
        } elsif ($line =~ /Number I\/O\'s per second: (.*)/) {
            $result->{io_per_sec} = $1;
        } elsif ($line =~ /Total passes completed: (.*)/) {
            $result->{total_passes} = $1;
        } elsif ($line =~ /Total errors detected: (\d+)\/(\d+)/) {
            $result->{error_count} = $1;
        } elsif ($line =~ /Total elapsed time: (.*)/) {
            $result->{total_elapsed_time} = $1;
        } elsif ($line =~ /Total system time: (.*)/) {
            $result->{total_system_time} = $1;
        } elsif ($line =~ /Total user time: (.*)/) {
            $result->{total_user_time} = $1;
        } elsif ($line =~ /Starting time: (.*)/) {
            $result->{start_time} = $1;
        } elsif ($line =~ /Ending time: (.*)/) {
            $result->{end_time} = $1;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $result;
}

###############################################################################
# Method:    _translate_result
# Objective: Translate result hash to generic result hash
# Details:   See POD at the end of this file
###############################################################################
sub _translate_result {
    my ($self, $result) = @_;

    my ($key, $i);
    my (@generic_result_keys);

    @generic_result_keys = qw (dev avg_transfer_rate io_per_sec total_elapsed_time
                               start_time end_time error_count);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # If number of procs is set it is translated differently because $result is
    # of format $result->{0}->{result_hash_ref_for_proc1}
    #
    # If no of procs is not set, keep all generic result keys directly under
    # $result and move rest of keys to $result->{tool_specific}
    # If multiple procs are involded, do the same for each result_hash_ref
    unless ($result->{procs}) {
        foreach $key (keys %{$result}) {
            if (List::MoreUtils::any {$_ eq $key} @generic_result_keys) {
                $self->{result}->{$key} = delete $result->{$key};
            }
        }
        $self->{result}->{tool_specific} = $result;
    } else {
        $i = 0;
        $self->{result}->{procs} = $result->{procs};
        while ($i < $result->{procs}) {
            foreach $key (keys %{$result->{$i}}) {
                if (List::MoreUtils::any {$_ eq $key} @generic_result_keys) {
                    $self->{result}->{$i}->{$key} = delete $result->{$i}->{$key};
                }
            }
            $self->{result}->{$i}->{tool_specific} = $result->{$i};
            $i++;
        }
    }

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

1;
