package IOMan::ClientIO::IO::Sio;

# Compiler directives
use strict;

use Data::Dumper;
use Params::Validate qw( :all );

our @ISA = qw(IOMan::ClientIO::IO);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create Sio object
# Details:   See POD at the end of this file
###############################################################################
sub new {
    my ($class, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Call base class IO.pm. It does basic param validation and sets all ioman
    # params in $self->{ioman_params}
    my $self = $class->SUPER::new(%args);
    if ($self->error()) {
        # Error set by IO.pm, just bubble up
        return $self;
    }

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # run_params contains all params required to run sio. The params here will
    # no longer be generic IOMan params but params specific to sio
    $self->{run_params} = {};

    $self->sanlog(-msg => "Translating generic IOMan params to sio specific " .
                          "params and then validating: " .
                           $self->dump_output($self),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # _translate_params() translates ioman params to sio specific params and
    # stores it in $self->{run_params}. _validate_params() validates
    # $self->{run_params}. Effectively, it validates params required for sio
    # command invocation
    $self->_translate_params();
    if ($self->error()) {
        # Error set by translate params, just bubble up
        return $self;
    }

    $self->_validate_params();
    if ($self->error()) {
        # Error set by translate params, just bubble up
        return $self;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $self;
}

###############################################################################
# Method:    _translate_params
# Objective: Translate generic IOMan params to sio specific params
# Details:   See POD at the end of this file
###############################################################################
sub _translate_params {
    my ($self) = @_;

    my ($ioman_params, $run_params, $param, $key);
    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $ioman_params = $self->{ioman_params};
    $self->sanlog(-msg => "Translating params:\n\t" . Dumper($ioman_params),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $run_params = $self->{run_params};

    # The syntax of IO.pm's ioman_params is different from params required for
    # sio command invocation. $run_params translates ioman_params to params
    # required for running sio command.  Also,
    # $self->{ioman_params}->{tool_params} are appropriately converted to
    # run_params. Generating $run_params is also necessary to validate params
    # required for sio command invocation (done in new())
    $self->{translate} = 1;
    $self->{failure_action} = $ioman_params->{failure_action};
    $self->{failure_policy} = $ioman_params->{failure_policy};
    $self->{testing_method} = $ioman_params->{testing_method};
    $run_params->{start_byte} = 0;
    $run_params->{bs} = $ioman_params->{bs};
    $run_params->{verify} = $ioman_params->{verify};
    $run_params->{device} = $self->{device};

    if (exists $ioman_params->{runtime}) {
        $run_params->{runtime} = $ioman_params->{runtime};
    }
    if (exists $ioman_params->{pattern}) {
        $run_params->{pattern} = $ioman_params->{pattern};
    }
    if (exists $ioman_params->{limit}) {
        $run_params->{file_size} = $ioman_params->{limit};
    }
    if (exists $ioman_params->{position}) {
        $run_params->{start_byte} = $self->{position};
    }
    if (defined $ioman_params->{tool_path}) {
        $run_params->{tool_path} = $ioman_params->{tool_path};
    }

    # If there is a conflict between tool params and generic params, tool params
    # takes precedence.
    # Allowed sio params: read_perc, random_perc, first_byte, threads, pass_thru
    # There is very minimal checking for pass_thru. We will just let it run and
    # and fail
    # Below code snippet takes all tool_params of $io_params and stores them
    # directly under $run_params
    if (exists $ioman_params->{tool_params}) {
        foreach $key (keys %{$ioman_params->{tool_params}}) {
            if (scalar $ioman_params->{tool_params}->{$key}) {
                $run_params->{$key} = $ioman_params->{tool_params}->{$key};
            }
        }
    }

    unless ($run_params->{read_perc}) {
        if ($ioman_params->{read}) {
            $run_params->{read_perc} = 100;
        } else {
            $run_params->{read_perc} = 0;
        }
    }

    unless ($run_params->{random_perc}) {
        if ($ioman_params->{iotype} eq "random") {
            $run_params->{random_perc} = 100;
        } elsif ($ioman_params->{iotype} eq "sequential")  {
            $run_params->{random_perc} = 0;
        }
    }

    if ($run_params->{pattern}) {
        $run_params->{pass_thru} = $run_params->{pass_thru} . " -fixedfill " .
                                   "-" . $run_params->{pattern};
    }
    if ($run_params->{verify}) {
        $run_params->{pass_thru} = $run_params->{pass_thru} . " -verify";
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE
}

###############################################################################
# Method:    validate_params
# Objective: Validate sio params
# Details:   See POD at the end of this file
###############################################################################
sub _validate_params{
    my ($self) = @_;

    my ($run_params);
    my (@input_params);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $run_params = $self->{run_params};
    $self->sanlog(-msg => "Validating sio params:" .
                          $self->dump_output($run_params),
                  -msg_level => $IOManUtil::DEBUG_HIGH);
    @input_params = ($self->{run_params});

    eval {
        (%$run_params) = Params::Validate::validate( @input_params,
                         {
                             runtime        => { type      => SCALAR,
                                                 optional  => 1,
                                                 default   => 0,
                                                 regex     => qr /^\d+$/ },
                             pattern        => { type      => SCALAR,
                                                 optional  => 1 },
                             start_byte     => { type      => SCALAR,
                                                 default   => 0,
                                                 regex     => qr /^\d+[kgmtKGMT]?$/ },
                             file_size      => { type      => SCALAR,
                                                 optional  => 1,
                                                 regex     => qr /^\d+[kgmtKGMT]?$/ },
                             passes         => { type      => SCALAR,
                                                 optional  => 1,
                                                 regex     => qr /^\d+$/ },
                             bs             => { type      => SCALAR,
                                                 optional  => 1},
                             verify         => { optional  => 1,
                                                 type      => SCALAR,
                                                 regex     => qr /(1|0)/ },
                             dispose        => { type      => SCALAR,
                                                 optional  => 1,
                                                 regex     => qr /keep|delete/ },
                             device         => { type      => SCALAR,
                                                 required  => 1},
                             threads        => { type      => SCALAR,
                                                 optional  => 1,
                                                 default   => 1 },
                             read_perc      => { type      => SCALAR,
                                                 default   => 100,
                                                 callbacks => {'read% must be within 0 and 100' => sub { $_[0] >= 0 and $_[0] <= 100}} },
                             random_perc    => { type      => SCALAR,
                                                 default   => 100,
                                                 callbacks => {'random% must be within 0 and 100' => sub { $_[0] >= 0 and $_[0] <= 100}} },
                             pass_thru      => { type      => SCALAR,
                                                  optional  => 1},

                             tool_path      => { type      => SCALAR,
								                 required  => 1,
                                                 callbacks => {'tool path not accessible' => sub { $self->tool_path_exists(tool_path => $_[0]) } }},
                         });
    };
    if ($@) {
        $self->sanlog(-msg => "Failed to validate params for sio run" .
                              $@,
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        $self->_set_err("Failed to validate parameters for sio run: ". $@);
        return $IOManUtil::FALSE;
    }

    # Validate if we have right combination of params. Potentially there will be
    # a lot but will keep this part simple for now
    if (($run_params->{runtime} == 0) and
        ($run_params->{pass_thru} !~ /(fillonce|numops)/)) {
        # If 100% writes, we'll add the -fillonce option to write once.
        if ($run_params->{read_perc} == 0) {
           $self->sanlog(-msg => "runtime is set to 0 and fillonce is not passed. " .
                                 "This will make sio run in an infinite loop. " .
                                 "Setting pass_thru to -fillonce to avoid.",
                         -msg_level => $IOManUtil::VERBOSE_LOW);
           $run_params->{pass_thru} = "-fillonce " . $run_params->{pass_thru};
           unless ($run_params->{file_size}) {
               $self->sanlog(-msg => "If runtime is 0, file_size must be passed ",
                             -msg_level => $IOManUtil::VERBOSE_LOW);
               $self->_set_err("If runtime is 0, file_size must be passed");
               return $IOManUtil::FALSE;
           }
           if ($run_params->{verify} == 1) {
              $self->sanlog(-msg => "Warning: Data will NOT be verified with -fillonce option!",
                            -msg_level => $IOManUtil::VERBOSE_LOW);
           }
        } else {
           $self->sanlog(-msg => "The runtime and number of operations is set to 0, ".
                                 "which will cause sio to run infinitely. Please " .
                                 "specify a non-zero runtime or -numops parameter!",
                         -msg_level => $IOManUtil::VERBOSE_LOW);
           $self->_set_err("Require runtime or number of operations to avoid infinite loop.");
           return $IOManUtil::FALSE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    start
# Objective: start
# Details:   See POD at the end of this file
###############################################################################
sub start {
    my ($self, %args) = @_;
    my ($run_params, $cmd, $status, $sio_path, $timeout, $line);
    my (@output);
    my ($result);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    %args = &SanUtil::strip_dash(%args);
    $run_params = $self->{run_params};

    $cmd = $run_params->{tool_path} . " " . $run_params->{read_perc} . " ".
           $run_params->{random_perc} . " " .$run_params->{bs} .
           " " . $run_params->{start_byte} . " ". $run_params->{file_size} .
           " " . $run_params->{runtime} . " ". $run_params->{threads} .
           " " . $run_params->{device} . " " . $run_params->{pass_thru};

    $timeout = $self->get_timeout();

    $self->sanlog(-msg => "Running sio cmd " . $cmd . "on device " .
                          $run_params->{device} . " with timeout " .
			              $timeout . " and params \n\t" .
                          $self->dump_output($run_params),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->sanlog( -msg => "Using conn obj" . $self->{conn_obj},
                   -msg_level => $IOManUtil::DEBUG_HIGH);

    ($status, @output) = &ClientUtil::execute($self->{conn_obj}, $cmd, $timeout);
    $self->sanlog(-msg => "sio run complete, result is " . @output,
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # $self->{result} parses the output and stores as follows:
    # 1) raw_output is stored directly under it
    # 2) All generic result params (those that are fairly common across
    #    different io_tools will be stored directly under $self->{result}
    #    So far, these are identified to be the common params:
    #    dev, avg_transfer_rate, io_per_sec, total_elapsed_time, start_time,
    #    end_time and error_count. If the output does not return any of these
    #    common params we set them to "NA"
    # 3) All remaining tool specific result output is stored in
    #    $self->{result}->{tool_specific}
    # The idea of the result hash is to store the params in a way that is
    # consistent across all tools so that Result.pm can expose interfaces that
    # the upper layers (ClientIO.pm and IOMan.pm) can benefit from
    #
    $self->{result}->{raw_output} = \@output;
    $self->{result}->{io_tool} = $self->{io_tool};
    $self->{result}->{client} = $self->{client};
    $self->{result}->{id} = $self->{id};
    $self->{result}->{error_count} = 0;

    #
    # Note: We now support negative and positive testing. For positive testing,
    # the default, we log a failure when an error occurs, as we've always done.
    # But for negative testing, if success is returned, we'll log a failure.
    #

    # If sio returned an error, set error count and return
    if ($status) {
        my $regexp = '(SIO_ERROR:.*)';
        my ($error) = grep(/$regexp/, @output);
        if ( !defined($error) ) {
            $error = "sio run failed";
        }
        if ($self->{testing_method} eq "positive") {
            $self->{result}->{error_count} = 1;
            $self->{result}->{error} = "sio run failed with output " . @output;
            $self->sanlog(-msg => "Error found in result" .
                                  $self->dump_output($self->{result}),
                          -msg_level => $IOManUtil::VERBOSE_LOW);
            # Note: We must execute before sending parent result, or parent will exit!
            $self->execute_failure_action($cmd, $error) if (length $self->{failure_action});
            if ($args{subtested}) {
                &Tharn::addresult(type => "FAIL", "io run failed");
                $Tharn::Parent->message_put(type    => "result",
                                            message => $self->{result});
                $Tharn::Parent->message_put(
                    type    => "failure_policy",
                    message => { failure_policy => $self->{failure_policy} });
            }
            $self->_set_err($self->{result}->{error});
            return $IOManUtil::FALSE;
        } else {
            # Expected failure occurred, so return success!
            $self->sanlog(-msg => "Failure occured as expected, result is " .
                                  $self->dump_output($self->{result}),
                          -msg_level => $IOManUtil::VERBOSE_LOW);
            $Tharn::Parent->message_put(type    => "result",
                                        message => $self->{result});
            return $IOManUtil::TRUE;
        }
    } elsif ($self->{testing_method} eq "negative") {
        my $error = "sio run succeeded but failure expected";
        # Expected failure, but success was returned!
        $self->{result}->{error_count} = 1;
        $self->{result}->{error} = $error;
        $self->sanlog(-msg => "Unexpected success, result is " .
                              $self->dump_output($self->{result}),
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        $self->execute_failure_action($cmd, $error) if (length $self->{failure_action});
        if ($args{subtested}) {
            &Tharn::addresult(type => "FAIL", "run succeeded, but failure expected");
            $Tharn::Parent->message_put(type    => "result",
                                        message => $self->{result});
            $Tharn::Parent->message_put(
                type    => "failure_policy",
                message => { failure_policy => $self->{failure_policy} }
                );
        }
        $self->_set_err($self->{result}->{error});
        return $IOManUtil::FALSE;
    }

    # Now we parse the output and build result hash. If we find
    # errors we just return failure
    $result = $self->_parse_output();

    # Copy result key value pairs to $self->{result}
    foreach my $key (keys %$result) {
        $self->{result}->{$key} = $result->{$key};
    }

    if ($result->{error_count}) {
        # return failure
        $self->sanlog(-msg => "Error found in result" .
                              $self->dump_output($self->{result}),
                      -msg_level => $IOManUtil::VERBOSE_LOW);
        $result->{error} = "Failed to build result " . $self->errmsg();
        if ($args{subtested}) {
            &Tharn::addresult(type => "FAIL", "io run failed");
            $self->sanlog(-msg => "Sending fail message to parent" .
                              $self->dump_output($self->{result}),
                      -msg_level => $IOManUtil::VERBOSE_LOW);
            $Tharn::Parent->message_put(type    => "result",
                                        message => $self->{result});
            $Tharn::Parent->message_put(
                type    => "failure_policy",
                message => { failure_policy => $self->{failure_policy} }
                );

        }
        return $IOManUtil::FALSE;
    }

    # The result hash that was generated using _parse_output() has one flat list
    # of key value pairs. Using translate_result() we convert the result hash
    # to a result hash that is common across different tools
    # Generating result hash is reverse of how we converted common io_params
    # to run_params to invoke the sio command
    # The sio output has sio specific result and we now convert it to generic
    # result hash
    $self->_translate_result($result);

    if ($args{subtested}) {
        &Tharn::addresult(type => "PASS", "run successful");
        $Tharn::Parent->message_put(type    => "result",
                                    message => $self->{result});
    }

    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _parse_output
# Objective: parse sio output and return result hash ref
# Details:   See POD at the end of this file
###############################################################################
sub _parse_output {
    my ($self) = @_;

    my ($line, $result);
    my (@output);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    @output = @{$self->{result}->{raw_output}};
    foreach $line (@output) {
      #
      # When sio finishes correctly, it reports stats.
      #
      # Version: 5.54.1
      # Read: 100 Rand: 0 BlkSz: 512 BegnBlk: 0 EndBlk: 10240 Secs: 5 Threads: 2 Devs: 1  ./tempfile
      # IOPS,   TPUT(KB/s),     LAT(ms),        READ,   RAND,   IOS,    SEC,    THDS    BLKSZ   CSW_IO, INT_IO, SYS,    USR,    WT,     IDL,    uIO,    uKB,    uuIO,   tuIO
      # 4784,   2392,           0.418[0.372],   100,    0,      24159,  5,      2,      512,    2.2,    1.7,    32,     68,     0,      0,      134,    267.6,  284,    418
      # Maximum latency observed:214.655000
      #
      # Terminating threads ...lash<133>

      #Version: 5.61
      #Read: 50 Rand: 50 BlkSz: 1024 BegnBlk: 0 EndBlk: 100 Secs: 120 Threads: 1 Devs: 1  \\\\.\\PHYSICALDRIVE2
      #IOPS,TPUT(KB/s),LAT(ms)Calc'd,LAT(ms)Actual,READ,RAND,IOS,SEC,THDS,BLKSZ,CSW_IO,INT_IO,SYS,USR,WT,IDL,uIO,uKB,uuIO,tuIO
      #504,504,1.984,[1.957],50,50,60496,120,1,1024,2.4,1.1,79,1,0,20,1567,1567.0,20,1587

      #504,    504,            2.0[2.0],       50,     50,     60485,  120,    1,      1024,   2.4,    1.1,    78,     1,      1,      20,     1547,   1547.5

      #openRead: 50 Rand: 50 BlkSz: 1024 BegnBlk: 0 EndBlk: 100 Secs: 90 Threads: 1 Devs: 1  \\\\.\\PHYSICALDRIVE1
      #: No error

      #Version: 6.15\r\r\n
      #Read: 100 Rand: 0 BlkSz: 1024 BegnBlk: 0 EndBlk: 4 Secs: 1073741823 Threads: 1 Devs: 1  \\\\.\\PhysicalDrive1\r\r\n
      #Execution less than expected (runtime 1073741823, actual 1)\r\r\n
      #IOPS,TPUT(KB/s),LAT(ms)Calc'd,LAT(ms)Actual,READ,RAND,IOS,SEC,THDS,BLKSZ\r\r\n
      #4790,4790,0.209,[0.264],100,0,4790,1,1,1024\r\r\n
      #Maximum latency observed:1000.000000\r\r\n
      #\r\r\n
      #Terminating threads ...
      # Normal output lines with no data to return
      # But it may be intermixed with STDERR text
      #$line =~ s/Read:\s+\d+\s*//;
      #$line =~ s/Rand:\s+\d+\s*//;
      #$line =~ s/BlkSz:\s+\d+\s*//;
      #$line =~ s/BegnBlk:\s+\d+\s*//;
      #$line =~ s/EndBlk:\s+\d+\s*//;
      #$line =~ s/Secs:\s+\d+\s*//;
      #$line =~ s/Threads:\s+\d+\s*//;
      #$line =~ s/Devs:\s+\d+\s+\S+//;
      # the above lines are trying to remove:
      # Read: 0 Rand: 0 BlkSz: 65536 BegnBlk: 0 EndBlk: 1600 Secs: 1073741823 Threads: 10 Devs: 1  /dev/rdsk/c3t1d1s6
      # Let's ignore the line
      if ($line =~ /^Read:\s+\d+\s+Rand:\s+/) {
         # this is the display of params info
      } elsif  ($line =~ /^(\d+),\s*(\d+),\s*([\.\d]+),?\[([\.\d]+)\],\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*([\.\d]+),\s*([\.\d]+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*([\.\d]+),\s*(\d+),\s*(\d+)/) {
         $result->{iops} = $1;                   # I/O operations per second
         $result->{tput} = $2;                   # KB per second
         $result->{latency_calculated} = $3;     # latency in milliseconds
         $result->{latency_actual} = $4;         # latency in milliseconds
         $result->{read} = $5;                   # read percentage
         $result->{random} = $6;                 # random percentage
         $result->{ios} = $7;                    # total I/O operations
         $result->{seconds} = $8;                # runlength in seconds
         $result->{threads} = $9;                # threads
         $result->{bs} = $10;                    # block size
         $result->{csw_io} = $11;
         $result->{int_io} = $12;
         $result->{sys} = $13;
         $result->{usr} = $14;
         $result->{wt} = $15;
         $result->{idl} = $16;
         $result->{uio} = $17;
         $result->{ukb} = $18;
         $result->{uuio} = $19;
         $result->{tuio} = $20;
      } elsif ($line =~ /^(\d+),\s*(\d+),\s*([\.\d]+),?\[([\.\d]+)\],\s*(\d+),
                        \s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),
                        \s*([\.\d]+),\s*([\.\d]+),\s*(\d+),\s*(\d+),\s*(\d+),
                        \s*(\d+),\s*(\d+),\s*([\.\d]+)/) {
         # Older versions only sent back 18 values
         $result->{iops} = $1;                   # I/O operations per second
         $result->{tput} = $2;                   # KB per second
         $result->{latency_calculated} = $3;     # latency in milliseconds
         $result->{latency_actual} = $4;         # latency in milliseconds
         $result->{read} = $5;                   # read percentage
         $result->{random} = $6;                 # random percentage
         $result->{ios} = $7;                    # total I/O operations
         $result->{seconds} = $8;                # runlength in seconds
         $result->{threads} = $9;                # threads
         $result->{blksize} = $10;               # block size
         $result->{csw_io} = $11;
         $result->{int_io} = $12;
         $result->{sys} = $13;
         $result->{usr} = $14;
         $result->{wt} = $15;
         $result->{idl} = $16;
         $result->{uio} = $17;
         $result->{ukb} = $18;
      } elsif ($line =~ /^(\d+),\s*(\d+),\s*([\.\d]+),?\[([\.\d]+)\],\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+)/) {
         # Older versions only sent back 18 values
         $result->{iops} = $1;                   # I/O operations per second
         $result->{tput} = $2;                   # KB per second
         $result->{latency_calculated} = $3;     # latency in milliseconds
         $result->{latency_actual} = $4;         # latency in milliseconds
         $result->{read} = $5;                   # read percentage
         $result->{random} = $6;                 # random percentage
         $result->{ios} = $7;                    # total I/O operations
         $result->{seconds} = $8;                # runlength in seconds
         $result->{threads} = $9;                # threads
         $result->{bs} = $10;                    # block size
      } elsif ($line =~ /^0,0,/) {
         # No I/O was performed!
         $result->{iops} = 0;                    # I/O operations per second
         $result->{tput} = 0;                    # KB per second
	 $result->{error_count}++;
	 $result->{error_msg} = "No I/O performed by sio";
      } elsif ($line =~ /^\s*$|^Terminating threads .../) {
         # Normal output lines with no data to return
      } elsif ($line =~ /^sio\s/) {
         # echo of sio input command
      } elsif ($line =~ /^IOPS,\s+TPUT|^IOPS,\s*TPUT\(KB\/s\)/) {
         # Normal output lines with no data to return
      } elsif ($line =~ /I\/O completed successfully|PASS/) {
         # If random options are used, this string is printed in
         # lieu of stats.
      } elsif ($line =~ /Version:\s*(\S+)/) {
         $result->{version} = $1;
      } elsif ($line =~ /Maximum latency observed:\s*(\d+)/) {
         $result->{max_latency} = $1;
      } elsif ($line =~ m/Execution less than expected/ &&
             $self->{run_params}->{pass_thru} =~ m/-numops/ ) {
             # because runlength=0, sec is default to 1073741823
             # use -numops to read
             # we should ignore
             next;
      # ignore extra output messages from verion 6.22
      } elsif ($line =~ /^(global_|Computed)/) {
         next;
      # Configuration Errors
      } elsif ($line =~ /host or user not listed in rhosts/) {
         $result->{error_count}++;
         $result->{error_msg} = "client needs to enable rsh access";
      } elsif ($line =~ / sio:  not found/) {
         $result->{error_count}++;
         $result->{error_msg} = "sio path not found on the client";
         # Report failure
      } elsif ($line =~ /Unknown size type /) {
         $result->{error_count} = 1;
         $result->{error_msg} = "Invalid block size parameter";
      # Other Errors
      } elsif ($line =~ /File open error|CreateFile fails with The system cannot find the file specified/) {
         $result->{error_count} = 1;
         $result->{error_msg} = "Client not configured to run sio on device";
      } elsif ($line =~ m/sio:\s+(.*)/) {
         $result->{error_count}++;
         $result->{error_msg} = "sio reported an error";
      #Version: 6.06\r\n
      #Read: 50 Rand: 50 BlkSz: 1024 BegnBlk: 0 EndBlk: 100 Secs: 240 Threads: 1 Devs: 1  /dev/rnate_lvlog
      #open: No such file or directory
      #File: /dev/rnate_lvlog
      #
      #sio: File open error
      } elsif ($line =~ /SIO_ERROR:/) {
            $result->{error_count} = 1;
            $result->{error_msg} = "The params passed might not be valid." .
                                   "Please double check";
      } elsif ($line =~ m/(open: No such file or directory)/) {
         $result->{error_count} = 1;
         $result->{error_msg} = "Client not configured to run sio on device";
      } elsif ($line =~ m/File: (\S+)/) {
         if ($result->{error_count} == 0) {
             # Previous message should have been seen first,
             # don't count this unless it is first
             $result->{error_msg} = "SIO reported an error on client";
             $result->{error_count}++;
         }
      # write fails
      } elsif ($line =~ m/error accessing file|WriteFile fails|Filename:/) {
         $result->{error_msg} = "SIO reported write error";
         $result->{error_count}++;
      } elsif ($line =~ m/error reading:/) {
         $result->{error_msg} = "SIO reported read error";
         $result->{error_count}++;
      } elsif ($line =~ m/connect to address|Connection refused|Trying\s+\S+\s+rsh/) {
         # connect to address 10.72.237.39 port 544: Connection refused\r\n
         # Trying krb4 rsh...\r\n
         # connect to address 10.72.237.39 port 544: Connection refused\r\n
         # trying normal rsh (/usr/bin/rsh)\r\n
         # just ignore these lines....
         next;
      } elsif ($self->{run_params}->{pass_thru} =~ /niceoutput/ and $line =~ /\w+.*(:).*/) {
        # skip this line
      }
    } # end of main foreach loop

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $result;
}

###############################################################################
# Method:    _translate_result
# Objective: Translate result hash to generic result hash
# Details:   See POD at the end of this file
###############################################################################
sub _translate_result {
    my ($self, $result) = @_;

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    # All generic result keys dev, avg_transfer_rate, io_per_sec,
    # total_elapsed_time, start_time end_time, error_count are stored directly
    # under result. Other result keys are pushed to tool_specific

    $self->{result}->{dev} = $self->{run_params}->{device};
    $self->{result}->{avg_transfer_rate} = delete $result->{tput};
    $self->{result}->{io_per_sec} = delete $result->{iops};
    $self->{result}->{total_elapsed_time} = "NA";
    $self->{result}->{start_time} = "NA";
    $self->{result}->{end_time} = "NA";
    $self->{result}->{tool_specific} = $result;

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

1;
