package IOMan::ClientIO::Manage;

# Compiler directives
use strict;

# Imported modules
use Data::Dumper;
use Data::Compare;
use IOMan::ClientIO::IO;
use San::ConnectionMgr;
use NATE::Events qw(listener_add eventloop_drain);
use Tharn;
use Subtest;
use NATE::STAF::Peer qw($Parent);
use San::SanUtil;
use ParamsMgr;

our @ISA = qw(IOMan::IOManUtil);


###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new Manage object
# Details:   See POD at the end of this file
###############################################################################
sub _new {

    my ($class, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my ($client, $debug_level, $conn_mgr);

    my $self = {
        io_objs            => [],
        client             => "",
        results            => []
    };

    bless $self, $class;

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    &ParamsMgr::process_args(%args);
    $debug_level    = &ParamsMgr::get_arg(-arg => "debug_level", -required => 0,
                                          -default => $IOManUtil::VERBOSE_LOW);
    $self->{client} = &ParamsMgr::get_arg(-arg => "client", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $self;
    }

    $self->set_debug_level(-debug_level => $debug_level);
    if ($self->error()) {
        $self->_set_err("Cannot set debug level", $self->errmsg());
        return $self;
    }

    $self->sanlog(-msg => "Creating Manage object with args " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::VERBOSE_LOW);

    # create conn_mgr. conn_mgr will be used to creating unique conn_obj for
    # each io_obj
    $conn_mgr = ConnectionMgr->new();
    if ($conn_mgr->error()) {
        $self->_set_err("Cannot create connectionMgr object",
                        $conn_mgr->errmsg());
        return $self;
    }
    $self->{conn_mgr} = $conn_mgr;

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $self;
}

###############################################################################
# Method:    manage
# Objective: Manages io objs continously.
#            Creates, starts and waits until io obj run is completed based on
#            messages received.
# Details:   See POD at the end of this file
###############################################################################
sub manage {
    my ($class, %args) = @_;

    my $self = {};
    my ($message, $create_listener, $start_listener, $wait_listener,
        $result_listener, $get_params_listener, $is_running_listener,
        $failure_policy_listener, $stop_listener, $set_siblings_listener);


    # Calls _new() to get a blessed obj. This is done to separate the creation
    # of obj and management of io objs. Also helps testing Manage.pm
    # independently without enterting infinite loop
    $self = IOMan::ClientIO::Manage->_new(%args);
    if ($self->error()){
        $self->_set_err("Failed to create Manage object for client" .
                         $self->{client}, $self->errmsg());
        return $self;
    }

    $self->sanlog(-msg => "Manage object created " . $self->dump_output($self),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    # The below code sets up event based handling of messages that gets sent to
    # to manage(). ClientIO sends messages to manage(), manage() takes
    # appropriate action and sends an acknowledgement back. The listener filters
    # the messages and calls a method based on the type of message
    # This method enters a infinite while loop, however since
    # eventloop_drain() is a blocking call for receiving any messages that
    # match the filters that are set up, this won't be waiting in while loop
    # continously
    #
    # Handles requests coming from client io obj and processes them.
    # Handles one io_obj per each request. ClientIO.pm does the necessary work
    # and sends request for one io_obj. The following requests
    # will be handled:
    # 1) If requests comes in to create io obj, ClientIO creates it and manages
    #    it by adding it to list of io_objs
    # 2) If requests comes in to start a io obj, then appropriate io obj is
    #    started
    # 3) If request comes in to wait until io objs are finished, then this waits
    #    until all io objs are finished and later sends a message back to
    #    ClientIO that wait_finish is done. This lets ClientIO to send
    #    wait_finish request and then do a blocking read to wait until all
    #    io_objs run if finished
    # 4) If request comes in to get results for io objs, then all the results
    #    from io objs are returned. wait_finish flag controls whether
    #    wait_finish is called before getting results.The result of each io_obj
    #    maintains a state - running, completed. These are the possible
    #    scenarios in which get_result can be called:
    #    a) when get_result is called, the message sent by start subtest is not
    #       read. This means this is the first time get_result is called after
    #       start subtest completed. In this case, the message from start
    #       subtest is read and result is built from it
    #       result is generated from it
    #    b) get_result is recalled - The message sent by start subtest is
    #       already read. In such a case, the cached result will be returned
    #    c) get_result is called with wait_finish and start subtest is still
    #       running. In such a case, a result will be returned with state set
    #       to running. This result with state=running is built just before
    #       start subtest is called in start(). This method just returns that
    #       result
    # 5) If request comes in for is_running, all the matched io_objs start
    #    subtests are queried to know whether they are running. If any one of
    #    them is not running, $IOManUtil::FALSE will be returned. Otherwise,
    #    $IOManUtil::TRUE is returned
    # 6) If request comes in to get_params, the matched io_obj ioman_params will
    #    be returned
    #
    # Apart from handling various requests from clientIO, this also handles
    # companion objs and results of all io obj runs.
    # TO DO: Companioin objs
    ## Handle Companion objs
    # 1) If there's a request to start a companion obj, subtest it and add it to
    #    list of companion_objs that needs to be managed
    # 2) Handle companion obj results
    ## Handle Result obj
    # 1) Collect result objs and send it back to ClientIO
    #

    $create_listener = {
        filter => sub { $_[0]->{type} eq "create_io_obj" },
        callback => sub {
                        $self->create_io_obj(%{$_[0]->{message}});
                        if ($self->error()) {
                            $Parent->message_put( type => "create_io_ack",
                                                  message => $self->errmsg()
                                              );
                        }
                        $Parent->message_put( type => "create_io_ack",
                                              message => "creation done" );
                    }
    };
    listener_add($create_listener);

    $start_listener = {
        filter => sub { $_[0]->{type} eq "start_io_obj" },
        callback => sub {
                        $self->start_io_obj(%{$_[0]->{message}});
                        if ($self->error()) {
                            $Parent->message_put( type => "start_io_ack",
                                                  message => $self->errmsg(),
                                              );
                        }
                        $Parent->message_put( type => "start_io_ack",
                                              message => "start done" );
                    }
    };
    listener_add($start_listener);

    $wait_listener = {
        filter => sub { $_[0]->{type} eq "wait_finish" },
        callback => sub {
                        $self->wait_finish(%{$_[0]->{message}});
                        $Parent->message_put( type => "wait_ack",
                                              message => "wait done" );
                    },
    };
    listener_add($wait_listener);

    $is_running_listener = {
        filter => sub { $_[0]->{type} eq "is_running" },
        callback => sub {
                        my $status = $self->is_running(%{$_[0]->{message}});
                        $Parent->message_put( type => "is_running",
                                              message => {status => $status}
                                                   );
                    },
    };
    listener_add($is_running_listener);

    $stop_listener = {
        filter => sub { $_[0]->{type} eq "stop" },
        callback => sub {
                        my $status = $self->stop(%{$_[0]->{message}});
                        $Parent->message_put( type => "stop",
                                              message => {status => $status}
                                          );
                    },
    };
    listener_add($stop_listener);


    $get_params_listener = {
        filter => sub { $_[0]->{type} eq "get_params" },
        callback => sub {
                        my $params_ref = $self->get_params(%{$_[0]->{message}});
                        $Parent->message_put( type => "get_params_ack",
                                              message => $params_ref );
                    },
    };
    listener_add($get_params_listener);

    $result_listener = {
        filter => sub { $_[0]->{type} eq "get_result" },
        callback => sub {
                        my $result = $self->get_result(%{$_[0]->{message}});

                        $Parent->message_put( type => "result_ack",
                                              message => $result );
                        },
    };
    listener_add($result_listener);

    $failure_policy_listener = {
        filter => sub { $_[0]->{type} eq "failure_policy" },
        callback => sub {
                            $self->failure_policy(%{$_[0]->{message}});
                        },
    };
    listener_add($failure_policy_listener);

    # eventloop_drain() processes queue events (sent to it by ClientIO),
    # passes them to filters that we setup (the listeners that are added)
    # If any of the filters match, the corresponding callback is called
    while (1) { eventloop_drain(); }
}

###############################################################################
# Method:    create_io_obj
# Objective: Creates io obj
# Details:   See POD at the end of this file
###############################################################################
sub create_io_obj {
    my ($self, %args) = @_;

    my ($io_obj, $key);
    my (@io_objs);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    $args{debug_level} = $self->{debug_level};

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Manage::create_io_obj called to create IO object
                          with params" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Prepare %args to create IO object
    foreach $key ('client', 'conn_mgr') {
        $args{$key} = $self->{$key};
    }

    # Call cleanup to cleanup the result entries for this io_obj, kill the start
    # process, and delete io_obj entry. See cleanup() for more details
    $self->cleanup(-ids_ref => [$args{id}]);
    if ($self->error()) {
        $self->_set_err("Not able to cleanup io_obj id", $self->errmsg());
        return $IOManUtil::FALSE;
    }

    # Create an io_obj based on the tool passed. All tool.pm's
    # (like sio.pm, dt.pm) inherit from IO.pm. IO.pm does the common work while
    # tool.pm does tool related work. Also, IO.pm calls appropriate tool.pm
    # based on the tool passed
    $io_obj = IOMan::ClientIO::IO->new(%args);
    if ($io_obj->error()) {
        $self->_set_err("Failed to create io obj ", $io_obj->errmsg());
        return $IOManUtil::FALSE;
    }

    # Push the io_obj on to the list of io_objs maintained by Manage.pm. Note
    # that an io_obj is uniquely identfied by client + tool + device name. If
    # an attempt is made to use the same combination of client + tool + device
    # name with different io params, we end up forming the same id. In such a
    # case, we will cleanup the previous id. Only one instance is allowed to
    # run
    push @{$self->{io_objs}}, $io_obj;
    $self->sanlog(-msg => "Pushed io_obj" . $self->dump_output($io_obj) .
                          "to io_objs list: " .
                          $self->dump_output($self->{io_objs}),
                  -msg_level=> $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    start_io_obj
# Objective: Creates a subtest and starts io obj
# Details:   See POD at the end of this file
###############################################################################
sub start_io_obj {
    my ($self, %args) = @_;

    my ($id, $io_obj, $index, $result);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Manage::start_io_obj called to start IO object ".
                          "with params" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    &ParamsMgr::process_args(%args);
    $id = &ParamsMgr::get_arg(-arg => "id", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Get io_obj associated with an id. Note that an id is client + tool +
    # device name and an io_obj is associated for each id
    $io_obj = $self->_get_io_obj(-id => $id);
    if($self->error()){
        # Error already set, just return
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "io_obj found for the $id Starting subtest for " .
                          "\$io_obj->start() with params" .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Creating a subtest for io_obj->start(). The onexit gets run after the
    # subtest finishes. We get the results and add a warning if the result
    # type is not pass. The objective here is to not let this process exit
    # if its subtest fails. We catch the failures, log them as warnings
    # (the log will turn yellow right away even if one io_obj->start()
    # fails) and continue the io runs of other io objs
    my $runid = $id;
    $runid =~ s/:/_/g;
    $io_obj->{subtest} = &Tharn::subtest (
        sub {$io_obj->start( -subtested => 1);},
        '-bg',
        -runid => $runid,
        -onexit => sub {
                        my ($st) = @_;
                        my ($result_ref, @results);
                        @results = $st->results();
                        foreach my $result (@results) {
                            if ($result->type() ne "PASS") {
                                &Tharn::addresult(type => "WARN",
                                                  "IO obj run failed");
                                last;
                            }
                        }
                        $st->exithandler(1);
                    },
    );

    $self->sanlog(-msg => "Created subtest successfully " .
                          $self->dump_output($io_obj),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Build the result hash, set the id and state to running, delete any
    # previous results that might exist. This needs to redone here even though
    # we do it in create method because user is allowed to run start() multiple
    # times on same io_obj
    $result->{id} = $id;
    $result->{state} = "running";

    $self->_delete_results(-ids_ref => [$id]);
    $self->_save_results(-results => [$result]);
    if ($self->error()) {
        # Error already set in _delete_results or _save_results
        return $IOManUtil::FALSE;
    }

    $self->sanlog(-msg => "After start, result is " .
                          $self->dump_output(@{$self->{results}}),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    get_params
# Objective: Returns generic params for given io_obj as hash ref
# Details:   See POD at the end of this file
###############################################################################
sub get_params {
    my ($self, %args) = @_;

    my ($id, $io_obj, $params_ref);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Process the params, id is required param. Only params of one io_obj can
    # be returned
    &ParamsMgr::process_args(%args);
    $id = &ParamsMgr::get_arg(-arg => "id", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    $self->sanlog(-msg => "Getting IO object params for id $id",
                  -msg_level => $IOManUtil::DEBUG_LOW);
    $io_obj = $self->_get_io_obj(-id => $id);
    if($self->error()){
        # Error already set, just return
        return {};
    }

    $self->sanlog(-msg => "io obj found for $id Getting params for io_obj" .
                          $self->dump_output($io_obj),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Call io_obj get_params and return
    $params_ref = $io_obj->get_params();

    if ($io_obj->error()) {
        $self->_set_err("Failed to get params from io_obj " .
                        $io_obj->errmsg());
        return {};
    }
    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return $params_ref;
}

###############################################################################
# Method:    wait_finish
# Objective: waits until all io obj runs are finished
# Details:   See POD at the end of this file
###############################################################################
sub wait_finish {
    my ($self, %args) = @_;

    my ($results_ref, $ids_ref);
    my (@subtests, @io_objs);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Waiting for io obj subtests to finish " .
                          $self->dump_output($self),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # ids_ref is an array ref to ids. This is a required param. Multiple ids
    # can be passed to Manage process. wait_finish will return after the start
    # subtests of all associated io_objs run to completion
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Get subtests associated with the ids passed
    @subtests = $self->_get_subtests(-ids_ref => $ids_ref);

    $self->sanlog(-msg => "ids_ref passed is " . $self->dump_output($ids_ref) .
                          "Waiting for following subtests to finish " .
                          $self->dump_output(@subtests),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Wait until all the subtests run to completion
    Subtest::wait_finish(subtests => [@subtests]);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    get_result
# Objective: returns an array ref of all io_obj result
# Details:   See POD at the end of this file
###############################################################################
sub get_result {
    my ($self, %args) = @_;

    my (@subtests, @results, @filtered_results);
    my ($subtest, $result, $ids_ref, $wait_finish);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    # ids_ref, a required param, is an array ref to ids. wait_finish controls
    # whether this method needs to wait until start subtests are finished.
    # By default, get_result waits until all io_obj io runs started that are
    # associated with ids_ref are completed
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    $wait_finish = &ParamsMgr::get_arg(-arg => "wait_finish", optional => 1,
                                       -default => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # If wait_finish is set, wait until io_runs of all io_objs start runs are
    # completed
    if ($wait_finish) {
        $self->wait_finish(-ids_ref => $ids_ref);
        if ($self->error()) {
            $self->_set_err("Failed to wait_finish", $self->errmsg());
            return [];
        }
    }

    # Get the NATE subtest objs associated with the io_objs. Use them to read
    # any pending messages that are not read from start subtest.
    @subtests = $self->_get_subtests(-ids_ref => $ids_ref);
    $self->sanlog(-msg => "Getting results for the following ids " .
                           $self->dump_output($ids_ref),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # The following are the scenarios that we can run into while getting the
    # result from io_obj
    # Scenario 1: io_obj start sent a result message which hasn't been read by
    #             Manage process yet. In this case, we won't get to on_death
    #             and results will be populated as expected
    # Scenario 2: User reran io_obj start/wait_finish. In this case, we will
    #             still have the result message from io_obj. However while
    #             storing the result to $self->{results}, we should override
    #             the previous result. We might have to maintain the previous
    #             io_obj result too, but based on user feedback this will be
    #             implemented later
    # Scenario 3: User already called get_result once, in this case there wont
    #             be message from io_obj and message_get fails. We handle such
    #             a scenario in on_death
    #
    # If for some reason we end up accessing a dead subtest or a timeout occurs
    # because of NATE issues, log them and proceed with the rest
    foreach $subtest (@subtests) {
        my $result = $subtest->message_get (
                         filter => sub { $_[0]->{type} eq "result" },
                         timeout => 2,
                         on_timeout => sub {
                                           $self->sanlog(-msg => " Subtest timed out " .
                                                                  $self->dump_output($subtest),
                                                         -msg_level => $IOManUtil::DEBUG_HIGH);
                                       },
                         on_death => sub {
                                         $self->sanlog(-msg => " Subtest already exited " .
                                                               $self->dump_output($subtest),
                                                      -msg_level => $IOManUtil::DEBUG_HIGH);
                                     }
                         );
        $self->sanlog(-msg => " Result for requested subtest " .
                              $self->dump_output($result),
                      -msg_level => $IOManUtil::DEBUG_HIGH);

        # Change the result state from running to completed and the push the
        # result to a temporary array
        if (ref ($result) eq "HASH") {
            $result->{state} = "completed";
            push (@results, $result);
        }
    }

    # Save the results, and get use filter_results to get the results of
    # requested io_objs from $self->{results}
    if (scalar @results) {
        $self->_save_results(-results => \@results);
    }

    @filtered_results = $self->_filter_results(-ids_ref => $ids_ref);

    $self->sanlog(-msg => " Results array after saving is " .
                          $self->dump_output($self->{results}),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->sanlog(-msg => "Results returned based on ids ref " .
                          $self->dump_output($ids_ref) . " is " .
                          $self->dump_output(\@filtered_results),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);

    return \@filtered_results;
}

###############################################################################
# Method:    is_running
# Objective: Returns IOManUtil::TRUE if all the io_objs passed are running
#            Otherwise retuns IOManUtil::FALSE
# Details:   See POD at the end of this file
###############################################################################
sub is_running {
    my ($self, %args) = @_;

    my ($results_ref, $ids_ref, $on_fail);
    my (@subtests, @subtests_live, @io_objs);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Verifying is_running for " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    $on_fail = &ParamsMgr::get_arg(-arg => "on_fail", -optional => 1,
                                   -default => undef);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Get the NATE subtest objs associated with the io_objs. Use them to query
    # status of subtests associated with requested io_objs
    @subtests = $self->_get_subtests(-ids_ref => $ids_ref);

    # Make sure there exists atleast one subtest for ids_ref passed. If not
    # return $IOManUtil::FALSE
    unless (scalar @subtests) {
        $self->sanlog(-msg => "No subtests found for ids_ref passed " .
                              $self->dump_output($ids_ref),
                      -msg_level => $IOManUtil::DEBUG_LOW);
        $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
        return $IOManUtil::FALSE;
    }

    # Get live subtests
    @subtests_live = Subtest::get_live();
    if ($self->error()) {
        # Error already set in _delete_results or _save_results
        return $IOManUtil::FALSE;
    }

    # Validate if the none of the io runs failed, If a failure is found return
    # false
    unless ($on_fail eq "ignore") {
        $results_ref = $self->get_result (-ids_ref => $ids_ref,
                                          -wait_finish => 0);
        if ($self->error()) {
            $self->_set_err("Failed to check if io run resulted in failure " .
                            $self->errmsg());
            return $IOManUtil::FALSE;
        }
        $self->sanlog(-msg => "Result is " . $self->dump_output ($results_ref),
                      -msg_level => $IOManUtil::DEBUG_LOW);
        foreach my $result (@$results_ref) {
            if ($result->{error_count}) {
                 $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
                 return $IOManUtil::FALSE;
            }
        }
    }

    # Make sure there is atleast one live subtest
    unless (scalar @subtests_live) {
        $self->sanlog(-msg => "No live subtests found " ,
                      -msg_level => $IOManUtil::DEBUG_LOW);
        $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
        return $IOManUtil::FALSE;
    }

    # Loop through all subtests of requested io_objs and compare them with
    # all live subtests. If a match is not found with any of the live subtests,
    # it means that we found atleast one io_obj subtest which is not live.
    # In such a case, return false
    foreach my $subtest (@subtests) {
        if (List::MoreUtils::any { Compare($subtest, $_) } @subtests_live) {
            $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
            return $IOManUtil::TRUE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::FALSE;
}

###############################################################################
# Method:    failure_policy
# Objective: Returns IOManUtil::TRUE if failure policy is executed successfully
#            Otherwise retuns IOManUtil::FALSE
# Details:   See POD at the end of this file
###############################################################################
sub failure_policy {
    my ($self, %args) = @_;

    my ($failure_policy);
    my (@ids);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Executing failure policy " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $failure_policy = &ParamsMgr::get_arg(-arg => "failure_policy",
                                          -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    if ($failure_policy eq "stop_all_client") {
        List::MoreUtils::all { push (@ids, $_->get_id()) } @{$self->{io_objs}};
        $self->sanlog(-msg => "ids is " .
                          $self->dump_output(\@ids),
                  -msg_level => $IOManUtil::DEBUG_HIGH);
        $self->stop(-ids_ref => \@ids, -warn => 1);
        if ($self->error()) {
            $self->_set_err("Failed to check if io run resulted in failure " .
                            $self->errmsg());
            return $IOManUtil::FALSE;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::FALSE;
}

###############################################################################
# Method:    stop
# Objective: Returns IOManUtil::TRUE if all the io_objs passed are stopped
#            Otherwise retuns IOManUtil::FALSE
# Details:   See POD at the end of this file
###############################################################################
sub stop {
    my ($self, %args) = @_;

    my ($results_ref, $ids_ref, $result, $warn, $message);
    my (@subtests, @subtests_live, @io_objs);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    $self->sanlog(-msg => "Request received to stop runs for " .
                          $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    $warn = &ParamsMgr::get_arg(-arg => "warn", -optional => 1, default => 0);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Set $message based on whether warn is set or not
    if ($warn) {
        $message = ["-warn", "stopping io"];
    } else {
        $message = ["-pass", "stopping io"]
    }

    $self->sanlog(-msg => "stop:: Results before resetting " .
                          $self->dump_output($self->{results}),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    # Loop through all ids, if id has a subtest associated with it that is
    # alive, then we need to stop it and reset the results to "stopped" state
    # Otherwise, there is no need to do anything
    foreach my $id (@{$ids_ref}) {
        my $result = {};
        my @subtests = $self->_get_subtests(-ids_ref => [$id]);
        unless (scalar @subtests and $subtests[0]->is_alive()) {
            next;
        }

        $subtests[0]->kill(message => $message);
        $subtests[0]->wait_finish();
        if ($subtests[0]->is_alive()) {
            $self->_set_err("Subtest still alive after killing");
            return $IOManUtil::FALSE;
        }

        $result->{id} = $id;
        $result->{state} = "stopped";

        $self->_delete_results(-ids_ref => [$id]);
        $self->_save_results(-results => [$result]);
        if ($self->error()) {
            # Error already set in _delete_results or _save_results
            return $IOManUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "Reset result " .
                          $self->dump_output($self->{results}),
                  -msg_level => $IOManUtil::DEBUG_HIGH);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    cleanup
# Objective: Returns IOManUtil::TRUE if all the io_objs passed are cleaned up
#            Otherwise retuns IOManUtil::FALSE
# Details:   See POD at the end of this file
###############################################################################
sub cleanup {
    my ($self, %args) = @_;

    my ($results_ref, $ids_ref);
    my (@subtests, @subtests_live, @io_objs);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);

    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # get all io_objs associated with the ids. If none are found then there is
    # no cleanup to do.
    @io_objs = $self->_get_io_objs(-ids_ref => $ids_ref);
    unless (scalar @io_objs) {
        $self->sanlog(-msg => "No io_objs found for ids" ,
                      -msg_level => $IOManUtil::DEBUG_LOW);
        return $IOManUtil::TRUE;
    }

    # The following are the steps followed to do a graceful cleanup of an
    # existing io_obj
    # Step 1: If the start subtests associated with the ids are still running,
    #         stop the subtests
    # Step 2: Delete the result of this io_obj from $self->{result}
    # Step 3: Delete io_obj from $self->{io_objs}

    # Stop the subtests associated with ids by calling stop()
    # TO DO: Gracefully stop the runs associated with the ids. This needs a
    # flag. The default behavior should be to just return a failure. Users
    # will not be able to start/create io_obj if the io_obj is still running
    # kill() method of subtest is not killing the subtest. This might be related
    # to multiple address spaces. Need further investigation. For now stop()
    # is being commented out


    $self->stop(-ids_ref => $ids_ref);
    if ($self->error()) {
        $self->_set_err("Not able to stop the subtest associated with ids " .
                        $self->errmsg());
        return $IOManUtil::FALSE;
    }

    # Delete the results associated with ids
    $self->_delete_results(-ids_ref => $ids_ref);
    if ($self->error()) {
        $self->_set_err("Not able to delete results associated with ids " .
                        $self->errmsg());
        return $IOManUtil::FALSE;
    }

    # Delete the io_objs associated with ids
    $self->_delete_io_objs(-ids_ref => $ids_ref);
    if ($self->error()) {
        $self->_set_err("Not able to delete io_objs associated with ids " .
                        $self->errmsg());
        return $IOManUtil::FALSE;
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _get_io_objs
# Objective: Returns io_objs
# Details:   See POD at the end of this file
###############################################################################
sub _get_io_objs {
    my ($self, %args) = @_;

    my ($ids_ref, $io_obj);
    my (@io_objs);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    # Loop through each id, use _get_io_obj to get io_obj for that id and store
    # the io_objs in an temporary array
    foreach my $id (@$ids_ref) {
        my $io_obj = $self->_get_io_obj(-id => $id);
        if (scalar keys %$io_obj) {
            push (@io_objs, $io_obj);
            if($self->error()){
                # Error already set, just return
                return ();
            }
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return @io_objs;
}

###############################################################################
# Method:    _get_io_obj
# Objective: Returns io_obj associated with the id passed
# Details:   See POD at the end of this file
###############################################################################
sub _get_io_obj {
    my ($self, %args) = @_;

    my ($id, $io_obj);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    # id is an unique identifier to io_obj and is a required param
    $id = &ParamsMgr::get_arg(-arg => "id", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Loop through all io_objs stored in Manage obj, if a match is found with
    # io_obj id, return that io_obj
    foreach $io_obj (@{$self->{io_objs}}) {
        if ($io_obj->get_id() eq $id) {
            $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
            return $io_obj;
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);

    return {};
}

###############################################################################
# Method:    _delete_io_objs
# Objective: Delete all io_objs asociated with ids_ref passed
# Details:   See POD at the end of this file
###############################################################################
sub _delete_io_objs {
    my ($self, %args) = @_;

    my ($ids_ref, $index);
    my (@io_objs, @results);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Loop through all io_objs, if a match is found (index != 1) delete it
    foreach my $id (@{$ids_ref}) {
        $index = List::MoreUtils::first_index {
                $_->{id} eq $id
             } @{$self->{io_objs}};

        if ($index != -1) {
            splice (@{$self->{io_objs}}, $index, 1);
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);

    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _get_subtests
# Objective: Returns subtests associated with io_objs
# Details:   See POD at the end of this file
###############################################################################
sub _get_subtests {
    my ($self, %args) = @_;

    my ($ids_ref, $id, $io_obj);
    my (@io_objs, @subtests);
    @subtests = ();

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    %args = &SanUtil::strip_dash(%args);


    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    # Get io_objs associated with ids_ref
    @io_objs =  $self->_get_io_objs(-ids_ref => $ids_ref);
    if($self->error()){
        $self->_set_err("Not able to get io_objs for ids passed");
        return ();
    }

    $self->sanlog(-msg => "io objs returned are " .
                          $self->dump_output(\@io_objs),
                  -msg_level => $IOManUtil::DEBUG_LOW);


    # subtest key in io_obj holds the NATE subtest obj. Loop through all
    # io_objs and get the subtest objs
    if (scalar @io_objs) {
        List::MoreUtils::all {
            push(@subtests, $_->{subtest}) if (defined $_->{subtest})
        } @io_objs;
    }

    $self->sanlog(-msg => "subtests returned are " .
                          $self->dump_output(\@subtests),
                  -msg_level => $IOManUtil::DEBUG_LOW);

    $self->exit(-msg_level => $IOManUtil::DEBUG_LOW);
    return @subtests;
}

###############################################################################
# Method:    _filter_results
# Objective: Returns results associated with ids passed
# Details:   See POD at the end of this file
###############################################################################
sub _filter_results {
    my ($self, %args) = @_;

    my ($ids_ref, $result);
    my (@io_objs, @results);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    # Loop through all ids, push the results if a match is found with id of
    # results stored in Manage obj to a temp array and return this array at
    # the end
    foreach my $id (@$ids_ref) {
        $result = List::MoreUtils::first_value {
                      $_->{id} eq $id
                  } @{$self->{results}};
        unless ($result) {
            $self->_set_err("Failed to find result for id" . $id);
            return ();
        }
        push (@results, $result);
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return @results;
}

###############################################################################
# Method:    _save_results
# Objective: saves results passed to internal results array that Manage obj
#            maintains
# Details:   See POD at the end of this file
###############################################################################
sub _save_results {
    my ($self, %args) = @_;

    my ($match_index, $results_ref);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);

    %args = &SanUtil::strip_dash(%args);

    # A result is a simple hash and results passed to this method is a reference
    # to an array of such result hashes
    &ParamsMgr::process_args(%args);
    $results_ref = &ParamsMgr::get_arg(-arg => "results", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Loop through the array of results passed through this method. If the id
    # of the resultis matched with any of the internal results array maintained
    # by Manage obj, then the existing result of Manage obj is overwritten.
    # Otherwise the results is pushed into internal results array of Manage obj
    foreach my $curr_result (@$results_ref) {
        $match_index = List::MoreUtils::first_index {
                          $curr_result->{id} eq $_->{id}
                       } @{$self->{results}};

        if ($match_index != -1) {
            ${$self->{results}}[$match_index] = $curr_result;
        } else {
            push (@{$self->{results}}, $curr_result);
        }
    }

    $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    _delete_results
# Objective: delete results associated with ids passed
# Details:   See POD at the end of this file
###############################################################################
sub _delete_results {
    my ($self, %args) = @_;

    my ($ids_ref, $index);
    my (@io_objs, @results);

    $self->entry(-msg_level => $IOManUtil::DEBUG_LOW);
    %args = &SanUtil::strip_dash(%args);

    # ids_ref, a required param, is an array ref to ids. Each id is an unique
    # identifier to io_obj
    &ParamsMgr::process_args(%args);
    $ids_ref = &ParamsMgr::get_arg(-arg => "ids_ref", -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # Loop through all ids and attempt to find a match with ids of results
    # maintained by Manage obj. If a match is found, delete that result
    foreach my $id (@{$ids_ref}) {
        $index = List::MoreUtils::first_index {
                          $id eq $_->{id}
                       } @{$self->{results}};
        if ($index != -1) {
            splice (@{$self->{results}}, $index, 1);
        }
    }

    return $IOManUtil::TRUE;
}

1;
