package IOMan::IOManUtil;

use strict;

use Data::Dumper;
use San::SanUtil;
use ParamsMgr;
use San::SanLog;
use Params::Validate qw ( :all );
use List::MoreUtils;
use Error;
use Tharn;
use NATE::Process;
use NATE::ProcManager;
use San::SanClient::ClientUtil;

use vars qw (@ISA);

@ISA = qw( Error SanLog );

$IOManUtil::VERBOSE_LOW  = 1;
$IOManUtil::VERBOSE_HIGH  = 3;
$IOManUtil::DEBUG_LOW  = 5;
$IOManUtil::DEBUG_HIGH  = 7;
$IOManUtil::RUN_TIMEOUT = 1800;
$IOManUtil::TRUE = $SanUtil::TRUE;
$IOManUtil::FALSE = $SanUtil::FALSE;
@IOManUtil::TOOLS = qw( sio dt );
%IOManUtil::TOOL_PATH = ('linux' =>   [{'dt'  => {'stable'  => '/x/eng/localtest/arch/i386-redhat-linux8/bin/dt.stable',
                                                  'current' => '/x/eng/localtest/arch/i386-redhat-linux8/bin/dt.latest',
                                                  'staging' => '/x/eng/localtest/arch/i386-redhat-linux8/bin/dt.staging',
                                                  'last'    => '/x/eng/localtest/arch/i386-redhat-linux8/bin/dt.last'}},
                                       {'sio' => {'stable'  => '/x/eng/localtest/arch/i386-redhat-linux8/bin/sio',
                                                  'current' => '/x/eng/localtest/arch/i386-redhat-linux8/bin/sio.current',
                                                  'staging' => '/x/eng/localtest/arch/i386-redhat-linux8/bin/sio.last',
                                                  'last'    => '/x/eng/localtest/arch/i386-redhat-linux8/bin/sio.last'}}
                                      ],
                         'windows' => [{'dt'  => {'stable'  => 'x:\SAN\bin\common\dt.stable\dt.exe',
                                                  'current' => 'x:\SAN\bin\common\dt.latest\dt.exe',
                                                  'staging' => 'x:\SAN\bin\common\dt.staging\dt.exe',
                                                  'last'    => 'x:\bin\dt.last.exe'}},
                                       {'sio' => {'stable'  => 'x:\bin\sio.exe',
                                                  'current' => 'x:\bin\sio.current.exe',
                                                  'staging' => 'x:\bin\sio.last.exe',
                                                  'last'    => 'x:\bin\sio.last.exe'}}
                                      ],
                        );

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:     new
# Objective:  Sets debug level of ParamsMgr, SanLog and SubIPC
# Parameters: Hash: -debug_level (Mandatory) Debug level to set
# Returns:    1|0
# Details:    See POD at the end of this file
###############################################################################
sub set_debug_level{
    my ($self, %args) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # Setting SanLog debug level
    $self->SUPER::set_debug_level(0);

    # Setting ParamsMgr debug level
    &ParamsMgr::debug_level(-debug_level => 0);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }
    $self->{debug_level} = $args{debug_level};

    return $IOManUtil::TRUE;
}

###############################################################################
# Method:    dump_output
# Objective: dump output from object
# Input:     Any hash or array object
# Return:    string from the dump
###############################################################################
sub dump_output
{
    my ($self, $obj) = @_;

    my ($d_obj);

    # lets' create new Data::Dumper object
    $d_obj = Data::Dumper->new([$obj], [ref($obj)]);

    if ($self->{debug_level} < $IOManUtil::DEBUG_HIGH) {
        $d_obj->Maxdepth(2);
    } else {
        $d_obj->Maxdepth(5);
    }

    return "\n\t" . $d_obj->Dump();
}

################################################################################
# Method:    tool_path_exists
# Objective: return true if tool path exists or if executable is passed
# Input:     Tool path
# Return:    IOManUtil::TRUE or IOManUtil::FALSE
################################################################################
sub tool_path_exists {
    my ($self, %args) = @_;

    my (@output, $status, $cmd, $tool_path);

    $self->entry(-msg_level => $IOManUtil::DEBUG_HIGH);
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    &ParamsMgr::process_args(%args);

    $tool_path = &ParamsMgr::get_arg(-arg => "tool_path",
                                     -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $IOManUtil::FALSE;
    }

    # If just the tool name is passed, then return true. This method only
    # makes sure that a path exists if a tool path is passed. This lets
    # IOMan to default to tool executable if tool path is not passed
    if (List::MoreUtils::any { $_ eq $tool_path } @IOManUtil::TOOLS) {
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return $IOManUtil::TRUE;
    }

    # For windows use dir command and other os's just use ls
    if ($self->{conn_obj}->{hostp}->{hosttype} =~ /windows/i) {
        $cmd = "dir " . $tool_path;
    } else {
        $cmd = "/bin/ls " . $tool_path;
    }

    ($status, @output) = &ClientUtil::execute($self->{conn_obj}, $cmd);
    if ($status != 0) {
        $self->exit(-msg_level => $IOManUtil::DEBUG_HIGH);
        return $IOManUtil::FALSE;
    }

    return $IOManUtil::TRUE;
}

###############################################################################
# Method:     _print_header (Helper function from FitMgr)
# Objective:  Adds a simple box around a message for logging purposes.
# Parameters: None.
# Returns:   The pretty message
###############################################################################
sub _print_header {
    my ($self, $msg) = @_;

    my $header  = "\n" x 2 . "#-" x 39 . "#\n";
    my $body    = "#  " . $msg . "\n";
    my $footer  = "\#-" x 39 . "#\n";

    # Prints something like (easier to identify FIT entries in the logs...):
    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
    #  Some message
    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
    return ($header . $body . $footer);
}


###############################################################################
# Method:     execute_script - Execute a script or binary.
# Objective:  Simple wrapper to execute an external script.
# Parameters: %args = Parameters for NATE::Process
# Returns:   The script results on failure or nothing on success.
###############################################################################
sub execute_script {
    my ($self, %args) = @_;

    my $proc = NATE::Process->new(
                                inherit_params => 1,
                                inherit_env => 1,
                                %args
                                );
    $self->sanlog(-msg => "NATE::Process params" . $self->dump_output(\%args),
                  -msg_level => $IOManUtil::DEBUG_HIGH);
    $proc->start();
    $proc->wait();
    my @results=$proc->own_results();
    my $resultsmsg = "";
    if ( scalar(@results) ) {
        $self->sanlog(-msg => "NATE::Returned results" . $self->dump_output(@results),
                      -msg_level => $IOManUtil::DEBUG_HIGH);
        foreach my $rslt (@results) {
            if ($rslt->type() !~ /(PASS|WARN|INFO|PARAM)/i) {
                my @msgs=$rslt->messages();
                $resultsmsg=$resultsmsg . "\n(" . $rslt->type() . ") " . join("\n", @msgs);
            }
        }
    }
    $proc->destroy();
    if ($resultsmsg) {
        $resultsmsg=~s/^\n//g;
    };
    return($resultsmsg);
}

1;
