#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/IOMgr.pm#1 $

## @summary Provide an interface to various IO tools across host systems
## @author lippe@netapp.com
## @status private
## @pod here

# Package Name
##############
package IOMgr;

use strict;
use Error;

use IOMgr::IO::Dt;
use IOMgr::IO::Random;
use San::SanClient::ClientUtil;

use vars qw(
	    $_CNT
	    @ISA
	    );

@ISA = qw(Error);

################################################################################
# Subroutine Name: new
# Objective:       Create new IOMgr object
# Details:         See POD documentation at the end of this file
################################################################################
sub new {
    my ($class, %args) = @_;
    my ($self, 
	$log);
    
    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    foreach (keys %args) {
	if (/-?loghdl$/) {
	    length($log = $args{$_}) or
		do {
		    $self->_set_err("IOMgr::new invalid value for \"log\" $args{$_}");
		    return $self;
		};
	} else {
	    $self->_set_err("IOMgr::new: Invalid parameter specified: $_");
	    return $self;
	}
    }

    $self = {};

    # Allow CLI commands to be logged;
    $self->{log} = $log if (defined $log);

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    return($self);
}

################################################################################
# Subroutine Name: datatest
# Objective:       Call datatest object constructor and return object handle
# Details:         See POD documentation at the end of this file
################################################################################
sub datatest {
    my ($self, %args) = @_;
    my ($ioObj,
	$mntObj,
	$mntprop,
	$name);

    $self->_clear_err();  

    $mntObj = $args{-mntobj};
    $mntprop = $mntObj->get_properties();
    
    $name = "datatest" . $_CNT++;

    # Mount object must be a reference and is requuired
    unless (defined $args{-mntobj}) {
	$self->_set_err("Error: IOMgr::datatest() Invalid value for argument '-mntobj'");
	return $self; 
    }

    # Specify a default filename to use
    unless (defined $args{-file} and length $args{-file}) {
	$args{-file} = $name . '.file';
    }

    # Log handle
    $args{-loghdl} = $self->{log} if (defined $self->{log});

    # Pass in appropriate args to DT constructor
    $ioObj = Dt->_new(-iomgr=>$self,
		      -name=>$name,
		      -verbose=>1,
		      %args);

    if ($ioObj->error){
        $self->_set_err($ioObj->errmsg());
        return $self;                                    
    }

    # Keep pointer to resulting object referenced by mount point
    push @{$self->{io}{$mntprop->{path}}}, $ioObj;

    return $ioObj;
}

################################################################################
# Subroutine Name: random
# Objective:       Call random I/O object constructor and return object handle
# Details:         See POD documentation at the end of this file
################################################################################
sub random {
    my ($self, %args) = @_;
    my ($ioObj,
	$mntObj,
	$mntprop,
	$name);

    $self->_clear_err();  

    $mntObj = $args{-mntobj};
    $mntprop = $mntObj->get_properties();
    
    $name = "random" . $_CNT++;

    # Mount object must be a reference and is requuired
    unless (defined $args{-mntobj}) {
	$self->_set_err("Error: IOMgr::random() Invalid value for argument '-mntobj'");
	return $self; 
    }

    # Log handle
    $args{-loghdl} = $self->{log} if (defined $self->{log});

    # Pass in appropriate args to DT constructor
    $ioObj = Random->_new(-iomgr=>$self,
			  -name=>$name,
			  -verbose=>1,
			  %args);
    
    if ($ioObj->error){
        $self->_set_err($ioObj->errmsg());
        return $self;                                    
    }

    # Keep pointer to resulting object referenced by mount point
    push @{$self->{io}{$mntprop->{path}}}, $ioObj;

    return $ioObj;
}

################################################################################
# Subroutine Name: get_io
# Objective:       Returns array of IO objects, each of which represents a 
#                  unique test.  If a mount object is passed, returns associated
#                  IO object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_io {
    my ($self, %args) = @_;
    my ($mntObj,
	@all_objs);

    $self->_clear_err();  

    foreach (keys %args) {
	if (/-?mntobj$/) {
	    length($mntObj = $args{$_}) or
		do {
		    $self->_set_err("IOMgr::get_io: Invalid value for \"mntobj\" $args{$_}");
		    return ()
		};
	} else {
	    $self->_set_err("IOMgr::get_io: Invalid parameter specified: $_");
	    return ();
	}
    }

    if (defined $self->{io}) {
	if (defined $mntObj) {
	    return (wantarray()) ? @{$self->{io}{$mntObj->{path}}} : \@{$self->{io}{$mntObj->{path}}};
	} else {
	    foreach (keys %{$self->{io}}) {
		foreach my $obj (@{$self->{io}{$_}}) {
		    push @all_objs, $obj;
		}
	    }
	    return (wantarray()) ? @all_objs : \@all_objs;
	}
    } else {
        $self->_set_err("Error: IOMgr::get_io() IO manager has no tests defined");
        return ();              
    }
}

################################################################################
# Subroutine Name: destroy
# Objective:       Iterates through contained objects, subsequently calling 
#                  individual destroy routines
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;
    my (@error,
	$fail_cnt,
	$mntpt);

    $self->_clear_err();  

    foreach $mntpt (keys %{$self->{io}}) {
	foreach (@{$self->{io}{$mntpt}}) {
	    $_->destroy();
	    $fail_cnt++ if ($_->error);
	}
	sleep 5;
    }

    return ($fail_cnt eq 0) ? $ClientUtil::TRUE : $ClientUtil::FALSE;
}

1;
