# this is basically -*-perl-*-# -*-perl-*x
#
# $Id: //depot/prod/test/main/lib/IPU/7mode_utils.pm# $
#
# Copyright (c) 2008 NetAp, Inc.
# All rights reserved.
#
## @summary
##     Library functions for 7mode LA IPU related testcases
##
## @description
##     This library will have functions to invoke all the new UI that has  
##     been added as a part of LA IPU feature 
##
## @usage
##     use IPU::7ModeIPU;
## 
##   @routines
##      create_multiple_aggrs
##      create_multiple_vols
##      cleanup_ipu_setup
##      aggr_upgrade_diag
##      aggr_upgrade_status_diag
##      parse_upgrade_check
##      aggr_upgrade_scan_start
##      aggr_upgrade_scan_stop
##      vol_upgrade_diag
##      get_upgrade_status
##
## @author nagpal@netapp.com
##
## @status private
## 

package IPU::7ModeIPU;

use constant FC_TRUE  => 1;
use constant FC_FALSE => 0;
use constant FC_ERROR => -1;

use strict;
use warnings;
use Data::Dumper;
use Carp;
use TCD;
use Aggregate qw(aggr_status aggr_exists aggr_create aggr_purge aggr_root
                aggr_disk_free aggr_add);
use Volume qw( vol_create_flex vol_purge vol_container vol_size vol_max_files
              vol_exists);
use Params::Validate qw(validate
                OBJECT
                SCALAR
                HASHREF
                ARRAYREF
                SCALARREF);
use SayWrap qw(sayEOF sayWAIT);
use Tharn qw (subtest host $Log);
use HybridBuftree;
use filer;
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(create_multiple_aggrs
                    create_multiple_vols
                    cleanup_ipu_setup
                    aggr_upgrade_diag
                    aggr_upgrade_status_diag
                    parse_upgrade_check
                    aggr_upgrade_scan_start
                    aggr_upgrade_scan_stop
                    vol_upgrade_diag
		    get_upgrade_status
                );

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

################################################################################
# Function:      create_multiple_aggrs (default 32 bit aggrs)  
#
# Synopsis:      Creates multiple  aggrs based on Input Parameters
#
# InputParameters:
#   Fc              => Filer connection object
#   aggr_size       => number of disks in an aggr (default 3)
#   aggr_count      => number of aggrs to be create (default 1)
#   block_type      => 64-bit or 32-bit aggr (default 32)
#                      possible values (64: 64-bit aggr, 32: 32-bit aggr)
#   aggr_prefix     => prefix to be added to each aggr (default testaggr)
#   purge_existing  => purge existing aggrs if they exists (default 1 i.e purge)
#                      possible values (0: no purge, 1: purge)
#   error_msg       => reference for error message to be captured
#
# Returns:          reference for array of aggr_names that were created
#		    die on failure
#
# Usage:
#               $aggr_list = create_multiple_aggrs(
#                              Fc        => $Fc,
#                              aggr_count  => '2',
#                              block_type  => '32',
#                              aggr_prefix => 'testaggr',
#                              purge_existing  => '1',
#                              error_msg => \$error_msg);
#
# Names of aggr created with default prefix:
#       <filername_prefix_#> => wqa_gx98_testaggr_1 ...
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub create_multiple_aggrs {
   my (%opts) = validate (@_,
            { Fc        =>  {type => HASHREF},
              aggr_size =>  {type => SCALAR, optional => 1, default => '3'},
              aggr_count  => {type => SCALAR, optional => 1, default => 1 },
              block_type  => {type => SCALAR, optional => 1, default => '32'},
              aggr_prefix => {type => SCALAR, optional => 1, default => 'testaggr'},
              purge_existing  => {type => SCALAR, optional => 1, default=>1},
              error_msg   => {type => SCALARREF},});


    my (@aggr_list);
    my $error_msg = $opts{error_msg};
    
    my $filername_with_underscores = filer_name($opts{Fc});
    $filername_with_underscores =~ s/-/_/g;
    # Create Aggregates while checking if they exist and if they are already $block_type
    eval {  
        for (my $i=1; $i<=$opts{aggr_count}; $i++) { 
            my $aggr_name=$filername_with_underscores."_".$opts{aggr_prefix}."_".$i;
            if(aggr_exists($opts{Fc}, $aggr_name)){ #Check if aggr exists
                if($opts{purge_existing}){
                        aggr_create($opts{Fc},
                                    $aggr_name,
                                    $opts{aggr_size},
                                    indirect_type => $opts{block_type},
                                    purge=>1)
                        or die "Failed to create aggr $@";
                }elsif ((aggr_is_convertible($opts{Fc},$aggr_name) && $opts{block_type} =~ /32/ ) ||
                         (!(aggr_is_convertible($opts{Fc},$aggr_name)) && $opts{block_type} =~ /64/)){
                    $Log->debug("$aggr_name exists..Reusing");
                }else{
                    $Log->debug("$aggr_name exists..but with incorrect block-type");
                    die "Existing aggr $aggr_name block_type does not match request,
                         you may want to use purge_existing=1 for function call";
                } 
            } else{
                $Log->debug("Creating aggregate $aggr_name\n");
                aggr_create($opts{Fc}, $aggr_name, $opts{aggr_size},
                            indirect_type => $opts{block_type},
                            purge=> $opts{purge_existing})
                or die "fail to create aggr $@";
            }
            push (@aggr_list, $aggr_name);
        }
    };
    $$error_msg = $@ if $@;
    return \@aggr_list;
}#End of function create_multiple_aggrs


################################################################################
# Function:      create_multiple_vols  
#
# Synopsis:      Creates multiple volume on each aggr based on Input Parameters
#
# InputParameters:
#   Fc              => Filer connection object
#   aggr_names      => one aggr name (scalar) or aggr_list (arrayref) to create
#                       volumes on.
#   vol_count       => number of volumes to be create (default 1)
#   vol_prefix      => prefix to be added to each vol (default testvol)
#   purge_existing  => purge existing vols if they exists (default 1 i.e purge)
#                      possible values (0: no purge, 1: purge)
#   vol_size        => size of each volume (default 20g)
#   error_msg       => reference for error message to be captured
#
# Returns:          reference for array of volume names that were created
#		    die on failure
#
# Usage:
#        $vol_list = create_multiple_vols(
#                        Fc        =>  $Fc,
#                        aggr_names => $aggr_list,
#                        vol_count  => '3',
#                        vol_prefix => 'testvol',
#                        vol_size => '1g',
#                        error_msg => \$error_msg);
#
# Names of aggr created with default prefix:
#       <aggrname_prefix_#> => wqa_gx98_testaggr_1_testvol_1 ...
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub create_multiple_vols {
     my (%opts) = validate (@_,
            { Fc          => {type => HASHREF},
              aggr_names  => {type => SCALAR | ARRAYREF},
              vol_count   => {type => SCALAR, optional => 1, default => 1 },
              vol_prefix  => {type => SCALAR,optional => 1,default => 'testvol'},
              purge_existing  => {type => SCALAR, optional => 1, default=>1},
              vol_size    => {type => SCALAR, optional => 1, default=>'20g'},
              error_msg   => {type => SCALARREF},});

    my (@aggrnames, @vol_list);
    my $aggr_names    = $opts{aggr_names};
    my $error_msg = $opts{error_msg};
        
    if(ref($aggr_names) eq "ARRAY") {
	push(@aggrnames,@$aggr_names);
    } else {
        push(@aggrnames,$aggr_names);
    }
    
    foreach my $aggr_name (@aggrnames) {
        for (my $count = 0; $count < $opts{vol_count}; $count++) {
            my $vol_name = $aggr_name."_".$opts{vol_prefix}."_".$count;
            if(vol_exists($opts{Fc}, $vol_name)){ #Check if aggr exists
                if($opts{purge_existing}){
                        vol_purge($opts{Fc}, $vol_name)
                        or die("Failed to purge $vol_name: $@");    
                    } else {
                        $Log->debug("$vol_name exists..Reusing");
                    }
            } else{
                $Log->debug("Creating volume $vol_name\n");
 
                vol_create_flex($opts{Fc},
                                $vol_name,
                                $aggr_name,
                                $opts{vol_size})
                or die "Failed to create vol $vol_name: $@";
            }
            push (@vol_list, $vol_name);
        }
    }
    $$error_msg = $@ if $@;
    return \@vol_list;;  
}#End of function create_multiple_vols

################################################################################
# Function:      cleanup_ipu_setup  
#
# Synopsis:      Delete a given number of vols on each of the aggregate 
#		   mentioned and the aggr itself
#
# InputParameters:
#   Fc          => Filer connection object
#   aggr_names  => one aggr name (scalar) or aggr_list (arrayref) to be deleted
#   error_msg   => reference for error message to be captured
#
# Returns:          1 if success
#		    die on failure
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub cleanup_ipu_setup {

    my (%opts)= validate (@_, {
                        Fc          =>  {type => HASHREF},
                        aggr_names   => {type => SCALAR | ARRAYREF},
                        error_msg   => {type => SCALARREF},});

    my (@aggrnames);
    
    my $aggr_list    = $opts{aggr_names};
    my $error_msg = $opts{error_msg};

    if(ref($aggr_list) eq "ARRAY") {
	push(@aggrnames,@$aggr_list);
    } else {
        push(@aggrnames,$aggr_list);
    }
    $Log->debug("Cleaning up aggrs created by script.\n");
    foreach my $aggr_name (@aggrnames) {
        aggr_purge($opts{Fc}, $aggr_name)
        or die "Aggr delete for $aggr_name failed: $@";
    }
    $$error_msg = $@ if $@; 
    return FC_TRUE;
}#End of function cleanup_ipu_setup

################################################################################
# Function:      aggr_upgrade_diag  
#
# Synopsis:      Extension of HybridBuftree::aggr_convert to take 
#                mode for IPU and not setup setflag in the call.
#
# InputParameters:
#   Fc          => Filer connection object
#   aggr_name   => name of aggr (scalar) to be upgraded to 64-bit
#   mode        => mode of upgrade (default normal)
#                  Possible volues:
#                   check   => check to see if aggr can be upgraded or
#                              the space needed for upgrade
#                 normal    => use this option if a user wants to have full
#                              control of the volume sizes for all volumes
#                              in an aggregate.
#               grow-all    => Use this option if a user wants to ensure the
#                               completion of format upgrade for the aggregate
#                               and all its contained volumes without
#                               fs_size_fixed option set) after "aggr add".
#            grow-reserved   => Use this option if a user wants to maintain the
#                               volumes sizes for existing non-required fixed
#                               sized volumes in an aggregate, except in cases
#                               when additional space must be added to volumes
#                               without the fs_size_fixed option set to
#                               accommodate the upgrade of the existing space
#                               reserved files. 
#   error_msg   => reference for error message to be captured
#
# Returns:      return values are all scalar and are baed on mode:
#                   check   => return raw output of -mode=check command on
#                              success.This output can be fed to
#                              parse_upgrade_check_diag to get specific values
#                              0 for failure
#                             -1 for command failure
#                 normal,grow-all,grow-reserved
#                           => retuns 1 for success,
#                                     0 for failure,
#                                     -1 for command failure
#
# Usage:
#       $status = aggr_upgrade_diag(
#                   Fc  => $Fc,
#                   aggr_name => $$aggr_list[0],
#                   mode => 'check',
#                   error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_diag {
  my (%opts) = validate (
        @_,{             
            Fc   =>	{ type => HASHREF},
            aggr_name => { type => SCALAR },
            mode => { type => SCALAR , optional=>1, default => 'normal'},
            error_msg   => {type => SCALARREF},});
  
    my $status;
    my $mode_str="-mode $opts{mode}";
    my $error_msg = $opts{error_msg};
    
    $Log->debug("Starting 64bit-upgrade on Aggr $opts{aggr_name}
                in $opts{mode} mode\n");
    eval {
	$status = sayEOF( $opts{Fc},
                         "priv set -q diag;
                         aggr 64bit-upgrade start $opts{aggr_name} $mode_str");
    };
    #checking the error msg.
    if ($status =~ /Start(ed|ing) 64-bit upgrade on aggregate/ && $opts{mode} !~ /check/){
        $Log->debug("64-bit upgrade started on $opts{aggr_name}");
        $$error_msg = $@ if $@; 
	return $status;   
    }
    elsif ($opts{mode} =~ /check/ && $status =~ /Checking for additional space/)
    {
        $Log->debug("64-bit upgrade status returned for $opts{aggr_name}");
        $$error_msg = $@ if $@; 
	return $status;   
    }
    else{
        $Log->debug("64-bit upgrade on $opts{aggr_name} Failed: ".$status);
        $$error_msg = $@ if $@; 
	return FC_ERROR; 
    }
    $$error_msg = $@ if $@; 
    return FC_FALSE;
}

################################################################################
# Function:      aggr_upgrade_status_diag  
#
# Synopsis:      Results from aggr 64bit-upgrade status command
#
# InputParameters:
#   Fc          => Filer connection object
#   aggr_name   => name of aggr(scalar) to be check for status of 64-bit upgrade
#   error_msg   => reference for error message to be captured
#
# Returns:      raw output of command on success. 
#               0 when not upgrading
#               -1 for command failure
#
# Usage:
#       $status = aggr_upgrade_status_diag(
#                   Fc  => $Fc,
#                   aggr_name => $$aggr_list[0],
#                   error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_status_diag {
my (%opts) = validate (
        @_,{             
            Fc   => { type => HASHREF},
            aggr_name => { type => SCALAR },
            error_msg   => {type => SCALARREF},});

    my $status;
    my $aggr_name = $opts{aggr_name};
    my $error_msg = $opts{error_msg};
  
    $Log->debug("Checking if 64bit-format upgrade is running\n");
    eval {
	$status = sayEOF( $opts{Fc}, "priv set -q diag;
                         aggr 64bit-upgrade status $aggr_name");
    };
    if ($status =~ /64-bit upgrade is not currently in progress on aggregate $aggr_name/) {
	$Log->debug("64-bit upgrade is not running ".$@);
        $$error_msg = $@ if $@; 
	return FC_FALSE;
    }
    elsif($status =~ /upgrading/){
        $$error_msg = $@ if $@; 
        return $status;
    }
    else{
	$Log->debug("64-bit upgrade status command FAILED: ".$status);
        $$error_msg = $@ if $@; 
	return FC_ERROR;   
    }
}

################################################################################
# Function:      parse_upgrade_check 
#
# Synopsis:      Parses the output from aggr_upgrade_diag or aggr_upgrade
#                with mode=check
#
# InputParameters:
#   output  => output from -mode check of aggr/vol 64bit-upgrade -mode=check
#                  or aggr add -64bit-upgrade check commands 
#   info    => type of information required
#              Posssible values:
#              VOL_INFO_AFTER_UPGRADE (default): gives the list(hash) of volumes
#                                      that will be upgraded and their
#                                      usage after upgrade
#              VOL_SPACE_INCREASE_BY : gives the hash of volumes which need
#                                       more space before upgrade
#              FILESYS_SIZE_FIXED   : gives the array of volumes with
#                                      fs_sized_fixed set
#   error_msg   => reference for error message to be captured
#
# Returns:  return hash ref or array ref or scalar based on info
#    VOL_INFO_AFTER_UPGRADE (HASHREF): example
#          'Used' => '204MB',
#          'fs_size_fixed' => 'no',
#          'Capacity' => '20%',
#          'Available' => '819MB',
#          'Volume Name' => 'wqa_gx98_testaggr_2_testvol_1'
#    VOL_SPACE_INCREASE_BY (HASHREF): example
#          'Volume size to be increased by' => '172KB',
#          'fs_size_fixed' => 'no',
#          'Volume Name' => 'wqa_gx60_testaggr_1_testvol_2'

#    FILESYS_SIZE_FIXED (ARRAYREF) : returns the array of volumes with
#                                      fs_sized_fixed set
#
# Usage:
#       my $hash = parse_upgrade_check(output => $status,
#                                 info =>'VOL_SPACE_INCREASE_BY',
#                                 error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub parse_upgrade_check {
    my %opts = validate(@_,
                        {output	  => {type => SCALAR},
			info	  => {type => SCALAR, optional=>1,
                                      default => 'VOL_INFO_AFTER_UPGRADE'},
                        error_msg   => {type => SCALARREF},});
    
    my ($output, $mode, $info
        , $error_msg, %vol_info, %vol_space, @vol_fs_size_fixed, $flag);
    $error_msg = $opts{error_msg};
    
    $output    = $opts{output};
    if($output =~ /fs_size_fixed/){
        $flag = 1 ;
        $output =~ s/^Warning:/:Warning/;
    }
    $info  = $opts{info};

    # just skip blank lines
    my @split_output = split (":", $output);
    
    if ($info eq "VOL_INFO_AFTER_UPGRADE"){     
        %vol_info = _parse_check_output('input' => $split_output[1] ,);
        return \%vol_info;
    }elsif($info eq "VOL_SPACE_INCREASE_BY"){
        if(defined $split_output[2]){
            %vol_space = _parse_check_output('input' => $split_output[2]);
            $$error_msg = $@ if $@; 
            return \%vol_space;
        }else{
            $$error_msg = $@ if $@; 
            return FC_FALSE;
        }
    }elsif($info eq "FILESYS_SIZE_FIXED"){
        if(defined $split_output[2] && $flag) {
            @vol_fs_size_fixed = _parse_check_output('input' => $split_output[2],
                                                     'flag'  =>  "1");
            $$error_msg = $@ if $@; 
            return \@vol_fs_size_fixed;
        }else{
            $$error_msg = $@ if $@; 
            return FC_FALSE;
        }
    }#End of elsif
    
    
}#End of function parse_upgrade_check
                
################################################################################
# Function:      aggr_upgrade_scan_stop  
#
# Synopsis:      Stop upgrade scanner on an aggr   
#
#   Fc          => Filer connection object
#   aggr_name   => name of aggr(scalar) to stop the upgrade scanner on
#   error_msg   => reference for error message to be captured
#
# Returns:          1 on Success
#		    0 on Failure
#
# Usage:
#       $status = aggr_upgrade_scan_stop(
#                              Fc => $Fc,
#                              aggr_name => $$aggr_list[0],
#                              error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_scan_stop {
  my (%opts) = validate (
        @_,{             
            Fc   =>	{ type => HASHREF},
            aggr_name => { type => SCALAR },
            error_msg   => {type => SCALARREF},});
  
    my $status;
    my $error_msg = $opts{error_msg};
    
    $Log->debug("Stop upgrade scanner on aggr $opts{aggr_name}\n");
    eval {
	$status = sayEOF( $opts{Fc}, "priv set -q diag;
                         aggr 64bit-upgrade scan-stop $opts{aggr_name}");
    };
    #checking the error msg.
    if ($status =~ /Stopp(ed|ing) the 64-bit upgrade scanner on aggregate $opts{aggr_name}/){
        $Log->debug("64-bit upgrade scanner stopped on $opts{aggr_name}");
        $$error_msg = $@ if $@; 
	return FC_TRUE;   
    }
    else {
        $Log->debug("Failed to stop upgrade scanner on aggregate $opts{aggr_name}\n");
        $$error_msg = $@ if $@; 
        return FC_FALSE;    
    }
    
}#End of function aggr_upgrade_scan_stop

################################################################################
# Function:      aggr_upgrade_scan_start  
#
# Synopsis:      Start upgrade scanner on an aggr   
#
#   Fc          => Filer connection object
#   aggr_name   => name of aggr(scalar) to start the upgrade scanner on
#   error_msg   => reference for error message to be captured
#
# Returns:          1 on Success
#		    0 on Failure
#
# Usage:
#       $status = aggr_upgrade_scan_start(
#                              Fc => $Fc,
#                              aggr_name => $$aggr_list[0],
#                              error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_scan_start {
  my (%opts) = validate (
        @_,{             
            Fc   =>	{ type => HASHREF},
            aggr_name => { type => SCALAR },
            error_msg   => {type => SCALARREF},});
  
    my $status;
    my $error_msg = $opts{error_msg};
    
    $Log->debug("Stop upgrade scanner on aggr $opts{aggr_name}\n");
    eval {
	$status = sayEOF( $opts{Fc}, "priv set -q diag; aggr 64bit-upgrade scan-start $opts{aggr_name}");
    };
    #checking the error msg.
    if ($status =~ /Start(ed|ing) the 64-bit upgrade scanner on aggregate $opts{aggr_name}/){
        $Log->debug("64-bit upgrade scanner set to start on $opts{aggr_name}");
        $$error_msg = $@ if $@; 
	return FC_TRUE;   
    }
    else {
        $Log->debug("Failed to start upgrade scanner on aggregate $opts{aggr_name}\n");
        $$error_msg = $@ if $@; 
        return FC_FALSE;    
    }
}#End of function aggr_upgrade_scan_start

################################################################################
# Function:      vol_upgrade_diag  
#
# Synopsis:      Check or start volume upgrade to 64-bit 
#
# InputParameters:
#   Fc          => Filer connection object
#   vol_name   => name of volume (scalar) to be upgraded to 64-bit
#   mode        => mode of upgrade (default normal)
#                  Possible volues:
#                   check   => check to see if aggr can be upgraded or
#                              the space needed for upgrade
#                 normal    => use this option if a user wants to have full
#                              control of the volume sizes for all volumes
#                              in an aggregate.
#   error_msg   => reference for error message to be captured
#
# Returns:      return values are all scalar and are baed on mode:
#                   check   => return raw output of -mode=check command on
#                              success.This output can be fed to
#                              parse_upgrade_check_diag to get specific values
#                              0 for failure
#                             -1 for command failure
#                 normal    => retuns 1 for success,
#                                     0 for failure,
#                                    -1 for command failure
#
# Usage:
#       $status = vol_upgrade_diag(
#                 Fc  => $Fc,
#                 vol_name => $$vol_list[0],
#                 mode => 'check',
#                 error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub vol_upgrade_diag {
  my (%opts) = validate (
        @_,{             
            Fc   =>	{ type => HASHREF},
            vol_name => { type => SCALAR },
            mode => { type => SCALAR , optional=>1, default => 'normal'},
            error_msg   => {type => SCALARREF},});
  
    my $status;
    my $mode_str="-mode $opts{mode}";
    my $error_msg = $opts{error_msg};
    
    $Log->debug("Starting 64bit-upgrade on volume $opts{vol_name} in $opts{mode} mode\n");
    eval {
	$status = sayEOF( $opts{Fc}, "priv set -q diag; vol 64bit-upgrade start $opts{vol_name} $mode_str");
    };
    #checking the error msg.
    if ($status =~ /Started 64-bit upgrade on volume/ && $opts{mode} !~ /check/){
        $Log->debug("64-bit upgrade started on $opts{vol_name}");
        $$error_msg = $@ if $@; 
	return $status;   
    }
    elsif ($opts{mode} =~ /check/ && $status =~ /Checking for additional space/){
        $Log->debug("64-bit upgrade status returned for $opts{vol_name}");
        $$error_msg = $@ if $@; 
	return $status;   
    }
    else{
        $Log->debug("64-bit upgrade on $opts{vol_name} Failed: ".$status);
        $$error_msg = $@ if $@; 
	return FC_ERROR; 
    }
    #more error handling should be added, failure should not only include timeout or system errors
    $$error_msg = $@ if $@; 
    return FC_FALSE;
} # End of vol_upgrade_diag

################################################################################
# Function:      get_upgrade_status  
#^M
# Synopsis:      return the command call of 64bit-upgrade status <aggr> -all 
#^M
# InputParameters:
#   Fc          => Filer connection object
#   aggr_name   => name of aggr (scalar) to be upgraded to 64-bit
#
# Returns:      return value is a string 
#
# Usage:
#       $status = get_upgrade_status(
#                 Fc  => $Fc,
#                 aggr_name => $aggr_name,
#		  );
# Notes:
#   - Global parameters referenced: None
################################################################################
sub get_upgrade_status {
$Log->debug("Parse the aggr-upgrade status with -all options output\n");
    #For the check-in to address burt456611 ,the function will just return the
    #raw-output as is ..But the plan is to parse the output and build a hash and
    #and return only the required data as in CmodeIPU.pm

 my (%opts) = validate (
         @_,{
             Fc   => { type => HASHREF},
             aggr_name => { type => SCALAR },}
             );
     my $status;
     eval {
         $status = sayEOF( $opts{Fc}, "priv set -q diag;
                          aggr 64bit-upgrade status $opts{aggr_name} -all");
     };
    confess "failed to get aggr upagrde status: $@" if ($@);
    return $status;
}

########################################################################
##HELPER FUNCTION
########################################################################
#This function parses the output of check and returns an hash of the table
#or an array of volumenames that has fs_size_fixed option set depending on
#the flag value 
sub _parse_check_output{
    my %opts = validate(@_,
                           {input =>{type => SCALAR},
                            flag  =>{type => SCALAR, optional => 1,default =>0},
                    });
    
    my $input    = $opts{input};
    my $flag     = $opts{flag};
    my %vol_info = ();
    my (@output, @tmp_output, @not_parsed,
        @fields_array, @values, @final_output,
        $fields, @vol_fs_size_fixed);
    my $partial_line = "";
        
    @output = split("\n",$input);
    @output = grep { /\w/ } @output;#remove all blank lines
    if (!scalar(@output) ) {
        return;
    }
    #remove all the messages that are displayed other than the required table
    foreach my $line1 (@output){
	if($line1 !~ /Before triggering 64-bit upgrade|
                    successful upgrade of each volume|
		    ^When triggering 64-bit upgrade|increased automatically|
		    ^Running this command|-+|to 64-bit.$/){
	    push (@not_parsed,$line1);
	}    
    } 
    #the first line will have headers ,get them to $fields ...
    ( $fields, @tmp_output ) = @not_parsed;
    if( defined $fields ) {
        $fields =~ s/Additional Space Required for Upgrade/
                        Additional_Space_Required_for_Upgrade/;
        @fields_array = split("  +", $fields );
    }
    $Log->debug("Fields:\n Dumper(@fields_array)");
  
    #Below foreach code is to make sure that all the elements of the row
    #are on the same line . If  forexample the volumename is long , the row gets
    #displayed in 2 lines. Since we are parsing line by line , will require all
    #elements to be on the same line .    

    foreach my $aline (@tmp_output){  
        _trim(string_ptr=>\$aline);
        @values = split("  +", $aline );
        if ( @values != @fields_array ) {
            $partial_line .= " ".$aline;
            _trim(string_ptr=>\$partial_line);
            my @partial = split(" +", $partial_line );
            next if (@partial != @fields_array);
	    push( @final_output, $partial_line );
	    $partial_line = "";	
	}else {
	    push (@final_output,$aline);
	}         
    }
    my $size = @fields_array;
    my $index = _indexArray(\@fields_array, 'Volume Name');
        
    foreach my $line (@final_output){
        my $i = 0;
        my @row = ( split( " +", $line));

        #we will only populate the hash with "fs_size_fixed" while parsing the
        #second table in the original output . This table has only 2 headers and
        #hence the number 2 below for comparing $size

        if( $size == 2 ){
            if($row[$index] =~ /\(\*\)$/){
                $row[$index] =~ s/\(\*\)$//;
                push(@vol_fs_size_fixed,$row[$index]);
                $vol_info{$row[$index]}{"fs_size_fixed"} = "yes";
            }else{
                $vol_info{$row[$index]}{"fs_size_fixed"} = "no";
            }
	}
        #Now populate the hash 
        while ($i != $size){
            $vol_info{$row[$index]}{$fields_array[$i]} = $row[$i];
            $i++;
        }
    }
    if($flag){
        return (@vol_fs_size_fixed);
    }else{
        return (%vol_info);
    }
}

sub _trim(%) {
    my %args=@_;
    my $item = $args{'string_ptr'};
    $$item =~ s/^\s+//;
    $$item =~ s/\s+$//;
}

#burt460639 - index not returning correctly . making these changes to fix the
#index issue
sub _indexArray
{
    my $array = shift;
    my $search = shift;
    
    my $size = @$array;
    for my $index (0..$size) {
        return $index if $array->[$index] eq $search;
    }
    return -1; # -1 means it's not in array
}

1;
