#: -*-perl-*x
#
# $Id: //depot/prod/test/main/dblade/wafl/LgAggr/CmodeUtils.pm
#
# Copyright (c) 2008 NetAp, Inc.
# All rights reserved.
#
## @summary
##     Library functions for Cmode LA IPU related testcases
##
## @description
##     This library will have functions to invoke all the new UI thats 
##     been added as a part of LA IPU feature and other utilities to
##     test LA-IPU feature 
##
## @usage
##     use IPU::CModeIPU;
##
## @routines
##      create_multiple_aggrs
##      create_multiple_vols
##      cleanup_ipu_setup
##      aggr_upgrade_diag
##      aggr_upgrade_status_diag
##      parse_upgrade_check    
##      aggr_upgrade_scan_start
##      aggr_upgrade_scan_stop
##      get_upgrade_status
##      vol_upgrade_diag
##      vol_upgrade_status_diag
##      volume_is_hybrid
##      volume_is_64bit
##
## @author asha@netapp.com
##
## @status private


package IPU::CModeIPU;

#Define constants

use constant FC_TRUE  => 1;
use constant FC_FALSE => 0;
use constant FC_ERROR => -1;

use constant AGGR_DOES_NOT_EXIST   => 0;
use constant EMPTY_AGGR_EXIST      => 1; 
use constant AGGR_EXIST_NOT_EMPTY  => 2;
use constant AGGR_EXIST_BUT_MODE_DIFFERENT => 3;

use constant VOL_EXIST          => 1;
use constant VOL_DOES_NOT_EXIST => 0;

#Export all other external libraries required

use strict;
use warnings;
use Data::Dumper;

use Client;
use Tharn qw (subtest host $Log);
use Params::Validate qw (validate SCALAR SCALARREF OBJECT ARRAYREF
                         HASHREF);
use SFOTest qw (create_aggregate create_volume delete_volume delete_aggregate);

#Global Variables

use vars (
          '@ISA', '@EXPORT'
);
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(create_multiple_aggrs
                    create_multiple_vols
                    cleanup_ipu_setup
                    aggr_upgrade_diag
                    aggr_upgrade_status_diag
                    parse_upgrade_check    
                    aggr_upgrade_scan_start
                    aggr_upgrade_scan_stop
                    get_upgrade_status
                    vol_upgrade_diag
                    vol_upgrade_status_diag
                    volume_is_hybrid
                    volume_is_64bit          
                    );

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

################################################################################
# Function:      create_multiple_aggrs  (default 32 bit aggrs) 
#
# Synopsis:      Creates multiple  aggrs based on Input Parameters
#			> checks if the aggr with the name already exists 
#			    if its of given block-type and no volume within ,
#                           we re-use
#                           else if purge_existing=1 , delete the existing 
#                           aggr and create new one.
#			> else create new aggr
# InputParameters:
#   serverobj       => Server object
#   filername       => Filername
#   aggr_prefix     => prefix to be added to each aggr (optional ,
#                       default testaggr)
#   aggr_size       => number of disks in an aggr (optional ,default 3)
#   block_type      => 64-bit or 32-bit aggr (optional , default 32)
#                      possible values (64: 64-bit aggr, 32: 32-bit aggr)
#   raid_type       =>  optional ,default to raid_dp 
#   purge_existing  => purge existing aggrs if they exists (optional ,default 1 
#                      i.e purge)possible values (0: no purge, 1: purge)
#   aggr_count      => number of aggrs to be create (optional ,default 1)
#   error_msg       => reference for error message to be captured
#
# Returns:          reference for array of aggr_names that were created
#		    die on failure
#
# Usage:
#	create_multiple_aggrs(
#			serverobj       => $Server,
#			filername       => $filername,
#			aggr_prefix     => $AGGR_NAME,
#			aggr_size       => $AGGR_SIZE,
#			raid_type       => $RAID_TYPE,
#			purge_existing  => 1,
#	    		aggr_count      => $AGGR_COUNT)
# Names of aggr created with default prefix:
#       <filername_prefix_#> => wqa_gx98_testaggr_1 ...
# Notes:
#   - Global parameters referenced: None
################################################################################
sub create_multiple_aggrs {
    my %options = validate(@_,
			   {serverobj       => {type => OBJECT},
			    filername       => {type => SCALAR},
			    aggr_prefix     => {type => SCALAR, optional => 1,
                                                default => "testaggr"},
			    aggr_size       => {type => SCALAR, optional => 1,
                                                default => '3'},
			    block_type      => {type => SCALAR,
                                                optional => 1,
                                                default =>"32-bit"},
			    raid_type       => {type => SCALAR,
                                                optional => 1,
                                                default =>"raid_dp"},
			    purge_existing  => {type => SCALAR, optional => 1,
                                                default => 1},
			    aggr_count      => {type => SCALAR, optional => 1,
                                                default => 1},
                            error_msg       => {type => SCALARREF},});

    my ($server,      $filername,      $aggr_name,
	$count,       $aggr_prefix,    $aggr_count,
	@aggrnames,   $aggr_size,      $raid_type,
	$result,      $create_aggr,    $purge_existing,
        $block_type,  $error_msg,
	$filername_with_underscores,);

    $filername      = $options{filername};
    $server         = $options{serverobj};
    $aggr_prefix    = $options{aggr_prefix};
    $aggr_count     = $options{aggr_count};
    $aggr_size      = $options{aggr_size};    
    $block_type     = $options{block_type};
    $raid_type      = $options{raid_type};
    $error_msg      = $options{error_msg};
    $purge_existing = $options{purge_existing};      
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;


    for ($count = 0; $count < $aggr_count; $count++) {
	$aggr_name = $filername_with_underscores."_".$aggr_prefix."_".$count ;
            $result = _aggr_exists( serverobj   => $server,
                                filername       => $filername,
                                aggr_name       => $aggr_name,
                                "block-type"    => $block_type,
                                error_msg       => $error_msg) ; 
            die ("_aggr_exists failed , error : $$error_msg") if ($result == FC_ERROR);
        
	if($result == EMPTY_AGGR_EXIST){
	    $Log->trace("Empty $block_type $aggr_name exists..Reusing");
	    $create_aggr = 0;
	    push(@aggrnames, $aggr_name);
	} elsif ($result == AGGR_EXIST_NOT_EMPTY ||
                    $result == AGGR_EXIST_BUT_MODE_DIFFERENT){
	    if($purge_existing){
		$Log->trace("$aggr_name exists..but doesn't match our criteria.."
                                                            . "Delete it");
	            cleanup_ipu_setup(serverobj     => $server,
					 filername  => $filername,
					 aggr_name  => $aggr_name,
                                         error_msg  => $error_msg) or 
                    die "Failed to delete $block_type $aggr_name containing "
                                . "volumes error : $$error_msg ";
		    
	        $Log->trace("Deleted the existing aggr ..now create it ");
	        $create_aggr = 1;	  
	    }else{	
		$Log->trace("Aggr already exists but not supposed to purge ..");
                $$error_msg = " Aggr exists but mode different and cant purge ";
		die "Aggr already exists but not supposed to purge ".
                                                        "error :$$error_msg ";
	    }  
	} elsif ($result == AGGR_DOES_NOT_EXIST){
	    $Log->trace("$block_type $aggr_name do not exists..create it ");
	    $create_aggr = 1;
	}#End of if($result) loop
	if($create_aggr){
	    $Log->trace("Creating new $block_type aggr $aggr_name");
            eval{
                create_aggregate(serverobj      => $server,
                                    filername   => $filername,
                                    aggr_name   => $aggr_name,
                                    aggr_size   => $aggr_size,
                                    "block_type"=> $block_type,
                                    raid_type   => $raid_type,);
            };
            if ($@){
                $$error_msg = $@ ;
                die "Create Aggregate $aggr_name FAILED ,error : $$error_msg" ; 
            }
	    push(@aggrnames, $aggr_name);
	}
    }#End of for loop

    return (\@aggrnames);
}#End of create_multiple_aggrs
################################################################################
# Function:        create_multiple_vols  
#
# Synopsis:      Creates multiple volume on each aggr based on Input Parameters
#
# InputParameters:
#   serverobj       => Server object
#   filername       => Filername
#   vol_prefix      => prefix to be added to each vol (default testvol)
#   aggr_names      => one aggr name (scalar) or aggr_list (arrayref) to create
#                       volumes on.
#   vol_size        => size of each volume (default 20g)
#   vserver_name    => name of the vserver the volume belongs to
#   vol_count       => number of volumes to be create (default 1)
#   purge_existing  => purge existing vols if they exists (default 1 i.e purge)
#                      possible values (0: no purge, 1: purge)
#   space_guaratee  => space guarantee of the volume (default  "volume")
#   error_msg       => reference for error message to be captured
#
# Returns:          reference for array of volume names that were created
#		    die on failure
# Usage:
# create_multiple_vols(serverobj    => $Server,
#                       filername   => $filername,
#		       aggr_names   => \@aggr_list,
#		       vol_prefix   => $VOL_NAME,
#		       vserver_name => $VSERVER,
#		       vol_size     => $VOL_SIZE,
#                    purge_existing => 1,
#		       vol_count    => $VOL_COUNT) 
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub create_multiple_vols {
    my %options = validate(@_,
			   {serverobj       => {type => OBJECT},
			    filername       => {type => SCALAR},
			    vol_prefix      => {type => SCALAR, optional => 1,
                                                default => "testvol"},
			    aggr_names      => {type => SCALAR | ARRAYREF},
			    vol_size        => {type => SCALAR, optional => 1,
                                                default => "20g"},
			    vserver_name    => {type => SCALAR},
			    vol_count       => {type => SCALAR, optional => 1,
                                                default => "1"},
			    space_guarantee => {type => SCALAR, optional => 1,
						default => "volume"},
                            purge_existing  => {type => SCALAR, optional => 1,
                                                default => "1"},});
    #space_guarantee is defaulted to "volume" since this is the system default. 
    #SFOTest.pm defaults it to none and hence making changes here to make sure
    #the system default is used #burt460639.
    my ($server,      $filername,     $vol_name,
	$count,       $vol_prefix,    $vol_count,
	$aggr_names,  @aggrnames,     $vol_size,
	$vserver_name, @volnames,     $purge_existing,
	$space_guarantee);

    $filername     = $options{filername};
    $server        = $options{serverobj};
    $vol_prefix    = $options{vol_prefix};
    $vol_count     = $options{vol_count};
    $vol_size      = $options{vol_size};
    $aggr_names    = $options{aggr_names};
    $vserver_name  = $options{vserver_name};
    $purge_existing= $options{purge_existing};
    $space_guarantee=$options{space_guarantee}; 
    if(ref($aggr_names) eq "ARRAY") {
	push(@aggrnames,@$aggr_names);
    } else {
        push(@aggrnames,$aggr_names);
    }

    foreach my $aggr_name (@aggrnames) {
        for ($count = 0; $count < $vol_count; $count++) {
            $vol_name = $aggr_name."_".$vol_prefix."_".$count;
	    #check volume exist 
	    if(!_vol_exists(
                            serverobj    => $server,
                            filername    => $filername, 
                            volume       => $vol_name)){
    	           die "Failed creating volume $vol_name"
		   if(create_volume(
                                serverobj    => $server,
                   	        filername    => $filername,
				vserver_name => $vserver_name,
                            	volume       => $vol_name,
                            	volume_size  => $vol_size,
			    space_guarantee  => $space_guarantee,
                            	aggr_name    => $aggr_name));
                   push(@volnames, $vol_name);
	    }else{
                if($purge_existing){
                    die "Volume deletion - $vol_name  FAILED"
                    if(delete_volume(
                                serverobj    => $server,
                                filername    => $filername,
                                volume       => $vol_name,
                                vserver_name => $vserver_name));
                    
                    die "Failed creating volume $vol_name"
                    if(create_volume(
                                serverobj    => $server,
                   	        filername    => $filername,
				vserver_name => $vserver_name,
                            	volume       => $vol_name,
                            	volume_size  => $vol_size,
			    space_guarantee  => $space_guarantee,
                            	aggr_name    => $aggr_name));
                    push(@volnames, $vol_name);
                }else{
                    die "volume $vol_name already exists but cant purge" ;
                }
            }
        }#End of for loop
    }#End of foreach loop  
    return \@volnames;  
}#End of function create_multiple_vols

################################################################################
# Function:        cleanup_ipu_setup
#
# Synopsis:     Delete a given number of vols on each of the aggregate 
#		   mentioned and the aggr itself
# InputParameters:  
#   serverobj   => Server object
#   filername   => Filername
#   aggr_name   => name of aggr (scalar) to be upgraded to 64-bit		   
#
# Returns:          die on failure
#		    1 on success
# Notes:
#   - Global parameters referenced: None
################################################################################
sub cleanup_ipu_setup {
    my %options = validate(@_,
			   {serverobj     => {type => OBJECT},
			    filername     => {type => SCALAR},
			    aggr_name     => {type => SCALAR},
                            error_msg     => {type => SCALARREF},});
    my ($server,      $filername,      $aggr_name,
	@vols,        $error_msg);
    $filername   = $options{filername};
    $server      = $options{serverobj};
    $aggr_name   = $options{aggr_name};
    $error_msg   = $options{error_msg};
    
    $server->set_execution_server(-Server => $filername);
    eval {
        @vols = $server->volume_show(aggregate=>$aggr_name);
    };
    if(!$@) {
        $Log->trace("Volume exists in the aggr $aggr_name..Deleting them");
        foreach my $vol_of (@vols) {
            my $volume_name  = $vol_of->{volume};
            my $vserver_name = $vol_of->{vserver};
            
            if(delete_volume(serverobj   => $server,
                            filername    => $filername,
                            volume       => $volume_name,
                            vserver_name => $vserver_name)){
                $$error_msg = $@;
                die "Volume deletion - $volume_name  FAILED . ";
            }
        }#End of foreach
    }#end of if    
    if(delete_aggregate(serverobj    => $server,
                        filername    => $filername,
                        aggr_name    => $aggr_name)){
            $$error_msg = $@;
            die "Could not delete aggr ,Aggregate $aggr_name still  exists ...";
    }
    return FC_TRUE;
}#End of function cleanup_ipu_setup

################################################################################
# Function:         aggr_upgrade_diag 
# Synopsis:         Executes the command "storage aggregate 64bit-upgrade start"
#                with specified options and returns the output
#
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   aggr_name  => name of aggr (scalar) to be upgraded to 64-bit
#   mode       => mode of upgrade (default normal)
#                  Possible volues:
#                   check   => check to see if aggr can be upgraded or
#                              the space needed for upgrade
#                 normal    => use this option if a user wants to have full
#                              control of the volume sizes for all volumes
#                              in an aggregate.
#               grow-all    => Use this option if a user wants to ensure the
#                               completion of format upgrade for the aggregate
#                               and all its contained volumes without
#                               fs_size_fixed option set) after "aggr add".
#            grow-reserved  => Use this option if a user wants to maintain the
#                               volumes sizes for existing non-required fixed
#                               sized volumes in an aggregate, except in cases
#                               when additional space must be added to volumes
#                               without the fs_size_fixed option set to
#                               accommodate the upgrade of the existing space
#                               reserved files. 
#   error_msg   => reference for error message to be captured
#
# Returns:      return values are all scalar and are based on mode:
#                   check   => return raw output of -mode=check command on
#                              success.This output can be fed to
#                              parse_upgrade_check to get specific values
#                              0 for failure
#                             -1 for command failure
#                 normal,grow-all,grow-reserved
#                           => retuns command output for success,
#                                     0 for failure,
#                                     -1 for command failure
#
# Usage:
#       $status = aggr_upgrade_diag(
#                 serverobj => $Server,
#                 filername => $NODES[0],
#                 aggr_name => $$aggr_list[0],
#                 mode      => 'check',
#                 error_msg => \$error_msg);
# Notes:
#   - Global parameters referenced: None
########################################################################
sub aggr_upgrade_diag {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR},
    			    mode	  => {type => SCALAR,
                                              optional => 1 ,
                                              default => "normal"},
			    error_msg     => {type => SCALARREF},});
    my ($server,     $filername,       $aggr_name,
        $mode,       $command_output,
	$error_msg);
    $filername  = $options{filername};
    $server     = $options{serverobj};
    $aggr_name  = $options{aggr_name};
    $mode       = $options{mode};
    $error_msg  = $options{error_msg};
    $$error_msg = "none";
    
    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
	my $mode_str = "-mode ".$mode ;
	$server->storage_aggregate_64bit_upgrade_start(
                                aggregate => $aggr_name,
		    		mode      => $mode,
                                output    => \$command_output);
    };
    if($@){
        $$error_msg = "aggr_upgrade_diag() failed, error :$@";
    }

    if ($command_output =~ /Start(ed|ing) 64-bit upgrade on aggregate/ &&
                            $mode !~ /check/) {
	$Log->trace("64-bit upgrade started on $aggr_name");
        return ($command_output);
    }elsif($mode =~ /check/ &&
                $command_output =~ /Checking for additional space/){
        $Log->trace("64-bit upgrade check status returned for $aggr_name");
        return ($command_output);
    }else{
        $Log->trace("Error while executing command ".$@);
	return FC_ERROR;
    }
    return FC_FALSE; 
}#End of function aggr_upgrade_diag
################################################################################
# Function:      vol_upgrade_diag  
#
# Synopsis:      Check or start volume upgrade to 64-bit 
#
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   vol_name   => name of volume (scalar) to be upgraded to 64-bit
#   mode       => mode of upgrade (optional,default normal)
#                  Possible volues:
#                   check   => check to see if aggr can be upgraded or
#                              the space needed for upgrade
#                   normal  => use this option if a user wants to have full
#                              control of the volume sizes for all volumes
#                              in an aggregate.
#   error_msg   => reference for error message to be captured
#
# Returns:      return values are all scalar and are based on mode:
#                 check   => return raw output of -mode=check command on
#                              success.This output can be fed to
#                              parse_upgrade_check to get specific values
#                    
#                             -1 for command failure
#                 normal  => returns 1 for success
#                                    -1 for command failure
#
# Usage:
#       $status = vol_upgrade_diag(
#                 serverobj => $Server,
#                 filername => $NODES[0],
#                 vol_name  => $$vol_list[0],
#                 mode      => 'check',
#                 error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub vol_upgrade_diag {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
						    vserver       => {type => SCALAR},
                            filername     => {type => SCALAR},
                            vol_name      => {type => SCALAR},
    			    mode	  => {type => SCALAR,
                                              optional => 1 ,
                                              default => "normal"},
			    error_msg     => {type => SCALARREF},});
    my ($server, $filername, $vol_name, $mode, $command_output,
	$error_msg,$vserver);
    $filername  = $options{filername};
	$vserver    = $options{vserver};
    $server     = $options{serverobj};
    $vol_name   = $options{vol_name};
    $mode       = $options{mode};
    $error_msg  = $options{error_msg};
    $$error_msg = "none";
    
    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
	my $mode_str = "-mode ".$mode ;
	$server->volume_64bit_upgrade_start(
                                volume    => $vol_name,
								vserver   => $vserver,
		    		mode      => $mode,
                                output    => \$command_output);
    };
    if ($command_output =~ /Start(ed|ing) 64-bit upgrade on volume/
                        && $mode !~ /check/) {
        $Log->trace("64-bit upgrade started on $vol_name");
	return $command_output;   	
    }elsif ($mode =~ /check/ &&
                $command_output =~ /Checking for additional space/){
        $Log->trace("64-bit upgrade status returned for $vol_name");
        return $command_output;
    }else{
        $Log->trace("Error while executing command ".$@);
	$$error_msg = "vol_upgrade_diag() failed, error :$@";
	return FC_ERROR;
    }
    $$error_msg = $@ if $@; 
    return FC_FALSE;
}#End of function vol_upgrade_diag
################################################################################
# Function:         parse_upgrade_check 
# Synopsis:      Parses the output from aggr_upgrade_diag or aggr_upgrade
#                with mode=check
#
# InputParameters:
#   output  => output from -mode check of aggr/vol 64bit-upgrade -mode=check
#                  or aggr add -64bit-upgrade check commands 
#   info    => type of information required
#              Posssible values:
#              VOL_INFO_AFTER_UPGRADE (default): gives the list(hash) of volumes
#                                      that will be upgraded and their
#                                      usage after upgrade
#              VOL_SPACE_INCREASE_BY : gives the hash of volumes which need
#                                       more space before upgrade
#              FILESYS_SIZE_FIXED   : gives the array of volumes with
#                                      filesys-size-fixed set and space required 
#                                      for upgrade to complete.
#   error_msg   => reference for error message to be captured
#
# Returns:  return hash ref or array ref or scalar based on info
#    VOL_INFO_AFTER_UPGRADE (HASHREF): example
#          'Used'       => '204.9MB',
#          'Total'      => '1GB',
#          'Volume'     => 'wqa_gx19_testaggr_0testvol1',
#          'Capacity'   => '20%',
#          'Available'  => '819.1MB',
#          'Vserver'    => 'vs0'
#    VOL_SPACE_INCREASE_BY (HASHREF): example
#          'Volume'        => 'wqa_gx19_testaggr_0testvol0',
#          'fs_size_fixed' => 'yes',
#          'Additional_Space_Required_for_Upgrade' => '1.09MB',
#          'Vserver'       => 'vs0'
#    FILESYS_SIZE_FIXED (ARRAYREF) : returns the array of volumes with
#                                      fs_sized_fixed set
#
# Usage:
#       my $hash = parse_upgrade_check(output => $status,
#                                 info        =>'VOL_SPACE_INCREASE_BY',
#                                 error_msg   => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub parse_upgrade_check {
    my %options = validate(@_,
                           {output	  => {type => SCALAR},
			    info	  => {type => SCALAR},});			    
    my ($output,            $mode,      $required, 
	$error_msg,         %vol_info,  %vol_space,
	@vol_fs_size_fixed, $flag );
    
    $output    = $options{output};
    $required  = $options{info};

    if($output =~ /filesys-size-fixed/){
        $flag = 1 ;
        $output =~ s/^Warning:/:Warning/;
    }
    # just skip blank lines
    my @split_output = split (":", $output);
    if ($required eq "VOL_INFO_AFTER_UPGRADE"){     
        %vol_info = _parse_check_output('input' => $split_output[1]);
        return \%vol_info;
    }elsif($required eq "VOL_SPACE_INCREASE_BY"){
        if(defined $split_output[2]){
            %vol_space = _parse_check_output('input' => $split_output[2]);
            return \%vol_space;
        }else{
            return FC_FALSE;
        }
    }elsif($required eq "FILESYS_SIZE_FIXED"){
        if(defined $split_output[2] && $flag) {
            @vol_fs_size_fixed = _parse_check_output('input' => $split_output[2],
                                                     'flag'  =>  "1");
            return \@vol_fs_size_fixed;
        }else{
            return FC_FALSE;
        }
    }#End of elsif 
}#End of function parse_upgrade_check

################################################################################
# Function:         aggr_upgrade_scan_start 
#
# Synopsis:      Start upgrade scanner on an aggr   
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   aggr_name  => name of aggr(scalar) on which scanner to be started
#   error_msg  => reference for error message to be captu/_parse_check_outputred
# Returns:          1 on Success
#		    0 on failure
#
# Usage:
#       $status = aggr_upgrade_scan_start(
#                              serverobj => $Server,
#			       filername => $NODES[0],
#			       aggr_name => $AggrNames[0],
#                              error_msg => \$error_msg);
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_scan_start {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR},
			    error_msg     => {type => SCALARREF},});
    my ($server, $filername, $aggr_name,
        $command_output,
	$error_msg);
    
    $filername = $options{filername};
    $server    = $options{serverobj};
    $aggr_name = $options{aggr_name};
    $error_msg = $options{error_msg};
    $$error_msg = "none";
    
    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
        $server->storage_aggregate_64bit_upgrade_scan_start(
                                            aggregate => $aggr_name,
                                            output    => \$command_output);
    };
    if ($command_output =~
        /Start(ed|ing) the 64-bit upgrade scanner on aggregate $aggr_name/) {
	$Log->trace("64-bit upgrade scanner stopped on $aggr_name");
	return FC_TRUE;
    } else {
        $Log->trace("Failed to start upgrade scanner on aggregate $aggr_name");
        $$error_msg = $@;
        return FC_FALSE;
    }
}#End of function aggr_upgrade_scan_start
################################################################################
# Function:         aggr_upgrade_scan_stop 
#
# Synopsis:      Stop upgrade scanner on an aggr   
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   aggr_name  => name of aggr(scalar) on which scanner to be stopped
#   error_msg  => reference for error message to be captured
# Returns:          1 on Success
#		    0 on failure
#
# Usage:
#       $status = aggr_upgrade_scan_stop(
#                              serverobj  => $Server,
#			        filername => $NODES[0],
#			        aggr_name => $AggrNames[0],
#                              error_msg  => \$error_msg);
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_scan_stop {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR},
			    error_msg     => {type => SCALARREF},});
    my ($server,         $filername, $aggr_name,
	$command_output, $error_msg);
    
    $filername = $options{filername};
    $server    = $options{serverobj};
    $aggr_name = $options{aggr_name};
    $error_msg = $options{error_msg};
    $$error_msg = "none";
    
    #Set the current server
    $server->set_execution_server(-Server => $filername);
    eval {
	    $server->storage_aggregate_64bit_upgrade_scan_stop(
                                aggregate => $aggr_name,
                                output    => \$command_output);
    };
    
    if ($command_output =~
        /Stopp(ed|ing) the 64-bit upgrade scanner on aggregate $aggr_name/) {
	$Log->trace("64-bit upgrade scanner stopped on $aggr_name");
	return FC_TRUE;
    } else {
        $Log->trace("Failed to stop upgrade scanner on aggregate $aggr_name");
        $$error_msg = $@;
        return FC_FALSE;
    }
}#End of function aggr_upgrade_scan_stop
################################################################################
# Function:         aggr_upgrade_status_diag
#
# Synopsis:          Results from aggr 64bit-upgrade status command 
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   aggr_name  => name of aggr(scalar) on which scanner to be stopped
#   error_msg  => reference for error message to be captured
# Returns:          reference to parsed command output on Success
#		    0 on failure
#
# Usage:
#       $status = aggr_upgrade_status_diag(
#                              serverobj  => $Server,
#				filername => $NODES[0],
#			        aggr_name => $AggrNames[0],
#                              error_msg  => \$error_msg);
# Notes:
#   - Global parameters referenced: None
################################################################################
sub aggr_upgrade_status_diag {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR},
			    error_msg     => {type => SCALARREF},});
    my ($server, $filername, $aggr_name,
	@status_messages,
	$error_msg);
    $filername = $options{filername};
    $server    = $options{serverobj};
    $aggr_name = $options{aggr_name};
    $error_msg = $options{error_msg};

    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
	@status_messages =
                $server->storage_aggregate_64bit_upgrade_status(
                                            aggregate => $aggr_name,
							);
    };
    $$error_msg = $@;
    if ($$error_msg =~
        /64-bit upgrade is not currently in progress on aggregate $aggr_name/) {
	$Log->trace("64-bit upgrade is not running ".$$error_msg);
	return FC_FALSE;
    }
    return \@status_messages;  
}#End of function aggr_upgrade_status_diag
################################################################################
# Function:         vol_upgrade_status_diag
#
# Synopsis:          Results from vol 64bit-upgrade status command 
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   vol_name   => name of vol(scalar) on which scanner to be stopped
#   error_msg  => reference for error message to be captured
# Returns:          reference to parsed command output on Success
#		    0 on failure
#
# Usage:
#       $status = aggr_upgrade_status_diag(
#                              serverobj  => $Server,
#				filername => $NODES[0],
#			        vol_name  => $VolNames[0],
#                              error_msg  => \$error_msg);
# Notes:
#   - Global parameters referenced: None
################################################################################
sub vol_upgrade_status_diag {
    my %options = validate(@_,
                           {serverobj   => {type => OBJECT},
                            filername   => {type => SCALAR},
                            vol_name    => {type => SCALAR},
			    error_msg   => {type => SCALARREF},});
    my ($server, $filername, $vol_name,
	@status_messages,
	$error_msg);
    $filername = $options{filername};
    $server    = $options{serverobj};
    $vol_name = $options{vol_name};
    $error_msg = $options{error_msg};

    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
	@status_messages =
                $server->volume_64bit_upgrade_status(volume => $vol_name,
							);
    };
    if ($@ =~
        /64-bit upgrade is not currently in progress on volume $vol_name/) {
	$Log->trace("64-bit upgrade is not running ".$@);	
	return FC_FALSE;
    }
    return \@status_messages;
    
}#End of function vol_upgrade_status_diag
################################################################################
# Function:         get_upgrade_status
#
# Synopsis:         Results from aggr 64bit-upgrade status command 
#                   
# InputParameters:
#   serverobj   => Server object
#   filername   => Filername
#   aggr_name   => name of aggr(scalar) on which scanner to be stopped
#   object_name => agregate/volume name for which the status is required(optional)
#    field      => (optional)field can be
#                        'block-format'
#                        'scanner-status'
#                        'aggregate'
#                        'volume'
#                        'vserver'
#                        'scan-time-to-complete'
#                        'scan-percent-completed'
#                        'scan-progress'
#   value       => (optional)the value ,  script is expecting for a given field
#                       for example , if $field = block-format
#                                       $value can be /64-bit|32-bit|upgrading|.
#   error_msg   => reference for error message to be captured
# Returns:
#       if (object_name ,field and value) are all undefined ,
#           return the reference to parsed command output
#       if (object_name ,field )are undefined and value is defined
#           return FC_ERROR , since this is not the correct usage.
#       if(object_name) is defined
#           check if both ($field and $value) are defined
#               if yes , return FC_TRUE if $field=$value , FC_FALSE otherwise
#               Usage:
#                        my $status = get_upgrade_status(serverobj => $Server,
#				filername => $NODES[0],
#				aggr_name => $AggrNames[0],
#                                object_name => "$AggrNames[0]",
#                                field     => "scanner-status",
#                                value     => "running",
#                                error_msg => \$error_msg);  
#           check if $field is defined and $value not defined
#               if yes , return value of $field.
#               Usage:
#                        my $status = get_upgrade_status(serverobj => $Server,
#				filername => $NODES[0],
#				aggr_name => $AggrNames[0],
#                                object_name => "$AggrNames[0]",
#                                field     => "scanner-status",
#                                error_msg => \$error_msg);  
#           else return an hash reference containing named-value pair for the
#                       object specified.
#               Usage:
#                        my $status = get_upgrade_status(serverobj => $Server,
#				filername => $NODES[0],
#				aggr_name => $AggrNames[0],
#                                object_name => "$AggrNames[0]",
#                                error_msg => \$error_msg);  
#       if (object_name is not defined )
#           but defined ($field and $value)
#              loop through all the elements of the output and
#              return FC_TRUE if $field=$value , FC_FALSE otherwise
#               Usage:
#                       my $status = get_upgrade_status(serverobj => $Server,
#				filername => $NODES[0],
#				aggr_name => $AggrNames[0],
#                                field     => "scanner-status",
#                                value     => "running",
#                                error_msg => \$error_msg);
#       if (object_name and $value not defined)
#           but defined $field
#              return a reference to a hash containing the fields value for all
#              the objects i.e(aggr and all the volumes within it )
#               Usage:
#                        my $status = get_upgrade_status(serverobj => $Server,
#				filername => $NODES[0],
#				aggr_name => $AggrNames[0],
#                                field     => "scanner-status",
#                                error_msg => \$error_msg);
#
# Notes:
#   - Global parameters referenced: None
################################################################################
sub get_upgrade_status {
	my %options = validate(@_,
			 	{
				serverobj     => {type => OBJECT},
				filername     => {type => SCALAR},
				aggr_name     => {type => SCALAR},
				field 	      => {type => SCALAR, optional => 1,
                                                   default => undef},
				object_name   => {type => SCALAR, optional => 1,
                                                   default => undef},	
				value	      => {type => SCALAR, optional => 1,
                                                   default => undef},
				error_msg     => {type => SCALARREF},
				 });
	my ($server,        $filername,     $aggr_name,
	    $object_name,   $field,         $value,
            $error_msg,     $is_aggr,       $obj,
            %ret_val,       @status_details );

	$filername 	= $options{filername};
        $server    	= $options{serverobj};
        $aggr_name 	= $options{aggr_name};
	$error_msg      = $options{error_msg};
        if (defined($options{field})){
	    $field= $options{field};
	}else {
            undef ($field);
        }
        if (defined($options{object_name})){
	    $object_name = $options{object_name};
	}else {
            undef $object_name;
        }
	if (defined($options{value})){
	    $value = $options{value};
	}else{
            undef $value ;
        }
        my $command_output;
        $is_aggr = 0;
        $obj     = "volume";
	if(defined $object_name && ($object_name eq $aggr_name)){
            $is_aggr = 1 ;
            $obj     = "aggregate";
        }
        if(!$is_aggr && (defined $object_name)){
            $Log->trace("Check if the volume exists in the aggregate");
            eval{
                $server->volume_show(volume    => $object_name,
                                     aggregate => $aggr_name );
            };
            if($@ =~ /no entries matching your query/){
                $Log->trace("Volume $object_name do not exists in $aggr_name".$@);
                $$error_msg = "get_upgrade_status() failed, error :$@";
                return FC_ERROR;
            }
        }
	my $i = 0;
	eval{
	     @status_details =
             $server->storage_aggregate_64bit_upgrade_status(
                                            aggregate => $aggr_name,
	 			'include-all-volumes' => "true");
	};
        
        $Log->trace("Printing output of upgrade status :\n" .
                                    Dumper(@status_details) );
	if($@){
	    $Log->trace("Error while getting status ".$@);
	    $$error_msg = "get_upgrade_status() failed, error :$@";
	    return FC_ERROR;
	}
        if(!defined($object_name) && !defined($value) && !defined($field)){
            return (\@status_details);
        }elsif(!defined($object_name) && defined($value) && !defined($field)){
            $$error_msg = "usage incorrect ..please specify field or objectname";
            return FC_ERROR;
        }
	
	while(defined($status_details[$i])) {
            if(defined $object_name){
                if($status_details[$i]->{$obj} eq $object_name){
                    if(defined $field && defined $value){
                        if($status_details[$i]->{$field} ne $value){
                            return FC_FALSE ;
                        }else{
                            return FC_TRUE ;
                        }
                    }elsif(defined $field && !defined $value){
                        return ($status_details[$i]->{$field}) ;
                    }else {
                        $Log->trace("Printing status_details[$i]:\n" .
                                    Dumper($status_details[$i]) );
                        return(\$status_details[$i]);      
                    }
                }  
            }elsif(defined $value && defined $field) {
                return FC_FALSE if($status_details[$i]->{$field} ne $value);
            }elsif(defined $field && !defined $value){
                if($status_details[$i]->{volume} =~ /-/){
                    $ret_val{$status_details[$i]->{aggregate}} =
                                        $status_details[$i]->{$field};
                }else{
                    $ret_val{$status_details[$i]->{volume}} =
                                        $status_details[$i]->{$field};
                }
            }
	    $i++ ;
        }
        if(!defined $object_name && defined $field && !defined $value){
            $Log->trace("Printing ret_val:\n" .Dumper(%ret_val)) ;
            return \%ret_val;
        }
        return FC_TRUE; 
}#End of function get_upgrade_status
################################################################################
# Function:         volume_is_hybrid
#
# Synopsis:         Returns true if volume is hybrid, false otherwise
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   vol_name   => name of vol(scalar) on which scanner to be stopped
#   vserver    => vserver name
#   error_msg  => reference for error message to be captured
# Returns:          1 if volume is hybrid
#		    0 if not
#
# Usage:
#       $status = volume_is_hybrid(
#                              serverobj => $Server,
#			       filername => $NODES[0],
#			        vol_name => $VolNames[0],
#                                vserver => "vs0",
#                              error_msg => \$error_msg);
# Notes:
#   - Global parameters referenced: None
################################################################################
sub volume_is_hybrid{
        my %options = validate(@_,
                           {serverobj   => {type => OBJECT},
                            filername   => {type => SCALAR},
                            volume      => {type => SCALAR},
                            vserver     => {type => SCALAR, },   
			    error_msg   => {type => SCALARREF},});
    my ($server, $filername, $vol_name,
	@status_messages, $vserver,
	$error_msg);
    $filername = $options{filername};
    $server    = $options{serverobj};
    $vol_name  = $options{volume};
    $vserver   = $options{vserver};
    $error_msg = $options{error_msg};

    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
	@status_messages =
                $server->volume_64bit_upgrade_status(volume => $vol_name,
                                                    vserver => $vserver
							);
    };
    if(defined($status_messages[0])){
        if($status_messages[0]->{"block-format"} =~ /upgrading/){
            return FC_TRUE;
        }else{
            return FC_FALSE;
        }
    }
    $$error_msg = $@;
    return FC_ERROR;
}#End of func volume_is_hybrid
################################################################################
# Function:         volume_is_64bit
#
# Synopsis:         Returns true if volume is 64bit, false otherwise
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   vol_name   => name of vol(scalar) on which scanner to be stopped
#   vserver    => vserver name
#   error_msg  => reference for error message to be captured
# Returns:          1 if volume is 64bit
#		    0 if not
#                   -1 if error
#
# Usage:
#       $status = volume_is_64bit(
#                              serverobj => $Server,
#			       filername => $NODES[0],
#			        vol_name => $VolNames[0],
#                               vserver  => "vs0",
#                              error_msg => \$error_msg);
# Notes:
#   - Global parameters referenced: None
################################################################################
sub volume_is_64bit{
        my %options = validate(@_,
                           {serverobj   => {type => OBJECT},
                            filername   => {type => SCALAR},
                            volume      => {type => SCALAR},
                            vserver     => {type => SCALAR, },   
			    error_msg   => {type => SCALARREF},});
    my ($server,          $filername, $vol_name,
	@status_messages, $vserver,
	$error_msg);
    $filername = $options{filername};
    $server    = $options{serverobj};
    $vol_name  = $options{volume};
    $vserver   = $options{vserver};
    $error_msg = $options{error_msg};

    #Set the current server
    $server->set_execution_server(-Server => $filername);

    eval {
	@status_messages =
                $server->volume_show(volume => $vol_name,
                                    vserver => $vserver
							);
    };
    if (defined($status_messages[0])){
        if($status_messages[0]->{"block-type"} =~ /64-bit/){
            return FC_TRUE;
        }else{
            return FC_FALSE;
        }
    }
    $$error_msg = $@;
    return FC_ERROR;
}#End of func volume_is_64bit
################################################################################
##HELPER FUNCTION
################################################################################
#This function parses the output of check and returns an hash of the table
#or an array of volumenames that has fs_size_fixed option set depending on
#the flag value 
sub _parse_check_output{
    my %options = validate(@_,
                           {input	  => {type => SCALAR},
                            flag          => {type => SCALAR,
                                              optional => 1 ,
                                              default => 0 },
                            });
    my $input           = $options{input};
    my $flag            = $options{flag};
    my %vol_info        = ();
    my $partial_line    = "";
    my (@output,       @tmp_output,        @not_parsed,
        @fields_array, @values,            @final_output,
        $fields,       @vol_fs_size_fixed);
         
    @output = split("\n", $input);
    @output = grep { /\w/ } @output;#remove all blank lines
    if (!scalar(@output) ) {
        return;
    }
    #remove all the messages that are displayed other than the required table
    foreach my $line1 (@output){
	if($line1 !~ /Before triggering 64-bit upgrade|
                    successful upgrade of these volumes|
                    ^When triggering 64-bit upgrade|increased automatically|
                    ^Running this command|-+|to 64-bit.$/){
	    push (@not_parsed, $line1);
	}    
    }
    $Log->trace("Fields:\n" .Dumper(@not_parsed) );
    
    #the first line will have headers ..get them to $fields ...
    ($fields, @tmp_output ) = @not_parsed;
    if( defined $fields ) {
        $fields =~ s/Additional Space Required for Upgrade/Additional_Space_Required_for_Upgrade/;
        @fields_array = split("  +", $fields );
    }
    $Log->trace("Fields:\n" .Dumper(@fields_array) );
    
    my $search = "Volume";
    my ($index) = grep { $fields_array[$_] eq $search } 0..$#fields_array;
    $Log->trace("index = $index ") ;

    #Below foreach code is to make sure that all the elements of the row
    #are on the same line . If  forexample the volumename is long , the row gets
    #displayed in 2 lines. Since we are parsing line by line , will require all
    #elements to be on the same line .
    
    foreach my $aline (@tmp_output){  
        _trim(string_ptr => \$aline);
        @values = split("  +", $aline );
        if ( @values != @fields_array ) {
            $partial_line .= "  ".$aline;
            _trim(string_ptr => \$partial_line);
            my @partial = split("  +", $partial_line );
            next if (@partial != @fields_array);
	    push(@final_output, $partial_line );
	    $partial_line = "";	
	}else {
	    push (@final_output, $aline);
	}         
    }
    $Log->trace("Fields:\n" .Dumper(@final_output) );
    
    my $size = @fields_array;
    foreach my $line (@final_output){
        my $i = 0;
        my @row = split(" +", $line);
        
        #we will only populate the hash with "fs_size_fixed" while parsing the
        #second table in the original output . This table has only 3 headers and
        #hence the number 3 below for comparing $size
        
        if($size == 3){
            if($row[$index] =~ /\(\*\)$/){
                $row[$index] =~ s/\(\*\)$//;
                push(@vol_fs_size_fixed, $row[1]);
                $vol_info{$row[$index]}{"fs_size_fixed"} = "yes";
            }else{
                $vol_info{$row[$index]}{"fs_size_fixed"} = "no";
            }
        }
        #Now populate the hash 
        while ($i != $size){
            $vol_info{$row[$index]}{$fields_array[$i]} = $row[$i];
            $i++;
        }
    }      
    $Log->trace("Parsed output is:\n" . Dumper(%vol_info));
       
    if($flag){
        $Log->trace("Parsed array for FILESYS_SIZE_FIXED is \n" .
                                Dumper(@vol_fs_size_fixed));
        return (@vol_fs_size_fixed);
    }else{
        return (%vol_info);
    }
}

################################################################################
# Function:        _aggr_exists 
#
# Synopsis:      Check if the aggr of specified name and block-type exists
#                and doesnot contain any volumes within 
#                   
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   aggr_name  => name of aggr(scalar) on which scanner to be started
#   block-type => reference for error message to be captured
#
# Returns:          EMPTY_AGGR_EXIST if aggr exists and is empty
#		    AGGR_EXIST_NOT_EMPTY if aggr exists and is not empty
#		    AGGR_DOES_NOT_EXIST if aggr does not exists
#                   AGGR_EXIST_BUT_MODE_DIFFERENT if aggr exist but of different
#                                               format
################################################################################
sub _aggr_exists {
    my %options = validate(@_,
			   {serverobj     => {type => OBJECT},
			    filername     => {type => SCALAR},
			    aggr_name     => {type => SCALAR},
			    "block-type"  => {type => SCALAR},
                            error_msg     => {type => SCALARREF},});
    my ($server,      $filername,      $aggr_name,
	$block_type,  $error_msg,
        @aggrs);
    $filername   = $options{filername};
    $server      = $options{serverobj};
    $aggr_name   = $options{aggr_name};
    $block_type  = $options{"block-type"};
    $error_msg  = $options{error_msg};
    $server->set_execution_server(-Server => $filername);
    eval {
        @aggrs = $server->storage_aggregate_show(aggregate=>$aggr_name);
    };
    if($@ =~ /no entries matching your query/){
        return AGGR_DOES_NOT_EXIST;
    } elsif (defined($aggrs[0])) {
        if($aggrs[0]->{"block-type"} ne $block_type){
            return AGGR_EXIST_BUT_MODE_DIFFERENT;
        }
        if($aggrs[0]->{volcount} ne "0"){
            return AGGR_EXIST_NOT_EMPTY;
        }else{
            return EMPTY_AGGR_EXIST;
        }   
    } else{
        $$error_msg = "storage_aggregate_show() failed, error :$@";
	return FC_ERROR;
    }
}#End of function _aggr_exists 

################################################################################
# Function:        _vol_exists 
#
# Description:     Check if the volume exists
#
# InputParameters:
#   serverobj  => Server object
#   filername  => Filername (required to set_execution_server)
#   vol_name   => name of vol(scalar) on which scanner to be started
# Returns:          VOL_EXIST if vol exists 
#		    VOL_DOES_NOT_EXIST if vol does not exists
################################################################################
sub _vol_exists {
    my %options = validate(@_,
			   {serverobj     => {type => OBJECT},
			    filername     => {type => SCALAR},
			    volume        => {type => SCALAR},
                            });
    my ($server,      $filername,      $vol_name);
    $filername   = $options{filername};
    $server      = $options{serverobj};
    $vol_name    = $options{volume};

    $server->set_execution_server(-Server => $filername);
    eval {
         $server->volume_show(volume=>$vol_name);
    };
    if($@ =~ /no entries matching your query/){
        return VOL_DOES_NOT_EXIST;
    }
    return VOL_EXIST;
}#End of function _vol_exists

sub _trim(%) {
    my %args=@_;
    my $item = $args{'string_ptr'};
    $$item =~ s/^\s+//;
    $$item =~ s/\s+$//;
}#End of function _trim

##end_of_file
1;
