#Hey emacs, this is basically -*- perl -*-
#
# $Id: //depot/prod/test/main/lib/IPU/IPU_Dataset.pm
#
# Copyright (c) 2010 NetAp, Inc.
# All rights reserved.
#
## @summary
##     Library functions generation of Datasets
##
## @description
##     This library will have functions to generate a variety of predefined
##     datasets as well as the capability to allow callers to define their
##     own.
##
## @usage
##     use IPU::IPU_Datasets;
##
## @routines
##      create_dataset_1_2_3
##      create_mixed_dataset_2_3
##      create_dataset
##      create_snapshot
##      start_IPU
##      get_buftree
##      get_inode
##      get_filenames
##
## @author nestorf@netapp.com
##
## @status private

=head1 IPU_Dataset
IPU_Dataset - Creates datasets and snapshots for use in automated NetApp 
in-place-upgrade (IPU) testing.

=head1 VERSION

    DRAFT: Please refer to section on BUGS AND LIMITATIONS

    This documentation refers to IPU_Dataset version 0.0.5

=head1 SYNOPSIS

    $ipu_obj = IPU_Dataset->new(filer   => $filer_name,
                                volume  => $volume_name,
                                client  => $client_name);

    $ipu_obj->create_dataset_1_2_3();
    $ipu_obj->create_mixed_dataset_2_3();

=head2 For users who want more fine grained control, lower level APIs are called:

    $ipu_obj = IPU_Dataset->new(filer  => $filer_name,
                                volume => $volume_name,
                                client => $client_name);
    
    $ipu_obj->create_dataset([{spec => $dataset1_spec,
                               name => "Name of dataset1",
                               dir  => "dataset1",
                               inplace => 0},
                              {spec => $dataset2_spec,
                               name => "Name of dataset2",
                               dir  => "dataset2",
                               inplace => 0}
                             ]);

    $ipu_obj->create_snapshot(snapshot => "snap_1");
    $ipu_obj->set_wafl_64bit_upgrade_scan(disable => 1);
    $ipu_obj->start_IPU();
    
    $ipu_obj->create_dataset([{spec => $mixed_dataset_spec,
                               name => "Name of mixed dataset2",
                               dir  => "dataset1",
                               inplace => 1}
                             ]);
    $ipu_obj->set_wafl_64bit_upgrade_scan(disable => 0);

    $ipu_obj->create_snapshot(snapshot => "snap_2");


=head1 DESCRIPTION

This module is used to create I<"interesting"> datasets which are intended to be
used for different types of testing such as in-place-upgrade (IPU) testing and 
volume commands.  

For a definition of the files in each dataset, see:
     L<http://wikid.netapp.com/w/WAFL/Larger_Aggregates/Mile3Tasks/TestPlan/IPU_dataset>.

What makes these datasets interesting is that the file characteristics are 
designed to force certain scenarios that are part of IPU.  For example, some of the
file sizes are such that they would force a new level created when going from 32-bit
to 64-bit volumes.  Another example, are files that are very holey and files that
are space reserved files.

The are two types of datasets, fixed and mixed.  The fixed datasets are created
on 32-bit volumes.  The mixed datasets are created by taking the fixed datasets,
turning on IPU and then performing operations on the files.  
is left off entirely. 


=head2 FIXED DATASETS

There are several predefined fixed datasets that B<must> be created before 
mixed datasets can be created.  These fixed datasets are listed below: 

=over 4

=item * Dataset_32_1: A collection of files of levels ranging from 0 to 3.  A mix of 
these are holey and space reserved.

=item * Dataset_32_2: Two files, one of size 4096 * 510 * 510 * 8 =  8323200k
and the other a space reserved file of size 4096 * 510 * 510 * 10 = 10404000k

=item * Dataset_32_3: A collection of 15 files; 5 ranging in size from 0 to 64bytes,
5 ranging in size from 65 to 32768 bytes, and 5 ranging in size from 32769 to 
2601000.

=back

=head2 MIXED DATASETS

Mixed datasets are those datasets which are derived from the fixed 
datasets or other mixed datasets.  The mixed datasets are created by turning 
on in-place-update (via the I<vol 64bit_upgrade> command and then modifying 
the existing datasets through calls to commands such as I<wafl touch> and 
I<wafl punch>.

The mixed datasets are listed below:

=over 4

=item * Dataset_Mixed_2: The files from Dataset_32_1 are touched via the I<wafl touch> 
command at various block numbers. 

=item * Dataset_Mixed_3: Some of the files from Dataset_Mixed_2 are then written-to 
(via the I<testsetattr -W> command) at different lengths and offsets.

=item * Dataset_Mixed_4: The command I<sio 0 random 0 end-of-file> is performed on each
file in Dataset_32_2.

=item * Dataset_Mixed_5: Random level 1 blocks are touched in each file of Dataset_32_3.

=back

=head1 BUGS AND LIMITATIONS

This is a work in progress.  There are some major caveats/limitations
listed below which will be address in a subsequent version.

=over 4

=item * Files that required allocation via fsetattr -a are omitted
for now due to requirement that the file remain open the life of the
test script.

=item * Error handling/exception handling has not been incorporated
yet.  That is, no checks are made to determine if a particular
commands such as mkfile succeeded.

=item * Assumes sufficient space in the volume to accomodate the dataset.
An option to vol_grow will be provided shortly.

=back

=cut

package IPU::Dataset;
use strict;
use warnings;

use Exporter qw(import);
our @EXPORT = qw/create_dataset_1_2_3
                 create_dataset_4
                 create_dataset_5
                 create_mixed_dataset_2_3
                 create_mixed_dataset_4
                 create_mixed_dataset_5
                 create_dataset
                 create_snapshot
                 start_IPU
                 wafl_upgrade_64bit_upgrade_scan
                 get_inode
                 get_buftree
                 get_filenames/;

use Tharn qw($Log);
use NACL::C::Client;
use NACL::C::Node;
use NACL::C::Snapshot;
use NACL::C::Interface;
use NACL::C::Client::Mount;
use NACL::APISet;
use NACL::C::Volume64bitUpgrade;
use NACL::STask::Volume;
use IPU::NACL_IPU;
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);
use Data::Dumper;


#
# Import Exceptions
#
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try); 
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

#forward declare the subroutines that get called from the dataset
sub _remake_dir;
sub _mkfile;
sub _testsetattr_s;
sub _testsetattr_W;
sub _testsetattr_f;
sub _wafl_punch;
sub _wafl_touch;
sub _filersio;

#this defines the dataset in terms of operations, in order, on files
#the fields are:
    #filename: the name of the file as a scalar to perform the operation
    #cmd     : reference to the subroutine that to call to
    #          that will check for optional parms and form the command
    #          string
    #parm    : reference to a hash of name/arguement pairs for the 
    #          respective command
our $_Dataset_32_1 = [
                {filename => "file_l0_1", 
                    cmd => \&_mkfile,           parms => {size => 0}},

                 {filename => "file_l0_2",
                    cmd => \&_mkfile,           parms => {size => 0}},
                 {filename => "file_l0_2",
                    cmd => \&_testsetattr_W,    parms => {offset => 0,
                                                          len    => 64}},

                 {filename => "file_l1_1",
                    cmd => \&_mkfile,           parms => {size => "8k"}},

                 {filename => "file_l1_2",
                    cmd => \&_mkfile,           parms => {size => "32k"}},

                 #{filename => "file_l1_sprs_3_past_eof",
                   #cmd => \&_mkfile,        parms => {size => "32k"}},
                 #{filename => "file_l1_sprs_3_past_eof",
                   #cmd => \&fsetattr_s,    parms => {size => "40k"}},

                 {filename => "file_l2_holey_1",
                    cmd => \&_mkfile,        parms => {size => 0}},
                 {filename => "file_l2_holey_1",
                    cmd => \&_testsetattr_s, parms => {size => "2040k"}},
                 {filename => "file_l2_sprs",
                    cmd => \&_mkfile,        parms => {size => "4080k"}},
                 {filename => "file_l2_sprs",
                    cmd => \&_testsetattr_f, parms => {enabled => "enable"}},
                 {filename => "file_l2_holey_sprs",
                    cmd => \&_mkfile,        parms => {size => "4800k"}},
                 {filename => "file_l2_holey_sprs",
                    cmd => \&_testsetattr_f, parms => {enabled => "enable"}},
                 {filename => "file_l2_holey_sprs",
                    cmd => \&_wafl_punch,    parms => {fbn   => 500,
                                                       count => 530}},
                 {filename => "file_l2_holey_sprs",
                    cmd => \&_wafl_punch,    parms => {fbn   => 1150,
                                                       count => 50}},
#TODO: get from Cindy how she wants to handle allocations i.e.fsetattr
                #"file_l2_holey_sprs_past_eof",
                #"file_l2_holey_sprs_past_eof_2",
                #"file_l2_holey_sprs_past_eof_3",
                 {filename => "file_l3_holey_sprs_1",
                    cmd => \&_mkfile,        parms => {size => 0}},
                 {filename => "file_l3_holey_sprs_1",
                    cmd => \&_testsetattr_s, parms => {size => "816000k"}},

                #was unsparse, needed to be mkfile as unsparse is being deprecated
                 {filename => "file_l3_holey_sprs_2",
                    cmd => \&_mkfile,      parms => {timeout => 1200, size => "816000k"}},
                 {filename => "file_l3_holey_sprs_2",
                    cmd => \&_wafl_punch,    parms => {fbn   => 203000,
                                                       count => 1000}},
                #was unsparse, needed to be mkfile as unsparse is being deprecated
                 {filename => "file_l3_sprs",
                    cmd => \&_mkfile,        parms => {timeout => 1200, size => "20400k"}},
                 {filename => "file_l3_sprs",
                    cmd => \&_testsetattr_f, parms => {enabled => "enable"}},

                 {filename => "file_l3_sprs_2",
                    cmd => \&_mkfile,        parms => {timeout => 1200, size => "2080800k"}},
                 {filename => "file_l3_sprs_2",
                    cmd => \&_testsetattr_f, parms => {enabled => "enable"}},
            ];

our $_Dataset_32_2 = [
                 {filename => "a_large_file",
                    cmd => \&_mkfile,        parms => {timeout => 1200, size => "8323200k"}},

                 {filename => "a_large_sprv_file",
                    cmd => \&_mkfile,        parms => {timeout => 1200, size => "10404000k"}},
                 {filename => "a_large_sprv_file",
                    cmd => \&_testsetattr_f, parms => {enabled => "enable"}},
            ];

our $_Dataset_32_3 = [
                {filename => "size_0",
                    cmd => \&_mkfile,        parms => {size => 0}},
                {filename => "size_4",
                    cmd => \&_mkfile,        parms => {size => 4}},
                {filename => "size_8",
                    cmd => \&_mkfile,        parms => {size => 8}},
                {filename => "size_32",
                    cmd => \&_mkfile,        parms => {size => 32}},
                {filename => "size_64",
                    cmd => \&_mkfile,        parms => {size => 64}},
                {filename => "size_65",
                    cmd => \&_mkfile,        parms => {size => 65}},
                {filename => "size_129",
                    cmd => \&_mkfile,        parms => {size => 129}},
                {filename => "size_256",
                    cmd => \&_mkfile,        parms => {size => 256}},
                {filename => "size_16k",
                    cmd => \&_mkfile,        parms => {size => "16k"}},
                {filename => "size_32k",
                    cmd => \&_mkfile,        parms => {size => "32k"}},
                {filename => "size_32769",
                    cmd => \&_mkfile,        parms => {size => 32769}},
                {filename => "size_33000",
                    cmd => \&_mkfile,        parms => {size => 33000}},
                {filename => "size_50000",
                    cmd => \&_mkfile,        parms => {size => 50000}},
                {filename => "size_2601000",
                    cmd => \&_mkfile,        parms => {size => 2601000}},
            ];

our $_Dataset_Mixed_2 = [
                {filename => "file_l1_1",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 0,
						                                level     => 0}},

                {filename => "file_l1_2",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 4,
						                                level     => 0}},

                {filename => "file_l2_holey_1",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 0,
						                                level     => 1}},

                {filename => "file_l2_sprs",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 100,
						                                level     => 0}},

                {filename => "file_l2_holey_sprs",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 600,
						                                level     => 1}},
                {filename => "file_l2_holey_sprs",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 1150,
						                                level     => 1}},


                #{filename => "file_l2_holey_sprs_past_eof",
                    #cmd => \&_wafl_touch,     parms => {fbn_start => 100},
                                                      #level => 1}},

                {filename => "file_l3_holey_sprs_1",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 1000,
                                                        level     => 1}},
                {filename => "file_l3_holey_sprs_2",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 203000,
                                                        level     => 1}},

                {filename => "file_l3_sprs",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 0,
                                                        level     => 2}},

                {filename => "file_l3_sprs_2",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 512,
                                                        level     => 1}},
                {filename => "file_l3_sprs_2",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 255000,
                                                        fbn_end   => 255100,
                                                        level     => 1}},
            ];

our $_Dataset_Mixed_3 = [
		{filename => "file_l0_1",
		 cmd => \&_testsetattr_s, parms => {size => "16320k"}},

        #{filename => "file_l1_sprs_3_past_eof",
		# cmd => \&_testsetattr_W, parms => {offset => 73729,
		#				   len => 1}},
        #{filename => "file_l2_holey_sprs_past_eof",
		# cmd => \&_testsetattr_W, parms => {offset => 8192001,
		#				   len => 1}},
        #{filename => "file_l2_holey_sprs_past_eof_2",
        # cmd => \&_testsetattr_W, parms => {offset => 9011300, len => 7000}},
        #{filename => "file_l2_holey_sprs_past_eof_3",
        # cmd => \&_testsetattr_W, # parms => {offset => 9011300, len => 1024000}},

        {filename => "file_l3_holey_sprs_1", 
         cmd => \&_testsetattr_W, parms => {offset => 4091904,
                                            len    => 12288}},
        {filename => "file_l3_holey_sprs_2",
         cmd => \&_testsetattr_W, parms => {offset => 830668800,
                                            len    => 2048000}},
        {filename => "file_l3_sprs",
         cmd => \&_testsetattr_W, parms => {offset => 266342401,
                                            len    => 1}},
    ];
        
my $_Dataset_Mixed_4 = [
                 {filename => "a_large_file",
                    cmd => \&_filersio,
                         parms => { read_percent => 0, #100% writes
                                    rand_percent => 100,
                                    start_byte   => 0,
                                    end_byte     => "8323200k",
                                    blk_size     => 4096,
                                    run_time     => 2,     #2 seconds
                                    num_active   => 1,
                                  }
                 },
                 {filename => "a_large_sprv_file",
                    cmd => \&_filersio,,     
                         parms => { read_percent => 0, #100% writes
                                    rand_percent => 100,
                                    start_byte   => 0,
                                    end_byte     => "10404000k",
                                    blk_size     => 4096,
                                    run_time     => 2, #2 seconds
                                    num_active   => 1,
                                  }
                },
            ];

my $_Dataset_Mixed_5 = [

                {filename => "size_65",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_129",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_256",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_16k",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_32k",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_32769",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_33000",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_50000",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
                {filename => "size_2601000",
                    cmd => \&_wafl_touch, parms => { fbn_start => 0,
                                                     level     => 1}},
            ];

# the following dataset is used mainly for IPU iron scenario tests:
# "file_l3_root_single_l2" has size of 20m (20971520 bytes.)
#         This file has 1 single L2 indirect block, underneath the root L3 buffer.
#         This file will serve the purpose of having VBN placeholder on L2,
#             in the middle of an indirect.
# "file_l3_with_2_full_l2":
#         size: 2130739200 (~2G) bytes.
#         This file has 2 full L2 indirects underneath the root L3 buffer.
#         This file will serve the purpose of having VBN placeholder on L3,
#             in the middle of an indirect.
# "file_l4_with_70_l2":
#         size: 18207000 L0 blocks (size: 74575872000 bytes.)
#         This file will serve the purpose of having VBN placeholder
# "file_l4_with_l2_with_index_lt_254":
#         size: 20889600 L0 blocks (85563801600 bytes)
#         Purpose: last vbn in L2 is a placeholder and its index is < 254.
#                  This has 1 L2 with 10 indirects pointing to L1 blocks. (???)
our $_Dataset_32_iron1 = [
                {filename => "file_l3_root_single_l2", 
                    cmd => \&_mkfile,           parms => {size => 0}},
                {filename => "file_l3_root_single_l2", 
                    cmd => \&_testsetattr_W,    parms => {offset => 0,
                                                          len    => 20971520 }},
            ];

our $Dataset_Mixed_iron1 = [
                {filename => "file_l3_root_single_l2",
                    cmd => \&_wafl_touch,     parms => {fbn_start => 0,
						                                level     => 2}},
            ];

our $_Dataset_32_iron2 = [
                {filename => "file_l3_with_2_full_l2", 
                    cmd => \&_mkfile,           parms => {size => 0}},
                {filename => "file_l3_with_2_full_l2", 
                    cmd => \&_testsetattr_W,    parms => {offset => 0,
                                                          len    => 2130739200 }},
            ];

## Dataset_32_iron2 has 2 full L2 branches,
## first L2 has fbn range from 0 to 260099 (== 510 * 510 -1)
## this operation attempts to upgrade the 2nd L2 branch.
# Dataset_Mixed_iron2 would upgrade fbn 260100 (= 510 * 510 )
## because fbn starts with 0 (zero.)
our $Dataset_Mixed_iron2 = [
    {filename => "file_l3_with_2_full_l2",
     cmd      => \&_wafl_touch,
     parms    => {fbn_start => 260100,
     level    => 2}
    },
    ];


## this set of data requires a long time to create...
our $_Dataset_32_iron3 = [
                {filename => "file_l4_with_70_l2", 
                    cmd => \&_filersio,
                         parms => { read_percent => 0, #100% writes
                                    rand_percent => 100,
                                    start_byte   => 0,
                                    end_byte     => "72828000k",
                                    blk_size     => 4096,
                                    run_time     => 10,     #10 seconds
                                    num_active   => 1,
                                  }
                 },
            ];

our $_Dataset_32_iron4 = [
                {filename => "file_l4_with_l2_with_index_lt_254", 
                    cmd => \&_filersio,
                         parms => { read_percent => 0, #100% writes
                                    rand_percent => 100,
                                    start_byte   => 0,
                                    end_byte     => "81600m",
                                    blk_size     => 4096,
                                    run_time     => 10,     #10 seconds
                                    num_active   => 1,
                                  }
                 },
            ];

#constructor for IPU dataset object.  Creates the necessary handles in order
#to communicate with the node and the host.  Also, creates the data structures
#used to store information about each dataset and each volume on which the datasets
#will be created.

=head1 METHODS

=head2 new()

To create an IPU_Dataset object used to manage the datasets:

        $ipu_obj = new(filer  => $filer_name, 
                       client => $client_name, 
                       volume => $volume_name);

$volume_name may be a scalar or an array reference containing a list
of volumes.
    i.e. $volume = "volume_name";
         -or-
         $volume = ["vol1","vol2","vol3"]; 

If $volume is an arrayref with more than one volume name, then 
subsequent lower-level API calls may require a volume name as a parameter.
See below.

Note that all volumes passed into the constructor must already exist.

The $filer and $volume_name are used to determine where to target the filer 
commands used to create the datasets.  

The $client_name is used to NFS mount the volume in order to create the 
directories where the datasets are created (see below).  Calling 
$ipu_obj->DESTROY() will unmount the volume.  

=cut

sub new {
    my $pkg = shift;
    my (%opts) = validate(@_,
                          {filer  => {type     => SCALAR},
                           volume => {type     => SCALAR | ARRAYREF},
                           client => {type     => SCALAR},
                          });
    my $client_host_obj;
    my $node_host_obj;
    my $node_hosttype;
    my $cmd;
    my $node_api_set_obj;
    my $client_api_set_obj;
    my $nacl_ipu_obj;

    my $self = {};
    bless $self, $pkg;

    #copy the client name 
    $self->{client} = $opts{client};

    #store the filer string as passed in 
    #also create a hostrec with the hostspec information and store that as well.
    $self->{filer}  = $opts{filer};
    $node_host_obj = main::host($self->{filer});
    $node_hosttype = $node_host_obj->hosttype();

    #if volume is a scalar then convert to an array ref with a single 
    #element so we can loop over it later when initializing the volume_data
    if (! ref $opts{volume} ) {
        $opts{volume} = [ $opts{volume} ];
    }
    $self->{volume_list} = $opts{volume};

    #set up the _dataset_data structure here each dataset here
    my %empty_hash = ();
    $self->{_dataset_data} = \%empty_hash;


    #get an APISet object against the client, it should be a Linux.
    $client_host_obj = main::host($self->{client});
    $client_api_set_obj = NACL::APISet->new(hostobj   => $client_host_obj,
                                            category  => "Host",
                                            interface => "CLI",
                                            set       => "Linux");
    $self->{host_api_set_obj} = $client_api_set_obj;

    #get the APISet object for the node for the majority of the dataset operations
    #that will be requested (e.g. mkfile, testsetattr, etc).  as part of creating the
    #APISet object, need to determine if the filer is 7mode or c-mode


    if ($node_hosttype =~ /filer/) {
        $self->{shell_type} = ($node_hosttype=~/ng/)? "Nodescope" : "7Mode";
        $Log->comment("shell type is set to: $self->{shell_type}");
    } else {
        NATE::BaseException->throw("Invalid filer hosttype, $node_hosttype");
    }
    
    #TODO: NOTE: for c-mode, all calls to execute_raw_command requite that that 
    #the command be prefaced with " system node run -node local -command "
    #this is a bug and is reported in BURT 485948.

    #get the node object for the filer and use that to get the 
    #Api_Set_Obj for issuing commands against the filer
    my $node_obj = NACL::C::Node->find();
    $self->{node_obj} = $node_obj;

    $node_api_set_obj = NACL::APISet->new(hostobj   => $node_host_obj,
                                          category  => "Node",
                                          interface => "CLI",
                                          set       => $self->{shell_type});
    $self->{node_api_set_obj} = $node_api_set_obj;

    #set up the _volume_data structure, this is a 2-dimensional
    #hash that contains the mount object and mount point on the host
    #and the path on the node

    #get the NACL IPU object for use later (wait for BURT 485259)
    $Log->comment("Get the NACL_IPU object");
    $nacl_ipu_obj = IPU::NACL_IPU->new(filer => $self->{filer},
                                       node => $node_obj,
                                       api_set_obj => $node_api_set_obj);
    $self->{nacl_ipu_obj} = $nacl_ipu_obj;

    #init the volume data structure
    #fix burt613330: make client usage consistent
    my $client = NACL::C::Client->new( hostrec => $client_host_obj );

    #for each volume, make sure the volume is 32-bit, throw if not.
    #exportfs the file as rw,anon=0 and then mount the file system on the 
    #host and use the returned mount object to get the mount point.  
    #also store the path on the node to the volume.  this is stored in 
    #$self->{_volume_data}{$volume_name} as a hash reference
    foreach my $volume_name (@ { $self->{volume_list} } ) {
        my $mount;
        my $export_path;
        my $export_node_name;
    
        #make sure the volume is 32-bit
        $Log->trace("check that volume $volume_name is 32bit and not " .
                        "already scanning");
        if (! $self->_volume_check_32bit($volume_name) ) {
            $Log->trace("Is not 32-bit");
        }
        else {
            $Log->trace("Is a 32-bit");
        }

        #will store the volume data while being created
        #the variable %volume_data is used for readability
        my %volume_data = (volume_name => $volume_name);
        $self->{_volume_data}{$volume_name} = \%volume_data;

        #exportfs the file system so can create dirs and files on it later and
        #mount the filer on the system and create an APISet object to the client
        #to be used when mkdir'ing.  store the APISet in the IPU_Dataset object

        $volume_data{volume_node_path} = $self->_make_node_path(\%volume_data);

        $Log->trace("attempt to exportfs $self->{filer} as " .
            $volume_data{volume_node_path});
        $self->_exportfs_volume($volume_data{volume_node_path});

        $Log->trace("attempt to mount $self->{filer} on " .
                                $volume_data{volume_node_path});

        if ($self->{shell_type} eq "Nodescope") {
            $Log->trace("attempt to get vserver and junction");
            $self->_get_vserver_junction(\%volume_data);

            $export_path = $volume_data{volume_junction};

            $Log->trace("attempt to get cmode export node name");
            $export_node_name = _get_cmode_export_node_name(
                                            $node_obj,
                                            \%volume_data);
        }
        else {
            $export_path = $volume_data{volume_node_path};
            $export_node_name = $node_host_obj->{default_ip};
        }

        $mount = NACL::C::Client::Mount->create(command_interface => $client,
                                                server            => $export_node_name,
                                                path              => $export_path);

        $Log->trace("mount point is $mount->{mount_point} ");
        $volume_data{volume_host_mount} = $mount;
        $volume_data{volume_host_mount_point} = $mount->{mount_point};
    }

    return $self;
}

sub DESTROY {
    my ($self) = @_;

    local $@; #avoid overwriting test script's $@
    $Log->enter();

    foreach my $volume_name (@ { $self->{volume_list} } ) {
        my $volume_data = $self->{_volume_data}{$volume_name};
        my $mount_obj = $volume_data->{volume_host_mount};

        #umount the volume if it had been mounted
        if ($mount_obj) {
            my $volume_mount_point = 
                        $volume_data->{volume_host_mount_point};
            $Log->trace("in DESTROY, will umount " .
                                    $volume_mount_point);

            #deleting the mount object will umount the file system
            #if the umount fails because the volume isn't mounted,
            #log a trace message but don't throw exception.
            try {
                $mount_obj->delete();
                $Log->trace("umount of $volume_mount_point complete");
            }
            catch NACL::APISet::Exceptions::CommandFailedException with {
                if ($@ =~ /not mounted/) {
                    $Log->trace("umount is skipped, " .
                                    $volume_mount_point .
                                    "no longer mounted");
                }
                else {
                    $Log->trace("umount throws unexpected exception");
                    $@->throw();
                }
            };
        }
        else {
            $Log->trace("umount skipped, $volume_name was never mounted");
        }
    }
    $Log->exit();
}

=head2 $ipu_obj->create_dataset_1_2_3()

To create the fixed datasets
        
    $ipu_obj->create_dataset_1_2_3(dir_prefix => $dir);

The optional dir_prefix parameter directs the subroutine under which top level 
directory the dataset will be created.  The default is "IPU".

Once the $ipu_obj is created by calling I<new()>, the first three fixed datasets 
are created by calling I<create_dataset_1_2_3()>.  Note that the files are 
created in the directories dataset1, dataset2, and dataset3, respectively.  

Another copy of the datasets can be created in the 
directories dataset1.bak, dataset2.bak, and dataset3.bak.   These 
directories, if they exist, are deleted and recreated before the 
datasets are created.

By default, the backup copies are NOT created.  To override the 
default and have the backup copies created, pass the option 
backup_create option as below:

    $ipu_obj->create_dataset_1_2_3(dir_prefix => $dir,
                                   backup_create => 1);

By default, a snapshot called I<base_snap> is created.  If the 
snapshot already exists, then that snapshot is deleted before being 
recreated.  To override the snapshot name, pass in the snapshot name 
in the option snapshot_name_1.

To override the default and NOT have the snapshot created, pass the 
option snapshot_create as below:
    $ipu_obj->create_dataset_1_2_3(dir_prefix => $dir,
                                   snapshot_create => 0);

=cut 

sub create_dataset_1_2_3 {
    my ($self) = shift;
    my %opts = validate(@_,
                        {dir_prefix      => {type    => SCALAR,
                                             default => "IPU"},
                         backup_create   => {type    => SCALAR,
                                             default => 0},
                         snapshot_create => {type    => SCALAR,
                                             default => 1},
                         snapshot_name_1 => {type    => SCALAR,
                                             default => "base_snap"},
                        }
                       );

    #list of datasets to create, first three fixed datasets
    my $datasets_to_create = [{spec    => $_Dataset_32_1,
                               name    => "Dataset_32_1",
                               dir     => "dataset1",
                               inplace => 0},

                              {spec    => $_Dataset_32_2,
                               name    => "Dataset_32_2",
                               dir     => "dataset2",
                               inplace => 0},

                              {spec    => $_Dataset_32_3,
                               name    => "Dataset_32_3",
                               dir     => "dataset3",
                               inplace => 0}];

    #list of datasets to create, first three fixed datasets
    #this time, however, create back ups
    my $bak_datasets_to_create = [{spec    => $_Dataset_32_1,
                                   name    => "Dataset_32_1 bak",
                                   dir     => "dataset1.bak",
                                   inplace => 0},

                                  {spec    => $_Dataset_32_2,
                                   name    => "Dataset_32_2 bak",
                                   dir     => "dataset2.bak",
                                   inplace => 0},

                                  {spec    => $_Dataset_32_3,
                                   name    => "Dataset_32_3 bak",
                                   dir     => "dataset3.bak",
                                   inplace => 0}];
                              

    #create the snapshot.  if $opt{backup_create} then make a second copy
    #using the .bak for the backup copy
    #if $opt{snapshot_create}, then make a snapshot
    $self->create_dataset(dir_prefix => $opts{dir_prefix},
                          dataset_spec => $datasets_to_create);
    if ($opts{backup_create}) {
        $self->create_dataset(dir_prefix => $opts{dir_prefix},
                              dataset_spec => $bak_datasets_to_create);
    }
    if ($opts{snapshot_create}) {
        $self->create_snapshot(snapshot => $opts{snapshot_name_1});
    }
}

=head2 $ipu_obj->create_mixed_dataset_2_3()

To create the mixed dataset 1, 2, and 3

    $ipu_obj->create_mixed_dataset_2_3(dir_prefix => $dir);

The optional dir_prefix parameter directs the subroutine under 
which top level directory the dataset will be created.  The 
default is "IPU".

Once the fixed datasets have been created, the mixed sets are created 
by calling I<create_mixed_dataset_2_3()>.  The mixed dataset 
Dataset_Mixed_2 will be created by performing the operations after
IPU scanning has been disabled and turning the 64 bit upgrade via
the I<vol 64bit_upgrade> command.  Note that the file operations are 
performed in-place (i.e. in the dataset1 directory).

Once Dataset_Mixed_2 has been completed, a snapshot, mixed_snap2 will
be created.  Note that if this snapshot already exists, it is deleted 
before being recreated.  The snapshot name can be overridden by passing
in the option snapshot_name_2.

If you wish for snapshot *NOT* to be taken, pass in the option 
snapshot_create as shown below:

    $ipu_obj->create_mixed_dataset_2_3(dir_prefix => $dir,
                                         snapshot_create => 0);

The mixed dataset Dataset_Mixed_3 is then created by performing a
prescribed set of operations on the Dataset_Mixed_2.  These file 
operations are also performed in-place (i.e. in the dataset1 
directory).

Once Dataset_Mixed_3 has been completed, a snapshot, mixed_snap3 will
be created.  Note that if this snapshot already exists, it is deleted 
before being recreated.   The snapshot name can be overridden by
passing in the option snapshot_name_3.

Again, use the snapshot_create => 0 option to supress the creation
of a snapshot.

=cut

sub create_mixed_dataset_2_3 {
    my ($self) = shift; 
    my %opts = validate(@_,
                        {dir_prefix      => {type    => SCALAR,
                                             default => "IPU"},
                         snapshot_create => {type    => SCALAR,
                                             default => 1},
                         snapshot_name_2 => {type    => SCALAR,
                                             default => "mixed_snap2"},
                         snapshot_name_3 => {type    => SCALAR,
                                             default => "mixed_snap3"}},
			     );

    #list of datasets to create.  since these are mixed datasets,
    #need to make them "inplace".  
    my $mixed_dataset_2 = [{spec    => $_Dataset_Mixed_2,
                            name    => "Dataset_Mixed_2",
                            dir     => "dataset1",
                            inplace => 1}
                          ];

    my $mixed_dataset_3 = [{spec    => $_Dataset_Mixed_3,
                            name    => "Dataset_Mixed_3",
                            dir     => "dataset1",
                            inplace => 1},
                          ];

    #start the in place upgrade and then create the mixed datasets
    #you want to create them separately so you can create snapshots
    #in between them
    $self->start_IPU(verify => 0);

    $self->create_dataset(dir_prefix   => $opts{dir_prefix},
                          dataset_spec => $mixed_dataset_2);

    if ($opts{snapshot_create}) {
        $self->create_snapshot(snapshot => $opts{snapshot_name_2});
    }

    $self->create_dataset(dir_prefix   => $opts{dir_prefix},
                          dataset_spec => $mixed_dataset_3);

    if ($opts{snapshot_create}) {
        $self->create_snapshot(snapshot => $opts{snapshot_name_3});
    }
} 

=head2 $ipu_obj->create_mixed_dataset_4()

To create the mixed dataset 4
    $ipu_obj->create_mixed_dataset_4()

The optional dir_prefix parameter directs the subroutine under 
which top level directory the dataset will be created.  The default 
is "IPU".

File operations are performed in-place on the fixed dataset 
Dataset_32_2 in the directory dataset2.

By default, a snapshot named mixed_snap4 is created after the
dataset is created.  To override the default behavior and NOT
take the snapshot, use the snapshot_create option as below:

    $ipu_obj->create_mixed_dataset_4(snapshot_create => 0);

=cut

sub create_mixed_dataset_4 {
    my ($self) = shift; 
    my %opts = validate(@_,
                        {dir_prefix      => {type    => SCALAR,
                                             default => "IPU"},
                         snapshot_create => {type    => SCALAR,
                                             default => 1},
                         snapshot_name_4 => {type    => SCALAR,
                                             default => "mixed_snap4"}}
		       );

    my $mixed_dataset_4 = [{spec    => $_Dataset_Mixed_4,
                            name    => "Dataset_Mixed_4",
                            dir     => "dataset2",
                            inplace => 1}];

    $self->start_IPU(verify => 0);
    $self->create_dataset(dir_prefix   => $opts{dir_prefix},
                          dataset_spec => $mixed_dataset_4);

    if ($opts{snapshot_create}) {
        $self->create_snapshot(snapshot => $opts{snapshot_name_4});
    }
}


=head2 $ipu_obj->create_mixed_dataset_5()

To create the mixed dataset 5
    $ipu_obj->create_mixed_dataset_5()

The optional dir_prefix parameter directs the subroutine under 
which top level directory the dataset will be created.  The default 
is "IPU".

File operations are performed in-place on the fixed dataset.

Dataset_32_3 is in the directory dataset3.

By default, a snapshot named mixed_snap5 is created after the
dataset is created.  To override the default behavior and NOT
take the snapshot, use the snapshot_create option as below:

    $ipu_obj->create_mixed_dataset_5(snapshot_create => 0);

=cut

sub create_mixed_dataset_5 {
    my ($self) = shift; 
    my %opts = validate(@_,
                         {dir_prefix      => {type    => SCALAR,
                                              default => "IPU"},
                          snapshot_create => {type    => SCALAR,
                                              default => 1},
                          snapshot_name_5 => {type     => SCALAR,
                                              default => "mixed_snap5"}}
		       );

    my $mixed_dataset_5 = [{spec    => $_Dataset_Mixed_5,
                            name    => "Dataset_Mixed_5",
                            dir     => "dataset3",
                            inplace => 1}];

    $self->start_IPU(verify => 0);
    $self->create_dataset(dir_prefix   => $opts{dir_prefix},
                          dataset_spec => $mixed_dataset_5);

    if ($opts{snapshot_create}) {
        $self->create_snapshot(snapshot => $opts{snapshot_name_5}); 
    }
    
}

=head2 $ipu_obj->create_dataset_iron1()

To create the iron dataset 1
    $ipu_obj->create_dataset_iron1()

Parameters:
       dir_prefix: optional
                   directs the subroutine under which top
                   level directory the dataset will be created.
		   The default is "IPU".
       backup_create: optional
                      default is 0, no backup will be created.
       snapshot_create: optional
                      default is 0, no snapshot will be created.
       snapshot_name_1: optional
                      default is "base_snap"

=cut

sub create_dataset_iron1 {
    my ($self) = shift;
    my %opts = validate(@_,
                        {dir_prefix => {type     => SCALAR,
                                        default  => "IPU"},
                         backup_create   => {type    => SCALAR,
                                             default => 0},
			             snapshot_create => {type    => SCALAR,
					                         default => 0},
			             snapshot_name_1 => {type    => SCALAR,
					                         default => "base_snap_iron1"},
			});

    # dataset specs to be used
    my $datasets_to_create = [{spec    => $_Dataset_32_iron1,
                               name    => "Dataset_32_iron1",
                               dir     => "dataset_iron1",
                               inplace => 0},
                               ];

    $self->create_dataset(dir_prefix => $opts{dir_prefix},
                          dataset_spec => $datasets_to_create);

    if ($opts{backup_create}) {
        # dataset specs to be used for backup
        my $bak_datasets_to_create = [{spec    => $_Dataset_32_iron1,
                                   name    => "Dataset_32_iron1 bak",
                                   dir     => "dataset_iron1.bak",
                                   inplace => 0},
                                   ];
                              
        $self->create_dataset(dir_prefix => $opts{dir_prefix},
                          dataset_spec => $bak_datasets_to_create);
    }

    if ($opts{snapshot_create}) {
        my $snap_name =  $opts{snapshot_name_1};
        $self->create_snapshot(snapshot => $snap_name);
    }
}
=head1 Lower Level APIs

    For lower level callers wanting more control, the following APIs are available:

    $ipu_obj->create_dataset(...);

    $ipu_obj->create_snapshot(snapshot => $snapshot_name);

    $ipu_obj->start_IPU();

    @filenames = $ipu_obj->get_filenames($dataset_name);

    %buftree_data = $ipu_obj->get_buftree($dataset_name, $filename);

    $inode = $ipu_obj->get_inode($dataset_name, $filename);

=head2 $ipu_obj->create_dataset(...)

This function accepts a hash with two named parameters, dir_prefix and 
dataset_spec.  The dir_prefix specifies a root directory where all 
datasets are created.  The dir_prefix is optional and is "IPU" by
default..

The dataset_spec an array of hash references.  Each each hash 
reference describes a dataset and in which directory the dataset 
either resides or where it should be created.

Each hash reference is evaluated according to its appearance in the 
referenced array.

The hash reference contains the following keys:

I<spec>   : Array reference to the sequence of operation to be 
performed to create this dataset.  There are currently several 
predefined datasets documented at
L<http://wikid.netapp.com/w/WAFL/Larger_Aggregates/Mile3Tasks/TestPlan/IPU_dataset>.

At some point, the spec will be defined so callers can create and 
pass in their own dataset specification.

I<name>   : String containing the name of the dataset.
       
I<dir>    : String containing the name of the dataset directory 
where the dataset should be created.  The directory is created 
by the library (unless the key inplace is set to true, see below)

I<inplace>: If false, specifies that the dataset directory should 
be deleted (if it exists) and the dataset directory recreated.
If true, specifies that the directory exists should be left intact.
By setting inplace to true, subsequent operations can be performed 
on files that were created in another dataset. 

Below is an example:    

    my $datasets_to_create = [{spec    => $_Dataset_32_1,
                               name    => "Dataset_32_1",
                               dir     => "dataset1",
                               inplace => 0},
                              {spec    => $_Mixed_Dataset_32_1,
                               name    => "MixedDataset",
                               dir     => "dataset1",
                               inplace => 1}
                             ]

    $ipu_obj->create_dataset(dataset_spec => $dataset_to_create,
                             dir_prefix   => "directory");

=cut

sub create_dataset {
    my $self = shift;
    my %opts = validate(@_,
                        {dir_prefix   => {type    => SCALAR,
                                          default => 'IPU'},
                         dataset_spec => {type    => ARRAYREF}
                        });
    #store the information for this dataset
    #note that any datasets with the same name get overwritten

    if (! scalar (@ { $opts{ dataset_spec } } ) ) {
        my $xx = Dumper($opts{dataset_spec});
        NATE::BaseException->throw( "No dataset passed in, check dataset_spec:$xx" );
    }
        
    #TODO check to see if volume is big enough and grow if option
    #provided.

    #TODO validate each entry in the dataset_spec contains
    # - a name and/or spec
    # - if only a name, that it be a well known spec and copied in
    # - either a dir_prefix and/or a dir default
    # - inplace, if omitted then default is true

    foreach my $dataset (@ { $opts{ dataset_spec } } ) {
        #for each dataset, attempt to create the dataset by first
        #creating the directory where the files are to reside
        #then perform the operation defined by the dataset
        #keep track of dataset information in %dataset_data, this
        #will be stored in the IPU object so we can retrieve info later
        #for users

        my $operation = $dataset->{spec};
        my $dataset_spec;
        my $dataset_directory;
        my $volume_list;
        my %_dataset_info = (filenames => {});  #store info for $self
        
        $Log->comment("Attempt to create dataset: $dataset->{name}");
        $Log->comment("Ref of operation is:" . ref $dataset->{spec}. "xx");

        #check that operations are there for this dataset
        if ( (! $operation ) || (ref $operation eq "ARRAYREF" ) ) {
            my $dataset_dump = Dumper($dataset);
            NATE::BaseException->throw( "The ops are not defined or " . 
                                        "not an arrayref " .
                                        "dataset_spec:$dataset_dump" );
        }
        
        #dataset bookkeeping, keep the name and init the structure for this
        #dataset.  have the field in $self reference $_dataset_info{name} for
        #readability.
        $_dataset_info{name} = $dataset->{name};
        $self->{_dataset_data}{$dataset->{name}} = \%_dataset_info;

        #create the path where the dataset files will be created
        #this will then be prefixed by either the host mount point
        #to form the $host_path and the node's volume path to form
        #the $node_path.  both are needed.
        $dataset_directory = "";
        if ($opts{dir_prefix}) {
            $dataset_directory = $opts{dir_prefix};
        }
        if ($dataset->{dir}) {
            $dataset_directory = $dataset_directory . '/' . $dataset->{dir};
        }
        $dataset_directory =~ s{//}{/}g; #replace any '//' with a single '/'
        if (length($dataset_directory) == 0) {
            NATE::BaseException->throw( "no directory specified for dataset");
        }
        
        #dataset bookkeeping, remember the directory passed in
        #it is used to reconstruct the directory on both the node
        #and the host
        $_dataset_info{dataset_directory} = $dataset_directory;

        $volume_list = $self->{volume_list};
        foreach my $volume_name (@ {$volume_list} ) {
            my $volume_data = $self->{_volume_data}{$volume_name};
            my $node_path;
            my $host_path;

            #path on the client box, will be passed into lower
            #level subroutines
            $host_path = $self->_make_host_path($volume_data,
                                                $dataset_directory);
            $Log->debug("host path: $host_path");

            #path on the filer, will be passed into lower
            #level subroutines
            $node_path = $self->_make_node_path($volume_data,
                                                $dataset_directory);
            $Log->debug("node path: $node_path");

            #make the directory, if inplace == 0, then will rm -Rf before 
            #the mkdir this is needed for mixed datasets that require that 
            #the file already exists
            #TODO if inplace==1 and directory doesn't exist, throws an exception
            _remake_dir($self, {inplace   => $dataset->{inplace}, 
                                host_path => $host_path});   

            #get the dataset definition from parms
            #for eeach operation specified, perform it, this is how the 
            #datasets get created
            $dataset_spec = $dataset->{spec};
    
            #perform each operation in the dataset definition
            foreach $operation (@ { $dataset_spec } ) {
                my $filename = $operation->{filename};
                $_dataset_info{filenames}{$filename} = 1;

                #TODO: turn the cmd_parms into a hash that gets sent down
                #      to the underlying _mkfile(), _testsetattr_W(), etc 
                #      commands
                my $cmd_parms = $operation->{parms};
                my $cmd = $operation->{cmd};
                & { $cmd } ($self,{node_path      => $node_path,
                                   host_path      => $host_path,
                                   volume         => $volume_name,
                                   filename       => $filename, 
                                   cmd_parms      => $cmd_parms});
            }
        }
        $Log->debug("self _dataset_data is: " .
                                 Dumper($self->{_dataset_data}));
    }
}

=head2 $ipu_obj->create_snapshot(snapshot => $snapshot_name)

A snapshot with the given name is created.

=cut

sub create_snapshot {
    my $self = shift;
    my %opts = validate(@_,
                        {snapshot => {type => SCALAR}});
    my $snap_name = $opts{snapshot};
    $Log->enter();

    #for each volume that was specified in the constructor, check to 
    #see if the snap shot exists and delete it if it exists.  if it 
    #doesn't, an exception will be thrown so catch it and continue
    foreach my $volume_name (@ { $self->{volume_list} } ) {
        my $cmd;
        my $snapshot;
        my $volume_data = $self->{_volume_data}{$volume_name};
        my $volume_vserver = $volume_data->{volume_vserver};

        $Log->trace("will find old snapshot $snap_name in $volume_name");
        try {
            $snapshot = NACL::C::Snapshot->find(command_interface => $self->{node_obj},
                                                filter => {
                                                    vserver  => $volume_vserver,
                                                    snapshot => $snap_name,
                                                    volume   => $volume_name
                                                });

            $snapshot = NACL::C::Snapshot->delete(command_interface => $self->{node_obj},
                                                  vserver           => $volume_vserver,
                                                  snapshot          => $snap_name,
                                                  volume            => $volume_name,
                                                  foreground        => "true");
                        
        }
        catch NACL::Exceptions::NoElementsFound with {
            $Log->trace("error: $volume_name, not existed, continue");
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            $Log->trace("error: $volume_name, not deleted");
        }
        finally {
            $Log->trace("error received from Snapshot delete on $volume_name is $@");
        };

        #attempt to create the snapshot 
        $snapshot = NACL::C::Snapshot->create(command_interface => $self->{node_obj},
                                              vserver           => $volume_vserver,
                                              snapshot          => $snap_name,
                                              volume            => $volume_name,
                                              foreground        => "true");

        $Log->trace("did create snapshot $snap_name in $volume_name");
    }
    $Log->exit();
}

=head2 $ipu_obj->start_IPU()

Starts IPU scanning.  This function is used to control when 64-bit mode 
is turned on.  When no parameters are passed in, start_IPU() begins IPU scan on 
each volume that was passed into the new() constructor.  

Alternately, callers may pass in an aggregate name, volume name, or ref to an
array of volume names.  In each of these cases, the IPU scan will be turned on
for the aggregrate, volume, or each volume in the array.

    $ipu_obj->start_IPU(aggregate => $aggr_name);

    $ipu_obj->start_IPU(volume => $volume_name);

    $ipu_obj->start_IPU(volume => \@volume_list);

Note, at most, one of aggregrate or volume may be passed in.

Another optional parameter is verify.  If verify is set to 1 (the default) then
the library will throw an exception if a start_IPU() is requested on a volume or
aggregrate which is already 64-bit.  If verify is set to 0, then a log message
is emitted but no exception thrown.

    $ipu_obj->start_IPU(volume => $volume_name, verify => 1);

=cut

sub start_IPU {
    my $self = shift;
    my (%opts) =  validate(@_,
                           {aggregate => {type => SCALAR,
                                          optional => 1},
                            volume    => {type => SCALAR | ARRAYREF,
                                          optional => 1},
                            verify    => {type => SCALAR,
                                          default => 1}
                  });
    my $volume_list = [];

    #param check, only accept an aggregate object or 
    #turn scanning is off
    if ( $opts{volume} && $opts{aggregate} ) {
        NATE::BaseException->throw("Cannot pass in both aggregate and volume");
    }

    #if aggregate passed in and 32-bit aggregate then do the upgrade
    #if aggregate is 64-bit, exception will be thrown
    if ($opts{aggregate}) {
        return $self->_aggregate_start_ipu($opts{aggregate},
                                           $opts{verify});
    }
    #else, if volume is passed in, copy to parms to use in starting scan
    elsif ($opts{volume}) {
        if (ref $opts{volume}) {
            $volume_list = $opts{volume};
        }
        else {
            $volume_list = [ $opts{volume} ];
        }
    }
    #no params so default to using the volume_list from constructor
    else {
        $volume_list = $self->{volume_list};
    }

    return $self->_volume_start_ipu($volume_list, $opts{verify});
}
    
=head2 $ipu_obj->set_wafl_64bit_upgrade_scan();

Sets the wafl_64bit_upgrade_scan_disable flag to the passed in value.

To set the disable flag to 1 (thereby disabling background scanning)

    $ipu_obj->set_wafl_64bit_upgrade_scan_disable(disable => 1);

To set the disable flag to 0 (thereby enabling background scanning)

    $ipu_obj->set_wafl_64bit_upgrade_scan_disable(disable => 0);

=cut
sub set_wafl_64bit_upgrade_scan{
    my $self = shift;
    my (%opts) = validate(@_, 
                          {disable => {type => SCALAR}}
                         );
    my $cmd;

    $cmd = "priv set -q test;" .
           "setflag wafl_64bit_upgrade_scan_disable " .
           $opts{disable};

    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node local -command \"" . $cmd .'"';
    }

    _execute_command($self->{node_api_set_obj}, {command => $cmd});
}

sub _aggregate_start_ipu {
    my ($self, $aggregate, $verify) = @_;
    my $cmd;
     
    if (! $self->_aggregate_check_32bit($aggregate) ) {
        if ($verify) {
            NATE::BaseException->throw("Cannot start IPU on aggregate $aggregate " .
                                            "already scanning or scanned"); 
        }
        else {
            $Log->trace("Aggregate $aggregate already 64-bit or scanning, " .
                            "verify off");
            return;
            }
    }

    $Log->trace("start IPU on aggregate $aggregate");
    $cmd = "priv set -q diag;aggr 64bit-upgrade start $aggregate";
    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node local -command \"" . $cmd .'"';
    }
    _execute_command($self->{node_api_set_obj}, {command => $cmd,
                                                 timeout => 300});
} 

sub _volume_start_ipu {
    my ($self, $volume_list, $verify) = @_;
    my $cmd;

    #YV: $cmd = "priv set -q test;setflag wafl_64bit_upgrade_scan_disable 1";
    #YV: if ($self->{shell_type} eq "Nodescope") {
    #YV:     $cmd = "system node run -node local -command \"" . $cmd .'"';
    #YV: }
    #YV: _execute_command($self->{node_api_set_obj}, {command => $cmd});

    foreach my $volume_name (@ { $volume_list } ) {
        #launch the IPU, the scanner should be off and should remain off

        $Log->trace("check if $volume_name is 32bit");
        if (! $self->_volume_check_32bit($volume_name)) {
            if ($verify) {
                NATE::BaseException->throw("Cannot start IPU on volume " .
                                            "$volume_name already scanning or scanned");
            }
            else {
                $Log->trace("Volume $volume_name already 64-bit or scanning, " .
                                "verify off");
                next;
            }
        }

        $Log->trace("start IPU on volume $volume_name");
        $cmd = "priv set -q diag;vol 64bit-upgrade start $volume_name";

        if ($self->{shell_type} eq "Nodescope") {
            $cmd = "system node run -node local -command \"" . $cmd .'"';
        }

        _execute_command($self->{node_api_set_obj}, {command => $cmd,
                                                     timeout => 300});
    }
}

sub _aggregate_check_32bit {
    my ($self, $aggregate, $verify) = @_;
    my $aggr;
    my $block_type;

    try {
        $aggr = NACL::C::Aggregate->find( 
            command_interface => $self->{node_obj},
            filter => { aggregate => $aggregate} 
                 );
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->trace("Checking for 32bit aggregate, $aggregate not found");
        $@->throw();
    };
    $block_type = $aggr->state()->block_type();
    $Log->debug("Block type for aggregate $aggregate is $block_type");

    #already a 64 bit aggregate or a scan is already in progress
    if ($block_type =~ /32-bit/) {
        return 1;
    }
    return 0;
}

sub _volume_check_32bit {
    my ($self, $volume_name) = @_;
    my $cmd;
    my $response;
    my $vol;
    my $block_type;
    
    try {
        $vol = NACL::C::Volume->find(
                    command_interface => $self->{node_obj},
                    filter => { volume => $volume_name}
                );
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->trace("Checking for 32bit volume, $volume_name not found");
        $@->throw();
    };

    $block_type = $vol->state()->block_type();
    $Log->debug("Block type for volume $volume_name is $block_type");

    if ($block_type =~ /32-bit/) {
        return 1;
    } 
    else {
        return 0;
    }
}

sub _exportfs_volume {
    my ($self, $node_path) = @_;
    my $cmd;

    $cmd = "priv set diag;exportfs -p rw,anon=0 " . $node_path;
    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node local -command \"" . $cmd .'"';
    }
    _execute_command($self->{node_api_set_obj},
                     {command => $cmd});
}
                    

=head2 %buftree_data = $ipu_obj->get_buftree($dataset_name, $filename, $volume)

$dataset_name: the name of the dataset containing the file.

$filename    : the name of the file under interest.

$volume      : (optional) if new() was called with $volume set to an arrayref containing
multiple volume names then $volume must be passed in and contain the name
of the volume containing the dataset.

Consider the output from the I<wafl buftree check> command:

    ----- ------------ ------------ ------------ ------------ ------------
    level 32-bit blks  32-bit holes 64-bit blks  64-bit holes     total   
    ----- ------------ ------------ ------------ ------------ ------------
        3            1            0            0            0            1
        2            0            1            0            0            1
        1            0          510            0            0          510
        0            0       260100            0            0       260100
    total            1       260611            0            0       260612
    ----- ------------ ------------ ------------ ------------ ------------

get_buftree() returns a hash containing keys: 

    max_level : Contains the maximum level for the file.  In the example above,
$buftree_data{max_level} would be 4.

    raw_output: Contains raw output of the I<wafl buftree check> command.  Note
that superfluous output may be included along with the buftree output.

get_buftree() also parses the raw output and returns a two-dimensional 
hash that mirrors the buftree.  The first dimension represents the level plus 
total.  In the case of the buftree above, the first dimension would be the 
values 3, 2, 1, 0, and "total".

The second dimension represents the blocks types.  The keys are 
for the second dimension are the following:

=over 4

=item * 32_blks

=item * 32_holes

=item * 64_blks

=item * 64_holes

=item * total

=back

In the example buftree above, $buftree{0}{32_holes} would be set to 260100.

=cut

sub get_buftree {
    my $self = shift;
    my %opts = validate(@_,
                        {dataset  => {type => SCALAR},
                         filename => {type => SCALAR},
                         volume   => {type => SCALAR,
                                      optional => 1}});
    my $dataset_info;
    my $cmd;
    my $inode;
    my $fullpath;
    my $response;
    my @buftree;
    my %buftree_data = ();

    #process parm
    my $dataset_name = $opts{dataset};
    my $filename = $opts{filename};
    my $volume_name = $self->_get_volume_name($opts{volume});
    

    #get the inode and then execute the wafl buftree check command
    #parse the output and then massage the results into a two dimensional
    #hash where the first dimension is the level or "total" for the total
    #and the second dimension is the type of block (32_blk, 32_hole, 64_blk,
    #64_hole, and total for grand totals)
    
    #below is the sample output:
    #----- ------------ ------------ ------------ ------------ ------------
    #level 32-bit blks  32-bit holes 64-bit blks  64-bit holes     total   
    #----- ------------ ------------ ------------ ------------ ------------
    #   3            1            0            0            0            1
    #   2            0            1            0            0            1
    #   1            0          510            0            0          510
    #   0            0       260100            0            0       260100
    #total           1       260611            0            0       260612

    $inode = $self->get_inode(dataset  => $dataset_name,
                              filename => $filename,
                              volume   => $volume_name);
    if (! $inode) {
        NATE::Exceptions::Argument->throw("Unable to get inode for file");
    }

    $cmd = "priv set -q test;" .
                "wafl buftree check -n -V $volume_name $inode"; 

    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node local -command \"" . $cmd .'"';
    }

    $response = _execute_command($self->{node_api_set_obj},
                                 {command => $cmd,
                                  timeout => 300}); 

    #will get all the lines of output execept for "total", get that 
    #later
    (@buftree) = $response =~ m{^\s+(\d+)   #level
                                 \s+(\d+)   #32-bit blks
                                 \s+(\d+)   #32-bit hole 
                                 \s+(\d+)   #64-bit blks
                                 \s+(\d+)   #64-bit hole
                                 \s+(\d+)$  #total blocks for this level
                               }gcxms;  
    if ( ! (scalar @buftree) ) {
        NATE::BaseException::->throw("Unable to parse buftree data" . 
                                     $response);
    }

    #should get a mod 6 since eacy line has 6 entries in it
    if (((scalar @buftree) % 6) != 0) {
        NATE::BaseException::->throw("Inconsistent amount of buftree data" . 
                                     $response);
    }

    #store the raw output in {raw_output}
    #the first alpha will have the highest level for the file
    my $i = 0;
    $buftree_data{raw_output} = $response;  #may contain other output as well 
    $buftree_data{max_level} = $buftree[0]; #first value is the highest level
    while ( $i < (scalar @buftree) ) {
        my $level = $buftree[$i++];       #first value on line is the level
        $buftree_data{$level}             = {};
        $buftree_data{$level}{"32_blks"}  = $buftree[$i++];
        $buftree_data{$level}{"32_holes"} = $buftree[$i++];
        $buftree_data{$level}{"64_blks"}  = $buftree[$i++];
        $buftree_data{$level}{"64_holes"} = $buftree[$i++];
        $buftree_data{$level}{"total"}    = $buftree[$i++];
    }  

    #get the totals by pulling from the output and then
    #adding to the buftree_data hash as "total"
    (@buftree) = $response =~ m{^total       #total     
                                 \s+(\d+)    #32-bit blks
                                 \s+(\d+)    #32-bit holes
                                 \s+(\d+)    #64-bit blks
                                 \s+(\d+)    #64-bit holes
                                 \s+(\d+)$   #total blocks for this file 
                               }gcxms;
    $Log->debug("buftree for total is " . Dumper(\@buftree));
    if ( ! (scalar @buftree) ) {
        NATE::BaseException::->throw("Unable to parse buftree total data"
                                        . $response);
    }
    $buftree_data{total}             = {};
    $buftree_data{total}{"32 blks"}  = $buftree[0];
    $buftree_data{total}{"32 holes"} = $buftree[1];
    $buftree_data{total}{"64 blks"}  = $buftree[2];
    $buftree_data{total}{"64 holes"} = $buftree[3];
    $buftree_data{total}{"total"}    = $buftree[4];

    $Log->debug("buftree data is " . Dumper(\%buftree_data));
    return %buftree_data;
}
    
    
=head2 $ipu_obj->get_inode($dataset_name, $filename, $volume);

Returns the inode number of $filename within $dataset_name.

$dataset_name: the name of the dataset containing the file.

$filename    : the name of the file under interest.

$volume      : (optional) if new() was called with $volume set to an arrayref containing
multiple volume names then $volume must be passed in and contain the name
of the volume containing the dataset.


=cut

sub get_inode {
    my $self = shift;
    my %opts = validate(@_,
                        {dataset  => {type     => SCALAR},
                         volume   => {type     => SCALAR,
                                      optional => 1},
                         filename => {type     => SCALAR}});
    my $volume_data;
    my $full_path_name;;
    my $dataset_info;
    my $filename = $opts{filename};
    my $dataset_name = $opts{dataset};
    my $volume_name = $self->_get_volume_name($opts{volume});
    
    #pull the node path out of the volume data, then pull the 
    #the dataset_directory to form the full path to the file
    #use this to get the inode from the lower-level call _get_inode()

    #get the volume data
    $volume_data = $self->{_volume_data}{$volume_name};
    if (! ref $volume_data ) {
        NATE::BaseException->throw("no volume data, for $volume_name");
    }

    #make sure this dataset was seen before
    if (! $self->{_dataset_data}{$dataset_name} ) {
        NATE::BaseException->throw("no dataset data found for $dataset_name");
    }

    #grab the structure for the dataset information
    $dataset_info = $self->{_dataset_data}{$dataset_name};

    #see if the filename is found within this dataset info
    if (! $dataset_info->{filenames}{$filename} ) {
        NATE::BaseException->throw("no filename data found for $filename" .
                                      "in dataset $dataset_name");
    }

    $full_path_name = $self->_make_node_path($volume_data,
                                             $dataset_info->{dataset_directory});
        
    $full_path_name = $full_path_name . $filename;
    return $self->_get_inode($self->{node_api_set_obj}, $full_path_name);
}

=head2 @filenames = $ipu_obj->get_datasets()

Returns an array with the list of dataset names

=cut
sub get_datasets {
    my $self = shift;
    my $ret = $self->{_dataset_data};
    my @return_value = sort keys( % {$ret} );
    return @return_value;
}

=head2 @filenames = $ipu_obj->get_filenames(dataset => $dataset_name
                                         , format=>$format);


Returns an array with the list of file names in the named dataset in the specified format,
'long' or 'short' (default).

If format => 'short' is specified, then an array containing only the filename is returned 
(e.g. file_l0).  

To retrieve the fully qualified file name on the filer, use the "format => 'long'" form.

@filenames = $ipu_obj->get_filenames(dataset => $dataset_name,
                                     format  => 'long');

This will return an array in the /vol/<volume_name>/<directory>/fname
format.  This is useful in cases where some other file operations are to be
performed on the file outside of this module.

If the dataset is created on multiple volumes (done by specifying a list of
volumes when calling the constructor, IPU::Dataset::new()), then the list
of filenames returned can be limited to those on a subset of volumes by
specifying a single or a list of volumes in the volumes parameter.

    $ipu_obj = IPU::Dataset->new(filer   => $filer_name,
                                 volume  => ['v1','v2','v3'],
                                 client  => $client_name);
    ...
    ...
    @filenames = $ipu_obj->get_filenames(dataset => $dataset_name,
                                         format  => 'long',
                                         volume  => ['v1','v2']);

=cut
sub get_filenames {
    my $self = shift;
    my %opts = validate(@_,
                        {dataset => {type => SCALAR},
                         format  => {type => SCALAR, default  => "short"},
                         volume  => {type => SCALAR | ARRAYREF,
                                     optional => 1},
                        }
                );
    my $dataset_name = $opts{dataset};
    my $dataset_info;
    my $volumes_to_get;
    my @return_values;

    #make sure this dataset was seen before
    if (! $self->{_dataset_data}{$dataset_name} ) {
        NACL::C::BaseException->throw(text =>
                                     "no dataset named $dataset_name");
    }

    #NOTE that a 'full' format may be implemented at some future time
    #to return more information still such as mounted file name,etc.  
    #this is deferred until the requirements for such are understood

    #if a volume or list of volumes is passed in, then return filenames
    #for that volume only.  if not volume name is passed in then get the
    #filenames for all the files in all the volumes
    if (defined $opts{volume} ) {
        if (! ref $opts{volume} ) {
            $volumes_to_get = [$opts{volume} ];
        }
        #if volumes were passed in, do sanity check to make sure
        #dataset was created for this volume
        foreach my $volume_name (@ {$volumes_to_get} ) {
            if (! defined $self->{_volume_data}{$volume_name} ) {
                NACL::C::BaseException->throw(text => 
                                         "no volume named $volume_name");
            }
        }
    }
    #if no list of volumes then get them all
    else {
        $volumes_to_get = $self->{volume_list};
    }

    #get the dataset information for this dataset
    $dataset_info = $self->{_dataset_data}{$dataset_name};

    #check what the format is, if short then do not need to
    #scan the list of volumes as there is only one name per
    #dataset
    if ($opts{format} eq "short") {
        #return just the names in it
        $Log->debug("filenames array:" . Dumper($dataset_info->{filenames}));
        @return_values = sort keys( % {$dataset_info->{filenames} } );
        return @return_values;
    }
    #if getting the long format, then need to scan the list of volumes
    #and get the full path name for each file in dataset
    elsif ($opts{format} eq "long") {
        my @full_path;
        my $dataset_directory = $dataset_info->{dataset_directory};

        #for each volume where the dataset was created, get the 
        #the path on the node (e.g. /vol/vol_name/IPU/dataset)
        foreach my $volume_name (@ {$volumes_to_get} ) {
           my $volume_data = $self->{_volume_data}{$volume_name};
           my $path_to_file = $self->_make_node_path($volume_data,
                                                     $dataset_directory);

           #for each file name in the volume, push the path concatenated
           #with the filename
           foreach my $filename (keys(%{$dataset_info->{filenames}})) {
               push @full_path, "${path_to_file}${filename}";
            }
        }
        @return_values = sort @full_path;
        return @return_values;
    }
    else {
        NACL::C::BaseException->throw(text =>
                                "No such format $opts{format} supported");
    }
}

#sub is called by subs that have volume name as optional parm
#sub checks if parm is not passed in that only one volume exists
#in the $self->{volume_list} and returns that o/w, throws exception
sub _get_volume_name {
    my ($self, $volume_parm) = @_;
    my $command = ( caller(1) )[3];
    my $volume_name;

    #if no volume name is passed in, then need to verify that only
    #one volume is in use and use that. otherwise, throw exception
    if (! $volume_parm )  {
        if ( scalar @{ $self->{volume_list} } != 1 ) {
            NATE::BaseException->throw("no volume name specified for $command");
        }
        else {
            $volume_name = $self->{volume_list}[0];
        }
    } 
    else {
        $volume_name = $volume_parm;
    }
    return $volume_name;
} 

#determine the path on the host by combining the mount point with the directory
sub _make_host_path {
    my ($self, $volume_data, $directory) = @_;
    my $full_host_path;

    $full_host_path = $volume_data->{volume_host_mount_point};
    if ($directory) {
        $full_host_path = $full_host_path . '/' . $directory . '/';
    }

    $full_host_path =~ s{//}{/}g; #replace any '//' with a single '/'
    return $full_host_path;
}

#determine the path on the node by combining the path to the volume along with the
#directory
sub _make_node_path {
    my ($self, $volume_data, $directory) = @_;
    my $full_node_path;

    my $volume_name = $volume_data->{volume_name};

    $full_node_path = '/vol/' . $volume_name . '/';

    if ($directory) {
             $full_node_path = $full_node_path . '/' . $directory . '/';
    }

    $full_node_path =~ s{//}{/}g; #replace any '//' with a single '/'
    $Log->trace("returning node path of $full_node_path");
    return $full_node_path;
}

sub _remake_dir {
    my ($self, $parms) = @_;
    my $cmd;
    my $response;

    #if inplace == 0, then delete the directory 
    #before trying to mkdir it
    $cmd = "";
    if ($parms->{inplace} == 0) {
        $cmd = "/usr/bin/sudo /bin/rm -Rf $parms->{host_path};" 
    }
    $cmd = $cmd . 
               "/usr/bin/sudo /bin/mkdir -p " .
               "$parms->{host_path};/bin/ls -d $parms->{host_path}";

    $response = _execute_command($self->{host_api_set_obj},
                                 {command => $cmd});
}

#for nodescope nodes, will get the junction point.  this is needed
#for determining where the filer should be exported from when
#mounting for NFS
sub _get_vserver_junction {
    my ($self, $volume_data) = @_;
    my $volume_name = $volume_data->{volume_name};
    my $cmd;
    my $response;
    my $junction;
    my $vserver;
   
    $Log->enter();
    my $vol_obj = NACL::STask::Volume->find(
                        command_interface   => $self->{node_obj},
                        filter              => { volume => $volume_name}
                  );
    
    $vserver    = $vol_obj->get_one_state_attribute("vserver");
    $junction   = $vol_obj->get_one_state_attribute("junction-path");

    if ($vserver && ($junction !~ /^-$/)) {
        $volume_data->{volume_junction} = $junction;
        $volume_data->{volume_vserver} = $vserver;
    }
    else {
        NATE::BaseException->throw("Junction is not active for volume");
    }
    $Log->trace("vserver and junction are $vserver and $junction");
    $Log->exit();
}

#lower level call to get the IP addresses for the cmode filer
#will use this to specify the correct server name on the mount
sub _get_cmode_export_node_name {

    my ($node_obj, $volume_data) = @_;
    
    my @Ifaces;
    my $IP_info;
    my $IP_addr;

    $Log->enter();

    @Ifaces = NACL::C::Interface->find(
                command_interface => $node_obj,,
                filter => 
                    {server => $volume_data->{volume_vserver}}
                                      );

    $Log->trace("num of cmode interface found " . scalar(@Ifaces));

    $IP_info = $Ifaces[0]->state();
    $IP_addr = $IP_info->{address};
    $Log->trace("cmode export IP address is $IP_addr");
    $Log->exit();

    return $IP_addr;
}

#lower level call to get the inode for the file
sub _get_inode {
    my ($self, $api_set_obj, $full_filename) = @_;
    my $response;
    my $inum;
    
    #get the inode number for the file
    $response = $self->{node_api_set_obj}->slist(path => $full_filename);
    $inum = $response->get_parsed_output()->[0]{'inode'};
    $Log->comment("inode = $inum");

    return $inum;
}

#TODO make this bullet proof and exception catching
sub _execute_command {
    my ($api_set_obj, $parms) = @_;
    my $response;

    $Log->trace("_execute_command:" . $parms->{command});
    $response = $api_set_obj->execute_raw_command(% { $parms } );
    $Log->trace("_execute_command response: $response");
    if ($response =~ /Error:/) {
        NATE::BaseException->throw("Error executing $parms->{command} " .
                                   "response is $response");
    }
    return $response;
}

#this routine should only be called by routines that implement the calls
#to the filer through the $api_set_obj (e.g. _mkfile(), _testsetattr_f)
#to make sure there isn't a missing parm in the dataset datastructure
sub _check_required_parms {
    my ($required_parms, $parms) = @_;

    #go through list of required params and check to see if
    #any of them are missing
    #TODO change this sub to return an array containing missing req. parms
    map {
        #if missing something, then return
        if (! defined $parms->{$_}) {
            my $command = ( caller(1) )[3];
            NATE::BaseException->throw("missing required paramater $_ " .
                                        "to commands: $command");
        }
    } @{$required_parms};

    #all required parms are there 
    return 1;
}

#help subroutine, will set to $timeout but return the previous value
sub _set_timeout {
    my ($self,$timeout) = @_;
    my $apiset = $self->{node_api_set_obj};     

    #get the old value
    my $old_timeout = $apiset->get_timeout();

    #set the new value
    $apiset->set_timeout('connectrec-timeout' => $timeout);
    my $tmp = $apiset->get_timeout();
    $Log->comment("timeout is now set to $tmp");

    #return the old value
    return $old_timeout;
}

sub _testsetattr_f {
    my ($self, $parms) = @_;
    my $cmd;
    my $full_filename;
    my $cmd_parms;

    $Log->enter();

    $cmd_parms = $parms->{cmd_parms};

    _check_required_parms(["enabled"], $cmd_parms);

    $full_filename = $parms->{node_path} . $parms->{filename};

    my $old_timeout = $self->_set_timeout(600);
    $self->{node_api_set_obj}->testsetattr('file-space-reservations' 
                                                            => $cmd_parms->{"enabled"},
                                            path => $full_filename);
    $self->_set_timeout($old_timeout);

    $Log->exit();
  }

sub _testsetattr_s {
    my ($self, $parms) = @_;
    my $cmd;
    my $full_filename;
    my $cmd_parms;

    $Log->enter();
    $cmd_parms = $parms->{cmd_parms};
    _check_required_parms(["size"], $cmd_parms);

    $full_filename = $parms->{node_path} . $parms->{filename};

    $self->{node_api_set_obj}->testsetattr("file-size" => $cmd_parms->{size},
                                            path => $full_filename);
    $Log->exit();
}

sub _testsetattr_W {
    my ($self, $parms) = @_;
    my $cmd;
    my $full_filename;
    my $cmd_parms;

    $Log->enter();
    $cmd_parms = $parms->{cmd_parms};
    _check_required_parms(["offset","len"], $cmd_parms);

    $full_filename = $parms->{node_path} . $parms->{filename};

    my $old_timeout = $self->_set_timeout(1800);
    $self->{node_api_set_obj}->testsetattr(path     => $full_filename,
                                           "write-offset-and-length" 
                                                    => "$cmd_parms->{offset} $cmd_parms->{len}");
    $self->_set_timeout($old_timeout);

    $Log->exit();
}

sub _wafl_touch {
    my ($self, $parms) = @_;
    my $cmd;
    my $full_filename;
    my $response;
    my $inum;
    my %default_parms = ();
    my $cmd_parms;

    $Log->enter();
    $cmd_parms = $parms->{cmd_parms};
    _check_required_parms(["level","fbn_start"], $cmd_parms);

    %default_parms = ("fbn_end" => $cmd_parms->{fbn_start});
    map { $cmd_parms->{$_} ||= $default_parms{$_} } keys(%default_parms);

    #get the inode number for the file
    $full_filename = $parms->{node_path} . $parms->{filename};
    if (! ($inum = $self->_get_inode($self->{node_api_set_obj}, $full_filename))) {
        $Log->trace("Unable to get inode number, return");
        return;
    }

    $response = $self->{node_api_set_obj}->wafl_touch(volume         => $parms->{volume},
                                                      'inode-number' => $inum,
                                                      'level-value'  => $cmd_parms->{level},
                                                      'start-fbn'    => $cmd_parms->{fbn_start},
                                                      'end-fbn'      => $cmd_parms->{fbn_end});
    $Log->exit();
}

sub _wafl_punch {
    my ($self, $parms) = @_;
    my $cmd;
    my $full_filename;
    my $cmd_parms = $parms->{cmd_parms};
  
    $Log->enter();
    _check_required_parms(["fbn","count"], $cmd_parms);

    $full_filename = $parms->{node_path} . $parms->{filename};
    $cmd = "priv set -q advanced;" .
        "wafl punch $full_filename $cmd_parms->{fbn} $cmd_parms->{count}";

    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node local -command \"" . $cmd .'"';
    }

    _execute_command($self->{node_api_set_obj}, {command => $cmd});
    $Log->exit();
}

sub _mkfile {
    my ($self, $parms) = @_;
    my $cmd_parms;
    my $full_filename;
    my $node_api_set_obj = $self->{node_api_set_obj};

    $Log->enter();
    $cmd_parms = $parms->{cmd_parms};

    #check required command parms
    _check_required_parms(["size"], $cmd_parms);

    $full_filename = $parms->{node_path} . $parms->{filename};
    
    #if a timeout value is passed in
    my $old_timeout;
    if ($cmd_parms->{timeout}) {
        $old_timeout = $self->_set_timeout($cmd_parms->{timeout});
    }

    #get the api set obj and execute the command
    $node_api_set_obj->mkfile(size      => $cmd_parms->{size},
                              filename  => $full_filename,
                              filecount => 1);

    #if a timeout value is passed in, restore to old value
    if ($parms->{timeout}) {
        $self->_set_timeout($old_timeout)
    }

    $Log->exit();
}

sub _filersio {
    my ($self, $parms) = @_;
    my $cmd;
    my $cmd_parms;
    my $full_filename;
    my $size;
    my %default_parms = (timeout => 300);

    $Log->enter();
    $cmd_parms = $parms->{cmd_parms};

    #default parm setting
    map { $parms->{$_} ||= $default_parms{$_} } qw/timeout/;

    #check required command parms
    _check_required_parms(["end_byte"], $cmd_parms);
    

    $full_filename = $parms->{node_path} . $parms->{filename};

    my $old_timeout = $self->_set_timeout($parms->{timeout});
    $self->{node_api_set_obj}->filersio_asyncio_active(filename => $full_filename,
                                                       "read-percentage" => $cmd_parms->{read_percent},
                                                       "rand-percentage" => $cmd_parms->{rand_percent},
                                                       "blk-size" =>  $cmd_parms->{blk_size}, 
                                                        "start-byte" => $cmd_parms->{start_byte},
                                                        "end-byte" => $cmd_parms->{end_byte},
                                                        "run-time" => $cmd_parms->{run_time}, 
                                                        "num-active" => $cmd_parms->{num_active},
                                                       );
                                                                        
    $self->_set_timeout($old_timeout);
    $Log->exit();
}
1;
