# $Id: //depot/prod/test/Rsierranevada/lib/IPU/NACL_IPU.pm#1 $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
## @summary
##
## How to use
##
## @status
#  2010 NetApp, Inc.  All rights reserved.
## @summary IPU object NACL format
## @status public
## @author vchan
# Have tharndoc use POD:
# - @summary, @author, @status still used when indexing
# - Lines below beginning with "##<any non-blank text>" will give tharndoc error
# - To display documentation:  tharndoc <name>.pm <==> perldoc -t <name>.pm
# Design assumptions, standards:
# - before NACL setup and Cleanup Sub-system Functions published,
#   the wrapper is responsible for cleaning up all the aggregates and volumes
# created for specific testcase usage.
# - http://wikid.netapp.com/w/TADA/NATE_LIB/Standards
#
## @pod here

=head1 NAME

NACL_IPU.pm

=head1 SYNOPSIS

Methods supporting IPU CLIs

 use IPU::NACL_IPU;
 $IPU_obj = IPU::NACL_IPU->new(%args);
 $Ret = $IPU_obj->ipu_setup(%args);
 $Ret = $IPU_obj->ipu_cleanup(%args);
 $Ret = $IPU_obj->create_aggr(%args);
 $Ret = $IPU_obj->create_vol(%args);
 $Ret = $IPU_obj->aggrs_delete(%args);
 $Ret = $IPU_obj->aggr_delete(%args);
 $Ret = $IPU_obj->vols_delete(%args);
 $Ret = $IPU_obj->aggr_upgrade_scan_start(%args);
 $Ret = $IPU_obj->aggr_upgrade_scan_stop(%args);
 $Ret = $IPU_obj->aggr_upgrade_diag(%args);
 $Ret = $IPU_obj->vol_upgrade_diag(%args);
 $Ret = $IPU_obj->vol_is_64bit(%args);
 $Ret = $IPU_obj->vol_is_mixed(%args);
 $Ret = $IPU_obj->disable_IPU_scanner(%args);
 $Ret = $IPU_obj->enable_IPU_scanner(%args);
 $Ret = $IPU_obj->get_upgrade_status(%args);
 $Ret = $IPU_obj->aggr_add_upgrade_start(%args);

=head1 DESCRIPTION

NACL_IPU is an Object-Oriented class inherited NACL structure that
provided a business/application logic and promoted re-usability of
small, well defined IPU related functionality that can performed on
setup testbed.

=head1 Terminology

=over

=item * Application Logic

Code that represents functionality that is needed for test case flow.
Differentiated from code that's needed to handle interface or version
logic.

=item * Component - A modeling of an element.

Components can encapsulate mode abstractions and interface
abstractions as well as other NACL services.

=item * Element

An external entity to test or manipulate in tests. Examples are
volumes, aggregates, mounts, etc.

=item * APISet

Code that adds consistency and makes available to Perl programs an
programitic interface to some external service. An example service
would be the CLI for a 7M filer.

=back

=head1 Dependencies

Due to the perl coding limitation (subroutine cannot be started with
digit).  the Autoload mechanism cannot be applied to upgrade cli
commands (ie. 64bit-upgrade).  NACL-DEV are currently implementing
separated components (Volume64bitUpgrade and
StorageAggregate64bitUpgrade ) to accommodate this issue and they
should be available for the next sprint.

Currently the library was trying to work around the issue (Components
not ready) by implementing the upgrade related subroutines with
APISets, but they will be replaced with Components when the Components
are available.

=head1 COPYRIGHT 2010-2011 NetApp, Inc.  All rights reserved.

=head1 REQUIRES

Perl 5.6.0, Getopt::Std, Pod::Usage

=head1 Author

vchan,shenry

=head1 PREJOBS

If the testbed was setup for CMode, then gx_boot.thpl and
deux_init.thpl need to be run as prejobs.  The default aggregate
(generated automatically through deux_init.thpl) will not be used
since it will be a 64.bit aggregate.  All of the other volumes and
aggregates shall be purged before moving on to the next test.

=head1 PUBLIC METHODS

=cut


package IPU::NACL_IPU;

use strict;
use Tharn;
use TCD;
use Carp qw(croak);
use Params::Validate qw(validate validate_with OBJECT SCALAR HASHREF ARRAYREF CODEREF UNDEF);
use NATE::BaseException qw( :try );
use NATE::Error::Simple;        # workaround for NACL Exception bug.
use NACL::APISet;
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::CS::Snapmirror;
use NACL::ComponentUtils qw (_optional_scalars);
use NACL::C::Vserver;
use NACL::C::Aggregate;
use NACL::C::Volume;
use Data::Dumper;
use NACL::C::Snapshot;
use NACL::C::StorageAggregate64bitUpgrade;
use NACL::C::Volume64bitUpgrade;

BEGIN {

} ## end BEGIN

########################################################################
#                            Public Methods
########################################################################

=head2 B<_new()>

Creates a _new IPU object on filer.  Give default values to all
supported attributes.

=head2 Arguments

=over 4

=item filer

Hostspec of a filer/node, defaults to $::FILER

=item node

The command interface for NACL Components

=item aggr_number

How many aggregates you want to create

=item aggr_disk

Size of each aggregate, in number of disks

=item vserver

Default is the vserver that was created by deux_init

=item aggr_prefix

Prefix name for creating multiple aggregates

=item vol_number

Number of volumes you want to create per aggregate. Volume names are
based on the aggr name

=item vol_size

Volume size of each volume

=item block_type

"32" or "64" for all aggregates created; Default is 32-bit since this is
for IPU testing

=item api_set_obj

APISet object created by the test.

=item RETURN

Returns an IPU object.

=back

=cut

sub new {
    my $class = shift;
    my $self = validate (
        @_,
        {
            filer       => { type => SCALAR,   default  => $::FILER },
            node        => { type => HASHREF,  optional => 0 },
            aggr_number => { type => SCALAR,   default  => 1 },
            aggr_disk   => { type => SCALAR,   default  => $::AGGR_SIZE },
            vserver     => { type => SCALAR,   default  => 'vs0' },
            aggr_prefix => { type => SCALAR,   default  => "aggr_nacl" },
            vol_number  => { type => SCALAR,   default  => 2 },
            vol_size    => { type => SCALAR,   default  => $::FLEX_SIZE },
            block_type  => { type => SCALAR,   default  => 32 },
            api_set_obj => { type => OBJECT,   optional => 1 },
            vol_list    => { type => ARRAYREF, optional => 1 },
            snap_list   => { type => ARRAYREF, optional => 1 },
            aggr_list   => { type => ARRAYREF, optional => 1 },
       node_api_set_obj => { type => OBJECT,   optional => 1 },
        }
    );
    my @aggrs_created_here;

    #always set $self->{aggr_list} to empty since we haven't
    #made any aggrs yet.  store away the value.
    if ( defined $self->{aggr_list} ) {
        @aggrs_created_here = @ {$self->{aggr_list} };

        #keep the two lists of aggrs made in sync
        #_aggr_hash represents the aggrs created by this library
        #so we can clean them up later
        $self->{_aggr_hash} = {};
        $self->{aggr_list} = [];
    }
    #keep track of the volumes created
    $self->{_vol_hash} = {};
    $self->{vol_list} = [];

    #create tha api set object if one was not passed in
    #determine what the mode is.
    if (! defined $self->{api_set_obj} ) {
        my $mode = $self->{node}->mode();
        eval {
            my $api_set_obj = NACL::APISet->new(hostobj   => 
                                                       $self->{node}->{hostrec},
                                                category  => "Node",
                                                interface => "CLI",
                                                set       => $mode);
            $self->{api_set_obj} = $api_set_obj;
        };
        if ($@) {
            croak "Unable to create apiset object:$@";
        }
    }
    if (! defined $self->{node_api_set_obj} ) {
        eval {
            $self->{shell_type} = ($self->{node}->mode() eq "CMode") ? "Nodescope" : "7Mode";
            $self->{node_api_set_obj} = $self->{node}->get_7m_or_nodescope_apiset();
        };
    }
    bless $self, $class;

    #foreach element aggregate that the caller will want
    #made, create the aggr
    foreach my $aggr_to_create  (@aggrs_created_here) {
        $Log->comment("Create aggr $aggr_to_create");
        $self->create_aggr($aggr_to_create)
    }
    return $self;
}

=head2 B<ipu_setup()>

Create resources on filers based on the attributes initiated in
new().  This is for setting up the IPU testbed. It might not fit into your
testcase scenario.  in that case, you should use aggr_create() and
vol_create() to configure your own testbeds.

Any provided arguments override the same value provided in B<new()>

=head2 Arguments

=over 4

=item aggr_number

How many aggregates you want to create

=item aggr_disk

Size of each aggregate, in number of disks

=item vserver

Default is the vserver that is created by deux_init

=item aggr_prefix

Prefix name for creating multiple aggregates

=item vol_number

Number of volume you want to create per aggregate. Volume names are
based on the aggr name

=item vol_size

Volume size of each volume

=item block_type

"32" or "64" for all aggregates created; Default is 32-bit since this is
for IPU testing

=item RETURN

Return 1 if successful.  Croak if something went
wrong.

=back

=cut

sub ipu_setup {
    my $self = shift;
    my %opts  = validate(
        @_,
        {
            aggr_number => { type => SCALAR,   optional => 1 },
            aggr_disk   => { type => SCALAR,   optional => 1 },
            vserver     => { type => SCALAR,   optional => 1 },
            aggr_prefix => { type => SCALAR,   optional => 1 },
            vol_number  => { type => SCALAR,   optional => 1 },
            vol_size    => { type => SCALAR,   optional => 1 },
            block_type  => { type => SCALAR,   optional => 1 },
        });

    #allow overwriting of $self since ipu_setup is the main
    #controller function and hence subsequent will rely on
    #values set here.

    for my $k (keys %opts) {
        $self->{$k} = $opts{$k};
    }

    my $node_hosttype = host( $self->{filer} )->hosttype();
    if (  $node_hosttype =~ /ng/  ) {
        $Log->comment("###Checking if VSERVER exists");
        eval {
            NACL::C::Vserver->find(
                'command_interface' => $self->{node},
                'filter'            => { 'vserver' => $self->{vserver} }
            )->state->vserver();
        };
        if ($@) {
            $Log->error( "no default vserver exists, ".
                         "please run gx_boot and deux_init first" );
            die($@);
        } else {
            $Log->comment("Vsever $self->{vserver} exists");
        }
    }
    $Log->comment("###Create aggr & vols");
    eval {

        my @aggrs_created_here;

        #create aggrs using the $self->{aggr_prefix}
        for (my $i=0; $i<$self->{aggr_number}; $i++) {
            my $aggr = "$self->{aggr_prefix}$i";
            $Log->comment("Create aggr $aggr");
            my $ret = $self->create_aggr($aggr);
            die "Failed to create aggr $aggr" if ( !($ret) );
            push @aggrs_created_here, $aggr;
        }

        #make sure all the newly created aggrs are online
        foreach my $aggr ( @aggrs_created_here ) {
            my $ret = $self->_check_aggr_status( aggr => $aggr );
            if (! $ret ) {
                die "aggr $aggr is not able to be online before timeout";
            }
        }

        #create vol based on the list of aggr
        $Log->comment("###Create vols");
        foreach my $aggr (@aggrs_created_here) {
            $Log->comment("Create vol list ");
            for ( my $i = 1 ; $i <= $self->{vol_number}; $i++ ) {
                my $volname = $aggr . "_vol" . $i;
                my $ret = $self->create_vol( vol  => $volname,
                                             aggr => $aggr );
                die "not able to create vol $volname" if ( !($ret) );
            }
        }
    };
    croak "ERROR: $@" if ($@);
    return $self;
}

# not used? deprecated!
sub create_multiple_aggrs {
    my $self = shift;
    my $output;
    my @aggr_to_create = @{ $self->{aggr_list} };
    my @aggr;
    foreach my $index ( 0 .. $#aggr_to_create ) {
        my $aggr = $aggr_to_create[$index];
        $Log->comment("###Create aggregate $aggr if it does not exist");
        eval {
            @aggr = NACL::C::Aggregate->find(
                command_interface => $self->{node},
                filter            => { aggregate => $aggr }
            );
        };
        if ( !($@) ) {
            #remove the existing aggr
            $Log->comment("remove the existing aggr");
            $self->aggr_delete($aggr);
        }
        eval {
            $aggr[0] = NACL::C::Aggregate->create(
                aggregate         => $aggr,
                command_interface => $self->{node},
                diskcount         => $self->{aggr_disk},
                "block-type"      => $self->{block_type},
            );
        };    #end of eval of creation
        if ($@) {
            #### something went wrong during creation, remove from the list
            delete $aggr_to_create[$index];
        }

        #update the list of aggrs created
        $self->_hash_add_to_list('aggr',$aggr);
    }
    return;
}

=head2 B<create_aggr()>

$IPU_obj->create_aggr($aggr, [aggr_disk=>$n, block_type=>32|64,
                              node_name=>$node_name, existing=>'purge'|'reuse']);

create aggregates on nodes based on the attributes initiated in new().

=head2 Arguments

=over 4

=item $aggr

The aggregate name to create. Uses the values set in new() for other options.

=item aggr_disk

Optional, number of disks to use to create the aggregate

=item block_type

Optional, '32' or '64', block type of aggregate to create

=item node_name

Optional, node to create aggregate on, if different

=item existing

Optional, 'purge' the existing aggr before creating the new one,
'reuse' an existing one, or 'die' if it already exists.
The Default is 'purge'

=item RETURN

Returns an Aggregate Component object if successful.

=back


=cut

sub create_aggr {
    my $self = shift;
    my $aggr = shift;
    my %opts = validate (@_, {
        aggr_disk  => { type => SCALAR, default => $self->{aggr_disk} },
        block_type => { type => SCALAR, default => $self->{block_type} },
        node_name  => { type => SCALAR, optional => 1 },
        wait       => { type => SCALAR, default => 1 },
        existing   => { type => SCALAR, default => "purge",
                        regexp => qr/^purge$|^reuse$|^die$/ },
    } );

    my $aggr_obj;
    #see if aggr already exists, if so check $opts{existing} option
    #to determine what to do

    $Log->enter();
    eval {
        $aggr_obj = NACL::C::Aggregate->find(
            command_interface => $self->{node},
            allow_empty       => "true",
            filter            => { aggregate => $aggr }
        );
    };
    if ( $aggr_obj ) {
        if ($opts{existing} eq 'purge') {
            #remove the existing aggr
            $Log->comment("remove the existing aggregate $aggr");
            eval {
                $self->aggr_delete($aggr);
            };
            croak "Unable to delete before adding $aggr" if ($@);

            #update the list of aggrs created
            $self->_hash_delete_from_list('aggr',$aggr);

        } elsif ($opts{existing} eq 'reuse') {
            $Log->comment("reusing aggregate $aggr");
            $Log->exit();
            return $aggr_obj;
        } else {
            die $@;
        }
    }
    #now create the aggr
    eval {
        my $nodename = $opts{node_name}
                                        ? $opts{node_name}
                                        : $self->{node}->{hostrec}->{hostname};

        $Log->comment("creating a new aggregate $aggr on node $nodename");
        $aggr_obj = NACL::C::Aggregate->create(
            aggregate         => $aggr,
            command_interface => $self->{node},
            diskcount         => $self->{aggr_disk},
            nodes             => [ $nodename ],
            "block-type"      => $opts{block_type},
        );
        # now wait for completion
        if ($opts{wait}) {
            my $state;
            do {
                Tharn::sleep(5);
                $state = $aggr_obj->state(requested_fields=>[ 'state' ]);
            } while ($state->{state} ne "online");
        }
    };    #end of eval of creation
    croak "Unable to create single aggr $aggr: $@" if $@;

    #update the list of aggrs created
    $self->_hash_add_to_list('aggr',$aggr);
    $Log->exit();
    return $aggr_obj;
}

# not used? deprecated? DO NOT USE, NOT TESTED
sub create_multiple_vols {
    my $self   = shift;
    my ($aggr) = @_;
    my $count  = $self->{vol_number};

    $Log->enter();
    for ( my $i = 1 ; $i <= $count ; $i++ ) {

        my $volname = $aggr . "_vol" . $i;
        eval {
            my $vol = NACL::C::Volume->find(
                command_interface => $self->{node},
                filter            => { volume  => $volname,
                                       vserver => $self->{vserver}
                                     },
            );
        };
        if ( !($@) ) {
            NACL::C::Volume->modify(
                command_interface => $self->{node},
                vserver           => $self->{vserver},
                volume            => $volname,
                state             => "offline",
            );
            NACL::C::Volume->delete(
                command_interface => $self->{node},
                volume            => "$volname",
                vserver           => $self->{vserver},
            );

            #now update the list of volumes
            $self->_hash_delete_from_list('vol', $volname);
        }

        eval {
            NACL::C::Volume->create(
                volume            => $volname,
                vserver           => $self->{vserver},
                aggregate         => $aggr,
                command_interface => $self->{node},
                size              => $self->{vol_size}
            );
        };
        croak "Unable to create vole $volname while creating mutliple: $@" if ($@);

        #update the list of volumes created
        $self->_hash_add_to_list('vol',$volname);
    }

    croak("failed to create vol $@") if ($@);

    $Log->exit();
    return;

}

=head2 B<create_vol()>

Create volumes on aggregates based on the attributes initiated in new().

=head2 Arguments

=over 4

=item vol

Required. Volume name to create.

=item aggr

Required. Aggregate to host the volume.

=item 'snapshot-policy'

Optional. What to do with snapshot-policy for c-mode, or nosnap option for 7-mode
The default is whatever default by ONTAP.

=item existing

Optional. What to do with pre-existing volumes: 'purge', 'reuse', or 'die'
The default is 'purge'.

=item RETURN

Returns a Volume Component object if successful.

=back

=cut

sub create_vol {
    $Log->enter();
    my $self = shift;
    my (%opts) = validate( @_, {
        vol                => { type => SCALAR },
        aggr               => { type => SCALAR },
        'snapshot-policy'  => { type => SCALAR, optional => 1 },
        existing           => { type => SCALAR, default => 'purge',
                                    regexp => qr/^purge$|^reuse$|^die$/ },
        type               => { type => SCALAR, optional => 1 },
    } );
    my $aggr    = $opts{aggr};
    my $volname = $opts{vol};
    my $vol_obj;

    $Log->enter();

    my %vol_create_opt = (
        volume            => $volname,
        vserver           => $self->{vserver},
        aggregate         => $aggr,
        command_interface => $self->{node},
        size              => $self->{vol_size},
        "junction-path"   => "/".$volname,
        "unix-permissions" => "---rwxrwxrwx",
    );

    if ( defined($opts{'snapshot-policy'}) ) {
        $vol_create_opt{'snapshot-policy'} = $opts{'snapshot-policy'};
    }
    if ( defined($opts{type}) ) {
        $vol_create_opt{type} = $opts{type};
        if ($opts{type} ne 'RW') {
            delete $vol_create_opt{'junction-path'};
            delete $vol_create_opt{'unix-permissions'};
            delete $vol_create_opt{'snapshot-policy'} if $vol_create_opt{'snapshot-policy'};
        }
    }

    try {
        $vol_obj = NACL::C::Volume->find(
            command_interface => $self->{node},
            filter            => { volume  => $volname,
                                   vserver => $self->{vserver} },
        );
    } catch NATE::BaseException with {
        my $e = shift;
        $Log->debug("Volume not found, so we will create a new one");
    };
    if ($vol_obj) {
        if ($opts{existing} eq 'purge') {
            $self->_vol_obj_delete( [ $vol_obj ] );
        } elsif ($opts{existing} eq 'reuse') {
            return $vol_obj;
        } else {
            die "Volume $volname already exists";
        }
    }

    eval {
        $vol_obj = NACL::C::Volume->create( %vol_create_opt );
    };
    croak "Unable to create single volume $volname: $@" if ($@);

    #update the list volumes created
    $self->_hash_add_to_list('vol',$volname);

    $Log->exit();
    return $vol_obj;
}

sub aggrs_delete {
    my $self       = shift;
    my ($aggrs)    = @_;
    my @aggrs;
    my $error_flag = 1;

    $Log->enter();

    if (!defined $aggrs) {
        @aggrs = @ {$self->{aggr_list} };
    } 
    else {
        @aggrs      = @{$aggrs};
    }

    foreach my $aggr (@aggrs) {
        if (defined($aggr)) {
            eval { $self->aggr_delete($aggr); };
            if ($@) {
                $Log->comment("Failed to delete aggr $aggr: $@");
                #set the error flag
                $error_flag = 0;
            }
        }
    }    #foreach

    $Log->exit();
    return $error_flag;
}

sub aggr_delete {
    my $self = shift;
    my ($aggr) = @_;
    my $aggr_obj;
    my @vols;

    $Log->enter();
    eval {
        $aggr_obj = NACL::C::Aggregate->find(
            command_interface => $self->{node},
            filter            => { aggregate => $aggr }
        );
    };
    ##### find all the vols in the container and delete them first ########
    if ( defined( $aggr_obj ) ) {
        my $state = $aggr_obj->state();
        if ($state->state() ne "online") {
            $aggr_obj->modify(state => "online");
        }
        eval {
            foreach my $i (1..5) {
                eval {
                    @vols = NACL::C::Volume->find(
                        command_interface => $self->{node},
                        filter            => { aggregate => $aggr ,
                                               vserver   => $self->{vserver},
                                             },
                        allow_empty       => "true",
                    );
                    last unless (@vols);
                    $self->_vol_obj_delete( \@vols );
                    Tharn::sleep(5);
                };
            }
            $aggr_obj->modify( state => "offline" );
            $aggr_obj->delete();

            #update the list of aggrs created in library
            $self->_hash_delete_from_list('aggr',$aggr);
        };
    }    #if

    if ($@) {
        croak "aggr_delete() fails: $@";
    } 

    $Log->exit();
    return;
}


sub _vol_obj_delete {
    my $self       = shift;
    my ($ref_vols) = @_;
    my @vols       = @{$ref_vols};
    my $mode       = $self->{node}->mode();

    foreach my $vol (@vols) {

        my $volname = $vol->volume();
        $Log->comment("volume obj is being deleted, the name is $volname");

        try {
            $Log->debug("**** unmount vol $vol");
           #vol unmount is Cmode specific and hence the check
            if ($mode eq "CMode"){
                $vol->unmount();
            }
            $Log->debug("**** offline vol $vol");
            $vol->modify( state => "offline" );
            $Log->debug("**** delete vol $vol");
            $vol->delete();

            #update the list of vols created in library
            $self->_hash_delete_from_list('vol', $volname);

        } catch NATE::BaseException with {
            my $e = shift;
            $Log->comment("---- problem purging volume: ".$e->text());
        };
    }
    return;

}

#single subroutine to add to either the list of aggrs or the
#the list of volumes.
sub _hash_add_to_list {
    my $self = shift;
    my ($which, $value) = @_;

    my $hash_key;
    my $list_key;

    #must be vol or aggr since var is used to create the hash
    $which = lc $which;
    if ( ($which ne "vol") &&
         ($which ne "aggr") )  {
        die "Must provide 'vol' or 'aggr' in delete";
    }

    $hash_key = "_${which}_hash"; #should be _vol_hash or _aggr_hash
    $list_key = "${which}_list";  #shoudl be vol_list or aggr_list

    #add to the hash by setting the value in the hash to 1
    $self->{$hash_key}{$value} = 1;

    #now get the keys, this ensures a single entry in list for
    #each aggr or vol

    $self->{$list_key} = [ keys % { $self->{$hash_key} } ];
    #TODO: sort the keys in alphabetical order so the return
    #is somewhat predictable
    #$self->{$list_key} = sort [ keys % { $self->{$hash_key} } ];
}

sub _hash_delete_from_list {
    my $self = shift;
    my ($which, $value) = @_;

    my $hash_key;
    my $list_key;

    #must be vol or aggr since use to create the hash
    $which = lc $which;
    if ( ($which ne "vol") &&
         ($which ne "aggr") )  {
        die "Must provide 'vol' or 'aggr' in delete";
    }

    $hash_key = "_${which}_hash"; #should be _vol_hash or _aggr_hash
    $list_key = "${which}_list";  #shoudl be vol_list or aggr_list

    #get rid of the value in the hash
    delete $self->{$hash_key}{$value};

    #keep the list up to date w/ the hash
    $self->{$list_key} = [ keys % { $self->{$hash_key} } ];
}

=head2 B<vols_delete()>

Delete the list of volumes

=head2 Arguments

=over 4

=item $vols

Array reference to a list of volume names to be destroyed.

=item vserver => 'vs0'

vserver where the volumes reside, defaulted to $self->{vserver}

=item RETURN

Errors are trapped; Nothing is returned.

=back

=cut

sub vols_delete {
    my $self  = shift;
    my $vols  = shift;
    my @vols;

    my %opts = validate(@_, { vserver => { type => SCALAR,
                                           default => $self->{vserver} }
                            }
                       );

    $Log->enter();
    if (! defined $vols) {
        @vols = @{$self->{vol_list}};
    }
    else {
        @vols   = @{$vols};
    }

    my $vserver = $opts{vserver};

    while ( $#vols >= 0 ) {
        my $vol = pop(@vols);
        try {
            my $vobj = NACL::C::Volume->find(
                command_interface => $self->{node},
                filter  => { vserver => $vserver,
                             volume  => $vol, },
            );
            $vobj->unmount();
            $vobj->modify( state => "offline" );
            $vobj->delete();

            #update the list of vols created in library
            $self->_hash_delete_from_list('vol', $vol);

        } catch NATE::BaseException with {
            $Log->comment("Error to delete $vol");
        };
    }
    $Log->exit();
    return;
}

=head2 B<ipu_cleanup()>

Destroy created resources on filers.  This includes aggrs, vols,
files, snapshots...etc.

=head2 Arguments

=over 4

=item RETURN

Return 1 if successful.  If anything goes wrong
during the purge, the error will be caught.

=back

=cut

sub ipu_cleanup {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            vols    => { type => ARRAYREF, default => $self->{vol_list} },
            vserver => { type => SCALAR, default => $self->{vserver} },
            aggrs   => { type => ARRAYREF, default => $self->{aggr_list} },
        }
    );
    $Log->enter();

    try {
        $self->vols_delete( $opts{vols},
                            vserver => $opts{vserver} );
    }
    catch NATE::Error::Simple with {
        my $e = shift;
        $Log->warn( "in ipu_cleanup: " . $e->text() );
    };

    try {
        $self->aggrs_delete( $opts{aggrs} );
    }
    catch NATE::Error::Simple with {
        my $e = shift;
        $Log->warn( "in ipu_cleanup: " . $e->text() );
    };
    $Log->exit();
    return;

}

=head2 B<aggr_upgrade_scan_start()>

fetch cli cmd "aggr 64bit upgrade scan start"

=head2 Arguments

=over 4

=item aggr

aggregate name

=item RETURN

Return if successful else throw an exception

=back

=cut

sub aggr_upgrade_scan_start {
    my $self = shift;
    my (%opts) = validate( @_, { aggr => { type => SCALAR }, } );
    my $aggr = $opts{aggr};

    $Log->enter();
    $Log->comment("aggr 64bit upgrade scan start ");
    NACL::C::StorageAggregate64bitUpgrade->scan_start(
                command_interface => $self->{node},
                aggregate         => $aggr );
    $Log->exit();
    return;
}

=head2 B<aggr_upgrade_scan_stop()>

Fetch result of cli cmd "aggr 64bit upgrade scan stop"

=head2 Arguments

=over 4

=item aggr

aggregate name

=item RETURN

Return if successful else throw an exception

=back

=cut

sub aggr_upgrade_scan_stop {
    my $self = shift;
    my (%opts) = validate( @_, { aggr => { type => SCALAR }, } );
    my $aggr = $opts{aggr};

    $Log->enter();
    $Log->comment("aggr 64bit upgrade scan stop ");
     NACL::C::StorageAggregate64bitUpgrade->scan_stop(
             command_interface  => $self->{node},
             aggregate          => $aggr );
    $Log->exit();
    return;
}

=head2 B<aggr_upgrade_diag()>

Fetch result of cli cmd "aggr 64bit upgrade start -mode <>"

=head2 Arguments

=over 4

=item aggr

Required. aggregate name

=item mode

Optional. One of 'check', 'normal', 'grow-reserved', 'grow-all'.
Default is 'normal'.

=item RETURN

Return check output if mode="check"
Return if successful else throw an exception.

=back

=cut

sub aggr_upgrade_diag {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            aggr => { type => SCALAR },
            mode => { type => SCALAR, default => "normal" },
        }
    );
    my @vols_upgrade_info;
    my $aggr = $opts{aggr};
    my $mode = $opts{mode};

    $Log->enter();
    $Log->comment("Start aggr 64-bit upgrade in mode $mode");
    if ($mode eq "check") {
        @vols_upgrade_info =
        NACL::C::StorageAggregate64bitUpgrade->start_check(
                command_interface  => $self->{node},
                        aggregate  => $aggr );
        $Log->trace("Dumping the check output". Dumper(@vols_upgrade_info));
        return \@vols_upgrade_info ;
    } else {
        NACL::C::StorageAggregate64bitUpgrade->start(
                command_interface  => $self->{node},
                        aggregate  => $aggr ,
                        mode       => $mode);
    }
    $Log->exit();
    return ;
}

=head2 B<aggr_add_upgrade_start()>

Fetch result of cli cmd "aggr add-disks  -diskcount <number> -64bit-upgrade <>"

=head2 Arguments

=over 4

=item aggr

Required. aggregate name

=item mode

Optional. One of 'check', 'normal', 'grow-reserved', 'grow-all'.
Default is 'normal'.

=item diskcount

Required. Number of disks that needs to be added to the aggr

=item RETURN

Return check output if mode="check"
Return if successful else throw an exception.

=back

=cut
sub aggr_add_upgrade_start{
    my $self = shift;
    my (%opts) = validate_with (
        params => \@_,
        spec => {
            aggr           => { type => SCALAR },
            diskcount      => { type => SCALAR },
            mode           => { type => SCALAR, default => "normal" },
        },
        allow_extra   => 1,
    );
    my @vols_upgrade_info;
    my $mode = delete $opts{mode};
    $opts{aggregate} = delete $opts{aggr};

    #TEMP_FIX: dirty workaround until NACL burt 517225 gets fixed
    my %zapi_hash = ( 'normal'        => 'grow_none',
                      'grow-reserved' => 'grow_reserved',
                      'grow-all'      => 'grow_all');

    $Log->enter();
    $Log->comment("Start aggr add 64-bit upgrade in mode $mode");
    if ($mode eq "check") {
        @vols_upgrade_info =
            NACL::C::Aggregate->add_disks_64bitupgrade_check(
                command_interface  => $self->{node},
                %opts,);
        $Log->trace("Dumping the check output". Dumper(@vols_upgrade_info));
        return \@vols_upgrade_info ;
    } else {
        #TEMP_FIX: dirty workaround until NACL burt 517225 gets fixed
        if ( defined($opts{apiset_should}{interface}) &&
             ($opts{apiset_should}{interface} eq 'ZAPI') ) {
            $Log->comment("upgrade-mode was $mode");
            $mode = $zapi_hash{$mode};
            $Log->comment("Running ZAPI, upgrade-mode changed to $mode");
        }
        NACL::C::Aggregate->add_disks(
                command_interface  => $self->{node},
                "64bit-upgrade"  => $mode,
                %opts,);
    }
    $Log->exit();
    return ;
}

=head2 B<vol_upgrade_diag()>

Fetch result of cli cmd "volume 64bit upgrade start -mode <>"

=head2 Arguments

=over 4

=item vol

volume name

=item mode

Optional; Upgrade mode, default is "normal"

=item RETURN

Returns check output if mode is check
Return if successful else throw an exception

=back

=cut

sub vol_upgrade_diag {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            vol  => { type => SCALAR },
            mode => { type => SCALAR, default => "normal" },
            vserver => { type => SCALAR, default => $self->{vserver} }
        }
    );
    my $vol  = $opts{vol};
    my $mode = $opts{mode};
    my $vserver = $opts{vserver};
    my $vols_upgrade_info;

    $Log->enter();
    $Log->comment("run vol 64-bit upgrade in $mode mode");
    if ($mode eq "check") {
        $vols_upgrade_info =
            NACL::C::Volume64bitUpgrade->start_check(
                command_interface => $self->{node},
                        volume    => $vol,
                        vserver   => $vserver);
        $Log->trace("Dumping the check output". Dumper($vols_upgrade_info));
        return $vols_upgrade_info ;
    } else {
        NACL::C::Volume64bitUpgrade->start(
                command_interface  => $self->{node},
                        volume     => $vol ,
                        vserver    => $vserver,
                        mode       => $mode);
    }
    $Log->exit();
    return ;
}

=head2 B<vol_is_64bit()>

Checks the block-type of the volume

=head2 Arguments

=over 4

=item vol

volume name

=item RETURN

Returns 1 if the volume is 64 bit, 0 if not

=back

=cut

sub vol_is_64bit {
    my $self = shift;
    my (%opts) = validate(
        @_,
        { vol     => { type => SCALAR },
          vserver => { type => SCALAR, default => $self->{vserver} }
        }
    );
    my $volname = $opts{vol};
    my $vserver = $opts{vserver};

    $Log->enter();
    my $vol = NACL::C::Volume->find(
        command_interface   => $self->{node},
        filter              => {
            volume          => $volname,
            vserver         => $vserver });
    my $format = $self->get_upgrade_status(
            aggr            => $vol->state()->aggregate(),
            container_name  => $volname,
            field           => "format");
    return 1 if ($format =~ /64-bit/) ;
    $Log->exit();

    return 0;
}
=head2 B<vol_is_mixed()>

Checks the block-type of the volume

=head2 Arguments

=over 4

=item vol

volume name

=item RETURN

Return 1 if the volume is in mixed format, return 0 if it's not

=back

=cut

sub vol_is_mixed {
    my $self = shift;
    my (%opts) = validate(
        @_,
        { vol     => { type => SCALAR },
          vserver => { type => SCALAR, default => $self->{vserver} }
        }
    );
    my $volname = $opts{vol};
    my $vserver = $opts{vserver};

    $Log->enter();
    my $vol = NACL::C::Volume->find(
        command_interface   => $self->{node},
        filter              => {
            volume          => $volname,
            vserver         => $vserver });
    my $format = $self->get_upgrade_status(
            aggr            => $vol->state()->aggregate(),
            container_name  => $volname,
            field           => "format");
    return 1 if ($format =~ /upgrading/) ;
    $Log->exit();

    return 0;
}

sub _check_aggr_status {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            aggr    => { type => SCALAR },
            status  => { type => SCALAR, default => "online" },
            timeout => { type => SCALAR, default => 600 },
        }
    );
    my @aggr;
    my $last   = $opts{timeout};
    my $ret    = 1;
    my $status = $opts{status};
    my $my_status;
    my $aggr = $opts{aggr};
    eval {
        @aggr = NACL::C::Aggregate->find(
            command_interface => $self->{node},
            filter            => { aggregate => $aggr }
        );
    };
    croak "Cannot locate the aggr $aggr: $@" if ($@);
    $my_status = $aggr[0]->state()->state();

    while ( ( !( $my_status =~ /$status/i ) ) && ( $last > 1 ) ) {
        $my_status = $aggr[0]->state()->state();
        Tharn::sleep(30);
        $last -= 30;
    }
    $ret = 0 if ( !( $my_status =~ /$status/i ) );
    return $ret;

}

=head2 B<wait_aggrupgrade_status()>

Wait until the aggregate has finished upgrading. If vol is passed, this
subroutine will return true when the volume is converted to the specified
format (obj); otherwise, it will return true when the aggregate and all of its
volumes are converted to the specified format.

=head2 Arguments

=over

=item obj

Aggr format to wait for: '64-bit', 'Upgrading', '32-bit'

=item aggr

Aggregate name

=item vol

Optional.  Volume name within the aggregate whose status is being checked.

=item timeout

Optional. Time to wait, default is $::TIMEOUT.

=item RETURN

1 for success, 0 for failure/timeout

=back

=cut

sub wait_aggrupgrade_status {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            obj     => { type => SCALAR },
            aggr    => { type => SCALAR },
            vol     => { type => SCALAR, optional => 1 },
            timeout => { type => SCALAR, default  => $::TIMEOUT },
        }
    );
    my $obj        = $opts{obj};    # "32-bit", "64-bit", or "upgrading".
    my $aggr       = $opts{aggr};
    my $vol        = $opts{vol};
    my $timeout    = $opts{timeout};
    my $endtime    = $timeout + time();
    my $done       = 0;

    $Log->enter();

    # If vol is passed, get_upgrade_status() will return true when the volume
    # is converted to 64-bits; otherwise, get_upgrade_status() will return true
    # when the aggregate and all of its volumes are converted to 64-bits.

    my %upgrade_opts = (
        aggr        => $aggr,
        field       => "format",
        value       => $obj);
    if (defined($vol)) {
        $upgrade_opts{container_name} = $vol;
    }

    while (!$done && (time() < $endtime)) {
        $done = $self->get_upgrade_status(%upgrade_opts);
        Tharn::sleep(10);
    }

    if ($done) {
        $Log->debug("Upgrade status is $obj.");
    } else {
        $Log->debug("Timed out waiting for upgrade status to be $obj.");
    }
    $Log->exit();
    return ($done);
}

=head2 B<wait_volupgrade_status()>

Wait until the volume has finished upgrading

=head2 Arguments

=over

=item obj

Volume format to wait for: '64-bit', 'Upgrading', '32-bit'

=item aggr

Aggregate name

=item vol

Optional.  Volume name whose status is being checked.

=item timeout

Optional. Time to wait, default is $::TIMEOUT.

=item RETURN

1 for success, 0 for failure/timeout.  Note that an exception will
be thrown if volume completes upgrade during this subroutine.

=back

=cut
sub wait_volupgrade_status {
    my $self = shift;
    my (%opts) = validate(
        @_,
        {
            obj     => { type => SCALAR },
            vol     => { type => SCALAR },
            timeout => { type => SCALAR, default => $::TIMEOUT },
        }
    );
    my $obj        = $opts{obj};
    my $volname    = $opts{vol};
    my $timeout    = $opts{timeout};
    my $endtime    = $timeout + time();
    my $done       = 0;

    $Log->enter();

    my $vol = NACL::C::Volume->find(
        command_interface   => $self->{node},
        filter              => {
            volume          => $volname,
            vserver         => $self->{vserver}});

    while (!$done && (time() < $endtime)) {
        $done = $self->get_upgrade_status(
            aggr            => $vol->state()->aggregate(),
            container_name  => $volname,
            field           => "format",
            value           => $obj);

        Tharn::sleep(10);
    }

    if ($done) {
        $Log->debug("Upgrade status is $obj.");
    } else {
        $Log->debug("Timed out waiting for upgrade status $obj.");
    }
    $Log->exit();
    return ($done);
}
=head2 B<disable_IPU_scanner()>

Disables the upgrade scanner on the node $node

=head2 arguments

=over

=item node

The node on which to disable the upgrade scanner

=item RETURN

1 on success, the underlying exception is raised on failure.

=back

=cut

sub disable_IPU_scanner {
    my $self = shift;
    my (%args) = validate(@_, {
        node => { TYPE => SCALAR },
    });
    my $node = $args{node};
    my $nodename = ref($node) ? $node->name() : $node;
    my $cmd;

    $Log->enter();
    $cmd = "priv set -q test;" .
           "setflag wafl_64bit_upgrade_scan_disable 1";

    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node $nodename -command \"" . $cmd .'"';
    }
    my $response = $self->{node_api_set_obj}->execute_raw_command(command => $cmd);
    $Log->trace("_execute_command response: $response");
    if ($response =~ /Error:/) {
        NATE::BaseException->throw("Error executing $cmd " .
                                   "response is $response");
    }
    $Log->exit();
    return 1;
}


=head2 B<enable_IPU_scanner()>

Enable the upgrade scanner on the node $node

=head2 Arguments

=over

=item node

The node on which to enable the upgrade scanner

=item RETURN

1 on success, the underlying exception is raised on failure.

=back

=cut

sub enable_IPU_scanner {
    my $self = shift;
    my (%args) = validate(@_, {
        node => { TYPE => SCALAR },
    });
    my $node = $args{node};
    my $nodename = ref($node) ? $node->name() : $node;
    my $cmd;

    $Log->enter();
    $cmd = "priv set -q test;" .
           "setflag wafl_64bit_upgrade_scan_disable 0";

    if ($self->{shell_type} eq "Nodescope") {
        $cmd = "system node run -node $nodename -command \"" . $cmd .'"';
    }
    my $response = $self->{node_api_set_obj}->execute_raw_command(command => $cmd);
    $Log->trace("_execute_command response: $response");
    if ($response =~ /Error:/) {
        NATE::BaseException->throw("Error executing $cmd " .
                                   "response is $response");
    }
    $Log->exit();
    return 1;
}


=head2 B<get_upgrade_status()>

This subroutine calls the "storage aggregate 64bit-upgrade status" command
(or its counterpart in 7-mode) and returns the results.

=head2 Arguments

=over 4

=item aggr (required)

The name of the aggregate where you want to get the status information about
the upgrade scan from.

=item container_name (optional)

The name of the agregate or volume where you want to get the status information
about the upgrade scan from. This will restrict the information being returned
to that object; otherwise, you will get information about the aggregate and all
its volumes.

If the name of the aggregate is the same as the name of one of the volumes
and it is passed to this parameter, this subroutine will assume it is the name
of the aggregate.  Note: Return type 6 (see below) will return an error if
there is a volume with the same name as the aggregate.

=item field (optional)

This is the name of the field that you are interested in. It can be any of the
following strings.

=over

    "aggregate"
    "format"
    "scan_percent_completed"
    "scan_progress"
    "scan_time_to_complete"
    "scanner_status"
    "volume"
    "vserver"

=back

=item value (optional)

This is the value that you expect for a given field. If this value is passed,
the return value will be true or false to indicate whether the this value
matches the in the field. For example, possible values for the "format" field
are "32-bit", "64-bit", or "upgrading".

=back

=over 4

=item Return

This subroutine will return different information depending upon which optional
parameters are passed. The following table summaries the combinations of
parameters.

    #  Agg  Obj  Fld  Val
    ---------------------
    1   X                   Return hash of aggregate information
    2   X    X              Return hash of values for specified object
    3   X    X    X         Return value of field
    4   X    X    X    X    Return true or false if field = value
    5   X    X         X    Error
    6   X         X         Return hash of field indexed by object
    7   X         X    X    Return true if field = value in all objects
    8   X              X    Error

If there is an error, this subroutine will throw an exception with a
wonderful error message.

The following items provide more detailed information on each of the types of
information returned by this subroutine.

=over

=item 1.

If only the aggregate name is passed (object name, field name, and value are
not passed), this subroutine will return a reference to a hash that contains
the information about the aggregate. One of the fields, volumes_64bit, is a
reference to an array of hashes, where each hash contains information about
one of the volumes in the aggregate.

For example:

    $out = $IPU_obj->get_upgrade_status(aggr => $Aggr->aggregate());

    my @vols;
    # Print the aggregate's attributes.
    foreach my $key (keys %{$out}) {
        if ($key eq "volumes_64bit") {
            @vols = @{$out->{$key}};
        } else {
            $Test->nlog("ATTR Key: $key = $out->{$key}");
        }
    }

    # Print the attributes for each volume.
    $Test->nlog("");
    foreach my $vol (@vols) {
        foreach my $key (keys %{$vol}) {
            $Test->nlog("VOL Key: $key = $vol->{$key}");
        }
        $Test->nlog("");
    }

=item 2.

If the object name is specified, but not the field or value parameter, this
subroutine will return a reference to a hash containing all the fields for that
object. Normally, you would specify the name of one of the volumes, but you can
specify the name of the aggregate, in which case, you will get the same values
as return type 1.

For example:

    $out = $IPU_obj->get_upgrade_status(
        aggr            => $Aggr->aggregate(),
        container_name  => $Vol1->volume());

    foreach my $key (keys %{$out}) {
        $Test->nlog("$key = $out->{$key}");
    }

=item 3.

If the object name and the field name are specified, but not the value,
this subroutine will return the value of the field for that object.

For example:

    $out = $IPU_obj->get_upgrade_status(
        aggr            => $Aggr->aggregate(),
        container_name  => $Vol1->volume(),
        field           => "format");

    $Test->nlog("Format = $out");

=item 4.

If the object name, field name, and value are all specified, this subroutine
will compare the value of the field with the value passed and return true or
false to indicate if the values are the same.

For example:

    $out = $IPU_obj->get_upgrade_status(
        aggr            => $Aggr->aggregate(),
        container_name  => $Vol1->volume(),
        field           => "format",
        value           => "32-bit");

    $Test->nlog("Format is \"32-bit\" = $out");

=item 5.

The value parameter is only used if the field parameter is also used; otherwise
this subroutine will return an error.

=item 6.

If only the aggregate and field name are passed, this subroutine will return a
hash that contains an entry (key) for each object and the value of the entry
will be the value of the field for that object.

Note: since the name of the aggregate or volume is used as the hash key, if
there is a volume with the same name as the aggregate when this return type is
used, this subroutine return an error.

For example:

    $out = $IPU_obj->get_upgrade_status(
        aggr        => $Aggr->aggregate(),
        field       => "format");

    foreach my $key (keys %{$out}) {
        $Test->nlog("$key = $out->{$key}");
    }

=item 7.

If the field name and value are specified, but not the object, this subroutine
will compare the value of the field in each object with the value passed. If
the value of the field in every object is the same as the value passed, this
function will return true, otherwise it will return false.

For example:

    $out = $IPU_obj->get_upgrade_status(
        aggr        => $Aggr->aggregate(),
        field       => "format",
        value       => "32-bit");

    $Test->nlog("All objects are 32-bit format = $out");

=item 8.

The value parameter is only used if the field parameter is also used; otherwise
this subroutine will return an error.

=back

=cut

sub get_upgrade_status {
    my $self = shift;
    my %opts = validate(@_, {
        aggr            => {type => SCALAR},
        container_name  => {type => SCALAR, optional => 1},
        field           => {type => SCALAR, optional => 1},
        value           => {type => SCALAR, optional => 1},
    });

    my $aggr            = $opts{aggr};
    my $container_name  = $opts{container_name};
    my $field           = $opts{field};
    my $value           = $opts{value};

    my $out;            # Output of command.
    my %ret_val;        # For return type 6.
    my $container_type  = "volume";

    $Log->debug("Parameter: aggr            = $aggr");
    $Log->debug("Parameter: container_name  = $container_name");
    $Log->debug("Parameter: field           = $field");
    $Log->debug("Parameter: value           = $value");

    if (defined($value) && !defined($field)) {
        # Return types 5 and 8.
        NACL::APISet::Exceptions::InvalidParamException->throw(
            text => "The value parameter is only used with the field parameter.");
    }

    $out = NACL::CS::StorageAggregate64bitUpgrade->fetch(
        command_interface           => $self->{node},
        filter => {
            aggregate               => $aggr,
            'include-all-volumes'   => "true"});

    $Data::Dumper::Maxdepth = 3;
    $Log->debug("Raw output using Dumper(): \n" . Dumper($out));

    if(!defined($container_name) && !defined($field) && !defined($value)) {
        return ($out);  # Return type 1.
    }

    if (defined($container_name) && ($container_name eq $aggr)) {
        $container_type = "aggregate";
    }

    # Search the aggregate and each of the volumes.
    foreach my $obj ($out, @{$out->{volumes_64bit}}) {
        if (defined $container_name) {
            if ($obj->{$container_type} eq $container_name) {
                if (defined($field) && defined($value)) {
                    if ($obj->{$field} ne $value) {
                        return 0;   # Return type 4.
                    } else {
                        return 1;   # Return type 4.
                    }
                } elsif (defined($field) && !defined($value)) {
                    return ($obj->{$field}); # Return type 3.
                }else {
                    # Neither field nor value passed.
                    return($obj);    # Return type 2.
                }
            }
        } elsif (defined($value) && defined($field)) {
            if ($obj->{$field} ne $value) {
                $Log->debug("Return type 7: container = $container_name, " .
                    "expected value = $value, actual value = $obj->{$field}");
                return 0;   # Return type 7.
            }
        } elsif (defined($field) && !defined($value)) {
            if ($obj->{aggregate}) {
                $ret_val{$obj->{aggregate}} = $obj->{$field};
            } else {
                if ($obj->{volume} eq $aggr) {
                    NACL::APISet::Exceptions::InvalidParamException->throw(
                        text => "When only the aggregate and field " .
                        "parameters are passed, there cannot be a volume " .
                        "with the same name as the aggregate.");
                }
                $ret_val{$obj->{volume}} = $obj->{$field};
            }
        }
    }

    if (!defined($container_name) && defined($field) && !defined($value)){
        return \%ret_val;   # Return type 6.
    }

    return 1;   # Return type 7.
}

1;
