# This is basically -*- perl -*-
# $Id: ontap/main/iconnect/lib/Command_Testing.pm
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#  Iconnect/Command_Testing.pm
##
## @summary
##    Library for Command Testing Scripts.
##
## @description
##    This library is for the Command Testing suite of scripts.
##
##    This file defines the subroutines and expected usage statements for the
##    'ic' family of ONTAP commands.  The subroutines in this file do the work
##    for the test cases for comparing expected and received usage statements,
##    as well as comparing expected and received output from command execution.
##
## @synopsis
##   use Iconnect::Command_Testing qw(:ALL);
#
#
##
## @example
##   use Iconnect::Command_Testing qw(:ALL);
##
## @burt 682902 Command_Testing Suite check-in
## @burt 785024 Move Command_Testing libs to /lib/Iconnect/
##
## @keywords
##   interconnect; ic; Command_Testing
##
## @status public
##
## @author Interconnect QA (dl-interconnect-qa@netapp.com)
## @author Paul Waters (pwaters@netapp.com)

use strict;
use warnings;

package Iconnect::Command_Testing;
use Exporter 'import';
use TCD;

our @EXPORT_OK = qw(
    %USAGE_TABLE
    driver_type
    get_driver_type
    get_APISet_and_mode
    get_ontap_name
    compare_usage_statements
    validate_cdef_parsed_values
);
our %EXPORT_TAGS = (
    ALL => [qw(
        %USAGE_TABLE
        driver_type
        get_driver_type
        get_APISet_and_mode
        get_ontap_name
        compare_usage_statements
        validate_cdef_parsed_values
    )],
);

use Iconnect::Command_Testing_Usage qw(:ALL);
use List::MoreUtils qw{any};
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::ParamSet;
use Readonly;

my $Log = log_global();

################################################################################
#  Name        : VERBOSE
#  Description : Enables verbose output during string comparisons
#  To Use      : Set nate env varable VERBOSE=1
################################################################################
my $VERBOSE = NATE::ParamSet->new(global=>1)->get('VERBOSE', default=>0);

################################################################################
#  Name        : PRIVS
#  Description : Array to hold know privilege levels
################################################################################
Readonly our @PRIVS => qw/admin advanced diag test/;

################################################################################
# compare_usage_statements()
# ----------
# Does the legwork for Command Refactoring Test.
# Input:
#    command     => (Required) command to execute
#    driver      => (Required) driver type (ConnectX, Iwarp, etc).
#    ontap_ver   => (Required) ONTAP version (FS0, SN1, etc)
#    kernel_type => (Required) DEBUG or NONDEBUG
#    controller  => (Required) APISet object for controller
#    privs       => (Optional) arrayref - list of privs to execute against
#    usage       => (Optional) hashref  - alternate expected usage string, by
#                                         priv level
# Returns:
#    scalar - $TCD::<status>
################################################################################
sub compare_usage_statements {
    my %args      = @_;
    my $fails     = 0;
    my ($full_cmd, $sub_cmd, $driver, $ontap_ver, $model, $apiset, $kernel_type,
        $alt_usage, );

    # Get user supplied args (required)
    if (!defined($args{command})) {
        $Log->error('\'command\' argument not supplied.');
        return $TCD::CONFIG;
    } elsif (!defined($args{driver})) {
        $Log->error('\'driver\' argument not supplied.');
        return $TCD::CONFIG;
    } elsif (!defined($args{ontap_ver})) {
        $Log->error('\'ontap_ver\' argument not supplied.');
        return $TCD::CONFIG;
    } elsif (!defined($args{kernel_type})) {
        $Log->error('\'kernel_type\' argument not supplied.');
        return $TCD::CONFIG;
    } elsif (!defined($args{controller})) {
        $Log->error('\'controller\' argument not supplied.');
        return $TCD::CONFIG;
    }
    $full_cmd    = $args{command};
    $driver      = $args{driver};
    $ontap_ver   = $args{ontap_ver};
    $kernel_type = $args{kernel_type};
    $apiset      = $args{controller};
    $alt_usage   = $args{usage};
    $model       = $apiset->get_version_manager()->get_version_attribute(
        'attribute' => 'modelinfo'
    );
    if (!defined($model)) { $model = 'undefined'; }

    # Get subcommand for help string
    # $full_cmd is executed; $sub_cmd is located in %USAGE_TABLE
    # If this subroutine is called with $full_cmd set to
    # "ic config e0e ip_addr <something invalid>", the expected help text is
    # based off of just the first-level ic subcommand.  So strip off the end,
    # and we'll compare against the help text entry for "ic config"
    $sub_cmd = $full_cmd;
    $sub_cmd =~ s/(ic \S*) .*/$1/;
    # Get user supplied args list (optional)
    my @priv_list = defined($args{privs}) ? @{$args{privs}} : @PRIVS;

    # Make sure entry exists in %USAGE_TABLE for command
    if (!defined($USAGE_TABLE{$sub_cmd})) {
        $Log->error("'$sub_cmd' does not appear to be defined in \%USAGE_TABLE!");
        return $TCD::FATAL;
    }

    # Loop through the privs
    foreach my $priv (@priv_list) {
        # Set up some vars
        my ($cmd_str, $raw_output, $status, $expected_usage);

        # Make sure there is a Usage Table entry defined for comparison in
        # either %USAGE_TABLE or $alt_usage, and assign to $expected_usage .
        if (defined($alt_usage->{$priv})) {
            # Alternate Usage has been defined - use it
            $expected_usage = $alt_usage->{$priv};
        } else {
            # No Alternate usage - check for entry in %USAGE_TABLE and assign
            # or bomb out if entry does not exist.
            if (
                defined($USAGE_TABLE{$sub_cmd})
                || defined($USAGE_TABLE{$sub_cmd}{$driver})
                || defined($USAGE_TABLE{$sub_cmd}{$driver}{$priv})
                || defined($USAGE_TABLE{$sub_cmd}{$driver}{$priv}{$ontap_ver})
                || defined($USAGE_TABLE{$sub_cmd}{$driver}{$priv}{$ontap_ver}{$kernel_type})
            ) {
                $expected_usage = $USAGE_TABLE{$sub_cmd}{$driver}{$priv}->
                    {$ontap_ver}{$kernel_type};
            } else {
                $Log->error(
                    'Unable to find text at this level. Exiting!'.
                    "\nLevel = $sub_cmd - $driver - $priv - $ontap_ver - $kernel_type "
                );
                return $TCD::FATAL;
            }
        }

        # Output what we're doing
        $Log->comment("COMMAND  : $full_cmd");
        $Log->comment("PRIVILEGE: $priv");
        $Log->comment("SYS INFO : $model / $driver / $ontap_ver / $kernel_type");

        # Build the command string
        $cmd_str    = "priv set -q $priv ; $full_cmd";

        # Execute the command
        $raw_output = $apiset->execute_command(
            command           => $full_cmd,
            'privilege-level' => $priv,
        );

        # Verify output
        $status = output_verify(
            expected   => $expected_usage,
            raw        => $raw_output,
            match_type => 'strict',
        );

        # Output test status
        if ($status) {
            # Pass
            $Log->comment("STATUS   : OK\n ");
        } else {
            # Fail
            $Log->comment("STATUS   : FAIL\n ");
            $fails++;
        }
    } # End foreach

    # Check fails counter. If not zero, return TCD FAIL
    if ($fails > 0) {
        # At least one failure
        $Log->comment("STATUS   : ERROR There were $fails failures-test failed");
        return $TCD::FAIL;
    } else {
        # No failures
        $Log->comment("STATUS   : OK\n----");
        return $TCD::PASS;
    }

} # End compare_usage_statements()

################################################################################
# get_driver_type()
# -----------------
# Returns string with driver type (eg., SINAI,TAVOR,CONNECTX,MVIA,IWARP)
#
# Inputs:
#    controller - APISet Object - APISet object for controller
# Returns:
#    string - driver type used
################################################################################
sub get_driver_type {
    my %args = @_;
    my $apiset = $args{controller};

    # A source: https://wikid.netapp.com/w/Internally_Supported_ONTAP_Platforms
    my $absolut       = qr/(?:FAS|V)62(?:[148])0/;
    my $absolut_ibm   = qr/N7[579]50T/;
    my $alpine        = qr/FAS2554/;
    my $alpine_ibm    = qr/NS2554/;
    my $bimota        = qr/(?:FAS|V)80(?:[46])0/;
    my $bimota_ibm    = qr/NS80[46]0/;
    my $buell         = qr/(?:FAS|V)8020/;
    my $buell_ibm     = qr/NS8020/;
    my $carnegie      = qr/(?:FAS|V)32(?:[147])0/;
    my $carnegie_ibm  = qr/N62[147]0/;
    my $delphi        = qr/(?:FAS|V)32(?:[25])0/;
    my $delphi_ibm    = qr/N62[25]0/;
    my $excalibur     = qr/(?:FAS|V)6080/;
    my $excalibur_ibm = qr/N7[6789]00/;
    my $reyka         = qr/(?:FAS|V)62(?:[259])0/;
#    my $reyka_ibm     =  same as Absolute
    my $ruckus        = qr/FAS2220/;
    my $ruckus_ibm    = qr/N3150/;
    my $spectre       = qr/(?:FAS|V)3170/;
    my $spectre_ibm   = qr/N60[467]0/;
    my $tenaya        = qr/FAS2552/;
    my $tenaya_ibm    = qr/NS2552/;
    my $tioga         = qr/FAS2520/;
    my $tioga_ibm     = qr/NS2520/;
    my $vespa         = qr/FAS2240/;
    my $vespa_ibm     = qr/N32[24]0/;
    my $vsim          = qr/SIMBOX/;
    my $vsa           = qr/(DO|FD)vM[12]00/;

    my $model = $apiset->get_version_manager()->get_version_attribute(
        'attribute' => 'modelinfo',
    );

    if (!defined($model)) {
        $Log->error(
            'version_manager did not return a model type.  Unable to determine model.'
        );
        die 'Unable to determine Model Type.';
    } elsif ($model =~ $vsim || $model =~ $vsa) {
        return ('MVIA');
    } elsif ( $model =~ $absolut || $model =~ $absolut_ibm ||
              $model =~ $bimota  || $model =~ $bimota_ibm  ||
              $model =~ $buell   || $model =~ $buell_ibm   ||
              $model =~ $reyka ) {
        return ('CONNECTX');
    } elsif ( $model =~ $carnegie || $model =~ $carnegie_ibm ||
              $model =~ $delphi   || $model =~ $delphi_ibm ) {
        return ('IWARP');
    } elsif ( $model =~ $vespa  || $model =~ $vespa_ibm  ||
              $model =~ $ruckus || $model =~ $ruckus_ibm ||
	      $model =~ $alpine || $model =~ $alpine_ibm ||
	      $model =~ $tenaya || $model =~ $tenaya_ibm ||
	      $model =~ $tioga  || $model =~ $tioga_ibm ) {
        return ('SINAI');
    } elsif ( $model =~ $excalibur || $model =~ $excalibur_ibm ||
              $model =~ $spectre   || $model =~ $spectre_ibm ) {
        return ('TAVOR');
    } else {
        $Log->error("get_driver_type(): Unknown Model Type identified: $model");
        die "Unknown Model Type identified: $model";
    }
} ## End get_driver_type()

################################################################################
# get_ontap_name()
# ----------------
# Returns shortened codename of ONTAP version (e.g., 'SN0' for "SierraNevada.0")
# Inputs:
#    controller => (required) APISet object for node
# Returns:
#    scalar - string - shortened codename of ONTAP version
################################################################################
sub get_ontap_name {
    my %args = @_;
    my ($version_string);
    my $controller = $args{controller};

    if    (!defined($controller)) {
        $Log->error('ERROR get_ontap_name(): controller argument not passed in!');
        die 'ERROR: get_ontap_name(): controller argument not passed in!';
    }

    # Get just v.r.m. info from version string
    $version_string = $controller->get_version_manager()->get_version_attribute(
        'attribute' => 'release'
    );
    $Log->comment("Release String received: $version_string");
    $version_string =~ s/.*__(\d[.]\d[.]\d)/$1/;

    # Based on v.r.m. info, return shortened codename for ONTAP version
    if    ($version_string =~ /^8.2/) {
        $version_string =~ s/8[.]2[.](\d).*/$1/;
        return 'SN'.$version_string;
    } elsif ($version_string =~ /^8.3/) {
        $version_string =~ s/8[.]3[.](\d).*/$1/;
        return 'FS'.$version_string;
    } else {
        $Log->error("ERROR get_ontap_name(): unknown version: $version_string");
        die "ERROR: get_ontap_name(): unknown version: $version_string";
    }
}

################################################################################
#  output_clean(): Strips extraneous white space
################################################################################
sub output_clean {

    my %args = @_;

    my $str = $args{string};

    $str =~ s/^\s+//gm;
    $str =~ s/\s+$//gm;
    #$str =~ s/\n\s+/\n/gm;
    #$str =~ s/\s+\n$/\n/gm;

    return $str;

}  ## End output_clean()

################################################################################
#  output_verify(): Compare expected/received output based on match_type
#  return:
#      1 - successful match
#      0 - unsuccessful match
################################################################################
sub output_verify {

    my %args = @_;

    my $expected     = $args{expected};
    my $raw_output   = $args{raw};
    my $match_type   = $args{match_type};
    my $clean_output = output_clean(string=>$raw_output,);

    # Throw incoming strings into some lists
    my @expect_lines = split /\n/, $expected;
    my @output_lines = split /\n/, $clean_output;

    # Attempt to match things
    if ($match_type eq 'strict') {
        # 'strict' means use 'eq' to match the lines.

        my ($end_print_index, @fails);
        my $logfile_line = sprintf(
            "%2s | %-50s | %-50s\n",
            'L#',
            'Expected Lines (strict)',
            'Received Lines',
        );
        $end_print_index = $#expect_lines>$#output_lines ? $#expect_lines : $#output_lines;
        for my $i (0..$end_print_index) {
            $logfile_line .= sprintf(
                "%2s | %-50s | %-50s\n",
                "$i",
                defined($expect_lines[$i]) ? $expect_lines[$i] : q{},
                defined($output_lines[$i]) ? $output_lines[$i] : q{},
            );
            if (defined($expect_lines[$i]) && defined($output_lines[$i])
                && $expect_lines[$i] eq $output_lines[$i]) { } #do nothing
            else {
                # Something didn't match. Append mis-match line to @fails
                push(@fails, $i);
            }
        }

        if ($#fails >= 0) {
            $Log->error(scalar @fails. ' mismatched lines found at: '
                .join(', ', @fails)
            );
            $Log->comment($logfile_line);
            return 0; #A line didn't match - return fail.
        } else {
            if ($VERBOSE) { $Log->comment($logfile_line); }
            return 1;  #If we're here, everything passed.
        }
    } elsif ($match_type eq 'regex') {
        my $modified_expect = $expected;
        my $modified_output = $clean_output;
        #Escape []| characters in the lines for comparison, otherwise regex will
        #try to use these characters as regex special characters
        $modified_expect =~ s/([[]|])/\\$1/g;
        $modified_output =~ s/([[]|])/\\$1/g;

        if ($modified_output =~ /$modified_expect/) {
            if ($VERBOSE) {
                $Log->comment("regex:\n$expected\n");
                $Log->comment("matches:\n$clean_output\n");
            }
            return 1;
        } else {
            my $end_print_index;
            $Log->error('Output does not regex-match Expected');
            my $logfile_line = sprintf(
                '%2s | %-50s | %-50s ',
                'L#',
                'Expected Lines (regex)',
                'Received Lines',
            );
            $Log->comment($logfile_line);

            $end_print_index = $#expect_lines>$#output_lines ? $#expect_lines : $#output_lines;
            for my $i (0..$end_print_index) {
                $logfile_line = sprintf(
                    '%2s | %-50s | %-50s ',
                    "$i",
                    defined($expect_lines[$i]) ? $expect_lines[$i] : q{},
                    defined($output_lines[$i]) ? $output_lines[$i] : q{},
                );
                $Log->comment($logfile_line);
            }
            return 0;
        }
    } else {
        # How did we get here?
        $Log->error('ERROR: match value \'strict\' or \'regex\' not found');
        return 0;
    }

}  ## End output_verify()

################################################################################
# _validate_cdef_output_enumerated()
# ----------------------------------
# Validates single entry in CDEF's parsed_output where entry is an enumerated type
#
# Inputs:
#    'key'      - scalar  - name of parse_output entry
#    'expected' - hashref - table containing expected values
#    'response' - scalar  - value returned from CDEF
# Returns:
#    1 - test passed
#    0 - test failed
################################################################################
sub _validate_cdef_output_enumerated {
    my %args = @_;
    my $entry_name      = $args{key};
    my @expected_values = @{$args{expected}};
    my $response        = $args{response};

    foreach my $expected_value (@expected_values) {
        if ($expected_value eq $response) {
            # We've found what we're looking for
            $Log->comment("ENUMERATED: $entry_name - PASSED");
            return 1;
        }
    }
    $Log->comment("ENUMERATED: $entry_name - FAILED");
    $Log->comment("  value: $response; expected: ".join(', ', @expected_values));
    # We didn't find a match between response and entries in expected list
    return 0;
}

################################################################################
# _validate_cdef_output_expected_key_list()
# -----------------------------------------
# Validates the list of keys returned in CDEF output matches the list of keys
# we expected.
#
# Inputs:
#    'expected'    - hashref - table containing expected values
#    'response'    - hashref - table returned from CDEF
#    'kernel_type' - scalar - kernel type ('DEBUG' or 'NONDEBUG')
#    'ontap_version' - scalar - kernel version ('SN1', 'FS0', etc.)
# Returns:
#    1 - test passed
#    0 - test failed
################################################################################
sub _validate_cdef_output_expected_key_list {
    my %args = @_;
    my %expected_output = %{$args{expected}};
    my %returned_output = %{$args{returned}};
    my $kernel_type     = $args{kernel_type};
    my $ontap_version   = $args{ontap_version};
    my $overall_pass    = 1;

    if (!%expected_output || !%returned_output || !defined($kernel_type)
        || !defined($ontap_version)) {
        $Log->error('_validate_cdef_output_expected_key_list(): required parameter missing');
    }

    my @no_check_list = ($kernel_type eq 'DEBUG') ?
        (@{$expected_output{'no_check'}}, @{$expected_output{'no_check_debug'}}) :
        @{$expected_output{'no_check'}};

    $Log->comment('STEP 1a: Check that the keys in response are expected.');
    foreach my $returned_key (sort(keys(%returned_output))) {
        # if:
        #    1. returned_key is not a valid key in %expected_output, AND
        #    2. returned_key is not in @no_check_list
        # elsif:
        #    1. ontap_versions list seen for key, AND
        #    2. current ontap version doesn't match valid versions for key
        if ( (!defined($expected_output{$returned_key}))
            && (!any {$_ eq $returned_key} @no_check_list) ) {
            $Log->error(" Received key '$returned_key' not found in list of expected keys! ");
            $overall_pass = 0;
        } elsif (defined($expected_output{$returned_key}->{ontap_versions})) {
            if (!any {$_ eq $ontap_version}
                @{$expected_output{$returned_key}->{ontap_versions}}) {
                # we get here if we receive a key that is seen on other ontap
                # versions, but not expected in the current ontap version
                $overall_pass = 0;
            }
        }
    }

    $Log->comment('STEP 1b: Check that all expected keys were returned');
    # First, check defined keys
    foreach my $expected_key (sort(keys(%expected_output))) {
        # if:
        #    1. ignore keys 'no_check_debug' and 'no_check'
        #    2. key not seen in returned output
        #    3. debug kernel or (nondebug & key is debug only)
        #    4. key is supposed to exist in current ontap version
        if ( ($expected_key ne 'no_check_debug' && $expected_key ne 'no_check')
            && (!defined($returned_output{$expected_key}))
            && ($kernel_type eq 'DEBUG'
                || ($kernel_type eq 'NONDEBUG'
                    && !$expected_output{$expected_key}->{is_debug_only}))
            && (!defined($expected_output{$expected_key}->{ontap_versions})
                || (defined($expected_output{$expected_key}->{ontap_versions})
                    && any {$_ eq $ontap_version}
                        @{$expected_output{$expected_key}->{ontap_versions}})) ) {
            $Log->error("'$expected_key' was expected but not seen!");
            $overall_pass = 0;
        }
    }

    # Second, check list of keys we aren't checking
    foreach my $expected_key (@no_check_list) {
        if (!defined($returned_output{$expected_key})) {
            $Log->error("'$expected_key'(no_check) was expected but not seen!");
            $overall_pass = 0;
        }
    }

    return $overall_pass;
}

################################################################################
# _validate_cdef_output_int_range()
# ---------------------------------
# Validates single entry in CDEF's parsed_output where entry is an integer range
#
# Inputs:
#    'key'      - scalar  - name of parsed_output entry
#    'expected' - hashref - table containing expected values
#    'response' - scalar  - value returned from CDEF
# Returns:
#    1 - test passed
#    0 - test failed
################################################################################
sub _validate_cdef_output_int_range {
    my %args = @_;
    my $entry_name      = $args{key};
    my @expected_values = @{$args{expected}};
    my $response        = $args{response};

    # Check both directions so that order of items doesn't matter
    if (
        ($response >= $expected_values[0] && $response <= $expected_values[1]) ||
        ($response <= $expected_values[0] && $response >= $expected_values[1])
    ) {
        $Log->comment("INT_RANGE: $entry_name - PASSED");
        return 1;
    }
    $Log->comment("INT_RANGE: $entry_name - FAILED");
    $Log->comment("  value: $response; range: [$expected_values[0],$expected_values[1]]");
    return 0;
}

################################################################################
# _validate_cdef_output_int_value()
# ---------------------------------
# Validates single entry in CDEF's parsed_output where entry should be a
# specific value
#
# Inputs:
#    'key'      - scalar  - name of parsed_output entry
#    'expected' - hashref - table containing expected values
#    'response' - scalar  - value returned from CDEF
# Returns:
#    1 - test passed
#    0 - test failed
################################################################################
sub _validate_cdef_output_int_value {
    my %args = @_;
    my $entry_name  = $args{key};
    my $expected    = $args{expected};
    my $response    = $args{response};

    if ($response == $expected) {
        $Log->comment("INT_VALUE: $entry_name - PASSED");
        return 1;
    }
    $Log->comment("INT_VALUE: $entry_name - FAILED");
    $Log->comment("  value: $response; expected: $expected");
    return 0;
}

################################################################################
# _validate_cdef_output_string()
# ------------------------------
# Validates single entry in CDEF's parsed_output where entry is a simple string
#
# Inputs:
#    'key'      - scalar  - name of parsed_output entry
#    'expected' - hashref - table containing expected values
#    'response' - scalar  - value returned from CDEF
# Returns:
#    1 - test passed
#    0 - test failed
################################################################################
sub _validate_cdef_output_string {
    my %args = @_;
    my $entry_name  = $args{key};
    my $expected    = $args{expected};
    my $response    = $args{response};

    if ($response eq $expected) {
        $Log->comment("STRING: $entry_name - PASSED");
        return 1;
    }
    $Log->comment("STRING: $entry_name - FAILED");
    $Log->comment("  value: '$response'; expected: '$expected'");
    return 0;
}

################################################################################
# validate_cdef_output()
# ----------------------
# Validates return values from CDEF for a command.  Typically, if we're called,
# we've already verified that there are the same number of entries in each
# expected and output table.
#
# Inputs:
#    'key'      - scalar  - name of parsed_output entry
#    'expected' - hashref - Expected table
#    'returned' - hashref - parsed_output table from CDEF
# Returns:
#      $TCD::<status>
################################################################################
sub validate_cdef_parsed_values {
    my $overall_pass = 1;
    my %args = @_;
    my $driver          = $args{driver};
    my %expected_output = %{$args{expected}};
    my %returned_output = %{$args{returned}};
    my $kernel_type     = $args{kernel_type};
    my $ontap_version   = $args{ontap_version};

    # Some basic checks before we continue
    if (!%expected_output || !%returned_output || !defined($kernel_type)) {
        $Log->error('validate_cdef_output(): Required parameters are missing!');
        die 'validate_cdef_output(): Required parameters missing.';
    }

    # First, check list of keys returned
    $overall_pass = _validate_cdef_output_expected_key_list(
        expected      => \%expected_output,
        returned      => \%returned_output,
        kernel_type   => $kernel_type,
        ontap_version => $ontap_version,
    );

    # If we've hit a failure, break.  Otherwise, keep going.
    if ($overall_pass == 0) { return $TCD::FAIL; }
    $Log->comment('STEP 1: PASSED');


    # Now, verify each value is what we expect
    $Log->comment('STEP 2: verify the values of returned keys');
    foreach my $key (keys(%expected_output)){
        my $response = $returned_output{$key};
        my $expected = $expected_output{$key};
        if ($key eq "no_check_debug" || $key eq "no_check") {
            #Skip these entries - just arrays of keys we aren't checking
        } elsif (defined($expected_output{$key}->{ontap_versions})
            && !any {$_ eq $ontap_version} @{$expected_output{$key}->{ontap_versions}} ) {
            #Skip these entries - should not exist in current ontap version
        } elsif ($kernel_type eq 'NONDEBUG' && $expected->{is_debug_only}) {
            # Skip it, as expected value is only visible in DEBUG kernel,
            # and current kernel is non-debug
        } elsif ($expected->{type} eq 'int_range') {
            if (!_validate_cdef_output_int_range(
                key      => $key,
                response => $response,
                expected => $expected->{values},
            )) { $overall_pass = 0; }
        } elsif ($expected->{type} eq 'int_value') {
            if (!_validate_cdef_output_int_value(
                key      => $key,
                response => $response,
                expected => $expected->{values},
            )) { $overall_pass = 0; }
        } elsif ($expected->{type} eq 'is_positive') {
            if ($response < 0) {
                $overall_pass = 0;
            }
        } elsif ($expected->{type} eq 'enumerated') {
            if (!_validate_cdef_output_enumerated(
                key      => $key,
                response => $response,
                expected => $expected->{values},
            )) { $overall_pass = 0; }
        } elsif ($expected->{type} eq 'string') {
            if (!_validate_cdef_output_string(
                key      => $key,
                response => $response,
                expected => $expected->{value},
            )) { $overall_pass = 0; }
        } elsif ($expected->{type} eq 'hashref') {
            $Log->comment('HASHREF: I don\'t know what to do with hashrefs yet, so skip it',);
        } else {
            $Log->error('Unknown expected type: '.$expected->{type});
            return $TCD::CONFIG;
        }

        $Log->comment(q{ });
    } #end foreach($key)

    if ($overall_pass) { return $TCD::PASS; }
    else               { return $TCD::FAIL; }
}

1;
## End of module
__END__

=head1 NAME

Iconnect::Command_Testing - common variables and subroutines for Command Testing scripts

=head1 SYNOPSIS

    use Command_Testing qw(:ALL);                          # this is all you need

=head1 DESCRIPTION

This module provides the following variables:
    B<%USAGE_TABLE>       - Table of expected usage statements

The module also exports the following subroutines:

=item B<compare_usage_statements()>
Input:
    command     = (required) command to execute on controller
    driver      = (required) driver type (CONNECTX, IWARP, etc.)
    ontap_ver   = (required) ONTAP version (FS0, SN1, etc)
    kernel_type = (required) DEBUG or NONDEBUG
    controller  = (required) APISet object for controller
    privs       = (optional) arrayref - list of priv levels to execute against
Returns:
    $TCD::<status>

=item B<get_driver_type()>
Input:
    controller  = APISet object for controller
Returns:
    scalar - string indicating IC Driver
        CONNECTX
        IWARP
        MVIA
        SINAI
        TAVOR

=item B<get_ontap_name()>
Input:
    controller = APISet object for node
Returns:
    scalar - string with shortened codename of ONTAP version (e.g., SN0)


=item B<validate_cdef_parsed_values()>
Input:
    key        = (required) name of parsed_output key being verified
    expected   = (required) hashref - table of Expected return key/values
    returned   = (required) hashref - parsed_output table from CDEF
Returns:
    $TCD::<status>

1;
