# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/ImageBackup.pm#1 $
##
## @summary ImageBackup
##	Core library module to support image backup commands for the Advanced Tape project
##
## @pod here
##
## @author sydneyw@netapp.com, dl-dpgqa-automation@netapp.com
##
## @status public
##

=head1 NAME

ImageBackup - Image Backup ONTAP interface library

=head1 DESCRIPTION

Provides ONTAP interface for Image Backup commands. This library supports both CLI, NDMP  and ZAPI
interface to the filer. The default is to use ZAPI and fallback to CLI if ZAPI is not available.

The following optional CLI parameter(s) are used by this module:

B<IB_CMD>

Used to specify the ONTAP interface to be use. The default is as above. Valid values are: cli, nbu or zapi.

=head1 ROUTINES

=cut

use strict;

package ImageBackup;

use Tharn;
use Carp;
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT);
use NBU::nbuUtils;
use SayWrap qw( sayEOF );

use base qw( Exporter );

use constant STU_EXT => "_stu";
use constant POL_EXT => "_pol";
use constant SCH_EXT => "_sch";
use constant NBU_BACKUP_TYPE  => 0 ;
use constant NBU_RESTORE_TYPE => 2 ;
use constant NBU_ACTIVE_STATE => 1;

BEGIN {
	param("IB_CMD", -default, undef);
	param("NDMP_DNA", -default, undef);

	Params::Validate::validation_options(
		on_fail => sub { my @msgs = @_; chomp(@msgs);
				Carp::confess(Tharn::script_result(@msgs)) }
	);

	our @EXPORT_OK = qw(
		nbu_job_active
		nbu_init
		ib_cmd_default
		ib_backup
		ib_restore
		ib_abort
		ib_continue
		ib_status
		ib_vol_geometry
		ib_tape_header
	);
}

END {
	# Cleanup the policies created by this run
	foreach my $device (keys %{ $::nbu{_POLICY}} ) {
		my $policy = $::nbu{_POLICY}{$device};
		logcomment("Deleting Policy $policy from NBU");
		NBU::nbuUtils::bppldelete($policy)
			if NBU::nbuUtils::bppllist($policy);
		delete $::nbu{_POLICY}{$device};
	}
}

my $ib_cmd = lc ($::IB_CMD) || "cli";
my %nbu; # Used to hold nbu related information

=head2 ::ib_cmd_default

Sets the default interface for ONTAP Image Backup commands and returns the default setting.

=over

=item Arguments

B<type> = I<cli|zapi>

I<Optional.>

Specify the default interface to use. If not supplied will just return the current value.

=item Examples

=back

	my $curr_default = ib_cmd_default("nbu");

=cut

sub ib_cmd_default {
	my ($type) = @_;

	if (defined $type) {
		croak "Invalid $type type. Must be cli, nbu or zapi. Aborting..."
			unless $type =~ /^(cli|nbu|zapi)$/i;
		if ( $::IB_CMD ) {
			logcomment("IB_CMD set from CLI. Ignoring default $type...");
		} else {
			$ib_cmd = lc($type);
			logcomment("Default ONTAP Interface is $ib_cmd");
		}
	}
	return $ib_cmd;
}

sub _policy {
	my ($nbu_client, $device) = @_;

	return $::nbu{_POLICY}{$device} = $nbu_client."_".$device.POL_EXT."_".time();
}

sub _nbu_job_id {
	my ($nbu_type, $device, $max_retry) = @_;

	# Attempt to find an active nbu job with max_retry times
	$max_retry = 60 unless defined $max_retry;
	do {
		my %status = NBU::nbuUtils::bpdbjobs();
		foreach my $id (keys %status) {
			if ( $status{$id}{state} == NBU_ACTIVE_STATE ) {
				my $line = join ",", %{$status{$id}};
				# Used policy as key for backup job and backup_id for restore job
				my $key = ($nbu_type == NBU_BACKUP_TYPE
							? $::nbu{_POLICY}{$device}
							: "restoring image ".$::nbu{_BACKUP_ID}{$device}
						);
				if ( $line =~ /$key/i ) {
					if ( $nbu_type == NBU_BACKUP_TYPE ) {
						# Compose and save backup id to use to identify restore job
						$::nbu{_BACKUP_ID}{$device} = $status{$id}{client}."_".$status{$id}{started};
					}
					logcomment("Found Active NBU Job id: $id");
					return $id;
				}
			}
		}
		sleep 1; # Pause a little before retrying
	} while $max_retry--;
	croak "No active NBU ".($nbu_type == NBU_BACKUP_TYPE ? "backup" : "restore")."job using $device";
}

=head2 ::nbu_job_active

Returns a boolean to indicate if an NBU job is active or not.

=over

=item Arguments

B<ib_id> = I<integer>

Image Backup Job id to determine the NBU job.

=back

=item Example

	my $is_active = nbu_job_active($ib_job_id);

=cut

sub nbu_job_active {
	my ($ib_id) = @_;
	
	my $nbu_id = $::nbu{_JOB_ID}{$ib_id};
	my %status = NBU::nbuUtils::bpdbjobs();
	my $active = $status{$nbu_id}{state} == NBU_ACTIVE_STATE;
	logcomment("NBU job $nbu_id (ib job id $ib_id) state ".$status{nbu_id}{state}.", is ".($active ? "" : "not ")."active");
	return $active;
}

# Translates an nbu job id to ib job id

sub _get_ib_jobid {
	my ($conn, $nbu_type, $nbu_id, $src, $dst) = @_;

	my $title = "NBU ".($nbu_type == NBU_BACKUP_TYPE ? "Backup" : "Restore")." job id $nbu_id, src $src, dst $dst";
	my $found_id;
	# Keep searching for IB job id as long as nbu job is active
	my $ib_jobid;
	do {
		# ignore 'smtape status' fails
		eval {
			# There can only be one active job on filer/device
			my @stat = ib_status($conn, spec=>$src, detail=>0);
			logcomment("Match nbu id $nbu_id to ib job id ".$stat[0]{'job id'});
			$ib_jobid = $stat[0]{'job id'};
		};
		return $ib_jobid
			if defined $ib_jobid;
		sleep 1; # Pause a little before retrying
		eval {
			$found_id = _nbu_job_id($nbu_type, $dst, 0); # croaks if nbu job is not active
		};
		croak "Unable to determine IB job id from $title. Error: $@"
			if $@;
	} while $found_id == $nbu_id;
	croak "Unable to determine IB job id from NBU job $nbu_id";
}

sub _nbu_client {
	my ($client) = @_;

	return (exists $::nbu{_NETNAME}{$client} ? $::nbu{_NETNAME}{$client} : "");
}

=head2 ::nbu_init

Function to initialise the NBU Master with relevant robot and tape device information so that it could be used for carry out IB jobs.

=over

=item Arguments

B<dma> = I<NBU Master>

Host containing NBU installation for which to run NBU CLIs from.

B<nbu_client> = I<filer>

The name of the filer to be use as an NBU client.

B<client_network> = I<Network Domain>

The network domain of the NBU client.

B<robot_name> = I<name>

The name of the robot tape library.

B<robot_type> = I<type>

The robot type, e.g. TLD, 8MM etc

B<device_type> = I<type>

The type of the tape devices, e.g. Type of the tape devices, e.g. 8mm, hcart

B<device_list> = I<array of tape devices to use>

List of the tape devices.

=back

=item Example

	@devices = ["rst1l", "rst2l"];
	nbu_init("sdqa-g68", "fsr-g21", "lab.netapp.com", "mc2", "TLD", "hcart", @devices);

=cut

sub nbu_init {
	my ($dma, $nbu_client, $client_network, $robot_name, $robot_type, $device_type, @device_list) = @_;

	my $stu_name = $nbu_client.STU_EXT;
	my $client_netname = $nbu_client;
	$client_netname .= ".".$client_network
		if $client_network && ! ($client_netname =~ /\./);

	# Configure the filer
	unless ( exists $::nbu{_NETNAME}{$nbu_client} ) {
		$::nbu{_DMA} = $dma;
		eval {
			# Add new filer / robot configuration
			my $robot_num = NBU::nbuUtils::add_tpconfig_list ($client_netname, $stu_name, undef, undef, $dma, $robot_name, $robot_type, $device_type, @device_list);
			NBU::nbuUtils::bpstuadd_tape($stu_name, $client_netname, $dma, $robot_num, $robot_type, $device_type, @device_list);

			$::nbu{_NETNAME}{$nbu_client} = $client_netname;
		};
		croak "NBU Setup for Filer $nbu_client failed. Error: $@"
			if $@;
	}
	# configure each device(s)
	foreach my $device (@device_list) {
		unless ( exists $::nbu{$nbu_client}{$device} ) {
			eval {
				my $policy_name = _policy($nbu_client, $device);
				NBU::nbuUtils::bppolicynew($policy_name, $dma);
				croak "Failed to add new policy $policy_name, for device $device. Error: $@"
					if $@;
				my $sched_name = $policy_name.SCH_EXT;
				NBU::nbuUtils::bpplsched($policy_name, $sched_name, $dma, "FULL", $stu_name);
				croak "Failed to add schedule to policy $policy_name. Error: $@"
					if $@;
				NBU::nbuUtils::bpplinfo_tape($policy_name, $dma, $stu_name);
				NBU::nbuUtils::bpplclients($policy_name, "-add", $client_netname, $dma, "NDMP", "NDMP");
			};
			if ( $@ ) {
				croak "Filer $nbu_client, Devices $device: Failed NBU setup. Error $@";
			} else {
				$::nbu{$nbu_client}{$device} = 1;
			}
		};
	};
}

sub _cmd_opts {
	my (%opts) = @_;

	my %args = ("block_size" => "-b"
	#	,	"mode"       => "-m"
	#	,	"type"       => "-t"
		,	"snapshot"   => "-s"
	);
	my $param = " ";
	foreach my $p ( keys %args ) {
		$param .= $args{$p}." ".$opts{$p}." " if exists $opts{$p};
	}
	return $param;
}

=head2 ::ib_backup

Starts an image backup from I<src_path> to I<dst_path>. I<src_path> is a volume path,
e.g. /vol/<name>, and I<dst_path> can be a tape device or pathname to a file.
Returns the job id that is assigned when the command starts running and croaks otherwise.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<src_path> => I<volume path>

The path of the volume that contains data to be backup, format of the form '/vol/<name>'.

B<dst_path> => I<device|file>

Name of the tape device or file to backup the data to.

B<block_size> => I<units of KB>

I<Optional.>

The tape record block size in KB units, must be multiples of 4. If not supplied ONTAP uses a
default of 256.

B<type> => I<dw|dw-metadata|mm>

I<Optional.>

For ONTAP test mode only, specifies which phase, datawarehouse, metadata blocks of the
datawarehouse or metadata map, to backup.

B<snapshot> => I<string>

I<Optional.>

Name of the snapshot to be use for the backup. If not specified a snapshot will be created to perform the
backup.

B<geometry> => I<disk geometry of volume>

I<Optional.>

Indicates destination volume's disk geometry of the destinational traditional volume.

B<all_snapshots> => I<boolean>

I<Optional.>

Indicates whether all snapshots, upto and including the specified snapshot, if a snapshot is included,
to backup. Default: 0

B<tape_seeding> => I<boolean>

I<Optional.>

Indicates whether backup is to be used for tape seeding, ie snapmirror to tape. Default: 0

B<inline> => I<boolean>

Indicates whether to use inline mode. Default: 0

=item Examples

=back

	$job_id = ib_backup($conn, src_path=>"/vol/src_vol", dst_path=>"rst01", mode=>"dw");

=cut

sub ib_backup {
	my $conn = shift;
	my (%opts) = validate (@_,
			{	src_path      => { type => SCALAR, optional => 0 }
			,	dst_path      => { type => SCALAR, optional => 0 }
			,	block_size    => { type => SCALAR, optional => 1 }
			,	type          => { type => SCALAR, optional => 1 }
			,	snapshot      => { type => SCALAR, optional => 1 }
			,	geometry      => { type => SCALAR, optional => 1 }
			,	all_snapshots => { type => BOOLEAN, optional => 1 }
			,	tape_seeding  => { type => BOOLEAN, optional => 1 }
			,	inline        => { type => BOOLEAN, optional => 1 }
			,	
			}
		);

	if ( $ib_cmd =~ /cli/i ) {
		my $priv_cmd = (exists $opts{type} ? "test" : "diag");
		my $cmd = ($opts{dst_path} =~ /^\// ? "priv set -q $priv_cmd; " : "")
			. "smtape backup"
			._cmd_opts(%opts)
			.($opts{geometry} ? "-g ".$opts{geometry}." " : "")
			#.($opts{all_snapshots} ? "-a " : "")
			#.($opts{tape_seeding} ? "" : "-l ")
			#.($opts{inline} ? "-n " : "")
			.$opts{src_path}." ".$opts{dst_path};
		my $text = sayEOF($conn, $cmd);
		if ( $text =~ /job (\w+) started/i ) {
			return $1;
		} else {
			croak "'$cmd' failed:\n$text.";
		}
	} elsif ( $ib_cmd =~ /nbu/i ) {
		my @params = (
			"\"set type=smtape\"",
			#"\"set SMTAPE_ALL_SNAPSHOTS=".($opts{all_snapshots} ? "yes" : "no")."\"",
			#"\"set SMTAPE_MODE=".($opts{tape_seeding} ? "DR" : "BR")."\"",
			#"\"set SMTAPE_NO_DISK_WRITE=".($opts{inline} ? "yes" : "no")."\"",
			($opts{snapshot} ? "\"set SMTAPE_SNAPSHOT_NAME=".$opts{snapshot}."\"" : "")
		);
		my $start_time = time();
		my $policy_name = $::nbu{_POLICY}{$opts{dst_path}};
		my $dma = $::nbu{_DMA};
		my $title =  "NBU backup from ".$opts{src_path}." to ".$opts{dst_path};
		NBU::nbuUtils::bpplinclude($policy_name, $dma, "-add", $opts{src_path}, @params);
		croak "$title. Error: $@"
			if $@;
		
		my ($ib_id, $nbu_id);
		eval {
			NBU::nbuUtils::bpbackup_tape($policy_name);
			$nbu_id = _nbu_job_id(NBU_BACKUP_TYPE, $opts{dst_path});
			$ib_id = _get_ib_jobid ($conn, NBU_BACKUP_TYPE, $nbu_id, $opts{src_path}, $opts{dst_path});
		};
		croak "$title failed. Error: $@"
			if $@;
		$::nbu{_JOB_ID}{$ib_id} = $nbu_id;
		$::nbu{_SRC}{$opts{dst_path}} = $start_time.",".$opts{src_path};
		return $ib_id;
	} else {
		croak "$ib_cmd not available...";
	}
}

=head2 ::ib_restore

Starts an image restore of data from I<src_path> to I<dst_path>. I<src_path> can be a tape device
or a path to a file. I<dst_path> is a volume path, e.g. /vol/<name>.
Returns the job id that is assigned when the command starts successfully and croaks otherwise.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<src_path> => I<device|file>

Name of the tape device or file holding the backup data.

B<dst_path> => I<volume path>

The path of the volume to be use to hold the restore data, format of the form '/vol/<name>'.

B<block_size> => I<units of KB>

I<Optional.>

The tape record block size in KB units, must be multiples of 4, this needs to match the block size
specified when the backup was created, otherwise it will fail. If not supplied ONTAP uses a
default of 256.

B<mode> => I<image|logical>

I<Optional.>

Image based restore is by physical blocks where as a logical restore is via the file system.

B<snapshot> => I<string>

I<Optional.>

The snapshot to be used in the restore.

=item Examples

=back

	$job_id = ib_restore($conn, src_path=>"rst0a", dst_path=>"/vol/dst_vol",
				block_size=>4, mode=>"image", snapshot=>"backup_ssn", list=>1);

=cut

sub ib_restore {
	my $conn = shift;
	my (%opts) = validate (@_,
			{	src_path   => { type => SCALAR, optional => 0 }
			,	dst_path   => { type => SCALAR, optional => 0 }
			,	block_size => { type => SCALAR, optional => 1 }
			,	mode       => { type => SCALAR, optional => 1 }
			,	snapshot   => { type => SCALAR, optional => 1 }
			,	list       => { type => BOOLEAN, optional => 1, default => 0 }
			}
		);

	if ( $ib_cmd =~ /cli/i ) {
		my $cmd = ($opts{src_path} =~ /^\// ? "priv set -q diag; " : "")
			. "smtape restore"
			._cmd_opts(%opts)
			.$opts{dst_path}." ".$opts{src_path};
		my $text = sayEOF($conn, $cmd);
		if ( $text =~ /job (\w+) started/i ) {
			return $1;
		} else {
			croak "'$cmd' failed:\n$text.";
		}
	} elsif ( $ib_cmd =~ /nbu/i ) {
		my $policy_name = $::nbu{_POLICY}{$opts{src_path}};
		my ($stime, $src) = split ",", $::nbu{_SRC}{$opts{src_path}};
		my $dma = $::nbu{_DMA};
		my ($ib_id, $nbu_id);
		eval {
			NBU::nbuUtils::bprestore_tape($policy_name, $dma, _nbu_client($conn->hostp()->hostname()), undef,
											_time_to_str($stime), _time_to_str(time()), $src, $opts{dst_path});
			$nbu_id = _nbu_job_id(NBU_RESTORE_TYPE, $opts{src_path});
			$ib_id = _get_ib_jobid ($conn, NBU_RESTORE_TYPE, $nbu_id, $opts{dst_path}, $nbu_id, $opts{src_path});
		};
		croak "NDMP restore from ".$opts{dst_path}." to ".$opts{src_path}." failed. Error: $@"
			if $@;
		$::nbu{_JOB_ID}{$ib_id} = $nbu_id;
		return $ib_id;
	} else {
		croak "$ib_cmd not available...";
	}
}

sub _time_to_str {
	my ($int_time) = @_;

	my ($sec, $min, $hour, $mday, $mon, $year) = localtime($int_time);
	return sprintf("%02d/%02d/%04d", $mon+1, $mday, 1900+$year);
}

=head2 ::ib_vol_geometry

Uses the '-g' option of 'smtape restore' to Returns the disk geometry of a traditional volume.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<vol_name> = I<string>

The name of the traditional volume to obtain disk geometry for.

=back

=itme Examples

	ib_vol_geometry($conn, "vol0"); # Retrieve disk geometry of a traditional root volume

=cut

sub ib_vol_geometry {
	my ($conn, $vol_name) = @_;

	if ( $ib_cmd =~ /(cli|nbu)/i ) {
		my $cmd = "smtape restore -g /vol/$vol_name";
		my $text = sayEOF($conn, $cmd);
		if ( $text =~ /-g (\d+)/i ) {
			return $1;
		} else {
			croak "'$cmd' failed:\n$text.";
		}
	} else {
		croak "$ib_cmd not available...";
	}
}

=head2 ::ib_tape_header

Returns the image header of a tape device.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<tape_device> = I<string>

Name of a tape device containing a tape of the backup to obtain the image header from.

=item Example

	ib_tape_header($fconn, "rst0l");

=cut

sub ib_tape_header {
	my ($conn, $device_name) = @_;

	if ( $ib_cmd =~ /(cli|nbu)/i ) {
		my $cmd = "smtape restore -h $device_name";
		my $text = sayEOF($conn, $cmd);
		if ( $text =~ /x/i ) {
			return $text;
		} else {
			croak "'$cmd' failed:\n$text.";
		}
	} else {
		croak "$ib_cmd not available...";
	}
}

=head2 ::ib_abort

Aborts a backup or restore operation. Croaks if abort fails.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<job_id> = I<id>

The backup or restore job id to be aborted.

=item Examples

=back

	ib_abort($conn, $job_id);

=cut

sub ib_abort {
	my ($conn, $id) = @_;

	croak "Missing job id..."
		unless defined $id;

	if ( $ib_cmd =~ /cli/i ) {
		my $cmd = "smtape abort $id";
		my $text = sayEOF($conn, $cmd);
		croak "'$cmd' failed: $text."
			unless $text =~ /job $id aborted/i;
	} elsif ( $ib_cmd =~ /nbu/i ) {
		if ( exists $::nbu{_JOB_ID}{$id} ) {
			my $nbu_id = $::nbu{_JOB_ID}{$id};
			eval {
				NBU::nbuUtils::bpdpjobs_cancel($nbu_id);
			};
			croak "Aborting IB Job $id (nbu job id: $nbu_id) failed. Error: $@"
				if $@;;
		} else {
			croak "No nbu job recorded for ib job id $id";
		}
	} else {
		croak "$ib_cmd not available...";
	}
	return 1;
}

=head2 ::ib_continue

Continue a backup or restore operation that is waiting for some manual action such as a tape
change. Croaks if the command fails.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<job_id> => I<integer>

I<Required.>

The backup or restore job id to be continued.

B<device> => I<device name>

I<Optional.>

Tape device to use for the remainder of the backup/restore command.
If not specified, will IB will use the current tape device for this job.

=item Examples

=back

	ib_continue($conn, job_id=>$id);
	ib_continue($conn, job_id=>$id, device=>$new_tape_device);

=cut

sub ib_continue {
	my $conn = shift;
	my (%opts) = validate (@_,
			{	job_id   => { type => SCALAR, optional => 0 }
			,	device   => { type => SCALAR, optional => 1 }
			}
		);

	croak "Missing job id..."
		unless exists $opts{job_id};
	my $id = $opts{job_id};

	if ( $ib_cmd =~ /cli/i ) {
		my $cmd = "smtape continue ".$id
				.(exists $opts{device} ? " ".$opts{device} : "");
		my $text = sayEOF($conn, $cmd);
		croak "'$cmd' failed: $text."
			if $text =~ /job $id continued/i;
	} elsif ( $ib_cmd =~ /nbu/i ) {
		croak "nbu not supported...";
	} else {
		croak "$ib_cmd not available...";
	}
	return 1;
}

=head2 ::ib_status

Returns the status of a backup or restore job in a pointer to an array of hashes.
Each entry representing the status of a single backup/restore job.  Croaks if no status
can be found.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<spec> => I<path|job_id>

I<Optional.>

This could be a volume path, of the form /vol/<name> or a job_id. And if not specify all
current jobs will be provided.

B<detail> => I<0|1>

I<Optional.>

Obtain detail status information on the job(s). Default: 0

=item Examples

=back

	$hash_array = ib_status($conn, spec=>$backup_id);
	$hash_array = ib_status($conn, spec=>$backup_id, detail=>1);

=cut

sub ib_status {
	my $conn = shift;
	my (%opts) = validate (@_,
			{	spec   => { type => SCALAR, optional => 0 }
			,	detail => { type => BOOLEAN, optional => 1, default => 0 }
			}
		);

	my @stats;
	if ( $ib_cmd =~ /(cli|nbu)/i ) {
		my $cmd = "smtape status ".($opts{detail} ? "-l " : "")
			.(exists $opts{spec}
				? ($opts{spec} =~ /^\// ? "-p " : "").$opts{spec}
				: ""
			);

		my $text = sayEOF($conn, $cmd);
		my @lines = split "\n", $text;
		if ( $opts{detail} ) {
			# Example of detail status:
			#	Job ID:         1
			#	Sequence No:    0
			#	Type:                  Backup
			#	Status:                Active
			#	Path:                  /vol/vol0
			#	Device(s):             urst0a
			#	Progress:              225680 KB
			#	Last Operation Begin:   Tue Jun 15 21:24:51 GMT 2007
			#	Last Operation End:    Tue Jun 15 21:35:20 GMT 2007
			#	Last Operation Size:   6244000 KB
			#	Last Operation Error:  -
			#	Stage:                 Metadata Map
			#	Data Warehouse Begin:  Tue Jun 16 21:24:51 GMT 2007
			#	Data Warehouse End:    Tue Jun 16 21:30:20 GMT 2007
			#	Data Warehouse Blocks: 50000
			#	Metadata Map Begin:    Tue Jun 16 21:30:21 GMT 2007
			#	Metadata Map End:      -
			#	Metadata Map Blocks:   1000
			#	Offset Map Begin:      -
			#	Offset Map End:        -
			#	Offset Map Blocks:     -
			#	
			#	Job ID:         2
			#	Sequence No:    0
			#	Type:                  Restore
			#	Status:                Active
			#	Path:                  /vol/vol1
			#	Device(s):             urst1a
			#	Progress:              -
			#	Last Operation Begin:   -
			#	Last Operation End:    -
			#	Last Operation Size:   -
			#	Last Operation Error:  -
			#	Stage:                 Tree Build
			#	Data Warehouse Begin:  Tue Jun 16 21:04:00 GMT 2007
			#	Data Warehouse End:    Tue Jun 16 21:05:00 GMT 2007
			#	Data Warehouse Blocks: 50000
			#	Metadata Map Begin:    Tue Jun 16 21:05:00 GMT 2007
			#	Metadata Map End:      Tue Jun 16 21:06:00 GMT 2007
			#	Metadata Map Blocks:   1000
			#	Offset Map Begin:      Tue Jun 16 21:06:00 GMT 2007
			#	Offset Map End:        Tue Jun 16 21:06:10 GMT 2007
			#	Offset Map Blocks:     500
			my $s = undef;
			foreach my $line (@lines) {
				if ( $line =~ /([^:]+):\s*(.+)/ ) {
					$s->{lc $1} = $2;
				} else {
					push @stats, $s if exists $s->{'job id'};
					$s = undef;
				}
			}
			push @stats, $s if exists $s->{'job id'};
   	 } else {
			# Example list 'summary' status
			#	Job ID  Seq No  Type    Status Path       Device(s)  Processed
			#	     1       0  Backup  Active /vol/vol0/ urst0a     240 MB Processed
			#	     2       0  Restore Active /vol/vol1/ urst1a     201 MB Restored
			my $title = lc(shift @lines);
			my @keys;
			if ( $title =~ /(\S+\s+\S+)\s+(\S+\s+\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/ ) {
				@keys = ($1, $2, $3, $4, $5, $6, $7);
			} else {
				croak "'smtape status' received unexpected output:\n$text";
			}
			foreach my $line (@lines) {
				$line =~ /\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.+)/;
				my @vals = ($1, $2, $3, $4, $5, $6, $7);
				my %s = map { lc($_) => shift @vals } @keys;
				push @stats, \%s;
			}
	 	}
	}
	croak "No status found"
		.(defined $opts{spec} ? " for spec/job id ".$opts{spec} : "")."..."
		unless @stats;
	return @stats;
}

=head2 ::device_action

Issue the specified tape device 'mt' command. Croak if commands are not successful

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<device> = I<string>

The logical or physical tape device name

B<action> = I<string>

The 'mt' action to issue to the filer

B<count> = I<integer>

I<Optional.>

Indicates the 'count', needed for some commands

=item Examples

=back

	device_action($conn, "rst3a", "eof", 1); # Write 1 eod-of-file marker

=cut

sub device_action {
	my ($conn, $device, $action, $count) = @_;

	my $cmd = "mt -t $device $action $count";
	my $text = sayEOF($conn, $cmd, timeout=>30);
	croak "Command failed: $text"
		if $text =~ /failed|missing command argument/i;
	return $text;
}

=head2 ::device_rewind

Rewind the tape device.

=over

=item Arguments

B<conn> = I<filer connect object>

Filer connection object.

B<device> = I<string>

The logical or physical tape device name

B<marker_cnt> = I<integer>

I<Optional.>

If present indicates to rewind the tape by <marker_cnt> file marker(s). Otherwise rewinds to the beginning
of the tape.

=item Examples

	device_rewind($conn, "rst1", 1); # Rewind 1 file marker
	device_rewind($conn, "rst1");    # Rewind to beginning of the tape

=back

=cut

sub device_rewind {
	my ($conn, $device, $marker_cnt) = @_;

	my $text;
	if ( $marker_cnt ) {
		# Rewind <marker_cnt> file marker(s)
		my $action = "bsf";;
		if ($marker_cnt < 0 ) {
			$action = "fsf";
			$marker_cnt = - $marker_cnt;
		}
		$text = device_action($conn, $device, $action, $marker_cnt);
	} else {
		# Rewind to beginning of tape
		$text = device_action($conn, $device, "rewind");
	}
	return 1;
}

1;

