###############################################################################
#
# InodeHack
#
# Exposes a consistent wrapper around dbladecli's "wreck" and coraltest's
# "readinode" facilities, to allow programmatic manipulation and testing
# of inode values.
#
###############################################################################

package InodeHack;

use strict;
use TracePkg;
use Server;
use Coral;
use InodeHack;

my $trace = new TracePkg;
my $coral = new Coral;

###############################################################################
#
# verifyHackReadiness - Checks to see if your filers are ready to run wafliron
# without panicking.  (Wafliron has a relatively new mode that causes it to
# panic on any corruption; it's on by default for developer builds, and off
# only in promoted builds.)
#
###############################################################################

sub verifyHackReadiness
{
    foreach my $filername ($coral->server()->servers()) {
        $_ = $coral->server()->shell(
                    cmd=>"/bin/kenv 'wafl-iron-panic-on-corruption?'",
                    Server=>$filername);

        $coral->die_config ("$filername is not configured to run wafliron tests yet.\n".
                            "You need to reboot it and issue:\n".
                            "   setenv wafl-iron-panic-on-corruption? false\n".
                            "at the CFE prompt.")
        unless /false/ || /0/;
    }
}

###############################################################################
#
# createInode - Creates a file within the specified volume, returning its inum
#
###############################################################################

sub createInode
{
    my %args = @_;
    my $dsid = $args{dsid};
    my $filename = $args{filename};
    my $geom = $args{geom}; # optional

    my $filer = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
    my $dbladename = $coral->find_volume_dbladename_by_dsid (dsid=>$dsid);

    $filename =~ s/^\///g;

    my $error = $coral->call_dbladecli(
                    server => $filer,
                    command => "priv set test ; mkfile 0 /vol/$dbladename/$filename");

    my $inumhex = undef;
    for (my $retry = 0; ($retry < 5) && (!defined $inumhex); ++$retry) {
        $_ = $coral->call_dbladecli(
                        server => $filer,
                        command => "priv set test ; showfh /vol/$dbladename/$filename");
        $inumhex = $1 if (/fileid=(0x[0-9a-f]*)/i);
    }

    $coral->die_config("unable to create file $filename within volume $dsid ($filer); $error")
        unless defined $inumhex;

    return hex $inumhex;
}

###############################################################################
#
# createManyInodes - Creates lots of files in the given volume, returning an
# array of inums.  This version is a lot more efficient if you're doing more
# than one file at a time.  If you supply a {geom} parameter, the files will
# be fully populated at all DVs.
#
###############################################################################

sub createManyInodes
{
    my %args = @_;
    my $volname = $args{volname};
    my $filename = $args{filename};
    my $testpath = $args{testpath};
    my $count = $args{count};
    my $geom = $args{geom};
    my %geom = %$geom unless not defined $geom;
    my @dsids = split /;/,$geom{dvlist} unless not defined %geom;

    # First create the files

    for (my $ii = 0; $ii < $count; ++$ii) {
        open (FH, '+>', "$testpath/$filename$ii") or die "gridiron test 4: unable to create $testpath/$filename$ii: $!";

        if (scalar (@dsids) != 0) {
            for (my $dvi = 0; $dvi < scalar(@dsids); ++$dvi) {
                sysseek (FH, $geom{width} * $dvi, 0);
                syswrite (FH, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!", 64);
            }
        }

        close (FH);
    }

    # Then go find their inums.  We'll query ten filenames at a time
    # from dbladecli, until we've gotten all of them.

    my $dsid = $coral->find_volume_dsid_by_name(volname=>$volname);
    my $server = $coral->find_filer_by_dsid(dsid=>$dsid, geom=>$geom);
    my $dbladename = $coral->find_volume_dbladename_by_dsid(dsid=>$dsid);

    my @fullresult = ( );
    for (my $start = 0; $start < $count; $start += 10) {

        my $command = "priv set test";
        my $expect = 0;
        for (my $ii = 0; $ii < 10; ++$ii) {
            my $index = $start + $ii;
            last if $index >= $count;
            $command .= " ; showfh /vol/$dbladename/$filename$index";
            ++$expect;
        }

        my @thisresult = ( );
        for (my $retry = 0; ($retry < 5) && (scalar(@thisresult) != $expect); ++$retry) {
            my $out = $coral->call_dbladecli(
                                server=>$server,
                                command=>$command);
            my @out = split(/\n/,$out);

            @thisresult = ( );
            foreach $_ (@out) {
                push @thisresult, hex $1 if /fileid=(0x[0-9a-f]*)/i;
            }
        }

        $coral->die_config("unable to find files within volume $volname ($server)")
            unless scalar(@thisresult) == $expect;

        @fullresult = ( @fullresult, @thisresult );
    }

    return @fullresult;
}

###############################################################################
#
# readInode - Returns the numeric value of a field in a particular inode.
# If the comma-separated "field" input has more than one entry, returns a list.
#
###############################################################################

sub readIntValue
{
    ($_, my $field) = @_;
    $coral->die_fatal("field $field not found in coraltest output ($_)") unless /   $field: (\S+)/;
    my $out = $1;
    chomp $out;

    if ($out =~ /^0x/) {
        return hex $out;
    } else {
        return $out;
    }
}

sub readTimeValue
{
    my $time = readIntvalue(@_);
    time =~ /(\S+)\/(\S+)/;
    return ($1 * 1000000) + ($2 / 1000);   # return time in usec
}

sub readInode
{
    my %args = @_;
    my $dsid = $args{dsid};
    my $inum = $args{inum};
    my $fields = $args{field};
    my $geom = $args{geom};  # optional

    my $filer = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
    my @fields = split /,/,$fields;

    $_ = $coral->call_coraltest(
                    server => $filer,
                    command => "readinode -d $dsid -i $inum -v");
    return undef unless /inode $dsid.$inum is cached/;

    my @output;
    while (scalar(@fields) != 0) {
        my $field = shift @fields;

        if ($field =~ /wi_generation/) {
            push @output, readIntValue($_, "generation");
        } elsif ($field =~ /wi_xinode/) {
            push @output, readIntValue($_, "xinode");
        } elsif ($field =~ /wi_type/) {
            push @output, readIntValue($_, "wafl_type");
        } elsif ($field =~ /wi_atime/) {
            push @output, readTimeValue($_, "atime");
        } elsif ($field =~ /wi_blk_cnt/) {
            push @output, readIntValue($_, "blocks");
        } elsif ($field =~ /wi_crtime/) {
            push @output, readTimeValue($_, "crtime");
        } elsif ($field =~ /wi_ctime/) {
            push @output, readTimeValue($_, "ctime");
        } elsif ($field =~ /wi_coral_dvmask/) {
            push @output, readIntValue($_, "dvmask");
        } elsif ($field =~ /wi_gid/) {
            push @output, readIntValue($_, "gid");
        } elsif ($field =~ /wi_coral_epoch/) {
            push @output, readIntValue($_, "layout_epoch");
        } elsif ($field =~ /wi_mtime/) {
            push @output, readTimeValue($_, "mtime");
        } elsif ($field =~ /wi_nlink/) {
            push @output, readIntValue($_, "nlink");
        } elsif ($field =~ /wi_perm/) {
            push @output, readIntValue($_, "perm");
        } elsif ($field =~ /wi_size/) {
            push @output, readIntValue($_, "size");
        } elsif ($field =~ /wi_subtype/) {
            push @output, readIntValue($_, "subtype");
        } elsif ($field =~ /wi_tid/) {
            push @output, readIntValue($_, "tid");
        } elsif ($field =~ /wi_diskflags/) {
            push @output, readIntValue($_, "true_flags");
        } elsif ($field =~ /wi_uid/) {
            push @output, readIntValue($_, "uid");
        } else {
            $coral->die_fatal("unrecognized field $field expected from InodeHack::readInode");
        }
    }

    return (scalar(@output) == 1) ? $output[0] : @output;
}

###############################################################################
#
# changeInode - Uses wreck to modify a field within an inode.  If you call
# changeInodeBegin first (and pass its return data as "job"), you can batch
# up changes and do them all at once.
#
###############################################################################

sub changeInodeBegin
{
    my %args = @_;
    my $geom = $args{geom};
    my %geom = %$geom;
    my @dsids = split /;/,$geom{dvlist};
    unshift @dsids, $geom{mdv_dsid};

    my %job = ( );
    foreach my $dsid (@dsids) {
        $job{$dsid} = "";
    }

    return \%job;
}

sub changeInodeEnd
{
    my %args = @_;
    my $volname = $args{volname};
    my $job = $args{job};
    my %job = %$job if defined $job;
    my $geom = $args{geom}; # optional

    foreach my $dsid (keys %job) {
        my $server = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
        $_ = $coral->call_coraltest(
                        server => $server,
                        command => "heartbeat off");
    }

    foreach my $dsid (keys %job) {
        my $server = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
        my $dbladename_mroot = $coral->find_mroot_dbladename (server=>$server);
        my $dbladename = $coral->find_volume_dbladename_by_dsid (dsid=>$dsid);

        $coral->server()->shell(
                        cmd=>"cat <<END >/mroot/wreckfile-$dsid\n$job{$dsid}\nEND",
                        Server=>$server);

        my $command = "priv set test ; wreck $dbladename /vol/$dbladename_mroot/wreckfile-$dsid";
        $trace->trace(msg=>"call '$command' in dbladecli");
        $_ = $coral->call_dbladecli(
                            server => $server,
                            timeout => 180,
                            command => $command);

        $coral->die_fatal("wreck (wreckfile-$dsid) failed: $_")
            if /Failed to read or process/;
    }

    foreach my $dsid (keys %job) {
        my $server = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
        $_ = $coral->call_coraltest(
                        server => $server,
                        command => "heartbeat on");
    }

    restoreVolumeAfterChanges (%args);
}

sub changeInode
{
    my %args = @_;
    my $dsid = $args{dsid};
    my $inum = $args{inum};
    my $field = $args{field};
    my $value = $args{value};
    my $job = $args{job};
    my %job = %$job if defined $job;
    my $geom = $args{geom}; # optional

    my $timesec = $value / 1000000;
    my $timensec = ($value % 1000000) * 1000;

    my $command = undef;
    $command = "inode_gen inum=$inum gen=$value" if ($field =~ /wi_generation/);
    $command = "inode_xinode inum=$inum inum2=$value" if ($field =~ /wi_xinode/);
    $command = "inode_type inum=$inum snapid=0 type=$value" if ($field =~ /wi_type/);
    $command = "inode_atime inum=$inum sec=$timesec nsec=$timensec" if ($field =~ /wi_atime/);
    $command = "inode_blk_cnt inum=$inum count=$value" if ($field =~ /wi_blk_cnt/);
    $command = "inode_crtime inum=$inum sec=$timesec nsec=$timensec" if ($field =~ /wi_crtime/);
    $command = "inode_ctime inum=$inum sec=$timesec nsec=$timensec" if ($field =~ /wi_ctime/);
    $command = "inode_coral_dvmask inum=$inum dvmask=$value" if ($field =~ /wi_coral_dvmask/);
    $command = "inode_coral_epoch inum=$inum epoch=$value" if ($field =~ /wi_coral_epoch/);
    $command = "inode_mtime inum=$inum sec=$timesec nsec=$timensec" if ($field =~ /wi_mtime/);
    $command = "inode_link inum=$inum link_cnt=$value" if ($field =~ /wi_nlink/);
    $command = "inode_size inum=$inum size=$value" if ($field =~ /wi_size/);
    $command = "inode_subtype inum=$inum subtype=$value" if ($field =~ /wi_subtype/);
    $command = "inode_tid inum=$inum tid=$value" if ($field =~ /wi_tid/);
    $command = "inode_flags inum=$inum flags=$value" if ($field =~ /wi_diskflags/);
    $command = "inode_perm inum=$inum" if ($field =~ /wi_perm/);  # wreck doesn't let you specify a new value, but just changes it

    $coral->die_fatal("unrecognized field $field expected from InodeHack::changeInode")
        unless defined $command;

    if (defined %job) {
        $job{$dsid} .= $command . "\n";
        return \%job;
    } else {
        my $filer = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
        my $dbladename = $coral->find_volume_dbladename_by_dsid (dsid=>$dsid);

        my $cmd = "priv set test ; wreck $dbladename \"$command\"";
        $trace->trace(msg=>"call '$command' in dbladecli");
        $_ = $coral->call_dbladecli(
                            server => $filer,
                            timeout => 90,
                            command => $cmd);

        $coral->die_fatal("wreck ($command) failed: $_")
            if /Failed to read or process/;
    }
}

###############################################################################
#
# restoreVolumeAfterChanges - Using changeInode drops a volume to Restricted
# mode.  You can do as many changes as you want in that mode, but you'll
# probably want to bring the volume back to Online sooner or later.
#
###############################################################################

sub restoreVolumeAfterChanges
{
    my %args = @_;
    my $aggregate = $args{aggregate};
    my $volname = $args{volname};
    my $vserver = $args{vserver};
    my $vserver = $coral->pick_vserver(volname=>$volname) unless defined $vserver;

    $coral->server()->storage_aggregate_modify(
                        aggregate => $aggregate,
                        state => "online");

    $coral->online(
                   vserver => $vserver,
                   volname => $volname,
                   timeout => 90);
}

###############################################################################
#
# findMaxfiles - Returns an array of integers, each representing the
# value of files-total (a.k.a. "maxfiles") as reported by zapi.
#
###############################################################################

sub findMaxfiles
{
    my %args = @_;
    my $volname = $args{volname};

    my $geom = $coral->get_geometry (volname=>$volname);
    my %geom = %$geom;
    my @dsids = split /;/,$geom{dvlist};
    unshift @dsids, $geom{mdv_dsid};

    my @result = ( );
    foreach my $dsid (@dsids) {
        my $server = $coral->find_filer_by_dsid (dsid=>$dsid, geom=>$geom);
        my $dsid_16 = sprintf("0x%08X", $dsid);

        $_ = $coral->call_zsmcli (
                server=>$server,
                command=>"d-volume-list-info id=$dsid_16 desired-attrs=files-total");

        $coral->die_fatal("unable to find maxfiles for dsid $dsid (asked server $server)")
            unless /<files-total>(\S+)<\/files-total>/;

        push @result, $1;
    }

    return @result;
}

###############################################################################
#
# changeMaxfiles - Adjusts the maxfiles value for a particular flexvol.
# Uses zapi directly, to introduce inconsistencies among constituent volumes.
#
###############################################################################

sub changeMaxfiles
{
    my %args = @_;
    my $volname = $args{volname};
    my $maxfiles = $args{maxfiles};

    my $dsid = $coral->find_volume_dsid_by_name (volname=>$volname);
    my $server = $coral->find_filer_by_dsid (dsid=>$dsid);
    my $dsid_16 = sprintf("0x%08X", $dsid);

    $coral->call_zsmcli (
            server=>$server,
            command=>"d-volume-set-info id=$dsid_16 desired-attrs=files-total volume-attrs=\'[d-volume-info=[files-total=$maxfiles]]\'");
}

###############################################################################
#
# calculateFavDsid - Returns the DSID of the FAV for a particular file.
# This calculation is fast but doesn't work with restriped volumes.  A slower
# but 100% reliable way to do this is to use Coral.pm's wrapper around
# the 'coraltest queryrole' function.
#
###############################################################################

sub calculateFavDsid
{
    my %args = @_;
    my $geom = $args{geom};
    my $inum = $args{inum};

    my %geom = %$geom;
    my @dsids = split /;/,$geom{dvlist};
    my $dvcount = scalar(@dsids);

    return $dsids[ ($inum % 4096) % $dvcount ];
}

###############################################################################
#
# calculateNonFavDsid - Returns the DSID of some DV that is *not* the FAV
# for a particular file.
#
###############################################################################

sub calculateNonFavDsid
{
    my %args = @_;
    my $geom = $args{geom};
    my %geom = %$geom;
    my @dsids = split /;/,$geom{dvlist};
    my $dsidFav = calculateFavDsid(%args);

    for (my $ii = 0; $ii < scalar(@dsids); ++$ii) {
        return $dsids[$ii] unless $dsids[$ii] == $dsidFav;
    }

    return undef;
}

###############################################################################
#
# calculateFavMask - Returns a dvmask showing the FAV for a particular file.
# This calculation is fast but doesn't work with restriped volumes.  A slower
# but 100% reliable way to do this is to use Coral.pm's wrapper around
# the 'coraltest queryrole' function.
#
###############################################################################

sub calculateFavMask
{
    my %args = @_;
    my $geom = $args{geom};
    my $inum = $args{inum};

    my %geom = %$geom;
    my @dsids = split /;/,$geom{dvlist};
    my $dvcount = scalar(@dsids);

    return 1 << ((($inum % 4096) % $dvcount) % 8);
}

###############################################################################
#
# calculateNonFavMask - Returns a dvmask showing a DV that is *not* the FAV
# for a particular file.
#
###############################################################################

sub calculateNonFavMask
{
    my %args = @_;
    my $geom = $args{geom};
    my %geom = %$geom;
    my @dsids = split /;/,$geom{dvlist};
    my $dsidFav = calculateFavDsid(%args);

    for (my $ii = 0; $ii < scalar(@dsids); ++$ii) {
        return (1 << ($ii % 8)) unless $dsids[$ii] == $dsidFav;
    }

    return undef;
}

1;
