## @summary     Provides the ability to create, and config cluster peering interfaces, and
##              relationships.
## @author      cwilcox@netapp.com,dl-cluster-peer@netapp.com
## @pod here

=head1 NAME

Intercluster.pm

=head1 SYNOPSIS

=over 4

 use Server;
 use Intercluster;

 $obj = new Intercluster();

 $obj = $obj->config_intercluster_lifs();

 $obj->create_intercluster_peers(clusters=>\@cluster_array);

 $obj->delete_intercluster_peers(clusters=>\@cluster_array);

 $obj->check_peer_connectivity(clusters=>\@cluster_array);

 $obj->restore_hijacked_lifs();

=back

=head1 DESCRIPTION

C<Intercluster> provides the ability to create, and config cluster peering interfaces, and
relationships.

=head1 REQUIREMENTS

MANDATORY:

=over 4

Create a cluster tharnhost file in the nate_hosts testbed (ie. /x/eng/rtp-testbeds/nate_hosts/rtpgx-3070-1112)
The cluster tharnhost filename must be unique, reside in the site-appropriate depot, and contain the following:

=over 4

  # CLUSTER THARN HOST FILE
  # This file defines specific configuration parameters for a single cluster.
  CLUSTER_NAME=radish-vsim12
  default_ip_gx=10.63.0.69
  hosttype=filer-ng-vsim

  # Required if filer is a VSIM
  conn console {conntype=pm_cons; connhost=radish-vsim1; username=admin ; password=netapp1! }

=back

B<CLUSTER_NAME> C<name-of-cluster> unique cluster name

B<default_ip_gx> is ideally the cluster-mgmt lif IP address. If a valid cluster-mgmt lif is not
available, then one of the node-mgmt lif IPs will work. However, node-mgmt lifs do not failover
in the event of a nodal failure. A cluster-mgmt lif will be available as long as there is a
functioning node in the cluster.

B<hosttype> C<filer-ng-cfe> is the hosttype we are using for a cluster-mode real hardware filer.



=back

OPTIONAL:

=over 4

Add the intercluster lif info to each node tharnhost file in the cluster. To ensure proper configuration,
every node in the cluster must have an intercluster lif which can reside on a port with either data, or
interlcuster roles. If this info does not exist in the node tharnhost file, we will hijack and delete an
existing data lif, and readd it as an intercluster lif.

(ie. /x/eng/rtp-testbeds/nate_hosts/
 rtpgx-3070-11
 rtpgx-3070-12)

 INTERCLUSTER_IP1=9.9.9.90
 INTERCLUSTER_PORT1=e0c
 INTERCLUSTER_MTU1=1500
 INTERCLUSTER_MASK1=255.254.0.0
 INTERCLUSTER_GATEWAY1=9.0.0.10

If a working intercluster lif already exists, we will ignore the INTERCLUSTER tharnhost config info and
use it.

The B<conn> specification is only required if you are working with vsims and
do not have -noconsole specified in your .ntest file.  This specifies which
console to connect to for the cluster.

B<conn> console {conntype=pm_cons; connhost=radish-vsim1; username=admin ; password=netapp1! }

=back

=over 4

Provide the intercluster lif information for each node, in each cluster as an input
hash table with key 'CLUSTERS'.
Each clusters names must be a key in the input hash and must contain as keys all the nodes
that are in the cluster.
Every node in the cluster must have an intercluster lif which can reside on a port with either data, or
interlcuster roles. If this info does not exist in the input varible, we will hijack and delete an
existing data lif, and readd it as an intercluster lif.

(i.e /p4/test/lib/San/UnitTest/Intercluster.thpl)

'CLUSTERS'  => {
				'radish-vsim12' => {
                                    'radish-vsim1' => {
                                                       'INTERCLUSTER_IP1'  =>  "9.9.9.90",
                                                       'INTERCLUSTER_PORT1'=>  "e0c",
                                                       'INTERCLUSTER_MTU1' =>  "1500",
                                                       'INTERCLUSTER_MASK1'=>  "255.254.0.0",
                                                       'INTERCLUSTER_GATEWAY1'=> "9.0.0.10"
                                                      }
	                                'radish-vsim2' => {
                                                       'INTERCLUSTER_IP1'  =>  ...,
                                                       'INTERCLUSTER_PORT1'=>  ...,
                                                        ...
                                                      }
									},
				 'radish-vsim34' => {
								     'radish-vsim3' => {
                                                        'INTERCLUSTER_IP1'  =>  "9.9.9.89",
                                                        'INTERCLUSTER_PORT1'=>  "e0c",
                                                        'INTERCLUSTER_MTU1' =>  "1500",
                                                        'INTERCLUSTER_MASK1'=>  "255.254.0.0",
                                                        'INTERCLUSTER_GATEWAY1'=> "9.0.0.11"
                                                       }
	                                 'radish-vsim4' => {
                                                        'INTERCLUSTER_IP1'  =>  ...,
                                                        'INTERCLUSTER_PORT1'=>  ...,
                                                        ...
                                                        }
									}
			}

If a working intercluster lif already exists, we will ignore the INTERCLUSTER tharnhost config info and
use it.

=back

=head1 FUNCTIONS

B<config_intercluster_lifs> - Creating and configuring lifs/ports to setup cluster peer infrastructure.

B<create_intercluster_peers> - Creating cluster peering relationships.

B<delete_intercluster_peers> - Deleting cluster peering relationships.

B<check_peer_connectivity> - Checking connectivity of the cluster peer networks.

B<restore_hijacked_lifs> - Restore data lifs which had been hijacked to be used as intercluster lifs

=cut

## @summary    Intercluster Library

package Intercluster;

#use statements
use strict;
use warnings;
use Data::Dumper;
use Server;
use Tharn qw($Log);
use Options;
use Params::Common;
use Params::Validate qw(:all);
use Tie::IxHash;
use NACL::GeneralUtils;

use TestUtils;
use TracePkg;
my $Trace = new TracePkg();

use VifMgr::VifConfig;

#my $Server   = new Server();

my $Vif_Cfg  = undef;

my $Options = new Options;
$Options->parameters(
    definitions=>
        [
           {
            name => "CLUSTERS",
            description => "The CLUSTERS that our scripts will operate on.  These must have cluster tharnhost (or cserver tharnhost) files defined for them.",
           },
        ]
);
#if CLUSTERS param not passed in, use CSERVERS param
$Options->compatability(definitions=>{
    CSERVERS => "CLUSTERS"
});
$Options->process();
my $Opt_Ref = $Options->get_hash();

my $Server;
my @Clusters;
my $Orig_Execution_Cluster;

=pod

=head2 new

=over 4

B<Description>
Creates the Intercluster object which is used to house the lif, port, and
peering configuration parameters.

B<Input>

B<Output>

=over 4

=item *

Blessed object reference

=back

=back

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};
    my $args_flag=0;

    # If user did NOT define parameter CLUSTERS but provided input parameters
    # (arguments), then populate the $Opt_ref accordingly to be used with
    # the existing logic.
    # Error out if neither was provided.
    unless ($Options->defined(option=>"CLUSTERS")) {
        if (%args){
            if (_constructor_params_helper(%args)){
                $args_flag = 1;
            } else {
                die "ERROR: While transforming input arguments to Options Reference value";
            }
        } else {
                die "ERROR: Either CLUSTERS=<cluster1>,<cluster2> param is required (a nate ".
                           "host file must exist for each <cluster> included in this parameter) ".
                           "OR an argument should be passed along with the clusters and nodes info. " .
                           "In both cases this library needs to talk to each cluster individually. ";
        }
    }

    # 1) Create the Server object
    $Server   = new Server(nodes=>$Opt_Ref->{CLUSTERS});
    @Clusters = $Server->servers();

    foreach my $cluster (@Clusters) {
        my @nodes;

        # Process the Params Options if no argument was passed.
        $Options->process(hostrec=>$cluster) unless($args_flag);
        $Opt_Ref = $Options->get_hash();

        # 2) Set execution server.
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server=>$cluster);

        # 3) Update the cluster name
        my $cluster_name = $Opt_Ref->{HOSTRECS}{$cluster}->{CLUSTER_NAME};

        if ( $cluster_name ) {
            my $set_name;
            eval {
                $set_name = $Server->cluster_identity_show( );
            };

            if ( $@ ) {
                $Log->comment("cluster_identity_show failed with $@");
            }

            if ( $set_name && ($set_name->[0]{name} ne $cluster_name) ) {
                $Server->cluster_identity_modify( name => $cluster_name );
            }
        }

        # 4) Populate @Nodes array
        my $clusterResult = $Server->cluster_show();
        foreach my $cnodes (@{$clusterResult}) {
            push @nodes, $cnodes->{node};
        }
        $Log->debug("This is the nodes param: ".Dumper(\@nodes));

        # Process the Params Options if no argument was passed.
        $Options->process(hostrec=>\@nodes) unless($args_flag);
        $Opt_Ref = $Options->get_hash();

        # 5) Populate the LIF values for each node
        # If intercluster variables are specified in the tharn host file, save them on the object
        foreach my $node (@nodes) {
            my @Lifs = grep(/^INTERCLUSTER_IP\d+$/, keys %{$Opt_Ref->{HOSTRECS}{$node}});
            if (!@Lifs) {$self->{$cluster}{NODES}{$node}{message} = "Intercluster config does not exist on $node."};
            foreach my $lif (sort _portsort @Lifs) {
                $lif =~ /(\d+)$/;
                my $lif_name = exists($Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_NAME".$1}) ? $Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_NAME".$1} :"intercluster_".$node."_".$1;
                my $port_name = $Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_PORT".$1};
                my $address = $Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_IP".$1};
                my $mtu = $Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_MTU".$1};
                my $mask = $Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_MASK".$1};
                my $gateway = $Opt_Ref->{HOSTRECS}{$node}{"INTERCLUSTER_GATEWAY".$1};

                $self->{$cluster}{NODES}{$node}{PORTS}{$port_name}{LIFS}{$lif_name}{IP} = $address;
                push @{$self->{$cluster}{PEERS}}, $address;
                $self->{$cluster}{NODES}{$node}{PORTS}{$port_name}{LIFS}{$lif_name}{MTU} = $mtu;
                $self->{$cluster}{NODES}{$node}{PORTS}{$port_name}{LIFS}{$lif_name}{MASK} = $mask;
                $self->{$cluster}{NODES}{$node}{PORTS}{$port_name}{LIFS}{$lif_name}{GATEWAY} = $gateway;
            }
        }

        # 6) Set back the execution server to the original value.
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    bless $self, $class;

    return $self;
}

=pod

=head2 config_intercluster_lifs

=over 4

B<Description>
This method is used to configure intercluster lifs and ports on a per node per cluster
basis. All nodes in all clusters need an intercluster lif to properly configure
cluster peering. If a working intercluster lif already exists, it will be used, and the
object updated accordingly.

If an intercluster lif does not exist, it will then look at the configuration information
stored in the object (taken from the node tharn host file).

If that config info does not exist, this method will:

1) If the ics_per_node parameter is specified, will create that many intercluster lifs per
   node, and name them "auto_intercluster_X where X is (0, 1, 2, ...)
   Note: You must specify both a DATA_INTERFACES and GATEWAY variable to define an IP
         pool from which to pull addresses.  Example:
         RTP_DATA_INTERFACES=10.227.71.71/64 RTP_GATEWAY=10.227.64.1
         PTC_DATA_INTERFACES=10.227.71.71/64 PTC_GATEWAY=10.227.64.1

2) If the ics_per_node parameter is not specified, a data lif on one of the data ports
   will be "hijacked" (i.e. it will be deleted/recreated as an intercluster lif)

3) If hijack_port is specified, we look for 2 data ports.  One of the data ports (src) will
   have it's lifs moved to the other port (dst), the src port will have it's role changed
   to "intercluster" and the intercluster lifs will be created on the src port.  If hijack_port
   is not specified, the intercluster lifs will be created on a data port.


B<Input>
hijack_port - boolean, 1 - move data lifs off of data port, change port role to be "intercluster"
                       0 - create intercluster lifs on data port.

ics_per_node - integer, - Number of intercluster lifs to create on each node

B<Output>

=over 4

=item *

B<object reference>
Updated Intercluster object

=back

=back

=cut

sub config_intercluster_lifs {
    my $self = shift;
    my $me = (caller(0))[3];
    my @hijacked_lifs;
    my $hijacked_route;
    my $intercluster_routing_group;
    my $i;

    my %params = validate(
        @_,
        {
           hijack_port  => { type  => SCALAR,
                             optional => 1,
                             default => undef,
                           },
           ics_per_node  => { type  => SCALAR,
                             optional => 1,
                             default => undef,
                           },
           move_cluster_mgmt_lifs  => { type  => SCALAR,
                             optional => 0,
                             default => undef,
                           },
           #if name_prefix is defined, any lifs created with that prefix will not
           #be cleaned up with the restore routine.  The name_prefix will only
           #be applied to lifs created by this library where the IP configuration
           #was passed into the script or exists in the nate_host but wasnt
           #already configured on the filer.
           name_prefix  => { type  => SCALAR,
                             optional => 1,
                             default => undef,
                           },
        }
    );

    # Revert all lifs and see if we need to clear out PEERS
    foreach my $cluster (keys %{$self}) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        # Make sure all lifs are home
        _revert_lifs( $Server );

        # If these addresses got in here via a previous call to config_intercluster_lifs()
        # delete them
        foreach my $node (keys %{$self->{$cluster}{NODES}}) {
            if (exists $self->{$cluster}{NODES}{$node}{message}) {
                if ( exists $self->{$cluster}{PEERS} ) {
                    delete $self->{$cluster}{PEERS};
                }
            }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    # Check to see if intercluster lif already exists.
    foreach my $cluster (keys %{$self}) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        foreach my $node (keys %{$self->{$cluster}{NODES}}) {
            my $inter_lifs;
            eval {
                $inter_lifs = $Server->network_interface_show(
                    server        => $node,
                    'home-node'   => $node,
                    role          => "intercluster",
                    'status-oper' => "up");
            };


            # If there are no intercluster lifs, and the tharn config
            # does not exist build intercluster lifs as described in the
            # comment header block.

            my $inter_lifs_find_rc = $@;
            my $no_inter_lifs_found = 0;
            if ($inter_lifs_find_rc ) {
                if ( $inter_lifs_find_rc =~ /There are no entries matching your query/) {
                    $no_inter_lifs_found = 1;
                } else {
                    die "$me: Attempted to query UP intercluster lifs on node $node, but failed with $@";
                }
            }

            if ( $no_inter_lifs_found ) {
                if (exists $self->{$cluster}{NODES}{$node}{message}) {
                    if ( $params{hijack_port} ) {
                        if ( $params{ics_per_node} ) {
                            $Log->comment("We have no intercluster lifs on $node. We will create $params{ics_per_node} ip addresses");
                        } else {
                            $Log->comment("We have no intercluster lifs on $node. We will hijack a data port");
                        }

                        my $data_ports;
                        eval {
                            $data_ports =
                                $Server->network_port_show(
                                    'node' => $node,
                                    'role' => "data",
                                    'link' => "up"
                                );
                        };

                        # If there are no data ports, error out
                        if ($@ =~ /There are no entries matching your query/) {
                            die "$me: ERROR: There are no data ports to hijack.";
                        } elsif ( $@ ) {
                            die "$me: ERROR: network_port_show failed with $@";
                        } elsif (scalar @{$data_ports} == 1) {
                            die "$me: ERROR: Only one data port remains. Need 2 so we can convert one into IC and migrate its lifs to the other.";
                        } else {

                            my $dst_port;
                            my $src_port;

                            _find_data_ports( $node, \$src_port, \$dst_port, $data_ports );

                            my $migrate_lifs;
                            if ( $src_port && $dst_port ) {
                                $Log->comment("MIGRATE: src_port $src_port, dst_port $dst_port");
                                eval {
                                    $migrate_lifs =
                                        $Server->network_interface_show(
                                            'home-node'   => $node,
                                            'home-port'   => $src_port,
                                            'curr-node'   => $node,
                                        );
                                };

                                if ( $@ ) {
                                    die "$me: ERROR: network_port_show failed with $@";
                                } elsif ( !$@ && $migrate_lifs ) {

                                    # Find an up data lif to hijack
                                    my $hijack_index = -1;
                                    my $num_lifs = scalar( @{$migrate_lifs} );

                                    # Only find a port to hijack if we are not creating multiple
                                    # intercluster lifs per node
                                    if ( !$params{ics_per_node} ) {
                                        for ( $i = 0; ($hijack_index < 0) && ($i < $num_lifs); $i++ ) {
                                            if ( ($migrate_lifs->[$i]{'role'} eq "data") &&
                                                 ($migrate_lifs->[$i]{'status-oper'} eq "up") ) {
                                                $hijack_index = $i;
                                            }
                                        }
                                    }

                                    # Need to pick one up data lif from $src_port as the hijack lif
                                    # Delete all other lifs from $src_port and recreate on $dst_port
                                    # with a name which will allow us to move them back later.
                                    # Then we delete the hijack lif from $src_port, change the hijack port
                                    # role to intercluster, then re-create the hijacked lif on $src_port
                                    # as role intercluster

                                    if ( ($hijack_index < 0) && !$params{ics_per_node} ) {
                                        die "$me: ERROR: Could not find an up data port to hijack.";
                                    } else {
                                        $num_lifs = scalar( @{$migrate_lifs} );

                                        if ( $hijack_index >= 0 ) {
                                            $Log->comment("Lif to be hijacked is ".$migrate_lifs->[$hijack_index]{lif});
                                        }

                                        $hijacked_route = undef;

                                        for ( $i = 0; $i < $num_lifs; $i++ ) {

                                            if ( $i == $hijack_index ) {
                                                # grab existing data lif routing-group info
                                                eval {
                                                    $hijacked_route = $Server->network_routing_groups_route_show(
                                                        'routing-group' => $migrate_lifs->[$i]{'routing-group'});
                                                };
                                                if ( $@ ) {
                                                    $Log->comment("network_routing_groups_route_show failed with $@");
                                                }
                                            }

                                            # Move all other lifs to $dst_port
                                            if ( $i != $hijack_index ) {
                                                $Server->network_interface_migrate(
                                                    server      => $migrate_lifs->[$i]{server},
                                                    lif         => $migrate_lifs->[$i]{lif},
                                                    'dest-node' => $migrate_lifs->[$i]{'home-node'},
                                                    'dest-port' => $dst_port
                                                    );
                                                $Server->network_interface_modify(
                                                    server      => $migrate_lifs->[$i]{server},
                                                    lif         => $migrate_lifs->[$i]{lif},
                                                    'home-port' => $dst_port
                                                    );
                                                $Server->network_interface_rename(
                                                    server      => $migrate_lifs->[$i]{server},
                                                    lif         => $migrate_lifs->[$i]{lif},
                                                    newname => "migrated__".
                                                                   $src_port."__".
                                                                   $migrate_lifs->[$i]{'server'}."__".
                                                                   $migrate_lifs->[$i]{lif}
                                                    );

                                            }
                                        }

                                        $Log->debug("Lifs have been migrated from src $src_port to dst $dst_port");

                                        if ( $hijack_index >= 0 ) {

                                            # Delete lif we are going to hijack
                                            my $retries = 3;
                                            while ( $retries > 0 ) {
                                                $retries--;
                                                eval {
                                                    $Server->network_interface_delete(
                                                        server => $migrate_lifs->[$hijack_index]{server},
                                                        lif    => $migrate_lifs->[$hijack_index]{lif},
                                                        -Timeout => 60 );
                                                };
                                                if ( !$@ ) {
                                                    last;
                                                } else {
                                                    my $err = "$me: Lif delete (migrate_lifs) failed, possibly due to BURT 582052 with $@";
                                                    if ( $retries > 0 ) {
                                                        $Log->warn($err);
                                                    } else {
                                                        die $err;
                                                    }
                                                    Tharn::sleep( 10 );
                                                }
                                            }
                                        }

                                        # Change hijacked data port role to intercluster
                                        $Server->network_port_modify(
                                            node      => $node,
                                            port      => $src_port,
                                            role      => "intercluster",
                                            );

                                        if ( $hijack_index >= 0 ) {
                                            # recreate lif as role=intercluster
                                            # new lif name will be like "hijacked_ic__vs0__user-vsim1_data2"

                                            $Server->network_interface_create(
                                                server      => $migrate_lifs->[$hijack_index]{'home-node'},
                                                lif         => "hijacked_ic__".
                                                               $migrate_lifs->[$hijack_index]{'server'}."__".
                                                               $migrate_lifs->[$hijack_index]{lif},
                                                address     => $migrate_lifs->[$hijack_index]{address},
                                                netmask     => $migrate_lifs->[$hijack_index]{netmask},
                                                'home-port' => $migrate_lifs->[$hijack_index]{'home-port'},
                                                'home-node' => $migrate_lifs->[$hijack_index]{'home-node'},
                                                'role'      => "intercluster",
                                                -Timeout    => 60 );

                                            push @{$self->{$cluster}{PEERS}}, $migrate_lifs->[$hijack_index]{address};
                                        } elsif ( $params{ics_per_node} ) {

                                            # Only initialize Vif_Config object when we are going to use it.
                                            $self->_init_vif_config();

                                            for ( my $lif_index = 0; $lif_index < $params{ics_per_node}; $lif_index++ ) {
                                                my $ip = $Vif_Cfg->get_address();

                                                $Vif_Cfg->create_vif(
                                                     server       => $node,
                                                     home_node    => $node,
                                                     home_port    => $src_port,
                                                     vif          => $params{name_prefix}."auto_intercluster_${node}_${lif_index}",
                                                     address      => $ip,
                                                     role         => "intercluster",
                                                );

                                                push @{$self->{$cluster}{PEERS}}, $ip;
                                            }
                                        }

                                        if ( $hijacked_route ) {
                                            # Create intercluster route
                                            my $intercluster_routing_group = $migrate_lifs->[$hijack_index]{'routing-group'};

                                            # Once IC lif was created, a route-group of the same name, but prefaced by "i"
                                            # instead of "d"
                                            $intercluster_routing_group =~ s/d/i/;

                                            $Server->network_routing_groups_route_create(
                                                server          => $migrate_lifs->[$hijack_index]{'home-node'},
                                                'routing-group' => $intercluster_routing_group,
                                                destination     => $hijacked_route->[0]{destination},
                                                gateway         => $hijacked_route->[0]{gateway});
                                        }
                                    }
                                }

                            } else {
                                die "$me: ERROR: Could not find 2 data ports, cannot hijack one of them.";
                            }
                        }

                    } else {  # Don't want to hijack a port and change it to be an IC port, just build IC lifs on a data port
                        # If we give the number of ICs per node, we will create these from scratch
                        if ( $params{ics_per_node} ) {
                            my $dst_port;
                            my $src_port;

                            _find_data_ports( $node, \$src_port, \$dst_port );

                            $Log->trace("We have no intercluster lifs on $node. We will auto-create $params{ics_per_node} IC lifs on port $src_port");

                            # Only initialize Vif_Config object when we are going to use it.
                            $self->_init_vif_config();

                            for ( my $lif_index = 0; $lif_index < $params{ics_per_node}; $lif_index++ ) {
                                my $ip = $Vif_Cfg->get_address();

                                $Vif_Cfg->create_vif(
                                     server       => $node,
                                     home_node    => $node,
                                     home_port    => $src_port,
                                     vif          => $params{name_prefix}."auto_intercluster_${node}_${lif_index}",
                                     address      => $ip,
                                     role         => "intercluster",
                                );

                                push @{$self->{$cluster}{PEERS}}, $ip;
                            }
                        } else {
                            # here we will hijack a single data lif and make it an IC lif
                            $Log->trace("We have no intercluster lifs on $node. We will have to hijack a data lif");
                            my $data_lifs;
                            eval {
                                $data_lifs = $Server->network_interface_show(
                                    'home-node'   => $node,
                                    'curr-node'   => $node,
                                    role          => "data",
                                    'status-oper' => "up");
                            };

                            # If there are no working data lifs to hijack, error out.
                            if ($@ =~ /There are no entries matching your query/) {
                                die "$me: ERROR: There are no remaining data lifs to hijack.";
                            } elsif ( $@ ) {
                                die "$me: ERROR: network_interface_show failed with $@";
                            } elsif (scalar @{$data_lifs} == 1) {
                                die "$me: ERROR: Only one data lif remains. Hijacking it would sever the node from the data network.";
                            } else {
                                # Hijack: grab existing data lif info, delete, and recreate as intercluster lif.
                                $i = rand @{$data_lifs};
                                push (@hijacked_lifs, $data_lifs->[$i]);

                                $hijacked_route = $Server->network_routing_groups_route_show(
                                        'routing-group' => $data_lifs->[$i]{'routing-group'});

                                push @{$self->{$cluster}{PEERS}}, $data_lifs->[$i]{address};


                                my $retries = 3;
                                while ( $retries > 0 ) {
                                    $retries--;
                                    eval {
                                        $Server->network_interface_delete(
                                            server   => $data_lifs->[$i]{server},
                                            lif      => $data_lifs->[$i]{lif},
                                            -Timeout => 60 );
                                    };
                                    if ( !$@ ) {
                                        last;
                                    } else {
                                        my $err = "$me: Lif delete (data_lifs) failed, possibly due to BURT 582052 with $@";
                                        if ( $retries > 0 ) {
                                            $Log->warn($err);
                                        } else {
                                            die $err;
                                        }
                                        Tharn::sleep( 10 );
                                    }
                                }

                                # new lif name will be like "hijacked_ic__vs0__user-vsim1_data2"
                                $Server->network_interface_create(
                                    server      => $data_lifs->[$i]{'home-node'},
                                    lif         => "hijacked_ic__".
                                                   $data_lifs->[$i]{'server'}."__".
                                                   $data_lifs->[$i]{lif},
                                    address     => $data_lifs->[$i]{address},
                                    netmask     => $data_lifs->[$i]{netmask},
                                    'home-port' => $data_lifs->[$i]{'home-port'},
                                    'home-node' => $data_lifs->[$i]{'home-node'},
                                    'role'      => "intercluster",
                                    -Timeout    => 60 );

                                # Configure intercluster routing group info.
                                $intercluster_routing_group = $data_lifs->[$i]{'routing-group'};
                                $intercluster_routing_group =~ s/d/i/;

                                $Server->network_routing_groups_route_create(
                                    server          => $data_lifs->[$i]{'home-node'},
                                    'routing-group' => $intercluster_routing_group,
                                    destination     => $hijacked_route->[0]{destination},
                                    gateway         => $hijacked_route->[0]{gateway});
                            }
                        }
                    }
                } else {
                    # If intercluster information exists in the node tharn file, use it to
                    # configure the intercluster lif.
                    foreach my $port (keys %{$self->{$cluster}{NODES}{$node}{PORTS}}) {
                        foreach my $lif (keys %{$self->{$cluster}{NODES}{$node}{PORTS}{$port}{LIFS}}) {
                            $Server->network_interface_create(
                                server => $node,
                                lif => $lif,
                                address => $self->{$cluster}{NODES}{$node}{PORTS}{$port}{LIFS}{$lif}{IP},
                                netmask => $self->{$cluster}{NODES}{$node}{PORTS}{$port}{LIFS}{$lif}{MASK},
                                'home-port' => $port,
                                'home-node' => $node,
                                'role' => "intercluster",
                                -Timeout => 60 );

                            my $inter_lif = $Server->network_interface_show(
                                server => $node,
                                lif => $lif);

                            # If two more or intercluster lifs share the same routing-group, this will
                            # fail with Error: command failed: duplicate entry
                            my $rc = eval { $Server->network_routing_groups_route_create(
                                server => $inter_lif->[0]{'home-node'},
                                'routing-group' => $inter_lif->[0]{'routing-group'},
                                gateway => $self->{$cluster}{NODES}{$node}{PORTS}{$port}{LIFS}{$lif}{GATEWAY});
                                1; };
                            if ( !$rc && $@ !~ /duplicate entry/i ) {
                                die "$me: $@";
                            }
                        }
                    }

                    eval {
                        $inter_lifs = $Server->network_interface_show(
                            server        => $node,
                            'home-node'   => $node,
                            role          => "intercluster",
                            'status-oper' => "up");
                    };
                    my $inter_lifs_find_rc = $@;
                    my $no_inter_lifs_found = 0;
                    if ($inter_lifs_find_rc ) {
                        if ( $inter_lifs_find_rc =~ /There are no entries matching your query/) {
                            $no_inter_lifs_found = 1;
                            $inter_lifs = [];
                        } else {
                            die "$me: Attempted to query UP intercluster lifs on node $node, but failed with $@";
                        }
                    }

                    #now check the ics_per_node requested, if more were requested than currently exist, lets try to create the new ones
                    my $num_existing_ic_lifs=@{$inter_lifs};
                    if ( $params{ics_per_node} and ($params{ics_per_node} > $num_existing_ic_lifs) ) {
                        if ( $no_inter_lifs_found ) {
                            die "$me: There should have been SOME UP intercluster lifs on node $node, but there were none";
                        }
                        my $src_port=$inter_lifs->[0]{'home-port'};
                        $Log->comment("interlifs less than ics_per_node, creating another ic lif \n");
                        $Log->comment("We will auto-create IC lifs on port $src_port");
                        #
                        ## Only initialize Vif_Config object when we are going to use it.
                        $self->_init_vif_config();
                        #
                        for ( my $lif_index = $num_existing_ic_lifs; $lif_index < $params{ics_per_node}; $lif_index++ ) {
                            my $ip = $Vif_Cfg->get_address();
                            $Log->comment("Creating lif $lif_index on $node on port $src_port with IP $ip \n");
                            $Vif_Cfg->create_vif(
                                server       => $node,
                                home_node    => $node,
                                home_port    => $src_port,
                                vif          => "auto_additional_intercluster_${node}_${lif_index}",
                                address      => $ip,
                                role         => "intercluster",
                            );
                            push @{$self->{$cluster}{PEERS}}, $ip;
                        }
                    }
                }
            } else {
                # If there is an existing intercluster lif, use it, and store the values
                # inside the object.
                $i = rand @{$inter_lifs};
                $Log->comment("Using existing intercluster lif: $inter_lifs->[$i]{lif}");
                if (exists $self->{$cluster}{NODES}{$node}) {
                    delete $self->{$cluster}{NODES}{$node};
                    delete $self->{$cluster}{PEERS};
                }
                $self->{$cluster}{NODES}{$node}{PORTS}{$inter_lifs->[$i]{'home-port'}}{LIFS}{$inter_lifs->[$i]{lif}}{IP} = $inter_lifs->[$i]{address};
                push @{$self->{$cluster}{PEERS}}, $inter_lifs->[$i]{address};
                $self->{$cluster}{NODES}{$node}{PORTS}{$inter_lifs->[$i]{'home-port'}}{LIFS}{$inter_lifs->[$i]{lif}}{MASK} = $inter_lifs->[$i]{netmask};
                my $inter_port = $Server->network_port_show(
                    port => $inter_lifs->[$i]{'home-port'},
                    node => $inter_lifs->[$i]{'home-node'});

                $self->{$cluster}{NODES}{$node}{PORTS}{$inter_lifs->[$i]{'home-port'}}{LIFS}{$inter_lifs->[$i]{lif}}{MTU} = $inter_port->[0]{mtu};
                my $inter_route = $Server->network_routing_groups_route_show(
                    'routing-group' => $inter_lifs->[$i]{'routing-group'},
                    server => $inter_lifs->[$i]{server});

                $self->{$cluster}{NODES}{$node}{PORTS}{$inter_lifs->[$i]{'home-port'}}{LIFS}{$inter_lifs->[$i]{lif}}{GATEWAY} = $inter_route->[0]{gateway};

                #now lets check if hijack_port was set, if so then we'll want to ensure the port the pre-configured intercluster lifs are on are intercluster ports.
                #If not, we'll move all interfaces off of that port, reconfigure as interlcuster, and move the intercluster lif back on.

                #1)  Check if hijack_port is set.
                #2)  If so, check if port intercluster lif is on is a intercluster port.  If it is, no action needed.
                #3)  If not an intercluster port, get a listing of the data lifs on that port.
                #4)  Migrate those data lifs to another data port (rename them so the restore routine works)
                #5)  Migrate the intercluster lif there for now.
                #6) Make the old port an intercluster port
                #7) Migrate the intercluster lif back there.

                if ($params{hijack_port}) {
                    if ($inter_port->[0]{role} ne "intercluster") {
                        my $avail_data_ports;
                        eval {
                            $avail_data_ports=$Server->network_port_show(
                                node => $inter_lifs->[$i]{'home-node'},
                                role => "data",
                                link => "up",
                                port => "!$inter_lifs->[$i]{'home-port'}");
                        };
                        # If there are no other data ports, error out
                        if ($@ =~ /There are no entries matching your query/) {
                            die "$me: ERROR: The current port is the only data port, so we can't convert it into an IC port and migrate the lifs.";
                        }
                        else {
                            #find data or cluster mgmt lifs on the port
                            my $lifs_to_move;
			    my $role;
			    #Optionally move cluster mgmt lifs as well
                            $role =
				($params{move_cluster_mgmt_lifs})  ?
                                     "data|cluster-mgmt"  :
                                     "data";  

                            eval {
                                $lifs_to_move=$Server->network_interface_show(
                                    role => $role,
                                    "home-node" => $inter_lifs->[$i]{'home-node'},
                                    "home-port" => $inter_lifs->[$i]{'home-port'});
                            };

                            #if there are no  lifs on our port, then we are good
                            if ($@ =~ /There are no entries matching your query/) {
                                $Log->comment("No lifs to move, proceeding to configure intercluster port\n");
                            }
                            #else, we'll need to move those lifs off the current port
                            else {
                                foreach my $lif_move(@{$lifs_to_move}) {
                                    $Server->network_interface_migrate(
                                                    server      => $lif_move->{server},
                                                    lif         => $lif_move->{lif},
                                                    'dest-node' => $lif_move->{'home-node'},
                                                    'dest-port' => $avail_data_ports->[0]{port}
                                                    );
                                    $Server->network_interface_modify(
                                                    server      => $lif_move->{server},
                                                    lif         => $lif_move->{lif},
                                                    'home-port' => $avail_data_ports->[0]{port}
                                                    );
                                    $Server->network_interface_rename(
                                                    server      => $lif_move->{server},
                                                    lif         => $lif_move->{lif},
                                                    newname => "migrated__".
                                                                   $lif_move->{'home-port'}."__".
                                                                   $lif_move->{'server'}."__".
                                                                   $lif_move->{lif}
                                                    );
                                }
                            }
                            #now move the current intercluster lif
                            foreach my $inter_lif_move(@{$inter_lifs}) {
                                $Server->network_interface_migrate(
                                                    server      => $inter_lif_move->{server},
                                                    lif         => $inter_lif_move->{lif},
                                                    'dest-node' => $inter_lif_move->{'home-node'},
                                                    'dest-port' => $avail_data_ports->[0]{port}
                                                    );
                                $Server->network_interface_modify(
                                                    server      => $inter_lif_move->{server},
                                                    lif         => $inter_lif_move->{lif},
                                                    'home-port' => $avail_data_ports->[0]{port}
                                                    );
                            }
                            # Change hijacked data port role to intercluster
                            $Server->network_port_modify(
                                            node => $inter_lifs->[$i]{'home-node'},
                                            role => "intercluster",
                                            port => $inter_lifs->[$i]{'home-port'}
                                            );
                            #migrate the ic lifs back
                            foreach my $inter_lif_move(@{$inter_lifs}) {
                                $Server->network_interface_migrate(
                                                    server      => $inter_lif_move->{server},
                                                    lif         => $inter_lif_move->{lif},
                                                    'dest-node' => $inter_lif_move->{'home-node'},
                                                    'dest-port' => $inter_lif_move->{'home-port'},
                                                    );
                                $Server->network_interface_modify(
                                                    server      => $inter_lif_move->{server},
                                                    lif         => $inter_lif_move->{lif},
                                                    'home-port' => $inter_lif_move->{'home-port'},
                                                    );
                                $Server->network_interface_rename(
                                                    server      => $inter_lif_move->{server},
                                                    lif         => $inter_lif_move->{lif},
                                                    newname => "hijacked_ic_migrated__".
                                                                   $inter_lif_move->{'server'}."__".
                                                                   #$inter_lifs->{'home-port'}."__".
                                                                   $inter_lif_move->{lif}
                                                    );
                            }
                        }
                    }
                }




                #now check the ics_per_node requested, if more were requested than currently exist, lets try to create the new ones
                my $num_existing_ic_lifs=@{$inter_lifs};
                if ( $params{ics_per_node} and ($params{ics_per_node} > $num_existing_ic_lifs) ) {
                    my $src_port=$inter_lifs->[$i]{'home-port'};
                    $Log->comment("interlifs less than ics_per_node, creating another ic lif \n");
                    $Log->comment("We will auto-create IC lifs on port $src_port");
                    #
                    ## Only initialize Vif_Config object when we are going to use it.
                    $self->_init_vif_config();
                    #
                    for ( my $lif_index = $num_existing_ic_lifs; $lif_index < $params{ics_per_node}; $lif_index++ ) {
                        my $ip = $Vif_Cfg->get_address();
                        $Log->comment("Creating lif $lif_index on $node on port $src_port with IP $ip \n");
                        $Vif_Cfg->create_vif(
                            server       => $node,
                            home_node    => $node,
                            home_port    => $src_port,
                            vif          => "auto_additional_intercluster_${node}_${lif_index}",
                            address      => $ip,
                            role         => "intercluster",
                        );
                        push @{$self->{$cluster}{PEERS}}, $ip;
                    }
                }
            }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }
}

=pod

=head2 restore_hijacked_lifs

=over 4

B<Description>
This method is used to restore data lifs that were hijacked to
become intercluster lifs, or lifs that were migrated
to another port so that port could be hijacked to become
an intercluster port.

The lifs have been renamed to contain information necessary to
restore them to what they were prior to running config_intercluster_lifs()

B<Input>

none

B<Output>

none

=back

=cut
sub restore_hijacked_lifs {
    my $self = shift;
    my $hijacked_lifs_restored = 0;
    my @hijacked_lifs;
    my $hijacked_route;
    my $data_routing_group;
    my $i;
    my $num_lifs;

    # config_intercluster_lifs() can work in 3 modes:
    # 1) hijack a extra data lif on a data port and change the role of that lif to be IC
    # 2) hijack an extra data port by migrating all but one of the data lifs to another
    #    data port, then changing the role of the port to be IC, and recreating
    #    the remaining data lif as role IC.
    # 3) If INTERCLUSTER configuration data is specified in the tharn hosts file, use that
    #    to configure the lifs.

    # Some things to be aware of:
    # 1) An IC lif can be created on an IC port or a data port, but if a port has lifs,
    #    the port role cannot be changed, even if it would normally be legal.  So when
    #    we change the port role from IC to data, we have to delete ALL lifs first.
    # 2) Lif creation will result in routing-group creation, but we will then have to
    #    create the route ourselves.
    # 3) When un-migrating, we do not delete routing-group or route as the lif role
    #    has not changed (data/data) but the port has.
    # 4) When un-hijacking, we DO need to delete the IC route and routing-group

    my %ic_routing_group;

    foreach my $cluster (keys %{$self}) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        # Make sure all lifs are home
        _revert_lifs( $Server );

        my $hijacked_lifs;
        my $auto_lifs;
        eval {
            $hijacked_lifs =
                $Server->network_interface_show(
                    lif => "hijacked_ic_*",
                    role          => "intercluster",
                );
        };

        # first delete lifs and save info in order to recreate the original lifs
        if ( $hijacked_lifs && ($@ !~ /There are no entries matching your query/) ) {

            $num_lifs = scalar( @{$hijacked_lifs} );
            for ( $i = 0; $i < $num_lifs; $i++ ) {

                $hijacked_lifs->[$i]{orig_vserver} = undef;
                $hijacked_lifs->[$i]{orig_lif}     = undef;

                # Do a minimal match on the port name, maximal on lif name
                if ( $hijacked_lifs->[$i]{lif} =~ /hijacked_ic__(\S+?)__(\S+)/ or
                        $hijacked_lifs->[$i]{lif} =~ /hijacked_ic_migrated__(\S+?)__(\S+)/) {
                    $hijacked_lifs->[$i]{orig_vserver} = $1;
                    $hijacked_lifs->[$i]{orig_lif}     = $2;
                }

                if ( $hijacked_lifs->[$i]{orig_vserver} &&
                     $hijacked_lifs->[$i]{orig_lif}) {

                    $hijacked_route = undef;
                    if ($hijacked_lifs->[$i]{lif} !~ /hijacked_ic_migrated__(\S+?)__(\S+)/) {
                        eval {
                            $hijacked_route = $Server->network_routing_groups_route_show(
                                'routing-group' => $hijacked_lifs->[$i]{'routing-group'},
                                'role' => "intercluster" );
                        };
                    }

                    my $retries = 3;
                    while ( $retries > 0 ) {
                        $retries--;
                        eval {
                            $Server->network_interface_delete(
                                server   => $hijacked_lifs->[$i]{server},
                                lif      => $hijacked_lifs->[$i]{lif},
                                -Timeout => 60 );
                        };
                        if ( !$@ ) {
                            last;
                        } else {
                            my $err = "restore_hijacked_lifs: Lif delete failed, possibly due to BURT 582052 with $@";
                            if ( $retries > 0 ) {
                                $Log->warn($err);
                            } else {
                                die $err;
                            }
                            Tharn::sleep( 10 );
                        }
                    }

                    # We may have to change the role of the port, so
                    # defer lif creation until AFTER the port role change

                    if ( $hijacked_route ) {

                        # need to save routing group info so we can delete it once all
                        # IC routes have been deleted
                        if ( !defined $ic_routing_group{$cluster} ) {
                            $ic_routing_group{$cluster} =
                                [ { "name"   => $hijacked_route->[0]{'routing-group'},
                                    "server" => $hijacked_route->[0]{server} } ];
                        } else {
                            push ( @{$ic_routing_group{$cluster}},
                                   { "name"   => $hijacked_route->[0]{'routing-group'},
                                     "server" => $hijacked_route->[0]{server} } );
                        }

                        eval {
                            # delete IC route
                            $Server->network_routing_groups_route_delete(
                                server          => $hijacked_route->[0]{server},
                                'routing-group' => $hijacked_route->[0]{'routing-group'},
                                destination     => $hijacked_route->[0]{destination},
                                gateway         => $hijacked_route->[0]{gateway},
                                -Timeout        => 60 );
                        };
                        if ( $@ ) {
                            $Log->warn("Routing groups route delete failed with $@");
                        }
                    }
                }
            }
        }

        #delete any additionally created auto lifs
        eval {
            $auto_lifs =
                $Server->network_interface_show(
                    lif => "auto_additional_intercluster_*",
                    role          => "intercluster",
                );
        };

        if ( $auto_lifs && ($@ !~ /There are no entries matching your query/) ) {

            $num_lifs = scalar( @{$auto_lifs} );

            # Only initialize Vif_Config object when we are going to use it.
            $self->_init_vif_config();

            for ( $i = 0; $i < $num_lifs; $i++ ) {

                # By default will delete the routing group too.
                $Vif_Cfg->delete_vif(
                    server => $auto_lifs->[$i]{server},
                    vif    => $auto_lifs->[$i]{lif});
            }
        }

        #delete auto created lifs on hijacked port
        eval {
            $auto_lifs =
                $Server->network_interface_show(
                    lif => "auto_intercluster_*",
                    role          => "intercluster",
                );
        };

        # delete auto created lifs
        if ( $auto_lifs && ($@ !~ /There are no entries matching your query/) ) {

            $num_lifs = scalar( @{$auto_lifs} );

            # Only initialize Vif_Config object when we are going to use it.
            $self->_init_vif_config();

            for ( $i = 0; $i < $num_lifs; $i++ ) {

                # By default will delete the routing group too.
                $Vif_Cfg->delete_vif(
                    server => $auto_lifs->[$i]{server},
                    vif    => $auto_lifs->[$i]{lif});
            }
        }

        my $lifs =  [];
        if ( $hijacked_lifs ) {
            push( @$lifs, @$hijacked_lifs );
        }

        if ( $auto_lifs ) {
            push( @$lifs, @$auto_lifs );
        }

        #
        # If we changed port role to be intercluster, restore it to be
        # data
        $num_lifs = scalar( @{$lifs} );
        for ( $i = 0; $i < $num_lifs; $i++ ) {

            my $hijacked_port;
            eval {
                $hijacked_port = $Server->network_port_show(
                    node      => $lifs->[$i]{'home-node'},
                    port      => $lifs->[$i]{'home-port'},
                    role      => "intercluster",
                    );
            };

            if ( $hijacked_port && ($@ !~ /There are no entries matching your query/) &&  ($lifs->[$i]{"lif"} =~ /^hijacked_ic/)) {
                # Change hijacked port's role
                $Server->network_port_modify(
                    node      => $lifs->[$i]{'home-node'},
                    port      => $lifs->[$i]{'home-port'},
                    role      => "data",
                    );
            }
        }

        # Restore any lifs that we hijacked to use as intercluster
        if ( $hijacked_lifs ) {
            $hijacked_lifs_restored++;
            $num_lifs = scalar( @{$hijacked_lifs} );
            for ( $i = 0; $i < $num_lifs; $i++ ) {
                # recreate the hijacked lifs as intercluster or data|cluster_mgmt lifs
                if ($hijacked_lifs->[$i]{lif}=~ /^hijacked_ic_migrated__/) {
                    $Server->network_interface_create(
                        server      => $hijacked_lifs->[$i]{orig_vserver},
                        lif         => $hijacked_lifs->[$i]{orig_lif},
                        address     => $hijacked_lifs->[$i]{address},
                        netmask     => $hijacked_lifs->[$i]{netmask},
                        'home-port' => $hijacked_lifs->[$i]{'home-port'},
                        'home-node' => $hijacked_lifs->[$i]{'home-node'},
                        'role'      => "intercluster",
                        -Timeout    => 60 );
                }
                else {
                    $Server->network_interface_create(
                        server      => $hijacked_lifs->[$i]{orig_vserver},
                        lif         => $hijacked_lifs->[$i]{orig_lif},
                        address     => $hijacked_lifs->[$i]{address},
                        netmask     => $hijacked_lifs->[$i]{netmask},
                        'home-port' => $hijacked_lifs->[$i]{'home-port'},
                        'home-node' => $hijacked_lifs->[$i]{'home-node'},
                        'role'      => $hijacked_lifs->[$i]{role},
                        -Timeout    => 60 );
                }
            }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    # Move migrated lifs back to their original port
    # both source and dest at this point will have the same role, i.e. data
    foreach my $cluster (keys %{$self}) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        my $migrated_lifs;
        eval {
            $migrated_lifs =
                $Server->network_interface_show(
                    lif => "migrated__*",
                );
        };

        if ( $migrated_lifs && ($@ !~ /There are no entries matching your query/) ) {

            $num_lifs = scalar( @{$migrated_lifs} );
            for ( $i = 0; $i < $num_lifs; $i++ ) {

                $migrated_lifs->[$i]{orig_port}    = undef;
                $migrated_lifs->[$i]{orig_vserver} = undef;
                $migrated_lifs->[$i]{orig_lif}     = undef;

                # Do a minimal match on the port and vserver names, maximal on lif name
                if ( $migrated_lifs->[$i]{lif} =~ /migrated__(\S+?)__(\S+?)__(\S+)/ ) {
                    # destination port is the original port
                    $migrated_lifs->[$i]{orig_port}    = $1;
                    $migrated_lifs->[$i]{orig_vserver} = $2;
                    $migrated_lifs->[$i]{orig_lif}     = $3;
                }

                if ( $migrated_lifs->[$i]{orig_port} &&
                     $migrated_lifs->[$i]{orig_vserver} &&
                     $migrated_lifs->[$i]{orig_lif} ) {

                    $Server->network_interface_migrate(
                        server      => $migrated_lifs->[$i]{server},
                        lif         => $migrated_lifs->[$i]{lif},
                        'dest-node' => $migrated_lifs->[$i]{'home-node'},
                        'dest-port' => $migrated_lifs->[$i]{orig_port},
                    );
                    $Server->network_interface_modify(
                        server      => $migrated_lifs->[$i]{server},
                        lif         => $migrated_lifs->[$i]{lif},
                        'home-port' => $migrated_lifs->[$i]{orig_port},
                    );
                    $Server->network_interface_rename(
                        server      => $migrated_lifs->[$i]{server},
                        lif         => $migrated_lifs->[$i]{lif},
                        newname     => $migrated_lifs->[$i]{orig_lif},
                    );
                }
            }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    # Delete the ic routing groups created when the hijacked lifs
    # were re-created as IC lifs

    foreach my $cluster (keys %ic_routing_group) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        foreach my $group ( @{$ic_routing_group{$cluster}} ) {
            eval {
                # delete IC route
                $Server->network_routing_groups_delete(
                    server          => $group->{server},
                    'routing-group' => $group->{name},
                    -Timeout        => 60 );
            };
            if ( $@ ) {
               $Log->warn("Routing groups delete failed with $@");
            }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    # added so that we won't try to re-use the IC lifs once we've restored the
    # hijacked lifs
    if ( $hijacked_lifs_restored ) {
        foreach my $cluster (keys %{$self}) {
            $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);
    
            foreach my $node (keys %{$self->{$cluster}{NODES}}) {
                if (exists $self->{$cluster}{NODES}{$node}) {
                    delete $self->{$cluster}{NODES}{$node};
                }
                # If we had to hijack lifs, then there must not have been any definitions
                $self->{$cluster}{NODES}{$node}{message} = "Intercluster config does not exist on $node.";
            }
            $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
        }
    }
}

=pod

=head2 create_intercluster_peers

=over 4

B<Description>
This method is used to create an intercluster peer relationship between the source and
destination clusters. Each cluster passed into the method will be used in configuring
the cluster peering.

B<Input>

=over 4

=item *

B<clusters> - Optional
An array of clusters that will be used in creating the peering relationships.

=back

B<Output>

none

=back

=cut

sub create_intercluster_peers {
    my $self = shift;
    my %params = validate(
    @_,
    {
       clusters => { type  => ARRAYREF,
                     optional => 1,
                     default => undef,
                   },
       timeout  => { type  => SCALAR,
                     optional => 1,
                     default => 25,
                   },
       username => { type => SCALAR,
                     optional => 1,
                     default => "admin",
                   },
       password => { type => SCALAR,
                     optional => 1,
                     default => "netapp1!",
                   },
       peer_addrs => { type => SCALAR,
                       optional => 1,
                       default => undef,
                     },
       use_names => { type => SCALAR,
                       optional => 1,
                       default => 0,
                     },
    }
    );

    # Normalize the cluster array to be all, or whatever the user
    # defines in the cluster parameter.
    my @Clusters;
    if ($params{clusters}) {
        @Clusters = @{$params{clusters}};
    } else {
        @Clusters = $Server->servers();
    }

    # Set the minimum number of clusters to be two.
    my $num_clusters = scalar(@Clusters);

    if ($num_clusters < 2) {
        die "ERROR: The number of clusters in the array must be two or more.";
    }

    my @all_clusters = @Clusters;

    # Use a cluster from the array to launch the peer creation from
    my $seed_cluster = shift (@Clusters);
    $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$seed_cluster);

    my $notice = 'Inter-cluster clock skew too great';

    foreach my $peer_cluster (@Clusters) {
        my $peer_address;
        my $output;
        $Log->debug("This is the peer cluster object: ".Dumper($self->{$peer_cluster}{PEERS}));
        my $i = rand @{$self->{$peer_cluster}{PEERS}};
        if ($params{peer_addrs}) {
            $peer_address = $params{peer_addrs};
        } else {
            $peer_address = $self->{$peer_cluster}{PEERS}[$i];
        }

        if ( $params{use_names}) {
            $peer_address = _get_hostname( $peer_address );
            if ( !$peer_address ) {
                $Log->error(die "ERROR: Could not get hostname for address |$peer_address|");
            }
        }

        eval { $Server->cli(cmd=>"debug mtrace buffer clear local mgwd xc_mgr"); };

        eval {
            $Server->cluster_peer_create(
                '-Verification' => 0,
                output => \$output,
                'peer-addrs' => $peer_address,
                username => $params{username},
                password => $params{password},
                timeout => $params{'timeout'},
            );
        };

        if ( $@ ) {
            $Log->warn("cluster peer create failed with $@");
            eval { $Server->cli(cmd=>"debug mtrace log show local mgwd xc_mgr -fields emit-time,filename,line-num,message"); };
        }

        if ($output =~ $notice) {
            # delete intercluster peers
            $Server->cluster_peer_delete(
                '-Verification' => 0,
                cluster => $peer_cluster,
            );

            # since we are going to die anyhow, go through all peers and show date
            # so we can verify that the clocks truely are out of sync
            my @times;
            foreach my $peer_cluster (@all_clusters) {
                $Server->set_execution_server(-Server =>$peer_cluster);
                my $date = $Server->system_node_date_show();
            }

            $Log->error(die "ERROR: The time differences between nodes in all clusters have caused the peer create to "
                    ."fail. Synchronize the \<system node date\> on each node, and try again.");
        }
    }
    $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
}

=pod

=head2 delete_intercluster_peers

=over 4

B<Description>
This method is used to delete an intercluster peer relationship between the source and
destination clusters. Each cluster passed into the method will be used in configuring
the cluster peering.

B<Input>

=over 4

=item *

B<clusters> - Optional
An array of clusters that will be used in creating the peering relationships.

=back

B<Output>
none

=back

=cut

sub delete_intercluster_peers {
    my $self = shift;
    my %params = validate(
    @_,
    {
       clusters => { type  => ARRAYREF,
                     optional => 1,
                     default => undef,
                   },
    }
    );

    # Normalize the cluster array to be all, or whatever the user
    # defines in the cluster parameter.
    my $peers;
    my @Clusters;
    if ($params{clusters}) {
        @Clusters = @{$params{clusters}};
    } else {
        @Clusters = $Server->servers();
    }

    # Set the minimum number of clusters to be two.
    my $num_clusters = scalar(@Clusters);
    if ($num_clusters < 2) {
        die "ERROR: The number of clusters in the array must be two or more.";
    }

    my $delete_peer_addresses = 0;

    # Foreach cluster, delete the peer relationships
    foreach my $peer_cluster (@Clusters) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$peer_cluster);

        # clear out so we don't have residue from a previous loop
        $peers = undef;
        eval {
            $peers = $Server->cluster_peer_show();
        };

        foreach my $peer (@{$peers}) {
            if ( $peer->{cluster} ) {
                $Server->cluster_peer_delete(
                    '-Verification' => 0,
                    cluster => $peer->{cluster},
               );
           } else {
               $Log->comment("Result from cluster peer show = ". Dumper( $peers ) );
               die "ERROR: The cluster name from cluster_peer_show is null.";
           }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

}

=pod

=head2 check_peer_connectivity

=over 4

B<Description>
This method is used to check connectivity between clusters by performing a full-mesh ping
between each cluster's dblade.

B<Input>

=over 4

=item *

B<clusters> - Optional
An array of clusters that will be used in creating the peering relationships.

=back

B<Output>

=over 4

=item *

B<success>
undef

=item *

B<failure>
aoh reference to cluster ping table output

=back

=back

=cut

sub check_peer_connectivity {
    my $self = shift;
    my %params = validate(
    @_,
    {
       clusters => { type  => ARRAYREF,
                     optional => 1,
                     default => undef,
                   },
    }
    );

    # Normalize the cluster array to be all, or whatever the user
    # defines in the cluster parameter.
    my @Clusters;
    if ($params{clusters}) {
        @Clusters = @{$params{clusters}};
    } else {
        @Clusters = $Server->servers();
    }

    my ($src_cluster, $peer_cluster) = (undef, undef);

    ## See burt 544751
    ## This code creates a race condition between the job manager and
    ## the script. I am commenting it out for now until a proper fix
    ## can be implemented.

#    # See if there is a Peer Manager job on each cluster
#    foreach $src_cluster (@Clusters) {
#        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$src_cluster);
#         foreach $peer_cluster (@Clusters) {
#             # Want to check that there is a peer manager job for each peer
#             if ( $src_cluster ne $peer_cluster ) {
#                 eval {
#                         $Server->job_show(
#                             affinity => "Cluster",
#                             name     => "\"Peer Manager for cluster $peer_cluster\"",
#                             state    => 'Queued|Running',
#                     );
#                 };
#
#                 # If no entries found, die
#                 if ($@) {
#                     if ($@ =~ /There are no entries matching your query/) {
#                         eval {
#                                 $Server->job_show(
#                                     affinity => "Cluster",
#                             );
#                         };
#                         die "ERROR: No Queued or Running Peer Manager job found on $src_cluster for $peer_cluster.";
#                     } else {
#                         die "ERROR: Encountered error \"$@\" when querying Peer Manager job on cluster $src_cluster";
#                     }
#                 } else {
#                     $Log->comment("Peer Manager job found on $src_cluster for $peer_cluster");
#                 }
#             }
#        }
#        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
#    }

    # Run the cluster ping from each cluster to its peer cluster.
    foreach $src_cluster (@Clusters) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$src_cluster);
        foreach $peer_cluster (@Clusters) {

            # No point in trying to peer ping to/from yourself
            if ( $src_cluster eq $peer_cluster ) {
                next;
            }

            my $ping_retries = 3;
PING:

            my $reachable  = $self->peer_ping( dest_cluster => $peer_cluster, type => "data" );
               $reachable &= $self->peer_ping( dest_cluster => $peer_cluster, type => "icmp" );

            if ( !$reachable && ($ping_retries > 0) ) {
                $ping_retries--;
                Tharn::snooze(10);
                goto PING;
            }

            if ( !$reachable ) {
                if ($peer_cluster eq $src_cluster) {
                    $Log->warn("Hit BURT 394219 cluster peer ping reports adjacent in-cluster interfaces as \'unreachable\'. Keep moving on!");
                } else {
                    die "ERROR: Could not reach cluster $peer_cluster from $src_cluster.";
                }
            }
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    return undef;
}

sub debug_info {

    my $self = shift;
    my @Clusters = $Server->servers();

    # Show debugging information
    foreach my $cluster (@Clusters) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        eval {$Server->cluster_show();};
        eval {$Server->cluster_ring_show();};
        eval {$Server->network_interface_show();};
        eval {$Server->network_port_show();};
        eval {$Server->cli(cmd=>"debug mtrace log show local mgwd xc_mgr -fields emit-time,filename,line-num,message");};
        eval {$Server->cli(cmd=>"debug mtrace log show local mgwd ZapiInterface -category spinnp-ping");};
        eval {$Server->cluster_peer_show();};
        eval {$Server->cluster_peer_health_show();};
        eval {$Server->cli(cmd=>"debug csm ct show");};
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }
}

sub clear_debug_info {

    my $self = shift;
    my @Clusters = $Server->servers();

    # Show debugging information
    foreach my $cluster (@Clusters) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        eval { $Server->cli(cmd=>"debug mtrace buffer clear local mgwd xc_mgr"); };
        eval { $Server->cli(cmd=>"debug mtrace buffer clear local mgwd ZapiInterface"); };
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }
}

# Peer ping the specified peer cluster
# Input parameters:
#    dest_cluster  - Name of the destination cluster to ping
#    type          - Type of ping (data|icmp)
#    count         - Number of times to ping
#
sub peer_ping {
    my $self = shift;
    my %params = validate(
        @_,
        {
           dest_cluster => { type  => SCALAR,
                             optional => 0,
                           },
           type         => { type  => SCALAR,
                             optional => 1,
                             default => "data",
                           },
           count        => { type  => SCALAR,
                              optional => 1,
                              default => undef,
                           },
           'packet-size' => { type  => SCALAR,
                              optional => 1,
                              default => undef,
                           },
           timeout      => { type  => SCALAR,
                              optional => 1,
                              default => undef,
                           },
        }
    );

    my $ping = undef;
    my @ping = ();
    my $reachable = 0;

    # only add count, size or timeout if they are not undef.
    my %opts;
    foreach my $opt ("count", "packet-size", "timeout" ) {
        if ( defined $params{$opt} ) {
            $opts{$opt} = $params{$opt};
        }
    }

    # ping default output is not strictly "show"
    # formatted, so we need to pass in the fields option.
    eval {
        $ping = $Server->cluster_peer_ping(
            'destination-cluster' => $params{dest_cluster},
            'fields'              => 'status',
            'type'                => $params{type},
            %opts
        );
    };

    my $cmd_err = $@;

    # need to jump through a few hoops since the automatic
    # string to array of hash conversion is only automatically
    # done when the command is "show".
    $Server->parse_cli_output(output_in=>$ping, want_array=>\@ping);

    if ( !$cmd_err && @ping ) {
        my $num_entries = scalar(@ping);
        my $num_up = 0;

        foreach my $item (@ping) {
            if ( ( "$item->{'status'}" =~ /interface[-_]reachable/i ) ) {
                $num_up++;
            }
        }

        if ( $num_entries  && ($num_up == $num_entries) ) {
            $Log->comment("Found $num_up, expected $num_entries entries to be \"interface\[-_\]reachable\" via \"$params{type}\" ping");
            $reachable = 1;
        } elsif ( $num_entries ) {
            $Log->warn("Found $num_up, expected $num_entries entries to be \"interface\[-_\]reachable\" via \"$params{type}\" ping");
        } else {
            $Log->warn("Ping via $params{type} returned $num_entries entries, should be non-zero");
        }
    } else {
        $Log->comment("Cluster peer ping failed with $cmd_err");
    }

    return $reachable;
}

# Check the health of the cluster peers
# Input parameters:
#   originating-node - Filter health check by the specified originating node
#   destination-cluster - filter health check to destination cluster
#   destination-node - Filter health check by destination node
#   destination-cluster-uuid - filter health check by destination uuid
#   data-ping - filter health check for data-ping results
#   icmp-ping - filter health check for icmp-ping results
#   node-health - filter by node health
#   cluster-health - filter by cluster-healt
#   availability - filter by availability
#   bypass-cache - enable health check real time rather than using the cache
#

sub health_check {
    my $self = shift;
    my %params = validate(
        @_,
        {
           'originating-node'       =>  { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'destination-cluster'    =>  { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'destination-node'       => { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'destination-cluster-uuid' => { type  => SCALAR,
                                            optional => 1,
                                            default => undef,
                                            },
           'data-ping'              => { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'icmp-ping'              => { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'node-health'              => { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'cluster-health'              => { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'availability'              => { type  => SCALAR,
                                        optional => 1,
                                        default => undef,
                                        },
           'bypass-cache'              => { type  => SCALAR,
                                        optional => 1,
                                        default => "true",
                                        },
        }
    );

    my $health_result;

    # only add defined params
    while ( my ($key, $value) = each(%params) ) {
        if (!defined($value)) {
            delete $params{$key}; #remove any params that don't have values configured
        }
    }

    #check cluster peer health
    eval {
        $health_result = $Server->cluster_peer_health_show(
            %params
        );
    };

    #if cluster peer health show returns an error, then return failure
    if ($@) {
        $Log->comment("Error returned while attempting cluster peer health show, $@\n");
        return 0;
    }

    foreach my $result (@$health_result) {
        if ($result->{'availability'} eq "false") {
            $Log->comment("Peer was unavailble\n");
            return 0;
        }
    }

    return 1;


} #end health_check


# Private Methods

# _portsort() is used in the new() method to sort the gathered configuration output.
# This keeps the config together so that it can be correctly processed and stored
# within the object.
sub _portsort {
    $a =~ /(\d+)$/;
    my $an = $1;
    $b =~ /(\d+)$/;
    my $bn = $1;

    $an <=> $bn;
}

# revert_lifs() will revert all lifs to their home nodes/ports
# Input parameters:
#    RETRIES  - Number of attempts (default 5)
#    SLEEP    - Number of seconds to sleep between attempts (default 5)
sub _revert_lifs {
    my $Server = shift;
    my %params =
        validate(
            @_,
            {
               RETRIES => { type  => SCALAR,
                             optional => 1,
                             default => 5,
                           },
               SLEEP => { type  => SCALAR,
                             optional => 1,
                             default => 5,
                           },
            }
        );

    my $not_homed_lifs = undef;

    # Make sure lifs are homed
    for ( my $home_retries = $params{RETRIES}; $home_retries > 0; $home_retries-- ) {

        $not_homed_lifs = undef;
        eval {
            $not_homed_lifs =
                $Server->network_interface_show(
                    'is-home'   => "false",
                );
        };

        # If we get an error other than "there are no entries", or there are
        # non-homed lifs, then try to revert
        if ( ($@ !~ /There are no entries matching your query/) ||
             ($not_homed_lifs) ) {

            eval {
                 $Server->network_interface_revert(server=>"*", lif =>"*");
            };

            # If there was an error, sleep and try again
            if ( $@ ) {
                $Log->comment("network_interface_revert failed with $@");
                Tharn::snooze($params{SLEEP});
            }
        } else {
            # don't try anymore
            last;
        }
    }

    if ( $not_homed_lifs ) {
        die("Could not home all lifs on cluster");
    }
}

# Enable or disable NTP
sub _set_ntp {
    my $control = shift;
    my $flag    = "true";
    my $enable  = "Enabling";

    if ( !$control ) {
        $flag   = "false";
        $enable = "Disabling";
    }

    my @Clusters = $Server->servers();
    foreach my $cluster (@Clusters) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        my $retries = 3;

RETRY:
        eval {
            $Log->comment($enable." NTP service on all nodes");
            $Server->system_service_ntp_config_modify( enabled => $flag );
        };

        if ( $@ && ($retries > 0) ) {
            $Log->comment("system_service_ntp_config_modify failed with $@");
            $retries--;
            Tharn::snooze(5);
            goto RETRY;
        }
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }
}

# Sync all nodes in all clusters with automation host local time
# Input parameters:
#    enable_ntp           - Should ntp be enabled? (default 1)
sub set_local_time {
    my $self = shift;

    my %params =
    (
        enable_ntp  => 1,
        @_,
    );

    my $enable_ntp = $params{enable_ntp};

    my @Clusters = $Server->servers();

    # Disable NTP
    eval { _set_ntp( 0 ) };
    if ( $@ ) {
        # In SN, they are turning NTP on by default, not allowing you to disable it
        # and allowing you to modify the date without explicitly disabling NTP
        if ( $@ =~ /NTP cannot be/ ) {
            $enable_ntp = 0;
        } else {
            $Log->warn("Intercluster::set_local_time() failed while disabling NTP with $@");
        }
    }

    my ($sec,$min,$hour,$day,$mon,$year,undef) = localtime();
    $year += 1900;
    $mon++;
    my $date = sprintf("\"%d/%d/%d %d:%02d:%02d\"", $mon, $day, $year, $hour, $min, $sec);

    foreach my $cluster (@Clusters) {
        $Orig_Execution_Cluster=$Server->set_execution_server(-Server =>$cluster);

        my $dt = $Server->system_node_date_show();

        $Server->system_node_date_modify(
            node     => "*",
            date     => $date,
            timezone => $dt->[0]{timezone}
        );
        $Server->set_execution_server(-Server=>$Orig_Execution_Cluster);
    }

    if ( $enable_ntp ) {
        # Enable NTP
        eval { _set_ntp( 1 ) };
    }

}

# Return all ports on a node that DATA ports
# Input parameters:
#    node           - Node name
#    src_port_ref   - Reference to source port
#    dst_port_ref   - Reference to destination port
#    data_ports     - Results from "network port show"
#
sub _find_data_ports {
    my $num_args = scalar( @_ );
    my $node = shift;
    my $src_port_ref = shift;
    my $dst_port_ref = shift;

    my $data_ports = undef;

    if ( $num_args >= 4 ) {
        $data_ports = shift;
    }

    # default src and dst ports
    $$src_port_ref = undef;
    $$dst_port_ref = undef;

    my $i;

    if ( !$data_ports ) {
        eval {
            $data_ports =
                $Server->network_port_show(
                    'node' => $node,
                    'role' => "data",
                    'link' => "up"
                );
        };

        # If there are no data ports, error out
        if ($@ =~ /There are no entries matching your query/) {
            die "ERROR: There are no data ports.";
        } elsif ( $@ ) {
            die "ERROR: network_port_show failed with $@";
        } elsif (scalar @{$data_ports} == 1) {
            die "ERROR: Only one data port found. Need 2 data ports.";
        }
    }

    if ( $data_ports ) {

        # Build  $port_spec = "p1|p2|p3" from @$data_ports
        my $port_spec = "";
        my $num_ports = scalar( @{$data_ports} );
        for ( $i = 0; $i < $num_ports; $i++ ) {
            if ( $port_spec ) {
                $port_spec = $port_spec."|".$data_ports->[$i]{'port'};
            } else {
                $port_spec = $data_ports->[$i]{'port'};
            }
        }
        $Log->debug("data ports are |$port_spec|");

        my $mgmt_lifs;
        eval {
            $mgmt_lifs =
                $Server->network_interface_show(
                    'home-node'   => $node,
                    'home-port'   => $port_spec,
                    'curr-node'   => $node,
                    role          => "node-mgmt",
                );
        };

        # There should be a data port with node-mgmt interfaces on it, if so, pick
        # the one with the node-mgmt as the dest, if not, pick
        # the first data port as dest and 2nd as src
        if ( $mgmt_lifs && ($@ !~ /There are no entries matching your query/) ) {
            $$dst_port_ref = $mgmt_lifs->[0]{'home-port'};

            for ( $i = 0; ($i < $num_ports) && !$$src_port_ref; $i++ ) {
                if ( $data_ports->[$i]{'port'} ne $$dst_port_ref ) {
                    $$src_port_ref = $data_ports->[$i]{'port'};
                }
            }
        } else {
            $$dst_port_ref = $data_ports->[0]{'port'};
            $$src_port_ref = $data_ports->[1]{'port'};
        }
    }
}

sub _get_hostname {
    my $name = shift;
    my $output;
    my $resolved = undef;

    $Server->shell(
            cmd   => "host $name",
            output => \$output,
    );

    if ( $output =~ /not found/ ) {
        $Log->error("ERROR: Original Name |$name| could not be resolved");
    } else {
        my @output = split(/ /, $output);
        my $resolved = $output[$#output];

        $resolved =~ s/\.$//;
        $Log->debug("Original Name |$name| Resolved name = |$resolved|");
    }

    return $resolved;
}

# Initialize the VifConfig object if not already initialized.
# Input parameters:
#    self           - Intercluster object reference
#
#  Note: VifConfig.pm requires SITE and DATA_MASK1 to be defined on the NODES
#        that Server.pm returns.  Trouble is, that for Intercluster tests, our
#        NODES aren't filers, they are clusters.  Since this is the case, the
#        cluster nate hosts files contain just enough information to connect
#        to a host.  To make VifConfig.pm happy, we copy the required parameters
#        over from one of the cluster's nodes.  This would imply a limitation
#        that all nodes in a cluster are in the same SITE (seems reasonable)
#        and that all nodes in a cluster share the same DATA_MASK1.  Not sure how
#        reasonable that is, but this is just a default and it can be overridden
#        in 2 different ways:
#
#        1) pass in DATA_IF_NETMASK on the command-line
#        2) pass in a node specific if_netmask when calling VifConfig::create_vif()
#
#        Because we have to mess around with the host info before we create the
#        VifConfig object, we can't just create this at global scope, but instead have
#        to create it the first time before we attempt to use it.
#
sub _init_vif_config {
    my $self = shift;
    if ( !$Vif_Cfg ) {
        foreach my $cluster (keys %{$self}) {
            my @nodes = (keys %{$self->{$cluster}{NODES}});
            foreach my $node (@nodes) {
                my $node_host    = Tharn::host($node);
                my $cluster_host = Tharn::host($cluster);

                # Set defaults on cluster_host that Vif_Config complains about
                # if they are missing.
                # If you specify DATA_IF_NETMASK on the command-line
                # when you run your script, that will be taken in preference
                # to DATA_MASK1.  Also, if you specify if_netmask when you create
                # your vif, that will also override.
                foreach my $var (qw( SITE DATA_MASK1 )) {
                    if ( !defined $cluster_host->{$var} || !$cluster_host->{$var} ) {
                        if ( $node_host->{$var} ) {
                            $cluster_host->{$var} = $node_host->{$var};
                        } elsif ( $var eq 'SITE' ) {
                            my $opt_ref = $Options->get_hash();
                            if ( $opt_ref->{SITE} ) {
                                $cluster_host->{$var} = $opt_ref->{SITE};
                            } else {
                                $Log->error("No valid SITE available for cluster $cluster: |$cluster_host->{$var}|");
                            }
                        }
                        $Log->debug("Copying node $node $var to cluster $cluster: |$cluster_host->{$var}|");
                    } else {
                        $Log->debug("Var $var was already defined on cluster $cluster: |$cluster_host->{$var}|");
                        $Log->debug("No need to copy node $node $var |$node_host->{$var}| to cluster $cluster");
                    }
                }
            }
        }

        $Vif_Cfg  = VifMgr::VifConfig->new();
    }
}


# Private Methods for the constructor new()
#
# _constructor_params_helper() is used in the new() method to populate in the
# $Options object instance with the attributes that the library is expected to
# retrieve from the nate files, when the nate files does not contain the information
# it has to be strip out of the input paramenters and adecuately inserted in the
# optrions reference object.
#
# This method keeps the original names, formats and tags that the $Options Reference
# object would have if the nate hosts files had been used.
sub _constructor_params_helper {
    my %params = validate(@_, { 'CLUSTERS' => {type  => HASHREF,
                                               optional => 0
                                              }
                               }
                        );
    # Loclas
    my $clusters_ref = $params{'CLUSTERS'};
    my $cluster_names = '';
    my @nodes = ();

    # Get the cluster names
    foreach my $cluster (keys %{$clusters_ref}) {
        $Opt_Ref->{HOSTRECS}{$cluster}->{CLUSTER_NAME} = $cluster;
        $cluster_names .= $cluster.',';
        foreach my $node (keys %{$clusters_ref->{$cluster}}) {
            foreach my $intercluster_key (keys %{$clusters_ref->{$cluster}->{$node}}) {
                my $value = $clusters_ref->{$cluster}->{$node}->{$intercluster_key};
                $Opt_Ref->{HOSTRECS}{$node}{$intercluster_key} = $value;
            }
        }
    }

    $Options->set('CLUSTERS'=> substr($cluster_names, 0, -1) );
    $Opt_Ref = $Options->get_hash();

    return 1;

} #END _constructor_params_helper

1;
