# Hey emacs, this is basically -*- perl -*-

#
## @summary
## @status public
## @pod here

=pod

=head1 NAME

DscSetup

=head1 DESCRIPTION

=over

=item Provide methods for DscSetup

=back

=head1 AUTHOR

  pavans@netapp.com

=head1 COPYRIGHT

  Copyright 2013, NetApp Inc., All Rights Reserved.

=cut

package DscSetup;

#
# Compiler directives.
#
use strict;
use warnings;
use Data::Dumper;
use Tharn qw($Log subtest);
use Params::Validate
  qw(validate validate_with OBJECT SCALAR ARRAYREF HASHREF BOOLEAN UNDEF);
use NACL::APISet;
use NACL::Transit;
use NACL::C::Slist;
use NACL::C::Client;
use NACL::C::Volume;
use NACL::C::Snapshot;
use NACL::C::DebugVreport;
use NACL::C::AggregateSnapshot;
use NACL::C::VolumeEfficiencyPolicy;
use NACL::STask::Lun;
use NACL::STask::Node;
use NACL::STask::Qtree;
use NACL::STask::Volume;
use NACL::STask::Aggregate;
use NACL::STask::VolumeFile;
use NACL::MTask::VolumeQuota;
use NACL::STask::VolumeClone;
use NACL::STask::Client::Mount;
use NACL::STask::VolumeEfficiency;
use NACL::STask::VolumeSnapshot;
use NACL::MTask::LunMount;
use NACL::MTask::LunSetup;
use NACL::MTask::Traffic;
use NACL::MTask::Wafliron;
use NACL::CS::Volume;
use NACL::CS::Ifconfig;
use NACL::CS::Aggregate;
use NACL::CS::Interface;
use NACL::CS::DebugVreport;
use NACL::CS::VolumeSnapshot;
use NACL::CS::Client::Sanlun;
use NACL::CS::AggregateSnapshot;
use NACL::CS::StorageAggregateWafliron;

use NATE::BaseException qw(:try);
use NACL::Exceptions::TrafficFailed;
use NACL::C::Exceptions::VolumeQtree::AlreadyExists ();
use NACL::GeneralUtils qw(random_name_generator);
use NACL::C::UnitNormalization qw(:convert :constants);
use File::Basename qw(fileparse);

use IronUtil::DSC::commonParams;
use constant SYSTEM_LOG_FILES_7M => 'messages';
use constant SYSTEM_LOG_FILES_CM => 'messages.log';

use TCD ();
use Readonly;
use NATE::ParamSet;

Readonly my %FILE_SIZES => (
    'k' => 1024,
    'm' => 1024 * 1024,
    'g' => 1024 * 1024 * 1024,
);

our $SUDO_PATH = 'sudo';

my $Test = new TCD( -testcases => [] );

########################################################################
#                            Public Methods
########################################################################

=head2 METHODS

=over

=item new()

Objective: Create new DscSetup object.

=back

=cut

sub new {
    my $class = shift;

    # Create a hash and bless the hash to create an object.
    my $self = {};
    bless $self, $class;

    my @nodes = NACL::STask::Node->find();
    $self->{node} = $nodes[0];    # Default to first node
    $self->{node}{apiset} = $self->{node}->get_7m_or_nodescope_apiset();
    $self->{node}{apiset}->set_timeout( 'connectrec-timeout' => 300 );

    my @clients = NACL::C::Client->find();
    my ( $Unix_client, $Windows_client ) = ();

    foreach my $client (@clients) {
        if ( $client->hostrec()->OS() =~ /Linux/ ) {
            $Unix_client = $client;
        }
        elsif ( $client->hostrec()->OS() =~ /Windows/ ) {
            $Windows_client = $client;
        }
    } ## end foreach

    my $conn          = $Unix_client->hostrec()->conn();
    my $Client_apiset = $Unix_client->apiset();
    while ( my ( $k, $v ) = each(%$conn) ) {
        if ( defined( $v->{username} ) && defined( $v->{password} ) ) {
            my $Username = $v->{username};
            my $Password = $v->{password};
            last;
        }
    } ## end while

    $self->{client_apiset} = $Client_apiset;
    $self->{client_conn}   = $conn;
    $self->{client}        = $Unix_client;
    $self->{client_apiset}
      ->set_privilege( 'privilege-level' => 'root' );

    # Init to store lun info
    $self->{lun_mountobj} = ();
    $self->{lun_device}   = ();
    $self->{lun_iohash}   = ();

    # Create client C object
    my $client_name = $Unix_client->client();
    $self->{client_obj} = NACL::C::Client->new(name => $client_name);

    # Include dsc default param setting
    $self->{iron_cparams} = IronUtil::DSC::commonParams->new();
    $Log->exit();

    return $self;
} ## end sub new

########################################################################
#  Member function of DscSetup class
########################################################################

# This sets up the container.
# It makes use of commonParams
sub setup_container {
    my $self              = shift;
    my $cont_params       = $self->{iron_cparams}{IRON_CONTAINER_PARAMS};
    my $aggr_created_flag = 0;
    my ( %aggr_options, %other_options );
    $Log->enter();

    $Log->debug( "DscSetup::setup_container - "
          . "IRON_CONTAINER_PARAMS = "
          . Dumper $cont_params);

    if (
        !defined( $cont_params->{IRON_CONTAINER_LARGE_AGGR} )
        || (   $cont_params->{IRON_CONTAINER_LARGE_AGGR}
            && $cont_params->{IRON_CONTAINER_LARGE_AGGR} !~ /32-bit/ )
      )
    {
        $self->{aggr_type}                        = "64-bit";
        $cont_params->{IRON_CONTAINER_LARGE_AGGR} = "64-bit";
    }
    else {
        $self->{aggr_type}                        = "32-bit";
        $cont_params->{IRON_CONTAINER_LARGE_AGGR} = "32-bit";
    }

    if ( defined $cont_params->{IRON_VVOL_NAME} ) {
        $Test->nlog("Using existing volume $cont_params->{IRON_VVOL_NAME}");

        my $vol_obj = NACL::C::Volume->find(
                command_interface => $self->{node},
                filter => {
                    volume  => $cont_params->{IRON_VVOL_NAME} });

        # Find aggr details
        $self->{aggregate} = NACL::STask::Aggregate->find(
                                 command_interface  => $self->{node},
                                 filter             => {
                                     aggregate => $vol_obj->aggregate()
                                 });
        $cont_params->{IRON_CONTAINER_NAME} = $self->{aggregate}{aggregate};
    }

    # aggr options
    $aggr_options{aggregate}    = $cont_params->{IRON_CONTAINER_NAME};
    $aggr_options{diskcount}    = $cont_params->{IRON_CONTAINER_SIZE};
    $aggr_options{raidtype}     = $cont_params->{IRON_CONTAINER_RAID_TYPE};
    $aggr_options{'block-type'} = $cont_params->{IRON_CONTAINER_LARGE_AGGR};

    if ( defined($main::SETUP_DONE) && $main::SETUP_DONE ) {
        $other_options{aggregate} = $cont_params->{IRON_CONTAINER_NAME};

        # Check if container exists
        $self->{aggregate} = NACL::STask::Aggregate->find_or_create(
                command_interface => $self->{node},
                _was_created => \$aggr_created_flag,
                %other_options,
                find_specific_options => {
                  'block-type' => $cont_params->{IRON_CONTAINER_LARGE_AGGR}
                },
                create_specific_options => {
                  diskcount    => $cont_params->{IRON_CONTAINER_SIZE},
                  raidtype     => $cont_params->{IRON_CONTAINER_RAID_TYPE},
                  'block-type' => $cont_params->{IRON_CONTAINER_LARGE_AGGR}
                },
                if_exists_with_different_attributes => 'purge'
        );
    } else {
        # Create container
        $self->{aggregate} = NACL::STask::Aggregate->create(
            command_interface => $self->{node},
            if_exists         => 'purge',
            _was_created      => \$aggr_created_flag,
            nacltask_wait     => 1,
            %aggr_options
        );
    }

    # Added aggr to the aggr-list of the vserver
    #$self->{aggregate}->add_to_aggr_list_of_vserver(
    #          vserver => $cont_params->{IRON_CONTAINER_VSERVER} );

    $self->{aggregate_type} =
      $self->{aggregate}->get_one_state_attribute("block-type");
    $self->{aggr_was_created} = $aggr_created_flag;
    $self->{noof_volsets}     = 0;    # Initialize

    $cont_params->{IRON_CONTAINER_CREATED} = $aggr_created_flag;
    $Log->debug( "DscSetup::setup_container"
               . " - aggregate = " . Dumper $self->{aggregate} );

    # Setup aggr options
    if ( defined($cont_params->{IRON_CONTAINER_OPTIONS}) ) {
        next if $cont_params->{IRON_CONTAINER_OPTIONS} =~ /undef/i;
        $Log->debug("setup_container: Setting aggr options...");
        #if ( $cont_params->{IRON_CONTAINER_OPTIONS} =~ /nosnap/i ) {
        #    $self->{aggregate}->modify( nosnap => 'off' );
        #}
    }

    $self->{aggr_snap_list} = ();
    if ( defined $cont_params->{IRON_CONTAINER_NUMOF_SNAPS}
             && $cont_params->{IRON_CONTAINER_NUMOF_SNAPS} ) {
        NACL::C::AggregateSnapshot->reserve(
                command_interface => $self->{node},
                aggregate         => $cont_params->{IRON_CONTAINER_NAME},
                percentage        => "5" );

        # Create snap
        my $snap_name = $cont_params->{IRON_CONTAINER_SNAP_NAME_PREFIX}
                        . $cont_params->{IRON_CONTAINER_NAME}
                        . "_1" ;

        try {
            NACL::C::AggregateSnapshot->create(
                      command_interface => $self->{node},
                      aggregate         => $cont_params->{IRON_CONTAINER_NAME},
                      snapshot          => $snap_name );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            # Only few cases are dependent on aggr snapshot
            # So, ignoring aggr snapshot create failure for now
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        }
        catch NACL::APISet::Exceptions::InvalidParamValueException with {
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        };

        my $aggrsnap_object = NACL::CS::AggregateSnapshot->fetch (
                        command_interface => $self->{node},
                        filter            => {
                          aggregate => $cont_params->{IRON_CONTAINER_NAME},
                          snapshot  => $snap_name },
                        allow_empty       => 1 );
        push(@{$self->{aggr_snap_list}}, $aggrsnap_object);
    }

    $Log->exit();
} ## end sub setup_container

# This will return the container object
sub get_aggr_object {

    my $self = shift;
    return $self->{aggregate};

} ## end sub get_aggr_object

# This sets up the volumes.
# It makes use of commonParams
sub setup_volumes {
    my $self            = shift;
    my $vol_created_flg = 0;
    my ( $volume_name, $sis );
    $Log->enter();

    my $if_exists;    # purge or reuse volumes
    if ( defined($main::SETUP_DONE) && $main::SETUP_DONE ) {
        $if_exists = "reuse";
    }
    else {
        $if_exists = "purge";
    }

    my $cont_params   = $self->{iron_cparams}{IRON_CONTAINER_PARAMS};
    my $vvol_params   = $self->{iron_cparams}{IRON_VVOL_PARAMS};
    my $vol_set_count = $vvol_params->{IRON_NUMOF_VVOL_SETS};
    my $ci_is_cmode   = $self->{node}->is_cmode();
    my $create_ulun   = 0;

    $self->{noof_volsets} = $vol_set_count;
    $self->{volset_list}  = ();

    if ( $vvol_params->{IRON_MULTI_TYPE_CORRUPTION} ) {
        # For injecting multiplt corruptions to diff volumes
        for ( my $i = 0; $i < $vol_set_count; $i++ ) {
            $vvol_params->{"VVOL_SET$i"}{IRON_VVOL_NUMOF_VVOLS} = 21;
            $vvol_params->{"VVOL_SET$i"}{IRON_VVOL_UNIXLUN}     = 0;
            $vvol_params->{"VVOL_SET$i"}{IRON_VVOL_SIZE}        = '10g';
            $vvol_params->{"VVOL_SET$i"}{IRON_VVOL_NUMOF_FILES} = 0;
            $vvol_params->{"VVOL_SET$i"}{IRON_VVOL_NUMOF_SNAPS} = 0;
        }
    }

    $self->set_flags( wafl_flags => {
        wafl_maxvvbn_disabled => $vvol_params->{IRON_MAXVVBN_DISABLE}
          }) if defined $vvol_params->{IRON_MAXVVBN_DISABLE};

    # Loop through all vvol sets
    for ( my $i = 0; $i < $vol_set_count; $i++ ) {
        my $vvol_ref = $vvol_params->{"VVOL_SET$i"};
        push( @{ $self->{volset_list} }, "VVOL_SET$i" );

        $self->{"VVOL_SET$i"}{noof_volumes} =
                       $vvol_ref->{IRON_VVOL_NUMOF_VVOLS};
        $Log->debug( "DscSetup::setup_volumes - vvol_ref "
                   . "for VVOL_SET$i = " . Dumper $vvol_ref );

        # Loop through each volume in volume set
        $self->{"VVOL_SET$i"}{vol_list} = ();
        for ( my $j = 0; $j < $vvol_ref->{IRON_VVOL_NUMOF_VVOLS}; $j++ ) {
            $volume_name = $vvol_ref->{IRON_VVOL_NAME_PREFIX} . "_VOL$i$j";
            $volume_name = $vvol_ref->{IRON_VVOL_NAME}
                               if defined $vvol_ref->{IRON_VVOL_NAME};
            $self->{"VVOL_SET$i"}{$volume_name}{qtree_list} = ();
            my $vol_obj;
            try {
                $vol_obj = NACL::C::Volume->find(
                        command_interface => $self->{node},
                        filter => {
                            vserver => $vvol_ref->{IRON_VVOL_VSERVER},
                            volume  => $volume_name,
                        });
            }
            catch NACL::Exceptions::NoElementsFound with {
                try {
                    # Could be with different vserver
                    $vol_obj = NACL::C::Volume->find(
                        command_interface => $self->{node},
                        filter => {
                            volume  => $volume_name,
                        });
                    $vvol_ref->{IRON_VVOL_VSERVER} = $vol_obj->vserver();
                }
                catch NACL::Exceptions::NoElementsFound with {
                    my $excpn_obj = shift;
                };
            };

            unless ( defined $vol_obj ) {
                # Create volume
                $Test->nlog("Creating volume $volume_name");
                $vol_obj = NACL::STask::Volume->create(
                    command_interface  => $self->{node},
                    volume             => $volume_name,
                    'junction-path'    => "/$volume_name",
                    vserver            => $vvol_ref->{IRON_VVOL_VSERVER},
                    aggregate          => $self->{aggregate}{aggregate},
                    size               => $vvol_ref->{IRON_VVOL_SIZE},
                    'space-guarantee'  => 'none',
                    'unix-permissions' => '---rwxrwxrwx',
                    'security-style'   => 'mixed',
                    'policy'           => 'default',
                    'type'             => 'RW',
                    _was_created       => \$vol_created_flg,
                    nacltask_if_exists => $if_exists,
                    nacltask_wait      => 1,
                    nacltask_destroy_corrupt_volume => 1,
                );
            }

            unless ($ci_is_cmode) {
                $Log->comment("export volume $volume_name");
                $self->{node}{apiset}->execute_raw_command(
                     command => "exportfs -p rw,anon=0,nosuid /vol/$volume_name");
                $self->{node}{apiset}->execute_raw_command(command => "exportfs");
            }

            push( @{$self->{"VVOL_SET$i"}{vol_list}}, $volume_name );
            $self->{"VVOL_SET$i"}{$volume_name}{is_created}  = $vol_created_flg;
            $self->{"VVOL_SET$i"}{$volume_name}{vol_obj_ref} = $vol_obj;
            $self->{"VVOL_SET$i"}{$volume_name}{vol_type}    = 'flex';
            $self->{"VVOL_SET$i"}{$volume_name}{vserver}     = $vol_obj->vserver();
            $self->{"VVOL_SET$i"}{$volume_name}{vol_aggr}
                                                    = $self->{aggregate}{aggregate};
            $self->{"VVOL_SET$i"}{$volume_name}{vol_aggr_obj}  = $self->{aggregate};
            $self->{"VVOL_SET$i"}{$volume_name}{vol_param_ref} = $vvol_ref;

            my %opts; # for Quota
            if ( $ci_is_cmode ) {
                $opts{nacltask_if_exists} = 'reuse';
            } else {
                $opts{nacltask_if_exists} = 'purge';
                $opts{qtree} = '';
            }

            # Create Qtree
            if ( defined $vvol_ref->{IRON_VVOL_QTREE}
                       && $vvol_ref->{IRON_VVOL_QTREE} ) {
                my (@qtree_list, $qtree_obj, $qtree);
                for ( my $i = 0; $i < $vvol_ref->{IRON_VVOL_QTREE}; $i++ ) {
                    $qtree = "vqtre$i";
                    $Test->nlog("Creating Qtree $qtree on vol $volume_name");

                    try {
                        $qtree_obj = NACL::STask::Qtree->create(
                                  command_interface  => $self->{node},
                                  volume             => $volume_name,
                                  vserver            => $vvol_ref->{IRON_VVOL_VSERVER},
                                  qtree              => $qtree,
                                  'unix-permissions' => '777',
                                  'security-style'   => 'unix',
                                  if_exists          => 'reuse' );
                    }
                    catch NACL::C::Exceptions::VolumeQtree::AlreadyExists with {
                        $qtree_obj = NACL::STask::Qtree->find(
                                  command_interface => $self->{node},
                                  filter => {
                                      volume  => $volume_name,
                                      vserver => $vvol_ref->{IRON_VVOL_VSERVER},
                                      qtree   => $qtree });
                    };
                    $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{qtree_name} = $qtree;
                    $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{qtree_obj}  = $qtree_obj;
                    push ( @qtree_list, $qtree );

                    # Quota 'ON' on request
                    if ( defined $vvol_ref->{IRON_VVOL_QUOTA}
                               && $vvol_ref->{IRON_VVOL_QUOTA} && $ci_is_cmode ) {
                        # TODO: 7-Mode give problem, need to fix
                        $Test->nlog("Enable Quotas on the Qtree $qtree...");
                        my $quota = NACL::MTask::VolumeQuota->setup(
                                  command_interface  => $self->{node},
                                  volume             => $volume_name,
                                  vserver            => $vvol_ref->{IRON_VVOL_VSERVER},
                                  target             => $qtree,
                                  type               => 'tree',
                                  'policy-name'      => 'default',
                                  nacltask_wait      => 1,
                                  %opts );

                        $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{quota_obj} = $quota;
                    }
                }
                $self->{"VVOL_SET$i"}{$volume_name}{qtree_list} = \@qtree_list;

            } elsif ( defined $vvol_ref->{IRON_VVOL_QUOTA}
                       && $vvol_ref->{IRON_VVOL_QUOTA} && $ci_is_cmode ) {
                # TODO: 7-Mode give problem, need to fix
                # Quota 'ON' on request
                $Test->nlog("Turn on Quotas on the volume $volume_name");
                my $quota = NACL::MTask::VolumeQuota->setup(
                                  command_interface    => $self->{node},
                                  volume               => $volume_name,
                                  vserver              => $vvol_ref->{IRON_VVOL_VSERVER},
                                  'policy-name'        => 'default',
                                  'type'               => 'user',
                                  'target'             => 'root',
                                  nacltask_create_user_or_group => 1,
                                  %opts );

                $self->{"VVOL_SET$i"}{$volume_name}{quota_obj} = $quota;
            }

            # Enable dense & compression
            if ( defined $vvol_ref->{IRON_VVOL_A_SIS}
                       && $vvol_ref->{IRON_VVOL_A_SIS} ) {
                $Test->nlog("Enabling storage efficiency on vol $volume_name");

                my $sis = $self->_storage_efficiency_config(
                    vvolref   => $vvol_ref,
                    volumeobj => $vol_obj
                );
                $self->{sisobjs}{$volume_name} = $sis;
            }

            # Create UNIX LUN if required
            if ( defined $vvol_ref->{IRON_VVOL_UNIXLUN}
                      && $vvol_ref->{IRON_VVOL_UNIXLUN} ) {
                for (my $u = 0; $u < $vvol_ref->{IRON_VVOL_UNIXLUN}; $u++) {
                    my $lun_name = $volume_name . "_ulun$u";
                    if ( defined $self->{"VVOL_SET$i"}{$volume_name}{qtree_list} ) {
                        my $qtree_list = $self->{"VVOL_SET$i"}{$volume_name}{qtree_list};
                        foreach my $qtree ( @$qtree_list ) {
                            $self->{lun_input}{$volume_name}{$qtree}{lun_name} = $lun_name;
                            $self->{lun_input}{$volume_name}{$qtree}{vol_obj}  = $vol_obj;
                            $self->{lun_input}{$volume_name}{$qtree}{lun_size}
                                      = $vvol_ref->{IRON_VVOL_UNIXLUN_SIZE};
                        }
                    } else {
                        $self->{lun_input}{$volume_name}{none}{lun_name} = $lun_name;
                        $self->{lun_input}{$volume_name}{none}{vol_obj}  = $vol_obj;
                        $self->{lun_input}{$volume_name}{none}{lun_size}
                                      = $vvol_ref->{IRON_VVOL_UNIXLUN_SIZE};
                    }
                }
                $create_ulun++;
            }
        } #end of for
    } #end of for

    # Setup VSM
    $self->setup_vol_snapmirror();

    # Setup vol clones
    $self->setup_vol_clones();

    for ( my $i = 0; $i < $vol_set_count; $i++ ) {
        $Log->debug(
            "DscSetup::setup_volumes - " . Dumper $self->{"VVOL_SET$i"} );
    }

    if ( $create_ulun ) {
        $Log->debug("lun_input for lun create - "
                           . Dumper $self->{lun_input});
        # Work with unix client
        my $client_ci  = $self->{client_apiset};
        my $client_obj = $self->{client};

        $self->{lunsetup_obj}  = ();
        $self->{ulun_obj_list} = ();

        # Get initiator from client
        my $uname = $client_ci->execute_command(command => "hostname -s");
        my $iname = $client_ci->execute_command(command => "/sbin/iscsi-iname");
        chomp $iname; chomp $uname;

        # Disable client multipath
        $client_ci->execute_raw_command(
                      command => "$SUDO_PATH /etc/init.d/multipathd stop" );

        my $cmd       = "echo InitiatorName=$iname.$uname";
        my $file_name = '/etc/iscsi/initiatorname.iscsi';
        my $initiator_name;
        if ( $client_ci->execute_command(
                             command => "/bin/cat $file_name" )
                                        =~ /InitiatorName=(.*)/ ) {
            $initiator_name = $1;
        } else {
            $client_ci->touch( path => $file_name );
            $client_ci->chmod( paths             => $file_name,
                               'privilege-level' => 'root',
                               mode              => 'a+rwx');

            $client_ci->execute_raw_command(
                           command => "$cmd > " . $file_name );

            if ( $client_ci->execute_command(
                             command => "/bin/cat $file_name" )
                                        =~ /InitiatorName=(.*)/ ) {
                $initiator_name = $1;
            } else {
                NATE::BaseException->throw("Could not find initiator name");
            }
        }

        # Setup lun(s)
        my $lunsetup_failed = 0;
        try {
            $self->_create_luns( initiator_name => $initiator_name );
        }
        catch NATE::BaseException with  {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
            $lunsetup_failed++;
        };

        unless ( $lunsetup_failed ) {
            my $lsobject  = ${$self->{lunsetup_obj}}[0];
            $self->_lun_client_config( lunsetup_obj => $lsobject );
        }
    } # end of if

    # Fill volumes & vol clones
    $Test->nlog("Fill test volumes with test data");
    $self->fill_volumes();

    # Create aggr snapshot after vol setup
    if ( defined $cont_params->{IRON_CONTAINER_NUMOF_SNAPS}
             && $cont_params->{IRON_CONTAINER_NUMOF_SNAPS} ) {
        # Create snap
        my $snap_name = $cont_params->{IRON_CONTAINER_SNAP_NAME_PREFIX}
                        . $cont_params->{IRON_CONTAINER_NAME} . "_2" ;

        try {
            NACL::C::AggregateSnapshot->create(
                command_interface => $self->{node},
                aggregate         => $cont_params->{IRON_CONTAINER_NAME},
                snapshot          => $snap_name );
        }
        catch NACL::APISet::Exceptions::InvalidParamValueException with {
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        };

        my $aggrsnap_object = NACL::CS::AggregateSnapshot->fetch (
                        command_interface => $self->{node},
                        filter            => {
                          aggregate => $cont_params->{IRON_CONTAINER_NAME},
                          snapshot  => $snap_name,
                        },
                        allow_empty       => 1 );
        push(@{$self->{aggr_snap_list}}, $aggrsnap_object);
    }

    # Get inode block count
    my $inode_count = $self->get_inode_block_count(
                        inum      => '73',
                        cont_type => 'aggregate',
                        cont_name => $cont_params->{IRON_CONTAINER_NAME} );

    # Perform wafl touch on activemap and then
    # this snapshot will not be sharing any block with active
    $Log->comment( "Performing wafl touch on activemap file" );
    try {
        $self->{node}{apiset}->wafl_touch(
                  'privilege-level' => 'test',
                  'aggregate'       => $cont_params->{IRON_CONTAINER_NAME},
                  'inode-number'    => '73',
                  'level-value'     => '0',
                  'private'         => '1',
                  'start-fbn'       => '0',
                  'end-fbn'         => $inode_count );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $excpn_obj = shift;
        $Log->warn( "Caught exception: " . $excpn_obj->text() );
    }
    catch NACL::APISet::Exceptions::TimeoutException with {
        my $excpn_obj = shift;
        $Log->debug( "Caught exception: " . $excpn_obj->text() );
    };

    # Help function for multi-type corruptions tests
    $self->_multi_type_corruption()
          if $vvol_params->{IRON_MULTI_TYPE_CORRUPTION};

    # Quota report for all files
    for ( my $i = 0; $i < $vol_set_count; $i++ ) {
        my @vol_list = $self->get_vvols(vvol_set => "VVOL_SET$i");
        foreach my $volume_name (@vol_list) {
            my @quota_list = ();
            push (@quota_list, $self->{"VVOL_SET$i"}{$volume_name}{quota_obj})
                if defined $self->{"VVOL_SET$i"}{$volume_name}{quota_obj};

            if ( defined $self->{"VVOL_SET$i"}{$volume_name}{qtree_list} ) {
                my $qtree_list = $self->{"VVOL_SET$i"}{$volume_name}{qtree_list};
                foreach my $qtree ( @$qtree_list ) {
                    push (@quota_list, $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{quota_obj})
                        if defined $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{quota_obj};
                }
            }

            $Log->comment("Running a Quota report...");
            foreach my $quota ( @quota_list ) {
                try {
                    my @reports = $quota->report();
                    $Log->debug("Quota report - " . Dumper \@reports);
                }
                catch NACL::Exceptions::NoElementsFound with {
                };
            }
        }
    }

    $Log->exit();
} ## end sub setup_volumes

# This creates clones of the volumes
# It makes use of commonParams
sub setup_vol_clones {
    my $self            = shift;
    my $vvol_params     = $self->{iron_cparams}{IRON_VVOL_PARAMS};
    my $ci_is_cmode     = $self->{node}->is_cmode();
    my $vol_created_flg = 0;
    $Log->enter();

    my $if_exists;    # purge or reuse volumes
    if ( defined($main::SETUP_DONE) && $main::SETUP_DONE ) {
        $if_exists = "reuse";
    }
    else {
        $if_exists = "purge";
    }

    # Loop through all vvol sets & Setup vol clones
    for ( my $i = 0, my $j = 0; $i < $self->{noof_volsets}; $i++ ) {
        my $vvol_ref = $vvol_params->{"VVOL_SET$i"};
        $self->{"VVOL_SET$i"}{cvol_list} = ();

        # Loop through each volume in volume set
        foreach my $parent_vol_name ( @{ $self->{"VVOL_SET$i"}{vol_list} } ) {
            for ( my $k = 0; $k < $vvol_ref->{IRON_VVOL_CLONES}; $k++, $j++ ) {
                my ( $volume_name, $vol_clone_obj );

                # clone for only flex vols
                next unless $vvol_ref->{IRON_VVOL_TYPE} =~ /flex/i;
                $volume_name = "SET$i" . "_CVOL$i$j$k";

                $Test->nlog("Creating volume clone for vol $parent_vol_name");
                $vol_clone_obj = NACL::STask::VolumeClone->create(
                    command_interface  => $self->{node},
                    flexclone          => $volume_name,
                    'junction-path'    => "/$volume_name",
                    vserver            => $vvol_ref->{IRON_VVOL_VSERVER},
                    'parent-volume'    => $parent_vol_name,
                    'space-guarantee'  => 'none',
                    'type'             => 'RW',
                    _was_created       => \$vol_created_flg,
                    nacltask_if_exists => $if_exists,
                    nacltask_wait      => 1,
                    #nacltask_destroy_corrupt_volume => 1,
                );

                push( @{$self->{"VVOL_SET$i"}{cvol_list}}, $volume_name );
                $self->{"VVOL_SET$i"}{$volume_name}{is_created}    = $vol_created_flg;
                $self->{"VVOL_SET$i"}{$volume_name}{vol_obj_ref}   = $vol_clone_obj;
                $self->{"VVOL_SET$i"}{$volume_name}{clone_parent}  = $parent_vol_name;
                $self->{"VVOL_SET$i"}{$volume_name}{vol_param_ref} = $vvol_ref;
                $self->{"VVOL_SET$i"}{$volume_name}{vol_type} = 'clone';
                $self->{"VVOL_SET$i"}{$volume_name}{vol_aggr} = $self->{aggregate}{aggregate};
                $self->{"VVOL_SET$i"}{$volume_name}{vol_aggr_obj} = $self->{aggregate};
                $self->{"VVOL_SET$i"}{$volume_name}{qtree_list} = ();

                if ( defined $vvol_ref->{IRON_VVOL_QTREE}
                           && $vvol_ref->{IRON_VVOL_QTREE} ) {
                    my (@qtree_list, $qtree_obj, $qtree);
                    for ( my $i = 0; $i < $vvol_ref->{IRON_VVOL_QTREE}; $i++ ) {
                        $qtree = "vqtre$i";

                        # find Qtree
                        try {
                            my $qtree_obj = NACL::STask::Qtree->find(
                                  command_interface => $self->{node},
                                  filter => {
                                      volume  => $volume_name,
                                      vserver => $vvol_ref->{IRON_VVOL_VSERVER},
                                      qtree   => $qtree });
                            $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{qtree_name} = $qtree;
                            $self->{"VVOL_SET$i"}{$volume_name}{$qtree}{qtree_obj}  = $qtree_obj;
                            push ( @qtree_list, $qtree );
                        }
                        catch NATE::BaseException with {
                            my $excpn_obj = shift;
                        };
                    }
                    $self->{"VVOL_SET$i"}{$volume_name}{qtree_list} = \@qtree_list;
                }
            }
        }
    }

    $Log->exit();
} ## end sub setup_vol_clones

# This creates snapmirror for each volume
# It makes use of commonParams
sub setup_vol_snapmirror {
    my $self            = shift;
    my $vvol_params     = $self->{iron_cparams}{IRON_VVOL_PARAMS};
    my $vol_created_flg = 0;
    my $is_cmode = $self->{node}->is_cmode();
    $Log->enter();

    my $if_exists;    # purge or reuse volumes
    if ( defined($main::SETUP_DONE) && $main::SETUP_DONE ) {
        $if_exists = "reuse";
    }
    else {
        $if_exists = "purge";
    }

    # Loop through all vvol sets & Setup vol clones
    for ( my $i = 0, my $j = 0; $i < $self->{noof_volsets}; $i++ ) {
        my $vvol_ref = $vvol_params->{"VVOL_SET$i"};
        $self->{"VVOL_SET$i"}{svol_list} = ();

        # Loop through each volume in volume set
        foreach my $parent_vol_name ( @{ $self->{"VVOL_SET$i"}{vol_list} } ) {
            for ( my $k = 0; $k < $vvol_ref->{IRON_VVOL_VSM}; $k++, $j++ )
            {
                my ( $vol_obj, $volume_name, $vol_snapmirror_obj );
                # snapmirror for only flex vols
                next unless $vvol_ref->{IRON_VVOL_TYPE} =~ /flex/i;
                $volume_name = "SET$i" . "_SVOL$i$j$k";

                $Test->nlog("Creating volume snapmirror for vol $parent_vol_name");
                $vol_obj = NACL::STask::Volume->create(
                    command_interface  => $self->{node},
                    volume             => $volume_name,
                    vserver            => $vvol_ref->{IRON_VVOL_VSERVER},
                    aggregate          => $self->{aggregate}{aggregate},
                    size               => $vvol_ref->{IRON_VVOL_SIZE},
                    'type'             => 'DP',
                    _was_created       => \$vol_created_flg,
                    nacltask_if_exists => $if_exists,
                    nacltask_wait      => 1,
                    nacltask_destroy_corrupt_volume => 1 );

                push( @{ $self->{"VVOL_SET$i"}{svol_list} }, $volume_name );
                $self->{"VVOL_SET$i"}{$volume_name}{is_created}  = $vol_created_flg;
                $self->{"VVOL_SET$i"}{$volume_name}{vol_obj_ref} = $vol_obj;
                $self->{"VVOL_SET$i"}{$volume_name}{source_vol}  = $parent_vol_name;
                $self->{"VVOL_SET$i"}{$volume_name}{vol_type} = 'snapmirror';
                $self->{"VVOL_SET$i"}{$volume_name}{vol_aggr} = $self->{aggregate}{aggregate};
                $self->{"VVOL_SET$i"}{$volume_name}{vol_param_ref} = $vvol_ref;

                # Create VSM relation
                my ($src_vol, $dst_vol);
                if ( $is_cmode ) {
                    $src_vol = "$vvol_ref->{IRON_VVOL_VSERVER}:$parent_vol_name";
                    $dst_vol = "$vvol_ref->{IRON_VVOL_VSERVER}:$volume_name";
                } else {
                    $src_vol = $parent_vol_name;
                    $dst_vol = $volume_name;
                }
                my $snapmirror_object = NACL::STask::Snapmirror->create(
                    command_interface  =>  $self->{node},
                    'source-path'      =>  $src_vol,
                    'destination-path' =>  $dst_vol,
                    if_exists          =>  $if_exists,
                    nacltask_wait      =>  '0',
                    'method-timeout'   =>  '300',
                );
                $snapmirror_object->wait_for_transfer_to_complete();
                $self->{"VVOL_SET$i"}{$volume_name}{snapmirror_object} = $snapmirror_object;
            }
        }
    }

    $Log->exit();
} ## end sub setup_vol_snapmirror

# This populates the dataset
# It makes use of commonParams
sub fill_volumes {
    my $self        = shift;
    my $vvol_params = $self->{iron_cparams}{IRON_VVOL_PARAMS};
    $Log->enter();

    # Loop through all vvol sets & Setup vol clones
    for ( my $i = 0, my $j = 0; $i < $self->{noof_volsets}; $i++ ) {
        my $vvol_ref = $vvol_params->{"VVOL_SET$i"};

        # Loop through each volume in volume set
        my (
            $total_vol_size, $used_vol_size,     $required_percentage,
            $required_size,  $size_to_fill,      $snapshot_count,
            $file_count,     $file_size_to_fill, $vol_obj
        );

        my @vol_list = $self->get_vvols( vvol_set => "VVOL_SET$i" );
        foreach my $vol_name (@vol_list) {

            my $fake_create = 0;
            $vol_obj = $self->{"VVOL_SET$i"}{$vol_name}{vol_obj_ref};

            # Vol info for debug
            NACL::CS::Volume->fetch(
                command_interface => $self->{node},
                filter            => {
                     vserver => $vol_obj->vserver(),
                     volume  => $vol_obj->volume()
                }
            );

            # Qtree info
            my @qtree_list = ();
            @qtree_list = @{$self->{"VVOL_SET$i"}{$vol_name}{qtree_list}}
                   if defined $self->{"VVOL_SET$i"}{$vol_name}{qtree_list};
            my $qtree_count = $#qtree_list + 1 ? $#qtree_list + 1 : 1;

            $snapshot_count = $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ?
                                    $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} : 1;
            $file_count     = $vvol_ref->{IRON_VVOL_NUMOF_FILES};
            $file_count     = int ($file_count/$qtree_count);

            # Cleanup all test files on request
            unless ( $vvol_ref->{IRON_VVOL_TESTFILE_REUSE} ) {
                # Vol snapshot
                my @snap_objects = NACL::CS::VolumeSnapshot->fetch(
                                command_interface => $self->{node},
                                filter            => {
                                     vserver => $vol_obj->vserver(),
                                     volume  => $vol_obj->volume()
                                },
                                allow_empty       => 1
                );

                # Remove snapshot
                foreach my $snap_obj ( @snap_objects ) {
                    next if $snap_obj->snapshot() =~ /clone_/;
                    NACL::STask::VolumeSnapshot->purge(
                                    command_interface  => $self->{node},
                                    volume             => $vol_obj->volume(),
                                    vserver            => $vol_obj->vserver(),
                                    snapshot           => $snap_obj->snapshot(),
                                    nacltask_if_purged => 'pass',
                                    nacltask_wait      => 1 );
                }

                my $volume_name = $vol_obj->volume();
                for ( my $s = 0; $s < $snapshot_count; $s++ ) {
                    for ( my $q = 0; $q < $qtree_count; $q++ ) {
                        my $qtree = $qtree_list[$q] if defined $qtree_list[$q];
                        for ( my $k = 0; $k < $file_count; $k++ ) {
                            for ( my $j = 0; $j < 2; $j++ ) {
                                my $full_file_name;
                                if (defined $qtree) {
                                    $full_file_name = "/vol/$volume_name/$qtree/file_s$j$s$k";
                                } else {
                                    $full_file_name = "/vol/$volume_name/file_s$j$s$k";
                                }

                                # Remove file
                                try {
                                    $self->{node}{apiset}
                                        ->rm('paths' => $full_file_name);
                                }
                                catch NATE::BaseException with {
                                    my $excpn_obj = shift;
                                };
                            }
                        }
                    }
                }
            }

            # Sync & Flush Cache
            $self->{node}{apiset}->sync();
            $self->{node}{apiset}->sync();
            $self->{node}{apiset}->wafl_flush_cache();

            # Find volume size
            $vol_obj = $vol_obj->cast_to_volume()
              if defined $self->{"VVOL_SET$i"}{$vol_name}{clone_parent};
            $total_vol_size = $vol_obj->get_one_state_attribute("total");
            $used_vol_size  = $vol_obj->get_one_state_attribute("used");
            $Log->debug(
                "total size = $total_vol_size & used_vol_size = $used_vol_size"
            );

            # Try to calculate how much percentage of data to fill.
            $required_percentage = $vvol_ref->{IRON_VVOL_PCT_USAGE};

            $required_size = ( $total_vol_size * $required_percentage ) / 100;
            $size_to_fill  = int ($required_size - $used_vol_size);
            $Log->debug("required_percentage = $required_percentage "
                                    . "& size_to_fill = $size_to_fill");
            if ( $size_to_fill <= 0 ) {
                $Log->warn( "Looks like the volume is already filled "
                          . "more than $required_percentage percentage" );
                $fake_create++;
            } elsif ( $vvol_ref->{IRON_VVOL_TESTFILE_REUSE} ) {
                $fake_create++;
            }

            $Log->comment("Create test files and snapshots on $vol_name");
            $self->_create_files_and_snapshots(
                volume_name    => $vol_name,
                vvol_ref       => $vvol_ref,
                size_to_fill   => $size_to_fill,
                file_count     => $vvol_ref->{IRON_VVOL_NUMOF_FILES},
                snapshot_count => $vvol_ref->{IRON_VVOL_NUMOF_SNAPS},
                fake_create    => $fake_create,
                vvol_set       => "VVOL_SET$i",
                qtree_list     => \@qtree_list
            ) if $vvol_ref->{IRON_VVOL_NUMOF_FILES};

            if ( defined $vvol_ref->{IRON_NUMOF_HOLEYFILES}
                && $vvol_ref->{IRON_NUMOF_HOLEYFILES} )
            {
                $Test->nlog("Create holey files on $vol_name");
                $file_count = $vvol_ref->{IRON_NUMOF_HOLEYFILES};
                $self->{"VVOL_SET$i"}{$vol_name}{holeyfile_list} = ();
                my $holey_file_size = 10;
                for ( my $h = 0; $h < $file_count; $h++ ) {
                    $self->_create_holey_file(
                        volume_name => $vol_name,
                        vserver     => $vol_obj->vserver(),
                        filename    => "holeyfile$h",
                        fake_create => $fake_create,
                        size        => $holey_file_size
                    ) unless $fake_create;

                    $holey_file_size += 10;
                    my @ret =
                      $self->{node}{apiset}->ls( paths => "/vol/$vol_name" )
                           ->get_processed_output();

                    push(
                        @{ $self->{"VVOL_SET$i"}{$vol_name}{holeyfile_list} },
                        "holeyfile$h"
                    ) if ( grep( "holeyfile$h", @ret ) );
                }
            }

            if ( defined $vvol_ref->{IRON_NUMOF_SPACERESFILES}
                       && $vvol_ref->{IRON_NUMOF_SPACERESFILES} )
            {
                $Test->nlog("Create space_reserved files on $vol_name");
                $file_count = $vvol_ref->{IRON_NUMOF_SPACERESFILES};
                $self->{"VVOL_SET$i"}{$vol_name}{spaceresfile_list} = ();
                my $space_reserved_file_size = 10240;
                for ( my $sp = 0; $sp < $file_count; $sp++ ) {
                    $self->_create_space_reserved_file(
                        volume_name => $vol_name,
                        vserver     => $vol_obj->vserver(),
                        filename    => "space_reserved$sp",
                        size        => $space_reserved_file_size
                    ) unless $fake_create;
                    $space_reserved_file_size += 10;
                    my @ret =
                      $self->{node}{apiset}->ls( paths => "/vol/$vol_name" )
                      ->get_processed_output();

                    push(
                      @{$self->{"VVOL_SET$i"}{$vol_name}{spaceresfile_list}},
                        "space_reserved$sp"
                    ) if ( grep( "space_reserved$sp", @ret ) );
                }
            }

            if ( defined $vvol_ref->{IRON_NUMOF_SPACERESFILES}
                && $vvol_ref->{IRON_NUMOF_SPACERESFILES} )
            {
                $Test->nlog("Create cifs enabled files on $vol_name");
                $file_count = $vvol_ref->{IRON_NUMOF_CIFSFILES};
                $self->{"VVOL_SET$i"}{$vol_name}{cifsfile_list} = ();

                # Qtree info
                my @qtree_list = ();
                @qtree_list = @{$self->{"VVOL_SET$i"}{$vol_name}{qtree_list}}
                   if defined $self->{"VVOL_SET$i"}{$vol_name}{qtree_list};

                for ( my $ce = 0; $ce < $file_count; $ce++ ) {
                    $self->_create_files_with_cifs_enabled(
                        volume_name => $vol_name,
                        vserver     => $vol_obj->vserver(),
                        filename    => "cifs_enabled$ce",
                        fake_create => $fake_create,
                        size        => '1g',
                        qtree_list  => \@qtree_list
                    ) unless $fake_create;

                    my @ret =
                      $self->{node}{apiset}->ls( paths => "/vol/$vol_name" )
                      ->get_processed_output();

                    push(
                        @{ $self->{"VVOL_SET$i"}{$vol_name}{cifsfile_list} },
                        "cifs_enabled$ce"
                    ) if ( grep( "cifs_enabled$ce", @ret ) );
                }
            }
        } ## end foreach
    } ## end for

    $Log->exit();
} ## end sub fill_volumes

# This starts sis waits for Idle state
# Accepts volume_name and returns the sis object
sub _storage_efficiency_config {
    my $self     = shift;
    my %options  = @_;
    my $vvolref  = $options{vvolref};
    my $vol_obj  = $options{volumeobj};
    my $vol_name = $vol_obj->volume();
    my $vserver  = $vol_obj->vserver();
    my $policy   = 'Policy_DscVolEfficiency';
    my $is_cmode = $self->{node}->is_cmode();

    $Log->comment('Find/Create volume efficiency policy');
    my $voleff_policy;
    my $voleff_policy_create = 0;
    try {
        $voleff_policy
             = NACL::C::VolumeEfficiencyPolicy->find(
                   command_interface => $self->{node},
                   filter => {
                       vserver => $vserver,
                       policy  => $policy,
                   });
    }
    catch NACL::Exceptions::NoElementsFound with {
        $voleff_policy_create = 1;
    }
    catch NACL::Exceptions::InvalidChoice with {
        $voleff_policy_create = 1;
    };

    if ( $voleff_policy_create && $is_cmode ) {
        # Create new policy
        $voleff_policy
            = NACL::C::VolumeEfficiencyPolicy->create(
                   command_interface => $self->{node},
                   vserver           => $vserver,
                   policy            => $policy,
                   schedule          => 'hourly',
                   enabled           => 'true',
                   'method-timeout'  => 60 );
    }

    # Enable Storage Efficiency
    my $sis = NACL::STask::VolumeEfficiency->on(
        command_interface => $self->{node},
        vserver           => $vserver,
        volume            => $vol_name,
        nacltask_if_enabled => 'pass' );

    # Enable compression
    $Log->comment("Link with policy only for $vol_name");
    if ( $is_cmode ) {
        $sis->config( policy => $policy );
    } else {
        my $schedule = 'sun-sat@0-23@~0-59/5';
        $Log->comment("Setup sis schedule for run at $schedule");
        $self->{node}{apiset}->execute_raw_command( 'command' =>
             "priv set test; sis config -s $schedule /vol/$vol_name" );
    }

    if ( defined $vvolref->{IRON_VVOL_COMPRESSION}
        && $vvolref->{IRON_VVOL_COMPRESSION}
        && $self->{aggr_type} =~ /64\-bit/ )
    {
        $Log->comment("Enable inline and background compression");
        $sis->config(
            'background-compression' => "true",
            'inline-compression'     => "true",
            'compression'            => "true",
        );
    }

    # Start sis
    try {
        $sis->start( nacltask_wait => '1', 'scan-old-data' => 'true' );
        $sis->wait_for_status( status => 'Idle' );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $excpn_obj = shift;
        $Log->debug( "Caught exception: " . $excpn_obj->text() );
    }

    return $sis;
} ## end sub _storage_efficiency_config

# This creates files and snapshots.
# It again indirectly uses commonParams from the caller.
sub _create_files_and_snapshots {
    my $self    = shift;
    my %options = @_;

    my $size_to_fill   = $options{size_to_fill};
    my $volume_name    = $options{volume_name};
    my $vvol_ref       = $options{vvol_ref};
    my $vvol_set       = $options{vvol_set};
    my $file_count     = defined($options{file_count}) ? $options{file_count} : 10;
    my $snapshot_count = defined($options{snapshot_count}) ? $options{snapshot_count} : 0;
    my $file_size      = defined($options{file_size}) ? $options{file_size} : 0;
    my $fake_create    = defined($options{fake_create}) ? $options{fake_create} : 0;
    my $postrun        = defined($options{postrun}) ? $options{postrun} : 0;
    my $create_method  = defined($vvol_ref->{IRON_FILE_CREATE_METHOD})
                            ? $vvol_ref->{IRON_FILE_CREATE_METHOD} : 'mkfile';
    my $vserver        = $vvol_ref->{IRON_VVOL_VSERVER};
    my $timeout        = $vvol_ref->{IRON_FILE_CREATE_TIMEOUT};
    my @qtree_list     = @{$options{qtree_list}};
    my $qtree_count    = $#qtree_list + 1 ? $#qtree_list + 1 : 1;

    my ( $file_name, $full_file_name, $snapshot, $file_size_to_fill );
    my @plist = ();

    $size_to_fill = int NACL::C::UnitNormalization::convert_units(
                NACL::C::UnitNormalization::strip_unit_spec($size_to_fill),
                guess_unit($size_to_fill), UNIT_B ) if defined $size_to_fill;

    $file_size = int NACL::C::UnitNormalization::convert_units(
                NACL::C::UnitNormalization::strip_unit_spec($file_size),
                guess_unit($file_size), UNIT_B ) if defined $file_size;

    if ( $file_count && $snapshot_count ) {
        $file_size_to_fill =
          int($size_to_fill / ($snapshot_count * $file_count));
    } elsif ( $file_count && $file_size && $postrun ) {
        $file_size_to_fill = $file_size;
    } elsif ( $snapshot_count ) {
        $file_size_to_fill =
          int($size_to_fill / ($snapshot_count * $file_count));
    } else {
        if ( $file_count ) {
            $file_size_to_fill = int($size_to_fill / $file_count);
        } else {
            $file_size_to_fill = int $size_to_fill;
        }
    }
    $Log->debug("file_size_to_fill = $file_size_to_fill");
    if ( defined $vvol_ref->{IRON_TEST_FILE_SIZE} && $vvol_ref->{IRON_TEST_FILE_SIZE} ) {
        $Log->comment("Each file will be of $vvol_ref->{IRON_TEST_FILE_SIZE}...");
        $file_size_to_fill = $vvol_ref->{IRON_TEST_FILE_SIZE};
    } elsif ( $file_size_to_fill < 0 ) {
        $Log->comment("Looks vol is already full (size_to_fill = $size_to_fill)."
                              . " Please cleanup vol and re-run");
        $fake_create++;
        $file_size_to_fill = '10M';
    }

    # Create snapshot?
    my $create_snapshot = $snapshot_count;
    $snapshot_count++ unless $snapshot_count;
    $file_count = int ($file_count/$qtree_count);

    # Loop through each snapshot
    $self->{$vvol_set}{$volume_name}{file_list} = () unless $postrun;
    $self->{$vvol_set}{$volume_name}{snap_list} = () unless $postrun;
    for ( my $s = 0; $s < $snapshot_count; $s++ ) {
        # Loop through each Qtree
        for ( my $q = 0; $q < $qtree_count; $q++ ) {
            my $qtree = $qtree_list[$q] if defined $qtree_list[$q];
            $file_count = $file_count ? $file_count : 1;

            if ( $create_method eq "filersio" ) {
                # Loop through each file
                for ( my $k = 0; $k < $file_count; $k++ ) {
                    $file_name = "file_s$postrun$s$k";
                    if (defined $qtree) {
                        $full_file_name = "/vol/$volume_name/$qtree/$file_name";
                    } else {
                        $full_file_name = "/vol/$volume_name/$file_name";
                    }

                    my $create_file = 1;
                    if ( $fake_create ) {
                        try {
                            $create_file = 0;
                            my $size = NACL::C::Slist->slist(
                                  command_interface => $self->{node},
                                  path              => $full_file_name,
                              )->get_parsed_output()->[0]->{file_size};
                            push(@{$self->{$vvol_set}{$volume_name}{file_list}}, $file_name);
                            $self->{$vvol_set}{$volume_name}{$file_name}{'name'} = $file_name;
                            $self->{$vvol_set}{$volume_name}{$file_name}{'size'} = $size;
                            $self->{$vvol_set}{$volume_name}{$file_name}{'full_name'} = $full_file_name;
                        }
                        catch NACL::APISet::Exceptions::InvalidParamValueException with {
                            my $excpn_obj = shift;
                            $create_file = 1;
                        };
                    }

                    if ( $create_file ) {
                        $self->{node}{apiset}->filersio_asyncio_active(
                               'read-percentage' => 0,
                               'rand-percentage' => 100,
                               'blk-size'        => '4096',
                               'start-byte'      => 0,
                               'end-byte'        => $file_size_to_fill,
                               'run-time'        => 10,
                               'num-active'      => 10,
                               filename          => $full_file_name,
                               create            => 1,
                               synchronous       => 1 ) if $create_file;

                        # Sync & Flush Cache
                        $self->{node}{apiset}->sync();
                        $self->{node}{apiset}->sync();
                        $self->{node}{apiset}->wafl_flush_cache();

                        my $size = NACL::C::Slist->slist(
                                command_interface => $self->{node},
                                path              => $full_file_name,
                            )->get_parsed_output()->[0]->{file_size};

                        unless ( $postrun ) {
                            push(@{$self->{$vvol_set}{$volume_name}{file_list}}, $file_name);
                            $self->{$vvol_set}{$volume_name}{$file_name}{'name'} = $file_name;
                            $self->{$vvol_set}{$volume_name}{$file_name}{'size'} = $size;
                            $self->{$vvol_set}{$volume_name}{$file_name}{'full_name'} = $full_file_name;
                        } else {
                            push (@plist, $full_file_name);
                        }
                    }
                }

            } else {
                $file_name = "file_s$postrun$s";
                if (defined $qtree) {
                    $full_file_name = "/vol/$volume_name/$qtree/$file_name";
                } else {
                    $full_file_name = "/vol/$volume_name/$file_name";
                }

                my $total_file_count = $file_count;
                my $file_suffix = 0;
                while ( $total_file_count ) {
                    my $new_file_name = $full_file_name;
                    if ( $total_file_count > 99 ) {
                        $file_count = 100;
                        $total_file_count = $total_file_count - $file_count;
                        $new_file_name    = $full_file_name . $file_suffix;
                        $file_suffix++;
                    }
                    my $create_file   = 1;
                    $total_file_count = 0;
                    if ( $fake_create ) {
                        for ( my $i = 0; $i < $file_count; $i++ ) {
                            my $ffile_name;
                            if ( $file_count <= 10 ) {
                                $ffile_name = $new_file_name . "\.$i";
                            } elsif ( $file_count <= 100 ) {
                                my $j = sprintf( "%02d", $i );
                                $ffile_name = $new_file_name . "\.$j";
                            }

                            my $base_name  = fileparse($ffile_name);
                            $create_file = 0;
                            try {
                                my $size = NACL::C::Slist->slist(
                                    command_interface => $self->{node},
                                    path              => $ffile_name,
                                    )->get_parsed_output()->[0]->{file_size};
                                push(@{$self->{$vvol_set}{$volume_name}{file_list}}, $base_name);
                                $self->{$vvol_set}{$volume_name}{$base_name}{'name'} = $base_name;
                                $self->{$vvol_set}{$volume_name}{$base_name}{'size'} = $size;
                                $self->{$vvol_set}{$volume_name}{$base_name}{'full_name'} = $ffile_name;
                            }
                            catch NACL::APISet::Exceptions::InvalidParamValueException with {
                                my $excpn_obj = shift;
                                $create_file = 1;
                            };
                            last if $create_file;
                        } # end of for
                    }

                    if ( $create_file ) {
                        my @file_objs = NACL::STask::VolumeFile->create(
                              command_interface     => $self->{node},
                              path                  => $new_file_name,
                              size                  => $file_size_to_fill,
                              vserver               => $vserver,
                              filecount             => $file_count,
                              'method-timeout'      => $timeout,
                              'step-size-in-blocks' => int(rand(5)),
                              nacltask_enable_space_reserve
                                  => $vvol_ref->{IRON_TEST_FILE_SPACERESERVE} );

                        # Sync & Flush Cache
                        $self->{node}{apiset}->sync();
                        $self->{node}{apiset}->sync();
                        $self->{node}{apiset}->wafl_flush_cache();

                        foreach my $volfile (@file_objs) {
                            my $fp = $volfile->path();
                            my $fn = fileparse($fp);
                            my $fs = NACL::C::Slist->slist(
                                command_interface => $self->{node},
                                path              => $fp,
                            )->get_parsed_output()->[0]->{file_size};

                            unless ( $postrun ) {
                                push(@{$self->{$vvol_set}{$volume_name}{file_list}}, $fn);
                                $self->{$vvol_set}{$volume_name}{$fn}{'name'} = $fn;
                                $self->{$vvol_set}{$volume_name}{$fn}{'size'} = $fs;
                                $self->{$vvol_set}{$volume_name}{$fn}{'full_name'} = $fp;
                            } else {
                                if ( $file_count == 1 ) {
                                    push (@plist, $new_file_name);
                                } elsif ( $file_count < 10 ) {
                                    my $i = 0;
                                    while ( $i < $file_count ) {
                                        $new_file_name = $new_file_name . '.' . $i;
                                        push (@plist, $new_file_name);
                                        $i++;
                                    }
                                } elsif ( $file_count < 100 ) {
                                    my $i = 0;
                                    my $j = sprintf( "%02d", $i );
                                    while ( $i < $file_count ) {
                                        $new_file_name = $new_file_name . '.' . $j;
                                        push (@plist, $new_file_name);
                                        $i++;
                                    }
                                }
                            }
                        } # end of foreach
                    } # end of if
                } # end of while
            }
        } ## end for

        # Sync & Flush Cache
        $self->{node}{apiset}->sync();
        $self->{node}{apiset}->sync();
        $self->{node}{apiset}->wafl_flush_cache();

        if ( $create_snapshot ) {
            # Create snapshot
            try {
                my $snap_name = $volume_name . "_snap.$postrun$s";
                $snapshot = NACL::STask::VolumeSnapshot->create(
                            command_interface  => $self->{node},
                            volume             => $volume_name,
                            vserver            => $vserver,
                            snapshot           => $snap_name,
                            nacltask_if_exists => "purge"
                        );

                unless ( $postrun ) {
                    push(@{$self->{$vvol_set}{$volume_name}{snap_list}}, $snap_name);
                    $self->{$vvol_set}{$volume_name}{$snap_name} = $snapshot;
                }
            }
            catch NACL::APISet::Exceptions::UnsupportedOptionException with {
                my $excpn_obj = shift;
                $Log->debug( "Caught exception: " . $excpn_obj->text() );
            };

            # Get inode block count
            my $inode_count = $self->get_inode_block_count(
                        inum      => '73',
                        cont_type => 'volume',
                        cont_name => $volume_name );

            # Perform wafl touch on activemap and then
            # this snapshot will not be sharing any block with active
            $Log->comment( "Performing wafl touch on activemap file" );
            try {
                $self->{node}{apiset}->wafl_touch(
                      'privilege-level' => 'test',
                      'volume'          => $volume_name,
                      'inode-number'    => '73',
                      'level-value'     => '0',
                      'private'         => '1',
                      'start-fbn'       => '0',
                      'end-fbn'         => $inode_count );
            }
            catch NACL::APISet::Exceptions::CommandFailedException with {
                my $excpn_obj = shift;
                $Log->debug( "Caught exception: " . $excpn_obj->text() );
            }
            catch NACL::APISet::Exceptions::TimeoutException with {
                my $excpn_obj = shift;
                $Log->debug( "Caught exception: " . $excpn_obj->text() );
            };
        }
    } ## end for

    # Clean up postrun files
    if ( $postrun ) {
        foreach my $file ( @plist ) {
            $self->{node}{apiset}->rm("paths" => $file);
        }
    }

    # Vol snapshot
    NACL::CS::VolumeSnapshot->fetch(
         command_interface => $self->{node},
         allow_empty       => 1
    );

} ## end sub _create_files_and_snapshots

# This cleans up the created resources
sub cleanup() {
    my $self = shift;
    $Log->enter();

    my $cmd_interface = $self->{node};
    my $client_ci     = $self->{client_apiset};
    my $ci_is_cmode   = $cmd_interface->is_cmode();

    # Stop traffic if running
    try {
        $self->stop_traffic();
    }
    catch NATE::BaseException with {
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
    };

    $Log->comment("Set default scan speed");
    try {
        $self->{node}{apiset}->wafl_scan_speed( 'speed' => '0' );
    }
    catch NATE::BaseException with {
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
        Tharn::sleep 60;
        $self->{node}{apiset}->wafl_scan_speed( 'speed' => '0' );
    };

    # Get wafliron status
    my @iron_status;
    try {
        @iron_status = NACL::CS::StorageAggregateWafliron->fetch(
            command_interface => $self->{node},
            requested_fields  => [ 'aggregate', 'iron-status' ]
        );
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        # Ignore NoMatchingEntriesException exception
    }
    catch NACL::Exceptions::NoElementsFound with {
        # Ignore NoElementsFound exception
    };

    # Stop iron if iron is still running
    foreach my $iron_status (@iron_status) {
        if ( $iron_status->iron_status() ne 'not_running' ) {
            $Log->comment("Wait for wafliron to complete...");
            $Log->comment("Stop iron if iron is still running...");
            if ( defined $self->{wafliron_obj}
                && $iron_status->aggregate() eq
                $self->{wafliron_obj}->aggregate() )
            {
                $self->{wafliron_obj}->wait_for_completion();
            }
            elsif ( defined $self->{wafliron_rerun_obj}
                && $iron_status->aggregate() eq
                $self->{wafliron_rerun_obj}->aggregate() )
            {
                $self->{wafliron_rerun_obj}->wait_for_completion();
            }
            else {
                # Create wafliron object
                my $iron_aggr    = $iron_status->aggregate();
                my $wafliron_obj = NACL::MTask::Wafliron->new(
                    command_interface => $self->{node},
                    aggregate         => $iron_aggr
                );

                # Create NACL::MTask::SystemLogDetector Object
                my $system_log_file
                            = ( $ci_is_cmode )
                              ? SYSTEM_LOG_FILES_CM
                              : SYSTEM_LOG_FILES_7M;
                my $log_detector = NACL::MTask::SystemLogDetector->new(
                                     command_interface => $cmd_interface,
                                     set_log_file      => [$system_log_file]
                                   );
                $log_detector->start();
                $wafliron_obj->system_log_object($log_detector);
                $wafliron_obj->wait_for_completion();
            }
        }
    }

    # Delete client lun traget
    if ( defined $self->{client}{lun_target}
                && $self->{client}{lun_target} ) {
        $Log->comment("Cleanup client settings...");
        $client_ci->set_privilege( 'privilege-level' => 'root' );

        $Log->comment("Disable automatic login...");
        $client_ci->iscsiadm(
                mode          => '1',
                node          => '1',
                'ip-port'     => $self->{client}{lun_address},
                'target-name' => $self->{client}{lun_target},
                operation     => 'update',
                name          => 'node.startup',
                value         => 'manual' );

        $Log->comment("Logout iscsi target...");
        my $logout_success = 1;
        while ($logout_success) {
            $logout_success = 0;
            try {
                $client_ci->iscsiadm(
                    mode          => '1',
                    node          => '1',
                    'ip-port'     => $self->{client}{lun_address},
                    'target-name' => $self->{client}{lun_target},
                    logout        => '1' );
            }
            catch NATE::BaseException with {
                my $excpn_obj = shift;
                $Log->debug( "Caught exception: " . $excpn_obj->text() );
                $logout_success = 1;
                Tharn::sleep 10;
            };
        }

        $Log->comment("Delete iscsi target...");
        $client_ci->iscsiadm(
                mode          => '1',
                node          => '1',
                'target-name' => $self->{client}{lun_target},
                operation     => 'delete' );

        $client_ci->iscsiadm(
                mode          => '1',
                discovery     => '1',
                'ip-port'     => $self->{client}{lun_address},
                operation     => 'delete' );

        $Log->comment("Umount lun mount(s)...");
        foreach my $mount ( @{$self->{lun_mountobj}} ) {
            try {
                $mount->purge();
            }
            catch NACL::APISet::Exceptions::CommandFailedException with {
                my $excpn_obj = shift;
                $Log->debug( "Caught exception: " . $excpn_obj->text() );
            };
        }
    }

    $Log->comment("Check for debug vreport & fix...");
    $self->_fix_debug_vreport( delete_lostvol => 1 );

    # Get aggr status (for debug)
    NACL::CS::Aggregate->fetch(
            command_interface => $self->{node},
            filter            => { nodes => [$self->{node}->node()] },
            requested_fields  => ['state'],
            allow_empty       => 1
    );

    # umount all client mounts
    if ( defined $self->{noof_volsets} ) {
        for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {
            # for all vols
            my $clone_vol_list = $self->{"VVOL_SET$i"}{cvol_list};
            my $vvol_vol_list  = $self->{"VVOL_SET$i"}{vol_list};
            foreach my $vol (@$clone_vol_list, @$vvol_vol_list) {
                # umount all client mounts
                if ( defined $self->{"VVOL_SET$i"}{$vol}{mount_obj}
                            && $self->{"VVOL_SET$i"}{$vol}{mount_obj} ) {
                    try {
                        $self->{"VVOL_SET$i"}{$vol}{mount_obj}->purge();
                    }
                    catch NACL::APISet::Exceptions::CommandFailedException with {
                        my $excpn_obj = shift;
                        $Log->debug( "Caught exception: " . $excpn_obj->text() );
                    };
                }
            }
        }
    }

    if (!defined($main::CLEANUP)
        || (defined($main::CLEANUP) && $main::CLEANUP ne 'none')) {

        # Delete lun(s) if any
        if (defined ${$self->{lunsetup_obj}}[0]) {
            $Log->comment("Deleting volume LUN");

            foreach my $lunsetup_obj ( @{$self->{lunsetup_obj}} ) {
                try {
                    $lunsetup_obj->cleanup(delete_all_objects => 1);
                }
                catch NACL::APISet::Exceptions::CommandFailedException with {
                    my $excpn_obj = shift;
                    $Log->debug( "Caught exception: " . $excpn_obj->text() );
                };
            }
        }

        if ( defined $self->{noof_volsets} ) {
            for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {
                # cleanup all clones
                my $clone_vol_list = $self->{"VVOL_SET$i"}{cvol_list};
                foreach my $vol (@$clone_vol_list) {
                    $Test->nlog("Deleting volume $vol");
                    my $vol_obj_ref = $self->{"VVOL_SET$i"}{$vol}{vol_obj_ref};
                    $self->_delete_vol( volume => $vol, vol_obj_ref => $vol_obj_ref );
                }

                # cleanup all snapmirror vols
                my $svol_list = $self->{"VVOL_SET$i"}{svol_list};
                foreach my $svol (@$svol_list) {
                    $Test->nlog("Break snapmirror on volume $svol");
                    my $snapmirror_object
                       = $self->{"VVOL_SET$i"}{$svol}{snapmirror_object};
                    try {
                        $snapmirror_object->break();
                    }
                    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
                        my $excpn_obj = shift;
                        $Log->comment( "Caught exception: " . $excpn_obj->text() );
                    }
                    catch NACL::C::Exceptions::Snapmirror::DataProtectionVolume with {
                        my $excpn_obj = shift;
                        $Log->debug( "Caught exception: " . $excpn_obj->text() );
                    }
                    catch NACL::Exceptions::InvalidChoice with {
                        my $excpn_obj = shift;
                        $Log->debug( "Caught exception: " . $excpn_obj->text() );
                    };

                    $snapmirror_object->release(
                          source_ci => $cmd_interface ) if $ci_is_cmode;

                    # Delete snapmirror
                    try  {
                        $snapmirror_object->delete();
                    }
                    catch NACL::Exceptions::InvalidChoice with {
                        my $excpn_obj = shift;
                        $Log->debug( "Caught exception: " . $excpn_obj->text() );
                    };
                    $Test->nlog("Deleting snapmirror volume $svol");
                    my $vol_obj_ref = $self->{"VVOL_SET$i"}{$svol}{vol_obj_ref};
                    $self->_delete_vol( volume => $svol, vol_obj_ref => $vol_obj_ref );
                }
            }

            for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {
                # cleanup all volumes
                my $vvol_vol_list  = $self->{"VVOL_SET$i"}{vol_list};
                foreach my $vol (@$vvol_vol_list) {
                    $Test->nlog("Deleting volume $vol");
                    my $vol_obj_ref = $self->{"VVOL_SET$i"}{$vol}{vol_obj_ref};
                    $self->_delete_vol( volume => $vol, vol_obj_ref => $vol_obj_ref );
                }
            }
        }

        # Skip aggr cleanup
        if ( defined($main::CLEANUP) && $main::CLEANUP ne 'volume' ) {
            if ( defined $self->{aggregate} && $self->{aggr_was_created} ) {
                $Test->nlog("Deleting Aggregate $self->{aggregate}{aggregate}");
                $self->{aggregate}->purge();
            }
        }
    }

    $Log->exit();
} ## end sub cleanup()

# This starts nfs traffic
sub start_traffic {
    my $self    = shift;
    my @args    = @_;
    my %options = validate(
        @args,
        {
            volume => { type => SCALAR | OBJECT },
            tool => { regex => '^(filersio|sio|hammer)$', default => 'sio' },
            fill  => { regex => qr/^[01]$/,      default  => 0 },
            mount => { type  => SCALAR | OBJECT, optional => 1 },
            mode  => {
                regex   => '^(mixed|creates|overwrites)$',
                default => 'mixed'
            }
        }
    );
    $Log->enter();

    my $flags_and_methods = $self->{iron_cparams}{IRON_FLAGS_AND_METHODS};
    my $iron_wafl_ops     = $self->{iron_cparams}{IRON_OPERATIONS};
    my %flags_and_ops     = ( %$flags_and_methods, %$iron_wafl_ops );
    my $iron_sio_runtime  = $flags_and_ops{IRON_FILERSIO_RUNTIME};
    my @subtests;

    my $io_tool     = $options{'tool'};     # filersio/sio/hammer
    my $volume_name = $options{'volume'};
    $volume_name    = $options{volume}->volume()
      if ( defined $options{volume} && ref $options{volume} );
    my $mount_path  = $options{'mount'};
    $mount_path     = $options{mount}->mount_point()
      if ( defined $options{mount} && ref $options{volume} );
    my $client_obj  = $self->{client};

    # Get random file size for io between 1g to 2g
    my $maximum_size = 2;
    my $minimum_size = 1;
    my $size         = int( rand($maximum_size) ) + $minimum_size;
    my $file_size    = $size . 'g';

    # Check if there is any LUN
    foreach my $lun_mount ( @{$self->{lun_mountobj}} ) {
        my $mount_point = $lun_mount->mount_point();
        my $is_traffic_running = 0;
        foreach my $traffic_href ( @{$self->{lun_iohash}} ) {
            next unless $$traffic_href{mount_point} eq $mount_point;
            $is_traffic_running
               = $$traffic_href{traffic_obj}->is_io_tool_running();
        }
        next if $is_traffic_running;

        my $file_name   = "file_" . random_name_generator();
        $Test->nlog("Try to start traffic on lun $mount_point");
        my $mount_path  =
                NACL::MTask::Traffic->get_abs_path($mount_point);
        $Log->debug("start_traffic::mount_point - $mount_point");

        # Create baseline file on client's mount point
        $file_name = "$mount_path/$file_name";

        # Run io
        my $traffic_obj = $self->_run_io (
                                 filename  => $file_name,
                                 filesize  => $file_size,
                                 mountpath => $mount_path,
                                 clientobj => $client_obj,
                                 runtime   => $iron_sio_runtime,
                                 iotool    => $io_tool,
                                 fill      => $options{'fill'},
                                 mode      => $options{'mode'} );
        if ( defined $traffic_obj ) {
            my %io_hash;
            $io_hash{traffic_tool} = 'mtask_traffic';
            $io_hash{traffic_obj}  = $traffic_obj;
            $io_hash{mount_point}  = $mount_point;
            push (@{$self->{lun_iohash}}, \%io_hash);
        }
    }

    for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {

        my @vol_list = $self->get_vvols( vvol_set => "VVOL_SET$i" );
        foreach my $vol (@vol_list) {

            next if ( defined $volume_name && $volume_name ne $vol );
            $Test->nlog("Try to start traffic on volume $volume_name");
            my $vol_obj = $self->{"VVOL_SET$i"}{$vol}{vol_obj_ref};

            $self->{"VVOL_SET$i"}{$vol}{traffic_list} = ()
              unless defined $self->{"VVOL_SET$i"}{$vol}{traffic_list};

            # Check for proper junction_path
            my @vols = NACL::CS::Volume->fetch(
                command_interface => $self->{node},
                filter            => { volume => $volume_name },
                requested_fields  => [ 'junction-path', 'state' ],
                allow_empty       => 1 );

            my $junction_path = $vols[0]->junction_path();
            my $state         = $vols[0]->state();
            my $interface     = $self->{node};
            my $vserver       = $vols[0]->vserver();
            my $file_name     = "file_" . random_name_generator();
            my ($traffic_obj, $mount_obj);

            # Qtree info
            my @qtree_list = ();
            @qtree_list = @{$self->{"VVOL_SET$i"}{$vol}{qtree_list}}
                    if defined $self->{"VVOL_SET$i"}{$vol}{qtree_list};

            # Use client traffic in case of luns
            $flags_and_ops{IRON_USE_FILER_IO_TOOL} = 0
                if defined $self->{"VVOL_SET$i"}{$vol}{ulun_obj};

            $junction_path = '-'
              if (defined $flags_and_ops{IRON_USE_FILER_IO_TOOL}
                 && $flags_and_ops{IRON_USE_FILER_IO_TOOL});

            #my @io_hash_list = ();
            unless ( $junction_path eq '-' ) {
                # Get filer interfaces
                my @nfs_lifs = NACL::CS::Interface->fetch(
                            command_interface => $interface,
                            filter            => {
                                vserver         => $vserver,
                                role            => 'data',
                                'data-protocol' => ['nfs'],
                                'status-admin'  => 'up'
                            });
                my $nfs_lif = shift(@nfs_lifs);

                $mount_obj = NACL::STask::Client::Mount->create(
                                server_command_interface => $interface,
                                command_interface        => $client_obj,
                                server                   => $nfs_lif->{address},
                                volume                   => $vol_obj,
                                chmod                    => '777' );

                # Get the mount point
                my $mount_point = $mount_obj->mount_point();
                my $mount_path =
                            NACL::MTask::Traffic->get_abs_path($mount_point);
                $Log->debug("start_traffic::mount_point - $mount_point");

                $self->{client_vol_mount_path} = $mount_path;
                $self->{client_mount_path_vol} = $vol_obj->volume();
                $self->{"VVOL_SET$i"}{$vol}{mount_obj}  = $mount_obj;
                $self->{"VVOL_SET$i"}{$vol}{mount_path} = $mount_path;

                my $qtree_count = $#qtree_list + 1;
                if ( $qtree_count ) {
                    foreach my $qtree ( @qtree_list ) {
                        # Create baseline file on client's mount point
                        my $fname = "$mount_path/$qtree/$file_name";

                        # Run io
                        $traffic_obj = $self->_run_io (
                                 filename  => $fname,
                                 filesize  => $file_size,
                                 mountpath => $mount_path,
                                 clientobj => $client_obj,
                                 runtime   => $iron_sio_runtime,
                                 iotool    => $io_tool,
                                 fill      => $options{'fill'},
                                 mode      => $options{'mode'} );

                        if ( defined $traffic_obj ) {
                            my %io_hash;
                            $io_hash{traffic_tool} = 'mtask_traffic';
                            $io_hash{traffic_obj}  = $traffic_obj;
                            push ( @{$self->{"VVOL_SET$i"}{$vol}{traffic_list}}, \%io_hash );
                        }
                    }
                } else {
                    # Create baseline file on client's mount point
                    $file_name = "$mount_path/$file_name";

                    # Run io
                    $traffic_obj = $self->_run_io (
                                 filename  => $file_name,
                                 filesize  => $file_size,
                                 mountpath => $mount_path,
                                 clientobj => $client_obj,
                                 runtime   => $iron_sio_runtime,
                                 iotool    => $io_tool,
                                 fill      => $options{'fill'},
                                 mode      => $options{'mode'} );

                    if ( defined $traffic_obj ) {
                        my %io_hash;
                        $io_hash{traffic_tool} = 'mtask_traffic';
                        $io_hash{traffic_obj}  = $traffic_obj;
                        push ( @{$self->{"VVOL_SET$i"}{$vol}{traffic_list}}, \%io_hash );
                    }
                }
            }
            else {
                # When you don't have proper junction path - use filer sio
                # Create sio baseline file on filer's mount point

                if ( $io_tool =~ /sio/i ) {
                    # SIO traffic
                    $Log->comment('Using filersio to create file');

                    push @subtests, subtest(
                       sub{ _filersio_asyncio_active($self, "VVOL_SET$i",
                             $volume_name, $file_name, $file_size); },
                                 -bg, -runid, "load_$volume_name",
                                '-iffatal', '-ok', '-ifscript', '-ok' );
                }
                else {
                    # HAMMER traffic
                    $Log->comment('NFS traffic tool : hammer');
                    $file_name = "/vol/$volume_name/$file_name";
                    $traffic_obj = $self->{node}{apiset}->hammer(
                        'force'      => "1",
                        'num-runs'   => "1",
                        'file-name'  => $file_name,
                        'num-blocks' => "1000"
                    );
                    my %io_hash;
                    $io_hash{traffic_tool} = 'hammer';
                    $io_hash{traffic_obj}  = $traffic_obj;
                    push ( @{$self->{"VVOL_SET$i"}{$vol}{traffic_list}}, \%io_hash );
                }
            }

        } ## end foreach
    } ## end for

    $Log->exit();
} ## end sub start_traffic

# filersio for traffic
sub _filersio_asyncio_active {
    my ($self, $vol_set, $vol_name, $file_name, $file_size) = @_;
    $Log->enter();

    my $filename = "/vol/$vol_name/$file_name";
    my $traffic_obj =
           $self->{node}{apiset}->filersio_asyncio_active(
                        'read-percentage' => 0,
                        'rand-percentage' => 100,
                        'blk-size'        => '4096',
                        'start-byte'      => 0,
                        'end-byte'        => $file_size,
                        'run-time'        => 10,
                        'num-active'      => 10,
                        filename          => $filename,
                        create            => 1,
                        synchronous       => 1 );

    my %io_hash;
    $io_hash{traffic_tool} = 'filersio';
    $io_hash{traffic_obj}  = $traffic_obj;
    push ( ${$self->{$vol_set}{$vol_name}{traffic_list}}, \%io_hash );

    $Log->exit();
}

# Online corrupted vol/aggr
sub _online_corrupted_object {
    my ( $self, @args ) = @_;
    my %options = validate(
        @args,
        {
            'corrupted_vol'  => { type => SCALAR | OBJECT, optional => 1 },
            'corrupted_aggr' => { type => SCALAR | OBJECT, optional => 1 },
            'access_file'    => { type => HASHREF,         optional => 1 },
        },
    );
    $Log->enter();

    my $corrupted_vol     = $options{'corrupted_vol'};
    my $corrupted_aggr    = $options{'corrupted_aggr'};
    my $access_file       = $options{'access_file'};
    my $flags_and_methods = $self->{iron_cparams}{IRON_FLAGS_AND_METHODS};
    my $iron_wafl_ops     = $self->{iron_cparams}{IRON_OPERATIONS};
    my %flags_and_ops     = ( %$flags_and_methods, %$iron_wafl_ops );
    my $online_all        = 0;
    my $corrupted_vol_obj;

    # Online corrupted vol/aggr if requested
    $Log->comment("Online corrupted vol/aggr");
    $online_all = 1
      if ( !defined $corrupted_vol && !defined $corrupted_aggr );

    # Store flags, to restore after filer reboot
    my %flags_hash = (
        'wafl_metadata_visible'           => undef,
        'wafl_iron_panic_on_corruption'   => undef,
        'wafl_inconsistent_panic_testing' => undef,
        'wafl_metadata_checks_in_raid'    => undef,
        'wafl_iron_lazy_dist_checks'      => undef,
        'wafl_snapiron_enabled'           => undef,
        'wafl_maxvvbn_disabled'           => undef,
        'wafl_lwr_trip'                   => undef
    );

    foreach my $flag ( keys %flags_hash ) {
        $flags_hash{$flag} = $self->{node}{apiset}->printflag(
            'flags'           => $flag,
            'privilege-level' => 'test'
        )->get_parsed_output()->[0]->{$flag};
    }
    $Log->comment( "flags saved: " . Dumper \%flags_hash );

    # We could expect panic here with corrupted vol/aggr
    $self->{node}{apiset}->bootargs_set(
        arg   => 'AUTOBOOT',
        value => 'true'
    );

    if ( defined $corrupted_aggr || defined $corrupted_vol || $online_all ) {

        # online aggr
        my @aggrs = NACL::CS::Aggregate->fetch(
            command_interface => $self->{node},
            filter            => {
                nodes => [ $self->{node}->node() ]
            },
            requested_fields  => ['state'],
            allow_empty       => 1
        );

        $corrupted_aggr = $self->{aggregate}{aggregate}
                          unless defined $corrupted_aggr;
        foreach my $aggr_obj (@aggrs) {
            next unless $aggr_obj->aggregate() eq $corrupted_aggr;

            $aggr_obj->get_component_instance()
                  ->modify( state => 'online' )
                     if $aggr_obj->state() ne 'online';
        }
    }

    if ( defined $corrupted_vol || $online_all ) {

        # online volume(s)
        my @vols = NACL::CS::Volume->fetch(
            command_interface => $self->{node},
            requested_fields  => ['state'],
            allow_empty       => 1
        );

        foreach my $vol_obj (@vols) {
            next
              unless (
                $online_all
                || ( defined $corrupted_vol
                    && $corrupted_vol eq $vol_obj->volume() )
              );
            $corrupted_vol_obj = $vol_obj;
            $vol_obj->get_component_instance()->modify( state => 'online' )
              if $vol_obj->state() ne 'online';
        }
    }

    # Access file requested
    my ( $access_method, $file_name, $full_file_name, $file_size, $volume_obj );
    my ( $read, $rand, $truncate, $file_list, $exception_flag );

    # Get right vol_obj
    my $client_obj           = $self->{client};
    my $found_test_file_flag = 0;
    my $found_test_file      = 0;
    my ( @vol_list, $vol_set, $volume_name );
    foreach my $volset ( @{ $self->{volset_list} } ) {
        $vol_set = $volset if defined $volset;
        @vol_list = $self->get_vvols( vvol_set => $vol_set );
        if (   defined $corrupted_vol
            && defined $self->{$vol_set}{$corrupted_vol}{vol_obj_ref} )
        {
            $volume_obj  = $self->{$vol_set}{$corrupted_vol}{vol_obj_ref};
            $file_list   = $self->{$vol_set}{$corrupted_vol}{file_list};
            $volume_name = $volume_obj->volume();
            if ( defined $options{'access_file'} ) {

                # Copy file details
                $file_name = $$access_file{'file_name'};
                $full_file_name
                  = $self->{$vol_set}{$corrupted_vol}{$file_name}{full_name};
                $file_size = $$access_file{'file_size'};
                $file_size = NACL::C::UnitNormalization::to_bytes($file_size);
                $file_size     = $file_size - 10;                  # avoid end
                $access_method = $$access_file{'access_method'};
                if (defined $self->{$vol_set}{$corrupted_vol}{$file_name}) {
                    $found_test_file++;
                    $full_file_name
                     = $self->{$vol_set}{$corrupted_vol}{$file_name}{full_name};
                } else {
                    $full_file_name = "/vol/$corrupted_vol/$file_name";
                }
            }
            last;
        }
    }

    unless ($found_test_file) {

        # Work in random volume
        my $index = rand @vol_list;
        my $vol   = $vol_list[$index];
        $volume_obj  = $self->{$vol_set}{$vol}{vol_obj_ref};
        $volume_name = $volume_obj->volume();

        # Random file read access
        $access_method  = 'read';
        $file_list      = $self->{$vol_set}{$vol}{file_list};
        $index          = rand @$file_list;
        $file_name      = $$file_list[$index];
        $full_file_name = $self->{$vol_set}{$vol}{$file_name}{full_name};
        $file_size      = $self->{$vol_set}{$vol}{$file_name}{'size'};
    }

    # See of file present
    try {
        NACL::C::Slist->slist(
            command_interface => $self->{node},
            path              => $full_file_name
        );
    }
    catch NATE::BaseException with {
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
        $exception_flag++;
    };

    # SIO command options
    if ( defined $access_method
        && $access_method eq 'read' )
    {
        $read     = '100';    # 100% read
        $rand     = '60';
        $truncate = '0';
    }
    elsif ( defined $access_method
        && $access_method eq 'write' )
    {
        $read     = '0';      # 0% read
        $rand     = '60';
        $truncate = '1';
    }
    else {
        $read     = '50';     # 0% read
        $rand     = '100';
        $truncate = '1';
    }

    try {
        $Test->nlog("Mounting volume on client ");

        my $vserver   = $volume_obj->vserver();
        my $mount_obj = NACL::STask::Client::Mount->create(
            command_interface        => $client_obj,
            server_command_interface => $self->{node},
            vserver                  => $vserver,
            volume                   => $volume_obj,
        );

        # Get the mount point
        my $mount_point = $mount_obj->mount_point();
        my $mount_path  = NACL::MTask::Traffic->get_abs_path($mount_point);
        my $file_path   = $mount_path . "/$file_name";

        # Start sio
        $Test->nlog("Running sio on $file_path");
        my $client_api = $self->{client_apiset};
        $client_api->sio(
            'read'               => $read,
            'rand'               => $rand,
            'block_size'         => '1k',
            'start_byte'         => 0,
            'file_size'          => $file_size,
            'duration_in_secs'   => 120,
            'devices_list'       => $file_path,
            'truncate'           => $truncate,
            'num_of_threads'     => 2,
            'connectrec-timeout' => 300
        );

    }
    catch NATE::BaseException with {
        # Ignore as we are expecting filer panic
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
        my $message = "Failed while on-lining corrupted vol/aggr";
        NACL::APISet::Exceptions::ConnectionFailedException->throw($message);
    };
    $Log->exit();
} ## end sub _online_corrupted_object

# Start wafliron/ioc on requested aggr
sub run_iron {
    my ( $self, @args ) = @_;
    my %options = validate(
        @args,
        {
            'command_interface' =>
              { type => OBJECT, isa => 'NACL::C::CommandInterface' },
            'aggregate'        => { type => SCALAR | OBJECT },
            'wait'             => { type => SCALAR, default => '0' },
            'scan_speed'       => { type => SCALAR, default => '-1' },
            'method-timeout'   => { type => SCALAR, default => '3600' },
            'include_mirrors'  => { type => SCALAR, default => '1' },
            'corrupted_vol'    => { type => SCALAR | OBJECT, optional => 1 },
            'corrupted_aggr'   => { type => SCALAR | OBJECT, optional => 1 },
            'root_aggr'        => { type => SCALAR | OBJECT, optional => 1 },
            'root_vol'         => { type => SCALAR | OBJECT, optional => 1 },
            'flags_to_set'     => { type => HASHREF, optional => 1 },
            'access_file'      => { type => HASHREF, optional => 1 },
            'return_iron_msgs' => { type => ARRAYREF, optional => 1 },
        },
    );
    $Log->enter();

    my $wait       = $options{wait};
    my $aggregate  = $options{aggregate};
    my $scan_speed = $options{scan_speed};

    my $include_mirrors   = $options{include_mirrors};
    my $command_interface = $options{command_interface};
    my $method_timeout    = $options{'method-timeout'};
    my $corrupted_vol     = $options{'corrupted_vol'};
    my $corrupted_aggr    = $options{'corrupted_aggr'};
    my $flags_to_set      = $options{'flags_to_set'};
    my $access_file       = $options{'access_file'};
    my $return_iron_msgs  = $options{'return_iron_msgs'};

    my $root_aggr = $options{root_aggr};
    $root_aggr = $options{root_aggr}->aggregate()
      if ( defined $options{root_aggr} && ref $options{root_aggr} );
    my $root_vol = $options{root_vol};
    $root_vol = $options{root_vol}->volume()
      if ( defined $options{root_vol} && ref $options{root_vol} );

    # Get all flags
    my $flags_and_methods = $self->{iron_cparams}{IRON_FLAGS_AND_METHODS};
    my $iron_wafl_ops     = $self->{iron_cparams}{IRON_OPERATIONS};
    my %flags_and_ops     = ( %$flags_and_methods, %$iron_wafl_ops );
    my $optional_commit    = 'false';

    # Check if this is c-mode filer
    my $cmd_interface = $self->{node};
    my $ci_is_cmode   = $cmd_interface->is_cmode();

    # Can not wait for wafliron complete
    # if scanspeed is negative, set default speed
    $scan_speed = '30000' if $wait;

    # Update requested flags
    $$flags_to_set{wafl_metadata_checks_in_raid}
                  = $flags_and_ops{IRON_MDLC_DISABLE}
        unless (defined $$flags_to_set{wafl_metadata_checks_in_raid});
    $$flags_to_set{wafl_iron_lazy_dist_checks}
                  = $flags_and_ops{IRON_LDC_ENABLE}
        unless (defined $$flags_to_set{wafl_iron_lazy_dist_checks});
    $$flags_to_set{wafl_snapiron_enabled}
                  = $flags_and_ops{IRON_SNAPIRON_ENABLE}
        unless (defined $$flags_to_set{wafl_snapiron_enabled});
    $$flags_to_set{wafl_iron_defer_summary_check}
                  = $flags_and_ops{IRON_DSC_ENABLED}
        unless (defined $$flags_to_set{wafl_iron_defer_summary_check});
    $$flags_to_set{wafl_iron_predictability_testing}
                  = $flags_and_ops{IRON_PREDICTABILITY_TESTING}
          if defined $flags_and_ops{IRON_PREDICTABILITY_TESTING};

    $method_timeout = $flags_and_ops{IRON_TIMEOUT}
             if ( defined $flags_and_ops{IRON_TIMEOUT} );

    my ($root_aggregate, $root_volume, %set_flags, %saved_flags);
    my ($root_vol_obj, $root_aggr_obj, $maint_mode_api);
    if ( defined $flags_and_ops{IRON_ON_ROOT}
        && $flags_and_ops{IRON_ON_ROOT} =~ /core/ ) {
        # Get the current root aggr & root volume
        $root_aggregate = NACL::STask::Aggregate->get_root(
                                command_interface => $command_interface );
        my $parsed_output = $self->{node}{apiset}->vol_status()
                                          ->get_parsed_output();
        foreach my $href ( @{$parsed_output} ) {
            $root_volume = $href->{'volume'}
                if $href->{'options'}->[0]->{'root'};
        }

        # check root_vol info
        if (defined $root_vol) {
            $Log->comment("Check root vol $root_vol volume exists");
            try {
                $root_vol_obj = NACL::CS::Volume->fetch(
                    command_interface => $command_interface,
                    filter            => {volume => $root_vol});
            }
            catch NACL::Exceptions::NoElementsFound with {
                my $exception = shift;
                $Log->exit();
                $exception->throw();
            };

            if ( defined $root_aggr && $root_aggr ne $root_vol_obj->aggregate() ) {
                $Log->exit();
                NATE::BaseException->throw(
                   "root vol $root_vol should be under root aggr $root_aggr"); 
            } else {
                $root_aggr = $root_vol_obj->aggregate();
            }
        }

        # Check type of root_aggr
        if (defined $root_aggr) {
            # Get aggr info
            try {
                $root_aggr_obj = NACL::CS::Aggregate->fetch(
                      command_interface => $command_interface,
                      filter            => {
                             nodes => [$self->{node}->node()],
                             aggregate => $root_aggr },
                      requested_fields  => ['ha-policy'],
                      allow_empty       => 1 );
            }
            catch NACL::Exceptions::NoElementsFound with {
                my $exception = shift;
                $Log->exit();
                $exception->throw();
            };

            # sfo cannot become root aggr
            if ( $root_aggr_obj->ha_policy() eq 'sfo' ) {
                $Log->exit();
                NATE::BaseException->throw(
                        "sfo aggr cannot become root aggr" );
            }
        }

        # corresponding env variables to flags
        if ( defined $flags_to_set ) {
            foreach my $flag ( keys %{$flags_to_set} ) {
                my $val;
                if ( $$flags_to_set{$flag} ) {
                    $val = 'true';
                } else {
                    $val = 'false';
                }

                if ( $flag eq 'wafl_iron_panic_on_corruption' ) {
                    $set_flags{'wafl-iron-panic-on-corruption?'} = $val;
                } elsif ( $flag eq 'wafl_inconsistent_panic_testing' ) {
                    $set_flags{'wafl-inconsistent-panic-testing?'} = $val;
                }
            }
            $set_flags{'wafl-ioc-root-rsh-disable?'} = 'true';
        }

        $Log->comment( "Saving the original value of environment"
                              . " variables to be reset in cleanup" );
        foreach my $var ( keys %set_flags ) {
            $saved_flags{$var} = $self->{node}{apiset}
                 ->bootargs_get( arg => $var )
                 ->get_parsed_output()->[0]->{$var};
        }
        $Log->debug("DscSetup: set_flags - ". Dumper \%set_flags);
        $Log->debug("DscSetup: saved_flags - ". Dumper \%saved_flags);

        my $transit_obj = NACL::Transit->new(
                      name => $command_interface->name() );
        my $conn = $transit_obj->{'conn'};
        $transit_obj->set( 'change_state_timeout' => 1200 );
        $transit_obj->change_state( to => 'FIRMWARE' );
        my $firmware_apiset_obj
                 = $command_interface->apiset(
                         interface => 'CLI', set => 'Firmware' );

        $Log->comment("Setting env flags...");
        foreach my $var ( keys %set_flags ) {
            $firmware_apiset_obj->setenv(
                      'varname' => $var,
                      'value'   => $set_flags{$var} );
        }

        if (defined $root_aggr && $root_aggregate ne $root_aggr) {
            # Setting new root aggr
            $Log->comment("Setting the $root_aggr as root aggregate");
            $transit_obj->change_state( 'to' => 'MAINT' );

            $maint_mode_api = $command_interface->apiset(
                         interface => 'CLI', set => 'Maintenance' );
            my $text;
            try {
                $text = $maint_mode_api->aggr_online(
                                    'aggregate'          => $root_aggr,
                                    'connectrec-timeout' => '300'
                                  )->get_raw_output();
            }
            catch NACL::APISet::Exceptions::CommandFailedException with {
                my $exception = shift;
                if ( $exception->text() !~ /already online/is ) {
                    $Log->exit();
                    $exception->throw();
                }
            };

            # set new aggr as root
            $text = $maint_mode_api->aggr_options(
                          'aggregate'   => $root_aggr,
                          'option-name' => 'root',
                        )->get_raw_output();
        }

        # Go to boot menu and start wafliron
        $transit_obj->change_state( to => 'BOOT_MENU' );

        # pick new root vol
        if (defined $root_vol && $root_volume ne $root_vol) {
            $Log->comment("Picking $root_vol as root volume");
            my $cmd  = "vol_pick_root $root_vol $root_aggr";
            $conn->execute(
                     'command'            => $cmd,
                     'additional_prompts' => [qr/Selection (1-8)?/i],
                     'timeout'            => 160 );
        }

        # Run iron from boot menu
        $conn->execute(
                'command'            => 'wafliron',
                'additional_prompts' => [qr/login/i],
                'timeout'            => 3000 );

        # Bring filer back to normal
        $transit_obj->change_state( to => 'CLI' );

        die("TODO FOR IOC CORE mode...");
    }

    # Will try to online corrupted vol/aggr
    if ( defined $flags_and_ops{IRON_ONLINE_CORRUPTED_VOL}
        && $flags_and_ops{IRON_ONLINE_CORRUPTED_VOL}
        && !$flags_and_ops{IRON_OPTIONAL_COMMIT} )
    {
        try {
            $Log->comment("Trying to online corrupted vol/aggr");
            if ( defined $options{corrupted_vol} ) {
                $self->_online_corrupted_object(
                    corrupted_vol => $corrupted_vol,
                    access_file   => $access_file
                );
            }
            elsif ( defined $options{corrupted_aggr} ) {
                $self->_online_corrupted_object(
                    corrupted_aggr => $corrupted_aggr );
            }
            else {
                $self->_online_corrupted_object();
            }
        }

        catch NATE::BaseException with {
            # Could be filer panic
            my $e = shift;
            $Log->comment( "Exception string: " . $e->text() );
            $self->_check_connection();
        };
    }

    $optional_commit = 'true'
      if ( defined $flags_and_ops{IRON_OPTIONAL_COMMIT}
                    && $flags_and_ops{IRON_OPTIONAL_COMMIT} );

    # Create/Use wafliron object
    my $wafliron_obj;
    if ( defined $self->{wafliron_obj} ) {
        $wafliron_obj = $self->{wafliron_obj};
    } else {
        $wafliron_obj = NACL::MTask::Wafliron->new(
                           command_interface => $command_interface,
                           aggregate         => $aggregate );
        $self->{wafliron_obj} = $wafliron_obj;
    }

    if ( defined $flags_to_set ) {
        # Start wafliron
        $wafliron_obj->start(
            nacltask_include_mirrors => $include_mirrors,
            'method-timeout'         => $method_timeout,
            nacltask_wait            => $wait,
            nacltask_wafl_scan_speed => $scan_speed,
            nacltask_optional_commit => $optional_commit,
            nacltask_flags_to_set    => $flags_to_set );
    } else {
        # Start wafliron
        $wafliron_obj->start(
            nacltask_include_mirrors => $include_mirrors,
            'method-timeout'         => $method_timeout,
            nacltask_wait            => $wait,
            nacltask_optional_commit => $optional_commit,
            nacltask_wafl_scan_speed => $scan_speed );
    }

    my $endtime = time() + $method_timeout;
    my $wait_for_mount = sub {
        while ( time() < $endtime ) {
            my $wafl_cs        = $wafliron_obj->wafliron_component->state;
            my $wafliron_state = $wafl_cs->iron_status;
            my @states         = ( 'mounting', 'unmounting' );
            my $states_regex   = '(' . ( join "\$\|\^", @states ) . ')';
            $states_regex      = qr/$states_regex/i;

            if ( $wafliron_state =~ $states_regex ) {
                $self->{node}{apiset}
                   ->wafl_scan_speed( 'speed' => '1' );
                Tharn::snooze 5;
            } else {
                last;
            }
        }};

    # Wait for mount phase to complete
    unless ( $wait ) {
        $Test->nlog("Run requested load & operations on filer");
        $wait_for_mount->();
        $self->{node}{apiset}->wafl_scan_speed(
               'speed' => $flags_and_ops{IRON_SCANNER_SPEED} );
        try {
            $self->run_filer_ops()
              if defined $flags_and_ops{IRON_COMMIT_CHANGES};
        }
        catch NACL::APISet::Exceptions::InvalidParamException with {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
        };
    }

    return $self->{wafliron_obj}
         if !defined $flags_and_ops{IRON_COMMIT_CHANGES};

    # Increase the wafl scan speed so that wafliron finishes
    unless ( $wait ) {
        try {
            $Test->nlog("Increase wafl scan speed");
            $self->{node}{apiset}->wafl_scan_speed('speed' => '30000');
        }
        catch NATE::BaseException with {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
            Tharn::sleep 20;
            $self->{node}{apiset}->wafl_scan_speed( 'speed' => '0' );
        };
    }

    # Wait for wafliron to complete
    $Test->nlog("Wait for wafliron to complete");
    try {
        $self->{wafliron_obj}
          ->wait_for_completion( 'method-timeout' => $method_timeout );
    }
    catch NATE::BaseException with {
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
    };

    # Check wafliron review
    if ( $optional_commit eq 'true' ) {
        my $review_messages = 'commit';
        $review_messages = $self->{wafliron_obj}->review()
                  unless $flags_and_ops{IRON_SKIP_REVIEW_CMD};
        unless ( $review_messages =~ /either no changes/ ) {
            if ( $flags_and_ops{IRON_COMMIT_CHANGES} ) {
                $self->{wafliron_obj}->commit();

                # online aggr
                my $aggr_obj = $self->{aggregate};
                $aggr_obj->modify( state => 'online' );
            } else {
                $self->{wafliron_obj}->reject();
            }
        } else {
            # online aggr
            my $aggr_obj = $self->{aggregate};
            $aggr_obj->modify( state => 'online' );
        }
    }

    # Stop traffic if running
    try {
        $self->stop_traffic() unless $wait;
    }
    catch NATE::BaseException with {
        my $excpn_obj = shift;
        $Log->comment( "Caught exception while stop "
                 . "traffic: " . $excpn_obj->text() );
    };

    # Check debug vreport & fix
    $self->_fix_debug_vreport( delete_lostvol => 0 );

    # Use log object & get wafliron messages
    my ( $log_obj, @message_list, @msgs );
    $log_obj      = $self->{wafliron_obj}->system_log_object();
    @message_list = ('wafl.check.info:error');

    # Check log for matched messages
    try {
        $log_obj->check( check_for_all_presence => \@message_list );

        # Get all matched messages
        @msgs = $log_obj->get_matched_logs();
        $Log->comment( "Iron/IOC Messages - \n" . Dumper \@msgs );
        $return_iron_msgs = \@msgs;
    }
    catch NACL::Exceptions::EventCheckFailure with {
        # Don't fail if no logs messages found
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
        my $w_msgs = $log_obj->get_logs();
        $Log->comment( "Iron/IOC Log - \n" . Dumper \$w_msgs );
    };

    if ( exists( $options{'return_iron_msgs'} ) ) {
        @{ $options{'return_iron_msgs'} } = @msgs;
    }

    # Check if EMS messages still throwing error messages
    $endtime = time() + $method_timeout;
    my $exit_while = 1;
    my $log_detector = $self->{wafliron_obj}->system_log_object();
    $log_detector->start();
    while ( time() < $endtime && $exit_while ) {
        try {
            Tharn::snooze 20;
            # Get wafliron messages
            my $w_msgs = $log_detector->get_logs( skip_cached_logs => 1 );
            $Log->debug("Log detector messages: " . Dumper $w_msgs);
            unless ( $w_msgs =~  m[wafl\.check\.info:(error|ALERT)] ) {
                Tharn::snooze 10;
                $exit_while = 0;
            }
        }
        catch NATE::BaseException with {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
        }
        catch NACL::Exceptions::NoElementsFound with {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
        };
    }

    # Vol info for debug
    NACL::CS::Volume->fetch(
                command_interface => $cmd_interface );

    my $volume_name = 'all';
    $volume_name = $options{corrupted_vol}
            if (defined $options{corrupted_vol});
    $volume_name = $options{corrupted_vol}->volume()
            if (defined $options{corrupted_vol}
                     && ref $options{corrupted_vol});

    # Delete all snapshots on request
    if ( defined $flags_and_ops{IRON_RECOVERY_SNAPSHOT_DELETE_ALL}
        && $flags_and_ops{IRON_RECOVERY_SNAPSHOT_DELETE_ALL} )
    {
        $Test->nlog("Delete all snapshots before iron rerun...");
        $self->_delete_snapshots(
                     command_interface => $command_interface,
                     aggregate         => $aggregate,
                     volume            => $volume_name );
    }

    $Log->exit();
    return $self->{wafliron_obj};
} ## end sub run_iron

# Start wafliron/ioc re-run on requested aggr
sub rerun_iron {
    my ( $self, @args ) = @_;
    my %options = validate(
        @args,
        {
            'command_interface' =>
              { type => OBJECT, isa => 'NACL::C::CommandInterface' },
            'aggregate'        => { type => SCALAR | OBJECT },
            'method-timeout'   => { type => SCALAR, default => '3600' },
            'include_mirrors'  => { type => SCALAR, default => '1' },
            'flags_to_set'     => { type => HASHREF, optional => 1 },
            'return_iron_msgs' => { type => ARRAYREF, optional => 1 },
        },
    );
    $Log->enter();

    my $aggregate         = $options{aggregate};
    my $include_mirrors   = $options{include_mirrors};
    my $command_interface = $options{command_interface};
    my $method_timeout    = $options{'method-timeout'};
    my $flags_to_set      = $options{'flags_to_set'};
    my $return_iron_msgs  = $options{'return_iron_msgs'};

    # Get all flags
    my $flags_and_methods = $self->{iron_cparams}{IRON_FLAGS_AND_METHODS};
    my $iron_wafl_ops     = $self->{iron_cparams}{IRON_OPERATIONS};
    my $optional_commit     = 'false';
    my %flags_and_ops     = ( %$flags_and_methods, %$iron_wafl_ops );

    $method_timeout = $flags_and_ops{IRON_RECOVERY_WAIT}
             if ( defined $flags_and_ops{IRON_RECOVERY_WAIT} );

    # Sync & Flush Cache
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->wafl_flush_cache();

    # Vol info for debug
    try {
        NACL::CS::Volume->fetch(
            command_interface => $self->{node},
            filter            => {aggregate => $aggregate});
    }
    catch NACL::Exceptions::NoElementsFound with {
        my $excpn_obj = shift;
        $Log->warn("Unable to find volume. "
              . "Got exception:" . $excpn_obj->text());
    };

    # Create new wafliron object
    $self->{wafliron_rerun_obj} = NACL::MTask::Wafliron->new(
        command_interface => $command_interface,
        aggregate         => $aggregate
    );

    $optional_commit = 'true'
      if ( defined $flags_and_ops{IRON_OPTIONAL_COMMIT}
            && $flags_and_ops{IRON_OPTIONAL_COMMIT} );

    if ( defined $flags_to_set ) {
        # Start wafliron
        $self->{wafliron_rerun_obj}->start(
            nacltask_include_mirrors => $include_mirrors,
            'method-timeout'         => $method_timeout,
            nacltask_flags_to_set    => $flags_to_set,
            nacltask_optional_commit => $optional_commit,
            nacltask_wait            => '0',
            nacltask_wafl_scan_speed => '30000'
        );
    } else {
        # Start wafliron
        $self->{wafliron_rerun_obj}->start(
            nacltask_include_mirrors => $include_mirrors,
            'method-timeout'         => $method_timeout,
            nacltask_optional_commit => $optional_commit,
            nacltask_wait            => '0',
            nacltask_wafl_scan_speed => '30000'
        );
    }

    # Wait for wafliron to complete
    $Test->nlog("Iron/IOC rerun. Wait for wafliron to complete");
    try {
        $self->{wafliron_rerun_obj}
             ->wait_for_completion( 'method-timeout' => $method_timeout );
    }
    catch NATE::BaseException with {
        my $excpn_obj = shift;
        $Log->comment( "Caught exception: " . $excpn_obj->text() );
    };

    # Set wafl scan speed to default
    $Test->nlog("Set default scan speed");
    try {
        $self->{node}{apiset}->wafl_scan_speed( 'speed' => '0' );
    }
    catch NATE::BaseException with {
        Tharn::sleep 20; # If fail, wait & try again
        $self->{node}{apiset}->wafl_scan_speed( 'speed' => '0' );
    };

    # Use log object & get wafliron messages
    # Check wafliron review
    if ( $optional_commit eq 'true' ) {
        my $review_messages = $self->{wafliron_obj}->review();
        unless ( $review_messages =~ /either no changes/ ) {
            if ( $flags_and_ops{IRON_COMMIT_CHANGES} ) {
                $self->{wafliron_obj}->commit();

                # Online aggr
                my $aggr_obj = $self->{aggregate};
                $aggr_obj->modify( state => 'online' );
            } else {
                $self->{wafliron_obj}->reject();
            }
        } else {
            # Online aggr
            my $aggr_obj = $self->{aggregate};
            $aggr_obj->modify( state => 'online' );
        }
    }

    my ( $log_obj, @message_list, @msgs, $w_msgs );
    $log_obj      = $self->{wafliron_rerun_obj}->system_log_object();
    #$w_msgs       = $log_obj->get_logs( skip_cached_logs => 1 );
    @message_list = ('wafl.check.info:error');

    # Check log for matched messages
    try {
        $log_obj->check( check_for_all_presence => \@message_list );

        # Get all matched messages
        @msgs = $log_obj->get_matched_logs();
        $Log->comment( "Iron/IOC Messages on rerun - \n" . Dumper \@msgs );
        $return_iron_msgs = \@msgs;
    }
    catch NACL::Exceptions::EventCheckFailure with {
        # Don't fail if no logs messages found
    };

    if ( exists( $options{'return_iron_msgs'} ) ) {
        @{ $options{'return_iron_msgs'} } = @msgs;
    }
    $Log->exit();
} ## end sub rerun_iron

# Perform requested operations on filer
sub run_filer_ops {
    my $self = shift;
    my %options = @_;
    $Log->enter();

    my $vvol_params       = $self->{iron_cparams}{IRON_VVOL_PARAMS};
    my $flags_and_methods = $self->{iron_cparams}{IRON_FLAGS_AND_METHODS};
    my $iron_wafl_ops     = $self->{iron_cparams}{IRON_OPERATIONS};
    my %flags_and_ops     = ( %$flags_and_methods, %$iron_wafl_ops );
    my $io_tool           = $flags_and_ops{IRON_CLIENT_IO_TOOL};

    # Get aggr status
    my $aggr_obj = $self->{aggregate};

    # Get volume list
    my @cs_objects;
    try {
        @cs_objects = NACL::CS::Volume->fetch(
            command_interface => $aggr_obj->command_interface(),
            filter            => {
                 aggregate => $aggr_obj->{aggregate} },
            allow_empty       => 1
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        my $excpn_obj = shift;
        $Log->warn("Unable to find volume. Got exception:" . $excpn_obj->text());

        # Check debug vreport & fix
        $self->_fix_debug_vreport( delete_lostvol => 0 );
        $Log->exit();
        return;
    };

    # Get vvol list from dataset
    my @vol_list = $self->get_vvols();
    my ( @vol_objs, @subtests );
    foreach my $cs_obj (@cs_objects) {
        my $vol = $cs_obj->volume();
        push (@vol_objs, $cs_obj->get_task_instance())
                if grep /^$vol$/, @vol_list;
    }

    foreach my $vol_obj (@vol_objs) {
        my $vol_name = $vol_obj->volume();
        my $vol_state = 'offline';
        my $vvol_ref;
        my @qtree_list = ();

        # skip offline/restricted vols
        try {
            $vol_state = $vol_obj->state()->state();
        }
        catch NACL::Exceptions::InvalidFilterField with {
            my $excpn_obj = shift;
            $Log->warn( "Caught exception: " . $excpn_obj->text() );
        };
        next if $vol_state !~ /online/i;

        for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {
            next unless defined $self->{"VVOL_SET$i"}{$vol_name};
            $vvol_ref = $vvol_params->{"VVOL_SET$i"};

            # Qtree info
            @qtree_list = @{$self->{"VVOL_SET$i"}{$vol_name}{qtree_list}}
                   if defined $self->{"VVOL_SET$i"}{$vol_name}{qtree_list};

            last;
        }

        # Start NFS HAMMER
        if ( defined $flags_and_ops{IRON_NFS_IO_LOAD}
               && $flags_and_ops{IRON_NFS_IO_LOAD} )
        {
            $Log->comment("Start NFS traffic on $vol_name...");
            try {
                $self->start_traffic(
                    volume => $vol_name,
                    tool   => $io_tool
                );
            }
            catch NATE::BaseException with {
                my $excpn_obj = shift;
                $Log->warn(
                    "Caught exception: " . $excpn_obj->text() );
            }
            catch NACL::Exceptions::TrafficFailed with {
                my $excpn_obj = shift;
                $Log->warn(
                    "Caught exception: " . $excpn_obj->text() );
            };
        } ## end if

        # Start file and snapshot creation
        if ( defined $flags_and_ops{IRON_SNAP_CREATE_POSTIRON}
               && $flags_and_ops{IRON_SNAP_CREATE_POSTIRON} )
        {
            $Test->nlog("Create files and snapshots on $vol_name");

            try {
                $self->_create_files_and_snapshots(
                    volume_name    => $vol_name,
                    vvol_ref       => $vvol_ref,
                    file_count     => '2',
                    snapshot_count => '5',
                    file_size      => "10m",
                    postrun        => '1',
                    qtree_list     => \@qtree_list
                );
            }
            catch NATE::BaseException with {
                my $excpn_obj = shift;
                $Log->warn(
                    "Caught exception while post iron start"
                    . " create file/snapshot: " . $excpn_obj->text() );
            };
        }
    } ## end foreach

    $Log->exit();
} ## end sub run_filer_ops

# This will return an array of all volume objects
sub get_vvols {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my ( @vols, $vvol_set );
    $vvol_set = $options{vvol_set};

    for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {
        my $clone_vol_list = $self->{"VVOL_SET$i"}{cvol_list};
        my $vvol_vol_list  = $self->{"VVOL_SET$i"}{vol_list};

        if ( defined $vvol_set ) {
            if ( $vvol_set eq "VVOL_SET$i" ) {
                if ( defined $clone_vol_list ) {
                    @vols = ( @$clone_vol_list, @$vvol_vol_list );
                }
                else {
                    @vols = @$vvol_vol_list if defined $vvol_vol_list;
                }
                $i = $self->{noof_volsets};
            }
        }
        else {
            if ( defined $clone_vol_list ) {
                @vols = ( @vols, @$clone_vol_list, @$vvol_vol_list );
            }
            else {
                @vols = ( @vols, @$vvol_vol_list ) if defined $vvol_vol_list;
            }
        }
    }

    $Log->debug("DscSetup::get_vvols - @vols");
    $Log->exit();
    return @vols;
} ## end sub get_vvols

# This will return an array of space_reserved files created
sub get_space_reservedfiles {
    my $self    = shift;
    my %options = @_;
    my ( $volume, @all_files );
    $Log->enter();

    if ( defined $options{volume} ) {
        if ( ref $options{volume} ) {
            $volume = $options{volume}->volume();
        }
        else {
            $volume = $options{volume};
        }
    }

    for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {

        my @vol_list = $self->get_vvols( vvol_set => "VVOL_SET$i" );
        foreach my $vol_name (@vol_list) {
            next
              unless
                defined $self->{"VVOL_SET$i"}{$vol_name}{spaceresfile_list};
            my $files = $self->{"VVOL_SET$i"}{$vol_name}{spaceresfile_list};
            if ( defined $volume && $volume eq $vol_name ) {
                @all_files = @$files;
                $i = $self->{noof_volsets};
            }
            else {
                @all_files = ( @all_files, @$files ) unless defined $volume;
            }
        }
    }

    $Log->debug("DscSetup::get_space_reservedfiles - @all_files");
    $Log->exit();
    return @all_files;    # returns empty array of no files present
} ## end sub get_space_reservedfiles

# This will return an array of cifs enabled files created
sub get_cifs_enabled_files {
    my $self    = shift;
    my %options = @_;
    my ( $volume, @all_files );
    $Log->enter();

    if ( defined $options{volume} ) {
        if ( ref $options{volume} ) {
            $volume = $options{volume}->volume();
        }
        else {
            $volume = $options{volume};
        }
    }

    for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {

        my @vol_list = $self->get_vvols( vvol_set => "VVOL_SET$i" );

        foreach my $vol_name (@vol_list) {
            next
              unless defined $self->{"VVOL_SET$i"}{$vol_name}{cifsfile_list};
            my $files = $self->{"VVOL_SET$i"}{$vol_name}{cifsfile_list};
            if ( defined $volume && $volume eq $vol_name ) {
                @all_files = @$files;
                last;
            }
            else {
                @all_files = ( @all_files, @$files ) unless defined $volume;
            }
        }
    }

    $Log->debug("DscSetup::get_cifs_enabled_files - @all_files");
    $Log->exit();
    return @all_files;    # returns empty array of no files present
} ## end sub get_cifs_enabled_files

# This will return an array of holey files created
sub get_holeyfiles {
    my $self    = shift;
    my %options = @_;
    my ( $volume, @all_files );
    $Log->enter();

    if ( defined $options{volume} ) {
        if ( ref $options{volume} ) {
            $volume = $options{volume}->volume();
        }
        else {
            $volume = $options{volume};
        }
    }

    for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {

        my @vol_list = $self->get_vvols( vvol_set => "VVOL_SET$i" );

        foreach my $vol_name (@vol_list) {
            next
              unless defined $self->{"VVOL_SET$i"}{$vol_name}{holeyfile_list};
            my $files = $self->{"VVOL_SET$i"}{$vol_name}{holeyfile_list};
            if ( defined $volume && $volume eq $vol_name ) {
                @all_files = @$files;
                last;
            }
            else {
                @all_files = ( @all_files, @$files ) unless defined $volume;
            }
        }
    }

    $Log->debug("DscSetup::get_holeyfiles - @all_files");
    $Log->exit();
    return @all_files;    # returns empty array of no files present
} ## end sub get_holeyfiles

# This will stop traffic
sub stop_traffic {
    my $self = shift;
    $Log->enter();

    $Log->comment("Stopping traffic...");
    return unless $self->{noof_volsets};
    for ( my $i = 0; $i < $self->{noof_volsets}; $i++ ) {

        my @vol_list = $self->get_vvols( vvol_set => "VVOL_SET$i" );
        foreach my $vol (@vol_list) {

            my @traffic_href_list = ();
            @traffic_href_list
                  = @{$self->{"VVOL_SET$i"}{$vol}{traffic_list}}
                      if defined $self->{"VVOL_SET$i"}{$vol}{traffic_list};
            foreach my $traffic_href ( @traffic_href_list )
            {
                next unless defined $$traffic_href{traffic_tool};
                try {
                    if ( $$traffic_href{traffic_tool} eq 'hammer' ) {
                        $self->{node}{apiset}->hammer_abort();
                    }
                    elsif ( $$traffic_href{traffic_tool} eq 'filersio' ) {
                        $self->{node}{apiset}->filersio_stop();
                    }
                    elsif ( $$traffic_href{traffic_tool} eq 'mtask_traffic' ) {
                        $$traffic_href{traffic_obj}->stop(
                                     graceful_shutdown_timeout => '3000' );
                    }
                }
                catch NATE::BaseException with {
                    my $excpn_obj = shift;
                    $Log->comment("stop_traffic: Caught exception: " . $excpn_obj->text());
                    Tharn::sleep 30;
                }
                catch NACL::APISet::Exceptions::CommandFailedException with {
                    my $excpn_obj = shift;
                    $Log->comment("stop_traffic: Caught exception: " . $excpn_obj->text());
                };
            }
            undef $self->{"VVOL_SET$i"}{$vol}{traffic_list};
        }
    }

    # Stop lun traffic
    $Log->comment("Stopping traffic on LUN(s)...");
    foreach my $traffic_href ( @{$self->{lun_iohash}} ) {
        try {
            $$traffic_href{traffic_obj}->stop(
                         graceful_shutdown_timeout => '3000' );
        }
        catch NATE::BaseException with {
            my $excpn_obj = shift;
            $Log->comment("stop_traffic: Caught exception: " . $excpn_obj->text());
        };
    }
    undef $self->{lun_iohash};

    $Log->exit();
} ## end sub stop_traffic

# This will create holey files
sub _create_holey_file {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $volume_name = $options{volume_name};
    my $vserver     = $options{vserver};
    my $filename    = $options{filename};
    my $size        = $options{size};
    my $path        = "/vol/$volume_name/$filename";

    NACL::STask::VolumeFile->create(
        command_interface     => $self->{node},
        path                  => $path,
        size                  => '0',
        vserver               => $vserver,
        filecount             => 1 );

    $self->{node}{apiset}->testsetattr(
        'path'            => $path,
        'privilege-level' => 'test',
        'permissions'     => '777',
        'file-size'       => $size
    );

    # Sync & Flush Cache
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->wafl_flush_cache();

    $Log->exit();
} ## end sub _create_holey_file

# This will create space_reserved files
sub _create_space_reserved_file {
    my $self        = shift;
    my %options     = @_;
    $Log->enter();

    my $volume_name = $options{volume_name};
    my $vserver     = $options{vserver};
    my $filename    = $options{filename};
    my $size        = $options{size};
    my $path        = "/vol/$volume_name/$filename";

    NACL::STask::VolumeFile->create(
        command_interface     => $self->{node},
        path                  => $path,
        size                  => '0',
        vserver               => $vserver,
        filecount             => 1 );

    $self->{node}{apiset}->testsetattr(
        'path'            => $path,
        'privilege-level' => 'test',
        'file-size'       => $size * 2
    );

    $self->{node}{apiset}->testsetattr(
        'path'                    => $path,
        'privilege-level'         => 'test',
        'permissions'             => '777',
        'file-space-reservations' => 'enable'
    );

    # Sync & Flush Cache
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->wafl_flush_cache();

    $Log->exit();
} ## end sub _create_space_reserved_file

# This will create files with cifs enabled
sub _create_files_with_cifs_enabled {
    my $self        = shift;
    my %options     = @_;
    $Log->enter();

    my $volume_name = $options{volume_name};
    my $vserver     = $options{vserver};
    my $filename    = $options{filename};
    my $size        = $options{size};
    my $path        = "/vol/$volume_name/$filename";

    NACL::STask::VolumeFile->create(
        command_interface     => $self->{node},
        path                  => $path,
        size                  => '0',
        vserver               => $vserver,
        filecount             => 1 );

    $self->{node}{apiset}->testsetattr(
        'path'            => $path,
        'privilege-level' => 'test',
        'file-size'       => $size
    );

    $self->{node}{apiset}->testsetattr(
        'path'                    => $path,
        'privilege-level'         => 'test',
        'write-offset-and-length' => '4096 40960'
    );

    $self->{node}{apiset}->testsetattr(
        'path'                    => $path,
        'privilege-level'         => 'test',
        'permissions'             => '777',
        'file-space-reservations' => 'enable cifs'
    );

    # Sync & Flush Cache
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->sync();
    $self->{node}{apiset}->wafl_flush_cache();

    $Log->exit();
} ## end sub _create_files_with_cifs_enabled

# This will punch holes at specified offset
sub _punch_holes {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $path  = $options{path};
    my $fbn   = $options{fbn};
    my $count = $options{count};

    $self->{node}{apiset}->file_punchhole(
        path   => $path,
        offset => $fbn,
        length => $count
    );
    $Log->exit();
} ## end sub _punch_holes

# This will perform various ops on files
sub perform_nfs_op_on_files {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $volume = $options{volume};
    my $op     = $options{op};

    if ( defined $options{volume} ) {
        unless ( ref $options{volume} ) {

            # Get vol object
            foreach my $vol_set ( @{ $self->{volset_list} } ) {
                my @vol_list = $self->get_vvols( vvol_set => $vol_set );
                foreach my $vol (@vol_list) {
                    $volume = $self->{$vol_set}{$vol}{vol_obj_ref}
                      if ( $vol eq $options{volume} );
                }
            }
        }
    }

    my $client_obj = $self->{client};
    my $interface  = $volume->command_interface();
    my $vserver    = $volume->vserver();

    $Log->comment( "Mounting volume " . $volume->volume() );
    my $mount_obj = NACL::STask::Client::Mount->create(
        command_interface        => $client_obj,
        server_command_interface => $interface,
        volume                   => $volume
    );

    # Get the mount point
    my $mount_point = $mount_obj->mount_point();
    my $mount_path  = NACL::MTask::Traffic->get_abs_path($mount_point);
    my $volume_name = $volume->volume();

    my @space_resfiles = $self->get_space_reservedfiles( volume => $volume );
    my @holey_files = $self->get_holeyfiles( volume => $volume );

    #my @cifs_files    = $self->get_cifs_enabled_files(
    #                                           volume => $volume );

    foreach my $file ( @space_resfiles, @holey_files ) {    # @cifs_files
        $Log->comment("Performing ops($op) on $file under $volume_name");
        my $file_path = $mount_path . "/$file";
        $client_obj->execute(
            command => "$SUDO_PATH $op $file_path 2>&1 /dev/null",
            timeout => 600
        );
    }
    $Log->exit();
} ## end sub perform_nfs_op_on_files

# This will perform ops at random offsets on all
# the files in the volume
sub perform_nfs_op_on_files_at_random_offsets {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $volume = $options{volume};
    if ( defined $options{volume} ) {
        unless ( ref $options{volume} ) {

            # Get vol object
            foreach my $vol_set ( @{ $self->{volset_list} } ) {
                my @vol_list = $self->get_vvols( vvol_set => $vol_set );
                foreach my $vol (@vol_list) {
                    $volume = $self->{$vol_set}{$vol}{vol_obj_ref}
                      if ( $vol eq $options{volume} );
                }
            }
        }
    }

    my $client_obj = $self->{client};
    my $client_api = $self->{client_apiset};
    my $interface  = $volume->command_interface();
    my $vserver    = $volume->vserver();

    $Log->comment( "Mounting volume " . $volume->volume() );
    my $mount_obj = NACL::STask::Client::Mount->create(
        command_interface        => $client_obj,
        server_command_interface => $interface,
        volume                   => $volume
    );

    # Get the mount point
    my $mount_point = $mount_obj->mount_point();
    my $mount_path  = NACL::MTask::Traffic->get_abs_path($mount_point);
    my $volume_name = $volume->volume();

    my @space_resfiles = $self->get_space_reservedfiles( volume => $volume );
    my @holey_files = $self->get_holeyfiles( volume => $volume );

    #my @cifs_files    = $self->get_cifs_enabled_files(
    #                                   volume => $volume );

    foreach my $file ( @space_resfiles, @holey_files ) {# @cifs_files
        $Log->comment(
            "Performing op(random offset) on $file under vol $volume_name");
        my $seek            = int( rand(10) );
        my $file_path       = $mount_path . "/$file";
        my $filer_file_path = "/vol/$volume_name/$file";

        # Run SIO
        $Log->comment("Run SIO on $file_path");
        $client_api->sio(
            'read'               => 0,
            'rand'               => 50,
            'block_size'         => "4k",
            'start_byte'         => 0,
            'file_size'          => "100m",
            'duration_in_secs'   => 10,
            'devices_list'       => $file_path,
            'num_of_threads'     => 2,
            'connectrec-timeout' => 120
        );

        # Truncate
        $Log->comment("Truncate $filer_file_path with testsetattr");
        $self->{node}{apiset}->testsetattr(
            'path'            => $filer_file_path,
            'privilege-level' => 'test',
            'file-size'       => 0
        );

        # DD from client
        $client_api->dd(
            'input-file'                  => "/dev/zero",
            'output-file'                 => $file_path,
            'skipped-output-blocks-count' => $seek,
            'bytes-rate'                  => 512,
            'block-count'                 => 2
        );
    } ## end foreach
    $Log->exit();
} ## end sub perform_nfs_op_on_files_at_random_offsets

# This returns the compressed field from df -Sb
sub get_compression_saved_blocks {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $inum   = $options{inum};
    my $volume = $options{volume};    # Object|name
    if ( defined $options{volume} ) {
        unless ( ref $options{volume} ) {

            # Get vol object
            foreach my $vol_set ( @{ $self->{volset_list} } ) {
                my @vol_list = $self->get_vvols( vvol_set => $vol_set );
                foreach my $vol (@vol_list) {
                    $volume = $self->{$vol_set}{$vol}{vol_obj_ref}
                      if ( $vol eq $options{volume} );
                }
            }
        }
    }

    # TODO
    # Since we don't have an API for getting the compressed saved blocks
    # at inode level, for now we getting from volume level.
    $Log->comment( "Executing df -Sb for " . $volume->volume() );
    my $parsed = $self->{node}{apiset}->df(
        'block-sharing-and-compression-statistics' => 1,
        'volume'                                   => 1,
        'name'                                     => $volume->volume()
    )->get_parsed_output();
    $Log->exit();
    return $$parsed[0]->{compressed};

} ## end sub get_compression_saved_blocks

# This returns the number of share blocks
sub get_shared_blocks {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $volume = $options{volume};
    if ( defined $options{volume} ) {
        unless ( ref $options{volume} ) {

            # Get vol object
            foreach my $vol_set ( @{ $self->{volset_list} } ) {
                my @vol_list = $self->get_vvols( vvol_set => $vol_set );
                foreach my $vol (@vol_list) {
                    $volume = $self->{$vol_set}{$vol}{vol_obj_ref}
                      if ( $vol eq $options{volume} );
                }
            }
        }
    }
    my $volume_name = $volume->volume();
    my $interface   = $volume->command_interface();

    #
    $Log->comment("Executing vol efficiency stat");
    my $parsed =
      NACL::STask::VolumeEfficiency->stat( command_interface => $interface );
    $Log->exit();
    return $parsed->{same_data_blocks};

} ## end sub get_shared_blocks

# This returns inode block count
sub get_inode_block_count {
    my $self    = shift;
    my %options = @_;
    my ($aggr, $parsed);
    $Log->enter();

    my $inum      = $options{inum};
    my $volume    = $options{volume};
    my $cont_name = $options{cont_name};
    my $snapid    = defined($options{snapid}) ? $options{snapid} : 0;
    my $private   = defined($options{private}) ? $options{private} : 1;
    my $level     = defined($options{level}) ? $options{level} : 99;
    my $cont_type = defined($options{cont_type})
                      ? $options{cont_type} : 'volume'; #aggregate/volume

    if ( defined $cont_name ) {
        $volume = $cont_name if $cont_type eq 'volume';
        $aggr   = $cont_name if $cont_type eq 'aggregate';
    }

    if ( defined $volume ) {
        unless ( ref $volume ) {
            # Get vol object
            foreach my $vol_set ( @{ $self->{volset_list} } ) {
                my @vol_list = $self->get_vvols( vvol_set => $vol_set );
                foreach my $vol (@vol_list) {
                    $volume = $self->{$vol_set}{$vol}{vol_obj_ref}
                      if ( $vol eq $volume );
                }
            }
        }
    }

    if ( defined $aggr ) {
        unless ( ref $aggr ) {
            # Get aggr object
            my @aggrs = NACL::CS::Aggregate->fetch(
                command_interface => $self->{node},
                filter            => { nodes => [ $self->{node}->node() ] },
                requested_fields  => ['state'],
                allow_empty       => 1
            );
            foreach my $aggr_obj (@aggrs) {
                $aggr = $aggr_obj if $aggr_obj->aggregate() eq $aggr;
            }
        }
    }

    $inum = $private ? "priv:$inum" : $inum;
    my $waitforbufcheck =  1;
    while ( $waitforbufcheck ) {
        $waitforbufcheck = 0;
        try {
            if ( $cont_type eq 'volume' ) {
                $Log->comment( "Getting inode block count for " . $volume->volume() );
                $parsed = $self->{node}{apiset}->wafl_buftree_check(
                      'no-panic'        => 1,
                      $cont_type        => $volume->volume(),
                      'inodespec'       => $inum,
                      'snapid'          => $snapid,
                      'privilege-level' => 'test'
                  )->get_parsed_output();
            } else {
                $Log->comment( "Getting inode block count for " . $aggr->aggregate() );
                $parsed = $self->{node}{apiset}->wafl_buftree_check(
                      'no-panic'        => 1,
                      $cont_type        => $aggr->aggregate(),
                      'inodespec'       => $inum,
                      'snapid'          => $snapid,
                      'privilege-level' => 'test'
                  )->get_parsed_output();
            }
        }
        catch NACL::APISet::Exceptions::TimeoutException with {
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
            $waitforbufcheck = 1;
            Tharn::sleep 30;
        };
    }

    $Log->exit();
    if ( $$parsed[0]->{buftree_info}[0]->{level} eq $level ) {
        return $$parsed[0]->{buftree_info}[0]->{total};
    } elsif ( $$parsed[0]->{buftree_info}[1]->{level} eq $level ) {
        return $$parsed[0]->{buftree_info}[1]->{total};
    } elsif ( $$parsed[0]->{buftree_info}[2]->{level} eq $level ) {
        return $$parsed[0]->{buftree_info}[2]->{total};
    } else {
        return $$parsed[0]->{total}[0]->{total};
    }
} ## end sub get_inode_block_count

# This returns non-hole fbns
sub get_non_hole_fbns {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $start_fbn = defined($options{'start_fbn'}) ? $options{'start_fbn'} : 0;
    my $end_fbn   = defined($options{'end_fbn'}) ? $options{'end_fbn'} : 10;
    my $cont_type = defined($options{'cont_type'}) ? $options{'cont_type'} : 'volume';
    my $level     = defined($options{'level'}) ? $options{'level'} : 0;
    my $private   = defined($options{'private'}) ? $options{'private'} : 1;
    my $snapid    = defined($options{'snapid'}) ? $options{'snapid'} : 0;
    my $count     = defined($options{'count'}) ? $options{'count'} : 1;
    my $cont_name = $options{'cont_name'}; # must
    my $inode_no  = $options{'inode_no'}; # must
    my @fbn_list  = ($options{'start_fbn'}..$options{'end_fbn'});

    my $fbn_count = 0;
    my @fbn_op_list;
    foreach my $fbn ( @fbn_list ) {
        try {
            my $pvbn_no = $self->{node}{apiset}->dumpblock_info(
                                         $cont_type  => $cont_name,
                                         fileid      => $inode_no,
                                         level       => $level,
                                         private     => $private,
                                         snapid      => $snapid,
                                         blockno     => $fbn,
                           )->get_parsed_output()->[0]->{'pvbn'};

            if ( $pvbn_no ) {
                $Log->debug("===== $fbn is not a hole =====");
                $Log->debug("For L0 FBN: $fbn, PVBN: $pvbn_no");
                push(@fbn_op_list, $fbn);
                $fbn_count++;
                if ( $count == 1 ) {
                    last if ( $fbn_count > 10 );
                } else {
                    last if ( $fbn_count > $count );
                }
            }
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            $Log->debug( "Error is: " . $exception_object->text() );
            $Log->debug("===== $fbn is a hole =====");
        };
    }

    $Log->exit();
    if ( $count eq 1 ) {
        return $fbn_op_list[int rand($#fbn_op_list)];
    } else {
        return ( wantarray() ? @fbn_op_list : "@fbn_op_list" );
    }
} ## end sub get_non_hole_fbns

# Get node stats
sub get_stats {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    # Default retrun stats of volume
    my $fields = $options{fields};

    my (%ret_hash);
    if ( defined $fields ) {
        foreach my $field_name (@$fields) {
            my @parsed = $self->{node}{apiset}->stats_show(
                'privilege-level' => 'test',
                'obj_inst_cnt'    => $field_name,
            )->get_parsed_output();
            my @out = $parsed[0];
            $ret_hash{$field_name} = $out[0][0]->{value};
        }
    }

    $Log->debug( "get_stats::stats - " . Dumper \%ret_hash );
    $Log->exit();
    return %ret_hash;
} ## end sub get_stats

# Get VVBN, BNO & INDEX of file for wreck
# TODO: this assumes level 1
sub get_vvbn_bno_index {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $file     = $options{file};
    my $response = $self->{node}{apiset}->wafl_showvbns(
        'path' => $file,
        'fbn'  => '0'
    );

    #my $pointers = $self->{aggregate_type} =~ /32-bit/i ? 1024 : 512;
    my $pointers = $self->{aggregate_type} =~ /32-bit/i ? 1024 : 510;
    my $vvbn     = $response->get_parsed_output()->[0]->{vvbn};
    my $bno      = ( int( $vvbn / $pointers ) ) * $pointers;
    my $index    = $vvbn % $pointers;

    $Log->exit();
    return ( $vvbn, $bno, $index );
} ## end sub get_vvbn_bno_index

# Set flags
sub misc_setflag {
    my $self    = shift;
    my %options = @_;
    $self->set_flags(@_);
}

# Set flags
sub set_flags {
    my $self    = shift;
    my %options = @_;
    $Log->enter();

    my $flags      = delete $options{wafl_flags};
    my $apiset_obj = $self->{node}{apiset};

    my %available_flags =
      %{ $apiset_obj->printflag( 'connectrec-log_level' => 'trace' )
          ->get_parsed_output()->[0] };
    foreach my $flag ( keys %$flags ) {
        if ( !exists $available_flags{$flag} ) {
            $Log->warn("Can't set the $flag, as this flag is unavailable");
            next;
        }
        $apiset_obj->setflag(
            flag  => $flag,
            value => $flags->{$flag},
        );
    }
    $Log->exit();
} ## end sub set_flags

# Reconnect the session if filer reboots
# We are reconnecting the APIset in while loop for 300 seconds.
sub _check_connection {
    my $self = shift;
    $Log->enter();

    my $apiset_obj = $self->{node}->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'Systemshell'
    );

    if ( !$apiset_obj->api_is_session_alive() ) {
        $Log->comment("Apiset connection is down. reconnecting...");
        my $transit_obj =
          NACL::Transit->new( name => $self->{command_interface}->name() );
        $transit_obj->change_state( to => 'CLI' );
        my $time = time() + 300;
        $self->{command_interface}->refresh_command_interface();
        while (1) {
            try {
                if ( $apiset_obj->api_is_session_alive() ) {
                    last;
                }
                else {
                    $apiset_obj->api_reconnect_session();
                }
            } ## end try
            catch NATE::BaseException with {
                $Log->debug("Retrying connection...");
            };
            if ( time() >= $time && !$apiset_obj->api_is_session_alive() ) {
                $Log->exit();
                NATE::BaseException->throw(
                    "Unable to reconnect to filer after connection loss");
            }
        } ## end while
    } ## end if

    $Log->exit();
} ## end sub _check_connection

# Fix VLDB inconsistencies
sub _fix_debug_vreport {
    my $self = shift;
    my %options = @_;
    my $delete_lostvol = delete $options{delete_lostvol};
    $Log->enter();

    # Check debug vreport & fix
    my %lost_vol_data;
    my $lost_vol_case = 0;
    my $cmd_interface = $self->{node};
    my $ci_is_cmode   = $cmd_interface->is_cmode();

    if ( $ci_is_cmode ) {
        $Log->comment("Checking for debug vreport...");
        try {
            my @DebugVreport_states
              = NACL::CS::DebugVreport->fetch( command_interface => $cmd_interface );
            foreach my $DebugVreport_obj ( @DebugVreport_states ) {
                $lost_vol_case = 1
                    if $DebugVreport_obj->object() =~ /lost_volume/;
                next if $DebugVreport_obj->type() eq 'volume';

                my $type = $DebugVreport_obj->type();
                my $DebugVreport = NACL::C::DebugVreport->new(
                               command_interface => $cmd_interface,
                               object            => $DebugVreport_obj->object(),
                               type              => $DebugVreport_obj->type() );

                $DebugVreport->fix(); # fix
            }

            if ( $lost_vol_case ) {
                foreach my $DebugVreport_obj ( @DebugVreport_states ) {
                    next unless $DebugVreport_obj->type() eq 'volume';
                    my $obj_name = $DebugVreport_obj->object();
                    if ( $obj_name =~ /.*:(lost_volume\d+)/ ) {
                        my $attrs = $DebugVreport_obj->attributes();
                        my ($uuid, $aggr_name, $aggr_uuid);
                        foreach my $var ( @$attrs ) {
                            $uuid      = $1 if $var =~ /\"UUID: (.*)\"/;
                            $aggr_name = $1 if $var =~ /\"Aggregate Name: (.*)\"/;
                            $aggr_uuid = $1 if $var =~ /\"Aggregate UUID: (.*)\"/;
                        }
                        $lost_vol_data{uuid}      = $uuid;
                        $lost_vol_data{aggr_name} = $aggr_name;
                        $lost_vol_data{aggr_uuid} = $aggr_uuid;
                    } else {
                        $obj_name =~ /(.*):(.*)/;
                        my $lost_vol_name = $2;
                        my $attrs = $DebugVreport_obj->attributes();
                        my ($dsid, $msid, $aggr_name, $aggr_uuid, $vs_uuid);
                        foreach my $var ( @$attrs ) {
                            if ( $var =~ /\"Volume DSID:(\d+) MSID:(\d+)\"/ ) {
                                $dsid  = $1; $msid = $2;
                            }
                            $aggr_name = $1 if $var =~ /\"Aggregate Name: (.*)\"/;
                            $aggr_uuid = $1 if $var =~ /\"Aggregate UUID: (.*)\"/;
                            $vs_uuid   = $1 if $var =~ /\"Vserver UUID: (.*)\"/;
                        }
                        $lost_vol_data{dsid}      = sprintf("0x%016x", $dsid);
                        $lost_vol_data{msid}      = sprintf("0x%016x", $msid);
                        $lost_vol_data{aggr_name} = $aggr_name;
                        $lost_vol_data{aggr_uuid} = $aggr_uuid;
                        $lost_vol_data{vol_name}  = $lost_vol_name;
                        $lost_vol_data{vserver_uuid} = $vs_uuid;
                    }

                    # Fix lost volume
                    if ( defined $lost_vol_data{uuid}
                        && defined $lost_vol_data{dsid}
                          && defined $lost_vol_data{msid} ) {

                        # Get Systemshell api
                        my $apiset_sys_obj = $self->{node}->apiset(
                                                    category  => 'Node',
                                                    interface => 'CLI',
                                                    set       => 'Systemshell' );
                        $Log->comment("Lost vol data - ". Dumper \%lost_vol_data);

                        $apiset_sys_obj->execute_raw_command(
                              command => "zsmcli d-volume-restrict id=$lost_vol_data{uuid}");
                        $apiset_sys_obj->execute_raw_command(
                              command => "zsmcli d-volume-set-info "
                                  . "id=$lost_vol_data{uuid} desired-attrs=name "
                                  . "volume-attrs=\"[d-volume-info=[name=$lost_vol_data{vol_name}]]\"");
                        $apiset_sys_obj->execute_raw_command(
                              command => "zsmcli d-volume-set-info "
                                  . "id=$lost_vol_data{uuid} desired-attrs=dsid "
                                  . "volume-attrs=\"[d-volume-info=[dsid=$lost_vol_data{dsid}]]\"");
                        $apiset_sys_obj->execute_raw_command(
                              command => "zsmcli d-volume-set-info "
                                  . "id=$lost_vol_data{uuid} desired-attrs=msid "
                                  . "volume-attrs=\"[d-volume-info=[msid=$lost_vol_data{msid}]]\"");
                        $apiset_sys_obj->execute_raw_command(
                              command => "zsmcli d-volume-set-info "
                                  . "id=$lost_vol_data{uuid} desired-attrs=owning-vserver-uuid "
                                  . "volume-attrs=\"[d-volume-info=["
                                  . "owning-vserver-uuid=$lost_vol_data{vserver_uuid}]]\"");

                        # Get vol info
                        my $vol_cs_obj = NACL::CS::Volume->fetch(
                                      command_interface => $cmd_interface,
                                      filter            => {
                                           volume  => $lost_vol_data{vol_name}
                                      });
                        $Log->comment("Lost vol vol_cs_obj - ". Dumper $vol_cs_obj);

                        if ( $delete_lostvol ) {
                            $Test->nlog("Deleting volume $lost_vol_data{vol_name}");
                            try {
                                NACL::STask::Volume->purge(
                                  command_interface => $cmd_interface,
                                  volume            => $lost_vol_data{vol_name},
                                  vserver           => $vol_cs_obj->vserver(),
                                  force             => 'true',
                                  allow_extra       => 1 );
                            }
                            catch NACL::C::Exceptions::Volume::DoesNotExist with {
                                my $excpn_obj = shift;
                                $Log->debug( "Caught exception: " . $excpn_obj->text() );
                            };
                        }
                    }
                }

                @DebugVreport_states = NACL::CS::DebugVreport->fetch(
                               command_interface => $cmd_interface );

                foreach my $DebugVreport_obj ( @DebugVreport_states ) {
                    my $type = $DebugVreport_obj->type();
                    my $DebugVreport = NACL::C::DebugVreport->new(
                               command_interface => $cmd_interface,
                               object            => $DebugVreport_obj->object(),
                               type              => $DebugVreport_obj->type() );

                    $DebugVreport->fix(); # fix
                }
            }
        }
        catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        }
        catch NACL::Exceptions::NoElementsFound with {
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        };
        catch NACL::Exceptions::NoElementsFound with {
            my $excpn_obj = shift;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        };
    } else {
        $Log->comment("Checking for lost volume(s)...");
    }

    $Log->exit();
} ## end sub _fix_debug_vreport

sub _get_iscsi_vserver {
    my $self = shift;
    my ($ivserver, $vserver);
    $Log->enter();

    try {
        $ivserver = NACL::C::VserverIscsi->find(
                         command_interface => $self->{node} );
    } catch NATE::BaseException with {
        # Not sure if this is preferred. Getting an existing vserver
        # and setting it for iscsi usage.
        my $e = shift;
        $Log->comment('Unable to find a ISCSI enabled Vserver');
        $Log->comment('iSCSI config failed with the error: ' . $e->text());
        try {
            $Test->nlog('Attempt to find and configure iSCSI vserver');
            $vserver = NACL::STask::Vserver->find_data_vservers(
                command_interface => $self->{node} );
            my $vsvr_name = $vserver->vserver();

            # Make sure we have a license
            $Test->nlog('Add or ck iSCSI is on the filer');
            NACL::STask::SystemLicense->add(
                command_interface    => $self->{node},
                feature              => 'iSCSI',
                nacltask_verify      => 1,
                nacltask_if_licensed => 'pass',);

            $ivserver = NACL::C::VserverIscsi->create(
                command_interface => $self->{node},
                vserver           => $vsvr_name,
                'target-alias'    => $vsvr_name,
                'status-admin'    => 'up'
            );
        } catch NATE::BaseException with {
            # Failed to create the iscsi vserver.
            undef($vserver);
        };
    };

    $Log->exit();
    return $ivserver;
} ## end sub _get_iscsi_vserver

sub _create_luns {
    my ($self, %opts)  = @_;
    my $initiator_name = $opts{initiator_name};
    my $igroup         = 'nacl_mtask_ig';
    my $ci_is_cmode    = $self->{node}->is_cmode();
    $Log->enter();

    my %lun_input = %{$self->{lun_input}};
    foreach my $volume (keys %lun_input) {
        my $vals = $lun_input{$volume};
        foreach my $qtree (keys %$vals) {
            my $val = ${$vals}{$qtree};

            my $vol_obj = $lun_input{$volume}{$qtree}{vol_obj};
            my (@lun_name, @lun_size, @lun_path, %lun_input, $lun);

            $Test->nlog("Creating LUN $$val{lun_name} on volume $volume");
            push (@lun_name, $$val{lun_name});
            push (@lun_size, $$val{lun_size});
            if ( $qtree eq 'none' ) {
                $lun = "/vol/$volume";
            } else {
                $lun = "/vol/$volume/$qtree";
            }
            push (@lun_path, $lun);

            $lun_input{command_interface}     = $self->{node};
            $lun_input{igroup}                = $igroup;
            $lun_input{data_protocol}         = 'iscsi';
            $lun_input{ostype}                = 'linux';
            $lun_input{protocol}              = 'mixed';
            $lun_input{comment}               = 'dsc_lun_testing';
            $lun_input{qtree}                 = $qtree unless $qtree eq 'none';
            $lun_input{nacltask_lun}          = \@lun_name;
            $lun_input{nacltask_size}         = \@lun_size;
            $lun_input{nacltask_path}         = \@lun_path unless $ci_is_cmode;
            $lun_input{initiator}             = [ $initiator_name ];
            $lun_input{volume}                = $vol_obj->volume();
            $lun_input{vserver}               = $vol_obj->vserver();
            $lun_input{nacltask_if_lif_exists}        = 'purge';
            $lun_input{nacltask_if_lun_exists}        = 'purge';
            $lun_input{nacltask_if_igroup_exists}     = 'reuse';
            $lun_input{nacltask_if_lun_mapped_igroup} = 'remap';

            my $lunsetup_object = NACL::MTask::LunSetup->setup(
                       command_interface     => $self->{node},
                       %lun_input );

            # Store lun obj
            push ( @{$self->{lunsetup_obj}},  $lunsetup_object );
            push ( @{$self->{ulun_obj_list}}, $lunsetup_object->{all_objects}->{lun} );
        }
    }

    $Log->exit();
} ## end sub _create_luns

sub _create_traffic_file {
    my ($self, %opts) = @_;
    my $fullname = $opts{file};
    my $size     = $opts{size};
    $Log->enter();

    # For read-write traffic, use NACL::C::Client::Remote::File
    # to remotely create and write to the file.
    require NACL::C::Client::Remote::File;

    my $client = $self->{client};
    my ($total_units, $one_unit);

    if ($size =~ /^(\d+)([kmg])$/) {
        my ($num, $unit) = ($1, lc $2);
        if ($unit eq 'g') {
            # I don't want to create any 1-gigabyte scalars, so just bump
            # up $total_units by a factor and use a megabyte scalar.
            $total_units = $num * 1024;
            $unit = 'm';
        } else {
            $total_units = $num;
        }
        $one_unit = chr(0) x $FILE_SIZES{$unit};
    } else {
        NATE::BaseException->throw("Value for 'size' ($size) is not valid");
    }

    # Get the remote file handle object
    my $fh_obj = NACL::C::Client::Remote::File->open(
                               command_interface => $client,
                               'privilege-level' => 'root',
                               file_name         => $fullname,
                               arguments         => {
                                       access_mode => 'write',
                                       open_flags  => 'create-new' });

    # Write the "$one_unit" at the very end of the file. This will
    # create the rest of the file by extending the file pointer out to
    # the end.
    $fh_obj->write(
                content => $one_unit,
                offset  => ($total_units - 1) * length($one_unit)
            );
    #$fh_obj->close;

    $Log->exit();
    return;

} ## end sub _create_traffic_file

sub _run_io {
    my ($self, %opts) = @_;
    my $file_name  = $opts{filename};
    my $file_size  = $opts{filesize};
    my $mount_path = $opts{mountpath};
    my $client_obj = $opts{clientobj};
    my $run_time   = $opts{runtime};
    my $fill       = $opts{fill};
    my $mode       = $opts{mode};
    my $io_tool    = defined($opts{iotool}) ? $opts{iotool} : 'sio';
    my $traffic_obj;
    $Log->enter();

    # Run io
    $Log->comment("Creating file $file_name");
    try {
        if ( $io_tool =~ /sio/i ) {
            # SIO traffic
            $Log->comment('NFS traffic tool : sio');
            # This will write out a sparse file for read-write
            $self->_create_traffic_file(
                          file  => $file_name,
                          size  => $file_size );

            # Initialize sio default arguments
            my %sio_default_args = (
                                    rand              => '60',
                                    block_size        => '4k',
                                    start_byte        => 0,
                                    verify            => 1,
                                    read              => 0,
                                    num_of_threads    => 4,
                                    execute_remotely  => 1,
                                    'privilege-level' => 'root',
                                    file_size         => $file_size,
                                    duration_in_secs  => $run_time );

            # Create sio nfs traffic object
            $Log->debug('traffic tool inputs: sio_default_args - '
                                         . Dumper \%sio_default_args);
            $Log->debug('traffic tool inputs: client_obj - '
                                         . Dumper $client_obj);

            $traffic_obj = NACL::MTask::Traffic->new (
                                     command_interface => $client_obj,
                                     paths             => $file_name );
            $traffic_obj->start(tool              => 'sio',
                                stop_gracefully   => 1,
                                %sio_default_args );
        }
        else {
            # HAMMER traffic
            $Log->comment('NFS traffic tool : hammer');
            $traffic_obj = NACL::MTask::Traffic->start(
                                command_interface => $client_obj,
                                paths             => $mount_path,
                                tool              => 'hammer',
                                'privilege-level' => 'root',
                                nofilercore       => 1,
                                #execute_remotely  => 1,
                                fill              => $fill,
                                mode              => $mode );
        }
    }
    catch NACL::Exceptions::TrafficFailed with {
        my $exception_obj = shift;
        $Log->comment('Failed to start traffic with '
                                   . $exception_obj->text());
    };

    $Log->exit();
    return $traffic_obj;

} ## end sub _run_io

sub _delete_snapshots {
    my ($self, %opts) = @_;
    my $command_interface = $opts{command_interface};
    my $aggregate   = $opts{aggregate};
    my $volume_name = defined($opts{volume}) ? $opts{volume} : 'all';
    $command_interface->apiset
                      ->set_timeout( 'connectrec-timeout' => 300 );

    $Log->enter();

    # Get volume list
    my @cs_objects;
    try {
        @cs_objects = NACL::CS::Volume->fetch(
                command_interface => $command_interface,
                filter            => {aggregate => $aggregate});
    }
    catch NACL::Exceptions::NoElementsFound with {
            my $excpn_obj = shift;
            $Log->warn("Unable to find volume. "
              . "Got exception:" . $excpn_obj->text());
    };

    # Delete all aggr snapshots
    NACL::C::AggregateSnapshot->delete(
                    command_interface => $command_interface,
                    aggregate         => $aggregate,
                    snapshot          => '*' );

    foreach my $cs_obj (@cs_objects) {
        my $vol_obj = $cs_obj->get_task_instance();
        next if ($volume_name ne 'all'
              && $volume_name ne $vol_obj->volume());

        # Vol snapshot
        my @snap_objects = NACL::CS::VolumeSnapshot->fetch(
                    command_interface => $command_interface,
                    filter            => {
                         vserver => $vol_obj->vserver(),
                         volume  => $vol_obj->volume(),
                         'is-7-mode' => 'false'
                    },
                    allow_empty       => 1 );

        # Remove snapshot
        foreach my $snap_obj ( @snap_objects ) {
            next if $snap_obj->snapshot() =~ /^clone_|^snapmirror/;
            NACL::STask::VolumeSnapshot->purge(
                        command_interface => $command_interface,
                        volume            => $vol_obj->volume(),
                        vserver           => $vol_obj->vserver(),
                        snapshot          => $snap_obj->snapshot(),
                        nacltask_if_purged         => 'pass',
                        nacltask_wait              => 1 );
        }
    }

    $Log->exit();

} ## end sub _delete_snapshots

sub _delete_vol {
    my ($self, %opts) = @_;
    my $volume_name    = $opts{volume};
    my $volume_obj_ref = $opts{vol_obj_ref};
    my $ci_is_cmode    = $self->{node}->is_cmode();

    $Log->enter();

    my $counter = 10;
    while ( $counter ) {
        try {
            $volume_obj_ref->purge( 'method-timeout' => '300' );
            $counter = 0;
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );

            # see what is holding vol ref
            my $cmd = "wafl vref status -V $volume_name";
            $cmd = "run local; wafl vref status -V $volume_name" if $ci_is_cmode;
            $self->{node}{apiset}->execute_raw_command( 'command' => $cmd );

            # sleep so that any pending vol ref gets released
            $Log->comment( "Sleeping for 5 sec before retrying volume delete" );
            Tharn::sleep(5) ;
            $counter--;
        }
        catch NACL::APISet::Exceptions::TimeoutException with {
            Tharn::sleep(30) ;
            $counter--;
        }
        catch NATE::Exceptions::Argument with {
            my $excpn_obj = shift;
            $counter = 0;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
        }
        catch NACL::C::Exceptions::Volume::DoesNotExist with {
            my $excpn_obj = shift;
            $counter = 0;
            $Log->debug( "Caught exception: " . $excpn_obj->text() );
        };
    }

    $Log->exit();

} ## end sub _delete_vol

sub _multi_type_corruption {
    my ($self, %opts) = @_;
    my $client_obj    = $self->{client};
    my $client_api    = $self->{client_apiset};

    my $vvol_params   = $self->{iron_cparams}{IRON_VVOL_PARAMS};
    my $set_pseudobad = $self->{iron_cparams}{IRON_FLAGS_AND_METHODS}{IRON_PSEUDOBAD};
    $Log->enter();

    my $vol_set_count = $vvol_params->{IRON_NUMOF_VVOL_SETS};
    for ( my $i = 0; $i < $vol_set_count; $i++ ) {
        my @vol_list = $self->get_vvols(vvol_set => "VVOL_SET$i");

        my $vol_count = 0;
        $Log->comment("Inject different corruptions to diff volumes");
        foreach my $vol_name (@vol_list) {
            my $vol_obj
                = $self->{"VVOL_SET$i"}{$vol_name}{vol_obj_ref};
            my $vol_aggr  = $self->{"VVOL_SET$i"}{$vol_name}{vol_aggr};
            my $interface = $vol_obj->command_interface();
            my $vserver   = $vol_obj->vserver();

            # mkdir l1_dir_2
            try {
                $vol_obj->directory_create(
                          dirname        => 'l1_dir_2',
                          perm           => '0777' );
            }
            catch NACL::C::Exceptions::VolumeDirectory::AlreadyExists with {
                my $excpn_obj = shift;
            };

            # mkdir l1_dir_1/l2_dir_1
            try {
                $vol_obj->directory_create(
                          dirname => 'l1_dir_1',
                          perm    => '0777' );
            }
            catch NACL::C::Exceptions::VolumeDirectory::AlreadyExists with {
                my $excpn_obj = shift;
            };
            try {
                $vol_obj->directory_create(
                          dirname => 'l1_dir_1/l2_dir_1',
                          perm    => '0777' );
            }
            catch NACL::C::Exceptions::VolumeDirectory::AlreadyExists with {
                my $excpn_obj = shift;
            };

            # mkfile l1_dir_1l/l2_dir_1/file, l1_dir_2/file
            my $file_path = "/vol/$vol_name/l1_dir_1/l2_dir_1/file";
            NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 1 );

            $file_path = "/vol/$vol_name/l1_dir_2/file";
            NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 1 );

            $file_path = "/vol/$vol_name/ind_trailer_fbn";
            NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 1 );

            if ( $vol_count < 3  ) {
                # dir_fileid inum=path:/l1_dir_1/l2_dir_1 dirent=. inum2=path:/l1_dir_2/file
                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => [
                        {
                            command => 'dir_fileid',
                            inum    => 'path:/l1_dir_1/l2_dir_1',
                            inum2   => "path:/l1_dir_2/file",
                            dirent  => '.',
                        }],
                    pseudo_bad       => $set_pseudobad,
                    'method-timeout' => '300' );

            } elsif ( $vol_count < 6  ) {
                # dir_gen inum=path:/l1_dir/l2_dir dirent=.. gen=12345678
                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => [
                        {
                            command => 'dir_gen',
                            inum    => 'path:/l1_dir_1/l2_dir_1',
                            gen     => '12345678',
                            dirent  => '..',
                        }],
                    pseudo_bad       => $set_pseudobad,
                    'method-timeout' => '300' );

            } elsif ( $vol_count < 9  ) {
                # ind_fbn inum=path:/ind_trailer_fbn level=1 bno=10 new_bno=0xffffffff
                my $bno = int rand(20);
                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => [
                        {
                            command => 'ind_fbn',
                            inum    => "path:/ind_trailer_fbn",
                            level   => 1,
                            bno     => $bno,
                            new_bno => '0xffffffff',
                        }],
                    pseudo_bad       => $set_pseudobad,
                    'method-timeout' => '300' );

            } elsif ( $vol_count < 12  ) {
                $file_path = "/vol/$vol_name/tfile";
                NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 5 );

                # inode_link inum=path:/$file link_cnt=$rand
                my @inj_cmd_list;
                for ( my $f = 0; $f < 5; $f++ ) {
                    my %inj_cmd;
                    $inj_cmd{command}  = 'inode_link';
                    $inj_cmd{inum}     = "path:/tfile.$f",
                    $inj_cmd{link_cnt} = '0xffffffff';
                    push(@inj_cmd_list, \%inj_cmd);
                }

                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => \@inj_cmd_list,
                    pseudo_bad               => $set_pseudobad,
                    'method-timeout'         => '300' );
            } elsif ( $vol_count < 15  ) {
                $file_path = "/vol/$vol_name/tfile";
                NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 5 );

                # inode_tid inum=path:/file.0 tid=400
                my @inj_cmd_list;
                for ( my $f = 0; $f < 5; $f++ ) {
                    my %inj_cmd;
                    my $tid = 400 + $f;
                    $inj_cmd{command} = 'inode_tid';
                    $inj_cmd{inum}    = "path:/tfile.$f",
                    $inj_cmd{tid}     = $tid,
                    push(@inj_cmd_list, \%inj_cmd);
                }

                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => \@inj_cmd_list,
                    pseudo_bad               => $set_pseudobad,
                    'method-timeout'         => '300' );
            } elsif ( $vol_count < 18  ) {
                $file_path = "/vol/$vol_name/tfile";
                NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 2 );

                # inode_xinode inum=path:/file.0 inum2=path:/file.1
                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => [
                        {
                            command => 'inode_xinode',
                            inum    => "path:/tfile.0",
                            inum2   => "path:/tfile.1",
                        }],
                    pseudo_bad       => $set_pseudobad,
                    'method-timeout' => '300' );
            } elsif ( $vol_count < 21  ) {
                $file_path = "/vol/$vol_name/tfile";
                NACL::STask::VolumeFile->create(
                             command_interface => $interface,
                             path              => $file_path,
                             size              => '100m',
                             vserver           => $vserver,
                             filecount         => 10 );

                # inode_blk_cnt inum=path:/$file count=$val
                my @inj_cmd_list;
                for ( my $f = 0; $f < 10; $f++ ) {
                    my %inj_cmd;
                    my $rand = int( rand(100) );
                    $inj_cmd{command} = 'inode_blk_cnt';
                    $inj_cmd{inum}    = "path:/tfile.$f",
                    $inj_cmd{count}   = $rand,
                    push(@inj_cmd_list, \%inj_cmd);
                }

                NACL::STask::Wreck->run(
                    command_interface        => $self->{node},
                    nacltask_volumes         => [$vol_name],
                    nacltask_inject_commands => \@inj_cmd_list,
                    pseudo_bad               => $set_pseudobad,
                    'method-timeout'         => '300' );
            }

            $vol_count++;
        }
    }

    $Log->exit();

} ## end sub _multi_type_corruption

sub _lun_client_config {
    my ($self, %opts) = @_;
    my $lsobject    = $opts{lunsetup_obj};
    # Work with unix client
    my $client_obj  = $self->{client};
    my $client_ci   = $self->{client_apiset};
    my $ci_is_cmode = $self->{node}->is_cmode();
    $Log->enter();

    # Re-start iscsi service
    try {
        $client_ci->service_iscsi_restart();
    }
    catch NACL::APISet::Exceptions::TimeoutException with {
        Tharn::sleep 30;
    };

    # Discover lun
    my ($discovery, $lif_state, $ivserver, $iaddress, $target_name);
    if ( $ci_is_cmode ) {
        $lif_state = $lsobject->{all_objects}->{lif}->state();
        $iaddress  = $lif_state->address();
        $discovery = $client_ci->iscsiadm(
                                type      => 'sendtargets',
                                mode      => '1',
                                discovery => '1',
                                'ip-port' => $iaddress );
        $ivserver    = $self->_get_iscsi_vserver();
        $target_name = $ivserver->state()->{nodename};
    } else {
        my $found_target = 0;
        my @output_7m    = NACL::CS::Ifconfig->fetch(
                                  command_interface => $self->{node} );
        foreach my $output_7m ( @output_7m ) {
            my $ipv4 = ${$output_7m->ipv4()}[0];
            if ( defined $ipv4 ) {
                try {
                    $discovery = $client_ci->iscsiadm(
                                type      => 'sendtargets',
                                mode      => '1',
                                discovery => '1',
                                'ip-port' => $ipv4 );

                    $target_name  = $self->{node}{node};
                    $found_target = 1;
                }
                catch NACL::APISet::Exceptions::TimeoutException with {
                    my $excpn_obj = shift;
                    $Log->comment( "Caught exception: " . $excpn_obj->text() );
                    $found_target = 0;
                }
                catch NACL::APISet::Exceptions::CommandFailedException with {
                    my $excpn_obj = shift;
                    $Log->comment( "Caught exception: " . $excpn_obj->text() );
                    $found_target = 0;
                };

                last if $found_target;
            }
        } # end of foreach
    } # end of else

    unless ( defined $discovery ) {
        $Test->nlog("Unable to discover the filer address "
                       . "on Client machine. Ignoring lun setup");
        return;
    }

    my $found = 0;
    my $parsed_output = $discovery->get_parsed_output();
    foreach my $output ( @{$parsed_output} ) {
        $target_name = $$output{iqn};
        $iaddress    = $$output{ipaddress};

        # Logout and login to discover the luns
        try {
            $client_ci->iscsiadm(
                                mode          => '1',
                                node          => '1',
                                'ip-port'     => $iaddress,
                                'target-name' => $target_name,
                                logout        => '1' );

            $client_ci->iscsiadm(
                                mode          => '1',
                                node          => '1',
                                'ip-port'     => $iaddress,
                                'target-name' => $target_name,
                                login         => '1' );

            $self->{client}{lun_target}  = $target_name;
            $self->{client}{lun_address} = $iaddress;
            $found = 1;
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $excpn_obj = shift;
            $Log->comment( "Caught exception: " . $excpn_obj->text() );
            $found = 0;
        };

        last if $found;
    }

    NATE::BaseException->throw("Unable to discover "
                . "LUN(s) on Client machine") unless $found;

    # Wait for the luns to be visible
    Tharn::sleep 30;

    # LUN list on client
    foreach my $lun_obj_ref ( @{$self->{ulun_obj_list}} ) {
        foreach my $lun_obj ( @$lun_obj_ref ) {
        my $lun_path   = $lun_obj->path();
        my $device_obj = NACL::CS::Client::Sanlun->fetch(
                 command_interface => $client_obj,
                 filter            => { lun_pathname => $lun_path });
        my $lun_device = $device_obj->device_filename();

        # Create new file system on LUN
        try {
            NACL::C::Client::Mkfs->mkfs(
                    command_interface => $client_obj,
                    verbose           => 1,
                    device            => $lun_device );
            }
            catch NACL::APISet::Exceptions::InvalidParamValueException with {
                my $excpn_obj = shift;
                $Log->comment( "Caught exception: " . $excpn_obj->text() );
                next;
            };

            # Mount LUN to client
            $Log->comment("Mount LUN to client...");
            my $mount = NACL::STask::Client::Mount->create(
                      command_interface => $client_obj,
                      path              => $lun_device,
                      server            => $self->{client}{lun_address},
                      type              => 'device',
                      chmod             => '777',
                      fstype            => 'ext2' );

            push (@{$self->{lun_mountobj}}, $mount);
            push (@{$self->{lun_device}}, $lun_device);
        }
    }

} ## end sub _lun_client_config

=begin COMMENTED_OUT

1;
