# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/IronUtil/MPT/MptTestUtil.pm#5 $
#
# $Change: 2244423 $
#
# $DateTime: 2013/11/21 23:02:16 $
#
# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.
#
#
## NAME
#      MptTestUtil.pm - provide interfaces required for MPT Test Automation
#
## DESCRIPTION
#       This perl module provide test interfaces, which are common
#       across multiple MPT Tests and help us to avoid code duplication.
#       All interfaces make use of single hash object ( say test_data),
#       passed by the script to library interfaces. This object consists of
#       all required parameters like various task/component/apiset objects,
#       timeout, flag hash etc, which are used either to perfrom an action or
#       used in actions.
#
## AUTHOR
#       tarung@netapp.com
#
##@pod here
##
##

=pod

=head1 NAME

    MptTestUtil.pm -  provide test interfaces required for MPT Test Automation

=head1 SYNOPSIS

    use IronUtil::MPT::MptTestUtil qw(:all);
    or
    use IronUtil::MPT::MptTestUtil qw(:mpt :constants);

    my $ret = fix_by_iron_verify_by_ioc( \%Test_data );

    my $ret = fix_by_ioc_verify_by_iron( \%Test_data );

    my $ret = perform_aggregate_cleanup( \%Test_data );

    my $ret = get_set_wafl_scan_speed( \%Test_data );

    my $ret = does_user_file_exist( \%Test_data );

    my $ret = set_mpt_test_flags( \%Test_data );

    my $ret = reset_default_mpt_test_flags( \%Test_data );

    my $ret = perform_cmode_validation( \%Test_data );

    my $ret = perform_iron_test_aggr_setup( \%Test_data );

    my $ret = perform_iron_test_vol_setup( \%Test_data );

    my $ret = change_iron_test_vol_property( \%Test_data );

    my $ret = change_iron_test_aggr_property( \%Test_data );

    my $ret = perform_iron_task_setup( \%Test_data );

    my $ret = perform_iron_test_setup( \%Test_data );

    my $ret = perform_iron_test_cleanup( \%Test_data );

    my $ret = mpt_run_iron( \%Test_data );

    my $ret = validate_mpt_test_usr_inputs( \%Test_data );

    my $ret = initialise_mpt_test_common_params( \%Test_data );

    my $test_id = volume_snap_create_in_background( \%Test_data );

    my $test_id = volume_snap_delete_in_background( \%Test_data );

    my $test_id = file_clones_create_in_background( \%Test_data );

    my $ret = stop_subtests(
                subtest_id => $snap_del_test_id,
            );

    my @file_names = generate_mkfile_filename_list(
                        filename  => "TstFile"
                        filecount => 100
                    );

    Keys used by %Test_data :
        aggregate         => name of test aggregate
        volume            => name of test volume
        apiset            => stores apiset object
        iron_task_obj     => stores iron task object
        aggr_task_obj     => stores aggr task object
        iron_timeout      => timeout in seconds for iron completion
        wait_timeout      => timeout in seconds for wait operation
        cmd_timeout       => timeout in seconds for command/component call execution
        sio_blk_size      => blk_size  used by filer_sio
        sio_file_size     => file_size used by filer_sio
        sio_file_name     => file name used by filer_sio
        start_io_ops      => flag to start ops during iron
        start_kireeti     => flag to start kireeti scan during iron
        max_scan_speed    => maximum wafl scan speed
        debug_msg         => flag to turn on/off printing of iron messages
        start_flag_hash   => set flags, before iron is run to fix corruption
        end_flag_hash     => set flags, before iron is run to verify fix
        action            => "get | set ", used to get/set  scan speed
        speed             => set scan speed to this value
        command_interface => node/filer object
        test_continue_flg => flag to continue test, if we miss corruption messages
        op_inode          => scalar reference to return inode of given input file
        crt_file_clone    => flag to start file clone creation [0]
        snapshot_delete   => flag to start snapshot delete ops [0]
        snapshot_interval => delay between two snapshot ops [5]
        snap_crt_prefix   => prefix used for snapshot creation [MptIronSnapshot]
        snap_del_prefix   => prefix used for snapshot deletion [vol_snap_]
        snap_start_index  => snapshot start index [1]
        file_clone_src    => source file path for clone ops
        file_clone_dst    => destination file path for clone ops
        vol_corr_snapname => test volume corrupted snapshot name
        aggr_corr_snapname => test aggregate corrupted snapshot name
        flag_hash_ref     => hash reference to test flags
        ldc_disabled      => flag to disable LDC before second iron run
        snaplock_type     => snaplock type [enterprise]



=head1 DESCRIPTION

    This library implements common interfaces using NACL Task/Component/Apiset,
    which are required for MPT test automation.

=head1 FUNCTIONS

MPT Test Utility Functions

=cut

package IronUtil::MPT::MptTestUtil;


#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($Validation, $VERSION);

#
# constants used by script
#
use Readonly;
Readonly::Scalar our $CONTINUE      => 1;
Readonly::Scalar our $NOT_CONTINUE  => 0;
Readonly::Scalar our $AGGR_ONLY     => 2;
Readonly::Scalar our $FLEXVOL_ONLY  => 3;
Readonly::Scalar our $AGGR_VOL_ONLY => 4;
Readonly::Scalar our $PARAM_NOT_SET => -1;

use warnings;
use strict qw(vars subs refs);
use base   qw(Exporter);

BEGIN {

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @mpt_sub = qw(
                        &fix_by_iron_verify_by_ioc
                        &fix_by_ioc_verify_by_iron
                        &fix_by_prev_cp_verify_by_iron
                        &perform_aggregate_cleanup
                        &wait_until_volume_is_online
                        &get_set_wafl_scan_speed
                        &does_user_file_exist
                        &set_mpt_test_flags
                        &reset_default_mpt_test_flags
                        &perform_cmode_validation
                        &perform_iron_test_aggr_setup
                        &perform_iron_test_vol_setup
                        &change_iron_test_vol_property
                        &change_iron_test_aggr_property
                        &perform_iron_task_setup
                        &perform_iron_test_setup
                        &perform_iron_test_cleanup
                        &mpt_run_iron
                        &validate_mpt_test_usr_inputs
                        &initialise_mpt_test_common_params
                        &volume_snap_create_in_background
                        &volume_snap_delete_in_background
                        &file_clones_create_in_background
                        &stop_subtests
                        &generate_mkfile_filename_list
                        &get_wafl_buftree_blk_info
                        &configure_feature_licenses
                        &delete_aggregate_snapshots
                        &aggregate_snap_create_in_background
                        &aggregate_snap_delete_in_background
                        &get_non_hole_fbns
                        &create_snapshot
                        &create_aggregate_snapshot
                        &does_files_exist_in_directory
                    );

    my @constants = qw( $CONTINUE $NOT_CONTINUE $AGGR_ONLY
                        $FLEXVOL_ONLY $AGGR_VOL_ONLY
                    );

    our %EXPORT_TAGS = (
        'mpt'       => [ @mpt_sub ],
        'constants' => [ @constants ],
    );

    $EXPORT_TAGS{all} = [ map{ @{$_} } values %EXPORT_TAGS ];

    Exporter::export_ok_tags('all');
}

#------------------------------------------------------
#   Imports
#------------------------------------------------------
use Tharn;
use TCD ();
use FS::commonParams;
use Data::Dumper;
use NATE::ParamSet qw();
use Params::Validate qw(validate_with validate :types);


use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException ();
use NACL::C::Wafliron;
use NACL::C::Slist;
use NACL::C::Options;
use NACL::C::VolumeFileClone;
use NACL::C::AggregateSnapshot;
use NACL::STask::Volume;
use NACL::STask::Vserver;
use NACL::MTask::Wafliron;
use NACL::STask::Aggregate;
use NACL::STask::VolumeSnapshot;
use NACL::STask::SystemLicense;
use NACL::STask::VolumeEfficiency;
use NACL::C::AggregateSnapshot;
use IronUtil::MPT::commonParams;

#------------------------------------------------------
#     Public Subroutines Prototype
#------------------------------------------------------
sub fix_by_iron_verify_by_ioc;
sub fix_by_ioc_verify_by_iron;
sub fix_by_prev_cp_verify_by_iron;
sub perform_aggregate_cleanup;
sub wait_until_volume_is_online;
sub get_set_wafl_scan_speed;
sub does_user_file_exist;
sub set_mpt_test_flags;
sub reset_default_mpt_test_flags;
sub perform_cmode_validation;
sub perform_iron_test_aggr_setup;
sub perform_iron_test_vol_setup;
sub change_iron_test_vol_property;
sub change_iron_test_aggr_property;
sub perform_iron_task_setup;
sub perform_iron_test_setup;
sub perform_iron_test_cleanup;
sub mpt_run_iron;
sub validate_mpt_test_usr_inputs;
sub initialise_mpt_test_common_params;
sub volume_snap_create_in_background;
sub volume_snap_delete_in_background;
sub file_clones_create_in_background;
sub stop_subtests;
sub generate_mkfile_filename_list;
sub get_wafl_buftree_blk_info;
sub configure_feature_licenses;
sub delete_aggregate_snapshots;
sub aggregate_snap_create_in_background;
sub aggregate_snap_delete_in_background;
sub get_non_hole_fbns;
sub create_snapshot;
sub create_aggregate_snapshot;
sub does_files_exist_in_directory;


#------------------------------------------------------
#     Private Subroutines Prototype
#------------------------------------------------------
sub _start_io_ops;


#------------------------------------------------------
#     Local Variables
#------------------------------------------------------
my $Tcd_Obj = new TCD( -testcases => [] );

my %Lic_7m_Options = (
                        'licensed_feature.disk_sanitization.enable' => 1,
                        'licensed_feature.fcp.enable' => 1,
                        'licensed_feature.flexcache_nfs.enable' => 1,
                        'licensed_feature.iscsi.enable' => 1,
                        'licensed_feature.multistore.enable' => 1,
                        'licensed_feature.nearstore_option.enable' => 1,
                        'licensed_feature.snaplock.enable' => 1,
                        'licensed_feature.snaplock_enterprise.enable' => 1,
                        'licensed_feature.vld.enable' => 1,
                     );

my $Global_Param_Set = NATE::ParamSet->new(global=>1);

#
# For v1 version this is set to feature
# For v2 version this is set to package
#
my $Lic_Stask_Attribute;



#------------------------------------------------------
#     Subroutines Definition
#------------------------------------------------------

=head2 fix_by_iron_verify_by_ioc

run iron to fix corruption and verifies it by runnig ioc

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset        => node scope or 7-mode apiset
    aggr_task_obj => aggregate task object
    iron_task_obj => iron task object
    iron_timeout  => timout used for iron completion [900]
    start_io_ops  => flag used to start read ops by filersio [0]
    start_flag_hash   => wafl flag setting before fixing the corruption
                         [ wafl_iron_panic_on_corruption : 0
                           wafl_inconsistent_panic_testing : 1 ]
    end_flag_hash     => wafl flag setting after fixing the corruption
                         [ wafl_iron_panic_on_corruption : 0
                           wafl_inconsistent_panic_testing : 1 ]
    test_continue_flg => flag used to continue test execution with
                         relaxed validation [$NOT_CONTINUE]
    debug_msg         => flag to enable iron message printing [0]
    max_scan_speed    => maximum scanner speed [10000]
    start_kireeti     => flag to start kireeti scan [0]
    crt_file_clone    => flag to start file clone creation [0]
    snapshot_create   => flag to start snapshot create ops [0]
    snapshot_delete   => flag to start snapshot delete ops [0]
    snapshot_count    => number of snapshots [50]
    snapshot_interval => delay between two snapshot ops [5]
    snap_crt_prefix   => prefix used for snapshot creation [MptIronSnapshot]
    snap_del_prefix   => prefix used for snapshot deletion [vol_snap_]
    snap_start_index  => snapshot start index [1]
    file_clone_src    => source file path for clone ops
    file_clone_dst    => destination file path for clone ops
    ldc_disabled      => flag to disable LDC before second iron run
    vol_corr_snapname => test volume corrupted snapshot name
    aggr_corr_snapname => test aggregate corrupted snapshot name
    container_type     => type of container [FLEXVOL_ONLY]
    perform_split      => flag to split clone volume [1]
    split_timeout      => timeout for split completion [3600]
    parnt_vol_task_obj => parent volume task object [undef]
    clone_vol_task_obj => clone volume task object [undef]

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise $TCD::FALSE

=back

=cut

sub fix_by_iron_verify_by_ioc {

    $Log->enter();

    my $start_flag_hash_ref = {
        wafl_iron_panic_on_corruption => 0,
        wafl_inconsistent_panic_testing => 1,
        wafl_iron_lazy_dist_checks      => 1,
        wafl_iron_defer_summary_check   => 1,
    };

    my $end_flag_hash_ref = {
        wafl_iron_panic_on_corruption => 0,
        wafl_inconsistent_panic_testing => 1,
        wafl_iron_lazy_dist_checks      => 1,
        wafl_iron_defer_summary_check   => 1,
    };

    my %args_validn_spec =
        (
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'aggr_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'parnt_vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                  default => undef,
                },
            'clone_vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::VolumeClone',
                  default => undef,
                },
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'iron_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::MTask::Wafliron',
                },
            'vserver'           => { type => SCALAR,  default => "vs0" },
            'iron_timeout'      => { type => SCALAR,  default => 900 },
            'wait_timeout'      => { type => SCALAR,  default => 900 },
            'cmd_timeout'       => { type => SCALAR,  default => 900 },
            'start_flag_hash'   => {
                                        type => HASHREF,
                                        default => $start_flag_hash_ref
                                   },
            'end_flag_hash'     => {
                                        type => HASHREF,
                                        default => $end_flag_hash_ref
                                   },
            'test_continue_flg' => { type => SCALAR,  default => $NOT_CONTINUE },
            'debug_msg'         => { type => SCALAR,  default => 0 },
            'max_scan_speed'    => { type => SCALAR,  default => 10000 },
            'file_clone_src'    => { type => SCALAR,  default => "none" },
            'file_clone_dst'    => { type => ARRAYREF,default => [] },
            'start_io_ops'      => { type => SCALAR,  default => 0 },
            'start_kireeti'     => { type => SCALAR,  default => 0 },
            'crt_file_clone'    => { type => SCALAR,  default => 0 },
            'snapshot_create'   => { type => SCALAR,  default => 0 },
            'snapshot_delete'   => { type => SCALAR,  default => 0 },
            'snapshot_count'    => { type => SCALAR,  default => 50 },
            'snapshot_interval' => { type => SCALAR,  default => 5 },
            'snap_crt_prefix'   => { type => SCALAR,  default => 'MptIronSnapshot' },
            'snap_del_prefix'   => { type => SCALAR,  default => 'vol_snap_' },
            'snap_start_index'  => { type => SCALAR,  default => 1 },
            'ldc_disabled'      => { type => SCALAR,  default => 0 },
            'vol_corr_snapname' => { type => SCALAR,  default => 'none' },
            'aggr_corr_snapname' => { type => SCALAR, default => 'none' },
            'container_type'    => { type => SCALAR,  default => $FLEXVOL_ONLY },
            'perform_split'     => { type => SCALAR,  default => 1 },
            'split_timeout'     => { type => SCALAR,  default => 3600 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj         = $options{'apiset'};
    my $iron_task_obj      = $options{'iron_task_obj'};
    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $vol_task_obj       = $options{'vol_task_obj'};
    my $iron_timeout       = $options{'iron_timeout'};
    my $wait_timeout       = $options{'wait_timeout'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $start_io_ops       = $options{'start_io_ops'};
    my $start_kireeti      = $options{'start_kireeti'};
    my $crt_file_clone     = $options{'crt_file_clone'};
    my $start_flag_hash    = $options{'start_flag_hash'};
    my $end_flag_hash      = $options{'end_flag_hash'};
    my $test_continue_flg  = $options{'test_continue_flg'};
    my $debug_msg          = $options{'debug_msg'};
    my $max_scan_speed     = $options{'max_scan_speed'};
    my $vserver            = $options{'vserver'};
    my @file_clone_dst     = @{$options{'file_clone_dst'}};
    my $file_clone_src     = $options{'file_clone_src'};

    my $snapshot_delete    = $options{'snapshot_delete'};
    my $snapshot_create    = $options{'snapshot_create'};
    my $snapshot_count     = $options{'snapshot_count'};
    my $snap_crt_prefix    = $options{'snap_crt_prefix'};
    my $snap_del_prefix    = $options{'snap_del_prefix'};
    my $snap_start_index   = $options{'snap_start_index'};
    my $snapshot_interval  = $options{'snapshot_interval'};
    my $ldc_disabled       = $options{'ldc_disabled'};
    my $vol_corr_snapname  = $options{'vol_corr_snapname'};
    my $aggr_corr_snapname = $options{'aggr_corr_snapname'};
    my $container_type     = $options{'container_type'};
    my $perform_split      = $options{'perform_split'};
    my $parnt_vol_task_obj = $options{'parnt_vol_task_obj'};
    my $clone_vol_task_obj = $options{'clone_vol_task_obj'};
    my $split_timeout      = $options{'split_timeout'};

    my $ret_val             = $TCD::TRUE ;
    my $iron_scan_speed     = $max_scan_speed;
    my $iron_wait           = 1;
    my $sio_failed_to_start = 0;
    my $iron_start_speed    = 1;
    my $scanner_speed       = $iron_start_speed;

    my ( $snap_crt_test_id, $snap_del_test_id );

    my (@snap_crt_subtests_list);
    my (@snap_del_subtests_list);

    my ($tmp_vol_name, $tmp_task_obj);


    my $mpt_disabled   = $Global_Param_Set->get('MPT_DISABLED', default=>0);
    my $dsc_disabled   = $Global_Param_Set->get('DSC_DISABLED', default=>0);

    if ($mpt_disabled) {
        $Tcd_Obj->nlog("=> ** MPT is disabled ** <=");
        $start_flag_hash_ref->{'wafl_iron_lazy_dist_checks'} = 0;
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}   = 0;
    }

    if ($dsc_disabled) {
        $Tcd_Obj->nlog("=> ** DSC is disabled ** <=");
        $Tcd_Obj->nlog("=> ** MPT will also be disabled ** <=");
        $start_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}    = 0;
        $start_flag_hash_ref->{'wafl_iron_defer_summary_check'} = 0;
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}      = 0;
        $end_flag_hash_ref->{'wafl_iron_defer_summary_check'}   = 0;
    }

    #
    # Add any scanner/ops flag, if we want to start it
    # after iron is started with reduced speed
    #
    if ( $start_io_ops
         || $crt_file_clone
         || $snapshot_delete
         || $snapshot_create
         || $start_kireeti ) {
        $iron_scan_speed = $iron_start_speed;
        $iron_wait  = 0;
        $scanner_speed = 100;
    }

    #
    ## Run wafliron with scanner speed set to 1 or $max_scan_speed
    #
    try {
        $Tcd_Obj->step("Run wafliron with scanner speed set to \"$iron_scan_speed\"");
        $iron_task_obj->start(
            nacltask_include_mirrors => 0,
            nacltask_optional_commit => 'false',
            nacltask_wait            => $iron_wait,
            nacltask_wafl_scan_speed => $iron_scan_speed,
            nacltask_flags_to_set    => $start_flag_hash,
            'method-timeout'         => $iron_timeout,
        );
    } catch NACL::Exceptions::Timeout with {
            my $excpn_obj = shift;
            $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
            $Tcd_Obj->nlog("Wait for some more time to complete iron");
            $iron_task_obj->wait_for_completion(
                'method-timeout' => $iron_timeout
            );
    };

    if ( 0 == $iron_wait ) {

        #
        ## Wait for volume to be online
        #
        $Tcd_Obj->step("Wait for test volume to be online");
        wait_until_volume_is_online( $test_data );

        sleep(10);

        #
        ## Start read/write ops using filersio
        #
        if ( $start_io_ops ) {
             $Tcd_Obj->step("Starting read/write ops using filersio");
            _start_io_ops( $test_data );
        }

        #
        ## Change the scanner speed, if required
        #
        if ( $scanner_speed != $iron_start_speed ) {
            $Tcd_Obj->step("Setting wafl scan speed to $scanner_speed");
            get_set_wafl_scan_speed(
                {
                    action  => "set",
                    speed   => $scanner_speed,
                    apiset  => $apiset_obj,
                });
        }

        #
        ## Start kireeti scanner
        #
        if ( $start_kireeti ) {
            $Tcd_Obj->step("Start kireeti scanner");
            $apiset_obj->wafl_scan_kireeti(
                'privilege-level' => "test",
                'volume-name' => $vol_task_obj->volume,
            );
        }

        #
        ## Start file clone creation
        #
        if ( $crt_file_clone ) {
            $Tcd_Obj->step("Start file clone creation");
            foreach my $dst_clone ( @file_clone_dst ) {
                $Tcd_Obj->nlog("Creating file clone: $dst_clone");
                my $fc1 = NACL::C::VolumeFileClone->create(
                    command_interface  => $vol_task_obj->command_interface,
                    vserver            => $vserver,
                    'source-path'      => $file_clone_src,
                    'destination-path' => $dst_clone,
                    'clone_type'       => "file",
                    'method-timeout'   => $cmd_timeout,
                );
            }
        }

        #
        ## Start snapshot delete operation
        #
        if ( $snapshot_delete ) {
            if ( $container_type =~ /($FLEXVOL_ONLY|$AGGR_VOL_ONLY)/ ) {
                $Tcd_Obj->step("Start flexvol's snapshots delete operation ");
                $snap_del_test_id = volume_snap_delete_in_background(
                    vol_task_obj      => $vol_task_obj,
                    snapshot_count    => $snapshot_count,
                    snapshot_interval => $snapshot_interval,
                    snap_del_prefix   => $snap_del_prefix,
                    snap_start_index  => $snap_start_index,
                );
                push(@snap_del_subtests_list,$snap_del_test_id);
            }

            if ( $container_type =~ /($AGGR_ONLY|$AGGR_VOL_ONLY)/ ) {
                $Tcd_Obj->step("Start aggr's snapshots delete operation ");
                $snap_del_test_id = aggregate_snap_delete_in_background(
                    aggr_task_obj     => $aggr_task_obj,
                    snapshot_count    => $snapshot_count,
                    snapshot_interval => $snapshot_interval,
                    snap_del_prefix   => $snap_del_prefix,
                    snap_start_index  => $snap_start_index,
                );
                push(@snap_del_subtests_list,$snap_del_test_id);
            }
        }

        #
        ## Start snapshot creation operation
        #
        if ( $snapshot_create ) {
            if ( $container_type =~ /($FLEXVOL_ONLY|$AGGR_VOL_ONLY)/ ) {
                $Tcd_Obj->step("Start flexvol's snapshots create operation ");
                $snap_crt_test_id = volume_snap_create_in_background(
                    vol_task_obj      => $vol_task_obj,
                    snapshot_count    => $snapshot_count,
                    snapshot_interval => $snapshot_interval,
                    snap_crt_prefix   => $snap_crt_prefix,
                    snap_start_index  => $snap_start_index,
                );
                push(@snap_crt_subtests_list,$snap_crt_test_id);
            }

            if ( $container_type =~ /($AGGR_ONLY|$AGGR_VOL_ONLY)/ ) {
                $Tcd_Obj->step("Start aggr's snapshots create operation ");
                $snap_crt_test_id = aggregate_snap_create_in_background(
                    aggr_task_obj     => $aggr_task_obj,
                    snapshot_count    => $snapshot_count,
                    snapshot_interval => $snapshot_interval,
                    snap_del_prefix   => $snap_del_prefix,
                    snap_start_index  => $snap_start_index,
                );
                push(@snap_crt_subtests_list,$snap_crt_test_id);
            }
        }

        $Tcd_Obj->step("Sleeping for 30 seconds");
        sleep(30);

        #
        ## Reset the iron speed to maximum
        #
        $Tcd_Obj->step("Reset wafl scan speed to \"$max_scan_speed\"");
        get_set_wafl_scan_speed(
            {
                action  => "set",
                speed   => $max_scan_speed,
                apiset  => $apiset_obj,
            });
        #
        ## Wait for wafliron completion
        #
        try {
            $Tcd_Obj->step("Waiting for iron completion");
            $iron_task_obj->wait_for_completion('method-timeout' => $iron_timeout);
        } catch NACL::Exceptions::Timeout with {
            my $excpn_obj = shift;
            $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
            $Tcd_Obj->nlog("Wait for some more time to complete iron");
            $iron_task_obj->wait_for_completion(
                'method-timeout' => $iron_timeout
            );
        };

        #
        ## Stop filersio ops after iron completion
        #
        if ( $start_io_ops ) {
            try {
                $Tcd_Obj->step("Stopping filersio after iron completion");
                my $resp_obj = $apiset_obj->filersio_stop(
                     'instance-id' => " ", );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $excpn_obj = shift;
                $Tcd_Obj->nlog( "Filersio status is:: " . $excpn_obj->text() );
            };
        }

        #
        ## Stop snapshot deletion subtest
        #
        if ( $snapshot_delete ) {
            $Tcd_Obj->step("Stopping snapshot delete operation");
            foreach my $test_id ( @snap_del_subtests_list ) {
                stop_subtests(
                    subtest_id => $test_id,
                );
            }
        }

        #
        ## Stop snapshot creation subtest
        #
        if ( $snapshot_create ) {
            $Tcd_Obj->step("Stopping snapshot create operation");
            foreach my $test_id ( @snap_crt_subtests_list ) {
                stop_subtests(
                    subtest_id => $test_id,
                );
            }
        }
    } # end if

    if ( $debug_msg ) {
        my $messages = $iron_task_obj->get_wafliron_messages();
        $Tcd_Obj->nlog("Iron Messages : \n", Dumper  $messages);
    }

    if ( ! $iron_task_obj->corruption_found() ) {
        $Tcd_Obj->nlog("Corruption is not found by regular iron");
        if( $NOT_CONTINUE == $test_continue_flg ) {
            $Log->exit();
            return $TCD::FALSE;
        }

        $Tcd_Obj->nlog("We will verify fix with IOC with "
            . "following flag setting");
        $Tcd_Obj->nlog("wafl_iron_panic_on_corruption=1,\n"
            . "wafl_inconsistent_panic_testing=0");
        $end_flag_hash->{'wafl_iron_panic_on_corruption'} = 1;
        $end_flag_hash->{'wafl_inconsistent_panic_testing'} = 0;
    } else {
        $Tcd_Obj->nlog("Corruption is found and fixed by regular iron");
    }

    #
    ## Perform clone split, if required
    #
    if ( $perform_split && $clone_vol_task_obj ) {
        $tmp_vol_name = $clone_vol_task_obj->volume;
        $Tcd_Obj->step("Splitting clone volume : \"$tmp_vol_name\"");
        $clone_vol_task_obj->split(
            nacltask_wait    => 1,
            'method-timeout' => $split_timeout,
        );
    }

    #
    ## Disable LDC, before next run of iron
    #
    if ( 1 == $ldc_disabled ) {
        $Tcd_Obj->step("Resetting MPT/LDC test flags, before second iron run");
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'} = 0;
    }

    #
    ## Delete aggregate corrupted snapshot before next iron run
    #
    if ( $aggr_corr_snapname ne 'none' ) {
        $Tcd_Obj->step("Deleting corrupted snapshot \"$aggr_corr_snapname\" "
            . " in test aggregate before second iron run");
        delete_aggregate_snapshots(
            command_interface => $aggr_task_obj->command_interface,
            aggregate         => $aggr_task_obj->aggregate,
            snapshot          => $aggr_corr_snapname,
            cmd_timeout       => $cmd_timeout,
        );
    }

    #
    ## Delete volume corrupted snapshot before next iron run
    #
    if ( $vol_corr_snapname ne 'none' ) {
        $tmp_task_obj = $vol_task_obj;
        $tmp_vol_name = $tmp_task_obj->volume;

        $Tcd_Obj->step("Deleting corrupted snapshot"
            . " \"$vol_corr_snapname\" in test volume :"
            . " \"$tmp_vol_name\" before second iron run");

        NACL::STask::VolumeSnapshot->purge(
            command_interface => $tmp_task_obj->command_interface,
            snapshot          => $vol_corr_snapname,
            volume            => $tmp_task_obj->volume,
            vserver           => $tmp_task_obj->vserver,
            nacltask_if_purged => 'pass',
        );

        if ( $parnt_vol_task_obj ) {
            $tmp_task_obj = $parnt_vol_task_obj;
            $tmp_vol_name = $tmp_task_obj->volume;

            $Tcd_Obj->step("Deleting corrupted snapshot"
                . " \"$vol_corr_snapname\" in parent volume :"
                . " \"$tmp_vol_name\" before second iron run");

            NACL::STask::VolumeSnapshot->purge(
                command_interface => $tmp_task_obj->command_interface,
                snapshot          => $vol_corr_snapname,
                volume            => $tmp_task_obj->volume,
                vserver           => $tmp_task_obj->vserver,
                nacltask_if_purged => 'pass',
            );
        }
    }

    #
    # Delete 'preiron*' snapshot taken by iron at the end
    #
    $Tcd_Obj->step("Deleting aggr's snapshot created by iron");
    delete_aggregate_snapshots(
        command_interface => $aggr_task_obj->command_interface,
        aggregate         => $aggr_task_obj->aggregate,
        snapshot          => "preiron*",
        cmd_timeout       => $cmd_timeout,
    );

    #
    ## Run wafliron with in optional commit mode
    #
    try {
        $Tcd_Obj->step("Run wafliron in optional commit mode");
        $iron_task_obj->start(
            nacltask_include_mirrors => 0,
            nacltask_optional_commit => 'true',
            nacltask_wait            => 1,
            nacltask_wafl_scan_speed => $max_scan_speed,
            nacltask_flags_to_set    => $end_flag_hash,
            'method-timeout'         => $iron_timeout,
        );
    } catch NACL::Exceptions::Timeout with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
        $Tcd_Obj->nlog("Wait for some more time to complete iron");
        $iron_task_obj->wait_for_completion(
            'method-timeout' => $iron_timeout
        );
    };

    $Tcd_Obj->nlog("Let's wait for 10 sec and then check aggregate state");
    sleep(10);

    #
    ## Verify, aggregate is in offline state
    #
    $Tcd_Obj->step("Verify, aggregate is in offline state");
    my $state =  $aggr_task_obj->state()->state();
    if ( $state eq "iron_restricted" ) {
        #
        # Don't worry about cleanup, it will be performed by
        # perform_aggregate_cleanup() call in test script cleanup()
        #
        $Tcd_Obj->nlog("Aggregate state is iron_restricted instead of offline");

        my $messages = $iron_task_obj->review();
        $Tcd_Obj->nlog("Iron review messages:\n $messages");

        $Log->exit();
        return $TCD::FALSE;
    } elsif( $state eq "offline" ) {
        $Tcd_Obj->nlog("Aggregate state is offline as expected");
    } else {
        $Tcd_Obj->nlog("Unexpected aggregate state: $state");
        $Log->exit();
        return $TCD::FALSE;
    }

    $Tcd_Obj->nlog("Let's wait for 10 sec, before making aggr online");
    sleep(10);

    #
    ## Bring aggregate to online state
    #
    $Tcd_Obj->step("Bringing the aggregate online");
    $aggr_task_obj->modify(
        'state'           => 'online',
        'unmount-volumes' => 'true',
    );

    #
    ## Verify corruption is not reported by IOC
    #
    $Tcd_Obj->step("Verify, corruption is not reported again by IOC");
    if ( $iron_task_obj->corruption_found() ) {
        $Tcd_Obj->nlog("Corruption fixed by iron is still reported by ioc");
        my $messages = $iron_task_obj->get_wafliron_messages();
        $Tcd_Obj->nlog("Iron Messages : \n", Dumper  $messages);
        $ret_val = $TCD::FALSE;
    } else {
        $Tcd_Obj->nlog("Corruption fixed by iron is not reported by ioc");
    }

    $Log->exit();
    return $ret_val;
} # end of fix_by_iron_verify_by_ioc()


=head2 fix_by_ioc_verify_by_iron

run ioc to fix corruption and verifies it by running iron

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    aggr_task_obj      => aggregate task object
    iron_task_obj      => iron task object
    vol_task_obj       => volume task object
    parnt_vol_task_obj => parent volume task object
    clone_vol_task_obj => clone  volume task object
    iron_timeout       => timout used for iron completion [900]
    max_scan_speed     => maximum scanner speed
    debug_msg          => flag to enable printing of iron messages
    ldc_disabled       => flag to disable LDC before second iron run [0]
    start_flag_hash    => wafl flag setting before fixing the corruption
                         [wafl_iron_panic_on_corruption:0
                          wafl_inconsistent_panic_testing:1 ]
    end_flag_hash      => wafl flag setting after fixing the corruption
                         [wafl_iron_panic_on_corruption:0
                          wafl_inconsistent_panic_testing:1 ]
    test_continue_flg  => flag used to continue test execution with
                         relaxed validation [$NOT_CONTINUE]
    vol_corr_snapname  => test volume corrupted snapshot name
    aggr_corr_snapname => test aggregate corrupted snapshot name

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise $TCD::FALSE

=back

=cut

sub fix_by_ioc_verify_by_iron {
    $Log->enter();

    my $start_flag_hash_ref = {
        wafl_iron_panic_on_corruption => 0,
        wafl_inconsistent_panic_testing => 1,
        wafl_iron_lazy_dist_checks      => 1,
        wafl_iron_defer_summary_check   => 1,
    };

    my $end_flag_hash_ref = {
        wafl_iron_panic_on_corruption => 0,
        wafl_inconsistent_panic_testing => 1,
        wafl_iron_lazy_dist_checks      => 1,
        wafl_iron_defer_summary_check   => 1,
    };

    my %args_validn_spec =
        (
            'aggr_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'parnt_vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                  default => undef,
                },
            'clone_vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::VolumeClone',
                  default => undef,
                },
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'iron_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::MTask::Wafliron',
                },
            'iron_timeout'      => { type => SCALAR, default => 900 },
            'start_flag_hash'   => {
                                        type => HASHREF,
                                        default => $start_flag_hash_ref
                                   },
            'end_flag_hash'     => {
                                        type => HASHREF,
                                        default => $end_flag_hash_ref
                                   },
            'test_continue_flg' => { type => SCALAR,  default => $NOT_CONTINUE },
            'max_scan_speed'    => { type => SCALAR,  default => 10000 },
            'debug_msg'         => { type => SCALAR,  default => 0 },
            'ldc_disabled'      => { type => SCALAR,  default => 0 },
            'vol_corr_snapname' => { type => SCALAR,  default => 'none' },
            'cmd_timeout'       => { type => SCALAR,  default => 900 },
            'aggr_corr_snapname' => { type => SCALAR, default => 'none' },
            'perform_split'     => { type => SCALAR,  default => 1 },
            'split_timeout'     => { type => SCALAR,  default => 3600 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $iron_task_obj      = $options{'iron_task_obj'};
    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $vol_task_obj       = $options{'vol_task_obj'};
    my $iron_timeout       = $options{'iron_timeout'};
    my $debug_msg          = $options{'debug_msg'};
    my $max_scan_speed     = $options{'max_scan_speed'};
    my $start_flag_hash    = $options{'start_flag_hash'};
    my $end_flag_hash      = $options{'end_flag_hash'};
    my $test_continue_flg  = $options{'test_continue_flg'};
    my $ldc_disabled       = $options{'ldc_disabled'};
    my $vol_corr_snapname  = $options{'vol_corr_snapname'};
    my $aggr_corr_snapname = $options{'aggr_corr_snapname'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $perform_split      = $options{'perform_split'};
    my $parnt_vol_task_obj = $options{'parnt_vol_task_obj'};
    my $clone_vol_task_obj = $options{'clone_vol_task_obj'};
    my $split_timeout      = $options{'split_timeout'};


    my ($messages, $tmp_vol_name, $tmp_task_obj);

    my $mpt_disabled   = $Global_Param_Set->get('MPT_DISABLED', default=>0);
    my $dsc_disabled   = $Global_Param_Set->get('DSC_DISABLED', default=>0);

    if ($mpt_disabled) {
        $Tcd_Obj->nlog("=> ** MPT is disabled ** <=");
        $start_flag_hash_ref->{'wafl_iron_lazy_dist_checks'} = 0;
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}   = 0;
    }

    if ($dsc_disabled) {
        $Tcd_Obj->nlog("=> ** DSC is disabled ** <=");
        $Tcd_Obj->nlog("=> ** MPT will also be disabled ** <=");
        $start_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}    = 0;
        $start_flag_hash_ref->{'wafl_iron_defer_summary_check'} = 0;
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}      = 0;
        $end_flag_hash_ref->{'wafl_iron_defer_summary_check'}   = 0;
    }

    #
    ## Run wafliron in optional commit mode
    #
    try {
        $Tcd_Obj->step("Run wafliron in optional commit mode");
        $iron_task_obj->start(
            nacltask_include_mirrors => 0,
            nacltask_optional_commit => 'true',
            nacltask_wait            => 1,
            nacltask_wafl_scan_speed => $max_scan_speed,
            nacltask_flags_to_set    => $start_flag_hash,
            'method-timeout'         => $iron_timeout,
        );
    } catch NACL::Exceptions::Timeout with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
        $Tcd_Obj->nlog("Wait for some more time to complete iron");
        $iron_task_obj->wait_for_completion(
            'method-timeout' => $iron_timeout
        );
    };

    #
    ## Verify, corruption is fixed by IOC
    #
    $Tcd_Obj->step("Verify, corruption is fixed by IOC");
    if ( ! $iron_task_obj->corruption_found() ) {
        $Tcd_Obj->nlog("Corruption is not found by ioc");
        if( $NOT_CONTINUE == $test_continue_flg ) {
            $Log->exit();
            return $TCD::FALSE;
        }

        $Tcd_Obj->nlog("We will verify fix with iron "
            . "with following flag setting");
        $Tcd_Obj->nlog("wafl_iron_panic_on_corruption=1,\n"
            . "wafl_inconsistent_panic_testing=0");
        $end_flag_hash->{'wafl_iron_panic_on_corruption'} = 1;
        $end_flag_hash->{'wafl_inconsistent_panic_testing'} = 0;
    } else {
        $Tcd_Obj->nlog("Corruption is found and fixed by ioc");
    }

    if ( $debug_msg ) {
        $messages = $iron_task_obj->get_wafliron_messages();
        $Tcd_Obj->nlog("Iron Messages : \n", Dumper  $messages);
        $messages = $iron_task_obj->review();
        $Tcd_Obj->nlog("Iron review messages:\n $messages");
    }

    $Tcd_Obj->nlog("Let's wait for 10 sec and then check aggregate state");
    sleep(10);

    #
    ## Verify, aggregate is in iron_restricted state
    #
    $Tcd_Obj->step("Verify, aggregate is in iron_restricted state");
    my $state =  $aggr_task_obj->state()->state();
    if ( $state eq "iron_restricted" ) {
        $Tcd_Obj->nlog("Aggregate state is \"$state\" as expected");
    } else {
        $Tcd_Obj->nlog("Aggregate state is not restricted, but $state");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    ## Commit the iron changes
    #
    $Tcd_Obj->step("Commiting the changes");
    $iron_task_obj->commit( nacltask_force => 1 );

    $Tcd_Obj->nlog("Let's wait for 20 sec to allow aggr to be offline");
    sleep(20);

    #
    ## Verify, aggregate is in offline state
    #
    $Tcd_Obj->step("Verify, aggregate is in offline state");
    $state =  $aggr_task_obj->state()->state();
    if ( $state eq "offline" ) {
        $Tcd_Obj->nlog("Aggregate state is \"$state\" as expected");
    } else {
        $Tcd_Obj->nlog("Aggregate state is not offline, but $state");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    ## Bring aggregate in online state
    #
    $Tcd_Obj->step("Bringing aggregate in online state");
    $aggr_task_obj->modify(
        'state'           => 'online',
        'unmount-volumes' => 'true',
    );

    #
    ## Perform clone split, if required
    #
    if ( $perform_split && $clone_vol_task_obj ) {
        $tmp_vol_name = $clone_vol_task_obj->volume;
        $Tcd_Obj->step("Splitting clone volume : \"$tmp_vol_name\"");
        $clone_vol_task_obj->split(
            nacltask_wait    => 1,
            'method-timeout' => $split_timeout,
        );
    }

    #
    ## Disable LDC, before next run of iron
    #
    if ( 1 == $ldc_disabled ) {
        $Tcd_Obj->step("Resetting MPT/LDC test flags, before second iron run");
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'} = 0;
    }

    #
    # Delete 'preiron*' snapshot taken by iron at the end
    #
    $Tcd_Obj->step("Deleting aggr's snapshot created by iron");
    delete_aggregate_snapshots(
        command_interface => $aggr_task_obj->command_interface,
        aggregate         => $aggr_task_obj->aggregate,
        snapshot          => "preiron*",
        cmd_timeout       => $cmd_timeout,
    );

    #
    ## Delete aggregate corrupted snapshot before next iron run
    #
    if ( $aggr_corr_snapname ne 'none' ) {
        $Tcd_Obj->step("Deleting corrupted snapshot \"$aggr_corr_snapname\" "
            . " in test aggregate before second iron run");
        delete_aggregate_snapshots(
            command_interface => $aggr_task_obj->command_interface,
            aggregate         => $aggr_task_obj->aggregate,
            snapshot          => $aggr_corr_snapname,
            cmd_timeout       => $cmd_timeout,
        );
    }

    #
    ## Delete volume corrupted snapshot before next iron run
    #
    if ( $vol_corr_snapname ne 'none' ) {
        $tmp_task_obj = $vol_task_obj;
        $tmp_vol_name = $tmp_task_obj->volume;

        $Tcd_Obj->step("Deleting corrupted snapshot"
            . " \"$vol_corr_snapname\" in test volume :"
            . " \"$tmp_vol_name\" before second iron run");

        NACL::STask::VolumeSnapshot->purge(
            command_interface => $tmp_task_obj->command_interface,
            snapshot          => $vol_corr_snapname,
            volume            => $tmp_task_obj->volume,
            vserver           => $tmp_task_obj->vserver,
            nacltask_if_purged => 'pass',
        );

        if ( $parnt_vol_task_obj ) {
            $tmp_task_obj = $parnt_vol_task_obj;
            $tmp_vol_name = $tmp_task_obj->volume;

            $Tcd_Obj->step("Deleting corrupted snapshot"
                . " \"$vol_corr_snapname\" in parent volume :"
                . " \"$tmp_vol_name\" before second iron run");

            NACL::STask::VolumeSnapshot->purge(
                command_interface => $tmp_task_obj->command_interface,
                snapshot          => $vol_corr_snapname,
                volume            => $tmp_task_obj->volume,
                vserver           => $tmp_task_obj->vserver,
                nacltask_if_purged => 'pass',
            );
        }
    }

    #
    ## Run wafliron
    #
    try {
        $Tcd_Obj->step("Run wafliron again");
        $iron_task_obj->start(
            nacltask_include_mirrors => 0,
            nacltask_optional_commit => 'false',
            nacltask_wait            => 1,
            nacltask_wafl_scan_speed => $max_scan_speed,
            nacltask_flags_to_set    => $end_flag_hash,
            'method-timeout'         => $iron_timeout,
        );
    } catch NACL::Exceptions::Timeout with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
        $Tcd_Obj->nlog("Wait for some more time to complete iron");
        $iron_task_obj->wait_for_completion(
            'method-timeout' => $iron_timeout
        );
    };

    #
    ## Verify, no corruption is reported by iron
    #
    $Tcd_Obj->step("Verify, no corruption is reported by iron");
    if ( $iron_task_obj->corruption_found() ) {
        $Tcd_Obj->nlog("Corruption fixed by ioc is still reported by iron");
        $messages = $iron_task_obj->get_wafliron_messages();
        $Tcd_Obj->nlog("Iron Messages : \n", Dumper  $messages);
        $Log->exit();
        return $TCD::FALSE;
    } else {
        $Tcd_Obj->nlog("Corruption fixed by ioc is not reported by iron");
    }

    $Log->exit();

    return $TCD::TRUE;
} # end of fix_by_ioc_verify_by_iron()

=head2 fix_by_prev_cp_verify_by_iron

run ioc with prev_cp to fix corruption and verifies it by running iron

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    aggr_task_obj     => aggregate task object
    iron_task_obj     => iron task object
    vol_task_obj      => volume task object
    iron_timeout      => timout used for iron completion [900]
    max_scan_speed    => maximum scanner speed
    debug_msg         => flag to enable printing of iron messages
    ldc_disabled      => flag to disable LDC before second iron run [0]
    start_flag_hash   => wafl flag setting before fixing the corruption
                         [wafl_iron_panic_on_corruption:0
                          wafl_inconsistent_panic_testing:1 ]
    end_flag_hash     => wafl flag setting after fixing the corruption
                         [wafl_iron_panic_on_corruption:0
                          wafl_inconsistent_panic_testing:1 ]
    aggr_resp_hash    => aggregate response hash
    vol_resp_hash     => volume response hash
    incmpl_vol_resp   => incomplete volume response string
    vol_snapid_list   => volume:snapid list
    review_msg_list   => array ref to list of messages to be checked
    check_iron_logs   => set this flag to 1 to check presence of iron logs [0]
    prev_cp_mode      => type of prev_cp  [semi_interactive]
    vol_corr_snapname => test volume corrupted snapshot name
    test_continue_flg => flag used to continue test execution with
                         relaxed validation [$NOT_CONTINUE]

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise $TCD::FALSE

=back

=cut

sub fix_by_prev_cp_verify_by_iron {

    $Log->enter();

    my $start_flag_hash_ref = {
        wafl_iron_panic_on_corruption => 0,
        wafl_inconsistent_panic_testing => 1,
        wafl_iron_lazy_dist_checks      => 1,
        wafl_iron_defer_summary_check   => 1,
    };

    my $end_flag_hash_ref = {
        wafl_iron_panic_on_corruption => 0,
        wafl_inconsistent_panic_testing => 1,
        wafl_iron_lazy_dist_checks      => 1,
        wafl_iron_defer_summary_check   => 1,
    };

    my %args_validn_spec =
        (
            'aggr_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'iron_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::MTask::Wafliron',
                },
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                  default => undef,
                },
            'iron_timeout'      => { type => SCALAR, default => 900 },
            'start_flag_hash'   => {
                                        type => HASHREF,
                                        default => $start_flag_hash_ref
                                   },
            'end_flag_hash'     => {
                                        type => HASHREF,
                                        default => $end_flag_hash_ref
                                   },
            'max_scan_speed'    => { type => SCALAR,  default => 10000 },
            'debug_msg'         => { type => SCALAR,  default => 1 },
            'ldc_disabled'      => { type => SCALAR,  default => 0 },
            'aggr_resp_hash'    => { type => HASHREF },
            'vol_resp_hash'     => { type => HASHREF,  default => undef },
            'incmpl_vol_resp'   => { type => SCALAR,   default => "" },
            'vol_snapid_list'   => { type => SCALAR,   default => undef },
            'check_iron_logs'   => { type => SCALAR,   default => 0 },
            'review_msg_list'   => { type => ARRAYREF, default => undef },
            'prev_cp_mode'      => { type => SCALAR,
                                     default => 'semi_interactive',
                                   },
            'cmd_timeout'       => { type => SCALAR,  default => 900 },
            'vol_corr_snapname' => { type => SCALAR,  default => 'none' },
            'test_continue_flg' => { type => SCALAR,  default => $NOT_CONTINUE },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );
    my $iron_task_obj      = $options{'iron_task_obj'};
    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $vol_task_obj       = $options{'vol_task_obj'};
    my $iron_timeout       = $options{'iron_timeout'};
    my $debug_msg          = $options{'debug_msg'};
    my $max_scan_speed     = $options{'max_scan_speed'};
    my $start_flag_hash    = $options{'start_flag_hash'};
    my $end_flag_hash      = $options{'end_flag_hash'};
    my $ldc_disabled       = $options{'ldc_disabled'};
    my $aggr_resp_hash     = $options{'aggr_resp_hash'};
    my $vol_resp_hash      = $options{'vol_resp_hash'};
    my $incmpl_vol_resp    = $options{'incmpl_vol_resp'};
    my $vol_snapid_list    = $options{'vol_snapid_list'};
    my $prev_cp_mode       = $options{'prev_cp_mode'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $vol_corr_snapname  = $options{'vol_corr_snapname'};
    my $test_continue_flg  = $options{'test_continue_flg'};


    my $mpt_disabled = $Global_Param_Set->get('MPT_DISABLED', default=>0);
    my $dsc_disabled = $Global_Param_Set->get('DSC_DISABLED', default=>0);

    if ($mpt_disabled) {
        $Tcd_Obj->nlog("=> ** MPT is disabled ** <=");
        $start_flag_hash_ref->{'wafl_iron_lazy_dist_checks'} = 0;
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}   = 0;
    }

    if ($dsc_disabled) {
        $Tcd_Obj->nlog("=> ** DSC is disabled ** <=");
        $Tcd_Obj->nlog("=> ** MPT will also be disabled ** <=");
        $start_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}    = 0;
        $start_flag_hash_ref->{'wafl_iron_defer_summary_check'} = 0;
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'}      = 0;
        $end_flag_hash_ref->{'wafl_iron_defer_summary_check'}   = 0;
    }

    my %prev_cp_options = ();

    $prev_cp_options{'aggregate_response'} = $aggr_resp_hash;

    if ($vol_resp_hash) {
        $Tcd_Obj->nlog("volume_response defined: ", Dumper $vol_resp_hash);
        $prev_cp_options{'volume_response'} = $vol_resp_hash;
    }

    if ( defined $incmpl_vol_resp ) {
        $prev_cp_options{'incomplete_volume_response'}
            = $incmpl_vol_resp;
    }

    if ( defined $vol_snapid_list ) {
        $Tcd_Obj->nlog("volume_with_snap_id_list defined: $vol_snapid_list");
        $prev_cp_options{'volume_with_snap_id_list'}
            = $vol_snapid_list;
    }

    #
    ## Run wafliron in prev_cp mode
    #
    try {
        $Tcd_Obj->step("Run wafliron in IOC with prev_cp mode");
        $iron_task_obj->start(
            nacltask_include_mirrors => 0,
            nacltask_optional_commit => 'prev_cp',
            nacltask_prev_cp => {
                prev_cp_mode  => $prev_cp_mode,
                %prev_cp_options,
            },
            nacltask_wait            => 1,
            nacltask_wafl_scan_speed => $max_scan_speed,
            nacltask_flags_to_set    => $start_flag_hash,
            'method-timeout'         => $iron_timeout,
        );
    } catch NACL::Exceptions::Timeout with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
        $Tcd_Obj->nlog("Wait for some more time to complete iron");
        $iron_task_obj->wait_for_completion(
            'method-timeout' => $iron_timeout
        );
    };

    #
    ## Verify, corruption is fixed by IOC
    #
    $Tcd_Obj->step("Verify, corruption is fixed by iron in prev_cp mode");

    if ( ! $iron_task_obj->corruption_found() ) {
        $Tcd_Obj->nlog("Corruption is not found by prev_cp iron");
        if( $NOT_CONTINUE == $test_continue_flg ) {
            $Log->exit();
            return $TCD::FALSE;
        }

        $Tcd_Obj->nlog("We will verify fix with IRON with "
            . "following flag setting");
        $Tcd_Obj->nlog("wafl_iron_panic_on_corruption=1,\n"
            . "wafl_inconsistent_panic_testing=0");
        $end_flag_hash->{'wafl_iron_panic_on_corruption'} = 1;
        $end_flag_hash->{'wafl_inconsistent_panic_testing'} = 0;
    } else {
        $Tcd_Obj->nlog("Corruption is found and fixed by prev_cp iron");
    }

    #
    ## Retrieve iron review messages
    #
    my $review_msg = $iron_task_obj->review();

    if ( $debug_msg ) {
        my $iron_console_msg = $iron_task_obj->get_wafliron_messages();
        $Tcd_Obj->nlog("Iron Messages : \n", Dumper  $iron_console_msg);
        $Tcd_Obj->nlog("Iron review messages:\n $review_msg");
    }

    #
    ## Verify iron review messages
    #
    if ( $options{'review_msg_list'} ) {
        $Tcd_Obj->step("Verifying message patterns in iron review messages");
        my @patrn_list = @{$options{'review_msg_list'}};
        my $found = 1;
        foreach my $patrn ( @patrn_list ) {
            if ( grep(/$patrn/, $review_msg) ) {
                $Tcd_Obj->nlog("Pattern: $patrn found in review messages");
            } else {
                $Tcd_Obj->nlog("Pattern: $patrn missing in review messages");
                $found = 0;
            }
        }

        return $TCD::FALSE if ( !$found );
    }

    #
    ## Check iron log files in metadir directory
    #
    if ( $options{'check_iron_logs'} ) {
        $Tcd_Obj->step("Checking presence of iron log files in metadir");
        my $status = does_files_exist_in_directory(
            command_interface => $aggr_task_obj->command_interface,
            dir_name => '/vol/vol0/metadir/irondir',
            file_list => [ 'OC-Change-Log', 'OC-LW-context', 'OC-Bplus' ],
        );
        return $TCD::FAIL if ( $TCD::FAIL == $status );
    }

    $Tcd_Obj->nlog("Let's wait for 10 sec and then check aggregate state");
    sleep(10);

    #
    ## Verify, aggregate is in iron_restricted state
    #
    $Tcd_Obj->step("Verify, aggregate is in iron_restricted state");
    my $state =  $aggr_task_obj->state()->state();
    if ( $state eq "iron_restricted" ) {
        $Tcd_Obj->nlog("Aggregate state is \"$state\" as expected");
    } else {
        $Tcd_Obj->nlog("Aggregate state is not restricted, but $state");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    ## Commit the iron changes
    #
    $Tcd_Obj->step("Commiting the changes");
    $iron_task_obj->commit( nacltask_force => 1 );

    $Tcd_Obj->nlog("Let's wait for 20 sec to allow aggr to be offline");
    sleep(20);

    #
    ## Verify, aggregate is in offline state
    #
    $Tcd_Obj->step("Verify, aggregate is in offline state");
    $state =  $aggr_task_obj->state()->state();
    if ( $state eq "offline" ) {
        $Tcd_Obj->nlog("Aggregate state is \"$state\" as expected");
    } else {
        $Tcd_Obj->nlog("Aggregate state is not offline, but $state");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    ## Bring aggregate in online state
    #
    $Tcd_Obj->step("Bringing aggregate in online state");
    $aggr_task_obj->modify(
        'state'           => 'online',
        'unmount-volumes' => 'true',
    );

    #
    ## Check iron log files are deleated in meta directory
    #
    if ( $options{'check_iron_logs'} ) {
        $Tcd_Obj->step("Checking iron log files are deleated");
        my $status = does_files_exist_in_directory(
            command_interface => $aggr_task_obj->command_interface,
            dir_name => '/vol/vol0/metadir/irondir',
            file_list => [ 'OC-Change-Log', 'OC-LW-context', 'OC-Bplus' ],
        );
        return $TCD::FAIL if ( $TCD::PASS == $status );
    }

    #
    ## Disable LDC, before next run of iron
    #
    if ( 1 == $ldc_disabled ) {
        $Tcd_Obj->step("Resetting MPT/LDC test flags, before second iron run");
        $end_flag_hash_ref->{'wafl_iron_lazy_dist_checks'} = 0;
    }

    #
    # Delete 'preiron*' snapshot taken by iron at the end
    #
    $Tcd_Obj->step("Deleting aggr's snapshot created by iron");
    delete_aggregate_snapshots(
        command_interface => $aggr_task_obj->command_interface,
        aggregate         => $aggr_task_obj->aggregate,
        snapshot          => "preiron*",
        cmd_timeout       => $cmd_timeout,
    );

    #
    ## Delete volume corrupted snapshot before next iron run
    #
    if ( $vol_corr_snapname ne 'none' ) {

        if ( $vol_task_obj ) {
            my $tmp_vol_name = $vol_task_obj->volume;

            $Tcd_Obj->step("Deleting corrupted snapshot"
                . " \"$vol_corr_snapname\" in test volume :"
                . " \"$tmp_vol_name\" before second iron run");

            NACL::STask::VolumeSnapshot->purge(
                command_interface => $vol_task_obj->command_interface,
                snapshot          => $vol_corr_snapname,
                volume            => $vol_task_obj->volume,
                vserver           => $vol_task_obj->vserver,
                nacltask_if_purged => 'pass',
            );
        }
    }

    #
    ## Run wafliron
    #
    try {
        $Tcd_Obj->step("Run wafliron again");
        $iron_task_obj->start(
            nacltask_include_mirrors => 0,
            nacltask_optional_commit => 'false',
            nacltask_wait            => 1,
            nacltask_wafl_scan_speed => $max_scan_speed,
            nacltask_flags_to_set    => $end_flag_hash,
            'method-timeout'         => $iron_timeout,
        );
    } catch NACL::Exceptions::Timeout with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("Caught exception: " . $excpn_obj->text());
        $Tcd_Obj->nlog("Wait for some more time to complete iron");
        $iron_task_obj->wait_for_completion(
            'method-timeout' => $iron_timeout
        );
    };

    #
    ## Verify, no corruption is reported by iron
    #
    $Tcd_Obj->step("Verify, no corruption is reported by iron");
    if ( $iron_task_obj->corruption_found() ) {
        $Tcd_Obj->nlog("Corruption fixed by prev_cp is still reported by iron");
        my $messages = $iron_task_obj->get_wafliron_messages();
        $Tcd_Obj->nlog("Iron Messages : \n", Dumper  $messages);
        $Log->exit();
        return $TCD::FALSE;
    } else {
        $Tcd_Obj->nlog("Corruption fixed by prev_cp is not reported by iron");
    }

    $Log->exit();

    return $TCD::TRUE;
} # end of fix_by_prev_cp_verify_by_iron()


=head2 perform_aggregate_cleanup

Checks, if iron is running on given aggregate and wait for iron completion
and sets the aggregate state to online state.

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset         => node scope or 7-mode apiset
    aggr_task_obj  => aggregate task object
    iron_timeout   => timout used for iron completion [600]
    max_scan_speed => maximum scanner speed

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise throws an exception

=back

=cut

sub perform_aggregate_cleanup {

    $Log->enter();

    my %args_validn_spec =
        (
            'apiset'        =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode'
                },
            'aggr_task_obj' =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'iron_timeout'   => { type => SCALAR, default => 900 },
            'max_scan_speed' => { type => SCALAR, default => 10000 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj         = $options{'apiset'};
    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $iron_timeout       = $options{'iron_timeout'};
    my $max_scan_speed     = $options{'max_scan_speed'};

    #
    ## Check, if iron is running on the given aggregate
    #
    my @iron_cmp_list;

    @iron_cmp_list = NACL::C::Wafliron->find(
        command_interface => $aggr_task_obj->command_interface,
        filter => { aggregate => $aggr_task_obj->aggregate},
        allow_empty => 1,
    );

    if ( ! @iron_cmp_list ) {
       $Tcd_Obj->nlog("Iron is not running");
    } else { # else-1

        my $irn_c_obj = $iron_cmp_list[0]; # iron component obj

        my $status = $irn_c_obj->get_one_state_attribute('iron-status');

        if ( $status =~ /^(waiting_for_commit|not_running)$/ ) {
            $Tcd_Obj->nlog("Iron status : $status");
            $Tcd_Obj->nlog("Iron is not running");
        } else { # else-2
            $Tcd_Obj->nlog("Iron state is $status, Let's wait for completion");

            #
            ## Increase the iron speed
            #
            $Tcd_Obj->step("Set wafl scan speed to \"$max_scan_speed\"");
            get_set_wafl_scan_speed(
                {
                    action  => "set",
                    speed   => $max_scan_speed,
                    apiset  => $apiset_obj,
                });

            #
            ## Wait for iron completion
            #
            $Tcd_Obj->step("Waiting for iron completion");

            my $end_time = time() + $iron_timeout;
            my $iron_completed = 0;
            while ( time() < $end_time ) {
                $status = $irn_c_obj->get_one_state_attribute('iron-status');
                if ( $status eq "not_running" ) {
                    $Tcd_Obj->nlog("Iron has finished");
                    $iron_completed = 1;
                    last;
                } else {
                    $Tcd_Obj->nlog("Iron is in $status state");
                    $Tcd_Obj->nlog("Sleeping for 10 seconds and rechecking again");
                    sleep(10);
                }
            }

            if ( $iron_completed == 0 ) {
                $Log->exit();
                NATE::BaseException->throw("Iron has not completed in $iron_timeout seconds");
            }
        } # end else-2
    } # end else-1

    #
    ## Retrieve aggregate state and commit changes, if
    ## state is iron_restricted
    #
    $Tcd_Obj->step("Retrieving aggregate state");
    my $state =  $aggr_task_obj->get_one_state_attribute('state');

    if ( $state eq "iron_restricted" ) {
        $Tcd_Obj->nlog("Aggregate state is iron_restricted");
        $Tcd_Obj->nlog("We need to commit the changes, so commiting");

        NACL::C::Wafliron->commit(
            command_interface => $aggr_task_obj->command_interface,
            aggregate         => $aggr_task_obj->aggregate,
        );

        $Tcd_Obj->nlog("Let's wait for 20 seconds for "
            . " aggr to be offline after commit");
        sleep(20);

        # Retrieve aggregate state again
        $state =  $aggr_task_obj->state()->state();
    } elsif( $state eq "offline" ) {
        $Tcd_Obj->nlog("Aggregate state is offline");
    } elsif( $state eq "online" ) {
        $Tcd_Obj->nlog("Aggregate state is online");
    } elsif( $state eq "restricted" ) {
        $Tcd_Obj->nlog("Aggregate state is restricted");
    } else {
        $Tcd_Obj->nlog("Unexpected aggregate state : $state");
        $Log->exit();
        NATE::BaseException->throw("Unexpected aggregate state : $state");
    }

    #
    ## Bring aggregate to online state
    #
    if ( $state =~ /^(offline|restricted)$/ ) {
        $Tcd_Obj->step("Bringing aggregate in online state");
        $aggr_task_obj->modify(
            'state'           => 'online',
            'unmount-volumes' => 'true',
        );

        $Tcd_Obj->nlog("Let's wait for 5 seconds to allow aggr to be online");
        sleep(5);
    }

    $Log->exit();

    return $TCD::TRUE;
} # end of perform_aggregate_cleanup()

=head2 wait_until_volume_is_online

wait for volume to be online

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    vol_task_obj  => volume task object
    wait_timeout  => timout used for volume to be online [600]

=back

=item Returns:

throws an exception if not successful, otherwise return $TCD::TRUE

=back

=cut

sub wait_until_volume_is_online {

    $Log->enter();

    my %args_validn_spec =
        (
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'wait_timeout'      => { type => SCALAR, default => 600 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $vol_task_obj = $options{'vol_task_obj'};
    my $wait_timeout = $options{'wait_timeout'};
    my $vol_name     = $vol_task_obj->volume;

    #
    ## Wait for test volume to be online
    #
    $Tcd_Obj->nlog("Waiting for \"$vol_name\" volume to be online");

    $vol_task_obj->wait_for_completion(
        attribute_to_check        => 'state',
        till_value                => 'online',
        'method-timeout'          => $wait_timeout,
    );

    $Log->exit();

    return $TCD::TRUE;
} # end of wait_until_volume_is_online()

=head2 get_set_wafl_scan_speed

get or set wafl scan speed

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    speed         => speed to set [2000]
    action        => type of action ( get or set) [get]

=back

=item Returns:

return speed, when action is get,
return $TCD::TRUE, when action is set

=back

=cut

sub get_set_wafl_scan_speed {

    $Log->enter();

    my %args_validn_spec =
        (
           'apiset'     =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode'
                },
            'speed'     => { type    => SCALAR, default => 2000,},
            'action'    => { type    => SCALAR,
                             regex   => qr/^(get|set)$/,
                             default => "get",
                           },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj = $options{'apiset'};
    my $speed      = $options{'speed'};
    my $action     = $options{'action'};

    if ( $action eq "get" ) {
        #
        ## Retrieve wafl scan speed
        #
        my $response = $apiset_obj->wafl_scan_speed();
        $Log->exit();
        return $response->{'parsed_output'}->[0]->{'wafl_scan_speed'};

    } elsif ( $action eq "set" ) {
        #
        ## Set wafl scan speed
        #
        $Tcd_Obj->nlog("Setting wafl scan speed to \"$speed\"");
        $apiset_obj->wafl_scan_speed( speed => $speed );
        $Log->exit();
        return $TCD::TRUE;
    }
} # end of get_set_wafl_scan_speed()

=head2 does_user_file_exist

checks, if given user file exist using slist command

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    sio_file_name => file name used by filersio to read data
    op_inode      => scalar reference to return inode [optional]
    vol_task_obj  => volume task object

=back

=item Returns:

Returns  $TCD::TRUE if successful, otherwise $TCD::FALSE

=back

=cut

sub does_user_file_exist {

    $Log->enter();

    my %args_validn_spec =
        (
            'sio_file_name'     => { type => SCALAR },
            'op_inode'          => { type => SCALARREF, optional => 1 },
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $sio_file_name      = $options{'sio_file_name'};
    my $vol_task_obj       = $options{'vol_task_obj'};
    my $ret_value          = $TCD::TRUE;

    try {
        my $response = NACL::C::Slist->slist(
            path              => $sio_file_name,
            command_interface => $vol_task_obj->command_interface,
            )->get_parsed_output();
        my $inode = $response->[0]->{inode};
        $Tcd_Obj->nlog("File \"$sio_file_name\" exist with inode:$inode");

        if ( exists($options{'op_inode'}) ) {
            ${$options{'op_inode'}} = $inode;
        }
    } # end try
    catch NACL::APISet::Exceptions::InvalidParamValueException with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("slist command returns error on file \"$sio_file_name\"");
        $Tcd_Obj->nlog("Error string: " . $excpn_obj->text());
        $Tcd_Obj->nlog("Exception: ". ref($excpn_obj));
        $ret_value = $TCD::FALSE;
    };

    $Log->exit();

    return $ret_value;

} # end of does_user_file_exist()

=head2 set_mpt_test_flags

set default flags only required for MPT
or set default flags along with user's flag
or set flags only given by user

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset            => Apiset object
    set_defaults      => flag to set default flags [1]
    flag_hash_ref     => reference to flag hash

=back

=item Returns:

Returns none

=back

=cut

sub set_mpt_test_flags {

    $Log->enter();

    my %def_flag_hash = (
            'wafl_maxvvbn_disabled'           => 0,
            'wafl_iron_lazy_dist_checks'      => 1,
            'wafl_disable_sis_clone_fixup'    => 0,
            'wafl_lwr_trip'                   => 1,
            'wafl_64bit_upgrade_scan_disable' => 0,
            'wafl_snapiron_enabled'           => 0,
            'wafl_iron_defer_summary_check'   => 1,
            'wafl_iron_status_local'          => 0,
            'wafl_metadata_visible'           => 1,
        );

    my %args_validn_spec =
        (
            'flag_hash_ref' => { type => HASHREF, default => undef },
            'set_defaults'  => { type => SCALAR,  default => 1 },
            'apiset'        =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );
    my $apiset_obj     = $options{'apiset'};
    my $flag_hash_ref  = $options{'flag_hash_ref'};
    my $set_defaults   = $options{'set_defaults'};

    my $mpt_disabled   = $Global_Param_Set->get('MPT_DISABLED', default=>0);
    my $dsc_disabled   = $Global_Param_Set->get('DSC_DISABLED', default=>0);

    my %act_flag_hash;

    if ( $options{'flag_hash_ref'} ) {
        if ( 1 == $set_defaults ) {
            %act_flag_hash = (%def_flag_hash, %{$flag_hash_ref});
        } else {
            %act_flag_hash = %{$flag_hash_ref};
        }
    } elsif ( 1 == $set_defaults ) {
            %act_flag_hash = %def_flag_hash;
    } else {
        $Tcd_Obj->nlog("No flags to set, so returning");
        $Log->exit();
        return;
    }

    if ($mpt_disabled) {
        $Tcd_Obj->nlog("=> ** MPT is disabled ** <=");
        $act_flag_hash{'wafl_iron_lazy_dist_checks'} = 0;
    }

    if ($dsc_disabled) {
        $Tcd_Obj->nlog("=> ** DSC is disabled ** <=");
        $Tcd_Obj->nlog("=> ** MPT will also be disabled ** <=");
        $act_flag_hash{'wafl_iron_lazy_dist_checks'} = 0;
        $act_flag_hash{'wafl_iron_defer_summary_check'} = 0;
    }

    foreach my $flag ( keys %act_flag_hash ) {
        $Tcd_Obj->nlog("Setting flag: $flag "
            . " to $act_flag_hash{$flag}");

        $apiset_obj->setflag(
            flag =>  $flag,
            value => $act_flag_hash{$flag},
            'privilege-level' => "test",
        );
    }

    $Log->exit();
} # end of set_mpt_test_flags()

=head2 reset_default_mpt_test_flags

reset default flags required for MPT Testing

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset    => Apiset object

=back

=item Returns:

Returns none

=back

=cut

sub reset_default_mpt_test_flags {

    $Log->enter();

    my %def_flag_hash = (
            'wafl_maxvvbn_disabled'           => 0,
            'wafl_iron_lazy_dist_checks'      => 1,
            'wafl_disable_sis_clone_fixup'    => 0,
            'wafl_lwr_trip'                   => 1,
            'wafl_64bit_upgrade_scan_disable' => 0,
            'wafl_snapiron_enabled'           => 1,
            'wafl_iron_defer_summary_check'   => 1,
            'wafl_iron_status_local'          => 0,
            'wafl_metadata_visible'           => 0,
            'wafl_iron_panic_on_corruption'   => 1,
            'wafl_inconsistent_panic_testing' => 0,
        );

    my %args_validn_spec =
        (
            'apiset'  =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj = $options{'apiset'};

    foreach my $flag ( keys %def_flag_hash ) {
        $Tcd_Obj->nlog("Setting flag: $flag "
            . " to $def_flag_hash{$flag}");

        $apiset_obj->setflag(
            flag =>  $flag,
            value => $def_flag_hash{$flag},
            'privilege-level' => "test",
        );
    }

    $Log->exit();
} # end of reset_default_mpt_test_flags()

=head2 perform_cmode_validation

perform cmode validation and and push the following keys in input hash reference
    vserver_task_obj  => key used to push vserver task object
    old_aggr_list     => key used to push vserver old aggr list

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset            => Apiset object
    cmd_timeout       => Timeout for cli command [300]
    vserver           => Vserver name [vs0]
    command_interface => Node component object

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub perform_cmode_validation {

    $Log->enter();

    my %args_validn_spec =
        (
            'command_interface'     =>
                { type => OBJECT,
                  isa  => 'NACL::C::Node',
                },
            'vserver'     => { type => SCALAR, default => "vs0" },
            'cmd_timeout' => { type => SCALAR, default => 300 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $node        = $options{'command_interface'};
    my $vserver     = $options{'vserver'};
    my $cmd_timeout = $options{'cmd_timeout'};

    #
    ## Do all required checks, if we are running test in c-mode
    #
    my ($vserver_task_obj, @old_aggr_list);

    if ( $node->mode() eq 'CMode' ) {
        #
        ## Check, if vserver exists or not
        #
        $Tcd_Obj->step("Checking, if vserver:\"$vserver\" exist");
        $vserver_task_obj = NACL::STask::Vserver->find(
                                command_interface => $node,
                                filter      => { vserver => $vserver },
        );

        $test_data->{'vserver_task_obj'} = $vserver_task_obj;

        $Tcd_Obj->nlog("Vserver:\"$vserver\" already exist");

        $Tcd_Obj->step("Retrieving vserver aggr-list");
        @old_aggr_list = $vserver_task_obj->get_one_state_attribute('aggr-list');
        $Tcd_Obj->nlog("Vserver \"$vserver\" old aggr list: @old_aggr_list");
        $test_data->{'old_aggr_list'} = \@old_aggr_list;
    } else {
        $Tcd_Obj->nlog("Filer is in 7-Mode, so skipping CMode check..");
    }

    $Log->exit();

    return $TCD::TRUE;
} # end of perform_cmode_validation()

=head2 perform_iron_test_aggr_setup

sets up test aggregate and push the following keys in input hash reference
    aggr_task_obj    => key used to push aggr task object
    aggr_created_flg => key used to push flag, which tells, if aggregate
                        was created or not

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset            => Apiset object
    command_interface => Node component object
    vserver_task_obj  => Vserver task object
    aggregate         => test aggregate name
    node_mode         => node's mode
    node_name         => node's name
    cmd_timeout       => Timeout for cli command [600]
    aggr_reuse        => Flag to reuse aggregate [reuse]
    raid_type         => raid type [raid4]
    disc_count        => number of disc used for aggregate creation
    large_aggr        => flag to create larger aggregate[1]
    old_aggr_list     => Array reference to store vserver's old aggr list
    aggr_compliance   => flag to create compliance type aggregate

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub perform_iron_test_aggr_setup {

    $Log->enter();

    my %args_validn_spec =
        (
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'command_interface' =>
                { type => OBJECT,
                  isa  => 'NACL::C::Node',
                },
            'vserver_task_obj'  =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Vserver',
                  optional => 1,
                },
            'aggregate'        => { type => SCALAR },
            'node_mode'        => { type => SCALAR },
            'node_name'        => { type => SCALAR },
            'cmd_timeout'      => { type => SCALAR, default => 600 },
            'aggr_reuse'       => { type => SCALAR, default => 'reuse' },
            'raid_type'        => { type => SCALAR, default => 'raid4' },
            'disc_count'       => { type => SCALAR, default => 3 },
            'large_aggr'       => { type => SCALAR, default => 1 },
            'aggr_compliance'  => { type => SCALAR, default => 0 },
            'snaplock_type'    => { type => SCALAR, default => 'enterprise' },
            'old_aggr_list'    => { type => ARRAYREF, optional => 1 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj         = $options{'apiset'};
    my $node               = $options{'command_interface'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $aggr_reuse         = $options{'aggr_reuse'};
    my $raid_type          = $options{'raid_type'};
    my $disc_count         = $options{'disc_count'};
    my $aggr_name          = $options{'aggregate'};
    my $node_mode          = $options{'node_mode'};
    my $node_name          = $options{'node_name'};
    my $large_aggr         = $options{'large_aggr'};
    my $aggr_compliance    = $options{'aggr_compliance'};
    my $snaplock_type      = $options{'snaplock_type'};

    my ($aggr_task_obj, @old_aggr_list, $vserver, $vserver_task_obj);

    my (@new_aggr_list);

    my $aggr_created_flg = 0;

    my $add_aggr_in_vserver = 0;

    my $block_type = $large_aggr ? 64 : 32;

    my %aggr_options = ();

    if ( $options{'old_aggr_list'} ) {
        @old_aggr_list = @{$options{'old_aggr_list'}};
    }

    if ( $options{'vserver_task_obj'} ) {
        $vserver_task_obj   = $options{'vserver_task_obj'};
    }

    if ( $aggr_compliance ) {
        # TODO
        #$Tcd_Obj->nlog("Since test aggregate has to be compliance type");
        #$Tcd_Obj->nlog("We will not reuse the aggregate, if existing");
        #$aggr_reuse = 'purge';
        $aggr_options{'snaplock-type'} = $snaplock_type;

        $Tcd_Obj->step('Initialising snaplock clock');
        $apiset_obj->execute_command(
            command => 'snaplock clock initialize -f',
        );
    }

    if ( $node_mode eq "CMode" ) {

        $Tcd_Obj->nlog("Verify, if cmode optional params are given");
        if ( ! $vserver_task_obj ) {
            $Tcd_Obj->nlog("vserver_task_obj is not given");
            $Log->exit();
            return $TCD::CONFIG;
        } elsif ( ! @old_aggr_list ) {
            $Tcd_Obj->nlog("old_aggr_list is not given");
            $Log->exit();
            return $TCD::CONFIG;
        }

        # get vserver name
        $vserver = $vserver_task_obj->vserver;
    }

    #
    ## Create Aggregate, reuse if already exists
    #
    my $msg;
    if ( $aggr_reuse =~ m/^reuse$/i ) {
        $msg = "Creating aggregate, reuse it, if aggregate already exists";
    } else {
        $msg = "Creating aggregate, purge it, if aggregate already exists";
    }

    $Tcd_Obj->step("$msg");
    $aggr_task_obj = NACL::STask::Aggregate->create(
        command_interface => $node,
        aggregate         => $aggr_name,
        raidtype          => $raid_type,
        'block-type'      => $block_type,
        diskcount         => $disc_count,
        if_exists         => $aggr_reuse,
        _was_created      => \$aggr_created_flg,
        %aggr_options,
    );

    #
    # place aggr task obj in input hash ref test_data
    #
    $test_data->{'aggr_task_obj'}    = $aggr_task_obj;
    $test_data->{'aggr_created_flg'} = $aggr_created_flg;

    if ( 0 == $aggr_created_flg ) {
        $Tcd_Obj->nlog("Reusing the aggregate \"$aggr_name\" ");

        $Tcd_Obj->nlog("Checking, wafliron status on aggregate \"$aggr_name\"");
        perform_aggregate_cleanup( $test_data );

        my $state =  $aggr_task_obj->state()->state();

        if ( $state !~  m/online/ ) {
            $Tcd_Obj->nlog("Aggregate state != online, but $state");
            $Log->exit();
            return $TCD::CONFIG;
        }

        if ( $node_mode eq 'CMode' ) {

            #
            # Checking, test aggregate supports flexvol
            #
            $Tcd_Obj->step("Checking, test aggregate supports flexvol");

            my $vol_type  = $aggr_task_obj->get_one_state_attribute('voltype');

            if ( $vol_type != "flex" ) {
                $Tcd_Obj->nlog("Aggregate vol_type != \"flex\", but  \"$vol_type\"");
                $Log->exit();
                return $TCD::CONFIG;
            }

            #
            # Checking, test aggregate exists on home node
            #
            $Tcd_Obj->step("Checking, test aggregate exists on home node");
            my @aggr_nodes = $aggr_task_obj->get_one_state_attribute('nodes');

            if ( (1 == scalar(@aggr_nodes)) and
                 ($aggr_nodes[0] eq  $node_name) ) {
                $Tcd_Obj->nlog("Aggr \"$aggr_name\"exist on home node $aggr_nodes[0]");
            } else {
                $Tcd_Obj->nlog("Aggr \"$aggr_name\" exist on "
                . " $aggr_nodes[0] instead of home node $node_name");

                return $TCD::CONFIG;
            }

            #
            # Check, if aggreate exists in vserver aggr-list
            #
            $Tcd_Obj->nlog("Checking, aggr \"$aggr_name\" exists in "
                . "Vserver: \"$vserver\" aggr-list");

            if ( grep ( /$aggr_name/, @old_aggr_list ) ) {
                 $Tcd_Obj->nlog("Aggr \"$aggr_name\" exists in "
                    . "Vserver: \"$vserver\" aggr-list");
            } else {
                 $Tcd_Obj->nlog("Aggr \"$aggr_name\" has to be added in "
                    . "Vserver: \"$vserver\" aggr-list");
                $add_aggr_in_vserver = 1;
            }
        }
    } else {
        $Tcd_Obj->nlog("Aggregate is created");
        $add_aggr_in_vserver = 1;

        # Retrieve aggr-list again for case,
        # if aggregate already existed and purged
        # before recreation
        if ( $node_mode eq 'CMode' ) {
            $Tcd_Obj->step("Retrieving vserver aggr-list again");
            @old_aggr_list = $vserver_task_obj->get_one_state_attribute('aggr-list');
            $Tcd_Obj->nlog("Vserver \"$vserver\" old aggr list: @old_aggr_list");
            $test_data->{'old_aggr_list'} = \@old_aggr_list;
        }
    }

    #
    # Add test aggregate into the vserver's "aggr-list"
    # This is necessary for SN.0 onwards.
    #
    if ( $add_aggr_in_vserver && $node_mode eq 'CMode' ) {
        $Tcd_Obj->step("Adding Aggr \"$aggr_name\" in "
            . "Vserver: \"$vserver\" aggr-list");

        if ( $old_aggr_list[0] eq '-' ) {
            # aggr-list is empty
            @new_aggr_list = ($aggr_name);
        } else {
            @new_aggr_list = (@old_aggr_list, $aggr_name);
        }
        $vserver_task_obj->modify('aggr-list'=> \@new_aggr_list);
    } # end if

    $Log->exit();

    return $TCD::TRUE;
} # end of perform_iron_test_aggr_setup()

=head2 perform_iron_test_vol_setup

sets up test volume and push the following keys in input hash reference
    vol_task_obj    => key used to push volume task object
    vol_created_flg => key used to push flag, which tells, if volume
                        was created or not
=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    aggr_task_obj     => Aggregate task object
    volume            => test volume name
    vol_size          => test volume size
    guarantee         => test volume guarantee [volume]
    vserver           => Vserver name [vs0]
    vol_reuse         => Flag to reuse volume [purge]

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub perform_iron_test_vol_setup {

    $Log->enter();

    my %args_validn_spec =
        (
            'aggr_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'volume'    => { type => SCALAR },
            'vol_size'  => { type => SCALAR },
            'guarantee' => { type => SCALAR, default => 'volume' },
            'vserver'   => { type => SCALAR, default => 'vs0' },
            'vol_reuse' => { type => SCALAR, default => 'purge' },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $vserver           = $options{'vserver'};
    my $aggr_task_obj     = $options{'aggr_task_obj'};
    my $vol_reuse         = $options{'vol_reuse'};
    my $vol_name          = $options{'volume'};
    my $vol_size          = $options{'vol_size'};
    my $vol_guarantee     = $options{'guarantee'};
    my $snap_policy       = "none"; # c-mode built-in policy
    my $vol_created_flg   = 0;

    #
    ## Create a test volume
    #
    my $msg;
    if ( $vol_reuse =~ m/^reuse$/i ) {
        $msg = "Creating volume, reuse it, if volume already exists";
    } else {
        $msg = "Creating volume, purge it, if volume already exists";
    }

    $Tcd_Obj->step($msg);

    $Tcd_Obj->nlog("Volume name: \"$vol_name\", volume size: \"$vol_size\"");

    my $vol_task_obj = NACL::STask::Volume->create(
            command_interface => $aggr_task_obj->command_interface,
            aggregate         => $aggr_task_obj->aggregate,
            volume            => $vol_name,
            vserver           => $vserver,
            size              => $vol_size,
            'space-guarantee' => $vol_guarantee,
            'snapshot-policy' => $snap_policy,
            _was_created      => \$vol_created_flg,
            nacltask_if_exists  => $vol_reuse,
            nacltask_destroy_corrupt_volume => 1,
        );

    if ( 0 == $vol_created_flg ) {
        $Tcd_Obj->nlog("Reusing the volume \"$vol_name\" ");
    }

    #
    # place aggr task obj in input hash ref test_data
    #
    $test_data->{'vol_task_obj'}    = $vol_task_obj;
    $test_data->{'vol_created_flg'} = $vol_created_flg;

    $Log->exit();

    return $TCD::TRUE;
} # end of perform_iron_test_vol_setup()

=head2 change_iron_test_vol_property

change volume property

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset            => Apiset object
    vol_task_obj      => Volume task object
    volume            => Test volume name
    vserver           => Vserver name
    node_mode         => Node's mode
    sis_on            => set this flag to 1 to turn on sis [0]
    compr_on          => set this flag to 1 to turn on compression [0]

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub change_iron_test_vol_property {

    $Log->enter();

    my %args_validn_spec =
        (
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                  optional => 1,
                },
            'volume'    => { type => SCALAR },
            'vserver'   => { type => SCALAR },
            'node_mode' => { type => SCALAR },
            'sis_on'    => { type => SCALAR, default => $PARAM_NOT_SET,
                             regex => qr/^(0|1|$PARAM_NOT_SET)$/,
                           },
            'compr_on'  => { type => SCALAR, default => $PARAM_NOT_SET,
                             regex => qr/^(0|1|$PARAM_NOT_SET)$/,
                           },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $vserver           = $options{'vserver'};
    my $apiset_obj        = $options{'apiset'};
    my $vol_name          = $options{'volume'};
    my $node_mode         = $options{'node_mode'};
    my $sis_on            = $options{'sis_on'};
    my $compr_on          = $options{'compr_on'};
    my $vol_task_obj      = $options{'vol_task_obj'};

    # if user has not passed sis_on and compr_on option
    # set it based on default global param setting
    if ( $PARAM_NOT_SET == $sis_on ) {
        $sis_on = $Global_Param_Set->get('SIS_ON', default=>0);
    }

    if ( $PARAM_NOT_SET == $compr_on ) {
        $compr_on = $Global_Param_Set->get('COMPR_ON', default=>0);
    }

    #
    ## Change volume properties
    #
    if ( $node_mode eq "CMode" ) {

        $Tcd_Obj->step("No change in flexvol properties in c-mode");
    } else {
        $Tcd_Obj->step("Turning off auto snapshot in \"$vol_name\"");
        $Tcd_Obj->nlog("Node mode : $node_mode");
        my $resp_obj = $apiset_obj->vol_options(
            'volume'       => $vol_name,
            'option-name'  => 'nosnap',
            'option-value' => 'on',
        );
    }

    if ($sis_on||$compr_on) {
        $Tcd_Obj->step("Turning on sis on $vol_name flexvol");
        NACL::STask::VolumeEfficiency->on(
            command_interface => $vol_task_obj->command_interface,
            volume            => $vol_task_obj->volume,
            vserver           => $vserver,
            ignore_enabled    => 1,
        );
    }

    if ($compr_on) {
        $Tcd_Obj->step("Turning on compression on $vol_name flexvol");
        NACL::STask::VolumeEfficiency->config(
            command_interface => $vol_task_obj->command_interface,
            volume            => $vol_task_obj->volume,
            vserver           => $vserver,
            'background-compression'=> 'true',
            compression             => 'true',
            'inline-compression'    => 'true',
        );
    }

    $Log->exit();

    return $TCD::TRUE;
} # end of change_iron_test_vol_property()

=head2 change_iron_test_aggr_property

change aggregate property

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset            => Apiset object
    aggr_task_obj     => Aggregate task object [optional]
    aggregate         => Test aggregate name
    node_mode         => Node's mode

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub change_iron_test_aggr_property {

    $Log->enter();

    my %args_validn_spec =
        (
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'aggr_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                  optional => 1,
                },
            'aggregate'  => { type => SCALAR },
            'node_mode'  => { type => SCALAR },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj        = $options{'apiset'};
    my $aggr_name         = $options{'aggregate'};
    my $node_mode         = $options{'node_mode'};

    #
    ## Change aggregate properties
    #
    $Tcd_Obj->step("Turning off auto snapshot in \"$aggr_name\"");
    $Tcd_Obj->nlog("Node mode : $node_mode");
    my $resp_obj = $apiset_obj->aggr_options(
        'aggregate'    => $aggr_name,
        'option-name'  => 'nosnap',
        'option-value' => 'on'
    );

    $Log->exit();

    return $TCD::TRUE;
} # end of change_iron_test_aggr_property()


=head2 perform_iron_task_setup

sets up iron task and push following keys in input hash reference
    iron_task_obj => key used to push iron task object

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    aggr_task_obj    => Aggregate task object

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub perform_iron_task_setup {

    $Log->enter();

    my %args_validn_spec =
        (
            'aggr_task_obj'     =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $aggr_task_obj     = $options{'aggr_task_obj'};

    #
    ## Initialise the iron task obj
    #
    $Tcd_Obj->step("Initialising the iron task obj");
    my $iron_task_obj = NACL::MTask::Wafliron->new(
        command_interface => $aggr_task_obj->command_interface,
        aggregate         => $aggr_task_obj,
    );

    #
    # place aggr task obj in input hash ref test_data
    #
    $test_data->{'iron_task_obj'} = $iron_task_obj;

    $Log->exit();

    return $TCD::TRUE;
} # end of perform_iron_task_setup()

=head2 perform_iron_test_setup

sets up test resources for iron testing

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    command_interface => Node component object
    apiset            => Apiset object
    node_mode         => Node's mode
    node_name         => Node's name
    aggregate         => Test aggregate name
    aggr_reuse        => Flag to reuse aggregate [reuse]
    raid_type         => raid type [raid4]
    disc_count        => number of disc used for aggregate creation
    large_aggr        => flag to create larger aggregate[1]
    aggr_compliance   => flag to create compliance type aggregate
    volume            => Test volume name
    vol_size          => Test volume size
    guarantee         => Test volume guarantee [volume]
    vserver           => Vserver name [vs0]
    vol_reuse         => Flag to reuse volume [purge]
    flag_hash_ref     => reference to flag hash [undef]
    cmd_timeout       => Timeout for cli command execution [300]
    vol_cfg_list      => array ref to list of volumes to be configured

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise exception will be thrown

=back

=cut

sub perform_iron_test_setup {

    $Log->enter();

    my %args_validn_spec =
        (
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'command_interface' =>
                { type => OBJECT,
                  isa  => 'NACL::C::Node',
                },
            'node_mode'        => { type => SCALAR },
            'node_name'        => { type => SCALAR },
            'aggregate'        => { type => SCALAR },
            'aggr_reuse'       => { type => SCALAR, default => 'reuse' },
            'raid_type'        => { type => SCALAR, default => 'raid4' },
            'disc_count'       => { type => SCALAR, default => 3 },
            'large_aggr'       => { type => SCALAR, default => 1 },
            'aggr_compliance'  => { type => SCALAR, default => 0 },
            'volume'           => { type => SCALAR|UNDEF, optional => 1 },
            'vol_size'         => { type => SCALAR|UNDEF, optional => 1 },
            'guarantee'        => { type => SCALAR, default => 'volume' },
            'vserver'          => { type => SCALAR, default => 'vs0' },
            'vol_reuse'        => { type => SCALAR, default => 'purge' },
            'flag_hash_ref'    => { type => HASHREF, default => undef },
            'cmd_timeout'      => { type => SCALAR, default => 300 },
            'vol_cfg_list'     => { type => ARRAYREF, optional => 1 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $test_data = $_[0];
    my $ret_code;

    #
    # Configure common feature licenses
    #
    $Tcd_Obj->step("Configure common feature licenses");
    my $lic_stask_list_ref = configure_feature_licenses( $test_data );

    #
    # Save license task list in test_data
    #
    $test_data->{'lic_stasks_list'} = $lic_stask_list_ref;

    #
    # Set MPT Test flags
    #
    $Tcd_Obj->step("Setting up required test flags");
    set_mpt_test_flags( $test_data );

    #
    # Retrieve wafl scan speed
    #
    $test_data->{'action'} = "get";
    $test_data->{'old_scan_speed'} = get_set_wafl_scan_speed( $test_data );
    $Tcd_Obj->nlog("Old scan speed : $test_data->{'old_scan_speed'}");

    #
    # Set scan speed to system default
    #
    my $default_scan_speed = 1300;
    if ( $test_data->{'old_scan_speed'} < $default_scan_speed ) {
        $Tcd_Obj->step("Setting scan speed to default: $default_scan_speed");
        $test_data->{'action'} = "set";
        $test_data->{'speed'}  = $default_scan_speed;
        get_set_wafl_scan_speed( $test_data );
    }

    #
    # Perform cmode validation
    #
    $Tcd_Obj->step("Performing cmode validation");
    $ret_code = perform_cmode_validation( $test_data );

    if ( $TCD::TRUE != $ret_code ) {
        $Tcd_Obj->nlog("perform_cmode_validation() failed");
        $Log->exit();
        return $ret_code;
    }

    #
    # Perform aggregate setup
    #
    $Tcd_Obj->step("Performing aggregate setup");
    $ret_code = perform_iron_test_aggr_setup( $test_data );

    if ( $TCD::TRUE != $ret_code ) {
        $Tcd_Obj->nlog("perform_iron_test_aggr_setup() failed");
        $Log->exit();
        return $ret_code;
    }

    #
    # Change aggregate properties
    #
    $Tcd_Obj->step("Setting aggregate properties");
    $ret_code = change_iron_test_aggr_property( $test_data );

    if ( $TCD::TRUE != $ret_code ) {
        $Tcd_Obj->nlog("change_iron_test_aggr_property() failed");
        $Log->exit();
        return $ret_code;
    }

    #
    # Perform volume setup
    #
    if ( $options{'vol_cfg_list'} ) {
        $Tcd_Obj->step("configuring multiple flexvols");

        my @vol_cfg_list = @{$options{'vol_cfg_list'}};
        my %vol_task_list;

        foreach my $vol_info ( @vol_cfg_list ) {
            my %temp_vol_info = %{$vol_info};

            $temp_vol_info{'aggr_task_obj'} = $test_data->{'aggr_task_obj'};

            $ret_code = perform_iron_test_vol_setup( \%temp_vol_info );

            if ( $TCD::TRUE != $ret_code ) {
                $Tcd_Obj->nlog("perform_iron_test_vol_setup() failed");
                $Log->exit();
                return $ret_code;
            }

            $vol_task_list{$temp_vol_info{'volume'}} =
                    { 'vol_task_obj' => $temp_vol_info{'vol_task_obj'},
                      'vol_created_flg' => $temp_vol_info{'vol_created_flg'},
                    };
            #
            # Change volume properties
            #
            $Tcd_Obj->step("Setting volume properties");
            $ret_code = change_iron_test_vol_property(
                    apiset => $test_data->{'apiset'},
                    node_mode => $test_data->{'node_mode'},
                    %temp_vol_info,
                );

            if ( $TCD::TRUE != $ret_code ) {
                $Tcd_Obj->nlog("change_iron_test_vol_property() failed");
                $Log->exit();
                return $ret_code;
            }
        } #end foreach

        # save vol stask objects in test_data
        $test_data->{'vol_stasks_list'} = \%vol_task_list;

    } else {
        $Tcd_Obj->step("Performing single volume setup");
        $ret_code = perform_iron_test_vol_setup( $test_data );

        if ( $TCD::TRUE != $ret_code ) {
            $Tcd_Obj->nlog("perform_iron_test_vol_setup() failed");
            $Log->exit();
            return $ret_code;
        }

        #
        # Change volume properties
        #
        $Tcd_Obj->step("Setting volume properties");
        $ret_code = change_iron_test_vol_property( $test_data );

        if ( $TCD::TRUE != $ret_code ) {
            $Tcd_Obj->nlog("change_iron_test_vol_property() failed");
            $Log->exit();
            return $ret_code;
        }
    }

    #
    # Perform iron task setup
    #
    $Tcd_Obj->step("Performing iron task setup");
    $ret_code = perform_iron_task_setup( $test_data );

    if ( $TCD::TRUE != $ret_code ) {
        $Tcd_Obj->nlog("perform_iron_task_setup() failed");
        $Log->exit();
        return $ret_code;
    }

    $Log->exit();

    return $TCD::TRUE;
} # end of perform_iron_test_setup()

=head2 perform_iron_test_cleanup

clean up test resources

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset             => Apiset object
    vserver_task_obj   => vserver   task object
    clone_vol_task_obj => clone     task object
    vol_task_obj       => volume    task object
    aggr_task_obj      => aggregate task object
    cleanup            => cleanup value [test]
    old_scan_speed     => old scan speed
    aggr_created_flg   => aggregate created flag
    node_mode          => node's mode
    old_aggr_list      => vserver old aggregate list
    lic_stasks_list    => reference to array having license stask objects
    vol_created_flg    => volume created flag
    vol_stasks_list    => hash ref to volume stask's information

=back

=item Returns:

Returns $TCD::TRUE

=back

=cut

sub perform_iron_test_cleanup {

    $Log->enter();

    my %args_validn_spec =
        (
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'vserver_task_obj'  =>
                { type    => OBJECT,
                  isa     => 'NACL::STask::Vserver',
                  default => undef,
                },
            'clone_vol_task_obj'    =>
                { type    => OBJECT,
                  isa     => 'NACL::STask::VolumeClone',
                  default => undef,
                },
            'vol_task_obj'      =>
                { type    => OBJECT,
                  isa     => 'NACL::STask::Volume',
                  default => undef,
                },
            'aggr_task_obj'     =>
                { type    => OBJECT,
                  isa     => 'NACL::STask::Aggregate',
                  default => undef,
                },
            'cleanup'          => { type => SCALAR|UNDEF,   default => 'test'},
            'old_scan_speed'   => { type => SCALAR|UNDEF,   default => 1300 },
            'aggr_created_flg' => { type => SCALAR|UNDEF,   default => 0 },
            'vol_created_flg'  => { type => SCALAR|UNDEF,   default => 1 },
            'node_mode'        => { type => SCALAR|UNDEF,   default => 'none' },
            'old_aggr_list'    => { type => ARRAYREF|UNDEF, optional => 1},
            'lic_stasks_list'  => { type => ARRAYREF|UNDEF, optional => 1},
            'vol_stasks_list'   => { type => HASHREF|UNDEF,  optional => 1},
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj         = $options{'apiset'};
    my $vserver_task_obj   = $options{'vserver_task_obj'};
    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $vol_task_obj       = $options{'vol_task_obj'};
    my $clone_vol_task_obj = $options{'clone_vol_task_obj'};
    my $cleanup            = $options{'cleanup'};
    my $old_scan_speed     = $options{'old_scan_speed'};
    my $node_mode          = $options{'node_mode'};
    my $aggr_created_flg   = $options{'aggr_created_flg'};
    my $vol_created_flg    = $options{'vol_created_flg'};

    my (@old_aggr_list, @lic_stasks_list);
    my ($vol_name, $aggr_name);

    if ( $options{'old_aggr_list'} ) {
        @old_aggr_list = @{$options{'old_aggr_list'}};
    } else {
        $Tcd_Obj->nlog("Vserver old aggr list is not given");
        undef(@old_aggr_list);
    }

    if ( $options{'lic_stasks_list'} ) {
        @lic_stasks_list = @{$options{'lic_stasks_list'}};
    } else {
        $Tcd_Obj->nlog("License stask objects list is not given");
        undef(@lic_stasks_list);
    }

    #
    # Stop filersio ops
    #
    try {
        $Tcd_Obj->step("Stopping filersio");
        my $resp_obj = $apiset_obj->filersio_stop(
             'instance-id' => " ", );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $excpn_obj = shift;
        $Tcd_Obj->nlog( "Filersio status is:: " . $excpn_obj->text() );
    };

    #
    # Perform iron cleanup
    #
    $Tcd_Obj->step("Performing iron cleanup on test aggregate, if required");
    if ( $aggr_task_obj ) {
        $aggr_name = $aggr_task_obj->aggregate;
        perform_aggregate_cleanup( $test_data );
        $Tcd_Obj->step("Iron cleanup on aggr \"$aggr_name\" is completed");
    } else {
        $Tcd_Obj->nlog("Aggregate task obj is not initialised");
        $Tcd_Obj->nlog("Skipping iron cleanup on aggregate");
    }

    #
    # Reset wafl scan speed
    #
    $Tcd_Obj->step("Restoring scan speed to old value");
    $test_data->{'action'} = "set";
    $test_data->{'speed'}  = $old_scan_speed;
    get_set_wafl_scan_speed( $test_data );

    #
    # reset test flags
    #
    $Tcd_Obj->step("Resetting mpt test flags");
    reset_default_mpt_test_flags( $test_data );

    if ( $cleanup !~ /^(none)$/i ) {

        # Delete clone volume
        $Tcd_Obj->step("Performing clone volume cleanup..");
        if ( $clone_vol_task_obj ) {
            $vol_name = $clone_vol_task_obj->volume;
            $Tcd_Obj->nlog("Deleting clone volume: $vol_name");
            $clone_vol_task_obj->purge();
            $Tcd_Obj->nlog("Clone volume is deleted");
        } else {
            $Tcd_Obj->nlog("Skipping clone volume deletion");
            $Tcd_Obj->nlog("Clone volume task obj is not initialised");
        }

        # Delete test volume
        $Tcd_Obj->step("Performing test volume cleanup..");

        if ( $options{'vol_stasks_list'} ) {
            my %vol_info = %{$options{'vol_stasks_list'}};
            my ($vol_obj, $vol_was_created);
            foreach my $vol_name ( keys %vol_info ) {
                $vol_obj = $vol_info{$vol_name}->{'vol_task_obj'};
                $vol_was_created = $vol_info{$vol_name}->{'vol_created_flg'};
                if ( $vol_was_created ) {
                    $Tcd_Obj->nlog("Deleting test volume: $vol_name");
                    $vol_obj->purge(nacltask_destroy_corrupt_volume => 0);
                    $Tcd_Obj->nlog("Volume is deleted");
                } else {
                    $Tcd_Obj->nlog("Test volume: $vol_name was reused");
                }
            }
        } else {
            if ( $vol_task_obj && $vol_created_flg ) {
                $vol_name = $vol_task_obj->volume;
                $Tcd_Obj->nlog("Deleting test volume: $vol_name");
                $vol_task_obj->purge(nacltask_destroy_corrupt_volume => 0);
                $Tcd_Obj->nlog("Volume is deleted");
            } else {
                $Tcd_Obj->nlog("Skipping volume deletion");
                $Tcd_Obj->nlog("Either volume task obj is not initialised");
                $Tcd_Obj->nlog("Or volume was not created by the setup()");
            }
        }

        # Restore vserver aggr-list
        if ( $node_mode eq 'CMode' ) {
            $Tcd_Obj->step('Performing vserver cleanup..');
            if ( $vserver_task_obj && @old_aggr_list ) {
                my $vserver   = $vserver_task_obj->vserver;
                $Tcd_Obj->nlog("Restoring old aggr list for vserver \"$vserver\"");
                $vserver_task_obj->modify( 'aggr-list' => \@old_aggr_list);
            } else {
                $Tcd_Obj->nlog("Skipping vserver aggr-list restoration");
            }
        } else {
            $Tcd_Obj->nlog("Node mode is not CMode");
            $Tcd_Obj->nlog("So vserver cleanup will not be done");
        }

        # Delete test aggregate
        $Tcd_Obj->step("Performing test aggregate cleanup..");
        if( $aggr_task_obj && $aggr_created_flg && ($cleanup !~ /volume/i) ) {
            $Tcd_Obj->nlog("Deleting test aggr: $aggr_name");
            $aggr_task_obj->purge();
            $Tcd_Obj->nlog("Aggregate is deleted");
        } else {
            $Tcd_Obj->nlog("Skipping aggregate deletion");
            $Tcd_Obj->nlog("Either aggregate task obj is not initialised");
            $Tcd_Obj->nlog("Or aggregate was not created by the setup()");
            $Tcd_Obj->nlog("Or cleanup was set to $cleanup");
        }

        # Delete license, which were configured during testing
        $Tcd_Obj->step("Performing license code cleanup..");
        if ( @lic_stasks_list ) {
            foreach my $stask ( @lic_stasks_list ) {
                my $feature = $stask->$Lic_Stask_Attribute;
                $Tcd_Obj->nlog("Cleaning up license for feature: $feature");
                if ( $stask->_was_added() ) {
                    $stask->purge( nacltask_verify => 1 );
                    $Tcd_Obj->nlog("\t => License for feature::$feature was removed");
                } else {
                    $Tcd_Obj->nlog("\t => License for feature::$feature was not added");
                }
            }
        }
    } else {
        $Tcd_Obj->nlog("Skipping cleanup, as cleanup is set to \"$cleanup\"");
    }

    $Log->exit();

    return $TCD::TRUE;
}

=head2 mpt_run_iron

run combinations of iron and ioc based on value of test_method
if test_method == IRON, run iron and then ioc
if test_method == IOC,  run ioc  and then iron

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    test_method  => test method [iron]

=back

=item Returns:

Returns $TCD::PASS if successful, otherwise $TCD::FAIL

=back

=cut

sub mpt_run_iron {

    $Log->enter();

    my %args_validn_spec =
        (
            'test_method' => { type => SCALAR, default => "iron"},
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $test_method = $options{'test_method'};

    my $ret_code;

    if ( $test_method =~m/^IRON$/i ) {
        $Tcd_Obj->step("Fixing corrupion by iron and verifying by IOC");
        $ret_code = fix_by_iron_verify_by_ioc( $test_data );
    } elsif ( $test_method =~m/^IOC$/i ) {
        $Tcd_Obj->step("Fixing corrupion by IOC and verifying by iron");
        $ret_code = fix_by_ioc_verify_by_iron( $test_data );
    } elsif ( $test_method =~m/^PREVCP_IRON$/i ) {
        $Tcd_Obj->step("Fixing corrupion by prev_cp and verifying by iron");
        $ret_code = fix_by_prev_cp_verify_by_iron( $test_data );
    }

    $Log->exit();

    return ($ret_code == $TCD::TRUE ? $TCD::PASS : $TCD::FAIL);
}

=head2 validate_mpt_test_usr_inputs

validate user inputs

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    test_method       => Test method
    raid_type         => Raid type
    aggr_size         => Aggregate size
    large_aggr        => flag to create large aggregate
    snaplock_type     => snaplock type
=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise $TCD::FALSE

=back

=cut

sub validate_mpt_test_usr_inputs {

    $Log->enter();

    my %args_validn_spec =
        (
            'test_method'       => { type => SCALAR, default => 'IRON'},
            'raid_type'         => { type => SCALAR, default => 'raid4'},
            'aggr_size'         => { type => SCALAR, default => 4},
            'mark_pseudobad'    => { type => SCALAR, default => 0},
            'mark_unverified'   => { type => SCALAR, default => 0},
            'large_aggr'        => { type => SCALAR, default => 1},
            'snaplock_type'     => { type => SCALAR, default => 'enterprise'},
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $test_method      = $options{'test_method'};
    my $raid_type        = $options{'raid_type'};
    my $aggr_size        = $options{'aggr_size'};
    my $mark_pseudobad   = $options{'mark_pseudobad'};
    my $mark_unverified  = $options{'mark_unverified'};
    my $large_aggr       = $options{'large_aggr'};
    my $snaplock_type    = $options{'snaplock_type'};

    #
    # validate test_method : TEST_METHOD
    #
    if ($test_method !~ /^(IRON|IOC)$/i ) {
        $Tcd_Obj->nlog("Wrong choice given for method");
        $Tcd_Obj->nlog("Please give \"IRON|IOC|iron|ioc\" ");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    # validate raid_type : RAID_TYPE
    #
    if ( $raid_type !~ /(raid4|raid_dp)/ ) {
        $Tcd_Obj->nlog("RAID_TYPE field should be \"raid4|raid_dp\"");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    # validate number of disk : AGGR_SIZE
    #
    my $raid_dp_min_disk_count = 3;
    if ( $raid_type eq "raid_dp" ) {
        if ( $aggr_size < $raid_dp_min_disk_count) {
            $Tcd_Obj->nlog("Minimun $raid_dp_min_disk_count disks"
                . " are required for raid_dp");
            $Log->exit();
            return $TCD::FALSE;
        }
    }

    #
    # validate pseudobad flag : MARK_PSEUDOBAD
    #
    if ( $mark_pseudobad !~ /(0|1)/ ) {
        $Tcd_Obj->nlog("mark_pseudobad != (0|1)");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    # validate unverified flag : MARK_UNVERIFIED
    #
    if ( $mark_unverified !~ /(0|1)/ ) {
        $Tcd_Obj->nlog("mark_unverified != (0|1)");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    # validate pseudobad & unverified flag, both are not true
    #
    if ( $mark_pseudobad and $mark_unverified ) {
        $Tcd_Obj->nlog("Both mark_pseudobad & mark_unverified can't be 1");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    # validate large_aggr flag : LARGE_AGGR
    #
    if ( $large_aggr !~ /(0|1)/ ) {
        $Tcd_Obj->nlog("large_aggr != (0|1)");
        $Log->exit();
        return $TCD::FALSE;
    }

    #
    # validate snaplock_type : SNAPLOCK_TYPE
    #
    if ( $snaplock_type !~ /(compliance|enterprise)/ ) {
        $Tcd_Obj->nlog("SNAPLOCK_TYPE field should be \"compliance|enterprise\"");
        $Log->exit();
        return $TCD::FALSE;
    }

    $Log->exit();

    return $TCD::TRUE;
}

=head2 initialise_mpt_test_common_params

initialise common parameters required for MPT testing

=over 4

=item Arguments:

=over 4

    None

=back

=item Returns:

    None

=back

=cut

sub initialise_mpt_test_common_params {

    $Log->enter();

    # Initialise MPT specific params setting
    IronUtil::MPT::commonParams->new(set_mpt_defaults => 1);

    # Initialise default params setting
    FS::commonParams->new(set_legacy_defaults => 1);

    $Log->exit();

    return $TCD::TRUE;
}

=head2 volume_snap_create_in_background

starts subtest to perform snapshot creation ops

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    vol_task_obj      => volume task object
    snapshot_count    => number of snapshots [255]
    snapshot_interval => delay between two snapshot creation [5]
    snap_crt_prefix   => snapshot prefix used for snapshot creation [vol_snap_]
    snap_start_index  => snapshot start index [1]
    cmd_timeout       => timeout used for snapshot creation [300]

=back

=item Returns:

Returns a reference to hash having keys: subtest_id', 'run_id'

=back

=cut

sub volume_snap_create_in_background {

    $Log->enter();

    my %args_validn_spec =
        (
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'snapshot_count'    => { type => SCALAR,  default => 255 },
            'snapshot_interval' => { type => SCALAR,  default => 5 },
            'snap_crt_prefix'   => { type => SCALAR,  default => 'vol_snap_' },
            'snap_start_index'  => { type => SCALAR,  default => 1 },
            'cmd_timeout'       => { type => SCALAR,  default => 300 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $vol_task_obj       = $options{'vol_task_obj'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $count              = $options{'snapshot_count'};
    my $interval           = $options{'snapshot_interval'};
    my $snap_crt_prefix    = $options{'snap_crt_prefix'};
    my $start_index        = $options{'snap_start_index'};
    my $volume             = $vol_task_obj->volume ;
    my $run_id             = "$volume" . "_snapshot_create_" . time();
    my $vserver            = 'none';

    if ($vol_task_obj->command_interface->mode() eq 'CMode' ) {
        $vserver = $vol_task_obj->vserver;
    }

    $Tcd_Obj->nlog("Starting snap create subtest with id : $run_id");
    my $test_id =
        subtest(
            sub {
                my $start_id = 1;
                my $name;

                while($start_id <= $count ) {
                    $name = "$snap_crt_prefix$start_index";
                    logcomment("Creating snapshot: $name on volume: $volume");
                    NACL::STask::VolumeSnapshot->create(
                        command_interface  => $vol_task_obj->command_interface,
                        volume             => $volume,
                        vserver            => $vserver,
                        snapshot           => $name,
                        nacltask_wait      => 1,
                        nacltask_verify    => 0,
                        nacltask_if_exists => 'reuse',
                        'method-timeout'   => $cmd_timeout,
                     );
                    sleep($interval);
                    $start_id++;
                    $start_index++;
                }
            },
            -bg,
            -runid, $run_id,
            -onexit, sub { $Tcd_Obj->nlog("Snapshot creation subtest exiting.."); },
        );

    sleep(2); # let's wait for 2 sec, before another call is made

    $Log->exit();

    return ( { 'subtest_id' => $test_id, 'run_id' => $run_id } );

} # end of volume_snap_create_in_background()

=head2 volume_snap_delete_in_background

starts subtest to perform snapshot delete ops

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    vol_task_obj      => volume    task object
    snapshot_count    => number of snapshots [255]
    snapshot_interval => delay between two snapshot deletion [5]
    snap_del_prefix   => snapshot prefix used for snapshot deletion [vol_snap_]
    snap_start_index  => snapshot start index [1]
    cmd_timeout       => timeout used for snapshot creation [300]

=back

=item Returns:

Returns a reference to hash having keys: 'subtest_id', 'run_id'

=back

=cut

sub volume_snap_delete_in_background {

    $Log->enter();

    my %args_validn_spec =
        (
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'snapshot_count'    => { type => SCALAR,  default => 255 },
            'snapshot_interval' => { type => SCALAR,  default => 5 },
            'snap_del_prefix'   => { type => SCALAR,  default => 'vol_snap_' },
            'snap_start_index'  => { type => SCALAR,  default => 1 },
            'cmd_timeout'       => { type => SCALAR,  default => 300 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $vol_task_obj       = $options{'vol_task_obj'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $count              = $options{'snapshot_count'};
    my $interval           = $options{'snapshot_interval'};
    my $snap_del_prefix    = $options{'snap_del_prefix'};
    my $start_index        = $options{'snap_start_index'};
    my $volume             = $vol_task_obj->volume ;
    my $vserver            = 'none';
    my $run_id             = "$volume" . "_snapshot_delete_" . time();

    if ($vol_task_obj->command_interface->mode() eq 'CMode' ) {
        $vserver = $vol_task_obj->vserver;
    }

    $Tcd_Obj->nlog("Starting snap delete subtest with id : $run_id");

    my $test_id =
        subtest(
            sub {
                my $start_id = 1;
                my $name;

                while($start_id <= $count ) {
                    $name = "$snap_del_prefix$start_index";
                    logcomment("Deleting snapshot: $name on volume $volume");
                    NACL::STask::VolumeSnapshot->purge(
                        command_interface  => $vol_task_obj->command_interface,
                        volume             => $volume,
                        vserver            => $vserver,
                        snapshot           => $name,
                        nacltask_wait      => 1,
                        nacltask_if_purged => 'pass',
                        'method-timeout'   => $cmd_timeout,
                     );
                    sleep($interval);
                    $start_id++;
                    $start_index++;
                }
            },
            -bg,
            -runid, $run_id,
            -onexit, sub { $Tcd_Obj->nlog("Snapshot deletion subtest exiting.."); },
        );

    sleep(2); # let's wait for 2 sec, before another call is made

    $Log->exit();

    return ( { 'subtest_id' => $test_id, 'run_id' => $run_id } );

} # end of volume_snap_delete_in_background()

=head2 file_clones_create_in_background

starts subtest to perform file clone creation ops

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    vol_task_obj      => volume task object
    src_dir_path      => path to source directory
    dst_dir_path      => path to destination directory
    src_file_list     => referece to array having list of source files
    dst_file_list     => referece to array having list of destination files
    clone_interval    => delay between two clone ops [1]
    cmd_timeout       => timeout used for file clone creation [300]

=back

=item Returns:

Returns a reference to hash having keys: subtest_id', 'run_id'

=back

=cut

sub file_clones_create_in_background {

    $Log->enter();

    my %args_validn_spec =
        (
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'src_dir_path'      => { type => SCALAR, },
            'dst_dir_path'      => { type => SCALAR, },
            'src_file_list'     => { type => ARRAYREF, },
            'dst_file_list'     => { type => ARRAYREF, },
            'clone_interval'    => { type => SCALAR, default => 1 },
            'cmd_timeout'       => { type => SCALAR, default => 300 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $vol_task_obj       = $options{'vol_task_obj'};
    my $src_dir_path       = $options{'src_dir_path'};
    my $dst_dir_path       = $options{'dst_dir_path'};
    my @src_file_list      = @{$options{'src_file_list'}};
    my @dst_file_list      = @{$options{'dst_file_list'}};
    my $clone_interval     = $options{'clone_interval'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $volume             = $vol_task_obj->volume ;
    my $run_id             = "$volume" . "_file_clone_create_" . time();
    my $clone_count        = scalar(@dst_file_list);
    my $vserver            = 'none';

    if ($vol_task_obj->command_interface->mode() eq 'CMode' ) {
        $vserver = $vol_task_obj->vserver;
    }

    if (  $clone_count == scalar (@src_file_list) ) {
        $Tcd_Obj->nlog("Creating 1(src)<->1(dst) $clone_count file clones");
    } elsif ( 1 == scalar (@src_file_list) ) {
        $Tcd_Obj->nlog("Creating 1(src)<->N(dst) $clone_count file clones");
        @src_file_list = (($src_file_list[0]) x $clone_count );
    } else {
        $Tcd_Obj->nlog("No of elements are not same in src and dst file list");
        $Log->exit();
        NATE::BaseException->throw("No of elements are not same in src and dst file list");
    }
    $Tcd_Obj->nlog("Starting file clone create subtest with id : $run_id");
    my $test_id =
        subtest(
            sub {
                my $start_id = 1;
                my ($clone_src, $clone_dst, $src_file, $dst_file);

                while($start_id <= $clone_count ) {
                    $src_file = shift @src_file_list;
                    $dst_file = shift @dst_file_list;
                    $clone_src = "$src_dir_path/$src_file";
                    $clone_dst = "$dst_dir_path/$dst_file";
                    logcomment("Creating file clone, "
                        . "src: $clone_src, dst:$clone_dst on volume: $volume");
                    NACL::C::VolumeFileClone->create(
                        command_interface  => $vol_task_obj->command_interface,
                        vserver            => $vserver,
                        volume             => $volume,
                        'source-path'      => $clone_src,
                        'destination-path' => $clone_dst,
                        'clone_type'       => "file",
                        'method-timeout'   => $cmd_timeout,
                    );
                    sleep($clone_interval);
                    $start_id++;
                }
            },
            -bg,
            -runid, $run_id,
            -onexit, sub { $Tcd_Obj->nlog("File clone creation subtest exiting.."); },
        );

    sleep(2); # let's wait for 2 sec, before another call is made

    $Log->exit();

    return ( { 'subtest_id' => $test_id, 'run_id' => $run_id } );

} # end of file_clones_create_in_background()

=head2 stop_subtests

stops subtests based on user input

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    subtest_id  => reference to hash having subtest id or
                   reference to array having list of subtest hashes

=back

=item Returns:

Returns $TCD::TRUE

=back

=cut

sub stop_subtests {

    $Log->enter();

    my %args_validn_spec =
        (
            'subtest_id'  => { type => HASHREF|ARRAYREF },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $subtest_id = $options{'subtest_id'};

    my @subtest_list;

    if ( ref($subtest_id) eq "HASH" ) {
        push(@subtest_list,$subtest_id);
    } else {
        @subtest_list = @{$subtest_id};
    }

    foreach my $test_ref ( @subtest_list ) {
        my $run_id   = $test_ref->{'run_id'} ;

        if ( $test_ref->{'subtest_id'}->isAlive() ) {
            $Tcd_Obj->nlog("Subtest::$run_id is running..");
            my $ret_val = eval {
                            $Tcd_Obj->nlog("Aborting subtest $run_id");
                            $test_ref->{'subtest_id'}->kill();
                            $Tcd_Obj->nlog("subtest $run_id is aborted");
                          };
            if ( $@ ) {
                $Tcd_Obj->nlog("Subtest::$run_id stop FAILED :: $ret_val, error:$@");
            }
        } else {
            $Tcd_Obj->nlog("Subtest::$run_id is not running..");
        }
    }

    $Log->exit();

    return $TCD::TRUE;
}

=head2 generate_mkfile_filename_list()

return a list of file names similar to file names
created by mkfile command on filer

=over 4

=item Arguments:

=over 4

takes a hash as input with following keys

    filename   =>  file prefix name used to create files
    filecount  =>  total number of files

    example :
         mkfile 4k /vol/TestVol/TstFile 100
         filename  => "TstFile"
         filecount => 100
=back

=item Returns:

Returns an array of file names

=back

=cut

sub generate_mkfile_filename_list {

   $Log->enter();

   my %args_validn_spec =
        (
            filename  => {type => SCALAR},
            filecount => {type => SCALAR},
        );

    my %options  = validate(@_, \%args_validn_spec);

    my $filename  = $options{'filename'}; # file name prefix
    my $filecount = $options{'filecount'}; # total number of files

    my (@file_list, $tmpname, $file_extn, $len, $loop);

    $len = length($filecount);

    if ( $filecount <= 1 ) {
        push @file_list, $filename;
    } else {
        for ( $loop = 0; $loop < $filecount; $loop += 1 ) {
            $file_extn = ("0" x ($len - length($loop))) . "$loop";
            $tmpname = $filename . ".$file_extn";
            push @file_list, $tmpname;
        }
    }

    $Log->exit();

    return @file_list;
}

=head2 get_wafl_buftree_blk_info

retrieve the wafl buftree block's information for given inode

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    cont_name         => Container name
    cont_type         => Container type (volume|aggregate)
    inode             => Inode number
    field             => Name of field to be returned
    private           => Set it to 1, if inode is private [0]
    level             => Reference to array having level information [0,1,2]
    apiset            => Apiset object

=back

=item Returns:

Returns a reference to hash having request field value for given level as key

=back

=cut

sub get_wafl_buftree_blk_info {

    $Log->enter();

    my %args_validn_spec =
        (
            'cont_type'    => { type => SCALAR,
                                regex => qr/^(volume|aggregate)$/,
                              },
            'cont_name'    => { type => SCALAR },
            'inode'        => { type => SCALAR },
            'snapid'       => { type => SCALAR, default => 0 },
            'field'        => { type => SCALAR, default => '64_bit_blks' },
            'private'      => { type => SCALAR, default => 0,
                                regex => qr/^(0|1)$/,
                              },
            'level'        => { type => ARRAYREF, default => [0,1,2] },
            'apiset'       => { type => OBJECT,
                                isa  => 'NACL::APISet::Node::CLI::7Mode',
                              },
            'field_list'   => { type => ARRAYREF, default => undef },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj  = $options{'apiset'};
    my $cont_type   = $options{'cont_type'};
    my $cont_name   = $options{'cont_name'};
    my $inode       = $options{'inode'};
    my $snapid      = $options{'snapid'};
    my $field       = $options{'field'};
    my $private     = $options{'private'};
    my @level_list  = @{$options{'level'}};
    my $field_list  = $options{'field_list'};

    my ($resp_obj, @parsed_op, @buftree_info, %blk_info);

    $inode = $private ? "priv:$inode" : $inode;

    $Tcd_Obj->nlog("Retrieve inode: $inode buftree information");
    $resp_obj = $apiset_obj->wafl_buftree_check(
                    'privilege-level' => 'test',
                    $cont_type        => $cont_name,
                    'inodespec'       => $inode,
                    'snapid'          => $snapid,
                );
    @parsed_op = @{$resp_obj->get_parsed_output()};
    @buftree_info = @{$parsed_op[0]->{'buftree_info'}};

    $Tcd_Obj->nlog("Buftree info : ", Dumper \@buftree_info);

    foreach my $level ( @level_list ) {
        foreach my $elm ( @buftree_info ) {
            if ( $elm->{'level'} eq $level ) {
                if ( $field_list ) {
                    foreach my $key ( @{$field_list} ) {
                        $blk_info{$level}->{$key} = $elm->{$key};
                    }
                    last;
                } else {
                    $blk_info{$level} = $elm->{$field};
                    last;
                }
            } # end-if
        } # end-foreach
    } # end-foreach

    $Log->exit();

    return \%blk_info;
}

=head2 configure_feature_licenses

configure feature licenses

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    command_interface => Node object
    cmd_timeout       => Timeout for cli command execution [300]

=back

=item Returns:

Returns reference to a list having license stask objects

=back

=cut

sub configure_feature_licenses {

    $Log->enter();

    my %args_validn_spec =
        (
            'command_interface'     =>
                { type => OBJECT,
                  isa  => 'NACL::C::Node',
                },
            'cmd_timeout' => { type => SCALAR, default => 300 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $node        = $options{'command_interface'};
    my $cmd_timeout = $options{'cmd_timeout'};
    my $node_mode   = $node->mode();

    my $feature_list_7_mode_v1 = [ qw(cifs nfs flex_clone snaplock
                                    snapmirror snaplock_enterprise ) ];

    my $feature_list_7_mode_v2 = [ qw( CIFS NFS FlexClone SnapLock
                                    iSCSI SnapLock_Enterprise
                                    SnapMirror SnapRestore SnapVault ) ];

    my $feature_list_c_mode_v1 = [ qw(Mirror FlexVol_HPO FlexClone CIFS
                                    NFS SnapMirror_DP SnapRestore ) ];

    my $feature_list_c_mode_v2 = [ qw(CIFS NFS FlexClone SnapLock iSCSI
                                    SnapMirror SnapRestore SnapVault ) ];

    my $lic_v2_supported = $node->is_feature_supported(feature => 'license_v2');

    $Lic_Stask_Attribute = $lic_v2_supported ? 'package' : 'feature';

    my ($feature_list_ref, @lic_stask_objs) ;

    if ($lic_v2_supported ) {
        $Tcd_Obj->nlog("license version v2 is supported on filer");
        if ( $node_mode eq 'CMode' ) {
            $feature_list_ref = $feature_list_c_mode_v2;
        } elsif ($node_mode eq '7Mode' ) {
            $feature_list_ref = $feature_list_7_mode_v2;
        }
    } else {
        $Tcd_Obj->nlog("license version v1 is supported on filer");
        if ( $node_mode eq 'CMode' ) {
            $feature_list_ref = $feature_list_c_mode_v1;
        } elsif ($node_mode eq '7Mode' ) {
            $feature_list_ref = $feature_list_7_mode_v1;
        }
    }

    $Tcd_Obj->nlog("Node mode is $node_mode");

    foreach my $feature ( @{$feature_list_ref} ) {
        $Tcd_Obj->nlog("Configuring license for feature: $feature");
        my $lic_task_obj;
        try {
            my $was_added;
            $lic_task_obj = NACL::STask::SystemLicense->add(
                        nacltask_verify      => 1,
                        _was_licensed        => 1,
                        command_interface    => $node,
                        $Lic_Stask_Attribute => $feature,
                        nacltask_if_licensed => 'pass',
                        'method-timeout'     => $cmd_timeout,
                        _was_added           => \$was_added,
                    );
            if ( $was_added ) {
                push( @lic_stask_objs, $lic_task_obj);
                $Tcd_Obj->nlog(" ** $feature license is configured");
            } else {
                $Tcd_Obj->nlog(" => $feature license is already installed");
            }
        } catch NACL::APISet::Exceptions::InvalidParamValueException with {
            my $excpn_obj = shift;
            if( $excpn_obj->text() =~ /already\s+installed/i ) {
                $Tcd_Obj->nlog("$feature license was already installed");
            } else {
                $Tcd_Obj->nlog("Exception: " . $excpn_obj->text());
                $Tcd_Obj->nlog("Rethrowing exception..");
                $excpn_obj->throw();
            }
        }
    }

    #
    # If node mode is 7-mode, then we need to enable
    # license options also
    #
    if ( $node_mode eq "7Mode" ) {
        foreach my $lic_option ( keys %Lic_7m_Options ) {
            NACL::C::Options->option(
                command_interface => $node,
                'option-name'     => $lic_option,
                'option-value'    => 'on',
            );
        }
    }

    $Log->exit();

    return \@lic_stask_objs;
}

=head2 delete_aggregate_snapshots

delete all matching snapshots on given aggregate, if exist

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    command_interface => Node object
    aggregate         => Aggregate name
    snapshot          => Snapshot name or prefix
    cmd_timeout       => Timeout for cli command execution [300]

=back

=item Returns:

None

=back

=cut

sub delete_aggregate_snapshots {

    $Log->enter();

    my %args_validn_spec =
        (
            'command_interface'     =>
                { type => OBJECT,
                  isa  => 'NACL::C::Node',
                },
            'cmd_timeout' => { type => SCALAR, default => 300 },
            'snapshot'    => { type => SCALAR },
            'aggregate'   => { type => SCALAR },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $node        = $options{'command_interface'};
    my $cmd_timeout = $options{'cmd_timeout'};
    my $snapshot    = $options{'snapshot'};
    my $aggregate   = $options{'aggregate'};

    try {
        my @snap_comp_obj_list =  NACL::C::AggregateSnapshot->find(
            command_interface => $node,
            filter => { aggregate => $aggregate, snapshot => $snapshot,},
            allow_empty  => 1,
            'method-timeout' => $cmd_timeout,
        );

        foreach my $snap_cmp_obj ( @snap_comp_obj_list) {
            try {
                $Tcd_Obj->nlog("Deleting snapshot: " . $snap_cmp_obj->snapshot);
                $snap_cmp_obj->delete();
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $excpn_obj = shift;
                $Tcd_Obj->nlog("Error during snapshot: "
                    . $snap_cmp_obj->snapshot . "deletion");
                $Tcd_Obj->nlog("Exception: " . $excpn_obj->text());
            };
        }
    } catch NACL::Exceptions::NoElementsFound with {
        $Tcd_Obj->nlog("snapshot(s) by name/prefix: $snapshot do not exit");
    };

    $Log->exit();
}

sub aggregate_snap_create_in_background {

    $Log->enter();

    my %args_validn_spec =
        (
            'aggr_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'snapshot_count'    => { type => SCALAR,  default => 255 },
            'snapshot_interval' => { type => SCALAR,  default => 5 },
            'snap_crt_prefix'   => { type => SCALAR,  default => 'aggr_snap_' },
            'snap_start_index'  => { type => SCALAR,  default => 1 },
            'cmd_timeout'       => { type => SCALAR,  default => 300 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $count              = $options{'snapshot_count'};
    my $interval           = $options{'snapshot_interval'};
    my $snap_crt_prefix    = $options{'snap_crt_prefix'};
    my $start_index        = $options{'snap_start_index'};
    my $aggregate          = $aggr_task_obj->aggregate ;
    my $run_id             = "$aggregate" . "_snapshot_create_" . time();

    $Tcd_Obj->nlog("Starting aggr's snap create subtest with id : $run_id");
    my $test_id =
        subtest(
            sub {
                my $start_id = 1;
                my $name;

                while($start_id <= $count ) {
                    $name = "$snap_crt_prefix$start_index";
                    logcomment("Creating snapshot: $name on aggregate: $aggregate");
                    NACL::C::AggregateSnapshot->create(
                        command_interface  => $aggr_task_obj->command_interface,
                        aggregate          => $aggregate,
                        snapshot           => $name,
                        'method-timeout'   => $cmd_timeout,
                     );
                    sleep($interval);
                    $start_id++;
                    $start_index++;
                }
            },
            -bg,
            -runid, $run_id,
            -onexit, sub { $Tcd_Obj->nlog("Aggr's snapshot creation subtest exiting.."); },
        );

    sleep(2); # let's wait for 2 sec, before another call is made

    $Log->exit();

    return ( { 'subtest_id' => $test_id, 'run_id' => $run_id } );

} # end of aggregate_snap_create_in_background()

sub aggregate_snap_delete_in_background {

    $Log->enter();

    my %args_validn_spec =
        (
            'aggr_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'snapshot_count'    => { type => SCALAR,  default => 255 },
            'snapshot_interval' => { type => SCALAR,  default => 5 },
            'snap_del_prefix'   => { type => SCALAR,  default => 'aggr_snap_' },
            'snap_start_index'  => { type => SCALAR,  default => 1 },
            'cmd_timeout'       => { type => SCALAR,  default => 300 },
        );

    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $aggr_task_obj      = $options{'aggr_task_obj'};
    my $cmd_timeout        = $options{'cmd_timeout'};
    my $count              = $options{'snapshot_count'};
    my $interval           = $options{'snapshot_interval'};
    my $snap_del_prefix    = $options{'snap_del_prefix'};
    my $start_index        = $options{'snap_start_index'};
    my $aggregate          = $aggr_task_obj->aggregate ;
    my $run_id             = "$aggregate" . "_snapshot_delete_" . time();

    $Tcd_Obj->nlog("Starting aggr's snap delete subtest with id: $run_id");

    my $test_id =
        subtest(
            sub {
                my $start_id = 1;
                my $name;

                while($start_id <= $count ) {
                    $name = "$snap_del_prefix$start_index";
                    logcomment("Deleting snapshot: $name on aggregate $aggregate");
                    try {
                        NACL::C::AggregateSnapshot->delete(
                            command_interface  => $aggr_task_obj->command_interface,
                            aggregate          => $aggregate,
                            snapshot           => $name,
                            'method-timeout'   => $cmd_timeout,
                        );
                    } catch NACL::APISet::Exceptions::ResponseException with {
                        logcomment("snapshot: $name does not exist");
                    } catch NACL::APISet::Exceptions::InvalidParamValueException with {
                        logcomment("snapshot:: $name does not exist");
                    };
                    sleep($interval);
                    $start_id++;
                    $start_index++;
                }
            },
            -bg,
            -runid, $run_id,
            -onexit, sub { $Tcd_Obj->nlog("Aggr's snapshot deletion subtest exiting.."); },
        );

    sleep(2); # let's wait for 2 sec, before another call is made

    $Log->exit();

    return ( { 'subtest_id' => $test_id, 'run_id' => $run_id } );

} # end of aggregate_snap_delete_in_background()

=head2 get_non_hole_fbns

check for non hole fbns from the given range of fbns defined
by start_fbn and end_fbn for the given container name and type

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset            => Apiset object
    snapid            => Snapshot id
    level             => Block level (0,1 etc)
    inode             => Inode number
    start_fbn         => Start FBN
    end_fbn           => End FBN
    cont_type         => Container type (volume|aggregate)
    private           => set this flag to 1, if inode is private
    count             => number of non-hole fbns to be returned [1]

=back

=item Returns:

returns the first non hole fbn from the given range, if called in
scalar context, Otherwise returns the list of all non hole fbns.

if non-hole fbn is not found, exception will be thrown.

=back

=cut

sub get_non_hole_fbns {

    my %args_validn_spec =
        (
            'apiset'       =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode',
                },
            'snapid'       => { type => SCALAR },
            'level'        => { type => SCALAR },
            'inode'        => { type => SCALAR },
            'start_fbn'    => { type => SCALAR },
            'end_fbn'      => { type => SCALAR },
            'cont_type'    => { type => SCALAR,
                                regex => qr/^(volume|aggregate)$/,
                              },
            'cont_name'    => { type => SCALAR },
            'private'      => { type => SCALAR,
                                default => 1,
                                regex   => qr/^(0|1)$/,
                              },
            'count'      =>   { type => SCALAR, default => 1 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my @fbn_list  = ($options{'start_fbn'}..$options{'end_fbn'});
    my $cont_type = $options{'cont_type'};
    my $inode     = $options{'inode'};
    my $start_fbn = $options{'start_fbn'};
    my $end_fbn   = $options{'end_fbn'};
    my $total_fbn = $options{'count'};

    my @fbn_op_list;

    my $fbn_is_found = 0;

    my $fbn_count = 0;

    if ( $options{'private'} ) {
        $inode = "priv:$options{'inode'}";
    }

    foreach my $fbn ( @fbn_list ) {
        try {
            $Tcd_Obj->step("Get the block info for inode:$inode fbn: $fbn");

            my $response = $options{'apiset'}->dumpblock_info(
                            $cont_type  => $options{'cont_name'},
                            fileid      => $options{'inode'},
                            level       => $options{'level'},
                            private     => $options{'private'},
                            snapid      => $options{'snapid'},
                            blockno     => $fbn,
            )->get_parsed_output();

            my $pvbn_No =  $response->[0]->{'pvbn'};

            if ( $pvbn_No != 0 ) {
                $Tcd_Obj->nlog("===== $fbn is not a hole =====");
                $Tcd_Obj->nlog("For L0 FBN: $fbn, PVBN: $pvbn_No");
                $fbn_is_found = 1;
                ++$fbn_count;
                push(@fbn_op_list, $fbn);

                if ( $fbn_count == $total_fbn ) {
                    last;
                }
            }
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            $Tcd_Obj->nlog( "Error is: " . $exception_object->text() );
            my $resp = $exception_object->get_response_object();
            $Tcd_Obj->nlog( "Raw output was:\n" . $resp->get_raw_output() );
            $Tcd_Obj->nlog("===== $fbn is a hole =====");
        };
    }

    $Log->exit();

    if ( 0 == $fbn_is_found ) {
        NATE::BaseException->throw("Non-hole fbn not found in "
            . "$start_fbn-$end_fbn fbn range");
    }

    return ( wantarray() ? @fbn_op_list : "@fbn_op_list" );
}

=head2 create_snapshot

create snapshot by given name for given volume
in case of timeout, exception will be thrown.

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    vol_task_obj   => Volume task object
    snapshot       => Snapshot name
    timeout        => Wait for snapshot completion [600]

=back

=item Returns:

returns snapshot physical id.

=back

=cut

sub create_snapshot {

    my %args_validn_spec =
        (
            'vol_task_obj' =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
            'snapshot'     => { type => SCALAR },
            'timeout'      => { type => SCALAR, default => 600 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $snapshot     = $options{'snapshot'};
    my $timeout      = $options{'timeout'};
    my $vol_task_obj = $options{'vol_task_obj'};
    my $volume       = $vol_task_obj->volume;
    my $node         = $vol_task_obj->command_interface;

    $Tcd_Obj->step("Creating a snapshot \"$snapshot\" on \"$volume\"");

    my $snap_cmp_obj =
        $vol_task_obj->snapshot_create( snapshot => $snapshot );

    my $flag = 0;
    my $t1   = time();
    my $start_time = time();
    my $end_time   = time() + $timeout;

    my $snap_id;
    my $status;

    while ($start_time <= $end_time) {
        try {
            my $snap = NACL::C::VolumeSnapshot->find(
                command_interface => $node,
                filter            => {
                    snapshot => $snapshot,
                    volume   => $volume,
                },
            );
            $snap_id = $snap->get_one_state_attribute('physical-snap-id');
            $Tcd_Obj->nlog("Snapshot id is : $snap_id");
            $flag = 1;
            $start_time = $end_time + 1;
        }
        catch NATE::BaseException with {
            my $exception_object = shift;
            $Tcd_Obj->nlog( "Error: ", $exception_object->text() );
            sleep(5);
            $start_time = time();
        };
    }

    if (!$flag) {
        $Log->exit();
        NATE::BaseException->throw("$snapshot is not created for"
            . " $volume vol in $timeout seconds");
    }

    my $t2 = time();
    $Tcd_Obj->nlog("Time taken for snapshot creation : " . ($t2 - $t1 ));

    $Log->exit();
    return $snap_id;
}

=head2 create_aggregate_snapshot

create snapshot by given name for given aggregate.
in case of timeout, exception will be thrown.

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    aggr_task_obj  => Aggregate task object
    snapshot       => Snapshot name
    timeout        => Wait for snapshot completion [600]

=back

=item Returns:

returns snapshot physical id.

=back

=cut

sub create_aggregate_snapshot {
    my %args_validn_spec =
        (
            'aggr_task_obj' =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Aggregate',
                },
            'snapshot'     => { type => SCALAR },
            'timeout'      => { type => SCALAR, default => 600 },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $snapshot      = $options{'snapshot'};
    my $timeout       = $options{'timeout'};
    my $aggr_task_obj = $options{'aggr_task_obj'};
    my $aggregate     = $aggr_task_obj->aggregate;
    my $node          = $aggr_task_obj->command_interface;

    $Tcd_Obj->step("Creating a snapshot \"$snapshot\" on \"$aggregate\"");

    my $snap_cmp_obj =
        NACL::C::AggregateSnapshot->create(
            command_interface  => $node,
            aggregate          => $aggregate,
            snapshot           => $snapshot,
        );

    my $flag = 0;
    my $t1   = time();
    my $start_time = time();
    my $end_time   = time() + $timeout;

    my $snap_id;

    while ($start_time <= $end_time) {
        try {
            my $snap = NACL::C::AggregateSnapshot->find(
                command_interface => $node,
                filter            => {
                    snapshot  => $snapshot,
                    aggregate => $aggregate,
                },
            );
            $snap_id = $snap->state()->snapid();
            $Tcd_Obj->nlog("Snapshot id is : $snap_id");
            $flag = 1;
            $start_time = $end_time + 1;
        }
        catch NATE::BaseException with {
            my $exception_object = shift;
            $Tcd_Obj->nlog( "Error: ", $exception_object->text() );
            sleep(2);
            $start_time = time();
        };
    }

    if (!$flag) {
        $Log->exit();
        NATE::BaseException->throw("$snapshot is not created for"
            . " $aggregate aggregate in $timeout seconds");
    }

    my $t2 = time();
    $Tcd_Obj->nlog("Time taken for snapshot creation : " . ($t2 - $t1 ));

    $Log->exit();
    return $snap_id;
}

=head2 does_files_exist_in_directory

returns true, if all files in given list are available
in given directory

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    command_interface => node object
    dir_name          => directory name
    file_list         => arrary ref to list of files to be checked

=back

=item Returns:

Returns $TCD::TRUE if successful, otherwise $TCD::FALSE

=back

=cut

sub does_files_exist_in_directory {

    $Log->enter();

    my %args_validn_spec =
        (
            'command_interface'     =>
                { type => OBJECT,
                  isa  => 'NACL::C::Node',
                },
            'dir_name'    => { type => SCALAR },
            'file_list'   => { type => ARRAYREF },
        );

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $node        = $options{'command_interface'};
    my $file_list   = $options{'file_list'};
    my $dir_name    = $options{'dir_name'};

    my $apiset = $node->get_7m_or_nodescope_apiset();
    $apiset->set_privilege('privilege-level' => 'test');

    my $resp = $apiset->ls( 'paths' => $dir_name );
    my $filer_file_list = $resp->get_parsed_output()->[0]->{'files'};

    $Tcd_Obj->nlog("List of files: ", Dumper($filer_file_list) );

    my $ret_val = $TCD::PASS;

    foreach my $file (@{$file_list}) {
        if ( ! grep(/$file/, @{$filer_file_list} ) ) {
            $Tcd_Obj->nlog("File pattern: $file not found");
            $ret_val = $TCD::FAIL;
        }
    }

    $Log->exit();
    return $ret_val;
}



# ------------------------------------------------------
#          Private Subroutines Defintion
# ------------------------------------------------------

=head2 _start_io_ops

starts read/write ops on given user file using filersio

=over 4

=item Arguments:

=over 4

takes a hash reference as input with following keys
    apiset        => node scope or 7-mode apiset
    vol_task_obj  => volume task object
    sio_blk_size  => block size used by filersio to read data
    sio_file_size => file size used by filersio to read data
    sio_file_name => file name used by filersio to read data
    sio_ops_type  => type of ops to be performed read | write [read]
    sio_timeout   => timeout for filersio ops completion [600]

=back

=item Returns:

throws an exception in case of error

=back

=cut

sub _start_io_ops {

    $Log->enter();

    my %args_validn_spec =
        (
            'sio_blk_size'      => { type => SCALAR },
            'sio_file_size'     => { type => SCALAR },
            'sio_file_name'     => { type => SCALAR },
            'sio_timeout'       => { type => SCALAR, default => 600 },
            'sio_ops_type'      => { type => SCALAR, default => "read" },
            'apiset'            =>
                { type => OBJECT,
                  isa  => 'NACL::APISet::Node::CLI::7Mode'
                },
            'vol_task_obj'      =>
                { type => OBJECT,
                  isa  => 'NACL::STask::Volume',
                },
        );

    #
    # save test data for latter use
    #
    my $test_data = $_[0];

    my %options = validate_with(
                    'params'      => \@_,
                    'spec'        => \%args_validn_spec,
                    'allow_extra' => 1,
                  );

    my $apiset_obj         = $options{'apiset'};
    my $vol_task_obj       = $options{'vol_task_obj'};
    my $sio_blk_size       = $options{'sio_blk_size'};
    my $sio_file_size      = $options{'sio_file_size'};
    my $sio_file_name      = $options{'sio_file_name'};
    my $sio_ops_type       = $options{'sio_ops_type'};
    my $sio_timeout        = $options{'sio_timeout'};

    my $inode; # scalar to store inode

    my @pattern = (66..90,97..122);
    my $fill_pattern = $pattern[int(rand(scalar @pattern))];
    #
    ## Perform read or write ops on file using filersio
    #
    my $resp_obj;
    try {
        if ( $sio_ops_type =~ m/^read$/i ) {
            $Tcd_Obj->step("Generating read ops on file \"$sio_file_name\""
                . " using filersio");
            $resp_obj = $apiset_obj->filersio_asyncio_active(
                'read-percentage' => 100, 'rand-percentage' => 100,
                'blk-size' => $sio_blk_size,
                'start-byte' => 0,   'end-byte' => $sio_file_size,
                'num-active' => 64, 'run-time' => 100,
                'filename'   => $sio_file_name, 'synchronous' => 0,
            );
        } elsif ( $sio_ops_type =~ m/^write$/i ) {
            $Tcd_Obj->step("Generating write ops on file \"$sio_file_name\""
                . " using filersio");
            $resp_obj = $apiset_obj->filersio_asyncio_active(
                'read-percentage' => 0, 'rand-percentage' => 0,
                'blk-size' => $sio_blk_size, 'create' => 1,
                'start-byte' => 0,  'end-byte' => $sio_file_size,
                'num-active' => 10, 'run-time' => 300,
                'fill-pattern' => $fill_pattern,
                'filename'   => $sio_file_name, 'synchronous' => 0,
                'connectrec-timeout' => $sio_timeout,
            );
        }

        my @parsed_op = @{$resp_obj->get_parsed_output()};
        my $sio_status = $parsed_op[0]->{'statistics'}->[0]->{'filersio_status'};
        if ( $sio_status =~ /workload initiated asynchronously/) {
            $Tcd_Obj->nlog("filersio started successfully");
        } else {
            $Tcd_Obj->nlog("filersio failed to start");
            $Tcd_Obj->nlog(Dumper \@parsed_op);
            $Log->exit();
            NATE::BaseException->throw("Filersio failed to start read/write ops");
        }
    } # end try
    catch NACL::APISet::Exceptions::InvalidParamValueException with {
        #
        # In few corruption cases, user data file is lost, so filersio
        # will fail to start ops on this user file. In this case
        # exception is accepted
        #
        my $excpn_obj = shift;
        $Tcd_Obj->nlog("Caught an exception thrown by filersio()");
        $Tcd_Obj->nlog("Let's verify, if user file was lost");

        if ( $TCD::FALSE == does_user_file_exist(
                                {
                                    vol_task_obj  => $vol_task_obj,
                                    sio_file_name => $sio_file_name,
                                    op_inode      => \$inode,
                                } ) ) {

            $Tcd_Obj->nlog("File \"$sio_file_name\" is really lost");
            $Tcd_Obj->nlog("Exception thrown by filersio is correct");
            $Tcd_Obj->nlog("Error string: " . $excpn_obj->text());
            $Tcd_Obj->nlog("Exception: ". ref($excpn_obj));

        } else {
            NATE::BaseException->throw("File \"$sio_file_name with inode:"
                . "$inode found, filersio failed to start");
        }
    };

    $Log->exit();
} #end of _start_io_ops()

1;


=head1 SEE ALSO

    None

=head1 STATUS

    Shared

=head1 AUTHOR

    Tarun Garg (tarung@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/IronUtil/MPT/MptTestUtil.pm#5 $
    Last $Change: 2244423 $
    Last Revised $DateTime: 2013/11/21 23:02:16 $

=head1 HISTORY

    2011-02-11 author: tarung      original lib created

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=cut
