
=pod

=head1 NAME

    JAMS.pm - Interface to JAMS database system.

=head1 SYNOPSIS

    use JAMS;

    my $JAMS = new JAMS();

    $JAMS->store_jams_result(tcrid => 22,
                            image => "/x/eng/rlse/DOT/RboilermakerN/export/foo.tgz",
                            change_nbr => 123456,
                            log_dir => "/x/testbeds/logdir/10245678",
                            result => "PASS",
                            run_time => 22);

    $JAMS->generate_remainder_plan(change_nbr => 123456),
                                    image => "/x/eng/rlse/DOT/RboilermakerN/export/foo.tgz",
                                    plan_name => "Rboilermaker_10M_Qualified.plan",
                                    directory => "/x/testbeds/JAMS15")
                                    );


    ($ok, $error) = $JAMS->jams_run_stats( 'action'   => <action>,
                                           '<column>' => <value>,
    );


    $JAMS->resolve_ids;

    $JAMS->resolve_run_spec;

    $JAMS->resolve_run_spec_name;

    $JAMS->disconnect;


=head1 Options

This module is also used to house options for use by the RegressionSystem*.thpl for JAMS
runs.


=over

=item JAMS DB Connection

=over

=item DB_DATABASE = JAMS database name

=item DB_HOSTNAME = Hostname of the database server

=item DB_USER = Username to log in as on database server

=item DB_PASSWORD = Password to log in as on database server

=back

=item JAMS Operations

=over

=item JAMS_PLAN_DIRECTORY = [/x/eng/rtp-testbeds/art/rtp_art/JAMS2.0]

=item REGRESSION_PROMOTED_IMAGE = promoted image path

=item REGRESSION_UPGRADE_IMAGE = upgrade image path

=item REGRESSION_TEST_MODE = filer operational mode [SEVEN|TEN]

=item REGRESSION_NACL_SHOC = [0|1]

=item JAMS_VSIM_NUM_NICS = [undef]

=back

=item JAMS DB IDs and Strings

=over

=item REGRESSION_TESTCASE_RELEASE_[ID|STR] = Testcase Release id

=item REGRESSION_SPECNAME_[ID|STR] = Specname of planfile

=item REGRESSION_PLATFORM_[ID|STR] = Platform id of plan file

=item REGRESSION_BRANCH_[ID|STR] = Branch id of plan file

=item REGRESSION_LINE_[ID|STR] = Line id of plan file

=item REGRESSION_MODE_[ID|STR] = Mode id of plan file

=item REGRESSION_TESTSET_[ID|STR] = Testset id of plan file

=item REGRESSION_BUILD_FLAVOR_[ID|STR] = Build id of plan file

=item REGRESSION_Q_STATUS_[ID|STR] = Qualifying Status of a runspec

=item REGRESSION_TESTCASE_[ID|STR] = Testcase id

=item REGRESSION_RUN_TO_PLAN_[ID|STR] = JAMS Run to Plan ID

=item REGRESSION_RUNSPEC_[ID|STR] = JAMS runspec ID/String

=back


=item JAMS counts

=over

=item REGRESSION_TEST_COUNT - count of test in runspec

=item  REGRESSION_TEST_SKIPPED - count of test in runspec marked skipped

=item  REGRESSION_TEST_IGNORED - count of test in runspec to ignore due to run to plan

=item REGRESSION_TEST_PREEXECUTED - count of test that have a pre-executed result

=item REGRESSION_TEST_TO_RUN - count of test that will be executed by plan file (count - skipped)

=back

=item JAMS System thpl's

=over

=item REGRESSION_CONLOG_START_THPL = [tools/smoke/conlog_start.thpl]

=item REGRESSION_CONLOG_END_THPL = [tools/smoke/conlog_end.thpl]

=item REGRESSION_BUILD_VERSION_THPL = [os/build_version.thpl]

=item REGRESSION_CLIENT_CHECK_THPL = [tools/client_check.thpl]

=item REGRESSION_VSA_DESTROY_THPL = [VSA/destroy.thpl]

=item REGRESSION_VSA_CREATE_THPL = [VSA/create.thpl]

=item REGRESSION_GX_BOOT_THPL = [os/boot/gx_boot.thpl]

=item REGRESSION_DEUX_INIT_THPL = [os/boot/deux_init.thpl]

=item REGRESSION_CREATE_VSIMS_THPL = [tools/jams/vsim/create_vsims.thpl]

=item REGRESSION_RESTORE_VSIMS_THPL = [tools/jams/vsim/restore_vsims.thpl]

=item REGRESSION_EXTEND_VSIMS_THPL = [tools/jams/vsim/extend_vsims.thpl]

=item REGRESSION_OTC_PREJOB_THPL = [tools/otc_prejob.thpl]

=item REGRESSION_SYSTEM_ABORT_THPL = [tools/RegressionSystemAbort.thpl]

=item REGRESSION_QC_PUSH_THPL = [qc_push.thpl]

=item REGRESSION_SETRAL_START_THPL = [os/setral_start.thpl]

=item REGRESSION_SETRAL_STOP_THPL  = [os/setral_stop.thpl]

=item REGRESSION_GCOV_THPL = [tools/gcov.thpl]

=item REGRESSION_LOG_ANALYSIS_PROCLOG_THPL = [mhost/log_analysis/proclog.thpl]

=item REGRESSION_CLEANUP_FIJIS_THPL = [tools/cleanup_fijis.thpl]

=item REGRESSION_GSI_THPL = [tools/gsi.thpl]

=item REGRESSION_DBH_THPL = [tools/smoke/did_bad_happen.thpl]

=item REGRESSION_CHECK_CLUSTER_THPL = [tools/check_cluster.thpl]

=item REGRESSION_OTC_POSTTEST_THPL = [tools/otc_posttest.thpl]

=item REGRESSION_HOLD_JIG_THPL = [tools/hold_jig.thpl]

=item REGRESSION_SPINCONTROL_THPL = [os/spincontrol.thpl]

=item REGRESSION_RESULTS_MAIL_THPL = [tools/results_mail.thpl]

=item REGRESSION_SHOC = [tools/shoc.thpl]

=item REGRESSION_NACL_SHOC_SAVE = [tools/shoc_save.thpl]

=item REGRESSION_NACL_SHOC_COMPARE = [tools/shoc_diff.thpl]

=item REGRESSION_SYSTEM_PREJOB = [tools/RegressionSystemPreJob.thpl]

=item REGRESSION_SYSTEM_POSTJOB = [tools/RegressionSystemPostJob.thpl]

=item REGRESSION_SYSTEM_PRETEST = [tools/RegressionSystemPreTest.thpl]

=item REGRESSION_SYSTEM_POSTTEST = [tools/RegressionSystemPostTest.thpl]

=back

=back


=head1 METHODS

=cut


package JAMS;

use strict;
use warnings;

use DBI;
use Options;
use Tharn;
use Data::Dumper;
use File::Temp qw/tempdir/;
use NATE::Inc qw/find_file/;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT);

our @EXPORT_OK = qw(get_jams_image_id
                generate_remainder_plan
                store_jams_result
                make_run_to_plan_id
);


my $Options = new Options();

$Options->parameters(definitions =>
    [
    {
        name => "CODE_COVERAGE",
        description => "Flag indicating to collect code coverage metrics",
        spec => "boolean",
        default => 0,
    },
    {
        name => "DB_DATABASE",
        description => "JAMS database name",
        spec => "s",
        default => "JAMS",
    },
    {
        name => "ALLNODES",
        description => "Full node list set by prejob.thpl",
        spec => "s",
        default => "",
    },
    {
        name => "DB_HOSTNAME",
        description => "Hostname of the database server",
        spec => "s",
        default => "csiqa-jams-admin-prod.rtp.netapp.com",
    },
    {
        name => "DB_USER",
        description => "Username to log in as on database server",
        spec => "s",
        default => "sqluser",
    },
    {
        name => "DB_PASSWORD",
        description => "Password to log in as on database server",
        spec => "s",
        default => "",
    },
    {
        name => "JAMS_PLAN_DIRECTORY",
        description => "Directory ndisp jobs will use plan files from",
        spec => "s",
        default => "/x/eng/rtp-testbeds/art/rtp_art/JAMS2.0",
    },
    {
        name => "REGRESSION_TESTCASE_RELEASE_ID",
        description => "testcase release id",
        spec => "i",
    },
    {
        name => "REGRESSION_SPECNAME_ID",
        description => "Specname of planfile",
        spec => "i",
    },
    {
        name => "REGRESSION_PLATFORM_ID",
        description => "Platform id of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_BRANCH_ID",
        description => "Branch id of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_LINE_ID",
        description => "line id of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_MODE_ID",
        description => "mode id of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_TESTSET_ID",
        description => "Testset id of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_BUILD_FLAVOR_ID",
        description => "Build id of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_Q_STATUS_ID",
        description => "Q status of plan file",
        spec => "i",
    },
    {
        name => "REGRESSION_TESTCASE_ID",
        description => "Testcase id",
        spec => "i",
    },
    {
        name => "REGRESSION_RUN_TO_PLAN_ID",
        description => "run to plan id for jams",
        spec => "i",
    },
    {
        name => "REGRESSION_RUNSPEC_ID",
        description => "jams runspec id",
        spec => "i",
    },
    {
        name => "REGRESSION_SCHEDULE_ID",
        description => "jams schedule id",
        spec => "i",
    },
    {
       name => "REGRESSION_CONLOG_START_THPL",
       description => "conlog_start.thpl path",
       default => 'tools/smoke/conlog_start.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_CONLOG_END_THPL",
       description => "tools/smoke/conlog_end.thpl path",
       default => 'tools/smoke/conlog_end.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_BUILD_VERSION_THPL",
       description => "build_version.thpl path",
       default => 'os/build_version.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_CLIENT_CHECK_THPL",
       description => "client_check.thpl path",
       default => 'tools/client_check.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_VSA_DESTROY_THPL",
       description => "VSA destroy.thpl path",
       default => 'VSA/destroy.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_VSA_CREATE_THPL",
       description => "vsa_create.thpl path",
       default => 'VSA/create.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_GX_BOOT_THPL",
       description => "gx_boot.thpl path",
       default => 'os/boot/gx_boot.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_CREATE_VSIMS_THPL",
       description => "create_vsims.thpl path",
       default => 'tools/jams/vsim/create_vsims.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_RESTORE_VSIMS_THPL",
       description => "restore_vsims.thpl path",
       default => 'tools/jams/vsim/restore_vsims.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_EXTEND_VSIMS_THPL",
       description => "extend_vsims.thpl path",
       default => 'tools/jams/vsim/extend_vsims.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_NTAP2IBM",
       description => "ntap2ibm.thpl path",
       default => 'tools/ntap2ibm.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_DEUX_INIT_THPL",
       description => "deux_init.thpl path",
       default => 'os/boot/deux_init.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_OTC_PREJOB_THPL",
       description => "otc_prejob.thpl path",
       default => 'tools/otc_prejob.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_SYSTEM_ABORT_THPL",
       description => "RegressionSystemAbort.thpl path",
       default => 'tools/RegressionSystemAbort.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_QC_PUSH_THPL",
       description => "qc_push.thpl path",
       default => 'qc_push.thpl',
       spec => "s",
    },
    {
        name => "IBM_CONVERT",
        description => "Flag indicating to run ntap2ibm branding conversion.",
        spec => "boolean",
        default => 0,
    },
    {
       name => "REGRESSION_SETRAL_START_THPL",
       description => "os/setral_start.thpl path",
       default => 'os/setral_start.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_SETRAL_STOP_THPL",
       description => "os/setral_stop.thpl path",
       default => 'os/setral_stop.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_GCOV_THPL",
       description => "tools/gcov.thpl path",
       default => 'tools/gcov.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_LOG_ANALYSIS_PROCLOG_THPL",
       description => "mhost/log_analysis/proclog.thpl path",
       default => 'mhost/log_analysis/proclog.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_CLEANUP_FIJIS_THPL",
       description => "tools/cleanup_fijis.thpl path",
       default => 'tools/cleanup_fijis.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_GSI_THPL",
       description => "tools/gsi.thpl path",
       default => 'tools/gsi.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_DBH_THPL",
       description => "tools/smoke/did_bad_happen.thpl path",
       default => 'tools/smoke/did_bad_happen.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_CHECK_CLUSTER_THPL",
       description => "tools/check_cluster.thpl path",
       default => 'tools/check_cluster.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_OTC_POSTTEST_THPL",
       description => "tools/otc_posttest.thpl path",
       default => 'tools/otc_posttest.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_HOLD_JIG_THPL",
       description => "tools/hold_jig.thpl path",
       default => 'tools/hold_jig.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_SPINCONTROL_THPL",
       description => "os/spincontrol.thpl path",
       default => 'os/spincontrol.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_RESULTS_MAIL_THPL",
       description => "tools/results_mail.thpl path",
       default => 'tools/results_mail.thpl',
       spec => "s",
    },
    {
       name => "REGRESSION_PROMOTED_IMAGE",
       description => "promoted image path",
       spec => "s",
    },
    {
       name => "REGRESSION_UPGRADE_IMAGE",
       description => "upgrade image path",
       spec => "s",
    },
    {
       name => "REGRESSION_SYSTEM_PREJOB",
       description => "regression prejob",
       default => "tools/RegressionSystemPreJob.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_SYSTEM_POSTJOB",
       description => "regression postjob",
       default => "tools/RegressionSystemPostJob.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_SYSTEM_PRETEST",
       description => "regression pretest",
       default => "tools/RegressionSystemPreTest.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_SYSTEM_POSTTEST",
       description => "regression posttest",
       default => "tools/RegressionSystemPostTest.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_SHOC",
       description => "regression show output comparator",
       default => "tools/shoc.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_NACL_SHOC",
       description => "Use NACL regression show output comparator",
       spec => "boolean",
       default => 1,
    },
    {
       name => "REGRESSION_NACL_SHOC_SAVE",
       description => "NACL regression show output save",
       default => "tools/shoc_save.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_NACL_SHOC_COMPARE",
       description => "NACL regression show output comparator",
       default => "tools/shoc_diff.thpl",
       spec => "s",
    },
    {
       name => "REGRESSION_CON_INIT_FAIL",
       description => "Maximum consecutive init failures allowed.",
       default => "4",
       spec => "i",
    },
    {
        name => "JAMS_CCL_RULESET_DIR",
        description => "Directory path JAMS will pull CCL files from",
        default => "tools/jams/ccl/",
        spec => "s",
    },
    {
        name => "CCL_RULESETS",
        description => "List of User Supplied CCL Rulesets to apply.",
        spec => "s@",
    },
    {
        name => "JAMS_CCL",
        description => "JAMS CCL Engine.",
        default => "tools/CCL/ccl_engine.thpl",
        spec => "s",
    },
    {   name        => "DEFAULT_RULESET",
        description => "Default CCL rulesets to be applied.",
        default     => "lib/Network/CCL/SCON/ifgrp.rs",
        spec        => "s",
    },
    {
        name => "JAMS_CCL_UNDO",
        description => "JAMS CCL Undo.",
        default => "tools/CCL/ccl_undo.thpl",
        spec => "s",
    },
    {
        name => "GCOV_DUMP_DIR",
        description => "User defined path for overiding gcov data, 0 default, 1 root, <user defined hierchy>",
        default => '0',
        spec => "s",
    },
    {
       name => "JAMS_RUN_TO_PLAN_NAME",
       description => "jams run to plan name",
       spec => "s",
    },
    {
       name => "JAMS_RUN_TO_PLAN_ID",
       description => "jams run to plan ID",
       spec => "i",
    },
    {
       name => "JAMS_SCHEDULE_ID",
       description => "jams schedule",
       spec => "i",
    },
    {
        name => "JAMS_VSIM_SNAPSHOT",
        description => "vsim snaphot name",
        spec => "s",
        default => "JAMS_SNAPSHOT",
    },
    {
        name => "JAMS_VSIM_CREATE_ARGS",
        description => "Arguments used to create vsims in JAMS",
        spec => "s",
        default => "",
    },
    {
       name => "JAMS_VSIM_NUM_NICS",
       description => "The number of NICS (ports) for the vsim, [valid values 5-8]",
       spec => "i",
    },
    {
        name => "BUILDROOT",
        description => "Ontap build directory root",
        spec => "s",
    },
    {
       name => "JOBID",
       description => "ndisp run jobid",
       spec => "s",
    },
    {
       name => 'REGRESSION_TEST_COUNT',
       description => "jams plan file test count",
       spec => "i",
    },
    {
       name => 'REGRESSION_TEST_SKIPPED',
       description => "jams plan file test skipped",
       spec => "i",
    },
    {
       name => 'REGRESSION_TEST_IGNORED',
       description => "jams plan file test ignored",
       spec => "i",
    },
    {
        name => 'REGRESSION_TEST_TO_RUN',
        description => "jams plan file test to run",
        spec => "i",
    },
    {
       name => 'REGRESSION_TEST_PREEXECUTED',
       description => "jams plan file pre-executed test count",
       spec => "i",
    },
    {
        name => "REGRESSION_TEST_MODE",
        description => "mode (e.g. SEVEN or TEN)",
        spec => "s",
        default => "noMode",
        valid_values => [qw( noMode SEVEN TEN )],
    },
    {
       name => 'REGRESSION_LINE_STR',
       description => 'JAMS line string',
       spec => "s",
    },
    {
       name => 'REGRESSION_MODE_STR',
       description => 'JAMS mode string',
       spec => "s",
    },
    {
       name => 'REGRESSION_BUILD_FLAVOR_STR',
       description => 'JAMS build string',
       spec => "s",
    },
    {
       name => 'REGRESSION_TESTSET_STR',
       description => 'JAMS testset string',
       spec => "s",
    },
    {
       name => 'REGRESSION_Q_STATUS_STR',
       description => 'JAMS q Status string',
       spec => "s",
    },
    {
       name => 'REGRESSION_SPECNAME_STR',
       description => 'JAMS specname string',
       spec => "s",
    },
    {
       name => 'REGRESSION_PLATFORM_STR',
       description => 'JAMS platform string',
       spec => "s",
    },
    {
       name => 'REGRESSION_BRANCH_STR',
       description => 'JAMS branch string',
       spec => "s",
    },
    {
       name => 'REGRESSION_RUNSPEC_STR',
       description => 'JAMS runspec name string',
       spec => "s",
    },
    {
       name => 'REGRESSION_NDISPC_EXECUTABLE',
       description => 'Full path to ndispc executable',
       spec => "s",
       default => "/x/eng/localtest/noarch/share/nate/latest/bin/ndispc",
    },
   ]);

$Options->process();

#
# JAMS enumerations
# tools/jams/database.sql
#
my %JAMS_run_status_enum = (
   1  => 'starting',
   2  => 'pretesting',
   3  => 'testing',
   4  => 'posttesting',
   5  => 'aborting',
   6  => 'idle',
   7  => 'ended',
   8  => 'unknown',
);

my %JAMS_end_status_enum = (
   1  => 'TBD',
   2  => 'Completed',
   3  => 'Stopped',
   4  => 'Killed',
   5  => 'Aborted',
   6  => 'Unknown',
);




=head2 new

Object constructor

    $JAMS = JAMS->new([db_database => "dbname"],
                      [db_hostname => "dbhost"],
                      [db_username => "dbuser"],
                      [db_password => "dbpassword"]);

=over 4

=item db_database

=item db_hostname

=item db_user

=item db_password

=back

=cut

sub new {
    my $n = shift;
    my (%args) = @_;

    my $class = ref $n || $n;
    my $self = {};
    $self = $n if ref $n;
    bless $self, $class;

    my $Cfg = { sql => {} };
    $Cfg->{sql}->{database} = $args{db_database} || $Options->get(option => "DB_DATABASE");
    $Cfg->{sql}->{server} = $args{db_hostname} || $Options->get(option => "DB_HOSTNAME");
    $Cfg->{sql}->{username} = $args{db_user} || $Options->get(option => "DB_USER");
    $Cfg->{sql}->{password} = $args{db_password} || $Options->get(option => "DB_PASSWORD");


    $self->{dbh} = DBI->connect("DBI:mysql:".$Cfg->{sql}->{database}.
                        ";host=".$Cfg->{sql}->{server},
                        $Cfg->{sql}->{username},
                        $Cfg->{sql}->{password},
                        { RaiseError => 1 });
    $self->{dbh}->{mysql_auto_reconnect} = 1;

    $self->{Cfg} = $Cfg;

    return $self;
}

=head2 disconnect

Disconnect from DB.

   $JAMS->disconnect;

=cut

sub disconnectDB {
    my $_self = shift;

    $Log->comment("disconnect from DB: [$_self->{'Cfg'}->{'sql'}->{'server'}]");
    $_self->{'dbh'}->disconnect();
    1;
}


=head2 get_jams_image_id

    $JAMS->get_jams_image_id->(image => $args{image}, change_nbr => $args{change_nbr});

=over 4

=item image - Full path to image.tgz

=item change_nbr - P4 change number of image

=back

=cut
sub get_jams_image_id {
    my $self = shift;
    my (%args) = @_;
    my $image_path = $args{image};
    my $image_change = $args{change_nbr};
    my $image_id;
    my $row;
    my $sth;

    # lookup/create image (name,p4_change_nbr)
    $Log->debug("Looking for {$image_path,$image_change}");
    $sth = $self->{dbh}->prepare("SELECT * FROM image WHERE name=? AND p4_change_nbr=?");
    $sth->execute($image_path,$image_change);
    if ($sth->rows == 0) {
        my $errorString = "No image_id found for image.name: $image_path image.p4_change_nbr: $image_change\n";
        $Log->warn($errorString);
        die $errorString;
    }

    $row = $sth->fetchrow_hashref();

    $image_id = 0;
    $image_id = $row->{id} if $row;

    return $image_id;
}


=head2 generate_remainder_plan

Create a plan file for the remaining tests not executed on image/change_nbr

   $JAMS->generate_remainder_plan(change_nbr => 123456),
                                   image => "/x/eng/rlse/DOT/RboilermakerN/export/foo.tgz",
                                   plan_name => "Rboilermaker_10M_Qualified.plan",
                                   directory => "/x/testbeds/JAMS15")
                                   );

=over 4

=item image - Full path to image.tgz

=item change_nbr - P4 change number of image

=item plan_name - Filename of planfile for the remainder run

=item directory - Directory to move the remainder planfile to

=back

=cut
sub generate_remainder_plan {
    my $self = shift;
    my (%args) = @_;
    my $official_dir = $args{directory};
    my $plan_name = $args{plan_name};
    my $image_id;

    # Look up the path to tools/jams/utility/jams_plangen.pl
    my $plangen = find_file("tools/jams/utility/jams_plangen.pl");

    # Image ID
    eval {
        $image_id = $self->get_jams_image_id(image => $args{image}, change_nbr => $args{change_nbr});
    };
    if ($@) {
        # Not a JAMS15 run
        return;
    }

    # Make a temp dir
    my $tempdir = tempdir( CLEANUP => 1 );

    # $plangen --plan_dir $tmp_dir --image_run_to_plan $image_id
    my $plangenCmd = "$plangen --plan_dir $tempdir --image_run_to_plan $image_id";
    $Log->comment("$plangenCmd");
    my $plangenOut = `$plangenCmd 2>&1`;
    $Log->comment("$plangenOut");

    # mv $tmp_dir/$PLAN_NAME $OFFICIAL_DIR
    my $moveCmd = "mv $tempdir/$plan_name $official_dir";
    $Log->comment("$moveCmd");
    my $moveOut = `$moveCmd 2>&1`;
    $Log->warn($moveOut) if $moveOut;
}

=head2 store_jams_result

Store the result of a JAMS testcase_release execution associated with (image,change_nbr)

    $JAMS->store_jams_result(tcrid => 22,
                            image => "/x/eng/rlse/DOT/RboilermakerN/export/foo.tgz",
                            change_nbr => 123456,
                            log_dir => "/x/testbeds/logdir/10245678",
                            result => "PASS",
                            run_time => 22);

=over 4

=item tcrid - testcase_release.id number, typically from "REGRESSION_TESTCASE_RELEASE_ID"

=item image - Full path to image.tgz

=item change_nbr - P4 change number of image

=item result - Result string PASS,FAIL,SCRIPT,WARN,...

=item run_time - Execution time in seconds

=item log_dir - NATE log directory

=back

=cut
sub store_jams_result {
    my $self = shift;
    my (%args) = Params::Validate::validate(@_,
                    {
                        tcrid => { type => SCALAR },
                        specname_id => { type => SCALAR },
                        platform_id => { type => SCALAR },
                        branch_id => { type => SCALAR },
                        image => { type => SCALAR },
                        line_id => { type => SCALAR },
                        mode_id => { type => SCALAR },
                        testset_id => { type => SCALAR },
                        build_flavor_id => { type => SCALAR },
                        q_status_id => { type => SCALAR },
                        testcase_id => { type => SCALAR },
                        image => { type => SCALAR },
                        change_nbr => { type => SCALAR },
                        result => { type => SCALAR },
                        run_time => { type => SCALAR },
                        run_to_plan_id => { type => SCALAR },
                        run_stats_id => { type => SCALAR },
                        log_dir => { type => SCALAR },
                        schedule_id => { type => SCALAR },
                    });
    my $tcrid;
    my $image_path;
    my ($row, $image_id);
    my $image_change;
    my $sql;
    my $sth;
    my $test_group;

    $tcrid = $args{tcrid};
    $image_path = $args{image};
    $image_change = $args{change_nbr};
    if (!$tcrid) {
        # Not a JAMS 1.5 Run
        $Log->warn("Asked to store a result without specifying tcrid, is this JAMS15?");
        return -1;
    }

    # Lookup the test group for the execution record
    $Log->debug("Looking for test_group");

    if ($args{'schedule_id'}) {
        my $sql= "
        SELECT groups.name AS execution_group
        FROM schedule
        JOIN groups ON groups.id=schedule.execution_group 
        WHERE schedule.id=$args{'schedule_id'} 
        ";
        $row = $self->{dbh}->selectrow_hashref($sql);
        if ($row) { $test_group=$row->{'execution_group'} }
    };

    # if test_group did not come from the schedule then compute it as usual from the test case release, testcase and group 
    if (!defined($test_group)) {
        my $sql = "
        SELECT  IFNULL(NULLIF(TR.test_group,''),IFNULL(NULLIF(TC.test_group,''),S.test_group)) as test_group
        FROM testcase_release TR 
        JOIN testcase AS TC ON TC.id = TR.testcase_id
        JOIN script AS S ON S.id = TC.script_id
        WHERE
        TR.testcase_id      =$args{'testcase_id'} AND
        TR.line_id          =$args{'line_id'} AND
        TR.mode_id          =$args{'mode_id'} AND
        TR.build_flavor_id  =$args{'build_flavor_id'} AND
        TR.testset_id       =$args{'testset_id'} AND
        TR.q_status_id      =$args{'q_status_id'}
        ";
        $row = $self->{dbh}->selectrow_hashref($sql);
        if ($row) { $test_group=$row->{'test_group'} } else { $test_group = "NULL"; }
    }

    # Lookup or create image_id using (path,change#)
    $Log->debug("Looking for {$image_path,$image_change}");
    $sth = $self->{dbh}->prepare("SELECT * FROM image WHERE name=? AND p4_change_nbr=?");
    $sth->execute($image_path,$image_change);
    if ($sth->rows == 0) {
        my $insertSth = $self->{dbh}->prepare("INSERT INTO image SET name=?,p4_change_nbr=?");
        $Log->debug("Adding image for $image_path,$image_change");
        $insertSth->execute($image_path,$image_change);
        $sth->execute($image_path,$image_change);
    }

    $row = $sth->fetchrow_hashref();
    if (!$row) {
        $Log->warn("Unable to get/make image.id for ".
                    "image.path: $image_path image.p4_change_nbr: $image_change");
    }
    $image_id = 0;
    $image_id = $row->{id} if $row;

    $Log->debug("Inserting result for testcase_execution_id: $tcrid image_id: $image_id");
    $sql = "INSERT INTO testcase_execution SET ".
                "specname_id=?, ".
                "platform_id=?, ".
                "branch_id=?, ".
                "image_id=?, ".
                "line_id=?, ".
                "mode_id=?, ".
                "testset_id=?, ".
                "build_flavor_id=?, ".
                "q_status_id=?, ".
                "testcase_id=?, ".
                "execution_date=NOW(), ".
                "result=?, ".
                "run_time=?, ".
                "run_to_plan_id=?, ".
                "run_stats_id=?, ".
                "test_group=?, ".
                "log=?";

    $self->{dbh}->do($sql,{},
                    $args{specname_id},
                    $args{platform_id},
                    $args{branch_id},
                    $image_id,
                    $args{line_id},
                    $args{mode_id},
                    $args{testset_id},
                    $args{build_flavor_id},
                    $args{q_status_id},
                    $args{testcase_id},
                    $args{result},
                    $args{run_time},
                    $args{run_to_plan_id},
                    $args{run_stats_id},
                    $test_group,
                    $args{log_dir}
    );
    return $self->{dbh}->{ q{mysql_insertid}};
}


sub _normalize_dbh_runid {
    my $arg = shift;

    my @interpret = (
        { regex => qr/did_bad_happen/xi, interpret => 'dbh' },
        { regex => qr/log-exists/xi, interpret => 'incomplete_gsi' },
        { regex => qr/console.log-exists/xi, interpret => 'console_log' },
        { regex => qr/quorum/xi, interpret => 'quorum' },
        { regex => qr/ls-vmware-core/xi, interpret => 'vmware_core' },
        { regex => qr/pre-mroot-cores|sim-cores|mroot-cores|mroot-etc-mcores|mroot-etc-crash-core|sim-cores-ucore-bz2|mroot-etc-crash-ucore-bz2/xi, interpret => 'user_cores' },
        { regex => qr/mroot-kcores|mroot-etc-crash-core-nz|sim-kcores/xi, interpret => 'panic' },
    );

    my $normalized = $arg;
    foreach my $match (@interpret) {
        if ($arg =~ $match->{regex}) {
            $normalized = $match->{interpret};
        }
    }

    return $normalized;
}

=head2 summarize_test

Summarize the test NATE::result

=over 4

=item 1 - NATE::Result

=back

=cut
sub summarize_test {
    return "test";
}



=head2 summarize_shoc

Summarize the shoc NATE::result

=over 4

=item 1 - NATE::Result

=back

=cut
sub summarize_shoc {
    return "shoc";
}

=head2 summarize_qc_push

Summarize the qc_push NATE::result

=over 4

=item 1 - NATE::Result

=back

=cut
sub summarize_qc_push {
    return "qc_push";
}

=head2 summarize_dbh

Summarize dbh Nate results into the format required JAMS::summarize_results, should be
passed to JAMS::summarize_results

=over 4

=item 1 - NATE::Result

=back

=cut
sub summarize_dbh {
    my $result = shift;

    my $runid = $result->runid();
    my @parsed;

    my @parse = split /\//, $runid;

    return _normalize_dbh_runid($parse[$#parse]);

}

=head2 summarize_results

    %ProblemSummary = $JAMS->summarize_results(
                summarize => "JAMS::summarize_dbh",
                results => [@DBH_Nate_Results],
                source => 'dbh',
                problems => {%ProblemSummary},
                worse_than => NATE::Result->severity()
            );

=over 4

=item summarize - method name to use for summarizing a nate result into a problem string

=item results - array ref of NATE results to store

=item source - source name of the results

=item problems - existing problems hash from summarize_results to append to

=item worse_than - Only summarize results worse than this severity

=back

=cut
sub summarize_results {
    my $self = shift;
    my %args = Params::Validate::validate(@_,
            {
            summarize => { type => SCALAR },
            results => { type => ARRAYREF },
            source => { type => SCALAR },
            problems => { type => HASHREF },
            worse_than => { type => SCALAR },
            }
        );

    my $summarize = $args{summarize};
    my $results = $args{results};
    my $source = $args{source};
    my %problems = %{$args{problems}};
    my $worse_than = $args{worse_than};

    foreach my $r (@{$results}) {
        no strict 'refs';

        if ($r->severity() < $worse_than) {
            next;
        }

        my $summary = &$summarize($r);
        my $key = $source.":".$summary;
        if (exists($problems{$key})) {
            push @{$problems{$key}->{results}}, $r;
        } else {
            $problems{$key} = {
                summary => $summary,
                results => [$r],
                source => $source,
            };
        }
    }
    return %problems;
}

#
# Function: _find_codeline_id
#
# Parameters:
#
#   1 - codeline name
#
# Returns:
#
#   codeline id
#
sub _find_codeline_id {
    my $self = shift;
    my $dbh = $self->{dbh};
    my $codeline = shift;
    my $codeline_id = -1;

    $codeline_id = $dbh->selectrow_array(
            qq{SELECT id
                 FROM triage_codeline
                 WHERE name = "$codeline"});
    if (defined $codeline_id) {
        #do nothing, we already got it.
        $Log->comment("codeline($codeline) exists and has id($codeline_id)");
    } else {
        my $rows_updated = $dbh->do(
                qq{
                INSERT INTO triage_codeline
                            (id, name)
                     VALUES (0,  "$codeline")});
        $codeline_id = $dbh->{ q{mysql_insertid}};
        $Log->comment("codeline($codeline) created with id($codeline_id)");
    }
    return($codeline_id);
}

#
# Function: _store_nate_result
#
# Parameters:
#
#   problem_log_id - the problem this nate result refers to
#
#   result - the NATE::Result to store a copy of
#
# Returns: none
#
sub _store_nate_result {
    my $self = shift;
    my %args = Params::Validate::validate(@_, {
                        problem_log_id => { type => SCALAR },
                        result => { type => OBJECT },
                    });

    my $dbh = $self->{dbh};
    my $result = $args{result};
    my $problem_log_id = $args{problem_log_id};

    my $sql = qq{
        INSERT INTO problem_nate_results
            (id,
            problem_log_id,
            messages,
            stacktrace,
            runid,
            timestamp,
            type,
            severity)
        VALUES
            (0,
            ?,
            ?,
            ?,
            ?,
            FROM_UNIXTIME(?),
            ?,
            ?)
    };

    my $rv = $dbh->do($sql,
                        {},
                        $problem_log_id,
                        (join "\n",$result->messages()),
                        $result->stacktrace(),
                        $result->runid(),
                        $result->timestamp(),
                        $result->type(),
                        $result->severity()
                        );


    return;
}

#
# Function: _find_problem_id
#
# Parameters:
#
#   summary - Summary text of the problem
#
#   source - Name of the problem src
#
# Returns:
#
#   Problem id
#
sub _find_problem_id {
    my $self = shift;
    my %args = Params::Validate::validate(@_, {
                            summary => { type => SCALAR },
                            source => { type => SCALAR },
                        });

    my $dbh = $self->{dbh};
    my $summary = $args{summary};
    my $source = $args{source};

    my $sql;
    my $row;

    # Look up dbh
    $sql = qq{
            SELECT * FROM problem WHERE summary = '$summary' AND source = '$source'
        };
    if ($row = $dbh->selectrow_hashref($sql)) {
        return $row->{id};
    } else {
        # If no rows INSERT
        $sql = qq{
                    INSERT INTO problem
                        (id,
                        summary,
                        source,
                        description,
                        css_class,
                        priority)
                    VALUES
                        (0,
                        ?,
                        ?,
                        ?,
                        ?,
                        ?)
        };
        my $ret = $dbh->do($sql,{},$summary,$source,'',$summary,99);
        if (!$ret) {
            $Log->error("Unable to insert $sql");
        }

        my $problem_id = $dbh->{ q{mysql_insertid}};
        return $problem_id;
    }

}

sub _valid_group {
    my $self = shift;
    my $group_name = shift;

    my $dbh = $self->{dbh};
    my $sql = qq{
        SELECT
            id
        FROM
            groups
        WHERE
            name = '$group_name'
    };

    my $row = $dbh->selectall_hashref($sql, ['id']);
    if (scalar(keys(%$row))) {
        return $group_name;
    }
    return;
}

sub _find_testgroup {
    my $self = shift;
    my $row = shift;
    my $schedule_id = shift;

    if ($schedule_id) {
        my $sql= "
        SELECT groups.name AS execution_group
        FROM schedule
        JOIN groups ON groups.id=schedule.execution_group 
        WHERE schedule.id=$schedule_id 
        ";
        my $row = $self->{dbh}->selectrow_hashref($sql);
        if ($row) { 
            my $test_group=$row->{'execution_group'};
            return $test_group if $self->_valid_group($test_group);
        }
    };

    if ($self->_valid_group($row->{tcr_tg})) {
        return $row->{tcr_tg};
    } elsif ($self->_valid_group($row->{tc_tg})) {
        return $row->{tc_tg};
    } elsif ($self->_valid_group($row->{s_tg})) {
        return $row->{s_tg};
    } else {
        return "Unknown";
    }
}

sub _find_tcr_row {
    my $self = shift;
    my $tcrid = shift;

    my $dbh = $self->{dbh};
    my $sql = qq{
        SELECT
            testcase_release.test_group AS tcr_tg,
            testcase.test_group AS tc_tg,
            script.test_group AS s_tg

        FROM
            testcase_release
        LEFT JOIN
            testcase
        ON
            testcase.id = testcase_release.testcase_id
        LEFT JOIN
            script
        ON
            script.id = testcase.script_id
        WHERE
            testcase_release.id = '$tcrid'
    };
    my $row;

    $row = $dbh->selectrow_hashref($sql);

    return $row;

}

=head2 store_triage_item

Store a triage item and associated dbh entries

$JAMS->store_triage_item(
                testcase_id => ...,
                testcase_release_id => ...,
                testcase_execution_id => ...,
                dbh => ...,
                testname => ...,
                mode => ...,
                testgroup  => ...,
                build => ...,
                codeline => ...,
                runid => ...,
                log => ...,
                history => ...,
                problems => ...,
            );
=cut

sub store_triage_item {
    my $self = shift;
    my %args = Params::Validate::validate(@_, {
                        testcase_id => { type => SCALAR },
                        testcase_release_id => { type => SCALAR },
                        testcase_execution_id => { type => SCALAR },
                        problems => { type => HASHREF },
# These should go away, available from testcase_id
                        testname => { type => SCALAR },

# This is derivable from testcase_execution_id or testcase_release_id
                        mode => { type => SCALAR },

# We aren't using testgroup anymore and we can find ownership via testcase_id
                        testgroup => { type => SCALAR },

# This should be in testcase_execution_id
                        build => { type => SCALAR },
                        codeline => { type => SCALAR },
                        runid => { type => SCALAR },
                        log => { type => SCALAR },

# Should be able to build this link on the fly
                        history => { type => SCALAR },
                        schedule_id => { type => SCALAR },
                    }
                    );

    my $dbh = $self->{dbh};
    my $sql;
    my %problems = %{$args{problems}};
    my $codeline_id = $self->_find_codeline_id($args{codeline});
    my $tcr_row = $self->_find_tcr_row($args{testcase_release_id});
    my $test_group = $self->_find_testgroup($tcr_row,$args{'schedule_id'});
    my $triage_id;

    $sql = qq{
            INSERT INTO triage_item
                            (
                            id,
                            triage_user_id,
                            triage_codeline_id,
                            test_name,
                            logdir,
                            log_link,
                            history_link,
                            triage_state,
                            result,
                            hours,
                            burts,
                            when_created,
                            build,
                            test_group,
                            mode,
                            testcase_id,
                            testcase_release_id,
                            testcase_execution_id
                            )
                    VALUES (
                            0,
                            -1,
                            ?,
                            ?,
                            ?,
                            ?,
                            ?,
                            'new',
                            'unknown',
                            '0',
                            '',
                            NOW(),
                            ?,
                            ?,
                            ?,
                            ?,
                            ?,
                            ?
                            )};
    my $rows_updated = $dbh->do($sql,{},
                            $codeline_id,
                            $args{testname},
                            $args{runid},
                            $args{log},
                            $args{history},
                            $args{build},
                            $test_group,
                            $args{mode},
                            $args{testcase_id},
                            $args{testcase_release_id},
                            $args{testcase_execution_id}
                            );
    if ($rows_updated == 1) {
        #great, the insert worked.
        $Log->comment("INSERT succeeded.");
        $triage_id = $dbh->{mysql_insertid};
        foreach my $problem (values %problems) {
            my $problem_id = $self->_find_problem_id(
                                summary => $problem->{summary},
                                source => $problem->{source},
                                #results => $problem->{results}
                            );
            $sql = qq{
                        INSERT INTO problem_log
                                (
                                id,
                                triage_id,
                                problem_id
                                )
                        VALUES (
                                0,
                                ?,
                                ?
                                )
                    };
            $dbh->do($sql,{},$triage_id,$problem_id);
            my $problem_log_id = $dbh->{mysql_insertid};
            foreach my $result (@{$problem->{results}}) {
                $self->_store_nate_result(problem_log_id => $problem_log_id, result => $result);
            }
        }
    } else {
        $Log->error("The insert failed. sql($sql), expected(1), got($rows_updated).")
    }
    return $triage_id;

}


=head2 jams_run_stats

Create and populate JAMS run_stats table.

    ($ok, $error) = $JAMS->jams_run_stats( 'action'   => <action>,
                                         '<column>' => <value>,
    );
Returns a 1 upon success in scalar context and (1, '') in list context.
Returns a 0 upon failure in scalar context and (0, '') in list context.


Each action will need a job_name and run_to_plan_name for updating and inserting to the correct row.
Different actions will work on different columns in the JAMS run_stats table.
The values used for the key shown as <column> above is the column of the same name in the
JAMS.run_stats table. 1-N columns can be given, required ones are checked for. If any are given
to an action that are not needed for that action they are not put into the sql statement.

See below for information on each action.

=head3 action = create row

Create a new run_stats row.

    ($ok, $error) = $JAMS->jams_run_stats(
          'action'           => 'create row',
          'job_name'         => $job_name,
          'run_to_plan_id' => $run_to_plan_id,
          'run_to_plan_name' => $run_to_plan_name,
          'run_spec_id'      => $run_spec_id,
          ['controller'      => $controllers],
          ['unixclients'     => $unixclients],
          ['winclients'      => $winclients],
          ['filers'          => $filers],
          ['plan_count'      => $num_of_test + $num_of_skips],
          ['plan_skipped'    => $num_of_skips],
          ['plan_ignored'    => $num_of_ignores],
          ['plan_to_test'    => $num_of_test],
          ['qc_push'         => $qc_push],
          ['triage'          => $triage],
          ['simplicity'      => $simplicity],
          ['line_str'        => $JAMS->{'run_spec'}{$run_spec_id}{'Line'}],
          ['mode_str'        => $JAMS->{'run_spec'}{$run_spec_id}{'Mode'}],
          ['build_str'       => $JAMS->{'run_spec'}{$run_spec_id}{'Build'}],
          ['suite_str'       => $JAMS->{'run_spec'}{$run_spec_id}{'Testset'}],
          ['q_status_str'    => $JAMS->{'run_spec'}{$run_spec_id}{'Q_Status'}],
          ['spec_str'        => $JAMS->{'run_spec'}{$run_spec_id}{'Specname'}],
          ['platform_str'    => $JAMS->{'run_spec'}{$run_spec_id}{'Platform'}],
          ['branch_str'      => $JAMS->{'run_spec'}{$run_spec_id}{'Branch'}],
      );

=over

=item action => 'create row'

Used to create new run_stats row.

=item job_name => <job_name>

The RUN_ID this test is being executed under

=item run_to_plan_id => <run to plan id>

The run_to_plan id , row index, this test is being executed under

=item run_to_plan_name => <run to plan name>

The run_to_plan name this test is being executed under

=item run_spec_id => <int>

Runspec index value

=item controller => <str>

The host controller of this test, will be set to 'none' if not given

=item unixclients => <str>

Comma separated list of unixclients (unix) used for this RUN_ID, set to 'none' if not given.

=item winclients => <str>

Comma separated list of window clients used for this RUN_ID, set to 'none' if not given.

=item filers  => <str>

Comma separated list of filers (nodes) used for this RUN_ID, set to 'none' if not given.

=item plan_count => <int>

Number of test mapped to this runspec (plan), set to 0 if not given

=item plan_skipped => <int>

Number of test skipped for this runspec (plan), set to 0 if not given

=item plan_ignored => <int>

Number of test ignored for this plan file generation, set to 0 if not given

=item plan_to_test => <int>

Number of test planned to be tested for this runspec (plan), set to 0 if not given

=item qc_push => <int>

QC_PUSH param setting, set to 0 if not given

=item triage => <int>

TRIAGE param setting, set to 0 if not given

=item simplicity => <int>

SIMPLICITY param setting, set to 0 if not given

=item line_str => <str>

The JAMS line char string

=item mode_str => <str>

The JAMS mode char string

=item build_str => <str>

The JAMS build_flavor char string

=item suite_str => <str>

The JAMS testset char string

=item q_status_str => <str>

The JAMS q_status char string

=item spec_str => <str>

The JAMS specname char string

=item platform_str => <str>

The JAMS platform char string

=item branch_str => <str>

The JAMS branch char string

=back

=over

=item end_status will be set to 'TBD' upon row creation.

=item run_status will be set to 'starting' upon row creation.

=item All test_<RESULT>, pretest_<RESULT> and posttest_<RESULT> will be set to 0.

=back


=head3 action => 'update hardware'

Update run_stats row for the respective job_name and run_to_plan_name with the given
hardware info. 1-N hardware columns can be given.

    ($ok, $error) = $JAMS->jams_run_stats(
                   'action'           => 'update hardware',
                   'job_name'         => $job_name,
                   'run_to_plan_name' => $run_to_plan_name,
                   ['controller'      => <str>],
                   ['unixclients']    => <str>],
                   ['winclients ']    => <str>],
                   ['filers']         => <str>],

    );

=over

=item action => 'update hardware'

Used to update run_stats row with new hardware info.

=item job_name => <job_name>

The RUN_ID this test is being executed under

=item run_to_plan_name => <run to plan id>

The run_to_plan ID this test is being executed under

=item controller => <str>

The host controller of this test.

=item unixclients => <str>

Comma separated list of unixclients (unix) used for this RUN_ID.

=item winclients => <str>

Comma separated list of window clients used for this RUN_ID.

=item filers  => <str>

Comma separated list of filers (nodes) used for this RUN_ID.

=back

=head3 $_action => 'update run_status'

Update run_stats row for the respective job_name and run_to_plan_name with the given
run_status. This is intended to be used to show the state of testing through
the pre and post stages and the actual testing.

    ($ok, $error) = $JAMS->jams_run_stats(
                   'action'           => 'update run_status',
                   'job_name'         => $job_name,
                   'run_to_plan_name' => $run_to_plan_name,
                   'run_status'       => <str>,
    );

See above actions for option description.

Supported run_status values:

=over

=item starting = Starting stage

=item pretesting = JAMS PreTest scripts are running

=item testing = JAMS Test is running

=item posttesting = JAMS PostTest is running

=item aborting = JAMS is aborting

=item idle = JAMS is idle

=item ended = JAMS has ended

=item unknown = Unknown state

=back

=head3  $_action => 'update end_status'

Update run_stats row for the respective job_name and run_to_plan_name with the given
end_status. This is intended to be used to show the state of the given RUN_ID


    ($ok, $error) = $JAMS->jams_run_stats(
                   'action'           => 'update end_status',
                   'job_name'         => $job_name,
                   'run_to_plan_name' => $run_to_plan_name,
                   'end_status'       => <str>,
    );

See above actions for option description.

Supported end_status values:

=over

=item TBD = end status to be determined

=item Completed = testing has completed, ie: plan file finished

=item Stopped = run has been stopped

=item Killed = testing has been killed before completion

=item Aborted = testing has been aborted

=item Unknown = state not known

=back

=head3 $_action => 'incr test count'

Update run_stats row for the respective job_name and run_to_plan_name for the
given test type and nate result. This is one action that takes options that are
not run_stats column names. This is inteneded to be used to show how many
test of each type have ran and and a talley of their results.



    ($ok, $error) = $JAMS->jams_run_stats(
                   'action'           => 'incr test count',
                   'job_name'         => $job_name,
                   'run_to_plan_name' => $run_to_plan_name,
                   'type'             => <pretest|test|posttest>,
                   'result'           => <nate result string, uppercase>,
    );

See above actions for option description.


Supported types:

=over

=item pretest

=item test

=item posttest

=back

Supported NATE Results:

=over

=item SCRIPT

=item CONFIG

=item FATAL

=item FAIL

=item INSPECT

=item NA

=item WARN

=item INFO

=item PARAM

=item PASS

=item OTHER - for any NATE result that is not defined in NATE

=back


=head3 $_action => 'update last testcase'

Update run_stats row for the respective job_name and run_to_plan_name with the
JAMS Testcase ID being tested. Only test of type=test have testcase IDs.
Pre and post tests in JAMS do not have ID.

    ($ok, $error) = $JAMS->jams_run_stats(
                   'action'           => 'foo',
                   'job_name'         => $job_name,
                   'run_to_plan_name' => $run_to_plan_name,
                   'last_testcase_id' => <int>,
    );

See above actions for option description.

=over

=item last_testcase_id = integer value of the JAMS test

=back


=head3 $_action => 'incr panic found'

Update run_stats row for the respective job_name and run_to_plan_name panic_found
counter.

    ($ok, $error) = $JAMS->jams_run_stats(
                   'action'           => 'foo',
                   'job_name'         => $job_name,
                   'run_to_plan_name' => $run_to_plan_name,
     );

=cut



sub jams_run_stats {

   my $self = shift;
   my (%_args) = @_;
   my $_return = 0;
   my $_error = '';
   my ($_sql, $_sth, $_str, $_t, $_r, $_tr, $_v, $_v2, $_ok, $_enum,
      $_col,
   );

   my $_action = delete($_args{'action'});
   my $_run_stats_tab = 'run_stats';


   $Log->debug("::: JAMS action [$_action] into $_run_stats_tab\n");
   #
   # create a row for a job
   #
   # TODO: should we have checks for run_spec_id and test_count, and throw warning
   if ($_action eq 'create row') {
      $Log->debug("::: if action: [$_action]");
      # make sure have run id
      if ( !defined($_args{'job_name'}) ) {
         $_error = 'need job_name string';
         return( wantarray ? ($_return, $_error) : $_return);
      }
      # make sure we have run_to_plan_name
      if ( !defined($_args{'run_to_plan_name'}) ) {
         $_error = 'need run_to_plan_name string';
         return( wantarray ? ($_return, $_error) : $_return);
      }
      # make sure we have run_to_plan_name
      if ( !defined($_args{'run_to_plan_id'}) ) {
         $_error = 'need run_to_plan_id value';
         return( wantarray ? ($_return, $_error) : $_return);
      }

      # make sure we have run_to_plan_name
      if ( !defined($_args{'run_spec_id'}) ) {
         $_error = 'need run_spec_id value';
         return( wantarray ? ($_return, $_error) : $_return);
      }

      $Log->debug("::: JAMS INSERT INTO $_run_stats_tab row " .
                    "job_name $_args{'job_name'} " .
                    "run_to_plan_name $_args{'run_to_plan_name'}"
      );

      # build sql insert statement
      my @_nate_results = ( 'PASS',    'PARAM', 'INFO',  'WARN',   'NA',
                            'INSPECT', 'FAIL',  'FATAL', 'CONFIG', 'SCRIPT',
                            'OTHER'
      );
      $_sql  = "INSERT INTO $_run_stats_tab SET ";
      $_sql .= sprintf("job_name = '%s', ", $_args{'job_name'});
      $_sql .= sprintf("run_to_plan_id = '%s', ", $_args{'run_to_plan_id'});
      $_sql .= sprintf("run_to_plan_name = '%s', ", $_args{'run_to_plan_name'});
      $_sql .= sprintf("when_started = NOW(), ");
      $_sql .= sprintf("last_updated = NOW(), ");
      $_sql .= sprintf("last_testcase_id = 0, "); ### SLP
      $_sql .= sprintf("controller =  '%s', ",  $_args{'controller'}  || 'none');
      $_sql .= sprintf("unixclients =  '%s', ", $_args{'unixclients'} || 'none');
      $_sql .= sprintf("winclients =  '%s', ",  $_args{'winclients'}  || 'none');
      $_sql .= sprintf("filers =  '%s', ",      $_args{'filers'}      || 'none');
      $_sql .= sprintf("end_status =  '%s', ",  'TBD');
      $_sql .= sprintf("run_status =  '%s', ",  'starting');
      $_sql .= sprintf("run_spec_id =  '%u', ",  $_args{'run_spec_id'});
      $_sql .= sprintf("plan_count =  '%u', ",   $_args{'plan_count'}   || 0);  # dummy check this
      $_sql .= sprintf("plan_skipped =  '%u', ", $_args{'plan_skipped'} || 0);
      $_sql .= sprintf("plan_ignored =  '%u', ", $_args{'plan_ignored'} || 0);
      $_sql .= sprintf("plan_to_test =  '%u', ", $_args{'plan_to_test'} || 0);
      foreach $_t ('test', 'posttest', 'pretest' ) {
         $_sql .= sprintf("%s =  '0', ", $_t);
         foreach $_v ( @_nate_results ) {
            $_sql .= sprintf("%s_%s =  '0', ", $_t, $_v);
         }
      }
      $_sql .= sprintf("qc_push =  '%u', ", $_args{'qc_push'} || 0);
      $_sql .= sprintf("triage =  '%u', ", $_args{'triage'} || 0);
      $_sql .= sprintf("simplicity =  '%u', ", $_args{'simplicity'} || 0);
      $_sql .= sprintf("line_str =  '%s', ", $_args{'line_str'} || 'line');
      $_sql .= sprintf("mode_str =  '%s', ", $_args{'mode_str'} || 'mode');
      $_sql .= sprintf("build_str =  '%s', ", $_args{'build_str'} || 'build');
      $_sql .= sprintf("suite_str =  '%s', ", $_args{'suite_str'} || 'suite');
      $_sql .= sprintf("q_status_str =  '%s', ", $_args{'q_status_str'} || 'q_status');
      $_sql .= sprintf("spec_str =  '%s', ", $_args{'spec_str'} || 'spec');
      $_sql .= sprintf("platform_str =  '%s', ", $_args{'platform_str'} || 'platform');
      $_sql .= sprintf("branch_str =  '%s', ", $_args{'branch_str'} || 'branch');


      # strip last ','
      $_sql =~ s/,\s$//;
      $Log->debug("sql:: $_sql");

      # send the query
      $_ok = $self->{'dbh'}->do($_sql);
      if ( defined($_ok) ) {
         $Log->debug("::: insert successful [$_action]");
         $_return = $self->{dbh}->{ q{mysql_insertid}};
      }
      else {
         $Log->warn("::: insert failed [$_action]");
         $_error = $self->{'dbh'}->errstr;
      }
   }
   #
   # update hardware
   #
   elsif ( $_action eq 'update hardware') {
      $Log->debug("::: if action: [$_action]");
      $_v = 0;
      # make check we have hardware columns
      foreach $_t ('controller', 'unixclients', 'winclients', 'filers' ) {
         $_v++ if defined($_args{$_t});
      }
      $Log->debug("::: update hardware columns $_v");

      if    ( !defined($_args{'job_name'}) )       { $_error = 'job_name not defined'; }
      elsif ( $_v == 0 )                           { $_error = 'no hardware given'; }
      else {
          # make the sql
          $Log->debug("::: update hardware");
          $_sql  = "UPDATE $_run_stats_tab SET ";
          $_sql  .= "controller   = '$_args{'controller'}', "  if  $_args{'controller'};
          $_sql  .= "unixclients  = '$_args{'unixclients'}', "     if  $_args{'unixclients'};
          $_sql  .= "winclients   = '$_args{'winclients'}', "  if  $_args{'winclients'};
          $_sql  .= "filers       = '$_args{'filers'}', "      if  $_args{'filers'};
          $_sql  .= "last_updated = NOW() ";
          #$_sql =~ s/,\s$/ /;
          $_sql  .= "WHERE job_name = '$_args{'job_name'}'";

          # strip last ','
          $Log->debug("sql:: $_sql");

          # send the query
          $_ok = $self->{'dbh'}->do($_sql);
          if ( defined($_ok) ) {
             $Log->debug("::: insert successful [$_action]");
             $_return = 1;
          }
          else {
             $Log->warn("::: insert failed [$_action]");
             $_error = $self->{'dbh'}->errstr;
          }
      }
   }
   #
   # update run_status
   #
   elsif ( $_action eq 'update run_status' ) {
      $Log->debug("::: if action: [$_action]");

      # find the run_status enumeration
      $_v = 0;
      foreach $_enum ( keys %JAMS_run_status_enum ) {
         if ( $_args{'run_status'} eq $JAMS_run_status_enum{$_enum} ) {
            $_v = $_enum;
         }
      }
      if    ( !defined($_args{'job_name'}) )         { $_error = 'job_name not defined'; }
      elsif ( !defined($_args{'run_to_plan_name'}) ) { $_error = 'run_to_plan_name not defined'; }
      elsif ( $_v == 0 )                             { $_error = "unmapped run_status: $_args{'run_status'}"; }
      else {
         $Log->debug("::: update run_status [$_v] $JAMS_run_status_enum{$_v}");
         # make the sql
         $_sql  = "UPDATE $_run_stats_tab SET run_status = $_v, ";
         $_sql  .= "      last_updated = NOW() ";
         $_sql  .= "WHERE job_name = '$_args{'job_name'}' ";
         $_sql  .= "  AND run_to_plan_name = '$_args{'run_to_plan_name'}'";
         # send the query
          $_ok = $self->{'dbh'}->do($_sql);
          if ( defined($_ok) ) {
             $Log->debug("::: update successful [$_action]");
             $_return = 1;
          }
          else {
             $Log->warn("::: update failed [$_action]");
             $_error = $self->{'dbh'}->errstr;
          }
      }
   }
   #
   # update end_status
   #
   elsif ( $_action eq 'update end_status' ) {
      $Log->comment("::: if action: [$_action]");

      # find the end_status enumeration
      $_v = 0;
      foreach $_enum ( keys %JAMS_end_status_enum ) {
         if ( $_args{'end_status'} eq $JAMS_end_status_enum{$_enum} ) {
            $_v = $_enum;
         }
      }
      if ( $_args{'run_status'} ) {
         $_v2 = 0;
         foreach $_enum ( keys %JAMS_run_status_enum ) {
            if ( $_args{'run_status'} eq $JAMS_run_status_enum{$_enum} ) {
               $_v2 = $_enum;
            }
         }

      }

      if    ( !defined($_args{'job_name'}) )         { $_error = 'job_name not defined'; }
      elsif ( !defined($_args{'run_to_plan_name'}) ) { $_error = 'run_to_plan_name not defined'; }
      elsif ( $_v == 0 )                             { $_error = "unmapped end_status: $_args{'end_status'}"; }
      elsif ( $_args{'run_status'} && $_v2 == 0 )    { $_error = "unmapped run_status: $_args{'run_status'}"; }
      else {
         $Log->comment("::: update end_status [$_v] $JAMS_end_status_enum{$_v}");
         # make the sql
         $_sql  = "UPDATE $_run_stats_tab SET ";
         $_sql  .= "end_status = $_v, ";
         $_sql  .= "run_status = $_v2, " if $_args{'run_status'};
         $_sql  .= "last_updated = NOW(), ";
         $_sql  .= "when_ended   = NOW() ";
         $_sql  .= "WHERE job_name = '$_args{'job_name'}' ";
         $_sql  .= "  AND run_to_plan_name = '$_args{'run_to_plan_name'}'";
         $Log->comment("::: update end_status sql: $_sql");
         # send the query
          $_ok = $self->{'dbh'}->do($_sql);
          if ( defined($_ok) ) {
             $Log->comment("::: update successful [$_action]");
             $_return = 1;
          }
          else {
             $Log->warn("::: update failed [$_action]");
             $_error = $self->{'dbh'}->errstr;
          }
      }
   }
   #
   # increment test count
   #
   elsif ($_action eq 'incr test count') {
      $Log->comment("::: if action: [$_action] type [$_args{'type'}]");
      ###if ($_args{'result'} eq 'FATAL') {
      ###   $_args{'result'} = 'FAIL';
      ###   $Log->comment("::: convert result FATAL to FAIL");
      ###}

      # make sure have the needed columns
      if    ( !defined($_args{'job_name'}) )         { $_error = 'job_name not defined'; }
      elsif ( !defined($_args{'run_to_plan_name'}) ) { $_error = 'run_to_plan_name not defined'; }
      elsif ( !defined($_args{'type'}) )             { $_error = 'test type not defined'; }
      elsif ( !defined($_args{'result'}) )           { $_error = 'result not defined'; }
      elsif ( $_args{'type'} !~ /[pre|post]?test/)   { $_error = 'type not known'; }
      elsif ( $_args{'result'} !~ /SCRIPT|CONFIG|FATAL|FAIL|CONFIG|INSPECT|NA|WARN|INFO|PARAM|PASS|OTHER/)
         { $_error = 'result not known'; }
      else {
        # make the column name from type and result
        $_col = sprintf("%s_%s", $_args{'type'}, $_args{'result'});
        $Log->comment("::: incr $_args{'job_name'} [$_col] \n");
        # get the current value for type and type_result
        $_sql  = "SELECT $_args{'type'} AS Type, ";
        $_sql  .= "      $_col AS TypeResult FROM $_run_stats_tab ";
        $_sql  .= "WHERE job_name = '$_args{'job_name'}' ";
        $_sql  .= "  AND run_to_plan_name = '$_args{'run_to_plan_name'}'";
        $_v = $self->{'dbh'}->selectrow_hashref($_sql);
        $Log->comment("::: got type $_args{'type'} =  $_v->{'Type'}, $_col = $_v->{'TypeResult'}");
        # increment value and update
        $_t = $_v->{'Type'} + 1;
        $_tr = $_v->{'TypeResult'} + 1;
        # update type and type_result with increment
        $Log->comment("::: set type $_args{'type'} => $_t, $_col => $_tr");
        $_sql  = "UPDATE $_run_stats_tab SET $_args{'type'} = $_t, ";
        $_sql  .= "      $_col = $_tr, ";
        $_sql  .= "      last_updated = NOW() ";
        $_sql  .= " WHERE job_name = '$_args{'job_name'}' ";
        $_sql  .= "  AND run_to_plan_name = '$_args{'run_to_plan_name'}'";

        $_ok = $self->{'dbh'}->do($_sql);
         if ( defined($_ok) ) {
            $Log->comment("::: update successful [$_action]");
            $_return = 1;
         }
         else {
            $Log->warn("::: insert failed [$_action]");
            $_error = $self->{'dbh'}->errstr;
         }
      }
   }
   #
   #
   # set the last testcase id
   #
   elsif ($_action eq 'update last testcase') {
      $Log->debug("::: if action: [$_action]");
      if    ( !defined($_args{'job_name'}) )           { $_error = 'job_name not defined'; }
      elsif ( !defined($_args{'run_to_plan_name'}) )   { $_error = 'run_to_plan_name not defined'; }
      elsif ( !defined($_args{'last_testcase_id'}))    { $_error = 'last_testcase_id not defined'; }
      else {
         $Log->debug("::: update last_testcase_id $_args{'last_testcase_id'} \n");
         $_sql  = "UPDATE $_run_stats_tab SET last_testcase_id = $_args{'last_testcase_id'}, ";
         $_sql  .= "      last_updated = NOW() ";
         $_sql .= " WHERE job_name = '$_args{'job_name'}' ";
         $_sql .= "   AND run_to_plan_name = '$_args{'run_to_plan_name'}'";
         $_ok = $self->{'dbh'}->do($_sql);
         if ( defined($_ok) ) {
            $Log->debug("::: update successful [$_action]");
            $_return = 1;
         }
         else {
            $Log->warn("::: insert failed [$_action]");
            $_error = $self->{'dbh'}->errstr;
         }
      }
   }
   #
   # incr panic found count
   #
   elsif ($_action eq 'incr panic found') {
      $Log->debug("::: if action: [$_action]");
      if    ( !defined($_args{'job_name'}) )           { $_error = 'job_name not defined'; }
      elsif ( !defined($_args{'run_to_plan_name'}) )   { $_error = 'run_to_plan_name not defined'; }
      else {
         $Log->debug("::: incr panic_found\n");
         # get the current value of panic_found
         $_sql  = "SELECT panic_found FROM $_run_stats_tab ";
         $_sql  .= "WHERE job_name = '$_args{'job_name'}' ";
         $_sql  .= "  AND run_to_plan_name = '$_args{'run_to_plan_name'}'";
         $_v = $self->{'dbh'}->selectrow_hashref($_sql);
         $_t = $_v->{'panic_found'} + 1;
         $Log->debug("::: incr panic_found $_v->{'panic_found'} -> $_t");
         # update value
         $_sql  = "UPDATE $_run_stats_tab SET panic_found = $_t, ";
         $_sql .= "       last_updated = NOW() ";
         $_sql .= " WHERE job_name = '$_args{'job_name'}' ";
         $_sql .= "   AND run_to_plan_name = '$_args{'run_to_plan_name'}'";
         $_ok = $self->{'dbh'}->do($_sql);
         if ( defined($_ok) ) {
            $Log->debug("::: update successful [$_action]");
            $_return = 1;
         }
         else {
            $Log->warn("::: update failed [$_action]");
            $_error = $self->{'dbh'}->errstr;
         }
      }
   }
   # get a run_stats row
   elsif ( $_action eq 'get row' ) {
      $Log->debug("::: if action: [$_action]");
      if    ( !defined($_args{'job_name'}) )           { $_error = 'job_name not defined'; }
      elsif ( !defined($_args{'run_to_plan_id'}) )     { $_error = 'run_to_plan_id not defined'; }
      elsif ( !defined($_args{'run_to_plan_name'}) )   { $_error = 'run_to_plan_name not defined'; }
      elsif ( !defined($_args{'href'}) )               { $_error = 'href not given'; }
      else {
         $Log->comment("::: get run_stats row\n");
         $_sql  = "SELECT * FROM $_run_stats_tab ";
         $_sql  .= "WHERE job_name = '$_args{'job_name'}' ";
         $_sql  .= "  AND run_to_plan_name = '$_args{'run_to_plan_name'}' ";
         $_sql  .= "  AND run_to_plan_id = '$_args{'run_to_plan_id'}'";
         $Log->debug("::: sql $_sql");
         $_ok = $self->{'dbh'}->selectrow_hashref($_sql);
         if ( defined($_ok) ) {
            $Log->debug("::: get successful [$_action]");
            %{$_args{'href'}} = %{$_ok};
            $_return = 1;
         }
         else {
            $Log->warn("::: get failed [$_action]");
            $_error = $self->{'dbh'}->errstr;
         }
      }
   }

   #
   # end run   TODO  make a one-stop method to closed out a row
   #
   elsif ($_action eq 'end run') {
      $Log->debug("::: action: [$_action]");
      $_return = 0;
      $_error  = 'action end run not implemented yet';
   }
   #
   # no action was given
   #
   else {
      $_return = 0;
      $_error  = 'no action given';
   }

   $Log->comment("::: JAMS action [$_action] done\n");
   $Log->comment("::: JAMS action error: $_error") if $_return == 0;

   return( wantarray ? ($_return, $_error) : $_return);
}


=head2 resolve_ids

Add hash to object to resolve id and names for given tables

    $JAMS->resolve_ids;


Hash syntax:

    $JAMS->{'id2name'}->{<table>}{'id'}{<id>} = name
    $JAMS->{'id2name'}->{<table>}{'name'}{<name>} = id

    <table> = line, mode, build_flavor, testset, q_status, branch, platform, specname
    <id> = ID values (int) from respective tables
    <name> = character string from the respective table for the given <id>

=cut

# Populate object with id-to-name translations
#
# $_[0] - object
#
# get base info to translate line, mode, build, testset index's
#
# $JAMS->{'id2name'}{<table>}{'id'}{<id>}      = name
#                            {'name'}{<name>}  = id
#
sub resolve_ids {

   my $_self = shift;

    my @_tables = ( 'line', 'mode', 'build_flavor', 'testset',
                    'branch', 'platform', 'q_status', 'specname'
    );

    my ($_sql, $_resp, $_aref, $_str);


    foreach $_str ( @_tables ) {
        $_sql = "SELECT * FROM $_str";
        $Log->debug("sql: $_sql");
        $_resp = $_self->{'dbh'}->selectall_arrayref($_sql);
        foreach $_aref ( @{$_resp} ) {
           $Log->debug("id2name:  table: $_str id: $_aref->[0]  name: $_aref->[1]");
           # hash->{<table>}{<id>} = name
           $_self->{'id2name'}{$_str}{'id'}{$_aref->[0]} = $_aref->[1];
           # hash->{<table>}{<name>} = id
           $_self->{'id2name'}{$_str}{'name'}{$_aref->[1]} = $_aref->[0];
        }
    }
   1;
}


=head2 resolve_run_spec

Add hash to object with run_spec table and resolved names

    $JAMS->resolve_run_spec;

Hash made is index by run_spec table index value,

    $JAMS->{'run_spec'}{<id>}{'line_id'}         = Line ID
                             {'Line'}            = Line name
                             {'mode_id'}         = Mode ID
                             {'Mode'}            = Mode name
                             {'build_flavor_id'} = Build flavor ID
                             {'Build'}           = Build flavor name
                             {'testset_id'}      = Testset ID
                             {'Testset'}         = Testset name
                             {'q_status_id'}     = Q_Status ID
                             {'Q_Status'}        = Q_Status name
                             {'specname_id'}     = Specname ID
                             {'Specname'}        = Specname name
                             {'platform_id'}     = Platform ID
                             {'Platform'}        = Platform name
                             {'branch_id'}       = Branch ID
                             {'Branch'}          = Branch name

=cut

# get run_spec table and name strings
#
#

sub resolve_run_spec {

    my $_self = shift;
    my $_sql  =
'SELECT run_spec.id,
       specname.name AS Specname, run_spec.specname_id,
       platform.name AS Platform, run_spec.platform_id,
       branch.name AS Branch, run_spec.branch_id,
       line.name AS Line, run_spec.line_id,
       mode.name AS Mode, run_spec.mode_id,
       build_flavor.name AS Build, run_spec.build_flavor_id,
       testset.name AS Testset, run_spec.testset_id,
       q_status.name AS Q_Status, run_spec.q_status_id
  FROM run_spec, specname, platform, branch, line, mode, build_flavor, testset, q_status
 WHERE run_spec.specname_id = specname.id
   AND run_spec.platform_id = platform.id
   AND run_spec.branch_id = branch.id
   AND run_spec.line_id = line.id
   AND run_spec.mode_id = mode.id
   AND run_spec.build_flavor_id = build_flavor.id
   AND run_spec.testset_id = testset.id
   AND run_spec.q_status_id = q_status.id';

    $Log->debug("sql: $_sql");
    my $_resp = $_self->{'dbh'}->selectall_hashref($_sql, ['id']);

    %{$_self->{'run_spec'}} = %{$_resp};

    1;

}

=head2 resolve_run_spec_name

For the run_spec information contained in $JAMS->{'runspec'}
cycle through the keys and generate a run_spec name and set it
in the respective hash

To get this hash in the object:

    $JAMS->resolve_run_spec_name;

Must of previously ran the resolve_run_spec mthod on this object,
this method will cycle through  the resolve_run_spec generated
hash, build a runspec name and add it to the hash.

Hash syntax:

    $JAMS->{'run_spec'}{<id>}{'run_spec_name'} = run_spec_name

=cut

# update object's run_spec info to contain the run_spec name
#
#
sub resolve_run_spec_name {

    my $_self = shift;
    my($_id);

    foreach $_id ( keys %{$_self->{'run_spec'}} ) {
       $_self->{'run_spec'}{$_id}{'run_spec_name'} = make_run_spec_name(
          $_self->{'run_spec'}{$_id}{'Line'},
          $_self->{'run_spec'}{$_id}{'Mode'},
          $_self->{'run_spec'}{$_id}{'Build'},
          $_self->{'run_spec'}{$_id}{'Testset'},
          $_self->{'run_spec'}{$_id}{'Q_Status'},
          $_self->{'run_spec'}{$_id}{'Specname'},
          $_self->{'run_spec'}{$_id}{'Platform'},
          $_self->{'run_spec'}{$_id}{'Branch'},
       );
    }
    1;
}



# TODO make exportable
#
# make a run_spec name
# Args are the char string for the following
#
# line, mode, build, suite, qstatus, spec, platform, branch
#
sub make_run_spec_name {

   sprintf("%s_%s_%s_%s_%s_%s_%s_%s",
      $_[0] || 'line',
      $_[1] || 'mode',
      $_[2] || 'build',
      $_[3] || 'suite',
      $_[4] || 'qstatus',
      $_[5] || 'spec',
      $_[6] || 'platform',
      $_[7] || 'branch',
   );
}



#
# Make run-to-paln-id from runid
# $_[0] run id
# return run-to-plan-id
#
sub make_run_to_plan_id {

   my $_rtpid = shift;
   $_rtpid =~ s/_/:/g;              # change underscore to colon
   $_rtpid =~ s/\.(\d+)$/:$1/g;  # change the trailing .<epoch> to :<epoch>

   $_rtpid;

}




1;

=head1 AUTHOR

Michael Stevens (stevensm@netapp.com)
Scott Parsons (parsons@netapp.com)

Mailto dl-coati-core

$Id$
Last Revised $Date$

=head1 STATUS

Shared

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=cut


