# -*- perl -*-
# $Id$

## @summary JAMS::Vsim
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

JAMS::Vsim - Class which interacts with vsims

=head1 SYNOPSIS

  use JAMS::Vsim;

=head1 DESCRIPTION

This module defines routines related to vsim creation, deletion and manipulation.

Note: 

No attempt is made here to document all vsim functionality.  For that please
see the following:

http://thebrewery-web.corp.netapp.com/Brewery/brewery/dev/Tools/VSim/index.html

=cut

package JAMS::Vsim;

use strict;
use Tharn qw($Log);
use Data::Dumper;
use Hostrec qw();
use IPC::Cmd qw(can_run);
use NACL::C::Options;
use NACL::C::SystemNodeDate;
use NACL::C::SystemServicesNtpConfig;
use NACL::STask::Node;
use NACL::Transit;
use NATE::BaseException qw(:try);

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => [
        {   '-default_ctor' => sub { $_[0]->_get_node_obj() },
            '-type' => 'NACL::STask::Node',
        },
    'naclnode'
    ],
    scalar => [ 'cmode' ],
    scalar => [ 'node' ],
    scalar => [ 'saved_env_HOME' ],
    scalar => [ 'saved_env_USER' ],
    scalar => [ 'uhome' ],
    scalar => [ 'user' ],
    scalar => [ 'id' ],
    scalar => [ 
        {   '-default_ctor' => sub { $_[0]->_create_transit() },
            '-type' => 'NACL::Transit',
        },
    'tran'
    ],
];

our $exec_cmd    = 'tools/jams/vsim/utility/exec_cmd.thpl';
our $unstick_cmd = 'tools/jams/vsim/utility/unstick_cmd.thpl';
our $vsim_cmd    = can_run('vsim');
our @nodes;

our $domainname_cmd = can_run('domainname');
our $domain = `$domainname_cmd`;
chomp($domain);

our $hostname_cmd = can_run('hostname');
our $hostname = `$hostname_cmd`;
chomp($hostname);

our $simdomain = "sim.netapp.com";

my $name_to_check = $domain;
if ( $hostname =~ /\.com$/ ) {
    $name_to_check = $hostname;
}

if ( $name_to_check =~ /rtp/ ) {
    $simdomain = "sim.rtp.netapp.com";
} elsif ( $name_to_check =~ /spin/ ) {
    $simdomain = "data.spin.eng.netapp.com";
} elsif ( $name_to_check =~ /btc/ ) {
    $simdomain = "sim.eng.btc.netapp.in";
} elsif ( $name_to_check =~ /nane/ ) {
    $simdomain = "sim.nane.netapp.com";
}

our $serial   = 1;

# -----------------------------------------------
# Function: get_seq
#   convenience routine to provide a serial (sequence) number
#   for subtest names
#
# Usage:
#
#    my $seq = $vsim->get_seq()
#
# Parameters:
#   
#    None
#
# Return:
#   
#    serial string in the range 0001 to 9999 incrementing each time
#    it is called.
#
# Assumptions:
# -----------------------------------------------
sub get_seq {
    my $self = shift;
    my $seq = sprintf( "%04d", $serial ); $serial++;
    return $seq;
}

# -----------------------------------------------
# Function: _get_node_obj
#   return a NACL::STask::Node object
#
# Usage:
#
#    my $node = $vsim->_get_node_obj()
#
# Parameters:
#   
#    None
#
# Return:
#   NACL::STask::Node object
#
# Assumptions:
# -----------------------------------------------

sub _get_node_obj {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    foreach my $node (@nodes) {
        if ( $node->node() eq $self->node() ) {
            $self->naclnode( $node );
            last;
        }
    }
    $self->{naclnode};
}

# -----------------------------------------------
# Function: _create_transit
#   return a NACL::Transit object
#
# Usage:
#
#    my $trans = $vsim->_create_transit()
#
# Parameters:
#
#    None
#
# Return:
#   NACL::Transit object
#
# Assumptions:
# -----------------------------------------------

sub _create_transit {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    return NACL::Transit->new( name => $self->node() );
}

# -----------------------------------------------
# Function: init
#   
#  routine used by the constructor
#
#  Note: Since there is not an explicit constructor
#        defined in this code (we're using init() and
#        MethodMaker's default constructor) I will 
#        provide the usage for the contructor here.
#
# Usage:
#   my $vsim = new( node => "ptcqa-vsim2" )
#
# Parameters:
#       node      - Name of the vsim
#
# Return: blessed reference
# -----------------------------------------------
sub init {
    my $self = shift;
    my $me=(caller(0))[3];

    my %opt = ( node => undef,
                @_ );

    my $node = $opt{node};

    my ($user, $ident);
    
    # in 7mode, or even Cmode, people can pass in 
    # hostrec overrides, like "radish-vsim7;hosttype=filer-vsim"
    #   or                    "radish-vsim7;MGMT_PORT=e0f"
    #   or                    "radish-vsim7{MGMT_PORT=e0f}"
    if ( $node =~ /(\S+);|{/ ) {
        $node = $1;
        $self->cmode(0);
    } else {
        $self->cmode(1);
    }
    $self->node( $node  );

    my $hostrec;

    # Not sure why hosttype isn't being overridden
    # from NODES parameter
    if ( $self->cmode() ) {
        $hostrec = Hostrec->new( id => $node );
    } else {
        $hostrec = Hostrec->new( id => $node, hosttype => 'filer-vsim' );
        if ( $hostrec->{hosttype} ne 'filer-vsim' ) {
            $Log->warn("$me: For some reason 7mode filer has hosttype of ".$hostrec->{hosttype}." instead of filer-vsim");
            $hostrec->{hosttype} = 'filer-vsim';
        }
    }

    if ( !@nodes ) {
        @nodes       = NACL::STask::Node->find();
    }

    if ( $node =~ /(\S+)-vsim(\d+)/ ) {
        $user  = $1;
        $ident = $2;
    } else {
        die "$me: Node name must be in the form <userid>-vsim<digit>";
    }

    $self->id(   $ident );
    $self->user( $user  );

    my $uid   = getpwnam($user);
    my $uhome = (getpwuid($uid))[7];

    $self->uhome( $uhome );

    $self->saved_env_HOME( $ENV{HOME} );
    $self->saved_env_USER( $ENV{USER} );

    $self;
}


# -----------------------------------------------
# Function: spoof_user_env
#   
#  Set ENV{HOME} and ENV{USER} to look like
#  we are run as the owner of vsim.
#
#  The vsim command actually looks at ENV for HOME and 
#  USER, even if you are subtesting and specifying the 
#  user as which to run the subtest correctly.
#
# Usage:
#   $vsim->spoof_user_env()
#
# Parameters:
#       Node
#
# Return: nothing
# -----------------------------------------------
sub spoof_user_env {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    $ENV{HOME} = $self->uhome();
    $ENV{USER} = $self->user();
}

# -----------------------------------------------
# Function: restore_user_env
#   
#  Restore ENV{HOME} and ENV{USER} settings
#  as they were when we first created the vsim.
#
#  The vsim command actually looks at ENV for HOME and 
#  USER, even if you are subtesting and specifying the 
#  user as which to run the subtest correctly.
#
# Usage:
#   $vsim->restore_user_env()
#
# Parameters:
#       Node
#
# Return: nothing
# -----------------------------------------------
sub restore_user_env {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    $ENV{HOME} = $self->saved_env_HOME();
    $ENV{USER} = $self->saved_env_USER();
}

# -----------------------------------------------
# Function: make
#   
#  Make a vsim.  All "vsim make" parameters can
#  be passed via the create_args option, but only
#  a small subset are passable via the other named
#  options.  The create_args option is not to be used
#  with any of the other options.
#
# Usage:
#   $vsim->make( num_nics => 8 );
#
# Parameters:
#       
#   build_root  => Root of the Ontap build tree
#   ibm         => Build with an IBM build? [0 | 1] (default 0)
#   mode        => Ontap Mode [10m | 7m]
#   platform    => [x86_64 | x86_64.gcov | x86_64.debug | x86_64.debug.gcov]
#   num_nics    => Number of ports for vsim, if not specified num_nics=4, [5 | 6 | 7 | 8]
#   timeout     => Number of seconds until command times out (default 600)
#   create_args => If you want full control over the arguments passed to make
#                  you must include all arguments except the vsim index.
#
# Return: A subtest reference for the backgrounded make.
# -----------------------------------------------
sub make {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( build_root  => undef,
                ibm         => 0,
                mode        => "10m",
                platform    => "x86_64",
                num_nics    => undef,
                create_args => undef,
                timeout     => 900,
                @_ );

    if ( !$opt{create_args} ) {
        my ($build,$mode,$flavor);
        if ( !$opt{build_root} ) {
            die "$me If \"create_args\" is not specified, \"build_root\" must be";
        } else {
            $build = $opt{build_root};
        }
        $opt{create_args} = "-pcons -quickl -".$opt{mode}." -p ".$opt{platform}." -ir ".$build;
        if ( $opt{ibm} ) {
            $opt{create_args} .= " -ibm";
        }
        if ( $opt{num_nics} ) {
            $opt{create_args} .= " -nics ".$opt{num_nics};
        }
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd make ".$opt{create_args}." ".$self->id(),
                    "EXPECTED=simulator created",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_make_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: state
#   
#  Invoke the "state" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   my $st = $vsim->state();
#   Subtest::wait_finish( subtests => [$st] );
#   @results = $st->results();
#   my $misc = $results[0]->misc();
#   my $out  = $misc->{VSIM_CMD_OUTPUT};
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 300)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub state {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 300,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $expected = "STATE";

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd state ".$self->id(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-timeout', $opt{timeout},
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_state_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: start
#   
#  Invoke the "start" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->start() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 300)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub start {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 300,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $expected = "Power on";

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd start ".$self->id(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-timeout', $opt{timeout},
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_start_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: stop
#   
#  Invoke the "stop" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->stop() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 300)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub stop {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 300,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $expected = "stopped";

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd stop ".$self->id(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-bg',
                    '-timeout', $opt{timeout},
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_stop_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rm
#   
#  Invoke the "rm" command on a vsim to remove it.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->rm() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   force       => Force removal even if cores are present (default 1)
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rm {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( force   => 1,
                timeout => 600,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();

    my $force = $opt{force} ? " -f" : "";

    my $expected = "simulator removed";
    if ( $force ) {
        $expected .= "\|does not exist";
    }

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd rm ".$force." ".$self->id(),
                    "EXPECTED=$expected",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_rm_".$self->node() );

    $self->restore_user_env();

    return $st;
}


# -----------------------------------------------
# Function: snapdelete
#   
#  Invoke the "snapdelete" command on a vsim to delete
#  a vsim snapshot.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->snapdelete( snapname => "post_deux_init" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   snapname    => Name of snapshot (default JAMS_SNAPSHOT)
#   timeout     => Number of seconds until command times out (default 1800)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub snapdelete {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( snapname => "JAMS_SNAPSHOT",
                timeout => 1800,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $snapshot = $opt{snapname};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd snapdelete ".$self->id()." $snapshot",
                    "EXPECTED=successful",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_snapdelete_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: snap
#   
#  Invoke the "snap" command on a vsim to create
#  a vsim snapshot.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->snap( snapname => "post_deux_init" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   snapname    => Name of snapshot (default JAMS_SNAPSHOT)
#   timeout     => Number of seconds until command times out (default 1800)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub snap {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( snapname => "JAMS_SNAPSHOT",
                timeout => 1800,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $snapshot = $opt{snapname};

    $self->spoof_user_env();

    # success message can be "successfully taken" or "successfully created"
    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd snap ".$self->id()." $snapshot",
                    "EXPECTED=successful",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_snap_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: snapback
#   
#  Invoke the "snapback" command on a vsim to restore
#  a previously saved vsim snapshot.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->snapback( snapname => "post_deux_init" ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   snapname    => Name of snapshot (default JAMS_SNAPSHOT)
#   timeout     => Number of seconds until command times out (default 1800)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub snapback {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( snapname => "JAMS_SNAPSHOT",
                timeout => 1800,
                @_ );

    my $node     = $self->node();
    my $id       = $self->id();
    my $snapshot = $opt{snapname};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd snapback ".$self->id()." $snapshot",
                    "EXPECTED=successful",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_snapback_".$self->node() );

    $self->restore_user_env();
    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: reserve
#   
#  Invoke the "reserve" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->reserve() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   simdomain   => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub reserve {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $simdomain,
                timeout => 600,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $simdomain = $opt{domain};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd reserve -restype vsimesx -domain $simdomain ".$self->id(),
                    "EXPECTED=successful",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_reserve_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: rehost
#   
#  Invoke the "rehost" command on a vsim.
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->rehost() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   simdomain   => An attempt is made to guess the correct simdomain
#                  based on the "domainname" of the client running
#                  this command.  Will be something like:
#                  "sim.rtp.netapp.com"
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub rehost {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( domain => $simdomain,
                timeout => 600,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $simdomain = $opt{domain};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().
                                  ";$vsim_cmd rehost -restype vsimesx -domain $simdomain ".$self->id(),
                    "EXPECTED=successful",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_rehost_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: extend or renew
#   
#  Invoke the "renew" command on a vsim.
#  Note: the extend() and renew() methods are
#  synonyms for one another.
#
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->extend( days => 3 ) );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#        
#   days        => 1-7 (default 7)
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub extend {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }
    return $self->renew( @_ );
}

sub renew {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( days => 7,
                timeout => 600,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();
    my $days      = $opt{days};

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd renew -res $days ".$self->id(),
                    "EXPECTED=Successful|End time must be less",
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_extend_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: halt
#   
#  Use NACL::Transit to transition a vsim to the LOADER
#  (FIRMWARE) prompt.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->halt() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#   maxidle     => Maximum number of seconds that no response from the filer
#                  will be tolerated (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub halt {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                maxidle => 600,
                @_ );

    # force $self->{tran} to be created via the default_ctor method
    my $tran      = $self->tran();
    my $node      = $self->node();
    my $st = Tharn::subtest( sub { $tran->change_state( to => 'FIRMWARE', timeout => $opt{timeout}, maxidle => $opt{maxidle} ); },
                 "-runid", get_seq()."_halt_$node",
                 '-timeout', $opt{timeout},
                 "-bg");

    return $st;
}


# -----------------------------------------------
# Function: boot
#   
#  Use NACL::Transit to transition a vsim to the CLI
#  prompt.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->boot() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#   maxidle     => Maximum number of seconds that no response from the filer
#                  will be tolerated (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub boot {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 900,
                maxidle => 600,
                @_ );

    # force $self->{tran} to be created via the default_ctor method
    my $tran      = $self->tran();
    my $node      = $self->node();
    my $naclnode  = $self->naclnode();

    my $st = Tharn::subtest( sub { $tran->change_state( to => 'CLI', timeout => $opt{timeout}, maxidle => $opt{maxidle} ); },
                 "-runid", get_seq()."_boot_$node",
                 '-timeout', $opt{timeout},
                 "-bg");

    return $st;
}

# -----------------------------------------------
# Function: stop_start_halt
#   
#  Stop then start a vsim, leaving it at the LOADER
#  prompt.  Very similar to gx_boot's shutdown_cold
#  task.
#
# Usage:
#
#   foreach $vsim (@vsims) {
#       push( @subtests, $vsim->stop_start_halt() );
#   }
#   Subtest::wait_finish( subtests => @subtest] ); @subtests = ();
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 900)
#   maxidle     => Maximum number of seconds that no response from the filer
#                  will be tolerated (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub stop_start_halt {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my @subtests;
    my %opt = ( maxidle => 600,
                timeout => 900,
                @_ );

    my $node      = $self->node();
    my $id        = $self->id();

    $self->spoof_user_env();

    my $st = Tharn::subtest( $exec_cmd,
                    "COMMAND=HOME=".$self->uhome().";USER=".$self->user().";$vsim_cmd stop ".$self->id().
                             ";$vsim_cmd start ".$self->id(),
                    'EXPECTED=Power on of virtual machine.*done',
                    'RETRY_IF=please rehost the vsim',
                    "BEFORE_RETRY=HOME=".$self->uhome().";USER=".$self->user().
                                       ";$vsim_cmd rehost -restype vsimesx -domain $simdomain ".$self->id(),
                    'NUM_RETRIES=3',
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_off_on_".$self->node() );

    push( @subtests, $st );

    #  need to stop the boot cycle ASAP, so need to halt
    Subtest::wait_finish( subtests => \@subtests );

    $st = $self->halt( timeout => $opt{timeout}, maxidle => $opt{maxidle} );

    push( @subtests, $st );

    $self->restore_user_env();

    if ( wantarray() ) {
        return @subtests;
    } else {
        return $st;
    }
}

# -----------------------------------------------
# Function: unstick
#   
#  Unstick, a "stuck" vsim.
#  Occasionally a vsim will become "stuck" when multiple
#  vsims attempt to write to the same pipe.  For a running
#  vsim this can be serious and should be looked into, however
#  a stuck vsim will even resist efforts to remove it.  This
#  routine will unstick a stuck vsim.
#
#  The output of any vsim command can be accessed
#  via the misc() call of the subtest result.
#
# Usage:
#
#   my $st = $vsim->unstick();
#   Subtest::wait_finish( subtests => [$st] );
#   @results = $st->results();
#   my $misc = $results[0]->misc();
#   my $out  = $misc->{VSIM_CMD_OUTPUT};
#
# Note: This routine is called automatically when another
#       command finds that the vsim is stuck, so it GENERALLY
#       does not need to be called explicitly.
#
# Parameters: 
#       
#   timeout     => Number of seconds until command times out (default 600)
#
# Return: A subtest reference for the backgrounded command.
# -----------------------------------------------
sub unstick {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( timeout => 600,
                @_ );

    $self->spoof_user_env();

    my $st = Tharn::subtest( $unstick_cmd,
                    "VSIM_INDEX=".$self->id(),
                    "VSIM_NODE=".$self->node(),
                    '-timeout', $opt{timeout},
                    '-bg',
                    '-username', $self->user(),
                    '-runid', $self->get_seq()."_unstick_".$self->node() );

    $self->restore_user_env();

    return $st;
}

# -----------------------------------------------
# Function: set_ntp
#   
#  Enable or disable NTP (Network Time Protocol)
#  Used in the clock syncing process.
#
# Usage:
#
#   $vsim->set_ntp( enable => 0 );
#
# Parameters: 
#       
#   enable     => [0|1] (default 1)
#
# Return: Nothing
# -----------------------------------------------
sub set_ntp {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( enable => 1, 
                @_ );

    my $flag    = "true";
    my $enable  = "Enabling";
    my $flag7   = "on";

    if ( !$opt{enable} ) {
        $flag   = "false";
        $flag7  = "off";
        $enable = "Disabling";
    }

    my $retries = 3;

RETRY:
 
    $Log->comment($enable." NTP service on all nodes");

    if ( $self->naclnode()->mode() eq '7Mode' ) {
        eval {
            NACL::C::Options->option(command_interface => $self->naclnode(),
                                    'option-name'      => "timed.enable" ,
                                    'option-value'     => $flag7 );
        };
    } else {
        eval {
            NACL::C::SystemServicesNtpConfig->modify( command_interface => $self->naclnode(), enabled => $flag );
        };
    }

    if ( $@ && ($retries > 0) ) {
        $Log->warn("$me: Failed with $@\nRetrying...");
        $retries--;
        Tharn::snooze(5);
        goto RETRY;
    } elsif ( $@ ) {
        die "$me: $@";
    }
}

# -----------------------------------------------
# Function: sync_clocks
#   
#  Synchronize the clocks of all cluster nodes for
#  a vsim with those of the client running this routine.
#  Useful after snapback of a previous vsim
#  snapshot.
#
# Usage:
#
#   $vsim->sync_clocks();
#
# Parameters: None
#       
# Return: Nothing
# -----------------------------------------------
sub sync_clocks {
    my $self  = shift;
    my $me=(caller(0))[3];
    if ( !ref($self) ) {
        die "$me: Can only call as instance method";
    }

    my %opt = ( @_ );

    my ($sec,$min,$hour,$day,$mon,$year,undef) = localtime();
    $year += 1900;
    $mon++;
    my $now = sprintf("\"%02d/%02d/%4d %02d:%02d:%02d\"", $mon, $day, $year, $hour, $min, $sec);

    sub _disable_ntp {
        $self->set_ntp( enable => 0 );
    }

    sub _enable_ntp {
        $self->set_ntp( enable => 1 );
    }

    my $ntp_must_be_disabled = 0;
    my $retries = 1;
    my @dates = NACL::C::SystemNodeDate->find( command_interface => $self->naclnode() );

AGAIN:

    if ( $ntp_must_be_disabled ) {
        _disable_ntp();
    }

    foreach my $date ( @dates ) {
        $Log->comment("Attempting to set date to $now on node ".$date->node());
        try {
            $date->modify( date => $now );
        } catch NATE::BaseException with {
            my $ex = shift;

            my $rethrow_exception = 1;
            if ( $ex->text() =~ /Cannot modify date on a node when ntp is enabled/i ) {
                if ( !$ntp_must_be_disabled ) {
                    $Log->comment("This build requires that NTP be disabled before the date can be modified, try again");
                    $ntp_must_be_disabled = 1;
                    $rethrow_exception = 0;
                }
            }

            if ( $rethrow_exception ) {
                $ex->throw();
            }
        };

        if ( $ntp_must_be_disabled && $retries ) {
            $retries = 0;
            goto AGAIN;
        }
    }

    if ( $ntp_must_be_disabled ) {
        _enable_ntp();
    }
}

1;
