#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/JobQueue.pm#1 $

#
## @summary     Manage jobs to be run on a list of clients
## @author      snider@netapp.com
## @pod here

=head1 NAME

JobQueue - Library to manage a queue of jobs to run on a list of clients

=head1 SYNOPSIS

    use JobQueue;
    my $Queue = new JobQueue(CLIENT=>\@clients, 
                             check_clients=>undef,
                             runid=>"myjob");
   $Queue->Loop(@list_of_jobs, \%nate_params);

=head1 DESCRIPTION

C<JobQueue> implements an object which is used to manage a list of jobs that
can run on any available client. This library is primarily used by 
L<TestManager> and has not been tested outside of it.

=head1 METHODS

=cut

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERLOADJOBQUEUE}) {
    my $newModule = $ENV{OVERLOADJOBQUEUE};
    delete $ENV{OVERLOADJOBQUEUE};
    print "The main JobQueue module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package JobQueue;
use Tharn qw(host);
use strict;
use POSIX ":sys_wait_h";
use TracePkg;
my $Trace = new TracePkg;

use Data::Dumper;
use RemoteExecution;
my $Remote;

my %maxProcs;
my %PIDS;
my $maxActiveJobs = 0;
my %clientLoads = ();
my @Clients;
my %clients = ();
my %nfsclients = ();
my %cifsclients = ();
my $numjobs;
my @jobs;
my %ActiveJobs;
my %Jobs;
my $errors = 0;
my $logfile = "/tmp/$0.log";
my ($Results_msg, $Success_msg);

=head2 new

    my $Queue = new JobQueue(CLIENT=>\@clients, 
                             check_clients=>undef,
                             runid=>"myjob");

C<new> is the invocant for the JobQueue object. This method will
create the object and setup the internal list of available clients.

=over

=item CLIENT

(Optional) This param is a reference to a list of available clients.

=item maxProcs

(Optional) This param indicates the maximum number of jobs to run on each
client at a time. The object only keeps track of the jobs that it submits
to the client itself. If another test is running with this client as well,
then this invocation is not aware of the other and will still allow this
number of jobs, just as the other test will. The param defaults to 10.

=item updateTime

(Optional) This param indicates how much time (in seconds) should elapse
in between status updates. See C<Loop>. The default is 10 seconds.

=item runid

(Optional) This param specifies the runid to pass into RemoteExecution
C<execute> routine. It will be used as part of the logfile names.

=item Output

(Optional) This param is obsolete.

=item check_clients

(Optional) This param indicates whether or not all of the clients should
be checked prior to queuing jobs to them. The default is to perform this 
check.

=back

If C<CLIENT>, C<nfsclient> or C<cifsclient> are not specified then no jobs
will be run.

=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
	CLIENT => [],       # a generic list of clients
	maxProcs  => 10,       # maximum processes per client
	updateTime => 10,      # how often to update the screen with the status
	runid => undef,
	Output => 1,
	check_clients => 1,
        nfsclient => [],
        cifsclient => [],
	@_,
    };
    bless ($self, $class);
    $self->{Output} = undef if (!$self->{Output});

    $Trace->entry();
    $Remote = new RemoteExecution;

    # parse the clients
    if (! scalar(@{$self->{CLIENT}})) {
	$Trace->fatal(msg=>"No clients were specified\n", result=>"CONFIG");
    }
    foreach my $client (@{$self->{CLIENT}}) {
       my $c = host($client);
       if ($c->hosttype() =~ /windows/) {
           push(@{$self->{cifsclient}}, $client);
       } elsif ($c->hosttype() =~ /unix/) {
           push(@{$self->{nfsclient}}, $client);
       } else {
           $Trace->fatal(msg=>"Unknown host type:".$c->hosttype);
       }
   }
   $Trace->debug(msg=>"nfsclient=".Dumper($self->{nfsclient}));
   $Trace->debug(msg=>"cifsclient=".Dumper($self->{cifsclient}));

    # for each client in the list, figure out if it has a specific maxProc set. If not, use the global default
    if (@{$self->{nfsclient}} != ()) {
	foreach my $host (@{$self->{nfsclient}}) {
	    foreach my $host (split(/,/, $host)) {
		my ($client, $maxProcs) = split(/:/, $host);
		$maxProcs{$client} = ($maxProcs) ? $maxProcs : $self->{maxProcs};
		$nfsclients{$client} = 1;
		push(@Clients, $client);
	    }
	}
    }
   $Trace->debug(msg=>"nfsclient=".Dumper(\%nfsclients));
    if (@{$self->{cifsclient}} != ()) {
	foreach my $host (@{$self->{cifsclient}}) {
	    foreach my $host (split /,/, $host) {
		my ($client, $maxProcs) = split(/:/, $host);
		$maxProcs{$client} = ($maxProcs) ? $maxProcs : $self->{maxProcs};
		$cifsclients{$client} = 1;
		push(@Clients, $client);
	    }
	}
    }
   $Trace->debug(msg=>"cifsclient=".Dumper(\%cifsclients));
    
    $Trace->exit();
    return $self;
}

=head2 Loop

   $Queue->Loop(@list_of_jobs, \%nate_params);

C<Loop> is the main method for the JobQueue object. The purpose of this 
method is to take the incoming list of jobs and run as many of them as 
possible at the same time on all of the available clients. It will spread
the jobs around all of the clients, trying to keep the load roughly the 
same. It is only keeping track of the number of jobs per client, rather 
than real load on the clients. 

The jobs are spread around by keeping track of how many jobs are executing
on each client. It will not add a new job to a client if that client
already has more jobs than other clients. Once all the clients are running
the maximim number of jobs (as specified in C<new>) it will no longer start
new jobs until others have finished.

After the C<maxUptime> amount of seconds has elapsed (as defined in C<new>)
the following type of message will be displayed:

  Jobs=500, Active Jobs=400, Errors=1

After all of the jobs have finished, a final result is returned to the
caller. The result is either a PASS result or a FAIL result with a
message similar to:

  1 / 500 job failures
$client returned FAIL

This method makes heavy use of routines defined in RemoteExecution.
It is primarily called by TestManager.
I do not believe it has a timeout mechanism internal to it.

=cut

sub Loop {
    my $self = shift;
    my $list = shift;
    my $nate_params = shift;
    my $timeout = shift || 1200;
    $Trace->entry();

    @jobs = @$list; # use the incoming list rather than copying it
    $maxActiveJobs = 0;
    %clientLoads = ();

    if (@jobs == ()) {
	$Trace->fatal(msg=>"Please provide a list of jobs.\n", result=>"SCRIPT");
	return undef;
    }

    my $clients = join(",", sort keys %clients);
    my @clients;
    
    # The caller has the ability to turn off the health-checking (maybe because it has already been performed)
    #  If the health-checking should occur now, set our local list to that returned from the health-checking 
    #  routine. Otherwise, set it to the master list which was provided by the caller
    if ($self->{check_clients}) {
	@clients = $Remote->check_clients(@Clients);
    } else {
	@clients = @Clients;
    }

    foreach my $host (@clients) {
	$maxActiveJobs += $maxProcs{$host};
	push(@{$clientLoads{'0'}}, $host);
	$clients{$host} = {};
    }
    $Trace->debug(msg=>"maxActiveJobs=$maxActiveJobs");
    if (! scalar(keys %clients) ) {
	$Trace->fatal(msg=>"None of the specified clients are running", result=>"CONFIG");
    }

    my $newclients = join(",", sort keys %clients);
    if ($newclients ne $clients) {
	$Trace->comment(msg=>"Now running with the following clients: $newclients\n");
    }
    # get a list of jobs that we are going to work with
    foreach my $job (@jobs) {
        if (ref($job) ne "HASH") {
            my $tjob = {
                        job => $job,
                        params => {},
                       };
            $job = $tjob;
        }
	if ($job->{job} =~ /nfs:\s*.+/ || $job->{job} =~ /cifs:\s*.+/ || $job->{job} =~ /any:\s*.+/) {
            $Trace->debug(msg=>"leaving job as $job->{job}");
	    # do nothing because this entry is OK
	} elsif ($job->{job} =~ /smb:\s*(.+)/) {
	    $job->{job} = "cifs:$1";
            $Trace->debug(msg=>"setting job to $job->{job}");
	} elsif ($job->{job} =~ /^(\S+):.+/) {
	    $Trace->fatal(msg=>"$1 is an unknown tag at for job '$job->{job}'\n", result=>"SCRIPT");
	} else {
	    $job->{job} = "any:$job->{job}";
            $Trace->debug(msg=>"setting job to $job->{job}");
	}
    }

    $Trace->debug(msg=>"Jobs=".Dumper(\@jobs));
    
    $Trace->comment(msg=>"Starting main loop\n");
    $numjobs = scalar(@jobs);
    my $orig_num_jobs = $numjobs;
    my $activeJobs = 0;

    # for each job fire off a client to process it trying to keep no more than the max per client
    my $start_time = localtime;
    my $counter = 0;
    while ($numjobs > 0) {
	if (localtime > $start_time + $timeout) {
	    $Trace->comment(msg=>"Jobs=$numjobs, Active Jobs=$activeJobs, Errors=$errors");
	    $Trace->fatal(msg=>"Timed out running the jobs after $timeout seconds");
	}
	if ($counter == $self->{updateTime}) {
	    $Trace->comment(msg=>"Jobs=$numjobs, Active Jobs=$activeJobs, Errors=$errors");
	    $counter = 0;
	    open(LOGFILE, ">$logfile") && do {
		foreach my $client (keys %ActiveJobs) {
		    foreach my $pid (keys %{$ActiveJobs{$client}}) {
			my %jobTable = %{$ActiveJobs{$client}};
			print LOGFILE "$client:$jobTable{$pid}\n";
		    }
		}
		close LOGFILE;
	    }
	}
	# process any child processes that are finished
	# activeJobs will be decremented by the number of jobs that have finished
	$activeJobs -= $self->ProcessChildren();
	
	if (($numjobs <= $activeJobs) || ($activeJobs >= $maxActiveJobs)) {
	    # we are at our max jobcount or there are no more jobs to obtain. Sleep and try again
	    $counter++;
	    sleep 1;
	} else {
	    # try to fire off a new job if jobs are available and we aren't at our max yet.
	    while (($numjobs > $activeJobs) && ($activeJobs < $maxActiveJobs)) {
		# find a client that has less that maxProcs running in it (could be smarter)
		my ($client,$protocol) = $self->GetClient();
		if ($client) {
		    my $job = $self->GetNextJob($protocol);
		    LAUNCH: {
			if ($job) {
			    $Trace->trace(msg=>"launch job on $client");
			    if ($self->LaunchProcess($client, $job, $activeJobs, $nate_params)) {
				$activeJobs++;
			    } else {
				# try another one client
                                delete $clients{$client};
                                if (! scalar keys %clients) {
                                    $Trace->fatal(msg=>"None of the clients are available any more");
                                }
				$Trace->warn(msg=>"$client failed so get a new client and try again");
				($client, $protocol) = $self->GetClient();
				redo LAUNCH;
			    }
			} else {
			    sleep 1;
			    $counter++;
			    last;
			}
		    }
		} else {
		    sleep 1;
		    $counter++;
		    last;
		}
	    }
	}
    }
    if ($errors) {
	$Trace->fatal(msg=>"$errors / $orig_num_jobs job failures\n" . $Results_msg);
    }

    $Trace->exit();
}

# This is an internal routine so does not have any POD

# return a random client that has less than maxProcs assigned to it
# this will actually search for a client that has the fewest number of jobs
sub GetClient {
    my $self = shift;
    my @availClients;
    my $client = undef;
    my $protocol = undef;

    $Trace->entry();
    foreach my $load (sort numerically keys %clientLoads) {
	if (@{$clientLoads{$load}} != ()) {	
	    @availClients = @{$clientLoads{$load}};
	    last;
	}
    }
    if (scalar @availClients) {
	my $randIndex = int(rand(scalar @availClients));
        $client = $availClients[$randIndex];
	if ($nfsclients{$client}) {
	    $protocol = "nfs";
	} elsif ($cifsclients{$client}) {
	    $protocol = "cifs";
	} else {
	    $Trace->fatal(msg=>"$client is neither CIFS nor NFS!\n", result=>"CONFIG");
	    $client = undef;
	}
    } else {
	sleep 1;
    }
    $Trace->exit(msg=>"return $client $protocol");
    return ($client, $protocol);
}

# This is an internal routine so does not have any POD

# given a client and a job, fire the job off and store some info on it
sub LaunchProcess {
    my $self = shift;
    my $client = shift;
    my $job = shift;
    my $active = shift;
    my $nate_params = shift;
    my $pid;
    
    my $runid = $self->{runid};

    while (my ($param, $value) = each (%{$job->{params}})) {
        $nate_params->{$param} = $value;
    }
    $job = $job->{job};

    if ($job =~ /(.+)xxQxx(.+)/) {
	($job, $runid) = ($1, $2);
    }
    $runid .= ".$active";
    $Trace->entry(msg=>"$client, $job, $active, $runid");
    
    return undef if (!$job);
    return undef if (!$client);

    eval {
	$pid = $Remote->execute(host=>$client, cmd=>$job, runid=>$runid, jobid=>$runid,
				nate_params=>$nate_params, output=>$self->{Output},
                                handler=>sub {});
    };
    if ($@ || !$pid) {
	$Trace->comment(msg=>"ERROR kicking off $job on client $client: $@\n");
	# remove this client from the available lists
	my $load = scalar(keys %{$clients{$client}});
	@{$clientLoads{$load}} = grep($_ ne $client, @{$clientLoads{$load}});
	$Trace->exit();
	return undef;
    } else {
	$Trace->trace(msg=> "kicking off $job ($runid) on client $client\n");
	my $load = scalar(keys %{$clients{$client}});
	@{$clientLoads{$load}} = grep($_ ne $client, @{$clientLoads{$load}});
	push(@{$clientLoads{++$load}}, $client);
	$clients{$client}{$pid} = 1;
	$PIDS{$pid} = $client;
	$Jobs{$pid} = $runid;
	$ActiveJobs{$client}{$pid} = $job;
	$Trace->exit();
	return 1;
    }
    $Trace->exit();
}

# This is an internal routine so does not have any POD

# look for any jobs that have finished
# returns the number of jobs that finished.
sub ProcessChildren {
    my $self = shift;
    my $rc;
    my $pids = 0;

    $Trace->entry();
    my @pids = $Remote->wait_pid(); # just process jobs that are finished
    $errors += $Remote->handle_results(\$Results_msg, \$Success_msg, @pids);

    foreach my $pid (@pids) {
	$rc = $Remote->exit_value(pid=>$pid);
	my $client = $PIDS{$pid};
	delete $PIDS{$pid};
	next if (!$client); # not sure how we get a bad one but we can
	if ($rc) {
	    $Trace->trace(msg=>"$pid for host $client ($Jobs{$pid}) returned with rc=$rc\n");
	    $Trace->comment(msg=>"$ActiveJobs{$client}{$pid}\n");
	} else {
	    $Trace->debug(msg=> "$pid for host $client returned with rc=$rc\n");
	}
	my $load = scalar(keys %{$clients{$client}});
	if ($load) {
	    @{$clientLoads{$load}} = grep($_ ne $client, @{$clientLoads{$load}});
	} else {
            print Dumper \%clients;
	    $Trace->fatal(msg=>"should not reach this point in ProcessChildren for client '$client'\n", result=>"SCRIPT");
	}
	push(@{$clientLoads{--$load}}, $client);
        delete $clients{$client}{$pid};
	delete $ActiveJobs{$client}{$pid};
	$numjobs--;
	$pids++;
	$Trace->debug(msg=>"removing '$pid' for host $client from the queues: numjobs=$numjobs\n");
    }
    $Trace->debug(msg=>"return $pids pids that have finished\n");
    $Trace->exit();
    return $pids;
}

# This is an internal routine so does not have any POD

# pick a job off of the joblist, based on the type of client that is free
sub GetNextJob {
    my $self = shift;
    my $protocol = shift;
    my $job;
    
    $Trace->entry();
    my $index = 0;
    return undef if (@jobs == ());
    foreach my $tjob (@jobs) {
	if ($tjob->{job} =~ /$protocol:\s*(.+)/ || $tjob->{job} =~ /any:\s*(.+)/) {
	    $tjob->{job} = $1;
            $job = $tjob;
	    last;
	}
	$index++;
    }
    if ($job) {
	splice(@jobs, $index, 1);
        $Trace->exit();
	return $job;
    }  
    $Trace->fatal(msg=>"could not find a matching job for $protocol\n", result=>"SCRIPT");
    return undef;
}

# This is an internal routine so does not have any POD

sub numerically { $a <=> $b }

=head1 PARAMS

none

=head1 SEE ALSO

L<TestManager>

L<RemoteExecution>

=head1 BUGS

This library is only being used by L<TestManager> and has not been tested
much without specifying SPINDEX=1 for L<RemoteExecution>

=head1 STATUS

Shared

=head1 ATTRIBUTES

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/JobQueue.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut
1;
