# this is basically -*- perl -*-
# $Id$
#
## @summary Kerberos Admin Class
## @author	dl-ssqa,gvg
## @status	public
## @pod here

=head1 NAME

Krb::Admin - Kerberos Admin Interface for Unix KDC and Krb Server/Clients

=head1 SYNOPSIS

  use Krb::Admin;

  $krbadmin = new Krb::Admin(kdc    => $host,
                             domain => $domain,
                             realm  => $DOMAIN,);

  $krbadmin->setup_kdc();

  $krbadmin->setup_filer(filer => $filername);

  $krbadmin->setup_host(host => $host,
                        user => $user);

=head1 TODO

  * Only Krb5 is supported.
  * KDC is tested and supported only on Solaris 10.
  * Only 7M mode is supported to setup filers.
  * Only Linux and Solaris are supported to setup krbhosts.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-ssqa,gvg
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2011, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Krb::Admin;

use strict;
use warnings;

use Tharn qw($Log);
use Krb::KDC;
use Krb::Host;
use GenericMount qw(mount umount);

# This wipes out CLIENTFILE but, 7mode uses it
#use nfs::GenericExport qw(export_path);
use Utils qw(make_conn);
use Carp qw(croak);
use filer qw(filer_options_get filer_options_set);
use Params::Validate
  qw(validate SCALAR BOOLEAN HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Options;
use Data::Dumper;
use NACL::C::VserverNfsKerberosConfig;
use NACL::C::VserverServicesKerberosRealm;
use NACL::C::Node;
use NACL::CS::Interface;
use NATE::BaseException qw(:try);
use NACL::C::UnixUser;
use NACL::CS::UnixUser;

my $Options = new Options;
$Options->parameters(
    definitions => [
        {name => "TOP_LOGDIR",
         description =>
           "ntest log directory where keytab files can be shared across machines",
        },
        {name        => "WEB_SERVER",
         default     => "10.60.132.98",
         description => "The webserver where one can access TOP_LOGDIR",},]);

my $sudo = "/usr/software/bin/sudo";

#---------------------------------------------------------------------

=head2 new

  $krbadmin = new Krb::Admin(kdc    => $host,
                             domain => $domain,
                             realm  => $realm,);

Krb::Admin object constructor. It connects to KDC server and initializes a few variables
such as krb5.conf, 

The following named arguments are accepted:

=over

=item Arguments

=over

=item kdc

(Required) : Hostname of KDC server

=item domain

(Required) : Kerberos domain name.

=item realm

(Required) : Kerberos realm name.

=back

=item Returns

Krb::Admin object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%;) {
    my $class = shift;
    my (%opts) =
      validate(@_,
               {kdc    => {type => SCALAR,},
                domain => {type => SCALAR, optional => 1},
                realm  => {type => SCALAR, optional => 1},});
    $Log->enter();

    my $self = bless {kdc => $opts{kdc},}, $class;

    # Create KDC Object and save it locally to use it later
    $self->{kdcobj} = new Krb::KDC(kdc => $opts{kdc});

    $Options->process();
    $self->{dir} =
        $Options->get(option => "TOP_LOGDIR")
      ? $Options->get(option => "TOP_LOGDIR")
      : "/tmp";
    $self->{web_dir} =
      'http://' . $Options->get(option => "WEB_SERVER") . $self->{dir};
    if ($self->{dir} =~ /^\/tmp$/) {
        $Log->warn(
            "The directory $self->{dir} is probably not web accessible, and therefore is not compatible with CMode Kerberos"
            );
    }

    # These are already on the KDC object, keep them here for ease.
    $self->{realm}  = $self->{kdcobj}->{realm};
    $self->{domain} = $self->{kdcobj}->{domain};

    $Log->exit(Dumper($self));
    return $self;
}

#---------------------------------------------------------------------

=head2 setup_kdc

  $krbadmin->setup_kdc();

Sets up KDC (Key Distribution Center) using the given values.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup_kdc {
    my $self = shift;

    croak "This function is not implemented yet";

    return;
}

#---------------------------------------------------------------------

=head2 setup_filer

  $krbadmin->setup_filer(filer => $filer,);

Sets up Filer for Kerberos/NFS.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup_filer {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {filer   => {type => SCALAR,},
                vserver => {type => SCALAR, optional => 1},
                service => {type => SCALAR, optional => 1, default => 'nfs'},
                keytab_uri => {type => BOOLEAN, default  => 1},
                kdc_user   => {type => SCALAR,  optional => 1, default => ""},
                kdc_pass   => {type => SCALAR,  optional => 1, default => ""},}
               );

    # If the vserver option was given, it is CMode
    if (my $vserver = $opts{vserver}) {
        my $node = NACL::C::Node->new(node => $opts{filer});

        # keytab uri is only applicable to CMode. In RR.0, I found that we
        # can either create a keytab on the kdc using kadmin.local and let
        # the filer download the filer via http OR we can omit the keytab
        # and allow the filer to create all its own service tickets and
        # principals on the KDC
        my $keytab_uri = $opts{keytab_uri};

        # only need to setup the kerberos realm on the filer one time. Tag
        # the object so we know this is done. multiple instantiations of this
        # object will cause this to run again.
        if (not $self->{filer_realm_configured}) {
            my $configName = 'kerberos-config';
            try {
                $self->{filer_realm_configured} = 1;
                my $obj =
                  NACL::C::VserverServicesKerberosRealm->create(
                       configname        => $configName,
                       command_interface => $node,
                       "kdc-vendor"      => $self->{kdcobj}->{vendor},
                       "kdc-ip" => $self->{kdcobj}->{host_object}->default_ip(),
                       realm    => $self->{realm},);
            }
            catch NATE::BaseException with {
                my $ex = shift;

                # ignore duplicate errors
                return if ($ex->text() =~ /Duplicate\sentry/i);
                die "EXCEPTION: " . $ex->text();
            };
        }

        # Create a user for the vserver that corresponds to the service name in the
        # kerberos ticket
        my @users =
          NACL::CS::UnixUser->fetch(command_interface => $node,
                                    allow_empty       => 1,
                                    filter => {vserver => $opts{vserver}});

        # Make a few reasonable tries to get a unique uid
        my $try       = 1;
        my $retry     = 3;
        my $userIdNum = undef;
        my $maxUserId = 65533;
        if (@users) {
            while ($try <= $retry) {
                $try++;
                my $tmp = int(rand($maxUserId));
                foreach my $hash (@users) {
                    if ($userIdNum == $hash->{id}) {
                        next;
                    }
                }
                $userIdNum = $tmp;
            }
        } else {
            $userIdNum = int(rand($maxUserId));
        }

        if (not defined($userIdNum)) {
            die config_result(
                "Unable to find a suitable user id number for the kerberos service $opts{service} account. "
                  . "I tried $try times to find a unique id.");
        }

        try {
            NACL::C::UnixUser->create(
                             user              => $opts{service},
                             command_interface => $node,
                             vserver           => $opts{vserver},
                             id                => $userIdNum,
                             "full-name" => $opts{service} . "_kerberosService",
                             "primary-gid" => "0",);
        }
        catch NATE::BaseException with {
            my $ex = shift;
            die "EXCEPTION: " . $ex->text();
        };

        # Get the lifs on this vserver
        # need an array of just the lif names
        my @lifs =
          NACL::CS::Interface->fetch(command_interface => $node,
                                     allow_empty       => 1,
                                     filter => {vserver => $opts{vserver}});
        if (not @lifs) {
            die config_result(
                "The vserver $opts{vserver} doesn't seem to have any lifs. That is bad."
                );
        }

        # Create Keytab file for each vserver LIF
        my $keytabfile;
        my $kdcTabFile;

        # On the KDC server, create a principle & a keytab for each vserver LIF
        if ($keytab_uri) {
            foreach my $lif (@lifs) {
                $lif->{dnsName} =
                  (gethostbyaddr(pack("C4", split(/\./, $lif->{address})), 2))
                  [0];

                $self->{kdcobj}->add_host(host    => $lif->{dnsName},
                                          service => "nfs");
                $Log->debug("Setting up keytab for hostname $lif->{dnsName}");
                $kdcTabFile =
                  $self->{kdcobj}->keytab(host    => $lif->{dnsName},
                                          service => $opts{service},);

                # copy each vserver LIF keytab to logs dir
                $keytabfile = $self->{dir} . "/" . $opts{vserver} . "_" .
                              $lif->{dnsName} . "_keytab." . $$;
                $self->{kdcobj}
                  ->keytab_copy(keytab => $kdcTabFile, to => $keytabfile);
            }
        }

        # Configure kerberos for each vserver lif
        foreach my $lif (@lifs) {
            $lif->{dnsName} =
              (gethostbyaddr(pack("C4", split(/\./, $lif->{address})), 2))[0]
              if not($lif->{dnsName});
            my %args;
            $args{command_interface} = $node;
            $args{vserver}           = $opts{vserver};
            $args{lif}               = $lif->{lif};
            $args{kerberos}          = "enable";
            $args{spn} =
              $opts{service} . '/' . $lif->{dnsName} . '@' . $self->{realm};
            if ($keytab_uri) {

                # the keytab was generated at the KDC and copied to logs dir, download it
                $args{"keytab-uri"} = $self->{web_dir} . '/' . $opts{vserver} . "_" .
                                      $lif->{dnsName} . "_keytab." . $$;
            } else {

                # Need the username and password to access the if the filer is going
                # to setup its own principals on the KDC. Not sure where to get them...
                $args{"admin-username"} = $opts{kdc_user};
                $args{"admin-password"} = $opts{kdc_pass};
                $Log->comment(  "Using user "
                              . $args{"admin-username"}
                              . " and pass "
                              . $args{"admin-password"}
                              . " to access the KDC kadmin tool.");
            }

            try {
                my $obj = NACL::C::VserverNfsKerberosConfig->modify(%args);

            }
            catch NATE::BaseException with {
                my $ex = shift;
                die "EXCEPTION: " . $ex->text();
            };
        }

    } else {

        # 7 Mode configuration starts here

        my $filer_options = {
                        "nfs.kerberos.enable"             => {required => "on"},
                        "nfs.kerberos.file_keytab.enable" => {required => "on"},
                        "nfs.kerberos.principal" =>
                          {required => $opts{filer} . "." . $self->{domain}},
                        "nfs.kerberos.realm" => {required => $self->{realm}},};

        # Create Keytab file for the Filer
        my $keytabfilename = $opts{filer} . "_keytab." . $$;
        my $keytabfile     = $self->{dir} . "/" . $keytabfilename;

        $self->{kdcobj}->add_host(host    => $opts{filer},
                                  service => "nfs");
        $self->{kdcobj}->keytab(host    => $opts{filer},
                                service => "nfs",
                                to      => $keytabfile);
        my $errmsg;
        my $krbconf = $self->{dir} . "/krb5.conf";
        my $volpath = "/vol/vol0";

        #
        # Get conf file from KDC
        #
        $self->{kdcobj}->copy_conf(to => $krbconf) if (!-f $krbconf);

        #
        # Copy conf file and keytab file to Filer
        #

        my $fc = make_conn($opts{filer});
        export_path($fc, $volpath, OPTIONS => "rw,anon=0");
        my $mntpnt = mount($opts{filer},
                           $volpath,
                           chmod   => "0755",
                           mntopts => "noac,proto=udp");

        my ($cmd, $out);
        $cmd = "$sudo cp $krbconf $mntpnt/etc/krb5.conf";
        $out = qx($cmd 2>&1);
        if ($?) {
            $errmsg = "Could not copy krb5.conf file into Filer/etc: $out";
            goto UMOUNT;
        }
        $Log->comment("cmd = $cmd, out = $out");

        $cmd = "$sudo mv $keytabfile $mntpnt/etc/UNIX_krb5.keytab";
        $out = qx($cmd 2>&1);
        if ($?) {
            $errmsg = "Could not copy keytab file into Filer/etc: $out";
            goto UMOUNT;
        }
        $Log->comment("cmd = $cmd, out = $out");

        #
        # Enable options on the filer to setup Kerberos
        #
        foreach my $option_key (keys %$filer_options) {
            $filer_options->{$option_key}{old_value} =
              filer::filer_options_get($fc, $option_key);
            filer::filer_options_set($fc, $option_key,
                                     $filer_options->{$option_key}{required});
        }

      UMOUNT:
        umount($mntpnt);

        croak $errmsg if ($errmsg);
    }

    return;
}

#---------------------------------------------------------------------

=head2 setup_host

  $krbadmin->setup_host(host => $host,
                        user => $user);

Sets up Kerberos Host.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup_host {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {host => {type => SCALAR,},
                user => {type => ARRAYREF, optional => 1},});
    $Log->enter();

    my $krbconf    = $self->{dir} . "/krb5.conf";
    my $keytabfile = $self->{dir} . "/" . $opts{host} . "_keytab." . $$;

    my $errmsg;

    #
    # Get conf file from KDC
    #
    $self->{kdcobj}->copy_conf(to => $krbconf) if (!-f $krbconf);

    #
    # Create Keytab file for the Host
    #
    $Log->debug("Creating Kerberos Principal and Keytab file for $opts{host}");
    my $fh = Tharn::host($opts{host});
    my $service = ($fh->hosttype =~ /solaris/i) ? "root" : "nfs";

    $self->{kdcobj}->add_host(host    => $opts{host},
                              service => $service);
    $self->{kdcobj}->keytab(host    => $opts{host},
                            service => $service,
                            to      => $keytabfile);

    #
    # Setup Kerberos Host
    #
    my $khost = Krb::Host->new(host   => $opts{host},
                               domain => $self->{domain},
                               realm  => $self->{realm});

    $khost->setup(krbconf => $krbconf,
                  keytab  => $keytabfile);

    foreach my $user (@{$opts{user}}) {
        $khost->krbinit(user => $user);
    }

    $Log->exit();
    return;
}

#---------------------------------------------------------------------

=head2 add_users

  $krbadmin->add_users(users => \@users);

Adds user principals to Kerberos.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub add_users {
    my $self = shift;
    my (%opts) = validate(@_, {users => {type => ARRAYREF,},});

    #
    # Create user principal
    #
    foreach my $user (@{$opts{users}}) {
        $self->{kdcobj}->add_user(user => $user);
    }

    return;
}

#---------------------------------------------------------------------

1;

