# this is basically -*- perl -*-
# $Id$
#
## @summary Kerberos Host Class
## @author	dl-ssqa,gvg
## @status	public
## @pod here

=head1 NAME

Krb::Host - Kerberos Host Interface

=head1 SYNOPSIS

  use Krb::Host;

  # Create Kerberos Host Object
  $khost = new Krb::Host(host    => $host,
                         domain => $domain,
                         realm  => $DOMAIN,);

  # Configure and Enable Kerberos Service on the Host
  $khost->setup();

  # Initialize Kerberos keys to start using service
  $khost->kinit();

=head1 TODO

  * Only Krb5 is supported.
  * KDC is tested and supported only on Solaris 10.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-ssqa,gvg
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Krb::Host;

use strict;
use warnings;

use Tharn qw($Log);
use Carp qw(croak);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Data::Dumper;

#---------------------------------------------------------------------

=head2 new

  $kdc = new Krb::Host(host   => $host,
                       domain => $domain,
                       realm  => $realm,);

Krb::Host object constructor based on host OS version.

The following named arguments are accepted:

=over

=item Arguments

=over

=item kdc

(Required) : Hostname of Kerberos Host

=item domain

(Required) : Kerberos domain name.

=item realm

(Required) : Kerberos realm name.

=back

=item Returns

Krb::KDC object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            host   => { type => SCALAR, },
            domain => { type => SCALAR, },
            realm  => { type => SCALAR, },
        }
    );
    $Log->enter();

    my $fh = Tharn::host($opts{host});
    $opts{conn} = Tharn::connect("#rlogin", '-timeout', 180);

    # Determine the right platform
    # and inherit correct OS dependent class.
    my $type = $fh->hosttype();
    if ( $type =~ /unix-solaris/i ) {
        $class = "Krb::Host::Solaris";
    } elsif ( $type =~ /unix-linux/i ) {
        $class = "Krb::Host::Linux";
    } else {
        croak "Krb::Host::new does not support hosttype: $type";
    }

    eval "require $class";
    my $self = $class->_new(%opts);

    $Log->exit(Dumper($self));
    return $self;
}

#---------------------------------------------------------------------

=head2 setup

    $khost->setup();

Sets up Kerberos service and enable it on the Host

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setup {
    my $self = shift;

    croak "This function is not implemented yet";

    return;
}

#---------------------------------------------------------------------

=head2 krbinit

    $khost->krbinit();

Runs kinit command to initialize kerberos keys to start using service.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub krbinit {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            user    => { type => SCALAR, },
            passwd  => { type => SCALAR, default => "netapp", },
        }
    );
    $Log->enter();

    # First destroy krb credentials, if exists already
    $self->krbdestroy(user => $opts{user});

    my $cmd = $self->{kinit}." -p ".$opts{user};

    my $pwprompt = "assword for ".$opts{user}."\@".$self->{realm}.":";
    my ($prompt, $out);
    my @outstr;
    (undef, undef, undef, $prompt) = 
        Tharn::say($self->{conn}, $cmd, '-prompt', $pwprompt);
    if ( $prompt =~ $pwprompt ) {
        @outstr = Tharn::say($self->{conn}, $opts{passwd});
    }
    $out = join("",@outstr);
    chomp($out);
    
    croak "$cmd failed: $out" if($out =~ /^\s+$/);

    $Log->exit();
}

#---------------------------------------------------------------------

=head2 krbdestroy

    $khost->krbdestroy();

Runs kdestroy command to destroy kerberos keys to cleanup or before init.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub krbdestroy {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            user    => { type => SCALAR, },
        }
    );
    $Log->enter();

    my $cmd = $self->{kdestroy};

    my $pwprompt = "assword for ".$opts{user}."\@".$self->{realm}.":";
    my (@outstr, $out);
    @outstr = Tharn::say($self->{conn}, $cmd);
    $out = join("",@outstr);
    chomp($out);

    croak "$cmd failed: $out" if(!$out && $out !~ /No credentials cache file found while destroying cache/i);

    $Log->exit();
}

#---------------------------------------------------------------------

1;
