# this is basically -*- perl -*-
# $Id$
#
## @summary Kerberos KDC Class
## @author	dl-ssqa,gvg,tronald
## @status	public
## @pod here

=head1 NAME

Krb::KDC - Kerberos KDC Interface for Unix KDC

=head1 SYNOPSIS

  use Krb::KDC;

  $kdc = new Krb::KDC(kdc    => $host,
                      domain => $domain,
                      realm  => $DOMAIN,);

  $kdc->copy_conf(to => $path);

  $kdc->add_principal(host => $host);

  $kdc->keytab(host => $host,
               to => $path);

=head1 TODO

  * Only Krb5 is supported.
  * KDC is tested and supported only on Solaris 10.
  * Added support for RHEL 6.3 (tronald) 2013/3/14

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-ssqa,gvg,tronald
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Krb::KDC;

use strict;
use warnings;

use Tharn qw($Log);
use Carp qw(croak);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);

use Krb::KDC::Solaris;
use Krb::KDC::Linux;

#---------------------------------------------------------------------

=head2 new

  $kdc = new Krb::KDC(kdc    => $host,
                      domain => $domain,
                      realm  => $realm,);

Krb::KDC object constructor based on host OS version.

The following named arguments are accepted:

=over

=item Arguments

=over

=item kdc

(Required) : Hostname of KDC server

=item domain

(Required) : Kerberos domain name.

=item realm

(Required) : Kerberos realm name.

=back

=item Returns

Krb::KDC object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%;) {
    my $class = shift;
    my (%opts) =
      validate(@_,
               {kdc    => {type => SCALAR,},
                domain => {type => SCALAR, optional => 1},
                realm  => {type => SCALAR, optional => 1},});
    $Log->enter();

    my $fh = Tharn::host($opts{kdc});
    $opts{conn} = Tharn::connect("#rlogin", '-timeout', 180);

    # Determine the right platform
    # and inherit correct OS dependent class.
    my $type = $fh->hosttype();
    if ($type =~ /unix-solaris/i) {
        $class = "Krb::KDC::Solaris";
    } elsif ($type =~ /unix-linux/i) {
        $class = "Krb::KDC::Linux";
    } else {
        croak "Krb::KDC::new does not support hosttype: $type";
    }

    eval "require $class";
    my $self = $class->_new(%opts);

    # save the host obj
    $self->{host_object} = $fh;

    # Use domain option if given
    if ($opts{domain}) {
        $self->{domain} = $opts{domain};

    } elsif ($self->{host_object}->{KRB_DOMAIN}) {

        # Get from the tharnhost object
        $self->{domain} = $self->{host_object}->{KRB_DOMAIN};

    } else {
        croak
          "Unable to determine the appropriate Kerberos domain from options or tharnhost";
    }

    # Use realm option if given
    if ($opts{realm}) {
        $self->{realm} = $opts{realm};

    } elsif ($self->{host_object}->{KRB_REALM}) {

        # Get from the tharnhost object
        $self->{realm} = $self->{host_object}->{KRB_REALM};

    } else {
        croak
          "Unable to determine the appropriate Kerberos realm from options or tharnhost";
    }

    # this is the Kerberos server vendor name.
    if ($type =~ /unix-solaris|unix-linux/) {
        $self->{vendor} = 'Other';
    } else {
        $self->{vendor} = 'Microsoft';
    }

    $Log->exit();
    return $self;
}

#---------------------------------------------------------------------

=head2 setup

    $kdc->setup();

Initial setup of KDC 

=for html <hr />

=cut

#---------------------------------------------------------------------

=head2 copy_conf

  $kdc->copy_cond(to => $path);

Copies kerberos config file to the location specified.

=for html <hr />

=cut

#---------------------------------------------------------------------

=head2 add_host

  $dc->add_host(host => $host,);

Adds principal for the given hostname.

=for html <hr />

=cut

#---------------------------------------------------------------------

=head2 delete_host

  $dc->delete_host(host => $host,);

Delete principal for the given hostname.

=for html <hr />

=cut

#---------------------------------------------------------------------

=head2 keytab

  $kdc->keytab(host => $host,
               to => $path);

Creates Keytab file for the given hostname and copies to the specified location.

=for html <hr />

=cut

#---------------------------------------------------------------------

=head2 add_user

  $kdc->add_user(user   => $user,
                 passwd => $passwd);

Adds user into Kerberos DB.

=for html <hr />

=cut

#---------------------------------------------------------------------

=head2 delete_user

  $kdc->delete_user(user   => $user,);

Deletes user from Kerberos DB.

=for html <hr />

=cut

#---------------------------------------------------------------------

1;

