# this is basically -*- perl -*-
# $Id$
#
## @summary Kerberos Admin Class
## @author	dl-ssqa,gvg
## @status	public
## @pod here

=head1 NAME

Krb::KDC::Solaris - Kerberos KDC Interface for Solaris KDC

=head1 SYNOPSIS

  use Krb::KDC;

  my $kdc = new Krb::KDC(kdc    => $solariskdc,
                         domain => $domain,
                         realm  => $realm);

Refer Krb::KDC

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-ssqa,gvg
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Krb::KDC::Solaris;
use base qw(Krb::KDC);

use strict;
use warnings;

use Tharn qw($Log);
use Carp qw(croak);
use SayWrap qw(sayEOF);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);

#---------------------------------------------------------------------
sub _new {
    my $class = shift;
    $Log->enter();

    my $self = bless {@_,
                      sudo    => "/usr/software/bin/sudo",
                      retry   => 5,
                      kadmin  => "/usr/sbin/kadmin.local",
                      krbconf => "/etc/krb5/krb5.conf",
                      kdcconf => "/etc/krb5/kdc.conf",}, $class;

    $Log->exit();
    return $self;
}

#---------------------------------------------------------------------
sub setup {
    my $self = shift;
    $Log->enter();

    croak "This function is not implemented yet";

    $Log->exit();
    return;
}

#---------------------------------------------------------------------
sub copy_conf {
    my $self = shift;
    my (%opts) = validate(@_, {to => {type => SCALAR,},});

    my $cmd =
        $self->{sudo} . " cp "
      . $self->{krbconf} . " "
      . $opts{to} . ';'
      . $self->{sudo}
      . " chmod 777 "
      . $opts{to};
    my $out = sayEOF($self->{conn}, $cmd);

    croak "$cmd failed: $out" if ($out =~ /fail|error/i);

    $Log->exit();
    return;
}

#---------------------------------------------------------------------
sub keytab {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {host    => {type => SCALAR,},
                service => {type => SCALAR, default => "root",},
                to      => {type => SCALAR, optional => 1},});
    $Log->enter();

    my $tmpkeytab = "/tmp/" . $opts{host} . "_keytab." . $$;

    my $qualifiedHostName = $opts{host};
    if ($qualifiedHostName !~ /$self->{domain}/i) {
        $qualifiedHostName .= '.' . $self->{domain};
    }

    #
    # Create keytab file
    #

    my $cmd =
        $self->{sudo} . " "
      . $self->{kadmin}
      . " -q 'ktadd -k $tmpkeytab -e \"des-cbc-crc:normal des-cbc-md5:normal\" "
      . $opts{service} . "/"
      . $qualifiedHostName . "'";
    my $out = sayEOF($self->{conn}, $cmd);

    my $expmsg =
        "Entry for principal "
      . $opts{service} . "/"
      . $qualifiedHostName
      . " with kvno 3, encryption type DES cbc mode with CRC-32 added to keytab WRFILE:${tmpkeytab}.";
    croak "$cmd failed: --$out-- did not match for --$expmsg--"
      if ($out !~ /$expmsg/);

    if ($opts{to}) {
        $Log->debug("Copy the keytab file because the to option was supplied.");
        $self->keytab_copy(keytab => $tmpkeytab, to => $opts{to});
    }

    return $tmpkeytab;
}

#---------------------------------------------------------------------
sub keytab_copy {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {keytab => {type => SCALAR},
                to     => {type => SCALAR,},});
    $Log->enter();
    my $tmpkeytab = $opts{keytab};

    #
    # Copy keytab to specified location with 777 permissions
    #
    my $cmd =
        $self->{sudo}
      . " chmod 777  "
      . $tmpkeytab . ';'
      . $self->{sudo} . ' mv '
      . $tmpkeytab . " "
      . $opts{to};
    my $out = sayEOF($self->{conn}, $cmd);
    croak "$cmd failed: $out" if ($out =~ /fail|error/i);

    $Log->exit();
    return;
}

#---------------------------------------------------------------------
sub add_host {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {host    => {type => SCALAR,},
                service => {type => SCALAR, default => "root",},});
    $Log->enter();

    my $qualifiedHostName = $opts{host};
    if ($qualifiedHostName !~ /$self->{domain}/i) {
        $qualifiedHostName .= '.' . $self->{domain};
    }

    eval { $self->delete_host(%opts); };
    if ($@ && $@ !~ /Principal does not exist while deleting principal/) {
        croak "add_host: delete_host failed: $@";
    }

    my $cmd =
        $self->{sudo} . " "
      . $self->{kadmin}
      . " -q 'add_principal -e \"des-cbc-crc:normal des-cbc-md5:normal\" -randkey "
      . $opts{service} . "/"
      . $qualifiedHostName . "'";
    my $out = sayEOF($self->{conn}, $cmd);

    my $expmsg =
        "Principal \""
      . $opts{service} . "/"
      . $qualifiedHostName . "\@"
      . $self->{realm}
      . "\" created.";
    croak "$cmd failed: --$out-- did not match for --$expmsg--"
      if ($out !~ /$expmsg/);

    $Log->exit();
    return;
}

#---------------------------------------------------------------------
sub delete_host {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {host    => {type => SCALAR,},
                service => {type => SCALAR, default => "root",},});
    $Log->enter();

    my $qualifiedHostName = $opts{host};
    if ($qualifiedHostName !~ /$self->{domain}/i) {
        $qualifiedHostName .= '.' . $self->{domain};
    }

    my $cmd =
        $self->{sudo} . " "
      . $self->{kadmin}
      . " -q 'delete_principal -force "
      . $opts{service} . "/"
      . $qualifiedHostName . "'";
    my $out = sayEOF($self->{conn}, $cmd);

    my $expmsg =
        "Principal \""
      . $opts{service} . "/"
      . $qualifiedHostName . "\@"
      . $self->{realm}
      . "\" deleted.";
    croak "$cmd failed: --$out-- did not match for --$expmsg--"
      if ($out !~ /$expmsg/);

    $Log->exit();
    return;
}

#---------------------------------------------------------------------
sub add_user {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {user   => {type => SCALAR,},
                passwd => {type => SCALAR, default => "netapp",},});
    $Log->enter();

    eval { $self->delete_user(user => $opts{user}); };
    if ($@ && $@ !~ /Principal does not exist while deleting principal/) {
        croak "add_host: delete_host failed: $@";
    }

    my $cmd =
        $self->{sudo} . " "
      . $self->{kadmin}
      . " -q 'add_principal "
      . $opts{user} . "'";

    my ($prompt, $match, $before, $after, $out);

    my $pwprompt =
      "password for principal \"" . $opts{user} . "\@" . $self->{realm} . "\":";
    (undef, undef, undef, $prompt) =
      Tharn::say($self->{conn}, $cmd, '-prompt', $pwprompt);
    while ($prompt =~ $pwprompt) {
        ($match, $before, $after, $prompt) =
          Tharn::say($self->{conn}, $opts{passwd}, '-prompt', $pwprompt);
    }
    $out = $before . $match . $after;

    my $expmsg =
      "Principal \"" . $opts{user} . "\@" . $self->{realm} . "\" created.";
    croak "$cmd failed: --$out-- did not match for --$expmsg--"
      if ($out !~ /$expmsg/);

    $Log->exit();
    return;
}

#---------------------------------------------------------------------
sub delete_user {
    my $self = shift;
    my (%opts) = validate(@_, {user => {type => SCALAR,},});
    $Log->enter();

    my $cmd =
        $self->{sudo} . " "
      . $self->{kadmin}
      . " -q 'delete_principal -force "
      . $opts{user} . "'";
    my $out = SayWrap::sayEOF($self->{conn}, $cmd);

    my $expmsg =
      "Principal \"" . $opts{user} . "\@" . $self->{realm} . "\" deleted.";
    croak "$cmd failed: --$out-- did not match for --$expmsg--"
      if ($out !~ /$expmsg/);

    $Log->exit();
    return;
}

#---------------------------------------------------------------------

1;

