#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     Common library to manage/maintain Limits of various ONTAP(NG)
##              subsystems.
##
## @author      dl-ssqa@netapp.com
##
## @pod here

=head1 NAME

Limits - Common library to manage/maintain Limits of various ONTAP(NG) subsystems.

=head1 DESCRIPTION

C<Limits.pm> can be called in two different ways as direct CLASS methods or OBJECT methods.
OBJECT support is in place to fulfill the need of the caller making many calls in the same
context with the same ONTAP release, platform, and the Mode.
Direct CLASS methods support is available, so one could make a bulk request to receive all
they need in one single call.

This library needs either the following 3 inputs:
(1) C<ONTAP release name>
(2) C<ONTAP Mode>
(3) C<Platform Model Nmae>
, or a single input
(1) C<Hostname of the ONTAP Node>

When hostname of the NODE is provided, Limits.pm will attempt to gather the required values from the ONTAP
NODE directly using C<NACL> library. If the NODE is not in the state where it can give out the required
details, the library will fail with appropriate errors.

=head1 SYNOPSIS

**************** OBJECT CONTEXT ********************

Example #1 - Get limit value for C<max_nfs_stateid>
    
    package Limits;
    my $Limits = new Limits(hw => "FAS6080", mode => "7Mode", sw => "rollingrock_0");
    my $max_nfs_stateid = $Limits->getvalue(limit => "max_nfs_stateid", subsystem => "nfs");
    print "Limit max_nfs_stateid has value of $max_nfs_stateid\n";

Example #2 - Get limit value for C<nfsv4_max_clients> and C<nfsv4_max_stateids>

    package Limits;
    my $Limits = new Limits(host=>"rtpqa-fas6080-8");
    my $max_nfs_stateid = $Limits->getvalue(limit => "max_nfs_stateid", subsystem => "nfs");
    my $max_nfs_clients = $Limits->getvalue(limit => "max_nfs_clients", subsystem => "nfs");
    print "Limit max_nfs_stateid has value of $max_nfs_stateid\n";
    print "Limit max_nfs_clients has value of $max_nfs_clients\n";

Example #3 - Get all C<wafl> Limits

    package Limits;
    my $Limits = new Limits(host=>"rtpqa-fas6080-8");
    my @Limits = $Limits->get(subsystem =>"wafl");
    foreach my $hashref (@Limits) {
        print "Limit $hashref->{name} has value of $hashref->{value}\n";
    }

**************** CLASS CONTEXT ********************

Example #4 - Get all C<wafl> Limits

    package Limits;
    my @Limits = Limits::get(host => "rtpqa-fas6080-8", subsystem =>"wafl");
    foreach my $hashref (@Limits) {
        print "Limit $hashref->{name} has value of $hashref->{value}\n";
    }

Example #5 - Get all C<san> Limits

    package Limits;
    my @Limits = Limits::get(hw => "FAS6040", mode => "7Mode", sw => "rollingrock_0", subsystem =>"san");
    foreach my $hashref (@Limits) {
        print "Limit $hashref->{name} has value of $hashref->{value}\n";
    }

Example #6 - Find out if a combination of ONTAP release and Platform is supported.

    package Limits;
    if(! Limits::is_supported(hw => "FAS3050", mode => "7Mode", sw => "rollingrock_0")) {
        die "FAS3050 is not supported in rollingrock_0";
    }

Example #7 - Get the list of platforms supported for a specific ONTAP release and a Mode.

     package Limits;
     my @platforms = Limits::supported_platforms(mode => "7Mode", sw => "rollingrock_0");
     print "rollingrock_0/7Mode support the following platforms:  @platforms\n";

Example #8 - Get the list of ONTAP releases supported for a specific Platform and a Mode.

    package Limits;
    my @releases = Limits::supported_releases(hw => "FAS6080", mode => "7Mode");
    print "FAS6080/7Mode can run the following releases: ". join(" ", @releases);

Example #9 - Get the list of subsystems available for a specified config.

    package Limits;
    my @subsystems = Limits::subsystems(hw => "FAS6080", mode => "7Mode");
    foreach my $subsys (@subsystems) {
        my @limits = Limits::get(hw => "FAS6080", mode => "7Mode",sw => "rollingrock_0", subsystem => "$subsys");
        foreach my $hashref (@limits) {
        print "Limit $hashref->{name} has value of $hashref->{value}\n";
     }
    }

=head1 TODO

  * Supporting hardware platforms without MODEL numbers assigned yet?

=head1 SEE ALSO

  NACL::APISet::Version
  http://wikid.netapp.com/w/Stress-Scalability_QA/Limits

=head1 BUGS

  Some probably exist but have not been listed yet.

=head1 STATUS

Shared

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id: $
  Last revised $Date: 2009/02/18 $

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved

=cut

package Limits;

use strict;
use warnings;

use Options;
use Tharn qw($Log host config_result script_result);
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use YAML;
use NACL::APISet::Version;
use Data::Dumper qw(Dumper);
use JSON;
use Carp qw(croak);
require LWP::UserAgent;
require HTTP::Request;

my @modes = ( "7Mode", "CMode" );

my $Options = new Options;
$Options->parameters(
    definitions => [
        {
            name        => "SERVER_NAME",
            spec        => "s",
            default     => "limitsfinder-prd.gdl.englab.netapp.com/limits",
            description => "Import from NATE params",
        },
    ]
);

$Options->process();
$Options->set( 'STOP_ON_ERROR', 0 );
my $Options_hash = $Options->get_hash();
my $server_name  = $Options_hash->{'SERVER_NAME'};
#######################Advertised Routines ################################

$Log->comment("Server name: $server_name");

=head1 METHODS

=head2 new

    my $Lim = new Limits(hw => "FAS6210", mode => "7Mode", sw => "rollingrock_0");
    (OR)
    my $Lim = new Limits(host => "rtpqa-fas6080-7");

This function returns Limits Object on success. By default, it loads Limits
values for all sub systems, if specific subsystem is not asked. If C<host> is
given instead of C<hw>, C<mode> and C<sw>, then those values are
found by using C<NACL> libraries to load appropriate Limits valueset.

On failure, it dies with approparite message.

=over

=item Options

=over

=item C<< host => $hostname >>

(Required). The hostname of the NODE.

=item C<< hw => $model >>

(Optional). The platform model name.

=item C<< mode => $node >>

(Optional). The mode whether CMode or 7Mode

=item C<< sw => $sw >>

(Optional). The release name.

=item C<< subsystem => $subsystem >>

(Optional). The name of the subsystem (wafl or nfs or cifs etc). Default: all

=back

=back

=cut

sub new {
    $Log->enter();
    my ( $invocant, %params ) = validate_mixed(
        @_,
        [ { type => SCALAR, } ],
        {
            hw        => { type => SCALAR, optional => 1 },
            mode      => { type => SCALAR, optional => 1 },
            sw        => { type => SCALAR, optional => 1 },
            host      => { type => SCALAR, optional => 1 },
            limit     => { type => SCALAR, optional => 1 },
            subsystem => { type => SCALAR, default  => "all" },
        }
    );

    my %cparams;
    if ( $params{host} ) {
        %cparams = _host2params( $params{host} );
    }
    my $self = { %params, %cparams };
    bless( $self, $invocant );
    if ( !$self->{hw} || !$self->{mode} || !$self->{sw} ) {

        # Make sure either 'host' or 'hw', 'mode' and 'sw' are given
        die script_result(
            "Make sure ('host') or ('hw','mode','sw') are given: "
              . Dumper($self) );
    }
    $Log->debug("HW: $self->{hw}, Mode: $self->{mode}, SW: $self->{sw}");

    #Replace an _ in release name to a .
    $self->{sw} =~ s/\_+/./g;
    if (
        !is_supported(
            hw   => $self->{hw},
            mode => $self->{mode},
            sw   => $self->{sw}
        )
      )
    {
        die
"Release $self->{sw} does not support $self->{hw} platform for $self->{mode}.";
    }
    $Log->debug( "Creating Limits object for " . Dumper($self) );

    #
    # Get list of supported platforms to save for later use.
    #
    my @splatforms =
      supported_platforms( mode => $self->{mode}, sw => $self->{sw} );
    $self->{_supported_platforms} = \@splatforms;

    #
    # Now, it's time to load all relevant limit values for this object
    # based on hardware and software versions
    #
    my @allsubsys = subsystems( mode => $self->{mode} );
    my @selected =
      grep { ( $self->{subsystem} eq "all" ) || ( $_ eq $self->{subsystem} ) }
      @allsubsys;
    if ( !@selected ) {
        die
          "None of the subsystems could be selected. Please verify subsystem=\""
          . $self->{subsystem} . "\"\n";
    }
    foreach my $subsys (@selected) {
        my @limits;
        my $deserialized =
          _json_call( 'js'
              . $self->{mode} . '/sw:'
              . $self->{sw}
              . '/subsys:'
              . $subsys
              . '/limit' );

        foreach my $hr ( @{$deserialized} ) {
            push @limits, $hr;
        }

        #
        # Save 'value' as well for the specified 'hw'
        #
        for ( my $i = 0 ; $i < scalar(@limits) ; $i++ ) {
            $limits[$i]{value} = $limits[$i]{hw_value}{ $self->{hw} };
            $limits[$i]{customer_value} =
              $limits[$i]{hw_customer_value}{ $self->{hw} };
        }

        push @{ $self->{_limits} }, @limits;
    }
    $Log->exit($self);
    return $self;
} ## end sub new

=head2 get

    my @limits = Limits::get(hw => "FAS6210", mode => "7Mode", sw => "rollingrock_0");
    (OR)
    my @limits = Limits::get(host => "rtpqa-fas6080-7", subsystem => "wafl");
    (OR)
    my $Limit = new Limits(hw => "FAS6210", mode => "7Mode", sw => "rollingrock_0");
    my @limits = $Limit->get();

This function returns list of Hashref each reprenting a limit with details.
By default, returns Limits for all sub systems, if specific subsystem is not asked.
If C<host> is given instead of C<hw>, C<mode> and C<sw> , then those values are
found by using C<NACL> libraries to return appropriate Limits valueset.

On failure, it dies with approparite message.

=over

=item Options

=over

=item C<< host => $hostname >>

(Required). The hostname of the NODE.

=item C<< hw => $model >>

(Optional). The platform model name.

=item C<< mode => $node >>

(Optional). The mode whether CMode or 7Mode

=item C<< sw => $sw >>

(Optional). The release name

=item C<< subsystem => $subsystem >>

(Optional). The name of the subsystem (wafl or nfs or cifs etc). Default: all

=item C<< limit => $limit >>

(Optional). The name of a specific limit. Default: undef

=back

=back

=cut

sub get {
    $Log->enter();
    my ($self, %params );
    $self = shift if ( ( @_ % 2 ) == 1 );
    if ( ref($self) ne "Limits" ) {
        $self = new Limits(@_);
    }
    else {
        %params = validate(
            @_,
            {
                limit     => { type => UNDEF | SCALAR, optional => 1 },
                subsystem => { type => UNDEF | SCALAR, default  => "all", },
            }
        );
    }
    my $limit  = $params{limit}     || $self->{limit};
    my $subsys = $params{subsystem} || $self->{subsystem};
    if ( $subsys && $limit ) {
        return grep {
            ( $_->{name} eq $limit )
              && ( ( $subsys eq "all" )
                || ( $_->{subsystem} eq $subsys ) )
        } @{ $self->{_limits} };
    }
    elsif ($subsys) {
        return
          grep { ( $subsys eq "all" ) || ( $_->{subsystem} eq $subsys ) }
          @{ $self->{_limits} };
    }
    elsif ($limit) {
        return grep { $_->{name} eq $limit } @{ $self->{_limits} };
    }
    else {
        return @{ $self->{_limits} };
    }
} ## end sub get

=head2 getvalue

    my @limits = Limits::get(hw => "FAS6210", mode => "7Mode", sw => "rollingrock_0");
    (OR)
    my @limits = Limits::get(host => "rtpqa-fas6080-7", subsystem => "wafl");

This function returns list of Hashref each reprenting a limit with details.
By default, returns Limits for all sub systems, if specific subsystem is not asked.
If C<host> is given instead of C<hw>, C<mode> and C<sw>, then those values are
found by using C<NACL> libraries to return appropriate Limits valueset.

On failure, it dies with approparite message.

=over

=item Options

=over

=item C<< host => $hostname >>

(Required). The hostname of the NODE.

=item C<< hw => $model >>

(Optional). The platform model name.

=item C<< mode => $node >>

(Optional). The mode whether CMode or 7Mode

=item C<< sw => $sw >>

(Optional). The release name.  

=item C<< subsystem => $subsystem >>

(Optional). The name of the subsystem (wafl or nfs or cifs etc). Default: all

=item C<< limit => $limit >>

(Required). The name of a specific limit. Default: undef

=back

=back

=cut

sub getvalue {
    $Log->enter();
    my ( $self, %params );
    $self = shift if ( ( @_ % 2 ) == 1 );
    if ( ref($self) ne "Limits" ) {
        $self = new Limits(@_);
    }
    else {
        %params = validate(
            @_,
            {
                limit     => { type => SCALAR,         optional => 0 },
                subsystem => { type => UNDEF | SCALAR, default  => "all", },
            }
        );
    }
    my $limit = $params{limit} || $self->{limit};
    my @limits;

    if ( $params{subsystem} && $limit ) {
        @limits = grep {
            ( $_->{name} eq $limit )
              && ( ( $params{subsystem} eq "all" )
                || ( $_->{subsystem} eq $params{subsystem} ) )
        } @{ $self->{_limits} };
    }
    elsif ($limit) {
        @limits = grep { $_->{name} eq $limit } @{ $self->{_limits} };
    }
    else {
        die "Required param \"limit\" was not specified\n";
    }
    return $limits[0]->{value};
} ## end sub getvalue

=head2  get_hw

C<<get_hw>> returns the limit values for all hardware supported for the release.
While C<<get()>> requires 'hw' and returns value for a specific hardware, C<<get_hw>>
is used to get values for all supported platforms.

=over

=item Options

=over

=item C<< mode => $node >>

(Required). The mode whether CMode or 7Mode

=item C<< subsystem => $subsystem >>

(Optional). The name of the subsystem (wafl or nfs or cifs etc). Default: all

=item C<< limit => $limit >>

(Required). The name of a specific limit. Default: undef

=back

=back

=cut

sub get_hw {
    $Log->enter();
    my ( $self, %params );
    $self = shift if ( ( @_ % 2 ) == 1 );
    %params = validate(
        @_,
        {
            mode  => { type => SCALAR },
            limit => { type => SCALAR },
        },
    );

    my $release = $params{sw} || $self->{sw};
    my @platforms =
      supported_platforms( mode => $params{mode}, sw => $release );
    my $hw     = $platforms[ rand($#platforms) ];
    my @limits = get(
        hw    => $hw,
        mode  => $params{mode},
        sw    => $release,
        limit => $params{limit},
    );

    for ( my $i = 0 ; $i < scalar(@limits) ; $i++ ) {
        my %temp;
        foreach my $hwl (@platforms) {
            if ( $limits[$i]{hw_value}{$hwl} ) {
                $temp{$hwl} = $limits[$i]{hw_value}{$hwl};
            }
            else {
                $temp{$hwl} = getvalue(
                    hw    => $hwl,
                    mode  => $params{mode},
                    sw    => $release,
                    limit => $limits[$i]{name}
                );
            }
        }
        $limits[$i]{value} = \%temp;
    }
    return @limits;
} ## end sub get_hw

=head2 is_supported

    Limits::is_supported(hw => "FAS6210", mode => "CMode", sw => "rollingrock_0");

Returns boolean whether or not the given combination is officially supported. If this library
does not support the given ONTAP release, then it dies with the error message.

=over

=item Options

=over

=item C<< hw => $model >>

(Required). The name of the Platform Model.

=item C<< mode => $mode >>

(Required). The mode whether CMode or 7Mode

=item C<< sw => $sw >>

(Required). The release name

=back

=back

=cut

sub is_supported {
    my (%params) = validate(
        @_,
        {
            hw   => { type => SCALAR, optional => 0 },
            mode => { type => SCALAR, optional => 0 },
            sw   => { type => SCALAR, optional => 0 },
        }
    );

    my @platforms =
      supported_platforms( mode => $params{mode}, sw => $params{sw} );
    return grep { $_ eq "$params{hw}" } @platforms;
} ## end sub is_supported

=head2 subsystems

    my @subsystems = Limits::subsystems(mode => "CMode", hw => "FAS6080");

Returns a list of subsystems supported by the ONTAP Release, Platform and the Mode.

=over

=item Options

=over

=item C<< mode => $mode >>

(Required). The mode whether CMode or 7Mode

=back

=back

=cut

sub subsystems {
    my (%params) = validate(
        @_,
        {
            hw   => { type => SCALAR, optional => 1 },
            mode => { type => SCALAR, optional => 0 },
        }
    );

    my $deserialized = _json_call( 'js' . $params{mode} . '/subsys' );

    my @subsystemarr = @{ $deserialized->{'subsystem_name'} };
    return @subsystemarr;
} ## end sub subsystems

=head2 supported_platforms

    my @platforms = Limits::supported_platforms(mode => "CMode", sw => "rollingrock_0");

Returns a list of platforms supported by the ONTAP Release and the Mode.

=over

=item Options

=over

=item C<< mode => $mode >>

(Required). The mode whether CMode or 7Mode

=item C<< sw => $sw >>

(Required). The release name.

=back

=back

=cut

sub supported_platforms {
    my (%params) = validate(
        @_,
        {
            mode => { type => SCALAR, optional => 0 },
            sw   => { type => SCALAR, optional => 0 },
        }
    );

    $params{sw} =~ s/\_+/./g;
    my $deserialized =
      _json_call( 'js' . $params{mode} . '/sw:' . $params{sw} );

    my @platformarr = @{ $deserialized->{'platform_name'} };
    my %platformhash;

    @platformhash{@platformarr} = ();
    if ( keys %platformhash ) {
        return sort ( keys %platformhash );
    }
    else {
        die "No platform found for $params{mode} and $params{sw}\n";
    }
} ## end sub supported_platforms

=head2 supported_releases

    my @releases = Limits::supported_releases(mode => "7Mode", hw => "DOvM100");

Returns a list of platforms supported by the ONTAP Release and the Mode.

=over

=item Options

=over

=item C<< mode => $mode >>

(Required). The mode whether CMode or 7Mode

=item C<< hw => $model >>

(Required). The platform model name.

=back

=back

=cut

sub supported_releases {
    my (%params) = validate(
        @_,
        {
            hw   => { type => SCALAR, optional => 0 },
            mode => { type => SCALAR, optional => 0 },
        }
    );

    my $deserialized =
      _json_call( 'js' . $params{mode} . '/hw:' . $params{hw} );

    my @releasesarr = @{ $deserialized->{'release_name'} };
    my %releaseshash;

    @releaseshash{@releasesarr} = ();

    if ( keys %releaseshash ) {
        return sort ( keys %releaseshash );
    }
    else {
        die "No release found for $params{hw} and $params{mode}\n";
    }
} ## end sub supported_releases

########################## Internal Routines ##################################

sub _host2params {
    my $hostname = shift;
    my %params;
    my $host = Tharn::host($hostname);
    $params{mode} = ( $host->hosttype() =~ /\-ng/ ) ? "CMode" : "7Mode";
    my $vm = NACL::APISet::Version->new(
        hostobj    => $host,
        category   => "Node",
        shell_type => $params{mode},
    );
    my $ver = $vm->get_version_attribute( attribute => "release" );
    $params{sw} = $vm->map_release_to_codename( release => $ver );
    $params{hw} = $host->{MODEL};
    return %params;
} ## end sub _host2params

sub _json_call {

    my $relative_path = shift;
    my $deserialized;
	my $url;
    eval {
        $url = 'http://'. $server_name . '/JSON/' . $relative_path;
        my $ua =
          LWP::UserAgent->new( 'default_headers' =>
              HTTP::Headers->new( Content_Type => 'application/json' ) );
        my $request =
          HTTP::Request->new( 'GET', $url,
            HTTP::Headers->new( Content_Type => 'application/json' ) );
        my $response    = $ua->request($request);
        my $json_string = $response->content;
        $deserialized = from_json($json_string);
    };

    croak( "json call failed with : " . $@ . "\n HTTP request that resulted in json fail is: $url !!") if ($@);

    return $deserialized;

}    ##end _json_call

1;
