package LogReader;

use LogReader::Base;
@ISA = qw( LogReader::Base );

use LogReader::Messages;
use LogReader::SnapMirror;
use LogReader::NBUSnapVault;
use LogReader::Audit;
use LogReader::Backup;
use LogReader::EMS;
use LogReader::MlogMessages;
use LogReader::MlogMgwd;
use LogReader::MlogVifmgr;
use LogReader::MlogVldb;
use LogReader::MlogBcomd;
use Carp;

use warnings;
use strict;

our $Index;

=head1 NAME

  LogReader - Read Filer Logs
  
=head1 DESCRIPTION

  This uses LogReader::* objects to access logfiles on the filer.
  
=head1 SYNOPSIS

  use LogReader;
  
  # For 7-mode filers or earlier
  host($FILER);
  my $filer = connect($FILER_CONN);
  my $f = LogReader->new( $filer );
  
  # For 7-mode, or 10-mode, or later filers
  my $Server = Server->new();
  my $f = LogReader->new( $Server );

  # During setup
  $f->start( [ 'snapmirror', 'messages', 'nbu_snapvault', 'auditlog', 'backup', 'ems', 'mlog_messages', 'mlog_mgwd' ] );

  #### Run test here ###

  # During cleanup
  $f->stop;
  $f->logcomment;	# Includes the logs in the NATE logs
  
  if ( $f->panicked ) {
      logresult( 'FAIL', 'PANIC detected in messages log.' );
  }

=head1 METHODS

=head2 new()

  The constructor.  Takes takes a Filer Connectrec object or Server object.
  Supports both 7 and 10-mode filers. Filer Connectrec object only works
  with 7-mode filers. Server object will work with both 7-mode and 10-mode
  filers.

=head2 start()

  Takes a reference to an array of log types (snapmirror|messages|nbu_snapvault|auditlog|backup|ems|mlog_messages|mlog_mgwd).  
  Returns nothing.

  Starts tracking log entries for each log type specified.

=cut

sub start {

    my ( $self, $types ) = @_;

    $self->{types} = $types;

    for (@$types) {

        if ( $_ eq 'messages' ) {
            $self->{logs}{messages}
                = LogReader::Messages->new( $self->{connectrec} );
            $self->{logs}{messages}->start;
        } elsif ( $_ eq 'snapmirror' ) {
            if ( $self->{ctype} ne 'nate' ) {
                croak "$_ not yet supported with Server.pm";
            }
            $self->{logs}{snapmirror}
                = LogReader::SnapMirror->new( $self->{connectrec} );
            $self->{logs}{snapmirror}->start;
        } elsif ( $_ eq 'nbu_snapvault' ) {
            if ( $self->{ctype} ne 'nate' ) {
                croak "$_ not yet supported with Server.pm";
            }
            $self->{logs}{nbu_snapvault}
                = LogReader::NBUSnapVault->new( $self->{connectrec} );
            $self->{logs}{nbu_snapvault}->start;
        } elsif ( $_ eq 'auditlog' ) {
            if ( $self->{ctype} ne 'nate' ) {
                croak "$_ not yet supported with Server.pm";
            }
            $self->{logs}{auditlog}
                = LogReader::Audit->new( $self->{connectrec} );
            $self->{logs}{auditlog}->start;
        } elsif ( $_ eq 'backup' ) {
            if ( $self->{ctype} ne 'nate' ) {
                croak "$_ not yet supported with Server.pm";
            }
            $self->{logs}{backup}
                = LogReader::Backup->new( $self->{connectrec} );
            $self->{logs}{backup}->start;
        } elsif ( $_ eq 'ems' ) {
            $self->{logs}{ems} = LogReader::EMS->new( $self->{connectrec} );
            $self->{logs}{ems}->start;
        } elsif ( $_ eq 'mlog_messages' ) {
            $self->{logs}{mlog_messages}
                = LogReader::MlogMessages->new( $self->{connectrec} );
            $self->{logs}{mlog_messages}->start;
        } elsif ( $_ eq 'mlog_mgwd' ) {
            $self->{logs}{mlog_mgwd}
                = LogReader::MlogMgwd->new( $self->{connectrec} );
            $self->{logs}{mlog_mgwd}->start;
        } elsif ( $_ eq 'mlog_vldb' ) {
            $self->{logs}{mlog_vldb}
                = LogReader::MlogVldb->new( $self->{connectrec} );
            $self->{logs}{mlog_vldb}->start;
        } elsif ( $_ eq 'mlog_vifmgr' ) {
            $self->{logs}{mlog_vifmgr}
                = LogReader::MlogVifmgr->new( $self->{connectrec} );
            $self->{logs}{mlog_vifmgr}->start;
        } elsif ( $_ eq 'mlog_bcomd' ) {
            $self->{logs}{mlog_bcomd}
                = LogReader::MlogBcomd->new( $self->{connectrec} );
            $self->{logs}{mlog_bcomd}->start;
        } else {
            croak
                "Unrecognized log type (messages|snapmirror|nbu_snapvault|auditlog|backup|ems|mlog_messages|mlog_mgwd|mlog_vifmgr|mlog_vldb|mlog_bcomd).";
        }
    }
} ## end sub start

=head2 stop()

  Takes nothing.  Returns nothing.

  Stops tracking log entries for each log type.  

=cut

sub stop {

    my ($self) = @_;

    for ( keys %{ $self->{logs} } ) {
        $self->{logs}{$_}->stop;
    }
} ## end sub stop

=head2 logcomment()

  Takes nothing.  Returns nothing.
  
  Makes a NATE log entry for each defined log type.

=cut

sub logcomment {

    my ($self) = @_;

    Tharn::logcomment("Test Execution Logs for Filer: \'$self->{filer}\'");

    for ( sort keys %{ $self->{logs} } ) {
        print $self->{logs}{$_}->to_log;
    }
    print "\n";
} ## end sub logcomment

=head2 panicked()

  Takes nothing.  Returns 1 if a PANIC was logged in the messages log, 
  otherwise 0.

=cut

sub panicked {

    my ($self) = @_;

    croak "You can't check for PANICs without tracking the messages log."
        unless ( $self->{logs}{messages} || $self->{logs}{mlog_messages} );

    if ( defined( $self->{logs}{messages} ) ) {
        if ( $self->{logs}{messages}->to_string =~ /PANIC/gs ) {
            return 1;
        }
    } elsif ( defined( $self->{logs}{mlog_messages} ) ) {
        if ( $self->{logs}{mlog_messages}->to_string =~ /Panic String/gs ) {
            return 1;
        }
    } else {
        return 0;
    }

} ## end sub panicked

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

    LogReader::Base;
    LogReader::Messages;
    LogReader::SnapMirror;
    LogReader::NBUSnapVault;
    LogReader::Audit;
    LogReader::Backup;
    LogReader::EMS;
    LogReader::MlogMessages;

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)

=head1 CHANGES

	08/05/2010 talluri Added support to read vldb.log,vifmgr.log,bcomd.log
  
=cut

