package LogReader::Backup;

use LogReader::Base;
@ISA = qw( LogReader::Base );

use Data::Dumper;
use strict;

our $Path = '/etc/log/backup';

=head1 NAME

  LogReader::Backup - Grab the Relevant Section of the backup Log(s)
  
=head1 DESCRIPTION

  This grabs the sections of the backup logs that were generated during
  the course of a test.  The object grabs the logs with 'rdfile' issued
  to a Filer. NOTE: Has not yet been ported to work with Server.pm.
  
=head1 SYNOPSIS

  use LogReader::Backup;

  host($FILER);
  my $filer = connect($FILER_CONN);
  
  my $f = LogReader::Backup->new( $filer );
  
  # Start recording.
  $f->start;
  
  ### Conduct your test here.
  
  # Stop recording
  $f->stop;

=head1 METHODS

=head2 new()

  The constructor.  Takes takes a Filer Connectrec object.

=head2 start()

  Takes no arguments.  Returns nothing.
  
  This begins recording the log(s).

=cut

sub start {
	
	my ( $self ) = @_;
	
	$self->{starttime}{time} = time();
	$self->{starttime}{localtime} = localtime( $self->{starttime}{time} );

	my $results = $self->rdfile( $Path );
	
	# This sets the point after which we will be interested in the log.
	
	$self->{mark}{index} = $#$results;
	$self->{mark}{string} = $results->[$#$results];
}	

=head2 stop()

  Takes no arguments.  Returns the log index indicating where the reader
  stopped reading.
  
  This ends recording the log(s).
  
=cut

sub stop {

	my ( $self ) = @_;

	$self->{endtime}{time} = time();
	$self->{endtime}{localtime} = localtime( $self->{endtime}{time} );

	my @snippet = ();

	my $null = $self->rdfile( $Path );
	
	$self->{mark}{index} = $LogReader::Index if defined $LogReader::Index;
	
	# This is the simplest case.  This means the mark is in the 'null' log (eg 
	# 'messages').  Therefore, the logs didn't rotate during test execution.  
	# Grab everything after the mark in the null log.
	
	if ( $self->{mark}{index} <= $#$null ) {
	
		unless ( $self->{mark}{index} == $#$null ) {
	 		@snippet = @$null[$self->{mark}{index}+1..$#$null];
		}
		
		$self->{current_log} = $Path;
	
	# The mark is NOT in the 'null' log.  It must have been rotated during the test.  
	# Therefore grab everything after the mark in 'messages.0' and concatenate 
	# the entire 'null' log onto it.
	
	} else {
	
		my $zero = $self->rdfile( "$Path.0" );
		
		unless ( $self->{mark}{index} == $#$zero ) {
			@snippet = @$zero[$self->{mark}{index}+1..$#$zero];
		}
		
		@snippet = ( @snippet, @$null );
		
		$self->{current_log} = "$Path.0";
	}
	
	$self->{snippet} = \@snippet;

	return $#$null;
}

=head2 to_string()

  Takes no arguments.  Returns a string representation of the log.
  
=cut

sub to_string {

	my ( $self ) = @_;
	
	my $string = "\nBackup Log:\n\n";
	for ( @{$self->{snippet}} ) {
		$string .= "$_";
	}
	
	$string .= "\n";
	
	return $string;
}

=head2 to_log()

  Takes no arguments.  Returns a copy of the log with each line
  terminated by a newline.
  
=cut

sub to_log {

	my ( $self ) = @_;
	
	my $string = "\nMessages Log:\n\n";
	for ( @{$self->{snippet}} ) {
		$string .= "$_\n";
	}
	
	return $string;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  LogReader::Base.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut
















