# $Id$
#
## @summary LogReader Base Class
##
## @status public
##
## @author Todd Shoenfelt (todd@netapp.com), dl-dpgqa-automation@netapp.com
##
## @pod here
##

=head1 NAME

  LogReader::Base - Abstract Base Class For Log Classes

=head1 DESCRIPTION

  Log classes inherit methods from this class.

=head1 SYNOPSIS

  use LogReader::Base;
  @ISA = qw( LogReader::Base );

=head1 METHODS

=head2 new()

  The constructor.  Takes takes a Filer Connectrec object.

=cut

package LogReader::Base;

use strict;
use Carp qw(croak);
use Server;
use SPARSE::Util::Logger;
use Tharn qw(logcomment);

our $Verbose = 1;
# Allow up to 5 minutes to dump the log file across the network
# for a Server.pm type connection
our $Cmd_timeout = 600;
# Set the NATE say maximum buffer size to 128MB. It defaults to
# 1MB which causes the read of large log files to be truncated.
our $Bufsize = 128 * 1024 * 1024;

sub new {
	my ( $class, $fobj ) = @_;
        my ( $fc, $out );

	my $self = bless {}, $class;

	$self->{connectrec} = $fobj;
	if ($fobj->isa( 'Connectrec' )) {
		$self->{filer} = $fobj->{hostp}{default_ip};
		$self->{username} = $fobj->username;
		$self->{password} = $fobj->password;
                # NATE connectrec is always 7-mode
		$self->{mode} = 7;
		$self->{ctype} = 'nate';
                $fobj->maxbufsize($Bufsize);
	} elsif ($fobj->isa( 'Server' )) {
		$self->{ctype} = 'server';
		$self->{filer} = ($fobj->servers)[0];
		$self->{execution_server} = $fobj->get_execution_server ();
                # A Server object could be used for a 7 or 10 mode filer.
		# Determine the mode of the filer.
		# Retry this 2 times because sometimes we get output
		# from the last command run on the server object and the
		# SAY_SWALLOW does not seem to help.
		for (1..2) {
			eval {
				$fobj->cli(cmd         => '?',
					   output      => \$out,
					   '-Timeout'  => $Cmd_timeout,
					   SAY_SWALLOW => 1);
			};
			if ($@) {
				croak "Could not send '?' command to filer to test " .
				      "for filer mode: $@";
			}
			if ($out =~ /up\s+Go up one directory/i) {
				$self->{mode} = 10;
				last;
			} elsif ($out =~ /version/) {
				$self->{mode} = 7;
				last;
			}
		}
		if (!defined($self->{mode})) {
			croak "Cannot figure out the filer mode (known types " .
			"are 10 or 7-mode):\n$out";
		}

                # If possible, set the maxbuf size on the connection object
		# 7-mode requires a cli connection while 10-mode
		# requires a Unix shell connection.
		if ($self->{mode} == 7) {
                    $fc = $fobj->get_connection(type => 'cli_connection');
                } else {
		    #
		    # 20090311	tyu	It's valid that the corresponding
		    #			shell_connection variable in the Node
		    #			object is not initialized by the
		    #			time this portion of code is being
		    #			exercised. Thus, the eval {} block
		    #			is introduced here to handle the
		    #			possible error condition, and try to
		    #			open a shell connectoin at this point.
		    #
		    eval {
			$fc = $fobj->get_connection(type => 'shell_connection');
		    };

		    my $possible_error = "\'shell_connection\' connection had not been created";
		    if ($@ =~ /$possible_error/) {
			$fc = $fobj->open_connection(type => 'shell_connection');
		    }
                }
                $fc->maxbufsize($Bufsize) if (defined($fc));
	} else {
		croak "LogReader::Base constructor requires a Connectrec " .
                      "or Server object.";
	}

	return $self;
}

=head2 hostname

	Returns the hostname of the filer that the logs are being read from

=over

=cut

sub hostname {
    my ($self) = @_;

    return $self->{filer};
}

# This reads the log via the Filer rdfile command, issued via rsh.

sub rdfile {
	my ( $self, $filepath ) = @_;
	my ( @data, $fobj, $out );

	$fobj = $self->{connectrec};

	# For backwards compatibility support 7mode with NATE connections
	if (($self->{mode} == 7) && ($fobj->isa( 'Connectrec' ))) {
		(undef, undef, $out, undef) = Tharn::say($fobj,
		  "priv set -q diag; rdfile $filepath", '-swallow',
		  '-timeout', $Cmd_timeout, '-noverbose');
	} else {
		# Using Server.pm to communicate with the filer
		# 7-mode requires a cli connection while 10-mode
		# requires a Unix shell connection.
		if ($self->{mode} == 7) {
			$fobj->cli(cmd          => "rdfile $filepath",
				   output       => \$out,
				   '-Noverbose' => 1,
				   '-Timeout'   => $Cmd_timeout);
		} else {
			my $orig_server = $fobj->set_execution_server (-Server=>$self->{execution_server});
			$fobj->shell(cmd          => "cat $filepath",
				     output       => \$out,
				     '-Noverbose' => 1,
				     '-Timeout'   => $Cmd_timeout);
			$fobj->set_execution_server (-Server=>$orig_server);
		}
	}
	@data = split(/(\n)/, $out);

	return \@data;
}

=head2 ::rdfilelines

Read portion of a file by means of the range of line numbers.

=over

=item Arguments

B<filepath> = path of the file to be examined.

I<Required>

Specify the path of the file where the block will be read from.

B<startpos> = Line number to identify the start position of the block to read

I<Required>

Specify the line number of the ending position of the block to be read from the file.

B<endpos> = Line number to identify the end position of the block to read

I<Required>

Specify the line number of the ending position of the block to be read from the file.

=item Examples

    my $result = $logrdr->rdfilelines (path=>'/etc/messages', $startpos=>10, $endpos=20);

    if (@$result) {
        my $report = join "\n", @$result;

	NOTE::logcomment ("Wafliron report :\n$report");
    }

=back

=cut

sub rdfilelines {
	my ( $self, $filepath, $startpos, $endpos ) = @_;
	my ( @data, $fobj, $out );

	$fobj = $self->{connectrec};

	# For backwards compatibility support 7mode with NATE connections
	if (($self->{mode} == 7) && ($fobj->isa( 'Connectrec' ))) {
		(undef, undef, $out, undef) = Tharn::say($fobj,
		  "priv set -q diag; rdfile $filepath", '-swallow',
		  '-timeout', $Cmd_timeout, '-noverbose');
		my @tempout = ();
		my $num_of_lines = $#$out;
		if ($num_of_lines >= $endpos) {
		    @tempout = @$out[$startpos..$endpos];
		    $out = \@tempout;
		} elsif ($num_of_lines >= $startpos) {
		    @tempout = @$out[$startpos..$num_of_lines];
		    $out = \@tempout;
		}
	} else {
		# Using Server.pm to communicate with the filer
		# 7-mode requires a cli connection while 10-mode
		# requires a Unix shell connection.
		if ($self->{mode} == 7) {
			$fobj->cli(cmd          => "rdfile $filepath",
				   output       => \$out,
				   '-Noverbose' => 1,
				   '-Timeout'   => $Cmd_timeout);

			my @tempout = ();
			my $num_of_lines = $#$out;
			if ($num_of_lines >= $endpos) {
			    @tempout = @$out[$startpos..$endpos];
			    $out = \@tempout;
			} elsif ($num_of_lines >= $startpos) {
			    @tempout = @$out[$startpos..$num_of_lines];
			    $out = \@tempout;
			}
		} else {
			my $linerange = "".$startpos.",".$endpos;
			logcomment ("linerange is ".$linerange);
			my $orig_server = $fobj->set_execution_server (-Server=>$self->{execution_server});
			$fobj->shell(cmd          => "sed -n '".$linerange."p ' ".$filepath,
				     output       => \$out,
				     '-Noverbose' => 1,
				     '-Timeout'   => $Cmd_timeout);
			$fobj->set_execution_server (-Server=>$orig_server);
		}
	}
	@data = split(/(\n)/, $out);

	return \@data;
}

=head2 ::linecount

Return the number of lines in the input file

=over

=item Arguments

B<path> = path of the file to be examined.

I<Required>

This is the path of the file to be examined.

=item Examples

    my $num_of_lines = $logrdr->linecount (path=>"/mroot/etc/messages");

=back

=cut

sub linecount {
	my ( $self, $filepath, ) = @_;
	my ( @data, $fobj, $out );

	$fobj = $self->{connectrec};

	# For backwards compatibility support 7mode with NATE connections
	if (($self->{mode} == 7) && ($fobj->isa( 'Connectrec' ))) {
		(undef, undef, $out, undef) = Tharn::say($fobj,
		  "priv set -q diag; wc -l $filepath", '-swallow',
		  '-timeout', $Cmd_timeout, '-noverbose');
	} else {
		# Using Server.pm to communicate with the filer
		# 7-mode requires a cli connection while 10-mode
		# requires a Unix shell connection.
		if ($self->{mode} == 7) {
			$fobj->cli(cmd          => "sed -n '$=' $filepath",
				   output       => \$out,
				   '-Noverbose' => 1,
				   '-Timeout'   => $Cmd_timeout);
		} else {
			my $orig_server = $fobj->set_execution_server (-Server=>$self->{execution_server});
			$fobj->shell(cmd          => "sed -n '\$=' $filepath",
				     output       => \$out,
				     '-Noverbose' => 1,
				     '-Timeout'   => $Cmd_timeout);
			$fobj->set_execution_server (-Server=>$orig_server);
		}
	}

	return $out;
}

=head2 matches()

	Captures the list of entries containing the search string
	in the logs. Returns the reference to the list.

=over

=item  matchstring

Required : Match string to look for in the logfile snippet

=cut

sub matches {
	my ( $self, $matchstring ) = @_;

	my @matches = ();
	foreach my $line ( @{$self->{snippet}} ) {
		logDbg( "LogReader::Base::matches() : searching '$line' for '$matchstring'" )
			if $Verbose == 1;
		push( @matches, $line ) if $line =~ /$matchstring/;
	}

	return \@matches;
}

=head2 verify()

	Searches for match string specified in the captured logs.
	Returns 1 if success, 0 on failure.

=over

=item  matchstring

Required : Match string to look for in the logfile snippet

=cut

sub verify {
	my ( $self, $matchstring ) = @_;

	my $matches = $self->matches( $matchstring );
	if ( $Verbose == 1 ) {
		my $msg = "LogReader::Base::verify() found " . @$matches . "\nMatches:\n";
		$msg .= $_ for @$matches;
		logDbg( $msg );
	}

	return @$matches ? 1 : 0;
}

=head2 restart()

  Takes nothing.  Increments the index, so it can check for incremental log
  changes in an event model.

=cut

sub restart {

	my ( $self ) = @_;

	# Call stop and reset the index to where stop() stopped.
	$self->{mark}{index} = $self->stop;
}


sub flush {

	my ( $self ) = @_;

        #don't bother in an empty file
        if ($self->{mark}{index} != 0) {
            $self->{mark}{index} =  $self->{mark}{index} + 1;
        }
}

1;

=head1 TODO

  Make a TODO list.

=head1 DEPENDENCIES

  Server.pm
  SPARSE::Util::Logger.pm

=head1 AUTHOR

  dl-dpgqa-automation@netapp.com

=cut

