package LogReader::EMS;

use LogReader::Base;
@ISA = qw( LogReader::Base );

use strict;
use Carp qw(croak);
use Tharn qw($Log);
use filer qw(filer_cli_say);


our $Path = '/etc/log/ems';
our $Path_10 = "/mroot$Path";

=head1 NAME

  LogReader::EMS - Grab the Relevant Section of the ems Log(s)
  
=head1 DESCRIPTION

  This grabs the sections of the ems logs that were generated during
  the course of a test.  Supports both 7 and 10-mode filers. For 7-mode and
  earlier filers the object grabs the logs with 'rdfile' issued to a Filer.
  For 10-mode or later filers it uses commands available via the Server.pm
  cli to get the needed information trying to avoid reading the whole
  log where possible.

  No special XML processing is performed on the ems log information it
  is returned as a string (scalar).
  
=head1 SYNOPSIS

  use LogReader::EMS;

  # For 7-mode filers or earlier
  host($FILER);
  my $filer = connect($FILER_CONN);
  my $f = LogReader::EMS->new( $filer );
  
  # For 7-mode, or 10-mode, or later filers
  my $Server = Server->new();
  my $f = LogReader::EMS->new( $Server );

  # Start recording.
  $f->start;
  
  ### Conduct your test here.
  
  # Stop recording
  $f->stop;

  # Pass in an array of event names to search for in the log
  # and then print all the error messages returned.
  $err_msgs = $f->find_events(@ems_events);
  foreach (@{$err_msgs}) {
      print "$_\n";
  }

=head1 METHODS

=head2 new()

  The constructor.  Takes a Filer Connectrec object or a Server object.

=head2 start()

  Takes no arguments.  Returns nothing.
  
  This begins recording the log(s).

=cut

sub start {
	
	my ( $self ) = @_;
	my ( $path );
	
        # Set the path based on the mode of the filer
	if ($self->{mode} < 10) {
		$path = $Path;
	} else {
		$path = $Path_10;
	}

	$self->{starttime}{time} = time();
	$self->{starttime}{localtime} = localtime( $self->{starttime}{time} );

	my $results = $self->rdfile( $path );

	# This sets the point after which we will be interested in the log.
	$self->{mark}{index} = $#$results;
	$self->{mark}{string} = $results->[$#$results];
}	

=head2 stop()

  Takes no arguments.  Returns the log index indicating where the reader
  stopped reading.
  
  This ends recording the log(s).
  
=cut

sub stop {

	my ( $self ) = @_;
	my ( $path );

	$self->{endtime}{time} = time();
	$self->{endtime}{localtime} = localtime( $self->{endtime}{time} );

	my @snippet = ();

        # Set the path based on the mode of the filer
	if ($self->{mode} < 10) {
		$path = $Path;
	} else {
		$path = $Path_10;
	}

	my $null = $self->rdfile( $path );
	
	$self->{mark}{index} = $LogReader::Index if defined $LogReader::Index;
	
	# This is the simplest case.  This means the mark is in the 'null' log (eg 
	# 'ems').  Therefore, the logs didn't rotate during test execution.  
	# Grab everything after the mark in the null log.
	
	if ( $self->{mark}{index} <= $#$null ) {
	
		unless ( $self->{mark}{index} == $#$null ) {
	 		@snippet = @$null[$self->{mark}{index}+1..$#$null];
		}
		
		$self->{current_log} = $path;
	
	# The mark is NOT in the 'null' log.  It must have been rotated during the test.  
	# Therefore grab everything after the mark in 'ems.0' and concatenate 
	# the entire 'null' log onto it.
	
	} else {
	
		my $zero = $self->rdfile( "$path.0" );
		
		unless ( $self->{mark}{index} == $#$zero ) {
			@snippet = @$zero[$self->{mark}{index}+1..$#$zero];
		}
		
		@snippet = ( @snippet, @$null );
		
		$self->{current_log} = "$path.0";
	}
	
	$self->{snippet} = \@snippet;

	return $#$null;
}

=head2 start_tail()

  Takes no arguments. Captures the time from when the ems messages are to be
  captured. The time that had passed after the call to start_tail, will be used 
  by 'ems log dump <value>' command to capture the ems messages.

=cut

sub start_tail {

        my ( $self ) = @_;
        $self->{time_start} = time();
}

=head2 stop_tail()

  Stops the capture of EMS logs. Returns what is captured as a scalar. Takes the
  following parameters:
  1. $poll_patterns - reference to an array of EMS messages to wait for 
  2. $poll_timeout - if $patterns was provided, this is the maximum time
     that will be spent waiting for the EMSes to arrive
  3. $poll_step - the amount of time to wait before checking the log again
  4. $timeout - timeout parameter to be used for "ems log dump" command.

=cut


sub stop_tail {

        my ( $self, $poll_patterns, $poll_timeout, $poll_step, $timeout) = @_;

        my ( $root_vol, $time, @snippet, $lines );

        if(!defined($poll_timeout)) {
            $poll_timeout = 300;
        }

        if(!defined($poll_step)) {
            $poll_step = 5;
        }

        # Using 'ems log dump' command to make sure that ems messages that may
        # have been generated but not flushed to /etc/log/ems are pushed into
        # the logs. This filer command makes use of emslog_flush() internally,
        # which helps us achieve this task.

        # Capturing the time that has passed since the call to start_tail
        my $ems_time = time() - $self->{time_start};
   
        # Using 'ignore => 'panic'' option so that filer_cli_say does not 
        # generate a fail result when ems log contains the string 'PANIC'
        # ems log dump faces timeout sometimes.
        # hence using timeout parameter to override default value of 60s.
        if(defined($timeout)){
            $lines = filer_cli_say($self->{connectrec},
                       "priv set -q diag; ems log dump $ems_time",
                       {timeout => $timeout}, {ignore => 'panic'});
        } else {
            $lines = filer_cli_say($self->{connectrec}, "priv set -q diag; "
                               . "ems log dump $ems_time", {ignore => 'panic'});
        }
        # Some ems messages are generated late and are delayed, the following
        # while loop will wait for specified ems messages for a specified amount
        # of time, before returning.

        if(defined($poll_patterns)) {
            while (   ($time <= $poll_timeout)
                   && (!_patterns_present($poll_patterns, $lines))) {
                sleep($poll_step);
                $time += $poll_step;

                # getting the EMSes generated during the last $poll_step seconds
                # ems log dump faces timeout sometimes.  
                # hence using timeout parameter to override default.
                my $new_lines;
                if(defined($timeout)){
                    $new_lines = filer_cli_say($self->{connectrec},
                                "priv set -q diag; ems log dump $poll_step",
                                {timeout => $timeout}, {ignore => 'panic'});
                } else {
                    $new_lines = filer_cli_say($self->{connectrec}, 
                                "priv set -q diag; ems log dump $poll_step", 
                                {ignore => 'panic'});
                }
                $lines .= $new_lines;
            }
        }

        # Storing the output into 'snippet' so that find_events() method
        # can access it
        @snippet = split(/(\n)/, $lines);
        $self->{snippet} = \@snippet;

        if($lines) {
            $Log->comment("EMS messages captured:\n $lines");
        } else {
            $Log->comment("No EMS messages captured");
        }

        # Return the captured entries
        return $lines;
}

=head2 _patterns_present()

  Private function, used by stop_tail to determine if an array of EMS 
  messages were generated or not.

=cut

sub _patterns_present {
        my ( $patterns, $lines ) = @_;
        foreach my $pattern (@$patterns) {
            if($lines !~ m/<$pattern/s) {
                return 0;
            }
        }
        return 1;
}

=head2 to_string()

  Takes no arguments.  Returns a string representation of the log.
  
=cut

sub to_string {

	my ( $self ) = @_;
	
	my $string = "\nEMS Log:\n\n";
	for ( @{$self->{snippet}} ) {
		$string .= "$_";
	}
	
	$string .= "\n";
	
	return $string;
}

=head2 to_log()

  Takes no arguments.  Returns a copy of the log with each line
  terminated by a newline.
  
=cut

sub to_log {

	my ( $self ) = @_;
	
	my $string = "\nEMS Log:\n\n";
	for ( @{$self->{snippet}} ) {
		$string .= "$_\n";
	}
	
	return $string;
}

=head2 find_events()

  Takes an array of event names as arguments.  Returns a reference
  to an array of the error messages if there were matches or an
  empty array if there were no matches.
  
=cut

sub find_events {

	my ( $self, @events ) = @_;
	my ($evn, $i, $line, $lr, $msg, @msgs);

	croak 'No events passed in to search for' if (@events == 0);

	# Replace all '.' with '_' in event name
	for ( $i=0; $i<@events; $i++) {
		$events[$i] =~ s/\./_/g;
	}

	for ( $i=0; $i < @{$self->{snippet}}; $i++ ) {
		$line = ${$self->{snippet}}[$i];
		if ($line =~ /^<LR\s+/) {
			# Save start of record position
			$lr = $i;
			next;
		}
		# Skip end of log record lines
		next if ($line =~ /^<\/LR>/);
                if ($line =~ /^</) {
			foreach $evn ( @events ) {
				if ($line =~ /^<$evn/) {
					# Hit an event, turn it into a string
					$msg = ${$self->{snippet}}[$lr];
					do {
						$msg .= $line;
						$i++;
						$line = ${$self->{snippet}}[$i];
					} while (($i < @{$self->{snippet}}) &&
						 ($line !~ /^<\/LR>/));
					push @msgs, $msg;
				}
			}
		}
	}
	
	return \@msgs;
}

=head2 match_events()

    Match occurence of an event in the EMS log
    Parameters:  
        1. $self - EMS log reader handle
        2. Event pattern to match
        3. Error message if pattern not found	
        4. Field pattern to the match - optional
    Returns: 
        Event matched (True/False) or event string (string) 

=cut

sub match_event {
        my ($self, $pattern, $errmsg, $field) = @_;
        my $err_msgs = $self->find_events(($pattern));
	if(@{$err_msgs}) {
		foreach (@{$err_msgs}) {
			$Log->comment("Matched Pattern : $_\n");
			if (defined($field)) {
                           if ($_ =~ m/$field/) {
                              $Log->comment("Matched Field : $_\n"); 
                              return $_;
			   }
                           else {
                              $Log->comment("Error: Cannot match Field $field");
                              return 0;
                           }
	                }
                }
	}
	else {
		$Log->comment($errmsg);
		return 0;
	}
	return 1;
}

1;

=head1 DEPENDENCIES

  LogReader::Base.pm

=head1 AUTHOR

  dl-platform-qa@netapp.com
  
=cut

