package LogReader::Messages;

use LogReader::Base;
@ISA = qw( LogReader::Base );

use strict;

our $Path = '/etc/log/messages';
our $Path_10 = '/mroot/etc/messages';

=head1 NAME

  LogReader::Messages - Grab the Relevant Section of the messages Log(s)
  
=head1 DESCRIPTION

  This grabs the sections of the messages logs that were generated during
  the course of a test. Supports both 7 and 10-mode filers. For 7-mode and
  earlier filers the object grabs the logs with 'rdfile' issued to a Filer.
  For 10-mode or later filers it uses commands available via the Server.pm
  cli to get the needed information trying to avoid reading the whole
  log where possible.
  
=head1 SYNOPSIS

  use LogReader::Messages;

  # For 7-mode filers or earlier
  host($FILER);
  my $filer = connect($FILER_CONN);
  my $f = LogReader::Messages->new( $filer );
  
  # For 7-mode, or 10-mode, or later filers
  my $Server = Server->new();
  my $f = LogReader::Messages->new( $Server );

  # Start recording.
  $f->start;
  
  ### Conduct your test here.
  
  # Stop recording
  $f->stop;

=head1 METHODS

=head2 new()

  The constructor.  Takes takes a Filer Connectrec object or a Server object.

=head2 start()

  Takes no arguments.  Returns nothing.
  
  This begins recording the log(s).

=cut

sub start {
	
	my ( $self ) = @_;
	my ( $path );
	
        # Set the path based on the mode of the filer
	if ($self->{mode} < 10) {
		$path = $Path;
	} else {
		$path = $Path_10;
	}

	$self->{starttime}{time} = time();
	$self->{starttime}{localtime} = localtime( $self->{starttime}{time} );

        # For 7-mode still have to read the whole messages file to figure
	# out the last line.
	my $results = $self->rdfile( $path );

	# This sets the point after which we will be interested in the log.
	$self->{mark}{index} = $#$results;
	$self->{mark}{string} = $results->[$#$results];
}	

=head2 stop()

  Takes no arguments.  Returns the log index indicating where the reader
  stopped reading.
  
  This ends recording the log(s).
  
=cut

sub stop {

	my ( $self ) = @_;
        my ( $path );

	$self->{endtime}{time} = time();
	$self->{endtime}{localtime} = localtime( $self->{endtime}{time} );

	my @snippet = ();

        # Set the path based on the mode of the filer
	if ($self->{mode} < 10) {
		$path = $Path;
	} else {
		$path = $Path_10;
	}

	my $null = $self->rdfile( $path );
	
	$self->{mark}{index} = $LogReader::Index if defined $LogReader::Index;
	
	# This is the simplest case.  This means the mark is in the 'null' log (eg 
	# 'messages').  Therefore, the logs didn't rotate during test execution.  
	# Grab everything after the mark in the null log.
	

	if ( $self->{mark}{index} <= $#$null ) {
	
		unless ( $self->{mark}{index} == $#$null ) {
	 		@snippet = @$null[$self->{mark}{index}+1..$#$null];
		}
		
		$self->{current_log} = $path;
	
	# The mark is NOT in the 'null' log.  It must have been rotated during the test.  
	# Therefore grab everything after the mark in 'messages.0' and concatenate 
	# the entire 'null' log onto it.
	
	} else {
	
		my $zero = $self->rdfile( "$path.0" );
		
		unless ( $self->{mark}{index} == $#$zero ) {
			@snippet = @$zero[$self->{mark}{index}+1..$#$zero];
		}
		
		@snippet = ( @snippet, @$null );
		
		$self->{current_log} = "$path.0";
	}
	
	$self->{snippet} = \@snippet;
	
	return $#$null;
}

=head2 to_string()

  Takes no arguments.  Returns a string representation of the log.
  
=cut

sub to_string {

	my ( $self ) = @_;
	
	my $string = "\nMessages Log:\n\n";
	for ( @{$self->{snippet}} ) {
		$string .= "$_";
	}
	
	$string .= "\n";
	
	return $string;
}

=head2 to_log()

  Takes no arguments.  Returns a copy of the log with each line
  terminated by a newline.
  
=cut

sub to_log {

	my ( $self ) = @_;
	
	my $string = "\nMessages Log:\n\n";
	for ( @{$self->{snippet}} ) {
		$string .= "$_\n";
	}
	
	return $string;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  LogReader::Base.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut
















