# $Id$
#
## @summary LogReader MlogBcomd Class
##
## @status review
##
## @author talluri@netapp.com, dl-mhost-qa@netapp.com
##
## @pod here
##

=head1 NAME

LogReader::MlogBcomd - Grab the Relevant Section of the MROOT's bcomd.log files

=head1 DESCRIPTION

This function grabs the sections of the MROOT bcomd.log file. The bcomd.log file 
is rotated close to midnight (00:00) everyday. The bcomd.log contains the
most current log file, and the bcomd.log.previous.1 contains the log file
that is the immediate previous one.

This function only supports 10-mode filers, since there's no mhost concept
in 7-mode.

=head1 SYNOPSIS

use LogReader::MlogBcomd;

# For 10-mode filers, that contain mlog message files
my $Server = Server->new();
my $f = LogReader::MlogBcomd->new( $Server );

# Start recording.
$f->start;

### Conduct your test here.

# Stop recording
$f->stop;

=head1 METHODS

=head2 new()

The constructor.  Takes takes a Server object.

=head2 start()

Takes no arguments.  Returns nothing.

This begins recording the log(s).

=head2 stop()

Takes no arguments.  Returns nothing.

This stops recording the log(s).

=head1 DEPENDENCIES

LogReader::Base.pm

=cut

package LogReader::MlogBcomd;

use LogReader::Base;
use Tharn qw(logcomment);
use Carp qw(confess);
@ISA = qw( LogReader::Base );

use strict;

our $PATH_BCOMD = '/mroot/etc/mlog/bcomd.log';
our $PATH_BCOMD_PREVIOUS = '/mroot/etc/mlog/bcomd.log.previous.1';
our $LINES_LIMIT_WHEN_FILE_SHRINK = 200;

=head2 ::start

Start recording the messages in the mlog messages file, by means of remembering the position of the file at the moment when this function is being called.

=over

=item Arguments

B<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogBcomd->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    my $lines = $logrdr->matches ('wafl');
    if (@$lines) {
        my $report = join "\n", @$lines;
	logcomment ("Report from Mlog Messages :\n$report");
    }

=back

=cut

sub start {
    
    my ( $self ) = @_;
    
    # Set the path based on the mode of the filer
    if ($self->{mode} < 10) {
	confess "LogReader::MlogBcomd does not support filers prior to 10 Mode. You pass in a ".$self->{mode}." mode system.";
    }

    $self->{starttime}{time} = time();
    $self->{starttime}{localtime} = localtime( $self->{starttime}{time} );

    #
    # 20090311	tyu@netapp.com
    #
    # Remember the line number of the file at the moment when this function
    # is being called.
    #
    # Calling start () multiple times will just mean that user wants to
    # skip lines and jump to new location.
    #
    $self->{mark}{index} = $self->linecount ($PATH_BCOMD);
}


=head2 ::stop

Returns the log index indicating the position where the reader is at the moment
when this function is being called.

=over

=item Arguments

B<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogBcomd->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    my $lines = $logrdr->matches ('wafl');
    if (@$lines) {
        my $report = join "\n", @$lines;
	logcomment ("Report from Mlog Messages :\n$report");
    }

=back

=cut

sub stop {

    my ( $self ) = @_;

    $self->{endtime}{time} = time();
    $self->{endtime}{localtime} = localtime( $self->{endtime}{time} );

    my @snippet = ();

    # Set the path based on the mode of the filer
    if ($self->{mode} < 10) {
	confess "LogReader::MlogBcomd does not support filers prior to 10 Mode. You pass in a ".$self->{mode}." mode system.";
    }

    my $begin_linecount = $self->{mark}{index};

    my $new_linecount = $self->linecount ($PATH_BCOMD);

    if ($begin_linecount > $new_linecount) {
	#
	# 20090411	tyu@netapp.com
	#
	# Entering this segment of code if the log file had shrink. In this
	# case, the assumption is that the rotation must have happened.
	# 
	# And thus, this function will tried to capture 2 segments of text :
	#
	# 1. the portion of text, in bcomd.log.previous.1, from the "start"
	#    pointer to the end of the file.
	# 2. the whole file, bcomd.log.
	#
	# And then, it will advance the pointer to point to the end of file
	# bcomd.log
	# 

	my $prev_linecount = $self->linecount ($PATH_BCOMD_PREVIOUS);
	my $prev_lines = $self->rdfilelines ($PATH_BCOMD_PREVIOUS, $begin_linecount, $prev_linecount);
	@snippet = ( @snippet, @$prev_lines );
	$begin_linecount = 0;
    }

    my $lines = $self->rdfilelines ($PATH_BCOMD, $begin_linecount, $new_linecount);
    @snippet = ( @snippet, @$lines );
    $self->{snippet} = \@snippet;
    
    return $new_linecount;
}

=head2 to_string()

Takes no arguments.  Returns a string representation of the log.

=item Arguments

I<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogBcomd->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    logcomment ("Mlog string :\n".$logrdr->to_string ());

=cut

sub to_string {

    my ( $self ) = @_;
    
    my $string = "\nBcomd Log:\n\n";
    for ( @{$self->{snippet}} ) {
	$string .= "$_";
    }
    
    $string .= "\n";
    
    return $string;
}

=head2 to_log()

Takes no arguments.  Returns a copy of the log with each line
terminated by a newline.

=item Arguments

I<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogBcomd->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    my $report = $logrdr->to_log ();
    logcomment ("Report from Mlog Messages :\n$report");

=cut

sub to_log {

    my ( $self ) = @_;
    
    my $string = "\nBcomd Log:\n\n";
    for ( @{$self->{snippet}} ) {
	$string .= "$_\n";
    }
    
    return $string;
}

1;
