# $Id: //depot/prod/test/Rsierranevada/lib/LogReader/MlogMessages.pm#1 $
#
## @summary LogReader MlogMessages Class
##
## @status public
##
## @author tyu@netapp.com, dl-dpgqa-automation@netapp.com
##
## @pod here
##

=head1 NAME

LogReader::MlogMessages - Grab the Relevant Section of the MROOT's message log

=head1 DESCRIPTION

This grabs the sections of the MROOT messages log that were generated during
the course of a test.

This special log file contains various kinds of important messages,
while one type of these is regarding the result of WAFLIRON operation.

This function only supports 10-mode filers, since there's no mhost concept
in 7-mode.

=head1 SYNOPSIS

use LogReader::MlogMessages;

# For 10-mode filers, that contain mlog message files
my $Server = Server->new();
my $f = LogReader::MlogMessages->new( $Server );

# Start recording.
$f->start;

### Conduct your test here.

# Stop recording
$f->stop;

=head1 METHODS

=head2 new()

The constructor.  Takes takes a Server object.

=head2 start()

Takes no arguments.  Returns nothing.

This begins recording the log(s).

=head2 stop()

Takes no arguments.  Returns nothing.

This stops recording the log(s).

=head1 DEPENDENCIES

LogReader::Base.pm

=cut

package LogReader::MlogMessages;

use LogReader::Base;
use Tharn qw(logcomment);
use Carp qw(confess);
@ISA = qw( LogReader::Base );

use strict;

our $Path_10 = '/mroot/etc/mlog/messages.log';
our $LINES_LIMIT_WHEN_FILE_SHRINK = 200;

=head2 ::start

Start recording the messages in the mlog messages file, by means of remembering the position of the file at the moment when this function is being called.

=over

=item Arguments

B<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogMessages->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    my $lines = $logrdr->matches ('wafl');
    if (@$lines) {
        my $report = join "\n", @$lines;
	logcomment ("Report from Mlog Messages :\n$report");
    }

=back

=cut

sub start {
    
    my ( $self ) = @_;
    my ( $path );
    
    # Set the path based on the mode of the filer
    if ($self->{mode} < 10) {
	confess "LogReader::MlogMessages does not support filers prior to 10 Mode. You pass in a ".$self->{mode}." mode system.";
    }

    $path = $Path_10;

    $self->{starttime}{time} = time();
    $self->{starttime}{localtime} = localtime( $self->{starttime}{time} );

    #
    # 20090311	tyu@netapp.com
    #
    # Remember the line number of the file at the moment when this function
    # is being called.
    #
    # Calling start () multiple times will just mean that user wants to
    # skip lines and jump to new location.
    #
    $self->{mark}{index} = $self->linecount ($path);
}


=head2 ::stop

Returns the log index indicating the position where the reader is at the moment
when this function is being called.

=over

=item Arguments

B<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogMessages->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    my $lines = $logrdr->matches ('wafl');
    if (@$lines) {
        my $report = join "\n", @$lines;
	logcomment ("Report from Mlog Messages :\n$report");
    }

=back

=cut

sub stop {

    my ( $self ) = @_;
    my ( $path );

    $self->{endtime}{time} = time();
    $self->{endtime}{localtime} = localtime( $self->{endtime}{time} );

    my @snippet = ();

    # Set the path based on the mode of the filer
    if ($self->{mode} < 10) {
	confess "LogReader::MlogMessages does not support filers prior to 10 Mode. You pass in a ".$self->{mode}." mode system.";
    }

    $path = $Path_10;

    my $begin_linecount = $self->{mark}{index};

    my $new_linecount = $self->linecount ($path);

    if ($begin_linecount > $new_linecount) {
	#
	# 20090311	tyu@netapp.com
	#
	# Entering this segment of code if the log file had unexpectedly shrink.
	#
	# Note that the maintanence mechanism (i.e. regarding how this file
	# grow, or shrink over time) of this file is unknown as of the time
	# when this function is being implemented. The assumption taken here
	# is that the file will keep growing over the lifetime of the test
	# that is calling this function, and there is no rotating mechanism
	# identified.
	#
	# The resolution taken here is that the function will return either
	# the last number of lines (defined by $LINES_LIMIT_WHEN_FILE_SHRINK),
	# or the whole file if the number of lines of the file is less than
	# the number sought.
	#
	if ($new_linecount > $LINES_LIMIT_WHEN_FILE_SHRINK) {
	    $begin_linecount = $new_linecount - $LINES_LIMIT_WHEN_FILE_SHRINK;
	} else {
	    $begin_linecount = 0;
	}
    }

    my $lines = $self->rdfilelines ($path, $begin_linecount, $new_linecount);
    $self->{snippet} = \@$lines;
    
    return $new_linecount;
}

=head2 to_string()

Takes no arguments.  Returns a string representation of the log.

=item Arguments

I<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogMessages->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    logcomment ("Mlog string :\n".$logrdr->to_string ());

=cut

sub to_string {

    my ( $self ) = @_;
    
    my $string = "\nMessages Log:\n\n";
    for ( @{$self->{snippet}} ) {
	$string .= "$_";
    }
    
    $string .= "\n";
    
    return $string;
}

=head2 to_log()

Takes no arguments.  Returns a copy of the log with each line
terminated by a newline.

=item Arguments

I<None>

=item Examples

    my $server = Server->new ("dpg-gx23");
    my $logrdr = LogReader::MlogMessages->new ($server);

    my $logrdr->start ();

    #
    # Do something ...
    #

    $logrdr->stop ();

    my $report = $logrdr->to_log ();
    logcomment ("Report from Mlog Messages :\n$report");

=cut

sub to_log {

    my ( $self ) = @_;
    
    my $string = "\nMessages Log:\n\n";
    for ( @{$self->{snippet}} ) {
	$string .= "$_\n";
    }
    
    return $string;
}

1;
