package LogReader::WORMLog;

use LogReader::Base;
@ISA = qw( LogReader::Base );

use Data::Dumper;
use SnapLock::WORMLogging qw(get_worm_log_file_path get_previous_log_file is_exists_log_file);
use strict;
use NOTE;


### WORM log path is not standard.
our $Path = '';

=head1 NAME

  LogReader::WORMLog - Grab the Relevant Section of the WORM Log(s)
  
=head1 DESCRIPTION

  This grabs the sections of the WORM logs(priv_delete & system_log) that were generated during
  the course of a test.  The object grabs the logs with 'rdfile' issued
  to a Filer via non-NATE rsh. 
  
=head1 SYNOPSIS

  use LogReader::WORMLog;

  host($FILER);
  my $filer = connect($FILER_CONN);
  
  my $f = LogReader::WORMLog->new( $filer );
  
  # Start recording.
  $f->start($file_path);
  
  ### Conduct your test here.
  
  # Stop recording
  $f->stop($file_path, $vol_obj);

=head1 METHODS

=head2 new()

  The constructor.  Takes takes a Filer Connectrec object.

=head2 start()

  Takes WORM log file path as arguments.  Returns nothing.
  
  This begins recording the log(s).

=cut

sub start($) {
	
    my ( $self , $Path) = @_;
	
    $self->{starttime}{time} = time();
    $self->{starttime}{localtime} = localtime( $self->{starttime}{time} );

    NOTE::logcomment("LOG READER STARTED READING LOG FILE");
    my $results = $self->rdfile( $Path );

    # This sets the point after which we will be interested in the log.
	
    $self->{mark}{index} = $#$results;
    $self->{mark}{string} = $results->[$#$results];
 
}	

=head2 stop($$)

  Takes  WORM Log file path and NOTE volume object arguments.  Returns the log index indicating where the reader
  stopped reading.
  
  This ends recording the log(s).
  
=cut

sub stop ($$){

    my ( $self,$Path ,$Vol_obj) = @_;

    my $filer_conn = $Vol_obj->Filer->conn;
    my $volume	= $Vol_obj->name;
    my $null;
    my $new_path;

    $self->{endtime}{time} = time();
    $self->{endtime}{localtime} = localtime( $self->{endtime}{time} );

    my @snippet = ();

    NOTE::logcomment("IN LOG READER STOP");

    my $check_path = is_exists_log_file($Path, $Vol_obj);
    if ( $check_path eq 1) {
	$null = $self->rdfile( $Path );
    } elsif ( $check_path ne 1 && $check_path =~ /priv_delete/ ) {
	my $Old_Path = get_previous_log_file($check_path, $Vol_obj, "priv_delete");
	$null = $self->rdfile( $Old_Path );
    } elsif ( $check_path ne 1 && $check_path =~ /system_log/ ) {
        NOTE::logcomment("My new System log file path = $check_path\n");
    }
   
    $self->{mark}{index} = $LogReader::Index if defined $LogReader::Index;
	
    # This is the simplest case.  This means the mark is in the 'null' log (eg 
    # 'messages').  Therefore, the logs didn't rotate during test execution.  
    # Grab everything after the mark in the null log.

    if ( $self->{mark}{index} <= $#$null ) {
	
   	unless ( $self->{mark}{index} == $#$null ) {
 		@snippet = @$null[$self->{mark}{index}+1..$#$null];
	}
    }

  

    # The mark is NOT in the 'null' log.  It must have been rotated during the test.  
    # Therefore grab everything after the mark in 'messages.0' and concatenate 
    # the entire 'null' log onto it.
	
    if ( $check_path ne 1 ){

 	NOTE::logcomment("Logfile has rotated/changed");
	my $basename;
	if ( $Path =~ /priv_delete/ ) {
	    $basename = "priv_delete";
	    $new_path = get_worm_log_file_path($filer_conn,$volume, $basename);
	    die "Failed to get Priv delete worm log file path\n" if ( !$new_path );
	} elsif ( $Path =~ /system_log/ ){
	    $basename = "system_log";
	    $new_path = get_worm_log_file_path( $filer_conn, $volume , $basename);
	    die "Failed to get System Worm log file path\n" if ( !$new_path );
	} 
	my $zero = $self->rdfile( $new_path );
	@snippet = ( @snippet, @$zero );
	$self->{current_log} = $new_path;
    }
    $self->{snippet} = \@snippet;
	
    return $#$null;
}

=head2 to_string()

  Takes no arguments.  Returns a string representation of the log.
  
=cut

sub to_string {

	my ( $self ) = @_;
	my $string;
	for ( @{$self->{snippet}} ) {
                ## Including two more conditionds, which got added from RR.
		if ( $_ =~/LogFileFooter LogFileCloseDate/i || $_ =~ /SnapLockLogFile/i || $_ =~ /LogFileHeader Date/i || $_ =~ /LogFileFooter LogVersion/i || $_ =~ /LogFileHeader LogVersion/ ) {
			## Do not add this line to string
		} else {
			$string .= "$_";
		}
	}
	return $string;
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  LogReader::Base.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  dl-snaplock-qa@netapp.com
  
=cut
















