
## @summary Library to map and unmap shares
## @author  dl-central-libs
## @pod here

=head1 NAME

MapShare - Mapping and Unmapping of a share

=head1 SYNOPSIS

    use MapShare qw(:all);
    eval{
    my $drive = map_cifs_share(server=>'tada-u12' , share=>'share1'
			       [,client => 'tada-c30']
			       [,domain=>'ad']
			       [,username=>'administrator']
			       [,password=>'sundance']
			       [,drive=>'*']);
    };
    croak($Log->error($@)) if($@);
    unmap_cifs_share(
		 mappath =>'E:',
                 drive => 'E:',(alias)
		 client => 'tada-c30'
		 )or $Log->warn('warn',"$@");

    my $map_found =
         map_exists(
                     drive => 'E:'
	   	     client => 'tada-c30'
		 );

    my @maps =
         list_maps(
	   	     client => 'tada-c30'
		 );

    unmap_cifs_shares_for_server(
		 server => 'tada-u12',
		 client => 'tada-c30'
		 )or $Log->warn('warn',"$@");

=head1 ROUTINES

=cut

package MapShare;
use strict;
use Params::Validate v0.70 qw(validate SCALAR HASHREF);
use Utils qw(make_conn);
use Carp qw(croak);
use Tharn qw(say $Log);
use base qw(Exporter);
BEGIN{
our %EXPORT_TAGS = (
    all => [ qw(map_cifs_share
                unmap_cifs_share
                map_exists
		unmap_cifs_shares_for_server
		list_maps )
           ] );
  Exporter::export_ok_tags('all');

}

=head2 map_cifs_share(%opts)

=over

=item Arguments

=over

=item share

(Required) Name of the share that needs to be mapped.
If param 'server' is set, please do not pass embedded server param
for 'share'( pass only share name ex: share1)
If param 'server' is NOT set, you'll have to pass the server in the 
'share' param ( ex: \\\\server\\share1)

=item client

(Optional) The client connection spec  on which the share is mapped, 
defaults to localhost

=item server

(Optional) Server's  DNS name, or the IP of the server.
Note:This is optional, if the param 'share' has an embedded 
server(ex: \\\\server\\share1)


=item workgroup

(Optional) The workgroup of which the filer(server) would be part of.
Defaults to "WORKGROUP".

=item domain

(Optional) Name of the domain where the username exists.

=item username

(Optional) The username on the domain/workgoup which has access 
to the share 

=item password

(Optional) The password of the "username". 

=item drive

(Optional) If not specified, it'll be a non persistent connection 
should be a letter in the range  (A-Z|a-z) or * for 
mapping the share to next available drive letter.

=item timeout

(Optional) Defaults to 100 seconds

=back

=item Examples

    eval{
        my $drive = map_cifs_share(server=>'tada-u12'
			       [,domain=>'ad']
			       [,username=>'administrator']
			       [,password=>'sundance']
			       [,share=>'share1']
			       [,drive=>'*']);
    } 
    croak($Log->error($@)); if($@);
         
Please note that only one of "workgroup" or "domain" params make sense and
should be specified. If  none of them is specified, we will try to map the share 
with the workgroup defaults.In case both are specified this routine would print a 
warning and will try to map the share with domain authentication.

=item Dependencies

rshd should be running on the client, if not setup rshd using min_win 
Refer http://web.netapp.com/engineering/hhg/Software-4.0/Tools-7.0/NATE-9.1/User_Guides-2.1/MinWin_Windows_Clients.html
cifs should be licenced and should be setup on the filer.

=item Notes

if argument 'share'(ex: \\\\server\\share) has embedded server in it, please 
do not pass 'server' argument.

=back

=for html 
<hr> </hr>

=cut

sub map_cifs_share {
    
    my (%args) = validate (@_, {
				server       => { type => SCALAR, optional => 1 , reqex => qr/\w+|\d+.\d+.\d+.\d+/, default=>undef },
				client       => { type => SCALAR|HASHREF ,default =>'localhost'},
				share        => { type => SCALAR },
				workgroup    => { type => SCALAR, optional => 1 },
				domain       => { type => SCALAR, optional => 1 },
				username     => { type => SCALAR, optional => 1},
				password     => { type => SCALAR, optional => 1 },
				drive        => { type => SCALAR, optional => 1},
				timeout      => { type => SCALAR, optional => 1, default => 100 },
				verbose      => { type => SCALAR, optional => 1, default => 1 }
			       });

    return eval{
	my $cc = make_conn($args{client}) or croak($@);
	$cc->timeout($args{timeout}) if ($args{timeout} > $cc->timeout());
	my $share_unc;
	# If both workgroup and domain are specified we use authentication via the domain.
	if ($args{workgroup} && $args{domain}) {
	    $Log->warn("Both workgroup and domain are given, Ignoring workgroup and setting as domain.");
	    delete($args{workgroup});
	}
	# Add ":" if it is a drive letter and ':' is not there.
	if ( $args{drive} =~ /^[A-Za-z]$/ ) {
	    $args{drive} = $args{drive}.":";
	}
	if (defined($args{server})){
	    my $first_two_letters = substr($args{share}, 0, 2);
	    if ($first_two_letters eq '\\\\') {
		$Log->warn("You have given server as \'$args{server}\' and the share" .
			  " as \'$args{share}\', this might not work, you should set only one of them arg");
	    }
	    #If server is defined,we make a unc path
	    $share_unc = $args{drive} . " \\\\" . $args{server} . "\\" . $args{share};
	}else{ #if server is undef, we assume share has embedded server in the path
	    $share_unc = $args{drive} . " " . $args{share};
	}
	my $make_share_path = "$share_unc";
	my $group = $args{domain} || $args{workgroup};
	if (defined ($args{password})) {
	    $make_share_path .= " \"$args{password}\"";
	}
	if ($args{username}) {
	    $make_share_path .=  $group ? " /user:$group\\$args{username}" : " /user:$args{username}";
	}
	
	my $net_use_cmd = "cmd /c net use $make_share_path";
         $Log->comment("$net_use_cmd");	
         my $verbose_flag = '-verbose';
        if (!$args{'verbose'}) { $verbose_flag = '-noverbose'; }
	#my (undef,undef,$after) = say($cc,$net_use_cmd,'-timeout',$args{timeout},$verbose_flag);
        my $after = $cc->execute(command => "$net_use_cmd");
        #my $net_use_cmd = 'net use * \\\10.226.114.81\CIFS_SHARE netapp*123*123 /user:ns\administrator';
        #my $after = $cc->execute(command => 'net use * \\\10.226.114.81\\CIFS_SHARE netapp*123*123 /user:ns\\administrator');
	if($after =~ /command\s+completed\s+successfully|now\s+connected/){
	    if (!$args{drive}){
		return($share_unc);
	    }else{
		#If we use *, we ask windows to chose the drive for us
		#extract the drive and send it to the calling routine
		if($args{drive} =~ /^\s*\*\s*$/ ){
		    if($after =~ /.*Drive\s+(\w:).*/){
			return($1);
		    }else{
			croak("$after\nnet_use_cmd (map_cifs_share):\n$net_use_cmd");
		    }
		}else{
		    return($args{drive});
		}
	    }
	}
        if (!$args{'verbose'}) { $Log->comment("$net_use_cmd\n $after"); }
	croak("$after\nnet_use_cmd (map_cifs_share):\n$net_use_cmd");
    };
}


=head2 unmap_cifs_share(%opts)

=over

=item Arguments

=over

=item client

(Optional) The client  connspec on which the share is mapped, defaults to 'localhost'

=item mappath

(Optional) This can be a drive letter(local) or a non persistent connection(remote, unc path)

=item drive

(Optional) This will be a drive letter(local). This is an alias,since you can pass a drive letter 
by setting 'mappath' to the drive letter,this is done so that it goes with map_cifs_share.

=item timeout

(Optional) Defaults to 100 seconds

=back

=item NOTE

One of the two params 'mappath'|'drive' is required
If both are set, we will use 'drive' to unmap the share

=item Examples

    unmap_cifs_share(
	   	 mappath => '\\172.16.70.112\share1'
		 client => 'tada-c30'

		 );
    unmap_cifs_share(
		 drive =>'E:'
		 client => 'tada-c30'
		 );		 

    unmap_cifs_share(
		 mappath =>'E:'
		 client => 'tada-c30'
		 );		 

=back

=cut
 
sub unmap_cifs_share {

    my (%args) = validate (@_, {
				

				mappath      => { type => SCALAR, optional => 1 },
				drive        => { type => SCALAR, optional => 1, regex => qr/^[*A-Za-z]:?$/ },
				client       => { type => SCALAR|HASHREF, optional => 1, default => 'localhost'},
				timeout      => { type => SCALAR, optional => 1, default => 100 },
				force        => { type => SCALAR, optional => 1, default => 0 },
				verbose      => { type => SCALAR, optional => 1, default => 1 }
				
			       });
    return eval{
	my $cc = make_conn($args{client}) or croak($@);
	$cc->timeout($args{timeout}) if ($args{timeout} > $cc->timeout());
	# Add ":" if it is a drive letter and ':' is not there.
	if(!$args{mappath} && !$args{drive}){
	    croak("One of two params \'mappath\' or \'drive\' is required");
	}elsif($args{drive} && $args{mappath}){
	    delete($args{mappath});	     
	}
	my $mappath = $args{mappath} || $args{drive};
	if($mappath =~ /^[A-Za-z]$/ ) {
	    $mappath = $mappath . ":";
	}elsif( $mappath eq '*'){
	    $Log->warn("Since specifying * will delete all shares, this routine will not handle *");
	}
	my $verbose_flag = '-verbose';
        if (!$args{'verbose'}) { $verbose_flag = '-noverbose'; }
	my $cmd = "cmd /c net use $mappath /delete";
	my $output2;
	my (undef,undef,$output) = say($cc, $cmd, "-timeout", $args{timeout}, $verbose_flag);
	if ( $output =~ /deleted\s+successfully/ ) {
	    return(1);
	} elsif ($output =~ /force them closed\?/ && $args{force}) {
	    (undef, undef, $output2) = say($cc, "Y", $verbose_flag);
	    if ($output2 =~ /deleted\s+successfully/) {
                return(1);
            }
	}
        if (!$args{'verbose'}) { $Log->comment("$cmd\n $output\n $output2"); }
        croak($output);
    };
}
  

=head2 map_exists(%opts)

=over

=item Arguments

=over

=item client

(Optional) The client  connspec on which the share is mapped, defaults to 'localhost'

=item drive

(Optional) This will be a drive letter(local). This is an alias,since you can pass a drive letter 
by setting 'mappath' to the drive letter,this is done so that it goes with map_cifs_share.

=item timeout

(Optional) Defaults to 100 seconds

=back

=item NOTE

=item Examples

    map_exists(
		 drive =>'E:'
		 client => 'tada-c30'
		 );		 

=back

=cut
 
sub map_exists {

    my (%args) = validate (@_, {
				drive        => { type => SCALAR, optional => 0, regex => qr/^[A-Za-z]:?$/ },
				client       => { type => SCALAR|HASHREF, optional => 1, default => 'localhost'},
				timeout      => { type => SCALAR, optional => 1, default => 100 },
				verbose      => { type => SCALAR, optional => 1, default => 1 }
				
			       });
    return eval{
	my $cc = make_conn($args{client}) or croak($@);
	$cc->timeout($args{timeout}) if ($args{timeout} > $cc->timeout());
	# Add ":" if it is a drive letter and ':' is not there.
	if(!$args{drive}){
	    croak("\'drive\' param is required");
	}
	my $drive = $args{drive};
	if($drive =~ /^[A-Za-z]$/ ) {
	    $drive .= ":";
	}elsif( $drive eq '*'){
	    croak("'wildcard' drive value not allowed with 'map_exists'");
	}
	my $verbose_flag = '-verbose';
        if (!$args{'verbose'}) { $verbose_flag = '-noverbose'; }
	my $cmd = "cmd /c net use $drive";
	my (undef,undef,$output) = say($cc, $cmd, "-timeout", $args{timeout}, $verbose_flag);
        if ($output =~ /completed\s+successfully/) {
            return (1);
        } elsif ($output =~ /network connection could not be found/) {
            if (!$args{'verbose'}) { $Log->comment("$cmd\n $output"); }
            return (0);
        } else {
            if (!$args{'verbose'}) { $Log->comment("$cmd\n $output"); }
            croak($output);
        }
    }
}


=head2 list_maps(%args)

=over

=item Arguments

=over

=item client

(Optional) The client connspec on which the shar(s) are mapped, defaults to 'localhost'

=item timeout

(Optional) Defaults to 100 seconds

=back

=item Examples

    list_maps(
		 client => 'tada-c30'
		 );		 

=back

=cut
 
sub list_maps {

    my (%args) = validate (@_, {

				client       => { type => SCALAR|HASHREF, optional => 1, default => 'localhost'},
				timeout      => { type => SCALAR, optional => 1, default => 100 },
				verbose      => { type => SCALAR, optional => 1, default => 1 }

			       });
    return eval{
	my $cc = make_conn($args{client}) or croak($@);
	$cc->timeout($args{timeout}) if ($args{timeout} > $cc->timeout());
	my $verbose_flag = '-verbose';
        if (!$args{'verbose'}) { $verbose_flag = '-noverbose'; }
	my $cmd = "cmd /c net use";
	my (undef,undef,$output) = say($cc, $cmd, "-timeout", $args{timeout}, $verbose_flag);
        if (!$args{'verbose'} && $output !~ /command completed successfully/) {
            $Log->comment("$cmd\n $output");
        }

	my @lines = split(/\n/, $output);
	my @maps = ();
	foreach my $line (@lines) {
		my ($rest, $path, $server, $share) = ($line =~ /(.*)\s+(\\\\([^\\]+)\\(\S+))/);
		next unless ($path);
		my ($status, $drive) = split(/\s+/, $rest);
		my $map = { server => $server, share => $share, drive => $drive, path => $path };
		$map->{status} = $status if ($status);
		push (@maps, $map);
	}
	return(@maps);
    };
}


=head2 unmap_cifs_shares_for_server(%args)

=over

=item Arguments

=over

=item client

(Optional) The client  connspec on which the share(s) are mapped, defaults to 'localhost'

=item Server

This is the name or IP address of the host which is serving the share(s).

=item timeout

(Optional) Defaults to 100 seconds

=back

=item Examples

    unmap_cifs_share(
	   	 server => '172.16.70.112'
		 client => 'tada-c30'
		 );

    unmap_cifs_share(
		 server =>'tada-u123'
		 client => 'tada-c30'
		 );		 

=back

=cut
 
sub unmap_cifs_shares_for_server {

    my (%args) = validate (@_, {
				server        => { type => SCALAR, optional => 0 },
				client       => { type => SCALAR|HASHREF, optional => 1, default => 'localhost'},
				timeout      => { type => SCALAR, optional => 1, default => 100 },
				force        => { type => SCALAR, optional => 1, default => 0 }

			       });
	my @maps = list_maps(client => $args{client});
	foreach my $map (@maps) {
		if ($map->{server} eq $args{server}) {
			if ($map->{drive}) {
				unmap_cifs_share(
		   	 		drive => $map->{drive},
					client => $args{client},
					timeout => $args{timeout},
					force => $args{force})
				    or $Log->warn("unmap_cifs_share failed: $@");
			} else {
				unmap_cifs_share(
		   	 		mappath => $map->{path},
					client => $args{client},
					timeout => $args{timeout},
					force => $args{force})
				    or $Log->warn("unmap_cifs_share failed: $@");
			}
		}
	}
}

1;

