#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/MdbTest.pm $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
## @summary Library of common subroutines used in all the MDB testcases.
## @author Jitendra Sasmal (jitendra.sasmal@netapp.com)
## @status public
## @pod here

package MdbTest;
require(Exporter);
@ISA = qw(Exporter);
our @EXPORT = qw(aggr_row_create aggr_row_delete_bykey aggr_row_delete_byname
        aggr_row_delete_bybladeid aggr_row_delete_bymemberid
        aggr_row_delete_bysnmpid aggr_show aggr_modify aggr_table_cleanup
        BYKEY BYNAME BYBLADEID BYMEMBERID BYSNMPID);
use warnings;
use strict qw(vars subs refs);
use Switch;
use Data::Dumper;
use Params::Validate     qw(validate validate_pos SCALAR ARRAYREF
                            SCALARREF OBJECT HASHREF CODEREF);

=head1 NAME
MdbTest - Library of common subroutines used in MDB testcases.

=head1 SYNOPSIS
This document is primarily intended for whoever wants to add or maintain MDB
testcases. Please refer the .thpl scripts under test/main/mhost/mdb to
know how to use these subroutines.

=head1 DESCRIPTION
This module defines routines used in the MDB testcases to create/delete/modify
rows in MDB aggr table. The CLIs as well as the aggr table are purely test level
i.e. they will be available only in test priviledge. So, don't expect
these subroutines to be functional in promoted builds as these builds don't
support test priviledge.

=head1 FUNCTIONS
All functions, unless otherwise stated, return 1 on success, and undef on
failure.

=cut

use constant BYKEY => 0;
use constant BYNAME => 1;
use constant BYBLADEID => 2;
use constant BYMEMBERID => 3;
use constant BYSNMPID => 4;


=head2 aggr_row_create(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    create_type => $create_type,
    id => $id, name => $name, 'blade-id' => $blade_id,
    nickname => $nickname, voltype => $voltype,
    'member-id' => $member_id, 'snmp-id' => $snmp_id,
    comment => $comment)

Creates a row in mdb aggr table.

This subroutine takes the TCD Object, Server Object, node name, row creation
type(BYKEY, BYNAME, BYBLADEID, BYMEMBERID, BYSNMPID), and values for all the
mandatory fields and/or the optional fields(nickname, comment) in mdb aggr table.

B<Usage:> C<
        $blade_id = $member_id = $snmp_id = $id = 0;
        $name = "aggr_$id";
        $nickname = "aggr";
        $voltype = "striped";
        $comment = $name;

        my $ret_status = aggr_row_create(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            create_type => BYNAME,
            id => $id, name => $name, 'blade-id' => $blade_id,
            nickname => $nickname,
            voltype => $voltype,
            'member-id' => $member_id,
            'snmp-id' => $snmp_id,
            comment => $comment);>

B<Returns:> C<$ret_status> 1 if success, undef if fails
=cut

sub aggr_row_create {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
            create_type =>1,
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            create_type     => {type => SCALAR},
            id              => {type => SCALAR},
            name            => {type => SCALAR},
            nickname        => {type => SCALAR, optional => 1},
            'blade-id'      => {type => SCALAR},
            'member-id'     => {type => SCALAR},
            'snmp-id'       => {type => SCALAR},
            comment         => {type => SCALAR, optional => 1},
        );


    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node
    my $create_type = $options{'create_type'}; #row creation by primary/byname
                                               #byblade-idetc

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    eval {
	    switch ($create_type) {
	        case BYKEY {
                $server->unit_test_mdb_aggr_create(%input_arg);
	        }
	        case BYNAME {
                $server->unit_test_mdb_aggr_by_name_create(%input_arg);
	        }
	        case BYBLADEID {
                $server->unit_test_mdb_aggr_by_bladeid_create(%input_arg);
	        }
	        case BYMEMBERID {
                $server->unit_test_mdb_aggr_by_memberid_create(%input_arg);
	        }
	        case BYSNMPID {
                $server->unit_test_mdb_aggr_by_snmpid_create(%input_arg);
	        }
	        else {
                return undef;
            }
	    }

    };

    if($@) {
        $tcd->nlog("aggr_row_create:Row creation failed, ERROR:$@");
        return undef;
    }


    my @mdb_aggr_row = ();
    #Check in DB that the row is added
    eval {
        @mdb_aggr_row = $server->unit_test_mdb_aggr_show(id => $options{'id'});
    };

    if($@) {
        $tcd->nlog("aggr_row_create:mdb-test mdb aggr show failed, ERROR:$@");
        return undef;
    } else {
        if (($mdb_aggr_row[0]->{id} != $options{'id'}) ||
            ($mdb_aggr_row[0]->{name} ne $options{'name'}) ||
            ($mdb_aggr_row[0]->{'blade-id'} != $options{'blade-id'}) ||
            ($mdb_aggr_row[0]->{'member-id'} != $options{'member-id'}) ||
            ($mdb_aggr_row[0]->{'snmp-id'} != $options{'snmp-id'}) ||
            ($mdb_aggr_row[0]->{comment} ne $options{'comment'})) {
                $tcd->nlog("aggr_row_create:values in db doesn't match with what ".
                            "entered");
                return undef;
        }
    }

    return 1;
}

=head2 aggr_row_delete_bykey(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    id => $id)

Deletes a row in mdb aggr table for the specified row id.

This subroutine takes the TCD Object, Server Object, node name, row id
to delete the row corresponding to the row id.

B<Usage:> C<
        my $id = 0;
        my $ret_status = aggr_row_delete_bykey(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            id => $id);>

B<Returns:> C<$ret_status> 1 if success, undef if fails
=cut

sub aggr_row_delete_bykey {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            id              => {type => SCALAR},
        );


    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    $tcd->nlog("deleting row by primary key, id = $options{'id'}");
    eval {
                $server->unit_test_mdb_aggr_delete(%input_arg);
    };

    if($@) {
        $tcd->nlog("aggr_row_delete_bykey failed, ERROR:$@");
        return undef;
    }


    my @mdb_aggr_row = ();
    #Check in DB that the row is deleted
    eval {
        @mdb_aggr_row = $server->unit_test_mdb_aggr_show(id => $options{'id'});
    };

    if ( ($@) && ($@ =~ /no\s+entr/)) {
        return 1;
    } else {
        return undef;
    }

}

=head2 aggr_row_delete_byname(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    name => $aggrname)

Deletes a row in mdb aggr by-name table for the specified aggregate name.

This subroutine takes the TCD Object, Server Object, node name, aggregate name
to delete the corresponding row.

B<Usage:> C<
        my $name = "aggr0";
        my $ret_status = aggr_row_delete_byname(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            name => $name);>

B<Returns:> C<$ret_status> 1 if success, undef if fails
=cut

sub aggr_row_delete_byname {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            name            => {type => SCALAR},
        );


    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    $tcd->nlog("deleting row by name, name = $options{'name'}");
    eval {
                $server->unit_test_mdb_aggr_by_name_delete(%input_arg);
    };

    if($@) {
        $tcd->nlog("aggr_row_delete_byname failed, ERROR:$@");
        return undef;
    }


    my @mdb_aggr_row = ();
    #Check in DB that the row is deleted
    eval {
        @mdb_aggr_row = $server->unit_test_mdb_aggr_by_name_show(name => $options{'name'});
    };

    if ( ($@) && ($@ =~ /no\s+entr/)) {
        return 1;
    } else {
        return undef;
    }

}

=head2 aggr_row_delete_bybladeid(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    id => $id,
    'blade-id' => $bladeid)

Deletes a row in mdb aggr by-bladeid table for the specified combination of row id
and blade-id.

This subroutine takes the TCD Object, Server Object, node name, row id and
blade-id to delete the corresponding row.

B<Usage:> C<
        my $id  = 1;
        my $bladeid= 1;
        my $ret_status = aggr_row_delete_bybladeid(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            id => $id,
            'blade-id'=> $bladeid);>

B<Returns:> C<$ret_status> 1 if success, undef if fails
=cut

sub aggr_row_delete_bybladeid {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            'blade-id'      => {type => SCALAR},
            id              => {type => SCALAR},
        );


    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    $tcd->nlog("deleting row by blade-id,  blade-id = $options{'blade-id'}");
    eval {
                $server->unit_test_mdb_aggr_by_bladeid_delete(%input_arg);
    };

    if($@) {
        $tcd->nlog("aggr_row_delete_bybladeid failed, ERROR:$@");
        return undef;
    }


    my @mdb_aggr_row = ();
    #Check in DB that the row is deleted
    eval {
        @mdb_aggr_row = $server->unit_test_mdb_aggr_by_bladeid_show('blade-id' =>
                                                            $options{'blade-id'});
    };

    if ( ($@) && ($@ =~ /no\s+entr/)) {
        return 1;
    } else {
        return undef;
    }

}

=head2 aggr_row_delete_bymemberid(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    name => $name,
    'member-id' => $memberid)

Deletes a row in mdb aggr by-memberid table for the specified combination of
aggregate name and member-id.

This subroutine takes the TCD Object, Server Object, node name, aggregate name and
member-id to delete the corresponding row.

B<Usage:> C<
        my $name = "aggr0";
        my $memberid = 0;
        my $ret_status = aggr_row_delete_bymemberid(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            name => $name,
            'member-id'=> $memberid);>

B<Returns:> C<$ret_status> 1 if success, undef if fails
=cut

sub aggr_row_delete_bymemberid {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            'member-id'     => {type => SCALAR},
            name            => {type => SCALAR},
        );


    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }


    $tcd->nlog("deleting row by member-id,  member-id = $options{'member-id'}");
    eval {
                $server->unit_test_mdb_aggr_by_memberid_delete(%input_arg);
    };

    if($@) {
        $tcd->nlog("aggr_row_delete_bymemberid failed, ERROR:$@");
        return undef;
    }


    my @mdb_aggr_row = ();
    #Check in DB that the row is deleted
    eval {
        @mdb_aggr_row = $server->unit_test_mdb_aggr_by_memberid_show('member-id' =>
                                                            $options{'member-id'});
    };

    if ( ($@) && ($@ =~ /no\s+entr/)) {
        return 1;
    } else {
        return undef;
    }

}

=head2 aggr_row_delete_bysnmpid(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    'snmp-id' => $snmpid)

Deletes a row in mdb aggr by-snmpid table for the specified snmp id.

This subroutine takes the TCD Object, Server Object, node name, snmp-id 
to delete the corresponding row.

B<Usage:> C<
        my $snmpid = 0;
        my $ret_status = aggr_row_delete_bysnmpid(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            'snmp-id'=> $snmpid);>

B<Returns:> C<$ret_status> 1 if success, undef if fails
=cut

sub aggr_row_delete_bysnmpid {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            'snmp-id'       => {type => SCALAR},
        );


    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }


    $tcd->nlog("deleting row by snmp-id,  snmp-id = $options{'snmp-id'}");
    eval {
                $server->unit_test_mdb_aggr_by_snmpid_delete(%input_arg);
    };

    if($@) {
        $tcd->nlog("aggr_row_delete_bysnmpid failed, ERROR:$@");
        return undef;
    }


    my @mdb_aggr_row = ();
    #Check in DB that the row is deleted
    eval {
        @mdb_aggr_row = $server->unit_test_mdb_aggr_by_snmpid_show('snmp-id' =>
                                                            $options{'snmp-id'});
    };

    if ( ($@) && ($@ =~ /no\s+entr/)) {
        return 1;
    } else {
        return undef;
    }

}

=head2 aggr_show(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    show_type = > $showtype,
    id => $id, name => $name,
    'blade-id' => $bladeid,
    voltype => $voltype,
    'member-id' => $memberid,
    'snmp=id' => $snmpid,
    comment => $comment)

Returns a reference to array of row(s) for the corresponding match of id, name etc.
show_type can be BYKEY, BYNAME, BYBLADEID, BYMEBMERID or BYSNMPID.
id, name, voltype, blade-id, member-id, snmp-id, comment are all optional
fields. If provided, any one or combination of these fields, aggr_show() will
return reference to the corresponding row(s) that matches exactly the values provided for
the fields. If non of the fields are provided, then it returns a reference to
the array of all the rows.

B<Usage:> C<
        my $ret_out = aggr_show(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            show_type => $BYKEY);>

        my $ret_out = aggr_show(tcd=>$Test,
            server => $Server,
            node => $Nodes[0],
            show_type => $BYKEY,
            id => $id);>

B<Returns:> C<$ret_out> Reference to the array of matched rows, undef if no
match

=cut

sub aggr_show {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
            show_type => 1
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            show_type       => {type => SCALAR},
            id              => {type => SCALAR, optional => 1},
            name            => {type => SCALAR, optional => 1},
            'blade-id'      => {type => SCALAR, optional => 1},
            'member-id'     => {type => SCALAR, optional => 1},
            'snmp-id'       => {type => SCALAR, optional => 1},
            comment         => {type => SCALAR, optional => 1 },
        );

    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node
    my $show_type = $options{'show_type'}; #show by key, by-name etc

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }


    my @show_output = ();
    eval {
        switch ($show_type) {
            case BYKEY {
                @show_output = $server->unit_test_mdb_aggr_show(%input_arg);
            }
            case BYNAME {
                @show_output = $server->unit_test_mdb_aggr_by_name_show(%input_arg);
            }
            case BYBLADEID {
                @show_output = $server->unit_test_mdb_aggr_by_bladeid_show(%input_arg);
            }
            case BYMEMBERID {
                @show_output = $server->unit_test_mdb_aggr_by_memberid_show(%input_arg);
            }
            case BYSNMPID {
                @show_output = $server->unit_test_mdb_aggr_by_snmpid_show(%input_arg);
            }
            else {
                $tcd->nlog("Invalid show type $show_type");
                return undef;
            }
        }
    };

    if ($@) {
        $tcd->nlog("aggr_show:ERROR $@");
        return undef;
    }

    return (\@show_output);
}

=head2 aggr_modify(
    tcd => $tcd_obj,
    server => $server_obj,
    node => $node,
    modify_type = > $modifytype,
    id => $id, name => $name,
    'blade-id' => $bladeid,
    voltype => $voltype,
    'member-id' => $memberid,
    'snmp=id' => $snmpid,
    comment => $comment)

Modify the field(s) for the corresponding match of id, name etc.
modify_type can be BYKEY, BYNAME, BYBLADEID, BYMEBMERID or BYSNMPID.
id, name, voltype, blade-id, member-id, snmp-id, comment are all optional
fields those need to be modified.

B<Usage:> C<
        $name = "aggr_$id";
        $voltype = "flex";
        $comment = $name;
        $id--;

        my $ret_status = aggr_modify(tcd => $Test, node => $Nodes[0],
                                    server => $Server,
                                    modify_type => BYKEY,
                                    id => $id, name => $name,
                                    'blade-id' => $blade_id,
                                    voltype => $voltype,
                                    'member-id' => $member_id,
                                    'snmp-id' => $snmp_id,
                                    comment => $comment);


B<Returns:> C<$ret_status> 1 if success, undef if fails.

=cut

sub aggr_modify {
    my %invalid_cmd_keys =
        (
            tcd => 1,
            server =>1,
            node =>1,
            modify_type => 1
        );

    my %args_validn_spec =
        (
            tcd             => {type => OBJECT},
            server          => {type => OBJECT},
            node            => {type => SCALAR},
            modify_type     => {type => SCALAR},
            id              => {type => SCALAR, optional => 1},
            name            => {type => SCALAR, optional => 1},
            nickname        => {type => SCALAR, optional => 1},
            'blade-id'      => {type => SCALAR, optional => 1},
            'member-id'     => {type => SCALAR, optional => 1},
            'snmp-id'       => {type => SCALAR, optional => 1},
            comment         => {type => SCALAR, optional => 1 },
        );

    my %options = validate(@_, \%args_validn_spec);
    my %input_arg; # hash of input argument to cli command
    my $tcd = $options{'tcd'}; #TCD Object
    my $server = $options{'server'}; # server object
    my $node = $options{'node'}; #Execution node
    my $modify_type = $options{'modify_type'}; # modify by key, by-name etc

    foreach my $key ( keys %args_validn_spec ) {
        if ( not defined($invalid_cmd_keys{$key}) ) {
            if ( defined($options{$key}) ) {
                $input_arg{$key} = $options{$key};
            }
        }
    }

    eval {
        switch ($modify_type) {
            case BYKEY {
                $server->unit_test_mdb_aggr_modify(%input_arg);
            }
            case BYNAME {
                $server->unit_test_mdb_aggr_by_name_modify(%input_arg);
            }
            case BYBLADEID {
                $server->unit_test_mdb_aggr_by_bladeid_modify(%input_arg);
            }
            case BYMEMBERID {
                $server->unit_test_mdb_aggr_by_memberid_modify(%input_arg);
            }
            case BYSNMPID {
                $server->unit_test_mdb_aggr_by_snmpid_modify(%input_arg);
            }
            else {
                $tcd->nlog("Invalid modify type $modify_type");
                return undef;
            }
        }
    };

    if ($@) {
        $tcd->nlog("aggr_modify:ERROR $@");
        return undef;
    }

    return (1);
}

=head2 aggr_table_cleanup(server => $server_obj)

Deletes all the rows in aggr table created by the test.

B<Usage:> C<$ret_status = aggr_table_cleanup(server => $Server);>

B<Returns:> C<$ret_status> 1 if success, undef if failed.

=cut

sub aggr_table_cleanup {
    my %args_validn_spec =
        (
            server => {type => OBJECT},
        );

    my %options = validate(@_, \%args_validn_spec);

    my $server = $options{'server'};

    eval {
        $server->unit_test_mdb_aggr_show();
    };

    if ((!$@) && ($@ !~ /empty/i)) { # Table is not empty
        # Delete the rows
        eval {
            $server->unit_test_mdb_aggr_delete(id => '*');
        };

        if ($@) {
            return undef;
        }
    }

    return 1;
}

1;
