#-------------------------------------------------------------------------------
# Author: Mauricio Sanchez
#   Date: 2012-Mar-29
#-------------------------------------------------------------------------------
# Description:
#   Update the build version in a .inf file.
#   Logic is based on "san_host_win_mpio" fixbld.pl script.
#
#   PERL module will generate the FlashAccel build number
#   based on number of days since January 1, 2006
#------------------------------------------------------------------------------

package Mercury::Build::DriverVersion;

use strict;
use warnings;

use POSIX qw(floor);
use Time::Local;

use Mercury::File::IO qw(file_get_contents file_put_contents);

#-------------------------------------------------------------------------------
#   Instatiate an object with the initial .inf and resc file
#------------------------------------------------------------------------------
sub new {

    # Get the hashed arguments
    my ($class, %args) = @_;

    # Create the actual object
    my $this = bless({}, $class);

    # Parse the hashed arguments
    $this->{inf1_file_in}  = $args{inf1_file_in} if ( $args{inf1_file_in} );
    $this->{inf2_file_in}  = $args{inf2_file_in} if ( $args{inf2_file_in} );
    $this->{resc_file_in} = $args{resc_file_in} if ( $args{resc_file_in} );

    # Read the entire inf files into arrays
    @{$this->{inf1_file_in_array}} = file_get_contents(file => $this->{inf1_file_in}) if ( $this->{inf1_file_in} );
    @{$this->{inf2_file_in_array}} = file_get_contents(file => $this->{inf2_file_in}) if ( $this->{inf2_file_in} );

    # Read the entire resc file into an array
    @{$this->{resc_file_in_array}} = file_get_contents(file => $this->{resc_file_in}) if ( $this->{resc_file_in} );

    return $this;
}

#-------------------------------------------------------------------------------
#   Update the Version info  in the arrayis; reflecting current time-stamp
#   Build Suffixes are added as follows, for promoted builds
#       - Release build            (no promote id)-    '0'
#       - Emergency release build  PromoteID 'R' -     '1'
#       - Patch                    PromoteID 'P' -     '2'
#       - Debug                    PromoteID 'D' -     '3'
#       - BlackDuck                PromoteID 'B' -     '6'
#       - Release Candidate        PromoteID 'RC'-     '7'
#       - Non-release build        PromoteID 'X'-      '8'
#       - Dev Build and all others PromoteID NA -      '9'
#------------------------------------------------------------------------------
sub set_build_num {

    # Initialize vars (these are relative to the product directory)
    my $build_id_file = './nightly_build/build_id.txt';
    my $build_type    = $ENV{BUILDTYPE};
    my $ver_file      = './nightly_build/version.txt';
	my $promote_build_flag = 0;
	my %build_suffixes = ('R' => '1', 'P' => '2', 'D' => '3', 'B' => '6', 'RC' => '7', 'X' => '8', 'N' => '9');

    # Get the hashed arguments
    my ($this, %args) = @_;

	# Check if the build is a promote build and get the type
	my $promote_version = $args{version};
	if ($promote_version eq '') {
		# Doing this to differentiate between a nightly and a release build
		$promote_version = 'N';
	}

	# Remove all . and numbers. Eg. 1.1.0xyz1 --> xyz
	$promote_version =~ s/[\d\.]//g;
	# Making this case insensitive
	$promote_version = uc($promote_version);

    # Define the version vars
    my ($ver_major, $ver_minor, $ver_micro, $ver_pico, $ver_date);

    # Only one line, but must use array context
    my @versions = grep(/^Version\s*=\s*/, file_get_contents(file => $ver_file));

    # Isolate the value (after the "=" without spaces)
    my $versions = $versions[$#versions];
    $versions =~ s/^Version\s*=\s*([\d\.]+)\s*/$1/;

    # Isolate 'major', 'minor' and 'release' portions
    ($ver_major, $ver_minor, $ver_micro) = split('\.', $versions);
    chomp $ver_micro;
    $this->{ver_major}   = $ver_major;
    $this->{ver_minor}   = $ver_minor;
    $this->{ver_micro} = $ver_micro;

    # Check if the $build_id_file has been created
    if ( -e $build_id_file ) {

        # Only one line, but must use array context
        my @build_ids = grep(/^Build\s*=\s*/, file_get_contents(file => $build_id_file));
        chomp $build_ids[$#build_ids];

        # If it was a three part version... $ver_micro will already be defined
        if ( defined $ver_micro )
        {
            # Only one portion... get 'pico' portion
            $ver_pico = $build_ids[$#build_ids];
			$ver_pico =~ s/^Build\s*=\s*([\d]+)\s*/$1/;

			chop($ver_pico);
			# Adding Build Suffixes according to specification
	        if ($ver_pico eq '10101010') {
	            $ver_pico = $ver_pico.'9';
		    } elsif ($promote_version eq '') {
				$ver_pico = $ver_pico.'0';
	        } elsif (!(defined $build_suffixes{$promote_version})) {
			   $ver_pico = $ver_pico.'9';
	        } else {
			    $ver_pico = $ver_pico.$build_suffixes{$promote_version};
	        }

			my @ver_build_out = (
		    "# Do not edit this file.\n",
		    "# Do not check-in.\n",
			"Build = " . $ver_pico . "\n"
			);

			# Update the $build_id_file
	        file_put_contents(file => $build_id_file, array => \@ver_build_out);

            $this->{ver_pico} = $ver_pico;
            $this->{build_id} = $ver_pico;
        }
        else
        # It was a two part build-id (for backward compatability)
        {
            # Isolate the 'micro' portion
            $ver_micro = $build_ids[$#build_ids];
            $ver_micro =~ s/^Build\s*=\s*([\d]+)\.[\d]+\s*/$1/;
            $this->{ver_micro} = $ver_micro;

            # Isolate the 'pico' portion
            $ver_pico = $build_ids[$#build_ids];
			$ver_pico =~ s/^Build\s*=\s*[\d]+\.([\d]+)\s*/$1/;
            $this->{ver_pico} = $ver_pico;
            $this->{build_id} = "$ver_micro.$ver_pico";
        }

    # Need to create $build_id_file
    } else {

        # If it was a three part version... $ver_micro will already be defined
        if ( defined $ver_micro )
        {
            # p4 commands
            my $p4 = '/usr/software/rats/bin/p4';
            my $ver_p4change;

            if ( -e $p4 ) {

                # Check if we have a p4 client
                my $p4client = `$p4 client -o 2>/dev/null| grep ^Client:| awk '{print \$2}'`;
                chomp($p4client);
                die "Not building in a P4 client" unless ( $p4client );

                # See if there are open p4 changes
                my $open_files = `$p4 opened -C $p4client 2>/dev/null`;
                if ( $open_files )
                {
                    print STDERR "This is a DevBuild!!!";
                    print STDERR "Setting p4Change to 10101010.";
                    $ver_p4change = '10101010';
                }
                else
                # Get the last p4Change
                {
                    $ver_p4change = `$p4 changes -m1 -s submitted ./...\@$p4client 2>/dev/null| awk '{print \$2}'`;
                    chomp $ver_p4change;
                
				}

            # No p4 command...
            } else {

                # Probably a window's build... See if P4CHANGELEVEL is set
				$ver_p4change = ( $ENV{P4CHANGELEVEL} )? $ENV{P4CHANGELEVEL}: '10101010';
            }
			
	        # Adding Build Suffixes according to specification
		    if ($ver_p4change eq '10101010') {
				$ver_p4change = $ver_p4change.'9';
			} elsif ($promote_version eq '') {
	            $ver_p4change = $ver_p4change.'0';
	        } elsif (!(defined $build_suffixes{$promote_version})) {
	            $ver_p4change = $ver_p4change.'9';
		    } else {
	            $ver_p4change = $ver_p4change.$build_suffixes{$promote_version};
	        }

            # For backwards compatability
			$ver_pico = $ver_p4change;
            $this->{ver_pico} = $ver_pico;
            $this->{build_id} = $ver_pico;
        }
        else
        # It was a two part version (for backward compatability)
        {
            # Parse the local-time
            my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = localtime(); 

            # Fix the internal representation to match real-world
            $mon++;
            $year += 1900;

            # Get our baseline epoch
            my $date1 = '2006-01-01_00:00:00';
            my $e1 = _date2epoch($date1);

            # Get our current-day epoch
            my $date2 = sprintf('%d-%02d-%02d_%02d:%02d:%02d', $year, $mon, $mday, $hour, $min, $sec);
            my $e2 = _date2epoch($date2);

            # Calculate delta in seconds
            my $diff = $e2 - $e1;

            # Capture a properly formatted date
            $ver_date = sprintf("%02d/%02d/%d", $mon, $mday, $year);

            # We add 2000 days making the reference July 11, 2000. (why? Because MPIO did it)
            $ver_micro = floor(($diff/86400) + 2000);
            $this->{ver_micro} = $ver_micro;

            # Create the 'build' portion
            $ver_pico = sprintf("%02d%02d", $hour, $min);
            $ver_pico = int($ver_pico);
            $this->{ver_pico}   = $ver_pico;
            $this->{build_id} = "$ver_micro.$ver_pico";
        }

        my @ver_build_out = (
            "# Do not edit this file.\n",
            "# Do not check-in.\n",
            "Build = " . $this->{build_id} . "\n"
        );

        # Create the $build_id_file
        file_put_contents(file => $build_id_file, array => \@ver_build_out);
    }

    # Check that we have a inf1 file
    if ( $this->{inf1_file_in} ) {

        # Iterate through the entire inf file
        for (@{$this->{inf1_file_in_array}}) {

            # Look for strings and replace all the fields
            s#(^DriverVer\s*=\s*)\d+/\d+/\d+(\D*)\d+\.\d+\.\d+\.\d+$#$1$ver_date$2$ver_major.$ver_minor.$ver_micro.$ver_pico#; 
        }
    }

    # Check that we have a inf2 file
    if ( $this->{inf2_file_in} ) {

        # Iterate through the entire inf file
        for (@{$this->{inf2_file_in_array}}) {

            # Look for strings and replace all the fields
            s#(^DriverVer\s*=\s*)\d+/\d+/\d+(\D*)\d+\.\d+\.\d+\.\d+$#$1$ver_date$2$ver_major.$ver_minor.$ver_micro.$ver_pico#; 
        }
    }

    # Check that we have a resc file
    if ( $this->{resc_file_in} ) {

        # Iterate through the entire resc file
        for (@{$this->{resc_file_in_array}}) {

            # Update all the values in the resc array
            s/(#define\s+VER_MAJOR\s+)\S+\s*$/$1$ver_major\n/;
            s/(#define\s+VER_MINOR\s+)\S+\s*$/$1$ver_minor\n/;
            s/(#define\s+VER_RELEASE\s+)\S+\s*$/$1$ver_micro\n/;
            s/(#define\s+VER_BUILD\s+)\S+\s*$/$1$ver_pico\n/;
            s/(#define\s+VER_SPECIAL_BUILD_STRING\s+)".+"\s*$/$1"$build_type"\n/;
        }
    }

    # Set the build_num
    $this->{build_num} = "$ver_major.$ver_minor.$ver_micro.$ver_pico";
}

#-------------------------------------------------------------------------------
#   Save the corresponding file from the array
#------------------------------------------------------------------------------
sub save {

    my $file_out = '';
    my $output_array;

    # Get the hashed arguments
    my ($this, %args) = @_;

    # See if an 'inf1' file was specified
    if ( $args{inf1_file_out} ) {

        # Set the output_file and the corresponding array
        $file_out = $args{inf1_file_out};
        $output_array = $this->{inf1_file_in_array};

    # See if an 'inf2' file was specified
    } elsif ( $args{inf2_file_out} ) {

        # Set the output_file and the corresponding array
        $file_out = $args{inf2_file_out};
        $output_array = $this->{inf2_file_in_array};

    # See if an 'resc' file was specified
    } elsif ( $args{resc_file_out} ) {

        # Set the output_file and the corresponding array
        $file_out = $args{resc_file_out};
        $output_array = $this->{resc_file_in_array};

    # Neither was specified... die
    } else {
        die "No output-file specified: $!\n";
    }

    # Send the array to the file
    file_put_contents(file => $file_out, array => $output_array);
}

#-------------------------------------------------------------------------------
#   Get build_num
#   This returns "$ver_major.$ver_minor.$ver_micro.$ver_pico"
#------------------------------------------------------------------------------
sub get_build_num {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{build_num};
}

#-------------------------------------------------------------------------------
#   Get ver_major
#------------------------------------------------------------------------------
sub get_ver_major {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_major};
}

#-------------------------------------------------------------------------------
#   Get ver_minor
#------------------------------------------------------------------------------
sub get_ver_minor {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_minor};
}

#-------------------------------------------------------------------------------
#   Get ver_micro
#------------------------------------------------------------------------------
sub get_ver_micro {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_micro};
}

#-------------------------------------------------------------------------------
#   Get ver_pico
#------------------------------------------------------------------------------
sub get_ver_pico {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_pico};
}

#-------------------------------------------------------------------------------
#   Get full_version
#------------------------------------------------------------------------------
sub get_version_only {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_major} . '.' . $this->{ver_minor} . '.' . $this->{ver_micro};
}

#-------------------------------------------------------------------------------
#   Get build_id
#------------------------------------------------------------------------------
sub get_build_id {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{build_id};
}

#-------------------------------------------------------------------------------
#   Get ver_release
#   (This is here for backward compatibility)
#------------------------------------------------------------------------------
sub get_ver_release {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_micro};
}

#-------------------------------------------------------------------------------
#   Get ver_build
#   (This is here for backward compatibility)
#------------------------------------------------------------------------------
sub get_ver_build {

    # Get the hashed arguments (don't expect any)
    my ($this, %args) = @_;

    # Return the Build Number
    return $this->{ver_pico};
}

#-------------------------------------------------------------------------------
#   Convert a string to an epoch using the timelocal function
#------------------------------------------------------------------------------
sub _date2epoch {
  my $date = shift;

  my @d = ();
  push @d, substr $date, 0, 4;
  push @d, substr $date, 5, 2;
  push @d, substr $date, 8, 2;
  push @d, substr $date, 11, 2;
  push @d, substr $date, 14, 2;
  push @d, substr $date, 17, 2;
  return timelocal $d[5], $d[4], $d[3], $d[2], $d[1] - 1, $d[0] - 1900;
}

1;
