#
# $Id: //depot/prod/test/main/lib/Mercury/ESXi.pm $
#
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc
#
## @summary      Library used to track ESXi properties and the devices under the server.
##
## @author       nkrajara@netapp.com
##
## @pod          here
##

=head1 NAME

ESXi.pm

=head1 AUTHOR

Narayan Rajaram

=head1 DESCRIPTION

C<ESXi> will maintain the properties of the ESXi machine.

1. Will maintain a list of SSD devices 

2. Will maintain a list of vlun created by the FA plugin

3. Will provide interfaces to get/set properties on the ESXi

=head1 SYNOPSIS

Library used to maintain the state and the properties of the Mercury Plugin in the ESXi host

=cut

package Mercury::ESXi;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT);
use Tharn;
use Mercury::SSD;
use Mercury::Vlun;
use Mercury::ESXi::FlashAccelCLI;
use Mercury::ESXi::CimClient;
use Error;
use San::SanLog;
use vars qw(@ISA);
@ISA = qw(Error SanLog);

# Netapp Specific Libs

use NACL::APISet;
use TracePkg;

# Global Variable
my $log = $Tharn::Log;

=head2 new

To instantiate and configure the ESXi object

=cut

#############################################################################
# Subroutine: new
# Parameters: None
# Returns:    Reference to this object
#############################################################################

sub new {

    my $class = shift(@_);
    my $self;
    my %params = validate(
        @_,
        {   host_name => {
                type     => SCALAR,
                optional => 0,
                default  => "*",
            },
            ssd_id => {
                type     => SCALAR,
                optional => 0,
                default  => "*",
            },
            user_name => {
                type     => SCALAR,
                optional => 1,
                default  => "root",
            },
            client_name => {
                type     => SCALAR,
                optional => 1,
                default  => "hg-qa-ndisp",
            },
            password => {
                type     => SCALAR,
                optional => 1,
                default  => "netapp1!",
            },
            faconfig => {
                type     => SCALAR,
                optional => 1,
                default  => "0",
            },
        }
    );

    # Copy specified parameters globally into object

    $self->{_host_name}    = $params{host_name};
    $self->{_ssd_id}       = $params{ssd_id};
    $self->{_user_name}    = $params{user_name} if defined($params{user_name});
    $self->{_client_name}  = $params{client_name} if defined($params{client_name});
    $self->{_password}     = $params{password} if defined($params{password});
    $self->{_faconfig}     = $params{faconfig} if defined($params{faconfig});

    #my %vluns = ();
    #    $self->{_vluns} = \%vluns;
    $self->{_vluns} = ();

    # Still need to figure out how to update the vm object
    $self->{_vms} = ();

    # Create a hostobj from the $self->{_host_name}
    my $Host       = Tharn::host($self->{_host_name});
    my $connection = Tharn::connect("#ssh");

    my $Api_Set_Obj = NACL::APISet->new(
        connection => $connection,
        category   => "Host",
        interface  => "CLI",
        set        => "Linux"
    );
    $self->{_shell} = $Api_Set_Obj;
    if($self->{_faconfig}){
        $self->{_facli}
            = Mercury::FlashAccelCLI->new(host_name => $self->{_host_name});
    }
    else {
        $log->log("HOT NAME = $self->{_host_name}");
        $self->{_facli}
            = Mercury::ESXi::CimClient->new(host_name => $self->{_host_name},
                                        client_name => $self->{_client_name},
                                        user_name => $self->{_user_name},
                                        pwd => $self->{_password});
    }

    # Creating a SSD objec
    $self->{ssd} = Mercury::SSD->new(
        id    => $self->{_ssd_id},
        shell => $Api_Set_Obj
    );
    $self->{ssd}->set_ssd();

    bless($self, $class);
    $log->log("Created a ESXi object\n");
    return $self;

} ## end sub new

#############################################################################
# Subroutine: check_ssd_flag
# Parameters: NAA ID
# Returns:    True if the device is a SSD
#############################################################################

sub check_ssd_flag {
    my ($self, %args) = @_;
    $self->_clear_err();
    if ($self->{ssd}->get_ssd_property(property => 'Is SSD') ne 'true') {
        $self->_set_err("$self->{_ssd_id} is not an SSD");
        return;
    }
    $log->log("$self->{_ssd_id} is an SSD");
    return 1;
} ## end sub check_ssd_flag

#############################################################################
# Subroutine: naa_exists
# Parameters: NAA ID
# Returns:    True if the device exists
#############################################################################

sub naa_exists {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $naa_id  = $args{naa_id};
    my $command = "esxcli storage core device list | grep -i $naa_id";
    my $output  = $self->{_shell}->execute_raw_command(command => $command);
    if ($output !~ /$naa_id/) {
        $self->_set_err("$naa_id is not present");
        return;
    }
    return 1;
} ## end sub naa_exists

#############################################################################
# Subroutine: set_vib
# Parameters:
# Returns:    Vib_version if the VIB is installed
#             Returns 0 otherwise
#############################################################################

sub set_vib {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $command = "esxcli software vib get -n netapp-flashaccel";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /Name\:\s+netapp-flashaccel/) {
        $output =~ /Version\:\s+(.*)\n/;
        $self->{vib_version} = $1;
    } else {
        $self->_set_err(
            "====FlashAccel not Installed on $self->{_host_name}");
        $log->error("====FlashAccel not Installed on $self->{_host_name}");
    }
} ## end sub set_vib

#############################################################################
# Subroutine: get_vib
# Returns:    Vib_version if the VIB is installed
#             Returns 0 otherwise
#############################################################################

sub get_vib {
    my ($self, %args) = @_;
    $self->_clear_err();
    return $self->{vib_version};
}

#############################################################################
# Subroutine: claim_device
# Parameters:
# Returns:    True, if we claimed the device successfully
#                 Returns False otherwise.
#############################################################################

sub claim_device {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $adapter;
    my $channel;
    my $target;
    my $lun;

    #check if the SSD is already claimed
    if ($self->{ssd}->get_ssd_property(property => "Multipath Plugin") eq
        "FlashAccel") {
        $self->_set_err(" DEVICE ALREADY CLAIMED BY FLASHACCEL");
    }

    # Command to get the path of the device to be claimed.
    my $command = "esxcli storage core path list --device=$self->{_ssd_id}";
    my $output = $self->{_shell}->execute_raw_command(command => $command);

    # Regular expression to get the Adapter, Channel, Target and Lun
    if ($output
        =~ /Adapter\:\s+(.*)\n\s+Channel\:\s+(\d+)\n\s+Target\:\s+(\d+)\n\s+LUN\:\s+(\d+)/
        ) {
        ($adapter, $channel, $target, $lun) = ($1, $2, $3, $4);
    } else {
        $self->_set_err("Not able to execute $command");
        return;
    }

    # Adding New claim rule
    $log->comment("ADDING FLASHACCEL CLAIM RULE");
    $command
        = "esxcli storage core claimrule add --autoassign -t location -A $adapter -C $channel -T $target -L $lun -P FlashAccel";
    $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /Claimrule is duplicate of rule/) {
        $log->warn("There is a claim rule already present \n");
    }

    $command = "esxcli storage core claimrule load";
    $output = $self->{_shell}->execute_raw_command(command => $command);
    my @outarray = split("\n", $output);
    if (scalar(@outarray > 1)) {
        $self->_set_err("Unable to load the claimrule");
        return;
    }

    $command  = "esxcli storage core claimrule run";
    $output   = $self->{_shell}->execute_raw_command(command => $command);
    @outarray = split("\n", $output);
    if (scalar(@outarray > 1)) {
        $self->_set_err("Unable to run the claim rule");
        return;
    }
    $log->comment("Loaded Claim Rule");
    $command = "esxcli storage core claimrule list";
    $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output !~ /FlashAccel/) {
        $self->_set_err("FLASH ACCEL CLAIM RULE IS NOT ADDED");
        return;
    }

    $command
        = "esxcli storage core claiming  reclaim --device=$self->{_ssd_id}";
    $output = $self->{_shell}->execute_raw_command(command => $command);

    $command
        = "esxcli storage core device list --device=$self->{_ssd_id}";
    $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output !~ /Unable to find device/i) {
        $self->_set_err("Device visible after claiming");
        return;
    }

} ## end sub claim_device

#############################################################################
# Subroutine: unclaim_device
# Returns:    True, if the device was unclaimed successfully
#                 Returns False otherwise.
#############################################################################

sub unclaim_device {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $adapter;
    my $channel;
    my $target;
    my $lun;

    #check if the SSD is not claimed
    my $flash = $self->{_facli}->get_claimed_devices();
    if ($self->error()){
        $self->_set_err("No device has been claimed");
        return ;
    }
    my $path = $flash->{HBTL};
    $log->comment("PATH = $path");

    #Command to unclaim the device
    my $command
        = "esxcli storage core claiming unclaim --type=path --path=$path";
    my $output = $self->{_shell}->execute_raw_command(command => $command);

    if ($output =~ /error/i) {
        $self->_set_err("Device unclaim failed with $output");
        return;
    }

    # Removing FlashAccel claim rule
    $log->comment("REMOVING FLASHACCEL CLAIM RULE");
    $command = "esxcli storage core claimrule remove --plugin FlashAccel";
    $output = $self->{_shell}->execute_raw_command(command => $command);

    # load and run the claimrule
    $command = "esxcli storage core claimrule load";
    $output  = $self->{_shell}->execute_raw_command(command => $command);
    $command = "esxcli storage core claimrule run";
    $output  = $self->{_shell}->execute_raw_command(command => $command);
    $log->comment("Loaded Claim Rule");

    # Reclaim the device which should assign NMP as the owner
    $command
        = "esxcli storage core claiming  reclaim --device=$self->{_ssd_id}";
    $output = $self->{_shell}->execute_raw_command(command => $command);

    my $retry_count = 10;
    while ($retry_count > 0) {
        $self->{ssd}->set_ssd();
        if ($self->{ssd}->get_ssd_property(property => "Multipath Plugin") ne
            "FlashAccel") {
            $log->log(" Device successfully unclaimed");
            return 1;
        }
        sleep(5);
        $retry_count = $retry_count - 1;
    }

    $self->_set_err("DEVICE UNCLAIM FALED");

} ## end sub unclaim_device

#############################################################################
# Subroutine: create_vlun
# Parameters: Number of blocks for the vlun
# Returns:    SSD ID of the new vlun
#############################################################################

sub create_vlun {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $blocks = $args{blocks};
    my $naa_id = $self->{_facli}->create_vlun(%args);
    if ($self->{_facli}->error()) {
        $self->_set_err($self->{_facli}->errmsg());
        return;
    }
    $log->comment("VLUN created with ID $naa_id");
    my $vlun = Mercury::Vlun->new(
        id    => $naa_id,
        shell => $self->{_shell}
    );
    $vlun->set_ssd();

    if ($vlun->get_ssd_property(property => "Is SSD") =~ /false/) {
        $self->_set_err("VLUN created as a non-SSD device");
        # Destroy this vlun as a part of the cleanup
        $self->{_facli}->destroy_vlun(naa_id => $naa_id);
        return ;
    }

    $self->{_vluns}->{$naa_id} = $vlun;
    return $naa_id;

} ## end sub create_vlun

#############################################################################
# Subroutine: destroy_vlun
# Parameters: NAA ID of the vlun to be destroyed
# Returns:    Nothing
#############################################################################

sub destroy_vlun {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $naa_id = $args{naa_id};
    $self->{_facli}->destroy_vlun(naa_id => $naa_id);
    if ($self->{_facli}->error()) {
        $self->_set_err($self->{_facli}->errmsg());
        return;
    }
    delete $self->{_vluns}->{$naa_id};
} ## end sub destroy_vlun

#############################################################################
# Subroutine: format_cache
# Returns:    1 when format is successful
#             Returns 0 otherwise
#############################################################################

sub format_cache {
    my ($self, %args) = @_;
    $self->_clear_err();
    $self->{_facli}->format_cache(%args);
    if ($self->{_facli}->error()) {
        $self->_set_err($self->{_facli}->errmsg());
        return;
    }
    my $key;
    foreach $key (keys %{$self->{_vluns}}) {
        delete $self->{_vluns}->{$key};
    }
} ## end sub format_cache

#############################################################################
# Subroutine: get_claimed_devices
# Returns:    Reference to an Hash
#############################################################################

sub get_claimed_devices {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $claimed_device = $self->{_facli}->get_claimed_devices();
    if ($self->{_facli}->error()) {
        $self->_set_err($self->{_facli}->errmsg());
        return;
    }
    return $claimed_device;
} ## end sub get_claimed_devices

#############################################################################
# Subroutine: get_cache_props
# Returns:    Refernce to an Hash with all the properties as the keys
#############################################################################

sub get_cache_props {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $props = $self->{_facli}->get_cache_props();
    if ($self->{_facli}->error()) {
        $self->_set_err($self->{_facli}->errmsg());
        return;
    }
    return $props;
} ## end sub get_cache_props

#############################################################################
# Subroutine: get_cache_spaces
# Returns:    Reference to a hash which internally is a hash of hashes
#             with the outer hash having the NAA_ID of the vluns as the key
#############################################################################

sub get_cache_spaces {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $cache_spaces = $self->{_facli}->get_cache_spaces();
    if ($self->{_facli}->error()) {
        $self->_set_err($self->{_facli}->errmsg());
        return;
    }
    return $cache_spaces;
} ## end sub get_cache_spaces

#############################################################################
# Subroutine: reboot_host
#############################################################################

sub reboot_host {
    my ($self, %args) = @_;
    $self->_clear_err();
    $self->{_shell}->execute_raw_command(command => "reboot");
    while ($self->{_shell}->api_is_session_alive()) {
        $log->log("HOST IS GOING DOWN!!!!");
        sleep(20);
    }
    $log->log("Machine has gone down");
    $self->{_shell}->api_reconnect_session(max_reconnect => 15,
                                           max_reconnect_timewait => 60);

} ## end sub reboot

#############################################################################
# Subroutine: remove_all_vluns
#############################################################################

sub remove_all_vluns {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $key;
    foreach $key (keys %{$self->{_vluns}}) {
        $self->{_facli}->destroy_vlun(naa_id => $key);
        if ($self->{_facli}->error()) {
            $self->_set_err($self->{_facli}->errmsg());
            return;
        }  
        delete $self->{_vluns}->{$key};
    }
} ## end sub remove_all_vluns

1;

