#
# $Id: //depot/prod/test/main/lib/Mercury/ESXi/Cim/CimClient.pm $
#
# Copyright (c) 2009 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc
#
## @summary      Library will provide interfaces to the CLI command provied by the CIM PROVIDER
##
## @author       nkrajara@netapp.com
##
## @pod          here
##

=head1 NAME

CimClient.pm

=head1 AUTHOR

Narayan Rajaram

=head1 DESCRIPTION

C  will provide interfaces to the python scripts which internally use pywbem

1.Interface for CIM provider interfaces

=head1 SYNOPSIS

Library provides interfaces to the python scrips which talk to CIM provider

Please note that the library internally uses python scripts.

These python scripts reside in /u/nkrajara/pywbem_scripts

=cut

package Mercury::ESXi::CimClient;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Tharn;
use Error;
use San::SanLog;
use vars qw(@ISA);
@ISA = qw(Error SanLog);

# Netapp Specific Libs

use NACL::APISet;

# This path will change once we know the correct path to add
use constant PATH => "CIM/";

# All the script calls
use constant CREATE_VLUN => "python " . PATH . "AllocateCacheSpace.py";
use constant DESTROY_VLUN => "python " . PATH . "FreeCacheSpace.py";
use constant FORMAT_CACHE => "python " . PATH . "InitCache.py";
use constant LIST_CACHE_SPACES => "python " . PATH . "ListCacheSpaces.py";
use constant LIST_CLAIMED_DEVICES => "python " . PATH . "ListFlashDevices.py";
use constant GET_PROPERTIES => "python " . PATH .  "GetProperties.py";

# Global Variable
my $log = $Tharn::Log;

=head2 new

To instantiate and configure an CIMCLIENT object

=cut

#############################################################################
# Subroutine: new
# Returns:    Reference to this object
#############################################################################

sub new {

    my $self;
    my ($class, %args) = @_;

    $self->{_host_name}   = $args{host_name};
    $self->{_client_name} = $args{client_name};
    $self->{_user_name}   = $args{user_name};
    $self->{_pwd}         = $args{pwd};

    # Create a hostobj from the $self->{_client_nam}
    my $Host       = Tharn::host($self->{_client_name});
    my $connection = Tharn::connect("#ssh");

    $self->{_shell} = NACL::APISet->new(
        connection => $connection,
        category   => "Host",
        interface  => "CLI",
        set        => "Linux"
    );

    bless($self, $class);
    $log->comment("Created a CimCLient object\n");
    return $self;

} ## end sub new

#############################################################################
# Subroutine: create_vlun
# Parameters: size of the lun to be created
# Returns:    Return the output from the command to create vlun
#############################################################################

sub create_vlun {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $blocks = $args{blocks};
    my $naa_id = $args{naa_id};
    $log->comment("username = $self->{_user_name}");
    $log->comment("hostname = $self->{_host_name}");

    #Create Vlun script which internally uses the pywbem client
    my $command
        = CREATE_VLUN ." -e $self->{_host_name} " . 
        "-u $self->{_user_name} -p $self->{_pwd} -s $blocks";
    $command = $command . " -d $naa_id" if defined($naa_id);
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /(naa\.\w+)$/) {
        return $1;
    }

    #If the IOCTL call fails python scrips issues error in the below format:
    # ERROR CODE: 0x10000009L
    # ERROR MSG: Cache has not been initialized
    if ($output =~ /error/i) {
        #Vlun Cretion failed, Setting the error message
        $self->_set_err($output);
    }
    return $self;
} ## end sub create_vlun

#############################################################################
# Subroutine: destroy_vlun
# Parameters: ID of the vlun to be destroyed
#############################################################################

sub destroy_vlun {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $naa_id = $args{naa_id};
    my $command = DESTROY_VLUN .
        " -e $self->{_host_name} -u $self->{_user_name} " .
        "-p $self->{_pwd} -d $naa_id";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err("$output");
    }
    return $self;
} ## end sub destroy_vlun

#############################################################################
# Subroutine: format_cache
# Parameters: flag to set the force option
#############################################################################

sub format_cache {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $command =FORMAT_CACHE . 
        " -e $self->{_host_name} -u $self->{_user_name} -p $self->{_pwd}";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err("$output");
    }
    return $self;
} ## end sub format_cache

#############################################################################
# Subroutine: get_claimed_devices
# Parameters: None
# Returns:    Returns a reference to an hash with the claimed device info.
#############################################################################

sub get_claimed_devices {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $claimed_devices = ();
    my $command = LIST_CLAIMED_DEVICES .
        " -e $self->{_host_name} -u $self->{_user_name} -p $self->{_pwd}";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err($output);
        return;
    }
    my @outarray = split /\n|\R/, $output;
    my $len      = scalar(@outarray);
    my $index    = 0;
    while ($index < $len) {

        # Each loop will parse output of the form
        if ($outarray[$index] =~ /^\s*(\w+)\s+\:\s+(.*)$/) {
            my $value = $2;
            my $key = $1;
            if ($value =~ /(vmhba.*L\d)/) {
                $value = $1;
            }
            $claimed_devices->{$key} = $value;
        }
        $index = $index + 1;
    }
    return $claimed_devices;
} ## end sub get_claimed_devices

#############################################################################
# Subroutine: get_cache_props
# Parameters: None
# Returns:    Returns the reference to an hash with the properties as the keys
#############################################################################

sub get_cache_props {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $cmd_status = 1;
    my $props      = ();
    print "get properties ".GET_PROPERTIES;
    my $command = GET_PROPERTIES .
         " -e $self->{_host_name} -u $self->{_user_name} -p $self->{_pwd}";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err($output);
        return;
    }
    my @outarray = split /\n|\R/, $output;
    foreach my $out (@outarray) {
        if ($out =~ /^(.*\w)\s+\:\s+(.*)$/) {
            $props->{$1} = $2;
        }
    }
    return $props;
} ## end sub get_cache_props

#############################################################################
# Subroutine: get_cache_spaces
# Parameters: None
# Returns:    Returns a reference to an Hash with an list of cache spaces.
#############################################################################

sub get_cache_spaces {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $cmd_status   = 1;
    my $cache_spaces = ();
    $log->comment("username = $self->{_user_name}");
    $log->comment("hostname = $self->{_host_name}");
    my $command = LIST_CACHE_SPACES .
         " -e $self->{_host_name} -u $self->{_user_name} -p $self->{_pwd}";
    my $output = $self->{_shell}->execute_raw_command(command => $command);

    if ($output =~ /(error)/i) {
        $self->_set_err("$output");
        return;
    }
    my @outarray = split /\n|\R/, $output;
    my $len      = scalar(@outarray);
    my $index    = 0;
    while ($index < $len) {

        #Every loop will parse output in the following form
        if ($outarray[$index] =~ /^(naa\.\w+)\s+\:\s+(\d+)/) {
            my $naa_id = $1;
            my $size   = $2;
            $cache_spaces->{$naa_id} = $size;
        }
        $index = $index + 1;
    }
    return $cache_spaces;
} ## end sub get_cache_spaces

1;
