#
# $Id: //depot/prod/test/main/lib/Mercury/FlashAccelCLI.pm $
#
# Copyright (c) 2009 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc
#
## @summary      Library will provide interfaces to the CLI command provied by the FACFG tool.
##
## @author       nkrajara@netapp.com
##
## @pod          here
##

=head1 NAME

SSD.pm

=head1 AUTHOR

Narayan Rajaram

=head1 DESCRIPTION

C<CLI> will provide interfaces to the CLI commands provided by the FACFG tool

1.Interface for CLI commands 

=head1 SYNOPSIS

Library provides interfaces to the CLI commands provided by the FACFG tool

=cut

package Mercury::FlashAccelCLI;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Tharn;
use Error;
use San::SanLog;
use vars qw(@ISA);
@ISA = qw(Error SanLog);

# Netapp Specific Libs

use NACL::APISet;

# Global Variable
my $log = $Tharn::Log;

=head2 new

To instantiate and configure an SSD onject

=cut

#############################################################################
# Subroutine: new
# Parameters: None
# Returns:    Reference to this object
#############################################################################

sub new {

    my $self;
    my ($class, %args) = @_;

    $self->{_host_name} = $args{host_name};

    # Create a hostobj from the $self->{_host_name}
    my $Host       = Tharn::host($self->{_host_name});
    my $connection = Tharn::connect("#ssh");

    $self->{_shell} = NACL::APISet->new(
        connection => $connection,
        category   => "Host",
        interface  => "CLI",
        set        => "Linux"
    );

    # We may add more FACFG specific parameters here.

    bless($self, $class);
    $log->comment("Created a FlashAccelCLI object\n");
    return $self;

} ## end sub new

#############################################################################
# Subroutine: faconfig_installed
# Parameters: None
# Returns:  Sets the error value if FACONFIG is not installed.
#############################################################################

sub faconfig_installed {
    my ($self, %args) = @_;

    # Check to see if we have the faconfig installed in the right directory
    $self->_clear_err();
    my $cmd     = "ls -al /opt/netapp/flashaccel/config | grep -i faconfig";
    my $command = "ls -al /opt/netapp/flashaccel/ | grep -i faconfig";
    my $output  = $self->{_shell}->execute_raw_command(command => $command);
    if ($output !~ /faconfig/) {
        $self->_set_err("CLI tool FACONFIG is not available");
    }
} ## end sub faconfig_installed

#############################################################################
# Subroutine: create_vlun
# Parameters: size of the lun to be created
# Returns:    Return the output from the command to create vlun
#############################################################################

sub create_vlun {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $blocks  = $args{blocks};
    my $command = "/opt/netapp/flashaccel/faconfig -c Allocate-CacheSpace -s $blocks";
    my $output  = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /(naa\.\w+)$/) {
        return $1;
    }

    #Vlun Cretion failed, Setting the error message
    $self->_set_err("VLUN CRETION FAILED with $output");
} ## end sub create_vlun

#############################################################################
# Subroutine: destroy_vlun
# Parameters: ID of the vlun to be destroyed
#############################################################################

sub destroy_vlun {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $naa_id  = $args{naa_id};
    my $command = "/opt/netapp/flashaccel/faconfig -c Free-CacheSpace -d $naa_id";
    my $output  = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err("DESTROY VLUN failed with $output");
    }
} ## end sub destroy_vlun

#############################################################################
# Subroutine: format_cache
# Parameters: flag to set the force option
#############################################################################

sub format_cache {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $force   = $args{force};
    my $command = "/opt/netapp/flashaccel/faconfig -c Initialize-cache";

    # append the -f option if the force flag is set
    $command = "echo yes | " . $command . " -f" if defined($force);
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err("FORMAT CACHE failed with $output");
        return;
    }
} ## end sub format_cache

#############################################################################
# Subroutine: get_claimed_devices
# Parameters: None
# Returns:    Returns a reference to an hash with the claimed device info.
#############################################################################

sub get_claimed_devices {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $cmd_status      = 1;
    my $claimed_devices = ();
    my $command         = "/opt/netapp/flashaccel/faconfig -c List-FlashDevices";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err("List-FlashDevices failed with $output");
        return;
    }
    my @outarray = split("\n", $output);
    my $len      = scalar(@outarray);
    my $index    = 0;
    while ($index < $len) {

        # Each loop will parse output of the form
        # Flash Device - naa.5000cca01307ee64 \r\n
        #     Initialized : Initialized \r\n
        #     Size        : 186.310753 GB
        if ($outarray[$index] =~ /(naa\.\w+)/) {
            my $naa_id = $1;
            $index = $index + 1;
            while ($outarray[$index] =~ /^\s+(\w+)\s+\:\s+(.*\w)\s*$/) {
                $claimed_devices->{$naa_id}->{$1} = $2;
                $index = $index + 1;
            }
            $cmd_status = 0;
        }
        $index = $index + 1;
    }
    if ($cmd_status) {
        $self->_set_err("List-Flashdevices failed with $output");
        return;
    }
    return $claimed_devices;
} ## end sub get_claimed_devices

#############################################################################
# Subroutine: get_cache_props
# Parameters: None
# Returns:    Returns the reference to an hash with the properties as the keys
#############################################################################

sub get_cache_props {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $cmd_status = 1;
    my $props      = ();
    my $command    = "/opt/netapp/flashaccel/faconfig -c List-CacheProps";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /error/i) {
        $self->_set_err("List-CacheProps failed with $output");
        return;
    }
    my @outarray = split("\n", $output);
    foreach my $out (@outarray) {

        #Regex to parse lines of the form
        # "\t Total Space                         : 186.250000 GB \r\n"
        if ($out =~ /\s+(.*\w)\s+\:\s+(.*\w)\s*/) {
            $props->{$1} = $2;
            $cmd_status = 0;
        }
    }
    if ($cmd_status) {
        $self->_set_err("List-CacheSpace failed with $output");
        return 0;
    }
    return $props;
} ## end sub get_cache_props

#############################################################################
# Subroutine: get_version
# Parameters: None
# Returns:    Returns the version
#############################################################################

sub get_version {

    my ($self, %args) = @_;
    $self->_clear_err();
    my $command = "/opt/netapp/flashaccel/faconfig -c Get-Version";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /^FlashAccel\sVersion\s(.*)$/) {
        $log->comment("VERSION = $1");
        return $1;
    }
    $self->_set_err("get-version failed with $output");
} ## end sub get_version

#############################################################################
# Subroutine: get_cache_spaces
# Parameters: None
# Returns:    Returns a reference to an Hash with an list of cache spaces.
#############################################################################

sub get_cache_spaces {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $cmd_status   = 1;
    my $cache_spaces = ();
    my $command      = "/opt/netapp/flashaccel/faconfig -c List-CacheSpaces";
    my $output = $self->{_shell}->execute_raw_command(command => $command);
    if ($output =~ /(error|empty)/i) {
        $self->_set_err("List-CacheSpaces returned $output");
        return;
    }
    my @outarray = split("\n", $output);
    my $len      = scalar(@outarray);
    my $index    = 0;
    while ($index < $len) {

        #Every loop will parse output in the following form
        #VM Space Tag - naa.600a098016063fdc4b67a2f1fc0227d0
        #         Size        : 5.000000 GB
        #         hbtl        : 00:00:03:00
        if ($outarray[$index] =~ /(naa\.\w+)/) {
            my $naa_id = $1;
            $index = $index + 1;
            while ($outarray[$index] =~ /^\s+(\w+)\s+\:\s+(.*\w)\s*$/) {
                $cache_spaces->{$naa_id}->{$1} = $2;
                $index = $index + 1;
            }
        }
        $cmd_status = 0;
        $index      = $index + 1;
    }
    if ($cmd_status) {
        $self->_set_err("List-CacheSpaces failed with $output");
        return;
    }
    return $cache_spaces;
} ## end sub get_cache_spaces
1;
