################################################################################
#
# $Id$
#
#      Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#      Any use, modification, or distribution is prohibited
#      without prior written consent from NetApp, Inc.
#
# NAME
#      Mercury::HGAdmin.pm - This module interacts with the hgadmin command
#      line tool. It will open a single connection per object instance.
#
# DESCRIPTION
#      This perl module launches an instance of hgadmin and intercats with
#      the executable via inter-process-communication. A method is mapped
#      to each command.
#
# AUTHOR
#       Mauricio.Sanchez@netapp.com
#
################################################################################
## @pod here
##
##

=pod

=head1 NAME

    Mercury::HGAdmin.pm - This module interacts with the hgadmin command
    line tool. It will open a single connection per object instance.

=head1 SYNOPSIS

    use Mercury::HGAdmin;

    $hgadmin = new Mercury::HGAdmin();

=head1 DESCRIPTION

    This perl module launches an instance of hgadmin and intercats with
    the executable via inter-process-communication. A method is mapped
    to each command.

=head1 METHODS

    set_host
    set_vm
    install_msi
    uninstall_msi
    show
    reboot
    wait_for_start

=cut

use strict;
use warnings;
use IPC::Run qw(start pump finish timeout);

package Mercury::HGAdmin;

# Holds the reference to the singleton object ($self)
my $singleton;

# Location of the hgadmin executable (IPC::Run wants it as reference to array)
# For now, runs out of mauricis/p4 directory (have ticket to move to network)
my $hgadmin = ['/u/mauricis/p4/prod/mercury/main/tools/hgadmin/hgadmin'];

# Setup hgadmin environment, allow ESXi 5 commands to work
# For now, located in mauricis/p4 directory (have ticket to move to network)
$ENV{PERL5LIB} = '/u/mauricis/vmware-perl-sdk-5.0/64bit/vmware-vsphere-cli-distrib/lib/VMware/share';

# Enable autoflush
$| = 1;

## Constructor... Connect/authenticate to a VCenter server
sub new {

    # Get $class, hashed arguments and declare $self
    my ($class, %args) = @_;
    my $self;

    # See if we've already been initialized... Just get the object!
    if ( $singleton ) {
        $self = $singleton;

    } else {

        # Initialize and bless $self
        $self = {
            debug        => 0,
            hgadmin_proc => '',
            hgadmin_in   => '',
            hgadmin_out  => '',
            hgadmin_err  => '',
            host         => '',
            vm           => '',
            is_connected => 0,
        };
        bless ($self, $class);

        # Store the object as the single instance
        $singleton = $self;

        # Check for arguments
        $self->{debug} = 1 if $args{debug};

        # Run the interactive script (pass references to vars)
        $self->{hgadmin_proc} = IPC::Run::start($hgadmin,
            \$self->{hgadmin_in},
            \$self->{hgadmin_out},
            \$self->{hgadmin_err});

        # Send a LF to force an empty ptompt
        $self->send_expect(send => '', expect => '');
    }

    ## Return this object
    return $self;
}

sub set_host {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # See if we got 'host' arg
    if ( $args{host} ) {

        # See if we've set a 'host'
        if ( $self->{host} ) {

            # See if we've set a 'vm'
            if ( $self->{vm} ) {

                # Go 'up' a level (unsets vm)
                $self->up;
            }

            # Go 'up' a level (unsets host)
            $self->up;
        }

        # Run the 'host' command
        $self->{host} = $args{host};
        $self->send_expect(send => "host $args{host}" , expect => "");
    }
}

sub set_vm {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # See if we got 'host' arg
    if ( $args{vm} ) {

        # See if we've set a 'host'
        if ( $self->{host} ) {

            # See if we've set a 'vm'
            if ( $self->{vm} ) {

                # Go 'up' a level (unsets vm)
                $self->up;
            }

            # Run the 'vm' command
            $self->{vm} = $args{vm};
            $self->send_expect(send => "vm $args{vm}" , expect => "");

        }

    }
}

sub uninstall_msi {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # See if we've set a 'host' and a 'vm'
    if ( $self->{host} && $self->{vm} ) {

        # Run the 'uninstall' command
        $self->send_expect(send => "msi uninstall" , expect => "");
    }

    if ( $self->{hgadmin_out} =~ /Reboot required/ ) {
        $self->reboot;
        sleep 15;

        $self->wait_for_start(sleep => 15);
    }
}

sub install_msi {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # See if we've set a 'host' and a 'vm'
    if ( $self->{host} && $self->{vm} ) {

        # Run the 'install' command
        $self->send_expect(send => "msi install $args{msi}" , expect => "");
    }

    if ( $self->{hgadmin_out} =~ /Reboot required/ ) {
        $self->reboot;
        sleep 15;

        $self->wait_for_start(sleep => 15);
    }
}

sub show {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # Run the 'show' command
    $self->send_expect(send => "show" , expect => "");
}

sub reboot {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # Run the 'show' command
    $self->send_expect(send => "restart" , expect => "");
}

sub wait_for_start {

    my $up = 0;

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # Look for sleep param... 15 seconds (default)
    my $sleep = ( $args{sleep} )? $args{sleep}: 15;

    # Look for timeout param... 120 seconds (default)
    my $timeout = ( $args{timeout} )? $args{timeout}: 120;

    my $timer = 0;
    while ( ! $up && $timer <= $timeout) {
        # Run the 'show' command
        $self->send_expect(send => "show" , expect => "");

        my @lines = split(/\n/, $self->{hgadmin_out});
        my @host_fields = split(/\s+/, $lines[2]);
        if ( $host_fields[3] =~ /[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/ ) {
          $up = 1;
        } else {
           sleep $sleep;
           $timer += $sleep;
        }
    }

    return $up;
}

sub up {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # See if we got 'host' arg
    if ( $self->{vm} || $self->{host} ) {
        if ( $self->{vm} ) {
            $self->{vm} = '';
        } else {
            $self->{host} = '';
        }


        # Run the 'up' command
        $self->send_expect(send => "up" , expect => '');
    }
}

sub run {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # See if we've set a 'host' and a 'vm'
    if ( $self->{host} && $self->{vm} ) {

        # Run the 'install' command
        $self->send_expect(send => "run $args{cmd}" , expect => "");
    }
}

sub send_expect {

    # Get $self and hashed arguments
    my ($self, %args) = @_;

    # Build up the expected prompt (grab all text up to prompt)
    my $command_end = "hgadmin";
    $command_end .= " host $self->{host}" if $self->{host};
    $command_end .= " vm $self->{vm}" if $self->{vm};
    $command_end .= "> \n";

    # Add a line-feed to the command, if not blank
    my $send = '';
    $send .= $args{send} . "\n" if $args{send};
    $self->{hgadmin_out} = '';

    # Run the command, add a line-feed for an empty prompt
    $self->{hgadmin_in} = $send . "\n";
    IPC::Run::pump $self->{hgadmin_proc} until $self->{hgadmin_out} =~ /$command_end/;
    print "$self->{hgadmin_out}" if $self->{debug};
}

## Get last error message
sub get_hgadmin_out {

    my $self = shift;

    return $self->{hgadmin_out};
}

## Get last command output
sub get_hgadmin_err {

    my $self = shift;

    return $self->{hgadmin_err};
}

1;
