# $Id: //depot/#1 $
#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc
#
## @summary      Library used to support Vmotion Automation for FA.
##
## @author       nkrajara@netapp.com
##
## @pod          here
##

=head1 NAME

Vmotion.pm

=head1 AUTHOR

Narayan Rajaram

=head1 DESCRIPTION

C<Vmotion> will provide the functionalities to support Vmotion for FA

1. Will provide the functionalities to support Vmotion for FA

=head1 SYNOPSIS

Library provides the functionalities to support Vmotion on FA

=cut

package Mercury::Vmotion;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Tharn;
use Error;
use San::SanLog;
use vars qw(@ISA);
@ISA = qw(Error SanLog);

# Netapp Specific Libs

use NACL::APISet;
use Mercury::ESXi;
use TracePkg;

# Global Variable
my $log = $Tharn::Log;

=head2 new

To instantiate and configure an Vmotion object

=cut

#############################################################################
# Subroutine: new
# Parameters: None
# Returns:    Reference to this object
#############################################################################

sub new {
    my $class = shift(@_);
    my $self;
    my %params = validate(
        @_,
        {   
            host_list => {
                type     => ARRAYREF,
                optional => 0,
                default  => "*",
            },
            ssd_list => {
                type     => ARRAYREF,
                optional => 0,
                default  => "*",
            },
            num_vms => {
                type     => SCALAR,
                optional => 0,
                default  => "*",
            },
        }
    );

    # Copy specified parameters globally into object
    # Might need to pass user credentials also

    $self->{_host_list}  = $params{host_list};
    $self->{_ssd_list}   = $params{ssd_list};
    $self->{_num_vms}    = $params{num_vms}; 

    my @host_list = @{$params{host_list}};
    my @ssd_list = @{$params{ssd_list}};

    my @esxi_list;
    my $esxi;
    # Create ESXi object for each host and append it to the array.
    for my $index (0 .. $#host_list) {
        $log->comment("Creating ESXI objects");
        $esxi = Mercury::ESXi->new(
                   host_name => $host_list[$index],
                   ssd_id    => $ssd_list[$index]
                );
        push(@esxi_list,$esxi);
    }
    $self->{_esxi_list} = \@esxi_list ;

    bless($self, $class);
    $log->comment("Created a VMOTION object\n");
    return $self;

} ## end sub new

=head claim_devices

Claim devices on the hosts

=cut

#############################################################################
# Subroutine: claim_devices
# Parameters: None
#############################################################################

sub claim_devices {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $esx;
    my @esxi_list = @{$self->{_esxi_list}};
    foreach $esx (@esxi_list) {
        $esx->claim_device();
        if($esx->error()) {
            $self->_set_err("Claim error: $esx->errmsg()");
        }
    }
} ## end sub claim_devices

=head unclaim_devices

Unclaim devices on the hosts

=cut

#############################################################################
# Subroutine: unclaim_devices
# Parameters: None
#############################################################################

sub unclaim_devices {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $esx;
    foreach $esx (@{$self->{_esxi_list}}) {
        $esx->unclaim_device();
        if($esx->error()) {
            $self->_set_err("unclaim error: $esx->errmsg()");
        }
    }
} ## end sub unclaim_devices

=head create_vluns

Create Vluns on all the hosts

=cut

#############################################################################
# Subroutine: create_vluns
# Parameters: None
#############################################################################

sub create_vluns {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $num_blocks = $args{blocks};
    my $esx;
    my @naa_list;
    my $naa_id;
    my $index;
    my @esx_list = @{$self->{_esxi_list}};

    # Remove the last esx object and create cachespaces on that first
    # Use the NAA IDs from this host to create VLUNs on other hosts
    my $esx_last = pop(@esx_list);
    for $index (1 .. $self->{_num_vms}) {
        $naa_id = $esx_last->create_vlun(blocks => $num_blocks);
        if($esx_last->error()) {
            $self->_set_err("Vlun Creation Failed: $esx_last->errmsg()");
        }
        push(@naa_list,$naa_id);
    }

    #Now create the same cachespaces on all the other hosts
    foreach $esx (@esx_list) {
        foreach $naa_id (@naa_list) {
            $esx->create_vlun(blocks => $num_blocks, naa_id => $naa_id);
            if($esx->error()) {
                $self->_set_err("unclaim error: $esx->errmsg()");
                $log->error("unclaim error: $esx->errmsg()");
            }
        }
    }
} ## end sub create_vluns

=head destroy_vluns

Destroy all Vluns on all the hosts

=cut

#############################################################################
# Subroutine: destroy_vluns
# Parameters: None
#############################################################################

sub destroy_vluns {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $esx;
    foreach $esx (@{$self->{_esxi_list}}) {
        $esx->remove_all_vluns();
        if($esx->error()) {
            $self->_set_err("Issue in removing cache space: $esx->errmsg()");
        }
    }
} ## end sub destroy_vluns

1;

