#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Provide access to cmdlets that will manage FlashAccel
## @author bhanup@netapp.com
## @status private
## @pod not here

package Mercury::Windows::hgCmdLets;

#
# Compiler Directives
#
use strict;
use warnings;

# NATE imports
#
use Hostrec;
use Tharn qw($Log);
use Params::Validate qw(:all);
use Data::Dumper;

# NACL imports
#
use NACL::APISet;
use NACL::APISet::Exceptions::TimeoutException;
use NATE::BaseException qw(:try);

################################################################################
# Method: new
# Objective: create a new hgCmdLets oject to manage a set of VMs using cmdlets  
# Details: Instanciate a mercury cmdlets skeleton object. setManagedVM is the
#          interface that does majority of the work for dealing with a VM
################################################################################
sub new {
    $Log->enter();
    my ($class, %args) = @_;

    my $self = {};
    bless($self, $class);

    $Log->exit();

    return $self;
} ## end sub new

################################################################################
#  Getters
################################################################################
sub get_dll_loc {
    my $self = shift @_;
    return $self->{dll_loc};
}

sub get_userid {
    my $self = shift @_;
    return $self->{userid};
}

sub get_password {
    my $self = shift @_;
    return $self->{password};
}

sub get_cmd_suffix {
    my $self = shift @_;
    return $self->{cmd_suffix};
}

sub get_powershell {
    my $self = shift @_;
    return $self->{powershell};
}

sub get_managed_vm {
    my $self = shift @_;
    return $self->{managed_vm};
}

sub get_vmname {
    my $self = shift @_;
    return $self->{vmname};
}

################################################################################
# Setters								       #
################################################################################
sub set_dll_loc {
    my ($self, $value) = @_;
    $self->{dll_loc} = $value if ($value);
}

sub set_vmname {
    my ($self, $value) = @_;
    $self->{vmname} = $value if ($value);
}

sub set_userid {
    my ($self, $value) = @_;
    $self->{userid} = $value if ($value);
}

sub set_password {
    my ($self, $value) = @_;
    $self->{password} = $value if ($value);
}

sub set_cmd_suffix {
    my ($self, $value) = @_;
    $Log->enter();

    # TODO...add remoting capabilities ...
    #    } elsif ($param =~ /^-?ComputerName$/) {
    #    } elsif ($param =~ /^-?Credential$/) {
    # for now there is no support for remote address as part of
    # cmdlets
    # $self->{cmd_suffix} = " -ComputerName ".$self->get_vmname() . " -force ";

    # by default always use the force option for cmdlets
    $self->{cmd_suffix} = " -force ";

    $Log->exit();
} ## end sub set_cmd_suffix

sub set_powershell {
    my ($self, $value) = @_;
    $self->{powershell} = $value if ($value);
}

################################################################################
# Method: setManagedVM
# Objective: set the VM connection specifics on which all the cmdlets are 
#            executed
# Details: 
#    This method will take a vm name and setup the hgcmdlets object's
#    access to powershell for that vm. It also maintains a list of
#    VMs that this object can manage, so that we reuse the persistant powershell
#    connection. 
#    So call this method with a VM name1, to send cmds to name1
################################################################################

sub setManagedVM {
    my ($self, $vm_name) = @_;

    $Log->enter();

    my $vm_rec;
    foreach $vm_rec (@{$self->{managed_vm_list}}) {
        if ($vm_rec->{name} eq $vm_name) {
            last;
        }
    }

    unless (defined $vm_rec) {
        my $hg_obj = Hostrec->new(id               => $vm_name,
                                  default_username => "administrator",
                                  default_password => "netapp1!");

        unless ($hg_obj) {
            $Log->comment("Hostrec() call failed!!");
        }

        my $pwrsh_obj = NACL::APISet->new(hostobj   => $hg_obj,
                                            category  => "Host",
                                            interface => "CLI",
                                            set       => "PowerShell",
                                            connid    => 'login',);
        $self->set_powershell($pwrsh_obj);
        $self->set_dll_loc('fa-pscmdlets');

        my $resp_obj;
        try {
            $resp_obj =
              $self->get_powershell()->import_module(name => $self->get_dll_loc());
        }
        catch NACL::APISet::Exceptions::TimeoutException with {
            my $ex_obj = shift;
            $Log->comment($ex_obj->text() . "\n" . $ex_obj->stacktrace());
            try {
                $resp_obj =
                  $self->get_powershell()
                  ->import_module(name => $self->get_dll_loc());
            }
            catch NACL::APISet::Exceptions::TimeoutException with {
                my $ex_obj_2 = shift;
                $Log->comment($ex_obj_2->text() . "\n" . $ex_obj_2->stacktrace());
                $Log->comment("Giving up on the second timeout");
                return;
            }
        } ## end with
        catch NATE::BaseException with {
            my $ex_obj = shift;
            $Log->comment($ex_obj->tesxt(). "\n". $ex_obj->stacktrace());
            return;
        }
        # for each of these hosts all we are interested in getting
        # the cerdentials . But these can be defaulted to administrator/netapp1!
        # in mercury labs
        
        $vm_rec->{powershell} = $pwrsh_obj;
        $vm_rec->{name}     = $hg_obj->hostname();
        $vm_rec->{obj}      = $hg_obj;
        $vm_rec->{userid}   = $hg_obj->default_username();
        $vm_rec->{password} = $hg_obj->default_password();
        $vm_rec->{ip}       = $hg_obj->default_ip();
        push @{$self->{managed_vm_list}}, $vm_rec;
    } ## end unless (defined $vm_rec)

    # update the vm_rec so we know which one is being managed at all times
    $self->{managed_vm} = $vm_rec;

    # for now we will only support one VM being managed by this object
    #.... as of 10/11/12
    # Assmption here is that the tharnhost file for each of hte VM
    # reflects the username and password in the tharnhost file
    # in the default-username and default-password entries
    $self->set_powershell($self->{managed_vm}->{powershell});
    $self->set_vmname($self->{managed_vm}->{name});
    $self->set_userid($self->{managed_vm}->{userid});
    $self->set_password($self->{managed_vm}->{password});

    # now set the cmd suffix
    $self->set_cmd_suffix();
    $Log->trace("cmds suffix is: " . $self->get_cmd_suffix());

    $Log->exit();
} ## end sub setManagedVM

################################################################################
# Method: getCacheFormatProgress
# Objective: Wrapper to get functionality similar to the webservice 
# Details: calls the getHgCache method, returns the formatinfo part of cache
#          properties.
################################################################################
sub getCacheFormatProgress {
    my ($self, %args) = @_;
    my (%ret_hash, $device_ret, $cache_ret);

    $Log->enter();
    
    $cache_ret  = $self->getHgCache(-formatInfo => 1,);
    
    $Log->exit();

    return \%ret_hash;
}

################################################################################
# Method: getCacheAndDeviceInfo
# Objective: Wrapper to get functionality similar to the webservice 
# Details: calls the getHgCache and getHgDevice cmdlets and then returns the 
#          combined hashs from the two calls.
################################################################################
sub getCacheAndDeviceInfo {
    my ($self, %args) = @_;
    my (%ret_hash, $device_ret, $cache_ret);

    $Log->enter();
    
    $cache_ret  = $self->getHgCache(-Detailed => 1,);
    $Log->comment("getCacheAndDeviceInfo cache data: ". Dumper($cache_ret));
    $device_ret = $self->getHgDevice(-Detailed => 1,);
    $Log->comment("getCacheAndDeviceInfo Device data: ". Dumper($device_ret));
    %ret_hash = (%$cache_ret, %$device_ret);
    $Log->comment("getCacheAndDeviceInfo returning this hash: ". Dumper(\%ret_hash));
    
    $Log->exit();

    return \%ret_hash;
}

################################################################################
# Method: setCacheID
# Objective: Wrapper to get functionality similar to the webservice 
# Details: uses the input param CacheID to call set-facache cmdlet
################################################################################
sub setCacheID {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?CacheID$/i) {
            $input_hash{CacheDeviceId} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    }

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_facache(\%input_hash);

    $Log->exit();

    return $response;
}

################################################################################
# Method: setCacheState
# Objective: Wrapper to provide webservice equivalence
################################################################################

sub setCacheState {
    my ($self, %args) = @_;

    $Log->enter();

    my $ret_val = $self->setHgCache(%args);

    $Log->exit();

    return $ret_val;
}

################################################################################
# Method: getHgCache
# Objective: exec the get-FAcache cmdlet and return parsed output 
# Details:
#PS C:\Users\Administrator> get-hgcache -detailed
#
#
#BuildVersion          : 1.0.4502.1417
#CacheDeviceId         : 600a0980cffffff8f5fa5efffc023750
#CacheNaaId            : 600a0980cffffff8f5fa5efffc023750
#CacheAdminState       : Enabled
#CacheOperationalState : Normal
#OperStateMessage      : WriteThru mode
#CacheMode             : WriteThru
#PerfState             : Enabled
#LogicalBlockAddress   : 20447231
#CurrentGenId          : 8
#
#ReplacementPolicy : 1
#BlockSize         : 4096
#BlocksPerChunk    : 128
#Chunks            : 19966
#FormattedChunks   : 19967
#UsableBlocks      : 2535682
#MaxDevices        : 1024
#
#HitLatencyAvg          : 0
#HitLatencyMin          : 0
#HitLatencyMax          : 0
#ChunkReadCount         : 0
#ChunkOverwriteCount    : 0
#ChunkSkipCount         : 0
#FullFlushLatencyAvg    : 0
#FullFlushLatencyMin    : 0
#FullFlushLatencyMax    : 0
#FullFlushCount         : 0
#PartialFlushLatencyAvg : 0
#PartialFlushLatencyMin : 0
#PartialFlushLatencyMax : 0
#PartialFlushCount      : 0
#FailedCacheReadOps     : 0
#FailedFullFlushes      : 0
#FailedPartialFlushes   : 0
#WriteQDepth            : 0
#WriteQDepthAvg         : 0
#WriteQDepthMax         : 0
#LogScanSecs            : 0
#
################################################################################

sub getHgCache {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?Metrics$/i) {
            $input_hash{Metrics} = $args{$param};
        } elsif ($param =~ /^-?Detailed$/i) {
            $input_hash{Detailed} = $args{$param};
        } elsif ($param =~ /^-?FormatInfo$/i) {
            $input_hash{FormatInfo} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    #
    if($input_hash{Detailed}) {
       $input_hash{pass_through} .= " -Detailed ";
       delete $input_hash{Detailed};
    } else {
        if($input_hash{FormatInfo}) {
            $input_hash{pass_through} .= " -FormatInfo ";
            delete $input_hash{FormatInfo};
        } 
        if($input_hash{Metrics}) {
            $input_hash{pass_through} .= " -Metrics ";
            delete $input_hash{Metrics};
        }
    }

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->get_facache(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    my @temp_array = split /\n/, $response->get_raw_output;

    my (%temp_hash);

    for my $one_line (@temp_array){

        #skip empty lines 
        #skip lines with spaces only
        next if ($one_line =~ /^$/);
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        # remove any leading or trailing spaces
        $key   =~ s/^\s+|\s+$//;
        $value =~ s/^\s+|\s+$//;
        $temp_hash{$key} = $value;

        # trying to build a return hash that has this structure
        # $ret_hash  =  { 
        #                    cache  =>  { 
        #                                   properties => 'hash_ref1',
        #                                   formatinfo => 'hash_ref2',
        #                                   metrics    => 'hash_ref3',
        #                               }
        #               }
        # so using the line order in the output array to figure out
        # when to start the next hash_ref to contain the next set of 
        # cache specifics

        if($key eq 'CurrentGenId') {
            my %shallow_copy = %temp_hash;
            $ret_hash{cache}{properties} = \%shallow_copy;
            %temp_hash = ();   
        } elsif($key eq 'MaxDevices') {
            my %shallow_copy = %temp_hash;
            $ret_hash{cache}{formatInfo} = \%shallow_copy;
            %temp_hash = ();   
        } elsif($key eq 'LogScanSecs') {
            my %shallow_copy = %temp_hash;
            $ret_hash{cache}{metrics} = \%shallow_copy;
            %temp_hash = ();   
        }

    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));
    
    $Log->exit();

    return \%ret_hash;

} ## end sub getHgCache

################################################################################
# Method: setHgCache
# Objective: calls the set-FAcache cmdlet
################################################################################
sub setHgCache {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?CacheMode$/i) {
            $input_hash{CacheMode} = $args{$param};
        } elsif ($param =~ /^-?CacheAdminState$/i) {
            $input_hash{CacheAdminState} = $args{$param};
        } elsif ($param =~ /^-?PerfState$/i) {
            $input_hash{PerfState} = $args{$param};
        } elsif ($param =~ /^-?CacheDeviceId$/) {
            $input_hash{CacheDeviceId} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_facache(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    my @temp_array = split /\n/, $response->get_raw_output;

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));
    
    $Log->exit();

    return \%ret_hash;

} ## end sub setHgCache

################################################################################
# Method: setDeviceState
# Objective: provide webservice equivalent using cmdlets
################################################################################
sub setDeviceState {
    my ($self, %args) = @_;

    $Log->enter();

    foreach my $param (keys %args) {
        if ($param =~ /^-?naaids$/i) {
            $args{NaaIds} = $args{$param};
            # now remove the naaids param
            delete $args{$param};
        }
    } 

    my $ret_val = $self->setHgDevice(%args);

    $Log->exit();
    
    return $ret_val;
}

################################################################################
# Method: setHgDevice
# Objective: calls the set-FAdevice cmdlet
################################################################################
sub setHgDevice {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?Disks$/i) {
            $input_hash{Disks} = $args{$param};
        } elsif ($param =~ /^-?NaaIds$/i) {
            $input_hash{NaaIds} = $args{$param};
        } elsif ($param =~ /^-?DeviceAdminState$/i) {
            $input_hash{DeviceAdminState} = $args{$param};
        } elsif ($param =~ /^-DeviceCacheMode$/i) {
            $input_hash{DeviceCacheMode} = $args{$param};
        } elsif ($param =~ /^-?PerfState$/i) {
            $input_hash{PerfState} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_fadevice(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    my @temp_array = split /\n/, $response->get_raw_output;

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));
    
    $Log->exit();

    return \%ret_hash;

} ## end sub setHgDevice

################################################################################
# Method: getHgDevice
# Objective: calls the get-FAdevice cmdlet with the -detailed flag to get all
#            state and metric info
# Details:
#PS C:\Users\Administrator> get-hgdevice -detailed
#
#
#Disk                   : Disk0
#SerialNumber           : 6000c29d8d20f7b609c474cd42d25e76
#VendorId               : VMware
#ProductId              : Virtual disk
#ProductRevision        : 1.0
#NaaId                  : 6000c29d8d20f7b609c474cd42d25e76
#DeviceAdminState       : Disabled
#DeviceOperationalState : Disabled
#OperStateMessage       : Admin disabled.
#CacheMode              : WriteThru
#PerfState              : Disabled
#SCSIAddress            : 00000000
#SectorSize             : 512
#LogicalBlockAddress    : 62914559
#LastGenId              : 0
#ReadOps                : 0
#SkippedReadOps         : 0
#ReadLatencyPerOp       : 0
#ReadBlocks             : 0
#ReadLatencyPerBlock    : 0
#ReadBytes              : 0
#ReadSmallCount         : 0
#ReadMediumCount        : 0
#ReadLargeCount         : 0
#ReadXLargeCount        : 0
#WriteOps               : 0
#SkippedWriteOps        : 0
#WriteLatencyPerOp      : 0
################################################################################
sub getHgDevice {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?Disks$/i) {
            $input_hash{Disks} = $args{$param};
        } elsif ($param =~ /^-?SerialNumbers$/i) {
            $input_hash{SerialNumbers} = $args{$param};
        } elsif ($param =~ /^-?Metrics$/i) {
            $input_hash{Metrics} = $args{$param};
        } elsif ($param =~ /^-?Detailed$/i) {
            $input_hash{Detailed} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    #
    if($input_hash{Detailed}) {
       $input_hash{pass_through} .= " -Detailed ";
       delete $input_hash{Detailed};
    } elsif($input_hash{Metrics}) {
       $input_hash{pass_through} .= " -Metrics ";
       delete $input_hash{Metrics};
    }

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->get_fadevice(\%input_hash);

    my @temp_array = split /\n/, $response->get_raw_output;

    $Log->trace(Dumper(\@temp_array));

    my (%temp_hash, $dealing_with_disk);

    for my $one_line (@temp_array){

        #skip empty lines 
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        # remove any leading or trailing spaces
        $key   =~ s/^\s+|\s+$//;
        $value =~ s/^\s+|\s+$//;
        if($key eq 'Disk') {
            $dealing_with_disk = $value;
            next;
        }
        # update the temp hash   
        $temp_hash{$key} = $value;

        if($key eq 'FailedWriteOps') {
            my %shallow_copy = %temp_hash;
            $ret_hash{$dealing_with_disk}{metrics} = \%shallow_copy;
            %temp_hash = ();   
        } elsif($key eq 'LastGenId') {
            my %shallow_copy = %temp_hash;
            $ret_hash{$dealing_with_disk}{properties} = \%shallow_copy;
            %temp_hash = ();   
        }

    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub getHgDevice

################################################################################
# Method: getCacheOptions
# Objective: Provide webservice equivalent functionality using cmdlets
################################################################################
sub getCacheOptions {
    my ($self, %args) = @_;

    $Log->enter();

    my $ret_hash = $self->getHgOptions(%args);

    $Log->exit();
    
    return $ret_hash;

}

################################################################################
# Method: getHgOptions
# Objective: calls the get-FAoptions cmdlet
################################################################################
sub getHgOptions {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->get_faoptions(\%input_hash);

    my @temp_array = split /\n/, $response->get_raw_output;

    $Log->trace(Dumper(\@temp_array));

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub getHgOptions

################################################################################
# Method: setHgOptions
# Objective: calls the set-FAoptions cmdlet
# Details:
#    Set-HgOptions [-ComputerName <string>] [-Credential <PsCredential>] [-DefaultDeviceState <enabled|disabled>] [-DefaultDeviceCacheMode <PassThru|WriteAround|WriteThru>] [-EventLogLevel <Fatal|Failure|Error|Warning|Informational|Debug>] [-ChecksumEnabled <enabled|disabled>] [-MetricsResetInterval <UInt32>] [-InternalTimeout <UInt32>] [-InternalRetryInterval <UInt32>] [-InternalRetryCount <UInt32>] [-CacheTimeout <UInt32>] [-CacheRetryCount <UInt32>] [-MinReadSize <UInt32>] [-MaxReadSize <UInt32>] [-MinWriteSize <UInt32>] [-MaxWriteSize <UInt32>] [-WriteQLimit <UInt32>] [-BlockUsageThreshold <UInt32>] [-ChunkMinBlockThreshold <UInt32>] [-ChunkMaxBlockThreshold <UInt32>] [<CommonParameters>]
# 
################################################################################
sub setHgOptions {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?DefaultDeviceState$/) {
            $input_hash{DefaultDeviceState} = $args{$param};
        } elsif ($param =~ /^-?DefaultDeviceCacheMode$/) {
            $input_hash{DefaultDeviceCacheMode} = $args{$param};
        } elsif ($param =~ /^-?EventLogLevel$/) {
            $input_hash{EventLogLevel} = $args{$param};
        } elsif ($param =~ /^-?ChecksumEnabled$/) {
            $input_hash{ChecksumEnabled} = $args{$param};
        } elsif ($param =~ /^-?MetricsResetInterval$/) {
            $input_hash{MetricsResetInterval} = $args{$param};
        } elsif ($param =~ /^-?InternalTimeout$/) {
            $input_hash{InternalTimeout} = $args{$param};
        } elsif ($param =~ /^-?InternalRetryInterval$/) {
            $input_hash{InternalRetryInterval} = $args{$param};
        } elsif ($param =~ /^-?InternalRetryCount$/) {
            $input_hash{InternalRetryCount} = $args{$param};
        } elsif ($param =~ /^-?CacheTimeout$/) {
            $input_hash{CacheTimeout} = $args{$param};
        } elsif ($param =~ /^-?CacheRetryCount$/) {
            $input_hash{CacheRetryCount} = $args{$param};
        } elsif ($param =~ /^-?MinReadSize$/) {
            $input_hash{MinReadSize} = $args{$param};
        } elsif ($param =~ /^-?MaxReadSize$/) {
            $input_hash{MaxReadSize} = $args{$param};
        } elsif ($param =~ /^-?MinWriteSize$/) {
            $input_hash{MaxWriteSize} = $args{$param};
        } elsif ($param =~ /^-?WriteQLimit$/) {
            $input_hash{WriteQLimit} = $args{$param};
        } elsif ($param =~ /^-?BlockUsageThreshold$/) {
            $input_hash{BlockUsageThreshold} = $args{$param};
        } elsif ($param =~ /^-?ChunkMinBlockThreshold$/) {
            $input_hash{ChunkMinBlockThreshold} = $args{$param};
        } elsif ($param =~ /^-?ChunkMaxBlockThreshold$/) {
            $input_hash{ChunkMaxBlockThreshold} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();


    # execute the command on the VM
    my $response = $self->get_powershell()->set_faoptions(\%input_hash);

    my @temp_array = split /\n/, $response->get_raw_output;

    $Log->trace(Dumper(\@temp_array));

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub setHgOptions

################################################################################
# Method: setCacheOptions
# Objective: provide webservice equivalent functionality using cmdlets 
################################################################################
sub setCacheOptions {
    my ($self, %args) = @_;

    $Log->enter();

    my $ret_val = $self->setHgOptions(%args);

    $Log->exit();

    return $ret_val;
}

################################################################################
# Method: clearHgMetrics
# Objective: calls the clear-fametrics cmdlet
################################################################################
sub clearHgMetrics {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?Disks$/) {
            $input_hash{Disks} = $args{$param};
        } elsif ($param =~ /^-?SerialNumbers$/) {
            $input_hash{SerialNumbers} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->clear_fametrics(\%input_hash);

    my @temp_array = split /\n/, $response->get_raw_output;

    $Log->trace(Dumper(\@temp_array));

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub clearHgMetrics

################################################################################
# Method: resetHgDevice
# Objective: calls the reset-faDevcie cmdlet
################################################################################
sub resetHgDevice {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?Disks$/) {
            $input_hash{Disks} = $args{$param};
        } elsif ($param =~ /^-?SerialNumbers$/) {
            $input_hash{SerialNumbers} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->reset_fadevice(\%input_hash);

    my @temp_array = split /\n/, $response->get_raw_output;

    $Log->trace(Dumper(\@temp_array));

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub resetHgDevice

################################################################################
# Method: startCacheFormat
# Objective: webservice equivalent using cmdlets
################################################################################

sub startCacheFormat {
    my ($self, %args) = @_;

    $Log->enter();

    my $ret_val = $self->formatHgCache(%args);   

    $Log->exit();

    return $ret_val;

} ## end sub startCacheFormat


################################################################################
# Method: formatHgCache
# Objective: calls teh format-FACache cmdlet
# Details:
#SYNTAX
#    Format-HgCache [-ComputerName <string>] [-Credential <PsCredential>] [-ReplacementPolicy <UInt64>] [-BlockSize <UInt32>] [-BlocksPerChunk <UInt32>] [-sizeLimitGb <UInt32>] [-Async <SwitchParameter>] [<CommonParameters>]
################################################################################
sub formatHgCache {
    my ($self, %args) = @_;

    $Log->enter();

    my (%input_hash, %ret_hash);

    foreach my $param (keys %args) {
        if ($param =~ /^-?ReplacementPolicy$/) {
            $input_hash{ReplacementPolicy} = $args{$param};
        } elsif ($param =~ /^-?BlockSize$/) {
            $input_hash{BlockSize} = $args{$param};
        } elsif ($param =~ /^-?BlocksPerChunk$/) {
            $input_hash{BlocksPerChunk} = $args{$param};
        } elsif ($param =~ /^-?sizeLimitGb$/) {
            $input_hash{sizeLimitGb} = $args{$param};
        } elsif ($param =~ /^-?Async$/) {
            $input_hash{Async} = $args{$param};
        } elsif ($param =~ /^-?pass_through$/i) {
            $input_hash{pass_through} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    if($input_hash{Async}) {
        $input_hash{pass_through} .= "-Async:\$True";
    }

    # now add the credentials and force stuff
    $input_hash{pass_through} .= $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->format_facache(\%input_hash);

    my @temp_array = split /\n/, $response->get_raw_output;

    $Log->trace(Dumper(\@temp_array));

    for my $one_line (@temp_array) {
        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end foreach my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;
} ## end sub formatHgCache

1;

