#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Use WCFClient to interface with the FlashAccel WebService
## @author bhanup@netapp.com
## @status private
## @pod not here

package Mercury::Windows::hgWebService;

#
# Compiler Directives
#
use strict;
use warnings;

# NATE imports
#
use Hostrec;
use Tharn qw($Log);
use Params::Validate qw(:all);
use Data::Dumper;

# NACL imports
#
use NACL::APISet;
use NACL::APISet::Exceptions::TimeoutException;
use NATE::BaseException qw(:try);

################################################################################
# Method: new
# Objective: create a new hgWebService oject
# Details: instantiates a NACL powershell persistent connection object using
#          the WCFClient to create the hostrec object. Then loads the WCFClient
#          dll which enables communication with FA webservice running on the
#          same or another windows VM
################################################################################
sub new {
    $Log->enter();
    my ($class, %args) = @_;

    my ($mercury_vms, $host_obj, $wcfClient_name, $use_ip);

    foreach my $param (keys %args) {
        if ($param =~ /^-?hg_VMs$/i) {
            $mercury_vms = $args{$param};
        } elsif ($param =~ /^-?WCFClient$/i) {
            $wcfClient_name = $args{$param};
        } elsif ($param =~ /^-?use_IP$/i) {
            $use_ip = $args{$param};
        }
    }

    my $self = {};
    bless($self, $class);

    # need to create a APISet to talk to the WCFClient VM
    # then need the set of VMs that WCFclient will interact
    # with. For these hosts weshould be able to either read their
    # tharnhost files or just their credentials and store them
    # for future use.
    # the third usecase is that these credentials can be set
    # per call

    # first deal with the WCFLClient
    my $wcfClient_obj = Hostrec->new(id => $wcfClient_name);
    unless ($wcfClient_obj) {
        $Log->comment("Hostrec failed!! for : " . $wcfClient_name);
        return;
    }

    my $Api_Set_Obj = NACL::APISet->new(hostobj   => $wcfClient_obj,
                                        category  => "Host",
                                        interface => "CLI",
                                        set       => "PowerShell",
                                        connid    => 'login',);
    $self->set_powershell($Api_Set_Obj);
    $self->set_dll_loc('c:\users\administrator\Downloads\WCFClient.dll');

    # set the dns flag, before the call to setManagedVM so that it picks it up
    $self->{use_ip} = $use_ip if ($use_ip);

    $Log->comment("Created an APISet object for the WCFClient communication");

    foreach my $hg_vm (@$mercury_vms) {
        # set the managed_vm_context and update the info for it
        $self->setManagedVM($hg_vm);
        last;
    }

    # now that the connection is there, lets import the dll we need to
    # to use the testclient
    # persistent powershell connection will need to handle the
    # NACL::APISet::Exceptions::TimeoutException exception and
    # then redo the importin of the test client dll
    # irrespective of the number of managed VMs there is only one
    # WCFClient object that needs to import the test client dll
    my $resp_obj;
    try {
        $resp_obj =
          $self->get_powershell()->import_module(name => $self->get_dll_loc());
    }
    catch NACL::APISet::Exceptions::TimeoutException with {
        my $ex_obj = shift;
        $Log->comment($ex_obj->text() . "\n" . $ex_obj->stacktrace());
        try {
            $resp_obj =
              $self->get_powershell()
              ->import_module(name => $self->get_dll_loc());
        }
        catch NACL::APISet::Exceptions::TimeoutException with {
            my $ex_obj_2 = shift;
            $Log->comment($ex_obj_2->text() . "\n" . $ex_obj_2->stacktrace());
            $Log->comment("Giving up on the second timeout");
        }
    } ## end with

    $Log->exit();

    return $self;
} ## end sub new

################################################################################
#  Getters
################################################################################
sub get_dll_loc {
    my $self = shift @_;
    return $self->{dll_loc};
}

sub get_userid {
    my $self = shift @_;
    return $self->{userid};
}

sub get_password {
    my $self = shift @_;
    return $self->{password};
}

sub get_svc_loc {
    my $self = shift @_;
    return $self->{svc_loc};
}

sub get_cmd_suffix {
    my $self = shift @_;
    return $self->{cmd_suffix};
}

sub get_powershell {
    my $self = shift @_;
    return $self->{wcfclient}->{powershell};
}

sub get_managed_vm {
    my $self = shift @_;
    return $self->{managed_vm};
}

################################################################################
# Setters
################################################################################
sub set_dll_loc {
    my ($self, $value) = @_;
    $self->{dll_loc} = $value if ($value);
}

sub set_userid {
    my ($self, $value) = @_;
    $self->{userid} = $value if ($value);
}

sub set_password {
    my ($self, $value) = @_;
    $self->{password} = $value if ($value);
}

sub set_svc_loc {
    my ($self, $value) = @_;
    $self->{svc_loc} = $value if ($value);
}

sub set_cmd_suffix {
    my ($self, $value) = @_;
    $Log->enter();

    $self->{cmd_suffix} =
        " -RemoteAddress "
      . $self->get_svc_loc()
      . " -UserName "
      . $self->get_userid()
      . " -PassWord "
      . $self->get_password();
    $Log->exit();
} ## end sub set_cmd_suffix

sub set_powershell {
    my ($self, $value) = @_;
    $self->{wcfclient}->{powershell} = $value if ($value);
}

sub setManagedVM {
    my ($self, $vm_name) = @_;

    $Log->enter();

    # TODO..
    # need to verify that the $value is a hostrec obj

    # create a list of mercury VMs that this test client object can manage
    # mark the current VM as the active one
    #
    my $vm_rec;
    foreach $vm_rec (@{$self->{managed_vm_list}}) {
        if ($vm_rec->{name} eq $vm_name) {
            last;
        }
    }

    unless (defined $vm_rec) {

        if ($self->{use_ip}) {
            $vm_rec->{name}     = $vm_name;
            $vm_rec->{userid}   = 'administrator';
            $vm_rec->{password} = 'netapp1!';
            $vm_rec->{ip}       = $vm_name;
        } else {
            my $hg_obj = Hostrec->new(id               => $vm_name,
                                      default_username => "administrator",
                                      default_password => "netapp1");

            unless ($hg_obj) {
                $Log->comment("Hostrec() call failed!!");
            }

            # for each of these hosts all we are interested in getting
            # the cerdentials . But these can be defaulted to administrator/netapp1!
            # in mercury labs

            $vm_rec->{name}     = $hg_obj->hostname();
            $vm_rec->{obj}      = $hg_obj;
            $vm_rec->{userid}   = $hg_obj->default_username();
            $vm_rec->{password} = $hg_obj->default_password();
            $vm_rec->{ip}       = $hg_obj->default_ip();
        } ## end else [ if ($self->{use_ip})

        push @{$self->{managed_vm_list}}, $vm_rec;
    } ## end unless (defined $vm_rec)

    # update the vm_rec so we know which one is being managed at all times
    $self->{managed_vm} = $vm_rec;

    $self->set_userid($self->{managed_vm}->{userid});
    $self->set_password($self->{managed_vm}->{password});

    # use the default_ip to rule out any DNS issues
    # but for now use the hostname()
    if ($self->{use_ip}) {
        $self->set_svc_loc('https://' . $self->{managed_vm}->{ip} . ':7010');
    } else {
        $self->set_svc_loc('https://' . $self->{managed_vm}->{name} . ':7010');
    }

    # now set the cmd suffix
    $self->set_cmd_suffix();
    $Log->trace("cmds suffix is: " . $self->get_cmd_suffix());

    $Log->exit();
} ## end sub setManagedVM

################################################################################
# Method: getCacheAndDeviceInfo
# Objective: Call the get-cacheinfo WCFClient interface
# Details:
#   Returns:
#
#$VAR1 = {
#          'Disk6' => {
#                       'metrics' => {
#                                       'WriteXLargeCount' => '1171',
#                                       'ReadLargeCount' => '9440921',
#....
#...
#                                       'AvgHitLatency' => '325'
#                                   },
#                       'properties' => {
#                                           'PerfState' => 'ADMIN_ENABLED',
#                                           'OperationalState' => 'OP_NORMAL',
#....
#...
#          'cache' => {
#                       'metrics' => {
#                                      'PartialFlushLatencyMax' => '0',
#                                      'FullFlushCount' => '4476902',
#......
#....
#                                      'FullFlushLatencyMin' => '0'
#                                    },
#                       'properties' => {
#                                         'BuildVersion' => '1.0.4473.434',
#                                         'PerfState' => 'ADMIN_ENABLED',
#.....
#...
#                                         'CacheMode' => 'CM_WRITETHRU'
#                                       }
#                     },
#          'Disk1' => {
#                       'metrics' => {
#                                      'WriteXLargeCount' => '0',
#
################################################################################

sub getCacheAndDeviceInfo {
    my ($self, %args) = @_;
    my %ret_hash;

    $Log->enter();

    # execute the command on the VM
    my $response =
      $self->get_powershell()
      ->get_cacheinfo(pass_through => $self->get_cmd_suffix());

    $Log->trace(Dumper($response->get_raw_output));

    my @temp_array = split /\n/, $response->get_raw_output;

    my (%temp_hash, $dealing_with_disk);

    for (my $index = 0 ; $index <= $#temp_array ; $index++) {

        my $one_line = $temp_array[$index];

        # skip all lines that start with ExtensionData\s+: but use that
        # pattern as the separation between blocks of information

        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        if ($one_line =~ /^ExtensionData\s+:/ or $index == $#temp_array) {

            # update the ret_hash if the delimiter is hit or if this is
            # the last line in array
            if (%temp_hash) {

                # if we are dealing with a non-empty hash then....

                # Create a shallow copy of the contents
                # from the temp hash
                my %shallow_hash_copy = %temp_hash;

                # now push the new hash into the return_hash
                $ret_hash{cache}{properties} = \%shallow_hash_copy
                  if (exists $shallow_hash_copy{BuildVersion});
                $ret_hash{cache}{metrics} = \%shallow_hash_copy
                  if (exists $shallow_hash_copy{FullFlushCount});

                # this output is dependent on the order in which the
                # webservice outputs the properties and metrics. Any
                # changes to the webservice will mean changes here as
                # well...
                $ret_hash{$dealing_with_disk}{properties} = \%shallow_hash_copy
                  if (exists $shallow_hash_copy{DiskID});
                $ret_hash{$dealing_with_disk}{metrics} = \%shallow_hash_copy
                  if (exists $shallow_hash_copy{Hits});

                # invalidate temp hash contents but do not undefine it
                %temp_hash = ();
            } ## end if (%temp_hash)
            next;
        } else {

            # then take the next n number of lines and parse them into
            # a hash
            my ($key, $value) = split /:/, $one_line;

            # remove any leading or trailing spaces
            $key   =~ s/^\s+|\s+$//;
            $value =~ s/^\s+|\s+$//;
            $temp_hash{$key} = $value;
            $dealing_with_disk = $value if ($key eq 'DiskID');
        } ## end else [ if ($one_line =~ /^ExtensionData\s+:/...
    } ## end for (my $index = 0 ; $index...

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub getCacheAndDeviceInfo

################################################################################
# Method: getCacheOptions
# Objective: Call the get-cacheoptions WCFClient interface
################################################################################

sub getCacheOptions {
    my ($self, %args) = @_;

    $Log->enter();

    my %ret_hash;

    # execute the command on the VM
    my $response =
      $self->get_powershell()
      ->get_cacheoptions(pass_through => $self->get_cmd_suffix());

    $Log->trace(Dumper($response->get_raw_output));

    my @temp_array = split /\n/, $response->get_raw_output;

    for my $one_line (@temp_array) {

        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # skip all lines that start with ExtensionData\s+:
        next if ($one_line =~ /^ExtensionData\s+:/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end for my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub getCacheOptions

################################################################################
# Method: getCacheFormatProgress
# Objective: Call the get-cacheformatprogress WCFClient interface
################################################################################

sub getCacheFormatProgress {
    my ($self, %args) = @_;

    $Log->enter();

    my %ret_hash;

    # execute the command on the VM
    my $response =
      $self->get_powershell()
      ->get_cacheFormatProgress(pass_through => $self->get_cmd_suffix());

    $Log->trace(Dumper($response->get_raw_output));

    my @temp_array = split /\n/, $response->get_raw_output;

    for my $one_line (@temp_array) {

        # skip blank lines
        next if ($one_line =~ /^$/);

        #skip lines with spaces only
        next unless ($one_line =~ /\S/);

        # skip all lines that start with ExtensionData\s+:
        next if ($one_line =~ /^ExtensionData\s+:/);

        # then take the next n number of lines and parse them into
        # a hash
        my ($key, $value) = split /:/, $one_line;
        $key   =~ s/\s+//g;
        $value =~ s/\s+//g;
        $ret_hash{$key} = $value;
    } ## end for my $one_line (@temp_array)

    $Log->trace(Dumper(\%ret_hash));

    $Log->exit();

    return \%ret_hash;

} ## end sub getCacheFormatProgress

################################################################################
# Method: startCacheFormat
# Objective: Call the start-CacheFormat WCFClient interface
################################################################################

sub startCacheFormat {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?ReplacementPolicy$/i) {
            $input_hash{ReplacementPolicy} = $args{$param};
        } elsif ($param =~ /^-?BlockSize$/i) {
            $input_hash{BlockSize} = $args{$param};
        } elsif ($param =~ /^-?BlocksPerChunk$/i) {
            $input_hash{BlocksPerChunk} = $args{$param};
        } elsif ($param =~ /^-?SizeLimitGb$/i) {
            $input_hash{SizeLimitGb} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->start_cacheformat(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub startCacheFormat

################################################################################
# Method: stopCacheFormat
# Objective: Call the stopCacheFormat WCFClient interface
################################################################################

sub stopCacheFormat {
    my ($self, %args) = @_;

    $Log->enter();

    # execute the command on the VM
    my $response =
      $self->get_powershell()
      ->stop_cacheformat(pass_through => $self->get_cmd_suffix());

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub stopCacheFormat

################################################################################
# Method: setCacheState
# Objective: Call the setCacheState WCFClient interface
################################################################################

sub setCacheState {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?CacheAdminState$/i) {
            $input_hash{CacheAdminState} = $args{$param};
        } elsif ($param =~ /^-?CacheMode$/i) {
            $input_hash{CacheMode} = $args{$param};
        } elsif ($param =~ /^-?PerfState$/i) {
            $input_hash{PerfState} = $args{$param};
        }
    }

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_cachestate(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub setCacheState

################################################################################
# Method: setDeviceState
# Objective: Call the setCacheState WCFClient interface
################################################################################

sub setDeviceState {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?DeviceAdminState$/i) {
            $input_hash{DeviceAdminState} = $args{$param};
        } elsif ($param =~ /^-?CacheMode$/i) {
            $input_hash{CacheMode} = $args{$param};
        } elsif ($param =~ /^-?PerfState$/i) {
            $input_hash{PerfState} = $args{$param};
        } elsif ($param =~ /^-?NaaIds$/i) {
            $input_hash{NaaIds} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_devicestate(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub setDeviceState

################################################################################
# Method: setCacheOptions
# Objective: Call the set-CacheOptions WCFClient interface
################################################################################

sub setCacheOptions {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?DefaultDeviceState$/i) {
            $input_hash{DefaultDeviceState} = $args{$param};
        } elsif ($param =~ /^-?DefaultDeviceCacheMode$/i) {
            $input_hash{DefaultDeviceCacheMode} = $args{$param};
        } elsif ($param =~ /^-?ChecksumEnabled$/i) {
            $input_hash{ChecksumEnabled} = $args{$param};
        } elsif ($param =~ /^-?EventLogLevel$/i) {
            $input_hash{EventLogLevel} = $args{$param};
        } elsif ($param =~ /^-?MetricsResetInterval$/i) {
            $input_hash{MetricsResetInterval} = $args{$param};
        }
    } ## end foreach my $param (keys %args)

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_cacheoptions(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub setCacheOptions

################################################################################
# Method: clearMetrics
# Objective: Call the clearMetrics WCFClient interface
################################################################################

sub clearMetrics {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?NaaIds$/i) {
            $input_hash{NaaIds} = $args{$param};
        }
    }

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->clear_metrics(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub clearMetrics

################################################################################
# Method: resetDevice
# Objective: Call the resetDevice WCFClient interface
################################################################################

sub resetDevice {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?NaaIds$/i) {
            $input_hash{NaaIds} = $args{$param};
        }
    }

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->reset_device(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub resetDevice

################################################################################
# Method: setCacheID
# Objective: Call the setCacheID WCFClient interface
################################################################################

sub setCacheID {
    my ($self, %args) = @_;

    $Log->enter();

    my %input_hash;

    foreach my $param (keys %args) {
        if ($param =~ /^-?CacheID$/i) {
            $input_hash{CacheID} = $args{$param};
        }
    }

    # now add the credentials stuff
    $input_hash{pass_through} = $self->get_cmd_suffix();

    # execute the command on the VM
    my $response = $self->get_powershell()->set_CacheID(\%input_hash);

    $Log->trace(Dumper($response->get_raw_output));

    $Log->exit();

} ## end sub setCacheID

1;
