#!/usr/software/bin/perl
# -*- perl -*-
# MERCURY::HGAdmin_Wrapper.pm
# Copyright (c) 2013 Network Appliance, Inc.
# All rights reserved.
## @summary
##      This module is designed to connect with the HgAdmin command-line 
##      interface and can execute the follow commands:
##
##          * vm credentials show
##
##          * vm credentials set
##
##          * vm msi install
##
##          * vm msi uninstall
##
##          * vm cache attach
##
##          * vm cache detach
##
##          * vm show
##
##          * vm msi show
##
##          * vm snapshot show
##
##          * vm snapshot create
##
##          * vm snapshot destroy
##
##          * vm snapshot revert
##
##          * vm start
##
##          * vm stop
##
##          * vm restart
##
#
## @description
##
##      This module is broken into the different functions of HGAdmin,
##      the command-line interface tool. Its functionality is similar
##      Mercury's FAMC web interface tool. 
##      
##
#
## @type
##      library(Perl)      
#
## @status
##      private
#
## @author
##      Natalie Kerby -
##      knatalie@netapp.com ::
##      natalie.kerby@netapp.com
package Mercury::hgadmin_wrapper;
use Data::Dumper;
use NATE::Log qw(log_global);
$Log = log_global();
use Error;
use San::SanLog;
use vars qw(@ISA);
@ISA = qw(Error SanLog);

#############################################################################
#
#
#############################################################################
# Global variables continued
my $hgadmin_path = "/u/knatalie/p4/FlashAccel_win/tools/hgadmin";
#############################################################################
# Method: new
# Objective: create a new hg_cmdlets oject
# Details:
## @usage NEW()
##
##      This constructor will build a hash that associated to each 
##      object. This method will also run the export command that will
##      link the appropriate perl library. 
##
##      The constructor accepts a reference to a hash that includes the 
##      following parameters: 
##      vCenter_name, vCetner_username, vCenter_password, host_name, 
##      vm_name, vm_username, vm_password. 
#
#############################################################################
sub new {
    my ($class, $credentials) = @_;
    my $self = { "vcenter_name" => undef,
                 "vcenter_username" => undef,
                 "vcenter_password" => undef,
                 "host_name" => undef,
                 "vm_name" => undef,
                 "vm_username" => undef,
                 "vm_password" => undef,
                   };
    foreach (keys %$credentials){
        $self->{$_} = $credentials->{$_};
    }
    bless($self, $class);
    if( defined $self->{vm_username} && defined $self->{vm_password}){
        $self->vm_credentials_set();
    }
    return $self;
} ## end sub new
#############################################################################
# The command methods are located below
#
#############################################################################
#############################################################################
# Method: set_export
#############################################################################
sub set_export {
    my ($self, %args) = @_;
    $self->vm_start();
    my $command = "export PERL5LIB=/u/narayanv/mercury/vmware-perl-sdk-5.0/" .
                  "64bit/vmware-vsphere-cli-distrib/apps:" .
                  "/u/narayanv/mercury/vmware-perl-sdk-5.0/" .
                  "64bit/vmware-vsphere-cli-distrib/lib/VMware/share";
    
    return $command;
} ## end sub set_export
#############################################################################
# Method: vm_show
## @usage VM_SHOW()
##
##      This subroutine will retrive a list of all vms located on the  
##      desired host using hgadmin.
##
##      The subroutine will return an array with all the vm names that
##      are listed on the desired host.
#############################################################################
sub vm_show {
    my ($self, %args) = @_;
    my @vm_names = undef;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm show";
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){
        push (@vm_names, $2) 
            if ($index =~ /((([a-zA-Z0-9-_]+ )+) {2,})/ && $2 ne 'VM Name ');
    }
    if(! @vm_names){
        $self->_set_err("ERROR FOUND");
    }
    return @vm_names;
} ## end sub vm_show
#############################################################################
# Method: vm_credentials_set
## @usage VM_CREDENTIALS_SET()
##
##      This subroutine will update the object's vm_name and/or 
##      vm_username & vm_password. It will then set the vm credentials
##      of the object in hgadmin.
##
##      This process only needs to be completed once.
##
##      The subroutine will return the new username and password 
##      associated with the vm.
#############################################################################
sub vm_credentials_set {
    my ($self, %credentials) = @_;
    my %set_credentials;
    foreach (keys %credentials){
        if ($_ =~ /^vm_name$/i) {
            $self->{vm_name} = $credentials{$_};            
        } elsif ($_ =~ /^vm_username$/i) {
            $self->{vm_username} = $credentials{$_};            
        } elsif ($param =~ /^vm_password$/i) {
            $self->{vm_password} = $credentials{$_};
        }  
    }
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm credentials set" .
                  " $self->{vm_name}" .
                  " $self->{vm_username}" .
                  " $self->{vm_password}";
    my $output = `./hgadmin $command`;
    my %credentials = $self->vm_credentials_show();
    return %credentials;
} ## end sub vm_credentials_set
#############################################################################
# Method: vm_credentials_show
## @usage VM_CREDENTIALS_SHOW()
##
##      This subroutine will retrive the current credentials that are set to 
##      the vm in hgadmin.
##
##      The subroutine will return a hash with the username and password 
##      associated with the vm.
#############################################################################
sub vm_credentials_show {
    my ($self, %args) = @_;
    my %credentials;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm credentials show" .
                  " $self->{vm_name}";
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){
        $credentials{'username'} = $2 if ($index =~ /^(Username: )(.*)$/);
        $credentials{'password'} = $2 if ($index =~ /^(Password: )(.*)$/);
    }
    if(! %credentials){
        $self->_set_err("ERROR FOUND");
    }
    return %credentials;
} ## end sub vm_credentials_show
#############################################################################
# Method: get_msi_path
## @usage GET_MSI_PATH()
##
##      This subroutine will get the file and path for the new windows build.
##
##      The subroutine will return a string of the path (with msi filename).
#############################################################################
sub get_msi_path {
    my ($self, $release_name) = @_;
    my $msi_file = '';
    my $path = "/x/eng/rlse/flashaccel_win/";
    my $release_path = $path . "$release_name/final/packages";
    my $std_path = $path . "mainN/final/packages";
    chdir($release_path) or chdir($std_path);
    chomp($file_path = `pwd`);
    chomp (my @file_names = split /\n/, `ls`);
    foreach my $index (@file_names){
        $msi_file = $index if (($index =~ /NetApp_FlashAccel-Win.*\.msi$/));        
    }
    $file_path .= '/' . $msi_file;
    return $file_path;
} ## end sub get_msi_path

#############################################################################
# Method: vm_msi_show
#
## @usage VM_MSI_SHOW()
##
##      This subroutine will retrieve the current install information on the
##      desired vm. 
##      
##      The subroutine will return a string containing all information about 
##      the install.
#############################################################################
sub vm_msi_show {
    my ($self, %args) = @_;
    my $install_info;
    chdir($hgadmin_path) or die;
    my $export_command= $self->set_export();
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm msi show".
                  " $self->{vm_name}";
    chomp ( my @output = `$export_command; ./hgadmin $command` );
    foreach my $index (@output){
        $install_info = "$index\n" 
            if ($index =~ /not installed/);  
        $install_info .= "$index\n" 
            if ($index =~ /^Name:/ || 
                $index =~ /^Version:/ || 
                $index =~ /^Install Date:/ || 
                $index =~ /^Package Name:/);
    }
    if(! defined $install_info){
        $self->_set_err("ERROR FOUND");
    }
    return $install_info;
} ## end sub vm_msi_show
#############################################################################
# Method: vm_msi_install
## @usage VM_MSI_INSTALL()
##
##      This subroutine will retrive the most recent .msi file and
##      install on the desired vm using hgadmin. 
##      
##      This subroutine accepts a string of the current release name
##
##      The subroutine will return the successful 
##      install message
#############################################################################
sub vm_msi_install {
    my ($self, %release_name) = @_;
    my $install;
    my $file_path = $self->get_msi_path($release_name{release_name});
    my $export_command= $self->set_export();
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm msi install --reboot" .
                  " $self->{vm_name}" .
                  " $file_path" ;
    chomp ( my @output = `$export_command; ./hgadmin $command` );
    foreach my $index (@output){
        $install = "$index\n" 
            if ($index =~ /Successfully/);
    }
    if(! defined $install){
        $self->_set_err("ERROR FOUND");
    }
    return $install;
} ## end sub vm_msi_install
#############################################################################
# Method: vm_msi_uninstall
## @usage VM_MSI_UNINSTALL()
##
##      This subroutine will uninstall Flash Accel on the desired 
##      vm using hgadmin. 
##
##      The subroutine will return the successful/unsuccessful 
##      uninstall message
#############################################################################
sub vm_msi_uninstall {
    my ($self, %args) = @_;
    my $uninstall;
    chdir($hgadmin_path) or die;
    my $export_command= $self->set_export();
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm msi uninstall" .
                  " $self->{vm_name}";
    chomp ( my @output = `$export_command; ./hgadmin $command` );
    foreach my $index (@output){
        $uninstall = "$index\n" 
            if ($index =~ /Successfully/);
    }
    if(! defined $uninstall){
        $self->_set_err("ERROR FOUND");
    }
    return $uninstall;
} ## end sub vm_msi_uninstall
#############################################################################
# Method: vm_cache_attach
## @usage VM_CACHE_ATTACH()
##
##      This subroutine will attach a desired cache to the desired vm. 
##      
##      The subroutine requires an input of an appropriate NAAID.
##
##      The subroutine will return the successful/unsuccessful 
##      attachment of the cache.
#############################################################################
sub vm_cache_attach {
    my ($self, %naaid) = @_;
    my $cache_attach_return;
    if(!%naaid){
        $Log->comment("-------------------");
        $Log->comment("Missing NaaID");
        $Log->comment("-------------------");
        return "Missing NaaId";
    }
    chdir($hgadmin_path) or die;
    #attach [--disable] <vm_name> <canonical_name> [ds_name]
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm cache attach";
    if($naaid{disable}){
        $command .= " --disable";
    }
    $command .=   " $self->{vm_name}" .
                  " $naaid{naaid}";             
    if($naaid{ds_name}){
        $command .= " $naaid{ds_name}";
    }
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){
        $cache_attach_return .= "$index\n" 
            if ($index =~ /$self->{vm_name}/);
    }
    if(! defined $cache_attach_return){
        $self->_set_err("ERROR FOUND");
    }
    return $cache_attach_return;
} ## end sub vm_cache_attach
#############################################################################
# Method: vm_cache_detach
## @usage VM_CACHE_DETACH()
##
##      This subroutine will detach the cache from the desired 
##      vm using hgadmin. 
##
##      The subroutine will return the successful/unsuccessful 
##      detach message
#############################################################################
sub vm_cache_detach {
    my ($self, %args) = @_;
    my $cache_detach_return;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm cache detach" .
                  " $self->{vm_name}";
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){
        $cache_detach_return = $index 
            if ($index =~ /\'$self->{vm_name}\'/);
    }
    if(! defined $cache_detach_return){
        $self->_set_err("ERROR FOUND");
    }
    return $cache_detach_return;
} ## end sub vm_cache_detach
#############################################################################
# Method: vm_snapshot_show
## @usage VM_SNAPSHOT_SHOW()
##
##      This subroutine will retrive a list of all snapshots located on the  
##      desired vm using hgadmin.
##
##      The subroutine will return an array with all the snapshot names that
##      are listed on the desired vm.
#############################################################################
sub vm_snapshot_show {
    my ($self, %args) = @_;
    my @snapshot_names;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm snapshot show" .
                  " $self->{vm_name}";
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){
        push (@snapshot_names, $2) 
            if ($index =~ /((([a-zA-Z0-9-_]+ )+) {7,})/ && $2 ne 'Name ');
    }
    if(! @snapshot_names){
        $self->_set_err("ERROR FOUND");
    }
    return @snapshot_names;
} ## end sub vm_snapshot_show

#############################################################################
# Method: vm_snapshot_create
## @usage VM_SNAPSHOT_CREATE()
##
##      This subroutine will create a snapshot with the desired name on the 
##      desired vm using hgadmin. 
##
##      The subroutine will return the successful/unsuccessful 
##      snapshot message.
#############################################################################
sub vm_snapshot_create {
    my ($self, $new_snapshot_name) = @_;
    my $create_return;
    chdir($hgadmin_path) or die;
    $self->vm_stop();
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm snapshot create" .
                  " $self->{vm_name}" .
                  " $new_snapshot_name";
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){  
        $create_return = $index if (($index =~ /Success/i));
    }
    $self->vm_start();
    if(! defined $create_return){
        $self->_set_err("ERROR FOUND");
    }
    return $create_return;
} ## end sub vm_snapshot_create

#############################################################################
# Method: vm_snapshot_destroy
## @usage VM_SNAPSHOT_DESTROY()
##
##      This subroutine will destroy a snapshot with the desired name on the 
##      desired vm using hgadmin. 
##
##      The subroutine will return the successful/unsuccessful 
##      snapshot message.
#############################################################################
sub vm_snapshot_destroy {
    my ($self, $destroy_snapshot_name) = @_;
    my $destroy_return;
    chdir($hgadmin_path) or die;
    $self->vm_stop();
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm snapshot destroy" .
                  " $self->{vm_name}" .
                  " $destroy_snapshot_name" ;
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){
        $destroy_return = $index if (($index =~ /Success/i));
    }
    $self->vm_start();
    if(! defined $destroy_return){
        $self->_set_err("ERROR FOUND");
    }
    return $destroy_return;
} ## end sub vm_snapshot_destroy
#############################################################################
# Method: vm_snapshot_revert
## @usage VM_SNAPSHOT_REVERT()
##
##      This subroutine will revert a snapshot with the desired name on the 
##      desired vm using hgadmin. 
##
##      The subroutine will return the successful/unsuccessful 
##      snapshot message.
#############################################################################
sub vm_snapshot_revert {
    my ($self, $revert_snapshot_name) = @_;
    my $revert_return;
    chdir($hgadmin_path) or die;
    $self->vm_stop();
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm snapshot revert" .
                  " $self->{vm_name}".
                  " $revert_snapshot_name" ;
    chomp ( my @output = `./hgadmin $command` );
    foreach my $index (@output){       
        $revert_return = $index if (($index =~ /Success/i));
    }
    $self->vm_start();
    if(! defined $revert_return){
        $self->_set_err("ERROR FOUND");
    }
    return $revert_return;
} ## end sub vm_snapshot_revert
#############################################################################
# Method: vm_start
## @usage VM_START()
##
##      This subroutine will start/boot the desired 
##      vm using hgadmin. 
##
##      The subroutine will not return anything 
#############################################################################
sub vm_start{
    my ($self, %args) = @_;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm start" .
                  " $self->{vm_name}";
    
    my $output = `./hgadmin $command`;
    sleep(30);
    return;
} ## end sub vm_start
#############################################################################
# Method: vm_stop
## @usage VM_STOP()
##
##      This subroutine will stop/shutdown the desired 
##      vm using hgadmin. 
##
##      --force can be used to do a force/quick shutdown (like
##      pulling the plug)
##
##      The subroutine will not return anything 
#############################################################################
sub vm_stop{
    my ($self, $force) = @_;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm stop" .
                  " $self->{vm_name}";
    if ($force) {
        $command .= " -force";
    }
    my $output = `./hgadmin $command`;
    sleep(30);
    return;
} ## end sub vm_stop

#############################################################################
# Method: vm_restart
## @usage VM_RESTART()
##
##      This subroutine will restart the desired 
##      vm using hgadmin. 
##
##      The subroutine will not return anything 
#############################################################################
sub vm_restart{
    my ($self, $force) = @_;
    chdir($hgadmin_path) or die;
    my $command = " -s $self->{vcenter_name}" .
                  " -u $self->{vcenter_username}" .
                  " -p $self->{vcenter_password}" .
                  " -h $self->{host_name}" .
                  " vm restart" .
                  " $self->{vm_name}";
    if ($force) {
        $command .= " -force";
    }
    my $output = `./hgadmin $command`;
    sleep(30);
    return;
} ## end sub vm_restart
1;

=pod

=head1 NAME

Mercury::hg_wrapper.pm

=head2 Author

Natalie Kerby
knatalie@netapp.com

=head1 DESCRIPTION

This module connect with the HgAdmin command-line interface and 
can execute the follow commands:

=head2 Included Commands

=over 4

=item * vm show

=item * vm credentials show

=item * vm credentials set

=item * vm msi install

=item * vm msi uninstall

=item * vm msi show

=item * vm cache attach

=item * vm cache detach

=back

=head1 vm credentials show

=head1 DESCRIPTION

This subroutine will return the current credentials set for the 
desired vm.

=head1 vm credentials set

=head1 DESCRIPTION

This subroutine will set the desired credentials for the 
desired vm.

=head1 vm msi install

=head1 DESCRIPTION

This subroutine will get the most recent .msi file and install 
it on the desired vm.

=head1 vm msi uninstall

=head1 DESCRIPTION

This subroutine will uninstall Flash Accel on the desired vm.

=cut

