# -*- perl -*-
# $Id$

## @summary Mhost::SubtestUtils
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

Mhost::SubtestUtils - 

=head1 SYNOPSIS

  use Mhost::SubtestUtils;

=head1 DESCRIPTION

This module 

Note: 

=cut

package Mhost::SubtestUtils;

use strict;
use NATE::OS;
use Tharn qw($Log);
use Data::Dumper;
use Connectrec qw();
use Hostrec qw();
use NATE::BaseException qw(:try);
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF CODEREF);

# Index to be used when creating runids
our $main_subtest_index = 0;

#
# Usage:
#
#   Mhost::SubtestUtils::wait_until_finished( subtests = \@subs );
#
# Parameters: 
#
#    subtests             => ARRAYREF of NATE Subtest objects (required)
#    worst_allowable      => NATE::Result Worst allowable result (default NATE::Result::NA)
#    errors               => ARRAYREF into which error strings will be stored (optional)
#    empty_errors_before  => SCALAR Should errors arrayref be emptied before subtests are checked? (default 1)
#    empty_subtests_after => SCALAR Should subtests arrayref be emptied when wait_until_finished() returns? (default 1)
#       
# Return: 
# -----------------------------------------------
sub wait_until_finished {
    my $me=(caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
                             # Any result with a severity greater than this will be
                             # considered a failure (i.e. Inspect and above)
            worst_allowable      => { isa => 'NATE::Result', default  => 'NATE::Result::NA' },
            subtests             => { type => ARRAYREF },
            errors               => { type => ARRAYREF, default => undef },
            die_if_errors        => { type => SCALAR,   default => 1 },
            empty_errors_before  => { type => SCALAR,   default => 1 },
            empty_subtests_after => { type => SCALAR,   default => 1 },
            timeout              => { type => SCALAR,   default => 0 },
        },
    );
    my $rc = 1;
    my $error_string;

    my $worst_allowable = $opt{worst_allowable};

    if ( $opt{empty_errors_before} ) {
        if ( $opt{errors} ) {
            @{$opt{errors}} = ();
        }
    }

    Subtest::wait_finish( subtests => $opt{subtests}, timeout => $opt{timeout} );

    foreach my $s (@{$opt{subtests}}) {
        my @subResults = $s->own_results();
        my $worst = NATE::Result::worst(@subResults);

        if ( $worst &&
            ($worst->isa( 'NATE::Result' ) ) &&
            ($worst->severity() > $worst_allowable->severity()) ) {

            $rc = 0;

            # Add result strings into errors array
            if ( $opt{errors} ) {
                push( @{$opt{errors}}, "Runid: ".$worst->runid()." Messages: ".join(',', $worst->messages()) );
            } elsif ( $rc ) {
                # If they don't want to accumulate all errors, at least save the first one
                $error_string = "Runid: ".$worst->runid()." Messages: ".join(',', $worst->messages());
            }
        }
    }

    if ( $opt{empty_subtests_after} ) {
        @{$opt{subtests}} = ();
    }

    if ( !$rc ) {
        my $errMsg = $error_string;
        if ( $opt{errors} ) {
            $errMsg = join('\n', @{$opt{errors}});
        }
        if ( $opt{die_if_errors} ) {
            die "$me failed with $errMsg";
        } else {
            $Log->warn("$me failed with $errMsg");
        }
    }

    return $rc;
}


#
# Usage:
#
#   Mhost::SubtestUtils::subtest_subref( subref        => $create_subref,
#                                        data          => \@data,
#                                        at_a_time     => $opt_href->{PARALLEL_SUBTESTS},
#                                        die_if_errors => 0,
#                                        errors        => \@errors,
#                                        topdir_prefix => 'create_node_backup',
#                                        subdir_prefix => 'create_backup' );
#
# Parameters: 
#
#    subref               => CODEREF of routine to be called (required)
#    data                 => ARRAYREF of data to be passed to the coderef
#                              If you want to call subref 10 times, there should be 10
#                              elements in data and those elements can be an array of data,
#                              or a hash of data (or if your coderef only takes a single
#                              item, the items can be scalars)
#    num                  => SCALAR Number of times to call subref.  Not needed if you are
#                              passing in "data"
#    die_if_errors        => NATE::Result Worst allowable result (default NATE::Result::NA)
#    worst_allowable      => NATE::Result Worst allowable result (default NATE::Result::NA)
#    errors               => ARRAYREF into which error strings will be stored (optional)
#    empty_errors_before  => SCALAR Should errors arrayref be emptied before subtests are checked? (default 1)
#    empty_subtests_after => SCALAR Should subtests arrayref be emptied when wait_until_finished() returns? (default 1)
#       
# Return: 
# -----------------------------------------------
sub subtest_subref {
    my $me=(caller(0))[3];
    my %opt = validate_with(
        params => \@_,
        spec   => {
            subref               => { type => CODEREF },
            data                 => { type => ARRAYREF, default => [] },
            num                  => { type => SCALAR,   default => undef },
            at_a_time            => { type => SCALAR,   default => 2 },
            die_if_errors        => { type => SCALAR,   default => 1 },
            errors               => { type => ARRAYREF, default => undef },
            topdir_index         => { type => SCALAR,   default => undef },
            topdir_prefix        => { type => SCALAR,   default => "topdir" },
            subdir_prefix        => { type => SCALAR,   default => "subdir" },
        },
    );

    my $subref        = $opt{subref};
    my $topdirpref    = $opt{topdir_prefix};
    my $subdirpref    = $opt{subdir_prefix};
    my $errors        = $opt{errors};
    my $die_if_errors = $opt{die_if_errors};
    my $num           = $opt{num};
    my $at_a_time     = $opt{at_a_time};
    my @data;
    my $rc            = 1;

    if ( $opt{data} ) {
        @data = @{$opt{data}};
        if ( !defined( $num ) ) {
            $num = scalar(@data);
        }
    }

    my $sub_count   = 0;
    my $runid_count = 0;
    my $pending     = 0;
    my @subtests;
    my %wait_opts  = (
                         die_if_errors  => $die_if_errors,
                     );

    if ( $errors ) {
        $wait_opts{errors}              = $opt{errors};
        $wait_opts{empty_errors_before} = 0;
    }

    if ( defined( $opt{topdir_index} ) ) {
        $main_subtest_index = $opt{topdir_index};
    } else {
        $main_subtest_index++;
    }

    my $dir    = sprintf( "%04d", $main_subtest_index )."_".$topdirpref;

    for ( my $i = 1; $i <= $num; $i++ ) {
        $runid_count++;
        $sub_count++;
        $pending = 1;

        my $runid  = $dir.'/'.sprintf( "%04d", $runid_count )."_".$subdirpref;

        my @entry;
        if ( @data ) {
            my $entry = shift( @data );
            @entry = @$entry;
        }
        my $st = Tharn::subtest( $subref,
                                 '-bg',
                                 '-onexit', sub { },
                                 '-runid', $runid,
                                 '--',
                                 @entry );
        push( @subtests, $st );

        # FWR Danny brought up a good point, if you have M tests to be run
        # N at a time, it might be better to constantly check for tests
        # that have completed, and then start up a new test from the queue.
        # For example, if you had 10 tests, rather than doing 2, waiting for
        # both to finish, then do do more, and so on, maybe do 2, then when the
        # first of those completes, start another without first waiting for the
        # second one to complete.
        if ( $sub_count >= $at_a_time ) {
            $sub_count = 0;
            $pending   = 0;
            $rc &= Mhost::SubtestUtils::wait_until_finished( subtests => \@subtests,
                                                             %wait_opts );
        }
    }

    if ( $pending ) {
        $rc &= Mhost::SubtestUtils::wait_until_finished( subtests  => \@subtests,
                                                         %wait_opts );
    }

    return $rc;
}

1;
