# -*- perl -*-
# $Id$

## @summary Mhost::TCDUtils
## @author  dl-coati-core
## @cf     false
## @status public
## @pod here

=head1 NAME

Mhost::TCDUtils - 

=head1 SYNOPSIS

  use Mhost::TCDUtils;

=head1 DESCRIPTION

This module 

Note: 

=cut

package Mhost::TCD;

use strict;
use Options;
use TCD;
use Tharn qw($Log);
use Data::Dumper;
use cluster::enterprise::lib::CalcUtils qw(normalize_time);
use Params::Validate
    qw(validate validate_with BOOLEAN SCALAR ARRAYREF HASHREF OBJECT SCALARREF);
use Mhost::SubtestUtils;

BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter TCD);
}

my $Options = new Options;

our $TestcaseIndex = 0;
our %TestcaseIndex;
our %TestcaseSteps;

sub new {
   my $class = shift;
   my @args  = @_;
   my $self = $class->SUPER::new( @args );
   # bless object
   bless($self, $class );
#print "$class new() self = ".Dumper($self)."\n";
   return $self;
}

#
# Usage:
#
#   $Test->run_for_duration( duration => 60 );
#
# Parameters: 
#
#    duration             => number of seconds to run TCD tests
#    worst_allowable      => NATE::Result Worst allowable result (default NATE::Result::NA)
#    errors               => ARRAYREF into which error strings will be stored (optional)
#    empty_errors_before  => SCALAR Should errors arrayref be emptied before subtests are checked? (default 1)
#    empty_subtests_after => SCALAR Should subtests arrayref be emptied when wait_until_finished() returns? (default 1)
#       
# Return: 
# -----------------------------------------------
sub run_for_duration {
    my $me=(caller(0))[3];
    my $Test = shift;
    my %opt = validate_with(
        params => \@_,
        spec   => {
            duration        => { type => SCALAR, default => 60 },
            stop_if_errors  => { type => SCALAR, default => 1 },
            sleep_between   => { type => SCALAR, default => 0 },
            max_iterations  => { type => SCALAR, default => undef },
        },
    );

    my @results;
    my $duration       = $opt{duration};
    my $stop_if_errors = $opt{stop_if_errors};
    my $sleep_between  = $opt{sleep_between};
    my $max_iterations = $opt{max_iterations};

    my $start_time     = time();
    my $iteration      = 0;
    my $last_iteration = 0;
    my $fail_count     = 0;

    my @orig_tests       = @{$Test->{tests}};
    my @orig_options     = @{$Test->{options}};
    my ($found_no_setup, $found_no_cleanup, $found_setup, $found_cleanup) = (0,0,0,0);

    my @loop_options;
    my @first_options;
    my @last_options;

    foreach my $opt (@{$Test->{options}}) {
        next if($opt !~ /^-/);
        if ($opt eq '-nsc') {
            $found_no_setup   = 1;
            $found_no_cleanup = 1;
        } elsif ( $opt eq '-ns') {
            $found_no_setup   = 1;
        } elsif ( $opt eq '-nc') {
            $found_no_cleanup = 1;
        } elsif ( $opt eq '-st') {
            $found_setup = 1;
            push( @loop_options, $opt );
        } elsif ( $opt eq '-cl') {
            $found_cleanup = 1;
            push( @loop_options, $opt );
        } elsif ( $opt eq '-rsc') {
            $found_setup = 1;
            $found_cleanup = 1;
            push( @loop_options, $opt );
        } else {
            push( @loop_options, $opt );
        }
    }

    @last_options  = @loop_options;
    @first_options = @loop_options;

    # We don't want the loop testcases to keep running setup and cleanup
    # If the user did NOT specify to, or not to run setup and cleanup
    if ( !($found_no_setup   && $found_setup &&
           $found_no_cleanup && $found_cleanup) ) {
        push( @first_options, '-nc' );
        push( @loop_options,  '-nsc' );
        push( @last_options,  '-ns' );
    # If the user did NOT specify to, or not to run setup
    } elsif ( !($found_no_setup && $found_setup) ) {
        push( @loop_options, '-ns' );
        push( @last_options, '-ns' );
    # If the user did NOT specify to, or not to run cleanup
    } elsif ( !($found_no_cleanup && $found_cleanup) ) {
        push( @first_options, '-nc' );
        push( @loop_options,  '-nc' );
    }

    $Test->{options} = \@first_options;

REPEAT:

    # Clear out any persistent flags that control how setup/cleanup are called
    foreach my $flag (qw(bRepeatSetup bSetup bCleanup 
                         bNoCleanup bNoSetup bNoSetClean 
                         bQuiet bNoPre bNoPost)) {
        if ( exists $Test->{$flag} ) {
            delete $Test->{$flag};
        }
    }

    # Clear out TestcaseIndex and Steps so that subtest indices increment
    foreach my $test (@orig_tests) {
        if ( exists( $TestcaseIndex{$test} ) ) {
            delete $TestcaseIndex{$test};
        }

        if ( exists( $TestcaseSteps{$test} ) ) {
            delete $TestcaseSteps{$test};
        }
    }

    $iteration++;

    # Performs method callbacks
    $Test->run_test();
    if ($Test->error) {
        $Test->nlog($Test->errmsg());
        $fail_count++;
        if ( $stop_if_errors ) {
            $Test->nlog("$me: called with stop_if_errors => 1 and errors were encountered");
            $duration = 0;
        }
    }

    push( @results, @{$Test->{result}} );

    if ( $max_iterations && ($iteration > $max_iterations) ) {
        $Test->nlog("$me: called with max_iterations => $max_iterations, now stopping");
        $duration = 0;
    }

    my $now                  = time();
    my $elapsed              = $now - $start_time;
    my $avg_runtime          = $elapsed/$iteration;
    my $elapsed_plus_average = $elapsed + $avg_runtime;

    # Is there enough time to do at least one more iteration?
    # Note, we can't predict the future.  If the next iteration would
    # take 1/2 as long as the current iteration, we could squeeze another in
    # but we can only go with what we know.
    if ( !$last_iteration && ($elapsed_plus_average  < $duration) ) {
        # The run_test() method keeps tacking on setup/cleanup each time through the loop
        # so we just want to remove what it is tacking on
        $Test->{tests}   = \@orig_tests;
        $Test->{options} = \@loop_options;
        if ( $sleep_between ) {
            $Test->nlog("$me: called with sleep_between => $sleep_between");
            Tharn::sleep( $sleep_between );
        }
        goto REPEAT;
    } elsif ( $last_iteration == 0 ) {
        if ( !grep(/^-nc$/, @last_options ) && 
             defined(&{*main::.'cleanup'}) ) {
            my $hRes = {};
            $hRes = $Test->_run_method('cleanup');
            # If not one of these "ok" statuses
            if ( !( ($hRes->{status} == $TCD::PASS) ||
                    ($hRes->{status} == $TCD::WARN) ||
                    ($hRes->{status} == $TCD::NA) ) ) {
                $fail_count++;
            }

            @{$Test->{result}} = @results;
            push( @{$Test->{result}}, $hRes );

            # Display overall summary of test results
            $Test->_run_summary($Test->{result});
        }
    }

    if ( $fail_count > 0 ) {
        return $TCD::FAIL;
    }

    return $TCD::PASS;
}

sub duration_to_seconds {
    my $duration = shift;

    my $seconds = normalize_time( time => $duration );
    
    $Options->set( DURATION => $seconds );
}

sub subtest_subref {
    my $me=(caller(0))[3];
    my $Test = shift;
    my %opt = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec   => {
            testcase_name   => { type => SCALAR,   default => undef },
            topdir_prefix   => { type => SCALAR,   default => "topdir" },
        },
    );

    my $testcase_name;
    my $testcase_index;
    if ( $opt{testcase_name} ) {
        $testcase_name = $opt{testcase_name};
    } else {
        $testcase_name = $Test->get_testcase();
    }

    delete $opt{testcase_name};

    if ( !exists $TestcaseIndex{$testcase_name} ) {
        $TestcaseIndex++;
        $TestcaseIndex{$testcase_name} = $TestcaseIndex;
    }
    if ( !exists $TestcaseSteps{$testcase_name} ) {
        $TestcaseSteps{$testcase_name} = [];
    }

    push( @{$TestcaseSteps{$testcase_name}}, $opt{topdir_prefix} );

    my $step_index = scalar( @{$TestcaseSteps{$testcase_name}} );
    my $topdir = $testcase_name."/".sprintf( "%04d", $step_index )."_".$opt{topdir_prefix};

    $testcase_index     = $TestcaseIndex{$testcase_name};
    $opt{topdir_index}  = $testcase_index;
    $opt{topdir_prefix} = $topdir;

    return Mhost::SubtestUtils::subtest_subref( %opt );
}

1;
