# $Id: //depot/prod/test/Rsierranevada/lib/MiniXFUN/Dispatch.pm#1 $
#
# TODO: Improve path handling. The code uses splitpath inconsistently.
# TODO: Make sure that $CONFIG matches the type of $IMAGE.
# TODO: Give access to $CONFIG via CLI and CGI.
# TODO: Improve error messages.

package MiniXFUN::Dispatch;

require Exporter;
@ISA       = qw(Exporter);
@EXPORT_OK = qw(add_request);

use strict;
use warnings;

use Cwd qw(cwd);
use File::Spec::Functions
  qw(rel2abs catdir catfile catpath splitdir splitpath updir);
use POSIX qw(strftime);

use lib "/usr/software/test/noarch/lib";
use ndisp::API;

use MiniXFUN::Config qw($default_testbed testbed_config);

my $ntest =
  exists $ENV{NTEST_BINARY} && -x $ENV{NTEST_BINARY}
  ? $ENV{NTEST_BINARY}
  : '/usr/software/test/share/nate/current/bin/ntest';

sub image_match($$) {
    my ($path, $match_path) = @_;
    my ($vol,       $dirs,       $filename)       = splitpath($path);
    my ($match_vol, $match_dirs, $match_filename) = splitpath($match_path);
    my @dirs       = splitdir($dirs);
    my @match_dirs = splitdir($match_dirs);

    # Check whether the image filename is correct.
    return undef unless $filename eq $match_filename;

    # Check whether the image directories have a valid suffix.
    return undef if @dirs < @match_dirs;
    while (@match_dirs) {
        my $dir       = pop(@dirs);
        my $match_dir = pop(@match_dirs);
        return undef unless $dir eq $match_dir;
    }

    # The image path matches. Return the path minus the suffix.
    return catpath($vol, catdir(@dirs), '');
} ## end sub image_match($$)

sub image_search($$) {
    my ($path, $search_path) = @_;
    my ($vol, $dirs) = splitpath($path, 'no file');
    my ($search_vol, $search_dirs, $search_filename) = splitpath($search_path);
    my @dirs        = splitdir($dirs);
    my @search_dirs = splitdir($search_dirs);

    # Search for a suffix of $search_path.
    my @test_dirs;
    while ('forever') {
        my $test_dirs = catdir(@dirs, @test_dirs);
        my $test_path = catpath($vol, $test_dirs, $search_filename);
        return $test_path if -f $test_path;
        last unless @search_dirs;
        unshift(@test_dirs, pop(@search_dirs));
    }

    # Not found.
    return undef;
} ## end sub image_search($$)

sub add_request(%) {
    my %arg = @_;

    # Read testbed configuration.
    $arg{TESTBED_DIR} = $default_testbed unless exists $arg{TESTBED_DIR};
    my %config;
    eval { %config = testbed_config($arg{TESTBED_DIR}); };
    return undef if $@ ne '';

    # Collect testbed parameters, except as overridden by add_request() args.
    while (my ($param, $value) = each %config) {
        $arg{$param} = $value unless exists $arg{$param};
    }
    my $maytag_filename = "maytag_kern.$arg{MAYTAG_SUFFIX}";
    my $converged_path =
      catfile('final', 'bedrock', 'export', $arg{CONVERGED_ARCH}, 'tarball',
              'image.tgz');
    my $classic_path = catfile('prod', 'build', $maytag_filename);

    # If the caller supplies a directory instead of a system image, try to find
    # the actual image. (If the caller didn't supply anything, start from the
    # current working directory.)
    $arg{IMAGE} = rel2abs(exists($arg{IMAGE}) ? $arg{IMAGE} : cwd());
    if (-d $arg{IMAGE}) {
        my $path = $arg{IMAGE};
        my $image;

        # Look for a converged tarball, then a classic (maytag) kernel.
        $image = image_search($path, $converged_path) unless defined $image;
        $image = image_search($path, $classic_path)   unless defined $image;
        if (defined($image)) {
            $arg{IMAGE} = $image;
        } else {
            $@ = "$path: Cannot find image.tgz or $maytag_filename\n";
            return undef;
        }
    } ## end if (-d $arg{IMAGE})

    # Validate system image path.
    unless (-f $arg{IMAGE}) {
        $@ = "$arg{IMAGE}: No such file or directory\n";
        return undef;
    }
    my ($vol, $dirs, $file) = splitpath($arg{IMAGE});
    if ($file eq 'image.tgz') {
        $arg{IMAGE_SUFFIX} = $converged_path;
        $arg{CONFIG} = 'converged' unless exists $arg{CONFIG};
    } elsif ($file eq $maytag_filename) {
        $arg{IMAGE_SUFFIX} = $classic_path;
        $arg{CONFIG} = 'classic' unless exists $arg{CONFIG};
    } else {
        $@ = "$arg{IMAGE}: Unsupported system image type\n";
        return undef;
    }
    my $build_dir = image_match($arg{IMAGE}, $arg{IMAGE_SUFFIX});
    unless (defined $build_dir) {
        $@ = "$arg{IMAGE}: Unsupported system image type\n";
        return undef;
    }
    $arg{BUILD_DIR} = $build_dir;

    # Validate configuration
    unless (exists($arg{CONFIG}) && $arg{CONFIG} =~ /^[[:alnum:]]+$/) {
        $@ = "$arg{CONFIG}: Invalid configuration ID\n";
        return undef;
    }
    my $startup_filename = "mini-xfun-$arg{CONFIG}.ntest";
    my $startup_path     = catfile($arg{TESTBED_DIR}, $startup_filename);
    my $release_filename = "mini-xfun-$arg{CONFIG}.release";
    my $release_path     = catfile($arg{TESTBED_DIR}, $release_filename);
    unless (-e $startup_path && -e $release_path) {
        $@ = "$arg{CONFIG}: Invalid configuration ID\n";
        return undef;
    }

    # User name and e-mail address
    $arg{username} = $ENV{LOGNAME} unless exists $arg{username};
    $arg{USER} = $arg{username};    # ndisp swallows $arg{username}
    $arg{MAIL_TO} = "$arg{username}\@netapp.com" unless exists $arg{MAIL_TO};
    $arg{MAILTO}  = $arg{MAIL_TO};

    # Request ID
    my $timestamp = strftime("%Y%m%d-%H%M%S", localtime());
    $arg{id} = "$arg{id}-$arg{username}-$timestamp";
    $arg{jobname}   = '%r';         # Request ID is already unique
    $arg{TEST_NAME} = '%j';         # Request ID is already unique

    # Workload
    my $stest_file;
    if (exists $arg{WORKLOAD}) {
        if ($arg{WORKLOAD} !~ /^[[:alnum:]]+$/) {
            $@ = "$arg{WORKLOAD}: Invalid workload ID\n";
            return undef;
        }
        $stest_file = "mini-xfun-$arg{WORKLOAD}.stest";
    } else {
        $stest_file = "mini-xfun.stest";
    }

    # Command
    $arg{command} = "$ntest -hw -cf -console -s $startup_filename "
      . "STEST_FILE=$stest_file stest/stest";

    my $api = ndisp::API->new($arg{NDISP_ADDR});
    my $result = $api->call("add request", \%arg);
    return defined $result ? $arg{id} : undef;
} ## end sub add_request(%)

1;
