# $Id: //depot/prod/test/main/lib/Model.pm#1 $

#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary Get filer Model number
## @author swati.bajaj@netapp.com, dl-central-libs
## @status shared
## @pod here

package Model;

BEGIN {
    require Exporter;
    @ISA = qw(Exporter);
    @EXPORT_OK = qw(get_model);    
}

use strict;
use warnings;
use XML::XPath;
use Tharn;
use NACL::APISet::Version;
use NACL::APISet::Tools::ZapiUtils::NaServer;

##############################################################################
# Method:    get_model
# Objective: To get the model number of the given filer/node
# Details:   See POD documentation at the end of this file.
##############################################################################

sub get_model {

    $Log->enter();
    my ($hostid, $tharnhost_path) = @_;
    my $error; 
    # Push the tharnhost_path to NATE_HOSTS if specified
    if (defined $tharnhost_path) {
        param('NATE_HOSTS', '-set', param('NATE_HOSTS').";$tharnhost_path");
    }
    my $modelinfo;
    my $host = host($hostid);
    # Get the model number directly from the hostrec object if present.
    # This avoids connecting to the filer/node unnecessarily
	
	$modelinfo = $host->MODEL();
    if ($modelinfo) {
        $Log->debug("Model information is obtained from Tharnhost");
        my $modelinfo = $host->{MODEL};
        return $modelinfo;
    } else {
        $Log->debug("Model information could not be obtained from Tharnhost."
            . " Trying other possibilities");

        # We will use ZAPI to get model number.
        my $hosttype = $host->{hosttype};
        my $shell_type = ($hosttype =~ /ng/)? "CMode" : "7Mode";
        eval {
            $modelinfo = _get_model_info_zapi($host, $shell_type); #does only zapi
        };

        $error = $@;
        if ($error) {
            $Log->debug("_get_model_info_zapi() returned $error");
            $@ = undef;
            $Log->debug("Trying Version manager using default connections\n");
            # try getting the model number using NACL Version manager (default connections).
            eval {
                $modelinfo = _get_model_info_cli($host, $shell_type);
            };
            $error = $@;
            if ($error) {
                $Log->debug("_get_model_info_cli() with default connections returned $error");
                # try getting the model number using NACL Version manager (connections overrides).
                $@ = undef;
                $Log->debug("Finally trying Version Manager with conection"
                    . " overrides\n");
                my ($connid, $conntype);
                ($connid, $conntype) = $shell_type eq "7Mode" ? ("7-login","libssh-persist") : ("c-exec", "libssh-persay");
                eval {
                    $modelinfo = _get_model_info_cli($host, $shell_type, $connid, $conntype);
                };
            }
        }
    }
    
    if (!$modelinfo) {
        die "Could not get modelinfo after trying all possibilities";
    }
    
    $@ = undef;

    $Log->exit();
    return $modelinfo;                          
}

# private method which invokes zapi to get the model number
sub _get_model_info_zapi {
    my ($host, $shell_type) = @_;
    my $command = $shell_type eq "7Mode" ? "system-get-info" : "system-get-node-info-iter";
    my $major_default = 1;
    my $minor_default = 0;
    my $modelinfo;
    
    my $NaServer_Obj = new NACL::APISet::Tools::ZapiUtils::NaServer ($host->default_ip(),$major_default,$minor_default);
    $NaServer_Obj->set_admin_user($host->default_username(),$host->default_password());
    my $NaElement = $NaServer_Obj->invoke($command);
    
    if ($NaElement->results_status() =~ /failed/) {
        die ("Could not invoke $command");
    }

    my $obj = XML::XPath->new($NaElement->sprintf());
    my @nodes = $obj->findnodes('//system-model');
    
    if(scalar @nodes) {
        $modelinfo = $nodes[0]->string_value;
    } else {
        die ("Could not get child nodes with system-model tag");
    }
    
    if (!$modelinfo) {
        die ("Could not get model number" . "Zapi reply was: " . $NaElement->results_reason());
    }

    return $modelinfo;
}

# private method which invokes NACL Version manager object to get the model number
sub _get_model_info_cli {
    my $host = shift @_;
    my $shell_type = shift @_; 
    my ($connid, $conntype) = @_;
    my $modelinfo;

    my $version_manager = NACL::APISet::Version->new( hostobj => $host,
                                                      category => "Node",
                                                      shell_type => $shell_type,
                                                      conn_overrides => {connid => $connid, conntype => $conntype});

    eval {
        $modelinfo = $version_manager->get_version_attribute( attribute => "modelinfo");
    };
    

    if (!$modelinfo) {
        die ("Could not get model number using Version Manager");
    }

    return $modelinfo;
}

1;

=head1 NAME

Model - Get model numbers of the filers

=head1 SYNOPSIS

use Model qw(get_model);

=head1 DESCRIPTION

This library defines methods to get the model number of the 
filer/node/vsim specified.

=head1 FUNCTIONS

=over

=item get_model

$model_number = get_model($FILER);
    or
$model_number = get_model($FILER,$natehost_path);

B<get_model> returns the model number of the node that has been 
specified in the $FILER parameter. get_model dies if it cannot get
the model number.

get_model first tries to obtain the model number from the natehost 
file specified in the MODEL parameter. If the information is not 
obtained, then zapi calls are made to get the model number. If this 
fails too, then NACL Version Manager is used to get the information.

=item Arguments

=over

=item $FILER

(mandatory)
Hostid of the filer/node/vsim for which the model number should be obtained.

=item $natehost_path

(Optional)
The path to the nate_host file corresponding to the $FILER parameter.
If not provided, it is assumed that the path is stored in the standard 
location under $NATE_HOSTS. 

=back

=item Returns

The model number.

=back

=head1 AUTHOR

swati.bajaj@netapp.com, dl-central-libs
 
=head1 STATUS

Shared
 
=head1 COPYRIGHT

Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut
