# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/ModelInfo.pm#1 $

## @summary Provides attributes of various filers/caches based on their model names or code names
## @author dl-central-libs
## @status Shared
## @pod here


=head1 NAME

ModelInfo - Provide attributes of various filers/caches based on their model names or code names

=head1 SYNOPSIS

   use ModelInfo;
   my $mi = new ModelInfo($modelname);

=head1 DESCRIPTION

Provide attributes of various filers/caches based on their model names or code names

Refer http://web.netapp.com/engineering/platformsw/Documentation/xref/xref.htm for
more information on filer models.
Refer http://wikid.netapp.com/w/NANEQA/VSERIES for information on V-series
(V-series is an evolution of gateways)
Refer http://web.netapp.com/engineering/design-depot/hw_platforms/projects/Deux/IBM/IBM-Deux%20Branding%2008jun05.pdf
and http://mfgeng-web.netapp.com/NPIEng/IBM/4
for information on IBM model names and machine types

It takes one argument, $modelname, which is the Model Name (e.g. "F840") or 
Code Name (e.g. "Rodan") or Machine Type Name (e.g. "IBM-2863-A10") of an
appliance.

The object created may have the following fields.  Any attributes
which could not be determined will be nonexistent.  Any model
which is completely unknown will produce an object with no fields.

=over

=item appliance

"filer", "netcache", "nearstore" , "gateway" or "sa_appliance".  Note that for some IBM
models, the model name (such as "N5200") is insufficient to determine
the appliance type.  If this field is needed, the Machine Name (such
as "IBM-2864-A10" or "IBM-2864-G10") should be used instead when
available.

=item arch

"alpha", "x86-64", "x86", "pc_elf" or "mips"

=item bootcard

boolean, whether this host has a flash card for booting

=item floppy

boolean, whether this host has a floppy disk for booting

=item firmware

"ofw", "bios", or "cfe"

=item boot

"ofw", "netboot", "grub", or "cfe".  These names
correspond to flavors of prompts that lib/boot.pm knows about;
see it for more info.

=item processors

number of processors on the appliance).  Note that hyperthreaded processors
count as multiple processors instead of 1 (for example StarGate is listed
as having 4 processors.

=item brand: 

"netapp" or "ibm".  Brand of machine.

=back

=cut


package ModelInfo;

use strict;
use warnings;

use constant filer => qw(appliance filer);
use constant netcache => qw(appliance netcache);
use constant nearstore => qw(appliance nearstore);
use constant gateway => qw(appliance gateway);
use constant sa_appliance => qw(appliance sa_appliance);

use constant alpha => qw(arch alpha);
use constant x86 => qw(arch x86);
use constant mips => qw(arch mips);
use constant pc_elf => qw(arch pc_elf);
use constant x86_64 => qw(arch x86-64);

use constant bootcard => qw(bootcard 1 floppy 0);
use constant floppy => qw(bootcard 0 floppy 1);
use constant nb_only => qw(bootcard 0 floppy 0);

use constant ofw => qw(firmware ofw boot ofw);
use constant netboot => qw(firmware bios boot netboot);
use constant grub => qw(firmware bios boot grub);
use constant cfe => qw(firmware cfe boot cfe);

use constant uni => qw(processors 1);
use constant mp => qw(processors 2);
use constant mp4 => qw(processors 4);
use constant mp8 => qw(processors 8);

use constant netapp => qw(brand netapp);
use constant ibm => qw(brand ibm);


my %Table = (
         # GX vmware simulator
         (map { $_ => [filer, pc_elf, bootcard, cfe, uni, netapp] } qw(SIMBOX VSIM)),

         # Rebranded to IBM (Nxxx) - IBM machine type encodes the appliance type
         # Machine type mapping at: http://mfgeng-web.netapp.com/NPIEng/IBM/4
         # http://mktg-web.netapp.com/products/Quote_Tool_Config_Rules/IBMConfigInfo/NTAPP-to-IBM-HWconfig-Info/4
         (map { $_ => [filer, pc_elf, bootcard, cfe, uni, ibm] } qw(N3300)), # only has filer configurations
         (map { $_ => [filer, pc_elf, bootcard, cfe, uni, ibm] }   qw(IBM-2859-A10 IBM-2859-A20)),   # FAS2020 Shuar-L 
         (map { $_ => [filer, pc_elf, bootcard, cfe, uni, ibm] } qw(N3600)), # only has filer configurations
         (map { $_ => [filer, pc_elf, bootcard, cfe, uni, ibm] }   qw(IBM-2862-A10 IBM-2862-A20)),   # FAS2050 Shuar-M
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp, ibm] } qw(N3400)), # only has filer configurations
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp, ibm] }   qw(IBM-2859-A11 IBM-2859-A21)),    # FAS2040 Shuar-XL
         (map { $_ => [filer, mips, bootcard, cfe, mp, ibm] }    qw(N3700)), # only has filer configurations
         (map { $_ => [filer, mips, bootcard, cfe, mp, ibm] }      qw(IBM-2863-A10 IBM-2863-A20)),   # FAS270
         (map { $_ => [pc_elf, bootcard, cfe, mp, ibm] }  qw(N5200)), # IBM machine types define the appliance
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp, ibm] }    qw(IBM-2864-A10 IBM-2864-A20)),   # FAS3020
         (map { $_ => [gateway, pc_elf, bootcard, cfe, mp, ibm] }  qw(IBM-2864-G10 IBM-2864-G20)),   # V3020
         (map { $_ => [pc_elf, bootcard, cfe, mp, ibm] }  qw(N5300)), # IBM machine types define the appliance 
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, ibm] }    qw(IBM-2869-A10 IBM-2869-A20)),   # FAS3040 Wahoo
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, ibm] }  qw(IBM-2869-G10 IBM-2869-G20)),   # V3040
         (map { $_ => [pc_elf, bootcard, cfe, mp4, ibm] } qw(N5500)), # IBM machine types define the appliance
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp4, ibm] }   qw(IBM-2865-A10 IBM-2865-A20)),   # FAS3050
         (map { $_ => [gateway, pc_elf, bootcard, cfe, mp4, ibm] } qw(IBM-2865-G10 IBM-2865-G20)),   # V3050
         (map { $_ => [pc_elf, bootcard, cfe, mp4, ibm] } qw(N5600)), # IBM machine types define the appliance 
         (map { $_ => [filer, x86_64, bootcard, cfe, mp4, ibm] }   qw(IBM-2868-A10 IBM-2868-A20)),   # FAS3070   
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp4, ibm] } qw(IBM-2868-G10 IBM-2868-G20)),   # V3070     
         (map { $_ => [x86, bootcard, cfe, mp, ibm] }     qw(N7600)), # IBM machine types define the appliance
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, ibm] }    qw(IBM-2866-A10 IBM-2866-A20)),   # FAS6030
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, ibm] }  qw(IBM-2866-G10 IBM-2866-G20)),   # V6030
         (map { $_ => [x86, bootcard, cfe, mp, ibm] }     qw(N7700)), # IBM machine types define the appliance
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, ibm] }    qw(IBM-2851-A10 IBM-2851-A20)),   # FAS6040
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, ibm] }  qw(IBM-2851-G10 IBM-2851-G20)),   # V6040
         (map { $_ => [x86, bootcard, cfe, mp4, ibm] }    qw(N7800)), # IBM machine types define the appliance
         (map { $_ => [filer, x86_64, bootcard, cfe, mp4, ibm] }   qw(IBM-2867-A10 IBM-2867-A20)),   # FAS6070
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp4, ibm] } qw(IBM-2867-G10 IBM-2867-G20)),   # V6070
         (map { $_ => [x86, bootcard, cfe, mp8, ibm] }    qw(N7900)), # IBM machine types define the appliance
         (map { $_ => [filer, x86_64, bootcard, cfe, mp8, ibm] }   qw(IBM-2852-A10 IBM-2852-A20)),   # FAS6080
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp8, ibm] } qw(IBM-2852-G10 IBM-2852-G20)),   # V6080
         (map { $_ => [x86_64, bootcard, cfe, mp, ibm] }  qw(N6040)), # IBM machine types define the appliance
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, ibm] }    qw(IBM-2869-A11 IBM-2869-A21)),   # FAS3140
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, ibm] }  qw(IBM-2869-G11 IBM-2869-G21)),   # V3140
         (map { $_ => [x86_64, bootcard, cfe, mp4, ibm] } qw(N6070)), # IBM machine types define the appliance
         (map { $_ => [filer, x86_64, bootcard, cfe, mp4, ibm] }   qw(IBM-2868-A11 IBM-2868-A21)),   # FAS3170
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp4, ibm] } qw(IBM-2868-G11 IBM-2868-G21)),   # V3170

         # Spectre and derivatives
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, netapp] }    qw(FAS3140)), # Spectre Low
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, netapp] }  qw(V3140)), # Spectre Low gateway
         (map { $_ => [filer, x86_64, bootcard, cfe, mp4, netapp] }   qw(FAS3170)), # Spectre High
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp4, netapp] } qw(V3170)), # Spectre High gateway

         # Tsantsa and derivatives
         # Note : There is no IBM gateway model
         (map { $_ => [filer, mips, bootcard, cfe, mp, netapp] } qw(Tsantsa FAS250 Jivaros FAS270)),
         (map { $_ => [filer, pc_elf, bootcard, cfe, uni, netapp] } qw(FAS2010 FAS2020 FAS2050)),
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp, netapp] } qw(FAS2040)), 
         (map { $_ => [gateway, mips, bootcard, cfe, mp, netapp] } qw(GF270)),
         (map { $_ => [netcache, mips, bootcard, cfe, mp, netapp] } qw(Csantsa)),
         (map { $_ => [filer, mips, nb_only, cfe, mp, netapp] } qw(Swarm)),

         # Deux and derivatives
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp4, netapp] } qw(FAS3050 FAS3050-M)),
         (map { $_ => [gateway, pc_elf, bootcard, cfe, mp4, netapp] } qw(V3050)),
         (map { $_ => [nearstore, pc_elf, bootcard, cfe, mp4, netapp] } qw(R230)),
         (map { $_ => [filer, pc_elf, bootcard, cfe, mp, netapp] } qw(DEUX FAS3020)),
         (map { $_ => [gateway, pc_elf, bootcard, cfe, mp, netapp] } qw(V3020)),
         (map { $_ => [nearstore, pc_elf, bootcard, cfe, mp, netapp] } qw(R210)),
         (map { $_ => [netcache, pc_elf, bootcard, cfe, uni, netapp] } qw(C2300 C3300 C6300)),
         
         # Excelsior derivatives in a Deux sized head http://wikid.netapp.com/w/Platform_SW/DrWho/faq
         (map { $_ => [filer, x86_64, bootcard, cfe, mp4, netapp] } qw(FAS3070)),    # Dr Who
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp4, netapp] } qw(V3070 V3070A)),
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, netapp] } qw(FAS3040)),    # Wahoo
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, netapp] } qw(V3040 V3040A)),
         
         # Excelsior and derivatives
         (map { $_ => [filer, x86_64, bootcard, cfe, mp, netapp] } qw(FAS6030 FAS6030A FAS6040 FAS6040A)),
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp, netapp] } qw(V6030 V6030A V6040 V6040A)),
         (map { $_ => [filer, x86_64, bootcard, cfe, mp4, netapp] } qw(EXCELSIOR FAS6070 FAS6070A)),
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp4, netapp] } qw(V6070 V6070A)),
         (map { $_ => [filer, x86_64, bootcard, cfe, mp8, netapp] } qw(FAS6080 FAS6080A)),
         (map { $_ => [gateway, x86_64, bootcard, cfe, mp8, netapp] } qw(V6080 V6080A)),
         
         # StoreVault Product Family - Puma, Lynx, Panther
         (map { $_ => [filer, x86, bootcard, grub, uni, netapp] } qw(S500 S300 S550)),
         
         # XFiler and derivatives
         (map { $_ => [filer, x86, bootcard, ofw, uni, netapp] } qw(FAS920 FAS940)),
         (map { $_ => [gateway, x86, bootcard, ofw, uni, netapp] } qw(GF940)),
         (map { $_ => [filer, x86, bootcard, ofw, mp, netapp] } qw(XFiler FAS9xx FAS960)),
         (map { $_ => [gateway, x86, bootcard, ofw, mp, netapp] } qw(GF960)),
         (map { $_ => [filer, x86, bootcard, ofw, mp4, netapp] } qw(StarGate FAS980)),
         (map { $_ => [gateway, x86, bootcard, ofw, mp4, netapp] } qw(GF980)),
         (map { $_ => [netcache, x86, bootcard, ofw, mp, netapp] } qw(C6200)),
         (map { $_ => [nearstore, x86, bootcard, ofw, mp, netapp] } qw(Cujo R150 Dogsled R200)),

         # Rodan and derivatives
         (map { $_ => [filer, x86, ofw, netapp] } qw(F8xx)),
         (map { $_ => [filer, x86, floppy, ofw, uni, netapp] } qw(Rodan F840 Minya F820 BladeRunner F810)),
         (map { $_ => [filer, x86, bootcard, ofw, mp, netapp] } qw(Gigan F880)),
         (map { $_ => [filer, x86, bootcard, ofw, uni, netapp] } qw(Duplo F825)),
         (map { $_ => [gateway, x86, bootcard, ofw, uni, netapp] } qw(GF825)),
         (map { $_ => [netcache, x86, floppy, ofw, uni, netapp] } qw(TopShelf C6100 Volcano C3100 C3xxx)),
         (map { $_ => [nearstore, x86, bootcard, ofw, mp, netapp] } qw(BigDog R100)),

         # Various 1U or 2U netcache
         (map { $_ => [netcache, x86, floppy, netboot, uni, netapp] } qw(Calypso C1100 C1105 C1100B)),
         (map { $_ => [netcache, x86, bootcard, grub, uni, netapp] } qw(MadDog C1200 C1300 Thunderbird C2100)),

         # Dell OEM (F8x, C8x)
         (map { $_ => [x86, floppy, netboot, uni, netapp] } qw(Tin)),
         (map { $_ => [filer, x86, floppy, netboot, uni, netapp] } qw(F85)),
         (map { $_ => [netcache, x86, floppy, netboot, uni, netapp] } qw(C85)),
         (map { $_ => [x86, bootcard, grub, uni, netapp] } qw(Rolex)),
         (map { $_ => [filer, x86, bootcard, grub, uni, netapp] } qw(F87)),
         (map { $_ => [netcache, x86, bootcard, grub, uni, netapp] } qw(C87)),
         
         # Gamera (F7xx)
         (map { $_ => [alpha, floppy, ofw, uni, netapp] } qw(Gamera)),
         (map { $_ => [filer, alpha, floppy, ofw, uni, netapp] } qw(F720 F740 F760 F7xx)),
         (map { $_ => [netcache, alpha, floppy, ofw, uni, netapp] } qw(C720 C720s C760 C7xx)),

         # SA Appliance (SA2xx, SA3xx, SA6xx)
         (map { $_ => [sa_appliance, pc_elf, bootcard, cfe, uni, netapp] } qw(SA200)),
         (map { $_ => [sa_appliance, x86_64, bootcard, cfe, mp4, netapp] } qw(SA300 SA300A SA310)),
         (map { $_ => [sa_appliance, x86_64, bootcard, cfe, mp8, netapp] } qw(SA600)),

        );

sub new {
    my ($proto, $model) = @_;
    my $class = ref($proto) || $proto;
    my $model_no_c = $model;
    $model_no_c =~ s/c$//;
    my $entry = $Table{$model_no_c};
    my $self = $entry ? {@{$entry}} : {};
    $self->{cluster} = ($model_no_c ne $model) ? 1 : 0;
    bless($self,$class);
    return $self;
}

sub get_Table {
	return \%Table;
}

1;


=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-central-libs
	$Id: //depot/prod/test/Rsierranevada/lib/ModelInfo.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

