# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Module/Common.pm#1 $

## @summary Common Idioms For NetApp Test Library 
## @author  dl-test-infra@netapp.com
## @status  Shared
## @pod

# XXX - WARNING WARNING WARNING
#
# Due to the nature of this module, Module::Common should NOT have any
# circular references (at compile time) to other NetApp modules.
# Please try very hard not to to include NetApp modules via use().
# Use require() at appropriate times instead.
#
# XXX - WARNING WARNING WARNING

package Module::Common;
use strict;
use warnings;
use base qw(Exporter);
BEGIN
{
	our $VERSION     = '0.01';
	our %EXPORT_TAGS = (
		validate => [ qw(validate_mixed) ],
#		raise => [ qw(eval_or_raise_error) ]
	);
	$EXPORT_TAGS{all} = [ map { @$_ } values %EXPORT_TAGS ];
	Exporter::export_ok_tags('all');
}
use Params::Validate qw(validate_pos validate_with ARRAYREF HASHREF);
use Carp qw(carp croak confess);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
			if(defined &Tharn::script_result) {
			    confess(Tharn::script_result(@msgs));
			} else {
			    confess(join(" ", @msgs));
			}}
        );
}


=head1 NAME

Module::Common - Common Idioms For NetApp Test Library 

=head1 SYNOPSIS

  package MyPackage;
  use Module::Common qw(validate_mixed);

  package MyPackage;
  use Module::Common;
  BEGIN
  {
     Module::Common::setup_export_tags(
        tag1 => [ qw( Symbols For Tag1 ) ],
        tag2 => [ qw( Symbols For Tag2 ) ]
     );
  }

=head1 DESCRIPTION

This module contains some oft-used idioms when writing modules.

=head1 FUNCTIONS

=head2 setup_export_tags(%tags)

This is a utility routine that will setup %EXPORT_TAGS and @EXPORT_OK
for you. This function is NOT exported. Use it via Module::Common::setup_export_tags().

  NOTE: YOU SHOULD CALL THIS FUNCTION FROM A BEGIN{} BLOCK

If you make sure to call this from a BEGIN block, BEFORE importing any
other NetApp modules, you will be safe from potential compile time errors
caused by circular relationships between modules.

The %tags argument is exactly what you would set %EXPORT_TAGS with,
except if you use this you don't have to declare any of those
variables in your module ;) On top of fixing up %EXPORT_TAGS and @EXPORT_OK
variables, this function also does the following:

- Sets up @ISA array of the calling package. If @ISA doesn't contain
some sort of Exporter, 'Exporter' will be pushed into the array so
you don't have to do it yourself.

- Sets up 'all' tag, which will contain all the symbols that you specified
in the tags, so that you can use ':all' to import everything with one call:

  use MyModule qw(:all);

=cut

# wrapper around the whole %EXPORT_TAGS and @EXPORT_OK business.
# This uses heavy voodoo with indirect access to another package's
# variables.
sub setup_export_tags
{
	my(%args) = @_;

    my $pkg = caller();

    # XXX - if the calling package's @ISA doesn't contain
    # 'Exporter', set it up for convenience
    if (!UNIVERSAL::isa($pkg, 'Exporter')) {
        no strict 'refs';
        push @{ "${pkg}::ISA" }, 'Exporter';
    }

    # get the caller's %EXPORT_TAGS and @EXPORT_OK
    my $caller_export_tags = eval {
        no strict 'refs';
        \%{ "${pkg}::EXPORT_TAGS" };
    };
    my $caller_export_ok = eval {
        no strict 'refs';
        \@{ "${pkg}::EXPORT_OK" };
    };

    if(exists $args{all}) {
        croak("tag name 'all' should not be used with this routine");
    }

    $caller_export_tags->{all} = [];

    while( my($tag, $symbols) = each %args) {
        if (exists $caller_export_tags->{ $tag }) {
            carp( "\%${pkg}::EXPORT_TAGS already contains tag '$tag'. Cowardly refusing to process this tag" );
            next;
        }
        $caller_export_tags->{ $tag } = [ @$symbols ];
        push @{ $caller_export_tags->{all} }, @$symbols;
    }

    push @{ $caller_export_ok }, @{ $caller_export_tags->{all}};
}

=head2 validate_mixed(@args,\@positional,\%named)

This is an attempt to make it easier to parse and validate argument list
that is consisted of mandatory positional parameters and optional
named parameters all at the same time.

Params::Validate doesn't allow you to do this (and for good reason), but
since this seems to be a very common idiom, well, here's the best effort
that we have :)

The arguments C<@positional> and C<%named> are exactly what you would
pass to C<Params::Validate::validate_pos()> and C<Params::Validate::validate()>,
respectively.  For more reference on how to setup the validation specs,
please see the Params::Validate perldoc.

For example, in Params::Validate, to specify two positional parameters
where one is scalar and other is an arrayref, you would write:

  my($a1, $a2) = validate_pos(@_, { type => SCALAR }, { type => ARRAYREF });

Now suppose that function may also take optional named parameters, such
as timeout. You with vanilla Params::Validate, you would need to separate
the first positional parameters from the rest, and use that as the argument:

  my @positional = splice(@_, 0, 2);
  my @named      = @_;
  my($a1, $a2) = validate_pos(
    @positional, { type => SCALAR }, { type => ARRAYREF }
  );
  my %args = validate(
    @named,
    {
      timeout => { type => SCALAR, default => 120 }
    }
  );

Which works, but is a bit cumbersome. C<validate_mixed()> partially solves
this by allowing you to specify both positional and named parameters
all at once:

  my($a1, $a2, %args) = validate_mixed(
    @_,
    [ { type => SCALAR }, { type => ARRAYREF } ], # positional
    {
      timeout => { type => SCALAR, default => 120 } # named
    }
  );

Note that when you only have one of each (positional or named parameters),
you shouldn't be using this function. Use C<validate_pos()> and C<validate()>
instead.

=cut

# XXX - for those who are not sure what the prototype means...
#       the prototype
#
#         validate_mixed(\@$$)
#
#       doesn't mean 1 arrayref, 2 scalars. well, it kinda does,
#       but the intended argument syntax is:
#
#         validate_mixed( @args, $arrayref, $hashref )
#
#       wacked, eh?

sub validate_mixed(\@$$)
{
	my $p = shift;
	my($pos_spec, $named_spec) = validate_pos(@_,
		{ type => ARRAYREF }, { type => HASHREF }
	);

	my $pos_arg_count = scalar(@$pos_spec);
	my @pos_args      = splice(@$p, 0, $pos_arg_count);
	my @named_args    = @$p;
	my $subname       = (caller(1))[3];

	my @validated_pos_args = validate_with(
		params => \@pos_args,
		spec   => $pos_spec,
		called => $subname
	);
	my %validated_named_args = validate_with(
		params => \@named_args,
		spec   => $named_spec,
		called => $subname
	);

	return (@validated_pos_args, %validated_named_args);
}

=head2 eval_or_raise_error BLOCK

  XXX - This function was created to promote use of exceptions,
  but since NetApp test code already heavily relies on returning
  success/failure with $@ optionally being set it has since been
  removed from all code that used to use this.

  It is left here more so for demonstration purposes. Do not use.
  (It is also intentionally not exported)

Runs BLOCK through C<eval()>, and either returns its result or raises
and exception if the caller package's $RaiseError variable and
it is set to a true value. (Note that you do NOT need to explicitly
define this variable... but if you want to be explicit, you should)

This function provides a way to make allow people who like to check
return values and people who like to raise exceptions to co-exist,
in a relatively painless way. The only change you need to make from
your "return eval { .. }" style code is to swap the C<eval()> with
C<eval_or_raise_error()>:

  # before
  return eval {
    some_operation() or die "blah";
  };

  # after
  return eval_or_raise_error {
    some_operation() or die "blah";
  };

The behavior changes depending on the calling package's $RaiseError
variable. If set to true, exceptions are raised:

  package MyPackage
  sub function {
    return eval_or_raise_error {
      some_operation() or die "blah";
    }
  }

  function(); # no exception
  local $MyPackage::RaiseError = 1;
  function(); # this will potentially die()

=head3 Motivation

Checking return values is the traditional way, and is widely supported
by the NetApp community (probably because of people's C-oriented roots).

  if( !function(...) ) {
    # handle error
  }

This is all fine, but when you have a string of operations, it quickly
starts being cumbersome:

  if( !f1(..) ) {
    # handle error
  }
  if( !f2(..) ) {
    # handle error
  }
  if( !f3(..) ) {
    # handle error
  }
  if( !f4(..) ) {
    # handle error
  }

With exceptions, you get to bundle everything in a eval block:

  eval {
    f1();
    f2();
    f3();
    f4();
  };
  if($@) {
    print "One of the functions died on me: $@";
  }

Also, in certain cases if you're no careful you may end up ignoring
an important return value. With exceptions, errors all jump at you unless
you explicitly state that you want to ignore that error. 

=cut
sub eval_or_raise_error(&)
{
	my($code) = @_;

	my($pkg)  = caller();
	my $raise = do {
		no strict 'refs';
		${"${pkg}::RaiseError"}
	};

	local $Carp::CarpLevel = 1;
	if(wantarray) {
		my @ret = eval {
			$code->();
		};
		if($raise && $@) {
			croak $@;
		}
		return @ret;
	} else {
		my $ret = eval {
			$code->();
		};
		if($raise && $@) {
			croak $@;
		}
		return $ret;
	}
}

# XXX - following was needed for eval_or_raise_error, but I'm
# disabling it
# sub import
#{
#	my $class = shift;
#	if(grep{ $_ eq ':raise_error' || $_ eq ':all'} @_) {
#		# install '$RaiseError' in the caller's namespace
#		no strict 'refs';
#		my($pkg) = caller();
#		my $varname = "${pkg}::RaiseError";
#		if(!defined( ${$varname} )) {
#			${$varname} = 0; # give it a default value and define
#		}
#	}
#	local $Exporter::ExportLevel = 1;
#	$class->SUPER::import(@_);
#}

1;

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Module/Common.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

__END__
