#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Mount/Aix.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package Mount::Aix;

## @summary
##   Class to handle OS dependent portion of mount and umount on AIX.
##
## @description
##   This is a sub-class of the MountBase package which provides
##   support for AIX specific portion of mount and umount. See
##   GenericMount.pm for more detailed documentation on features.
##
## @usage
##   use Mount::Aix;
##
## @status	review
##
## @author  dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.
##
## @notes
##    Written for AIX 5.3

use strict;
use Tharn;
use Carp qw( croak );
use Mount::MountBase;
use vars qw( @ISA );

#---------------------------------------------------------------------
# Inherit all functionality from MountBase class.
#---------------------------------------------------------------------
our @ISA = qw(Mount::MountBase);

## @name ::new
##
## @description
##  AIX specific class constructor. It overrides default
##  behavior of the base mount class.
##
## @return
##  AIX specific mount object.

sub new ( $ ) {
    my $class = shift;

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class (
        rpcinfo_cmd    => '/usr/sbin/rpcinfo',
    );
}

## @name ::mount
## @description
##  Handles the AIX dependent portion of mounting.
##
##  AIX just ignores options passed in that it doesn't understand so there
##  is no need to filter out any options. AIX supports most of the ONC
##  NFS mount options so there is no need for mappings. See GenericMount.pm
##  for a description of the generic parameters.
##
## @arg
##  $this	Required: The name of the class. Automatically passed in
##		by PERL.
##
##  $opts_ref	Required: reference to mount library options hash
##
##  $os_mntopts_ref Required: reference to OS dependent mount command
##		options hash
##
## @return
##  $mountpoint	Set if success else set to undef

sub mount ( $\%\% ) {
    my ($this, $opts_ref, $os_mntopts_ref) = @_;
    my ($cmdline, $mountfrom, $opt, $out);

    # Set mountfrom based on whether a server was specified or not.
    if ($$opts_ref{server}) {
        $mountfrom = "$$opts_ref{server}:$$opts_ref{path}";
    } else {
        $mountfrom = $$opts_ref{path};
    }

    if (!$$opts_ref{mountpoint}) {
        my $suffix;

        if ($$os_mntopts_ref{proto}) {
            $suffix .= $$os_mntopts_ref{proto};
        } else {
            $suffix .= 'tcp';
        }
        if ($$os_mntopts_ref{vers}) {
            $suffix .= "_v$$os_mntopts_ref{vers}";
        } else {
            $suffix .= "_v4";
        }
        $$opts_ref{mountpoint} = $this->get_mountpoint($opts_ref, $suffix);
    }

    # Build the command line mapping from generic options to AIX Options
    $cmdline = $this->{mount_cmd};
    if ($$opts_ref{fstype}) {
        $cmdline .= " -v $$opts_ref{fstype} ";
    }
    # Add the user/default options to the list
    if (keys %$os_mntopts_ref != 0) {
        $cmdline .= " -o ";
    }
    for $opt ( keys %$os_mntopts_ref ) {
        my $nopt = $opt;
        $cmdline .= "$nopt";
        if ($$os_mntopts_ref{$opt} ne $MountClass::novalue) {
            $cmdline .= "=$$os_mntopts_ref{$opt}";
        }
        $cmdline .= ",";
    }
    # Remove a trailing ,
    $cmdline =~ s@,$@@;
    # Add the mounting from and to arguments
    $cmdline .= " $mountfrom $$opts_ref{mountpoint}";

    # Even if the caller requested no retries we can end up retrying
    # a mount if the previous mount attempt returned timeout, stale mount,
    # or server not responding.
    my $timeout_retry = 0;
    my $stale_retry = 0;
    my $srv_not_resp_retry = 0;
    for (my $i = 0; $i <= $$opts_ref{mountretry}; $i++) {
        $out = $this->do_mount($opts_ref, $cmdline, $$opts_ref{mountpoint});

        # The order of error processing is very significant
        # First handle class of errors returned by say. AIX
        # is funny in that the mount timeouts when doing a V4 mount
        # but the server doesn't support V4.
        if ($out =~ /Err=TIMEOUT/i || $out =~ /NFS service not available/i) {
            # Reset the output string because say will have dumped the whole
            # command line which contains mount_ok.
            $out = 'TIMEOUT';

            # Send a SIGINT on the client connection in-case the sayPerl
            # script is still running.
            $$opts_ref{cc}->send_intr();

            if ($timeout_retry == 0) {
                $timeout_retry = 1;
                $i--;
                logcomment("AIX mount retrying after timeout.\n");
            }
            next;
        }
        if ($out =~ /_-mount_ok-_/) {
            $out = undef;
            last;
        }
        # On AIX if the mkdir fails and we see the special message then
        # the mountpoint is stale and it is flagged.
        if ($out =~ /_-mkdir_failed-_/ &&
            $out =~ /file system .* no longer available/) {
            $out = '_-stale_nfs_filehandle-_';
        }
        if ($out =~ /_-stale_nfs_filehandle-_/) {
            # turn on the force flag for umount and then call generic umount.
            $$opts_ref{force} = 1;
            $this->umount($opts_ref);
            $$opts_ref{force} = undef;
            if ($stale_retry == 0) {
                $stale_retry = 1;
                $i--;
                logcomment("AIX mount retrying after stale NFS filehandle.\n");
            }
            next;
        }
        # Handle server not responding, want to always retry these up to a point
        # then catch it here.
        if ($out =~ /server not responding/i) {
            if ($srv_not_resp_retry == 0) {
                $srv_not_resp_retry = 1;
                $i--;
                logcomment("AIX mount retrying after server not responding.\n");
            }
            next;
        }
        if ($out =~ /_-chmod_failed-_/ ||
            $out =~ /_-mode_-chmod_failed-_/) {
            # turn on the force flag for umount and then call generic umount.
            $$opts_ref{force} = 1;
            $this->umount($opts_ref);
            $$opts_ref{force} = undef;
            last;
        }
        if ($out =~ /_-exclusive_failed-_/ ||
            $out =~ /_-parent_stat_failed-_/ ||
            $out =~ /_-mountpoint_stat_failed-_/ ||
            $out =~ /_-mount_cmd_failed-_/) {
            last;
        }
    }
    croak $out if ($out);
    return ($$opts_ref{mountpoint});
}

1;
