#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Mount/Linux.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package Mount::Linux;

## @summary
##   Class to handle OS dependent portion of mount and umount on Linux.
##
## @description
##   This is a sub-class of the MountBase package which provides
##   support for Linux specific portion of mount and umount. See
##   GenericMount.pm for more detailed documentation on features.
##
## @usage
##   use Mount::Linux;
##
## @status	review
##
## @author  dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.
##
## @notes
##    Written for Fedora Core 3.
## @Modifications
## Modified line 222 as a fix for burt 367196. Author: Meharpraveen.Maddineni@netapp.com

use strict;
use Tharn;
use Carp qw( croak );
use Mount::MountBase;
use vars qw( @ISA );

#---------------------------------------------------------------------
# Inherit all functionality from MountBase class.
#---------------------------------------------------------------------
our @ISA = qw(Mount::MountBase);

## @name ::new
##
## @description
##  Linux specific class constructor. It overrides default
##  behavior of the base mount class.
##
## @return
##  Linux specific mount object.

sub new ( $ ) {
    my $class = shift;

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class (
        mount_cmd      => '/bin/mount',
        umount_cmd     => '/bin/umount',
        rpcinfo_cmd    => '/usr/sbin/rpcinfo',
    );
}

## @name ::mount
## @description
##  Handles the Linux dependent portion of mounting.
##
##  The documentation below discusses any options that have to be mapped
##  from the generic value to Linux specific. See GenericMount.pm for a
##  description of the generic parameters.
##
##  The following generic mount options are not supported by Linux:
##	forcedirectio
##	grpid
##	quota
##	xattr
##
## @arg
##  $this	Required: The name of the class. Automatically passed in
##		by PERL.
##
##  $opts_ref	Required: reference to mount library options hash
##
##  $os_mntopts_ref Required: reference to OS dependent mount command
##		options hash
##
## @return
##  $mountpoint	Set if success else set to undef
##
## @note
## 1) On Solaris the retry options is the number of retries, on Linux
##    it is the number of minutes to retry. Right now we do not distinguish
##    between the two.

sub mount ( $\%\% ) {
    my ($this, $opts_ref, $os_mntopts_ref) = @_;
    my ($cmdline, $errstr, $mountfrom, $opt, $out);

    # Error out on any options Linux does not support
    my @unsupp_opt = qw(forcedirectio grpid quota xattr);
    foreach $opt (@unsupp_opt) {
        if ($$os_mntopts_ref{$opt}) {
            $errstr .= "$opt ";
        }
    }
    if ($errstr) {
        croak "Linux mount does not support these options: $errstr";
    }

    # Remove any options Linux does not support but we won't fail
    @unsupp_opt = qw(noforcedirectio noquota noxattr);
    foreach $opt (@unsupp_opt) {
        delete $$os_mntopts_ref{$opt};
    }

    # If v4 is set then set fstype to nfs4
    if ($$os_mntopts_ref{vers} == 4) {
        $$opts_ref{fstype} = 'nfs4';
    }

    # Set mountfrom based on whether a server was specified or not.
    if ($$opts_ref{server}) {
        $mountfrom = "$$opts_ref{server}:$$opts_ref{path}";
    } else {
        $mountfrom = $$opts_ref{path};
    }

    if (!$$opts_ref{mountpoint}) {
        my $suffix;

        if ($$os_mntopts_ref{proto}) {
            $suffix .= $$os_mntopts_ref{proto};
        } else {
            $suffix .= 'tcp';
        }
        if ($$os_mntopts_ref{vers}) {
            $suffix .= "_v$$os_mntopts_ref{vers}";
        } else {
            $suffix .= "_v3";
        }
        $$opts_ref{mountpoint} = $this->get_mountpoint($opts_ref, $suffix);
    }

    # Build the command line mapping from generic options to Linux Options
    $cmdline = $this->{mount_cmd};
    if ($$opts_ref{fstype}) {
        $cmdline .= " -t $$opts_ref{fstype} ";
    }
    # Add the user/default options to the list
    if (keys %$os_mntopts_ref != 0) {
        $cmdline .= " -o ";
    }
    for $opt ( keys %$os_mntopts_ref ) {
        my $nopt = $opt;
        if ($opt eq "vers") {
            # Linux doesn't support setting nfsvers=4, have to use fstype
            # so skip the option here
            if ($$os_mntopts_ref{vers} == 4) {
                next;
            }
            $nopt = "nfsvers";
        }
        $cmdline .= "$nopt";
        if ($$os_mntopts_ref{$opt} ne $MountClass::novalue) {
            $cmdline .= "=$$os_mntopts_ref{$opt}";
        }
        $cmdline .= ",";
    }
    # Remove a trailing ,
    $cmdline =~ s@,$@@;
    # Add the mounting from and to arguments
    $cmdline .= " $mountfrom $$opts_ref{mountpoint}";

    # Even if the caller requested no retries we can end up retrying
    # a mount if the previous mount attempt returned timeout, stale mount,
    # or server not responding. A weak cred error can retry up to 5 times.
    my $timeout_retry = 0;
    my $stale_retry = 0;
    my $srv_not_resp_retry = 0;
    my $weak_cred_retry = 5;
    for (my $i = 0; $i <= $$opts_ref{mountretry}; $i++) {
        $out = $this->do_mount($opts_ref, $cmdline, $$opts_ref{mountpoint});

        # The order of error processing is very significant
        # First handle say timeouts and weak creds
        if ($out =~ /credential too weak/i ||
            $out =~ /Permission denied/i ||
            $out =~ /bindresvport: Address already in use/i ||
            $out =~ /Err=TIMEOUT/i) {
            # First use netstat to look to see if there is a TCP reserve
            # port exhaustion problem. If so then sleep to let linux reclaim
            # some reserved TCP ports in TIME_WAIT state. Do this even if
            # this is a UDP mount because only NFS uses UDP, mountd still
            # used TCP on Linux.
            my $tmp;
            (undef, undef, $tmp) = say($$opts_ref{cc},
                                       "netstat -a | grep tcp | wc",
                                       '-swallow', '-timeout', 5,
                                       '-eval_error');
            if (!$@) {
                $tmp =~ /(\s+)(\d+)(\s+)(\d+)(\s+)(\d+)/;
                if ($2 > 300) {
                    logcomment("Linux mount detected TCP reserve port exhaustion\n");
                    sleep 15;
                    if ($weak_cred_retry-- > 0) {
                        $i--;
                    }
                    next;
                }
            }
            # If the timeout wasn't due to reserve port exhaustion then
            # decrement the retry counter.
            if ($out =~ /Err=TIMEOUT/i) {
                if ($timeout_retry == 0) {
                    $timeout_retry = 1;
                    $i--;
                    logcomment("Linux mount retrying after timeout.\n");

                    # Send a SIGINT on the client connection in-case the sayPerl
                    # script is still running.

		    ## The following eval block is added as a fix for burt 367196
                    eval { $$opts_ref{cc}->send_intr(); };
                }
                # If say wrote errors into $out then it contains the command
                # string which means mount_ok is in there, reset string.
                $out = 'TIMEOUT';
            }
            next;
        }
        if ($out =~ /_-mount_ok-_/) {
            $out = undef;
            last;
        }
        if ($out =~ /_-stale_nfs_filehandle-_/) {
            # turn on the force flag for umount and then call generic umount.
            $$opts_ref{force} = 1;
            $this->umount($opts_ref);
            $$opts_ref{force} = undef;
            if ($stale_retry == 0) {
                $stale_retry = 1;
                $i--;
                logcomment("Linux mount retrying after stale filehandle.\n");
            }
            next;
        }
        # Handle server not responding, want to always retry these up to a point
        # then catch it here.
        if ($out =~ /server not responding/i) {
            if ($srv_not_resp_retry == 0) {
                $srv_not_resp_retry = 1;
                $i--;
                logcomment("Linux mount retrying after server not responding.\n");
            }
            next;
        }
        if ($out =~ /_-chmod_failed-_/ ||
            $out =~ /_-mode_-chmod_failed-_/) {
            # turn on the force flag for umount and then call generic umount.
            $$opts_ref{force} = 1;
            $this->umount($opts_ref);
            $$opts_ref{force} = undef;
            last;
        }
        if ($out =~ /_-exclusive_failed-_/ ||
            $out =~ /_-parent_stat_failed-_/ ||
            $out =~ /_-mountpoint_stat_failed-_/ ||
            $out =~ /_-mkdir_failed-_/ ||
            $out =~ /_-mount_cmd_failed-_/) {
            last;
        }
    }
    croak $out if ($out);
    return ($$opts_ref{mountpoint});
}

1;
