# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Mount/MountBase.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package Mount::MountBase;

## @summary
##  Base class to the OS dependent Mount classes
##
## @description
##  Provides a default behavior based on Solaris 10 for the subroutines
##  that perform mount and umount operations. These subroutines should
##  only be called directly or indirectly from the GenericMount.pm library.
##  This should be the parent class for any OS-dependent class.
##
## @usage
##  use nfs::MountBase;
##
## @status	review
##
## @author  dl-maryv (FSS Protocols QA)
## @notes       Created by EP Network Storage Performance Lab, Inc.
## @notes
## 1) The Solaris mount command will take any version number you give
##    it whether it is valid or not. If the number is invalid then it
##    uses the default version. We don't verify that the version requested
##    is supported by the version of Solaris being used.

use strict;
use Tharn;
use Carp qw( croak );
use SayWrap qw( sayPerl );

# Number of times to retry an operation due to a dropped connection
my $drop_conn_retries = 3;

# File that gets created in a mount point directory to let umount
# know that the mountpoint needs to be deleted after an umount.
my $created_file = 'mntlib_created';

# Root path used for mounts
my $mnt_root_path = '/t';

## @name ::new
##
## @description
##  Mount base class constructor which initializes default behavior
##  based on Solaris 10. Class behavior is controlled by the blessed
##  hash of the constructor.
##
## @synopsis
##  $mount_obj = Mount::MountBase::->new( %override_defaults );
##
## @arg
##  options		Optional: Any named parameter found in a Mount
##			object can be specified here as the default value.
##
##  mount_cmd		Optional: Full path of mount command.
##
##  rpcinfo_cmd		Optional: Full Path of rpcinfo command.
##
##  umount_cmd		Optional: Full path of umount command.
##
##  umount_force_flg	Optional: Flag used to force an umount
##
## @return
##  Mount base object.
##
## @example
##  $obj = Mount::MountBase::->new( rpcinfo_cmd => '/usr/local/bin/rpcinfo' );
sub new($;%) {
    my $class = shift;

    my $sudo_cmd = param('SUDO_PATH', '-mesg','(NATE is expected to have set this)');

    bless {
        # Defaults
        mount_cmd        => '/usr/sbin/mount',
        rpcinfo_cmd      => '/bin/rpcinfo',
        sudo             => $sudo_cmd,
        umount_cmd       => '/usr/sbin/umount',
        umount_force_flg => '-f',

        # Override defaults
        @_
    }, $class;
}

## @name ::mountlist
## @description
##  Return the output from running the mount command without any parameters.
##
## @arg
##  $this	Required: The name of the class. Automatically passed in
##		by PERL.
##
##  $opts_ref	Required: reference to mount library options hash
##
## @return
##  Output of mount command run without any parameters

sub mountlist ( $\% ) {
    my ($this, $opts_ref) = @_;
    my ($out);

    (undef, undef, $out) = say($$opts_ref{cc}, $this->{mount_cmd},
                               '-swallow', '-timeout', $$opts_ref{timeout});
    return ($out);
}

## @name ::mount
## @description
##  Handles the Solaris compatible portion of mounting.
##
## @arg
##  $this	Required: The name of the class. Automatically passed in
##		by PERL.
##
##  $opts_ref	Required: reference to mount library options hash
##
##  $os_mntopts_ref Required: reference to OS dependent mount command
##		options hash
##
## @return
##  $mountpoint	Set if success else set to undef

sub mount ( $\%\% ) {
    my ($this, $opts_ref, $os_mntopts_ref) = @_;
    my ($cmdline, $mountfrom, $opt, $out);

    # Set mountfrom based on whether a server was specified or not.
    if ($$opts_ref{server}) {
        $mountfrom = "$$opts_ref{server}:$$opts_ref{path}";
    } else {
        $mountfrom = $$opts_ref{path};
    }

    if (!$$opts_ref{mountpoint}) {
        my $suffix;

        if ($$os_mntopts_ref{proto}) {
            $suffix .= $$os_mntopts_ref{proto};
        } else {
            $suffix .= 'tcp';
        }
        if ($$os_mntopts_ref{vers}) {
            $suffix .= "_v$$os_mntopts_ref{vers}";
        } else {
            $suffix .= "_v4";
        }
        $$opts_ref{mountpoint} = $this->get_mountpoint($opts_ref, $suffix);
    }

    # Build the command line mapping from generic options to Solaris Options
    $cmdline = $this->{mount_cmd};
    if ($$opts_ref{fstype}) {
        $cmdline .= " -F $$opts_ref{fstype} ";
    }
    # Add the user/default options to the list
    if (keys %$os_mntopts_ref != 0) {
        $cmdline .= " -o ";
    }
    for $opt ( keys %$os_mntopts_ref ) {
        my $nopt = $opt;
        $cmdline .= "$nopt";
        if ($$os_mntopts_ref{$opt} ne $MountClass::novalue) {
            $cmdline .= "=$$os_mntopts_ref{$opt}";
        }
        $cmdline .= ",";
    }
    # Remove a trailing ,
    $cmdline =~ s@,$@@;
    # Add the mounting from and to arguments
    $cmdline .= " $mountfrom $$opts_ref{mountpoint}";

    # Even if the caller requested no retries we can end up retrying
    # a mount if the previous mount attempt returned timeout, stale mount,
    # or server not responding.
    my $timeout_retry = 0;
    my $stale_retry = 0;
    my $srv_not_resp_retry = 0;
    for (my $i = 0; $i <= $$opts_ref{mountretry}; $i++) {
        $out = $this->do_mount($opts_ref, $cmdline, $$opts_ref{mountpoint});

        # The order of error processing is very significant
        # First handle class of errors returned by say. Solaris 10 is
        # funny in that the mount will timeout when doing a V4 mount
        # if the server doesn't support V4.
        if ($out =~ /NFS service not available/i ||
            $out =~ /Err=TIMEOUT/i) {
            # Reset the output string because say will have dumped the whole
            # command line which contains mount_ok.
            $out = 'TIMEOUT';

            # Send a SIGINT on the client connection in-case the sayPerl
            # script is still running.
            $$opts_ref{cc}->send_intr();

            if ($timeout_retry == 0) {
                $timeout_retry = 1;
                $i--;
                logcomment("Solaris mount retrying after timeout.\n");
            }
            next;
        }
        if ($out =~ /_-mount_ok-_/) {
            $out = undef;
            last;
        }
        if ($out =~ /_-stale_nfs_filehandle-_/) {
            # turn on the force flag for umount and then call generic umount.
            $$opts_ref{force} = 1;
            $this->umount($opts_ref);
            $$opts_ref{force} = undef;
            if ($stale_retry == 0) {
                $stale_retry = 1;
                $i--;
                logcomment("Solaris mount retrying after stale NFS filehandle.\n");
            }
            next;
        }
        # Handle server not responding, want to always retry these up to a point
        # then catch it here.
        if ($out =~ /server not responding/i) {
            if ($srv_not_resp_retry == 0) {
                $srv_not_resp_retry = 1;
                $i--;
                logcomment("Solaris mount retrying after server not responding.\n");
            }
            next;
        }
        if ($out =~ /_-chmod_failed-_/ ||
            $out =~ /_-mode_-chmod_failed-_/) {
            # turn on the force flag for umount and then call generic umount.
            $$opts_ref{force} = 1;
            $this->umount($opts_ref);
            $$opts_ref{force} = undef;
            last;
        }
        if ($out =~ /_-exclusive_failed-_/ ||
            $out =~ /_-parent_stat_failed-_/ ||
            $out =~ /_-mountpoint_stat_failed-_/ ||
            $out =~ /_-mkdir_failed-_/ ||
            $out =~ /_-mount_cmd_failed-_/) {
            last;
        }
    }
    croak $out if ($out);
    return ($$opts_ref{mountpoint});
}

## @name ::umount
## @description
##  Contains the OS dependent call to the umount command.
##  It handles the following:
##   1) Maps the generic force option to the OS dependent flag.
##   2) Call the OS specific umount
##
## @arg
##  $this	Required: The name of the class. Automatically passed in
##		by PERL.
##
##  $opts_ref	Required: reference to mount library options hash
##
## @return
##  1 if succeeds else croaks

sub umount ( $\% ) {
    my ($this, $opts_ref) = @_;
    my ($cmd, $out, $retries);

    # Create the umount command
    $cmd = "$this->{umount_cmd} ";
    if ($$opts_ref{force}) {
        $cmd .= " $this->{umount_force_flg} ";
    }
    $cmd .= $$opts_ref{mountpoint};

    # Create a Perl script to run the umount command on the
    # specified client. The script looks to see if the created file
    # is there after the umount and if so it removes the mountpoint by
    # recursively removing the last directory name in the path until
    # the rmdir fails (indicated directory not empty).
    $cmd = qq{
        use File::Basename;
        \$out = `$cmd 2>&1 && echo _-umount_ok-_`;
        if ( \$out =~ /_-umount_ok-_/ ) {
            if ( -f "$$opts_ref{mountpoint}/$created_file" ) {
                unlink "$$opts_ref{mountpoint}/$created_file";
                \$dir = "$$opts_ref{mountpoint}";
                \$err = 0;
                while(\$err == 0 and \$dir ne "/") {
                    rmdir \$dir or \$err = \$OS_ERROR;
                    \$dir = dirname(\$dir);
                }
            }
            print "_-umount_ok-_";
        } else {
            print "\$out _-umount_cmd_failed-_";
        }
    };

    # Execute script to check the mount on the client. The sayPerl may
    # return no error and no output here. If that is the case then
    # the connection was most likely dropped and the command needs to
    # be retried.
    $out = undef;
    $retries = $drop_conn_retries;

    # Control the say() output from NATE
    my $disable_umount_say = param("DISABLE_UMOUNT_SAYLOG", -default, 0);
    my $option = $disable_umount_say ? "-noverbose" : "";

    while (!$out && $retries > 0) {
        (undef, undef, $out) = sayPerl($$opts_ref{cc}, "$cmd",
                                "$this->{sudo}", '-swallow', '-timeout',
                                $$opts_ref{timeout}, '-eval_error', "$option");
        $out = $@ if ($@);
        $retries--;
    }
    if (!$out) {
        croak "umount of \"$$opts_ref{mountpoint}\" failed.\ncommand: $cmd\n";
    } elsif ($out !~ /_-umount_ok-_/) {
        $out =~ s/^\s+//;
        $out =~ s/\s+$//;
        croak "umount of \"$$opts_ref{mountpoint}\" failed.\ncommand: $cmd\noutput: $out\n";
    }
    return 1;
}

## @name ::get_mountpoint
## @description
##  Creates the name of the mountpoint given a list of various arguments
##  that are used to build the path of the mountpoint. An example follows:
##	/t/$server/$path/share_type/$suffix
##  If $path is "/" then it is mapped to "root".
##
## @synopsis
## 	$mountpoint = $this->get_mountpoint(\%opts, $suffix);
##
## @arg
##	$this		Required: Name of class
##
##	$opts_ref	Required: Reference to the library options hash
##
##	$suffix		Optional: suffix to add to end of directory name
##
## @return
##  returns the mountpoint

sub get_mountpoint( $\%;$ ) {
    my ($this, $opts_ref, $suffix) = @_;
    my ($path, $tpath);

    # Map '/' to the word 'root'
    if ($$opts_ref{path} eq '/') {
        $path = 'root';
    } else {
        $path = $$opts_ref{path};
    }

    # Remove any preceding / on path and trailing / at end
    $path =~ s@^/@@;
    $path =~ s@/$@@;

    # Add sharing type and suffix to the path
    $path .= '/' . $$opts_ref{sharing};
    if ($suffix) {
        $path .= "/$suffix";
    }

    $tpath = $mnt_root_path . '/';
    if ($$opts_ref{server}) {
        $tpath .= "$$opts_ref{server}/";
    }
    $tpath .= $path;
    return ($tpath);
}

## @name ::do_mount
## @description
##  Runs the mount command line on the client.
##
## @synopsis
## 	Mount->do_mount($opts_ref, $cmdline, $mountpoint);
##
## @arg
##	$opts_ref	Required: Reference to the library options hash
##
##	$cmdline	Required: full command line to execute
##
##	$mountpoint	Required: The mountpoint being mounted.
##
## @return
##  returns output from the command execution

sub do_mount( $\%$$ ) {
    my ($this, $opts_ref, $cmdline, $mountpoint) = @_;
    my ($chmod_val, $cmd, $out, $parent, $retries);

    # If chmod value isn't defined set it to a value we won't use
    # so the perl script knows it is not set.
    if (defined($$opts_ref{chmod})) {
        $chmod_val = $$opts_ref{chmod};
    } else {
        $chmod_val = -1;
    }
    # The following is a perl script to do the mount on the client and
    # change the mode if needed. A script is used instead of running
    # multiple say commands because it runs faster. Mount sharing is
    # supported by this script. If the script creates the directory for
    # the mountpoint then it creates a file in that directory. This lets
    # the unmount command know that it needs to remove the directory.
    $parent = $mountpoint;
    $parent =~ s@/[^/]+$@@;

    # Control the say() output from NATE
    my $disable_mount_say = param("DISABLE_MOUNT_SAYLOG", -default, 0);
    my $option = $disable_mount_say ? "-noverbose" : "";

    # The SayPerl normally returns output but sometimes the connection
    # is dropped and no output is received. This causes a problem because
    # we don't know if the mount was done or not. Therefore we need to handle
    # retrying this command if no output is received. This creates a
    # problem for exclusive mounts because the mount should only be done once.
    # That case is handled here.

    if ($$opts_ref{sharing} eq "exclusive") {
        $cmd = qq{
            sub is_mounted() {
                (\$pardev)=stat("$parent") or
                    die "\$OS_ERROR:_-parent_stat_failed-_";
                (\$mntdev)=stat("$mountpoint") or
                    die "\$OS_ERROR:_-mountpoint_stat_failed-_";
                return (\$pardev != \$mntdev);
            }
            \$out = `ls $$opts_ref{mountpoint} 2>&1`;
            if (\$out =~ /stale nfs|I.O error/i) {
                print "_-stale_nfs_filehandle-_";
                exit 1;
            } elsif (\$out =~ /No such/i) {
                \$poss_stale = 1;
            }
            if ( not -d "$$opts_ref{mountpoint}" ) {
                \$out = `mkdir -p "$$opts_ref{mountpoint}"`;
                if ( not -d "$$opts_ref{mountpoint}" ) {
                    print "_-mkdir_failed-_";
                    exit 1;
                }
            } else {
                if (\$poss_stale == 1) {
                    print "_-stale_nfs_filehandle-_";
                    exit 1;
                }
                if (is_mounted()) {
                    print "Already mounted_-exclusive_failed-_";
                    exit 1;
                }
            }
            print "_-mountchk_ok-_";
            exit 0;
        };

        # Execute script to check the mount on the client. The SayPerl may
        # return no error and no output here. If that is the case then
        # the connection was most likely dropped and the command needs to
        # be retried.
        $out = undef;
        $retries = $drop_conn_retries;
        while (!$out && $retries > 0) {
            (undef, undef, $out) = sayPerl($$opts_ref{cc}, "$cmd",
                                    "$this->{sudo}", '-swallow', '-timeout',
                                    $$opts_ref{timeout}, '-eval_error', "$option");
            $out = $@ if ($@);
            $retries--;
        }
        $out = '_-mount_cmd_failed-_' if (!$out);

        # If we tried to make the directory then we think it didn't exist
        # but it did - good sign this is a stale mount.
        if ($out =~ /_-mkdir_failed-_/ &&
            $out =~ /File exists/) {
            $out = '_-stale_nfs_filehandle-_';
        }
        return ($out) if ($out !~ /_-mountchk_ok-_/);
    }

    $cmd = qq{
        sub is_mounted {
            (\$pardev)=stat("$parent") or
                die "\$OS_ERROR:_-parent_stat_failed-_";
            (\$mntdev)=stat("$mountpoint") or
                die "\$OS_ERROR:_-mountpoint_stat_failed-_";
            return (\$pardev != \$mntdev);
        }
        \$out = `ls $$opts_ref{mountpoint} 2>&1`;
        if (\$out =~ /stale nfs|I.O error/i) {
            print "_-stale_nfs_filehandle-_";
            exit 1;
        } elsif (\$out =~ /No such/i) {
            \$poss_stale = 1;
        }
        if ( not -d "$$opts_ref{mountpoint}" ) {
            \$out = `mkdir -p "$$opts_ref{mountpoint}"`;
            if ( not -d "$$opts_ref{mountpoint}" ) {
                print "_-mkdir_failed-_";
                exit 1;
            }
            \$out = `echo > $$opts_ref{mountpoint}/$created_file`;
        } else {
            if (\$poss_stale == 1) {
                print "_-stale_nfs_filehandle-_";
                exit 1;
            }
            if (is_mounted()) {
                print "Already mounted_-mount_ok-_";
                exit 0;
            }
        }
        \$out = `$cmdline 2>&1 && echo _-mount_ok-_`;
        if (not is_mounted()) {
            if (\$out !~ /_-mount_ok-_/) {
                print "\$out _-mount_cmd_failed-_";
            } else {
                print "\$out _-mount_failed-_";
            }
            exit 1;
        }
        if ($chmod_val != -1) {
            \$out = `chmod $chmod_val "$$opts_ref{mountpoint}" && echo _-chmod_ok-_`;
            if (\$out !~ /_-chmod_ok-_/) {
                die "\$out _-chmod_failed-_";
            }
            \$mode = (stat("$mountpoint"))[2] & 07777 or
                die "\$OS_ERROR:_-stat_fail_-chmod_failed-_";
            if (\$mode != oct("$chmod_val")) {
                die "exp: $chmod_val got: \$mode:_-mode_-chmod_failed-_";
            }
        }
        print "_-mount_ok-_";
        exit 0;
    };

    # Execute script to do the mount on the client. The SayPerl may return
    # no error and no output here. If that is the case then the
    # connection was most likely dropped and the command needs to be retried.
    $out = undef;
    $retries = $drop_conn_retries;
    while (!$out && $retries > 0) {
        (undef, undef, $out) = sayPerl($$opts_ref{cc}, "$cmd",
                                "$this->{sudo}", '-swallow', '-timeout',
                                $$opts_ref{timeout}, '-eval_error', "$option");
        $out = $@ if ($@);
        $retries--;
    }

    # If we tried to make the directory then we think it didn't exist
    # but it did - good sign this is a stale mount.
    if ($out =~ /_-mkdir_failed-_/ &&
        $out =~ /File exists/) {
        $out = '_-stale_nfs_filehandle-_';
    }

    $out = '_-mount_cmd_failed-_' if (!$out);
    return $out;
}

1;
