#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NASClient.pm#1 $

## @summary     Base class for access to NAS clients
## @author      snider@netapp.com,gvg@netapp.com
## @status in progress
## @pod here

=head1 NAME

NASClient - Class definition for a NAS Client

=head1 SYNOPSIS

    use NASClient;
    my $client = new NASClient();

=head1 DESCRIPTION

C<NASClient> 

=head1 METHODS

=cut

#***********************************************************************
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDE_NAS_CLIENT}) {
    my $newModule = $ENV{OVERRIDE_NAS_CLIENT};
    delete $ENV{OVERRIDE_NAS_CLIENT};
    print "The main NASClient module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package NASClient;
use strict;
use warnings;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper qw(Dumper);
use Params::Common;
use TracePkg;
use Options;
my $Trace = new TracePkg;
my $Options = new Options;

my $Sudo = "/usr/local/bin/sudo";
my $Ifconfig = "/sbin/ifconfig";

my %Clients = (
#               "unix-linux-fedora3" => "NASClient::NFS::Synth::Fedora",
               "unix-linux-rhel3"   => "NASClient::NFS::Synth::Rhel3",
               "unix-linux-rhel4"   => "NASClient::NFS::Synth::Rhel4",
               "unix-linux-rhel5"   => "NASClient::NFS::Synth::Rhel5",
               "nfs-local"          => "NASClient::NFS::Local",
               );

$Options->parameters(definitions=>
                     [
                      {
                       name => "CLIENT_TYPE",
                       description => "the type of client to create",
                      },
                      ]);
#---------------------------------------------------------------------

=head2 new

    use NASClient;
    my $client = NASClient->new();

    use NASClient;
    my $client = NASClient->new(type => "unix-linux-rhel5");

    use NASClient;
    use Options;
    my $Options = new Options;
    $Options->process();
    my $client_type = $Options->get("CLIENT_TYPE");
    my $client = NASClient->new(type => $client_type);    

C<new> is the invocant for the C<NASClient> class. Creates appropriate
C<NASClient> object based on named param C<type> or nate param C<CLIENT_TYPE>

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Optional) : The type of the client. E.g unix-linux-rhel5

=back

=item Returns

NASClient Object on Success.

=back

=for html <hr />

=cut
    
sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my %params = (
                  type => undef,
                  @_,
                  );
    my $self = {
                type => $params{type},
            };
    $Trace->entry();
    bless($self, $class);

    $Options->process();
    my $options = $Options->get_hash();
    $self->{type} = $options->{CLIENT_TYPE} if ($options->{CLIENT_TYPE});
    if (!$self->{type}) {
        $Trace->fatal(msg=>"Please specify the type of client to $class->new() or by nate param CLIENT_TYPE. The supported clients are: ".join(",", keys %Clients), result=>"SCRIPT");
    }
    if (!grep($self->{type} eq $_, keys %Clients)) {
        $Trace->fatal(msg=>"client type '$self->{type}' is not supported. The supported clients are: ".join(",", keys %Clients), result=>"SCRIPT");
    }
    
    # we made it past the checks so create the new object and return it
    my $client_class = $Clients{$self->{type}};
    $Trace->debug(msg=>"Load $client_class library");
    eval "use $client_class";

    if ($@) {
        $Trace->fatal(msg=>"Encountered an error while trying to load the $client_class module:\n$@");
    }
    my $client;
    eval {
        $Trace->debug(msg=>"Create a new $client_class object");
        delete $params{type};
        $client = $client_class->new(%params);
    };
    if ($@) {
        $Trace->fatal(msg=>"Encountered an error while trying to generate a new $client_class object:\n$@");
    }

    $client->{options} = $Options;
    $Trace->exit();
    return $client;
}

# the following are simple stubs for supported methods of the underlying objects

###########################################################################
####      Perl Func Interfaces (Need to be decided)
###########################################################################

=head1 Perl func METHODS

=cut

#---------------------------------------------------------------------

=head2 mount

    my $mntpnt = $client->mount( export => "/vol/vol0" , mntpnt => "/t/mnt" );

The following named arguments are accepted:

=over

=item Arguments

=over

=item export

(Required) : The export path on the server

=item mntpnt

(Optional) : The path to use for the local mountpoint.
A default value will be built up if it is not specified.

=item chmod

(Optional) : This parameter specifies that a chmod should be
performed on the mountpoint after it has been mounted. The
value must be specified as a string and should begin with "0"
if the value is being specified in octal.

=back

=item Returns

The path to the local mountpoint.

=back

=for html <hr />

=cut

sub mount {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->mount() has not been implemented yet.", result=>"SCRIPT");
}

#---------------------------------------------------------------------

=head2 umount

    $client->umount( mntpnt => "/t/mnt" );

The following named arguments are accepted:

=over

=item Arguments

=over

=item mntpnt

(Required) : The local mntpnt to unmount

=back

=item Returns

None on success.

=back

=for html <hr />

=cut

sub umount {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->umount() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 ls
    
    @entries = $client->ls( directory => "$mntpnt" );

Lists and returns all directory entries on for the given directory.

The following named arguments are accepted:

=over

=item Arguments

=over

=item directory

(Required) : The full path to the directory.

=back

=item Returns

All entries in the directory. Each entry in the array is a hashref
with the following keys:

=over

=item name

The name of the entry

=item fhandle

The filehandle for the entry

=item name_attr

The named attributes for the entry. This needs further documentation.

=back

=back

=for html <hr />

=cut

sub ls {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->ls() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 read_file
    
    $bytes = $client->read_file( file => "$mntpnt/file" );
    ($bytes, $contents) = $client->read_file( file => "$mntpnt/file" );

Opens a file for reading, reads the file and then closes it.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : The full path to the file to be read.

=back

=item Returns

In a scalar context, it just returns the number of bytes that have
been read. In an array context, it will return the number of bytes
and a string representing the contents of the file.

=back

=for html <hr />

=cut

# The underlying code has support for offset, count, buf_size but they have
# not yet been tested.
sub read_file {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->write_file() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 write_file
    
    $client->write_file( file => "$mntpnt/file" , data => $data_string);

Opens a file for writing, writes the file and then closes it.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : The full path to the file to be written.

=item data

(Required) : A string of data to be written into the file.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

# The underlying code has support for offset, count, buf_size but they have
# not yet been tested.
sub write_file {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->write_file() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 unlink
    
    $client->unlink( file => "$mntpnt/file");

Removes a file.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : The full path of the file to be removed.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

sub unlink {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->is_file() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 is_file
    
    $client->is_file( object => "$mntpnt/file");

This routine is used to determine if a given directory entry
is a file or not. This is analgous to the '-f' file test operator
in perl.

The following named arguments are accepted:

=over

=item Arguments

=over

=item object

(Required) : The full path of the object to be checked.

=back

=item Returns

1 on success

=back

=for html <hr />

=cut

sub is_file {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->unlink() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 is_readable
    
    $client->is_readable( object => "$mntpnt/file");

This routine is used to determine if a given file can be
read. This is analgous to the '-r' file test operator in perl.

The following named arguments are accepted:

=over

=item Arguments

=over

=item object

(Required) : The full path of the object to be checked.

=back

=item Returns

1 on success

=back

=for html <hr />

=cut

sub is_readable {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->is_readable() has not been implemented yet.", result=>"SCRIPT");
}
#-------------------------------------------------------------------

=head2 is_writeable
    
    $client->is_writeable( object => "$mntpnt/file");

This routine is used to determine if a given file can be
written. This is analgous to the '-w' file test operator in perl.

The following named arguments are accepted:

=over

=item Arguments

=over

=item object

(Required) : The full path of the object to be checked.

=back

=item Returns

1 on success

=back

=for html <hr />

=cut

sub is_writeable {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->is_writeable() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 is_dir
    
    $client->is_dir( object => "$mntpnt/file");

This routine is used to determine if a given object is a
directory. This is analgous to the '-d' file test operator
in perl.

The following named arguments are accepted:

=over

=item Arguments

=over

=item object

(Required) : The full path of the object to be checked.

=back

=item Returns

1 on success

=back

p=for html <hr />

=cut

sub is_dir {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->is_dir() has not been implemented yet.", result=>"SCRIPT");
}

#---------------------------------------------------------------------


###########################################################################
####      Unix Commandline Style Interfaces
###########################################################################

=head1 Unix Commandline Style METHODS

=cut


#---------------------------------------------------------------------

=head2 getfacl

    $client->getfacl( name => "myfile", [ dhandle => $dhandle ] );

Get NFSv4 ACLs and return the array of ACEs.

The following named arguments are accepted:

=over

=item Arguments

=over

=item name

(Required) : File name

=item dhandle

(Optional) : Directory handle where file exists.

=back

=item Returns

Hashref of ACL

=back

=for html <hr />

=cut

sub getfacl {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->getfacl() has not been implemented yet.", result=>"SCRIPT");
}

#---------------------------------------------------------------------

=head2 setfacl

    $client->setfacl( name => "myfile", acl => $acl, [ dhandle => $dhandle, ] );

Set NFSv4 ACLs.

The following named arguments are accepted:

=over

=item Arguments

=over

=item name

(Required) : File name

=item acl

(Required) : Hashref containing all ACLs to be set.

=item dhandle

(Optional) : Directory handle where file exists.

=back

=item Returns

None if successful. fatals on failure.

=back

=for html <hr />

=cut

sub setfacl {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->setfacl() has not been implemented yet.", result=>"SCRIPT");
}



###########################################################################
####      Special Case Utilities
###########################################################################

=head1 Special Case Util METHODS

=cut

#-------------------------------------------------------------------

=head2 fpolicy_read

    $client->fpolicy_read( name => "myfile", [ dhandle => $dhandle ] );

Read the first 1024 byts to make sure the file is not a stub file
but real file. FPolicy server creates the stubfile when real file
is migrated to secondary filer as part of NetApp Transparent 
Migration solution. If it is found a stub file, this function
warns with appropriate error message.

The following named arguments are accepted:

=over

=item Arguments

=over

=item name

(Required) : File name

=item dhandle

(Optional) : Directory handle where file exists.

=back

=item Returns

Nothing if successful. Else it warns with error message.

=back

=for html <hr />

=cut

sub fpolicy_read {
    my $self = shift;
    my $class = ref($self);
    $Trace->fatal(msg=>"$class->fpolicy_read() has not been implemented yet.", result=>"SCRIPT");
}

#-------------------------------------------------------------------

=head2 getnfs

    $client->getnfs();

Returns NFS specific object for the client to perform NFS specific calls, 
if any.

The following named arguments are accepted:

=over

=item Returns

NfsClient Object if successful. Else undef.

=back

=for html <hr />

=cut

sub getnfs {
    my $self = shift;
    my $class = ref($self);
    return $self->{nfs};
}


#-------------------------------------------------------------------

=head2 getcifs

    $client->getcifs();

Returns CIFS specific object for the client to perform CIFS specific calls, 
if any.

The following named arguments are accepted:

=over

=item Returns

CifsClient Object if successful. Else dies.

=back

=for html <hr />

=cut

sub getcifs {
    my $self = shift;
    my $class = ref($self);
    return $self->{cifs} if $self->{cifs};
    $Trace->fatal(msg=>"$class does not contain CIFS object.", result=>"SCRIPT");
}

#---------------------------------------------------------------------

# The following three routines might better belong in Synth.pm with
# some basic wrappers here. The problem is that this code does not know
# if it will be tied to a real client or not

=head2 create_ip_alias

    NASClient::create_ip_alias( ip => "192.168.1.1",
                                device => "eth0",
                                netmask => "255.255.0.0");

This subroutine will add the provided IP to the provided device. This step
is required if a script is going to use a source IP in a synthetic client
and the IP is not already configured.

This code currently only works on Linux.

=over

=item Arguments

=over

=item ip

(Required) : The IP address to configure on the client.

=item device

(Required) : The device to use when adding the IP alias.

=item netmask

(Required) : The netmask to use with the IP.

=item broadcast

(Optional) : The broadcast address to use with the IP.

=back

=item Returns

None

=back

=for html <hr />

=cut

sub create_ip_alias {
    my (%params) = validate(
			    @_,
			    {
                             ip => {type => SCALAR, },
                             device => {type => SCALAR, },
                             netmask => {type => SCALAR, },
                             broadcast => {type => SCALAR|UNDEF, optional => 1},
			    }
                           );
    $Trace->entry();

    my $device = $params{device};
    my $netmask = $params{netmask};
    my $broadcast = $params{broadcast};
    my $ip = $params{ip};

    my %ips;

    if ($^O !~ /linux/i) {
        $Trace->fatal(msg=>"This routine is currently only supported on linux");
    }

    # check the current ifconfig to see if the device is available 
    #  and whether or not we already have this virtual device
    my $out = qx($Sudo $Ifconfig 2>&1);
    if ($?) {
        $Trace->fatal(msg=>"Could not get interface information: $out");
    }

    # Create virtual ip addresses on client
    my $alias = _find_device_alias($device);

    if ($alias) {
        $Trace->trace(msg=>"Found $ip:$netmask on $alias\n");
    } else {
        # find the next available dev alias
        my $dev = (sort { $a <=> $b } ($out =~ /^$device:(\S+)/mg))[-1];
        $dev = defined($dev) ? $dev+1 : 0;
        
        # Command to create virtual device
        my $cmd = "$Sudo $Ifconfig $device:$dev $ip netmask $netmask 2>&1";
        $Trace->trace(msg=>"$cmd");
        $out = qx($cmd);
        $Trace->fatal(msg=>"Could not create virtual interface: $out") if $?;
    }

    $Trace->exit();
}

#---------------------------------------------------------------------

=head2 delete_ip_alias

    NASClient::delete_ip_alias( ip => "192.168.1.1",
                                device => "eth0");

This subroutine will remove the provided IP from the provided device. This step
is required if a script is going to use a source IP in a synthetic client
and the IP is not already configured.

This code currently only works on Linux.

=over

=item Arguments

=over

=item ip

(Required) : The IP address to remove from the client.

=item device

(Required) : The device to use when removing the IP alias.

=item Returns

None

=back

=for html <hr />

=cut

sub delete_ip_alias {
    my (%params) = validate(
			    @_,
			    {
                             ip => {type => SCALAR, },
                             device => {type => SCALAR, },
			    }
                           );
    $Trace->entry();

    if ($^O !~ /linux/i) {
        $Trace->fatal(msg=>"This routine is currently only supported on linux");
    }

    my $device = $params{device};
    my $ip = $params{ip};

    my $alias = _find_device_alias($device);

    # Command to delete virtual device
    my $cmd = "$Sudo $Ifconfig $alias del $ip 2>&1";
    $Trace->trace(msg=>"$cmd");
    my $out = qx($cmd);
    $Trace->fatal(msg=>"Could not delete virtual interface: $out") if $?;

    $Trace->exit();
}

sub _find_device_alias {
    my $device = shift;
    $Trace->entry(msg=>"$device");

    # check the current ifconfig to see if the device is available 
    #  and whether or not we already have this virtual device
    my $out = qx($Sudo $Ifconfig 2>&1);
    if ($?) {
        $Trace->fatal(msg=>"Could not get interface information: $out");
    }

    my $alias = undef;
    my $found = 0;
    my $ip;
    foreach (split("\n", $out)) {
        if (/^$device\s+/) {
            $found = 1;
            next;
        }
        if (/^$device:(\S+)/) {
            $alias = $1;
            next;
        }
        if (defined($alias) && /inet addr:(\S+)\s+.+Mask:(\S+)/) {
            $ip = $1;
        }
    }
    
    if (!$found) {
        $Trace->fatal(msg=>"Did not find device '$device' in ifconfig: $out");
    }

    my $dev_alias = ($ip) ? "$device:$alias" : "";
    $Trace->exit(msg=>"return $dev_alias");
    return $dev_alias;
}
############################# END ####################################
1;

