#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NASClient/NFS/Local.pm#1 $

## @summary local NFS client
## @author snider@netapp.com
## @status in progress
## @pod here

=head1 NAME

NASClient::NFS::Local - Class definition for a local NFS client

=head1 DESCRIPTION

Please see the documentation in C<NASClient> which describes the API for all
of the NASClient objects

=cut

package NASClient::NFS::Local;
use strict;
use warnings;
use base qw(NASClient::NFS);
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF GLOBREF);
use TestUtils qw(size_in_unit do_lockunlock);
use GenericMount;
use IO::File;
use File::stat;
use TracePkg;
use Options;
use nfs::NfsDef qw(NFREG NFDIR);
my $Trace = new TracePkg;
my $Options = new Options;
$Options->parameters(definitions=>
                     [
                      {
                       name => "USE_O_DIRECT",
                       default => 1,
                      },
                      {
                       name => "FILE_READ_LOCK",
                       default => 0,
                      },
                      ]);

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    $Trace->entry();
    $Options->process();
    my $options = $Options->get_hash();
    my $self = { version => $options->{nfsversion} || 3,
                 proto => $options->{protocol},
                 srcip => $options->{SRCIP},
                 server => $options->{DATA_INTERFACES}[0] || $options->{default_data_interface},
                 retries => 6,
                 timeout => 3,#60
                 acregmin => 3, # this is for linux
               @_};
    bless($self, $class);
    
    if (! grep($_ == $self->{version}, qw(2 3 4))) {
        $Trace->fatal(msg=>"NFS version '$self->{version}' is not supported yet");
    }
    if (!$self->{server}) {
        $Trace->fatal(msg=>"Please specify a server in your call to new(), a FILER nate param or a DATA_INTERFACES nate param");
    }

    if ($self->{srcip}) {
        $Trace->fatal(msg=>"A client-type nfs-local cannot use a srcip");
    }
    $self->_set_defaults();
    return $self;
}    

#-------------------------------------------------------------------
sub mount {
    my $self = shift;
    my (%params) = 
        validate(
                 @_,
                 {
                  export => {type => SCALAR},
                  mntpnt => {type => SCALAR | UNDEF, optional => 1},
                  chmod  => {type => SCALAR, optional => 1 },
                 }
                );
    $Trace->entry();

    my $options = "vers=$self->{version},proto=$self->{proto}";
    $options .= ",rsize=$self->{rsize}" if ($self->{rsize});
    $options .= ",wsize=$self->{wsize}" if ($self->{wsize});
    $options .= ",hard" if ($self->{hard_mount});

    my %mount_options = (
                         mntopts => $options,
                         sharing => "exclusive",
                         #		   retries => $retries,
                        );
    $mount_options{chmod} = $params{chmod} if (defined($params{chmod}));
    $mount_options{mountpoint} = $params{mntpnt} if (defined($params{mntpnt}));
    my $mnt = GenericMount::mount($self->{server},
                                  $params{export},
                                  %mount_options,
                                 );
    if (!$mnt) {
        $Trace->fatal(msg=>"Failed to mount $self->{server}:$params{export}: $@");
    }
    $self->{mountpoint} = [$mnt, $params{export}];

    $Trace->exit(msg=>"return $mnt");
    return $mnt;
}

#-------------------------------------------------------------------
sub umount {
    my $self = shift;
    my (%params) = 
        validate(
                 @_,
                 {
                  mntpnt => {type => SCALAR | UNDEF},
                 }
                );
    $Trace->entry();

    if (!$self->{mountpoint}) {
        $Trace->fatal(msg=>"Did not find a cached mntpnt", result=>"SCRIPT");
    }
    my ($mount, $export) = @{$self->{mountpoint}};
    if ($mount ne $params{mntpnt}) {
        $Trace->fatal(msg=>"cached mntpnt '$mount' does not equal mntpnt '$params{mntpnt}'", result=>"SCRIPT");
    }

    $Trace->trace(msg=>"umount $mount");
    # umount() is one of the routines which will return eval {}. This means that it will
    #  not actually die but will return undef on failure with the error in $@.
    GenericMount::umount($mount);
    if ($@ && $@ !~ /not found||not mounted/) {
        $Trace->fatal(msg=>"Failed to umount '$mount': $@");
    }

    delete $self->{mountpoint};
    $Trace->exit(msg=>);
    return;
}

#-------------------------------------------------------------------
sub open {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
                  file    => {type => SCALAR,},
                  mode    => {type => SCALAR,},
              });

    $Trace->entry(msg=>"$params{file}, $params{mode}");
    my $fh = new IO::File;
    my $mode;
    if ($params{mode} eq "r") {
        $mode |= O_RDONLY;
        $mode |= O_DIRECT if ($self->{options}->{USE_O_DIRECT}); # comes from TestUtils
    } elsif ($params{mode} eq "w") {
        $mode |= O_WRONLY;
    } else {
        $Trace->fatal(msg=>"The mode '$params{mode}' is not supported.\n".
                      "Only 'r' and 'w' are currently supported.");
    }
    
    if (!$fh->open($params{file}, $params{mode})) {
        $Trace->fatal(msg=>"Failed to open file '$params{file}' for mode '$params{mode}': $!");
    }

    do_lockunlock($fh, $params{file}) if ($params{mode} eq "r");;

    $self->{handles}{$params{file}} = $fh;
    $Trace->exit();
    return $fh;
}

#-------------------------------------------------------------------
sub close {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
               file => {type => SCALAR},
              });
    $Trace->entry();
    
    my $fd = $self->{handles}{$params{file}};

    if (!$fd) {
        $Trace->fatal(msg=>"Could not find the handle for $params{file}", result=>"SCRIPT");
    }
    undef $fd;
    delete $self->{handles}{$params{file}};
    $Trace->exit();
}

#-------------------------------------------------------------------
sub read {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
                  dhandle => {type => SCALAR, optional=>1}, # maybe cache the current directory???
                  file    => {type => SCALAR},
                  offset  => {type => SCALAR, default => 0},
                  count   => {type => SCALAR | UNDEF, optional => 1},
                  buf_size => {type => SCALAR | UNDEF, default => 8192},
              });
    $Trace->entry();

    my $fname = $params{file};
    my $offset = $params{offset};
    my $length = $params{count};
    my $bufSize = $params{buf_size};
    my $fd = $self->{handles}{$params{file}};
    my $contents;

    if (!$fd) {
        $Trace->fatal(msg=>"Could not find the handle for $params{file}. Maybe a call to open() has not been made yet.", result=>"SCRIPT");
    }
                              
    
    if (!defined($length)) {
        my $st = stat $fname;
        $length = $st->size;
    }
    my $bytesRemaining = $length || 0;
    my $bytesRead = 0;
    $Trace->trace(msg=>sprintf("    reading %3.2f%s at offset $offset from $fname in $bufSize increments\n", size_in_unit($bytesRemaining)));

    if (!defined(sysseek($fd, $offset, 0))) {
        $Trace->fatal(msg=>"Error seeking to offset $offset in $fname: $!\n", result=>"FAIL");
    }

    while ($bytesRemaining > 0) {
        my $bytes = ($bytesRemaining < $bufSize) ? $bytesRemaining : $bufSize;
        my $buf;
        my $read;
        $Trace->debug(msg=>"      reading $bytes bytes, $bytesRemaining still to go\n");

        if (!($read=sysread($fd, $buf, $bytes))) {
            $Trace->fatal(msg=>"Error reading $bytes bytes from $fname: $!\n", result=>"FAIL");
        }

        $bytesRemaining -= $read;
        $bytesRead += $read;
        $Trace->debug(msg=>"       $read bytes read, $bytesRead completed\n");
        $contents .= $buf;
    }        

    $Trace->exit(msg=>"return $bytesRead bytes read");
    return (wantarray) ? ($bytesRead, $contents) : $bytesRead;
}

#-------------------------------------------------------------------
sub write {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
               dhandle => {type => SCALAR, optional=>1}, # maybe cache the current directory???
               file    => {type => SCALAR},
               offset  => {type => SCALAR, default => 0},
               count   => {type => SCALAR | UNDEF, optional => 1},
               data  => {type => SCALAR | UNDEF, optional => 1},
               buf_size => {type => SCALAR | UNDEF, default => 8192},
              });
    $Trace->entry();

    my $fname = $params{file};
    my $offset = $params{offset};
    my $bufSize = $params{buf_size};
    my $fd = $self->{handles}{$params{file}};
    my $written;
    my $buf = $params{data};

    my $length;
    if (!defined($length)) {
        $length = length($params{data});
    }
    my $bytesRemaining = $length;

    if (!$fd) {
        $Trace->fatal(msg=>"Could not find the handle for $params{file}", result=>"SCRIPT");
    }
    
    $Trace->trace(msg=>sprintf("    writing %3.2f%s at offset $offset from $fname in $bufSize increments\n", size_in_unit($length)));

    if (!defined(sysseek($fd, $offset, 0))) {
        $Trace->fatal(msg=>"Error seeking to offset $offset in $fname: $!\n", result=>"FAIL");
    }

    while ($bytesRemaining > 0) {
        my $bytes = ($bytesRemaining < $bufSize) ? $bytesRemaining : $bufSize;
        $Trace->debug(msg=>"      writing $bytes bytes at offset $offset, $bytesRemaining still to go\n");
        if (!($written=syswrite($fd, $buf, $bytes, $offset))) {
            $Trace->fatal(msg=>"Error writing $bytes bytes from $fname: $!\n", result=>"FAIL");
        }

        $offset += $written;
        $bytesRemaining -= $written;
        $Trace->debug(msg=>"       $written bytes written, $bytesRemaining remaining\n");
    }        

    $Trace->exit();
}

#-------------------------------------------------------------------
sub opendir {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
                             directory => {type => SCALAR },
			    }
			    );
    $Trace->entry(msg=>"directory=$params{directory}");
    my $handle;
    opendir($handle, $params{directory}) ||
    $Trace->fatal(msg=>"Could not open $params{directory}: $!");
    
    $Trace->exit(msg=>"return $handle");
    return $handle;
}

#-------------------------------------------------------------------
sub readdir {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
				handle => {type => GLOBREF},
                                directory => {type => SCALAR},
                                stat => {type => SCALAR | UNDEF, optional => 1 ,
                                        default => 1},
			    }
			    );
    $Trace->entry(msg=>"handle=$params{handle}");
    my @entries;

    foreach my $entry (readdir $params{handle}) {
        my %entry = ();
        $entry{name} = $entry;
        $entry{fhandle} = undef; # what should this be?
        if ($params{stat}) {
            $entry{name_attr} = $self->_stat(object=>"$params{directory}/$entry");
        }
        push(@entries, \%entry);
    }

    $Trace->exit();
    return \@entries;
}

#-------------------------------------------------------------------
sub closedir {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
				handle => {type => GLOBREF},
			    }
			    );
    $Trace->entry(msg=>"handle=$params{handle}");

    closedir($params{handle}) || $Trace->fatal(msg=>"Could not close the directory: $!");
    
    $Trace->exit();
}


#-------------------------------------------------------------------
sub ls {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
                             directory => {type => SCALAR },
                             stat => {type => SCALAR | UNDEF, optional => 1 },
			    }
			    );
    $Trace->entry();

    my $dir_handle = $self->opendir(directory => $params{directory});
    my $entries = $self->readdir(handle=>$dir_handle, %params);
    $self->closedir(handle=>$dir_handle);
    
    $Trace->exit();
    return $entries;
}

#-------------------------------------------------------------------
sub read_file {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
                  file    => {type => SCALAR},
                  offset  => {type => SCALAR, default => 0},
                  count   => {type => SCALAR | UNDEF, optional => 1},
                  buf_size => {type => SCALAR | UNDEF, default => 8192},
              });
    $Trace->entry();
    my $handle = $self->open(file=>$params{file}, mode=>"r");
    my ($bytes, $contents) = $self->read(%params);
    $self->close(file=>$params{file});
    $Trace->exit(msg=>"return $bytes bytes");
    return (wantarray) ? ($bytes, $contents) : $bytes;
}

#-------------------------------------------------------------------
sub write_file {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
                  file    => {type => SCALAR},
                  offset  => {type => SCALAR, default => 0},
                  count   => {type => SCALAR | UNDEF, optional => 1},
                  buf_size => {type => SCALAR | UNDEF, default => 8192},
                  data => {type => SCALAR | UNDEF, optional => 1},
              });
    $Trace->entry();
    my $handle = $self->open(file=>$params{file}, mode=>"w");
    $self->write(%params);
    $self->close(file=>$params{file});
    $Trace->exit();
}

#-------------------------------------------------------------------
sub unlink {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
                  file    => {type => SCALAR},
              });
    $Trace->entry();
    
    if (!unlink($params{file})) {
        $Trace->fatal(msg=>"Unable to rm '$params{file}': $!");
    }
    $Trace->exit();
}

#-------------------------------------------------------------------
sub _stat {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
               object    => {type => SCALAR},
              });
    $Trace->entry();
    
    my $stats = {};
    my $st = stat($params{object});
    $stats->{crtime} = {seconds=>$st->ctime, nseconds=>0};
    $stats->{mtime} = {seconds=>$st->mtime, nseconds=>0};
    $stats->{atime} = {seconds=>$st->atime, nseconds=>0};
    $stats->{rdev} = {specdata1=>$st->rdev, specdata2=>$st->rdev}; # not sure about this one
    $stats->{uid} = $st->uid;
    $stats->{gid} = $st->gid;
    $stats->{fileid} = $st->ino;
    $stats->{size} = $st->size;
    $stats->{nlink} = $st->nlink;
    $stats->{mode} = $st->mode & 07777;
    $stats->{used} = $st->blocks * 512;
    $stats->{fsid} = undef; # not sure about this one, should be $st->dev;
    if (-f $params{object}) {
        $stats->{type} = "1"; # FILE
    } elsif (-d $params{object}) {
        $stats->{type} = 2; # DIR
    } elsif (-b $params{object}) {
        $stats->{type} = 3; # BLK
    } elsif (-c $params{object}) {
        $stats->{type} = 4; # CHR
    } elsif (-l $params{object}) {
        $stats->{type} = 5; # LINK
    } elsif (-S $params{object}) {
        $stats->{type} = 6; # SOCK
    } elsif (-p $params{object}) {
        $stats->{type} = 7; # named pipe
    }
    
    $Trace->exit();
    return $stats;
}

sub is_file {
    my $self = shift;
    my (%params) = validate(
                            @_,
                            {
                             object => {type => HASHREF|SCALAR},
                            }
                           );
    $Trace->entry();
    my $rc = undef;
    my $object = $params{object};

    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            $Trace->fatal(msg=>"The passed-in object did not have the name_attr hash", result=>"SCRIPT");
        }
        if (!exists($object->{name_attr}{type})) {
            $Trace->fatal(msg=>"The passed-in object did not have the type field in the name_attr hash", result=>"SCRIPT");
        }
        $rc = $object->{name_attr}{type} == NFREG();
    } else {
        $rc = (-f $object);
    }
    $Trace->exit(msg=>($rc) ? "1" : "undef");
    return $rc;
}

sub is_readable {
    my $self = shift;
    my (%params) = validate(
                            @_,
                            {
                             object => {type => HASHREF|SCALAR},
                            }
                           );
    $Trace->entry();
    my $rc = undef;
    my $object = $params{object};

    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            $Trace->fatal(msg=>"The passed-in object did not have the name_attr hash", result=>"SCRIPT");
        }
        if (!exists($object->{name_attr}{type})) {
            $Trace->fatal(msg=>"The passed-in object did not have the type field in the name_attr hash", result=>"SCRIPT");
        }
        my $attr = $object->{name_attr};
        if ($attr->{mode} & 044) {
            $rc = 1;
        }
    } else {
        $rc = (-r $object);
    }
    $Trace->exit(msg=>($rc) ? "1" : "undef");
    return $rc;
}

sub is_writeable {
    my $self = shift;
    my (%params) = validate(
                            @_,
                            {
                             object => {type => HASHREF|SCALAR},
                            }
                           );
    $Trace->entry();
    my $rc = undef;
    my $object = $params{object};

    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            $Trace->fatal(msg=>"The passed-in object did not have the name_attr hash", result=>"SCRIPT");
        }
        if (!exists($object->{name_attr}{type})) {
            $Trace->fatal(msg=>"The passed-in object did not have the type field in the name_attr hash", result=>"SCRIPT");
        }
        my $attr = $object->{name_attr};
        if ($attr->{mode} & 022) {
            $rc = 1;
        }
    } else {
        $rc = (-w $object);
    }
    $Trace->exit(msg=>($rc) ? "1" : "undef");
    return $rc;
}

sub is_dir {
    my $self = shift;
    my (%params) = validate(
                            @_,
                            {
                             object => {type => HASHREF|SCALAR},
                            }
                           );
    $Trace->entry();
    my $rc = undef;
    my $object = $params{object};
    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            $Trace->fatal(msg=>"The passed-in object did not have the name_attr hash", result=>"SCRIPT");
        }
        if (!exists($object->{name_attr}{type})) {
            $Trace->fatal(msg=>"The passed-in object did not have the type field in the name_attr hash", result=>"SCRIPT");
        }
        $rc = $object->{name_attr}{type} == NFDIR();
    } else {
        $rc = (-d $object);
    }
    $Trace->exit(msg=>($rc) ? "1" : "undef");
    return $rc;
}
1;

