#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NASClient/NFS/Synth.pm#1 $

## @summary synthetic
## @author snider@netapp.com,gvg@netapp.com
## @status in progress
## @pod here

=head1 NAME

NASClient::NFS::Synth - Class definition for a Synthetic NFS clients

=head1 DESCRIPTION

Please see the documentation in C<NASClient> which describes the API for all
of the NASClient objects. This class file defines methods that support all
of the synthetic NFS clients.

=cut

package NASClient::NFS::Synth;
use strict;
use warnings;
use base qw(NASClient::NFS);
use Tharn qw($Log fatal_result config_result);
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use nfs::NfsDef qw(NFREG NFDIR);
use nfs::GenericAcl; # This is temp. Should be removed as soon as GenericAcl is fixed or other means are found
use Options;
my $Options = new Options;

$Options->parameters(
    definitions => [
        {name=>"SRCIP",
         description=>"the source IP address to use for the synthetic client",},
        {name        => "NFS_TIMEOUT",
         description => "NFS Timeout.",
         default     => 3,},
        {name        => "NFS_RETRIES",
         description => "Number of retries.",
         default     => 6,},]);

#-------------------------------------------------------------------
sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    $Log->enter();
    $Options->process();
    my $options = $Options->get_hash();
    my $self = { version => $options->{nfsversion} || 3,
                 proto => $options->{protocol},
                 srcip => $options->{SRCIP},
                 server => $options->{DATA_INTERFACES}[0] || $options->{default_data_interface},
                 retries => $options->{NFS_RETRIES},
                 timeout => $options->{NFS_TIMEOUT},
                 acregmin => 3, # this is for linux
               @_};
    bless($self, $class);
    
    if (! grep($_ == $self->{version}, qw(3 4))) {
        die config_result("NFS version '$self->{version}' is not supported yet");
    }
    if (!$self->{server}) {
        die config_result("Please specify a server in your call to new(), a FILER nate param or a DATA_INTERFACES nate param");
    }
    
    my %opts = ();

    if ($self->{srcip}) {
        %opts = (localaddr => $self->{srcip});
    }

    # NFSv4 requests need to be resent after reconnecting, otherwise
    # requests will be treated as bad ones and hence server will reset
    # the connection.
    if ( ($self->{version} == 4) && (! defined($opts{reconnect}) ) ) {
        $opts{reconnect} = 1;
    }

    eval {
        $self->{nfs} = nfs::NfsClient->new(hostname => $self->{server},
                                           version=>$self->{version}, 
                                           proto => $self->{proto},
                                           resvport => 0,
                                           retry => $self->{retries},
                                           timeout => $self->{timeout},
                                           %opts,
                                          );
    };
    if ($@) {
        die fatal_result("New NfsClient creation failed. Error: $@");
    }
    $self->_set_defaults();
    return $self;
}    


# A few generic Util functions 
#-------------------------------------------------------------------
# Convert ace list to compatible keys for GenericAcl library.
# GenericAcl library uses 'flags' instead of 'flag'
sub _synth_to_acl {
    my ($self, $acl) = @_;
    my $ace_cnt = 0;

    foreach (@{$acl->{ace}}) {
        $ace_cnt++;
        $_->{flags} = $_->{flag};
        delete $_->{flag};
    }
    $acl->{ace_cnt} = $ace_cnt;

    return $acl;
}

#-------------------------------------------------------------------
# Convert ace list to compatible keys for perl synthetic client.
# Synthetic client uses 'flag' instead of 'flags'
sub _acl_to_synth( $ ) {
    my ($self, $acl) = @_;
    
    foreach (@{$acl->{ace}}) {
        $_->{flag} = $_->{flags};
        delete $_->{flags};
    }

    return $acl;
}

# Not sure if these functions should live here - XXX - GV
#-------------------------------------------------------------------
sub print_acl {
    my ($self, %args) = @_;

    # XXX - GV
    # Hack to make this hashref into OBJECT to make GenericAcl happy for now
    my $acl = $args{acl};
    delete $args{acl};
    bless $acl;
    $args{acl} = $acl;

    nfs::GenericAcl::print_acl(%args);
}

#-------------------------------------------------------------------
sub set_acl_ace {
    my ($self, %args) = @_;

    # XXX - GV
    # Hack to make this hashref into OBJECT to make GenericAcl happy for now
    my $acl = $args{acl};
    delete $args{acl};
    bless $acl;
    $args{acl} = $acl;

    nfs::GenericAcl::set_acl_ace(%args);
}

#-------------------------------------------------------------------
sub get_stateid  {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
				fhandle => {type => SCALAR, },
			    }
			    );
    return $self->{_stateids}{$params{fhandle}};
}

#-------------------------------------------------------------------
sub set_stateid  {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
				fhandle => {type => SCALAR, },
                                stateid => {type => HASHREF, },
			    }
			    );
    $self->{_stateids}{$params{fhandle}} = $params{stateid};

    return;
}

#-------------------------------------------------------------------
sub delete_stateid  {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
				fhandle => {type => SCALAR, },
			    }
			    );
    delete $self->{_stateids}{$params{fhandle}};

    return;
}

#-------------------------------------------------------------------
# It would be nice to implement a timeout in the cache and perform the getattr when necessary for existing files
sub _get_handle {
    my $self = shift;
    my $class = ref($self);
    my (%params) = validate(
			    @_,
			    {
                             object => {type => SCALAR },
			    }
			    );
    $params{object} =~ s!/$!!;
    $Log->enter(msg=>"object=$params{object}");
    
    my $handle = "";
    my $attr = "";
    my $out;
    my $nfs = $self->getnfs();

    # need to find the current directory handle
    my $mountpoint = $self->{mountpoint}[0];
    my $cache_entry = $self->_get_cache_entry(path=>$mountpoint);
    my $dir_handle = $cache_entry->{handle};
    if (!$dir_handle) {
        die script_result("Did not find a filehandle for '$mountpoint' in the cache");
    }

    my $object = $params{object};
    if ($object eq $mountpoint) {
        $handle = $dir_handle;
        $attr = $cache_entry->{attr};
        goto done;
    }
        
    $object =~ s!$mountpoint/!!;
    $Log->trace("mountpoint=$mountpoint, object=$object");
    
    my $path = $mountpoint;
    foreach my $component (split(m!/!, $object)) {
        $path .= "/$component";
        my $cache_entry = $self->_get_cache_entry(path=>$path);
        if ($cache_entry->{handle}) {
            $handle = $cache_entry->{handle};
            $attr = $cache_entry->{attr};
            $Log->debug("found handle for $path");
        } else {
            $Log->trace("lookup $path $component");
            eval {
                $self->_exe(_cmd => "access",
                            fhandle=>$dir_handle, 
                            access => 0x1f );    ## 0x1f => read|lookup|modify|extend|delete|not-execute
                $out = $self->_exe(_cmd => "lookup",
                                   _error_ok => 1,
                                   name    => $component,
                                   dhandle => $dir_handle);
                $handle = $out->{fhandle};
                $attr = $out->{fattr};
            };
            if ($@) {
                $Log->trace("Lookup failed with " . Dumper($@));
                $handle = undef;
                $attr = undef;
                last;
            } elsif ($handle) {
                $self->_add_to_cache(path=>$path,
                                     handle => $handle,
                                     attr => $attr,
                                    );
            }
        }
        $dir_handle = $handle;
    }
  done:
    $Log->exit(($handle) ? "success" : "failure");
    return (wantarray) ? ($handle, $attr) : $handle;
}

sub is_file {
    my $self = shift;
    my (%params) = validate_mixed(
                                  @_,
                                  [],
                                  {
                                   object => {type => HASHREF|SCALAR},
                                  }
                                 );
    $Log->enter();
    my $rc = undef;
    my $object = $params{object};

    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            die fatal_result("The passed-in object did not have the name_attr hash");
        }
        if (!exists($object->{name_attr}{type})) {
            die fatal_result("The passed-in object did not have the type field in the name_attr hash");
        }
        $rc = $object->{name_attr}{type} == NFREG();
    } else {
        my ($handle,$attr) = $self->_get_handle(object=>$object);
        if ($attr && $attr->{type} == NFREG()) {
            $rc = 1;
        }
    }
    $Log->exit(($rc) ? "1" : "undef");
    return $rc;
}

sub is_readable {
    my $self = shift;
    my (%params) = validate_mixed(
                                  @_,
                                  [],
                                  {
                                   object => {type => HASHREF|SCALAR},
                                  }
                                 );
    $Log->enter();
    my $rc = undef;
    my $object = $params{object};
    my $attr;
    my $handle;
    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            die config_result("The passed-in object did not have the name_attr hash");
        }
        if (!exists($object->{name_attr}{type})) {
            die config_result("The passed-in object did not have the type field in the name_attr hash");
        }
        $attr = $object->{name_attr};
    } else {
        ($handle,$attr) = $self->_get_handle(object=>$object);
        $attr = $self->{attrs}{$handle} if ($handle);
    }
    if ($attr && $attr->{mode} & 044) {
        $rc = 1;
    }
    $Log->exit(($rc) ? "1" : "undef");
    return $rc;
}

sub is_writeable {
    my $self = shift;
    my (%params) = validate_mixed(
                                  @_,
                                  [],
                                  {
                                   object => {type => HASHREF|SCALAR},
                                  }
                                 );
    $Log->enter();

    my $rc = undef;
    my $object = $params{object};
    my $attr;
    my $handle;
    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            die config_result("The passed-in object did not have the name_attr hash");
        }
        if (!exists($object->{name_attr}{type})) {
            die config_result("The passed-in object did not have the type field in the name_attr hash");
        }
        $attr = $object->{name_attr};
    } else {
        ($handle,$attr) = $self->_get_handle(object=>$object);
    }
    if ($attr->{mode} & 022) {
        $rc = 1;
    }
    $Log->exit(($rc) ? "1" : "undef");
    return $rc;
}

sub is_dir {
    my $self = shift;
    my (%params) = validate_mixed(
                                  @_,
                                  [],
                                  {
                                   object => {type => HASHREF|SCALAR},
                                  }
                                 );
    $Log->enter();
    my $rc = undef;
    my $object = $params{object};
    if (ref($object) eq "HASH") {
        if (ref($object->{name_attr}) ne "HASH") {
            die config_result("The passed-in object did not have the name_attr hash");
        }
        if (!exists($object->{name_attr}{type})) {
            die config_result("The passed-in object did not have the type field in the name_attr hash");
        }
        $rc = $object->{name_attr}{type} == NFDIR();
    } else {
        my ($handle,$attr) = $self->_get_handle(object=>$object);
        if ($attr && $attr->{type} == NFDIR()) {
            $rc = 1;
        }
    }
    $Log->exit(($rc) ? "1" : "undef");
    return $rc;
}

#-------------------------------------------------------------------
sub _exe {
    # not using validate because %params will also contain unknown keys
    my ($self, %params) = @_;
    my $nfs = $self->getnfs();
    my $cmd = $params{_cmd};
    delete $params{_cmd};
    my $error_ok = $params{_error_ok};
    delete $params{_error_ok};

    if (!$cmd) {
        die config_result("Did not get a valid '_cmd' from \%params:".Dumper \%params);
    }
    my $out = undef;
    
    my $sleep = $self->{jukebox_sleep};
    do {
        sleep($sleep) if($@ =~ /ERR_JUKEBOX/);
        eval { $out = $nfs->$cmd(%params); };
    } while($self->{jukebox} && $@ =~ /ERR_JUKEBOX/);
        die config_result("'$cmd' failed with error \" $@ \" at \n".$self->get_caller_info(3)) if ($@ && !$error_ok);

    return $out;
}

sub _add_to_cache {
    my $self = shift;
    my (%params) = validate_mixed(
                                  @_,
                                  [],
                                  {
                                   path => {type => SCALAR},
                                   handle => {type => SCALAR},
                                   attr => {type => HASHREF | UNDEF},
                                  }
                                 );
    $Log->enter($params{path});
    my $handle = $params{handle};
    my $attr = $params{attr};
    my $path = $params{path};

    $self->{Cache}{$path} = {
                             attr => $attr,
                             handle => $handle,
                             entry => time,
                            };
    $Log->exit();
}

sub _get_cache_entry {
    my $self = shift;
    my (%params) = validate_mixed(
                                  @_,
                                  [],
                                  {
                                   path => {type => SCALAR},
                                  }
                                 );
    $Log->enter($params{path});
    my $path = $params{path};
    my $handle = $self->{Cache}{$path}{handle};
    my $attr = $self->{Cache}{$path}{attr};

    if ($handle) {
        # we need to timeout dir entries as well
        # timeout the file attributes
        # what happens if the filehandle changes?
        if ((time - $self->{Cache}{$path}{entry}) > $self->{acregmin}) {
            my $nfs = $self->getnfs();
            my $result = $self->_exe(_cmd => "getattr",
                                     fhandle=>$handle);
            $self->{Cache}{$path}{attr} = $result->{fattr};
            $attr = $result->{fattr};
        }
    }

    $Log->exit(($handle) ? "success" : "failure");
    return {handle=>$handle, attr=>$attr};
}

sub _clear_cache {
    my $self = shift;
    $Log->enter();
    $self->{Cache} = {};
    $Log->exit();
}

1;

__DATA__

# this code is in a DATA section and not loaded by the script. It is
# only left in the script for a while in case we need to reuse some
# of it
sub _try_nfs {
    my $self = shift;
    my %params = @_;

    $Log->enter();
    $self->{error} = undef;

    my $method = $params{_method};
    delete $params{_method};
    my $namespace = $params{_namespace};
    delete $params{_namespace};

    # remove our name from the namespace to force the call back through the autoloader
    # this has to occur very early before a die is called because it might get caught and this entry would still remain in the namespace
    no strict 'refs';
    if ($namespace->{$method}) {
#        print "deleting $method out of $namespace\n";
#        *{$namespace->{$method}} = undef;
#        delete $namespace->{$method};
    }
    my $nfs = $self->{nfs};
    if ($method && $nfs->can($method)) {
        my $retries = $self->{retries};
        my $timeout = $self->{timeout};

        my $out;
        do {
            eval {
                $out = $nfs->$method(%params);
                $Log->debug(Dumper($out));
                $self->{error} = $out->{status};
            };
            if ($@ || grep($self->{error}, qw(10008))) {
                if ($self->{hard_mount} || $retries > 0) {
                    if ($@) {
                        $Log->comment("status = $out->{status}");
                        $Log->warn("Retrying op '$method' due to error: $@");
                    } else {
                        $Log->warn(msg=>"Retrying op '$method' due to status code: $self->{error}");
                    }
                    sleep $timeout;
                }
            }
        } until (!$@ || (!$self->{hard_mount} && $retries-- == 0));
        die fatal_result("Server not responding:\n$@") if (!$retries);
        $Log->exit();
        return $out;
    } else {
        die fatal_result("Could not find '$method'");
    }
}

sub AUTOLOAD {
    my $method = our $AUTOLOAD;
    return if ($method =~ /::DESTROY/);
    $Log->enter();

    my $namespace;
    if ($method =~ /^(.+\:\:)(.+)$/) {
	$namespace = $1;
	$method = $2;
    } else {
	die config_result("do not know how to parse method: $method");
    }	

    # do this after cleaning up the namespace
    no strict 'refs';
    if ($method =~ /^_nfs_(.+)/) {
        $method = $1;
        $Log->trace("looking for $method");
        *$AUTOLOAD = sub { _try_nfs(@_, _method=>$method, _namespace=>$namespace); };
        goto &$AUTOLOAD;
    } else {
        my $object = shift;
        my $class = ref($object);
        die fatal_result("Attempt to call $class->$method() failed.\n");
    }
    $Log->exit();
} # end AUTOLOAD

1;
