package NBU::Backup;

use NBU::Command;
@ISA = qw(NBU::Command);

use strict;
use Data::Dumper;

############################# Accessors ##############################

sub set_i { 
	my ( $self, $i ) = @_;
	die "i must be equal to 1 or 0" unless $i == 1;
	$self->{i} = $i;
}
sub get_i { my ( $self ) = @_; return $self->{i} }

sub set_L {
	my ( $self, $path ) = @_;
	die "L must have a length" unless length $path;
	$self->{L} = $path;
}
sub get_L { my ( $self ) = @_; return $self->{L} }

sub set_s {
	my ( $self, $s ) = @_;
	die "s must be an instance of NBU::Schedule" unless ref($s) eq 'NBU::Schedule';
	$self->{s} = $s;
}
sub get_s { my ( $self ) = @_; return $self->{s} }

sub set_S {
	my ( $self, $S ) = @_;
	die "S must have a length" unless length $S;
	$self->{S} = $S;
}
sub get_S { my ( $self ) = @_; return $self->{S} }

sub set_w {
	my ( $self, $w ) = @_;
	die "w must have a true value" unless $w;
	$self->{w} = $w;
}
sub get_w { my ( $self ) = @_; return $self->{w} }

sub set_policy {
	my ( $self, $p ) = @_;
	die "policy must be an instance of NBU::Policy" unless ref( $p ) eq 'NBU::Policy';
	$self->{policy} = $p;
}
sub get_policy { my ( $self ) = @_; return $self->{policy} }

sub set_job {
	my ( $self, $job ) = @_;
	die "job must be an instance of Job" unless ref($job) eq 'Job';
	$self->{job} = $job;
}
sub get_job { my ( $self ) = @_; return $self->{job} }

sub to_create_string {

	my ( $self ) = @_;
	
	my $string = $NBU::Command::ROOT . 'bin/bpbackup';

	$string .= ' -i' if defined $self->{i};
	$string .= " -p $self->{policy}->{name}";
	$string .= " -L $self->{L}" if $self->{L};
	$string .= " -s $self->{s}->{name}";
	$string .= " -S $self->{S}";
	$string .= " -w" if defined $self->{w};

	return $string;
}

# No-op to preserver the Command.pm interface.
sub to_delete_string {

	my ( $self ) = @_;
	
	return undef;
}

sub create_results {

	my ( $self ) = @_;
	
	return 0;
}

sub delete_results {

	my ( $self ) = @_;
	
	return 0;
}

1;

=head1 NAME

  NBU::Backup - Start a NetBackup
  
=head1 DESCRIPTION

  This is a wrapper around the bpbackup command.  The object caches command 
  results internally.  For the most part, the variable names are the same as 
  the NBU command switches.
  
=head1 SYNOPSIS

  use NBU::Command;
  use NBU::StorageUnit;
  use NBU::Policy;
  use NBU::Schedule;  
  use NBU::Backup;

  my $s = NBU::StorageUnit->new;
  ... set values

  my $p = NBU::Policy->new;
  ... set values

  my $c = NBU::Schedule->new;
  ... set values

  my $b = NBU::Backup->new;
  $b->set_i( 1 );                           # Start immediate manual backup
  $b->set_policy( $p );                     # The policy to apply to this backup
  $b->set_L( '/u/todd/nbu/log' );           # The path to the backup log
  $b->set_s( $c );                          # The schedule to apply to this backup
  $b->set_S( 'perf280er.lab.netapp.com' );  # The master server
  $b->set_w( 1 );                           # Wait for the backup to complete.
  
  # Note: If you don't set_w, checking command results (eg $b->passed) isn't usefull.

  # Issue the create command
  $s->execute_create;

=head1 TODO

  - Handle the file paths.
  
=head1 DEPENDENCIES

  NBU::Command.pm
  Data::Dumper.pm
  
=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut

__END__


Backup for UNIX Client

/usr/openv/netbackup/bin/bpbackup \
	-i \
	-p djohnP \
	-s Full \
	-S perf280er.lab.netapp.com \
	-w

-i		Starts an Immediate-Manual Backup
-c or -p	Policy (GUI uses -c; man page says -p)	
-L <file>	Generate Progress Log (stored on client)
-s		Schedule
-S		Master Server
-w		Wait for Completion
-f <file>	File or Files to Restore (or you can just specify a filename)

In the above command the backup will include files that were included in
the policy, in this case djohnP. You can give the bpbackup command a list
of files and directories on the command line or by a file (-f). This may
be a better way to go than including everything in the policy.


/usr/openv/netbackup/bin/bpbackup \
	-i \
	-p djohnP \
	-s Differential-Inc \
	-S perf280er.lab.netapp.com \
	-w

See above.


Backup for Windows Client (same as UNIX Client)

For Windows clients to backup data over CIFS you must change who the
NBU Windows service (NetBackup Client Service) logs in as. By default
it logs in as "system" I changed it to "./administrator".

/usr/openv/netbackup/bin/bpbackup \
	-i \
	-p djohnP-W \
	-s Full \
	-S perf280er.lab.netapp.com \
	-w

-i		Starts an Immediate-Manual Backup
-c or -p	Policy (GUI uses -c; man page says -p)	
-L <file>	Generate Progress Log (stored on client)
-s		Schedule
-S		Master Server
-w		Wait for Completion
-f <file>	File or Files to Restore (or you can just specify a filename)

See UNIX comments about specifying files to be backed up on the command
line. It is as true for Windows as it is for UNIX.


/usr/openv/netbackup/bin/bpbackup \
	-i \
	-p djohnP-W \
	-s Differential-Inc \
	-S perf280er.lab.netapp.com \
	-w

See above.
