package NBU::Command;

use strict;
use Data::Dumper;

our $ROOT = '/usr/openv/netbackup/';
our @COMMANDS = ();
our $VERBOSE = 0;
our $NOOP = 0;
our $WARNED = 0;
our $SLEEP = 0;
our $SUDO = '/usr/local/bin/sudo';
our $STEM = "$SUDO $ROOT";

=head1 NAME

  NBU::Command - A NetBackup Command
  
=head1 DESCRIPTION

  This is an abstract base class from which all NBU commands inherit.
  
=head1 SYNOPSIS

  use NBU::Command;
  @ISA = qw(NBU::Command);

  # View CLI commands as they are executed and their results
  $NBU::Command::VERBOSE = 1;
  
  # execute_*() methods don't actually issue commands
  # Useful for debugging.
  $NBU::Command::NOOP = 1;
  
  # Pause 3 seconds between each command execution
  $NBU::Command::SLEEP = 3

=head1 METHODS

  The following methods are inherited:

=head2 new()

  Create the object instance.  Intended to be inherited.  Warns if the
  $NBU::Command::ROOT directory doesn't exist.
  
=cut

sub new {

	my ( $class ) = @_;
	
	my $self = bless {}, $class;

	if ( !-e $NBU::Command::ROOT && !$WARNED ) {
		warn "The netbackup directory doesn't exist: $NBU::Command::ROOT.  Commands can't be executed";
		$WARNED++;
	}

	# Keep track of every child instance.
	push( @NBU::Command::COMMANDS, $self );

	return $self;
}

=head2 get_results()
  
  Get the results of the most recent command execution.  Intended to 
  be inherited.  
  
=cut

sub get_results { my ( $self ) = @_; return $self->{results} }

=head2 passed()

  Takes no arguments.

  Checks whether the results returned by a command contained a string.  
  That's assumed to be an NBU error.  Returns 1 if no error (passed),
  0 if error.  Intended to be inherited.

=cut

sub passed {

	my ( $self ) = @_;

	my $results = $self->get_results;

	return 0 if $results =~ /\w/;
	return 1;
}

=head2 execute_create()

  Call the to_create_string() instance method.
  Execute the command.
  Cache the results in the 'results' instance variable.
  
  Intended to be inherited.

=cut

sub execute_create {

	my ( $self ) = @_;
	
	die "Can't execute create command." unless $self->can( 'to_create_string' );

	my $command = $self->to_create_string;
	
	return unless defined $command;

	if ( $NBU::Command::VERBOSE ) {
		print "\n";
		print "Type: " . ref( $self ) . "\n";
		print "Executing Create ";
		print "(DEBUGGING - NOT EXECUTING)" if $NBU::Command::NOOP;
		print ":\n";
		print "$command\n\n";
	}

	$self->{result} = `$NBU::Command::SUDO $command 2>&1` unless $NBU::Command::NOOP;
	
	if ( $NBU::Command::VERBOSE && !$NBU::Command::NOOP ) {
		print "Result:\n";
		print "$self->{result}\n\n";
	}
	
	die "execute_create() command failed in " . ref( $self ) 
		if ( $self->create_results && !$NBU::Command::NOOP );
}

=head2 execute_delete()
  
  Call the to_delete_string() instance method.
  Execute the command.
  Cache the results in the 'results' instance variable.
  
  Intended to be inherited.

=cut

sub execute_delete {

	my ( $self ) = @_;
	
	die "Can't execute delete command." unless $self->can( 'to_delete_string' );

	my $command = $self->to_delete_string;
	
	return unless defined $command;

	if ( $NBU::Command::VERBOSE ) {
		print "Type: " . ref( $self ) . "\n";
		print "Executing Delete ";
		print "(DEBUGGING - NOT EXECUTING)" if $NBU::Command::NOOP;
		print ":\n";
		print "$command\n\n";
	}
	
	$self->{result} = `$NBU::Command::SUDO $command` unless $NBU::Command::NOOP;
	
	if ( $NBU::Command::VERBOSE && !$NBU::Command::NOOP ) {
		print "Result:\n";
		print "$self->{result}\n\n";
	}
	
	die "execute_delete() command failed in " . ref( $self ) 
		if ( $self->delete_results && !$NBU::Command::NOOP );
}

=head1 CLASS METHODS

=head2 execute_create_all()

  Run execute_create() for all objects in the order in which the objects
  were instantiated.

=cut

sub execute_create_all {

	for ( @NBU::Command::COMMANDS ) {
		$_->execute_create;
		if ( $NBU::Command::SLEEP ) {
			sleep $NBU::Command::SLEEP;
			print "Sleeping $NBU::Command::SLEEP seconds\n" if $NBU::Command::VERBOSE;
		}
	}
}

=head2 execute_delete_all()

  Run execute_delete() for all objects in the REVERSE order in which the 
  objects were instantiated.
  
=cut

sub execute_delete_all {

	for ( reverse @NBU::Command::COMMANDS ) {
		$_->execute_delete;
		if ( $NBU::Command::SLEEP ) {
			sleep $NBU::Command::SLEEP;
			print "Sleeping $NBU::Command::SLEEP seconds\n" if $NBU::Command::VERBOSE;
		}
	}
}

1;

=head1 CLASS VARIABLES
  
  The following class (static) variables are available:

  $NBU::Command::ROOT     # Directory where NBU is installed
  $NBU::Command::VERBOSE  # Echo CLI commands when executed
  $NBU::Command::NOOP     # Don't execute commands
  $NBU::Command::WARNED   # Positive value if NBU not installed
  $NBU::Command::SLEEP    # Seconds to sleep between command execution
  $NBU::Command::SUDO     # Path to sudo command

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  
=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut
