package NBU::JobMonitor;

use Data::Dumper;
use strict;

use NBU::Command;
use NBU::Job;

=head1 NAME

  NBU::JobMonitor - Monitor NBU Backup and Restore Statuses

=head1 DESCRIPTION

  This starts out as a wrapper around the bpdpjobs command.  This class
  works differently than the rest of the NBU classes.  It does NOT 
  inherit from NBU::Command.
  
  It is used to monitor the status of ongoing and completed jobs.

=head1 SYNOPSIS

  use NBU::JobMonitor;
  
  my $m = NBU::JobMonitor->new;
  
  # Fetch the jobs from the database; cache them internally
  $m->get_jobs;		
  
  # Fetch job details with the highest JobID
  my $last = $m->get_last_job;
  
  # Are the jobs done?
  print "All Done\n" if $m->all_done( \@job_objects );
  
  # Are the jobs done and successfully completed (Status = 0)?
  print "All Done and Successful\n" if $m->all_done_success( \@job_objects );
  
  # Check 5 times whether jobs are finished; sleep 300 seconds between checks
  print "All jobs have finished\n" if $m->wait_all( \@jobs, 5, 300 );
  
  # For debugging
  print $m->to_string, "\n";

  ################ Here's the way this is used with Backup Objects ##################
  
  # Admittedly, not pretty.  This is the consequence of ommitting the -m flag.
  # This means the Backup or Restore commands return immediately.  They don't 
  # block and wait for the job to finish.  That means WE have to wait for the 
  # jobs to finish instead.
  
  # Check job status PRIOR to backup
  my $m0 = NBU::JobMonitor->new;
  $m0->get_jobs;
  my $j0 = $m0->get_last_job;

  # Make a Backup object.  Note, the absence of set_m().  
  my $b = NBU::Backup->new;
  $b->set_i( 1 );
  $b->set_policy( $setup->{$p_name} );
  $b->set_L( $setup->{log} );
  $b->set_s( $setup->{$s_name} );
  $b->set_S( $setup->{client_name} );
  $b->execute_create;
	
  # Check job status AFTER backup
  my $m1 = NBU::JobMonitor->new;
  $m1->get_jobs;
  my $j1 = $m1->get_last_job;
		
  my @jobs = ();
  
  # Validate the job ID, just in case.
  if ( $m1->job_id_valid( $m0 ) ) {
		
     $b->set_job( $j1 );
     push( @jobs, $j1 );
		
  } else {
     print "Couldn't reliably obtain a JobID.\n"; 
  }
  
  my $iter = 5;    # Check 5 times for the jobs to finish
  my $time = 50;   # Wait 50 seconds between checks
  
  if ( $m2->wait_all( \@jobs, $iter, $time ) ) {
			
     if ( $m2->all_done_success( \@jobs ) ) {
        Tharn::logresult( 'FAIL', "At least one completed process finished with non-zero status." );			
     } else {}   # PASS by default
		
  } else {
     Tharn::logresult( 'WARN', "Not all of the backups finished before timeout." );
  }
  
=head1 METHODS

=head2 new()

  Constructor.
  
  Takes no arguments.  Checks whether NBU is installed.  Returns the object reference.
    
=cut

sub new {
	my ( $class ) = @_;	
	
	if ( !-e $NBU::Command::ROOT && !$NBU::Command::WARNED ) {
		die "The netbackup directory doesn't exist: $NBU::Command::ROOT.  Commands can't be executed";
		$NBU::Command::WARNED++;
	}

	return bless {}, $class;
}

=head2 get_jobs()

  Instance method.

  Takes no arguments.  Returns a hash reference, keyed by JobID, containing
  the history of NBU jobs (the equivalent of Activity Monitor in the GUI).

=cut

sub get_jobs {

	my ( $self ) = @_;

	my $command = $NBU::Command::STEM . "bin/admincmd/bpdbjobs";
	
	my $results = `$command`;

	my $data = {};

	while ( $results =~ /(.*?)\n/g ) {
	
		next if $1 =~ /JobID/;
		my $line = $1;
		#print "*$1*\n";
		#$line =~ /\s+(.+)\s+(.+)\s+(.+)\s+(.{7})(.{11})(.{16})(.*?) (.*?) (.*?)$/;
		$line =~ /^\s+(\d+?) +(\w+?) +(\w+?) +(\d*)/;
		#print "$1|$2|$3|$4|$5|$6|$7|$8|$9\n";

		my $j = Job->new;
		my $JobID = $self->strip($1);
		$j->set_JobID( $JobID );
		$j->set_Type( $self->strip($2) );
		$j->set_State( $self->strip($3) );
		$j->set_Status( $self->strip($4) );
		
		$data->{$JobID} = $j;
	}
	
	#return $data;
	$self->{jobs} = $data;
}

# Strip leading or trailing whitespace.

sub strip {

	my ( $self, $string ) = @_;
	
	$string =~ s/(.*?)\s+/$1/;
	$string =~ s/\s+(.*?)/$1/;
	
	return $string;
}


=head2 get_last_job()

  Instance method.

  Takes a reference to a hash as returned by get_jobs() and returns the JobID
  of the last job (highest JobID) and its parameters.
  
=cut

sub get_last_job {

	my ( $self ) = @_;
	
	for ( reverse sort { $a <=> $b } keys %{$self->{jobs}} ) {
		return $self->{jobs}{$_};
	}
}

=head2 all_done()

  Instance method.

  Takes a reference to a list of JobIDs.  Returns one if all jobs are 'Done', 
  zero otherwise.
  
=cut

sub all_done {

	my ( $self, $jobs ) = @_;
	
	for ( @$jobs ) {
		next unless defined $_;
		return 0 unless $_->get_State eq 'Done';
	}
	
	return 1;
}	

=head2 all_done_success()

  Instance method.
  
  Takes a reference to a list of JobIDs.  Returns one if all jobs' States are
  'Done' and their statuses are 0.

=cut

sub all_done_success {

	my ( $self, $jobs ) = @_;

	for ( @$jobs ) {
		next unless defined $_;
		return 0 unless $_->done_success;
	}

	return 1;
}

=head2 wait_all()

  Instance method.
  
  Takes a reference to an array of jobs, a loop count, and a sleep time.  Returns 
  1 when all jobs are done.  Returns 0 if it times out.
  
=cut

sub wait_all {

	my ( $self, $jobs, $iter, $sleep ) = @_;

	for ( 1..$iter ) {
	
		print "JobMonitor is sleeping $sleep seconds...\n";
		sleep $sleep;

		my $m = NBU::JobMonitor->new;
		$m->get_jobs;
		my $new_jobs = $m->update_status( $jobs ); 
	
		if ( $m->all_done( $new_jobs ) ) {
			return 1;
		}
	}

	return 0;	
}

=head2 wait()

  Instance method.
  
  Takes a reference to an array of Jobs.  Returns 1 when all jobs are done.  Returns
  0 if it times out.  Time out is 2 hours.
  
=cut

sub wait {

	my ( $self, $jobs ) = @_;

	return $self->wait_all( $jobs, 120, 60 );
}

=head2 to_string()

  Instance method.
  
  Takes no arguments.  Returns a string representation of all jobs.
  
=cut

sub to_string {

	my ( $self ) = @_;
	
	my $string = "All Jobs:\n\n";
	
	for ( keys %{$self->{jobs}} ) {
		$string .= $self->{jobs}{$_}->to_string . "\n";
	}
	
	return $string;
}

=head2 job_id_valid()

  Instance method.
  
  Takes a JobMonitor object.  Compares its last job's JobID to this object's last
  job's JobID to assure it's valid.  Returns 1 if valid.  Returns 0 if not.

=cut

sub job_id_valid {

	my ( $self, $m ) = @_;
	
	if ( ( $m->get_last_job->get_JobID + 1 ) == $self->get_last_job->get_JobID ) {
		return 1;
	}	
	
	return 0;
}

=head2 update_status()

  Instance method.
  
  Takes an array of jobs, and returns an equivalent array from its internal cache,
  effectively updating the job statuses.

=cut

sub update_status {

	my ( $self, $jobs ) = @_;
	
	my @new = ();
	for ( @$jobs ) {
		next unless defined $_;
		my $JobID = $_->get_JobID;
		push( @new, $self->{jobs}{$JobID} ); 		 
	}
	
	return \@new;
}

1;

__END__

-bash-2.05b$ r /usr/openv/netbackup/bin/admincmd/bpdbjobs
JobID    Type  State Status                            Policy Schedule                   Client           Dest Media Svr Active PID
    2 Restore   Done      0                                            perf280er.lab.netapp.com                                2485
    3  Backup   Done      0    p2_basic_smoke_1123200370_2890        - perf280er.lab.netapp.com perf280er.lab.netapp.com       3284
    4 Restore   Done      0                                            perf280er.lab.netapp.com                                3413
    5  Backup   Done      0    p2_basic_smoke_1123261643_8965        - perf280er.lab.netapp.com perf280er.lab.netapp.com      10188
    6 Restore   Done      0                                            perf280er.lab.netapp.com                               10697
    7  Backup   Done      0   p2_basic_smoke_1123262076_11613        - perf280er.lab.netapp.com perf280er.lab.netapp.com      12544
    8 Restore   Done      0                                            perf280er.lab.netapp.com                               13003
    9  Backup   Done    156 p2_four_backups_1123277379_176581        - perf280er.lab.netapp.com perf280er.lab.netapp.com      19673
   10  Backup   Done      0  p2_four_backups_1123280849_19451        - perf280er.lab.netapp.com perf280er.lab.netapp.com       3879
   11  Backup   Done      0  p2_four_backups_1123280849_19452        - perf280er.lab.netapp.com perf280er.lab.netapp.com       4446
   12  Backup   Done      0  p2_four_backups_1123280849_19453        - perf280er.lab.netapp.com perf280er.lab.netapp.com       4989
   13  Backup   Done      0  p2_four_backups_1123280849_19454        - perf280er.lab.netapp.com perf280er.lab.netapp.com       5559
   14  Backup   Done      0 p2_four_backups_1123283064_133611        - perf280er.lab.netapp.com perf280er.lab.netapp.com      15206
   15  Backup   Done      0 p2_four_backups_1123283064_133612        - perf280er.lab.netapp.com perf280er.lab.netapp.com      15962
   16  Backup   Done      0 p2_four_backups_1123283064_133613        - perf280er.lab.netapp.com perf280er.lab.netapp.com      16628
   17  Backup   Done      0 p2_four_backups_1123283064_133614        - perf280er.lab.netapp.com perf280er.lab.netapp.com      17323
   18  Backup   Done      0  p2_four_backups_1123295920_20731        - perf280er.lab.netapp.com perf280er.lab.netapp.com       3877
   19  Backup   Done      0  p2_four_backups_1123295920_20732        - perf280er.lab.netapp.com perf280er.lab.netapp.com       4649
   20  Backup   Done      0  p2_four_backups_1123295920_20733        - perf280er.lab.netapp.com perf280er.lab.netapp.com       5402
   21  Backup   Done      0  p2_four_backups_1123295920_20734        - perf280er.lab.netapp.com perf280er.lab.netapp.com       6172
   22  Backup   Done      0  five_gig_backup_1123299010_16074        - perf280er.lab.netapp.com perf280er.lab.netapp.com      16862
   23 Restore   Done      0                                            perf280er.lab.netapp.com                               17690
   24  Backup   Done      0   five_gig_backup_1123347729_5728        - perf280er.lab.netapp.com perf280er.lab.netapp.com       6246
   25 Restore   Done      0                                            perf280er.lab.netapp.com                                6758
   26  Backup Active          five_gig_backup_1123350353_7383        - perf280er.lab.netapp.com perf280er.lab.netapp.com       7907
