package NBU::PerfStats;



use strict;
use Data::Dumper;


=head1 NAME

  NBU::PerfStats - Parse a Filer Performance Statistics File
  
=head1 DESCRIPTION

  This is intended to be used on the output of perf-sysstat.thpl, which produces
  a file containing sysstat outpout.

=head1 SYNOPSIS

  use NBU::PerfStats;
  
=head2 new()

  Constructor.
  
  Takes no arguments.  Returns the object reference.
  
=cut

sub new {	
	my ( $class ) = @_;
	
	my $self = {};

	$self->{total_net_in} = 0;
	$self->{avg_net_in} = 0;
	$self->{lines} = [];
	$self->{good_data} = [];

	return bless $self, $class;
}

# Accessors

sub set_filename {
	my ( $self, $filename ) = @_;
	die "filename must have a length" unless length $filename;
	$self->{filename} = $filename;
}
sub get_filename { my ( $self ) = @_; return $self->{filename} }

sub get_total_net_in { my ( $self ) = @_; return $self->{total_net_in} }

sub get_avg_net_in { my ( $self ) = @_; return $self->{avg_net_in} }

sub add_line {
	my ( $self, $line ) = @_;
	push( @{$self->{lines}}, $line );
}
sub get_lines { my ( $self ) = @_; return $self->{lines} }

sub get_good_data { my ( $self ) = @_; return $self->{good_data} }

=head2 parse_file()

  Instance method.

  Takes no arguments.  Parses and caches the file named in 'filename', ignoring
  headings.  Returns nothing.

=cut

sub parse_file {
	my ( $self ) = @_;
	
	open( STATS, "<$self->{filename}" ) || die "Couldn't open file: $self->{filename}";
	while ( <STATS> ) {
		
		next unless /\s*\d/;
		#print;	
		
		my @line = split( /\s+/, $_ );
		shift @line;
		
		$self->add_line( \@line );
		
		#print Dumper \@line;
	}
	close STATS;
}

=head2 trim_data()

  Instance method.
  
  Takes an optional significance level in kilobytes/second.  This removes statistics
  from both ends of the statistics array and saves the modified data internally.
  Returns nothing.
  
=cut

sub trim_data {

	my ( $self, $sig ) = @_;	
	
	$sig ||= 5;
	
	my @data = @{$self->{lines}};
	my @fore = ();
	my @aft = ();
	
	# Remove insignificant stats from the top
	my $count = 0;
	for ( @data ) {
		
		if ( $_->[5] < $sig ) {
			$count++;
			next;
		} else {
			last;
		}
	}
	
	for ( 1..$count ) {
		shift @data;
	}
	
	# Remove insignificant stats from the bottom
	my $rev_count = 0;
	for ( reverse @data ) {
	
		if ( $_->[5] < $sig ) {
			$rev_count++;
			next;
		} else {
			last;
		}
	}
	
	for ( 1..$rev_count ) {
		pop @data;
	}
	
	my $total = 0;
	my $average = 0;
	
	$self->{good_data} = \@data;
}

=head2 calc_net_in()

  Instance method.
  
  Takes no arguments.  Calculates total net in kilobytes and average net in kilobytes
  and stores the values internally.

=cut

sub calc_net_in {

	my ( $self ) = @_;
	
	for ( @{$self->{good_data}} ) {	
		$self->{total_net_in} += $_->[5];
	}
		
	if ( @{$self->{good_data}} ) {
		$self->{avg_net_in} = $self->{total_net_in} / @{$self->{good_data}};
	}
}

=head2 to_string()

  Instance method.
  
  Takes no arguments.  Returns a string representation of the object.

=cut

sub to_string {

	my ( $self ) = @_;
	
	my $string = '';
	$string .= "Total Net In (kilobytes):       " . $self->get_total_net_in . "\n";
	$string .= "Average Net In (kilobytes/sec): " . $self->get_avg_net_in . "\n";

	return $string;
}


1;


__END__

  0     1     2     3      4       5    6      7    8        9   10    11    12   13   14   15    16     17    18    19   20
  =============================================================================================================================

 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     0     0     0       0     1     2      8     32     0     0   >60  100%   0%  -    3%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     0     0     0       0     1     2      8      0     0     0   >60  100%   0%  -    2%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     0     0     0       0     1     2     16      0     0     0   >60  100%   0%  -    0%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     0     0     0       0     1     2      8     24     0     0   >60  100%   0%  -    2%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     0     0     0       0     1     2      0      8     0     0   >60  100%   0%  -    6%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  3%     0     0     0       0     1     2     71    230     0     0   >60  100%   7%  T    6%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     0     0     0       0     1     2      8     24     0     0   >60  100%   0%  -    2%      0     0     0     0     0
 CPU   NFS  CIFS  HTTP   Total    Net kB/s   Disk kB/s     Tape kB/s Cache Cache  CP   CP Disk   DAFS   FCP iSCSI   FCP  kB/s
                                  in   out   read  write  read write   age   hit time  ty util                       in   out
  2%     1     0     0       1     1     2     16      0     0     0   >60  100%   0%  -    0%      0     0     0     0     0
