package NBU::Policy;

use NBU::Command;
@ISA = qw(NBU::Command);

use strict;
use Data::Dumper;

############################# Accessors ##############################

sub set_name {
	my ( $self, $name ) = @_;
	die "name must have a length" unless length $name;
	$self->{name} = $name;
}
sub get_name { my ( $self ) = @_; return $self->{name} }

sub set_storage_unit {
	my ( $self, $storage_unit ) = @_;
	die "storage_unit must be an instance of NBU::StorageUnit" unless ref( $storage_unit ) eq 'NBU::StorageUnit';
	$self->{storage_unit} = $storage_unit;
}
sub get_storage_unit { my ( $self ) = @_; return $self->{storage_unit} }

# String to create a policy.
sub to_create_string {

	my ( $self ) = @_;
	
	my $string = $NBU::Command::STEM . 'bin/admincmd/bppolicynew';
	$string .= " $self->{name}";
	$string .= " -M $self->{storage_unit}{M}";
	
	return $string;
}

# String to delete an existing policy.
sub to_delete_string {

	my ( $self ) = @_;

	my $string = $NBU::Command::STEM . 'bin/admincmd/bppldelete';
	$string .= " $self->{name}";
	$string .= " -M $self->{storage_unit}{M}";
	
	return $string;
}

sub create_results {
	
	my ( $self ) = @_;
		
	return 0 if $NBU::Command::NOOP;
		
	my $string = $NBU::Command::STEM . 'bin/admincmd/bppllist';
	
	my $out = `$string`;
	
	while ( $out =~ /(.*)\n/g ) {
		return 0 if $1 eq $self->{name};
	}
	
	return 1;
}

sub delete_results {

	my ( $self ) = @_;
	
	return 0 if $NBU::Command::NOOP;
	
	my $string = $NBU::Command::STEM . 'bin/admincmd/bppllist';
	
	my $out = `$string`;
	
	while ( $out =~ /(.*)\n/g ) {
		return 1 if $1 eq $self->{name};
	}
	
	return 0;
}

1;

=head1 NAME

  NBU::Policy - A NetBackup Policy
  
=head1 DESCRIPTION

  This is a wrapper around the bppolicynew and bppldelete commands.  The object
  caches command results internally.  For the most part, the variable names
  are the same as the NBU command switches.
  
  Note, in order to have a useable policy, you must also 'decorate' the policy
  with object from these classes:
  
  NBU::Info         # General configuration options
  NBU::Schedule     # The backup schedule
  NBU::Include      # Directories/Files to backup
  NBU::Client       # Clients from which to backup
  
=head1 SYNOPSIS

  use NBU::Command;
  use NBU::Policy;
  use NBU::StorageUnit;
  
  $NBU::Command::VERBOSE = 1;

  my $s = NBU::StorageUnit->new;
  ... set values

  # For both p2 and p3
  
  my $p = NBU::Policy->new;         
  $p->set_name( 'todd_policy' );   # The policy name
  $p->set_storage_unit( $s );      # The storage unit to which to apply the policy

  # Create the policy
  $p->execute_create;
  print $p->get_results unless $p->passed;
  
  # Delete the policy
  $p->execute_delete;
  print $p->get_results unless $p->passed;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  NBU::Command.pm
  Data::Dumper.pm
  
=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut

__END__


Add/Configure Policy

/usr/openv/netbackup/bin/admincmd/bppolicynew djohnP \
	-M perf280er.lab.netapp.com

djohnP		Policy Name
-M		Master Server


Add/Configure Policy

/usr/openv/netbackup/bin/admincmd/bppolicynew djohnP-W \
	-M perf280er.lab.netapp.com

djohnP-W	Policy Name
-M		Master Server

Removing Policies and Storage Units

/usr/openv/netbackup/bin/admincmd/bppldelete djohnP \
	-M perf280er.lab.netapp.com
	
------------------------------------------------------------------------

Here is the commands I use to create a policy.

Configure NDMP Host
-------------------

/usr/openv/volmgr/bin/tpconfig -add -nh <NAS HOST> -user_id <userid> -password <password> 
/usr/openv/volmgr/bin/tpconfig -add -nh <NAS HOST> -user_id <userid> -password <password>

/usr/openv/volmgr/bin/tpautoconf -verify <secondary nas host>

Create a snapvault storage unit
-------------------------------

/usr/openv/netbackup/bin/admincmd/bpstuadd -label <name of storage unit> -path <secondary volume> 
-nh <NAS secondary name>  -dt 3 

Create a policy 
---------------

/usr/openv/netbackup/bin/admincmd/bppolicynew <name of policy> 
/usr/openv/netbackup/bin/admincmd/bpplinclude <name of policy>  -add "qtree cifs or shared name" 
/usr/openv/netbackup/bin/admincmd/bpplinfo <name of policy> -modify -enable_pfi 1 -use_data_mover 1 -data_mover_type 3 -fi 1 -offhost_backup 1 -snapshot_method NAS_Snapshot -snapshot_method_args max_snapshots=<max number of snapshots to keep>  -ct Standard -multiple_streams 1 -collect_tir_info 0 -collect_bmr_info 0 -residence <storage utit> 
/usr/openv/netbackup/bin/admincmd/bpplclients <name of policy> -add steeljacket Solaris Solaris8 
/usr/openv/netbackup/bin/admincmd/bpplsched <name of policy> -add full -st FULL -rl 3  -synthetic 0 -pfi_fast_recovery 1 

Start backup
------------

/usr/openv/netbackup/bin/bpbackup -i -p <name of policy> -w 

Start restore: 
/usr/opnev/netbackup/bin/bprestore  -C <client name> -D <client name> <restore directory or file>

I hope this helps.
