package NBU::StorageUnit;

use NBU::Command;
@ISA = qw(NBU::Command);

use strict;
use Data::Dumper;

my @order = qw(M label path odo host);

############################# Accessors ##############################

sub set_M {
	my ( $self, $M ) = @_;
	die "M must have a length" unless length $M;
	$self->{M} = $M;
}
sub get_M { my ( $self ) = @_; return $self->{M} }

sub set_label {
	my ( $self, $label ) = @_;
	die "label must have a length" unless length $label;
	$self->{label} = $label;
}
sub get_label { my ( $self ) = @_; return $self->{label} }

sub set_path {
	my ( $self, $path ) = @_;
	die "path must have a length" unless length $path;
	$self->{path} = $path;
}
sub get_path { my ( $self ) = @_; return $self->{path} }

sub set_cj {
	my ( $self, $cj ) = @_;
	die "cj must be a positive integer" unless $cj > 0;
	$self->{cj} = $cj;
}
sub get_cj { my ( $self ) = @_; return $self->{cj} }

sub set_odo {
	my ( $self, $odo ) = @_;
	die "odo must be 1 or 0" unless ( $odo == 1 || $odo == 0 );
	$self->{odo} = $odo;
}
sub get_odo { my ( $self ) = @_; return $self->{odo} }

sub set_host {
	my ( $self, $host ) = @_;
	die "host must have a length" unless length $host;
	$self->{host} = $host;
}
sub get_host { my ( $self ) = @_; return $self->{host} }

sub set_maxmpx {
	my ( $self, $maxmpx ) = @_;
	die "maxmpx must be an integer between 0 and 32, inclusive" unless ( $maxmpx >= 0 && $maxmpx <= 32 );
	$self->{maxmpx} = $maxmpx;
}
sub get_maxmpx { my ( $self ) = @_; return $self->{maxmpx} }

sub set_mfs {
	my ( $self, $mfs ) = @_;
	die "mfs must be and integer between 20 and 524288, inclusive" unless ( $mfs >= 20 && $mfs <= 524288 );
	$self->{mfs} = $mfs;
}
sub get_mfs { my ( $self ) = @_; return $self->{mfs} }

sub set_dt {
	my ( $self, $type ) = @_;
	die "dt must equal 2 or 3 if set" unless ( $type == 2 || $type == 3 );
	$self->{dt} = $type;
}
sub get_dt { my ( $self ) = @_; return $self->{dt} }

sub set_nh {
	my ( $self, $nearstore ) = @_;
	die "nh must have a length" unless length $nearstore;
	$self->{nh} = $nearstore;
}
sub get_nh { my ( $self ) = @_; return $self->{nh} }

sub set_cf {
	my ( $self, $cf ) = @_;
	die "cf must be a positive integer" unless $cf > 0;
	$self->{cf} = $cf;
}
sub get_cf { my ( $self ) = @_; return $self->{cf} }

# Create the create storage unit command.
sub to_create_string {

	my ( $self ) = @_;
	
	my $string = $NBU::Command::ROOT . 'bin/admincmd/bpstuadd';
	for ( @order ) {
		$string .= " -$_ $self->{$_}";
	}
	
	$string .= " -cf $self->{cf}" if defined $self->{cf};
	$string .= " -maxmpx $self->{maxmpx}" if defined $self->{maxmpx};
	$string .= " -mfs $self->{mfs}" if defined $self->{mfs};
	$string .= " -cj $self->{cj}" if defined $self->{cj};
	$string .= " -dt $self->{dt}" if defined $self->{dt};
	$string .= " -nh $self->{nh}" if defined $self->{nh};
	
	return $string;
}

# Create the delete storage unit command.
sub to_delete_string {

	my ( $self ) = @_;

	my $string = $NBU::Command::ROOT . 'bin/admincmd/bpstudel';
	for ( 'M', 'label' ) {
		$string .= " -$_ $self->{$_}";
	}

	return $string;
}

# Affirm the label exists
sub create_results {

	my ( $self ) = @_;
	
	return 0 if $NBU::Command::NOOP;
	
	my $string = $NBU::Command::STEM . 'bin/admincmd/bpstulist';
	
	my $out = `$string`;
	
	while ( $out =~ /(.*)\n/g ) {
		my @line = split( /\s+/, $1 );
		return 0 if $line[0] eq $self->{label};
	}
	
	return 1;
}

# Affirm the label doesn't exist
sub delete_results {

	my ( $self ) = @_;
	
	return 0 if $NBU::Command::NOOP;
	
	my $string = $NBU::Command::STEM . 'bin/admincmd/bpstulist';
	
	my $out = `$string`;
	
	while ( $out =~ /(.*)\n/g ) {
		my @line = split( /\s+/, $1 );	
		return 1 if $line[0] eq $self->{label};
	}

	return 0;
}

1;

=head1 NAME

  NBU::StorageUnit - A NetBackup Storage Unit
  
=head1 DESCRIPTION

  This is a wrapper around the bpstuadd and bpstudel commands.  The object
  caches command results internally.  For the most part, the variable names
  are the same as the NBU command switches.
  
=head1 SYNOPSIS

  use NBU::Command;
  use NBU::StorageUnit;

  $NBU::Command::VERBOSE = 1;

  # The only difference between p2 and p3 is 'dt'
  
  my $s = NBU::StorageUnit->new;
  $s->set_M( 'perf280er.lab.netapp.com' );    # The Master Server
  $s->set_label( 'browne1' );                 # The storage unit label
  $s->set_path( '/vol/nbumods' );             # The destination path on the Filer
  $s->set_cj( 6 );                            # Maximum number of jobs
  $s->set_cf( 1 );                            # Enable block sharing
  $s->set_odo( 1 );                           # On demand only
  $s->set_host( 'perf280er.lab.netapp.com' ); # The name of the backed-up host
  $s->set_maxmpx( 130 );                      # Multiplexing factor
  $s->set_mfs( 2000 );                        # Max fragment size
  $s->set_dt( 2 );                            # 2 (Nearstore - p3), 3 (Snapvault - p2)
  $s->set_nh( 'browne' );                     # The fully qualified name of the NearStore
  
  # Create the Storage Unit
  $s->execute_create;
  print $s->get_results unless $s->passed;
  
  # Delete the Storage Unit
  $s->execute_delete;
  print $s->get_results unless $s->passed;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  NBU::Command.pm
  Data::Dumper.pm
  
=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut

__END__

Add a Storage Unit

/usr/openv/netbackup/bin/admincmd/bpstuadd \
	-M perf280er.lab.netapp.com \
	-label djohnSU \
	-path "/djohn" \
	-cj 1 \
	-odo 1 \
	-host perf280er.lab.netapp.com \
	-maxmpx 1 \
	-mfs 2000


-M			Master Server
-label 		Storage Unit Name
-path		Path to Directory
-cj			Maximum Jobs
-odo		On Demand Only; 1 = only when requested
-host		Media Server
-maxmpx		Multiplexing Factor
-mfs		Max Frament Size; Units of 1 MB; Disk is 20 to 2000

Delete a Storage Unit

/usr/openv/netbackup/bin/admincmd/bpstudel \
	-M perf280er.lab.netapp.com \
	-label djohnSU
