#  $Id: $

## @summary Utils - Catchall NBU Utility Functions
## @author todd@netapp.com
## @pod here

package NBU::Utils;

use NBU::Command;
use Data::Dumper;
use strict;

our $VERBOSE = 0;
my $STEM = "$NBU::Command::SUDO $NBU::Command::ROOT";
our $TESTBIN = "/usr/local/test/bin";

=head1 NAME

  Utils - Catchall NBU Utility Functions
  
=head1 DESCRIPTION

  Catchall NBU Utility Functions.
  
=head1 SYNOPSIS

  use NBU::Utils;

  # Is NBU installed?
  die "NBU is not installed" unless $NBU::Utils::nbu_installed();

  # Are required NBU daemons running?
  if ( NBU::Utils::daemons_okay() ) {
      ... Issue NBU commands
  } else {
  
      # Stop and restart daemons softly
      NBU::Utils::bounce_daemons();
	  
	  # That must not have worked.
	  unless ( NBU::Utils::daemons_okay() ) {
	      
		  # Now use 'kill'
		  NBU::Utils::bounce_daemons( 'hard' );
	  }
  }
  
  # Get the NBU version
  my $version = NBU::Utils::version();

  # Lookup up an NBU error code.
  print NBU::Utils::lookup_error( 4 ) . "\n";
  
  # Get job history
  my $jobs = NBU::Utils::get_jobs();
  # Get the last job details
  my ( $id, $details ) = NBU::Utils::last_job( $jobs );

=head1 METHODS

=head2 nbu_installed()

  Takes no arguments.  Does a cursory check to make sure NBU
  is installed.  Returns 1 if installed, 0 if not.

=cut

sub nbu_installed {

	if ( -e $NBU::Command::ROOT . 'bin/bpbackup' ) {
		return 1;
	}
	
	return 0;
}

=head2 version() 

  Takes no arguments.  Reads the NBU version file and returns a reference
  to a hash of version attributes.

=cut

sub version {

	my $version_file =  $NBU::Command::ROOT . 'version';

	die "version file doesn't exist." unless -e $version_file;
	
	my $params = {};
	
	open( FILE, "$version_file" ) || die "Couldn't open version file: $version_file";
	
	while ( <FILE> ) {
		chomp;
		/(.+?) (.+)/;
		$params->{$1} = $2;
	}
	
	close FILE;
	
	return $params;
}

=head2 daemons_okay()

  Takes no arguments.  Check whether the important NBU daemons are running.
  Return 1 if okay, 0 otherwise.

=cut

sub daemons_okay {

	my $command = $STEM . 'bin/bpps';

	my $services = `$command`;
	
	print "Current Services:\n$services" if $NBU::Utils::VERBOSE;

	my $nbpem = $NBU::Command::ROOT . 'bin/nbpem';
	my $nbrb = $NBU::Command::ROOT . 'bin/nbrb';
	my $nbjm = $NBU::Command::ROOT . 'bin/nbjm';
	
	if ( 
		$services =~ /$nbpem/s && 
		$services =~ /$nbrb/s && 
		$services =~ /$nbjm/s
	) {
		return 1;
	} else {
		return 0;
	}
}

=head2 kill_daemons_soft()

  Takes no arguments and returns nothing.  Gentle shutdown of all deamons.

=cut

sub kill_daemons_soft {

	my $command = $STEM . 'bin/goodies/netbackup stop';

	my $stop = `$command`;
	
	print "Stopping Services:\n$stop" if $NBU::Utils::VERBOSE;
}

=head2 kill_deamons_hard()

  Takes no arguments and returns nothing.  Hard shutdown of all daemons 
  using 'kill'.

=cut

sub kill_daemons_hard {

	my $command = $STEM . 'bin/bpps';
	
	my $bpps = `$command`;

	print "Killing Services:\n$bpps" if $NBU::Utils::VERBOSE;
	
	while ( $bpps =~ /\s*(.*?)\s*(\d+)(.*)\n/g ) {
		my $pid = $2;
		my $remainder = $3;
		next if $remainder =~ /java/i;
		my $kill = `/usr/local/bin/sudo kill -9 $pid`;
		print "Killing Service:\n$kill" if $NBU::Utils::VERBOSE;
	}
}

=head2 start_daemons()

  Takes no arguments and returns nothing.  Starts all daemons.
  
=cut

sub start_daemons {

	my $command = $STEM . 'bin/goodies/netbackup start';
	
	print "Starting Services:\n" if $NBU::Utils::VERBOSE;

	my $start = `$command`;
	
	print "$start\n" if $NBU::Utils::VERBOSE;

}

=head2 bounce_daemons()

  Takes a string.  
  
  If the string equals 'hard', this shuts down all daemons gently by 
  calling kill_daemons_soft() then shuts down remaining daemons
  mercilessly using kill_daemons_hard().  Otherwise, this shuts down 
  all daemons softly by calling kill_daemons_soft().
  
  Then all daemons are restarted.
  
  ( NOTE:  This does not kill processes matching the string 'java' in 
  order to leave the NBU GUI running. )

=cut

sub bounce_daemons {

	my ( $severe ) = @_;

	kill_daemons_soft();
	
	kill_daemons_hard() if $severe eq 'hard';
	
	start_daemons();
}

=head2 lookup_error()

  Takes an integer error code.  Returns a string for the error code.
  If you pass a second argument that evaluates true, the recommended
  fix is included in the return string.

=cut

sub lookup_error {

	my ( $error, $recommend ) = @_;

	my $command = $STEM . "bin/admincmd/bperror -S $error";
	$command .= " -r" if $recommend;

	my $results = `$command`;
	
	return $results;
}

=head2 get_jobs()

   Takes no arguments.  Returns a hash reference, keyed by JobID, containing
   the history of NBU jobs (the equivalent of Activity Monitor in the GUI).
   
   {
     '1433' => {
                 'Client' => 'perf280er.lab.netapp.com',
                 'Dest' => 'perf280er.lab.netapp.com',
                 'Status' => '0',
                 'Policy' => 'k-test',
                 'Schedule' => '-',
                 'PID' => 1181,
                 'Type' => 'Backup',
                 'State' => 'Done'
			   }
   };
   
=cut

sub get_jobs {

	my $command = $STEM . "bin/admincmd/bpdbjobs";
	
	my $results = `$command`;

	my $data = {};

	while ( $results =~ /(.*?)\n/g ) {
		next if $1 =~ /JobID/;
		my $line = $1;
		#print "*$1*\n";
		$line =~ /(.{5})(.{8})(.{6})(.{7})(.{11})(.{16})(.*?) (.*?) (.*?)$/;
		#print "$1|$2|$3|$4|$5|$6|$7|$8|$9\n";
		my $key = strip($1);
		$data->{$key}{Type} = strip($2);
		$data->{$key}{State} = strip($3);
		$data->{$key}{Status} = strip($4);
		$data->{$key}{Policy} = strip($5);
		$data->{$key}{Schedule} = strip($6);
		$data->{$key}{Client} = strip($7);
		$data->{$key}{Dest} = strip($8);
		$data->{$key}{PID} = strip($9);
	}
	
	return $data;
}

# Strip leading or trailing whitespace.

sub strip {

	my ( $string ) = @_;
	
	$string =~ s/(.*?)\s+/$1/;
	$string =~ s/\s+(.*?)/$1/;
	
	return $string;
}

=head2 last_job()

  Takes a reference to a hash as returned by get_jobs() and returns the JobID
  of the last job (highest JobID) and its parameters.
  
=cut

sub last_job {

	my ( $jobs ) = @_;
	
	for ( reverse sort { $a <=> $b } keys %$jobs ) {
		return $_, $jobs->{$_};
	}
}

=head2 all_done()

  Takes a reference to a hash as returned by get_jobs() and a reference to a list 
  of JobIDs.  Returns one if all jobs are 'Done', zero otherwise.
  
=cut

sub all_done {

	my ( $jobs, $jobids ) = @_;
	
	for ( @$jobids ) {
		return 0 unless $jobs->{$_}{State} eq 'Done';
	}
	
	return 1;
}	

=head2 check_nbu()

  Convenience setup function.

  Takes no arguments.  Checks whether NBU is installed, the required daemons
  are running, and checks and logs the NBU version.
  
=cut

sub check_nbu {

	Tharn::logcomment( "Check whether NBU is installed" );
	
	die "NBU isn't installed" unless NBU::Utils::nbu_installed();
	
	Tharn::logcomment( "Start NBU daemons" );
	
	NBU::Utils::start_daemons();
	die "Couldn't start required daemons" unless NBU::Utils::daemons_okay();
	
    Tharn::logcomment( "The NBU software version" );
	
	my $version = NBU::Utils::version();
	my $version_string;
	for ( keys %$version ) {
		$version_string .= "$_ $version->{$_}\n";
	}
	print $version_string, "\n";

}

=head2 p2_check_setup()

  Takes a hash reference to setup details for p2.
  
  - Sets up NDMPCredentials on the Master Server for both Filers.
  - Licenses snapvault on both Filers.
  - Sets options on both Filers.
  - Turns on the ndmp daemon on both Filers.
  
=cut

sub p2_check_setup {

	my ( $setup ) = @_;
	
	die "Both source and destination Filers are required." unless $setup->{filer1} && $setup->{filer2};
	
	# 1) This approach assumes the NDMP credentials and Filer's normal logon credentials
	# are the same.  
	# 2) I'm sucking the Filer rsh connection parameters out of the Connectrec objects.
	Tharn::logcomment( "Create NDMP Credentials" );
	
	my $ndmp1 = NDMPCredentials->new;
	$ndmp1->set_hostname( $setup->{filer1_name} );
	$ndmp1->set_username( 'root' );
	my $passwd1 = $setup->{filer1}{password} || $setup->{filer1}{hostp}{connection_fields}{console}{filer_conn}{password};
	$ndmp1->set_password( $passwd1 );		
	$ndmp1->execute_create;
	$setup->{filer1_ndmp} = $ndmp1;
	
	my $ndmp2 = NDMPCredentials->new;
	$ndmp2->set_hostname( $setup->{filer2_name} );
	$ndmp2->set_username( 'root' );
	my $passwd2 = $setup->{filer2}{password} || $setup->{filer2}{hostp}{connection_fields}{console}{filer_conn}{password};
	$ndmp2->set_password( $passwd2 );
	$ndmp2->execute_create;
	$setup->{filer2_ndmp} = $ndmp2;	Tharn::logcomment( "License SnapVault on source and destination filers" );
	
	Admin::License::license( $setup->{filer1}, 'sv_ontap_pri' );
	Admin::License::license( $setup->{filer1}, 'nfs' );
	Admin::License::license( $setup->{filer1}, 'cifs' );
	Admin::License::license( $setup->{filer2}, 'sv_ontap_sec' );
	Admin::License::license( $setup->{filer2}, 'nfs' );
	Admin::License::license( $setup->{filer2}, 'cifs' );
	
	Tharn::logcomment( "Set the NDMP options" );
	
	for ( $setup->{filer1}, $setup->{filer2} ) {
		SayWrap::sayEOF( $_, 'ndmpd on' );
		SayWrap::sayEOF( $_, 'options ndmpd.access                 all' );
		SayWrap::sayEOF( $_, 'options ndmpd.authtype               challenge,plaintext' );
		SayWrap::sayEOF( $_, 'options ndmpd.connectlog.enabled     off' );
		SayWrap::sayEOF( $_, 'options ndmpd.enable                 on' );
		SayWrap::sayEOF( $_, 'options ndmpd.ignore_ctime.enabled   off' );        
		SayWrap::sayEOF( $_, 'options ndmpd.offset_map.enable      on' );         
		SayWrap::sayEOF( $_, 'options ndmpd.password_length        16' );         
		SayWrap::sayEOF( $_, 'options ndmpd.preferred_interface    disable' );   
	}
	
	Tharn::logcomment( "Set the SnapVault options" );
	
	for ( $setup->{filer1}, $setup->{filer2} ) {
		SayWrap::sayEOF( $_, 'options snapvault.access             all' );
		SayWrap::sayEOF( $_, 'options snapvault.enable             on' );
	}
	
	Tharn::logcomment( "Start the NDMP daemons" );
	
	for ( $setup->{filer1}, $setup->{filer2} ) {
		SayWrap::sayEOF( $_, 'ndmpd on' );	
	}		

}

=head2 p2_setup_simple()

  Takes a reference to a setup hash and returns nothing.  The StorageUnit,
  Policy, Info, Include, Client, and Schedule objects are added to the
  setup hash.
  
  This sets up the basic storage unit and policy for a simple p2 test.
  
=cut

sub p2_setup_simple {

	my ( $setup ) = @_;

	$NBU::Command::VERBOSE = 1;
	#$NBU::Command::SLEEP = 5;

	## @step Create STU
	Tharn::logcomment( "Create a Storage Unit" );

	my $s = NBU::StorageUnit->new;
	$s->set_M( $setup->{client_name} );					# The Master Server
	$s->set_label( $setup->{prefix} );					# The storage unit label
	$s->set_path( "/vol/$setup->{dst_vol}" );			# The destination path on the Filer
	$s->set_odo( 1 );									# On demand only
	$s->set_host( $setup->{client_name} );				# The name of the backed-up host
	$s->set_dt( 3 );									# Disk Type 3 is a SnapVault
	$s->set_nh( $setup->{filer2_name} );				# The fully qualified name of the NearStore

	# These may cause an NBU bug.  The STU doesn't get created.	
	$s->set_cj( 130 );									# Maximum number of jobs
	$s->set_maxmpx( 32 );								# Multiplexing factor
	$s->set_mfs( 524288 );								# Max fragment size

	$s->execute_create;
	$setup->{storage_unit} = $s;

	## @step Create the Policy
	Tharn::logcomment( "Create the Policy" );

	my $p = NBU::Policy->new;
	$p->set_name( $setup->{prefix} );					# The policy name
	$p->set_storage_unit( $s );							# The storage unit to which to apply the policy
	$p->execute_create;
	$setup->{policy} = $p;
	
	my $i = NBU::Info->new;
	$i->set_policy( $p );								# The policy to which to add the client
	$i->set_collect_bmr_info( 0 );						# Whether to collect bare metal restore info
	$i->set_collect_tir_info( 0 );						#
	$i->set_offhost_backup( 1 );						#
	$i->set_use_data_mover( 1 );						#
	$i->set_data_mover_type( 3 );						# Same as 'dt'.  ?
	$i->set_snapshot_method( 'NAS_Snapshot' );			#
	$i->set_multiple_streams( 1 );						# Whether to allow multiple streams
	$i->set_fi( 1 );									# Perform snapshot backups
	$i->set_enable_pfi( 1 );							# Retain snapshots for instant recovery
	$i->set_ct( 'Standard' );							#
	$i->set_ef( '01/01/1970' ); 						#
	$i->execute_create;
	$setup->{info} = $i;

	my $c = NBU::Schedule->new;
	$c->set_name( $setup->{prefix} );					# The schedule name
	$c->set_type( 'FULL' );								# The type of backup
	$c->set_cal( 0 );									# The frequency, for frequency-based schedules
	$c->set_policy( $p );								# The policy to which to apply the schedule
	$c->set_rl( 3 );									#
	$c->set_pfi_fast_recovery( 1 );						#
	$c->set_synthetic( 0 );								#
	$c->set_st( 'FULL' );								#
	$c->execute_create;									#
	$setup->{schedule} = $c;							#

	my $d = NBU::Include->new;
	$d->set_path( $setup->{qtree1_mnt} );				# The path to backup
	$d->set_policy( $p );								# The policy to which to include the path
	$d->execute_create;
	$setup->{include} = $d;

	my $version = `uname -a`;
	$version =~ / \d+.(\d+) /;
	my $num = $1;

	my $t = NBU::Client->new;
	$t->set_policy( $p );								# The policy to which to add the client
	$t->set_name( $setup->{client_name} );				# The client host name
	$t->set_os( "Solaris$num" );						# The Solaris version
	$t->execute_create;
	$setup->{client} = $t;
}

=head2 discover_gig_ip()

  Takes a filer name.

  This discovers and returns the gigabit ethernet IP on a Filer, if one exists.
  
=cut

sub discover_gig_ip {

	my ( $filer ) = @_;
	
	my @ifconfig = ();
	for ( '', 'netapp' ) {
		@ifconfig = `rsh $filer -l root:$_ "ifconfig -a" 2>&1`;
		if ( @ifconfig && $ifconfig[0] !~ /Login incorrect/ ) {
			last;
		}
	}

	chomp for @ifconfig;

	#print Dumper \@ifconfig;

	my $ifs = {};
	my $key = '';
	for ( @ifconfig ) {
		if ( /^(.+?): (.*)$/ ) {
			$key = $1;
			$ifs->{$1} = [ $2 ];
		} else {
			s/\s+?(.*)/$1/;
			push( @{$ifs->{$key}}, $_ );
		}
	}

	#print Dumper $ifs;

	my $g;
	for my $if ( keys %$ifs ) {
		for ( @{$ifs->{$if}} ) {
			if ( /1000/ ) {
				$g = $if;
			}
		}
	}

	#print Dumper $g, $ifs->{$g};

	my $ip;
	for ( @{$ifs->{$g}} ) {
		if ( /inet (\d+?\.\d+?\.\d+?.\d+?) / ) {
			$ip = $1;
			last;
		}
	}

	#print Dumper $ip;
	
	return $ip, $g, $ifs;
}

=head2 compare_data()

  Takes tree1 and tree2 to compare.

  This compares the source tree and destination tree using "compare" unix utitlity
	based off of DataManip.pm routines.
  
=cut

sub compare_data {

  my ($tree1, $tree2) = @_;
  my $diff = 1; # no diffs at this time.

  if (-e "$TESTBIN/compare") {
    my $results = `$NBU::Command::SUDO $TESTBIN/compare $tree1 $tree2 2>&1`;

    if ($results =~ /Comparison of $tree1 to $tree2 failed/) {
      Tharn::logresult("FATAL", "Compare failed ($results)");
      return 0;
    }
    Tharn::logcomment( "Compare ($results)");
    my @lines = split(/\n/,$results);
    foreach my $text (@lines) {
      if ($text =~ /^(\d+).*/) {
        $diff = 0
          if $1;
        last
          unless $diff;
      }
    }
    return $diff ? 1 : 0;
  } else {
      Tharn::logresult("FATAL", "$TESTBIN/compare utility not found");
      return 0;
  }
}

1;

=head1 TODO

  Make a TODO list.
  
=head1 DEPENDENCIES

  Data::Dumper.pm
  NBU::Command.pm

=head1 AUTHOR

  Todd Shoenfelt (todd@netapp.com)
  
=cut

__END__

    root 14386     1  0   Oct 29 ?        0:01 /usr/openv/netbackup/bin/bpdbm
    root 14379     1  0   Oct 29 ?        0:00 /usr/openv/netbackup/bin/bprd
    root 17208 17204  0 17:46:31 ?        0:04 /usr/openv/netbackup/bin/bpjava-susvc root -1 -1 C /usr/openv/java/auth.conf 1 
    root 17204     1  0 17:46:28 ?        0:00 /usr/openv/netbackup/bin/bpjava-susvc root -1 -1 C /usr/openv/java/auth.conf 1 
    root 17206 17204  0 17:46:30 ?        0:00 /usr/openv/netbackup/bin/bpjava-susvc root -1 -1 C /usr/openv/java/auth.conf 1 
    root 17221 17204  0 17:46:41 ?        0:00 /usr/openv/netbackup/bin/bpjava-susvc root -1 -1 C /usr/openv/java/auth.conf 1 
    root 14388 14386  0   Oct 29 ?        0:02 /usr/openv/netbackup/bin/bpjobd
    root 14393 14392  0   Oct 29 ?        0:00 /usr/openv/netbackup/bin/nbproxy dblib
    root 14352     1  0   Oct 29 ?        0:00 /usr/openv/netbackup/bin/nbrb
    root 14397     1  0   Oct 29 ?        1:30 /usr/openv/netbackup/bin/nbpem
    root 14342     1  0   Oct 29 ?        0:09 /usr/openv/netbackup/bin/nbemm
    root 14338     1  0   Oct 29 ?        0:00 /usr/openv/netbackup/bin/nbnos
    root 14390     1  0   Oct 29 ?        0:02 /usr/openv/netbackup/bin/nbjm
    root 14402     1  0   Oct 29 ?        0:00 /usr/openv/netbackup/bin/nbsl
    root 14400 14399  0   Oct 29 ?        0:01 /usr/openv/netbackup/bin/nbproxy dblib
    root 17190 17161  0 17:46:19 pts/13   0:54 /usr/openv/java/jre/bin/java -Dvrts.NBJAVA_CONF=/usr/openv/java/nbj.conf -showv
    root 14334     1  0   Oct 29 ?        0:04 /opt/openv/db/bin/NB_dbsrv -n VERITAS_NB_fsr-c38 -x tcpip(ServerPort=13785) -gp

*JobID    Type State Status     Policy        Schedule                   Client           Dest Media Svr Active PID*
* 1367  Backup  Done     50 btu1-11781 btu1-11781sched perf280er.lab.netapp.com perf280er.lab.netapp.com      15619*
* 1411 Restore  Done      5                            perf280er.lab.netapp.com                                8361*
* 1414 Restore  Done      5                            perf280er.lab.netapp.com                                8884*
* 1419  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com      26566*
* 1420  Backup  Done      0     k-test            full perf280er.lab.netapp.com perf280er.lab.netapp.com      26601*
* 1421  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com      27925*
* 1422  Backup  Done      0     k-test            full perf280er.lab.netapp.com perf280er.lab.netapp.com      27950*
* 1423  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com      28138*
* 1424  Backup  Done      0     k-test            incr perf280er.lab.netapp.com perf280er.lab.netapp.com      28172*
* 1425  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com      28685*
* 1426  Backup  Done      0     k-test            incr perf280er.lab.netapp.com perf280er.lab.netapp.com      28704*
* 1427  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com      29094*
* 1428  Backup  Done      0     k-test            incr perf280er.lab.netapp.com perf280er.lab.netapp.com      29115*
* 1429  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com      29654*
* 1430  Backup  Done      0     k-test            incr perf280er.lab.netapp.com perf280er.lab.netapp.com      29672*
* 1431  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com        386*
* 1432  Backup  Done      0     k-test            incr perf280er.lab.netapp.com perf280er.lab.netapp.com        414*
* 1433  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com       1181*
* 1434  Backup  Done      0     k-test            full perf280er.lab.netapp.com perf280er.lab.netapp.com       1203*
* 1435  Backup  Done      0     k-test               - perf280er.lab.netapp.com perf280er.lab.netapp.com       1369*
* 1436  Backup  Done      0     k-test            incr perf280er.lab.netapp.com perf280er.lab.netapp.com       1384*
* 1437  Backup  Done    156  p2_policy               - perf280er.lab.netapp.com perf280er.lab.netapp.com       4257*
* 1438  Backup  Done    156  p2_policy               - perf280er.lab.netapp.com perf280er.lab.netapp.com       5689*
* 1439  Backup  Done      0  p2_policy               - perf280er.lab.netapp.com perf280er.lab.netapp.com       8252*
