# $Id: $

## @summary Base class for synthetic NFS clients
## @author snider@netapp.com
## @status Shared
## @pod here

package NFSClient;

use strict;
use nblade::nfsErrors;
use Params::NFS;
use Options;
use TracePkg;

my $Options;
my $Trace = new TracePkg;

$Options = new Options;
$Options->parameters(definitions=>
		     [
		      {
			  name => "VERBOSITY",
			  default => 2,
		      },
		      {
			  name => "DEVICE",
			  description => "The ethernet device to use",
		      },
		      {
			  name => "RETRIES",
			  description => "The number of retries before failing",
			  default => 6,
		      },
		      {
			  name => "TIMEO",
			  description => "The number of seconds between retries",
			  default => 3,
		      },
		      {
			  name => "TYPE",
			  default => "hard",
			  description => "a hard or soft mount type",
		      },
		      ]
		     );

=head1 NAME

new - this is the object instantiation routine for a synthetic client

=head1 SYNOPSIS

   use base "NFSClient";

=head1 DESCRIPTION

This package is the base class for the actual synthetic NFS client objects.
Test scripts should not use this packages. Only the synthetic NFS clients will
use this package.

=head1 ROUTINES

=head2 new

   my $linux_client = new NFSClient::Linux(srcip=>$client, uid=>0, gid=>0);

B<new> causes all parameters to be processed and will then create its own
internal nfsErrors object. This nfsErrors object is the actual interface
for creating NFS packets and sending them to the remote server. This package
should only be used by packages that are providing a synthetic client; not
actual test scripts.

The following named arguments are accepted:

=over

=item version

(Optional) : version of the NFS protocol (2 or 3)

=item retries 

(Optional) : the number of retries on failures. Defaults to the client default

=item timeo 

(Optional) : the number of seconds between retries. Defaults to the client default

=item hard_mount

(Optional) : act like a hard mount and continue retrying. Defaults to 1.

=back

=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = { version => 3 };
    $Trace->entry();
    bless($self, $class);

    $Options->process();
    $self->{nfs} = nfsErrors->new(version=>$self->{version}, mount=>undef, 
				  retries=>$Options->get(option=>"RETRIES"),
				  timeo=>$Options->get(option=>"TIMEO"),
				  hard_mount=>$Options->get(option=>"TYPE") eq "hard",
				  synthetic=>1,
				  chmod => undef,
				  @_,
				  );
    $self->{options} = $Options;
    return $self;
}    

=head2 nfs_code

   my $code = $client->nfs_code();

B<nfs_code> will return the last error code seen by the client.

=cut

sub nfs_code {
    my $self = shift;
    $self->{nfs}->printStatus();
}

=head2 attempts

   my $attempts = $client->attempts();

B<nfs_code> will return the number of attempts for the last call.

=cut

sub attempts {
    my $self = shift;
    $self->{nfs}->{attempts};
}

=head2 set_uid

  $client->set_uid(uid=>0);

B<set_uid> will set the default userid to be used in the authuid field
of each packet sent.

=over

=item uid

The uid to use as the new default authuid

=item back

=cut

sub set_uid {
    my $self = shift;
    my %params = (
		  uid => 0,
		  @_,
		  );
		  
    $self->{nfs}->{uid} = $params{uid};
}

=head2 get_attributes

  my %attr = $client->get_attributes();

B<get_attributes> will return the attributes for the last call.

=cut

sub get_attributes {
    my $self = shift;
    my %params = (
		  @_,
		  );
		  
    return $self->{nfs}->getAttributes();
}

1;

=head1 STATUS

   Shared

=head1 ATTRIBUTES

   ONTAP_COMMANDS = none

   ONTAP_ZAPI_APIS = none

=head1 AUTHOR

   snider@netapp.com
   $Id: $
   Last revised $Date: $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut
