# $Id: $

## @summary Synthetic FedoraCore3 Linux client
## @author snider@netapp.com
## @status Shared
## @pod here

## Fedora Core 3
package NFSClient::Linux;
use strict;
use base qw(NFSClient);
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use TracePkg;
use Options;
my $Trace = new TracePkg;
my $Options = new Options;

BEGIN {
    $Options = new Options;
    $Options->parameters(definitions=>
			 [
			  {
			      name => "RETRIES",
			      default => 6,
			  },
			  {
			      name => "TIMEO",
			      default => 3,
			  },
			  ]
			 );
}

my %common_params = (
		     src_ip => { type => SCALAR | UNDEF, optional => 1},
		     );
		     
sub mount {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT } ],
					 {
					     %common_params,
					     export => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    $nfs->get_port(port=>"nfs");
    $nfs->nfs_null();
    $nfs->get_port(port=>"mount");
    $nfs->mount_null();
    
    my $retries = $Options->{RETRIES};
    my $mountfh = $nfs->mount(share=>$params{export});
    while ($retries-- > 0) {
	if (!$mountfh) {
	    sleep $Options->{TIMEO};
	    $mountfh = $nfs->mount(share=>$params{export});
	} else {
	    last;
	}
    } 
    $Trace->fatal(msg=>"Server not responding\n") if (!$retries);
    $nfs->fsinfo(handle=>$mountfh) if ($self->{nfs}->{version} == 3);
    $Trace->exit(msg=>"return $mountfh");
    return $mountfh;
}

sub cd {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    $nfs->getattr(handle=>$params{handle});
    my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
    if ($access !~ /read/ && $access !~ /lookup/) {
	$Trace->fatal(msg=>"Access not allowed on the dir: $access\n");
    }
    $Trace->exit();
}


#ls
# NFS Getattr
# NFS Access 0x1f
# Readdirplus
sub ls {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					 }
					 );
    # cached ls will simply perfrom the getattrs rather than a readdirplus again
    $Trace->entry();
    my @entries;
    my $nfs = $self->{nfs};
    my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
    if ($access =~ /read/ && $access =~ /lookup/) {
	$nfs->getattr(handle=>$params{handle});
	@entries = $nfs->readdirplus(handle=>$params{handle});
	# should handle NULL attributes
    } else {
	$Trace->fatal(msg=>"Access not allowed on the dir: $access\n");
    }
    $Trace->exit();
    return @entries;
}

sub stat {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my %stat;
    my $nfs = $self->{nfs};
    if (0) { # newer linux
	my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
	if ($access !~ /read/) {
	    $Trace->fatal(msg=>"Access not allowed on the dir: $access\n");
	}
    }
    %stat = $nfs->getattr(handle=>$params{handle});
    
    $Trace->exit();
    return %stat;
}

sub create {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					     file => {type => SCALAR},
					     ok_to_exist => {type => SCALAR | UNDEF, optional => 1},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
    if ($access !~ /extend/) {
	$Trace->fatal(msg=>"'extend' access not allowed on the dir: $access\n");
    }
    
    my $fh;
    $nfs->getattr(handle => $params{handle}); 
    if ($fh = $nfs->lookup(handle => $params{handle},name => $params{file},
			   acceptable_codes=>[0,2])) {
	if (!$params{ok_to_exist}) {
	    $Trace->fatal(msg=>"'$params{file}' already exists\n");
	}
    } else {
	$fh = $nfs->create(handle => $params{handle}, file => $params{file});
    }
    
    $Trace->exit();
    return $fh;
}

sub remove {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					     file => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
    if ($access !~ /delet/) {
	$Trace->fatal(msg=>"delete access not allowed on the dir: $access\n");
    }
    
    my $fh;
    $nfs->getattr(handle => $params{handle}); 
    if (!($fh = $nfs->lookup(handle => $params{handle},name => $params{file}, fatal=>0))) {
	$Trace->fatal(msg=>"'$params{file}' did not exist\n");	
    }

    $nfs->getattr(handle => $fh); 
    $access = $nfs->access(handle=>$fh, access=>0x2d);
    if ($access !~ /read/) {
	$Trace->fatal(msg=>"read access not allowed on file '$params{file}': $access\n");
    }
    $nfs->remove(handle=>$params{handle}, file => $params{file});

    $Trace->exit();
}

sub write {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					     file_handle => {type => SCALAR | UNDEF, optional => 1 },
					     file => {type => SCALAR | UNDEF, optional => 1 },
					     data => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
    if ($access !~ /modify/) {
	$Trace->fatal(msg=>"Write access not allowed on the dir: $access\n");
    }
    
    $nfs->getattr(handle => $params{handle}); 
    my $fh = $params{file_handle};
    if (!$fh) {
	if (!$params{file}) {
	    $Trace->fatal(result=>"SCRIPT", msg=>"'file_handle' or 'file' must be passed in\n");
	}
	if (!($fh = $nfs->lookup(handle => $params{handle},name => $params{file}, fatal=>0))) {
	    $fh = $nfs->create(handle => $params{handle}, file => $params{file});
	}
    }
    $nfs->write(handle=>$fh, data=>$params{data});
    
    $Trace->exit();
}

sub read {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR},
					     file_handle => {type => SCALAR | UNDEF, optional => 1 },
					     file => {type => SCALAR | UNDEF, optional => 1 },
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    my %attr = $nfs->getattr(handle => $params{handle}); 
    my $access = $nfs->access(handle=>$params{handle}, access=>0x2d);
    if ($access !~ /read/) {
	$Trace->fatal(msg=>"Read access not allowed on the dir: $access\n");
    }
    
    my $fh = $params{file_handle};

    if (!$fh) {
	if (!$params{file}) {
	    $Trace->fatal(result=>"SCRIPT", msg=>"'file_handle' or 'file' must be passed in\n");
	}
	if (!($fh = $nfs->lookup(handle => $params{handle},name => $params{file}, fatal=>0))) {
	    $Trace->fatal(msg=>"file '$params{file}' does not exist: $access\n");
	}
    }
    my $data = $nfs->read(handle=>$fh, count=>$attr{size});
    $Trace->exit();
    return $data;
}

sub chmod {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     handle => {type => SCALAR | UNDEF, optional => 1 },
					     file_handle => {type => SCALAR | UNDEF, optional => 1 },
					     file => {type => SCALAR | UNDEF, optional => 1 },
					     mode => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};

    my $fh = $params{file_handle};

    if (! $fh) {
	if (!$params{file}) {
p	    $Trace->fatal(result=>"SCRIPT", msg=>"'file_handle' or 'file' must be passed in\n");
	}
	if (!$params{handle}) {
	    $Trace->fatal(result=>"SCRIPT", msg=>"'handle' must be passed in with 'file'\n");
	}
	my $access = $nfs->access(handle=>$params{handle}, access=>0x1f);
	if ($access !~ /read/ && $access !~ /lookup/) {
	    $Trace->fatal(msg=>"Read/lookup access not allowed on the dir: $access\n");
	}
	if (!($fh = $nfs->lookup(handle => $params{handle},name => $params{file}))) {
	    $Trace->fatal(msg=>"file '$params{file}' does not exist: $access\n");
	}
    }
    my %attr = $nfs->getattr(handle => $params{handle}); 
    $nfs->setattr(handle=>$fh, mode=>$params{mode});
    $Trace->exit();
}

#umount
# GetPort MOUNT
# MOUNT NULL
# MOUNT UMNT
sub umount {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT,  } ],
					 {
					     %common_params,
					     export => {type => SCALAR},
					 }
					 );
    $Trace->entry();
    my $nfs = $self->{nfs};
    $nfs->get_port(port=>"mount");
    $nfs->mount_null();
    $nfs->umount(share=>$params{export});

    $Trace->exit();
}

# TCP sends one MNT request

# GetPort MOUNT
# Mount EXPORT

# foreach export
#v2
# GetPort NFS
# NFS Null
# GetPort MOUNT
# Mount NULL
# Mount MNT
# NFS Getattr
# NFS Statfs

1;
