# -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE.pm#1 $
#
## @summary NOTE Netapp Objects for Test Engineering
##
## @author dl-dpgqa-automation@netapp.com
## @status Private (for now)
## @pod here

=head1 NAME

NOTE - A thin object wrapper around the test library

=head1 SYNOPSIS

If you find writing test scripts tedious, if you think they
require much too much effort to code and debug, if you don't
actually, really, like to code very much and prefer to
get the job done with as little effort as possible, then this
package is for you.

=head1 DESCRIPTION

This is the base class for the NOTE package, which is intended
to be a very thin object layer around the test library. Its purpose
is to make it possible to write test code using objects without
have a full-blown object model for the filer.

=head1 METHODS

=cut

# Base class for Netapp Objects for Test Engineering
package NOTE;

use Tharn;
BEGIN {
use GNOTE::Log;
    $NOTE::TRACELOG            = NOTE::Log->new(); # sets up trace log file
    $NOTE::TRACELOG->init_log();
}

use NOTE::Filer;
use NOTE::ImageBackup;
use NOTE::Client;
use NOTE::Aggregate;
use NOTE::Aggregate::Copy;
use NOTE::Volume;
use NOTE::Volume::Copy;
use NOTE::Qtree;
use NOTE::SnapMirror;
use NOTE::SnapVault;
use NOTE::LockVault;
use NOTE::FileSystem;
use NOTE::FileSystem::NFS;
use NOTE::Fiji;
use NOTE::FcSwitch;
use NOTE::Printf;
use NOTE::CFO;
use NOTE::FcMetrocluster;
use Utils::Units qw();
use Getopt::Std;
use IO::File;

use Data::Dumper qw(Dumper);

use filer;  # this is needed for filer_cli_say

our $_SPARE_ALL          = 0; # If set to 1, destructors do nothing to filer
our $_SPARE_ALL_ON_ERROR = 0; # If set to 1, spare all if error encountered
our $_CONSUMED_RESOURCES = 0;
our $_PANICKED = 0;
our $USAGE_ANALYSIS = defined $ENV{'NOTE_STATS'} ? $ENV{'NOTE_STATS'} : 0;
our $_MEM_STATS = defined $ENV{'NOTE_MEMSTAT'} ? $ENV{'NOTE_MEMSTAT'} : 0;
our $_SAVE_LOGS = defined $ENV{'NOTE_SAVE_LOGS'} ? $ENV{'NOTE_SAVE_LOGS'} : undef;
our $LOG = defined $ENV{'NOTE_LOG'} ? $ENV{'NOTE_LOG'} : undef;
our $LOGLEV  = undef;
our $LOGFILE = undef;
our $_POSTTEST = defined $ENV{'NOTE_POSTTEST'} ? $ENV{'NOTE_POSTTEST'} : undef;
our $Trace = TracePkg->new();

# This creates an object representing the tied handle which
# is now the default file handle for print, printf, and write.
# This the handlers for this tied hash redirect the output
# data to NOTE::logcomment. When the object is destroyed,
# prints are restored to STDOUT.

if ( defined $LOG ) {
    if ( $LOG =~ /^(\d+)$/ ) {
	$LOGLEV  = $1; 
	$LOGFILE = "/dev/null";
    } elsif ( $LOG =~ /^(\d+),(.+)$/ ) {
	$LOGLEV  = $1; 
	$LOGFILE = $2;
    } elsif ( $LOG =~ /^(.+)$/ ) {
	$LOGLEV  = '0';
	$LOGFILE = $1;
    } else {
	die "NOTE_LOG ($LOG) is incomprehensible\n";
    }
}
our $stdout = NOTE::Printf->new->start;
our $STDOUT = $stdout->raw;


my $init = 0;
my $teardown_called = 0;

# If you want to create objects within a subtest, call 
# NOTE::init as the first line of the subtest.
# You can also do this within the script if you want to call NOTE::teardown
# multiple times within the same process.
sub init {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $NOTE::init = $$;
    $teardown_called = 0;
}

# This BEGIN and END block make it possible to write
# a NOTE script without explicitly calling NOTE::teardown.
# Care is taken to make sure that teardown is called only
# once and that it is not executed in the context of a
# subtest (where $$ != $NOTE::init).
BEGIN {
    NOTE::init;
}

END {
    return unless ( $$ == $NOTE::init );
    eval { NOTE::teardown( exitcode => $? ) };
    NOTE::logcomment("Errors during teardown: $@") if ($@);
    $NOTE::TRACELOG->log_call_count();
    $NOTE::TRACELOG->close_log();
    $NOTE::TRACELOG->write_csv_log();
    $NOTE::TRACELOG->write_central_logs();
    posttest_cleanup();
}

use strict;

sub posttest_cleanup {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( defined $_POSTTEST ) {
	foreach my $script (split /;/, $_POSTTEST) {
		NOTE::logcomment("Running NOTE_POSTTEST script: $script");
		eval {
			subtest($script);
		};
	}
	NOTE::logcomment("Completed NOTE_POSTTEST script(s)");
    } else {
	NOTE::logcomment("No NOTE_POSTTEST scripts found");
    }
}

sub newName {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$descriptor) = @_;
    my $name;
    $name = sprintf( "%05d_%04x_%08x", $$, scalar( time ) & 0xFFFF, int(rand(0xffffffff)) );
    $name = $descriptor.$name if ( $descriptor );
    return $name;
}

sub datasetDecode {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name) = @_;

    if ( $name =~ /..(.....)_(....)_(........)_(.*)/ ) {
	return _datasetDecode( $4 );
    }
    return "";
}

sub datasetEncode {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name,$dataset) = @_;

    return $name."_"._datasetEncode( $dataset );
}

# Returns TRUE if an object *looks* anonymous according to
# its name. This is used by 'rename' methods to determine
# whether a renamed object should be spared or not.
sub _looksAnonymous {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $id;

    if ( $self->isa( "NOTE::Aggregate" ) ) {
	$id = "ag";
    } elsif ( $self->isa( "NOTE::Volume" ) ) {
	$id = "vl";
    } elsif ( $self->isa( "NOTE::Qtree" ) ) {
	$id = "qt";
    }
    return ( $self->name =~ /^(_|)$id(.....)_(....)_(........)/i );
}

# Returns TRUE if an object *looks* reusable according to
# its name.
sub looksReusable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    return $self->_looksAnonymous && (($self->name =~ /^_/) || ($self->name =~ /^AG/));
}

my @base64 = split //,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_+";

sub _datasetEncode {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $code = shift;

    # unpack into array of 8 bit bytes
    my @bytes = split //,$code;

    # Pad with whitespace to a multiple of 6 bytes
    push @bytes, ' ' for ( 1 .. (6-(((scalar @bytes) + 6) % 6)) );

    my $bitstr = "";
    for my $byte ( @bytes ) {
	my $bstr = unpack( "b7", $byte );
	$bitstr .= $bstr;
    }
    my $index = 0;
    my $nextChunk;
    my $bytestr = "";
    my $b6;
    while ( $index < length $bitstr ) {
	$nextChunk = substr( $bitstr, $index, 6 );
	$index += 6;
	$b6 = unpack( "C", pack( "b6", $nextChunk ) ) & 0x3f;
	my $c = $base64[$b6];
        if ( $c eq "_" ) {
	    $c = "__";
	} elsif ( $c eq "+" ) {
	    $c = "_P";
	}
	$bytestr .= $c;
    }
    return $bytestr;
}

sub _datasetDecode {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $bytestr = shift;
    my $bitstr = "";
    my $b7;
    $bytestr =~ s/__/=/g;
    $bytestr =~ s/_P/+/g;
    $bytestr =~ s/=/_/g;
    my @bytes = split //,$bytestr;
    for my $byte ( @bytes ) {
	my $pos;
	for ( $pos = 0; $pos < @base64; $pos++ ) {
	    last if ( $byte eq $base64[$pos] );
	}
	my $bstr = unpack( "b6", pack( "C", $pos & 0x3f ) );
	$bitstr .= $bstr;
    }
    my $index = 0;
    my $nextChunk;
    $bytestr = "";
    while ( $index < length $bitstr ) {
	$nextChunk = substr( $bitstr, $index, 7 );
	$index += 7;
	$b7 = unpack( "C", pack( "b7", $nextChunk ) ) & 0x7f;
	$bytestr .= chr($b7);
    }
    $bytestr =~ s/\s*$//; # Remove trailing whitespace
    return $bytestr;
}

sub _hashOverlay {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %args = @_;
    my $source = $args{source};
    my $target = $args{target} || {};
    my $overwrite = defined $args{overwrite} ? $args{overwrite} : 1;

    return unless $source;
    return unless (ref $target eq 'HASH');
    for my $key ( keys %{$source} ) {
	if ( ref $source->{$key} eq 'HASH' ) {
	    $target->{$key} = _hashOverlay(
		target    => $target->{$key},
		source    => $source->{$key},
		overwrite => $overwrite,
	    );
	} else {
	    if ( !defined($target->{$key}) || $overwrite ) {
		$target->{$key} = $source->{$key};
	    }
	}
    }
    return $target;
}

my @line2step = ();
my $_initted = 0;

sub _getSteps {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $file = shift;
    my $curstep;
    my $stepno = undef;
    my @steps;
    my $offset;

    return if $file =~ /^\(/; # Ignore invalid name; inline NOTE script invoked in subtest mode
    open( F, $file ) or die "Can't open $file";
    while ( <F> ) {
	if ( /^\s*##\s*\@step\s+(.*)/ ) {
	    $curstep = $1;
	} elsif ( $curstep && /^\s+##\s*(.*)/ ) {
	    $curstep .= " ".$1;
	} else {
	    if ( $curstep ) {
		push @steps, $curstep;
		$stepno = $#steps;
		$offset = 0;
	    }
	    $curstep = undef;
	}
	if ( defined $stepno ) {
	    my $n = $.;
	    $line2step[$n] = {
		step => "[".$steps[$stepno]."]",
		offset => $offset
	    };
	    $offset++;
	}
    }
}

# Given a string of the form
#    x,y=1,z=abcde
# will return a reference to a hash:
#    x => undef,  # i.e. exists, but undefined
#    y => 1,
#    z => abcde
sub opts {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $string = shift;
    my %opts = ();

    for my $p ( split( /,/, $string ) ) {
	if ( $p =~ /^([^=]+)=(.*)$/ ) {
	    $opts{$1} = $2;
	} else {
	    $opts{$p} = 1;
	}
    }
    return \%opts;
}

=head2 ::option

Uses L<C<Tharn::param>|engineering::nate::releases::latest::html::Tharn/param>
to process command line parameters.

If the argument is of the form
     $param1:$param2:...:$paramN[:default]
returns the first parameter that resolves as defined using Tharn::param
or the last element in the list (if it is not preceeded by a $).
Otherwise returns undef.

=over

=item Arguments

B<param1:...:paramN>

Returns the first parameter that resolves to a value.

=item Example:

    my $filer = NOTE::option('$FILER_DST:$FILER_SRC:$FILER');
    my $type = NOTE::option('$VOL_TYPE:flex');

=back

=cut

sub option {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $pstring = shift;
    return $pstring unless ( defined($pstring) && $pstring =~ /^\$/ );
    my @plist = split( /:/, $pstring );
    my ($pkey,$pval);
    while ( $pkey = shift @plist ) {
	return $pkey unless ( $pkey =~ /^\$/ );
	$pkey =~ s/^\$//;
	$pval = Tharn::param( $pkey, '-default', undef );
	return $pval if ( defined $pval );
    }
    return undef;
}

my $_lastHeader = undef;
my $_lastLine = undef;
my @_lastFrame = ();
my $_lastStep = "";
my $_logInitted = 0;
my $_trace_enabled  = 0;
my $_curphase = "";
my ($_odev,$_oino);
my ($_edev,$_eino);

our (
    $opt_b, # print backtrace
    $opt_s, # print steps
    $opt_p, # print test phase
    $opt_r, # trace resource usage
);

=head2 ::logcomment

Uses L<C<Tharn::logcomment>|engineering::nate::releases::latest::html::Tharn/logcomment>
with additional NOTE information.

=over

=item Example

    NOTE::logcomment("Create Volume");

=back

=cut

sub logcomment (@) {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @args = @_;
    if ( ! $_logInitted ) {
	($_odev,$_oino) = stat(STDOUT);
	($_edev,$_eino) = stat(STDERR);
	$_logInitted = 1;
	$_trace_enabled = defined $ENV{NOTE_TRACE};
	if ( $_trace_enabled ) {
	    local @ARGV = split / /, $ENV{NOTE_TRACE};
	    getopts( 'bspr' );
	}
    }
    my $s;
    if ( $args[0] =~ /%/ && (scalar @args > 1) ) {
	my $f = shift( @args );
	$s = sprintf( $f, @args );
    } else {
	$s = join '', @args;
    }
    my $newStep;
    if ( $_trace_enabled ) {
	my $prefix = "";
	my $level = 0;
	my $step;
	my @frames = ();
	my $lastLine = "";
	my $lastFile = "";
	my $nearestStep = undef;
	my $inMain = 0;
	my $inTeardown = 0;
	my ($package, $filename, $line, $subroutine,
	    $hasargs, $wantarray, $evaltext, $is_require);
	my $i = 0;
	while ( 1 ) {
	    ($package, $filename, $line, $subroutine,
	     $hasargs, $wantarray, $evaltext, $is_require) = caller($i);
	    $i++;
	    last unless $package;
	    if ( $inMain ) {
		if ( $package ne 'main' ) {
		    if ( $opt_p ) { # Display test phase, if new
			my $subname = $subroutine;
			$subname =~ s/^main:://; # Phase is top-level subroutine name in ::main
			$subname = "Unknown" if ( $subname eq '(eval)' );
			if ( $subname ne $_curphase ) {
			    # Entering new phase
			    my $out = sprintf( "\n\n%s %s %s\n", '%%%% Phase: ', $subname, '%%%%' );
			    if ( ($_odev != $_edev) || ($_oino != $_eino) ) {
				if ( $NOTE::LOGLEV eq '1' ) {
				    print $NOTE::STDOUT $out, "\n";
				    STDERR->print( $out, "\n" );
				}
			    }
			    Tharn::logcomment( $out );
			    $_curphase = $subname;
			}
		    }
		    last;
		}
	    }
	    next if ( $subroutine eq "(eval)" );
	    next if ( $is_require );
	    $subroutine =~ s/^NOTE::/::/;

	    if ( $subroutine ne "::logcomment" ) {
		unshift @frames, $subroutine." [".$lastLine."]";
		$level++;
	    }
	    if ( $package eq 'main' ) {
		$inMain++;
		$inTeardown++ if ( $subroutine eq '::teardown' );
		if ( $opt_s ) {
		    if ( ! $_initted ) {
			$_initted++;
			_getSteps( $filename );
		    }
		    if ( !defined $nearestStep ) {
			if ( $line2step[$line]->{offset} < 10 ) { # ADJUST this
			    $nearestStep = $line;
			}
		    }
		}
	    }
	    $lastLine = $line;
	    $lastFile = $filename;
	}
	my $ind = 0;
	while ( (defined $_lastFrame[$ind] && defined $frames[$ind]) &&
	        ( $_lastFrame[$ind] eq $frames[$ind] ) ) {
	    $ind++
	}

	my $_nextHeader = "";
	if ( $opt_b ) {
	    for my $tlev ( $ind .. ($level-1) ) {
		$_nextHeader .= "    " for ( 0 .. $tlev );
		$_nextHeader .= $frames[$tlev];
		$_nextHeader .= "\n";
	    }
	    chomp( $_nextHeader );
	}

	my $_nextLine;
	if ( $opt_s && defined ($step = $line2step[$nearestStep]->{step}) ) {
	    # Hack because teardown usually doesn't have @step:
	    $step = "[Teardown]" if ( $inTeardown );
	    $_nextLine  = $step;
	    my $offset = $line2step[$nearestStep]->{offset};
	    $_nextLine .= "+$offset" if ( $offset );
	    $_nextLine .= " [$nearestStep]";
	} else {
	    $_nextLine = "$lastFile [$lastLine]";
	}
	$newStep = 0;
	if ( $_lastLine ne $_nextLine ) {
	    $prefix .= "\n$_nextLine\n";
	    $newStep = 1;
	    $_lastLine = $_nextLine;
	}
	if ( $_lastHeader ne $_nextHeader ) {
	    $prefix .= "$_nextHeader\n" if ( $_nextHeader );
	    $_lastHeader = $_nextHeader;
	}
	if ( $opt_b ) {
	    my $pad = "    " x ($level+1);
	    $s =~ s/\n/\n$pad/g;
	    $s = $prefix.$pad.$s;
	} else {
	    my $pad = "    ";
	    $s =~ s/\n/\n$pad/g;
	    $s = $prefix.$pad.$s;
	}
	@_lastFrame = @frames;
    }
    if ( ($_odev != $_edev) || ($_oino != $_eino) ) {
	if ( $NOTE::LOGLEV eq '1' ) {
	    print $NOTE::STDOUT $s, "\n";
#	    STDERR->print( $s, "\n" );
	}
    }
    if ( $_trace_enabled ) {
        $s = "\n" . $s if ( $s !~ /^\n/ );
	if ( $newStep ) {
	    $s = "%%%" . $s;
	} else {
	    $s = "%%" . $s;
	}
    }
    Tharn::logcomment( $s );
}

=head2 ::units

If passed a value with a unit specifier, (i.e. t = terabytes, g = gigabytes, m = megabytes, etc.. ),
returns the value in kilobytes (without a unit specifier). If no unit specifier is present,
the value is assumed to already be in kilobytes.

=over

=item Argument

I<value>

=item Example

     my $kb = NOTE::units("200m");

=back

=cut

sub units {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $size = shift;
    return $size if ( $size =~ /^\d+$/ );
    my $bytes = Utils::Units::convert_to_bytes( $size );
    my $kb = Utils::Units::convert_units(
	$bytes,
	Utils::Units::UNIT_B,
	Utils::Units::UNIT_KB
    );
    return $kb;
}


my @_objList = ();

sub mark {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mark = shift;
    NOTE::reference( \$mark );
    return $mark;
}

sub reference {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $ref = shift;

    push @_objList, $ref if ( $ref );
    if ( wantarray ) {
	return @_objList;
    } else {
	return scalar @_objList;
    }
}

sub unreference {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    for my $i ( 0..$#_objList ) {
	if ( $_objList[$i] == $self ) {
	    delete $_objList[$i];
	}
    }
}


my $_resourceSerial = 0x0000;
sub resource {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$mode) = @_;

    return unless ( $opt_r );

    my $props = "";
    # Trace resource usage
    if ( $mode eq 'allocate' ) {
	$self->{_serialNumber} = $_resourceSerial++;
    }
    if ( ref $self eq 'NOTE::Aggregate' ) {
	$props .= "Aggregate $mode: ";
	$props .= sprintf( "filer=%s ", $self->Filer->host );
	$props .= sprintf( "serialNumber=%d ", $self->{_serialNumber} );
	$props .= sprintf( "disks=%d ", scalar $self->disks );
	$props .= sprintf( "name=%s ", $self->name );
	if ( $mode eq 'allocate' ) {
	    my %properties = $self->Properties;
	    $props .= sprintf( "share=%d ", $self->is_shareable );
	    $props .= sprintf( "recycle=%d ", $self->is_recyclable );
	    $props .= sprintf( "anonymous=%d ", $self->is_anonymous );
	    $props .= sprintf( "snaplock=%s ", $properties{snaplock} );
	    $props .= sprintf( "raidtype=%s ", $properties{raidtype} );
	    $props .= sprintf( "raidsize=%s ", $properties{rgsize} );
	}
    } elsif ( ref $self eq 'NOTE::Volume' && $self->type eq 'trad' ) {
	$props .= "Volume $mode: ";
	$props .= sprintf( "filer=%s ", $self->Filer->host );
	$props .= sprintf( "serialNumber=%d ", $self->{_serialNumber} );
	$props .= sprintf( "disks=%d ", scalar $self->disks );
	$props .= sprintf( "name=%s ", $self->name );
	if ( $mode eq 'allocate' ) {
	    my %properties = $self->Properties;
	    $props .= sprintf( "type=%s ", $self->type );
	    $props .= sprintf( "recycle=%d ", $self->is_recyclable );
	    $props .= sprintf( "anonymous=%d ", $self->is_anonymous );
	    $props .= sprintf( "snaplock=%s ", $properties{snaplock} );
	    $props .= sprintf( "raidtype=%s ", $properties{raidtype} );
	    $props .= sprintf( "raidsize=%s ", $properties{rgsize} );
	}
    } else {
	return;
    }
    NOTE::logcomment( "Resource: $props" );
}

sub RECYCLE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return defined $ENV{NOTE_RECYCLE} ? $ENV{NOTE_RECYCLE} : 1;
}

sub DATAMANIP {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return defined $ENV{NOTE_DATAMANIP} ? $ENV{NOTE_DATAMANIP} : 1;
}

sub ALL_SPARED {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $NOTE::_SPARE_ALL || $ENV{NOTE_SPARE};
}

sub SPARE_ALL_ON_ERROR {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $NOTE::_SPARE_ALL_ON_ERROR || $ENV{NOTE_SPARE_ALL_ON_ERROR};
}

sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "NOTE::SPARE: Disabling object destruction on teardown" );
    $NOTE::_SPARE_ALL = 1;
}

sub consume {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $NOTE::_CONSUMED_RESOURCES = 1;
}

sub panicked {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $NOTE::_PANICKED++;
}

sub save_logs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $filer ) = @_;

    my $root_vol  = $filer->vol_root;
    return unless( $root_vol );

    # wait for core files to be saved if in progress, so cores will be collected as well
    my $tried=0;
    while( $tried++ < 3 ) {
	# do not rely on NFS, NFS has cache and may have delays, we do not want to miss anythihng
	my $cores = filer_cli_say($filer->conn, "priv set -q advanced;ls /vol/$root_vol/etc/crash");
	if( $cores =~ /saving/ ) {
	    NOTE::logcomment( "NOTE: waiting for core to be saved:\n$cores");
	    $tried=0;
	    sleep 5;
	}
    }

    my ($presnap,$postsnap) = ( '_pretest_'.$filer->{_root_snapshots}, '_posttest_'.$filer->{_root_snapshots} );
    my $root = NOTE::Volume->attach( Filer => $filer, name => $root_vol );

    my $presnapflag = 0;    # we need to check if pretest snapshot exists
    my $snapretries = 5;    # and we need to retry, see BURT 231315
    do {
	sleep 10 if( $snapretries < 5 );
	eval{ $presnapflag = $root->snapshot_exists($presnap) };
    } while( $@ && $snapretries-- );

    return unless( $presnapflag );  # If we can not get hold of snapshots then there is nothing we can do

    my $LOGDIR = param('LOGDIR','-default',undef);
    my $saveto = $filer->host . "_filerlogs";			    # where WE actually want to save files
    $saveto = "$_SAVE_LOGS/$saveto" if( $_SAVE_LOGS =~ /[a-z]/i );  # use directory given by SAVE_LOGS
    $saveto = "$LOGDIR/$saveto" if( $LOGDIR =~ /\w/ );		    # but add/use LOGDIR also if exist

    # and finally, create posttest snapshot only after all cores are saved

    if( $root->snapshot_create($postsnap) ) {
	my $srcpath = "/vol/$root_vol/.snapshot/$presnap";
	my $dstpath = "/vol/$root_vol/.snapshot/$postsnap";

	my $out ='';
	eval { $out = filer_cli_say(    # list of changed files in '$out'
		$filer->conn,
		"priv set -q test; treecompare start -V -c -V $srcpath $dstpath",
		{ timeout => 7200 }
	    );
	};

	my %files; # use hash of names to prevent multiple entries due to multiple mismatch like mtime and data

	foreach( split('[\n\r]',$out )) {
	    s/(^\s+|\s+$)//g;
	    if( /(Found tree2 only directory entry|size mismatch|data mismatch)/ ) {	# pay attention to this miscompare only
		s/.*\s+//; s/^\.//; # clean up - leave filename only
		$files{$_} = '';    # and register name in our hash
	    }
	}

	if( keys %files ) {
	    my $mount = undef;
	    eval { $mount = $root->mount; };	# if mount will fail for some reason then we just exit
	    unless( $@ ) {
		`mkdir -p $saveto`;
		if( -d $saveto && $root->mounted ) {
		    NOTE::logcomment("Saving logs from filer " . $filer->host . " to $saveto directory");
		    foreach my $file (keys %files) {
			my $newname = $file;
			$newname =~ s/\//_/g;
			`/usr/local/bin/sudo cp $mount/$file $saveto/$newname` if( -f "$mount/$file" );
		    }
		    `/usr/local/bin/sudo chmod 666 $saveto/*`;
		} else {
		    logcomment("Unable to save log files for filer " . $filer->host . ". Unable to mount or $saveto is not created");
		}
	    }
	}

	unless( NOTE->ALL_SPARED ) {	# do not delete snapshots if we want to spare rest of stuff.
	    $root->snapshot_delete($presnap);
	    $root->snapshot_delete($postsnap);
	}
    } else {
	NOTE::logcomment("Unable to save filer logs for filer ".$filer->host.", unable to create post test snapshot $postsnap");
    }
}

=head2 ::teardown

Cleanup all NOTE objects by calling their DESTROY methods in reverse order of creation.
Attempts to restart any non-pingable filers before cleanup of their objects.

Cleanup will be skipped if CLI parameter NOTE_SPARE is set to 1, which can be overridden
within a script by calling NOTE::SPARE;

If NOTE_SPARE_ALL_ON_ERROR is set to 1, teardown will set NOTE_SPARE to 1 if
the exit code passed to it is non-zero (i.e. a failure code) or if any object
destructor croaks.

=over

=item Argument

B<exitcode> => I<value>

I<Optional> Optional value to be use as NOTE exit code

=item Example

      NOTE::teardown(exitcode=>255);

=back

=cut

my $tear_errors = "";

# Takes an optional argument, that if specified, will cause the stack
# to be torn down only until a matching reference is encountered.
# This can be used to teardown all the objects created by a
# particular code block:
#
#    NOTE::mark( "here" );
#
#    $sm = NOTE::SnapMirror->initialize;
#    ...
#
#    NOTE::down( "here" );
#
sub down {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $mark = shift;
    unless ($NOTE::_SPARE_ALL) {
    	#
    	#	This is required because telnet does not work with a taken over filer
    	#	If ME takes over PARTNER, you cannot telnet to PARTNER.
    	#
    	#	find all CFO objects on the list and call
    	#
    	#		->giveback(safe=>1, force => 1)
    	#
    	#	which puts the CFO pair back into UP & UP if it isn't already
    	#
    	for my $CFO_obj  (@_objList) {
    		if (ref($CFO_obj) =~ /^NOTE::CFO/ ) {
    			NOTE::logcomment ("CFO object forcing " . $CFO_obj->me->host . " and " .
    					$CFO_obj->partner->host . " into UP & UP state ");
    			$CFO_obj->giveback(safe => 1, force => 1);
    			$CFO_obj->monitor;
    		}
    	}
    }

    while ( @_objList ) {
	my $errmsg = "";
	my $o = pop @_objList;
	next unless defined $o;
	if ( ref $o eq 'CODE' ) {
	    eval { &$o() };
	    $errmsg .= "    - ".$@ if ( $@ );
	} elsif ( ref $o eq 'SCALAR' ) {
	    return if ( $$o eq $mark );
	} else {
	    if ( $o->{_teardown} ) {
		if ( ref $o->{_teardown} eq 'HASH' && $o->{_teardown}->{pre} ) {
		    eval{ &{$o->{_teardown}->{pre}}( $o ) };
		    $errmsg .= "    - ".$@ if ( $@ );
		}
	    }
	    eval { $o->DESTROY };
	    $errmsg .= $@;
	    if ( $o->{_teardown} ) {
		if ( ref $o->{_teardown} eq 'HASH' && $o->{_teardown}->{post} ) {
		    eval{ &{$o->{_teardown}->{post}}( $o, $errmsg ) };
		    $errmsg .= "    - ".$@ if ( $@ );
		} elsif ( ref $o->{_teardown} eq 'CODE' ) {
		    eval{ &{$o->{_teardown}}( $o, $errmsg ) };
		    $errmsg .= "    - ".$@ if ( $@ );
		}
	    }
	}
	if ( $errmsg ) {
	    $tear_errors .= $errmsg;
	    NOTE::logcomment( $errmsg );
	    if ( NOTE->SPARE_ALL_ON_ERROR ) {
		NOTE::logcomment( "NOTE::SPARE_ALL_ON_ERROR is set" );
		NOTE->SPARE;
	    }
	}
    }
}

sub teardown {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    local $?; # Don't allow exit code to be changed during teardown
    my %args = @_;

    return if ( $teardown_called );
    $teardown_called = 1;

    $stdout->flush if ( $stdout );
    
    #
    # 20090413		tyu@netapp.com
    #
    # The following block handle the situation where the user (test executor)
    # indicate he/she wants to preserve the configured entities (i.e. bypass
    # teardown) if there's any failure occurred during the test.
    #
    if (NOTE->SPARE_ALL_ON_ERROR()) {
	my @result_objects = own_results ();

	if (@result_objects) {
	    my $flag_need_spare = 0;
	    my $num_of_results = @result_objects;
	    my $idx = 0;

	    NOTE::logcomment( "(NOTE::teardown): num_of_results is ".$num_of_results );

	    #
	    # 20090413		tyu@netapp.com
	    #
	    # The following block handles the condition where croak, confess,
	    # or die was called during the course of the test. This is working
	    # in connection with TestWrapper.pm, which is responsible to detect
	    # the calling of croak, confess, and die, and posted a special INFO
	    # result object into the Test_DB. This enable the block here to
	    # determine whether the test is in the failure condition or not.
	    #

	    while (($flag_need_spare != 1) && ($idx < $num_of_results)) {
		my @the_messages = $result_objects[$idx]->message();
		if (@the_messages) {
		    if ($the_messages[0] eq "Error detected for NOTE_SPARE_ALL_ON_ERROR") {
			$Trace->debug (msg=>"(NOTE::teardown): Need spare, A");
			$flag_need_spare = 1;
		    }
		}
		$idx ++;
	    }

	    #
	    # 20090413		tyu@netapp.com
	    #
	    # The following block handles the condition where the test script
	    # specifically log a "FAIL" result into the TestDB.
	    #

	    if ($flag_need_spare != 1) {
		my $worst_result_obj = NATE::Result::worst (@result_objects);

		$Trace->debug (msg=>"(NOTE::teardown): worst severity is ".$worst_result_obj->severity ());
		if ($worst_result_obj->severity () > NATE::Result::Warn::severity ()) {
		    $Trace->debug (msg=>"(NOTE::teardown): Need spare, B");
		    $flag_need_spare = 1;
		}
	    }

	    if ($flag_need_spare == 1) {
		$Trace->debug (msg=>"(NOTE::teardown): Before NOTE::SPARE");
		NOTE->SPARE();
	    }
	}
    }

    if ( defined $args{exitcode} ) {
	# probably called from an END block
	if ( $args{exitcode} == 0 ) {
	    NOTE::logcomment( "NOTE: (Success) Teardown started" );
	} else {
	    NOTE::logcomment( "NOTE: (Failure[$args{exitcode}]) Teardown started" );
	    if ( NOTE->SPARE_ALL_ON_ERROR ) {
		NOTE::logcomment( "NOTE::SPARE_ALL_ON_ERROR is set" );
		NOTE->SPARE;
	    }
	}
    } else {
	NOTE::logcomment( "NOTE: Teardown started" );
    }

    if ( NOTE->ALL_SPARED ) {
        NOTE::logcomment( "NOTE::SPARE is set: no objects will be destroyed" );
    }

    # Check that all filers are up. If not, bring them up
    my @SaveLogsPid = ();
    for my $obj ( @_objList ) {
	next unless defined $obj;
	next if ( ref $obj eq 'CODE' );
	next if ( ref $obj eq 'SCALAR' );
	if ( $obj->isa( "NOTE::Filer" ) ) {
	    if( ! $obj->ping ) {
		if ( NOTE->ALL_SPARED || $obj->{_spare_me} ) {
		    NOTE::logcomment( "NOTE: %s seems to have panicked.", $obj->hostspec );
		    next;
		}
		NOTE::logcomment( "NOTE: Bringing up %s because it seems to have panicked.", $obj->hostspec );
		$obj->panicked;
		eval { $obj->transit_to( "UP" ) };
		if ( $@ ) {
		    NOTE::logcomment( "NOTE: Bringup of %s failed; $@", $obj->hostspec );
		} else {
		    NOTE::logcomment( "NOTE: Bringup of %s successful.", $obj->hostspec );
		}
	    }

	    if( $obj->ping && $_SAVE_LOGS ) {	# try to save logs...
		my $logsto = $obj->host . '_filerlogs';	# this will effect directory name
		push @SaveLogsPid, subtest( \&save_logs, '-runid', $logsto, '-bg', '-onexit', sub{}, '--', $obj );
	    }
	}
    }
    foreach( @SaveLogsPid ) { $_->wait_finish if( defined $_ ); }

    NOTE::down();

    if ( $NOTE::USAGE_ANALYSIS ) {
	NOTE::logcomment( "NOTE: %s resources were consumed.", $NOTE::_CONSUMED_RESOURCE ? "Some" : "No" );
	if ( $NOTE::_PANICKED ) {
	    NOTE::logcomment( "NOTE: %s filer%s may have panicked.",
		$NOTE::_PANICKED,
		$NOTE::_PANICKED == 1 ? "" : "s"
	    );
	}
    }
    if ( $tear_errors ) {
	if ( defined $args{exitcode} && $args{exitcode} == 0 ) {
	    die $tear_errors;
	} else {
	    NOTE::logcomment( "Errors during teardown:\n%s", $tear_errors );
	}
    }
}

=head1 STATUS

Public.

=head1 AUTHOR

dl-dpgqa-automation@netapp.com

=cut

1;
