#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Aggregate.pm#2 $
#
## @summary  NOTE Aggregate module
## @author   dl-dpgqa-automation@netapp.com
## @status   Public  
## @pod      here
=head1 NAME

NOTE::Aggregate

=head1 DESCRIPTION

This is the NOTE class for working with Aggregates.

=head1 METHODS

=cut

use NOTE;
package NOTE::Aggregate;
our @ISA = qw(NOTE);

use strict;
use Carp qw(croak confess);

# NOTE: Do *NOT* import any external functions into this module!
# This is to avoid namespace collisions namely prototype mismatches and
# all that. Call the external function directly by prepending the package 
# name.
use Aggregate qw();
use Aggregate::Copy qw();
use cluster::VolUtils qw( aggr_mirror );
use Wafliron qw();
use disk qw();
use Data::Dumper;

use constant FILERS_WITH_SINGLE_FAKE_RAID_GROUPS   => qw(FAS3040 FAS3140 FAS3050);

=head2 ::create

Returns an object representing an Aggregate. Uses 
L<C<aggr_create>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_create($conn,$aggr,$disklist[,%opts])>
function from lib/Aggregate.pm

=over

=item Syntax

C<NOTE::Aggregate-E<gt>create(%args)>

The B<%arg> hash contains NOTE's members as well as members for 
L<C<Aggregate::aggr_create>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_create($conn,$aggr,$disklist[,%opts])>
function. The NOTE::Aggregate::create will process and delete NOTE's members
from B<%arg> hash first, then B<%arg> will be passed to Aggregate::aggr_create
as B<%opts> hash as is.

=item Arguments

B<Filer> = I<filer object> | <filer spec>

I<Required> The filer on which to create the aggregate.

B<share> = 0 | 1

I<Optional> If 1, the aggregate will be sharable, which means that it
will be persistant (not destroyed at the end of the script) and that
it will be useable by other scripts that require an aggregate with
the same characteristics (raid size, raid type, number of disks).
Volumes created in shared aggregates should generally use
C<guarantee> => C<volume> (the default) so that an insufficient space
condition can be detected during volume create. If this condition
occurs when creating a volume in a shared aggregate, a disk will
be added to the aggregate and the create will be attempted again.
Note that this happens only once per volume create attempt.
C<share> => C<1> is the default setting, unless $WAFLIRON (the
flag indicating that wafliron is to run at the end of the script)
is set, since snapmirror activity in other volumes in a shared
aggregate may cause wafliron to fail.

B<name> = string

I<Optional> Optional aggregate name. If name is provided, then script will search
for aggregate with same name and will use it if found. If aggregate with given
name is not found then new aggregate will be created.

B<tag> = Comments

I<Optional> Comment tag. Will be printed in log, very useful when name is created automatically.

B<disks> = number

I<Optional> Number of disks to use, default is 3.

B<disklist> = list

I<Optional> List is string of space separated disk IDs. For example '2.10 2.11' etc.

B<recycle> = 0|1

I<Optional.>
A flag to indicate whether or not the created aggregate
should be recycled when its current life ends.
Requesting recycling also means that a previously
existing aggregate may be reused instead of actually
creating a new one.
The default behavior is to recycle anonymous objects
and not recycle named objects.

Aggregates with datasets may be recycled.
This is implemented by taking a snapshot
immediately after the dataset is created
and reverting to it when the aggregate is reused.
Dataset specifiers must be identical strings
to be considered exact matches and thus
equivalent for the purposes of recycling.

B<raidtype> = 'raid_dp'|'raid4'

I<Optional> Force specific RAID type. Default is 'raid_dp'

B<snaplock> = 'compliance'|'enterprise'

I<Optional> Create snaplock aggregate of given type. Licenses will be applied automatically

B<dataset> => I<dataset specifier>

I<Optional.>
A string containing perl code that will generate a dataset
for the aggregate. This string will be used to construct a
perl function which will be called with the aggregate object
reference as the first parameter just after the aggregate
has been created.

B<indirect_type> = 32 | 64

I<Optional> Specify either a 32-bit or 64-bit aggregate. 
This argument is valid only for ONTAP releases >= BR.0
Default is 64 if fake_large is set to 1, else it is 32.

B<fake_large> = 0|1

I<Optional> If 1, a Fake large aggregate will be created/used. Default is 0.

B<hole_size> = size spec

I<Optional> Size of hole for fake large aggregate. 
If undefined, a default size of 32TB will be used for aggregrate creation.

=item Arguments from Aggregate.pm

B<mirror>

I<Optional> Enable/disable aggregate mirroring

B<force>

I<Optional> Force aggregate creation

B<language>

I<Optional> Specify the language on the aggregate

B<rgsize>

I<Optional> Specify the raidgroup size.

B<randomize_args> = 0|1

I<Optional> Randomizes the command line arguments (except for disk-list)

=item Examples

Explicit declaration:

    $srcagr = NOTE::Aggregate->create(
	    Filer => { host => 'fsr-u60' },
	    tag => '8 disks large!',
	    disks => 8,
	    share => 1
	);
    $VolA = NOTE::Volume->create(
	    type => 'flex',
	    size => '10g',
	    Aggregate => $srcagr
	);
    $VolB = NOTE::Volume->create(
	    type => 'flex',
	    size => '100m',
	    Aggregate => $srcagr
	);

Implicit as part of NOTE::Volume->create

    $srcvol = NOTE::Volume->create(
	Filer => { host => $FILER_SRC },
	name => $FILER_SRC_VOL,
	type => 'flex',
	Aggregate => { recycle => 1, name => $FILER_SRC_AGGR },
    );
    logcomment("Aggregate name:" . $srcvol->Aggregate->name);

or

    $src = NOTE::Volume->create(
	Filer => { host => $FILER_SRC, stats => $SYSSTAT, },
	name => $FILER_SRC_VOL, recycle => 1, type => $VOL_TYPE,
	flex => {
	    Aggregate => { recycle => 1, name => $FILER_SRC_AGGR },
	    size => "10g",
	},
	trad => { disks => 3, raidtype => 'raid_dp' },
    );
    logcomment("Aggregate name:" . $src->Aggregate->name);

=back

=cut

# This constructor creates an object representing an aggregate.
# It can be passed either a filer object or a filer descriptor.
# NOTE::Aggregate( Filer => { host => $FILER } );
sub create {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($class,%args) = @_;
    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;
    $self->{_tag}   = delete $args{tag};

    my $share = NOTE::option(delete $args{share});

    # default to 1 for share flag, unless $WAFLIRON is true, since
    # snapmirror activity in a shared aggregate may cause wafliron to fail.
    $self->{_share} = ((!defined $share) &&
		        !NOTE::option(q/$WAFLIRON/)) || $share;


    exists $args{name}                  and $args{name}                 = NOTE::option( $args{name} );
    exists $args{snaplock}              and $args{snaplock}             = NOTE::option( $args{snaplock} );
    exists $args{raidtype}              and $args{raidtype}             = NOTE::option( $args{raidtype} );
    exists $args{rgsize}                and $args{rgsize}               = NOTE::option( $args{rgsize} );
    exists $args{disks}                 and $args{disks}                = NOTE::option( $args{disks} );
    exists $args{anonymous}             and $args{anonymous}            = NOTE::option( $args{anonymous} );
    exists $args{dataset}               and $args{dataset}              = NOTE::option( $args{dataset} );
    exists $args{recycle}               and $args{recycle}              = NOTE::option( $args{recycle} );
    exists $args{indirect_type}         and $args{indirect_type}        = NOTE::option( $args{indirect_type} );
    exists $args{fake_large}            and $args{fake_large}           = NOTE::option( $args{fake_large} );
    exists $args{hole_size}             and $args{hole_size}            = NOTE::option( $args{hole_size} );
    exists $args{set_default_reserve}   and $args{set_default_reserve}  = NOTE::option( $args{set_default_reserve} );

    my $default_snap_reserve = 1;
    $default_snap_reserve = 0 if (defined $args{set_default_reserve});
    delete $args{set_default_reserve};
     
    $self->{_teardown} = delete $args{teardown};

    # Default to 0 for fake large aggegate flag. 
    # This is required so that we recycle the right type of aggregate.
    $args{fake_large} = 0 if (! defined $args{fake_large});

    # hole size needs to be a multiple of 4K (WAFL block size).
    if (defined $args{hole_size}) {
        if ((NOTE::units($args{hole_size})) % 4) {
            confess "Specified hole size ($args{hole_size}) is not a multiple of 4K blocks.";
        }
    }

    # TODO:
    #  To support datasets strings that are longer than 200 or so bytes
    #  (the limit is related to the length limit of aggregate/volume names),
    #  the dataset string could be converted to a fingerprint:
    #    if ( $args{dataset} ) {
    #	     $args{dataset} = Digest::MD5->new->add( $args{dataset} )->hexdigest;
    #    }
    #  The disadvantage is that the conversion is only one-way.
    #  It wouldn't be as easy to determine the contents of an aggregate
    #  or a volume given its name only. Perhaps a best-of-both-worlds
    #  solution is to convert to a fingerprint only of the dataset string
    #  exceeds a certain length..
    if ( $args{dataset} ) {
	$args{dataset} =~ s/\s+/ /g; # Reduce multiple whitespace to one space
	$args{dataset} =~ s/^\s+//g; # Remove leading whitespace
	$args{dataset} =~ s/\s+$//g; # Remove trailing whitespace
    }


    if ( $args{name} ) {
	$self->{_name}  = $args{name};
	$self->{_share} = 0; # Share makes no sense with named aggregates
	# Allow named aggregates that behave like anonymous aggregates:
	$self->{_anonymous} = $args{anonymous};
	if ( $args{dataset} && !$self->{_anonymous} ) {
	    # Don't allow non-anonymous aggregates with datasets.
	    croak "Recycled datasets can only exist in anonymous aggegates";
	}
    } else { # Create a unique name
	if ( $args{dataset} ) {
	    # Don't allow shared aggregates with datasets.
	    $self->{_share} = 0;
	}
	$self->{_name} = NOTE::Aggregate->newName( share => $self->{_share} );
	$self->{_anonymous} = 1;
    }
    delete $args{name};
    delete $args{anonymous};


    # Default to recycle=>1 for aggregates
    $args{recycle} = 1 unless ( defined $args{recycle} );

    # Don't print name yet; it's not necessarily set in stone
    NOTE::logcomment( "Require aggregate ". ($self->{_tag} ? "(".$self->{_tag}.")" : "" ) );

    # if the $args{Filer} is a filer object, use it,
    # otherwise try to create one by calling the NOTE::Filer constructor.
    if ( !defined($args{Filer}) ) {
	# In this case, there are no arguments (i.e. no hostname) for the
        # filer. I call the filer method anyway and let it decide to
        # handle it or croak.
        $self->{_Filer} = NOTE::Filer->connect();
    } elsif ( ref $args{Filer} eq "HASH" ) {
	$self->{_Filer} = NOTE::Filer->connect( %{$args{Filer}} );
    } elsif ( ref $args{Filer} eq "NOTE::Filer" ) {
	$self->{_Filer} = $args{Filer}; # This is already an object
    } elsif ( ref \$args{Filer} eq "SCALAR" ) {
	# Is just a hostspec
	$self->{_Filer} = NOTE::Filer->connect( host => $args{Filer} );
    } else {
	croak "Invalid filer specification: ".ref \$args{Filer}."\n";
    }
    delete $args{Filer};
    $args{raidtype} ||= NOTE::option(q/$RAIDTYPE:raid_dp/); # Default

    if ($args{fake_large}) {

        # Currently we don't have the ability in NOTE::Aggregate.pm to set bootarg 'raidtop_fake_rg_max_count', 
        # reboot for it to take effect & repeat this during cleanup. 
        # So we will confess if this bootarg is not set correctly.
        my $filer_model = $self->{_Filer}->model();
        my $current_bootargs = $self->{_Filer}->bootargs();
        my $incorrect_fake_rg = 0;

        if (List::Util::first {/^$filer_model$/} FILERS_WITH_SINGLE_FAKE_RAID_GROUPS) {
            if (($current_bootargs->{'raidtop_fake_rg_max_count'} == 1) ||
                !defined ($current_bootargs->{'raidtop_fake_rg_max_count'}) ||
                ($current_bootargs->{'raidtop_fake_rg_max_count'} eq 'undefined')) {
                $incorrect_fake_rg = 1;
            }
        } else {
            if ($current_bootargs->{'raidtop_fake_rg_max_count'} == 1) {
                $incorrect_fake_rg = 1;
            }
        }
        if ($incorrect_fake_rg) {
            confess "bootarg raidtop_fake_rg_max_count = $current_bootargs->{'raidtop_fake_rg_max_count'}
                     In order to split up the vvols VVBN space so as to get allocations with large vvbns in a 
                     fake large aggregate this bootarg should be set to a value > 1.
                     To do this, either:
                     1. Use a platform that has this bootarg set to > 1 by default OR
                     2. Manually set this bootarg to the required value and reboot the filer for the bootarg to take effect. 
                     Remember to reset this bootarg to its default value and reboot the filer after the test completes.
                     (Note: This bootarg affects all vvols on the filer)\n";
        }

        if (!defined $args{indirect_type}) {
            NOTE::logcomment("Setting aggregate indirect_type to 64 to create fake large aggregate.");
            $args{indirect_type} = 64;
        } elsif ($args{indirect_type} != 64) {
            confess "Fake Large aggregate can only be created with aggregate indirect_type = 64-bit.\n";
        }
    } else {
        my $version = $self->{_Filer}->version();
        my $compare = Filer::Version::version_compare($version, Filer::Version::WV_BOILERMAKER_MIN);

        # Set indirect_type to default (32-bit) for ONTAP versions >= BR. 
        # This is required so that we recycle the right type of aggregate.
        if ($compare != -1 && !defined $args{indirect_type}) {
            $args{indirect_type} = 32;
        }
    }

    if ( !($args{disks}) && !($args{disklist}) ) {
	if ( $args{raidtype} eq "raid4" ) {
	    $args{disks} = "2";
	} elsif ( $args{raidtype} eq "raid_dp" ) {
	    $args{disks} = "3"; # Need at least 3 disks for raid_dp
	} else {
	    croak "Invalid raidtype: '".$args{raidtype}."'";
	}
    }

    my $numdisks;
    my @disk_list;
    if ( $args{disklist} ){
        @disk_list=split(/ /,$args{disklist});
        $numdisks = @disk_list;
    } 
    $numdisks ||= $args{disks};
    NOTE::logcomment( "Require a %s aggregate with %d disks",
	$args{raidtype},
	$numdisks,
    );

    my $diskspec = $args{disklist} ? \@disk_list : $args{disks};
    $args{disks} = @disk_list
	if exists $args{disklist};
    $args{timeout} = 3600 * $args{disks}; # 1 hour per disk

    delete $args{disks};
    delete $args{disklist};

    $self->{_recycle} = delete $args{recycle};
    $self->{_dataset} = delete $args{dataset};

    if ( $self->{_recycle} || $self->{_share} ) {
	if ( !$self->{_anonymous} ) {
	    if ( Aggregate::aggr_exists( $self->{_Filer}->conn, 
				         $self->{_name} ) ) {
		my $desc = $self->{_name};
		$desc .= " (".$self->{_tag}.")" if ( $self->{_tag} );
		NOTE::logcomment( "Re-using Aggregate $desc" );
		bless $self, $class;
		NOTE::reference( $self ) if ( $self->{_teardown} );
		$self->resource( 'allocate' );
		return $self;
	    }
	}
	if ( _recycle( $self, $diskspec, %args ) ) {
	    my $desc = $self->{_name};
	    $desc .= " (".$self->{_tag}.")" if ( $self->{_tag} );
	    NOTE::logcomment( "Using aggregate $desc" );
	    bless $self, $class;
	    NOTE::reference( $self );
	    $self->resource( 'allocate' );
	    return $self;
	}
	# This is a bit tricky: If I wasn't able to recycle
	# an aggregate that was populated with a dataset, try to
	# recycle an empty one and then populate it as if
	# it was just created.
	if ( my $dataset = $self->{_dataset} ) {
	    $self->{_dataset} = undef;
	    my $gotit = _recycle( $self, $diskspec, %args );
	    $self->{_dataset} = $dataset;
	    goto initaggr if ( $gotit );
	}
    }
    if ( $self->{_anonymous} && $self->{_share} && ! NOTE->RECYCLE ) {
	# If we're here, we are about to create a shared aggregate
	# when recycling has been disabled. In this case, we
	# should create a regular aggregate instead, so it will
	# be destroyed when the script exits.  $self->{_share} = 0;
	$self->{_name} = NOTE::Aggregate->newName( share => $self->{_share} );

    }
    my $desc = $self->{_name};
    $desc .= " (".$self->{_tag}.")" if ( $self->{_tag} );
    NOTE::logcomment( "Creating Aggregate $desc" );

    if ( my $snaptype = $args{snaplock} ) {
	# Make sure proper licences are installed.
	# Fix this up the way Aggregate::aggr_create wants them:
	delete $args{snaplock};
	if ( $snaptype eq 'compliance' ) {
	    $args{snaplock} = '1';
	    $self->{_Filer}->check_license( 'snaplock' );
	} elsif ( $snaptype eq 'enterprise' ) {
	    $self->{_Filer}->check_license( 'snaplock_enterprise' );
	    $args{snaplock_enterprise} = '1';
	}
    }

    # Creating Fake large aggregate.
    # To create a fake large aggregate, we will create an aggregate with 
    # the least number of disks required for the raid type specified. 
    # Then we will add a hole of specified size, followed by the remaining disks.
    my $remaining_disks;
    my $fake_large = delete $args{fake_large};
    my $hole_size = delete $args{hole_size};

    if ($fake_large) {

        # bootarg ntap.dblade.fake_large_aggr does not require a reboot.
        my $new_bootargs = {'ntap.dblade.fake_large_aggr' => 'true', 
                            'bootarg.dblade.fake_large_aggr' => 'true'};
        $self->{_Filer}->bootargs(args=>$new_bootargs);

        # If hole_size is not defined, default hole_size=32TB for new aggr creation
        $hole_size = '32t' unless ( defined $hole_size );
        my $total_disks = $diskspec;
        $diskspec = undef;

         # Create the aggr with the minimum number of disks required for the raid type.
	if ( $args{raidtype} eq "raid4" ) {
            if (ref($total_disks) eq "ARRAY") {
                @$diskspec = splice(@$total_disks, -2);
                $remaining_disks = $total_disks;
            } else {
                $diskspec = "2";
                $remaining_disks = $total_disks - $diskspec;
            }
	} elsif ( $args{raidtype} eq "raid_dp" ) {
            if (ref($total_disks) eq "ARRAY") {
                @$diskspec = splice(@$total_disks, -3);
                $remaining_disks = $total_disks;
            } else {
                $diskspec = "3";    # Need at least 3 disks for raid_dp
                $remaining_disks = $total_disks - $diskspec;
            }
	} else {
	    croak "Invalid raidtype: '".$args{raidtype}."'";
	}
    }

    my ($success, $err);
    do {
	$success = Aggregate::aggr_create( $self->{_Filer}->conn, $self->{_name}, $diskspec, %args );
        $err = $@;
    } while ( (!$success) && (_releaseSpareAggr( $self->{_Filer})) );

    croak "Create of aggregate failed ($err)" unless $success;
initaggr:
    bless $self, $class;
    NOTE::reference( $self );
    $self->resource( 'allocate' );

    # Add hole for fake LA
    if ($fake_large) {
        my $hole_size_in_blks = (NOTE::units($hole_size))/4;      #hole size in WAFL blocks

        $self->add(disks=>$remaining_disks, skip=>$hole_size_in_blks, keep_recycle=>1);
    }

    # Set snap reserve on aggregate to default to 5%
    if (($self->Filer->wafl_version >= Filer::Version::WV_ROLLINGROCK_MIN) && $default_snap_reserve){
	NOTE::logcomment("NOTE** Setting snap reserve on aggregate to 5% \n This is applicable to RR release\n");
	my $new_reserve = Aggregate::aggr_snap_reserve($self->Filer->conn,$self->name,reserve=>5);
	NOTE::logcomment("Default snap reservation on aggregate : $new_reserve %\n");
	}

    # Create the __empty__ snapshot
    if ( $self->is_anonymous && !$self->is_shareable ) {
	eval{ $self->snapshot_create( "__empty__" ) };
	if ( $@ ) {
	    croak $@ unless ( $@ =~ /Snapshot already exists/ );
	}
    }

    if ( $self->{_dataset} ) {
        NOTE::logcomment( "Initializing dataset of ".$self->description );
	my $func = eval 'sub {'.$self->{_dataset}.'}';
	croak $@ if $@;
	eval { $self->$func };
	croak $@ if $@;
	if ( $self->_looksAnonymous ) {
	    $self->_rename( NOTE::datasetEncode( $self->{_name}, $self->{_dataset} ) );
	}
	# We need to turn this off, otherwise the __clean__ snapshot
	# might be auto-deleted.
	$self->options( "snapshot_autodelete", "off" );
    }
    # Only save the state of private aggregates
    if ( $self->is_anonymous && !$self->is_shareable ) {
	eval{ $self->snapshot_delete( "__clean__" ) }; # Don't care if this fails
	$self->snapshot_create( "__clean__" );
    }
    return $self;
}

=head2 ::attach

Attach to existing aggregate instead of creating new one. Uses
L<C<aggr_exists>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_exists($conn,$aggr)>
function from lib/Aggregate.pm

=over

=item Arguments

B<Filer> = I<filer object> | <filer spec>

I<Required> The filer on which to attach to aggregate.

B<name> = string

I<Required> Name of the aggregate to attach to, aggregate must be already present on filer

B<tag> = string

I<Optional> Comments. Comments will be printed in log file along with aggregate name

=back

=cut

sub attach {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($class,%args) = @_;
    my $self = {};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;
    $self->{_tag}   = $args{tag};
    delete $args{tag};

    $args{name}     = NOTE::option( $args{name} ) if defined $args{name};

    # if the $args{Filer} is a filer object, use it,
    # otherwise try to create one by calling the NOTE::Filer constructor.
    if ( !defined($args{Filer}) ) {
	# In this case, there are no arguments (i.e. no hostname) for the
        # filer. I call the filer method anyway and let it decide to
        # handle it or croak.
        $self->{_Filer} = NOTE::Filer->connect();
    } elsif ( ref $args{Filer} eq "HASH" ) {
	$self->{_Filer} = NOTE::Filer->connect( %{$args{Filer}} );
    } elsif ( ref $args{Filer} eq "NOTE::Filer" ) {
	$self->{_Filer} = $args{Filer}; # This is already an object
    } elsif ( ref \$args{Filer} eq "SCALAR" ) {
	# Is just a hostspec
	$self->{_Filer} = NOTE::Filer->connect( host => $args{Filer} );
    } else {
	croak "Invalid filer specification: ".ref \$args{Filer}."\n";
    }
    delete $args{Filer};

    croak "Aggregate attach requires a name" unless ( $args{name} );
    $self->{_name} = $args{name};
    $self->{_dataset} = NOTE::datasetDecode( $self->{_name} );

    unless ( Aggregate::aggr_exists( $self->{_Filer}->conn, 
				     $self->{_name} ) ) {
	croak "Aggregate $self->{_name} does not exist.";
    };
    bless $self, $class;
    return $self;
}

=head2 ::dup

Create (or reuse) an aggregate with attributes similar to an existing aggregate.

=over

=item Arguments

Takes the same arguments as NOTE::Aggregate::create.

=item Examples

  $aggr_dst = $aggr_src->dup;  # Make a destination aggregate similar to the source

  $aggr_dst = $aggr_src->dup( Filer => q/$FILER_DST/ );  # The same, but different filer

=back

=cut

sub dup {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($origin,%args) = @_;
    my %properties = $origin->Properties;

    %properties = %{NOTE::_hashOverlay( target => \%properties, source => \%args, overwrite => 1 )};
    $properties{Filer} ||= $origin->Filer;
    return NOTE::Aggregate->create( %properties );
}

sub Properties {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $opts = $self->options;
    my %properties = (
	recycle   => $self->{_recycle},
	anonymous => $self->{_anonymous},
	snaplock  => $self->{_snaplock},
	disks     => scalar $self->disks,
	raidtype  => $opts->{raidtype},
	rgsize    => $opts->{raidsize},
    );
    return %properties;
}

=head2 ::volumes

Returns array of volume names found in aggregate. Array is simply array of strings.

=cut

sub volumes {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return @{shift->status->volumes} }

=head2 ::Volumes

Returns array of
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm>
objects found in aggrgate.

=over

=item Arguments

B<spare_only> = 0|1

I<Optional> If set to 1, then it will return list of
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm>
objects of recycled type which have not yet been reused. By default all volumes are retured.

=back

=cut

sub Volumes {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    my @objList = ();
    for my $vol ( $self->volumes ) {
	next if ( $args{'spare_only'} && $vol !~ /^_/ );
	push @objList, NOTE::Volume->attach(
	    Filer => $self->Filer,
	    name => $vol,
	);
    }
    return @objList;
}

=head2 ::add

Add disks to aggregate. Adds one disk only if called without arguments.
This method is a wrapper around
L<C<Aggregate::aggr_add>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_add($conn,$aggr,$disklist[,%opts])>
. All arguments are
passed as hash to it.

=over

=item Arguments (NOTE::Aggregate.pm)

B<disks> = number

I<Optional> Number of disks to add. This argument has no effect if
B<disklist> is also present.

B<disklist> = array

I<Optional> Array of disk ids. Uses the '-d' option to add each disk via its disk
name taken from each element of the array.

B<keep_recycle> = 0 | 1

I<Optional> Do not turn off the recycle flag when adding disk(s) to anonymous,
non-shareable, recyclable aggregate. Default is 0, i.e. by default, when 
disk(s) are added to anonymous, non-shareable, recyclable aggregate, the 
recycle flag is turned off, so aggregate will be destroyed at teardown.

=item Arguments inherited from
L<C<Aggregate::aggr_add>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_add($conn,$aggr,$disklist[,%opts])>

B<force> = 1|0

I<Optional> Force operation

B<timeout> = seconds

I<Optional> Wait for validation timeout. By default it waits for all disks to be
zeroed if needed.

=back

=cut

sub add {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    my $ret;
    my $keep_recycle = delete $args{keep_recycle};

    my $diskspec = $args{disks} || $args{disklist};
    delete $args{disks};
    delete $args{disklist};
    $diskspec = "1" unless ( $diskspec );

    if ( $diskspec =~ /^\d+$/ ) {
	my $what = $diskspec == 1 ? "disk" : "disks";
        NOTE::logcomment( "Adding $diskspec $what to ".$self->description );
    } else {
	my $what = $diskspec == 1 ? "a disk" : "disks";
        NOTE::logcomment( "Adding $what to ".$self->description );
    }
    $ret = Aggregate::aggr_add(
	$self->{_Filer}->conn,
	$self->{_name},
	$diskspec,
	%args
    );
    croak $@ if $@;
    $self->{_disks} = scalar $self->disks;

    if ( !$self->is_shareable && ($self->{_anonymous} && $self->{_recycle})) {
        if ($keep_recycle) {
            NOTE::logcomment(
                "Added disk(s) to anonymous, non-shareable, recyclable aggregate.\n".
                "Keeping recycle flag on so aggregate will not be destroyed at teardown."
                );
        } else {
            NOTE::logcomment(
                "Added disk(s) to anonymous, non-shareable, recyclable aggregate.\n".
                "Turning recycle flag off so aggregate will be destroyed at teardown.\n".
                "Name the aggregate or set the 'keep_recycle' flag if you want to get around this."
            );
            $self->{_recycle} = 0;
        }
    }
    return $ret;
}

=head2 ::online

Change the state of the aggregate to "online".
Returns success (if the aggregate was changed or is already in the desired
state).
or failure otherwise (if the aggregate couldn't change or didn't exist).

Wrapper around
L<C<Aggregate::aggr_online>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_online($conn,$aggr[,%opts])>

=over

=item Examples

     $aggr=>online;

     $aggr->online(force=>1);

=back

=cut

sub online {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    Aggregate::aggr_online( $self->{_Filer}->conn, $self->{_name}, @_ );
    croak $@ if $@;
    return $self;
}

=head2 ::offline

Change the state of the aggregate to "offline".
Returns success (if the aggregate was changed or is already in the desired
state).
or failure otherwise (if the aggregate couldn't change or didn't exist).
Wrapper around
L<C<Aggregate::aggr_offline>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_offline($conn,$aggr[,%opts])>

=over

=item Examples

     $aggr->offline;

     $aggr->offline(force=>1);

=back

=cut

sub offline {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    croak "Cannot offline a sharable aggregate" if $self->is_shareable;
    Aggregate::aggr_offline( $self->{_Filer}->conn, $self->{_name}, @_ );
    croak $@ if $@;
    return $self;
}

=head2 ::restrict

Change the state of the aggregate to "restricted".
Returns success (if the aggregate was changed or is already in the desired
state).
or failure otherwise (if the aggregate couldn't change or didn't exist).

Wrapper around
L<C<Aggregate::aggr_restrict>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_restrict($conn,$aggr[,%opts])>

=over

=item Example

     $aggr->restrict;

=back

=cut

sub restrict {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    croak "Cannot restrict a sharable aggregate" if $self->is_shareable;
    Aggregate::aggr_restrict( $self->{_Filer}->conn, $self->{_name}, @_ );
    croak $@ if $@;
    return $self;
}

=head2 ::snaplock

Returns snaplock type if snaplock is licensed and enabled for given aggregate. Snaplock
returned types are compliance or enterprise. Returns zero if no snaplock present.

=over

=item Example

      my $snaplock_type = $aggr->snaplock;

=back

=cut

sub snaplock {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $opts = $self->options;
    return 'compliance' if ( $opts->{'snaplock_compliance'} );
    return 'enterprise' if ( $opts->{'snaplock_enterprise'} );
    return 0;
}

=head2 ::options

Sets or gets the options of an aggregate. Wrapper around
L<C<Aggregate::aggr_options>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_options($conn,$aggr[, %opts])>.

=over

=item Examples

Reading example:

    my $time = $aggr->options('resyncsnaptime');

Setting example:

    $aggr->options(resyncsnaptime => 100);

The options "root, ignore_inconsistent" can be changed in 
maintainance mode only.

=back

=cut

sub options {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$which_one,$value) = @_;
    my ($options,$ret);
    if ( @_ == 3 ) {
	$ret = Aggregate::aggr_options (
	    $self->Filer->conn,
	    $self->{_name},
	    $which_one => $value,
	);
	croak $@ if $@;
	return $ret;
    } else {
	$options = Aggregate::aggr_options (
	    $self->Filer->conn,
	    $self->{_name},
	);
	croak $@ if $@;
	if ( $which_one ) {
	    return $options->{ $which_one };
	} else {
	    return $options;
	}
    }
}

=head2 ::status

Returns the state, status, host aggregate name, and a hash of aggregate options
of the given aggregate. Wrapper around 
L<C<Aggregate::aggr_status>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_status($conn,$aggr[,$timeout])>.

=over

=item Examples

  my $status_obj = $aggr->status;

  my $type      = $status_obj->type();
  my $state     = $status_obj->state();
  my $status    = $status_obj->status();
  my %options   = $status_obj->options();
  my $vols_list = $status_obj->volumes(); # returns reference to an array

=back

=cut

sub status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    
    if ( wantarray ) {
        my ($state, $status, $aggr, $options) = Aggregate::aggr_status(
	    $self->{_Filer}->conn, $self->{_name},
        );
	return ( $state, $status, $aggr, $options );
    } else {
        return Aggregate::aggr_status(
	    $self->{_Filer}->conn, $self->{_name},
        );
    }
}

=head2 ::hostpath

Returns the name of the aggregate
prepended by its filer's name.

=over

=item Arguments

None.

=item Example

=back

     $hp = $aggr->hostpath;

=cut

sub hostpath {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return $self->{_Filer}->host.":".$self->{_name};
}

#
# $fsid = $aggr->fsid;       # Read fsid
#
# or
#
# $aggr->fsid( $newfsid );   # Write fsid
#

=head2 ::fsid

Returns the FSID (in hexadecimal) of an aggregate or rewrites the FSID of the desired aggregate.
When rewrites FSID it returns the *original* FSID of the aggregate prior to the FSID rewrite.

Do not force a rewrite of FSID of an aggregate if you have no idea what
you're doing. Note: You need to restrict the volume(s) first prior to calling this API.

When setting new fsid, additional parameters are passed to
L<C<Aggregate::aggr_rewrite_fsid>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_rewrite_fsid($conn,$aggr,[%opts])>.

=over

=item Arguments

B<fsid> = id

I<Otional.> Sets FSID to one provided

B<force> = 1|0

I<Otional.> Force rewriting of the FSID even though an existing of the
same FSID exists on the system.

=item Examples

B<Read FSID>

    $fsid = $aggr->fsid;

B<Set FSID>

    $aggr->fsid( $newfsid );

    $aggr->fsid( $newfsid, force => 1 );

=back

=cut

sub fsid {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$newfsid,%args) = @_;
    my $oldfsid;
    $oldfsid = Aggregate::aggr_read_fsid(
        $self->{_Filer}->conn,
	$self->{_name}
    );
    croak $@ if $@;
    if ( $newfsid ) {
	Aggregate::aggr_rewrite_fsid(
	    $self->{_Filer}->conn,
	    $self->{_name},
	    fsid => $newfsid,
	    %args
	);
    }
    croak $@ if $@;
    return $oldfsid;
}

=head2 ::split

Split a aggregate. After the split of the syncmirrored aggregate, a
NOTE::Aggregate object would be returned to caller for the split-off
aggregate.

=over

=item Syntax

$aggr->split( $plex, $newaggr, %args );

=item Argments

B<plex> = name 

I<Optional> Plex to use. If no plex given then automatically grab the last plex as the one to be split off

B<newaggr> = name

I<Optional> New name for splitted half

B<debug> = 1|0

I<Optional> Print extra debugging info

B<recycle> = 1|0

I<Optional> Set or clear recycle flag

B<anonymous> = 1|0

I<Optional> Generate anonymous name for new aggregate. This is default behavior

=item Example

my $aggr_b = $aggr_a->split();

=back

=cut

sub split {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($parent,$plex,$newaggr,%args) = @_;
    my $self = {};

    $parent->Filer->check_license( "syncmirror_local" );
    $self->{_debug}    = $args{debug};
    $self->{_owner}    = $$;
    $self->{_dataset}  = $parent->{_dataset};
    $self->{_tag}      = "SyncMirrored of ".$parent->description;
    $self->{_recycle}  = $args{recycle};
    delete $args{recycle};

    NOTE::logcomment( "Splitting ".$parent->description );
    if ( $newaggr ) {
	$self->{_name} = $newaggr;
	$self->{_anonymous} = $args{anonymous};
	delete $args{newaggr};
	delete $args{anonymous};
    } else {
	# Create a unique name
	$self->{_name} = NOTE::datasetEncode(
	    NOTE::Aggregate->newName,
	    $self->{_dataset}
	);
	$self->{_anonymous} = 1;
    }

    $self->{_Filer} = $parent->Filer;

    # automatically grab the last plex as the one to be split off
    if ( !$plex ) {
	my $status = Aggregate::aggr_raidgroup_status( 
            $parent->Filer->conn,
	    $parent->name 
        ) or croak $@;

	foreach my $plex_info ( @{$status->{plex}} ) {
	    $plex = $plex_info->{name};
	}
    }

    Aggregate::aggr_split( $parent->Filer->conn, 
			   $plex,
			   $self->{_name} );
    croak $@ if $@;

    bless $self, ref($parent);
    NOTE::reference( $self );
    $self->resource( 'allocate' );

    return $self;
}

sub _rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$name) = @_;
    Aggregate::aggr_rename( $self->{_Filer}->conn, $self->{_name}, $name );
    croak $@ if $@;
    $self->{_name} = $name;
}

=head2 ::rename

Renames the aggregate to one given by argument or to another
anonymous name if no new name is given.

Anonymous aggregates retain the anonymous flag after being renamed
to a specific name if the new name looks like that of an anonymous
aggregate. This is needed for tests that 'swap' names between two
anonymous aggregates.

=over

=item Argument

B<newname> = string

I<Optional> Renames aggregate to 'string'

=item Examples

    $aggr->rename( 'Some_New_Name' );

    $aggr->rename;

=back

=cut

sub rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$newname) = @_;
    if ( defined $newname ) {
	$self->_rename( $newname );
	# Allow renaming an aggregate to a name that appears
	# to be anonymous without making it named. This is
	# needed for tests that 'swap' names between two anonymous
	# aggregates.
	if ( $self->{_anonymous} && !$self->_looksAnonymous ) {
	    $self->{_anonymous} = 0;
	}
    } else {
	$newname = $self->newName;
	$self->{_anonymous} = 1;
	$self->_rename( $newname );
    }
}

=head2 snapshot_create

Create a snapshot on the aggregate

=over

=item Argument

B<snapname> - I<name>

Optional

If not set then a anonymous name is used to create the snapshot with

=item Examples

=back

     $aggr->snapshot_create;

     $aggr->snapshot("special");

=cut

sub snapshot_create {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$snapname) = @_;

    $snapname = $self->SUPER::newName( "sn" ) unless( $snapname );

    NOTE::logcomment( "Creating snapshot ".$snapname." of ".$self->description );
    SnapCmds::snapshot_create(
        $self->Filer->conn,
        "-A ".$self->name,
	$snapname,
    );
    croak $@ if ($@);
    return $snapname;
}

=head2 ::snapshot_restore

Restored a named snapshot

=over

=item Arguments

B<snapname> - I<name>

Name of the snapshot to restore

=item Example

=back

     $aggr->snapshot_restore("special");

=cut

sub snapshot_restore {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$snapname,%opts) = @_;

    croak "Need snapshot name" unless ( $snapname );
    NOTE::logcomment( "Restoring ".$self->description." using snapshot ".$snapname );

    $self->Filer->check_license( "snaprestore" );
    SnapCmds::snaprestore_volume(
        $self->Filer->conn,
        "-A ".$self->name,
	$snapname,
	%opts,
    );
    croak $@ if ($@);
    return $self;
}

=head2 ::snapshot_list

Returns an array of the snapshots that exists on this aggregate

=over

=item Arguments

None.

=item Example

=back

     $aggr->snapshot_list;

=cut

sub snapshot_list {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my @list = SnapCmds::snapshot_list(
        $self->Filer->conn,
        "-A ".$self->name
    );
    croak $@ if ($@);
    return @list;
}

=head2 ::snapshot_status

Returns a reference to a hash of the status with the
snapshot names as keys.

=over

=item Arguments

None

=item Example

=back

     $aggr->snapshot_status;

=cut

sub snapshot_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %status = ();
    my @status;

    @status = SnapCmds::snapshot_status(
	$self->Filer->conn,
	"-A ".$self->name
    );
    croak $@ if ($@);

    for ( @status ) {
	$status{$_->{name}} = $_;
    }
    return \%status;
}

=head2 ::snapshot_exists

Returns a boolean to indicate whether a given named snapshot exists or not

=over

=item Arguments

B<snapshot> - I<name>

Name of the snapshot to check

=item Example

=back

     $aggr->snapshot_exists('special');

=cut

sub snapshot_exists {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$snapshot) = @_;
    my $ret;

    $ret = SnapCmds::snapshot_exists(
	$self->Filer->conn,
	"-A ".$self->name,
	$snapshot
    );
    croak $@ if ($@);
    return $ret;
}

=head2 ::snapshot_delete

Delete a named snapshot

=over

=item Arguments

B<snapshot> - I<name>

Name of the snapshot to delete

=item Example

=back

      $vol->snapshot_delete('special');

=cut

sub snapshot_delete {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$snapshot,%args) = @_;

    NOTE::logcomment( "Deleting snapshot ".$snapshot );

    SnapCmds::snapshot_delete(
	$self->Filer->conn,
	"-A ".$self->name,
	$snapshot,
	# %args
    );
    croak $@ if ($@);
    return $self;
}

=head2 ::snap_delta

Returns the output of the snap delta command for the aggregate specified, 
in an array of hash references. Keys are taken from header.
Wrapper around Aggregate::aggr_snap_delta.

=over

=item Arguments

B<beginning_snapshot> - I<name>

I<Optional> Name of the beginning_snapshot for the delta

B<ending_snapshot> - I<name>

I<Optional> Name of the ending_snapshot for the delta

=item Examples

=back

      $aggr->snap_delta();
      $aggr->snap_delta('beginning_snapshot' =>'snap1');
      $aggr->snap_delta('beginning_snapshot' =>'snap1', 
                        'ending_snapshot' => 'snap2');

=cut

sub snap_delta {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    my $ret;

    $ret = Aggregate::aggr_snap_delta(
	    $self->{_Filer}->conn,
            $self->{_name},
            %args
            );
    croak $@ if $@;
    return $ret;
}

sub _spareAggrs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $filer = shift;
    my @spareAggrs = ();
    my ($aggrname,$aggr);

    # Get a list of spare aggregates (or tradvols) and their associated disks
    for $aggrname ( $filer->aggregates ) {
	next unless ( $aggrname =~ /^(_ag|_vl)/ );
	eval { $aggr = NOTE::Aggregate->attach( Filer => $filer, name => $aggrname ) };
	next if $@;
	next unless ( $aggr->writeable );
	$aggr->{_disks} = $aggr->disks;
	push @spareAggrs, $aggr;
    }
    return sort { $a->{_disks} <=> $b->{_disks} } @spareAggrs;
}

sub _shareableAggrs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $filer = shift;
    my @shareableAggrs = ();
    my ($aggrname,$aggr);

    # Get a list of shareable aggregates and their associated disks
    for $aggrname ( $filer->aggregates ) {
	next unless ( $aggrname =~ /^AG/ );
	eval { $aggr = NOTE::Aggregate->attach( Filer => $filer, name => $aggrname ) };
	next if $@;
	next unless ( $aggr->writeable );
	$aggr->{_disks} = $aggr->disks;
	push @shareableAggrs, $aggr;
    }
    # Probably should sort according to capacity (most free space first)
    return sort { $a->{_disks} <=> $b->{_disks} } @shareableAggrs;
}

# Release the smallest spare aggregate, if one exists
sub _releaseSpareAggr {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $filer = shift;
    my @spareAggrs = _spareAggrs( $filer );
    return 0 unless ( @spareAggrs );
    my $aggr = pop @spareAggrs;
    if ( $aggr->name =~ /^_ag/i ) {
	NOTE::logcomment( "Releasing spare aggregate ".$aggr->name );
	$aggr->purge;
    } else {
	NOTE::logcomment( "Releasing spare tradvol ".$aggr->name );
	eval { $aggr->offline }; # Don't care if this fails
	$aggr->destroy;	     # Do care if this fails
    }
    return 1;
}

#
# Choose the closest fitting free aggregate.
# If $diskspec is defined, match must be perfect.
# Match against raidtype and raidgroup size.
#
sub _recycle {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, $disks, %args) = @_;

    if ( $self->{_share} ) {
	NOTE::logcomment( "Looking for existing shared aggregate" );
	for my $aggr ( _shareableAggrs( $self->{_Filer} ) ) {
	    if ( $aggr->snaplock || $args{ snaplock } ) {
		next unless ( $aggr->snaplock eq $args{ snaplock } );
	    }
	    my $aggropts = $aggr->options; # So we only have to do this once
	    if ( $args{ raidtype } ) {
		next unless ( $aggropts->{ raidtype } eq $args{ raidtype } );
	    }
	    if ( $args{ rgsize } ) {
		next unless ( $aggropts->{ raidsize } eq $args{ rgsize } );
	    }
	    if ( $disks && ($aggr->{_disks} < $disks) ) {
		# At this point, the only thing wrong with this
		# shared aggregate is that it doesn't contain enough disks.
		# Try to add enough disks to accomodate the request.
		my $err;
		do {
		    eval{ $aggr->add( disks => $disks - $aggr->{_disks} ) };
		    $err = $@;
		} while ( $err && (_releaseSpareAggr( $aggr->Filer )) );
		croak $err if $err;
	    }

            my $aggr_status = $aggr->status->status();  # So we only have to do this once
                if ( $args{ indirect_type } ) {
                    next unless ( $aggr_status =~ /$args{ indirect_type }-bit/);
                }

	    # Don't rename the aggregate, rename the object
	    $self->{_name} = $aggr->{_name};
	    NOTE::logcomment( "Found shared aggregate %s", $aggr->description );
	    return 1;
	}
    }
    if ( $self->{_recycle} ) {
	NOTE::logcomment( "Looking for recyclable aggregate" );
	my @spares = grep { $_->name =~ /^_ag/ } _spareAggrs( $self->{_Filer} );
        my $desperate = 0;
again:
	for my $aggr ( @spares ) {
	    eval { # If any error occurs, simply try the next candidate
		if ( !$desperate ) {
		    # Don't bother with this aggr at all, if the datasets aren't
		    # the same.
		    if ( $self->{_dataset} ) {
			next unless ($aggr->dataset eq $self->{_dataset});
		    } else {
			next if ($aggr->dataset);
		    }
		}

                my $aggr_status = $aggr->status->status();  # So we only have to do this once
                if ( $args{ indirect_type } ) {
                    next unless ( $aggr_status =~ /$args{ indirect_type }-bit/);
                }

                # Find out if the aggregate is a fake large aggr and determine its hole size.
                my $aggr_space = $aggr->show_space();
                my $aggr_size = $aggr_space->{totalSpaceValue};
                $aggr_size =~ s/KB//;

                my @disks = $aggr->disks();
                my $disks_total_size = 0;
                $disks_total_size  = List::Util::sum map { $_->{used_mb}  }  grep { $_->{type} !~ /parity/ } @disks;
                my $disk_size_in_kb = NOTE::units($disks_total_size . 'm');

                $aggr->{_hole_size} = $aggr_size - $disk_size_in_kb;

                if ($aggr->{_hole_size} > 0 && $aggr_status =~ /64-bit/) {
                    $aggr->{_fake_large} = 1;
                } else {
                    $aggr->{_fake_large} = 0;
                }

		if ( $disks ) {
		    next unless ( $aggr->{_disks} >= $disks );
		}
		if ( $args{ snaplock } ) {
		    next unless ( $aggr->snaplock eq $args{ snaplock } );
		}
                next unless ( $aggr->{_fake_large} == $args{fake_large} );

                if ($args{ hole_size } ) {
                    my $reqd_holesize_in_kb = NOTE::units($args{hole_size});
                    next unless ( ($aggr->{_hole_size} >= 0.99 * $reqd_holesize_in_kb) && 
                                  ($aggr->{_hole_size} <= $reqd_holesize_in_kb) );
                }
		my $aggropts = $aggr->options; # So we only have to do this once
		if ( $args{ raidtype } ) {
		    next unless ( $aggropts->{ raidtype } eq $args{ raidtype } );
		}
		if ( $args{ rgsize } ) {
		    next unless ( $aggropts->{ raidsize } eq $args{ rgsize } );
		}
		# Everything looks OK..

		if ( $desperate ) {
		    $aggr->rename; # Grab it
		    if ( $aggr->dataset ) { # Datasets won't match
			$aggr->empty;
		    }
		    if ( $self->{_dataset} ) {
			$aggr->{_dataset} = $self->{_dataset};
			NOTE::logcomment( "Initializing dataset of ".$aggr->description );
			my $func = eval 'sub {'.$aggr->{_dataset}.'}';
			croak $@ if $@;
			eval { $aggr->$func };
			croak $@ if $@;
			if ( $aggr->_looksAnonymous ) {
			    $aggr->_rename( NOTE::datasetEncode( $aggr->{_name}, $aggr->{_dataset} ) );
			    eval{ $aggr->snapshot_delete( "__clean__" ) }; # Don't care if this fails
			    $aggr->snapshot_create( "__clean__" );
			}
		    }
		}

		if ( $self->{_dataset} ) {
		    $self->{_name} = NOTE::datasetEncode( $self->{_name}, $self->{_dataset} );
		}
		NOTE::logcomment(
		    "Recycling Aggregate; ".
		    $aggr->name, " --> ".  $self->{_name}
		);
		$aggr->_rename( $self->{_name} );

		if ( $aggr->is_shareable ) {
		    # Delete the __clean__ and __empty__ snapshots if we
		    # are recycling an aggregate in order to make a shareable one.
		    eval{ $self->snapshot_delete( "__clean__" ) }; # Don't care if this fails
		    eval{ $self->snapshot_delete( "__empty__" ) }; # Don't care if this fails
		} else {
		    if ( ! $aggr->snapshot_exists( "__empty__" ) ) {
			$aggr->purge;
			next;
		    }
		    if ( ! $aggr->snapshot_exists( "__clean__" ) ) {
			$aggr->empty;
			# If a dataset is required, this aggregate is no longer appropriate
			next if ( $self->{_dataset } );
		    }
	        }
		$self->{_was_recycled} = 1;
		goto gotOne; # got one
	    };
	    NOTE::logcomment( "Attempted recycle failed: $@" ) if ($@);
	}
	# If I get here, no candidates were found.
	# Try once more, but this time don't reject aggregates with
	# datasets we're not looking for. If we find one that meets all
	# the other requirements, just change the dataset to be what
	# we want it to be.
	if ( !$desperate ) {
	    NOTE::logcomment( "No candidate found; trying harder" );
	    $desperate = 1;
	    # Re-sort @spares according to creation difficulty:
	    # Lowest number of disks and smallest population of data first.
	    for (@spares) {
		# Cache 'used', so the sort is efficient.
		$_->{_used} = $_->used;
	    }
	    @spares = sort {
		$a->{_disks} <=> $b->{_disks}
			     ||
		$a->{_used}  <=> $b->{_used}
	    } @spares;
	    goto again;
        }
    }
    NOTE::logcomment( "No candidate found" );
    # XXX If I want an aggregate with a dataset, I should be
    # able to convert an empty aggregate to a populated aggregate
    # if no appropriate populated aggregate exists.
    # And vice-versa - If I want an empty aggregate,
    # and none exist, I might want to consider removing
    # data from a populated aggregate, if there are no more
    # disks available.
    return 0;
gotOne:
    return 1;
}

=head2 ::dataset

String (of code) representing the dataset of this object.

=over

=item Arguments

None.

=item Example

=back

     printf( "Dataset of this aggregate: %s\n", $aggr->dataset );

=cut

sub dataset {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_dataset} }

=head2 ::tag

Sets new tag if given or reads aggregate tag if called without arguments

=over

=item Argument

B<newtag> = string

I<Optional> Set tag to given value. Tag will printed during script execution
along with aggregate name instead of just faceless anonymous name

=item Examples

    $aggr->tag( 'Source Aggregate' );

    my $atag = $aggr->tag;

=back

=cut

sub tag {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$newtag) = @_;

    if ( defined $newtag ) {
	$self->{_tag} = $newtag;
    }
    return $self->{_tag};
}

=head2 ::description
Returns a description of the aggregate consisting of
its filer's name, its name and its tag, if any.

=over

=item Arguments

None.

=item Examples

    logcomment( $aggr->description );

=back

=cut

sub description {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $txt;

    $txt = $self->hostpath;
    if ( $self->{_tag} ) {
	$txt .= " (".$self->{_tag}.")";
    }
    return $txt;
}

# Does a aggr copy ($self is destination)

=head2 ::copy

Execute aggr copy command from source to self. Brings aggregate online
after aggr copy command is completed. If aggr online failed then
command is retried up to 5 times before failing

=over

=item Arguments

B<src> aggregate object

I<Required> Source aggregate object. Object is the NOTE's object.

B<background> = 1|0

I<Optional> Background or foreground execution flag

=item Examples

    $dst_aggr->copy( src => $src_aggr );

=back

=cut

sub copy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    my $background = $args{background};
    delete $args{background};
    my $src = $args{src};
    delete $args{src};
    my %ARGS = ();
    my $was_restricted = $self->restricted;

    croak "src must be specified for copy" unless ( $src );
    
    NOTE::logcomment( "Copying ".$src->description." -> ".$self->description );
    $ARGS{uc $_}    = $args{$_} for ( keys %args );
    $ARGS{COMPLETE} = $background ? 0 : 1;
    $ARGS{SOURCE}   = $src->name;

    $self->restrict unless ( $was_restricted );
    Aggregate::Copy::aggrcopy_start(
	$self->Filer->conn,
	$self->name,
	%ARGS,
    );
    croak $@ if $@;

    unless ( $was_restricted ) {
	my $try = 0;
	for $try ( 1..5 ) {
	    eval { $self->online };
	    last unless ( $@ );
	    sleep 1;
	}
	if ( $try >= 5 ) {
	    croak "Couldn't bring ".$self->name." online after aggrcopy: $@";
	}
    }
    return $self;
}

=head2 ::newName

Generates new anonymous name for aggregate. Name starts with 'AG' for
shared aggregate or with 'ag' for non shared aggregate. This method
doesn't execute any filer's command

=over

=item Arguments

B<share> = 1|0

I<Optional> Define aggregate type. By default share=0.

=item Examples

    $aggr->newName( shared => 1 );

=back

=cut

sub newName {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    if ( $args{share} ) {
	return shift->SUPER::newName( "AG" )
    } else {
	return shift->SUPER::newName( "ag" )
    }
}

=head2 ::Filer

Returns NOTE filer's object of the filer aggregate belong to.

=over

=item Arguments

None.

=item Examples

    my $filer = $aggr->Filer;

    $aggr->Filer->reboot;

=back

=cut

sub Filer {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_Filer} };

=head2 ::name

Returns name of aggregate.

=over

=item Arguments

None.

=item Examples

    printf( "Aggregate name is: %s\n", $aggr->name );

=back

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_name} };

=head2 ::is_recyclable

Returns 1 if the aggregate is a recyclable aggregate.
This is also assignable, if it is desired to change the
status.

=over

=item Arguments

None.

=item Examples

    if ( $aggr->is_recyclable ) {
	printf( "Aggregate %s is recyclable\n", $aggr->name );
    }

    $aggr->is_recyclable = 0;

=back

=cut

sub is_recyclable :lvalue {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    # [See comments in ::is_anonymous]
    my $r = \$self->{_recycle};
    $$r;
};

=head2 ::is_anonymous

Returns 1 if aggregate is anonymous. This is also assignable,
if it is desired to change the status. Anonymous aggregates are
destroyed or recycled when the script exits.

=over

=item Arguments

None.

=item Examples

    unless( $aggr->is_anonymous ) {
	...
    }

    $aggr->is_anonymous = 1;

=back

=cut

sub is_anonymous :lvalue {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    # Must do this trickery due to a bug in perl regarding
    # tied hashes and lvalues.
    # Actually, for NOTE::Aggregate.pm, this isn't strictly necessary
    # because while NOTE::Volume objects are tied hashes, NOTE::Aggregate
    # objects are not (nothing interesting in the registry about them).
    # But I put this workaround here anyway, in case they ever become
    # tied hashes.
    my $r = \$self->{_anonymous};
    $$r;
};


=head2 ::is_free

Returns 1 if the aggregate is available for reuse. An aggregate is recyclable
if the name starts with '_ag'

=over

=item Arguments

None.

=item Examples

    for my $aggr ( $filer->Aggregates ) {
	next unless ( $aggr->is_free );
	printf( "Aggregate %s is available for reuse\n", $aggr->name );
    }

=back

=cut

sub is_free {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_name} =~ /^_ag/i }

=head2 ::is_recycled

Returns 1 if aggregate was recycled, i.e. obtained from the
pool of available aggregates. Returns 0, otherwise. A shared
aggregate is not considered recycled.

=cut

sub is_recycled {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_was_recycled} }

=head2 ::is_shareable

Returns 1 if aggregate is shared aggregate. Aggregate is
shareable if aggregate name starts with 'AG'

=cut

sub is_shareable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_name} =~ /^AG/ }

=head2 ::writeable

Returns 1 if aggregate is writable. Aggregate is 
writable if state is 'online' and it's not 'read-only'
and it's not 'restricted'

=cut


sub writeable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $state = $self->status->{'state'};

    return (
    	$state =~ /online/ &&
    	$state !~ /read-only/ &&
    	$state !~ /restricted/
    );
}

=head2 ::blocks

Returns total nummber of blocks in aggregate. Output is taken from 'aggr status -b' command

=cut

sub blocks {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $blocks;
    eval {
	$blocks = Aggregate::aggr_blocks( $self->{_Filer}->conn, 
					  $self->{_name} );
    };
    croak $@ if $@;
    return $blocks->{blocks};
}


=head2 ::block_size

Returns number of bytes per block. Output is taken from 'aggr status -b' command

=cut

sub block_size {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $blocks;
    eval {
	$blocks = Aggregate::aggr_blocks( $self->{_Filer}->conn, 
					  $self->{_name} );
    };
    croak $@ if $@;
    return $blocks->{block_size};
}

=head2 ::fs_size

Returns aggregate size in blocks. Output is taken from 'aggr status -b' command

=cut

sub fs_size {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $blocks;
    eval {
	$blocks = Aggregate::aggr_blocks( $self->{_Filer}->conn, 
					  $self->{_name} );
    };
    croak $@ if $@;
    return $blocks->{fs_size};
}


=head2 ::disks

Returns number of disks in aggregate. Output is taken from 'aggr status $aggr_name -d' command.

=cut

sub disks {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    if ( wantarray ) {
	my @disks = Aggregate::aggr_disks( $self->{_Filer}->conn, $self->{_name} );
	croak $@ if $@;
	if ( scalar @disks == 0 ) {
	    croak "Aggregate ".$self->description." reports ZERO disks";
	}
	$self->{_disks} = scalar @disks;
	return @disks;
    } else {
	unless ( $self->{_destroyed} ) {
	    # Unfortunately, buggy formating code in ONTAP sometimes generates
	    # unparsable output for the aggr status <name> -d command, so if I just
	    # want a count of disks, this is a much more reliable method:
	    # Added FC metro cluster disk output format
	    # Examples:
	    #  Regular:
	    # Device          HA    SHELF BAY CHAN    Disk Vital Product Information
	    # ----------      --------------- -----   ------------------------------
	    # 0a.35   0a    2   3     FC:A    5QD2754D

	    # Metro cluster format:
	    # Device          HA    SHELF BAY CHAN    Disk Vital Product Information
	    # ----------      --------------- -----   ------------------------------
	    # mc-sw14:5.22    0b    1   6     FC:A    3LN3TRM600009829K6U4

	    $self->{_disks} = scalar grep /^\s*(.*:|[v]?)\d/, split( /\n/,
		filer::filer_cli_say( $self->Filer->conn, "aggr status ".$self->name." -d" )
	    );
	}
	return $self->{_disks}; # return last known number of disks
    }
}


=head2 ::total

Returns overall aggregate size in KB. Output is taken from 'df -A -k' command

=cut

sub total    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return (shift->disk_free)[0] }

=head2 ::used

Returns used size in KB. Output is taken from 'df -A -k' command

=cut

sub used     {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return (shift->disk_free)[1] }

=head2 ::avail

Returns available space in KB. Output is taken from 'df -A -k' command

=cut

sub avail    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return (shift->disk_free)[2] }

=head2 ::capacity

Returns capacity in '%'. Output is taken from 'df -A -k' command.

=cut

sub capacity {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return (shift->disk_free)[3] }


=head2 ::disk_free

Returns values of 'total', 'used', 'avail', 'capacity' as hash for self or given path.
Output is taken from 'df -A -k' command.

=over

=item Syntax

$aggr->disk_free( path=>'/vol/vol0/etc' );

=item Arguments

B<path> string

I<Optional> Path for 'df -A -k' command. Default is plain aggregate name

=back

=cut

sub disk_free {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;

    if ( defined $args{path} ) {
        croak "\"path\" may not be a key in the options hash";
    }
    $args{path} = $self->name;
    my $out = Aggregate::aggr_disk_free(
        $self->Filer->conn,
        %args
    );
    croak $@ if $@;
    return wantarray ? ($out->{total}, $out->{used}, $out->{avail}, $out->{capacity}) : $out;
}


=head2 ::show_space

Parses the output of the aggr show_space command and returns a hashref. All sizes are in KB.

=over

=item Syntax

my $aggr_space = $aggr->show_space();

=item Examples
Each field and its value reported by aggr show_space is stored in the hash as key/value pairs.
The space usage for all volumes in the aggregate is stored within an array in this hash.
This is what the hash looks like.

{
  name => aggr1,
  totalSpaceValue => 36030780416KB,
  waflReserveValue => 3603078040KB,
  snapReserveValue => 3621385116KB,
  usableSpaceValue => 30806317260KB,
  bsrNVLOGValue => 0KB,
  asisValue => 0KB,
  imageBackupValue => 0KB,

  vols => [
    {
      allocated => 25912969348KB,
      guarantee => volume,
      name => vol1,
      used => 712KB
    },
    {
      allocated => 3742969348KB,
      guarantee => volume,
      name => vol2,
      used => 12KB
    }
  ],

  totalSpace => {
    allocated => 25912969348KB,
    avail => 4892224064KB,
    used => 712KB
  },
  snapReserve => {
    allocated => 1621385116KB,
    avail => 1621384872KB,
    used => 244KB
  },
  waflReserve => {
    allocated => 3603078040KB,
    avail' => 3601952052KB,
    used => 1125988KB
  },
}

=back

=cut

sub show_space {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $out = Aggregate::aggr_show_space(
        $self->Filer->conn,
        '-k',
        $self->name
    );

    croak $@ if $@;
    return $out->[0];
}


=head2 ::exists

Returns true if aggregate still exists.

=over

=item Examples

    eval { $aggr->destroy };
    if ( $aggr->exists ) {
        croak "Unable to destroy aggregate";
    }

=back

=cut

sub exists {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $ret;
    my $name = $self->{_name};
    $ret = Aggregate::aggr_exists( $self->{_Filer}->conn, $name );
    if ( $@ ) {
	# aggr_exists sets $@ if the volume doesn't exist
	# Croak only if that's not the reason.
	croak $@ unless ( $@ =~ /No .* named '$name' exists/ );
    }
    return $ret;
}


=head2 ::restricted

Returns true if aggregate is restricted

=cut

sub restricted {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my ($state, $status, $aggr, $options) = Aggregate::aggr_status(
        $self->Filer->conn, $self->name
    );
    croak $@ if $@;
    return $state =~ /restricted/;
}


=head2 ::wafliron_start

Start WAFL iron on aggregate. By default it runs in the background. Set I<wait> arguement
to 1 if foreground execution is required. If executed as a background process then 
the subtest object is saved in $self->{_wafliron_wait} private variable.

This method will croak if the aggregate is shared, because snapmirror activity by other
scripts will cause this to fail. Wafliron can only reliably be run in private anonymous
aggregates.

=over

=item Arguments

B<wait> = 1|0

I<Optional> Wait for wafliron to finish and exit. By default it start wafliron as background process.

=item Examples

    $aggr->wafliron_start( wait => 1 );

=back

=cut

sub wafliron_start {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    # Don't allow wafliron to run on a shared aggregate. This is
    # unreliable.
    croak "Cannot start wafliron on a shared aggregate ".$self->description
	if ( ref($self) =~ /NOTE::Aggregate/ && $self->is_shareable );
    NOTE::logcomment( "Starting Wafliron on ".$self->description );
    my @out = Wafliron::wafliron_start( $self->Filer->conn, 
					$self->name, 
					%args );

    # user request not to wait for wafliron to complete
    if ( !$args{wait} ) {
	$self->{_wafliron_wait} = [ @out ];
	# Cause it to be stopped on teardown, if still running
	NOTE::reference( sub { $self->wafliron_stop if $self->{_wafliron_wait} } )
		if @out < 2;
    }
    return @out;
}


=head2 ::wafliron_stop

Termintate WAFL iron process.

=cut

sub wafliron_stop {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    NOTE::logcomment( "Stopping Wafliron on ".$self->description );
    if ( my $id = _save_wafliron_id($self) ) {
	$args{id} = $id;
    }
    return Wafliron::wafliron_stop( $self->Filer->conn, $self->name, %args );
}


=head2 ::wafliron_status

Return WAFL iron status string

=cut

sub wafliron_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return Wafliron::wafliron_status( $self->Filer->conn, $self->name );
}

=head2 ::wafliron_wait

Wait for WAFL iron to finish or exit with error "I<Wafliron doesn't seem to be running on this aggregate>"

=cut

sub wafliron_wait {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    if ( my $id = _save_wafliron_id($self) ) {
	NOTE::logcomment( "Waiting for Wafliron on ".$self->description );
	return Wafliron::wafliron_wait( $id )
    } else {
	my ($parent, $obj_name) = split /:/, ref($self);
	croak "Wafliron doesn't seem to be running on this ".lc($obj_name);
    }
}

sub _save_wafliron_id {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self) = @_;

    my $id = undef;
    if ( exists $self->{_wafliron_wait} ) {
	my @out = @{$self->{_wafliron_wait}};
	delete $self->{_wafliron_wait};
	# array will have only 1 elements if it is an id
	$id = $out[0] if @out < 2;
    }
    return $id;
}

=head2 ::mirror

Mirrors aggregate using syncmirror if not already mirrored.

If syncmirror is licensed then it simply runs aggr mirror command. If syncmirror is not licensed
then this method will license syncmirror_local service and will reboot filer because it's
required by syncmirror. Aggregate will be mirrored after reboot. Relies on cluster::VolUtils::aggr_mirror.

=over

=item Arguments

B<timeout> = seconds

I<Optional> sync timout in seconds, default is 7200 seconds (2 hrs).

=item Examples

    $aggr->mirror;

    $aggr->mirror( timeout=>300 );

=back

=cut

sub mirror {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    my $sync_timeout = $args{timeout} ? $args{timeout} : 7200;

    return 1 if( $self->is_mirrored );
    return cluster::VolUtils::aggr_mirror( fc => $self->Filer->conn, aggr => $self->name, timeout => $sync_timeout );
}

=head2 ::is_mirrored

Checks if agregate is syncmirrored. Returns 1 if it's mirrored.

=over

=item Arguments

None

=item Examples

    if( $aggr_a->is_mirrored ) {
	$aggr_b = $aggr_a->split;
    } else {
	croak "Unable to split aggregate, aggregate is not mirrored";
    }

=back

=cut

sub is_mirrored {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return ((shift->status)[1] =~ /mirrored/ ) }

=head2 ::upgrade_64bit_start

Upgrades aggregate to 64 bit

=over

=item Arguments

<B>mode => <check/normal/grow-reserved/grow-all> 
Defaults to normal

=item Examples

    $aggr->upgrade_64bit_start;

=back

=cut

sub upgrade_64bit_start() {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    my $count = 0;

    ##If mode is not specified, default it to normal mode
    my $mode = exists $args{mode} ? $args{mode} : "normal";
    my $filer = $self->{_Filer};
    my $aggr_name = $self->name();

    ##Trigger 64 bit upgrade on aggregate
    NOTE::logcomment("Upgrading aggregate $aggr_name to 64 bit\n");
    $filer->say("priv set -q diag; aggr status $aggr_name");
    my $text = $filer->say("priv set -q test;aggr 64bit-upgrade start $aggr_name -mode $mode");

    if ($text =~ /Checking for additional space required to upgrade/i){
        ##If one or more volume in aggregate are %100 full, throw a warning on check
        if($mode =~ /check/i){
            if ($text =~ /increase the volume size/i){
            NOTE::logcomment("** WARN:\n Some of the volumes in $aggr_name do not have enough free space: \n$text\n");
	    return 1;
            }
	    else{
	        NOTE::logcomment("Output of check on aggregate before 64 bit upgrade:\n $text\n");
	        }
            }
        else {
            if($mode =~ /normal/i){
   	        ##If one or more volume in aggregate are %100 full, upgrade fails
	        if ($text =~ /One or more volumes do not have enough free space to support the 64-bit upgrade/i) {
	            croak "64 bit upgrade on aggregate $aggr_name failed: \n $text\n";
	            }
	        ##aggr 64 bit upgrade is not supported in trad vols
	        elsif ($text =~ /Invalid operation on traditional volume/i) {
	            croak "64 bit upgrade is not supported on traditional volume: \n$text\n";
	            }
                }

	    ##Verfying whether upgrade succeeded	
            if ($text =~ /Started 64-bit upgrade on aggregate $aggr_name/i) {
	        NOTE::logcomment("*** 64-bit upgrade started on aggregate $aggr_name");
	        do {
	            if($count >2){
		        croak "64-bit upgrade on aggregate $aggr_name failed\n";
		        }
		    sleep 5;	
		    $self->upgrade_64bit_status;
		    $text = $filer->say("priv set -q diag; aggr status $aggr_name");
		    $count++;
		    }while($text !~ /64-bit/i);
	        NOTE::logcomment("64-bit upgrade on aggregate $aggr_name succeeded\n");
	        }
	   else { 
	       croak "Failed to start upgrading aggregate to 64 bit\n Error String: \n $text \n";
               }
	   }
        }
        else {
		croak "Upgrade on aggregate $aggr_name failed\n $text\n";
	     }
        return 1;
      }

=head2 ::upgrade_64bit_scan_start

Starts 64 bit upgrade scanner on aggregate

=over

=item Arguments

None.

=item Examples

=back

    $aggr->64bit_upgrade_scan_start();

=cut

sub upgrade_64bit_scan_start() {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $aggr=$self->name;
    my $result = $self->Filer->say("priv set -q diag; aggr 64bit-upgrade scan-start ".$aggr." ");
    unless ($result =~ /(Started|Starting) the 64-bit upgrade scanner on aggregate $aggr/i){
        croak "Starting 64-bit upgrade scanner on $aggr failed:\n $result\n";
	}
    NOTE::logcomment("64-bit upgrade scanner start on $aggr succeeded\n");
    return 1;
    }

=head2 ::upgrade_64bit_scan_stop

Stops 64 bit upgrade scanner on aggregate

=over

=item Arguments

None.

=item Examples

=back

    $aggr->64bit_upgrade_scan_stop();

=cut

sub upgrade_64bit_scan_stop() {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $aggr=$self->name;
    my $result = $self->Filer->say("priv set -q diag; aggr 64bit-upgrade scan-stop ".$aggr." ");
    unless ($result =~ /(Stopping|Stopped) the 64-bit upgrade scanner on aggregate $aggr/i){
        croak "Stopping 64-bit upgrade scanner on $aggr failed:\n $result\n";
        }
    NOTE::logcomment("64-bit upgrade scanner stop on $aggr succeeded\n");
    return 1;
    }

=head2 ::upgrade_64bit_status

Returns the status of 64 bit upgrade on the aggregate

=over

=item Arguments

None.

=item Examples

=back

    $aggr->64bit_upgrade_status();

=cut

sub upgrade_64bit_status{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
   
    my %hash;
    my $aggr=$self->name;

    #Check aggregate upgrade status
    my $text = $self->Filer->say("priv set -q diag; aggr 64bit-upgrade status ".$aggr);
    if ($text eq "64-bit upgrade is not currently in progress on aggregate $aggr or its contained volumes") {
        return undef;
        }
    elsif ($text =~ /No aggregate named $aggr was found/){
	croak "Aggregate not found\n";
	}
    else { #Return the status in hash
    	    my @lines = split("\n",$text);
    	    foreach my $line (@lines){
	        if ($line =~ /^Volume\s+Format\s+Scanner Status\s+\% Completed\s+Time to Completion\s+Progress$/){
	            next;
	            }	
                if ($line =~ /^(\w+)\s+(\w+)\s+(\w+)\s+(\d+)\s+(\w+\/\w+|\w+)\s+fbn (\d+), inode (\d+) of (\d+), (\w+)$/){
                    $hash{"Volume"} = $1;
                    $hash{"Format"} = $2;
                    $hash{"Scanner_status"} = $3;
                    $hash{"Percentage_Completed"} = $4;
                    $hash{"Time_for_completion"} = $5;
                    $hash{"fbn"} = $6;
                    $hash{"inode"} = $7;
                    $hash{"total_inodes"} = $8;
                    $hash{"type"} = $9;
		    return %hash;
		    }
                }
          }
    }

=head2 ::mixed_aggregate

Converts aggregate from 32 bit to mixed aggregate format.

=over

=item Arguments

None.

=item Examples

=back

     $aggr->mixed_aggregate();

=cut

sub mixed_aggregate(){
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    # Reduce speed of wafl scan speed
    $self->Filer->set_wafl_scan_speed(speed=>25);

    # Start upgrading aggregate in background
    $self->upgrade_64bit_start(mode=>"grow-all",background=>1);

    # Make sure aggregate is upgrading & wafl scan is in progress
    my %status;
    do {
        %status = $self->upgrade_64bit_status;
        sleep 3;
    }while ( ($status{"Scanner_status"} != "running") && ($status{"Format"} != "upgrading") );

    # Stop the wafl scan
    $self->upgrade_64bit_scan_stop;

    # Make sure wafl scan stopped & aggregate is in upgrading state
    %status = $self->upgrade_64bit_status;
    unless ( ($status{"Scanner_status"} == "stopped") && ($status{"Format"} == "upgrading") ){
        croak "unexpected behaviour";
        }

    # Use vol_db command to check whether aggregate is in mixed format
    my $name = $self->name;
    my $handle = $self->Filer->say("priv set -q diag;vol_db lookup name $name");
    croak("Unable to lookup the handle for $name.\n")
      if ( $handle =~ m/not found/i );

    my $output = $self->Filer->say("priv set -q diag;vol_db get buftree_type $handle");
    croak("Unable to convert $name into mixed aggregate\n $output")
      if ( ($output =~ m/trad64(upgrading)/) || ($output =~ m/trad32(upgrading)/) || ($output =~ m/dual32(upgrading)/) || ($output =~ m/dual64(upgrading)/) );

    return 1;
}

=cut

=head2 ::purge

Delete all volumes within aggregate and then destroy aggregate itself. Uses
L<C<aggr_purge>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_purge($conn,$aggr,[%opts])>
function from lib/Aggregate.pm

=over

=item Arguments

B<vol_only> = 1|0

I<Optional> Delete all volumes within aggregate but do not destroy aggregate itself. Default is 0.

=item Examples

    $aggr->purge( vol_only => 1 );

=back

=cut

sub purge {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    if ( $args{vol_only} == 1 ) {
	NOTE::logcomment( "Deleting contents of aggregate ".$self->description );
    } else {
	NOTE::logcomment( "Purging Aggregate ".$self->description );
    }
    Aggregate::aggr_purge( $self->{_Filer}->conn, $self->{_name}, %args );
    croak $@ if $@;
    # if vol_only => 1 was specified, only the volumes will be destroyed
    # NOT the aggregate
    unless ( $self->exists ) {
	$self->{_destroyed} = 1;
	$self->resource( 'destroy' );
	$self->{_Filer}->zero_spares; # Zero spares on teardown
    }
    return $self;
}


=head2 ::destroy

Destroy aggregate. This method uses 
L<C<aggr_destroy>|engineering::ontap_tests::agnostic::docs::lib-Aggregate-pm/aggr_destroy($conn,$aggr)>
function from lib/Aggregate.pm

=cut

sub destroy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    NOTE::logcomment( "Destroying Aggregate ".$self->description );
    Aggregate::aggr_destroy( $self->{_Filer}->conn, $self->{_name} );
    croak $@ if $@;
    $self->{_destroyed} = 1;
    $self->resource( 'destroy' );
    $self->{_Filer}->zero_spares; # Zero spares on teardown
    return $self;
}

=head2 ::empty

If an __empty__ snapshot exists, the aggregate will be restored
using this snapshot. 

Croaks on error.
Returns $self.

=over

=item Examples

=back

     $aggr->empty;

=cut

sub empty {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    unless ( $self->is_anonymous && !$self->is_shareable ) {
	croak "Cannot empty a non-anonymous or shareable aggregate: ".$self->description;
    }
    if ( $self->snapshot_exists( "__empty__" ) ) {
	$self->snapshot_restore( "__empty__" );
	if ( $self->{_dataset} ) {
	    delete $self->{_dataset};
	    $self->rename if ( $self->_looksAnonymous );
	}
	eval{ $self->snapshot_delete( "__clean__" ) }; # Don't care if this fails
	$self->snapshot_create( "__clean__" );
    } else {
	croak "Could not restore aggregate: __empty__ snapshot missing: ".$self->description;
    }
    return $self;
}

=head2 ::clean

If a __clean__ snapshot exists, the aggregate will be restored
using this snapshot. 

Croaks on error.
Returns $self.

=over

=item Examples

=back

     $aggr->clean;

=cut

sub clean {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    unless ( $self->is_anonymous && !$self->is_shareable ) {
	croak "Cannot clean a non-anonymous or shareable aggregate: ".$self->description;
    }
    if ( $self->snapshot_exists( "__clean__" ) ) {
	$self->snapshot_restore( "__clean__" );
    } else {
	croak "Could not restore aggregate: __clean__ snapshot missing: ".$self->description;
    }
    return $self;
}

=head2 ::SPARE

Don't physically destroy (or recycle), even if anonymous. Clears recycle flag and sets spare flag.

=cut

# Don't physically destroy (or recycle), even if anonymous
sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    $self->{_recycle}  = 0;
    $self->{_spare_me} = 1;
}


sub release {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    $self->_release;
    $self->unreference; # Remove from teardown stack
    return $self;
}

sub _release {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return unless ( $self->{_anonymous} );
    return if ( $self->{_share} );
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    $self->{_destroyed} = 1;
    if ( NOTE->RECYCLE && $self->{_recycle} ) {
	eval { $self->online };
	if ( $@ && $@ !~ /is already online/ ) {
	    NOTE::logcomment( "Online of ".$self->{_name}." failed: $@" );
	} else {
	    eval { $self->clean };
	    unless ( $@ ) {
		NOTE::logcomment( "Returning Aggregate ".$self->{_name}." to free pool" );
		if ( !$self->_looksAnonymous ) {
		    $self->_rename( "_".$self->newName );
		} else {
		    $self->_rename( "_".$self->name );
		}
		$self->resource( 'free' );
		return;
	    }
	    NOTE::logcomment( "Restore to clean state of ".$self->{_name}." failed: $@" );
	    # restore using __clean__ failed, so just destroy the aggregate
	}
    }
    eval { $self->purge };
}

=head2 ::meta_i2p

Return hashref of meta_i2p data for the aggregate,
with fileids as keys, and filenames as values.

=over

=item Arguments

None.

=item Examples

=back

     $data = $aggr->meta_i2p();

=cut

sub meta_i2p {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $data = Aggregate::aggr_meta_i2p($self->Filer->conn(), $self->name());

    if (!$data) {
        croak 'Cannot fetch inode->pathname mapping for aggregate ' . $self->name();
    }
    return $data;
}

=head2 ::DESTROY

Object destroy for aggregate object.

=cut

sub DESTROY {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    return unless ( $self->{_anonymous} );
    return if ( $self->{_share} );
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );
    return if ( NOTE->ALL_SPARED );
    if ( $self->{_spare_me} ) {
    	NOTE::logcomment( "Aggregate ".$self->{_name}." has been spared" );
	return;
    }
    $self->release;
}


=head1 STATUS

Public.

$Date: 2013/04/24 $

=cut

1;
