## -*- perl -*-
##$Id: //depot/prod/test/Rsierranevada/lib/NOTE/CFO.pm#1 $
##
##
##Copyright (c) 2008 Network Appliance, Inc.
##All rights reserved.
##
## @summary  NOTE Cluster Fail Over (CFO) module
## @status public
## @pod    here
## @author pimin@netapp.com, dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::CFO

=head1 DESCRIPTION

This package provides basic CFO functionalities. 

Given a <filer> the CFO configuration for <filer> is discovered and an object representing 
the CFO filer pair is created.  The CFO::new method requires the name of a filer, or a NOTE::Filer 
object, which is configured with a CFO partner.  In many thpl scripts the script is passed 
FILER_SRC and FILER_DST.  Either FILER_SRC or FILER_DST may be partnered with each other or 
with another filer, which is neither FILER_SRC nor FILER_DST.  The invoking test must determine 
what is a legitimate configuration for that specific test.  

=head1 METHODS

=cut

=head2 B<::new>

Given a <filer> CFO::new discovers the CFO configuration for that <filer> and returns an object 
representing the CFO filer pair, if it exists.  The method requires a <filer> which is configured 
with a CFO partner.  <filer> is referred to as ME and ME's partner is referred to as PARTNER.  The
filer pair are partners.

The ME|PARTNER designation is without regard for any other relationship(s) between the filers.

CFO::new will return with both filers in the 'UP' state.  If the CFO pair is in 'TAKEOVER' & 'DEAD' 
states CFO::new will attempt to giveback and wait for the 'UP' &'UP' states.  If both filers in the 
'UP' state cannot be achieved the method will croak.  Filers in other transition states between 
TAKEOVER & DEAD (or DEAD & TAKEOVER) and 'UP & 'UP' will not be detected as correctable and the 
method will croak.

This method will croak if 

	a partner cannot be discovered,
	<Filer> partners with itself,
	a connection to either filer cannot be made,
	partners were in 'TAKEOVER' & 'DEAD' states and could not be forced into 'UP' & 'UP' states,
	either or both partners are not in the 'UP' state or
	partners are not configured too failover to each other

=over 

=item Arguments

B<Filer> => I<host spec>

I<Required.>
Returns a class CFO which coordinates a <filer> and it's partner for takeover and giveback.
Given a <filer>, known as "ME", the partner filer, known as "PARTNER", is discovered and a 
CFO object is constructed which coordinates the partners.

B<%args>

I<Optional.>
All parameters, in the form of a hash, are passed to NOTE::Filer->connect.

=item Examples

	my $pair = NOTE::CFO->new(Filer => $FILER_SRC);

=back

=cut

use strict;
use NOTE;
package NOTE::CFO;
our @ISA = qw(NOTE);

use Data::Dumper;
use Carp      qw();
use Params::Validate qw( validate validation_options SCALAR BOOLEAN HASHREF OBJECT );
use Sysconfig qw();

sub new {
	my ($class, %args)  = @_;
	if (ref $class) {
		%args = $class;
		$class = 'CFO';
	}
	$Carp::Verbose = 1;
	my ($A, $B);
	my $ME_filer = delete $args{Filer};
	#
	#	Check what kind of object has been passed as the <filer>
	#
	if (ref ($ME_filer) eq "NOTE::Filer") {
		$A = $ME_filer;
	} else {
		$A = NOTE::Filer->connect(host =>  $ME_filer, %args);
	}
	#
	#    set _CFO_B to  _CFO_A's partner
	#
	my $sysConfig =  Sysconfig->new($A->host);
	my $partner_name = $sysConfig->partner_name;
	if (!defined $partner_name || $partner_name eq '') {
		Carp::croak ("CFO::new Can not get PARTNER name for filer ". $A->host);
	}
	if ($partner_name eq $A->host) {
		Carp::croak ("CFO::new filer pair is single filer ME = \"" . $A->host . "\" PARTNER = \"" . $partner_name . "\"");
	}

	my $B_sysConfig = Sysconfig->new($partner_name);
	my $partner_partner = $B_sysConfig->partner_name;
	
	if ($A->host ne $partner_partner) {
		my $me = $A->host;
		Carp::croak("CFO::new ME ($me) fails over to PARTNER ($partner_name) which fails over to $partner_partner\n");
	}
	$B = NOTE::Filer->connect(host => $partner_name);

	my $self = {
		_CFO_A   => $A,		# NOTE::Filer object
		_CFO_B   => $B,		# NOTE::Filer object
#		_TAKEN   => undef,	# if this exists then it is the filer connection
		              		# for the filer that did the takeover and MUST do the giveback
		_owner  => $$
	};

	bless ( $self );
	NOTE::reference( $self );

	$self->{_destroy} = 1;	# there really isn't anything to destroy yet

	my ($A_state, $B_state) = $self->status;
	my $AB_state = $A_state . $B_state;
	#
	#	Check for filers in UP and UP states
	#
	unless ($AB_state eq "UPUP") {
		if ($AB_state =~ /(TAKEOVERDEAD|DEADTAKEOVER)/ ) {
			NOTE::logcomment ("CFO::new attempting to force a giveback on filers in states \"$A_state\" \"$B_state\"");
			#
			# pick which filer is in TAKEOVER state (it has taken over)
			#
			$self->{_TAKEN} = $AB_state =~ /TAKEOVERDEAD/ ?  $self->{_CFO_A} : $self->{_CFO_B};
			#
			# tell it to giveback
			#
			$self->giveback(force => 1);
			#
			#wait for it
			#
			$self->monitor;
		}
		($A_state, $B_state) = $self->status;
		$AB_state = $A_state . $B_state;
		if ($AB_state !~ /UPUP/) {
			Carp::croak ("CFO::new Both filers must be 'UP' to start - \"$A_state\" \"$B_state\"");
		}
		NOTE::logcomment ("CFO::new did giveback work?? \"$A_state\" \"$B_state\"");
	} else {
		NOTE::logcomment ("CFO::new found filers in UP & UP \"$A_state\" \"$B_state\"");
	}
	delete $self->{_destroy};	# there still isn't much to destroy

    my $A_version = Filer::Version::get_ontap_version( $A->hostrec );
    my $B_version = Filer::Version::get_ontap_version( $B->hostrec );

    if (Filer::Version::version_compare($A_version, '8.0') < 0) {
        $self->{_CFO_A}->license('cluster');
    } elsif(Filer::Version::version_compare($A_version, '8.2') == 0) {
        $self->{_CFO_A}->registry->{'options.cf.mode'} = "ha";
    } else {
        $self->{_CFO_A}->license('cf');
    }

    if (Filer::Version::version_compare($B_version, '8.0') < 0) {
        $self->{_CFO_B}->license('cluster');
    } elsif(Filer::Version::version_compare($B_version, '8.2') == 0) {
        $self->{_CFO_B}->registry->{'options.cf.mode'} = "ha";
    } else {
        $self->{_CFO_B}->license('cf');
    }

	if (cluster::cf_enabled ($self->{_CFO_A}->hostrec, $self->{_CFO_B}->hostrec, 10, 10) != 1 ) {
		Carp::croak ("CFO::new Both filers must be CFO enabled");
	}

	return $self;
};

=head2 B<::me>


Returns the name of the <filer> passed to CFO::new.

=over

=item Examples

   $pair->me;

Will return the filer object for "ME", the filer used with the CFO::new method.

=back

=cut 

sub me { return (shift)->{_CFO_A}; };

=head2 B<::partner>

Returns the filer object for the partner to the <filer> passed to CFO::new.

=over

=item Examples

	$pair->partner

Returns the name of "ME"'s partner.

=back 

=cut 

sub partner { return (shift)->{_CFO_B}; };

=head2 B<::takeover>

Have "ME" or "PARTNER" takeover "PARTNER" or "ME".
By default "ME" takes over "PARTNER".
ME => <non zero value> causes "PARTNER" to takeover "ME".  

=over

=item Arguments

B<ME> => I<boolean>

I<Optional.> If B<ME> is not specified or is set to zero (0) "ME" will takeover "PARTNER".  
Otherwise, "PARTNER" will takeover "ME".

B<background> => I<boolean>

I<Optional.> If B<background> is specified and set to non-zero, the takeover request is run 
in the background and the method will return immediately.

=item Examples

   $pair->takeover(ME => 0, background => 1);

is the same as

   $pair->takeover(background => 1);

Returns immediately after filer "ME" is told to initiate a takeover of "PARTNER".

   $pair->takeover(ME=>1);

Returns after "PARTNER" has completed a takeover of "ME".

=back

=cut 

sub takeover {
	my $self = shift;
	my (%args) = Params::Validate::validate (@_, {
		ME         => {type=>BOOLEAN, optional=>1},
		background => {type=>BOOLEAN, optional=>1}
	});
	if ($self->{_TAKEN}) {
		Carp::croak ("CFO::takeover " . $self->{_TAKEN}->host . " Already in takeover state");
	} else {

		$self->monitor;

		$self->{_TAKEN} = $args{ME} ? $self->{_CFO_B} :  $self->{_CFO_A} ;

		if (defined cluster::cf_takeover($self->{_TAKEN}->conn, "cf takeover")) {
			my $state = cluster::get_cf_state ($self->{_TAKEN}->conn);
			if (! exists $args{background}) {
				$self->monitor;
			}
			$state = cluster::get_cf_state ($self->{_TAKEN}->conn);
			return 1;
		} else {
			Carp::croak ("CFO::takeover Takeover failed");
		}
	}
};

=head2 B<::giveback>

Initiates a giveback by the <filer> that did the takeover.  
If neither filer is in takeover mode the giveback method will croak.

=over

=item Arguments

B<background> => I<boolean>

I<Optional.> If B<background> is defined and non-zero, giveback request is run
in the background and the method will return immediately.

B<force> => I<boolean>

I<Optional.> If B<force> is defined and non-zero, giveback will use the "-f" force option.

B<safe> => I<boolean>

I<Optional.> If B<safe> is defined and non-zero, giveback will not croak if the the 
current state of the CFO pair is NOT taken over.


=item Examples

  $pair->giveback (background => 1, force => 1);

If one of the filers has been taken over it will be told to giveback.  background
will return as soon as the giveback has been issued, otherwise giveback will not 
return until the giveback is complete.

=back

=cut 

sub giveback {
	my $self = shift;
	my (%args) = Params::Validate::validate (@_, {
		background => {type=>BOOLEAN, optional=>1},
		force      => {type=>BOOLEAN, optional=>1},
		safe       => {type=>BOOLEAN, optional=>1}
	});

	if (!$self->{_TAKEN}) {
		if ($args{safe} != 0) {
			#
			#	Calling method has no idea what the state of the CFO
			#	object is but, wants any takeover to be given back.
			#	Instead of a method to determine the state and then
			#	perhaps a call to giveback the 'safe' option is
			#	provided.  If a giveback is required it will not get
			#	here it will just go on with the giveback.  If a
			#	giveback is not requried this check insures the method
			#	will not croak and the caller is assured the desired
			#	state (UP & UP) has been achieved.
			#
			NOTE::logcomment ("NOTE::CFO giveback ( safe => 1) already given back " . $self->{_CFO_A}->host . " " . $self->{_CFO_B}->host . " ");
			return 1;
		}
		Carp::croak ("CFO::giveback failed not taken over");
	}
	my $opt = "";
	if ($args{force}) {
		$opt = {
			force => "-f"
		}
	};
	$self->monitor;
	#
	#    One filer is in TAKEOVER state
	#    Other filer is in DEAD state
	#
	NOTE::logcomment ("NOTE::CFO giveback ( safe => 1) giveback initiated " . $self->{_CFO_A}->host . " " . $self->{_CFO_B}->host . " ") if ($args{safe} !=0 );
	if (cluster::cf_giveback ($self->{_TAKEN}->conn, $opt) != 0) {
		delete $self->{_TAKEN};
		unless ($args{background}) {
			NOTE::logcomment ("NOTE::CFO giveback ( safe => 1) giveback monitoring " . $self->{_CFO_A}->host . " " . $self->{_CFO_B}->host . " ") if ($args{safe} !=0 );;
			$self->monitor;
		}
	} else {
		Carp::croak ("CFO::giveback failed");
	}
	return 1;
};

=head2 B<::status>

Returns a list of the current states of the filer pair, returns (ME, PARTNER).
Most useful when takeover or giveback were called with 'background' => 1.

=over

=item Examples

  ($me_filer_state, $partner_filer_state) = $pair->CFO::status;

=back

=cut 

sub status {
	my $self= shift;
	my $stateA = cluster::get_cf_state ($self->{_CFO_A}->conn);
	my $stateB = cluster::get_cf_state ($self->{_CFO_B}->conn);
	return ($stateA, $stateB);
};

=head2 B<::monitor>

Waits until the states of the filers are as requested in the call.
Missing states are determined by a set of default rules stated below.
Most useful when takeover or giveback are called with background.

=over

=item Arguments

B<ME> => I<state>

I<Optional.> If specified <state> is the state that monitor will wait for the filer <ME>
to be in before returning to the caller.  See Examples for the default and possible values.

B<PARTNER> => I<state>

I<Optional.> If specified <state> is the state that monitor will wait for the filer <PARTNER>
to be in before returning to the caller.  See Examples for the default and possible values.

B<TIMEOUT> => I<time>

I<Optional.> If specified the value of <time>, in seconds, is the length of time that 
the monitor method will wait for filers to achieve the desired states.  TIMEOUT defaults
to 30 seconds.

=item Examples

  ($me_state, $part_state) = $pair->CFO::monitor (ME => 'STATE-M', PARTNER => 'STATE-P');

Returns when the ME filer is in STATE-M and the PARTNER filer is in STATE-P.

  $pair->monitor;

In the above case (no arguments) arguments default to:

     ('TAKEOVER', 'DEAD') if filer ME has taken over
     ('DEAD', 'TAKEOVER') if filer PARTNER has taken over

     ('UP', 'UP') if filers have given back

Useful states are 'UP', 'TAKEOVER', 'DEAD', 'UNREACH', 'DISABLED'.

The value of missing arguments is determined by the default rules stated above.

A TIMEOUT value of 0 (zero) is the largest timeout value possible.

=back

=cut 

sub monitor {
	my $self = shift;
	my (%args) = Params::Validate::validate (@_, {
		ME      => {type => SCALAR, optional=>1},
		PARTNER => {type => SCALAR, optional=>1},
		TIMEOUT => {type => SCALAR, optional=>1}
	});
	if ($self->{_TAKEN}) {
		if ($self->{_TAKEN} == $self->{_CFO_A}) {
			if (!$args{ME})      { $args{ME}      = 'TAKEOVER'};
			if (!$args{PARTNER}) { $args{PARTNER} = 'DEAD'};
		} else {
			if (!$args{ME})      { $args{ME}      = 'DEAD'};
			if (!$args{PARTNER}) { $args{PARTNER} = 'TAKEOVER'};
		}
	} else {
		if (!$args{ME})      {$args{ME}      = 'UP'};
		if (!$args{PARTNER}) {$args{PARTNER} = 'UP'};
	}
	my ($state, $part_state);
	my $TIMEOUT = 30;
        # Fixed as a part of burt484854. Now $sub_time will be set to 1 which would result in croak if the filers don't reach the state by $TIMEOUT
	if ($args{TIMEOUT}) {
		$TIMEOUT = $args{TIMEOUT};
	}
	my $sub_time = $TIMEOUT > 0 ? 1 : 0;
	do {
		if ($self->{_TAKEN} == $self->{_CFO_A}) {
			cluster::wait_state($self->{_CFO_B}->host, $args{PARTNER}, 1);
			$part_state = $args{PARTNER};
			$state      = cluster::get_cf_state ($self->{_CFO_A}->conn);
		} else {
			cluster::wait_state($self->{_CFO_A}->host, $args{ME}, 1);
			$state = $args{ME};
			$part_state = cluster::get_cf_state ($self->{_CFO_B}->conn);
		}
		#	sub_time == 0 means no timeout will be applied
		if (($sub_time != 0) && ($TIMEOUT-- < 0)) {
			my $me      = $args{ME};
			my $partner = $args{PARTNER};
			Carp::croak("CFO::monitor timed out waiting for ME = \"$me\", APRTER = \"$partner\"\n");
		}
	} until (($state eq $args{ME}) && ($part_state eq $args{PARTNER}));
	return ($state, $part_state);
};

=head2 B<::DESTROY>

Marks the CFO object for destruction.

=over

=item Examples

  $pair->DESTROY;

=back

=cut

sub DESTROY {
	my ($self, %args) = @_;

	NOTE::logcomment("CFO:: DESTROY has been called");

	return if ( $$ != $self->{_owner} );
	return if ( $self->{_destroyed} );
	return if ( NOTE->ALL_SPARED );

	if ( $self->{_spare_me} ) {
		NOTE::logcomment("CFO::DESTROY " . $self->{_CFO_A}->host . " has been spared");
		return;
	}
	#
	#	undo any takeover that was in place
	#	(even if spared so that the filer is in a defined known state)
	#
	if ($self->{_TAKEN}) {
		NOTE::logcomment("CFO::DESTROY  \"" . $self->{_TAKEN}->host . "\" starting giveback ");
		$self->giveback(force => 1);
		$self->monitor;
		NOTE::logcomment("CFO::DESTROY  \"" . $self->{_TAKEN}->host . "\" finished giveback");
	}

	$self->{_destroy} = 1;

	NOTE::logcomment("CFO:: DESTROY has completed");
}
1;
