# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Client.pm#2 $
##
## @summary note Client class
##
## @status public
##
## @pod    here
##
## @author sydneyw@netapp.com, dl-dpgqs-automation@netapp.com
##

=head1 NAME

NOTE::Client

=head1 DESCRIPTION

Provides APIs for various actions on a client machine, such as Unix or Windows.
Provides support for streams, ACLs and LUNs.

=head1 METHODS

=cut

use POSIX qw();
package NOTE::Client;
use NOTE;

our @ISA = qw(NOTE);
use Params::Validate qw( validate validation_options SCALAR BOOLEAN HASHREF );
use strict;
use File::Basename;
use Data::Dumper;
use Carp;
use Tharn;
use dp::snapmirror::SetupCifs;
use nfs::ACL::AclNFSV4 qw( :ACE4_TYPES :ACE4_FLAGS :ACE4_MASK );
use blocks_filer;
use SayWrap;
use Options;

my $options = new Options;

$options->parameters(
	    definitions => [
		{
		    name        => 'CIFS_DOMAIN',
		    description => 'CIFS domainname. Default: LAB',
		    default     => "LAB.NETAPP.COM",
		},

		{
		    name        => 'CIFS_USER',
		    description => 'CIFS username. Default: administrator',
		    default     => "Administrator",
		},
		{
		    name        => 'CIFS_PASSWD',
		    description => 'CIFS password. Default: sundance',
		    default     => "sundance",
		},

		{
		    name        => 'CIFS_DOMAIN_ADMIN',
		    description => 'CIFS domainname admin. Default: labclient',
		    default     => "labclient",
		},
		{
		    name        => 'CIFS_DOMAIN_ADMIN_PASSWD',
		    description => 'CIFS domainname admin password. Default: sundance',
		    default     => "sundance",
		},

		{
		    name        => 'FILER_CONN',
		    description => 'Filer\'s connection type. Default is \'#rsh-socket\'',
		    default     => "#rsh-socket",
		},
		{
		    name        => 'CMDTIMEOUT',
		    description => 'Per CMD timeout. Default is 3 minutes',
		    default     => 200,
		},
		{
		    name        => 'TIMEOUT',
		    description => 'All purpose timeout. Default is 10 minutes',
		    default     => 600,
		},
	    ]
);

$options->process();
my %param = %{ $options->get_hash() };

=head2 ::new

Given a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
or a
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object, returns a client object that uses it as to root for all operations.

=over

=item arguments

B<name> => I<string>

Host name of client. Default: localhost

B<storage> => I<filer storage object>

A NOTE::Volume or NOTE::Qtree object when the client will create 'objects' on.

=item Examples

=back

    my $client = NOTE::Client->new(storage=>$qt_obj);

    my $cifs = NOTE::Client->new(name=>"win_client", storage=>$qt_obj);

=cut

use TestEnv::MinWin qw(:all);

sub _gx_storage {
    my ( $self, $storage ) = @_;

    $storage ||= $self->{_storage};
    return (ref $storage) =~ /^NOTE::GX::/;
}

sub new {
    my ($class, %args) = @_;
    my $self = {};

    croak "storage must be a NOTE::Volume or NOTE:Qtree object, Object is a ".(ref $args{storage})
	unless (ref $args{storage}) =~ /^NOTE::.*(Volume|Qtree)/;

    $self->{_storage} = $args{storage};
    $self->{_args}{client} = $args{name} || "localhost";
    NOTE::logcomment("Using Client: ".$self->{_args}{client});
    my $is_gx = _gx_storage($class, $args{storage});
    if ( host($self->{_args}{client})->hosttype =~ /windows/i ) {
	NOTE::logcomment("Client: CIFS start setup");
	if ( $is_gx ) {
		# XXX Assumed manual cifs setup so create and map share only
        	$self->{_cifs} = "have_gx_cifs"; # Dummy cifs entry
		$self->{_drive} = $self->{_storage}->map_share(client=>$self->{_args}{client});
	} else {
		$self->{_storage}->NOTE::Qtree::security("mixed");
		$self->{_cifs} = dp::snapmirror::SetupCifs->new(
					filer=>$self->{_storage}->Filer->hostspec,
					client=>$self->{_args}{client}, params=>\%param);
		$self->{_drive} = $self->{_cifs}->create_and_map_share(qtree=>$self->{_storage});
	}
    } else {
	NOTE::logcomment("Client: Remount storage on client ".$self->{_args}{client});
	my %opts;
	# Filer / Node options
	if ( $is_gx ) {
		$opts{'security-style'} = 'mixed';
		$opts{'unix-permissions'} = '---rwxrwxrwx';
	} else {
		$opts{exportfs} = "rw,nosuid,anon=0";
	}
	# nfs side options
	$opts{chmod} = "777";

	$self->{_storage}->umount if $self->{_storage}->mounted;
	$self->{_storage}->mount(client=>$self->{_args}{client}, %opts);
    }

    $self->{_owner} = $$;
    bless $self, $class;
    NOTE::reference($self);

    return $self;
}

=head2 ::mountpoint

If Unix client then returns the mountpoint. For a windows client returns the drive letter

=over

=item Arguments

None.

=item Example

      $client->mountpoint;

=back

=cut

sub mountpoint {
    my ($self) = @_;

    return ($self->{_cifs} ? $self->{_drive} : $self->{_storage}->mountpoint);
}

=head2 ::fullpath

Given a path will return the full path from the root of the 'storage' tree.

=over

=item Arguments

B<rel_path> = I<path>

Relative path

=item Example

      my $fpath = $client->fullpath("/abc/efg");

=back

=cut

sub fullpath {
    my ($self, $rel_path) = @_;

    $rel_path = "/".$rel_path if $rel_path !~ /^\//;
    my $fp = $self->mountpoint.$rel_path;
    $fp =~ s/\//\\/g
	if $self->{_cifs};
    NOTE::logcomment("fullpath: $fp");
    return $fp;
}

=head2 ::storage

Returns the storage object being used by this object. It is either a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
or a
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object.

=over

=item Argument

None.

=item Example

      my $note_obj = $client->storage

=back

=cut

sub storage { return shift->{_storage}; }

=head2 ::get_handle

Returns an object to be used by this object as a handle to the 'path' name supplied
much like 'open' returning a file handle. This handle will be used by other methods
to identify the object represented by pathname 'path'.

'path' is relative to the root of the 'storage' object used by this Client object

=over

=item Argument

B<path> = I<string>

Relative pathname to root of storage object in Client.

=item Example

      my $handle = $client->("/a_file");

=back

=cut

sub get_handle {
    my ($self, $path) = @_;

    my $handle = {};
    $handle->{path} = $path;
    bless $handle, "NOTE::Client::handle";

    $path = $self->storage->mount_path."/".$path;
    my $txt;
    my $priv_cmd = "priv set -q diag"; #jawwad changing from test to diag
    if ( $self->_gx_storage () ) {
	# XXX Replace with ngsh slist code when BURT 267245 have been implemented
	my $vol_name = $self->storage->name;
	my $uuid = $self->storage->uuid_name;;
	$path =~ s/$vol_name/$uuid/;
	$txt = $self->storage->server->dblade(-Server=>$self->storage->Aggregate->dblade_node,
				cmd=>"$priv_cmd; slist /vol$path");
    } else {
	$txt = SayWrap::sayEOF($self->storage->Filer->conn, "$priv_cmd; slist $path");
    }
    unless ($txt =~ /invalid path/i) {
	my @line = split "\n", $txt;
	my $i = 0;
	# The format of 'slist' is as follows:
	#   Inode#        File size      Name
	# -----------  ---------------   ----
	# <fileInode>  <file size>       <fullpath of file>
	# <streamInode><stream size>     <stream name> 
	# <next stream entry, etc>
	while ($i < @line) {
		if ($i > 2) { # Skip header and file entry
			$line[$i] =~ s/\s+\d+\s+\d+\s+(\S+)/$1/; # obtain just the stream name
			$handle->{stream}{$line[$i]} = 1;
		}
		$i++;
	}
    }
    return $handle;
}

=head2 ::populate

Depending on the type of the client it will generate NFS or CIFS data, using dataMaip or RishData

=over

=item Arguments

B<%args> - I<dataManip argument Hash|RichData argument Hash>

I<Optional.>

All parameters, in the form of a hash is passed to DataManip::populate_tree or RichData::populate for
processing

=item Examples

=back

    $client->populate;
    $client->populate(path=>"/subdir"); # Populate sub directory 'subdir' within storage

=cut

sub populate {
    my ($self, %args) = @_;

    eval {
	if ( host($self->{_args}{client})->hosttype =~ /windows/i ) {
    		$args{path} = $self->mountpoint.'\\\\'.$args{path};
		NOTE::logcomment("Use RichData to populate data at ".$args{path});
		RichData::populate(client=>$self->{_args}{client}, %args);
	} else {
		my $old_pwd;
		if ( $args{path} ) {
    			$old_pwd = $self->storage->pwd;
			$self->storage->cd($args{path});
			delete $args{path}
		}
		NOTE::logcomment("Populate data at ".$self->storage->pwd);
		$self->storage->populate(%args);
		$self->storage->cd($old_pwd) if $old_pwd;
	}
    };
    croak $@ if $@;
}

=head2 ::modify

Depending on the type of the client it will use dataMaip or RishData to modify data generated by
the previous call to 'populate'

=over

=item Arguments

B<%args> - I<dataManip argument hash|RichData argument Hash>

I<Optional.>

All parameters, in the form of a hash is passed to DataManip::modify_tree or RichData::modify for
processing

=item Examples

=back

    $client->modify;
    $client->modify(path=>"/subdir"); # Modify sub directory 'subdir' within storage

=cut

sub modify {
    my ($self, %args) = @_;

    eval {
	if ( host($self->{_args}{client})->hosttype =~ /windows/i ) {
    		$args{path} = $self->mountpoint.'\\\\'.$args{path};
		NOTE::logcomment("Use RichData to modify data at ".$args{path});
		RichData::modify(path=>$self->mountpoint."\\\\", client=>$self->{_args}{client}, %args);
	} else {
		my $old_pwd;
		if ( $args{path} ) {
    			$old_pwd = $self->storage->pwd;
			$self->storage->cd($args{path});
			delete $args{path}
		}
		NOTE::logcomment("Modify data at ".$self->storage->pwd);
		$self->storage->modify(%args);
		$self->storage->cd($old_pwd) if $old_pwd;
	}
    };
    croak $@ if $@;
}

=head2 ::write_file

Writes data to file given by '$handle'. Wrapper to
L<C<FileManip::write_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_file>

Croaks if failed.

=over

=item Arguments

B<handle> = I<file handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>
method

B<args> = I<args>

Hash of arguments to determine what and how it is written. See
L<C<FileManip::write_file>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_file>
for details.

=item Example

     $client->write_file($handle, contents => "some data");

=back

=cut

sub write_file {
    my ($self, $handle, %args) = @_;

    croak "<client handle> is required"
	unless $handle;

    $args{file} = $self->fullpath($handle->{path});
    $args{client} = $self->{_args}{client};
    if ( FileManip::write_file(%args) ) {
	return $handle;
    } else {
	croak "Failed to write data to file $handle->{path}";
    }
    croak "Error detected while writing file $handle->{path}: $@" if $@;
}

=head2 ::add_stream

Add a stream file to an object indicated by '$handle'. Wrapper to
L<C<FileManip::add_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/add_stream>

Returns a modified handle for the object. Croaks if failed.

=over

=item Argument

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stream_name> = <string>

Name of the stream file to create

=item Example

      my $handle = $client->get_handle("/a_file_with_stream");
      $handle = $client->add_stream($handle, "stream_file");

=back

=cut

sub add_stream {
    my ($self, $handle, $stream_name) = @_;

    my %info;
    croak "stream $stream_name already exists for file ".$handle->{path}
	if exists $handle->{stream}{$stream_name};

    return $self->_write_stream($handle, $stream_name, 0, undef, 0);
}

=head2 ::write_stream

Writes to the named stream. Returns handle or croaks if detected error. Wrapper to
L<C<FileManip::write_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/write_stream>

=over

=item Arguements

B<handle> = <object handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stram_name = <string>

Name of the stream to be written to.

B<offset> = <integer>

Offset to start writing, starting from 0. Default 0

B<size> = <integer>

Number of bytes to write. Default 1

=item Example

      $handle = $client->write_stream ($handle, "stream_file", 4, 5);

=back

=cut

sub write_stream {
    my ($self, $handle, $stream_name, $offset, $size) = @_;

    $offset = $offset || 0;
    $size = $size || 1;
    return $self->_write_stream($handle, $stream_name, 0, $offset, $size);
}
    
sub _write_stream {
    my ($self, $handle, $stream_name, $append, $offset, $size) = @_;

    my $data = "a" x $size;
    my %opts;
    $opts{file} = $self->fullpath($handle->{path});
    $opts{stream} = $stream_name;
    $opts{append} = $append;
    $opts{offset} = $offset if defined $offset;
    $opts{contents} = $data;
    FileManip::write_stream(%opts, %{$self->{_args}});
    $handle->{stream}{$stream_name} = 1;

    return $handle;
}

=head2 ::delete_stream

Removed stream(s) from '$handle' object and returns the modified '$handle' object. Wrapper to
L<C<FileManip::remove_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/remove_stream>

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stream_name> = <string>

If supplied named the stream to be remove, otherwise removes all streams from $handle

=item Example

     $handle = $client->delete_stream($handle, 'a_stream_to_be_remove');
     $handle = $client->delete_stream($handle);

=back

=cut

sub delete_stream {
    my ($self, $handle, $stream_name) = @_;

    my @list;
    if ( defined $stream_name ) {
	croak "No stream $stream_name found for file ".$handle->{path}
		unless defined $handle->{stream}{$stream_name};
	push @list, $stream_name;
    } else {
	push @list, keys %{$handle->{stream}};
    }

    foreach my $stream (@list) {
	eval {
	    	FileManip::remove_stream(file=>$self->fullpath($handle->{path}),
				 stream=>$stream, %{$self->{_args}});
	};
	croak $@ if $@;

	delete $handle->{stream}{$stream};
    }
    return $handle;
}

=head2 :rename_stream

Renames a named stream. Wrapper for
L<C<FileManip::rename_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/rename_stream>
Returns the modified '$handle' object or croaks.

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<old_stream> = <string>

stream to rename

B<new_stream> = <string>

new name of stream

=item Example

      $handle = $cleint->rename_stream($handle, 'this_stream', 'new_stream_name');

=back

=cut

sub rename_stream {
    my ($self, $handle, $old_stream, $new_stream) = @_;

    # XXX This supposedly only works for unix-solaris m/c, but it does not work
    croak "stream $old_stream does not exists for file ".$handle->{path}
	unless  exists $handle->{stream}{$old_stream};
    croak "new stream $new_stream already exists for file ".$handle->{path}
	if exists $handle->{steam}{$new_stream};

    eval {
	FileManip::rename_stream(file=>$self->fullpath($handle->{path}),
			stream=>$old_stream, nstream=>$new_stream);
    };
    croak $@ if $@;

    $handle->{stream}{$new_stream} = 1;
    return $handle;
}

=head2 ::set_stream_size

Sets the size of a stream. Uses
L<C<FileManip::setattr_stream>|engineering::ontap_tests::agnostic::docs::lib-FileManip-pm/setattr_stream>
to set the stream size

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<stream_name> = I<string>

Stream to set the size to

B<size> = <integer>

Size to set the stream to.

=item Example

      $handle = $client->set_stream_size($handle, "stream_to_change", 4096);

=back

=cut

sub set_stream_size {
    my ($self, $handle, $stream_name, $size) = @_; 

    croak "stream $stream_name does not exists for file ".$handle->{path}
	unless  exists $handle->{stream}{$stream_name};

    eval {
	FileManip::setattr_stream(file=>$self->fullpath($handle->{path}),
			stream=>$stream_name, size=>$size, %{$self->{_args}});
    };
    croak $@ if $@;

    return $handle;
}

=head2 ::create_with_acl

Creates a file/directory with pre-determined ACLs. Wrapper for
L<C<nfs::GenericAcl::create_with_acl>|engineering::ontap_tests::agnostic::docs::lib-nfs-GenericAcl-pm/create_with_acl>
Returns modified '$handle' or croaks.

This operation is carried remotely on the client

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<type> = "file|dir"

Indicates whether to create a 'file' or 'directory'/

=item Example

    $handle = $client->create_with_acl($handle, "file");

=back

=cut

sub create_with_acl {
    my ($self, $handle, $type) = @_;

    my %opts = $self->_acl_params($handle);
    $opts{filetype} = lc ($type || "file");
    croak "Invalid type: $type. Must be 'file' or 'dir'"
	unless $opts{filetype} =~ /file|dir/i;

    $handle->{acl} = $self->_run_ipc("create_with_acl", %opts);
    return $handle;
}

=head2 ::get_acl

Returns the ACL object for the '$handle' object. Acl return from
L<C<nfs::GenericAcl::get_acl>|engineering::ontap_tests::agnostic::docs::lib-nfs-GenericAcl-pm/get_acl>
or croak.

This operation is carried remotely on the client

=over

=item Argument

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

=item Example

    my $acl = $client->get_acl($handle);

=back

=cut

sub get_acl {
    my ($self, $handle) = @_;

    unless ( defined $handle->{acl} ) {
	my %opts = $self->_acl_params($handle);
        $handle->{acl} = $self->_run_ipc("get_acl", %opts);
    }
    return $handle->{acl};
}

=head2 ::set_acl

Sets the ACL with a hash containing an ACE. Wrapper for
L<C<nfs::GenericAcl::set_acl>|engineering::ontap_tests::agnostic::docs::lib-nfs-GenericAcl-pm/set_acl>.
Returns the modified '$handle' or croaks.

See
L<C<nfs::ACL::AclNFSV4>|engineering::ontap_tests::agnostic::docs::lib-nfs-ACL-AclNFSV4-pm>.
for constants to use in the <ace> hash.

This operation is carried remotely on the client

=over

=item Arguments

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

B<ace> = <hash>

Hash representing ACE entry to set the ACL to

=item Example

     $acl_file = $cifs->set_acl ($acl_file,
			type => NOTE::Client::ACE4_ACCESS_ALLOWED_ACE_TYPE,
			mask => NOTE::Client::ACE4_READ_DATA |
				NOTE::Client::ACE4_EXECUTE,
			mask_how => 'SET',
			who => 'OWNER@');

=back

=cut

sub set_acl {
    my ($self, $handle, %ace) = @_;

    $self->get_acl($handle)
	unless defined $handle->{acl};
    croak "Missing ACE"
	unless %ace;
    my %opts;
    $opts{ace} = \%ace;
    $opts{acl} = $handle->{acl};
    $handle->{acl} = $self->_run_ipc("set_acl", %opts);
    return $handle;
}

=head2 ::clear_acl

Clears an ACL from the file/dir in '$handle'. Wrapper to
L<C<nfs::GenericAcl::clear_acl>|engineering::ontap_tests::agnostic::docs::lib-nfs-GenericAcl-pm/clear_acl>.
Returns the modified '$handle' or croaks.

This operation is carried remotely on the client

=over

=item Argument

B<handle> = I<file/directory handle>

Obtained previously from 
L<C<NOTE::Client::get_handle>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Client-pm/get_handle>

=item Example

     $handle = $client->clear_acl($handle);

=back

=cut

sub clear_acl {
    my ($self, $handle) = @_;

    $self->get_acl($handle)
	unless defined $handle->{acl};
    $self->_run_ipc("clear_acl", $handle->{acl});
    delete $handle->{acl};
    return $self->get_acl($handle);
}

sub _acl_params {
    my ($self, $handle) = @_;

    my %opts;
    $opts{filename} = $self->fullpath($handle->{path});
    $opts{filer_path} = $self->{_storage}->mount_path;
    $opts{filer_vol} = _Volume($self->{_storage})->name;
    $opts{mountpoint} = $self->mountpoint;
    $opts{use_synthetic} = 1;
    return %opts;
}

sub _Volume {
    my ($obj) = @_;

    return ((ref $obj) =~ "NOTE::.*Volume" ? $obj : $obj->Volume);
}

sub _run_ipc {
    my ($self, $cmd, @args) = @_;

    unless ( $self->{_ipc} ) {
	my @extra_options;
	my $script = qq{ \{
		use Tharn;
		use nfs::ACL::AclNFSV4 qw( :ACE4_TYPES :ACE4_FLAGS :ACE4_MASK );
		use nfs::ACL::Acl qw( get_ace_array );
		use nfs::GenericAcl qw( get_acl get_acl_array create_with_acl
					set_acl set_acl_ace clear_acl );

		Subtest::ipc()->serve();

		sub result \{
		    my (\$data) = \@_;

		    my \$rc = \{\};
		    \$rc->\{error\} = \$\@;
		    \$rc->\{data\} = \$data;
		    bless \$rc, "ipc_rc";
		    return \$rc;
		\}

		sub get_fconn \{
		    my (\$conn_spec) = \@_;

		    return connect(\$conn_spec, -timeout, 60);
		\}

		sub remote_get_acl \{
			my (\$conn_spec, %args) = \@_;

			\$args\{fc\} = get_fconn(\$conn_spec);
			my \$acl;
			eval \{ \$acl = get_acl(%args); \};
			delete \$acl->\{fc\} if \$acl;
			return result(\$acl);
		\}

		sub remote_create_with_acl \{
			my (\$conn_spec, %args) = \@_;

			\$args\{fc\} = get_fconn(\$conn_spec);
			my \@file_aces = get_ace_array(type=>\$args\{filetype\}."_aces");
			\$args\{ace\} = \\\@file_aces;
			my \$acl;
			eval \{ \$acl = create_with_acl(%args); \};
			delete \$acl->\{fc\} if \$acl;
			return result(\$acl);
		\}

		sub remote_set_acl \{
			my (\$conn_spec, %args) = \@_;

			my \%ace = %\{\$args\{ace\}\};
			my \$acl = \$args\{acl\};
			\$acl->\{fc\} = get_fconn(\$conn_spec);
			eval \{ set_acl_ace (acl=>\$acl, %ace); \};
			return result() if \$\@;
			my \$rc; 
			eval \{ \$rc = set_acl (acl=>\$acl); \};
			return result() if \$\@;
			\$\@ = \$rc unless \$rc =~ /NFS4_OK/;
			delete \$acl->\{fc\};
			return result(\$acl);
		\}

		sub remote_clear_acl \{
			my (\$conn_spec, \$del_acl) = \@_;

			\$del_acl->\{fc\} = get_fconn(\$conn_spec);
			eval \{ clear_acl(acl=>\$del_acl); \};
			return result();
		\}

	\} };
	NOTE::logcomment("remote script: $script");
	$self->{_ipc} = Tharn::subtest( $script, '-bg', '-ipc',
					'-client', $self->{_args}{client},
					'-runid', "01_Client",
					'-onexit', sub {},
					@extra_options );
    }
#    my $conn_spec = $self->{_storage}->Filer->hostspec.
#			"#".$self->{_storage}->Filer->conn->conntype();
    NOTE::logcomment("Calling $cmd, with: ", Dumper(@args));
#    my $rc = $self->{_ipc}->ipc->call("remote_".$cmd, $conn_spec, @args);
    my $rc = $self->{_ipc}->ipc->call("remote_".$cmd, $self->{_storage_address}, @args);
    croak "$cmd failed. Error: ".$rc->{error} if $rc->{error};
    return $rc->{data};
}

=head2 ::lun_path

Returns the LUN path on the filer given a LUN name.

=over

=item Argument

B<lun_name> = <string>

Lun name

=item Example

     my $filer_path = $client->lun_path("a_lun");

=back

=cut

sub lun_path {
    my ($self, $lun_name) = @_;

    return $self->{_storage}->mount_path."/".$lun_name;
}

=head2 ::create_lun

Creates a LUN on the Filer. The LUN is create on the root directory of the volume or Qtree.
Wrapper for
L<C<blocks_filer::filer_lun_create>|engineering::ontap_tests::agnostic::docs::lib-blocks_filer-pm/filer_lun_create>.
Croaks if error.

=over

=item Argument

B<name> => I<string>

Required. Lun name

B<size> => I<lun size>

Optional. Lun size. Default "32M' for a windows LUN and "4m" otherwise

B<ostype> => I<lun type>

Optional. OS type of lun. Default linux

=item Example

    $client->create_lun("a_lun", "32m", "windows");

=back

=cut

sub create_lun {
    my ($self, %args) = @_;

    croak "Missing name" unless defined $args{name};
    $args{ostype} = $args{ostype} || "linux";
    $args{size} = $args{size} || ($args{ostype} =~ /windows/i ? "32m" : "4m");
    eval {
	blocks_filer::filer_lun_create($self->{_storage}->Filer->conn,
					$self->lun_path($args{name}),
					$args{size}, $args{ostype});
    };
    croak "Lun create failed: $@" if $@;
    NOTE::logcomment("Lun created. Parameter: ". Dumper(%args));
}

=head2 ::resize_lun

Resizes the specified LUN on the filer. Wrapper for
L<C<blocks_filer::filer_lun_size_change>|engineering::ontap_tests::agnostic::docs::lib-blocks_filer-pm/filer_lun_size_change>.
Croaks if failed.

=over

=item Arguments

B<name> => I<string>

Required. Name of Lun

B<size> => I<value>

Required. New size of the LUN to change to

=item Examples

      $client->resize_lun("a_lun", "8m");

=back

=cut

sub resize_lun {
    my ($self, %args) = @_;

    croak "Missing name" unless defined $args{name};
    croak "Missing size" unless defined $args{size};

    eval {
	blocks_filer::filer_lun_size_change($self->{_storage}->Filer->conn,
					$self->lun_path($args{name}),
					$args{size});
    };
    croak "Lun resize failed: $@" if $@;
    NOTE::logcomment("Lun set to $args{size}");
}

=head2 ::modify_lun

Given a LUN, with a start block number and number of blocks, will modify specific location on the LUN

=over

=item Arguments

B<name> => I<string>

Name of the LUN to modify

B<start> => I<integer>

Optional. Block to start modification. Default 0

B<count> => I<integer>

Optional. Number of blocks to modify. Default 1

=item Examples

    $client->modify_lun(name=>"a_lun"); # Modify block 0

    $client=>modify_lun(name=>"a_lun", start=>20, count=> 5); # modify 5 blocks starting from block 20

=back

=cut

sub modify_lun {
    my ($self, %args) = @_;

    $args{start} = 0 unless defined $args{start};
    $args{count} = 1 unless $args{count};

    croak "modify_lun not supported on Windows client" if $self->{_cifs};
    croak "Missing name" unless defined $args{name};
    croak "start must be >= 0" unless $args{start} >= 0;
    croak "count must be +ve" unless $args{count} > 0;

    my $blk_size = $self->{_storage}->Filer->block_size;
    
    my $rsh_prefix = "rsh ".$self->{_args}{client}." ";
    my $cln_lun_path = $self->{_storage}->mountpoint."/".$args{name};

    # Create a file of 'count' blocks in size to modify LUN with
    my $in_file = $self->{_storage}->mountpoint."/data_for_lun_".$args{name};
    
    my $script = qq { \{
    			my \$string = 1 x ($blk_size*$args{count});
			open F, ">", '$in_file' || die "open failed for input file to LUN";
			syswrite(F, \$string) || die "syswrite failed while creating input file to LUN";
			close F;
			\} };
    Tharn::subtest($script, -client, $self->{_args}{client});
    NOTE::logcomment("Modify LUN $args{name} from block $args{start} for $args{count} block(s)");
    eval {
	# Enable read / write access to LUN
	blocks_filer::filer_lun_share_all($self->{_storage}->Filer->conn,
				$self->lun_path($args{name}));
	# Use 'dd' to modify LUN with it
	my $cmd = "/usr/local/bin/sudo dd conv=notrunc "
			."seek=$args{start} count=$args{count} bs=$blk_size "
			."if=$in_file of=$cln_lun_path";
	NOTE::logcomment("LUN modify cmd: $cmd");
	system($rsh_prefix.$cmd);
	croak "dd failed, rc = $?" if $?;
    };
    system($rsh_prefix."rm $in_file");
    croak "LUN modify failed: $@" if $@;

    return 1;
}

=head2 ::move_lun

Renames a LUN within a qtree.
Wrapper for
L<C<blocks_filer::filer_lun_size_change>|engineering::ontap_tests::agnostic::docs::lib-blocks_filer-pm/filer_lun_move>.
Croaks if error.

=over

=item Argument

B<from> => I<string>

Required. Name of LUN to move.

B<to> => I<string>

Required. New name to move LUN to.

=Item Example

    $client->create_lun("old_lun", ostype=>'unix');
    $client->move_lun(from=>'old_lun', to=>'new_lun');

=back

=cut

sub move_lun {
    my ($self, %args) = @_;

    croak "Missing name of LUN to move..."
	unless defined $args{from};
    croak "Missing name to move LUN to..."
	unless defined $args{to};

    my $title = "Lun move from ".$args{from}." to ".$args{to};
    eval {
	blocks_filer::filer_lun_move($self->{_storage}->Filer->conn,
				$self->lun_path($args{from}),
				$self->lun_path($args{to}));
    };
    croak $title." failed: $@" if $@;
    NOTE::logcomment($title." succeed");
}

=head2 ::destroy_lun

Removes a LUN.
Wrapper for
L<C<blocks_filer::filer_lun_size_change>|engineering::ontap_tests::agnostic::docs::lib-blocks_filer-pm/filer_lun_destroy>.
Croaks if error.

=over

=item Argument

B<name> => I<string>

Required. Name of LUN to remove.

=Item Example

    $client->remove_lun(name=>'a_lun');

=back

=cut

sub destroy_lun {
    my ($self, %args) = @_;

    croak "Missing LUN name..."
	unless defined $args{name};

    my $title = "Destroy Lun ".$args{name};
    eval {
	blocks_filer::filer_lun_destroy($self->{_storage}->Filer->conn, $self->lun_path($args{name}));
    };
    croak $title." failed: $@" if $@;
    NOTE::logcomment($title." succeed");
}

sub SPARE {
    my ($self) = @_;

    $self->{_spare_me} = 1;
}

sub destroy {
    my ($self) = @_;

    $self->{_destroy} = 1;

    $self->{_ipc}->ipc()->quit() if defined $self->{_ipc};
    if ( defined $self->{_cifs} ) {
	if ( $self->_gx_storage () ) {
		$self->{_storage}->delete_share();
	} else {
		$self->{_cifs}->teardown ();
	}
    }
}

sub DESTROY {
    my ($self) = @_;

    return if $$ != $self->{_owner}
	   or $self->{_spare_me}
	   or $self->{_destroy}
	   or NOTE->ALL_SPARED;

    $self->destroy;
}

=head2 ::dd

Executes the unix dd command to create/modify a particular file.
If you are looking to create this file on a Volume or Qtree or even a Lun,you
are probably looking for the dd methods in those packages and not this method.

=over

=item Arguments

B<output_file> = <file_name>

I<Required> Absolute path name of the file to be created or modified. This must
be the absolute path of file with respect to the client.
For example to create file file1 on Volume src, the absolute path on client 
would be <mountpath of volume src>/file1.

B<count> = <number_of_blocks>

I<Required> Number of blocks to write

B<input_file> = <file_name>

I<Optional> Input file on client from where the blocks are read. By default,
'/dev/zero' file is taken as input file

B<seek> = <offset>

I<Optional> offset to modify the blocks. By Default, 0 is taken as the offset.

B<block_size> = <block_size>

I<Optional> Block Size. Default is 4096 bytes.

B<conv> = <keyword>

I<Optional> Conv keyword for dd command. By default, it is 'notrunc'.

=item Examples

=back
    
    $client->dd (output_file => '/vol/vol1/file1',
                      count       =>  4081,);
                      
    $client->dd (output_file => '/vol/vol1/qtree/file1',
                      count       =>  2000,
                      seek        =>  2000);                      
    
=cut

sub dd {
    my $self = shift;
    my (%args) =
      validate(@_,
               {output_file => {type => SCALAR, optional => 0, regex=>qr/^\//},
                count       => {type => SCALAR, optional => 0},
                input_file =>
                  {type => SCALAR, default => '/dev/zero', optional => 1},
                seek       => {type => SCALAR, default => 0,    optional => 1},
                block_size => {type => SCALAR, default => 4096, optional => 1},
                conv => {type => SCALAR, default => 'notrunc', optional => 1,},}
              );

    # Get the dd command
    my $cmd =
      "r dd if=$args{input_file} of=$args{output_file} bs=$args{block_size} count=$args{count} seek=$args{seek} conv=$args{conv}";
    NOTE::logcomment($cmd);
      # Execute dd command on client
    my $out = system($cmd);

    # Validate command output
    croak $args{output_file}.' not created' if ($out != 0);
} ## end sub dd

1;
