## @status   public
## @pod      here
## @summary  NOTE ESX class
## @author   chetang@netapp.com, dl-bijaya@netapp.com
=head1 NAME

NOTE::ESX

=head1 DESCRIPTION

This is the NOTE class for working with ESX.

=head1 METHODS

=cut


package NOTE::ESX;

use strict;
use warnings;

use San::San;
use San::SanFiler;
use San::SanClient;
use San::SanFiler7M;
use San::SanFiler::IGroup;
use San::SanUtil;
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::ESXDev;
use San::SanClient::DevMgr::ESXDevMgr;

use Carp qw(croak);
use Params::Validate qw[validate OBJECT SCALAR ARRAYREF];
use Data::Dumper;

use NOTE::LUN;
use NOTE::Volume;

use vars qw( @EXPORT @ISA );

=head2 ::new

Creates an object of ESX.pm

=over
=item Arguments

B<host> => I<string>
I<Required.>
The host name of the ESX Client.

B<parentobj> => I<string>

I<Required.>
The object of Volume or Qtree on which LUN would be created.

=item Examples

$esx = NOTE::ESX->new(
        host => 'esx230-30',
        conn_type  => 'ssh',
        username   => 'root',
        password   => 'root123',
        parentobj  => $vol
     );

=back

=cut

sub new {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $class = shift;
    my %args = @_;
    my $self;
    %args  = validate(
        @_,
        {  host      => { type => SCALAR, optional => 0 },
           conn_type => { type => SCALAR, optional => 0 },
           username  => { type => SCALAR, optional => 0 },
           password  => { type => SCALAR, optional => 0 },
           parentobj => { type => OBJECT, optional => 0 },

        }
           );

    $self->{_owner} = $$;
    $self->{_host}          = $args{host};
    $self->{_conn_type}     = $args{conn_type};
    $self->{_username}      = $args{username};
    $self->{_password}      = $args{password};
    $self->{_parentobj}     = $args{parentobj};


    bless $self,$class;
    return $self;
}



=head2 ::create_lun

Creates a LUN of type VMWare and user defined name and size.

=over

=item Arguments

B<lun_name> => I<string>

I<Required.>
The name of the LUN.

B<lun_size> => I<string>

I<Required.>
The size of the LUN.

=item Examples

$esx->create_lun(lun_name   => 'lun1', lun_size   => '5m');

=back

=cut

sub create_lun {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my %args  = validate(
        @_,
        {  lun_name      => { type => SCALAR, optional => 0 },
           lun_size      => { type => SCALAR, optional => 0 },
        }
           );
    $self->{_lun_name} = $args{lun_name};
    $self->{_lun_size} = $args{lun_size};

    NOTE::logcomment( "Creating a Lun ".$self->{_lun_name} ." of type vmware and of size ".  $self->{_lun_size} );
    my $lun = NOTE::LUN->new(
         Parent => $self->{_parentobj},
         name   => $self->{_lun_name},
         ostype => 'vmware',
         size   => $self->{_lun_size},
         tag    => 'source_dataset'
      );

    $lun->create(
         Parent => $self->{_parentobj},
         tag    => 'created using parent object'
    );

    croak $@ if $@;
}


=head2 ::create_igoup

Creates an igroup for ESX hosts.

=over

=item Arguments

B<igroup_name> => I<string>

I<Required.>
The name of the igroup that needs to be created.

B<protocol> => I<string>

I<Required.>
The protocol with which igroup needs to be created.

=item Examples

$esx->create_igroup(
         igroup_name       => 'igroup1',
         protocol          => 'fcp');

=back

=cut

sub create_igroup {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $self = shift;

    my %args  = validate(
        @_,
        {  igroup_name   => { type => SCALAR, optional => 0 },
           protocol      => { type => SCALAR, optional => 0 },
        }
           );
    $self->{_igroup_name}    = $args{igroup_name};
    $self->{_protocol}       = $args{protocol};

    $self->{_clientobj} = new SanClient(
             -host      => $self->{_host},
             -conn_type => $self->{_conn_type},
             -user      => $self->{_username},
             -passwd    => $self->{_password} ,
             -protocol  => $self->{_protocol}
             );

    if(ref($self->{_clientobj}) !~ /SanClient/i)
    {
        croak "ERROR: SanClient object could not be created.";
    }

    NOTE::logcomment("Creating an object of SanFiler");
    $self->{_filerobj} = new SanFiler(-filer => $self->{_parentobj}->Filer->hostrec);
    if ($self->{_filerobj}->error()) {
      my $error_msg = join "\n", $self->{_filerobj}->errmsg();
      croak $error_msg;
     }

    NOTE::logcomment("Creating an igroup ".$self->{_igroup_name}. " of type vmware on SanFiler ");
    my $IgroupObj = $self->{_filerobj}->create_igroup(
              -name => $self->{_igroup_name},
              -initObjs => [$self->{_clientobj}->get_initiators()]
              );
    if ($self->{_filerobj}->error()) {
      my $error_msg = join "\n", $self->{_filerobj}->errmsg();
      croak $error_msg;
     }

}


=head2 ::map

Maps LUN to designated igroup.

=over

=item Arguments

B<igroupe> => I<string>

I<Required.>
The name of the igroup that needs to be mapped.

=item Examples

$esx->map(
         igroup       => 'igroup1'  );

=back

=cut

sub map {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args  = validate(
        @_,
        {  igroup  => { type => SCALAR, optional => 0 },
        }
           );

    $self->{_igroup}    = $args{igroup};
    my $filer  = $self->{_parentobj}->Filer->conn;
    my $mount_path = $self->{_parentobj}->mount_path."/".$self->{_lun_name};

    NOTE::logcomment("Mapping the LUN ".$mount_path. " to the igroup ".$self->{_igroup});
    eval {
           blocks_filer::filer_lun_map(
                             $filer,
                             $mount_path,
                             $self->{_igroup},
                             1);
      };

    croak $@ if $@;
}


=head2 :: get_devices

Get devices from the device manager form the ESX Client.

A device manager object would be created that looks for the luns
that are mapped to ESX Client filtering by the filer name and lun mount path regex.

Returns an array of devices.

=over

=item Arguments

B<device_type> => I<string>

I<Required.>
The device type. For example, it could be of type "vmfs"

=item Examples

$esx->get_devices(
         device_type  => 'vmfs'  );

=back

=cut

sub get_devices {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $self = shift;
    my %args  = validate(
        @_,
        {  device_type  => { type => SCALAR, optional => 0 },
        }
           );
    $self->{_device_type}    =     $args{device_type};

    my @iscsi_targets = $self->{_filerobj}->get_iscsi_targets(
                                            -state    => "enabled",
                                            -valid_ip => 1);
    if ($self->{_filerobj}->error()) {
      my $error_msg = join "\n", $self->{_filerobj}->errmsg();
      croak $error_msg;
     }

    NOTE::logcomment("Enabling iSCSI discovery on the Client");
    $self->{_clientobj}->{iscsi}->{CONFIG_ISCSI_DISCOVERY} = 1;

    # option to enable/disable subnet matching when choosing IP addresses # disabling this because your config has the initiator and target # routed over different subnets.
    $self->{_clientobj}->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} = 0;

    # configure the client with these iscsi targets

    $self->{_clientobj}->configure(-targets => \@iscsi_targets);

    # rescan HBA's on the client so new LUNs show up

    $self->{_clientobj}->discover();


    # generating the LUN mount path.
    my $mount_path = $self->{_parentobj}->mount_path."/".$self->{_lun_name};

    NOTE::logcomment("Creating a device manager object");
    my $volname = $self->{_parentobj}->name;

    my $esx_dev_mgr = $self->{_clientobj}->create_devmgr(-filers => [$self->{_parentobj}->Filer->host],
                                       -lun_path_regex_filter => qr/\/vol\/${volname}/);

    if ($self->{_clientobj}->error()) {
      my $error_msg = join "\n", $self->{_clientobj}->errmsg();
      croak $error_msg;
     }

    NOTE::logcomment("Device manager object created");
    my @vmfs_devices = $esx_dev_mgr->get_free_devices(-device_type => $self->{_device_type},
                                                  -hosttype => 'vmware',
                                                  -filers => [$self->{_parentobj}->Filer->host]);
    if ($esx_dev_mgr->error()) {
      my $error_msg = join "\n", $esx_dev_mgr->errmsg();
      croak $error_msg;
     }
    return @vmfs_devices;

}

=head2 ::create_datastore

Create datastores on ESX host of user defined specifications.
Returns an array of datastores.

=over

=item Arguments

B<vmfs_devices> => I<string>

I<Required.>
The reference to an array containing list of vmfs devices.

B<vmfs_type> => I<string>

I<Required.>
The type of Virtual Machine File System. (vmfs2 or vmf3)

B<datastore_type> => I<string>

I<Required.>
The type of datastore needed.

B<num_luns> => I<string>

I<Required.>
The number of LUNs required per datastore.

B<block_size> => I<string>

I<Required.>
The size of the block.


=item Examples

$esx->create_datastore(
         vmfs_devices => \@vmfs_devices,
         vmfs_type  => 'vmfs3',
         datastore_type  => "one-to-one",
         num_luns   =>  1,
         block_size  => "1m"
     );

=back

=cut

sub create_datastore {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args  = validate(
        @_,
        {
           vmfs_devices         => { type => ARRAYREF, optional => 0 },
           vmfs_type            => { type => SCALAR, optional => 0 },
           datastore_type       => { type => SCALAR, optional => 0 },
           num_luns_per_dstore  => { type => SCALAR, optional => 0 },
           datastore_name       => { type => SCALAR, optional => 0 },
           block_size           => { type => SCALAR, optional => 0 },
        }
           );

    $self->{_vmfs_devices}        =    $args{vmfs_devices};
    $self->{_vmfs_type}           =    $args{vmfs_type};
    $self->{_datastore_type}      =    $args{datastore_type};
    $self->{_num_luns_per_dstore} =    $args{num_luns_per_dstore};
    $self->{_datastore_name}      =    $args{datastore_name};
    $self->{_block_size}          =    $args{block_size};

    NOTE::logcomment("Creating a LVM Object on host ".$self->{_host}." of type vmfs");
    $self->{_vmfs_lvm_obj}     = $self->{_clientobj}->create_lvm(-lvm => "vmfs");

    if ($self->{_clientobj}->error()) {
      my $error_msg = join "\n", $self->{_clientobj}->errmsg();
      croak $error_msg;
     }

    NOTE::logcomment("Creating datastores of vmfs type ".$self->{_vmfs_type}. " on ESX Host");
    my @datastores = $self->{_vmfs_lvm_obj}->create_datastores(-vmfs_devices => $self->{_vmfs_devices},
                                                  -vmfs_type => $self->{_vmfs_type},
                                                  -datastore_type => $self->{_datastore_type},
                                                  -num_luns_per_dstore => $self->{_num_luns_per_dstore},
                                                  -hostname => $self->{_datastore_name},
                                                  -block_size => $self->{_block_size});
    if ($self->{_vmfs_lvm_obj}->error()) {
      my $error_msg = join "\n", $self->{_vmfs_lvm_obj}->errmsg();
      croak $error_msg;
     }

    NOTE::logcomment("Datastores created");


    return @datastores;
}

=head2 ::create_vmdk

Create vmdk file on datastores.
Returns an array of vmdk files.

=over
=item Arguments

B<datastores > => I<string>

I<Required.>
The reference to an array containing list of datastores.

B<controller> => I<string>

I<Required.>
The name of the controller.

B<datastore_type> => I<string>

I<Required.>
The type of datastore needed.

B<num_vmfs_vmdks > => I<string>

I<Required.>
The number of vmfs vmdk required.

=item Examples

$esx->create_vmdk(
         datastores => \@datastores,
         controller  => 'lsilogic',
         datastore_type  => "one-to-one",
         num_vmfs_vmdks    =>  1,
     );

=back

=cut


sub create_vmdk {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $self = shift;
    my %args  = validate(
        @_,
        {
           datastores      => { type => ARRAYREF, optional => 0 },
           controller      => { type => SCALAR, optional => 0 },
           datastore_type  => { type => SCALAR, optional => 0 },
           num_vmfs_vmdks  => { type => SCALAR, optional => 0 },

        }
           );

    $self->{_datastores}       =    $args{datastores};
    $self->{_controller}       =    $args{controller};
    $self->{_datastore_type}   =    $args{datastore_type};
    $self->{_num_vmfs_vmdks}   =    $args{num_vmfs_vmdks};


    NOTE::logcomment("Creating vmdk files on ESX Host");

    my @vmdk_files = $self->{_vmfs_lvm_obj}->create_vmdks(-datastores => $self->{_datastores},
                                             -controller => $self->{_controller},
                                             -datastore_type => $self->{_datastore_type},
                                             -num_vmfs_vmdks => $self->{_num_vmfs_vmdks});
    if ($self->{_vmfs_lvm_obj}->error()) {
      my $error_msg = join "\n", $self->{_vmfs_lvm_obj}->errmsg();
      croak $error_msg;
     }

    NOTE::logcomment("vmdk files created");

    return @vmdk_files;
}

1;
