#
# $Id: //depot/prod/test/main/lib/NOTE/FastClones.pm#1 $
#
## @summary  NOTE Constant Time Sis Clone class
## @status   Review
## @pod      here
## @author   dl-dpg-nb-automation@netapp.com


=head1 NAME

NOTE::FastClones

=head1 DESCRIPTION

This package provides methods for working with Constant Time Sis Clones

=head1 METHODS

=cut

package NOTE::FastClones;
our @ISA = qw(NOTE);

use POSIX qw();
use strict;
use filer;
use Dense::Fast_Clones qw();
use NOTE;
use NOTE::Qtree;
use Tharn;
use Carp;
use Utils qw(make_conn);

=head2 B<Constant Time Sis Clone Operations>


=head2 ::start

Creates a Constant Time Sis Clone Object on specified volume

=over

=item Arguments
B<Filer> => I<host spec>

I<Required.>
The filer on which to create the volume.

B<source> => I<Source File Or Lun Path>

I<Required.>
Source Filer Or Lun Path

B<clone> => I<Clone Path>

I<Optional.>
Clone Path

B<partner> => C<0> | C<1>

I<Optional.>
if set to 1 Runs the command in the partner filer, defaults to 0.

B<options> => I<string>

I<Optional.>
[-n] - Do not use temporary snapshot for clone creation.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is 1800 seconds.

B<snapshot> => I<name>

I<Optional.>
Specify snapshot name for clone from snapshot

B<range> => I<number>

I<Optional.>
Set of triplets <src_blk>:<dest_blk>:<blk_cnt> 
Where
<src_blk>  - Starting FBN of the source block range.
<dest_blk> - Starting FBN of the destination block range.
<fbn_blk>  - Number of blocks to be cloned.

=item Examples

=back

    $sis_clone = NOTE::FastClones->start( Filer  => 'f6070-230-33',
	                                  source =>  "/vol/src/f",
					  clone  =>  "/vol/src/f_clone"           
                                         );
									
    $sis_clone = NOTE::FastClones->start(
        Filer  => 'f6070-230-33',
        source => "/vol/src/f",
        clone  => "/vol/src/f",
	snapshot => "snap1",
    );
	
	
=cut

sub start {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $class, %args ) = @_;
    my $self = {};
    my $options,

        $self->{_owner} = $$;

    if ( ref $args{Filer} eq "NOTE::Filer" ) {
        $self->{_Filer} = $args{Filer}->hostspec;    # This is NOTE object
    } elsif ( ref \$args{Filer} eq "SCALAR" ) {

        # Is already a hostspec
        $self->{_Filer} = $args{Filer};
    } else {
        croak "Invalid filer specification: " . ref \$args{Filer} . "\n";
    }
    $self->{_source}   = $args{source}   if defined $args{source};
    $self->{_clone}    = $args{clone}    if defined $args{clone};
    $self->{_partner}  = $args{partner}  if defined $args{partner};
    $self->{_options}  = $args{options}  if defined $args{options};
    $self->{_snapshot} = $args{snapshot} if defined $args{snapshot};
    $self->{_range}    = $args{range}    if defined $args{range};
    $self->{_timeout}  = $args{timeout}  if defined $args{timeout};
    $self->{_timeout}  = 1800            if !defined $self->{_timeout};

    # Bless this object
    bless $self, $class;
    NOTE::reference($self);

    my $conn = make_conn( $args{Filer}, "#rsh" );
    my $cmd;
    my $cmd_suffix;

    if ( defined( $args{options} ) ) {
        $options = " " . $args{options} . " ";
    } else {
        $options = " ";
    }

    if ( defined( $args{snapshot} ) ) {
        $options = $options . " " . '-s' . " " . $args{snapshot} . " ";
    }

# "clone" or [dest_path] need not be defined for
# clone start <src_path> [dest_path] [-n] <-r <src_blk>:<dest_blk>:<blk_cnt> ...>
# clone start <src_path> [dest_path] [ -s <snapshot_name> ] <-r <src_blk>:<dest_blk>:<blk_cnt> ...
# in all other cases check for clone/dest_path
#
    if ( defined( $args{range} ) ) {
        $cmd_suffix = $args{range};
        $options    = $options . " " . '-r' . " ";
    }

    if ( $args{partner} ) {
        $cmd
            = "partner "
            . "clone start" . " "
            . $args{source} . " "
            . $args{clone}
            . $options
            . $cmd_suffix;
    } else {
        $cmd
            = "clone start" . " "
            . $args{source} . " "
            . $args{clone}
            . $options
            . $cmd_suffix;
    } ## end else [ if ( $args{partner} )
    my ($return_text) = $self->Filer->say( "priv set -q advanced;" . $cmd,
        -timeout, $self->{_timeout} );

    if ($return_text =~ m/Clone operation started successfully. ID:\s*(\d+)/i )
    {
        logcomment("Clone Operation Started Sucessfully");
    } else {
        die("Unable to start clone operation: $return_text");
    }

    return $self;
} ## end sub start

=head2 B<Constant Time Sis Clone Commands>

=head2 ::list

Runs a 'clone list' command returns 1 is the opearation completed and returns 0 if the operation completed

=over

B<inode_num> => I<number>

I<Optional.>
Inode number of the file 

B<snapshot> => I<name>

I<Optional.>
Specify snapshot name 

B<start_fbn> => I<number>

I<Optional.>
Start fbn

B<fbn_count> => I<number>

I<Optional.>
Fbn count

=item Examples

=back

    $sis_clone->list(inode_num=>$inode_num);
	
    $sis_clone1->list(inode_num=>$inode_num,snapshot=>"special");
	
    my $return = $sis_clone2->list(start_fbn=>10,fbn_count=>30);
    if ( $return){
	logcomment("Clone operation completed");
    }
	
	
=cut

sub list {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;

    my $volume = $self->_get_volume_name_from_file_or_lun_path;
    my $conn   = $self->Filer;
    my $cmd;

# clone list <[-v <vol-name> -i <inode_number> [-s <snapshot_name>]] | [-p <path>]> [-r <start_fbn>:<fbn_count>]
    if ( exists $args{inode_num} ) {
        $cmd
            = "priv set test; "
            . "clone list "
            . "-v $volume -i args{inode_num} "
            . defined $args{snapshot} ? "-s $args{snapshot} " : " "
            . (
            ( defined $args{start_fbn} && defined $args{fbn_count} )
            ? "-r $args{start_fbn}:$args{start_fbn_count} "
            : ""
            );
    } else {
        $cmd 
            = "priv set test; " 
            . "clone list " 
            . "-p $self->{_clone} "
            . (
            ( defined $args{start_fbn} && defined $args{fbn_count} )
            ? "-r $args{start_fbn}:$args{start_fdn_count} "
            : ""
            );
    } ## end else [ if ( exists $args{inode_num...

    my $return_text = $conn->say("$cmd");
     $return_text =~s/^Warning.*systems.\n//gsm;
    if ( $return_text =~ m/No entries found!/ ) {
        logcomment("Clone operation has completed");
        return 1;
    } elsif ( $return_text
        =~ /(.*)Clone-inum\s+=\s+(\d)+\s+\(Public\),\s+start-fbn\s+=\s+(\d)+\s+blk-count\s+=\s+(\d)+\s+Parent\s+inum\s+=\s+(\d)+\s+\(P(rivate|ublic)\),\s+start-fbn\s+=\s+(\d)+(.*)/g
        )
    {
        logcomment("Clone operation is still running");
        return 0;
    } else {
        croak("Unmatched Output");
    }
} ## end sub list

=head2 B<Waiting>

=head2 ::wait

Waits for the clone operation to complete

=over

=item Arguments

B<timeout> => I<number>

I<Optional.>
Set default timeout for the object.

=item Example

=back

$sis_clone1->wait();

=cut

sub wait {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    my ( $timeout, $return );
    $self->{_timeout} = $args{timeout} if defined $args{timeout};

    $timeout = $self->{_timeout};

    $return = $self->list;
    while ( !$return && $timeout ) {
        $return  = $self->list;
        $timeout = -5;
    }

    if ( $timeout == 0 ) {
        croak "Clone did not complete within specified time";
    }
} ## end sub wait

=head2 B<Component Objects>

=head2 ::Volume

Returns the Volume object.

=over

=item Arguments

None.

=item Examples

=back

     $vol = $sis_clone->Volume;

=cut

sub Volume {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $vol_name = $self->_get_volume_name_from_file_or_lun_path;

    return NOTE::Volume->attach(
        Filer => $self->{_Filer},
        name  => $vol_name
    );
} ## end sub Volume

=head2 ::Qtree

Returns the Qtree object.

=over

=item Arguments

None.

=item Examples

=back

     $qtree = $sis_clone->Qtree;

=cut

sub Qtree {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $qtree_name = $self->_get_qtree_name_from_file_or_lun_path;

    return NOTE::Qtree->attach(
        Volume => $self->Volume,
        name   => $qtree_name
    );
} ## end sub Qtree

=head2 ::Filer

Returns the Filer object.

=over

=item Arguments

None.

=item Examples

=back

     $src = $sis_clone->Filer;

=cut

sub Filer {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return NOTE::Filer->connect( host => $self->{_Filer} );
}

=head2 B<Object Configuration>

=head2 ::timeout

Sets or returns the current default timeout for the object before being set.
A value of -1 is to wait forever.

=over

=item Arguments

B<timeout> => I<number>

I<Optional.>
Set default timeout for the object.

=item Examples

=back

    $src = $sis_clone->timeout;
    $sis_clone->timeout( timeout => 300 );
	
=cut

sub timeout {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;

    $self->{_timeout} = $args{timeout} if exists $args{timeout};
    $self->{_timeout} = 1800           if !defined $self->{_timeout};
    return $self->{_timeout};
} ## end sub timeout

=head2 B<Object Destruction>

=head2 ::destroy

Destroy the clone

=over

=item Examples

=back
    
	$sis_clone->destroy;

=cut

sub destroy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $vol_obj  = $self->Volume();
    my $hostpath = $self->{_clone};

    my $file_or_lun = $self->_get_file_or_lun_name_from_path;
    my $file        = $vol_obj->nfs->attach($file_or_lun);

    $self->Filer->say( "priv set test; ls %s",
        ( "/vol/" . ( $vol_obj->name ) ) );
    NOTE::logcomment("Destroying Sis Clone $self->{_clone}");
    $file->unlink();

    $self->Filer->say( "priv set test; ls %s",
        ( "/vol/" . ( $vol_obj->name ) ) );

} ## end sub destroy

# Object destructor
sub DESTROY {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    # Only the creator can destroy the object
    return if ( $$ != $self->{_owner} );
    return if ( $self->_destroyed );       # Already Clone destroyed
    return if ( NOTE->ALL_SPARED );

    # Destroy object
    $self->destroy;
} ## end sub DESTROY

# Private functions
sub _get_volume_name_from_file_or_lun_path {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my @split_components = split( /\//, $self->{_source} );

    # Return Volume name
    return $split_components[2];
} ## end sub _get_volume_name_from_file_or_lun_path

sub _get_qtree_name_from_file_or_lun_path {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my @split_components = split( /\//, $self->{_source} );

    # Return Volume name
    return $split_components[3];
}    ##

sub _get_file_or_lun_name_from_path {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    if ( $self->{_clone} =~ m/^(\/vol\/\w+)(\/.*)/ ) {
        return "$2";
    } else {
        croak "Wrong source path";
    }
}    ##

sub _destroyed {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    if ( filer_file_exists( $self->Filer->conn, $self->{_clone} ) ) {
        return 0;
    } else {
        return 1;
    }

}    ##

=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm/>

=head1 STATUS

Review

$Id: //depot/prod/test/main/lib/NOTE/FastClones.pm#1 $

$Date: 2010/09/18 $

=head1 AUTHOR

bkumaran@netapp.com, dl-dpg-nb-automation@netapp.com

=cut

1;
