#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Fiji.pm#1 $
#
## @summary  NOTE Fiji module
## @author   tmichael@netapp.com
## @status   Public
## @pod      here

=head1 NAME

NOTE::Fiji

=head1 DESCRIPTION

This module contains methods for working with NetApp FIJI's.

In order to be used, a FIJI must first be registered. Registering
a FIJI makes it possible to create FIJI objects (using C<::new>)

FIJIs would typically be pre-registered by this module, but NOTE
scripts may also register their own FIJI's.

FIJI objects are associated with a particular filer, and support
methods such as C<::set>, C<::nvset>, C<::clear>, C<::active>,
C<::wait>, and C<::SPARE>.

During the teardown process, FIJI objects will be destroyed like
any other object; even if they are in NVRAM, they will be cleared
unless spared with C<::SPARE> or by using the global spare feature
(setting environment variable NOTE_SPARE to 1)

See C<NOTE::Filer::fiji_set> and C<NOTE::Filer::fiji_nvset> for
methods that will create a new fiji object and also set it active
in one method.

=head1 METHODS

=cut

use NOTE;
package NOTE::Fiji;
our @ISA = (NOTE);

use strict;
use Carp;
my %_FIJI_table = ();

#
# Pre-registered FIJI's
#
#-------------------------

NOTE::Fiji->register(
	quiet => 1, # Don't log the registration
	tag   => 'Simulate NVLOG out of space',
	prog  => '(rsm.nvlog.write.fail.no.space(eq @vol %s) (return 1))',
	param => 'volname',
);

use constant QSM_FIJI_SRC_START_OF_NEG 		=> 1;
use constant QSM_FIJI_SRC_DURING_NEG_2		=> 2;
use constant QSM_FIJI_SRC_DURING_NEG_3		=> 3;
use constant QSM_FIJI_SRC_DURING_NEG_5		=> 5;
use constant QSM_FIJI_SRC_END_OF_NEG		=> 8;
use constant QSM_FIJI_SRC_DURING_PASS_1 	=> 49;
use constant QSM_FIJI_SRC_END_OF_PASS_1 	=> 50;
use constant QSM_FIJI_SRC_BEFORE_PASS_2 	=> 51;
use constant QSM_FIJI_SRC_END_OF_PASS_2 	=> 99;
use constant QSM_FIJI_SRC_BEFORE_FINAL_ACK 	=> 100;
use constant QSM_FIJI_DST_BEFORE_NEG 		=> 1001;
use constant QSM_FIJI_DST_DURING_NEG 		=> 1003;
use constant QSM_FIJI_DST_AFTER_NEG 		=> 1005;
use constant QSM_FIJI_DST_DURING_INOMAP_FLIP 	=> 1010;
use constant QSM_FIJI_DST_MV_STREAM_TO_PURGATORY=> 1019;
use constant QSM_FIJI_DST_AFTER_STREAM_DELETE	=> 1020;
use constant QSM_FIJI_DST_DURING_PASS_1 	=> 1025;
use constant QSM_FIJI_DST_PROCESS_STREAM_DIR	=> 1030;
use constant QSM_FIJI_DST_AFTER_STREAM_DIR	=> 1035;
use constant QSM_FIJI_DST_AFTER_PASS_1 		=> 1050;
use constant QSM_FIJI_DST_DURING_PASS_2 	=> 1075;
use constant QSM_FIJI_DST_BEFORE_JUMP_AHEAD 	=> 1095;
use constant QSM_FIJI_DST_DURING_JUMP_AHEAD 	=> 1097;
use constant QSM_FIJI_DST_AFTER_JUMP_AHEAD 	=> 1099;
use constant QSM_FIJI_DST_BEFORE_FINAL_ACK 	=> 1100;

use constant HISTORY_LOG_WRITE_FAIL => "history log write fail";
use constant HISTORY_LOG_BYPASS => "history log bypass";

# Physical checker FIJIs
use constant CHECKER_FAIL            => "checker testpoint 1 - Checker Fail";
use constant CHECKER_CORRUPT_DST     => "checker testpoint 2 - Corrupt Dst";
use constant CHECKER_MISMATCH_VBN    => "checker testpoint 3 - Mismatch VBN";
use constant CHECKER_CORRUPT_VOLINFO => "checker testpoint 4 - Corrupt VOLINFO";

use constant CHECKER_CORRUPT_HEADER  => "checker corrupt verifier log header";
use constant CHECKER_CORRUPT_DETAIL  => "checker corrupt verifier log detail";

# Fiji Actions:
use constant QSM_ABORT	=> "return 10001";
use constant BLOCK	=> "block";
use constant DELAY	=> "delay";
use constant ECHO	=> "echo";
use constant HALT	=> "halt";
use constant LOG	=> "log";
use constant NIL	=> "nil";
use constant PANIC	=> "panic";
use constant PANIC_LATER=> "panic_later";
use constant REBOOT	=> "reboot";
use constant REBOOT_LATER=> "reboot_later";
use constant RETURN	=> "return";
use constant SKMSGSTATUS=> "skMsgStatus";
use constant SET	=> "set";
use constant SLEEP	=> "sleep";
use constant SPIN	=> "spin";
use constant RANDOM	=> "random";

NOTE::Fiji->register(
	quiet => 1,
	tag   => 'QSM action at state',
	prog  => '(lrep.at.state (&& (eq @dstpath %s_%s) (== @state %s)) (%s))',
	param => [
		'prefix',	# 'src' or 'dst'
		'dstpath',
		'state',
		'action',
	],
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => 'Cause resync to fail by making marking of volume as a duplicate fail',
	prog  => '(idump.resync.fail.mark (eq @vol %s)(return 1))',
	param => [
		'volname',
	],
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => HISTORY_LOG_WRITE_FAIL,
	prog  => '(repeat %d (idump.operlog.full (true) (return true)))',
	param => { 'count' => 100 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => HISTORY_LOG_BYPASS,
	prog  => '(repeat %d (idump.operlog.fullbypass (true) (return true)))',
	param => { 'count' => 5 }
);

# To cause snapmirror check to simulate mismatches during check operations
NOTE::Fiji->register(
	quiet => 1,
	tag   => CHECKER_FAIL,
	prog  => '(repeat %d (idump.checker.fail ' .
	                          '(== @vbn (* 4 (/ @vbn 4))) ' .
	                          '(return true)))',
	param => { 'count' => 100 }
);
# To cause snapmirror check to simulate corruption in stream after checksum comparison
NOTE::Fiji->register(
	quiet => 1,
	tag   => CHECKER_CORRUPT_DST,
	prog  => '(repeat %d (idump.dst.corrupt.chk.data ' .
	                          '(== @vbn (* 4 (/ @vbn 4))) ' .
	                          '(return true)))',
	param => { 'count' => 100 }
);
# To cause snapmirror check to simulate corruption in metadata blocks of a transfer stream
NOTE::Fiji->register(
	quiet => 1,
	tag   => CHECKER_MISMATCH_VBN,
	prog  => '(repeat %d (idump.dst.corrupt.data ' .
	                          '(&& (== @vbn (* 4 (/ @vbn 4))) (== @meta 0)) ' .
	                          '(return true)))',
	param => { 'count' => 100 }
);

# To cause snapmirror check to simulate volinfo corruption.
NOTE::Fiji->register(
	quiet => 1,
	tag   => CHECKER_CORRUPT_VOLINFO,
	prog  => '(idump.dst.corrupt.volinfo (== @offset %d) ' .
	                          '(return true))',
	param => { 'offset' => 8 }
);

# To cause snapmirror check to corrupt verifier log header
NOTE::Fiji->register(
	quiet => 1,
	tag   => CHECKER_CORRUPT_HEADER,
	prog  => '(repeat %d (idump.checker.log.write.bad.header (true) (return 1)))',
	param => { 'count' => 2 }
);

# To cause snapmirror check to corrupt verifier log detail
NOTE::Fiji->register(
	quiet => 1,
	tag   => CHECKER_CORRUPT_DETAIL,
	prog  => '(repeat %d (idump.checker.log.write.bad.chunk (true) (return 1)))',
	param => { 'count' => 2 }
);

# Image Backup list of FIJIs:
# Prefix:
#    BACKUP  - Applies to the 'ib backup' command
#    IB      - Applies to both backup and restore
#    RESTORE - Applies to the 'ib restore -m logical' command

use constant BACKUP_FOP_RB_REPLY_FAILURE   => "ib_fop_rb_reply_failure";
use constant BACKUP_MM_MODECHECK_IFREE     => "ib_bkp_mm_modecheck_ifree";
use constant BACKUP_MM_READ_FAILURE        => "ib_bkp_mm_read_failure";
use constant BACKUP_MM_WRITE_FAILURE       => "ib_bkp_mm_write_failure";
use constant BACKUP_OM_READ_FAILURE        => "ib_bkp_om_read_failure";
use constant IB_FOP_IO_REPLY_FAILURE       => "image_backup_fop_io_reply";
use constant IB_FUNC_CALL_FAILURE          => "ib_func_call_failure";
use constant IB_MEM_ALLOC_FAILURE          => "ib_mem_alloc_failure";
use constant RESTORE_BAD_IMAGE             => "restore_bad_image"; # Used only for logical restore
use constant RESTORE_DW_CHUNK_FAILURE      => "restore_dw_chunk_failure";
use constant RESTORE_FOP_BW_REPLY_FAILURE  => "ib_fop_bw_reply_failure";
use constant RESTORE_FOP_MB_REPLY_FAILURE  => "ib_fop_mb_reply_failure";
use constant RESTORE_FOP_RCV_REPLY_FAILURE => "ib_fop_rcv_reply_failure";
use constant RESTORE_MM_NO_DIRENT          => "restore_mm_no_dir_entry";
use constant RESTORE_MM_NO_INODE           => "restore_mm_no_inode";
use constant RESTORE_MM_CHUNK_FAILURE      => "restore_mm_chunk_failure";
use constant RESTORE_OM_CHUNK_FAILURE      => "restore_om_chunk_failure";
use constant IB_SESSION_PAUSE              => "ib.session.pause";
use constant IB_SESSION_RESUME             => "ib.session.resume";
use constant IB_BACKUP_DW_FAILURE	   => "ib.backup.dw.failure";
use constant IB_BKP_VSM_INTERACTION	   => "ib.bkp.vsm.ib.interaction";

NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_BKP_VSM_INTERACTION,
	prog  => '(ib.bkp.vsm.ib.interaction(true) (return (true)))',
	param => {}
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_BACKUP_DW_FAILURE,
	prog  => '(ib.backup.dw.failure (> @count %s) (return true))',
	param => { count => 1000 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => BACKUP_MM_MODECHECK_IFREE,
	prog  => '(ib.bkp.mm.modecheck.ifree (true) (return 5))',
	param => {}
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_MM_NO_INODE,
	prog  => '(ib.rst.mm.no.inode (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_MM_NO_DIRENT,
	prog  => '(ib.rst.mm.no.dirent (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_BAD_IMAGE,
	prog  => '(ib.rst.bad.image (true) (return true))',
	param => {}
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_MM_CHUNK_FAILURE,
	prog  => '(ib.rst.mm.chunk.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_DW_CHUNK_FAILURE,
	prog  => '(ib.rst.dw.chunk.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_OM_CHUNK_FAILURE,
	prog  => '(ib.rst.om.chunk.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_FOP_IO_REPLY_FAILURE,
	prog  => '(ib.fop.io.reply.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_FOP_BW_REPLY_FAILURE,
	prog  => '(ib.fop.bw.reply.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_FOP_RCV_REPLY_FAILURE,
	prog  => '(ib.fop.rcv.reply.failure (&& (> @count %d) (eq @tag %s)) (return true))',
	param => [
			{ count => 1 }
		,	{ 'type' => 'new_file' }
		]
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => BACKUP_FOP_RB_REPLY_FAILURE,
	prog  => '(ib.fop.rb.reply.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => RESTORE_FOP_MB_REPLY_FAILURE,
	prog  => '(ib.fop.mb.reply.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

# Operations for IB_MEM_ALLOC_FAILURE Fiji below:
#	Restore Only
use constant FOP_NEWACL      => "fop_newacl";
use constant FOP_WMDP        => "fop_wmdP";
use constant DENSE_ARG       => "dense_arg";
#	Backup only
use constant BKP_MAIN        => "bkp_main";
#	Both Backup and Restore
use constant MGR_NEW_SESSION => "mgr_new_session";
#   Applies to 'ib status'
use constant MGR_LIST_ALL    => "mgr_list_all";
use constant MGR_LIST_ONE    => "mgr_list_one";

NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_MEM_ALLOC_FAILURE,
	prog  => '(ib.mem.alloc.failure (eq @tag %s) (return true))',
	param => { 'op' => MGR_NEW_SESSION }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => BACKUP_MM_READ_FAILURE,
	prog  => '(ib.bkp.mm.read.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => BACKUP_MM_WRITE_FAILURE,
	prog  => '(ib.bkp.mm.write.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => BACKUP_OM_READ_FAILURE,
	prog  => '(ib.bkp.om.read.failure (> @count %d) (return true))',
	param => { 'count' => 1 }
);

# Tags for IB_FUNC_CALL_FAILURE Fiji below:
use constant IB_DW_INIT             => "ib_dw_init";
use constant BKP_WRITE_IMAGE_HEADER => "ib_bkp_write_image_header";

NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_FUNC_CALL_FAILURE,
	prog  => '(ib.func.call.failure (eq @tag %s) (return true))',
	param => { 'func' => IB_DW_INIT }
);

use constant BKP_MSGLOOP   => "bkp_msgloop";
use constant RST_MMPREPROC => "rst_mmpreproc";
use constant RST_DIRWALK   => "rst_dirwalk";
use constant POPULATE_DATA => "populate_data";
NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_SESSION_PAUSE,
	prog  => '(ib.session.pause (&& (== @job_id %d) (eq @tag %s)) (return true))',
	param => [
			{ 'job_id' => 1 }
		,	{ 'pause_at'    => BKP_MSGLOOP }
		]
);

NOTE::Fiji->register(
	quiet => 1,
	tag   => IB_SESSION_RESUME,
	prog  => '(ib.session.resume (== @job_id %d) (return true))',
	param => { 'job_id' => 1 }
);

	 ### This list should grow..

#-------------------------


=head2 ::register

Register a FIJI

=over

=item Arguments

B<tag> => I<string>

I<Required.>
A descriptive name of the fiji

B<prog> => I<string>

I<Required.>
The actual fiji program, with sprintf-like format descriptors where arguments
should be substituted.

B<param> => I<param name>

I<Optional.>
If value of param is a scalar, it is taken to mean that the FIJI program
has a single, non-optional parameter, and the value of param is its name.
For example: param => 'volname'.

If the value of param is a hash with a single element, it is taken to mean
that the FIJI program has a single parameter, but it is optional.
The name of the parameter is the hash key and the default value is the
hash value. For example: param => { action => 'return 1' }.
This hash may contain only one element.

If the value of param is an array, the FIJI may have more than one parameter.
Each element of the array may be either a scalar or a hash with a single
element (as described above). The order of the elements of the array
specify the order of the parameter substitution.

=item Examples

=back

	 NOTE::Fiji->register(
		 tag   => 'Simulate nvlog out of space',
		 prog  => '(rsm.nvlog.write.fail.no.space(eq @vol %s) (return 1))',
		 param => 'volname',
	 );

	 NOTE::Fiji->register(
		 tag   => 'Simulate disk failure on write',
		 prog  => '(raid.disk.write.done (eq @disk %s) (return buf_Errno E_DISK_FAILED))',
		 param => { 'diskid' => 'v0.00' },
	 );

	 NOTE::Fiji->register(
		 tag    => 'My generic FIJI',
		 prog   => '(%s (%s) (%s))',
		 param => [
		     'name',
		     { 'condition' => '1' },
		     { 'action'    => 'nil' },
		 ],
	 );


=cut

sub register {
	my ($self,%args) = @_;
	my $tag = lc $args{tag};

	croak "Registering FIJI requires tag" unless $tag;
	croak "Registering FIJI requires FIJI program" unless $args{prog};

	if ( exists $_FIJI_table{$tag} ) {
		NOTE::logcomment( "Overriding previously registered FIJI: \"$args{tag}\"" );
	} else {
		NOTE::logcomment( "Registering FIJI: \"$args{tag}\"" )
		    unless( $args{quiet} );
	}
	delete $args{quiet};
	$_FIJI_table{$tag} = \%args;
	
	return $self;
}

sub _assemble {
	my ($self,$ftag,%args) = @_;
	my $tag = lc $ftag;
	my @arguments = ();

	croak "No registered FIJI with tag \"$ftag\"" unless ( exists $_FIJI_table{$tag} );
	my $ftemp = $_FIJI_table{$tag};

	if ( ! ref $ftemp->{param} ) {
		# single parameter, non-optional, no default
		croak $ftemp->{param}." requires a value" unless exists $args{$ftemp->{param}};
		push @arguments, $args{$ftemp->{param}};
	} elsif ( ref $ftemp->{param} eq 'HASH' ) {
		croak "Too many parameters in argument hash for FIJI \"$ftag\""
		    if ( scalar %{$ftemp->{param}} > 1 );
		# single parameter, optional with default
		my ($key,$value) = %{$ftemp->{param}};
		if ( exists $args{$key} ) {
		    push @arguments, $args{$key};
		} else {
		    push @arguments, $value;
		}
	} elsif ( ref $ftemp->{param} eq 'ARRAY' ) {
		my $params = $ftemp->{param};
		for my $ptemp ( @{$params} ) {
		    if ( ! ref $ptemp ) {
			# parameter, non-optional, no default
			croak $ptemp." requires a value" unless exists $args{$ptemp};
			push @arguments, $args{$ptemp};
		    } elsif ( ref $ptemp eq 'HASH' ) {
			# parameter, optional with default
			croak "Too many parameters in argument hash for FIJI \"$ftag\""
			    if ( scalar %{$ptemp} > 1 );
			# single parameter, optional with default
			my ($key,$value) = %{$ptemp};
			if ( exists $args{$key} ) {
			    push @arguments, $args{$key};
			} else {
			    push @arguments, $value;
			}
		    } else {
			croak "Bad type for $ptemp";
		    }
		}
	} else {
		# invalid
		croak "Bad type for ".$ftemp->{param};
	}
	my $fiji_string = sprintf( $ftemp->{prog}, @arguments );
	croak "Can't assemble fiji string" unless $fiji_string;
	return $fiji_string;
}

=head2 ::new

Create a FIJI object

=over

=item Arguments

B<Filer> => I<host spec> | <filer object>

I<Required.>
The filer on which this FIJI will be associated

B<tag> => I<string>

I<Required.>
A descriptive name of the FIJI (case is not significant)

Further arguments are used to assemble the FIJI. They should be
the parameters specified by the C<param> argument with the
FIJI was registered.

Note that this does NOT set the FIJI on the filer; it only
creates the object.

=item Examples

=back

	 my $fiji = NOTE::Fiji->new(
		Filer   => { host => $FILER_SRC },
		tag     => 'Simulate NVLOG out of space',
		volname => $vol->name,
	 );

=cut

sub new {
	my ($class,%args) = @_;
	my $self = {};

	$self->{_owner} = $$;

	if ( !defined($args{Filer}) ) {
		# In this case, there are no arguments (i.e. no hostname) for the
		# filer. I call the filer method anyway and let it decide to
		# handle it or croak.
		$self->{_Filer} = NOTE::Filer->connect();
	} elsif ( ref $args{Filer} eq "HASH" ) {
		$self->{_Filer} = NOTE::Filer->connect( %{$args{Filer}} );
	} elsif ( ref $args{Filer} eq "NOTE::Filer" ) {
		$self->{_Filer} = $args{Filer}; # This is already an object
	} elsif ( ref \$args{Filer} eq "SCALAR" ) {
		# Is just a hostspec
		$self->{_Filer} = NOTE::Filer->connect( host => $args{Filer} );
	} else {
		croak "Invalid filer specification: ".ref \$args{Filer}."\n";
	}
	$self->{_set} = 0;
	delete $args{Filer};
	$self->{_tag} = $args{tag};
	delete $args{tag};
	croak "Cannot create a FIJI object without a tag" unless ( $self->{_tag} );
	$self->{_prog} = NOTE::Fiji->_assemble( $self->{_tag}, %args );

	bless $self, $class;
	NOTE::reference( $self );
	return $self;
}

sub _set {
	my ($self,%args) = @_;

	$self->{_nvram} = $args{nvram};
	my $filter = $self->{_nvram} ? 'nvfilter' : 'filter';
	my $cmd = "priv set -q diag; fiji $filter \"".$self->{_prog}."\"";

	NOTE::logcomment( "Setting FIJI %s'%s':\n    %s",
		$self->{_nvram} ? "(in NVRAM) " : "",
		$self->{_tag},
		$self->{_prog}
	);
	my $text = SayWrap::sayEOF( $self->{_Filer}->conn, $cmd );
	croak "Error setting FIJI: $text" unless ( $text =~ /filter set/ );
	$self->{_set} = 1;
	if ( $text =~ /slot (\d+)/ ) {
		$self->{_slot} = $1;
	} else {
		$self->{_slot} = undef;
	}
}

=head2 ::set

Set a FIJI. 

This will activate the FIJI on filer with which it was associated
when the object was created (see ::new).

=over

=item Arguments

None.

=item Examples

=back

	 $fiji->set;

=cut
sub set {
	my ($self) = @_;

	_set( $self, nvram => 0);
}

=head2 ::nvset

Set a FIJI in NVRAM.

This will activate the FIJI on filer with which it was associated
when the object was created (see C<::new>). The FIJI will survive
filer reboots, but will be cleared when the script exits UNLESS
it is spared with C<::SPARE>.

=over

=item Arguments

None.

=item Examples

=back

	 $fiji->nvset;
	 $fiji->Filer->reboot;
	 # FIJI will still be active

=cut
sub nvset {
	my ($self) = @_;

	_set( $self, nvram => 1);
}

=head2 ::clear

Clear a FIJI. 

This will clear the FIJI whether it is in NVRAM or not.

=over

=item Arguments

None.

=item Examples

=back

	 $fiji->clear;

=cut
sub clear {
	my ($self) = @_;

	NOTE::logcomment( "Clearing FIJI %s'%s':\n    %s",
		$self->{_nvram} ? "(in NVRAM) " : "",
		$self->{_tag},
		$self->{_prog}
	);
	my $slot = $self->slot;
	if ( defined $slot ) {
		my $clear = $self->{_nvram} ? "nvclear" : "clear";
		my $cmd   = "priv set -q diag; fiji $clear $slot";
		my $text  = SayWrap::sayEOF( $self->{_Filer}->conn, $cmd );
		unless ( $self->{_nvram} ) { # nvclear doesn't print anything at all
		    croak "Error clearing FIJI: $text" unless ( $text =~ /filter cleared/ );
		};
	}
	$self->{_set} = 0;
}

=head2 ::slot

Returns the slot in the FIJI table that contains the FIJI, and undef
if the FIJI is not in the table.

=over

=item Arguments

None.

=item Examples

=back

	 my $slot = $fiji->slot;
	 if ( defined $slot ) {
	     NOTE::logcomment( "Fiji is active in slot $slot" );
	 } else {
	     NOTE::logcomment( "Fiji not active" );
	 }

=cut
sub slot {
	my ($self) = @_;
	my $expected_slot = defined $self->{_slot} ? $self->{_slot} : ""; 

	my $cmd = "priv set -q diag; fiji show $expected_slot";
	my $text = SayWrap::sayEOF( $self->{_Filer}->conn, $cmd );
	foreach ( split( /\n/, $text ) ) {
		if ( /(\d+).*:\s*\"(.*)\"/ ) {
		    my ($slot,$prog) = ($1,$2);
		    if ( $self->{_prog} eq $prog ) {
			# If we don't know exactly which slot it should be in
			# (older versions of ONTAP), but the fiji program matches,
			# return the current slot.
			# If we do know the slot, only return the slot number if
			# the program AND the slot number match
			if ( !defined $self->{_slot} || ($slot eq $expected_slot) ) {
			    return $slot
			}
		    }

		}
	}
	return undef; # not found
}

=head2 ::active

Returns true if the FIJI is active (i.e. it has not triggered yet).
Returns false if the FIJI has been cleared or has triggered.

=over

=item slot

Specifies the specific fiji instance when multiple instances of a fiji exists
on the filer.

=item Examples

=back

	 my $slot = $fiji->slot;
	 while ( $fiji->active ) {
	     NOTE::logcomment( "Fiji is still active in slot $slot" );
	     sleep 5;
	 }

=cut
sub _active {
	my ($self,$slot) = @_;

	return 0 unless ( $self->{_set} && defined $slot );
	my ($curslot,$line);
	my $cmd = "priv set -q diag; fiji dump $slot";
	my $text = SayWrap::sayEOF( $self->{_Filer}->conn, $cmd );
	for $line ( split( /\n/, $text ) ) {
		$curslot = $1 if ( $line =~ /^(\d+):/ );
		if ( ($line =~ /^------>/) && ( $curslot == $slot ) ) {
		    return 1;
		}
	}
	return 0;
};

sub active {
	my ($self,%args) = @_;

	return $self->_active( $self->slot );
};

=head2 ::wait

Returns when FIJI is no longer active (i.e. it has triggered)

=over

=item Arguments

B<interval> => I<seconds>

I<Optional.>
Specifies the poll interval. Default is 5 seconds.

=item Examples

=back

	 $fiji->wait;

	 $fiji->wait( interval => 10 );

=cut
sub wait {
	my ($self,%args) = @_;
	my $slot;
	my $interval = $args{interval} || 5; # default to 5 second poll interval

	unless ( $self->{_set} ) {
		croak "Can't wait for FIJI (".$self->{_tag}.") because it hasn't been set";
	}
	NOTE::logcomment( "Waiting for FIJI %s'%s':\n    %s",
		$self->{_nvram} ? "(in NVRAM) " : "",
		$self->{_tag},
		$self->{_prog}
	);
	my $start = time;
	while ( $self->_active( $slot ) ) {
		my $elapsed = (time - $start);
		NOTE::logcomment( "%4d seconds elapsed waiting for FIJI: '%s'", $elapsed, $self->{_tag} );
		if ( $args{timeout} && ($elapsed >= $args{timeout}) ) {
		    croak sprintf "Waiting for FIJI %s'%s':\n    %s",
				  $self->{_nvram} ? "(in NVRAM) " : "",
				  $self->{_tag},
				  $self->{_prog}
		}
		sleep $interval;
	}
	NOTE::logcomment( "FIJI %s'%s' is no longer active:\n    %s",
		$self->{_nvram} ? "(in NVRAM) " : "",
		$self->{_tag},
		$self->{_prog}
	);
}

=head2 ::SPARE

Specifies that a FIJI should NOT be cleared when the script exits.

=over

=item Examples

=back

	 $fiji->SPARE;

=cut
sub SPARE { shift->{_spare_me} = 1 }

sub DESTROY {
	my $self = shift;

	return if ( $$ != $self->{_owner} );
	return if ( $self->{_destroyed} );
	$self->{_destroyed} = 1;
	return if ( NOTE->ALL_SPARED );

	if ( $self->{_spare_me} ) {
		NOTE::logcomment( "FIJI %s'%s':\n    %s\nhas been spared.",
		    $self->{_nvram} ? "(in NVRAM) " : "",
		    $self->{_tag},
		    $self->{_prog}
		);
		return;
	}
	# $self->clear( nvram => 1 ) if ( $self->{_nvram} );
	$self->clear;
}

=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::main::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm/>

L<C<NOTE::SnapMirror>|engineering::ontap_tests::main::docs::lib-NOTE-SnapMirror-pm/>

=head1 STATUS

Public.

$Id: //depot/prod/test/Rsierranevada/lib/NOTE/Fiji.pm#1 $

$Date: 2013/01/22 $

=head1 AUTHOR

Author information is in the
L<NOTE overview|engineering::ontap_tests::main::docs::lib-NOTE-pm/::author>.

=cut

1;
