# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/FileSystem.pm#1 $
#
# The purpose of this package is to implement a 'generic' filesystem
# layer -- that is one that only maintains the structure of
# the filesystem without implementing any operations specific
# to any particular filesystem.
#
# It works by maintaining a lazily-built, sparse representation of the
# target filesystem (i.e. nodes are only added to the model as they are
# referenced).
#
# This package is intended for internal use. It is currently used
# by NOTE::FileSystem::NFS.
#
## @status Private
## @author tmichael@netapp.com
#
package NOTE::FileSystem;
our @ISA = qw(NOTE);
use strict;

my %FS = ();

# Return a new filesystem node
sub new {
    my ($class,%args) = @_;
    my $self = {
	_name     => $args{'name'},
	_is_dir   => $args{'is_dir'},
	_parent   => $args{'parent'},
	_children => {},
    };
    bless $self, $class;
    return $self;
}

sub _addpath {
    my ($e,@elems) = @_;
    my $name = shift @elems;

    return $e unless ( $name );
    if ( $name eq '.' ) {
	# Ignore this element
	return _addpath( $e, @elems );
    }
    if ( $name eq '..' ) {
	my $parent = $e->parent;
	# Add remaining elements starting from parent
	return _addpath( $parent, @elems );
    }
    my $child;
    unless ( $child = $e->children->{$name} ) {
	$child = NOTE::FileSystem->new(
	      name => $name,
	      is_dir => scalar @elems ? 1 : undef,
	      parent => $e
	 );
	$e->children->{$name} = $child;
    }
    return _addpath( $child, @elems );
}

sub _getroot {
    my $self = shift;

    while ($self != $self->parent) {
	$self = $self->parent;
    }
    return $self;
}

# Given a path, returns 0 if it doesn't exist. Otherwise,
# adds each element to the model (if not already present)
# and returns a reference to the element representing the
# final path element.
sub lookup {
    my ($self,$path) = @_;

    my @elems = split /\//, $path;
    shift @elems if ( $elems[0] eq '' );
    my $e = _addpath( _getroot( $self ), @elems );
    return undef unless $e;
    return $e;
}

sub delete {
    my $self = shift;
    if ( $self->is_dir ) {
	if ( scalar %{$self->children} ) {
	    NOTE::logcomment( "Warning: deleting directory %s with %d children", $self->path, %{$self->children} );
	}
    }
    delete $self->parent->children->{$self->name};
    $self->{_parent} = $self;
}

sub move {
    my ($self,$path) = @_;

    my $where = $self->lookup( $path );
    if ( $where ) {
	# move self to become child of 'where'
	# unlink from parent
	delete $self->parent->children->{$self->name};
	if ( $where->is_dir ) {
	    # make new parent
	    $self->parent( $where );
	    $self->parent->children->{$self->name} = $self;
	} else {
	    # Replaces 'where'
	    delete $where->parent->children->{$where->name};
	    $self->parent( $where->parent );
	    $self->parent->children->{$self->name} = $self;
	    $self->name( $where->name );
	}
    }
}

sub _getpath {
    my ($self,@elems) = @_;

    if ( $self != $self->parent ) {
        if ( $self->is_dir ) {
	    unshift @elems, $self->name . "/";
	} else {
	    unshift @elems, $self->name;
	}
	return _getpath( $self->parent, @elems );
    }
    return "/".join( '', @elems );
}

sub path { return _getpath( shift ) }

sub is_dir {
    my ($self,$is_dir) = @_;
    if ( defined( $is_dir ) ) {
	$self->{_is_dir} = $is_dir;
    }
    return $self->{_is_dir};
}

sub name {
    my ($self,$name) = @_;
    if ( defined( $name ) ) {
	$self->{_name} = $name;
    }
    return $self->{_name};
}

sub parent {
    my ($self,$parent) = @_;
    if ( defined( $parent ) ) {
	$self->{_parent} = $parent;
    }
    return $self->{_parent};
}

sub children { return shift->{_children} }

sub init {
    my ($self,$FSID) = @_;
    croak( "FSID not specified" ) unless ( $FSID );
    $FS{$FSID} = NOTE::FileSystem->new(
	parent => undef,
	name   => '/',
	is_dir => 1,
    );
    $FS{$FSID}->parent( $FS{$FSID} );
}

1;
