#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Filer.pm#1 $
#
## @summary  NOTE Filer class
## @status   Public
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME
 
NOTE::Filer

=head1 DESCRIPTION

This is the NOTE class for working with filers.

=head1 METHODS

=cut

package NOTE::Filer;
our @ISA = qw(NOTE);

#use strict;
use NOTE::Registry;
use Carp qw(croak);
use List::MoreUtils qw/ uniq /;

# NOTE: Do *NOT* import any external functions into this module!
# This is to avoid namespace collisions namely prototype mismatches and
# all that. Call the external function directly by prepending the package 
# name.
use Tharn qw();
use Aggregate qw();
use Volume qw();
use Admin::License qw();
use SnapMirror qw();
use SnapVault qw();
use boot qw();
use filer qw();
use Filer::Version qw();
use SayWrap qw();
use RaidInfo qw();
use Net::Ping qw();
use Date::Manip qw();
use fileSystem qw();
my %_FILER_CONNS = ();

use Params::Validate qw[validate OBJECT SCALAR ARRAYREF];

=head2 B<Basic Operations>

=head2 ::connect

Creates an object representing a filer.
Uses L<C<Tharn::connect>|engineering::nate::releases::latest::html::Tharn/connect>
to establish a connection to the filer.

=over

=item Arguments

B<host> => I<host spec>

I<Required.>
The host spec for a filer.

B<mode> => C<#rsh> | C<#telnet> | C<#console>

I<Optional.>
The connection method to use.
The default is C<#rsh-socket>.
Very rarely is this option needed.
One time when it is needed
is when a test requires a direct console connection.

B<stats> => C<0> | C<1> | C<[-r] [-n]>

I<Optional.>
Enables or disables recording/printing statistics.
If C<0> is specified, statistics will be disabled.
If C<1> is specified, statistics will be enabled and displayed as
per/second values. NFS statistics will not be displayed.
If C<-r> is specified, statistics will be enabled and printed out
as total interval values (i.e. they will reflect what happened in
the entire interval between the first connect and the last disconnect).
If C<-n> is specified, statistics will be enabled and NFS statistics
will be included.


=item Example

=back

	 $filer = NOTE::Filer->connect(
	     host => $FILER,
	     mode => '#console',
		 stats => "-r -n"
	 );

=cut

sub connect {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $class = shift;
	my %args;

	# Allow for single argument to specify hostspec:
	if ( scalar @_ == 1 ) {
		$args{host} = shift;
	} else {
		%args = @_;
	}
	my $self = {};

	$args{timeout} = $args{timeout} || 300;

	$args{host}        ||= q/$FILER/;
	$self->{_owner}    = $$;
	$self->{_debug}    = $args{debug};
	$self->{_hostspec} = NOTE::option( $args{host} ) or
		croak "Host specification '".$args{host}."' doesn't resolve to anything";
	$self->{_hostrec}  = Tharn::host($self->{_hostspec});

	$self->{_defaultip}  = $self->{_hostrec}->default_ip();
	$self->{_conn}     = Tharn::connect($args{mode} || "#rsh-socket")
		or croak "Can't connect to ".$self->{_hostspec};

	# If there already was an object with this same connection object, return
	# a reference to it, rather than creating a new one. This is useful
	# because when this object gets destroyed, we know it is the last reference,
	# so we can print out status before actually disconnecting.
	return $_FILER_CONNS{"$self->{_conn}"} if ( $_FILER_CONNS{"$self->{_conn}"} );

	NOTE::logcomment( "Connected to filer using hostspec ".$self->{_hostspec} );
	$self->{_stats}    = $args{stats};
	$self->{_host}     = $self->{_hostrec}->hostname();
	$self->{_do_zero_spares} = 0;
	$self->{_license_policy} = {};
	$self->{_root_snapshots} = NOTE::newName();	# suffix for _pretest_ and _posttest_ root snapshots
	
	bless $self, $class;

	my %reg = ();
	tie %reg, 'NOTE::Registry', $self;
	$self->{_registry} = \%reg;
	$self->{_Filer} = $self;
	tie %{$self}, 'NOTE::Registry::Slice', $self, undef;
	NOTE::reference( $self );

	$_FILER_CONNS{"$self->{_conn}"} = $self;
	
	$self->_create_pretest_snapshot if( $NOTE::_SAVE_LOGS );

	$self->_memstat_begin if ( $NOTE::_MEM_STATS );
	$self->_status_begin if ( $NOTE::USAGE_ANALYSIS );
	return $self;
}

=head2 ::reconnect

Reconnect to host with different connection method. Default method is '#rsh'.
Convinient way to switch from RSH to CONSOLE and back without need to create
extra instance of filer's object.
This method effects '_hostrec' and '_conn' private variables.

=over

=item Arguments

B<mode> => C<#rsh> | C<#telnet> | C<#console>

I<Optional.>
The connection method to use.
The default is C<#rsh>.

=item Examples

=back

	$filer->connect( host=>'fsr-u60', mode=>'#rsh' );
	... do something ...
	# switch connection to console
	$filer->reconnect( mode=>'#console' );
	# bring network down
	filer::filer_cli_say( $filer->conn, "priv set -q diag; ifconfig eth0 down" );
	... do something ...
	# bring network up
	filer::filer_cli_say( $filer->conn, "priv set -q diag; ifconfig eth0 up" );
	# switch connection back to rsh
	$filer->reconnect();

=cut

sub reconnect {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%args) = @_;

	$self->{_hostrec}  = Tharn::host($self->{_hostspec});
	$self->{_conn}     = Tharn::connect($args{mode} || "#rsh")
		or croak "Can't connect to ".$self->{_hostspec};
	return $self;
}

=head2 ::ping

Ping filer on default ip. Uses Perl 'Net::Ping->new->ping' method. All arguments
are passed to Net::Ping as is.

=cut

sub ping {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 Net::Ping->new('external')->ping( shift->conn->hostp->default_ip, @_ ) }

sub _create_pretest_snapshot {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;
	my $LOGDIR = Tharn::param('LOGDIR','-default',undef);

	return unless ( $NOTE::_SAVE_LOGS );
	return if ( ! $LOGDIR && $NOTE::_SAVE_LOGS eq '1' );

	if( $NOTE::_SAVE_LOGS || $LOGDIR ) {
		my $snap_name = '_pretest_' . $self->{_root_snapshots};
		my $root = $self->vol_root;
		my $vl = NOTE::Volume->attach( Filer => $self, name => $root );
		$vl->snapshot_delete($snap_name) if( $vl->snapshot_exists($snap_name) );
		$vl->snapshot_create($snap_name);
	}
}

sub _get_memstats {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($filer,$keypattern) = @_;
	my @stats = grep { /\s*\d+\s+\d+\s+[\*]*\s*/ } split( /\n/,
		filer::filer_cli_say( $filer->conn, "priv set -q diag; mem_stats"
	) );
	return map { my @f = split; my $k = pop @f; $k => [@f[0..1]]} grep {/$keypattern/} @stats;
}

sub _compare_stats {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($filer_name,$before,$after) = @_;
	my %diffs = ();
	my $k;
	for $k ( keys %{$after} ) {
		if ( ($before->{$k}->[0] < $after->{$k}->[0]) ||
		     ($before->{$k}->[1] < $after->{$k}->[1]) ) {
		    $diffs{$k} = { before => $before->{$k}, after => $after->{$k} };
		}
	}
	if ( scalar %diffs ) {
	    my $msg;
		for $k ( sort keys %diffs ) {
		    my $diff = $after->{$k}->[1] - $before->{$k}->[1];
		    $msg .= sprintf( "    %6d [%s%-5d] bytes: %s\n",
			$after->{$k}->[1],
			$diff < 0 ? "-" : "+", $diff,
			$k
		    );
		}
		NOTE::logcomment( "Memstat diffs for $filer_name:\n".$msg );
	} else {
		NOTE::logcomment( "Memstat OK (no leaks)" );
	}
}

=head2 ::memstats

Returns reference to memory statistic hash taken from 'C<priv set -q diag; mem_stats>' command

=cut

sub memstats {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;
	my %memstats = _get_memstats( $self, '.*' );
	return \%memstats;
}

=head2 ::memstat_diffs

Prints memstat differences taken before and after tests. Can be used to detect memory leaks.

=cut

sub memstat_diffs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 shift->_memstat_end }

sub _memstat_begin {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;
	my %memstats_begin = _get_memstats( $self, '.*' );
	$self->{_memstats_begin} = \%memstats_begin;
}

sub _memstat_end {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;
	my %memstats_end = _get_memstats( $self, '.*' );
	_compare_stats( $self->{_host}, $self->{_memstats_begin}, \%memstats_end );
}

sub _status_begin {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my $version = $self->version;
	my @aggrs   = $self->Aggregates;
	my @vols    = $self->Volumes;

	my $sv_lic = Admin::License::is_licensed($self->{'_conn'}, 'sv_ontap_pri')
	          || Admin::License::is_licensed($self->{'_conn'}, 'sv_ontap_sec');

	$self->{_num_aggrs}    = scalar @aggrs;
	$self->{_num_aggrs_nr} = scalar grep( ! $_->looksReusable, @aggrs );
	$self->{_num_vols}     = scalar @vols;
	$self->{_num_vols_nr}  = scalar grep( ! $_->looksReusable, @vols );
	$self->{_num_qtrees} = scalar $self->qtrees;
	$self->{_num_sms}    = scalar $self->snapmirror_destinations;
	$self->{_num_svs}    = $sv_lic ? scalar $self->snapvault_destinations : 0;
	my $disk_count = $self->_disk_count;
	$self->{_num_disks}        = $disk_count->{'total'};
	$self->{_num_disk_spares}  = $disk_count->{'spare'};
	$self->{_num_disk_zeroed}  = $disk_count->{'zeroed'};
	NOTE::logcomment(
		"HOST %s (ONTAP version %s) has:\n".
	    "     %2d disk%s; %2d %s in use; %2d %s, of which %d %s zeroed\n".
		"     %2d Aggregate%s [%d recyclable, %d non-recyclable]\n".
		"     %2d Volume%s    [%d recyclable, %d non-recyclable]\n".
	    "     %2d Qtree%s\n".
		"     %2d SnapMirror relation%s\n".
		"     %2d SnapVault relation%s",
		$self->{_hostspec}, $version,
		$self->{_num_disks}, $self->{_num_disks} == 1 ? " ":"s",
		$self->{_num_disks} - $self->{_num_disk_spares},
		($self->{_num_disks} - $self->{_num_disk_spares}) == 1 ? "is":"are",
		$self->{_num_disk_spares}, $self->{_num_disk_spares} == 1 ? "is a spare":"are spares",
		$self->{_num_disk_zeroed}, $self->{_num_disk_zeroed} == 1 ? "is":"are",
		$self->{_num_aggrs},  $self->{_num_aggrs}  == 1 ? " ":"s",
		$self->{_num_aggrs} - $self->{_num_aggrs_nr}, $self->{_num_aggrs_nr},
		$self->{_num_vols},   $self->{_num_vols}   == 1 ? " ":"s",
		$self->{_num_vols} - $self->{_num_vols_nr}, $self->{_num_vols_nr},
		$self->{_num_qtrees}, $self->{_num_qtrees} == 1 ? " ":"s",
		$self->{_num_sms},    $self->{_num_sms}    == 1 ? " ":"s",
		$self->{_num_svs},    $self->{_num_svs}    == 1 ? " ":"s" );
	if ( $self->{_stats} ) {
		filer::filer_statit_start( $self->conn );
	}
}

sub _status_end {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	if ( $self->{_stats} ) {
		if ( $self->{_stats} eq "1" ) {
		    filer::filer_cli_diag_say( $self->conn, "statit -e" );
		} else {
		    filer::filer_cli_diag_say( $self->conn, "statit -e ".$self->{_stats} );
		}
	}
	my @aggrs  = $self->Aggregates;
	my @vols   = $self->Volumes;
	my $sv_lic = Admin::License::is_licensed($self->{'_conn'}, 'sv_ontap_pri')
	          || Admin::License::is_licensed($self->{'_conn'}, 'sv_ontap_sec');

	my $num_aggrs      = scalar @aggrs;
	my $num_aggrs_nr   = scalar grep( ! $_->looksReusable, @aggrs );
	my $num_vols       = scalar @vols;
	my $num_vols_nr    = scalar grep( ! $_->looksReusable, @vols );
	my $num_qtrees     = scalar $self->qtrees;
	my $num_sms        = scalar $self->snapmirror_destinations;
	my $num_svs        = $sv_lic ? scalar $self->snapvault_destinations : 0;
	my $delta_aggrs    = $num_aggrs    - $self->{_num_aggrs};
	my $delta_aggrs_nr = $num_aggrs_nr - $self->{_num_aggrs_nr};
	my $delta_vols     = $num_vols     - $self->{_num_vols};
	my $delta_vols_nr  = $num_vols_nr  - $self->{_num_vols_nr};
	my $delta_qtrees   = $num_qtrees   - $self->{_num_qtrees};
	my $delta_sms      = $num_sms      - $self->{_num_sms};
	my $delta_svs      = $num_svs      - $self->{_num_svs};
	my $disk_count     = $self->_disk_count;
	my $delta_used     = ($disk_count->{total} - $disk_count->{spare}) -
			         ($self->{_num_disks}  - $self->{_num_disk_spares});
	my $delta_spares   = $disk_count->{spare}  - $self->{_num_disk_spares};
	my $delta_zeroed   = $disk_count->{zeroed} - $self->{_num_disk_zeroed};
	NOTE::logcomment(
		"HOST %s has:\n".
	    "     %2d disk%s; %2d(%+d) %s in use; %2d(%+d) %s, of which %d %s zeroed\n".
	    "     %2d(%+2d) Aggregate%s [%d(%+d) recyclable, %d(%+d) non-recyclable]\n".
	    "     %2d(%+2d) Volume%s    [%d(%+d) recyclable, %d(%+d) non-recyclable]\n".
	    "     %2d(%+2d) Qtree%s\n".
	    "     %2d(%+2d) SnapMirror relation%s\n".
	    "     %2d(%+2d) SnapVault relation%s",
		$self->{_hostspec},
		$self->{_num_disks}, $self->{_num_disks} == 1 ? " ":"s",
		$self->{_num_disks} - $disk_count->{spare}, $delta_used,
		($self->{_num_disks} - $disk_count->{spare}) == 1 ? "is":"are", 
		$disk_count->{spare}, $delta_spares, $disk_count->{spare} == 1 ? "is a spare":"are spares",
		$disk_count->{zeroed}, $disk_count->{zeroed} == 1 ? "is":"are",
		$num_aggrs,  $delta_aggrs,  $num_aggrs  == 1 ? " ":"s",
		$num_aggrs - $num_aggrs_nr, ($num_aggrs - $num_aggrs_nr) - ($self->{_num_aggrs} - $self->{_num_aggrs_nr}),
		$num_aggrs_nr, $delta_aggrs_nr,
		$num_vols,   $delta_vols,   $num_vols   == 1 ? " ":"s",
		$num_vols - $num_vols_nr, ($num_vols - $num_vols_nr) - ($self->{_num_vols} - $self->{_num_vols_nr}),
		$num_vols_nr,  $delta_vols_nr,
		$num_qtrees, $delta_qtrees, $num_qtrees == 1 ? " ":"s", 
		$num_sms,    $delta_sms,    $num_sms    == 1 ? " ":"s",
		$num_svs,    $delta_svs,    $num_svs    == 1 ? " ":"s" );
	if ( $self->{_panicked} ) {
		NOTE::logcomment( "     NOTE: This host may have panicked." );
		NOTE::panicked();
	}
	NOTE::consume() if (
		($delta_aggrs > 0) ||
		($delta_vols  > 0) ||
		($delta_sms   > 0) ||
		($delta_svs   > 0) ||
		($delta_used  > 0)
	);
}

sub _disk_count {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;

	my $ri = new RaidInfo($self->conn);
	my %disks = $ri->all_disks();
	my @all_disks = keys %disks;
	my %disk_count = ();

	$disk_count{total}  = scalar @all_disks;
	$disk_count{spare}  = 0;
	$disk_count{zeroed} = 0;

	foreach my $name (@all_disks) {
	    my $disk = $ri->disk($name);

	    if ( $disk->parameter("spare") eq "1" ) {
		    $disk_count{spare}++;
		    if ( $disk->parameter("fsm_state") eq "ZEROED" ) {
			$disk_count{zeroed}++;
		    }
	    }
	}
	return \%disk_count;
}

=head2 ::version

Returns the ONTAP version running on the filer.

=over

=item Arguments

None.

=item Example

=back

	 $version = $filer->version;

=cut

sub version {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	return Filer::Version::get_ontap_version( $self->hostrec );
}

=head2 ::wafl_version

Returns a numberic file system version akin to ontap version. Calls get_wafl_version method to option the wafl version of the filer

=over

=item Arguemtns

None.

=item Example

=back

	if ( $filer->wafl_version >= Filer::Version::WV_IRONCITY_MIN ) {
		# post-ironcity related logic
	} else 
		# pre-ironcity logic
	}

=cut

sub wafl_version {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	return Filer::Version::get_wafl_version($self->hostrec);
}

=head2 ::is_a_simulator

Returns boolean indicating whether filer is a simulator

=over

=item Arguments

None.

=item Example

=back

		$sim_filer = $filer->is_a_simulator;

=cut

sub is_a_simulator {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;

	my $result = $self->hostrec->hosttype =~ /(simulator|sim)/i;
		NOTE::logcomment("Host ".$self->hostrec->hostname
						.", (type ".$self->hostrec->hosttype
						.") is ".($result ? "" : "not")." a simulator");
		return $result;
}

=head2 B<Compatibility>

=head2 ::conn

Returns the Tharn connection object created
when L<C<::connect>|/::connect> was called.
This is for use with non-NOTE libraries.

=over

=item Arguments

None.

=item Example

=back

	 $conn = $filer->conn;

=cut

sub conn       {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_conn} }

=head2 ::hostrec

Returns the host rec for the filer.
This is for use with non-NOTE libraries.

=over

=item Arguments

None.

=item Example

=back

	 $hostrec = $filer->hostrec;

=cut

sub hostrec    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_hostrec} }

=head2 ::hostspec

Returns the host spec for the filer.
This is for use with non-NOTE libraries.

=over

=item Arguments

None.

=item Example

=back

	 $hostspec = $filer->hostspec;

=cut

sub hostspec   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_hostspec} }

=head2 ::defaultip

Returns the ip addr of the filer.

=over

=item Arguments

None.

=item Example

=back

	 $defaultip = $filer->defaultip;

=cut

sub defaultip   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_defaultip} }

=head2 B<Filer Attributes>

=head2 ::host

Returns the host name of the filer.

=over

=item Arguments

None.

=item Example

=back

	 $hostname = $filer->host;

=cut

sub host       {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->hostrec->hostname }

=head2 ::registry

Returns a reference to a hash of the registry for the filer.

=over

=item Arguments

None.

=item Examples

=back

	 $reg = $filer->registry;
	 $reg->{'options.snapmirror.vsm.restart_legacy_enable'} = 'off';

	 delete $filer->registry->{'options.files.snapmirror.conf.*'};

=cut

sub registry   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return shift->{_registry} }

=head2 ::tag

A tag is a text description for an object.
It is only stored in the object referring to the entity.
The purpose of the tag is to improve readability of the log.
If an object is tagged,
NOTE will include the tag in comments it logs about the object.

Returns the tag for the object.

=over

=item Arguments

I<string>

I<Optional.>
A tag to store in the object.

=item Examples

=back

	 $filer->tag( 'The source filer' );
	 printf( "This is %s\n", $filer->tag );

=cut

sub tag {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$newtag) = @_;

	if ( defined $newtag ) {
		$self->{_tag} = $newtag;
	}
	return $self->{_tag};
}

=head2 ::description

Returns a description of the filer consisting of
its filer's host spec, its connection mode and
the tag, if any.

=over

=item Arguments

None.

=item Example

=back

	 printf( "Using %s\n", $filer->description );

=cut

sub description {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my $txt;

	$txt = $self->host;
	if ( $self->{_tag} ) {
		$txt .= " (".$self->{_tag}.")";
	}
	return $txt;
}

=head2 B<Filer Entities>

=head2 ::aggregates

Returns a list of the names of aggregates extant on the filer.

=over

=item Arguments

None.

=item Example

=back

	 @aggr_names = $filer->aggregates;

=cut

sub aggregates {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return Aggregate::aggr_list( shift->{_conn} ) }

=head2 ::Aggregates

Returns a list of objects representing the aggregates extant on the filer.
Equivalent to calling
L<C<NOTE::Aggregate::attach>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Aggregate-pm/::attach>
on each name in the list returned by L<C<::aggregates>|/::aggregates>.

=over

=item Arguments

None.

=item Example

=back

	 @aggr_objs = $filer->Aggregates;

=cut

sub Aggregates    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my @objList = ();
	for my $aggr ( $self->aggregates ) {
		my $obj;
		eval {
		    $obj = NOTE::Aggregate->attach(
			Filer => $self,
			name => $aggr,
		    )
		};
		next if ( $@ );
		push @objList, $obj;
	}
	return @objList;
}

=head2 ::aggr_root

Returns the name of the root aggregate of the filer.

=over

=item Arguments

None.

=item Example

=back

	 $ag_root = $filer->aggr_root;

=cut

sub aggr_root {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 
	return Aggregate::aggr_root(shift->{_conn});
}

=head2 ::volumes

Returns a list of the names of volumes extant on the filer.

=over

=item Arguments

None.

=item Example

=back

	 @vol_names = $filer->volumes;

=cut

sub volumes    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my @list = Volume::vol_list( shift->{_conn} );
	croak $@ if $@;
	return @list;
}

=head2 ::Volumes

Returns a list of objects representing the volumes extant on the filer.
Equivalent to calling
L<C<NOTE::Volume::attach>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/::attach>
on each name in the list returned by L<C<::volumes>|/::volumes>.

=over

=item Arguments

B<spare_only> => C<0> | C<1>

I<Optional.>
If set to one, only spare recyclable volumes are included in the list.

=item Arguments

None.

=item Example

=back

	 @vol_objs = $filer->Volumes;

=cut

sub Volumes    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%args) = @_;
	my @objList = ();
	for my $vol ( $self->volumes ) {
	    next if ( $args{spare_only} && $vol !~ /^_/ );
		my $obj;
		eval {
		    $obj = NOTE::Volume->attach(
			Filer => $self,
			name => $vol,
		    )
		};
		next if ( $@ );
		push @objList, $obj;
	}
	return @objList;
}

=head2 ::qtrees

Returns a list of the names of qtrees extant on the filer.

=over

=item Arguments

None.

=item Example

=back

	 @qtree_names = $filer->qtrees;

=cut

sub qtrees {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 return QtreeCmds::qtreeList( shift->{_conn} ) }

=head2 ::Qtrees

Returns a list of objects representing the qtrees extant on the filer.
Equivalent to calling
L<C<NOTE::Qtree::attach>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/::attach>
on each name in the list returned by L<C<::qtrees>|/::qtrees>.

=over

=item Arguments

B<spare_only> => C<0> | C<1>

I<Optional.>
If set to one, only spare recyclable qtrees are included in the list.

=item Arguments

None.

=item Example

=back

	 @qtree_objs = $filer->Qtrees;

=cut

sub Qtrees    {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%args) = @_;
	my @objList = ();
	for my $qtree ( $self->qtrees ) {
	    next if ( $args{spare_only} && $qtree !~ /^_/ );
		my $obj;
		eval {
		    $obj = NOTE::Qtree->attach(
			Filer => $self,
			name => $qtree,
		    )
		};
		next if ( $@ );
		push @objList, $obj;
	}
	return @objList;
}

=head2 ::vol_root

Returns the name of the root volume of the filer.

=over

=item Arguments

None.

=item Example

=back

	 $vol_root = $filer->vol_root;

=cut

sub vol_root {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $name = Volume::vol_root(shift->{_conn});
	croak $@ if ( $@ );
	return $name;
}

=head2 ::Root

Returns an object that refers to the filer's root volume

=over

=item Arguments

None.

=item Example

=back

	 $root_vol = $filer->Root;

=cut

sub Root {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $filer = shift;
	return NOTE::Volume->attach( Filer => $filer, name => $filer->vol_root );
}

=head2 ::fiji_set, ::fiji_nvset

Creates a new FIJI object (using the FIJI tag to determine the
registered FIJI), and sets (activates) it on the filer.
The new object is returned.

The first argument must be the FIJI's tag. This must be the
same string (case insensitive) that was specified by the C<tag>
parameter when the FIJI was registered. Note this is NOT a named
parameter.

All other arguments are those required by the FIJI. These are
specified by the C<param> argument when the FIJI is registered.

C<::fiji_nvset> will also store the FIJI in NVRAM, so it will
survive filer reboots. However, the object must be spared
using C<NOTE::Fiji::SPARE> if it is to be preserved when the
script exits.

=over

=item Arguments

None.

=item Example

=back

	 my $fiji;
	 $fiji = $filer->fiji_set(
	    'Simulate nvlog out of space',
	    volname => $vol->name
	 );
	 $sm->update( background => 1 );
	 $fiji->wait;

=cut

sub fiji_set {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$tag,%args) = @_;

	my $fiji = NOTE::Fiji->new( Filer => $self, tag => $tag, %args );
	$fiji->set;
	return $fiji;
}

sub fiji_nvset {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$tag,%args) = @_;

	my $fiji = NOTE::Fiji->new( Filer => $self, tag => $tag, %args );
	$fiji->nvset;
	return $fiji;
}

=head2 B<Licensing>

=head2 ::check_license

Licenses a feature if it's not already licensed
and the license policy for that feature is not 'unlicense'.
There is no notification if the feature
can't be licensed because the license policy is 'unlicense'.

=over

=item Arguments

I<string>

I<Required.>
The name of the feature to check and license if necessary.

=item Example

=back

	 $filer->check_license( 'nfs' );

=cut

sub check_license {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$feature) = @_;

	if ( $self->{_license_policy}{$feature} ne "unlicense" ) {
		if ( $self->{_license_policy}{$feature} ne "licensed" ) {
		    Admin::License::license( $self->{'_conn'}, $feature ) or croak $@;
		    $self->{_license_policy}{$feature} = "licensed";
		}
	}
}

=head2 ::license

Licenses a feature regardless of
the license policy for that feature.

=over

=item Arguments

I<string>

I<Required.>
The name of the feature to license.

=item Example

=back

	 $filer->license( 'nfs' );

=cut

sub license {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$feature) = @_;
	Admin::License::license( $self->{'_conn'}, $feature ) or croak $@;
	$self->{_license_policy}{$feature} = "licensed";
}

=head2 ::unlicense

Unlicenses a feature.

=over

=item Arguments

I<string>

I<Required.>
The name of the feature to unlicense.

=item Example

=back

	 $filer->unlicense( 'nfs' );

=cut

sub unlicense {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$feature) = @_;
	Admin::License::unlicense( $self->{'_conn'}, $feature ) or croak $@;
	$self->{_license_policy}{$feature} = "unlicense";
}

=head2 B<Booting and State Control>

=head2 ::powercycle

Temporarily shuts off the power to the filer.
If called with no arguments
returns the filer to a normal up and running state.
The arguments control the selection
of other states to put the filer in
and the process by which states are attained.

=over

=item Arguments

I<boot state constant>

I<Optional, positional.>
The final state the filer should be left in.
The valid boot states are described in
L<C<lib::boot>|engineering::ontap_tests::agnostic::docs::lib-boot-pm>.
Must be given before any named options.

I<named options>

I<Optional.>
There are many named options which are rarely needed.
These options are described in
L<C<lib::boot::transit_to>|engineering::ontap_tests::agnostic::docs::lib-boot-pm>.

=item Examples

=back

	 $filer->powercycle;
	 $filer->powercycle( 'BOOT_MENU' );
	 $filer->powercycle( 'UP', timeout => 300 );
	 $filer->powercycle( timeout => 300 );   # Same as above; default state is UP

=cut

sub powercycle {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = shift;
	my ($to, %args) = @_;

	$to = 'UP' if ( !defined $to );

	$self->{_hostrec}->powercycle;
	$self->transit_to( $to, %args );
}

=head2 ::reboot

Reboots the filer if it is up or boots it if it was not up
and returns the filer to a normal up and running state.
The arguments control the process by which states are attained.

=over

=item Arguments

I<named options>

I<Optional.>
There are many named options which are rarely needed.
These options are described in
L<C<lib::boot::transit_to>|engineering::ontap_tests::agnostic::docs::lib-boot-pm>.

=item Examples

=back

	 $filer->reboot;
	 $filer->reboot( timeout => 300 );

=cut

sub reboot {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%args) = @_;
	boot::reboot( $self->{_hostrec}, %args ) or croak $@;
}

=head2 ::tape_ready

Determine if a tape device is ready.
Croaks if it cannot find status for tape device

=over

=item Argument

B<tape_device> => <tape device name>

=item Example

=back

	$status = $filer->tape_ready(tape_device=>"rst01");

=cut

sub tape_ready {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self, %args) = @_;

	croak "Missing tape device parameter" unless exists $args{tape_device};

	my %info = Tapeinfo::getTapeStatus($self->conn, $args{tape_device});
	croak $@ if $@;

	if ( $info{status} =~ /ready/i ) {
		return 1;
	} else {
		NOTE::logcomment("Tape device ".$args{tape_device}." not ready, status: ".$info{status});
		return 0;
	}
}

=head2 ::sync

Force a CP on the filer to sync dirty FS blocks to disk. Two 'sync' commands
will be sent when this method is invoked.

=over

=item Arguments

B<wait> => C<O> | C<1>

I<Optional.>
Waits for the CP to be completed before returning control to user.
Defaults to 1.

=item Examples

=back

	 $filer->sync;              # synchronous
	 $filer->sync( wait => 0 ); # asynchronous

=cut

sub sync {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%args) = @_;

	my $wait = ( exists($args{wait}) ) ? $args{wait} : 1;
	my $cmd = "priv set -q diag; sync; sync";

	if ( $wait ) {
		SayWrap::sayEOF( $self->conn, $cmd );
	} else {
		main::say( $self->conn, $cmd, -blind );
	}
}

=head2 ::say

Sends a command to the filer's CLI.
Returns the output from the command. This
is essentially a wrapper around 
L<C<SayWrap::sayEOF>|engineering::ontap_tests::agnostic::docs::lib-SayWrap-pm/sayEOF>

=over

=item Arguments

Takes arguments like printf.

=item Example

=back

	 $filer->say( 'snapmirror destinations' );

	 $rawout = $filer->say( 'vol status %s', $myvol->name );

=cut

sub say {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$fmt,@args) = @_;
	return unless ( $fmt );

        # Fixed as a part of burt 502722.Passing timeout value to sayEOF method.    
	my( $index )= grep { $args[$_] eq 'timeout'} 0..$#args;
	if(defined($index)) {
       	    $index = $index + 1;
	    return SayWrap::sayEOF( $self->conn, sprintf( $fmt, @args ), timeout => $args[$index] );
	} else {
	    return SayWrap::sayEOF( $self->conn, sprintf( $fmt, @args ));
	}
}

=head2 ::panic

Causes the filer to panic.
Sleeps for 320 seconds after the panic.

=over

=item Arguments

None.

=item Example

=back

	 $filer->panic;

=cut

sub panic {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	NOTE::logcomment( "Issuing panic on ".$self->host."; will sleep for 320 seconds" );
	# This does a panic, then sleeps for 320 seconds:
	filer::filer_panic( $self->conn );
}

=head2 ::transit_to

Transition the filer to a specific state.
If the filer is already in the requested state
no action is taken.
The arguments control the selection
of states to put the filer in
and the process by which states are attained.

=over

=item Arguments

I<boot state constant>

I<Required, positional.>
The final state the filer should be left in.
The valid states are described in
L<C<lib::boot>|engineering::ontap_tests::agnostic::docs::lib-boot-pm>.
Must be given before any named options.

I<named options>

I<Optional.>
There are many named options.
These options are described in
L<C<lib::boot::transit_to>|engineering::ontap_tests::agnostic::docs::lib-boot-pm>.

=item Examples

=back

	 $filer->transit_to( 'FIRMWARE' );
	 $filer->transit_to( 'UP', timeout => 300 );

=cut

sub transit_to {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my ($to, %args) = @_;

	boot::transit_to( $self->{_hostrec}, $to, %args ) or croak $@;
}

=head2 B<SnapMirror>

=head2 ::snapmirror_enable

Turns on SnapMirror on the filer.
Returns the output of the C<snapmirror on> command.

=over

=item Arguments

None.

=item Example

=back

	 $ret = $filer->snapmirror_enable;

=cut

sub snapmirror_enable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my $ret;
	eval {
	    $ret = SnapMirror::snapmirror_enable( host => $self->hostrec );
	};
	croak $@ if ( $@ );
	return $ret;
}

=head2 ::snapmirror_disable

Turns off SnapMirror on the filer.
Returns the output of the C<snapmirror off> command.

=over

=item Arguments

None.

=item Example

=back

	 $ret = $filer->snapmirror_disable;

=cut

sub snapmirror_disable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my $ret;
	eval {
	    $ret = SnapMirror::snapmirror_disable( host => $self->hostrec );
	};
	croak $@ if ( $@ );
	return $ret;
}

=head2 ::snapmirror_destinations

Returns a list of NOTE::SnapMirror objects that refer to relationships that exist on the filer.

Note that the destructor for these objects
does B<not> destroy the relationship,
although there is a C<::destroy> method
that can be used to do so.

=over

=item Arguments

None.

=item Example

=back

	 @dests = $filer->snapmirror_destinations;

=cut

sub snapmirror_destinations {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my $list;
	$self->check_license( 'snapmirror' );
	eval {
	    $list = SnapMirror::snapmirror_destinations_list( 
	        host => $self->hostrec 
	    );
	};
	croak $@ if ( $@ );
	return unless ( $list );
	# Return number of destinations if called in
	# scalar context.
	return scalar @{$list} unless ( wantarray );
	my @dests = ();
	for my $item ( @{$list} ) {
		push @dests, NOTE::SnapMirror->attach(
		    srcpath => $item->{'source-location'},
		    dstpath => $item->{'destination-location'},
		);
	}
	return @dests;
}

=head2 ::snapmirror_status

Returns a pointer to an array of hashes representing the statuses of all
snapmirror relationships on the filer

=over

=item Arguments

None.

=item Example

	$statuses = $filer->snapmirror_status;

	$is_idle = $statuses->[0]->{status} eq 'idle';

=back

=cut

sub snapmirror_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	$self->check_license( 'snapmirror' );
	my $list;
	eval {
		$list = SnapMirror::snapmirror_status_list(
				host => $self->hostrec
		);
	};
	croak $@ if $@;
	return $list;
}

=head2 B<SnapVault>

=head2 ::snapvault_enable

Turns on SnapVault on the filer.

=over

=item Arguments

None.

=item Example

=back

	 $ret = $filer->snapvault_enable;

=cut

sub snapvault_enable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;

	return SnapVault::sv_toggle( $self->conn, 1 );
}

=head2 ::snapvault_disable

Turns off SnapVault on the filer.

=over

=item Arguments

None.

=item Example

=back

	 $ret = $filer->snapvault_disable;

=cut

sub snapvault_disable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;

	return SnapVault::sv_toggle( $self->conn, 0 );
}

=head2 ::snapvault_destinations

Returns a list of NOTE::SnapVault 
objects that refer to relationships that exist on the filer.

Note that the destructor for these objects does B<not> destroy the relationship, although there is a C<::destroy> method that can be used to do so.

=over

=item Arguments

None.

=item Example

=back

	 @dests = $filer->snapvault_destinations;

=cut

sub snapvault_destinations {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	my ($list, @dests, $src, $dst, $item);

	# Get list of SnapVault destinations
	$list = SnapVault::snapvault_destinations( $self->conn ) or croak $@;
	return unless ( $list );

	# Return number of destinations if called in
	# scalar context.
	return scalar(keys %$list) unless ( wantarray );

	# Create a SnapVault object for each destination
	foreach $src ( keys %$list ) {
	    foreach $dst ( @{$list->{$src}} ) {
	        push @dests, NOTE::SnapVault->attach(
	            src => $self->host . ':' . $src,
	            dst => $dst,
	        );
	    }
	}

	return @dests;
}

=head2 ::snapvault_status

Returns a pointer to an array of hashes representing the statuses of all
snapvault relationships on the filer

=over

=item Arguments

None.

=item Example

	$statuses = $filer->snapvault_status;

	$is_idle = $statuses->[0]->{status} eq 'idle';

=back

=cut

sub snapvault_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	my $list;
	$list = SnapVault::snapvault_status_list($self->conn);
	croak $@ if $@;
	return $list;
}

=head2 B<VolCopy>

=head2 ::volcopy_status

Returns a pointer to an array of hashes representing the statuses of all
vol copy relationships on the filer

=over

=item Arguments

None.

=item Example

	$statuses = $filer->volcopy_status;

	$is_idle = $statuses->[0]->{status} eq 'idle';

=back

=cut

sub volcopy_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	my ($list);
	$list = Volume::Copy::volcopy_status_list($self->conn);
	croak $@ if $@;
	return [ map { {op_num => $_, progress => $list->{$_}} } keys %$list ];
}

=head2 B<AggrCopy>

=head2 ::aggrcopy_status

Returns a pointer to an array of hashes representing the statuses of all
aggr copy relationships on the filer

=over

=item Arguments

None.

=item Example

	$statuses = $filer->aggrcopy_status;

	$is_idle = $statuses->[0]->{status} eq 'idle';

=back

=cut

sub aggrcopy_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	my ($list);
	$list = Aggregate::Copy::aggrcopy_status_list($self->conn);
	croak $@ if $@;
	return [ map { {op_num => $_, progress => $list->{$_}} } keys %$list ];
}

=head2 B<Good Citizenship>

=head2 ::zero_spares

Enables or disables the scheduling of spare disk zeroing
when the filer object is destroyed
(either explicitly or upon test completion).
If called with no arguments zeroing is enabled.

Starting disk zeroing as soon as possible
can save a significant amount of time during later test runs,
and thus is recommended unless there's a reason not to.
The default setting for all filer objects
is not to zero spares,
however destroying a traditional volume or an aggregate
will enable zeroing.

=over

=item Arguments

C<0> | C<1>

I<Optional.>
A flag to turn on (C<1>) or off (C<0>)
the scheduling of spare disk zeroing.

=item Examples

=back

	 $filer->zero_spares;        # Schedule zeroing
	 $filer->zero_spares( 0 );   # Changed her mind

=cut

sub zero_spares {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$flag) = @_;
	if ( defined $flag ) {
		$self->{_do_zero_spares} = $flag;
	} else {
		$self->{_do_zero_spares} = 1;
	}
}

sub do_zero_spares {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	filer::filer_cli_say(shift->{_conn},"disk zero spares");
}

=head2 ::languages

Returns hash reference of languages used by volumes.  Relies on Volume::vol_lang call.

=over

=item Arguments

None.

=item Example

	my @langs = $filer->languages;

=back

=cut

sub languages {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	return Volume::vol_lang($self->{_conn});
}

=head2 ::rdate

Adjust real time clock on filer using remote time server

=over

=item Syntax

$filer->rdate($time_server);

=item Arguments

B<time_server> host/ip

I<Required> Host or IP address of remote time server. For example any solaris client
acts as time server by default. No changes or installation needed on Solaris box.

=back

=cut

sub rdate {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$server) = @_;

	filer::filer_cli_say( $self->{_conn}, "rdate $server" );
}

=head2 ::time

Read or set filer's real time clock

=over

=item Arguments

B<clock> => 'compliance' | 'system' | 'skew'

I<Optional> Clock type. See examples for details

B<set> => 1|0

I<Optional> Set/Read clock

=item Examples

	printf( "System Seconds since epoch: %d\n", $filer->time );

	printf( "Compliance seconds since epoch: %d\n", $filer->time( clock => 'compliance' ) );

	printf( "Clock skew: %d\n", $filer->time( clock => 'skew' ) );

	$filer->time( clock => 'compliance', set => 1 );     # initialize the compliance clock

	$filer->time( clock => 'system', set => 1 );   # set the system clock to the client's clock

	$filer->time( clock => 'system', set => $filer->time( clock => 'compliance' ) ); # set the system clock to the compliance clock

	$filer->time( clock => 'skew', set => 0 ); # set system clock to compliance clock

=back

=cut

sub time {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%opts) = @_;
	if ( !exists($opts{'clock'}) || ($opts{'clock'} eq 'system') ) {
		if ( !$opts{'set'} ) {
		    return Date::Manip::UnixDate( filer::filer_date_get( $self->{_conn} ), '%s' );
		} else {
		    if ( ref $opts{'set'} eq 'ARRAY' ) {
			# Set the clock with rdate, if an array was specified
			eval { $self->rdate( $opts{'set'}->[0] ) } ||
			eval { $self->rdate( $opts{'set'}->[1] ) };
			croak $@ if ( $@ );
		    } elsif ( $opts{'set'} eq 1 ) {
			# Just '1' means sets to localtime of client
			return filer::filer_date_set( $self->{_conn} );
		    } else {
			# Otherwise, the time was specified
		        my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($opts{'set'});
			my $ds = sprintf( "-u %4d%02d%02d%02d%02d.%02d",
				1900 + $year, $mon + 1, $mday, $hour, $min, $sec );
			return filer::filer_date_set( $self->{_conn}, $ds );
		    }
		}
	} elsif ( $opts{'clock'} eq 'compliance' ) {
		$self->check_license( 'snaplock' );
		my $epoch_secs;
		if ( !$opts{'set'} ) {
		    eval {
			$epoch_secs = snaplock::compliance_clock_status( $self->conn );
		    };
		    croak $@ if ( $@ );
		    croak "Compliance clock not initialized" unless defined $epoch_secs;
		} else {
		    eval {
			$epoch_secs = snaplock::compliance_clock_initialize( $self->conn, reinit_ok => 1, strict => 0 );
		    };
		    croak $@ if ( $@ );
		}
		return $epoch_secs;
	} elsif ( $opts{'clock'} eq 'skew' ) {
		unless ( defined $opts{'set'} ) {
		    # Returns 'undef' if compliance clock is not initialized
		    my $comp_val = eval { $self->time( clock => 'compliance' ) };
		    return undef if ( $@ );
		    return $self->time( clock => 'system' ) - $comp_val;
	    }
		# Set the skew
		return $self->time( clock => 'system', set => $self->time( clock => 'compliance' ) + $opts{'set'} );
	} else {
		croak 'Unknown clock type: '.$opts{'clock'};
	}
}

=head2 ::date

Does the same as FILER::time except that it uses and returns
formatted date strings:

=over

=item Argument

B<set> => I<date string>

=item Example

	$filer->date( set => 'Aug  4 11:23:05 PDT 2006' );

=back

=cut

sub date {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,%opts) = @_;
	if ( !exists($opts{'clock'}) || ($opts{'clock'} eq 'system') ) {
		if ( !$opts{'set'} ) {
		    return scalar localtime( $self->time( clock => 'system' ) );
		} else {
		    return $self->time( set => Date::Manip::UnixDate($opts{'set'}) );
		}
	} elsif ( $opts{'clock'} eq 'compliance' ) {
		my $compliance_time = $self->time( set => $opts{'set'}, clock => 'compliance' );
		return "Undefined" unless ( defined $compliance_time );
		return scalar localtime( $compliance_time );
	} else {
		croak 'Unknown clock type: '.$opts{'clock'};
	}
}

=head2 ::block_size

Returns BLOCK_SIZE for filer

=over

=item Arguments

None.

=item Example

	$nfs->block_size;

=back

=cut

sub block_size {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self) = @_;

	# Use defined constant until there is a need to return different filer block_size
	return fileSystem::BLOCK_SIZE;
}

=head2 ::file_punchhole

Punch a hole in a file. The hole size must be whole blocks.

=over

=item Arguments

B<path> => I<file pathname>

Path name of file

B<start_block> => I<start block number>

Starting number number. First block of file starts at 0. Defaults: 0.

B<size> => i<number of blocks>

Number of blocks to punch hole in. Default: 1.

=item Example

	# punch a hole of 3 blocks in size starting from block 2
	$filer->file_punchhole(path=>"/vol/src/holy_file.lst", start_block=>2, size=>3);

=back

=cut

sub file_punchhole {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self, %opts) = @_;

	croak ("missing <path> of file")
		unless defined $opts{path};

	my $start_blk = $opts{start_block} || 0;
	my $start_byte = $start_blk * $self->block_size;
	my $size_blk = $opts{size} || 1;
	my $size_byte = $size_blk * $self->block_size;

	NOTE::logcomment("Punching hole in file $opts{path} "
		."at block $start_blk (byte $start_byte) "
		."for $size_blk blocks (byte $size_byte)");
	# XXX This needs to be a wrapper to a new function in lib/filer.pm
	#     but was not able to change that file so adding here for now.
	my $ret = SayWrap::sayEOF($self->conn,
			"priv set test; file punchhole $opts{path} $start_byte $size_byte");
	croak "file punchhole failed: $ret" if $ret;

        #Getting file id of the file to verify whether hole punched in
        my $out1 = $self->say("priv set test; showfh $opts{path}");
        my @fields = split(" ",$out1);
        my @file_id = split("=",$fields[2]);
        my @vol_name = split("/",$opts{path});
        my $block_num = $start_byte/4096;

        #using blockinfo command to verify whether punching hole succeeded
        my $result = $self->say("priv set test; dumpblock info $vol_name[2] $file_id[1] $block_num");
        unless ($result =~ /block is a hole/i){
                croak "punching hole on a file failed:\n ";
                }
        NOTE::logcomment("Punching hole on a file succeeded\n");
        return 1;
}

=head2 ::panicked

Sets / Returns TRUE if filer did panic during test execution. There is no automatic check if
filer did panic during test itself accept at the very end of test execution. Normally used
to bring filer up and wait for cores being saved.

=over

=item Arguments

None.

=item Examples

	$filer->panicked(1);
	... some FIJI code causing panic ...

=back

=cut

sub panicked {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my ($self,$value) = @_;
	return $self->{_panicked} ? 1 : 0 unless ( defined $value );
	$self->{_panicked} = $value;
}

=head2 ::bootargs

Sets or Gets the bootargs of the filer.
If no arguments are specified, returns a hash of bootarg arg/value pairs. 
If arguments are specified, sets the bootarg to $arg{value}.
To unset a bootarg, specify its value as 'undefined', in the $arg{value} pair.

=over

=item Arguments

B<args> => hashref to arg/value pairs

I<Optional> A hashref to arg/value pairs to be set.

=item Examples

	$filer->bootargs();
	$filer->bootargs(args => $args);

=back

=cut

sub bootargs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        my ($self,%params) = @_;
        my $args = $params{args};

        if (defined $args) {
            filer::filer_bootargs_set($self->conn, $args);
           }
        else {
            my $out = filer::filer_bootargs_get($self->conn);
            return $out;
        }
}

=head2 ::bootargs_unset

Unsets the bootargs of the filer, using the 'bootargs unset' command. ie. sets them to 'undefined'

=over

=item Arguments

B<args> => arrayref of bootargs to unset

I<Optional> A arrayref of bootargs to unset.

=item Examples

	$filer->bootargs(args => $args);

=back

=cut

sub bootargs_unset {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        my ($self,%params) = @_;
        my $args = $params{args};

        filer::filer_bootargs_unset($self->conn, $args);
}

=head2 ::model

Returns the model of the filer. 

=over

=item Examples

	$filer->model();

=back

=cut

sub model {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return filer::filer_model_get($self->conn);
}

=head2 ::SPARE

Sets spare flag to 1 preventing clean up after test is complete.

=over

=item Examples

	if($error) {
		$filer->SPARE;
	}

=back

=cut

sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 shift->{_spare_me} = 1 };

sub DESTROY {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;

	return if ( $$ != $self->{_owner} );
	return if ( $self->{_destroyed} );
	$self->_status_end if ( $NOTE::USAGE_ANALYSIS );
	$self->_memstat_end if ( $NOTE::_MEM_STATS );
	$self->{_destroyed} = 1;
	unless ( $self->{_spare_me} || NOTE->ALL_SPARED ) {
		$self->do_zero_spares if ( $self->{_do_zero_spares} );
	}
	NOTE::logcomment( "Disconnecting from filer ".$self->host );
	Tharn::disconnect( $self->{_hostconn} ) if ( $self->{_hostconn} );
}

=head2 B<FCNIC OOD>

=head2 ::ood_status

To find the status of ood interface specifed.

=over

=item Arguments

B<interface> => I<string>

I<Required.>
Interface to be queried for status

=item Return Value

Reference to hash containing the status

             $VAR1 = {
                      'OOD'        => 'Enabled',
                      'Coalescing' => 'Enabled'
                     };

=item Examples

=back

    my $flr = NOTE::Filer->connect( FILER => $FILER );
    $flr->ood_status( interface => 'ql5a' );

=cut

sub ood_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { interface => { type => SCALAR, optional => 0 } } );
    my $retval;

    my $output = $self->say( 'fcnic ood status ' . $args{'interface'} );

    if ( $output =~ /not a FCVI interface/ ) {
        Carp::croak(
            "Interface " . $args{'interface'} . " is not a FCVI interface" );
    } elsif ($@) {
        Carp::croak( "Unknown Output: " . $@ );
    } else {
        my @lines = split( /\n|\r/, $output );
        $retval->{'OOD'}        = ( split( /OOD Status\s+/, $lines[2] ) )[1];
        $retval->{'Coalescing'} = ( split( /Coalescing\s+/, $lines[3] ) )[1];

        return $retval;
    }
}

=head2 ::ood_enable

To Enable OOD on the interface specifed.

=over

=item Arguments

B<interface> => I<string>

I<Required.>
Interface to be enabled

=item Return Value

croaks if failed

=item Examples

=back

    my $flr = NOTE::Filer->connect( FILER => $FILER );
    $flr->ood_enable(interface => 'ql5a');

=cut

sub ood_enable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { interface => { type => SCALAR, optional => 0 } } );

    my $status = $self->ood_status( interface => $args{'interface'} );

    if ( $status->{'OOD'} eq 'Enabled' ) {
        NOTE::logcomment(
            "OOD already enabled on the interface, no changes made!");
    } else {
        my $output = $self->say( 'fcnic ood on ' . $args{'interface'} );
        sleep(5);
        if ( $output =~ /not a FCVI interface/ ) {
            Carp::croak( "Interface "
                  . $args{'interface'}
                  . " is not a FCVI interface" );
        }

        $status = $self->ood_status( interface => $args{'interface'} );

        Carp::croak( "Failed to enable OOD on " . $args{'interface'} )
          if ( $status->{'OOD'} ne 'Enabled' );
    }
}

=head2 ::ood_disable

To disable OOD on the interface specifed.

=over

=item Arguments

B<interface> => I<string>

I<Required.>
Interface to be disabled

=item Return Value

croaks if failed

=item Examples

=back

    my $flr = NOTE::Filer->connect( FILER => $FILER );
    $flr->ood_disable(interface => 'ql5a');

=cut

sub ood_disable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { interface => { type => SCALAR, optional => 0 } } );

    my $status = $self->ood_status( interface => $args{'interface'} );

    if ( $status->{'OOD'} eq 'Disabled' ) {
        NOTE::logcomment(
            "OOD already disabled on the interface, no changes made!");
    } else {
        my $output = $self->say( 'fcnic ood off ' . $args{'interface'} );
        sleep(5);
        if ( $output =~ /not a FCVI interface/ ) {
            Carp::croak( "Interface "
                  . $args{'interface'}
                  . " is not a FCVI interface" );
        }

        $status = $self->ood_status( interface => $args{'interface'} );

        Carp::croak( "Failed to disable OOD on " . $args{'interface'} )
          if ( $status->{'OOD'} ne 'Disabled' );
    }
}

=head2 ::ood_print_stats

To print OOD statistics on the interface specifed.

=over

=item Arguments

B<interface> => I<string>

I<Required.>
Interface for which stats have to be printed

=item Return Value

Total number of OOD requests received

=item Examples

=back

    my $flr = NOTE::Filer->connect( FILER => $FILER );
    $flr->ood_print_stats(interface => 'ql5a');

=cut

sub ood_print_stats {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { interface => { type => SCALAR, optional => 0 } } );

    my $output = $self->say(
        'priv set -q diag;fcnic ood print_stats ' . $args{'interface'} );

    if ( $output =~ /not a FCVI interface/ ) {
        Carp::croak(
            "Interface " . $args{'interface'} . " is not a FCVI interface" );

    } elsif ($@) {
        Carp::croak( "Unknown Output: " . $@ );
    } else {
        my @lines = split( /\n|\r/, $output );
        my $last = @lines - 1;

        return (
            split(
                /Total number of out of order requests on receive side:/,
                $lines[$last]
            )
        )[1];
    }
}

=head2 ::ood_zero_stats

To zero OOD statistics on the interface specifed.

=over

=item Arguments

B<interface> => I<string>

I<Required.>
Interface for which stats have to be set to 0

=item Examples

=back

    my $flr = NOTE::Filer->connect( FILER => $FILER );
    $flr->ood_zero_stats(interface => 'ql5a');

=cut

sub ood_zero_stats {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { interface => { type => SCALAR, optional => 0 } } );

    my $output = $self->say(
        'priv set -q diag;fcnic ood zero_stats ' . $args{'interface'} );
    sleep(5);
    if ( $output =~ /not a FCVI interface/ ) {
        Carp::croak(
            "Interface " . $args{'interface'} . " is not a FCVI interface" );
    }

    Carp::croak( "Failed to zero OOD stats on " . $args{'interface'} )
      unless ( $self->ood_print_stats( interface => $args{'interface'} ) == 0 );
}

=head2 ::memstats_start

To start capturing memory data by issuing 'mem_stats -c' & 'mem_stats -b' commands on the filer

=over

=item Arguments

None

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->memstats_start() ;

=cut

sub memstats_start {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $error_strings =
      qr{(?im)(Invalid option|unrecognized privilege|not found)};

    my $cmdout = $self->say("priv set -q diag;mem_stats -c");
    croak("Unable to set privilege/mem_stats -c command on the filer")
      if ( $cmdout =~ $error_strings );

    $cmdout = $self->say("priv set -q diag;mem_stats -b");
    croak("Unable to set privilege/mem_stats -b command on the filer")
      if ( $cmdout =~ $error_strings );
}

=head2 ::memstats_report

To report memory usage data by issuing 'mem_stats -e' command on the filer

=over

=item Arguments

B<keylist> => I<Array Reference>

I<Optional.>
Reference to array containing heads to be monitored, if omitted all functions are returned

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->memstats_report() ;

    my @array = ("vfiler_create_etc", "split_namelist", "in_pcballoc");
    $filer_object->memstats_report( keylist => \@array ) ;

=cut

sub memstats_report {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { keylist => { type => ARRAYREF, optional => 1 } } );

    my @list;

    #If keylist is given, it should be an array reference
    ( exists $args{'keylist'} )
      ? @list = @{ $args{'keylist'} }
      : NOTE::logcomment("No keys passed, generating output for all functions");

    my $error_strings =
qr{(?im)(No saved copy of memory statistics|Invalid option|unrecognized privilege|not found)};

    my $cmdout = $self->say("priv set -q diag;mem_stats -e");
    croak("Unable to set privilege or run mem_stats -e command on the filer")
      if ( $cmdout =~ $error_strings );

    my @output_lines = split( /\n/, $cmdout );

    # If keylist was passed, find out the function names
    if ( scalar(@list) > 0 ) {
        my @return_array;
        foreach my $key (@list) {
            my $notFound = 1;

            # Check each element in the command output status
            foreach my $values (@output_lines) {
                if ( $values =~ m/(\s+)(.*$key.*)/i ) {
                    push( @return_array, $2 );
                    $notFound = 0;
                }
            }
            push( @return_array, "0\t" . $key ) if ($notFound);
        }
        return \@return_array;
    }
    return \@output_lines;
}

=head2 ::get_buftree_type

To determine the aggregate or volume type using vol_db command on the filer.

=over

=item Arguments

B<name> => I<Array Reference> or I<String>

I<Required.>
Reference to array containing names of aggregates/volumes or name of individual aggregate or volume
whose type is to be determined.

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->get_buftree_type(name => 'aggr0') ;

    my @array = ("aggr0", "vol0", "vol1");
    $filer_object->get_buftree_type( name => \@array ) ;

=cut

sub get_buftree_type {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args =
      validate( @_, { name => { type => ARRAYREF | SCALAR, optional => 0 } } );

    my @namelist;
    @namelist = @{ $args{name} } if ( ref( $args{name} ) eq 'ARRAY' );

    if ( scalar(@namelist) > 0 ) {
        my $retval;
        foreach my $key (@namelist) {
           $retval->{$key} = $self->_buftree_type($key);
        }
        return $retval;
    } else {
        return $self->_buftree_type( $args{name} );
    }
}

# pvt method for determining buftree type via handle
sub _buftree_type {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

  my ( $self, $name ) = @_;

    my $handle = $self->say("priv set -q diag;vol_db lookup name $name");
    croak("Unable to lookup the handle for $name.\n")
      if ( $handle =~ m/not found/i );

    my $output = $self->say("priv set -q diag;vol_db get buftree_type $handle");
       $output =~ s/\s+//g;

    ( $output =~ m/(buftree_type:\s*)(\w+(\(upgrading\))?)/i )
      ? return $2
      : croak("Unexpected output, $output for $name & handle $handle\n");
}

=head2 ::get_wafl_scan_speed

    Returns wafl scan speed on the filer.

=over

=item Arguments

None

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->get_wafl_scan_speed;

=cut

sub get_wafl_scan_speed{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $text = $self->say("priv set -q diag;wafl scan speed");
    unless ($text =~ /WAFL scan speed is/){
        croak "Unable to get wafl scan speed on the filer\n";
        }
    my @arr = split(" ",$text);
    return $arr[4];
    }

=head2 ::set_wafl_scan_speed

    Sets wafl scan speed on the filer.

=over

=item Arguments

B<speed> => I<String>

I<Required>
    Sets wafl speed to specified speed on the filer.

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->set_wafl_scan_speed(speed=>10);

=cut

sub set_wafl_scan_speed{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    croak "Specify speed to be set\n" if (!exists $args{speed});
    my $speed = $args{speed};
    $self->say("priv set -q diag;wafl scan speed $speed");
    if ($speed != $self->get_wafl_scan_speed){
        croak "Unable to set wafl scan speed on the filer\n";
        }
    return 1;
    }

=head2 ::printflag

    Returns the value of flag.

=over

=item Arguments

B<flag> => I<String>

I<Required.>
    Flag to be returned

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->printflag(flag=>"wafl_reclaim_threshold_s");

=cut

sub printflag{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    croak "Specify the flag\n" if (!exists $args{flag});
    my $flag = $args{flag};

    #Print the flag
    my $text = $self->say("priv set -q diag;printflag $flag");
    if ($text =~  /no such flag/){
	$text = $self->say("priv set -q test;printflag $flag");
	if ($text =~  /no such flag/){
	    croak "Incorrect flag\n";
	    }
	}
    my @arr = split(" ",$text);
    return $arr[2];
    }

=head2 ::setflag

    Sets the value of flag.

=over

=item Arguments

B<flag> => I<String>

I<Required.>
    Flag to be set.

B<value> => I<String>

I<Required.>
    Value of flag to be set.

=item Examples

=back

    my $filer_object = NOTE::Filer->connect( FILER => $FILER );
    $filer_object->setflag(flag=>"wafl_reclaim_threshold_s",value=>1);

=cut

sub setflag{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    croak "Specify the flag\n" if (!exists $args{flag});
    croak "Specify the value of the flag\n" if (!exists $args{value});

    my $flag = $args{flag};
    my $value = $args{value};
    #Set the flag
    my $text = $self->say("priv set -q diag;setflag $flag $value");
    if ($text =~ /invalid integer value/){
        croak "Invalid value is trying to be set";
    }
    if ($text =~ /no such flag|cannot access flag/){
	$text = $self->say("priv set -q test;setflag $flag $value");
        if ($text =~ /invalid integer value/){
            croak "Invalid value is trying to be set";
        }
        if($text =~ /usage: setflag <flag> <value>/){
           croak "Invalid value is trying to be set";
        }
	if ($text =~ /no such flag/){
	    croak "No such flag\n";
	    }
	}
    #Verify whether flag is set
    my $ret_val = $self->printflag(flag=>$flag);
    $value =~ s/^\s+//; #remove leading spaces
    $value =~ s/\s+$//; #remove trailing spaces
    $ret_val=lc($ret_val);
    $value=lc($value);
    if((($value =~ /^0x/i) && ($ret_val =~ /^0x/i) ) || (($value !~ /^0x/i) && ($ret_val !~ /^0x/i))){
        if ($ret_val ne $value){
            croak "Setting flag failed";
        }
    }
    elsif($value !~ /^0x/i){
        if(hex($ret_val) ne $value){
            croak "Setting flag failed";
        }
    }
    else{
        if($ret_val ne hex($value)){
            croak "Setting flag failed"
        }
    }
    return 1;
}

=head2 ::filer_host_entry

Writes supplied hostname & ip into /etc/hosts file of the Filer

=over

=item Arguments

B<host> => I<String>
Name of the host

B<host> => I<integer_in_ip_format>
IP of the host

=item Examples

    $filer->filer_host_entry(host => "filer1", ip => "10.72.230.10");
    $filer->filer_host_entry(host => $filer1->host, ip => $filer1->defaultip);

=back

=cut

sub filer_host_entry{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;

    my ( $hostname, $hostip, $hostfile, $host_entry );

    unless ( exists $args{host} && exists $args{ip} ) {
        die "Insufficient no of arguements\n";
    }

    $hostname = $args{host};
    $hostip   = $args{ip};

    #Writes into /etc/hosts
    $hostfile    = "/etc/hosts";
    $host_entry  = "$hostip" . " $hostname";
    my $cmd = "wrfile -a $hostfile $host_entry";

    #Make sure entry is not present already
    my $ret = $self->say( "priv set -q test;rdfile %s", $hostfile );
    if ( $ret =~ /$host_entry/ ) {
	NOTE::logcomment("Entry in /etc/hosts file already present\n");
	return;
	}

    $self->say( "priv set -q test;%s", $cmd );

    my $text = $self->say( "priv set -q test;rdfile %s", $hostfile );
    unless ( $text =~ /$host_entry/ ) {
        die "Writing to hostfile failed";
    }

} ## end sub filer_host_entry

=head2 ::filer_host_access

Sets access permission on Filer

=over

=item Arguments

B<access_to> => I<String>
Access permission to be set on Filer. Applicable values "all","none"&hostname. Defaults to "all".

=item Examples

    $filer->filer_host_access(access_to => $filer1->name);
    $filer->filer_host_access(access_to => "all");
    $filer->filer_host_access(access_to => "none");

=back

=cut

sub filer_host_access {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;

    unless ( exists $args{access_to} ) {
        die "Insufficient no of arguements\n";
    }

    #Set option snapmirror.access on Filer
    $args{access_to} = "host=" . $args{access_to};
    if (!filer::filer_options_set($self->conn,'snapmirror.access',$args{access_to},1)){
        die "Unable set access permission\n";
        }
    NOTE::logcomment("Option snapmirror.access is set to ".$args{access_to}." successfully\n");
} ## end sub filer_host_access

=head2 ::validate_mode

Validates the connection mode of the transfers on the filer. Checks the output 
of the 'netstat -an' command on the filer to determine if the transfers are 
going through the said connection mode (or local address) and port.

=over

=item Arguments

B<connection_mode> => C<inet> | C<inet6>

I<Required>
The connection mode through which the transfers are happening

Or

B<local_address> => C<ipv4 or ipv6 address of the filer> 

I<Required>
The local address on filer through which the transfers are happening

B<port> => C<port> 

I<Required>
The local port through which transfers are happening. Can be regex also.

=item Arguments

None.

=item Example

=back

     $filer->validate_mode(connection_mode => inet);
     
     $filer->validate_mode(connection_mode => inet6, port => 1056.); 
     
     $filer->validate_mode(local_address => 10.225.215.230 , port => 10567);

=cut

sub validate_mode {
    my ($self, %args) = @_;
    my ($addr_matched, $port_matched, @match_addr,
        $filer_name,   $output,       @uniq_addr,
        %nestat_hash,  $nestat_val,   $netstat_ref);
    $addr_matched = $port_matched = 0;

    # Validate if either argument connection_mode or local_address exist
    if (!exists $args{connection_mode} && !exists $args{local_address}) {
        croak 'Neither connection_mode nor local_address were provided';
    }

    # Validate if right connection mode was passed
    if (   $args{connection_mode} ne 'inet'
        && $args{connection_mode} ne 'inet6'
        && !exists $args{local_address}) {
        croak 'Improper connection mode was passed';
    }
    
    # Issue netsat -an on the filer.
    $netstat_ref = filer::filer_netstat_get($self->conn, 1);
    %nestat_hash = %{$netstat_ref};
    ($nestat_val) = values(%nestat_hash);
    
    # If local_address is provided this is the only address to match.
    # Else determine the addresses according to connection mode using the
    # command getXXbyYY getaddrinfo
    if (exists $args{local_address}) {
        push @match_addr, $args{local_address};
    } else {
        $filer_name = $self->host;
        $output =
          $self->say("priv set -q diag; getXXbyYY getaddrinfo $filer_name");

        if ($output =~ /getaddrinfo($filer_name)\s+failed/) {
            croak 'getXXbyYY did not return the filer address';
        }
        while ($output =~ /address:\s+(.*)/g) {
            $addr = $1;
            if ($args{connection_mode} eq 'inet') {
    # If connection_mode is inet get the list of ipv4 addresses on filer
                push @match_addr, $addr
                  if ($addr =~ /^[\d]+\.[\d]+\.[\d]+\.[\d]+$/);
            } else {
    # Else get list of ipv6 addresses on filer
                push @match_addr, $addr
                  if ($addr !~ /^[\d]+\.[\d]+\.[\d]+\.[\d]+$/);
            }
        } ## end while ($output =~ /address:\s+(.*)/g)
    } ## end else [ if (exists $args{local_address...

    # Get the list of unique filer addresses
    @uniq_addr = uniq @match_addr;
    
    # Validate if transfer are happening through the specified filer addresses
    foreach my $address (@uniq_addr) {
        foreach my $nestat (@$nestat_val) {
            if (   ($nestat->{'Local Address'} =~ /$address/)
                && ($nestat->{'State'} eq 'ESTABLISHED')) {
                $addr_matched = $address;
    # Validate if transfer is happening through specified port.
                if (exists $args{port}
                    && $nestat->{'Local Address'} =~ /\.($args{port})$/) {
                    $port_matched = $1;
                    last;
                }
            }
        } ## end foreach my $nestat (@$nestat_val)
    } ## end foreach my $address (@uniq_addr)
    
    # Croak here if connection mode or port were not validated.
    # Else print appropriate message.
    if (exists $args{port} && $port_matched != 0) {
        NOTE::logcomment(  'Connection mode was verified using address '
                         . $addr_matched
                         . ' and port '
                         . $port_matched);
    } elsif (   exists $args{port}
             && ($addr_matched ne '0')
             && ($port_matched eq '0')) {
        croak 'Connection mode could not be verified using port ' . $args{port};
    } elsif ($addr_matched) {
        NOTE::logcomment(
                 'Connection mode was verified using address ' . $addr_matched);
    } else {
        croak 'Connection mode is not verified';
    }

} ## end sub validate_mode

1;
