# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/ImageBackup.pm#1 $
##
## @summary note ImageBackup class
##
## @status public
##
## @pod    here
##
## @author sydneyw@netapp.com, dl-dpgqa-automation@netapp.com
##

=head1 NAME

NOTE::ImageBackup

=head1 DESCRIPTION

This module provides Image Backup functionalities. The module requires the name of a
tape device or file path, and optionally a
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
storage object that hold the source of the data being saved or destination
to restore to. If storage is not specified then internal storage will be generated for it's use.

In addition internally generated storage maybe directed to contain pre-generated dataset.

If optional parameters are not supplied, this module will use the following CLI parameters
to determine the behavior for objects of this class:

B<DEVICE>

Specify the tape device or file to be use to backup to or restore from.

B<VSERVER|FILER>

Specify the Vserver or Filer to create the associated storage to be used if not specified.

B<IB_CMD>

Determines FILER interface to use. Valid values are: cli, nbu or zapi. Default: cli

B<NBU_SPEC>

If IB_CMD is set to 'nbu', additional information needs to be to supplied to the NBU application.
Parameter format is as follows:

<NBU_SPEC>=<nbu master>:<filer domain>:<robot name>:<robot type>:<drive type>

where:
<nbu master>   Host where the NBU application is running on, e.g. sdqa-c68
<filer domain> The domain of the filer, e.g. lab.netapp.com
<robot name>   The tape libary robot name containing the tape drive to use.
<robot type>   The robot type, e.g. TLD, 8MM etc
<drive typer>  The type of the tape drive, e.g. hcart, 8MM etc

Example:
	IB_CMD=nbu NBU_SPEC=sdqa-c68:lab.netapp.com:mc2:TLD:hcart

Also NBU application must be installed on the host that the script is to be run on or use the '-client' cli parameter to specify a host
where the NBU application exists.

B<DATASET>

Indicates what dataset to pre-create on internally created storage. To use existing master dataset
specify [<filer>:<volume name>]. This will cause the contents of the specified volume to be copied
to the internally created storage. To specify data generation using existing tool, use the key
word "DS-populate", "DS-small" and "DS-SS<count>", where:

DS-populate is a standard dataset generated using DataManip and/or RichData tools.
DS-small    is generated in the same way as DS-populate, only a minimal size speed up data generation time.
DS-SS"spec" generates a dataset with multiple snapshots, and specifies which snapshot to be used for the
            restore. The format of <spec> is <count>:<restore_ss>, where <count> is an integer between 20 and 254.
            If the integer falls outside of this range, then it defaults to 20. And <restore_ss> is an integer
            between 1 and <count>, defaults random number between 1 and <count>.
DS-DIRECT:<volname>
            Uses the <volname> directly without copying. The volume must be on the filer that IB will run backup from.

Override the behavior of this object from a batch, or 'automatic' mode, to 'manual' mode.
In 'manual' mode input is requested after manual intervention such as changing tape,
rather then aborting the test.

B<MODE>

Determines 'restore' mode to use when verifying backup data.
The can be one of 'all', 'image' or 'logical'.
Defaults to 'all', which means both 'image' and 'logical'.

B<CLIENT>

Determine the data generation tool(s) to use when DATASET is set to the key word "DS-populate".
Used to contain list of comma separated clients to be used to generate data. Each client type,
Unix, Linux or Windows, will determine whether NFS or CIFS data will be generated.

B<VOL_TYPE>

Volume type, flex or trad. Default flex.

B<VOL_SIZE>

Volume size of the associated storage to create.

B<COMPARE>

Determine if treecompare will be executed to verify data from restore. Default: 1

B<WAFLIRON>

Determine if wafliron as part of verification of data from restore. Default: 0

B<FILER_AGGR>

Name of specific aggregate to create volumes on if specified.

=head1 METHODS

=cut

use POSIX qw();
package NOTE::ImageBackup;
use NOTE;
our @ISA = qw(NOTE);

use strict;
use Carp;
use Tharn;
use ImageBackup;
use LogReader::Backup;

use fileSystem qw( treecompare );

use constant SS_PREFIX => "ib_ss_";

sub _is_note_volume {
	my ($vol) = @_;

	return ref ($vol) =~ /^NOTE::Volume/;
}

=head2 ::new

Creates a NOTE::ImageBackup object containing a storage (volume) and tape device / file.
A storage can be pass in to be use by this object or it can be created by the object.
Internally created storage may be pre-populated with a dataset via data generation tools
or copied from a master dataset.

=over

=item Arguments

B<tag> => I<string>

An internal name for this object to be used to identify the object.

B<device> => I<device spec>

This is the name of the tape device to use. It could be a physical/logical tape device name, or
it could be a file spec of the form /vol/<vol>/<file>.

B<storage> => I<storage spec>

I<Optional.>

This is a
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
object to be use as the associated storage for this object.

If <storage> is not supplied than the following CLI parameters, in order of precedence, will be
used to identified where to create the required storage on: VSERVER, NODE, FILER

B<size> => I<volume size>

I<Optional.>

Volume size of the associated volume to be created. Default: 20G

B<populate> => I<0|1|"HASH">

I<Optional.>

Determines if the internally generated storage will be pre-populated using the 'DATASET' CLI value.
The default internal generation will use the L<C<::populate>|/::populate> method to carryout data
generation.

In addition this parameter can be used to pass specifications to the standard data generation tools,
L<C<DataManip>|lib-DataManip-local-pm>
or
L<C<RichData>|lib-RichData-pm>.

Default: 0

B<client> => I<list of clients>

I<Optional.>

List of comma separated clients where data will be generated. The L<C<::populate>|/::populate> method will
be used to generate data, and the data generation tools used will be as defined by that method.
Default: localhost

B<manual> => I<0|1>

I<Optional.>

Determines whether to prompt after completion of manual task such as tape change.
In automatic mode, script is aborted if manual intervention is encountered. Default: 0.

=item Examples

=back

	# Uses CLI parameters to specify object spec
	# CLIs used: DEVICE for tape device. FILER and VOL_SIZE for storage object
	$atape = NOTE::ImageBackup->new();

	# Uses the supplied NOTE::Volume storage object
	$bu = NOTE::ImageBackup->new(device=>"rst01", storage=>$vol_obj);

	# Use standard data generation tools to generate data on a 50GB volume
	$bu = NOTE::ImageBackup->new(device=>"rst0a", size=>"50G",
			populate=>1, client=>"unix_host,windows_client");

	# Use dataManip with specific parameter
	$bu = NOTE::ImageBackup->new(device=>"rst0a", client=>"unix_host", populate=>{DIR_AVG=>1});

=cut

sub new {
	my ($class, %args) = @_;

	my $self = {};

	$self->{_owner} = $$;
	$self->{_tag} = delete $args{tag};

	$self->{_compare} = NOTE::option( q/$COMPARE:1/ );
	$self->{_wafliron} = NOTE::option( q/$WAFLIRON/ ) || 0;

	$self->{_device} = $args{device} || NOTE::option( q/$DEVICE/ );
	croak "Unable to determine tape device..."
		unless $self->{_device};

	if ( exists $args{storage} ) {
		if ( _is_note_volume($args{storage}) ) {
			# For NOTE::Volume object: Can call NOTE::Volume::copy directly 
			#                       or NOTE::ImageBackup::populate after object creation
			croak "Cannot populate pre-existing NOTE::Volume object..."
				if $args{populate};
			$self->{_storage} = $args{storage};
		} else {
			croak "storage must be a NOTE::Volume object, aborting...";
		}
	} else {
		# Create internal storage VSERVER, NODE or FILER, VOL_SIZE and DATASET
		my %opts;
		if ( ref $args{Filer} eq "NOTE::Filer" ) {
			$opts{Filer} = $args{Filer};
		} else {
			$opts{Filer} = NOTE::Filer->connect($args{Filer} || NOTE::option( q/$VSERVER:$NODE:$FILER/ ));
		}
		$opts{type} = NOTE::option( q/$VOL_TYPE/ ) || "flex";
		my $for_sim = $opts{Filer}->is_a_simulator;
		if ( $opts{type} =~ /flex/i ) {
			if ( my $aggr = NOTE::option( q/$FILER_AGGR/ ) ) {
				$opts{Aggregate}{name} = $aggr;
			} else {
				$opts{Aggregate} = { recycle => 1, disks => ($for_sim ? 6 : 5) };
			}
			$opts{size} = $args{size} || NOTE::option( q/$VOL_SIZE/ ) || ($for_sim ? "500M" : "20G");
		} else {
			$opts{disks} = $args{size} || NOTE::option( q/$VOL_SIZE/ ) || ($for_sim ? 6 : 3);
			$opts{zeroed} = 1;
		}
		my $dataset = NOTE::option( q/$DATASET:DS-populate/ );
		if ( $dataset =~ /DS-SS(.*)/ ) {
			my ($max_ss, $restore_num) = split /:/, $1;
			$max_ss = 254 if $max_ss > 254; # Leave one free snapshot for backup to use
			$max_ss= 20 if $max_ss < 20;
			if ( $max_ss) {
				$self->{_max_ss} = $max_ss;
				NOTE::logcomment("Create ".$self->{_max_ss}." snapshot(s)");
				$restore_num = int (($max_ss * rand ()) + 0.5) if $restore_num < 1 || $restore_num > $max_ss;
				$self->{_restore_ss} = SS_PREFIX.$restore_num;
				NOTE::logcomment("Restore to snapshot ".$self->{_restore_ss});
			}
		}
		if ( $args{populate} ) {
			if ( $dataset =~ /DS-(populate|small|SS)/ ) {
				my $pop_args;
				$self->{_client} = $args{client} || NOTE::option( q/$CLIENT/ ) || "localhost";
				$args{populate} = { dir_avg => 4 } if $for_sim || $dataset =~ /DS-(small|SS)/;
				$pop_args = join ", ", (map { $_."=>'".$args{populate}{$_}."'"; } keys %{$args{populate}})
					if (ref $args{populate}) =~ /HASH/;
				$pop_args .= ", max_ss=>".$self->{_max_ss};
				$opts{dataset} = "NOTE::ImageBackup::_populate(shift, \'$self->{_client}\', $pop_args)";
			} elsif ( $dataset =~ /DS-DIRECT:(\w+)/ )  {
				$opts{name} = $1;
			} else {
				$opts{dataset} = "NOTE::ImageBackup::_copy_master(shift, \'$dataset\')";
			}
			$opts{recycle} = 1;
		} else {
			$opts{recycle} = 0;
		}
		$self->{_storage} = NOTE::Volume->create(%opts);
		$self->{_storage}->snapshot_sched(sched=>"0 0 0"); # Turn off automatic snapshot generation
	}

	$self->{_manual} = $args{manual} || NOTE::option( q/$MANUAL/ ) || 0;

	bless $self, $class;
	NOTE::reference($self);

	NOTE::logcomment("Created ".$self->description);

	# burt293726 workaround - Create a vol to hold tapefile instead of using root volume
	if (  $self->{_storage}->Filer->is_a_simulator && $self->{_device} =~ /\//  ) {
		# Create a volume to hold tapefile if using root volume of a simulator
		$self->{_tape_vol} = NOTE::Volume->create(
									name    => "vl_imagebackup_vol",
									type    => $self->{_storage}->type,
									size    => $self->{_storage}->size,
									Filer   => $self->{_storage}->Filer,
									recycle => 1
							 );
		$self->{_device} = $self->{_tape_vol}->mount_path.$self->{_device};
		NOTE::logcomment("Relocating tapefile: ".$self->{_device});
	}

	# NBU Spec consists of 5 parts: <dma>:<client network>:<robot name>:<robot type>:<device type>
	#                      Example: sdqa-c68.lab.netapp.com:mc1:TL8:8mm
	if ( my $nbu_spec  = NOTE::option( q/$NBU_SPEC/ ) ) {
		my $reg = $self->storage->Filer->registry;
		$reg->{'ndmp.enable'} = "on";
		$reg->{'ndmp.access'} = "all";
		$reg->{'ndmp.authtype'} = "challenge,plaintext";

		my ($dma, $client_network, $robot_name, $robot_type, $device_type) = split /:/, $nbu_spec;
		my $nbu_client = $self->storage->Filer->hostrec->hostname;
		my @devices = ( $self->{_device} ) ;
		# Wait for all devices to be offline before proceeding
		foreach my $device (@devices) {
			$self->device_wait(status=>'offline');
		}
		NOTE::logcomment("Initialize NBU for DMA $dma, Client $nbu_client. "
						."Robot: Name $robot_name, Type $robot_type. "
						."Device Type $device_type, Device(s): ".(join ", ", @devices));
		ImageBackup::nbu_init($dma, $nbu_client, $client_network, $robot_name, $robot_type, $device_type, @devices);
		$self->{_nbu} = 1;
	}

	return $self;
}

sub _copy_master {
	my ($vol, $dataset) = @_;

	my %opts;
	if ( $dataset =~ /:/ ) {
		my ($filer, $vol) = split ":", $dataset;
		$opts{Filer} = $filer;
		$opts{name} = $vol;
	} else {
		$opts{name} = $dataset;
	}
	my $src_vol = NOTE::Volume->attach(%opts);
	croak "Mismatch type: src vol ".$src_vol->name.", type ".$src_vol->type.", dst vol ".$vol->name.", type ".$vol->type
		unless $src_vol->type eq $vol->type;
	croak "Src vol too large: src vol ".$src_vol->name.", size ".$src_vol->size.", dst vol ".$vol->name.", size ".$vol->size
		if $src_vol->size > $vol->size;
	$vol->copy(src=>$src_vol);
}

=head2 ::description

Returns a description for this object, including the 'tag' is specified.

=over

=item Arguments

B<None.>

=item Examples 

=back

	$atape->description;

=cut

sub description {
	my ($self) = @_;

	return "Image Backup Object (".(exists $self->{_tag} ? $self->{_tag} : "").") : "
			."\n	Device: ".$self->device_name
			.", Storage: ".$self->storage->description;
}

=head2 ::tag

Return this object's tag if exists

=over

=item Argments

B<None.>

=item Example

=back

	$atape->tag;

=cut

sub tag { return shift->{_tag}; }

=head2 ::populate

Populate the associated storage with standard data generation tools. Depending on the clients'
host type will either use
L<C<DataManip>|lib-DataManip-local-pm>
or
L<C<RichData>|lib-RichData-pm>
to generate data. For each client, creates a qtree named after the client, and generate
data within that qtree.

Parameters for
L<C<DataManip>|lib-DataManip-local-pm>
or
L<C<RichData>|lib-RichData-pm>
will be pass-throught to the data generation tools.

=over

=item Arguments

B<client> => I<list of names>

I<Optional.>

List of comma separated client names to be used to execute the data generation tools. The tool
used will be appropriate to the client's host type.
For Unix or Linux hosts,
L<C<DataManip>|lib-DataManip-local-pm>
will be used. For Windows,
L<C<RichData>|docs::lib-RichData-pm>
will be used.

If a client name is listed twice, only one instance will be invoked.

Default: localhost

=item Examples

=back

	# Use localhost to generate data
	$atape->populate();

	# Generate both NFS (wildwood) and CIFS (win3) data.
	$atape->populate(client=>"wildwood,win3");

=cut

sub populate {
	my ($self, %args) = @_;

	NOTE::logcomment("Image Backup Populate: ".$self->storage->description);
	my @client = $self->_client_list(delete $args{client});
	$self->{_client} = join ",", @client; # Save client list
	_populate($self->storage, $self->{_client}, %args);
}

sub _populate {
	my ($storage, $hosts, %args) = @_;

	my $max_ss = delete $args{max_ss};

	my @clients = split ",", $hosts;
	foreach my $client (@clients) {
		my $qt = NOTE::Qtree->create(Volume=>$storage, name=>$client, anonymous=>1);
		my $host = NOTE::Client->new(storage=>$qt, name=>$client);
		my $ds = NOTE::option( q/$DATASET/ );
		if ( $ds =~ /^DS-SS/ && host($client)->hosttype =~ /windows/i ) {
			# Resduce data for DS-SS dataset
			NOTE::logcomment("Reduce windows RichData setting for DS_SS");
			$args{dir_avg} = 2.2;
			$args{file_size_avg} = 16384;
		}
		if ( $storage->size >= (5 + @clients) * 1024 * 1024
		&& ! $storage->Filer->is_a_simulator ) {
			# Increase data generation for hardware filers only
			if ( host($client)->hosttype =~ /windows/i ) {
				$args{dir_avg} = 3.2 unless exists $args{dir_avg};
				$args{file_size_avg} = 65536 unless exists $args{file_size_avg};
			} else {
				$args{dir_avg} = 5 unless exists $args{dir_avg};
			}
		}
		$host->populate(%args);
		$host->DESTROY();
	}

	if ( $max_ss ) {
		my $ss_cnt = $storage->snapshot_list + 1;
		$storage->snapshot_create(SS_PREFIX.$ss_cnt++);
		my $clnt = NOTE::Client->new(storage=>$storage);
		_generate_ss($storage, $ss_cnt, $max_ss, $clnt, @clients);
	}
}

sub _generate_ss {
	my ($vol, $ss_cnt, $max_ss, $clnt, @clients) = @_;

	_modify($vol, client_list=>\@clients);
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	my $nfs = $vol->nfs;

	my $lun = "a_lun".$ss_cnt;
#	$clnt->create_lun(name =>$lun, size =>"16m");
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

#	$clnt->resize_lun(name=>$lun, size=>"8m");
	my @file = $nfs->create_files(10);
	my $dir = $nfs->mkdir;
	my $sub_dir = $nfs->mkdir->mkdir->mkdir("sub_dir");
	my @dir_f = $dir->create_files(10);
	my $hd_ln = $file[0]->link;
	my $sym_ln = $file[1]->symlink;
	my $dir_to_rm = $nfs->mkdir("dir_to_rm");
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	$sym_ln->mv("/sym_ln_to_root_".$ss_cnt);
	my $sub_f = $dir->create;
	$sub_f->mv("..");
	$file[8]->mv($dir->path."new_name_".$ss_cnt);
	my $f = pop @file; $f->unlink;
	$file[3]->touch;
	$file[4]->chmod('0444'); 
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	my $hd_ln2 = $file[8]->link;
	$nfs->create();
	$hd_ln->unlink;
	$file[5]->unlink;
	$dir->chmod('0744');
	my $sub_f = $sub_dir->create;
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	$file[6]->mv('move_within_dir_'.$ss_cnt);
	$sub_f->mv('/move_to_root_'.$ss_cnt);
	$file[7]->mv($sub_dir->path."/move_to_sub_dir_".$ss_cnt);
	$sub_dir->mv("move_to_new_subdir");
	my $qt_1 = NOTE::Qtree->create(Volume=>$vol);
	(pop @file)->unlink;
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	$nfs->create();
	$nfs->mkdir();
	$dir_to_rm->rmdir();
	$nfs->create_socket("socket_".$ss_cnt);
	$file[0]->atime;
	$file[1]->mtime;
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	# A snapshot with no activities
	$vol->snapshot_create(SS_PREFIX.$ss_cnt++); return if $ss_cnt > $max_ss;

	_generate_ss($vol, $ss_cnt, $max_ss, $clnt, @clients);
}

=head2 ::modify

Use standard data generation tools to modify the data that was created using the
L<C<::populate>|/::populate> method.

Parameters for
L<C<DataManip>|lib-DataManip-local-pm>
or
L<C<RichData>|lib-RichData-pm>
will be pass-throught to the data generation tools.

=over

=item Arguments

B<client> => I<list of names>

This is either the same list as that used by the previous call to L<C<::populate>|/::populate> or a subset
so that the data modification tool can be pointed to the appropriate data set.

=item Examples

=back

	$atape->modify(); # changes all data created with populate

	# Modify a subset of the populated data
	$atape->populate(client=>"wildwood,win3");
	$atape->modify(client=>"wildwood", DIV_AVG=>1); # DIR_AVG will be passed to dataManip
	$atape->modify(); # Use wildwood and win3 to modify data

=cut

sub modify {
	my ($self, %args) = @_;

	NOTE::logcomment("Image Backup Modify: ".$self->storage->description);
	my @clients = $self->_client_list(delete $args{client});
	_modify($self->storage, client_list=>\@clients, %args);
}

sub _modify {
	my ($storage, %args) = @_;

	my $clients = delete $args{client_list};
	foreach my $client (@{$clients}) {
		my $qt = NOTE::Qtree->attach(Volume=>$storage, name=>$client);
		my $host = NOTE::Client->new(storage=>$qt, name=>$client);
		$host->modify(%args);
		$host->DESTROY();
	}
}

=head2 ::client_list

List of clients used to pre-populate data for this object

=over

=item Arguments

B<None.>

=item Examples

=back

	@list = $atape->client_list;

=cut

sub client_list {
	my ($self) = @_;

    return (exists $self->{_client} ? $self->_client_list() : ());
}

sub _client_list {
	my ($self, $client) = @_;

	$client = ($self->{_client} || "localhost") unless $client;
	my %list = map { $_ => 1; } split ",", $client;
	return keys %list;
}

=head2 ::backup

Initiate a backup of data from storage to the device.
Control is returned the backup command have started.
If called in background mode, the job id is returned and progress of the backup can be
monitored using the L<C<::status>|/::status> method or automatically using the
L<C<::monitor>|/::monitor> method.

This method is a wrapper for
L<C<ib_backup>|lib-ImageBackup-pm/::ib_backup>
The following are pass-through parameters: I<block_size>, I<mode>, I<snapshot>

=over

=item Arguments

B<background> => I<0|1>

I<Optional.>

Indicate whether backup job should be run in background mode. Default: 0

B<timeout> => I<integer>

I<Optional.>

Integer in seconds to timeout job that has idled.

B<interval> => I<integer>

I<Optional.>

Monitor check interval in seconds.

B<manual> => I<0|1>

I<Optional.>

Flag to indicate manual or automatic mode. 0 indicates automatic mode.
In manual mode input is required to continued operation after manual interaction such as tape change.
Default: 0

B<force> => I<boolean>

I<Optional.>

Boolean to force execution of the backup command whether the tape device is ready or not. Default 0

=item Examples

=back

	$atape->backup();           # Regular full backup of association storage
	my $job_id = $atape->backup(background=>1); # Regular full backup in background mode
	$atape->backup(block_size=>1024); # Backup using tape block size of 1024KB
	$atape->backup(manual=>1);  # Prompts for input when manual operation have been completed

=cut

sub backup {
	my ($self, %args) = @_;

	# Default src and dst paths unless supplied
	$args{src_path} = $self->storage->mount_path unless exists $args{src_path};
	$args{dst_path} = $self->device_name unless exists $args{dst_path};
	$args{all_snapshots} = 1 if $self->{_max_ss} && ! exists $args{all_snapshots};

	my $force = delete $args{force};

	my $background = delete $args{background};
	# Compose monitor options
	my %opts;
	$opts{manual}   = (exists $args{manual} ? delete $args{manual} : $self->{_manual});
	$opts{timeout}  = delete $args{timeout} if exists $args{timeout};
	$opts{interval} = delete $args{interval} if exists $args{interval};

	unless ( $force ) {
		# Check if device is ready unless in 'force' mode
		eval {
			$self->device_wait;
		};
		croak $self->description."\n$@" if $@;
	}

	$self->{_job_type} = "BKP";
	$self->{_log} = ""; # Clears the log
	NOTE::logcomment("Starting backup for ".$self->description);
	$self->_start_bu_log();
	eval {
		$self->{_job_id} = ImageBackup::ib_backup($self->storage->Filer->conn, %args);
	};
	croak $self->description."\nBackup failed to start: $@" if $@;

	if ( $background ) {
		NOTE::logcomment("backup started. Job ID is ".$self->{_job_id});
		return $self->{_job_id};
	} else {
		sleep 5; # pause a little before monitoring
		$self->monitor(%opts);
		return $self;
	} 
}

=head2 ::set_restore_snapshot

Given a list of snapshots, randomly selects one to be used in subsequent restore and compare.
If no snapshoit list is supplied, it will clear previously set snapshot selection.

By default a random selection is made when using DS-SS dataset, which contains multiple snapshots.

=over

=item Arguments

B<snap_list> = I<array of snapshot names>

A list of snapshots to select from. Existing selection will be removed if no snapshot is passed.

=item Examples

=back

	$re->set_restore_snapshot(@snapshot_list);

=cut

sub set_restore_snapshot {
	my ($self, @snap_list) = @_;

	if ( @snap_list ) {
		$self->{_restore_ss} = $snap_list[int (rand() * @snap_list)];
		NOTE::logcomment("Set restore ss to ".$self->{_restore_ss});
	} else {
		delete $self->{_restore_ss};
		NOTE::logcomment("Cleared restore ss");
	}
	return $self->{_restore_ss};
}

=head2 ::restore

Initiate a restore of data from the device to the associated storage. Default to 'image' restore unless
specified.

Control is returned once the restore command have started.
If called in background mode, the job id is returned, and progress of the restore can be monitored
using the L<C<::status>|/::status> method or automatically using the L<C<::monitor>|/::monitor> method.

This method is a wrapper for
L<C<ib_restore>|lib-ImageBackup-pm/::ib_restore>
The following are pass-through parameters: I<block_size>, I<mode>, I<snapshot>

=over

=item Arguments

B<restrict> => I<0|1>

I<Optional.>

Indicate whether to restrict the volume or not. Default: Restrict if using 'image' restore mode

B<background> => I<0|1>

I<Optional.>

Indicate whether backup job should be run in background mode. Default: 0

B<timeout> => I<integer>

I<Optional.>

Integer in seconds to timeout job that has idled.

B<interval> => I<integer>

I<Optional.>

Monitor check interval in seconds.

B<manual => I<0|1>

I<Optional.>

Flag to indicate manual or automatic mode. 0 indicates automatic mode.
In manual mode input is required to continued operation after manual interaction such as a tape
change. Default: 0

B<skip_rewind> => I<boolean>

I<Optional>

Indicates not to rewind the tape first. Default: 0. (ie rewinds the tape)

B<force> => I<boolean>

I<Optional.>

Boolean to force execution of the restore command and skip setup of tape device,
such as rewinding tape device or volume cannot be restricted for image restore, etc. Default 0

=item Examples

=back

	$atape->restore(mode=>"image"); # Use image restore mode
	$atape->restore(manual=>1); # Prompt after manual intervention to continue restore

=cut

sub restore {
	my ($self, %args) = @_;

	# In BR the default mode is image. logical is not supported.
	# remove next 2 lines once logical mode is supported.
	croak "logical mode not supported" if($args{mode} =~ /logical/i);
	$args{mode} = 'image';

	croak "Incorrect restrict/mode combination: (".($args{restict} ? "restrict" : "not restrict").", ".$args{mode}.")"
		if ( exists $args{retrict} && exists $args{mode}
		&&	(	($args{restrict} && $args{mode} =~ /logical/i)
			||	($args{mode} =~ /image/i && ! $args{restrict})
			)
		);
		
	# Default src and dst paths unless supplied
	$args{src_path} = $self->device_name unless exists $args{src_path};
	$args{dst_path} = $self->storage->mount_path unless exists $args{dst_path};

	# Default is to rewind the tape before a restore
	$self->{_job_type} = "RST";
	$args{mode} = NOTE::option( q/$MODE/ ) unless exists $args{mode};

	my $restrict = (exists $args{restrict} ? delete $args{restrict} : $args{mode} =~ /image/i);
	delete $args{mode}
		if exists $args{mode};

	# Set snapshot to restore when using multiple snapshot dataset such as DS-SS
	$args{snapshot} = $self->{_restore_ss}
		if exists $self->{_restore_ss} && ! ( $restrict || exists $args{snapshot} );

	my $force = delete $args{force};

	my $background = delete $args{background};
	# Compose monitor options
	my %opts;
	$opts{manual}   = (exists $args{manual} ? delete $args{manual} : $self->{_manual});
	$opts{timeout}  = delete $args{timeout} if exists $args{timeout};
	$opts{interval} = delete $args{interval} if exists $args{interval};

	unless ( $force ) { # skip if in force mode
		my $skip_rewind = delete $args{skip_rewind};
		eval {
			$self->device_rewind() unless $skip_rewind;
			$self->device_wait;
		};
		croak $self->description."\n$@" if $@;
	};

	eval {
		# Restrict volume for image restore only
		$self->storage->restrict
			if $restrict && ! $self->storage->restricted;
	};
	croak $self->description."\n$@" if $@ && ! $force;

	$self->{_log} = ""; # Clears the log
	NOTE::logcomment("Starting restore for: ".$self->description);
	$self->_start_bu_log();
	eval {
		$self->{_job_id} = ImageBackup::ib_restore($self->storage->Filer->conn, %args);
	};
	croak $self->description."\nRestore failed to start: $@" if $@;

	if ( $background ) {
		NOTE::logcomment("Restore started. Job ID is ".$self->{_job_id});
		return $self->{_job_id};
	} else {
		sleep 5; # pause a little before monitoring
		$self->monitor(%opts);
		return $self;
	}
}

=head2 ::compare

Verify that data on associated storage matches data on the supplied src volume.
Uses treecompare to match the data.
Croaks if mismatches are found.

IF a snapshot is speicifed, it will compare all snapshots upto specified snapshot. If not, the active file system 
is verified. Unless a restore is for a tape seeding backup the IB created snapshot will be ignored.

When using the multi-snapshot dataset DS-SS, a random snapshot is pre-selected for restore and verification.

The CLI parameter COMPARE can be used to override execution of treecompare.

=over

=item Arguments

B<src> => I<vol spec>

The volume containing the data that is to be compared with the restored data. This could be a
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
object or a path of the form /vol/<src>.

B<restore_ss> => I<snapshot name>

Compare all snapshots upto, and including this snapshot. If not specified verify all snapshots
Default: For DS-SS uses randomly selected snapshot, otherwise none.

B<snapshot> => I<snapshot name>

Compare only snapshot supplied. If not set, compare snapshots as appropriate.

B<break> => I<1|0>

Break the tape seeding mirror before performing treecompare. Default 0

B<wafliron> => I<boolean>

Flag to indicate whether to run wafliron against this object's storage as part of data verification.
If not set, CLI parameter WAFLIRON will determine this value.
Default: 0

B<timeout> => I<integer>

Timeout in secs. Default: calculated using data size at 0.1 GB / min

=item Examples

=back

	 $atape->restore(); # Restore data to the associated storage volume of this object
	 $atape->compare(src=>$src_vol); # Verify data between supplied src vol and associated storage.

=cut

sub compare {
	my ($self, %args) = @_;

	croak "src missing or incorrect type..."
		unless _is_note_volume($args{src});

	$args{break} = 1 unless exists $args{break}; # Setting this as default due to ONTAP change to image backup as the default mode

	my $src = delete $args{src};
	my $wafliron = $self->{_wafliron};
	$wafliron = delete $args{wafliron} if exists $args{wafliron};
	
	my $timeout = delete $args{timeout};
	unless ( $timeout ) {
		# Calculate timeout based on 0.1 GB / min
		$timeout = (($self->storage->used / ($self->storage->Filer->is_a_simulator ? 500000 : 100000)) * 60) + 60;
		# Add some sanity to calculated value. Hardcode timeout in script if longer timeout is needed
		if ( $self->storage->Filer->is_a_simulator ) {
			$timeout = 8 * 60 * 60 if $timeout > 8 * 60 * 60;
			$timeout = 30 * 60 if $timeout < 30 * 60;
		} else {
			$timeout = 4 * 60 * 60 if $timeout > 4 * 60 * 60;
			$timeout = 15 * 60 if $timeout < 15 * 60;
		}
	}

	my $tape_seeding_backup = 0;
	if ( $args{break}  ) {
		$self->_remove_mirror();
		$tape_seeding_backup = 1;
	}

	NOTE::logcomment("Compare storage ".$self->storage->description
				."\n	to src ".$src->description
				."\n	Timeout $timeout");
	if ( $self->{_compare} ) {
		my @rs_list = $self->storage->snapshot_list();
		my @bu_list = $src->snapshot_list();

		# Determine if restore is to a specific snapshot
		my $last_ss = ( exists $args{restore_ss} ? delete $args{restore_ss} : $self->{_restore_ss} );
		$last_ss = $args{snapshot}
			if defined $args{snapshot};
		# Perform treecompare on all snapshots upto last_ss
		NOTE::logcomment("Compare snapshot upto $last_ss")
			if $last_ss;
		my $done = 0;
		while ( my $rs_ss = pop @rs_list ) {
			if ( ($rs_ss =~ /snapshot_for_smtape\..*/) && ! $tape_seeding_backup ) {
				# Unless tape seeding backup skip ib created snapshot
				NOTE::logcomment("Skipping IB created backup $rs_ss...");
				next;
			} else {
				croak "Unexpected snapshot $rs_ss"
					if $done && $rs_ss !~ /__clean__/i; # Check Clean snapshot too
			}
			my $bu_ss = pop @bu_list;
			croak "No more snapshot at src volume."
				unless defined $bu_ss;
			NOTE::logcomment("Verifying snapshot $rs_ss");
			my $good = 0;
			eval {
				$good = treecompare($self->storage->Filer->conn,
									$src->mount_path()."/.snapshot/".$bu_ss,
									$self->storage->mount_path()."/.snapshot/".$rs_ss,
									args=>"-c -V",											# Workaround to mask ctime mismatches
									timeout=>$timeout
								);
			};
			croak $self->description."\nSnapshot $rs_ss compare failed: $@"
				if $@ || ! $good;
			$done = 1
				if $rs_ss eq $last_ss;
		}
		croak "Did not find restored snapshot $last_ss"
			if $last_ss && ! $done;
		unless ( $last_ss ) {
			# Verify active file system if no snapshot specified or an image restore
			eval {
				$self->storage->treecompare($src, args=>"-c -V", timeout=>$timeout); # Workaround to mask ctime mismatches
			};
			croak $self->description."\nCompare failed: $@" if $@;
		}
	} else {
		NOTE::logcomment("Skipping treecompare as directed");
	}

	if ( $wafliron ) {
		eval {
			$self->storage->wafliron_start(wait=>1);
		};
		croak ($self->description." Wafliron failed. Error: $@") if $@
	}

	return 1;
}

sub _remove_mirror {
	my ($self) = @_;

	my ($state, $status, $aggr, $options) = $self->storage->status;
	if ( $state =~ /online/ && $status =~ /read-only/ ) {
		# Break tape seeding mirror if read-only volume
		SnapMirror::snapmirror_break(
			host=>$self->storage->Filer->hostrec,
			dstpath=>$self->storage->hostpath);
		croak "Failed breaking the tape seeding mirror. Error: $@"
			if $@;
	} else {
		NOTE::logcomment("No need to break mirror as volume not read-only");
	}
}

=head2 ::verify_device_data

Performs a restore to the associated storage and then compares the restored data with
the supplied source volume. Croaks if mismatches are found.

The purpose of the method is to act as a 'short-hand' for the L<C<::restore>|/::restore>
and L<C<::compare>|/::compare>
methods, and may contain parameters appropriate for those methods.

The default is to verify the data using each of the restore mode, ie image and logical.

=over

=item Arguments

B<src> => I<vol spec>

The volume containing the data that is to be compared with the restored data. This could be a
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
object or a path of the form /vol/<src>.

B<mode> => I<image|logical|all>

I<Optional.>

Restoration mode to use. With 'all', data will be restored using each of the existing modes and
data comparison is performed after each restoration. Default: all

B<skip_rewind> => I<boolean>

I<Optional>

If set will assume the tape is already position at the begining of the backup and skip the rewind
command. Default 0. (ie to rewinds the tape)

B<wafliron>

Determines whether wafliron will be run against the restored volume as part of the data verification.
If not set, CLI parameter WAFLIRON will determine this value.
Default: 0

=item Examples

=back

	# compare data after an image and a logical restore
	$atape->verify_device_data(src=>$srv_vol);

	# compare data after an image restore only
	$atape->verify_device_data(src=>$src_vol, mode=>"image");

	# To restore from a specific snapshot
	$atape->verify_device_data(src=>$src_vol, snapshot=>"this_sssn");

=cut

sub verify_device_data {
	my ($self, %args) = @_;

	croak "Missing src volume..."
		unless _is_note_volume($args{src});
	my $src = delete $args{src};

	NOTE::logcomment("Verify device data for ".$self->description."\nSrc data: ".$src->description);

	# Default mode is image.
	my $mode = delete $args{mode} || NOTE::option( q/$MODE/ ) || "image";
	$mode = "logical,image" if $mode =~ /^all$/i;

	# In BR the default mode is image. logical is not supported.
	# remove next 2 lines once logical mode is supported.
	croak "logical mode not supported" if($args{mode} =~ /logical/i);
	unless($args{restrict}) {
		$args{restrict} = 1;
		NOTE::logcomment("Restrict mode turned on by default");
	} 

	my $wafliron = $self->{_wafliron};
	$wafliron = delete $args{wafliron} if exists $args{wafliron};

	my $restored;
	foreach my $m (split ",", $mode) {
		$self->replace_storage
			if $restored;
		my $skip_flg = delete $args{skip_rewind};
		my %re_opts = %args;
		my %cmp_opts;
		$cmp_opts{snapshot} = $args{snapshot}
								if exists $args{snapshot};
		if ( $m =~ /image/i ) {
			if ( exists $args{snapshot} ) {
				# don't set snapshot for restore
				delete $re_opts{snapshot}
			} else {
				# check all snapshots if there is an internal default ss for logical restore otherwise no snapshot is being restored
				$cmp_opts{snapshot} = SS_PREFIX.$self->{_max_ss}
										if $self->{_restore_ss};
			}
			# Treecompare creates a snapshot, so break the volume
			# for image restores.
			$cmp_opts{break} = 1;
		}
		eval {
			$self->restore(%re_opts, mode=>$m, skip_rewind=>$skip_flg);
			$self->compare(%cmp_opts, src=>$src, wafliron=>$wafliron);
		};
		croak "Verify Device Data: mode $m. Error:\n	$@"
			if $@;
		$restored = 1;
	}
	return $self;
}

=head2 ::abort

Abort the on-going backup or restore operation.

This method is a wrapper for
L<C<ib_abort>|lib-ImageBackup-pm/::ib_abort>

=over

=item Arguments

B<None.>

=item Examples

=back

	 $atape->backup(background=>1);
	 sleep(3);
	 $atape->abort(); # Aborts an on-going backup

=cut

sub abort {
	my $self = shift;

	croak "No running Image Backup job..."
		unless defined $self->{_job_id};

	NOTE::logcomment("Abort ".$self->description);
	eval {
		ImageBackup::ib_abort($self->storage->Filer->conn, $self->{_job_id});
	};
	croak $self->description."\nAbort failed: $@" if $@;
	return $self;
}

=head2 ::resume

Resume the on-going backup or restore operation. To be used to resume the backup/restore
after a tape change.

This method is a wrapper for
L<C<ib_continue>|lib-ImageBackup-pm/::ib_continue>

=over

=item Arguments

B<None.>

=item Examples

=back

	$atape->resume();

=cut

sub resume {
	my ($self, $new_device) = @_;

	croak "No running Image Backup job..."
		unless defined $self->{_job_id};
	my %opts;
	$opts{job_id} = $self->{_job_id};
	$opts{device} = $new_device if $new_device;
	NOTE::logcomment("Resume ".$self->description);
	eval {
		ImageBackup::ib_continue($self->storage->Filer->conn, %opts);
	};
	croak $self->description."\nResume failed: $@" if $@;
	return $self;
}

=head2 ::status

Returns the status of the backup or restore.

This method is a wrapper for
L<C<ib_status>|lib-ImageBackup-pm/::ib_status>.
The following are pass-throught parameters for that routine:
B<detail>

=over

=item Arguments

B<column> => I<name>

I<Optional.>

If supplied, the indicated column value is returned as opposed to a hash of the status.

B<manual> => I<0|1>

I<Optional.>

Prompt user for input after manual action such as a tape change, otherwise aborts the scripts.
Default: 0

=item Examples 

=back

	my $hash_ptr = $atape->status; # Return all statuses
	my $stat = $atape->status(column=>"progress"); # Returns 'progress' column
	my $stat = $atape->status(detail=>1); # Request detail status
	$atape->status(manual=>1); # Prompt for user input after manual actions if detected

=cut

sub status {
	my ($self, %args) = @_;

	croak "No active Image Backup job..."
		unless defined $self->{_job_id};

	NOTE::logcomment("Status ".$self->description."\njob id ".$self->{_job_id});
	my @stat;
	my $column = lc (delete $args{column}) if exists $args{column};
	my $manual = (exists $args{manual} ? delete $args{manual} : $self->{_manual});
	eval {
		@stat = ImageBackup::ib_status($self->storage->Filer->conn, spec=>$self->{_job_id}, %args);
	};
	croak $self->description."\nError: $@" if $@;

	if ( $stat[0]{'status'} =~ /waiting/i ) {
		my $msg = "Current status: ".$stat[0]{'status'}.", Progress: ".$stat[0]{'progress'};
		if ( $manual ) {
			NOTE::logcomment("$msg. Please type device name if different, followed by <enter> to continue: ");
			<>; chomp $_;
			$self->resume($_);
		} else {
			croak $self->description."\nMsg: $msg\nWhich is blocked. Aborting...";
		}
	}

	if ( defined $column ) {
		return $stat[0]{$column};
	} else {
		return %{$stat[0]};
	}
}

=head2 ::monitor

Provides automatic monitoring of backup / restore operations.

=over

=item Arguments

B<status> => I<string>

I<Optional.>

Expected status. Continuous polling until the expected status is reached or timeout.
Default: <job id> [Backup|Restore] End

B<timeout> => I<integer>

I<Optional.>

Integer in seconds to timeout job that has idled. Default: 180

B<interval> => I<integer>

Integer in seconds to pause in between polling for status. If not set, the default starts
from a low value, and gradually increased to 1mins. If set, then a fixed interval will be used.
Default: 5, gradually increased to 120.

B<manual> => I<0|1>

I<Optional.>

Prompt user for input after manual action such as a tape change, otherwise aborts the scripts.
Default: 0

=item Examples

=back

	my $job_id = $bu->backup(background=>1);
	$bu->monitor(status=>$job_id."BKP-End"); # Returns control when the backup completes.

=cut

sub monitor {
	my ($self, %args) = @_;

	croak "No active job..."
		unless defined $self->{_job_id};

	NOTE::logcomment("Monitor Job ".$self->{_job_id}.": ".$self->description);
	my $job_id = $self->{_job_id};
	my $status = $args{status} || $self->{_job_type}."-End";
	my $abort_stat = $self->{_job_type}."-Abort";
	my $remain = (exists $args{timeout} ? $args{timeout} : 180);
	my $interval = $args{interval} || 15;
	my $manual = (exists $args{manual} ? delete $args{manual} : $self->{_manual});

	my ($pause, $start, $init_mb, $last_mb, $last_dw, $last_upd_time) = (0, undef, undef, undef, undef, undef);
	while ($remain > 0) {
		# Collect log entries for this job
		$self->{_log_rdr}->restart();
		# Process new log entries
		my $abort_log;
		foreach my $line (@{$self->{_log_rdr}->matches("^$job_id ")}) {
			$self->{_log} .= $line;
			$abort_log .= $line
				if $line =~ /$abort_stat/i;
		}
		NOTE::logcomment("Job id $job_id logs found:\n".$self->{_log});
		croak "Image Backup job $job_id aborted, Error: $abort_log"
			if $abort_log;
		my (%stat, $curr_mb, $curr_dw, $curr_upd_time);
		eval {
			%stat = $self->status(detail=>1, manual=>$manual);
		};
		if ( $@ ) {
			croak "Failed to obtain status for job $job_id. Error: $@"
				unless $@ =~ /no status found/i;
			$remain -= $pause; # count down the clock if no status found
		} else {
			$curr_dw = $stat{'data warehouse blocks'};
			$curr_upd_time = $stat{'last update time'};
			my $progress = $stat{progress};
			# Determine progress and xfer rate
			if ( $progress =~ /(\d+\.*\d*)\s+([KMG])B/ ) {
				my ($kb, $unit) = ($1, lc($2));
				$curr_mb = ($unit =~ /g/ ? $kb * 1024 : ($unit =~ /m/ ? $kb : $kb / 1024.0));
				if ( $init_mb == undef ) {
					$init_mb = $curr_mb;
					$start = time;
					$last_upd_time = $curr_upd_time;
				} else {
					my $total_gb = ($curr_mb - $init_mb) / 1024.0;
					my $elapse = time - $start;
					NOTE::logcomment($self->description
						.":\nProcessed $progress. Approx. Xfer Rate: "
						.sprintf ("%.3lf", ($total_gb * 3600 / $elapse))." GB/hr");
				}
			} else {
				NOTE::logcomment($self->description.". Unexpected progress format: $progress");
			}
			# Countdown the clock only if no progress.
			$remain -= $pause
				if ($curr_mb <= $last_mb && $curr_dw <= $last_dw && $curr_upd_time eq $last_upd_time);
		}
		# Return success if expected status has been found
		if ( $self->{_log} =~ /$status/ ) {
			$self->_nbu_job_wait() if $self->{_nbu}; # Wait for NBU job to end
			NOTE::logcomment($self->description.": $status status found");
			delete $self->{_job_id}
				if $self->{_log} =~ /-(end|abort)/i; # Remove internal job id upon job end/abort
			return 1;
		} else {
			NOTE::logcomment($self->description."\n	Waiting for $status status."
					." Idle Time Remaining ".($remain < 0 ? 0 : $remain)." Sec(s)");
			if ( $remain ) {
				$pause = sleep($interval);
				# Default is to start low and progressively increased to a max.
				$interval += 15 unless ($interval > 119 || exists $args{interval});
			}
		}
		$last_mb = $curr_mb;
		$last_dw = $curr_dw;
		$last_upd_time = $curr_upd_time;
	}
	croak ($self->description."\nTimeout waiting for $status status");
}

sub _nbu_job_wait {
	my ($self) = @_;

	my $active;
	my $interval = 5;
	do {
		$active = ImageBackup::nbu_job_active($self->{_job_id});
		sleep $interval if $active; # Pause a little
		$interval += 5 unless $interval > 60;
	} while $active;
}

=head2 ::geometry

Returns the disk geometry of the associated storage volume.
Aborts if the the storage volume is not a traditional volume or other
failures.

=over

=item Argument

B<None.>

=item Example

=back

	my $geomtry = $re->geometry();

=cut

sub geometry {
	my ($self) = @_;

	croak "Storage must be a traditional volume to obtain geometry"
		unless $self->storage->type =~/trad/i;

	my $geometry;
	eval {
		$geometry = ImageBackup::ib_vol_geometry($self->storage->Filer->conn, $self->storage->name);
	};
	croak "Error obtaining geometry for vol ".$self->storage->description.". Error: $@"
		if $@;
	return $geometry;
}

=head2 ::tape_header

Returns the image header of the tape located in the tape device
associated with this object

=over

=item Argument

B<None.>

=item Example

=back

	my $header = $re->tape_header();

=cut

sub tape_header {
	my ($self) = @_;

	my $header;
	eval {
		$header = ImageBackup::ib_tape_header($self->storage->Filer->conn, $self->device_name);
	};
	croak "Failed to obtain image header from device ".$self->device_name.". Error: $@"
		if $@;
	return $header;
}

=head2 ::log

Returns the logs of the current or last image backup command

=over


=item Arguments

B<None.>

=item Examples

=back

	my $log = $bu->log;

=cut

sub log { return shift->{_log}; }

sub _start_bu_log {
	my ($self) = @_;

	unless ( exists $self->{_log_rdr} ) {
		$self->{_log_rdr} = LogReader::Backup->new($self->{_storage}->Filer->conn);
#		$LogReader::Backup::Path = "/etc/log/image_backup"; # Replace /etc/log/backup
	}

	$LogReader::Base::Verbose = 0;
	$self->{_log_rdr}->start();
	delete $self->{_log};
}

=head2 ::storage

Returns the associated
L<C<NOTE::Volume>|lib-NOTE-Volume-pm>
storage object.

=over

=item Arguments

B<None.>

=item Examples

=back

	$vol = $atape->storage;

=cut

sub storage { return shift->{_storage}; }


=head2 ::replace_storage

Replaces existing storage to get an empty volume, typically used to get 'clean' storage for
subsequent restores.

If storage is not supplied then an empty volume will be created to replace the existing storage.

=over

=item Arguments

B<new_storage> => NOTE::Volume object to use as this oobject's storage

I<Optional.>

If it is not specified than a new empty volume is generated to replace the existing volume

B<purge> => <boolean>

I<Optional.>

Indicate whether to purge the orginal storae. Default: 1
=item Example

=back

	$im->replace_storage;           # Internally generate a new storage
	$im->replace_storage($new_vol); # Replace storage with supplied NOTE::Volume object

=cut

sub replace_storage {
	my ($self, %args) = @_;

	if ( exists $args{new_storage} ) {
		croak "storage must be a NOTE::Volume object, aborting..."
			unless _is_note_volume($args{new_storage});
	} else {
		# Replace volume to make sure volume is empty
		my %opts;
		$opts{Aggregate} = $self->storage->Aggregate
			if $self->storage->type =~ /flex/;
		$args{new_storage} = $self->storage->dup(%opts);
	}
	$args{new_storage}->snapshot_sched(sched=>"0 0 0"); # Turn off automatic snapshot generation
	unless ( $args{purge} == 0 ) {
		$self->storage->purge;
		$self->storage->Aggregate->DESTROY
			if $self->storage->type =~ /flex/;
	}
	$self->{_storage} = $args{new_storage};
	return $self;
}

=head2 ::device_name

Returns the name of the tape device / file path being used to hold the backup data

=over

=item Arguments

B<None.>

=item Examples

=back

	my $name = $bu->device_name();

=cut

sub device_name { return shift->{_device}; }

=head2 ::device_status

Returns the status of the tape device. This returns 'ready' if a file is used as the device.

=over

=item Arguments

B<None.>

=item Examples

=back

	 my $dev_stat = $bu->device_status;

=cut

sub device_status {
	my ($self) = @_;

	return "ready"
		if $self->device_name =~ /^(\/|null)/ # Tapefile is always ready
		&& ! ($self->device_name =~ /^\/dev\//); # WWN device name is not a tapefile
	my $status;
	if ( $self->{_device} =~ /\/vol\// ) {
		$status = "ready";
	} else {
		eval {
			my %info = Tapeinfo::getTapeStatus($self->storage->Filer->conn, $self->{_device});
			$status = $info{status};
			croak "Device Status is $status"
				if $status =~ /unknown/i;
		};
		croak ($self->description."device status check failed. Error: $@")
			if $@;
	}

	return lc ($status);
}

=head2 ::device_wait

Wait until the device reaches a specific status.

=over

=item Arguments

B<status> => I<string>

I<Optional.>

The status to wait for. Default: ready

B<timeout> => I<integer>

I<Optional.>

Timeout in seconds. Default: 60

B<interval> => I<integer>

I<Optional.>

Interval between polling in secs. The default is to start from a low value and gradually
increase to a max of 1min interval. If set, the value does not change.

Default: 5

=item Examples

=back

	$self->device_wait;
	$self->device_wait(status=>"ready", interval=>30); # Fix polling of 30 sec for ready status.

=cut

sub device_wait {
	my ($self, %args) = @_;

	my $wait_status = $args{status} || ($self->{_nbu} ? "offline" : "ready");
	my $remain = $args{timeout} || 900;
	my $interval = $args{interval} || 5;

	my ($done, $curr_status);
	do {
		$curr_status = $self->device_status;
		if ( $curr_status =~ /$wait_status/i || $remain == 0 ) {
			$done = 1;
		} else {
			NOTE::logcomment($self->description.", Device status: $curr_status, "
					."waiting for $wait_status...");
			$remain -= sleep($interval);
			# Default is to start low and gradually increase to 60 secs
			$interval += 5 unless ($interval > 59 || exists $args{interval});
		};
	} while $remain > 0 && ! $done;
	croak ($self->description."\nTimeout waiting for device status ($wait_status). "
			."Last status: $curr_status.")
		unless $curr_status =~ /$wait_status/i;
	return 1;
}

=head2 ::device_rewind

Rewind the tape device to either the begining of the tape or previous file marker.

=over

=item Arguments

B<start_of_tape> => I<boolean>

I<Optional>

If set indicate to rewind to begining of the tape. Otherwise, the default, rewinds 1 file marker.

=item Examples

	$bu->device_rewind(); # Rewind 1 file marker
	$bu->device_rewind(start_of_tape=>1); # Rewind to beinging of the tape

=back

=cut

sub device_rewind {
	my ($self, %args) = @_;

	if ( $self->{_nbu} ) {
		NOTE::logcomment("For NBU job no device rewind needed for device ". $self->device_name);
		return;
	}
	if ( $self->device_name =~ /^(\/|r|\/dev\/r)/ ) {
		NOTE::logcomment("Tape device ".$self->device_name." does not need rewinding...");
		return;
	}
	if ( $args{start_of_tape} ) {
		NOTE::logcomment("Rewind device ".$self->device_name." to the begining of the tape...");
		return ImageBackup::device_rewind($self->storage->Filer->conn, $self->device_name);
	} else {
		NOTE::logcomment("Rewind device ".$self->device_name." by 1 file marker...");
		my ($rc, %info);
		%info = Tapeinfo::getTapeStatus($self->storage->Filer->conn, $self->{_device});
		if ( $info{fileno} > 1 ) {
			# Rewind 2 filemarkers and forward 1
			ImageBackup::device_rewind($self->storage->Filer->conn, $self->device_name, 2);
			$rc = ImageBackup::device_rewind($self->storage->Filer->conn, $self->device_name, -1);
		} else {
			# Rewind to begining of tape
			$rc = ImageBackup::device_rewind($self->storage->Filer->conn, $self->device_name);
		}
		return $rc;
	}
}

sub DESTROY {
	my $self = shift;

	return if ( $$ != $self->{_owner} );
	return if ( $self->{_destroyed} );
	return if ( NOTE->ALL_SPARED );

	if ( $self->{_spare_me} ) {
		NOTE::logocmment($self->description."\nHas been spared");
		return;
	}

	if ( exists $self->{_job_id} ) {
		NOTE::logcomment($self->description."\nAbort existing job");
		eval {
			$self->abort;
			$self->monitor;
		}
	}
	
	$self->{_destroy} = 1;
} 

1;

