
# $Id: //depot/prod/test/main/lib/NOTE/LUN.pm#1 $
#
## @summary  NOTE LUN class
## @status   Public
## @pod      here
## @author   indra@netapp.com, dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::LUN

=head1 DESCRIPTION

This package provides methods for creating and manipulating luns.
There are three constructor methods for lun objects;
C<::create>,
C<::attach>,
and
C<::new>.
The appropriate one to use depends on
the circumstances as explained for each method.

LUN objects support many of the same methods that Qtree objects
do.

=head1 METHODS

=cut

package NOTE::LUN;
our @ISA = qw(NOTE);         #FORME Modification to NOTE.pm (automatic include)

# NOTE: Do *NOT* import any external functions into this module!
# This is to avoid namespace collisions namely prototype mismatches and
# all that. Call the external function directly by prepending the package 
# name.

use Carp qw(croak);
use Params::Validate qw( validate validation_options SCALAR UNDEF BOOLEAN HASHREF );
use NOTE;
use strict;
use Data::Dumper;
use SayWrap qw(sayEOF);
use blocks_filer qw();
use Utils qw(make_host make_conn make_conn_and_complain);
use RichData qw(populate modify);
use Tharn;
use Net::Ping qw();

## Attributes
# _anonymous 
# _debug
# _destroyed
# _igroup
# _mntconfig 
# _name
# _file (if created from a file)
# _noreserve
# _online
# _ostype
# _owner
# _recycle 
# _size
# _spare_me
# _tag
# _teardown 
# _Volume
# _Qtree (if parent was a qtree)
# _Parent
# _was_recycled

## Exposed methods
# sub name;
# sub newName;   
# sub file;
# sub Volume ;   
# sub Parent;         
# sub Qtree;     
# sub Filer;     
# sub Aggregate; 
# sub size;      
# sub ostype;    
# sub mount_path; 
# sub description;
# sub mounted;    
# sub mountpoint; 
# sub new; 
# sub create; 
# sub attach; 
# sub offline; 
# sub online; 
# sub resize; 
# sub exists; 
# sub destroy; 
# sub SPARE; 
# sub DESTROY; 
# sub clone; 
# sub split; 
# sub populate; 
# sub modify; 
# sub mount;
# sub umount;
# sub devicename;
# sub dd;

## Unexposed methods
# sub _map;
# sub _unmap; 
# sub _mountfs; 
# sub _umount; 
# sub _disks;

BEGIN {
    our @EXPORT_OK = qw(
        name
        newName   
        file
        Volume    
        Parent         
        Qtree     
        Filer     
        Aggregate 
        size      
        ostype    
        mount_path 
        description
        mounted    
        mountpoint 
        new 
        create 
        attach 
        offline 
        online 
        resize 
        exists 
        destroy 
        SPARE 
        DESTROY 
        clone 
        split 
        populate 
        modify
        mount
        umount 
        devicename
        dd
    );
}


=head2 ::name

Returns the name of the LUN.

=over

=item Arguments

    None.

=item Example

=back

    $name = $lun->name;    

=cut

sub name      {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_name} }

=head2 ::newName

Returns a new name

=over

=item Argument

None.

=item Example

=back

	  $lun->newName;

=cut

sub newName   {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->SUPER::newName( "lun" ) }

=head2 ::file

Returns the path of the file (if created from one)

=over

=item Argument

None.

=item Example

=back

	  $lun->file;

=cut

sub file      {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_file} }

=head2 B<Related Objects>

=head2 ::Volume

Returns an object representing the volume containing the lun.

=over

=item Arguments

None.

=item Example

=back

     $vol = $lun->Volume;

=cut

sub Volume    {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_Volume} }

=head2 ::Parent

Returns an object representing the container of the LUN.

=over

=item Arguments

None.

=item Example

=back

     $parent = $lun->Parent;

=cut

sub Parent    {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_Parent} }

=head2 ::Qtree

Returns an object representing the Qtree containing the LUN.
If a qtree is not the container returns undef.

=over

=item Arguments

None.

=item Example

=back

     $qt = $lun->Qtree;

=cut

sub Qtree     { 
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    if (ref $self->Parent eq 'NOTE::Qtree') { 
        return $self->{_Qtree}  
    } else {
        croak "This LUN does not reside under a Qtree!";
    }
}

=head2 ::Filer

Returns an object representing the filer containing the LUN

=over

=item Arguments

None.

=item Example

=back

     $filer = $lun->Filer;

=cut

sub Filer     {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_Parent}->Filer }

=head2 ::Aggregate

Returns an object representing the aggregate
containing the volume containing the LUN.

Croaks if the containing volume is not a flex volume.

=over

=item Arguments

None.

=item Example

=back

     $aggr = $lun->Aggregate;

=cut

sub Aggregate {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    if ( $self->Volume->type ne 'flex' ) {
        croak "Container of ".$self->description." is not a flex volume.";
    }
    return $self->Volume->Aggregate;
}


=head2 ::size

Returns the size of the lun.

=over

=item Arguments

None

=item Examples

=back

	 $size = $lun->size;

=cut

sub size {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;    
    return blocks_filer::filer_lun_size_get(
                             $self->Filer->conn, 
                             $self->mount_path);
}

=head2 ::ostype

Returns the OS type of the lun.

=over

=item Arguments

None

=item Examples

=back

	 $ostype = $lun->ostype;

=cut

sub ostype    {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_ostype} }

=head2 ::mount_path

Returns the full path name of the LUN,
i.e. C</vol/><I<containing volume name>>C<[/>I<<Qtree name>>C<]/>I<<LUN nameE<gt>>

=over

=item Arguments

None.

=item Example

=back

     $mntpt = $lun->mount_path;

=cut

sub mount_path {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return $self->{_Parent}->mount_path."/".$self->{_name};
}

=head2 ::description

Returns a description of the lun consisting of
its filer's name, its mount path, and its tag, if any.

=over

=item Arguments

None.

=item Example

=back

     $desc = $lun->description;

=cut

sub description {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $txt;

    $txt = $self->{_Parent}->Filer->host.":".$self->mount_path;
    if ( $self->{_tag} ) {
        $txt .= " (".$self->{_tag}.")";
    }
    return $txt;
}

=head2 ::mounted

Returns the NFS mount point and NFS client of the lun.
Returns C<undef> if the lun is not mounted.

=over

=item Arguments

None.

=item Example

=back

	 %mntpt = $lun->mounted;
	 $client = $mntpt{'client'};
	 $mount_point = $mntpt{'mount'};

=cut

sub mounted {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_mntconfig} }

=head2 ::mountpoint

Returns the NFS mount point of the lun.
Returns C<undef> if the lun is not mounted.

=over

=item Arguments

None.

=item Example

=back

	 $mntpt = $lun->mountpoint;

=cut

sub mountpoint {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
   return shift->{_mntconfig}{'mount'} }

=head2 ::new

Creates an object representing a LUN,
but does B<not> cause the creation of a LUN on the filer.

=over

=item Arguments

All of the arguments to L<C<::create>|/::create> are also
valid arguments to C<::new> and have the same effect


=item Examples

=back

     $lun = NOTE::LUN->new(
         Parent => $vol,
         tag    => 'source_dataset'
     );

=cut

sub new {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($class,%args) = @_;
    my $self = {};

    exists $args{name}     and $args{name} = NOTE::option( $args{name} );
    exists $args{snaplock} and $args{snaplock} = NOTE::option( $args{snaplock} );

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    $self->{_tag}      = delete $args{tag};
    $self->{_teardown} = delete $args{teardown};
    $self->{_size}     = defined $args{size} ? delete $args{size}:"5m";
    $self->{_ostype}   = defined $args{ostype} ? delete $args{ostype}:"linux";

    # optional params
    $self->{_noreserve} = $args{noreserve};
    $self->{_recycle} = $args{recycle};

    if ( defined($args{Parent}) ) {
        if( ref $args{Parent} eq 'NOTE::Volume' ){
            $self->{_Volume} = $args{Parent};
            $self->{_Parent} = $self->{_Volume};
        } elsif (ref $args{Parent} eq 'NOTE::Qtree'){
            $self->{_Qtree}  = $args{Parent};
            $self->{_Volume} = $self->{_Qtree}->Volume;
            $self->{_Parent} = $self->{_Qtree};            
        } elsif (ref $args{Parent} eq 'HASH'){
            if( defined($args{Parent}->{Volume})) {
                $self->{_Volume} = NOTE::Volume->create(%{$args{Parent}->{Volume}});
                $self->{_Parent} = $self->{_Volume};                
            } elsif (defined($args{Parent}->{Qtree})){
                $self->{_Qtree} = NOTE::Qtree->create(%{$args{Parent}->{Qtree}});
                $self->{_Volume} = $self->{_Qtree}->Volume;
                $self->{_Parent} = $self->{_Qtree};
            } else {
                croak 'Unable to understand Parent argument'; 
            }
        } else {
            croak 'Parent is neither a Volume or a Qtree';
        }
    } else {
        croak "Parent (NOTE::Volume or NOTE::Qtree obj) not specified!";
    }

    bless $self, $class;
    NOTE::reference( $self );

    if ( $args{name} ) {
        $self->{_name} = $args{name};
        # Allow named luns that behave like anonymous luns:
        $self->{_anonymous} = (exists $args{anonymous}
                             ? $args{anonymous}
                             : $self->_looksAnonymous           #FORME Modification to NOTE.pm
                             );
        delete $args{name};
    } else { # Create a unique name
        $self->{_name} = NOTE::LUN->newName;
        $self->{_anonymous} = 1;
    }
    delete $args{anonymous};

    # If the luns isn't anonymous, make certain the parent
    # container is spared.
    if ( ! $self->{_anonymous} ) {
        $self->{_Parent}->SPARE;
    }

    return $self;
}

=head2 ::create

Creates a lun and returns
an object representing the created lun.

=over

=item Arguments

B<Parent> => I<object> | I<spec>

I<Required.>
The volume/qtree in which to create the LUN.
May be a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
object or a volume spec or 
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object or a qtree spec

B<name> => I<string>

I<Optional.>
The name of the LUN.
If absent, an anonymous LUN name will be generated.

B<tag> => I<string>

I<Optional.>
A text description of the object.
See L<C<::tag>|/::tag> for details.

=item Examples

=back

     $lun = NOTE::LUN->create(
         Parent => $vol,
         tag    => 'created using NOTE::Volume object'
     );

     $lun = NOTE::lun->create(
         Parent => { Volume => {name=>"vol0",
                                Filer => { host => $params->{FILER_SRC} }
                                }
                   }
         tag    => 'created using NOTE::Volume spec'
     );

=cut

sub create {
   my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;

    unless ( ref $self eq "NOTE::LUN" ) { # Already a lun object
        if ( $args{recycle} && $args{name} ) {
            $self = eval { NOTE::LUN->attach( %args ) };
            unless( $@ ) {
                NOTE::logcomment( "Re-using existing lun ".$self->Filer->host.":".$self->mount_path );
                $self->{_was_recycled} = 1;
                goto skipluncreate;
            }
            # Otherwise, create one
        }
        $self = NOTE::LUN->new( %args );
    }

    if(exists $args{'file'}){
        NOTE::logcomment( "Creating ".$self->{_ostype}." LUN ".$self->Filer->host.
                          ":".$self->mount_path." from file ".$args{'file'});
        if($self->{_noreserve}) {
            blocks_filer::filer_lun_create_from_file_with_noreserve( 
                             $self->Filer->conn,
                             $self->mount_path,
                             $args{'file'},
                             $self->{_ostype} );
        } else {
            blocks_filer::filer_lun_create_from_file( 
                             $self->Filer->conn,
                             $self->mount_path,
                             $args{'file'},
                             $self->{_ostype} );
        }
        $self->{_file} = $args{'file'};
    } else {
        NOTE::logcomment( "Creating ".$self->{_ostype}." LUN ".$self->Filer->host.
                             ":".$self->mount_path." (".$self->{_size}.")" );
        if($self->{_noreserve}) {
            blocks_filer::filer_lun_create_with_noreserve( 
                             $self->Filer->conn, 
                             $self->mount_path, 
                             $self->{_size}, 
                             $self->{_ostype} );
        } else {
            blocks_filer::filer_lun_create( 
                             $self->Filer->conn, 
                             $self->mount_path, 
                             $self->{_size}, 
                             $self->{_ostype} );
        } 
    }
    croak "Lun creation failed!" unless($self->exists);

skipluncreate:

    $self->{_recycle}  = delete $args{recycle};
    $self->{_online} = 1 if($self->exists);
    return $self;

}

=head2 ::attach

Creates an object representing an existing LUN.

=over

=item Arguments

B<name> => I<string>

I<Required.>
The name of an existing lun.

B<Parent> => I<object> | I<spec>

I<Required.>
The volume/qtree in which to create the LUN.
May be a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
object or a volume spec or 
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object or a qtree spec

=item Examples

=back

     $lun = NOTE::LUN->attach(
         Parent => $qt,
         name   => 'lun32'
     );

=cut

sub attach {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($class,%args) = @_;
    my $self = {};
    my $force = delete $args{force};

    $args{name} = NOTE::option( $args{name} ) if exists $args{name};

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;
    $self->{_tag}  = delete $args{tag};

    ## XXX Why defined? check this
    unless ( defined($args{name}) ) {
        croak "Attaching a lun requires a lun name";
    }

    $self->{_name} = $args{name};

    unless ( defined($args{'Parent'}) ) {
        croak "Attaching a lun requires a parent object (NOTE::Volume or NOTE::Qtree)";
    }

    if ( defined($args{Parent}) ) {
        if( ref $args{Parent} eq 'NOTE::Volume' ){
            $self->{_Volume} = $args{Parent};
            $self->{_Parent} = $self->{_Volume};
        } elsif (ref $args{Parent} eq 'NOTE::Qtree'){
            $self->{_Qtree}  = $args{Parent};
            $self->{_Volume} = $self->{_Qtree}->Volume;
            $self->{_Parent} = $self->{_Qtree};            
        } elsif (ref $args{Parent} eq 'HASH'){
            if( defined($args{Parent}->{Volume})) {
                $self->{_Volume} = NOTE::Volume->attach(%{$args{Parent}->{Volume}});
                $self->{_Parent} = $self->{_Volume};                
            } elsif (defined($args{Parent}->{Qtree})){
                $self->{_Qtree} = NOTE::Qtree->attach(%{$args{Parent}->{Qtree}});
                $self->{_Volume} = $self->{_Qtree}->Volume;
                $self->{_Parent} = $self->{_Qtree};
            } else {
                croak 'Unable to understand Parent argument'; 
            }
        } else {
            croak 'Parent is neither a Volume or a Qtree';
        }
    } else {
        croak "Parent (NOTE::Volume or NOTE::Qtree obj) not specified!";
    }

    bless $self, $class;
    unless ( $force || $self->exists ) {
        # $@ is not always set above when it should be, so here's a back up.
        croak "LUN ".$self->mount_path." does not exist";
    }
    $self->{_online} = 1;
    return $self;
}

=head2 ::offline, ::online

Sets volume offline and online respectively

=over

=item Arguments

None

=item Examples

=back

	 $lun->offline;

	 $lun->online;

=cut

sub offline {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    blocks_filer::filer_lun_offline( $self->Filer->conn, $self->mount_path);
    $self->{_online} = 0;
    return $self;
}

sub online {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    blocks_filer::filer_lun_online( $self->Filer->conn, $self->mount_path);
    $self->{_online} = 1;
    return $self;
}

=head2 ::resize

Resizes the lun with given size spec

=over

=item Arguments

I<size spec>

I<Required.>
A new size to make the lun

=item Examples

=back

	 $lun->resize('100m');
	 
=cut

sub resize {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,$newsize) = @_;
    blocks_filer::filer_lun_size_change( $self->Filer->conn, $self->mount_path, $newsize);
    $self->{_size} = $newsize;
    return $self;    
}

=head2 ::exists

Returns a boolean indicating is this lun exists on the filer

=over

=item Arguments

None.

=item Example

=back

	 $lun->exists;

=cut

sub exists {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return blocks_filer::filer_lun_exists( $self->Filer->conn,
                             $self->mount_path );
}

=head2 ::destroy

Destroys the lun and renders the lun object unuseable.
Unmaps and unmounts the lun if mounted.
=over

=item Arguments

None.

=item Example

=back

	 $lun->destroy;

=cut

sub destroy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self,%args) = @_;
    my $force = delete $args{'force'};
    NOTE::logcomment( "mount config " . $self->{_mntconfig});

    eval{ $self->_umount if( defined $self->{_mntconfig})};
    $self->{_mntconfig} = undef;
    eval{ $self->_unmap if( defined $self->{_igroup})};

    NOTE::logcomment( "Destroying LUN ".$self->description );
    blocks_filer::filer_lun_destroy( $self->Filer->conn, $self->mount_path);
    $self->{_destroyed} = 1;
    return $self;
}


=head2 ::SPARE

Don't destroy / recycle the LUN and it's container

=over

=item Arguments

None.

=item Example

=back

     $lun->SPARE;

=cut

sub SPARE {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

	my $self = shift;
	$self->{_recycle}  = 0;
	$self->{_spare_me} = 1;
        $self->{_Parent}->SPARE;
	return $self;
}

sub DESTROY {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    return if ( $$ != $self->{_owner});
    return if ( ! $self->{_anonymous} );
    return if ( $self->{_destroyed} );
    return if ( NOTE->ALL_SPARED );

    if ( $self->{_spare_me} ) {
        NOTE::logcomment( "LUN ".$self->description." has been spared");
        return;
    } else {
    	$self->destroy;
    }
}

sub _map {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my (%args) = @_;
    my $lun_map_status = undef;
    
    my $ostype = 'linux';
    my $igname = $args{'igroup'} || $args{'client'};
    my $ignode = $args{'initiator'};
    my $filer  = $self->Filer->conn;
    
    #Checking for iscsi license
    $self->Filer->check_license('iscsi');

    #Checking whether iscsi is running
    blocks_filer::filer_iscsi_start($filer);
    
    unless(blocks_filer::filer_igroup_is_defined($filer, $igname)){
        blocks_filer::filer_iscsi_igroup_create_for_ostype($filer, $igname, $ostype, $ignode);
    }
    
    eval {
        if($args{'lun'}) {
            blocks_filer::filer_lun_map(
                             $filer,
                             $self->mount_path,
                             $igname,
                             $args{'lun'});
        } else {
            $lun_map_status = blocks_filer::filer_lun_map_low(
                             $filer,
                             $self->mount_path,
                             $igname);            
        }
    };
    
    #croak "$@" unless ($@ =~ /LUN already mapped to this group/i);
    
    $self->{_igroup} = $igname;

    return $self;    
}

sub _unmap {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift; 
    blocks_filer::filer_lun_unmap($self->Filer->conn, $self->mount_path, $self->{_igroup});
    $self->{_igroup} = undef;
    return $self;
}


=head2 ::clone

Create a clone of this lun

=over

=item Arguments

B<name> => I<string>

I<Optional.>
The name of the clone

B<Parent> => I<object> 

I<Optional.>
The volume/qtree in which to create the clone.
May be a
L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>
object or  
L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>
object

B<snapshot> => I<string>

I<Optional.>
Backing snapshot to be used for cloning

B<splitclone> => I<boolean>

I<Optional.>
To split the clone or not

=item Example

=back

	 $lun_clone = $lun->clone;

=cut

sub clone {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my (%args) = validate (@_, {
                             name => {type=>SCALAR, default => $self->SUPER::newName("lun"),optional=>0},
                             Parent => {type=>HASHREF, default => $self->Parent, optional=>0},
                             snapshot => {type=>SCALAR, default => $self->Volume->snapshot_create, optional=>0},
                             splitclone => {type=>BOOLEAN, default=>1, optional=>0} });
    if(!defined $args{'snapshot'}){
        $args{'snapshot'} = $self->Volume->snapshot_create;
    }elsif(!$self->Volume->snapshot_exists($args{'snapshot'})){
        $self->Volume->snapshot_create($args{'snapshot'});
    }

    my $clone_path = $args{'Parent'}->mount_path."/".$args{'name'};
    
    blocks_filer::filer_lun_clone_create($self->Filer->conn,
                             $clone_path,
                             $self->mount_path, 
                             $args{'snapshot'});
    if($args{'splitclone'}) {                                       
        blocks_filer::filer_lun_clone_split_start($self->Filer->conn,
                             $clone_path);
    
        blocks_filer::filer_lun_clone_wait($self->Filer->conn,
                             $clone_path);
    } 
    return NOTE::LUN->attach(Parent => $args{'Parent'},
                             name => $args{'name'},);
}

=head2 ::split

splits this lun clone

=over

=item Arguments

None.

=item Example

=back

	 $lun_clone->split;

=cut


sub split {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    blocks_filer::filer_lun_clone_split_start($self->Filer->conn,
                         $self->mount_path);

    blocks_filer::filer_lun_clone_wait($self->Filer->conn,
                         $self->mount_path);
    return $self;
}

sub _mountfs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    croak "Cannot mount an offlined lun!" unless($self->{_online});

    my $sudo = "sudo";
    my $mount = "/bin/mount";
    my (%args) = @_;
    my $iscsid = defined $args{iscsid_init_path} ? $args{iscsid_init_path} : '/etc/init.d/iscsi';
    my $timeout = $args{timeout};
    my ($device, $iscsi_conf, @filer_ipv4_addr, $client_conn, $filer_ipv4_addr, $igroup_status,
        $discovery_pattern, $iscsi_cmd);

    # Making connection with the client
    if ( !defined($args{client})) {
       croak 'Pass a valid client name...';
    }
    NOTE::logcomment($args{client});
    if($args{client} eq 'localhost') {
        $client_conn = make_conn($args{client},"#rsh-persist");
    } else {
        NOTE::logcomment("the value :::". $args{client});
        $client_conn = make_conn($args{client},"#telnet");
        NOTE::logcomment(Dumper($client_conn));
    if(!defined $client_conn) {
        ($args{client}) = $args{client} =~/^([\w\d\-]*)\./m;
        $client_conn = make_conn($args{client},"#telnet");
    }
    }
    NOTE::logcomment(Dumper($client_conn));
    if(!defined $client_conn) {
         croak 'Not able to create a Client object... Make Sure the client and its tharnhost is proper with hosttype, rsh and telnet entries before execution'; 
    }
    # Checks whether iscsi initiator exists in the client or not
    my $out = SayWrap::sayEOF($client_conn, 'rpm -qa | grep iscsi');
    if ($out !~ /iscsi-initiator-utils/) {
        croak 'Iscsi initiator utils not configured in the client';
    }

    # Retrieving the ip which is needed to pass to the iscsi conf file with help of iscsi portal command

    my %portal_out = blocks_filer::filer_iscsi_portal_get( $self->Filer->conn );
    NOTE::logcomment(Dumper(\%portal_out));
    my @filer_ip_addr = keys %portal_out;
    foreach (@filer_ip_addr) {
      if($_=~/\d+\.\d+\.\d+\.\d+/) {
      push(@filer_ipv4_addr, $_);
      }
    }

    # Checking the filer whether pingable ip exist or not
    foreach my $ipv4 (@filer_ipv4_addr) {
     if(Net::Ping->new('external')->ping($ipv4)) {
           $filer_ipv4_addr = $ipv4;
           last;
     }
    }
   croak 'Need pingable ip other than management port ip to discover network for iscsi' if(!defined $filer_ipv4_addr);
   NOTE::logcomment("array ::: $filer_ipv4_addr");

    # Defining the variable iscsi_conf depending upon the rhel version
    my $version = SayWrap::sayEOF($client_conn, 'sudo cat /etc/redhat-*');
    if ( defined($args{iscsi_conf_path}) ) {
        $iscsi_conf = $args{iscsi_conf_path};
        $iscsi_cmd = "/sbin/iscsiadm -m node -p $filer_ipv4_addr -R";    
        $discovery_pattern = 'DiscoveryAddress = '. "$filer_ipv4_addr"; 
    } elsif ($version =~/release\s+5/) {
        $iscsi_conf = '/etc/iscsi/iscsid.conf';
        $iscsi_cmd = "/sbin/iscsiadm -m node -p $filer_ipv4_addr -R";
        $discovery_pattern = 'DiscoveryAddress = '. "$filer_ipv4_addr";
    } elsif ($version =~/release\s+4/) {
        $iscsi_conf = '/etc/iscsi.conf';
        $iscsi_cmd = "$iscsid reload";
        $discovery_pattern = 'DiscoveryAddress='. "$filer_ipv4_addr";
    } else {
        croak 'Pass either the valid path for Iscsi_conf_path argument or ' 
              .'client of rhel4 or 5 version. Version of the client passed :::'
              ." $version";
    }

    # Check whether the client is added as DiscoveryAddress in /etc/iscsi.conf of ISCSI HOST
    $out = SayWrap::sayEOF($client_conn, "$sudo egrep \"$discovery_pattern\" $iscsi_conf");
    if($out !~ /$filer_ipv4_addr/){
        SayWrap::sayEOF($client_conn, "$sudo echo $discovery_pattern >> $iscsi_conf");
        $out = SayWrap::sayEOF($client_conn, "$sudo egrep \"$discovery_pattern\" $iscsi_conf");
	croak "Unable to add DiscoveryAddress to $iscsi_conf" if($out !~ /$filer_ipv4_addr/);
    }

    # Discovering the target in rhel5 version
    if($version =~/release\s+5/) {
        SayWrap::sayEOF($client_conn, "$sudo /sbin/iscsiadm -m discovery -t st -p $filer_ipv4_addr --login", timeout => $timeout);
     sleep 5;
    } else {
        SayWrap::sayEOF($client_conn, "$sudo $iscsid restart", timeout => $timeout);

    }

    # Try iscsi reload/rescan for the new LUN to be recognised
    $out = SayWrap::sayEOF($client_conn, "$sudo $iscsi_cmd", timeout => $timeout);
    if(($out=~/FAILED/) && ($iscsi_cmd =~/reload/)) {
        SayWrap::sayEOF($client_conn, "$sudo $iscsid restart", timeout => $timeout);
    }
    sleep 5;

    # Check whether the client is logged in to ISCSI HOST
    my $i = 0;
    while ($i <= 4) {
    sleep 5;
    $igroup_status = $self->Filer->say('igroup show '.$self->{_igroup});
    if($igroup_status =~ /\(logged\s+in\s+on:/){
       last;
    } else {
        warn "igroup is not logged into iscsi host.. sleeping for 5 seconds before rechecking";
        sleep 5;
        $out = SayWrap::sayEOF($client_conn, "$sudo $iscsi_cmd", timeout => $timeout);
        sleep 5;
    }
    $i++;
    }
    $igroup_status = $self->Filer->say('igroup show '.$self->{_igroup});
    croak $self->{_igroup}." igroup is not logged into ISCSI Host: $igroup_status" if($igroup_status =~ /not logged in/);

    # Try iscsi reload/rescan for the new LUN to be recognised
    my $i = 0;
    while ($i <= 5) {
    $out = SayWrap::sayEOF($client_conn, "$sudo $iscsi_cmd", timeout => $timeout);
    sleep 5;
        if ($self->exists) {
        my @lun_dev = $self->_disks(client => $args{client}, sanlun_path=>$args{'sanlun_path'});
        NOTE::logcomment("Disk list: @lun_dev");
        $device = $lun_dev[0]; # pick first disk
    }
    if(defined $device) {
    last;
    } 
    NOTE::logcomment("New lun is not detected... sleeping for 5 seconds before rechecking");
    sleep 5; 
    $i++;
    }
    croak "Device list is empty! Lun Undetected" unless($device);
    NOTE::logcomment("Device is... $device");
   
    #Check whether chosen device is already mounted
    my $sayout = SayWrap::sayEOF($client_conn, "$sudo $mount | grep $device");
    logcomment("Device ".$device." out: ".$sayout);
    croak "Device already mounted: $sayout" if($sayout =~ /($device)/);

    #Check whether file system exist in the chosen device or not 
    my $sayout = SayWrap::sayEOF($client_conn, "$sudo /sbin/dumpe2fs $device | grep UUID");
    logcomment("Device ".$device." out: ".$sayout);
    if($sayout !~ /UUID/) {
        # Make the file system
        Tharn::say($client_conn, "$sudo /sbin/mkfs -t ext2 $device", '-prompt', "Proceed anyway\?");
        my (undef, undef, $out) = Tharn::say($client_conn, "y", '-timeout', $args{timeout});
        croak("$out\n") if ($out =~ /ERROR/i);
    } else {
        NOTE::logcomment('Skipping the file system creation... since fs already exist');
    }

    # Mount the filesystem
    my $mountdir = "/t/".$args{client}."/dev_".$self->name;
    $sayout = SayWrap::sayEOF($client_conn, "$sudo mkdir -p $mountdir");
    croak $sayout if($sayout =~ /ERROR|FAILED/i);
    $sayout = SayWrap::sayEOF($client_conn, "$sudo $mount -t ext2 $device $mountdir");
    croak $sayout if($sayout =~ /ERROR|FAILED/i);
    $sayout = SayWrap::sayEOF($client_conn, "$sudo chmod 777 $mountdir");
    croak $sayout if($sayout =~ /ERROR|FAILED/i);

    $self->{_mntconfig} = {%args, mount => $mountdir};
    NOTE::logcomment("Mount Point is ".$self->{_mntconfig}{'mount'});
    NOTE::logcomment("client is ".$self->{_mntconfig}{'client'});

    return $self;
}

sub _umount {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $sudo = "sudo";
    my $umount = "/bin/umount";
    my $mount_point = $self->{_mntconfig}{'mount'};
    my $client_conn = make_conn($self->{_mntconfig}{'client'});
    
    my $out = SayWrap::sayEOF($client_conn, "$sudo $umount -f $mount_point");
    croak $out if($out =~ /ERROR|FAILED/i);
    $self->{_mntconfig}{'mount'} = undef;
}

=head2 ::populate

Populate lun with data using RichData::populate.

=over

=item Arguments

B<initiator> - I<string>

I<Required.>
Node name of the ISCSI initiator

B<igroup> - I<string>

I<Required.>
Name of the ISCSI igroup to be created on Filer
Uses the same if this igroup is existing -  doesnot verify 
whether the node name is the same as given initiator.

B<client> - I<string>

I<Optional.>
Name of the ISCSI linux host. 
This module currently supports only linux hosts

B<timeout> - I<string>

I<Optional.>
Timeout for the operation in seconds

B<sanlun_path> - I<string>

I<Optional.>
Path of the sanlun binary in the ISCSI host m/c.

B<pop_opts> - I<RichData argument Hash>

Optional.

All parameters, in the form of a hash is passed to RichData::populate for processing.

=item Examples

=back

	  $lun->populate;

	  $lun->populate("file_dev" => 100, "dir_avg" => 3);

=cut

sub populate {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my (%args) = validate (@_, { initiator => {type=>SCALAR, optional=>0}, 
                             igroup => {type=>SCALAR, optional=>1},
                             client => {type=>SCALAR|UNDEF, default=>"localhost", optional=>1},
                             timeout => {type=>SCALAR|UNDEF, default=>600, optional=>1},
                             pop_opts => {type => HASHREF, default => undef, optional => 1},
                             sanlun_path => {type=>SCALAR|UNDEF, default=> '/usr/sbin', optional=>1,
                             regex=>qr/^(\/\w+)+$/ }});
    $self->_map(%args);
    $self->_mountfs(%args);
    my $mount_point = $self->{_mntconfig}{'mount'};
    my $client = $self->{_mntconfig}{'client'};

    RichData::populate(path => $mount_point."/data", client => $client, %{$args{'pop_opts'}});

   my $client_conn = make_conn($args{client},"#telnet");

   # Added the below lines of code ie., executing the ls command only for
   # debug purpose
   my $sout = SayWrap::sayEOF($client_conn, "ls $mount_point/data");
   NOTE::logcomment("content ::: $sout");

    eval{ $self->_umount if( defined $self->{_mntconfig})};
    return $self;
}

=head2 ::modify

Modify lun data using RichData::modify.

=over

=item Arguments

B<pop_opts> - I<RichData argument Hash>

Optional.

All parameters, in the form of a hash is passed to RichData::populate for processing.

=item Examples

=back

	  $lun->modify;

=cut

sub modify {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = @_;

    $self->_mountfs(%{$self->{_mntconfig}});
    
    my $mount_point = $self->{_mntconfig}{'mount'};
    my $client = $self->{_mntconfig}{'client'};
    NOTE::logcomment("the client is $client and $mount_point");

    my $client_conn = make_conn($self->{_mntconfig}{'client'},"#telnet");

    RichData::modify(path => $mount_point."/data", client => $client, %args);

    # Added the below lines of code ie., executing the ls command only for 
    # debug purpose
    my $sout = SayWrap::sayEOF($client_conn, "ls $mount_point/data");
    NOTE::logcomment("content ::: $sout");

    eval{ $self->_umount if( defined $self->{_mntconfig})};
    return $self;
}

sub _disks{
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my (%args) = validate(@_, { client => {type=>SCALAR, default=>"localhost", optional=>0},
                             sanlun_path => { type=>SCALAR, optional=>1, default=>'/usr/sbin', 
                                regex=>qr/^(\/\w+)+$/ },
                             });

    NOTE::logcomment($args{'client'});
    my $sudo = "/usr/local/bin/sudo";
    host($args{'client'});
    my $client_conn = connect("#rsh-persist");
    my $filer = $self->Filer->host;
    my $lun_path = $self->mount_path;

    NOTE::logcomment(Dumper($client_conn)); 
    # do multiple times since sanlun fails on first attempt and then passes on linux iscsi
    sayEOF($client_conn, "$sudo $args{sanlun_path}/sanlun lun show $filer:$lun_path");
    sleep 5;
    sayEOF($client_conn, "$sudo $args{sanlun_path}/sanlun lun show $filer:$lun_path");
    sleep 5;
    my $out = sayEOF($client_conn, "$sudo $args{sanlun_path}/sanlun lun show $filer:$lun_path");
    croak("Host attach kit 'sanlun' missing on ".$args{'client'}.": $out\n") 
        if ($out =~ /command not found/i);
        
    my @lun_dev;
    while ($out =~ /$filer:\s+$lun_path\s+(\/dev(\/\w+)+)\s+.*GOOD/gm) {
        push(@lun_dev,$1) if ($1);
    }
    unless(@lun_dev) {
	while ($out =~ /$filer\s+$lun_path\s+(\/dev(\/\w+)+)\s+.*/gm) {
            push(@lun_dev,$1) if ($1);
        }
    }

    return(@lun_dev);
}

=head2 ::rename

Changes the name of the LUN.

=over

=item Arguments

I<string>

I<Optional.>
The new name for the Lun.
If absent a new anonymous name is generated
and the lun becomes anonymous.
If present the lun will no longer be anonymous,
causing its container to be spared.

=item Examples

=back

         $lun->rename;
         $lun->rename( 'newName' );

=cut

sub rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        my ($self,$newname) = @_;
        my $oldname = $self->{_name};
        if ( defined $newname ) {
                $self->_rename( $newname );
                if ( $self->{_anonymous} && !$self->_looksAnonymous ) {
                    # Converting from anonymous to non-anonymous
                    $self->SPARE;
                    $self->{_anonymous} = 0;
                }
        } else {
                $newname = $self->newName;
                $self->{_anonymous} = 1;
                $self->_rename( $newname );
        }
        return $oldname;
}

sub _rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

        my ($self,$newname) = @_;
        blocks_filer::filer_lun_move( $self->Filer->conn, $self->mount_path, $self->Parent->mount_path."/".$newname );
        croak $@ if $@;
        $self->{_name} = $newname;
}

=head2 ::mount

Mounts the lun and returns the object with mount_point.
This method doesnot support 2 mounts for writing the data in rhel4 clients. 
In such case use dd function to create/modify data else get the current 
devicename of the lun using devicename method and write data. Writing the data
by mounting 2 or more luns at the same time is supported only with rhel5 client. 
=over

=item Arguments

B<initiator> - I<string>

I<Required.>
Node name of the ISCSI initiator

B<igroup> - I<string>
I<Required.>
Name of the ISCSI igroup to be created on Filer
Uses the same if this igroup is existing -  doesnot verify
whether the node name is the same as given initiator.

B<client> - I<string>

I<Optional.>
Name of the ISCSI linux host.
This module currently supports only linux hosts

B<timeout> - I<string>

I<Optional.>
Timeout for the operation in seconds

B<iscsid_init_path> - I<string>

I<Optional.>
Path of the iscsi init binary in the ISCSI host m/c.

B<iscsi_conf_path> - I<string>

I<Optional.>
Path of the iscsi conf binary in the ISCSI host m/c.

B<sanlun_path> - I<string>

I<Optional.>
Path of the sanlun binary in the ISCSI host m/c.

=item Examples

=back

    $lun->mount(initiator => initiator name,
                client => client_name,);

    $lun->mount(initiator => initiator name,
                client => client_name,
                igroup => igroup name);

=cut

sub mount {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
 
    my $self = shift;
    my (%args) = validate (@_, { initiator => {type=>SCALAR, optional=>0},
                             igroup => {type=>SCALAR, optional=>1},
                             client => {type=>SCALAR, optional=>0},
                             timeout => {type=>SCALAR|UNDEF, default=>600, optional=>1},
                             iscsid_init_path => {type=>SCALAR|UNDEF, default=> '/etc/init.d/iscsi', optional=>1, regex=>qr/^(\/\w+)+$/},
                             iscsi_conf_path => {type=>SCALAR|UNDEF, optional=>1, regex=>qr/^(\/\w+)+$/},
                             sanlun_path => {type=>SCALAR|UNDEF, default=> '/usr/sbin', optional=>1,
                             regex=>qr/^(\/\w+)+$/ }});

  if($args{client} eq 'localhost') {
      croak 'Pass a valid client name other than localhost ...';
  }

   $self->_map(%args);
   $self->_mountfs(%args);
   my $mount_point = $self->{_mntconfig}{'mount'};
   my $client = $self->{_mntconfig}{'client'};
   NOTE::logcomment("Mount Point is $mount_point, $client");
   return $self;
}

=head2 ::devicename

Executes the sanlun lun show command to get the current device name on which lun
is mapped. Returns the device name only when the lun is mounted to that client
else it will fails with empty device list

=over

=item Arguments

B<client> = <client on which LUN is mounted>

I<Required> Client on which LUN is mounted 

B<sanlun_path> = <path of sanlun>

I<Optional> Path of the sanlun command. By default,
'/usr/sbin' is the path in which sanlun command resides

=item Examples

=back

    $lun->devicename(client => client_name,);

    $lun->dd (client => client_name,
              sanlun_path => '/usr/sbin');

=cut
sub devicename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my (%args) = validate (@_, { client => {type=>SCALAR, optional=>0},
                             sanlun_path => {type=>SCALAR|UNDEF, default=> '/usr/sbin', optional=>1,
                             regex=>qr/^(\/\w+)+$/ }});

   my $device;
        if ($self->exists) {
            my @lun_dev = $self->_disks(client => $args{client}, sanlun_path=>$args{'sanlun_path'});
            NOTE::logcomment("Disk list: @lun_dev");
            $device = $lun_dev[0]; # pick first disk
    }
    croak "Device list is empty! Lun Undetected" unless($device);
    NOTE::logcomment("Device is... $device");
    $self->{_mntconfig}{'device'}= $device;
    return $self->{_mntconfig}{'device'};
}

=head2 ::dd

Executes the unix dd command to fill the LUN with data. Writes to the device
file the LUN has been mapped to.

=over

=item Arguments

B<client> = <client to which LUN is mapped>

I<Required> The Client to which LUN is mapped.

B<count> = <number_of_blocks>

I<Required> Number of blocks to write

B<input_file> = <file_name>

I<Optional> Input file on client from where the blocks are read. By default,
'/dev/zero' file is taken as input file

B<seek> = <offset>

I<Optional> offset to modify the blocks. By default, offset is considered as
zero.

B<block_size> = <block_size>

I<Optional> Block Size. Default is 4096 bytes.

B<conv> = <keyword>

I<Optional> Conv keyword for dd command. By default, it is 'notrunc'.

=item Examples

=back

    $lun ->dd (client => client_name,
               count  =>  4081,);

    $lun->dd (client => client_name,
              count       =>  2000,
              seek        =>  2000);

=cut

sub dd {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

     my ($self,%args) = @_;

    # Check if all required parameters are present.
    if (!exists($args{client}) || !exists($args{count})){
        croak 'Client and count are required parameter for LUN::dd';
    }

    my $device_name = $self->devicename(client => $args{client});

    # Create a client object
    my $client = NOTE::Client->new(storage => $self->Volume);

    # Append the mount point of Volume to get absolute path of the output file.
    $args{output_file} = $device_name;

    # Delete the client parameter before invoking dd method from Client.pm
    delete $args{client};

    # Invoke the client dd command
    $client->dd(%args);

} ## end sub dd

=head2 ::umount

Umounts the Lun.

=over

=item Arguments

None.

=item Example

=back

     $lun->umount;

=cut

sub umount {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    eval{ $self->_umount if( defined $self->{_mntconfig})};
    $self->{_mntconfig} = undef;
}

=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::agnostic::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::FileSystem::NFS>|engineering::ontap_tests::agnostic::docs::lib-NOTE-FileSystem-NFS-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Volume-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::agnostic::docs::lib-NOTE-Qtree-pm/>

=head1 STATUS

Public.

=head1 AUTHOR

indra@netapp.com

=cut

1;
