# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/LockVault.pm#1 $
##
## @summary note lockvault class
##
## @status public
##
## @pod    here
##
## @author sydneyw@netapp.com, dl-dpgqs-automation@netapp.com
##

=head1 NAME

NOTE::LockVault

=head1 DESCRIPTION

this is the note class for creating and maniplating lockvault relationships.
it's methods are mostly wrapper methods for the lib/snaplock.pm library

=head1 METHODS

=cut

use NOTE;
use POSIX qw();
package NOTE::LockVault;
our @ISA = qw(NOTE);

use strict;
use Carp;
use File::Find;
use FSR::FSRutils qw (subtestResults);
use snaplock qw();
use SnapVault qw();
use SayWrap qw();

=head2 ::new

creates an object representing a lockvault relationship,
but does B<not> cause the creation of the lockvault relationship on the filer.

actions include:
- initializing the compliance clock unless already initialized
- setting snaplock licenses for src and dst filers
- setting retention periods for lockvult objects
- setting the lockvault log volume

=over

=item arguments

B<src> => I<qtree object>

a qtree object to be used as the lockvault source.

B<dst> => I<qtree object>

a qtree object to be used as the lockvault destination.

B<log> => I<volume object>

a volume object to be used as the lockvault log volume

B<snaplock_minimum_period> => I<time spec>

I<optional.>
the lockvault minimum retention period. the specification is given
as a number vault followed by <s|m|h> for seconds, minutes or hours
respectively. the default is 60s.

B<snaplock_maximum_period> => I<time spec>

I<optional.>
the lockvault maximum retention period. defaults to minimum retention
setting

B<snaplock_default_period> => I<min|max|time spec>

I<optional.>
the lockvault default retention period. can be set to 'min' or 'max'
to signify using the minimum and maximum retention spec. respectively,
or it could have given a time spec.

=item Examples

=back

    $lv = NOTE::SnapMirror->new(src=>$src_qt, dst=>$dst_qt, log=>$log_vol);

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    $self->{_src} = $args{src};
    $self->{_dst} = $args{dst};
    $self->{_log} = $args{log};

    croak "src must be a NOTE::Qtree object"
	unless $self->{_src}->isa("NOTE::Qtree");
    croak "dst must be a NOTE::Qtree object"
	unless $self->{_dst}->isa("NOTE::Qtree");
    croak "log must be a NOTE::Volume object"
	unless $self->{_log}->isa("NOTE::Volume");
    croak "dst and log must be on the same filer"
	unless $self->{_dst}->Filer->hostspec eq $self->{_log}->Filer->hostspec;
    croak "dst must be on a compliance or enterprise volume"
	unless _snaplock_type($self->{_dst}) =~ /compliance|enterprise/i;
    croak "log must be same the lockvault type as dst"
	unless _snaplock_type($self->{_dst}) eq $self->{_log}->snaplock;

    $self->{_owner} = $$;

    bless $self, $class;

    snaplock::compliance_clock_initialize($self->{_dst}->Filer->conn, reinit_ok=>1);

    NOTE::logcomment("set src filer lockvault licenses");
    $self->_set_lv_licenses("pri");
    NOTE::logcomment("set dst filer lockvault licenses");
    $self->_set_lv_licenses("sec");

    # check src filer access permission for dst filer and set if needed
    $self->_set_access($self->{_src}->Filer, $self->{_dst}->Filer);
    # set the retention periods
    my $min = $args{snaplock_minimum_period} || "60s"; # defaults to 1 minute
    my $max = $args{snaplock_maximum_period} || $min;
    my $def = $args{snaplock_default_period} || "min";
    $self->{_retention_period} = ( $def eq "min"
					? $min
					: ( $def eq "max" ? $max : $def )
				);
    _set_vol_retention($self->{_dst}, $min, $max, $def);
    _set_vol_retention($self->{_log}, $min, $max, $def);

    $self->set_log_volume;

    return $self;
}

sub _snaplock_type {
    my ($obj) = @_;

    return ( $obj->isa("NOTE::Volume")
	? $obj->snaplock : $obj->Volume->snaplock );
}

=head2 ::set_log_volume

sets the lockvault log volume.

=over

=item B<new_log>

I<optional.>
the internal log volume object is used to set the log volume log unless already
set. if a new log volume object is supplied, it will replace the internal log
object with the given object.

=item Example

=back

     $lv->set_log_volume($new_log_vol);

=cut

sub set_log_volume {
    my ($self, $new_log) = @_;

    if ( defined $new_log ) {
	# set new log volume if supplied
	_same_type($self->{_log}, $new_log) if exists $self->{_log};
	$self->{_log} = $new_log;
    }

    croak "LockVault log volume not set"
	unless exists $self->{_log};

    return $self->_set_log_vol($self->{_log}->Filer->conn, $self->{_log}->name);
}

sub _set_log_vol {
    my ($self, $fconn, $log_vol_name) = @_;

    my $key = "snapvault.lockvault_log_volume";
    my $log_vol = _option_value($fconn, $key);
    # check if current log vol is the same
    if ($log_vol_name eq "''" && $log_vol eq "" || $log_vol eq $log_vol_name) {
	NOTE::logcomment("log volume already set, ignoring...");
    } else {
	my $text = SayWrap::sayEOF ($fconn, "options $key ".$log_vol_name);
	$log_vol = _option_value($fconn, $key);
	croak "Failed to set $key to volume ".$log_vol_name
		.", current value is $log_vol. error: $text"
		unless ($log_vol_name eq "''" && $log_vol eq "" || $log_vol eq $log_vol_name);
	delete $self->{_xfer}; # clear out stored xfer log info
	NOTE::logcomment("log volume set to $log_vol");
    }
    return 1;
}

=head2 ::attach

calls new to create a new lockvault object without starting the lockvault
arguments are the same as those for method B<new>.

=over

=item Example

=back

     NOTE::LockVault->attach(src=>$src_qt, dst=>$dst_qt, log=>$log_vol);

=cut

sub attach {
    my ($class, %args) = @_;
    my $self = NOTE::LockVault->new(%args);

    return $self;
}

=head2 ::start

creates a lockvault relationship out of the supplied objects.
calls method B<new> to create a lockvault object, unless it is already an object.
and then calls 'snapvault start' to create the baseline lockvault transfer.

parameters are the same as those to method B<new>, in addition to parameters passed
to the lower level function.

=over

=item %args

B<src> => I<src qt object>

The lockvault src NOTE::Qtree object

B<dst> => I<dst qt object>

The LockVault dst NOTE::Qtree object

B<log> => I<log vol object>

The LockVault log NOTE::Volume object

B<snaplock_minimum_period> = I<time>

I<Optional.>
Minimum retention period. Time given is seconds or any units of m(inutes) or h(ours).
Default 60s

B<snaplock_maximum_period> = I<time>

I<Optional.>
Maximum retention period. Default snaplock_minimum_period

B<snaplock_default_period>

I<Optional.>
Default retention period. Default snaplock_minimum_period.

=item Examples

=back

    my $lv = NOTE::LockVault->start(src=>$src_qt, dst=>$dst_qt, log=>$log_vol);

    my $lv = NOTE::LockVault->start(src=>$src_qt, dst=>$dst_qt, log=>$log_vol,
                   snaplock_minimum_period=>"5m");

=cut

sub start {
    my ($class, %args) = @_;
    my $self = {};

    if ( ref $class ne '' ) {
	$self = $class;
	croak "Cannot override existing lockvault 'src', 'dst' or log' attribute"
		if ( exists $args{src} || exists $args{dst} || exists $args{log} );
    } else {
	$self = NOTE::LockVault->new(%args);
    }
    NOTE::reference($self);

    # initiate snapvault transfer in subtest
    my $op = "Snapvault Start".$self->description;
    NOTE::logcomment($op);
    my %sv_args;
    $sv_args{SOURCE} = $self->src_path;
    $sv_args{THROTTLE} = $args{THROTTLE} if $args{THROTTLE};
    $sv_args{TRIES_COUNT} = $args{TRIES_COUNT} if $args{TRIES_COUNT};
    $sv_args{TIMEOUT} = $args{TIMEOUT} if $args{TIMEOUT};
    $sv_args{COMPLETE} = 0;
    my $ret = SnapVault::sv_start($self->{_dst}->Filer->conn, $self->{_dst}->mount_path, %sv_args);
    croak $@ if $@;
    # monitor transfer until completion unless 'background' option set
    push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    $self->monitor()
	unless $args{background};
    return $self;
}

sub _set_vol_retention {
    my ($obj, $min, $max, $def) = @_;

    my $vol = ($obj->isa("NOTE::Volume") ? $obj : $obj->Volume);
    $vol->{'options.snaplock_minimum_period'} = $min;
    $vol->{'options.snaplock_maximum_period'} = $max;
    $vol->{'options.snaplock_default_period'} = $def;
}

sub _set_lv_licenses {
    my ($self, $role) = @_;

    my $tag = ($role =~ /pri/i ? "_src" : "_dst");
    $self->{$tag}->Filer->check_license("nfs");
    $self->{$tag}->Filer->check_license("snapmirror");
    $self->{$tag}->Filer->check_license("sv_ontap_".$role);
}

=head2 ::log_expiration

returns the expected log expiration time based on the atime of the
supplied log file

=over

=item file

log file of the expected retention expiration.

=item Example

=back

     my $expected_expire_time = $lv->log_expiration($history_log_file);

=cut

sub log_expiration {
    my ($self, $file) = @_;

    croak "LockVault log volume not set"
	unless exists $self->{_log};
    my ($dev, $ino, $mode, $nlnk, $uid, $gid, $rdev, $size, $atime)
	= stat $file;
    return $atime;
}

=head2 ::expiration_time

Expiration time using retention period from compliance time

=over

=item Example

=back

     my $expected_expiration = $lv->expiration_time;

=cut

sub expiration_time {
    my ($self) = @_;

    return $self->compliance_time + $self->retention_period;
}

=head2 ::retention_period

Returns retention_period in seconds of this lockvault.
retention_period is defined by snaplock_default_period
when LockVault object is created

=over

=item Example

=back

      my $period = $lv->retention_period;

=cut

sub retention_period {
    my ($self) = @_;

    croak "LockVault retention_period not set"
	unless exists $self->{_retention_period};
    return snaplock::parse_retention($self->{_retention_period});
}

=head2 ::compliance_time

current compliance time in secs since epoch.

=over

=item Example

=back

      my $comp_time = $lv->compliance_time

=cut

sub compliance_time {
    my ($self) = @_;

    croak "LockVault dst volume not set"
	unless exists $self->{_dst};
    return snaplock::compliance_clock_status($self->{_dst}->Filer->conn);
}

=head2 ::history_logs

Returns a list of history logs relative to log volume's 'logroot' dir

=over

=item Example

=back

      my @logs = $lv->history_logs;

=cut

sub history_logs {
    my ($self) = @_;

    croak "LockVault log volume not set"
	unless exists $self->{_log};

    my @list;
    opendir logdir, $self->logroot()
	or croak "Failed to open history log dir ".$self->logroot;
    foreach my $f (readdir logdir) {
	# files in this directory represent history logs
	# Add in reverse order so the first one represent the latest history log
	push @list, $f
		if -f $self->logroot().$f;
    }
    closedir logdir;
    NOTE::logcomment("History Log(s) found:\n	".(join "\n	", @list));
    return @list;
}

=head2 ::transfer_logs

Returns a list of transfer logs relative to log volume's 'logroot' dir

=over

=item arguments

B<type>

I<optional>

type of the logs to be return:
	'all' - all transfer logs,
	'old' - existing transfer logs
	'new' - created since last call
	'del' - removed since last call
default: 'all'

=item Examples

=back

     my $all_xfr_logs = $lv->transfer_logs;

     my $new_since_last_call = $lv->transfer_logs(type=>'new');

=cut

sub transfer_logs {
    my ($self, %args) = @_;

    $args{type} = "all" unless exists $args{type};
    croak "Type must be: 'all', 'old', 'new' or 'del'"
	unless $args{type} =~ /all|old|new|del/;

    croak "LockVault log volume not set"
	unless exists $self->{_log};

    $self->log->umount; # forces remounting to get most up to date info

    # replace entries in 'del' list with entries from 'old' and 'new'
    # move entries from 'del' to 'old' during the search
    # so that at the end, remaining entries in 'del' will be deleted ones
    my %del;
    foreach my $f ( @{$self->{_xlog}{new}}, @{$self->{_xlog}{old}} ) {
	$del{$f} = 1;
    }
    delete $self->{_xlog};
    my $wanted = sub { my $f = $File::Find::name;
			if ( -f && $f =~ /snapvault_filelog/ ) {
				$f =~ s/^.*\/etc\/log\///;
				if ( exists $del{$f} ) {
					# move to old if entry exists in 'del' list
					push @{$self->{_xlog}{old}}, $f;
					delete $del{$f};
				} else {
					push @{$self->{_xlog}{new}}, $f;
				}
			}
		};
    find ($wanted, $self->logroot);
    push @{$self->{_xlog}{del}}, keys %del;
    my @list;
    foreach my $type (keys %{$self->{_xlog}}) {
	push @list, @{$self->{_xlog}{$type}}
		if $args{type} =~ /all/i || $args{type} =~ /$type/;
    }
    NOTE::logcomment("For type ".$args{type}." found ".@list." xfer log(s):\n"
			.(join "	\n", @list));

    return @list;
}

=head2 ::logroot

The root directory of the lockvault's logs

=over

=item Example

=back

     my $log_root_dir = $lv->logroot;

=cut

sub logroot {
    my ($self) = @_;

    croak "LockVault log volume not set"
	unless exists $self->{_log};

    $self->{_log}->mount
	unless $self->{_log}->mounted;

    return $self->{_log}->mountpoint."/etc/log/";
}

=head2 ::src

Sets or returns this lockvault's src object
Returns the old src object if a new one is being set.

=over

=item $newsrc

I<Optional.>

If supplied will set LockVault src to this NOTE::Qtree object

=item Examples

=back

     my $src = $lv->src;

     my $old_src = $lv->src($new_src_qt);

=cut

sub src {
    my ($self, $newsrc) = @_;

    if ( defined $newsrc ) {
	croak "New src must be a NOTE::Qtree object"
		unless $newsrc->isa("NOTE::Qtree");
	_same_type($self->{_src}, $newsrc) if exists $self->{_src};

	my $oldsrc = $self->{_src};
	$self->{_src} = $newsrc;
	return $oldsrc;
    } else {
	return shift->{_src};
    }
}

=head2 ::dst

Sets or returns this lockvault's dst object.
Returns the old dst object if a new one is being set.

=over

=item $newdst

I<Optional.>

If supplied will LockVault dst to this NOTE::Qtree object

=item Examples

=back

     my $dst = $lv->dst;

     my $old_dst = $lv->dst($new_dst_qt);

=cut

sub dst {
    my ($self, $newdst) = @_;

    if ( defined $newdst ) {
	croak "New dst must be a NOTE::Qtree object"
		unless $newdst->isa("NOTE::Qtree");
	_same_type($self->{_dst}, $newdst) if exists $self->{_dst};
	croak "New object must be ".$self->{_log}->snaplock
		unless _snaplock_type($newdst) eq $self->{_log}->snaplock;

	my $olddst = $self->{_dst};
	$self->{_dst} = $newdst;
	return $olddst;
    } else {
	return shift->{_dst};
    }
}

sub _same_type {
    my ($old_obj, $new_obj) = @_;

    my $exp_type = ($old_obj->isa("NOTE::Volume") ? "NOTE::Volume" : "NOTE::Qtree");
    croak "New object must be a $exp_type object"
	unless $new_obj->isa($exp_type);
}

=head2 ::log

Sets or returns this lockvault's log object
Returns the old log object if a new one is being set.

=over

=item $newlog

I<Optional.>

If supplied will LockVault log to this NOTE::Volume object

=item Examples

=back

     my $log = $lv->log;

     my $old_log = $lv->log($new_log_vol);


=cut
 
sub log {
    my ($self, $newlog) = @_;

    if ( defined $newlog ) {
	croak "New log must be a NOTE::Volume object"
		unless $newlog->isa("NOTE::Volume");
	my $oldlog = $self->{_log};
	$self->{_log} = $newlog;
	$self->set_log_volume;
	return $oldlog;
    } else {
	return shift->{_log};
    }
}

=head2 ::src_path

Returns <filer>:<filer path> of the src object

=over

=item Example

=back

     my $path = $lv->src_path;

=cut

sub src_path {
    my $self = shift;

    croak "LockVault src object not set"
	unless exists $self->{_src};

    return $self->{_src}->Filer->host.":".$self->{_src}->mount_path;
}

=head2 ::dst_path

Returns <filer>:<filer path> of the dst object

=over

=item Example

=back

      my $path = $lv->dst_path;

=cut

sub dst_path {
    my $self = shift;

    croak "LockVault dst object not set"
	unless exists $self->{_dst};

    return $self->{_dst}->Filer->host.":".$self->{_dst}->mount_path;
}

=head2 ::log_path

Returns <filer>:<filer path> of the log volume

=over

=item Example

=back

     my $path = $lv->log_path;

=cut

sub log_path {
    my $self = shift;

    croak "LockVault log object not set"
	unless exists $self->{_log};

    return $self->{_log}->Filer->host.":".$self->{_log}->mount_path;
}

=head2 ::description

Returns a description of the lockvault relationship

=over

=item Example

=back

      my $lv_desc = $lv->description;

=cut

sub description {
    my ($self) = @_;

    return "\n   +-- "
	   .$self->src_path.($self->src->tag ? " (".$self->src->tag.") " : "")
	   ."\n   +-> "
	   .$self->dst_path.($self->dst->tag ? " (".$self->dst->tag.") " : "")
	   ."\n   "
	   ."Log Volume: ".$self->log_path."\n";
    ;
}

=head2 ::update

Perform a LockVault update

=over

=item Arguments

B<background>

I<Optional>

Boolean representing whether to run the update in the background
Defaults 0.

Perform update in background

=item Example

=back

     $lv->update;

=cut

sub update {
    my ($self, %args) = @_;

    $self->_chk_lv_objects;

    my $op = "Snapvault Update".$self->description;
    NOTE::logcomment($op);

    $args{COMPLETE} = ($args{background} ? 0 : 1); delete $args{background};
    my $ret = SnapVault::sv_update($self->{_dst}->Filer->conn,
				$self->{_dst}->mount_path,
				%args);
    croak $@ if $@;

    push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    $self->monitor()
	if $args{COMPLETE};

    return $self;
}

=head2 ::abort

Cancels transfers to destination

=over

=item Example

=back

     $lv->abort;

=cut

sub abort {
    my ($self, %args) = @_;

    $self->_chk_lv_objects;
    NOTE::logcomment("Abort LockVault transfer of ".$self->dst->description);
    my $stat = SnapVault::snapvault_abort($self->{_dst}->Filer->conn,
				$self->{_dst}->Filer->host.":".$self->{_dst}->mount_path,
				%args);
    if ( $@ ) {
	croak "Abort Failed with unexpected error: $@"
		unless $@ =~ /process was aborted|could not read from socket/i;
    }

    if ( $stat ) {
	return $self;
    } else {
	croak "Abort Failed";
    }
}

=head2 ::compare

Compares LockVault's src and dst to determine if there are differences

=over

=item Example

=back

     $lv->compare;

=cut

sub compare {
    my ($self) = @_;

    $self->_chk_lv_objects;
    NOTE::logcomment("Compare LockVault ".$self->description);

    return $self->src->compare($self->dst);
}

=head2 ::stop

Unconfigures the lockvault, and remove the dst object

=over

=item Example

=back

     $lv->stop;

=cut

sub stop {
    my ($self, %args) = @_;

    $self->_chk_lv_objects;

    my $op = "Snapvault Stop".$self->description;
    NOTE::logcomment($op);

    $args{COMPLETE} = ($args{background} ? 0 : 1); delete $args{background};
    my $ret = SnapVault::sv_stop($self->{_dst}->Filer->conn,
				$self->{_dst}->mount_path, %args);
    croak $@ if $@;

    push @{$self->{_subtests}}, { operation => $op,
				  subtest => $ret,
				  state => undef,
				  status => undef  };
    $self->monitor(%args)
   	 if $args{COMPLETE};

    return $self;
}

=head2 ::release

Indicates that there will be not more updates to this lockvault

=over

=item Example

=back

      $lv->release;

=cut

sub release {
    my ($self, %args) = @_;

    $self->_chk_lv_objects;

    SnapVault::snapvault_release($self->{_src}->Filer->conn,
				$self->{_src}->mount_path,
				$self->dst_path);
    croak $@ if $@;
    return $self;
}

=head2 ::restore

Replaces src data with the dst's data

=over

=item %args

B<src> => I<newsrc object>

I<Optional.>

If supplied will set the LockVault src to this NOTE::Qtree object
and perform restore from this object. If not will use the src object
within this LockVault object.

=item Example

=back

     $lv->restore;

     $lv->restore(src=>$new_src);

=cut

sub restore {
    my ($self, %args) = @_;

    $self->_chk_lv_objects;

    $self->{_src} = $args{src} if exists $args{src};
    my $op = "Snapvault Restore".$self->description;
    NOTE::logcomment($op);

    $args{SOURCE} = $self->dst_path;
    $args{COMPLETE} = ($args{background} ? 0 : 1); delete $args{background};
    # check dst filer access permission for src filer and set if needed
    $self->_set_access($self->{_dst}->Filer, $self->{_src}->Filer);
    my $ret = SnapVault::sv_restore($self->{_src}->Filer->conn, $self->src_path, %args);
    croak $@ if $@;
    
    push @{$self->{_subtests}}, { operation => $op,
				  subtest => $ret,
				  state => "Broken-off",
				  status => "Idle",
				  fconn => $self->src->Filer->conn,
				  src_path => $self->dst_path,
				  dst_path => $self->src_path
				};
    $self->monitor()
	if $args{COMPLETE};

    return $self;
}

=head2 ::resync

Restarts backup to the dst from the src

=over

=item %args

B<src> => I<newsrc object>

I<Optional.>

If supplied will set the LockVault src to this NOTE::Qtree object
and perform resync from this object. If not will use the src object
within this LockVault object.

=item Example

=back

     $lv->resync;

     $lv->resync(src=>$new_src);

=cut

sub resync {
    my ($self, %args) = @_;

    $self->_chk_lv_objects;

    $self->{_src} = $args{src} if exists $args{src};
    my $op = "Snapvault Resync".$self->description;
    NOTE::logcomment($op);

    $args{SOURCE} = $self->src_path;
    $args{COMPLETE} = ($args{background} ? 0 : 1); delete $args{background};
    my $ret = SnapVault::sv_resync($self->{_dst}->Filer->conn, $self->dst_path, %args);
    croak $@ if $@;

    push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    $self->monitor()
	if $args{COMPLETE};

    return $self;
}

=head2 ::monitor

Monitors the lockvault status, and returns when the expected status is reached or
times out. Can Only monitor the last LockVault operation.

=over

=item Example

=back

     $lv->monitor;

=cut

sub monitor {
    my ($self, %args) = @_;

    my $test = $self->{_subtests}->[-1]; # Monitor only the more recent subtest
    return unless $test;
    my $op = $test->{operation};
    my $subtest = $test->{subtest};
    my $exp_state = $args{state} || (exists $test->{state} ? $test->{state} : "Snapvaulted");
    my $exp_status = $args{status} || (exists $test->{status} ? $test->{status} : "Idle");
    my $fconn = $test->{fconn} || $self->dst->Filer->conn;
    my $src_path = $test->{src_path} || $self->src_path;
    my $dst_path = $test->{dst_path} || $self->dst_path;
    $args{TIMEOUT} = 300 unless exists $args{TIMEOUT};
    $args{DELAY} = 5;
    NOTE::logcomment("Monitor $op");
    my $wait_time = $args{TIMEOUT};
    my $mb = 0;
    do {
	my ($src, $dst, $state, $lag, $stat)
		= $self->_status($fconn, $src_path, $dst_path);
	NOTE::logcomment("Expected ($exp_state, $exp_status), "
			."Actual ($state, $stat)");
	if ( lc($state) eq lc($exp_state) && lc($stat) eq lc($exp_status) ) {
		# Verify check sub test for error(s)
		if ( ref $subtest =~ /Subtest/ ) {
			NOTE::logcomment("Wait for subtest to exit...");
			$subtest->wait_finish();
			croak $@ unless subtestResults($subtest);
		}
		return $self;
        }
	my $amt = $mb;
	if ( $stat =~ /Transferring/ ) {
		$stat =~ s/Transferring +\(//;
		$stat =~ s/B done\)//;
		my $unit;
		($amt, $unit) = split / /, $stat;
		$amt = $amt / 1000 unless $unit =~ /m/i;
	}
	if ( $amt > $mb ) {
		$mb = $amt;
	} else {
		# count down the clock if no progress is being made
		$wait_time -= $args{DELAY};
	}
	NOTE::logcomment("Transferred $mb MB");
	sleep($args{DELAY});
    } while $wait_time > 0;
    croak $@ if $@;
    croak "Timeout waiting for lockvault to complete...";
}

=head2 ::status

Returns lockvault status

=over

=item arg

Indicate which field to return
Must be one of: 'src', 'dst', 'state', 'lag' or 'status'.

If not supplied, an array representing these fields are returned

=item Examples

=back

    my ($src, $dst, $state, $lag, $stat) = $lv->status;

    my $state = $lv->status('state');

=cut

sub status {
    my ($self, $arg) = @_;

    my @stat = $self->_status;
    my %type = ( src=>0, dst=>1, state=>2, lag=>3, status=>4 );
    if ( $arg ) {
	$arg = "status" unless exists $type{$arg};
	return $stat[$type{$arg}];
    } else {
	return @stat;
    }
}

sub _status {
    my ($self, $fconn, $src_path, $dst_path) = @_;

    $self->_chk_lv_objects;

    $fconn = $self->dst->Filer->conn unless $fconn;
    $src_path = $self->src_path unless $src_path;
    $dst_path = $self->dst_path unless $dst_path;

    my @stat = SnapVault::snapvault_status($fconn, $dst_path, $src_path);
    croak $@ if $@;

    return @stat;
}

sub _chk_lv_objects {
    my ($self) = @_;

    croak "LockVault src not set" unless $self->{_src};
    croak "LockVault dst not set" unless $self->{_dst};
    croak "LockVault log not set" unless $self->{_log};
}

=head2 ::destroy

Cleanup the lockvault by stopping any activities, removing the dst and clearing the log
volume setting

=over

=item Example

=back

     $lv->destroy;

=cut

sub destroy {
    my ($self) = @_;

    $self->{_destroy} = 1;

    my ($stat, $error);
    if ( $self->status('status') ) {
	NOTE::logcomment("Destroy LockVault ".$self->description);
	# Set log vol in case not set, maybe removed by another lockvault object destroy.
	$self->_set_log_vol($self->{_log}->Filer->conn, $self->{_log}->name);
	# Cleanup lockvault configuration
	eval {
		$stat = $self->status('status');
		$self->abort
			if $stat =~ /transferring/i;
		$self->stop;
		$self->release;
	};
	$error = $@ if $@;
	if ( $self->log->snaplock eq "compliance" ) {
		# This is so that compliance dst and log volumes can be destroyed
		NOTE::logcomment("Wait for retention period to pass before continuing...");
		sleep($self->retention_period + 5);
	}
    } else {
	NOTE::logcomment("No lockvault status for ".$self->description
		.", skipping...");
    }
    $self->_set_log_vol($self->{_log}->Filer->conn, "''");
    croak "Error during lockvault cleanup: $error" if $error;
}

sub DESTROY {
    my ($self) = @_;

    return if $$ != $self->{_owner}
	   or $self->{_spare_me}
	   or $self->{_destroy}
	   or NOTE->ALL_SPARED;

    $self->destroy;
}

sub _set_access {
    my ($self, $srcF, $dstF) = @_;

    my $f = $srcF->conn;
    my $key = "snapvault.access";
    my $value = _option_value($f, $key);
    $value =~ s/^host=//;
    my $fname = $dstF->hostspec;
    unless ( $value =~ /$fname/ or $value eq "all" ) {
	NOTE::logcomment("Add filer ($fname), to filer ".$srcF->hostspec."'s $key");
	$value .= "," if $value;
	$value .= $fname;
	my $text = SayWrap::sayEOF($f, "options $key host=$value");
	croak $text if $text;
    }
}

sub _option_value {
   my ($f, $key) = @_;

   my $text = SayWrap::sayEOF ($f, "options $key");
   my ($return_key, $value) = split " ", $text;
   croak $text
	unless $return_key eq $key;
   return $value;
}

