#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/NBU/NetBackup.pm#1 $
#
## @summary  NOTE NetBackup class
## @status   Production
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::NBU::NetBackup

=head1 DESCRIPTION

This is the NOTE class for working with Veritas NetBackup, ntest client must have NBU installed, this is a requiremnt for NBU::nbuUtils !!!

=head1 METHODS
stuadd, pathadd, pathdel, clientadd, clientdel, stuinfo, stulist, studel, sturep, policyadd, policyinfo, policylist, policydel, backup, restore, schedadd, scheddel, comparedata, backupIDlist, imageExpire

=cut

package NOTE::NBU::NetBackup;
our @ISA = qw(NOTE);    # This library inherits NOTE

use strict;
use Carp qw(croak);    # Enhanced error handling. Prints backtraces
#use Config;        # Pull OS and hardware from Perl itself
use Tharn qw();        # Use NetApp core as well

use NDMP::NdmpCmds;    # /rlse/test/testN/NDMP/NdmpCmds.pm
use NBU::nbuUtils;    # /rlse/test/testN/lib/NBU/nbuUtils.pm

use Data::Dumper;    # Used for debugging, remove if not needed anymore

my $regSlice = {
    'snapvault.access' => sub {
        'options.snapvault.access'
    },
    'snapvault.enable' => sub {
        'options.snapvault.enable'
    },
    'snapvault.nbu.trace_level' => sub {
        'options.snapvault.nbu.trace_level'
    },
};

sub _get_host_rec {
    my( $host ) = @_;
    my $rec = undef;
    if( ref( $host ) =~ /^Hostrec::/ ) {
        $rec = $host;
    } else {
        $rec = Tharn::host($host) || croak "Host specification '$host' doesn't resolve to anything";
    }
    return $rec;
}


=head2 ::stuadd

Creates storage unit object. Licenses sv_ontap_sec,nearstore_option,a_sis and adds configuration to NBU server.
Returns storage unit object.

=over

=item Syntax

    NOTE::NBU::NetBackup->stuadd(
        volume=>$vol,
        server=>$client,
        maxfragsize=>20,
        maxjobs=>1,
        client=>$client,
        cf=>$cf );

=item Arguments

B<volume> => I<NOTE::Volume> Required parameter. Volume object where backup data will be stored

B<maxfragsize> => I<number> Optional parameter. Maximum fragment size for disk storage units must be between 20 MB and 524288 MB. Default value is 1024 (Mb).

B<maxjobs> => I<number> Optional parameter. Default is 10.

B<server> => I<string> Required parameter. Fully qualified NBU server (solaris) host name or hostrec

B<client> => I<string> Optional parameter. Fully qualified client host name or hostrec, default is B<server>'s hostname.

B<cf> => I<string> Optional parameter. Possible value is 0|1|2, 0 for image backup, 1 for clearfs backup, 2 for soi backup.  default is 1.
=item Examples

    my $vol = NOTE::Volume->create( Filer => { host => $FILER } );
    my $nbu = NOTE::NBU::NetBackup->stuadd(volume=>$vol, server=>$client );


=item Dependencies

 /rlse/test/testN/lib/NOTE.pm, /rlse/test/testN/lib/NBU/nbuUtils.pm, /rlse/test/testN/NDMP/NdmpCmds.pm
 NBU::nbuUtils::add_tpconfig, NdmpCmds::IsNdmp, NdmpCmds::enableNdmp, NOTE::Registry::Slice

=back

=cut

sub stuadd {
    my ($class,%args) = @_;
    my $self = {};

    croak "Missing argument 'server', fully qualified NBU server (solaris) host name or hostrec must be provided" unless($args{server});
    $self->{_Server} = _get_host_rec($args{server});
    $self->{_Client} = $args{client} ? _get_host_rec($args{client}) : $self->{_Server};

    $self->{_StuLabel} = 'stu_' . NOTE::newName();
    
    $self->{_Volume} = $args{volume};
    $self->{_Filer} = $self->{_Volume}->Filer;
    $self->{_PolicyList} = ();
    my $cf = defined $args{cf} ? $args{cf} : 1;

    $self->{_bkup_rstr_log} = "$ENV{'LOGDIR'}/p3backup_restore_log."."$$";  

    my $maxjobs = $args{maxjobs} ? $args{maxjobs} : 10;
    # Maximum fragment size for disk storage units must be between 20 MB and 524288 MB (512GB)
    my $maxfragsize = $args{maxfragsize} ? $args{maxfragsize} : 1024;

    bless $self, $class;
    tie %{$self}, 'NOTE::Registry::Slice', $self, $regSlice;

    $self->{_Filer}->check_license($_) foreach( qw( sv_ontap_sec nearstore_option a_sis ) );

    $self->{'snapvault.access'} = 'all';
    $self->{'snapvault.enable'} = 'on';
    $self->{'snapvault.nbu.trace_level'} = 150;

    NdmpCmds::IsNdmp($self->{_Filer}->conn);
    NdmpCmds::enableNdmp($self->{_Filer}->conn);

    eval { NBU::nbuUtils::add_tpconfig($self->{_Filer}->host, 'root', $self->{_Filer}->conn->{default_password}); };
    croak("Unable to add config: $@") if($@ && $@ !~ /Credentials already exist/);

    NOTE::reference( $self );

    $self->{_Volume}->snapshot_sched(sched=>"0 0 0");

    bpstuadd(
        $self->{_StuLabel},
        $self->{_Server}->{'hostname'},
        $self->{_Volume}->mount_path,
        $self->{_Client}->{'hostname'},
        $maxjobs,
        $self->{_Filer}->host,
        $maxfragsize,
        $cf
    );
    croak("Unable to add storage unit...") unless( $self->stuinfo ); # this will dump details to log

    croak($@) if($@);

    return $self;
}

#--------------------------------------------------------

=head2 ::pathadd

Add specified source pathname to the policy file list. The pathname can
be a directory, a filename, a script, or a directive.

=over

=item Syntax

=item Arguments

B<policy> => I<string> Required parameter. NetBackup policy name.

B<path> => I<string> Required parameter. Path to backup. Space separated list of
directories. Path must be a full path starting with C<'/'>!

=item Examples

 my $nbu = NOTE::NBU::NetBackup->stuadd(volume=>$vol, server=>$client ); # Creat object
 my $policy = $nbu->policyadd;                                           # Add anonymous policy
 $nbu->pathadd(policy=>$policy,path=>'/usr/local/software');             # Add directory we want to backup

=item Dependencies

=back

=cut

sub pathadd {
    my ($self,%args) = @_;

    croak("Missing policy") unless($args{policy});
    croak("Missing path") unless( $args{path} );

    bpplinclude(
    $args{policy},
    $self->{_Client}->{'hostname'},
    '-add',
    $args{path}
    );
    croak($@) if($@);
}
#--------------------------------------------------------

=head2 ::pathdel



=over

=item Syntax

=item Arguments

B<policy> => I<string> Required parameter. NetBackup policy name.

B<path> => I<string> Required parameter. Path to delete.

=item Examples

 $nbu->pathdel( policy=>$policy,path=>'/usr/local/software' ); 

=item Dependencies

=back

=cut

sub pathdel {
    my ($self,%args) = @_;

    croak("Missing policy") unless($args{policy});
    croak("Missing path") unless( $args{path} );

    bpplinclude(
    $args{policy},
    $self->{_Client}->{'hostname'},
    '-delete',
    $args{path}
    );
    croak($@) if($@);
}
#--------------------------------------------------------

=head2 ::clientadd

Add storage unit client to given policy

=over

=item Syntax

=item Arguments

B<policy> => I<string> Required Optional parameter.

=item Examples

 $nbu->clientadd( policy=>$policy );

=item Dependencies

=back

=cut

sub clientadd {
    my ($self,%args) = @_;
    croak("Missing policy") unless($args{policy});
    bpplclients(
    $args{policy},
    '-add',
    $self->{_Client}->{'hostname'} ? $self->{_Client}->{'hostname'} : $self->{_Server}->{'hostname'},
    $self->{_Server}->{'hostname'},
    (split('-',$self->{_Client}->{hosttype}))[1] eq 'linux' ? 'pc'    : (split('-',$self->{_Client}->{hosttype}))[1],		                                   # Pull HW from host record
    (split('-',$self->{_Client}->{hosttype}))[1] eq 'linux' ? 'linux' : (split('-',$self->{_Client}->{hosttype}))[1] . (split(/\./,$self->{_Client}->{OS_VER}))[1] # Pull OS from host record
);
    croak($@) if($@);
}
#--------------------------------------------------------

=head2 ::clientdel

Remove storage unit client from given policy

=over

=item Syntax

=item Arguments

B<policy> => I<string> Required parameter. Policy name

=item Examples

=item Dependencies

=back

=cut

sub clientdel {
    my ($self,%args) = @_;
    croak("Missing policy") unless($args{policy});
    bpplclients(
    $args{policy},
    '-delete',
    $self->{_Client}->{'hostname'} ? $self->{_Client}->{'hostname'} : $self->{_Server}->{'hostname'},
    $self->{_Server}->{'hostname'},
    (split('-',$self->{_Client}->{hosttype}))[1] eq 'linux' ? 'pc'    : (split('-',$self->{_Client}->{hosttype}))[1],		                                   # Pull HW from host record
    (split('-',$self->{_Client}->{hosttype}))[1] eq 'linux' ? 'linux' : (split('-',$self->{_Client}->{hosttype}))[1] . (split(/\./,$self->{_Client}->{OS_VER}))[1] # Pull OS from host record
);
    croak($@) if($@);
}



#--------------------------------------------------------

=head2 ::stuinfo

Returns storage unit information in long format as hash

=over

=item Syntax

 $nbu->stuinfo();

=item Arguments

None

=item Examples

 logcomment( $nbu->stuinfo );

=item Dependencies

 Calls ::stulist

=back

=cut

sub stuinfo {
    my $self = shift;
    return $self->stulist( long=>1, label=>$self->{_StuLabel} );
}
#--------------------------------------------------------

=head2 ::stulist

Displays one or all of NetBackup storage units or storage unit groups

Returns array of storage units names if C<long> option is not set. Or returns hash of storage units as shown below
if C<long> options is set:

'p3stu2' => {
    'Low_Water_Mark' => '80',
    'Host_Connection' => 'nam29.lab.netapp.com',
    'High_Water_Mark' => '98',
    'On_Demand_Only' => 'yes',
    'Concurrent_Jobs' => '5',
    'Max_Fragment_Size' => '524288 MB',
    'Ok_On_Root' => 'no',
    'Path' => '"/vol/p3ie"',
    'Stage_data' => 'no',
    'Storage_Unit_Type' => 'Disk',
    'Storage_Unit_Subtype' => 'NearStore',
    'Storage_Server' => 'dpg-drwho6 (NearStore)',
    'Block_Sharing' => 'yes',
    'Max_MPX' => '1'
},
'p2test' => {
    'Low_Water_Mark' => '80',
    'Host_Connection' => 'nam29.lab.netapp.com',
    'High_Water_Mark' => '98',
    'Transfer_Throttle' => '0',
    'On_Demand_Only' => 'yes',
    'Concurrent_Jobs' => '1',
    'Max_Fragment_Size' => '524288 MB',
    'Ok_On_Root' => 'no',
    'Path' => '"/vol/p2dst"',
    'Stage_data' => 'no',
    'Storage_Unit_Type' => 'Disk',
    'Storage_Unit_Subtype' => 'SnapVault',
    'Storage_Server' => 'dpg-drwho6 (SnapVault)',
    'Block_Sharing' => 'no',
    'Max_MPX' => '1'
}

=over

=item Syntax
 
 $result = stulist( long=>0, label => 'whatever');

=item Arguments

B<long> => C<1|0> (I<optional>) Display output in long format, default is short

B<label> => C<string> (I<optional>) Display info for one storage unit only which is defined by label

=item Examples

=item Dependencies

=back

=cut

sub stulist {
    my($self,%args) = @_;
    my $result = bpstulist(%args);
    croak ($@) if($@);
    return $result;
}
#--------------------------------------------------------

=head2 ::stuupdate

Update NetBackup storage unit with new volume object 

=over

=item Syntax

 $nbu->stuupdate(volume=>$new_vol);

=item Arguments

B<volume> => C<string> (I<required>) new volume object 

=item Examples

 $nbu->stuupdate(volume=>$new_vol);

=item Dependencies

=back

=cut

sub stuupdate {

    my ($self, %args) = @_;
    croak "Missing new volume object" unless ($args{volume});

    $self->{_Volume} = $args{volume};
    $self->{_Filer} = $self->{_Volume}->Filer;
    my $filer_name = ($self->{_Filer})->host;
    NOTE::logcomment("MY NEW FILER NAME IS $filer_name"); 
    print Dumper($self);

    my $filer_obj =  $self->{_Filer};
    $filer_obj->check_license($_) foreach( qw( sv_ontap_sec nearstore_option a_sis ) );
    $self->{'snapvault.access'} = 'all';
    $self->{'snapvault.enable'} = 'on';
    $self->{'snapvault.nbu.trace_level'} = 150;

    my $conn = $filer_obj->conn;

    NdmpCmds::IsNdmp($conn);
    NdmpCmds::enableNdmp($conn);

    eval { NBU::nbuUtils::add_tpconfig($filer_name, 'root', $conn->{default_password}); };
    croak("Unable to add config: $@") if($@ && $@ !~ /Credentials already exist/);

    bpsturep(
        $self->{_StuLabel},
        $self->{_Server}->{'hostname'},
        $self->{_Volume}->mount_path,
        $self->{_Client}->{'hostname'},
        $self->{_Filer}->host,
    );
    croak($@) if ($@);
}
    
    
#-------------------------------------------
=head2 ::studel

Remove storage unit from server. If STU has policies, then policies will be removed first

=over

=item Syntax

 $nbu->studel;

=item Arguments

None

=item Examples

 $nbu->studel;

=item Dependencies

=back

=cut

sub studel {
    my($self,%args) = @_;
    if( $self->{_PolicyList} ) {
        NOTE::logcomment("Destroying policies");
        $self->policydel(policy=>$_) foreach(keys %{$self->{_PolicyList}});
    }
    NOTE::logcomment("Destroying storage unit:" . $self->{_StuLabel});
    my $result = bpstudel($self->{_StuLabel});
    croak ($@) if($@);
    return $result;
}
#--------------------------------------------------------

=head2 ::policyadd

Adds new policy to storage unit.

=over

=item Syntax

 $nbu->policyadd( policy=>'NeaPolicyName' );

=item Arguments

B<policy> => I<string> Optional policy name. If name is not given, then name will be generated

=item Examples

 my $policy = $nbu->policyadd;

=item Dependencies

=back

=cut

sub policyadd{
    my($self,%args) = @_;
    my $policy = $args{policy} ? $args{policy} : 'plcy_' . NOTE::newName();
    NOTE::logcomment("Adding policy $policy to " . $self->{_StuLabel} );
    my $result = NBU::nbuUtils::bppolicynew( $policy, $self->{_Server}->{'hostname'} );
    croak ($@) if($@);

    $self->{_PolicyList}{$policy} = $self->policylist(policy=>$policy);
    
    croak("Unable to add policy: $policy to ".$self->{_StuLabel}) unless( $self->{_PolicyList}{$policy} );
    return $policy;
}


#--------------------------------------------------------

=head2 ::policyinfo

Set or modify the defined attributes to the policy file list with the defined collect_tir_info(0, 1 or 2). 

=over

=item Syntax

=item Arguments

B<policy> => I<string> Required parameter. NetBackup policy name.

=item Examples

 my $nbu = NOTE::NBU::NetBackup->stuadd(volume=>$vol, server=>$client ); # Creat object
 my $policy = $nbu->policyadd;                                           # Add anonymous policy
 $nbu->policyinfo(policy=>$policy,tir=>'1');             #set collect_tir_info to 1 ; it does "bpplinfo $policy -M $master -set -follownfs 1 -collect_tir_info 1 -residence $storageunit"
 or 
 $nbu->policyinfo(policy=>$policy);             #set collect_tir_info to 2; it does "bpplinfo $policy -M $master -set -follownfs 1 -collect_tir_info 2 -residence $storageunit"

=item Dependencies

=back

=cut

sub  policyinfo{

    my($self,%args) = @_;

    croak("Missing policy") unless($args{policy});

    my $tir = $args{tir} ? $args{tir}: 2;
    NOTE::logcomment("Setting attributes to policy $args{policy} with tir = ", $tir);
    my $result = NBU::nbuUtils::bpplinfo(
        $args{policy},
        $self->{_Server}->{'hostname'},
        $args{option} ? $args{options} : '-set',
        $self->{_StuLabel},
        $tir,
    );
    croak ($@) if($@);
    $self->{_PolicyList}{$args{policy}} = $self->policylist(policy=>$args{policy }); # renew info;
    return 1;
}



#--------------------------------------------------------

=head2 ::policylist

Returns hash of policy lists

=over

=item Syntax

 $nbu->policylist( policy => $policy );

=item Arguments

B<policy> => I<string> Optional parameter.

=item Examples

 my %AllPolicies = $nbu->policylist;

=back

=cut

sub policylist {
    my($self,%args) = @_;
    my %list = NBU::nbuUtils::bppllist($args{policy});
    croak ($@) if($@);
    return %list;
}
#--------------------------------------------------------

=head2 ::policydel

Deletes policy from storage unit. You can only delete policy you've created

=over

=item Syntax

 $nbu->policydel( policy => $mypolicy );

=item Arguments

B<policy> => I<string> Required parameter. Policy name

=back

=cut

sub policydel{
    my($self,%args) = @_;

    if( $self->{_PolicyList}{ $args{policy} } ) {
        NBU::nbuUtils::bppldelete( $args{policy} );
        croak ($@) if($@);
    } else {
        croak("Can not delete policy, you're not owner of $args{policy}");
    }
}

#--------------------------------------------------------

=head2 ::backup

Backup given policy/schedule

=over

=item Syntax

=item Arguments

B<policy> => I<string> Optional parameter if schedule is unique within storage unit

B<schedule> => I<string> Required parameter. Name of the schedule to backup now.

=item Examples
 $nbu->backup( policy=>$policy, schedule=>$schedl );

=item Dependencies

=back

=cut

sub backup{
    my($self,%args) = @_;
    my $result = NBU::nbuUtils::bpbackup( 
        $args{policy}, 
        #$args{logfile}, 
        $self->{_bkup_rstr_log},
        $args{schedule}, 
        $self->{_Server}->{'hostname'} );
    croak ($@) if($@);
    return $result;
}

#--------------------------------------------------------

=head2 ::restore

Restore given policy/schedule

=over

=item Syntax

=item Arguments

B<policy> => I<string> Optional parameter if schedule is unique within storage unit

B<schedule> => I<string> Required parameter. Name of the schedule to backup now.

=item Examples
 $nbu->restore( policy=>$policy, schedule=>$schedl );

=item Dependencies

=back

=cut

sub restore{
    my($self,%args) = @_;
    
    croak("Missing policy") unless($args{policy});
    croak("Missing source path") unless( $args{srcpath} );
    
    my $result = NBU::nbuUtils::bprestore( 
        $args{policy}, 
        $args{starttime} ? $args{starttime} : "01/01/1970", 
        $args{endtime}? $args{endtime} : currentDate(), 
        $self->{_Server}->{'hostname'}, 
        $self->{_Server}->{'hostname'},
        #$args{logfile}, 
        $self->{_bkup_rstr_log},
        $args{srcpath}, 
        $args{dstpath} );
    croak ($@) if($@);
    return $result;
}

#--------------------------------------------------------

=head2 ::schedadd

Add schedule to policy. Operation will fail if you're not the owner of given policy

=over

=item Syntax

=item Arguments

B<policy> => I<string> Required parameter.

B<type> => I<string> Optional parameter. The type of the schedule. The default schedule
type is FULL. The following list contains the possible values for this attribute with
their meanings:

    FULL - full
    INCR - differential incremental
    CINC - cumulative incremental
    UBAK - user backup
    UARC - user archive

B<schedule> => I<string> Optional parameter. Name of the schedule. Will generate schedule name if nothing is given.

=item Examples

=item Dependencies

=back

=cut

sub schedadd{
    my($self,%args) = @_;

    croak("Can not use policy $args{policy}, not an owner") unless( $self->{_PolicyList}{$args{policy}} );
    $args{schedule}  = 'schd_' . NOTE::newName() unless( $args{schedule} =~ /\w+/ );

    my $result = NBU::nbuUtils::bpplsched(
        $args{policy},
        $args{schedule},
        $self->{_Server}->{'hostname'},
        #$args{type} =~ /(FULL|INCR|CINC|UBAK|UARC)/ ? $args{type} : 'FULL',
        $args{type} ? $args{type} : 'FULL',
        $self->{_StuLabel},
    );
    croak ($@) if($@);
    $self->{_PolicyList}{$args{policy}} = $self->policylist(policy=>$args{policy}); # renew info;
    return $args{schedule};
}


=head2 ::scheddel

Delete schedule from given policy

=over

=item Syntax

=item Arguments

B<policy> => I<string> Required parameter. Policy name

B<schedule> => I<string> Required parameter. Schedule name

=item Examples

=item Dependencies

=back

=cut

sub scheddel{
    my($self,%args) = @_;

    croak("Incorrect policy or schedule") unless( $self->{_PolicyList}{$args{policy}}{$args{schedule}} );

    NBU::nbuUtils::bpplscheddel(
    $args{policy},
    $args{schedule}
    );

    croak ($@) if($@);
}


=head2 ::comparedata

compare the data of two given trees 

=over

=item Syntax

=item Arguments

B<src_path> => I<string> Required parameter. path of source tree

B<dst_path> => I<string> Required parameter. path of destination tree

=item Examples

=item Dependencies

=back

=cut

sub comparedata {
    my($self, %args) = @_;

    croak("Missing source path") unless($args{src_path});
    croak("Missing destination path") unless( $args{dst_path} );
    
    my $result = NBU::nbuUtils::compareData( 
        $args{src_path}, 
        $args{dst_path} 
    );
    croak ("DATA COMPARE BETWEEN $args{src_path} AND $args{dst_path} FAILED! - $result" ) if(!$result);
}

###############################
=head2 ::backupIDlist

returns a list of backup id of a given policy and schedule type.

=over

=item Syntax

 @backup_ids = backupIDlist(policy => $policy ,schd_type =>$schd_type);

=item Arguments

B<policy> => C<string> (I<Required>) NetBackup policy name

B<schd_type> => C<string> (I<Optional>) schedule type, see bpimagelist
man page for details FULL|INCR|CINC|UBAK|UARC|NOT_ARCHIVE, default is FULL

=back

=cut

sub backupIDlist {
    my($self, %args) = @_;

    croak("Missing policy") unless($args{policy});
    croak("Incorrect policy") unless( $self->{_PolicyList}{$args{policy}} );

    my @list = NBU::nbuUtils::backupIDs(
        $args{policy},
        $args{schd_type}
    );
    croak ($@) if($@);
    return @list;
}

=head2 ::imageExpire

expire a backup image

=over

=item Syntax

 $result = imageExpire(backup_id =>$backupID);

=item Arguments

B<backup_id> => C<string> (I<Required>) Backup ID

B<expr_time> => C<string> (I<Optional>) expiration time, see bpexprdate man page for details , default is 0

B<$nodelete> => C<0|1> (I<Optional>) To delete the image from storage disk or not. default is 0, that image will be deleted from catalog as well as storage disk, if set to 1, image will be only deleted from catalog but not from storage disk
=back

=cut

sub imageExpire {
    my($self, %args) = @_;

    croak("Missing backup id") unless($args{backup_id});

    NBU::nbuUtils::exprImage(
        $args{backup_id},
        $args{expr_time},
        $args{nodelete}
    );
    croak ($@) if($@);
}

    
#sub Volume {
#    my $self = shift;
#    return $self->{_Volume};
#}
    
        
sub DESTROY {
    my $self = shift;

    unlink </usr/tmp/renameFile*> || warn "having trouble deleteing, $@";
    
    #$self->studel unless( NOTE::SPARE );
    #$self->studel unless( $NOTE::_NOTE_SPARE_ALL );
    $self->studel unless( NOTE::ALL_SPARED );
    $self->unreference;
}

1;
