#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/NBU/P3Utils.pm#1 $
#
## @summary  P3 related library routines.
## @status   Public
## @pod      here
## @author   weili@netapp.com

=head1 NAME

NOTE::NBU::P3Utils

=head1 DESCRIPTION

This is the NOTE class for working with P3 utils.

=head1 METHODS

=cut

package NOTE::NBU::P3Utils;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(	create_aggregate
			p3_setup 
			create_p3_volume 
			delete_p3_volume
			snapvault_status
			delete_qtrees
			list_delete_images
			write_p3_image
			close_p3_image
			read_p3_image
			delete_p3_image
			list_p3_image
			compare_images
			vsm_p3_volume
			delete_snapvault_config
		);

use strict;
use Carp;
use Tharn;
use SayWrap;
use FSR::FSRutils;
use Aggregate;
use Volume;
use Dense::Utils;
use SPARSE::Resource::Filer;
use Admin::License qw(license unlicense is_licensed);
use filer;
use Filer::Version;
use NOTE;
use QtreeCmds;


my $TOOLSDIR = "/usr/local/test/bin";

=head2 B<Basic Operations>

=head2 ::p3_setup

Do the basic setup for p3: install required licenses (

=over

=item Arguments

B<filer> => I<object>

I<Required.>
The NOTE filer object.

B<options_snapvault_access> => I<string>

I<Optional.>
Set up snapvault.access option.

B<options_snapvault_enable> => I<string>

I<Optional.>
Set up snapvault.enable option.

B<options_snapvault_nbu_trace_category> => I<string>

I<Optional.>
Set up snapvault.nbu.trace_category option.

B<options_snapvault_nbu_trace_level> => I<string>

I<Optional.>
Set up snapvault.nbu.trace_level option.

B<options_snapvault_nbu_verify_backup> => I<string>

I<Optional.>
Set up snapvault.nbu.verify_backup option.

B<options_snapvault_nbu_verify_duplicate> => I<string>

I<Optional.>
Set up snapvault.nbu.verify_duplicate option.

=item Examples

=back

	p3_setup( $filer, 
		  options_snapvault_access=>'all', 
		  options_snapvault_enable=>'on', 
		  options_snapvault_nbu_trace_category=>'all'
		);




=cut

sub p3_setup{

	my $access= "all";
	my $enable= 'on';
	my $category= 'all';
	my $level= '100';
	my $duplicate = '';
	
	my ($filer, %args) = @_; 
	$access=$args{options_snapvault_access};
	$enable=$args{options_snapvault_enable};
	$category=$args{options_snapvault_nbu_trace_category};
	$level=$args{options_snapvault_nbu_trace_level};
	$duplicate =$args{options_snapvault_nbu_verify_duplicate};

	eval {$filer->check_license('sv_ontap_sec') };		
        if (Filer::Version::get_wafl_version($filer->hostrec)>73){
		$filer->check_license('sv_nbu_sec');
        }

	## license nearstore options if not a nearstore filer
	my $filer_model = filer::filer_model_get($filer->conn);
        if ($filer_model =~ (/^FAS30/ || /^FAS60/) ) {
 		$filer->check_license('nearstore_option') ;
        }

        ##need to find out which version the a_sis started 
        if (Filer::Version::get_wafl_version($filer->hostrec)>71){
                $filer->check_license('a_sis');
        }
        else {
		$filer->check_license('nearstore_asis2');
        }

	## set options snapvault value
	sayEOF($filer->conn, "options snapvault.access $access"); 
	sayEOF($filer->conn, "options snapvault.enable $enable");
	sayEOF($filer->conn, "options snapvault.nbu.trace_category $category");
	sayEOF($filer->conn, "options snapvault.nbu.trace_level $level");
	sayEOF($filer->conn, "options snapvault.nbu.verify_duplicate $duplicate");
}


=head2 ::create_p3_volume

Creates a p3 volume.

=over

=item Arguments

B<filer> => I<object>

I<Required.>
The NOTE filer object.

B<vol_name> => I<string>

I<Required.>
The name of the volume need to be created.

B<aggr_name> => I<string>

I<Optional.>
The name of the aggregate that the volume resides. The default value is "aggr_p3"

B<vol_size> => I<string>

I<Optional.>
The size of the volume to be created. The default value is 3g 

B<force_delete> => C<0 | 1>

I<Optional.>
Destroy the volume before create it, if set to 1 

=item Return Value
Return a NOTE volume object 


=item Examples

=back

	$vol = create_p3_volume(
				$filer, 
				force_delete=>1, 
				vol_name=>'vol_tt', 
				aggr_name=>'aggr_p3'
				);

=cut

sub create_p3_volume{

	my $delete = 0;

	my ($filer, %args) = @_; 
	my $vname = $args{vol_name};
	my $vsize = $args{vol_size};
	my $aggname = $args{aggr_name};
	$delete = $args{force_delete}; 

	if (!defined($vsize)){
		$vsize = '5g';
	}
	if (!defined($aggname)){
		$aggname = 'aggr_p3';
	}

	##get or create the aggregate
	my $aggr = NOTE::Aggregate->create(Filer=>$filer, name=>$aggname);

	## destroy the volume before creation if choose to 
	my $vol; 
	if ($delete && Volume::vol_exists( $filer->conn, $vname )){
		$vol = NOTE::Volume->attach(Filer=>$filer, name=>$vname);
		$vol->purge;
	}

	##Create volume
	$vol = NOTE::Volume->create (
		Filer => $filer,
		name => $vname,
		size => $vsize,
		Aggregate => $aggr);

	
	## we don't need any regular snap shot
	sayEOF($filer->conn, "snap sched $vname 0 0 0"); 

        return $vol;

}

=head2 ::delete_p3_volume

Destroy a p3 volume.

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<vol> => I<object>

I<Required.>
A NOTE volume object

B<> => I<string> 

I<Optional.>

B<1.> If set to 'QTREES': will delete all the qtrees in the volume using "snapvault stop -f", then destroy teh volume.
B<2.> If set to 'IMAGES': will get a list of all images on the volume, deletes them in order and then destroy the volume
B<3.> If not set: will destroy the volume and cleans the snapvault config on the volume 

=item Examples

=back

	delete_p3_volume(
			$filer, 
			$vol, 
			'QTREES'
		);

=cut

sub delete_p3_volume{
        my ($filer, $vol, $options) = @_;
        my $result = 0;
	####my $vol = NOTE::Volume->attach(Filer=>$filer, name=>$vname);

	## QTREES option: destroy the volume and clean up snapvault config
	if ($options =~ /QTREES/){
	##	$vol = NOTE::Volume->attach(Filer=>$filer, name=>$vname);
		delete_qtrees($filer, $vol);
		$vol->purge;	
		
	}
	elsif ($options =~ /IMAGES/){ ## list and delete all images, then destroy volume  
                list_delete_images($filer, $vol);
		###$vol->purge;	
        }
        else {   ## destroy the volume and clean up the qtrees
		my $vol_name = $vol->name;
                $vol->purge;
		delete_snapvault_config($filer, $vol_name);
        }
}

=head2 ::delet_qtrees

delete all qtrees in the volume

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<vol> => I<object>

I<Required.>
A NOTE volume object

=item Examples

=back
	delete_qtrees ($filer, $vol);

=cut

sub delete_qtrees{
        my ($filer, $vol) = @_;
	my ($qtr, @qtrees, $result);

	@qtrees = QtreeCmds::qtreeList ($filer->conn, $vol->name);

	foreach $qtr (@qtrees){
	print "\n\n qtree:$qtr\n\n";
		$result = SnapVault::snapvault_stop($filer->conn, $qtr);
        }

        return $result;
}


=head2 ::delete_snapvault_config

clean up snapvault config

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<vol> => I<object>

I<Required.>
A NOTE volume object

=item Examples

=back
        delete_snapvault_config ($filer, $vol);

=cut

sub delete_snapvault_config{
	my ($filer, $vol) = @_;
	my ($result, $res, $qtr);
        my $status=0;

	$result = sayEOF($filer->conn, "snapvault status -c");

	my @qtrees = split (/\n/, $result);
	foreach $qtr (@qtrees){
		if ($qtr =~ /\/vol\/$vol\/(\S+)/) {
			$res = SnapVault::snapvault_stop($filer->conn, $&);
			if ($res =~ /Could not delete qtree/ || $status == 1){
               			$status = 1;
        		}
		}
       	}
        return $status;
}

=head2 ::list_delete_images

Get a list of all images on the volume, delete them in order

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<vol> => I<object>

I<Required.>
A NOTE volume object

=item Examples

=back
         list_delete_images($filer, $vol);

=cut

sub list_delete_images{
	my ($filer, $vol) = @_;
	my ($basename, $policy, $date);


	my @list = list_p3_image($filer,"basename", "date", "tarfile.img", $vol, "clear", "full" );

      while (@list){
                my $tmp = shift @list;
                if ( $tmp =~ /(Basnename:Policy is )/ ){
			$tmp = $';
                	($basename, $policy)= split/:/, $tmp;
                }
                my $tmp = shift @list;
                if ( $tmp =~ /\Date: (\d+)/){
                        $date = $1;
                }
                delete_p3_image($filer,$basename, $date, "tarfile.img", $vol, "clear", "full", $policy );

      }
}

=head2 ::execute_image_test

This is the internal function to excute image_test with different types

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<filer_with_cred > => I<string>

I<Required.>
Filer with user name and password. i.e.: filer:root@netapp

B<basename> => I<string>

I<Required.>
Image backup's basename

B<policy> => I<string>

I<Required.>
Image backup's policy name

B<img_name> => I<string>

I<Required.>
Image backup's image name

B<tarfile> => I<string>

I<Required.>
Image backup's tar file and path

B<metafile> => I<string>

I<Required.>
Image backup's meta file and path   

B<type> => I<string>

I<Required.>
Image backup's type: <r (read)|w (write)|c (start unpacking)|  d (delete)| l (list)| p (properties)>   

B<vol> => I<object>

I<Required.>
A NOTE volume object

B<image_type> => I<string>

I<Required.>
Image backup's image type: <clear (image unpackable)| image (not unpackable)|soi>

B<backup_type> => I<string>

I<Required.>
Backup type: <incr|full> 

=cut

sub execute_image_test{
# please use real time for $date to call this subroutine, so you will never get the same timestamp. 

	my ($transfer_type, $filer, $tarId, $date, $tarfile, $vol, $image_type, $backup_type, $policy) = @_;

 my $vol_name = $vol->name;
##print "\n\n $transfer_type,$tarId,$date,$tarfile,$vol_name,$image_type,$backup_type\n\n";

	my $filer_with_cred = $filer->hostspec.':root@'.param( 'PASSWORD', '-default', 'netapp' );
	my $filer_dst_vol = '/vol/'.$vol_name;

	##my $filer_dst_vol = '/vol/'.$vol->name;

	my $metafile = $tarfile.".output";
	$policy = $tarId."daily-$$" unless defined ($policy);

	## make sure image and image_test are exist.
if ($transfer_type != 'd'){
	check_p3_image($tarfile, $metafile);
}

	my $command = "/usr/local/bin/sudo $TOOLSDIR/image_test $filer_with_cred $tarId $policy $date $tarfile $metafile $transfer_type $filer_dst_vol $image_type $backup_type log_writes=no log_msgs=no";

	logcomment("command: $command");

	my $result = `$command 2>&1`;
	return $result;
}

# Check pre-requisites prior to running.
sub check_p3_image ($$){
        my ($tarfile,$metafile) = @_;

        unless (-e $tarfile) {
                logresult( 'FAIL', "Specified tar doesn't exist: $tarfile" );
                return 1;
        }
        unless (-e $metafile) {
                logresult( 'FAIL', "Specified tar metafile doesn't exist: $metafile" ) ;
                return 1;
        };
        unless (-e "$TOOLSDIR/image_test") {
                logresult( 'FAIL', "image_test doesn't exist" ) ;
                return 1;
        }
        return 0;
}


=head2 ::write_p3_image

image write 

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<basename> => I<string>

I<Required.>
Image backup's basename

B<policy> => I<string>

I<Required.>
Image backup's policy name

B<img_name> => I<string>

I<Required.>
Image backup's image name

B<tarfile> => I<string>

I<Required.>
Image backup's tar file and path

B<vol> => I<object>

I<Required.>
A NOTE volume object for destination volume

B<image_type> => I<string>

I<Required.>
Image backup's image type: <clear (image unpackable)| image (not unpackabl
e)|soi>

B<backup_type> => I<string>

I<Required.>
Backup type: <incr|full>

=item Examples

=back
       write_p3_image (
		$filer, 
		"basename1",
		"Policy1
		$timeofday, 
		$tarfile,
		$vol, 
		"clear", 
		"full");
 

=cut

sub write_p3_image{
	logcomment("writing p3 images");
	my $write_result=execute_image_test('w', @_);	

        print STDERR "==>image_test Create Result: \n$write_result\n";
  if ($write_result =~ /.+Message reply error: (\d+).+/sm) {
    print "write failed: aborting with error=$write_result\n";
    return 1;
  }
  return 0;

}


=head2 ::close_p3_image

image close

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<basename> => I<string>

I<Required.>
Image backup's basename

B<policy> => I<string>

I<Required.>
Image backup's policy name

B<img_name> => I<string>

I<Required.>
Image backup's image name

B<tarfile> => I<string>

I<Required.>
Image backup's tar file and path

B<vol> => I<object>

I<Required.>
A NOTE volume object for destination volume

B<image_type> => I<string>

I<Required.>
Image backup's image type: <clear (image unpackable)| image (not unpackabl
e)|soi>

B<backup_type> => I<string>

I<Required.>
Backup type: <incr|full>

=item Examples

=back
       close_p3_image (
                $filer,
                "basename1",
                "Policy1
                $timeofday,
                $tarfile,
                $vol,
                "clear",
                "full");


=cut

sub close_p3_image{
	logcomment("closing p3 images");
        my $close_result=execute_image_test('c', @_);

        my $error_code;
  print STDERR "==>image_test closeImage Result: \n$close_result\n";
  if ($close_result =~ /.+Message reply error: (\d+).+/sm) {
                $error_code = $1;
                if ($close_result =~ /.+qtree is currently involved in another operation
.+close retry worked.+/sm) {
                        print "closeImage(): Encountered QTREE_BUSY_ERR but retry image_c
lose succeeded\n";
                return 0;
                } else {
        print "closeImage() failed: aborting with error=$error_code...\n";
        return 1;
                }
                # This is to gaurd against any other errors.
        return 1;
  }
        return 0;
} 


=head2 ::read_p3_image

image read

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<basename> => I<string>

I<Required.>
Image backup's basename

B<policy> => I<string>

I<Required.>
Image backup's policy name

B<img_name> => I<string>

I<Required.>
Image backup's image name

B<tarfile> => I<string>

I<Required.>
Image backup's tar file and path

B<vol> => I<object>

I<Required.>
A NOTE volume object for destination volume

B<image_type> => I<string>

I<Required.>
Image backup's image type: <clear (image unpackable)| image (not unpackabl
e)|soi>

B<backup_type> => I<string>

I<Required.>
Backup type: <incr|full>

=item Examples

=back
       read_p3_image (
                $filer,
                "basename1",
                "Policy1
                $timeofday,
                $tarfile,
                $vol,
                "clear",
                "full");


=cut

sub read_p3_image{
        my $read_result=execute_image_test('r', @_);

        print STDERR  "image_test Read Result: \n$read_result\n";
  	return 1 if ($read_result =~ /.+Message reply error: \d+.+/sm);
  	return 0;
}

=head2 ::delete_p3_image

delete given image

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<basename> => I<string>

I<Required.>
Image backup's basename

B<policy> => I<string>

I<Required.>
Image backup's policy name

B<img_name> => I<string>

I<Required.>
Image backup's image name

B<tarfile> => I<string>

I<Required.>
Image backup's tar file and path

B<vol> => I<object>

I<Required.>
A NOTE volume object for destination volume

B<image_type> => I<string>

I<Required.>
Image backup's image type: <clear (image unpackable)| image (not unpackabl
e)|soi>

B<backup_type> => I<string>

I<Required.>
Backup type: <incr|full>

=item Examples

=back
       delete_p3_image (
                $filer,
                "basename1",
                "Policy1
                $timeofday,
                $tarfile,
                $vol,
                "clear",
                "full");


=cut

sub delete_p3_image{
        my $delete_result=execute_image_test('d', @_);

        print STDERR "===>image_test Delete command: $delete_result\n";

	if ($delete_result=~/.+Message reply error: (\d+).+/sm) {
		print "delete image failed: error= $1\n";
		return $1;
	}
	return 0;

}

=head2 ::list_p3_image

image list

=over

=item Arguments

B<filer> => I<object>

I<Required.>
A NOTE filer object

B<basename> => I<string>

I<Required.>
Image backup's basename

B<policy> => I<string>

I<Required.>
Image backup's policy 


B<img_name> => I<string>

I<Required.>
Image backup's image name

B<tarfile> => I<string>

I<Required.>
Image backup's tar file and path

B<vol> => I<object>

I<Required.>
A NOTE volume object for destination volume

B<image_type> => I<string>

I<Required.>
Image backup's image type: <clear (image unpackable)| image (not unpackabl
e)|soi>

B<backup_type> => I<string>

I<Required.>
Backup type: <incr|full>

=item Examples

=back
       list_p3_image (
                $filer,
                "basename1",
		"policy1",
                $timeofday,
                $tarfile,
                $vol,
                "clear",
                "full");


=cut

sub list_p3_image{
        my $list_result=execute_image_test('l', @_);
	print STDERR "===> image_test List:\n$list_result\n";
	
	my @list = split(/\n/, $list_result);
	@list = grep {/(\bBasnename:|\bDate:)/ } @list;

	return @list;
}

