#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Registry/Slice.pm#1 $
#
## @summary NOTE Registry/Slice module
##
## @description
##    An internal module implementing NOTE::Registry::Slice using Perl's tie hash
##    feature. Module provide 'shorthand' access to registry entries for a Volume
##    / Qtree objects within a filer. Module is meant for internal use within NOTE
##    modules, and are not to be called directly.
##   
## @status Private
##
## @author dl-dpgqa-automation@netapp.com
##

# This package makes it possible to use an object (like a Volume or Qtree object)
# as a hash to reference elements related to it in the registry.
#
# When objects are tied to this package, in addition to the standard arguments
# $pkg and $self, a hash reference ($registrySlice) is passed as an
# argument to the TIEHASH function. This refers to a hash whose keys
# are shorthand notation for the relevant element in the registry
# related to that object and whose values are anonymous subroutines
# that accept the object reference as the first argument and who
# return a key appropriate to use as an index into the actual filer
# registry.
#
# @author tmichael@netapp.com


use NOTE;
use strict;
use warnings;
package NOTE::Registry::Slice;

sub TIEHASH {
    my ($pkg, $realSelf,$registrySlice) = @_;
    my $self = {};
    
    %{$self} = %{$realSelf};
    $self->{_registrySlice} = $registrySlice;
    $self->{_realSelf} = $realSelf;
    $self->{_name} = $realSelf->{_name};
    bless $self, $pkg;
#    NOTE::reference( $self );
    return $self;
}

sub DESTROY {
    my $self = shift;
    
    return unless $self->{_destroyed};
    $self->{_destroyed} = 1;
    delete $self->{_realSelf};
}

sub EXISTS {
    my ($self,$key) = @_;
    my $regSlice = $self->{_registrySlice};

    if ( ! defined $regSlice ) {
	return NOTE::Registry::EXISTS( $self, $key );
    } elsif ( exists $regSlice->{$key} ) {
	return NOTE::Registry::EXISTS(
	    $self,
	    $regSlice->{$key}($self->{_realSelf})
	);
    } else {
        return exists $self->{$key};
    }
}

sub FIRSTKEY {
    my $self = shift;
    my $regSlice = $self->{_registrySlice};
    if ( ! defined $regSlice ) {
	NOTE::Registry::FIRSTKEY( $self );
    } else {
	my $a = keys %{$regSlice};          # reset each() iterator
	each %{$regSlice};
    }
}

sub NEXTKEY  {
    my $self = shift;
    my $regSlice = $self->{_registrySlice};
    if ( ! defined $regSlice ) {
	NOTE::Registry::NEXTKEY( $self );
    } else {
	each %{$regSlice};
    }
}

sub FETCH {
    my ($self, $key) = @_;
    return unless ( defined $self );
    my $regSlice = $self->{_registrySlice};
    if ( ! defined $regSlice ) {  # The slice is of the entire registry
	if ( $key =~ /\./ ) {
	    # If the key has any dots in it, assume it refers to the registry
	    return NOTE::Registry::FETCH( $self, $key );
	}
    } elsif ( exists $regSlice->{$key} ) {
	if ( ref $regSlice->{$key} eq "HASH" ) {
	    my $get = $regSlice->{$key}->{'get'};
	    return &{$get}($self->{_realSelf}) if $get;
	    return undef;
	}
	return NOTE::Registry::FETCH(
	    $self,
	    $regSlice->{$key}($self->{_realSelf})
	);
    }
    return $self->{$key};
}

sub STORE {
    my ($self, $key, $value) = @_;
    my $regSlice = $self->{_registrySlice};

    if ( ! defined $regSlice ) {  # The slice is of the entire registry
	if ( $key =~ /\./ ) {
	    return NOTE::Registry::STORE( $self, $key, $value );
	}
    } elsif ( exists $regSlice->{$key} ) {
	if ( ref $regSlice->{$key} eq "HASH" ) {
	    my $set = $regSlice->{$key}->{'set'};
	    return &{$set}( $self->{_realSelf}, $value ) if $set;
	    return undef;
	}
	return NOTE::Registry::STORE(
	    $self,
	    $regSlice->{$key}($self->{_realSelf}),
	    $value
	);
    }
    $self->{$key} = $value;
}

sub DELETE {
    my ($self, $key) = @_;
    my $regSlice = $self->{_registrySlice};

    if ( ! defined $regSlice ) {  # The slice is of the entire registry
	if ( $key =~ /\./ ) {
	    return NOTE::Registry::DELETE( $self, $key );
	}
    } elsif ( exists $regSlice->{$key} ) {
	if ( ref $regSlice->{$key} eq "HASH" ) {
	    my $delete = $regSlice->{$key}->{'delete'};
	    return $delete( $self->{_realSelf} ) if $delete;
	    return undef;
	}
	return NOTE::Registry::DELETE(
	    $self,
	    $regSlice->{$key}($self->{_realSelf})
	);
    }
    delete $self->{$key};
}

#sub CLEAR   { %{shift->{_realSelf}} = () }
sub CLEAR   {}

1;
