#
# $Id: //depot/prod/test/main/lib/NOTE/Sis.pm#1 $
#
## @summary  NOTE Sis class
## @status   Review
## @pod      here
## @author   dl-dpg-nb-automation@netapp.com

=head1 NAME

NOTE::Sis

=head1 DESCRIPTION

This is the NOTE class for working with Sis

=head1 METHODS

=cut

use strict;

package NOTE::Sis;

use NOTE;
use Carp qw(croak);
use POSIX qw();
use Dense qw();
use ZapiUtil qw(zapi_say zapi_parse_output);
use Params::Validate qw(validate validate_pos BOOLEAN SCALAR OBJECT HASHREF);
use Data::Dumper;

# Expose the following registry keys on the object
my $regSlice = {
    'sis.allow_platforms' => sub {
        'options.sis.allow_platforms'
    },
    'sis.async_requests' => sub {
        'options.sis.async_requests'
    },
    'sis.change_log_blks' => sub {
        'options.sis.change_log_blks'
    },
    'sis.clog_bkt_count' => sub {
        'options.sis.clog_bkt_count'
    },
    'sis.deleted_inode_hash' => sub {
        'options.sis.deleted_inode_hash'
    },
    'sis.dense_use_extended_memory' => sub {
        'options.sis.dense_use_extended_memory'
    },
    'sis.gatherer_readahead_size' => sub {
        'options.sis.gatherer_readahead_size'
    },
    'sis.max_active_ops' => sub {
        'options.sis.max_active_ops'
    },
    'sis.max_share_blks' => sub {
        'options.sis.max_share_blks'
    },
    'sis.max_share_workers' => sub {
        'options.sis.max_share_workers'
    },
    'sis.max_sv_active_ops' => sub {
        'options.sis.max_sv_active_ops'
    },
    'sis.max_vfiler_active_ops' => sub {
        'options.sis.max_vfiler_active_ops'
    },
    'sis.max_vol_limit' => sub {
        'options.sis.max_vol_limit'
    },
    'sis.min_share_blks' => sub {
        'options.sis.min_share_blks'
    },
    'sis.num_gatherer_scanners' => sub {
        'options.sis.num_gatherer_scanners'
    },
    'sis.overlap_requests' => sub {
        'options.sis.overlap_requests'
    },
    'sis.pass1_prefetch_factor' => sub {
        'options.sis.pass1_prefetch_factor'
    },
    'sis.pass2_prefetch_factor' => sub {
        'options.sis.pass2_prefetch_factor'
    },
    'sis.pass2_stage_size' => sub {
        'options.sis.pass2_stage_size'
    },
    'sis.throttle_msec' => sub {
        'options.sis.throttle_msec'
    },
    'sis.verify_trigger_change_rate' => sub {
        'options.sis.verify_trigger_change_rate'
    },
    'sis.wafl_max_blks' => sub {
        'options.sis.wafl_max_blks'
    },
};

=head2 B<Sis Operations>

=head2 ::start

Creates a Sis Object on specified volume

=over

=item Arguments

B<Volume> => I<Volume Object>

I<Required.>
Object of Volume on which SIS should be started.

B<scan> => C<0> | C<1>

I<Optional.>
if set to 1 uses the -s -f option with sis start, defaults to 0 .

B<nosync> => C<0> | C<1>

I<Optional.>
if set to 1 uses the -d -f option with sis start, defaults to 0.
This operation will delete SIS checkpoint and starts from the beginning.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is 600 seconds.

=item Examples

=back

     $sis = NOTE::Sis->start(
         Volume => $vol_src,
         scan => 1,
         fresh => 1,
     );

=cut

sub start {
    my ($class, %args) = @_;
    my $self = {};
    my (%sis_args, $ret, $op);

    # Instantiate this object
    if( ref $class ne '' ) {
        $self = $class;
        if ( exists $args{Volume} ) {
            croak "Cannot override existing 'Volume' attribute ".
                  "in existing object.";
        }
    } else {
        $self = NOTE::Sis->new( %args );
    }

    delete $args{Volume};
    
    $op = "Sis Start" . $self->description;
    NOTE::logcomment( $op );

    # Override default timeout
    $self->timeout(%args);
    
    $args{scan} = defined $args{scan}? $args{scan}:0;
    $args{nosync} = defined $args{nosync}? $args{nosync}: 0;

    # Enable Sis on filer: license, enable
    $self->_license_and_enable;

    # Start Sis
    Dense::dense_start(
        host        => $self->Filer->hostspec,
        path_name   => $self->path,
        nosync      => $args{nosync},
        scan        => $args{scan},
    );
    
    # Check for errors
    croak $@ if $@;

    return $self;
}

=head2 ::new

Creates an object representing sis
but does B<not> cause the starting
of sis on the volume.
The L<C<::start>|/::start> method
of the returned object may be used to start sis on the volume.

This method should not normally be used;
use L<C<::start>|/::start> directly instead.

=over

=item Arguments

B<Volume> => I<Volume Object>

I<Required.>
Object of Volume on which SIS object is to be created.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is 600 seconds.

B<preprocess> => I<1|0>

I<Optional.>
Boolean flag to indicate whether preprocessing of
admin commands such as adding licensing and enabling
should be performed. Default: 0

=item Example

=back

     $sis = NOTE::Sis->new(
         Volume => $vol_src,
     );

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    # Check arguments
    if ( !exists $args{Volume} ) {
        croak "Volume object missing";
    }

    # Set private variables for object
    $self->{_owner}   = $$;
    $self->{_Volume}  = $args{Volume};
    $self->{_Filer}   = $self->{_Volume}->Filer;
    $self->{_timeout} = defined $args{timeout} ? $args{timeout} :600;

    # Bless this object
    bless $self, $class;

    # Make this object a tie variable so registry options can be accessed
    # via a hash element of the object
    tie %{$self}, 'NOTE::Registry::Slice', $self, $regSlice;

    if ( $args{preprocess} ) {
        # Enable Sis on both filers: license, open access and enable
        $self->_license_and_enable;
    }

    NOTE::reference( $self );
    
    return $self;
}

# Internal function to license and enable Sis
sub _license_and_enable {
    my ($self, %args) = @_;

    # Add Sis license
    $self->{_Volume}->Filer->check_license( "nearstore_option" );
    $self->{_Volume}->Filer->check_license( "a_sis" );

    # Get _Filer from Volume object
    $self->{_Filer} = $self->{_Volume}->Filer;

    # Enable Sis
    $self->on;
    return;
}

=head2 B<Sis Commands>

The methods names in this section correspond to
ONTAP's C<sis> subcommands.
Administrator privilege level subcommands are listed in this section.
These methods eventually call the self-named subcommands.
Only the briefest mention of what these subcommands do is given here.
See the sis documentation for more information on them.

=head2 ::on

Turns sis on

=over

=item Arguments

None

=item Examples

=back

     $sis->on;
     
=cut

sub on {
    my ($self) = @_;

    NOTE::logcomment( "Enable Sis on " . $self->description );

    # Turn on sis
    Dense::dense_toggle(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path, 
        state=>'on');

    Dense::dense_verify_flag(
        host=> $self->Filer->hostspec, 
        vol_name=>$self->{_Volume}->name, 
        state=>'ON');
        
    Dense::dense_verify_state(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path, 
        state=>"ENABLED");    
}

=head2 ::off

Turns sis off

=over

=item Arguments

None

=item Examples

=back

     $sis->off;
     
=cut

sub off {
    my ($self) = @_;

    NOTE::logcomment( "Disable Sis on " . $self->description );

    # Turn on sis
    Dense::dense_toggle(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path, 
        state=>'off');
        
    Dense::dense_verify_state(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path, 
        state=>"DISABLED");
    
}

=head2 ::stop

Aborts the currently active SIS operation on the volume.

=over

=item Arguments

None

=item Example

=back

     $sis->stop;

=cut

sub stop {
    my ($self) = @_;

    my $op = "Sis Stop" . $self->description;
    NOTE::logcomment( $op );

    # Stop sis on the volume
    my $ret = $self->Filer->say("sis stop %s",$self->{_Volume}->mount_path);
    
    die "sis is currently idle" if($ret =~ /Operation is currently idle/);

    warn "Last Operation Error does not match string: Operation was stopped" 
        if($self->status->{"last-operation-error"} !~ /Operation was stopped/i);

    # Set flag that this relationship has already been stopped
    $self->{_stopped} = 1;

    return $ret;
}

=head2 ::monitor_snap_replace

Tracks whether specified snapshot in the volume has been replaced by sis.
Returns a subtest object that can intimate whether snapshot was replaced 
within timeout expiry or not.

=over

=item Arguments

None

=item Example

=back

     $sis->monitor_snap_replace(snapshot => $snap);

=cut

sub monitor_snap_replace {
    my ($self, %args) = @_;
    croak "snapshot to be monitored is not specified" if(!defined $args{snapshot});
    NOTE::logcomment(Dumper($self));
    my $timestamp_at_start = $self->get_snapshot_timestamp($args{snapshot});
    
    $args{timeout} = 60; #defined $args{timeout}? $args{timeout}: 120;
    my $timestamp_after_replace = $timestamp_at_start;
    NOTE::logcomment("Time stamp at start: ".$timestamp_at_start);
    my $waittime = 0;
    my $subtest = Tharn::subtest( 
        sub{
            while($timestamp_after_replace =~ /$timestamp_at_start/ && $waittime <= $args{timeout}){
                sleep 5;
                $waittime += 5;
                $timestamp_after_replace = $self->get_snapshot_timestamp($args{snapshot});
                NOTE::logcomment("Time stamp at start:".$timestamp_at_start." Vs now:".$timestamp_after_replace);
            }
            croak $args{snapshot}." was not replaced by sis" if($waittime >= $args{timeout});
        }, 
        -bg, -runid , "monitor:".$args{snapshot}
    );    
    return $subtest;
}


sub get_snapshot_timestamp(){
    my ($self, $snapshot) = @_;
    NOTE::logcomment(Dumper($self));
    my ($out, $errno, $reason) = zapi_say($self->Filer->hostspec, "snapshot-list-info", 'volume',$self->name);
    my %snap_access = zapi_parse_output($out,"name", "access-time"); 
    return $snap_access{$snapshot};    
}

=head2 ::status

Returns a hash reference to the status of sis.

=over

=item Arguments

None

=item Examples

=back

     $status = $sis->status;   # Get status from destination filer
     
     $status->{'last-operation-error'} #Gets Last operation error
    
=cut

sub status {
    my ($self) = @_;

    # Get status for this object
    my %status = Dense::dense_status(
        host=> $self->Filer->hostspec, 
        path=>$self->{_Volume}->mount_path,
        long => 1);

    # The above function may return undef with no error
    # so croak on error if it exists
    croak $@ if $@;
        
    my $status = $status{$self->{_Volume}->mount_path};
    foreach(keys %{$status}){
        my $key = $_;
        $key =~ s/\s/-/g;
        $key = lc($key);
        my $value = $status->{$_};
        delete $status->{$_};
        $status->{$key} = $value;
    }
    return $status;
}

=head2 ::stat

Returns a hash reference to the stats of sis.

=over

=item Arguments

B<block> => C<0> | C<1>

I<Optional.>
Optional, either 0 or 1. 
If set to 1, will use -b to display in block format. 
Default is 0.

=item Examples

=back

     $stat = $sis->stat;   # Get long sis status from filer

     $stat = $sis->stat(block => 1);   # Gets the sis info in blocks
     
     $stat->{'Total Process Time'} #Gets Total Process Time
    
=cut

sub stat {
    my ($self, %args) = @_;
    $args{block} = 0 if(!exists $args{block});

    # Get stat for this object
    my %stat = Dense::dense_stat(
        host=> $self->Filer->hostspec, 
        path=>$self->{_Volume}->mount_path,
        block => $args{block},
        long => 1,
        diag_mode => 1,
    );

    # The above function may return undef with no error
    # so croak on error if it exists
    croak $@ if $@;
        
    my $stat = \$stat{$self->{_Volume}->mount_path};

    return $stat;
}


=head2 B<Waiting>

=head2 ::status_wait

Wait for the sis to reach specific states.

Note that there are many sis status elements,
one of which is thoughtfully named 'status'.
Another one is named 'state'.
There are none however named 'element' (not yet, anyway).

The status elements checked are the same ones
that may appear in the hash generated by L<C<::status>|/::status>.
(Internally C<::status> is used; this is a convenience method.)

Because you are a valued NOTE user,
the values for the status elements may be given
as regular expressions.

=over

=item Arguments

I<status element name> => I<status element value> | I<regex>

I<Required.>
The status elements to check are given as named arguments.
Their values may be given as literal strings
or regular expressions that (hopefully) match valid values.

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 5 seconds.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sis->status_wait(
         status => 'idle'
     );

     $sis->status_wait(
         status => 'active'
     );

     $sis->status_wait(
         'blocks-de-duplicated' => '\d{4}'
     );

=cut

sub status_wait {
    my ($self, %args) = @_;
    my ($wait_for, $matchall, $havekeys, $status, $etime, $delta);
    my $sleepis = 5;    # sleep number of seconds between status read

    if( exists $args{'timeout'} ) {
        $etime = $args{'timeout'};
        delete $args{'timeout'};
    } else {
        $etime = $self->{_timeout};
    }

    if( exists $args{'sleep'} ) {
        $sleepis = $args{'sleep'};
        delete $args{'sleep'};
    }

    $wait_for = ''; # Show what we are waiting for
    foreach( keys %args ) { $wait_for .= "$_ =~ /$args{$_}/\n" }

    $matchall = 0;
    $havekeys = scalar( keys %args );

    # Set expiration time
    $etime += time;

    while( $matchall < $havekeys ) {
        NOTE::logcomment("status_wait:\n$wait_for");

        croak "Status Timeout waiting for:\n$wait_for" if time > $etime;

        $status = $self->status();
        $matchall = 0;

        # check keys one by one and match all of them
        foreach( keys %args ) {
            my $item =  $status->{lc($_)};
            my $expected = lc($args{$_});
            $matchall++ if( $item =~ qr/$expected/i );
        }

        # Do not wait anymore if everything has been found
        last unless ( $matchall < $havekeys );

        # Wait for next iteration
        $delta = $etime - time;
        $delta = 0 if $delta < 0;
        $sleepis = $delta if $delta < $sleepis;
        sleep $sleepis if( $sleepis > 0 );
    }
    
    return;
}

=head2 ::idle_wait

Waits for the sis on the volume to become idle.

=over

=item Arguments

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 5 seconds.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sis->idle_wait;

=cut

sub idle_wait {
    my ($self, %args) = @_;
    my (%sis_arg);
    $args{delay}    = 5;
    $args{complete} = 1;

    # Process the sleep option
    if( exists $args{'sleep'} ) {
        $args{delay} = $args{'sleep'};
        delete $args{'sleep'};
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    Dense::dense_wait_for_status(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path,
        time_gap=>$args{delay},
        time_out=>$args{timeout},
        status => 'IDLE');

    Dense::dense_verify_status_error(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path,
        verify_error=>"-");

    # Check for error
    croak $@ if $@;
}


=head2 ::active_wait

Waits for the sis on the volume to become active.

=over

=item Arguments

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 1 second.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sis->active_wait;

=cut

sub active_wait {
    my ($self, %args) = @_;
    my (%sis_args);
    $args{delay}    = 1;
    $args{complete} = 1;

    # Process the sleep option
    if( exists $args{'sleep'} ) {
        $args{delay} = $args{'sleep'};
        delete $args{'sleep'};
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    Dense::dense_wait_for_status(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path,
        time_gap=>$args{delay},
        time_out=>$args{timeout},
        status => 'ACTIVE');

    Dense::dense_verify_status(
        host=> $self->Filer->hostspec, 
        path_name=>$self->{_Volume}->mount_path,
        verify_against=>"ACTIVE");

    # Check for error
    croak $@ if $@;
}

=head2 B<Component Objects>

=head2 ::Volume

Returns the Volume object.

=over

=item Arguments

None.

=item Examples

=back

     $src = $sis->Volume;

=cut

sub Volume {
    my $self = shift;
    return $self->{_Volume};
}

=head2 ::Filer

Returns the Filer object.

=over

=item Arguments

None.

=item Examples

=back

     $src = $sis->Filer;

=cut

sub Filer {
    my $self = shift;
    return $self->{_Filer};
}


=head2 B<Names and Paths>

=head2 ::path

Returns the mount_path of corresponding Volume.

=over

=item Arguments

None.

=item Examples

=back

     $path = $sis->path;

=cut

sub path {
    my $self = shift;
    return $self->{_Volume}->mount_path;
}


=head2 ::name

Returns the name of the Volume.

=over

=item Arguments

None.

=item Example

=back

     $srcname = $sis->name;

=cut

sub name {
    my $self = shift;
    return $self->{_Volume}->name;
}

=head2 B<Object Identification>

=head2 ::description

Returns a description of the Sis Object
consisting of the tag volume,
a rightward pointing arrow,
and the tag for the destination.

=over

=item Arguments

None.

=item Example

=back

     $desc = $sis->description;

=cut

sub description {
    my $self = shift;
    my $desc = "SIS OBJ of ".$self->Volume->description;
    return $desc;
}

=head2 B<Object Configuration>

=head2 ::timeout

Sets or returns the current default timeout for the object before being set.
A value of -1 is to wait forever.

=over

=item Arguments

B<timeout> => I<number>

I<Optional.>
Set default timeout for the object.

=item Example

=back

     $curr_timeout = $sis->timeout;    # Get current default timeout

     $orig_timeout = $sis->timeout(
         timeout => 300               # Set default timeout
     );

=cut

sub timeout {
    my ($self, %args) = @_;

    $self->{_timeout} = $args{timeout} if exists $args{timeout};
    $self->{_timeout} = 600 if ! defined $self->{_timeout};
    return $self->{_timeout};
}

=head2 B<Object Destruction>

=head2 ::destroy

Destroy the sis by unconfiguring it on the volume.

=over

=item Arguments

None.

=item Example

=back

     $sis->destroy;

=cut


=head2 ::reallocate_start

Enable and start reallocate on specified sis volume

=over

=item Arguments

B<sched> - I<string>

Optional

If set will have this schedule for reallocate operation.
If not set, schedule is "10 * * *".

=item Example

=back

     $sis->reallocate_start;

=cut

sub reallocate_start {
    my ($self, %args) = @_;
    my $sched = (exists $args{sched} ? $args{sched} : "10 * * *");
    my $cmd   = "reallocate schedule -s";
    my $vol_path = $self->{_Volume}->mount_path;
    my $out;
    $out = $self->Filer->say("reallocate on");
    if( $out =~ /Reallocation scans are on/i) {
    NOTE::logcomment("Reallocation is enabled\n");
    }else{
      croak "FAIL::Unable to enable reallocate\n";
      }
    NOTE::logcomment("Starting reallocate on ASIS volume %s",$vol_path);
    $out =  $self->Filer->say("reallocate start $vol_path");
    if( $out =~ /Reallocation scan will be started on/i) {
    NOTE::logcomment("Reallocation is started on volume $vol_path\n");
    }else{
      croak "FAIL::Unable to start reallocate on volume $vol_path\n";
      }
    eval {
      $self->Filer->say( $cmd . " " . "'" . $sched . "'" . " " . $vol_path );
    };
    croak "FAIL::Unable to schedule for reallocation due to: $@\n" if ($@);
    eval { $self->Filer->say("reallocate status -v $vol_path"); };
    croak "FAIL::Unable to check the reallocate status due to: $@\n" if ($@);
    return;
   }



sub destroy {
    my $self = shift;
    my $conf;
    $self->{_destroyed} = 1;

    NOTE::logcomment( "Destroying sis object " . $self->description );

    # Abort any transfer currently active, ignore error
    $self->stop unless ($self->{_stopped} || $self->status->{"status"} =~ /idle/i );

    $self->status_wait(
        status => 'idle'
    );

    NOTE::logcomment("Sis removed from ".$self->Volume->description);

    return;
}

# Object destructor
sub DESTROY {
    my $self = shift;

    # Only the creator can destroy the object
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_destroyed} );   # Already dead
    return if ( NOTE->ALL_SPARED );

    # Destroy object
    $self->destroy;
}

=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::main::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>

=head1 STATUS

Review

$Id: //depot/prod/test/main/lib/NOTE/Sis.pm#1 $

$Date: 2009/02/06 $

=head1 AUTHOR

indra@netapp.com, dl-dpg-nb-automation@netapp.com

=cut

1;

