#
#$Id: //depot/prod/test/main/lib/NOTE/Vfiler.pm#19 $
#
## @summary  NOTE Vfiler class
## @status   public
## @pod      here
## @author   saurabhg@netapp.com, dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::Vfiler

=head1 DESCRIPTION

This is the NOTE class for working with Vfilers.

=head1 METHODS

=cut

package NOTE::Vfiler;
use strict;
use warnings;
no warnings qw(redefine);
use Data::Dumper;
use VFILER::storagemotion;
use Params::Validate qw[validate OBJECT SCALAR];
use constant FEATURE => q[multistore];
use constant RSH     => q[#rsh-socket];
use RegistryCmds qw[getOption setOption registryGet registrySet
    registryDelete registryWalk];

our $iparegex = qr/^(\d\d?\d?)\.(\d\d?\d?)\.(\d\d?\d?)\.(\d\d?\d?)$/;

BEGIN {
    our @EXPORT_OK = qw(
        add
        attach
        basefiler
        conn
        connection
        create
        destroy
        getoption
        ipa
        name
        password
        purge
        remove
        rename
        run
        setoption
        start
        stop
        username
        fill
        exportfs
        migrate_prepare
    );
} ## end BEGIN

BEGIN {

    # Die with a script error when an input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess( Tharn::script_result(@msgs) );
        }
    );
} ## end BEGIN

our $NETMASK       = '255.255.255.128';
our $vfuser        = 'vftestuser';
our $vfpasswd      = 'vftestuser123';
our $nfsMountPoint = undef;

=head2 ::create

Constructor for Vfiler class, to create a new Vfiler on the given hosting filer.

If Vfiler is to be created in non-default IPspace, then the gateway also needs to
be specified.

=over

=item Arguments

B<name> => I<string>

I<Required.>
Name of the Vfiler to be created

B<interface> => I<string>

I<Required.>
Interface to be used for binding Vfiler

B<ipaddr> => I<string>

I<Required.>
IP address to be assigned to Vfiler

B<Rootvol> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path where Vfiler root volume should reside.
Object created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<netmask> => I<string>

I<Optional.>
Netmask to be assigned to Vfiler. The default is 255.255.255.128

B<username> => I<string>

I<Optional.>
Vfiler user to be used for access. The default username is vftestuser.

B<password> => I<string>

I<Optional.>
Vfiler user password to be used for access. The default user's password is vftestuser123.

B<ipspace> => I<string>

I<Optional.>
IP space to be used for Vfiler in order to use non-default IP space.

B<gateway> => I<string>

I<Optional.>
Gateway to be used for the Vfiler when using non-default IP space.

B<v6_address> => I<string>

I<Optional.>
The IPv6 address that needs to be configured on the Vfiler

B<prefixlen> => I<string>

I<Optional.>
The IPv6 prefixlen that needs to be configured on the Vfiler. Default is 64.

B<purge_existing> => I<0|1>

I<Optional.>
To purge an existing vfiler with same name. The default value is 1.

B<recycle> => I<0|1>

I<Optional.>
To avoid destroying the Vfiler created for reuse. The default value is 0.

=item Return Value

Vfiler object on successful creation, undef otherwise.

=item Examples

=back

    NOTE::Vfiler->create(
        name           => 'vfiler101',
        interface      => 'e0a',
        ipaddr         => '10.72.192.84',
        Rootvol       =>  $obj,
        );

     NOTE::Vfiler->create(
        name           => 'vfiler101',
        interface      => 'e0a',
        ipaddr         => '10.72.192.84',
        Rootvol       =>  $obj,
        netmask        => '255.255.255.0',
        username       => 'test',
        password       => 'test1test',
        purge_existing => 1,
        recycle        => 0
        );

     NOTE::Vfiler->create(
        name           => 'vfiler101',
        interface      => 'e0a',
        ipaddr         => '10.72.192.84',
        Rootvol       =>  $obj,
        ipspace        => 'ipsxyz'
        gateway        => '10.72.230.1'
        purge_existing => 1,
        recycle        => 0
        );

=cut

sub create {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $class = shift;
    my %args  = validate(
        @_,
        {   name      => { type => SCALAR, optional => 0 },
            interface => { type => SCALAR, optional => 0 },
            ipaddr    => {
                type     => SCALAR,
                optional => 0,
                regex    => $iparegex
            },
            netmask => {
                type     => SCALAR,
                optional => 1,
                regex    => $iparegex,
                default  => $NETMASK
            },
            Rootvol  => { type => OBJECT, optional => 0 },
            username => {
                type => SCALAR,
                ,
                optional => 1,
                default  => $vfuser,
                depends  => ['password']
            },
            password =>
                { type => SCALAR,, optional => 1, default => $vfpasswd },
            purge_existing => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^[01]$/,
                default  => 1
            },
            recycle => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^[01]$/,
                default  => 0
            },
            ipspace => {
                type     => SCALAR,
                optional => 1,
                default  => 'default-ipspace',
                depends  => ['gateway']
            },
            gateway => {
                type     => SCALAR,
                optional => 1,
                regex    => $iparegex
            },
            v6_address =>
                { type => SCALAR, optional => 1, depends => ['prefixlen'] },
            prefixlen => {
                type     => SCALAR,
                optional => 1,
                default  => 64,
                regex    => qr/^\d+$/
                }

        }
    );

    my $self = {};
    $self->{_owner} = $$;

    $self->{_name}           = $args{name};
    $self->{_if}             = $args{interface};
    $self->{_ipa}            = $args{ipaddr};
    $self->{_netmask}        = $args{netmask};
    $self->{_Rootvol}        = $args{Rootvol};
    $self->{_username}       = $args{username};
    $self->{_password}       = $args{password};
    $self->{_ipspace}        = $args{ipspace};
    $self->{_gateway}        = $args{gateway};
    $self->{_purge_existing} = $args{purge_existing};
    $self->{_recycle}        = $args{recycle};
    $self->{_prefixlen}      = $args{prefixlen};
    $self->{_v6_address}     = $args{v6_address};

    $self->{_pfiler} = $args{Rootvol}->Filer->host;

    my $pfiler
        = NOTE::Filer->connect( host => $self->{_pfiler}, mode => '#telnet' );

    $pfiler->check_license(FEATURE);

    if ( VFILER::Setup::vfiler_present( $pfiler->conn, $self->{_name} ) ) {
        if ( $self->{_purge_existing} ) {
            VFILER::Setup::vfiler_stop( $pfiler->conn, $self->{_name} );
            sleep(5);
            VFILER::Setup::vfiler_destroy( $pfiler->conn, $self->{_name} );
        } else {
            NOTE::logcomment( "Vfiler "
                    . $self->{_name}
                    . " already exists on filer "
                    . $self->{_pfiler} );

            NOTE::logcomment( "Attaching to Vfiler "
                    . $self->{_name}
                    . " with username "
                    . $self->{_username}
                    . " & password "
                    . $self->{_password} );
            return NOTE::Vfiler->attach(
                filer    => $self->{_pfiler},
                name     => $self->{_name},
                username => $self->{_username},
                password => $self->{_password}
            );

        } ## end else [ if ( $self->{_purge_existing...
    } ## end if ( VFILER::Setup::vfiler_present...

    my $pathObj = $self->{_Rootvol};

    unless ( defined($pathObj) ) {
        NOTE::logcomment( "Invalid path specified for Vfiler root volume : "
                . $self->{_Rootvol} );
        return undef;
    }

# Changed by saurabhg on 21 Nov 2010 - not needed as using 'depends' keyword in validate
#    if (   ( $self->{_username} ne $vfuser )
#        && ( $self->{_password} eq $vfpasswd ) )
#    {
#        NOTE::logcomment(
#"Password should be specified if non-default Vfiler user is being used. Create failed!"
#        );
#        return undef;
#    } ## end if ( ( $self->{_username...

    if ( $self->{_ipspace} ne 'default-ipspace' ) {

#        if ( !defined( $self->{_gateway} ) ) {
#            NOTE::logcomment(
#"Gateway should be defined if non-default IPspace is being used. Create failed!"
#            );
#            return undef;
#        } ## end if ( !defined( $self->...

        NOTE::logcomment("Creating Vfiler in non-default IPspace");
        VFILER::Setup::setup_vfilers(
            $pfiler->conn,
            {   vfiler         => $self->{_name},
                ip             => $self->{_if},
                ipaddr         => $self->{_ipa},
                netmask        => $self->{_netmask},
                path           => $pathObj->mount_path,
                ipspace        => $self->{_ipspace},
                gateway        => $self->{_gateway},
                purge_existing => $self->{_purge_existing}
            }
        );
    } else {
        NOTE::logcomment("Creating Vfiler in default IPspace");
        $self->{_ipspace} = 'default-ipspace';
        $self->{_gateway} = _getgateway($pfiler);
        VFILER::Setup::setup_vfilers(
            $pfiler->conn,
            {   vfiler         => $self->{_name},
                ip             => $self->{_if},
                ipaddr         => $self->{_ipa},
                netmask        => $self->{_netmask},
                path           => $pathObj->mount_path,
                purge_existing => $self->{_purge_existing}
            }
        );
    } ## end else [ if ( $self->{_ipspace}...

    if ( _ping( $pfiler, $self->{_ipa} ) ) {
        NOTE::logcomment( "Vfiler "
                . $self->{_name}
                . " successfully created on filer "
                . $self->{_pfiler} );

        bless( $self, $class );
        NOTE::reference($self);
        croak $@ if $@;

        # Now enable IPv6 for this Vfiler unit
        if ( defined( $self->{_v6_address} ) ) {
            $self->configure_ipv6(
                v6_address => $self->{_v6_address},
                interface  => $self->{_if},
                prefixlen  => $self->{_prefixlen}
            );
        } ## end if ( defined( $self->{...

        $self->connect;
        return $self;
    } else {
        NOTE::logcomment( "Failed to create vfiler "
                . $self->{_name}
                . " on filer "
                . $self->{_pfiler} );
        return undef;
    } ## end else [ if ( _ping( $pfiler, $self...
} ## end sub create

=head2 ::attach

To attach to an existing Vfiler

=over

=item Arguments

B<filer> => I<hostspec>

I<Required.>
hostspec of the hosting filer

B<name> => I<string>

I<Required.>
Name of the Vfiler to be attached to

B<username> => I<string>

I<Required.>
Vfiler user to be used for access.

B<password> => I<string>

I<Required.>
Vfiler user password to be used for access.

=item Return Value

Vfiler object on success, undef otherwise

=item Examples

=back

    NOTE::Vfiler->attach(
        filer          => 'f3070-192-95',
        name           => 'vfiler101',
        username       => 'test',
        password       => 'test1test'
        );

=cut

sub attach {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $class = shift;
    my %args  = validate(
        @_,
        {   name     => { type => SCALAR, optional => 0 },
            username => { type => SCALAR, optional => 0 },
            password => { type => SCALAR, optional => 0 },
            filer    => { type => SCALAR, optional => 0 }
        }
    );

    my $self = {};
    $self->{_owner} = $$;

    $self->{_name}     = $args{name};
    $self->{_username} = $args{username};
    $self->{_password} = $args{password};
    $self->{_pfiler}   = $args{filer};

    my $pfiler
        = NOTE::Filer->connect( host => $self->{_pfiler}, mode => '#rsh' );

    my %sts = _status_a( $pfiler->conn, $self->{_name} );

    if ( $sts{status} !~ /Running/i ) {
        NOTE::logcomment( $sts{status} );
        return undef;
    }

    $self->{_if}      = $sts{if};
    $self->{_ipa}     = $sts{ipaddr};
    $self->{_Rootvol} = $sts{path};
    $self->{_ipspace} = $sts{ipspace};
    $self->{_netmask}
        = _getnetmask( $self->{_pfiler}, $self->{_ipa}, $self->{_if} );
    $self->{_gateway}        = _getgateway($pfiler);
    $self->{_v6_address}     = _getv6address( $pfiler, $self->{_name} );
    $self->{_prefixlen}      = _getprefixlen( $pfiler, $self->{_v6_address} );
    $self->{_purge_existing} = 0;
    $self->{_recycle}        = 1;

    bless( $self, $class );
    NOTE::reference($self);

    if ($@) {
        NOTE::logcomment( "Failed to attach to the vfiler "
                . $self->{_name}
                . " hosted on filer "
                . $self->{_pfiler} );
        return undef;
    } ## end if ($@)
    $self->connect;
    return $self;
} ## end sub attach

=head2 ::name

To fetch the Vfiler name

=over

=item Arguments

None

=item Return Value

String containing the name of Vfiler represented by the object.

=item Examples

=back

        $vfiler_object->name ;

=cut

sub name {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_name};
}

=head2 ::ipa

To fetch the IP Address of the Vfiler

=over

=item Arguments

None

=item Return Value

String containing the IP Address of Vfiler represented by the object.

=item Examples

=back

        $vfiler_object->ipa ;

=cut

sub ipa {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_ipa};
}

=head2 ::username

To fetch the name of the Vfiler user

=over

=item Arguments

None

=item Return Value

String containing the username of Vfiler user.

=item Examples

=back

        $vfiler_object->username ;

=cut

sub username {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_username};
}

=head2 ::password

To fetch the password of the Vfiler user

=over

=item Arguments

None

=item Return Value

String containing the password of Vfiler user.

=item Examples

=back

        $vfiler_object->password ;

=cut

sub password {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_password};
}

=head2 ::connect

To establish rsh connection to the Vfiler

=over

=item Arguments

None

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->connect ;

=cut

sub connect {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $pfiler
        = NOTE::Filer->connect( host => $self->{_pfiler}, mode => "#rsh" );

    if (_create_vfiler_user(
            $pfiler, $self, $self->{_username}, $self->{_password}
        )
        )
    {
        NOTE::logcomment( "User for Vfiler rsh is "
                . $self->{_username}
                . " with password "
                . $self->{_password} );
    } else {
        NOTE::logcomment("Unable to create vfiler rsh user, connect failed");
        return 0;
    }

    my $fullconnspec
        = $self->{_name}
        . ";hosttype=filer;default_ip="
        . $self->{_ipa} . ";"
        . RSH
        . ";username="
        . $self->{_username}
        . ";password="
        . $self->{_password}
        . ";timeout=90;";

    VFILER::Vfiler::vfiler_allow( $pfiler->conn, $self->{_name},
        rsh => "proto=rsh" );

    eval { $self->{_conn} = Tharn::connect($fullconnspec); };
    warn $@ if $@;

    ( defined( $self->{_conn} ) ) ? return 1 : return 0;
} ## end sub connect

=head2 ::conn

To return rsh connection object to the Vfiler.

=over

=item Arguments

None

=item Return Value

Tharn::connect object with rsh connection to the Vfiler

=item Examples

=back

        $vfiler_object->conn ;

=cut

sub conn {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_conn};
}

=head2 ::basefiler

To fetch hostspec hosting filer

=over

=item Arguments

None

=item Return Value

String containing the hostspec of the hosting filer represented by the object.

=item Examples

=back

        $vfiler_object->basefiler ;

=cut

sub basefiler {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_pfiler};
}

=head2 ::stop

To stop the Vfiler

=over

=item Arguments

None

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->stop ;

=cut

sub stop {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );

    if ( VFILER::Setup::vfiler_stop( $pfiler->conn, $self->name ) ) {
        NOTE::logcomment( "Vfiler " . $self->name . " stopped successfully" );
        return 1;
    } else {
        NOTE::logcomment( "Unable to stop Vfiler " . $self->name );
        return 0;
    }
} ## end sub stop

=head2 ::run

To run a command on the Vfiler

=over

=item Arguments

B<command> => I<string>

I<Required.>
Command to be run in the Vfiler context

=item Return Value

Output of the command

=item Examples

=back

        $vfiler_object->run(command => "hostname") ;

=cut

sub run {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args
        = validate( @_, { command => { type => SCALAR, optional => 0 }, } );

    my $hostname = $self->name;
    my $otpt     = "No command given to be run";

    if ( defined( $args{command} ) ) {
        NOTE::logcomment( "Running command "
                . $args{command}
                . " on the Vfiler "
                . $hostname );

        my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
        my %result = VFILER::Vfiler::vfiler_run( $pfiler->conn, $hostname,
            $args{command} );

        $otpt = $result{output};
        $otpt = substr( $otpt,
            index( $otpt, $hostname ) + 1 + length($hostname) );

        if ( $otpt =~ /.+/ ) {
            $otpt =~ s/^\s+//;
        } else {
            $otpt = "Command did not return any output";
        }
    } ## end if ( defined( $args{command...
    return $otpt;
} ## end sub run

=head2 ::destroy

To destroy a stopped Vfiler

=over

=item Arguments

None

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->destroy ;

=cut

sub destroy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );

    if ( $self->status eq "running" ) {
        NOTE::logcomment(
            "Cannot destroy a running Vfiler, please stop the Vfiler first");
        return 0;
    } elsif ( $self->status eq "Vfiler not found" ) {
        NOTE::logcomment("Cannot destroy a non-existent Vfiler");
        return 0;
    }

    if ( VFILER::Setup::vfiler_destroy( $pfiler->conn, $self->name ) ) {
        NOTE::logcomment(
            "Vfiler " . $self->name . " destroyed successfully" );
        return 1;
    } else {
        NOTE::logcomment( "Unable to destroy Vfiler "
                . $self->name
                . "Please Stop the Filer before invoking Destroy" );
        return 0;
    } ## end else [ if ( VFILER::Setup::vfiler_destroy...
} ## end sub destroy

=head2 ::getoption

To fetch value of specified option from the Vfiler

=over

=item Arguments

B<option> => I<string>

I<Required.>
Option name

=item Return Value

String containing the current value of that Vfiler option

=item Examples

=back

        $vfiler_object->getoption(option => 'ftpd.enable') ;

=cut

sub getoption {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args
        = validate( @_, { option => { type => SCALAR, optional => 0 }, } );

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
    my $result
        = VFILER::Vfiler::vfiler_options_get( $pfiler->conn, $self->name,
        $args{option} );

    if ( $result eq "" ) {
        NOTE::logcomment("Invalid option specified !");
        return undef;
    }

    return $result;
} ## end sub getoption

=head2 ::setoption

To set the value of the specified option on the Vfiler

=over

=item Arguments

Two arguments are needed.

B<option> => I<string>

I<Required.>
Name of the option to be set

B<value> => I<string>

I<Required.>
Value to be assigned to the Vfiler option

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->setoption(option => 'ftpd.enable', value => 'on') ;

=cut

sub setoption {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   option => { type => SCALAR, optional => 0 },
            value  => { type => SCALAR, optional => 0 }
        }
    );

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
    eval {
        VFILER::Vfiler::vfiler_options_set( $pfiler->conn, $self->name,
            $args{option}, $args{value} );
    };

    if ($@) {
        NOTE::logcomment(
            "Cannot set the option $args{option} to value $args{value}"
                . $@ );
        return 0;
    } ## end if ($@)

    return 1;
} ## end sub setoption

=head2 ::status

To determine current status of the Vfiler

=over

=item Arguments

None

=item Return Value

String containing the current status of the Vfiler

=item Examples

=back

        $vfiler_object->status ;

=cut

sub status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %hash;

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
    eval { %hash = _status_a( $pfiler->conn, $self->name ); };

    return $hash{status};
} ## end sub status

=head2 ::start

To start the Vfiler

=over

=item Arguments

None

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->start ;

=cut

sub start {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %hash;

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
    my $vfstatus = $self->status;

    if ( $vfstatus eq "Vfiler not found" ) {
        NOTE::logcomment("Vfiler does not exist");
        return 0;
    } elsif ( $vfstatus eq "running" ) {
        NOTE::logcomment(
            "Vfiler is already running, nothing needs to be done");
    } elsif ( $vfstatus eq "stopped" ) {
        NOTE::logcomment( "Starting Vfiler "
                . $self->name
                . "on filer "
                . $self->basefiler );
        if ( VFILER::Vfiler::vfiler_start( $pfiler->conn, $self->name ) ) {
            NOTE::logcomment( "Vfiler "
                    . $self->name
                    . " successfully started on filer "
                    . $self->basefiler );
            return 1;
        } else {
            NOTE::logcomment( "Vfiler "
                    . $self->name
                    . " failed to start on filer "
                    . $self->basefiler );
            return 0;
        } ## end else [ if ( VFILER::Vfiler::vfiler_start...
    } ## end elsif ( $vfstatus eq "stopped")
} ## end sub start

=head2 ::purge

To stop and destroy the Vfiler

=over

=item Arguments

None

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->purge ;

=cut

sub purge {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self   = shift;
    my $retval = 0;

    NOTE::logcomment( "Purging vfiler "
            . $self->name
            . " hosted on filer "
            . $self->basefiler );

    if ( $self->status eq "Vfiler not found" ) {
        NOTE::logcomment("Vfiler does not exist");
        return 1;
    }

    $retval = 1 if ( $self->stop );
    sleep(5);
    ($retval) && ( $self->destroy ) ? $retval = 1 : $retval = 0;

    return $retval;
} ## end sub purge

=head2 ::rename

To rename a Vfiler

=over

=item Arguments

B<name> => I<string>

I<Required.>
New name for the Vfiler

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

        $vfiler_object->rename(name => 'newname') ;

=cut

sub rename {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate( @_, { name => { type => SCALAR, optional => 0 }, } );

    if ( $self->status eq "Vfiler not found" ) {
        NOTE::logcomment("Vfiler does not exist");
        return 0;
    }

    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
    my $retval = 0;
    eval {
        $retval
            = VFILER::Vfiler::vfiler_rename( $pfiler->conn, $self->name,
            $args{name} );
    };

    if ($retval) {
        $self->{_name} = $args{name};
        return 1;
    } else {
        return 0;
    }
} ## end sub rename

=head2 ::add

To add resources to the Vfiler

=over

=item Arguments

B<vol> => I<Volume or Qtree object>

I<Optional.>
Volume or Qtree objected that is attached to the path where Vfiler root volume should reside.
Object created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<ipaddr> => I<string>

I<Optional.>
IP Address to be added to the Vfiler

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

       $vfiler_object->add(ipaddr => '10.72.10.84'
                           vol => $vol);

       $vfiler->add(ipaddr => '10.72.10.84' );

       $vfiler_object->add(vol => $vol);

=cut

sub add {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   vol    => { type => OBJECT, optional => 1 },
            ipaddr => { type => SCALAR, optional => 1 },
        }
    );

    my $filer = NOTE::Filer->connect( host => $self->basefiler );

    if ( ( !defined( $args{'vol'} ) ) && ( !defined( $args{'ipaddr'} ) ) ) {
        NOTE::logcomment("Neither path or IP address to be added specified!");
        return 0;
    } elsif ( ( defined( $args{'vol'} ) ) && ( !defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment(
            "Adding resource " . $args{'vol'}->mount_path . " to Vfiler" );
        eval {
            VFILER::Vfiler::vfiler_add(
                $filer->conn, $self->name,
                option => "-f",
                path   => $args{'vol'}->mount_path
            );
        };
        if ($@) {
            warn $@;
            return 0;
        }
    } elsif ( ( !defined( $args{'vol'} ) ) && ( defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment(
            "Adding resource " . $args{'ipaddr'} . " to Vfiler" );
        eval {
            VFILER::Vfiler::vfiler_add(
                $filer->conn, $self->name,
                option => "-f",
                ipaddr => "-i " . $args{'ipaddr'}
            );
        };
        if ($@) {
            warn $@;
            return 0;
        }
    } elsif ( ( defined( $args{'vol'} ) ) && ( defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment( "Adding resources "
                . $args{'vol'}->mount_path . " & "
                . $args{'ipaddr'}
                . " to Vfiler" );
        eval {
            VFILER::Vfiler::vfiler_add(
                $filer->conn, $self->name,
                option => "-f",
                ipaddr => "-i " . $args{'ipaddr'},
                path   => $args{'vol'}->mount_path

            );
        };
        if ($@) {
            warn $@;
            return 0;
        }

    } else {
        return 0;
    }
    return 1;
} ## end sub add

=head2 ::remove

To remove resources from the Vfiler

=over

=item Arguments

B<vol> => I<Volume or Qtree object>

I<Optional.>
Volume or Qtree objected that is attached to the path where Vfiler root volume should reside.
Object created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<ipaddr> => I<string>

I<Optional.>
IP Address to be added to the Vfiler

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

       $vfiler_object->remove( ipaddr => '10.72.10.84'
                               vol => $vol);

       $vfiler->remove(ipaddr => '10.72.10.84' );

       $vfiler_object->remove(vol => $vol);

=cut

sub remove {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   vol    => { type => OBJECT, optional => 1 },
            ipaddr => { type => SCALAR, optional => 1 },
        }
    );

    my $filer = NOTE::Filer->connect( host => $self->basefiler );

    if ( ( !defined( $args{'vol'} ) ) && ( !defined( $args{'ipaddr'} ) ) ) {
        NOTE::logcomment(
            "Neither path or IP address to be removed specified!");
        return 0;
    } elsif ( ( defined( $args{'vol'} ) ) && ( !defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment( "Removing resource "
                . $args{'vol'}->mount_path
                . " from the Vfiler" );
        eval {
            VFILER::Vfiler::vfiler_remove(
                $filer->conn, $self->name,
                option => "-f",
                paths  => $args{'vol'}->mount_path
            );
        };
        if ($@) {
            warn $@;
            return 0;
        }
    } elsif ( ( !defined( $args{'vol'} ) ) && ( defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment(
            "Removing IP address " . $args{'ipaddr'} . " from the Vfiler" );
        eval {
            VFILER::Vfiler::vfiler_remove(
                $filer->conn, $self->name,
                option  => "-f",
                ipaddrs => "-i " . $args{'ipaddr'}
            );
        };
        if ($@) {
            warn $@;
            return 0;
        }
    } elsif ( ( defined( $args{'vol'} ) ) && ( defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment( "Removing resource "
                . $args{'vol'}->mount_path
                . " & IP address "
                . $args{'ipaddr'}
                . " from the Vfiler" );
        eval {
            VFILER::Vfiler::vfiler_remove(
                $filer->conn, $self->name,
                option  => "-f",
                ipaddrs => "-i " . $args{'ipaddr'},
                paths   => $args{'vol'}->mount_path

            );
        };
        if ($@) {
            warn $@;
            return 0;
        }

    } else {
        return 0;
    }
    return 1;
} ## end sub remove

=head2 ::move

To move resources from one Vfiler to another Vfiler

=over

=item Arguments

B<tovfiler> => I<Vfiler object>

I<Required.>
NOTE::Vfiler object for the Vfiler to which the resources will be moved to. Note that
this method needs to be called from Vfiler object that is relinquishing the resources.

B<vol> => I<Volume or Qtree object>

I<Optional.>
Volume or Qtree objected that is attached to the path where Vfiler root volume should reside.
Object created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<ipaddr> => I<string>

I<Optional.>
IP Address to be added to the Vfiler

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

       $vfiler_object->move(  tovfiler => $to_vfiler,
                              ipaddr => '10.72.10.84'
                              vol => $vol);

       $vfiler->move(   tovfiler => $to_vfiler,
                        ipaddr => '10.72.10.84' );

       $vfiler_object->move(   tovfiler => $to_vfiler,
                               vol => $vol);

=cut

sub move {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   tovfiler => { type => OBJECT, optional => 0 },
            vol      => { type => OBJECT, optional => 1 },
            ipaddr   => { type => SCALAR, optional => 1 },
        }
    );

    my $vfto  = $args{tovfiler};
    my $filer = NOTE::Filer->connect( host => $self->basefiler );
    my $moved = 0;

    if ( ( !defined( $args{'vol'} ) ) && ( !defined( $args{'ipaddr'} ) ) ) {
        NOTE::logcomment("Neither path or IP address to be moved specified!");
        return 0;
    } elsif ( ( defined( $args{'vol'} ) ) && ( !defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment( "Moving resource "
                . $args{'vol'}->mount_path
                . " from the Vfiler "
                . $self->name );
        $moved = $self->remove( vol => $args{'vol'} );
        $moved = $vfto->add( vol => $args{'vol'} );

    } elsif ( ( !defined( $args{'vol'} ) ) && ( defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment(
            "Moving IP address " . $args{'ipaddr'} . " from the Vfiler" );
        $moved = $self->remove( ipaddr => $args{'ipaddr'} );
        $moved = $vfto->add( ipaddr => $args{'ipaddr'} );

    } elsif ( ( defined( $args{'vol'} ) ) && ( defined( $args{'ipaddr'} ) ) )
    {
        NOTE::logcomment( "Moving resource "
                . $args{'vol'}->mount_path
                . " & IP address "
                . $args{'ipaddr'}
                . " from the Vfiler" );
        $moved = $self->remove(
            vol    => $args{'vol'},
            ipaddr => $args{'ipaddr'}
        );
        $moved = $vfto->add(
            vol    => $args{'vol'},
            ipaddr => $args{'ipaddr'}
        );

    } else {
        return 0;
    }
    return $moved;
} ## end sub move

sub DESTROY {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return if ( $$ != $self->{_owner} );

    return if ( NOTE->ALL_SPARED );

    my $pfiler = NOTE::Filer->connect( host => $self->{_pfiler} );

        if ( defined($nfsMountPoint) ) {
            NOTE::logcomment(
                "Trying to unmount the mounted path " . $nfsMountPoint );
            GenericMount::umount( $nfsMountPoint, force => "1" );
            $nfsMountPoint = undef;
        } ## end if ( defined($nfsMountPoint...

    if ( $self->{_recycle} ) {
        NOTE::logcomment( "Sparing vfiler "
                . $self->{_name}
                . " hosted on filer "
                . $self->{_pfiler} );
    } else {
        $self->purge;
    }
} ## end sub DESTROY

=head2 ::fill

To fill the specified Vfiler volume with data over NFS protocol. This method exports the volume
path then mounts it before running the traffic.

=over

=item Arguments

B<client> => I<hostspec of the Linux Client>

I<Required.>
Hostspec of the client to be used for running the dd command to fill the Vfiler path

B<container> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path which is to be filled with data. Object
created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<fillsize> => I<String>

I<Optional.>
Size in KB to be filled. if nothing is specified all the available space is filled.

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

   NOTE::Vfiler->fill( container =>  $obj, client => 'lnx205-140' );
   NOTE::Vfiler->fill( container =>  $obj, client => 'lnx205-140', fillsize => 1000);

=cut

sub fill {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   container => { type => OBJECT, optional => 0 },
            client    => { type => SCALAR, optional => 0 },
            fillsize  => { type => SCALAR, optional => 1, default => 0 }
        }
    );

    my $mntpnt = $self->mount( container => $args{'container'} );

    Carp::croak(
        "Failed to mount the path " . $args{'container'}->mount_path )
        unless ( defined($mntpnt) );

    my $avlblKB = $args{'container'}->avail;
    my $tofill  = 0;

    if ( ( $args{'fillsize'} == 0 ) || ( $args{'fillsize'} >= $avlblKB ) ) {
        NOTE::logcomment(
            "Specified fill size is more than available free space. Hence, filling up to maximum available freespace"
        ) if ( $args{'fillsize'} >= $avlblKB );
        $tofill = $avlblKB;
    } else {
        $tofill = $args{'fillsize'};
    }
    NOTE::logcomment( "Disk space to be filled is " . $tofill . " KB !" );
    my $fname = "Vfiler_fill_" . time();

    my $host    = Tharn::host( $args{'client'} );
    my $lnxclnt = Tharn::connect('#rsh');

    Tharn::say( $lnxclnt,
              "/bin/dd if=/dev/urandom of=" 
            . $fname . " bs=" 
            . $tofill
            . "k count=1" );

    return 1;

} ## end sub fill

=head2 ::exportfs

To export the specified Vfiler path so that they can be mounted through NFS protocol

=over

=item Arguments

B<container> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path which is to be exported. Object created
through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

=item Return Value

1 on success, croaks with error message otherwise

=item Examples

=back

       $vfiler_object->exportfs( conatiner => $obj);

=cut

sub exportfs {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args
        = validate( @_, { container => { type => OBJECT, optional => 0 } } );

    my $filer = NOTE::Filer->connect( host => $self->basefiler );
    my $path = $args{'container'}->mount_path;

    eval {
        VFILER::Vfiler::vfiler_run_vfiler_exportfs( $filer->conn, $self->name,
            $path );
    };

    Carp::croak( "Unable to export the path" . $path . "\n" . $@ ) if ($@);

    NOTE::logcomment( "Successfully exported path " . $path . " for mount" );
    return 1;
} ## end sub exportfs

=head2 ::mount

To mount the specified Vfiler path. This method exports the volume path specified before
attempting to mount it.

=over

=item Arguments

B<container> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path which is to be exported. Object created
through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<chmod> => I<string>

I<Optional.>
The mode (permissions) to set on the mountpoint after mounting. The chmod is only performed if
the mountpoint was not already mounted. The mode should be an octal number inside a string.

Default permission are "0755".

B<timeout> => I<string>

I<Optional.>
How many seconds to wait for a mount or umount attempt to finish. The default is 30 seconds.
"mount" may take up to mountretry of these periods before it completely fails. This options
is used to set the timeout on the say command.

B<mountpoint> => I<string>

I<Optional.>
The directory to mount this file system onto. The default is "/t/server/path/{exclusive|shared}/extra"
where "server" is the given "$server" value, and "path" is a string generated from the "$path", followed
by the sharing type of the mount, and then OS dependent options used to ensure that the mountpoint is unique.
As described in "Sharing of Mountpoints", above, one should generally let "mount" choose the value of this
option or else do at least as good a job of avoiding ever having separate tests choose identical mountpoint
paths but non-identical mount options. If a mountpoint is specified it should not contain any of the sharing
options {shared|exclusive}.

=item Return Value

mountpoint on success, croaks on failure

=item Examples

=back

       $vfiler_object->mount( container => $obj );
       $vfiler_object->mount( container => $obj, chmod => "0755", timeout => 100, mountpoint => "/mnt/somedir" );

=cut

sub mount {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   container => { type => OBJECT, optional => 0 },
            chmod     => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^\d+$/,
                default  => "0755"
            },
            timeout => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^\d+$/,
                default  => 30
            },
            mountpoint => { type => SCALAR, optional => 1 },
        }
    );

    $self->exportfs( container => $args{'container'} );
    eval {
        if ( exists( $args{'mountpoint'} ) )
        {
            $nfsMountPoint = GenericMount::mount(
                $self->ipa,
                $args{'container'}->mount_path,
                {   chmod      => $args{'chmod'},
                    timeout    => $args{'timeout'},
                    mountpoint => $args{'mountpoint'},
                    sharing    => "exclusive"
                }
            );
            croak( "Unable to mount at the specified mountpoint "
                    . $args{'mountpoint'} )
                if ( ( !defined($nfsMountPoint) )
                || ( $args{'mountpoint'} ne $nfsMountPoint ) );
        } else {
            $nfsMountPoint = GenericMount::mount(
                $self->ipa, $args{'container'}->mount_path,
                chmod   => $args{'chmod'},
                timeout => $args{'timeout'}
            );
        } ## end else [ if ( exists( $args{'mountpoint'...
    };

    Carp::croak( "Error in mounting:\n " . $@ ) if ($@);
    NOTE::logcomment( "Successfully mounted path "
            . $args{'container'}->mount_path . "to "
            . $nfsMountPoint );

    return $nfsMountPoint;
} ## end sub mount

=head2 ::nfs_hammer

To run hammer over NFS on the specified Qtree or Volume

=over

=item Arguments

B<container> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path which is to be filled with data. Object
created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<runtime> => I<String>

I<Required.>
Time in seconds for which hammer has to be run.

=item Examples

=back

   $vfiler_obect->nfs_hammer( container =>  $obj, runtime => 100 );

=cut

sub nfs_hammer {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   container => { type => OBJECT, optional => 0 },
            runtime   => { type => SCALAR, optional => 0, regex => qr/^\d+$/ }
        }
    );

    my $mntpnt = $self->mount( container => $args{'container'} );

    NOTE::logcomment("Starting NFS Hammer");

    my $cmdHammer
        = "/usr/local/bin/hammer " 
        . $mntpnt
        . " -nofilercore -mode mixed -runtime "
        . $args{'runtime'};

    my $host    = Tharn::host('localhost');
    my $lnxclnt = Tharn::connect('#rsh');
    Tharn::say( $lnxclnt, $cmdHammer, -noprompt );
    return 1;

} ## end sub nfs_hammer

=head2 ::cf_create

Constructor for Vfiler class, to create a new Vfiler on a cf partner in giveback or takeover mode.

If Vfiler is to be created in non-default IPspace, then the gateway also needs to
be specified.

=over

=item Arguments

B<name> => I<string>

I<Required.>
Name of the Vfiler to be created

B<interface> => I<string>

I<Required.>
Interface to be used for binding Vfiler

B<ipaddr> => I<string>

I<Required.>
IP address to be assigned to Vfiler

B<Rootvol> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path where Vfiler root volume should reside.
Object created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<netmask> => I<string>

I<Optional.>
Netmask to be assigned to Vfiler. The default is 255.255.255.128

B<username> => I<string>

I<Optional.>
Vfiler user to be used for access. The default username is vftestuser.

B<password> => I<string>

I<Optional.>
Vfiler user password to be used for access. The default user's password is vftestuser123.

B<ipspace> => I<string>

I<Optional.>
IP space to be used for Vfiler in order to use non-default IP space.

B<gateway> => I<string>

I<Optional.>
Gateway to be used for the Vfiler when using non-default IP space.

B<purge_existing> => I<0|1>

I<Optional.>
To purge an existing vfiler with same name. The default value is 1.

B<recycle> => I<0|1>

I<Optional.>
To avoid destroying the Vfiler created for reuse. The default value is 0.

=item Return Value

Vfiler object on successful creation, undef otherwise.

=item Examples

=back

    NOTE::Vfiler->cf_create(
        name           => 'vfiler101',
        interface      => 'e0a',
        ipaddr         => '10.72.192.84',
        Rootvol       =>  $obj,
        );

     NOTE::Vfiler->cf_create(
        name           => 'vfiler101',
        interface      => 'e0a',
        ipaddr         => '10.72.192.84',
        Rootvol       =>  $obj,
        netmask        => '255.255.255.0',
        username       => 'test',
        password       => 'test1test',
        purge_existing => 1,
        recycle        => 0
        );

     NOTE::Vfiler->cf_create(
        name           => 'vfiler101',
        interface      => 'e0a',
        ipaddr         => '10.72.192.84',
        Rootvol       =>  $obj,
        ipspace        => 'ipsxyz'
        gateway        => '10.72.230.1'
        purge_existing => 1,
        recycle        => 0
        );

=cut

sub cf_create {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $class = shift;
    my %args  = validate(
        @_,
        {   name      => { type => SCALAR, optional => 0 },
            interface => { type => SCALAR, optional => 0 },
            ipaddr    => {
                type     => SCALAR,
                optional => 0,
                regex    => $iparegex
            },
            netmask => {
                type     => SCALAR,
                optional => 1,
                regex    => $iparegex,
                default  => $NETMASK
            },
            Rootvol => { type => OBJECT, optional => 0 },
            username =>
                { type => SCALAR,, optional => 1, default => $vfuser },
            password =>
                { type => SCALAR,, optional => 1, default => $vfpasswd },
            purge_existing => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^[01]$/,
                default  => 1
            },
            recycle => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^[01]$/,
                default  => 0
            },
            ipspace => {
                type     => SCALAR,
                optional => 1,
                default  => 'default-ipspace'
            },
            gateway => {
                type     => SCALAR,
                optional => 1,
                regex    => $iparegex
            }
        }
    );

    my $self = {};
    $self->{_owner} = $$;

    $self->{_name}           = $args{name};
    $self->{_if}             = $args{interface};
    $self->{_ipa}            = $args{ipaddr};
    $self->{_netmask}        = $args{netmask};
    $self->{_Rootvol}        = $args{Rootvol};
    $self->{_username}       = $args{username};
    $self->{_password}       = $args{password};
    $self->{_ipspace}        = $args{ipspace};
    $self->{_gateway}        = $args{gateway};
    $self->{_purge_existing} = $args{purge_existing};
    $self->{_recycle}        = $args{recycle};

    $self->{_pfiler} = $args{Rootvol}->Filer->host;

    my $pfiler
        = NOTE::Filer->connect( host => $self->{_pfiler}, mode => '#rsh' );

    $pfiler->check_license(FEATURE);

    if ( VFILER::Setup::vfiler_present( $pfiler->conn, $self->{_name} ) ) {
        if ( $self->{_purge_existing} ) {
            VFILER::Setup::vfiler_stop( $pfiler->conn, $self->{_name} );
            sleep(5);
            VFILER::Setup::vfiler_destroy( $pfiler->conn, $self->{_name} );
        } else {
            NOTE::logcomment( "Vfiler "
                    . $self->{_name}
                    . " already exists on filer "
                    . $self->{_pfiler} );
            return undef;
        } ## end else [ if ( $self->{_purge_existing...
    } ## end if ( VFILER::Setup::vfiler_present...

    my $pathObj = $self->{_Rootvol};

    unless ( defined($pathObj) ) {
        NOTE::logcomment( "Invalid path specified for Vfiler root volume : "
                . $self->{_Rootvol} );
        return undef;
    }

    if (   ( $self->{_username} ne $vfuser )
        && ( $self->{_password} eq $vfpasswd ) )
    {
        NOTE::logcomment(
            "Password should be specified if non-default Vfiler user is being used. Create failed!"
        );
        return undef;
    } ## end if ( ( $self->{_username...

    if ( $self->{_ipspace} ne 'default-ipspace' ) {
        if ( !defined( $self->{_gateway} ) ) {
            NOTE::logcomment(
                "Gateway should be defined if non-default IPspace is being used. Create failed!"
            );
            return undef;
        } ## end if ( !defined( $self->...

        NOTE::logcomment("Creating Vfiler in non-default IPspace");
        VFILER::Setup::setup_vfilers(
            $pfiler->conn,
            {   vfiler         => $self->{_name},
                ip             => $self->{_if},
                ipaddr         => $self->{_ipa},
                netmask        => $self->{_netmask},
                path           => $pathObj->mount_path,
                ipspace        => $self->{_ipspace},
                gateway        => $self->{_gateway},
                purge_existing => $self->{_purge_existing}
            }
        );
    } else {
        NOTE::logcomment("Creating Vfiler in default IPspace");
        $self->{_ipspace} = 'default-ipspace';
        $self->{_gateway} = _getgateway($pfiler);

        VFILER::Setup::setup_vfilers(
            $pfiler->conn,
            {   vfiler         => $self->{_name},
                ip             => $self->{_if},
                ipaddr         => $self->{_ipa},
                netmask        => $self->{_netmask},
                path           => $pathObj->mount_path,
                purge_existing => $self->{_purge_existing}
            }
        );
    } ## end else [ if ( $self->{_ipspace}...

    my $if_alias
        = "ifconfig "
        . $self->{_if}
        . " alias "
        . $self->{_ipa}
        . " netmask "
        . $self->{_netmask};

    $pfiler->say($if_alias);

    #burt 456712: Making entry to /etc/rc in basefiler
    _append_tofile(
        $pfiler,
        filepath => '/etc/rc',
        data     => $if_alias
    );

    NOTE::logcomment( "Vfiler "
            . $self->{_name}
            . " successfully created on filer "
            . $self->{_pfiler} );

    bless( $self, $class );
    NOTE::reference($self);
    croak $@ if $@;

    $self->connect;
    return $self;
} ## end sub cf_create

=head2 ::migrate_prepare

To do a vfiler migrate prepare on a filer for vfilers on other filer

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.


B<remote_login> => I<login name for the dst filer>

I<Optional.>
Login for the filer where vfiler is to be moved. Defaults to login credential from tharnhost.


B<use_secure_channel> => I<use secure channel for migrate prepare>

default is false i.e vfiler migrate prepare happens through rsh.
true i.e vfiler migrate prepare happens through ssl

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

     $VFiler_NOTE1= NOTE::Vfiler->create(
        Rootvol   => $Volume1,
        name      => 'vfiler1',
        interface => 'e0a',
        ipaddr    => '10.225.167.216'
    );

$Filer1 = NOTE::Filer->connect( host => $FILER_DST);
$VFiler_NOTE1->migrate_start(filer=>$Filer1,if_to_use=>'e0a',  );
$VFiler_NOTE1->migrate_prepare(filer=>$Filer1  );

=cut

sub migrate_prepare {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $self = shift;
    my %args = @_;

    #Validating the hash
    my %args_hash = validate(
        @_,
        {   filer        => { type => OBJECT, optional => 0 },
            remote_login => {
                type     => SCALAR,
                optional => 1,
                default  => $args{'filer'}->conn->{'username'}
            },
            use_secure_channel => {
                type     => SCALAR,
                optional => 1,
                default  => 'false',
                regex    => qr/(0|1)/
            },

        }
    );
    my $dst_filer_name = $args_hash{'filer'}->host;
    my $remote_user    = $args_hash{'remote_login'};
    my $remote_passwd  = '';
    my $src_vfiler     = $self->name;
    my $src_filer_name = $self->basefiler;
    my $src_filer      = NOTE::Filer->connect( FILER => $src_filer_name );

    my $src_user   = $src_filer->conn->{'username'};
    my $src_passwd = '';
    my $secure     = $args_hash{'use_secure_channel'};
    my $return_value;
    eval {

        $return_value = VFILER::storagemotion::zapi_vfiler_migrate_prepare(
            $dst_filer_name, $remote_user, $remote_passwd,
            remote_vfiler      => $src_vfiler,
            remote_filer       => $src_filer_name,
            remote_login       => $src_user,
            remote_passwd      => $src_passwd,
            use_secure_channel => $secure
        );
    };
    die "VFiler migrate prepare could not start because of reason : $@\n"
        if ($@);
    my $count = 30;
    while ( $count > 0 ) {
 	my $status=$src_filer->say("vfiler status -a $src_vfiler");	
	
	if ( $status  =~ /migrate_prepare/ ) {  $count = 0; }
        else {
            NOTE::logcomment(
                "Waiting for vfiler status to become preparing,sleeping for 10 seconds"
            );
            sleep 10;
        } ## end else [ if ( $status =~ /migrate_prepare/)
    } ## end while ( $count > 0 )
    return $return_value;

} ## end sub migrate_prepare

=head2 ::migrate_start

To start migrate of Vfiler on one filer to another Filer.

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

B<if_to_use> => I<interface to use>

I<Required>
Interface to be used for migration

B<ipaddr> => I<Ip address >

I<Optional.>
IP address of the vfiler on destination filer. Defaults to source vfiler ip addr.

B<netmask> => I<netmask  to be used>

I<Optional.>
netmask of the vfiler on destination filer. Defaults to source vfiler netmask.

B<remote_login> => I<login name for the dst filer>

I<Optional.>
Login for the filer where vfiler is to be moved. Defaults to login credential from tharnhost.


B<remote_passwd> => I<password for the dst filer>

I<Optional.>
Password for the filer where vfiler is to be moved. Defaults to password from tharnhost.


B<create_vol> => I<Flag to indicate where volumes has to be created>

I<Optional.> 1|0
If this is set, then volumes of the source vfiler will be created on the destination filer if it is not already existing. If set to 0 will not create 
any volume and if destination filer doesn't have all the volumes as that in source vfiler then test will die.Defaults to 1


B<method> => I<Method to use>

I<Optional.> 'nocopy'
'nocopy' can be passed as a method.

B<timeout> => I<timeout value for the method>

I<Optional.> 
Timeout value for the method. Defaults to 240.

=item Return Value

1 on success, 0 otherwise

=item Examples

=back

     $VFiler_NOTE1= NOTE::Vfiler->create(
        Rootvol   => $Volume1,
        name      => 'vfiler1',
        interface => 'e0a',
        ipaddr    => '10.225.167.216'
    );
    
$Filer1 = NOTE::Filer->connect( host => $FILER_DST);
$VFiler_NOTE1->migrate_start(filer=>$Filer1,if_to_use=>'e0a',  );

=cut 

sub migrate_start {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics


    my $self = shift;
    my %args = @_;

    #Validating the hash
    my %args_hash = validate(
        @_,
        {   filer     => { type => OBJECT, optional => 0 },
            if_to_use => { type => SCALAR, optional => 0 },
            ipaddr    => {
                type     => SCALAR,
                optional => 1,
                regex    => $iparegex,
                default  => $self->ipa,
            },
            netmask => {
                type     => SCALAR,
                optional => 1,
                regex    => $iparegex,
                default  => $self->{_netmask},
            },
            remote_login => {
                type     => SCALAR,
                optional => 1,
                default  => $args{'filer'}->conn->{'username'}
            },
            remote_passwd => {
                type     => SCALAR,
                optional => 1,
                default  => $args{'filer'}->conn->{'password'}
            },
            create_vol => {
                type     => SCALAR,
                optional => 1,
                default  => '1',
                regex    => qr/0|1/
            },
            method  => { type => SCALAR,  optional => 1, default => '' },
            timeout => { type => SCALAR,, optional => 1, default => '240' },
            use_secure_channel => {
                type     => SCALAR,
                optional => 1,
                default  => 'false',
                regex    => qr/(0|1)/
            },

        }
    );
    my $fc         = $args_hash{'filer'}->conn;
    my $src_vfiler = $self->name;
    my $src_filer  = $self->basefiler;
    my $dst_filer  = $args_hash{'filer'};
    my $create     = $args_hash{'create_vol'};

#deleting the keys that will not be required when calling the vfiler_migrate_start subroutine
    delete( $args_hash{'create_vol'} );
    my ( $vol, $vol_object, $size, $vol_dst_object );

    my @src_volumes = $self->vfiler_volumes;
    my @dst_volumes = $dst_filer->volumes;

#If create variable is set then this method will try to create the volumes with the same name as that are present inside the vfiler.
    if ($create) {
        foreach $vol (@src_volumes) {

            #Get an object handle for the existing volumes inside a Vfiler.
            $vol_object = NOTE::Volume->attach(
                Filer => $self->basefiler,
                name  => $vol,
            );
            $size = $vol_object->size;

#Check if a volume with a same name is present on the filer where vfiler is to be moved.
            if ( grep { $_ eq $vol } @dst_volumes ) {
                $vol_dst_object = NOTE::Volume->attach(
                    Filer => $dst_filer,
                    name  => $vol,
                );

                if ( $vol_dst_object->size < $size ) {

#If a volume with the same name is present on the new filer but is of lesser size then die is called
                    die
                        "A volume with the same name exists on the dst filer but is of smaller size";
                } ## end if ( $vol_dst_object->size...
            } ## end if ( grep { $_ eq $vol...

            #Volumes with the same name will be created on the 2ndfiler.
            else {
                NOTE::Volume->create(
                    Filer => $dst_filer,
                    name  => $vol,
                    size  => $size,
                    type  => 'flex'
                );
            } ## end else [ if ( grep { $_ eq $vol...
        } ## end foreach $vol (@src_volumes)
        @dst_volumes = $dst_filer->volumes;
    } ## end if ($create)

    #Checking one more time if all the vols are present .
    foreach $vol (@src_volumes) {
        unless ( grep { $_ eq $vol } @dst_volumes ) {
            die "\nSome Volumes are still not there on the 2nd Filer \n";
        }
    }

    $args_hash{'ifnames_ipaddrs_netmasks'}
        = $args_hash{'if_to_use'} . "\:"
        . $args_hash{'ipaddr'} . "\:"
        . $args_hash{'netmask'};
    delete( $args_hash{'ipaddr'} );
    delete( $args_hash{'netmask'} );
    delete( $args_hash{'filer'} );
    delete( $args_hash{'if_to_use'} );
    my $return_value;
    eval {
        $return_value
            = VFILER::Vfiler::vfiler_migrate_start( $fc, $src_vfiler,
            $src_filer, %args_hash );
    };
    die "VFiler migrate could not start because of reason : $@\n" if ($@);
    return $return_value;
} ## end sub migrate_start

=head2 ::vfiler_volumes

Returns an array with the names of volumes inside the given vfiler obect.

=over

=item Arguments

None

=item Return Value

Return an array with the name of volumes on the vfiler

=item Examples

=back

     $VFiler_NOTE1= NOTE::Vfiler->create(
        Rootvol   => $Volume1,
        name      => 'vfiler1',
        interface => 'e0a',
        ipaddr    => '10.225.167.216'
    );
my @vols=$VFiler_NOTE1->vfiler_volumes;

=cut

sub vfiler_volumes {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self          = shift;
    my $return_string = $self->run( command => "vol status" );
    my @list          = split( /\n/, $return_string );
    my ( @split_list, @volumes );
    my $i = 0;
    foreach (@list) {
        if ( $_ =~ /(online|offline|restricted)/ ) {
            @split_list = split( / (online|offline|restricted) /, $_ );
            $volumes[$i] = $split_list[0];
            $volumes[$i] =~ s/\s//g;
            $i++;
        } ## end if ( $_ =~ /(online|offline|restricted)/)
    } ## end foreach (@list)
    return @volumes;
} ## end sub vfiler_volumes

=head2 ::migrate_status

Runs migrate_status on all the volumes, dies if volume is not snapmirrored.

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

=item Return Value

None

=item Examples

=back

     $VFiler_NOTE1= NOTE::Vfiler->create(
        Rootvol   => $Volume1,
        name      => 'vfiler1',
        interface => 'e0a',
        ipaddr    => '10.225.167.216'
    );
   $VFiler_NOTE1->migrate_status(filer=>$Filer1);

=cut

sub migrate_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   filer => { type => OBJECT, optional => 0 },

        }
    );
    my $fc         = $args{'filer'}->conn;
    my $src_vfiler = $self->name;
    my $src_filer  = $self->basefiler;
    my @status;
    my $i       = 0;
    my @volumes = $self->vfiler_volumes;
    foreach (@volumes) {
        eval {
            VFILER::Vfiler::vfiler_migrate_status( $fc, $src_vfiler,
                $src_filer, $_ );
        };
        die "\n migrate status for volume $_ failed with reason :$@ \n"
            if ($@);
        $i++;
    } ## end foreach (@volumes)
} ## end sub migrate_status

=head2 ::migrate_cancel

Cancels the vfiler migrate operation

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

=item Return Value

None

=item Examples

=back

   $VFiler_NOTE1->migrate_cancel(filer=>$Filer1);

=cut

sub migrate_cancel {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   filer => { type => OBJECT, optional => 0 },

        }
    );
    my $fc         = $args{'filer'}->conn;
    my $src_vfiler = $self->name;
    my $src_filer  = $self->basefiler;
    eval {
        VFILER::Vfiler::vfiler_migrate_cancel( $fc, $src_vfiler, $src_filer );
    };
    die "\n Migrate cancel failed with reason :$@ \n" if ($@);
} ## end sub migrate_cancel

=head2 ::migrate_complete

Completes the data movement

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

B<transparent> => I<use transparent mode >

I<Optional.>default is false.

=item Return Value

None

=item Examples

=back

     $VFiler_NOTE1= NOTE::Vfiler->create(
        Rootvol   => $Volume1,
        name      => 'vfiler1',
        interface => 'e0a',
        ipaddr    => '10.225.167.216'
    );
   $VFiler_NOTE1->migrate_complete(filer=>$Filer1);

=cut

sub migrate_complete {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   filer => { type => OBJECT, optional => 0 },
	    transparent => { type => SCALAR, optional => 1 },
	    use_secure_channel => { type => SCALAR, optional => 1 },

        }
    );
    my $fc         = $args{'filer'}->conn;
    my $src_vfiler = $self->name;
    my $src_filer  = $self->basefiler;
    my $transparent= $args{transparent} || 'false';
    my $coption="";
    my $use_secure_channel=$args{use_secure_channel} || 'false';
    if ($transparent eq 'true') {
        if ($use_secure_channel eq 'true') {
                $coption='-c secure';
        }
	eval {
	my (undef, undef, $after) = Tharn::say($fc,
                        "vfiler migrate complete -t $coption $src_vfiler\@$src_filer",
                        -swallow);
	sleep(4);

        die "Vfiler migration can't be completed\n"
                        unless ($after =~ /Vfiler\s+$src_vfiler\s+migrated/gm);

        return 1;
        }
    } else {
    eval {
        VFILER::Vfiler::vfiler_migrate_complete( $fc, $src_vfiler,
            $src_filer );
    };
    die "\n Migrate complete failed with reason :$@ \n" if ($@);
    }
    sleep 60;
    return 1;
} ## end sub migrate_complete

=head2 ::convert_relationships

Converts all the relationship create while data movement to semi-sync, and wait till all relationships are in-sync.

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

=item Return Value

None

=item Examples

=back

   $VFiler_NOTE1->convert_relationships(filer=>$Filer1);

=cut

sub convert_relationships {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   filer => { type => OBJECT, optional => 0 },

        }
    );
    my @volumes = $self->vfiler_volumes;
    die "Filer object not passed" unless ( $args{'filer'} );
    my ( $reg_key, $sm );
    my $fc         = $args{'filer'}->conn;
    my $filer_host = $args{'filer'}->host;
    my $base_filer = $self->basefiler;
    foreach (@volumes) {
        $reg_key = "options.files.snapmirror.conf." . $filer_host . "\:" . $_;
        my $mode = registryGet( $fc, "$reg_key.sched" ) or die $@;
        if ( $mode !~ 'semi-sync' ) {
            registrySet( $fc, "$reg_key.sched", 'semi-sync' ) or die $@;
        }
    } ## end foreach (@volumes)
    $args{'filer'}->snapmirror_enable;
    foreach (@volumes) {
        $sm = NOTE::SnapMirror->attach(
            srcpath => "$base_filer:$_",
            dstpath => "$filer_host:$_",
        );
        $sm->wait_sync( cp_sync => 1 );
    } ## end foreach (@volumes)

} ## end sub convert_relationships

=head2 ::vis_snapshot_disable

Calls vis_snapshot disable on all the volumes inside the vfiler

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

=item Return Value

None

=item Examples


=back

   $VFiler_NOTE1->vis_snapshot_disable(filer=>$Filer1);

=cut

sub vis_snapshot_disable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   filer => { type => OBJECT, optional => 0 },

        }
    );

    my @volumes = $self->vfiler_volumes;
    die "Filer object not passed" unless ( $args{'filer'} );
    foreach (@volumes) {
        my $return = $args{'filer'}
            ->say("priv set test;rsm vis_snapshot disable $_");
        die "vis_snapshot disable failed with reason :$return " if ($return);
    }
} ## end sub vis_snapshot_disable

=head2 ::vis_snapshot_enable

Calls vis_snapshot enable on all the volumes inside the vfiler

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

=item Return Value

None

=item Examples

=back

   $VFiler_NOTE1->vis_snapshot_enable(filer=>$Filer1);

=cut

sub vis_snapshot_enable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   filer => { type => OBJECT, optional => 0 },

        }
    );
    my @volumes = $self->vfiler_volumes;
    foreach (@volumes) {
        my $return
            = $args{'filer'}->say("priv set test;rsm vis_snapshot enable $_");
        die "vis_snapshot enable failed with reason :$return " if ($return);
    }
} ## end sub vis_snapshot_enable

=head2 ::vol_ssm_offline

Calls vol ssm offline  on the mentioned vols or on all the volumes inside the vfiler if volume name is not passed

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

B<vol> => I<vol name>

I<Optional.>
Volume name which has to be offlined


=item Return Value

None

=item Examples

=back

   $VFiler_NOTE1->vol_ssm_offline(filer=>$Filer1,vol=>$volume1);
   $VFiler_NOTE1->vol_ssm_offline(filer=>$Filer1);

=cut

sub vol_ssm_offline {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    my ( $vol, @volumes );
    if ( $args{'vol'} ) {
        $vol = $args{'vol'}->name if ( $args{'vol'} );
    } else {
        @volumes = $self->vfiler_volumes;
        $vol = join( "\,", @volumes );
    }
    my $return_string
        = $args{'filer'}->say("priv set test;vol ssm offline $vol");
    if (@volumes) {
        foreach (@volumes) {
            if ( ( $return_string !~ /Volume $_: offline successful/ ) ) {
                die "\n Could not offline volume $_ :$return_string \n";
            }
        }
    } else {
        if ( ( $return_string !~ /Volume $vol: offline successful/ ) ) {
            die "\n Could not offline volume $_ :$return_string \n";
        }
    }
    NOTE::logcomment("Volumes(s) succesfully offlined");
    return 1;
} ## end sub vol_ssm_offline

=head2 ::vol_ssm_online

Calls vol ssm online  on the mentioned vols or on all the volumes inside the vfiler if volume name is not passed

=over

=item Arguments

B<filer> => I<Filer Object >

I<Required.>
Filer object where vfiler has to be moved.

B<vol> => I<vol name>

I<Optional.>
Volume name which has to be onlined


=item Return Value

None

=item Examples

=back

   $VFiler_NOTE1->vol_ssm_online(filer=>$Filer1,vol=>$volume1);
   $VFiler_NOTE1->vol_ssm_online(filer=>$Filer1);

=cut

sub vol_ssm_online {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    my ( $vol, @volumes );
    if ( $args{'vol'} ) {
        $vol = $args{'vol'}->name if ( $args{'vol'} );
    } else {
        @volumes = $self->vfiler_volumes;
    }
    $vol = join( "\,", @volumes );
    my $return_string
        = $args{'filer'}->say("priv set test;vol ssm online $vol");
    if (@volumes) {
        foreach (@volumes) {
            if ( ( $return_string !~ /Volume $_: online successful/ ) ) {
                die "\n Could not online volume $_ :$return_string \n";
            }
        }
    } else {
        if ( ( $return_string !~ /Volume $vol: online successful/ ) ) {
            die "\n Could not online volume $_ :$return_string \n";
        }
    }
    NOTE::logcomment("Volumes(s) succesfully onlined");
    return 1;
} ## end sub vol_ssm_online

# Internal function to ping a given IPA from filer
# Usage:
#       _ping($filer,'10.72.72.100');
# return Value:
#              1 if IP is pingable, 0 otherwise

sub _ping {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $filerobj, $ip ) = @_;
    my $result = $filerobj->say( "ping " . $ip );
    unless ( $result =~ /alive/ ) {
        NOTE::logcomment( "Unable to ping " . $ip );
        return 0;
    }
    return 1;
} ## end sub _ping

# Internal function to create a given user on filer
# Usage:
#       _create_vfiler_user($filer,$vfiler,'testuser','password');
# return Value:
#              1 on success, 0 otherwise

sub _create_vfiler_user {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $basefiler, $vf, $user, $pwd ) = @_;

    $basefiler->say( "vfiler run "
            . $vf->{_name}
            . " useradmin user add "
            . $user . " -p "
            . $pwd
            . " -g Administrators" );

    (   $basefiler->say(
            "vfiler run " . $vf->{_name} . " useradmin user list " . $user
            ) =~ /User does not exist/
    ) ? return 0 : return 1;
} ## end sub _create_vfiler_user

# Internal function to return full Vfiler status using -a option
# Usage:
#       _status_a($filer, $vfilername);
# return Value:
#      Hash with all values,  hash keys: status, ipaddr, ipspace & path
#      otherwise status like Vfiler not found/not configured etc in case of errors

sub _status_a {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $filer, $vfname ) = @_;
    my %hash;
    my %retHash = (
        status  => '',
        ipaddr  => '',
        if      => '',
        ipspace => '',
        path    => ''
    );
    eval { %hash = VFILER::Setup::vfiler_status( $filer, $vfname ); };

    if ( $@ =~ /Vfiler not found/ ) {
        $retHash{status} = "Vfiler not found";
    } elsif ( ( $@ =~ m/Unexpected line/ )
        || ( $@ =~ m/Insufficient output/ ) )
    {
        $retHash{status} = "Vfiler not Configured properly";
    } else {
        $retHash{status}  = $hash{$vfname}{status};
        $retHash{if}      = $hash{$vfname}{ip0};
        $retHash{ipaddr}  = $hash{$vfname}{ipaddr0};
        $retHash{ipspace} = $hash{$vfname}{ipspace};
        $retHash{path}    = $hash{$vfname}{path0};
    } ## end else [ if ( $@ =~ /Vfiler not found/)
    return %retHash;
} ## end sub _status_a

# Internal function to return netmask being used by Vfiler
# Usage:
#       _getnetmask($filer, $vfilerIP, 'e0a');
# return Value:
#       Value of Netmask being used

sub _getnetmask {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $flr, $vfip, $intrfc ) = @_;

    my $filer = NOTE::Filer->connect( host => $flr );
    my $result = Network::Utils::get_ifconfig_info( $filer->conn, $intrfc );

    return Socket::inet_ntoa(
        pack( "N",
            hex( $result->{$intrfc}->{'alias'}->{$vfip}->{'netmask'} ) )
    );
} ## end sub _getnetmask

# Internal function to return netmask being used by Vfiler
# Usage:
#       _getgateway($filer);
#               where $filer is Filer NOTE object
# return Value:
#       Value of gateway set on the filer

sub _getgateway {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($flr) = @_;
    my $routeinfo = $flr->say("route -ns");

    if ( $routeinfo
        =~ m/(default\s+)((\d\d?\d?)\.(\d\d?\d?)\.(\d\d?\d?)\.(\d\d?\d?))/ )
    {
        return $2;
    } else {
        NOTE::logcomment(
            "Failed to obtain route table information through 'route -ns' command"
        );
        return undef;
    } ## end else [ if ( $routeinfo =~ ...
} ## end sub _getgateway

=head2 ::snapmirror_enable

Enables snapmirror on the vFiler.

=over

=item Arguments

None.

=item Examples

    $vfiler->snapmirror_enable;

=back

=cut

sub snapmirror_enable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    #Set option snapmirror.enable to on
    $self->setoption( option => 'snapmirror.enable', value => 'on' );

    my $text = $self->getoption( option => 'snapmirror.enable' );
    die "Enabling snapmirror failed\n" if ( $text !~ 'on' );
    return 1;
} ## end sub snapmirror_enable

=head2 ::snapmirror_disable

Disables snapmirror on the vFiler.

=over

=item Arguments

None.

=item Examples

    $vfiler->snapmirror_disable;

=back

=cut

sub snapmirror_disable {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    #Set option snapmirror.enable to off
    $self->setoption( option => 'snapmirror.enable', value => 'off' );

    my $text = $self->getoption( option => 'snapmirror.enable' );
    die "Disabling snapmirror failed\n" if ( $text !~ 'off' );
    return 1;
} ## end sub snapmirror_disable

=head2 ::vfiler_host_entry

Writes supplied hostname & ip into /etc/hosts file of the vFiler

=over

=item Arguments

B<host> => I<String>
Name of the host

B<host> => I<integer_in_ip_format>
IP of the host

=item Examples

    $vfiler->vfiler_host_entry(host => "Vfiler1", ip => "10.72.230.10");
    $vfiler->vfiler_host_entry(host => $vfiler1->name, ip => $vfiler1->ipa);

=back

=cut

sub vfiler_host_entry {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;

    my ( $hostname, $hostip, $pfiler, %vfiler_info, $root_vol, $hostfile,
        $host_entry );

    unless ( exists $args{host} && exists $args{ip} ) {
        die "Insufficient no of arguements\n";
    }

    $hostname = $args{host};
    $hostip   = $args{ip};

    $pfiler = NOTE::Filer->connect( host => $self->basefiler );

    #Writes into /vol/<root_vol_of_vFiler>/etc/hosts
    %vfiler_info = _status_a( $pfiler->conn, $self->{_name} );
    $root_vol    = $vfiler_info{path};
    $hostfile    = $root_vol . "/etc/hosts";
    $host_entry  = "$hostip" . " $hostname";
    my $cmd = "wrfile -a $hostfile $host_entry";
    $pfiler->say( "priv set -q test;%s", $cmd );

    my $text = $pfiler->say( "priv set -q test;rdfile %s", $hostfile );
    unless ( $text =~ /$host_entry/ ) {
        die "Writing to hostfile failed";
    }

} ## end sub vfiler_host_entry

=head2 ::vfiler_host_access

Sets access permission on vFiler

=over

=item Arguments

B<access_to> => I<String>
Access permission to be set on vFiler. Applicable values "all","none"&hostname. Defaults to "all".

=item Examples

    $vfiler->vfiler_host_access(access_to => $vfiler1->name);
    $vfiler->vfiler_host_access(access_to => "all");
    $vfiler->vfiler_host_access(access_to => "none");

=back

=cut

sub vfiler_host_access {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;

    unless ( exists $args{access_to} ) {
        die "Insufficient no of arguements\n";
    }

    #Set option snapmirror.access on vFiler
    $args{access_to} = "host=" . $args{access_to};
    $self->setoption(
        option => 'snapmirror.access',
        value  => $args{access_to}
    );

    my $text = $self->getoption( option => 'snapmirror.access' );
    die "Unable set access permission\n" if ( $text !~ $args{access_to} );
    return 1;
} ## end sub vfiler_host_access

=head2 ::host

To fetch the Vfiler hostname

=over

=item Arguments

None

=item Return Value

String containing the name of Vfiler represented by the object.

=item Examples

=back

        $vfiler_object->host ;

=cut

sub host {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return $self->name;
}

=head2 ::defaultip

To fetch the default IP Address of the Vfiler

=over

=item Arguments

None

=item Return Value

String containing the IP Address of Vfiler represented by the object.

=item Examples

=back

        $vfiler_object->defaultip ;

=cut

sub defaultip {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    return $self->ipa;
}

=head2 ::configure_ipv6

To configure IPv6 Address on the Vfiler

=over

=item Arguments

B<v6_addresss> => I<String>
I<Required.>
IPv6 address to be assigned to Vfiler

B<interface> => I<String>
I<Required.>
Interface on which Vfiler IPv6 address is to be configured

B<prefixlen> => I<String>
I<Optional.>
Prefixlen for IPv6 address to be assigned to Vfiler.
Default value is 64.

B<remove_v4> => I<String>
I<Optional.>
if IPv4 address is to be removed after successful configuration of IPv6 address
Default value is 0, which means IPv4 address will be retained.

=item Return Value

None, croaks on failure.

=item Examples

=back

        $vfiler_object->configure_ipv6(v6_address => 'fd20:8b1e:b255:8301:2a0:98ff:fe27:fa',
                                        interface => 'e0a',
                                         prefixlen => 64
                                       ) ;

=cut

sub configure_ipv6 {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    my %args = validate(
        @_,
        {   v6_address => { type => SCALAR, optional => 0 },
            interface  => { type => SCALAR, optional => 0 },
            prefixlen  => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^\d+$/,
                default  => 64
            },
            remove_v4 => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/^[01]$/,
                default  => 0
            },
        }
    );

    #check basefiler v6 options before attempting to enable IPv6 for Vfiler
    my $pfiler = NOTE::Filer->connect( host => $self->basefiler );
    RegistryCmds::setOption( $pfiler->conn, 'ip.v6.ra_enable', 'on' );
    RegistryCmds::setOption( $pfiler->conn, 'ip.v6.enable',    'on' );

    $self->add( ipaddr => $args{v6_address} );
    my $if_alias
        = "ifconfig "
        . $args{interface}
        . " alias "
        . $args{v6_address}
        . " prefixlen "
        . $args{prefixlen};
    $pfiler->say($if_alias);

    if ( $args{remove_v4} == 1 ) {
        $self->remove( ipaddr => $self->ipa );
    }
} ## end sub configure_ipv6

=head2 ::v6_address

To fetch the IPv6 Address of the Vfiler

=over

=item Arguments

None

=item Return Value

String containing the IPv6 Address of Vfiler

=item Examples

=back

        $vfiler_object->v6_address ;

=cut

sub v6_address {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return shift->{_v6_address};
}

=head2 ::sio

To run sio traffic over NFS on the specified Qtree or Volume

=over

=item Arguments

B<container> => I<Volume or Qtree object>

I<Required.>
Volume or Qtree objected that is attached to the path which is to be filled with data. Object
created through constructors create or attach of classes NOTE::Volume or NOTE::Qtree respectively.

B<runtime> => I<String>

I<Required.>
Time in seconds for which hammer has to be run.

=item Examples

=back

   $vfiler_obect->sio( container =>  $obj, runtime => 100 );

=cut

sub sio {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my %args = validate(
        @_,
        {   container => { type => OBJECT, optional => 0 },
            runtime => { type => SCALAR, optional => 0, regex => qr/^\d+$/ },
            client  => { type => SCALAR, optional => 0 }
        }
    );

    my $timeout = $args{runtime} + 120;
    my $mntpnt = $self->mount( container => $args{'container'} );

    NOTE::logcomment("Starting NFS sio traffic");
    my $cmdsio
        = "sudo /usr/software/test/bin/sio 50 50 100K 0 10m "
        . $args{runtime} . " 1 "
        . $mntpnt
        . "/sio.tmp -verify -niceoutput";

    my $host    = Tharn::host( $args{client} );
    my $lnxclnt = Tharn::connect('#rsh');
    Tharn::say( $lnxclnt,
              "sudo /bin/dd if=/dev/urandom of=" 
            . $mntpnt
            . "/sio.tmp bs=10M count=1" );

    my $cmdoutput = Tharn::say( $lnxclnt, $cmdsio, -timeout, $timeout );
    NOTE::logcomment($cmdoutput);

    return 1;

} ## end sub sio

# Internal function to check and append data if needed
# Usage:
#       _append_tofile($filer,$data);
#               where $filer is Filer NOTE object & $data is data to be appended
# return Value:
#       none

sub _append_tofile {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $self, %args ) = @_;
    my $overwrite = 0;

    croak("Mandatory parameters to the function not specified")
        if ( ( !defined( $args{'filepath'} ) )
        || ( !defined( $args{'data'} ) ) );

    my $existing_data = $self->say( "rdfile " . $args{'filepath'} );
    $existing_data = "" if ( $existing_data =~ /No such file or Directory/ );

    if ( $existing_data =~ m/$args{'data'}/i ) {
        NOTE::logcomment( "Data \""
                . $args{'data'}
                . "\" already present in the file \""
                . $args{'filepath'}
                . "\", no updates needed !" );
        return 1;
    } else {

        #data does not exist we need to write in the file
        $self->say( "wrfile -a " . $args{'filepath'} . " " . $args{'data'} );
    }
} ## end sub _append_tofile

# Internal function to fetch the IPv6 address from vfiler status -r command
# Usage:
#   _getv6address($filer,$vfilername);
#     where $filer is Filer NOTE object & $vfilername is the name of vfiler unit
# return Value:
#       IPv6 address of the Vfiler

sub _getv6address {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $pfiler, $vfilername ) = @_;
    my $output = $pfiler->say("vfiler status -r $vfilername");

    if ($output =~ /(?mx)(
                  (IP\s+address:\s+)
                  ( ( ([a-fA-F0-9]){1,4}(:)? )+ )
                  ( \s+ \[)
                  )/
        )
    {
        return $3;
    } ## end if ( $output =~ /(?mx)( )
    return undef;
} ## end sub _getv6address

# Internal function to fetch the IPv6 address from vfiler status -r command
# Usage:
#   _getprefixlen($filer,$ipv6_addr);
#     where $filer is Filer NOTE object & $ipv6_addr is the v6 address whose
#           prefixlen is to determined
# return Value:
#       prefixlen of IPv6 address of the Vfiler

sub _getprefixlen {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ( $pfiler, $v6addr ) = @_;
    my $output = $pfiler->say("ifconfig -a");

    if ( $output =~ /(?mx)((inet6\s+ $v6addr\s+ prefixlen\s+)(\d+))/ ) {
        return $3;
    }
    return undef;
} ## end sub _getprefixlen

=head1 SEE ALSO

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm>

=head1 STATUS

Under Review.

$Id: //depot/prod/test/main/lib/NOTE/Vfiler.pm#13 $

$Date: 2010/08/26 $

=cut

1;
