#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Vfiler/SnapVault.pm#1 $
#
## @summary  NOTE Vfiler/SnapVault class
## @status   Public
## @pod      here
## @author   Akshatha.G@netapp.com Sumukh.Bhardwaj@netapp.com

=head1 NAME

NOTE::Vfiler::SnapVault

=head1 DESCRIPTION

This is the NOTE class for working with Vfiler SnapVault relationships.

=head1 METHODS

=cut

package NOTE::Vfiler::SnapVault;

use strict;
use warnings;
use Tharn;
use NOTE;
use Data::Dumper;
use Carp qw(croak);
use POSIX qw();
use SnapVault qw();
use FSR::FSRutils qw(subtestResults);
use FSR::rpl_cmds qw(:all);
use filer qw(filer_file_read );
use CLI::FileIO qw(wrfile);
use SnapCmds qw(snapshot_delete snapshot_status snapshot_exists snapshot_list);
use Wreck qw(wreck);
use NOTE::DataIntegrity;

# Expose the following registry keys on the object
my $regSlice = {
    'snapvault.access' => sub {
        'options.snapvault.access';
    },
    'snapvault.lockvault_log_volume' => sub {
        'options.snapvault.lockvault_log_volume';
    },
    'snapvault.nbu.trace_category' => sub {
        'options.snapvault.nbu.trace_category';
    },
    'snapvault.nbu.trace_level' => sub {
        'options.snapvault.nbu.trace_level';
    },
    'snapvault.nbu.verify_backup' => sub {
        'options.snapvault.nbu.verify_backup';
    },
    'snapvault.nbu.verify_duplicate' => sub {
        'options.snapvault.nbu.verify_duplicate';
    },
    'snapvault.offline_reboot.enable' => sub {
        'options.snapvault.offline_reboot.enable';
    },
    'snapvault.repair_bug137685' => sub {
        'options.snapvault.repair_bug137685';
    },
    'snapvault.repair_bug137685_stats' => sub {
        'options.snapvault.repair_bug137685_stats';
    },
    'snapvault.suppress_bli_warnings' => sub {
        'options.snapvault.suppress_bli_warnings';
    },
    'snapvault.conf.qtree' => sub {
        'options.snapvaultconf.qtree.' . shift->dst_name;
    },};

=head2 B<Relationship Creation>

=head2 ::start

Creates a SnapVault relationship between vfiler and pfiler and returns an object
representing the relationship.

=over

=item Arguments

B<src> => I<Qtree object>

I<Dependent.>
A Qtree object.

B<dst> => I<Qtree object>

I<Dependent.>
A Qtree object.

B<src_host> => I<Note Filer or vfiler object>

I<Dependent.>
The Filer/Vfiler host object from which the relation needs to be established.

B<dst_host> => I<Note Filer or vfiler object>

I<Dependent.>
The Filer/Vfiler host object from which the relation needs to be established.

All the four arguments are required when C<::start> is called as a class method
and neither may be used when called via an object.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<tries_count> => I<number>

I<Optional.>
Snapvault tries count for this relationship.

B<background> => C<0> | C<1>

I<Optional.>
Run the initial transfer in the background. The default is 0.

B<parallel> => C<0> | C<1>

I<Optional.>
Start the transfer and return. The background option is ignored
and the licenses are not checked to speed up start of transfer.
Use only when starting a lot of transfers concurrently. Default is 0.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Examples

=back

  $sv = NOTE::Vfiler::SnapVault->start(
         src      => $src_qtree,
         dst      => $dst_qtree,
         src_host => $vfiler_src,
         dst_host => $vfiler_dst,
     );

  $sv = NOTE::Vfiler::SnapVault->start(
         src      => $src_qtree,
         dst      => $dst_qtree,
         src_host => $pfiler_src,
         dst_host => $vfiler_dst,
     );

  $sv = NOTE::Vfiler::SnapVault->start(
         src        => $src_qtree,
         dst        => $dst_qtree,
         src_host   => $vfiler_src,
         dst_host   => $pfiler_dst,
         throttle   => 100,
         background => 1
     );

=cut

sub start {

    my ($class, %args) = @_;
    my $self = {};
    my (%sv_args, $ret, $op, %vf_args);

    # Most likely an object. note that 'ref $class' is not compared
    # to NOTE::Vfiler::SnapVault because this method can be inherited by a
    # subclass and 'ref $class' will not reflect 'NOTE::Vfiler::SnapVault'
    # in the subclass.
    if (ref $class ne q{}) {
        $self = $class;

        if (   exists $args{src}
            || exists $args{dst}
            || exists $args{src_host}
            || exists $args{dst_host}) {
            croak 'Cannot override existing src or dst attributes '
              . 'in existing object.';
        }
    } else {

        # Instantiate this object
        $self = NOTE::Vfiler::SnapVault->new(%args);
    }

    NOTE::reference($self);

    # These are flags that are used to indicate whether the source and
    # destination are vfilers or not.
    $self->{_vfiler_src} = (ref $self->{_src_host} eq 'NOTE::Vfiler') ? 1 : 0;
    $self->{_vfiler_dst} = (ref $self->{_dst_host} eq 'NOTE::Vfiler') ? 1 : 0;
=head
    # This check has been removed since we might want to use this library 
    # for relations of the type pfiler->pfiler where the hosts are pfilers 
    # but volumes belong to Vfiler.
    #
    # Beware using this library in other cases!!
    # Please use  NOTE::SnapVault.pm wherever possible.
    #
    # You are allowed to use this library only if either of source or
    # destinationNOTE::Vfiler::SnapVaultlse use NOTE::SnapVault.pm
    if (!($self->{_vfiler_src} || $self->{_vfiler_dst})) {
        croak 'Either of the source or dest host has to be a vfiler';
    }
=cut
    # We will not accept qtree path as a valid 'src' or 'dst' parameter.
    # This is because the qtree path has to be relevant wrt the source/dest
    # hosts. Hence checking if they are NOTE::Qtree objects.
    if (   ref $self->{_src} ne 'NOTE::Qtree'
        || ref $self->{_dst} ne 'NOTE::Qtree') {
        croak 'Both the source and destination have to be NOTE::Qtree types';
    }

    delete $args{src};
    delete $args{dst};
    delete $args{src_host};
    delete $args{dst_host};

    # Set flags for normal transfer
    $self->{_rxfer}    = 0;
    $self->{_released} = 0;
    $self->{_stopped}  = 0;

    $op = 'Snapvault Start' . $self->description;
    NOTE::logcomment($op);

    # Process background option
    if (exists $args{background}) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Override default timeout
    $self->timeout(%args);
    $args{timeout} = $self->{_timeout};

    if (!$args{parallel}) {

        # Enable SnapVault on both filers: license, open access and enable
        $self->_license_and_enable;
    }

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;
    if (exists $args{connection_mode}) {
        $self->{connection_mode} = $args{connection_mode};
        $sv_args{connection_mode} = $args{connection_mode};
    }

    elsif (defined NOTE::option(q/$CONN_MODE/)) {
        $sv_args{connection_mode} = NOTE::option(q/$CONN_MODE/);
        $self->{connection_mode} = $args{connection_mode};
    }

    # If the destination host is a vfiler object, Set the vfiler flag in
    # SnapVault.pm. Also set all other vfiler arguments here.
    $self->_set_vfiler_context($self->{_dst_host});

    # Start snapvault relationship
    $ret = SnapVault::sv_start($self->{_dst_host}->conn,
                               $self->dst_path,
                               SOURCE => $self->src_path,
                               %sv_args);

    # Check for errors
    croak $@ if $@;

    if (ref $ret eq 'Subtest') {    # Save, if it's an object
        push @{$self->{_subtests}}, {operation => $op, subtest => $ret};
    }

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    return $self;
} ## end sub start

=head2 ::new

Creates an object representing a SnapVault relationship,but does B<not> cause
the creation of a SnapVault relationship on the filer.
The L<C<::start>|/::start> method of the returned object may be used to create
the relationship.

This method should not normally be used;
use L<C<::start>|/::start> directly instead.

=over

=item Arguments

B<src_host> => I<Filer object> | I<Vfiler object>

I<Required>
A Note Filer/Vfiler object of the source on which the snapvault relation is to
be hosted.

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Required>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
to be hosted.

B<src> => I<Qtree object>

I<Required.>
A Qtree object.

B<dst> => I<Qtree object>

I<Required.>
A Qtree object.

B<preprocess> => C<0> | C<1>

I<Optional.>
If true, verifies licenses on both filers/vfilers.
Use only in conjunction with parallel option on L<C<::start>|/::start>
to create a lot of snapvault objects and start all transfers faster.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Example

=back

     $sv = NOTE::Vfiler::SnapVault->new(
         src_host => $vfiler_src,
         dst_host => $vfiler_dst,
         src      => $src_qtree,
         dst      => $dst_qtree,
     );

     $sv = NOTE::Vfiler::SnapVault->new(
         src_host => $pfiler_src,
         dst_host => $vfiler_dst,
         src      => $src_qtree,
         dst      => $dst_qtree,
     );

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    # Check arguments
    croak 'Src or dst host argument is missing'
      if (!exists $args{src_host} || !exists $args{dst_host});

    croak 'Src or dst argument is missing'
      if (!exists $args{src} || !exists $args{dst});

    # Set private variables for object
    $self->{_owner}    = $$;
    $self->{_timeout}  = exists $args{timeout} ? $args{timeout} : 600;
    $self->{_src}      = NOTE::SnapVault::_qtree_path_attach($args{src});
    $self->{_dst}      = NOTE::SnapVault::_qtree_path_attach($args{dst});
    $self->{_src_host} = $args{src_host};
    $self->{_dst_host} = $args{dst_host};
    $self->{_Filer} = $self->{_dst_host};    # needed by NOTE::Registry.pm

    # Bless this object
    bless $self, $class;

    # Make this object a tie variable so registry options can be accessed
    # via a hash element of the object
    tie %{$self}, 'NOTE::Registry::Slice', $self, $regSlice;

    # Add the default schedule to cleanup list
    push @{$self->{_schedules}}, {};
    $self->{_schedules}[-1]{q{:} . $self->dst->hostpath} = {
                                                   volume => $self->dst->Volume,
                                                   source => 0,};

    if ($args{preprocess}) {

        # Enable SnapVault on both filers: license, open access and enable
        $self->NOTE::Vfiler::SnapVault::_license_and_enable;
    }

    return $self;
} ## end sub new

=head2 B<SnapVault Commands>

The methods names in this section correspond to
ONTAP's C<snapvault> subcommands.
Only administrator privilege level subcommands are listed in this section.
These methods eventually call the self-named subcommands.
Only the briefest mention of what these subcommands do is given here.
See the SnapVault documentation for more information on them.

SnapVault commands that apply to the filer as a whole are part of
L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>
and are described in the
L<C<NOTE::Filer::SnapVault>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/snapvault>
section.

=head2 ::abort

Aborts an active SnapVault transfer on the destination.
If this function is called after L<C<::restore>|/::restore>, it aborts the
restore transfer on the source.
Returns 1 on success.

=over

=item Arguments

B<hard> => C<0> | C<1>

I<Optional.>
If set to 1, does a hard abort, such that the transfer is not restartable.
The default is 0.

B<src_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the source host from which the relation is
aborted.

B<dst_host> => I<Filer object> | I<Vfiler object>

A Note Filer/Vfiler object of the destination host from which the relation is
aborted.

B<source> => C<0> | C<1>

I<Optional.>
Abort the relationship from the source host instead of destination host.
The default is 0 (destination host).
The argument, if specified, will be the arbiter as to whether source or destination
host is to be used, even if src_host or dst_host are specified.
If the argument is not specified, the src_host or dst_host will be used (accordingly
if only one of them is specified); or only dst_host will be used (if both are specified
or neither is specified)

=item Examples

=back

    # aborts the snapvault transfer by running ZAPI/CLI command on the host vfiler_new
    $sv->abort(dst_host => vfiler_new);

    $sv->abort(
        hard => 1
    );

    $sv->abort( source => 1 );

=cut

sub abort {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret);
    my $qtree    = $self->{_dst};
    my $source   = delete $args{source};
    my $host     = ( exists $args{dst_host} )? $args{dst_host}: $self->dst_host;
    my $hostpath = $self->dst_path;

    # Assign default host to src_host based on 'source'
    if ( defined $source ? $source : 0 ) {
        $host = ( exists $args{src_host} )? $args{src_host}: $self->src_host;
    }

    if ( exists $args{src_host}
      and (ref $args{src_host} ne 'NOTE::Vfiler')
      and (ref $args{src_host} ne 'NOTE::Filer') ){
        croak 'The source host has to be a Filer/Vfiler object';
    }

    if ( exists $args{dst_host}
      and (ref $args{dst_host} ne 'NOTE::Vfiler')
      and (ref $args{dst_host} ne 'NOTE::Filer') ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    NOTE::logcomment( "Snapvault Abort" . $self->description );

    # Check if coming from a restore
    if ( $self->{_rxfer} && exists($self->{_restore}) ) {
        # Abort restore transfer on the source
        $qtree    = $self->{_restore};
        $source   = 0 unless defined($source);
        $host     = ( exists $args{dst_host} )? $args{dst_host}: $self->{_restore_host};
        $hostpath = $self->restore_path;

        if ( defined $source ? $source : 0 ) {
            $host = ( exists $args{src_host} )? $args{src_host}: $self->{_dst_host};
        }

        # Save qtree for restore restart
        $self->{_restore_restart} = $self->{_restore};
        delete $self->{_restore};
    }

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($host);

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Abort transfer
    $ret = SnapVault::snapvault_abort(
      { conn => $host->conn,
        source => defined $source ? $source : 0
      },
      $hostpath,
      %sv_args
    );

    # Set back the context to filer context
    $self->_set_vfiler_context;

    # The above function may return undef with no error, in that case croak
    # Also croak if any errors exist
    croak $@ if $@;
    croak "Abort did not succeed" if ! $ret;

    return $ret;
}

=head2 ::update

Starts an incremental transfer for a SnapVault relationship.
Returns a 'subtest' object, which may be ignored if the transfer completes
before the call to C<::update> returns.

=over

=item Arguments

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which the relation needs to be updated.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the source to synchronize from.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<background> => C<0> | C<1>

I<Optional.>
Run the update transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

B<no_lun_clone_expansion> => C<0> | C<1>

I<Optional.>
If SnapVault::use_zapi_only() is set this param will make sure
<no-lun-clone-expansion> tag is sent along with
<snapvault-secondary-initiate-incremental-transfer>.
By default this value is 0.So transfer will not happen in background.

=item Examples

=back
     $sv->update(dst_host => $vfiler);

     $sv->update;

     $subtest = $sv->update(
         background => 1,
         throttle   => 100,
     );

=cut

sub update {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret, $op, $sv_dest);

    if (exists($args{dst_host})
        && (   (ref $args{dst_host} ne 'NOTE::Vfiler')
            && (ref $args{dst_host} ne 'NOTE::Filer'))
      ) {
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    # Get the right destination host.
    $sv_dest = (exists $args{dst_host}) ? $args{dst_host} : $self->{_dst_host};

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    # Set flag for normal transfer
    $self->{_rxfer} = 0;

    $op = 'Snapvault Update' . $self->description;
    NOTE::logcomment($op);

    # Process background option
    if (exists $args{background}) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if !exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Start snapvault update
    $ret = SnapVault::sv_update($sv_dest->conn, $self->dst_path, %sv_args);

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    if (ref $ret eq "Subtest") {    # Save, if it's an object
        push @{$self->{_subtests}}, {operation => $op, subtest => $ret};
    }

    return $ret;
} ## end sub update

=head2 ::stop

Stop SnapVault relationship and unconfigure the destination.
Returns a 'subtest' object, which may be ignored if the stop completes
before the call to C<::stop> returns.

=over

=item Arguments

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which the relation needs to be stopped from.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<background> => C<0> | C<1>

I<Optional.>
Run the stop command in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Example

=back

     $sv->stop(dst_host => $vfiler);

     $sv->stop;

     $subtest = $sv->stop(
         background => 1,
     );

=cut

sub stop {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret, $op, $sv_dest);

    if (exists($args{dst_host})
        && (   (ref $args{dst_host} ne 'NOTE::Vfiler')
            && (ref $args{dst_host} ne 'NOTE::Filer'))
      ) {
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    # Get the right destination host.
    $sv_dest = (exists $args{dst_host}) ? $args{dst_host} : $self->{_dst_host};

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    $op = "Snapvault Stop" . $self->description;
    NOTE::logcomment($op);

    # Process background option
    if (exists $args{background}) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if !exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Unconfigure destination
    $ret = SnapVault::sv_stop($sv_dest->conn, $self->dst_path, %sv_args);

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    # Set flag that this relationship has already been stopped
    $self->{_stopped} = 1;

    if (ref $ret eq "Subtest") {    # Save, if it's an object
        push @{$self->{_subtests}}, {operation => $op, subtest => $ret};
    }

    return $ret;
} ## end sub stop

=head2 ::status

Returns a hash reference to the status of the relationship.

=over

=item Arguments

B<src_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the source on which the snapvault relation is
hosted.

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
hosted.

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the relationship from the source host instead of destination host.
The default is 0 (destination host).
The argument, if specified, will be the arbiter as to whether source or destination
host is to be used, even if src_host or dst_host are specified.
If the argument is not specified, the src_host or dst_host will be used (accordingly
if only one of them is specified); or only dst_host will be used (if both are specified
or neither is specified)

B<restore> => C<0> | C<1>

I<Optional.>
Get the status of the restore transfer.
The default is 0.

=item Examples

=back

    $status = $sv->status;     # Get status for that particular relation


    $status = $sv->status(     # Get status from source filer vfiler_src
        src_host => vfiler_src
    );

    $status = $sv->status(     # Get status from destination filer vfiler_new
        dst_host => vfiler_new
    );

=cut

sub status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my $status;
    my $host;
    my $is_source = $args{source};
    my ( $src_host, $dst_host );

    if ( exists $args{src_host}
      and (ref $args{src_host} ne 'NOTE::Vfiler')
      and (ref $args{src_host} ne 'NOTE::Filer') ){
        croak 'The source host has to be a Filer/Vfiler object';
    }
    if ( exists $args{dst_host}
      and (ref $args{dst_host} ne 'NOTE::Vfiler')
      and (ref $args{dst_host} ne 'NOTE::Filer') ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    if( $args{restore} ) {
        $dst_host = ( exists $args{dst_host} )? $args{dst_host}: $self->{_restore_host};
        $src_host = ( exists $args{src_host} )? $args{src_host}: $self->dst_host;
    } else {
        $dst_host = ( exists $args{dst_host} )? $args{dst_host}: $self->dst_host;
        $src_host = ( exists $args{src_host} )? $args{src_host}: $self->src_host;
    }

    # Decide which host to use for getting snapvault status
    if ( exists $args{source} ) {
        # 'source' parameter is specified
        $host = $args{source} ? $src_host : $dst_host;
    } else {
        # 'source' parameter is not specified
        if ( exists $args{src_host} and not exists $args{dst_host} ) {
            $host = $src_host;
            $is_source = 1;
        } else {
            $host = $dst_host;
        }
    }

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($host);

    my $dst_path = $self->dst_path;
    my $src_path = $self->src_path;

    if ( $args{restore} ) {
        croak "Option 'restore' is only valid after calling restore"
          if ! $self->{_restore};

        $src_path = $self->dst_path;
        $dst_path = $self->restore_path;
    }

    # Get status of snapvault relationship
    $status = SnapVault::snapvault_status_detailed(
        { conn => $host->conn, source => $is_source },
        $dst_path,
        $src_path
    );

    # Set back the context to filer context
    $self->_set_vfiler_context;

    # The above function may return undef with no error, in that case croak
    # Also croak if any errors exist
    croak $@ if $@;
    croak "Could not return status" if ! $status;

    return $status;
}

=head2 ::configuration

Returns a hash reference to the configuration of the relationship.

=over

=item Arguments

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
hosted.

=item Examples

=back

    $conf = $sv->configuration;  # Get configuration for that particular relation

    $conf = $sv->configuration(  # Get configuration from dst_host vfiler_dst
        dst_host => vfiler_dst
    );

=cut

sub configuration {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my %config;
    my $dst_host;

    if ( exists $args{dst_host}
      and (ref $args{dst_host} ne 'NOTE::Vfiler')
      and (ref $args{dst_host} ne 'NOTE::Filer') ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    $dst_host = ( exists $args{dst_host} )? $args{dst_host}: $self->dst_host;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($dst_host);

    # Get configuration of snapvault relationship
    %config = SnapVault::snapvault_status_config(
        $dst_host->conn,
        dstpath => $self->dst_name,
        source  => $self->src_path
    );

    # Set back the context to filer context
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;
    croak "Could not return configuration" if ! %config;

    return \%config;
}

=head2 ::modify

Modifies the existing SnapVault configuration for the relationship.

=over

=item Arguments

B<src> => I<Qtree object> | I<string>

I<Optional.>
New source qtree or host path.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<tries_count> => I<number>

I<Optional.>
Snapvault tries count for this relationship.

B<back_up_open_files> => C<0> | C<1>

I<Optional.>
Allow the inclusion of open files on the primary system at the time
of the transfer.
ze> => I<number>

I<Optional.>
Window size for the snapvault transfer.

B<ignore_atime> => C<0> | C<1>

I<Optional.>
When set to 1, snapvault primary does not send files with only access time
changes during incremental transfers,
The default value is 0.

B<utf8_primary_path> => C<0> | C<1>

I<Optional.>
Specifies encoding format for the primary pathname. When set to 1 value
indicates that the primary pathname is in UTF8 format.
The default encoding format is ASCII/extended ASCII. The default value is 0.

=item Examples

=back

     $sv->modify(
         throttle => 100,
     );

     $sv->modify(
    back_up_open_files => 0,
        connection_mode    => 'inet',
    ignore_atime       => 1
     );

=cut

sub modify {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret, $sv_dest);

    if (exists($args{dst_host})
        && (   (ref $args{dst_host} ne 'NOTE::Vfiler')
            && (ref $args{dst_host} ne 'NOTE::Filer'))
      ) {
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    # Get the right destination host.
    $sv_dest = (exists $args{dst_host}) ? $args{dst_host} : $self->{_dst_host};

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    # Translate options to proper arguments for function
    # This function is not consistent with the rest of the
    # functions in SnapVault.pm
    $sv_args{kbs}   = $args{throttle}    if exists $args{throttle};
    $sv_args{tries} = $args{tries_count} if exists $args{tries_count};
    $sv_args{back_up_open_files} = $args{back_up_open_files}
      if exists $args{back_up_open_files};
    $sv_args{wsize}        = $args{wsize}        if exists $args{wsize};
    $sv_args{ignore_atime} = $args{ignore_atime} if exists $args{ignore_atime};
    $sv_args{utf8_primary_path} = $args{utf8_primary_path}
      if exists $args{utf8_primary_path};

    # Change the source qtree if option is given
    if (exists($args{src})) {
        $self->{_src} = NOTE::SnapVault::_qtree_path_attach($args{src});
    }
    $sv_args{source} = $self->src_path;
    if (exists $args{connection_mode}) {
        $self->{connection_mode} = $args{connection_mode};
        $sv_args{connection_mode} = $args{connection_mode};
    }

    elsif (defined NOTE::option(q/$CONN_MODE/)) {
        $sv_args{connection_mode} = NOTE::option(q/$CONN_MODE/);
        $self->{connection_mode} = $args{connection_mode};
    }

    # Modify configuration of snapvault relationship
    $ret =
      SnapVault::snapvault_modify($sv_dest->conn, $self->dst_name, %sv_args);

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    return $ret;
} ## end sub modify

=head2 ::restore

Starts a baseline restore transfer for a SnapVault relationship.
Returns a 'subtest' object, which may be ignored if the transfer completes
before the call to C<::restore> returns.

=over

=item Arguments

B<dst> => I<Qtree object> | I<string>

I<Dependent.>
A Qtree object or host path. This is where the data will be restored to
so this will become the new primary (src) for the SnapVault relationship
after calling L<C<::resync>|/::resync>.

Argument is optional when restarting a previously aborted restore,
otherwise it is required.

B<incremental> => C<0> | C<1>

I<Optional.>
Run baseline restore transfer if 0 is specified.
Run incremental restore transfer if 1 is specified.
The default is 0.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the destination to synchronize from.

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<background> => C<0> | C<1>

I<Optional.>
Run the restore transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->restore(
         dst => new_qtree_src,
     );

     $sv->restore(
         dst         => qtree_src,
     incremental => 1,
     );

     $subtest = $sv->restore(
         dst        => new_qtree_src,
         throttle   => 100,
         background => 1,
     );

     # Restart previously aborted restore
     $sv->restore;

=cut

sub restore {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);

    # Set flag for normal transfer
    $self->{_rxfer} = 1;

    if (exists($self->{_restore_restart})) {

        # Restart restore on same qtree
        $args{dst} = $self->{_restore_restart} unless exists $args{dst};
        $args{dst_host} = $self->{_restore_host} unless exists $args{dst_host};
        delete $self->{_restore_restart};
    }

    croak "Argument dst must be specified" unless exists $args{dst};

    if (!exists $args{dst_host}) {
        $args{dst_host} = $self->{_src_host};
    }

    if (ref $args{dst} ne 'NOTE::Qtree') {
        croak 'Destination has to be a NOTE::Qtree type only';
    }

    if (   (ref $args{dst_host} ne 'NOTE::Vfiler')
        && (ref $args{dst_host} ne 'NOTE::Filer')) {
        croak
          'Argument dst_host has to be either a NOTE::Filer or NOTE::Vfiler';
    }

    # The dst argument will become the new primary when resync is called
    $self->{_restore}      = NOTE::SnapVault::_qtree_path_attach($args{dst});
    $self->{_restore_host} = $args{dst_host};
    delete $args{dst};
    delete $args{dst_host};

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($self->{_restore_host});

    $op = "Snapvault Restore" . $self->description;
    NOTE::logcomment($op);

    # Process incremental option
    $args{incremental} = 0 unless exists $args{incremental};

    # Process background option
    if (exists $args{background}) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if !exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;
    if (exists $args{connection_mode}) {
        $self->{connection_mode} = $args{connection_mode};
        $sv_args{connection_mode} = $args{connection_mode};
    }

    elsif (defined NOTE::option(q/$CONN_MODE/)) {
        $sv_args{connection_mode} = NOTE::option(q/$CONN_MODE/);
        $self->{connection_mode} = $args{connection_mode};
    }

    my $host_name =
      (ref $self->{_restore_host} eq 'NOTE::Filer')
      ? $self->{_restore_host}->host
      : $self->{_restore_host}->name;
    my $host_path = $host_name . q{:} . $self->{_restore}->mount_path;

    # Start snapvault restore
    $ret = SnapVault::sv_restore($self->{_restore_host}->conn,
                                 $host_path,
                                 SOURCE => $self->dst_path,
                                 %sv_args);

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    if (ref $ret eq "Subtest") {    # Save, if it's an object
        push @{$self->{_subtests}}, {operation => $op, subtest => $ret};
    }

    return $ret;
} ## end sub restore

=head2 ::resync

Starts a resync-rollback transfer for a SnapVault relationship.
Returns a 'subtest' object, which may be ignored if the transfer completes
before the call to C<::resync> returns.

=over

=item Arguments

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<tries_count> => I<number>

I<Optional.>
Snapvault tries count for this relationship.

B<background> => C<0> | C<1>

I<Optional.>
Run the resync transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $sv->resync;

     $subtest = $sv->resync(
         background => 1,
         throttle   => 100,
     );

=cut

sub resync {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret, $op);

    # Check if coming from a restore
    if (exists($self->{_restore})) {

        # The qtree where data was restored to becomes the new primary
        $self->{_src}      = $self->{_restore};
        $self->{_src_host} = $self->{_restore_host};
        delete $self->{_restore};
        delete $self->{_restore_host};
    }

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($self->{_dst_host});

    $op = "Snapvault Resync" . $self->description;
    NOTE::logcomment($op);

    # Process background option
    if (exists $args{background}) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if !exists $args{timeout};

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;
    if (exists $args{connection_mode}) {
        $self->{connection_mode} = $args{connection_mode};
        $sv_args{connection_mode} = $args{connection_mode};
    }

    elsif (defined NOTE::option(q/$CONN_MODE/)) {
        $sv_args{connection_mode} = NOTE::option(q/$CONN_MODE/);
        $self->{connection_mode} = $args{connection_mode};
    }

    # Start snapvault resync
    $ret = SnapVault::sv_resync($self->{_dst_host}->conn,
                                $self->dst_path,
                                SOURCE => $self->src_path,
                                %sv_args);

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    if (ref $ret eq "Subtest") {    # Save, if it's an object
        push @{$self->{_subtests}}, {operation => $op, subtest => $ret};
    }

    # Create another entry on cleanup list
    push @{$self->{_schedules}}, {};

    return $ret;
} ## end sub resync

=head2 B<Checker/Fixer>

=head2 ::check

Checks the consistency of the SnapVault relationship.

The destination must exist and be in a 'snapvaulted' or 'quiesced' state.
The checker process runs on the destination host.

Returns a 'subtest' object if background => 1.

=over

=item Arguments

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<check_level> => C<data> | C<checksum>

I<Optional.>
Specify data comparison level as either data or checksum.
The default is C<data>.

B<quick_mode> => C<0> | C<1>

I<Optional.>
If 1, Starts C<snapvault check> in quick mode.
The default is 0.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the source to use.
By default the base snapshot for the SnapVault relationship will be used.

B<ignore> => C<ctime>

I<Optional.>
Ignores some type of comparisons during the check session. At present only
C<ctime> can be ignored.

B<background> => C<0> | C<1>

I<Optional.>
If 1, returns while checking is still running in the background.
The result of the check must be checked later using
L<C<::wait>|/::wait> or the C<wait> method of the returned object.
If the status is not checked using one of the wait methods
any errors in the check will go unnoticed.
The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
hosted.

=item Examples

=back

     $sv->check;

     $subtest = $sv->check(
         background => 1
     );

=cut

sub check {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $item, $ret, $op);
    my %tr_args = (
        fix         => 'VERIFIER_FIX',
        quick_mode  => 'VERIFIER_QUICK',
        check_level => 'VERIFIER_CKSUM',
        snapshot    => 'VERIFIER_SNAPSHOT',
        ignore      => 'VERIFIER_IGNORE',
    );
    my $dst_host;

    if ( exists $args{dst_host}
      and (ref $args{dst_host} ne 'NOTE::Vfiler')
      and (ref $args{dst_host} ne 'NOTE::Filer') ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    $dst_host = ( exists $args{dst_host} )? $args{dst_host}: $self->{_dst_host};

    $op = "Snapvault Check" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{complete};
    } else {
        $args{complete} = 1;
    }

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    # Process the check_level option
    $args{check_level} = 'data' if ! exists $args{check_level};

    # Translate arguments
    foreach $item ( keys %args ) {
        if ( exists $tr_args{$item} ) {
            $args{$tr_args{$item}} = $args{$item};
            delete $args{$item};
        }
    }

    # Convert all arguments to upper case
    %sv_args = map { uc($_) => $args{$_} } keys %args;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context( $dst_host );

    # Run the snapvault check
    if (ref $dst_host eq 'NOTE::Filer') {
        $ret = SnapVault::snapvault_check(
            $self->dst->Filer->conn,
            $dst_host->host . q{:} . $self->{_dst}->mount_path,
            SOURCE => $self->src_path,
            %sv_args,
        );
    } else {
        $ret = _snapvault_vfiler_check(
            $self->dst->Filer->conn,
            $dst_host->name,
            $dst_host->name . q{:} . $self->{_dst}->mount_path,
            SOURCE => $self->src_path,
            %sv_args,
        );
    }

    # Set back the context to filer context
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    return $ret;
} ## end sub check

=head2 ::fix

Checks the consistency of the SnapVault relationship
while repairing any inconsistencies.

The destination must exist and be in a 'snapvaulted' or 'quiesced' state.
The fixer process runs on the destination host.

Returns a 'subtest' object if background => 1.

=over

=item Arguments

B<throttle> => I<number>

I<Optional.>
Maximum transfer rate allowed, in KB/sec.

B<check_level> => C<data> | C<checksum>

I<Optional.>
Specify data comparison level as either data or checksum.
The default is C<data>.

B<quick_mode> => C<0> | C<1>

I<Optional.>
If 1, Starts C<snapvault check> in quick mode.
The default is 0.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot on the source to use.
By default the base snapshot for the SnapVault relationship will be used.

B<ignore> => C<ctime>

I<Optional.>
Ignores some type of comparisons during the check session. At present only
C<ctime> can be ignored.

B<background> => C<0> | C<1>

I<Optional.>
If 1, returns while checking is still running in the background.
The result of the check must be checked later using
L<C<::wait>|/::wait> or the C<wait> method of the returned object.
If the status is not checked using one of the wait methods
any errors in the check will go unnoticed.
The default is 0.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for command to complete, in seconds.
The default is object's default timeout.

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
hosted.

=item Examples

=back

     $sv->fix;

     $subtest = $sv->fix(
         background => 1
     );

=cut

sub fix {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    return $self->check(fix => 1, %args);
} ## end sub fix

=head2 B<Waiting>

=head2 ::wait

Waits for all background processes to complete.
These are the processes created by using the B<background> option of
L<C<::start>|/::start>, L<C<::stop>|/::stop>, L<C<::update>|/::update>,
L<C<::restore>|/::restore>, L<C<::resync>|/::resync>
or L<C<::snap_create>|/::snap_create>.

This method is completely unrelated to L<C<::status_wait>|/::status_wait>.

=over

=item Arguments

None.

=item Example

=back

     $sv->wait;

=cut

sub wait {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my ($obj, $test, $op);

    while ( $test = pop @{$self->{_subtests}} ) {
        $op  = $test->{operation};
        $obj = $test->{subtest};
        if ( $obj->is_alive() ) {
            NOTE::logcomment( "Waiting for: ".$op );
        }
        $obj->wait_finish();
        croak $@ unless FSR::FSRutils::subtestResults($obj);
    }

    return;
} ## end sub wait

=head2 ::status_wait

Wait for the SnapVault relationship to reach specific states.
Returns when all of the specified status elements
have their respective values I<at the same time>.
(Hint: don't use element/value combinations
that are mutually exclusive.)

Note that there are many SnapVault status elements,
one of which is thoughtfully named 'status'.
Another one is named 'state'.
There are none however named 'element' (not yet, anyway).

The status elements checked are the same ones
that may appear in the hash generated by L<C<::status>|/::status>.
(Internally C<::status> is used; this is a convenience method.)

Because you are a valued NOTE user,
the values for the status elements may be given
as regular expressions.

This method is completely unrelated to L<C<::wait>|/::wait>.

=over

=item Arguments

I<status element name> => I<status element value> | I<regex>

I<Required.>
The status elements to check are given as named arguments.
Their values may be given as literal strings
or regular expressions that (hopefully) match valid values.

B<src_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the source on which the snapvault relation is
hosted.

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
hosted.

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the relationship from the source host instead.
The default is 0 (destination host).
The argument, if specified, will be the arbiter as to whether source or destination
host is to be used, even if src_host or dst_host are specified.
If the argument is not specified, the src_host or dst_host will be used (accordingly
if only one of them is specified); or only dst_host will be used (if both are specified
or neither is specified)

B<restore> => C<0> | C<1>

I<Optional.>
Get the status of the restore transfer if set to 1.
The default is 0.

B<ignore_transfer_error> => C<0> | C<1>

I<Optional.>
If 1, don't fail if the status contains the C<current-transfer-error> element.
The default is 0.

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 5 seconds.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

B<dst_host> => I<Filer object> | I<Vfiler object>

I<Optional.>
A Note Filer/Vfiler object of the destination on which the snapvault relation is
hosted.

=item Examples

=back

    $sv->status_wait(
        status => 'idle'
    );

 $sv->status_wait(
        status => 'transferring'
    );

 $sv->status_wait(
        status => 'idle',
        dst_host => $vfiler_new,
    );


=cut

sub status_wait {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my (%opts, $wait_for, $matchall, $havekeys, $status, $etime, $delta);
    my $sleepis = 5;    # sleep number of seconds between status read
    my $noerror = 0;    # croak by default if 'current-transfer-error'

    $opts{source} = exists $args{source} ? $args{source} : 0;
    delete $args{source};
    $opts{restore} = exists $args{restore} ? $args{restore} : 0;
    delete $args{restore};

    if ( exists $args{src_host} ) {
        if ( (ref $args{src_host} ne 'NOTE::Vfiler')
          and (ref $args{src_host} ne 'NOTE::Filer') ){
            croak 'The source host has to be a Filer/Vfiler object';
        }
        $opts{src_host} = $args{src_host};
        delete $args{src_host};
    }

    if ( exists $args{dst_host} ) {
        if ( (ref $args{dst_host} ne 'NOTE::Vfiler')
          and (ref $args{dst_host} ne 'NOTE::Filer') ){
            croak 'The destination host has to be a Filer/Vfiler object';
        }
        $opts{dst_host} = $args{dst_host};
        delete $args{dst_host};
    }

    if( exists $args{'timeout'} ) {
        $etime = $args{'timeout'};
        delete $args{'timeout'};
    } else {
        $etime = $self->{_timeout};
    }

    if( exists $args{'sleep'} ) {
        $sleepis = $args{'sleep'};
        delete $args{'sleep'};
    }

    if( exists $args{'ignore_transfer_error'} ) {$noerror++;
        delete $args{'ignore_transfer_error'};
    }

    $wait_for = ''; # Show what we are waiting for
    foreach( keys %args ) { $wait_for .= "$_ =~ /$args{$_}/\n" }

    $matchall = 0;
    $havekeys = scalar( keys %args );

    # Set expiration time
    $etime += time;

    while( $matchall < $havekeys ) {
        NOTE::logcomment("status_wait:\n$wait_for");

        croak "Status Timeout waiting for:\n$wait_for" if time > $etime;

        $status = $self->status(%opts);
        $matchall = 0;

        if( $status->{'current-transfer-error'} && $noerror < 1 ) {
            croak $status->{'current-transfer-error'};
        }

        # check keys one by one and match all of them
        foreach( keys %args ) {
            $matchall++ if( $status->{$_} =~ qr/$args{$_}/i );
        }

        # Do not wait anymore if everything has been found
        last unless ( $matchall < $havekeys );

        # Wait for next iteration
        $delta = $etime - time;
        $delta = 0 if $delta < 0;
        $sleepis = $delta if $delta < $sleepis;
        sleep $sleepis if( $sleepis > 0 );
    }

    return;
} ## end sub status_wait

=head2 ::snap_sched

Configures or lists the SnapVault backup schedule for either the destination
or the source volume.

=over

=item Arguments

B<src_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the source host used in the new/
start method is only used. Only applicable if used with source argument

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot to configure or list.

B<sched> => I<string> | I<schedule_info>

I<Optional.>
Schedule to set for the given snapshot.

B<xfer> => C<0> | C<1>

I<Optional.>
Configure a 'xfer' snapshot instead of a 'create' snapshot. The default is 0.

B<source> => C<0> | C<1>

I<Optional.>
Configure the source volume. The default is 0 (destination volume).

B<preserve> => C<on> | C<off> | C<default>

I<Optional.>
If set to 'on' Create and retain the SnapVault archive Snapshots till the
allocated count, fail archive snapshot creation thereafter.
If set to 'off' Auto delete the oldest archive snapshot to create new archive
snapshot, the old snapvault behavior.
If set to 'default' Take the preserve value from global snapvault.snappreserve
setting and thus behavior will be as mentioned above depending on the value.

B<warn> => C<n>

I<Optional.>
Warn when remaining archive snapshot count reaches less than 'n'
=item Examples

=back

     @schedules = $sv->snap_sched(dst_host => vfiler);

     # List all backup schedules for the destination volume
     @schedules = $sv->snap_sched;

     # Get the schedule entry for the 'snapshot1' snapshot on the source volume
     $sched_h = $sv->snap_sched(
         source   => 1,
         snapshot => 'snapshot1',
     );

     # Set the schedule for the 'snapshot1' snapshot on the destination volume
     $sv->snap_sched(
         snapshot => 'snapshot1',
         sched    => '2@1,2'
     );

=cut

sub snap_sched {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my (@opts, @schedules, $sched_h,$sv_dest,$src_filer,$dst_filer);

    if ( exists($args{dst_host}) && ((ref $args{dst_host} ne 'NOTE::Vfiler') && (ref $args{dst_host} eq 'NOTE::Filer')) ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }
    if ( exists($args{src_host}) && ((ref $args{src_host} ne 'NOTE::Vfiler') && (ref $args{src_host} eq 'NOTE::Filer')) ){
        croak 'The source host has to be a Filer/Vfiler object';
    }
    $src_filer = (exists $args{src_host})? $args{src_host}:$self->{_src_host};
    $dst_filer = (exists $args{dst_host})? $args{dst_host}:$self->{_dst_host};

    # Get the right destination host.
    $sv_dest = $args{source} ? $src_filer: $dst_filer;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    my $qtree = $args{source} ? $self->{_src} : $self->{_dst};

    # Check for proper options
    if ( exists($args{sched}) && ! exists($args{snapshot}) ) {
        croak "Option 'sched' requires 'snapshot' to be defined";
    }

    if ( exists($args{xfer}) && ! exists($args{sched}) ) {
        croak "Option 'xfer' requires 'sched' to be defined";
    }

    # Add fixed parameters
    @opts = (
        { conn => $sv_dest->conn, source => $args{source} },
        $qtree->Volume->name,
    );

    # Add optional parameters
    push @opts, $args{snapshot}       if exists $args{snapshot};
    push @opts, $args{sched}          if exists $args{sched};
    push @opts, ($args{xfer} ? 1 : 0) ;
    push @opts, $args{preserve} if exists $args{preserve};
    push @opts, $args{warn} if exists $args{warn};

    # Configure or list the backup schedule
    # Use the '&' to avoid prototype of function so the
    # optional array '@opts' is not evaluated in a scalar context
    if ( exists($args{snapshot}) ) {
        # Must be evaluated in scalar context to get a hash reference
        $sched_h = &SnapVault::snapvault_snap_sched(@opts);
    } else {
        @schedules = &SnapVault::snapvault_snap_sched($opts[0],$opts[1]);
    }

    # Check for errors
    croak $@ if $@;

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    if ( exists $args{sched} ) {
        # Add schedule to cleanup hash
        $self->{_schedules}[-1]{$args{snapshot}.':'.$qtree->hostpath} = {
            volume => $qtree->Volume,
            source => $args{source},
        };
    }

    if ( exists($args{snapshot}) ) {
        return $sched_h;
    } else {
        return @schedules;
    }
}

=head2 ::snap_unsched

Unconfigures the SnapVault backup schedule for the given snapshot on either
the destination or the source volume.

=over

=item Arguments

B<src_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the source host used in the new/
start method is only used. Only applicable if used with source argument

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot to unconfigure.

B<source> => C<0> | C<1>

I<Optional.>
Unconfigure the source volume snapshot. The default is 0 (destination volume).

=item Examples

=back

     # Unconfigure snapshot on destination
     $sv->snap_unsched(
         snapshot => 'snapshot1',
     );

     # Unconfigure snapshot on destination
     $sv->snap_unsched(
         snapshot => 'snapshot1',
         dst_host => $new_vfiler,
     );

     # Unconfigure snapshot on source
     $sv->snap_unsched(
         source   => 1,
         snapshot => 'snapshot1',
     );

     # Unconfigure snapshot on source
     $sv->snap_unsched(
         source   => 1,
         snapshot => 'snapshot1',
         src_host => $new_vfiler,
     );

     # Unconfigure all snapshots created by this object
     $sv->snap_unsched;

=cut

sub snap_unsched {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my ($item, $key, $snapname, %removed, $ret,$src_filer,$dst_filer,$sv_dest);
    if ( exists($args{dst_host}) && ((ref $args{dst_host} ne 'NOTE::Vfiler') && (ref $args{dst_host} eq 'NOTE::Filer')) ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }
    if ( exists($args{src_host}) && ((ref $args{src_host} ne 'NOTE::Vfiler') && (ref $args{src_host} eq 'NOTE::Filer')) ){
        croak 'The source host has to be a Filer/Vfiler object';
    }
    $src_filer = (exists $args{src_host})? $args{src_host}:$self->{_src_host};
    $dst_filer = (exists $args{dst_host})? $args{dst_host}:$self->{_dst_host};

    # Get the right destination host.
    $sv_dest = $args{source} ? $src_filer: $dst_filer;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    my $qtree = $args{source} ? $self->{_src} : $self->{_dst};

    $ret = SnapVault::snapvault_snap_unsched(
            { conn => $sv_dest->conn, source => $args{source} },
            $qtree->Volume->name,
            $args{snapshot},
        );

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    if ( exists $args{snapshot} ) {
        # Remove schedule from cleanup hash
        delete $self->{_schedules}[-1]{$args{snapshot}.':'.$qtree->hostpath};
    } else {
        # Delete all schedules created by this object
        while ( @{$self->{_schedules}} ) {
            $item = $self->{_schedules}[0];
            while ( $key = each %$item ) {
                if ( ! $removed{$key} ) {
                    ($snapname) = ($key =~ /([^:]*):/);
                    if(!$snapname)
                    {
                        croak $@ if($@ && ($@!~/No matching snapshot schedule found/));
                    }
                    else {
                        croak $@ if ($@ && !$args{ignore_errors}) ;
                    }
                    # Keep track of schedules already removed
                    $removed{$key} = 1;
                }
                delete $item->{$key};
            }
            splice(@{$self->{_schedules}}, 0, 1);
        }
        $ret = 1;
    }

    return $ret;
}

=head2 ::release

Releases the source-side resources being held
for a particular SnapVault destination.

=over

=item Arguments

B<src_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which the relation needs to be stopped from.
If this parameter is not specified than the source host used in the new/
start method is only used.

B<restore> => C<0> | C<1>

I<Optional.>
Release the destination-side resources of a restore relationship.
The default is 0.

=item Example

=back

     $sv->release(src_host => $new_vfiler);

     $sv->release;

     $sv->release(
         restore => 1,
         src_host => $vfiler_dst_new
     );

=cut

sub release {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my ($ret,$op);

    $op = "Snapvault Release" . $self->description;
    NOTE::logcomment($op);

    if (exists($args{src_host})
        && (   (ref $args{src_host} ne 'NOTE::Vfiler')
            && (ref $args{src_host} ne 'NOTE::Filer'))
      ) {
        croak 'The destination host has to be a Filer/Vfiler object';
    }

    # Get the right destination host.
    my $sv_dest = (exists $args{src_host}) ? $args{src_host} : $self->{_src_host};
    my $src_name = $self->src_name;
    my $dst_path = $self->dst_path;

    NOTE::logcomment( "Snapvault Release" . $self->description );

    if ( $args{restore} ) {
        croak "Option 'restore' is only valid after calling restore"
          if ! $self->{_restore};

        $sv_dest = (exists $args{src_host}) ? $args{src_host} : $self->{_dst_host};;
        $src_name = $self->dst_name;
        $dst_path = $self->restore_path;
    }

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    # Release the source-side resources for snapvault
    $ret = SnapVault::snapvault_release(
        { conn => $sv_dest->conn, source => $args{restore} ? 0 : 1 },
        $src_name,
        $dst_path,
    );

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    # Set flag that this relationship has already been released
    $self->{_released} = 1;

    return $ret;
}

=head2 ::snap_create

Creates a configured SnapVault snapshot schedule on either
the destination or the source volume.
Returns a 'subtest' object, which may be ignored if the command completes
before the call to C<::snap_create> returns.

=over

=item Arguments

B<src_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the source host used in the new/
start method is only used. Only applicable if used with source argument

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<snapshot> => I<string>

I<Required.>
The name of the configured snapshot to create.

B<source> => C<0> | C<1>

I<Optional.>
Create the snapshot on the source volume.
The default is 0 (destination volume).

B<background> => C<0> | C<1>

I<Optional.>
Run the command in the background.
Only used when snapshot is created on the destination. The default is 0.

B<timeout> => I<number>

I<Optional.>
Timeout period waiting for the command to complete.
Only used when snapshot is created on the destination.
The default is object's default timeout.

=item Examples

=back

     # Create snapshot on the destination
     $sv->snap_create(
         snapshot => 'snapshot1',
     );

     # Create snapshot on the destination (background)
     $subtest = $sv->snap_create(
         snapshot   => 'snapshot1',
         dst_host   => $vfiler_dst,
         background => 1
     );

     # Create snapshot on the source
     $sv->snap_create(
         snapshot => 'snapshot1',
         source   => 1,
         src_host => $vfiler_src,
     );

=cut

sub snap_create {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, $ret, $src_filer,$dst_filer,$sv_dest,$op);
    my $snapshot = $args{snapshot};

    # Set flag for normal transfer
    $self->{_rxfer} = 0;

    $op = "Snapvault Snapshot Create\n" . $self->description;
    NOTE::logcomment($op);

    if ( exists($args{dst_host}) && ((ref $args{dst_host} ne 'NOTE::Vfiler') && (ref $args{dst_host} eq 'NOTE::Filer')) ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }
    if ( exists($args{src_host}) && ((ref $args{src_host} ne 'NOTE::Vfiler') && (ref $args{src_host} eq 'NOTE::Filer')) ){
        croak 'The source host has to be a Filer/Vfiler object';
    }
    $src_filer = (exists $args{src_host})? $args{src_host}:$self->{_src_host};
    $dst_filer = (exists $args{dst_host})? $args{dst_host}:$self->{_dst_host};

    # Get the right destination host.
    $sv_dest = $args{source} ? $src_filer: $dst_filer;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    # Process the timeout option
    $args{timeout} = $self->{_timeout} if ! exists $args{timeout};

    if ( $args{source} ) {
        # Create snapshot on the source volume
        $ret = SnapVault::snapvault_snap_create(
            $sv_dest->conn,
            $self->{_src}->Volume->name,
            $snapshot,
        );

        # Check for errors
        croak $@ if $@;
    } else {
        delete $args{snapshot};
        delete $args{source};

        # Process background option
        if ( exists $args{background} ) {
            $args{complete} = $args{background} ? 0 : 1;
            delete $args{background};
        } else {
            $args{complete} = 1;
        }

        # Convert all arguments to upper case
        %sv_args = map { uc($_) => $args{$_} } keys %args;

        # Create snapshot on the destination volume
        $ret = SnapVault::sv_snap_create(
            $sv_dest->conn,
            $self->{_dst}->Volume->name,
            $snapshot,
            %sv_args,
        );

        # Check for errors
        croak $@ if $@;
    }

    # Save destination snapshot for use with ::snap_abort
    $self->{_snapshot} = $snapshot;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    return $ret;
}

=head2 ::snap_status

Returns a list of configured SnapVault snapshot targets for the destination
or source volume of the relationship or a hash reference of the schedule if
the 'snapshot' argument is specified.

=over

=item Arguments

B<src_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the source host used in the new/
start method is only used. Only applicable if used with source argument

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<snapshot> => I<string>

I<Optional.>
The name of the configured snapshot.

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the snapshot from the source filer instead.
The default is 0 (destination filer).

=item Examples

=back

     # Get list of snapshot targets for destination volume
     @schedules = $sv->snap_status;

     # Get the status info of the specified snapshot target on the source
     $status = $sv->snap_status(
         snapshot => 'snap_name',
         source => 1
     );

     # Get list of snapshot targets for destination volume
     @schedules = $sv->snap_status(dst_host => $pfiler_dst);

     # Get the status info of the specified snapshot target on the source
     $status = $sv->snap_status(
         snapshot => 'snap_name',
         source   => 1,
         src_host => $pfiler_src,
     );

=cut

sub snap_status {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my (%sv_args, @ret,$src_filer,$dst_filer,$sv_dest);

    if ( exists($args{dst_host}) && ((ref $args{dst_host} ne 'NOTE::Vfiler') && (ref $args{dst_host} ne 'NOTE::Filer')) ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }
    if ( exists($args{src_host}) && ((ref $args{src_host} ne 'NOTE::Vfiler') && (ref $args{src_host} ne 'NOTE::Filer')) ){
        croak 'The source host has to be a Filer/Vfiler object';
    }
    $src_filer = (exists $args{src_host})? $args{src_host}:$self->{_src_host};
    $dst_filer = (exists $args{dst_host})? $args{dst_host}:$self->{_dst_host};

    # Get the right destination host.
    $sv_dest = $args{source} ? $src_filer: $dst_filer;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    my $qtree = $args{source} ? $self->{_src} : $self->{_dst};

    $sv_args{SNAPSHOT} = $args{snapshot} if exists $args{snapshot};

    # Get status of snapvault relationship
    @ret = SnapVault::snapvault_status_snapshot(
        { conn => $sv_dest->conn, source => $args{source} },
        VOLUME => $qtree->Volume->name,
        %sv_args
    );

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for error
    croak $@ if $@;

    return (exists $args{snapshot} ? $ret[0] : @ret);
}

=head2 ::snap_abort

Aborts the creation of an active snapshot target.

=over

=item Arguments

B<src_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the source host used in the new/
start method is only used. Only applicable if used with source argument

B<dst_host> => I<Note Filer or vfiler object>

I<Optional.>
The Filer/Vfiler host object from which schedules can be set or changed etc.
If this parameter is not specified than the destination host used in the new/
start method is only used.

B<snapshot> => I<string>

I<Optional.>
The name of a snapshot to abort.
The default is the snapshot specified in the last call to
L<C<::snap_create>|/::snap_create>.

B<source> => C<0> | C<1>

I<Optional.>
Unconfigure the source volume snapshot. The default is 0 (destination volume).

=item Examples

=back

     # Abort the snapshot create previously created by ::snap_create
     $sv->snap_abort;

     # Abort the snapshot create on the source
     $sv->snap_abort(
         snapshot => 'snapshot1',
         source   => 1
     );

     # Abort the snapshot create previously created by ::snap_create
     $sv->snap_abort(dst_host => $pfiler_dst);

     # Abort the snapshot create on the source
     $sv->snap_abort(
         snapshot => 'snapshot1',
         source   => 1,
         src_host => $pfiler_src,

     );

=cut

sub snap_abort {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($self, %args) = @_;
    my $ret;
    my $snapshot = $args{snapshot} || $self->{_snapshot};
    my $qtree = $args{source} ? $self->src : $self->dst;

    croak "Snapshot name must be specified" unless $snapshot;

    if ( exists($args{dst_host}) && ((ref $args{dst_host} ne 'NOTE::Vfiler') && (ref $args{dst_host} eq 'NOTE::Filer')) ){
        croak 'The destination host has to be a Filer/Vfiler object';
    }
    if ( exists($args{src_host}) && ((ref $args{src_host} ne 'NOTE::Vfiler') && (ref $args{src_host} eq 'NOTE::Filer')) ){
        croak 'The source host has to be a Filer/Vfiler object';
    }
    my $src_filer = (exists $args{src_host})? $args{src_host}:$self->{_src_host};
    my $dst_filer = (exists $args{dst_host})? $args{dst_host}:$self->{_dst_host};

    # Get the right destination host.
    my $sv_dest = $args{source} ? $src_filer: $dst_filer;

    # Set the right vfiler context depending on the destination host
    $self->_set_vfiler_context($sv_dest);

    # Abort the snapshot create
    $ret = SnapVault::snapvault_abort_snapshot(
        { conn => $sv_dest->conn, source => $args{source} },
        $qtree->Volume->name,
        $snapshot,
    );

    # Set back the right vfiler context wrt the established relation
    $self->_set_vfiler_context();

    # Check for errors
    croak $@ if $@;

    return $ret;
}

=head2 B<Component Objects>

=head2 ::src

Returns the Qtree object representing the current source.

=over

=item Arguments

None.

=item Examples

=back

     $src = $sv->src;

=cut

sub src {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    return $self->{_src};
}

=head2 ::dst

Returns the Qtree object representing the current destination.

=over

=item Arguments

None.

=item Examples

=back

     $dst = $sv->dst;

=cut

sub dst {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    return $self->{_dst};
}

=head2 ::src_host

Returns the Filer/Vfiler object representing the current source.

=over

=item Arguments

None.

=item Examples

=back

     $src = $sv->src_host;

=cut

sub src_host {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    return $self->{_src_host};
}

=head2 ::dst_host

Returns the Filer/Vfiler object representing the current destination.

=over

=item Arguments

None.

=item Examples

=back

     $dst = $sv->dst_host;

=cut

sub dst_host {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    return $self->{_dst_host};
}

=head2 B<Names and Paths>

=head2 ::src_name

Returns the mount_path of the source Qtree.This is of the form
/vol/<vol-name>/<qtree-name>

=over

=item Arguments

None.

=item Example

=back

     $srcname = $sv->src_name;

=cut

sub src_name {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    return $self->{_src}->mount_path;
}

=head2 ::dst_name

Returns the mount path of the destination Qtree.This is of the form
/vol/<vol-name>/<qtree-name>

=over

=item Arguments

None.

=item Example

=back

     $dstname = $sv->dst_name;

=cut

sub dst_name {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    return $self->{_dst}->mount_path;
}

=head2 ::src_path

Returns the host-qualified path name of the source Qtree.This is not the same as
the qtree->hostpath.Here, the src_path returns qtree path wrt the snapvault
source.Like, for example, vfiler:/vol/src/qt.

=over

=item Arguments

None.

=item Example

=back

     $srcpath = $sv->src_path;

=cut

sub src_path {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    my $host_name =
        $self->{_vfiler_src}
      ? $self->{_src_host}->name
      : $self->{_src_host}->host;
    my $hostpath = $host_name . q{:} . $self->{_src}->mount_path;
    return $hostpath;
} ## end sub src_path

=head2 ::dst_path

Returns the host-qualified path name of the destination Qtree.This is not the
same as the qtree->hostpath.Here, the src_path returns qtree path wrt the
snapvault destination.Like, for example, vfiler:/vol/dst/qt.

=over

=item Arguments

None.

=item Example

=back

     $dstpath = $sv->dst_path;

=cut

sub dst_path {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    my $host_name =
        $self->{_vfiler_dst}
      ? $self->{_dst_host}->name
      : $self->{_dst_host}->host;
    my $hostpath = $host_name . q{:} . $self->{_dst}->mount_path;
    return $hostpath;
} ## end sub dst_path

=head2 ::restore_path

Returns the host-qualified path name of the restored Qtree.This is not the
same as the qtree->hostpath.Here, the src_path returns qtree path wrt the
snapvault destination.Like, for example, vfiler:/vol/src/qt.

=over

=item Arguments

None.

=item Example

=back

     $dstpath = $sv->restore_path;

=cut

sub restore_path {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    my $host_name =
      (ref $self->{_restore_host} eq 'NOTE::Filer')
      ? $self->{_restore_host}->host
      : $self->{_restore_host}->name;
    my $hostpath = $host_name . q{:} . $self->{_restore}->mount_path;
    return $hostpath;
} ## end sub restore_path

=head2 B<Object Identification>

=head2 ::description

Returns a description of the SnapVault relationship consisting of the tag for
the source, a rightward pointing arrow, and the tag for the destination.

=over

=item Arguments

None.

=item Example

=back

     $desc = $sv->description;

=cut

sub description {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my $self = shift;
    my $desc;

    if ($self->{_rxfer} && exists($self->{_restore})) {
        $desc =
            "\n    +---"
          . $self->dst_path . q{(}
          . $self->dst->tag . q{)}
          . "\n    +-> "
          . $self->restore_path . q{(}
          . $self->{_restore}->tag . q{)};
    } else {
        $desc =
            "\n    +---"
          . $self->src_path . q{(}
          . $self->src->tag . q{)}
          . "\n    +-> "
          . $self->dst_path . q{(}
          . $self->dst->tag . q{)};
    }

    return $desc;
} ## end sub description

=head2 B<Object Configuration>

=head2 ::timeout

Sets or returns the current default timeout for the object before being set.
A value of -1 is to wait forever.

=over

=item Arguments

B<timeout> => I<number>

I<Optional.>
Set default timeout for the object.

=item Example

=back

     $curr_timeout = $sv->timeout;    # Get current default timeout

     $orig_timeout = $sv->timeout(
         timeout => 300               # Set default timeout
     );

=cut

sub timeout {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;

    $self->{_timeout} = $args{timeout} if exists $args{timeout};
    $self->{_timeout} = -1             if !defined $self->{_timeout};
    return $self->{_timeout};
}

=head2 B<Object Destruction>

=head2 ::destroy

Destroy the SnapVault relationship by unconfiguring both the source
and destination and removing the registry entries for the relationship.
It also deletes all snapshot schedules created by this object.

=over

=item Arguments

None.

=item Example

=back

     $sv->destroy;

=cut

sub destroy {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;
    my $conf;
    $self->{_destroyed} = 1;

    NOTE::logcomment( "Destroying Snapvault " . $self->description );

    # Check if this is a valid relationship
    $conf = $self->configuration;

    # Abort any transfer currently active, ignore error
    eval {
        if ( $conf && $self->abort(hard => 1) ) {
            $self->status_wait(
                status => 'idle',
                ignore_transfer_error => 1,
            );
        }
    };

    # Delete all schedules created by this object
    eval { $self->snap_unsched(ignore_errors => 1) };

    # Unconfigure destination
    $self->stop unless $self->{_stopped};

    # Release the source-side resources for snapvault
    $self->release unless $self->{_released};

    NOTE::logcomment("Snapvault relationship destroyed");

    return;
}

# Internal function to license, enable and open access for SnapVault
sub _license_and_enable {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($self, %args) = @_;
    my $dst_enable = exists($args{dst_enable}) ? $args{dst_enable} : 1;
    my ($host_entries, $host_ip, $host_name);

    # Enable SnapVault licenses on both source and destination
    $self->{_src}->Filer->check_license('sv_ontap_pri');
    $self->{_dst}->Filer->check_license('sv_ontap_sec');

    # Make /etc/host entries on both the src and dst host

    # Making entries on the snapvault source
    $host_name =
        $self->{_vfiler_dst}
      ? $self->{_dst_host}->name
      : $self->{_dst_host}->host;
    $host_ip =
        $self->{_vfiler_dst}
      ? $self->{_dst_host}->ipa
      : $self->{_dst_host}->defaultip;

    # If the snapvault source is a vfiler,
    if ($self->{_vfiler_src}) {

        # Make host entry for the snapvault destination
        $self->{_src_host}
          ->vfiler_host_entry(host => $host_name, ip => $host_ip);
    }

    # else if snapvault source is pfiler,
    else {

        # Read the host entries on the source
        $host_entries = filer_file_read($self->{_src_host}->conn, '/etc/hosts');

        # If the source does not contain any host entries for destination,
        # Then make these entries in /etc/hosts of vfiler0.
        if ($host_entries !~ /$host_ip\s+$host_name/) {
            $host_entries = $host_entries . "\n" . "$host_ip  $host_name\n";
            wrfile($self->{_src_host}->conn, '/etc/hosts', $host_entries)
              or croak $@;
        }
    } ## end else [ if ($self->{_vfiler_src...

    # Repeat same procedure on the snapvault destiation
    $host_name =
        $self->{_vfiler_src}
      ? $self->{_src_host}->name
      : $self->{_src_host}->host;
    $host_ip =
        $self->{_vfiler_src}
      ? $self->{_src_host}->ipa
      : $self->{_src_host}->defaultip;

    if ($self->{_vfiler_dst}) {
        $self->{_dst_host}
          ->vfiler_host_entry(host => $host_name, ip => $host_ip);
    } else {
        $host_entries = filer_file_read($self->{_dst_host}->conn, '/etc/hosts');
        if ($host_entries !~ /$host_ip\s+$host_name/) {
            $host_entries = $host_entries . "\n" . "$host_ip  $host_name\n";
            wrfile($self->{_dst_host}->conn, '/etc/hosts', $host_entries)
              or croak $@;
        }
    }

    # Open SnapVault access to all on both source and destination
    # Also for vfiler objects, enable httpd and rsh and snapvault options.
    if ($self->{_vfiler_src}) {
        $self->{_src_host}->set_options();
    } else {
        $self->{_Filer} = $self->{_src_host};
        $self->{'snapvault.access'} = 'all';
        $self->{_src_host}->snapvault_enable;
    }

    if ($self->{_vfiler_dst}) {
        $self->{_dst_host}->set_options();
    } else {
        $self->{_Filer} = $self->{_dst_host};
        $self->{'snapvault.access'} = 'all';
        $self->{_dst_host}->snapvault_enable if $dst_enable;
    }

    return;
} ## end sub _license_and_enable

# Internal function to set the vfiler context depending on the vfiler host
# This in turn calls the SnapVault::set_vfiler_args to set the vfiler flag in
# SnapVault.pm and to set the vfiler_args.
sub _set_vfiler_context {
    my ($self, $destination_is) = @_;

    my %vf_args;

    if (!defined($destination_is)) {
        SnapVault::set_vfiler_args(0, undef);
    }

    if (ref $destination_is eq 'NOTE::Vfiler') {
        $vf_args{_user_name} = $destination_is->username;
        $vf_args{_password}  = $destination_is->password;
        SnapVault::set_vfiler_args(1, \%vf_args);
    } elsif (ref $destination_is eq 'NOTE::Filer') {
        SnapVault::set_vfiler_args(0, undef);
    }
    return;
} ## end sub _set_vfiler_context

# Internal function to perform snapvault check in vfiler context
# This method is required because the NATE method snapvault_check
# does a priv set -q test on the conn object; and priv is not supported
# on vfilers; so snapvault_check cannot be used. So on the filer conn,
# a priv set -q test and a vfiler run is done by this method to perform
# the snapvault check
#
# name ::_snapvault_vfiler_check
# usage _snapvault_vfiler_check($filerConn, $vfiler, $dst, %options)
# synopsis
#    Starts a snapvault check session for the specified secondary qtree.
#    This function can only be run in vfiler context on the secondary filer.
# param SOURCE Required: the source <filer>:<path> for the SnapVault relationship
# param THROTTLE Optional: kbs rate to configure and throttle transfers with
# param VERIFIER_FIX Optional: run in fix mode.
# param VERIFIER_QUICK Optional: run in quick mode.
# param VERIFIER_CKSUM Optional: data comparison level.
# param VERIFIER_SNAPSHOT Optional: specific snapshot on the secondary
#        filer to be used.
# param VERIFIER_IGNORE Optional: ignore some type of comparisons (ctime).
# param TIMEOUT Optional: timeout period waiting for the command to complete
# param COMPLETE Optional: wait for transfer to complete before returning if set
# result
#    The subtest waiting for the transfer to finish logs a FATAL result
#    with key EVAL_ERROR if the transfer fails.  Under normal circumstances,
#    the calling script should wait for the subtest to complete and then call
#    FSR::FSRutils::subtestResults() to determine whether the transfer succeeded.
sub _snapvault_vfiler_check($$;%) {
    my $trace = $NOTE::TRACELOG->start_timer();    # log internal metrics

    my ($conn, $vfiler, $dst, %param) = @_;
    my ($cmd, $text, $obj, %opts);
    my $sv = 'snapvault';    # SnapVault command on the filer

    return eval {
        $param{'WAIT'} = 1;  # force ONTAP command to wait for completion
        $param{'FORCE'} = 1; # never answer questions, just do it
        $cmd = rpl_cmd_from_opts("$sv check", 'start', $dst, %param);

        # Make changes to the command generated
        $cmd =~ s/priv set -q diag ; /priv set -q diag ; vfiler run $vfiler /;

        # Set options for rpl commands
        $opts{TIMEOUT} = $param{TIMEOUT} if exists $param{TIMEOUT};

        # Try starting the check
        my $command = '[Ss]napvault';

        # This is a hardcoded regexp which corresponds to the ONTAP code.
        # The parentheses also determine how we determine errors below.
        # Don't change these unless you know exactly what you're doing.

        # check successfully started.
        my $waitstring = "(?m)" .
          "^===== $vfiler\\s+" .
          "$command check started\\.\\s+" .
          "Monitor progress with '$sv status(?: -l)?' " .
          "or the snapvault log\\.\\s*(?:\\n|\$)";

        $text = rpl_start_op($conn, $cmd, $waitstring, %opts);
        croak $@ if ($@);
        croak $text if $text;

        # The check successfully started,
        # so now we just need to wait for the completion string.
        $obj = subtest( sub { rpl_finish_xfer($conn, $sv, %opts); },
                           '-bg', '-onexit', sub{} );

        # If specified, wait for the transfer to complete
        # and check the results once it does.
        if ($param{'COMPLETE'}) {
            $obj->wait_finish();
            subtestResults($obj) or croak $@;
        }

        return $obj;
    };
} ## end sub _snapvault_vfiler_check

# Object destructor
sub DESTROY {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $self = shift;

    # Only the creator can destroy the object
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_spare_me} );    # Means don't destroy
    return if ( $self->{_destroyed} );   # Already dead
    return if ( NOTE->ALL_SPARED );

    # Destroy object
    $self->destroy;
}

=head1 SEE ALSO

L<C<NOTE::SnapVault>|engineering::ontap_tests::main::docs::lib-NOTE-SnapVault-pm/>

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::main::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Qtree>|engineering::ontap_tests::main::docs::lib-NOTE-Qtree-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>

=head1 STATUS

Public

$Id: //depot/prod/test/Rsierranevada/lib/NOTE/Vfiler/SnapVault.pm#1 $

$Date: 2013/01/22 $

=head1 AUTHOR

Akshatha.G@netapp.com

Sumukh.Bhardwaj@netapp.com

=cut

1;
