#
# $Id: //depot/prod/test/Rsierranevada/lib/NOTE/Volume/Copy.pm#1 $
#
## @summary  NOTE Volume copy class
## @status   Public
## @pod      here
## @author   dl-dpgqa-automation@netapp.com

=head1 NAME

NOTE::Volume::Copy

=head1 DESCRIPTION

This is the NOTE class for working with volume copy relationships.

=head1 METHODS

=cut

use strict;

package NOTE::Volume::Copy;

use NOTE;
use Carp;
use Volume::Copy qw();
use FSR::FSRutils qw();

=head2 B<Relationship Creation>

=head2 ::start

Creates a volume copy relationship
and returns an object representing the relationship.
Uses L<C<volcopy_start>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm/volcopy_start($conn,$volname[,%opts])>
function from
L<C<Volume::Copy>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm>

=over

=item Arguments

B<src> => I<Volume object> | I<string>

I<Dependent.>
A Volume object or a host path.

B<dst> => I<Volume object> | I<string>

I<Dependent.>
A Volume object or a host path.

Both arguments are required when C<::start> is called as a class method
and neither may be used when called via an object.

A host path is a full path name prepended with the filer name.

B<copy_all> => C<0> | C<1>

I<Optional.>
Copy all the snapshots from the source volume.

B<snapshot> => I<string>

I<Optional.>
The name of the snapshot on the source volume to copy.

B<background> => C<0> | C<1>

I<Optional.>
Run the transfer in the background. The default is 0.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Examples

=back

     $vc = NOTE::Volume::Copy->start(
         src => $src_vol,
         dst => $dst_vol,
     );

     $vc = NOTE::Volume::Copy->start(
         src        => 'filer_src:vol_src',
         dst        => 'filer_dst:vol_dst',
         background => 1
     );

     $vc->start;

=cut

sub start {
    my ($class, %args) = @_;
    my $self = {};
    my (%vc_args, $ret, $op_num, $op);

    # Most likely an object. note that 'ref $class' is not compared
    # to NOTE::Volume::Copy because this method can be inherited by a
    # subclass and 'ref $class' will not reflect 'NOTE::Volume::Copy'
    # in the subclass.
    if ( ref $class ne '' ) {
        $self = $class;

        if ( exists $args{src} || exists $args{dst} ) {
            croak "Cannot override existing 'src' or 'dst' attributes ".
                  "in existing object.";
        }
    } else {
        # Instantiate this object
        $self = NOTE::Volume::Copy->new( %args );
    }
    NOTE::reference( $self );

    delete $args{src};
    delete $args{dst};

    $op = "Volume Copy Start" . $self->description;
    NOTE::logcomment( $op );

    # Process background option
    if ( exists $args{background} ) {
        $args{complete} = $args{background} ? 0 : 1;
        delete $args{background};
    } else {
        $args{complete} = 1;
    }

    # Override default timeout
    $self->timeout(%args);
    $args{timeout} = $self->{_timeout};

    # Convert all arguments to upper case
    %vc_args = map { uc($_) => $args{$_} } keys %args;

    # When doing a volcopy from and to the local filer
    # the volcopy commmand output is not going to include
    # the name of the filer but 127.0.0.1 instead
    # To make sure volcopy_start does the right thing
    # do not include the filer name in the source path
    my $SOURCE = $self->src_path;
    if ($self->{_dst}->Filer->host eq $self->{_src}->Filer->host){
        $SOURCE = $self->{_src}->name;
    }

    # Start vol copy relationship
    ($ret, $op_num) = Volume::Copy::volcopy_start(
        $self->{_dst}->Filer->conn,
        $self->{_dst}->name,
        SOURCE => $SOURCE,
        RESTRICT => 1,
        %vc_args
    );

    # Check for errors
    croak $@ if $@;

    if ( ref $ret eq "Subtest" ) { # Save, if it's an object
        push @{$self->{_subtests}}, { operation => $op, subtest => $ret };
    }

    # Save the operation number for this transfer
    $self->{_op_num} = $op_num;

    return $self;
}

=head2 ::attach

Creates an object representing a volume copy relationship,
but does B<not> affect or cause the creation
of a volume copy relationship on the filer.
If the source and destination given
are already in a volume copy relationship with each other
then other C<NOTE::Volume::Copy> methods may be used
to control and manipulate the relationship
just as if L<C<::start>|/::start>
was used to create the relationship.
If the source and destination given
are not already in a volume copy relationship with each other
the L<C<::start>|/::start> method
of the returned object may be used to create a relationship.

=over

=item Arguments

B<src> => I<Volume object> | I<string>

I<Dependent.>
A Volume object or a host path.

B<dst> => I<Volume object> | I<string>

I<Dependent.>
A Volume object or a host path.

Both arguments are required when C<::attach> is called as a class method
and neither may be used when called via an object.

A host path is a full path name prepended with the filer name.

B<op_num> => I<number>

I<Optional.>
Operation number of currently transferring vol copy.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Example

=back

     $vc = NOTE::Volume::Copy->attach(
         src => $src_vol,
         dst => $dst_vol,
         op_num => 1,
     );

     $vc = NOTE::Volume::Copy->attach(
         src => 'filer_src:vol_src',
         dst => 'filer_dst:vol_dst',
         op_num => 0,
     );

     $vc->attach(
         timeout => 600,
     );

=cut

sub attach {
    my ($class, %args) = @_;
    my $self = {};

    # Most likely an object. note that 'ref $class' is not compared
    # to NOTE::Volume::Copy because this method can be inherited by a
    # subclass and 'ref $class' will not reflect 'NOTE::Volume::Copy'
    # in the subclass.
    if ( ref $class ne '' ) {
        $self = $class;

        if ( exists $args{src} || exists $args{dst} ) {
            croak "Cannot override existing 'src' or 'dst' attributes ".
                  "in existing object.";
        }

        # Override defaults
        $self->timeout(%args);
        $self->{_op_num} = $args{op_num} if exists $args{op_num};
    } else {
        # Instantiate this object
        $self = NOTE::Volume::Copy->new( %args );
    }

    # Do not destroy relationship when the object is destroyed
    $self->{_spare_me} = 1;

    return $self;
}

=head2 ::new

Creates an object representing a vol copy relationship,
but does B<not> cause the creation
of a vol copy relationship on the filer.
The L<C<::start>|/::start> method
of the returned object may be used to create the relationship.

This method should not normally be used;
use L<C<::start>|/::start> directly instead.

=over

=item Arguments

B<src> => I<Volume object> | I<string>

I<Required.>
A Volume object or a host path.

B<dst> => I<Volume object> | I<string>

I<Required.>
A Volume object or a host path.

A host path is a full path name prepended with the filer name.

B<op_num> => I<number>

I<Optional.>
Operation number of currently transferring vol copy.

B<timeout> => I<number>

I<Optional.>
Default timeout for methods which take a timeout option.
The default is 600 seconds.

=item Example

=back

     $vc = NOTE::Volume::Copy->new(
         src => $src_vol,
         dst => $dst_vol,
     );

     $vc = NOTE::Volume::Copy->new(
         src => 'filer_src:vol_src',
         dst => 'filer_dst:vol_dst',
     );

=cut

sub new {
    my ($class, %args) = @_;
    my $self = {};

    # Check arguments
    if ( !exists $args{src} || !exists $args{dst} ) {
        croak "Src or dst argument is missing";
    }

    # Set private variables for object
    $self->{_owner}   = $$;
    $self->{_timeout} = exists $args{timeout} ? $args{timeout} : 600;
    $self->{_src}     = _vol_path_attach($args{src});
    $self->{_dst}     = _vol_path_attach($args{dst});
    $self->{_op_num}  = $args{op_num};

    # Get total used space on the source volume
    $self->{_src_used} = $self->{_src}->used;

    # Bless this object
    bless $self, $class;

    return $self;
}

=head2 B<Volume::Copy Commands>

The methods names in this section correspond to ONTAP's
L<C<vol copy>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm>
subcommands.
Only administrator privilege level subcommands are listed in this section.
These methods eventually call the self-named subcommands.
Only the briefest mention of what these subcommands do is given here.
See the vol copy documentation for more information on them.

Vol copy commands that apply to the filer as a whole are part of
L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>
and are described in the
L<C<NOTE::Filer::VolCopy>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/volcopy>
section.

=head2 ::abort

Aborts an active vol copy transfer on the destination.
Uses L<C<volcopy_abort>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm/volcopy_abort($conn,$operation)>
function from
L<C<Volume::Copy>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm>

=over

=item Examples

=back

     $vc->abort;

=cut

sub abort {
    my ($self, %args) = @_;
    my $ret;

    croak "Transfer for " . $self->description . " is not active"
      unless defined $self->{_op_num};

    NOTE::logcomment( "Vol Copy Abort" . $self->description );

    # Abort transfer
    $ret = Volume::Copy::volcopy_abort(
        $self->{_dst}->Filer->conn,
        $self->{_op_num},
    );

    # Check for errors
    croak $@ if $@;

    # Undefine the operation number so it won't accidentally abort another
    # transfer started by someone else
    undef $self->{_op_num};

    return $ret;
}

=head2 ::status

Returns a hash reference to the status of the relationship.
Uses L<C<volcopy_status>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm/volcopy_status($conn,$operation)>
function from
L<C<Volume::Copy>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm>

=over

=item Examples

=back

     $status = $vc->status;

=cut

sub status {
    my ($self, %args) = @_;
    my $ret;
    my $status = {};

    if ( ! defined $self->{_op_num} ) {
        $status->{status} = 'idle';
    } else {
        $ret = Volume::Copy::volcopy_status(
            $self->{_dst}->Filer->conn,
            $self->{_op_num},
        );

        # Check for errors
        croak $@ if $@;

        if ( defined $ret ) {
            $status->{status} = 'transferring';
            $status->{'transfer-progress'} = int($ret*$self->{_src_used}/100);
        } else {
            $status->{status} = 'idle';
        }
    }

    return $status;
}

=head2 B<Waiting>

=head2 ::wait

Waits for all background processes to complete.
These are the processes created by using the B<background> option of
L<C<::start>|/::start>.

This method is completely unrelated to L<C<::status_wait>|/::status_wait>.

=over

=item Arguments

None.

=item Example

=back

     $vc->wait;

=cut

sub wait {
    my $self = shift;
    my ($obj, $test, $op);

    while ( $test = pop @{$self->{_subtests}} ) {
        $op  = $test->{operation};
        $obj = $test->{subtest};
        if ( $obj->is_alive() ) {
            NOTE::logcomment( "Waiting for: ".$op );
        }
        $obj->wait_finish();
        croak $@ unless FSR::FSRutils::subtestResults($obj);
    }

    return;
}

=head2 ::status_wait

Wait for the vol copy relationship to reach specific states.
Returns when all of the specified status elements
have their respective values I<at the same time>.
(Hint: don't use element/value combinations
that are mutually exclusive.)

The status elements checked are the same ones
that may appear in the hash generated by L<C<::status>|/::status>.
(Internally C<::status> is used; this is a convenience method.)

Because you are a valued NOTE user,
the values for the status elements may be given
as regular expressions.

This method is completely unrelated to L<C<::wait>|/::wait>.

=over

=item Arguments

I<status element name> => I<status element value> | I<regex>

I<Required.>
The status elements to check are given as named arguments.
Their values may be given as literal strings
or regular expressions that (hopefully) match valid values.

B<source> => C<0> | C<1>

I<Optional.>
Get the status of the relationship from the source filer instead.
The default is 0 (destination filer).

B<restore> => C<0> | C<1>

I<Optional.>
Get the status of the restore transfer.
The default is 0.

B<ignore_transfer_error> => C<0> | C<1>

I<Optional.>
If 1, don't fail if the status contains the C<current-transfer-error> element.
The default is 0.

B<sleep> => I<number>

I<Optional.>
Number of seconds to sleep in between polling the filer's status.
The default is 5 seconds.

B<timeout> => I<number>

I<Optional.>
Maximum time allowed for commands to complete, in seconds.
The default is object's default timeout.

=item Examples

=back

     $vc->status_wait(
         status => 'idle'
     );

     $vc->status_wait(
         status => 'transferring'
     );

=cut

sub status_wait {
    my ($self, %args) = @_;
    my (%opts, $wait_for, $matchall, $havekeys, $status, $etime, $delta);
    my $sleepis = 5;    # sleep number of seconds between status read
    my $noerror = 0;    # croak by default if 'current-transfer-error'

    $opts{source} = exists $args{source} ? $args{source} : 0;
    delete $args{source};
    $opts{restore} = exists $args{restore} ? $args{restore} : 0;
    delete $args{restore};

    if( exists $args{'timeout'} ) {
        $etime = $args{'timeout'};
        delete $args{'timeout'};
    } else {
        $etime = $self->{_timeout};
    }

    if( exists $args{'sleep'} ) {
        $sleepis = $args{'sleep'};
        delete $args{'sleep'};
    }

    if( exists $args{'ignore_transfer_error'} ) {
        $noerror++;
        delete $args{'ignore_transfer_error'};
    }

    $wait_for = ''; # Show what we are waiting for
    foreach( keys %args ) { $wait_for .= "$_ =~ /$args{$_}/\n" }

    $matchall = 0;
    $havekeys = scalar( keys %args );

    # Set expiration time
    $etime += time;

    while( $matchall < $havekeys ) {
        NOTE::logcomment("status_wait:\n$wait_for");

        croak "Status Timeout waiting for:\n$wait_for" if time > $etime;

        $status = $self->status(%opts);
        $matchall = 0;

        if( $status->{'current-transfer-error'} && $noerror < 1 ) {
            croak $status->{'current-transfer-error'};
        }

        # check keys one by one and match all of them
        foreach( keys %args ) {
            $matchall++ if( $status->{$_} =~ qr/$args{$_}/i );
        }

        # Do not wait anymore if everything has been found
        last unless ( $matchall < $havekeys );

        # Wait for next iteration
        $delta = $etime - time;
        $delta = 0 if $delta < 0;
        $sleepis = $delta if $delta < $sleepis;
        sleep $sleepis if( $sleepis > 0 );
    }

    return 1;
}

=head2 B<Component Objects>

=head2 ::src

Returns the Volume object representing the current source.

=over

=item Arguments

None.

=item Examples

=back

     $src = $vc->src;

=cut

sub src {
    my $self = shift;
    return $self->{_src};
}

=head2 ::dst

Returns the Volume object representing the current destination.

=over

=item Arguments

None.

=item Examples

=back

     $dst = $vc->dst;

=cut

sub dst {
    my $self = shift;
    return $self->{_dst};
}

=head2 B<Names and Paths>

=head2 ::src_name

Returns the full name of the source Volume: '/vol/<name>'.

=over

=item Arguments

None.

=item Example

=back

     $srcname = $vc->src_name;

=cut

sub src_name {
    my $self = shift;
    return $self->{_src}->mount_path;
}

=head2 ::dst_name

Returns the full name of the destination Volume: '/vol/<name>'.

=over

=item Arguments

None.

=item Example

=back

     $dstname = $vc->dst_name;

=cut

sub dst_name {
    my $self = shift;
    return $self->{_dst}->mount_path;
}

=head2 ::src_path

Returns the host-qualified path name of the source Volume.

=over

=item Arguments

None.

=item Example

=back

     $srcpath = $vc->src_path;

=cut

sub src_path {
    my $self = shift;
    return $self->{_src}->hostpath;
}

=head2 ::dst_path

Returns the host-qualified path name of the destination Volume.

=over

=item Arguments

None.

=item Example

=back

     $dstpath = $vc->dst_path;

=cut

sub dst_path {
    my $self = shift;
    return $self->{_dst}->hostpath;
}

=head2 B<Object Identification>

=head2 ::description

Returns a description of the vol copy relationship consisting
of the source volume description, a rightward pointing arrow,
and the destination volume description. See
L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>
for volume description format.

=over

=item Arguments

None.

=item Example

=back

     $desc = $vc->description;

=cut

sub description {
    my $self = shift;

    return "\n    +---" . $self->src->description .
           "\n    +-> " . $self->dst->description;
}

=head2 ::op_num

Returns the operation number for this transfer.

=over

=item Arguments

None.

=item Example

=back

     $op_num = $vc->op_num;

=cut

sub op_num {
    my $self = shift;
    return $self->{_op_num};
}

=head2 B<Object Configuration>

=head2 ::timeout

Sets or returns the current default timeout for the object before being set.
A value of -1 is to wait forever.

=over

=item Arguments

B<timeout> => I<number>

I<Optional.>
Set default timeout for the object.

=item Example

=back

     $curr_timeout = $vc->timeout;    # Get current default timeout

     $orig_timeout = $vc->timeout(
         timeout => 300               # Set default timeout
     );

=cut

sub timeout {
    my ($self, %args) = @_;

    $self->{_timeout} = $args{timeout} if exists $args{timeout};
    $self->{_timeout} = -1 if ! defined $self->{_timeout};
    return $self->{_timeout};
}

=head2 B<Object Destruction>

=head2 ::destroy

Destroy the vol copy relationship by aborting the active transfer.

=over

=item Arguments

None.

=item Example

=back

     $vc->destroy;

=cut

sub destroy {
    my $self = shift;
    $self->{_destroyed} = 1;

    NOTE::logcomment( "Destroying Volume Copy " . $self->description );

    # Abort any transfer currently active, ignore error
    eval {
        if ( defined $self->{_op_num} && $self->abort ) {
            $self->status_wait(
                status => 'idle',
                ignore_transfer_error => 1,
            );
        }
    };

    NOTE::logcomment("Volume Copy relationship destroyed");

    return;
}

# Internal function which takes as argument a Volume object or a volume
# host path string and returns the Volume object attached to the volume
sub _vol_path_attach {
    my $vol = shift;
    my ($vol_obj, $filer_name, $vol_name);

    if ( ref($vol) ) {
        if ( ! $vol->isa('NOTE::Volume') ) {
            croak "Argument must be a Volume object: <" . ref($vol) . ">";
        }
        # It is already a Volume object
        $vol_obj = $vol;
    } elsif ( $vol =~ /([^:]+):(?:\/vol\/)?(.*)/ ) {
        # This is a host path so create the Volume object
        ($filer_name, $vol_name) = ($1, $2);
        # Attach the volume
        $vol_obj = NOTE::Volume->attach(
            Filer => { host => $filer_name },
            name  => $vol_name,
        );
    } else {
        croak "Invalid volume path: $vol";
    }

    return $vol_obj;
}

# Object destructor
sub DESTROY {
    my $self = shift;

    # Only the creator can destroy the object
    return if ( $$ != $self->{_owner} );
    return if ( $self->{_spare_me} );    # Means don't destroy
    return if ( $self->{_destroyed} );   # Already dead
    return if ( NOTE->ALL_SPARED );

    # Destroy object
    $self->destroy;
}

=head1 SEE ALSO

L<C<Volume::Copy>|engineering::ontap_tests::agnostic::docs::lib-Volume-Copy-pm>

L<C<NOTE::Filer::VolCopy>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/volcopy>

L<C<NOTE>|engineering::ontap_tests::main::docs::lib-NOTE-pm/>

L<C<NOTE::Aggregate>|engineering::ontap_tests::main::docs::lib-NOTE-Aggregate-pm/>

L<C<NOTE::Filer>|engineering::ontap_tests::main::docs::lib-NOTE-Filer-pm/>

L<C<NOTE::Volume>|engineering::ontap_tests::main::docs::lib-NOTE-Volume-pm/>

=head1 STATUS

Review

$Id: //depot/prod/test/Rsierranevada/lib/NOTE/Volume/Copy.pm#1 $

$Date: 2013/01/22 $

=head1 AUTHOR

Author information is in the
L<NOTE overview|engineering::ontap_tests::main::docs::lib-NOTE-pm/::author>.

=cut

1;

