# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NX/Aggr.pm#4 $
#
# Copyright (c) 2012 NetApp
# All rights reserved.
#
## @summary
## Aggregate specified functions to support NACL Translator framework
##
## @status    private
##

package NX::Aggr;

use strict 'vars';
use Tharn;

use Data::Dumper;

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::StorageAggregate;
use NACL::STask::Aggregate;
use NX::Helper;
use NACL::STask::StorageAggregateRelocation;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
    create
    make_hybrid
    add_disks
    relocate
    scrub
);

my @Pkg_Workers;
sub stop_workers {
    $Log->comment("Detected interrupt signal");
    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}

=pod

=head1 NAME

NX::Aggr

=head1 DESCRIPTION

Aggregate related routines for use in NX framework;

=head1 METHODS

=head2 create

Create an aggregate on the filer.

The input is a hash (%opts) which must include the command_interface:
    command_interface => <array ref of nodes>

=head3 opts

%opts can be used to provide additional options for aggregate creation
    nx_context       => <NX context reference>
    base_name        => <base_name for aggregate>
    disktype         => <disk type>
    'block-type'     => <block type>
    hybrid_raidtype  => <raid type for SSD raidgroup if converting to hybrid>

%opts can also be used to override default values, which include
    offset           => obtained from NX context reference if not specified
    validate_only    => 0
    qty              => 1
    diskcount        => 3
    if_exists        => reuse
    raidtype         => raid_dp
    hybrid           => 0 (Whether or not to conver to hybrid)
    hybrid_diskcount => 3 (SSDs to use if converting to hybrid)
    chksumstyle      => 'block' (checksum can be block,zoned,advanced_zoned)*

    *chksumstyle will purge aggr if an existing one is of a different checksum type

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX create
Create
description
Create aggregate(s)
required
nodes;Node Name;;Node(s) to create aggr on
optional
base_name;Aggregate name;;If quantity is greater than 1, a number will be appended to the basename, otherwise the basename will be used.
disktype;Disk Type;;Disk types can be of FCAL|BSAS|SATA|ATA|FSAS|LUN|MSATA|SAS|SSD
block-type;Block Type;;Block type for aggregate. 64-bit or 32-bit
chksumstyle;Checksum;block;The default checksum type is 'block', other value can be 'zoned'
qty;Quantity;1;How many to create
diskcount;Number of Disks;5;The default diskcount is 5 for raid_dp
raidtype;Raid Type;raid_dp;The default raid type is raid_dp
if_exists;If Exists;reuse;The default action if it exists is to reuse, can be 'purge'
hybrid;Convert to Hybrid;0;Set to 1 to convert the new aggregate to hybrid
hybrid_raidtype;Hybrid Raid Type;;Raid Type for SSD configuration
hybrid_diskcount;SSD Count;3;Amount of SSDs to use for aggregate creation, default is 3
=end NX create
=cut

sub create {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 1;
    }

    my $my_return; #Hash reference containing return values
    my $exception;
    my $my_context;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF, optional => 0},
                nx_context        => {type => HASHREF,  optional => 1},
                base_name         => {type => SCALAR,   optional => 1},
                disktype          => {type => SCALAR,   optional => 1},
                'block-type'      => {type => SCALAR,   optional => 1},
                offset            => {type => SCALAR,   optional => 1},
                chksumstyle       => {type => SCALAR,   optional => 1, default => 'block'},
                validate_only     => {type => SCALAR,   optional => 1, default => 0},
                qty               => {type => SCALAR,   optional => 1, default => 1},
                diskcount         => {type => SCALAR,   optional => 1, default => 3},
                if_exists         => {type => SCALAR,   optional => 1, default => 'reuse'},
                raidtype          => {type => SCALAR,   optional => 1, default => 'raid_dp'},
                # Hybrid Aggr options
                hybrid            => {type => SCALAR,   optional => 1, default => 0},
                hybrid_raidtype   => {type => SCALAR,   optional => 1},
                hybrid_diskcount  => {type => SCALAR,   optional => 1, default => '3'}
            }
        );


        foreach my $key (keys %input){
            print "key: |$key|\nvalue: |$input{$key}|\n";
            if( $key eq "if_exists"){
                if ($input{$key} eq "reuse" ||
                    $input{$key} eq "purge"){
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: reuse,purge ");
            } elsif( $key eq "chksumstyle"){
                if ($input{$key} ne "block" &&
                    $input{$key} ne "zoned" &&
                    $input{$key} ne "advanced_zoned") {
                    NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: zoned, block, advanced_zoned");
                }
            } elsif( $key eq "disktype"){
                if($input{$key} eq "FCAL"  ||
                   $input{$key} eq "BSAS"  ||
                   $input{$key} eq "SATA"  ||
                   $input{$key} eq "ATA"   ||
                   $input{$key} eq "FSAS"  ||
                   $input{$key} eq "LUN"   ||
                   $input{$key} eq "MSATA" ||
                   $input{$key} eq "SAS"   ||
                   $input{$key} eq "SSD"){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: FCAL,BSAS,SATA,ATA,FSAS,LUN,MSATA,SAS,SSD");
            } elsif( $key eq "raidtype"){
                if($input{$key} eq "raid4" ||
                   $input{$key} eq "raid_dp"){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: raid4,raid_dp ");
            } elsif( $key eq "diskcount"){
                if($input{$key} < 2) {
                    NATE::BaseException->throw("Aggr create requires more 2 or more disks.");
                }
                if($input{'raidtype'} eq "raid4"){
                    next;
                } elsif($input{'raidtype'} eq "raid_dp"){
                    if($input{$key} < 3) {
                        NATE::BaseException->throw("Raid_dp aggr create requires 3 or more disks.");
                    }
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) aggr create option <$key>");
            } elsif( $key eq "qty"){
                if($input{$key} >= 1){
                   next;
                }
                NATE::BaseException->throw("Quantity for aggr create must be 1 or greater");
            } elsif( $key eq "block-type"){
                if($input{$key} eq '32-bit' ||
                   $input{$key} eq '64-bit'){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: 32-bit,64-bit ");
            } elsif( $key eq "hybrid"){
                if($input{$key} == 0 ||
                   $input{$key} == 1){
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: 0,1 ");
            } elsif( $key eq "hybrid_raidtype"){
                if($input{$key} eq "raid4" ||
                   $input{$key} eq "raid_dp"){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. Accepted values: raid4,raid_dp ");
            } elsif( $key eq "hybrid_diskcount"){
                if($input{$key} < 2) {
                    NATE::BaseException->throw("SSD diskcount for hybrid aggregate must be 2 or greater");
                }
                if($input{'hybrid_raidtype'} eq "raid4" || !defined $input{'hybrid_raidtype'}){
                    next;
                } elsif($input{'hybrid_raidtype'} eq "raid_dp"){
                    if($input{$key} < 3) {
                        NATE::BaseException->throw("SSD diskcount for hybrid aggregate with raid_dp must be 3 or greater ");
                    }
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr create option <$key>. ");
            }
        }
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Aggregate create action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Aggr::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            # create aggregate
            my %aggr_opts = %input;
            $aggr_opts{command_interface} = [ $n ];

            my $id = 'AC_N_DaC.'.$proc_cnt;
            push(@ids, $id);

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Aggr',
                method  => 'create',
                args    => [ %aggr_opts ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        #Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
            context         => $input{nx_context},
            merge_context   => 1,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $node_ref = $res->{$rid}->{objects}->{node};
                foreach my $node_name (keys %{$node_ref} ) {
                    my @res_obj = @{$node_ref->{$node_name}};
                    push( @{$obj_ref->{node}->{$node_name}}, @res_obj );
                }
            }

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Created aggregates each on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
            $my_return->{context} = $input{nx_context};
        }
        return $my_return;
    } else { #Single node
        my $num_aggrs    = delete $input{qty};
        my $context      = delete $input{nx_context};
        my $cur_offset   = delete $input{offset};
        my $make_hya     = delete $input{hybrid};
        my $hya_disknum  = delete $input{hybrid_diskcount};
        my $hya_raidtype = delete $input{hybrid_raidtype};
        delete $input{validate_only};

        my $n         = $nodes[0];
        my $node_name = $n->name();
        my $node_name_xlated = $node_name;
        $node_name_xlated =~ s/\-//g;

        # Get base_name established
        my $base_name;
        my $base_name_defined = 0;
        if(!defined($input{base_name})) {
            $base_name = "nx_".$node_name_xlated."_a";
        } else {
            $base_name = delete $input{base_name};
            $base_name_defined = 1;
        }

        # Get current offset or set new one to 1
        if ( !defined($cur_offset) ) {
            $cur_offset = $context->{nodes}->{$node_name}->{aggregate}->{$base_name}->{offset};
            $cur_offset = 1 if !defined($cur_offset);
        }

        my $obj_ref;
        try {
            $input{command_interface} = $n;
            for( 1..$num_aggrs ) {
                if($num_aggrs > 1) {
                    $input{aggregate} = $base_name."$cur_offset";
                } else {
                    if($base_name_defined){
                        $input{aggregate} = $base_name;
                    } else {
                        $input{aggregate} = $base_name."$cur_offset";
                    }
                }
                $cur_offset++;

                my $purged = 0;
                my $all_valid_disk_names;

                #Purge aggr if exists and is not of desired checksum type
                my $api_set = $n->get_7m_or_nodescope_apiset();
                my $chksum = _if_exists_return_checksum(
                        api  => $api_set,
                        aggr => $input{aggregate},
                    );
 
                if($chksum && ($chksum ne $input{chksumstyle})) {
                    $Log->warn("Aggr $input{aggregate} exists, but checksum is different, will purge.");
                    NACL::STask::Aggregate->purge(
                        command_interface => $input{command_interface},
                        aggregate => $input{aggregate},
                        );
                    $purged = 1;
                }

                if(!$chksum || $purged) {  #If aggr needs to be created or will be purged
                    #Getting disk spares
                    $Log->comment("Checking for enough spares");
                    my %disks;
                    my $cs = $input{chksumstyle};

                    my $spares = _get_spares(
                        node => $input{command_interface},
                        checksum_type => $cs,
                        );
                    @{$disks{z}} = ();
                    @{$disks{nz}} = ();
                    @{$disks{zeroing}} = ();
                    @{$disks{ssd}} = ();
                    # check for disks that are ready and zeroed, store SSDs elsewhere
                    foreach(@{$spares}) {
                        if($_->{disk_type} !~ /SSD/i) {
                            if($_->{fsm_state} =~ /^zeroed|fakezeroed/i){
                                push(@{$disks{z}},[$_->{name},$_->{checksum_type}]);
                                push(@$all_valid_disk_names, $_->{name});
                            } else {
                                push(@{$disks{nz}},[$_->{name},$_->{checksum_type}]);
                            }
                        } else {
                            push(@{$disks{ssd}},[$_->{name},$_->{checksum_type}]);
                        }
                    }

                    my $d_total = 0;

                    if(@{$disks{z}} + @{$disks{nz}} < $input{diskcount}) {
                        $Log->comment("Not enough disks, searching for more applicable disks.");
                        my $disk_obj = _get_spares( node => $input{command_interface} );
                        foreach(@{$disk_obj}) {
                            #Do not double count the disks
                            next if($input{chksumstyle} =~ /zoned/i && $_->{'disk_type'} !~ /BSAS|MSATA|ATA|SATA/i);
                            next if($input{chksumstyle} =~ /zoned/i && $_->{checksum_type} eq 'advanced_zoned');
                            next if($input{chksumstyle} !~ /zoned/i && $_->{checksum_type} ne 'advanced_zoned');
                            push(@{$disks{nz}},[$_->{name},$_->{checksum_type}]);
                        }
                        $d_total = @{$disks{z}} + @{$disks{nz}};
                        if($d_total < $input{diskcount}) {  # Still not enough
                            NATE::BaseException->throw("Need $input{diskcount} disks, only $d_total 
                                                       applicable disks for $input{aggregate} creation!");
                        }
                    }

                    ## Check if enough SSDs for hybrid requirement
                    if($make_hya) {
                        if(@{$disks{ssd}} < $hya_disknum) {
                            NATE::BaseException->throw("Not enough SSDs available for $input{aggregate} creation!");
                        }
                    }

                    ## Start zeroing the disks
                    for(my $i=0;$i<($input{diskcount}-@{$disks{z}});$i++) {
                        my $d = shift @{$disks{nz}};
                        if($d) {
                            my $cs = 'zoned';
                            if($input{chksumstyle} eq 'block') {
                                $cs = 'block';
                            }
                            my %c_mode_owner;
                            if($input{command_interface}->mode() =~ /cmode/i) {
                                $c_mode_owner{owner} = $node_name;
                            }
                            # $d is an arrayref of (name,checksumtype)
                            if($d->[1] !~ $cs) {
                                $Log->comment("Disk($d->[0]) is of different checksum ($d->[1] !~ $cs)");
                                NACL::C::StorageDisk->assign(
                                        command_interface => $input{command_interface},
                                        disk => $d->[0],
                                        checksum => $cs,
                                        %c_mode_owner,
                                    );
                            } else {
                                $Log->comment("Disk($d->[0]) is of same checksum($d->[1]), skipping assignment.");
                            }
                            #doesn't need to push checksum type, doesn't matter anymore
                            push(@{$disks{zeroing}},$d->[0]);
                            push(@$all_valid_disk_names, $d->[0]);
                        }
                    }
                    sleep 10;

                    # See if filer is fake zeroed, if so, don't check zeroing status,
                    # as it may be false info
                    my @cur_aggrs = NACL::STask::Aggregate->find(
                        command_interface => $input{command_interface}
                    );
                    my $rs = $cur_aggrs[0]->state()->raidstatus();
                    $Log->comment("Zeroing spares");
                    NACL::C::StorageDisk->zerospares( command_interface => $input{command_interface} );
                    $Log->comment("RAID status of aggr ".$cur_aggrs[0]->aggregate().": *$rs*");
                    if ($rs =~ /uninit/i) {
                        $Log->comment("Disks are fake-zeroed");
                        sleep 30;
                        # Don't bother waiting for zeroing status
                    } elsif (@{$disks{zeroing}}) {
                        $Log->comment("Disks are real-zeroed");
                        sleep 5;
                        my $elapsed = 0;
                        my $slp = 60;
                        while(1) {
                            last if(!@{$disks{zeroing}});
                            if($elapsed == 180) {
                                # Issue another disk zero just in case
                                NACL::C::StorageDisk->zerospares( command_interface => $input{command_interface} );
                            }
                            my $d = shift @{$disks{zeroing}};
                            my $obj = _get_spares( node => $input{command_interface}, disk_name => $d );
                            my $state = $obj->[0]->{fsm_state};
                            $Log->comment("Current State: $state");
                            if($state =~ /zeroing|nonzeroed|notzeroed/i){
                                $Log->comment("State is zeroing or needs to be zeroed");
                                push(@{$disks{zeroing}},$d);
                                sleep $slp;
                                $elapsed += $slp;
                            } elsif($state =~ /^zeroed|fakezeroed/i) {
                                $Log->comment("State matched zeroed or fakezeroed");
                            } else {
                                NATE::BaseException->throw("Unexpected state while zeroing: $state");
                            }
                            if($elapsed > 3600*24){ ##Time out of 24 hours.
                                NATE::BaseException->throw("Diskzeroing has timed out!");
                            }
                        }
                    }
                }

                my $aggr_obj;
                if(!$chksum || $purged) {
                    my $master_disk_list;
                    for (1..$input{diskcount}) {
                        push(@$master_disk_list, shift(@$all_valid_disk_names));
                    }
                    foreach (@$master_disk_list) {
                        $Log->comment("Using disk $_ to create aggr $input{aggregate}");
                    }
                    $input{disklist} = $master_disk_list;
    
                    # Pass a modified copy of %input, as we can't
                    # delete values, since we need them for creating the
                    # other aggrs
                    my %aggr_create_input = %input;
                    delete $aggr_create_input{diskcount};
                    delete $aggr_create_input{chksumstyle};
                    $aggr_obj = NACL::STask::Aggregate->create(%aggr_create_input);
                } else {
                    $aggr_obj = NACL::STask::Aggregate->create(%input);
                }

                # Convert to hybrid if options are set and aggr is not already hybrid
                if($make_hya && !$aggr_obj->is_hybrid()) {
                    $Log->comment("Converting ".$aggr_obj->aggregate()." to a hybrid aggregate");
                    if(defined($hya_raidtype) && ($hya_raidtype ne $input{raidtype})){
                        $aggr_obj->convert_to_hybrid(
                            nacltask_add_disks => {
                                raidtype  => $hya_raidtype,
                                diskcount => $hya_disknum,
                                disktype  => 'SSD',
                            }
                        );
                    } else {
                        $aggr_obj->convert_to_hybrid(
                            nacltask_add_disks => {
                                diskcount => $hya_disknum,
                                disktype  => 'SSD',
                            }
                        );
                    }
                }
                push(@{$obj_ref->{node}->{$node_name}}, $aggr_obj);
            }
            #Store context offsets per aggr per node
            $my_context->{nodes}->{$node_name}->{aggregate}->{$base_name}->{offset} = $cur_offset;
        } otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Created aggregates each on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
            $my_return->{context} = $my_context;
        }
        return $my_return;
    }
} ## End create

=pod

=head2 make_hybrid

Make an aggregate into a hybrid aggregate/flash pool.

The input is a hash (%opts) which must include the command_interface and
aggregates to change:
    command_interface => <array ref of nodes>
    aggregates        => <array ref of aggregates>

=head3 opts

%opts can be used to provide additional options for aggregate creation
    raidtype   => <raid type of new raid group of SSDs>

%opts can also be used to override default values, which include
    diskcount  => 3

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX make_hybrid
Make Hybrid
description
Make an aggregate into a hybrid aggregate/flash pool
required
nodes;Node(s);;Node(s) to run current action on. Specified as $N<#> e.g. $N1.
aggregates;Aggregate(s);;Aggregate(s) to run current action on. Specified as $N<#>/$A<#> e.g. $N1/$A1 will be on the first aggregate(created or used in this scenario) of the first node.
optional
diskcount;Diskcount;3;Amount of SSDs to use, default is 3
raidtype;Raidtype;;Raid type for the SSDs
=end NX make_hybrid
=cut

sub make_hybrid {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return; #Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF, optional => 0},
                aggregates        => {type => ARRAYREF, optional => 0},
                validate_only     => {type => SCALAR,   optional => 1, default => 0},
                diskcount         => {type => SCALAR,   optional => 1, default => 3},
                raidtype          => {type => SCALAR,   optional => 1},
            }
        );


        foreach my $key (keys %input){
            print "key: |$key|\nvalue: |$input{$key}|\n";
            if( $key eq "raidtype"){
                if($input{$key} eq "raid4" ||
                   $input{$key} eq "raid_dp"){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr make_hybrid option <$key>. Accepted Values: raid4,raid_dp");
            } elsif( $key eq "diskcount"){
                if($input{'raidtype'} eq "raid4" || !defined $input{'raidtype'}){
                    if($input{$key} < 2) {
                        NATE::BaseException->throw("Invalid input($input{$key}) for aggr make_hybrid option <$key>. Diskcount must 2 or greater. ");
                    }
                    next;
                } elsif($input{'raidtype'} eq "raid_dp"){
                    if($input{$key} < 3) {
                        NATE::BaseException->throw("Invalid input($input{$key}) for aggr make_hybrid option <$key>. Raid_dp diskcount must 3 or greater. ");
                    }
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr make_hybrid option <$key>");
            }
        }
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Aggregate convert to hybrid action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Aggr::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;
    my $node_to_aggrs;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my @aggrs = @{$input{aggregates}};
        foreach my $a ( @aggrs ) {
            my $node_name = $a->command_interface()->name();
            push(@{$node_to_aggrs->{$node_name}}, $a);
            $Log->comment("Associating aggr ".$a->aggregate()." to node $node_name\n");
        }
        foreach my $n (@nodes) {
            if ( !$node_to_aggrs->{$n->name()} ) {
                $Log->comment("No aggregates associated with node ".$n->name()
                    . ", skipping the node");
                next ;
            }
            my $id = 'HyA_C_DaC.'.$proc_cnt; # HybridAggr Conversion Divide and Conquer
            push(@ids, $id);

            $input{command_interface} = [ $n ];
            $input{aggregates} = $node_to_aggrs->{$n->name()};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Aggr',
                method  => 'make_hybrid',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }
        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Converted aggregates to hybrid each on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { # Single node
        my $hya_raidtype = delete $input{raidtype};
        my $hya_disknum  = delete $input{diskcount};
        delete $input{validate_only};
        delete $input{command_interface};

        my $node_name = $nodes[0]->name();
        my $num_aggrs;
        try {
            my @aggrs  = @{$input{aggregates}};
            $num_aggrs = @aggrs;
            delete $input{aggregates};
            foreach my $aggr_obj ( @aggrs ) {
                if($hya_raidtype ne $input{raidtype}){
                    $aggr_obj->convert_to_hybrid(
                        nacltask_add_disks => {
                            raidtype  => $hya_raidtype,
                            diskcount => $hya_disknum,
                            disktype  => 'SSD',
                        }
                    );
                } else {
                    $aggr_obj->convert_to_hybrid(
                        nacltask_add_disks => {
                            diskcount => $hya_disknum,
                            disktype  => 'SSD',
                        }
                    );
                }
            }
        } otherwise {
            $exception = shift;
        };
        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Converted $num_aggrs to hybrid on $node_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} ## End make_hybrid

=pod

=head2 add_disks

Add disks to existing aggregates

The input is a hash (%opts) which must include the command_interface and
aggregates to change:
    command_interface => <array ref of nodes>
    aggregates        => <array ref of aggregates>
    diskcount         => <amount of disks to add>

=head3 opts

%opts can be used to provide additional options for aggregate creation
    raidgroup       => <raidgroup to add disks to>
    disktype        => <Type of disk to add>
    raidtype        => <raid type of new raid group for disks added>
                        (Not supported for 7mode at the moment, burt625792)
    'allow-mixed'   => (Cmode) <if mixed raid types are allowed>
    '64bit_upgrade' => <upgrade to 64bit if 32bit>

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX add_disks
Add Disks
description
Add disks to existing aggregates.
required
nodes;Node(s);;Node(s) to run current action on. Specified as $N<#> e.g. $N1.
aggregates;Aggregate(s);;Aggregate(s) to run current action on. Specified as $N<#>/$A<#> e.g. $N1/$A1 will be on the first aggregate(created or used in this scenario) of the first node.
diskcount;Diskcount;;Integer value of amount of disks to add
optional
raidgroup;Raidgroup;;Raidgroup to add disks to
disktype;Disktype;;Type of disk to add (depends on what is available in your system)
raidtype;Raidtype;;Raid type of new raid group for disks added
allow-mixed;Allow Mixed;;Only available in cmode, specify 1 if allowing mixed raidtypes
64bit_upgrade;64bit Upgrade;;Specify 1 if upgrading from 32bit to 64bit aggregate
=end NX add_disks
=cut

sub add_disks {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return; #Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF, optional => 0},
                aggregates        => {type => ARRAYREF, optional => 0},
                diskcount         => {type => SCALAR,   optional => 0},
                validate_only     => {type => SCALAR,   optional => 1, default => 0},
                raidgroup         => {type => SCALAR,   optional => 1},
                'allow-mixed'     => {type => SCALAR,   optional => 1},
                '64bit-upgrade'   => {type => SCALAR,   optional => 1},
                disktype          => {type => SCALAR,   optional => 1},
                raidtype          => {type => SCALAR,   optional => 1},

            }
        );


        foreach my $key (keys %input){
            print "key: |$key|\nvalue: |$input{$key}|\n";
            if( $key eq "raidtype"){
                if($input{$key} eq "raid4" ||
                   $input{$key} eq "raid_dp"){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Accepted values: raid4,raid_dp ");
            } elsif( $key eq "diskcount"){
                if($input{$key} < 1) {
                    NATE::BaseException->throw("At least 1 disk required for aggr add_disks.")
                }
                if($input{'raidtype'} eq "raid4"){
                    if($input{$key} < 2) {
                        NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Raid4 requires 2 or more disks.");
                    }
                    next;
                } elsif($input{'raidtype'} eq "raid_dp"){
                    if($input{$key} < 3) {
                        NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Raid_dp requires 3 or more disks.");
                    }
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key> ");
            } elsif( $key eq "raidgroup"){
                if($input{$key} eq "new" || $input{$key} eq "all" || $input{$key} =~ /^rg\d+$/) {
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Accepted values: new,all,existing raidgroups (e.g. rg0)");
            } elsif( $key eq "disktype"){
                if($input{$key} eq "FCAL"  ||
                   $input{$key} eq "BSAS"  ||
                   $input{$key} eq "SATA"  ||
                   $input{$key} eq "ATA"   ||
                   $input{$key} eq "FSAS"  ||
                   $input{$key} eq "LUN"   ||
                   $input{$key} eq "MSATA" ||
                   $input{$key} eq "SAS"   ||
                   $input{$key} eq "SSD"){
                   next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Accepted values: FCAL,BSAS,SATA,ATA,FSAS,LUN,MSATA,SAS,SSD");
            } elsif( $key eq 'allow-mixed'){
                if($input{$key} eq "true") {
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Accepted values: true");
            } elsif( $key eq '64bit-upgrade'){
                if($input{$key} eq "check"    ||
                   $input{$key} eq "grow-all" ||
                   $input{$key} eq "normal"){
                    next;
                }
                NATE::BaseException->throw("Invalid input($input{$key}) for aggr add_disks option <$key>. Accepted values: check,normal ");
            }
        }
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Aggregate add_disks action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Aggr::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;
    my $node_to_aggrs;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my @aggrs = @{$input{aggregates}};
        foreach my $a ( @aggrs ) {
            my $node_name = $a->command_interface()->name();
            push(@{$node_to_aggrs->{$node_name}}, $a);
            $Log->comment("Associating aggr ".$a->aggregate()." to node $node_name");
        }
        foreach my $n (@nodes) {
            if ( !$node_to_aggrs->{$n->name()} ) {
                $Log->comment("No aggregates associated with node ".$n->name()
                    . ", skipping the node");
                next ;
            }
            my $id = 'A_D_DaC.'.$proc_cnt; # Add Disk Divide and Conquer
            push(@ids, $id);

            $input{command_interface} = [ $n ];
            $input{aggregates} = $node_to_aggrs->{$n->name()};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Aggr',
                method  => 'add_disks',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }
        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Added disks each on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { # Single node
        my $node_name = $nodes[0]->name();
        my $num_aggrs;
        delete $input{validate_only};
        try {
            my @aggrs  = @{delete $input{aggregates}};
            $num_aggrs = @aggrs;
            foreach my $aggr_obj ( @aggrs ) {
                my %add_opts = %input;
                $add_opts{command_interface} = $nodes[0];
                $add_opts{aggregate} = $aggr_obj->aggregate();
                $Log->comment("ABC: Dumper(%add_opts)\n");
                NACL::C::StorageAggregate->add_disks( %add_opts );
            }
        } otherwise {
            $exception = shift;
        };
        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Added disks to $num_aggrs aggrs on $node_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} ## End add_disks


=pod

=head2 relocate

Relocates specified aggregates from one node to another.

The input is a hash (%opts) which must include:
    dst_nodes  => [ $node ]  # Array ref of 1 node object
    aggregates => [ $aggr ]  # Array ref of 1 or more aggregate objects
                             # to relocate

=head3 opts

%opts can also be used to override default values, which include
    iterations      => <integer>|infinite (default is 1)
    interval        => odd/even/all/<value in seconds>
                       Every odd, even, or every clock hour,
                       or specified value in seconds
    start_minute    => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
                       a two-digit value representing the clock minute
    do_flip         => 0|1 (default is 1, to relocate aggrs back to
                       their original node per iteration)
    flip_delay      => <value in seconds> (time to wait before flipping)
    validate_only   => default 0

This method returns a hash reference containing a list of volume objects

=over

=back

=cut

=cut begin NX relocate
Relocate
description
Relocates specified aggregates from one node to another.
required
dst_nodes;Dst Node(s);;Destination Node(s) to run current action on. Specified as $N<#> e.g. $N1.
aggregates;Aggregate(s);;Aggregate(s) to run current action on. Specified as $N<#>/$A<#> e.g. $N1/$A1 will be on the first aggregate(created or used in this scenario) of the first node.
optional
iterations;Iterations;1;Integer value of iterations to run or 'infinite', default is 1
interval;Interval;;Acceptable values odd|even|all or interger value(in seconds)
start_minute;Start Minute;;Required if "interval" is odd|even|all, otherwise it is ignored.
do_flip;Do Flip;1;Acceptable values are 0|1. 1 designates to relocate aggrs to original node per iteration, default is 1
flip_delay;Flip Delay;60;Integer(in seconds) value of delay before flipping, default is 60
=end NX relocate
=cut

sub relocate {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                dst_nodes            => { type => ARRAYREF, optional => 0 },
                aggregates           => { type => ARRAYREF, optional => 0 },
                iterations           => { type => SCALAR,   optional => 1, default => 1 },
                interval             => { type => SCALAR,   optional => 1 },
                start_minute         => { type => SCALAR,   optional => 1 },
                do_flip              => { type => SCALAR,   optional => 1, default => 1 },
                flip_delay           => { type => SCALAR,   optional => 1, default => 60 },
                validate_only        => { type => SCALAR,   optional => 1, default => 0 },
            }
        );

        foreach my $arg (keys %input ) {
            my $val = $input{$arg};
            if ( $arg eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer or \'infinite\'");
                }
            } elsif ( ($arg eq "iterations") && ($val > 1) && !defined($input{interval}) ) {
                NATE::BaseException->throw("\'repeat\' parameter requires a value "
                        . "for the \'interval\' parameter");
            } elsif ( $arg eq "interval" ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            } elsif ( $arg eq 'dst_nodes' ) {
                if ( @{$input{dst_nodes}} > 1 ) {
                    NATE::BaseException->throw("Only one node is allowed "
                        . "for the \'dst_nodes\' parameter");
                }
            } elsif ( $arg eq 'flip_delay' ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer in seconds");
                }
            } elsif ( $arg eq 'do_flip' ) {
                if ( $val !~ /^(0|1)$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be \'0\' or \'1\'");
                }
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Aggregate relocation action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Aggr::stop_workers";
    }

    my @nodes = @{$input{dst_nodes}};
    my $dst_node = $nodes[0];
    my $dst_node_name = $dst_node->name();
    my @aggrs = @{$input{aggregates}};
    my $num_aggrs = @aggrs;

    # Ensure all aggregates belong to the same node and is different
    # from the specified destination node
    my $src_node = $aggrs[0]->command_interface();
    my $src_node_name = $src_node->name();
    if ($src_node_name eq $dst_node_name) {
        NATE::BaseException->throw("The node that all input aggregates "
            . "belong to initially, must be different than the specified "
            . "destination node");
    }

    my @found_aggrs = NACL::STask::Aggregate->find(
        command_interface => $src_node,
        filter            => { 'owner-name' => $src_node_name },
    );
    foreach my $a (@aggrs) {
        my $aggr_name = $a->aggregate();
        if (!grep{ $_->aggregate() eq $aggr_name } @found_aggrs) {
            NATE::BaseException->throw("All specified aggregates for ARL "
                . "must be from the same node");
        }
    }

    # Store start tick for return later
    my $start_tick = time();
    $my_return->{start_tick} = $start_tick;

    my $done = 0;
    my $iter = 1;
    my $cur_src_node = $src_node;
    my $cur_dst_node = $dst_node;
    my $arl_timeout = 720;
    my $num_arls = 1;
    $num_arls = 2 if $input{do_flip};
    while (!$done) {
        try {
            $Log->comment("Starting relocation iteration $iter");

            for (1..$num_arls) {
                my $flip_string;
                $flip_string = ' (flipping)' if ($_ == 2);
                # Get list of aggrs on current src node
                my $cur_dst_node_name = $cur_dst_node->name();
                my $cur_src_node_name = $cur_src_node->name();
                my @aggr_list;
                my @aggr_names;
                my @tmp = NACL::STask::Aggregate->find(
                    command_interface => $cur_src_node,
                    filter            => { 'owner-name' => $cur_src_node->name() },
                );

                foreach my $ta (@tmp) {
                    my $aggr_name = $ta->aggregate();
                    if (grep{$aggr_name eq $_->aggregate()} @aggrs) {
                        push(@aggr_list, $ta);
                        push(@aggr_names, $aggr_name);
                    }
                }

                if ($aggr_list[0]) {
                    $Log->comment("Relocating".$flip_string." following "
                        . "aggregates from $cur_src_node_name to "
                        . "$cur_dst_node_name: @aggr_names");

                    NACL::STask::StorageAggregateRelocation->start(
                          command_interface => $cur_src_node,
                          node              => $cur_src_node_name,
                          destination       => $cur_dst_node_name,
                          'aggregate-list'  => \@aggr_list,
                          'method-timeout'  => $arl_timeout,
                          nacltask_wait     => 1,
                    );
                    $Log->comment("Success at relocation".$flip_string." of "
                        . "following aggregates from $cur_src_node_name to "
                        . "$cur_dst_node_name: @aggr_names");
                } else {
                    $Log->comment("No aggregates to relocate or flip from node "
                        . "$cur_src_node_name");
                }

                # Flip src/dst nodes
                if ( $cur_src_node_name eq $src_node_name ) {
                    $cur_src_node = $dst_node;
                } else {
                    $cur_src_node = $src_node;
                }

                if ( $cur_dst_node_name eq $dst_node_name ) {
                    $cur_dst_node = $src_node;
                } else {
                    $cur_dst_node = $dst_node;
                }

                sleep $input{flip_delay} if $input{do_flip};
            }

            if ( ($input{iterations} eq 'infinite') || ($iter < $input{iterations}) ) {
                wait_for_interval(
                    interval     => $input{interval},
                    start_minute => $input{start_minute},
                    start_tick   => $start_tick,
                );
                $iter++;
            } else {
                $done = 1;
            }
        } otherwise {
            $exception = shift;
            if ($::IGNORE_ERRORS) {
                $Log->warn("Got following exception, ignoring and "
                    . "continuing: ".$exception->text());
                if ( ($input{iterations} eq 'infinite') || ($iter < $input{iterations}) ) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $start_tick,
                    );
                    $iter++;
                } else {
                    $done = 1;
                }
            } else {
                $done = 1;
            }
        };
    }

    $my_return->{end_tick} = time();
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Relocated $num_aggrs aggregates on $dst_node_name "
            . "for $iter iterations";
        $Log->comment( $msg );
        $my_return->{msg} = $msg;
    }
    return $my_return;

} # End relocate


=pod

=head2 scrub

Calls scrub command against specified aggregates.

The input is a hash (%opts) which must include:
    aggregates => [ $aggr ]  # Array ref of 1 or more aggregate objects
                             # to scrub

=head3 opts

%opts can also be used to override default values, which include
    iterations      => <integer>|infinite (default is 1)
    interval        => odd/even/all/<value in seconds>
                       Every odd, even, or every clock hour,
                       or specified value in seconds
    start_minute    => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
                       a two-digit value representing the clock minute
    validate_only   => default 0

=over

=back

=cut

=cut begin NX scrub
Scrub
description
Calls scrub command against specified aggregates.
required
aggregates;Aggregate(s);;Aggregate(s) to run current action on. Specified as $N<#>/$A<#> e.g. $N1/$A1 will be on the first aggregate(created or used in this scenario) of the first node.
optional
iterations;Iterations;1;Integer value of iterations to run or 'infinite', default is 1
interval;Interval;;Acceptable values odd|even|all or interger value(in seconds)
start_minute;Start Minute;;Required if "interval" is odd|even|all, otherwise it is ignored.
=end NX scrub
=cut

sub scrub {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                aggregates           => { type => ARRAYREF, optional => 0 },
                iterations           => { type => SCALAR,   optional => 1, default => 1 },
                interval             => { type => SCALAR,   optional => 1 },
                start_minute         => { type => SCALAR,   optional => 1 },
                validate_only        => { type => SCALAR,   optional => 1, default => 0 },
                _extracted_nodes     => { type => SCALAR,   optional => 1, default => 0 },
            }
        );

        foreach my $arg (keys %input ) {
            my $val = $input{$arg};
            if ( $arg eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer or \'infinite\'");
                }
            } elsif ( ($arg eq "iterations") && ($val > 1) && !defined($input{interval}) ) {
                NATE::BaseException->throw("\'repeat\' parameter requires a value "
                        . "for the \'interval\' parameter");
            } elsif ( $arg eq "interval" ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Aggregate scrub action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Aggr::stop_workers";
    }

    my @aggrs = @{$input{aggregates}};
    my $num_aggrs = @aggrs;
    my $num_nodes = 0;
    my $node_to_aggrs;
    my @nodes;
    if ( !$input{_extracted_nodes} ) {
        # Initial NX call.  Extract number of unique nodes
        my $node_hash;
        foreach my $a ( @aggrs ) {
            my $node_obj = $a->command_interface();
            my $node_name = $node_obj->node();
            if ( !defined($node_hash->{$node_name}) ) {
                $num_nodes++;
                $node_hash->{$node_name} = 1;
                $Log->comment("Associating aggr ".$a->aggregate()
                    ." to node $node_name");
                push(@nodes, $node_obj);
            }
            push(@{$node_to_aggrs->{$node_name}}, $a);
        }
    }


    # Store start tick for return later
    my $start_tick = time();
    $my_return->{start_tick} = $start_tick;

    if ( !$input{_extracted_nodes} && ( $num_nodes > 1 ) ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n ( @nodes) {
            my $id = 'AScrub_N_DaC.'.$proc_cnt; # (N)ode_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            my $node_name = $n->node();
            $input{_extracted_nodes} = 1;
            $input{aggregates} = $node_to_aggrs->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Aggr',
                method  => 'scrub',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Success at aggr scrub on $num_nodes nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { # Single node, one or more aggrs
        my $node = $aggrs[0]->command_interface();
        my $api  = $node->get_7m_or_nodescope_apiset();

        my $done = 0;
        my $iter = 1;
        while (!$done) {
            try {
                $Log->comment("Starting scrub iteration $iter");

                foreach my $aggr (@aggrs) {
                    try {
                        $aggr->scrub( action => 'start' );
                    } otherwise {
                        my $e = shift;
                        if ($e->text() !~ /already been started/i) {
                            $e->throw();
                        } else {
                            $Log->comment("Scrub already started on "
                                . $aggr->aggregate());
                        }
                    };
                }

                my $active;
                my $num_scrubbed = 0;
                while ($num_scrubbed != $num_aggrs) {
                    sleep 60;
                    my $output;
                    foreach my $aggr (@aggrs) {
                        my $aggr_name = $aggr->aggregate();
                        my $resp = $api->aggr_scrub_status(
                            aggregate => $aggr_name,
                            verbose   => 1,
                        );
                        my $parsed = $resp->get_parsed_output();
                        $output->{$aggr_name} = $parsed;
                    }

                    foreach my $aggr_name (keys(%$output)) {
                        my $parsed = $output->{$aggr_name};
                        my $cur_status = $parsed->[0]->{current_scrub};
                        if ( $active->{$aggr_name} &&
                             $cur_status =~ /not active/i ) {
                            $Log->comment("Aggregate $aggr_name scrubbed");
                            $num_scrubbed++;
                        } elsif ($cur_status =~ /\d+\%/) {
                            $active->{$aggr_name} = 1;
                        }
                    }
                }

                if (   ($input{iterations} eq 'infinite') ||
                       ($iter < $input{iterations}) ) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $start_tick,
                    );
                    $iter++;
                } else {
                    $done = 1;
                }
            } otherwise {
                if ($::IGNORE_ERRORS) {
                    my $e = shift;
                    $Log->warn("Got following exception, ignoring and "
                        . "continuing: ".$e->text());
                    if ( ($input{iterations} eq 'infinite') ||
                         ($iter < $input{iterations}) ) {
                        wait_for_interval(
                            interval     => $input{interval},
                            start_minute => $input{start_minute},
                            start_tick   => $start_tick,
                        );
                        $iter++;
                    } else {
                        $done = 1;
                    }
                } else {
                    $exception = shift;
                    $done = 1;
                }
            };
        }
    }

    $my_return->{end_tick} = time();
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Scrubbed $num_aggrs aggregates for $input{iterations} "
            . "iterations";
        $Log->comment( $msg );
        $my_return->{msg} = $msg;
    }
    return $my_return;

} # End scrub


###############################################################################
###########               Helper Subroutines                        ###########
###############################################################################

sub _if_exists_return_checksum() {
    my %opts    = @_;
    my $api_obj = $opts{api};
    my $aggregate = $opts{aggr};
    my $output;

    try {
        $output = $api_obj->aggr_status(
                        'raid-disk' => 1,
                        aggregate   => $aggregate,
                  )->get_parsed_output();
    } catch NACL::APISet::Exceptions::InvalidParamValueException with {
        return 0;
    };

    return $output->[0]->{aggregate}->[0]->{checksums};
}

sub _get_spares () {
    my %opts       = @_;
    my $node       = $opts{node};
    my $nn         = $node->node();
    my $checksum   = $opts{checksum_type};
    my $disk_name  = $opts{disk_name};
    my %args;

    if($checksum){
        $args{'checksum-compatibility'} = $checksum;
    }
    if($disk_name){
        $args{'disk'} = $disk_name;
    }

    my @disk_list = NACL::STask::StorageDisk->find(
        command_interface => $node,
        filter            => {
            'container-type'   => 'spare',
            home                => $nn,
            %args,
        },
        allow_empty       => 1,
    );

    my $disk_info;
    my $ret_disks;
    my $cnt = 0;
    my $num = 0;
    foreach (@disk_list) {
        $disk_info .= "\n";
        
        my $cs = $_->state();
        $disk_info .= "DT: *".$cs->type()."*\n";
        $ret_disks->[$cnt]->{disk_type} = $cs->type();

        $disk_info .= "CK: *".$cs->checksum_compatibility()."*\n";
        $ret_disks->[$cnt]->{checksum_type} = $cs->checksum_compatibility();

        my $z = $cs->zeroed();
        if ($z =~ /true/i) {
            $ret_disks->[$cnt]->{fsm_state} = 'zeroed';
        } elsif ($cs->zeroing_percent() eq '-') {
            $ret_disks->[$cnt]->{fsm_state} = 'notzeroed';
        } else {
            $ret_disks->[$cnt]->{fsm_state} = 'zeroing';
        }
        $disk_info .= "FS: *".$ret_disks->[$cnt]->{fsm_state}."*\n";

        $disk_info .= "SN: *".$cs->serialnumber()."*\n";
        $ret_disks->[$cnt]->{serialno} = $cs->serialnumber();

        $disk_info .= "IS_SPARE: *".$cs->container_type()."*\n";
        
        $disk_info .= "NAME_XLATED: *".$_->disk()."*\n";
        $ret_disks->[$cnt]->{name} = $_->disk();
        $cnt++;
        $num++;
    }

    $Log->comment("$disk_info\nNUM_DISKS = $num\n");

    return $ret_disks;
}

1;
