# this is basically -*- perl -*-
# $Id$
#
# Copyright (c) 2013 NetApp
# All rights reserved.
#
## @summary
## Helper functions for NACL XML translation framework.
##
## @status    private
##
package NX::Disk;

use strict 'vars';
use Tharn;
use Data::Dumper;
use NX::Helper;
use NX::Volume;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::STask::Volume;
use NACL::C::StorageAggregate;
use NACL::C::StorageDisk;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
  fail
);

my @Pkg_Workers;

sub stop_workers {
    $Log->comment("Detected interrupt signal");

    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}

=pod

=head1 NAME

NX::Disk

=head1 DESCRIPTION

Disk related routines for use in NX framework;

=head1 METHODS

=head2 fail

Fail disks in aggregate.

The input is a hash (%opts) which must include the command_interface:
    command_interface => $node

=head3 opts

The input is a hash (%opts) which must include:
    aggregates          => Arrayref of one or more aggregate objects

%opts can also be used to override default values, which include
    disk_position       => Optional
                        # Specifies what disk position to use
    disk_type           => Optional
                        # Specifies what disk type to use
    num_disks_per_rg    => Optional, defaults to 1
                        # Specifies number of disks to fail per raid group
    num_rg              => Optional, defaults to 1
                        # Specifies number of raid groups to fail
    wait_time           => Optional, defaults to 1800 seconds
                        # Specifies how long to wait for raid groups to recover
    start_minute        => Required if <interval> is odd|even|all, otherwise
                        ignored if <interval> is an integer value.  This is
                        a two-digit value representing the clock minute
    interval            => odd/even/all/<value in seconds>
                        # Every odd, even, or every clock hour,
                        # or specified value in seconds
    iterations          => integer value or <infinite>
                        # requires interval if defined

A sample call would be:

%opts = {
    aggregates          => [ $aggr_obj1 ],
    num_disks_per_rg    => 2,
    num_rg              => 2,
    wait_time           => 100
    disk_position       => data
    disk_type           => SSD
    interval            => 60,
    iterations          => 5,
}

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

sub fail {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    # Hash reference containing return values
    my $my_return;
    my $exception;
    my $interval;

    try {

        # Validate parameters
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                aggregates    => { type => ARRAYREF, optional => 0 },
                disk_type     => { type => SCALAR,   optional => 1 },
                disk_path     => { type => SCALAR,   optional => 1 },
                disk_position => { type => SCALAR,   optional => 1 },
                wait_time => { type => SCALAR, optional => 1, default => 1800 },
                num_disks_per_rg =>
                  { type => SCALAR, optional => 1, default => 1 },
                num_rg     => { type => SCALAR, optional => 1, default => 1 },
                iterations => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 1
                },
                interval     => { type => SCALAR, optional => 1 },
                start_minute => { type => SCALAR, optional => 1 },
                validate_only =>
                  { type => SCALAR, optional => 1, default => 0 },
                _num_nodes =>
                  { type => SCALAR, optional => 1, default => undef },
            }
        );

        $interval = $input{interval};
        foreach my $arg ( keys %input ) {
            if (   ( $arg eq "iterations" )
                && ( $input{$arg} > 1 )
                && !defined($interval) )
            {
                # validate that <interval> is defined if <iterations> is defined
                NATE::BaseException->throw(
"Invalid value: \'$input{$arg}\'. \'iterations\' parameter requires a value "
                      . "for the \'interval\'' parameter" );
            } elsif ( ( $arg eq 'iterations' )
                && ( ( $input{$arg} !~ /^(\d+|infinite)$/ ) || !$input{$arg} ) )
            {
                NATE::BaseException->throw(
"The \'iterations\' parameter value must be either a positive integer or \'infinite\'"
                );
            } elsif ( $arg eq "interval" ) {

                # wait_for_interval() validates start_minute
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            } elsif ( $arg eq "num_disks_per_rg" ) {

                # wait_for_interval() validates start_minute
                if ( $input{num_disks_per_rg} < 0 ) {
                    NATE::BaseException->throw(
"The \'num_disks_per_rg\' parameter value must be greater than 0"
                    );
                }
            } elsif ( $arg eq "num_rg" ) {

                # wait_for_interval() validates start_minute
                if ( $input{num_rg} < 0 ) {
                    NATE::BaseException->throw(
"The \'num_disks_per_rg\' parameter value must be greater than 0"
                    );
                }
            } elsif ( $arg eq "wait_time" ) {

                # wait_for_interval() validates start_minute
                if ( $input{wait_time} < 0 ) {
                    NATE::BaseException->throw(
"The \'wait_time\' parameter value must be greater than 0"
                    );
                }
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Disk start action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Disk::stop_workers";
        delete $input{validate_only};
    }

    my @aggrs      = @{ $input{aggregates} };
    my $iterations = $input{iterations};

    my $num_aggr = @aggrs;
    my @nodes;
    my $_num_nodes;
    my $node_to_aggrs;

    # Extract number of unique nodes, if we haven't already
    # determined that from a previous call
    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $a (@aggrs) {
            my $node = $a->command_interface();
            my $name = $node->name;
            if ( !defined( $node_hash->{$name} ) ) {
                push( @nodes, $node );
                $node_hash->{$name} = 1;
            }
            push( @{ $node_to_aggrs->{$name} }, $a );
        }
        $_num_nodes = @nodes;
    }
    foreach my $a (@aggrs) {
        $Log->comment( "Aggregate: " . $a->aggregate() );
    }
    foreach my $n (@nodes) {
        $Log->comment( "Node: " . $n->name() );
    }

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_num_nodes} && ( $_num_nodes > 1 ) ) {

        # multiple nodes, multiple aggrs
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            my $node_name = $n->name();
            my $id        = 'SDF_N_DaC.' . $proc_cnt;
            push( @ids, $id );

            # Already parsed out unique nodes, don't need to do again
            $input{_num_nodes} = $_num_nodes;
            $input{aggregates} = $node_to_aggrs->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Disk',
                method  => 'fail',
                args    => [%input],
            );
            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;

        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }
        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran disk fail on $_num_nodes nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } elsif ( $num_aggr > 1 ) {

        # single node, multiple aggr
        my $node      = $aggrs[0]->command_interface();
        my $node_name = $node->name();
        my $idx       = 0;
        my $done      = 0;
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $input_ref = \%input;

        # aggr related
        my $aggr_name;
        my $aggr_obj;
        my $aggr_raid_type;
        my $aggr_rg_count;

        $my_return->{start_tick} = time();

        # Work on current batch of aggrs
        foreach my $a (@aggrs) {

            # validate <num_disks_per_rg> and <num_rg> by checking aggrs
            $aggr_name = $a->aggregate();
            $Log->comment("Validating config for $aggr_name");
            $aggr_obj = NACL::C::StorageAggregate->new(
                command_interface => $node,
                aggregate         => $aggr_name,
            );

            $aggr_raid_type = $aggr_obj->state()->raidtype();
            $aggr_rg_count  = scalar( $aggr_obj->state()->raidgroups() );

            if ( $aggr_raid_type eq "raid_dp" && $input{num_disks_per_rg} > 2 )
            {
                $Log->warn(
"The \'num_disks_per_rg\' parameter value must not be greater than 2 for $aggr_raid_type"
                );
                next;
            } elsif ( $aggr_raid_type eq "raid4"
                && $input{num_disks_per_rg} > 1 )
            {
                $Log->warn(
"The \'num_disks_per_rg\' parameter value must not be greater than 1 for $aggr_raid_type"
                );
                next;
            }

            if ( $input{num_rg} > $aggr_rg_count ) {
                $Log->warn(
"The \'num_rg\' parameter value must not be greater than $aggr_rg_count for $aggr_name"
                );
                next;
            }

            my $id = 'SDF_A_DaC' . $proc_cnt;
            push( @ids, $id );
            $input{aggregates} = [$a];
            $input{_num_nodes} = 1;

            # added delay to offset disk runs
            sleep(10);
            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Disk',
                method  => 'fail',
                args    => [%input],
            );
            $Log->comment("$id is running");
            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        $Log->comment("collecting results");

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        my $msg;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }
        $my_return->{end_tick} = time();

        if ($had_failure) {
            $my_return->{result} = 0;
            $msg = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            $msg = "Failed disks successfully on $node_name";
        }
        $Log->comment($msg);
        $my_return->{msg} = $msg;
        return $my_return;
    } else {
        # single node, single aggr
        my $node      = $aggrs[0]->command_interface();
        my $node_name = $node->name();
        my $aggr_name = $aggrs[0]->aggregate();

        # status message
        my $msg;

        # aggr related var
        my $aggr_obj;
        my $aggr_raid_type;
        my $aggr_rg_count;
        my @rg_avail;

        # disk related var
        my @disk_list;
        my $disk_obj;
        my $disk_name;
        my $disk_type;
        my $disk_position;
        my $raid_group;

        # all the rg in aggr
        my $rg_list;

        # search criteria for disks
        my %disk_filter;

        # disks to fail, should contain NACL::C::StorageDisk objs
        my @fail_list;

        # poll interval to check when there are disks available
        my $disk_poll = $input{wait_time};

        # for store NACL::CS obj for later printing
        my @print_buffer;

        my $done = 0;
        my $i    = 1;

        my $temp;
        my @temp;

        my $raid_group_selected = 0;
        my $start_tick          = time();
        $my_return->{start_tick} = $start_tick;

        #                               1   2   3
        # *steps*
        # get total disk count          x   x   x
        # select disks to fail          x   x   x
        # check for rg degradation      x   x   x
        # cycle through rg
        # fail disks                    x   x   x
        # unfail disks                  x   x   x
        # zero spares                   x   x   x

        # *cases*
        # 1. disk name defined
        # 2. parity disk
        # 3. any other type of disk

        # validate <num_disks_per_rg> and <num_rg> by checking aggrs
        $Log->comment("Validating config for $aggr_name");
        $aggr_obj = NACL::C::StorageAggregate->new(
            command_interface => $node,
            aggregate         => $aggr_name,
        );

        $aggr_raid_type = $aggr_obj->state()->raidtype();
        @temp           = $aggr_obj->state()->raidgroups();
        $aggr_rg_count  = scalar(@temp);

        if ( $aggr_raid_type eq "raid_dp" && $input{num_disks_per_rg} > 2 ) {
            $Log->warn(
"The \'num_disks_per_rg\' parameter value must not be greater than 2 for $aggr_raid_type"
            );
        } elsif ( $aggr_raid_type eq "raid4" && $input{num_disks_per_rg} > 1 ) {
            $Log->warn(
"The \'num_disks_per_rg\' parameter value must not be greater than 1 for $aggr_raid_type"
            );
        }

        if ( $input{num_rg} > $aggr_rg_count ) {
            $Log->warn(
"The \'num_rg\' parameter value must not be greater than $aggr_rg_count for $aggr_name"
            );
        }

        # outer try to catch exceptions thrown for $::IGNORE_ERRORS
        try {
            while ( !$done ) {
                @fail_list   = ();
                @disk_list   = ();
                %disk_filter = ();

                try {
                    $disk_filter{aggregate} = $aggr_name;

                    # if no disk is specified, fail random disk
                    # if disk specified, disk name takes precedence
                    # over any other filters
                    if ( defined( $input{disk_path} ) ) {

                        # select disk
                        $disk_name = $input{disk_path};
                        my $disk_search = NACL::C::StorageDisk->find(
                            command_interface => $node,
                            filter            => { disk => $disk_name },
                        );
                        $disk_obj = NACL::C::StorageDisk->new(
                            command_interface => $node,
                            disk              => $disk_search->disk(),
                        );

                        push( @fail_list, $disk_obj );

                    } elsif ( $input{disk_position} eq "parity" ) {
                        my @parity_temp;

                        # if parity disk specified
                        if ( defined( $input{disk_type} ) ) {
                            $disk_filter{type} = $input{disk_type};
                        }
                        if ( defined( $input{disk_position} ) ) {
                            $disk_filter{position} = $input{disk_position};
                        }

                        # if no disks available, poll in X min intervals
                        while (1) {
                            $Log->comment("Finding parity disks in $aggr_name");

                   # ignore exception because we still want to check for dparity
                            try {
                                #TODO investigate this
                                # append dparity disk to parity disk list
                                @disk_list = NACL::C::StorageDisk->find(
                                    command_interface => $node,
                                    filter            => {%disk_filter},
                                );
                            }
                            catch NACL::Exceptions::NoElementsFound with {

                                # disk list is empty
                                $Log->comment(
                                    "No parity disks found available");
                            } otherwise {

                                # REALLY BAD ERROR
                                $Log->comment("Error in disk find");
                            };

                        # ignore exception because we may have found parity disk
                            try {
                                $disk_filter{position} = "dparity";
                                @temp = NACL::C::StorageDisk->find(
                                    command_interface => $node,
                                    filter            => {%disk_filter},
                                );
                                push( @disk_list, @temp );
                            }
                            catch NACL::Exceptions::NoElementsFound with {

                                # disk list is empty
                                $Log->comment(
                                    "No dparity disks found available");
                            } otherwise {

                                # REALLY BAD ERROR
                                $Log->comment("Error in disk find");
                            };

                            # number of disks available has to be greater than
                            # max number of disks failable per raid group
                            if (scalar(@disk_list) >= $input{num_disk_per_rg}) {
                                $Log->comment("Disks available, proceeding");
                                last;
                            } else {
                                $Log->comment("No parity disks available, "
                                    . "sleeping for $disk_poll minutes");
                                sleep($disk_poll);
                            }
                        }    # while
                    } else { # any kind of disk
                        if ( defined( $input{disk_type} ) ) {
                            $disk_filter{type} = $input{disk_type};
                        }
                        if ( defined( $input{disk_position} ) ) {
                            $disk_filter{position} = $input{disk_position};
                        }
                        $Log->comment("Finding disks in $aggr_name");
                        try {
                            @disk_list = NACL::C::StorageDisk->find(
                                command_interface => $node,
                                filter            => {%disk_filter},
                            );
                        }
                        catch NACL::Exceptions::NoElementsFound with {

                            # disk list is empty
                            $Log->comment("No disks found available");
                        } otherwise {

                            # REALLY BAD ERROR
                            $Log->comment("Error in disk find");
                        };
                    }    # else

                    # select raid group
                    if ( !$input{disk_path} ) {

                        # disk path not specified
                        for ( my $i = 0 ; $i < $aggr_rg_count ; $i++ ) {
                            my $disks_added = 0;
                            $Log->comment("Determining raid group to use");
                            $raid_group = "rg$i";
                            $Log->comment(
                                "Raid group selected: " . $raid_group );
                            $Log->comment("Pushing disks into fail list");
                            foreach (@disk_list) {

              # only push disks of the selected raid group into fail list
              # by now, disks in will be filtered already to user specifications
                                my $disk_name = $_->disk();
                                $Log->comment("Checking if $disk_name matches "
                                    . "requirements");

                                $temp = $_->state()->raid_group();
                                if ( $temp eq $raid_group ) {

                                    $disk_obj = NACL::C::StorageDisk->new(
                                        command_interface => $node,
                                        disk              => $disk_name,
                                    );
                                    $Log->comment(
                                        "Pushing $disk_name into fail list");
                                    push( @fail_list, $disk_obj );
                                    $disks_added++;
                                } else {
                                    next;
                                }

                                if ( $disks_added == $input{num_disks_per_rg} )
                                {

                              # reached number of disks in to fail per iteration
                                    $Log->comment("Number of disks requested "
                                        . "per raid group to fail reached.");
                                    last;
                                }
                            }    # foreach

                            if (
                                scalar(@fail_list) >=
                                ( $input{num_disks_per_rg} * $input{num_rg} ) )
                            {

                              # reached number of disks in to fail per iteration
                                $Log->comment("Number of disks requested total "
                                    . "to fail reached.");
                                last;
                            }
                        }    # for
                        if ( !scalar(@fail_list) ) {
                           # throw exception if there are no disks by this point
                            NATE::BaseException->throw(
                                "Error: No disks in fail list");
                        }
                    }    # if

                  # disks to fail
                  # at this point, there should always be some disks in the list
                    @print_buffer = ();
                    foreach (@fail_list) {
                        $temp =
                            $_->state()->aggregate() . "\t"
                          . $_->disk() . "\t"
                          . $_->state()->raid_group();
                        push( @print_buffer, $temp );
                    }
                    $Log->comment("Disk Fail List");
                    foreach (@print_buffer) {
                        $Log->comment($_);
                    }

                    foreach (@fail_list) {

                        # raid group degradation check
                        while (1) {
                            @print_buffer = ();
                            $raid_group   = $_->state()->raidgroup();
                            $Log->comment(
                                "Checking for raid group degradation");
                            my $reconstructing_count = 0;

                            # get available disks to compare
                            %disk_filter = (
                                aggregate    => $aggr_name,
                                'raid-group' => $raid_group,
                            );
                            try {
                                @temp = NACL::C::StorageDisk->find(
                                    command_interface => $node,
                                    filter            => {%disk_filter},
                                );

                                # search for reconstructing disks
                                foreach (@temp) {
                                    $temp =
                                      $_->state()->reconstruction_percent();
                                    if ( $temp ne '-' ) {
                                        $temp =
                                            $_->disk() . "\t"
                                          . $_->state()->raid_group() . "\t"
                                          . $_->state()
                                          ->reconstruction_percent() . "%";
                                        push( @print_buffer, $temp );
                                        $reconstructing_count++;
                                    }
                                    sleep(10);
                                }
                                if ( scalar(@print_buffer) >= 1 ) {
                                    $Log->comment("Reconstructing disks:");
                                } else {
                                    $Log->comment("No disks reconstructing");
                                }
                                foreach (@print_buffer) {
                                    $Log->comment($_);
                                }

                            } otherwise {

                                # REALLY BAD ERROR
                                $Log->comment("Error looking for disks");
                            };

                        # check if raid group is degraded
                        #if ($reconstructing_count >= $input{num_disks_per_rg}){
                            if ( $reconstructing_count != 0 ) {
                                $Log->comment(
"Sleeping $disk_poll seconds to wait until raid group recover"
                                );
                                sleep($disk_poll);
                            } else {
                                $Log->comment(
"Raid group ready. Proceeding to fail disks..."
                                );
                                last;
                            }    # else
                        }    # while
                        my $disk_name = $_->disk();
                        $Log->comment("Failing disk $disk_name on $aggr_name");
                        $_->fail( immediate => "true" );

                        # check if disk has been failed
                        while (1) {
                            try {
                                $Log->comment("Checking for disk failure...");
                                NACL::C::StorageDisk->find(
                                    command_interface => $node,
                                    filter            => {
                                        disk             => $disk_name,
                                        'container-type' => "broken"
                                    },
                                );
                                $Log->comment("$disk_name failed successfully");
                                last;
                            }
                            catch NACL::Exceptions::NoElementsFound with {

                                # disk list is empty
                                $Log->comment("$disk_name not yet failed...");
                                sleep(10);
                            }
                        }    # while

                        $Log->comment(
                            "Unfailing disk $disk_name on $aggr_name");
                        $_->unfail();

                        # check if disk has been unfailed
                        while (1) {
                            try {
                                $Log->comment(
                                    "Checking to for disk Unfailing...");
                                NACL::C::StorageDisk->find(
                                    command_interface => $node,
                                    filter => { disk => $disk_name, },
                                );
                                $Log->comment(
                                    "$disk_name unfailed successfully");
                                last;
                            }
                            catch NACL::Exceptions::NoElementsFound with {

                                # disk list is empty
                                $Log->comment("$disk_name not yet unfailed...");
                                sleep(10);
                            }
                        }    # while
                    }    # foreach

                    # longer delay so the disks will be zeroed, else does not start zeroing
                    sleep(20);
                    NACL::C::StorageDisk->zerospares(
                        command_interface => $node, );
                    $Log->comment("Zeroing spares on $node_name");

                    # don't call interval if last iteration
                    if ( $input{iterations} eq 'infinite' || $i != $iterations )
                    {
                        $Log->comment("Successfully completed iteration #$i");
                        $Log->comment("Sleeping for $interval seconds "
                            . "before next iteration on $aggr_name");
                        wait_for_interval(
                            interval     => $input{interval},
                            start_minute => $input{start_minute},
                            start_tick   => $start_tick,
                        );
                    } else {
                        $Log->comment("Last iteration #$i");
                        $done = 1;
                    }
                } otherwise {
                    my $exception = shift;
                    if ($::IGNORE_ERRORS) {

                    } else {
                        $exception->throw();
                    }
                };
                $i++;
            }    # end main loop for disk fail execution
        } otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $msg                    = "Failed disk fail action on $aggr_name";

            # disks aren't zeroed if action fails
            sleep(10);
            NACL::C::StorageDisk->zerospares( command_interface => $node, );
            $Log->comment("Zeroing spares on $node_name");
        } else {
            $my_return->{result} = 1;
            $msg =
              "Successful disk fail action on $aggr_name $iterations times";
        }
        $Log->comment($msg);
        $my_return->{msg} = $msg;
        return $my_return;
    }
}    # End of fail

1;
