#
# Copyright 2013 NetApp, Inc.
#
# File related actions for NX or general autmoation use.
#
#
use strict;

package NX::File;

use Tharn;
use NACL::STask::Node;
use NX::Helper;
use NX::IO;
use NATE::PerlRPC::Worker::Process;
use NACL::C::VolumeFileMove;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);
use Data::Dumper;

our @ISA = ('Exporter');
our @EXPORT
    = qw( sfmod );

my @Pkg_Workers;

sub stop_workers {
    $Log->comment("Detected interrupt signal");

    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
} ## end sub stop_workers

=pod

=head1 NAME

NX::File

=head1 DESCRIPTION

File related routines for use in NX framework.

=head1 METHODS

=head2 sfmod

Perform file overwrites on specified volumes while performing single file move
operation. The specified file will move sequentially from the 1st volume
specified to the 2nd, and so forth until the last volume. Finally, the file
will then be moved back to the 1st volume. This constitutes one iteration.

=head3 opts

%opts can be used to provide additional options for sfmod
    validate_only            => <VALIDATE_ONLY option>
    client_command_interface => <array ref of client objects>
    volumes                  => <arrayref of volume objects>
    iterations               => <integer>|infinite (default is 1)
    interval                 => odd/even/all/<value in seconds>
                                Every odd, even, or every clock hour,
                                or specified value in seconds
    start_minute             => Required if "interval" is odd|even|all, otherwise
                                ignored if "interval" is an integer value.  This is
                                a two-digit value representing the clock minute
    sio_file_size            => Size of each siofile, defaults to 10m
    sio_num_files            => Number of sio files to move. Defaults to 5 per
                                volume.
    sio_duration             => Duration in seconds for sio to run. Defaults to 600.
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX sfmod
Perform IO and single file move operation between volumes.
description
Perform file overwrites on specified files, while performing file move operation between specified volumes.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volume(s);;Volume(s) for file to move between. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
iterations;Iterations;1;Number of times to run the clone split operation on each volume
interval;Interval;1;interval of time to wait between clone split operations
start_minute;Start Minute;1;Required if 'interval' is odd|even|all, otherwise ignored if 'interval' is an integer value.
sio_file_size;File Size for sio;10m;Defaults to '10m'
sio_num_files;Number of files for sio and to move at once;5;Defaults to 5
sio_duration;SIO duration;600;Duration in seconds for sio to run. Defaults to 600
=end NX sfmod
=cut

sub sfmod {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                client_command_interface =>
                    { type => ARRAYREF, optional => 0 },
                volumes =>
                    { type => ARRAYREF, optional => 0 },
                iterations =>
                    { type => SCALAR, optional => 1, default => 1 },
                interval     =>
                    { type => SCALAR, optional => 1 },
                start_minute =>
                    { type => SCALAR, optional => 1 },
                sio_file_size =>
                    { type => SCALAR, optional => 1, default => '10m' },
                sio_num_files =>
                    { type => SCALAR, optional => 1, default => 5 },
                sio_duration =>
                    { type => SCALAR, optional => 1, default => 600 },
                _ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
            }
        );

        foreach my $key ( keys %input ) {
            my $val = $input{$key};
            if ( $key eq "sio_file_size" ) {
                if ( ( $input{$key} !~ /\d+[kKmMgG]?$/ ) ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer optionally followed by k,m or g"
                    );
                }
            } elsif ( $key eq 'sio_num_files' ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw( "'$val' value for '$key' "
                            . "parameter must be an integer" );
                }
            } elsif ( $key eq 'sio_duration' ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw( "'$val' value for '$key' "
                            . "parameter must be an integer" );
                }
            } elsif ( $key eq 'iterations'
                      && ( $val > 1 )
                      && !defined( $input{interval} ) )
            {
                NATE::BaseException->throw(
                          "\'iterations\' parameter requires a"
                        . " positive integer for the \'interval\' parameter"
                );
            } elsif ( $key eq 'interval' ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute}
                );
            } elsif ( $key eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/i ) {
                    NATE::BaseException->throw( "'$val' value for '$key' "
                            . "parameter must be an integer or 'infinite'" );
                }
            }
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "SFMOD passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX:File::stop_workers";
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;
    if ( $num_volumes < 2 ) {
        NATE::BaseException->throw( "At least two volumes must be "
            . "specified to perform file move. Only $num_volumes "
            . "was specified." );
    }
    my $qty_threshold = 2;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    my $n              = $volume_objs[0]->command_interface();
    my $iterations     = $input{iterations};
    my $interval       = $input{interval};
    my $cur_iter       = 0;
    my $done           = 0;
    my $sio_duration   = $input{sio_duration};
    my @files;
    for (0..($input{sio_num_files} - 1)) {
        push(@files, "sfmod_file_".$_);
    }
    my $num_files = @files;

    while ( ($iterations > 0 || $iterations eq 'infinite') && !$done ) {
        $cur_iter++;
        try {
            my $tracker;
            my $num_files_looped = 0;
            my $cur_move = 0;
            while ( $num_files_looped != $num_files ) {
                my $file_to_vol;
                my $vol_to_files;
                my $volume_files;
                $cur_move++;

                # Figure out where the files are currently
                foreach my $f (@files) {
                    my $vol_pos = 0;
                    foreach my $v (@volume_objs) {
                        my $vol_name = $v->volume();
                        my $vs = $v->vserver();
                        my $api = $v->command_interface()->get_7m_or_nodescope_apiset();
                        my $parsed = $api->ls(paths => '/vol/'.$vol_name)->get_parsed_output();
                        if (grep{$_ eq $f} @{$parsed->[0]->{files}}) {
                            $file_to_vol->{$f}->{cur_vol_path} = $vs.':'.$vol_name.'/';
                            push(@{$vol_to_files->{$vol_name}}, $f);
                            if (!$volume_objs[$vol_pos+1]) {
                                $file_to_vol->{$f}->{next_vol_path} =
                                    $vs.':'.$volume_objs[0]->volume().'/';
                                $tracker->{$f}->{looping} = 1;
                            } else {
                                $file_to_vol->{$f}->{next_vol_path} =
                                    $vs.':'.$volume_objs[$vol_pos+1]->volume()
                                    . '/';
                            }
                            $Log->comment("File $f is on $vol_name. Will be "
                                . "moved to " . $file_to_vol->{$f}->{next_vol_path});
                            last;
                        }
                        $vol_pos++;
                    }
                }

                # For brand new files, initialize tracker to indicate
                # existing and next vol destination
                foreach my $f (@files) {
                    if (!defined($file_to_vol->{$f})) {
                        my $vol_name = $volume_objs[0]->volume();
                        my $vs = $volume_objs[0]->vserver();
                        push(@{$vol_to_files->{$vol_name}}, $f);
                        $file_to_vol->{$f}->{cur_vol_path} = $vs.':'.$vol_name.'/';
                        $file_to_vol->{$f}->{next_vol_path} =
                            $vs.':'.$volume_objs[1]->volume().'/';
                        $Log->comment("File $f initially on $vol_name. Will "
                            . "be moved to " . $file_to_vol->{$f}->{next_vol_path});
                    }
                }

                # Make sure any same named file is removed from the
                # next volume that file is being moved to
                my $volume_files;
                foreach my $v (@volume_objs) {
                    my $vol_name = $v->volume();
                    my $api = $v->command_interface()->get_7m_or_nodescope_apiset();
                    my $parsed = $api->ls(paths => '/vol/'.$vol_name)->get_parsed_output();
                    push(@{$volume_files->{$v->volume()}}, @{$parsed->[0]->{files}});
                }

                foreach my $f (@files) {
                    my $path = $file_to_vol->{$f}->{next_vol_path};
                    $path =~ /\:(\S+)\//;
                    my $vol_name = $1;
                    foreach my $v (@volume_objs) {
                        if ($v->volume() eq $vol_name) {
                            if (grep{$_ eq $f} @{$volume_files->{$vol_name}}) {
                                my $api = $v->command_interface()->get_7m_or_nodescope_apiset();
                                $Log->comment("Removing $f from volume "
                                    . $vol_name .", because file will be "
                                    . "moved to this volume");
                                $api->rm( paths => '/vol/'.$vol_name.'/'.$f );
                                last;
                            }
                        }
                    }
                }

                # Start traffic on volumes
                my @client_objs = @{ $input{client_command_interface} };
                my $client = $client_objs[0];
                my $end = $input{sio_file_size};
                my $futures;
                my @workers;

                # Convert end from k|m|g to bytes
                if ( $end =~ s/k//i ) {
                    $end = $end * 1024;    #k == 1024
                } elsif ( $end =~ s/m//i ) {
                    $end = $end * 1024 * 1024;    #m == 1024*1024
                } elsif ( $end =~ s/g//i ) {
                    $end = $end * 1024 * 1024 * 1024;   #g == 1024*1024*1024
                }

                my %input_sio;
                $input_sio{reads}                    = 50;
                $input_sio{start}                    = 0;
                $input_sio{end}                      = $end;
                $input_sio{secs}                     = $sio_duration;
                $input_sio{filesize}                 = $input{sio_file_size};
                $input_sio{client_command_interface} = $input{client_command_interface};

                foreach my $v (@volume_objs) {
                    next if !defined($vol_to_files->{$v->volume()});
                    my @files = @{$vol_to_files->{$v->volume()}};
                    my $file_list = join(',', @files);
                    $Log->comment("Start IO against vol ".$v->volume()
                        ." on these files: $file_list");
                    $input_sio{file_names}               = $file_list;
                    $input_sio{volumes}                  = [ $v ];

                    my $sio_id = 'SFMOD_IO_DaC_i'.$cur_iter.'_m'.$cur_move
                        . '_'.$v->volume();
                    my $sio_worker = NATE::PerlRPC::Worker::Process->new(
                        runid => $sio_id);
                    push(@workers,      $sio_worker);
                    push(@Pkg_Workers,  $sio_worker );

                    my $sio_future = $sio_worker->submit(
                        package => 'NX::IO',
                        method  => 'sio',
                        args    => [%input_sio],
                    );
                    $futures->{$sio_id}->{obj} = $sio_future;
                }

                # Perform file move halfway into SIO run time
                sleep $sio_duration / 2;

                my $num_moves = 0;
                foreach my $f (@files) {
                    if ($tracker->{$f}->{looped}) {
                        $Log->comment("File $f already looped back to "
                            . "initial volume, skipping move");
                        $tracker->{$f}->{obj} = undef;
                    } else {
                        $tracker->{$f}->{move_done} = undef;
                        $tracker->{$f}->{obj} = NACL::C::VolumeFileMove->start(
                           command_interface => $n,
                           'source-paths'    => [ $file_to_vol->{$f}->{cur_vol_path}.$f ],
                           'destination-paths' => [ $file_to_vol->{$f}->{next_vol_path}.$f ],
                        );
                        $num_moves++;
                    }
                }

                my $num_done = 0;
                while ($num_done != $num_moves) {
                    sleep 30;
                    foreach my $f (keys(%$tracker)) {
                        next if ($tracker->{$f}->{move_done}
                            || $tracker->{$f}->{looped});
                        my $obj = $tracker->{$f}->{obj};
                        my $percent = $obj->state()->scanner_percent();
                        my $status = $obj->state()->scanner_status();
                        if (($status =~ /complete/i) && ($percent =~ /100/)) {
                            $num_done++;
                            $tracker->{$f}->{move_done} = 1;
                            if ($tracker->{$f}->{looping}) {
                                $tracker->{$f}->{looped} = 1;
                                $num_files_looped++;
                            }
                        }
                    }
                    $Log->comment("$num_done of $num_moves file moves completed");
                }

                my $res = collect_worker_results(
                    workers       => \@workers,
                    futures       => $futures,
                    merge_context => 0,
                );
                # Don't bother with checking results of IO, as we expect
                # errors due to file being moved to another volume.
            }

            if ( $iterations ne 'infinite' ) {
                $iterations--;
            }

            if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                wait_for_interval(
                    interval     => $input{interval},
                    start_minute => $input{start_minute},
                    start_tick   => $my_return->{start_tick},
                );
            }
        } otherwise {
            my $e = shift;
            if ($::IGNORE_ERRORS) {
                my $msg = $e->text();
                $Log->warn("Hit exception during SFMOD action: $msg");

                if ( $iterations ne 'infinite' ) {
                    $iterations--;
                }

                if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $my_return->{start_tick},
                    );
                }
            } else {
                $Log->error("Caught exception during SFMOD. "
                            ."Error was: ". $e->text() );
                $exception = $e;
                $done      = 1;
            }
        };
    }

    $my_return->{end_tick} = time();
    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Finished SFMOD of $input{sio_num_files} files across "
            ."$num_volumes volumes";
        $Log->comment($msg);
        $my_return->{msg}    = $msg;
    } ## end else [ if ($exception)
    return $my_return;

}    # End sfmod

1;
