# this is basically -*- perl -*-
# $Id: $
#
# Copyright (c) 2012 NetApp
# All rights reserved.
#
## @summary
## File clone functions to support NACL Translator framework
##
## @status    private
##

package NX::File_clone;

use strict 'vars';
use Tharn;

use Data::Dumper;

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::StorageAggregate;
use NACL::STask::Aggregate;
use NACL::STask::Vserver;
use NX::Helper;
use NACL::STask::SystemLicense;
use NACL::C::VolumeFileClone;
use NACL::APISet::Node::CLI::7Mode::default::filersio;
use NACL::APISet::Node::CLI::7Mode::default::slist;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
    start
);

my @Pkg_Workers;

sub stop_workers {
    $Log->comment("Detected interrupt signal");

    # Stop any workers from divide/conquer logic
    _nx_stop_workers(workers => \@Pkg_Workers);
    exit;
} ## end sub stop_workers

=pod

=head1 NAME

NX::File_clone

=head1 DESCRIPTION

File clone functions to support NACL Translator framework;

=head1 METHODS

=head2 start

Create a file clone on the volume.

The input is a hash (%opts)

=head3 opts

%opts can be used to provide options for file clone creation
    volume                  => volume to create file clone in 
    source_file             => path to source file for clone
    snapshot                => name of snapshot if clone is to be created from file in snapshot
    destination_file        => path to clone file
    #If number of clones is specified then destination file is used as base name
    number_of_clones        => number of clones to be created
    source_start_range      => start range of source file for sub file clone
    destination_start_range => start range of destination file for sub file clone
    num_of_blocks           => num of files to clone for sub file clone   
    vserver                 => vserver is required if Filer is in Cmode
    #The following 2 parameters are used for wait_for_interval function
    #Check Helper.pm for further info
    interval                => odd/even/all/<value in seconds>
                               # Every odd, even, or every clock hour,
                               # or specified value in seconds
    start_minute            => Required if "interval" is odd|even|all, otherwise
                               ignored if "interval" is an integer value.  This is
                               a two-digit value representing the clock minute
    iterations              => Number of times to repeat the process 


    
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX start
Start
description
Create file clones on volumes
required
source_file;Source File;;Path to source file for clone
volumes;Volume(s);;Volumes to create file clone in
optional
snapshot;Snapshot;;Name of snapshot if clone is to be created from file in snapshot
vserver;Vserver;;Vserver is required if filer is in CMode
destination_file;Destination File;;Path to clone file. Note: If number of clones is specified then destination file is used as a base name
source_start_range;Source Start Range;;Start range of source file
destination_start_range;Destination Start Range;;Start range of destination file
num_of_blocks;Num Of Blocks;;Number of 4k blocks to clone in file
num_of_clones;Num Of Clones;;Number of clones
interval;Interval;;Acceptable values: odd|even|all|<integer in seconds>
start_minute;Start Minute;;Required if "interval" is an integer value.
iterations;Iterations;1;Number of times to repeat this process
=end NX start
=cut

sub start {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;
    my $my_return;

    if ($input{nx_context_inquire}) {
        return 0;
    }

    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   command_interface => {type => ARRAYREF, optional => 1},
                num_nodes         => {type => SCALAR,   optional => 1},
                source_file       => {type => SCALAR,   optional => 0},
                volumes           => {type => ARRAYREF, optional => 0},
                snapshot          => {type => SCALAR,   optional => 1},
                vserver           => {type => SCALAR,   optional => 1},
                destination_file  => {type => SCALAR,   optional => 1},
                number_of_clones =>
                    {type => SCALAR, optional => 1, default => 0},
                source_start_range      => {type => SCALAR, optional => 1},
                destination_start_range => {type => SCALAR, optional => 1},
                num_of_blocks           => {type => SCALAR, optional => 1},
                validate_only =>
                    {type => SCALAR, optional => 1, default => 0},
                interval     => {type => SCALAR, optional => 1},
                start_minute => {type => SCALAR, optional => 1},
                iterations   => {type => SCALAR, optional => 1, default => 1},
            }
        );
        foreach my $key (keys %input) {

            # Check to make sure number_of_clones,
            # source_start_range, num_of_blocks,
            # destination_start_range and number_of_clones
            # are numbers
            # In case of sub file clones make sure triplet
            # is passed
            if (   ($key eq "number_of_clones")
                || ($key eq "source_start_range")
                || ($key eq "num_of_blocks")
                || ($key eq "destination_start_range")) {
                if ($key ne "number_of_clones") {
                    if (   !exists($input{num_of_blocks})
                        || !exists($input{source_start_range})
                        || !exists($input{destination_start_range})) {
                        NATE::BaseException->throw(
                                  " Variables num_of_blocks , "
                                . "source_start_range and destination_start_range"
                                . "are all required if one of them is passed"
                        );
                    }
                }
                if ($input{$key} =~ /^\d+$/) {
                    next;
                }
                NATE::BaseException->throw(
                    "Invalid input($input{$key}) for clone start option <$key>. should be an integer "
                );
            }
            if ($key eq "interval") {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            }
            if ($key eq "iterations") {
                if ($input{$key} !~ /^\d+$/) {
                    NATE::BaseException->throw(
                              "$input{$key} value for interval "
                            . "parameter must be an integer ");
                }
            }
        }
    }
    otherwise {
        $exception = shift;
    };

    # End validating parameters
    # Take care of errors
    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ($input{validate_only}) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "File clone create passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::File_clone::stop_workers";
    }

    # Done taking care of validating parameters and error handling
    # Time to do some useful work
    my @volumes     = @{$input{volumes}};
    my $num_volumes = @volumes;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;
    my $mode;

    # Need to find mode so that we can give the
    # absolute path in 7mode
    $mode = $volumes[0]->command_interface()->mode();

    # Extract number of unique nodes, if we haven't already
    # determined that from a previous call
    if (!$input{num_nodes}) {
        my $node_hash;
        foreach my $v (@volumes) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if (!defined($node_hash->{$name})) {
                push(@nodes, $node);
                $node_hash->{$name} = 1;
            }
            push(@{$node_to_volumes->{$name}}, $v);
        }
        $num_nodes = @nodes;

        foreach my $n (@nodes) {
            NACL::STask::SystemLicense->add(
                command_interface    => $n,
                'package'            => ['flexclone'],
                nacltask_verify      => 1,
                nacltask_if_licensed => "pass",
            );
        }
    } else {
        # Irrelevant
    }

    if (!$input{num_nodes} && ($num_nodes > 1)) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        $my_return->{start_tick} = time();

        foreach my $n (@nodes) {
            my $node_name = $n->name();
            my $id = 'FC_N_DaC.' . $proc_cnt;  # (N)ode_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Indicate we already parsed out unique nodes, no need
            # to do it again
            $input{num_nodes} = $num_nodes;
            $Log->comment("FC_N_DaC"
                    . $proc_cnt
                    . " doling "
                    . @{$node_to_volumes->{$node_name}}
                    . " volumes");
            $input{volumes} = $node_to_volumes->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::File_clone',
                method  => 'start',
                args    => [%input],
            );

            push(@workers,     $worker);
            push(@Pkg_Workers, $worker);
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ($res->{$rid}->{result} != 1) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Created File clones on $num_nodes nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } elsif ($num_volumes > 5) {

        # Multiple volumes on one node
        # Split into batches of 5 volumes
        my $node_name = $volumes[0]->command_interface()->name();
        my $done      = 0;
        my $idx       = 0;
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $had_failure = 0;

        while (!$done) {
            if ($volumes[$idx]) {
                my @cur_vols;
                for (1 .. 5) {
                    if ($volumes[$idx]) {
                        push(@cur_vols, $volumes[$idx]);
                        $idx++;
                    } else {
                        $done = 1;
                        last;
                    }
                }
                my $num_cur_vols = @cur_vols;

                # Work on current batch of volumes
                my $id
                    = 'FC_V_DaC.' . $proc_cnt; # (V)ode_(D)ivide(a)nd(C)onquer
                push(@ids, $id);

                # Indicate we already parsed out unique nodes, no need
                # to do it again
                $input{volumes}   = \@cur_vols;
                $input{num_nodes} = 1;

                my $worker
                    = NATE::PerlRPC::Worker::Process->new(runid => $id);
                my $future = $worker->submit(
                    package => 'NX::File_clone',
                    method  => 'start',
                    args    => [%input],
                );

                push(@workers,     $worker);
                push(@Pkg_Workers, $worker);
                $futures->{$id}->{obj} = $future;
                $proc_cnt++;

                # Collect results
                my $res = collect_worker_results(
                    workers       => \@workers,
                    futures       => $futures,
                    merge_context => 0,
                );

                foreach my $w (@workers) {
                    my $rid = $w->runid();

                    if ($res->{$rid}->{result} != 1) {
                        $Log->error("$rid had a failure");
                        $had_failure = 1;
                    } else {
                        $Log->comment("$rid ran successfully");
                    }
                }
            } else {
                $done = 1;
            }
        }

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Created File Clones on all $num_volumes volumes on $node_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        }
        return $my_return;

    } else {
        $my_return->{result} = 1;
        my @range_triplet;

        # If file does not exist we need to
        # calculate size of file to be created
        # default 1M
        my $source_file_size = 1048576;
        if (exists($input{num_of_blocks})) {
            @range_triplet = (
                $input{source_start_range},
                $input{destination_start_range},
                $input{num_of_blocks}
            );
            $source_file_size
                = ($input{source_start_range} + $input{num_of_blocks} + 1)
                * 4096;
        }
        my $node      = $volumes[0]->command_interface();
        my $node_name = $node->name();
        $input{command_interface} = $node;
        my $node_apiset_obj = $node->get_7m_or_nodescope_apiset();
        $my_return->{start_tick} = time();

        for (my $r = 1; $r <= $input{iterations}; $r++) {
            foreach my $cur_vol (@volumes) {

                # Single node, Max 5 volumes
                my $vol_name = $cur_vol->volume();

                # Get the source file name for node scope
                my $sf = "/vol/$vol_name/$input{source_file}";

                # check if source file exists
                try {
                    $node_apiset_obj->slist("path" => $sf);

                    # Source file exists, remove prior clones
                    for (my $j = 0; $j < $input{number_of_clones}; $j++) {
                        my $prior_file = $sf . "$j";
                        if ($input{iterations} != 1) {
                            $prior_file .= "i$r";
                        }
                        try {
                            $node_apiset_obj->slist("path" => $prior_file);
                            $Log->comment("Removing prior clone: $prior_file");
                            $node_apiset_obj->rm("paths" => $prior_file);
                        } otherwise {
                            # didn't exist
                        };
                    }
                }
                otherwise {
                    try {
                        $node_apiset_obj->filersio_asyncio_active(
                            'read-percentage'    => '0',
                            'rand-percentage'    => '100',
                            'blk-size'           => '4096',
                            'start-byte'         => '0',
                            'end-byte'           => $source_file_size,
                            'run-time'           => 10,
                            'num-active'         => 5,
                            'filename'           => $sf,
                            'connectrec-timeout' => '600',
                            'create'             => '-create',
                        );
                        sleep(10);

                    }
                    otherwise {
                        if ($::IGNORE_ERRORS) {
                            $Log->comment("Ignoring errors, but exception"
                                    . "message while creating source file is: "
                                    . $exception->text());
                        } else {
                            $exception = shift;
                            $exception->throw();

                        }
                    };

                };
                $sf = $input{source_file};
                my $df = $input{source_file};

                #In 7mode File names have to be absolute path
                if ($mode =~ /7Mode/) {
                    if (exists($input{destination_file})) {
                        $df = "/vol/$vol_name/$input{destination_file}";
                    } else {
                        $df = "/vol/$vol_name/$input{source_file}";
                    }
                    $sf = "/vol/$vol_name/$input{source_file}";
                } else {
                    if (exists($input{destination_file})) {
                        $df = "$input{destination_file}";
                    }
                }
                my $orig_df = $df;
                try {
                    if ($input{number_of_clones} > 1) {
                        for (my $j = 0; $j < $input{number_of_clones}; $j++) {
                            $df = "$orig_df" . "$j";
                            if ($input{iterations} != 1) {
                                $df .= "i$r";
                            }
                            my %fc_opts = (
                                command_interface  => $node,
                                volume             => $vol_name,
                                vserver            => $input{vserver},
                                clone_type         => "file",
                                'source-path'      => $sf,
                                'destination-path' => $df
                            );
                            if (exists($input{snapshot})) {
                                $fc_opts{'snapshot-name'} = $input{snapshot};
                            }
                            if (exists($input{num_of_blocks})) {
                                $fc_opts{range} = \@range_triplet;
                            }
                            try {
                                NACL::C::VolumeFileClone->create(%fc_opts);
                            }
                            otherwise {
                                my $e = shift;
                                if ($::IGNORE_ERRORS) {
                                    $Log->comment(
                                              "Ignoring errors, but exception"
                                            . "message is : ".$e->text());
                                } else {
                                    $e->throw($exception);
                                }
                            };
                        }
                    } else {
                        if ($input{iterations} != 1) {
                            $df = "$orig_df" . "i$r";
                        }
                        my %fc_opts = (
                            command_interface  => $node,
                            volume             => $vol_name,
                            vserver            => $input{vserver},
                            clone_type         => "file",
                            'source-path'      => $sf,
                            'destination-path' => $df
                        );
                        if (exists($input{snapshot})) {
                            $fc_opts{'snapshot-name'} = $input{snapshot};
                        }
                        if (exists($input{num_of_blocks})) {
                            $fc_opts{range} = \@range_triplet;
                        }

                        try {
                            NACL::C::VolumeFileClone->create(%fc_opts);
                        }
                        otherwise {
                            my $e = shift;
                            if ($::IGNORE_ERRORS) {
                                $Log->comment("Ignoring errors, but exception"
                                        . "message is : ".$e->text());
                            } else {
                                $e->throw();
                            }
                        };

                    }
                } otherwise {
                    $exception = shift;
                };
                last if $exception;
            }    # end for loop cur_vol
            if (($input{iterations} ne '1') && !$exception
                && ($r ne $input{iterations})) {
                if (exists($input{interval})) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $my_return->{start_tick},
                    );
                }
            }
        }
        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
        }
        return $my_return;

    }

}    ## End create

1;
