# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NX/HA.pm#4 $
#
# Copyright (c) 2012 NetApp
# All rights reserved.
#
## @summary
## HA specified functions to support NACL Translator framework
##
## @status    private
##
## @author Paul Chen (chenp@netapp.com)

package NX::HA;

use strict 'vars';
use Tharn;
use Data::Dumper;

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::MTask::HA;
use NACL::STask::StorageDisk;
use NACL::STask::StorageFailover;
use NX::Helper;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);
use FS::commonParams;

our @ISA    = ('Exporter');
our @EXPORT = qw( takeover_and_giveback );

my $params = NATE::ParamSet->new(global => 1);
my $IGNORE_ERRORS  = $params->get("IGNORE_ERRORS",  default => '0');

my @Pkg_Workers;
sub stop_workers {
    $Log->comment("Detected interrupt signal");
    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}

=pod

=head1 NAME

NX::HA

=head1 DESCRIPTION

HA related routines for use in NX framework;
This library performs takeover and auto-giveback. Please use caution when
assigning test intervals.
For SN.0, auto-giveback is limited to 2 attempts for every 60 minutes.

=head1 METHODS
=head2 takeover_and_giveback

Perform takeover and giveback on specified nodes.

The input is a hash (%opts) which must include the command_interface:
    src_nodes      => <array ref of source nodes>
    dst_nodes      => <array ref of destination nodes>

=head3 opts

%opts can also be used to override default values, which include
    iterations     => <integer>|'infinite' # Number of cycles for failover.
    interval       => odd/even/all/<value in seconds>
                       # Every odd, even, or every clock hour,
                       # or specified value in seconds
    start_minute   => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
                       a two-digit value representing the clock minute
    action_timeout => 500
    override-vetoes => 'true|false' # Override vetoes during giveback
                                      Default is 'true'
    option         => 'normal|immediate' # Option for takeover.
                                             Default is 'normal'.
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX takeover_and_giveback
Takeover and Giveback
description
Run takeover and giveback HA operations
required
src_nodes;Source nodes;;name of source nodes, currently supports only one. Syntax is $N<#> e.g. $N1
dst_nodes;Destination nodes;;name of destination nodes, currently supports only one. Syntax is $N<#> e.g. $N2
optional
iterations;Test Iterations;1;number of cycles for failover and giveback. Syntax is <integer> or 'infinite'
interval;Test Interval;3600;interval between failovers, can be every odd, even, or every clock hour, or specified value in seconds. Syntax is 'odd' or 'even' or 'all' or <value in seconds>
start_minute;Start Minute;;Required if 'interval' is odd|eve|all, otherwise ignored if 'interval' is an interger value. This is a two-digit value representing the clock minute. Syntax is <\d\d> e.g 10 or 20.
action_timeout;Action Timeout;500;Timeout value in seconds for checking failover and giveback.
option; HA Option;normal;Options for failover and giveback. Supports 'normal' and 'immediate'
=end NX takeover_and_giveback
=cut

sub takeover_and_giveback {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;
    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return; #Hash reference containing return values
    my $exception;
    my ( @src_nodes, @dst_nodes, $num_dst_nodes, $num_src_nodes );
    $num_dst_nodes = @{$input{dst_nodes}};
    $num_src_nodes = @{$input{src_nodes}};
    my $dst_node = ${$input{dst_nodes}}[0];
    my $src_node = ${$input{src_nodes}}[0];

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validate_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                dst_nodes      => {type => ARRAYREF, optional => 0},
                src_nodes      => {type => ARRAYREF, optional => 0},
                iterations     => {type => SCALAR,   optional => 1, default => '1'},
                interval       => {type => SCALAR,   optional => 1, default => '3600'},
                start_minute   => {type => SCALAR,   optional => 1 },
                option         => {type => SCALAR,   optional => 1, default => 'normal'},
                action_timeout => {type => SCALAR,   optional => 1, default => 500},
                'override-vetoes' => {type => SCALAR,   optional => 1, default => 'true'},
                validate_only  => {type => SCALAR,   optional => 1, default => 0},
            }
        );

        if ($num_dst_nodes != 1 ){
            NATE::BaseException->throw("Unable to perform CFO/SFO with $num_dst_nodes destination nodes! Currently supports cluster with 2 nodes, 1 destination and 1 source.");
        } elsif ($num_src_nodes != 1 ){
            NATE::BaseException->throw("Unable to perform CFO/SFO with $num_src_nodes source nodes! Currently supports cluster with 2 nodes, 1 destination and 1 source.");
        }

        for my $key (keys %input){
            $Log->comment("key: |$key|\nvalue: | $input{$key}|\n");
            if ( ($key eq 'iterations') && ($input{$key} > 1) && !defined($input{'interval'}) ) {
                NATE::BaseException->throw("\'iterations\' parameter requires a value "
                        . "for the \'interval\' parameter");
            } elsif ( ($key eq 'iterations') && ($input{$key} ne 'infinite') && ($input{$key} !~ /\d+/) ) {
                NATE::BaseException->throw("Invalid input ($input{$key}) for HA takeover option <$key>. Accepted values: 'infinite' or integer.");
            } elsif ( $key eq 'interval' ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
                # This library utilizes auto-giveback, need to make sure interval
                # given in seconds is > ( attempts_time  * 60 / attempts )
                # The scale for attempts_time is minute.
                my $ha_cs = NACL::CS::StorageFailover->fetch(
                            command_interface => $dst_node,
                            filter            => {
                                node => $dst_node->node() },
                            requested_fields  => [
                                qw(attempts attempts-time enabled) ],
                );
                my $interval_allow =  $ha_cs->{attempts_time} * 60 / $ha_cs->{attempts} ;
                if ( $input{'interval'} =~ /\d+/ && $input{'interval'} <= $interval_allow ) {
                    NATE::BaseException->throw("Invalid input ($input{$key}) for HA takeover option <$key>. It has to be more than $interval_allow seconds to avoid hitting auto-giveback limits");
                }
            } elsif ( $key eq 'option' ) {
                if ($input{$key} ne 'normal' && $input{$key} ne 'immediate'){
                NATE::BaseException->throw("Invalid input ($input{$key}) for HA takeover option <$key>. Accepted values: normal, immediate");
                }
            } elsif ( $key eq 'action_timeout' && $input{$key} !~ /(\d+)/) {
                NATE::BaseException->throw("Invalid input ($input{$key}) for HA takeover option <$key>. Accept numbers only");
            } elsif ( $key eq 'override-vetoes' && $input{$key} !~ /^(true|false)$/) {
                NATE::BaseException->throw("Invalid input ($input{$key}) for HA takeover option <$key>. Accept values: true, false");
            } else {
                next;
            }
        }

        if ( $input{validate_only} ) {
            my @nodes = ( $src_node, $dst_node );
            for my $n ( @nodes ) {
                my %mod_opts = (command_interface => $n, node => $n->node());
                my $subr = sub { my %new_opt = @_;
                    NACL::C::StorageFailover->modify(%mod_opts, %new_opt);
                };
                &$subr('onreboot' => 'true');
                &$subr('auto-giveback' => 'true');
                &$subr('auto-giveback-override-vetoes' => $input{'override-vetoes'});
            }

            $Log->comment("Check if HA is enabled");
            my $ha_cs = NACL::CS::StorageFailover->fetch(
                        command_interface => $src_node,
                        filter            => {
                            node => $src_node->node() },
                        requested_fields  => [
                            qw(attempts attempts-time enabled) ],
            );
            if ( $ha_cs->{enabled} ne 'true' ) {
                # Enable HA on both nodes if not enabled yet
                NACL::STask::StorageFailover->enable(
                    command_interface => $src_node,
                    node              => $src_node,
                    nacltask_modify_mode_option => 1,
                );
            }

            # Check for disk cross connection
            NACL::STask::StorageDisk->check_cross_connect(
                command_interface       => $dst_node,
                node1                   => $dst_node->node(),
                node2                   => $src_node->node(),
            );

            # Make sure the node pair is UP
            NACL::STask::StorageFailover->wait_for_state(
                command_interface => $dst_node,
                node              => $dst_node,
                partner           => $src_node,
                node_state        => 'up',
                partner_state     => 'up',
                'method-timeout'  => '1800',
            );
        }
    } # end try
    otherwise {
        $exception = shift;
    }; #end otherwise

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "HA takeover action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::HA::stop_workers";
    }

    # Store start tick for return later
    my $start_tick = time();
    $my_return->{start_tick} = $start_tick;
    my $iteration = 1;
    my $done = 0;

    # Start failover test
    while(!$done) {
        try {
            # Don't perform HA if wafliron is in progress
            my $proceed = 0;
            while (!$proceed) {
                $src_node->refresh_command_interface( no_version_change => 1 );
                my $running = 0;
                my $api = $src_node->apiset();
                my $parsed = $api->storage_aggregate_wafliron_show()->get_parsed_output();
        
                foreach my $entry (@$parsed) {
                    if ($entry->{'iron-status'} !~ /not_running/i) {
                        $Log->comment("Wafliron is running on ".$entry->{aggregate});
                        $running = 1;
                        last;
                    }
                }
        
                if (!$running) {
                    $Log->comment("Wafliron is NOT running");
                    $proceed = 1;
                } else {
                    $Log->comment("Wafliron is running on node "
                        . $src_node->node() . ", checking again in 5 minutes");
                    sleep 300;
                }
            }

            $Log->comment("Starting takeover and giveback test iteration $iteration");
            # Create a MTask::HA obj for failover
            my $failover_obj = NACL::MTask::HA->new( node => $dst_node );

            # Start takeover
            $Log->comment("Start takeover action");
            $failover_obj->takeover(
                option                            => $input{'option'},
                method_timeout                    => $input{'action_timeout'},
                nacltask_verify_partner_sfo_state => '1',
            );
            $Log->comment("Success at takeover");

            $Log->comment("Start giveback action");
            $failover_obj->giveback(
                method_timeout            => $input{'action_timeout'},
                nacltask_verify_sfo_state => '1',
                auto_giveback_expected    => '1',
                'override-vetoes'         => $input{'override-vetoes'},
            );
            $Log->comment("Success at giveback");

            # Check if iteration has met requirement
            if ($input{'iterations'} ne 'infinite' && $iteration >= $input{'iterations'}){
                $done = 1;
            } else {
                $iteration++;
                wait_for_interval(
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                    start_tick    => $start_tick,
                );
            }
        } ## end try
        otherwise {
            $Log->comment("Perform a failover show-takeover for debugging purpose");
            NACL::C::StorageFailover->show_takeover(command_interface => $dst_node);
            $Log->comment("Perform a failover show-giveback for debugging purpose");
            NACL::C::StorageFailover->show_giveback(command_interface => $dst_node);
            my $e = shift;
            if (($input{'override-vetoes'} eq 'true') || ($IGNORE_ERRORS)) {
                if ($input{'override-vetoes'} eq 'true') {
                    $Log->comment("Override veto set to true, ignoring error: "
                        . $e->text());
                } else {
                    $Log->warn("Ignoring caught exception. Error was: " . $e->text());
                }
                sleep 600; # Let cluster stablize
                if ($input{'iterations'} ne 'infinite' && $iteration >= $input{'iterations'}){
                    $done = 1;
                } else {
                    $iteration++;
                    wait_for_interval(
                        interval      => $input{interval},
                        start_minute  => $input{start_minute},
                        start_tick    => $start_tick,
                    );
                }
            } else {
                $Log->error("Caught exception. Error was: " . $e->text());
                $exception = $e;
            }
        }; ## end otherwise
    } ## end while

    $my_return->{end_tick} = time();
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
    } else {
        $Log->comment("Completed $iteration iterations of takeover and giveback"
                      . " on ".$dst_node->node() ." and ". $src_node->node());
        $my_return->{result} = 1;
    }
    return $my_return;
} ## end takeover_and_giveback
1;
