# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NX/Helper.pm#6 $
#
# Copyright (c) 2012-2013 NetApp
# All rights reserved.
#
## @summary
## Helper functions for NACL XML translation framework.
##
## @status    private
##

package NX::Helper; 

use strict 'vars';
use Tharn;
use NACL::ComponentUtils;

use Data::Dumper;
use XML::Smart;
use File::Copy;
use NACL::C::JobScheduleCron;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
    parse_input_file    nx_log                      calc_duration
    merge_context       collect_worker_results      wait_for_interval
    create_cron_sched    _nx_stop_workers
);

sub _nx_stop_workers {
    my (%input) = @_;
    my @workers = @{$input{workers}};
    $Log->comment("Stopping all outstanding workers");
    if ( $workers[0] ) {
        foreach my $w ( @workers ) {
            $Log->comment("Stopping worker ".$w->runid());
            $w->stop();
        }
        while ( my $w = shift(@workers) ) {
            while ( $w->is_busy() ) {
                $Log->comment($w->runid()." still busy");
                sleep 1;
            }
            $Log->comment("Worker ".$w->runid()." no longer running");
        }
    }
    $Log->comment("No more workers to kill");
    return;
}

=pod

=head1 NAME

NX::Helper

=head1 DESCRIPTION

Helper routines used by the NX tool.

=head1 METHODS

=head2 parse_input_file

Parse input file and return a nested hash reference containing the contents.

The input is a hash (%opts) which must include:
    path        => $input_file_path

Example:   

my $struct = parse_input_file( path => $Path );

=over

=back

=cut
sub parse_input_file {
    my $parsed;
    my $contents;
	my %opts = validate(
        @_,
        {
		 	path         => { type => SCALAR,   optional => 0 },
		}
	);

    if ( !-e $opts{path} ) {
        NATE::BaseException->throw("Input file $opts{path} does not exist");
    } elsif ( !open(IFILE, $opts{path}) ) {
        NATE::BaseException->throw("Could not open input file $opts{path}");
    } else { # File exists and open successful
        close(IFILE);
        # Make a copy of the input file for debug
        my $dst_dir = $::LOGDIR;
        $dst_dir .= '/' if $dst_dir; # Prevent trying to copy file to / if
                                     # ntest run with LOGDIR=-
        copy( $opts{path}, $dst_dir."orig_input.xml");
        if ( $opts{path} =~ /\.xml/ ) {
            $contents = XML::Smart->new($opts{path}, 'XML::Smart::Parser');
        } else {
            NATE::BaseException->throw("Input file $opts{path} must have xml file extension");
        }
    }

    my $root = $contents->{scenario};
    if ( !$root ) {
        NATE::BaseException->throw("Input file $opts{path} is missing "
            . "root <scenario> tag");
    }
    my @sections = $root->order();

    my $sec_cnt = 0;
    foreach my $sec ( @sections ) {
        $Log->comment("SECTION: *$sec*");
        my @sub_sections = $root->{$sec}->order();
        if ( $#sub_sections < 1 ) {
            $Log->comment("<1 sub sections for $sec, using nodes_keys sub");
            @sub_sections = $root->{$sec}->nodes_keys();
        } else {
            $Log->comment(">1 sub sections ".$#sub_sections." for $sec");
        }
        $Log->comment("---->> ALL sub sections: @sub_sections");

        if ( $sec eq 'summary') {
            $parsed->{$sec} = $root->{$sec}->content();
            next;
        }

        my $sub_sec_cnt = 0;
        my $exec_order_cnt = 0;
        my $cfg_exec_order_cnt = 0;
        my $tag_idx;
        my $cfg_tag_idx;
        my $p_idx = 0;
        my $np_sub_sec_action_name_cnt;
        my $p_sub_sec_action_name_cnt;
        my $cfg_sub_sec_tag_name_cnt;
        foreach my $sub_sec ( @sub_sections ){
            $Log->comment("SUBSECTION: *$sub_sec*");
            if ( $sec eq 'config' ) {
                if ( $sub_sec ne 'cluster' ) {
                    my @cfg_sub_sec_params = $root->{$sec}->{$sub_sec}->order();
                    if ( $#cfg_sub_sec_params < 1 ) {
                        $Log->comment("<1 cfg sub sec params for $sub_sec, using "
                            . "nodes_keys sub");
                        @cfg_sub_sec_params = $root->{$sec}->{$sub_sec}->nodes_keys();
                    } else {
                        $Log->comment(">1 cfg sub sections ".$#sub_sections
                            ." for $sub_sec");
                    }
                    foreach my $sub_sec_key ( @cfg_sub_sec_params ) {
                        $Log->comment("CFG SUB SEC KEY: $sub_sec_key");
                        $parsed->{$sec}->{$sub_sec}->{$sub_sec_key} =
                            $root->{$sec}->{$sub_sec}->{$sub_sec_key}->content();
                        $Log->comment("  ".$parsed->{$sec}->{$sub_sec}->{$sub_sec_key});
                    }
                } else {
                    my $cur_cfg_idx = 0;
                    if ( $cfg_sub_sec_tag_name_cnt->{$sub_sec} ) {
                        $cur_cfg_idx = $cfg_sub_sec_tag_name_cnt->{$sub_sec};
                        $cfg_sub_sec_tag_name_cnt->{$sub_sec}++;
                    } else {
                        $cfg_sub_sec_tag_name_cnt->{$sub_sec} = 1;
                    }
    
    
                    my $tag_name;
                    my $t_names = $root->{$sec}->{$sub_sec}->[$cur_cfg_idx];
                    foreach my $n (keys %{$t_names} ) {
                        $tag_name = $n;
                        $Log->comment("CFG TAG NAME: *$tag_name*");
        
                        my @tag_args = $root->{$sec}->{$sub_sec}->[$cur_cfg_idx]->{$tag_name}->order();
                        if ( $#tag_args < 1 ) {
                            $Log->comment("<1 args for config tag "
                                . "$tag_name, using nodes_keys sub");
                            @tag_args = $root->{$sec}->{$sub_sec}->[$cur_cfg_idx]->{$tag_name}->nodes_keys();
                        } else {
                            $Log->comment(">1 args for config tag "
                                . "$tag_name ".$#tag_args);
                        }
                        if ( !defined($cfg_tag_idx->{$tag_name}) ) {
                            $cfg_tag_idx->{$tag_name} = 1;
                        } else {
                            $cfg_tag_idx->{$tag_name}++;
                        }
        
                        if ( @tag_args < 1 ) {
                            NATE::BaseException->throw($sub_sec.'::'.$tag_name
                                . " does not seem like it's a valid config tag arg");
                        } else {
                            foreach my $t_arg ( @tag_args ) {
                                $Log->comment("CFG TAG ARG: $t_arg");
                                $Log->comment("  ARG VALUE: ".
                                    $root->{$sec}->{$sub_sec}->[$cur_cfg_idx]->{$tag_name}->{$t_arg}->content());
                                $parsed->{$sec}->{$sub_sec}->{$tag_name}->{$t_arg} =
                                    $root->{$sec}->{$sub_sec}->[$cur_cfg_idx]->{$tag_name}->{$t_arg}->content();
                            }
                        }
                    }
                }





            } elsif ( $sec eq 'post_process' ) {
                my @pp_sub_sec_params = $root->{$sec}->{$sub_sec}->order();
                if ( $#pp_sub_sec_params < 1 ) {
                    $Log->comment("<1 pp sub sec params for $sub_sec, using "
                        . "nodes_keys sub");
                    @pp_sub_sec_params = $root->{$sec}->{$sub_sec}->nodes_keys();
                } else {
                    $Log->comment(">1 pp sub sections ".$#sub_sections
                        ." for $sub_sec");
                }
                foreach my $sub_sec_key ( @pp_sub_sec_params ) {
                    $Log->comment("PP SUB SEC KEY: $sub_sec_key");
                    $parsed->{$sec}->{$sub_sec}->{$sub_sec_key} =
                        $root->{$sec}->{$sub_sec}->{$sub_sec_key}->content();
                    $Log->comment("  ".$parsed->{$sec}->{$sub_sec}->{$sub_sec_key});
                }
            } elsif ( $sec eq 'actions' ) {
                if ( $sub_sec ne 'parallel' ) {
                    # $sub_sec is an action type
                    my $cur_sub_sec_idx = 0;
                    if ( $np_sub_sec_action_name_cnt->{$sub_sec} ) {
                        $cur_sub_sec_idx = $np_sub_sec_action_name_cnt->{$sub_sec};
                        $np_sub_sec_action_name_cnt->{$sub_sec}++;
                    } else {
                        $np_sub_sec_action_name_cnt->{$sub_sec} = 1;
                    }

                    if ( $sub_sec eq 'sleep' ) {
                        $parsed->{$sec}->[$exec_order_cnt]->[0]->{$sub_sec} =
                            $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->content();
                    } else {
                        my $action_name;
                        my $a_names = $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx];
                        foreach my $n (keys %{$a_names} ) {
                            $action_name = $n;
                        }
                        $Log->comment("ACTION NAME: *$action_name*");

                        my @action_args = $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->order();
                        if ( $#action_args < 1 ) {
                            $Log->comment("<1 action args for action "
                                . "$action_name, using nodes_keys sub");
                            @action_args = $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->nodes_keys();
                        } else {
                            $Log->comment(">1 action args for action "
                                . "$action_name ".$#action_args);
                        }
                        if ( !defined($tag_idx->{$action_name}) ) {
                            $tag_idx->{$action_name} = 1;
                        } else {
                            $tag_idx->{$action_name}++;
                        }
    
                        if ( @action_args < 1 ) {
                            NATE::BaseException->throw($sub_sec.'::'.$action_name
                                . " does not seem like it's a valid action");
                        }

                        my $kv_cnt = 0;
                        my $elem_cnt = 0;
                        foreach my $a_arg ( @action_args ) {
                            $Log->comment("ACTION ARG: $a_arg");
                            if ( $a_arg eq 'key_value' ) {
                                print "\nKEY VALUE: *".$root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->[$kv_cnt]->content()."*\n";
                                $parsed->{$sec}->[$exec_order_cnt]->[0]->{$sub_sec.'::'.$action_name}->{$a_arg}->[$kv_cnt] =
                                    $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->[$kv_cnt]->content();
                                $kv_cnt++;
                            } elsif ( $a_arg eq 'list_element') {
                                print "\nLIST ELEMENT: *".$root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->[$elem_cnt]->content()."*\n";
                                $parsed->{$sec}->[$exec_order_cnt]->[0]->{$sub_sec.'::'.$action_name}->{$a_arg}->[$elem_cnt] =
                                    $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->[$elem_cnt]->content();
                                $elem_cnt++;
                            } else {
                                $Log->comment("  ACTION ARG VALUE: ".
                                    $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->content());
                                $parsed->{$sec}->[$exec_order_cnt]->[0]->{$sub_sec.'::'.$action_name}->{$a_arg} =
                                    $root->{$sec}->{$sub_sec}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->content();
                            }
                        }
                    }
                } else { 
                    # $sub_sec is parallel
                    # Get all action info for the parallel contents at corresponding parallel idx
                    my $p_tag_idx;

                    # Get action types at parallel $p_idx
                    my @action_types = $root->{$sec}->{parallel}->[$p_idx]->order();
                    $Log->comment("!Num Action types => ".@action_types);
                    if ( $#action_types < 1 ) {
                        $Log->comment("<1 action types for parallel idx $p_idx, using nodes_keys sub");
                        @action_types = $root->{$sec}->{parallel}->[$p_idx]->nodes_keys();
                        $Log->comment("!Num Action types NODE_KEYS => ".@action_types);
                    } else {
                        $Log->comment(">1 action types for parallel idx $p_idx: ".$#action_types);
                    }

                    my $action_order_idx = 0;
                    my $cur_action_type_idx;#xxx
                    foreach my $a_type ( @action_types ) {
                        if ( !defined($cur_action_type_idx->{$a_type}) ) {
                            $cur_action_type_idx->{$a_type} = 0;
                        } else {
                            $cur_action_type_idx->{$a_type}++;
                        }
                        my $cur_sub_sec_idx = $cur_action_type_idx->{$a_type};

                        my $action_name;
                        my $a_names = $root->{$sec}->{parallel}->[$p_idx]->{$a_type}->[$cur_sub_sec_idx];
                        foreach my $n (keys %{$a_names} ) {
                            $action_name = $n;
                        }
                        $Log->comment("PARALLEL ACTION TYPE/NAME: "
                            . "*$a_type/$action_name*");

                        my @action_args = $root->{$sec}->{parallel}->[$p_idx]->{$a_type}->[$cur_sub_sec_idx]->{$action_name}->order();
                        if ( $#action_args < 1 ) {
                            $Log->comment("<1 action args for action "
                                . "$action_name, using nodes_keys sub");
                            @action_args = $root->{$sec}->{parallel}->[$p_idx]->{$a_type}->[$cur_sub_sec_idx]->{$action_name}->nodes_keys();
                        } else {
                            $Log->comment(">1 action args for action "
                                . "$action_name ".$#action_args);
                        }
                        if ( !defined($p_tag_idx->{$action_name}) ) {
                            $p_tag_idx->{$action_name} = 1;
                        } else {
                            $p_tag_idx->{$action_name}++;
                        }

                        if ( ( @action_args < 1 ) && ( $a_type ne 'sleep') ) {
                            NATE::BaseException->throw($a_type.'::'.$action_name
                                . " does not seem like it's a valid action");
                        }

                        if ( $a_type eq 'sleep' ) {
                            $parsed->{$sec}->[$exec_order_cnt]->[$action_order_idx]->{$a_type} =
                                $root->{$sec}->{parallel}->[$p_idx]->{$a_type}->[$cur_sub_sec_idx]->content();
                        } else {
                            foreach my $a_arg ( @action_args ) {
                                $Log->comment("ACTION ARG: $a_arg");
                                $Log->comment("  ACTION ARG VALUE: ".
                                    $root->{$sec}->{parallel}->[$p_idx]->{$a_type}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->content());
                                $parsed->{$sec}->[$exec_order_cnt]->[$action_order_idx]->{$a_type.'::'.$action_name}->{$a_arg} =
                                    $root->{$sec}->{parallel}->[$p_idx]->{$a_type}->[$cur_sub_sec_idx]->{$action_name}->{$a_arg}->content();
                            }
                        }
                        $action_order_idx++;
                    }
                    
                    $p_idx++;
                }
                $exec_order_cnt++;
            }
            $sub_sec_cnt++;
        }
        $sec_cnt++;
    }

    $Log->comment(">>> Dump of post-processed XML:\n" . Dumper($parsed));

	return $parsed;
} ## end parse_input_file


sub nx_log {
    my $msg = shift;
    $Log->comment(
        "\n###############################################################".
        "\n$msg".
        "\n###############################################################\n");
    return 1;
}

sub calc_duration {
    my $tick = shift;
    return $tick if ( $tick !~ /^\d+$/ );
    
    my $d = '00';
    my $h = '00';
    my $m = '00';
    my $s = '00';
    
    $d = int($tick / 86400);
    $d = '0'.$d if ($d < 10);
    my $mod = $tick % 86400;
    if ($mod)
    {
        $h = int($mod / 3600);
        $h = '0'.$h if ($h < 10);
        $mod = $mod % 3600;
        if ($mod)
        {
            $m = int($mod / 60);
            $m = '0'.$m if ($m < 10);
            $mod = $mod % 60;
            if ($mod)
            {
                $s = $mod;
                $s = '0'.$s if ($s < 10);
            }
        }
    }
    
    return $d.'d:'.$h.'h:'.$m.'m:'.$s.'s';
}

sub merge_context {
    my (%input) = @_;
    my $orig = $input{orig};
    my $new  = $input{new};

    my @new_values;
    _get_context_data(
        junction => $new,
        change_array => \@new_values,
    );

    foreach my $c ( @new_values ) {
        my @arr = split(/:/, $c);
        my $for_eval;
        my $last = pop(@arr);
        foreach (@arr) {
            $for_eval .= "->{\'$_\'}";
        }
        $for_eval = '$orig' . $for_eval;
        $for_eval .= " = \'$last\'";
        $Log->comment("Merging context: $for_eval");
        eval ( $for_eval );
    }
    return $orig;
}


# returns hash reference containing worker results:
#   $hash->{workers}->{runid}->... # where is all the designated return data
sub collect_worker_results {
    my %input = validate(
        @_,
        {
            workers         => { type => ARRAYREF,  optional => 0 },
            futures         => { type => HASHREF,   optional => 0 },
            context         => { type => HASHREF,   optional => 1, default => undef },
            merge_context   => { type => SCALAR,    optional => 1, default => 0 },
            sample          => { type => SCALAR,    optional => 1, default => 0 },
        }
    );

    my @workers = @{$input{workers}};
    my $futures = $input{futures};
    my $context = $input{context};
    my $merge   = $input{merge_context};
    my $all_results;
    my @completed_workers;

    my $busy_timers = {
        1 => 1,     2 => 5,
        3 => 10,    4 => 30,
        5 => 60,    6, 300,
        7 => 600,
    };
    my $b_max = 7;

    # Poll for worker status round-robin style
    if ( !$input{sample} ) {
        nx_log("Waiting for all workers to complete");
    }
    my $b_cnt = 1;
    my $s_tick = time();
    my $num_done = 0;
    my $num_workers = @workers;
    my $status;
    my $last_busy_log_tick = $s_tick;
    while ( $num_done != $num_workers ) {
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            next if $status->{$rid};
            if ( !$w->is_busy() ) {
                $status->{$rid} = 1;
                $Log->comment("\'$rid\' is done");
                $num_done++;
                push(@completed_workers, $w);
            }
        }

        if ( $input{sample} ) {
            last;
        } else {
            my $c_tick = time();
            if ( ($c_tick - $last_busy_log_tick) >= $busy_timers->{$b_cnt} ) {
                $Log->comment("$num_done/$num_workers workers done, waiting...");
                $b_cnt++ if ($b_cnt < $b_max);
                $last_busy_log_tick = time();
            }
            sleep 1;
        }
    }

    if ( ( $num_done == $num_workers ) && !$input{sample} ) {
        nx_log("All workers completed. Retrieving results");
    }

    # Retrieve results
    foreach my $w ( @completed_workers ) {
        my $rid = $w->runid();
        my $cur_tick = time();
        my $start_tick;
        my $end_tick;
        my $res;

        $Log->comment("Getting worker results for \'$rid\'");
        try {
            $res = $futures->{$rid}->{obj}->result();
            foreach my $k (keys %{$res} ) {
                $all_results->{$rid}->{$k} = $res->{$k};
                $Log->comment("   $k => $res->{$k}");
            }

            # Merge updated context if desired
            if ( $merge && $res->{context} && $context ) {
                merge_context(
                    orig => $context,
                    new  => $res->{context},
                );
            }
        } otherwise {
            my $e = shift;
            my $msg = $e->text();
            $Log->comment("Exception string from \'$rid\': *$msg*");
            $all_results->{$rid}->{result} = -1;
            $all_results->{$rid}->{exception} = $e;
        };

        my $s_tick = $res->{start_tick};
        my $e_tick = $res->{end_tick};
        if ( defined($s_tick) && defined($e_tick) ) {
            $all_results->{$rid}->{duration} = $e_tick - $s_tick;
            $res->{duration} = $e_tick - $s_tick;
            $Log->comment("   duration => ".($e_tick - $s_tick)."s");
        } else {
            $all_results->{$rid}->{duration} = "n/a";
        }
        $w->stop();
    }
    return $all_results;
}

=pod

=head2 wait_for_interval

Sleeps until next designated interval time is hit and returns back to the caller

The input is a hash (%opts) which must include:
    interval        => odd/even/all/<value in seconds>
                       # Every odd, even, or every clock hour,
                       # or specified value in seconds

%opts can also be used to override default values, which include
    start_minute    => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
                       a two-digit value representing the clock minute
    start_tick      => Required if "interval" is an integer.  This is the
                       starting tick value
    validate_only   => 0/1 (default is 0).  Set to 1 to validate
                       "interval" and "start_minute" parameter values

Examples:

# Wait until next even hour at 15 minute marker (e.g. 2:15)
    wait_for_interval( interval => "even", start_minute => 15 );

# Wait until next odd hour at 15 minute marker (e.g. 3:15)
    wait_for_interval( interval => "odd", start_minute => 15 );

# Wait until next 15 minute marker (e.g. *:15)
    wait_for_interval( interval => "all", start_minute => 15 );

# Wait until next tick value that is a multiple of 15 seconds from
# specified starting tick value and current tick value. (e.g. if
# start tick equates to 12:00:00, and current tick equates to 02:00:00,
# then next multiple will be at 02:00:15
    wait_for_interval( start_tick => $start_tick, interval => 15 );

=over

=back

=cut
sub wait_for_interval {
    my (%input) = @_;
    my $start_tick = $input{start_tick};
    my $interval = $input{interval};
    my $minute = $input{start_minute};
    my $validate = $input{validate_only};

    if ($validate) {
        if (!defined($interval)) {
            NATE::BaseException->throw("\'interval\' parameter is required");
        } elsif ( ($interval !~ /^\d+$/) && !defined($minute) ) {
            NATE::BaseException->throw("\'start_minute\' parameter is required");
        } elsif ( $interval !~ /^(odd|even|all|\d+)$/ ) {
            NATE::BaseException->throw("\'interval\' parameter value must be "
                . "\'odd\', \'even\', \'all\', or an integer value in seconds");
        } elsif ( defined($minute) && ($minute !~ /^[0-5]\d$/) ) {
            NATE::BaseException->throw("\'start_minute\' parameter value must "
                . "be a two-digit minute value");
        }
    } else {
        if ( ($interval =~ /^\d+$/) && !$start_tick ) {
            NATE::BaseException->throw("\'start_tick\' parameter is required");
        }
        my $ready = 0;
        my $clock;
        while (!$ready) {
            sleep 1;
            my $cur_tick = time();
            my @tmp = split(/\s+/, localtime($cur_tick));
            $clock = $tmp[3];
            $clock =~ /(\d+):(\d+):(\d+)/;
            my $cur_hour = $1;
            my $cur_min = $2;
            my $cur_sec = $3;

            if ( $interval eq 'odd' ) {
                if ( ($cur_hour % 2) && ($cur_min eq $minute) ) {
                    $ready = 1;
                }
            } elsif ($interval eq 'even' ) {
                if ( (($cur_hour % 2) == 0) && ($cur_min eq $minute) ) {
                    $ready = 1;
                }
            } elsif ( $interval eq 'all' ) {
                if ( $cur_min eq $minute ) {
                    $ready = 1;
                }
            } else { # interval is an integer
                if ( (($cur_tick - $start_tick) % $interval ) == 0 ) {
                    $ready = 1;
                }
            }
        }
        $Log->comment("Hit next interval at $clock");
    }
    return 1;
}


=pod

=head2 create_cron_sched

Creates a cron schedule with an intuitive and legible name (if possible)
and returns the name back to the caller.

The input is a hash (%opts) which must include:
    command_interface => Node object
    hours   => One or more comma separated hour values for the schedule
               (e.g. "1,2,3"), or the string "all", which equates to
               every hour.
    minutes => One or more comma separated minute values for the schedule
               (e.g. "0,30"), or the string "all", which equates to
               makes the function look strictly at the "hours" argument.

Examples:

# A schedule named "every_10_min" is created:
    create_cron_sched(
        command_interface => $node,
        hours             => "all",
        minutes           => "0,10,20,30,40,50"
    );

# A schedule named "hourly" is created:
    create_cron_sched(
        command_interface => $node,
        hours             => "all",
        minutes           => "30"
    );

# A schedule named "every_2_hour" is created:
    create_cron_sched(
        hours   => "0,2,4,6,8,10,12,14,16,18,20,22",
        minutes => "all"
    );

# A schedule with the format "nx_sched_<time_tick>" is created
# if there is no discernable and repeatable pattern in the "hours"
# and "minutes" arguments.

=over

=back

=cut
sub create_cron_sched {
    my %input = validate(
        @_,
        {
            command_interface => { type => OBJECT, optional => 0 },
            hours             => { type => SCALAR, optional => 0 },
            minutes           => { type => SCALAR, optional => 0 },
        }
    );

    # Create schedule and use this schedule
    my $sched_name = 'nx_sched_'.time();
    my $sched_exists = 0;
    if ( $input{hours} eq 'all' ) {
        my @arr = split(/,/, $input{minutes});
        if (@arr == 1) {
            $sched_name = 'hourly';
        } else {
            my $every_interval = 1;
            my $interval_val = $arr[1] - $arr[0];
            for(my $i = 0; $i < $#arr; $i++) {
                if (($arr[$i+1] - $arr[$i]) != $interval_val) {
                    $every_interval = 0;
                    last;
                }
            }
            if ($every_interval
                && ((60 % $interval_val) == 0)
                && (60/$interval_val == @arr)) {
                $sched_name = 'every_'.$interval_val."_min";
            } else {
                $sched_name = join('_', @arr).'_min';
            }
        }
    } elsif (($input{minutes} eq 'all') || ($input{minutes} !~ /,/)) {
        my @arr = split(/,/, $input{hours});
        if (@arr == 1) {
            $sched_name = $arr[0].'_hour';
        } else {
            my $every_interval = 1;
            my $interval_val = $arr[1] - $arr[0];
            for(my $i = 0; $i < $#arr; $i++) {
                if (($arr[$i+1] - $arr[$i]) != $interval_val) {
                    $every_interval = 0;
                    last;
                }
            }
            if ($every_interval
                && ((24 % $interval_val) == 0)
                && (24/$interval_val == @arr)) {
                $sched_name = 'every_'.$interval_val."_hour";
            } else {
                $sched_name = join('_', @arr).'_hour';
            }
        }
    }

    # Check to see if sched name already exists
    $sched_exists = NACL::C::JobScheduleCron->find(
        command_interface => $input{command_interface},
        filter            => { name => $sched_name },
        allow_empty       => 1,
    );

    if (!$sched_exists) {
        NACL::C::JobScheduleCron->create(
            command_interface   => $input{command_interface},
            name                => $sched_name,
            hour                => [ $input{hours} ],
            minute              => [ $input{minutes} ],
        );
    } else {
        $Log->comment("Schedule $sched_name already exists");
    }
    return $sched_name;
} # End create_cron_sched


sub _get_context_data {
    my (%input) = @_;
    my $junc = $input{junction};
    my $path = $input{path};
    my $chg_arr = $input{change_array};

    my $type = ref($junc);
    my $val;

    if ( $type eq 'HASH' ) {
        foreach my $k (keys %{$junc} ) {
            $val = _get_context_data(
                junction => $junc->{$k},
                path     => $path.':'.$k,
                change_array => $chg_arr
            );
        }
    } else {
        $path =~ s/^://;
        push(@{$chg_arr}, $path.':'.$junc);
        return $junc;
    }
}


1;