# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NX/IO.pm#12 $
#
# Copyright (c) 2012 NetApp
# All rights reserved.
#
## @summary
## IO specific functions to support NACL Translator framework
##
## @status    private
##

package NX::IO;

use strict 'vars';
use Tharn;
use NACL::ComponentUtils;
use File::Basename;
use NATE::ParamSet;

# This has been replaced by NACL::ComponentUtils::Dumper
use Data::Dumper;
use Devel::StackTrace;
use File::Spec::Functions;

use NX::Helper;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::MTask::Traffic;
use NACL::STask::VserverNFS;
use NACL::STask::Node;
use NACL::STask::Client::Mount;
use NACL::C::VserverCifs;
use NACL::STask::Volume;
use NACL::STask::VserverCifsShare;
use NACL::APISet::Node::CLI::7Mode::default::mkfile;
use NACL::APISet::Node::CLI::7Mode::default::ls;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
    hammer      fops    sio     filersio_asyncio_active     diskhogger
    _mount_client_volumes       _purge_mounts    cifs_hammer
);

my @Traffic_Objs;
my @Pkg_Workers;
my $Params = NATE::ParamSet->new(global => 1);

sub stop_workers {
    $Log->comment("Detected interrupt signal");
    if ( $Traffic_Objs[0] ) {
        my $num_running = 0;
        foreach my $to (@Traffic_Objs) {
            if ( $to->is_io_tool_running() ) {
                $num_running++;
                $to->stop();
            }
        } ## end foreach my $to (@Traffic_Objs)
        $Log->comment("Stopped $num_running traffic objects in progress")
            if $num_running;
        while ( my $to = shift(@Traffic_Objs) ) {
            while ( $to->is_io_tool_running() ) {
                sleep 1;
            }
        }
    } ## end if ( $Traffic_Objs[0] )

    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
} ## end sub stop_workers

=pod

=head1 NAME

NX::IO

=head1 DESCRIPTION

IO-related routines for use in NX framework.

=head1 METHODS

=head2 hammer

Run hammer on specified volumes on specified clients in parallel.  Each
hammer/volume instance requires one login session to the client.

!!!NOTE!!! If the "background" parameter is not set, then resulting hammer
log files could be potentially huge and take up large amounts of space.

The input is a hash (%input) which must include:
    client_command_interface => [ $client ] # Array ref of client objects
    volumes                  => [ $vol ]    # Array ref of volume objects

%opts can be used to provide additional options for hammer, which include
    mapping         => default '1-to-many'
    api             => -api param
    background      => -bg param
    checkinodes     => -checkinodes parma
    filercore       => -filercore param
    nofilercore     => -nofilercore param
    fill            => -fill param
    iterations      => -iterations param
    logfile         => -logfile param
    mode            => -mode param
    nocleanup       => -nocleanup param
    noflush         => -noflush param
    onlyflush       => -onlyflush param
    nostreams       => -nostreams param
    runtime         => -runtime param
    seed            => -seed param
    blocksize       => -blocksize param
    minbsize        => -minbsize param
    maxbsize        => -maxbsize param
    minfsize        => -minfsize param
    maxfsize        => -maxfsize param

If "filercore" is not specified, then "nofilercore" must be set.

Hammer will run forever if "runtime" is not specified.

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=cut

=cut begin NX hammer
Hammer
description
Run hammer on specified volumes on specified clients in parallel. If the "background" parameter is not set, then resulting log files could potentially take up large amounts of space.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volumes;;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
api;Api;;hammer: -api, use specified API, accepted values are win32|posix
background;Run in Background;;hammer: -bg, does not use stdin or stdout, set as 1 to enable this flag
checkinodes;Check Inodes;;hammer: -checkinodes, Check if file system reusing inodes, set as 1 to enable this flag
filercore;Filercore;;hammer: -filercore, have filer coredump upon corruption, cannot be used with nofilercore. This is the default setting.
nofilercore;No Filercore;;hammer: -nofilercore, disable default -filercore option, cannot be used with filercore.
fill;Fill;;hammer: -fill, fill disk and then keep it full, set as 1 to enable this flag
iterations;Iterations;;hammer: -iterations, integer value of interations to execute
logfile;Log File;;hammer: -logfile, specify file to use as logfile
mode;Mode;;hammer: -mode, use specified mode, acceptable values are mixed|creates|overwrites
nocleanup;No Cleanup;;hammer: -nocleanup, do not clean up upon completion, set as 1 to enable this flag
noflush;No Flush;;hammer: -noflush, do not use FlushFileBuffers(), set as 1 to enable this flag
onlyflush;Only Flush;;hammer: -onlyflush, always use FlushFileBuffers(), set as 1 to enable this flag
onlystreams;Only Streams;;hammer: use only NT stream files, set as 1 to enable this flag
nostreams;No Streams;;hammer: -nostreams, do not use NT stream files, set as 1 to enable this flag
runtime;Runtime;;hammer: -runtime, stop hammer after integer value of seconds
seed;Seed;;hammer: -seed, set initial random seed to integer value
blocksize;Block Size;;hammer: -blocksize, use blocksize of integer value in bytes
minbsize;Min Block Size;;hammer: -minbsize, set minimum block size of integer value in bytes
maxbsize;Max Block Size;;hammer: -maxbsize, set maximum block size of integer value in bytes
minfsize;Min File Size;;hammer: -minfsize, set minimum file size of integer value in bytes
maxfsize;Max File Size;;hammer: -maxfsize, set maximum file size of integer value in bytes
mapping;Mapping;1-to-many;Mapping of operations to volumes. Acceptable values 1-to-many|1-to-1 default is 1-to-many.
=end NX hammer
=cut

sub hammer {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};    # No context requirements

    my $my_return;    # Hash reference containing return values
    my $exception;
    my $obj_ref;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   client_command_interface =>
                    { type => ARRAYREF, optional => 0 },
                volumes => { type => ARRAYREF, optional => 0 },
                validate_only =>
                    { type => SCALAR, optional => 0, default => 0 },
                mapping =>
                    { type => SCALAR, optional => 1, default => '1-to-many' },
                api         => { type => SCALAR, optional => 1 },
                background  => { type => SCALAR, optional => 1 },
                checkinodes => { type => SCALAR, optional => 1 },
                filercore   => { type => SCALAR, optional => 1 },
                nofilercore => { type => SCALAR, optional => 1 },
                fill        => { type => SCALAR, optional => 1 },
                iterations  => { type => SCALAR, optional => 1 },
                logfile     => { type => SCALAR, optional => 1 },
                mode        => { type => SCALAR, optional => 1 },
                nocleanup   => { type => SCALAR, optional => 1 },
                noflush     => { type => SCALAR, optional => 1 },
                onlyflush   => { type => SCALAR, optional => 1 },
                onlystreams => { type => SCALAR, optional => 1 },
                nostreams   => { type => SCALAR, optional => 1 },
                runtime     => { type => SCALAR, optional => 1 },
                seed        => { type => SCALAR, optional => 1 },
                blocksize   => { type => SCALAR, optional => 1 },
                minbsize    => { type => SCALAR, optional => 1 },
                maxbsize    => { type => SCALAR, optional => 1 },
                minfsize    => { type => SCALAR, optional => 1 },
                maxfsize    => { type => SCALAR, optional => 1 },
                skip_nfs_setup =>
                    { type => SCALAR, optional => 1, default => 0 },
                mapping_position => { type => SCALAR, optional => 1 },
            }
        );

        foreach my $key ( keys %input ) {
            if ( $key eq "api" ) {
                if ( $input{$key} !~ /^(win32|posix)$/ ) {
                    NATE::BaseException->throw(
                        "\'$key\' value must be win32 or posix");
                }
            } elsif ( $key eq "mode" ) {
                if ( $input{$key} !~ /^(mixed|creates|overwrites)$/ ) {
                    NATE::BaseException->throw( "\'$key\' value must be "
                            . "\'mixed\', \'creates\', or \'overwrites\'" );
                }
            } elsif ( $key eq "filercore" ) {

                #TODO figure out why filercore is not getting param
                if ( exists( $input{'nofilercore'} ) ) {
                    NATE::BaseException->throw( "filercore and nofilercore "
                            . "params are mutually exclusive" );
                }
            } elsif ( $key eq "nofilercore" ) {
                if ( exists( $input{'filercore'} ) ) {
                    NATE::BaseException->throw( "filercore and nofilercore "
                            . "params are mutually exclusive" );
                }
            } elsif ( $key eq "logfile" ) {
                if ( $input{$key} !~ /\S+/ ) {
                    NATE::BaseException->throw("\'$key\' value required");
                }
            } elsif ( $key eq "runtime" ) {
                if ( $input{$key} !~ /^\d+s*/ ) {
                    NATE::BaseException->throw( "\'$key\' value must be "
                            . "in integer or integer followed by \'s\'" );
                }
            } elsif ( ( $key =~ /^(iterations|seed|blocksize)$/ )
                || ( $key =~ /^(minbsize|maxbsize|minfsize|maxfsize)$/ ) )
            {
                if ( ( $input{$key} !~ /^\d+$/ ) || ( $input{$key} < 1 ) ) {
                    NATE::BaseException->throw( "\'$key\' value must be "
                            . "integer greater than 1" );
                }
            } elsif ( $key eq "minfsize" ) {
                if ( defined( $input{'maxfsize'} )
                    && ( $input{$key} > $input{'maxfsize'} ) )
                {
                    NATE::BaseException->throw( "\'$key\' value must be less "
                            . "than \'maxfsize\' value" );
                } ## end if ( defined( $input{'maxfsize'...
            } elsif ( $key eq "maxfsize" ) {
                if ( defined( $input{'minfsize'} )
                    && ( $input{$key} < $input{'minfsize'} ) )
                {
                    NATE::BaseException->throw(
                              "\'$key\' value must be greater "
                            . "than \'minfsize\' value" );
                } ## end if ( defined( $input{'minfsize'...
            } ## end elsif ( $key eq "maxfsize")
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "hammer action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::IO::stop_workers";
        delete $input{validate_only};
    } ## end else [ if ($exception)

    # Log start tick
    $my_return->{start_tick} = time();

    my @client_objs = @{ $input{client_command_interface} };
    my $num_clients = @client_objs;
    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;

    # On all volumes, set up NFS and export
    _setup_nfs( volumes => $input{volumes} ) if !$input{skip_nfs_setup};

    if ( $num_clients > 1 ) {

        # More than one client specified, divide and conquer by splitting
        # work off on each client in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $c (@client_objs) {
            my $id = 'hammer_C_DaC.'
                . $proc_cnt;    # (C)lient_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{client_command_interface} = [$c];
            $input{mapping_position}         = ( $proc_cnt - 1 );
            $input{skip_nfs_setup}           = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'hammer',
                args    => [%input],
            );
            push( @Pkg_Workers, $worker );
            push( @workers,     $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $c (@client_objs)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $client_ref = $res->{$rid}->{objects}->{client};
                foreach my $client_name ( keys %{$client_ref} ) {
                    my @mnt_objs = @{ $client_ref->{$client_name} };
                    push( @{ $obj_ref->{client}->{$client_name} },
                        @mnt_objs );
                }
            } ## end if ( $res->{$rid}->{objects...
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            my $msg = 'One or more procs had failures';
            $Log->error($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran hammer on $num_clients clients";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( $num_volumes > 1 ) {

        # Single client, but more than one volume specified, divide and
        # conquer by splitting work off for each volume in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $client      = shift(@client_objs);
        my $client_name = $client->name();
        my $api         = $client->apiset();

        if ( $input{mapping} eq '1-to-1' ) {

            # Perform IO to volume at same array position as
            # the 'mapping_position' param value
            @volume_objs = $volume_objs[ $input{mapping_position} ];
        } ## end if ( $input{mapping} eq...

        my $mounts_and_paths = _mount_client_volumes(
            client  => $client,
            volumes => \@volume_objs
        );
        push(
            @{ $obj_ref->{client}->{$client_name} },
            @{ $mounts_and_paths->{mounts} }
        );
        my @all_paths;

        foreach my $v (@volume_objs) {
            my $id
                = 'hammer_P_DaC.' . $proc_cnt; # (P)ath_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            my $path = $mounts_and_paths->{paths}->{ $v->volume() };
            $path .= '/' . $client_name;

            # If mapping is 1-to-many, create a subdir
            if ( $input{mapping} eq '1-to-many' ) {
                $path .= '/hammer_P_DaC.' . $proc_cnt;
            }

            # Append random subdir, in case multiple hammer
            # instances on same volume
            srand();
            my $suffix = int( rand(1000000) );
            $path .= '/' . $suffix;

            # Create directory
            $api->execute_raw_command( command => "mkdir -p $path" );

            # Remove files from working dir for clean run
            if ($path) {    # Provide some safety against
                            # possible null value
                $Log->comment("Removing files before hammer from $path");
                $api->execute_raw_command( command => "rm -rf $path/*" );
                push( @all_paths, $path );
            } ## end if ($path)

            $input{volumes}        = [$path];
            $input{skip_nfs_setup} = 1;

            # Stagger execution to avoid 10 sessions within single tick
            my $slp = int( rand(3) );
            sleep 1 if $slp;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'hammer',
                args    => [%input],
            );
            push( @Pkg_Workers, $worker );
            push( @workers,     $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $v (@volume_objs)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my @res_obj
                    = @{ $res->{$rid}->{objects}->{client}->{$client_name} };
                push( @{ $obj_ref->{client}->{$client_name} }, @res_obj );
            }
        } ## end foreach my $w (@workers)

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            my $msg = 'hammer reported failures against one or more volumes';
            $Log->error($msg);
            $my_return->{msg} = $msg;
        } else {

            # Remove all files if everything passed
            $Log->comment("Removing files after hammer from all paths");
            foreach my $p (@all_paths) {
                $api->execute_raw_command( command => "rm -rf $p/*" );
            }
            $my_return->{result} = 1;
            my $msg
                = "Successfully ran hammer on $client_name against "
                . @volume_objs
                . " volumes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {

        # Single client, single path
        $input{tool} = 'hammer';

        my $client       = shift(@client_objs);
        my $client_name  = $client->name();
        my $found_errors = 0;
        my $api          = $client->apiset();
        $api->set_timeout( 'connectrec-timeout' => 600 );
        my $to = $api->get_timeout();
        if ( $to == 600 ) {
            $Log->comment("Increased client timeout to 600");
        } else {
            $Log->error("Failed to increaes client timeout");
        }

        try {

            # Translate 'volumes' arg into 'paths' arg
            my $input_type = 'recurse';
            $input{'paths'} = $input{volumes}->[0];
            if ( ref( $input{'paths'} ) ) {

                # It's a vol object, mount it and get path
                # This is the case if initial NX call was for a single volume
                my $vol_obj          = $input{'paths'};
                my $mounts_and_paths = _mount_client_volumes(
                    client  => $client,
                    volumes => $input{volumes}
                );
                $input{'paths'}
                    = $mounts_and_paths->{paths}->{ $vol_obj->volume() };
                push(
                    @{ $obj_ref->{client}->{$client_name} },
                    @{ $mounts_and_paths->{mounts} }
                );
                $input_type = 'discrete';
            } else {

                # It's an actual path string that's already mounted,
                # most likely from path divide & conquer
            }

            # Ensure at a minimum the client's name is in the path
            if ( $input{'paths'} !~ /$client_name/ ) {
                $input{'paths'} .= '/' . $client_name;
                $Log->comment( "Appending client name to path: *"
                        . $input{'paths'}
                        . "*" );

                # Append random subdir, in case multiple hammer
                # instances on same volume
                srand();
                my $suffix = int( rand(1000000) );
                $input{'paths'} .= '/' . $suffix;

                # Create directory
                $api->execute_raw_command(
                    command => "mkdir -p $input{paths}" );

                # Remove files from working dir for clean run
                if ( $input{'paths'} ) {    # Provide some safety against
                                            # possible null value
                    $Log->comment(
                        "Removing files before hammer from $input{paths}");
                    $api->execute_raw_command(
                        command => "rm -rf $input{paths}/*" );
                } ## end if ( $input{'paths'} )
            } ## end if ( $input{'paths'} !~...

            delete $input{volumes};
            delete $input{mapping};
            delete $input{mapping_position};
            delete $input{client_command_interface};
            delete $input{skip_nfs_setup};
            $input{command_interface} = $client;

            ## MTask call is non-blocking
            my $traffic_obj = NACL::MTask::Traffic->start(%input);
            push( @Traffic_Objs, $traffic_obj );
            my $resp = $traffic_obj->wait_for_completion();
            my $raw  = $resp->get_raw_output();
            if ( $raw =~ /error/i ) {
                $found_errors = 1;
            }
            if ( !$found_errors && ( $input_type eq 'discrete' ) ) {
                $Log->comment(
                    "Removing files after hammer from $input{paths}");
                $api->execute_raw_command(
                    command => "rm -rf $input{paths}/*" );
            }
        } ## end try
        otherwise {
            $exception = shift;
        };

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } elsif ($found_errors) {
            $my_return->{result} = 0;
            my $msg
                = "Found error string(s) in hammer output on $client_name";
            $Log->error($msg);
            $my_return->{msg} = $msg;
        } else {    # Good IO without errors
            $my_return->{result} = 1;
            my $msg
                = "hammer ran successfully without errors on $client_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( $num_clients > 1 )
}    # end sub hammer

=head2 cifs_hammer

Run cifs_hammer on specified volumes on specified clients in parallel.  Each
hammer/volume instance requires one login session to the client.

!!!NOTE!!! If the "background" parameter is not set, then resulting hammer
log files could be potentially huge and take up large amounts of space.

The input is a hash (%input) which must include:
    client_command_interface => [ $client ] # Array ref of client objects
    volumes                  => [ $vol ]    # Array ref of volume objects

%opts can be used to provide additional options for hammer, which include
    mapping         => default '1-to-many'
    api             => -api param
    background      => -bg param
    checkinodes     => -checkinodes parma
    filercore       => -filercore param
    nofilercore     => -nofilercore param
    fill            => -fill param
    iterations      => -iterations param
    logfile         => -logfile param
    mode            => -mode param
    nocleanup       => -nocleanup param
    noflush         => -noflush param
    onlyflush       => -onlyflush param
    nostreams       => -nostreams param
    runtime         => -runtime param
    seed            => -seed param
    blocksize       => -blocksize param
    minbsize        => -minbsize param
    maxbsize        => -maxbsize param
    minfsize        => -minfsize param
    maxfsize        => -maxfsize param
    pretest_mount_cleanup => default '0', clear other mounts prior to hammer run


If "filercore" is not specified, then "nofilercore" must be set.

Hammer will run forever if "runtime" is not specified.

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=cut

=cut begin NX cifs_hammer
CIFS Hammer
description
Run CIFS hammer on specified volumes from specified clients. CIFS clients required. If the "background" parameter is not set, then resulting log files could potentially take up large amounts of space.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volumes;;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
api;Api;;hammer: -api, use specified API, accepted values are win32|posix
background;Run in Background;;hammer: -bg, does not use stdin or stdout, set as 1 to enable this flag
checkinodes;Check Inodes;;hammer: -checkinodes, Check if file system reusing inodes, set as 1 to enable this flag
filercore;Filercore;;hammer: -filercore, have filer coredump upon corruption, cannot be used with nofilercore. This is the default setting.
nofilercore;No Filercore;;hammer: -nofilercore, disable default -filercore option, cannot be used with filercore.
fill;Fill;;hammer: -fill, fill disk and then keep it full, set as 1 to enable this flag
iterations;Iterations;;hammer: -iterations, integer value of interations to execute
logfile;Log File;;hammer: -logfile, specify file to use as logfile
mode;Mode;;hammer: -mode, use specified mode, acceptable values are mixed|creates|overwrites
nocleanup;No Cleanup;;hammer: -nocleanup, do not clean up upon completion, set as 1 to enable this flag
noflush;No Flush;;hammer: -noflush, do not use FlushFileBuffers(), set as 1 to enable this flag
onlyflush;Only Flush;;hammer: -onlyflush, always use FlushFileBuffers(), set as 1 to enable this flag
onlystreams;Only Streams;;hammer: use only NT stream files, set as 1 to enable this flag
nostreams;No Streams;;hammer: -nostreams, do not use NT stream files, set as 1 to enable this flag
runtime;Runtime;;hammer: -runtime, stop hammer after integer value of seconds
seed;Seed;;hammer: -seed, set initial random seed to integer value
blocksize;Block Size;;hammer: -blocksize, use blocksize of integer value in bytes
minbsize;Min Block Size;;hammer: -minbsize, set minimum block size of integer value in bytes
maxbsize;Max Block Size;;hammer: -maxbsize, set maximum block size of integer value in bytes
minfsize;Min File Size;;hammer: -minfsize, set minimum file size of integer value in bytes
maxfsize;Max File Size;;hammer: -maxfsize, set maximum file size of integer value in bytes
mapping;Mapping;1-to-many;Mapping of operations to volumes. Acceptable values 1-to-many|1-to-1 default is 1-to-many.
pretest_mount_cleanup;Pretest Mount Cleanup;0;Cleanup existing mounts except for X: prior to running cifs hammer. Default is set to 0. Set as 1 to enable.
=end NX cifs_hammer
=cut

sub cifs_hammer {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};    # No context requirements

    my $my_return;    # Hash reference containing return values
    my $exception;
    my $obj_ref;
    my $cifs;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                client_command_interface => { type => ARRAYREF, optional => 0 },
                volumes                  => { type => ARRAYREF, optional => 0 },
                validate_only =>
                  { type => SCALAR, optional => 0, default => 0 },
                validate_only_params =>
                  { type => SCALAR, optional => 0, default => 0 },
                mapping =>
                  { type => SCALAR, optional => 1, default => '1-to-many' },
                api         => { type => SCALAR, optional => 1 },
                background  => { type => SCALAR, optional => 1 },
                checkinodes => { type => SCALAR, optional => 1 },
                filercore   => { type => SCALAR, optional => 1 },
                nofilercore => { type => SCALAR, optional => 1 },
                fill        => { type => SCALAR, optional => 1 },
                iterations  => { type => SCALAR, optional => 1 },
                logfile     => { type => SCALAR, optional => 1 },
                mode        => { type => SCALAR, optional => 1 },
                nocleanup   => { type => SCALAR, optional => 1 },
                noflush     => { type => SCALAR, optional => 1 },
                onlyflush   => { type => SCALAR, optional => 1 },
                onlystreams => { type => SCALAR, optional => 1 },
                nostreams   => { type => SCALAR, optional => 1 },
                runtime     => { type => SCALAR, optional => 1 },
                seed        => { type => SCALAR, optional => 1 },
                blocksize   => { type => SCALAR, optional => 1 },
                minbsize    => { type => SCALAR, optional => 1 },
                maxbsize    => { type => SCALAR, optional => 1 },
                minfsize    => { type => SCALAR, optional => 1 },
                maxfsize    => { type => SCALAR, optional => 1 },
                mapping_position => { type => SCALAR, optional => 1 },
                skip_cifs_setup =>
                  { type => SCALAR, optional => 1, default => 0 },
                pretest_mount_cleanup =>
                  { type => SCALAR, optional => 1, default => 0 },
                skip_tool_map =>
                  { type => SCALAR, optional => 1, default => 0 },
            }
        );

        foreach my $key ( keys %input ) {
            if ( $key eq "api" ) {
                if ( $input{$key} !~ /^(win32|posix)$/ ) {
                    NATE::BaseException->throw(
                        "\'$key\' value must be win32 or posix");
                }
            } elsif ( $key eq "mode" ) {
                if ( $input{$key} !~ /^(mixed|creates|overwrites)$/ ) {
                    NATE::BaseException->throw( "\'$key\' value must be "
                          . "\'mixed\', \'creates\', or \'overwrites\'" );
                }
            } elsif ( $key eq "filercore" ) {
                if ( exists( $input{'nofilercore'} ) ) {
                    NATE::BaseException->throw( "filercore and nofilercore "
                          . "params are mutually exclusive" );
                }
            } elsif ( $key eq "nofilercore" ) {
                if ( exists( $input{'filercore'} ) ) {
                    NATE::BaseException->throw( "filercore and nofilercore "
                          . "params are mutually exclusive" );
                }
            } elsif ( $key eq "logfile" ) {
                if ( $input{$key} !~ /\S+/ ) {
                    NATE::BaseException->throw("\'$key\' value required");
                }
            } elsif ( $key eq "runtime" ) {
                if ( $input{$key} !~ /^\d+s*/ ) {
                    NATE::BaseException->throw( "\'$key\' value must be "
                          . "in integer or integer followed by \'s\'" );
                }
            } elsif (( $key =~ /^(iterations|seed|blocksize)$/ )
                || ( $key =~ /^(minbsize|maxbsize|minfsize|maxfsize)$/ ) )
            {
                if ( ( $input{$key} !~ /^\d+$/ ) || ( $input{$key} < 1 ) ) {
                    NATE::BaseException->throw(
                        "\'$key\' value must be " . "integer greater than 1" );
                }
            } elsif ( $key eq "minfsize" ) {
                if ( defined( $input{'maxfsize'} )
                    && ( $input{$key} > $input{'maxfsize'} ) )
                {
                    NATE::BaseException->throw( "\'$key\' value must be less "
                          . "than \'maxfsize\' value" );
                } ## end if ( defined( $input{'maxfsize'...
            } elsif ( $key eq "maxfsize" ) {
                if ( defined( $input{'minfsize'} )
                    && ( $input{$key} < $input{'minfsize'} ) )
                {
                    NATE::BaseException->throw(
                            "\'$key\' value must be greater "
                          . "than \'minfsize\' value" );
                } ## end if ( defined( $input{'minfsize'...
            } ## end elsif ( $key eq "maxfsize")
        } ## end foreach my $key ( keys %input)

        ## Check CIFS parameters
        $cifs = _get_cifs_args(clients => $input{client_command_interface});

    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only_params} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "cifs hammer action passed param validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::IO::stop_workers";
        delete $input{validate_only_params};
    } ## end else [ if ($exception)

    # Log start tick
    $my_return->{start_tick} = time();

    my @client_objs = @{ $input{client_command_interface} };
    my $num_clients = @client_objs;
    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;

    try {
        # On all clusters, set up CIFS
        $Log->comment("Setting up CIFS");
        _setup_cifs( volumes => $input{volumes}, params => $cifs ) if !$input{skip_cifs_setup};

        if($input{validate_only}) {
            # On all clients, map hammer tool directory
            $Log->comment("Mapping CIFS Tools if necessary");
            _map_cifs_tools( clients => $input{client_command_interface} )
            if !$input{skip_tool_map};

            # On all clients, cleanup other mounts if pretest_mount_cleanup is 1
            $Log->comment("Purge previous mounts if specified");
            _pretest_mount_cleanup( clients => $input{client_command_interface} )
            if $input{pretest_mount_cleanup};
        }
    } ## end second try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "cifs hammer action passed parameter ".
                               "and config validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        delete $input{validate_only};
    } ## end else [ if ($exception)

    $Log->comment("Starting...");

    if ( $num_clients > 1 ) {

        # More than one client specified, divide and conquer by splitting
        # work off on each client in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $c (@client_objs) {
            my $id = 'cifs_hammer_C_DaC.'
              . $proc_cnt;    # (C)lient_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{client_command_interface} = [$c];
            $input{mapping_position}         = ( $proc_cnt - 1 );
            $input{skip_cifs_setup}          = 1;
            $input{skip_tool_map}            = 1;
            $input{pretest_mount_cleanup}    = 0;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'cifs_hammer',
                args    => [%input],
            );
            push( @Pkg_Workers, $worker );
            push( @workers,     $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $c (@client_objs)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $client_ref = $res->{$rid}->{objects}->{client};
                foreach my $client_name ( keys %{$client_ref} ) {
                    my @mnt_objs = @{ $client_ref->{$client_name} };
                    push( @{ $obj_ref->{client}->{$client_name} }, @mnt_objs );
                }
            } ## end if ( $res->{$rid}->{objects...
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            my $msg = 'One or more procs had failures';
            $Log->error($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran cifs hammer on $num_clients clients";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( $num_volumes > 1 ) {
        # Single client, but more than one volume specified, divide and
        # conquer by splitting work off for each volume in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $client      = shift(@client_objs);
        my $client_name = $client->name();
        my $api         = $client->apiset();

        if ( $input{mapping} eq '1-to-1' ) {

            # Perform IO to volume at same array position as
            # the 'mapping_position' param value
            @volume_objs = $volume_objs[ $input{mapping_position} ];
        } ## end if ( $input{mapping} eq...

        my $mounts_and_paths = _mount_cifs_volumes(
            client  => $client,
            volumes => \@volume_objs,
            params  => $cifs,
        );
        push(
            @{ $obj_ref->{client}->{$client_name} },
            @{ $mounts_and_paths->{mounts} }
        );
        my @all_paths;

        foreach my $v (@volume_objs) {
            my $id =
              'cifs_hammer_P_DaC.' . $proc_cnt;  # (P)ath_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            my $path = $mounts_and_paths->{paths}->{ $v->volume() };
            $path .= '\\' . $client_name;

            # If mapping is 1-to-many, create a subdir
            if ( $input{mapping} eq '1-to-many' ) {
                $path .= '\\cifs_hammer_P_DaC.' . $proc_cnt;
            }

            # Append random subdir, in case multiple hammer
            # instances on same volume
            srand();
            my $suffix = int( rand(1000000) );
            $path .= '\\a' . $suffix;

            # Create directory
            $api->mkdir( path => $path );

            # Remove files from working dir for clean run
            if ($path) {    # Provide some safety against
                            # possible null value
                $Log->comment("Removing files before hammer from $path");
                $api->del(
                    'force-delete'          => 1,
                    'delete-specified-file' => 1,
                    'quiet-mode'            => 1,
                    'files'                 => "$path > nul",
                );
                push( @all_paths, $path );
            } ## end if ($path)

            $input{volumes}               = [$path];
            $input{skip_cifs_setup}       = 1;
            $input{skip_tool_map}         = 1;
            $input{pretest_mount_cleanup} = 0;

            # Stagger execution to avoid 10 sessions within single tick
            my $slp = int( rand(5) );
            sleep 1 if $slp;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'cifs_hammer',
                args    => [%input],
            );
            push( @Pkg_Workers, $worker );
            push( @workers,     $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $v (@volume_objs)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my @res_obj =
                  @{ $res->{$rid}->{objects}->{client}->{$client_name} };
                push( @{ $obj_ref->{client}->{$client_name} }, @res_obj );
            }
        } ## end foreach my $w (@workers)

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            my $msg =
              'cifs hammer reported failures against one or more volumes';
            $Log->error($msg);
            $my_return->{msg} = $msg;
        } else {

            # Remove all files if everything passed
            $Log->comment("Removing files after cifs hammer from all paths");
            foreach my $p (@all_paths) {
                try {
                    $api->rmdir(
                        'recursive'  => 1,
                        'quiet-mode' => 1,
                        'path'       => $p,
                    );
                }
                otherwise {
                    $exception = shift;
                    if ( $exception =~ /\'text\' field was not provided/i ) {
                        $exception = '';    #Clear exception
                    }
                };
            }
            $my_return->{result} = 1;
            my $msg =
                "Successfully ran cifs hammer on $client_name against "
              . @volume_objs
              . " volumes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {

        # Single client, single path
        $input{tool} = 'hammer';

        my $client       = shift(@client_objs);
        my $client_name  = $client->name();
        my $found_errors = 0;
        my $api          = $client->apiset();
        $api->set_timeout( 'connectrec-timeout' => 600 );
        my $to = $api->get_timeout();
        if ( $to == 600 ) {
            $Log->comment("Increased client timeout to 600");
        } else {
            $Log->error("Failed to increase client timeout");
        }

        try {

            # Translate 'volumes' arg into 'paths' arg
            my $input_type = 'recurse';
            $input{'paths'} = $input{volumes}->[0];
            if ( ref( $input{'paths'} ) ) {

                # It's a vol object, mount it and get path
                # This is the case if initial NX call was for a single volume
                my $vol_obj          = $input{'paths'};
                my $mounts_and_paths = _mount_cifs_volumes(
                    client  => $client,
                    volumes => $input{volumes},
                    params  => $cifs,
                );
                $input{'paths'} =
                  $mounts_and_paths->{paths}->{ $vol_obj->volume() };
                push(
                    @{ $obj_ref->{client}->{$client_name} },
                    @{ $mounts_and_paths->{mounts} }
                );
                $input_type = 'discrete';
            } else {

                # It's an actual path string that's already mounted,
                # most likely from path divide & conquer
            }

            # Ensure at a minimum the client's name is in the path
            if ( $input{'paths'} !~ /$client_name/ ) {
                $input{'paths'} .= '\\' . $client_name;
                $Log->comment( "Appending client name to path: *"
                      . $input{'paths'}
                      . "*" );

                # Append random subdir, in case multiple hammer
                # instances on same volume
                srand();
                my $suffix = int( rand(1000000) );
                $input{'paths'} .= '\\b' . $suffix;

                # Create directory
                $api->mkdir( path => $input{paths} );

                # Remove files from working dir for clean run
                if ( $input{'paths'} ) {    # Provide some safety against
                                            # possible null value
                    $Log->comment("Removing files before cifs hammer from $input{paths}");
                    $api->del(
                        'force-delete'          => 1,
                        'delete-specified-file' => 1,
                        'quiet-mode'            => 1,
                        'files'                 => $input{paths} . " > nul",
                    );
                } ## end if ( $input{'paths'} )
            } ## end if ( $input{'paths'} !~...

            delete $input{volumes};
            delete $input{mapping};
            delete $input{mapping_position};
            delete $input{client_command_interface};
            delete $input{skip_cifs_setup};
            delete $input{skip_tool_map};
            delete $input{pretest_mount_cleanup};
            $input{command_interface} = $client;

            ## Using hammer_19.exe until a better version of hammer is released
            $input{tool_full_path} = 'X:\\\\bin\\hammer_19.exe';

            ## MTask call is non-blocking
            my $traffic_obj;
            try {
                $traffic_obj = NACL::MTask::Traffic->start(%input);
                push( @Traffic_Objs, $traffic_obj );
                if ( $input{runtime} ) {
                    my $dur        = 0;
                    my $time_limit = $input{runtime} + 120;
                    $Log->comment("Duration will be padded with two minutes for overhead");
                    while ( $dur <= $time_limit ) {
                        if ( $traffic_obj->is_io_tool_running ) {
                            $Log->comment("CIFS Hammer is still running(duration: $dur/$time_limit)");
                            if ( $dur + 1 > $time_limit ) {
                                $Log->comment("Tool has run past duration, exiting");
                                if ( !$::IGNORE_ERRORS ) {
                                    $found_errors = 1;
                                }
                                $traffic_obj->stop();
                                last;
                            }
                        } else {
                            $Log->comment("CIFS Hammer is no longer running(duration: $dur/$time_limit");
                            last;
                        }
                        sleep(30);
                        $dur += 30;
                    }
                }
            } catch NACL::Exceptions::TrafficFailed with {
                my $e     = shift;
                my $e_raw = $e->stacktrace_chain();
                if ( $e_raw =~ /Reference to nonexistent group in regex/i ) {
                    $Log->comment("Ignoring Known Issue: nonexistent group in regex");
                } elsif ( $e_raw =~ /The I\/O tool \'hammer\' has stopped running/i ) {
                    $Log->comment("CIFS hammer has stopped running.");
                } else {
                    if ($::IGNORE_ERRORS) {
                        $Log->warn("IGNORE_ERRORS is set, ignoring error: $e_raw");
                    } else {
                        $exception    = $e_raw;
                        $found_errors = 1;
                        $Log->error("Exception detected: $e_raw");
                    }
                }
            } otherwise {
                my $e = shift;
                if ($::IGNORE_ERRORS) {
                    $Log->warn("IGNORE_ERRORS is set, ignoring error: ".$e->text());
                } else {
                    $exception    = $e;
                    $found_errors = 1;
                }
            };

            if (   ( !$found_errors || $::IGNORE_ERRORS )
                && ( $input_type eq 'discrete' ) ) {
                $Log->comment("Removing files after hammer from $input{paths}");
                $traffic_obj->wait_for_completion();
                try {
                    $api->rmdir(
                        'recursive'  => 1,
                        'quiet-mode' => 1,
                        'path'       => $input{paths},
                    );
                } otherwise {
                    my $e = shift;
                    if ( $e->text() !~ /\'text\' field was not provided/i ) {
                        $e->throw();
                    }
                };
            }

        } otherwise {
            $exception = shift;
        };

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } elsif ($found_errors) {
            $my_return->{result} = 0;
            my $msg = "Found error in cifs hammer output on $client_name";
            $Log->error($msg);
            $my_return->{msg} = $msg;
        } else {    # Good IO without errors
            $my_return->{result} = 1;
            my $msg = "cifs hammer ran successfully without errors on $client_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    }
}    # end sub cifs_hammer

=pod

=head2 fops

Run fops on specified volumes on specified clients in parallel.  Each client
requires one login session to perform all fops operations.  Each fops
invocation is run as a background process on the client via "&".

The input is a hash (%input) which must include:
    client_command_interface => [ $client ] # Array ref of client objects
    volumes                  => [ $vol ]    # Array ref of volume objects
    operations               => operation string (-o param)
    num-of-files             => # of files (-n param)

%opts can be used to provide additional options for fops, which include
    mapping                     => default "1-to-many"
    repeat                      => default 1
    height                      => -H param
    width                       => -W param
    num-of-threads              => -t param
    size-string                 => -s param
    dir-prefix                  => -P param
    file-prefix                 => -p param
    rename-file-prefix          => -q param
    level                       => -l param
    block-size                  => -b param
    delay-in-operation          => -d param
    delay-between-operations    => -D param
    expression                  => -x param
    offset-string               => -f param
    metrics-string              => -M param
    error-tolerance-percentage  => -X param
    file-open-options           => -O param
    use-existing-dir            => -e param
    pick-random-files           => -r param

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=cut

=cut begin NX fops
FOps
description
Run fops from specified clients on specified volumes.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
operations;Operations;;FOPS: -o, operation string, sequence of supported operation codes
num-of-files;Num of Files;;FOPS: -n, number of files in each directory
optional
size-string;Size String;;FOPS: -s, Comma separated operations size string, default 1byte
height;Height;;FOPS: -H, height of directory tree, default 0
width;Width;;FOPS: -W, width of tree at each level, default NULL
use-existing-dir;Use Existing Directory;;FOPS: -e, Use existing directory tree
num-of-threads;Num of Threads;;FOPS: -t, Number of threads to use, default 0
dir-prefix;Directory Prefix;;FOPS: -P, Directory name prefix, default DIR
file-prefix;File Prefix;;FOPS: -p, File name prefix, default FILE
rename-file-prefix;Rename-File Prefix;;FOPS: -q, File prefix for renamed file name, default NEW_FILE
level;Level;;FOPS: -l, Level in the directory tree to work, default leaf level
block-size;Block Size;;FOPS: -b, Block size to use for operations, default 256k
delay-in-operation;Delay in Operation;;FOPS: -d, Delay before operating on next file, default 0
delay-between-operations;Delay Between Operations;;FOPS: -D, Delay before each file operation, default 0
expression;Expression;;FOPS: -x, Expression to write and append, default A
offset-string;Offset String;;FOPS: -f, Comma separated string of offsets, default 0
metrics-string;Metrics String;;FOPS: -M, Comma separated time ranges, default 0,3000,100
error-tolerance-percentage;Error Tolerance %;;FOPS: -X, Percentage of errors to ignore, default 10
file-open-options;File Open Options;;FOPS: -O, Open system call options, default NULL
pick-random-files;Pick Random Files;;FOPS: -r, Flag to pick random files
repeat;Repeat;1;Amount of times to repeat FOPS
mapping;Mapping;1-to-many;Mapping of operations to volumes. Acceptable values 1-to-many|1-to-1 default is 1-to-many.
=end NX fops
=cut

sub fops {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};    # No context requirements

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   client_command_interface =>
                    { type => ARRAYREF, optional => 0 },
                validate_only =>
                    { type => SCALAR, optional => 0, default => 0 },

# Working-dir param will be created on the fly based on volumes:
# -> Until we find out how to handle mounting and how to associate mount paths
#    with volumes, take an ARRAYREF of path strings for now
#
                mapping =>
                    { type => SCALAR, optional => 1, default => '1-to-many' },
                'volumes'            => { type => ARRAYREF, optional => 0, },
                'operations'         => { type => SCALAR,   optional => 0, },
                'num-of-files'       => { type => SCALAR,   optional => 0, },
                'size-string'        => { type => SCALAR,   optional => 1, },
                'height'             => { type => SCALAR,   optional => 1, },
                'width'              => { type => SCALAR,   optional => 1, },
                'use-existing-dir'   => { type => SCALAR,   optional => 1, },
                'num-of-threads'     => { type => SCALAR,   optional => 1, },
                'dir-prefix'         => { type => SCALAR,   optional => 1, },
                'file-prefix'        => { type => SCALAR,   optional => 1, },
                'rename-file-prefix' => { type => SCALAR,   optional => 1, },
                'level'              => { type => SCALAR,   optional => 1, },
                'block-size'         => { type => SCALAR,   optional => 1, },
                'delay-in-operation' => { type => SCALAR,   optional => 1, },
                'delay-between-operations' =>
                    { type => SCALAR, optional => 1, },
                'expression'     => { type => SCALAR, optional => 1, },
                'offset-string'  => { type => SCALAR, optional => 1, },
                'metrics-string' => { type => SCALAR, optional => 1, },
                'error-tolerance-percentage' =>
                    { type => SCALAR, optional => 1, },
                'file-open-options' => { type => SCALAR, optional => 1, },
                'pick-random-files' => { type => SCALAR, optional => 1, },
                mapping_position    => { type => SCALAR, optional => 1, },
                repeat => { type => SCALAR, optional => 1, default => 1 },

      #mounts_and_paths                => { type => SCALAR, optional => 1,  },
                skip_nfs_setup =>
                    { type => SCALAR, optional => 1, default => 0 },
            }
        );

        foreach my $key ( keys %input ) {
            if ( $key eq "operations" ) {
                if ( $input{$key} !~ m/[crwatdns]/ ) {
                    NATE::BaseException->throw(
                        "Invalid value for $key parameter");
                }
            } elsif ( $key eq 'mapping' ) {
                if ( $input{$key} !~ /^(1\-to\-1|1\-to\-many)$/i ) {
                    NATE::BaseException->throw(
                        "$key value must be \'1-to-1\' or \'1-to-many\'");
                }
            } elsif ( $key eq "num-of-files" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "size-string" ) {

                #
            } elsif ( $key eq "height" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "width" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "num-of-threads" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "repeat" ) {
                if ( $input{$key} < 1 ) {
                    NATE::BaseException->throw(
                        "$key value must be 1 or more");
                }
            } elsif ( $key eq "dir-prefix" ) {

                #TODO failing test, fix it
                if ( $input{$key} !~ m/S+/ ) {
                    NATE::BaseException->throw(
                        "Invalid value for $key parameter");
                }
            } elsif ( $key eq "file-prefix" ) {
                if ( $input{$key} !~ m/S+/ ) {
                    NATE::BaseException->throw(
                        "Invalid value for $key parameter");
                }
            } elsif ( $key eq "rename-file-prefix" ) {
                if ( $input{$key} !~ m/S+/ ) {
                    NATE::BaseException->throw(
                        "Invalid value for $key parameter");
                }
            } elsif ( $key eq "level" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "block-size" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "delay-in-operation" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "delay-between-operations" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } elsif ( $key eq "expression" ) {

                #
            } elsif ( $key eq "offset-string" ) {

                #
            } elsif ( $key eq "metrics-string" ) {

                #
            } elsif ( $key eq "error-tolerance-percentage" ) {
                if ( $input{$key} < 0 ) {
                    NATE::BaseException->throw(
                        "$key value must be zero or more");
                }
            } ## end elsif ( $key eq "error-tolerance-percentage")
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "fops action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::IO::stop_workers";
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    my $obj_ref;
    my @volumes     = @{ $input{volumes} };
    my $num_volumes = @volumes;
    my @clients     = @{ $input{client_command_interface} };
    my $num_clients = @clients;
    $my_return->{start_tick} = time();

    # On all volumes, set up NFS and export
    _setup_nfs( volumes => $input{volumes} ) if !$input{skip_nfs_setup};

    if ( $num_clients > 1 ) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $c (@clients) {
            my $id
                = 'fops_C_DaC.' . $proc_cnt; # (C)lient_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{client_command_interface} = [$c];
            $input{mapping_position}         = ( $proc_cnt - 1 );
            $input{skip_nfs_setup}           = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'fops',
                args    => [%input],
            );
            push( @Pkg_Workers, $worker );
            push( @workers,     $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $c (@clients)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $client_ref = $res->{$rid}->{objects}->{client};
                foreach my $client_name ( keys %{$client_ref} ) {
                    my @mnt_objs = @{ $client_ref->{$client_name} };
                    push( @{ $obj_ref->{client}->{$client_name} },
                        @mnt_objs );
                }
            } ## end if ( $res->{$rid}->{objects...
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            my $msg = 'One or more procs had failures';
            $Log->error($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran fops on $num_clients clients";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {    # Single client
        my $num_volumes = @volumes;
        my $client      = shift(@clients);
        my $client_name = $client->name();
        my $api         = $client->apiset();
        $api->set_timeout( 'connectrec-timeout' => 600 );
        my $to = $api->get_timeout();
        if ( $to == 600 ) {
            $Log->comment("Increased client timeout to 600");
        } else {
            $Log->error("Failed to increaes client timeout");
        }
        if ( $input{mapping} eq '1-to-1' ) {

            # Perform IO to volume at same array position as
            # the 'mapping_position' param value
            @volumes = $volumes[ $input{mapping_position} ];
        } ## end if ( $input{mapping} eq...

        my $mounts_and_paths = _mount_client_volumes(
            client  => $client,
            volumes => \@volumes
        );
        push(
            @{ $obj_ref->{client}->{$client_name} },
            @{ $mounts_and_paths->{mounts} }
        );

        my $num_loops = delete $input{repeat};
        $num_loops = 1 if !$num_loops;
        my $num_errors = 0;
        my $loop_cnt   = 1;
        my $vol_fops_info;
        my $ofile_base = 'fops_';
        $ofile_base .= localtime( time() );
        $ofile_base =~ s/[\s:]//g;
        srand();
        $ofile_base .= '_' . ( int( rand(10000) ) + 1 ) . '/';

        # Create directory to store fops output logs
        $api->execute_raw_command( command => "mkdir -p $ofile_base" );

        my $active_fops_pids;
        foreach my $v (@volumes) {
            my $vol_name = $v->volume();
            my $path     = $mounts_and_paths->{paths}->{$vol_name};
            $path .= '/' . $client_name;
            my $ofile = $ofile_base . $vol_name;

            # Append random subdir, in case multiple fops
            # instances on same volume
            my $suffix = int( rand(1000000) );
            $path .= '/' . $suffix;

            my %fops_opts = (
                'operations'                 => '-o',
                'num-of-files'               => '-n',
                'height'                     => '-H',
                'width'                      => '-W',
                'num-of-threads'             => '-t',
                'size-string'                => '-s',
                'dir-prefix'                 => '-P',
                'file-prefix'                => '-p',
                'rename-file-prefix'         => '-q',
                'level'                      => '-l',
                'block-size'                 => '-b',
                'delay-in-operation'         => '-d',
                'delay-between-operations'   => '-D',
                'expression'                 => '-x',
                'offset-string'              => '-f',
                'metrics-string'             => '-M',
                'error-tolerance-percentage' => '-X',
                'file-open-options'          => '-O',
                'use-existing-dir'           => '-e',
                'pick-random-files'          => '-r',
            );

            my $fops_cmd = "/usr/software/test/bin/fops ";
            foreach my $arg_name ( keys(%fops_opts) ) {
                if ( defined( $input{$arg_name} ) ) {
                    $fops_cmd .= "$fops_opts{$arg_name} $input{$arg_name} ";
                }
            }
            $fops_cmd .= $path;
            $fops_cmd .= " > $ofile \&";

            # Create directory
            $api->execute_raw_command( command => "mkdir -p $path" );

            try {
                $Log->comment(
                    "Beginning iteration $loop_cnt for volume $vol_name");
                $Log->comment("Removing files before IO from $path");
                $api->execute_raw_command( command => "rm -rf $path/*" );

                my $pid = _run_cmd_get_pid(
                    api_set   => $api,
                    cmd       => $fops_cmd,
                    proc_name => 'fops',
                );

                $vol_fops_info->{$vol_name}->{ofile}     = $ofile;
                $vol_fops_info->{$vol_name}->{iteration} = 1;
                $vol_fops_info->{$vol_name}->{pid}       = $pid;
                $vol_fops_info->{$vol_name}->{path}      = $path;
                $vol_fops_info->{$vol_name}->{cmd}       = $fops_cmd;
                $Log->comment("PID for $vol_name iter 1 = $pid");
                $active_fops_pids->{$pid} = 1;
            } ## end try
            otherwise {
                $exception = shift;
                last;
            };
        } ## end foreach my $v (@volumes)

        if ($exception) {
            $my_return->{objects}   = $obj_ref if $obj_ref;
            $my_return->{end_tick}  = time();
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            return $my_return;
        } ## end if ($exception)

        my $num_done    = 0;
        my $done        = 0;
        my $had_failure = 0;
        try {
            while ( !$done ) {

                # Get list of currently active fops pids
                my @cur_active_pids = _get_proc_pids(
                    api_set   => $api,
                    proc_name => 'fops',
                );

                # Find out which fops processes completed
                my @completed_fops_pids;
                foreach my $a_pid ( keys( %{$active_fops_pids} ) ) {
                    if ( !grep { $_ eq $a_pid } @cur_active_pids ) {
                        push( @completed_fops_pids, $a_pid );
                    }
                }

                # Process each completed fops process
                $Log->comment("Current active PIDs: *@cur_active_pids*");
                $Log->comment("Completed PIDs: *@completed_fops_pids*");
                sleep 5;    # Allow cool down for tool and time for
                            # output to fully redirect to log file
                foreach my $c_pid (@completed_fops_pids) {
                    $Log->comment("Processing completed PID $c_pid");
                    delete $active_fops_pids->{$c_pid};
                    my $cur_vol_name;
                    foreach my $vname ( keys( %{$vol_fops_info} ) ) {
                        if ( $vol_fops_info->{$vname}->{pid} eq $c_pid ) {
                            $cur_vol_name = $vname;
                        }
                    }
                    my $cur_path = $vol_fops_info->{$cur_vol_name}->{path};
                    my $cur_iter
                        = $vol_fops_info->{$cur_vol_name}->{iteration};
                    my $cur_ofile = $vol_fops_info->{$cur_vol_name}->{ofile};

                    my $valid_line = 0;
                    my $ran_fine = $api->execute_raw_command(
                        command => "grep \'xecution time\' $cur_ofile", );
                    $valid_line = 1 if ($ran_fine =~ /seconds/i);

                    my $raw_results = $api->execute_raw_command(
                        command => "grep \'no of errors *\[1-9\]\' $cur_ofile" );

                    my @lines = split( /\n/, $raw_results );
                    foreach my $l (@lines) {
                        if ( $l =~ /no of errors\s+(\d+)/i ) {
                            $num_errors += $1;
                        }
                    } ## end foreach my $l (@lines)

                    if ( !$valid_line ) {
                        $Log->warn("Did not find any recognizable fops output "
                                . "for volume $cur_vol_name on iteration "
                                . $cur_iter);
                    } elsif ($num_errors) {
                        my $msg = "fops reported $num_errors errors for "
                                . "vol $cur_vol_name on iteration $cur_iter";
                        if ($::IGNORE_ERRORS) {
                            $Log->warn($msg);
                        } else {
                            $Log->error($msg);
                        }
                        $had_failure = 1;
                    } ## end elsif ($num_errors)

                    if ( $num_errors && !$::IGNORE_ERRORS) {
                        $done = 1;
                        last;
                    } elsif ( !$num_errors ) {
                        $Log->comment( "Successful fops on $cur_vol_name, "
                            . "iteration $cur_iter" );
                    }

                    # Remove files from working dir
                    $Log->comment("Removing files from $cur_path");
                    $api->execute_raw_command(
                        command => "rm -rf $cur_path/*" );

                    if ( $cur_iter != $num_loops ) {

                        # Start next iteration
                        my $cmd = $vol_fops_info->{$cur_vol_name}->{cmd};
                        my $next_iter = $cur_iter + 1;
                        $Log->comment(
                            "Beginning iteration $next_iter for volume $cur_vol_name"
                        );
                        my $pid = _run_cmd_get_pid(
                            api_set   => $api,
                            cmd       => $cmd,
                            proc_name => 'fops',
                        );
                        $vol_fops_info->{$cur_vol_name}->{pid} = $pid;
                        $active_fops_pids->{$pid} = 1;
                        $Log->comment(
                                  "PID for $cur_vol_name iter $next_iter "
                                . "= $pid" );
                    } else {

                        # Completed all loops, remove log
                        $api->execute_raw_command(
                            command => "rm -rf $cur_ofile" );
                        $num_done++;
                    } ## end else [ if ( $cur_iter != $num_loops)
                    $vol_fops_info->{$cur_vol_name}->{iteration}++;

                } ## end foreach my $c_pid (@completed_fops_pids)

                if ( $num_done == $num_volumes ) {
                    $done = 1;
                    last;
                }
                sleep 30;
            } ## end while ( !$done )
        } ## end try
        otherwise {
            $exception = shift;
        };

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{objects}   = $obj_ref if $obj_ref;
            return $my_return;
        } elsif ($had_failure) {
            # Kill all active fops pids
            foreach my $a_pid ( keys( %{$active_fops_pids} ) ) {
                $api->execute_raw_command( command => "kill -9 $a_pid" );
            }
            my $msg = 'fops reported failures against one or more volumes';
            if ($::IGNORE_ERRORS) {
                $my_return->{result} = 1;
                $Log->warn($msg);
            } else {
                $my_return->{result} = 0;
                $Log->error($msg);
            }
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref if $obj_ref;
        } else {
            # Remove fops log file base dir
            $api->execute_raw_command( command => "rm -rf $ofile_base" );
            $my_return->{result} = 1;
            my $msg = "Ran fops on $client_name against $num_volumes volumes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref if $obj_ref;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( $num_clients > 1 )
} ## end sub fops

=pod

=head2 filersio_asyncio_active

Run filersio asyncio_active on the filer.

=head3 opts

%opts can be used to provide additional options for filersio asyncio_active
    nx_context       => <NX context reference>
    validate_only    => <VALIDATE_ONLY option>
    volumes          => <volume array references>
    read_percentage  => <read percentage on traffic>
    rand_percentage  => <random percentage on traffic>
    exp-constant     => <exp_constant>
    blk-size	  	 => <block size>
    start-byte       => <starting byte for the file>
    end-byte         => <ending byte for the file>
    run-time         => <time to run filersio asyncio_active>
    num_active       => <number of active threads>
    filename         => <name of the file>
    runonce          => <1 or 0>
    fill-pattern     => <file content pattern>
    verify-pattern   => <file content pattern>
    synchronous      => <1 or 0>
    interval         => <interval to print stats>
    create           => <1 or 0>
    check_status_interval
    				 => <run filersio status interval>
    num_file         => <number of file to create>
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX filersio_asyncio_active
Filersio Asyncio Active
description
Run filersio asyncio_active on the filer.
required
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
filename;Filename;;Name of the file to be target for workload
start-byte;Start Byte;;Lower bound for I/O access in bytes
end-byte;End-Byte;;Upper bound for I/O access in bytes
run-time;Run Time;;Run time in seconds
optional
create;Create;;Acceptable values are 0|1, which designates if the file is to be created should it not exist
interval;Interval;;Integer value(in seconds) of interval to print statistics
runonce;Runonce;;If the endbyte is reached the test will stop, set 1 to enable the flag
num_file;Number of Files;1;Number of files to create, default is 1
synchronous;Synchronous;;Run the workload synchornously, default is asynchronous. Acceptable values are 0|1
num-active;Number Active;5;Number of active threads for this filersio run, default is 5
rand-percentage;Randomizing Percentage;50;Percentage of randomized traffic, default is 50
read-percentage;Read Percentage;50;Percentage of traffic that is read, default is 50
verify-pattern;Verify Pattern;;Specify file content pattern
blk-size;Block Size;64k;I/O Size, default is 64k
fill-pattern;Fill Pattern;;Specify file content pattern
=end NX filersio_asyncio_active
=cut

sub filersio_asyncio_active {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return;    #Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                volumes => { type => ARRAYREF, optional => 0 },
                create  => { type => SCALAR,   optional => 1 },
                'num-active' =>
                    { type => SCALAR, optional => 1, default => 5 },
                'rand-percentage' =>
                    { type => SCALAR, optional => 1, default => 50 },
                'read-percentage' =>
                    { type => SCALAR, optional => 1, default => 50 },
                'verify-pattern' => { type => SCALAR, optional => 1 },
                'run-time'       => { type => SCALAR, optional => 0 },
                interval         => { type => SCALAR, optional => 1 },
                'blk-size' =>
                    { type => SCALAR, optional => 1, default => '64k' },
                'fill-pattern' => { type => SCALAR, optional => 1 },
                runonce        => { type => SCALAR, optional => 1 },
                'exp-constant' => { type => SCALAR, optional => 1 },
                filename       => { type => SCALAR, optional => 0 },
                num_file => { type => SCALAR, optional => 1, default => 1 },
                synchronous  => { type => SCALAR, optional => 1 },
                'end-byte'   => { type => SCALAR, optional => 0 },
                'start-byte' => { type => SCALAR, optional => 0 },
                _num_nodes   => { type => SCALAR, optional => 1 },
            }
        );

        foreach my $key ( keys %input ) {
            if ( $key eq "create" ) {
                if ( defined( $input{$key} )
                    && ( ( $input{$key} != 1 ) && ( $input{$key} != 0 ) ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be either 1 or 0"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "num-active" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be a positive integer"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "rand-percentage" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be a positive integer"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "read-percentage" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be a positive integer"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "verify-pattern" ) {
                if ( defined( $input{$key} )
                    && ( ( $input{$key} != 1 ) && ( $input{$key} != 0 ) ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be either 1 or 0"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "run-time" ) {
                if ( !defined( $input{$key} ) || ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "input($input{$key}) value need to be defined and to be a positive integer"
                    );
                } ## end if ( !defined( $input{...
            } elsif ( $key eq "interval" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be a positive integer"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "blk-size" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+k$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be in a format of a positive integer followed by k"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "exp-constant" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be a positive integer"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "filenames" ) {
                if ( !defined( $input{$key} ) ) {
                    NATE::BaseException->throw(
                        "input($input{$key}) need to be defined");
                }
            } elsif ( $key eq "synchronous" ) {
                if ( defined( $input{$key} )
                    && ( ( $input{$key} != 1 ) && ( $input{$key} != 0 ) ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be either 1 or 0"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "start-byte" ) {
                if ( !defined( $input{$key} )
                    || ( $input{$key} !~ /^\d+[kmgtKMGT]*$/ ) )
                {
                    NATE::BaseException->throw(
                        "input($input{$key}) value need to be defined and to be a positive integer followed by a unit"
                    );
                } ## end if ( !defined( $input{...
            } elsif ( $key eq "end-byte" ) {
                if ( !defined( $input{$key} )
                    || ( $input{$key} !~ /^\d+[kmgtKMGT]*$/ ) )
                {
                    NATE::BaseException->throw(
                        "input($input{$key}) value need to be defined and to be a positive integer"
                    );
                } ## end if ( !defined( $input{...
            }
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };
    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}
            = "filersio asyncio_active action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::IO::stop_workers";
    }

    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;

    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v (@volume_objs) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined( $node_hash->{$name} ) ) {
                push( @nodes, $node );
                $node_hash->{$name} = 1;
            }
            push( @{ $node_to_volumes->{$name} }, $v );
        } ## end foreach my $v (@volume_objs)
        $num_nodes = @nodes;
    } else {

        # Irrelevant
    }

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {

            my %filersio_opts = %input;

            my $id = 'Filersio_N_DaC' . $proc_cnt;
            push( @ids, $id );

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            $filersio_opts{_num_nodes} = $num_nodes;
            $filersio_opts{volumes}    = $node_to_volumes->{ $n->node() };
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'filersio_asyncio_active',
                args    => [%filersio_opts],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $n (@nodes)

        #Collect results
        my $res = collect_worker_results(
            workers => \@workers,
            futures => $futures,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran filersio asyncio_active on " . @nodes . " nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {    #Single node
        delete $input{nx_context};
        delete $input{validate_only};
        delete $input{volumes};
        delete $input{command_interface};
        my $check_status_interval = $input{check_status_interval};
        delete $input{check_status_interval};
        my $num_file = $input{num_file};
        delete $input{num_file};
        my $root_filename = $input{filename};
        delete $input{filename};
        delete $input{_num_nodes};

        my $n         = $volume_objs[0]->command_interface();
        my $node_name = $n->name();

        try {
            $input{'privilege-level'} = 'test';
            my $apiset_obj = $n->get_7m_or_nodescope_apiset();
            my $volume_obj;
            foreach $volume_obj (@volume_objs) {
                for ( my $i = 0; $i < $num_file; $i++ ) {
                    $input{'filename'}
                        = "/vol/"
                        . $volume_obj->{'volume'} . "/"
                        . $root_filename . "_$i";
                    $apiset_obj->filersio_asyncio_active(%input);
                } ## end for ( my $i = 0; $i < $num_file...
            } ## end foreach $volume_obj (@volume_objs)

            # Check to make sure it's still running halfway and 80% into
            # the run-time
            sleep int($input{'run-time'} / 2);
            $apiset_obj->filersio_status("connectrec-timeout" => 600);
            sleep int($input{'run-time'} * 0.3);
            $apiset_obj->filersio_status("connectrec-timeout" => 600);
        } otherwise {
            my $e = shift;
            if ($::IGNORE_ERRORS) {
                $Log->warn("Recevied following error running filersio: "
                    . $e->text());
            } else {
                $exception = $e;
            }
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran filersio on " . @volume_objs . " volumes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( !$input{_num_nodes...
}    ## End filersio asyncio_active

sub _setup_nfs {
    my %input   = @_;
    my @volumes = @{ $input{volumes} };
    my $unique_nodes;
    my $vs_on_nodes;

    foreach my $v (@volumes) {

        # Get all unique node command_interfaces
        my $n      = $v->command_interface();
        my $n_name = $n->name();
        if ( !$unique_nodes->{$n_name} ) {
            $unique_nodes->{$n_name} = $n;
        }

        if ( $n->mode =~ /cmode/i ) {
            my $vs = $v->vserver();
            push( @{ $vs_on_nodes->{$n_name} }, $vs );
        } else {    # 7Mode
            my $api      = $n->get_7m_or_nodescope_apiset();
            my $vol_name = $v->volume();

            # Note, this may leave an extra line in /etc/exports
            $Log->comment("Changing export rules for volume $vol_name");
            $api->exportfs(
                "path"               => "/vol/$vol_name",
                "persistent"         => 1,
                "persistent-options" => 'anon=0',
            );
        } ## end else [ if ( $n->mode =~ /cmode/i)
    } ## end foreach my $v (@volumes)

    foreach my $n_name ( keys %{$unique_nodes} ) {
        if ( $vs_on_nodes->{$n_name} ) {

            # This node is CMode and has VServers
            my @vs = @{ $vs_on_nodes->{$n_name} };
            foreach my $vs (@vs) {
                _configure_cmode_nfs(
                    command_interface => $unique_nodes->{$n_name},
                    vserver           => $vs,
                    client_type       => 'unix',
                );
            } ## end foreach my $vs (@vs)
        } ## end if ( $vs_on_nodes->{$n_name...
    } ## end foreach my $n_name ( keys %...
}    # End of _setup_nfs

sub _configure_cmode_nfs {
    my %input   = @_;
    my $ci      = $input{command_interface};
    my $type    = $input{client_type};
    my $vserver = $input{vserver};

    $Log->comment(
        "Configuring CMode NFS on Vserver $vserver on node " . $ci->name() );

    if ( $type =~ /unix/i ) {

        # Configure VServer to support NFS
        my $vserver_nfs_exists = 0;
        try {
            my $cmode_apiset = $ci->apiset();
            my $tmp          = $cmode_apiset->vserver_nfs_show(
                vserver => $vserver,
                fields  => "access,v3,v4.0,tcp,udp",
            )->get_parsed_output()->[0];
            $vserver_nfs_exists = 1;
        } ## end try
        catch NACL::APISet::Exceptions::NoMatchingEntriesException with {

            # The exception will be thrown if the vserver doesn't exist
            NACL::STask::VserverNFS->create(
                command_interface => $ci,
                vserver           => $vserver,
                access            => "true",
                'v4.0'            => "enabled",
                'v3'              => "enabled",
                udp               => "enabled",
                tcp               => "enabled"
            );
        };
        if ($vserver_nfs_exists) {
            $Log->comment(
                "Vserver NFS settings exist for $vserver, modifying...");
            NACL::STask::VserverNFS->modify(
                command_interface => $ci,
                vserver           => $vserver,
                access            => "true",
                'v4.0'            => "enabled",
                'v3'              => "enabled",
                udp               => "enabled",
                tcp               => "enabled"
            );
        } ## end if ($vserver_nfs_exists)
    } else {
        $Log->warn("Don't know how to configure client type $type");
    }
} ## end sub _configure_cmode_nfs

=head2 diskhogger

This will call the diskhogger tool from /usr/software on the client with the
specified parameters.

The input is a hash (%opts) which must include the following:

    no_of_root_dirs  => number of root directories to create
    dir_depth        => number of nested directories in each root directory
    file_count       => number of files to be created in each directory
    size             => kb/mb/gb/cu
                        kb - creates 4Kb file
                        mb - creates 1Mb file
                        gb - creates 1Gb file
                        cu - custom file size in multiples of 4Kb
    no_of_4kb_blocks => (required if size=cu) number of 4Kb blocks
    volumes          => array ref of volume object
    client_command_interface => array ref of client objects

head3 %opts, additional options that can be overridden
    dedup_percent   => no default, dedup percentage for custom file size
                        Note: Can only be used if no_of_4kb_blocks is >= 10
    skip_nfs_setup  => default 0, change to 1 if volumes are already set for nfs export
    path            => dh <default_mount_path>/diskhogger, no slashes required,
                        slashes will be stripped
    nocleanup       => default 0, 0 to remove files after diskhogger, 1 to leave it
    duration        => no default value, specify integer of seconds to run
    repeat          => default 1, amount of times to run diskhogger with specified params

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX diskhogger
Diskhogger
description
Run the diskhogger tool from /usr/software from a client.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
no_of_root_dirs;Number Of Root Directories;;Number of root directories to create
dir_depth;Directory Depth;;Number of nested directories in each root directory
file_count;File Count;;Number of files to be created in each directory
size;Size;;Size of files to create, Acceptable values are kb|mb|gb|cu. kb=4Kb, mb=1Mb, gb=1Gb, cu = size in multiples of 4k
optional
no_of_4kb_blocks;Number Of 4kb Blocks;;Number of 4k blocks. Required if size=cu.
dedup_percent;Dedup Percent;;dedup percentage for custom file size, can only be used if number of 4k blocks is >= 10
duration;Duration;;Integer value of run time in seconds
path;Diskhogger Directory;dh;Name of directory to run diskhogger, default is dh. Slashes will be stripped.
nocleanup;No Cleanup;;Remove files after diskhogger, default is 0. Set as 1 to enable
mapping;Mapping;1-to-many;Mapping of operations to volumes. Acceptable values 1-to-many|1-to-1 default is 1-to-many.
repeat;Repeat;1;Amount of times to run diskhogger with specified parameters. Default is 1.
=end NX diskhogger
=cut

sub diskhogger {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire}; # No context requirements

    my $my_return; # Hash reference containing return values
    my $exception;
    my $obj_ref;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                validate_only    => { type => SCALAR,   optional => 0, default => 0 },
                client_command_interface => { type => ARRAYREF, optional => 0 },
                volumes          => { type => ARRAYREF, optional => 0 },
                no_of_root_dirs  => { type => SCALAR,   optional => 0 },
                dir_depth        => { type => SCALAR,   optional => 0 },
                file_count       => { type => SCALAR,   optional => 0 },
                size             => { type => SCALAR,   optional => 0 },
                no_of_4kb_blocks => { type => SCALAR,   optional => 1 },
                dedup_percent    => { type => SCALAR,   optional => 1 },
                mapping_position => { type => SCALAR,   optional => 1 },
                duration         => { type => SCALAR,   optional => 1 },
                skip_nfs_setup   => { type => SCALAR,   optional => 1, default => 0 },
                path             => { type => SCALAR,   optional => 1, default => 'dh' },
                nocleanup        => { type => SCALAR,   optional => 1, default => 0 },
                mapping => { type => SCALAR,   optional => 1, default => '1-to-many'},
                repeat           => { type => SCALAR,   optional => 1, default => 1 },
            }
        );

        foreach my $key (keys %input) {
            if( $key eq "path") {
                #strip the slashes off the path
                $input{'path'} =~ s/\/+//g;
                if ($input{$key} !~ /\w+/) {
                    NATE::BaseException->throw("\'$key\' only accepts alphanumeric, dashes and underscores");
                }
            } elsif ( $key eq "no_of_root_dirs"
                    || $key eq "dir_depth"
                    || $key eq "file_count"
                    || $key eq "repeat"
                    || $key eq "duration") {
                if ($input{$key} !~ /^\d+$/) {
                    NATE::BaseException->throw("\'$key\' value must be an integer.");
                }
            } elsif ( $key eq "size") {
                if ($input{$key} !~ /[kb|mb|gb|cu]$/i) {
                    NATE::BaseException->throw("\'$key\' value must be kb,mb,gb or cu.");
                }
            } elsif ( $key eq "no_of_4kb_blocks"
                     || $key eq "dedup_percent") {
                if ($input{size} !~ /^cu$/i) {
                    NATE::BaseException->throw("To use \'$key\', size($input{size}) must be \'cu\'.");
                } elsif ($input{$key} !~ /^\d+$/) {
                    NATE::BaseException->throw("\'$key\' value must be an integer.");
                } elsif ($key eq "dedup_percent" && ($input{$key} > 100)) {
                    NATE::BaseException->throw("\'$key\' cannot be greater than 100.");
                }
            } elsif ( $key eq "skip_nfs_setup" || $key eq "nocleanup") {
                if ($input{$key} != 0 && $input{$key} != 1) {
                    NATE::BaseException->throw("\'$key\' must be 0 or 1.");
                }
            } elsif ( $key eq "mapping") {
                if ($input{$key} ne "1-to-many" && $input{$key} ne "1-to-1") {
                    NATE::BaseException->throw("\'$key\' must be \'1-to-many\' or \'1-to-1\'")
                }
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "diskhogger action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::IO::stop_workers";
        delete $input{validate_only};
    }

    $my_return->{start_tick} = time();

    my @client_objs = @{$input{client_command_interface}};
    my $num_clients = @client_objs;
    my @volume_objs = @{$input{volumes}};

    #Set up NFS and export all volumes
    _setup_nfs( volumes => $input{volumes} ) if !$input{skip_nfs_setup};

    if ($num_clients > 1) {
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $c ( @client_objs ) {
            my $id = 'diskhogger_C_DaC.'.$proc_cnt; # (C)lient_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{client_command_interface} = [ $c ];
            $input{skip_nfs_setup}           = 1;
            $input{mapping_position}         = ( $proc_cnt - 1);

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'diskhogger',
                args    => [ %input ],
            );
            push( @Pkg_Workers, $worker );
            push( @workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
            merge_context   => 0,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $client_ref = $res->{$rid}->{objects}->{client};
                foreach my $client_name (keys %{$client_ref} ) {
                    my @mnt_objs = @{$client_ref->{$client_name}};
                    push( @{$obj_ref->{client}->{$client_name}}, @mnt_objs );
                }
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            my $msg = 'One or more procs had failures';
            $Log->error( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran diskhogger on $num_clients clients";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
        }
        return $my_return;
    } else {
        my $client      = shift(@client_objs);
        my $client_name = $client->name();
        my $vol_ref     = delete $input{volumes};
        my $path        = delete $input{path};
        my @vols        = @{$vol_ref};
        my $nocleanup   = delete $input{nocleanup};
        my $num_loops   = delete $input{repeat};
        $num_loops = 1 if !$num_loops;
        my $duration    = delete $input{duration};
        delete $input{client_command_interface};
        delete $input{skip_nfs_setup};
        delete $input{validate_only};
        my $vol_dh_info;

        #Increase Timeout
        my $api = $client->apiset();
        $api->set_timeout('connectrec-timeout' => 600);

        my $to = $api->get_timeout();
        if ($to == 600) {
            $Log->comment("Increased client timeout to 600");
        } else {
            $Log->error("Failed to increaes client timeout");
        }

        #Change @vols if mapping is 1-to-1
        if($input{mapping} eq '1-to-1') {
            @vols = $vols[ $input{mapping_position} ];
        }
        my $num_volumes = @vols;
        delete $input{mapping_position};
        delete $input{mapping};

        #Mount volumes
        my $mounts = _mount_client_volumes( client => $client, volumes => \@vols );
        push(@{$obj_ref->{client}->{$client_name}}, @{$mounts->{mounts}});

        #Create log directory
        my $ofile_base = 'diskhogger_';
        $ofile_base .= localtime(time());
        $ofile_base =~ s/[\s:]//g;
        srand();
        $ofile_base .= '_' . (int(rand(10000))+1) . '/';
        $api->execute_raw_command( command => "mkdir -p $ofile_base" );

        my $active_dh_pids;
        my @dh_list; #Track individual diskhogger paths for cleanup
        foreach my $v (@vols) {
            my $vol_name = $v->volume();
            my $p = $mounts->{paths}->{$vol_name};
            srand();
            my $r = int(rand(99)); #avoid collision on timestamp
            $input{path} = $p.'/'.$path.'/'.$client_name.'_'.time().$r;
            my $ofile = $ofile_base.$vol_name;

            my $dh_cmd = 'diskhogger32 ';
            my $cmd_rsp = $api->execute_command( command => $dh_cmd );
            if($cmd_rsp =~ /Command not found/i) {
                $Log->comment("Diskhogger32 command not found, using default location");
                $dh_cmd = '/u/phillipg/c/diskhogger32 ';
            }
            my $params = '';
            foreach my $k (keys %input) {
                $params .= "$k=$input{$k} ";
            }

            $dh_cmd .= $params;
            # 2>&1 outputs stdout and stderr
            $dh_cmd .= " > $ofile 2>\&1 \&";

            $api->execute_command( command => "mkdir -p $input{path}",
                                   'privilege-level' => "root" );
            $api->execute_command( command => "chmod 777 $input{path}",
                                   'privilege-level' => "root" );

            try {
                $Log->comment("Starting diskhogger for $vol_name.");
                my $pid = _run_cmd_get_pid(
                    api_set   => $api,
                    cmd       => $dh_cmd,
                    proc_name => 'diskhogger',
                );

                $vol_dh_info->{$vol_name}->{ofile}     = $ofile;
                $vol_dh_info->{$vol_name}->{pid}       = $pid;
                $vol_dh_info->{$vol_name}->{path}      = $input{path};
                $vol_dh_info->{$vol_name}->{cmd}       = $dh_cmd;
                $vol_dh_info->{$vol_name}->{iteration} = 1;
                $Log->comment("PID for $vol_name = $pid");
                $active_dh_pids->{$pid} = 1;
                push(@dh_list,$input{path});
            } otherwise {
                $exception = shift;
                if ($::IGNORE_ERRORS) {
                    $Log->warn( "Got following exception, ignoring and "
                              . "continuing: " . $exception->text() );
                    $exception = "";    #Reset the exception.
                }
            };
        }

        if ($exception) {
            $my_return->{objects} = $obj_ref if $obj_ref;
            $my_return->{end_tick} = time();
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
            return $my_return;
        }

        my $num_done = 0;
        my $num_errors = 0;
        my $had_failure = 0;
        my $done = 0;
        my $start = time();
        my $duration_met = 0;
        try {
            while (!$done) {
                # Get list of currently active diskhogger pids
                my @cur_active_pids = _get_proc_pids(
                    api_set     => $api,
                    proc_name   => 'diskhogger',
                );

                # Find out which diskhogger processes completed
                my @completed_dh_pids;
                foreach my $a_pid (keys(%{$active_dh_pids})) {
                    if ((!grep {$_ eq $a_pid} @cur_active_pids) || $duration_met ) {
                        $Log->comment("Pushing $a_pid into completed list");
                        push(@completed_dh_pids, $a_pid);
                    }
                }
                # Process each completed diskhogger process
                $Log->comment("Current active PIDs: *@cur_active_pids*");
                $Log->comment("Completed PIDs: *@completed_dh_pids*");
                sleep 5; # Allow cool down for tool and time for
                         # output to fully redirect to log file
                foreach my $c_pid (@completed_dh_pids) {
                    delete $active_dh_pids->{$c_pid};
                    my $cur_vol_name;
                    foreach my $vname (keys(%{$vol_dh_info})) {
                        if ( $vol_dh_info->{$vname}->{pid} eq $c_pid ) {
                            $cur_vol_name = $vname;
                        }
                    }
                    my $cur_path  = $vol_dh_info->{$cur_vol_name}->{path};
                    my $cur_ofile = $vol_dh_info->{$cur_vol_name}->{ofile};
                    my $cur_iter  = $vol_dh_info->{$cur_vol_name}->{iteration};

                    my $raw_results = $api->execute_raw_command(
                        command => "cat $cur_ofile",
                    );
                    my $err_msg = '';
                    my $valid_line = 0;
                    my @lines = split(/\n/, $raw_results);
                    if ((@lines == 1) && $lines[0] !~ /^Start\:/i) {
                        $num_errors++;
                        $err_msg = "Diskhogger failed to start: $raw_results";
                        $had_failure = 1;
                    } elsif ((@lines > 1) && $lines[1] !~ /^End\:/i) {
                        $num_errors++;
                        $err_msg = "Diskhogger failed midway: $raw_results";
                        $had_failure = 1;
                    }

                    if($had_failure) {
                        if($::IGNORE_ERRORS) {
                            $Log->warn("Diskhogger encountered an error, "
                                      ."but will continue: $err_msg ");
                        } else {
                            $Log->error("Diskhogger encountered an error: $err_msg");
                        }
                    }

                    if( !$num_errors ) {
                        $Log->comment("Successfully executed diskhogger on $cur_vol_name");
                    }
                    if( !$num_errors || ($num_errors && $::IGNORE_ERRORS) ) {
                        $Log->comment("Removing logs for $cur_vol_name");
                        $api->execute_raw_command(  command => "rm -rf $cur_ofile" );
                        
                        $num_done++;

                        if( $cur_iter != $num_loops ) {
                            $Log->comment("Cleanup folder before new iteration");
                            $api->execute_command( command => "rm -rf $cur_path/*",
                                               'privilege-level' => 'root');
                            my $next = $cur_iter + 1;
                            my $cmd  = $vol_dh_info->{$cur_vol_name}->{cmd};
                            $Log->comment("Beginning iteration $next for volume $cur_vol_name");
                            my $pid = _run_cmd_get_pid(
                                api_set   => $api,
                                cmd       => $cmd,
                                proc_name => 'diskhogger',
                            );
                            $vol_dh_info->{$cur_vol_name}->{pid} = $pid;
                            $active_dh_pids->{$pid} = 1;
                            $Log->comment("$cur_vol_name iteration $next has pid $pid");
                            $vol_dh_info->{$cur_vol_name}->{iteration}++;
                        }
                    } else {
                        $done = 1;
                        last;
                    }
                }

                if($num_done == ($num_volumes * $num_loops)) {
                    $done = 1;
                    last;
                }

                if($duration) {
                    $Log->comment("Duration is set at $duration");
                    $Log->comment("time is: ".time());
                    my $end = $start + $duration;
                    $Log->comment("Finish time should be: $end");
                }

                ## Duration has been met
                if($duration && (time() >= ($start + $duration))) {
                    $Log->comment("Diskhogger duration met, exiting now");
                    foreach my $a_pid (keys(%{$active_dh_pids})) {
                        $api->execute_raw_command(command => "kill -9 $a_pid");
                    }
                    $duration_met = 1;
                }
                sleep 30;
            }
        } otherwise {
            $exception = shift;
            if ($::IGNORE_ERRORS) {
                $Log->warn( "Got following exception, ignoring and "
                            . "continuing: " . $exception->text() );
                $exception = "";    #Reset the exception.
            }
        };

        ##CLEANUP
        if(!$nocleanup) {
            foreach my $dh_path (@dh_list) {
                $Log->comment("Removing files from the diskhogger runs");
                $api->execute_command( command => "rm -rf $input{path}",
                                      'privilege-level' => "root",
                                      'connectrec-timeout' => 3600 );
             }
        }

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } elsif ( $had_failure ) {
            $my_return->{result} = 0;
            my $msg = "diskhogger reported failures in one or more volumes";
            if($::IGNORE_ERRORS) {
                $my_return->{result} = 1;
                $msg .= " and will be ignored.";
                $Log->warn( $msg );
            } else {
                #Kill remaining active diskhogger pids
                foreach my $a_pid (keys(%{$active_dh_pids})) {
                    $api->execute_raw_command(command => "kill -9 $a_pid");
                }
                $Log->error( $msg );
            }
            $my_return->{msg} = $msg;
        } else { # Good IO without errors
            $api->execute_raw_command(command => "rm -rf $ofile_base");
            $my_return->{result} = 1;
            my $msg = "diskhogger ran successfully without errors on $client_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
}; # End diskhogger

# Returns hash reference with keys:
# 'paths' -> $vol_name -> $path
# 'mounts' -> [ mount_objects ]
# 'NFS_VERSION' -> [v3 or v4 or v3,v4]
sub _mount_client_volumes {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    my $client = $input{client};
    my $client_name = $client->name();
    my @volumes = @{$input{volumes}};
    my $my_return;
    my $exception;
    my @vers_array;
    my $numvers = 0;
    my $currver = 0;

    if (!defined ($::NFS_VERSION)) {
        @vers_array = ("3");
    } elsif ($::NFS_VERSION eq "v3,v4") {
        @vers_array = ("3", "4");
    } elsif ($::NFS_VERSION eq "v3") {
        @vers_array = ("3");
    } elsif ($::NFS_VERSION eq "v4") {
        @vers_array = ("4");
    } else {
        @vers_array = ("3");
    }
    
    $numvers = @vers_array;

    foreach my $vol_obj ( @volumes ) {
        my $node = $vol_obj->command_interface();
        # Mount volume on client
        my $vol_name = $vol_obj->volume();

        try {
            $Log->comment("Mounting volume $vol_name on $client_name");
    
            # Check if volume is part of an LS snapmirror relationship
            my $src_sm = NACL::C::Snapmirror->find(
                command_interface => $node,
                filter            => {
                    'source-volume' => $vol_name,
                    'type'          => 'LS',
                },
                allow_empty       => 1,
            );
        
            my $dst_sm = NACL::C::Snapmirror->find(
                command_interface => $node,
                filter            => {
                    'destination-volume' => $vol_name,
                    'type'          => 'LS',
                },
                allow_empty       => 1,
            );

            my $mount;
            my %opts = (
                command_interface   => $client,
                volume              => $vol_obj,
                chmod               => "777",
                vers                => $vers_array[$currver],
            );

            $vol_obj->mount( 'junction-path' => "/$vol_name" );            
            if (($src_sm) || ($dst_sm)) {
                $Log->comment("Volume $vol_name is part of an LS snapmirror"
                    . "relationship, mounting with .admin path");
                $opts{path} = "/\.admin/$vol_name";
                $opts{vserver} = $vol_obj->vserver();
                $opts{server_command_interface} = $node;
            }
            $mount = NACL::STask::Client::Mount->create( %opts );

            $mount->purge();
            sleep 2;
            $mount = NACL::STask::Client::Mount->create( %opts );

            my $path = $mount->mount_point();
            $my_return->{paths}->{$vol_name} = $path;

            push ( @{$my_return->{mounts}}, $mount ); 
            $currver = ($currver + 1) % $numvers;
        } catch NATE::BaseException with {
            $exception = shift;
            if ($::IGNORE_ERRORS){
                $Log->warn ("Got the following exception, ignoring and continuing: ".$exception->text());
                $exception = "";#Reset the exception
            }
        };
    }
    if ($exception) {
        $my_return->{exception} = $exception;
    }

    return $my_return;
} # End _mount_client_volumes

sub _purge_mounts {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    my @mounts     = @{ $input{mounts} };
    my $num_mounts = @mounts;
    my $my_return;
    $my_return->{start_tick} = time();

    my $c           = $mounts[0]->command_interface();
    my $client_name = $c->name();
    my $exception;

    try {
        my $api = $c->apiset();
        $api->set_timeout( 'connectrec-timeout' => 300 );
        foreach my $m (@mounts) {

            # Remove client directory name
            my $mp = $m->mount_point();
            if ( $mp && $client_name ) {
                $api->execute_raw_command(
                    command => "rm -rf $mp/$client_name" );
            }
            $m->purge();
        } ## end foreach my $m (@mounts)
    } ## end try
    otherwise {
        $exception = shift;
    };

    $my_return->{end_tick} = time();
    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Purged $num_mounts mounts on client $client_name";
        $Log->comment($msg);
        $my_return->{msg} = $msg;
    } ## end else [ if ($exception)
    return $my_return;

}    # End _purge_mounts

sub _run_cmd_get_pid {
    my (%input)   = @_;
    my $api       = $input{api_set};
    my $proc_name = $input{proc_name};

    my @pids_before = _get_proc_pids(
        api_set   => $api,
        proc_name => $proc_name,
    );

    # Run command
    $api->execute_raw_command( command => $input{cmd} );
    my $the_pid;
    my $attempts = 0;
    while (!$the_pid) {
        $attempts++;

        my @pids_after = _get_proc_pids(
            api_set   => $api,
            proc_name => $proc_name,
        );
    
        foreach my $p (@pids_after) {
            if ( !grep { $_ eq $p } @pids_before ) {
                $the_pid = $p;
            }
        }
        last if ($attempts > 120); # 2 minutes
        sleep 1;
    }
    return $the_pid;
} ## end sub _run_cmd_get_pid

sub _get_proc_pids {
    my (%input) = @_;
    my @pids;
    my $api       = $input{api_set};
    my $proc_name = $input{proc_name};

    my $raw
        = $api->execute_raw_command( command => "ps -ef|grep $proc_name" );
    my @lines = split( /\n/, $raw );
    foreach my $l (@lines) {
        if ( ( $l =~ /^\S+\s+(\d+)\s+\d+\s+\d+/ ) && ( $l !~ /grep/ ) ) {
            push( @pids, $1 );
        }
    }
    return @pids;
} ## end sub _get_proc_pids

###########################################################

=pod

=head2 sio

Run sio on specified volumes on specified clients in parallel. 
Each client requires one login session to perform all sio operations.

Either of the sets (start, end) or (num_offset_runs,range_min_size,range_type)
should be specified in XML file.
<start> & <end> are specified for single sio run on each file.
(num_offset_runs, range_min_size, range_type) are specified to perform 
multiple sio runs on each file. 

Either <path> or <volumes> should be specified in the XML file.
If <path> is specified, sio is run on file mentioned in the path.
If <volumes> is specified, function creates sio file on volume, mounts 
the volume and runs sio on the created files.

The input is a hash (%input) which must include:
    client_command_interface => [ $client ] # Array ref of client objects

%inputs can be used to provide additional options for sio, which include
    reads             => <read%>   param                     default =>50
    randpercent       => <rand%>   param                     default => 80
    blksz             => <blksz>   param                     default => '4k'
    start             => <start>   param                     
    end               => <end>     param                     
    secs              => <secs>    param                     default => 10
    threads           => <threads> param                     default => 2
    filesize          => Size of each siofile                default => 100k
    filecount         => Number of sio files on each volume  default => 1
    should_remove     => Set to 1 if you want to remove
                         siofile after sio run completes     default => 0
    volumes           => [ $vol ]  # Array ref of volume
                         objects after sio run completes
    file_names        => Comma delimited file names if sio should be run against
                         specific files on the specified volumes.
    num_offset_runs   => Number of sio runs on each siofile
    range_type        => range type can be split | random
                         split  => split <start> and <end> equally
                         random => split <start> and <end> randomly
                         with 'range_min_size' diff between <start> and <end>
    range_min_size    => Difference between <start> and <end> when 'range_type'
                         is random
    if_exists         => If the SIO file already exists, then either 'reuse'
                         or 'purge'. Defaults to reuse.
    base_name         => Base name of files for sio. Default is siofile, or
                         siofile.<x> for multiple files. Specify this param
                         to replace "siofile" with a custom base name.

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=cut

=cut begin NX sio
Sio
description
Run sio on specified volumes on specified clients in parallel.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
optional
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
reads;Read Percentage;50;Percentage of access that are reads, default is 50
blksz;Block Size;4k;I/O size, default is 4k
start;Start Byte;;Lower bound of I/O access in bytes
end;End Byte;;Upper bound of I/O access in bytes
secs;Run Time;10;Run time in seconds, default is 10
threads;Threads;2;Amount of threads to use, default is 2
filecount;Filecount;1;Number of sio files on each volume, default is 1
file_names;File Names;;Comma delimited file names if sio should be run against specific files on the specified volumes
num_offset_runs;Number of Offset Runs;;Number of sio runs on each siofile
range_type;Range Type;;Acceptable values are split|random
range_min_size;Range Min Size;;Minimum difference between the end byte and start byte when range type is random
randpercent;Randomize Percentage;;Percentage of randomization in I/O
should_remove;Postrun Cleanup;0;Remove siofile after sio run, default is 0. Set to 1 to enable flag
filesize;File Size;100k;Size for siofile, default is 100k
base_name;File base name;siofile.<x>;Base name of files for sio. Default is siofile, or siofile.<x> for multiple files. Specify this param to replace "siofile" with a custom base name.
if_exists;If file exists;Re-use;If the SIO file already exists, then either 'reuse' or 'purge'. Defaults to reuse.
=end NX sio
=cut

sub sio {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};    # No context requirements

    my $my_return;    # Hash reference containing return values
    my $exception;
    my $obj_ref;
    my $should_remove;

    my $num_multirun_params_found  = 0;
    my $using_multitrun_mode       = 0;
    my $num_singlerun_params_found = 0;
    my $using_singlerun_mode       = 0;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   client_command_interface =>
                    { type => ARRAYREF, optional => 0 },
                volumes => { type => ARRAYREF, optional => 0 },
                validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                reads => { type => SCALAR, optional => 1, default => 50 },
                randpercent =>
                    { type => SCALAR, optional => 1, default => 80 },
                blksz   => { type => SCALAR, optional => 1, default => '4k' },
                start   => { type => SCALAR, optional => 1 },
                end     => { type => SCALAR, optional => 1 },
                secs    => { type => SCALAR, optional => 1, default => 10 },
                threads => { type => SCALAR, optional => 1, default => 2 },
                skip_nfs_setup =>
                    { type => SCALAR, optional => 1, default => 0 },
                filecount => { type => SCALAR, optional => 1, default => 1 },
                filesize =>
                    { type => SCALAR, optional => 1, default => '100k' },
                mapping_position => { type => SCALAR, optional => 1, },
                should_remove =>
                    { type => SCALAR, optional => 1, default => 0 },
                file_names      => { type => SCALAR, optional => 1 },
                num_offset_runs => { type => SCALAR, optional => 1 },
                range_type      => { type => SCALAR, optional => 1 },
                range_min_size  => { type => SCALAR, optional => 1 },
                ofileid         => { type => SCALAR, optional => 1 },
                if_exists       => { type => SCALAR, optional => 1, default => 'reuse' },
                base_name       => { type => SCALAR, optional => 1, default => 'siofile' },
            }
        );

        foreach my $key ( keys %input ) {
            if ( $key eq "reads" ) {
                if (   ( $input{$key} !~ /^\d+$/ )
                    || ( 0 > $input{$key} )
                    || ( $input{$key} > 100 ) )
                {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer in the range [0,100]");
                } ## end if ( ( $input{$key} !~...
            } elsif ( $key eq "randpercent" ) {
                if (   ( $input{$key} !~ /^\d+$/ )
                    || ( 0 > $input{$key} )
                    || ( $input{$key} > 100 ) )
                {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer in the range [0,100]. 0% implies sequential accesses"
                    );
                } ## end if ( ( $input{$key} !~...
            } elsif ( $key eq "blksz" ) {
                if ( $input{$key} !~ /^\d+[kKmMgG]?$/ ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer optionally followed by k,m or g"
                    );
                }
            } elsif ( $key eq "start" ) {
                if ( $input{$key} !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$key\' must be an integer");
                }
                $num_singlerun_params_found++;
            } elsif ( $key eq "end" ) {
                if ( $input{$key} !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$key\' must be an integer");
                }
                $num_singlerun_params_found++;
            } elsif ( $key eq "secs" ) {
                if ( $input{$key} !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$key\' must be an integer");
                }
            } elsif ( $key eq "threads" ) {
                if ( $input{$key} !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$key\' must be a number");
                }
            } elsif ( $key eq "filesize" ) {
                if ( ( $input{$key} !~ /\d+[kKmMgG]?$/ ) ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer optionally followed by k,m or g"
                    );
                }
            } elsif ( $key eq "filecount" ) {
                if ( $input{$key} !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$key\' must be a number");
                }
            } elsif ( $key eq "num_offset_runs" ) {
                if ( $input{$key} !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$key\' must be a number");
                }
                $num_multirun_params_found++;
            } elsif ( $key eq "range_type" ) {
                if ( $input{$key} !~ /^split|random$/i ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be 'split' or 'random'");
                }
                $num_multirun_params_found++;
            } elsif ( $key eq "range_min_size" ) {
                if ( $input{$key} !~ /\d+[kKmMgG]?$/ ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be a number optionally followed by k,m or g"
                    );
                }
                $num_multirun_params_found++;
            } elsif ( $key eq "if_exists" ) {
                if ( $input{$key} !~ /^reuse|purge$/i ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be 'reuse' or 'purge'");
                }
            } elsif ( $key eq 'file_names' ) {
                if ( $input{$key} !~ /\S+/ ) {
                    NATE::BaseException->throw(
                        "File name param is empty. Please specifiy one or "
                        . "more file names");
                }
                $Log->comment("\'file_names\' param specified, ignoring "
                    . "filecount and base_name params");
            } elsif ( $key eq 'base_name' ) {
                if ( $input{$key} !~ /\S+/ ) {
                    NATE::BaseException->throw("Base name param is empty.");
                }
            }
        } ## end foreach my $key ( keys %input)

        if ( $num_multirun_params_found == 3 ) {
            $using_multitrun_mode = 1;
        }
        if ( $num_singlerun_params_found == 2 ) {
            $using_singlerun_mode = 1;
        }

        # Check to see both (start & end) and (num_offset_runs,range_min_size,range_type)
        # are not specified in same XML file.
        if ( $using_multitrun_mode && $using_singlerun_mode ) {
            NATE::BaseException->throw(
                "You can only specify either single run parameters (start & end) "
                    . "or multirun parameters(num_offset_runs,range_min_size,range_type)"
                    . " at one time" );
        } elsif ( !$using_multitrun_mode && !$using_singlerun_mode ) {
            NATE::BaseException->throw("Start/end parameters not specified");
        }
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "sio action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::IO::stop_workers";
        delete $input{validate_only};
    } ## end else [ if ($exception)

    my @client_objs = @{ $input{client_command_interface} };
    my $num_clients = @client_objs;
    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;
    my $filecount   = $input{'filecount'};
    my $filesize    = $input{'filesize'};
    $should_remove = $input{should_remove};
    my $start;
    my $end;
    my @offsets;
    my $range_min_size;
    my $num_offset_runs = 1;
    my $range_type      = $input{'range_type'};
    my $curr_sio_cmd;
    my $failing_msg;

    sleep 10;

    # On all volumes, set up NFS and export
    _setup_nfs( volumes => $input{volumes} ) if !$input{skip_nfs_setup};

    # create <$filecount> siofile(s) using mkfile on each volume
    my $mkfile_to = 60;
    $filesize =~ /(\d+)(\D+)/i;
    my $val = $1;
    my $unit = $2;
    if ($unit =~ /^g|gb$/i) {
        $mkfile_to += (300 * $val) * $filecount;
    } else {
        $mkfile_to += $filecount * 120;
    }
    $Log->comment("Setting mkfile timeout to $mkfile_to seconds");

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_clients > 1 ) {
        # More than one client specified, divide and conquer by splitting
        # work off on each client in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $c (@client_objs) {
            my $id
                = 'sio_C_DaC.' . $proc_cnt;  # (C)lient_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single client.  Keep all other original
            # inputs to pass in
            $input{client_command_interface} = [$c];
            $input{mapping_position}         = ( $proc_cnt - 1 );
            $input{skip_nfs_setup}           = 1;
            $input{ofileid}                  = $id
                ;  #Will be used to distinguish different sio files worked on.

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::IO',
                method  => 'sio',
                args    => [%input],
            );
            push( @Pkg_Workers, $worker );
            push( @workers,     $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $c (@client_objs)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $client_ref = $res->{$rid}->{objects}->{client};
                foreach my $client_name ( keys %{$client_ref} ) {
                    my @mnt_objs = @{ $client_ref->{$client_name} };
                    push( @{ $obj_ref->{client}->{$client_name} },
                        @mnt_objs );
                }
            } ## end if ( $res->{$rid}->{objects...
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            my $msg = 'One or more procs had failures';
            $Log->error($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran sio on $num_clients clients";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {
        srand();
        # Single client, but more than one volume specified, divide and
        # conquer by splitting work off for each volume in parallel
        # Convert filesize from k|m|g to bytes
        if ( $filesize =~ s/k//i ) {
            $filesize = $filesize * 1024;    #k == 1024
        } elsif ( $filesize =~ s/m//i ) {
            $filesize = $filesize * 1024 * 1024;    #m == 1024*1024
        } elsif ( $filesize =~ s/g//i ) {
            $filesize = $filesize * 1024 * 1024 * 1024;   #g == 1024*1024*1024
        }

        # For multi run mode
        # Generate and store offsets, based on range_type specified
        if ( $num_multirun_params_found == 3 ) {
            $using_multitrun_mode = 1;
            $Log->comment(
                "Found multirun parameters specified. Ignoring start and end parameters"
            );
            $num_offset_runs = $input{num_offset_runs};

            if ( $range_type eq "split" ) {
                for ( my $i = 0; $i < $num_offset_runs; $i++ ) {
                    my $offset = int( $i * ( $filesize / $num_offset_runs ) );
                    push( @offsets, $offset );
                }
                $range_min_size = int( $filesize / $num_offset_runs );
            } elsif ( $range_type eq "random" ) {
                $range_min_size = $input{range_min_size};
                if ( $range_min_size =~ s/k//i ) {
                    $range_min_size *= 1024;
                } elsif ( $range_min_size =~ s/m//i ) {
                    $range_min_size *= ( 1024 * 1024 );
                } elsif ( $range_min_size =~ s/g//i ) {
                    $range_min_size *= ( 1024 * 1024 * 1024 );
                }

                srand();
                for ( my $i = 0; $i < $num_offset_runs; $i++ ) {
                    my $offset = int( rand( $filesize + 1 ) );
                    push( @offsets, $offset );
                }
            } ## end elsif ( $range_type eq "random")
        } else {
            $Log->comment(
                "Running in singlerun mode,using start and end parameters\n");
        }

        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $client       = shift(@client_objs);
        my $client_name  = $client->name();
        my $api          = $client->apiset();
        $api->set_timeout( 'connectrec-timeout' => $mkfile_to );
        my $to = $api->get_timeout();
        if ( $to == $mkfile_to ) {
            $Log->comment("Increased client timeout to $mkfile_to");
        } else {
            $Log->warn("Failed to increase client timeout to $mkfile_to");
        }
        my $found_errors = 0;

        my $siocmd;
        my @siocmd;
        my @ofile;
        my $ofile;
        my $pid;
        my $raw_results;
        my $input_type;
        my $active_sio_pids;

        if ( $input{mapping} eq '1-to-1' ) {
            # Perform IO to volume at same array position as
            # the 'mapping_position' param value
            @volume_objs = $volume_objs[ $input{mapping_position} ];
        }

        my $mounts_and_paths = _mount_client_volumes(
            client  => $client,
            volumes => \@volume_objs
        );
        push(
            @{ $obj_ref->{client}->{$client_name} },
            @{ $mounts_and_paths->{mounts} }
        );

        my $bindir = $Params->get('TESTBINDIR');
        my $mkfile = catdir($bindir, 'mkfile');
        foreach my $v (@volume_objs) {
            my $id = 'sio_P_DaC.'
                . $proc_cnt;    # (P)ath_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            my $path    = $mounts_and_paths->{paths}->{ $v->volume() };
            my $ofileid = $input{ofileid} . "_" . $id;

            # If mapping is 1-to-many, create a subdir
            if ( $input{mapping} eq '1-to-many' ) {
                $path .= '/sio_P_DaC.' . $proc_cnt;
            }

            # Create directory
            $api->execute_raw_command( command => "mkdir -p $path" );

            # Get existing list of files 
            my $ls = $api->execute_raw_command(command => "ls -l $path");
            my @ls_entries = split(/\n/, $ls);
            my @ls_files;
            foreach my $e (@ls_entries) {
                my @tmp = split(/\s+/, $e);
                push(@ls_files, $tmp[-1]);
            }

            # Create the files
            my @file_names;
            if ( $input{file_names} ) {
                @file_names = split(/,/, $input{file_names});
            } else {
                if ( $filecount == 1 ) {
                    push( @file_names, $input{base_name} );
                } else {
                    for ( my $i = 0; $i < $filecount; $i++ ) {
                        push( @file_names, $input{base_name}.'.'.$i );
                    }
                }
            }

            foreach my $f (@file_names) {
                my $file_path = catdir($path, $f);
                my $create = 0;
                if (grep{$_ eq $f} @ls_files) {
                    # File exists
                    if ($input{if_exists} eq 'purge') {
                        $Log->comment("File $file_path exists, purging");
                        $api->execute_command(
                            command => "rm -rf $file_path",
                            'privilege-level' => 'root'
                        );
                        $create = 1;
                    }
                } else { # File doesn't exist
                    $create = 1;
                }

                if ($create) {
                    #$api->execute_command(command => "$mkfile $filesize $file");
                    # Workaround until EngSup installs mkfile on rh5 i386 mounts
                    # Divide filesize into 4096 byte blocks to speed up dd command
                    $Log->comment("Creating $file_path");
                    my $count = $filesize / 4096;
                    $api->execute_command(command => "dd count=$count bs=4096 if=/dev/zero of=$file_path");
                } else {
                    $Log->comment("$file_path exists, re-using");
                }
            }

            try {
                if ( !$using_multitrun_mode ) {
                    # for single run mode, construct sio command
                    $siocmd
                        = "sio $input{reads} $input{randpercent} $input{blksz} $input{start} $input{end} $input{secs} $input{threads} ";
                    foreach my $file_name (@file_names) {
                        $siocmd .= " " . "/" . $path . "/" . $file_name;
                        $api->execute_raw_command( command =>
                                "sudo chmod 777 $path/$file_name" );
                    }
                    $ofile = "sioout_" . $ofileid . "_" . int(rand(1000000)) . ".txt";
                    push( @siocmd, $siocmd );
                    push( @ofile,  $ofile );
                    $pid = _run_cmd_get_pid(
                        api_set   => $api,
                        cmd       => "$siocmd  >\& $ofile \&",
                        proc_name => 'sio'
                    );
                    $active_sio_pids->{$pid} = 1;
                } else {
                    # for multi-run mode, construct multiple sio command from
                    # offsets generated earlier.
                    for ( my $i = 0; $i < $num_offset_runs; $i++ ) {

                        $start = $offsets[$i];
                        $end   = $start + $range_min_size - 1;
                        if ( $end >= $filesize ) {
                            $end = $filesize
                                - 2;    #bytes range from 0 to filezie-1;
                        }

                        $siocmd
                            = "sio $input{reads} $input{randpercent} $input{blksz} $start $end $input{secs} $input{threads} ";
                        foreach my $file_name (@file_names) {
                            $siocmd
                                .= " " . "/" . $path . "/" . $file_name;
                            $api->execute_raw_command( command =>
                                    "sudo chmod 777 $path/$file_name" );
                        } ## end foreach my $file_name (@file_names)
                        $ofile = "sioout_" . $ofileid . "_$i" ."_" . int(rand(1000000)) . ".txt";

                        push( @siocmd, $siocmd );
                        push( @ofile,  $ofile );

                        my $pid = _run_cmd_get_pid(
                            api_set   => $api,
                            cmd       => "$siocmd  >\& $ofile \&",
                            proc_name => 'sio'
                        );
                        $active_sio_pids->{$pid} = 1;
                    } ## end for ( my $i = 0; $i < $num_offset_runs...
                } ## end else [ if ( !$using_multitrun_mode)
            } otherwise {
                $exception = shift;
                if ($::IGNORE_ERRORS) {
                    $Log->warn( "Got following exception, ignoring and "
                            . "continuing: "
                            . $exception->text() );
                    $exception = "";    #Reset the exception.
                } ## end if ($::IGNORE_ERRORS)
            };

            $proc_cnt++;
        }    #end running sio on each volume.

        #Wait till all sio commands for this volume are done.
        _wait_till_all_sio_complete( $api, $active_sio_pids );

        # Make sure that the results of sio commands will come in log files.
        my $num = @siocmd;
        for ( my $i = 0; $i < $num; $i++ ) {
            $raw_results
                .= $api->execute_raw_command( command => "echo $siocmd[$i]" );
            $raw_results
                .= $api->execute_raw_command( command => "cat $ofile[$i]" );
        } ## end for ( my $i = 0; $i < $num...

        sleep 10;

        #Kill any remaining (for whatever reason) sio threads.
        foreach my $a_pid ( keys( %{$active_sio_pids} ) ) {
            $api->execute_raw_command( command => "kill -9 $a_pid" );
        }

        #Analyze results and report errors.
        #Note that only the last exception survives.
        my @lines = split( /\n/, $raw_results );
        my $cur_vol;
        foreach my $l (@lines) {
            if ( $l =~ /(sio .*)/ ) {
                $curr_sio_cmd = $1;
                $cur_vol = $curr_sio_cmd;
                $cur_vol =~ /(.+\/t\/\d+\.\d+\.\d+\.\d+\/)/;
                $cur_vol =~ s/$1//;
                my @arr = split(/\//, $cur_vol);
                $cur_vol = $arr[0];
                $cur_vol =~ s/\-(3|4)$//;
                $Log->comment("SIO info for vol $cur_vol");
            }
            if ( $l =~ /SIO_ERROR/i ) {
                $found_errors = 1;               
                $failing_msg .= "\n## Found error on volume $cur_vol\n";
                $failing_msg .= $curr_sio_cmd . "\n";
                $failing_msg .= "Error line is $l" . "\n";
            }
        } ## end foreach my $l (@lines)

        if ( !$found_errors ) {
            if ($should_remove) {
                $Log->comment("Removing files after sio from $input{paths}");
                $api->execute_raw_command(
                    command => "rm -rf $input{paths}/*" );
            }
            # rm the files used to direct output to.
            foreach $ofile (@ofile) {
                $Log->comment("removing file $ofile");
                $api->execute_raw_command( command => "rm $ofile" );
            }
        } ## end if ( !$found_errors )

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } elsif ($found_errors) {
            my $msg
                = "sio reported failures against one or more volumes. "
                . "Found error string(s) in sio output on $client_name. Fail info follows:"
                . $failing_msg;
            if ($::IGNORE_ERRORS) {
                $my_return->{result} = 1;
                $Log->warn($msg);
            } else {
                $my_return->{result} = 0;
                $Log->error($msg);
            }
            $my_return->{msg} = $msg;
        } else {    # Good IO without errors
            $my_return->{result} = 1;
            my $msg = "sio ran successfully without errors on $client_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( $num_clients > 1 )
}    # end sub sio

sub _wait_till_all_sio_complete {

    my ( $api, $active_sio_pids ) = @_;

    my $done = 0;
    my @completed_sio_pids;
    my $num_done = 0;

    # Wait loop to check if all sio threads are finished.
    while ( !$done ) {

        # Get list of currently active sio pids
        my @cur_active_pids = _get_proc_pids(
            api_set   => $api,
            proc_name => 'sio',
        );

        # Find out which sio processes completed
        foreach my $a_pid ( keys( %{$active_sio_pids} ) ) {
            if ( !grep { $_ eq $a_pid } @cur_active_pids ) {
                push( @completed_sio_pids, $a_pid );
            }
        }

        # Process each completed sio process
        sleep 5;    # Allow cool down for tool and time for
                    # output to fully redirect to log file
        foreach my $c_pid (@completed_sio_pids) {
            delete $active_sio_pids->{$c_pid};
            $num_done++;
        }

        if ( keys( %{$active_sio_pids} ) == 0 )
        {           #no more entries left in active_sio_pids.
            $done = 1;
            last;
        }
        sleep 30;
    } ## end while ( !$done )
} ## end sub _wait_till_all_sio_complete

##CIFS HELPERS
###################################################################
# sub: _setup_cifs
# summary: 1) Checks if vserver is running,if not,starts it
#          2) Checks if cifs vserver exists, if not, creates it
#          3) Checks if cifs vserver is running, starts it, if not
#
# input : hash of volume objects (volumes => \@volumes)
#         and hash reference to cifs config params
#         (params => \$cifs);
# output: n/a
###################################################################
sub _setup_cifs {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    my @vols     = @{ $input{volumes} };
    my $domain   = $input{params}->{domain};
    my $username = $input{params}->{user};
    my $password = $input{params}->{password};
    my %setup;
    my $e;

    foreach my $v (@vols) {
        my $n      = $v->command_interface();
        my $n_name = $n->name();
        my $c = NACL::CS::ClusterIdentity->fetch( command_interface => $n, );
        my $c_name = $c->name();

        #Setup Vserver and CIFS only once per cluster
        if ( !$setup{$c_name} ) {

            #Check if Vserver is running, if not, start vserver
            my $vs     = $v->vserver();
            my $vs_obj = NACL::C::Vserver->find(
                command_interface => $n,
                filter            => { vserver => $vs }
            );
            if ( $vs_obj->state()->admin_state() !~ /^running/i ) {
                $Log->comment("Vserver($vs) is not running: Will start it.");
                $vs_obj->start();
            }
            else {
                $Log->comment("Vserver($vs) is running.");
            }

            #Check if cifs vserver exists, if not, create it
            my $cifs_vs;
            try {
                $cifs_vs = NACL::C::VserverCifs->find(
                    command_interface => $n,
                    filter            => { vserver => $vs }
                );
                $Log->comment("CIFS Vserver is found!");
            }
            otherwise {
                $e = shift;
                if ( $e =~ /No matching vserver cifs\(s\) found/i ) {
                    $Log->comment("CIFS vserver does not exist!");
                    $cifs_vs = NACL::C::VserverCifs->create(
                        command_interface     => $n,
                        vserver               => $vs,
                        domain                => $domain,
                        username              => $username,
                        password              => $password,
                        'cifs-server'         => $c_name . '_vs_cifs',
                        reuse_machine_account => "y",
                    );
                }
                else {
                    $e->throw();
                }
            };

            #Check if cifs vserver is running, if not, start it
            if ( $cifs_vs->state->status_admin() !~ /^up/i ) {
                $Log->comment("CIFS Vserver is not running: Will start it.");
                $cifs_vs->start();
            }
            else {
                $Log->comment("CIFS Vserver is running.");
            }
            $setup{$c_name} = 1;
        }    #end if(!$setup{$c_name})
    }    #end foreach @vol
}    #end _setup_cifs

#############################################################
# sub: _map_cifs_tools
# summary: 1) Gets SITE info from natehost params
#          2) maps X: to appropriate server location
#
# input : hash of client objects (clients => \@clients)
# output: n/a
#############################################################
sub _map_cifs_tools {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    my @clients  = @{ $input{clients} };
    my %mapped;

    my %site_map = (
        SVL => {
            server => 'svlinf06.eng.netapp.com',
            domain => 'svlinf06',
        },
        RTP => {
            server => 'rtpinf02.rtp.netapp.com',
            domain => 'rtpinf02',
        },
        PTC => {
            server => 'ptcinf02.spin.eng.netapp.com',
            domain => 'ptcinf02',
        },
        NANE => {
            server => 'naninf02.nane.netapp.com',
            domain => 'naninf02',
        },
        BTC => {
            server => 'btcinf02.eng.btc.netapp.in',
            domain => 'btcinf02',
        }
    );

    my $site = uc $ENV{SITE};
    $site =~ s/\s+//;
    if ( $site eq '' ) {
        $site = 'SVL';
    }
    elsif ( $site eq 'NB' ) {
        $site = 'BTC';
    }

    foreach my $c (@clients) {
        my $c_name = $c->name();
        next if $mapped{$c_name};    #Only map X: once

        my $found = 0;
        my $api   = $c->apiset();
        my $xdrive;
        try {
            $xdrive = $api->net_use( sharename => 'X:' )->get_raw_output();
            if ( $xdrive =~ /localtest\\arch\\win32/i ) {
                $found = 1;
                $Log->comment("X: is mapped correctly for $c_name");
            }
            else {
                $Log->comment("X: is mapped incorrectly to $xdrive");
                NACL::STask::Client::Mount->purge(
                    command_interface => $c,
                    mount_point       => 'X:',
                );
            }
        }
        otherwise {
            $Log->comment("X: is not mapped!");
        };

        if ( !$found ) {
            $Log->comment("Need to map X:");
            NACL::STask::Client::Mount->create(
                command_interface => $c,
                user              => $site_map{$site}{domain} . '\minwin',
                password          => 'qa4testing',
                mount_point       => 'X:',
                server            => $site_map{$site}{server},
                share             => 'localtest\arch\win32',
            );
        }
        $mapped{$c_name} = 1;
    }
} #end _map_cifs_tools

#############################################################
# sub: _pretest_mount_cleanup
# summary: 1) Purges all other mounts on client except X:
#
# input : hash of client objects (clients => \@clients)
# output: n/a
#############################################################
sub _pretest_mount_cleanup {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;
    my %cleaned_up;

    my @clients = @{ $input{clients} };
    foreach my $c (@clients) {
        my $c_name = $c->name();
        next if $cleaned_up{$c_name};

        my @mounts =
          NACL::STask::Client::Mount->find( command_interface => $c, );
        foreach my $m (@mounts) {
            my $mp = NACL::MTask::Traffic->get_abs_path($m);
            if ( $mp !~ /^X\:$/i ) {
                $Log->comment("Purging $mp on $c_name");
                $m->purge();
            }
        }
        $cleaned_up{$c_name} = 1;
    }
} #end _pretest_mount_cleanup

#############################################################
# sub: _mount_cifs_volumes
# summary: 1) modifies volume unix-permissions to 777
#          2) Creates cifs share on filer
#          3) mounts volumes on cifs client
#
# input : hash of volume objects (volumes => \@volumes)
#         and hash reference to cifs config params
#         (params => \$cifs);
# output: array of hashes of mount paths and mount objects
#############################################################
sub _mount_cifs_volumes {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    my $client = $input{client};
    my $c_name = $client->name();
    my @vols   = @{ $input{volumes} };
    my $domain   = $input{params}->{domain};
    my $username = $input{params}->{user};
    my $password = $input{params}->{password};
    my $my_return;

    my %share_info = (
        user => $domain. '\\' .$username, password => $password);

    foreach my $v (@vols) {
        my $vol_name = $v->volume();
        my $node     = $v->command_interface();
        my $vs       = $v->vserver();
        NACL::STask::Volume->modify(
            command_interface  => $node,
            vserver            => $vs,
            volume             => $vol_name,
            'unix-permissions' => '777',
        );

        my $share_obj = NACL::STask::VserverCifsShare->create(
            volume             => $v,
            'share-name'       => $vol_name,
            nacltask_if_exists => 'reuse',
        );
        $share_info{share} = $share_obj;

        my $mount = NACL::STask::Client::Mount->create(
            command_interface => $client,
            %share_info,
        );

        my $path = NACL::MTask::Traffic->get_abs_path($mount);

        $my_return->{paths}->{$vol_name} = $path;
        push( @{ $my_return->{mounts} }, $mount );
    }
    return $my_return;
} #end _mount_cifs_volumes

###########################################################
# sub: _get_cifs_args
# summary: 1) cycles through ntest,nodes and clients to
#             gather CIFS_DOMAIN_LONG, CIFS_USER, CIFS_PASSWD
#             AD_SERVER_USER and AD_SERVER_PASSWORD
#
# input:   hash array of clients (clients => \@clients)
# output:  hash of CIFS_DOMAIN_LONG, AD_SERVER_USER|CIFS_USER,
#              CIFS_PASSWD|AD_SERVER_PASSWORD
###########################################################
sub _get_cifs_args {
    my %opts = @_;
    my @a = qw ( CIFS_USER CIFS_PASSWD AD_SERVER_USER AD_SERVER_PASSWORD CIFS_DOMAIN_LONG );
    my %list = map { $_ => 1 } @a;
    my @clients = @{$opts{clients}};
    my @nodes   = NACL::STask::Node->find();
    my (%temp, %config);

    foreach my $k (keys %list) {
        if($ENV{$k}) {
            $Log->comment("Looking at $ENV{$k}");
            $temp{$k} = $ENV{$k};
            delete $list{$k};
        }
    }

    foreach my $c (@clients) {
        foreach my $k (keys %list) {
            $Log->comment("Client at $c->hostrec->{$k}");
            if($c->hostrec->{$k}) {
                $temp{$k} = $c->hostrec->{$k};
                delete $list{$k};
            }
        }
    }

    foreach my $n (@nodes) {
        foreach my $k (keys %list) {
            if($n->hostrec->{$k}) {
                $temp{$k} = $n->hostrec->{$k};
                delete $list{$k};
            }
        }
    }

    if(%list) {
        my $err = 'CIFS parameters missing in natehosts or .ntest:';
        my $missing_param = 0;
        if($list{AD_SERVER_USER} && $list{CIFS_USER}) {
            $err .= ' CIFS_USER';
            $missing_param = 1;
        }
        if($list{AD_SERVER_PASSWORD} && $list{CIFS_PASSWD}) {
            $err .= ' CIFS_PASSWD';
            $missing_param = 1;
        }
        if($list{CIFS_DOMAIN_LONG}) {
            $err .= ' CIFS_DOMAIN_LONG';
            $missing_param = 1;
        }
        if($missing_param) {
            NATE::BaseException->throw($err);
        }
    }

    $config{user} = $temp{CIFS_USER} ? $temp{CIFS_USER} : $temp{AD_SERVER_USER};
    $config{password} = $temp{CIFS_PASSWD} ? $temp{CIFS_PASSWD} : $temp{AD_SERVER_PASSWORD};
    $config{domain} = $temp{CIFS_DOMAIN_LONG};

    return \%config;
} #end _get_cifs_args
###########################################################

1;
