# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/NX/Node.pm#5 $
#
# Copyright (c) 2013 NetApp
# All rights reserved.
#
## @summary
## Node specified functions to support NACL Translator framework
##
## @status    private
##

package NX::Node;

use strict 'vars';
use Tharn;

use Data::Dumper;
use NX::Helper;
use NATE::PerlRPC::Worker::Process;
use NACL::Transit;
use Params::Validate qw(validate SCALAR SCALARREF HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
    options_set     bootargs_set        flags_set   systemshell_cmds
);

my @Pkg_Workers;
sub stop_workers {
    $Log->comment("Detected interrupt signal");
    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}

=pod

=head1 NAME

NX::Aggr

=head1 DESCRIPTION

Node related routines for use in NX framework;

=head1 METHODS

=head2 options_set

Modifies specified option names to specified values.

The input is a hash (%opts) which must include the command_interface:
    command_interface => <array ref of nodes>
    key_value         => <hash ref to option names and their values>

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

sub options_set {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; #Hash reference containing return values
    my $exception;
    my $settings;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF,  optional => 0},
                key_value         => {type => HASHREF,   optional => 0},
                validate_only     => {type => SCALAR,    optional => 1, default => 0},
            }
        );

        foreach my $k (keys(%input)) {
            if ($k eq 'key_value') {
                foreach my $opt (keys(%{$input{$k}})) {
                    if ($opt =~ /\s+/) {
                        NATE::BaseException->throw("The option name \'$opt\' "
                            . "seems invalid as it contains spaces.");
                    } elsif ($input{$k}->{$opt} =~ /\s+/) {
                        NATE::BaseException->throw("The value \'"
                            . $input{$k}->{$opt} . "\' for option name "
                            . "\'$opt\' seems invalid as it contains spaces.");
                    }
                }
                $settings = $input{$k};
            }
        }
                
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "options_set action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Node::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            # create aggregate
            $input{command_interface} = [ $n ];

            my $id = 'Option_Set_N_DaC.'.$proc_cnt;
            push(@ids, $id);

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Node',
                method  => 'options_set',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        #Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Set options on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { #Single node
        delete $input{validate_only};
        my $node = $nodes[0];
        my $api = $node->get_7m_or_nodescope_apiset();

        foreach my $opt (keys(%{$settings})) {
            try {
                # Verify if it was already set
                my $po = $api->options("$opt" => undef)->get_parsed_output();
                my $cur_val = $po->[0]->{$opt};
                if ($cur_val eq $settings->{$opt}) {
                    $Log->comment("Option $opt is already set to "
                        . "\'$settings->{$opt}\' on ". $node->node());
                } else {
                    # Set option
                    $api->options("$opt" => $settings->{$opt});

                    # Verify it was indeed set
                    my $po = $api->options("$opt" => undef)->get_parsed_output();
                    my $cur_val = $po->[0]->{$opt};
                    if ($cur_val eq $settings->{$opt}) {
                        $Log->comment("Successfully set option $opt to "
                            . "\'$settings->{$opt}\' on ". $node->node());
                    } else {
                        NATE::BaseException->throw("Failed to set option "
                            . "$opt with a value of \'$settings->{$opt}\' "
                            . " on ". $node->node());
                    }
                }
            } otherwise {
                my $e = shift;
                if ($::IGNORE_ERRORS) {
                    $Log->warn("Got exception while trying to set option "
                        . "$opt to \'$settings->{$opt}\' on ". $node->node()
                        . ": " . $e->text());
                } else {
                    $exception = shift;
                }
            };
            last if $exception;
        }

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Set options on ".$node->node();
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} ## End options_set

=pod

=head2 bootargs_set

Modifies specified bootarg names to specified values.

The input is a hash (%opts) which must include the command_interface:
    command_interface => <array ref of nodes>
    key_value         => <hash ref to bootarg names and their values>

=head3 opts
    
%opts can also be used to override default values, which include
    force_setenv      => If set to 1, then function will boot to the bootloader
                         and use 'setenv' to set the boot arg values.
                         If set to 0, then function will use the 'bootargs'
                         CLI command to set the value, if it is not already
                         set to the specified value.  Default is 0.

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut
sub bootargs_set {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; #Hash reference containing return values
    my $exception;
    my $settings;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF,  optional => 0},
                key_value         => {type => HASHREF,   optional => 0},
                force_setenv      => {type => SCALAR,    optional => 1, default => 0},
                validate_only     => {type => SCALAR,    optional => 1, default => 0},
            }
        );

        foreach my $k (keys(%input)) {
            if ($k eq 'key_value') {
                foreach my $opt (keys(%{$input{$k}})) {
                    if ($opt =~ /\s+/) {
                        NATE::BaseException->throw("The bootarg name \'$opt\' "
                            . "seems invalid as it contains spaces.");
                    } elsif ($input{$k}->{$opt} =~ /\s+/) {
                        NATE::BaseException->throw("The value \'"
                            . $input{$k}->{$opt} . "\' for bootarg name "
                            . "\'$opt\' seems invalid as it contains spaces.");
                    }
                }
                $settings = $input{$k};
            } elsif ($k eq 'force_setenv') {
                if ($input{$k} !~ /^0|1$/) {
                    NATE::BaseException->throw("The \'$k\' parameter must be "
                        . "either 0 or 1.");
                }
            }
        }
                
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "bootargs_set action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Node::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            $input{command_interface} = [ $n ];

            my $id = 'Bootargs_Set_N_DaC.'.$proc_cnt;
            push(@ids, $id);

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Node',
                method  => 'bootargs_set',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        #Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Set bootargs on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { #Single node
        my $node = $nodes[0];
        my $api = $node->get_7m_or_nodescope_apiset();
        my $transit_obj;
        my $console_api;

        if ($input{force_setenv}) {
            $transit_obj = NACL::Transit->new(name => $node->hostrec());
            $transit_obj->change_state(to => 'FIRMWARE', inhibit_takeover => 1);
        
            # Need a console connection to send setenv commands
            my $host_obj = Tharn::host($node->node());
        
            ## Create a connectrec object here to the console
            my $conn = $host_obj->connect(
                template => 'console',
                id       => 'console',
                login    => 1
            );
        
            ## Create an APISet object here for console connection
            $console_api = NACL::APISet->new(
                connection => $conn,
                category   => 'Node',
                interface  => 'CLI',
                set        => 'Firmware',
            );
        }

        foreach my $ba (keys(%{$settings})) {
            try {
                if ($input{force_setenv}) {                
                    $console_api->execute_raw_command(
                        command => "setenv $ba $settings->{$ba}"
                    );

                    my $raw_output = $console_api->execute_raw_command(
                        command => "printenv $ba"
                    );

                    my @raw = split(/\n/, $raw_output);
                    my $value_line = $raw[-1];
                    $Log->comment("Value line from printenv: $value_line");
                    $value_line =~ s/\s+$//;
                    if (($value_line eq $settings->{$ba}) || ($value_line =~ /\s+$settings->{$ba}$/)) {
                        $Log->comment("Successfully bootarg \'$ba\' to "
                            . "\'$settings->{$ba}\'");
                    } else {
                        NATE::BaseException->throw("Failed to set bootarg "
                            . "\'$ba\' to \'$settings->{$ba}\'");
                    }
                } else {
                    # Verify if it was already set
                    my $po = $api->bootargs_print(arg => $ba)->get_parsed_output();
                    my $cur_val = $po->[0]->{$ba};
                    if (($cur_val eq $settings->{$ba}) && !$input{force_setenv}) {
                        $Log->comment("Bootarg \'$ba\' is already set to "
                            . "\'$settings->{$ba}\' on ". $node->node());
                        next;
                    } else {
                        # Set bootarg
                        $api->bootargs_set(arg => $ba, value => $settings->{$ba});
                    }
                }
            } otherwise {
                $exception = shift;
            };
            last if $exception;
        }

        if (!$exception) {
            if ($input{force_setenv}) {
                $transit_obj->reboot();
            }
    
            try {
                if ($input{force_setenv}) {
                    $node->refresh_command_interface( no_version_change => 1 );
                }
                foreach my $ba (keys(%{$settings})) {
                    # Verify it was indeed set
                    my $po = $api->bootargs_print(
                        arg => $ba
                    )->get_parsed_output();

                    my $cur_val = $po->[0]->{$ba};
                    if ($cur_val eq $settings->{$ba}) {
                        $Log->comment("Successfully set bootarg $ba to "
                            . "\'$settings->{$ba}\' on ". $node->node());
                    } else {
                        NATE::BaseException->throw("Failed to set bootarg "
                            . "$ba with a value of \'$settings->{$ba}\' "
                            . " on ". $node->node());
                    }
                }
            } otherwise {
                $exception = shift;
            }
        }

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Set bootargs on ".$node->node();
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} ## End bootargs_set


=pod

=head2 flags_set

Modifies specified flag names to specified values.

The input is a hash (%opts) which must include the command_interface:
    command_interface => <array ref of nodes>
    key_value         => <hash ref to flag names and their values>

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut
sub flags_set {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; #Hash reference containing return values
    my $exception;
    my $settings;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF,  optional => 0},
                key_value         => {type => HASHREF,   optional => 0},
                validate_only     => {type => SCALAR,    optional => 1, default => 0},
            }
        );

        foreach my $k (keys(%input)) {
            if ($k eq 'key_value') {
                foreach my $opt (keys(%{$input{$k}})) {
                    if ($opt =~ /\s+/) {
                        NATE::BaseException->throw("The flag name \'$opt\' "
                            . "seems invalid as it contains spaces.");
                    } elsif ($input{$k}->{$opt} =~ /\s+/) {
                        NATE::BaseException->throw("The value \'"
                            . $input{$k}->{$opt} . "\' for flag name "
                            . "\'$opt\' seems invalid as it contains spaces.");
                    }
                }
                $settings = $input{$k};
            }
        }
                
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "flags_set action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Node::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            # create aggregate
            $input{command_interface} = [ $n ];

            my $id = 'Flags_Set_N_DaC.'.$proc_cnt;
            push(@ids, $id);

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Node',
                method  => 'flags_set',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        #Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Set flags on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { #Single node
        delete $input{validate_only};
        my $node = $nodes[0];
        my $api = $node->get_7m_or_nodescope_apiset();

        foreach my $flag (keys(%{$settings})) {
            try {
                # Verify if it was already set
                my $po = $api->printflag(
                    flags               => $flag,
                    "privilege-level"   => "test"
                )->get_parsed_output();

                my $cur_val = $po->[0]->{$flag};
                if ($cur_val eq $settings->{$flag}) {
                    $Log->comment("Flag $flag is already set to "
                        . "\'$settings->{$flag}\' on ". $node->node());
                } else {
                    # Set flag
                    $api->setflag(
                        flag                => $flag,
                        value               => $settings->{$flag},
                        "privilege-level"   => "test",
                    );

                    # Verify it was indeed set
                    my $po = $api->printflag(
                        flags               => $flag,
                        "privilege-level"   => "test"
                    )->get_parsed_output();

                    my $cur_val = $po->[0]->{$flag};
                    if ($cur_val eq $settings->{$flag}) {
                        $Log->comment("Successfully set flag $flag to "
                            . "\'$settings->{$flag}\' on ". $node->node());
                    } else {
                        NATE::BaseException->throw("Failed to set flag "
                            . "$flag with a value of \'$settings->{$flag}\' "
                            . " on ". $node->node());
                    }
                }
            } otherwise {
                my $e = shift;
                if ($::IGNORE_ERRORS) {
                    $Log->warn("Got exception while trying to set flag "
                        . "$flag to \'$settings->{$flag}\' on ". $node->node()
                        . ": " . $e->text());
                } else {
                    $exception = shift;
                }
            };
            last if $exception;
        }

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Set flags on ".$node->node();
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} ## End flags_set


=pod

=head2 systemshell_cmds

Executes specified systemshell commands

The input is a hash (%opts) which must include the command_interface:
    command_interface => <array ref of nodes>
    list_element      => <array ref to systemshell commands to execute>

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut
sub systemshell_cmds {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; #Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface => {type => ARRAYREF,  optional => 0},
                list_element      => {type => ARRAYREF,  optional => 0},
                validate_only     => {type => SCALAR,    optional => 1, default => 0},
            }
        );
                
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "systemshell_cmds action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Node::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            $input{command_interface} = [ $n ];

            my $id = 'Systemshell_Cmds_N_DaC.'.$proc_cnt;
            push(@ids, $id);

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Node',
                method  => 'systemshell_cmds',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        #Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Execute systemshell commands on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else { #Single node
        my @cmds = @{$input{list_element}};
        my $node = $nodes[0];
        my $system_shell = $node->apiset(set => "Systemshell");

        foreach my $c (@cmds) {
            $system_shell->execute_command( command => $c );
        }

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Executed systemshell commands on ".$node->node();
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} ## End systemshell_cmds


1;
