# this is basically -*- perl -*-
# $Id: $
#
# Copyright (c) 2012 NetApp
# All rights reserved.
#
## @summary
## Snapmirror specified functions to support NACL Translator framework
##
## @status    private
##
## @author Paul Chen (chenp@netapp.com)

package NX::Snapmirror;

use strict 'vars';
use Tharn;
use Data::Dumper;

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::STask::SystemLicense;
use NACL::MTask::SnapmirrorSet;
use NACL::C::VolumeSnapshotPolicy;
use NACL::C::SnapmirrorPolicy;
use NACL::C::Vserver;
use NX::Helper;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);
use FS::commonParams;

our @ISA    = ('Exporter');
our @EXPORT = qw( snapmirror_life_cycle );

my $params = NATE::ParamSet->new(global => 1);
my $IGNORE_ERRORS = $params->get("IGNORE_ERRORS", default => '0');

my @Pkg_Workers;

sub stop_workers {
    $Log->comment("Detected interrupt signal");

    # Stop any workers from divide/conquer logic
    _nx_stop_workers(workers => \@Pkg_Workers);
    exit;
} ## end sub stop_workers

my $Host_Info;

=pod

=head1 NAME

NX::Snapmirror

=head1 DESCRIPTION

HA related routines for use in NX framework;
This library performs takeover and auto-giveback. Please use caution when
assigning test intervals.
For SN.0, auto-giveback is limited to 2 attempts for every 60 minutes.

=head1 METHODS
=head2 snapmirror_life_cycle 

Perform snapmirror fanout and life cycle on specified nodes.

The input is a hash (%opts) which must include the command_interface:
    src_nodes      => <array ref of source nodes>
    dst_nodes      => <array ref of destination nodes>
    volumes        => <array ref of all volume objects for test>
=head3 opts

%opts can also be used to override default values, which include
    hours          => One or more comma separated hour values for the
                      schedule (e.g. "1,2,3"), or the string "all", which
                      equates to every hour. Default is 'all'.
    minutes        => One or more comma separated minute values for the schedule
                      (e.g. "0,30"), or the string "all", which equates to
                      makes the function look strictly at the "hours" argument.
                      Default is "0,30".
    iterations     => <integer>|'infinite' # Number of cycles for failover.
    interval       => odd/even/all/<value in seconds>
                       # Every odd, even, or every clock hour,
                       # or specified value in seconds
    poll_interval  => <integer>   # Number of seconds to wait before checking
                         snapmirror transfer. Default is '300'.
    max_wait       => <integer>   # Number of seconds of max wait for checking
                         snapmirror transfer. Default is '7200'.
    start_minute   => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
                       a two-digit value representing the clock minute
    type           => 'LS|DP|XDP' # Type of snapmirror. Default is 'XDP'.
    num_fanout     => <integer>   # Number of fanout of snapmirror. Default is '2'.
    num_sm_pairs   => <integer>   # Number of snapmirror pairs. Default is '1'.
    do_cleanup     => 0|1         # Set to 1 to release and break snapmirror
                                  # relationships when function is done.
                                  # Default is 0.
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX snapmirror_life_cycle
Snapmirror_Life_Cycle
description
Set up snapmirror between a source and destination
required
src_volumes;Source volumes;;Syntax is $N<#>/$A<#>/$V<#> e.g. $N1/$A1/$V1-$V2;$N2/$A1/$V1-$V4 or $ALL
dst_volumes;Destination volumes;;Syntax is $N<#>/$A<#>/$V<#> e.g. $N1/$A1/$V1-$V2;$N2/$A1/$V1-$V4 or $ALL
optional
iterations;Test Iterations;1;number of cycles for snapmirror life cycle test. Syntax is <integer> or 'infinite'
interval;Test Interval;3600;interval between failovers, can be every odd, even, or every clock hour, or specified value in seconds. Syntax is 'odd' or 'even' or 'all' or <value in seconds>
start_minute;Start Minute;;Required if 'interval' is odd|eve|all, otherwise ignored if 'interval' is an interger value. This is a two-digit value representing the clock minute. Syntax is <\d\d> e.g 10 or 20.
poll_interval;Polling Interval;300;number of seconds between checking snapmirror transfers.
max_wait;Max Wait;300;Timeout for checking snapmirror transfer.
type;SM Type;XDP;Snapmirror type,string value, supports LS|DP|XDP, default is XDP.
num_fanout;Number of fanouts;2;number of snapmirror fanout, default to 2.
num_sm_pairs;Number of snapmirror pairs;1;number of snapmirror pairs, default to 1.
=end NX snapmirror_life_cycle
=cut

sub snapmirror_life_cycle {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if (@args % 2);
    my %input = @args;
    return 0 if $input{nx_context_inquire};

    my $my_return;    #Hash reference containing return values
    my $exception;
    my (@src_nodes,     @dst_nodes, $num_dst_nodes,
        $num_src_nodes, $node_ci,   @nodes,         $base_node,
    );

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validate_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                dst_nodes  => {type => ARRAYREF, optional => 1},
                src_nodes  => {type => ARRAYREF, optional => 1},
                volumes    => {type => ARRAYREF, optional => 1},
                src_volumes => {type => ARRAYREF, optional => 1},
                dst_volumes => {type => ARRAYREF, optional => 1},
                hours      => {type => SCALAR,   optional => 1, default => 'all'},
                minutes    => {type => SCALAR,   optional => 1, default => '0,30'},
                iterations => {type => SCALAR, optional => 1, default => 1},
                interval   =>
                    {type => SCALAR, optional => 1, default => 3600},
                poll_interval =>
                    {type => SCALAR, optional => 1, default => 300},
                max_wait   => {type => SCALAR, optional => 1, default => 3600},
                start_minute => {type => SCALAR, optional => 1},
                type => {type => SCALAR, optional => 1, default => 'XDP'},
                num_fanout => {type => SCALAR, optional => 1, default => 2},
                num_sm_pairs =>
                    {type => SCALAR, optional => 1, default => 1},
                do_cleanup =>
                    {type => SCALAR, optional => 1, default => 0},
                _extracted_sm =>
                    {type => SCALAR, optional => 1, default => 0},
                validate_only =>
                    {type => SCALAR, optional => 1, default => 0},
            }
        );

        if ((!$input{src_volumes} && !$input{dst_volumes})
            && (!$input{dst_nodes} || !$input{src_nodes} || !$input{volumes}) ) {
            NATE::BaseException->throw(
                "dst_nodes, src_nodes, and volumes parameters are required"
            );
        } elsif ((!$input{dst_nodes} && !$input{src_nodes} && !$input{volumes})
            && (!$input{src_volumes} || !$input{dst_volumes}) ) {
                NATE::BaseException->throw("src_volumes and dst_volumes "
                    . "parameters are required"
            );
        }

        for my $key (keys %input) {
            $Log->comment("key: |$key|\nvalue: | $input{$key}|\n");
            if (($key eq 'iterations') && !defined($input{'interval'})) {
                NATE::BaseException->throw(
                          "\'iterations\' parameter requires a value "
                        . "for the \'interval\' parameter");
            } elsif ($key eq 'interval') {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            } elsif ($key eq 'type'
                && $input{$key} ne 'LS'
                && $input{$key} ne 'DP'
                && $input{$key} ne 'XDP') {
                NATE::BaseException->throw(
                    "Invalid input ($input{$key}) for snapmirror option <$key>. Accepted types: LS, DP, and XDP"
                );
            } elsif ($key eq 'iterations' && $input{$key} !~ /^(\d+)$/) {
                NATE::BaseException->throw(
                    "Invalid input ($input{$key}) for snapmirror option <$key>. Accept integers only"
                );
            } elsif ($key eq 'num_fanout' && $input{$key} !~ /^(\d+)$/) {
                NATE::BaseException->throw(
                    "Invalid input ($input{$key}) for snapmirror option <$key>. Accept integers only"
                );
            } elsif ($key eq 'num_sm_pairs' && $input{$key} !~ /^(\d+)$/) {
                NATE::BaseException->throw(
                    "Invalid input ($input{$key}) for snapmirror option <$key>. Accept integers only"
                );
            } elsif ($key eq 'poll_interval' && $input{$key} !~ /^(\d+)$/) {
                NATE::BaseException->throw(
                    "Invalid input ($input{$key}) for snapmirror option <$key>. Accept integers only"
                );
            } elsif ($key eq 'max_wait' && $input{$key} !~ /^(\d+)$/) {
                NATE::BaseException->throw(
                    "Invalid input ($input{$key}) for snapmirror option <$key>. Accept integers only"
                );
            } elsif ( $key =~ /^(hours|minutes)$/ ) {
                my @tmp = split(/,/, $input{$key});
                foreach my $v (@tmp) {
                    if (($v !~ /^\d+$/) && ($v ne 'all')) {
                        NATE::BaseException->throw("\'$key\' value must be "
                            . "comma delimited integers or \'all\'");
                    }
                }
            } else {
                next;
            }
        }
    }    # end try
    otherwise {
        $exception = shift;
    };    #end otherwise

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ($input{validate_only}) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "snapmirror_life_cycle passed validation";
        return $my_return;
    }

    # Store start tick for return later
    my $start_tick = time();
    $my_return->{start_tick} = $start_tick;

    my @src_vol_objs;
    my @dst_vol_objs;
    if ($input{volumes}) {
        # Keep this around for backwards compatibility
        push(@nodes, @{$input{src_nodes}});
        push(@nodes, @{$input{dst_nodes}});
        my $src_node = ${$input{src_nodes}}[0];
        my $dst_node = ${$input{dst_nodes}}[0];

        my @vol_objs = @{$input{volumes}};
        foreach (@vol_objs) {
            my $n = $_->state()->node();
            my $type = $_->state()->type();
            $Log->comment("Vol: ".$_->volume()."\n".
                "Node: $n\n".
                "Type: $type\n");
            if (($n eq $src_node->node()) && ($type eq 'RW')) {
                $Log->comment("Pushing ".$_->volume()." as a src volume");
                push(@src_vol_objs, $_);
            }
        }
    
        foreach (@vol_objs) {
            my $n = $_->state()->node();
            my $type = $_->state()->type();
            $Log->comment("Vol: ".$_->volume()."\n".
                "Node: $n\n".
                "Type: $type\n");
            if (($n eq $dst_node->node()) && ($type =~ /DP|LS/)) {
                $Log->comment("Pushing ".$_->volume()." as a dst volume");
                push(@dst_vol_objs, $_);
            }
        }
    } else {
        @src_vol_objs = @{$input{src_volumes}};
        foreach my $sv (@src_vol_objs) {
            my $n = $sv->state()->node();
            my $vol_name = $sv->volume();
            if (!$node_ci->{$n}) {
                $node_ci->{$n} = $sv->command_interface();
                push(@nodes, $sv->command_interface());
            }
            my $type = $sv->state()->type();
            $Log->comment("Src vol: $vol_name\n".
                "Node: $n\n".
                "Type: $type\n");
            if ($type ne 'RW') {
                NATE::BaseException->throw(
                    "Volume $vol_name is not RW, thus cannot be a src volume"
                );
            }
        }

        @dst_vol_objs = @{$input{dst_volumes}};
        foreach my $dv (@dst_vol_objs) {
            my $n = $dv->state()->node();
            my $vol_name = $dv->volume();
            if (!$node_ci->{$n}) {
                $node_ci->{$n} = $dv->command_interface();
                push(@nodes, $dv->command_interface());
            }
            my $type = $dv->state()->type();
            $Log->comment("Dst vol: $vol_name\n".
                "Node: $n\n".
                "Type: $type\n");
            if ($type !~ /DP|LS/) {
                NATE::BaseException->throw(
                    "Volume $vol_name is not DP|LS, thus cannot be a dst volume"
                );
            }
        }
    }

    my $num_src_vols = @src_vol_objs;
    my $num_dst_vols = @dst_vol_objs;
    my $num_nodes = @nodes;
    $base_node = $nodes[0];
    
    my ($sm_policy, $snap_policy, $schedule_name);
    my ($sm_policy_name, $snapshot_policy_name, $label_name)
        = ('nx_sm_policy', 'nx_sm_test', 'sm_policy_label');
    my $vserver_admin = NACL::C::Vserver->find(
        command_interface => $base_node,
        filter            => {type => 'admin'}
    );

    my (@sm_objs, @src_vols);
    my $iter = 1;

    try {
        $sm_policy_name       .= '_' . $input{type};
        $snapshot_policy_name .= '_' . $input{type};
        $label_name           .= '_' . $input{type};
        # Create unique names for label

        # Make sure we have enough volumes for test
        # $input{'num_sm_pairs'} of source volumes are required
        # $input{'num_sm_pairs'} * $input{'num_fanout'} of destination volumes are
        # required
        $Log->comment("Make sure we have enough volumes to start!");
        if ($num_src_vols < $input{'num_sm_pairs'}) {
            NATE::BaseException->throw("Requires "
                    . $input{'num_sm_pairs'}
                    . " of source volumes , only $num_src_vols available.");
        } elsif (
            $num_dst_vols < ($input{'num_sm_pairs'} * $input{'num_fanout'})) {
            NATE::BaseException->throw("Requires "
                    . $input{'num_sm_pairs'} * $input{'num_fanout'}
                    . " of source volumes , only $num_dst_vols available.");
        } else {
            $Log->comment("We have enough volumes to start!");
        }

        # Start snapmirror test

        # Add snapmirror license
        for my $node (@nodes) {
            NACL::STask::SystemLicense->add(
                command_interface    => $node,
                'package'            => ['snapmirror', 'snapvault'],
                nacltask_verify      => 1,
                nacltask_if_licensed => "pass",
            );
        }

        $schedule_name = create_cron_sched(
            command_interface   => $base_node,
            hours               => $input{hours},
            minutes             => $input{minutes},
        );

        $Log->comment("Create snapshot schedule for Snapmirror test");


        # Ensure snapshot policy exists, if not, create it on CMode
        $snap_policy = NACL::C::VolumeSnapshotPolicy->find(
            command_interface => $base_node,
            filter            => { policy => $snapshot_policy_name },
            allow_empty       => 1,
        );

        if ($snap_policy) {
            $Log->comment("Snapshot policy $snapshot_policy_name already exists");
        } else {
            $snap_policy = NACL::C::VolumeSnapshotPolicy->create(
                command_interface   => $base_node,
                policy              => $snapshot_policy_name,
                enabled             => 'true',
                schedule1           => $schedule_name,
                count1              => '10',
                'snapmirror-label1' => $label_name,
            );
        }

        if ($input{type} eq 'XDP') {
            $Log->comment("Create a new snapmirror policy for XDP");

            # Ensure snapmirror policy exists, if not, create it on CMode
            $sm_policy = NACL::C::SnapmirrorPolicy->find(
                command_interface => $base_node,
                filter            => { policy => $sm_policy_name },
                allow_empty       => 1,
            );

            if ($sm_policy) {
                $Log->comment("Snapmirror policy $sm_policy_name already exists");
            } else {
                $sm_policy = NACL::C::SnapmirrorPolicy->create(
                    command_interface => $base_node,
                    vserver           => $vserver_admin,
                    policy            => $sm_policy_name,
                );
            }

            my @labels = $sm_policy->state()->snapmirror_label();
            if (grep{$_ eq $label_name} @labels) {
                $Log->comment("SM label $label_name already "
                    . "part of policy $sm_policy_name");
            } else {
                $Log->comment("Add rule to snapmirror policy for XDP");
                $sm_policy->add_rule(
                    'snapmirror-label' => $label_name,
                    keep               => '10',
                );
            }
        }

        my $existed_prior = 0;
        $Log->comment(
            "Modify snapshot policy of source volumes and create snapmirror objects"
        );
        for (1 .. $input{num_sm_pairs}) {
            my $src_vol = shift(@src_vol_objs);
            push @src_vols, $src_vol;

            $Log->comment(
                "Modify volume snapshot policy to $snapshot_policy_name");
            $src_vol->modify('snapshot-policy' => $snapshot_policy_name);
            for (1 .. $input{num_fanout}) {
                my $dst_vol = shift(@dst_vol_objs);

                # Create snapmirror objects
                my $sm_obj;
                try {
                    $sm_obj = NACL::STask::Snapmirror->create(
                        command_interface   => $base_node,
                        'source-path'       => $src_vol,
                        'destination-path'  => $dst_vol,
                        type                => $input{type},
                        do_initialize       => 0,
                    );
                    $Log->comment("Created snapmirror relationship between "
                        . $src_vol->volume() . " and " . $dst_vol->volume());
                } otherwise {
                    my $e = shift;
                    if ($e =~ /already exist/i) {
                        $Log->comment("Snapmirror relationship already exists "
                            . "between ".$src_vol->volume()." and "
                            . $dst_vol->volume());
                        $sm_obj = NACL::STask::Snapmirror->create(
                            command_interface   => $base_node,
                            'source-path'       => $src_vol,
                            'destination-path'  => $dst_vol,
                            type                => $input{type},
                            if_exists           => 'reuse',
                            do_initialize       => 0,
                        );
                        $existed_prior = 1;
                    } else {
                        $e->throw();
                    }
                };
                push @sm_objs, $sm_obj;
            }
        }

        $Log->comment("Modify schedule and policy to all snapmirror objects");
        for my $sm_obj (@sm_objs) {
            if ($input{type} eq 'XDP') {
                $sm_obj->modify(
                    'schedule' => $schedule_name,
                    'policy'   => $sm_policy_name,
                );
            } else {
                $sm_obj->modify('schedule' => $schedule_name,);
            }
        }

        if (!$existed_prior) {
            $Log->comment("Initialize all snapmirror objects");
            my %unique;
            for my $sm_obj (@sm_objs) {
                if ($input{type} eq 'LS') {
                    # Filter out any non-unique source-paths
                    # initialize_ls_set() will initialize all destinations for a
                    # particular source
                    if (!exists $unique{$sm_obj->source_path()}) {
                        NACL::STask::Snapmirror->initialize_ls_set(
                            command_interface => $base_node,
                            'source-path'     => $sm_obj->source_path(),
                        );
                        $unique{$sm_obj->source_path()} = 1;
                    }
                } else {
                    $sm_obj->initialize();
                }
            }
        }

        while (1) {
            try {
                my $sm_obj_info;
                $Log->comment(
                    "Start snapmirror life cycle test on iteration $iter");

                $Log->comment("Verify snapmirror transfer happens between "
                        . $input{poll_interval}
                        . " seconds.");
                _snapmirror_transfer_check(
                    sm_objs       => \@sm_objs,
                    type          => $input{type},
                    max_wait      => $input{max_wait},
                    poll_interval => $input{poll_interval},
                );

                # Skip break and resync since don't work on LS snapmirror
                if ($input{type} ne 'LS') {
                    $Log->comment("Break all snapmirrors");
                    for my $sm_obj (@sm_objs) {
                        $Log->comment(
                            "Wait for state and status to be Snapmirrored and"
                                . " Idle before snapmirror break");
                        $sm_obj->monitor();
                    }
                    for my $sm_obj (@sm_objs) {
                        $sm_obj->break();
                    }

                    $Log->comment("Wait for 60 seconds after break");
                    sleep 60;

                    $Log->comment("Resync all snapmirrors");
                    for my $sm_obj (@sm_objs) {
                        $sm_obj->resync();
                    }

                    $Log->comment(
                        "Wait for state and status to be Snapmirrored and Idle after resync"
                    );
                    for my $sm_obj (@sm_objs) {
                        $sm_obj->monitor();
                    }
                }

                $Log->comment("Verify snapmirror transfer happens between "
                        . $input{poll_interval}
                        . " seconds.");
                _snapmirror_transfer_check(
                    sm_objs       => \@sm_objs,
                    type          => $input{type},
                    max_wait      => $input{max_wait},
                    poll_interval => $input{poll_interval},
                );

                $Log->comment("Perform snapmirror check on all snapmirrors");
                for my $sm_obj (@sm_objs) {
                    $sm_obj->check();
                }

                # Exit while loop and delete snapmirror and snapshot policies
                # if required iterations has been met
                if ($iter eq $input{iterations}) {
                    if ($input{do_cleanup}) {
                        $Log->comment(
                            "Modify snapshot policy on all source volume to default"
                        );
                        for my $src_vol (@src_vols) {
                            $src_vol->modify('snapshot-policy' => 'default');
                        }
    
                        $Log->comment("Purge all snapmirrors");
                        for my $sm_obj (@sm_objs) {
                            $sm_obj->purge();
                        }
    
                        $Log->comment("Delete snapmirror policy for XDP");
                        $sm_policy->delete() if $input{type} eq 'XDP';
    
                        $Log->comment(
                            "Delete snapshot policy for snapmirror test");
                        $snap_policy->delete();
                    }
                    last;
                }

                $Log->comment("Wait until next iteration starts");
                $iter++;
                wait_for_interval(
                    interval     => $input{interval},
                    start_minute => $input{start_minute},
                    start_tick   => $start_tick,
                );
            } ## end try
            otherwise {
                my $exception = shift;
                $Log->warn(
                    "Caught exception. Error was: " . $exception->text());
                if ($::IGNORE_ERRORS) {
                    if (   $input{'iterations'} ne 'infinite'
                        && $iter >= $input{'iterations'}) {

                        if ($input{do_cleanup}) {
                            $Log->comment(
                                "Modify snapshot policy on all source volume to default"
                            );
                            for my $src_vol (@src_vols) {
                                $src_vol->modify('snapshot-policy' => 'default');
                            }
    
                            $Log->comment("Purge all snapmirrors");
                            for my $sm_obj (@sm_objs) {
                                $sm_obj->purge();
                            }
    
                            $Log->comment("Delete snapmirror policy for XDP");
                            $sm_policy->delete() if $input{type} eq 'XDP';
    
                            $Log->comment("Delete snapshot policy");
                            $snap_policy->delete();
                        }
                    } else {
                        $iter++;
                        wait_for_interval(
                            interval     => $input{interval},
                            start_minute => $input{start_minute},
                            start_tick   => $start_tick,
                        );
                    }
                } else {

                    # For debugging purpose, current design will stop running
                    # snapmirror test and keep testbeds at the failure state
                    $exception->throw();
                }
            };
        } ## end while (1)

        $Log->comment("Snapmirror life cycle passed");

    } ## end try
    otherwise {
        my $e = shift;
        if (!$::IGNORE_ERRORS) {
            $exception = $e;
        } else {
            $Log->warn("Caught exception. Error was: " . $e->text());
        }
    };

    $my_return->{end_tick} = time();
    if ($exception) {
        my $msg = "Test failed with error!";
        $Log->comment($msg);
        $my_return->{msg}       = $msg;
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
    } else {
        my $msg
            = "Completed $iter iterations of "
            . $input{'type'}
            . " snapmirror_life_cycle test";
        $Log->comment($msg);
        $my_return->{msg}    = $msg;
        $my_return->{result} = 1;
    }
    return $my_return;
} ## end sub snapmirror_life_cycle

sub _snapmirror_transfer_check {

# This sub takes an input hash and verify snapnirror transfer happens within
# given time from $input{poll_interval}
# 1. $opts{sm_objs}       => A reference of an array of snapmirror objects.
# 2. $opts{type}          => Snapmirror type of the test.
# 3. $opts{max_wait}      => Maximum time allowed for transfer to complete.
# 4. $opts{poll_interval} => Polling interval of checking snapmirror transfers.

    my %opts = @_;

    my @sm_objs       = @{$opts{sm_objs}};
    my $sm_type       = $opts{type};
    my $max_wait      = $opts{max_wait};
    my $poll_interval = $opts{poll_interval};
    my $iter_max      = int($max_wait / $poll_interval);
    my $sm_obj_info;

    $Log->comment(
        "Save transfer timestamp from each destination for comparison later");
    for my $sm_obj (@sm_objs) {
        my $src_vol_name =
            $sm_obj->get_one_state_attribute('source-volume');
        my $dst_vol_name = $sm_obj->get_one_state_attribute('destination-volume');
        my $transfer_timestamp;
        if ($sm_type eq 'LS') {
            $transfer_timestamp = $sm_obj->get_one_state_attribute(
                'exported-snapshot');
            $Log->comment("Last snapshot exported between source "
                    . $src_vol_name
                    . " and destination "
                    . $dst_vol_name . " is "
                    . $transfer_timestamp);
        } else {
            $transfer_timestamp = $sm_obj->get_one_state_attribute(
                'last-transfer-end-timestamp');
            $Log->comment("Last transfer between source "
                    . $src_vol_name
                    . " and destination "
                    . $dst_vol_name
                    . " happened on "
                    . $transfer_timestamp);
        }
        $sm_obj_info->{$dst_vol_name}->{transfer_timestamp}
            = $transfer_timestamp;
    }

    $Log->comment(
        "Wait for $poll_interval seconds and check for transfer again");
    sleep $poll_interval;

    $Log->comment("Retrieve transfer timestamp for comparison");
    for my $sm_obj (@sm_objs) {
        my $iter = 0;
        while (1) {
            my $src_vol_name
                = $sm_obj->get_one_state_attribute('source-volume');
            my $dst_vol_name =
                $sm_obj->get_one_state_attribute('destination-volume');

            my $previous_transfer_timestamp
                = $sm_obj_info->{$dst_vol_name}->{transfer_timestamp};
            $Log->comment("Previously recorded transfer between source "
                    . $src_vol_name
                    . " and destination "
                    . $dst_vol_name
                    . " happened on "
                    . $previous_transfer_timestamp);

            my $latest_transfer_timestamp;
            if ($sm_type eq 'LS') {
                $latest_transfer_timestamp
                    = $sm_obj->get_one_state_attribute(
                    'exported-snapshot');
                $Log->comment("Last snapshot exported between source "
                        . $src_vol_name
                        . " and destination "
                        . $dst_vol_name . " is "
                        . $latest_transfer_timestamp);
            } else {
                $latest_transfer_timestamp
                    = $sm_obj->get_one_state_attribute(
                    'last-transfer-end-timestamp');
                $Log->comment("Latest transfer between source "
                        . $src_vol_name
                        . " and destination "
                        . $dst_vol_name
                        . " happened on "
                        . $latest_transfer_timestamp);
            }
            if ($previous_transfer_timestamp eq
                $latest_transfer_timestamp) {
                if ($iter >= $iter_max) {
                    my $err_msg
                        = "No transfer happened during past $max_wait seconds!";
                    $Log->error("$err_msg");
                    NATE::BaseException->throw("$err_msg");
                } else {
                    my $msg
                        = "No transfer detected during past $poll_interval"
                        . " seconds, check again in $poll_interval seconds";
                    $Log->comment("$msg");
                    $iter++;
                    sleep $poll_interval;
                }
            } else {
                $Log->comment("Snapmirror transfer detected!");
                last;
            }
        }
    }

    $Log->comment(
        "Wait for state and status to be Snapmirrored and Idle on all snapmirrors"
    );
    for my $sm_obj (@sm_objs) {
        $sm_obj->monitor();
    }

} ## end sub _snapmirror_transfer_check

sub _host_file_operation {

# This sub takes an input hash and backs up, add entry, or restore host file
#   1. $opts{'opt'}       : options of host file operation
#                           a. bak_add : back up and add new entry to host file.
#                           b. restore : restore host file.
#   2. $opts{'node_obj'}  : node object to run host file operation
#   3. $opts{'host_node'} : host name of the node object
#   4. $opts{'host_add'}  : host name to be added to node host file
#   5. $opts{'host_ip'}   : ip address to be added to node host file

    my %opts = @_;

    my $opt       = $opts{'opt'};
    my $node_obj  = $opts{'node_obj'};
    my $host_node = $opts{'host_node'};
    my $host_add  = $opts{'host_add'};
    my $host_ip   = $opts{'host_ip'};
    my $err_msg;

    if ($opt eq 'bak_add') {
        $Log->comment("Back up host file of $host_node");
        my $host_add_req = 1;
        my @parsed_output = @{$node_obj->rdfile('file' => "/etc/hosts",)
                ->get_parsed_output()};
        push(
            @{$Host_Info->{"$host_node"}->{contents}},
            split(/\n/, $parsed_output[0]->{'contents'})
        );

        $Log->comment(
            "Check if $host_add already on host file of $host_node");
        for my $content (@{$Host_Info->{"$host_node"}->{contents}}) {
            if ($content eq "$host_ip $host_add") {
                $Log->comment(
                    "Host info of $host_add found in host file of $host_node"
                );
                $Host_Info->{"$host_node"}->{no_modify} = 1;
                $host_add_req = 0;
            }
        }

        if ($host_add_req) {
            $Log->comment(
                "Add host $host_add entry to host file on $host_node");
            $node_obj->wrfile(
                'append'  => 1,
                'file'    => '/etc/hosts',
                'content' => "$host_ip $host_add",
            );
        }
        $node_obj->rdfile('file' => "/etc/hosts",);

    } elsif ($opt eq 'restore' && @{$Host_Info->{"$host_node"}->{contents}}) {
        if ($Host_Info->{"$host_node"}->{no_modify}) {
            $Log->comment("No restore is needed for $host_node");
        } else {
            $Log->comment("Remove host file with extra entry");
            $node_obj->rm(
                'privilege-level' => 'test',
                'paths'           => '/etc/hosts',
            );

            $Log->comment("Restore host file of $host_node from back up");
            for my $content (@{$Host_Info->{"$host_node"}->{contents}}) {
                $node_obj->wrfile(
                    'append'  => 1,
                    'file'    => '/etc/hosts',
                    'content' => "\'$content\'",
                );
            }
        }
    } else {
        if ($opt eq 'restore') {
            $err_msg = "No host file back up found for $host_node!";
        } else {
            $err_msg = "Option $opt is not supported!";
        }
        $Log->error("$err_msg");
        NATE::BaseException->throw("$err_msg");
    }
} ## end sub _host_file_operation

1;
