#
# Copyright 2012 NetApp, Inc.
#
# Snapshot related actions for NX or general autmoation use.
#
#
package NX::Snapshot;

use strict 'vars';
use Tharn;
use Data::Dumper;
use NX::Helper;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::STask::VolumeSnapshot;
use NACL::STask::VserverNFS;
use NACL::STask::Client::Mount;
use NACL::C::VolumeSnapshotPolicy;
use NACL::C::JobScheduleCron;
use NACL::C::Cluster;
use NACL::STask::SystemLicense;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
    create     schedule     delete      incore_buf  io_vol_restore
);


my @Pkg_Workers;
sub stop_workers {
    $Log->comment("Detected interrupt signal");
    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}


=pod

=head1 NAME

NX::Snapshot

=head1 DESCRIPTION

Snapshot-related routines for use in NX framework.

=head1 METHODS

=head2 create

Create specified quantity of snapshots per volume on specified volumes.  Each
volume spawns a separate process, up to 50 to create snapshots in parallel.

The input is a hash (%opts) which must include:
    volumes     => [ $vol_obj ] # Array ref of 1 or more volume objects
    qty         => # of snapshots to create per volume, or "infinite"

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX create
Create
description
Creates snapshots on volumes
required
volumes;Volume(s);;Volume(s) to create snapshots on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
qty;Qty;;Integer value of snapshots to create per volume, or "infinite"
optional
base_name;Base Name;ss_<vol_name>_<random>;Base name for snapshots. Snapshot names will be <base_name>_<vol_name>_<random>.
interval;Interval;;Interval(in seconds) between taking snapshots
=end NX create
=cut

sub create{
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;
    my $obj_ref;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                volumes           => { type => ARRAYREF, optional => 0 },
                qty               => { type => SCALAR,   optional => 0 },
                base_name         => { type => SCALAR,   optional => 1 },
                interval          => { type => SCALAR,   optional => 1 },
                validate_only     => { type => SCALAR,   optional => 1, default => 0 },
                _num_nodes        => { type => SCALAR,   optional => 1, default => undef },
            }
        );

        foreach my $arg (keys %input ) {
            my $val = $input{$arg};
            if ( $arg =~ /^(qty)$/ ) {
                if ( ( $val !~ /^\d+$/ ) && ( $val ne 'infinite' ) ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer or \'infinite\'");
                }
            } elsif ( $arg eq 'interval' ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer");
                }
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Snapshot create action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Snapshot::stop_workers";
        delete $input{validate_only};
    }
    

    my @volumes = @{$input{volumes}};
    my $num_volumes = @volumes;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;

    # Extract number of unique nodes, if we haven't already
    # determined that from a previous call
    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v ( @volumes ) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined($node_hash->{$name}) ) {
                push(@nodes, $node);
                $node_hash->{$name} = 1;
            }
            push(@{$node_to_volumes->{$name}}, $v);
        }
        $num_nodes = @nodes;
    } else {
        # Irrelevant
    }

    $my_return->{start_tick} = time();
    if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n ( @nodes) {
            my $node_name = $n->name();
            my $id = 'SSC_N_DaC.'.$proc_cnt; # (N)ode_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Indicate we already parsed out unique nodes, no need
            # to do it again
            $input{_num_nodes} = $num_nodes;
            $Log->comment("SSC_N_DaC".$proc_cnt." doling ".
                @{$node_to_volumes->{$node_name}}. " volumes");
            $input{volumes}   = $node_to_volumes->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Snapshot',
                method  => 'create',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }
    
        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
            merge_context   => 0,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $node_ref = $res->{$rid}->{objects}->{node};
                foreach my $node_name (keys %{$node_ref} ) {
                    my @res_obj = @{$node_ref->{$node_name}};
                    push( @{$obj_ref->{node}->{$node_name}}, @res_obj );
                }
            }

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Created snapshots on $num_nodes nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref if $obj_ref;
        }
        return $my_return;
    } elsif ( $num_volumes > 1 ) {
        # Multiple volumes on one node
        # Split into batches of 50 volumes
        my $node_name = $volumes[0]->command_interface()->name();
        my $idx = 0;
        my $done = 0;
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        while ( !$done ) {
            my @cur_vols;
            for (1..50) {
                if ( $volumes[$idx] ) {
                    push(@cur_vols, $volumes[$idx]);
                    $idx++;
                } else {
                    $done = 1;
                    last;
                }
            }
            my $num_cur_vols = @cur_vols;

            # Work on current batch of volumes
            foreach my $v (@cur_vols) {
                my $id = 'SSC_V_DaC.'.$proc_cnt; # (V)ode_(D)ivide(a)nd(C)onquer
                push(@ids, $id);
    
                # Indicate we already parsed out unique nodes, no need
                # to do it again
                $input{volumes}   = [ $v ];
                $input{_num_nodes} = 1;
    
                my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
                my $future = $worker->submit(
                    package => 'NX::Snapshot',
                    method  => 'create',
                    args    => [ %input ],
                );
    
                push( @workers, $worker );
                push( @Pkg_Workers, $worker );
                $futures->{$id}->{obj} = $future;
                $proc_cnt++;
            }
        
            # Collect results
            my $res = collect_worker_results(
                workers         => \@workers,
                futures         => $futures,
                merge_context   => 0,
            );

            my $had_failure = 0;
            my $obj_ref;
            foreach my $w ( @workers ) {
                my $rid = $w->runid();
    
                # Bundle objects from all the workers
                if ( $res->{$rid}->{objects} ) {
                    my $node_ref = $res->{$rid}->{objects}->{node};
                    foreach my $node_name (keys %{$node_ref} ) {
                        my @res_obj = @{$node_ref->{$node_name}};
                        push( @{$obj_ref->{node}->{$node_name}}, @res_obj );
                    }
                }
    
                if ( $res->{$rid}->{result} != 1 ) {
                    $Log->error("$rid had a failure");
                    $had_failure = 1;
                } else {
                    $Log->comment("$rid ran successfully");
                }
            }

            if ( $had_failure ) {
                $my_return->{end_tick} = time();
                $my_return->{result} = 0;
                $my_return->{msg} = 'One or more procs had failures';
                return $my_return;
            }
        }

        $my_return->{end_tick} = time();
        $my_return->{result} = 1;
        my $msg = "Created snapshots on all $num_volumes volumes on $node_name";
        $Log->comment( $msg );
        $my_return->{msg} = $msg;
        $my_return->{objects} = $obj_ref if $obj_ref;
        return $my_return;
        
    } else {
        # Single node, single volume
        delete $input{volumes};
        delete $input{_num_nodes};
        my $qty = delete $input{qty};
        my $base_name = delete $input{base_name};
        my $node = $volumes[0]->command_interface();
        my $node_name = $node->name();
        my $vol_name = $volumes[0]->volume();
        $input{command_interface} = $node;
        srand();

        my $count = 0;
        try {
            while (1) {
                last if (($count == $qty) && ($qty ne 'infinite'));
                my $snap_name;
                my $t = time();
                $t =~ /([\d]{7})$/;
                $t = $1;
                my $suffix = int(rand(100)) . int(rand(100));
                if (!$base_name) {
                    $snap_name = "ss_".$vol_name."_".$t."_".$suffix;
                } else {
                    $snap_name = $base_name."_".$vol_name."_".$t."_".$suffix;
                }
                $input{vserver} = $volumes[0]->vserver();
                my $snapshot_obj = NACL::STask::VolumeSnapshot->create(
                    volume              => $vol_name,
                    snapshot            => $snap_name,
                    'method-timeout'    => 300,
                    %input,
                );
                $count++;
                sleep $input{interval} if $input{interval};
                #push(@{$obj_ref->{node}->{$node_name}}, $snapshot_obj);
            }
        } otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Created $qty snapshots on volume $vol_name on node $node_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
        }
        return $my_return;
    }
} # End of create

=pod

=head2 schedule

Set snapshot schedule for specified volumes.

The input is a hash (%opts) which must include:
    volumes     => [ $vol_obj ] # Array ref of 1 or more volume objects

=head3 opts

%opts can be used to override default values, which include
    hours   => default 0
    minutes => default 0
    count   => default 1
    policy  => 

=over

=back

=cut

=cut begin NX schedule
Schedule
description
Set snapshot schedule for specified volumes.
required
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
hours;Hours;;Hours to create snapshot, comma delimited integers or 'all', default 0
minutes;Minutes;;Minutes on the hour to create snapshot, comma delimited integers or 'all', default 0
count;Count;1;Amount of snapshots to create
policy;Policy;;Policy Name
=end NX schedule
=cut

sub schedule {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                volumes             => { type => ARRAYREF, optional => 0 },
                validate_only       => { type => SCALAR,   optional => 1, default => 0 },
                hours               => { type => SCALAR,   optional => 1, default => 0 },
                minutes             => { type => SCALAR,   optional => 1, default => 0 },
                count               => { type => SCALAR,   optional => 1, default => 1 },
                policy              => { type => SCALAR,   optional => 1, },
                _num_nodes          => { type => SCALAR,   optional => 1, },
            }
        );

        foreach my $arg (keys %input ) {
            if ( $arg =~ /^(hours|minutes)$/ ) {
                my @tmp = split(/,/, $input{$arg});
                foreach my $v (@tmp) {
                    if (($v !~ /^\d+$/) && ($v ne 'all')) {
                        NATE::BaseException->throw("\'$arg\' value must be "
                            . "comma delimited integers or \'all\'");
                    }
                }
            }
        }
        if ( !$input{hours} && !$input{minutes} ) {
            NATE::BaseException->throw("Either the \'hours\' or \'minutes\' "
                . "parameter must be specified");
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Snapshot schedule action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Snapshot::stop_workers";
        delete $input{validate_only};
    }

    my @volumes = @{$input{volumes}};
    my $num_volumes = @volumes;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;
    # Extract number of unique nodes, if we haven't already
    # determined that from a previous call
    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v ( @volumes ) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined($node_hash->{$name}) ) {
                push(@nodes, $node);
                $node_hash->{$name} = 1;
            }
            push(@{$node_to_volumes->{$name}}, $v);
        }
        $num_nodes = @nodes;
    } else {
        # Irrelevant
    }

    $my_return->{start_tick} = time();
    #if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {
    if ( !$input{_num_nodes} ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $cluster_members;
        my $cmode_node_handled;
        my $had_failure = 0;
        my $obj_ref;

        try {
            # Figure out which nodes belong to cluster
            foreach my $n ( @nodes ) {
                if ( $n->mode() =~ /cmode/i ) {
                    my $cur_name = $n->name();
                    my @members = NACL::C::Cluster->nodes( command_interface => $n );
                    foreach ( @members ) {
                        push( @{$cluster_members->{$cur_name}}, $_ )
                            if ( $_ ne $cur_name);
                    }
                }
            }
    
            foreach my $n ( @nodes) {
                my $node_name = $n->name();
                my $id = 'SSSched_N_DaC.'.$proc_cnt; # (N)ode_(D)ivide(a)nd(C)onquer
                push(@ids, $id);
    
                # Indicate we already parsed out unique nodes, no need
                # to do it again
                $input{_num_nodes} = $num_nodes;
                $Log->comment("SSSched_N_DaC".$proc_cnt." doling ".
                    @{$node_to_volumes->{$node_name}}. " volumes");
                $input{volumes} = $node_to_volumes->{$node_name};
    
                # If cmode, create schedule and policy as these are cluster wide
                if ( $n->mode() =~ /cmode/i ) {
                    if ( grep { $cmode_node_handled->{$_} } @{$cluster_members->{$node_name}} ) {
                        $Log->comment("Schedule and policy already created on "
                            . " cluster for node $node_name");
                    } else {
                        my $sched_name = create_cron_sched(
                            command_interface => $n,
                            hours             => $input{hours},
                            minutes           => $input{minutes},
                        );
                        $cmode_node_handled->{$node_name} = 1;
    
                        my $pol_exists = 0;
                        try {
                            # Ensure policy exists, if not, create it on CMode
                            NACL::C::VolumeSnapshotPolicy->find(
                                command_interface => $n,
                                filter => {
                                    policy   => $input{policy},
                                }
                            );
                            $pol_exists = 1;
                        } catch NACL::Exceptions::NoElementsFound with {
                            # Policy doesn't exist
                        };
        
                        if ( !$pol_exists ) {
                            # Create the policy with the cron job schedule
                            # previously created
                            my %opts = (
                                command_interface   => $n,
                                policy              => $input{policy},
                                enabled             => 'true',
                                schedule1           => $sched_name,
                                count1              => $input{count},
                            );
                            NACL::C::VolumeSnapshotPolicy->create( %opts );
                        } else {
                            my $pol = NACL::C::VolumeSnapshotPolicy->find(
                                command_interface => $n,
                                filter => { policy => $input{policy} }
                            );
                            my @scheds = $pol->state()->schedules();
                            if (grep{$_ eq $sched_name} @scheds) {
                                $Log->comment("Schedule $sched_name already "
                                    . "part of policy $input{policy}");
                            } else {
                                $Log->comment("Adding schedule $sched_name "
                                    . "to policy $input{policy}");
                                NACL::C::VolumeSnapshotPolicy->add_schedule(
                                    command_interface => $n,
                                    policy            => $input{policy},
                                    schedule          => $sched_name,
                                    count             => $input{count}
                                );
                            }
                        }
                    }
                }
    
                my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
                my $future = $worker->submit(
                    package => 'NX::Snapshot',
                    method  => 'schedule',
                    args    => [ %input ],
                );
    
                push( @workers, $worker );
                push( @Pkg_Workers, $worker );
                $futures->{$id}->{obj} = $future;
                $proc_cnt++;
                sleep 5;
            }
        
            # Collect results
            my $res = collect_worker_results(
                workers         => \@workers,
                futures         => $futures,
                merge_context   => 0,
            );

            foreach my $w ( @workers ) {
                my $rid = $w->runid();
    
                # Bundle objects from all the workers
                if ( $res->{$rid}->{objects} ) {
                    my $node_ref = $res->{$rid}->{objects}->{node};
                    foreach my $node_name (keys %{$node_ref} ) {
                        my @res_obj = @{$node_ref->{$node_name}};
                        push( @{$obj_ref->{node}->{$node_name}}, @res_obj );
                    }
                }
    
                if ( $res->{$rid}->{result} != 1 ) {
                    $Log->error("$rid had a failure");
                    $had_failure = 1;
                } else {
                    $Log->comment("$rid ran successfully");
                }
            }
        } otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
            return $my_return;
        } elsif ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Set volume snapshot schedule on $num_nodes nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref if $obj_ref;
        }
        return $my_return;
    } else {
        # Single node, mulitple volumes
        my $node = $volumes[0]->command_interface();
        $input{command_interface} = $node;
        my $node_name = $node->name();
        my $snap_name = "nx_sss_".time();

        if ( $node->mode() =~ /cmode/i ) {
            try {
                foreach my $vol_obj (@volumes) {    
                    $vol_obj->modify( 'snapshot-policy' => $input{policy} );
                }
            } otherwise {
                $exception = shift;
            };
        } else { # 7mode
            try {
                foreach my $vol_obj ( @volumes ) {
                    $input{snapshot}    = $snap_name;
                    $input{volume}      = $vol_obj->volume();
                    NACL::STask::VolumeSnapshot->schedule( %input );
                }
            } otherwise {
                $exception = shift;
            };
        }

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Set volume snapshot schedule on $num_volumes volumes on node $node_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
} # End of schedule

=pod

=head2 delete

Delete specified quantity of snapshots per volume on specified volumes.  Each
volume spawns a separate process, up to 50 to create snapshots in parallel.
Function polls for snapshot list on each volume.  If no snapshots exist,
keep polling.  This is done until we delete the specified quantity.

The input is a hash (%opts) which must include:
    volumes     => [ $vol_obj ] # Array ref of 1 or more volume objects
    qty         => # of snapshots to delete per volume, or "infinite"

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX delete
Delete
description
Deletes snapshots on volumes
required
volumes;Volume(s);;Volume(s) to delete snapshots on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
qty;Qty;;Integer value of snapshots to delete per volume, or "infinite"
optional
base_name;Base Name;;Base name for snapshots. Snapshots will be searched for as <base_name>*.
interval;Interval;;Interval(in seconds) between deleting snapshots
=end NX delete
=cut

sub delete {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;
    my $obj_ref;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                volumes           => { type => ARRAYREF, optional => 0 },
                qty               => { type => SCALAR,   optional => 0 },
                base_name         => { type => SCALAR,   optional => 1 },
                interval          => { type => SCALAR,   optional => 1 },
                validate_only     => { type => SCALAR,   optional => 1, default => 0 },
                _num_nodes        => { type => SCALAR,   optional => 1, default => undef },
            }
        );

        foreach my $arg (keys %input ) {
            my $val = $input{$arg};
            if ( $arg =~ /^(qty)$/ ) {
                if ( ( $val !~ /^\d+$/ ) && ( $val ne 'infinite' ) ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer or \'infinite\'");
                }
            } elsif ( $arg eq 'interval' ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer");
                }
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Snapshot create action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Snapshot::stop_workers";
        delete $input{validate_only};
    }

    my @volumes = @{$input{volumes}};
    my $num_volumes = @volumes;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;
    # Extract number of unique nodes, if we haven't already
    # determined that from a previous call
    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v ( @volumes ) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined($node_hash->{$name}) ) {
                push(@nodes, $node);
                $node_hash->{$name} = 1;
            }
            push(@{$node_to_volumes->{$name}}, $v);
        }
        $num_nodes = @nodes;
    } else {
        # Irrelevant
    }

    $my_return->{start_tick} = time();
    if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n ( @nodes) {
            my $node_name = $n->name();
            my $id = 'SSD_N_DaC.'.$proc_cnt; # (N)ode_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Indicate we already parsed out unique nodes, no need
            # to do it again
            $input{_num_nodes} = $num_nodes;
            $Log->comment("SSD_N_DaC".$proc_cnt." doling ".
                @{$node_to_volumes->{$node_name}}. " volumes on $node_name");
            $input{volumes}   = $node_to_volumes->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Snapshot',
                method  => 'delete',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }
    
        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
            merge_context   => 0,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $node_ref = $res->{$rid}->{objects}->{node};
                foreach my $node_name (keys %{$node_ref} ) {
                    my @res_obj = @{$node_ref->{$node_name}};
                    push( @{$obj_ref->{node}->{$node_name}}, @res_obj );
                }
            }

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Deleted snapshots on $num_nodes nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref if $obj_ref;
        }
        return $my_return;
    } elsif ( $num_volumes > 1 ) {
        # Multiple volumes on one node
        # Split into batches of 50 volumes
        my $node_name = $volumes[0]->command_interface()->name();
        my $idx = 0;
        my $done = 0;
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        while ( !$done ) {
            my @cur_vols;
            for (1..50) {
                if ( $volumes[$idx] ) {
                    push(@cur_vols, $volumes[$idx]);
                    $idx++;
                } else {
                    $done = 1;
                    last;
                }
            }
            my $num_cur_vols = @cur_vols;

            # Work on current batch of volumes
            foreach my $v (@cur_vols) {
                my $id = 'SSD_V_DaC.'.$proc_cnt; # (V)ode_(D)ivide(a)nd(C)onquer
                push(@ids, $id);
    
                # Indicate we already parsed out unique nodes, no need
                # to do it again
                $input{volumes}   = [ $v ];
                $input{_num_nodes} = 1;
    
                my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
                my $future = $worker->submit(
                    package => 'NX::Snapshot',
                    method  => 'delete',
                    args    => [ %input ],
                );
    
                push( @workers, $worker );
                push( @Pkg_Workers, $worker );
                $futures->{$id}->{obj} = $future;
                $proc_cnt++;
            }

            # Collect results
            my $res = collect_worker_results(
                workers         => \@workers,
                futures         => $futures,
                merge_context   => 0,
            );

            my $had_failure = 0;
            my $obj_ref;
            foreach my $w ( @workers ) {
                my $rid = $w->runid();
    
                # Bundle objects from all the workers
                if ( $res->{$rid}->{objects} ) {
                    my $node_ref = $res->{$rid}->{objects}->{node};
                    foreach my $node_name (keys %{$node_ref} ) {
                        my @res_obj = @{$node_ref->{$node_name}};
                        push( @{$obj_ref->{node}->{$node_name}}, @res_obj );
                    }
                }
    
                if ( $res->{$rid}->{result} != 1 ) {
                    $Log->error("$rid had a failure");
                    $had_failure = 1;
                } else {
                    $Log->comment("$rid ran successfully");
                }
            }

            if ( $had_failure ) {
                $my_return->{end_tick} = time();
                $my_return->{result} = 0;
                $my_return->{msg} = 'One or more procs had failures';
                return $my_return;
            }
        }

        $my_return->{end_tick} = time();
        $my_return->{result} = 1;
        my $msg = "Deleted snapshots on all $num_volumes volumes on $node_name";
        $Log->comment( $msg );
        $my_return->{msg} = $msg;
        $my_return->{objects} = $obj_ref if $obj_ref;
        return $my_return;
        
    } else {
        # Single node, single volume
        delete $input{volumes};
        delete $input{_num_nodes};
        my $qty = delete $input{qty};
        my $node = $volumes[0]->command_interface();
        my $node_name = $node->name();
        my $vol_name = $volumes[0]->volume();
        srand();

        my $count = 0;
        while (1) {
            try {
                if (($count == $qty) && ($qty ne 'infinite')) {
                    $Log->comment("Met snapshot delete qty of $qty for "
                        . "volume $vol_name");
                    last;
                } else {
                    $Log->comment("$count / $qty snapshots deleted so far "
                        . "for volume $vol_name");
                }

                my @snaps;

                if ($input{base_name}) {
                    # Find snapshots beginning with <base_name>
                    @snaps = NACL::STask::VolumeSnapshot->find(
                        command_interface => $node,
                        filter => {
                            volume   => $vol_name,
                            snapshot => $input{base_name}.'*',
                        }
                    );
                } else {
                    # Find any/all snapshots for the volume
                    @snaps = NACL::STask::VolumeSnapshot->find(
                        command_interface => $node,
                        filter => {
                            volume => $vol_name,
                        }
                    );
                }

                foreach my $ss ( @snaps ) {
                    $ss->purge();
                    $count++;
                    last if ( $count == $qty );
                }
                sleep $input{interval} if $input{interval};
            } catch NACL::Exceptions::NoElementsFound with {
                # Didn't find any matching snapshots, that's
                # okay, try again
            } otherwise {
                $exception = shift;
            };
        }

        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Deleted $qty snapshots on volume $vol_name on node $node_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
        }
        return $my_return;
    }
} # End of delete


sub _configure_cmode_nfs {
    my %input = @_;
    my $ci      = $input{command_interface};
    my $type    = $input{client_type};
    my $vserver = $input{vserver};

    $Log->comment("Configuring CMode NFS on Vserver $vserver on node "
        . $ci->name());

    if ( $type =~ /unix/i ) {
        # Configure VServer to support NFS
        my $vserver_nfs_exists = 0;
        try {
            my $cmode_apiset = $ci->apiset();
            my $tmp = $cmode_apiset->vserver_nfs_show(
                vserver => $vserver,
                fields  => "access,v3,v4.0,tcp,udp",
            )->get_parsed_output()->[0];
            $vserver_nfs_exists = 1;
        }
        catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            # The exception will be thrown if the vserver doesn't exist
            NACL::STask::VserverNFS->create(
                command_interface => $ci,
                vserver           => $vserver,
                access            => "true",
                'v4.0'            => "enabled",
                'v3'              => "enabled",
                udp               => "enabled",
                tcp               => "enabled"
            );
        };
        if ($vserver_nfs_exists) {
            $Log->comment("Vserver NFS settings exist for $vserver, modifying...");
            NACL::STask::VserverNFS->modify(
                command_interface => $ci,
                vserver           => $vserver,
                access            => "true",
                'v4.0'            => "enabled",
                'v3'              => "enabled",
                udp               => "enabled",
                tcp               => "enabled"
            );
        }
    } else {
        $Log->warn("Don't know how to configure client type $type");
    }
} ## end sub _configure_cmode_nfs()

sub _setup_nfs {
    my %input = @_;
    my @volumes = @{$input{volumes}};
    my $unique_nodes;
    my $vs_on_nodes;

    foreach my $v ( @volumes ) {
        # Get all unique node command_interfaces
        my $n = $v->command_interface();
        my $n_name = $n->name();
        if ( !$unique_nodes->{$n_name} ) {
            $unique_nodes->{$n_name} = $n;
        }

        if ( $n->mode =~ /cmode/i ) {
            my $vs = $v->vserver();
            push (@{$vs_on_nodes->{$n_name}}, $vs);
        } else { # 7Mode
            my $api = $n->get_7m_or_nodescope_apiset();
            my $vol_name = $v->volume();
            # Note, this may leave an extra line in /etc/exports
            $Log->comment("Changing export rules for volume $vol_name");
            $api->exportfs(
                "path"               => "/vol/$vol_name",
                "persistent"         => 1,
                "persistent-options" => 'anon=0',
            );
        }
    }

    foreach my $n_name (keys %{$unique_nodes} ) {
        if ( $vs_on_nodes->{$n_name} ) {
            # This node is CMode and has VServers
            my @vs = @{$vs_on_nodes->{$n_name}};
            foreach my $vs ( @vs ) {
                _configure_cmode_nfs(
                    command_interface => $unique_nodes->{$n_name},
                    vserver           => $vs,
                    client_type       => 'unix',
                );
            }
        }
    }
} # End of _setup_nfs


#=pod

=head2 incore_buf

Run sio on specified number of files on specified clients in parallel.
Each client requires one login session to perform all sio operations.

The input is a hash (%input) which must include:
    client_command_interface => [ $client ] # Array ref of client objects

%inputs can be used to provide additional options for sio, which include
    blksz             => <blksz>      param               default => '4k'
    start             => <start_byte> param               default => 0
    end               => <end_byte>   param               default => 8000
    secs              => <secs>       param               default => 10
    threads           => <threads>    param               default => 2
    volumes           => [ $vol ]  # Array ref of volume
                         objects after sio run completes
    num_files         => [$num_files] # num of files      default => 3
                         on which sio should be run

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=cut

=cut begin NX incore_buf
Incore Buf
description
Run sio on specified number of files on specified clients in parallel. Each client requires one login session to perform all sio operations.
required
client_command_interface;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
blksz;Blksz;4k;Block size for sio tool, default 4k
start;Start;0;Start position for sio tool, default 0
end;End;8000;End position for sio tool, default 8000
threads;Threads;2;Number of sio threads to run
secs;Secs;10;Duration to run(in seconds), default 10
num_files;Num Files;3;Number of files to use, default 3
=end NX incore_buf
=cut

sub incore_buf {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;
    my $obj_ref;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                client_command_interface => { type => ARRAYREF, optional => 0 },
                volumes           => { type => ARRAYREF, optional => 0 },
                blksz             => { type => SCALAR, optional => 1, default => '4k' },
                start             => { type => SCALAR, optional => 1, default => 0 },
                end               => { type => SCALAR, optional => 1, default => '8000' },
                threads           => { type => SCALAR, optional => 1, default => 2 },
                secs              => { type => SCALAR, optional => 1, default => 10 },
                ofileid           => { type => SCALAR, optional => 1 },
                num_files         => { type => SCALAR, optional => 1, default => 3 },
                skip_nfs_setup    => { type => SCALAR,   optional => 1, default => 0 },
                validate_only     => { type => SCALAR, optional => 1, default => 0 }
            }
        );

        foreach my $key (keys %input) {
           if ( $key eq "blksz") {
                if ($input{$key} !~ /^\d+[kKmMgG]?$/) {
                   NATE::BaseException->throw("\'$key\' must be an integer optionally followed by k,m or g");
                }
           } elsif ( $key eq "start") {
                if ($input{$key} !~ /^\d+$/) {
                   NATE::BaseException->throw("\'$key\' must be an integer");
                }
           } elsif ( $key eq "end") {
                if ($input{$key} !~ /^\d+$/) {
                   NATE::BaseException->throw("\'$key\' must be an integer");
                }
           } elsif ( $key eq "secs") {
                if ($input{$key} !~ /^\d+$/) {
                   NATE::BaseException->throw("\'$key\' must be an integer");
                }
           } elsif ( $key eq "threads") {
                if ($input{$key} !~ /^\d+$/) {
                    NATE::BaseException->throw("\'$key\' must be an integer");
                }
           } elsif ( $key eq "num_files") {
                if ($input{$key} !~ /^\d+$/) {
                    NATE::BaseException->throw("\'$key\' must be an integer");
                }
           }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "incore_buf action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Snapshot::stop_workers";
        delete $input{validate_only};
    }

    my @client_objs = @{$input{client_command_interface}};
    my $num_clients = @client_objs;
    my @volume_objs = @{$input{volumes}};
    my $num_volumes = @volume_objs;
    my $num_files   = $input{'num_files'};
    my $active_sio_pids;

    # On all volumes, set up NFS and export
    _setup_nfs( volumes => $input{volumes} ) if !$input{skip_nfs_setup};

    if ($num_clients > 1) {
        # More than one client specified, divide and conquer by splitting
        # work off on each client in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $c ( @client_objs ) {
            my $id = 'incore_buf_C_DaC.'.$proc_cnt; # (C)lient_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Replace command_interface with a reference to an array
            # with just the single client.  Keep all other original
            # inputs to pass in
            $input{client_command_interface} = [ $c ];
            $input{ofileid} =  $id; # Will be used to distinguish different incore_buf files worked on.
            $input{skip_nfs_setup}   = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Snapshot',
                method  => 'incore_buf',
                args    => [ %input ],
            );
            push( @Pkg_Workers, $worker );
            push( @workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
            merge_context   => 0,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();
            $had_failure = 1 if ( $res->{$rid}->{result} != 1 );

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $client_ref = $res->{$rid}->{objects}->{client};
                foreach my $client_name (keys %{$client_ref} ) {
                    my @mnt_objs = @{$client_ref->{$client_name}};
                    push( @{$obj_ref->{client}->{$client_name}}, @mnt_objs );
                }
            };
        }

        if ( $had_failure ) {
            $my_return->{result} = 0;
            my $msg = 'One or more procs had failures';
            $Log->error( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran sio on $num_clients clients";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{objects} = $obj_ref;
        }
        return $my_return;
    } else  {
        # Single client, one or more volumes specified, divide and
        # conquer by splitting work off for each volume in parallel
        my $client = shift (@client_objs);
        my $client_name = $client->name();
        my $api = $client->apiset();
        my $found_errors = 0;
        my $ofile_idx=0;

        my @siocmds;
        my $siocmd;
        my @ofiles;
        my $ofile;
        my $pid;
        my $raw_results;
        my $input_type;
        my $active_sio_pids;

        my $ofileid = $input{ofileid};
        my $mounts_and_paths =
             _mount_client_volumes( client => $client, volumes => \@volume_objs );
        push( @{$obj_ref->{client}->{$client_name}}, @{$mounts_and_paths->{mounts}} );

        my @paths_to_ls;
        # perform operations on each volume
        foreach my $v ( @volume_objs ) {
           try {
                my $path = $mounts_and_paths->{paths}->{$v->volume()};
                $path .= "/.snapshot";
                push(@paths_to_ls, $path);
                my $snapdir;
                #Look for (most recently modified) directories
                my $raw_output_1 = $api->execute_raw_command
                    ("command" => "ls -tAp $path| grep '\/\$' | head -n 2");

                if ($raw_output_1 eq '') {
                    NATE::BaseException->throw ("No directories found in $path");
                } else {
                    # Extract out just the directory, in case
                    # prior outstanding pid complete message becomes
                    # part of stdout and messes up our 'ls' output
                    my @lines = split(/\n/, $raw_output_1);
                    my $num_lines = @lines;
                    if ($num_lines > 1) {
                        $snapdir = $lines[1];
                    } else {
                        $snapdir = $lines[0];
                    }
                    $snapdir =~ s/^\s+//;
                    my @tmp = split(/\s/, $snapdir);
                    $snapdir = $tmp[0];
                }

                # Append snapdir to $path.
                $Log->comment("Snap directory: *$snapdir*");
                $path .= "/".$snapdir;
                push(@paths_to_ls, $path);

                #Look for $num_files most recently modified files in the directory
                my $raw_output_2 = $api->execute_raw_command
                    ("command" => "ls -tAp $path | grep -v '\/\$' |head -n $num_files");
                if ($raw_output_2 eq '') {
                    NATE::BaseException->throw ("No files found in $path");
                }

                my @lines = split(/\n/, $raw_output_2);
                if (@lines < $num_files) {
                    #
                } elsif (@lines != $num_files) {
                    NATE::BaseException->throw("There are less than $num_files in path $path");
                }
                my $raw_output_3 = $api->execute_raw_command  ("command" => "cd $path");
                #Print pwd value so that reading log files is easier
                my $pwd_output = $api->execute_raw_command ("command" => "pwd");

                #Each line $raw_output_2 represents a file.
                # Run sio on each of these files, in parallel
                foreach my $l (@lines) {
                    my $filepath = $path."/".$l;
                    $ofile = "/tmp/sioout_".$ofileid."_$ofile_idx.txt";
                    $ofile_idx++;
                    my $siocmd = "sio 100 0 $input{blksz} $input{start} $input{end} $input{secs} $input{threads} ";
                    $siocmd .= $filepath;

                    push (@siocmds,  $siocmd);
                    push (@ofiles, $ofile);
                    my $pid = _run_cmd_get_pid (
                               api_set => $api,
                               cmd     => "$siocmd  >\& $ofile \&",
                               proc_name => 'sio');
                    $active_sio_pids->{$pid} = 1;
                }
            } otherwise {
                $exception = shift;
                if ($::IGNORE_ERRORS) {
                    $Log->warn("Got following exception, ignoring and "
                        . "continuing: ".$exception->text());
                    $exception = "";#Reset the exception.
                }
            }
        } # end foreach loop, running sio on each volume.

        # Wait till all sio commands for this volume are done.
        _wait_till_all_sio_complete ($api,$active_sio_pids);

        # Make sure that the results of sio commands will come in log files.
        my $num = @siocmds;
        $raw_results = "";
        for (my $i=0;$i<$num;$i++) {
            $raw_results .= $api->execute_raw_command (command => "echo $siocmds[$i]");
            $raw_results .= $api->execute_raw_command (command => "cat $ofiles[$i]");
        }
        # rm the files used to direct output to.
        foreach $ofile (@ofiles){
            $Log->comment("removing file $ofile");
            $api->execute_raw_command (command => "rm $ofile")
        }

        # Analyze results and report errors.
        # Note that only the last exception survives.
        my @lines = split(/\n/, $raw_results);
        my $curr_sio_cmd;
        my $failing_msg;
        foreach my $l ( @lines ) {
            if ($l =~ /echo (sio .*)/) {
                $curr_sio_cmd = $1;
            }
            if ($l =~ /SIO_ERROR/i) {
                $found_errors = 1;
                $failing_msg .= $curr_sio_cmd."\n";
                $failing_msg .= "Error line is $l"."\n";
            }
        }

        $my_return->{objects} = $obj_ref if $obj_ref;
        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } elsif ( $found_errors ) {
            $Log->comment("Performing ls of directories to see their "
                . "contents for debugging");
            foreach (@paths_to_ls) {
                $api->execute_raw_command (command => "ls -l $_");
            }
            my $msg = "sio reported failures against one or more volumes. "
                     ."Found error string(s) in sio output on $client_name. Fail info follows:"
                     .$failing_msg;
            if ($::IGNORE_ERRORS){
                $my_return->{result} = 1;
                $Log->warn( $msg );
            } else {
                $my_return->{result} = 0;
                $Log->error( $msg );
            }
            $my_return->{msg} = $msg;
        } else {
            $my_return->{result} = 1;
            my $msg = "Incore_buf ran successfully without errors on $client_name";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }

        return $my_return;
    } # End of else loop

} #end sub incore_buf


=head2 io_vol_restore

Perform file overwrites on specified volumes. Creates a snapshot, and then
a subsequent volume snapshot restore.

=head3 opts

%opts can be used to provide additional options for io_vol_restore
    validate_only            => <VALIDATE_ONLY option>
    client_command_interface => <array ref of client objects>
    volumes                  => <arrayref of volume objects>
    iterations               => <integer>|infinite (default is 1)
    interval                 => odd/even/all/<value in seconds>
                                Every odd, even, or every clock hour,
                                or specified value in seconds
    start_minute             => Required if "interval" is odd|even|all, otherwise
                                ignored if "interval" is an integer value.  This is
                                a two-digit value representing the clock minute
    sio_file_size            => Size of each siofile, defaults to 10m
    sio_num_files            => Number of files for sio
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX io_vol_restore
Perform IO and volume snapshot restore
description
Perform file overwrites on specified volumes. Creates a snapshot, and then a subsequent volume snapshot restore.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volume(s);;Volume(s) to create clones and run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
iterations;Iterations;1;Number of times to run the clone split operation on each volume
interval;Interval;1;interval of time to wait between clone split operations
start_minute;Start Minute;1;Required if 'interval' is odd|even|all, otherwise ignored if 'interval' is an integer value.
sio_file_size;File Size for sio;1m;Defaults to '10m'
sio_num_files;Number of files for sio;5;Defaults to 5
=end NX io_vol_restore
=cut

sub io_vol_restore {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                client_command_interface =>
                    { type => ARRAYREF, optional => 0 },
                volumes =>
                    { type => ARRAYREF, optional => 0 },
                iterations =>
                    { type => SCALAR, optional => 1, default => 1 },
                interval     =>
                    { type => SCALAR, optional => 1 },
                start_minute =>
                    { type => SCALAR, optional => 1 },
                sio_file_size =>
                    { type => SCALAR, optional => 1, default => '10m' },
                sio_num_files =>
                    { type => SCALAR, optional => 1, default => 5 },
                _ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
                _num_nodes =>
                    { type => SCALAR, optional => 1 },
            }
        );

        foreach my $key ( keys %input ) {
            my $val = $input{$key};
            if ( $key eq "sio_file_size" ) {
                if ( ( $input{$key} !~ /\d+[kKmMgG]?$/ ) ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer optionally followed by k,m or g"
                    );
                }
            } elsif ( $key eq 'sio_num_files' ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw( "'$val' value for '$key' "
                            . "parameter must be an integer" );
                }
            } elsif ( $key eq 'iterations'
                      && ( $val > 1 )
                      && !defined( $input{interval} ) )
            {
                NATE::BaseException->throw(
                          "\'iterations\' parameter requires a"
                        . " positive integer for the \'interval\' parameter"
                );
            } elsif ( $key eq 'interval' ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute}
                );
            } elsif ( $key eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/i ) {
                    NATE::BaseException->throw( "'$val' value for '$key' "
                            . "parameter must be an integer or 'infinite'" );
                }
            }
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Volume snap restore passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Snapshot::stop_workers";
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;

    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v (@volume_objs) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined( $node_hash->{$name} ) ) {
                push( @nodes, $node );
                $node_hash->{$name} = 1;
            }
            push( @{ $node_to_volumes->{$name} }, $v );
        } ## end foreach my $v (@volume_objs)
        $num_nodes = @nodes;

        # Add snaprestore license
        NACL::STask::SystemLicense->add(
            command_interface => $nodes[0],
            feature           => 'snaprestore',
            nacltask_verify   => 1,
        );
    } else {
        # Irrelevant
    }

    my $qty_threshold = 3;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            my %vol_restore_opts = %input;
            $vol_restore_opts{_num_nodes} = $num_nodes;
            $vol_restore_opts{volumes}    = $node_to_volumes->{ $n->node() };
            my $id = 'Vol_Snap_Restore_N_DaC' . $proc_cnt;
            push( @ids, $id );

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Snapshot',
                method  => 'io_vol_restore',
                args    => [%vol_restore_opts],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $n (@nodes)

        #Collect results
        my $res = collect_worker_results(
            workers => \@workers,
            futures => $futures,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran vol_restore on " . @nodes . " nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( ( $num_volumes > $qty_threshold )
        && ( !$input{_ignore_qty_threshold} ) )
    {
        # Single node, but lots of volumes. Divide and conquer by
        # splitting work into batches, each with a reasonable
        # # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $n        = $volume_objs[0]->command_interface();
        my $name     = $n->name();

        # Throttle number of parallel processes
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt ( 2 .. $max_procs ) {
            $num_procs = $cnt;
            last if ( int( $num_volumes / $cnt ) < $qty_threshold );
            last if ($num_volumes == $num_procs);
        }

        my $vols_per_process = int($num_volumes / $num_procs);
        my $left_over = $num_volumes % $num_procs;
        $Log->comment( "Splitting vol restore into $num_procs "
                . "processes, left over = $left_over" );

        my $proc_to_vols;
        foreach my $pn (1 .. $num_procs) {
            for (1 .. $vols_per_process) {
                push(@{$proc_to_vols->{$pn}}, shift(@volume_objs));
            }
            if ($pn == $num_procs) {
                for (1 .. $left_over) {
                    push(@{$proc_to_vols->{$pn}}, shift(@volume_objs));
                }
            }
            $Log->comment("Pushed "
                    . @{$proc_to_vols->{$pn}}
                    . " volumes for proc $pn");
        }

        foreach my $pn ( 1 .. $num_procs ) {
            my $num_vols = @{$proc_to_vols->{$pn}};
            last if !$num_vols;
            my $id = 'Vol_Snap_Restore_V_DaD.'.$pn; # (V)olume_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            $input{volumes}              = $proc_to_vols->{$pn};
            $input{_ignore_qty_threshold} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Snapshot',
                method  => 'io_vol_restore',
                args    => [%input],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end for ( 1 .. $num_procs )

        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{msg}       = 'One or more procs had failures';
        } else {
            $my_return->{result}    = 1;
            my $msg = "Completed batched vol snap restore of $num_volumes "
                      ."volumes on $name";
            $Log->comment($msg);
            $my_return->{msg}       = $msg;
        }
        return $my_return;
    } else {    # Single node with qty < divide/conquer threshold
        my $flag      = delete $input{if_exists};

        my $n         = $volume_objs[0]->command_interface();
        my $node_name = $n->name();

        my $iterations     = $input{iterations};
        my $interval       = $input{interval};
        my $done           = 0;
        my $sio_duration    = 600;

        while ( ($iterations > 0 || $iterations eq 'infinite') && !$done ) {
            try {
                # Start traffic on volumes
                my @client_objs = @{ $input{client_command_interface} };
                my $client = $client_objs[0];
                my $end = $input{sio_file_size};
                my $futures;
                my @workers;

                # Convert end from k|m|g to bytes
                if ( $end =~ s/k//i ) {
                    $end = $end * 1024;    #k == 1024
                } elsif ( $end =~ s/m//i ) {
                    $end = $end * 1024 * 1024;    #m == 1024*1024
                } elsif ( $end =~ s/g//i ) {
                    $end = $end * 1024 * 1024 * 1024;   #g == 1024*1024*1024
                }

                my %input_sio;
                $input_sio{reads}                    = 50;
                $input_sio{start}                    = 0;
                $input_sio{end}                      = $end;
                $input_sio{secs}                     = $sio_duration;
                $input_sio{filesize}                 = $input{sio_file_size};
                $input_sio{filecount}                = $input{sio_num_files};
                $input_sio{volumes}                  = \@volume_objs;
                $input_sio{client_command_interface} = $input{client_command_interface};

                $Log->comment("Start IO on each volume");
                my $sio_id = 'Vol_Restore_IO_DaC_sio';
                my $sio_worker = NATE::PerlRPC::Worker::Process->new(runid => $sio_id);
                push(@workers,      $sio_worker);
                push(@Pkg_Workers,  $sio_worker );

                my $sio_future = $sio_worker->submit(
                    package => 'NX::IO',
                    method  => 'sio',
                    args    => [%input_sio],
                );
                $futures->{$sio_id}->{obj} = $sio_future;

                # Create snapshot halfway into SIO run time
                sleep $sio_duration / 2;

                my $base_name = 'io_restore_'.time();
                my %input_snapshot;
                $input_snapshot{qty}        = 1;
                $input_snapshot{base_name}  = $base_name;
                $input_snapshot{volumes}    = \@volume_objs;

                $Log->comment("Create snapshot on each volume");
                my $snapshot_id = 'Vol_Restore_Create_Snapshot_DaC';
                my $snapshot_worker = NATE::PerlRPC::Worker::Process->new(runid => $snapshot_id);
                push(@workers,      $snapshot_worker);
                push(@Pkg_Workers,  $snapshot_worker );

                my $snapshot_future = $snapshot_worker->submit(
                    package => 'NX::Snapshot',
                    method  => 'create',
                    args    => [%input_snapshot],
                );
                $futures->{$snapshot_id}->{obj} = $snapshot_future;

                my $res = collect_worker_results(
                    workers       => \@workers,
                    futures       => $futures,
                    merge_context => 0,
                );

                my $had_failure = 0;
                foreach my $w (@workers) {
                    my $rid = $w->runid();
                    if ($res->{$rid}->{exception}) {
                        my $e = $res->{$rid}->{exception};
                        $Log->comment(
                            "Found exception for runid $rid: "
                                . $e->text());
                        $e->throw() if !$::IGNORE_ERRORS;
                    } elsif ( $res->{$rid}->{result} != 1 ) {
                        my $msg = $res->{$rid}->{msg};
                        $Log->comment(
                            "Runid $rid stopped with error message: $msg!");
                        NATE::BaseException->throw("$rid failed with: $msg")
                            if !$::IGNORE_ERRORS;
                    } else {
                        $Log->comment("$rid ran successfully");
                    }
                } ## end foreach my $w (@workers)

                # Perform volume restore
                foreach my $v (@volume_objs) {
                    my $snapshot_obj = NACL::STask::VolumeSnapshot->find(
                        command_interface => $n,
                        filter            => {
                            volume        => $v->volume(),
                            snapshot      => $base_name.'*',
                        },
                    );
                    $snapshot_obj->restore('method-timeout' => 300);
                    sleep 5;
                    $snapshot_obj->purge( nacltask_retry_sleep_time => 30 );
                }

                if ( $iterations ne 'infinite' ) {
                    $iterations--;
                }

                if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $my_return->{start_tick},
                    );
                }
            } ## end try
            otherwise {
                my $e = shift;
                if ($::IGNORE_ERRORS) {
                    my $msg = $e->text();
                    $Log->warn("Hit exception during io vol restore action: $msg");

                    if ( $iterations ne 'infinite' ) {
                        $iterations--;
                    }

                    if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                        wait_for_interval(
                            interval     => $input{interval},
                            start_minute => $input{start_minute},
                            start_tick   => $my_return->{start_tick},
                        );
                    }
                } else {
                    $Log->error("Caught exception during io vol restore. "
                                ."Error was: ". $e->text() );
                    $exception = $e;
                    $done      = 1;
                }
            };
        }

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Finished io_vol_restore of $num_volumes volumes on $node_name";
            $Log->comment($msg);
            $my_return->{msg}    = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( !$input{_num_nodes...
}    # End io_vol_restore


# Returns hash reference with keys:
# 'paths' -> $vol_name -> $path
# 'mounts' -> [ mount_objects ]
sub _mount_client_volumes {
    my %input = @_;
    my $client = $input{client};
    my $client_name = $client->name();
    my @volumes = @{$input{volumes}};
    my $my_return;

    foreach my $vol_obj ( @volumes ) {
        # Mount volume on client
        my $vol_name = $vol_obj->volume();

        $Log->comment("Mounting volume $vol_name on $client_name");
        $vol_obj->mount( 'junction-path' => "/$vol_name" );
        my $mount = NACL::STask::Client::Mount->create(
            command_interface => $client,
            volume            => $vol_obj,
            chmod             => "777",
        );
        my $path = $mount->mount_point();

        $my_return->{paths}->{$vol_name} = $path;
        push ( @{$my_return->{mounts}}, $mount );
    }
    return $my_return;
} # End _mount_client_volumes

sub _run_cmd_get_pid {
    my (%input) = @_;
    my $api = $input{api_set};
    my $proc_name = $input{proc_name};

    my @pids_before = _get_proc_pids(
        api_set     => $api,
        proc_name   => $proc_name,
    );

    # Run command
    $api->execute_raw_command( command => $input{cmd} );
    sleep 5; # Allow some time for PID to register in "ps" output

    my @pids_after = _get_proc_pids(
        api_set     => $api,
        proc_name   => $proc_name,
    );

    my $pid;
    foreach my $p (@pids_after) {
        if (!grep{$_ eq $p} @pids_before) {
            $pid = $p;
        }
    }
    return $pid;
}


sub _get_proc_pids {
    my (%input) = @_;
    my @pids;
    my $api = $input{api_set};
    my $proc_name = $input{proc_name};

    my $raw = $api->execute_raw_command( command => "ps -ef|grep $proc_name" );
    my @lines = split(/\n/, $raw);
    foreach my $l (@lines) {
        if (($l =~ /^\S+\s+(\d+)\s+\d+\s+\d+/) && ($l !~ /grep/)) {
            push(@pids, $1);
        }
    }

    return @pids;
}

sub _wait_till_all_sio_complete {

    my ($api, $active_sio_pids) = @_;

    my $done = 0;
    my @completed_sio_pids;
    my $num_done = 0;
    # Wait loop to check if all sio threads are finished.
    while (!$done) {
        # Get list of currently active sio pids
        my @cur_active_pids = _get_proc_pids(
            api_set     => $api,
            proc_name   => 'sio',
        );

        # Find out which sio processes completed
        foreach my $a_pid (keys(%{$active_sio_pids})) {
            if ( !grep {$_ eq $a_pid} @cur_active_pids ) {
                push(@completed_sio_pids, $a_pid);
            }
        }

        # Process each completed sio process
        sleep 5; # Allow cool down for tool and time for
                 # output to fully redirect to log file
        foreach my $c_pid (@completed_sio_pids) {
            delete $active_sio_pids->{$c_pid};
            $num_done++;
        }

        if (keys (%{$active_sio_pids}) == 0){ #no more entries left in active_sio_pids.
            $done = 1;
            last;
        }
        sleep 30;
    }
}

#################################################################################################
1;
