use strict;
package NX::Support;

use Tharn;
use NACL::STask::Node;
use NACL::STask::Aggregate;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);
use Cwd;
use Data::Dumper;
use NX::Helper;
use Time::Local;
use NACL::MTask::SCP;
use NACL::C::Client::Ping;
use MIME::Lite;

our @ISA    = ('Exporter');
our @EXPORT = qw( run core_monitor );


my @Pkg_Workers;
sub stop_workers {
    $Log->comment("Detected interrupt signal");
    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}

=pod

=head1 NAME

NX::Support

=head1 DESCRIPTION

Support routines for use in NX framework.

=head1 METHODS

=head2 run

Executes specified script as a separate process.

The input is a hash (%input) which must include:
    client          => client host rec
    executable      => Path to file
    args            => Arguments to the executable

=over

=back

=cut
sub run {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return if $input{nx_context_inquire}; # No context requirements

    my $my_return; # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                executable        => { type => SCALAR, optional => 0 },
                args              => { type => SCALAR, optional => 1 },
                client            => { type => SCALAR, optional => 1 },
                validate_only     => { type => SCALAR, optional => 1, default => 0 },
            }
        );
        if ( !$input{executable} ) {
            NATE::BaseException->throw("An executable name must be specified");
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "run action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Support::stop_workers";
    }

    # Store start time
    $my_return->{start_tick} = time();

    my $is_thpl = 1;
    $is_thpl = 0 if ($input{executable} !~ /\.thpl$/);

    my $proc;
    my $proc_id = $input{executable};
    my @arr = split(/\//, $proc_id);
    $proc_id = $arr[-1];
    $proc_id =~ s/\.\w+$//;
    $proc_id = $proc_id . '_' . time();

    my %opts;
    $opts{workdir}   = cwd();
    $opts{client}    = $input{client} if $input{client};
    $opts{codespec}  = $input{executable};
    $opts{runid}     = $proc_id;

    if ( $input{args} ) {
        my @args_list = split(/\s+/, $input{args});
        if ( $is_thpl ) {
            my %args_hash;
            foreach my $a ( @args_list ) {
                my ($key, $val) = split(/=/, $a);
                $args_hash{$key} = $val;
            }
            $opts{params} = \%args_hash;
        } else {
            $opts{args} = \@args_list;
        }
    }

    $proc = NATE::Process->new( %opts );
    $proc->start;
    $proc->wait;

    my $parsed_proc_result;
    #if ( $is_thpl && ($input{executable} !~ /stest\.thpl/)) {
    if ( $is_thpl ) {
        my $res_obj = $proc->worst_result();
        if (!defined($res_obj) && ( $input{executable} =~ /stest\.thpl/)) {
            # An undef for stest is a good thing (sigh)
            $Log->comment("Got undef for worst_result call for stest, "
                . "interpreting this as success");
            $parsed_proc_result = 'PASS';
        } else {
            $parsed_proc_result = $res_obj->type();
        }
    } else {
        my $res = $proc->results();
        if ( $res != 0 ) { # 0 means passed
            $parsed_proc_result = 'FAIL';
        } else {
            $parsed_proc_result = 'PASS';
        }
    }

    $proc->destroy;

    $my_return->{end_tick} = time();
    if ( $parsed_proc_result ne 'PASS' ) {
        $my_return->{result} = 0;
        my $msg = "$input{executable} executable returned "
            . "$parsed_proc_result result instead of \'PASS\'";
        $Log->error( $msg );
        $my_return->{msg} = $msg;
    } else {
        $my_return->{result} = 1;
        my $msg = "Successfully ran $input{executable} executable";
        $Log->comment( $msg );
        $my_return->{msg} = $msg;
    }
    return $my_return;
}

=pod

=head2 core_monitor

Monitors the filer for cores and extracts them.

The input is a hash (%input) which must include:
    command_interface => ARRAYREF of nodes
    client_command_interface => ARRAYREF of clients

%opts can also be used to override default values, which include
    save_path => /x/eng/cores/nx_<nodename>_<timestamp>
    poll_interval => 10 (interval in minutes to check if a filer has panicked)

A global param $::CONTINUE_ON_PANIC can be set to 1, to allow for core_monitor
to run continuously until your test is complete. Otherwise, it will exit upon
the first core retrieved.

If this module is called separate of nx.thpl, create a file called .nx_poll_once
in the log directory when the rest of your test is complete. This will signal
for core_monitor to finish in the event $::CONTINUE_ON_PANIC is set to 1.

=over

=back

=cut

sub core_monitor() {
    @Pkg_Workers = ();
    my (@args) = @_;
    shift(@args) if (@args % 2);
    my %input = @args;
    my $thefile = $::LOGDIR.'/.core_monitor_found_core';

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return; #Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validate_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                command_interface        => {type => ARRAYREF, optional => 0},
                client_command_interface => {type => ARRAYREF, optional => 0},
                validate_only            => {type => SCALAR,   optional => 1, default => 0},
                save_path                => {type => SCALAR,   optional => 1},
                poll_interval            => {type => SCALAR,   optional => 1, default => 10},
            }
        );

        foreach my $key (keys %input){
            $Log->comment("key: |$key|\nvalue: |$input{$key}|\n");
            if( $key eq "poll_interval"){
                if ($input{$key} !~ /^\d+$/) {
                    NATE::BaseException->throw("Invalid input($input{$key}) for core_monitor, poll_interval must be an integer.");
                }
            }
        }
    } otherwise {
        $exception = shift;
    };
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Core_monitor passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Support::stop_workers";
    }

    my @nodes = @{$input{command_interface}};
    my $num_nodes = @nodes;

    $my_return->{start_tick} = time();

    if ( $num_nodes > 1 ) {
        my @workers;
        my $futures;
        my @ids;
        my @cores;

        foreach my $n (@nodes) {
            my $id = 'Core_Mon_DaC.'.$n->name(); #Core Monitor Divide and Conquer per node
            push(@ids, $id);

            $input{command_interface} = [ $n ];

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Support',
                method  => 'core_monitor',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
        }
        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{cores} ) {
                push( @cores, $res->{$rid}->{cores} );
            }

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Core_monitor ran successfully on ".@nodes." nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{cores} = \@cores;
        }
        return $my_return;
    } else { # Single Node
        my @node_arr      = @{$input{command_interface}};
        my @client_arr    = @{$input{client_command_interface}};
        my $node          = $node_arr[0];
        my $client        = $client_arr[0];
        my $poll_interval = $input{poll_interval};
        my $path          = $input{save_path};
        my $time          = time() - 180; #adding tolerance for early panics

        my (%cores, $saving, $node_ip);
        my $core_found = 0;
        my $nx_poll_once = 0;
        my $continue_on_panic = 0;
        my $a_core_was_extracted = 0;

        if($::CONTINUE_ON_PANIC == 1) {
            $continue_on_panic = $::CONTINUE_ON_PANIC;
        }

        if($node->mode() =~ /7mode/i) {
            $node_ip = $node->hostrec()->default_ip_7g;
        } else {
            $node_ip = $node->hostrec()->default_ip_gx;
        }

        my $poll_once = $::LOGDIR.'/.nx_poll_once';
        my $core_extracted = $::LOGDIR.'/.core_extracted';

        try {
            while(1) {
                $Log->comment("**Entering core polling loop**");
                my $alive = 0;
                my $t_elapsed = 0;

                # If another process/action has exited, .nx_poll_once will be created
                $Log->comment("Checking for $poll_once");
                if(-e($poll_once)) {
                    $Log->comment("NX exited in another area, exit after checking for core.");
                    $nx_poll_once = 1;
                }

                $Log->comment("Checking if any partner nodes exited");
                if(-e($core_extracted)) {
                    $a_core_was_extracted = 1;
                }

                # Poll for up to 2 hours, to allow for filer to come back up after panic
                while(!$alive && ($t_elapsed < 60)) {
                    try {
                        my $res = NACL::C::Client::Ping->ping(
                            command_interface => $client,
                            destination => $node_ip,
                            count => 2,
                        );
                        $alive = 1;
                    }
                    catch NACL::APISet::Exceptions::CommandFailedException with {
                        $Log->comment("Node still down: sleeping for 2 minutes.");
                        $t_elapsed++;
                        sleep 120;
                    };
                }

                # Throw exception and exit because node is still down, timeout reached
                if(!$alive) {
                    NATE::BaseException->throw("Node: $node->name is still down. Timeout reached, core_monitor exiting.");
                }

                # Node is still alive, refresh connection in case it had rebooted
                # Leave refresh_command_interface empty, defaults to checking if version changed
                $node->refresh_command_interface();
                my $api = $node->get_7m_or_nodescope_apiset();
                do {
                    $saving = $api->savecore('display-progress' => 1)->get_parsed_output();
                    foreach my $cs (@{$saving}) {
                        $Log->comment("$cs->{'core_name'} is saving ($cs->{'saved_percent'}\%).");
                        `touch $thefile`;
                        sleep 120; #Check every 2 minutes
                    }
                } while( @{$saving} > 0 );

                ## Start the copy process for ones that aren't copied out
                my $list = $api->savecore('list' => 1)->get_parsed_output();
                foreach my $c (@{$list}) {
                    my $c_name  = $c->{'core_name'};
                    my $c_panic = $c->{'panic_string'};
                    my $c_time  = $c->{'panic_time'};
                    my $c_unix_t = _parse_time($c_name);

                    $Log->comment("Core Name: $c_name\nCore Time: $c_time\nCore Unix Time: $c_unix_t\nCurrent Unix Time: $time");

                    if($time > $c_unix_t) {
                        $Log->comment("***Skipping over old core: $c_name");
                        next;
                    }
                    if($cores{$c_name}{'state'} eq 'complete') {
                        $Log->comment("Core already copied: $c_name");
                        next;
                    }

                    my $f = `touch $thefile`;  #Core has been found, let NX know
                    if(-e($thefile)) {
                        $Log->warn("Unable to create touch file: $f");
                    }

                    my %hash = (
                            node      => $node,
                            node_name => $node->name(),
                            client    => $client,
                            save_path => $path,
                            core      => $c_name,
                            time      => $time,
                    );

                    my $core_path = _extract_core(%hash);
                    $hash{save_path} = $core_path;
                    _verify_core(%hash);

                    $cores{$c_name}{'core_path'} = $core_path;
                    $cores{$c_name}{'panic_string'} = $c_panic;
                    $cores{$c_name}{'panic_time'}   = $c_time;
                    $cores{$c_name}{'state'} = 'complete';
                    $cores{$c_name}{'node'} = $node->name();
                    $core_found = 1;

                    if($::EMAIL_IDS) {
                        my %panic_email = (
                            'node'  => $node->name(),
                            'path'  => $core_path,
                            'core'  => $c_name,
                            'time'  => $c_time,
                            'panic' => $c_panic,
                        );
                        _send_panic_email(%panic_email);
                    }
                    `touch $core_extracted`;
                    if(!-e($core_extracted)) {
                        $Log->warn("Unable to create touch file: $core_extracted");
                    }
                }
                # Leaving the loop exit criteria at the bottom, so sleep isn't run when core is detected
                if((($a_core_was_extracted || $core_found) && !$continue_on_panic) || $nx_poll_once) {
                    last;
                } else {
                    $Log->comment("****Sleeping $poll_interval minute(s)*****");
                    for(my $sl=1; $sl<=$poll_interval*2; $sl++) {
                        my $s_time = $sl*30;
                        sleep $s_time;
                        $Log->comment("Slept($s_time seconds), checking to wake early.");
                        if(-e($poll_once)) {
                            $Log->comment("Core monitor: Waking Early");
                            last;
                        }
                    }
                }

                if(-e($poll_once)) {
                    $Log->comment("NX exited in another area, exit after checking for core.");
                    $nx_poll_once = 1;
                }
            } #outer while loop
        } otherwise {
            $exception = shift;
        };
        $my_return->{end_tick} = time();
        if ( $exception ) {
            $my_return->{result} = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Core monitoring finished on ".$node->name();
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
            $my_return->{cores} = \%cores;
        }
        return $my_return;
    }
} ## End core_monitor


############################################################################
##################### Private supporting subroutines #######################
############################################################################
sub _parse_time() {
    my $core = shift;
    $core =~ /(\d+)-(\d+)-(\d+)\.(\d+)_(\d+)_(\d+)/;
    my $year = $1;
    my $mon  = $2-1; #timegm goes from 0-11 for months
    my $day  = $3;
    my $hr   = $4;
    my $min  = $5;
    my $sec  = $6;

    return timegm($sec,$min,$hr,$day,$mon,$year);
}

sub _extract_core() {
    my %opts      = @_;
    my $node      = $opts{node},
    my $client    = $opts{client},
    my $path      = $opts{save_path};
    my $core      = $opts{core};
    my $time      = $opts{time};
    my $node_name = $opts{node_name};


    ##Try to create savepath on client
    my $c_api = $client->apiset();
    if(!$path) {
        $path = "/x/eng/cores/nx_$node_name"."_$time";
    } else {
        $path .= "/nx_$node_name"."_$time";
    }

    my $output = $c_api->execute_raw_command( command => "mkdir $path");
    if($output =~ /permission/i) {
        NATE::BaseException->throw("Unable to create directory: $path, $output");
    }

    my $scp = NACL::MTask::SCP->new(
        source_command_interface => $node,
        target_command_interface => $client,
        target_username => 'root',
        target_password => 'sundance',
    );

    $scp->copy_from_source_to_target(
        files_to_copy  => [$core],
        set_directory_on_source => '/mroot/etc/crash',
        set_directory_on_target => $path,
        timeout => 1800,  ## 30 minute timeout
    );

    return $path;
}

sub _verify_core() {
    my %opts      = @_;
    my $node      = host($opts{node_name}),
    my $client    = $opts{client},
    my $path      = $opts{save_path};
    my $core      = $opts{core};
    my $time      = $opts{time};
    my $node_name = $opts{node_name};

    my $api = NACL::APISet->new(
            hostobj   => $node,
            category  => "Node",
            interface => "CLI",
            set       => "Systemshell",
    );
    my $filer_md5;
    my $client_md5;

    my $output = $api->execute_command(command => "cd /mroot/etc/crash; md5 $core");
    if($output =~ /=\s+(.*?)$/) {
        $filer_md5 = $1;
    } else {
        NATE::BaseException->throw("Unable to obtain md5 checksum of corefile from filer: $core");
    }

    my $c_api = $client->apiset();
    $output = $c_api->execute_raw_command(command => "cd $path;sudo chmod 777 $core;md5sum $core");
    if($output =~ /(.*?)\s+/) {
        $client_md5 = $1;
    } else {
        NATE::BaseException->throw("Unable to obtain md5 checksum of corefile from client: $core");
    }

    if($filer_md5 eq $client_md5) {
        $Log->comment("Core extraction successful, md5 checksums are same on client and filer.");
    } else {
        NATE::BaseException->throw("Checksum for $core, not the same! Client: $client_md5 Filer: $filer_md5");
    }

}

sub _send_panic_email() {
    my %opts  = @_;
    my $node  = $opts{node_name};
    my $path  = $opts{path};
    my $core  = $opts{core};
    my $time  = $opts{time};
    my $panic = $opts{panic};

    my @email = (
        "<p>\n",
        "Panic detected on $node: <br>\n",
        "- panic_string: $panic <br>\n",
        "- core_path: $path <br>\n",
        "- core_name: $core <br>\n",
        "- panic_time: $time <br>\n,"
    );
    my $smtp = '10.57.159.114';
    my $em_list = $::EMAIL_IDS;
    if ($em_list =~ /,/) {
        $em_list =~ s/,/\@netapp\.com,/g;
    } else {
        $em_list .= '@netapp.com';
    }
    $em_list =~ s/,$//;

    my $msg = MIME::Lite->new (
        Subject => "NX detected a panic on $node ($time )",
        From    => 'nx-no-reply@netapp.com',
        To      => $em_list,
        Type    => 'text/html',
        Data    => "@email",
    );

    $msg->send('smtp',$smtp, Timeout => 60);
}

1;
