#
# Copyright 2012 NetApp, Inc.
#
# Volume related actions for NX or general autmoation use.
#
#
use strict;

package NX::Volume;

use Tharn;
use NACL::STask::SystemLicense;
use NACL::STask::Node;
use NACL::STask::Aggregate;
use NACL::STask::VolumeClone;
use NACL::STask::VolumeEfficiency;
use NACL::C::VolumeEfficiencyPolicy;
use NX::Helper;
use NX::IO;
use NATE::PerlRPC::Worker::Process;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA = ('Exporter');
our @EXPORT
    = qw( create destroy efficiency move create_flex_cache create_clone modify clone_split volume_operation);

my @Pkg_Workers;

sub stop_workers {
    $Log->comment("Detected interrupt signal");

    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
} ## end sub stop_workers

=pod

=head1 NAME

NX::Volume

=head1 DESCRIPTION

Volume related routines for use in NX framework.

=head1 METHODS

=head2 create

Creates volumes on one or more nodes.  This function also implements
divide and conquer logic to split work in parallel for efficient execution.
This returns a hash reference containing standard information per
the NX action module guidelines.

The input is a hash (%opts) which must include:
    aggregates => [ $aggr ]  # Array ref of 1 or more aggr objects
    qty        => number of volumes to create per aggr
    nx_context => hash reference containing context information for offsets

=head3 opts
    
%opts can also be used to override default values, which include
    offset              => obtained from nx_context reference if not specified
    validate_only       => default 0
    size                => default 20m
    vserver             => default vs0
    if_exists           => default reuse
    type                => default RW
    autosize            => default false (true,false,off,grow, or grow_shrink)
                           NOTE: If true, then value set to 'grow_shrink'
    name_suffix         => default _v
    max-autosize        => max. vol size limit during autosizing
    autosize-increment  => vol size increment during autosizing

This method returns a hash reference containing a list of volume objects

=over

=back

=cut

=cut begin NX create
Create
description
Create volume(s)
required
qty;Number of Volumes;;Number of volumes to create per aggregate
aggregates;Volume's Aggregate;;Aggregate is specified as $N#/$A#
optional
base_name;Volume Name;;Base name of volumes or name of volume
size;Volume Size;20m;Size of volume, suffix required
vserver;Vserver;vs0;Vserver of volume
if_exists;If Exists;reuse;Defaults to 'reuse'
type;Type;RW;Type defaults to RW
autosize;Autosize;false;Autosize the volume?
name_suffix;Volume Suffix;_v;Only used if base_name is not provided
max-autosize;Max Autosize;;Max size for autosize, requires autosize=true
autosize-increment;Autosize Increment;;Increment for autosize, requires autosize=true
=end NX create
=cut

sub create {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 1;
    }

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   qty          => { type => SCALAR,   optional => 0 },
                aggregates   => { type => ARRAYREF, optional => 0 },
                nx_context   => { type => HASHREF,  optional => 0 },
                offset       => { type => SCALAR,   optional => 1 },
                aggr_offsets => { type => HASHREF,  optional => 1 },
                base_name    => { type => SCALAR,   optional => 1 },
                validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                size => { type => SCALAR, optional => 1, default => '20m' },
                vserver =>
                    { type => SCALAR, optional => 1, default => 'vs0' },
                if_exists =>
                    { type => SCALAR, optional => 1, default => 'reuse' },
                type => { type => SCALAR, optional => 1, default => 'RW' },
                autosize =>
                    { type => SCALAR, optional => 1, default => 'false' },
                name_suffix =>
                    { type => SCALAR, optional => 1, default => '_v' },
                ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
                'max-autosize'       => { type => SCALAR, optional => 1 },
                'autosize-increment' => { type => SCALAR, optional => 1 },
                num_nodes            => { type => SCALAR, optional => 1 },

              # Backwards compatibility, we'll simply ignore command_interface
              # as we can get this from aggr objects.
                command_interface => { type => ARRAYREF, optional => 1 },
            }
        );

        foreach my $arg ( keys %input ) {
            my $val = $input{$arg};
            if ( $arg =~ /^(qty|offset)$/ ) {
                if ( $val !~ /^\d+$/ ) {
                    NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                            . "parameter must be an integer" );
                }
            } elsif ( $arg eq 'size' ) {
                if ( $val !~ /^\d+(m|mb|g|gb|t|tb)$/i ) {
                    NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                            . "parameter must be an integer followed by m, mb, "
                            . "g, gb, or t, tb" );
                }
            } elsif ( $arg eq 'if_exists' ) {
                if ( $val !~ /^(reuse|purge)$/ ) {
                    NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                            . "parameter must be either \'reuse\' or \'purge\'"
                    );
                }
            } elsif ( $arg eq 'autosize' ) {
                if ( $val !~ /^(true|false|off|grow|grow_shrink)$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be either \'true\', \'false\', "
                        . "\'off\', \'grow\', or \'grow_shrink\'");
                }
            } elsif ( $arg =~ /^(max\-autosize|autosize\-increment)$/ ) {
                if ( ( $input{autosize} ne 'true' ) && $input{validate_only} )
                {
                    $Log->warn( "\'$arg\' parameter specified even though "
                            . "\'autosize\' option is false" );
                }
                if ( $val !~ /^\d+(k|m|g|t|p)b*$/i ) {
                    NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                            . "parameter must be an integer followed by either "
                            . "\'kb\', \'mb\', \'gb\', \'tb\', or \'pb\'" );
                }
            } ## end elsif ( $arg =~ /^(max\-autosize|autosize\-increment)$/)
        } ## end foreach my $arg ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Volume create action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Volume::stop_workers";
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    my $num_nodes = 0;
    my $node_hash;
    my @nodes;
    if ( !$input{num_nodes} ) {

        # Initial NX call.  Extract number of unique nodes
        my @aggrs = @{ $input{aggregates} };
        foreach my $a (@aggrs) {
            my $node_obj  = $a->command_interface();
            my $node_name = $node_obj->name();
            if ( !defined( $node_hash->{$node_name} ) ) {
                $num_nodes++;
                $node_hash->{$node_name} = 1;
                push( @nodes, $node_obj );
            }
        } ## end foreach my $a (@aggrs)
    } else {
        @nodes = @{ $input{command_interface} };
    }

    my $qty_threshold = 5;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{num_nodes} && ( $num_nodes > 1 ) ) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $cur_offset = 1;
        my @offsets;

        # Bundle aggrs to their nodes
        my $node_to_aggrs;
        my @aggrs = @{ $input{aggregates} };
        foreach my $a (@aggrs) {
            my $node_name = $a->command_interface()->name();
            push( @{ $node_to_aggrs->{$node_name} }, $a );
            $Log->comment( "Associating aggr "
                    . $a->aggregate()
                    . " to node $node_name" );
        } ## end foreach my $a (@aggrs)

        foreach my $n (@nodes) {
            if ( !$node_to_aggrs->{ $n->name() } ) {
                $Log->comment( "No aggregates associated with node "
                        . $n->name()
                        . ", skipping the node" );
                next;
            } ## end if ( !$node_to_aggrs->...
            my $id = 'VC_N_DaC.' . $proc_cnt;  # (N)ode_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{command_interface} = [$n];
            $input{num_nodes}         = $num_nodes;
            $input{aggregates}        = $node_to_aggrs->{ $n->name() };

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'create',
                args    => [%input],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $n (@nodes)

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            context       => $input{nx_context},
            merge_context => 1,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my $node_ref = $res->{$rid}->{objects}->{node};
                foreach my $node_name ( keys %{$node_ref} ) {
                    my @res_obj = @{ $node_ref->{$node_name} };
                    push( @{ $obj_ref->{node}->{$node_name} }, @res_obj );
                }
            } ## end if ( $res->{$rid}->{objects...

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Created volumes each on " . @nodes . " nodes";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
            $my_return->{context} = $input{nx_context};
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( ( $input{qty} > $qty_threshold )
        && ( !$input{ignore_qty_threshold} ) )
    {

        # Single node, but lots of volumes. Divide and conquer by
        # splitting work into batches, each with a reasonable
        # # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $n          = $nodes[0];
        my $name       = $n->name();
        my $cur_offset = $input{offset};
        my $orig_qty   = $input{qty};
        my @aggrs      = @{ $input{aggregates} };    # All aggrs on this node

        # Throttle number of parallel processes
        my $my_context;
        my $vols_per_process;
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt ( 2 .. $max_procs ) {
            $num_procs = $cnt;
            last if ( int( $input{qty} / $cnt ) < $qty_threshold );
        }

        $vols_per_process = int( $input{qty} / $num_procs );
        my $left_over = $input{qty} % $num_procs;
        $Log->comment( "Splitting volume creation into $num_procs "
                . "processes, $vols_per_process volumes each, left over "
                . " = $left_over" );

        my $aggr_to_starting_offset;
        foreach my $aggr (@aggrs) {
            my $offset;
            my $aggr_name = $aggr->aggregate();

            # Assign offset for current aggr
            if ( !defined($cur_offset) ) {
                $offset = $input{nx_context}->{nodes}->{$name}->{volume}
                    ->{$aggr_name}->{offset};
                $offset = 1 if !defined($offset);
            }
            $aggr_to_starting_offset->{$aggr_name} = $offset;

        } ## end foreach my $aggr (@aggrs)

        my $aggr_to_proc_offset;
        for ( 1 .. $num_procs ) {
            foreach my $aggr (@aggrs) {
                my $aggr_name = $aggr->aggregate();
                if ( $_ eq '1' ) {
                    $aggr_to_proc_offset->{$_}->{$aggr_name}
                        = $aggr_to_starting_offset->{$aggr_name};
                } else {
                    $aggr_to_proc_offset->{$_}->{$aggr_name}
                        = $aggr_to_proc_offset->{ $_ - 1 }->{$aggr_name}
                        + $vols_per_process;
                }
                $Log->comment( "Offset for aggr $aggr_name for proc $_ = "
                        . $aggr_to_proc_offset->{$_}->{$aggr_name} );
            } ## end foreach my $aggr (@aggrs)
        } ## end for ( 1 .. $num_procs )

        for ( 1 .. $num_procs ) {
            my $id = 'VC_V_DaC.' . $_;    # (V)olume_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{command_interface} = [$n];

            # Replace offset and qty with batch specific values
            $input{qty} = $vols_per_process;
            if ( $proc_cnt == $num_procs ) {
                $input{qty} += $left_over;
                $Log->comment(
                    "Adding $left_over left over to proc $proc_cnt");
            }
            $input{aggr_offsets}         = $aggr_to_proc_offset->{$_};
            $input{ignore_qty_threshold} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'create',
                args    => [%input],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end for ( 1 .. $num_procs )

        ## Collect results. Don't merge volume offset context, we'll do it
        ## because only we know what the final offset for the node
        foreach my $aggr (@aggrs) {
            my $aggr_name = $aggr->aggregate();
            $my_context->{nodes}->{$name}->{volume}->{$aggr_name}->{offset}
                = $aggr_to_starting_offset->{$aggr_name}
                + ( $vols_per_process * $num_procs )
                + $left_over;
        } ## end foreach my $aggr (@aggrs)

        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my @res_obj = @{ $res->{$rid}->{objects}->{node}->{$name} };
                push( @{ $obj_ref->{node}->{$name} }, @res_obj );
            }

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{msg}       = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Completed batched creation of $orig_qty volumes on $name";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
            $my_return->{context} = $my_context;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {    # Single node with qty < divide/conquer threshold
                # Examples of a.
        my $num_vols     = delete $input{qty};
        my $cur_offset   = delete $input{offset};
        my $aggr_offsets = delete $input{aggr_offsets};
        my $context      = delete $input{nx_context};
        delete $input{validate_only};
        delete $input{ignore_qty_threshold};
        delete $input{num_nodes};
        my $suffix    = delete $input{name_suffix};
        my $base_name = delete $input{base_name};

        # Example of b.
        $input{nacltask_if_exists} = delete $input{if_exists};

        my $n         = $nodes[0];
        my $node_name = $n->name();

        # Determine correct autosize param name based on release
        my $api = $n->get_7m_or_nodescope_apiset();
        my $ver_str = $api->version()->get_raw_output();
        $ver_str =~ s/\s+$//;
        $ver_str =~ s/^\s+//;
        my $is_rr = 0;

        if (($ver_str =~ /8\.1/) || ($ver_str =~ /rollingrock/i)) {
            $Log->comment("Node is running RR: $ver_str");
            $is_rr = 1;
        } else {
            $Log->comment("Node is running non-RR: $ver_str");
        }

        my $obj_ref;
        my $my_context;
        my $num_aggrs;
        try {
            $input{command_interface} = $n;
            my @aggrs = @{ $input{aggregates} };
            $num_aggrs = @aggrs;
            delete $input{aggregates};
            my $as_arg = $input{autosize};
            foreach my $aggr_obj (@aggrs) {

                # Determine offset from context if offset param was specified
                my $aggr_name = $aggr_obj->aggregate();
                if ( !$aggr_offsets ) {
                    if ( !defined($cur_offset) ) {
                        $cur_offset
                            = $context->{nodes}->{$node_name}->{volume}
                            ->{$aggr_name}->{offset};
                        $cur_offset = 1 if !defined($cur_offset);
                    } ## end if ( !defined($cur_offset...
                } else {    # Must be call from volume DnC
                    $cur_offset = $aggr_offsets->{$aggr_name};
                }

                $input{aggregate} = $aggr_name;
                # Map autosize arg to autosize-mode on non-RR releases
                if ($is_rr) {
                    if ($input{autosize} ne 'true') {
                        $input{autosize} = 'false';
                    }
                } else {
                    delete $input{autosize};
                    if ($as_arg eq 'true') {
                        $input{'autosize-mode'} = 'grow_shrink';
                    } elsif (($as_arg eq 'false') || ($as_arg eq 'off')) {
                        $input{'autosize-mode'} = 'off';
                    } else {
                        $input{'autosize-mode'} = $as_arg;
                    }
                }
                $Log->comment("autosize: *".$input{autosize}."*");
                $Log->comment("autosize-mode: *".$input{'autosize-mode'}."*");

                for ( 1 .. $num_vols ) {
                    if ( !$base_name ) {
                        $input{volume} = $aggr_name . $suffix . $cur_offset;
                    } else {
                        $input{volume} = $base_name . $cur_offset;
                    }
                    my $was_created = 0;
                    $input{_was_created} = \$was_created;
                    my $volume = NACL::STask::Volume->create(%input);
                    push( @{ $obj_ref->{node}->{$node_name} }, $volume );
                    $cur_offset++;

                    # Set options if specified and we re-used volume
                    if ( !$was_created ) {
                        # Modify options only if volume is not part of a snapmirror
                        # relationship, otherwise, modify is not allowed
                        my $is_src_sm = NACL::C::Snapmirror->find(
                            command_interface => $n,
                            filter            => {
                                'source-volume' => $volume->volume(),
                            },
                            allow_empty       => 1,
                        );
                    
                        my $is_dst_sm = NACL::C::Snapmirror->find(
                            command_interface => $n,
                            filter            => {
                                'destination-volume' => $volume->volume(),
                            },
                            allow_empty       => 1,
                        );

                        if (!$is_src_sm && !$is_dst_sm) {
                            $Log->comment(
                                "Volume $input{volume} was reused, setting options"
                            );
                            my %options;
                            if ($is_rr) {
                                if ($input{autosize} ne 'true') {
                                    $options{autosize} = 'false';
                                } else {
                                    $options{autosize} = 'true';
                                }
                            } else {
                                $options{'autosize-mode'} = $input{'autosize-mode'};
                                $options{'max-autosize'} = $input{'max-autosize'}
                                    if $input{'max-autosize'};
                                $options{'autosize-increment'}
                                    = $input{'autosize-increment'}
                                    if $input{'autosize-increment'};
                            }
                            $volume->modify(%options);
                        } else {
                            $Log->comment("Volume ".$volume->volume()." is "
                                . "already part of a snapmirror relationship, "
                                . "skipping vol modify");
                        }
                    } ## end if ( !$was_created )
                } ## end for ( 1 .. $num_vols )
                $my_context->{nodes}->{$node_name}->{volume}->{$aggr_name}
                    ->{offset} = $cur_offset;
            } ## end foreach my $aggr_obj (@aggrs)
        } ## end try
        otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Created $num_vols volumes on each of $num_aggrs aggrs on $node_name";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
            $my_return->{context} = $my_context;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( !$input{num_nodes...
}    # End create

=pod

=head2 destroy

Destroy volumes on one or more nodes.  This function also implements
divide and conquer logic to split work in parallel for efficient execution.

The input is a hash (%opts) which must include:
    command_interface => [ $node ]  # Array ref of 1 or more node objects
    volumes           => [ $vol ]   # Array ref of 1 or more volume objects

=over

=back

=cut

=cut begin NX destroy
Destroy
description
Destroy volumes on one or more nodes.
required
nodes;Node(s);;Node(s) to run current action on. Specified as $N<#> e.g. $N1.
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
ignore_qty_threshold;Ignore Qty Threshold;0;Ignore the quantity threshold. Acceptable values are 0|1. Default is 0.
=end NX destroy
=cut

sub destroy {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   command_interface => { type => ARRAYREF, optional => 0 },
                volumes           => { type => ARRAYREF, optional => 0 },
                validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
            }
        );
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Volume deletion action passed validation";
        return $my_return;
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    # Examples of a.
    my @nodes    = @{ $input{command_interface} };
    my @volumes  = @{ $input{volumes} };
    my $num_vols = @volumes;

    my $qty_threshold = 5;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if (   ( @volumes > $qty_threshold )
        && ( !$input{ignore_qty_threshold} ) )
    {

        # Single node, but lots of volumes. Divide and conquer by
        # splitting work into batches, each with a reasonable
        # # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $n        = $nodes[0];
        my $name     = $n->name();
        my $num_vols = @volumes;

        # Throttle number of parallel processes
        my $my_context;
        my $vols_per_process;
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt ( 2 .. $max_procs ) {
            $num_procs = $cnt;
            last if ( int( $num_vols / $cnt ) < $qty_threshold );
        }

        $vols_per_process = int( $num_vols / $num_procs );
        my $left_over = $num_vols % $num_procs;
        $Log->comment(
                  "Splitting deletion of $num_vols volumes into $num_procs "
                . "processes, $vols_per_process volumes each" );

        my $proc_to_vols;
        for my $pn ( 1 .. $num_procs ) {
            my @vols;
            for ( 1 .. $vols_per_process ) {
                push( @{ $proc_to_vols->{$pn} }, shift(@volumes) );
            }
            if ( $pn == $num_procs ) {
                for ( 1 .. $left_over ) {
                    push( @{ $proc_to_vols->{$pn} }, shift(@volumes) );
                }
            }
            $Log->comment( "Pushed "
                    . @{ $proc_to_vols->{$pn} }
                    . " volumes for proc $pn" );
        } ## end for my $pn ( 1 .. $num_procs)

        for my $pn ( 1 .. $num_procs ) {
            my $id
                = 'VD_DaC.' . $pn;  # (V)olume(D)estroy_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{command_interface}    = [$n];
            $input{volumes}              = $proc_to_vols->{$pn};
            $input{ignore_qty_threshold} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'destroy',
                args    => [%input],
            );

            push( @workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end for my $pn ( 1 .. $num_procs)

        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{msg}       = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Completed batched deletion of $num_vols volumes on $name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {    # Single node with qty < divide/conquer threshold
                # Examples of a.
        delete $input{volumes};
        delete $input{validate_only};
        delete $input{ignore_qty_threshold};

        my $n         = $nodes[0];
        my $node_name = $n->name();
        my $unexport  = 0;
        my $api;
        if ( $n->mode() =~ /7mode/i ) {
            $unexport = 1;
            $api      = $n->get_7m_or_nodescope_apiset();
        }

        try {
            foreach my $v (@volumes) {
                if ($unexport) {
                    try {

                        # Unexport volume
                        $api->exportfs(
                            "unexport" => 1,
                            "path"     => "/vol/" . $v->volume(),
                        );
                    } ## end try
                    otherwise {

                        # Don't care if we can't unexport
                    };
                } ## end if ($unexport)

                # Don't proceed further if volume in question
                # no longer exists
                my $vol_exists = NACL::STask::Volume->find(
                    command_interface => $n,
                    filter            => { 'volume' => $v->volume() },
                    allow_empty       => 1,
                );
                if (!$vol_exists) {
                    $Log->comment("Volume ".$v->volume()." no longer "
                        . "exists, skipping purge");
                    next;
                }
                
                ## Purge any associated clones first, in reverse
                ## order of index suffix
                my @clone_vols = NACL::STask::Volume->find(
                    command_interface => $n,
                    filter            => {
                        'volume'      => $v->volume().'_clone_*',
                    },
                    allow_empty       => 1,
                );

                my @sort_order;
                my $clone_vol_hash;
                foreach my $c_v (@clone_vols) {
                    if ($c_v->volume() =~ /clone_(\d+)$/) {
                        push(@sort_order, $1);
                        $clone_vol_hash->{$1} = $c_v;
                    }
                }

                @sort_order = reverse(sort{$a <=> $b} @sort_order);
                foreach my $idx (@sort_order) {
                    $Log->comment("Purging clone volume "
                        . $clone_vol_hash->{$idx}->volume()
                        . " associated with original volume ".$v->volume());
                    NACL::STask::Volume->purge(
                        command_interface => $n,
                        volume            => $clone_vol_hash->{$idx},
                        vserver           => $clone_vol_hash->{$idx}->vserver(),
                        'method-timeout'  => 300,
                    );
                }

                ## Purge any associated flex cache vols first
                my @flex_vols = NACL::STask::Volume->find(
                    command_interface => $n,
                    filter            => {
                        'volume'      => $v->volume().'_flexcache*',
                    },
                    allow_empty       => 1,
                );

                foreach my $f_v (@flex_vols) {
                    $Log->comment("Purging flex cache volume ".$f_v->volume()
                        . " associated with origin volume ".$v->volume());
                    NACL::STask::Volume->purge(
                        command_interface => $n,
                        volume            => $f_v->volume(),
                        vserver           => $f_v->vserver(),
                        'method-timeout'  => 300,
                    );
                }

                NACL::STask::Volume->purge(
                    command_interface => $n,
                    volume            => $v->volume(),
                    vserver           => $v->vserver(),
                    'method-timeout'  => 300,
                );
            } ## end foreach my $v (@volumes)
        } ## end try
        otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Deleted $num_vols volumes on node $node_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( ( @volumes > $qty_threshold...
}    # End delete

=pod

=head2 efficiency

Volume efficiency related actions.  This function also implements
divide and conquer logic to split work in parallel for efficient execution.

The input is a hash (%opts) which must include:
    volumes           => [ $vol ]   # Array ref of 1 or more volume objects
    mode              => enable/disable/config/start/stop/undo

=head3 opts
    
%opts can also be used to override default values, which include:

If "mode" is "start":
    scan-old-data      => true/false (default is false)
    wait_til_done      => true/false (default is false)

If "mode" is "config":
    compression        => true/false (default is false)
    inline-compression => true/false (default is false)
    policy             => vol efficiency policy name
    hours              => hours value for schedule
    minutes            => minutes value for schedule

If "mode" is "undo":
    compression        => true/false (default is false)
    dedupe             => true/false (default is false)

=over

=back

=cut

=cut begin NX efficiency
Efficiency
description
Volume efficiency related actions.  This function also implements divide and conquer logic to split work in parallel for efficient execution.
required
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
mode;Mode;;Acceptable values are enable|disable|config|start|stop|undo
optional
compression;Compression;;Acceptable values are true|false, only available when mode is config or undo, default is false
dedupe;Dedupe;;Acceptable values are true|false, only available when mode is undo, default is false
policy;Policy;;Vol efficiency policy name. Only available when mode is config.
hours;Hours;;Hours to create snapshot, comma delimited integers or 'all', default 0
minutes;Minutes;;Minutes on the hour to create snapshot, comma delimited integers or 'all', default 0
wait_til_done;Wait Til Done;;Acceptable values are true|false, only available if mode is start. Default value is false.
scan-old-data;Scan Old Data;;Acceptable values are true|false, only available if mode is start. Default value is false.
inline-compression;Inline Compression;;Enable inline compression. Acceptable values are true|false, only available if mode is config. Default value is false.
=end NX efficiency
=cut

sub efficiency {
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   volumes              => { type => ARRAYREF, optional => 0 },
                mode                 => { type => SCALAR,   optional => 0 },
                compression          => { type => SCALAR,   optional => 1 },
                dedupe               => { type => SCALAR,   optional => 1 },
                'inline-compression' => { type => SCALAR,   optional => 1 },
                policy               => { type => SCALAR,   optional => 1, default => 'nx_ve_policy' },
                'scan-old-data'      => { type => SCALAR,   optional => 1 },
                hours => { type => SCALAR, optional => 1, default => 'all' },
                minutes => { type => SCALAR, optional => 1, default => 30 },
                wait_til_done => { type => SCALAR, optional => 1 },
                validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                _extracted_nodes =>
                    { type => SCALAR, optional => 1, default => 0 },
                _ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
            }
        );
        foreach my $arg ( keys %input ) {
            my $val = $input{$arg};
            if (   ( $arg eq 'mode' )
                && ( $val !~ /^(start|stop|config|enable|disable|undo)$/ ) )
            {
                NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                        . "parameter must be \'start\', \'stop\', \'config\', "
                        . "\'enable\', \'disable\', or \'undo\'" );
            } elsif (
                (   $arg
                    =~ /^(dedupe|compression|inline-compression|scan-old-data|wait_til_done)$/
                )
                && ( $val !~ /^(true|false)$/ )
                )
            {
                NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                        . "parameter must be \'true\' or \'false\'" );
            } elsif ( $arg =~ /^(hours|minutes)$/ ) {
                my @tmp = split( /,/, $input{$arg} );
                foreach my $v (@tmp) {
                    if ( ( $v !~ /^\d+$/ ) && ( $v ne 'all' ) ) {
                        NATE::BaseException->throw( "\'$arg\' value must be "
                                . "comma delimited integers or \'all\'" );
                    }
                } ## end foreach my $v (@tmp)
            } ## end elsif ( $arg =~ /^(hours|minutes)$/)
        } ## end foreach my $arg ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Dedupe $input{mode} action passed validation";
        return $my_return;
    }

    my @volumes   = @{ $input{volumes} };
    my $num_vols  = @volumes;
    my $num_nodes = 0;
    my $node_hash;
    my $node_to_vols;
    my @nodes;
    if ( !$input{_extracted_nodes} ) {

        # Initial NX call.  Extract number of unique nodes
        foreach my $v (@volumes) {
            my $node_obj  = $v->command_interface();
            my $node_name = $node_obj->name();
            if ( !defined( $node_hash->{$node_name} ) ) {
                $num_nodes++;
                $node_hash->{$node_name} = 1;
                $Log->comment( "Associating vol "
                        . $v->volume()
                        . " to node $node_name" );
                push( @nodes, $node_obj );
            } ## end if ( !defined( $node_hash...
            push( @{ $node_to_vols->{$node_name} }, $v );
        } ## end foreach my $v (@volumes)
    } ## end if ( !$input{_extracted_nodes...

    my $qty_threshold = 5;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if (!$input{_extracted_nodes}) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        # Figure out which nodes belong to cluster
        my $cluster_members;
        my $cmode_node_handled;
        foreach my $n (@nodes) {
            if ( $n->mode() =~ /cmode/i ) {
                my $cur_name = $n->name();
                my @members
                    = NACL::C::Cluster->nodes( command_interface => $n );
                foreach (@members) {
                    push( @{ $cluster_members->{$cur_name} }, $_ )
                        if ( $_ ne $cur_name );
                }
            } ## end if ( $n->mode() =~ /cmode/i)
        } ## end foreach my $n (@nodes)

        foreach my $n (@nodes) {
            my $node_name = $n->name();
            if ( !$node_to_vols->{$node_name} ) {
                $Log->comment( "No volumes associated with node $node_name "
                        . ", skipping the node" );
                next;
            }
            my $id
                = 'VEff_N_DaC.' . $proc_cnt;   # (N)ode_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            $input{_extracted_nodes} = 1;
            $input{volumes}          = $node_to_vols->{$node_name};

            # If cmode, create schedule and policy as these are cluster wide
            #if ( ( $input{mode} eq 'config' ) && ( $n->mode() =~ /cmode/i ) )
            if ($n->mode() =~ /cmode/i) {
                if ( grep { $cmode_node_handled->{$_} }
                    @{ $cluster_members->{$node_name} } )
                {
                    $Log->comment( "Schedule and policy already created on "
                            . " cluster for node $node_name" );
                } else {
                    my $tstamp      = time();
                    my $pol_exists  = 0;
                    my $policy_name = 'nx_ve_policy_' . $tstamp;
                    if ( $input{policy} ) {
                        $policy_name = $input{policy};
                        try {
                            # Ensure policy exists, if not, create it on CMode
                            NACL::C::VolumeEfficiencyPolicy->find(
                                command_interface => $n,
                                filter => { policy => $policy_name, }
                            );
                            $pol_exists = 1;
                            $Log->comment("VE policy $policy_name "
                                . "already exists");
                        } ## end try
                        catch NACL::Exceptions::NoElementsFound with {
                            # Policy doesn't exist
                        };
                    } else {
                        $input{policy} = $policy_name;
                    }

                    # Create schedule and use this schedule
                    my $sched_name = create_cron_sched(
                        command_interface => $n,
                        hours             => $input{hours},
                        minutes           => $input{minutes},
                    );

                    $cmode_node_handled->{$node_name} = 1;

                    # Get all VServer names
                    my @vserver;
                    my %vs_proc;
                    foreach my $v (@volumes) {
                        my $vs = $v->vserver();
                        if ( !$vs_proc{$vs} ) {
                            push( @vserver, $vs );
                            $vs_proc{$vs} = 1;
                        }
                    } ## end foreach my $v (@volumes)
                    my $num_vs = @vserver;
                    if ( !$pol_exists ) {
                        foreach my $vs (@vserver) {

                            # Create the policy with the cron job schedule
                            # previously created
                            my %opts = (
                                command_interface => $n,
                                enabled           => 'true',
                                schedule          => $sched_name,
                                vserver           => $vs,
                            );
                            if ( $num_vs > 1 ) {
                                $opts{policy} = $policy_name . '_' . $vs;
                            } else {
                                $opts{policy} = $policy_name;
                            }
                            NACL::C::VolumeEfficiencyPolicy->create(%opts);
                        } ## end foreach my $vs (@vserver)
                    } else {

                        # Get all VServer names
                        foreach my $vs (@vserver) {
                            my %opts = (
                                command_interface => $n,
                                schedule          => $sched_name,
                                vserver           => $vs,
                            );
                            if ( $num_vs > 1 ) {
                                $opts{policy} = $policy_name . '_' . $vs;
                            } else {
                                $opts{policy} = $policy_name;
                            }
                            NACL::C::VolumeEfficiencyPolicy->modify(%opts);
                        } ## end foreach my $vs (@vserver)
                    } ## end else [ if ( !$pol_exists )
                } ## end else [ if ( grep { $cmode_node_handled...
            } ## end if ( ( $input{mode} eq...

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'efficiency',
                args    => [%input],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $n (@nodes)

        # Collect results
        my $res = collect_worker_results(
            workers => \@workers,
            futures => $futures,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();
            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Success at vol efficiency $input{mode} on " 
                . @nodes
                . " nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( ( $num_vols > $qty_threshold )
        && ( !$input{_ignore_qty_threshold} ) )
    {
        # Single node, but lots of volumes. Divide and conquer by
        # splitting work into batches, each with a reasonable
        # # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $n    = $volumes[0]->command_interface();
        my $name = $n->name();

        # Throttle number of parallel processes
        my $my_context;
        my $vols_per_process;
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt ( 2 .. $max_procs ) {
            $num_procs = $cnt;
            last if ( int( $num_vols / $cnt ) <= $qty_threshold );
        }

        $vols_per_process = int( $num_vols / $num_procs );
        my $left_over = $num_vols % $num_procs;
        $Log->comment(
                  "Splitting deletion of $num_vols volumes into $num_procs "
                . "processes, $vols_per_process volumes each" );

        my $proc_to_vols;
        for my $pn ( 1 .. $num_procs ) {
            my @vols;
            for ( 1 .. $vols_per_process ) {
                push( @{ $proc_to_vols->{$pn} }, shift(@volumes) );
            }
            if ( $pn == $num_procs ) {
                for ( 1 .. $left_over ) {
                    push( @{ $proc_to_vols->{$pn} }, shift(@volumes) );
                }
            }
            $Log->comment( "Pushed "
                    . @{ $proc_to_vols->{$pn} }
                    . " volumes for proc $pn" );
        } ## end for my $pn ( 1 .. $num_procs)

        for my $pn ( 1 .. $num_procs ) {
            my $id = 'VEff_V_DaC.' . $pn;
            push( @ids, $id );

            $input{volumes}               = $proc_to_vols->{$pn};
            $input{_ignore_qty_threshold} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'efficiency',
                args    => [%input],
            );

            push( @workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end for my $pn ( 1 .. $num_procs)

        my $res = collect_worker_results(
            workers => \@workers,
            futures => $futures,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{msg}       = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Completed batched vol efficiency $input{mode} of "
                . "$num_vols volumes on $name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {    # Single node with qty < divide/conquer threshold
        my $mode = delete $input{mode};

        my $n         = $volumes[0]->command_interface();
        my $node_name = $n->name();

        # Determine if we need to set efficiency policy param for
        # each volume in FS+
        my $api = $n->get_7m_or_nodescope_apiset();
        my $ver_str = $api->version()->get_raw_output();
        $ver_str =~ s/\s+$//;
        $ver_str =~ s/^\s+//;
        my $set_e_pol = 0;

        if (($ver_str !~ /8\.1|8\.2|8\_1|8\_2/) && ($ver_str !~ /rollingrock|sierranevada|rr\d|sn\d/i)) {
            $Log->comment("If necessary, will set volume efficiency policy, "
                . "because node is running post RR/SN: $ver_str");
            $set_e_pol = 1;
        } else {
            $Log->comment("Node is running RR or SN: $ver_str");
        }

        try {
            foreach my $v (@volumes) {
                # Is volume already SIS enabled?
                my $on = $v->state()->is_sis_volume();
                if ($on =~ /^true|on$/i) {
                    $on = 1;
                } else {
                    $on = 0;
                }

                my %opts = (
                    command_interface => $n,
                    volume            => $v->volume(),
                );
                $opts{vserver} = $v->vserver() if $v->vserver();
                my $vol_eff_obj = NACL::STask::VolumeEfficiency->new( %opts );

                if ($set_e_pol && !$on && ($mode =~ /start|enable|config/)) {
                    $v->modify( 'efficiency-policy' => $input{policy} );
                }

                if ( $mode eq 'start' ) {
                    NACL::STask::VolumeEfficiency->on(
                        nacltask_if_enabled => 'pass',
                        %opts,
                    ) if !$on;

                    my @params = qw( scan-old-data );
                    foreach my $p (@params) {
                        $opts{$p} = $input{$p} if $input{$p};
                    }
                    $opts{force}         = 'true';
                    $opts{dedupe}        = 'true';
                    $opts{nacltask_wait} = 0
                        if ( $input{wait_til_done} ne 'true' );

                    try {
                        NACL::STask::VolumeEfficiency->start(%opts);
                    }
                    catch NACL::APISet::Exceptions::CommandFailedException
                        with {
                        my $e   = shift;
                        my $err = $e->text();
                        if ($err =~ /currently active/i) {
                            # Already started, error is okay
                            $Log->comment("SIS operation already in progress");
                        } elsif ($err =~ /pending/i) {
                            # One already pending, error is okay
                            $Log->comment("SIS operation currently pending");
                        } else {
                            $e->throw();
                        }
                    };
                } elsif ( $mode eq 'enable' ) {
                    NACL::STask::VolumeEfficiency->on(
                        nacltask_if_enabled => 'pass',
                        %opts,
                    ) if !$on;
                } elsif ( $mode eq 'disable' ) {
                    NACL::STask::VolumeEfficiency->off(
                        ignore_disabled => 1,
                        %opts,
                    );
                } elsif ( $mode eq 'stop' ) {
                    try {
                        NACL::STask::VolumeEfficiency->stop(%opts);
                    }
                    catch NACL::APISet::Exceptions::CommandFailedException
                        with {
                        my $e   = shift;
                        my $err = $e->text();
                        if ( $err =~ /currently idle/i ) {
                            # No active ops, error is okay
                        } elsif ($err =~ /not enabled/i) {
                            # Efficiency not turned on, error is okay
                            $Log->comment("Volume ".$v->volume()." does not "
                                . "have efficiency turned on");
                        } else {
                            $e->throw();
                        }
                    };
                } elsif ($mode eq 'undo') {
                    # Stop any operation in progress
                    my $idle = 0;
                    while (!$idle) {
                        if ($vol_eff_obj->state()->op_status() !~ /idle/i) {
                            try {
                                NACL::STask::VolumeEfficiency->stop(%opts);
                            }
                            catch NACL::APISet::Exceptions::CommandFailedException
                                with {
                                my $e   = shift;
                                my $err = $e->text();
                                if ( $err =~ /currently idle/i ) {
                                    # No active ops, error is okay
                                } elsif ($err =~ /not enabled/i) {
                                    # Efficiency not turned on, error is okay
                                    $Log->comment("Volume ".$v->volume()." does not "
                                        . "have efficiency turned on");
                                } else {
                                    $e->throw();
                                }
                            };
                            sleep 5;
                        } else {
                            $idle = 1;
                        }
                    }

                    my @params = qw( compression dedupe );
                    foreach my $p (@params) {
                        $opts{$p} = $input{$p} if $input{$p};
                    }
                    NACL::STask::VolumeEfficiency->undo(%opts);
                } else {    # must be config
                    my @params = qw( compression inline-compression );
                    foreach my $p (@params) {
                        $opts{$p} = $input{$p} if $input{$p};
                    }
                    NACL::STask::VolumeEfficiency->config(%opts);
                } ## end else [ if ( $mode eq 'start' )
            } ## end foreach my $v (@volumes)
        } ## end try
        otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Performed vol efficiency $mode on $num_vols volumes on "
                . "node $node_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( !$input{_extracted_nodes...
}    # End efficiency

=head2 move

Moves volumes from one aggregate to another. This function also implements
the ability to move them back and forth for an N amount of times.
This returns a hash reference containing standard information per
the NX action module guidelines.

The input is a hash (%opts) which must include:
    volumes           => [ $vol ]          # Array ref of 1 or more volume objects
    dst_aggregate     => [ $dst_aggregate] # Array ref of 1 aggregate object

=head3 opts

%opts can also be used to override default values, which include
    interval        => Scalar containing an integer in seconds
    iterations      => Scalar containing an integer
    timeout         => Scalar containing an integer in seconds
    start_minute    => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
    abort           => Set to 1 to abort volume move operation on every other
                       volume. These volumes will essentially never move.
                       Default is 0.

=over

=back

=cut

=cut begin NX move
Volume Move
description
Move Volume(s) to new aggregates
required
volumes;Volume(s);;Volumes(s) to move to the destination aggregate. Syntax is $V<#> e.g. $V1
dst_aggregate;Destination Aggregate;;Destination Aggregate the volume(s) will be moved to. Syntax is $A<#> e.g. $A2
optional
abort;Abort move;0;Aborts move attempt of every other volume on each iteration
interval;Interval;1;interval of time to wait between volume moves
iterations;Iterations;1;How many times to move the volumes between the source aggregate and the destination aggregate
timeout;Timeout;14400;How many seconds to wait before timing out the volume moves
start_minute;Start Minute;1;Required if 'interval' is odd|even|all, otherwise ignored if 'interval' is an integer value.
=end NX move
=cut

sub move {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   interval      => { type => SCALAR,   optional => 1 },
                start_minute  => { type => SCALAR,   optional => 1 },
                iterations    => { type => SCALAR,   optional => 1, default => 1 },
                volumes       => { type => ARRAYREF, optional => 0 },
                dst_aggregate => { type => ARRAYREF, optional => 0 },
                abort         => { type => SCALAR,   optional => 1, default => 0 },
                timeout       => {
                    type     => SCALAR,
                    optional => 1,
                    default  => '14400'
                },
                validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
            }
        );

        foreach my $arg ( keys %input ) {
            my $val = $input{$arg};
            if (   $arg eq 'iterations'
                && ( $val > 1 )
                && !defined( $input{interval} ) )
            {
                NATE::BaseException->throw(
                          "\'iterations\' parameter requires a"
                        . " positive integer for the \'interval\' parameter"
                );
            } elsif ( $arg eq 'interval' ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute}
                );
            } elsif ( $arg eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/i ) {
                    NATE::BaseException->throw( "'$val' value for '$arg' "
                            . "parameter must be an integer or 'infitine'" );
                }
            } elsif ( $arg eq 'timeout' ) {
                if ( $val !~ /^\d+$/i ) {
                    NATE::BaseException->throw( "'$val' value for '$arg' "
                            . "parameter must be an integer for seconds" );
                }
            } ## end elsif ( $arg eq 'timeout')
        } ## end foreach my $arg ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Volume move action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Volume::stop_workers";
    }

    my @volumes = @{ $input{volumes} };

    #dst_aggregate is an arrayref with one aggregate in it
    my @dst_aggr        = @{ $input{dst_aggregate} };
    my $dst_aggr        = $dst_aggr[0]->aggregate();
    my $src_aggr        = $volumes[0]->aggregate();
    my %vol_to_dst_aggr = ();
    my @vol_moves_complete;
    my $vol_moves_abort;
    my $iterations = $input{iterations};
    my $interval   = $input{interval};
    my $timeout    = $input{timeout};

    # Store start tick for return later
    $my_return->{start_tick} = time();

    my $mode = $volumes[0]->command_interface()->mode();

    foreach my $v (@volumes) {
        if ( $v->aggregate() ne $src_aggr ) {
            NATE::BaseException->throw( "Volume: "
                    . $v->volume()
                    . " is on a differant aggregate: "
                    . $v->aggregate()
                    . " than the first volume. Please have all volumes"
                    . " originate from the same aggregate." );
        } ## end if ( $v->aggregate() ne...
    } ## end foreach my $v (@volumes)

    #populate the hash that contains the destination aggregate for
    #each volume
    foreach my $v (@volumes) {
        $vol_to_dst_aggr{ $v->volume() } = $dst_aggr;
    }

    try {
        while ( $iterations > 0 || $iterations eq 'infinite' ) {
            #cmode can handle multiple vol moves at a time
            my $vol_move;
            my $vol_cnt = 1;
            my $num_vols_to_abort = 0;
            foreach my $v (@volumes) {
                try {
                    $vol_move = NACL::STask::VolumeMove->start(
                        'command_interface' => $v->command_interface(),
                        'volume'            => $v->volume(),
                        'destination-aggregate' =>
                            $vol_to_dst_aggr{ $v->volume() },
                        'vserver'       => $v->vserver(),
                        'nacltask_wait' => 0,
                        'method-timeout'  => '600',
                    );
                    if ($input{abort} && (($vol_cnt % 2) == 0)) {
                        $vol_moves_abort->{$v->volume()}->{obj} = $vol_move;
                        $num_vols_to_abort++;
                    } else {
                        push(@vol_moves_complete, $vol_move);
                    }
                } ## end try
                otherwise {
                    $exception = shift;
                    if ($::IGNORE_ERRORS) {
                        $Log->comment( "Ignorning errors, but exception"
                                . "message is : "
                                . $exception->text() );
                    } else {
                        $exception->throw();
                    }
                };
                $vol_cnt++;
            } ## end foreach my $v (@volumes)

            if ($input{abort}) {
                my $done = 0;
                my $num_processed = 0;
                while (!$done) {
                    sleep 30;

                    # Update percentages
                    foreach my $vol (keys(%$vol_moves_abort)) {
                        my $cur_vm = $vol_moves_abort->{$vol}->{obj};
                        $vol_moves_abort->{$vol}->{pc} =
                            $cur_vm->state()->percent_complete();
                    }

                    # Perform aborts on those moves passed 35%, but < 90%
                    foreach my $vol (keys(%$vol_moves_abort)) {
                        my $cur_vm = $vol_moves_abort->{$vol}->{obj};
                        if (($vol_moves_abort->{$vol}->{pc} =~ /(\d+)\%/)
                            && ($1 >= 90)) {
                            # To late to abort, move it to other bucket to
                            # let it complete
                            push(@vol_moves_complete, $cur_vm);
                            delete $vol_moves_abort->{$vol};
                            $num_processed++;
                            $Log->comment("Volume $vol is >= 90% for move, "
                                . "skipping abort");
                        } elsif (($vol_moves_abort->{$vol}->{pc} =~ /(\d+)\%/)
                            && ($1 > 35)) {
                            $Log->comment("Volume move for $vol is > 35%, "
                                . "aborting");
                            $cur_vm->abort('method-timeout' => 300);
                            delete $vol_moves_abort->{$vol};
                            $num_processed++;
                        }
                    }
                    $done = 1 if ($num_processed == $num_vols_to_abort);
                }
            }

            foreach my $vm (@vol_moves_complete) {
                try {
                    $vm->wait_for_move(
                        'method-timeout' => $timeout,
                        'destination-aggregate' =>
                            $vol_to_dst_aggr{ $vm->volume() },
                    );

                    $vm->verify( 'method-timeout' => $timeout, );

                    #switch aggregates after successful move
                    if ( $vol_to_dst_aggr{ $vm->volume() } eq $dst_aggr )
                    {
                        $vol_to_dst_aggr{ $vm->volume() } = $src_aggr;
                    } else {
                        $vol_to_dst_aggr{ $vm->volume() } = $dst_aggr;
                    }
                } ## end try
                catch NACL::APISet::Exceptions::NoMatchingEntriesException
                    with {
                    $Log->comment( "Volume Move of volume: "
                            . $vm->volume()
                            . " already "
                            . "happened" );

                    #added in case volume move already happened and the
                    #above exception was thrown
                    $vm->verify( 'method-timeout' => $timeout, );

                    #switch aggregates after successful move
                    if ( $vol_to_dst_aggr{ $vm->volume() } eq $dst_aggr )
                    {
                        $vol_to_dst_aggr{ $vm->volume() } = $src_aggr;
                    } else {
                        $vol_to_dst_aggr{ $vm->volume() } = $dst_aggr;
                    }

                    } ## end with
                otherwise {
                    $exception = shift;
                    if ($::IGNORE_ERRORS) {
                        $Log->comment( "Ignoring errors, but exception"
                                . "message is : "
                                . $exception->text() );
                    } else {
                        $exception->throw();
                    }
                };
                $vol_cnt++;
            } ## end foreach my $vm (@vol_moves)

            #empty the array for next time
            undef(@vol_moves_complete);
            undef($vol_moves_abort);

            if ( $iterations ne 'infinite' ) {
                $iterations--;
            }

            if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                wait_for_interval(
                    interval     => $input{interval},
                    start_minute => $input{start_minute},
                    start_tick   => $my_return->{start_tick},
                );
            } ## end if ( ( $iterations > 0...
        } ## end while ( $iterations > 0 ||...
    } ## end try
    otherwise {
        $exception = shift;
    };

    $my_return->{end_tick} = time();
    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Volume Move completed";
        $my_return->{msg} = $msg;
    }
    return $my_return;
}    # End move

=head2 create_flex_cache

Creates flexcache volumes on any node. This function will have serial execution.
The only type allowed for flexcache volumes in ONTAP is 'DC'. Only one flexcache
volume per volume. The naming convention will be <original_vol>_flexcache.

The input is a hash (%opts) which must include:
    origin_volumes => arrayref of volume objects as the origin for the flexcache
                      (volumes must exist)

=head3 opts

%opts can also be used to override default values, which include
    size      => default 20m
    if_exists => default 'reuse'

=over

=back

=cut

=cut begin NX create_flex_cache
Create Flex Cache
description
Creates flexcache volumes on any node. Only one flexcache volume per volume and naming convention is <original_vol>_flexcache.
required
origin_volumes;Origin Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
size;Size;20m;Size of flexcache volume to create. Default is '20m'
if_exists;If Exists;reuse;Action if it exists. Default is 'reuse.'
=end NX create_flex_cache
=cut

sub create_flex_cache {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    my $my_return;    # Hash reference containing return values
    my $exception;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   origin_volumes => { type => ARRAYREF, optional => 0 },
                size => { type => SCALAR, optional => 1, default => '20m' },
                validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                if_exists =>
                    { type => SCALAR, optional => 1, default => 'reuse' },
            }
        );

        foreach my $arg ( keys %input ) {
            my $val = $input{$arg};
            if ( $arg eq 'size' ) {
                if ( $val !~ /^\d+(k|m|g|t|p)b*$/i ) {
                    NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                            . "parameter must be an integer followed by either "
                            . "\'kb\', \'mb\', \'gb\', \'tb\', or \'pb\'" );
                }
            } elsif ( $arg eq 'if_exist' ) {
                if ( $val !~ /^reuse|purge|die/i ) {
                    NATE::BaseException->throw( "\'$val\' value for \'$arg\' "
                            . "paramter must be 'reuse', 'die' or 'purge' " );
                }
            } ## end elsif ( $arg eq 'if_exist')
        } ## end foreach my $arg ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}
            = "Volume create_flex_cache action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Volume::stop_workers";
    }

    $my_return->{start_tick} = time();

    my @volumes;

    try {
        @volumes = @{ $input{origin_volumes} };
        foreach my $v (@volumes) {
            my $aggr     = $v->get_aggregate();
            my $node     = $aggr->command_interface();
            my $vol_name = $v->volume() . '_flexcache';
            NACL::STask::Volume->create(
                command_interface         => $node,
                size                      => $input{size},
                aggregate                 => $aggr,
                volume                    => $vol_name,
                if_exists                 => $input{if_exists},
                vserver                   => $v->vserver(),
                'flexcache-origin-volume' => $v->volume(),
                'junction-path'           => undef,
            );
        } ## end foreach my $v (@volumes)
    } ## end try
    otherwise {
        $exception = shift;
    };

    $my_return->{end_tick} = time();
    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Successfully created " . @volumes . " flexcache volumes.";
        $Log->comment($msg);
        $my_return->{msg} = $msg;
    } ## end else [ if ($exception)
    return $my_return;
}    # End create_flex_cache

=head2 create_clone

Creates clones on one or more nodes.  This function also implements
divide and conquer logic to split work in parallel for efficient execution.
This returns a hash reference containing standard information per
the NX action module guidelines.

=head3 opts

%opts can be used to provide additional options for create_clone
    validate_only    => <VALIDATE_ONLY option>
    volumes          => <arrayref of volume objects>
    ignore_qty_threshold
    				 => <either 1 or 0>
    num_layers       => <layer of cascade clones>
    if_exists        => <reuse or purge>
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX create_clone
Create Clone
description
Creates clones on one or more nodes.
required
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
num_layers;Num Layers;1;Integer value of layers to cascade clones, default is 1.
optional
if_exists;If Exists;;Default value is 'reuse.'
ignore_qty_threshold;Ignore Qty Threshold;0;Ignore the quantity threshold. Acceptable values are 0|1. Default is 0.
=end NX create_clone
=cut

sub create_clone {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                volumes => { type => ARRAYREF, optional => 0 },
                ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
                num_layers => { type => SCALAR, optional => 0, default => 1 },
                if_exists =>
                    { type => SCALAR, optional => 1, default => 'reuse' },
                _num_nodes => { type => SCALAR, optional => 1 },
                _license_handled =>
                    { type => SCALAR, optional => 1, default => 0 },
            }
        );

        foreach my $key ( keys %input ) {
            if ( $key eq "num_layers" ) {
                if ( defined( $input{$key} ) && ( $input{$key} !~ /^\d+$/ ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be a positive integer"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "ignore_qty_threshold" ) {
                if (   defined( $input{$key} )
                    && ( $input{$key} != 0 )
                    && ( $input{$key} != 1 ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be either 1 or 0"
                    );
                } ## end if ( defined( $input{$key...
            } elsif ( $key eq "if_exists" ) {
                if (   defined( $input{$key} )
                    && ( $input{$key} ne "reuse" )
                    && ( $input{$key} ne "purge" ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be either reuse or purge"
                    );
                } ## end if ( defined( $input{$key...
            } ## end elsif ( $key eq "if_exists")
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Create clone action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Volume::stop_workers";
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    # Add flexclone license
    my @volume_objs = @{ $input{volumes} };
    if (!$input{_license_handled}) {
        my $node = $volume_objs[0]->command_interface();
        NACL::STask::SystemLicense->add(
            command_interface => $node,
            'package'         => ['flexclone'],
            nacltask_verify   => 1,
        );
    }

    my $num_volumes = @volume_objs;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;

    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v (@volume_objs) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined( $node_hash->{$name} ) ) {
                push( @nodes, $node );
                $node_hash->{$name} = 1;
            }
            push( @{ $node_to_volumes->{$name} }, $v );
        } ## end foreach my $v (@volume_objs)
        $num_nodes = @nodes;
    } else {

        # Irrelevant
    }

    my $qty_threshold = 5;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            my %create_clone_opts = %input;
            $create_clone_opts{_license_handled} = 1;
            $create_clone_opts{_num_nodes} = $num_nodes;
            $create_clone_opts{volumes}    = $node_to_volumes->{ $n->node() };
            my $id = 'Create_Clone_N_DaC' . $proc_cnt;
            push( @ids, $id );

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'create_clone',
                args    => [%create_clone_opts],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $n (@nodes)

        #Collect results
        my $res = collect_worker_results(
            workers => \@workers,
            futures => $futures,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran create_clone on " . @nodes . " nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( ( $num_volumes > $qty_threshold )
        && ( !$input{ignore_qty_threshold} ) )
    {

        # Single node, but lots of volumes. Divide and conquer by
        # splitting work into batches, each with a reasonable
        # # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $n        = $volume_objs[0]->command_interface();
        my $name     = $n->name();
        my $orig_qty = $num_volumes;

        # Throttle number of parallel processes
        my $vols_per_process;
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt ( 2 .. $max_procs ) {
            $num_procs = $cnt;
            last if ( int( $num_volumes / $cnt ) < $qty_threshold );
        }

        $vols_per_process = int( $num_volumes / $num_procs );
        my $left_over = $num_volumes % $num_procs;
        $Log->comment( "Splitting clone creation into $num_procs "
                . "processes, $vols_per_process cascade clones each, left over "
                . " = $left_over" );

        for ( 1 .. $num_procs ) {
            my $id = 'Create_Clone_V_DaD.'
                . $_;    # (V)olume_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            $input{ignore_qty_threshold} = 1;
            $input{_license_handled} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'create_clone',
                args    => [%input],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end for ( 1 .. $num_procs )

        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        my $obj_ref;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            # Bundle objects from all the workers
            if ( $res->{$rid}->{objects} ) {
                my @res_obj = @{ $res->{$rid}->{objects}->{node}->{$name} };
                push( @{ $obj_ref->{node}->{$name} }, @res_obj );
            }

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{msg}       = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Completed batched creation of $orig_qty cascade clones on $name";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } ## end else [ if ($had_failure)
        return $my_return;
    } else {    # Single node with qty < divide/conquer threshold
                # Examples of a.
        my $num_layers = delete $input{num_layers};
        my $flag       = delete $input{if_exists};

        my $n         = $volume_objs[0]->command_interface();
        my $node_name = $n->name();

        my $obj_ref;
        my $parent_vol;
        my $clone_name;
        my $flag_exist_layer;
        my @vols = NACL::C::Volume->find( command_interface => $n );
        foreach my $volume_obj (@volume_objs) {
            try {
                $flag_exist_layer = -1;
                for ( my $i = 0; $i < $num_layers; $i++ ) {
                    $clone_name = $volume_obj->{volume} . "_clone_$i";
                    foreach my $vol (@vols) {
                        if ( $vol->volume() eq $clone_name ) {
                            $flag_exist_layer = $i;
                            last;
                        }
                    } ## end foreach my $vol (@vols)
                } ## end for ( my $i = 0; $i < $num_layers...
                if ( $flag eq "reuse" ) {
                    my $count;
                    if ( $flag_exist_layer == -1 ) {
                        $count = 0;
                    } else {
                        $count = $flag_exist_layer + 1;
                    }
                    $parent_vol = $volume_obj->{volume};
                    for ( my $i = $count; $i < $num_layers; $i++ ) {
                        $clone_name = $volume_obj->{volume} . "_clone_$i";
                        NACL::C::VolumeClone->create(
                            command_interface => $n,
                            flexclone         => $clone_name,
                            'parent-volume'   => $parent_vol,
                            vserver           => $volume_obj->vserver(),
                        );
                        $parent_vol = $clone_name;
                        sleep(10);
                    } ## end for ( my $i = $count; $i...
                } elsif ( $flag eq "purge" ) {
                    if ( $flag_exist_layer >= 0 ) {
                        for ( my $i = $flag_exist_layer; $i >= 0; $i-- ) {
                            $clone_name = $volume_obj->{volume} . "_clone_$i";
                            NACL::STask::Volume->purge(
                                command_interface => $n,
                                volume            => $clone_name,
                                vserver           => $volume_obj->vserver(),
                                'method-timeout'  => 300,
                            );
                        } ## end for ( my $i = $flag_exist_layer...
                    } ## end if ( $flag_exist_layer...
                    $parent_vol = $volume_obj->{volume};
                    for ( my $i = 0; $i < $num_layers; $i++ ) {
                        $clone_name = $volume_obj->{volume} . "_clone_$i";
                        NACL::C::VolumeClone->create(
                            command_interface => $n,
                            flexclone         => $clone_name,
                            'parent-volume'   => $parent_vol,
                            vserver           => $volume_obj->vserver(),
                        );
                        $parent_vol = $clone_name;
                        sleep(10);
                    } ## end for ( my $i = 0; $i < $num_layers...
                } ## end elsif ( $flag eq "purge" )
            } ## end try
            otherwise {
                my $e = shift;
                if ($::IGNORE_ERRORS) {
                    my $msg = $e->text();
                    $Log->warn( "Got error trying to create clone for "
                            . $volume_obj->volume() );
                } else {
                    $exception = $e;
                }
            };
            last if $exception;
        } ## end foreach my $volume_obj (@volume_objs)

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Created $num_volumes cascade clones on each of $node_name";
            $Log->comment($msg);
            $my_return->{msg}     = $msg;
            $my_return->{objects} = $obj_ref;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( !$input{_num_nodes...
}    # End create_clone

=pod

=head2 modify

Modifies specified volume attributes

The input is a hash (%opts) which must include:
    volumes    => [ $vol ]   # Array ref of 1 or more volume objects
                             # to modify
    attribute  => $attr      # Scalar - name of attribute to modify
    values     => $val       # Scalar - comma-separated list of 1 or more
                             # values to set on the attribute

=head3 opts

%opts can also be used to override default values, which include
    iterations      => <integer>|infinite (default is 1)
    interval        => odd/even/all/<value in seconds>
                       Every odd, even, or every clock hour,
                       or specified value in seconds
    start_minute    => Required if "interval" is odd|even|all, otherwise
                       ignored if "interval" is an integer value.  This is
                       a two-digit value representing the clock minute
    validate_only   => default 0

This returns a hash reference containing standard information per
the NX action module guidelines.

=over

=back

=cut

=cut begin NX modify
Modify
description
Modifies specified volume attributes
required
volumes;Volume(s);;Volume(s) to run current action on. Specified as $V<#> e.g. $V1.
attribute;Attribute;;Volume attribute name to modify.
values;Value(s);;Value(s) to modify the current attribute to. Specified as a comma delimited set of attribute values.
optional
iterations;Iterations;1;Integer value of iterations to run or 'infinite', default is 1
interval;Interval;;Acceptable values odd|even|all or integer value(in seconds)
start_minute;Start Minute;;Required if "interval" is odd|even|all, otherwise it is ignored.
=end NX modify
=cut

sub modify {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return; # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                volumes              => { type => ARRAYREF, optional => 0 },
                values               => { type => SCALAR,   optional => 0 },
                attribute            => { type => SCALAR,   optional => 0 },
                iterations           => { type => SCALAR,   optional => 1, default => 1 },
                interval             => { type => SCALAR,   optional => 1 },
                start_minute         => { type => SCALAR,   optional => 1 },
                validate_only        => { type => SCALAR,   optional => 1, default => 0 },
            }
        );

        foreach my $arg (keys %input ) {
            my $val = $input{$arg};
            if ( $arg eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/ ) {
                    NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be an integer or \'infinite\'");
                }
            } elsif ( ($arg eq "iterations") && ($val > 1) && !defined($input{interval}) ) {
                NATE::BaseException->throw("\'iterations\' parameter requires a value "
                        . "for the \'interval\' parameter");
            } elsif ( $arg eq "interval" ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg} = "Volume modify action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Volume::stop_workers";
    }

    # Store start tick for return later
    my $start_tick = time();
    $my_return->{start_tick} = $start_tick;

    my %vserver_volumes; # Group volumes by vserver
    my @volumes  = @{ $input{volumes} };
    my $num_vols = @volumes;
    my @values   = split(',',$input{values});
    my $num_vals = @values;
    my $done = 0;
    my $iter = 1;
    my @vol_names;
    foreach my $vol (@volumes) {
        push @vol_names, $vol->volume();
        push @{$vserver_volumes{$vol->vserver()}}, $vol->volume();
    }

    while (!$done) {
        try {
            $Log->comment("Starting modify iteration $iter");

            my $cur_value = $values[$iter % $num_vals];

            $Log->comment("Modifying $input{attribute} to $cur_value "
                . "in the following volumes: @vol_names");

            foreach my $vserver (keys %vserver_volumes) {
                my @vol_group = @{$vserver_volumes{$vserver}};
                NACL::STask::Volume->modify(
                    command_interface => $volumes[0]->command_interface(),
                    vserver           => $vserver,
                    volume            => join(',',@vol_group),
                    $input{attribute} => $cur_value,
                );
                $Log->comment("Modify of $input{attribute} successful on "
                    . "the following volumes in vserver $vserver: @vol_group");
            }

            if ( ($input{iterations} eq 'infinite') || ($iter < $input{iterations}) ) {
                wait_for_interval(
                    interval     => $input{interval},
                    start_minute => $input{start_minute},
                    start_tick   => $start_tick,
                );
                $iter++;
            } else {
                $done = 1;
            }
        } otherwise {
            $exception = shift;
            if ($::IGNORE_ERRORS) {
                $Log->warn("Got following exception, ignoring and "
                    . "continuing: ".$exception->text());
                if ( ($input{iterations} eq 'infinite') || ($iter < $input{iterations}) ) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $start_tick,
                    );
                    $iter++;
                } else {
                    $done = 1;
                }
                $exception = undef;
            } else {
                $done = 1;
            }
        };
    }

    $my_return->{end_tick} = time();
    if ( $exception ) {
        $my_return->{result} = 0;
        $my_return->{exception} = $exception;
    } else {
        $my_return->{result} = 1;
        my $msg = "Modified $input{attribute} on $num_vols volumes "
            . "for $iter iterations";
        $Log->comment( $msg );
        $my_return->{msg} = $msg;
    }
    return $my_return;

} # End modify

=head2 clone_split

Creates clones for the given list of volumes and splits the clones. Supports
the ability to abort the split operation.

=head3 opts

%opts can be used to provide additional options for clone_split
    validate_only            => <VALIDATE_ONLY option>
    client_command_interface => <array ref of client objects>
    volumes                  => <arrayref of volume objects>
    abort                    => <either 1 or 0>
    interval                 => odd/even/all/<value in seconds>
                                Every odd, even, or every clock hour,
                                or specified value in seconds
    iterations               => <integer>|infinite (default is 1)
    start_minute             => Required if "interval" is odd|even|all, otherwise
                                ignored if "interval" is an integer value.  This is
                                a two-digit value representing the clock minute
    if_exists                => <reuse or purge>
    sio_file_size            => Size of each siofile, defaults to 1m
If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX clone_split
Split volume clones
description
Creates clones for given volumes and splits them. Supports the ability to abort the split operation.
required
volumes;Volume(s);;Volume(s) to create clones and run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
abort;Abort;0;Abort the clone split operation while in progress. Abort will occur on every other iteration.
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
iterations;Iterations;1;Number of times to run the clone split operation on each volume
interval;Interval;1;interval of time to wait between clone split operations
start_minute;Start Minute;1;Required if 'interval' is odd|even|all, otherwise ignored if 'interval' is an integer value.
if_exists;If Exists;reuse;Defaults to 'reuse'
sio_file_size;File Size for sio;1m;Defaults to '1m'
=end NX clone_split
=cut

sub clone_split {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    if ( $input{nx_context_inquire} ) {
        return 0;
    }

    my $my_return;    # Hash reference containing return values
    my $exception;

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   validate_only =>
                    { type => SCALAR, optional => 1, default => 0 },
                client_command_interface =>
                    { type => ARRAYREF, optional => 0 },
                volumes =>
                    { type => ARRAYREF, optional => 0 },
                abort =>
                    { type => SCALAR, optional => 1, default => 0 },
                iterations =>
                    { type => SCALAR, optional => 1, default => 1 },
                interval     =>
                    { type => SCALAR, optional => 1 },
                start_minute =>
                    { type => SCALAR, optional => 1 },
                if_exists =>
                    { type => SCALAR, optional => 1, default => 'reuse' },
                sio_file_size =>
                    { type => SCALAR, optional => 1, default => '1m' },
                _ignore_qty_threshold =>
                    { type => SCALAR, optional => 1, default => 0 },
                _num_nodes =>
                    { type => SCALAR, optional => 1 },
            }
        );

        foreach my $key ( keys %input ) {
            my $val = $input{$key};
            if ( $key eq "if_exists" ) {
                if (   defined( $input{$key} )
                    && ( $input{$key} ne "reuse" )
                    && ( $input{$key} ne "purge" ) )
                {
                    NATE::BaseException->throw(
                        "Invalid input($input{$key}) value, need to be either reuse or purge"
                    );
                }
            } elsif ( $key eq "sio_file_size" ) {
                if ( ( $input{$key} !~ /\d+[kKmMgG]?$/ ) ) {
                    NATE::BaseException->throw(
                        "\'$key\' must be an integer optionally followed by k,m or g"
                    );
                }
            } elsif ( $key eq 'iterations'
                      && ( $val > 1 )
                      && !defined( $input{interval} ) )
            {
                NATE::BaseException->throw(
                          "\'iterations\' parameter requires a"
                        . " positive integer for the \'interval\' parameter"
                );
            } elsif ( $key eq 'interval' ) {
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute}
                );
            } elsif ( $key eq 'iterations' ) {
                if ( $val !~ /^(\d+|infinite)$/i ) {
                    NATE::BaseException->throw( "'$val' value for '$key' "
                            . "parameter must be an integer or 'infinite'" );
                }
            } elsif ( $key eq 'abort' ) {
                if ($val != 0 && $val != 1) {
                    NATE::BaseException->throw(
                        "Invalid input($val) value, need to be either 0 or 1"
                    );
                }
            }
        } ## end foreach my $key ( keys %input)
    } ## end try
    otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Clone split action passed validation";
        return $my_return;
    } else {
        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Volume::stop_workers";
    }

    # a. Strip out all parameters that do not relate to underlying
    #    NACL calls
    # b. Ensure any NACL-related params are properly named and
    #    translated as necessary

    my @volume_objs = @{ $input{volumes} };
    my $num_volumes = @volume_objs;
    my $num_nodes;
    my @nodes;
    my $node_to_volumes;

    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v (@volume_objs) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined( $node_hash->{$name} ) ) {
                push( @nodes, $node );
                $node_hash->{$name} = 1;
            }
            push( @{ $node_to_volumes->{$name} }, $v );
        } ## end foreach my $v (@volume_objs)
        $num_nodes = @nodes;
    } else {
        # Irrelevant
    }

    my $qty_threshold = 5;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_num_nodes} && ( $num_nodes > 1 ) ) {

        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            my %clone_split_opts = %input;
            $clone_split_opts{_num_nodes} = $num_nodes;
            $clone_split_opts{volumes}    = $node_to_volumes->{ $n->node() };
            my $id = 'Clone_Split_N_DaC' . $proc_cnt;
            push( @ids, $id );

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'clone_split',
                args    => [%clone_split_opts],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end foreach my $n (@nodes)

        #Collect results
        my $res = collect_worker_results(
            workers => \@workers,
            futures => $futures,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran clone_split on " . @nodes . " nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        } ## end else [ if ($had_failure)
        return $my_return;
    } elsif ( ( $num_volumes > $qty_threshold )
        && ( !$input{_ignore_qty_threshold} ) )
    {

        # Single node, but lots of volumes. Divide and conquer by
        # splitting work into batches, each with a reasonable
        # # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $n        = $volume_objs[0]->command_interface();
        my $name     = $n->name();

        # Throttle number of parallel processes
        my $splits_per_process;
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt ( 2 .. $max_procs ) {
            $num_procs = $cnt;
            last if ( int( $num_volumes / $cnt ) < $qty_threshold );
        }

        $splits_per_process = int( $num_volumes / $num_procs );
        my $left_over = $num_volumes % $num_procs;
        $Log->comment( "Splitting clone split into $num_procs "
                . "processes, $splits_per_process clones each, left over "
                . " = $left_over" );

        my $proc_to_vols;
        foreach my $pn (1 .. $num_procs) {
            for (1 .. $splits_per_process) {
                push(@{$proc_to_vols->{$pn}}, shift(@volume_objs));
            }
            if ($pn == $num_procs) {
                for (1 .. $left_over) {
                    push(@{$proc_to_vols->{$pn}}, shift(@volume_objs));
                }
            }
            $Log->comment("Pushed "
                    . @{$proc_to_vols->{$pn}}
                    . " volumes for proc $pn");
        }

        foreach my $pn ( 1 .. $num_procs ) {
            my $id = 'Clone_Split_V_DaD.'.$pn; # (V)olume_(D)ivide(a)nd(C)onquer
            push( @ids, $id );

            $input{volumes}              = $proc_to_vols->{$pn};
            $input{_ignore_qty_threshold} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'clone_split',
                args    => [%input],
            );

            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        } ## end for ( 1 .. $num_procs )

        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        } ## end foreach my $w (@workers)

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $my_return->{msg}       = 'One or more procs had failures';
        } else {
            $my_return->{result}    = 1;
            my $msg = "Completed batched clone split of $num_volumes "
                      ."volumes on $name";
            $Log->comment($msg);
            $my_return->{msg}       = $msg;
        }
        return $my_return;
    } else {    # Single node with qty < divide/conquer threshold
        my $flag      = delete $input{if_exists};

        my $n         = $volume_objs[0]->command_interface();
        my $node_name = $n->name();

        my $iterations     = $input{iterations};
        my $interval       = $input{interval};
        my $abort_interval = 0;
        my $done           = 0;

        while ( ($iterations > 0 || $iterations eq 'infinite') && !$done ) {
            try {
                my $clone_name;
                my $flag_clone_exist;
                my $is_clone;
                my $clone_obj;
                my @vols = NACL::STask::Volume->find( command_interface => $n );
                my @clone_vols;
                foreach my $volume_obj (@volume_objs) {
                    $flag_clone_exist = 0;

                    $clone_name = $volume_obj->volume() . "_clonesplit";
                    foreach my $vol (@vols) {
                        if ( $vol->volume() eq $clone_name ) {
                            $flag_clone_exist = 1;
                            $is_clone = $vol->get_one_state_attribute('clone-volume');
                            $clone_obj = $vol;
                            last;
                        }
                    } ## end foreach my $vol (@vols)

                    if ( $flag_clone_exist
                         && $flag eq "reuse" && $is_clone eq 'true' ) {
                        # Save the clone object to reuse
                        push @clone_vols, $clone_obj;
                    } else {
                        if ( $flag_clone_exist
                             && ($flag eq 'purge' || $is_clone eq 'false') ) {
                            # Purge if specified or if the found volume is no longer a clone
                            NACL::STask::Volume->purge(
                                command_interface => $n,
                                volume            => $clone_name,
                                vserver           => $volume_obj->vserver(),
                                'method-timeout'  => 300,
                            );
                        }

                        # Clone doesn't exist, create it
                        $clone_obj = $volume_obj->clone_create( flexclone => $clone_name );
                        push @clone_vols, $clone_obj;
                        sleep(10);
                    }
                } ## end foreach my $volume_obj (@volume_objs)

                # Start traffic on clones
                my @client_objs = @{ $input{client_command_interface} };
                my $client = $client_objs[0];

                my $end = $input{sio_file_size};

                # Convert end from k|m|g to bytes
                if ( $end =~ s/k//i ) {
                    $end = $end * 1024;    #k == 1024
                } elsif ( $end =~ s/m//i ) {
                    $end = $end * 1024 * 1024;    #m == 1024*1024
                } elsif ( $end =~ s/g//i ) {
                    $end = $end * 1024 * 1024 * 1024;   #g == 1024*1024*1024
                }

                my %input_sio;
                $input_sio{reads}                    = 0;
                $input_sio{start}                    = 0;
                $input_sio{end}                      = $end;
                $input_sio{secs}                     = 600;
                $input_sio{filesize}                 = $input{sio_file_size};
                $input_sio{volumes}                  = \@clone_vols;
                $input_sio{client_command_interface} = $input{client_command_interface};

                $Log->comment("Start IO on each volume");
                my $future;
                my $id = 'CloneSplit_IO_DaC_sio';
                my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);

                $future = $worker->submit(
                    package => 'NX::IO',
                    method  => 'sio',
                    args    => [%input_sio],
                );

                # Allow the IO to get started
                sleep 60;

                my $abort_flag = $input{abort};
                my @split_jobs;

                # Start vol clone split
                foreach my $clone_obj (@clone_vols) {
                    my %clone_to_job;
                    my $job_component;

                    NACL::STask::VolumeClone->split(
                        command_interface => $clone_obj->command_interface(),
                        flexclone         => $clone_obj->volume(),
                        vserver           => $clone_obj->vserver(),
                        foreground        => 'false',
                        job_component     => \$job_component,
                        nacltask_wait     => 0,
                    );

                    $clone_to_job{clone_obj}     = $clone_obj;
                    $clone_to_job{job_component} = $job_component;

                    push @split_jobs, \%clone_to_job;
                }

                # Abort every other time if specified
                if ($abort_flag && $abort_interval % 2 == 0) {
                    # Allow the split time to get started
                    sleep 10;

                    $Log->comment("Aborting clone splits");
                    foreach my $split_job (@split_jobs) {
                        try {
                            NACL::C::VolumeCloneSplit->stop(
                                command_interface => $split_job->{clone_obj}->command_interface(),
                                flexclone         => $split_job->{clone_obj}->volume(),
                                vserver           => $split_job->{clone_obj}->vserver(),
                            );
                        } catch NACL::APISet::Exceptions::CommandFailedException with {
                            my $e = shift;
                            # the split job finished already
                            if ($e->text() !~ /volume is not a clone/) {
                                $e->throw();
                            }
                        };
                    }

                    # Allow time for split to abort - this can take awhile
                    sleep 300;
                } else {
                    $Log->comment("Waiting for clone splits to complete");
                    foreach my $split_job (@split_jobs) {
                        NACL::STask::VolumeClone->wait_for_split(
                            command_interface => $split_job->{clone_obj}->command_interface(),
                            flexclone         => $split_job->{clone_obj}->volume(),
                            vserver           => $split_job->{clone_obj}->vserver(),
                            job_component     => $split_job->{job_component},
                            'method-timeout'  => 2400,
                        );
                    }
                }

                $abort_interval++;

                $Log->comment("Check and stop IO on all volumes");
                $id = $worker->runid();
                if ($worker->is_busy()) {
                    $worker->stop();
                } else {
                    # Check if worker stopped due to exception
                    if ($future->result()->{exception}) {
                        my $e = $future->result()->{exception};
                        $Log->comment(
                            "Found IO failure with exception for runid $id: "
                                . $e->text());
                        $e->throw() if !$::IGNORE_ERRORS;
                    } elsif (!$future->result()->{result}) {
                        my $msg = $future->result()->{msg};
                        $Log->comment(
                            "IO runid $id stopped with error message: $msg!");
                        NATE::BaseException->throw("$id failed with: $msg")
                            if !$::IGNORE_ERRORS;
                    }
                }

                if ( $iterations ne 'infinite' ) {
                    $iterations--;
                }

                if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                    wait_for_interval(
                        interval     => $input{interval},
                        start_minute => $input{start_minute},
                        start_tick   => $my_return->{start_tick},
                    );
                }
            } ## end try
            otherwise {
                my $e = shift;
                if ($::IGNORE_ERRORS) {
                    my $msg = $e->text();
                    $Log->warn("Hit exception during clone split action: $msg");

                    if ( $iterations ne 'infinite' ) {
                        $iterations--;
                    }

                    if ( ( $iterations > 0 ) || ( $iterations eq 'infinite' ) ) {
                        wait_for_interval(
                            interval     => $input{interval},
                            start_minute => $input{start_minute},
                            start_tick   => $my_return->{start_tick},
                        );
                    }
                } else {
                    $Log->error("Caught exception during clone split. "
                                ."Error was: ". $e->text() );
                    $exception = $e;
                    $done      = 1;
                }
            };
        }

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg = "Finished clone split of $num_volumes volumes on $node_name";
            $Log->comment($msg);
            $my_return->{msg}    = $msg;
        } ## end else [ if ($exception)
        return $my_return;
    } ## end else [ if ( !$input{_num_nodes...
}    # End clone_split

=pod

=head2 volume_operation

Volume operation
This sub performs volume operation required for Bento stress which includes
user specify IO, volume online/offline, and delete. This function also
implements divide and conquer logic to split work in parallel for efficient
execution.

User is required to provide IO specific params/options for desired IO.
Currently supports hammer and fops.

The input is a hash (%opts) which must include:
    client_command_interface => <array ref of client objects>
    volumes                  => <array ref of all volume objects>

=head3 opts

%opts can also be used to override default values, which include
    enable_batching => Enable batching to create more parallel process, default is 0.
    iterations      => Iteration of volume_operation test.
    io_run_time     => Number of seconds for user specify IO to run
                                before offining/deleting volumes.
    io_type         => Type of IO for test.

For other io specific options, please refer to IO.pm.

If an option does not does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX volume_operation
Volume Ops
description
Volume operations including IO, online/offline and delete functionalities.
required
clients;Client(s);;Client(s) to run current action on. Specified as $C<#> e.g. $C1
volumes;Volume(s);;Volume(s) to run current action on. Specified as $N<#>/$A<#>/$V<#> e.g. $N1/$A2/$V1 will be on first node, first volume on second aggregate created.
optional
enable_batching;Enable Batching;0;Enable batching to create more parallel processes, default is 0, set as 1 to enable
iterations;Iterations;1;Iterations to run volume operations
io_run_time;I/O Run Time;600;Run time in seconds, default is 600
io_type;I/O Type;fops;I/O tool to use, default is fops. Acceptable values: fops|hammer
mapping;Mapping;1-to-many;Mapping of operations to volumes. Acceptable values 1-to-many|1-to-1 default is 1-to-many.
ignore_qty_threshold;Ignore Qty Threshold;;Ignore the quantity threshold. Acceptable values are 0|1.
#Hammer optional
api;Api;;hammer: -api, use specified API, accepted values are win32|posix
background;Run in Background;;hammer: -bg, does not use stdin or stdout, set as 1 to enable this flag
checkinodes;Check Inodes;;hammer: -checkinodes, Check if file system reusing inodes, set as 1 to enable this flag
filercore;Filercore;;hammer: -filercore, have filer coredump upon corruption, cannot be used with nofilercore. This is the default setting.
nofilercore;No Filercore;;hammer: -nofilercore, disable default -filercore option, cannot be used with filercore.
fill;Fill;;hammer: -fill, fill disk and then keep it full, set as 1 to enable this flag
iterations;Iterations;;hammer: -iterations, integer value of interations to execute
logfile;Log File;;hammer: -logfile, specify file to use as logfile
mode;Mode;;hammer: -mode, use specified mode, acceptable values are mixed|creates|overwrites
nocleanup;No Cleanup;;hammer: -nocleanup, do not clean up upon completion, set as 1 to enable this flag
noflush;No Flush;;hammer: -noflush, do not use FlushFileBuffers(), set as 1 to enable this flag
onlyflush;Only Flush;;hammer: -onlyflush, always use FlushFileBuffers(), set as 1 to enable this flag
onlystreams;Only Streams;;hammer: use only NT stream files, set as 1 to enable this flag
nostreams;No Streams;;hammer: -nostreams, do not use NT stream files, set as 1 to enable this flag
runtime;Runtime;;hammer: -runtime, stop hammer after integer value of seconds
seed;Seed;;hammer: -seed, set initial random seed to integer value
blocksize;Block Size;;hammer: -blocksize, use blocksize of integer value in bytes
minbsize;Min Block Size;;hammer: -minbsize, set minimum block size of integer value in bytes
maxbsize;Max Block Size;;hammer: -maxbsize, set maximum block size of integer value in bytes
minfsize;Min File Size;;hammer: -minfsize, set minimum file size of integer value in bytes
maxfsize;Max File Size;;hammer: -maxfsize, set maximum file size of integer value in bytes
#FOPs optional
operations;Operations;;FOPS: -o, operation string, sequence of supported operation codes
num-of-files;Num of Files;;FOPS: -n, number of files in each directory
size-string;Size String;;FOPS: -s, Comma separated operations size string, default 1byte
height;Height;;FOPS: -H, height of directory tree, default 0
width;Width;;FOPS: -W, width of tree at each level, default NULL
use-existing-dir;Use Existing Directory;;FOPS: -e, Use existing directory tree
num-of-threads;Num of Threads;;FOPS: -t, Number of threads to use, default 0
dir-prefix;Directory Prefix;;FOPS: -P, Directory name prefix, default DIR
file-prefix;File Prefix;;FOPS: -p, File name prefix, default FILE
rename-file-prefix;Rename-File Prefix;;FOPS: -q, File prefix for renamed file name, default NEW_FILE
level;Level;;FOPS: -l, Level in the directory tree to work, default leaf level
block-size;Block Size;;FOPS: -b, Block size to use for operations, default 256k
delay-in-operation;Delay in Operation;;FOPS: -d, Delay before operating on next file, default 0
delay-between-operations;Delay Between Operations;;FOPS: -D, Delay before each file operation, default 0
expression;Expression;;FOPS: -x, Expression to write and append, default A
offset-string;Offset String;;FOPS: -f, Comma separated string of offsets, default 0
metrics-string;Metrics String;;FOPS: -M, Comma separated time ranges, default 0,3000,100
error-tolerance-percentage;Error Tolerance %;;FOPS: -X, Percentage of errors to ignore, default 10
file-open-options;File Open Options;;FOPS: -O, Open system call options, default NULL
pick-random-files;Pick Random Files;;FOPS: -r, Flag to pick random files
repeat;Repeat;;FOPS: Amount of times to repeat fops
=end NX volume_operation
=cut

sub volume_operation {
    my @args = @_;
    shift(@args) if (@args % 2);
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    my $my_return;    # Hash reference containing return values
    my $exception;

    my $num_of_client = @{$input{client_command_interface}};

    my %io_input = %input;
    delete $io_input{io_type};
    delete $io_input{iterations};
    delete $io_input{io_run_time};
    delete $io_input{enable_batching};
    delete $io_input{_extracted_nodes};
    delete $io_input{_num_nodes};
    delete $io_input{validate_only};
    delete $io_input{ignore_qty_threshold};

    my %input_fops   = %io_input;
    my %input_hammer = %io_input;
    $input_hammer{'iterations'} = $input_hammer{'hammer_iterations'};
    delete $input_hammer{'hammer_iterations'};

    # Validate parameters
    try {

        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {   client_command_interface => {type => ARRAYREF, optional => 0},
                volumes                  => {type => ARRAYREF, optional => 0},
                enable_batching          => {type => SCALAR, optional => 1, default => 0},
                iterations               => {type => SCALAR, optional => 1, default => 1},
                io_run_time              => {type => SCALAR, optional => 1, default => 600},
                io_type                  => {type => SCALAR, optional => 1, default => 'fops'},
                mapping                  => {type => SCALAR, optional => 1, default => '1-to-many'},
                validate_only            => {type => SCALAR, optional => 1, default => 0},
                _extracted_nodes         => {type => SCALAR, optional => 1, default => 0},
                _num_nodes               => {type => SCALAR, optional => 1, default => undef },
                ignore_qty_threshold     => {type => SCALAR, optional => 1, default => 0},

                # hammer only agruments
                api              => {type => SCALAR, optional => 1},
                background       => {type => SCALAR, optional => 1},
                checkinodes      => {type => SCALAR, optional => 1},
                filercore        => {type => SCALAR, optional => 1},
                nofilercore      => {type => SCALAR, optional => 1},
                fill             => {type => SCALAR, optional => 1},
                hammer_iterations => {type => SCALAR, optional => 1},
                logfile          => {type => SCALAR, optional => 1},
                mode             => {type => SCALAR, optional => 1},
                nocleanup        => {type => SCALAR, optional => 1},
                noflush          => {type => SCALAR, optional => 1},
                onlyflush        => {type => SCALAR, optional => 1},
                onlystreams      => {type => SCALAR, optional => 1},
                nostreams        => {type => SCALAR, optional => 1},
                runtime          => {type => SCALAR, optional => 1},
                seed             => {type => SCALAR, optional => 1},
                blocksize        => {type => SCALAR, optional => 1},
                minbsize         => {type => SCALAR, optional => 1},
                maxbsize         => {type => SCALAR, optional => 1},
                minfsize         => {type => SCALAR, optional => 1},
                maxfsize         => {type => SCALAR, optional => 1},
                skip_nfs_setup   => {type => SCALAR, optional => 1, default => 0},
                mapping_position => {type => SCALAR, optional => 1},

                # fops only arguments
                # Set optional to 1 to accommodate other IO options
                'operations' => {type => SCALAR, optional => 1,},

                # Set optional to 1 to accommodate other IO options
                'num-of-files'               => {type => SCALAR, optional => 1,},
                'size-string'                => {type => SCALAR, optional => 1,},
                'height'                     => {type => SCALAR, optional => 1,},
                'width'                      => {type => SCALAR, optional => 1,},
                'use-existing-dir'           => {type => SCALAR, optional => 1,},
                'num-of-threads'             => {type => SCALAR, optional => 1,},
                'dir-prefix'                 => {type => SCALAR, optional => 1,},
                'file-prefix'                => {type => SCALAR, optional => 1,},
                'rename-file-prefix'         => {type => SCALAR, optional => 1,},
                'level'                      => {type => SCALAR, optional => 1,},
                'block-size'                 => {type => SCALAR, optional => 1,},
                'delay-in-operation'         => {type => SCALAR, optional => 1,},
                'delay-between-operations'   => {type => SCALAR, optional => 1,},
                'expression'                 => {type => SCALAR, optional => 1,},
                'offset-string'              => {type => SCALAR, optional => 1,},
                'metrics-string'             => {type => SCALAR, optional => 1,},
                'error-tolerance-percentage' => {type => SCALAR, optional => 1,},
                'file-open-options'          => {type => SCALAR, optional => 1,},
                'pick-random-files'          => {type => SCALAR, optional => 1,},
                mapping_position             => {type => SCALAR, optional => 1,},
                repeat                       => {type => SCALAR, optional => 1,},
                skip_nfs_setup               => {type => SCALAR, optional => 1, default => 0},

            }
        );
        foreach my $arg (keys %input) {
            $Log->comment("key: |$arg|\nvalue: | $input{$arg}|\n");
            my $val = $input{$arg};
            if ( $arg eq 'client_command_interface' && $num_of_client ne '1' ) {
                NATE::BaseException->throw(
                    "Only one client per run is supported! Found $num_of_client client from input file. Please check your input file!"
                );
            } elsif (   $arg eq 'io_run_time'
                && $val !~ /^(\d+)$/) {
                NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be integers");
            } elsif ($arg eq 'iterations'
                && $val !~ /^(\d+)$/) {
                NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be integers");
            } elsif ($arg eq 'enable_batching'
                && $val != 0 && $val != 1) {
                NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be 1 or 0");
            } elsif ($arg eq 'io_type'
                && $val !~ /^(hammer|fops)$/) {
                NATE::BaseException->throw("\'$val\' value for \'$arg\' "
                        . "parameter must be \'hammer\' or \'fops\'");
            } elsif ($arg eq 'io_type' && $val eq 'fops') {
                my $res = NX::IO::fops(%input_fops, validate_only => 1,);
                if (!$res->{result}) {
                    $res->{exception}->throw();
                }
            } elsif ($arg eq 'io_type' && $val eq 'hammer') {
                my $res = NX::IO::hammer(%input_hammer, validate_only => 1,);
                if (!$res->{result}) {
                    $res->{exception}->throw();
                }
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ($input{validate_only}) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "volume_operation action passed validation";
        return $my_return;
    }

    my @volumes   = @{$input{volumes}};
    my $num_vols  = @volumes;
    my $num_nodes;
    my @nodes;
    my $enable_batching  = $input{enable_batching};
    my $node_to_volumes;

    # Extract number of unique nodes, if we haven not already
    # determined that from a previous call
    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $v ( @volumes ) {
            my $node = $v->command_interface();
            my $name = $node->name;
            if ( !defined($node_hash->{$name}) ) {
                push(@nodes, $node);
                $node_hash->{$name} = 1;
            }
            push(@{$node_to_volumes->{$name}}, $v);
        }
        $num_nodes = @nodes;
    } else {
        # Irrelevant
    }

    my $qty_threshold = 5;

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_extracted_nodes} && ( $num_nodes > 1 ) ) {
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n ( @nodes) {
            my $node_name = $n->name();
            my $id = 'VolOp_N_DaC.'.$proc_cnt; # (Vol)ume_(Op)eration_(N)ode_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Indicate we already parsed out unique nodes, no need
            # to do it again
            $input{_num_nodes} = $num_nodes;
            $Log->comment("VolOp_N_DaC".$proc_cnt." is running with ".
                @{$node_to_volumes->{$node_name}}. " volumes");
            $input{volumes}   = $node_to_volumes->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'volume_operation',
                args    => [ %input ],
            );

            push( @workers, $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers         => \@workers,
            futures         => $futures,
            merge_context   => 0,
        );

        my $had_failure = 0;
        foreach my $w ( @workers ) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ( $had_failure ) {
            $my_return->{result} = 0;
            $my_return->{msg} = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Volume operation completed on $num_nodes nodes";
            $Log->comment( $msg );
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } elsif (   (@volumes > $qty_threshold)
        && (!$input{ignore_qty_threshold}) && $enable_batching ) {

        # Lots of volumes. Divide and conquer by splitting work into batches,
        # each with a reasonable # of volumes
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        my $node = $volumes[0]->command_interface();
        my $node_name = $node->name();
        my $num_vols = @volumes;

        # Throttle number of parallel processes
        my $my_context;
        my $vols_per_process;
        my $num_procs;

        # Figure out how many processes needed. Keep incrementing
        # # of procs until we have less than the threshold qty of volumes.
        # Otherwise, use max # of procs regardless of the threshold.
        my $max_procs = 24;  # Max number of processes, e.g. set to some filer
                             # connection limit
        for my $cnt (2 .. $max_procs) {
            $num_procs = $cnt;
            last if (int($num_vols / $cnt) < $qty_threshold);
        }

        $vols_per_process = int($num_vols / $num_procs);
        my $left_over = $num_vols % $num_procs;
        $Log->comment(
            "Splitting volume operations of $num_vols volumes into $num_procs "
                . "processes, $vols_per_process volumes each");

        my $proc_to_vols;
        for my $pn (1 .. $num_procs) {
            my @vols;
            for (1 .. $vols_per_process) {
                push(@{$proc_to_vols->{$pn}}, shift(@volumes));
            }
            if ($pn == $num_procs) {
                for (1 .. $left_over) {
                    push(@{$proc_to_vols->{$pn}}, shift(@volumes));
                }
            }
            $Log->comment("Pushed "
                    . @{$proc_to_vols->{$pn}}
                    . " volumes for proc $pn");
        }

        for my $pn (1 .. $num_procs) {
            my $id = 'VolOp_DaC.'
                . $pn;    # (Vol)ume(Op)eration_(D)ivide(a)nd(C)onquer
            push(@ids, $id);

            # Replace command_interface with a reference to an array
            # with just the single node.  Keep all other original
            # inputs to pass in
            $input{command_interface}    = [$node];
            $input{volumes}              = $proc_to_vols->{$pn};
            $input{ignore_qty_threshold} = 1;

            my $worker = NATE::PerlRPC::Worker::Process->new(runid => $id);
            my $future = $worker->submit(
                package => 'NX::Volume',
                method  => 'volume_operation',
                args    => [%input],
            );

            push(@workers, $worker);
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ($res->{$rid}->{result} != 1) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }

        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Completed batched volume operation of $num_vols volumes on $node_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } else {    # Volume qty < divide/conquer threshold or $enable_batching = 0
        my $node = $volumes[0]->command_interface();
        my $node_name = $node->name();
        my $unexport  = 0;
        my $api;
        if ($node->mode() =~ /7mode/i) {
            $unexport = 1;
            $api      = $node->get_7m_or_nodescope_apiset();
        }

        $Log->comment("Save volume names and sizes");
        my @vol_infos;
        for my $v (@volumes) {
            my $vol_info;
            $vol_info->{name} = $v->volume();
            $vol_info->{size} = $v->state()->size();
            $vol_info->{aggr} = $v->aggregate();
            push(@vol_infos, $vol_info);
        }

        for my $iter (1 .. $input{iterations}) {
            try {
                $Log->comment("Start volume operation test iteration $iter");

                my @vols;
                $Log->comment("Re-create volumes if doesn't exit");
                for my $vol_info (@vol_infos) {
                    my $vol = NACL::STask::Volume->create(
                        command_interface  => $node,
                        volume             => $vol_info->{name},
                        size               => $vol_info->{size},
                        aggregate          => $vol_info->{aggr},
                        nacltask_if_exists => 'reuse',
                    );
                    push(@vols, $vol);
                }

                my @client_objs = @{ $input{client_command_interface} };
                my $client = $client_objs[0];

                $Log->comment("Set up NFS before mounting volumes");
                NX::IO->_setup_nfs(
                    volumes => \@vols,
                );

                $Log->comment("Mount volumes to client for IO");
                my $return = NX::IO->_mount_client_volumes(
                        client  => $client,
                        volumes => $input{volumes},
                );
                my $mounts = $return->{mounts};

                $Log->comment("Start IO on each volumes");
                my $future;
                my $id = 'VolOp_IO_DaC_'
                    . $input{'io_type'}
                    ;    # (Vol)ume(Op)eration_IO_(D)ivide(a)nd(C)onquer
                my $worker
                    = NATE::PerlRPC::Worker::Process->new(runid => $id);
                if ($input{'io_type'} eq 'fops') {
                    $input_fops{volumes} = \@vols;
                    $future = $worker->submit(
                        package => 'NX::IO',
                        method  => 'fops',
                        args    => [%input_fops],
                    );
                } elsif ($input{'io_type'} eq 'hammer') {
                    $input_hammer{volumes} = \@vols;
                    $future = $worker->submit(
                        package => 'NX::IO',
                        method  => 'hammer',
                        args    => [%input_hammer],
                    );
                }
                $Log->comment(
                    "Keep IO running for $input{io_run_time} seconds");
                sleep $input{'io_run_time'};

                $Log->comment("Check IO status before offlining volumes");
                $id = $worker->runid();
                if ($worker->is_busy()) {
                    $Log->comment("IO with runid $id is still running!");
                } else {

                    # Check if worker stopped due to exception
                    if ($future->result()->{exception}) {
                        my $e = $future->result()->{exception};
                        $Log->comment(
                            "Found IO failure with exception for runid $id: "
                                . $e->text());
                        $e->throw() if !$::IGNORE_ERRORS;

                        # Check if worker stopped due to other failures
                    } elsif (!$future->result()->{result}) {
                        my $msg = $future->result()->{msg};
                        $Log->comment(
                            "IO runid $id stopped with error message: $msg!");
                        NATE::BaseException->throw("$id failed with: $msg")
                            if !$::IGNORE_ERRORS;
                    }
                }

                $Log->comment("Offline each volume");
                for my $v (@vols) {
                    $v->unmount();
                    $v->offline();
                }

                $Log->comment("Wait for 2 minutes after offlining volumes");
                sleep 120;

                $Log->comment("Online each volume");
                for my $v (@vols) {
                    $v->online();
                }

                $Log->comment(
                    "Keep IO running for $input{io_run_time} seconds");
                sleep $input{'io_run_time'};

                $Log->comment("Check and stop IO on all volumes");
                $id = $worker->runid();
                if ($worker->is_busy()) {
                    $worker->stop();
                } else {
                    # Check if worker stopped due to exception
                    if ($future->result()->{exception}) {
                        my $e = $future->result()->{exception};
                        $Log->comment(
                            "Found IO failure with exception for runid $id: "
                                . $e->text());
                        $e->throw() if !$::IGNORE_ERRORS;
                    } elsif (!$future->result()->{result}) {
                        my $msg = $future->result()->{msg};
                        $Log->comment(
                            "IO runid $id stopped with error message: $msg!");
                        NATE::BaseException->throw("$id failed with: $msg")
                            if !$::IGNORE_ERRORS;
                    }
                }

                $Log->comment("Purge mounts after stopping IO");
                my $res = NX::IO->_purge_mounts(
                        mounts => $mounts,
                );
                if ( !$res->{result}) {
                    $Log->comment("Mount purging failed");
                    $res->{exception}->throw();
                }

                $Log->comment("Purge all volume");
                for my $v (@volumes) {
                    if ($unexport) {
                        try {
                            # Unexport volume
                            $api->exportfs(
                                "unexport" => 1,
                                "path"     => "/vol/" . $v->volume(),
                            );
                        } otherwise {
                            # Don't care if we can't unexport
                        };
                    }
                    NACL::STask::Volume->purge(
                        command_interface => $node,
                        volume            => $v->volume(),
                        vserver           => $v->vserver(),
                        'method-timeout'  => 300,
                    );
                }
            } ## end try
            otherwise {
                if (!$::IGNORE_ERRORS) {
                    $exception = shift;
                    last;
                } else {
                    my $e = shift;
                    $Log->error("Volume operation failed with $e->text()");
                }
            }; ## end otherwise
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
        } else {
            $my_return->{result} = 1;
            my $msg
                = "Completed $input{iterations} iteration of volume_operation"
                . " test with $num_vols volumes on node $node_name";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        }
        return $my_return;
    }
}    # End volume_operation

1;
