package NX::Wafliron;

use strict 'vars';
use Tharn;
use Data::Dumper;
use NX::Helper;
use NX::Volume;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::MTask::Wafliron;
use NACL::STask::Volume;
use NACL::C::StorageAggregate;
use Params::Validate qw(validate SCALAR OBJECT HASHREF ARRAYREF);

our @ISA    = ('Exporter');
our @EXPORT = qw(
  start
);

my @Pkg_Workers;

sub stop_workers {
    $Log->comment("Detected interrupt signal");

    # Stop any workers from divide/conquer logic
    _nx_stop_workers( workers => \@Pkg_Workers );
    exit;
}

=pod

=head1 NAME

NX::Wafliron

=head1 DESCRIPTION

Wafliron related routines for use in NX framework;

=head1 METHODS

=head2 start

Run Wafliron on an aggregate(s).

The input is a hash (%opts) which must include the command_interface:
    command_interface => $node

=head3 opts

The input is a hash (%opts) which must include:
    aggregates          => Arrayref of one or more aggregate objects

%opts can also be used to override default values, which include
    optional_commit    => Used for wafliron -optional commit.
                        # true|false (default is false)
    ioc_if_error_found => Requred if <optional_commit> is defined. This tag
                          determines what NX does with wafliron errors
                        # "commit" to commit fixes
                        # "reject" to reject fixes
                        # "fail" allows the action to fail
                        # use "fail" with <continue_on_failure> to keep prevent other
                          actions from being stopped
    start_minute       => Required if <interval> is odd|even|all, otherwise
                       ignored if <interval> is an integer value.  This is
                       a two-digit value representing the clock minute
    interval           => odd/even/all/<value in seconds>
                        # Every odd, even, or every clock hour,
                        # or specified value in seconds
    iterations         => integer value or <infinite>
                        # requires interval if defined
    timeout            => Timeout in seconds to wait for each wafliron to complete
                          Default: 7200 seconds
    delay_start        => Specify 1 to wait for the specified interval before
                          starting first wafliron iteration. Default is 0.

A sample call would be:

%opts = {
    aggregates          => [ $aggr_obj1 ],
    optional_commit    => 'true',
    ioc_if_error_found => 'fail',
    interval           => 60,
    iterations         => 5,
}

If an option does not pass validation, a NATE::BaseException will be thrown.

=over

=back

=cut

=cut begin NX start
Start
description
Run wafliron on specified aggregate(s).
required
aggregates;Aggregate(s);;Aggregate(s) to fail disks on. Sample syntax is $N<#>/$A<#> e.g. $N1/$A1
optional
optional_commit;Optional commit;false;Flag to set run wafliron IOC. Set value to 'true' to run wafliron IOC.
ioc_if_error_found;IOC if error found;;Specifies the subsequent action if wafliron errors are found. This parameter is required if optional_commit is set to true. Sample value would be 'commit', meaning to commit the fixes.
start_minute;Start minute;;Specifies the two digit clock minute to start the next iteration. Sample value would be 05, meaning iterations would start on 00:05, 01:05, etc.
interval;Interval;;Specifies the interval to start the next iteration. Sample value would be odd|even|all, meaning iterations would start on odd intervals, even intervals, or all intervals.
iterations;Iterations;1;Specifies how many times to execute the action. Sample value could be 5, meaning to execute this action 5 times.
timeout;Timeout;7200;Timeout in seconds to wait for each wafliron to complete.
delay_start;Delay wafliron start;0;If set to 1, then wait for the specified interval before starting first wafliron iteration.
=end NX start
=cut

sub start {
    @Pkg_Workers = ();
    my @args = @_;
    shift(@args) if ( @args % 2 );
    my %input = @args;

    return 0 if $input{nx_context_inquire};

    # Hash reference containing return values
    my $my_return;
    my $exception;
    my $interval;

    try {

        # Validate parameters
        # Call NATE validate() to ensure variable type correctness and set
        # defaults. NACL objects (command_interfaces, volumes, etc.) are
        # always passed in as an ARRAYREF, even if there is only one.
        #
        # The values in the ARRAYREF (other than command_interfaces) may not
        # actually be an object during the validation phase (i.e. when
        # validation_only => 1).
        #
        # NACL method itself will throw an exception if an object is expected,
        # therefore it is not necessary to do object validation.
        %input = validate(
            @args,
            {
                aggregates => { type => ARRAYREF, optional => 0 },
                optional_commit =>
                  { type => SCALAR, optional => 1, default => 'false', },
                ioc_if_error_found => { type => SCALAR, optional => 1, },
                iterations         => {
                    type     => SCALAR,
                    optional => 1,
                    default  => 1
                },
                timeout      => { type => SCALAR, optional => 1, default => 7200 },
                interval     => { type => SCALAR, optional => 1 },
                start_minute => { type => SCALAR, optional => 1 },
                delay_start  => { type => SCALAR, optional => 1, default => 0 },
                validate_only =>
                  { type => SCALAR, optional => 1, default => 0 },
                _num_nodes =>
                  { type => SCALAR, optional => 1, default => undef },
            }
        );

        $interval = $input{interval};
        foreach my $arg ( keys %input ) {
            if (   ( $arg eq "iterations" )
                && ( $input{$arg} > 1 )
                && !defined($interval) )
            {

                # validate that <interval> is defined if <iterations> is defined
                NATE::BaseException->throw(
"Invalid value: \'$input{$arg}\'. \'iterations\' parameter requires a value "
                      . "for the \’interval\’ parameter" );
            } elsif ( ( $arg eq 'iterations' )
                && ( ( $input{$arg} !~ /^(\d+|infinite)$/ ) || !$input{$arg} ) )
            {
                NATE::BaseException->throw(
"The \'iterations\' parameter value must be either a positive integer or \'infinite\'"
                );
            } elsif ( $arg eq "interval" ) {
                # wait_for_interval() validates start_minute
                wait_for_interval(
                    validate_only => 1,
                    interval      => $input{interval},
                    start_minute  => $input{start_minute},
                );
            } elsif ( ( $arg eq "optional_commit" )
                && ( $input{$arg} !~ m/^(true|false)$/ ) )
            {
                NATE::BaseException->throw(
                    "The \'optional_commit\' parameter value must be true|false"
                );
            } elsif ( ( $arg eq "ioc_if_error_found" )
                && ( $input{$arg} !~ m/^(commit|reject|fail)$/ ) )
            {
                NATE::BaseException->throw(
"The \'ioc_if_error_found\' parameter value must be commit|reject|fail"
                );
            }
        }
        if ( $input{optional_commit} eq 'true' ) {
            if ( !defined $input{ioc_if_error_found} ) {
                NATE::BaseException->throw(
"Missing tag: <ioc_if_error_found> must be defined if <optional_commit> is true"
                );
            }
        }
    } otherwise {
        $exception = shift;
    };

    if ($exception) {
        $my_return->{result}    = 0;
        $my_return->{exception} = $exception;
        return $my_return;
    } elsif ( $input{validate_only} ) {
        $my_return->{result} = 1;
        $my_return->{msg}    = "Wafliron start action passed validation";
        return $my_return;
    } else {

        # Proceed to real run, register interrupt handler
        $SIG{'INT'} = "NX::Wafliron::stop_workers";
        delete $input{validate_only};
    }

    my @aggrs      = @{ $input{aggregates} };
    my $iterations = $input{iterations};

    my $num_aggr = @aggrs;
    my @nodes;
    my $_num_nodes;
    my $node_to_aggrs;

    # Extract number of unique nodes, if we haven't already
    # determined that from a previous call
    if ( !$input{_num_nodes} ) {
        my $node_hash;
        foreach my $a (@aggrs) {
            my $node = $a->command_interface();
            my $name = $node->name;
            if ( !defined( $node_hash->{$name} ) ) {
                push( @nodes, $node );
                $node_hash->{$name} = 1;
            }
            push( @{ $node_to_aggrs->{$name} }, $a );
        }
        $_num_nodes = @nodes;
    }
    foreach my $a (@aggrs) {
        $Log->comment( "Aggregate: " . $a->aggregate() );
    }
    foreach my $n (@nodes) {
        $Log->comment( "Node: " . $n->name() );
    }

    # Store start tick for return later
    $my_return->{start_tick} = time();

    if ( !$input{_num_nodes} && ( $_num_nodes > 1 ) ) {

        # multiple nodes, multiple aggrs
        # More than one node specified, divide and conquer by splitting
        # work off on each node in parallel
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;

        foreach my $n (@nodes) {
            my $node_name = $n->name();
            my $id        = 'WIRONC_N_DaC.' . $proc_cnt;
            push( @ids, $id );

            # Already parsed out unique nodes, don't need to do again
            $input{_num_nodes} = $_num_nodes;
            $input{aggregates} = $node_to_aggrs->{$node_name};

            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Wafliron',
                method  => 'start',
                args    => [%input],
            );
            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;

        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }
        $my_return->{end_tick} = time();
        if ($had_failure) {
            $my_return->{result} = 0;
            $my_return->{msg}    = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            my $msg = "Ran wafliron on $_num_nodes nodes";
            $Log->comment($msg);
            $my_return->{msg} = $msg;
        }
        return $my_return;
    } elsif ( $num_aggr > 1 ) {

        # single node, multiple aggr
        my $node_name = $aggrs[0]->command_interface()->name();
        my $idx       = 0;
        my $done      = 0;
        my @workers;
        my $futures;
        my $proc_cnt = 1;
        my @ids;
        my $input_ref = \%input;
        $my_return->{start_tick} = time();

        # Work on current batch of aggrs
        foreach my $a (@aggrs) {
            my $id = 'WIRONC_A_DaC' . $proc_cnt;
            push( @ids, $id );
            $input{aggregates} = [$a];
            $input{_num_nodes} = 1;

            # added delay to offset wafliron runs
            sleep(10);
            my $worker = NATE::PerlRPC::Worker::Process->new( runid => $id );
            my $future = $worker->submit(
                package => 'NX::Wafliron',
                method  => 'start',
                args    => [%input],
            );
            $Log->comment("$id is running");
            push( @workers,     $worker );
            push( @Pkg_Workers, $worker );
            $futures->{$id}->{obj} = $future;
            $proc_cnt++;
        }

        $Log->comment("collecting results");

        # Collect results
        my $res = collect_worker_results(
            workers       => \@workers,
            futures       => $futures,
            merge_context => 0,
        );

        my $had_failure = 0;
        my $msg;
        foreach my $w (@workers) {
            my $rid = $w->runid();

            if ( $res->{$rid}->{result} != 1 ) {
                $Log->error("$rid had a failure");
                $had_failure = 1;
            } else {
                $Log->comment("$rid ran successfully");
            }
        }
        $my_return->{end_tick} = time();

        if ($had_failure) {
            $my_return->{result} = 0;
            $msg = 'One or more procs had failures';
        } else {
            $my_return->{result} = 1;
            $msg = "Ran wafliron on all $num_aggr aggregates on $node_name";
        }
        $Log->comment($msg);
        $my_return->{msg} = $msg;
        return $my_return;
    } else {

        # single node, single aggr
        my $node      = $aggrs[0]->command_interface();
        my $aggr_name = $aggrs[0]->aggregate();
        my $msg;
        my @vol_list;
        my $wafliron_obj;
        my $wafliron_review;

        # flag to initialize wafliron obj once
        my $ran_once = 0;

        # flag to online aggr
        my $online;

        # counter for loop
        my $i    = 1;
        my $done = 0;

        my $start_tick = time();
        $my_return->{start_tick} = $start_tick;

        # outer try to catch exceptions thrown for $::IGNORE_ERRORS
        try {
            if (($i eq '1') && $input{delay_start}) {
                wait_for_interval(
                    interval     => $input{interval},
                    start_minute => $input{start_minute},
                    start_tick   => $start_tick,
                );
            }

            while ( !$done ) {
                my $was_down = 0;
                try {
                    # Make sure node is up
                    my $node_up = 0;
                    while (!$node_up) {
                        my $obj = NACL::Transit->new(name => $node->hostrec());
                        my $state = $obj->get_state(transient => 1);
                        if ($state =~ /ontap|cli|username/i) {
                            $node_up = 1;
                            $Log->comment("Node ". $node->node() . " is up. "
                                . "Proceeding with wafliron.");
                        } else {
                            $Log->comment("Node ". $node->node()
                                . " is not fully up. State is $state. Waiting"
                                . " for 5 minutes");
                            sleep 300;
                            $was_down = 1;
                        }
                    }

                    if (($was_down) || (!$ran_once)) {
                        # These only need to be run once if node is up
                        $wafliron_obj = NACL::MTask::Wafliron->new(
                            command_interface => $node,
                            aggregate         => $aggr_name,
                        );

                        @vol_list = NACL::STask::Volume->find(
                            command_interface => $node,
                            filter            => { aggregate => $aggr_name, }
                        );
                        $ran_once = 1;
                    }

                    $Log->comment("Start wafliron run # $i");

                    # stop efficiency operations on all volumes on aggregate
                    # it seems like NACL::MTask::Wafliron already stops it but
                    # this is a backup
                    NX::Volume::efficiency(
                        mode    => 'stop',
                        volumes => \@vol_list,
                    );

                    $wafliron_obj->start(
                        nacltask_optional_commit => $input{optional_commit}, 
                        'method-timeout'         => $input{timeout},
                        nacltask_timeout         => 300,
                    );

                    if ( $input{optional_commit} eq 'true' ) {
                        $wafliron_review = $wafliron_obj->review();
                        $Log->comment("**** wafliron review ****");
                        $Log->comment($wafliron_review);

                        if ( $wafliron_review =~ m/committed/ ) {

                            # there are no errors found
                            $Log->comment(
"**** wafliron optional commit did not find any errors ****"
                            );
                            $online = 1;
                        } elsif ( $wafliron_review =~ m/corrected/ ) {

                            # there are errors present
                            if ( $input{ioc_if_error_found} eq 'commit' ) {
                                $Log->comment(
"**** wafliron found errors, committing fixes ****"
                                );
                                $wafliron_obj->commit( nacltask_force => 1 );
                                $online = 1;
                            } elsif ( $input{ioc_if_error_found} eq 'reject' ) {
                                $Log->comment(
"**** wafliron found errors, but rejecting fixes ****"
                                );
                                $wafliron_obj->reject();
                                $online = 1;
                            } else {

                                # raise failure exception
                                $Log->comment(
"**** wafliron found errors, marking wafliron action as failure ****"
                                );
                                NATE::BaseException->throw(
"Wafliron IOC error. Preserving state for review"
                                );
                            }
                        }

                        if ($online) {

                            # online aggr after running
                            NACL::C::StorageAggregate->online(
                                command_interface => $node,
                                aggregate         => $aggr_name,
                            );

                            # reset flag
                            $online = 0;
                        }
                    }
                    $Log->comment("Completed wafliron run #$i");

                    # don't call interval if last iteration
                    if ( $i != $iterations || $input{iterations} eq 'infinite' )
                    {
                        $Log->comment(
"Sleeping for $interval seconds before next run #$i on $aggr_name"
                        );
                        wait_for_interval(
                            interval     => $input{interval},
                            start_minute => $input{start_minute},
                            start_tick   => $start_tick,
                        );
                    } else {
                        $Log->comment("Last iteration #$i");
                        $done = 1;
                    }
                } otherwise {
                    my $exception = shift;
                    if ($::IGNORE_ERRORS) {
                        $Log->warn(
"Exception on $aggr_name on run #$i. Ignoring and continuing: "
                              . $exception->text() );

                        # needs a wait interval even with exception
                        $Log->comment(
"Sleeping for $interval seconds before next run #$i on $aggr_name"
                        );

                        # don't call interval if last iteration
                        if (   $i != $iterations
                            || $input{iterations} eq 'infinite' )
                        {
                            $Log->comment(
"Sleeping for $interval seconds before next run #$i on $aggr_name"
                            );
                            wait_for_interval(
                                interval     => $input{interval},
                                start_minute => $input{start_minute},
                                start_tick   => $start_tick,
                            );
                        } else {
                            $Log->comment("Last iteration #$i");
                            $done = 1;
                        }
                    } else {
                        $exception->throw();
                    }
                };
                $i++;
            }    # end main loop for wafliron execution
        } otherwise {
            $exception = shift;
        };

        $my_return->{end_tick} = time();
        if ($exception) {
            $my_return->{result}    = 0;
            $my_return->{exception} = $exception;
            $msg                    = "Failed wafliron run on $aggr_name";
        } else {
            $my_return->{result} = 1;
            $msg = "Successful wafliron run on $aggr_name $iterations times";
        }
        $Log->comment($msg);
        $my_return->{msg} = $msg;
        return $my_return;
    }
}    # End of start

1;
