# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/CDP.pm#1 $
#
# Copyright (c) 2004 Network Appliance Inc.
# All right reserved.
#
## @summary Routines for running network utilities
##
## @description
##         These functions are used to run various networking
##         utilities on devices independent of the device type.
##        These functions help in abstracting the test scripts
##        from subtle differences in command usage and command
##        outputs on various platforms.
##
## @dependencies
##
## use Tharn;
##
## @change 2004-10-05 gajria: fixed error match string in config_ipfw_*
## and added return eval.
##
## @change 2005-03-19 jcorbin: added function get_subnet.
##
## @author  dl-maryv (FSS Protocols QA)

package Network::CDP;

require Exporter;
use strict;
use Data::Dumper;
use Tharn;
use filer qw(filer_options_get filer_options_set);

use vars qw (@ISA @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT_OK = qw (cdp_is_supported 
		cdp_is_ready
		cdp_is_enabled 
		cdp_enable
                cdp_disable 
		cdp_zero_stats 
		cdp_test 
		cdp_stats_get
		cdp_show_neighbors
		cdp_show_neighbors_7mode
		cdp_show_neighbors_detail_7mode
		cdp_option_enable_disable_partner_7mode
		cdp_statistics_7mode
            );

my $debug = 1;

## @name ::cdp_is_supported()
##
## @summary verifies whether CDP is supported on the release of Ontap
##
## @description
##
## this function takes a host connection object, queries the registry
## and determines whether or not CDP is supported
##
## returns 1 if CDP is supported, 0 if not
##
## @arg $fc Required: connectspec of filer
##

sub cdp_is_supported($) {

    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();

    if ( $hosttype =~ /filer-cfe/ ) {
	return filer_options_get($fc,"cdpd.enable");
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
	my $hostname = $fc->hostp->hostname();
        my (undef, undef, $line) = say ($fc,"node run $hostname options cdpd.enable");
        if ($line =~ /No such option cdpd/) {
	    return 0;
        } else {
	    return 1;
	}
    }
    return 0;
}

## @name ::cdp_is_ready()
##
## @summary verifies whether CDP has completed initialization
##
## @description
##
## this function takes a host connection object, runs the cdp show-stats
## command and determines whether or not CDP is initialized
##
## returns 1 if CDP is initialized, 0 if not
##
## @arg $fc Required: connectspec of filer
##

sub cdp_is_ready($) {

    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();
    my $cmd;

    if ( $hosttype =~ /filer-cfe/ ) {
        $cmd = "priv set test; cdpd test 1";
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        $cmd = "node run $hostname priv set test; cdpd test 1";
    } else {
        die fail_result("Unsupported platform type for cdpd");
    }

    my (undef, undef, $line) = say ($fc, $cmd);

    if ($line =~ /not found/) {
        die fail_result("CDPD is not supported in this release of Ontap");
    } elsif (($line =~ /cdpd: /) ||($line =~ /Usage: /)) {
        die fail_result("Error running cdpd test 1: $@");
    } else {
	return ($line !~ /CDP not initialized/);
    }
}

## @name ::cdp_is_enabled()
##
## @summary verifies whether CDP is enabled
##
## @description
##
## this function takes a host connection object, queries the registry
## and determines whether or not CDP is enabled
##
## returns 1 if CDP is enabled, 0 if not
##
## @arg $fc Required: connectspec of filer
##

sub cdp_is_enabled ($){
    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();
    my $cdp_enabled;

    if ( $hosttype =~ /filer-cfe/ ) {
        $cdp_enabled = filer_options_get($fc,"cdpd.enable");
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        (undef, undef, $cdp_enabled) = say ($fc,"node run $hostname options cdpd.enable");
        if ($cdp_enabled =~ /No such option/) {
	    die fail_result("CDPD is not supported in this release of Ontap");
        } elsif ($cdp_enabled =~ /option cdpd.enable: /) {
	    die fail_result("Error viewing options cdpd.enable: $cdp_enabled");
        } 
    } else {
        die fail_result("Unsupported platform type for cdp_enable");
    }

    return ($cdp_enabled =~ /cdpd.enable\s+on[\n\r]/);

}


## @name ::cdp_enable()
##
## @summary enables CDP
##
## @description
##
## this function takes a host connection object, sets the registry
## option cdpd.enable to on
##
## returns 1 if CDP is successfully enabled, dies a horrible death if unsuccessful
##
## @arg $fc Required: connectspec of filer
##

sub cdp_enable ($){
    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();

    if ( $hosttype =~ /filer-cfe/ ) {
        return filer_options_set($fc,"cdpd.enable", "on");
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        my (undef, undef, $line) = say ($fc,"node run $hostname options cdpd.enable on");
        if ($line =~ /No such option/) {
            die fail_result("CDPD is not supported in this release of Ontap");
        } elsif ($line =~ /option cdpd.enable: /) {
            die fail_result("Error setting options cdpd.enable on: $@");
        } else {
            return 1;
        }
    }
    die fail_result("Unsupported platform type for cdp_enable");
}


## @name ::cdp_disable()
##
## @summary disables CDP
##
## @description
##
## this function takes a host connection object, sets the registry
## option cdpd.enable to off
##
## returns 1 if CDP is successfully disabled, dies a horrible death if unsuccessful

##
## @arg $fc Required: connectspec of filer
##

sub cdp_disable ($){
    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();

    if ( $hosttype =~ /filer-cfe/ ) {
        return filer_options_set($fc,"cdpd.enable", "off");
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        my (undef, undef, $line) = say ($fc,"node run $hostname options cdpd.enable off");
        if ($line =~ /No such option/) {
            die fail_result("CDPD is not supported in this release of Ontap");
        } elsif ($line =~ /option cdpd.enable: /) {
            die fail_result("Error setting options cdpd.enable off: $@");
        } else {
            return 1;
        }
    }
    die fail_result("Unsupported platform type for cdp_disable");

}


## @name ::cdp_zero_stats()
##
## @summary Zeros the CDP stats.
##
## @description
##
## this function calls cdpd zero-stats
##
## returns 1 
##
## @arg $fc Required: connectspec of filer
##

sub cdp_zero_stats ($){
    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();
    my $cmd;

    if ( $hosttype =~ /filer-cfe/ ) {
	$cmd = "cdpd zero-stats";
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
	$cmd = "node run $hostname cdpd zero-stats";
    } else {
	die fail_result("Unsupported platform type for cdpd");
    }

    my (undef, undef, $line) = say ($fc, $cmd);

    if ($line =~ /not found/) {
	die fail_result("CDPD is not supported in this release of Ontap");
    } elsif (($line =~ /cdpd: /) || ($line =~ /Usage:/)) {
	die fail_result("Error running cdpd zero-stats: $line");
    } else {
        return 1;
    }
}


## @name ::cdp_test()
##
## @summary runs the CDPD test specified
##
## @description
##
## this function takes a host connection object, a test id
## and runs the cdpd test <test_id> command
##
## returns the results of the test command
##
## @arg $fc Required: connectspec of filer
## @arg $test_id Required: number of test to run (1..12)
##

sub cdp_test ($$){
    my ($fc, $test_id) = @_;
    my $hosttype = $fc->hostp->hosttype();
    my $cmd;

    if ( $hosttype =~ /filer-cfe/ ) {
        $cmd = "priv set test; cdpd test $test_id";
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        $cmd = "node run $hostname priv set test; cdpd test $test_id";
    } else {
        die fail_result("Unsupported platform type for cdpd");
    }

    my (undef, undef, $line) = say ($fc, $cmd);

    if ($line =~ /not found/) {
        die fail_result("CDPD is not supported in this release of Ontap");
    } elsif (($line =~ /cdpd: /) ||($line =~ /Usage: /)) {
        die fail_result("Error running cdpd test $test_id: $@");
    } else {
        return ($line =~ /CDP receive queing passed/);
    }
}


## @name ::cdp_stats_get()
##
## @summary Runs the cdpd show-stats command and extracts the value of the named stat.
##
## @description
##
## this function takes a host connection object, the name of a cdpd stat
## and runs the cdpd show-stats command and returns the value of the named stat
##
## returns the value of the cdpd stat
##
## @arg $fc Required: connectspec of filer
## @arg $stat Required: name of cdpd stat.  Must be one of these: rx_pkts, rx_inval_len, 
## 	rx_mtlv, rx_csum_errs, rx_malformed, rx_cache_over, rx_unsupp_vers, rx_mem_alloc_fail, 
##	rx_misc, tx_pkts, tx_trunc, tx_xmit_fail, tx_mem_alloc_fail, tx_no_hostname, tx_misc, init_fail.
##

sub cdp_stats_get($$){
    my ($fc, $stat) = @_;
    my $hosttype = $fc->hostp->hosttype();
    my $cmd;
    my %stats;

    if ( $hosttype =~ /filer-cfe/ ) {
        $cmd = "cdpd show-stats";
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        $cmd = "node run $hostname cdpd show-stats";
    } else {
        die fail_result("Unsupported platform type for cdpd");
    }

    my (undef, undef, $line) = say ($fc, $cmd);

    if ($line =~ /not found/) {
        die fail_result("CDPD is not supported in this release of Ontap");
    } elsif (($line =~ /cdpd: /) ||($line =~ /Usage: /)) {
        die fail_result("Error running cdpd show-stats: $line");
    } else {
	my @stats_out = split(/\n/, $line);
	foreach my $i (0..$#stats_out){
	    if ($stats_out[$i] =~ /Packets:\s+(\d)+\s+.\s+Csum Errors:\s+(\d)+\s+.\s+Unsupported Vers:\s+(\d)\s*$/){
		$stats{'rx_pkts'} = $1;
		$stats{'rx_csum_errs'} = $2;
		$stats{'rx_unsupp_vers'} = $3;
		print "match 1: $stats_out[$i]\n" if $debug;
	    } elsif ($stats_out[$i] =~ /Invalid length:\s+(\d)+\s+.\s+Malformed:\s+(\d)+\s+.\s+Mem alloc fails:\s+(\d)\s*$/){
		$stats{'rx_inval_len'} = $1;
                $stats{'rx_malformed'} = $2;
                $stats{'rx_mem_alloc_fail'} = $3;
		print "match 2: $stats_out[$i]\n" if $debug;
	    } elsif ($stats_out[$i] =~ /Missing TLVs:\s+(\d)+\s+.\s+Cache overflow:\s+(\d)+\s+.\s+Other errors:\s+(\d)\s*$/){
                $stats{'rx_mtlv'} = $1;
                $stats{'rx_cache_over'} = $2;
                $stats{'rx_misc'} = $3;
		print "match 3: $stats_out[$i]\n" if $debug;
            } elsif ($stats_out[$i] =~ /Packets:\s+(\d)+\s+.\s+Xmit fails:\s+(\d)+\s+.\s+No hostname:\s+(\d)\s*$/){
                $stats{'tx_pkts'} = $1;
                $stats{'tx_xmit_fail'} = $2;
                $stats{'tx_no_hostname'} = $3;
		print "match 4: $stats_out[$i]\n" if $debug;
            } elsif ($stats_out[$i] =~ /Packet truncated:\s+(\d)+\s+.\s+Mem alloc fails:\s+(\d)+\s+.\s+Other errors:\s+(\d)\s*$/){
                $stats{'tx_trunc'} = $1;
                $stats{'tx_mem_alloc_fail'} = $2;
                $stats{'tx_misc'} = $3;
		print "match 5: $stats_out[$i]\n" if $debug;
	    } elsif ($stats_out[$i] =~ /Init failures:\s+(\d)\s*$/){
		$stats{'init_fail'} = $1;
		print "match 6: $stats_out[$i]\n" if $debug;
	    } else {
		print "nomatch: $stats_out[$i]\n" if $debug;
	    }

	}
	
	print "stats =",Data::Dumper->Dump([\%stats]),"\n" if $debug;

	if (defined $stats{$stat}){
	    return $stats{$stat};
	} else {
	    die fail_result("CDPD show-stats did not return a value of type $stat");
	}
    }
}

## @name ::cdp_show_neighbors()
##
## @summary Runs the cdpd show-neighbors command and returns a hash
##
## @description
##
## this function takes a host connection object, 
## and runs the cdpd show-neighbors command and returns a hash with the neighbors
##
## returns the neighbors hash
##
## @arg $fc Required: connectspec of filer
## @arg $stat Required: name of cdpd stat.  Must be one of these: rx_pkts, rx_inval_len,
##      rx_mtlv, rx_csum_errs, rx_malformed, rx_cache_over, rx_unsupp_vers, rx_mem_alloc_fail,
##      rx_misc, tx_pkts, tx_trunc, tx_xmit_fail, tx_mem_alloc_fail, tx_no_hostname, tx_misc, init_fail.
##

sub cdp_show_neighbors($){
    my ($fc) = @_;
    my $hosttype = $fc->hostp->hosttype();
    my $cmd;
    my $neighbors = {};

    if ( $hosttype =~ /filer-cfe/ ) {
        $cmd = "cdpd show-neighbors -v";
    } elsif ( $hosttype =~ /filer-ng-cfe/ ) {
        my $hostname = $fc->hostp->hostname();
        $cmd = "node run $hostname cdpd show-neighbors -v";
    } else {
        die fail_result("Unsupported platform type for cdpd");
    }

    my (undef, undef, $line) = say ($fc, $cmd);

    my @neighbors_out = split(/\n/, $line);
    my $remote_device = "";

    foreach my $i (0..$#neighbors_out){
	if ($neighbors_out[$i] =~ /Remote Device:\s+(.+)$/){
	    $remote_device = $1;
	    $$neighbors{$remote_device} = {};
	    $$neighbors{$remote_device}{addresses} = [];
	} elsif ($neighbors_out[$i] =~ /Local Port:\s+(.+)$/){
	    $$neighbors{$remote_device}{local_port} = $1;
	} elsif ($neighbors_out[$i] =~ /Holdtime:\s+(.+)$/){
	    $$neighbors{$remote_device}{holdtime} = $1;
	} elsif ($neighbors_out[$i] =~ /Capabilities:\s+(.+)$/){
	    $$neighbors{$remote_device}{capabilities} = $1;
	} elsif ($neighbors_out[$i] =~ /Platform:\s+(.+)$/){
	    $$neighbors{$remote_device}{platform} = $1;
	} elsif ($neighbors_out[$i] =~ /Remote Port:\s+(.+)$/){
	    $$neighbors{$remote_device}{remote_port} = $1;
	} elsif ($neighbors_out[$i] =~ /Version:\s+(.+)$/){
	    $$neighbors{$remote_device}{version} = $1;
	} elsif ($neighbors_out[$i] =~ /(\d+\.\d+\.\d+\.\d+)/){
	    push (@{$$neighbors{$remote_device}{addresses}}, $1);
	}
    }
    print "neighbors=",Data::Dumper->Dump([\%$neighbors]),"\n" if $debug;
    return $neighbors;

}
##Subroutine to get the details of a cdpd show neighbors command in filer and show cdp-neighbors commenad in filer.
##Arguments
## $conn Conection handle -> Switch or filer.
## $args -> partner (1 or 0)
## Usage.
##cdp_show_neighbors_7mode($fc2);
##cdp_show_neighbors_7mode($sc);
##cdp_show_neighbors($fc2,{partner=>1});
sub cdp_show_neighbors_7mode
{
	my $conn = shift;
	my $args = shift;
	my $cdp_output;
	my $hosttype = $conn->hostp->hosttype();
	my %cdpd_vals=();
	if($hosttype =~ m/filer/i)
	{
		my $partner="";
		$partner = "partner" if($args->{partner});
		(undef,undef,$cdp_output) = say($conn,"$partner cdpd show-neighbors",'-swallow');
		$cdp_output =~ s/cdpd show-neighbors\n*//;
		my @lines = split(/\n/,$cdp_output);
		foreach(@lines)
		{
			if($_ !~ m/^(Local|Port|--|\s+|\n)/i)
			{
				my @vals=split(/\s{2,}/,$_);
				$cdpd_vals{$vals[0]}{'Remote_Device'} = $vals[1];
				$cdpd_vals{$vals[0]}{'Remote_Interface'} = $vals[2];
				$cdpd_vals{$vals[0]}{'Remote_Platform'} = $vals[3];
				$cdpd_vals{$vals[0]}{'Hold_Time'} = $vals[4];
				$cdpd_vals{$vals[0]}{'Remote_Capability'} = $vals[5];
			}
		}
	}
	elsif($hosttype =~ m/switch/i)
	{
		(undef,undef,$cdp_output) = say($conn,"show cdp neighbors");
		$cdp_output =~ s/\#show cdp neighbors\n*//;
		my @lines = split(/\n/,$cdp_output);
		foreach(@lines)
		{
		if($_ !~ m/^(\s*Capability Codes|\s+|\n|\s*Device ID)/i)
		{
			my @vals=split(/\s{2,}/,$_);
			$vals[1]=~s/\s+//g;
			$vals[1]=~s/Gig/GigabitEthernet/g;
			$cdpd_vals{$vals[1]}{'Device_ID'} = $vals[0];
			$cdpd_vals{$vals[1]}{'Hold_Time'} = $vals[2];
			$cdpd_vals{$vals[1]}{'Remote_Platform'} = $vals[4];
			$cdpd_vals{$vals[1]}{'Remote_Capability'} = $vals[3];
			$cdpd_vals{$vals[1]}{'Remote_Port'} = $vals[5];
		}
	}
}
	return ( \%cdpd_vals );
}
##Subroutine to get the details of a cdpd show neighbors -v command in filer and show cdp-neighbors detail command in filer.
##Arguments
## $conn Conection handle -> Switch or filer.
## $args -> partner (1 or 0)
## Usage.
##cdp_show_neighbors_detail_7mode($fc2);
##cdp_show_neighbors_detail_7mode($sc);
##cdp_show_neighbors_detail_7mode($fc2,{partner=>1});

sub cdp_show_neighbors_detail_7mode
{
	my $conn = shift;
	my $args = shift;
	my ($cdp_output,$key,$key_val);
	my $hosttype = $conn->hostp->hosttype();
	my %cdpd_vals=();
	if($hosttype =~ m/filer/i)
	{
		my $partner="";
		$partner = "partner" if($args->{partner});
		(undef,undef,$cdp_output) = say($conn,"$partner cdpd show-neighbors -v",'-swallow');
		$cdp_output =~ s/cdpd show-neighbors -v\n//;
		my @lines = split(/\n/,$cdp_output);
		foreach(@lines)
		{
			my $value=$_;
			$value=~s/\n|\r//g;
			if($value =~ m/Local/)
			{
				my @vals=split(/:\s+/,$value);
				$key = $vals[1];
				if($cdp_output=~ m/Remote Device\:(.*?)\nLocal Port\:\s*$key/i)
				{
					$cdpd_vals{$key}{Remote_Device} = $1;
				}
			}
			elsif($value =~ m/^(Holdtime|Capabilities|Platform|Remote Port)/i)
			{
				my @vals=split(/:\s+/,$value);
				$vals[0] =~ s/\s+/_/g;
				$cdpd_vals{$key}{$vals[0]}=$vals[1];
			}
			elsif($value =~ m/^Entry Address/i)
			{
				if($cdp_output =~ m/$key.*?Entry Address\(es\)\:\s*((\d+\.?)+)/s)
				{
				     $cdpd_vals{$key}{'Entry_Address'} = $1;
				}
			}
			elsif($value =~ m/^Version/i)
			{
				if($cdp_output =~ m/$key.*?Version\:(.*?)Technical Support/s)#(.*)Version/) #:(.*)Technical Support/s)
				{
				    $cdpd_vals{$key}{'Version'}=$1;
				} 
			}
		}
	}
	elsif($hosttype =~ m/switch/i)
	{
		my (undef,undef,$switch_val) = say($conn,"show cdp neighbors detail");
		$switch_val =~ s/\#show cdp neighbors detail//;
		my @lines_switch= split(/-{3,}\n?/,$switch_val);
		foreach my $res (@lines_switch)
		{
			if($res =~ m/Interface:\s*(.*?),\s*Port ID \(outgoing port\):\s*(.*)\n/i)
			{
				$key=$1;
				$cdpd_vals{$key}{'Port_Id'} = $2;
			}
			if($res =~ m/^Platform\s*:(.*?),\s*Capabilities:\s*(.*)/s)
			{
				$cdpd_vals{$key}{'Platform'}=$1;
				$cdpd_vals{$key}{'Capabilities'}=$2;
			}
			if($res =~ m/Version\s*:(.*?)advertisement/s)
			{
				$cdpd_vals{$key}{'Version'}=$1;
				$cdpd_vals{$key}{'Version'} =~ s/\r|\n//g;
			}
			foreach(split(/\n/,$res))
			{    
				$_=~s/\n|\r//g;
				$_ =~s/^\s+//g;
				if($_ =~ m/^(Device ID|IP address|IPv6 address|Holdtime|advertisement version|VTP Management Domain|Native VLAN|Duplex)/)
				{
					my @vals2=split(/\s*:\s+/,$_);
					$vals2[0] =~ s/\s+/_/g;
					if($vals2[0] =~ m/IP_address/) 
					{
						push @{ $cdpd_vals{$key}{IP_address} }, $vals2[1];
					}
					elsif($vals2[0] =~ m/IPv6_address/)
					{
						 push @{ $cdpd_vals{$key}{IPv6_address}},$vals2[1];
					}
					else
					{
						$cdpd_vals{$key}{$vals2[0]}=$vals2[1];
					}
				}
			}
		}
	}
return ( \%cdpd_vals );
}
##Subroutine to enable and disable cdp in both filer and partner filer
##Arguments:-
## Required :- $conn - > Connection handle
## %args -> To specify the value of the option and whether to do it in the partner context or in the present cintext.
##Usage:-
##_cdp_option_enable_disable_partner_7mode($fc2,{partner=>1,value=>"off"})
##_cdp_option_enable_disable_partner_7mode($fc2,{value=>"off"})
##
sub cdp_option_enable_disable_partner_7mode
{
	my $conn =shift;
	my $args =shift;
	my $partner="";
	if($args->{partner})
	{
		$partner="partner ";
	}
	say($conn,"$partner options cdpd.enable $args->{value}");
	my (undef,undef,$res) = say($conn,"options cdpd.enable",'-swallow');
	if($res !~ m/$args->{value}/)
	{
		return 0;
	}
	return 1;
}

##Subroutine to get the statistics 
##Arguments:-
## Required :- $conn - > Connection handle
## %args -> To specify the value of the option and whether to do it in the partner context or in the present cintext.
##Usage:-
## cdp_statistics_7mode($fc2,{partner=>1});
## cdp_statistics_7mode($fc2);
##
sub cdp_statistics_7mode
{
	my $conn = shift;
        my $args = shift;
        my ($cdp_stat_output,$section,@fields,$label,$val);
	my %cdpd_statistics=();
	my $partner="";
        $partner = "partner" if($args->{partner});
        (undef,undef,$cdp_stat_output) = say($conn,"$partner cdpd show-stats",'-swallow');
	my @info = split(/\n/,$cdp_stat_output);
        foreach my $line (@info)
	{
		if ($line =~ /(RECEIVE|TRANSMIT|OTHER)/)
		{
		        $section = $1;
			next;
		}
		next if (!defined($section));
		@fields = split(/\|/, $line) if($line !~ m/(^\s*$|^\n$|^\r$|^$)/);
		foreach $line (@fields)
		{
			#$line =~ s/\s*?$//;
			$line =~ /\s+(.*):\s+(.*)$/;
			$label = $1;
			$val = $2;
			$label =~ s/ /_/g;
			my ($num, $unit) = $val =~ /^(\d+)\s*(t|g|m|k)*$/;
			if (defined($unit))
			{
				if ($unit eq 't')
				{
					$num *= 1000000000000;
				}
				elsif ($unit eq 'g')
				{
					$num *= 1000000000;
				}
				elsif ($unit eq 'm')
				{
					$num *= 1000000;
				}
				elsif ($unit eq 'k')
				{
					$num *= 1000;
				}
			}
			$val = $num if (defined($num));
			$cdpd_statistics{$section}{$label} = $val;
		}
	}
	return(\%cdpd_statistics);
}
1;
