#!/usr/local/bin/perl
# $Id: //depot/prod/test/main/lib/CSM.pm#4 $
## @summary     Provides the NACL api to monitor performance, isolate issues, gather error counters, and surmise CSM health.
## @author    
## @pod here

=head1 NAME

CSM.pm

=head1 SYNOPSIS

	use Network::CSM;
	$obj = Network::CSM->new();
	$obj->initialize();
	$obj->csm_health_summary(
				level=> "low"|"medium"|"high",
				);
	$obj->SessionDecoder(uniquifier => $uniquifier );
	$ResponseonCreate=$obj->SessionCreate(connectionType => "NullTx"|"Rc"|"Ct"|"IntraclusterCt",
					  node1 => $active_node,
					  node2 => $passive_node,
					  numChannelsLow => $numLowPriorityChannels,
					  numChannelsMed => $numMediumPriorityChannels,
					  numChannelsHigh => $numHighPriorityChannels, -- See BURT 517395
					  windowSizeLow => $lowPriorityWindowSize,
					  windowSizeMed => $mediumPriorityWindowSize,
					  windowSizeHigh => $highPriorityWindowSize,
					  timeoutMs => $TimeOutInMillisecondsForCreate,
					  disableCongestionControl => "true"|"false",
					  sourceBlade => "Nblade|Dblade|SpinVFS,SCSIBlade", -- See BURT 520858, 520861
					  destBlade => "Nblade|Dblade|SpinVFS,SCSIBlade"); -- See BURT 520858, 520861

	$uniquifier=$obj->SessionCreateCheck(connectionType => "NullTx"|"Rc"|"Ct"|"IntraclusterCt",`
						node1 => $active_node,
						node2 => $passive_node,
						response => $ResponseonCreate,
						timeoutMs => $TimeOutInMillisecondsForCreate,
						bombPointCheck => "yes|no");
	

	$csmObj->SessionSendTest(uniquifier => $uniquifier
                             node => $active_node|$passive_node,
                             priority => "low"|"medium"|"high",
                             timeoutMs => $timeoutofSendCommand,
                             trackEachRequest => "true"|"false",
                             requestSize => $requestSize,
                             responseSize => $responseSize,
                             numPackets => $numPackets,
                             responseDelayMs => $delayBetweenEachResponse);

	$csmObj->ClobberTest(node => $active_node|$passive_node,
				uniquifier => $uniquifier,
			        remoteNode => "$active_node|$passive_node");

	$csmObj->SetFlowControl(node => $passive_node,
			 totalPackets => $totalPackets,
			 blade => "Dblade|Nblade|SpinVFS,SCSIBlade",
			 medHiReserved => med_high_priority_packetlimit,
			 highReserved => high_priority_packetlimit);

	$csmObj->FlowControlStats(node => $active_node|$passive_node,
				blade => "Dblade|Nblade|SpinVFS|SCSIblade,
				interfaceId => [1-13]); -- See BURT 520042	

	$csmObj->CsmBombPoint(node => $active_node|$passive_node,
				enable => 1|0,
			        value => $value,
				forever => 1|0,
				name => $bombPointName);

	$csmObj->BombPointCheck(node => $active_node | $passive_node,
				  bombPoint => $bombPointName,
				  response => $responseOfCreateorSend);
	

	C<CSM> Cluster Session Manager, or CSM provides a managed abstraction layer or tunnel for SpinNP sessions between the blades of each node in the cluster.
	Local node session communication between Nblade, and Dblade is handled by NullTX transport, while RC (Reliable Connection) transports
	internodal session messaging between the cluster blades. CSM uses UUID endpoints to map paths between blades. CSM connects blades... not
	nodes through mapped Logical Interfaces (LIFs). Interaction exists with VifMgr, VLDB, and MGWD. 

	Internodal Blade Session Communication:
	SpinNP <-> Blade UUID <-> RC <-> UDP <-> IP <-> Ethernet <-> IP <-> UDP <-> RC <-> Blade UUID <-> SpinNP

       Local Blade Session Communication:
       SpinNP <-> Blade UUID <-> NullTX <-> Blade UUID <-> SpinNP

       http://wikid.netapp.com/w/CSM 

=head1 DESCRIPTION

C<CSM.pm> CSM basically creates a tunnel by managing network resources and state in real time for applications. Because the transport
          layer is constantly building and tearing down connections in response to application requests/responses, enforcing flow-control,
          and dynamically adjusting to network conditions, a good litmus test is validating the performance of the applications that are
          being transported across the session.

          The CSM.pm library provides API for:
          1. Searching for errored, and/or troubled sessions
          2. Verifying network cluster lif connectivity using 'cluster-ping'.
          3. Verifying the CSM tunnel linking striped volumes across the cluster to all msid(s) using 'csmping'.
          4. Verifying network cluster port health.
          5. Gathering cluster-wide error counters using variances on the following commands:
             a. statistics show -category {striped|csm}
             b. debug csm session show
             c. debug csm counters show -category {nulltx|session|rc|csm|cache}
          6. Provides a low, medium, and high level impact CSM Health Summary which increases the granularity, and scope of potential CSM issues.
	  7. Creating and checking the status of the created 'test' sessions over NullTx, RC and ct
	  8. Sending requests of varying sizes and priorities over the created sessions 
	  9. Clobbering the created sessions
	 10. Setting flow control parameters for blades and retrieving flow control stats
	 11. Enabling and checking the impact of bomb points
	 12. Decoding the type of traffic over sessions by parsing the session uniquifiers



=head1 FUNCTIONS

=cut

package Network::CSM;

#use statements
use Data::Dumper;
use Tharn qw($Log);
use Options;
use Params::Common;
use Params::Validate qw(:all);
use NACL::C::Node;
use Carp;
use NACL::CS::DebugCsmSession;
use NACL::CS::Statistics;
use NACL::CS::Volume;
use NACL::CS::NetworkPort;
use NACL::CS::Interface; 
use NACL::APISet;
use NACL::CS::DebugCsmCounters;
use NACL::CS::ClusterIdentity;
use NACL::CS::DebugCsmSession;
use NACL::C::DebugCsmSession;
use NACL::CS::DebugCsmNulltx;
use NACL::CS::DebugCsmLocalBlade;
use NACL::CS::DebugCsmRc;
use NACL::CS::DebugCsmCt;
use NACL::CS::ClusterPeerHealth;
use NACL::CS::ClusterPeer;
use NACL::C::NetworkInterface;
use NACL::CS::NetworkInterface;
use Switch;
use strict;


=pod

=head2 new

C<new>

B<Description>
Creates, and initializes the CSM object which collects data for the primary litmus test
in determining CSM health using the 'debug csm session show' command. During "low", and
"medium" levels for CSM Health Summary, the state of these sessions will determine how 
deep the inspection will drill for errors. 

B<Output>
Blessed hash structure as a CSM object

=cut

sub new {
    my $class = shift;

    my $self = {_sessions => undef};
    bless $self,$class;

    $self->{_sessions} = $self->initialize();

    return $self;
}


=pod

=head2 initialize

C<initialize>

B<Description>
Used to initialize the CSM object which collects data for the primary litmus test
in determining CSM health using the 'debug csm session show' command. The scope of
the statistics is filtered via the @SESSION_FIELDS array.

B<Output>
Cluster CSM sessions stored within an array-of-hash.

=cut

sub initialize {
my @SESSION_FIELDS = qw(node uniquifier state type
                        initiator-string target-string
                        initiator target req-rcvd resp-sent
                        resp-rcvd req-sent req-pending
                        rsp-pending requests timedout-resp
                        timeout
                       );
    my $self = shift;
    my @session_aoh;
    my @output;
    my @Nodes = NACL::C::Node->find();
    eval {        
        @output = NACL::CS::DebugCsmSession->fetch(command_interface=>$Nodes[0]);

    };
    
    if ($@) {
        $Log->error("The 'debug csm session show' command failed due to an unknown error: $@");
    }
    foreach my $rec (@output) {
        my %session_fields;
        foreach my $field (@SESSION_FIELDS) {
            $session_fields{$field} = $rec->{$field};
        }
        push @session_aoh, {%session_fields};
    }
    return \@session_aoh;
}



=pod

=head2 SessionDecoder

C<SessionDecoder>

B<Description>
Retreives the type of traffic on  a session

B<Input>

=over 4

=item * uniquifier
	Uniquifier that uniquely identifies a session

=back

B<Output>
Traffic type corresponding to the first byte of the uniquifier string

=cut

sub SessionDecoder {

	my $self=shift;
	my %params = validate(@_,
				{
				uniquifier => { type => SCALAR, }
				});
	my $sessionMap;
	my $uniquifier=$params{uniquifier};
	while(length $uniquifier < 16) {
		$uniquifier="0".$uniquifier;
	}
	$Log->comment("Uniquifier is $uniquifier");
	my $sessionDecode={"00"=>"RC",
			"01"=>"Test",
			"02"=>"Coral Low Priority",
			"03"=>"Coral Medium Priority",
			"04"=>"Coral High Priority",
			"05"=>"Coral Extreme",
			"06"=>"Local Vol",
			"07"=>"Intercluster Paloma",
			"08"=>"Intracluster Paloma",
			"09"=>"SAN Local",
			"0a"=>"SAN Remote",
			"0b"=>"CTRAN",
			"0c"=>"vstorage"};
	($uniquifier=~/^(..)/) && ($sessionMap="$uniquifier indicates $sessionDecode->{$1} traffic");
	$Log->comment("Session Map is $sessionMap"); 
	
	return $sessionMap;
}

=pod

=head2 csm_health_summary

C<csm_health_summary>

B<Description>
Method used to gather and aggregate cluster health records into a summarized 
report for determining the overall health of CSM. The intensity, and impact
can be controlled by a user param-defined level of low, medium, or high. 
Each error instance found will generate a THARN WARN result.

B<Input>

=over 4
=item * B<level>

=over 4
=item * B<low> (default)
        If any CSM sessions are not "ESTABLISHED", cluster connectivity checks
        are performed, and errored session counters are returned.

=item * B<medium>
        If any CSM sessions are not "ESTABLISHED", OR troubled connections exist,
        cluster connectivity checks are performed, and errored counters are returned
        for session, csm, and lif categories.

=item * B<high>
        Checks for CSM sessions not in "ESTABLISHED" state.
        Checks for troubled connections.
        Performs cluster connectivity checks.
        Performs striped volume connectivity checks.
        Returns errored counters for: 
            1. csm
            2. session
            3. striped
            4. nulltx
            5. cache
            6. lif
            7. rc
        
=back

=back

B<Output>
Returns all cluster-wide error status, and counters in a hash of array.

Example output:
CSM Health Summary: $VAR1 = {
'rc_error_counters' => [
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-6070-13 -name BindSentOpAbort\' = 315',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-1 -name BindSentOpAbort\' = 354',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-1 -name LoSndAbort\' = 48',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-2 -name BindSentOpAbort\' = 72',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-2 -name LoPassiveOpenNoMem\' = 2',
  'WARNING: RC Errors from command \'debug csm counters show -category rc -node cqa-perf-2 -name LoSndAbort\' = 84'
 ],
'cache_error_counters' => [
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-6070-13 -name dsToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-6070-13 -name vifToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-1 -name dsToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-1 -name vifToBladeInvalidates\' = 2',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-2 -name dsToBladeInvalidates\' = 1',
  'WARNING: Cache Errors from command \'debug csm counters show -category cache -node cqa-perf-2 -name vifToBladeInvalidates\' = 1'
 ],
'striped_error_counters' => [
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-heartbeat-connect-failure\' = 266',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-req-send-retried\' = 10',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-req-send-failure\' = 21',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter client-csm-resp-failure\' = 75',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-13 -counter server-heartbeat-stalled-wafl-status\' = 2431981',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-6070-14 -counter server-heartbeat-stalled-wafl-status\' = 1157151',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-heartbeat-connect-failure\' = 437',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-req-send-retried\' = 10',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-req-send-failure\' = 41',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter client-csm-resp-failure\' = 100',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-1 -counter server-heartbeat-stalled-wafl-status\' = 2325788',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-heartbeat-connect-failure\' = 460',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-req-send-retried\' = 10',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-req-send-failure\' = 39',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter client-csm-resp-failure\' = 81',
   'WARNING: Striped Errors from command \'statistics show -category striped -object striped-statistics -node cqa-perf-2 -counter server-heartbeat-stalled-wafl-status\' = 2389427'
 ],
'cluster_errored_ports' => 0,
'cluster_connectivity_errors' => 0,
'session_error_counters' => [
   'WARNING: Session Errors from command \'debug csm counters show -category session -node cqa-6070-13 -name closefailed\' = 7',
   'WARNING: Session Errors from command \'debug csm counters show -category session -node cqa-perf-1 -name closefailed\' = 6',
   'WARNING: Session Errors from command \'debug csm counters show -category session -node cqa-perf-2 -name closefailed\' = 7'
 ],
'nulltx_error_counters' => [],
'csmping_striped_failures' => 0,
'lif_error_counters' => [],
'csm_error_counters' => [
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-6070-13 -name activeConnFailures\' = 72',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-6070-13 -name activeConnStartFailures\' = 361',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-6070-13 -name getSessionForDataSetFailRsp\' = 115',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-1 -name activeConnFailures\' = 72',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-1 -name activeConnStartFailures\' = 383',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-1 -name getSessionForDataSetFailRsp\' = 157',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name activeConnFailures\' = 72',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name activeConnStartFailures\' = 114',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name connReject_NoBlade\' = 84',
   'WARNING: CSM Errors from command \'debug csm counters show -category csm -node cqa-perf-2 -name getSessionForDataSetFailRsp\' = 127'
 ]
};

=cut

sub csm_health_summary {
	my $self = shift;
	my %summary;
	my %params = validate(@_,
			{
			level => { type => SCALAR,
			default => "low",
				},
			});
	$Log->comment("We are performing a $params{level} impact CSM health summary.\n");
	my $badness_check = $self->get_bad_sessions();
	if ((defined(@$badness_check)) || ($params{level} eq "high") || ($params{level} eq "medium")) {
		$Log->comment("We will report CSM sessions that are not ESTABLISHED, and"."cluster network connectivity checks will be executed.\n");
		$summary{bad_sessions} = @$badness_check;
		$summary{cluster_connectivity_errors} = $self->get_cluster_connectivity_errors();
		$summary{cluster_errored_ports} = $self->get_cluster_errored_ports();
		$Log->comment("Gathering Session Errors:\n");
		$summary{session_error_counters} = $self->get_error_counters(category=>"session");

		if ($params{level} ne "low") {
			$summary{troubled_sessions} = $self->get_troubled_sessions();
			$Log->comment("Gathering CSM Errors:\n");    
			$summary{csm_error_counters} = $self->get_error_counters(category=>"csm");
			$Log->comment("Gathering LIF Errors:\n");
			$summary{lif_error_counters} = $self->get_lif_error_counters();
		}
		if ($params{level} eq "high") {
			$Log->comment("Csmping Striped Volumes:\n");
			$summary{csmping_striped_failures} = $self->csmping_striped_vols();
			$Log->comment("Gathering Striped Errors:\n");
			$summary{striped_error_counters} = $self->get_striped_error_counters();
			$Log->comment("Gathering RC Errors:\n");
			#$summary{rc_error_counters} = $self->get_error_counters(category=>"rc"); #commented becuse of burt 514740
			$Log->comment("Gathering Cache Errors:\n");
			$summary{cache_error_counters} = $self->get_error_counters(category=>"cache");
			$Log->comment("Gathering Nulltx Errors:\n");
			$summary{nulltx_error_counters} = $self->get_error_counters(category=>"nulltx");
		}
	}
	
	else {
		$Log->comment("Passed $params{level} CSM Health Summary!");
		$summary{passed} = "Passed";   
	}
	
	return \%summary;
}
    
=pod

=head2 csmping_striped_vols

C<csmping_striped_vols>

B<Description>
Used to verify CSM connectivity to striped volumes across the cluster using the  
'csmping' utility for acro style tests.
Each error instance found will generate a THARN WARN result.

coraltest csmping -a -m 2147484747
Will check connectivity between localhost and all constituent volumes of the coral set 2147484747
contacting localhost...
sending csmping from localhost to owner of volume 1625...: Reachable
sending csmping from localhost to owner of volume 1626...: Reachable
sending csmping from localhost to owner of volume 1627...: Reachable
sending csmping from localhost to owner of volume 1628...: Reachable
sending csmping from localhost to owner of volume 1629...: Reachable
sending csmping from localhost to owner of volume 1630...: Reachable
sending csmping from localhost to owner of volume 1631...: Reachable
sending csmping from localhost to owner of volume 1632...: Reachable
8 Volumes Reachable, 0 Volume unreachable in this coralset from localhost

B<Input>

=over 4

=item *  0|1
        If bool turned on, the csmping binary will ping all striped volumes from each
        node. If bool turned off, it will choose a random volume to send a csmping.
        The default is off. 

=back

B<Output>
Array of striped volumes found to be unreachable.

=cut

sub csmping_striped_vols {
	my $self = shift;
	my @errored_stripe_vols;
	my @striped_vols;
	my %params = validate(@_,
				{
				ping_all_vols => { type => BOOLEAN,
				default => 0,
				}});
	my @Nodes =NACL::C::Node->find();
	eval{
	@striped_vols = NACL::CS::Volume->fetch(command_interface=>$Nodes[0],filter=>{'volume-style' => "striped",
											'state' => "online",
											'vsroot' => "false"});
	$Log->comment("dumper of striped vols array".Dumper(@striped_vols));
	};

	if ($@) {
		if ($@ =~ /No\smatching\svolume/) {
			$Log->warn("There were no striped volumes found upon which to execute the csmping test");
		}
		else {
			$Log->error("The 'volume show' command for striped volumes failed due to an unknown error: $@");
		}
		return;
	}

	if (!$params{ping_all_vols}) {
		my $sum_vol = int(@striped_vols);
		my $index = int(rand($sum_vol));
		my $volume = $striped_vols[$index];
		undef(@striped_vols); 
		push @striped_vols, {%$volume};
	}
	foreach my $node (@Nodes) {
		foreach my $vol (@striped_vols) {
			my $api = $node->apiset(category => "Node", interface => "CLI", set => "Systemshell");
			my $csm_ping_result = $api->execute_raw_command( command => "coraltest csmping -a -m $vol->{msid}");
			$csm_ping_result =~ /(\d+)\sVolume\sunreachable/;
			if ($1 > 0) {
				push @errored_stripe_vols, "WARNING: $vol->{vserver}:$vol->{volume} failed coraltest csmping on $node: $&";
			}
		}
	}
	if (@errored_stripe_vols) {
		foreach (@errored_stripe_vols) {
			$Log->warn($_);
		}
		return \@errored_stripe_vols;
	}
}

=pod

=head2 get_cluster_connectivity_errors

C<get_cluster_connectivity_errors>

B<Description>
Used to verify whether cluster connectivity issues could be accountable for CSM problems by using the 'cluster ping-cluster'
command. 
Each error instance found will generate a THARN WARN result.

Example command output:
Host is cqa-6070-13
Getting information from ngsh
Local = 192.168.20.140 192.168.21.140 192.168.20.141 192.168.21.141
Remote = 192.168.20.176 192.168.21.176 192.168.20.178 192.168.22.178 192.168.20.142 192.168.21.142 192.168.20.143 192.168.21.143
Ping status:
32 paths up, 0 paths down at 1500 size
32 paths up, 0 paths down at 4500 size
32 paths up, 0 paths down at 9000 size
RPC status:
8 paths up, 0 paths down (tcp check)
8 paths up, 0 paths down (udp check)

B<Output>
Array of path failures from 'cluster ping-cluster' command.

=cut

sub get_cluster_connectivity_errors {
	my $self = shift;
	my @cluster_connectivity_errors;
	my @Nodes =NACL::C::Node->find();
	my $Pings;
	my  $api = $Nodes[0]->apiset(category => "Node", interface => "CLI", set => "CMode");
	foreach my $ping_from_node (@Nodes) {
		$Log->comment("Node  is".$ping_from_node->node);
		my $node=$ping_from_node->node;
		eval{
			$Pings = $api->execute_raw_command( command => "cluster ping-cluster -node $node -use-sitelist true");
		};
		if ($@) {
			$Log->error("The 'cluster ping cluster' command failed due to an unknown error: $@");
		}
		# Validate Ping Results for paths down for cluster ports 
		while ($Pings =~ /Basic\sconnectivity\sfails\son\s(\d+)\spath/g) {
			if ($1 > 0) {
				push @cluster_connectivity_errors, "WARNING: There were $& packets while performing a cluster ping from $ping_from_node->{node} .";
			}
		}
		# Validate Ping Results for paths down tcp|udp check
		while ($Pings =~ /(\d+)\spaths\sdown\s\((tcp|udp)/g) {
			if ($1 >  0) {
				push @cluster_connectivity_errors, "WARNING: $& while performing a cluster ping from $ping_from_node->{node} .";
			}
		}
	}
	if (@cluster_connectivity_errors) {
		foreach (@cluster_connectivity_errors) {
			$Log->warn($_);
		}
		return \@cluster_connectivity_errors;
	}
}

=pod

=head2 get_cluster_errored_ports

C<get_cluster_errored_ports>

B<Description>
Used to verify network cluster ports are "up", and configured the same throughout
the cluster. 
These can cause an environment that affects the performance, and health of CSM.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the port attribute(s) in error, or misconfigured.

=cut


sub get_cluster_errored_ports {
	my $self = shift;
	my @cluster_errored_ports;
	my @Nodes =NACL::C::Node->find();
	my @Ports;
	eval{
		@Ports = NACL::CS::NetworkPort->fetch(command_interface=>$Nodes[0],filter=>{role => "cluster"});
	};
	if ($@) {
		$Log->error("The 'network port show' command failed due to an unknown error: $@");
	}

	my $compare_a = shift(@Ports);
	foreach my $compare_b (@Ports) {
		if ($compare_a->{mtu} != $compare_b->{mtu}) {
			push @cluster_errored_ports, "ERROR: The MTU setting on cluster port $compare_b->{port}, on $compare_b->{node}"
							."does not match $compare_a->{mtu} which was derived from $compare_a->{port} on"
							."$compare_a->{node}.\n"
		}
	}

	my @ports_down = grep($_->{link} != "up", @Ports);
	foreach (@ports_down) {
		push @cluster_errored_ports, "ERROR: Port $_->{port} on $_->{node} link is in a $_->{link} state.\n";
	}

	if (@cluster_errored_ports) {
		foreach (@cluster_errored_ports) {
			$Log->error($_);
		}
		return \@cluster_errored_ports;
	}
}


=pod

=head2 get_csm_error_counters

C<get_csm_error_counters>

B<Description>
Used to derive errors from specific csm counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero csm error counter.

=cut

sub get_csm_error_counters {
	my @CSM_ERROR_COUNTERS = qw(activeConnFailures activeConnStartFailures connReject_BindToSession connReject_BrokenSession
					connReject_Mallocation connReject_NoBlade connReject_ObsoleteConn connReject_SessionInit
					connReject_SharedSessionExists connReject_UniquifierClash getSessionForDataSetFailRsp 
					getSessionForVifFailRsp registerRcListenAddressFailed);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @csm_counter_error;
	foreach my $node (@Nodes) {
		foreach my $counter_string (@CSM_ERROR_COUNTERS) {
			eval{
				@csm_counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$node,filter=>{category=>"csm",name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			} 
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: CSM Errors from command 'debug csm counters show -category csm -node $_->{node} -name $_->{name}' = $_->{value}");
						push @errored_counters, "WARNING: CSM Errors from command 'debug csm counters show -category csm -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @csm_counter_error;
			}
		}
	}
	return \@errored_counters;
}

=pod

=head2 get_ct_error_counters

C<get_ct_error_counters>

B<Description>
Used to derive errors from specific CT counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero CT error counter.

=cut

sub get_ct_error_counters {
	my @CT_ERROR_COUNTERS = qw(BuildPktNoBuf LoActiveOpenPCPConnError RcvInvalidOp);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @ct_counter_error;
	foreach my $node (@Nodes) {
		foreach my $counter_string (@CT_ERROR_COUNTERS) {
			eval{
				@ct_counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$Nodes[0],filter=>{category=>"ct",name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			} 
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: CT Errors from command 'debug csm counters show -category ct -node $_->{node} -name $_->{name}' = $_->{value}");
						push @errored_counters, "WARNING: CT Errors from command 'debug csm counters show -category ct -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @ct_counter_error;
			}
		}
	}
	return \@errored_counters;
}

=pod

=head2 get_rc_error_counters

C<get_rc_error_counters>

B<Description>
Used to derive errors from specific RC counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero RC error counter.

=cut

sub get_rc_error_counters {
	my @RC_ERROR_COUNTERS = qw(AcceptWaitLoAborted AcceptWaitLoAborted2 BadBuf BadDataVerifierAck BadHdrVerifier BadLen BadMagic
				BadVersion BindRcvdAckBad BindRcvdOpAbort BindRcvdOpAbort2 BindRcvdSeqBad BindSentAckBad BindSentAckBadAbort
				BindSentInvalid BindSentOpAbort BuildPktNoBuf ConnUserWaitDrop ConnUserWaitLoAborted ConnUserWaitLoAborted2
				FastTimerClosedBusyBad ListenOpBad LoActiveOpenRspNoMem LoPassiveOpenNoMem LoRcvRcInputNull LoRemoteCloseAbort
				LoRemoteCloseAbort2 LoSndAbort LocalCloseAbort LocalCloseBadState SyncInvalid SyncOpAbort SyncOpAbort2
				SyncRcvBadDataVerifier SyncRcvBadDataVerifierBind SyncRcvBadDataVerifierBrsp SyncRcvGbufTypeMismatch
				SyncRcvStrangeBuf Unknown);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @rc_counter_error;
	foreach my $node (@Nodes) {
		foreach my $counter_string (@RC_ERROR_COUNTERS) {
			eval{
				@rc_counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$Nodes[0],filter=>{category=>"rc",name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			}
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: RC Errors from command 'debug csm counters show -category rc -node $_->{node} -name $_->{name}' = $_->{value}");
						push @errored_counters, "WARNING: RC Errors from command 'debug csm counters show -category rc -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @rc_counter_error;
			}
		}
	}
	return \@errored_counters;
}

=pod

=head2 get_session_error_counters

C<get_session_error_counters>

B<Description>
Used to derive errors from specific session counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero session error counter.

=cut

sub get_session_error_counters {
	my @SESSION_ERROR_COUNTERS = qw(badchannel badinitop badinterface badprototag badseq badstateforinit badstaterequest
					bindfailed closefailed createfailed createrespfailed marshalfailure negoifacefailed negosessfailed
					noreqforresp spinnp_abort_operation_error spinnp_bind_connection_error spinnp_close_session_error
					spinnp_create_session_error spinnp_get_session_info_error spinnp_gss_auth_begin_error
					spinnp_gss_auth_complete_error spinnp_gss_continue_init_error spinnp_gss_destroy_error
					spinnp_gss_init_error spinnp_negotiate_interface_error spinnp_negotiate_resp_store_error
					spinnp_negotiate_session_error spinnp_null_proc_error spinnp_recovery_complete_error
					spinnp_replay_response_error spinnp_response_ids_error spinnp_set_seq_window_error
					spinnp_unbind_connection_error unbindfailed unmarshalfailure);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @session_counter_error;
	foreach my $node (@Nodes) {
		foreach my $counter_string (@SESSION_ERROR_COUNTERS) {
			eval{
				@session_counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$Nodes[0],filter=>{category=>"session",name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			} 
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: Session Errors from command 'debug csm counters show -category session -node $_->{node} -name $_->{name}' = $_->{value}");
						push @errored_counters, "WARNING: Session Errors from command 'debug csm counters show -category session -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @session_counter_error;
			}
		}
	}
	return \@errored_counters;
}

=pod

=head2 get_nulltx_error_counters

C<get_nulltx_error_counters>

B<Description>
Used to derive errors from nulltx striped counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero nulltx error counter.

=cut

sub get_nulltx_error_counters {
	my @NULLTX_ERROR_COUNTERS = qw(noMemActiveConn noMemPassiveConn);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @nulltx_counter_error;
	foreach my $node (@Nodes) {
		foreach my $counter_string (@NULLTX_ERROR_COUNTERS) {
			eval{
				@nulltx_counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$Nodes[0],filter=>{category=>"nulltx",name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			}
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: Nulltx Errors from command 'debug csm counters show -category nulltx -node $_->{node} -name $_->{name}' = $_->{value}");
						push @errored_counters, "WARNING: Nulltx Errors from command 'debug csm counters show -category nulltx -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @nulltx_counter_error;
			}
		}
	}
	return \@errored_counters;
}

=pod

=head2 get_error_counters

C<get_error_counters>

B<Description>
Used to derive errors from specific counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero cache error counter.

=cut

sub get_error_counters {
	my @CACHE_ERROR_COUNTERS = qw(dsToBladeInvalidates vifToBladeInvalidates);
	my @NULLTX_ERROR_COUNTERS = qw(noMemActiveConn noMemPassiveConn);
	my @SESSION_ERROR_COUNTERS = qw(badchan badchannel badinitop badinterface badprototag badseq badstateforinit badstaterequest
					bindfailed closefailed createfailed createrespfailed marshalfailure negoifacefailed negosessfailed
					noreqforresp spinnp_abort_operation_error spinnp_bind_connection_error spinnp_close_session_error
					spinnp_create_session_error spinnp_get_session_info_error spinnp_gss_auth_begin_error
					spinnp_gss_auth_complete_error spinnp_gss_continue_init_error spinnp_gss_destroy_error
					spinnp_gss_init_error spinnp_negotiate_interface_error spinnp_negotiate_resp_store_error
					spinnp_negotiate_session_error spinnp_null_proc_error spinnp_recovery_complete_error
					spinnp_replay_response_error spinnp_response_ids_error spinnp_set_seq_window_error
					spinnp_unbind_connection_error unbindfailed unmarshalfailure);
	my @RC_ERROR_COUNTERS = qw(AcceptWaitLoAborted AcceptWaitLoAborted2 BadBuf BadDataVerifierAck BadHdrVerifier BadLen BadMagic
					BadVersion BindRcvdAckBad BindRcvdOpAbort BindRcvdOpAbort2 BindRcvdSeqBad BindSentAckBad BindSentAckBadAbort
					BindSentInvalid BindSentOpAbort BuildPktNoBuf ConnUserWaitDrop ConnUserWaitLoAborted ConnUserWaitLoAborted2
					FastTimerClosedBusyBad ListenOpBad LoActiveOpenRspNoMem LoPassiveOpenNoMem LoRcvRcInputNull LoRemoteCloseAbort
					LoRemoteCloseAbort2 LoSndAbort LocalCloseAbort LocalCloseBadState SyncInvalid SyncOpAbort SyncOpAbort2
					SyncRcvBadDataVerifier SyncRcvBadDataVerifierBind SyncRcvBadDataVerifierBrsp SyncRcvGbufTypeMismatch
					SyncRcvStrangeBuf Unknown);
	my @CT_ERROR_COUNTERS = qw(BuildPktNoBuf LoActiveOpenPCPConnError RcvInvalidOp);
	my @CSM_ERROR_COUNTERS = qw(activeConnFailures activeConnStartFailures connReject_BindToSession connReject_BrokenSession
					connReject_Mallocation connReject_NoBlade connReject_ObsoleteConn connReject_SessionInit
					connReject_SharedSessionExists connReject_UniquifierClash getSessionForDataSetFailRsp 
					getSessionForVifFailRsp registerRcListenAddressFailed);
	my @ERROR_COUNTERS;
	my $self=shift; 
	my %params = validate(@_,
			{
				category => { type => SCALAR,
				default => "rc",
				}
			}
			);
	my $category=$params{category};
	SWITCH: {
	($category eq "rc") && (@ERROR_COUNTERS=@RC_ERROR_COUNTERS);
	($category eq "cache") && (@ERROR_COUNTERS=@CACHE_ERROR_COUNTERS);
	($category eq "nulltx") && (@ERROR_COUNTERS=@NULLTX_ERROR_COUNTERS);
	($category eq "session") && (@ERROR_COUNTERS=@SESSION_ERROR_COUNTERS);
	($category eq "ct") && (@ERROR_COUNTERS=@CT_ERROR_COUNTERS);
	($category eq "csm") && (@ERROR_COUNTERS=@CSM_ERROR_COUNTERS);
	}	
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @counter_error;
	foreach my $node (@Nodes){
		foreach my $counter_string (@ERROR_COUNTERS) {
			eval{
				@counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$node,filter=>{category=>$category,name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			}
			else {
				grep {
					if ($_->{value} > 0) {
					$Log->warn("WARNING: $category Errors from command 'debug csm counters show -category $category -node $_->{node} -name $_->{name}' = $_->{value}");
					push @errored_counters, "WARNING: Cache Errors from command 'debug csm counters show -category $category -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @counter_error; 
			}
		}
	}
	return \@errored_counters;
}


=pod

=head2 get_cache_error_counters

C<get_cache_error_counters>

B<Description>
Used to derive errors from specific cache counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero cache error counter.

=cut

sub get_cache_error_counters {
	my @CACHE_ERROR_COUNTERS = qw(dsToBladeInvalidates vifToBladeInvalidates);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @cache_counter_error;
	foreach my $node (@Nodes){
		foreach my $counter_string (@CACHE_ERROR_COUNTERS){
			eval{
				@cache_counter_error =NACL::CS::DebugCsmCounters->fetch(command_interface=>$node,filter=>{category=>"cache",name=>$counter_string}); 
			};
			if ($@) {
				$Log->error("The 'debug csm counters show' command failed for an unknown error: $@");
			}
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: Cache Errors from command 'debug csm counters show -category cache -node $_->{node} -name $_->{name}' = $_->{value}");
						push @errored_counters, "WARNING: Cache Errors from command 'debug csm counters show -category cache -node $_->{node} -name $_->{name}' = $_->{value}";
					}
				} @cache_counter_error; 
			}
		}
	}
	return \@errored_counters;
}


=pod

=head2 get_striped_error_counters

C<get_striped_error_counters>

B<Description>
Used to derive errors from specific striped counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero striped error counter.

=cut

sub get_striped_error_counters {
	my @STRIPED_ERROR_COUNTERS = qw(server_csm_req_stalled server_csm_req_discarded server_csm_req_unmarshal_error
					client_csm_resp_failure client_csm_resp_unmarshal_error client_csm_req_send_failure client_csm_req_send_retried
					client_csm_heartbeat_connect_gaveup client_csm_heartbeat_connect_failure client_csm_multicast_connect_failure
					client_csm_multicast_connect_gaveup server_heartbeat_stalled_wafl_status server_heartbeat_resumed_wafl_status
					server_heartbeat_stalled_adminlock server_heartbeat_resumed_adminlock server_heartbeat_resumed_offline);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @striped_stats_error;
	foreach my $node (@Nodes){
		foreach my $counter_string (@STRIPED_ERROR_COUNTERS){
			eval{
				@striped_stats_error = NACL::CS::Statistics->fetch(command_interface=>$node,filter=>{
											object  => "striped",
											instance  => "striped-statistics",
											node => $node,
											counter => $counter_string,} );
			};
			if ($@) {
				if ($@ =~ /There are no entries matching your query/) {
					$Log->error("Requested striped statistics do not exist. High probability: No striped volumes provisioned in cluster.");
				}
				else {
				$Log->error("The 'statistics show' command failed for an unknown error: $@");
				}
				return;
			}
			else {
				grep {
					if ($_->{value} > 0) {
						$Log->warn("WARNING: Striped Errors from command 'statistics show -category striped -object striped-statistics -node $_->{node} -counter $_->{counter}' = $_->{value}");
						push @errored_counters, "WARNING: Striped Errors from command 'statistics show -category striped -object striped-statistics -node $_->{node} -counter $_->{counter}' = $_->{value}";
					}
				} @striped_stats_error;
			}
		}
	}
	return \@errored_counters;
}

=pod

=head2 get_lif_error_counters

C<get_lif_error_counters>

B<Description>
Used to derive errors from specific lif counters which may pinpoint the root cause of the CSM issue.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages pinpointing the non-zero lif error counter.

=cut

sub get_lif_error_counters {
	my @LIF_ERROR_COUNTERS = qw(recv-errors sent-errors);
	my $self = shift;
	my @errored_counters;
	my @Nodes =NACL::C::Node->find();
	my @lifs;
	my @lif_stats_error;
	foreach my $node (@Nodes) {
		foreach my $counter_string (@LIF_ERROR_COUNTERS) {
			@lifs = NACL::CS::Interface->fetch(command_interface=>$node,filter=>{
												server => $node,
												role => "cluster",
											});
			foreach my $lif (@lifs) {
				eval{
					@lif_stats_error=NACL::CS::Statistics->fetch(command_interface=>$node,filter=>{
					object => "lif",
					instance=> $lif->{server}.":".$lif->{lif},
					counter => $counter_string,});
				};
				if ($@) {
					$Log->error("The 'statistics show' command failed due to an unknown error: $@");
				}
				else {
					grep {
						if ($_->{value} >  0) {
							$Log->warn("WARNING: LIF Errors from command 'statistics show -object lif -instance $_->{instance} -counter $_->{counter}' = $_->{value}");
							push @errored_counters, "WARNING: LIF Errors from command 'statistics show -object lif -instance $_->{instance} -counter $_->{counter}' = $_->{value}";
							$Log->comment("Dumping netstat to logs:\n");
							my $api = $Nodes[0]->apiset(category => "Node", interface => "CLI", set => "Systemshell");
							$api->execute_raw_command( command => "netstat -p udp");
							$api->execute_raw_command( command => "netstat -i");
							$api->execute_raw_command( command => "netstat -s");
						}
					} @lif_stats_error;
				}
			}
		}
	}
	return \@errored_counters;
}



=pod

=head2 get_bad_sessions

C<get_bad_sessions>

B<Description>
Used to find any CSM session that is NOT in an "ESTABLISHED" state.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages reporting any CSM sessions NOT in an "ESTABLISHED" state. Also details
the blade interface endpoints which make up that specific CSM session.

=cut

sub get_bad_sessions {
	my $self = shift;
	my @bad_sessions;
	my @not_established = grep($_->{state} != "ESTABLISHED",@{$self->{_sessions}});
	if (@not_established) {
		foreach my $errored (@not_established) {
			my $initiator = $errored->{'initiator-string'};
			my $initiator_uuid = $errored->{'initiator'};
			my $target_uuid = $errored->{'target'};
			my $target = $errored->{'target-string'};
			my $state = $errored->{state};
			push @bad_sessions, "WARNING: Session from $initiator:$initiator_uuid to $target:$target_uuid is in a $state state!";
		}
		if (@bad_sessions) {
			foreach (@bad_sessions) {
				$Log->warn($_);
			}
			return \@bad_sessions;
		}
	} 
	else {
		return undef; 
	}
}


=pod

=head2 get_troubled_sessions

C<get_troubled_sessions>

B<Description>
Used to find CSM related stats which might reveal intermittent, or underlying issues in the cluster.
Each error instance found will generate a THARN WARN result.

B<Output>
Array of warning messages reporting CSM sessions performance dispairity. Also details
the blade interface endpoints which make up that specific CSM session.

=cut

sub get_troubled_sessions {
	my $self=shift; 
	my @troubled_sessions;
	my @outstanding_resp = grep(($_->{'req-rcvd'} != $_->{'resp-sent'}),@{$self->{_sessions}});
	if (@outstanding_resp) {
		foreach my $troubled (@outstanding_resp) {
			my $req_rcvd = $troubled->{'req-rcvd'};
			my $resp_sent = $troubled->{'resp-sent'};
			my $target = $troubled->{'target-string'};
			my $target_uuid = $troubled->{'target'};
			my $initiator = $troubled->{'initiator-string'};
			my $initiator_uuid = $troubled->{'initiator'};
			push @troubled_sessions, "WARNING: Requests received: $req_rcvd do not equal Responses sent: $resp_sent.\n"
						."$initiator:$initiator_uuid is having problems communicating with $target:$target_uuid.\n";
		}
	}

	my @outstanding_req = grep(($_->{'req-sent'} != $_->{'resp-rcvd'}),@{$self->{_sessions}});
	if (@outstanding_req) {
		foreach my $troubled (@outstanding_req) {
			my $resp_rcvd = $troubled->{'resp-rcvd'};
			my $req_sent = $troubled->{'req-sent'};
			my $initiator = $troubled->{'initiator-string'};
			my $initiator_uuid = $troubled->{'initiator'};
			my $target = $troubled->{'target-string'};
			my $target_uuid = $troubled->{'target'};
			push @troubled_sessions, "WARNING: Responses received: $resp_rcvd does not equal Requests sent: $req_sent.\n"
						."$initiator:$initiator_uuid is having problems communicating with $target:$target_uuid.\n";
		}
	}
	if (@troubled_sessions) {
		foreach (@troubled_sessions) {
			$Log->warn($_);
		}
		return \@troubled_sessions;
	}
	else {
		return undef;
	}
}


=pod

=head2 SessionCreateCheck 

C<SessionCreateCheck>

B<Description>
Checks if the test session create was successful 
Queries the status of the create commnd to gather stats
Calls NumConnectionsCreatedCheck to ensure that the expected number of connections were created

B<Input>

=over 4

=item * B<node1>
	Node that originated the create 

=item * B<node2> 
	Passive node 

=item * B<response>
	Status response from the create command

=item * B<connectionType>
        Type of connection

=over 4

=item * B<NullTx> 
	Null Transport connection 

=item * B<Rc>
	Rc Connection	

=item * B<Ct>
	Ct connection        

=item * B<IntraclusterCt>
	Intra-cluster Ct connection        

=back

=item * B<bombPointCheck>
	Indicates if the status of the create commnd is being checked after a bomb point is set      

=over 4

=item * B<yes> 
	Checking create status after setting a bomb point 

=item * B<no>
	Checking create status independent of any bomb point setting

=back

=back

B<Output>
Returns the uniquifier of the session if no bomb point setting in involved, otherwise returns the status of the create command

=cut

sub SessionCreateCheck{
	my $class=shift;
	my %params = validate(@_,
			{
				node1=>{type=>HASHREF},
				node2=>{type=>HASHREF, optional=>1},
				response=>{type=>SCALAR},
				connectionType=>{type=>SCALAR},
				bombPointCheck=>{type=>SCALAR,optional=>1,default=>"no"},
				minNumConnectionsLow=>{type=>SCALAR,optional=>1,default=>0},
				minNumConnectionsMed=>{type=>SCALAR,optional=>1,default=>0},
				minNumConnectionsHigh=>{type=>SCALAR,optional=>1,default=>0},
			});

	my $node1=$params{node1};
	defined($params{node2}) && (my $node2=$params{node2});
	my $api1 = $node1->apiset(category => "Node", interface => "CLI", set => "CMode");
	defined($params{node2}) && (my $api2 = $node2->apiset(category => "Node", interface => "CLI", set => "CMode"));
	my $Response=$params{response};
	my $connectionType=$params{connectionType};
	my $filer=$params{node1}->{node};
	my $bombPointCheck=$params{bombPointCheck};
	my $minNumConnectionsLow=$params{minNumConnectionsLow};
	my $minNumConnectionsMed=$params{minNumConnectionsMed};
	my $minNumConnectionsHigh=$params{minNumConnectionsHigh};
	my $createRequestID;
	if($Response=~/${filer}\s+(\d+)\s+CSM_OK/) {
		$Log->comment("############# Detected CSM OK....$1 ################");
		$createRequestID=$1;
	}
	else {
		if($bombPointCheck eq "yes"){ return $Response;} else {	$Log->error("Did not detect CSM OK..");}
	}
	my $statusResponseOnCreate;
	my $uniquifier;
	my $delta;
	do {
		$statusResponseOnCreate = $api1->execute_raw_command(command=>"debug csm session status -node $filer -ui-req-id $createRequestID");
	}while($statusResponseOnCreate=~/pending/);

	$statusResponseOnCreate=~tr/X;"/ /;
	$Log->comment("#################### Status response on Create: $statusResponseOnCreate ###############");
	if($bombPointCheck eq "yes") { ### RETURN status response after create for bomb points to check the ba;ue
		return $statusResponseOnCreate;
	}
	if($statusResponseOnCreate=~/${createRequestID}\s+0\s+create\s+(\w+).*\s+(\d+)\s+complete\s+CSM_OK/) {
			$uniquifier=$1;
			$delta=$2;
			$Log->comment("################## The status of the create command is complete...The time it took was $delta microseconds, and the session uniquifier is $uniquifier ############");
	}
	else {
		$Log->error("Error..The create request was not successful....Response from the create request: $statusResponseOnCreate");
	}

	if(defined($params{node2})) {
		NumConnectionsCreatedCheck($class,node1=>$node1,node2=>$node2,connectionType=>$connectionType,uniquifier=>$uniquifier,minNumConnectionsLow=>$minNumConnectionsLow,minNumConnectionsMed=>$minNumConnectionsMed,minNumConnectionsHigh=>$minNumConnectionsHigh);
	}

	else {
		NumConnectionsCreatedCheck($class,node1=>$node1,connectionType=>$connectionType,uniquifier=>$uniquifier,minNumConnectionsLow=>$minNumConnectionsLow,minNumConnectionsMed=>$minNumConnectionsMed,minNumConnectionsHigh=>$minNumConnectionsHigh);
	}
	return $uniquifier;

}




=pod

=head2 NumConnectionsCreatedCheck 

C<NumConnectionsCreatedCheck>

B<Description>
Checks if the number of connections were created as expected 
Only one NullTx connection is created at each priority
Number of Rc connections at each priority the maxium number of cluster lifs on the node with the maximum number of cluster lifs
Number of Ct connections at each priority the maxium number of intercluster lifs on the node with the maximum number of intercluster lifs
Number of IntraclusterCt connections at each priority the maxium number of cluster lifs on the node with the maximum number of cluster lifs

B<Input>

=over 4

=item * B<node1>
	Node that originated the send 

=item * B<node2> 
	Passive node 

=item * B<uniquifier>
	Uniquifier that uniquely identifies the created session

=item * B<connectionType>
        Type of connection

=over 4

=item * B<NullTx> 
	Null Transport connection 

=item * B<Rc>
	Rc Connection	

=item * B<Ct>
	Ct connection        

=item * B<IntraclusterCt>
	Intra-cluster Ct connection        

=back

=back

B<Output>

Does not return a value. Logs warnings if the number of connections created is not as expected.

=cut

sub NumConnectionsCreatedCheck {

	my $class=shift;
	my %params = validate(@_,
			{
			node1=>{type=>HASHREF},
			node2=>{type=>HASHREF, optional=>1},
			connectionType=>{type=>SCALAR},
			uniquifier=>{type=>SCALAR},
			minNumConnectionsLow=>{type=>SCALAR,optional=>1,default=>0},
			minNumConnectionsMed=>{type=>SCALAR,optional=>1,default=>0},
			minNumConnectionsHigh=>{type=>SCALAR,optional=>1,default=>0},
			});

	my $node1=$params{node1};
	my $node2=$params{node2};
	my $connectionType=$params{connectionType};
	my $uniquifier=$params{uniquifier};
	my $minNumConnectionsLow=$params{minNumConnectionsLow};
	my $minNumConnectionsMed=$params{minNumConnectionsMed};
	my $minNumConnectionsHigh=$params{minNumConnectionsHigh};
	my $clusterPorts;
	my @clusterPorts1=NACL::CS::NetworkInterface->fetch(command_interface=>$node1,filter=>{role=>"cluster",vserver=>$node1->{node}});
	$Log->comment("#############3 Number of Cluster ports on node 1: ". int(@clusterPorts1));	
	if(defined($node2)) {
		my @clusterPorts2=NACL::CS::NetworkInterface->fetch(command_interface=>$node2,filter=>{role=>"cluster",vserver=>$node2->{node}});
		$Log->comment("#############3 Number of Cluster ports on node 2: ". int(@clusterPorts2));
		$clusterPorts=int(@clusterPorts1)>=int(@clusterPorts2)?int(@clusterPorts1):int(@clusterPorts2);
	}

	else {
		$clusterPorts=int(@clusterPorts1);
	}

	switch($connectionType) {
		case /NullTx/ {

			my @LowPriorityConnections= NACL::CS::DebugCsmNulltx->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"low"});
			my @MediumPriorityConnections= NACL::CS::DebugCsmNulltx->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"medium"});
			my @HighPriorityConnections= NACL::CS::DebugCsmNulltx->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"high"});
			$Log->comment("Number of low priority connections created ".int(@LowPriorityConnections));	
			$Log->comment("Number of medium priority connections created ".int(@MediumPriorityConnections));	
			$Log->comment("Number of high priority connections created ".int(@HighPriorityConnections));	
			(int(@LowPriorityConnections)!=2) && $Log->warn("Number of low priority connections established was not equal to the number of cluster lifs");
			(int(@MediumPriorityConnections)!=2) && $Log->warn("Number of medium priority connections established was not equal to the number of cluster lifs");
			(int(@HighPriorityConnections)!=2) && $Log->warn("Number of high priority connections established was not equal to the number of cluster lifs");
		}
		case /Rc/ {
			my $lowPriorityCxns=$minNumConnectionsLow>$clusterPorts?$minNumConnectionsLow:$clusterPorts;
			my $medPriorityCxns=$minNumConnectionsMed>$clusterPorts?$minNumConnectionsMed:$clusterPorts;
			my $highPriorityCxns=$minNumConnectionsHigh>$clusterPorts?$minNumConnectionsHigh:$clusterPorts;


			my @LowPriorityConnections= NACL::CS::DebugCsmRc->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"low",state=>"ESTABLISHED"});
			my @MediumPriorityConnections= NACL::CS::DebugCsmRc->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"medium",state=>"ESTABLISHED"});
			my @HighPriorityConnections= NACL::CS::DebugCsmRc->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"high",state=>"ESTABLISHED"});

			$Log->comment("Number of low priority connections created ".int(@LowPriorityConnections));	
			$Log->comment("Number of medium priority connections created ".int(@MediumPriorityConnections));	
			$Log->comment("Number of high priority connections created ".int(@HighPriorityConnections));	
			(($lowPriorityCxns*2)!=int(@LowPriorityConnections)) && $Log->warn("Number of low priority connections established was not equal to the number of cluster lifs");
			(($medPriorityCxns*2)!=int(@MediumPriorityConnections)) && $Log->warn("Number of medium priority connections established was not equal to the number of cluster lifs");
			(($highPriorityCxns*2)!=int(@HighPriorityConnections)) && $Log->warn("Number of high priority connections established was not equal to the number of cluster lifs");
		}

		case /IntraclusterCt/ { #having to log into both nodes because of burt 520038
			
			


			my @LowPriorityConnections= NACL::CS::DebugCsmCt->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"low",state=>"ESTABLISHED",node=>$node1->{node}});
  			my @MediumPriorityConnections= NACL::CS::DebugCsmCt->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"medium",state=>"ESTABLISHED",node=>$node1->{node}});
  			my @HighPriorityConnections= NACL::CS::DebugCsmCt->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"high",state=>"ESTABLISHED",node=>$node1->{node}});
			$Log->comment("Number of low priority connections created ".int(@LowPriorityConnections));	
			$Log->comment("Number of medium priority connections created ".int(@MediumPriorityConnections));	
			$Log->comment("Number of high priority connections created ".int(@HighPriorityConnections));	
    	
  			my @LowPriorityConnectionsRemote= NACL::CS::DebugCsmCt->fetch(command_interface=>$node2,filter=>{uniquifier=>$uniquifier,priority=>"low",state=>"ESTABLISHED",node=>$node2->{node}});
  			my @MediumPriorityConnectionsRemote= NACL::CS::DebugCsmCt->fetch(command_interface=>$node2,filter=>{uniquifier=>$uniquifier,priority=>"medium",state=>"ESTABLISHED",node=>$node2->{node}});
  			my @HighPriorityConnectionsRemote= NACL::CS::DebugCsmCt->fetch(command_interface=>$node2,filter=>{uniquifier=>$uniquifier,priority=>"high",state=>"ESTABLISHED",node=>$node2->{node}});
			$Log->comment("Number of remote low priority connections created ".int(@LowPriorityConnectionsRemote));	
			$Log->comment("Number of remote medium priority connections created ".int(@MediumPriorityConnectionsRemote));	
			$Log->comment("Number of remote high priority connections created ".int(@HighPriorityConnectionsRemote));	
			# DOING THIS BECAUSE OF BURT 520038 #  			
			my @clusterPorts1=NACL::CS::NetworkInterface->fetch(command_interface=>$node1,filter=>{role=>"cluster",vserver=>$node1->{node}});
			my @clusterPorts2=NACL::CS::NetworkInterface->fetch(command_interface=>$node2,filter=>{role=>"cluster",vserver=>$node2->{node}});
			$Log->comment("#############3 Number of  Cluster ports on node 1: ". int(@clusterPorts1));	
			$Log->comment("#############3 Number of  Cluster ports on node 2: ". int(@clusterPorts2));
			$clusterPorts=int(@clusterPorts1)>=int(@clusterPorts2)?int(@clusterPorts1):int(@clusterPorts2);

			my $lowPriorityCxns=$minNumConnectionsLow>$clusterPorts?$minNumConnectionsLow:$clusterPorts;
			my $medPriorityCxns=$minNumConnectionsMed>$clusterPorts?$minNumConnectionsMed:$clusterPorts;
			my $highPriorityCxns=$minNumConnectionsHigh>$clusterPorts?$minNumConnectionsHigh:$clusterPorts;

			(($lowPriorityCxns*2)!=int(@LowPriorityConnections)+int(@LowPriorityConnectionsRemote)) && $Log->warn("Number of low priority connections established was not equal to the number of cluster lifs");
			(($medPriorityCxns*2)!=int(@MediumPriorityConnections)+int(@MediumPriorityConnectionsRemote))  && $Log->warn("Number of medium priority connections established was not equal to the number of cluster lifs");
			(($highPriorityCxns*2)!=int(@HighPriorityConnections)+int(@HighPriorityConnectionsRemote))  && $Log->warn("Number of high priority connections established was not equal to the number of cluster lifs");
		}

		case /Ct/ {

			my @LowPriorityConnections= NACL::CS::DebugCsmCt->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"low",state=>"ESTABLISHED"});
			my @MediumPriorityConnections= NACL::CS::DebugCsmCt->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"medium",state=>"ESTABLISHED"});
			my @HighPriorityConnections= NACL::CS::DebugCsmCt->fetch(command_interface=>$node1,filter=>{uniquifier=>$uniquifier,priority=>"high",state=>"ESTABLISHED"});
			$Log->comment("Number of low priority connections created ".int(@LowPriorityConnections));	
			$Log->comment("Number of medium priority connections created ".int(@MediumPriorityConnections));	
			$Log->comment("Number of high priority connections created ".int(@HighPriorityConnections));	
    	
			my @LowPriorityConnectionsRemote= NACL::CS::DebugCsmCt->fetch(command_interface=>$node2,filter=>{uniquifier=>$uniquifier,priority=>"low",state=>"ESTABLISHED"});
			my @MediumPriorityConnectionsRemote= NACL::CS::DebugCsmCt->fetch(command_interface=>$node2,filter=>{uniquifier=>$uniquifier,priority=>"medium",state=>"ESTABLISHED"});
			my @HighPriorityConnectionsRemote= NACL::CS::DebugCsmCt->fetch(command_interface=>$node2,filter=>{uniquifier=>$uniquifier,priority=>"high",state=>"ESTABLISHED"});
			$Log->comment("Number of remote low priority connections created ".int(@LowPriorityConnectionsRemote));	
			$Log->comment("Number of remote medium priority connections created ".int(@MediumPriorityConnectionsRemote));	
			$Log->comment("Number of remote high priority connections created ".int(@HighPriorityConnectionsRemote));	
			my @interclusterPorts1=NACL::CS::NetworkInterface->fetch(command_interface=>$node1,filter=>{role=>"intercluster",vserver=>$node1->{node}});
			my @interclusterPorts2=NACL::CS::NetworkInterface->fetch(command_interface=>$node2,filter=>{role=>"intercluster",vserver=>$node2->{node}});
			$Log->comment("############# Number of Inter Cluster ports on node 1: ". int(@interclusterPorts1));	
			$Log->comment("############# Number of Inter Cluster ports on node 2: ". int(@interclusterPorts2));

			my $interclusterPorts=int(@interclusterPorts1)>=int(@interclusterPorts2)?int(@interclusterPorts1):int(@interclusterPorts2);
			my $lowPriorityCxns=$minNumConnectionsLow>$interclusterPorts?$minNumConnectionsLow:$interclusterPorts;
			my $medPriorityCxns=$minNumConnectionsMed>$interclusterPorts?$minNumConnectionsMed:$interclusterPorts;
			my $highPriorityCxns=$minNumConnectionsHigh>$interclusterPorts?$minNumConnectionsHigh:$interclusterPorts;
			(($lowPriorityCxns*2)!=(int(@LowPriorityConnections)+int(@LowPriorityConnectionsRemote))) && $Log->warn("Number of low priority connections established was not as expected");
			(($medPriorityCxns*2)!=(int(@MediumPriorityConnections)+int(@MediumPriorityConnectionsRemote))) && $Log->warn("Number of med  priority connections established was not as expected");
			(($highPriorityCxns*2)!=(int(@HighPriorityConnections)+int(@HighPriorityConnectionsRemote))) && $Log->warn("Number of high  priority connections established was not  as expected");
		}

	} #end of switch

}

=pod

=head2 SessionSendTest 

C<SessionSendTest>

B<Description>
Sends requests over a created session
Checks the status of the send command
Tracks statistics of the send command 

B<Input>

=over 4

=item * B<node>
	Node originating the send 

=item * B<priority> 
	Passive node 

=over 4

=item * B<low> 
	low priority connection
 
=item * B<medium> 
	low priority connection

=item * B<high> 
	low priority connection

=back

=item * B<uniquifier>
	Uniquifier that uniquely identifies the created session

=item * B<requestSize>
	Size of the request to be sent        

=item * B<responseSize>
	Size of the response        

=item * B<responseDelayMs> 
	Number of milliseconds to delay subsequent responses by 

=item * B<requestDelayMs> 
	Number of milliseconds to delay subsequent requests by 

=item * B<timeoutMs>
	Timeout in Milliseconds for the send command	

=item * B<numPackets>
	Number of requests to send        

=item * B<trackEachRequest>
	Specifies if a status table entry has to be created for each request        

=over 4

=item * B<true> 
	Create a status table entry for each request
 
=item * B<false> 
	Do not create a status table entry for each request

=back

=back

B<Output>

If send was successful, method returns a hash with statistics of the send command- like backpressure count, success percent, rtt values

=cut

sub SessionSendTest{
	my $class=shift; 
	my %params = validate(
				@_,
				{
				node=>{type=>HASHREF},
				priority=>{type=>SCALAR, optional=>1, default=>"low"},
				requestSize=>{type=>SCALAR, optional=>1, default=>"0"},
				responseSize=>{type=>SCALAR, optional=>1, default=>"0"},
				responseDelayMs=>{type=>SCALAR, optional=>1, default=>"10"},
				requestDelayMs=>{type=>SCALAR, optional=>1, default=>"10"},
				timeoutMs=>{type=>SCALAR, optional=>1, default=>"30000"},
				numPackets=>{type=>SCALAR, optional=>1, default=>"1"},
				uniquifier=>{type=>SCALAR},
				bombPointCheck=>{type=>SCALAR,optional=>1,default=>"no"},
				trackEachRequest=>{type=>SCALAR,optional=>1,default=>"false"}
				});
	my $bombPointCheck=$params{bombPointCheck};
	my $node=$params{node};
	my $filer=$params{node}->{node};
	my $api = $node->apiset(category => "Node", interface => "CLI", set => "CMode");
	my $priority=$params{priority};
	my $numPackets=$params{numPackets};
	my $requestSize=$params{requestSize};
	my $responseSize=$params{responseSize};
	my $responseDelayMs=$params{responseDelayMs};
	my $requestDelayMs=$params{requestDelayMs};
	my $timeoutMs=$params{timeoutMs};
	my $numPackets=$params{numPackets};
	my $uniquifier=$params{uniquifier};
	my $trackEachRequest=$params{trackEachRequest};
	my $sendRequestID;
	#removing request delay ms as it does not llow me to simulate backpressures, and high packet counts brings nodes down due to out of memory error
	my $sendResponse=$api->execute_raw_command(command=>"debug csm session send -node $node->{node} -uniquifier $uniquifier -priority $priority -request-size $requestSize -response-size $responseSize -response-delay-ms $responseDelayMs -count $numPackets -timeout-ms $timeoutMs -track-each-request $trackEachRequest");
	my %sendStats;
	if($sendResponse=~/${filer}\s+(\d+)\s+CSM_OK/) {
		$Log->comment("############## Detected CSM OK...".$1."####################");
		$sendRequestID=$1;
	}
	else { 
		$Log->error("Did not detect CSM OK..");
	}
	my $statusResponseOnSend;  
	my $sendDelta;
	my $count;
	do {
		$statusResponseOnSend = $api->execute_raw_command(command=>"debug csm session status -node $filer -ui-req-id $sendRequestID");
	}while($statusResponseOnSend=~/pending/);
	$statusResponseOnSend=~tr/X;"/ /;
	$Log->comment("############### Status response on send..".$statusResponseOnSend."############### ");
	if($bombPointCheck eq "yes") {
		return $statusResponseOnSend;
	} 
	if($statusResponseOnSend=~/${sendRequestID}\s+0\s+send\s+${uniquifier}.*\s+(\d+)\s+complete\s+count:\s(\d+)/) {
		$sendDelta=$1;
		$count=$2;		
		if($count > 0) {		
			$Log->comment("################### The status of the send  command is complete...The time it took to send $count packets was $sendDelta ms, and the session uniquifier is $uniquifier ################");
		}
		if($statusResponseOnSend=~/CSM_OK:\s(\d+)/) {
			$sendStats{csmOK}=$1;
			$sendStats{csmOK}>1 && $Log->comment("############### $sendStats{csmOK} requests were successfully sent ###############");	

		}
		if($statusResponseOnSend=~/CSM_BACKPRESSURE:\s(\d+)/) {
			$sendStats{csmBACKPRESSURE}=$1;
			$Log->comment("################### $sendStats{csmBACKPRESSURE} requests were subjected to backpressure ################");
		}	
		if($statusResponseOnSend=~/CSM_INVALID:\s(\d+)/) {
			$sendStats{csmINVALID}=$1;
			$Log->warn("############ $sendStats{csmINVALID} requests were flagged invalid #############");
                }	
		if($statusResponseOnSend=~/CSM_TIMEMOUT:\s(\d+)/) {
			$sendStats{csmTIMEOUT}=$1;
			$Log->warn("############ $sendStats{csmTIMEOUT} requests timed out #############");
                }	
		if($statusResponseOnSend=~/(\d+)%\s+success/) {
			$sendStats{successPercent}=$1;
			$Log->comment("############ Success percent of the send command is $sendStats{successPercent} ##################");
                }	
		if($statusResponseOnSend=~/total\stime:\s+(\d+)\s(\w+),\stotal/) { 
			$sendStats{totalTime}=$1;
			$sendStats{totalTimeUnits}=$2;
			$Log->comment("################ Total time taken for the send is $sendStats{totalTime} $sendStats{totalTimeUnits} ################");
		}

		if($statusResponseOnSend=~/total\ssent\/received:\s+(\d+)\/(\d+)\sbytes/) { 
			$sendStats{totalBytesSent}=$1;
			$sendStats{totalBytesReceived}=$2;
			$Log->comment("############### Total bytes sent were $sendStats{totalBytesSent} and total bytes received were $sendStats{totalBytesReceived} #############");
		}
		if($statusResponseOnSend=~/transfer\srate\ssend\/receive:\s+(\d+)\/(\d+)\sbytes/) { 
			$sendStats{sendTransferRate}=$1;
			$sendStats{receiveTransferRate}=$2;
			$Log->comment("############### Send Transfer Rate is $sendStats{sendTransferRate} bytes/sec and Receive transfer rate is $sendStats{receiveTransferRate} bytes/sec ###############");
		}
		if($statusResponseOnSend=~/rtt\s+min\/avg\/max:\s+(\d+)\.(\d+)\/(\d+)\.(\d+)\/(\d+)\.(\d+)\s+(\D+)/) { 
			$sendStats{rttMin}=$1.".".$2;
			$sendStats{rttAvg}=$3.".".$4;
			$sendStats{rttMax}=$5.".".$6;
			$sendStats{rttUnits}=$7;
			$Log->comment("################## The minimum, average and maximum RTT values are $sendStats{rttMin} $sendStats{rttUnits} , $sendStats{rttAvg} $sendStats{rttUnits} and $sendStats{rttMax} $sendStats{rttUnits} respectively ##################");
		}
		
              	return \%sendStats;

	}
    	else
	{
		$Log->error("Send requestwas not completed successfully... $statusResponseOnSend"); 
	}

   }

=pod

=head2 ClobberTest 

C<ClobberTest>

B<Description>
Issues a clobber command for a particular session
Checks to see if the session has been clobbered

B<Input>

=over 4

=item * B<node>
	Node whose session has to be clobbered 

=item * B<remoteNode>
	Remote Node whose session has to be clobbered 
	Applicable only for a CT connection

=item * B<uniquifier>
	Uniquifier that uniquely identifies the created session

=back


B<Output>

If the clobber command was unsuccessful, this method fails the test by flagging an error

=cut

#### Clobber session and confirm if clobber was successful ####
sub ClobberTest{
	my $class=shift; 
	my %params = validate(
	@_,
	{ 
	node=>{type=>HASHREF},
	uniquifier=>{type=>SCALAR},
	remoteNode=>{type=>HASHREF,optional=>1}
	});
	my $node=$params{node};
	my $filer=$node->{node};
	my $uniquifier=$params{uniquifier};
	my $clobberSession= NACL::C::DebugCsmSession->clobber(command_interface=>$node,node=>$filer,uniquifier=>$uniquifier);
	my @sessionShow=NACL::CS::DebugCsmSession->fetch(command_interface=>$node);
	my $unclobberedSession=0;
	foreach my $row(@sessionShow) {
		if(($row->{uniquifier} eq $uniquifier) && ($row->{state} eq "ESTABLISHED")) {
			$Log->error("Session with uniquifier $uniquifier could not be clobbered successfully");
			$unclobberedSession=1;
		}
	}
	if(defined($params{remoteNode})) {
		my $remoteNode=$params{remoteNode};
		my $remoteFiler=$remoteNode->{node};
		my @remoteSessionShow=NACL::CS::DebugCsmSession->fetch(command_interface=>$remoteNode);
		foreach my $row(@remoteSessionShow) {
			if($row->{uniquifier} eq $uniquifier && $row->{state} eq "ESTABLISHED") {
				$Log->error("Session with uniquifier $uniquifier could not be clobbered successfully");
				$unclobberedSession=1;
			}
		}
	}
	if ($unclobberedSession==0) {
		$Log->comment("############### Session clobber was successful #####################");
	}
}

=pod

=head2 SetFlowControl 

C<SetFlowControl>

B<Description>
Sets flow control for a particular blade

B<Input>

=over 4

=item * B<node>
	Node on which the flow control setting has to be made 

=item * B<blade>
	Blade for which flow control has to be set` 

=item * B<totalPackets>
	Total packets that can be sent up to the blade	

=item * B<medHiReserved>
	Packets reserved for medium and high priority only	

=item * B<highReserved>
	Packets reserved for high priority only	

=back

=cut

sub SetFlowControl {
	my $class=shift;
	my %params = validate(
				@_,
				{ 
				node=> {type => HASHREF},
				blade=>{ type=> SCALAR, default=>"Dblade",optional=>1}, 
				totalPackets=>{ type=> SCALAR, default=>0, optional=>1}, 
				medHiReserved=>{ type=> SCALAR, default=>0,optional=>1 }, 
				highReserved=>{ type=> SCALAR, default=>0,optional=>1}, 
				}
			);

	my $Node=$params{node};  
	my $filer=$Node->{node}; 
	my $api = $Node->apiset(category => "Node", interface => "CLI", set => "CMode");
	my $Response = $api->execute_raw_command(command=>'set -confirmations off');
	my $Response = $api->execute_raw_command(command=>'set test');
	$api->execute_raw_command(command=>"debug csm set-flow-control -node $filer -blade $params{blade} -total-packets $params{totalPackets} -medhi-reserved $params{medHiReserved} -hi-reserved $params{highReserved}");
         
	
} #End of Method



=pod

=head2 FlowControlStats 

C<FlowControlStats>

B<Description>
Retrieves the flow control for a particular blade

B<Input>

=over 4

=item * B<node>
	Node on which the flow control setting has to be made 

=item * B<blade>
	Blade for which flow control has to be set` 

=item * B<interfaceId>
	Interface for which the flow control stats are to be retrieved	

=back

B<Output>
Returns a has ref to the flow control stats

=cut

sub FlowControlStats {
	# replace interfacename 13 once burt is fixed
	my $class=shift;
	my %params = validate(@_,
				{ 
			blade=>{type=>SCALAR,default=>"Dblade"}, 
			interfaceId=>{type=>SCALAR,default=>"13"}, 
			node=> {type => HASHREF},
			});
	my $Node=$params{node};
	my $blade=$params{blade};
	my $interfaceId=$params{interfaceId};
	my $filer=$Node->{node}; 
	my $totalFC=NACL::CS::DebugCsmLocalBlade->fetch(command_interface=>$Node,filter=>{'interface-id'=>"13",'node'=>$filer,'blade-type'=>"Dblade"});
	$Log->comment("####################### TOTAL FLOW CONTROLLED PACKETS : $totalFC->{'total'} ###########################");
	$Log->comment("####################### PACKETS FLOW CONTROLLED AT LOW PRIORITY : $totalFC->{'low_priority_packets'} ###########################");
	$Log->comment("####################### PACKETS FLOW CONTROLLED AT MEDIUM PRIORITY : $totalFC->{'med_priority_packets'} ###########################");
	$Log->comment("####################### PACKETS FLOW CONTROLLED AT HIGH PRIORITY : $totalFC->{'high_priority_packets'} ###########################");
 
	return $totalFC; 
} #End of Flow Control Stats 

=pod

=head2 SessionCreate 

C<SessionCreate>

B<Description>
Attempts to create a session over a particular connection
Returns the status of the create commanD 

B<Input>

=over 4

=item * B<node1>
	Node that originated the create 

=item * B<node2> 
	Passive node 

=item * B<connectionType>
        Type of connection

=over 4

=item * B<NullTx> 
	Null Transport connection 

=item * B<Rc>
	Rc Connection	

=item * B<Ct>
	Ct connection        

=item * B<IntraclusterCt>
	Intra-cluster Ct connection        

=back

=item * B<numChannelsLow> 
	Number of low priority channels to be created (default :500 ) 

=item * B<numChannelsMed> 
	Number of medium priority channels to be created (default :30 ) 

=item * B<numChannelsHigh> 
	Number of high priority channels to be created (default :30 ) 

=item * B<disableCongestionControl> 
	Specifies if session layer congestion control has to be disabled

=over 4

=item * B<true>
	Disable session layer congestion control	 

=item * B<false> default
	Do not disable session layer congestion control	 

=back

=item * B<windowSizeLow> 
	Low priority window size (default :100 ) 

=item * B<windowSizeMed> 
	Med priority window size (default :100 ) 

=item * B<windowSizeHigh> 
	High priority window size (default :100 ) 

=item * B<numChannelsMed> 
	Number of medium priority channels to be created (default :30 ) 

=item * B<numChannelsHigh> 
	Number of high priority channels to be created (default :30 ) 

=item * B<sourceBlade>
	Blade from which the session has to be created 

=over 4

=item * B<Dblade>
	Dblade (default)

=item * B<Nblade>
	Nblade 

=item * B<SCSI blade>
	SCSI blade

=item * B<SpinVFS>
	SpinVFS blade 

=back

=item * B<destBlade>
	Blade to which the session has to be created 

=over 4

=item * B<Dblade>
	Dblade (default)

=item * B<Nblade>
	Nblade 

=item * B<SCSI blade>
	SCSI blade

=item * B<SpinVFS>
	SpinVFS blade 

=back

=item * B<timeoutMs>
	Number of milliseconds for the create command to time out 

=item * B<minNumConnectionsLow>
	Minimum number of connections to be created for this session at low priority 

=item * B<minNumConnectionsMedium>
	Minimum number of connections to be created for this session at medium priority 

=item * B<minNumConnectionsHigh>
	Minimum number of connections to be created for this session at high priority 

=back

B<Output>
Returns the status response the create command

=cut

sub SessionCreate {
	my $class=shift;
	my %params = validate(
			@_,
			{ 
			connectionType=>{type=>SCALAR},
			node1=> {type => HASHREF, optional=>1},
			node2=>{type=>HASHREF, optional=>1},
			numChannelsLow=>{ type=> SCALAR, default=>500, optional=>1 }, 
			numChannelsMed=>{ type=> SCALAR, default=>30, optional=>1  }, 
			numChannelsHigh=>{ type=> SCALAR, default=>30, optional=>1   }, 
			disableCongestionControl=>{ type=> SCALAR, default=>"false" , optional=>1  }, 
			windowSizeLow=>{ type=> SCALAR, default=>100 , optional=>1  }, 
			windowSizeMed=>{ type=> SCALAR, default=>100 , optional=>1  }, 
			windowSizeHigh=>{ type=> SCALAR, default=>100, optional=>1  }, 
			sourceBlade=>{ type=> SCALAR, default=>"Dblade", optional=>1  }, 
			destBlade=>{ type=> SCALAR, default=>"Dblade", optional=>1  }, 
			timeoutMs=>{ type=> SCALAR, default=>"30000", optional=>1  },
			minNumConnectionsLow=>{type=>SCALAR,default=>0,optional=>1},
      			minNumConnectionsMed=>{type=>SCALAR,default=>0,optional=>1},
      			minNumConnectionsHigh=>{type=>SCALAR,default=>0,optional=>1}, 
		}
	);
	my $numChannelsLow=$params{numChannelsLow};
	my $numChannelsMed=$params{numChannelsMed};
	my $numChannelsHigh=$params{numChannelsHigh};
	my $disableCongestionControl=$params{disableCongestionControl};
	my $windowSizeLow=$params{windowSizeLow};
	my $windowSizeMed=$params{windowSizeMed};
	my $windowSizeHigh=$params{windowSizeHigh};
	my $connectionType=$params{connectionType};
	my $sourceBlade=$params{sourceBlade};
	my $destBlade=$params{destBlade};
	my $timeoutMs=$params{timeoutMs};
	my $minNumConnectionsLow=$params{minNumConnectionsLow};
    	my $minNumConnectionsMed=$params{minNumConnectionsMed};
    	my $minNumConnectionsHigh=$params{minNumConnectionsHigh};

	switch($connectionType) {
		case /NullTx/i {
			my $Nodes=$params{node1};
			my $clusterIdentity=NACL::CS::ClusterIdentity->fetch(command_interface=>$Nodes);
			my $api = $Nodes->apiset(category => "Node", interface => "CLI", set => "CMode");
			$api->execute_raw_command(command=>'set -confirmations off');
			$api->execute_raw_command(command=>'set test');

			############### CREATE A SESSION FROMN TO D BLADE OF THE SAME NODE ###########################

			my $Response = $api->execute_raw_command(command=>"debug csm session create -node $Nodes->{node} -destination-cluster $clusterIdentity->{name} -passive-node $Nodes->{node} -source-blade $sourceBlade -dest-blade $destBlade -num-channels-low $numChannelsLow -num-channels-med $numChannelsMed -num-channels-high $numChannelsHigh -disable-congestion-control $disableCongestionControl -window-size-low $windowSizeLow -window-size-med $windowSizeMed -window-size-high $windowSizeHigh"); 
			return $Response;
		} #End if connectionType=NullTx

		case /(Rc|IntraclusterCt)/ {
			my $Nodes0=$params{node1};
			my $Nodes1=$params{node2};
			my $clusterIdentity1=NACL::CS::ClusterIdentity->fetch(command_interface=>$Nodes0);
			my $clusterIdentity2=NACL::CS::ClusterIdentity->fetch(command_interface=>$Nodes1);
			if($clusterIdentity1->{name} ne $clusterIdentity2->{name}){
				$Log->warn("Cannot run this with nodes in two different clusters... $connectionType connections are meant to be for nodes within the same cluster");
			}
			my $api = $Nodes0->apiset(category => "Node", interface => "CLI", set => "CMode");
			$api->execute_raw_command(command=>'set -confirmations off');
			$api->execute_raw_command(command=>'set test');

			############### CREATE A SESSION FROM TO ONE NODE TO ANOTHER ###########################
			my $Response = $api->execute_raw_command(command=>"debug csm session create -node $Nodes0->{node} -destination-cluster $clusterIdentity1->{name} -passive-node $Nodes1->{node} -source-blade $sourceBlade -dest-blade $destBlade -num-channels-low $numChannelsLow -num-channels-med $numChannelsMed -num-channels-high $numChannelsHigh -disable-congestion-control $disableCongestionControl -window-size-low $windowSizeLow -window-size-med $windowSizeMed -window-size-high $windowSizeHigh -min-num-connections-low $minNumConnectionsLow -min-num-connections-med $minNumConnectionsMed -min-num-connections-high $minNumConnectionsHigh"); 
			return $Response;
		} #End If connectionType=Rc


  		case /Ct/ {
			my $Nodes0=$params{node1};
			my $Nodes1=$params{node2};
   			#### cluster identity show to check if nodes are in different clusters ####
			my $clusterIdentity1=NACL::CS::ClusterIdentity->fetch(command_interface=>$Nodes0);
			my $clusterIdentity2=NACL::CS::ClusterIdentity->fetch(command_interface=>$Nodes1);
			if($clusterIdentity1->{name} eq $clusterIdentity2->{name}) {
				$Log->warn("Cannot run this with nodes in the same clusters...CT connections are meant to be for nodes acrosss different clusters");
			}
   			#### find cluster peer health #####
			my $remoteCluster=$clusterIdentity2->{name};
			my @clusterPeer=NACL::CS::ClusterPeer->fetch(command_interface=>$Nodes0);
			my $peerClusterFound=0;
			foreach my $clusterPeerRow(@clusterPeer) {
				$Log->comment("rows dest cluster".$clusterPeerRow->{cluster});
				$Log->comment("remote cluster".$remoteCluster);
				if($clusterPeerRow->{cluster} eq $remoteCluster) {
					$peerClusterFound=1;
					if($clusterPeerRow->{availability} ne "Available") {
						$Log->warn("Clusterpeer show result: Availability to cluster $remoteCluster is false");
					}
				}
			}
			if($peerClusterFound==0) {
				$Log->warn("Peer cluster $remoteCluster could  not be located...");
			}
			my $api = $Nodes0->apiset(category => "Node", interface => "CLI", set => "CMode");
			$api->execute_raw_command(command=>'set -confirmations off');
			$api->execute_raw_command(command=>'set test');

			############### CREATE A SESSION FROMLOCAL TO REMOTE NODE ###########################

			my $Response = $api->execute_raw_command(command=>"debug csm session create -node $Nodes0->{node} -destination-cluster $clusterIdentity2->{name}  -passive-node $Nodes1->{node} -source-blade $sourceBlade -dest-blade $destBlade -num-channels-low $numChannelsLow -num-channels-med $numChannelsMed -num-channels-high $numChannelsHigh -disable-congestion-control $disableCongestionControl -window-size-low $windowSizeLow -window-size-med $windowSizeMed -window-size-high $windowSizeHigh -min-num-connections-low $minNumConnectionsLow -min-num-connections-med $minNumConnectionsMed -min-num-connections-high $minNumConnectionsHigh"); 
			return $Response;
		}
	}
}



=pod

=head2 CsmBombPoint 

C<CsmBombPoint>

B<Description>
Enables or disables a CSM bomb point

B<Input>

=over 4

=item * B<node>
	Node on which the bomb point has to be enabled / disabled 

=item * B<name>
	Name of the bomb point to be enabled or disabled 

=item * B<value>
	Value that will trigger the bomb point	

=item * B<enable>
	Value of enable decides if the bomb point is being enabled or disabled  	

=over 4

=item * B<1>
	Enable bomb point	 

=item * B<0>
	Disable bomb point (default)

=back 

=item * B<forever>
	Value of enable decides if the bomb point is being enabled forever  	

=over 4

=item * B<1>
	Enable bomb point forever	 

=item * B<0>
	Do not enable bomb point forever (default)

=back 

=back

B<Output>

Does not return anything

=cut

sub CsmBombPoint {
	my $class=shift;
	my %params = validate(@_,
			{
			node=>{type=> HASHREF},
			name => { type => SCALAR},
			value => {type=> SCALAR, optional=>1},
			enable => {type=> SCALAR, default=>0},
			forever => {type=> SCALAR, default=>0},
			test =>{ type=>SCALAR, optional=>1, default=>"none"}	
			}
		);

	my $Nodes0=$params{node};
	my $name=$params{name};
	my $value=$params{value};
	my $enable=$params{enable};
	my $forever=$params{forever};
	my $filer0=$Nodes0->{node}; 
	my $api_bsd_node1 = $Nodes0->apiset(category => "Node", interface => "CLI", set => "Systemshell");
	$api_bsd_node1->get_connection()->discardbuff();
	my @clusterPortsNode0;
	my @responseclusterPortsNode0=NACL::CS::NetworkInterface->fetch(command_interface=>$Nodes0,filter=>{role=>"cluster",vserver=>$filer0});
	if(($params{test} eq "IntraclusterCt") || ($params{test} eq "Ct")) {	
		foreach my $clusterLif(@responseclusterPortsNode0){
			push(@clusterPortsNode0,$clusterLif->{lif});
		}
		foreach my $clif0(@clusterPortsNode0) {
			NACL::C::NetworkInterface->modify(command_interface=>$Nodes0,server=>$filer0,lif=>$clif0,"status-admin"=>"down");
		}
	}
	$api_bsd_node1->execute_raw_command(command=>"sudo sysctl sysvar.csm.bomb.enable=$enable");
	$api_bsd_node1->execute_raw_command(command=>"sudo sysctl sysvar.csm.bomb.0.name=$name");
	$api_bsd_node1->execute_raw_command(command=>"sudo sysctl sysvar.csm.bomb.0.set=$enable");
	$api_bsd_node1->execute_raw_command(command=>"sudo sysctl sysvar.csm.bomb.0.forever=$forever");

	if(($params{test} eq "IntraclusterCt") || ($params{test} eq "Ct")) {	
		foreach my $clif0(@clusterPortsNode0) {
			NACL::C::NetworkInterface->modify(command_interface=>$Nodes0,server=>$filer0,lif=>$clif0,"status-admin"=>"up");
		}
	}
} 


=pod

=head2 BombPointCheck 

C<BombPointCheck>

B<Description>
Checks if a bomb point has fired by returning the impact of setting the bomb point

B<Input>

=over 4

=item * B<node>
	Node on which the bomb point has been enabled  

=item * B<bombPoint>
	Name of the bomb point 

=item * B<response>
	Response of a command which has to be parsed to retrieve impact of setting the bomb point	

=back

B<Output>

Returns the impact of the bomb point setting or flags an error if expected behavior is not seen 

=cut

sub BombPointCheck {
	my $class=shift;
	my %params = validate(
				@_,
				{ 
				node=>{type=> HASHREF},
				bombPoint => { type => SCALAR},
				response => {type=> SCALAR, optional=>1},
				}
			);

	my $node=$params{node};
	my $bombPoint=$params{bombPoint};
	my $response=$params{response};
	my $status;
	switch($bombPoint) {
		case "rc_RcActiveOpen::loActiveOpenRsp_noMem" { 
			#	$rcCounter=NACL::CS::DebugCsmCounters->fetch(command_interface=>$node,filter=>{category=>"rc",name=>"LoActiveOpenRspNoMem"});
			#	Waiting for burt 514740 fix	
			# $Log->comment("LoActiveOpenRspNoMem value = $rcCounter");
			if($response=~/CSM_NO_MEMORY/ || $response=~/Error/) { 
				$status="Session create failed  because the bomb point $bombPoint was set.. Response $response";
				$Log->comment("$status");	
			}
		   	else {
				$Log->error("The bomb point $bombPoint did not have any effect"); 
			}
		}
	
		case "rc_RcConnection::activeOpenInit_forceReBind" {
			if($response=~/NOT\sREGISTERED/ || $response=~/Error/) { 
				$status="Session create failed  because the bomb point $bombPoint was set.. Response $response";
				$Log->comment("$status");	
			}
			else {
				$Log->error("The bomb point $bombPoint did not have any effect"); 
			}

		}
	
		case "CsmListenUser::passiveOpen_NO_BLADE" {
			if($response=~/CSM_NO_BLADE/ || $response=~/Error/) { 
				$status="Session was not registered because of the bomb point $bombPoint.. Response $response"; 
				$Log->comment("$status");	
			}
			else {
				$Log->error("The bomb point $bombPoint did not have any effect"); 
			}

		}	

		case "ct_CtActiveOpen::loActiveOpenRsp_noMem" { 
			my $ctCounter=NACL::CS::DebugCsmCounters->fetch(command_interface=>$node,filter=>{category=>"ct",name=>"LoActiveOpenRspNoMem"});
			$Log->comment("LoActiveOpenRspNoMem value = $ctCounter->{value}");
			if($response=~/CSM_NO_MEMORY/ || $response=~/Error/) { 
				$status="Session failed because the bomb point $bombPoint was set.. Response $response"; 
				$Log->comment("$status");	
			}
			else {
				$Log->error("The bomb point $bombPoint did not have any effect"); 
			}
		}	

		case "ct_CtConnection::activeOpenInit_forceReBind" {
			if($response=~/NOT\sREGISTERED/ || $response=~/Error/) { 
				$status="Session was not registered because of the bomb point $bombPoint .. Response $response"; 
				$Log->comment("$status");
			}
			else {
				$Log->error("The bomb point $bombPoint did not have any effect"); 
			}

		}	
		case "ct_allocOntapGbuf_fail" {
			if($response=~/CSM_BACKPRESSURE/) { 
				$status="CSM_BACKPRESSURE was returned because of the bomb point $bombPoint.. Response from the send command was $response";
				$Log->comment("$status");
			} 
			else {
				$Log->error("The bomb point $bombPoint did not have any effect"); 
			}

		}	
	}
	return $status;
}
1;

=pod

=head1 LIMITATIONS

The initial offering of the CSM.pm library utilizes "current" values, and is meant mainly as a post-processing tool. Future versions
will attempt to take samples and report delta values in-paralell.
The lib uses NACL APISet as NACL components were unavailable.


=cut

