# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Dns.pm#1 $
#
# Copyright (c) 2004 Network Appliance Inc.
# All right reserved.
#
## @summary Routines for running network utilities
##
## @description		
## 		These functions are used to run various dns
## 		utilities on devices independent of the device type.
##		These functions help in abstracting the test scripts
##		from subtle differences in command usage and command
##		outputs on various platforms.
##
## @dependencies
##
## use Tharn;	
##
## @author  dl-maryv (FSS Protocols QA)

package Network::Dns;

require Exporter;
use strict;
use Tharn;
use Data::Dumper;
use Carp;
use SayWrap qw(sayEOF);
use Params::Validate qw(:all);
use Filer::Version qw(get_wafl_version WV_TSINGTAO_MIN
		      UNSUPPORTED_FEATURE_EXCEPTION);
use filer qw(filer_options_set filer_options_get);
use CLI::FileIO qw(rdfile wrfile);

use vars qw (@ISA @EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT_OK = qw (
		 	dns_command
		 	dns_get_active_server
		 	dns_dig
		 	dns_configure
		 	ddns_is_supported
		 	dns_get_options
                 	dns_restore_options
		    );

# path for resolv.conf file on the filer
my $resolv_conf = '/etc/resolv.conf';


## @name ::dns_command()
##
## @summary runs dns commands
##
## @description
##
## The function runs "dns" commands and returns command output
##
## @returns output of specified command
##
## @arg $conn  Required: connectspec of device
## @arg $command  Required: command to run.
##                        eg. "info"
##
## @arg %args  Optional: Hash of arguments
##             options => command specific option string
##

sub dns_command($$%) {

    my ($conn,$command,%args) = @_;
    my $result = "";
    my $command_options = $args{'options'};

    $result = sayEOF($conn,
		     "priv set advanced ; dns $command $command_options");

    return($result);

}

## @name ::dns_get_active_server()
##
## @summary finds out the current active DNS server
##
## @description
##
## The function runs "dns info" and returns the ip address of the active DNS
## server
##
## @returns ip address of active DNS server
##
## @arg $conn  Required: connectspec of device
##


sub dns_get_active_server( $ ) {

    my $conn = shift;
    my $active_server = "";
    my $output = sayEOF($conn,"dns info");

    if ($output !~ /DNS is enabled/) {
	croak "DNS is disabled"
    }

    if ($output =~ m/((\d+\.){3}\d+)\s+UP/) {
	$active_server = $1;
    } else {
	croak "no active DNS server";
    }

    return ($active_server);
}



## @name ::dns_dig()
##
## @summary runs dig to get domain name information for given name
##
## @description this function runs dig, parses the output of dig and returns
## the entire output of dig and ip addresses returned by the dns server
##
## @returns a reference to hash that contains the following keys
##          full_output : entire output of dig command
##          <ip_addr> : hash containing 
##                      'ttl': time to live
##                      'class': address class (eg IN,internet)
##                      'family': A or AAAA (v4 or v6)
##          1 <ip_addr> hash is returned for each address returned by the 
##          dns server.
##
## @arg %args  Required: Hash of arguments
##             args{'conn'} : connection to client on which to run dig
##             args{'server'} : name of dns server (Start of Authority)
##             args{'name'} : hostname to lookup using dig
##
## @example my $dns_dig = dns_dig(conn => $cc,
##			          server => "nwk-pc75", 
##			          name => "nwk-u18.netqasec.netapp.com");

sub dns_dig(%) {

    my $return_hash = {};
    my $cmd;

    my %args = Params::Validate::validate(@_, {conn => { type => OBJECT },
					       server => { type => SCALAR },
					       name => { type => SCALAR }
					      });

    my $conn = $args{'conn'};

    if ($conn->hostp()->hosttype() eq 'unix-solaris') {
	$cmd = "/usr/sbin/dig \@$args{'server'} -c any $args{'name'}";
    } elsif ($conn->hostp()->hosttype() eq 'unix-linux') {
	$cmd = "/usr/bin/dig \@$args{'server'} -c any $args{'name'}";
    } else {
       	$cmd = "dig \@$args{'server'} -c any $args{'name'}";
    }

    my $dig_out= sayEOF($conn,$cmd);

    $return_hash->{'full_output'} = $dig_out;

    if ($dig_out =~ m/;; ANSWER SECTION:(.*);;/si) {

	# strip of blank lines
	my $answer_section = $1;
	$answer_section =~ s/\n\n/\n/g;
	foreach my $line (split /\n/,$answer_section) {
	    if ($line =~ 
		/^$args{'name'}\.\s+(\S+)\s+(\S+)\s+(\S+)\s+((\d+.){3}\d+)/) {
		$return_hash->{$4}{'ttl'} = $1;
		$return_hash->{$4}{'class'} = $2;
		$return_hash->{$4}{'family'} = $3;
	    }
	}
    }

    return $return_hash;

}

##
## @name ::dns_configure()
##
## @summary configure dns on a filer
##
## @description this routine configured dns and dns update options
##
## @returns 1 on success.
##
## @arg %args  Required: Hash of arguments
##             args{'conn'} : connection to client on which to run dig
##             args{'enable'} : value for option dns.enable
##             args{'domain'} : value for option dns,domainname
##             Optional: arguments
##             args{'cache'} : value for option dns.cache
##             args{'update'} : value for option dns.update
##             args{'ttl'} : value for option dns.update.ttl
##             args{'srv_ip_addr'} : DNS server IP address
##
##
## @example my $dns_dig = dns_configure(conn => $fc,
##                                enable => "on",
##                                domain => "netqasec.netapp.com",
##                                cache => "off",
##                                update => "on",
##                                ttl => "10m");
##
##

sub dns_configure(%) {

    my %args = Params::Validate::validate(@_, {conn => { type => OBJECT },
					       enable => { type => SCALAR },
					       domain => { type => SCALAR },
					       cache => { type => SCALAR|UNDEF,
							  default => 'on'},
					       update => { type =>
							   SCALAR|UNDEF,
							   default => 'off' },
					       ttl => { type =>
							SCALAR|UNDEF,
							optional => 1},
                                               srv_ip_addr => { type => SCALAR,
                                                   regex=>qr/(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/,
                                                   optional => 1 },
					      });

    my $ddns = 1;
    my $conn = $args{'conn'};

    my $version= get_wafl_version($conn->hostp());

    if ($version < WV_TSINGTAO_MIN) {
	$ddns = 0;
    }

    # set dns options
    filer_options_set($conn,"dns.domainname",$args{'domain'}); 
    filer_options_set($conn,"dns.enable",$args{'enable'});
    filer_options_set($conn,"dns.cache.enable",$args{'cache'});

    # ddns options
    if ($ddns == 1) {
	filer_options_set($conn,"dns.update.enable",$args{'update'});
	if (defined($args{'ttl'})) {
	    filer_options_set($conn,"dns.update.ttl",$args{'ttl'});
	}
    }

    # Change the server resolv.conf if DNS server IP address specified
    if (defined($args{srv_ip_addr})) {
        wrfile($conn, $resolv_conf, "nameserver $args{srv_ip_addr}\n");
        croak("dns_configure: wrfile: $@") if ($@);
    }
    return 1;
}

##
## @name ::ddns_is_supported()
##
## @summary check if ddns is supported on filer ontap version 
##
## @description ddns is supported only in TsingTao and newer releases
##
## @returns 1 on success
##          croaks on failure
##
## @arg %args  Required: Hash of arguments
##             args{'hostobj'} : filer host object
##
## @example ddns_is_supported(hostobj => $fh);
##

sub ddns_is_supported {

    my %args = validate(@_,{ hostobj => {type => OBJECT}});

    my $hostobj = $args{'hostobj'};

    my $version = get_wafl_version($hostobj);
    if ($version < WV_TSINGTAO_MIN) {
	croak(UNSUPPORTED_FEATURE_EXCEPTION . 
	      ": ddns unsupported in release prior to TsingTao");
    }
    return 1;
}

##
## @name ::dns_get_options()
##
## @summary extract dns options and return a hash of options
##
## @description this function runs options dns and parses out the values of
## all dns options. 
##
## @note This will need to be updated everytime dns options are changed
##
## @returns hash reference containing dns options
##
##
## @arg %args  Required: %args hash containing arguments
##             args{'conn'} 
##
## @example dns_get_options(conn => $fc)
##

sub dns_get_options {

    my %args = validate(@_,{ conn => {type => OBJECT}});

    my $conn = $args{'conn'};

    my $dns_options = {};

    $dns_options->{'domain'} = filer_options_get($conn,"dns.domainname");
    $dns_options->{'enable'} = filer_options_get($conn,"dns.enable");
    $dns_options->{'cache'} = filer_options_get($conn,"dns.cache.enable");
    $dns_options->{'update'} = filer_options_get($conn,"dns.update.enable");
    $dns_options->{'ttl'} = filer_options_get($conn,"dns.update.ttl");
    $dns_options->{'resolv_conf'} = rdfile($conn, $resolv_conf);

    return $dns_options;
}

##
## @name ::dns_restore_options()
##
## @summary Restore filer DNS configuration
##
## @description Restore the DNS options specified by the dns_options hash.
##  Use dns_get_options() to get the hash.
##
## @arg %args  Required: Hash of arguments
##             args{'conn'} : connection to filer
##             args{'options'} : The DNS options returned from dns_get_options()
##
## @example dns_restore_options(conn => $fc, options => $saved_dns_options);
##

sub dns_restore_options(%) {
    my %args = validate(
        @_,
        {
            conn    => { type => OBJECT },
            options => { type => HASHREF },
        }
    );

    my $dns_options = $args{options};

    if (defined($dns_options->{'resolv_conf'})) {
        wrfile($args{conn}, $resolv_conf, $dns_options->{'resolv_conf'});
        croak("dns_restore_options: wrfile: $@") if ($@);
    }

    dns_configure(conn => $args{conn},
                  domain => $dns_options->{'domain'},
                  enable => $dns_options->{'enable'},
                  cache  =>  $dns_options->{'cache'},
                  update =>  $dns_options->{'update'},
                  ttl    =>  $dns_options->{'ttl'});

    return;
}
