# this is basically -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/Network/DnsResolver.pm#1 $
#
# Copyright (c) 2008 Network Appliance, Inc., All Rights Reserved
# Permission is granted to use this script on a Network Appliance Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from Network Appliance, Inc
#
## @summary      Subroutines to verify DNS Name Resolution
## @dependencies That the server is correctly configured to handle the domains we are
##               querying for
## @author       dl-cluster-qa
## @pod          here

=head1 NAME

DnsResolver.pm

=head1 AUTHOR

dl-cluster-qa

=head1 DESCRIPTION

C<DnsResolver> will handle issuing commands and parsing the output sent to and received 
from the client to verify DNS Name Resolution. This class can be similar to 
../lib/Network/Dns.pm (so you may want to read that library as well before deciding 
that this is what you need). In addition, this class will:
    * Allow the calling method to verify load balancing via several methods:
        - dig
        - host
        - nslookup
    * Repetively query the filer to associate weights with the IP addresses returned
    * Convert weights to percentages (TODO)
    * Check for an even distribution (TODO)
    * Check if an IP is simply handed out or not handed out
It is also worth noting that this class handles nothing on the filer end (e.g. no DNS 
configuration, VifConfig::VifDns handles that), this is purely for IP address verification.

=head1 SYNOPSIS

    use Network::DnsResolver;
    my $Dns_Resolver = new Network::DnsResolver();
    my $zone = "Test.com";
    my $mount_to = $Dns_Resolver->resolve_domain(domain => $zone);
    my @ips = (1.1.1.1, 1.1.1.2, 1.1.1.3);
    my $result = $Dns_Resolver->verify_ip_range_exists(ips => \@ips, domain => $zone);
    $result = $Dns_Resolver->verify_ip(ips => \@ips, domain => $zone);
    $result = $Dns_Resolver->verify_ip(ips => \@ips, domain => $zone, state => 0);

=head1 ARGUMENTS

=over 4

=item DNS_COMMAND
(Optional)

Usage: [DNS_COMMAND=dig]

Description: Specifies the DNS Command to use (select from: dig, host, or nslookup 

=item DNS_SERVER
(Optional)

Usage: [DNS_SERVER=undef]

Description: Specifies the DNS Server for ALL clients to use. If this is undefined, the
first data LIF from "network interface show" will be used since DNS is always enabled
on all data LIFs.

=item CLIENT_CONN
(Optional)

Usage: [CLIENT_CONN=#rsh]

Description: Type of connection to make to the client

=item DNS_TIMEOUT
(Optional)

Usage: [DNS_TIMEOUT=120]

Description: Timeout fir issuing DNS Commands on a client

=back

=head1 PUBLIC METHODS

=over 4

=item set_dns_command

Parameters: Name of the DNS command to be used going forward. Returns: 1 on 
Success, 0 on Failure. Purpose: So the calling method is able to goggle the 
dns_command parameter that this class examines each time is is about to 
issue a DNS command.

=item verify_ip

Parameters: IP to evaluate, Domain IP should be in, State (exists=1 to verify
IP is handed out, or exists=0 to verify IP is not handed out.
Returns: 1 on Success (specified IP is found if state=>1 OR specified is not
found if state=>0), O on Failure. Purpose: To verifiy a specified IP address 
is or is not part of a DNS Zone.

=item verify_ip_range_exists

Parameters: IP range to verify, and any other parameters that need to be 
passed to verify_ip. Returns: 1 on Success for ALL IPs or 0 for 1 failed 
IP. Purpose: To have a wrapper for verify_ip that allows the user to verify a 
range exists (only used to verify IPs exist, since we want to verify each 
IP in the range). 

=item resolve_domain

Parameters: Domain name to resolve. Returns: The IP address of the domain
Purpose: To allow the calling method to perform the dig/host/etc. 
themselves.

=back

=head1 PRIVATE METHODS

=over 4

=item _dig

Parameters: Reference to self, Domain to dig on, optional server. Returns: 
IP address found from issuing the 'dig'. Purpose: To perform dig on a 
client and parse out the IP address.

=item _host

Parameters: Reference to self, Domain to issue host on, optional server.
Returns: IP address found from issuing the 'host'. Purpose: To perform 
host on a client and parse out the IP address.

=item _nslookup

Parameters: Reference to self, Domain to issue nslookup on, optional server.
Returns: IP address found from issuing the 'nslookup'. Purpose: To perform 
nslookup on a client and parse out the IP address.

=item _get_client

Parameters: None. Returns: A client name (from CLIENT=...). Purpose: To 
extract a client from within the object. Assumes that $self->{_clients} was 
already initialized (in new). In addition to returning a client, the method 
shifts the first client off of the beginning of the array and pushes it onto 
the end of the array to try to get a uniform distribution across all clients 
for issuing DNS commands.

=back

=cut

package Network::DnsResolver;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Switch;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);

# Netapp Specific Libs

use TracePkg;
use Server;
use Options;
use SayWrap qw(sayEOF);
use FS::Util::Utilities qw(uSystem);
use Utils qw(make_conn);
use Tharn;

# Class Data, only accessible from this class
{
    my $refCount = 0;

    sub incRef {
        $refCount++;
    }

    sub getRef {
        return $refCount;
    }
}

# Globals

my $Trace   = new TracePkg();
my $Server  = new Server();
my $Options = new Options();
my $self;

#############################################################################
# Subroutine: new
# Parameters: None
# Returns: Reference to this object
# Purpose: To instantiate and configure the DnsResolver object
#############################################################################

sub new {

    $Trace->entry();
    my $class = shift(@_);

    # Check to see if this class has already been declared. If so, then we can
    # just return its reference.

    if (getRef()) {
        $Trace->exit(msg => "Package has already been declared. Returning...");
        return $self;
    }

    # Get the command line options. These will become the defaults for
    # the DnsResolver. However, the calling method is still capable of
    # passing in values to override these to each of the relative subroutines.

    $Options->parameters(
        definitions => [
            {
                name => "DNS_COMMAND",
                description =>
                  "Specifies the DNS Command to use (select from: dig, "
                  . "host, or nslookup",
                default => "dig",
            },
            {
                name => "DNS_SERVER",
                description =>
                  "Specifies the DNS Server for ALL clients to use",
                default => undef,
            },
            {
                name        => "CLIENT_CONN",
                description => "Type of connection to make to the client",
                default     => "#rsh",
            },
            {
                name        => "DNS_TIMEOUT",
                description => "Timeout for issuing DNS command on client",
                default     => "120",
            },
        ]
    );
    $Options->process();
    $Trace->debug(msg => "Options = " . Dumper($Options));

    # Configure the DnsResolve object (done by copying over the options
    # locally).

    my $opts = $Options->get_hash();
    $self->{_dns_command} = $opts->{DNS_COMMAND};
    $self->{_clients}     = $opts->{CLIENT};
    $self->{_client_conn} = $opts->{CLIENT_CONN};
    $self->{_dns_timeout} = $opts->{DNS_TIMEOUT};

    # Find the DNS Server if it was not specified. This will just default to
    # the first data VIF (non-cservser) on the cluster.

    if (!defined($opts->{DNS_SERVER})) {
        my @vifs = $Server->network_interface_show(
            role              => "data",
            'firewall-policy' => "data",
            vif               => "!p_*",
        );
        $self->{_server} = $vifs[0]->{address};
    } else {
        $self->{_server} = $opts->{DNS_SERVER};
    }

    # Bless and return

    bless($self, $class);
    incRef();

    $Trace->debug(msg => "Initialized Network::DnsResolver object with the "
          . "following configuration:\n"
          . Dumper($self));

    $Trace->exit();
    return $self;
} ## end sub new

############################################################################
# Subroutine: set_dns_command
# Parameters: Name of the DNS command to be used going forward
# Returns: 1 on Success, 0 on Failure
# Purpose: So the calling method is able to goggle the dns_command parameter
#          that this class examines each time is is about to issue a DNS
#          command
#############################################################################

sub set_dns_command {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(@_, {new_cmd => {type => SCALAR},});

    # The only valid commands for our purposes here are: dig, host, and nslookup.
    # Verify it is one of those passed in.

    if (   ($params{new_cmd} !~ m/^dig$/)
        && ($params{new_cmd} !~ m/^host$/)
        && ($params{new_cmd} !~ m/^nslookup$/)) {
        $Trace->warn(msg => "Encountered invalid command. Expected: dig/host/"
              . "nslookup. Found: $params{new_cmd}.");
        return 0;
    }

    # Now that we know we have a valid command, we can perform the set and
    # return

    my $old_command = $self->{_dns_command};
    $self->{_dns_command} = $params{new_command};
    $Trace->exit(
        msg => "Successfully switched DNS command from $old_command " . "to "
          . $self->{_dns_command});
    return 1;
} ## end sub set_dns_command

###############################################################################
# Subroutine: verify_ip
# Parameters: IP to evaluate, Domain IP should be in, State (exists=1 to verify
#             IP is handed out, or exists=0 to verify IP is not handed out.
# Returns: 1 on Success (specified IP is found if state=>1 OR specified is not
#          found if state=>0), O on Failure
# Purpose: To verifiy a specified IP address is or is not part of a DNS Zone
###############################################################################

sub verify_ip {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(
        @_,
        {
            ips    => {type => SCALAR | ARRAYREF},
            domain => {type => SCALAR},
            server =>
              {type => SCALAR, optional => 1, default => $self->{_server}},
            state     => {type => SCALAR, optional => 1, default => 1},
            num_tries => {type => SCALAR, optional => 1, default => 200},
            sleep     => {type => SCALAR, optional => 1, default => 15},
        }
    );

    # If the state is 0 and the user did not override the default value of num_tries
    # we will turn it down here to cut back on test execution time. Since we are
    # grepping through an array of IPs to make sure none of them exists, we don't
    # need to try 200 times (unless otherwise specified).

    if (($params{state} == 0) && ($params{num_tries} == 200)) {
        $params{num_tries} = 5;
    }
    $Trace->comment(msg => "Retry count set to: $params{num_tries}");

    # Cast to an array if not already

    if (ref($params{ips}) ne 'ARRAY') {
        $params{ips} = [$params{ips}];
    }

    # Open the connection. We will simply verify the IP on one client.
    # Connection done here so we can iterate through clients continiously
    # in case of verify_ip_range_exists

    my $client = _get_client($self);

    # Verify that we are not executing "dig" on a windows client

    my $host = host($client);
    if (($host->hosttype() =~ /^win/) && ($self->{_dns_command} =~ /^dig$/)) {
        $Trace->warn(
            msg => "Cannot execute dig on a windows client, changing to "
              . "nslookup");
        set_dns_command(new_cmd => "nslookup");
    }

    my $connection = make_conn($client, $self->{_client_conn});

    # There is an dampening interval of 10 seconds before the VifMgr updates bind.
    # We will sleep for that here, unless the calling method specifies otherwise.

    $Trace->comment(
        msg => "Sleeping for $params{sleep} seconds before querying zone.");
    sleep $params{sleep};

    # Check for the IP address (at most) the specified number of tries

    $Trace->comment(msg => "Checking domain: $params{domain}, IP(s): "
          . Dumper($params{ips}));

    for (my $i = 0 ; $i < $params{num_tries} ; $i++) {

        # To avoid issues with the 'protocol circuit' failing on the client, we'll
        # give it a second before retrying.

        sleep 1;

        # Switch based on the command

        my $ip_returned = 0;
        switch ($self->{_dns_command}) {
            case ("dig") {
                $ip_returned = _dig(
                    $self,
                    domain     => $params{domain},
                    connection => $connection
                );
            }
            case ("host") {
                $ip_returned = _host(
                    $self,
                    domain     => $params{domain},
                    connection => $connection
                );
            }
            case ("nslookup") {
                $ip_returned = _nslookup(
                    $self,
                    domain     => $params{domain},
                    connection => $connection
                );
            }
            else {
                $Trace->warn(
                        msg => "Encounted unexpected DNS command. Found:" . " "
                      . $self->{_dns_command}
                      . " Expected: "
                      . "dig/host/nslookup");
                return 0;
            }
        } ## end switch

        # Now we will check if we have a match. If the state is flagged to
        # 1, we are verifying the IP exists. In that case, if a match is
        # seen we can successfuly return, otherwise, contine looping.
        # If we are verifying a match does not exist (the state will be
        # flagged to 0) and we found a match, we can retry to give the service.
        # time to update.

        my $match = 0;

        # If the ip_returned matches anything in the $params{ips}, flag as a match.

        if (   ($ip_returned ne "0")
            && (scalar(grep(/^$ip_returned$/, @{$params{ips}})))) {
            $Trace->debug(msg => "Found match for: $ip_returned");
            $match = 1;
        }

        if ($params{state}) {

            # Find a match when we expect or retry

            if ($match) {
                $Trace->exit(
                    msg => "Successfully matched IP address: " . $ip_returned);
                return 1;
            } else {

                # Did not find a match when expected to, but we will retry

                next;
            }
        } else {

            # In the case where we are verify that the IP does not exist,
            # we want to fail on a match. Since the DNS Zone can can return
            # a different IP each time, we want to go through all of num_tries
            # and make sure that we never get a match. Generally, for this
            # case, num_tries should be a lower number than the default.

            if ($match) {

                # Dump additional information
                
                $Server->cli(cmd => "debug mtrace log show -node * -process ".
                             "vifmgr -buffer loadbalance -fields message,emit-time");
                $Server->cli(cmd => "debug mtrace log show -node * -process mgwd ".
                             "-buffer loadbalance -fields message,emit-time");

                # Fail

                $Trace->exit(msg => "IPs in "
                      . Dumper($params{ips})
                      . " are still being "
                      . "resolved in DNS Zone $params{domain} when they were expected not to. "
                      . "Exiting.");

                return 0;

            } else {

                # Retry to verify its not returned at all

                next;
            }
        } ## end else [ if ($params{state})
    } ## end for (my $i = 0 ; $i < $params...

    # Now do one last check. If we at this point, and we were seeing if the
    # IP address existed (i.e. state=1), then we have failed. If we are at
    # this point and we were verifying that the IP address did not exist,
    # then we will return in success because we didnt not find a match in the
    # specified number of retries.

    if ($params{state}) {

        # Dump additional information before returning
        
        $Server->cli(cmd => "debug mtrace log show -node * -process ".
                     "vifmgr -buffer loadbalance -fields message,emit-time");
        $Server->cli(cmd => "debug mtrace log show -node * -process mgwd ".
                     "-buffer loadbalance -fields message,emit-time");

        $Trace->exit(msg => "Unable to find match on IP addresses: "
              . Dumper($params{ips}) . "via "
              . $self->{_dns_command}
              . " for zone: "
              . $params{domain});

        return 0;

    } else {
        $Trace->exit(msg => "IPs in: "
              . Dumper($params{ips})
              . " were not found "
              . "in zone: $params{domain} via "
              . $self->{_dns_command}
              . " as expected. Returning success.");
        return 1;
    }
} ## end sub verify_ip

#############################################################################
# Subroutine: verify_ip_range_exists
# Parameters: IP range to verify, and any other parameters that need to be
#             passed to verify_ip
# Returns: 1 on Success for ALL IPs or 0 for 1 failed IP
# Purpose: To have a wrapper for verify_ip that allows the user to verify a
#          range exists (only used to verify IPs exist, since we want to
#          verify each IP in the range).
###############################################################################

sub verify_ip_range_exists {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(
        @_,
        {
            ips    => {type => ARRAYREF},
            domain => {type => SCALAR},
            server => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_server}
            },
            state => {
                type     => SCALAR,
                optional => 1,
                default  => 1
            },
            num_tries => {
                type     => SCALAR,
                optional => 1,
                default  => 200
            },
        }
    );

    # Iterate through all of the IPs and perform a verify

    foreach my $ip (@{$params{ips}}) {

        if (
            !verify_ip(
                $self,
                ips       => $ip,
                domain    => $params{domain},
                server    => $params{server},
                num_tries => $params{num_tries},
                state     => $params{state}
            )
          ) {
            $Trace->warn(
                msg => "Failure verifying IP with: " . Dumper(\%params));
            return 0;
        } ## end if (!verify_ip($self, ...
    } ## end foreach my $ip (@{$params{ips...

    $Trace->exit(msg => "Success verifying IP with: " . Dumper(\%params));
    return 1;
} ## end sub verify_ip_range_exists

#############################################################################
# Subroutine: resolve_domain
# Parameters: Domain name to resolve
# Returns: The IP address of the domain
# Purpose: To allow the calling method to perform the dig/host/etc. themselves
############################################################################

sub resolve_domain {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(
        @_,
        {
            domain => {type => SCALAR},
            server => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_server}
            },
        }
    );

    # Open the connection. We will simply verify the IP on one client

    my $client = _get_client($self);
    my $connection = make_conn($client, $self->{_client_conn});

    # Resolve the domain

    my $ip_returned = 0;
    switch ($self->{_dns_command}) {
        case ("dig") {
            $ip_returned = _dig(
                $self,
                domain     => $params{domain},
                connection => $connection
            );
        }
        case ("host") {
            $ip_returned = _host(
                $self,
                domain     => $params{domain},
                connection => $connection
            );
        }
        case ("nslookup") {
            $ip_returned = _nslookup(
                $self,
                domain     => $params{domain},
                connection => $connection
            );
        }
        else {
            $Trace->warn(msg => "Encounted unexpected DNS command. Found:" . " "
                  . $self->{_dns_command}
                  . " Expected: "
                  . "dig/host/nslookup");
            return 0;
        }
    } ## end switch

    $Trace->exit(msg => "Returning with: $ip_returned");
    return $ip_returned;
} ## end sub resolve_domain

##############################################################################
# Subroutine: _dig
# Parameters: Reference to self, Domain to dig on, optional server
# Returns: IP address found from issuing the 'dig'
# Purpose: To perform dig on a client and parse out the IP address
##############################################################################

sub _dig {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(
        @_,
        {
            domain => {type => SCALAR},
            server => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_server}
            },
            connection => {type => HASHREF},
        }
    );

    # Set the command based on the OS

    my $dig_cmd;
    if ($params{connection}->hostp()->hosttype() eq 'unix-solaris') {
        $dig_cmd =
            "/usr/local/bin/dig \@"
          . $params{server}
          . " -c any "
          . $params{domain};
    } elsif ($params{connection}->hostp()->hosttype() eq 'unix-linux') {
        $dig_cmd =
          "/usr/bin/dig \@" . $params{server} . " -c any " . $params{domain};
    } else {
        $dig_cmd = "dig \@" . $params{server} . " -c any " . $params{domain};
    }
    $Trace->debug(msg => "Issuing command $dig_cmd");

    # Issue the command

    my $dig_output =
      sayEOF($params{connection}, $dig_cmd, timeout => $self->{_dns_timeout});
    $Trace->debug(msg => "dig retured: $dig_output");

    # Extract the IP returned (e.g. Pull out the ANSWER SECTION, strip the
    # blank lines, and parse the IP).

    my $found_ip = 0;
    if ($dig_output =~ m/;; ANSWER SECTION:(.+?);;/si) {
        my $answer_section = $1;
        $answer_section =~ s/\n\n/\n/g;

        if ($answer_section =~
            /$params{domain}\.\s+(\w+)\s+(\w+)\s+(\w+)\s+((\d+.){3}\d+)/) {
            $found_ip = $4;
        }
    } elsif (
        $dig_output =~ m/connection timed out; no servers could be reached/) {
        $Trace->fatal(
            msg    => "DNS Server unreachable. Exiting.",
            result => "FAIL"
        );
    }

    # Make sure we found something and return. No warning here because we
    # may be expecting the IP to NOT exist

    if (!$found_ip) {
        $Trace->exit(msg => "Dig did not find an IP in Domain: "
              . "$params{domain}, Server: $params{server}");
        return 0;
    }

    $Trace->exit(msg => "Dig found: $found_ip in Domain: $params{domain}, "
          . "Server: $params{server}");
    return $found_ip;
} ## end sub _dig

##############################################################################
# Subroutine: _host
# Parameters: Reference to self, Domain to issue host on, optional server
# Returns: IP address found from issuing the 'host'
# Purpose: To perform host on a client and parse out the IP address
##############################################################################

sub _host {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(
        @_,
        {
            domain => {type => SCALAR},
            server => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_server}
            },
            connection => {type => SCALAR},
        }
    );

    # Set the command based on the OS

    my $host_cmd = "host $params{domain} $params{server}";
    $Trace->debug(msg => "Issuing command $host_cmd");

    # Issue the command

    my $host_output = sayEOF($params{connection}, $host_cmd);
    $Trace->debug(msg => "host retured: $host_output");

    # Extract the IP returned (e.g. Pull out the 'has address', and match
    # on the the IP address). Host only returns one IP at a time.

    my $found_ip = 0;
    if ($host_output =~ m/(\S)+ has address ((\d+.){3}\d+)/si) {
        $found_ip = $1;
    }

    # Make sure we found something and return

    if (!$found_ip) {
        $Trace->warn(msg => "No IP address found for Server: $params{server},"
              . " Domain: $params{domain}");
        return 0;
    }

    $Trace->exit(msg => "Host found: $found_ip in Domain: $params{domain}, "
          . "Server: $params{server}");
    return $found_ip;
} ## end sub _host

##############################################################################
# Subroutine: _nslookup
# Parameters: Reference to self, Domain to issue nslookup on, optional server
# Returns: IP address found from issuing the 'nslookup'
# Purpose: To perform nslookup on a client and parse out the IP address
##############################################################################

sub _nslookup {

    $Trace->entry();
    my $self = shift(@_);

    # Initialize the parameters

    my %params = validate(
        @_,
        {
            domain => {type => SCALAR},
            server => {
                type     => SCALAR,
                optional => 1,
                default  => $self->{_server}
            },
            connection => {type => SCALAR},
        }
    );

    # Set the command based on the OS

    my $nslookup_cmd = "nslookup $params{domain} $params{server}";
    $Trace->debug(msg => "Issuing command $nslookup_cmd");

    # Issue the command

    my $nslookup_output = sayEOF($params{connection}, $nslookup_cmd);
    $Trace->debug(msg => "nslookup retured: $nslookup_output");

    # Extract the IP returned (e.g. Pull out the 'has address', and match
    # on the the IP address). Host only returns one IP at a time.

    my $found_ip = 0;
    if ($nslookup_output =~ m/Name:(\S*)\nAddress: ((\d+.){3}\d+)/si) {
        $found_ip = $2;
    }

    # Make sure we found something and return

    if (!$found_ip) {
        $Trace->warn(msg => "No IP address found for Server: $params{server},"
              . " Domain: $params{domain}");
        return 0;
    }

    $Trace->exit(msg => "Host found: $found_ip in Domain: $params{domain}, "
          . "Server: $params{server}");
    return $found_ip;
} ## end sub _nslookup

###############################################################################
# Subroutine: _get_client
# Parameters: none
# Returns: A client name (from CLIENT=...)
# Purpose: To extract a client from within the object. Assumes that $self->
#          {_clients} was already initialized (in new). In addition to
#          returning a client, the method shifts the first client off of the
#          beginning of the array and pushes it onto the end of the array to
#          try to get a uniform distribution across all clients for issuing
#          DNS commands.
###############################################################################

sub _get_client {

    $Trace->entry();
    my $self = shift(@_);

    # Pull out the existing array of clients

    my @client_array = @{$self->{_clients}};

    # Shift the client off the beginning and push it back on the end of
    # the array to rotate

    my $client = shift(@client_array);
    push(@client_array, $client);

    # Save the values for the next run

    $self->{_clients} = \@client_array;

    $Trace->exit(msg => "Found client: $client");
    return $client;
} ## end sub _get_client

1;
