# $Id: $

## @summary Failover - library to support failover groups and targets
## @author dl-nwkg-test@netapp.com
## @pod here


=head1 NAME

Failover - library to manage failover groups and targets.   A failover target is a node/port pair.   It is represented
internally as a string of 'node:pair' similarly to the way the filer displays it.

=head1 SYNOPSIS

    use Network::Failover qw(create_failover_group, delete_failover_group);


    my $ret  = create_failover_group($group => $group_name,....);
    #check return
    #do other stuff
    my $ret  = delete_failover_group($group => $group_name,....);


=head1 DESCRIPTION

	This library has a variety of methods to handle dealing with failover groups and targets.
	A failover target is a string of the format 'node:port'.


=head1 ROUTINES

=cut

package Network::Failover_Utils;

use strict;
use warnings;
use Tharn;

use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF BOOLEAN);
use base qw(Exporter);
use Data::Dumper;
our @EXPORT_OK = qw(get_node_for_failover_target, get_port_for_failover_target, get_failover_group,
		get_lif_failover_targets, failover_target_state_change, modify_all_failover_targets, fail_lif_to_incorrigible,
		get_failover_group_targets, create_failover_group, delete_failover_group );
		
#############################################################################
#                                NACL STUFF
#############################################################################
use NACL::C::Node;
use NACL::C::Interface;
use NACL::C::NetworkPort;


#############################################################################
#                            FEATURE SPECIFIC LIBS
#############################################################################
use Network::SCON_Utils;

my $clus_node = NACL::C::Node->find();

# Explicitly create a CMode/CLI APISet for the node
my $APISet = NACL::APISet->new(
    hostobj 	=> $clus_node->hostrec(),
    category 	=> "Node",
    interface 	=> "CLI",
    set 		=> "CMode"
);

=head2 get_node_for_failover_target

    my $ret = get_node_for_failover_target(
			target => $target);

This routine returns the node of a failover target of the form 'node:port'

It returns the node as a string on success and dies on failure.

The following named arguments are accepted,

=over

=item target

(Required) : failover target in the form of 'node:port'

=back 

=for html <hr>

=cut
sub get_node_for_failover_target {
	my %params = validate (
		@_,
		{
			target  => {type => SCALAR, optional => 0},
		}
	);
	
	$params{target} =~ m/(.*):.*/;
	#$Log->debug("For target $params{target} returning node $1");
	return $1; 	
}

=head2 get_port_for_failover_target

    my $ret = get_port_for_failover_target(
			target => $target);

This routine returns the port of a failover target of the form 'node:port'

It returns the port as a string on success and dies on failure.

The following named arguments are accepted,

=over

=item target

(Required) : failover target in the form of 'node:port'

=back 

=for html <hr>

=cut
sub get_port_for_failover_target {
	my %params = validate (
		@_,
		{
			target  => {type => SCALAR, optional => 0},
		}
	);
	
	$params{target} =~ m/.*:(.*)/;
	#$Log->debug("For target $params{target} returning port $1");
	return $1; 	
}


=head2 get_failover_group

    my $ret = get_failover_group(
    		clus_node	=> $node,
			lif 		=> $lif_name,
			vserver 	=> $vserver_name);

This routine returns the name of the failover group for the given lif

It returns the group as a string on success and dies on failure.

The following named arguments are accepted,

=over

=item clus_node

(Required) : cluster node for context

=item lif

(Required) : name of the lif

=item  vserver

(Optional) : vserver name. Default is 'undef'.

=back 

=for html <hr>

=cut
sub get_failover_group
{
	
	my %params = validate (
		@_,
		{
			clus_node	=> (type => SCALAR, optional => 0),
			lif  		=> {type => SCALAR, optional => 0},
			vserver 	=> {type => SCALAR, optional => 1, default => undef},
		}
	);
	
	my @vserver_option;
	if (defined($params{vserver})) {
		push(@vserver_option, "vserver" => $params{vserver});
	} 
	
	
	my $node_obj = NACL::C::Node->find(
		command_interface	=> $clus_node,
        filter  => {
            node => $params{clus_node},
        });
	
	my $lif = NACL::CS::Interface->fetch(
				command_interface => $node_obj,
				filter => {
					lif => $params{lif}, 
					@vserver_option,
				},
			);
			
	my $group = $lif->failover_group();
	#$Log->debug("The failover group of lif $params{lif} is $group");
	return $group;
	
}


=head2 get_lif_failover_targets

    my $ret = get_lif_failover_targets(
			lif 	=> $lif_name,
			vserver => $vserver_name);

This routine returns an arrayref of failover targets in order of priority

It returns the arrayref on success and dies on failure.

The following named arguments are accepted,

=over

=item lif

(Required) : name of the lif

=item  vserver

(Optional) : vserver name. Default is 'undef'

=back 

=for html <hr>

=cut
sub get_lif_failover_targets
{
	my %params = validate (
		@_,
		{
			lif  	=> {type => SCALAR, optional => 0},
			vserver => {type => SCALAR, optional => 1, default => undef},
		}
	);
	
	my @failover_targets = ();
	
	# Get failover targets
	my @vserver_option;
	if (defined($params{vserver})) {
		push(@vserver_option, "vserver" => $params{vserver});
	} 
	
	Network::SCON_Utils::debug_info();
	
	my $output = $APISet->network_interface_show(
	    	failover => "",
	    	lif      => $params{lif},
	    	@vserver_option,
	);
	#$Log->debug("The response  is " . Dumper($output));
	if ($output->{output} =~ m/.*Failover Targets: (.*)/s) {
		@failover_targets = split(',\s+', $1);
	
		#trim
		foreach my $val (@failover_targets) {
			$val =~ s/\s+$//;
  		};
	} else {
		$Log->debug("No specific failover targets specificed for lif $params{lif}");
	};

	return \@failover_targets;
}


=head2 failover_target_state_change

    my $ret = failover_target_state_change(
			target 	=> $target_spec,
			state   => $admin_state);

This routine sets the given failover target to the given state

It returns 1 on success and dies on failure.

The following named arguments are accepted,

=over

=item target

(Required) : failover target in the form of 'node:port'

=item  vserver

(Required) : state to modify port to

=back 

=for html <hr>

=cut
sub failover_target_state_change {
	my %params = validate (
		@_,
		{
			target  => {type => SCALAR, optional => 0},
			state	=> {type => SCALAR, optional => 0},
			sleep	=> {type => SCALAR, optional => 1, default => 60},
		}
	);
	
	my $target_port = get_port_for_failover_target(target => $params{target});
	my $target_node = get_node_for_failover_target(target => $params{target});

	my $cmd_int = NACL::C::Node->find(
					command_interface => $clus_node,
					filter => {node => $target_node},
				  );
	my $port_obj = NACL::C::NetworkPort->find(
					command_interface => $cmd_int,
					filter => {	
						port => $target_port,
						node => $target_node,
					},
	);	
	
	$port_obj->modify('up-admin' => $params{state});
	my $incr   = 5;
	my $ticks  = $params{sleep};
	my $status = 0;
	
	while ($ticks) {
		sleep($incr);
		my $port_state = $port_obj->state();
		my $state = $port_state->{up_admin};
		if ($port_state->up_admin() eq $params{state}) {
			$ticks  = 0;
			$status = 1;
		} else {
			$ticks = $ticks - $incr;
		}
	}
	
	if (!$status){
		$Log->debug("Unable to change $target_port on $target_node to state $params{state}");
	}
	return $status;		  
	
}



=head2 modify_all_failover_targets

    my $ret = modify_all_failover_targets(
    		clus_node	=> $node,
			lif 		=> $lif_name,
			vserver 	=> $vserver_name,
			state  		=> $admin_state);

This routine sets the failover targets of the given lif to the given state

It returns 1 on success and dies on failure.

The following named arguments are accepted,

=over

=item clus_node

(Required) : cluster node for context

=item lif

(Required) : name of the lif

=item  vserver

(Required) : vserver name. 

=item  state

(Required) : state to set failover targets to. 

=back 

=for html <hr>

=cut
sub modify_all_failover_targets {
	my %params = validate (
		@_,
		{
			clus_node	=> {type => SCALAR, optional => 0},
			lif  		=> {type => SCALAR, optional => 0},
			vserver 	=> {type => SCALAR, optional => 0},
			state		=> {type => SCALAR, optional => 0},
			sleep		=> {type => SCALAR, optional => 1, default => 60},
		}
	);
	
	my $targets = get_lif_failover_targets(
						lif 	=> $params{lif}, 
						vserver => $params{vserver},
	);
	
	foreach my $target (@$targets) {
		#$Log->debug ("Target is $target");
		failover_target_state_change(
			target	=> $target,
			state 	=> $params{state},
			sleep 	=> $params{sleep},
		);
	};
	
	my $node_obj = NACL::C::Node->find(
        command_interface => $clus_node,
        filter  => {
        	node => $params{clus_node},
        });
	
	my $status   = NACL::CS::Interface->fetch(
					command_interface => $node_obj,
					filter => {
						lif 	=> $params{lif}, 
						vserver => $params{vserver},
					}
	);
	
	my $state = $params{state} eq 'true' ? 'up' : 'down';
	 
	if ($status->status_oper() ne $state) {			
		$Log->warn("Fail: The lif $params{lif} did not move to state $state.  "
			. "State is " . $status->status_oper());
		Network::SCON_Utils::debug_info();
		return 0;
	}
	return 1;
	
	
}

=head2 fail_lif_to_incorrigible

    my $ret = fail_lif_to_incorrigible(
    		clus_node	=> $node,
			lif 		=> $lif_name,
			vserver 	=> $vserver_name,);

This routine fails the given lif to incorrigible

It returns 1 on success and dies on failure.

The following named arguments are accepted,

=over

=item clus_node

(Required) : cluster node for context

=item lif

(Required) : name of the lif

=item  vserver

(Required) : vserver name. 

=back 

=for html <hr>

=cut
sub fail_lif_to_incorrigible {
		
	return modify_all_failover_targets(
			@_,
			state	  	=> 'false',
		);
}



=head2 get_failover_group_targets

    my $ret = get_failover_group_targets(
			group => $group_name,);

This routine returns an array ref to the failover targets in the given group

It returns arrayref on success and dies on failure.

The following named arguments are accepted,

=over

=item  group

(Required) : failover group name. 

=back 

=for html <hr>

=cut
sub get_failover_group_targets {
	my %params = validate (
		@_,
		{
			group  	=> {type => SCALAR, optional => 0},
		}
	);
	my $targets = $APISet->network_interface_failover_groups_show('failover-group' => $params{group});
    #$Log->debug("The response  is " . Dumper($targets));
    
	my $target_array = $targets->{parsed_output};
	my @target_list  = ();
	foreach my $target (@$target_array) {
		push(@target_list, $target->{node} . ':' . $target->{port});
	}
	
	#$Log->debug("The response  is " . Dumper(\@target_list));
	return \@target_list;

}


=head2 create_failover_group

    my $ret = create_failover_group(
			group => $group_name,);

This routine creates a failover group with the given name on the given node and using the given failover targets

It returns 1 on success and dies on failure.

The following named arguments are accepted,

=over

=item  group

(Required) : failover group name. 

=item  node

(Optional) : node for failover target, default is random cluster node

=item  port

(Optional) : port to add to the group, default is all up data ports

=back 

=for html <hr>

=cut
sub create_failover_group {
	
	my %params = validate (
		@_,
		{
			group  	=> {type => SCALAR,   optional => 0},
			node	=> {type => SCALAR,   optional => 0},
			port 	=> {type => SCALAR,   optional => 1},
		}
	);
	
    my @target_list = ();
    
    my $node_obj = NACL::C::Node->find(
    	command_interface => $clus_node,
        filter  => {
            node => $params{node},
        });
    
    # if the ports aren't requested, build a list of all up data ports 
    if (!defined($params{port})) {
    	my @port_objs = NACL::C::NetworkPort->find(
    		command_interface => $node_obj,
    		filter => { role  => 'data', 
   						link  => 'up',
    		},
    	);
    	foreach my $port_obj (@port_objs) {
    		push(@target_list, $port_obj->{node} . ':' . $port_obj->{port});
    	};
    } else {
    	push(@target_list, $params{node} . ':' . $params{port});
    };
    
    # Add each failover target to the group
	foreach my $target (@target_list) {	
		my $target_port;
		my $target_node;
		try {
			$target_port = get_port_for_failover_target(target => $target);
			$target_node = get_node_for_failover_target(target => $target);
	
	   		$APISet->network_interface_failover_groups_create(
	    			'failover-group' 	=> $params{group},
	    			node				=> $target_node,
	    			port				=> $target_port,
			);
  			$Log->debug("Adding port $target_port on node $target_node");		
		} otherwise {
			$Log->comment("Port $target_port on node $target_node was not successfully added to failover group $params{group}");
			my $result = shift();
			$result->throw();
		};
	};
	return 1;
}

=head2 delete_failover_group

    my $ret = delete_failover_group(
			group => $group_name,
			node  => $node_name);

This routine deletes a failover group with the given name and hosted on the given node

It returns 1 on success and dies on failure.

The following named arguments are accepted,

=over

=item  group

(Required) : failover group name. 

=item  node

(Optional) : node(or list of nodes) for failover target, default is all nodes

=item  port

(Optional) : port(or list of ports) for failover target, default is all ports

=back 

=for html <hr>

=cut
sub delete_failover_group{
	my %params = validate (
		@_,
		{
			group  	=> {type => SCALAR, optional => 0},
			node	=> {type => SCALAR, optional => 1, default => "*"},
			port    => {type => SCALAR, optional => 1, default => "*"},
		}
	);


	$APISet->network_interface_failover_groups_delete('failover-group' => $params{group},
														port => $params{port},
														node => $params{node});
													
	return 1;
}

=head2 show_failover_group

    my $ret = show_failover_group(
			group => $group_name,);

This routine shows a failover group with the given values

It returns 1 on success and dies on failure.

The following named arguments are accepted,

=over

=item  group

(Required) : failover group name. 

=item  node

(Optional) : node(or list of nodes) for failover target, default is all nodes

=item  port

(Optional) : port(or list of ports) for failover target, default is all ports

=back 

=for html <hr>

=cut
sub show_failover_group {
	my %params = validate (
		@_,
		{
			group  	=> {type => SCALAR, optional => 0, default => "*"},
			node	=> {type => SCALAR, optional => 1, default => "*"},
			port    => {type => SCALAR, optional => 1, default => "*"},
		}
	);

    $APISet->network_interface_failover_groups_show(
    		'failover-group' 	=> $params{group},
    		node				=> $params{node},
		   	port				=> $params{port},
	);
	
	return 1;
	
}



=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES



=head1 AUTHOR

    dl-nwkg-test (dl-nwkg-test@netapp.com)
    $Id: $
    Last revised $Date: $

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut

1;