# Perl Module
# $Id:
# $Date 
# vim: ft=perl
#
# Purpose:  Wrapper around other IP address libs from CPAN to 
#           get IP addresses into a common format for use in thpl's
#
#
## @summary IP Adrress utility
##
## @description Can take multiple ip address syntax to expan an ip list
##
## @author      parsons@netapp.com
##
## @pod here
#
# POD at end of file

package Network::IPAddress;

use strict;

use Net::IP;
use NetAddr::IP;
use Data::Dumper; $Data::Dumper::Sortkeys = 1;
use Tharn; 


#=============================================================================
#
#  Constructor
#
# Args
#    address  => <str>   1.1.1.1-1.1.1.10
#                           1.1.1.1+9        # start at 1.1.1.1 and add next 9 addresses
#                           1.1.1.1,1.1.1.2,1.1.1.3,...
#                           1.1.1.1
#
#    netmask => <str>  ip netmask to use
#    gateway => <str>  ip addresses gateway address
#
#    add_host_list => <0|1>
#
# Return
#   wantarray ? ($_self, $_error) : $_self
#
#  $obj->{'addresses'}    = number of addresses (hashes) in $obj->{IPAddr}
#  $obj->{'next_address'} = next index to the hash in AoH ($obj->{IPAddr}) 
#
#  $obj->{IPAddr} = [ {
#      '_first'        => '3.3.3.1',
#      '_hosts'        => '14',
#      '_last'         => '3.3.3.14',
#      '_usable'       => 1,
#      'address'       => '3.3.3.1',
#      'address_len'   => 32,
#      'broadcast'     => '3.3.3.15/28',
#      'broadcast_id'  => '3.3.3.15',
#      'broadcast_len' => '28',
#      'gateway'       => '3.3.3.1',
#      'netmask'       => '255.255.255.240',
#      'netmask_len'   => 28,
#      'network'       => '3.3.3.0/28',
#      'network_id'    => '3.3.3.0',
#      'network_len'   => '28',
#      'prefix'        => '3.3.3.1-15',
#      'range'         => '3.3.3.0 - 3.3.3.15',
#      'version'       => 4
#   },
#   {}, 
#   {},
#  ]
#
sub new {
   my $_proto = shift;
   my $_class = ref($_proto) || $_proto;
   my $_self  = {};
   # bless object
   bless($_self, $_class);

   # get object arguments
   my %_arg = @_;

   my ($_a, $_stat, $_err);

   # Init object values
   @{$_self->{'IPAddr'}}        = ();
   $_self->{'addresses'}        = 0;
   $_self->{'next_address_idx'} = 0;

   # load addresses into object   
   ($_stat, $_err) = $_self->add_ip_address(@_);

   if ($_stat) { return (wantarray ? ($_self, $_err) : $_self); }
   else        { return (wantarray ? (undef, "IPAddress: $_err") : undef); }

}
#
#-----------------------------------------------------------------------------
#
# Method to add IP address to object.
# For the most part constructs the object but is separate so that
# it can be called as a separate method to 'add to' the object
#
# See constructor comments
# 
# Args
#   address => <str>
#   netmask => <str>
#   gateway => <str> 
#   add_host_list => <0|1>
#   
# 
sub add_ip_address {

   # not all perl distro's have  both Net::IP and NetAddr::IP
   # usually NetAddr::IP is missing
   # this code should make it so if perl distro is missing
   # the priv libs get used until all distro's are updated
   #$Log->comment("### IPAddr::add_ip_address: determine NetAddr::IP location");
   #
   #{eval 'require NetAddr::IP';}
   #
   #if ( $@ ) {
   #   $Log->warn("### IPAddr::add_ip_address: Using priv copy of NetAddr::IP " 
   #             ."since Perl distro on this host does not contain it"
   #  );
   #   use lib qw(/x/eng/kc-users/parsons/perl/lib/lib/perl5/site_perl/5.8.8/x86_64-linux-thread-multi
   #              /x/eng/kc-users/parsons/perl/lib/lib/perl5/site_perl/5.8.8
   #   );
   #}
   #else {
   #   $Log->comment("### IPAddr::add_ip_address: Using NetAddr::IP from Perl distro"); 
   #}
   #use NetAddr::IP;


  my $_obj = shift;
  my %_arg = @_;
  my ($_a, $_ok, $_error, $_ip,
       $_t, %_hash, $_h,
       $_gw, $_v, $_netmask
   );

   my $_added = 0;
   my @_addrs = ();
 
   # load supported args into object
   foreach $_a (keys %_arg) {
      if    ($_a =~ /^address$/i)       { $_obj->{$_a}  = delete($_arg{$_a}); }
      elsif ($_a =~ /^netmask$/i)       { $_obj->{$_a}  = delete($_arg{$_a}); }
      elsif ($_a =~ /^gateway$/i)       { $_obj->{$_a}  = delete($_arg{$_a}); }
      elsif ($_a =~ /^add_host_list$/i) { $_obj->{$_a}  = delete($_arg{$_a}); }
      else {
         $_error = "IPAddress::add_ip_address: unsupported arg: $_a";
         return (wantarray ? (undef, $_error) : undef );
      }
   }

   # must have IP netmask
   if ( !defined($_obj->{'netmask'}) ) {
      $_error = "IPAddress::add_ip_address: netmask is a required arg";
      $Log->warn("IPAddr::: $_error");
      return (wantarray ? (undef, $_error) : undef );
   }

   # backporting a general netmask function from IPManager 2.0 libs
   # need to normalizt netmask to an address and netmask_bits to an integer
   # so need to make use of object values IPM2.0 contains in it object
   #
   # determine version manually
   # Since there are multiple input syntax, can not use any Net::IP functions
   # Safe to assume a v4 address will have a '.'  anything eles is v6
   if    ( $_obj->{'address'} =~ /\./ )            { $_obj->{'_version'} = 4; }
   elsif ( $_obj->{'address'} =~ /:/ )             { $_obj->{'_version'} = 6; }
   elsif ( $_obj->{'address'} =~ /[0-9a-fA-F:]+/)  { $_obj->{'_version'} = 6; }
   else {
      $_error =  "IPAddr::add_ip_address: unable to determine ip version ";
      $_error .= "from [ $_obj->{'address'}]";
      return (wantarray ? (undef, $_error) : undef );
   }
   $Log->comment("IPAddr::dd_ip_address: IPM 2.0 netmask checking");
   $Log->comment("IPAddr::add_ip_address: IP version determined to be $_obj->{'_version'}");

   # Set address length based on derived version
   $_obj->{'_bits'} =  32 if $_obj->{'_version'} == 4;
   $_obj->{'_bits'} = 128 if $_obj->{'_version'} == 6;

   ($_ok, $_error) = $_obj->my_netmask;
   unless ( $_ok ) {
      return (wantarray ? (undef, $_error) : undef );
   }
   $Log->comment(sprintf("IPAddr::add_ip_address: netmask string [%s] int [%s]\n",
                        $_obj->{'last'}{'netmask'},  $_obj->{'last'}{'netmask_bits'}
                 )    
   );
   # end of IPM2.0 logic port




   $Log->comment("IPAddr::: Make IP Address Data Structure for [$_obj->{'address'}]");

   # Expand in to a list with Net::IP
   #
   # look for comma separated list
   if ( $_obj->{'address'} =~ /,/ ) {
      $Log->comment("IPAddr::: Address seen as list: $_obj->{'address'}");
      $_obj->{'address'} =~s/\s+//g;  # in case there is some spaces
      @_addrs = split(',', $_obj->{'address'});
   }
   # look for range ip1 - ipN
   #                ip1 + count
   elsif ( $_obj->{'address'} =~ /\-/  || $_obj->{'address'} =~ /\+/) {
      $Log->comment("IPAddr::: Address seen as range: $_obj->{'address'}");
      if ( $_obj->{'address'} =~ /\+/) {
         # the '+' notation needs spaces
         $_obj->{'address'} =~ s/(\S)\+(\S)/$1 \+ $2/ if $_obj->{'address'} =~ /(\S)\+(\S)/;
         $_obj->{'address'} =~ s/(\S)\+\s+/$1 \+ /    if $_obj->{'address'} =~ /(\S)\+\s+/;
         $_obj->{'address'} =~ s/\s+\+(\S)/ \+ $1/    if $_obj->{'address'} =~ /\s+\+(\S)/;
         $Log->comment("IPAddr::: changed to  [$_obj->{'address'}]");
      }
      ($_ok, $_error) = $_obj->net_ip($_obj->{'address'}, \@_addrs);
      if ($_ok) { $Log->comment("IPAddr::: Addresses extracted: $_ok"); }
      else      { return (wantarray ? (undef, "IPAddress::add_ip_address: $_error") : undef ); }
   }
   # look for a block  ip/mask
   elsif ( $_obj->{'address'} =~ /\// ) {
      $Log->comment("IPAddr::: Address seen as block: $_obj->{'address'}");
      ($_ok, $_error) = $_obj->net_ip($_obj->{'address'}, \@_addrs);
      if ($_ok) { $Log->comment("IPAddr::: Addresses expanded: $_ok"); }
      else      { return (wantarray ? (undef, "IPAddress::add_ip_address: $_error") : undef ); }
   }
   # single ip address
   elsif ( &Net::IP::ip_is_ipv4($_obj->{'address'}) || &Net::IP::ip_is_ipv6($_obj->{'address'})   ) {
      $Log->comment("IPAddr::: Address seen as single address: $_obj->{'address'}");
      ($_ok, $_error) = $_obj->net_ip($_obj->{'address'}, \@_addrs);
      if ($_ok) { $Log->comment("IPAddr::: Addresses expanded: $_ok"); }
      else      { return (wantarray ? (undef, "IPAddress::add_ip_address: $_error") : undef ); }
   }
   # do not know what it is
   else {
      $_error = "IPAddress::add_ip_address: Unsupported syntax: $_obj->{'address'}";
      $Log->warn("IPAddr::: $_error");
      return (wantarray ? (undef, $_error) : undef );
   }
   $Log->debug("IPAddr::: Addresses put in to list: " . Dumper \@_addrs);

   # make sure we have a list of IPs in @_addrs
   unless( scalar(@_addrs) ) {
      $_error = 'IPAddress::add_ip_address: No IPs found in args loaded into IP list';
      return (wantarray ? (undef, $_error) : undef );
   }

   # make sure each addr is valid
   foreach $a (@_addrs) {
      unless ( &Net::IP::ip_is_ipv4($a) || &Net::IP::ip_is_ipv6($a) ) {
         $_error = "IPAddress::add_ip_address: address [$a] not seen as v4 or v6";
         return (wantarray ? (undef, $_error) : undef ); 
      }
   } 
   # check validity of gateway
   if ( $_obj->{'gateway'} ) {
      $a = $_obj->{'gateway'};
      unless ( &Net::IP::ip_is_ipv4($a) || &Net::IP::ip_is_ipv6($a) ) {
         $_error = "IPAddress::add_ip_address: gateway address [$a] not seen as v4 or v6";
         return (wantarray ? (undef, $_error) : undef );
      }
   }

   # check the mask, a mask should not have a 1 follow a 0
   $_v = &Net::IP::ip_get_version($_obj->{'netmask'});
   $_a = &Net::IP::ip_iptobin($_obj->{'netmask'}, $_v);
   if ( $_a =~ /01/ ) {
      $_error = "IPAddress::add_ip_address: Invalid netmask [$_obj->{'netmask'}] [$_a] ver $_v";
      return (wantarray ? (undef, $_error) : undef );
   }

   # Populate info for the IP address in the Object
   # use NetAddr::IP to get specific info on the IP
   $Log->comment("IPAddr::: Expand addresses:"); 
   foreach $_a (@_addrs) {
      %_hash = ();
      # address
      $Log->debug("IPAddr::: NetAddr address: $_a");
      $_hash{'address'} = $_a;

      # NetAddr::IP does not like when a mask has a slash
      $_netmask = $_obj->{'netmask'};
      if ( $_netmask =~ /\/\d+/ ) {
         $_netmask =~ s/\///;
         $Log->debug("IPAddr:::    Netmask: changed given $_obj->{'netmask'} to $_netmask");
      }
      $Log->debug("IPAddr:::    Netmask: $_netmask");
      # create NetAddr::IP object
      $_ip = NetAddr::IP->new($_a, $_netmask);
      # address lenth
      $_hash{'address_len'} = $_ip->bits();
      # version
      $_hash{'version'} = $_ip->version();

      # netmask, always expanded
      # make sure v6 is expanded
      if ( $_hash{'version'} == 6 ) {$_hash{'netmask'} = &Net::IP::ip_expand_address($_ip->mask(), 6);}
      else                          {$_hash{'netmask'} = $_ip->mask();}
      $_hash{'netmask_len'} = $_ip->masklen();
      $Log->debug("IPAddr:::    netmask: $_hash{'netmask'}  len: $_hash{'netmask_len'}");
      # TODO compare arg and object value
      # network syntax network/mask
      $_t = sprintf("%s", $_ip->network());
      $_hash{'network'} = $_t;
      ($_hash{'network_id'}, $_hash{'network_len'}) = split('\/', $_t);
      $Log->debug("IPAddr:::    network: $_t  [$_hash{'network_id'}] [$_hash{'network_len'}]");
      # broadcast
      $_t = sprintf("%s", $_ip->broadcast());
      $_hash{'broadcast'} = $_t;
      ($_hash{'broadcast_id'}, $_hash{'broadcast_len'}) = split('\/', $_t);
      # prefix
      $_hash{'prefix'} = $_ip->prefix();
      # range 
      $_hash{'range'} = $_ip->range();
      # first and last
      $_hash{'_first'} = $_ip->first()->addr();
      $_hash{'_last'}  = $_ip->last()->addr();
      $_hash{'_hosts'} = $_ip->num();

      # see if address is usable
      if ( $_hash{'address'} ne $_hash{'network_id'} && $_hash{'address'} ne $_hash{'broadcast_id'} ) {
         $_hash{'_usable'} = 1;
      }
      else {
         $_hash{'_usable'} = 0;
      }

      # check the gateway
      if ( $_obj->{'gateway'} ) {
         $Log->debug("IPAddr:::    Gateway: $_obj->{'gateway'} Mask: $_netmask [$_obj->{'netmask'}]");
         $_gw = NetAddr::IP->new($_obj->{'gateway'}, $_netmask);
         # Loop thru usable address to see if gw is in the list of usable addresses
         # Was hard to use a Net Object method for this
         #   - case where using network id or broadcast as gatway was seen as a
         #     valid ip in the objects, since it was string matches
         #   - did not use NetAddr::IP::hostenum, so as to not generate big
         #     list in memeory when the gw is usually the first usable
         $_ok = 0;
         # TODO fix in ipm2.0, v6 max value is to big of an int
         #foreach $_t ( 1..$_ip->num() ) {
         foreach $_t (1..10001) {
             $Log->debug(sprintf("IPAddr:::      addr:%-6u  ip: %s",
                                     $_t,  $_ip->nth($_t-1)->addr()
                                  )
             );
             if ( $_gw->addr() eq $_ip->nth($_t-1)->addr() ) {
                $_ok = 1;
                $Log->debug(sprintf("IPAddr:::      gw matched [%u] %s", $_t, $_gw->addr()));
                last;
             }
         }
         if ( $_ok ) {
            $Log->debug("IPAddr:::    Gateway: added to hash");
            $_hash{'gateway'} = $_obj->{'gateway'};
         }
         else {
            $Log->warn("IPAddr:::    Gateway: not a usable host");
         }
      }
      # make sure v6 is expanded
      if ( $_hash{'version'} == 6 ) {
         $Log->comment("IPAddr::: Ensuring IPv6 addresses expanded and lowercase");
         $_hash{'address'}      = &Net::IP::ip_expand_address($_hash{'address'}, 6);
         $_hash{'broadcast'}    = &Net::IP::ip_expand_address($_hash{'broadcast'}, 6);
         $_hash{'broadcast_id'} = &Net::IP::ip_expand_address($_hash{'broadcast_id'}, 6);
         $_hash{'gateway'}      = &Net::IP::ip_expand_address($_hash{'gateway'}, 6);
         $_hash{'network'}      = &Net::IP::ip_expand_address($_hash{'network'}, 6);
         $_hash{'network_id'}   = &Net::IP::ip_expand_address($_hash{'network_id'}, 6);
         $_hash{'_first'}       = &Net::IP::ip_expand_address($_hash{'_first'}, 6);
         $_hash{'_last'}        = &Net::IP::ip_expand_address($_hash{'_last'}, 6);
      }
      # store hash in object 
      push(@{$_obj->{'IPAddr'}}, {%_hash});
      $_obj->{'addresses'}++;
      $_added++;
   }

   $Log->debug("IPAddr::: Object IPAddr: " . Dumper $_obj->{'IPAddr'});

   return( wantarray ? ($_added, '') : $_added); 

}

#-----------------------------------------------------------------------------
#
# Method to give a list of IP Adresses only
# Populates:
#   @{$_obj->{'ip_address_list'}} with unique IPs 
#
# Args
#   none
#
# Return
#   wantarray ? ( count, error) ? count
#
# $obj->{'ip_address_list'} = [ip1, ip2, ip3, ...]
# 
sub ip_address_list {

   my $_obj = shift;
   my $_href;
   my %_cache = ();
   my $_c = 0;

   $Log->comment("IPAddr::: Extract IP address from object into a list");

   # loop thru object and extract uniqes IPs
   foreach  $_href ( @{$_obj->{'IPAddr'}} ) {
      if ( exists($_cache{$_href->{'address'}}) ) {
         $Log->warn(sprintf("IPAddr::: Duplicate IP: %s at element %u also at element %u",
                               $_href->{'address'}, 
                               $_c,
                               $_cache{$_href->{'address'}},
                       )
         );          
      }
      else {
         push(@{$_obj->{'ip_address_list'}}, $_href->{'address'}); 
         $_cache{$_href->{'address'}} = $_c;
      }
      $_c++;
   }
  
   $Log->debug("IPAddr::: Object ip_address_list: " . Dumper $_obj->{'ip_address_list'});
   $_c = $#{$_obj->{'ip_address_list'}} + 1;

   return (wantarray ? ($_c, 'noError') : $_c );
}


#-----------------------------------------------------------------------------
#

sub isa_ipv6 {
   my $_obj = shift;
   $Log->comment("... IPAddress: isa IPv6: $_[0]");
   if ( &Net::IP::ip_is_ipv6($_[0]) ) { $Log->comment(" its 6"); }
   else                              { $Log->comment(" its 4"); }
}


#-----------------------------------------------------------------------------
#
# Method to give a list of unique Network IDs
# Populate:
#   @{$_obj->{'network_id_list'}}     # 3.3.3.0
#   @{$_obj->{'network_list'}}        # 3.3.3.0/25
# 
# Args
#   none
#
# Return
#   wantarray ? ( count, error) ? count
#
# $obj->{'network_id_list'} = [ 'net_id', ...]                      3.3.3.0
# $obj->{'network_list'}    = [ 'net_id/length', ...]               3.3.3.0/28
# $obj->{'Network'}{<net_id>}{'broadcast'}   = broadcast address    3.3.3.15
#                            {'gateway'}     = gateway address      3.3.3.7,
#                            {'netmask'}     = network mask         255.255.255.240
#                            {'netmask_len'} = mask length          28
#                            {'hosts'}     = [ip1, ip2, ip3, ...]   3.3.3.1 .. 3.3.3.14
#
sub network_id_list {

   my $_obj = shift;
   my ($_href, $_ip);
   my %_cache = ();
   my $_c = 0;

   $Log->comment("IPAddr::: Extract Network IDs from object into a list");

   # loop thru object and extract uniqes IPs
   foreach  $_href ( @{$_obj->{'IPAddr'}} ) {

      next if $_cache{$_href->{'network_id'}};

      $Log->comment("IPAddr::: push network_id $_href->{'network_id'}");
      push(@{$_obj->{'network_id_list'}}, $_href->{'network_id'});
      push(@{$_obj->{'network_list'}},    $_href->{'network'});
      $_cache{$_href->{'network_id'}}++;

      $_obj->{'Network'}{$_href->{'network_id'}}{'broadcast'}   = $_href->{'broadcast_id'};
      $_obj->{'Network'}{$_href->{'network_id'}}{'netmask'}     = $_href->{'netmask'};
      $_obj->{'Network'}{$_href->{'network_id'}}{'netmask_len'} = $_href->{'netmask_len'};
      if ( $_href->{'gateway'} ) {
         $_obj->{'Network'}{$_href->{'network_id'}}{'gateway'} = $_href->{'gateway'};
      }

      # add_host_list flag set
      # make list of valid host for each unique network id
      $_c = 1;
      if ( $_obj->{'add_host_list'} ) {
         $Log->comment("IPAddr::: Get host list for net: $_href->{'network_id'}");
         $_ip = Net::IP->new("$_href->{'_first'} - $_href->{'_last'}");
         do {
            push(@{$_obj->{'Network'}{$_href->{'network_id'}}{'hosts'}}, $_ip->ip());
            $Log->debug(sprintf("   IPAddr::: host %-6u %s", $_c, $_ip->ip()));
            $_c++;
         } while ++$_ip;
      }
   }

   $Log->debug("IPAddr::: Object network_id_list: " . Dumper $_obj->{'network_id_list'});
   $Log->debug("IPAddr::: Object network_list: "    . Dumper $_obj->{'network_list'});
   $Log->debug("IPAddr::: Object Network: "         . Dumper $_obj->{'Network'}); 
   $_c = $#{$_obj->{'network_id_list'}} + 1;

   return (wantarray ? ($_c, 'noError') : $_c );
}

#-----------------------------------------------------------------------------
# $_[]  = object
# $_[0] = arg for Net::IP
# $_[1] = ref to array to populate (ip1, ip2, ip3, ..., ipN)
#

sub net_ip {
   my $_obj = shift;
   my $_c = 0;
   my $_err = '';

   # Net::IP object
   my $_nip = Net::IP->new($_[0]);
   if ($_nip) {
         # loop thru Net::IP object
         do {
           $Log->debug( sprintf("IPAddr:::   push IP on list [%u]: %s ", $_c, $_nip->ip(), ) );
           push(@{$_[1]}, $_nip->ip());
           $_c++;
         } while (++$_nip);
   }
   else {
      $_err = Net::IP::Error();
      $Log->comment("IPAddr::: Net::IP::Error: $_err");
      return (wantarray ? (undef, $_err) : undef );
   }
   $_c = scalar(@{$_[1]});

   return( wantarray ? ($_c, 'noError') : $_c);
}

#-----------------------------------------------------------------------------
#
# Get addresses from the object
# Will cycle thru the object ip addresses
#
#
sub next_address {

   my $_obj = shift;
   my $_i   = \$_obj->{'next_address_idx'};

   my $_c = $_obj->{'addresses'} - 1;
   $Log->comment("IPAddr::: Next address index set to: $$_i of $_c"); 

   if ( defined($_obj->{'next_address_idx'}) ){
      $_obj->{'next_address'}   = $_obj->{'IPAddr'}[$$_i]{'address'}      || '';
      $_obj->{'next_netmask'}   = $_obj->{'IPAddr'}[$$_i]{'netmask'}      || '';
      $_obj->{'next_broadcast'} = $_obj->{'IPAddr'}[$$_i]{'broadcast_id'} || '';
      $_obj->{'next_network'}   = $_obj->{'IPAddr'}[$$_i]{'network_id'}   || '';
      $_obj->{'next_usable'}    = $_obj->{'IPAddr'}[$$_i]{'_usable'}      || '';

      $_obj->{'next_address_idx'}++; 
      if ( $_obj->{'next_address_idx'} == $_obj->{'addresses'} + 1 ) {
         $Log->comment("IPAddr::: Out of addresses");
         $_obj->{'next_address'} = '';
         $_obj->{'next_netmask'} = '';
         $_obj->{'next_broadcast'} = '';
         $_obj->{'next_network'} = '';
         $_obj->{'next_usable'} = '';
         $_obj->{'next_address_idx'} = undef;
      }
   }
   $_obj->{'next_address'};
}

#
#----------------------------------------------------------------------------
# Determine Netmask and Netmask bits
# Determine mask from various input scenerios
# Need to normalize to a netmask, could of been entered multip ways
#   arg{'address'} network block ip/mask
#   $_arg{'netmask'} 255.255.240.0
#   $_arg{'netmask_bits'} /20
#
# Will populate the object
# Looks at all possible syntaxs and checks them
#   if netmask is given multiple ways, they are checked to resolve the same
#   if a netmask is only given on way, the other values are filled in
#   if address syntax is ip/mask, then this mask value has precedence
#     if netmask and netmask_bits given, they are compared
#     if not given, they are populated from this value
# Operation
#   Look to see what is defined and not defined
#   If something is not defined, then derive it from the other
#   These derived values are held in temp vars and assigned
#   to the object, if need to be at end of function
# 
# Args
#   [] = object
#
# Return
#   return (wantarray ? (undef, $ERROR) : undef );   # Error
#   return (wantarray ? (1, '') : 1 );               # Success
sub my_netmask {
   
   my $_obj = shift;
   my ( $_a, $_n, $_nm,
       $_a_nm, $_a_nm_bits,
       $_bin_str, $_error,
   );
   # init the temp vars
   my $_nm_bits = 0;
   my $_nm_addr = '';
   
   if ( $_obj->{'netmask_len'} =~ /^\// ) { 
          $_obj->{'netmask_len'} =~ s/^\///;
          $Log->comment("IPAddr::my_netmask: strip \/ from netmask_len $_obj->{'netmask_len'}\n");
   }
   if ( $_obj->{'netmask'} =~ /^\// ) {
          $_obj->{'netmask'} =~ s/^\///;
          $Log->comment("IPAddr::my_netmask: strip \/ from netmask $_obj->{'netmask'}\n");
   }


   # if both string and bits are given, make sure they are equivalent
   if ( defined($_obj->{'netmask_len'}) &&  defined($_obj->{'netmask'}) ) {
       $Log->comment(sprintf("IPAddr::my_netmask: check netmask_len [%s] and netmask [%s]\n",
                        $_obj->{'netmask_len'}, $_obj->{'netmask'}
                     )
       );

       $_bin_str =  &Net::IP::ip_iptobin($_obj->{'netmask'}, $_obj->{'_version'});
       # a mask should not have a 1 follow a 0 in the binary string
       if ( $_bin_str  =~ /01/ ) {
          $_error = sprintf("IPAddress::my_netmask: netmask: [%s] result bad bin string [%s]",
             $_obj->{'netmask'},
             $_bin_str,
          );
          return (wantarray ? (undef, $_error) : undef );
       }
       # get bit count based on netmask string and check to given netmask_len
       $_bin_str =~ s/0//g;
       if ( length($_bin_str) ne "$_obj->{'netmask_len'}" ) {
          $_error = sprintf("IPAddress::my_netmask: netmask: [%s] [%s] and netmask_len [%s] diff",
             $_obj->{'netmask'},
             length($_bin_str),
             $_obj->{'netmask_len'}
          );
          return (wantarray ? (undef, $_error) : undef );
       }
   }
   # netmask_len
   # make sure netmask_len is an integer
   # derive netmask string
   elsif ( defined($_obj->{'netmask_len'}) && !defined($_obj->{'netmask'}) ) {
      if ( $_obj->{'netmask_len'} !~ /^\d+$/ ) {
         $_error =  "IPAddress::my_netmask: netmask_len not an int ";
         $_error .= "[$_obj->{'netmask_len'}]";
         return (wantarray ? (undef, $_error) : undef );
      }
      $_nm_bits = $_obj->{'netmask_len'};
      # figure out netmask string
      $_bin_str = sprintf("%s%s",
         '1' x $_obj->{'netmask_len'},
         '0' x ($_obj->{'_bits'} - $_obj->{'netmask_len'})
      );
      $_nm_addr = &Net::IP::ip_bintoip($_bin_str, $_obj->{'_version'});

      $Log->comment("IPAddr::my_netmask: set from netmask_len: nm_bits $_nm_bits nm_addr $_nm_addr\n");
   }

   # netmask
   # parse different syntax for netmask
   #    strip '/'
   #    allow netmask to be an integer, change to string
   # derive netmask_len
   elsif ( defined($_obj->{'netmask'}) && !defined($_obj->{'netmask_len'}) ) {
      # netmask given as int, change to string
      if (  $_obj->{'netmask'} =~ /^\d+$/ ) {
         $_bin_str = sprintf("%s%s",
            '1' x  $_obj->{'netmask'},
            '0' x ($_obj->{'_bits'} - $_obj->{'netmask'})
         );
         $_nm_bits = $_obj->{'netmask'};
         $_nm_addr = &Net::IP::ip_bintoip($_bin_str, $_obj->{'_version'});
      }
      else {
         $_nm_addr =  $_obj->{'netmask'};
         if ( !defined( $_obj->{'netmask_len'}) ) {
            $_bin_str =  &Net::IP::ip_iptobin($_obj->{'netmask'}, $_obj->{'_version'});
            $_bin_str =~ s/0//g;
            $_nm_bits = length($_bin_str);
         }
      }
      $Log->comment("IPAddr::my_netmask: set from netmask: nm_addr $_nm_addr nm_bits $_nm_bits\n");
   }
   # address given as IP/mask
   if (  $_obj->{'address'} =~ /\// ) {
      ($_a, $_nm) = split(/\//,  $_obj->{'address'});

      $Log->comment(sprintf("IPAddr::my_netmask: network address, splitting [%s] [%s]\n", $_a, $_nm));

      # ip/<int>  make a netmask string
      if ( $_nm =~ /^\d+$/ )  {
         $_a_nm_bits = $_nm;
         $_bin_str = sprintf("%s%s",
            '1' x $_a_nm_bits,
            '0' x ($_obj->{'_bits'} - $_a_nm_bits ),
         );
         $_a_nm =  &Net::IP::ip_bintoip($_bin_str, $_obj->{'_version'});
      }
      # ip/<str> v4   make a netmask_len
      elsif (  $_nm =~ /\d+\./ ) {
         $_a_nm = $_nm;
         $_bin_str =  &Net::IP::ip_iptobin($_nm, $_obj->{'_version'});
         $_bin_str =~ s/0//g;
         $_a_nm_bits = length($_bin_str);
         $_obj->{'address'} = sprintf("%s/%s", $_a, $_a_nm_bits);
         $Log->comment("IPAddr::my_netmask: converted v4 netmask string to int: [$_obj->{'address'}]");
      }
      # ip/<str> v6
      elsif (  $_nm =~ /[0-9a-fA-F:]+/ ) {
         $_a_nm = $_nm;
         $_bin_str =  &Net::IP::ip_iptobin($_nm, $_obj->{'_version'});
         $_bin_str =~ s/0//g;
         $_a_nm_bits = length($_bin_str);
         $_obj->{'address'} = sprintf("%s/%s", $_a, $_a_nm_bits);
         $Log->comment("IPAddr::my_netmask: converted v6 netmask string to int: [$_obj->{'address'}]");
      }
      else {
         $_error =  "IPAddress::my_netmask: unable to parse netmask from given ";
         $_error .= "network address [$_obj->{'address'}]";
         return (wantarray ? (undef, $_error) : undef );
      }

      $Log->comment(sprintf("--- set from network address: a_nm [%s] a_nm_bits [%s]",
                       $_a_nm, $_a_nm_bits
                   )
      );


      # make sure netmask from IP/mask  same as netmask_len if given
      if ( defined($_obj->{'netmask_len'}) ) {
        $Log->comment(sprintf("IPAddr::my_netmask: check given netmask_len %s and parsed bits %s\n",
                         $_obj->{'netmask_len'}, $_a_nm_bits,
                     )
         );
         if ( $_obj->{'netmask_len'} ne "$_a_nm_bits" ) {
            $_error = sprintf("IPAddress::my_netmask: [ip/mask] %s diff mask than netmask_len [%s]",
               $_obj->{'address'},
               $_a_nm_bits,
            );
            return (wantarray ? (undef, $_error) : undef );
         }
      }
      # make sure netmask from IP/mask  same as netmask if defined
      if ( defined($_obj->{'netmask'}) ) {
         if ( $_obj->{'netmask'} ne "$_a_nm" ) {
            $_error = sprintf("IPAddress::my_netmask: ip/mask [%s] yields diff netmask than given [%s]",
               $_obj->{'address'},
               $_a_nm,
            );
            return (wantarray ? (undef, $_error) : undef );
         }
      }
   }


   # if we are this far then all the check pass
   # just make sure the object knows about netmask and netmask_len
   if ( $_a_nm and  $_a_nm_bits ) {
      $Log->comment("IPAddr::my_netmask: set object a_nm  $_a_nm a_nm_bits $_a_nm_bits\n");
      #if ( !defined($_obj->{'netmask'}) )     { $_obj->{'netmask'} = $_a_nm; }
      #if ( !defined($_obj->{'netmask_len'}) ) { $_obj->{'netmask_len'} = $_a_nm_bits; }
      $_obj->{'netmask'} = $_a_nm;
      $_obj->{'netmask_len'} = $_a_nm_bits; 
   }
   else {
      $Log->comment("IPAddr::my_netmask: set object nm_addr $_nm_addr nm_bits $_nm_bits\n");
      #if ( !defined($_obj->{'netmask'}) )      { $_obj->{'netmask'} = $_nm_addr; }
      #if ( !defined($_obj->{'netmask_len'}) )  { $_obj->{'netmask_len'} = $_nm_bits; }
      $_obj->{'netmask'} = $_nm_addr;
      $_obj->{'netmask_len'} = $_nm_bits;
   }

   # validate netmask,
   # a mask should not have a 1 follow a 0 in the binary string
   $_bin_str =  &Net::IP::ip_iptobin($_obj->{'netmask'}, $_obj->{'_version'});
   if ( $_bin_str  =~ /01/ ) {
      $_error = sprintf("IPAddress::my_netmask: invalid netmask: [%s]  [%s]",
         $_obj->{'netmask'},
         $_bin_str,
      );
      return (wantarray ? (undef, $_error) : undef );
   }

   # make sure we defined a netmask and netmask_len
   if ( !defined($_obj->{'netmask'}) || !$_obj->{'netmask_len'} ) {
      $_error = sprintf("IPAddress::my_netmask: unable to define netmask [%s] or netmask_len [%s]",
         $_obj->{'netmask'} || '-',
         $_obj->{'netmask_len'} || '-',
      );
      return (wantarray ? (undef, $_error) : undef );
   }

   $Log->comment(sprintf("IPAddr::my_netmask: netmask set [%s] [%s]\n", 
                    $_obj->{'netmask'}, $_obj->{'netmask_len'}
                 )
   );

   return (wantarray ? (1, '') : 1 );
}

#
#.............................................................................
#
# simple subs to give acces to object data structure
#
# next_xxx
sub next_netmask   { my $_obj = shift; $_obj->{'next_netmask'};}
sub next_broadcast { my $_obj = shift; $_obj->{'next_broadcast'};}
sub next_network   { my $_obj = shift; $_obj->{'next_network'};}
sub next_usable    { my $_obj = shift; $_obj->{'next_usable'};}
sub next_reset     { 
   my $_obj = shift;
   $Log->comment("IPAddr::: Reset next_address index start");
   $_obj->{'next_address_idx'} = 0;
}

# return counters/stats or work on object
sub address_count { my $_obj = shift; $_obj->{'addresses'}}; 

sub init_object { my $_obj = shift; $_obj = {}; }


# network
# return the list
sub get_network_id_list { my $_obj = shift; @{$_obj->{'network_id_list'}}; }
sub get_network_list    { my $_obj = shift; @{$_obj->{'network_list'}}; }
# return the hash for the given network id
sub get_network {
   my $_obj = shift;
  if ( exists($_obj->{'Network'}{$_[0]}) ) { return($_obj->{'Network'}{$_[0]}); }
  else                                     { return(undef); }
}

# ip addresses
sub get_ip_address_list { my $_obj = shift; @{$_obj->{'ip_address_list'}}; }
#
# find the hash for given IP
# $_[0] = ip
sub get_ip_address_hash {
   my $_obj = shift;
   my $_href;
   # init the object entry
   %{$_obj->{'current_ip'}} = (); 
   $Log->comment("IPAddr::: Get IP Address hash: $_[0]"); 
   # look for the ip 
   foreach $_href ( @{$_obj->{'IPAddr'}} ) {
      if ( $_[0] eq $_href->{'address'} ) {
         # get network info if exist
         if ( exists($_obj->{'Network'}{$_href->{'network_id'}}{'hosts'}) ) {
            $Log->comment("IPAddr::: Network hash exist for ip $_[0] net $_href->{'network_id'}");
            @{$_href->{'hosts'}} = @{$_obj->{'Network'}{$_href->{'network_id'}}{'hosts'}};
         }
         %{$_obj->{'current_ip'}} = %{$_href};
      }
   }
   $_obj->{'current_ip'};
}

# IP utilities
sub ip_version {
   shift;
   if ( &Net::IP::ip_is_ipv4($_[0]) ) { return(4); }
   if ( &Net::IP::ip_is_ipv6($_[0]) ) { return(6); }
   0;
}


#=============================================================================
#
#     Wrappers to Net::IP Functions
#     Since the Net::IP functions are not directly  available

sub netip_ip_version {
   if ( &Net::IP::ip_is_ipv4($_[0]) ) { return(4); }
   if ( &Net::IP::ip_is_ipv6($_[0]) ) { return(6); }
   0;
}

#-----------------------------------------------------------------------------
# compress an ipv6 address
# Args
#  $_[0] = address
#
sub netip_compress_address {
   my $_ret = $_[0];

   #$Log->comment("IPAddress: compress in $_[0]");  
   if (  &Net::IP::ip_is_ipv6($_[0]) ) {  
      $_ret = &Net::IP::ip_compress_address($_[0], 6);
   }
   #$Log->comment("IPAddress: compress out $_ret");
   $_ret;
}
#-----------------------------------------------------------------------------
sub netip_expand_address {
   my $_ret = $_[0];

   #$Log->comment("IPAddress: expand in $_[0]");  
   if (  &Net::IP::ip_is_ipv6($_[0]) ) {
      $_ret = &Net::IP::ip_expand_address($_[0], 6);
   }
   #$Log->comment("IPAddress: expand out $_ret");
   $_ret;
}


1;

__END__

=pod

=head1 NAME

Network::IPAddress

=head1 DESCRIPTION

Wrapper around Perl's Net::IP and NetAddr::IP Modules to provide
methods to handle IP Address information in QA thpl's. As these
libs both support Ipv4 and IPv6, this lib inherently supports both IP 
versions. Conversly, any problems with this lib propagate up into
Network::IPAddress. Their code may use C code that does not work
or is buggy in Win32 environemnts.

This module supports different types of IP Address syntax for input
and will expand upon the given information and provide accessor
methods to this information. 

=head1 SYNOPSIS

    use strict;
    use Tharn;

    my ($ip, $error) = Network::IPAddress->new(
        address       => <ip address str>,
        gateway       => <ip gateway>,
        netmask       => <dotted decimal netmask>,
        add_host_list => <0|1> flag 
    );

    ($ok, $error) = $ip->ip_address_list;

    ($ok, $error) = $ip->network_id_list;

    $addres    = $ip->next_address;
    $gateway   = $ip->next_gateway;
    $broadcast = $ip->next_broadcast;
    $network   = $ip->next_network;

=head1 PARAMS

None - Intentionally so if exist different params for inputting IP info
into a thpl, did not want this lib to be the policer of the params. Thus
it is up to the user to get the params and supply them to the object constructor.

=head1 Methods

=head2 new()

Object constructor. Expands the given IP address and netmask information into 
individual addresses, broadcast, network, range, usable host and other bits
or information needed for dealing with IP addresses. If given a gateway this
lib will also check the gatway exist in the same network as the given IP 
addresses.

    my ($ip, $error) = Network::IPAddress->new(
       address        => <address str>,
       netmask        => <dotted decimal netmask>,
       [gateway       => <ip gateway>,]
       [add_host_list => <0|1>,]
    );

=head3 Args

=over

=item address => <address str>  

a string to input 1 to N IPv4 or IPv6 address where 
<address string> supports the following syntax:

=over

=item 3.3.3.1-3.3.3.5

A inclusive range where all IP address between and including the start and end 
values are stored in the object. 

=item 3.3.3.1 + 4 

A range of address starting at the full address given and incrementing through
consecutive addresses until N+1 address are reached or till the end of the IP 
range.

=item 3.3.3.1, 3.3.3.2, 3.3.3.3, 3.3.3.4, 3.3.3.5

A comma separated list of valid IP addresses. Do not have to be consecutive
nor the same subnet.

=item 3.3.3/28 

A block of IPs for the given network prefix and mask length.


=item 3.3.3.1

A single valid IP address.

=back

=item netmask => <dotted decimal netmask>

The dotted decimal notation for for the subnet mask

=item gateway => <ip gateway> 

Optional, IP address of the subnet gateway. If given it is checked to be 
a usable host address in the sunet based on the IP address and Netmask given.
If it a usable host it is stored in the object, if not a warning is generated
and it is not stored in the object.

=item add_host_list => <0|1>

Optional flag to store all usable hosts for the subnet in the objects network
section. Default is 0, since doing large subnets would require large lists
of IPs to be stored. This slows down methods that have to cycle through and
is a memory waste if the information is not desired by the user. 

=back

Upon success in array context, this will return a reference to the object and
error string. In scalar context this will return a reference to the object.

Upon failure in array context, this will return undef and an error string.
In scalar context this return undef.

=head3 Operation

The new() method will check the arguments and error if address and netmask 
arguments are not given. Object will then init its IP list then look at 
the address argument. Net::IP will be used to expand the argument to a list
of IP addresses. Net::IP will be used to validate addresses are either
v4 or v6 including the gateway and will check that the netmask
is a sting of 1's followed by 0's. Will then loop through the list of IP 
addresses to build a AoH to store in the object. Information in the 
hash will include address, netmask, broadcast, network, prefix, range, 
version, first and last address, number of hosts, usable host setting and 
gateway (if given). If the address is seen not to be the network and 
broadcast address then the '_usable' is set to 1, otherwise it is 0.

This will stored in the object as:
    $obj->{IPAddr} = [ 
       {
           '_first'        => '3.3.3.1',
           '_hosts'        => '14',
           '_last'         => '3.3.3.14',
           '_usable'       => 1,
           'address'       => '3.3.3.1',
           'address_len'   => 32,
           'broadcast'     => '3.3.3.15/28',
           'broadcast_id'  => '3.3.3.15',
           'broadcast_len' => '28',
           'gateway'       => '3.3.3.1',
           'netmask'       => '255.255.255.240',
           'netmask_len'   => 28,
           'network'       => '3.3.3.0/28',
           'network_id'    => '3.3.3.0',
           'network_len'   => '28',
           'prefix'        => '3.3.3.1-15',
           'range'         => '3.3.3.0 - 3.3.3.15',
           'version'       => 4
       },
       { }, 
       { },
    ]

=head2 add_ip_address()

Method to add address to the object. This is called by the new()
and has the same arguments, see new() for description.

    ($ok, $error) = $ip->add_ip_address( address => <str>,
                                         netmask => <str>,
                                         gateway => <str>,
                                         add_host_list => <0|1>,
    ); 

Upon success, in list context will return how many addresses were added
and error string, in scalar context will return how many addresses were 
added.

Upon failure, in list context will return undef and error string,
in scalar context will return undef.



=head2 ip_address_list()

   ($ok, $error) = $ip->ip_address_list;

Method to create a list of IP addresses from the addresses stored in the 
object. This list will be a simple list of just IPs which are the address
field in each hash of the IPAddr list in the object (see new()).

If a duplicate IP address is found, only the first instance
is stored, meaning the hash is stored in the list at the index
respective to the first instance. Upon seeing a duplicate IP
a warning is generated.

Upon success in list context the count of addresses and 'noError' string
are returned, in scalar context the count of addresses.

Upon failure in list context undef and an error string are returned,
in scalar context just undef.

To access the list directly

   @{$obj->{'ip_address_list'}}


=head2 get_ip_address_list()

    @ip_addresses = $ip->get_address_list;

Method to return the IP address list created with ip_address_list()
method.


=head2 network_id_list()

   ($ok, $error) = $ip->network_id_list;

Method to store information specific to each unique network found.
If multiple IPs are all in one network, then only one network is stored.

If the object has its 'add_host_list' flag set, an argument to new(),
then all usable host addresses are stored.

Upon success in list context the count of networks and 'noError' string
are returned, in scalar context the count of networks.

Upon failure in list context undef and an error string are returned,
in scalar context just undef.

To access directly in the object:

    $obj->{'network_id_list'} = [ 'net_id', ...]                      3.3.3.0
    $obj->{'network_list'}    = [ 'net_id/length', ...]               3.3.3.0/28
    $obj->{'Network'}{<net_id>}{'broadcast'}   = broadcast address    3.3.3.15
                               {'gateway'}     = gateway address      3.3.3.7,
                               {'netmask'}     = network mask         255.255.255.240
                               {'netmask_len'} = mask length          28
                               {'hosts'}     = [ip1, ip2, ip3, ...]   3.3.3.1 .. 3.3.3.14

=head2 get_network_id_list()

   @list = $ip->get_network_id_list();

Method to return the Network ID list created with network_id_list() method.
Network ID strings are the 1st address in the subnet, not a usable address.

=head2 get_network_list()

   @list = $ip->get_network_list();

Method to return the Network list created with network_id_list() method.
Network strings are the <1st address in the subnet>/<mask length>.


=head2 next_xxx()

   $address   = $ip->next_address();
   $netmask   = $ip->next_netmask();
   $broadcast = $ip->next_broadcast();
   $network   = $ip->next_network();

Methods to access the address, netmask, broadcast, network stored in the 
objects 'IPAddr' section. For each call to next_address(), the object 
increment is index for access to the next hash containing the IP information.
See new() for structure of hash. When the last address is reached
the index is set to undef, otherwords, not cyclic loop. 

Each call to next_address increments the counter, the next netmask,
broadcast and network values are respective to the address indexed.

=head3 next_usable()

    $usable = $ip->next_usable();

Method to get the flag to indicate if given IP is a usable host address 
in its network. 1 if usable, 0 if not usable.

=head3 $next_reset()
   
    $ip->next_reset();

Method to reset the index counter to the start for the next_xxx() methods.


=head3 get_ip_address_list()

    $ip->get_ip_address_list();

Method to return the list content of the objects 'ip_address_list' 

=head3 get_ip_address_hash()

   $href = $ip->get_ip_address_hash('3.3.3.4');

Method to get a ref to the hash for the given IP address, this is the hash
in the objects IPAddr list. The usable host list for the network_id stored in 
this hash (for the respective IP) is checked for existence and if exist the
list is added to the hash with the key 'hosts'. This hash is a separate
hash in the object keyed as 'current_ip', thus the other hashes
are not altered. Direct access to this in the object is 
$obj->{'current_ip}, the content is from the last call on 
get_ip_address_hash() 

This href looks as:

   $href = {
        '_first' => '3.3.3.1',
        '_hosts' => 14,
        '_last' => '3.3.3.14',
        '_usable' => 1,
        'address' => '3.3.3.4',
        'address_len' => 32,
        'broadcast' => '3.3.3.15/28',
        'broadcast_id' => '3.3.3.15',
        'broadcast_len' => '28',
        'gateway' => '3.3.3.7',
        'hosts' => [
                     '3.3.3.1',
                     '3.3.3.2',
                     '3.3.3.3',
                     '3.3.3.4',
                     '3.3.3.5',
                     '3.3.3.6',
                     '3.3.3.7',
                     '3.3.3.8',
                     '3.3.3.9',
                     '3.3.3.10',
                     '3.3.3.11',
                     '3.3.3.12',
                     '3.3.3.13',
                     '3.3.3.14'
                   ],
        'netmask' => '255.255.255.240',
        'netmask_len' => 28,
        'network' => '3.3.3.0/28',
        'network_id' => '3.3.3.0',
        'network_len' => '28',
        'prefix' => '3.3.3.1-15',
        'range' => '3.3.3.0 - 3.3.3.15',
        'version' => 4
   };

=head3 get_network_id_list()

    $ip->get_network_id_list();

Method to return the list content of the objects network_id_list.
The content of this list can be used to access the network hash. See 
network_id_list().

=head3 get_network_list

    $ip->get_network_list();

Method to return the list content of the objects network_list.
See network_id_list().


=head3 get_network()

    $href = $ip->get_network(<network_id>);

Method to take a given network id, as seen in the network_id_list, and return
a reference to the objects Network hash indexed by network id. 
See network_id_list().


=head3 init_object()

    $ip->init_object();

Method to init (clear) the data in the object. Once cleared
add_ip_address() can be used to put data back in to the object.

=head3 ip_version()

   $v = $ip->ip_version(<ip>);

Method to determin IP version. Returns 4 if IPv4, 6 if IPv6
otherwise 0.


=head1 Using Net::IP functions.

Net::IP has some functions that you may want to use as Network::IPAddress
does not provide an API for. You can call the Net::IP function directly
then. Note that the Net::IP doc has some functions listed that are
not in their .pm file. So if they seem to not work, grep the
Net::IP code or look in Net::IP's export list.

Example of using Net::IP ip_is_ipv6 function:

    $n = 'dead:beef::';
    $ok = &Net::IP::ip_is_ipv6($n);
    $Log->comment("=== Net::IP::ip_is_ipv6 $n  [$ok]");

To create your own Net::IP objectin a thpl:

    $ip = Net::IP->new('5.5.5.5');
    $Log->comment("=== Net::IP " . Dumper $ip);

=head1 Using NetAddr::IP

NetAddr::IP is a total OO package, thus it can be used to
create an object in your script and then called upon through
its methods.

Example using NetAddr::IP

   $ip = NetAddr::IP->new('3.3.3.4', '255.255.255.240');
   $Log->comment("=== NetAddr::IP " . Dumper $ip);
   $Log->comment("=== NetAddr::IP->hostenumref " . Dumper $ip->hostenumref());



=head1 Functions

No exported functions, call speciffically.

=head2 netip_compress_address

   $compresss_addr = Network::IPAddress::netip_ip_version(<addr>);

Given an IP address will return a v6 address in compress syntax or
given a v4 address will return the address.


=head2 netip_expand_address

   $expanded_addr = Network::IPAddress::netip_ip_version(<addr>);

Given an IP address will return a v6 address in expanded syntax or
given a v4 address will return the address.


=head1 Example

    use strict;
    use Tharn;
    use Data::Dumper;
    use Options;
    use Params::Common;
    
    use Network::IPAddress;
    
    my $Option = Options->new();
    
    my ( $params_href,
         $ok, $error, 
         $addr, @list, $n, $href
    );
    
    $Option->parameters( 'definitions' =>
       [
          { 'name'         => 'foo',
            'description'  => 'some foo var',
            'spec'         => 'i',
            'default'      => 1,
          },
       ]
    );
    
    
    $Option->process();
    $params_href = $Option->get_hash();
    
    # create object
    my ($ip, $error) = Network::IPAddress->new(
       'address'  => '3.3.3.1 - 3.3.3.5',
       #'address'  => '3.3.3.1 + 4',
       #'address'   => '3.3.3.1, 3.3.3.2, 3.3.3.3, 3.3.3.4, 3.3.3.5, 3.3.3.1',
       #'address'   => '3.3.3/28',
       #'address'   => '3.3.3.2',
    
       'gateway'   => '3.3.3.7',
       'netmask' => '255.255.255.240',
    
       'add_host_list' => 1,
    );
    
    if ($ip) { $Log->comment('=== IP Object made'); }
    else     { $Log->comment("=== IP Object Error: $error"); exit(0); }
    
    # add another ip to object
    ($ok, $error) = $ip->add_ip_address( 'address' => '4.4.4.1-4.4.4.5',
                         'gateway' => '4.4.4.1',
    );
    unless($ok) {
       $Log->comment("ERROR: add_ip_address: $error");
       exit(1);
    }
    # Get and Dumper some of the objects data structures
    if ( 1 ){
    # ip address list
       $Log->comment('=== Make IP Address list');
       ($ok, $error) = $ip->ip_address_list;
       $Log->comment("=== IP Address count: $ok");
    
       # network id list
       $Log->comment('=== Make Network ID list');
       ($ok, $error) = $ip->network_id_list;
       $Log->comment("=== Network ID count: $ok"); 
    
       @list = $ip->get_ip_address_list();
       $Log->comment("=== get_ip_address_list:" . Dumper \@list );
    
       @list = $ip->get_network_id_list();
       $Log->comment("=== get_network_id_list:" . Dumper \@list );
    
       @list = $ip->get_network_list();
       $Log->comment("=== get_network_list:" . Dumper \@list );
    }
    
    # cycle thru more addresses than in object
    $ok = $ip->{'addresses'} + 2;
    foreach (1..$ok) {
       $addr = $ip->next_address;
       $Log->comment("=== Next addr:: $addr");
    }
    # reset next counter
    $ip->next_reset;
    $addr = $ip->next_address;
    $Log->comment("=== Next addr:: $addr");
    
    # get the first network in the list
    $n = ($ip->get_network_id_list())[0];
    $Log->comment("=== Get network: $n");
    
    # get a ref to info about this network
    $href = $ip->get_network($n);
    $Log->comment("=== Network $n hash: " . Dumper $href);

    # version utility
    $n = 'dead:beef::';
    $ok = $ip->ip_version($n);
    $Log->comment("=== IP version $n  -> $ok");

    
    # get the hash for the given ip
    $href = $ip->get_ip_address_hash('3.3.3.4');
    $Log->comment("=== IP hash: " . Dumper $href);

    # access to libs in IPAddress.pm
    ##$ip = Net::IP->new('5.5.5.5');
    ##$Log->comment("=== Net::IP " . Dumper $ip); 

    $n = 'dead:beef::';
    $ok = &Net::IP::ip_is_ipv6($n);
    $Log->comment("=== Net::IP::ip_is_ipv6 $n  [$ok]");
    
    #$ip = NetAddr::IP->new('3.3.3.4', '255.255.255.240');
    #$Log->comment("=== NetAddr::IP " . Dumper $ip);
    #$Log->comment("=== NetAddr::IP->hostenumref " . Dumper $ip->hostenumref());

    exit(0);



=head1 STATUS

   Current

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=cut





