# -*- perl -*-
# $Id$
#
## @summary library to manage IP pools for the lab
##
## @description
##   This library contains routines that will be used to manage
##	 pools of IP addresses.   The initial use will be for data lifs.
##   It will be written generically so that other pools can be serviced
##   as well.
##
## @usage
##   use IP_Pool;
##
## @keywords  IP lifs test library
##
## @dependencies
## use config_base
## use Tharn
## 
## @author dl-cluster-qa@netapp.com
## @status testme
## 


package IP_Pool;

use strict;
use warnings;
use vars ('@ISA','@EXPORT');

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
			 new
             get_ip_pool
             return_ip_pool
	     );


use Options;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper;
use TracePkg;
use Tharn;

# Globals

my $Trace   = new TracePkg();
my $Options = new Options();

#############################################################################
# Subroutine: new
# Parameters: None
# Returns: Reference to this object
# Purpose: To instantiate and configure the ip pool object
#############################################################################
sub new {

    $Trace->entry();
    my $class = shift(@_);
    my $self = {};
    
    my %params = validate (
     	@_,
     	{
	         range => {
	             type     => SCALAR,
	             optional => 1,
	             default  => "1.2.3.1 - 1.2.3.200",
	         },
     	},  
	);
	
	my $ip_obj  = new Net::IP($params{range}) || die;
	my $ip_pool = [];
	
	do {
		push( @$ip_pool, $ip_obj->ip() );
	} while ( ++$ip_obj );
	
	$Log->comment(Dumper($ip_pool));
	$self->{ip_pool} = $ip_pool;
	bless($self, $class);
    $Trace->exit(msg => "Returning blessed reference: $self");
    return $self;
}


#############################################################################
# Subroutine: get_ip_pool
# Parameters: 
#		num_addrs 	- number of addresses requested
#		type		- type of IP address(only data supported currently)
# Returns: Reference to an array of IP address
# Purpose: To retrieve a set of IP addresses
#############################################################################
sub get_ip_pool {
	my $self = shift(@_);
	
	my $pool = [];

    $Trace->entry();
    my %params = validate (
        @_,
        {
            num_addrs => {
                type     => SCALAR,
                optional => 1,
                default  => "data",
            },
        },
        
    );
    
    my $num_addrs = $params{num_addrs};
	my $ip_pool   = $self->{ip_pool};
	
	$Log->comment("Requesting  IP addresses from pool");
	
	#Error if requested more than exist
	my $dp_size = @$ip_pool;
	if ($dp_size < $params{num_addrs}) {
		$Log->error( "Not enough IPs in pool for LIFS\n" );
		return 1;	
	}
	while ($num_addrs) {
		push(@$pool, shift(@$ip_pool));
		$num_addrs--;
	}
	
	$Trace->exit(); 
	return $pool;

}

#############################################################################
# Subroutine: put_ip_pool
# Parameters: 
#		pool		- pool of IPS to return
# Returns: Reference to an array of IP addresses
# Purpose: To release a set of IP addresses
#############################################################################
sub put_ip_pool {
   
   my $self = shift(@_);
   my $addr;
   
   $Trace->entry();
   my %params = validate (
        @_,
        {
            pool 	  => {
                type     => ARRAYREF,
                optional => 0,
            },
        },
        
    );	
    my $pool    = $params{pool};
    my $ip_pool = $self->{ip_pool};

	foreach $addr (@$pool) {
		push(@$ip_pool, $addr);
	}

	$Trace->exit(); 
	return 0;
}

1;