# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Ldap.pm#1 $

## @summary LDAP Setup routines
## @author dl-test-infra@netapp.com
## @status Shared
## @pod here


=head1 NAME

Network::Ldap - Filer setup routines for LDAP.

=head1 SYNOPSIS

    use Network::Ldap qw(ldap_setup);

    my $host = host($FILER);

    eval {
        ldap_setup(host=>$host, 
            vfiler=>$vfiler,
            ldap_base=>$ldap_base,
            ldap_servers=>$ldap_servers,
            ldap_pref_servers=>$ldap_pref_servers,
            ldap_name=>$ldap_name,
            ldap_passwd=>$ldap_passwd,
            ldap_port=>$ldap_port
        );
    };
    logresult('fatal', $@) if ($@);


=head1 DESCRIPTION

    This library contains LDAP setup routines.

=head1 ROUTINES

=cut

package Network::Ldap;

use strict;
use warnings;

use base qw(Exporter);
our @EXPORT_OK = qw(ldap_setup);

use Params::Validate v0.70 qw(validate SCALAR HASHREF);
use Carp qw(croak confess);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}

use Tharn qw(say logcomment script_result);
use Utils qw(make_host make_conn);
use CLI::FileIO qw(rdfile wrfile);
use VFILER::Setup qw(vfiler_status);


=head2 ldap_setup

This routine sets up LDAP on the filer.

On error the routine will die.

It takes the following named arguments

=over

=item ldap_servers

(Required) : Name of the ldap server/servers (comma separated) 
to be used for authentication

=item ldap_base 

(Required) : suffix of the ldap server, suffix that is defined 
in the conf file of ldap server.

=item ldap_port

(Optional): 

=item ldap_name 

(Optional): Distiguished name of the ldap server, ie rootdn that is
defined in conf file of ldap server.

=item ldap_passwd

(Optional): ldap password to be used for authentication.

=item ldap_pref_servers

(Optional): If you have multiple servers, specify the order of preference

=back

=cut

sub ldap_setup(%) {
    my (%args) = validate (@_, { 
                    host => { type=>SCALAR|HASHREF },
                    vfiler => { type=>SCALAR, optional=>1 },
                    ldap_base => { type=>SCALAR },
                    ldap_servers => { type=>SCALAR }, # comma separated list of servers
                    ldap_pref_servers => { type=>SCALAR, optional=>1 },
                    ldap_name =>{ type=>SCALAR, optional=>1 },
                    ldap_passwd => { type=>SCALAR, optional=>1 },
                    ldap_port => { type=>SCALAR, optional=>1 },
                    });
    
    my $cmd_prefix = ($args{vfiler}) ? "vfiler run $args{vfiler}" : "";

    # Create the connection for the filer
    my $host = make_host($args{host});
    my $conn = make_conn($args{host});

    # Create a hash based on what is specified by the user
    my %ldap_options = ('ldap.enable'=> 'on', 
                        'ldap.base' => $args{ldap_base},
                        'ldap.servers' => $args{ldap_servers});
    $ldap_options{'ldap.pref.servers'} = $args{ldap_pref_servers} if ($args{ldap_pref_servers});
    $ldap_options{'ldap.name'} = $args{ldap_name} if $args{ldap_name};
    $ldap_options{'ldap.passwd'} = $args{ldap_passwd} if $args{ldap_passwd};
    $ldap_options{'ldap.port'} = $args{ldap_port} if $args{ldap_port};

    # create an array of options to be setup.
    foreach my $ldap_option_key (keys %ldap_options) {
        my (undef,undef,$after) = say($conn, 
                            "$cmd_prefix options $ldap_option_key $ldap_options{$ldap_option_key}");
            croak($after) if ($after =~ /No such option/);
    }

    # edit /etc/nsswitch.conf file to reflect ldap first for passwd line
    # If vfiler_path option is specified, we assume that nsswitch.conf file 
    # of vfiler needs to be modified, so append  /etc/nsswitch.conf to
    # vfiler_path.
    my $nsswitch_file;
    if ($args{vfiler}) {
        my $vfiler_name = $args{vfiler};
        my %vfiler_attr  = vfiler_status($conn,$vfiler_name);
        logcomment("This is the path $vfiler_attr{$vfiler_name}{path0}");
        $nsswitch_file = "$vfiler_attr{$vfiler_name}{path0}" . "/etc/nsswitch.conf";
    } else {
        $nsswitch_file = "/etc/nsswitch.conf";
    }

    # do not cmd_prefix this, since 'rdfile' is not available 
    # in vfiler context.  
    my $output = rdfile($conn, "$nsswitch_file") or die($@);
    my @nsswitch_array = split("\n",$output);
    my @write_contents = "\n";
    foreach my $line(@nsswitch_array) {
        push @write_contents , "\n";
        if ($line =~ /passwd/) {
            push @write_contents , "passwd: ldap files nis";
        } else {
            push @write_contents , "$line";
        }
    }
    my $input = join("\n",@write_contents);
    # do not cmd_prefix this, since 'wrfile' is not available 
    # in vfiler context.  
    wrfile($conn,$nsswitch_file,$input) or die($@);
}

1;


=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = options ldap

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Network/Ldap.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

