# $Id: //depot/prod/test/Rsierranevada/lib/Network/Netgroup.pm#1 $
#
## @summary Routines for running Netgroup operations
##
## @description
##   These routines are used by netgroup tests for tasks like modifying
##   the local netgroup file, etc
##
## use Tharn;
##
## @author  dl-maryv (FSS Protocols QA)
##
##

package Network::Netgroup;

require Exporter;
use strict;
use Tharn;
use Data::Dumper;
use SayWrap qw(sayEOF);
use Carp qw(croak);
use SPARSE::Util::Logger qw(logDbg logTrc);
use filer qw(filer_options_set);
use Filer::Version qw(get_wafl_version WV_TSINGTAO_MIN
		      UNSUPPORTED_FEATURE_EXCEPTION);
use Network::Utils qw(run_getXXbyYY config_ipfw_rule run_ping);

use vars qw (@ISA @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT_OK = qw (
		 netgroup_db_is_supported
		 netgroup_add_entries_to_file
		 netgroup_create_large_file
		 verify_netgroup_entries
		 confirm_netgroup_db_created
		);

##
## @name ::netgroup_db_is_supported
##
## @summary
##    Check if netgroup database conversion is supported
##
## @description
##     Checks if netgroup db feature is supported in current ontap and
##     returns 1 or exception
## 
## @arg $fh  Required Filer host object
##
## @return
##     1 if netgroup database conversion is supported
##     UNSUPPORTED_FEATURE_EXCEPTION if netgroup db is not supported
##

sub netgroup_db_is_supported {

    my $fh = shift;

    my $version = get_wafl_version($fh);
    if ($version < WV_TSINGTAO_MIN) {
	croak(UNSUPPORTED_FEATURE_EXCEPTION . 
	      ": netgroup db unsupported in release prior to TsingTao.0");
    }
    return 1;
}


## @name ::netgroup_add_entries_to_file
##
## @summary
##    add netgroup entries to file
##
## @description
##     extract netgroup entries from provided hash and add them into the 
##   local netgroup file.
##
## @arg $mount_dir  Required mount point of root vol
##      $entries    Required reference to hash of netgroup entries
##      %options    Optional
##           overwrite => yes or no. If yes overwrite file,else append
##
## @return
## 
##

sub netgroup_add_entries_to_file {

    my ($mount_point,$entries,%options) =  @_;

    my $overwrite = $options{'overwrite'} || 'no';
    my $openflag = "";

    if (!$mount_point) {
	croak "mount point does not exist";
    }

    if ($overwrite =~ m/yes/i) {
	$openflag = ">";
    } else {
	$openflag = ">>"
    }

    # open netgroup file write/append/create nonexistant/no clobber
    open(FH,"$openflag"."$mount_point/etc/netgroup") 
	 or croak "Could not open netgroup file";
    # always start on a new line 
    print FH "\n";

    foreach my $netgroup (keys %$entries) {
	print FH "$netgroup $entries->{$netgroup}\n";
    }

    close FH;

    return 1;
}


## @name ::netgroup_create_large_file
## 
## @summary 
##    create a large netgroup file with dummy entries
##
## @description
##    create netgroup entries based on arguments provided and 
##	  dump them into a file. The number of entries, enrties per 
##	  line and nesting level determine how the entries will be 
##	  arranged. 
##
## @returns 
##		%netgroup_entries() hash of the netgroup entries 
##		that can be passed to verify_netgroup_entries()
##
## @args %options Required
##       out_file - file in which to create entries
##       srcip -  start of range of ip addresses used in 
##		 netgroup entries
##       count - number of netgroup entries (not number of netgroups)
##       entries_per_line - number of netgroup entries per netgroup
##       nesting_level - number of levels of nesting
##       nis_domains_string - comma separated list of nis domains to use
##                            in entries 
##       users_string - comma separated list of user names 
##       dns_domain - dns domainname to create hostnames of different 
##		 formats
##
##

sub netgroup_create_large_file (%) {

    my (%options) = @_;

    my $out_file = $options{'out_file'};
    my $srcip = $options{'src_ip'};
    my $count = $options{'count'};
    my $entries_per_line = $options{'entries_per_line'} || '70';
    my $nesting_level = $options{'nesting_level'} || '3';
    my $nis_domains_string = $options{'nis_domains'} || ",";
    my $users_string = $options{'users'} || 'user1,user2,user3';
    my $dns_domain = $options{'dns_domain'} || $main::DNS_DOMAIN;

    my $currentsrcip = $srcip;
    my $current_netgroup = 1;
	my %netgroup_entries = ();

    # create arrays of nis domains and users
    my @nis_domains = split /,/, $nis_domains_string;
    my @users = split /,/, $users_string;
    logTrc(Dumper(\@nis_domains));
    logTrc(Dumper(\@users));

    # process dns domain to create domain extensions of different lengths
    # eg lab.netapp.com, lab, lab.netapp
    my @dns_domain_split = split /./,$dns_domain;
    my $dns_ext_1 = $dns_domain_split[0];
    my $dns_ext_2 = $dns_domain_split[0].".".$dns_domain_split[1];
    my $dns_ext_3 = $dns_domain;


    open (FH,">$out_file")
	or croak "Could not open or create file for writing netgroup entries";

    # print comments about the file
    my ($s,$m,$h,$md,$mon,$yr) = localtime(time);
    print FH "# netgroup file generated by generate_$out_file on ";
    printf FH "%4d-%02d-%02d-%02d:%02d:%02d\n",$yr+1900,$mon+1,$md,$h,$m,$s; 
    print FH "# number of hosts = $count\n";
    print FH "# nesting level = $nesting_level\n";
    print FH "# Entries per line = $entries_per_line\n";

    my $line_count = 0;
    my $nesting_count = 0;
    my $line_length = 0;
    my $entry;
    my $pending = 0;
	my $actual_current;

    for (1..$count) {
	
	if ($line_count == 0) {
	    $line_length = 0;
		$actual_current = $current_netgroup;
	    print FH "ng-$current_netgroup ";

	    # last created entry was not printed. print it here
	    if ($pending) {
		print FH $entry;
		$netgroup_entries{"ng-$actual_current"} .= $entry;
	    }
	}

	# randomly pick one domainname from the array	
	my $domain = $nis_domains[rand(scalar(@nis_domains))];
	my $user = $users[rand(scalar(@users))];
	my $hostname;

	my $x = $_ % 4;

	if ($x == 1) {
	  $hostname = "nwk-fake-$currentsrcip";
	} elsif ($x == 2) {
	  $hostname = "nwk-fake-$currentsrcip.".$dns_ext_1;
	} elsif ($x == 3) {
	  $hostname = "nwk-fake-$currentsrcip.".$dns_ext_2;
	} else {
	  $hostname = "nwk-fake-$currentsrcip.".$dns_ext_3;
	}
	
	# create the netgroup entry 
        $entry = "(".$hostname.",".$user.",".$domain.")"." ";
	$line_length += length($entry);

	$currentsrcip = _get_src_ip($currentsrcip);
	++$line_count;

	# print no more than 4096 characters on each entry.a netgroup that
	# has more than 4096 characters will result in an illegal netgroup
	if ($line_length < 4000) {
	    # print no more than $per_line entries per netgroup line
	    if ($line_count > $entries_per_line) {
		$line_count = 0;
		$pending = 1;
		my $previous_netgroup = $current_netgroup - 1;
		
		# we want some levels of nested netgroups
		if ($nesting_count < $nesting_level) {
		    print FH "ng-$previous_netgroup\n";
			$netgroup_entries{"ng-$actual_current"} .= 
					"ng-$previous_netgroup";
		} else {
		    print FH "\n";
		    $nesting_count = 0;
		}
	    } else {
		print FH $entry;
		$netgroup_entries{"ng-$actual_current"} .= "$entry";
	    }
	    ++$nesting_count;
	    ++$current_netgroup;
	} else {
	    $line_count = 0;
	    $pending = 1;
	    my $previous_netgroup = $current_netgroup - 1;
		
	    # we want some levels of nested netgroups
	    if ($nesting_count < $nesting_level) {
		print FH "ng-$previous_netgroup\n";
		$netgroup_entries{"ng-$actual_current"} .= 
				"ng-$previous_netgroup";
	    } else {
		print FH "\n";
		$nesting_count = 0;
	    }
	    ++$nesting_count;
	    ++$current_netgroup;
	}
    }

    ++$current_netgroup;
    print FH "\n";

    my $i;
    $line_length = 0;
    print FH "ng-$current_netgroup ";
	$actual_current = $current_netgroup;
    for ($i=1;$i<$current_netgroup;$i++) {
	my $entry = "ng-$i ";
	$line_length += length($entry);
	if ($line_length < 4000) {
	    print FH $entry;
		$netgroup_entries{"ng-$actual_current"} .= "$entry";
	} else {
	    last;
	}
    }

    close FH;

    #return 1;
    return \%netgroup_entries;

}

# increment the ip address by 1

sub _get_src_ip($) {

    my $srcip = shift;

    # incr the src ip by 1.

    my @octets = split /\./, $srcip;

    $octets[3]++;
    if ( ($octets[3]) > 254 ) {
	$octets[3] = 1;
	$octets[2]++;
	if (($octets[2]) > 254) {
	    $octets[2] = 1;
	    $octets[1]++;
	    if (($octets[1]) > 254) {
		$octets[1] = 1;
		$octets[0]++;
		if (($octets[0]) > 254) {
		    $octets[0] = 1;
		}
	    }
	}
    }

    return($octets[0].".".$octets[1].".".$octets[2].".".$octets[3]);
}


## @name ::verify_netgroup_entries
##
## @summary
##  expand a netgroup hash and verify that each entry in a netgroup 
##  is correctly looked up
##
## @description
##  given a hash of netgroup entries, this function expands the netgroup
##  (if it contains other netgroups) and then runs getXXbyYY for each
##  netgroup entry to verify that the local netgroup entry can be looked
##  up correctly
##
## @args  $conn              Required Filer connection object
##        $netgroup_entries  Required Reference to Hash of netgroup entries
##
## @returns $result_hash reference to hash of arrays of host-lookup 
##          result pairs
##          croaks upon failures

sub verify_netgroup_entries($$;%) {

    my ($conn,$netgroup_entries,%options) = @_;
    my $nis_domain = $options{'nis_domain'} || $main::NIS_DOMAIN;
    my $vfiler_name = $options{'vfiler_name'} || undef;

    my %result_hash = ();
    my %expanded_netgroup_entries =();
    my $netgroup;
    my @entries;
    my $expanded_entry;
    my $result = "";
    my $output;
    my $bad_entry;

    foreach $netgroup (keys %$netgroup_entries) {
	my $nesting = 0;
	if (!exists $expanded_netgroup_entries{$netgroup}) {
	    $expanded_entry = _expand_netgroup($netgroup,
					       $netgroup_entries,
					       \%expanded_netgroup_entries,
					       \$nesting,
					       $nis_domain);
	    $expanded_netgroup_entries{$netgroup} = $expanded_entry;
	}
    }
	
    logTrc(Dumper(\%expanded_netgroup_entries));

    # run getXXbyYY
    foreach $netgroup (keys %expanded_netgroup_entries) {
	@entries =();
	@entries = split /\s+/,$expanded_netgroup_entries{$netgroup};
	foreach my $entry (@entries) {
	    if ($entry =~ /\((\S+),.*,/) {
		my $host = $1;
		if ($vfiler_name) {
			$output = run_getXXbyYY($conn,"netgrp","$netgroup $host",
					vfiler_name => $vfiler_name);
		} else {
			$output = run_getXXbyYY($conn,"netgrp","$netgroup $host");
		}
		if ($output =~ /is not in/) {
		    croak ("netgroup match failed for".
			   " $host in netgroup $netgroup ");
		} else {
		    push @{$result_hash{$host}},"$output\n"
		}
	    } else {
		croak ("Bad entry $entry in netgroup hash");
	    }
	}

	# check for negative entries too
	$bad_entry = time();
	if ($vfiler_name) {
		$output = run_getXXbyYY($conn,"netgrp","$netgroup $bad_entry", 
			vfiler_name => $vfiler_name);
	} else {
		$output = run_getXXbyYY($conn,"netgrp","$netgroup ".$bad_entry);
	}
	if ($output =~ /is in/) {
	    croak ("netgroup match incorrectly succeeded ".
		   " for netgroup $netgroup");
	} else {
	    push @{$result_hash{$bad_entry}}, "$output\n"
	}
    }

    logTrc(Dumper(\%result_hash));
    return \%result_hash;
}


# recursive function that expands netgroup entries that contain other
# netgroups. The function takes a netgroup name, reference to the netgroup
# hash, reference to a placeholder hash for expanded entries and 
# a scalar counter for maintaining the level of nesting that is being
# processed. the function looks at each entry in a netgroup and matches it
# to netgroup names in the hash. If a match is found, it means this netgroup
# needs expansion. If the entry that matched already has an expanded entry
# in the expanded entries hash, it uses the expanded entry directly. If not,
# the function is called with the netgroup that needs expansion. The counter
# is used to keep track of how many times this function has been called
# recursivelty. At the 1000th iteration the function will croak because
# we could either be processing a cyclic netgroup

sub _expand_netgroup() {
    my ($netgroup,$netgroup_entries,$expanded_netgroup_entries,$nesting,
	$nis_domain) = @_;

    if ($$nesting > 999) {
	# exit if we are entering this recursive function the 1000th time
	# even the filer allows only 1000 nesting levels. We could be in 
	# an infinite loop due to cyclic netgroup entries!
	croak ("Too much nesting (>=1000) or cyclic netgroups encountered");
    }

    my @entries=();
    my ($expand,$expanded_entry) = ("","");

    ++$$nesting;
    @entries = split  /\s+/,$netgroup_entries->{$netgroup};
    foreach my $entry (@entries) {
	if (exists $netgroup_entries->{$entry}) {
	    if (exists $expanded_netgroup_entries->{$entry}) {
		$expanded_entry .= $expanded_netgroup_entries->{$entry}." ";
	    } else {
		# recursive call
		$expanded_entry .=
		    _expand_netgroup($entry,
				     $netgroup_entries,
				     $expanded_netgroup_entries,
				     $nesting,
				     $nis_domain)." ";
	    }
	} else {
	    # strip off ( and ) from entry
	    if ($entry =~ /\((.*)\)/) {
		my $stripped_entry = $1;
		my @entry_array = split /,/,$stripped_entry;

		# use this entry only is the nis domain matches or the 
		# nis domain is wildcard (blank)
		my $entry_nis = $entry_array[2];
		# get rid of blank spaces
		$entry_nis =~ s/\s+//g;

		if (($entry_nis eq $nis_domain) or
		    ($entry_nis eq "")) {
		    $expanded_entry .= $entry." ";
		} else {
		    next;
		}
	    }

	}
    }

    return $expanded_entry;
}


## @name ::confirm_netgroup_db_created
##
## @summary
##    confirm if netgroup db files have been created
##
## @description
##     This subroutine checks if the two netgroup db files are created
##  in the /etc/yp/file directory under the root volume
##
## @arg $mount_dir  Required mount point of root vol
##
## @return
## 
##

sub confirm_netgroup_db_created($) {

    my $mount_dir = shift;

    if (!$mount_dir) {
        croak "mount point does not exist";
    }
    my ($dir_handle,@allfiles);

    # check directory contents to confirm netgroup db is created
    opendir($dir_handle,"$mount_dir/etc/yp/file")
        or die fatal_result("Could not open netgroup db directory $!");
    @allfiles = readdir($dir_handle);
    closedir($dir_handle);

    # pop off . and ..
    shift(@allfiles);
    shift(@allfiles);

    if (scalar(@allfiles) == 0) {
        die fail_result("netgroup db not created in".
            "$mount_dir/etc/yp/file");
    } else {
        # pop the remaining two files and check names
        foreach my $file (@allfiles) {
            if ($file !~ m/netgroup.*.db/i) {
                die fail_result("unexpected netgroup db file name");
            }
        }
        logcomment("Netgroup db files created correctly");
    }
    
    return 1;
}

