# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Nexus/Nxinterface.pm#3 $
#
## @summary Nxinterface -- Provide a class for the interfaces in the Nexus switch
## @author  Nayan Patel(nayankumar.patel@netapp.com)
## @status  public
## @notes   Created by NICQA Group
## @pod here

=head1 NAME

    Nxinterface - Provide a class for the Nexus 5010 switch interface sub-mode

=head1 SYNOPSIS

    use Network::Nexus::Nxinterface;

# Switch interface constructor

    $if_obj = Network::Nexus::Nxinterface->new(conn => $sw_obj{conn}, 
                    if_name => $interface_name[,if_type => $interterface_type]);

 

=head1 DESCRIPTION

    Provides a class for the cisco Nexus switch interface sub-mode.

=head1 STATUS

    Shared

=head1 AUTHOR

    Nayan patel (NIC QA)
    
=head1 COPYRIGHT

    Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Nexus::Nxinterface;

use strict;
use warnings;

use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT);
use SPARSE::Util::Logger qw(logStep logWarn);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};
#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $VERSION = '1.0';
our @ISA;

#---------------------------------------------------------------------

=head2 new

    $if_obj = Network::Nexus::Nxinterface->new(conn => $sw_obj{conn}, 
                    if_name => $interface_name[,if_type => $interterface_type]);


Creates an independent hardware Nexus switch interface object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item switch_conn_obj

    (Required) : switch connection object from the switch object (EX: $opts{conn}).

=item interface_name

    (required) : Name of the interface (EX: 1/1).

=item type

    (optional) : interface type. default will be "Ethernet"

=back

=item Returns

    Switch interface object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub new {

    my $class = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                conn  => { type => HASHREF },
                                if_name => { type => SCALAR },
                                if_type => { type => SCALAR, default => 'ethernet' },
                            }

                          );

     # get the hosttype from the nate_host file of this switch
     my $type = $opts{conn}->hostp->hosttype;
     if ( $type =~ /switch-cisco/ ) {
          require Network::Nexus::Nxswitch;
          @ISA = qw(Network::Nexus::Nxswitch);
     } else {
          croak "Switch type is not supported: $type";
     }

    return bless \%opts, $class;
}

#--------------------------------------------------------------------
=head2 show_interface

    $out = $if_obj->show_interface(info_type => 'brief');

    show the configuration of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item info_type

    (optional) : type of information (EX: brief, capability, etc).

=back

=item Returns

   return the interface configurations. 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_interface {

    my $self = shift;
    my (%opts) = validate (
                            @_,
                            {
                                info_type => { type => SCALAR, optional => 1 }
                            }
                          );

    my $cmd;
    if(!defined($opts{info_type})) {
        $cmd = "show interface $self->{if_type} $self->{if_name} | no-more";
    } else {
        $cmd = "show interface $self->{if_type} $self->{if_name} $opts{info_type}";
    }
    return $self->show_switch_config(cmd => $cmd);
}

#--------------------------------------------------------------------
=head2 show_interface_brief

    %out = $if_obj->show_interface_brief();

    show the configuration of the interface in brief.

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

   return the brief interface configurations in HASH.
    EX:
    %out = (
            Ethernet  => 'Eth1/20',
            Status => 'up',
            Port => '--',
            Reason => 'none',
            Type => 'eth',
            VLAN => '1',
            Speed => '10G(D)',
            Mode => 'access'
           );

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_interface_brief {

    my $self = shift;
    
    my $config =  $self->show_interface(info_type => "brief");
    my @temp = split(/\-{5,}\n/,$config);
    shift(@temp);
    my ($temp3, $temp4) = split(/\n/,$temp[0]); 
    my @key = split(/\s+/,$temp3);
    my @value = split(/\s+/,$temp[1], 6);
    push(@value,split(/\s{2,}/,pop(@value))); 
    push(@value,split(/\s/,pop(@value))); 
    
    my $i = 0;
    my %out;
    for($i=0; $i< scalar(@key);$i++){
        $out{$key[$i]} = $value[$i];
    }
    
    return %out;
}

#--------------------------------------------------------------------
=head2 show_interface_capabilities

    %out = $if_obj->show_interface_capabilities();

    gives the interface capabilities.

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

   return the interface capabilities in HASH.
    EX:
    %out = (
             'UDLD' => 'yes',
             'Trunk encap. type' => '802.1Q',
             'ToS rewrite' => 'no',
             'QOS scheduling' => 'rx-(6q1t),tx-(1p6q0t)',
             'Model' => 'MOC12400903',
             'SPAN' => 'yes',
             'Type' => '10Gbase-(unknown)',
             'Speed' => '10000',
             'MDIX' => 'no',
             'Flowcontrol' => 'rx-(off/on),tx-(off/on)',
             'Duplex' => 'full',
             'Broadcast suppression' => 'percentage(0-100)',
             'FEX Fabric' => 'yes',
             'Rate mode' => 'none',
             'CoS rewrite' => 'no',
             'Channel' => 'yes',
             'Link Debounce Time' => 'yes',
             'Link Debounce' => 'yes'
           );

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub show_interface_capabilities {
    
    my $self = shift;

    my $config = $self->show_interface(info_type => 'capabilities');
    
    my @temp = split(/\n/,$config);
    shift(@temp);
    my ($i,@key,@value);
    for($i=0;$i<scalar(@temp);$i++) {
        ($key[$i],$value[$i]) = split(/\:/,$temp[$i]);
        #remove leading and trailing spaces
        $key[$i] =~ s/^\s+//;
        $value[$i] =~ s/^\s+//;
        $value[$i] =~ s/\s+$//;
    }
    my %out;
    for($i=0; $i< scalar(@key);$i++){
        $out{$key[$i]} = $value[$i];
    }
    
    return %out;
}

#--------------------------------------------------------------------
=head2 show_interface_flowcontrol

    %out = $if_obj->show_interface_flowcontrol();

    gives flow control setting of the interface .

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

   return the flowcontrol setting of interface in HASH.
    EX:
    %out = (
            'TxPause' => '0',
            'RxAdmin' => 'off',
            'port' => 'Eth1/20',
            'TxAdmin' => 'off',
            'TxOper' => 'off',
            'RxOper' => 'off',
            'RxPause' => '0'
          );

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub show_interface_flowcontrol {

    my $self = shift;

    my $config = $self->show_interface(info_type => "flowcontrol");
    my @temp = split(/\-+/,$config);
    my $fc = pop(@temp);
    $fc =~ s/\n+//;

    my @value = split(/\s+/,$fc);
    my %out;
    $out{port} = $value[0];
    $out{TxAdmin} = $value[1];
    $out{TxOper} = $value[2];
    $out{RxAdmin} = $value[3];
    $out{RxOper} = $value[4];
    $out{RxPause} = $value[5];
    $out{TxPause} = $value[6];

    return %out;
}

#--------------------------------------------------------------------
=head2 show_interface_status

    %out = $if_obj->show_interface_status();

    gives status of the interface .

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

   return the status of interface in HASH.
    EX:
    %out = (
              'Duplex' => 'full',
              'Type' => '10g',
              'Speed' => '10G',
              'Status' => 'up',
              'Port' => 'Eth1/20',
              'Name' => '--',
              'Vlan' => '1'
          );

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub show_interface_status {

    my $self = shift;
    
    my $config = $self->show_interface(info_type => "status");
    my @temp = split(/\-{5,}\n/,$config);
    shift(@temp);
    chomp($temp[0]);
    chomp($temp[1]);
    my @key = $temp[0] =~ /(\w+\s+|\w+)/g;
    my @headings;
    map {push(@headings,$_);} @key;
    foreach (@headings) {
        $_ =~ s/^\s+//;
        $_ =~ s/\s+$//;
    }

    my %hash;
    my $l = 0;
    my $line = $temp[1];
    for (my $i = 0; $i < scalar(@key); $i++) {
        if ($i != scalar(@key) - 1) {
            $hash{$headings[$i]} = substr($line, $l, length($key[$i]));
        } else {
            $hash{$headings[$i]} = substr($line, $l);
        }
        $l = $l + length($key[$i]);
        $hash{$headings[$i]} =~ s/^\s+//;
        $hash{$headings[$i]} =~ s/\s+$//;
    }

    return %hash;
}

#--------------------------------------------------------------------
=head2 show_interface_switchport

    %out = $if_obj->show_interface_switchport();

    gives switchport configuration of the interface .

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

   return the switchport configuration of interface in HASH.
    EX:
    %out = (
             'Operational Mode' => 'access',
             'Administrative private-vlan trunk private VLANs' => 'none',
             'Administrative private-vlan primary mapping' => 'none',
             'Administrative private-vlan secondary mapping' => 'none',
             'Administrative private-vlan trunk encapsulation' => 'dot1q',
             'Trunking Native Mode VLAN' => '1 (default)',
             'Unknown unicast blocked' => 'disabled',
             'Administrative private-vlan trunk normal VLANs' => 'none',
             'Trunking VLANs Enabled' => '1-3967,4048-4093',
             'Switchport Monitor' => 'Not enabled',
             'Unknown multicast blocked' => 'disabled',
             'Switchport' => 'Enabled',
             'Access Mode VLAN' => '1 (default)',
             'Administrative private-vlan trunk native VLAN' => 'none',
             'Operational private-vlan' => 'none',
             'Administrative private-vlan primary host-association' => 'none',
             'Name' => 'Ethernet1/20',
             'Administrative private-vlan secondary host-association' => 'none'
            );

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub show_interface_switchport {

    my $self = shift;

    my $config = $self->show_interface(info_type => 'switchport');
    my @temp = split(/\n+/,$config);

    my ($i,@key,@value);
    for($i=0;$i<scalar(@temp);$i++) {
        ($key[$i],$value[$i]) = split(/\:/,$temp[$i]);
        #remove leading and trailing spaces
        $key[$i] =~ s/^\s+//;
        $value[$i] =~ s/^\s+//;
        $value[$i] =~ s/\s+$//;
    }
    
    my %out;
    for($i=0; $i< scalar(@key);$i++) {
        $out{$key[$i]} = $value[$i];
    }
    return %out;
}


#--------------------------------------------------------------------
=head2 show_running_config

    @out = $if_obj->show_running_config();

    gives running configuration of the switch port.
    it gives output of "show running-config interface ethernet <interface name>"

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

   return the running configuration of the switch port in array.
 =back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_running_config {

    my $self = shift;
    
    my $cmd = "show running-config interface $self->{if_type} $self->{if_name}";
    my $temp = $self->show_switch_config(cmd => $cmd);
    my @temp1 = split(/\n\n/,$temp); 
    my @config = split(/\n/,$temp1[-1]);
    shift(@config);
    foreach my $conf (@config){
        $conf =~ s/^\s+//;
    }

    return @config;
}

#--------------------------------------------------------------------
=head2 show_port_channel_db

    %out = $if_obj->show_port_channel_db();

    gives the information about the port-channel.
    it gives output of "show port channel database nterface port-channel <port-channel number>"

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns
   return the information of the port-channel in the HASH.
   EX: following is a example of a return HASH
          'First operational port' => 'Ethernet1/13',
          'ports_up' => '3',
          'Last bundled member' => '',
          'total_ports' => '4',
          'ports' => [
                       {
                         'status' => 'up',
                         'port' => 'Ethernet1/13'
                       },
                       {
                         'status' => 'up',
                         'port' => 'Ethernet1/14'
                       },
                       {
                         'status' => 'up',
                         'port' => 'Ethernet1/15'
                       },
                       {
                         'status' => 'individual',
                         'port' => 'Ethernet1/16'
                       }
                     ],
          'Operational channel mode' => 'active',
          'Time since last bundle' => '0d:01h:53m:51s',
          'Administrative channel mode' => 'active',
          'Age of the port-channel' => '0d:01h:52m:34s',
          'Last membership update' => 'successful'
 =back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_port_channel_db {
    
    my $self = shift;
    my $cmd = "show port-channel database interface $self->{if_type} $self->{if_name}";
    my $temp = $self->show_switch_config(cmd => $cmd);
    my @temp1 = split(/Ports\:/,$temp); 
    
    my @temp2 = split(/\n/,$temp1[0]);
    #remove first element, which is port-channel<num>
    shift(@temp2);
    #remove last element, which is just the empty space
    pop(@temp2);
    my ($i,@key,@value,$total_ports,$ports_up);
    for($i=0;$i<scalar(@temp2);$i++) {
        $temp2[$i] =~ s/^\s+//;
        if($temp2[$i] =~ /(\d+)\sports\sin\stotal\,\s(\d+)\sports\sup/) {
            $key[$i] = "total_ports";
            $value[$i] = $1;
            $ports_up = $2;
        } else {
            ($key[$i],$value[$i]) = split(/\sis\s/,$temp2[$i]);
        }    
    }
    $key[$i] = "ports_up"; 
    $value[$i] = $ports_up;

    my ($j,@ports);
    my @temp3 = split(/\n/,$temp1[1]);
    for($j=0;$j<scalar(@temp3);$j++) {
        $temp3[$j] =~ s/^\s+//;
        if($temp3[$j] =~ /(\w+(\d+\/)?\d+\/\d+)\s+\[(\w+)\s?\]\s?\[?(\w+)?\]?(\s\*)?/) {
            $ports[$j]{port} = $1;
            if (defined($3)) {
                $ports[$j]{status}= $4;
                $ports[$j]{mode}= $3;
                
            } else {
                $ports[$j]{status}= $3;
            }    
        }
    }

    $key[$i+1] = "ports";
    $value[$i+1] = \@ports;
    
    my %out;
    for($i=0; $i< scalar(@key);$i++) {
        $out{$key[$i]} = $value[$i];
    }
 
    return %out;
}

#--------------------------------------------------------------------
=head2 config_interface

    $out = $if_obj->config_interface(@commands);

    configure the interface in configure terminal mode with specified command list.

The following named arguments are accepted:

=over

=item Arguments

=over

=item commands
    (Required): Array of commands, you want to use to configure interface.

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub config_interface {

    my ($self,@cmds) = @_;

    if(!defined($cmds[0])){
        croak "Please provide commands to configure interface";
    }
    
    my ($cmd,$out);
    #go to interface sub-mode
    $out = $self->nexus_say(
                            cmd => "interface $self->{if_type} $self->{if_name}",
                            mode => 1
                           );
    if($out =~ /(^ERROR(.*) | invalid)/i) {
        die "Not able to go into the interface sub-mode."
    }
    foreach $cmd (@cmds){ 
        $out = $self->nexus_say(cmd => $cmd, mode => 0);
        if($out =~ /(^ERROR(.*) | invalid | ^command\sfailed)/) {
            croak "command failed: $out ";
        }
    }

    return;
}

#--------------------------------------------------------------------
=head2 config_flowcontrol

    $out = $if_obj->config_flowcontrol(send => 'on', receive => 'on');

    configure flowcontrol settings on the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item send
    (optional): send flowcontrol setting
=item receive
    (optional):receive flowcontrol setting
    
    EX: (send => 'on', receive => 'on')
=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_flowcontrol {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                send  => { type => SCALAR, optional => 1 },
                                receive => { type => SCALAR, optional => 1 },
                            }

                          );


    my @cmds;
    if(defined($opts{send})){
        $cmds[0] = "flowcontrol send $opts{send}";
    }

    if(defined($opts{receive})){
        $cmds[1] = "flowcontrol receive $opts{receive}";
    }

    if(scalar(@cmds) == 0) {
        croak "You have not specified correct send or receive flowcontrol settings."
    }
    eval { $self->config_interface(@cmds); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_speed
    $out = $if_obj->config_speed(speed => '1000');

    configure speed settings on the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item speed
    (Required): speed of the interface.
    EX: 1000 or 10000


=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_speed {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                speed => { type => SCALAR },
                            }

                          );


    if(!defined($opts{speed})) {
        croak "Please specify the speed you want to set.";
    }
    eval { $self->config_interface("speed $opts{speed}"); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_status
    $out = $if_obj->config_status(status => 'up');

    configure status of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item status
    (Required): status of the interface
    EX: up or down


=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_status {
    
    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                status => { type => SCALAR },
                            }

                          );


    if(!defined($opts{status})) {
        croak "Please provide the status(up/down), you want to set.";
    } elsif($opts{status} eq 'down') {
        eval { $self->config_interface("shutdown"); };
        croak $@ if $@;
    } elsif($opts{status} eq 'up') {
        eval { $self->config_interface("no shutdown"); };
        croak $@ if $@;
    } else {
        croak "The status, you want to set, is not up or down, it is $opts{status}.";
    }

    return;
}

#--------------------------------------------------------------------
=head2 config_bandwidth
    $out = $if_obj->config_bandwidth(bandwidth => 10000, inherit => 0);

    configure bandwidth of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item bandwidth
    (Required): bandwidth of the interface
    EX: 1 to 10000000

=item inherit
    (optional): whether you want to inherit bandwidth or not.
    EX: inherit => 0 -- for not inherit
        inherit => 1 -- for inherit
=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_bandwidth {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                bandwidth => { type => SCALAR },
                                inherit => {type => SCALAR, optional => 1 }
                            }

                          );


    if(!defined($opts{bandwidth})) {
        croak "Please specify the bandwidth you want to set.";
    } elsif(!defined($opts{inherit}) or $opts{inherit} == 0) {
        eval { $self->config_interface("bandwidth $opts{bandwidth}"); };
        croak $@ if $@;
    } elsif($opts{inherit} == 1) {
        eval { $self->config_interface("bandwidth inherit $opts{bandwidth}"); };
        croak $@ if $@;
    } else {
        eval { $self->config_interface("bandwidth $opts{bandwidth}"); };
        croak $@ if $@;
    }

    return;
}

#--------------------------------------------------------------------
=head2 config_channel_group
    $out = $if_obj->config_channel_group(port_ch => 1, mode => 'passive');

    configure channel group of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port_ch
    (Required): port channel number
    EX: 1 to 4096

=item mode
    (optional): mode of the port channel.
    EX: active, on or passive

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_channel_group {
    
    my $self= shift;
    my %opts = validate(
                            @_,
                            {
                                port_ch => { type => SCALAR },
                                mode => { type => SCALAR, optional => 1 } 
                            }
                      );
 
    if($opts{port_ch} < 1 or $opts{port_ch} > 4096) {
        croak "Please Specify the port channel number between 1 to 4096. ".
              "you have specified $opts{port_ch}";
    }

    if(!defined($opts{mode})) {
        eval { $self->config_interface("channel-group $opts{port_ch}"); };
        croak $@ if $@;
    } elsif ($opts{mode} =~ /(^active$|^on$|^passive$)/) {
        eval { $self->config_interface("channel-group $opts{port_ch} mode $opts{mode}"); };
        croak $@ if $@;
    } else {
        croak "The mode is worng. it should be active, on or passive. you specified $opts{mode}.";
    }
    return;
}

#--------------------------------------------------------------------
=head2 remove_channel_group
    $out = $if_obj->remove_channel_group(channel_group => 100');

    remove channel group from the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item channel_group
    (Required): port channel number
    EX: 1 to 4096

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub remove_channel_group {

    my $self= shift;
    my %opts = validate(
                            @_,
                            {
                                channel_group => { type => SCALAR }
                            }
                      );
    my $cmd = "no channel-group $opts{channel_group}";                   
    eval {
        $self->config_interface($cmd);
    };
    croak $@ if $@;
    
    return;
}

#--------------------------------------------------------------------
=head2 config_switchport
    $out = $if_obj->config_switchport(cmd => 'access valn 1');

    configure switchport setting of the interface. It will prefix the 
    "switchport" to the specified command and execute the command.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cmd
    (Required): specify the command under switchport
    EX: "access vlan 1" or "host" or "monitor" or "mode trunk" etc

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_switchport {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                cmd => { type => SCALAR },
                            }
                          );
  
    if(!defined($opts{cmd})) {
        croak "Please provide config, you want to set.";
    }
    eval { $self->config_interface("switchport $opts{cmd}"); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_switchport_access_vlan
    $out = $if_obj->config_switchport_access_vlan(vlan => '1');

    configure switchport access vlan setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vlan 
    (Required): vlan id
    EX: 1 to 3967 or 4048 to 4093

=back

=item Returns 

    Retrun the out put of the command "switchport access vlan <vlan id>"

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_switchport_access_vlan {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                vlan => { type => SCALAR },
                            }
                          );

    if(!defined($opts{vlan})) {
        croak "Please provide the vlan id, you want to set.";
    }
    if($opts{vlan} < 1 or ($opts{vlan} > 3967 and $opts{vlan} < 4048) or $opts{vlan} > 4093) {
        croak "Please provide vlan ID in the <1-3967, 4048-4093> range. ".
              "you have provided $opts{vlan}";
    }
    
    eval { $self->config_switchport(cmd => "access vlan $opts{vlan}"); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_switchport_mode
    $out = $if_obj->config_switchport_mode(mode => "access");

    configure switchport mode setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mode 
    (Required): provide the mode, you want to set.
    EX: access or trunk

=back

=item Returns 

    Retrun the out put of the command "switchport mode <mode>"

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_switchport_mode {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                mode => { type => SCALAR },
                            }
                          );
    
    if(!defined($opts{mode})) {
        croak "please provide the mode, you want to set.";
    }

    if($opts{mode} !~ /(^access$|^trunk$)/) {
        croak "please provide correct mode(access or trunk). you have provided $opts{mode}";
    }

    eval { $self->config_switchport(cmd => "mode $opts{mode}"); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_switchport_trunk_native_vlan
    $out = $if_obj->config_switchport_trunk_native_vlan(vlan => 1);

    configure switchport trunk native vlan setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vlan
    (Required): vlan id.
    EX: 1 to 3967,4048 to 4093

=back

=item Returns 

    Retrun the out put of the command "switchport trunk native vlan <vlan id>"

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_switchport_trunk_native_vlan {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                vlan => { type => SCALAR },
                            }
                          );

   
    if(!defined($opts{vlan})) {
        croak "Please provide vlan id.";
    }

    if($opts{vlan} < 1 or ($opts{vlan} > 3967 and $opts{vlan} < 4048) or $opts{vlan} > 4093) {
        croak "Please provide vlan ID in the <1-3967, 4048-4093> range. ".
              "you have provided $opts{vlan}";
    }
    
    eval { $self->config_switchport(cmd => "trunk native vlan $opts{vlan}"); };
    croak $@ if $@;

    return;
}


#--------------------------------------------------------------------
=head2 config_switchport_trunk_allowed_vlan
    $out = $if_obj->config_switchport_trunk_allowed_vlan(vlan => 'all');

    configure switchport trunk allowed vlan setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vlan
    (Required): vlan 
    EX: 1 to 3967,4048 to 4093, all or none

=back

=item Returns 

    Retrun the out put of the command "switchport trunk allowed vlan <vlan>"

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_switchport_trunk_allowed_vlan {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                vlan => { type => SCALAR },
                            }
                          );

    if(!defined($opts{vlan})) {
        croak "Please provide vlan vlaue(<1-3967, 4048-4093>, 'all', or 'none').";
    }

    if(($opts{vlan} =~ /(^all$|^none$|(([1-9]{1,4})(\-[1-9]{1,4})?(\,)?)+)/)) {
        eval { $self->config_switchport(cmd => "trunk allowed vlan $opts{vlan}"); };
        croak $@ if $@;
    } else {
        croak "Please provide correct value(<1-3967, 4048-4093>, 'all', or 'none') for vlan.";   
    }

    return;
}

#--------------------------------------------------------------------
=head2 config_cdp
    $out = $if_obj->config_cdp(status => 'enable');

    enable/disable cdp of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item status
    (Required): status of the cdp 
    EX: enable or disable

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_cdp {
    
    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                status => { type => SCALAR },
                            }
                          );
   
    my $cmd;

    if(!defined($opts{status}) or $opts{status} eq 'enable') {
        $cmd = "cdp enable";    
    } elsif($opts{status} eq 'disable') {
        $cmd = "no cdp enable";
    } else {
        croak "Please provide correct value of cdp status(enable or disable). ".
              "you have provided $opts{status}.";
    }

    eval { $self->config_interface($cmd); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_spanning_tree
    $out = $if_obj->config_spanning_tree(cmd => "bpduguard enable");

    configure the spanning tree setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cmd
    (Required): command  
    EX: bpduguard enable or
        port type network         

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_spanning_tree {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                cmd => { type => SCALAR },
                            }
                          );

   
    if(!defined($opts{cmd})) {
        croak "Please command you want to execute under spanning tree directory."
    }
    eval { $self->config_interface("spanning-tree $opts{cmd}"); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_spanning_tree_bpduguard
    $out = $if_obj->config_spanning_tree_bpduguard(status => "enable");

    configure the spanning tree bpduguard setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item status
    (Required): status of the bpduguard  
    EX: enable or disable         

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_spanning_tree_bpduguard {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                status => { type => SCALAR },
                            }
                          );
   
    my $cmd;
    if(!defined($opts{status})) {
        croak "Please provide the status(enable or disable) for the bpduguard.";
    }

    if($opts{status} =~ /(^enable$|^disable$)/) {
        $cmd = "bpduguard $opts{status}";    
    } else {
        croak "Please provide correct value of bpduguard status(enable or disable). ".
              "you have provide $opts{status}";
    }

    eval { $self->config_spanning_tree(cmd => $cmd); };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 config_spanning_tree_port_type
    $out = $if_obj->config_spanning_tree_port_type(type => "edge");

    configure the spanning tree port type setting of the interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type
    (Required): port type of the interface.  
    EX: edge or network or normal

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_spanning_tree_port_type {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                type => { type => SCALAR },
                            }
                          );

  
    if(!defined($opts{type})) {
        croak "Please provide the port type('edge' or 'network' or 'normal').";
    }

    if($opts{type} =~ /(^edge$|^network$|^normal$|^edge\strunk$)/) {
        eval { $self->config_spanning_tree(cmd => "port type $opts{type}"); };
        croak $@ if $@;
    } else {
        croak "Please provide the correct port type('edge','network','normal'). ".
              "You have provide $opts{type}.";
    }

    return;
}

#--------------------------------------------------------------------
=head2 verify_port_configs 
    $out = $if_obj->verify_port_configs;

    verify following configuration if the switch port.
    1. it should not have switchport mode trunk
    2. it should have following three configuration
        b. spanning-tree port type edge
        c. spanning-tree bpduguard enable
    3. if port is on channel-group than it will save the channel group seting 
       on port and remove port from that channel group for the test and at the 
       end of the test, it will restore back this channel-group setting.

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub verify_port_configs {

    my $self = shift;
    
    my @configs = $self->show_running_config();
    if(grep(/switchport\smode\strunk/,@configs)) {
        croak "verify_port_configs:port set to trunk mode. it should be in access mode";
    }
    
    if(!grep(/spanning\-tree\sport\stype\sedge/,@configs)) {
        croak "verify_port_configs:spanning-tree port type should be edge";
    }
    
    # FEX module does not support following config on port.
    # if(!grep(/spanning\-tree\sbpduguard\senable/,@configs)) {
    #    croak "verify_port_configs:spanning-tree bpduguard should be enable";
    # }
    
    my @found;
    #verify, save and remove if port is on any channel-group

    if(@found = grep(/^channel\-group(.*)/,@configs)) {
        logStep("verify_port_configs:port found on channne group: $found[0] >> save ".
                "and remove channel-group from port");
        $self->{ch_group} = $found[0];    
        $self->config_interface("no $found[0]");
    }
    
    return; 
}
#--------------------------------------------------------------------
=head2 save_port_setting 
    $out = $if_obj->save_port_setting;
    save the initial setting of the port. it will restore these at cleanup. 
    It will save the setting in interface object itself. it will store in 
    port_setting HASH reference.

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub save_port_setting {

    my $self = shift;

    my %fc = $self->show_interface_flowcontrol();
    my %brief = $self->show_interface_brief();

    $self->{port_setting} = { TxAdmin => $fc{TxAdmin},
                              RxAdmin => $fc{RxAdmin},
                              Speed => $brief{Speed},
                              VLAN => $brief{VLAN},
                              Mode => $brief{Mode},
                            };
    return;
}

#--------------------------------------------------------------------
=head2 restore_port_setting 
    $out = $if_obj->restore_port_setting;
    It will Restore the setting back after finishing the test. 
    it will restore setting exactly as it was beginning of test.

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub restore_port_setting {

    my $self = shift;

    my %fc = $self->show_interface_flowcontrol();
    my %brief = $self->show_interface_brief();
    
    my %cur_set = ( TxAdmin => $fc{TxAdmin},
                    RxAdmin => $fc{RxAdmin},
                    Speed => $brief{Speed},
                    VLAN => $brief{VLAN},
                    Mode => $brief{Mode},
                   );
    
    my (@cmds,$value,$key);

    foreach $key (keys %{$self->{port_setting}}) {
        $value = $self->{port_setting}->{$key}; 
        if($value ne $cur_set{$key}) {
            if($key eq 'TxAdmin') {
                push(@cmds,"flowcontrol send $value");
            } elsif($key eq 'RxAdmin') {
                push(@cmds,"flowcontrol receive $value");
            } elsif($key eq 'Speed') {
                if($value eq '1000(D)') {
                    $value = '1000';
                } elsif($value eq '10G(D)') { 
                    $value = '10000';
                }
                push(@cmds,"speed $value");
            } elsif($key eq 'VLAN') {
                push(@cmds,"switchport access vlan $value");
            } elsif($key eq 'Mode') {
                push(@cmds,"switchport mode $value");
            } else {
                
            }
        }
    }
   
    if(scalar(@cmds) != 0) {
        eval { $self->config_interface(@cmds); };
        croak $@ if $@;
    }
    
    eval {
        #restore channel-group if it exists before test
        if(defined($self->{ch_group})) {
            $self->config_interface($self->{ch_group});
        }

        $self->config_status(status => "up");
    };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 set_port_setting
    $out = $if_obj->set_port_setting( fc_send => 'on',
                                      fc_rec => 'on',
                                      status => 'up',
                                      vlan => 1,
                                      mode => 'access',
                                      speed => 10000
                                    );
    set the port with specified setting.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc_send
    (optional): flowcontrol send value(on/off).  

=item fc_rec
    (optional):flowcontrol receive value(on/off)

=item status 
    (optional): port status(up/down)

=item vlan
    (optional): vlan id, port will allow

=item mode
    (optional): port mode(access/trunk)

=item speed
    (optional): port speed(10000/1000)

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub set_port_setting {

    my $self = shift;

    my (%opts) = validate (
                            @_,
                            {
                                fc_send => { type => SCALAR, optional => 1},
                                fc_rec => { type => SCALAR, optional => 1 },
                                status => { type => SCALAR, optional => 1 },
                                vlan => { type => SCALAR, optional => 1 },
                                mode => {type => SCALAR, optional => 1 },
                                speed => { type => SCALAR, optional => 1},
                            }
                          );
                          
    my @cmds;

    if(defined($opts{fc_send})) {
        push(@cmds,"flowcontrol send $opts{fc_send}");
    }
    if(defined($opts{fc_rec})) {
        push(@cmds,"flowcontrol receive $opts{fc_rec}");
    }
    if(defined($opts{status})) {
        if($opts{status} =~ /up/i) {
            push(@cmds,"no shutdown");
        } else {
            push(@cmds,"shutdown");
        }
    }
    if(defined($opts{vlan})) {
        push(@cmds,"switchport access vlan $opts{vlan}");
    }
    if(defined($opts{mode})) {
        push(@cmds,"switchport mode $opts{mode}");
    }
     if(defined($opts{speed})) {
        push(@cmds,"speed $opts{speed}");
    }
    
    if(scalar(@cmds) != 0) {
      eval{  $self->config_interface(@cmds); };
      croak "$@" if $@;
    }

    return;
}
#--------------------------------------------------------------------
=head2 add_port_to_vlan
    $out = $if_obj->add_port_to_vlan(vlan_obj => $vlan_obj);
    add the port to specified vlan.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vlan_obj
    vlan object of Nxvlan.pm.  

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub add_port_to_vlan {

    my $self = shift;
    my (%opts) = validate (
                            @_,
                            {
                                vlan_obj => { type => OBJECT }
                            }
                          );
    
    eval {
	    #config the port mode to trunk
	    $self->config_switchport_mode(mode => 'trunk');
	    
	    my $vlan = '1,' . $opts{vlan_obj}->{vlan_id};
	
	    #allowed vlan will be 1 and the specified one. if specified is 999 than it is 1,999
	    $self->config_switchport_trunk_allowed_vlan(vlan => $vlan);
	
	    #enable portfast 
	    $self->config_spanning_tree_port_type(type => 'edge trunk');
    };
    croak $@ if $@;

    return;
}

#--------------------------------------------------------------------
=head2 delete_port_from_vlan
    $out = $if_obj->delete_port_from_vlan(vlan_obj => $vlan_obj);
    delete port from vlan.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vlan_obj
    (optional): vlan object of Nxvlan.pm.  

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub delete_port_from_vlan {

    my $self = shift;
    my (%opts) = validate (
                            @_,
                            {
                                vlan_obj => { type => OBJECT, optional => 1 }
                            }
                          );
    
    eval {
	    #config the port mode to access
	    $self->config_switchport_mode(mode => 'access');
	
	    #remove the allowed vlan from the trunk port
	    $self->config_interface("no switchport trunk allowed vlan");
	
	    #disable portfast
	    $self->config_spanning_tree_port_type(type => 'edge');
    };
    croak $@ if $@;

    return;
}

1;
