# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Nexus/Nxswitch.pm#2 $
#
## @summary Nxswitch -- Provide a class for the Nexus switch
## @author  Nayan Patel(nayankumar.patel@netapp.com)
## @status  public
## @notes   Created by NICQA Group
## @pod here

=head1 NAME

    Nxswitch - Provide a class for the Nexus 5010 switch

=head1 SYNOPSIS

    use Network::Nexus::Nxswitch;

# Switch constructor
  
    $switch_obj = Network::Nexus::Nxswitch->new(hostspec => $sw_name);

=head1 DESCRIPTION

    Provides a class for the cisco Nexus switch.

=head1 STATUS

    Shared

=head1 AUTHOR

    Nayan patel (NIC QA)
    
=head1 COPYRIGHT

    Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Nexus::Nxswitch;

use strict;
use warnings;

use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT);
use SPARSE::Util::Logger qw(logStep logWarn);
use Network::Nexus::Nxinterface;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};


#---------------------------------------------------------------------

=head2 new

  $switch_obj = Network::Nexus::Nxswitch->new(hostspec => $sw_name
                                       [, connspec => $connspec ]);

Creates an independent hardware Nexus switch object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostspec

    (Required) : switch name(EX: net-5010-1).

=item connspec

    (Optional) : Connection specification for the switch. Defaults to '#ssh-persist'.

=item timeout

    (Optional) : Timeout for commands to the switch. Defaults to 120 secs.

=back

=item Returns

    Switch object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub new {

    my $class = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                hostspec => { type => SCALAR },
                                connspec => { type => SCALAR, default => '#ssh-persist' },
                                timeout => { type => SCALAR, default => 120 },
                            }
                          );

    #Get the connection object to switch
    Tharn::host($opts{hostspec});
    $opts{conn} = Tharn::connect("$opts{connspec};username=admin;password=corduroy");
    $opts{conn}->timeout($opts{timeout});

    # set Nexus OS version from nate_host file
    my %fields = $opts{conn}->hostp->param_fields();
    $opts{nx_os_version} = $fields{OS_VER};

    # get the hosttype from the nate_host file of this switch
    my $type = $opts{conn}->hostp->hosttype;
    if($type !~ /switch-cisco/) {

        croak "Switch type is not supported: $type";
    }

    return bless \%opts, $class;

}

#---------------------------------------------------------------------

=head2 nexus_say

  $out = $sw_obj->nexus_say(
                            cmd => 'show policy-map cluster',
                            mode => 2
                           );

Execute the specified command on the switch and give the output of that command.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cmd

    (Required) : Command, you want to execute on the switch(EX:show interface).

=item mode

    (Optional) : It specified the mode on the switch, on which you want to execute specified command. 
    mode = 0 or nothing -- means use whatever mode is on this connection
    mode = 1 -- means go to config terminal mode
    mode = 2 -- means go to EXEC mode
    
=back

=item Returns

   
  switch output of the specified command

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub nexus_say {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                cmd => { type => SCALAR },
                                mode => { type => SCALAR, optional => 1, default => '0' },
                            }
                          );

   
    my (@op_list, $qcmd);
    my $goto_mode;
    # $mode = 0 or nothing -- means use whatever mode is on this connection
    # $mode = 1 -- means go to config terminal mode
    # $mode = 2 -- means go to EXEC mode
    

    if(defined($opts{mode}) and $opts{mode} == 1){ 
        $goto_mode = 'configure terminal';
    } elsif(defined($opts{mode}) and $opts{mode} == 2){
        $goto_mode = 'end';
    } else {       
        $goto_mode = undef;
    }

    if(defined($goto_mode)){
        #go to the specified mode
        Tharn::say($self->{conn},$goto_mode);
    }

    # send command to the nexus switch
    @op_list = Tharn::say($self->{conn},$opts{cmd},'-swallow');
    
    if(defined $op_list[2]){

        #remove all '\r' from the output
        $op_list[2] =~ s/\r+//g;
    
        #rint "op_list 2 is $op_list[2]\n";
        #remove echo of command from the output
        #############################################
        #  there r problem in following removing. error message in system jumbomtu 5000 is completely remove in following action, which should not.
        ############################################
        $qcmd = quotemeta($opts{cmd});
        #print "qcmd is $qcmd\n";
        $op_list[2] =~ s/.*$qcmd\n*//s;
        
        #print "op_list 2 is $op_list[2]\n";
        #remove console messages if any
        #not implemented for NOW
        
    }

    return $op_list[2];
}

#---------------------------------------------------------------------

=head2 show_switch_config

  $out = $sw_obj->show_switch_config(cmd => 'show interface e1/2 brief');

excute the command in the EXEC mode and give the command output. normally, it is use to get the switch configuration.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cmd

    (Required) : Command, you want to execute on the switch(EX:show interface).

=back

=item Returns
   
  output of the command.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_switch_config {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                cmd => { type => SCALAR },
                            }
                          );
   
    return $self->nexus_say(cmd => $opts{cmd}, mode => 2);
}


#---------------------------------------------------------------------

=head2 show_policy_map

  $out = $sw_obj->show_policy_map(policy_map => 'cluster');
    It use to get the information about the specified policy map.

The following named arguments are accepted:

=over

=item Arguments

=over

=item policy_map

    (Required) : Name of the policy map.

=back

=item Returns
   
  return the output of the command "show policy-map <policy_map>".

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
   
sub show_policy_map {
    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                policy_map => { type => SCALAR },
                            }
                          );
    
    return $self->show_switch_config(cmd => "show policy-map $opts{policy_map}");
}

#---------------------------------------------------------------------

=head2 verify_policy_map

  $out = $sw_obj->verify_policy_map(policy_map => 'cluster', class_map => 'default-class');
    Verify the following information in the policy map.
    1. class map should the one which is specified in the parameter. normally, it is class-default.
    2. mtu = 9216 and pause = no-drop

The following named arguments are accepted:

=over

=item Arguments

=over

=item policy_map

    (Required) : Name of the policy map.

=item class_map

    (Required) : Name of the class map in the policy map.

=back

=item Returns
   
    Return 1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub verify_policy_map {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                policy_map => { type => SCALAR },
                                class_map => { type => SCALAR }
                            }
                          );

    my $policy_map = $opts{policy_map};
    my $class_map = $opts{class_map};

    my $config = $self->show_policy_map(policy_map => $policy_map);
    if(($config =~ /Policy map not created/) or ($config =~ /ERROR/)){
        croak "Problem in a requested policy map : $config";
    }

    my @text = split(/\n\n/,$config);

    shift(@text);
    pop(@text);

    my @class_map_info;
    my $temp_class_map;
    my $i;

    for($i=0; $i<scalar(@text);$i++){
        @class_map_info = split(/\n/,$text[$i]);
        $temp_class_map = shift(@class_map_info);
        $temp_class_map =~ s/^\s+//;
        if($temp_class_map eq "Class $class_map"){
            last;
        }
    }

    if($i == scalar(@text)){
        croak "Could not find the required class map in the policy map\n";
    }
    #change in golder config, pause no-drop has been removed.
    #my($mtu, $pause); 
    my($mtu); 
    foreach my $info (@class_map_info){
        $info =~ s/^\s+//;
        my ($key, $value) = split(/\s/,$info);
        if($key =~ /mtu/){
            if($value != 9216){
                croak "MTU is $value. It should be 9216 in class map $class_map under policy map $policy_map.";
            }
            $mtu = 1;
        }elsif($key =~ /pause/){
     #       if($value !~ /no-drop/){
     #           croak "PAUSE is $value. It should be no-drop in class map $class_map under policy map $policy_map.";
     #       }
     #       $pause = 1;
        }elsif($key =~ /bandwidth/){
            if($value != 50){
                croak "BANDWIDTH is $value. It should be 50 percent in class map $class_map under policy map $policy_map.";
            }
        }else{
            
           #key not match with above conditions. 
        }

    }

    if($mtu != 1){
        croak "Not all the parameters are set in the class map $class_map. required parameters are MTU,PAUSE.";
    }
    return 1; 
}


#---------------------------------------------------------------------

=head2 show_class_map

  $out = $sw_obj->show_class_map(class_map => 'default-class');
    return the information of the specified class map.

The following named arguments are accepted:

=over

=item Arguments

=over

=item class_map

    (Required) : Name of the class map.

=back

=item Returns
   
    Return the output of the command "show class-map <class_map>".

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_class_map {
    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                class_map => { type => SCALAR }
                            }
                          );
 
    return $self->show_switch_config(cmd => "show class-map $opts{class_map}");
}


#---------------------------------------------------------------------

=head2 verify_class_map

  $out = $sw_obj->verify_class_map(class_map => 'default-class');
    verify whetehr the specified class_map is exists or not.

The following named arguments are accepted:

=over

=item Arguments

=over

=item class_map

    (Required) : Name of the class map.

=back

=item Returns
   
    Return 1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub verify_class_map {

    my $self = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                class_map => { type => SCALAR }
                            }
                          );
 
    my $config = $self->show_class_map(class_map => $opts{class_map});
    if(($config =~ /Class map not created/) or ($config =~ /ERROR/)){
        croak "Problem in a requested class map : $config";
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 show_version

  $out = $sw_obj->show_version;
    show the version of the switch software.

The following named arguments are accepted:

=over

=item Arguments

=over

=back
=item Returns
   
    Retrun the output of the command "show version".

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub show_version {
    
    my $self = shift;
    return $self->show_switch_config(cmd => "show version | no-more");
}

#---------------------------------------------------------------------

=head2 show_system_qos

  $out = $sw_obj->show_system_qos;
        show the information about the system qos. 

The following named arguments are accepted:

=over

=item Arguments

=over

=back
=item Returns
   
    Retrun the output of the command "show system qos".

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_system_qos {
    
    my $self = shift;
    return $self->show_switch_config(cmd => "show system qos");

}


#---------------------------------------------------------------------

=head2 verify_system_qos

  $out = $sw_obj->verify_system_qos(service_policy => 'cluster');
        verify that specified service policy is available in the system qos.

The following named arguments are accepted:

=over

=item Arguments

=over

=item service_policy

    (Required) : Name of the service policy.

=back

=item Returns
   
    Retrun the output of the command "show system qos".

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub verify_system_qos {

    my $self = shift;
    my (%opts) = validate (
                            @_,
                            { 
                                service_policy => { type => SCALAR, default => 'cluster' },
                            }
                         );
    
    my $config = $self->show_system_qos;
    if($config =~ /ERROR/){
        croak "Problem in a requested system qos : $config";
    }
    my @text = split(/\n/,$config);
    my $temp = $text[-1];
    my($temp1,$sp) = split(/\:\s/,$temp);

    if($sp ne $opts{service_policy}){
        croak "system qos is not set with $opts{service_policy} service policy.";
    }

    return 1;
}

#------------------------------------------------------------------------------------------------
=head2 config_switch

  $out = $sw_obj->config_switch(cmd => 'configure terminal');
    configure the switch in the config terminal mode. Actually, it is execute the specified command in the configure terminal mode.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cmd

    (Required) : command, you want to execute in configure terminal mode.

=back

=item Returns
   
    Retrun the output of the specified command.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_switch {
    my $self = shift;
    my (%opts) = validate (
                            @_,
                            { 
                                cmd => { type => SCALAR },
                            }
                         );
    
    return $self->nexus_say(cmd => $opts{cmd}, mode => 1);
}

#--------------------------------------------------------------------
=head2 config_feature

  $out = $sw_obj->config_feature(feature => 'lacp');
    enable the feature on the switch.

The following named arguments are accepted:

=over

=item Arguments

=over

=item feature

    (Required) : name of a feature, you want to enable.

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub config_feature {
    my $self = shift;
    my (%opts) = validate (
                            @_,
                            { 
                                feature => { type => SCALAR },
                            }
                         );
    
    my $out = $self->config_switch(cmd => "feature $opts{feature}");
    if($out =~ /ERROR/){
        croak "Problem in configuring a feature $opts{feature} : $out";
    }
 
    return;
}


#--------------------------------------------------------------------
=head2 config_system_jumbomtu

  $out = $sw_obj->config_system_jumbomtu(mtu => 9216);
    set the jumbo mtu on the system.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mtu 

    (optional) : mtu, you want to set. default is 9216.

=back

=item Returns 

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_system_jumbomtu {
    my $self = shift;
    my (%opts) = validate (
                            @_,
                            { 
                                mtu => { type => SCALAR, optional => 1 },
                            }
                         );
 
    if(!defined($opts{mtu})){
        $opts{mtu} = 9216;
    }
    if($opts{mtu} < 2240 or $opts{mtu} > 9216){
    
        croak "SORRY: you can not configure mtu $opts{mtu}. allowed value is 2240 to 9216.";
    }
    my $out = $self->config_switch(cmd => "system jumbomtu $opts{mtu}");
    if($out =~ /^ERROR/){
        croak "Problem in configuring system mtu to $opts{mtu} : $out";
    }
 
    return;
}

#--------------------------------------------------------------------
=head2 get_port_from_mac

  %out = $sw_obj->get_port_from_mac(mac => '00:11:22:33:44:55');
    find the directly connected switch port of the filer port with specified MAC address.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mac

    (requierd) : MAC address of the filer port.

=back

=item Returns 

    name of the directly connected switch port.
    EX: 2/2
=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub get_port_from_mac {

    my $self = shift;
    my (%opts) = validate (
                            @_,
                                {
                                    mac => { type => SCALAR },
                                }
                           );

    my $hw_addr = $opts{mac};

    if(!defined($hw_addr)) {
        croak "Please provide the hardware address.";
    }
   
    # add extra zero, if the mac is not in a proper format
    # following three line will converts 0:c:29:c2:5c:6b to 00:0c:29:c2:5c:6b
    $hw_addr =~ s/^(\w)\:/0$1\:/i;
    $hw_addr =~ s/\:(\w)\:/\:0$1\:/ig;
    $hw_addr =~ s/\:(\w)$/\:0$1/i;

    if($hw_addr !~ /^([0-9a-f]{2}([:-]|$)){6}$/i) {
        croak "You have provided wrong MAC address, which is $hw_addr";
    }

    $hw_addr =~ s/://g;
    substr($hw_addr,4,0) = '.';
    substr($hw_addr,-4,0) = '.';

    my $cmd = "show mac address-table address $hw_addr";
    if ($self->{nx_os_version} =~ /^4.*/) {
        $cmd = "show mac-address-table address $hw_addr";
    }

    my $config = $self->show_switch_config(cmd => $cmd);
   
    if($config =~ /Total MAC Addresses: 0/) {
        croak "You need to ping from the machine to find out connected switch interface.";
    }

    my @temp = split(/\n/,$config);
    my @line = grep($_ =~ /$hw_addr/i,@temp);
    my (@value);
    if (@line) {
        @value = split(/\s+/,$line[0]);
    } else {
        croak "interface is not directly connected to the specified switch.";
    }

    # we only need last value which is the actual port, 
    # split that value in interface type and name
    $value[-1] =~ m/([a-zA-Z]+)(.+)/;
    my %if = (type=>$1,name=>$2);
    
    if($if{type} !~ /^Eth/i) {
        croak "Your interface is not directly connected to the specified switch.";
    }

    return $if{name};
}

#--------------------------------------------------------------------
=head2 show_running_config

  @out = $sw_obj->show_running_config(grep_text => 'feature lacp');
    get all the running configuration of the switch.
    it may grep the output based on the text you have provided.

The following named arguments are accepted:

=over

=item Arguments

=over

=item grep_text
    (optional) : the text on which it will grep the o/p of show running-config command.

=back

=item Returns 

    array of the running configs on the switch. if you have provided grep text than it will give array of the greped running configs.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub show_running_config {

    my $self = shift;
    my (%opts) = validate (
                            @_,
                                {
                                    grep_text => { type => SCALAR, optional => 1 },
                                }
                           );

    my $cmd = "show running-config | no-more";
    if(defined($opts{grep_text})) {
        $cmd = "show running-config | grep $opts{grep_text} | no-more";
    }

    my $out = $self->show_switch_config(cmd => $cmd);
    
    my @out1 = split(/\n/,$out);
    foreach my $ele (@out1) {
            $ele =~ s/^\s+//;
    }
    
    return @out1;
}

1;
