# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Nexus/Nxvlan.pm#1 $
#
## @summary Nxvlan -- Provide a class for the vlan in the Nexus switch
## @author  Nayan Patel(nayankumar.patel@netapp.com)
## @status  public
## @notes   Created by NICQA Group
## @pod here

=head1 NAME

    Nxvlan - Provide a class for the Nexus 5010 switch vlan sub-mode

=head1 SYNOPSIS

    use Network::Nexus::Nxvlan;

# Switch vlan constructor

    $vlan_obj = Network::Nexus::Nxvlan->new(conn => $sw_obj{conn}, vlan_id => $vlan_id);

=head1 DESCRIPTION

    Provides a class for the cisco Nexus switch vlan sub-mode.

=head1 STATUS

    Shared

=head1 AUTHOR

    Nayan Patel (NIC QA)
    
=head1 COPYRIGHT

    Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Nexus::Nxvlan;

use strict;
use warnings;

use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT);
use SPARSE::Util::Logger qw(logStep logWarn);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};
#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $VERSION = '1.0';
our @ISA;

#---------------------------------------------------------------------

=head2 new

    $vlan_obj = Network::Nexus::Nxvlan->new(conn => $sw_obj{conn}, vlan_id => $vlan_id);


Creates an independent hardware Nexus switch vlan object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item switch_conn_obj

    (Required) : switch connection object from the switch object (EX: $opts{conn}).

=item vlan_id

    (required) : It is a vlan id. (EX: 1 to 3967, 4048 to 4093).

=back

=item Returns

    Switch vlan object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub new {

    my $class = shift;
    my (%opts) = validate (
                            @_, 
                            {
                                conn  => { type => HASHREF },
                                vlan_id => { type => SCALAR },
                            }

                          );

     # get the hosttype from the nate_host file of this switch
     my $type = $opts{conn}->hostp->hosttype;
     if ( $type =~ /switch-cisco/ ) {
          require Network::Nexus::Nxswitch;
          @ISA = qw(Network::Nexus::Nxswitch);
     } else {
          croak "Switch type is not supported: $type";
     }

    return bless \%opts, $class;
}

#---------------------------------------------------------------------

=head2 config_vlan

    $vlan_obj->config_vlan(<array of commands>);

    configure the vlan on the nexus switch

The following named arguments are accepted:

=over

=item Arguments

=over

=item <commands>
    (Required) : array of commands.
    EX: @cmds = ('state active', 'name vlan1', 'no shutdown')
    $vlan_obj->config_vlan(@cmds);

=back

=item Returns

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_vlan {

    my ($self,@cmds) = @_;

    if(!defined($cmds[0])){
        croak "Please provide commands to configure vlan";
    }
    
    my ($cmd,$out);
    #go to vlan sub-mode
    $out = $self->nexus_say(cmd => "vlan $self->{vlan_id}", mode => 1);
    if($out =~ /(^ERROR(.*) | invalid)/i) {
        die "Not able to go into the vlan sub-mode."
    }
    foreach $cmd (@cmds){ 
        $out = $self->nexus_say(cmd => $cmd, mode => 0);
        if($out =~ /(^ERROR(.*) | invalid)/) {
            croak "command failed: $out ";
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 config_state

    $vlan_obj->config_state(state => 'active');

    configure the state of the vlan on the nexus switch

The following named arguments are accepted:

=over

=item Arguments

=over

=item state
    (Required) : state of the vlan

=back

=item Returns

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_state {

    my $self = shift;

    my %opts = validate (
                            @_,
                            {
                                state => { type => SCALAR }
                            }
                        );
    
    my $cmd = "state $opts{state}";
    $self->config_vlan($cmd);

    #save state to the vlan object
    $self->{state} = $opts{state};

    return;
}   

#---------------------------------------------------------------------

=head2 config_name

    $vlan_obj->config_name(name => 'vlan1');

    configure the name of the vlan on the nexus switch

The following named arguments are accepted:

=over

=item Arguments

=over

=item name
    (Required) : name of the vlan

=back

=item Returns

=back

=for html <hr />

=cut

#---------------------------------------------------------------------


sub config_name {

    my $self = shift;
    my %opts = validate (
                            @_,
                            {
                                name => { type => SCALAR }
                            }
                        );
    
    my $cmd = "name $opts{name}";
    $self->config_vlan($cmd);
    
    #save name to the vlan object
    $self->{name} = $opts{name};

    return;
}   

#---------------------------------------------------------------------

=head2 enable_vlan

    $vlan_obj->enable_vlan();

    enable vlan on the nexus switch

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub enable_vlan {

    my $self = shift;

    my $cmd = "no shutdown";
    $self->config_vlan($cmd);

    return;
}

#---------------------------------------------------------------------

=head2 disable_vlan

    $vlan_obj->disable_vlan();

    disable vlan on the nexus switch

The following named arguments are accepted:

=over

=item Arguments

=over

=back

=item Returns

=back

=for html <hr />

=cut

#---------------------------------------------------------------

sub disable_vlan {

    my $self = shift;

    my $cmd = "shutdown";
    $self->config_vlan($cmd);

    return;
}

1;
