# $Id: //depot/prod/test/Rsierranevada/lib/Network/Nis.pm#1 $
#
## @summary Routines for invoking NIS operations.
##
## @description
##              These functions are used primarily by NIS tests
##              by setting options, /etc/nsswitch values, configuring
##              NIS servers, NIS domains and enabling NIS.
##
## @dependencies
##
## use Tharn;
##
## @author  dl-maryv (FSS Protocols QA)
##
## @change 2005-05-03 gajria: added nis_slave_is_supported()
##                            added nis_slave_map_download_verify()
##
##

package Network::Nis;

require Exporter;
use strict;
use Tharn;
use Data::Dumper;
use SayWrap qw(sayEOF);
use Carp qw(croak);
use SPARSE::Util::Logger qw(logDbg logTrc);
use filer qw(filer_options_set);
use Filer::Version qw(get_wafl_version WV_TSINGTAO_MIN
		      UNSUPPORTED_FEATURE_EXCEPTION);
use Network::Utils qw(run_getXXbyYY config_ipfw_rule run_ping);

use vars qw (@ISA @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT_OK = qw (
		 get_nis_info
		 nis_binding_check
		 nis_block_traffic_to_server
		 nis_command
		 nis_configure
		 nis_get_master
		 nis_rebuild_maps
		 nis_slave_is_supported
		 nis_slave_map_download_verify
		 nis_config_verify
		 );


## @name ::nis_configure
##
## @summary configures NIS by assigning values to the nis.* options.
## gets nis_domain_name and nis_servers as arguments. Other values are
## set to default values.
## After assigning the values, invokes nis_config_verify() to verify if the
## values are displayed by 'nis info' command.
##
## @arg $fc, %options
##  $fc - required. filer connection object
##  %options - Optional.  defaults to NATE command line parameters
##  nis_domain_name - string. name of the nis domain name
##  nis_servers - string containing IP addresses of servers seperated by commma
##
##  example: nis_configure($fc, nis_domain_name => 'domain_name',
##                              nis_servers => '1.2.3.4,5.6.7.8.9');
##

sub nis_configure
{
    my ($fc, %options) = @_;

    my %nis_options = ();

    my $version = get_wafl_version($fc->{hostp});

    # Initialize the option strings for all the NIS options
    my $nis_domain_name_str = "nis.domainname";
    my $nis_enable_str = "nis.enable";
    my $nis_group_update_enable_str = "nis.group_update.enable";
    my $nis_group_update_schedule_str = "nis.group_update_schedule";
    my $nis_servers_str = "nis.servers";
    my $nis_slave_enable_str = "nis.slave.enable";



#    Set the default values; nis_enable = on;  nis_group_update = on;
#    nis_group_update_schedule  = 24;
#    $nis_options{$nis_enable_str}  =  'on';
#    $nis_options{$nis_group_update_enable_str}  =  'on';
#    $nis_options{$nis_group_update_schedule_str}  =  24;

#    if ( defined($options{nis_domain_name}) ) {
#        $nis_options{$nis_domain_name_str}  =  $options{nis_domain_name};
#    } else {
#        $nis_options{$nis_domain_name_str}  =  $main::NIS_DOMAIN;
#    }

#    if ( defined($options{nis_servers}) ) {
#        $nis_options{$nis_servers_str} = $options{nis_servers};
#    } else {
#        $nis_options{$nis_servers_str} = $main::NIS_SERVER;
#    }

    $nis_options{$nis_domain_name_str} = 
	$options{nis_domain_name} || $main::NIS_DOMAIN;
    $nis_options{$nis_servers_str} = 
	$options{nis_servers} || $main::NIS_SERVER;
    $nis_options{$nis_enable_str}  =  $options{nis_enable} || 'on';
    $nis_options{$nis_group_update_enable_str}  =
	$options{nis_group_update_enable} || 'on';
    $nis_options{$nis_group_update_schedule_str} = 
	$options{nis_group_update_schedule} ||  24;

    if ($version >= WV_TSINGTAO_MIN) {
	$nis_options{$nis_slave_enable_str} = 
	    $options{nis_slave_enable} ||  'off';
    }

    # Set the option values.
    foreach my $option_str ( $nis_domain_name_str, $nis_servers_str, 
			     $nis_group_update_enable_str, 
			     $nis_group_update_schedule_str, 
			     $nis_enable_str ) {
        my $option_val = $nis_options{$option_str};
	logTrc("configure_NIS: setting $option_str to $option_val");
        filer_options_set($fc, $option_str, $option_val);
    }
    # Set the NIS domain name first

    #Verify the values are displayed in nis info command.
    sleep(10);	
    nis_config_verify($fc, 
		      nis_domain_name => 
		      $nis_options{$nis_domain_name_str},
		      nis_servers => $nis_options{$nis_servers_str})
        or croak "NIS config verification failed";

    return 1;
}


## @name ::nis_config_verify
##
## @summary Invokes "nis info" cli command on the filer and compares the
## info for domain name and nis servers.
## This function is private and invoked by nis_configure()
##
## @arg $fc, %options
##  $fc - required. filer connection object
##  %options - required.  
##  nis_domain_name - string. name of the nis domain name
##  nis_servers - string containing IP addresses of servers seperated by commma
##
##

sub nis_config_verify
{
    my ($fc, %options) = @_;

    if ( !defined($options{nis_domain_name}) ||
         !defined($options{nis_servers}) ) {
	logDbg("cfg_verify_NIS: function invoked without parameters");
        return 0;
    }

    my $NIS_ERROR_STRING="NIS is not enabled";
    my $nis_domain = undef;
    my $nis_servers = undef;

    # Send cli command to the filer to get nis info.
    my $filer_output = sayEOF($fc, "nis info");

    my @output_lines = split /\n/, $filer_output;

    # If NIS is not enabled, return 0, else extract the domain and
    # nis_servers information and compare with expected values.
    # if they don't match return 0.

    if ( $output_lines[0] =~ m/$NIS_ERROR_STRING/ )  {

        logDbg("cfg_verify_NIS: error in nis : $output_lines[0]");
        return 0;

    } else {
        my $seperator = "";
	# multiple nis servers are seperated by ,
        foreach my $lines ( @output_lines ) {
            # make sure domain doesn't contain white spaces at the end.
            if ( $lines =~ m/(.*) domain is (.*[a-z])/ ) {

                $nis_domain = $2;

            } elsif ( $lines =~ m/^(\d+).(\d+).(\d+).(\d+)/) {
		
                # since there could be multiple servers, concatenate them
                # into a single string seperated by ,
                if ( defined($nis_servers) ) {
                    $seperator = ",";
                }
                $nis_servers = $nis_servers . $seperator . "$1.$2.$3.$4";
            } elsif ( $lines =~ m/^([\da-f:]+)/) {
                #if the server address is in ipv6 format
                if ( defined($nis_servers) ) {
                    $seperator = ",";
                }
                $nis_servers = $nis_servers . $seperator . "$1";
            }
        }

        if ( $nis_domain ne $options{nis_domain_name} ) {
            logDbg
		("cfg_verify_NIS: nis_domain = $nis_domain not expected");
            return 0;
        } elsif ( $nis_servers ne $options{nis_servers} ) {
            if ( $options{nis_servers} !~ m/\*/ ) {
                logDbg("cfg_verify_NIS: nis_servers = $nis_servers not ".
		       "expected");
                return 0;
	    } else {
		logDbg
		    ("cfg_verify_NIS: nis_servers mis match because of *");
		return 1;
	    }
        } else {
            return 1;
        }
    }
}

## @name :: nis_binding_check
##
## @summary
##   Check if NIS binding exists and return name/ip of bound nis server
##
## @args $conn Required:  connection object
## 

sub nis_binding_check( $ ) {

    my $conn = shift;

    my $line = sayEOF($conn,"nis info");
    $line = sayEOF($conn,"ypwhich");

    if ($line =~ m/not bound to any nis server/i) {
        croak "NIS client not bound to any NIS server";
    }

    if ($line =~ m/nis is not enabled/i) {
        croak "NIS not enabled";
    }

    return($line);
}

## @name ::nis_command
##
## @summary 
##   Invokes the specified nis command (ypmatch,ypcat,ypwhich,ypgroup)
##   and returns the output from the filer
##
## @arg $fc, $command, %options
##  $fc - required. filer connection object
##  $command - actual command to be invoked on the filer.
##  %options - Optional.  options and values that are command specific.
##
##  example: 
##        $filer_output = nis_command($fc, "ypmatch", 
##		key => "nwk-c44", map_name => "hosts");
##
sub nis_command
{
    my ($fc, $command, %options) = @_;

    my $vfiler_name = $options{'vfiler_name'};
    my $map_name = $options{map_name};
    my $key = $options{key};
    my $flags = $options{flag};
    my $command_str = "";

    if ( $command !~ /(nis info|ypmatch|ypcat|ypwhich|ypgroup)/ ) {
	logDbg("nis_command: $command not supported");	
	return 0;
    }

    # construct the command str
    if ($vfiler_name) {
	$command_str = "vfiler run $vfiler_name $command $flags $key $map_name";
    } else {
	$command_str = "$command $flags $key $map_name";
    }

    logTrc("nis_command: $command_str");
    my $filer_output = sayEOF($fc, $command_str);
    return ($filer_output);
}


## @name ::nis_slave_is_supported
##
## @summary
##    Check if nis slave is supported
##
## @description
##     Checks if nis slave feature is supported in current ontap and 
##     returns 1 or exception
## 
## @arg $fh  Required Filer host object
##
## @return
##     1 if nis slave is supported
##     UNSUPPORTED_FEATURE_EXCEPTION is nis slave is not supported
##

sub nis_slave_is_supported {

    my $fh = shift;

    my $version = get_wafl_version($fh);
    if ($version < WV_TSINGTAO_MIN) {
	croak(UNSUPPORTED_FEATURE_EXCEPTION . 
	      ": nis slave unsupported in release prior to TsingTao");
    }
    return 1;
}


## @name ::nis_slave_map_download_verify
##
## @summary
##    Check if nis maps have been downloaded.
##
## @description
##    runs nis info command to find out if nis map download has 
##	  completed.
##    The function will sleep 10 seconds on each non-fatal failure until
##    timeout is reached. Fatal errors results in a croak if nis or
##    nis slave is disabled.
##    If domain is passed, this function checks if maps actually got
##	  downloaded into the appropriate domain directory
##
## @args $fc      Required. Filer connection object
## @args %options Optional.
##       keys : vfiler_name => 'vfiler_name', defaults to 'undef'
##              wait => 'yes' or 'no'. If 'yes' , keep checking till 
##				timeout expires
##              If 'no' return immediately after one check.
##				Defaults to yes
##              timeout => 'seconds', Amount of time to wait for map 
##                         download before reporting failure defaults 
##						   to 180 seconds. 
##							Applies only if 'wait' is 'yes'
##				domain => Used to check if maps actually got
##				downloaded into the appropriate domain directory.
##				Defaults to undef
##				path => Used to access the map directory
##				in vfilers.
##				Defaults to ''
##
##
## @return
##      on success: timestamp ("NIS Map last checked time")
##
##      on failure: croak
##
# XXX\gajria: enhance this with a double checking using ls on /etc/yp/<domain> dir
# to ensure that maps were really downloaded. Also time stamp is returned in the same

sub nis_slave_map_download_verify( $;% ) {

    my ($conn,%options,$output) = @_;
    my ($timestamp,$nis_disabled);

    my $vfiler_name = $options{'vfiler_name'} || undef;
    my $timeout     = $options{'timeout'} || 180;
    my $wait        = $options{'wait'} || 'yes';
    my $domain      = $options{'domain'} || undef;
    my $path        = $options{'path'} || '';

    my $starttime = time();
    my $endtime = $starttime + $timeout;


    while ($endtime > time()) {

	if (defined($vfiler_name)) {
	    $output = sayEOF($conn, "vfiler run $vfiler_name nis info");
	} else {
	    $output = sayEOF($conn, "nis info");
	}

	logTrc($output);
	
	if ($output =~
	   m/Time\s+((\w+)\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\w+)\s+(\d+))/) {
	    $timestamp = $1;
	    last;
	}

	if ($output =~ m/NIS slave disabled|NIS is not enabled/i) {
	    $nis_disabled = 1;
	    last;
	}
	if ($output =~ m/NIS Slave No Maps found|not yet checked/i) {
	    if ($wait eq 'yes') {
		logTrc("maps not downloaded.. will check after 10 seconds");
		sleep 10;
		next;
	    } else {
		logTrc("Returning because wait = $wait");
		last;
	    }
	}
    }
	
	## Checking if maps actually got downloaded in the <domain>dir
	## "if" condition is used as existing tests don't pass domain
	if (defined($domain)) {
		my $maps_dir = $path."/etc/yp/$domain";
		$output = sayEOF($conn, "priv set -q advanced;ls $maps_dir");
		my @lines = split(/\n/, $output);
		my $count = 0;
		foreach my $line (@lines) {
			$count++ if ($line =~ /.db/);
		}
		if ( $count <=2 ){
			die fail_result("No NIS maps found in $maps_dir");
		} else {
			logcomment("Found ".$count." map files in $maps_dir");
		}
	}

	if ($timestamp) {
		return $timestamp
	} elsif ($nis_disabled) {
		croak "NIS or NIS Slave is disabled";
	} else {
		croak "NIS map download not completed";
	}
}


## @name ::nis_rebuild_maps
##
## @summary
##     rebuild nis maps on nis server
##
## @description
##     This function runs the necessary steps to build nis maps on a nis master
##     server and verifies if maps were built correctly.
##
##
## @arg $conn Required. Client Connect object
## 
## @arg %options Optional. Hash of options
##      $slave_servers: comma separated list of slave servers
##
## @return
##     on success returns 1
##
##     on failure croak
##
sub nis_rebuild_maps($;%) {

    my ($conn,%options) = @_;
    my $slave_servers = $options{'slave_servers'};

    my @slaves = split /,/,$slave_servers;
    logTrc(Dumper(\@slaves));

    my $hosttype = $conn->hostp()->hosttype();

    if ($hosttype eq 'unix-solaris') {
	say($conn,"ypinit -m", '-noprompt', '-wait',"next host to add:");
	if ($#slaves > 0) {
	    for (my $i=0;$i <= $#slaves - 1 ;$i++) {
		say($conn,"$slaves[$i]",'-noprompt','-wait',
		    "next host to add:");
	    }
	}
	say($conn,"$slaves[$#slaves]",'-noprompt','-wait',
	    "next host to add:");
	say($conn,"\r",'-nocr','-noprompt','-wait',
	    "Is this correct.*");
	say($conn,"\r",'-nocr','-noprompt','-wait',
	    "non-fatal errors.*");
	say($conn,"\r",'-nocr','-noprompt','-wait',
	    "and its contents\?.*");
	say($conn,"y");

    } elsif ($hosttype eq 'unix-linux') {
	
	say($conn,"cd /var/yp");
	say($conn,"make");
    }

# XXX\gajria verify if maps were built
}


## @name ::nis_get_master
##
## @summary
##    get nis master information
##
## @description
##    run "ypmatch YP_MASTER_NAME <map>" and get the nis master name.
##    run getXXbyYY on nis master name and extract the ip address
##
## @arg $fc Required connect object of filer
##
## @return
##     list containing master server name and ip address
##     ($nis_master_name, $nis_master_ip)
##
## @example
##     my ($name,$ip) = nis_get_master($fc);

sub nis_get_master {

    my $fc = shift;

    my ($nis_master_name,$nis_master_ip);

    my $output = nis_command($fc,"ypmatch",
			     map_name => "hosts.byname",
			     key => "YP_MASTER_NAME");

    if ($output =~ m/Can\'t match key YP_MASTER_NAME/) {
	croak ("Could not get nis master information: $output");
    } else {
	$nis_master_name = $output;

	# get master ip address
	my $getXXbyYY_output = 
	    run_getXXbyYY($fc,"gethostbyname_r",$nis_master_name);
	
	# get first returned ip address
	if ($getXXbyYY_output =~ m/not found/i) {
	    croak "Could not resolve nis master name $nis_master_name";
	} elsif ($getXXbyYY_output =~ 
		 m/addresses: ((\d+).(\d+).(\d+).(\d+))\s*/ ) {
	    $nis_master_ip = $1;
	}
	
	return ($nis_master_name,$nis_master_ip);
    }
}


## @name ::nis_block_traffic_to_server
##
## @summary
##    block all traffic to nis server
##
## @description
##    run ypwhich and get nis server information, block traffic to nis
##    server ip using ipfw rule
##
## @arg $fc Required connect object of filer
##      $rule_id Required rule id for ipfw rule
##
##

sub nis_block_traffic_to_server( $$ ) {

    my $conn = shift;
    my $rule_id = shift;

    my $nis_server_ip;
    my $nis_server = nis_binding_check($conn);

    my $ypmatch_host_output = nis_command($conn,"ypmatch",
					  map_name => "hosts.byname",
					  key => $nis_server);

    if ( $ypmatch_host_output =~ m/^((\d+).(\d+).(\d+).(\d+))/) {
	$nis_server_ip = $1;
    } else {
	croak ("nis server name could not be resolved");
    }

    config_ipfw_rule ($conn, 'add', $rule_id,
		      proto => 'ip',
		      to_ip => $nis_server_ip,
		      from_ip => $conn->hostp()->default_ip(),
		      action => 'deny',
		     ) 
	or croak 
	    "Could not set ipfw rule to block traffic to nis server: $@";

    return 1;
}


#
# Name: get_nis_info
# Args: $fc
# Desc: invokes nis info and gets the data for each of listed servers
#       in a hash.
# Returns: the hash_reference.
#

## @name ::get_nis_info
##
## @summary 
##   Invokes the cli command nis info, parses the output, populates a hash
##   with available information for each server. Returns the hash reference.
##
## @arg $fc
##  $fc - required. filer connection object
##
##  example: 
##        $hash_ref = get_nis_info($fc) or die $@;
##
sub get_nis_info
{
    my ($fc) = @_;
    my $nis_info = {};	# Hash ref to  be returned.

    my $filer_out = sayEOF($fc, "nis info");

    if ( $filer_out =~ m/NIS is not enabled/ ) {
	croak "$filer_out";
    }
    my @output_lines = split /\n/, $filer_out;

    #
    # Parse the output from NIS info and gather necessary information
    #
    foreach my $lines ( @output_lines ) {
        #
        # Look for lines beginning with IP addr only
        #
	if ( $lines =~ m/^(\d+).(\d+).(\d+).(\d+)/) {
	    my $ip_addr = "$1.$2.$3.$4";
            $lines =~ 
		m/$ip_addr\s+(PREF|BCAST)\s+(ALIVE|NO RESP)\s+(NO|YES)\s+(.*)/;
            if ( $4 =~ m/not yet sent/ ) {
            	logDbg("Polling is incomplete");
               	return 0;
            }
            my $type = $1;
            my $state = $2;
            my $is_bound = $3;
	
            $nis_info->{$ip_addr}{Type} = $type;
            $nis_info->{$ip_addr}{State} = $state;
            $nis_info->{$ip_addr}{Bound} = $is_bound;
            logDbg("get_nis_info: Info for $ip_addr added: ".
		   "$type;$state;$is_bound");
            if ( $nis_info->{$ip_addr}{Bound} eq "YES" ) {
                logDbg("Bound server is $ip_addr");
                $nis_info->{active_server} = $ip_addr;
            }
	}


	# look for slave information as well. If found, add to hash
	if ($lines =~ m/NIS Master Server (\S+)/){
	    $nis_info->{master} = $1;
	}
	if ($lines =~ m/NIS Map Last Checked Time (.*)$/){
	    $nis_info->{maps_time} = $1;
	}
    }
    return($nis_info);
}
