#
# $Id: //depot/prod/test/Rsierranevada/lib/Network/NodeTrace.pm#2 $
#

## @summary NodeTrace - a simple class to run cluster side packet tracing
## @author  collinm@netapp.com
## @pod here

=head1 NAME

Network::NodeTrace - a simple class to run pktt cluster side packet tracing

=head1 SYNPOSIS

{

    my $trace = new Network::NodeTrace();

    # do something you want to trace
    
    # optionally you can stop the trace with:
    $trace->end_and_copy();
    # however, if you don't explicitly make
    # this call, it will be called by the
    # DESTROY function when the object instance
    # falls out of scope.
}

=head1 DESCRIPTION

This class will trace network activity to the specified node.  The trace files
will be copied to the LOGDIR after the tracing stops.  If an object reference
falls out of scope, the trace will be stopped gracefully.

=head1 METHODS

=cut

package Network::NodeTrace;

use strict;
use warnings;
use TracePkg;
use Options;
use Sys::Hostname;
use Scalar::Util qw/weaken isweak/;

my $Trace = new TracePkg();
$Trace->set_options(separateNewline => 1);

my $Server;
my @nodes;

# Class data to prevent multiple instances
my $Singleton = undef;

my $Options = new Options();
$Options->parameters(
    definitions => 
    [
    {
        name        => 'LOGDIR',
        description => "The location of the logs",
        default     => undef,
    },
    {
        name        => "LOCAL_TEMP_DIR",
        description => "Local directory on node where packet traces are stored.",
        default     => undef,
    },
    {
        name        => "PKTT_TRACE",
        description => "Enable/disable pktt network tracing",
        default     => 0,
    },
    {
        name        => "PKTT_IF",
        description => "Interface name where the trace will be taken from.",
        default     => "all",
    },
    {
        name        => "PKTT_LEN",
        description => "Length at which packets will be truncated.",
        default     => undef,
    },
    {
        name        => "PKTT_IP",
        description => "Comma separated list of IP addresses to specifically target.",
        default     => undef,
    },
    ]
);

$Options->process();

=head2 new

This is the constructor for the Network::NodeTrace package. The following 
parameters are availalble:

=over

=item PKTT_TRACE

(optional) : Enable/disable node side tracing.

=item PKTT_IF

(optional) : Interface name where the trace will be taken from.

=item PKTT_IP

(optional) : Comma separated list of IP addresses to specifically target.

=item PKTT_LEN

(optional) : Length at which packets will be truncated.

=item LOCAL_TEMP_DIR

(optional) : Local directory on node where packet traces are stored.

=back

=for html <hr>

=cut

sub new {

    $Trace->entry();
    my $class = shift();
    my $self = ();

    # Save off additional command line parameters
    $self->{logdir} = $Options->get(option => [qw(LOGDIR)]);
    $self->{pktt_trace} = $Options->get(option => [qw(PKTT_TRACE)]);
    $self->{pktt_if} = $Options->get(option => [qw(PKTT_IF)]);
    $self->{pktt_ip} = $Options->get(option => [qw(PKTT_IP)]);
    $self->{pktt_len} = $Options->get(option => [qw(PKTT_LEN)]);
    $self->{temp_dir} = $Options->get(option => [qw(LOCAL_TEMP_DIR)]);
    $self = { %$self, @_ }; 

    # Return existing instance if this is not the first instantiation
    if (defined $Singleton) {
        $Trace->comment(msg => "Returning pre-existing NodeTrace instance.");
        $Trace->exit();
        return $Singleton;
    }
    
    # If tracing is disabled, return a reference to an EmptyTrace subclass object.
    if (! $self->{pktt_trace}) {
        $Trace->exit();
        return new Network::EmptyNodeTrace;
    }
    
    # We can't load Server.pm from Windows
    if ($^O =~ /win/i) {
        $Trace->warn(msg => "NodeTrace.pm cannot run on a Windows machine.");
        $Trace->exit();
        return new Network::EmptyNodeTrace;
    } else {
        require Server;
        $Server = new Server();
        @nodes = $Server->get_clustered_nodes();
    }
    
    # We can't save the trace files if there is no LOGDIR
    if (! defined($self->{logdir}) || $self->{logdir} eq '-') {
        $Trace->config(msg => "You must supply a LOGDIR other than '-' where the pktt trace files will be copied.");
    }

    # Bless the class
    bless($self, $class);
    
    # Store a reference to this instance and
    # weaken the reference so that DESTROY will get called
    $Singleton = $self;
    weaken($Singleton) unless isweak($Singleton);

    # Initiate node tracing
    $self->_start_trace();

    # Return (tracing will still be running in the background)
    $Trace->exit();
    return $self;
}

=head2 end_and_copy

This function dumps the trace buffer to disk, stops the trace, and copies 
the trace files to LOGDIR.

=for html <hr>

=cut

sub end_and_copy {
    $Trace->entry();
    my $self = shift();

    $self->_stop_trace();
    $Singleton = undef;

    $Trace->exit();
}

#
# DESTORY
#
# Destructor: Stop the trace when the object falls out of scope.
#
sub DESTROY {
    $Trace->entry();
    my $self = shift();

    $self->_stop_trace();
    $Singleton = undef;

    $Trace->exit();
}

#
# _start_trace
#
# Make sure that tracing is not already running, then start tracing on the nodes.
#
sub _start_trace {
    $Trace->entry();
    my $self = shift();
    
    $self->_stop_pktt();
    $self->_start_pktt();
    
    $Trace->exit();
}

#
# _stop_trace
#
# Dump the tracing to local files on the nodes, stop the tracing, and copy the logs.
#
sub _stop_trace {
    $Trace->entry();
    my $self = shift();
    
    $self->_dump_pktt();
    $self->_stop_pktt();
    $self->_copy_trace();
    
    $Trace->exit();
}

#
# _stop_pktt
#
# Stop PKTT on the nodes.
#
sub _stop_pktt {
    $Trace->entry();
    my $self = shift();
    
    # Build pktt command
    my $pktt_cmd = 'pktt stop '.$self->{pktt_if};
    
    for my $node (@nodes) {
        $Server->set_execution_server(-Server => $node);
        
        my $output = $Server->dblade(cmd => $pktt_cmd);
    }
    
    $Trace->exit();
}

#
# _start_pktt
#
# Start PKTT on the nodes.
#
sub _start_pktt {
    $Trace->entry();
    my $self = shift();
    
    # Build pktt command
    my $pktt_cmd = 'pktt start '.$self->{pktt_if};
    $pktt_cmd .= ' -d '.$self->{temp_dir} if ($self->{temp_dir});
    $pktt_cmd .= ' -m '.$self->{pktt_len} if ($self->{pktt_len});
    
    # Append each IP address, if they are specified
    if ($self->{pktt_ip}) {
        my @ip_list = split(',', $self->{pktt_ip});
        for my $ip (@ip_list) {
            $pktt_cmd .= ' -i '.$ip;
        }
    }
    
    for my $node (@nodes) {
        $Server->set_execution_server(-Server => $node);
        
        my $output = $Server->dblade(cmd => $pktt_cmd);
    }
    
    $Trace->exit();
}

#
# _dump_pktt
#
# Dump the tracing to local files on the nodes.
#
sub _dump_pktt {
    $Trace->entry();
    my $self = shift();
    
    # Build pktt command
    my $pktt_cmd = 'pktt dump '.$self->{pktt_if};
    $pktt_cmd .= ' -d '.$self->{temp_dir} if ($self->{temp_dir});
    
    for my $node (@nodes) {
        $Server->set_execution_server(-Server => $node);
        
        my $output = $Server->dblade(cmd => $pktt_cmd);
    }
    
    $Trace->exit();
}

#
# _copy_trace
#
# Copy the trace files from the nodes to the LOGDIR.
#
sub _copy_trace {
    $Trace->entry();
    my $self = shift();
    
    # Where am I?
    my $hostname = hostname();
    
    # Directory where trace files are stored
    my $trace_dir;
    if (defined ($self->{temp_dir}) ) {
        $trace_dir = $self->{temp_dir};
    }else {
        $trace_dir = '/mroot';
    }
    $trace_dir .= '/';
    
    for my $node (@nodes) {
        $Server->set_execution_server(-Server => $node);
        
        eval {
            # Prefix the trace file name with the node name
            my $rename_cmd = 'cd '.$trace_dir.';ls -1 *.trc| xargs -I TRC mv TRC '.$node.'_TRC';
            my $output = $Server->shell(cmd => $rename_cmd);
            
            # rcp the trace files to the log directory
            my $copy_cmd = 'rcp '.$trace_dir.'*.trc root@'.$hostname.':'.$self->{logdir}.'/';
            $output = $Server->shell(cmd => $copy_cmd);
            
            # Remove the local copy of the trace files
            my $remove_cmd = 'rm '.$trace_dir.'*.trc';
            $Server->shell(cmd => $remove_cmd);
        };
        if ($@) {
            $Trace->warn(msg => $@);
        }
    }
    
    $Trace->exit();
}

#
# Empty package implementing disabled tracing.
#
package Network::EmptyNodeTrace;

sub new ()
{
    $Trace->entry();
    my $class = shift();
    my $self = [];
    bless($self, $class);
    $Trace->trace(msg => "Network node trace disabled.");
    $Trace->exit();
    return $self;
}

sub end_and_copy {}

1;
