# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Pktt.pm#1 $
#
## @summary Pktt - Provide packet trace class
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Pktt - Provide a packet trace class

=head1 SYNOPSIS

  use Network::Pktt;

  # Pktt constructor when using a filer
  $pktt_obj = Network::Pktt->new(
      hostspec => $FILER,
      ifname   => 'e0a',
  );

  # Pktt constructor using a trace file
  # In this form, some methods are invalid like ifname,
  # and others like start and stop are a NOP
  $pktt_obj = Network::Pktt->new(
      file => 'trace_file.trc',
  );

  # Start the packet trace on the filer
  $out = $pktt_obj->start();

  # Stop the packet trace on the filer and move the trace files to the logdir
  $out = $pktt_obj->stop();

  # Find the next packet that matches the input
  # Match the first IP packet either TCP or UDP
  $pkt = $pktt_obj->match(
      'Ethernet:IP:protocol' => [6, 17],
  );

  # Get the next packet from the trace file
  $pkt = $pktt_obj->next_pkt();

  # Get the 10th packet from the trace file
  $pkt = $pktt_obj->get_pkt(pkt_index => 9);

  # Rewind the trace file, by setting the file pointer to the offset
  # of the first packet
  $out = $pktt_obj->rewind();

  # Set the current interface name, it closes the current trace file
  # and opens the trace file for the specified interface
  $ifname = $pktt_obj->ifname(ifname => 'e0b');

=head1 DESCRIPTION

Pktt - Provide a packet trace class.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id: //depot/prod/test/Rsierranevada/lib/Network/Pktt.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Pktt;

use strict;
use warnings;

use Carp qw(croak);
use SayWrap qw(sayEOF);
use GenericMount qw(mount umount);
use Params::Validate qw(validate SCALAR ARRAYREF);
use Net::IP qw(:PROC);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $VERSION = '1.0';
our $AUTOLOAD;

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------
my $root_vol = '/vol/vol0';
my $sudo = '/usr/local/bin/sudo';

# Supported packet trace file formats
my %format = (
    "\324\303\262\241" => {
        bigendian  => 0,
        header_fmt => 'vvVVVV',
        header_rec => 'VVVV',
    },
    "\241\262\303\324" => {
        bigendian  => 1,
        header_fmt => 'nnNNNN',
        header_rec => 'NNNN',
    }
);

#---------------------------------------------------------------------

=head2 new

  $pktt_obj = Network::Pktt->new(hostspec => $hostspec | file => $file
                                 [, ifname => 'all'|$ifname|\@ifname_list ]
                                 [, pklen => $pklen ]
                                 [, logdir => $logdir ]);

Creates a packet trace object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostspec

(Optional) : Hostspec of filer where the trace will be collected.
This option is required when 'file' is not specified.

=item file

(Optional) : File name of trace file on localhost.
This option is required when 'hostspec' is not specified.

=item ifname

(Optional) : Interface name where the trace will be taken from.
This argument takes 'all', the interface name or a list of interface names.
The current interface for the object will be set to this argument if it is
an interface name. If this argument is a list of interfaces, the current
interface will be set to the first interface name specified in the list.
If this argument is set to 'all', the current interface will be set randomly
to a valid interface.
Defaults to 'all'.

=item pklen

(Optional) : Length at which packets will be truncated.
Defaults to 10000.

=item logdir

(Optional) : Directory where the trace files will be moved.
Default is the value given by the $main::TOP_LOGDIR if exists
or 'traces' in the current working directory.

=back

=item Returns

Packet trace object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            hostspec => { type => SCALAR, optional => 1 },
            file     => { type => SCALAR, optional => 1 },
            ifname   => { type => SCALAR|ARRAYREF, default => 'all' },
            pklen    => { type => SCALAR, default => 10000 },
            logdir   => { type => SCALAR,
                       default => (defined $main::TOP_LOGDIR ?
                                   "$main::TOP_LOGDIR/traces" : 'traces') },
        }
    );

    # Check options
    if ( defined $opts{file} && defined $opts{hostspec} ) {
        croak "Options 'file' and 'hostspec' are mutually exclusive";
    } elsif ( !defined $opts{file} && !defined $opts{hostspec} ) {
        croak "Option 'file' or 'hostspec' is required";
    }

    if ( defined $opts{file} ) {
        $opts{logdir} = '';
    }

    # Initialize offset
    $opts{offset} = 0;

    # Initialize packet number
    $opts{pkt_number} = 0;

    # Initialize interface list
    if ( ref $opts{ifname} eq 'ARRAY' ) {
        push @{$opts{_if_list}}, @{$opts{ifname}};
    } else {
        push @{$opts{_if_list}}, $opts{ifname};
    }

    if ( defined $opts{hostspec} ) {
        # Create connectrec object for filer
        Tharn::host($opts{hostspec});
        $opts{_conn} = Tharn::connect('#rsh-socket');
    }

    return bless \%opts, $class;
}

#---------------------------------------------------------------------

=head2 start

  $out = $pktt_obj->start();

Starts the packet trace on the filer.

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub start {
    my $self = shift;
    my $ifname;

    # Ignore this method if the file option was specified
    return 1 if !$self->{logdir};

    # Get current list of trace files on the root directory
    $self->{_trace_files_curr} = { map {$_ => 1} @{$self->_pktt_list} };

    foreach $ifname ( @{$self->{_if_list}} ) {
        $self->{_if_hash}{$ifname} = 1;

        # Stop packet trace on all the interfaces of interest
        $self->_pktt_stop($ifname);

        # Start packet trace on all the interfaces of interest
        $self->_pktt_start($ifname);
    }

    # Flag to make sure this method is called before stop
    $self->{_started} = 1;

    return 1;
}

#---------------------------------------------------------------------

=head2 stop

  $out = $pktt_obj->stop();

Stops the packet trace on the filer, mounts the root volume on localhost
and moves all the trace files created to the logdir.

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub stop {
    my $self = shift;
    my ($pktt_file, $trc_file_orig, $trc_file_new, $ifname, $out);
    my ($mountpoint, $err, @new_traces);

    # Ignore this method if the file option was specified
    return 1 if !$self->{logdir};

    # Check that trace was initialized
    if ( !defined $self->{_started} ) {
        croak "Must initialize trace using the 'start' method";
    }

    # Initialize trace files hash
    $self->{_trace_files} = {};

    # Initialize offset
    $self->{offset} = 0;

    # Initialize packet number
    $self->{pkt_number} = 0;

    # Shut down pktt tracing and generate reports
    foreach $ifname ( @{$self->{_if_list}} ) {
        # Stop packet trace
        $self->_pktt_stop($ifname);
    }

    # Get list of trace files
    $out = $self->_pktt_list;

    # Get list of all new trace files
    # (trace files which did not exist before the 'pktt start')
    @new_traces = grep { !exists $self->{_trace_files_curr}{$_} } @$out;

    # Get interface names for all new trace files
    foreach $trc_file_new (@new_traces) {
        ($ifname) = ($trc_file_new =~ /^([^_]+)_.+\.trc$/);
        # Ignore trace files which are not properly formated
        next if !defined $ifname;
        if ($self->{_if_hash}{all} || $self->{_if_hash}{$ifname}) {
            # Add this trace file from an expected interface
            croak "Could not determine trace file for interface '$ifname': " .
                  "more than one trace file was found for interface " .
                  "so packet trace must have been started on the same " .
                  "interface from a different process"
                  if exists $self->{_trace_files}{$ifname};
            $self->{_trace_files}{$ifname} = $trc_file_new;
        }
    }

    # Mount the root volume on the local client
    $mountpoint = mount($self->{_conn}->connhost, $root_vol, mntopts => 'noac')
      or croak "Could not mount root volume: $@";

    # Catch any errors to unmount the root volume
    eval {
        # Create the traces directory in the test log directory
        if ( !-e $self->{logdir} ) {
            mkdir $self->{logdir}
              or croak "Could not create traces log directory " .
                       "'$self->{logdir}': $!";
        } elsif ( !-d $self->{logdir} ) {
            croak "Traces log directory '$self->{logdir}' exists as a file " .
                  "not a directory";
        }

        foreach $ifname ( keys %{$self->{_trace_files}} ) {
            $pktt_file = $self->{_trace_files}{$ifname};
            # Change permissions on any new files and
            # move them into the trace directory
            $trc_file_orig = "$mountpoint/$pktt_file";
            $trc_file_new = "$self->{logdir}/$pktt_file";
            $out = qx($sudo chmod 444 $trc_file_orig 2>&1);
            croak $out if $?;
            $out = qx($sudo mv $trc_file_orig $trc_file_new 2>&1);
            croak $out if $?;
        }
    };
    $err = $@;

    # Unmount the volume
    umount($mountpoint);
    croak $err if $err;
    croak "Could not unmount root volume: $@" if $@;

    # Check if there is at least one trace file
    if ( ! %{$self->{_trace_files}} ) {
        croak "No trace files were found";
    }

    # Set the current trace file
    if ( ref $self->{ifname} eq '' && $self->{ifname} ne 'all' ) {
        # The ifname parameter to the constructor is an interface name
        # the current interface is already set correctly
        $self->{file} = $self->{_trace_files}{$self->{ifname}};
    } else {
        if ( ref $self->{ifname} ) {
            # The ifname parameter to the constructor is a list or interfaces
            # set the current interface to the first interface name in the list
            $self->{ifname} = $self->{ifname}[0];
        } else {
            # The ifname parameter to the constructor is 'all'
            # just select an interface at random
            $self->{ifname} = (keys %{$self->{_trace_files}})[0];
        }
        $self->{file} = $self->{_trace_files}{$self->{ifname}};
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 match

  $pkt = $pktt_obj->match(%match);

Find the next packet that matches the input hash.
The following examples are equivalent:

  $pkt = $pktt_obj->match(
      'Ethernet:type' => 2048,  # IP protocol
      'Ethernet:IP:protocol' => [6, 17],  # Match for TCP or UDP
      'Ethernet:IP:src' => '172.17.153.181',
      'Ethernet:IP:UDP:dst_port' => 137,
      'Ethernet:IP:UDP' => {
          checksum   => 15755,
          src_port   => 55270,
          total_size => 76,
      },
  );

  $pkt = $pktt_obj->match(
      Ethernet => {
          type => 2048,  # IP protocol
          IP => {
              protocol => [6, 17],  # Match for TCP or UDP
              src => '172.17.153.181',
              UDP => {
                  dst_port   => 137,
                  checksum   => 15755,
                  src_port   => 55270,
                  total_size => 76,
              },
          },
      },
  );

=over

=item Returns

Packet hash reference if there is a match, else undef.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub match {
    my $self = shift;
    # Not using validate, the input hash could be anything
    my %opts = @_;
    my (@list, $pkt, $ptr, $key, $found);
    my $pkt_number = $self->{pkt_number};

    # Convert input hash to a standard hash
    my $match_h = _process_input_hash(\%opts);

    # Get one packet at a time starting a current position
    while ($pkt = $self->next_pkt) {
        # Return packet if it matches
        return $pkt if _compare_hashes($match_h, $pkt);
    }

    # No packet matched, so position the packet number and offset
    # to what it was before the search
    seek($self->{fh}, $self->{pkt_map}[$pkt_number], 0)
      or croak "Could not position file pointer: $!";

    # Reset the current packet number and offset
    $self->{pkt_number} = $pkt_number;
    $self->{offset} = $self->{pkt_map}[$pkt_number];

    return;
}

#---------------------------------------------------------------------

=head2 next_pkt

  $pkt = $pktt_obj->next_pkt();

Get the next packet from the trace file.

=over

=item Returns

Packet hash reference or undef if there is at the end of the trace file.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub next_pkt {
    my $self = shift;
    my (%record, $header, $data, $count);

    # Save file offset for this packet
    $self->{pkt_map}[$self->{pkt_number}] = $self->{offset};

    # Get record header
    $count = $self->_read(\$header, 16);
    return if $count < 16;

    # Unpack record header
    @record{qw(seconds msecs length_inc length_orig)} =
      unpack($format{$self->{ident}}{header_rec}, $header);

    # Get the data
    $count = $self->_read(\$data, $record{length_inc});

    # Calculate number of drops
    $record{drops} = $record{length_orig} - $record{length_inc};

    if ( $self->{tcpdump_header}{link_type} == 1 ) {
        # Decode ethernet layer
        $record{Ethernet} = $self->_decode_ethernet_layer(\$data);
    } else {
        # Unknown link layer
        $record{data} = $data;
    }

    # Increment packet number
    $record{pkt_number} = $self->{pkt_number}++;

    return \%record;
}

#---------------------------------------------------------------------

=head2 get_pkt

  $pkt = $pktt_obj->get_pkt(pkt_index => $pkt_index);

Get the packet from the trace file given by the index.

The following named arguments are accepted:

=over

=item Arguments

=over

=item pkt_index

(Required) : Index of packet to get.

=back

=item Returns

Packet hash reference or undef if the index is out of bounds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_pkt {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            pkt_index => { type => SCALAR },
        }
    );
    my $pkt;
    my $index = $opts{pkt_index};

    if (defined $self->{pkt_map} && $index < @{$self->{pkt_map}}) {
        # The index is less than the current packet offset so position
        # the file pointer to the offset of packet given by index
        seek($self->{fh}, $self->{pkt_map}[$index], 0)
          or croak "Could not position file pointer: $!";

        # Reset the current packet number and offset
        $self->{pkt_number} = $index;
        $self->{offset} = $self->{pkt_map}[$index];
    }

    # Get one packet at a time until the packet given by the index is found
    do {
        $pkt = $self->next_pkt;
    } while (defined $pkt && $pkt->{pkt_number} < $index);

    return $pkt;
}

#---------------------------------------------------------------------

=head2 rewind

  $out = $pktt_obj->rewind();

Rewind the trace file. It sets the file pointer to the offset
of the first packet.

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub rewind {
    my $self = shift;

    # Initialize filehandle
    $self->_get_filehandle;

    # Position the file pointer to the offset of the first packet
    seek($self->{fh}, $self->{pkt_map}[0], 0)
      or croak "Could not position file pointer: $!";

    # Reset the current packet number and offset
    $self->{pkt_number} = 0;
    $self->{offset} = $self->{pkt_map}[0];

    return 1;
}

#---------------------------------------------------------------------

=head2 ifname

  $ifname = $pktt_obj->ifname([ ifname => $ifname ]);

Get or set the current interface name.
When setting the current interface name, it closes the current trace file
and opens the trace file for the specified interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ifname

(Optional) : Name to set the current interface.

=back

=item Returns

The current interface name.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ifname {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            ifname => { type => SCALAR, optional => 1 },
        }
    );

    if ( defined $opts{ifname} ) {
        croak "Must stop trace using the 'stop' method"
          if !defined $self->{_trace_files};

        # Get the trace file for given interface
        $self->{file} = $self->{_trace_files}{$opts{ifname}};
        croak "No trace file found for interface $opts{ifname}"
          if !defined $self->{file};

        # Set the interface name
        $self->{ifname} = $opts{ifname};

        # Close the current filehandle
        if (defined $self->{fh}) {
            close($self->{fh});
            undef $self->{fh};
        }

        # Clear packet map
        $self->{pkt_map} = [];

        # Initialize filehandle
        $self->_get_filehandle;
    }

    return $self->{ifname};
}

#---------------------------------------------------------------------
# Internal method
# Decodes the ethernet layer
#---------------------------------------------------------------------
sub _decode_ethernet_layer {
    my ($self, $data) = @_;
    my (%ret, $header, $tci, $type, $size, $delta);
    my $decode = 1;
    my $header_size = 14;

    # Get total size of packet
    $ret{packet_size} = length $$data;

    # Decode ethernet layer (RFC 894) Ethernet II
    $header = substr($$data, 0, 14, '');
    @ret{qw(dst src type)} = unpack('H12H12n', $header);

    # Add the ':' to the MAC addresses
    $ret{src} = join(':', $ret{src} =~ /(..)/g);
    $ret{dst} = join(':', $ret{dst} =~ /(..)/g);

    if ($ret{type} <= 1500) {
        # IEEE 802.2/802.3 Encapsulation (RFC 1042)
        $ret{size} = $ret{type};
        $header = substr($$data, 0, 8, '');
        @ret{qw(DSAP SSAP cntl org_code type)} = unpack('CCCa3n', $header);
        $ret{org_code} = unpack('N', "\000$ret{org_code}");
    }

    if ( $ret{type} == 0x8100 ) {
        # 802.1Q Virtual LAN packet
        $header_size += 4;
        $header = substr($$data, 0, 4, '');
        ($tci, $type) = unpack('nn', $header);
        $ret{VLAN} = {
            priority => ($tci >> 13),
            CFI      => ($tci >> 12) & 0x1,
            VID      => ($tci & 0xFFF),
        };
        if ( $type <= 1500 ) {
            $ret{VLAN}{size} = $type;
            $ret{VLAN}{trailer} = substr($$data, $type, 4, '');
            $ret{VLAN}{data} = $$data;
            substr($$data, 0, $type+4, '');
        } else {
            $ret{type} = $type;
        }
    }

    if ( $ret{type} == 0x0800 ) {
        # Type of ethernet packet is IP (Internet Protocol)
        $ret{IP} = $self->_decode_ip_layer($data);
    } elsif ( $ret{type} == 0x86dd ) {
        # Type of ethernet packet is IPv6 (Internet Protocol version 6)
        $ret{IPv6} = $self->_decode_ipv6_layer($data);
    } elsif ( $ret{type} == 0x0806 ) {
        # Type of ethernet packet is ARP (Address Resolution Protocol)
        $ret{ARP} = $self->_decode_arp_layer($data);
    } elsif ( $ret{type} == 0x8035 ) {
        # Type of ethernet packet is RARP (Reverse Address Resolution Protocol)
        $ret{RARP} = $self->_decode_arp_layer($data);
    } else {
        # Type does not have any decoding method so just return the data
        $ret{data} = $$data;
        $decode = 0;
    }

    # Get actual packet size
    if ( $decode && ($size = length $$data) ) {
        $ret{packet_size} -= $size;
        if ( $ret{packet_size} < 46 + $header_size ) {
            # This packet is < 46 bytes, get the padding
            $delta = 46 + $header_size - $ret{packet_size};
            $ret{trailer} = substr($$data, 0, $delta, '');
            $size -= $delta;
            $ret{packet_size} += $delta;
        }
        # Get the CRC if any
        $ret{CRC} = unpack('N', $$data) if $size >= 4;
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the IP layer
#---------------------------------------------------------------------
sub _decode_ip_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header, $count, $temp, $ip_data);

    # Get the IP version and header length
    $temp = unpack('C', $$data);
    # Header length in bytes
    $count = 4*($temp & 0x0F);
    # Get full IP header
    $header = substr($$data, 0, $count, '');
    # Decode IP header
    @list = unpack('CCnnnCCna4a4', $header);

    %ret = (
        version         => ($temp >> 4),
        IHL             => ($temp & 0x0F),
        header_size     => $count,
        total_size      => $list[2],
        id              => $list[3],
        fragment_offset => ($list[4] & 0x1FFF),
        TTL             => $list[5],
        protocol        => $list[6],
        checksum        => $list[7],
        src             => join('.', unpack('CCCC', $list[8])),
        dst             => join('.', unpack('CCCC', $list[9])),
    );

    # Include both TOS and DSCP/ECN from the same byte
    $temp = $list[1];
    # Add TOS (Type of Service) from RFC 1349
    $ret{TOS} = {
        precedence    => ($temp >> 5),
        delay         => (($temp >> 4) & 0x01),
        throughput    => (($temp >> 3) & 0x01),
        reliability   => (($temp >> 2) & 0x01),
        monetary_cost => (($temp >> 1) & 0x01),
    };

    # Add DSCP (Differentiated Services Code Point)
    # and ECN (Explicit Congestion Notification) from RFC 3168
    $ret{DSCP} = ($temp >> 2);
    $ret{ECN}  = ($temp & 0x03);

    $temp = ($list[4] >> 13);
    $ret{flags} = {
        DF => (($temp >> 1) & 0x01),
        MF => ($temp & 0x01),
    };

    if ( $count > 20 ) {
        # Save IP header options
        $ret{options} = substr($header, 20);
    }

    # Get full data for IP packet
    $ip_data = substr($$data, 0, $ret{total_size} - $ret{header_size}, '');

    if ( $ret{protocol} == 6 ) {
        # Type of IP packet is TCP (Transmission Control Protocol)
        $ret{TCP} = $self->_decode_tcp_layer(\$ip_data);
    } elsif ( $ret{protocol} == 17 ) {
        # Type of IP packet is UDP (User Datagram Protocol)
        $ret{UDP} = $self->_decode_udp_layer(\$ip_data);
    } elsif ( $ret{protocol} == 1 ) {
        # Type of IP packet is ICMP (Internet Control Message Protocol)
        $ret{ICMP} = $self->_decode_icmp_layer(\$ip_data);
    } elsif ( $ret{protocol} == 2 ) {
        # Type of IP packet is IGMP (Internet Group Management Protocol)
        $ret{IGMP} = $self->_decode_igmp_layer(\$ip_data);
    } else {
        # Type does not have any decoding method so just return the data
        $ret{data} = $ip_data;
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the IPv6 layer
#---------------------------------------------------------------------
sub _decode_ipv6_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header, $temp, $ip_data, $src, $dst, @groups);
    
    # Get full IPv6 header of fix size of 40 bytes
    $header = substr($$data, 0, 40, '');
    
    # Decode IPv6 header
    @list = unpack('HChCnCCH32H32', $header);
    
    foreach ($list[7],$list[8]) {
        @groups = unpack "a4" x 8, $_;
        $_ = join(':', @groups);
        $_ = Net::IP::ip_compress_address($_,6);
    }

    %ret = (
        version => $list[0],
        traffic_class => $list[1],
        flow_lable => ($list[2] << 16) | $list[3],
        payload_length => $list[4],
        next_header => $list[5],
        hop_limit => $list[6],
        src => $list[7],
        dst => $list[8],
    );
    
    # Get full data for IP packet
    $ip_data = $$data;
   
    if ( $ret{next_header} == 6 ) {
        # Type of IPv6 packet is TCP (Transmission Control Protocol)
        $ret{TCP} = $self->_decode_tcp_layer(\$ip_data);
    } elsif ( $ret{next_header} == 17 ) {
        # Type of IP packet is UDP (User Datagram Protocol)
        $ret{UDP} = $self->_decode_udp_layer(\$ip_data);
    } elsif ( $ret{next_header} == 1 ) {
        # Type of IP packet is ICMPv6 (Internet Control Message Protocol version 6)
        $ret{ICMP} = $self->_decode_icmp_layer(\$ip_data);
    } elsif ( $ret{next_header} == 58 ) {
        # Type of IP packet is ICMPv6 (Internet Control Message Protocol version 6)
        $ret{ICMPv6} = $self->_decode_icmpv6_layer(\$ip_data);
    } elsif ( $ret{next_header} == 2 ) {
        # Type of IP packet is IGMP (Internet Group Management Protocol)
        $ret{IGMP} = $self->_decode_igmp_layer(\$ip_data);
    } else {
        $ret{data} = $ip_data;
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the TCP layer
#---------------------------------------------------------------------
sub _decode_tcp_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header, $count, $rpc, $temp);

    # Decode the TCP layer header
    (undef, $temp) = unpack('a12C', $$data);
    $temp >>= 4;
    $count = 4*$temp;
    $header = substr($$data, 0, $count, '');
    @list = unpack('nnNNnnnn', $header);

    %ret = (
        src_port    => $list[0],
        dst_port    => $list[1],
        seq_number  => $list[2],
        ack_number  => $list[3],
        window_size => $list[5],
        checksum    => $list[6],
        urgent_ptr  => $list[7],
        hl          => $temp,
        header_size => $count,
        flags => {
            FIN => ($list[4] & 0x01),
            SYN => (($list[4] >> 1) & 0x01),
            RST => (($list[4] >> 2) & 0x01),
            PSH => (($list[4] >> 3) & 0x01),
            ACK => (($list[4] >> 4) & 0x01),
            URG => (($list[4] >> 5) & 0x01),
        },
    );

    if ( $rpc = $self->_decode_rpc_layer($data, 6) ) {
        $ret{RPC} = $rpc;
    } else {
        $ret{data} = $$data;
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the UDP layer
#---------------------------------------------------------------------
sub _decode_udp_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header, $out);

    # Decode the UDP layer header
    $header = substr($$data, 0, 8, '');
    @list = unpack('nnnn', $header);

    %ret = (
        src_port   => $list[0],
        dst_port   => $list[1],
        total_size => $list[2],
        checksum   => $list[3],
    );

    if ( $ret{src_port} == 520 && $ret{dst_port} == 520 ) {
        $out = $self->_decode_rip_layer($data);
        $ret{RIP} = $out if defined $out;
    } elsif ( $out = $self->_decode_rpc_layer($data, 17) ) {
        $ret{RPC} = $out;
    }

    if ( length $$data ) {
        # Add the raw data if unknown format
        $ret{data} = $$data;
    }

    return \%ret;
}
#---------------------------------------------------------------------
# Internal method
# Decodes the ICMPv6 layer
#---------------------------------------------------------------------
sub _decode_icmpv6_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header);

    $header = substr($$data, 0, 4, '');
    @list = unpack('CCn', $header);

    %ret = (
        type     => $list[0],
        code     => $list[1],
        checksum => $list[2],
        data     => $$data,
    );

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the ICMP layer
#---------------------------------------------------------------------
sub _decode_icmp_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header);

    $header = substr($$data, 0, 4, '');
    @list = unpack('CCn', $header);

    %ret = (
        type     => $list[0],
        code     => $list[1],
        checksum => $list[2],
        data     => $$data,
    );

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the IGMP layer
#---------------------------------------------------------------------
sub _decode_igmp_layer {
    my ($self, $data) = @_;
    my (%ret, @list, $header, %data);

    $header = substr($$data, 0, 8, '');
    @list = unpack('CCna4', $header);

    if ( $list[0] >= 0x11 && $list[0] <= 0x13 ) {
        # Version 1
        %data = (
            version => 1,
            type    => ($list[0] >> 4),
        );
    } else {
        # Version 2
        %data = (
            version      => 2,
            type         => $list[0],
            max_response => $list[1],
        );
    }

    %ret = (
        %data,
        checksum => $list[2],
        address  => join('.', unpack('CCCC', $list[3])),
    );

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the ARP/RARP layer
#---------------------------------------------------------------------
sub _decode_arp_layer {
    my ($self, $data) = @_;
    my (%ret, $header, $uformat);

    # Decode ARP/RARP header
    $header = substr($$data, 0, 8, '');
    @ret{qw(hardware_type protocol_type hardware_size protocol_size operation)} =
      unpack('nnCCn', $header);

    # Get the right format for the hardware and protocol addresses
    $uformat = ('H' . (2*$ret{hardware_size}) . 'a' . $ret{protocol_size}) x 2;

    # Decode the hardware and protocol addresses
    @ret{qw(sender_ether_addr sender_proto_addr target_ether_addr target_proto_addr)} =
      unpack($uformat, $$data);

    if ( $ret{hardware_type} == 1 ) {
        # Add the ':' to the MAC addresses
        $ret{sender_ether_addr} = join(':', $ret{sender_ether_addr} =~ /(..)/g);
        $ret{target_ether_addr} = join(':', $ret{target_ether_addr} =~ /(..)/g);
    }

    if ( $ret{protocol_type} == 0x0800 ) {
        # Make the IP protocol addresses human readable
        $ret{sender_proto_addr} = join('.', unpack('CCCC', $ret{sender_proto_addr}));
        $ret{target_proto_addr} = join('.', unpack('CCCC', $ret{target_proto_addr}));
    }

    # Remove full data for ARP/RARP packet from the data
    substr($$data, 0, 8 + 2*($ret{hardware_size} + $ret{protocol_size}), '');

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes the RPC call and reply messages
#---------------------------------------------------------------------
sub _decode_rpc_layer {
    my ($self, $data, $proto) = @_;
    my (%ret, $buffer, $psize);
    my $size = length $$data;
    my $offset = 0;

    if ( $proto == 6 ) {
        # TCP packet
        return if $size < 12;

        # Decode fragment header
        $buffer = substr($$data, 0, 4);
        $psize = unpack('N', $buffer);
        $ret{fragment_hdr}{size} = $psize & 0x7FFFFFFF;
        $ret{fragment_hdr}{last_fragment} = $psize >> 31;
        $offset += 4;
    } else {
        # UDP packet
        return if $size < 8;
    }

    # Decode the XID and RPC type
    $buffer = substr($$data, $offset, 8);
    @ret{qw(xid type)} = unpack('NN', $buffer);
    $offset += 8;

    if ( $ret{type} == 0 ) {
        # RCP call
        return if $size < $offset + 32;
        $buffer = substr($$data, $offset, 16);
        $offset += 16;
        @ret{qw(rpc_version program version procedure)} = unpack('NNNN', $buffer);
        $ret{credential} = _rpc_credential($data, \$offset, $size);
        return if !$ret{credential};
        $ret{verifier} = _rpc_credential($data, \$offset, $size);
        return if !$ret{verifier};
    } elsif ( $ret{type} == 1 ) {
        # RCP reply
        return if $size < $offset + 12;
        $buffer = substr($$data, $offset, 4);
        $offset += 4;
        $ret{reply_status} = unpack('N', $buffer);
        if ( $ret{reply_status} ) {
            $buffer = substr($$data, $offset, 4);
            $offset += 4;
            $ret{rejected_status} = unpack('N', $buffer);
            if ( $ret{rejected_status} ) {
                $buffer = substr($$data, $offset, 4);
                $offset += 4;
                $ret{auth_status} = unpack('N', $buffer);
            } else {
                return if $size < $offset + 8;
                $buffer = substr($$data, $offset, 8);
                $offset += 8;
                @{$ret{rpc_mismatch}}{qw(low high)} = unpack('NN', $buffer);
            }
        } else {
            $ret{verifier} = _rpc_credential($data, \$offset, $size);
            return if !$ret{verifier};
            return if $size < $offset + 4;
            $buffer = substr($$data, $offset, 4);
            $offset += 4;
            $ret{accepted_status} = unpack('N', $buffer);
            if ( $ret{accepted_status} == 2 ) {
                return if $size < $offset + 8;
                $buffer = substr($$data, $offset, 8);
                $offset += 8;
                @{$ret{prog_mismatch}}{qw(low high)} = unpack('NN', $buffer);
            }
        }
    } else {
        # Not an RPC packet
        return;
    }

    # Remove the RPC header
    substr($$data, 0, $offset, '');
    $ret{data} = $$data;

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decodes RIP (Routing Information Protocol)
#---------------------------------------------------------------------
sub _decode_rip_layer {
    my ($self, $data) = @_;
    my (%ret, $header, @list);
    my $offset = 0;

    # Decode RIP header
    $header = substr($$data, $offset, 4);
    $offset += 4;
    @list = unpack('CCn', $header);

    return if ($list[0] == 1 && $list[2]);
    %ret = (
        command => $list[0],
        version => $list[1],
    );

    if ( $ret{version} == 2 ) {
        $ret{domain} = $list[2];
    }

    while ( length $$data >= $offset + 20 ) {
        $header = substr($$data, $offset, 20);
        $offset += 20;
        if ( $ret{version} == 1 ) {
            # RIP-v1
            @list = unpack('nna4NNN', $header);
            # Check correct format
            if ( $list[1] || $list[3] || $list[4]
              || $list[5] < 1 || $list[5] > 16 ) {
                return;
            }
            if ( $list[0] == 2 ) {
                # Address family identifier is for an IP address
                $list[2] = join('.', unpack('CCCC', $list[2]));
            }
            push @{$ret{routes}}, {
                AFI     => $list[0],
                address => $list[2],
                metric  => $list[5],
            };
        } elsif ( $ret{version} == 2 ) {
            # RIP-v2
            @list = unpack('nna4a4a4N', $header);
            if ( $list[0] == 2 ) {
                # Address family identifier is for IP addresses
                $list[2] = join('.', unpack('CCCC', $list[2]));
                $list[3] = join('.', unpack('CCCC', $list[3]));
                $list[4] = join('.', unpack('CCCC', $list[4]));
            }
            push @{$ret{routes}}, {
                AFI      => $list[0],
                tag      => $list[1],
                address  => $list[2],
                subnet   => $list[3],
                next_hop => $list[4],
                metric   => $list[5],
            };
        }
    }

    # Check for invalid extra data
    return if length $$data > $offset;

    # There is no more data
    $$data = '';

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Decode the RPC credentials
#---------------------------------------------------------------------
sub _rpc_credential {
    my ($data, $offset, $size) = @_;
    my (%ret, $header);

    # Get the flavor and size of credential
    $header = substr($$data, $$offset, 8);
    $$offset += 8;
    @ret{qw(flavor size)} = unpack('NN', $header);

    # Check for correct size
    return if $size < $$offset + $ret{size};

    # Just get the credential data
    $ret{data} = substr($$data, $$offset, $ret{size});
    $$offset += $ret{size};

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Get filehandle of the current trace file
#---------------------------------------------------------------------
sub _get_filehandle {
    my $self = shift;
    my ($file, $ident, $header, $count);

    if (! defined $self->{fh}) {
        # Open trace file
        $file = $self->{logdir} ? "$self->{logdir}/$self->{file}" : $self->{file};
        open($self->{fh}, "<$file") or croak "Could not open trace file: $!";
        binmode($self->{fh});

        # Initialize offset
        $self->{offset} = 0;

        # Get file type identifier
        $count = $self->_read(\$ident, 4);
        croak "Not a tcpdump file" if !exists $format{$ident};

        # Get header information
        $count = $self->_read(\$header, 20);
        croak "Not a tcpdump file" if $count < 20;

        # Unpack header
        @{$self->{tcpdump_header}}{qw(major minor zone_offset accuracy dump_length link_type)} =
          unpack($format{$ident}{header_fmt}, $header);

        # Initialize packet number
        $self->{pkt_number} = 0;
        $self->{ident} = $ident;
        $self->{pkt_map}[0] = $self->{offset};
    }

    return $self->{fh};
}

#---------------------------------------------------------------------
# Internal method
# Wrapper for the system call read() to update the internal offset
#---------------------------------------------------------------------
sub _read {
    my ($self, $buffer, $count) = @_;

    my $out = read($self->_get_filehandle, $$buffer, $count);
    croak "Could not read trace file: $!" if !defined $out;
    $self->{offset} += $out;

    return $out;
}

#---------------------------------------------------------------------
# Internal function
# Compare two hashes -- returns 1 if h1 is a subset of h2
# The h1 hash could have an array reference as the value so it
# could match any of the values in the array
#---------------------------------------------------------------------
sub _compare_hashes {
    my ($h1, $h2) = @_;
    my ($key, $item, $list, $found);

    # Check each key in the h1 hash
    foreach $key ( keys %$h1 ) {
        # Compare fails if the key does not exists on the h2 hash
        return 0 if !exists $h2->{$key};

        if ( ref $h1->{$key} eq 'HASH' && ref $h2->{$key} eq 'HASH' ) {
            # If both values are hash references, call this function recursively
            return 0 if !_compare_hashes($h1->{$key}, $h2->{$key});
        } elsif ( ref $h2->{$key}
          || (ref $h1->{$key} && ref $h1->{$key} ne 'Regexp' && ref $h1->{$key} ne 'ARRAY') ) {
            # Compare fails if value is a reference except for valid values
            # on h1 that could be a compile regular expression or an array
            # reference for OR style compare
            return 0;
        } else {
            # Get list of values to match
            if ( ref $h1->{$key} eq 'ARRAY' ) {
                # This is already an array
                $list = $h1->{$key};
            } else {
                # Make an array with single value
                $list = [ $h1->{$key} ];
            }
            # Compare all values in list with the value on h2
            $found = 0;
            foreach $item ( @$list ) {
                if ( ref $item eq 'Regexp' ) {
                    $found = 1 if $h2->{$key} =~ /$item/;
                } elsif ( $item eq $h2->{$key} ) {
                    $found = 1;
                }
                last if $found;
            }
            return 0 if !$found;
        }
    }

    # Everything matched
    return 1;
}

#---------------------------------------------------------------------
# Internal function
# Convert input hash to a standard hash, e.g.:
#   %input = (
#       'Ethernet:type' => 2048,
#       'Ethernet:IP:protocol' => [6, 17],
#       'Ethernet:IP:src' => '172.17.153.181',
#       'Ethernet:IP:UDP:dst_port' => 137,
#       'Ethernet:IP:UDP' => {
#           checksum   => 15755,
#           src_port   => 55270,
#           total_size => 76,
#       },
#   );
#   %output = (
#       Ethernet => {
#           type => 2048,
#           IP => {
#               protocol => [6, 17],
#               src => '172.17.153.181',
#               UDP => {
#                   dst_port   => 137,
#                   checksum   => 15755,
#                   src_port   => 55270,
#                   total_size => 76,
#               },
#           },
#       },
#   );
#---------------------------------------------------------------------
sub _process_input_hash {
    my $hash = shift;
    my (%ret, @list, $key, $item, $ptr, $last_item, $ref_k, $ref_p);

    foreach $key ( keys %$hash ) {
        # Split input key
        @list = split /:/, $key;
        # Initialize pointer to top level of hash
        $ptr = \%ret;
        # Get the key for the last leg
        $last_item = pop @list;
        foreach $item ( @list ) {
            # Initialize hash for each leg of the input key
            $ptr->{$item} = {} if !exists $ptr->{$item};
            # Set the pointer to each level
            $ptr = $ptr->{$item};
        }
        if ( ref $ptr ne 'HASH' ) {
            croak "Bad match key for '$key'";
        } elsif ( exists $ptr->{$last_item} ) {
            # The last leg already exists, so check for compatibility
            $ref_k = ref $hash->{$key};
            $ref_p = ref $ptr->{$last_item};
            if ( $ref_k ne $ref_p ) {
                croak "Bad match key for '$key'";
            } elsif ( $ref_k eq 'HASH' ) {
                # Combine both hashes
                $ptr->{$last_item} = { %{$ptr->{$last_item}}, %{$hash->{$key}} };
            } elsif ( $ref_k eq 'ARRAY' ) {
                # Combine both arrays
                $ptr->{$last_item} = [ @{$ptr->{$last_item}}, @{$hash->{$key}} ];
            } else {
                # The reference type is not supported
                croak "Unknown reference $ref_k";
            }
        } else {
            # Set the value for this key
            $ptr->{$last_item} = $hash->{$key};
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal function
# Start packet trace on the given interface
#---------------------------------------------------------------------
sub _pktt_start {
    my ($self, $ifname) = @_;
    my $out;

    # Send command to filer
    $out = sayEOF($self->{_conn}, "pktt start $ifname -m $self->{pklen} -d /");

    if ( $out !~ /started packet trace/ ) {
        croak "Could not start packet trace for interface '$ifname'";
    }

    return 1;
}

#---------------------------------------------------------------------
# Internal function
# Stop packet trace on the given interface
#---------------------------------------------------------------------
sub _pktt_stop {
    my ($self, $ifname) = @_;
    my $out;

    # Send command to filer
    $out = sayEOF($self->{_conn}, "pktt stop $ifname");

    if ( $out !~ /(Tracing stopped|pktt not active)/ ) {
        croak "Could not stop packet trace for interface '$ifname'";
    }

    return 1;
}

#---------------------------------------------------------------------
# Internal function
# Pause packet trace on the given interface
#---------------------------------------------------------------------
sub _pktt_pause {
    my ($self, $ifname) = @_;
    my $out;

    # Send command to filer
    $out = sayEOF($self->{_conn}, "pktt pause $ifname");

    if ( $out !~ /Tracing suspended/ ) {
        croak "Could not pause packet trace for interface '$ifname'";
    }

    return 1;
}

#---------------------------------------------------------------------
# Internal function
# Dump packet trace for the given interface to a file, it returns
# a hash of interface names and their respective trace dump files
#---------------------------------------------------------------------
sub _pktt_dump {
    my ($self, $ifname) = @_;
    my ($out, %ret);

    # Send command to filer
    $out = sayEOF($self->{_conn}, "pktt dump $ifname");

    # Get file names for each interface
    %ret = ($out =~ m{\(if ([^\)]+)\) already writing to file /(\S+)}g);

    if ( !%ret ) {
        croak "Could not dump packet trace for interface '$ifname'";
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal function
# Get list of all trace files in the root directory, it returns
# a hash of interface names and their respective trace dump files
#---------------------------------------------------------------------
sub _pktt_list {
    my $self = shift;
    my ($out, @ret);

    # Send command to filer
    $out = sayEOF($self->{_conn}, "pktt list");

    # Get all trace files
    @ret = split ' ', $out;

    return \@ret;;
}

#---------------------------------------------------------------------
# AUTOLOAD to exposed all keys in the blessed object as read-only methods
# Do not expose private keys (the ones which start with a '_')
#---------------------------------------------------------------------
sub AUTOLOAD {
    my $self = shift;
    my $method = $AUTOLOAD;
    $method =~ s/.*:://;
    if ( $method =~ /^_/ || !exists $self->{$method} ) {
        croak qq(Can't locate object method "$method" via package "Network::Pktt");
    }
    return $self->{$method};
}

#---------------------------------------------------------------------
# Destructor
# Close fh and unmount root volume
#---------------------------------------------------------------------
sub DESTROY {
    my $self = shift;

    # Close trace file
    if (defined $self->{fh}) {
        close($self->{fh});
        undef $self->{fh};
    }
}

1;

