# $Id: $

## @summary Port_Utils - library to implement utilities for Network Ports
## @author dl-nwkg-test@netapp.com
## @pod here


=head1 NAME
SCON_Utils - library to implement utilities for SCON testing

=head1 SYNOPSIS

    use Network::Port_Utils qw(get_port_list);

=head1 DESCRIPTION

    This library has a variety of methods to deal with network ports.

=head1 ROUTINES

=cut

package Network::Port_Utils;

use strict;
use warnings;
use Tharn;

use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF BOOLEAN);
use base qw(Exporter);
use Data::Dumper;
our @EXPORT_OK = qw(get_port_list, revive_port);
use NATE::BaseException qw(:try);
use POSIX qw(strftime);

#############################################################################
#                                NACL STUFF
#############################################################################
use NACL::C::Node;
use NACL::C::Interface;
use NACL::C::NetworkPort;

#############################################################################
#                           FEATURE SPECIFIC LIBS
#############################################################################
use Network::Failover_Utils;

my $Node = NACL::C::Node->find();

=head2 get_port_list

    my $ret = get_port_list(node_name => "my_filer_1");

This routine returns a arrayref of ports for the given arguments

It throws exceptions on error.

The following named arguments are accepted,

=over

=item node_name

(Required) : Name of the node where ports reside

=item role

(Optional) : Role of the port (default data)

=item link

(Optional) : Link state of port (default 'up')

=item type

(Optional) : Type of port (default 'physical')

=back 

=for html <hr>

=cut
############################################################################
# Subroutine: get_port_list
# Parameters: node_name (name of the node)
#				role 	(role of the ports, optional)
#			 	link 	(link state of port)
#				type	(type of port)
# Returns: Reference to an array of port names
# Purpose: Find all ports on the given node, with given parameters
############################################################################
sub get_port_list {
	my %params = validate (
		@_,
		{
			node_name	=> {type => SCALAR, optional => 0},
			role		=> {type => SCALAR, optional => 1, default => 'data'},
			link  		=> {type => SCALAR, optional => 1, default => 'up'},
			type  		=> {type => SCALAR, optional => 1, default => '*'},
		}
	);

	my $node_obj = NACL::C::Node->find(
		command_interface	=> $Node,
        filter  => {
            node => $params{node_name},
        });
	
	my @ports = NACL::C::NetworkPort->find(
 			command_interface	=> $node_obj,
 			filter =>	{
 				role => $params{role},
 				type => $params{type},
 				node => $params{node_name},
 				link => $params{link},
 			},
 	);	
	my @port_list  = ();
	foreach my $port (@ports) {
		push(@port_list, $port->{port});
	}

	return \@port_list;
	
}


=head2 revive_port 

    my $ret = revive_port(node_name = $node, port = $port);

This routine attempts to bring a port to linke state 'up'

Returns: 1 on success, 0 failure

The following named arguments are accepted,

=over

=item node_name

(Required) : name of the node of the port

=item port

(Required) : port to revive

=item sleep

(Optional) : # seconds to wait before checking port state (Default 5)

=back 

=for html <hr>

=cut
############################################################################
# Subroutine: revive_port
# Parameters: node_name (name of the node)
#				port 	(port to revive)
#			 	sleep 	(how long to wait before checking state)
# Returns: 1 on success, 0 failure
# Purpose: Attempt to set the state of a port to link = 'up'
############################################################################
sub revive_port {
	
	my %params = validate (
		@_,
		{
			node_name	=> {type => SCALAR, optional => 0},
			port		=> {type => SCALAR, optional => 0},
			sleep		=> {type => SCALAR, optional => 1, default => 5},
		}
	);
	
	my $node = $params{node_name};
	my $port = $params{port};
	
	my $status = 1;
	
	my $node_obj = NACL::C::Node->find(
		command_interface	=> $Node,
        filter  => {
            node    => $node,
        });

		
	my $port_obj = NACL::C::NetworkPort->find(
		command_interface 	=> $node_obj,
		filter	=> {
			port	=> $port,
			node	=> $node,
		});	
		
	my $port_state = $port_obj->state();

	if ($port_state->{link} ne 'up') {
		Network::Failover_Utils::failover_target_state_change(
			target 	=> $node . ":" . $port,
			state	=> "true",
			sleep 	=> $params{sleep},
		);
		$port_state = $port_obj->state();
		if ($port_state->{link} ne 'up') {
			$status = 0;
		}
			
	};
	return $status;		
}


=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES



=head1 AUTHOR

    dl-nwkg-test (dl-nwkg-test@netapp.com)
    $Id: $
    Last revised $Date: $

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut

1;