# Perl Module
# $Id: //depot/prod/test/Rsierranevada/lib/Network/SCONTopology.pm#4 $ 
# vim: ft=perl tabstop=3 expandtab
#
# Purpose:   SCON Topology Object.
# TODO
#  add IPV6 
#     - check ipv6 supportability
#     - do ndp to get prefix per port
# 
#
## @summary
##
## @description
##
## @author      parsons@netapp.com
##
## @pod here
#
# POD at end of file

package Network::SCONTopology;

use strict;

# Distro
use Time::Local;
use Date::Parse;
use File::Path qw(make_path);
use Data::Dumper;
   $Data::Dumper::Indent   = 1;
   $Data::Dumper::Sortkeys = 1;

# NATE
use Options;
use Params::Common;
#use Tharn qw($Log);
use NATE::Log qw(log_global);

# NACL
use NACL::CS::ClusterIdentity;
use NACL::CS::NetworkInterface;
use NACL::CS::NetworkPort;
use NACL::CS::NetworkPortIfgrp;
use NACL::CS::NetworkPortVlan;
use NACL::CS::Vserver;
use NACL::CS::NetworkIpspace;             # IPSpace only
use NACL::CS::NetworkRoutingGroupsRoute;  # non-IPSpace
use NACL::CS::NetworkRoute;               # IPSpace
use NACL::C::Job;
use NACL::MTask::SCP;

BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter);
}

# Define TAGs
our @LogMsg = qw(decode_logmsg_hex 
                parse_logmsg_datestr 
                parse_logmsg_hexstamp
);
our @TAG_2 = qw();

# Exports

our @EXPORT  = (
   @LogMsg,
   @TAG_2,
);

our @EXPORT_OK  = qw();
our %EXPORT_TAGS = (
   'logmsg'   => [@LogMsg],
   'tag_2'    => [@TAG_2],
);

our $ERROR;

my $Log = log_global(1);
my $Alt_log_path = sprintf("%s/%s/topology.log", Tharn::param('LOGDIR'), Tharn::param('RUNID'));



my $Options = new Options;
$Options->parameters(definitions=> [
      { 'name'        => 'foo',
        'description' => 'foo foo',
        'default'     => 1,
      },
      { 'name'        => 'TOPOLOGY_MLOG_SCP_SLEEP',
        'description' => 'Amount of time in seconds to sleep between triage tool running and scp of log',
        'default'     => 15,
      },
   ]
);
$Options->process();
my $options = $Options->get_hash();


#=============================================================================
#
#  Constructor
#  Create object to be used to query filers and create SCON Object.
#  Methods will aceess the object to compare/lookup/check config values
#
#  Args
#   nacl_c_node   => <ref>    ref to list made by  NACL::C::Node->find();  
#   fetch_port    => <0|1>    fetch port/interface objects
#   fetch_vserver => <0|1>    fetch vserver/ipspace objects      
#   fetch_shell   => <0|1>
#
# This module will check forIPSpace support on each node given


sub new {
   my $_proto = shift;
   my $_class = ref($_proto) || $_proto;
   my $_self  = {};
   # bless object
   bless($_self, $_class);

   $ERROR = '';
   # get object arguments
   my %_arg = @_;

   my ($_a, $_n, $_ok);

   foreach $_a (keys %_arg) {
      if    ($_a =~ /^-?nacl_c_node$/i)    { $_self->{'input'}{'nacl_nodes'}    = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?fetch_port$/i)     { $_self->{'input'}{'fetch_port'}    = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?fetch_vserver$/i)  { $_self->{'input'}{'fetch_vserver'} = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?fetch_shell$/i)    { $_self->{'input'}{'fetch_shell'}   = delete($_arg{$_a}); }
      elsif ($_a =~ /^-?bar$/i)            { $_self->{'input'}{'bar'}           = delete($_arg{$_a}); }
      else {
         $ERROR = "unsupported option: [$_a]";
         return( wantarray ? (undef, $ERROR) : undef );
      }
   }

   # Check Args
   #
   # check given node nacl node list
   if ( !defined($_self->{'input'}{'nacl_nodes'}) ) {
      $ERROR = "option required:  nacl_c_node";
      return( wantarray ? (undef, $ERROR) : undef );
   }
   # validate it is a NACL::C::Node object
   # If CLI is given as FILER_LIST=NODES NODES=f1, 
   # this will still work in list context 
   foreach $_n ( @{$_self->{'input'}{'nacl_nodes'}} ) {
      if ( !Scalar::Util::blessed($_n) || !$_n->isa('NACL::C::Node')) {
         $ERROR = "nacl_c_node element not a NACL::C::Node isa " . ref($_n);
         return( wantarray ? (undef, $ERROR) : undef );
      }
   }

   # Determine same cluster
   ($_ok, $ERROR) = $_self->determine_cluster_identity;
   unless($_ok) { 
      return( wantarray ? (undef, $ERROR) : undef );
   }



   # Determine IPSpace support
   # not concern with pass/fail, just want the object to know
   $_ok = $_self->determine_ipspace_support;

   # fetch_port flag set
   if ( $_self->{'input'}{'fetch_port'} ) {
      ($_ok, $ERROR) = $_self->fetch_port_if();
      unless($_ok) {
          return( wantarray ? (undef, $ERROR) : undef );
      }
   }

   # fetch_vserver
   if ( $_self->{'input'}{'fetch_vserver'} ) {
      ($_ok, $ERROR) = $_self->fetch_vserver_ipspace();
      unless($_ok) {
          return( wantarray ? (undef, $ERROR) : undef );
      }
   }

   # fetch_shell
   if ( $_self->{'input'}{'fetch_shell'} ) {
      ($_ok, $ERROR) = $_self->fetch_shell_commands();
      unless($_ok) {
          return( wantarray ? (undef, $ERROR) : undef );
      }
   }


   return( wantarray ? ($_self, $ERROR) : $_self )
}

#---------------------------------------------------------------------------
# Update object
# Each fetch method sets a flag when it runs
# Look for set flags and re-run the fetch method to update
# 
#
sub update_topology {

   my $_obj = shift;
   my ($_k, $_ok);
   my $_ret = 1;

   foreach $_k ( keys %{$_obj->{'_Config'}{'fetch'}} ) {
      $Log->comment("Update topology: $_k");
      if ( $_k eq 'PortInfo' ) { 
         ($_ok, $ERROR) = $_obj->fetch_port_if;
         unless ( $_ok ) { $_ret = 0; last; }
      }
      if ( $_k eq 'VServer' )  { 
         ($_ok, $ERROR) = $_obj->fetch_vserver_ipspace;
         unless ( $_ok ) { $_ret = 0; last; } 
      }
      if ( $_k eq 'Shell' ) { 
         ($_ok, $ERROR) = $_obj->fetch_shell_commands;
         unless ( $_ok ) { $_ret = 0; last; } 
      }
   }

   return( wantarray ? ($_ret, $ERROR) :  $_ret);
} 



#---------------------------------------------------------------------------
# Return error string
#
sub error_string { my $_obj= shift; return($ERROR); }



#---------------------------------------------------------------------------
# Function to handle switching of log destination
# (switching back is a one-liner)
#
# $_[0] = log switch message
#
# not documented
#
sub switch_logging {
   my $_str = sprintf("%s %s", $_[0], time);

   $Log->comment("--- Log switch [$_str] to: $Alt_log_path");
   $Log->redirect('path' => $Alt_log_path);
   $Log->comment(sprintf("%s %s %s", '='x20, $_str, '='x20));

   1;
}
#---------------------------------------------------------------------------
# Set node index for each node name
#    $_obj->{'_Config'}{'NodeIndex'}{<nodename>} = <int>
#
# Args
#   none work on object
# Return
#   how many node indexed
# 
sub set_node_index {
   my $_obj = shift;
   my ($_node);
   my $_c = 0;
   # init the hash
   $_obj->{'_Config'}{'NodeIndex'} = ();
   # loop through node list and set index
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_obj->{'_Config'}{'NodeIndex'}{$_node->hostrec->{'hostname'}} = $_c;
      $Log->comment("set_node_index: $_c  " . $_node->hostrec->{'hostname'});
      $_c++;
   }

   $_c;
}

#---------------------------------------------------------------------------
# get ref to node index hash
#
# Arg
#  none
#
# Return
#  <ref>  ref to $_obj->{'_Config'}{'NodeIndex'}
#
#
sub get_node_index_href { 
   my $_obj = shift;
   if( !defined($_obj->{'_Config'}{'NodeIndex'}) ) { $_obj->set_node_index; }
   return $_obj->{'_Config'}{'NodeIndex'};
}

#---------------------------------------------------------------------------
# get ref to node index hash
#
# Arg
#  $_[0] = node name
#
# Return
#  <int>  index in @{$_obj->{'input'}{'nacl_nodes'}} to given node
#  undef if not found
#
sub get_node_index {
   my $_obj = shift;
 
   if( !defined($_obj->{'_Config'}{'NodeIndex'}) ) { $_obj->set_node_index; }
 
   if ( exists($_obj->{'_Config'}{'NodeIndex'}{$_[0]}) ) {
      return($_obj->{'_Config'}{'NodeIndex'}{$_[0]});
   } 
   $ERROR = "get_node_index: node index for $_[0] not fond in object";
   return(undef);
}


#
#-----------------------------------------------------------------------------
#
# Check for IPSpace support
# Populate the object
#    $_obj->{'_Config'}{'support'}{'IPSpace'}{<node>} = <0|1>
# 
# Args 
#  none
#
# Return
#    wantarray ? ($count, $ERROR) : $count );
#
# This module only prints comments, it does not make pass/fail decisions
# This is up to user bassed on return, off which the error string will 
# indicate the problem for the caller to log result as desired
#
sub determine_ipspace_support {
   my $_obj = shift;
   my $_node;
   my $_bool = undef;
   my $_c = 0;
   my $_s = 0;
   my @_n = ();
   $ERROR = '';

   # check each node for IPSpace support
   $Log->comment("### Verifying IPSpace feature support");
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      # init the support flag
      $_obj->{'_Config'}{'support'}{'IPSpace'}{$_node->hostrec->{'hostname'}} = 0;
      $_bool = $_node->is_feature_supported( 'feature' => 'ipspaces_feature');
      # ipspace supported
      if ( $_bool ) {
         $_obj->{'_Config'}{'support'}{'IPSpace'}{$_node->hostrec->{'hostname'}} = 1;
         $_c++;
         $_s++;
         $Log->comment("### IPSpace support on " . $_node->hostrec->{'hostname'} . " "
                      ."[$_bool] is supported"
         );
      }
      # ipspace not supported
      else {
         $_obj->{'_Config'}{'support'}{'IPSpace'}{$_node->hostrec->{'hostname'}} = 0;
         $_c++;
         push(@_n, $_node->hostrec->{'hostname'});
         $Log->comment("### IPSpace support on " . $_node->hostrec->{'hostname'} . " "
                      ."[$_bool] is not_supported"
         );
      }
   }

   # Check the info gather
   # Somehow a node got skipped being checked
   # Count of nodes checked should equal the count of nodes in the list
   if ( $_c != scalar(@{$_obj->{'input'}{'nacl_nodes'}}) ) {
      $_c = 0;
      $ERROR = 'IPSpaces support not  verified on all nodes';
   }
   # Support not seen on all nodes
   # Support count should  equal the count of nodes in the list
   elsif ($_s != scalar(@{$_obj->{'input'}{'nacl_nodes'}}) ) {
      $_c = 0;
      $ERROR = 'IPSpaces not supported on nodes: ';
      foreach $_node ( @_n ) 
         {$ERROR .= "$_node, "}
      $ERROR =~ s/,\s+$//;
   }

   return( wantarray ? ($_c, $ERROR) : $_c );
}

#-----------------------------------------------------------------------------
# Determine cluster affiliation
# All nodes required to be in the same cluster, if not then error
#
# Popultate object
#   $_obj->{'Cluster'}{<node>} = <cluster object>;
#
# Args
#   none:   work on cluster object
#
#  wantarray ? (count, error) : count
#


sub determine_cluster_identity {

   my $_obj = shift;
   my ($_node, $_cluster);
   my $_last_cluster;
   my $_c = 0;

   # get the cluster object foreach node
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_cluster = NACL::CS::ClusterIdentity->fetch(
            'command_interface'  => $_node,
      );
      delete($_cluster->{'command_interface'});
      $_obj->{'Cluster'}{$_node->hostrec->{'hostname'}} = $_cluster;
      $_last_cluster = $_cluster->{'name'};
   }
  
   # check all the cluster name
   $Log->comment("### Check cluster name to be $_last_cluster");
   foreach $_node  (sort keys %{$_obj->{'Cluster'}} ) {
      $Log->comment("cluster name: $_node " . $_obj->{'Cluster'}{$_node}->{'name'});
      if ( $_obj->{'Cluster'}{$_node}->{'name'} ne $_last_cluster ) {
         $ERROR = sprintf("cluster name: %s %s mistmatch expect %s",
            $_node,  $_obj->{'Cluster'}{$_node}->{'name'}, $_last_cluster
         );  
         $Log->error("determine_cluster_identity: $ERROR");
         return( wantarray ? (undef, $ERROR) : undef );
      }
      $_c++;
   } 
   return( wantarray ? ($_c, '') : $_c );
}


#-----------------------------------------------------------------------------
# Given an API, cli command with underscore, get a list of 
# accetpable arguments to the options.
# Store in object per node
#   - different nodes may have different release with new or removed fields
#
# Populate Object
#     %obj->{'Args'}{<api>}{<node>}{<field>} = (arg,1, arg2, ...);
# 
# Args
#   api  => <str>   api, cli command with underscores (required)
# Return
#   return( wantarray ? (count, $ERROR) : count ); 
#
sub determine_cli_args {

   my $_obj = shift;
   my %_args = @_;
   my ($_node, $node_apiset, 
       $_cli, $_f, $_str, 
       $_parsed, @_tokens,
   );
   my $_c = 0;
   $ERROR = 'noErr';

   # check api
   if (!defined($_args{'api'}) ) {
      $ERROR = 'arg  api required';
      return( wantarray ? (undef, $ERROR) : undef );
   }

   $Log->comment("Determine option arguments for $_args{'api'}");
   # foreach node, get an apiset and use APISet::Node::CLI::CMode::cli_command_helper
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $Log->comment("Determine node: " . $_node->hostrec->{'hostname'} . " " 
                   ."API: $_args{'api'} arguments"
      ); 
      $node_apiset = $_node->apiset( 'interface' => 'CLI' );
      $_cli = $node_apiset->cli_command_helper('api' => $_args{'api'});
      $Log->debug("CLI Command Helper: " . Dumper $_cli);

      # loop throug each field of cli_command
      foreach $_f ( keys %{$_cli->{'fields'}} ) {
         # description strin seen to have the values
         # cat the description to one big string and parse out values
         if ( defined($_cli->{'fields'}{$_f}{'description'}) ) {
            $_str = join(' ', @{$_cli->{'fields'}{$_f}{'description'}});
            @_tokens = ();
            if ($_str =~ /possible value.s.\sare\s+(.+)/ ) {
               $_parsed = $1;
               @_tokens = split(/,/, $_parsed);
               $Log->comment("Parsed: API: $_args{'api'} Option: $_f Args: [" 
                            . join(' -  ', @_tokens) . "]"
               );
               @{$_obj->{'Args'}{$_args{'api'}}{$_node->hostrec->{'hostname'}}{$_f}} = @_tokens;
               $_c = $_c + scalar(@_tokens);
            }
            else {
               $Log->comment("Parsed: API: $_args{'api'} Option: $_f  no pattern match")
            }
         }
         else {
            $Log->comment("Parsed: API: $_args{'api'} Option: $_f  no description");
         }
      }
   }
   $ERROR = "no option arguments parsed for api: $_args{'api'}" if $_c==0; 
   return( wantarray ? ($_c, $ERROR) : $_c);

}


#
#-----------------------------------------------------------------------------
#
# Fetch the Port and Interface Object and populate object
# 
# $obj->{'PortInfo}{<node>}{<port name>}{'port'} = <port object>
#                                       {'if'}   = [<IF object>, <>, ... ]
#                                       {'_stat'}{'if_count'}  = <int> number of interfaces on port
#                                       {'_stat'}{'NodeIndex'} = <int> index in NACL @Nodes   
#                                       {'_stat'}{'InIFgrp'}   = <str> IFGrp name containing port  
#                                       {'_stat'}{'InVLAN'}    = [];   List of VLANs containing this port
#                                       {'_stat'}{'isaPort'}   = 0|1   seen in 'port show'
#                                       {'_stat'}{'isaIFgrp'}  = 0|1   seen in 'ifgrp show'
#                                       {'_stat'}{'isaVlan'}   = 0|1   seen in 'vlan show' 
#                                       {'_stat'}{'inaVlan'}   = <int>; number of vlan seen in
#                                       {'_stat'}{'inaIFgrp'}  = <int>; number of ifgrp seen in
#       {'IFGroup'}{<node>}{<ifgr name>}{'ifgrp'} = <ifgrp object>
#       {'VLAN'}{<vlan-id>} = [ <vlan obj>, <>, ..]
#       {'_Config'}{'fetch'}{'PortInfo'} = <int>   epoch when last ran
#
# Args: none
#
# Return  
#    return( wantarray ? ($_port_count, $ERROR) : $_port_count );

sub fetch_port_if {

   my $_obj = shift;
   my($_node, $_port, $_if, $_ifgrp, $_vlan, $_cs);
   my @_Ports;
   my @_Interfaces;
   my @_IFGrps;
   my @_Vlans;
   my $_node_idx = 0;
   my $_port_count = 0;
   $ERROR = "noError port count $_port_count";

   &switch_logging('Fetch Port and Interface');

   $Log->comment("### Getting port and interface info");

   # init the storage and set timer   
   %{$_obj->{'PortInfo'}} = ();
   %{$_obj->{'IFGroup'}}  = ();
   %{$_obj->{'VLAN'}}     = ();

   $_obj->{'_Config'}{'fetch'}{'PortInfo'} = time;

   # loop thru nodes and get local ports and IFs
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {

      $Log->comment("### Fetch Port and Interface info from " . $_node->hostrec->{'hostname'});
      @_Ports = NACL::CS::NetworkPort->fetch(
            'command_interface'  => $_node,
            'filter' => {'node'  => 'local',},
      );

      @_Interfaces = NACL::CS::NetworkInterface->fetch(
            'command_interface'  => $_node,
             'allow_empty'       => 1,
            'filter' => {'home-node' => 'local',},
      );
      @_IFGrps = NACL::CS::NetworkPortIfgrp->fetch(
            'command_interface' => $_node,
            'allow_empty'       => 1,
            'filter' => {'node' => 'local',},
      );
      @_Vlans = NACL::CS::NetworkPortVlan->fetch(
            'command_interface' => $_node,
            'allow_empty'       => 1,
            'filter' => {'node' => 'local',},
      );

      # build %PortInfo
      $Log->comment("### Building PortInfo:");
      # ports
      foreach $_port (@_Ports) {
         $Log->comment("###  Indexing with Node [" . $_port->node() . "] Port [" . $_port->port() ."]");
         delete($_port->{'command_interface'});
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'port'} = $_port;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'NodeIndex'} = $_node_idx;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'InIFgrp'}   = '';
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'InVLAN'}    = [];
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'isaPort'}   = 1;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'isaIFgrp'}  = 0;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'isaVlan'}   = 0;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'inaVlan'}   = 0;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'inaIFgrp'}  = 0;
         $_obj->{'PortInfo'}{$_port->node()}{$_port->port()}{'_stat'}{'if_count'}  = 0;
         $_port_count++;
      }
      # interfaces
      foreach $_if ( @_Interfaces ) {
          $Log->comment("### Index IF Node ["  . $_if->home_node() . "] "
                             . "Port [" . $_if->home_port() . "] "
                             . "IF ["   . $_if->lif() . "] "
          );
          delete($_if->{'command_interface'});
          push (@{$_obj->{'PortInfo'}{$_if->home_node()}{$_if->home_port()}{'if'}}, $_if);
          $_obj->{'PortInfo'}{$_if->home_node()}{$_if->home_port()}{'_stat'}{'if_count'}++;
      }
      # if groups
      foreach $_ifgrp ( @_IFGrps ) {
         $Log->comment("### Add IFGrp Node ["  . $_ifgrp->node() . "] "
                            . "IFGrp [" . $_ifgrp->ifgrp() . "] "
         );
         delete($_ifgrp->{'command_interface'});
         $_obj->{'IFGroup'}{$_ifgrp->node()}{$_ifgrp->ifgrp()}{'ifgrp'} = $_ifgrp;

         $_obj->{'PortInfo'}{$_ifgrp->node()}{$_ifgrp->ifgrp()}{'_isaIFgrp'}  = 1;

         # port can only be in one IFGroup
         foreach $_port ( @{$_ifgrp->ports()} ) {
            $Log->comment("###   port [$_port] in ifgrp [" . $_ifgrp->ifgrp() ."]");
            next if $_ifgrp->ifgrp() eq '-';
            $_obj->{'PortInfo'}{$_ifgrp->node()}{$_port}{'_stat'}{'InIFgrp'} =  $_ifgrp->ifgrp();
            $_obj->{'PortInfo'}{$_ifgrp->node()}{$_port}{'_stat'}{'inaIFgrp'}++;
         }
      }
      # vlans
      # vlans can be made from an ifgroup port
      foreach $_vlan ( @_Vlans ) {
         $Log->comment("### Add VLAN  Node ["  . $_vlan->node() . "] "
                      . "VLAN [" .  $_vlan->vlan_name() . "] "
         );
         delete($_vlan->{'command_interface'});
         ###$_obj->{'VLAN'}{$_vlan->node()}{$_vlan->vlan_name()}{'vlan'} = $_vlan;
         push(@{$_obj->{'VLAN'}{$_vlan->vlan_id()}}, $_vlan);
         
         # populate the PortInfo tracking
         ###push(@{$_obj->{'PortInfo'}{$_vlan->node()}{$_vlan->port()}{'_stat'}{'InVLAN'}}, $_vlan->vlan_name());
         push(@{$_obj->{'PortInfo'}{$_vlan->node()}{$_vlan->port()}{'_stat'}{'InVLAN'}}, $_vlan->vlan_id());
         $_obj->{'PortInfo'}{$_vlan->node()}{$_vlan->port()}{'_stat'}{'inaVlan'}++;
         $_obj->{'PortInfo'}{$_vlan->node()}{$_vlan->vlan_name()}{'_stat'}{'isaVlan'} = 1;
      }

      $_node_idx++;
   }
   #$Log->comment("### PortInfo hash: " . Dumper $_obj->{'PortInfo'}); 

   $Log->reset_path();
   return( wantarray ? ($_port_count, $ERROR) : $_port_count );
}



#-----------------------------------------------------------------------------
# Get VServer and IP Space info from ngsh#
# Include SMF table dumps
# debug smdb table vserver_for_ipspace show
# {   'ipspace-id' => '4294967292',
#     'type'       => 'system',
#     'uuid'       => '4d16178b-2d7f-11e1-8fbf-123478563412',
#     'vserver'    => 'ACP',
#     'vserver-id' => '4294967292'
# },
#
# debug smdb table vserver_by_name  show
# {  'id'          => 'ACP',
#    'id-unsigned' => '4294967292',
#    'ipspace-id'  => '4294967292',
#    'language'    => 'en_US',
#    'type'        => 'system',
#    'uuid'        => '4d16178b-2d7f-11e1-8fbf-123478563412',
#    'vif-owner'   => 'ACP',
#    'vserver'     => 'ACP'
#  },
#
#
#  %obj->{'VServer'}{<vserver>}{'vserver'} = <vserver object>
#                              {'ipspace'} = <ipspace object>
#                              {'_vserver_by_name'}     = href      hash of vserver_by_name
#                              {'_vserver_for_ipspace'} = href  hash of vserver_for_ipspace
#        {'IPSpace'}{<ipspace>}{'ipspace'}            = <nacl ipspace object>
#                              {'vserver'}{<vserver>} = <nacl vserver object>
#        {'Maps'}{'IPSpace'}{<id>} = <name>
#                           {<name>} = <id>
#  {'_Config'}{'fetch'}{'VServer'} = <int>;    

sub fetch_vserver_ipspace  {

   my $_obj = shift;

   my ( $_node, $_vserver, $_ipspace, $_route,
        $_apiset, 
        $_vserver_for_ipspace, $_vserver_for_ipspace_raw, 
        $_vserver_by_name, $_vserver_by_name_raw,
        @_vservers,
        $_v, $i, $_vbn, $_vfi, $_r,
   );
   my @_Vservers = ();
   my @_IPSPaces = ();
   my @_Routes   = ();
   my $_verserv_count = 0;
   $ERROR = '';

   &switch_logging('### Fetch VServer and IPspace');

   $_obj->{'_Config'}{'fetch'}{'VServer'} = time;

   %{$_obj->{'VServer'}} = ();
   %{$_obj->{'IPSpace'}} = ();
   %{$_obj->{'Maps'}{'IPSpace'}} = ();

   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      # needed for smf debug tables
      $_apiset = $_node->apiset(category=> 'Node',
                                interface => 'CLI',
                                set       => 'CMode',
      );

      # VServer
      $Log->comment("### VServers CS on " . $_node->hostrec->{'hostname'});
      @_Vservers =  NACL::CS::Vserver->fetch(
         'command_interface'  => $_node,
         'allow_empty'        => 1,
      );
      # vserver smf tables
      $_vserver_by_name_raw = $_apiset->execute_raw_command(
         'command' => 'debug smdb table vserver_by_name show',
      );
      $_vserver_by_name = &topo_ngsh_parser($_vserver_by_name_raw);


      # IPspaces info
      if ( $_obj->{'_Config'}{'support'}{'IPSpace'}{$_node->hostrec->{'hostname'}} ) {
         # Net Routes
         $Log->comment("### IPSpace Routes on " . $_node->hostrec->{'hostname'});
         @_Routes = NACL::CS::NetworkRoute->fetch(
            'command_interface'  => $_node,
            'allow_empty'        => 1,
         );

         $Log->comment("### IPSpaces CS on " . $_node->hostrec->{'hostname'});
         @_IPSPaces =  NACL::CS::NetworkIpspace->fetch(
            'command_interface'  => $_node,
            'allow_empty'        => 1,
         );
   
         # ipspaces smf tables
         $_vserver_for_ipspace_raw = $_apiset->execute_raw_command(
            'command' => 'debug smdb table vserver_for_ipspace show',
         );
         $_vserver_for_ipspace = &topo_ngsh_parser($_vserver_for_ipspace_raw);
      }
      else {
         # Net Routes
         $Log->comment("### Routes on " . $_node->hostrec->{'hostname'});
         @_Routes = NACL::CS::NetworkRoutingGroupsRoute->fetch(
            'command_interface'  => $_node,
            'allow_empty'        => 1,
         );
      }

 
      # this is cluster-wide, so once populated, exit loop
      if (scalar(@_Vservers) && scalar(@_IPSPaces) && scalar(@_Routes) ) {last;}
   }

   # remove command_interface
   # TODO make this configurable
   foreach $_vserver ( @_Vservers ) { delete($_vserver->{'command_interface'}); }
   foreach $_ipspace ( @_IPSPaces ) { delete($_ipspace->{'command_interface'}); }
   foreach $_route   ( @_Routes )   { delete($_route->{'command_interface'}); }

   # VServer
   #  keys: 'vserver', 'ipspace', 'type', 'uuid'
   # %obj->{'VServer'}{<vserver>}{'vserver'} = <vserver object>
   #                             {'ipspace'} = <ipspace object>
   #                             {'_vserver_by_name'} = href      hash of vserver_by_name
   #                             {'_vserver_for_ipspace'} = href  hash of vserver_for_ipspace
   #
   foreach $_vserver ( @_Vservers ) {
      $_obj->{'VServer'}{$_vserver->{'vserver'}}{'vserver'} = $_vserver;
      $_verserv_count++;
      $Log->comment("### Match VServer: " . $_vserver->{'vserver'} );
      ##$Log->comment(Dumper $_vserver);
      # match a vserver component to ngsh ipspace
      $Log->comment("### Match VServer to IPSpace");
      foreach $_ipspace ( @_IPSPaces ) {
         $Log->comment("###    Look at IPSpace: " . $_ipspace->ipspace());
         # ipspace vserver list 
         if ( ref($_ipspace->{'vservers'}) eq 'ARRAY' ) {
            $Log->comment("###    multiple vservers in ipspace: " . $_ipspace->ipspace());
            foreach $_v ( @{$_ipspace->{'vservers'}} ) {
               if ( $_vserver->{'vserver'} eq $_v ) {
                  $_obj->{'VServer'}{$_vserver->{'vserver'}}{'ipspace'} = $_ipspace;
                  $_vserver->{'_matched_to_ipspace'}++;
                  $Log->comment("###      matched vserver $_v from list to ipspace: " . $_ipspace->{'ipspace'});
                  ##$Log->comment(Dumper $_ipspace);
               }
            }
         }
         # ipspace vserver
         elsif ( $_ipspace->{'vservers'} eq $_vserver->{'vserver'} ) {
            $_obj->{'VServer'}{$_vserver->{'vserver'}}{'ipspace'} = $_ipspace;
            $_vserver->{'_matched_to_ipspace'}++;
            $Log->comment("###       matched single vserver: " . $_vserver->{'vserver'}  
                                             ."ipspace: " . $_ipspace->{'ipspace'}
            );
            ##$Log->comment(Dumper $_ipspace);
         }
         else {
            $Log->comment("###        not matched to this ipspace");
         }
      } 
      # Match VServer to vserver_by_name
      foreach $_vbn ( @{$_vserver_by_name} ) {
         if ( &match_vserver_to_vserver_by_name($_vserver, $_vbn) ) {
              $_vbn->{'_matched_to_vserver'}++;
              $_vserver->{'_matched_to_vserver_by_name'}++;
              $_obj->{'VServer'}{$_vserver->{'vserver'}}{'_vserver_by_name'} = $_vbn;
              $Log->comment("###    Found vserver_by_name ");
              #$Log->comment(Dumper $_vbn);
              ### TODO put a last ???
         }
      }
      # Match VServer to vserver_for_ipspace
      foreach $_vfi ( @{$_vserver_for_ipspace} ) {
         if ( &match_vserver_to_vserver_for_ipspace($_vserver, $_vfi) ) {
            $_vfi->{'_matched_to_vserver'}++;
            $_vserver->{'_matched_to_vserver_for_ipspace'}++;
            $_obj->{'VServer'}{$_vserver->{'vserver'}}{'_vserver_for_ipspace'} = $_vfi;
            $Log->comment("###    Found vserver_for_ipspace ");
            #$Log->comment(Dumper $_vfi);
         } 
      }
   }

   ### TODO store these unmatched ones in the object  $obj->{'_unmatched'}{'foo'}
   foreach $_vbn ( @{$_vserver_by_name} ) {
      next if $_vbn->{'_matched_to_vserver'};
      $_vbn->{'_matched_to_vserver'} = 0;
      $Log->comment("###    Unmatched vserver_by_name: " .Dumper $_vbn);
   }
   foreach $_vfi ( @{$_vserver_for_ipspace} ) {
      next if $_vfi->{'_matched_to_vserver'};
      $_vfi->{'_matched_to_vserver'} = 0;
      $Log->comment("###    Unmatched vserver_for_ipspace: " .Dumper $_vfi);
   }

   # Routes
   # IPSpaces use NetworkRoute
   # Non-IPspace uses  NetworkRoutingGroupRoute
   # The NACL object dif is only NetworkRoute has same fields + 'ipspace'
   foreach $_route ( @_Routes ) {
      push(@{$_obj->{'VServer'}{$_route->{'vserver'}}{'routes'}}, $_route); 
      $Log->comment("### Route entry : " . $_route->{'destination'} . " " 
                    ."for vserver: " . $_route->{'vserver'}
      );
   }

 
   # IPSpace
   # keys 'ipspace', 'vservers' scalar or array  'ports' list 'uuid'
   # %obj->{'IPSpace'}{<ipspace>}{'ipspace'} = <nacl ipspace object>
   #                             {'vserver'}{<vserver>} = <nacl vserver object>
   # IPSPaces can have multiple vserver
   #
   foreach $_ipspace ( @_IPSPaces ) {
      $_obj->{'IPSpace'}{$_ipspace->{'ipspace'}}{'ipspace'} = $_ipspace;

      $Log->comment("### Match IPSpace: " . $_ipspace->{'ipspace'} );
      ##$Log->comment(Dumper $_ipspace);

      # extract the vserver names, could be list or scalar, yea for nacl
      if ( ref($_ipspace->{'vservers'}) eq 'ARRAY' ) { @_vservers = @{$_ipspace->{'vservers'}}; }
      else                                           { push(@_vservers, $_ipspace->{'vservers'}); }
      $Log->comment("### Match IPSpace [$_ipspace->{'ipspace'}] with VServers "
          . sprintf("[%s] ", join(', ', @_vservers))
      );

      # loop thru each vserver name
      foreach $_v ( @_vservers ) {
         $Log->comment("###    Match IPSpace: [$_ipspace->{'ipspace'}]  Vserver: [$_v]");
         # look at each vserver
         #    $_i->{'vservers'}  $_i->{'ipspace'} 
         #    $_v->{'vserver'}   $_v->{'ipspace'}
         foreach $_vserver ( @_Vservers ) {
            if ( $_v eq $_vserver->{'vserver'} && $_ipspace->{'ipspace'} eq $_vserver->{'ipspace'} ) {
               $_obj->{'IPSpace'}{$_ipspace->{'ipspace'}}{'vserver'}{$_v} = $_vserver;
               $_ipspace->{'_matched_to_vserver'}++;
               $_vserver->{'_matched_to_ipspace'}++;
               $Log->comment("###       matched ipspace [" . $_ipspace->{'ipspace'}
                            . "] to vserver: [" . $_vserver->{'vserver'} . "]"
               );
               ##$Log->comment(Dumper $_vserver);
            }

            # the vserver_by_name vserver_for_ipspace are already matched in $_obj->{'VServer'}
            # if want to also add under vserver here, can copy the code above
            # down to here
         }
      }
   }

   $_obj->map_ipspace_name_id;   

   $Log->reset_path();
   return( wantarray ? ($_verserv_count, $ERROR) : $_verserv_count  );

}
#
#-----------------------------------------------------------------------------
# Map IPSpace Name to ID
#
# Args none
# Return none, populate object
#
sub map_ipspace_name_id {

   my $_obj = shift;
   my ($_i);
   my $_r = 0;
   %{$_obj->{'Maps'}{'IPSpace'}} = ();
   if ( defined($_obj->{'IPSpace'}) ) {
      foreach $_i ( keys %{$_obj->{'IPSpace'}} ) {
         $_obj->{'Maps'}{'IPSpace'}{$_i} = $_obj->{'IPSpace'}{$_i}{'ipspace'}->id();
         $_obj->{'Maps'}{'IPSpace'}{$_obj->{'IPSpace'}{$_i}{'ipspace'}->id()} = $_i;
         $_r++;
      }
   }
   $_r;
}

#
#-----------------------------------------------------------------------------
# Get data from bsd and sk stack
#
# Use systemshell or nodescope NACL command interface and get
# command output and parse into a data structures
# Poplate the object
#
#    systemshell  = BSD shell diag user  ngsh: set diag; system node systemshell local
#    nodescope    = SK stack  7M         ngsh: run local <command>   
#
# $_obj->{'Shell'}{<node>}{'systemshell'}{'ifconfig'} = systemshell ifconfig hash
#                                        {'route'}    = systemshell route hash
#
#                         {'nodescope'}{'ifconfig'}     = nodescope ifconfig hash
#                                      {'route'}        = nodescope route hash
#                                      {'ipspace_list'} = nodescope ipspace_list hash
# 
# Args 
#   none
# Return 
#   1
# 
# TODO: make this a return(wantarray?(y,z):y)

sub fetch_shell_commands {

   my $_obj = shift;
   my ($_node, $_n, $_file,
       $_systemshell_apiset, $_systemshell_response,
       $_nodescope_apiset,   $_nodescope_response,
   );
   my %_out = ();

   &switch_logging('Fetch Shell');

   %{$_obj->{'Shell'}} = ();
   $_obj->{'_Config'}{'fetch'}{'Shell'} = time;

   $Log->comment("### Getting systemshell and nodescope command output");

   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}}  ) {
      %_out = ();
      $_n = $_node->hostrec->{'hostname'}; 

      $Log->comment("### Current node: " .  $_node->hostrec->{'hostname'} );
      #
      # systemshell (BSD)
      #
      $Log->comment("### systemshell [bsd] command gathering");
      # apiset
      $_systemshell_apiset = $_node->apiset(category  => 'Node',
                                            interface => 'CLI',
                                            set       => 'Systemshell',
      );
      $_systemshell_apiset->set_privilege("privilege-level" => "root");

      # bsd ifconfig
      $Log->comment("### systemshell [bsd] get ifconfig");
      $_systemshell_response                            = $_systemshell_apiset->ifconfig(all => 1);
      $_out{'systemshell'}{'ifconfig'}{'raw'}           = $_systemshell_response->get_raw_output();
      $_obj->{'Shell'}{$_n}{'systemshell'}{'ifconfig'}  = $_systemshell_response->get_parsed_output();

      # bsd route  (no nacl method)
      $Log->comment("### systemshell [bsd] get netstat -rn");
      $_out{'systemshell'}{'route'}{'raw'} = $_systemshell_apiset->execute_raw_command(
                                                    'command' => 'sudo netstat -rn'
      );
      $Log->comment("### systemshell [bsd] store parsed output: [$_n]");
      $_obj->{'Shell'}{$_n}{'systemshell'}{'route'} = &topo_systemshell_route_parser(
                                                        $_out{'systemshell'}{'route'}{'raw'}
      );
      #
      # nodescope (SK)
      #
      $Log->comment("### nodescope [sk] command gathering");
      # apiset
      $_nodescope_apiset = $_node->get_7m_or_nodescope_apiset();
      # sk ifconfig
      $Log->comment("### nodescope [sk] get ifconfig");
      $_nodescope_response = $_nodescope_apiset->ifconfig(all=>"-a");
      $_out{'nodescope'}{'ifconfig'}{'raw'}          = $_nodescope_response->get_raw_output();
      $_obj->{'Shell'}{$_n}{'nodescope'}{'ifconfig'} = $_nodescope_response->get_parsed_output();

      # sk route -sgn
      # $_obj->{'Shell'}{$_n}{'nodescope'}{'route'} = AoHoH due to get_parsed_output 
      #   ...{'route'} =>[{'internet' => [{},{}],
      #                    'internet_v6' => [[{},{}],
      #                   },
      #                  ],
      $Log->comment("### nodescope [sk] get route -sgn");
      $_nodescope_response = $_nodescope_apiset->route_show( 'group' => 1, 'num-mode' => 1);
      $_out{'nodescope'}{'route'}{'raw'} = $_nodescope_response->get_raw_output();
      #$_obj->{'Shell'}{$_n}{'nodescope'}{'route'} = $_nodescope_response->get_parsed_output();
      $_obj->{'Shell'}{$_n}{'nodescope'}{'route'}  = topo_nodescope_route_parser($_out{'nodescope'}{'route'}{'raw'});
      

      # ipspace list
      if ( $_obj->{'_Config'}{'support'}{'IPSpace'}{$_node->hostrec->{'hostname'}} ) {
         $Log->comment("### nodescope [sk] get ipspace list");
         $_nodescope_response = $_nodescope_apiset->ipspace_list();
         $_out{'nodescope'}{'ipspace_list'}{'raw'} = $_nodescope_response->get_raw_output();
         #$_out{'nodescope'}{'ipspace_list'}{'parsed'} = $_nodescope_response->get_parsed_output();
         $_obj->{'Shell'}{$_n}{'nodescope'}{'ipspace_list'} = topo_nodescope_ipspace_list_parser(
              $_out{'nodescope'}{'ipspace_list'}{'raw'}
         );
      }
      if(0){
      $_file = sprintf("%s/db/%s_command_out.txt", $ENV{'LOGDIR'}, $_node->hostrec->{'hostname'});
      open(FH, ">>$_file");
      printf FH ("%s", Dumper \%_out);
      close (FH);
      }
   }

   #$Log->comment("### systemshell output hash: " . Dumper \%_out);

   $Log->reset_path();
   return(1);
}

#
#-----------------------------------------------------------------------------
# Print node name to log
# Args
#   none - work on object
# Return node names or count
#
sub get_node_list {

   my $_obj = shift;
   my $_node;
   my @_nodes = ();

   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $Log->debug("Found Node in object: " . $_node->hostrec->{'hostname'});
      push(@_nodes, $_node->hostrec->{'hostname'});
   }
   return(wantarray ? (@_nodes) : scalar(@_nodes));
}

# TODO
# make method to take a string and see if it is a node in the object 


#
#-----------------------------------------------------------------------------
# Get ports for objects
# Use NACL::CS::ComponentState->_apply_filter for
# filtering the NACL port objects.
# Problem with NACL::CS::ComponentState->_apply_filter
# is that it will remove the object not matching the filter.
# Resolve by making cloning the NACL objects, pushing into
# a temp list applying filtering to that.
#
# Args 
#   node   => <str>   node name
#   filter => <ref>   ref to filter hash
#   return => <ports|objects>  set return type flag
#                 ports returns ([node, port, role], [], [],... 
#                 objects returns (<port object>, <>, ...)
# TODO remove this after making a comparison tempplate on this
#      keeping for now to show a different way to filter, albeit wastefull
#-#sub _get_port_list {
#-#
#-#   my $_obj = shift;
#-#   my %_args = @_;
#-#   my @_nodes = ();
#-#   my ($_n, $_p, $_h, $_clone);
#-#   my @_clones = ();
#-#   my @_return = ();
#-#
#-#   # TODO do want to keep this, the filtering will cover it
#-#   # get the nodes to query for from arg
#-#   #   node => n1,n2,n3
#-#   #   node => n1
#-#   if ( $_args{'node'} ) {
#-#      if    ( $_args{'node'} =~ /,/)   { @_nodes = split(',', $_args{'node'}); }
#-#      elsif ( $_args{'node'} =~ /\w+/) { push(@_nodes, $_args{'node'}); }
#-#   }
#-#   else {@_nodes = $_obj->get_node_list;}
#-#
#-#   # make sure have return type
#-#   if ( !defined($_args{'return'}) ) { $_args{'return'} = 'ports'; } 
#-#
#-#
#-#   # Access all the port objects in the Topology object
#-#   foreach $_n ( @_nodes ) {
#-#      foreach $_p ( keys %{$_obj->{'PortInfo'}{$_n}} ) {
#-#         $Log->comment("get_port_list checking: {$_n}{$_p} object: " 
#-#            . $_obj->{'PortInfo'}{$_n}{$_p}{'port'}->{'port'} . " "
#-#            . $_obj->{'PortInfo'}{$_n}{$_p}{'port'}->{'role'} . " "
#-#         );
#-#         # clone the object
#-#         push(@_clones, Storable::dclone($_obj->{'PortInfo'}{$_n}{$_p}{'port'}));
#-#      }
#-#   }
#-#   # apply filter if given
#-#   if ( defined($_args{'filter'}) && scalar(@_clones) ) {
#-#      $Log->comment("Filter " . scalar(@_clones) . " objects");
#-#      NACL::CS::ComponentState->_apply_filter(
#-#         'filter'     => $_args{'filter'},
#-#         'state_objs' => \@_clones
#-#      );
#-#      $Log->comment("Filtered out " . scalar(@_clones) . " objects");
#-#   }
#-#
#-#   # See what to return
#-#   # Return reference to nacl port objects
#-#   if ( $_args{'return'} eq 'objects' ) { return(\@_clones); }  
#-#   # return AoA of port info [node,port,role] 
#-#   else {
#-#      foreach $_h ( @_clones ) {
#-#         $Log->comment("returning cloned object values: " 
#-#            . $_h->{'node'} . " " . $_h->{'port'} . " " . $_h->{'role'} 
#-#         );
#-#         push(@_return, [$_h->{'node'}, $_h->{'port'}, $_h->{'role'}] );
#-#      }
#-#      return(\@_return);
#-#   }
#-#}


#..........................................................................
# Get port list from object
# Use NACL::CS::ComponentState verify_fields for
# filtering the NACL port objects.
#
# Populate user reference list to be AoA
#    ([node,port], [], [], ...)
#
# Args 
#   aref    => ref to a list to populate with objects port indexing (required)
#   node    => <str>   node name (optional)
#   filter  => <ref>   ref to filter hash, same as NACL filter (optional)
#   unbound => <0|1>   port is not in a vlan,ifgrp or has an IF 
# Return 
#    wantarray ( count, error) : count
#

sub get_port_list {
   my $_obj = shift;
   my %_args = @_;
   my @_nodes = ();
   my ($_n, $_p, $_keep);

   # check given array to populate
   if ( !defined($_args{'aref'}) ) {
      $ERROR = 'arg aref not defined';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   if ( ref($_args{'aref'}) ne 'ARRAY') {
      $ERROR = 'arg aref not arrary reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   # init the port list that is to be populated
   @{$_args{'aref'}} = ();
   $ERROR = 'noError';  

   # XXX do want to keep this, the filtering will cover it
   # get the nodes to query for from arg
   #   node => n1,n2,n3
   #   node => n1
   if ( $_args{'node'} ) {
      if    ( $_args{'node'} =~ /,/)   { @_nodes = split(/,/, $_args{'node'}); }
      elsif ( $_args{'node'} =~ /\w+/) { push(@_nodes, $_args{'node'}); }
   }
   else {@_nodes = $_obj->get_node_list;}

   _role_ipspace_translate($_args{'filter'}) if defined($_args{'filter'});


   # loop through object to get ports
   $Log->comment("### Getting ports from object: " 
                . (defined($_args{'filter'})
                     ? 'apply filtering: ' . &_filter_str_comment($_args{'filter'})
                     : 'no filtering'
                 )
   );
   # Access all the port objects in the Topology object
   foreach $_n ( @_nodes ) {
      foreach $_p ( keys %{$_obj->{'PortInfo'}{$_n}} ) {
         $_keep = 0;
         # apply filter
         if ( defined($_args{'filter'}) ) {
            eval{
               $_obj->{'PortInfo'}{$_n}{$_p}{'port'}->verify_fields(%{$_args{'filter'}});
            };
            if ($@) { $_keep = 0; }
            else    { $_keep++; }
         }
         # node filter
         else {
            $_keep++;
         }
         $Log->comment("### get_port_list checking: {$_n}{$_p} object: "
            . $_obj->{'PortInfo'}{$_n}{$_p}{'port'}->{'port'} . " "
            . $_obj->{'PortInfo'}{$_n}{$_p}{'port'}->{'role'} . ": "
            . ($_keep ? 'matched object ' : 'unmatched object ')
            . ( defined($_args{'filter'})
               ? _filter_match_str_comment($_args{'filter'}, $_obj->{'PortInfo'}{$_n}{$_p}{'port'})
               : 'not filtered'
            )
         );

         if ( $_keep ) {
            push(@{$_args{'aref'}}, [$_n, $_p]);
         }
      }
   }
   return( wantarray ? (scalar(@{$_args{'aref'}}), $ERROR) :   scalar(@{$_args{'aref'}}) ); 
}
#-----------------------------------------------------------------------------
# Get port name list for a node
#
# Args
#    $_[0] = node name
#
# Return
#   (name, name, name,....)
#
sub get_portname_list {

   my $_obj= shift;
   my ( $_p); 
   my @_ports = ();
   if ( defined($_obj->{'PortInfo'}{$_[0]}) ) {
      foreach $_p ( sort keys %{$_obj->{'PortInfo'}{$_[0]}} ) {
         push(@_ports, $_p);
      }
      return(@_ports);
   }
   else {
      $ERROR = "get_portname_list: node: $_[0] not defined";
      return(@_ports);
   }
}

#-----------------------------------------------------------------------------
# Return hash of the stats for given node::port
# Not giving ref to hash, since $_obj->{'PortInfo'} can be cleared
#
# Args
#    $_[0] = node name
#    $_[1] = port name
#
# Return 
#  wantarray ? ( ref, error) : ref
#
sub get_port_stats {

   my $_obj= shift;

   if ( defined($_obj->{'PortInfo'}{$_[0]}{$_[1]}{'_stat'}) ) {
      return( wantarray ? ($_obj->{'PortInfo'}{$_[0]}{$_[1]}{'_stat'}, '' )
                        : $_obj->{'PortInfo'}{$_[0]}{$_[1]}{'_stat'} 

      ); 
   }
   else {
      $ERROR = "get_port_stats: node: $_[0] port: $_[1] stats hash not defined";
      return( wantarray ? (undef, $ERROR) : undef ); 
   }
}


#-----------------------------------------------------------------------------
# Get port object given the indexing as returned by get_port_list()
# Args
#  $_[0] = node name
#  $_[1] = port name 
#
# Return 
#    wantarray ? (ref, err) : ref
#
sub get_port_object {

   my $_obj = shift;
   $ERROR = '';
   if ( defined($_obj->{'PortInfo'}{$_[0]}{$_[1]}{'port'}) ) {
      return( wantarray ? ($_obj->{'PortInfo'}{$_[0]}{$_[1]}{'port'}, $ERROR) 
                        : $_obj->{'PortInfo'}{$_[0]}{$_[1]}{'port'}  
      );
   }
   else {
      $ERROR = "object port info not defined for node:$_[0] port:$_[1]";
      return( wantarray ? (undef, $ERROR) : undef);     
   }
}


#-----------------------------------------------------------------------------
# Get the ports profile
#  real - is not an ifgrp or vlan port
#  bound - is vlan or ifgrp OR has IFs OR in vlan or ifgrp
#  member - is in vlan or in ifgrp
#
# Args
#  $_[0] = node name
#  $_[1] = port name 
#
# Return 
#  (profile, real, bound, member, ip_version)
#  if profile is undef, then have to get the error with get_error method
#
sub get_port_profile {

   my $_obj = shift;

   my $_profile = undef;
   my $_real    = 0;
   my $_bound   = 0;
   my $_member  = 0;
   my $_ip_ver  = undef;
   my ($_ref, $_err, $_if, $_inet4, $_inet6 );

   # get ref to port stats
   ($_ref, $ERROR)  = $_obj->get_port_stats($_[0], $_[1]);

   # profile the eport
   if ( $_ref ) {
      $_profile = 1;
      if ( $_ref->{'isaPort'} )    { $_real = 1; $_bound = 0;}
      if ( $_ref->{'isaIFgrp'} )   { $_real = 0; $_bound = 1; }
      if ( $_ref->{'isaVlan'} )    { $_real = 0; $_bound = 1; }
      if ( $_ref->{'if_count'} )   {             $_bound = 1; }
      if ( $_ref->{'inaIFgrp'} )   {             $_bound = 1; $_member = 1; }
      if ( $_ref->{'inaVlan'} )    {             $_bound = 1; $_member = 1; }
      # see what ip versions are on the LIFs
      if (  $_ref->{'if_count'} ) {
         $_inet4 = 0;
         $_inet6 = 0;
         foreach  $_if ( @{$_obj->{'PortInfo'}{$_[0]}{$_[1]}{'if'}} ) {
            if ( $_if->{'address_family'} =~ /4/ ) { $_inet4++; }
            if ( $_if->{'address_family'} =~ /6/ ) { $_inet6++; } 
         }
      }
      # set ipversion profile
      if    ( $_inet4 && $_inet6 )  { $_ip_ver = 'mixed'; }
      elsif ( $_inet4 && !$_inet6 ) { $_ip_ver = 'inet4'; }
      elsif ( !$_inet4 && $_inet6 ) { $_ip_ver = 'inet6'; }
      else                          { $_ip_ver = 'unknown'; }
 
   }

   return($_profile, $_real, $_bound, $_member, $_ip_ver);
}

#-----------------------------------------------------------------------------
# Get port profile hash
# Wrapper around get_port_profile to return as hash
#
# Args
#  $_[0] = node name
#  $_[1] = port name 
#
# Return 
#  hash{'profile'}
#      {'real'}
#      {'bound'}
#      {'member'}
#      {'ip_version'}
#
sub get_port_profile_hash {

   my $_obj = shift;
   my @_a = $_obj->get_port_profile($_[0], $_[1]);

   my %_h = (
      'profile' => $_a[0],
      'real'    => $_a[1],
      'bound'   => $_a[2],
      'member'  => $_a[3],
      'ip_ver'  => $_a[4],
   );
   
   return(%_h);

}




#-----------------------------------------------------------------------------
# Get interface list, this is under ports so allow for filtering of ports 
# Use NACL::CS::ComponentState verify_fields for
# filtering the NACL port objects.
# Make use of get_port_list method
#
# Populate user reference list to be AoA
#    ([node,port, index], [], [], ...)
#
# Args 
#   aref        => ref to a list to populate with objects interface indexing (required)
#   node        => <str> node name (optional)
#   filter      => <ref> ref to filter hash for IF, same as NACL filtering (optional)
#   port_filter => <ref> ref to filter hash for ports, same as NACL filtering (optional)
#   port_list   => <ref> ref to port list ([node,port], [],..) ((optional)
#                        same as list made with get_port_list method 
#    Note 
#    - content of node arg (if given) will also be used with port_filter (if given)
#      both are independent
#    - port_filter will not be used if port_list is given

# Return 
#    wantarray ( count, error) : count
#

sub get_interface_list {
   my $_obj = shift;
   my %_args = @_;
   my @_nodes = ();
   my @_ports = ();
   my ($_n, $_p, $_i, $_keep, 
       $_ok, $_error, $_aref, $_if,
   );
   

   # check given array to populate
   if ( !defined($_args{'aref'}) ) {
      $ERROR = 'arg aref not defined';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   if ( ref($_args{'aref'}) ne 'ARRAY') {
      $ERROR = 'arg aref not arrary reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   # filtering for ports
   if ( defined($_args{'port_filter'}) && ref($_args{'port_filter'}) ne 'HASH' ) {
      $ERROR = 'arg port_filter defined and not hash reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   if ( defined($_args{'port_list'}) && ref($_args{'port_list'}) ne 'ARRAY' ) {
      $ERROR = 'arg port_list defined and not an array reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }

   # init the port list that is to be populated
   @{$_args{'aref'}} = ();
   $ERROR = 'noError';

   # XXX do want to keep this, the filtering will cover it
   # get the nodes to query for from arg
   #   node => n1,n2,n3
   #   node => n1
   if ( $_args{'node'} ) {
      if    ( $_args{'node'} =~ /,/)   { @_nodes = split(/,/, $_args{'node'}); }
      elsif ( $_args{'node'} =~ /\w+/) { push(@_nodes, $_args{'node'}); }
   }
   else {@_nodes = $_obj->get_node_list;}

   # see what port list to use
   # user already got port list or made their own
   if (defined($_args{'port_list'})) {
      @_ports = @{$_args{'port_list'}};
   }
   else {
      ($_ok, $_error) = $_obj->get_port_list(
         'aref'   => \@_ports,
         'nodes'  => join(',', @_nodes),
         defined($_args{'port_filter'}) ? ('filter' => $_args{'port_filter'}) : (),   
      );
   }
   # loop through object to get IFs
   $Log->comment("### Getting interfaces from object: "
                . (defined($_args{'filter'})
                     ? 'apply filtering: ' . &_filter_str_comment($_args{'filter'})
                     : 'no filtering'
                 )
   );
   # loop thru the ports
   foreach $_aref ( @_ports ) {
      $Log->comment("### get_interface_list checking: port [$_aref->[0] $_aref->[1]]");
      # loop thru IF for this port
      $_i    = 0;
      foreach $_if ( @{$_obj->{'PortInfo'}{$_aref->[0]}{$_aref->[1]}{'if'}} ) {
         $_keep = 0;
         # apply filter
         if ( defined($_args{'filter'}) ) {
            eval{
               $_if->verify_fields(%{$_args{'filter'}});
            };
            if ($@) { $_keep = 0; }
            else    { $_keep++; }
         }
         # no filter
         else {
            $_keep++;
         }
         $Log->comment("### get_interface_list checking: IF {$_aref->[0]}{$_aref->[1]}{'if'}[$_i] object: "
            . $_obj->{'PortInfo'}{$_aref->[0]}{$_aref->[1]}{'if'}->[$_i]->{'lif'} . " "
            . $_obj->{'PortInfo'}{$_aref->[0]}{$_aref->[1]}{'if'}->[$_i]->{'address'} . ": "
            . ($_keep ? 'matched object ' : 'unmatched object ')
            . ( defined($_args{'filter'})
                 ? _filter_match_str_comment($_args{'filter'}, $_if)
                 : 'not filtered'
            )
         );
         # store [node, port, if_obj_index]
         if ( $_keep ) {
            push(@{$_args{'aref'}}, [$_aref->[0], $_aref->[1], $_i] );
         }
         $_i++;
      } 
   }

   return( wantarray ? (scalar(@{$_args{'aref'}}), $ERROR) :   scalar(@{$_args{'aref'}}) );

}
#-----------------------------------------------------------------------------
# Get interface object given the indexing as returned by get_port_list()
# Args
#  $_[0] = node name
#  $_[1] = port name
#  $_[2] = ports IF list index 
#
# Return 
#    wantarray ? (ref, err) : ref
#
sub get_interface_object {

   my $_obj = shift;
   $ERROR = '';
   if ( defined($_obj->{'PortInfo'}{$_[0]}{$_[1]}{'if'}->[$_[2]]) )  {
      return( wantarray ? ($_obj->{'PortInfo'}{$_[0]}{$_[1]}{'if'}->[$_[2]], $ERROR)
                        : $_obj->{'PortInfo'}{$_[0]}{$_[1]}{'if'}->[$_[2]]
      );
   }
   else {
      $ERROR = "object interface info not defined for node:$_[0] port:$_[1] IF index $_[2]";
      return( wantarray ? (undef, $ERROR) : undef);
   }
}

#-----------------------------------------------------------------------------
# Get the VLAN list
# Use NACL::CS::ComponentState verify_fields for
# filtering the NACL port objects.
# Populate user reference list to be AoA
#    ([node,port], [], [], ...)
#
# Args 
#   aref   => ref to a list to populate with objects port indexing (required)
#   filter => <ref>   ref to filter hash, same as NACL filter (optional)
#
# Return 
#    wantarray ( count, error) : count
#
sub get_vlan_list {

   my $_obj = shift;
   my %_args = @_;
   my ($_id, $_e, $_keep, $_ref);


   # check given array to populate
   if ( !defined($_args{'aref'}) ) {
      $ERROR = 'arg aref not defined';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   if ( ref($_args{'aref'}) ne 'ARRAY') {
      $ERROR = 'arg aref not arrary reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   # init the port list that is to be populated
   @{$_args{'aref'}} = ();
   $ERROR = 'noError';


   # loop through object to get vlans
   $Log->comment("### Getting VLAN IDs from object: "
                . (defined($_args{'filter'})
                     ? 'apply filtering: ' . &_filter_str_comment($_args{'filter'})
                     : 'no filtering'
                 )
   );
   foreach $_id ( sort keys %{$_obj->{'VLAN'}} ) {
      $_keep = 0;
      $_e = 0;
      foreach $_ref ( @{$_obj->{'VLAN'}{$_id }} ) {
         if ( defined($_args{'filter'}) ) {
            eval{
               $_ref->verify_fields(%{$_args{'filter'}});
            };
            if ($@) { $_keep = 0; }
            else    { $_keep++; }
         }
         else {
            $_keep++;
         }
         $Log->comment("### get_vlan_list id=$_id "
                      ."node: $_ref->{'node'} port: $_ref->{'port'} "
                      . ($_keep ? 'matched object ' : 'unmatched object ')
                      . ( defined($_args{'filter'})
                           ? _filter_match_str_comment($_args{'filter'}, $_ref)
                           : 'not filtered'
                        )
         );
         if ( $_keep ) {
            push(@{$_args{'aref'}}, [$_id, $_e]);
         }
         $_e++;
      }
   }

   return( wantarray ? (scalar(@{$_args{'aref'}}), $ERROR) :   scalar(@{$_args{'aref'}}) );
}

#-----------------------------------------------------------------------------
# Get VLAN object
# Args
#  $_[0] = vlan-id
#  $_[1] = vlan list index 
# Return 
#    wantarray ? (ref, err) : ref
#

sub get_vlan_object {
   my $_obj = shift;
   $ERROR = '';
   if ( defined($_obj->{'VLAN'}{$_[0]}->[$_[1]]) ) {
      return( wantarray ? ( $_obj->{'VLAN'}{$_[0]}->[$_[1]], $ERROR)
                        : $_obj->{'VLAN'}{$_[0]}->[$_[1]]
      );
   }
   else {
      $ERROR = "object vlan info not defined for vlan-id: $_[0]  aref index: $_[1]";
      return( wantarray ? (undef, $ERROR) : undef);
   }

}

#-----------------------------------------------------------------------------
# Get Route list
# Given a ref to a list, populate with the key|indexes to the route objects
# Object keys ar VServer name and list index of route object
#    {VServer}{$_v}{'routes'}[$_c]
#
# Args 
#   aref    => ref to a list to populate with objects port indexing (required)
#   filter  => <ref>   ref to filter hash, same as NACL filter (optional)
# Return 
#    wantarray ( count, error) : count
#


sub get_route_list {
   my $_obj = shift;
   my %_args = @_;
   my ($_v, $_href, $_keep, $_c);
   $ERROR = 'noErr';

   # check given array to populate
   if ( !defined($_args{'aref'}) ) {
      $ERROR = 'arg aref not defined';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   if ( ref($_args{'aref'}) ne 'ARRAY') {
      $ERROR = 'arg aref not arrary reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   # init the port list that is to be populated
   @{$_args{'aref'}} = ();
   $ERROR = 'noError';


   # loop through object to get routes object from vserver tree
   $Log->comment("### Getting routes from object:"
                . (defined($_args{'filter'})
                     ? 'apply filtering: ' . &_filter_str_comment($_args{'filter'})
                     : 'no filtering'
                 )
   );

   foreach $_v ( keys %{$_obj->{'VServer'}} ) {
      $_c = 0;
      foreach $_href ( @{$_obj->{'VServer'}{$_v}{'routes'}} ) {
         $_keep = 0;
         if ( defined($_args{'filter'}) ) {
            eval{
               $_href->verify_fields(%{$_args{'filter'}});
            };
            if ($@) { $_keep = 0; }
            else    { $_keep++; }
         }
         else {
            $_keep++;
         }
         $Log->comment("### get_route_list checking: {VServer}{$_v}{'routes'}[$_c] "
                      ."vserver: " . $_href->{'vserver'} . " "
                      ."dest: " . $_href->{'destination'} . " "
                      . ($_keep ? 'matched object ' : 'unmatched object ')
                      . ( defined($_args{'filter'})
                          ? _filter_match_str_comment($_args{'filter'}, $_href)
                          : 'not filtered'
                       )
         );
         if ( $_keep ) {
            push(@{$_args{'aref'}}, [$_v, $_c]);
         }

         $_c++; 
      } 
   }
   return( wantarray ? (scalar(@{$_args{'aref'}}), $ERROR) :   scalar(@{$_args{'aref'}}) );
}

#-----------------------------------------------------------------------------
# Get route object  given the indexing as returned by get_route_list()
#
# Args
#  $_[0] = vserver name
#  $_[1] = route array index 
#
# Return 
#    wantarray ? (ref, err) : ref
#
sub get_route_object {
   my $_obj = shift;
   $ERROR = '';

   if ( defined($_obj->{'VServer'}{$_[0]}{'routes'}[$_[1]]) ) {
      return( wantarray ? ( $_obj->{'VServer'}{$_[0]}{'routes'}[$_[1]], $ERROR)
                        : $_obj->{'VServer'}{$_[0]}{'routes'}[$_[1]]
      );
   }
   else {
      $ERROR = "object vserver route info not defined for vserver: $_[0] index: $_[1]";
      return( wantarray ? (undef, $ERROR) : undef);
   }
}

#-----------------------------------------------------------------------------
# Match ngsh route object to systemshell route entry
# Args
#   route_ref  => <ref>  $_[0] ref to nacl route object
#   vserver_id => <int>  
#

sub find_systemshell_route {

   my $_obj = shift;
   my ($_node, $_r, $_c,
       $_dest, $_inet, $_netif,
       $_matched 
   );
   my %_return = ();
   $ERROR = '';

   my %_arg = @_;

   # make sure given a route    
   if ( !defined($_arg{'route_ref'}) ) {
      $ERROR = "find_systemshell_route: arg route_ref not defined";
      return ( wantarray ? (undef, $ERROR) : undef);
   }
   

   $Log->comment("find_systemshell_route: given route object: " . Dumper $_arg{'route_ref'} );

   # Normalize to how data presented in systemshell
   #
   # normalize destination
   # ngsh gives in ip addr syntax
   # systemshell uses string 'default' 
   if    ( $_arg{'route_ref'}->{'destination'} eq '0.0.0.0/0' ) { $_dest = 'default'; }
   elsif ( $_arg{'route_ref'}->{'destination'} eq '::/0' )      { $_dest = 'default'; }
   else                                            { $_dest =  $_arg{'route_ref'}->{'destination'}; }

   # normalize address family
   if    ( $_arg{'route_ref'}->{'address_family'} eq 'ipv4' ) { $_inet = 'Internet'; }
   elsif ( $_arg{'route_ref'}->{'address_family'} eq 'ipv6' ) { $_inet = 'Internet6'; }
   else {
      $ERROR = "Unknown address_family in NACL route object: $_arg{'route_ref'}->{'address_family'}";
      return ( wantarray ? (undef, $ERROR ) : undef);
   }

   # normalize systemshell Netif
   # routing-groups
   if ( defined($_arg{'route_ref'}->{'routing_group'}) ) {
      $_netif = sprintf("%s_%s", $_arg{'route_ref'}->{'vserver'}, $_arg{'route_ref'}->{'routing_group'}); 
   }
   # ipspace 
   elsif ( defined($_arg{'route_ref'}->{'ipspace'})  ) {
      if ( $_arg{'vserver_id'} ) {
         $_netif = $_arg{'vserver_id'};
      }
      else {
         $ERROR = "find_systemshell_route: ipspace route needs vserver_id arg";
         return ( wantarray ? (undef, $ERROR ) : undef);
      }
   }
   # assert
   else {
      $ERROR = "find_systemshell_route: unknown route object";
      return ( wantarray ? (undef, $ERROR ) : undef);
   }

   # loop through systemshell route table and match
   $Log->comment("find_systemshell_route: match values "
                ."inet: $_inet dest: $_dest " 
                ."gw: $_arg{'route_ref'}->{'gateway'} netif: $_netif"
   );
   $_matched = 0;
   foreach $_node ( sort keys %{$_obj->{'Shell'}} ) {
      $_c = 0;
      foreach  $_r ( @{$_obj->{'Shell'}{$_node}{'systemshell'}{'route'}{$_inet}} ) { 
         if (     $_dest                          eq $_r->{'Destination'} 
               && $_arg{'route_ref'}->{'gateway'} eq $_r->{'Gateway'}
               && $_netif                         eq $_r->{'Netif'} ) {
            push( @{$_return{$_node}}, $_r );
            $Log->comment("find_systemshell_route: matched route entry on $_node entry $_c " . Dumper $_r);
            last; 
         }
         $_c++;
      }
   }

   return ( wantarray ? ( \%_return, undef ) : \%_return);
}


#-----------------------------------------------------------------------------
# Match ngsh route object to nodescope (7M) route entry
#
# route_ref  => <ref>  $_[0] ref to nacl route object
# vserver    => <str>
#

sub find_nodescope_route {

   my $_obj = shift;
   my($_dest, $_inet, $_node,
      $_matched, $_c, $_display_it, 
      $_href, $_r,
      $_route_grp
   );
   my %_return = ();
   $ERROR = '';

   my %_arg = @_;

   # make sure given a route    
   if ( !defined($_arg{'route_ref'}) ) {
      $ERROR = "find_nodescope_route: arg route_ref not defined";
      return ( wantarray ? (undef, $ERROR) : undef);
   }
   
   # Normalize to how data presented in systemshell
   #
   # normalize destination
   # ngsh gives in ip addr syntax
   # systemshell uses string 'default' 
   if    ( $_arg{'route_ref'}->{'destination'} eq '0.0.0.0/0' ) { $_dest = 'default'; }
   elsif ( $_arg{'route_ref'}->{'destination'} eq '::/0' )      { $_dest = 'default'; }
   else                                            { $_dest =  $_arg{'route_ref'}->{'destination'}; }

   # normalize address family
   if    ( $_arg{'route_ref'}->{'address_family'} eq 'ipv4' ) { $_inet = 'internet'; }
   elsif ( $_arg{'route_ref'}->{'address_family'} eq 'ipv6' ) { $_inet = 'internet6'; }
   else {
      $ERROR = "Unknown address_family in NACL route object: $_arg{'route_ref'}->{'address_family'}";
      return ( wantarray ? (undef, $ERROR ) : undef);
   }

   # normalize for nodescope route entry route_group
   #
   # non-ipspace
   if ( defined($_arg{'route_ref'}->{'routing_group'}) ) {
      $_route_grp = sprintf("%s_%s", 
         $_arg{'route_ref'}->{'vserver'} || $_arg{'route_ref'}->{'server'}, 
         $_arg{'route_ref'}->{'routing_group'} 
      );
   }
   # ipspace  XXX need to be vservername + subnet 
   elsif (defined($_arg{'route_ref'}->{'ipspace'}) ) {
      $_route_grp = sprintf("%s_%s",
         $_arg{'route_ref'}->{'vsid'},
         $_arg{'route_ref'}->{'destination'},
      );
   }
   # assert
   else {
      $ERROR = "find_nodescope_route: unknown route object";
      return ( wantarray ? (undef, $ERROR ) : undef);
   }

   $Log->comment("find_nodescope_route: vserver: $_arg{'vserver'} "
                ."addr_family: $_arg{'route_ref'}->{'address_family'} [$_inet] "
                ."dest: $_arg{'route_ref'}->{'destination'} [$_dest] "
                ."gw: $_arg{'route_ref'}->{'gateway'} "
                ."route_group: [$_route_grp] " 
                # ."given route object: " . Dumper $_arg{'route_ref'} 
   );


   $_matched = 0;
   foreach $_node ( sort keys %{$_obj->{'Shell'}} ) {
      $_c = 0;
      foreach  $_href ( @{$_obj->{'Shell'}{$_node}{'nodescope'}{'route'}} ) {
         foreach $_r ( @{$_href->{$_inet}} ) {
            if (    $_dest                           eq $_r->{'Destination'}
                 && $_arg{'route_ref'}->{'gateway'}  eq $_r->{'Gateway'}
                 && $_route_grp                      eq $_r->{'routing_group'} ) {
               push( @{$_return{$_node}}, $_r );
               $Log->comment("find_nodescope_route: matched route entry on $_node entry $_c ");
               $Log->comment("find_nodescope_route: route = "
                  ."dest: $_r->{'Destination'} "
                  ."gw: $_r->{'Gateway'} "
                  ."IF: $_r->{'Interface'} "
                  ."rtgrp: $_r->{'routing_group'} "
               );
               last;
            }
            $_c++;
         }
      }
   }


   sub __display_it {
      $Log->comment("find_nodescope_route: route = "
               ."dest: $_r->{'Destination'} "
               ."gw: $_r->{'Gateway'} "
               ."IF: $_r->{'Interface'} "
               ."rtgrp: $_r->{'routing_group'} "
      );
   }

}






#-----------------------------------------------------------------------------
#
# Run the VIFMgr TriageTool
# - links
#   http://wikid.netapp.com/w/SCON/SierraNevada/IpSpaces/SN_tasklist/UI-report-node-config
#   http://wikid.netapp.com/w/SCON/SierraNevada/IpSpaces/SN_tasklist/VM-check-node-config
#   http://wikid.netapp.com/w/SCON/FullSteam/TriageTool
#
# - check /mroot/etc/mlog/vifmgr.log for 'Starting debug network report' string
#    - if exist get timestamp of the last start.
#    - ending seen at 'Completed debug network report'
# - in ngsh run 'debug network report -node local'
#    - will give a job id and run the app in the background
# - wait for job id to finish (success) 
#   job show -cat "Networking Diagnostic"
#   or get status of job id
# - parse vifmgr.log for 'vifmgr_debug_report'
#   The string to grep for is "Inconsistency found by vifmgr_debug_report" in vifmgr.log
# - if messages return as errors 
#
# Populate object
#  @{$_obj->{'_TriageTool'}{'msg'}}          = (msg1, msg2, ...)
#                          {'line_count'}    = <int>  lines counted in vifmgr.log
#                          {'matched_count'} = <int>  triage-tools lines counted in vifmgr.log
 
# Args
#   'node' => <str>   optional, only do this node if given
# 
# Return
#   wantarray ? ( <matched>, <block>, <lines>) : matched
#
#        <matched> int, number of lines matched indicating a problem
#        <block> 0|1,  true if start and end delimiter lines seen
#        <line> int, number of lines checked in vifmgr.log
#
sub vifmgr_triage_tool {

   my $_obj = shift;
   my ($_node, $_resp, $_raw,
      $_nodename, $_job_id, $_job, $_job_state,
      $_apiset, $_flag, $_sleep,
      $_scp_aref, $_line, $_lc, $_mc, $_hs, $_ds, $_epoch,
      $_start, $_end,
      %_strings,
   );
   $ERROR = '';

   my %_arg = @_;

   # hash to hold settings to control exiting job watch loop
   my %_exit_loop = (
        'Initial'    => 0,        'Queued'     => 0,
        'Running'    => 0,        'Waiting'    => 0,
        'Pausing'    => 0,        'Paused'     => 1,
        'Quitting'   => 0,        'Success'    => 1,
        'Failure'    => 1,        'Reschedule' => 1,
        'Error'      => 1,        'Quit'       => 1,
        'Dead'       => 1,        'Unknown'    => 1,
        'Restart'    => 1,        'Dormant'    => 1,
   );

   %{$_obj->{'_TriageTool'}} = ();

   # store the last line of the mlog files
   $_obj->store_last_mlog_line;

   # vifmgr is cluster wide, so only need to do on one filer
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = $_node->hostrec->{'hostname'};

      if ( defined($_arg{'node'}) && $_arg{'node'} ne $_nodename) {
         $Log->comment("vifmgr_triage_tool: skip running on node $_nodename");
         next;
      }
      $Log->comment("vifmgr_triage_tool: running on node $_nodename");

      # use apiset to kick off vifmgr triage tool
      $_apiset = $_node->apiset( 'interface' => 'CLI' );
      #$_raw =  $_apiset->execute_raw_command(
      #   'command' => 'debug network report -node local',
      #);
      $_resp = $_apiset->debug_network_report('node'=> 'local');
      $_raw = $_resp->get_raw_output();
      $Log->comment("vifmgr_triage_tool: debug_network_report output: $_raw");
      # parse the jobid
      if (  $_raw =~ /\[Job\s+(\d+)\]/ ) {
         $_job_id = $1;
         $Log->comment("vifmgr_triage_tool: parsed job id: $_job_id");
         $_flag  = 0;
         $_sleep = 1;
         sleep 2;
         # Watch the vifmgr triage tool job
         # check job status, set flag when job state = success to exit loop
         $Log->comment("vifmgr_triage_tool: monitor triage-tool job ...");
         do  {
            $_job = NACL::C::Job->new('command_interface' => $_node, 'id' => $_job_id);
            $_job_state = $_job->state();
            #delete($_job_state->{'command_interface'});
            #$Log->comment("job state: " . Dumper $_job_state);
            $Log->comment("vifmgr_triage_tool: job watch loop "
                            ."state: $_job_state->{'state'} "
                            ."queuetime: $_job_state->{'queuetime'} "
                            ."starttime: $_job_state->{'starttime'} "
                            ."schedule: $_job_state->{'schedule'} "

            );
            # job is done, success, get vifmgr.log file
            # there does seem to be an issue where:
            #   - see realtime tail -f vifmgr.log and the triage tool lines
            #   - the scp'd file does not contain the line
            #   - may be the scp is happen quicker than messages getting dumped into file
            # so putting in a sleep
            if ( $_job_state->{'state'} eq 'Success') {
               $_flag++;
               $Log->comment("vifmgr_triage_tool: job watch state=Success get files "
                             ."in $options->{'TOPOLOGY_MLOG_SCP_SLEEP'} sec"
               );
               sleep $options->{'TOPOLOGY_MLOG_SCP_SLEEP'};
               $_scp_aref = &scp_mlog_file($_node, 'vifmgr.log');
            }
            # job is in a state that we should abort watching the status 
            elsif ( $_exit_loop{$_job_state->{'state'}} ) { 
               $_flag++;
            }
            # job stat still acceptable, keep looping
            else {
               sleep $_sleep++;
               $_flag++ if $_sleep > 10;   # infinite loop protection
            }
         } until $_flag > 0;

         # done watching
         $Log->comment("vifmgr_triage_tool: job watch loop exited "
                      ."state: $_job_state->{'state'}"
         );
         last;
      }
      else {
         $ERROR = "job id not parsed from instantiating the triage tool";
         $Log->comment("vifmgr_triage_tool: parsed job id: failed");
      }
   }
   #
   # Get time stats for known last entry before triage tool was ran
   #
   $_line = \$_obj->{'_Config'}{'mlog'}{$_nodename}{'vifmgr.log'};
   if ( !defined($_line) ) {
      $ERROR = "last line not defined for vifmgr.log for node $_nodename";
      return(wantarray ? (0,0,0) : 0);
   }

   $_strings{'start'}{'hex_str'}  =  &parse_logmsg_hexstamp($$_line);
   @{$_strings{'start'}{'hex_decode'}} = &decode_logmsg_hex($_strings{'start'}{'hex_str'});
   ($_strings{'start'}{'date_str'}, $_strings{'start'}{'epoch'}) = &parse_logmsg_datestr($$_line);   

   $Log->comment("vifmgr_triage_tool: last vifmgr.log: " 
                ."hex: $_strings{'start'}{'hex_str'} "
                ."date: $_strings{'start'}{'date_str'} "
                ."epoch: $_strings{'start'}{'epoch'} "
   );

   # init line and message counters
   $_lc = 0;
   $_mc = 0;
   $_start = 0;
   $_end   = 0;
   # parse the vifmgr.log file scp'd to LOGDIR
   if ( scalar(@{$_scp_aref}) ) {
      $Log->comment("vifmgr_triage_tool: parse vifmgr file $_scp_aref->[0]");
      open(FH, shift(@{$_scp_aref}) );
      while ($_line =<FH>) {
         $_flag = 0;
         %{$_strings{'this'}} = ();
         # get timstamp info from current line
         ##$_strings{'this'}{'hex_str'}       = &parse_logmsg_hexstamp($_line);
         ##@{$_strings{'this'}{'hex_decode'}} = &decode_logmsg_hex($_strings{'this'}{'hex_str'});
         ($_strings{'this'}{'date_str'}, $_strings{'this'}{'epoch'}) = &parse_logmsg_datestr($_line);

         # all triage tool lines after known start point
         if ( $_strings{'this'}{'epoch'} > $_strings{'start'}{'epoch'} ) {
            # set flag to display start/end strings
            if ( $_line =~ /Starting debug network report/)   { $_flag++; $_start = $.;}
            if ( $_line =~ /Completed debug network report/)  { $_flag++; $_end   = $.;}
            # found triage_tool line,
            # flag to print it, incr match counter, store it in object
            if ( $_line =~ /Inconsistency found by vifmgr_debug_report/) { 
               $_mc++; 
               $_flag++;
               chomp($_line);
               push(@{$_obj->{'_TriageTool'}{'msg'}}, $_line);
            }
         }
         if ( $_flag ) { $Log->comment("vifmgr_triage_tool: linum $.  $_line"); }
         $_lc++;
      }
      close(FH);
      $Log->comment("vifmgr_triage_tool: file parse count "
                   ."lines: $_lc matched: $_mc "
                   ."start: $_start end: $_end "
      );
   } 
   $_obj->{'_TriageTool'}{'line_count'} = $_lc;
   $_obj->{'_TriageTool'}{'matched_count'} = $_mc;

   # return based on what is seen 
   if ( !$_start ) {
      $ERROR = "vifmgr_triage_tool: start line never seen in logfile "
               ."after $_strings{'start'}{'date_str'}";
      return(wantarray ? (0, $_start, $_lc) : 0);
   }
   elsif ( !$_end ) {
      $ERROR = "vifmgr_triage_tool: end line never seen in logfile "
              ."after $_strings{'start'}{'date_str'}";
      return(wantarray ? (0, $_end, $_lc) : 0);
   }
   else {
      return( wantarray ? ($_mc, $_end, $_lc) : $_mc);
   }
}

#-----------------------------------------------------------------------------
# Method to get the stored triage tool messages
#
#
sub vifmgr_triage_tool_messages {
   my $_obj = shift;
   return($_obj->{'_TriageTool'});
}  


#-----------------------------------------------------------------------------
# Method to get the last timestamp of /mroot/etc/mlog files
#
sub store_last_mlog_line {

   my $_obj = shift;
   my $_filer_dir   = '/mroot/etc/mlog';
   my $_cmd         = sprintf("tail -1 %s/*.log", $_filer_dir);
   my ( $_node, $_systemshell_apiset, 
        $_out, $_nodename, 
        $_line, $_filename, $_hex_str,
   );
   my $_r=0;
   &switch_logging('Store last mlog line');


   # loop through each node
   foreach $_node ( @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = \$_node->hostrec->{'hostname'};
      $_systemshell_apiset = $_node->apiset(category  => 'Node',
                                            interface => 'CLI',
                                            set       => 'Systemshell',
      );
      $_out = $_systemshell_apiset->execute_raw_command(
            'command' => $_cmd,
      );

      # parse the output and store it in object
      foreach $_line ( split(/\n/, $_out) ) {
         next if $_line =~ /^$/;
         next if $_line =~ /^\s+$/;
         $Log->comment("store_last_mlog_line: $_line");
         chomp($_line);
         if ( $_line =~ /==>\s+(.+)\s+<==/ ) {
            $_filename = $1;
            $_filename =~ s/$_filer_dir\///;
            $Log->comment("store_last_mlog_line: $_filename");
         }
         # must be a line from a log get hex stamp
         $_obj->{'_Config'}{'mlog'}{$$_nodename}{$_filename} = $_line;
         $Log->comment("store_last_mlog_line: node: $$_nodename "
                      ."file: $_filename "
                      ."$_obj->{'_Config'}{'mlog'}{$$_nodename}{$_filename}"
         );
         $_r++;
      }
   }
   $Log->reset_path();
   $_r;
}

#-----------------------------------------------------------------------------
# Given an API, cli command with underscore, get a list of 
# accetpable arguments to the options
#
# 
# Args
#   node => <ref>   ref to node object  (optional)
#   api  => <str>   api, cli command with underscores (required)
#   
#
sub get_cli_args {

   my $_obj = shift;
   my %_args = @_;
   my ($cli_apiset, $_href, $_f, $_str, $_parsed, @_tokens);

   # if no node, then use 1st one in object
   if ( !defined($_args{'node'}) ) { $_args{'node'} = $_obj->{'input'}{'nacl_nodes'}[0]; } 

   # check api
   if (!defined($_args{'api'}) ) {
      $ERROR = 'arg  api required';
      return( wantarray ? (undef, $ERROR) : undef );
   }  


   $cli_apiset = $_args{'node'}->apiset( interface => 'CLI' );
   $_href = $cli_apiset->cli_command_helper('api' => $_args{'api'});
   #$Log->comment("CLI_APIset: " . Dumper $_href);

   foreach $_f ( keys %{$_href->{'fields'}} ) {
      if ( defined($_href->{'fields'}{$_f}{'description'}) ) {
         $_str = join(' ', @{$_href->{'fields'}{$_f}{'description'}});
         @_tokens = ();
         if ($_str =~ /possible value.s.\sare\s+(.+)/ ) {
            $_parsed = $1;
            @_tokens = split(/,/, $_parsed);
            $Log->comment("API: $_args{'api'} Option: $_f Args: " . join(' ', @_tokens) );
            @{$_obj->{'Args'}{$_args{'api'}}{$_f}} = @_tokens;
         }
      }
   }

   $Log->comment("CLI static Args" . Dumper $_obj->{'Args'});

   return( wantarray ? (1, $ERROR) : 1 );

}



#-----------------------------------------------------------------------------
# Local object filter routine
# Using NACL::CS::ComponentState->_apply_filter instead
# Problem is this will remove objects.
# Keeping this around just in case want to use it
#
# Args
#  object = <ref>  ref to nacl object
#  filter = <ref> ref to filter hash
#
sub filter_object {

   my %_a = @_;
   my $_k;
   my $_r = 0;
   $Log->comment("Filter object: " . ref($_a{'object'}) );
   foreach $_k ( keys %{$_a{'filter'}} ) {
      $Log->comment("filter for key: [$_k] value: [$_a{'filter'}->{$_k}] ");
      
      if ( $_a{'object'}->{$_k} =~ /$_a{'filter'}->{$_k}/) {
         $Log->comment("filter key: [$_k] =~ [$_a{'filter'}->{$_k}] found");
         $_r++;
         last; 
      }
   }
   $_r;
}


#
#-----------------------------------------------------------------------------
# Extract IPSpace names from object
# Return array of names or scalar string
# 
#
sub extract_ipspace_names {
   my $_obj = shift;
   my $_i;
   my @_names =  sort keys %{$_obj->{'IPSpace'}};
   return( wantarray ? @_names :  join(', ', @_names) );
}
#
#-----------------------------------------------------------------------------
# Extract IPSpace names from object
# Return array of names or scalar string
# 
#
sub extract_vserver_names {
   my $_obj = shift;
   my $_i;
   my @_names =  sort keys %{$_obj->{'VServer'}};
   return( wantarray ? @_names :  join(', ', @_names) );
}


#..........................................................................
# Get vserver list from object
# Use NACL::CS::ComponentState verify_fields for
# filtering the NACL port objects.
#
# Populate user reference list to be AoA
#    (vserver, vserver, ... )
#
# Args 
#   aref    => ref to a list to populate with objects port indexing (required)
#   filter  => <ref>   ref to filter hash, same as NACL filter (optional)
#
sub get_vserver_list {

   my $_obj = shift;
   my %_args = @_;
   my ($_v, $_keep, $_k);

   # check given array to populate
   if ( !defined($_args{'aref'}) ) {
      $ERROR = 'arg aref not defined';
      return(wantarray ? (undef, $ERROR) : undef);
   }
   if ( ref($_args{'aref'}) ne 'ARRAY') {
      $ERROR = 'arg aref not arrary reference';
      return(wantarray ? (undef, $ERROR) : undef);
   }

   # init the port list that is to be populated
   @{$_args{'aref'}} = ();
   $ERROR = 'noError';

   # loop through object to get vserver
   $Log->comment("### Getting vserver from object: "
                . (defined($_args{'filter'}) 
                     ? 'apply filtering: ' . &_filter_str_comment($_args{'filter'}) 
                     : 'no filtering'
                 )
   );

   # Access all the port objects in the Vservers object
   foreach $_v  ( sort keys %{$_obj->{'VServer'}} ) {
      $_keep = 0;
      if ( defined($_args{'filter'}) ) {
            eval{
               $_obj->{'VServer'}{$_v}{'vserver'}->verify_fields(%{$_args{'filter'}});
            };
            if ($@) { $_keep = 0; }
            else    { $_keep++; }
      }
      # no filter
      else {
         $_keep++;
      }
      $Log->comment("### get_vserver_list checking vserver [$_v] object: "
         . ($_keep ? 'matched object ' : 'unmatched object ')
         . ( defined($_args{'filter'}) 
               ? _filter_match_str_comment($_args{'filter'}, $_obj->{'VServer'}{$_v}{'vserver'})
               : 'not filtered'
           )
      );

      if ( $_keep ) {
         push(@{$_args{'aref'}}, $_v);
      }
   }
   return( wantarray ? (scalar(@{$_args{'aref'}}), $ERROR) :   scalar(@{$_args{'aref'}}) );
}


#-----------------------------------------------------------------------------
# Get vserver name list for cluster
#
# Args
#  none
#
# Return
#   (name, name, name,....)
#
sub get_vservername_list {

   my $_obj= shift;
   my ( $_v);
   my @_vservers = ();
   if ( defined($_obj->{'VServer'}) ) {
      @_vservers = sort keys %{$_obj->{'VServer'}};
      return(@_vservers);
   }
   else {
      $ERROR = "get_vservername_list: empty list";
      return(@_vservers);
   }
}

#-----------------------------------------------------------------------------
# Get port object given the indexing as returned by get_port_list()
# Args
#  $_[0] = vserver name
#
# Return 
#    wantarray ? (ref, err) : ref
#
sub get_vserver_object {

   my $_obj = shift;
   $ERROR = '';
   if ( defined($_obj->{'VServer'}) ) {
      return( wantarray ? ($_obj->{'VServer'}{$_[0]}{'vserver'}, $ERROR)
                        : $_obj->{'VServer'}{$_[0]}{'vserver'}
      );
   }
   else {
      $ERROR = "object vserver info not defined for vserver: $_[0]";
      return( wantarray ? (undef, $ERROR) : undef);
   }
}


#
#-----------------------------------------------------------------------------
# Given a port, node and ipspace name
# Check if port is in given IPSpace on that node
#
# Args
#   'port'    => <port name>
#   'node'    => <node name>
#   'ipspace' => <ipspace name>
#
# Return
#   wantarray ? ( 1|0, error) : 1|0
#
sub is_port_in_ipspace {

   my $_obj = shift;
   my %_args = @_;
   my $_ret = 0;
   my ($_a, $_p, 
       $_ips_href, $ips_port_aref,
       $_ips_list_aref,
       $_href, $_aref,
   );


   foreach $_a ( 'port', 'node', 'ipspace' ) {
      if (!defined($_args{$_a})) {
         $ERROR = "required arg $_a not given";
         return(wantarray ? ( undef, $ERROR) : undef);
      } 
   }

   # check IPSpace known by object
   if ( !defined($_obj->{'IPSpace'}{$_args{'ipspace'}})) {
      $ERROR = "IPSpace [$_args{'ipspace'}] not known to object";
      return(wantarray ? ( undef, $ERROR) : undef);
   }

   # get port list from NACL::CS::NetworkIpspace object
   # port listing in this object is an aref
   # port name syntax is:  node:port   parsons-vsim1:e0a
   #
   $_ips_href     = $_obj->{'IPSpace'}{$_args{'ipspace'}}{'ipspace'};
   $ips_port_aref = $_obj->{'IPSpace'}{$_args{'ipspace'}}{'ipspace'}{'ports'};

   $Log->comment("is_port_in_ipspace: ipspace $_args{'ipspace'} "
                ."ports: " . sprintf("[%s] ", join(', ', @{$ips_port_aref}) )
   ); 

   # get port list from ipspace list
   # $_obj->{'Shell'}{<node>}{'nodescope'}{'ipspace_list'}{'parsed'} = AoH
   $_ips_list_aref = $_obj->{'Shell'}{$_args{'node'}}{'nodescope'}{'ipspace_list'}{'parsed'};
   foreach $_href ( @{$_ips_list_aref} ) {
      # find ipspace_list entry from match ipspace id
      if ( $_href->{'name'} eq $_ips_href->{'id'} ) {
         $Log->comment("is_port_in_ipspace: ipspace_list name [$_href->{'name'}] "
                      ."matched ipspace id [$_ips_href->{'id'}]"
         );
         foreach $_p ( @{$_href->{'interfaces'}} ) {
            if ($_args{'port'} eq $_p ) {
               $_ret = 1;
               $Log->comment("is_port_in_ipspace: ipspace_list interfcace [$_p] "
                            ."matched given port [$_args{'port'}]"
               );
               last; 
            }
         }
      }
      last if $_ret;
   }
   
   return(wantarray ? ( $_ret, $ERROR) : $_ret);
}


#-----------------------------------------------------------------------------
#
# Find the IPSpace containing given node, port
#
# Args 
#   'node'   => <node name>
#   'port'   => <port name>
#

sub find_port_ipspace {

   my $_obj = shift;
   my %_args = @_;
   my %_return = ();
   my $_found = 0;
   my ($_a, $_ips, $_hp, $_h, $_p,
       $_href, $_aref,
   );


   foreach $_a ( 'port', 'node' ) {
      if (!defined($_args{$_a})) {
         $ERROR = "required arg $_a not given";
         return(wantarray ? ( undef, $ERROR) : undef);
      }
   }

   # check port is known to object
   if ( !defined($_obj->{'PortInfo'}{$_args{'node'}}{$_args{'port'}})) {
      $ERROR = "Port $_args{'node'}:$_args{'port'} not known to object";
      return(wantarray ? ( undef, $ERROR) : undef);
   }

   # port info
   $_href = $_obj->{'PortInfo'}{$_args{'node'}}{$_args{'port'}}{'port'};
   %{$_return{'port'}} = (
     'ifgrp_mode' => $_href->ifgrp_mode(),
     'ifgrp_node' => $_href->ifgrp_node(),
     'ifgrp_port' => $_href->ifgrp_port(),
     'mtu'        => $_href->mtu(),
     'mac'        => $_href->mac(),
     'ipspace'    => $_href->ipspace(),
     'vlan_node' => $_href->vlan_node(),
     'vlan_port' => $_href->vlan_port(),
   );


   # get the ports IF
   foreach $_href ( @{$_obj->{'PortInfo'}{$_args{'node'}}{$_args{'port'}}{'if'}} ) {
      push(@{$_return{'interfaces'}},  
         { 'address' => $_href->address(),
           'netmask' => $_href->netmask(),
           'routing_group' => $_href->routing_group(),
           'ipspace'  => $_href->ipspace(),
           'vserver'  => $_href->vserver(),
           'vif' => $_href->vif()
         },
      );
   }

   # look in NACL::CS::NetworkIpspace object
   # ports are in a list with syntax <filer>:<port>  [parsons-vsim1:e0a]
   foreach $_ips ( sort keys %{$_obj->{'IPSpace'}} ) {
      foreach  $_hp ( @{$_obj->{'IPSpace'}{$_ips}{'ipspace'}{'ports'}} ) {
         if ( $_hp =~ /\w:\w\d+\w$/ ) {
            ($_h, $_p) = split(/:/, $_hp);
            if ( $_args{'node'} eq $_h && $_args{'port'} eq $_p ) {
               $_found++;
               $_return{'ipspace'} = $_obj->{'IPSpace'}{$_ips}{'ipspace'}->ipspace();
            } 
         }
         last if $_found;
      }
   }

   # get from ipspace list
   $_href = $_obj->{'Shell'}{$_args{'node'}}{'nodescope'}{'ipspace_list'}; 
   $_found = 0;
   foreach $_hp ( @{$_href->{'parsed'}} ) {
      foreach $_p ( @{$_hp->{'interfaces'}} ) {
         if ( $_args{'port'} eq $_p ) {
             $_found++;
             $_return{'ipspace_list_id'} = $_hp->{'name'};
             $_return{'ipspace_list_name'} = $_obj->{'Maps'}{'IPSpace'}{$_hp->{'name'}};
             
             last;
         }
      }
      last if $_found;
   }

   
   $Log->comment("find_port_ipspace IF: " . Dumper \%_return); 

   # XXX why returning this and not \%_return
   return(wantarray ? (1,'') : 1);
}


#-----------------------------------------------------------------------------
#
#                          Functions
#
#---------------------------------------------------------------------------
sub topo_ngsh_parser {

   my ($_l, $_c, $_i, @_val, %_h,  );
   my @_arr = ();
   my @_lines = split(/\n/, $_[0]);
   my @_col   = split(/X;X/, shift(@_lines));
   my @_desc  = split(/X;X/, shift(@_lines));

   foreach $_l ( @_lines ) {
      @_val =  split(/X;X/, $_l);
      %_h = ();
      $_i = 0;
      foreach $_c ( @_col ) {
         $_h{$_c} = $_val[$_i];
         $_i++;
      }
      push(@_arr, {%_h});
   }

   return \@_arr;
}


#-----------------------------------------------------------------------------
#
# parse systemshell netstat -rn output
#
# $_[0] = block of raw text
sub topo_systemshell_route_parser {

   my ($_line, $_header, $_c, $_i);
   my $_get_column = 0;
   my @_columns = ();
   my @_values = ();
   my %_hash = ();
   my @_array = ();
   my %_temp = ();

   $Log->comment("### parsing systemshell route output with local parser");
   foreach $_line ( split(/\n/, $_[0]) ) {
      #$Log->comment("### line $_line");

      next if $_line =~/^sudo/;
      next if $_line =~/^Routing tables/;
      next if $_line =~ /^$/;
      next if $_line =~ /^\s+$/;

      # find the Internet: and Internet6:  section demarc's
      if ( $_line =~ /^(\w+):$/) {
         $_header = $1;
         $Log->comment("### line  heading: [$_header]");
         $_get_column = 1;
         next;
      }
      # Internet:
      #    Destination  Gateway Flags Refs Use  Netif Expire
      # Internet6:
      #    Destination Gateway Flags  Netif Expire
      if ( $_get_column && $_line =~ /Destination\s+Gateway\s+Flags/ ) {
         @_columns = split(/\s+/, $_line);
         $_get_column = 0;
         $Log->comment("### line column: " . sprintf("[%s]", join(', ', @_columns)) )
      }
      
      else {
          @_values = split(/\s+/, $_line);
          if ( scalar(@_values)  >= 4 ) {
             $_i = 0;
             %_hash = ();
             foreach $_c ( @_columns ) {
                if ( defined($_values[$_i]) ) { $_hash{$_c} = $_values[$_i]; }
                else                          { $_hash{$_c} = ''; }
                $_i++;
             }
             push(@{$_temp{$_header}}, {%_hash});
             #$Log->comment("### line add $_header hash values " . Dumper \%_hash ); 
          }
      }
   }
   #$Log->comment("### temp hash " . Dumper \%_temp);

   return(\%_temp);
}
#------------------------------------------------------------------------
# parse nodescope 'route -sng' output
# taken from NACL/APISet/Node/CLI/7Mode/default/route.pm and added my changes 
# burt 562783
#   nacl does not account for routing group info
#
# Args
#    $_[0] = block of raw text
# Return
#    ref to AoHoAoH
#   [ {'internet' => [{}, {}, {}, ...],
#      'internet6' => [{}, {}, {}, ...],
#     }
#   ]
# 
sub topo_nodescope_route_parser {

   my ($_line, @_headings, @_values, %_hash,
       $_routing_group, $_routing_group_id, $_cmvfid,
       $_i, $_h,
   );
   my $_inet_flag = 'internet';
   my %_temp = ();
   my @_return_array = ();
   

   $Log->comment("### parsing nodescope 'route -sgn' output with local parser");
   foreach $_line ( split(/\n/, $_[0]) ) {
      next if $_line =~ /^$/;
      next if $_line =~ /^\s+$/;
      next if $_line =~ /^Routing tables/;

      # init table description vars
      if ($_line =~ /Routing group:/ ) {
         $_routing_group    = ''; 
         $_routing_group_id = '';
         $_cmvfid           = '';
      }

      # inet_4 or inet_6
      if ($_line =~ /Internet:/)     { $_inet_flag = 'internet'; next; }
      if ($_line =~ /Internet\sv6:/) { $_inet_flag = 'internet_v6'; next; }

      # routing group ( may not have CMVFID)
      if ($_line =~ /CMVFID\s+(\w+)/ )               { $_cmvfid = $1; }
      if ($_line =~ /Routing group:\s+(.+)$/ )       { $_routing_group = $1; next; }
      if ($_line =~ /Routing group-id:\s+(\w+)/ )    { $_routing_group_id = $1; next; }

      if ($_line =~ /Destination\s+Gateway\s+Flags\s+Refs\s/) {
         @_headings = split(/\s+/,$_line);
      }
      elsif ( $_line =~ /(.*)\s+(\w+)\s+(\d+)\s+(\d+)\s+(\w+)\s+(\w+)|(.*)\s+(\w+)\s+(\d+)\s+(\d+)\s+(\w+)\s+=>\s+(\w+)/){
         %_hash = ();
         @_values = split(/\s+/, $_line);
         $_i = 0;
         foreach $_h ( @_headings ) {
            $_hash{$_h} = $_values[$_i];
            $_i++;
         }
         # put routing group info on the 
         $_hash{'routing_group'}    = $_routing_group;
         $_hash{'routing_group_id'} = $_routing_group_id;
         $_hash{'cmvfid'}           = $_cmvfid;
         push(@{$_temp{$_inet_flag}}, {%_hash} );
      }
      else {
         $Log->comment("### parsing nodescope 'route -sgn' skipped [$_line]");
      }
   }
   push @_return_array ,\%_temp ;
   return \@_return_array;

}

#------------------------------------------------------------------------
# parse nodescope 'ipspace list' output
#
# Args
#    $_[0] = block of raw text
# Return
#    ref to AoH
# 
# Sample (indented here for readability)
#   Number of ipspaces configured: 3
#   default-ipspace                  0   (losk)
#   4294967294                       1   (e0a e0b)
#   4294967295                       2   (e0c e0d e0e e0f e0g e0h)
#
# AoH
# array = [ { 'index'  => <int>,
#             'interfaces' => [p1, p2,p3,..., pN],
#             'name'  => <str>,
#           },
#           {}, {},
#         ] 

sub topo_nodescope_ipspace_list_parser {

   my ($_line, @_v, $_port_str );
   my %_h = ();
   my $_num_ipspaces = 0;
   my @_return = ();

   $Log->comment("### parsing nodescope 'ipspace list' output with local parser");
   foreach $_line ( split(/\n/, $_[0]) ) {
      next if $_line =~ /^$/;
      next if $_line =~ /^\s+$/;
      if ( $_line =~ /Number\s+of\s+ipspaces\s+configured:\s+(\d+)/ ) {
         $_num_ipspaces = $1;
         next;
      }
      %_h = ();
      if ( $_line =~ /(.+\w)\s+(\d+)\s+\((.+)\)/ ) {
         $_h{'name'}  = $1;
         $_h{'index'} = $2;
         $_port_str   = $3;
         # the output has a unseen chars at the end messing things up
         # printing in hex can seen string ending with hex 2008
         #    x20 = (space)
         #    x08 = BS
         # need to strip these off
         ## print "x", unpack("H*", $_port_str), "\n";
         $_port_str =~ s/\x08$//;
         $_port_str =~ s/\x20$//;
         if ( $_port_str =~ /\s/) { push(@{$_h{'interfaces'}}, split(/\s/, $_port_str)); }
         else                     { push(@{$_h{'interfaces'}}, $_port_str); }
         # push hash onto AoH
         push(@_return, {%_h});
      }
      else {
         $Log->info("### unknown line syntax for [ipspace list]: $_line");
      }
   }
   # check number of parse lines match number given as ipspace count
   # since making an AoH would be odd to insert this one scalar somewhere
   # but since it represent the count use it to check the parser
   if ( $_num_ipspaces == scalar(@_return) ) {
      return(\@_return);
   }
   else {
      return( undef );
   }

}
#
#------------------------------------------------------------------------
#
# Match NACL VServer component to parsed output of 
# 'debug smdb table vserver_by_name  show'
# Args
#   $_[0] ref to nacl vserver  
#   $_[1] ref to vserver_by_name entry
#
# account for uuid = 00000000-0000-0000-0000-000000000000 od '-'
sub match_vserver_to_vserver_by_name {

   my $_r = 0;
   if ( $_[0]->{'vserver'} eq $_[1]->{'vserver'} ) {
      if (  $_[0]->{'type'}  eq $_[1]->{'type'} )  {
         if ( $_[0]->{'uuid'}  eq $_[1]->{'uuid'} ) {
            $_r = 1;
         }
         elsif ( $_[0]->{'uuid'} =~ /^[0\-]+$/ && $_[1]->{'uuid'} eq '-' ) {
            $_r = 1;
         }
      }
   }
   $_r;
}
#
#------------------------------------------------------------------------
#
# Match NACL VServer component to parsed output of 
# 'debug smdb table vserver_for_ipspace  show'
# Args
#   $_[0] ref to nacl vserver  
#   $_[1] ref to vserver_for_ipspace entry
#
sub match_vserver_to_vserver_for_ipspace {

   my $_r = 0;
   if ( $_[0]->{'vserver'} eq $_[1]->{'vserver'} ) {
      if (  $_[0]->{'type'}  eq $_[1]->{'type'} )  {
         if ( $_[0]->{'uuid'}  eq $_[1]->{'uuid'} ) {
            $_r = 1;
         }
         elsif ( $_[0]->{'uuid'} =~ /^[0\-]+$/ && $_[1]->{'uuid'} eq '-' ) {
            $_r = 1;
         }
      }
   }
   $_r;
}

#-----------------------------------------------------------------------------
#  SCP logs files from /mroot/etc/mlog 
# Args 
#   $_[0]   = node ref
#   $_[1-N] = list of files
#
sub scp_mlog_file {

   my $_node_ref    = shift;
   my @_filer_files = @_;
   my $_nodename    = \$_node_ref->hostrec->{'hostname'};
   my $_filer_dir   = '/mroot/etc/mlog';
   my $_client_dir  = &_gen_logdir_name($$_nodename);
   my @_scp         = ();

   $Log->comment("scp_mlog_file: node $$_nodename files: "
                . join(', ', @_filer_files)
   );

   # make sure client side dir exists
   if (!-e  $_client_dir ) {
      $Log->comment("scp_mlog_file: need to make local dir: $_client_dir");
      make_path($_client_dir,{'verbose'  => 1, 'mode' => 0755});
   }

   # SCP from target (filer) to source (client)
   $Log->comment("scp_mlog_file: copy files from filer");
   @_scp = NACL::MTask::SCP->copy_to_localhost(
      target_command_interface => $_node_ref,
      set_directory_on_target  => $_filer_dir,
      set_directory_on_source  => $_client_dir,
      files_to_copy            => \@_filer_files
   );
   $Log->comment("scp_mlog_file: files copy_to_localhost from $$_nodename to");
   $Log->comment(Dumper \@_scp);

   return( wantarray ? (\@_scp, '') : \@_scp);

   #---------------------- priv local functions -----------------------------
   # $_[0] = node name
   # $_[1] = user given logdir site
   sub _gen_logdir_name {
      sprintf("%s/%s/mlogs/%s",
         $_[1] || Tharn::param('LOGDIR'),
         Tharn::param('RUNID') || 'runid',
         $_[0]
      );
   }
}

#-----------------------------------------------------------------------------
# parse the hexstamp at beginning of log message
# $_[0] = line from log file
#
sub parse_logmsg_hexstamp {
   my $_r = '';
   if ( $_[0] =~ /([0-9a-f\s\.]+)\s[MTWFS][oueiahr]/) {
      $_r = $1;
   }
   $_r;
}

#-----------------------------------------------------------------------------
# parse the unix date string at beginning of log message and convert to epoch
# Args
#   $_[0] = line from log file
# Return
#   (date string, epoch)
#
sub parse_logmsg_datestr {
   my ($_ds, $_t) = ('', 0);

   if ( $_[0] =~ /([MTWFS]\w\w\s[JFMASOND]\w\w\s[\d\s\-:+]+)/ ) {
      $_ds = $1;
      $_t = str2time($_ds);
   }

   return($_ds, $_t);
}

#-----------------------------------------------------------------------------
# Decoding file log messages
# Given in hex, parse and convert to dec
# Return both hex and dec values
#
# http://wikid.netapp.com/w/Log_format
# Sample message 
# 00000007.000000a9 00721f21 Thu Sep 20 2012 13:37:20 +00:00 [mgr.boot.reason_ok:notice] blah
#
# Message format
# BOOT#.SEQ# UP# DOW MMM DD YYYY HH:MM:SS ZZZZZ [EMS:SEVERITY:PID]: TEXT
# BOOT# is a zero filled hex value for a 32-bit number which increments 
#       each time the system boots and the log daemon starts
# SEQ# is a zero filled hex value for a 32-bit number which increments 
#      when each new message is recorded. 
#      The combination of BOOT# and SEQ# creates a unique key for message interleaving.
# UP# is a zero filled hex value for a 32-bit number which represents 
#     the current uptime in 1/10th of a second as represented by the 
#     cycle timer when the entry is made by the log daemon (or recording process).
#
# Args
#  $_[0] = hex string from log
#
# Return
#  (boot_hex, seq_hex, up_hex, boot_dec, seq_dec, up_dec)
#
sub decode_logmsg_hex {

    # set boot, seq,  up vars for hex and dec
    my ( $_b_h, $_s_h, $_u_h, $_b_d, $_s_d, $_u_d ) = ('','','',0,0,0);

    # parse hex from the string
    if ( $_[0] =~ /([0-9a-f]+)\.([0-9a-f]+)\s([0-9a-f]+)/ ) {
        $_b_h = $1;
        $_s_h = $2;
        $_u_h = $3;
        # convert hex to dec
        $_b_d = hex($_b_h);
        $_s_d = hex($_s_h);
        $_u_d = hex($_u_h);
    }

    return($_b_h, $_s_h, $_u_h, $_b_d, $_s_d, $_u_d);
}

#--------------------------------------------------------------------------
#    Priv functions
#
#------------------------------------------------------------------------------
# $_[0] = ref to filter hash
#
sub _filter_str_comment {
   my $_s = '';
   my $_k;
   foreach $_k ( sort keys %{$_[0]} ) {
      $_s .= sprintf("%s=%s ", $_k, $_[0]->{$_k});
   }
   $_s;
}
#-------------------------------------------------------------------------------
# $_[0] = ref to filter hash
# $_[1] = ref to nacl object
#
sub _filter_match_str_comment {
   my $_s = 'filtered: ';
   my $_k;
   foreach $_k ( sort keys %{$_[0]} ) {
      $_s .= sprintf("%s=%s ", $_k, $_[1]->{$_k} || '??');
   }
   $_s;
}

#-------------------------------------------------------------------------------
# Ensure translation between role and ipspace
#
# $_[0] = ref to hash with role|ipspace key
#
# 
sub _role_ipspace_translate {

   my $_obj = shift;
  
   if ( defined($_[0]->{'role'}) && !defined($_[0]->{'ipspace'}) ) {
      $_[0]->{'ipspace'} = $_[0]->{'role'};
   }
   elsif ( !defined($_[0]->{'role'}) && defined($_[0]->{'ipspace'}) ) {
      $_[0]->{'role'} = $_[0]->{'ipspace'};
   }


   1;
}


1;
__END__

=pod

=head1 NAME Network::SCONTopology

=head1 DESCRIPTION

Perl Object to gather Topology information from a cluster for user to 
use to determine topology of the cluster. Information gathered and 
merged into a hash for later use includes port, interface, ifgroup, 
vlan, vserver, ipspace. Lower layer systemshell (bsd) and nodescope (SK, 7M) 
shell ifconfig and route output is parsed and put into hash.

Oject provides user control over what information is gather
and provides methods to derive information such as stats and 
profiles.


=head1 SYNOPSIS

    use Network::SCONTopology;

    ($Topology, $error) = Network::SCONTopology->new(
        'nacl_c_node'   => \@Nodes,
        'fetch_port'    => 1,
        'fetch_vserver' => 0,
        'fetch_shell'   => 0,
    );


    ($ok, $_error) = $Topology->get_port_list(
        'aref'   => \@_data_ports,
        'filter' => { 'role' => 'data'},
    );

    ($_ok, $_error) = $Topology->update_topology;

    ($ok, $error) = $Topology->get_interface_list(
        'aref'   => \@if_list,
        'filter' => {'is_home' => 'true' },
        'port_filter' =>  { 'role' => 'data' },
    );


=head1 PARAMS

=head2 TOPOLOGY_MLOG_SCP_SLEEP

    TOPOLOGY_MLOG_SCP_SLEEP=<int>

    Amount of time in seconds to sleep between triage tool running and scp of log 



=head1 Methods

Method functionality is reflected in the naming scheme of the method.

=over

=item fetch_*  using NACL ComponenetState methods, fetch information 
from cluster

=item get_* get the information from the object and present to user in 
as a usable data structure

=item find_* will run through all object data to find info

=item parse_* will parse blocks of text block of filer output

=back


=head2 new

    Constructor to create Topology object.

    ($Topology, $error) = Network::SCONTopology->new(
        'nacl_c_node'   => \@Nodes,
        'fetch_port'    => 1,
        'fetch_vserver' => 1,
        'fetch_shell'   => 0,
    );

Object will determine cluster identity and verify that all nodes in the 
NACL::C::Nodes object are in the same cluster. If you desire multiple
clusters, then make multiple objects.

=head3 Args

=over

=item nacl_c_node => <ref>

    Reference to NACL::C::Nodes->find() object.
    Will be used by the object to provide command interface
    to NACL objects and  as a node list to work on

    Required and will be checked to be a NACL::C::Node object.

=item fetch_port => 0|1

    Enable/disable the constructor to call fetch_port_if() 
    to fetch network port and interface info.

    Optional default=0 

=item fetch_vserver => 0|1

    Enable/disable the constructor to call fetch_vserver_ipspace() 
    to fetch vserver and ipspace info. 

    Optional default=0

=item fetch_shell => 0|1

    Enable/disable the constructor to call fetch_shell_commands() 
    to fetch vserver and ipspace info.

    Optional default=0

=back

=head3 Return

    Upon success a reference to the object and empty error string.
    Upon failure the reference is undefined and a descriptive error string

    In list context will return a reference to the object and error
    string.

    In scalar context will return a reference to the object and
    the error string can be retrieved with the error_string() method. 


=head2 update_topology

    Update the object by running the info gathering methods

    $Topology->update_topology;
    $ok = $Topology->update_topology;
    ($ok, $error) = $Topology->update_topology;


    The object maintains a list of which fetch methods have been
    used. Calling update_topology() will re-run any previously
    called methods so as to update the topoplogy info.

=head3 Return

    Upon success a reference to the object and empty error string.
    Upon failure the reference is undefined and a descriptive error string

    In list context will return a reference to the object and error
    string.

    In scalar context will return a reference to the object and
    the error string can be retrieved with the error_string() method.


=head2 error_string

    Return last error encountered by a method or function

    $error = $Topology->error_string;
    $Log->error("ERROR: $Topology->error_string"):

=head2 set_node_index

    Populate the object with a node name to index hash

    $Topology->set_node_index;
    $ok = $Topology->set_node_index;

    Will loop through the node list given to the object with the 
    nacl_c_node arg (NACL::C::Node) and make a hash whose
    nodename key is what is pulled from $obj->hostrec->{'hostname'}.
    Will return the number of nodes indexed in the object
    Any call to this will init the previous content of the 
    node index hash.

=head2 get_node_list

    Get list of node name stored in object

    @nodenames = $Topology->get_node_list;

=head3 Args

    none

=head3 Return

    list of node names

=head2 get_node_index_href

    Get a reference to the node index hash

    $ref = $Topology->get_node_index_href;

=head3 Args

    none

=head3 Return

    Node index hash has the node name as the key and the index
    value is the index in the NACL::C:Node list

=head2 get_node_index

    Get the index value for the given node name

    $i = $Topology->get_node_index( 'node' => <nodename> );

    If the hash made with set_node_index is not populated, then 
    this method is called. The given node name is then looked for in 
    this hash.

=head3 Args

=over

=item node => <nodename>  

    Name of node expected to be NACL::C:Node object
    hostrec hostname value ($node->hostrec->{'hostname'}.

=back

=head3 Return

    If node name is found then the index in the NACL::C::Node list.
    Otherwise udef is returned.

=head2 get_node_index_href

    Get a ref to the hash holding the nodename to NACL::C::Node list index.

    $ref = $Topology->get_node_index_href

=head3 Args

    none

=head3 Return

    ref to hash where 

        $ref->{<nodename>} = index in NACL::C::Node

=head2 determine_ipspace_support

    Check all nodes known by the object for IPSpace support

    $Topology->determine_ipspace_support;
    $ok = $Topology->determine_ipspace_support;
    ($ok, $error) = $Topology->determine_ipspace_support;

    Will loop through all nodes known to the object and use the 
    NACLs is_feature_supported() feature to determine IPSpace support.
    The object will store info indicating support.

=head3 Return

    In list context the number of nodes supporting IPSpaces and empty 
    error string. 

    In scalar context the number of nodes supporting IPSpaces.

    If not all nodes had IPSpace support verified then O and
    error string will be returned.

    If any one node does not support IPSpaces then 0 and an error
    string indicating which nodes do not support IPSpaces is returned.

=head2 determine_cluster_identity

    Determine that all nodes in the object are in the same cluster

    $ok = $Topology->determine_cluster_identity;
    ($ok, $error) = $Topology->determine_cluster_identity;

    For the object all nodes are required to be in the same cluster.
    This is called by the constructor (new()) but is a method
    available to the user.

=head3 Return

    In list context the number of nodes in the cluster and an
    error string.


    Upon success the number of nodes in the cluster and an empty
    error string.

    Upon failure undef and error string containing node name
    and the two differnt cluster name.


=head2 determine_cli_args

    Determine all possible arguments for a filer ui command.

    $ok = $Topology->determine_cli_args( 'api' => <str> );
    ($ok, $error) = $Topology->determine_cli_args( 'api' => <str> );

    User may want to run through all argumnet permutations of an 
    UI command option. NACL does not yet provide method to do this.
    User can suppy an API string to this method and all allowable
    field values will be storted in the object.

    Each node is queried and NACL apiset is made from the NACL node 
    object and used to call the cli_command_helper() method. The output
    is parsed and stored in the Network::SCONTopology object

        %obj->{'Args'}{<api>}{<node>}{<field>} = (arg,1, arg2, ...)

=head3 Args

=over

=item 'api' => <str>

    UI command with underscores (required)

=back

=head3 Return

    In list context the number of values found across all nodes and
    an error string.

    In scalar context the number of values found across all nodes.

    If no values are found then 0 and error string indicating such.

=head2 fetch_port_if

    Method to fetch NACL Port, Interface, IFGroup and VLAN components
    using NACL::CS::NetworkPort,  NACL::CS::NetworkInterface,
    NACL::CS::NetworkPortIfgrp, NACL::CS::NetworkPortVlan.

    $Topology->fetch_port_if;
    $ok = $Topology->fetch_port_if;
    ($ok, $error) = $Topology->fetch_port_if; 

    Object will store NACL CS object and relate IFs to their ports.
    Stats for each port will be derived for other uses.

=head3 Stats

=over

=item NodeIndex - the index of the home node in objecs NACL::C::Node list

=item if_count - the count of LIFs using this port as home node|port

=item isaPort - a network port (network port show)

=item isaIFgrp - an IFGroup port (network port ifgrp show)

=item isaVlan - a VLAN port (network port vlan show)

=item inaVlan - port is a member of a VLAN, tagged vlan

=item inaIFgrp - port is a member of an IFGroup, port given in IFGroup port list 

=item InIFgrp - the ifgroup this port is member of

=item InVLAN - list of Tagged VLANs this port belongs to 

=back

=head3 Args

    none

=head3 Return

    In list context the port count and error string. 
    In scalar context the port count.

    Upon success port count will be an integer since a valid port count could 
    be 0.


=head2 fetch_vserver_ipspace

    Method to fetch NACL Vserver, Route and IPspace components 
    using NACL::CS::Vserver, NACL::CS::NetworkRoute, NACL::CS::NetworkIpspace.
    
    $Topology->fetch_vserver_ipspace;
    $ok = $Topology->fetch_vserver_ipspace;
    ($ok, $error) = $Topology->fetch_vserver_ipspace;

    Method will also gather vserver information from 
    'debug smdb table vserver_by_name show'. If IPSpace
    is seen to be supported on the node then IPSpace
    info is gathered along with 
    'debug smdb table vserver_for_ipspace show'.

    This method will call the map_ipspace_name_id() method
    so as to populate the object with a mapping.

=head3 Args

    none

=head3 Return

    In list context the vserver count and error string. 
    In scalar context the vserver count.

    Upon success port count will be an integer since a valid vserver count 
    could be 0.


=head2 map_ipspace_name_id

    Map IPSpace name to its ID

    $Topology->map_ipspace_name_id;

    Populates the object.

=head3 Args

    none

=head3 Return

    The number of IPSpace IDs mapped. 

=head2 fetch_shell_commands

    Method to connect to systemshell (bsd) and nodescope
    (7M, SK) shell to get ifconfig and route tables.

    $Topology->fetch_shell_commands;
    $ok = $Topology->fetch_shell_commands;

    Will create needed NACL connections to respective shell to run the 
    commands, parse the output and store in the object.

=head3 Args

    none

=head3 Return

    1

=head2 get_port_list

    Populate a user array with port names.

    @port_list = ();
    ($ok, $error) = $Topology->get_port_list(
        'aref'   => \@port_list,
        'filter' =>  { 'link' => 'up' },
    );
    unless ($ok) {
        $Log->error("ERROR: get_port_list: $error");
    }

    Will populate the list given with a set of array references
    where each array ref holds the node name and port name. Thus
    user will have an AoA = ([node,port], [], ...).
    Each array ref in the list holds the index used to get the 
    NACL port object from Topology object. 

=head3 Args

=over

=item aref => <aref>

    Reference to user array populate with port values 


=item node => <nodenames>

    Optional: name of nodes, comma separated to get port info from.

=item filter => <href>

    Optional: reference to user hash to supply as filter argumnet to 
    NACL port object.  $port->verify_fields(<href>);
    
=back

=head3 Return

    The number of ports added to the user port list and error string

    The user list populated by this method will be a AoA such as
    ( [node,port], [node,port],  [node,port], ...) 


=head2 get_portname_list

    Get a port name list for a given node

    @portnames = $Topology->get_portname_list(<nodename>);

    For the given node name, will return all port names for the node. 


=head3 Args

=over

=item <nodename> 

    Node name as known in the hostrec for the node.

=back 

=head3 Return

    List of port names as a simple list
    ( port, port, port, ...)

=head2 get_port_object

    Get the NACL::CS::NetworkPort object for the given node, port.

    $port_obj = $Topology->get_port_object(<node>, <port>);

    foreach $aref ( \@port_list ) {
        $port_obj = $Topology->get_port_object(@{$aref});
    }

=head3 Args

=over

=item <node>

    Node name

=item <port>

    Port name such as: e0M, e0a, e0b 

=back

=head3 Return

    In list context a reference to NACL::CS::NetworkPort object 
    and error string. In scalar context just the NACL::CS::NetworkPort object
    reference.

    If the node:port pairing is not stored in the object then the returned
    ref is undef and descriptive error string.

=head2 get_port_stats

    Return a ref to the hash holding the port stats.

    $port_stats = $Topology->get_port_stats(<node>, <port>);

    foreach $aref ( \@port_list ) {
        $port_stats = $Topology->get_port_stats(@{$aref});
    }

    See fetch_port_if() for description of port stats

=head3 Args

=over

=item <node>

    Node name

=item <port>

    Port name such as: e0M, e0a, e0b 

=back

=head3 Return

    In list context a reference to stats hash as described in fetch_port_if() 
    and error string. In scalar context just the stats hash.

    If the node:port pairing is not stored in the object then the returned
    ref is undef and descriptive error string.


=head2 get_port_profile 

    Get port profile setting

    ($_profile, $_real, $_bound, $_member, $_ip_ver) = $Topology->get_port_stats(<node>, <port>);

    From the port stats determine a profile for the given port.

=over

=item profile = 1 if port is profiled otherwise 0

=item real = 1 if the isaPort is true otherwise 0

=item bound = 1 if isaIFgrp or isaVlan or if_count or inaIFgrp or inaVlan is true otherwise 0

=item member = 1 if inaIFgrp or inaVlan is true otherwise 0

=item ip version 

=over

=item inet4 if all LIFs are IPv4

=item inetv6 if all LIFs are IPv6

=item mixed if IPv4 and IPv6 LIFs are on the port

=item unknown otherwise 

=back

=back

=head3 Args

=over

=item <node>

    Node name

=item <port>

    Port name such as: e0M, e0a, e0b 

=back

=head3 Return

    List of profile values (profile, real, bound, member, ip_ver)

=head2 get_port_profile_hash

    Get port profile hash

    %profile_hash = $Topology->get_port_stats(<node>, <port>);

    See get_port_profile() description for more detail

=head3 Return

    Return a hash with profile settings whose hash keys are 
    $profile_hash{'profile'}
    $profile_hash{'real'}
    $profile_hash{'bound'}
    $profile_hash{'member'}
    $profile_hash{'ip_ver'}


=head2 get_interface_list

    Populate a user array with interface port names and inexes.

    @if_list = ();
    ($ok, $error) = $Topology->get_interface_list(
        'aref'        => \@if_list,
        'filter'      =>  { 'role' => 'data' },
        'port_filter' =>  { 'ipspace' => 'Default'},
        'port_list'   =>  \@port_list,
    );
    unless ($ok) {
        $Log->error("ERROR: get_interface_list: $error");
    }

    Will populate the list given with a set of array references
    where each array ref holds the node name, port name, list index . 
    Thus user will have an AoA = ([node,port, index], [], ...).
    Each array ref in the list holds the index used to get the 
    NACL:CS:NetworkInterface object from Topology object. 

=head3 Args

=over

=item aref => <aref>

    Reference to user array  to populate with port values 


=item node => <nodenames>

    Optional: name of nodes, comma separated to get port info from.

=item filter => <href>

    Optional: reference to user hash to supply as filter argumnet to 
    NACL:CS:NetworkInterface object.  $if->verify_fields(<href>);

=item port_filter => <href>

    Optional: reference to user hash to supply as filter argumnet to 
    NACL port object.  $port->verify_fields(<href>);

=item port_list'   =>  <aref>

    Optional: reference to user array to supply port_list.
    This would be a AoA made with $obj->get_port_list() or 
    user made list. See get_port_list() for details on 
    AoA structure. 

=back

=head3 Return

    The number of IFs added to the user IF list and error string

    The user list populated by this method will be a AoA such as
    ( [node,port,if_index], [node,port,if_index]], ...) 

    if_index is the index value used in this objects storing
    of IFs and not reflective of any index of an IF in the cluster


=head2 get_interface_object

    Get the NACL::CS::NetworkInterface object for the given node, port, index.

    $if_obj = $Topology->get_interface_object(<node>, <port>, index);

    @if_list   = ();
    ($ok, $error) = $Topology->get_interface_list('aref' => \@if_list)
    foreach $aref ( \@if_list ) {
        $if_obj = $Topology->get_interface_object(@{$aref});
    }

=head3 Args

=over

=item <node>

    Node name

=item <port>

    Port name such as: e0M, e0a, e0b 

=item <index>

    Index value the IF is stored in object with.

=back

=head3 Return

    In list context a reference to NACL::CS::NetworkInterface object 
    and error string. In scalar context just the NACL::CS::NetworkInterface object
    reference.

    If the node:port:index grouping is not stored in the object then the returned
    ref is undef and descriptive error string.

=head2 get_vlan_list 

    Populate a user array with vlan id names

    ($ok, $error) = $Topology->get_vlan_list(
        aref   => \@vlan_list,
        filter => { port => <str> },
    );

    Will populate the list given with a set of array references
    where each array ref holds the vlan name and vlan_id.
    Thus user will have an Array = (vlan-id...).
    Each array ref in the list holds indexing needed to get the 
    NACL::CS::NetworkPortVlan object. The vlan-id
    is that as give by the NACL::CS::NetworkPortVlan->vlan_id()
    

=head3 Args

=over

=item aref => <aref>

Reference to user array  to populate with port values

=item filter => <href>

    Optional: reference to user hash to supply as filter argumnet to
    NACL::CS::NetworkPortVlan object.  $if->verify_fields(<href>);

=item node => <nodenames>

    Optional: name of nodes, comma separated to get port info from.

=head3 Return

    In list context the number of vlans found in the object and error string.
    In scalar context just the number of vlans found in the object and error 
    string. 

=head2 get_vlan_object

    Get the NACL::CS::NetworkPortVlan object from the Topology object.

    ($vlan, $error) = $Topology->get_vlan_object( vlan, id)

    ($ok, $error) = $Topology->get_vlan_list( aref => \@vlan_list);
    foreach $aref ( @vlan_list ) {
        $Topology->get_vlan_object(@{$aref});
    }

=head3 Args

=over

=item vlan

    The VLAN 

=item  vlan-id

=back

=head3 Return

    In list context a reference to the NACL::CS::NetworkPortVlan object
    and error string. In scalar context the reference to the 
    NACL::CS::NetworkPortVlan object.

    If the given values are not found the the Topology object
    then the ref value is undef and error string is made.


 
=head1 Functions


=head1 Misc

=head1 STATUS

   Current 

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date: 2013/03/04 $

=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved

=cut

