## @summary SCON Trafic object
##
## @description SCON Traffic 
##
## @author      isarika@netapp.com
##
## @pod here
# POD at end of file

package SCONTraffic;

# Compiler directives.
use strict;

# Module import.
use TCD;
use San::SanMgr;
use IOMan;
use Data::Dumper;
use Data::Compare;
use NACL::C::Node;
use NACL::C::Statistics;
use NACL::CS::Statistics;
use NACL::STask::Client::Mount;
use NACL::C::Client;
use NACL::STask::StorageFailover;
use NATE::BaseException qw(:try);
use NACL::C::ClusterRing;
use NACL::STask::Cluster;
use NACL::C::Volume;
use NACL::C::Aggregate;
use Tharn;

sub new {
	my $class = shift;
	my $args = shift;

	my $Debug_Level = 1;

	# Create IOMan Obj
	my $IOMan_Obj = IOMan->new(-debug_level => $Debug_Level,
				-testing_method => $args->{'testing_method'});
	if ($IOMan_Obj->error()) {
		$Log->warn("Failed in IOMan constructor". $IOMan_Obj->errmsg());
		$IOMan_Obj=undef;
		}
	
	my $self = {
		nodes 		=> $args->{'nodes'},
		client		=> $args->{'client'},
		devices 	=> $args->{'devices'},
		IOMan_Obj 	=> $IOMan_Obj,
		aggrnode 	=> 'aggrnode',
		lifnode		=> 'lifnode',
		data_protocol	=> '',
		address_family	=> '',
		hosttype	=> '',
		junction_path	=> '',
		traffic_started => 0,
		nfsops		=> 0,
		cifsops		=> 0,
		iscsiops	=> 0,
		traffic_running	=> 0,
		traffic_tool		=> $args->{'traffic_tool'}
		};

	$Log->comment("Nodes:\t\t\t@{$self->{nodes}}\nClients:\t\t$self->{client}\nDevices:\t\t@{$self->{devices}}\n");
       
	bless $self, $class;
	
	my $clienthost = Tharn::host($self->{client});
	$self->{hosttype} = $clienthost->{hosttype};
	$self = $self->_get_params_for_verify();
	$self = $self->_create_io_records();	
	return $self;
}

sub start {
	my ( $self ) = @_;
	# Start io on each io_record
	$Log->comment("Starting $self->{data_protocol} traffic");
	$self->{IOMan_Obj}->start();
	if ($self->{IOMan_Obj}->error()) {
		$Log->warn("Failed to start io for $self->{data_protocol} mount points ".$self->{IOMan_Obj}->errmsg());
		return 0;
		}
	$self->{traffic_started} = 1;
	return 1;
}

sub play {
	my ( $self ) = @_;
	if($self->{traffic_started} == 1){
		$Log->warn("Traffic is already running");
		return 1;
	}
	# Start io on each io_record
	$Log->comment("Starting $self->{data_protocol} traffic");
	$self->{IOMan_Obj}->start();
	if ($self->{IOMan_Obj}->error()) {
		$Log->warn("Failed to start io for $self->{data_protocol} mount points ".$self->{IOMan_Obj}->errmsg());
		return 0;
		}
	$self->{traffic_started} = 1;
	$Log->comment("Sleeping for 120 secs");
	sleep(120);
	return 1;
}

sub pause {
	my ( $self ) = @_;
	if($self->{traffic_started} == 0){
		$Log->warn("Traffic is already stopped");
		return 1;
	}
        my $client=$self->{client};
	my @devices=@{$self->{devices}};
	
	$Log->comment("Stopping $self->{data_protocol} traffic");
	$self->{IOMan_Obj}->stop();
	if ($self->{IOMan_Obj}->error()) {
		$Log->warn("Failed to stop io recs ".$self->{IOMan_Obj}->errmsg());
		return 0;
		}
	$self->{traffic_started} = 0;
	$Log->comment("Sleeping for 120 secs");
	sleep(120);
	return 1;
}

sub stop {	
	my ( $self ) = @_;
	
        my $client=$self->{client};
	my @devices=@{$self->{devices}};
	
	$Log->comment("Stopping $self->{data_protocol} traffic");
	$self->{IOMan_Obj}->stop();
	if ($self->{IOMan_Obj}->error()) {
		$Log->warn("Failed to stop io recs ".$self->{IOMan_Obj}->errmsg());
		return 0;
		}
	
	$Log->comment("Deleting files created for SIO");
	my $Client_obj = NACL::C::Client->new(name => $client);
	my $i=1;
	if($self->{hosttype} =~ /windows/) {
		$Client_obj->apiset_must(connid => "rsh");
		foreach my $device (@devices){
			my $siofile='\testsio_win';
			$device =~ s/^(\w+:)\\.*/$1/g;
			$device=$device.$siofile.$i;
			try{
				my $output = $Client_obj->execute(command => "del $device");
			} catch NATE::BaseException with {
				$Log->warn("Failed to delete the file created for SIO");
			};
			$i++;
		}
			
	}
	else {
		$Client_obj->apiset_must(connid => "ssh");
		foreach my $device (@devices){
			my $siofile='/testsio';
			$device=$device.$siofile.$i;
			try{
				my $output = $Client_obj->execute(command => "rm -rf $device");
			} catch NATE::BaseException with {
				$Log->warn("Failed to delete the file created for SIO");
			};
			$i++;
		}
	}	
	
	$self->{traffic_started} = 0;
	return 1;
}

sub _create_io_records {
	my ( $self ) = @_;
        my $client=$self->{client};
	my @devices=@{$self->{devices}};
	my $traffic_tool = $self->{'traffic_tool'};
	
	# Create the file in the mount point to run sio.
	$Log->comment("Creating file for SIO");
	my $Client_obj = NACL::C::Client->new(name => $client);
	my $i=1;
	if($self->{hosttype} =~ /windows/) {
		$Client_obj->apiset_must(connid => "rsh");      
		foreach my $device (@devices){
			my $siofile='\testsio_win';
			$device =~ s/^(\w+:)\\.*/$1/g;
			$device=$device.$siofile.$i;
			try {
				my $output = $Client_obj->execute(command => "fsutil file createnew $device 10000");
				} catch NATE::BaseException with {
					$Log->warn("Failed to create file $device"); 
				};
			$i++;
			} 
	}
	else {
		$Client_obj->apiset_must(connid => "ssh");      
		foreach my $device (@devices){
			my $siofile='/testsio';
			$device=$device.$siofile.$i;
			try {
				my $output = $Client_obj->execute(command => "touch $device");
				} catch NATE::BaseException with {
					$Log->warn("Failed to create file $device"); 
				};
			$i++;
			} 
	}
	# Params for creating io recs. Two io_records will be created, one for each
	# device.
	my $pattern = "a";
	my $limit = "10000m";
	my @io_records = $self->{IOMan_Obj}->create_io_records(
			  -client    => $client,
			  -devices   => \@devices,
			  -io_tool   => $traffic_tool,
			  -iotype    => "random",
			  -pattern   => $pattern,
			  -bs        => "4k",
			  -limit     => $limit,
			  -passes    => 1,
			  -read      => 0,
			  -verify    => 0
		      );

	if ($self->{IOMan_Obj}->error()) {
		$Log->warn("Failed to create io recs ".$self->{IOMan_Obj}->errmsg());
		return 0;
		}

	return $self;
}

sub update_verify_params {
	my ( $self ) = @_;
	$self = $self->_get_params_for_verify();
	return $self;
}

sub _get_params_for_verify {
	my ( $self ) = @_;
        my $client=$self->{client};
	my @devices=@{$self->{devices}};
        my $mount_point=$devices[0];
	my @nodes=@{$self->{nodes}};
	my $up_node=$nodes[0];
	$up_node=NACL::C::Node->new(name => $up_node);
		
        $Log->comment("Finding the junction path of the volume ...");
        $Data::Dumper::Maxdepth = 3;
        my $Client_obj = NACL::C::Client->new(name => $client);

        if($self->{hosttype} =~ /unix/) {
		$Client_obj->apiset_must(connid => "ssh");
	} else {
		$Client_obj->apiset_must(connid => "rsh");	
		if($mount_point =~ m/(^\w+:).*/){
			$mount_point = $1;
		}
	}

 
        my $mount_obj = NACL::STask::Client::Mount->new(
                       command_interface => $Client_obj,
                       mount_point       => $mount_point
                   );
	my $mount_state = $mount_obj->state();
        my $lif_server = $mount_state->{'server'};
        my $data_protocol = $mount_state->{'type'};
        my $junction_path = $mount_state->{'path'};
        if($self->{hosttype} =~ /windows/) {
        	$data_protocol = 'cifs';
		$junction_path = $devices[0];
		if($junction_path =~ m/^\w+:\\(.*)/){
			$junction_path = '/'.$1;
		}
		$lif_server =~ s/-/:/g;
		$lif_server =~ s/\.ipv6:literal\.net//g;
		$Log->comment("lif for windows is now : $lif_server");	
	}

	$lif_server =~ s/\[//g;	
	$lif_server =~ s/\]//g;	


	$Log->comment("Volume Junction Path: $junction_path, LIF IP Address: $lif_server, Data Protocol: $data_protocol");

        $Log->comment("Locating the node on which the aggregate is present");
        my $volobj=NACL::C::Volume->find(command_interface => $up_node, filter => { 'junction-path' => $junction_path });
        my $aggrname=$volobj->{'aggregate'};
        my $aggrobj=NACL::C::Aggregate->find(command_interface => $up_node, filter => { 'aggregate' => $aggrname });
        my $aggrnode = $aggrobj->get_local_node_name();
        $Log->comment("Aggregate is on $aggrnode");

        $Log->comment("\nLocating the Node on which the LIF is currently placed ...");
        my $mount_lif = NACL::C::NetworkInterface->find(command_interface => $up_node, filter => { 'address' => $lif_server });
        my $mount_lif_state = $mount_lif->state();
        my $current_node = $mount_lif_state->{'curr_node'};
        my $address_family = $mount_lif_state->{'address_family'};
        $Log->comment("LIF is placed on $current_node, Address Family: $address_family");

        my %values = (  "aggrnode"      => $aggrnode,
                        "lifnode"       => $current_node,
                        "data_protocol" => $data_protocol,
                        "address_family" => $address_family
                        );
	
	$self->{aggrnode} = $aggrnode;
	$self->{lifnode} = $current_node;
	$self->{data_protocol} = $data_protocol;
	$self->{address_family} = $address_family;	
	$self->{junction_path} = $junction_path;

        return $self;
}

sub verify_traffic
{
	my ( $self ) = @_;
        
	my $aggrnode = $self->{aggrnode};
        my $lifnode = $self->{lifnode};
        my $protocol = $self->{data_protocol};
        my $address_family = $self->{address_family};

        my $Hostobj=Tharn::host($aggrnode);
        my $Node=NACL::C::Node->new(name => $lifnode);
        my $api_obj = NACL::APISet->new(hostobj   => $Hostobj,
                        category  => "Node",
                        interface => "CLI",
                        set       => "Nodescope",
                        );

        my $nfs_threshold=0;
        my $cifs_threshold=0;
        my $iscsi_threshold=0;
        my $cpu_threshold=0;

        $Log->comment("Sleeping for 5 seconds");
        sleep(5);

        $Log->comment("Checking sysstat on $aggrnode");
        my $response =  $api_obj->sysstat ('count'=> 5,
                'interval'    => 1,
                "print-all"   => 1,
                "print-summary"=> 1,
                "SAN-statistics" => 1
                );
        my $parsed_op = $response->get_parsed_output();
        if($protocol =~ /nfs/i){
                my $nfs_stat  = $parsed_op->[0]->{'Avg'}{'NFS'};
                $self->{nfsops} = $nfs_stat;
		if($nfs_stat > $nfs_threshold){
                        $Log->comment("NFS(ops per sec)\t$nfs_stat");
                        $self->{traffic_running} = 1;
			} else {
                        $Log->comment("NFS(ops per sec)\t$nfs_stat (WARNING - NFS ops per second is less than $nfs_threshold)");
                        $self->{traffic_running} = 0;
			}
        } elsif ($protocol =~ /cifs/i) {
                my $cifs_stat = $parsed_op->[0]->{'Avg'}{'CIFS'};
                $self->{cifsops} = $cifs_stat;
                if($cifs_stat > $cifs_threshold){
                        $Log->comment("CIFS(ops per sec)\t$cifs_stat");
                        $self->{traffic_running} = 1;
                        } else {
                        $Log->comment("CIFS(ops per sec)\t$cifs_stat (WARNING - CIFS ops per second is less than $cifs_threshold)");
                        $self->{traffic_running} = 0;
			}
        } elsif ($protocol =~ /iscsi/i) {
                my $iscsi_stat =$parsed_op->[0]->{'Avg'}{'iSCSI'};
                $self->{iscsiops} = $iscsi_stat;
                if($iscsi_stat > $iscsi_threshold){
                        $Log->comment("iSCSI(ops per sec)\t$iscsi_stat");
                        $self->{traffic_running} = 1;
                        } else {
                        $Log->comment("iSCSI(ops per sec)\t$iscsi_stat (WARNING - iSCSI ops per second is less than $iscsi_threshold)");
                        $self->{traffic_running} = 0;
			}
        }

	$Log->comment("\nChecking Multiprocessor capabilities on $aggrnode");
        $response =  $api_obj->sysstat ('count' => 3,
                'interval'    => 1,
                "print-summary"=> 1,
                'multiprocessor-statistcs' => 1
                );
        $parsed_op = $response->get_parsed_output();
        my $response2 =  $api_obj->sysconfig ();
        my $parsed_op2 = $response2->get_parsed_output();
        my $cores  = $parsed_op2->[0]->{'slot_information'}[0]{'processors'};
        my $count=0;
        $Log->comment("Processor\tUtilization");
        while($count<$cores)
        {
                my $cpu  = substr($parsed_op->[0]->{'Avg'}{"CPU$count"},0,-1);
                $Log->comment("CPU$count\t\t$cpu%");
                if($cpu <= $cpu_threshold) {
                        $Log->warn("WARNING - CPU".$count." is not being utilized, value is $cpu% and threshold value is $cpu_threshold%");
                }
                $count++;
        }

	if(0){
	$Log->comment("\nChecking IPv4 over TCP on $lifnode");
	my $stats=NACL::C::Statistics->show_periodic(
					command_interface       =>      $Node,
					'object'                =>      'oncrpcmsg',
					'instance'              =>      'oncrpc_tcp_cpu0',
					'counter'               =>      'num_calls_IPv4',
					'interval'              =>      '3',
					'iterations'            =>      '5',
					'summary'               =>      'true',
					'node'                  =>      $Node->node()
					);
	my $max = $stats->{'maximums'}->{'num_calls_ipv4'};
	my $min = $stats->{'minimums'}->{'num_calls_ipv4'};
	my $calls = $max - $min;
	if($calls>0){
		$Log->comment("Number of IPv4 calls over TCP: $calls");
		}
	else    {
		$Log->warn("Number of IPv4 calls over TCP: $calls (WARNING: This value should be greater than 0)");
		}
	}
	if(0){
	$Log->comment("\nChecking IPv6 over TCP on $lifnode");
	my $stats=NACL::C::Statistics->show_periodic(
					command_interface       =>      $Node,
					'object'                =>      'oncrpcmsg',
					'instance'              =>      'oncrpc_tcp_cpu0',
					'counter'               =>      'num_calls_IPv6',
					'interval'              =>      '3',
					'iterations'            =>      '5',
					'summary'               =>      'true',
					'node'                  =>      $Node->node()
					);
	my $max = $stats->{'maximums'}->{'num_calls_ipv6'};
	my $min = $stats->{'minimums'}->{'num_calls_ipv6'};
	my $calls = $max - $min;
	if($calls>0){
		$Log->comment("Number of IPv6 calls over TCP: $calls");
		}
	else    {
		$Log->warn("Number of IPv6 calls over TCP: $calls (WARNING: This value should be greater than 0)");
		}
	}

        return $self;
}


1;


=pod

=head1 NAME

Network::SCONTraffic


=head1 DESCRIPTION

SCON traffic.

=head1 SYNOPSIS

    use Network::SCONTraffic;
    $scontraffic = Network::SCONTraffic->new();


=head1 STATUS

    public

=head1 AUTHOR

    isarika@netapp.com

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved


=cut


