# Perl Module
# $Id$
# Last Revised $Date$
# vim: ft=perl tabstop=3 expandtab
#
# Module for access to SCON DB
#
# Non-NATE lib but can be used by NATE 
#
#
package Network::SCON_DB::IPAddrManager;

use strict;
use Network::SCON_DB::IPMsite;
use DBI;

BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter);
   our $VERSION       = '1.0.0';
}   

# place holder for future module TAGS
our @TAGS_1 = qw(ip_manager_validate_ip_version);
our @TAGS_2 = ();

# Export functions
our @EXPORT = (@TAGS_1, @TAGS_2);

our @EXPORT_OK   = qw();
our %EXPORT_TAGS = ( 'iptools'  => [@TAGS_1],
                     'tag2'     => [@TAGS_2],
);

our $ERROR = '';

our %IPAddrManagercfg = (
    'hostname' => 'csiqa-jams-admin-prod.rtp.netapp.com',
    'database' => 'SCON',
    'username' => 'sqluser',
    'password' => '',
);

our @IP_ADDRESS_COLUMNS = (
      ['ip_version',        'str'],
      ['ip_address',        'str'],
      ['ip_address_bits',   'int'],
      ['ip_netmask',        'str'],
      ['ip_netmask_bits',   'int'],
      ['ip_network',        'str'],
      ['ip_network_bits',   'int'],
      ['ip_broadcast',      'str'],
      ['ip_broadcast_bits', 'int'],
      ['ip_gateway',        'str'],
      ['hosts',             'int'],
      ['host_index',        'int'],
      ['host_type',         'str'],
      ['role',              'str'],
      ['vlan_id',           'int'],
      ['passphrase',        'str'],
);


#=============================================================================
#
#  Constructor:  create an object for JAMS DB connection
#
# Args:
#    site_id   =>  NATE site is, will define DB connct values
#  DB connection 
#    hostname  => db server hostname
#    database  => databsase name
#    username  => database username
#    password  => username password for the database
#  Misc
#    debug     => int
#
#
# Object structure
#    $_obj->{'dbh'} = datbase handle
#    $_obj->{'cfg'}{'hostname'}           # DB info
#                  {'database'}
#                  {'username'}
#                  {'password'}
#    $_obj->{'last_sql'}           =>  <last sql statement> 
#    $_obj->{'ip_manager'}{'role'} => [r1, r2, ..]           list of roles form enum list
#                         {'host_type'} = [t1, t2, t3, ...]  type of ip adderess
#                         {'ip_version'} = [v1, v2, v3, ...] ip version
#           {'ip_address_range'}   => '1.0.0.6-1.0.0.10',
#           {'ip_address_request'} => [ [], [], ]
#           {'ip_address_list'}    => [ip, ip, ip,...]
#           {'attrib'}{'hosts'}    => number of host
#           {'attrib'}{'ip_version'} => <ip version>
#           {'attrib'}{'ip_network'} => <network id>
#           {'attrib'}{ip_netmask}   => <netmask>
#           {'attrib'}{'ip_address_bits'} => <address bits>
#           {'attrib'}{'ip_broadcast'}  => <broadcast address>
#           {'attrib'}{'ip_broadcast_bits'}
#           {'attrib'}{'ip_gateway'}  => <gateway>
#           {'attrib'}{'ip_network_bits'} 
#           {'attrib'}{'ip_netmask_bits'}
#    $_obj->{''_ip_addr_req_boundries'}{'end'}{'id'} = row id of ending ip addr
#                                             {'host'} = host index of ending ip addr
#    $_obj->{''_ip_addr_req_boundries'}{'start'}{'id'} = row id of starting ip addr
#                                               {'host'} = host index of starting ip addr
#
sub new {

   my $_proto = shift;
   my $_class = ref($_proto) || $_proto;
   my $_self  = {};
   # bless object
   bless($_self, $_class);

   # get object arguments
   my %_arg = @_;

   my($_a, $_k, $_dsn, $_sql, $_resp, @_row_array, $_site_href);

   $ERROR = '';

   # check args

   # site arg will set the DB server info first
   # can be overwrriten by individual setting though
   if ( defined($_arg{'site_id'}) ) {
      ($_site_href, $ERROR)  = &ipm_site($_arg{'site_id'});
   }
   else {
      ($_site_href, $ERROR)  = &ipm_site($ENV{'SITE'});
   }

   # check to see if there any args to overwrite 
   foreach $_a (keys %_arg) {
      if    ( $_a eq 'hostname' )     { $_self->{'cfg'}{'hostname'} = delete($_arg{$_a}); }
      elsif ( $_a eq 'database' )     { $_self->{'cfg'}{'database'} = delete($_arg{$_a}); }
      elsif ( $_a eq 'username' )     { $_self->{'cfg'}{'username'} = delete($_arg{$_a}); }
      elsif ( $_a eq 'password' )     { $_self->{'cfg'}{'password'} = delete($_arg{$_a}); }
      elsif ( $_a eq 'debug' )        { $_self->{'debug'}           = delete($_arg{$_a}); }
      elsif ( $_a eq 'site_id' )      { $_self->{'cfg'}{'site_id'}  = delete($_arg{$_a}); }
      else {
         $ERROR = "unsupported argument: $_a";
         return( wantarray ? (undef, $ERROR ) : undef);
      }
   }

   # make sure we have config info either from options or site_id
   foreach $_a (keys %{$_site_href}) {
      if (!defined( $_self->{'cfg'}{$_a})) 
         { $_self->{'cfg'}{$_a} = $_site_href->{$_a}; }
   }

   # check we have a host and db name for connection
   foreach $_a ( 'hostname', 'database' ) {
      if ( !defined( $_self->{'cfg'}{$_a}) ) {
         $ERROR = "site info for $_a not defined";
         return( wantarray ? (undef, $ERROR ) : undef);
      } 
   }
   # make DB connection
   $_dsn = sprintf("DBI:mysql:database=%s;host=%s",
      $_self->{'cfg'}{'database'},
      $_self->{'cfg'}{'hostname'}
   );
   # make DB conenction
   $_self->{'dbh'} = DBI->connect( $_dsn,
                                   $_self->{'cfg'}{'username'},
                                   $_self->{'cfg'}{'password'}
   );
   # validate connection 
   unless ( $_self->{'dbh'} )  {
      $ERROR = $DBI::errstr;
      return( wantarray ? (undef, $ERROR ) : undef);
   }
   # TODO this could be controlled by constructor args
   $_self->{'dbh'}->{mysql_auto_reconnect} = 1;
   $_self->{'dbh'}->{RaiseError}           = 1;
   $_self->{'dbh'}->{PrintError}           = 0;
   $_self->{'dbh'}->{PrintWarn}            = 0;
   #$_self->{'dbh'}->{ShowErrorStatement}   = 1;


   # populate object with common values (enums)
   #
   # roles: this may in future be populated from a DB query
   # 
   #  another way
   #     describe ip_manager role; 
   #     then parse the output      
   $_sql =  "SELECT SUBSTRING(COLUMN_TYPE,5) AS RoleList ";
   $_sql.=  "FROM information_schema.COLUMNS ";
   $_sql.=  "WHERE TABLE_SCHEMA=" . sprintf("'%s' ", $_self->{'cfg'}{'database'}) ;
   $_sql.=  "  AND TABLE_NAME='ip_manager' ";
   $_sql.=  "  AND COLUMN_NAME='role'";

   @_row_array = $_self->{'dbh'}->selectrow_array($_sql);
   if (scalar(@_row_array) == 0 ){
      $ERROR = "failed to get ip_manager.roles list: $DBI::errstr";
      return( wantarray ? (undef, $ERROR ) : undef);
   }
   @{$_self->{'ip_manager'}{'role'}} = eval $_row_array[0];
   #
   # host types
   #
   $_sql =~ s/role/host_type/;
   @_row_array = $_self->{'dbh'}->selectrow_array($_sql);
   if (scalar(@_row_array) == 0 ){
      $ERROR = "failed to get ip_manager.host_type list: $DBI::errstr";
      return( wantarray ? (undef, $ERROR ) : undef);
   }
   @{$_self->{'ip_manager'}{'host_type'}} = eval $_row_array[0];
   #
   # versions
   #
   $_sql =~ s/host_type/ip_version/;
   @_row_array = $_self->{'dbh'}->selectrow_array($_sql);
   if (scalar(@_row_array) == 0 ){    
      $ERROR = "failed to get ip_manager.ip_version list: $DBI::errstr";
      return( wantarray ? (undef, $ERROR ) : undef);
   }
   @{$_self->{'ip_manager'}{'ip_version'}} = eval $_row_array[0];


   @{$_self->{'messages'}} = ();

   return ( wantarray ? ($_self, '' ) : $_self);

}

#
#-----------------------------------------------------------------------------
#

sub scon_db_disconnect{
   my $_self = shift;

   printf("jams_db_disconnect: [%s] [%s]\n",
      $_self->{'cfg'}{'hostname'},
      $_self->{'cfg'}{'database'}
   ) if $_self->{'debug'};
   $_self->{'dbh'}->disconnect();
   1;
}
#
#-----------------------------------------------------------------------------
# Insert IP address row into ip_mamager table
#
# Args
#   'ip_version'         => '',
#   'ip_address'         => '',
#   'ip_address_bits'    => '',
#   'ip_netmask'         => '',
#   'ip_netmask_bits'    => '',
#   'ip_network'         => '',
#   'ip_network_bits'    => '',
#   'ip_broadcast'       => '',
#   'ip_broadcast_bits'  => '',
#   'ip_gateway'         => '',
#   'hosts'              => '',
#   'host_index'         => '',
#   'host_type'          => '',
#   'role'               => '',
#   'vlan_id'            => '',
#   'passphrase'         => '',
#
# Return
#   return( wantarray ? (<status>, $ERROR ) : <status>);
#

sub insert_ip_manager {

   my $_obj = shift;
   my %_args = @_;
   $ERROR = '';

   my ($_sql, $_ok, $_aref, $_r, $_t );

   $_args{'role'} = &normalize_role($_args{'role'});

   # default the passphrase
   if (!defined($_args{'passphrase'})) { $_args{'passphrase'} = 'default'; } 

   # loop thru the columns and
   # make sql statement and check for all args
   $_sql = 'INSERT INTO ip_manager SET ';
   foreach $_aref ( @IP_ADDRESS_COLUMNS ) {
      if ( !defined($_args{$_aref->[0]}) ) {
         $ERROR = "arg $_aref->[0] not given";
         return( wantarray ? (undef, $ERROR ) : undef);         
      }
      next if $_aref->[0] eq 'role';
      next if $_aref->[0] eq 'host_type';
      if ( $_aref->[1] eq 'int') {  $_sql .= sprintf("`%s` = %u, ",   $_aref->[0], $_args{$_aref->[0]}); }
      if ( $_aref->[1] eq 'str') {  $_sql .= sprintf("`%s` = '%s', ", $_aref->[0], $_args{$_aref->[0]}); }
   }
   # check enumerated types columns
   foreach $_t ( 'role', 'host_type') {
      # role is an enum, so verify enumerated values
      if ($_t eq 'role' )      { $_ok = $_obj->ip_manager_validate_role($_args{'role'}); }
      if ($_t eq 'host_type' ) { $_ok = $_obj->ip_manager_validate_host_type($_args{'host_type'}); }

      unless ($_ok) {
         $ERROR = sprintf("arg %s: [$_args{$_t}] not supported");
         return( wantarray ? (undef, $ERROR ) : undef);
      }
      $_sql .= sprintf("`%s` = '%s', ", $_t, $_args{$_t});
   }
   $_sql =~ s/,\s+$//;

   # store the SQL statement   
   $_obj->{'last_sql'} = $_sql;

   # Send the sql insert
   eval {
      $_ok = $_obj->{'dbh'}->do($_sql);
      # get row id
      if ($_ok)
         { $_r = $_obj->{'dbh'}->last_insert_id(undef, undef, undef,undef); }
   };

   # return
   # error
   if ($@) {
      $ERROR= $DBI::err . ": " . $@;
      return( wantarray ? (undef, $ERROR) : undef);
   }
   #success
   else {
      return( wantarray ? ($_r, '') : $_r);
   }

}

#-----------------------------------------------------------------------------
# Insert IP address row into ip_mamager table value list
# Value list will be used, when called to make bulk sql
# Insert statement
#   INSERT INTO foo ( var1, var2, var3) 
#   VALUES
#     ( val1.1, val1.2, val1.3),
#     ( val2.1, val2.2, val2.3)
#
# 
# API same as &insert_ip_manager 
#
sub ip_manager_store_value {

   my $_obj = shift;
   my %_args = @_;
   my @_tmp = ();
   my ($_aref, $_ok);
   $ERROR = '';

   $_args{'role'} = &normalize_role($_args{'role'});

   # default the passphrase
   if (!defined($_args{'passphrase'})) { $_args{'passphrase'} = 'default'; }

   # check args and populate VALUES list for SQL statement
   foreach $_aref ( @IP_ADDRESS_COLUMNS ) {
      if ( !defined($_args{$_aref->[0]}) ) {
         $ERROR = "arg $_aref->[0] not given";
         return( wantarray ? (undef, $ERROR ) : undef);
      }
      if ($_aref->[0] eq 'role' ) {
         $_ok = $_obj->ip_manager_validate_role($_args{'role'});
         unless ($_ok) {
            $ERROR = "role: [$_args{'role'}] not supported";
            return( wantarray ? (undef, $ERROR ) : undef);
         }
      }
      if ($_aref->[0] eq 'host_type' ) {
         $_ok = $_obj->ip_manager_validate_host_type($_args{'host_type'});
         unless ($_ok) {
            $ERROR = "host_type: [$_args{'host_type'}] not supported";
            return( wantarray ? (undef, $ERROR ) : undef);
         }
      }

      push(@_tmp, $_args{$_aref->[0]}); 
   }
   # populate object value list
   push( @{$_obj->{'_ip_manager_values'}}, [@_tmp]);
 
   return( wantarray ? (scalar(@_tmp), '') : scalar(@_tmp));
}
#-----------------------------------------------------------------------------
# Clear objects IP Manager Value list
#
sub ip_manager_clear_value {
   my $_obj = shift;
   @{$_obj->{'_ip_manager_values'}} = ();
   return( wantarray ? (1, '') : 1);
}
#-----------------------------------------------------------------------------
# IP Manager Vlaue list count
#
sub ip_manager_value_count {
   my $_obj = shift;
   scalar(@{$_obj->{'_ip_manager_values'}});
}

#-----------------------------------------------------------------------------
#  Insert values stoed in @{$_obj->{'_ip_manager_values'}} 
#   INSERT INTO foo ( var1, var2, var3) 
#   VALUES
#     ( val1.1, val1.2, val1.3),
#     ( val2.1, val2.2, val2.3)
# Args
#   none
#
# Return
#
sub insert_ip_manager_values {
   my $_obj = shift;
   $ERROR = '';
   my ($_aref, $_v, $_i, $_ok);

   # make list of columns
   my $_sql = "INSERT INTO ip_manager \n( ";
   foreach $_aref ( @IP_ADDRESS_COLUMNS ) {
      $_sql .= sprintf("`%s`, ", $_aref->[0]);
   }
   $_sql =~ s/,\s+$//;
   $_sql .= ") \nVALUES\n ";

   # add to VALUES list
   foreach $_aref ( @{$_obj->{'_ip_manager_values'}} ) {
      $_sql .= "   (";
      $_i=0;
      foreach $_v ( @{$_aref} ) {
         if ( $IP_ADDRESS_COLUMNS[$_i][1] eq 'str')  { $_sql .= sprintf("'%s', ", $_v); }
         if ( $IP_ADDRESS_COLUMNS[$_i][1] eq 'int')  { $_sql .= sprintf("%u, ", $_v); } 
         $_i++;
      }
      $_sql =~ s/,\s+$//;
      $_sql .= "), \n ";
   }
   $_sql =~ s/,\s+$//s;

   # store sql statement in object
   $_obj->{'last_sql'} = $_sql;

   # Send the sql insert
   eval {
      $_ok = $_obj->{'dbh'}->do($_sql);
   };

   # return
   # error 
   if ($@) {
      $ERROR= $DBI::err . ": " . $@;
      return( wantarray ? (undef, $ERROR) : undef);
   }
   #success
   else {
      return( wantarray ? ($_ok, '') : $_ok);
   }
}

#
#-----------------------------------------------------------------------------
#
# Get IP addresses for given network_id
#
# Args (defaults in [])
#   'hosts'        => <int>        how many host    [1]
#   'host_type'    => <str>,      ('net','bcast','host') [host]
#   'query'        => <any|order>  [any]
#   'network_id'   => <str>,
#   'vlan_id'      => <int>,
#   'version'      = '4|6',       ('4','6')  [v4]
#   'role'         => <str>,      ('free','locked','data','clus','mgmt','inter-clus') [data]
#   'job_id'       => <str>,
#   'passphrase'   => <str>,
# Return
#  return( wantarray ? (status, $ERROR ) : status);
#
sub ip_address_request {

   my $_obj = shift;
   my %_args = @_;
   $ERROR = '';

   my ($_aref, $_sql, $_resp, 
       @_start, @_end,
       $_last_idx, $_ok, $_consect, $_consect_max,
       $_r, $_m, $_u,
       $_href, @_addrs, $_remote_ip, $_local_ip, $_l, $_f, 
   );
   # init the messages
   @{$_obj->{'messages'}} = ();
   push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: check args');

   # check args and set defaults
   #
   # host 
   # default to 1
   if ( !defined( $_args{'hosts'}) ) {  $_args{'hosts'} = 1; }
   else {
      if ( $_args{'hosts'} !~ /^\d+$/ ) {
         $ERROR = "IPAddrManager::ip_address_request: hosts arg not an int [$_args{'hosts'}]";
         return( wantarray ? (undef, $ERROR ) : undef); 
      }
   }
   # host_type 
   # default to host
   if ( !defined( $_args{'host_type'}) ) {  $_args{'host_type'} = 'host'; }
   else {
      $_ok = $_obj->ip_manager_validate_host_type($_args{'host_type'});
      unless ($_ok) {
         $ERROR = "IPAddrManager::ip_address_request: host_type option not supported: [$_args{'host_type'}]";
         return( wantarray ? (undef, $ERROR ) : undef);
      }      
   }
   # query
   # default to any
   if ( !defined( $_args{'query'}) ) {  $_args{'query'} = 'any'; }
   else {
      if ( $_args{'query'} !~ /^any|order$/ ) {
         $ERROR = "IPAddrManager::ip_address_request: query option not supported: [$_args{'query'}]";
         return( wantarray ? (undef, $ERROR ) : undef);
      }
   }
   # version 
   # default to v4 
   if ( !defined( $_args{'version'}) ) {  $_args{'version'} = '4'; }
   else {
      ($_args{'version'}, $ERROR) = &ip_manager_validate_ip_version($_args{'version'});
      unless ($_args{'version'} ) {
         return( wantarray ? (undef, "IPAddrManager::ip_address_request: $ERROR" ) : undef);
      }
   }
   # role 
   # default data
   if ( !defined( $_args{'role'}) ) {  $_args{'role'} = 'data'; }
   else {
      $_args{'role'} = &normalize_role($_args{'role'});
      $_ok = $_obj->ip_manager_validate_role($_args{'role'});
      unless ($_ok) {
         $ERROR = "IPAddrManager::ip_address_request: role option not supported: [$_args{'role'}]";
         return( wantarray ? (undef, $ERROR ) : undef);
      }
   }
   # network and vlan
   # no defaults, required arg
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: network_id: [$_args{'network_id'}]");
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: vlan_id: [$_args{'vlan_id'}]");
   if ( !defined( $_args{'network_id'}) && !defined($_args{'vlan_id'}) ) {
      $ERROR = "IPAddrManager::ip_address_request: network_id and vlan_id arg not given";
      return( wantarray ? (undef, $ERROR ) : undef);
   }

   # job_id
   # no defaults, required arg
   if ( !defined( $_args{'job_id'}) ) {
      $ERROR = "IPAddrManager::ip_address_request: job_id arg not given";
      return( wantarray ? (undef, $ERROR ) : undef);
   }

   # passphrase
   if ( !defined( $_args{'passphrase'}) ) { $_args{'passphrase'} = 'default'; }


   # set some defaults
   # no run id use username
   if ( !defined( $_args{'run_id'}) ) {
      $_args{'run_id'} = getlogin() || (getpwuid($<))[0];
   }
   # no server, use hostname, if localhost then use IP
   # using a priv function to parse netstat output, 
   # then resolve the DB's IP and look for connection in netstat
   # better way would be to find the socket connection in the DBI and look at its
   # local ip, but not directly given by the DBI
   if ( !defined( $_args{'ndisp_server'}) ) {    
      $_args{'ndisp_server'} = sprintf("host:%s",gethostbyname('localhost')); 
      if ( $_args{'ndisp_server'} =~ /localhost/ || $_args{'ndisp_server'} =~ /template/) {
         $_href = &_parse_netstat;
         @_addrs = (gethostbyname($_obj->{'cfg'}{'hostname'}))[4];
         $_remote_ip = sprintf("%s", join('.', unpack('C4', $_addrs[0])));
         $_local_ip = '';
         foreach $_l (keys %{$_href} ) {
            foreach $_f (keys %{$_href->{$_l}} ) {
               if ( $_f eq $_remote_ip ) {
                  $_local_ip = $_l;
                  last; 
               }
            }
            last if $_local_ip;
         }
         $_args{'ndisp_server'} = sprintf("host:%s", $_local_ip) if $_local_ip;
         push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: hostname of localhost converted to ip $_local_ip");
      }
   }   
   # no port use PID
   if ( !defined( $_args{'ndisp_server_port'}) ) {
      $_args{'ndisp_server_port'} = $$;
   } 


   # init any object parts
   %{$_obj->{'_ip_addr_req_boundries'}} = ();
   %{$_obj->{'attrib'}} = ();

   # lock the table
   # TODO put a flag in the object if the table is locked
   # can be used to see if UNLOCK TABLES should be sent
   push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: LOCK TABLES ip_manager LOW_PRIORITY WRITE');
   $_obj->{'dbh'}->do('LOCK TABLES ip_manager LOW_PRIORITY WRITE');

   # either sql gets the same columns
   $_sql = "SELECT id, host_index, ip_version, ip_address, ip_address_bits, ip_netmask, ip_netmask_bits, "
            . "ip_network, ip_network_bits, ip_broadcast, ip_broadcast_bits, ip_gateway, "
            . "hosts, host_type, role, job_id, vlan_id, passphrase "
            . "FROM ip_manager ";
   #
   # requesting list of any addresses
   #
   if ( $_args{'query'} eq 'any' ) {
      # continue building sql
      $_sql .= sprintf("WHERE ISNULL\(job_id\) ");
      $_sql .= sprintf("  AND `ip_network` = '%s' ",  $_args{'network_id'}) if defined($_args{'network_id'});
      $_sql .= sprintf("  AND `vlan_id` = '%s' ",     $_args{'vlan_id'}) if defined($_args{'vlan_id'});
      $_sql .= sprintf("  AND `host_type` = '%s' ",   $_args{'host_type'});
      $_sql .= sprintf("  AND `ip_version` = '%s' ",  $_args{'version'});
      $_sql .= sprintf("  AND `role` = '%s' ",        $_args{'role'});
      $_sql .= sprintf("  AND `passphrase` = '%s' ",  $_args{'passphrase'});
      $_sql .= sprintf("ORDER BY host_index ASC ");
      $_sql .= sprintf("LIMIT %u", $_args{'hosts'});
      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: sql query for any: $_sql");
   }
   #
   # requesting list of ordered (consectutive)
   #
   elsif ( $_args{'query'} eq 'order' ) { 
      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: query = order  get available addresses');
      # make sql statement to get all available addresses
      # will loop through to find contiguous block of addresses
      #
      $_obj->{'last_sql'} = "SELECT id, host_index FROM ip_manager ";
      $_obj->{'last_sql'} .= sprintf("WHERE ISNULL\(job_id\) ");
      $_obj->{'last_sql'} .= sprintf("  AND `ip_network` = '%s' ",  $_args{'network_id'}) if defined($_args{'network_id'});
      $_obj->{'last_sql'} .= sprintf("  AND `vlan_id` = '%s' ",     $_args{'vlan_id'}) if defined($_args{'vlan_id'});
      $_obj->{'last_sql'} .= sprintf("  AND `ip_version` = '%s' ",  $_args{'version'});
      $_obj->{'last_sql'} .= sprintf("  AND `host_type` = '%s' ",   $_args{'host_type'});
      $_obj->{'last_sql'} .= sprintf("  AND `role` = '%s' ",        $_args{'role'});
      $_obj->{'last_sql'} .= sprintf("  AND `passphrase` = '%s' ",  $_args{'passphrase'});
      $_obj->{'last_sql'} .= sprintf("ORDER BY host_index ASC ");

      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request:sql query for order: $_obj->{'last_sql'}");
 
      # send the sql
      eval {
         $_resp = $_obj->{'dbh'}->selectall_arrayref($_obj->{'last_sql'});  
      };
      # error
      if ($@) {
         $ERROR= "select first list: " . $DBI::err . ": " . $@ . ": " . $_obj->{'last_sql'};
         $_obj->{'dbh'}->do('UNLOCK TABLES');
         push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
         return( wantarray ? (undef, $ERROR) : undef);
      }

      # check we got rows
      $_r = scalar(@{$_resp});
      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: sql response for available rows: $_r"); 
      unless ( $_r ) {
         $_obj->{'dbh'}->do('UNLOCK TABLES');
         push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
         return( wantarray ? ($_r, "IPAddrManager::ip_address_request: DB query return no rows: $_obj->{'last_sql'}") : $_r);
      }


      # find consectutive block   
      @_start    = ();
      @_end      = ();
      $_last_idx = 0;
      $_consect  = 0;
      $_consect_max = 0;

      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: loop through available addresses');
      # loop thru response 
      foreach $_aref ( @{$_resp} ) {
         # start of a block
         if ( scalar(@_start) == 0 ) {
            $_consect++;
            @_start = @{$_aref};
            @_end   = @{$_aref};
         }
         # diff of 1 between host_index
         elsif ( abs($_aref->[1] - $_last_idx) == 1 ) {
            $_consect++;
            @_end   = @{$_aref};
         } 
         # diff is greater than 1, not contiguous, but start over
         else {
            $_consect = 1;
            @_start = @{$_aref};
            @_end   = @{$_aref};
         }
         # next iteration needs previous value
         $_last_idx = $_aref->[1];
   
         # found desired amount
         last if $_consect == $_args{'hosts'};

         $_consect_max = $_consect if $_consect > $_consect_max;
      }
      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: Consecutive count: $_consect max: $_consect_max");

      # Did not find enough
      # 
      if ( $_consect != $_args{'hosts'} ) {
         $ERROR = "IPAddrManager::ip_address_request: requested $_args{'hosts'} not equal to consectutive max $_consect_max";
         $_obj->{'dbh'}->do('UNLOCK TABLES');
         return( wantarray ? (undef, $ERROR) : undef);    
      } 

      # populate boundary data structures, inited above
      $_obj->{'_ip_addr_req_boundries'}{'start'}{'id'}   = $_start[0];
      $_obj->{'_ip_addr_req_boundries'}{'start'}{'host'} = $_start[1];
      $_obj->{'_ip_addr_req_boundries'}{'end'}{'id'}     = $_end[0];
      $_obj->{'_ip_addr_req_boundries'}{'end'}{'host'}   = $_end[1];

      # build out sql statement to match given criteria
      # and to get the addresses within the found boundaries
      $_sql .= sprintf("WHERE ISNULL(job_id) ");
      $_sql .= sprintf("  AND `ip_network` = '%s' ",  $_args{'network_id'}) if defined($_args{'network_id'});
      $_sql .= sprintf("  AND `vlan_id` = '%s' ",     $_args{'vlan_id'}) if defined($_args{'vlan_id'});
      $_sql .= sprintf("  AND `host_type` = '%s' ",   $_args{'host_type'});
      $_sql .= sprintf("  AND `ip_version` = '%s' ",  $_args{'version'});
      $_sql .= sprintf("  AND `role` = '%s' ",        $_args{'role'});
      $_sql .= sprintf("  AND `host_index` >= %u ",   $_start[1]);
      $_sql .= sprintf("  AND `host_index` <= %u ",   $_end[1]);
      $_sql .= sprintf("ORDER BY host_index ASC ");

      # store sql in object
      $_obj->{'last_sql'} = $_sql;
      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: sql query for order: $_sql");
   }
   # unknow query 
   # we should not get here cause of previous arg check
   # but put the code here anyways
   else {
      $ERROR = "IPAddrManager::ip_address_request: query option not supported [$_args{'query'}]";
      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES'); 
      $_obj->{'dbh'}->do('UNLOCK TABLES');
      return( wantarray ? (undef, $ERROR ) : undef);
   }

   # store sql in object
   $_obj->{'last_sql'} = $_sql;
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: dbh sql: $_obj->{'last_sql'}");
 
   # init the data structure
   @{$_obj->{'ip_address_request'}} = ();

   # send the sql
   eval {
      $_obj->{'ip_address_request'} = $_obj->{'dbh'}->selectall_arrayref($_sql);
   };

   # return if error
   if ($@) {
      $ERROR= 'IPAddrManager::ip_address_request:' . $DBI::err . ": " . $@ . ": " . "sql: $_obj->{'last_sql'}";
      $_obj->{'dbh'}->do('UNLOCK TABLES');
      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # set row count
   $_r = scalar(@{$_obj->{'ip_address_request'}});
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: ip_address_request rows: $_r");

   # check we have desired amount of addresses
   if ( $_r != $_args{'hosts'} ) {
      $ERROR = "IPAddrManager::ip_address_request: ip_address_request only has $_r rows, requesting $_args{'hosts'}";
      $_obj->{'dbh'}->do('UNLOCK TABLES');
      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
      return( wantarray ? (undef, $ERROR) : undef);
   }
   #
   # Update the DB rows so as to reserve the addresses
   #
   my @_map = (
      [2, 'ip_version'],
      [4, 'ip_address_bits'],
      [5, 'ip_netmask'],
      [6, 'ip_netmask_bits'],
      [7, 'ip_network'],
      [8, 'ip_network_bits'],
      [9, 'ip_broadcast'],
      [10, 'ip_broadcast_bits'],
      [11, 'ip_gateway'],
      [12, 'hosts'],
      [17, 'vlan_id'],
   );

   # sql statement to update desired row
   $_sql = sprintf("UPDATE ip_manager SET ");
   $_sql .= sprintf(" `job_id` = '%s', ", $_args{'job_id'});
   $_sql .= sprintf(" `run_id` = '%s', ", $_args{'run_id'});
   $_sql .= sprintf(" `ndisp_server` = '%s', ", $_args{'ndisp_server'}); 
   $_sql .= sprintf(" `ndisp_server_port` = '%s', ", $_args{'ndisp_server_port'}); 
   $_sql .= sprintf(" `date_given` = NOW\(\), ");
   $_sql .= sprintf(" `date_returned` = NULL ");
   $_sql .= sprintf(" WHERE `id` IN \(");


   # loop thru addresses make list and pull out singular data defined in @_map
   foreach $_aref ( @{$_obj->{'ip_address_request'}} ) {
      push( @{$_obj->{'ip_address_list'}}, $_aref->[3] );   # add ip to list
      $_sql .= "$_aref->[0], ";                             # add row id to sql

      # make sure all info is same and store it in object
      # if singular data has a diff, error out
      foreach $_m ( @_map ) {
         if ( !defined($_obj->{$_m->[1]}) ) { $_obj->{'attrib'}{$_m->[1]} = $_aref->[$_m->[0]]; }
         elsif ( $_obj->{$_m->[1]} ne $_aref->[$_m->[0]]  ) {
            $ERROR = sprintf("IPAddrManager::ip_address_request: diff: %s from row %u [%s] [%s]",
               $_m->[1],
               $_aref->[0],
               $_obj->{'attrib'}{$_m->[1]},
               $_aref->[$_m->[0]]
            );
            $_obj->{'dbh'}->do('UNLOCK TABLES');
            push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
            return( wantarray ? (undef, $ERROR) : undef);  
         }
      }
   }
   # remove last ',' and close the sql list
   $_sql =~ s/,\s+$//;
   $_sql .= ")";


   $_obj->{'attrib'}{'_host_counts'} = $_r;
   $_obj->{'ip_address_range'} = sprintf("%s-%s", 
      $_obj->{'ip_address_request'}[0][3],
      $_obj->{'ip_address_request'}[-1][3],
   );

   # store sql in object
   $_obj->{'last_sql'} = $_sql;
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_request: dbh sql: $_obj->{'last_sql'}");

   # send the update sql
   # send the sql
   eval {
      $_ok = $_obj->{'dbh'}->do($_sql);
   };
   # return if error
   if ($@) {
      $ERROR= "IPAddrManager::ip_address_request: update ip_manager: " . $DBI::err . ": " . $@ . ": " . "sql: $_obj->{'last_sql'}";
      $_obj->{'dbh'}->do('UNLOCK TABLES');
      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # check that $_r == $ok
   #    $_r is how many rows expected to update
   #    $_ok is how many rows got updated
   if ( $_r != $_ok ) {
      $ERROR = "update expected on $_r rows, only did $_ok";
      $_obj->{'dbh'}->do('UNLOCK TABLES');
      push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
      return( wantarray ? (undef, $ERROR) : undef);
   }


   $_obj->{'dbh'}->do('UNLOCK TABLES');
   push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_request: dbh do: UNLOCK TABLES');
   return( wantarray ? ($_r,'') : $_r);

};

#
#-----------------------------------------------------------------------------
# 
# Return IP address, clear the job_id of the row so as to make address available
#
# Args
#   'job_id'       => <str>
#   'address'      => <str>   single address
#   'address_aref' => <ref>   ref to list of addresses
# 

sub ip_manager_return {

    my $_obj = shift;
    my %_args = @_;
    $ERROR = '';
   
   my ($_sql, $_ok);

   # TODO when status table has a 'free' status, update it here 
   # make sql statement
   $_sql  = sprintf("UPDATE ip_manager SET `job_id` = NULL, ");
   $_sql .= sprintf("       `date_returned` = NOW\(\), ");
   $_sql .= sprintf("       `date_given` = NULL ");

   # job_id given
   if (  $_args{'job_id'} ) {
      $_sql .= sprintf(" WHERE `job_id` = '%s' ", $_args{'job_id'} );
   }
   # single address
   elsif ( $_args{'address'}  ) {
      $_sql .= sprintf(" WHERE `ip_address` = '%s' ", $_args{'address'} );
   }
   # list of addresses
   #  ... WHERE ip_address IN ( ip1, ip2, ip3, ..)
   elsif ( ref($_args{'address_aref'}) eq 'ARRAY'  ) {
      $_sql .= sprintf(" WHERE `ip_address` IN \(");      
      foreach $_ok ( @{$_args{'address_aref'}} ) 
      { $_sql .= "'$_ok', "; }
      $_sql =~ s/,\s+$//;   # strip trailing ','
      $_sql .= "\)";        # close the paren
   }
   else  {
      $ERROR = "IPAddrManager::ip_manager_return: no address given";
      return( wantarray ? (undef, $ERROR ) : undef);
   }

   # store sql in object
   $_obj->{'last_sql'} = $_sql;

   # send the sql
   eval {
      $_ok = $_obj->{'dbh'}->do($_sql);
   };

   # return if error
   if ($@) {
      $ERROR= "IPAddrManager::ip_manager_return: " . $DBI::err . ": " . $@;
      $_obj->{'dbh'}->do('UNLOCK TABLES');
      return( wantarray ? (undef, $ERROR) : undef);
   }
   return( wantarray ? ($_ok,'') : $_ok);
}

#-----------------------------------------------------------------------------
# given job_id, ipversion and role
# find existing reservation
# 
# Args
#   'job_id'   => <str>
#   'version'  => <4|6>
#    'role'    => <dat|cluster|inter-cluster|management>
#
# for now i'm keeping the passphase arg out of this
# so that reuse is solely on runid
#
sub ip_address_reuse {

   my $_obj = shift;
   my %_args = @_;
   $ERROR = '';
   my ($_sql, $_resp, $_aref, $_r, $_m );


   if (!defined($_args{'job_id'})) { 
      $ERROR = 'IPAddrManager::ip_address_reuse: requires arg: job_id';
      return( wantarray ? (undef, $ERROR) : undef); 
   }

   $_sql = "SELECT id, host_index, ip_version, ip_address, ip_address_bits, ip_netmask, ip_netmask_bits, "
            . "ip_network, ip_network_bits, ip_broadcast, ip_broadcast_bits, ip_gateway, "
            . "hosts, host_type, role, job_id, vlan_id "
            . "FROM ip_manager "; 
   $_sql .= sprintf("WHERE job_id = '%s' ", $_args{'job_id'});
   $_sql .= sprintf("  AND ip_version  = '%s' ", $_args{'version'})   if $_args{'version'};
   $_sql .= sprintf("  AND role = '%s' ", $_args{'role'})             if $_args{'role'};
   $_sql .= sprintf("  AND ip_network = '%s' ", $_args{'network_id'}) if $_args{'network_id'};
   $_sql .= sprintf("  AND vlan_id = '%s' ", $_args{'vlan_id'})       if $_args{'vlan_id'};
   $_sql .= sprintf("ORDER BY ip_network,host_index ASC ");

   # store sql in object but do not over-write more important sql
   $_obj->{'last_sql'} = $_sql;
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_reuse: dbh sql: $_sql");

   # init the data structure
   @{$_obj->{'ip_address_request'}} = ();

   # send the sql
   eval {
      $_obj->{'ip_address_request'} = $_obj->{'dbh'}->selectall_arrayref($_sql);
   };

   # return if error
   if ($@) {
      $ERROR= 'IPAddrManager::ip_address_reuse:' . $DBI::err . ": " . $@ . ": " . "sql: $_obj->{'last_sql'}";
      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_reuse: $ERROR");
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # set row count
   $_r = scalar(@{$_obj->{'ip_address_request'}});
   push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_reuse: ip_address_request rows: $_r");

   # check we have desired amount of addresses
   unless ( $_r ) {
      $ERROR = "IPAddrManager::ip_address_reuse: ip_address_request only has $_r rows, requesting $_args{'hosts'}";
      push(@{$_obj->{'messages'}}, "IPAddrManager::ip_address_reuse: dbh do: $ERROR");
      return( wantarray ? (undef, $ERROR) : undef);
   }

   my @_map = (
      [2, 'ip_version'],
      [4, 'ip_address_bits'],
      [5, 'ip_netmask'],
      [6, 'ip_netmask_bits'],
      [7, 'ip_network'],
      [8, 'ip_network_bits'],
      [9, 'ip_broadcast'],
      [10, 'ip_broadcast_bits'],
      [11, 'ip_gateway'],
      [12, 'hosts'],
      [17, 'vlan_id'],
   );

   # loop thru addresses make list and pull out singular data defined in @_map
   foreach $_aref ( @{$_obj->{'ip_address_request'}} ) {
      push( @{$_obj->{'ip_address_list'}}, $_aref->[3] );   # add ip to list
      # make sure all info is same and store it in object
      # if singular data has a diff, error out
      foreach $_m ( @_map ) {
         if ( !defined($_obj->{$_m->[1]}) ) { $_obj->{'attrib'}{$_m->[1]} = $_aref->[$_m->[0]]; }
         elsif ( $_obj->{$_m->[1]} ne $_aref->[$_m->[0]]  ) {
            $ERROR = sprintf("IPAddrManager::ip_address_reuse: diff: %s from row %u [%s] [%s]",
               $_m->[1],
               $_aref->[0],
               $_obj->{'attrib'}{$_m->[1]},
               $_aref->[$_m->[0]]
            );
            push(@{$_obj->{'messages'}}, 'IPAddrManager::ip_address_reuse: dbh do: UNLOCK TABLES');
            return( wantarray ? (undef, $ERROR) : undef);
         }
      }
   }

   $_obj->{'attrib'}{'_host_counts'} = $_r;
   $_obj->{'ip_address_range'} = sprintf("%s-%s",
      $_obj->{'ip_address_request'}[0][3],
      $_obj->{'ip_address_request'}[-1][3],
   );

   return( wantarray ? ($_r,'') : $_r);

}



#-----------------------------------------------------------------------------
# 
# Populate object with vlan_id in DB
#
#  $_obj->{'ip_manager'}{'vlan_id'}{<vlan id>}{'Count'} = <int>    # num rows
#                                             {'vlan_id'} = <int>  # vlan id (2300)
#

sub ip_manager_vlan_ids {


   my $_obj = shift;
   $ERROR = '';
   my $_sql;

   $_sql  = sprintf("SELECT COUNT(*) AS 'Count', vlan_id FROM ip_manager ");
   $_sql .= sprintf("GROUP BY vlan_id");

   # store sql in object but do not over-write more important sql
   $_obj->{'last_sql_chk'} = $_sql;



   # send the sql
   eval {
      $_obj->{'ip_manager'}{'vlan_id'} = $_obj->{'dbh'}->selectall_hashref($_sql,['vlan_id']);
   };


   # return if error
   if ($@) { 
      $ERROR= "IPAddrManager::ip_manager_vlan_ids: " . $DBI::err . ": " . $@;
      return( wantarray ? (undef, $ERROR) : undef);
   }
   else {
      return( wantarray ? (1, '') : 1);
   }
}

#-----------------------------------------------------------------------------
# 
# Populate object with network_id in DB
#
#  $_obj->{'ip_manager'}{'ip_network'}{<network>}{'Count'} = <int>       # num rows
#                                                {'ip_network'} = <str>  # network (1.0.16.0)

sub ip_manager_network_ids {
   
   my $_obj = shift;
   $ERROR = '';
   my $_sql;

   $_sql  = sprintf("SELECT COUNT(*) AS 'Count', ip_network FROM ip_manager ");
   $_sql .= sprintf("GROUP BY ip_network");

   # store sql in object but do not over-write more important sql
   $_obj->{'last_sql_chk'} = $_sql;

   # send the sql
   eval {
      $_obj->{'ip_manager'}{'ip_network'} =  $_obj->{'dbh'}->selectall_hashref($_sql,['ip_network']);
   };


   # return if error
   if ($@) {
      $ERROR= "IPAddrManager::ip_manager_network_ids: " . $DBI::err . ": " . $@; 
      return( wantarray ? (undef, $ERROR) : undef);
   }
   else {
      return( wantarray ? (1, '') : 1);
   }
}

#-------------------------------------------------------------------------
#
# Populate object with vlan to network info
#
# Query returns
#    [vlan_id, ip_network, Count]
#
# $_obj->{'ip_manager'}{'vlan2network'}{<vlan_id>}{<ip_network>} = <int> count of network entries
#
sub ip_manager_vlan_to_network {
   my $_obj = shift;
   $ERROR = '';
   my ($_sql, $_resp, $_aref);

   $_sql  = "select DISTINCT vlan_id, ip_network, COUNT(*) AS 'Count' "
          . "FROM ip_manager GROUP BY ip_network ORDER BY  vlan_id";

   # store sql in object but do not over-write more important sql
   $_obj->{'last_sql'} = $_sql;

   # send the sql
   eval {
      $_resp = $_obj->{'dbh'}->selectall_arrayref($_sql);
   };

   # return if error
   if ($@) {
      $ERROR= "IPAddrManager::ip_manager_vlan_to_network: " . $DBI::err . ": " . $@;
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # store in object a bit more efficiently
   my $_r = 0;
   foreach $_aref ( @{$_resp} ) {
      $_obj->{'ip_manager'}{'vlan2network'}{$_aref->[0]}{$_aref->[1]} = $_aref->[2];
      $_r++;
   } 

   return( wantarray ? ($_r, '') : $_r);

}


#-------------------------------------------------------------------------
#
# Get data relating vlan, version, role
# 
# args
#   'vlan_id'    => <int>
# 
# Make sql statement union'ing bunch of separate statement
# SELECT v,h,i,r,C       FROM ip_manager WHERE v=2300 h=host i=4 r=data   GROUP BY r 
# UNION SELECT v,h,i,r,C FROM ip_manager WHERE v=2300 h=host i=6 r=data   GROUP BY r 
# UNION SELECT v,h,i,r,C FROM ip_manager WHERE v=2300 h=host i=6 r=static GROUP BY r
# UNION SELECT v,h,i,r,C FROM ip_manager WHERE v=2300 h=net  i=4 r=data   GROUP BY r
# UNION SELECT v,h,i,r,C FROM ip_manager WHERE v=2300 h=net  i=6 r=data   GROUP BY r
#
#  +---------+-----------+------------+--------+-------+
#  | vlan_id | host_type | ip_version | role   | Count |
#  +---------+-----------+------------+--------+-------+
#  |    2300 | host      | 4          | data   |  1728 |
#  |    2300 | host      | 4          | static |   308 |
#  |    2300 | host      | 6          | data   |  5001 |
#  +---------+-----------+------------+--------+-------+
#
#
sub ip_manager_vlan_type_version_role {

   my $_obj = shift;
   my %_args = @_; 
   my ($_role, $_ver, $_type, $_sql, $_resp, $_aref);

   $ERROR = '';
 
   unless ( $_args{'vlan_id'}  ) {
      $ERROR = "IPAddrManager::ip_manager_vlan_type_version_role requires vlan_id arg";
      return( wantarray ? (undef, $ERROR) : undef);
   } 

   # loop throug all values and make sql
   foreach $_type ( sort  @{$_obj->{'ip_manager'}{'host_type'}} ) {
      foreach $_ver ( sort  @{$_obj->{'ip_manager'}{'ip_version'}} ) {
         foreach $_role ( sort @{$_obj->{'ip_manager'}{'role'}} ) {
            $_sql .= sprintf("SELECT vlan_id, host_type, ip_version, role, COUNT( * ) AS `Count` ");
            $_sql .= sprintf("FROM ip_manager ");
            $_sql .= sprintf("WHERE vlan_id = '%u' ", $_args{'vlan_id'} );
            $_sql .= sprintf("AND host_type = '%s' ", $_type);
            $_sql .= sprintf("AND ip_version = '%u' ", $_ver);
            $_sql .= sprintf("AND role = '%s' ", $_role);
            $_sql .= sprintf("GROUP BY role \n");
            $_sql .= sprintf("UNION ");
         }
      }
   }    
   $_sql =~ s/UNION\s+$//;
   $_sql =~ s/\s+$//;

   # store sql in object but do not over-write more important sql
   $_obj->{'last_sql'} = $_sql;

   # send the sql
   eval {
      $_resp = $_obj->{'dbh'}->selectall_arrayref($_sql);
   };

   # return if error
   if ($@) {
      $ERROR= "IPAddrManager::ip_manager_vlan_to_network: " . $DBI::err . ": " . $@;
      return( wantarray ? (undef, $ERROR) : undef);
   }
   return( wantarray ? ($_resp, '')  : $_resp);
}

#-------------------------------------------------------------------------
#
# Get data relating vlan, version, role
# 
# args
#   'network_id'    => <int>
# 
# Make sql statement union'ing bunch of separate statement
# SELECT n,h,i,r,C       FROM ip_manager WHERE n=1.0.0.0 h=host i=4 r=data   GROUP BY r 
# UNION SELECT n,h,i,r,C FROM ip_manager WHERE n=1.0.0.0 h=host i=6 r=data   GROUP BY r 
# UNION SELECT n,h,i,r,C FROM ip_manager WHERE n=1.0.0.0 h=host i=6 r=static GROUP BY r
# UNION SELECT n,h,i,r,C FROM ip_manager WHERE n=1.0.0.0 h=net  i=4 r=data   GROUP BY r
# UNION SELECT n,h,i,r,C FROM ip_manager WHERE n=1.0.0.0 h=net  i=6 r=data   GROUP BY r
#
#  +------------+-----------+------------+--------+-------+
#  | network_id | host_type | ip_version | role   | Count |
#  +------------+-----------+------------+--------+-------+
#  |  1.0.0.0   | host      | 4          | data   |  1728 |
#  |  1.0.0.0   | host      | 4          | static |   308 |
#  |  1.0.0.0   | host      | 6          | data   |  5001 |
#  +------------+-----------+------------+--------+-------+
#
#
sub ip_manager_network_type_version_role {

   my $_obj = shift;
   my %_args = @_;
   my ($_role, $_ver, $_type, $_sql, $_resp, $_aref);


   $ERROR = '';

   unless ( $_args{'network_id'}  ) {
      $ERROR = "IPAddrManager::ip_manager_network_type_version_role requires network_id arg";
      return( wantarray ? (undef, $ERROR) : undef);
   }

   my $_addr_version = &Network::IPAddress::netip_ip_version($_args{'network_id'});


   # loop throug all values and make sql
   foreach $_type ( sort  @{$_obj->{'ip_manager'}{'host_type'}} ) {
      foreach $_ver ( sort  @{$_obj->{'ip_manager'}{'ip_version'}} ) {
         # filter versions
         next if $_ver ne "$_addr_version";
         foreach $_role ( sort @{$_obj->{'ip_manager'}{'role'}} ) {
            $_sql .= sprintf("SELECT ip_network, host_type, ip_version, role, COUNT( * ) AS `Count` ");
            $_sql .= sprintf("FROM ip_manager ");
            $_sql .= sprintf("WHERE ip_network = '%s' ", $_args{'network_id'} );
            $_sql .= sprintf("AND host_type = '%s' ", $_type);
            $_sql .= sprintf("AND ip_version = '%u' ", $_ver);
            $_sql .= sprintf("AND role = '%s' ", $_role);
            $_sql .= sprintf("GROUP BY role \n");
            $_sql .= sprintf("UNION ");
         }
      }
   }
   $_sql =~ s/UNION\s+$//;
   $_sql =~ s/\s+$//;

   # store sql in object but do not over-write more important sql
   $_obj->{'last_sql'} = $_sql;

   # send the sql
   eval {
      $_resp = $_obj->{'dbh'}->selectall_arrayref($_sql);
   };

   # return if error
   if ($@) {
      $ERROR= "IPAddrManager::ip_manager_network_to_network: " . $DBI::err . ": " . $@;
      return( wantarray ? (undef, $ERROR) : undef);
   }
   return( wantarray ? ($_resp, '')  : $_resp);
}

#==========================================================================
#
#                       Accessors
#
#
sub ip_address_range   { my $_o = shift;   $_o->{'ip_address_range'}; }
sub ip_address_list    { my $_o = shift;   $_o->{'ip_address_list'}; }
sub ip_address_request_aref { my $_o = shift;   $_o->{'ip_address_request'}; }


sub ip_address_version        { my $_o = shift;  $_o->{'attrib'}{'ip_version'}; }
sub ip_address_address_bits   { my $_o = shift;  $_o->{'attrib'}{'ip_address_bits'}; }
sub ip_address_network        { my $_o = shift;  $_o->{'attrib'}{'ip_network'}; }
sub ip_address_network_bits   { my $_o = shift;  $_o->{'attrib'}{'ip_network_bits'}; }
sub ip_address_broadcast      { my $_o = shift;  $_o->{'attrib'}{'ip_broadcast'}; }
sub ip_address_broadcast_bits { my $_o = shift;  $_o->{'attrib'}{'ip_broadcast_bits'}; }
sub ip_address_gateway        { my $_o = shift;  $_o->{'attrib'}{'ip_gateway'}; }


#===========================================================================
#
#                        Helpers
#
# Normalize role shortcuts to the enum types in the DB
#
# $_[0] = role
# return role
#
sub normalize_role {

   my $_r = $_[0];
   if    ( $_[0] eq 'clus' )          { $_r = 'cluster'; }
   elsif ( $_[0] =~ /cluster.mgmt/ )  { $_r = 'cluster_management'; }
   elsif ( $_[0] eq 'mgmt' )          { $_r = 'management'; }
   elsif ( $_[0] =~ /node.mgmt/ )     { $_r = 'management'; }
   elsif ( $_[0] eq 'inter-clus' )    { $_r = 'intercluster'; }
   
   $_r;
}


#-----------------------------------------------------------------------------
# Validate ip manager roles
# Args
#   $_[0] = <role>
#
# Return
#   0 - no match
#   1 - match
#
sub ip_manager_validate_role {

   my $_obj = shift;
   my $_input = $_[0];

   my ($_r,);
   my $_ok = 0;

 
   # find a match of input to DB 
   foreach $_r ( @{$_obj->{'ip_manager'}{'role'}} ) {
      if ( $_input eq $_r) {
         $_ok++;
         last;
      }
   }
   return ($_ok);
}

#-----------------------------------------------------------------------------
# Validate ip manager host_type
# Args
#   $_[0] = <host type>
#
# Return
#   0 - no match
#   1 - match
#
sub ip_manager_validate_host_type {

   my $_obj = shift;
   my ($_r,);
   my $_ok = 0;
   foreach $_r ( @{$_obj->{'ip_manager'}{'host_type'}} ) {
      if ($_[0] eq $_r) {
         $_ok++;
         last;
      }
   }
   return ($_ok);
}

#-----------------------------------------------------------------------------
# validate and nomalize ip version string
#  normalize to 4|6, strip off 'v' or 'inet' strings  
#  only stored as int in DB
#
# Args 
#   $_[0] = <str>
# Return
#   0    - no match
#   4|6  - matched
#
sub ip_manager_validate_ip_version {

   my $_v = $_[0];

   if    ( $_v =~ /^[4|6]$/ )         { return(wantarray ? ($_v, ''): $_v); }
   elsif ( $_v =~ /^v([4|6])$/i)      { return(wantarray ? ($1, ''): $1); }
   elsif ( $_v =~ /^inet([4|6]$)/i )  { return(wantarray ? ($1, ''): $1); }
   elsif ( $_v =~ /^inet_([4|6]$)/i ) { return(wantarray ? ($1, ''): $1); }
   else {
     return(wantarray ? (undef, "invalid ip_version string [$_v]"): undef);
   }
}
#
#-----------------------------------------------------------------------------
#
#  Parse netstat -r to find local address
#  netstat -r supported on win and unix but the clumns are different
#  This accounts for both
# Unix
# netstat -n | head -20
# Active Internet connections (w/o servers)
# Proto Recv-Q Send-Q Local Address               Foreign Address             State      
# tcp        0      0 10.227.76.243:924           10.60.118.11:2049           ESTABLISHED 
# tcp        0      0 10.227.76.243:774           10.63.63.12:2049            ESTABLISHED 
#
# Win
# C:\temp>netstat -n
# Active Connections
#   Proto  Local Address          Foreign Address        State
#   TCP    10.98.32.162:3390      10.106.52.53:59531     ESTABLISHED
#   TCP    10.98.32.162:3394      10.106.52.53:59531     ESTABLISHED
#
# Return reference to hash
#  $href->{<local addr>}{<foreign add>} = count
#
sub _parse_netstat {
   my ($_l, $_c, $_i, $_local, $_foreign);
   my @_columns = ();
   my @_values;
   my %_addr =();
   
   open(NS, 'netstat -n|');
   while ( $_l = <NS> ) {
      # find the columns
      if ( $_l =~ /Local Address/i ) {
         # strip out 'Address' so can split the columns on space
         $_l =~ s/Address//ig;
         # get column list
         @_columns = split(/\s+/, $_l);
         $_i = 0;
         # find the column for local and foreign address, keep the index
         foreach $_c (@_columns) {
            if ( $_c  =~ /Local/ )   { $_local = $_i; }
            if ( $_c  =~ /Foreign/)  { $_foreign = $_i; }
            $_i++; 
         } 
      }
      # skip if not listing protocol in first column
      next if $_l !~ /^(\s+)?tcp|udp/i;

      # line should be one hold local and foreign address   
      @_values = split(/\s+/, $_l);
   
      # strip out port at end of IP
      if ( $_values[$_local] =~ /:(\d+)$/ )   {$_values[$_local] =~ s/:\d+$//; }
      if ( $_values[$_foreign] =~ /:(\d+)$/ ) {$_values[$_foreign] =~ s/:\d+$//; }
      # populate the hash 
      $_addr{$_values[$_local]}{$_values[$_foreign]}++;
   }
   close(NS);
   return(\%_addr);
}



1;
__END__
# end of module

=head1 NAME Network::SCON_DB::IPAddrManager

=head1 DESCRIPTION

Perl module providing methods to get IP addresses for testing.


=head1 SYNOPSIS

    use Network::SCON_DB::IPAddrManager;

    my ($DB, $error)  = Network::SCON_DB::IPAddrManager->new();
    unless ($DB) { print "ERROR: $error\n"; }

    ($ok, $error) = $DB->ip_address_request(
        'hosts'        => <int>,
        'host_type'    => 'host|net|bcast|mcast',
        'query'        => 'any|order',
        'network'      => <network id>,
        'version'      => 'v4|v6',
        'role'         => 'data|clus|mgmt|free|locked|inter-clus',
        'job_id'       => <str>,
    );
    unless ($ok) { &my_exit(1, $error);}


    $DB->scon_db_disconnect;


=head1 PARAMS

=head1 Methods

=head2 new()

Constructor, wrapper around DBI to create DB connection

    $DB = Network::SCON_DB::IPAddrManager->new(hostname  => <db server hostname>
        database  => <databsase name>
        username  => <database username>
        password  => <username password for the database>
    );


=head2 scon_db_disconnect();

Method to have object disconnect from DB

    $DB->scon_db_disconnect();
 
=head2 Insert IP information

=head3 insert_ip_manager()

Method to insert data into DB to populate the DB.

    $DB->insert_ip_manager(
        'ip_version'         => <4|6>,
        'ip_address'         => <ip address>,
        'ip_address_bits'    => <int>,
        'ip_netmask'         => <ip netmask>,
        'ip_netmask_bits'    => <int>,
        'ip_network'         => <ip network>,
        'ip_network_bits'    => <int>,
        'ip_broadcast'       => <ip broadcast>,
        'ip_broadcast_bits'  => <int>,
        'ip_gateway'         => <ip gateway>,
        'hosts'              => <int>,
        'host_index'         => <int>,
        'host_type'          => <net|host|bcast|mcast>,
        'role'               => <free|locked|data|clus|mgmt|inter-clus>,
    );


=head3 ip_manager_store_value ()

Method to store values in object for later insert, to be use for bulk inserts

    $DB->ip_manager_store_value( args )

Same arguments as insert_ip_manager()



=head3 ip_manager_clear_value();

Method to clear values stored in object for bulk update.

    $DB->ip_manager_clear_value;


=head3 ip_manager_value_count()

Method to return the number of rows to be inserted in to DB

    $int = $DB->ip_manager_value_count;


=head3 insert_ip_manager_values()

Method to insert all info stored in object, as entered with ip_manager_store_value().

   $ok = $DB->insert_ip_manager_values;
 

=head2 Request and Return IP Addresses

=head3 ip_address_request()

Method to request a list of IP addresses fromt eh DB based on network.

    ($ok, $error) = $DB->ip_address_request(
        'hosts'        => <int>,       # how many host    [1]
        'host_type'    => <str>,       # enum('net','bcast','host') [host]
        'query'        => <any|order>, # query for a list or continuous block [any]
        'network'      => <str>,       # network id
        'version'      => <'4|6'>,        # 4|6|v4|v6|inet4|inet6 [4]
        'role'         => <str>,       # enum('free','locked','data','clus','mgmt','inter-clus') [data]
        'job_id'       => <str>,       # job id
        # optional 
        'vlan'         => <int>,       # vlan id, use instead of network
        'run_id'       => <str>,       # nate RUN ID , if not given username or user id used
        'ndisp_server' => <str>,       # ndisp server, not given then hostname or local ip used
        ndisp_server_port => <int>,    # ndisp server port, not given then PID used 

    );

=head3 ip_manager_return();

Method to return an ip address to the pool based on runid or single address or
ref to list of addresses

    ($ok, $error) = $DB->ip_manager_return( 'job_id' => <str> );
    ($ok, $error) = $DB->ip_manager_return( 'address' => <str> );
    ($ok, $error) = $DB->ip_manager_return( 'address_aref' => <ref> );

Only one argument is worked on, precedence is job_id, address, address_aref.

Returns upon success in array context the number of addresses returned and
error string, in scalar context just the number of addresses returned. 
Upon failure in array context the number of addresses returned and
error string, in scalar context undef.



=head2 Accessor Methods

Methods to access network information from the object

=head3 ip_address_range

Returns the ip address range as a string start->end

    $str = $DB->ip_address_range;


=head3 ip_address_list

Returns a reference to a list of ip addresses (ip1, ip2, ip3, ..)

    $aref = $DB->ip_address_list;

=head3 ip_address_request_aref

Returns a ref to the AoA containing the IP info ([], [], [], ...)

    $aref = $DB->ip_address_request_aref;

=head3 ip_address_version

Returns the IP version as a string.

    $str = $DB->ip_address_version;

=head3 ip_address_address_bits

Returns the integer value of bit in the ip address 

    $int = $DB->ip_address_address_bits;

=head3 ip_address_network

Returns the IP network id as a string.

    $str = $DB->ip_address_network

=head3 ip_address_network_bits 

Returns the integer value of bit in the ip network id.

    $int = $DB->ip_address_network_bits

=head3 ip_address_broadcast 

Returns the IP address broadcast as a string.

    $str = $DB->ip_address_broadcast;

=head3 ip_address_broadcast_bits

Returns the integer value of bit in the ip broadcast address.

    $int = $DB->ip_address_network_bits;

=head3 ip_address_gateway 

Return the gateway for the  IP address as a string

   $str = $DB->ip_address_gateway;

=head1 Functions

=head2 normalize_role

Note exported but accessible, use to provide the user with shortcut string
for the role setting. function will translated given shortcuts to the
proper strings in the DB.

   $role = &Network::SCON_DB::IPAddrManager::normalize_role(<role>);

=over

=item clus => cluster

=item cluster[-_]mgmt => cluster_management

=item mgmt => management

=item node[-_]mgmt => management

=inter-clus => intercluster

=back





=head1 STATUS

   Current

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date:$

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=cut







