# Perl Module
# $Id$
# Last Revised $Date$
# vim: ft=perl
#
# Module providing wrapper method/functions for IP reservations
#
## @summary IPManger access for NATE
##
## @description Reads in IP info or queries DB for IP info
##
## @author      parsons@netapp.com
##
## @pod here
#
# POD at end of file


package Network::SCON_DB::IPReservation;

use strict;
use Options;
use Tharn; #qw($Log);
use Data::Dumper;

use Network::IPAddress;
use Network::SCON_DB::IPAddrManager;


BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter Network::IPAddress);
   our $VERSION       = '1.0.0';
}

## place holder for future module TAGS
our @TAGS_1 = ();
our @TAGS_2 = ();

# Export functions
our @EXPORT = (@TAGS_1, @TAGS_2);

our @EXPORT_OK  = qw();
our %EXPORT_TAGS = ( 'tag1'  => [@TAGS_1],
                     'tag2'  => [@TAGS_2],
);

our $ERROR = '';




# Define the options need for IP addressing
my $Options  = new Options;
$Options->parameters(definitions=>
  [
    {
        name            => "DATA_IP_RANGE",
        description     => "Data IPS",
        spec            => 's',
    },
    {
        name            => "CLUSTER_IP_RANGE",
        description     => "CLUSTER IPS",
        spec            => 's',
    },

    # vars from GDL csiqa ndisp
    {
       name             => 'FREE_DATA_IP',
       description      => 'Data IPs',
       spec             => 's',
    },
    {
       name             => 'FREE_DATA_IP_NETMASK',
       description      => 'Data IP Netmask',
       spec             => 's',
    },
    {
       name             => 'FREE_DATA_IP_GATEWAY',
       description      => 'Data IP Gateway',
       spec             => 's',
    },
    {
       name             => 'FREE_DATA_IP_CONFIGS',
       description      => 'Data IP Configs',
       spec             => 's',
    },
    {
       name             => 'FREE_CLUSTER_IP',
       description      => 'Cluster IPs',
       spec             => 's',
    },
    {
       name             => 'FREE_CLUSTER_IP_NETMASK',
       description      => 'Cluster IP Netmask',
       spec             => 's',
    },
    {
       name             => 'FREE_CLUSTER_IP_GATEWAY',
       description      => 'Cluster IP Gateway',
       spec             => 's',
    },
    {
       name             => 'FREE_CLUSTER_IP_CONFIGS',
       description      => 'Cluster IP Configs',
       spec             => 's',
    },
    {
       name             => 'FREE_MGMT_IP',
       description      => 'MGMT IPs',
       spec             => 's',
    },
    {
       name             => 'FREE_MGMT_IP_NETMASK',
       description      => 'MGMT IP Netmask',
       spec             => 's',
    },
    {
       name             => 'FREE_MGMT_IP_GATEWAY',
       description      => 'MGMT IP Gateway',
       spec             => 's',
    },
    {
       name             => 'FREE_MGMT_IP_CONFIGS',
       description      => 'MGMT IP Configs',
       spec             => 's',
    },
    {
       name             => 'IPM_PASSPHRASE',
       description      => 'IP Manager Pass Phrase',
       spec             => 's',
    },
    {
       name             => 'IPM_COUNT',
       description      => 'IP Manager Host count',
       spec             => 'i',
    },
    {
       name             => 'IPM_VERSION',
       description      => 'IP Manager IP version',
       spec             => 's',
    },
    {   
       name              => 'IPM_ROLE',
       description       => 'Role of IP address',
       spec              => 's',
    },
    {   
       name              => 'IPM_QUERY',
       description       => 'Query type',
       spec              => 's',
    },
    {
       name              => 'IPM_NETWORK_MATCH',
       description       => 'Boolean to check network ids',
       spec              => 'i',
    },
    {
       name              => 'IPM_SITE',
       description       => 'IPManager DB site',
       spec              => 's',
    },
    {
       name              => 'IPM_NETWORK_ID',
       description       => 'IPManager DB lookup network id',
       spec              => 's',
    },
    {
       name              => 'IPM_VLAN_ID',
       description       => 'IPManager DB lookup vlan id',
       spec              => 's',
    },
    {
       name              => 'IPM_JOB_ID',
       description       => 'IPManager reservation job id',
       spec              => 's',
    },
    {
       name              => 'IPM_REUSE_RESERVATION',
       description       => 'IPManager flag to look for existing job_id first',
       spec              => 'i',
    },
  ]
);

# translate old params to new
#   xxxx_IP_RANGE are the original params
#   FREE_xxxx_IP are params distributed by NDISP in GDL
$Options->compatability(
   definitions => {
      'DATA_IP_RANGE'     => 'FREE_DATA_IP',
      'CLUSTER_IP_RANGE'  => 'FREE_CLUSTER_IP' 
   }
);

#=============================================================================
#
# Constructor
# Will create the object and call supporting method based on input.
#
#   hostrec    => <ref to hostrec>
#   network_id => <str>
#   vlan_id    => <int>
#
#   role       => <cluster|data|intercluster>
#   count      => <int>              how many addresses
#   query      => <any|order>        
#   version    => <4|6|v4|v6|inet4|inet6>  IP version
#   job_id     => <job id>
#
#   run_id            => <run id>
#   ndisp_server      => <server name>
#   ndisp_server_port => <ndisp server port> 
#   reuse_reservation => 0|1 
#   passphrase        => <str>
#   network_match     => <0|1>
# 

sub ip_addresses {

   my $_proto = shift;
   my $_class = ref($_proto) || $_proto;
   my $_self  = {};
   # bless object
   bless($_self, $_class);

   # get object arguments
   my %_args = @_;

   my ($_ok, $_error);

   $Log->enter("IPReservation: ip address constructor");

   # Check Common args 
   # role (required)
   if ( !defined($_args{'role'}) ) {
      $ERROR = 'IPReservation::ip_addresses: role arg not given';
      return( wantarray ? (undef, $ERROR) : undef);
   }
   # count default=1
   if ( !defined($_args{'count'}) ) {
      $Log->comment("IPReservation::ip_addresses: count arg not given, set to 1");
      $_args{'count'} = 1;
   }
   # query default=order
   if ( !defined($_args{'query'}) ) {
      $Log->comment("IPReservation::ip_addresses: query arg not given, set to order");
      $_args{'query'} = 'any';
   }
   # version default=4
   if ( !defined($_args{'version'}) ) {
      $Log->comment("IPReservation::ip_addresses: version arg not given, set to 4");
      $_args{'version'} = '4';
   }
   # passphrase default=default
   if ( !defined($_args{'passphrase'}) ) {
      $Log->comment("IPReservation::ip_addresses: passphrase arg not given, set to default");
      $_args{'passphrase'} = 'default';
   }
   # network_match  default=1;
   if ( !defined($_args{'network_match'}) ) {
      $Log->comment("IPReservation::ip_addresses: network_match arg not given, set to 1");
      $_args{'network_match'} = 1;
   }

   # job id  (required) 
   if ( !defined($_args{'job_id'}) ) {
      $ERROR = "IPReservation::ip_addresses: job_id arg not given";
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # check/set site param
   if ( !defined($_args{'site_id'}) ) {
        $Log->comment("IPReservation::ip_addresses: site_id not given");
        $_args{'site_id'} = param('IPM_SITE') || param('SITE') || $ENV{'SITE'};
    }
    $Log->comment("IPReservation::ip_addresses: site_id set to $_args{'site_id'}");


   # Normalize role
   #
   $_args{'role'} = &Network::SCON_DB::IPAddrManager::normalize_role($_args{'role'});

   # Validate IPVersion
   ($_args{'version'}, $ERROR) = ip_manager_validate_ip_version($_args{'version'});
   unless ($_args{'version'}) {
      return( wantarray ? (undef, $ERROR) : undef);
   }

   $Log->comment("IPReservation: Determine method to get addresses with "
                ."version $_args{'version'} " 
                ."network_id " . ($_args{'network_id'} ? "$_args{'network_id'} " : '- ')
                ."vlan_id "    . ($_args{'vlan_id'}    ? "$_args{'vlan_id'} " : '- ') 
                ."hostrec "    . (defined($_args{'hostrec'}) ? "$_args{'hostrec'}->{'hostname'} " : '- ')
                ."reuse "      . ($_args{'reuse_reservation'} ? 'yes ' : '- ')
   );


   # see which method to call
   if ( $_args{'reuse_reservation'} ) {
      ($_ok, $ERROR) = $_self->ip_addresses_reuse(%_args);
      unless( $_ok ) {
         $Log->comment("IPReservation:::ip_addresses: not re-using a reservation");
         if ( $_args{'network_id'} || $_args{'vlan_id'} ) {
            ($_ok, $ERROR) = $_self->ip_addresses_by_id(%_args);
         }
         elsif ( defined($_args{'hostrec'}) ) {
            ($_ok, $ERROR) = $_self->ip_addresses_by_hostrec(%_args);
         }
         else {
            $ERROR = "IPReservation::ip_addresses: failed with reuse_reservation";
            ($_ok, $ERROR) = (0, $ERROR);
         }
      }
   }
   elsif ( $_args{'network_id'} || $_args{'vlan_id'} ) {
      ($_ok, $ERROR) = $_self->ip_addresses_by_id(%_args);
   }
   elsif ( defined($_args{'hostrec'}) ) {
      ($_ok, $ERROR) = $_self->ip_addresses_by_hostrec(%_args);
   }
   else {
      ($_ok, $ERROR) = (0, "required arg not given");
   }

   if ($_ok) { return ( wantarray ? ($_self, $ERROR ) : $_self); }
   else      { return( wantarray ? (undef, $ERROR) : undef); }

}


#-----------------------------------------------------------------------------
# Get IP address alrady reserved given the info from previous reservation
#
sub ip_addresses_reuse {

   my $_obj = shift;
   my %_arg = @_;
   my %_ip_opt_hash = ();
   my($_ok, $_error, $_m);

   $Log->comment("IPReservation::ip_addresses_reuse");

   # get access to the options hash
   my $_options = $Options->get_hash();

   # role (required)
   if ( !defined($_arg{'role'}) ) {
      $ERROR = 'IPReservation::ip_addresses_reuse: role arg not given';
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # version default=4
   if ( !defined($_arg{'version'}) ) {
      $ERROR = "IPReservation::ip_addresses_reuse: version arg not given";
      return( wantarray ? (undef, $ERROR) : undef);
   }
   # job id  (required) 
   if ( !defined($_arg{'job_id'}) ) {
      $ERROR = "IPReservation::ip_addresses_reuse: job_id arg not given";
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # Normalize role
   #
   $_arg{'role'} = &Network::SCON_DB::IPAddrManager::normalize_role($_arg{'role'});

   # Validate IPVersion
   ($_arg{'version'}, $ERROR) = ip_manager_validate_ip_version($_arg{'version'});
   unless($_arg{'version'}) {
      return( wantarray ? (undef, $ERROR) : undef);
   }

   ($_obj->{'IP'}, $_error)  = Network::SCON_DB::IPAddrManager->new(
        'site_id' => $_arg{'site_id'},
   );

   unless ($_obj->{'IP'}) {
         $ERROR = "IPReservation::ip_addresses_reuse: IPAddrManager object: $_error";
         $Log->error("DB connection failed: $_error");
         return( wantarray ? (undef, $ERROR) : undef);      
   }
   $Log->comment("IPReservation::ip_addresses_reuse: connected to "
                 . "site: " . $_obj->{'IP'}{'cfg'}{'site_id'} . " "
                 . "for "   . $_obj->{'IP'}{'cfg'}{'description'} . " "
   );  


   # populate args for ip request
   %_ip_opt_hash = (
            'version'      => $_arg{'version'},
            'role'         => $_arg{'role'},
            'job_id'       => $_arg{'job_id'},
   );

   $Log->comment("IPReservation::ip_addresses_reuse: checking to reuse reservation "
                ."job_id: $_arg{'job_id'} "
                ."version: $_arg{'version'} "
                ."role: $_arg{'role'} "
   );
   ($_ok, $_error) = $_obj->{'IP'}->ip_address_reuse(%_ip_opt_hash);
  
    unless ($_ok) {
         $ERROR = "IPReservation::ip_addresses_reuse: ip_address_request failed: $_error";
         # list the messages
         foreach $_m ( @{$_obj->{'IP'}->{'messages'}} ) { $Log->comment("msg: $_m");}
         # error out
         $Log->comment("$ERROR");
         return( wantarray ? (undef, $ERROR) : undef);
      }
 

   return( wantarray ? ($_ok, $ERROR) : $_ok);

}



#-----------------------------------------------------------------------------
# Get IP address given a hostrec or use of FREE_<role>_IP_* params
#
#   hostrec => <ref to hostrec>
#   role    => <cluster|data|intercluster>
#   count   => <int>              how many addresses
#   query   => <any|order>        
#   version => <4|6|v4|v6|inet4|inet6>  IP version
#   job_id  => <job id>
#
#   run_id  => <run id>
#   ndisp_server => <server name>
#   ndisp_server_port => <ndisp server port> 
#   passphrase        => <str>
#   network_match     => <0|1>
#   
sub ip_addresses_by_hostrec {

   my $_obj = shift;


   # get object arguments
   my %_arg = @_;
   my %_ip_opt_hash = ();
   my ($_ip, $_ok, $_error, $_error2,
       $_hostrec_net, $_hostrec_vlan_id, $_hostrec_ip, $_hostrec_mask, $_hostrec_gw, 
       $_param_ip_str, $_param_mask_str, $_param_gw_str, $_param_vlan_str, 
       $_free_ip_str, $_free_ip_gw_str, $_free_ip_mask_str,
       $_href, $_id, $_m, $_h, $_str, $_ipv,
   );

   # get access to the options hash
   my $_options = $Options->get_hash();

   $Log->comment("IPReservation::ip_addresses_by_hostrec");
   #
   # check args
   #
   # hostrec (required) 
   if ( !defined($_arg{'hostrec'})) {
      $ERROR = 'IPReservation::ip_addresses_by_hostrec: hostrec arg not given';
      return( wantarray ? (undef, $ERROR) : undef);
   } 
   # role (required)
   if ( !defined($_arg{'role'}) ) {
      $ERROR = 'IPReservation::ip_addresses_by_hostrec: role arg not given';
      return( wantarray ? (undef, $ERROR) : undef);
   }
   # count default=1
   if ( !defined($_arg{'count'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_hostrec: count arg not given, set to 1");
      $_arg{'count'} = 1;
   }
   # query default=order
   if ( !defined($_arg{'query'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_hostrec: query arg not given, set to order");
      $_arg{'query'} = 'any';
   }
   # version default=4
   if ( !defined($_arg{'version'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_hostrec: version arg not given, set to 4");
      $_arg{'version'} = '4';
   }
   # passphrase default=default
   if ( !defined($_arg{'passphrase'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_hostrec: passphrase arg not given, set to default");
      $_arg{'passphrase'} = 'default';
   }
   # network_match  default=1;
   if ( !defined($_arg{'network_match'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_hostrec: network_match arg not given, set to 1");
      $_arg{'network_match'} = 1;
   }

   # job id  (required) 
   if ( !defined($_arg{'job_id'}) ) {
      $ERROR = "IPReservation::ip_addresses_by_hostrec: job_id arg not given";
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # Normalize role
   #
   $_arg{'role'} = &Network::SCON_DB::IPAddrManager::normalize_role($_arg{'role'});

   # Validate IPVersion
   ($_arg{'version'}, $ERROR) = ip_manager_validate_ip_version($_arg{'version'});
   unless($_arg{'version'}) { 
      return( wantarray ? (undef, $ERROR) : undef);
   } 

   # Translate needed strings and values from hostrec
   #
   if ( $_arg{'role'} eq 'cluster'  ||  $_arg{'role'} eq 'clus') {
      $_param_ip_str   = 'CLUSTER_IP1';
      $_param_mask_str = 'CLUSTER_MASK1';
      $_param_gw_str   = 'CLUSTER_GATEWAY1';
      $_param_vlan_str = 'CLUSTER_VLAN';

      $_free_ip_str      = 'FREE_CLUSTER_IP';
      $_free_ip_gw_str   = 'FREE_CLUSTER_IP_GATEWAY';
      $_free_ip_mask_str = 'FREE_CLUSTER_IP_NETMASK';
   }
   elsif  ( $_arg{'role'} eq 'data' ) {
      $_param_ip_str   = 'DATA_IP1';
      $_param_mask_str = 'DATA_MASK1';
      $_param_gw_str   = 'DATA_GATEWAY1';
      $_param_vlan_str = 'DATA_VLAN';

      $_free_ip_str      = 'FREE_DATA_IP';
      $_free_ip_gw_str   = 'FREE_DATA_IP_GATEWAY';
      $_free_ip_mask_str = 'FREE_DATA_IP_NETMASK';
   }
   elsif  ( $_arg{'role'} eq 'intercluster' || $_arg{'role'} eq 'cluster_management' ) {
      $_param_ip_str   = 'CLUSTER_MGMT_IP';
      $_param_mask_str = 'CLUSTER_MGMT_MASK';
      $_param_gw_str   = 'CLUSTER_MGMT_GATEWAY';
      $_param_vlan_str = 'CLUSTER_MGMT_VLAN';

      $_free_ip_str      = 'FREE_CLUSTER_MGMT_IP';
      $_free_ip_mask_str = 'FREE_CLUSTER_MGMT_IP_NETMASK';
      $_free_ip_gw_str   = 'FREE_CLUSTER_MGMT_IP_GATEWAY';
   }
   elsif  ( $_arg{'role'} eq 'management'  ) {
      $_param_ip_str   = 'MGMT_IP';
      $_param_mask_str = 'MGMT_MASK';
      $_param_gw_str   = 'MGMT_GATEWAY';
      $_param_vlan_str = 'MGMT_VLAN';

      $_free_ip_str      = 'FREE_MGMT_IP';
      $_free_ip_mask_str = 'FREE_MGMT_IP_NETMASK';
      $_free_ip_gw_str   = 'FREE_MGMT_IP_GATEWAY';
   }


   else {
      $ERROR = "IPReservation::ip_addresses_by_hostrec: "
             . "role [$_arg{'role'}] not supported, not mapped to tharn host entries";
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # Look in hostrec net records for desired based on 'netid' 
   $_ok = ip_manager_validate_ip_version($_arg{'version'});
   $_str = sprintf("inet%s_%s1", $_ok, $_arg{'role'});
   $Log->comment("IPReservation: look in hostrec net records for netid: $_str");
   foreach $_h ( @{$_arg{'hostrec'}->{'net'}} ) {
      if ( $_h->{'netid'} eq $_str ) { 
         $_href = $_h;
         $Log->comment("IPReservation: [$_str] hostrec net record [$_str] found: " . Dumper $_href); 
         last;
      }
   }  
   # if a net record did not exist
   if ( ref($_href) ne 'HASH' ) {
      $ERROR = "IPReservation::ip_addresses_by_hostrec: net record not found with netid: $_str";
      #$Log->comment("no net record $_str " . Dumper $_arg{'hostrec'});
      $Log->error($ERROR);
      foreach $_h ( @{$_arg{'hostrec'}->{'net'}} ) {
         $Log->comment("found netid: $_h->{'netid'}");
      }
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # Set hostrec info from record, then from scalar
   $_hostrec_ip      = $_href->{'address'} || $_arg{'hostrec'}->{$_param_ip_str};
   $_hostrec_mask    = $_href->{'netmask'} || $_arg{'hostrec'}->{$_param_mask_str};
   $_hostrec_gw      = $_href->{'gateway'} || $_arg{'hostrec'}->{$_param_gw_str};
   $_hostrec_vlan_id = $_arg{'hostrec'}->{$_param_vlan_str} || '';

   # make sure hostrec or given ipv6 address is in expanded syntax
   if ( $_ok == 6 ) {
      $Log->comment("IPReservation: ensure ipv6 address is expanded");
      $_hostrec_ip   = &Net::IP::ip_expand_address($_hostrec_ip, 6);
      $_hostrec_mask = &Net::IP::ip_expand_address($_hostrec_mask, 6);
      $_hostrec_gw   = &Net::IP::ip_expand_address($_hostrec_gw, 6);
   }


   # Info
   $Log->comment("IPReservation: hostrec role=$_arg{'role'} "
                ."IP: [$_hostrec_ip] "
                ."Mask: [$_hostrec_mask] "
                ."GW: [$_hostrec_gw] "
                ."VLAN_ID: [" . ($_hostrec_vlan_id ? $_hostrec_vlan_id : '-') . "] "
   );
 
   # Network ID derived from hostrec info
   ($_ip, $_error) = Network::IPAddress->new(
      'address' => $_hostrec_ip,
      'netmask' => $_hostrec_mask,
      'gateway' => $_hostrec_gw,
      'add_host_list' => 0
   );
   unless ($_ip) {
      $ERROR = "IPReservation::ip_addresses_by_hostrec: Network::IPAddress "
             . "object from hostrec values error: $_error";
      return( wantarray ? (undef, $ERROR) : undef);
   }
   $_hostrec_net = $_ip->get_ip_address_hash($_hostrec_ip);
   $Log->comment("IPReservation: hostrec role=$_arg{'role'} network_id: "
                . sprintf("[%s] ",  $_hostrec_net->{'network_id'} ? $_hostrec_net->{'network_id'} : '-')
   );

   # VLAN ID
   $Log->comment("IPReservation: hostrec role=$_arg{'role'} vlan_id: "
                . sprintf("[%s] ", $_hostrec_vlan_id ? $_hostrec_vlan_id  : '-')
   ); 
   #
   # Expand given FREE_<role>_IP else get from DB
   # &Network::IPAddress::netip_ip_version($_addr)
   #
   # first check the ip version compatibility if FREE_*_IP is in use
   if ( defined($_options->{$_free_ip_str}) ) {
       $Log->comment("IPReservation: param given $_free_ip_str, check version compatibility"); 
       # the given $_free_ip_str has multiple syntax
       # but a v4 address will always contain '.' 
       if ( $_options->{$_free_ip_str} =~ /\./ ) { $_ipv = 4; }
       else                                      { $_ipv = 6; }
       $Log->comment("IPReservation: param given $_free_ip_str determined to be "
                     ."ver: $_ipv arg [$_arg{'version'}]");

       if ( $_arg{'version'} eq $_ipv ) {
          $Log->comment("IPReservation: param $_free_ip_str will be used, ip versions match");
       }
       else {
          $Log->comment("IPReservation: param $_free_ip_str will not be used, ip versions mismatch");
          $Log->comment("IPReservation: DB will be queried for ver: $_arg{'version'} addresses");
       }
   }

   if ( defined($_options->{$_free_ip_str})  && $_arg{'version'} eq $_ipv) {
      # Given FREE_[DATA|CLUSTER|INTERCLUSTER]_IP
      $Log->comment("IPReservation: $_free_ip_str given as param: $_options->{$_free_ip_str}");

      # ensure we have a mask and gateway
      if ( !defined($_options->{$_free_ip_mask_str}) ) {
         $Options->set("$_free_ip_mask_str" => $_hostrec_mask);
         $Log->comment("IPReservation: $_free_ip_mask_str mask set to $_options->{$_free_ip_mask_str} from hostrec");
      }
      if ( !defined($_options->{$_free_ip_gw_str}) ) {
         $Options->set("$_free_ip_gw_str" => $_hostrec_gw);
         $Log->comment("IPReservation: $_free_ip_gw_str gateway set to $_options->{$_free_ip_gw_str} from hostrec");
      }

      # expand the given address since can be given in multiple syntax's
      ($_obj->{'IP'}, $_error) = Network::IPAddress->new(
         'address' => $_options->{$_free_ip_str},
         'netmask' => $_options->{$_free_ip_mask_str},
         'gateway' => $_options->{$_free_ip_gw_str},
         'add_host_list' => 0,
      );
      unless ($_obj->{'IP'}) {
            $ERROR = "IPReservation::ip_addresses_by_hostrec: $_arg{'role'}  Network::IPAddress:  $_error";
            return( wantarray ? (undef, $ERROR) : undef);
      }
      $Log->comment("IPReservation: $_free_ip_str address count: $_obj->{'IP'}->{'addresses'}");
      $_ok = $_obj->{'IP'}->{'addresses'};
      $Log->comment("IPReservation: make address and network list");
      # IP Address list
      # stored in object $_obj->{'ip_address_list'} = (ip, ip, ip, ..)
      $_obj->{'IP'}->ip_address_list;

      # have object filter all the network id
      $_obj->{'IP'}->network_id_list;

      # check gateway and network id
      if ( $_arg{'network_match'} ) {
         $Log->comment("IPReservation: check gateway and network id");
         foreach $_id ( @{$_obj->{'IP'}->{'network_id_list'}} ) {
            # network id not the same
            if ( $_id ne $_hostrec_net->{'network_id'} ) {
               $ERROR = sprintf("IPReservation::ip_addresses_by_hostrec: net id [%s] not equal hostrec net id [%s]", 
                        $_id, $_hostrec_net->{'network_id'}
               );
               return( wantarray ? (undef, $ERROR) : undef);
            }
            # only valid gateway get put into $_obj->{'IP'}->{'Network'}{$_id}{'gateway'}
            # so an invalid gateway will never match this
            if ( $_options->{$_free_ip_gw_str} ne $_obj->{'IP'}->{'Network'}{$_id}{'gateway'} ) {
               $ERROR = sprintf("IPReservation::ip_addresses_by_hostrec: given %s  gateway [%s] not in network [%s]",
                  $_arg{'role'},
                  $_options->{$_free_ip_gw_str},
                  $_id
               );
               return( wantarray ? (undef, $ERROR) : undef);
            }
         }
      }
      else {
         $Log->comment("IPReservation: not checking gateway and network id");
      }
      #$Log->debug("IPReservation: given $_free_ip_str: " . Dumper $_obj->{'IP'});  ###SLP
   } else {
      # Query DB
      $Log->comment("IPReservation: role=$_arg{'role'} addresses needed from DB ");
      
      # IPs need to be requested
      ($_obj->{'IP'}, $_error)  = Network::SCON_DB::IPAddrManager->new(
            'site_id' => $_arg{'site_id'},
      );

      unless ($_obj->{'IP'}) {
         $ERROR = "IPReservation::ip_addresses_by_hostrec: Network::SCON_DB::IPAddrManager: $_error";
         $Log->error("DB connection failed: $_error");
         return( wantarray ? (undef, $ERROR) : undef);
      }
      $Log->comment("IPReservation::ip_addresses_by_hostrec: connected to "
                 . "site: " . $_obj->{'IP'}{'cfg'}{'site_id'} . " "
                 . "for "   . $_obj->{'IP'}{'cfg'}{'description'} . " "
      );  


      # make hash of args for ip_address_request
      %_ip_opt_hash = (
            'hosts'        => $_arg{'count'},
            'host_type'    => 'host',
            'query'        => $_arg{'query'},
            'version'      => $_arg{'version'},
            'role'         => $_arg{'role'},
            'job_id'       => $_arg{'job_id'},
            'passphrase'   => $_arg{'passphrase'},
      );
      $_ip_opt_hash{'run_id'} = $_arg{'run_id'} if defined( $_arg{'run_id'});
      $_ip_opt_hash{'ndisp_server'} = $_arg{'ndisp_server'} if defined( $_arg{'ndisp_server'});
      $_ip_opt_hash{'ndisp_server_port'} = $_arg{'ndisp_server_port'} if defined( $_arg{'ndisp_server_port'});

      # add to hash either vlan or network id to use in the sql query
      # use vlan_id if given 
      if ( $_hostrec_vlan_id ) {
         $_ip_opt_hash{'vlan_id'} = $_hostrec_vlan_id;
         $Log->comment("IPReservation: using hostrec "
                      . "role: [$_arg{'role'}] "
                      . "vlan_id: [$_ip_opt_hash{'vlan_id'}]"
         );
         # check DB for supported vlan ids
         ($_ok, $_error)  = $_obj->{'IP'}->ip_manager_vlan_ids;
         if ( $_ok) {
           if ( $_obj->{'IP'}->{'ip_manager'}{'vlan_id'}{$_ip_opt_hash{'vlan_id'}}{'Count'} ) {
              $Log->comment("IPReservation: vlan_id supported [$_ip_opt_hash{'vlan_id'}] with "
                 . $_obj->{'IP'}->{'ip_manager'}{'vlan_id'}{$_ip_opt_hash{'vlan_id'}}{'Count'}
                 . " rows"
              );
           }
           else {
              $ERROR = "IPReservation::ip_addresses_by_hostrec: "
                      ."vlan_id [$_ip_opt_hash{'vlan_id'}] not found in DB";
              $_obj->{'IP'}->scon_db_disconnect;
              return( wantarray ? (undef, $ERROR) : undef);
           }
         } 
      }
      # network id
      elsif ( $_hostrec_net->{'network_id'} )  {
         $_ip_opt_hash{'network_id'} = $_hostrec_net->{'network_id'};
         $Log->comment("IPReservation: use hostrec "
                       ."role: [$_arg{'role'}] "
                       ."derived network: [$_ip_opt_hash{'network_id'}] "
         );
         # check DB for supported network ids
         ($_ok, $_error)  = $_obj->{'IP'}->ip_manager_network_ids;
         if ( $_ok) {
           if ( $_obj->{'IP'}->{'ip_manager'}{'ip_network'}{$_ip_opt_hash{'network_id'}}{'Count'} ) {
              $Log->comment("IPReservation: ip_network [$_ip_opt_hash{'network_id'}] supported with "
                 . $_obj->{'IP'}->{'ip_manager'}{'ip_network'}{$_ip_opt_hash{'network_id'}}{'Count'}
                 . " rows"
              );
           }
           else {
              $ERROR = "IPReservation::ip_addresses_by_hostrec: "
                     . "ip_network [$_ip_opt_hash{'network_id'}] not found in DB";
              $_obj->{'IP'}->scon_db_disconnect;
              return( wantarray ? (undef, $ERROR) : undef);
           }
         } 
      }
      else {
         $ERROR = "IPReservation::ip_addresses_by_hostrec: hostrec "
                . "vlan_id or  network_id not found";
         $Log->comment("ERROR: $_error: missing args for ip_address_request call");
         $_obj->{'IP'}->scon_db_disconnect;
         return( wantarray ? (undef, $ERROR) : undef);
      }
      # get the IP address from DB
      # stored in object $_obj->{'ip_address_list'} = (ip, ip, ip, ..)
      $Log->comment("IPReservation: DB request for [$_free_ip_str] "
                   ."role: [$_arg{'role'}] "
                   ."hosts: [$_arg{'count'}] "
                   ."query: [$_arg{'query'}] "
                   ."ip ver: [$_arg{'version'}] "
                   ."job_id: [$_arg{'job_id'}] "
                   ."network_id [" . ($_ip_opt_hash{'network_id'} ? $_ip_opt_hash{'network_id'} : '-') ."] "
                   ."vlan_id [" . ($_ip_opt_hash{'vlan_id'} ? $_ip_opt_hash{'vlan_id'} : '-') ."] "
                   ."passphrase [$_arg{'passphrase'}] "

      );
      ($_ok, $_error) = $_obj->{'IP'}->ip_address_request( %_ip_opt_hash );

      # ip address query returned no address
      unless ($_ok) {
         $ERROR = "IPReservation::ip_addresses_by_hostrec: ip_address_request failed: $_error";
         # list the messages
         foreach $_m ( @{$_obj->{'IP'}->{'messages'}} ) { $Log->comment("msg: $_m");}
         # error out
         $Log->error("$ERROR");
         return( wantarray ? (undef, $ERROR) : undef);
      }

      $Log->comment("IPReservation: $_free_ip_str request returned $_ok addresses");

      # store the address
      #$Log->debug("IPReservation: requested $_arg{'role'}: " . Dumper $_obj->{'IP'});  ###SLP

      $_obj->{'IP'}->scon_db_disconnect;
   }

   return ( wantarray ? ($_ok, $ERROR ) : $_ok);
}
#-----------------------------------------------------------------------------
#
# Get IP addresses given network or vlan id
#
#   vlan_id    => <int>
#   network_id => <str>
#   role       => <cluster|data|intercluster>
#   count      => <int>              how many addresses
#   query      => <any|order>        
#   version    => <4|6|v4|v6|inet4|inet6>  IP version
#   job_id     => <job id>
#
#   run_id  => <run id>
#   ndisp_server => <server name>
#   ndisp_server_port => <ndisp server port> 
#
sub ip_addresses_by_id {

   my $_obj = shift;

   # get object arguments
   my %_arg = @_;
   my %_ip_opt_hash = ();

   my( $_ok, $_error,
       $_m,
   );

   $Log->comment("IPReservation::ip_addresses_by_id ");


   # role (required)
   if ( !defined($_arg{'role'}) ) {
      $ERROR = 'IPReservation::ip_addresses_by_id: role arg not given';
      return( wantarray ? (undef, $ERROR) : undef);
   }
   # vlan_id or network_id required
   if ( !defined($_arg{'vlan_id'}) || !defined($_arg{'network_id'}) ) {
      $ERROR = "IPReservation::ip_addresses_by_id: "
             . "missing required arg: vlan_id: [$_arg{'vlan_id'}] "
             . "network_id: [$_arg{'network_id'}] ";
   }

   # count default=1
   if ( !defined($_arg{'count'}) ) {
      $Log->comment("IPReservation: :ip_addresses_by_id: count arg not given, set to 1");
      $_arg{'count'} = 1;
   }
   # query default=order
   if ( !defined($_arg{'query'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_id: query arg not given, set to order");
      $_arg{'query'} = 'any';
   }
   # version default=4
   if ( !defined($_arg{'version'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_id: version arg not given, set to 4");
      $_arg{'version'} = '4';
   }
   # passphrase default=default
   if ( !defined($_arg{'passphrase'}) ) {
      $Log->comment("IPReservation::ip_addresses_by_id: passphrase arg not given, set to default");
      $_arg{'passphrase'} = 'default';
   }
   # job id  (required) 
   if ( !defined($_arg{'job_id'}) ) {
      $ERROR = "IPReservation::ip_addresses_by_id: job_id arg not given";
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # Normalize role
   $_arg{'role'} = &Network::SCON_DB::IPAddrManager::normalize_role($_arg{'role'});

   # Validate IPVersion
   ($_arg{'version'}, $ERROR) = ip_manager_validate_ip_version($_arg{'version'});
   unless($_arg{'version'}) {
      return( wantarray ? (undef, $ERROR) : undef);
   }

   # IPAddrManager object
   ($_obj->{'IP'}, $_error)  = Network::SCON_DB::IPAddrManager->new(
      'site_id' => $_arg{'site_id'}
   );
   unless ($_obj->{'IP'}) {
         $ERROR = "IPReservation::ip_addresses_by_id: IPAddrManager object: $_error";
         $Log->error("DB connection failed: $_error");
         return( wantarray ? (undef, $ERROR) : undef);
   }
   $Log->comment("IPReservation::ip_addresses_by_id: connected to "
                 . "site: " . $_obj->{'IP'}{'cfg'}{'site_id'} . " "
                 . "for "   . $_obj->{'IP'}{'cfg'}{'description'} . " "
   );
 

   # populate args for ip request
   %_ip_opt_hash = (
            'hosts'        => $_arg{'count'},
            'host_type'    => 'host',
            'query'        => $_arg{'query'},
            'version'      => $_arg{'version'},
            'role'         => $_arg{'role'},
            'job_id'       => $_arg{'job_id'},
            'passphrase'   => $_arg{'passphrase'},
   );


   # check vlan_id
   if ( $_arg{'vlan_id'} )    { 
      $_ip_opt_hash{'vlan_id'} = $_arg{'vlan_id'};
      # check DB for supported vlan ids
      ($_ok, $_error)  = $_obj->{'IP'}->ip_manager_vlan_ids;
      if ( $_ok) {
         if ( $_obj->{'IP'}->{'ip_manager'}{'vlan_id'}{$_ip_opt_hash{'vlan_id'}}{'Count'} ) {
            $Log->comment("IPReservation: vlan_id supported [$_ip_opt_hash{'vlan_id'}] with "
                 . $_obj->{'IP'}->{'ip_manager'}{'vlan_id'}{$_ip_opt_hash{'vlan_id'}}{'Count'}
                 . " rows"
            );
         }
         else {
            $ERROR = "IPReservation::ip_addresses_by_id: "
                   . "vlan_id [$_ip_opt_hash{'vlan_id'}] not found in DB";
            $_obj->{'IP'}->scon_db_disconnect;
            return( wantarray ? (undef, $ERROR) : undef);
         }
      }
   }
   # check network_id (translate to network)
   if ( $_arg{'network_id'} ) { 
      $_ip_opt_hash{'network_id'} = $_arg{'network_id'};
      ($_ok, $_error)  = $_obj->{'IP'}->ip_manager_network_ids;
      if ( $_ok) {
         if ( $_obj->{'IP'}->{'ip_manager'}{'ip_network'}{$_ip_opt_hash{'network_id'}}{'Count'} ) {
            $Log->comment("IPReservation: ip_network [$_ip_opt_hash{'network_id'}] supported with "
               . $_obj->{'IP'}->{'ip_manager'}{'ip_network'}{$_ip_opt_hash{'network_id'}}{'Count'}
               . " rows"
            );
         }
         else {
            $ERROR = "IPReservation::ip_addresses_by_id: "
                   . "ip_network [$_ip_opt_hash{'network_id'}] not found in DB";
            $_obj->{'IP'}->scon_db_disconnect;
            return( wantarray ? (undef, $ERROR) : undef);
         }
      }
   }

   
   $_ip_opt_hash{'run_id'} = $_arg{'run_id'} if defined( $_arg{'run_id'});
   $_ip_opt_hash{'ndisp_server'} = $_arg{'ndisp_server'} if defined( $_arg{'ndisp_server'});
   $_ip_opt_hash{'ndisp_server_port'} = $_arg{'ndisp_server_port'} if defined( $_arg{'ndisp_server_port'});

   # get the IP address from DB
   # stored in object $_obj->{'ip_address_list'} = (ip, ip, ip, ..)
      $Log->comment("IPReservation: DB request for "
                   ."role: [$_arg{'role'}] "
                   ."hosts: [$_arg{'count'}] "
                   ."query: [$_arg{'query'}] "
                   ."ip ver: [$_arg{'version'}] "
                   ."job_id: [$_arg{'job_id'}] "
                   ."network_id [" . ($_ip_opt_hash{'network_id'} ? $_ip_opt_hash{'network_id'} : '-') ."] "
                   ."vlan_id [" . ($_ip_opt_hash{'vlan_id'} ? $_ip_opt_hash{'vlan_id'} : '-') ."] "
                   ."passphrase [$_arg{'passphrase'}] "
      );

   ($_ok, $_error) = $_obj->{'IP'}->ip_address_request( %_ip_opt_hash );

   unless ($_ok) {
      $ERROR = "IPReservation::ip_addresses_by_id: ip_address_request failed: $_error";
      # list the messages
      foreach $_m ( @{$_obj->{'IP'}->{'messages'}} ) { $Log->comment("msg: $_m");}
      # error out
      $Log->error("$ERROR");
      return( wantarray ? (undef, $ERROR) : undef);

      $Log->comment("IPReservation: request returned $_ok addresses");

      # store the address
      #$Log->debug("IPReservation: requested $_arg{'role'}: " . Dumper $_obj->{'IP'});  ###SLP
      $_obj->{'IP'}->scon_db_disconnect;
   }

   return ( wantarray ? ($_ok, $ERROR ) : $_ok);
}



#-----------------------------------------------------------------------------
#
# Return address to DB
#
# Args
#  'job_id'       => <job_id>
#  'address'      => <str>   single address
#  'address_aref' => <ref>   ref to list of addresses
#  Return 
#   wantarray ? (status, ERROR) : status
#
sub ip_address_return {

   my $_obj = shift;

   my $_site = param('IPM_SITE') || param('SITE') || $ENV{'SITE'};

   my %_arg = @_;
   my %_opt = ();
   my ($_ok, $_k);
   $ERROR = '';

   $Log->comment("IPReservation::ip_address_return: site [$_site]");

   # DB object
   my ($_db, $_error)  = Network::SCON_DB::IPAddrManager->new(
      'site_id' => $_site,
   );

   # DB object failed
   unless($_db) {
      $ERROR = "IPReservation::ip_address_return:  DB connection setup: $_error";
      return( wantarray ? (undef, $ERROR) : undef);
   }
   $Log->comment("IPReservation::ip_addresses_return: connected to "
                 . "site: " . $_db->{'cfg'}{'site_id'} . " "
                 . "for "   . $_db->{'cfg'}{'description'} . " "
   );

   # job_id
   if ( $_arg{'job_id'} ) {
      $Log->comment("IPReservation::ip_address_return: return addresses under job_id [$_arg{'job_id'}]");
      %_opt = ( 'job_id' => $_arg{'job_id'} ); 
      $_k = 'job_id';
   }
   # single address 
   elsif  ( $_arg{'address'}  ) {
      $Log->comment("IPReservation::ip_address_return: return single address [$_arg{'address'}]");
      %_opt = ( 'address' => $_arg{'address'} );
      $_k = 'address';
   }
   # multiple address 
   elsif  ( ref($_arg{'address_aref'}) eq 'ARRAY'  ) {
      $Log->comment("IPReservation: return address list");
      %_opt = ( 'address_aref' => $_arg{'address_aref'} );
      $_k = 'address_aref';
   }
   else  {
      $ERROR = "IPReservation::ip_address_return: ip_address_return: unsupported return argument";
      return( wantarray ? (undef, $ERROR ) : undef);
   }


   # return address and disconnect from DB 
   ($_ok, $_error) = $_db->ip_manager_return( %_opt );
   $_db->scon_db_disconnect;

   # return
   if ( $_ok ) {
      $Log->comment("IPReservation::ip_address_return: $_ok ip address returned with $_k arg");
       return ( wantarray ? ($_ok, "IPReservation::ip_address_return: $ERROR") : $_ok);
   }
   else {
      $ERROR = "IPReservation::ip_address_return: $_error";
      return( wantarray ? (undef, $ERROR) : undef);
   } 
}

#-----------------------------------------------------------------------------
#
# method return list of ip addresses
# 
sub ip_address_list {
   my $_obj = shift;
   return ( @{$_obj->{'IP'}{'ip_address_list'}} );
}
#-----------------------------------------------------------------------------
#
# method return list of ip addresses aref
#  
sub ip_address_list_aref {
   my $_obj = shift;
   return($_obj->{'IP'}{'ip_address_list'});
}

#-----------------------------------------------------------------------------
sub ip_address_list_reserved {
   my $_obj = shift;
   return ( @{$_obj->{'IP'}{'ip_reserved'}} );
}

#-----------------------------------------------------------------------------

sub ip_address_list_reserved_aref {
   my $_obj = shift;
   return ( $_obj->{'IP'}{'ip_reserved'} );
}




#-----------------------------------------------------------------------------
#
# method to compress object address list 

sub compress_ip_address_list {
   my $_obj = shift;
   my $_addr;
   my $_c = 0;
   my $_i = 0;
   my $_n;

   $Log->comment("IPReservation: compress addresses:");
   # compress ip address list
   foreach $_addr ( @{$_obj->{'IP'}{'ip_address_list'}} ) {
      #$Log->comment("IPReservation: compress IP ? $_addr");
      if ( &Network::IPAddress::netip_ip_version($_addr) == 6 ) {
         $_n = &Network::IPAddress::netip_compress_address($_addr);
         #$Log->comment("IPReservation: compressed it: $_n");
         $_obj->{'IP'}{'ip_address_list'}[$_i] = $_n;
         $_c++;
      }
      $_i++;
   }

   # if this is a DB query then compress addresses in $_obj->{'IP'}{'attrib'}
   if ( defined( $_obj->{'IP'}{'attrib'}) ) {
      # if ip_version is 6
      if ( $_obj->{'IP'}{'attrib'}{'ip_version'} == 6 ) {
         # change ip_broadcast
         $_n = &Network::IPAddress::netip_compress_address($_obj->{'IP'}{'attrib'}{'ip_broadcast'});
         $_obj->{'IP'}{'attrib'}{'ip_broadcast'} = $_n;
         # change ip_gateway
         $_n = &Network::IPAddress::netip_compress_address($_obj->{'IP'}{'attrib'}{'ip_gateway'});
         $_obj->{'IP'}{'attrib'}{'ip_gateway'} = $_n;
         # change ip_netmask
         $_n = &Network::IPAddress::netip_compress_address($_obj->{'IP'}{'attrib'}{'ip_netmask'});
         $_obj->{'IP'}{'attrib'}{'ip_netmask'} = $_n;
         # change ip_network
         $_n = &Network::IPAddress::netip_compress_address($_obj->{'IP'}{'attrib'}{'ip_network'});
         $_obj->{'IP'}{'attrib'}{'ip_network'} = $_n;
      }
   }

   $_c;
}
#-----------------------------------------------------------------------------
#
# method to compress object address list 

sub expand_ip_address_list {
   my $_obj = shift;
   my $_addr;
   my $_c = 0;
   my $_i = 0;
   my $_n;

   $Log->comment("IPReservation: expand addresses:");
   # compress ip address list
   foreach $_addr ( @{$_obj->{'IP'}{'ip_address_list'}} ) {
      #$Log->comment("IPReservation: compress IP ? $_addr");
      if ( &Network::IPAddress::netip_ip_version($_addr) == 6 ) {
         $_n = &Network::IPAddress::netip_expand_address($_addr);
         #$Log->comment("IPReservation: compressed it: $_n");
         $_obj->{'IP'}{'ip_address_list'}[$_i] = $_n;
         $_c++;
      }
      $_i++;
   }

   # if this is a DB query then compress addresses in $_obj->{'IP'}{'attrib'}
   if ( defined( $_obj->{'IP'}{'attrib'}) ) {
      # if ip_version is 6
      if ( $_obj->{'IP'}{'attrib'}{'ip_version'} == 6 ) {
         # change ip_broadcast
         $_n = &Network::IPAddress::netip_expand_address($_obj->{'IP'}{'attrib'}{'ip_broadcast'});
         $_obj->{'IP'}{'attrib'}{'ip_broadcast'} = $_n;
         # change ip_gateway
         $_n = &Network::IPAddress::netip_expand_address($_obj->{'IP'}{'attrib'}{'ip_gateway'});
         $_obj->{'IP'}{'attrib'}{'ip_gateway'} = $_n;
         # change ip_netmask
         $_n = &Network::IPAddress::netip_expand_address($_obj->{'IP'}{'attrib'}{'ip_netmask'});
         $_obj->{'IP'}{'attrib'}{'ip_netmask'} = $_n;
         # change ip_network
         $_n = &Network::IPAddress::netip_expand_address($_obj->{'IP'}{'attrib'}{'ip_network'});
         $_obj->{'IP'}{'attrib'}{'ip_network'} = $_n;
      }
   }

   $_c;
}

#-------------------------------------------------------------------------------
#     Accessors
#  
# If address is given the info is in Network::IPAddress object
# If address is queried from DB the info is in Network::SCON_DB::IPAddrManager object
#
# Gateway
sub ip_address_gateway {
   my $_obj = shift;
   if    ( defined($_obj->{'IP'}{'gateway'}) )              { return($_obj->{'IP'}{'gateway'}); }
   elsif ( defined($_obj->{'IP'}{'attrib'}{'ip_gateway'}))  { return($_obj->{'IP'}{'attrib'}{'ip_gateway'}); }
   else                                                     { return (undef); }

}
# Network
sub ip_address_network {
   my $_obj = shift;
   if    ( defined($_obj->{'IP'}->{'network_id_list'}) ) { return( $_obj->{'IP'}->{'network_id_list'}[0]); }
   elsif ( $_obj->{'IP'}->{'attrib'}{'ip_network'})      { return( $_obj->{'IP'}->{'attrib'}{'ip_network'}); }
   else                                                  { return( undef); }
}
# Netmask
sub ip_address_netmask {
   my $_obj = shift;
   if    ( $_obj->{'IP'}->{'netmask'} )              { return( $_obj->{'IP'}->{'netmask'}); }
   elsif ( $_obj->{'IP'}->{'attrib'}{'ip_netmask'})  { return( $_obj->{'IP'}->{'attrib'}{'ip_netmask'}); }
   else                                              { return( undef); }
}

# Netmask length, integer value for mask
sub ip_address_netmask_length {
   my $_obj = shift;
   if    ( $_obj->{'IP'}->{'netmask'} ) 
      { return( $_obj->{'IP'}->{'Network'}{$_obj->ip_address_network}{'netmask_len'}); }
   elsif ( $_obj->{'IP'}->{'attrib'}{'ip_netmask_bits'})  
      { return( $_obj->{'IP'}->{'attrib'}{'ip_netmask_bits'}); }
   else 
      { return( undef); }
}


sub ip_address_source {
    my $_obj = shift;
    if    ( defined($_obj->{'IP'}{'gateway'}) )              { return('param');}
    elsif ( defined($_obj->{'IP'}{'attrib'}{'ip_gateway'}))  { return('db');}
    else                                                     { return (undef); }
}



# end of  module
1;
__END__




=head1 NAME Network::SCON_DB::IPReservation

=head1 DESCRIPTION

Perl module providing methods to get IP addresses for testing.


=head1 SYNOPSIS

    use Network/SCON_DB/IPReservation;

    # Using a hostrec
    ($DataIP, $error) = Network::SCON_DB::IPReservation->ip_addresses(
        'hostrec' => $hostrec_object,
        'role'    => 'data',
        'count'   => 254,
        'job_id'  => $ENV{'JOBID'}
    );
    unless( $DataIP ) {
        $Log->comment("=== ERROR Data IP: $error");
    }

    # Using a network_id string
    ($DataIP, $error) = Network::SCON_DB::IPReservation->ip_addresses(
        'network_id' => '1.0.16.0',
        'role'    => 'data',
        'count'   => 254,
        'job_id'  => $ENV{'JOBID'}
    );
    unless( $DataIP ) {
        $Log->comment("=== ERROR Data IP: $error");
    }

    # Using a VLAN_ID 
    ($DataIP, $error) = Network::SCON_DB::IPReservation->ip_addresses(
        'vlan_id' => <int>,
        'role'    => 'data',
        'count'   => 254,
        'job_id'  => $ENV{'JOBID'}
    );
    unless( $DataIP ) {
        $Log->comment("=== ERROR Data IP: $error");
    }


    ($cleanup, $error) = $DataIP->ip_address_return('job_id' => $ENV{'JOBID'});


    @ip_list = $DataIP->ip_address_list;

    $ip_list_aref = $DataIP->ip_address_list_aref;


    $gateway  = $DataIP->ip_address_gateway;

    $network =  $DataIP->ip_address_network;

    $netmask =  $DataIP->ip_address_netmask;


=head1 PARAMS

=over

=item DATA_IP_RANGE =  Data network IP address range

=item CLUSTER_IP_RANGE =  Cluster network IP address range 

=item FREE_DATA_IP = free data network IP address(es)

=item FREE_DATA_IP_NETMASK = free data network IP address(es) netmask

=item FREE_DATA_IP_GATEWAY = free data network IP address(es) gateway

=item FREE_DATA_IP_CONFIGS Cluster 

=item FREE_CLUSTER_IP = free cluster network IP address(es) 

=item FREE_CLUSTER_IP_NETMASK = free cluster network IP address(es) netmask 

=item FREE_CLUSTER_IP_GATEWAY = free cluster network IP address(es) gateway

=item FREE_DATA_IP_CONFIGS

=back


=head1 Methods

=head2  ip_addresses 

Object constructor making use of Network::IPAddress and Network::SCON_DB::IPAddrManager.


    ($DataIP, $error) = Network::SCON_DB::IPReservation->ip_addresses(
        'hostrec' => $hostrec_object,
        'role'    => 'data',
        'count'   => 254,
        'job_id'  => $ENV{'JOBID'}
    );

    ($DataIP, $error) = Network::SCON_DB::IPReservation->ip_addresses(
        'network_id' => '1.0.16.0',
        'role'    => 'data',
        'count'   => 254,
        'job_id'  => $ENV{'JOBID'}
    );

    # Using a VLAN_ID 
    ($DataIP, $error) = Network::SCON_DB::IPReservation->ip_addresses(
        'vlan_id' => <int>,
        'role'    => 'data',
        'count'   => 254,
        'job_id'  => $ENV{'JOBID'}
    );


Will look in hostrec for network info based on role given, then query the DB 
for addresses as defined by the arguments or defaults. IP addresses are tracked 
in the DB per NATE JOBID. 

Instead of using hostrec info could supply just a network_id or vlan_id
to query the DB for.

Data Structure

If addresses are given
    $DataIP->{'IP'}->{'IPAddr'} = [ {}, {}, {}]    Network::IPAddress hashes
    $DataIP->{'IP'}->{'Network'}{<network id>}{'broadcast'}
    $DataIP->{'IP'}->{'Network'}{<network id>}{'gateway'}
    $DataIP->{'IP'}->{'Network'}{<network id>}{'netmask'}
    $DataIP->{'IP'}->{'Network'}{<network id>}{'netmask_len'}
    $DataIP->{'IP'}->{'add_host_list'} = 0|1
    $DataIP->{'IP'}->{'address'} = <str>   address string given
    $DataIP->{'IP'}->{'addresses'} = <int>   number of addresses
    $DataIP->{'IP'}->{'gateway'}   = <str>   network gateway
    $DataIP->{'IP'}->{'ip_address_list' => [<ip>, <ip>, <ip>]
    $DataIP->{'IP'}->{'netmask'} => <str>  address netmask
    $DataIP->{'IP'}->{'network_id_list'} => [<ip>, <ip>, <ip>]
    $DataIP->{'IP'}->{'network_list'} => [<ip/len>, ,<ip/len>, <ip/len>]
    $DataIP->{'IP'}->{'next_address_idx'} => <int>

If addresses are queried from DB
   $DataIP->{'IP'}->{'_ip_addr_req_boundries'} => {} 
   $DataIP->{'IP'}->{'attrib'}{'_host_counts'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'hosts'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'ip_address_bits'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'ip_broadcast'} => <ip>
   $DataIP->{'IP'}->{'attrib'}{'ip_broadcast_bits'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'ip_gateway'} => <ip>
   $DataIP->{'IP'}->{'attrib'}{'ip_netmask'} => <ip>
   $DataIP->{'IP'}->{'attrib'}{'ip_netmask_bits'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'ip_network'} => <ip>
   $DataIP->{'IP'}->{'attrib'}{'ip_network_bits'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'ip_version'} => <int>
   $DataIP->{'IP'}->{'attrib'}{'vlan_id'} => <int>
   $DataIP->{'IP'}->{'dbh'} = DBI object
   $DataIP->{'IP'}->{'ip_address_list'} = [ip>, <ip>, <ip>]
   $DataIP->{'IP'}->{'ip_address_range'] = <ip1-ipN>
   $DataIP->{'IP'}->{'ip_address_request'} = [ [], [], []]
   $DataIP->{'IP'}->{'last_sql'} = <str>
   $DataIP->{'IP'}->{'messages'} = [mesg, mesg, mesg]
   
   ip_address_request contains row id, host index, version, ip_addr, addr_bits,
   netmask, netmask bits, network, network bits, broadcast, broadcast bits
   gateway, host count, role, job id, vlan id 

=over

=item hostrec => <ref to hostrec>  required if not giving an ID, reference to a hostrec

=item network_id => <string>  required if not giving hostrec or vlan_id, IP network ID

=item vlan_id => <int>  required if not giving hostrec or network_id, VLAN ID 

=item job_id => <job id>    required, job id used to track IP reservation

=item role => <clus|data>    required, role of IP address to request (cluster, Data)

=item count => <int>    number of IP s to request, default=1

=item query   => <any|order>   IP list query, any for a list, order for continuous, default=order

=item version => <4|6|v4|v6|inet4|inet6>   IP version,  default=4

=item network_match => <0|1>   match hostrec network id with FREE_<role>_IP args 

=back


=head2 ip_address_return

    ($cleanup, $error) = $DataIP->ip_address_return('job_id' => <str>);
    ($cleanup, $error) = $DataIP->ip_address_return('address' => <str>);
    ($cleanup, $error) = $DataIP->ip_address_return('address_aref' => ref);

Return (unreserve) IP addresses in bulk or singularly.

=over

=item 'job_id' => <str>      # all IPs with given JOB ID will be returned

=item 'address' => <str>     # given single address will be returned

=item 'address_aref' => ref  # array ref to list of IPs will return each address

=back


Upon suceess will return how many DB entries have been returned.


=head2 ip_address_list

     @ip_list = $DataIP->ip_address_list;

Return the list of IP Addresses obtained from the DB.

=head2 ip_address_list_aref

    $ip_list_aref = $DataIP->ip_address_list_aref;

Return a reference to the list of IP Addresses obtained from the DB.


=head2 compress_ip_address_list

   $DataIP->compress_ip_address_list;

Will update the addresses contained in the address list and the 
broadcast, gateway, netmask and network syntax to a compressed form.

Only compress an address if it is v6.

Only supported for IPs obtained from the DB.

=head2 expand_ip_address_list

   $DataIP->expand_ip_address_list;

Will update the addresses contained in the address list and the 
broadcast, gateway, netmask and network syntax to an expanded form.

Only expands an address if it is v6.

Only supported for IPs obtained from the DB.



=head2 Accessor

Returns a scalar containing the gateway,network,netmask,netmask length 
and address source string respectively.

=head3 ip_address_gateway 

=head3 ip_address_network

=head3 ip_address_netmask

=head3 ip_address_netmask_length

=head3 ip_address_source

    Returns 'param' if IP Addresses are from the FREE_<role>_IP param
    or 'db' if IP Addresses are from the DB

=head1 STATUS

   Current

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date:$

=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=cut

