# $Id: $

## @summary SCON_Utils - library to implement utilities for SCON testing
## @author dl-nwkg-test@netapp.com
## @pod here


=head1 NAME
SCON_Utils - library to implement utilities for SCON testing

=head1 SYNOPSIS

    use Network::SCON_Utils qw(ping_checker, get_release_name, debug_info, get_nodes_in_cluster);

=head1 DESCRIPTION

    This library has a variety of helper methods to deal with verification of network entities.

=head1 ROUTINES

=cut

package Network::SCON_Utils;

use strict;
use warnings;
use Tharn;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF BOOLEAN);
use base qw(Exporter);
use Data::Dumper;
our @EXPORT_OK = qw(ping_checker get_nodes_in_cluster debug_info get_codename get_port_list);
use NATE::BaseException qw(:try);
use POSIX qw(strftime);


#############################################################################
#                                NACL STUFF
#############################################################################
use NACL::C::Node;
use NACL::C::Interface;

my $Node = NACL::C::Node->find();

# Explicitly create a CMode/CLI APISet for the node
my $APISet = NACL::APISet->new(
    hostobj 	=> $Node->hostrec(),
    category 	=> "Node",
    interface 	=> "CLI",
    set 		=> "CMode"
);

=head2 ping_checker 

    my $ret = ping_checker(ping_address => $addr));

This routine does a ping check of the address passed in

It throws exceptions on error.

The following named arguments are accepted,

=over

=item ping_address  

(Required) : lif address to ping

=item client_array   

(Required) : client(s) to perform the ping on

=item ping_count    

(Optional) : number of packets to send(default 5)

=item ping_interval 

(Optional) : delay between send packets(default 1)

=item ping_runid  
  
(Optional) : process id prefix to use with subtest(default 'ping_')

=item network_ping_fail     

(Optional) : if 1, network poing causes test failure (Default 1) 

=item node

(Optional) : Node to perfrom network ping from (default random)

=item debug_mode

(Optional) : If 1, add extra debug info (default 1)

=back 

=for html <hr>

=cut
#############################################################################
# Subroutine: ping_checker
# Parameters: 
#   ping_address  : lif address to ping
#   client_array  : client to perform the ping on
#   ping_count    : number of packets to send
#   ping_interval : delay between send packets
#   ping_runid    : process id to use with subtest
#   network_ping_fail : True if ping failure is test failure
#	node		  : Filer node to ping from
#   debug_mode	  : If true, emit extra debug info
# Purpose: Performs a ping subtest from the client to the lif address 
#          which gives the ability to verify the datapath to the node.
# Returns: Exceptions on error
#############################################################################

sub ping_checker {
    my $timestamp = strftime(".%m%d%y.%H%M%S",localtime) ;

    my %params = validate(
        @_,
        {
             ping_address      => {type => SCALAR, optional => 0},
             client_array      => {type => ARRAYREF|UNDEF, optional => 0},
             ping_count        => {default => 5},
             ping_interval     => {default => 1},
             ping_runid        => {default => 'ping_'},
             ping_fail		   => {type => SCALAR, optional => 1, default => 0},
             network_ping_fail => {type => SCALAR, optional => 1, default => 0},
             node              => {type => SCALAR|UNDEF, optional => 1, default => $Node->name()},
             debug_mode		   => {type => SCALAR, optional => 1, default => 0},
        }
    ); 

    # If we are in debug mode, we want to dump this information before
    # the ping

    if ($params{debug_mode}) {
        $Log->debug(msg => "Dumping debugging... DEBUG_MODE flagged to ".
                        "$params{debug_mode}");
        _dump_port_info(ip => $params{ping_address});
    }

    # Get the arguments for ping test.

    my $ping_args = " -c $params{ping_count} -i $params{ping_interval} $params{ping_address}";

    # Execute ping subtest on client.

    foreach my $client (@{$params{client_array}}) {
        my $ping = Tharn::subtest(         
            "ping $ping_args",
            "-search", [],
            -runid  => $params{ping_runid}.$client.$timestamp,        
            -onexit => sub {},
            -client => $client,         
        );

        # Evaluate subtest results.
        # exit_status() will return 0 on success, 1 on failure
        
        if ($ping->exit_status() && !$params{ping_fail}) {

	    $Log->comment("Check again for a sleep of 30 seconds to ensure LIF responds");
	    sleep (30);
            my $ping = Tharn::subtest(         
            "ping $ping_args",
            "-search", [],
            -runid  => $params{ping_runid}.$client.$timestamp,        
            -onexit => sub {},
            -client => $client,         
        );
	if ($ping->exit_status() && !$params{ping_fail}) {

            # Dump debugging and fatal

            _dump_port_info(ip => $params{ping_address});
            NATE::BaseException->throw("Ping check of $params{ping_address} ".
                              "failed when expected to pass.");
        };
	}
    }

    # Issue a "network ping" from the cluster as well

    # If we are in debug mode, we want to dump this information before
    # the ping

    if ($params{debug_mode}) {
        $Log->debug(msg => "Dumping debugging... DEBUG_MODE flagged to ".
                            "$params{debug_mode}");
        _dump_port_info(ip => $params{ping_address});
    }

    my $node_ping = $APISet->network_ping(
        node        => $params{node},
        destination => $params{ping_address},
        count       => $params{ping_count}, 
        wait        => $params{ping_interval}, 
    );

    if (($node_ping->{processed_output} !~ /alive/) && !$params{network_ping_fail}) {

        # Dump debugging and fatal

        _dump_port_info(ip => $params{ping_address});
        NATE::BaseException->throw("Ping failed from node $params{node} ".
                      "when expected to pass");

    } elsif (($node_ping !~ /alive/) && $params{network_ping_fail}) {
        $Log->comment(msg => "Ping failed from node $params{node} ".
                       "when expected to");
    } elsif (($node_ping =~ /alive/) && $params{network_ping_fail}) {

        # Dump debugging and fatal
 
        _dump_port_info(ip => $params{ping_address});
        NATE::BaseException->throw("Ping succeeded from node $params{node} ".
                      "when expected to fail");
   };
}


=head2 get_nodes_in_cluster 

    my $ret = get_nodes_in_cluster();

This routine returns a arrayref of nodes in the cluster

It throws exceptions on error.

=over

=back 

=for html <hr>

=cut
sub get_nodes_in_cluster {
    my $response = $APISet->cluster_show();
    my $parsed   = $response->get_parsed_output();
    my @nodes = ();
    foreach my $cluster (@{$parsed}){
        push @nodes, $cluster->{'node'};
    }
    return \@nodes;
}

=head2 get_codename

    my $ret = get_codename();

This routine returns aa string of the SUT release

It throws exceptions on error.

=over

=item node_name

(Optional) : name of node to retrieve codename from, defaults to random node

=back 

=for html <hr>

=cut
sub get_codename {
	
	my %params = validate(
        @_,
        {   node_name => {type => SCALAR|UNDEF, optional => 1, default => $Node->name()},
        }
    ); 
    
    	my $node_obj = NACL::C::Node->find(
    		command_interface => $Node,
       		filter  => {
            	node => $params{node_name},
        	});
    
    # These tests only work with releases post BR.   Check that we have 
    # an appropriate build.
	my $version_manager 	= $node_obj->get_version_manager();
    my $release 			= $version_manager->get_version_attribute(attribute => "release");
    return $version_manager->map_release_to_codename(release => $release);
    
}


=head2 debug_info 

    my $ret = debug_info();

This prints out network related state to help with debuggin.

It throws exceptions on error.

=for html <hr>

=cut
sub debug_info {
	$APISet->network_interface_show();
	$APISet->network_port_show();
}


#######################################################################
# Function:    _dump_port_info
# Description: On a connectivity failure, we may need the following information.
# Parameters:  ip: IP address to issue a FreeBSD ping to
# Returns:     Void
#
# Note:        This was originally added for connectivity_check failures,
#              but can be used for anything.
########################################################################

sub _dump_port_info {

    # Define the parameters
    
    my %params = validate (
        @_,
        {
            ip 		=> {type => SCALAR, optional => 1, default  => undef},
            node 	=> {type => SCALAR, optional => 1, default  => undef},
            sudo 	=> {type => SCALAR, optional => 1, default  => "/usr/bin/sudo"},
        },
    );

    # If the calling method specifies a node, we will only run the commands
    # on that node. Otherwise, we will run them on all nodes.

    my @nodes;
      
    if (!defined($params{node})) {
        my @cluster_nodes = NACL::C::Node->find();
        foreach my $node (@cluster_nodes) {
        	push(@nodes, $node);      	
        }
    } else {
    	my $node = NACL::C::Node->find(
    			command_interface => $Node,
    			filter	=> {
    				node => $params{node},
    			});
        push(@nodes, $node);
    }
    
    # Certain commands need sudo access

    my $sudo = $params{sudo};

    foreach my $node_obj (@nodes) {

   		my $sys_shell = NACL::APISet->new(
   		 	hostobj => $node_obj->hostrec(),
            category => 'Node',
            interface => 'CLI',
            set => 'Systemshell',
            shell_type => "CMode",
        );
        # Netstat commands
        my $node = $node_obj->name();
        		
		$sys_shell->execute_raw_command(command => "netstat -s -p ip");
		$sys_shell->execute_raw_command(command => "netstat -s -p icmp");	
		$sys_shell->execute_raw_command(command => 
						"run -node $node -command \"netstat -s -p ip\"");
		$sys_shell->execute_raw_command(command => 
						"run -node $node -command \"netstat -s -p icmp\" ");

        if (defined($params{ip})) {

            # Start tcpdump. Give it a little more time than the default
            
            my $timestamp = strftime("%m%d%y.%H%M%S",localtime) ;
            my $device    = _get_pseudo_device(ip => $params{ip});

            $sys_shell->execute_raw_command(command => "$sudo tcpdump -i $device -w ".
                           "/mroot/tcpdump.$timestamp.pseudo_dev&",
            );
            $sys_shell->execute_raw_command(command => "$sudo tcpdump -i lo0 -w ".
                           "/mroot/tcpdump.$timestamp.lo0&",
            );
            $sys_shell->execute_raw_command(command => "$sudo tcpdump -i lofb -w ".
                           "/mroot/tcpdump.$timestamp.lofb&",
            );
            $sys_shell->execute_raw_command(command => "$sudo tcpdump -i ips0 -w ".
                           "/mroot/tcpdump.$timestamp.ips0&",
            );

            # Start pktt

            $sys_shell->execute_raw_command(command => 
            	"run -node $node -command \"pktt start all\"");

        }
        
        # If a ping IP is specified, we will issue that from FreeBSD as well

        if (defined($params{ip})) {
            $sys_shell->execute_raw_command(command => "ping -c 10 $params{ip}");
        }
        
        # Netstat commands after ping

        $sys_shell->execute_raw_command(command  => "netstat -s -p ip");
        $sys_shell->execute_raw_command(command  => "netstat -s -p icmp");
        
		$sys_shell->execute_raw_command(command => 
						"run -node $node -command \"netstat -s -p ip\"");
		$sys_shell->execute_raw_command(command => 
						"run -node $node -command \"netstat -s -p icmp\"");
        
        if (defined($params{ip})) {

            # Kill tcpdump
        
            $sys_shell->execute_raw_command(command => "$sudo killall tcpdump",);
            
            # Stop pktt
        
            $sys_shell->execute_raw_command(command => 
            	"run -node $node -command \"pktt dump all\"");
            	
            $sys_shell->execute_raw_command(command => 
            	"run -node $node -command \"pktt stop all\"");
        }

        # Get interface information

        $sys_shell->execute_raw_command(command  => "$sudo pcpconfig 0 stat_vif all");
        $sys_shell->execute_raw_command(command  => "ifconfig");
        $sys_shell->execute_raw_command(command => 
            	"run -node $node -command \"ifconfig -a\"");
        $sys_shell->execute_raw_command(command => 
            	"run -node $node -command \"ifstat -a\"");
            	
        # Get routing information

        $sys_shell->execute_raw_command(command => 
            	"run -node $node -command \"route -gsn\"");
        $sys_shell->execute_raw_command(command  => "$sudo netstat -rn");
        $sys_shell->execute_raw_command(command  => "$sudo netstat -in");
		
    } ## end foreach my $node_obj (@nodes)

    return;

} ## end sub _dump_port_info



#############################################################################
# Subroutine: _get_pseudo_device
# Parameters: IP address
# Returns: Name of pseudo device that the IP address resides on
# Purpose: Utility to find the psuedo device (used for running FreeBSD 
#          commands)
#############################################################################
sub _get_pseudo_device {

    # Define the parameters
    
    my %params = validate (
        @_,
        {
            ip 			=> {type => SCALAR, optional => 0,},
       		node_name 	=> {type => SCALAR|UNDEF, optional => 1, default => $Node->name()},
        }
    ); 
    
    my $node_obj = NACL::C::Node->find(
    	command_interface => $Node,
        filter  => {
            node => $params{node_name},
        });
    
    
    # The device is comprised of the server name and the routing-group

	my $lif_obj = NACL::C::Interface->find(
		command_interface 	=> $node_obj,
		filter	=>	{
			address	=> $params{ip},
		},
	);
    
    # Extract what we need and create the device name

	my $lif_state = $lif_obj->state();
    my $p_device = $lif_state->{vserver} . "_" . $lif_state->{routing_group};

    $Log->debug("Returning pseudo device $p_device for IP $params{ip}");
    return $p_device;

} ## end sub _get_pseudo_device

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES



=head1 AUTHOR

    dl-nwkg-test (dl-nwkg-test@netapp.com)
    $Id: $
    Last revised $Date: $

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut

1;
