## @summary SCON Config object
##
## @description Configure a cluster for scon testing 
##
## @author      isarika@netapp.com
##
## @pod here
#
package Network::SconConfig;

# Compiler directives.
use strict;
use warnings;
use Net::IP qw(:PROC);

use TracePkg;
use Options;
use Data::Dumper;
use Tharn qw($Staf subtest $Log host connect);
use Network::SCON_DB::IPReservation;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);


# Module import.
use NACL::APISet;

use NACL::C::Node;
use NACL::C::Client;
use NACL::C::Vserver;
use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::C::NetworkPort;
use NACL::C::Interface;
use NACL::C::Snapmirror;
use NACL::C::VserverNFS;
use NACL::C::VserverNfs;
use NACL::C::Route;
use NACL::C::ExportPolicy;
use NACL::C::ExportPolicyRule;
use NACL::C::VserverExportPolicyRule;
use NACL::C::NetworkRoutingGroupsRoute;
use NACL::C::VserverCifs;
use NACL::C::VserverCifsOptions;
use NACL::C::NetworkOptionsIpv6;
use NACL::C::LunIgroup;
use NACL::C::JobScheduleCron;
use NACL::C::VserverPeer;
use NACL::CS::NetworkPort;
use NACL::CS::NetworkPortIfgrp;
use NACL::CS::ClusterIdentity;
use NACL::CS::NetworkOptionsIpv6;
use NACL::CS::VserverIscsi;

use NACL::STask::Aggregate;
use NACL::STask::NetworkInterface;
use NACL::STask::NetworkPortIfgrp;
use NACL::STask::Snapmirror;
use NACL::STask::Volume;
use NACL::STask::Client::Mount;
use NACL::STask::VserverNFS;
use NACL::STask::NetworkRoutingGroupsRoute;
use NACL::STask::VserverCifsShare;
use NACL::STask::StorageFailover;

use NACL::MTask::AggregateMulti;
use NACL::MTask::VserverMulti;
use NACL::MTask::VolumeMulti;
use NACL::MTask::NFS;
use NACL::MTask::CIFS;
use NACL::MTask::SnapmirrorSet;
use NACL::MTask::Traffic;
use NACL::MTask::LunSetup;
use NACL::MTask::NetworkUtils;
use NACL::MTask::EventLogDetector;

use NACL::Exceptions::NoElementsFound qw(:try);

=head1 NAME

Network::SconConfig

=head1 DESCRIPTION

C<Network::SconConfig> provides methods to configure aggregates,vservers,lifs,volumes 
and snapmirror relationships.Its provides mechanism to configure NFS,CIFS or ISCSI 
configuration based on the inputs provided during SconConfig creation.
This library is built to aid quick and easy ways to configure the ONTAP filer.

It is created on top of NACL Tasks/Componets. 

=head1 ATTRIBUTES

=head2 ENABLE_V6

(Optional)
Flag to decide whether IPv6 should be enabled on the FILER or NOT
Defaults to 1

=head2 SNAPMIRROR_TEST 

(Optional)
Flag to decide whether to enable snapmirror relationship between nodes in the cluster
and provide scheduled updates of 20M.
Defaults to 1

=head2 WINCLIENT

(Optional)
Name of Windows client for CIFS MAP Shares
Defaults to ""

=head2 PROTOCOL

(Optional)
Provide whether to configure NAS or SAN on FILER to facilitate protocol access
Defaults to "NAS"
Valid Values : NAS/SAN

=head2 TYPE

(Optional)
Configure NFS or CIFS or BOTH
Defaults to "NFS,CIFS"
Valid Values : NFS/CIFS/NFS,CIFS 


=head1 METHODS

=head2 new

 # Minimal construction option:
 $sconObj = Network::SconConfig->new();

 # Create for NFS configuration:
 $sconObj = new Network::SconConfig(protocol => "NAS",
  				    type => "NFS");

 # Create for CIFS configuration:
 $sconObj = new Network::SconConfig(protocol => "NAS",
  				    type => "CIFS",
				    WINCLIENT => "csiqa-vmw225");

 # Create for NAS configuration(for NFS and CIFS):
 $sconObj = new Network::SconConfig(protocol => "NAS",
				    WINCLIENT => "csiqa-vmw225");

 type by default is NFS,CIFS

B<new> creates a new object of this class and returns a reference to it. 

=head2 enableV6

 Works on the object created with new to enable IPv6 capabilities on the CLUSTER

 $sconObj->enableV6();

=head2 commonSetup

 Uses the object reference to configure the following on the cluster
 1) aggregates on all the nodes of the cluster
 2) volumes on all these aggregates
 3) one vserver 
 4) 1 IPv4 LIF and 1 IPv6 LIF for the created vserver with appropriate routing entries

 $sconObj->commonSetup();  

=head2 nfsSetup

 Configures NFS server and export policies for V4 and V6 (if ENABLE_V6 = 1) and mounts the data volumes
 on the LINUX Clients

 returns array reference of nfs Mount points 

 $nfsMounts = $sconObj->nfsSetup(
                 vserver => $sconObject->{vservers}->[0],
		 volume  => $sconObject->{$vserver}->{vols},
		 action  => "reuse",
		);

=head2 cifsSetup

 Configures CIFS server and NAME MAPPINGS or V4 and V6 (if ENABLE_V6 = 1) and maps the shares on the
 on the Windows Clients

 returns array reference of CIFS MAP shares 

 $cifsShares = $sconObj->cifsSetup(
                 vserver => $sconObject->{vservers}->[0],
		 volume  => $sconObject->{$vserver}->{vols},
		 action  => "reuse",
		);

=head2 vserverPeerSetup

 Configure vserver Peer relationship with another vserver on remote node of the cluster for intracluster
 snapmirror relationships

  $sconObj->vserverPeerSetup( vserver => $vserver );

=head2 snapMirrorSetup

 Configures Snapmirror relationship of type DP between 2 vservers 

 $sconObj->snapMirrorSetup( vserver => $vserver );

=cut

use constant SNAP_SCHEDULE => "15M";

my $Options = new Options();
my $Trace   = new TracePkg();

my %node_details = ();
my %data_lifs    = ();
my $job_id       = $ENV{'JOBID'};

use vars qw(
  $node
  $nodeobj
  $nodesobj
  $hostobj
  $Client
  $Client_win
  $WINCLIENT
  $CLIENT
  @dstVolObjs
  $dstVolObjs
  @vol_objects
  @s_dvs_names
  @vserver_names
  $dvs_names
  $mount_obj
  $num_vserver
  $num_vols
  @aggr_names
  $nfs_mounts
  $cifs_mounts
  $cluster_ips
  $data_ips
  $node_mgmt_ips
  $ClusIP
  $DataIPV4
  $DataIPV6
  $error
  $detector
  @d_dvs_names
);

sub new {

    $Trace->entry();
    my $class       = shift;
    my $Debug_Level = 1;

    my $self  = {};
    my @nodes = NACL::C::Node->find();
    $Options->parameters(
        definitions => [
            {
                name => "SNAPMIRROR_TEST",
                description =>
"Request for snap mirror configuration and updates verification as part of base setup",
                valid_values => [qw(0 1)],
                default      => "1",
            },
            {
                name        => "CLIENT_USERNAME",
                description => "Username of clients to mount volumes and luns",
                default     => "root",
            },
            {
                name        => "CLIENT_PASSWORD",
                description => "Password credentials of clients",
                default     => "netapp1!",
            },
            {
                name         => "ENABLE_V6",
                description  => "Enable IPv6",
                valid_values => [qw(0 1)],
                default      => "1",
            },
            {
                name        => "CIFS_USERNAME",
                description => "CIFS Username details.Required to purge vs0",
                default     => "administrator",
            },
            {
                name        => "CIFS_PASSWORD",
                description => "CIFS Password details.",
                default     => "netapp"
            },
            {
                name        => "CIFS_DOMAIN",
                description => "CIFS DOMAIN",
                default     => "csiqa.rtp.netapp.com"
            },
            {
                name        => "CIFS_SERVER",
                description => "CIFS Password details.",
                default     => "CIFSTest"
            },
            {
                name        => "DNS_DOMAIN",
                description => "DNS Domain entries",
                default     => "csiqa.rtp.netapp.com"
            },
            {
                name        => "WINCLIENT",
                description => "Windows client details",
		default     => "",
            },
            {
                name        => "PROTOCOL",
                description => "Define the data protocol to use",
		valid_values=> [qw(NAS SAN)],
		default     => "NAS",
            },
            {
                name        => "TYPE",
                description => "Define the type of data protocol to use",
		default     => "NFS,CIFS",
            },

        ]
    );

    $Options->process();
    $Log->debug( "Options = " . Dumper($Options) );
    my $opts = $Options->get_hash();

    $Log->comment("Check if IPManager support is available, if NOT, populate using Tharnhosts");
    #init the ip lists
    my $clus_ip_pool   = [];
    my $data_ipv4_pool = [];
    my $data_ipv6_pool = [];
    my $clus_netmask = "";
    my $data_netmask = "";
    my $data_gateway = "";
    my $data_netmask_len = "";
    my $data_v6_gateway = "";

    my $hostrec_object = host $nodes[0]->node();

    my ($dataIP, $err) = Network::SCON_DB::IPReservation->ip_addresses(
        'hostrec' => $hostrec_object,
        'role'    => 'data',
        'count'   => 1,
        'job_id'  => $job_id,
        'version' => "v4"
    );
    if ($err) {
       $Log->comment("IPManager support not available,derive details from Tharnhosts");
       my @filtered_networks;
       push @$clus_ip_pool ,  $hostrec_object->{CLUSTER_IP1};
       push @$clus_ip_pool ,  $hostrec_object->{CLUSTER_IP2};
       push @$data_ipv4_pool , $hostrec_object->{DATA_IP1};
       push @$data_ipv4_pool , $hostrec_object->{DATA_IP2};
       $clus_netmask = $hostrec_object->{CLUSTER_MASK1};
       $data_netmask = $hostrec_object->{DATA_MASK1};
       $data_gateway = $hostrec_object->{DATA_GATEWAY1};
       if ($opts->{ENABLE_V6}) {
          my @filter = ({ netid => "inet6_data1" },
	                { netid => "inet6_data2" },);
	  @filtered_networks =
	                $hostrec_object->net_filter(include => \@filter);
          if (scalar(@filtered_networks) <= 0)
         {
             $Log->warn(
		"Error : No inet_6 net records found in the tharnhosts.Please update tharnhosts with inet6 records"
             );
           return;
         }
         foreach my $filtered_network (@filtered_networks) {
	     push @$data_ipv6_pool  , $filtered_network->{address};
	     $data_netmask_len = $filtered_network->{netmask};
	     $data_v6_gateway = $filtered_network->{gateway};
	 }
      }
    } else {
       $Log->comment("IPManager support available");
    #IPManager to allocate IPS

    $Log->comment("=== Get IP addressing");
##Commenting out the cluster ip part, as we are not using cluster ip's currently in this library
=cut
    $Log->comment("=== Cluster Addressing");
    ( $ClusIP, $error ) = Network::SCON_DB::IPReservation->ip_addresses(
        'hostrec' => $hostrec_object,
        'role'    => 'clus',
        'count'   => 2,
        'job_id'  => $job_id
    );
    unless ($ClusIP) {
        $Log->comment("=== ERROR Clus IP: $error");
    }
=cut

    $Log->comment("=== Data Addressing");
    ( $DataIPV4, $error ) = Network::SCON_DB::IPReservation->ip_addresses(
        'hostrec' => $hostrec_object,
        'role'    => 'data',
        'count'   => 2,
        'job_id'  => $job_id,
        'version' => "v4"
    );
    unless ($DataIPV4) {
        $Log->comment("=== ERROR Data IP: $error");
    }

    if ( $opts->{ENABLE_V6} ) {
        ( $DataIPV6, $error ) = Network::SCON_DB::IPReservation->ip_addresses(
            'hostrec' => $hostrec_object,
            'role'    => 'data',
            'count'   => 2,
            'job_id'  => $job_id,
            'version' => "v6"
        );
        unless ($DataIPV6) {
            $Log->comment("=== ERROR Data IP: $error");
        }
    }
    $data_netmask = $DataIPV4->ip_address_netmask;
    $data_gateway = $DataIPV4->ip_address_gateway;
    # load the ip list
    #@$clus_ip_pool   = $ClusIP->ip_address_list;
    @$data_ipv4_pool = $DataIPV4->ip_address_list;
    if ($DataIPV6) {
        @$data_ipv6_pool = $DataIPV6->ip_address_list;
    	$data_netmask_len = $DataIPV6->ip_address_netmask_length;
    	$data_v6_gateway = $DataIPV6->ip_address_gateway;
    $Log->comment(
        sprintf( "=== IPv6 Network: %s", $DataIPV6->ip_address_network ) );
    $Log->comment(
        sprintf( "=== IPv6   Netmask: %s", $DataIPV6->ip_address_netmask ) );
    }

    # cluster range string
    #my $size = scalar @$clus_ip_pool;
    #$Log->comment("Size of cluster ips is $size");
    #my $cluster_range = $clus_ip_pool->[0] . "\/$size";

    # data range strings
    my $size = scalar @$data_ipv4_pool;
    $Log->comment("Size of data ips is $size");
    my $data_range = $data_ipv4_pool->[0] . "\/$size";
 }

    my %data = ();
    $data{v4}{ips}     = $data_ipv4_pool;
    $data{v4}{netmask} = $data_netmask;
    $data{v4}{gateway} = $data_gateway;

      $data{v6}{ips}     = $data_ipv6_pool;
      $data{v6}{netmask} = $data_netmask_len;
      $data{v6}{gateway} = $data_v6_gateway;


    $self->{num_vservers}    = $opts->{NUM_VSERVERS};
    $self->{num_vols}        = $opts->{NUM_VOLS};
    $self->{snapmirror_test} = $opts->{SNAPMIRROR_TEST};
    $self->{client_username} = $opts->{CLIENT_USERNAME};
    $self->{client_password} = $opts->{CLIENT_PASSWORD};
    $self->{data}            = \%data;
    $self->{nodes}           = \@nodes;
    $self->{enable_v6}       = $opts->{ENABLE_V6};
    #$self->{cifs_username}   = $opts->{CIFS_USERNAME};
    #$self->{cifs_password}   = $opts->{CIFS_PASSWORD};
    #$self->{cifs_domain}     = $opts->{CIFS_DOMAIN};
    #$self->{cifs_server}     = $opts->{CIFS_SERVER};
    #$self->{dns_domain}      = $opts->{DNS_DOMAIN};
    $self->{cifs_username}   = Tharn::param('CIFS_USER') ||  Tharn::param('CIFS_SUPERUSERS');
    $self->{cifs_password}   = Tharn::param('CIFS_PASSWD') || "netapp";
    $self->{cifs_domain}     = Tharn::param('CIFS_DOMAIN_LONG'); 
    $self->{cifs_server}     = $opts->{CIFS_SERVER};
    $self->{dns_domain}      = Tharn::param('FILER_DNS_DOMAINNAME');
        ##### Pushing only two values from DNS_SERVER_IP list ###########
    if ( Tharn::param('SITE') eq "rtpeng" ) {
    my @ad_server_ip         = split(/,/,Tharn::param('AD_SERVER_IP'));
    $self->{name_servers}    ="$ad_server_ip[0]" . "," . "$ad_server_ip[1]";
    }
    else {
    my @dns_server_ip         = split(/,/,Tharn::param('FILER_DNS_SERVER_LIST'));
    $self->{name_servers}    ="$dns_server_ip[0]" . "," . "$dns_server_ip[1]";
    }
    $self->{protocol}        = $opts->{PROTOCOL};
    $self->{protocol_type}   = $opts->{TYPE};



    $Log->comment("Create Client connection");
    $Client = NACL::C::Client->find();

    $Client->apiset_must(
        connid   => "rlogin",
        username => 'root',
    );
    $self->{client} = $Client;
    if ( $opts->{WINCLIENT} ) {
        $Log->comment("Windows client object creation");
        $Client_win = NACL::C::Client->new( name => $opts->{WINCLIENT} );
        $self->{win_client} = $Client_win;
    }

    foreach $nodesobj (@nodes) {
        $node = $nodesobj->node();
        my $hostobj = host $node;
        my $aggr    = $node . '_aggr1';
        $aggr =~ s/-/_/g;
        $node_details{$node}{aggr}         = $aggr;
        $node_details{$node}{igroup}       = "igroup_" . _random_string();
        $node_details{$node}{data_port1}   = $hostobj->{DATA_PORT1},
          $node_details{$node}{data_port2} = $hostobj->{DATA_PORT2},
          ;
    }
    $self->{node_details} = \%node_details;

    $Log->debug("Start logging events from now on .. ");
    $detector = NACL::MTask::EventLogDetector->new(
            command_interface => $nodes[0]);

     my $begin_date
             = NACL::CS::SystemNodeDate->fetch( 'command_interface' => $nodes[0] )
	             ->date();

    $detector->start();

    $self->{detector} = $detector;

    bless $self, $class;
    $Log->debug( "Initialized baseC object with the "
          . "following configuration:\n"
          . Dumper($self) );



    $Log->exit("Returning blessed reference: $self");
    return $self;
}

sub commonSetup {

    $Trace->entry();
    $Log->debug("Entering commonSetup");
    my $self   = shift(@_);
    my $action = "reuse";
    my $nodes  = $self->{nodes};
    foreach $nodesobj (@$nodes) {
        my $node = $nodesobj->node();
        $Log->debug("\n\n--------------- node = $node -------------------\n");

        # Create an aggregate
        $Log->debug(" Create Aggregate ...");
        @aggr_names = NACL::MTask::AggregateMulti->create_multi(
            command_interface  => $nodesobj,
            aggregate          => $self->{node_details}->{$node}{aggr},
            diskcount          => 5,
            nacltask_if_exists => $action,

        );

    }

    if ( !vserverOps( $self, op => "create" ) ) {
        return 0;
    }
    return 1;
}

sub vserverOps {

    $Log->comment("Vserver operations");
    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        op => {
            type     => SCALAR,
            required => 1
        },
        lifs_create => {
            type     => SCALAR,
            required => 0,
            default  => 1
        },
        dataVol_create => {
            type     => SCALAR,
            required => 0,
            default  => 1
        },
        action => {
            type     => SCALAR,
            required => 0,
            default  => "reuse"
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $self->_set_err( "failed to validate parameter: " . $@ );
        return 0;
    }

    my $nodes = $self->{nodes};
    my $node  = $nodes->[0]->node();
    if ( $args{op} eq "create" ) {
        $Log->comment(
" Create a vserver hosting lifs for $self->{protocol_type}"
        );
        my @s_dvs_names = NACL::MTask::VserverMulti->create_multi(
            command_interface => $nodes->[0],

            #vserver =>  ["vs_" . _random_string()],
            vserver =>  ["vs11"],
            aggregate          => $self->{node_details}->{$node}->{aggr},
            nacltask_if_exists => $args{action},

        );

        $self->{vservers}                    = \@s_dvs_names;
        #$self->{ $s_dvs_names[0] }{protocol} = "NAS";dd
        #$self->{ $s_dvs_names[1] }{protocol} = "SAN";
        if ( $args{dataVol_create} ) {
            $self->volOps( op => "create", vserver => \@s_dvs_names );
        }

        if ( $args{lifs_create} ) {

            my $protocol   = lc("$self->{protocol_type}"); 
            my $autorevert = "";
            my $i          = 1;
            foreach my $vs (@s_dvs_names) {

                if ( $self->{protocol} eq "NAS" ) {
                    $autorevert = "true";
                }
                else {
                    $autorevert = "false";
                }

                #my $v4_pool         = $self->{data}->{v4}{ips};
                my $ipv4         = shift @{$self->{data}->{v4}{ips}};
                #my $ipv4            = shift @$v4_pool;
                my $interfaceObj_v4 = NACL::STask::NetworkInterface->create(
                    command_interface => $nodes->[0],
                    lif               => "lif_$node" . _random_string(),
                    vserver           => $vs,
                    role              => "data",
                    'home-port' =>
                      $self->{node_details}->{$node}->{"data_port$i"},
                    'home-node'        => $node,
                    'data-protocol'    => ["$protocol"],
                    'auto-revert'      => $autorevert,
                    nacltask_if_exists => "purge",
                    address            => $ipv4,
                    'netmask'          => $self->{data}->{v4}{netmask},
		     nacltask_if_exists => "reuse"
                );
                NACL::STask::NetworkRoutingGroupsRoute->create(
                    command_interface => $nodes->[0],
                    vserver           => $vs,
                    'routing-group' =>
                      $interfaceObj_v4->state()->{'routing_group'},
                    destination        => "0.0.0.0/0",
                    gateway            => $self->{data}->{v4}->{gateway},
                    nacltask_if_exists => "purge",
                );
		$Log->comment("Ping the gateway .. prior to mount. Causing mount problems due to reachability issues");
		my $nodescope_apiset_node = $nodes->[0]->get_7m_or_nodescope_apiset();
		try {
		  my $response_ping = $nodescope_apiset_node->ping('destination' => "$self->{data}->{v4}->{gateway}",
		  						    count        => "3",
								    interface    => $self->{node_details}->{$node}->{"data_port$i"});

		}
		catch NACL::APISet::Exceptions::ResponseException with {
		        my $exception_object = shift;
		        my $resp             = $exception_object->get_response_object();
		        $Log->comment( "Caught a " . ref($exception_object) . "!" );
			$Log->comment( "Error is: " . $exception_object->text() );
		        $Log->comment( "Raw output was:\n" . $resp->get_raw_output() );
		};
                my $interfaceObj_v6;
                if ( $self->{enable_v6} ) {
		    $Log->comment("Creating v6 lif");
                    #my $v6_pool = $self->{data}->{v6}{ips};
                    my $ipv6 = shift @{$self->{data}->{v6}{ips}};;
                    #my $ipv6    = shift @$v6_pool;
                    my $netmask = $self->{data}->{v6}{netmask};
                    $netmask =~ s/\///g;
                    $Log->comment("netmask length is now $netmask");
                    $interfaceObj_v6 = NACL::STask::NetworkInterface->create(
                        command_interface => $nodes->[0],
                        lif               => "lif_$node" . _random_string(),
                        vserver           => $vs,
                        role              => "data",
                        'home-port' =>
                          $self->{node_details}->{$node}->{"data_port$i"},
                        'home-node'        => $node,
                        'data-protocol'    => ["$protocol"],
                        'auto-revert'      => $autorevert,
                        nacltask_if_exists => "purge",
                        address            => $ipv6,
                        'netmask-length'   => $netmask,
			nacltask_if_exists => "reuse"
                    );

                    NACL::STask::NetworkRoutingGroupsRoute->create(
                        command_interface => $nodes->[0],
                        vserver           => $vs,
                        'routing-group' =>
                          $interfaceObj_v6->state()->{'routing_group'},
                        destination        => "::/0",
                        gateway            => $self->{data}->{v6}->{gateway},
                        nacltask_if_exists => "purge",
                    );
		    
		try {
		  my $dest = Net::IP::ip_compress_address( $self->{data}->{v6}->{gateway}, 'IPv6' );
		  $ipv6 =  Net::IP::ip_compress_address( $ipv6, 'IPv6' );
		  my $response_ping = $nodescope_apiset_node->ping6(destination => $dest, 
		  						     'source-address' => $ipv6,
								     count	      => "3");
		}
		catch NACL::APISet::Exceptions::ResponseException with {
		        my $exception_object = shift;
		        my $resp             = $exception_object->get_response_object();
		        $Log->comment( "Caught a " . ref($exception_object) . "!" );
			$Log->comment( "Error is: " . $exception_object->text() );
		        $Log->comment( "Raw output was:\n" . $resp->get_raw_output() );
		};

                }

                $self->{$vs}{lif_v4} = $interfaceObj_v4;
                if ($interfaceObj_v6) {
                    $self->{$vs}{lif_v6} = $interfaceObj_v6;
                }
                $i++;
                my $policy = "policy" . _random_string();
                $self->{$vs}{exportpolicy} = $policy;
            }

        }
    }
    else {

        $Log->comment(
            "Delete any vserver peering prior to deletion of vserver");
        my $vservers = $self->{vservers};
        foreach my $vs (@$vservers) {
            try {
                my $VserverPeer = NACL::C::VserverPeer->find(
                    command_interface => $nodes->[0],
                    filter            => { vserver => $vs },
                );
                my $peer_vserver = $VserverPeer->{'peer_vserver'};
                $Log->comment("Peer vserver is $peer_vserver");
                NACL::C::VserverPeer->delete(
                    command_interface => $nodes->[0],
                    'peer-vserver'    => $peer_vserver,
                    vserver           => $vs
                );
            }
            catch NACL::Exceptions::NoElementsFound with {
                $Log->comment("No vserver peering relationship");
            };
            my $vserver_obj = NACL::STask::Vserver->find(
                command_interface => $nodes->[0],
                filter            => { vserver => $vs },
                allow_empty       => 1
            );

            $vserver_obj->purge(
                nacltask_cifs_username => $self->{cifs_username},
                nacltask_cifs_password => $self->{cifs_password}
            );
        }
    }

}

sub volOps {
    $Log->comment("Vol operations");
    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        op => {
            type     => SCALAR,
            required => 1
        },
        vserver => {
            type     => ARRAYREF,
            required => 1
        },
        size => {
            type     => SCALAR,
            required => 0,
            default  => "20g"
        },
        dataVol_type => {
            type     => SCALAR,
            required => 0,
            default  => "RW"
        },
        action => {
            type     => SCALAR,
            required => 0,
            default  => "reuse"
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {

        #$self->_set_err("failed to validate parameter: ". $@);
        $Log->error( "failed to validate parameter :" . $@ );
        return 0;
    }

    my $nodes = $self->{nodes};
    my $node  = $nodes->[0]->node();

    my $vservers = $args{vserver};
    $Log->comment(
        " Create data volumes spanning across all nodes in the cluster ...");
    foreach my $vs (@$vservers) {

        # Empty the vol_objects once the per vserver volumes are added
        my @vol_objects = ();
        foreach my $nodeobj (@$nodes) {
            my $node        = $nodeobj->node();
	    my $vol         = $node . '_vol';
	    $vol =~ s/-/_/g;

            my @s_vol_names = NACL::MTask::VolumeMulti->create_multi(
                command_interface  => $nodeobj,
                aggregate          => $self->{node_details}->{$node}->{aggr},
                vserver            => $vs,
                #volume             => "vol_" . _random_string(),
                volume             => $vol,
                type               => "RW",
                "unix-permissions" => "---rwxrwxrwx",
                nacltask_num_vols  => 1,
                if_exists          => $args{action},
                size               => $args{size},
            );

            # Creating source volume object
            my $svol = NACL::STask::Volume->find(
                command_interface => $nodeobj,
                filter            => { volume => $s_vol_names[0] },
            );
            push @vol_objects, $svol;
        }
        $self->{$vs}{vols} = \@vol_objects;
        $Log->comment("vserver=$vs and volumes=");

    }
}

sub vserverPeerSetup {

    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        vserver => {
            type     => SCALAR,
            required => 1
        },
	node => {
	    type    => HASHREF,
	    default => "",
	}
    );

    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }

    my $nodes   = $self->{nodes};
    my $size = scalar @{$nodes};
    my $nodeobj;
    my $node;
    if (!$args{node}) {
    $nodeobj = $nodes->[ $size - 1 ];
    $node    = $nodeobj->name();
    } else {
      $nodeobj = $args{node};
      $node    = $nodeobj->name();
    }

    try {
        @d_dvs_names = NACL::MTask::VserverMulti->create_multi(
            command_interface  => $nodeobj,
            vserver            => [ "drVserver1", "drVserver2" ],
            aggregate          => $self->{node_details}->{$node}{aggr},
            nacltask_if_exists => "reuse",
        );

        #my @vservers = $self->{vservers};
	# some issue with SnapMirrorSet hence setting dp volumes uniqiue fr each vserver 
	my $i = "x";
        foreach my $vserver (@d_dvs_names) {
	    my $vol_suffix = "dp". $i;
            $Log->comment("Create destination volumes of type DP for vserver $vserver");
	    #push @{$self->{vservers}}, $vs;
            my $job;
	    my @vol_names = NACL::MTask::VolumeMulti->create_multi(
	            command_interface    => $nodeobj,
		    aggregate            => $self->{node_details}->{$node}{aggr},
		    vserver              => $vserver,
	            volume               => $vol_suffix,
	            nacltask_compression => "30", 
	            nacltask_compression_inline => "30",
		    type                        => "DP",
	            nacltask_guarantee_file     => "30",
	            nacltask_num_vols           => "2", 
		    size			=> "20g"
           );
            $Log->comment(" Destination DP Volumes created.");
	    @dstVolObjs = ();
	    $self->{dstVservers}{$vserver} = ();
	    #printf Dumper($self->{dstVservers}{$vserver});
	    foreach my $vol (@vol_names) {
	       my $dvol = NACL::STask::Volume->find(
	                       command_interface => $nodeobj,
			       filter            => { volume => $vol ,
			       			      vserver => $vserver},
					         );

            push @dstVolObjs, $dvol;
	    my $key = $vserver;
	    $Log->comment("Key=$key***");
	    push(@{$self->{dstVservers}{$key}}, $dvol);
	    }
	    $Log->comment("now after adding to self .. details are ");
	    #printf Dumper($self->{dstVservers}); 
	    $i = "y";
        }

	foreach my $v (@d_dvs_names) {
	   my $vols = $self->{dstVservers}{$v} ;
	   #printf Dumper($vols);
	}
        foreach my $vs (@d_dvs_names) {
            try {
                my $peerObj = NACL::C::VserverPeer->find(
                    command_interface => $nodeobj,
                    filter            => { vserver => $vs },
                );
            }
            catch NACL::Exceptions::NoElementsFound with {
                my $peer = NACL::C::VserverPeer->create(
                    command_interface => $nodeobj,
                    'peer-vserver'    => $args{vserver},
                    'applications'    => ["snapmirror"],
                    vserver           => $vs
		    );
	    };
	}
	}
	catch NATE::BaseException with {
		my $exception = shift;
	        printf Dumper($exception);
    };
    }

sub snapMirrorSetup {

    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        vserver => {
            type     => SCALAR,
            required => 1
        },
        action => {
            type     => SCALAR,
            required => 0,
            default  => "reuse"
        },
	isScheduled => {
	    type    => SCALAR,
	    required => 0,
	    default  => 1
	}
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }

    my $Job;
    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[0];
    my $node    = $nodeobj->node();
    my $vs      = $args{vserver};

    $Log->comment("Setup Snapmirror schedule");
    my @min =
      #( "5", "10", "15", "20", "25", "30", "35", "40", "45", "50", "55" );
      ( "15", "30", "45");
    try {
        $Job = NACL::C::JobScheduleCron->find(
            command_interface => $nodeobj,
            filter            => { name => SNAP_SCHEDULE },
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Job = NACL::C::JobScheduleCron->create(
            command_interface => $nodeobj,
            name              => SNAP_SCHEDULE,
            minute            => \@min,
        );
    };

    $Log->comment("\nCreate and Initialize SnapMirror relationships..");
    my $vols = $self->{$vs}->{vols};
    foreach my $v (@$vols) {
       my $dstVols = "";
       
       my $dvs = pop @d_dvs_names;
       $dstVols = $self->{dstVservers}{$dvs};
       #printf Dumper($dstVols);
       my $smset = "";
    if ($args{isScheduled}) {
    $Log->comment("SnapMirror with Schedule");
    $smset = NACL::MTask::SnapmirrorSet->new(
        source             => $v,
        destinations       => { DP => $dstVols },
        snapmirror_options => { all => { schedule => SNAP_SCHEDULE }, }
    );
    } else {
    $Log->comment("SnapMirror without Schedule");
    $smset = NACL::MTask::SnapmirrorSet->new(
        source             => $v,
        destinations       => { DP => $dstVols },
    );
    }
    $Log->comment("Snapmirror initialize");
    $smset->initialize( nacltask_wait => 0.5 );
    $Log->comment("Snapmirror Monitor");
    $smset->monitor;
    $Log->comment("Snapmirror relationship Established.");
    $self->{snapMirrors}{$v} = $smset;
    return $smset;
  }

}

sub enableV6 {

    #($nodeobj) = @_;
    #my ($nodeobj) = @_;
    my $self    = shift(@_);
    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[0];

    $Log->comment("\n Enable IPv6 ...");

    my $ipv6_status =
      NACL::CS::NetworkOptionsIpv6->fetch( command_interface => $nodeobj );
    if ( $ipv6_status->{'enabled'} =~ m/false/ ) {
        $Log->comment("\n Enable IPv6 Option ...");
        my $ipv6_enable = NACL::C::NetworkOptionsIpv6->modify(
            command_interface => $nodeobj,
            enabled           => "true"
        );
        my $ipv6_status =
          NACL::CS::NetworkOptionsIpv6->fetch( command_interface => $nodeobj );
        if ( $ipv6_status->{'enabled'} =~ m/false/ ) {
            $Log->comment(" FAILED to Enable Ipv6 option!!");
            return 0;
        }
    }
    return 1;
}

sub nfsSetup {
    $Log->comment("\n NFS Setup");
    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        vserver => {
            type     => SCALAR,
            required => 1
        },
        action => {
            type     => SCALAR,
            required => 0,
            default  => "reuse"
        },
        volume => {
            type     => ARRAYREF,
            required => 1
        },
        addrType => {
            type     => SCALAR,
            required => 0,
            default  => "v4,v6"
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }

    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[0];
    my $node    = $nodeobj->node();
    my $vs      = $args{vserver};
    my $addr;
    my %opts;

    my $setupV4_V6 = 0;
    if ( $args{addrType} eq "v4" ) {
        $addr = $self->{$vs}->{lif_v4}->state()->{'address'};
        $Log->comment("addr is $addr");
        %opts = (
            'address' => $addr,

            #'netmask' => $self->{node_details}->{$node}->{data_netmask1},
            'gateway' => $self->{data}->{v4}->{gateway},

            #'home-port' => $self->{node_details}->{$node}->{data_port1},
            #'home-node' => $nodeobj,
            'clientmatch' => "0.0.0.0/0",
        );
    }
    elsif ( $args{addrType} eq "v6" ) {
        $addr = $self->{node_details}->{$node}->{data_ip_inet6_data1};
        %opts = (
            'address' => $self->{node_details}->{$node}->{data_ip_inet6_data1},
            'netmask' =>
              $self->{node_details}->{$node}->{data_netmask_inet6_data1},
            'gateway' =>
              $self->{node_details}->{$node}->{data_gateway_inet6_data1},
            'home-port'   => $self->{node_details}->{$node}->{data_port1},
            'home-node'   => $nodeobj,
            'clientmatch' => "::/0",
        );
    }
    else {
        $Log->comment("NFS Setup for both v4 and v6");
        $addr = $self->{$vs}{lif_v4}->state()->{'address'};
        my $lif = $self->{$vs}->{lif_v4}->state()->{'lif'};
        %opts = (
            'address' => $addr,
            lif       => $lif,

            #'netmask' => $self->{node_details}->{$node}->{data_netmask1},
            'gateway' => $self->{data}->{v4}->{gateway},

            #'home-port' => $self->{node_details}->{$node}->{data_port1},
            #'home-node' => $nodeobj,
            'clientmatch' => "0.0.0.0/0",
        );

        $setupV4_V6 = 1;

    }

    $Log->comment(" NFS Setup for volumes of vserver $args{vserver}  ...");
    my $vols = $self->{ $args{vserver} }->{vols};
    foreach my $v (@$vols) {

        my $vol    = $v->{volume};
        my $policy = $self->{ $args{vserver} }->{exportpolicy};

        my $nfs_obj = NACL::MTask::NFS->setup(
            'command_interface'                    => $nodeobj,
            'vserver'                              => $args{vserver},
            address                                => $addr,
            'ruleindex'                            => 777,
            'policyname'                           => $policy,
            'protocol'                             => ["any"],
            'rorule'                               => ["any"],
            'rwrule'                               => ["any"],
            'volume'                               => $vol,
            'junction-path'                        => "$vol",
            'anon'                                 => 0,
            'superuser'                            => ["any"],
            'nacltask_if_lif_exists'               => $args{action},
            'nacltask_if_routinggrouproute_exists' => $args{action},
            'v4.0'                                 => 'enabled',
            'nacltask_if_exportpolicy_exists'      => $args{action},
            'nacltask_if_exportpolicyrule_exists'  => $args{action},
            'nacltask_if_exists'                   => $args{action},
            'nacltask_if_mount_exists'             => $args{action},
            %opts,
        );

        if ($setupV4_V6) {
            $Log->comment("Explicitly add v6 export policy");

            # Export policy rule create for IPv6
            $Log->comment(" Explicit - Export policy rule create for IPv6 ...");
            NACL::C::VserverExportPolicyRule->create(
                command_interface => $nodeobj,
                vserver           => $args{vserver},
                policyname        => $policy,
                ruleindex         => 778,
                clientmatch       => "::/0",
                superuser         => ["any"],
                protocol          => ["any"],
                rorule            => ["any"],
                rwrule            => ["any"],
            );
        }
        $Log->comment(" Creating NFS mounts ...");
        $Log->comment(" Find the volume to be mounted ...");
        $vol = NACL::STask::Volume->find(
            command_interface => $nodeobj,
            filter            => { volume => $v },
        );
        $Log->comment(" Create the mount point ...");

        # Mount the volume
        $mount_obj = NACL::STask::Client::Mount->create(
            command_interface => $self->{client},
            volume            => $vol,
	    #nacltask_address_family => "ipv6",
        );
        $Log->comment(" Get the mount point ...");

        # Get the mount point
        my $mount_point = $mount_obj->mount_point();
        $Log->comment("Mount point is $mount_point");

        push my @logical_interfaces, "ipv4";

        # enable when v6 ips are ready
        if ( $self->{enable_v6} ) {
            push @logical_interfaces, "ipv6";
        }

        my $verify;
        my $Mount_pt;
        my $mpt;
        my $version;
        my $protocol;
        my @versions;
        my @protocols;

        #my @logical_interfaces;
        my $logical_interface;
        my $lif_ip;

        #$nfs_mounts = [];
        foreach $logical_interface (@logical_interfaces) {
            print "isarika>> inside foreach of logical_interfaces loop\n";
            print "isarika>> ip = $logical_interface *****\n";
            if ( $logical_interface eq "ipv6" ) {

                $lif_ip    = _getIP( $self->{ $args{vserver} }->{lif_v6} );
		#$lif_ip   = $lif_ip;
                @protocols = ( "udp6", "tcp6" );
                $verify    = 0;
            }
            else {
                $lif_ip = _getIP( $self->{ $args{vserver} }->{lif_v4} );
                @protocols = ( "udp", "tcp" );

                #$verify    = 1;
                $verify = 0;
            }
            $Log->comment(
" Mount the volume using same $logical_interface LIF over different NFS versions from the Client ..."
            );
            my $mpt1 = "$mount_point" . "_tcp_nfsv3_" . "$logical_interface";
            my $mpt2 = "$mount_point" . "_udp_nfsv3_" . "$logical_interface";
            my $mpt3 = "$mount_point" . "_tcp_nfsv4_" . "$logical_interface";

            #my $mpt4 = "$mount_point"."_udp_nfsv4_"."$logical_interface";
            foreach $protocol (@protocols) {
                print "isarika>> In protocols loop of foreach\n";
                if ( $protocol =~ m/udp/ ) {
                    @versions = ("nfsv3");
                }
                else {
                    @versions = ( "nfsv3", "nfsv4" );
                }
                foreach $version (@versions) {
                    if ( $version =~ m/4/ ) {
                        if ( $protocol =~ m/tcp/ ) {
                            $mpt = $mpt3;
                        }
                        else {

                            # nfsv4 over udp is not supported in ONTAP
                        }
                    }
                    else {
                        if ( $protocol =~ m/tcp/ ) {
                            $mpt = $mpt1;
                        }
                        else {
                            $mpt = $mpt2;
                        }
                    }
                    $version =~ s/nfsv//g;
                    print "isarika>> Mounting the details";
                    $Mount_pt = NACL::STask::Client::Mount->create(
                        command_interface  => $self->{client},
                        volume             => $vol,
                        mount_point        => $mpt,
                        server             => $lif_ip,
                        vers               => $version,
                        proto              => $protocol,
                        nacltask_if_exists => "purge",
                        nacltask_verify    => $verify,
                    );
                    my $mountpt = $Mount_pt->mount_point();
                    $Log->comment("Mount point is $mountpt");
                    push @$nfs_mounts, $mountpt;
                    printf Dumper($nfs_mounts);
                }
            }
        }
    }
    return $nfs_mounts;

}

sub _getIP {

    my $interface_obj = shift(@_);

    $Log->comment("ISAIKA>> IP = $interface_obj->state()->{address}");
    return $interface_obj->state()->{address};

}

sub nfsCleanup {
    $Log->comment("\n NFS Cleanup");
    my $self    = shift(@_);
    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[1];
    $Log->comment(" Destroying NFS mounts ...");
    $Log->comment(" Find the volume that is mounted ...");
    printf Dumper($self->{vservers});
    my $vservers  = $self->{vservers};
    my $vs = shift(@$vservers);
    my $vols = $self->{$vs}->{vols} ;
    #my @vol = NACL::STask::Volume->find(
    #    command_interface => $nodeobj,
    #    filter            => { vserver => $vs,
    #			       'junction-active' => "true"},
    #);
    $Log->comment(" Find & get the mount point ...");

    foreach my $vol (@$vols) {
        my $mount_obj = NACL::STask::Client::Mount->create(
            command_interface => $self->{client},
            volume            => $vol,
        );
        my $mpt                = $mount_obj->mount_point();
        my @logical_interfaces = ("ipv4");
        if ( $self->{enable_v6} ) {
            @logical_interfaces = ( "ipv4", "ipv6" );
        }

        foreach my $logical_interface (@logical_interfaces) {
            $Log->comment(
                " Destroy the $logical_interface mounts on client ...");
            my $mpt1  = "$mpt" . "_tcp_nfsv3_" . "$logical_interface";
            my $mobj1 = NACL::STask::Client::Mount->purge(
                command_interface => $self->{client},
                mount_point       => $mpt1,
                nacltask_verify   => 1,
            );
            my $mpt2  = "$mpt" . "_udp_nfsv3_" . "$logical_interface";
            my $mobj2 = NACL::STask::Client::Mount->purge(
                command_interface => $self->{client},
                mount_point       => $mpt2,
                nacltask_verify   => 1,
            );
            my $mpt3  = "$mpt" . "_tcp_nfsv4_" . "$logical_interface";
            my $mobj3 = NACL::STask::Client::Mount->purge(
                command_interface => $self->{client},
                mount_point       => $mpt3,
                nacltask_verify   => 1,
            );
            my $mpt4  = $mpt;
            my $mobj4 = NACL::STask::Client::Mount->purge(
                command_interface => $self->{client},
                mount_point       => $mpt4,
                nacltask_verify   => 1,
            );
        }
    }
}

sub cifsSetup {

    $Log->comment("\n CIFS Setup");
    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        vserver => {
            type     => SCALAR,
            required => 1
        },
        action => {
            type     => SCALAR,
            required => 0,
            default  => "reuse"
        },
        volume => {
            type     => ARRAYREF,
            required => 1
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }

    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[0];
    my $node    = $nodeobj->node();

    my $Task_obj;
    my @cifs_mounts_obj;

    my $cifs_pattern= Tharn::param('CIFS_DOMAIN') . '\\' . Tharn::param('CIFS_SUPERUSERS');

    try {
        my $cifs_obj = NACL::MTask::CIFS->setup(
            'command_interface' => $nodeobj,
            'vserver'           => $args{vserver},
            'if_exists'         => $args{action},
            'cifs_domain'       => $self->{cifs_domain},
            'cifs-server'       => $self->{cifs_server},
            'username'          => $self->{cifs_username},
            'password'          => $self->{cifs_password},
            'dns_domain'        => $self->{dns_domain},
            #'name-servers'      => [ '10.227.64.22', '10.227.64.21' ],
            'name-servers'      => [ $self->{name_servers} ],
            'policyname'        => 'test_policy',
            'unix_groups'       => [
                { 'name' => 'nobody', 'id' => 65534, },
                { 'name' => 'daemon', 'id' => 1, },
                { 'name' => 'root',   'id' => 0, }
            ],
            'unix_users' => [
                { 'user' => 'nobody', 'id' => 65534, 'primary-gid' => 65534 },
                { 'user' => 'root',   'id' => 0,     'primary-gid' => 0 },
            ],
            'vserver_name_mapping' => [
                {
                    'pattern'     => $cifs_pattern,
                    'direction'   => 'win-unix',
                    'replacement' => 'root',
                    'position'    => 1,
                },
                {
                    'pattern'     => 'root',
                    'direction'   => 'unix-win',
                    'replacement' => $cifs_pattern,
                    'position'    => 1,
                },
            ],
        );
        my $cifs_options_object = NACL::C::VserverCifsOptions->modify(
            command_interface   => $nodeobj,
            vserver             => $args{vserver},
            'default-unix-user' => 'nobody',
        );

        try {
            my $pol = NACL::C::ExportPolicy->find(
                command_interface => $nodeobj,
                filter            => {
                    vserver    => $args{vserver},
                    policyname => $self->{ $args{vserver} }->{exportpolicy}
                },
            );
        }
        catch NACL::Exceptions::NoElementsFound with {
            $Log->comment("Policy not found .. creating");

            my $newpolicy = NACL::C::ExportPolicy->create(
                command_interface => $nodeobj,
                vserver           => $args{vserver},
                ,
                policyname => $self->{ $args{vserver} }->{exportpolicy},
            );

            my $pol = NACL::C::VserverExportPolicyRule->create(
                command_interface => $nodeobj,
                vserver           => $args{vserver},
                policyname        => $self->{ $args{vserver} }->{exportpolicy},
                clientmatch       => "0.0.0.0/0",
                superuser         => ["any"],
                protocol          => ["any"],
                rorule            => ["any"],
                rwrule            => ["any"],
            );
            NACL::C::VserverExportPolicyRule->create(
                command_interface => $nodeobj,
                vserver           => $args{vserver},
                policyname        => $self->{ $args{vserver} }->{exportpolicy},
                clientmatch       => "::/0",
                superuser         => ["any"],
                protocol          => ["any"],
                rorule            => ["any"],
                rwrule            => ["any"],
            );
        };

        # Creating/identifying source volume object
        $Log->comment(' Re-use same Nfs vol for Cifs setup ...');
        my $vols = $args{volume};
        $cifs_mounts = [];
        foreach my $v (@$vols) {
            my $vol = $v->{volume};
            $v->modify( policy => $self->{ $args{vserver} }->{exportpolicy} );

            my $Volume_cifs = $v;
            $Log->comment(' Create CIFS share ...');
            my $Share_name = $vol . "_cifsshare";
            my $path = $Volume_cifs->get_one_state_attribute('junction-path');
            $path =~ s/\(//g;
            $path =~ s/\)//g;
            my $cifs_share_object = NACL::STask::VserverCifsShare->create(
                command_interface  => $nodeobj,
                path               => $path,
                vserver            => $args{vserver},
                'share-name'       => $Share_name,
                nacltask_verify    => 1,
                nacltask_if_exists => $args{action}
            );
            my @map_lifip;
            push (@map_lifip,$self->{ $args{vserver} }->{lif_v4}->state()->{'address'});
            if ( $self->{enable_v6} ) {
            push (@map_lifip,$self->{ $args{vserver} }->{lif_v6}->state()->{'address'});
            }
            foreach my $lif_ip (@map_lifip) {
                $Log->comment(" Map the vol across IP : $lif_ip");
                my $Mount_cifs = NACL::STask::Client::Mount->create(
                    command_interface => $self->{win_client},
                    volume            => $Volume_cifs
                );
                $Log->comment(
                    " Mount point is  " . $Mount_cifs->mount_point() );

                $Mount_cifs = NACL::STask::Client::Mount->create(
                    command_interface  => $self->{win_client},
                    volume             => $Volume_cifs,
                    mount_point        => $Mount_cifs->mount_point(),
                    server             => $lif_ip,
                    nacltask_verify    => 1,
                    nacltask_if_exists => 'purge'
                );
                push @cifs_mounts_obj, $Mount_cifs;
                my $mountpt = $Mount_cifs->mount_point();
                ## Change the mountpt prior to push to enable CIFS traffic to run on the mount point
                my $mod_junc_path =
                  $Volume_cifs->get_one_state_attribute('junction-path');
                $mod_junc_path =~ s/\///g;
                $mountpt = "$mountpt\\$mod_junc_path";
                push @$cifs_mounts, $mountpt;
            }


            printf Dumper($cifs_mounts);
            $self->{cifs_mounts} = \@cifs_mounts_obj;
        }
    }
    catch NATE::BaseException with {
        my $exception = shift;
        $Log->error( ' Failed due to a ' . ref $exception );
        $Log->error( ' Exception text is : ' . $exception->text );
        return;
    };

    return $cifs_mounts;
}

sub iscsiSetup {
}

sub iscsiCleanup {
}
sub enableSFO {

    my ( $self, %args ) = @_;
    my @input_params = (%args);
    my $nodes = $self->{nodes};

    $Log->comment("Enable SFO...");
    NACL::STask::StorageFailover->enable(
        command_interface => $self->{nodes}->[0],
        node              => $self->{nodes}->[0],
    );
    $Log->comment("Enable HA...");
    NACL::STask::StorageFailover->modify(
        command_interface => $self->{nodes}->[0],
        'auto-giveback'   => "false",
        ha_enable         => 1,
        nacltask_verify   => 1,
    );

}

sub sfo {

    my ( $self, %args ) = @_;
    my @input_params = (%args);
    my $nodes = $self->{nodes};

    #validate_params is the arguments to validate
    my %validate_params = (
        method => {
            type     => SCALAR,
            required => 1
        },
        ofNode => {
            type     => HASHREF,
            required => 1
        },
        byNode => {
            type     => HASHREF,
            required => 1
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }
    # We always want to connect to the live Node in case of 4x node cluster.
    my $node = $args{byNode};
    my $partner = $args{ofNode};

    $Log->comment("Setting AUTOBOOT to true prior to takeover");
    my $Api_Set_Obj_dbladecli = NACL::APISet->new(
                 hostobj   => $node->{hostrec},
                 category  => "Node",
                 interface => "CLI",
                 set       => "Nodescope"
              );

    $Api_Set_Obj_dbladecli->bootargs_set(arg   => 'AUTOBOOT',value => 'true');


    $Log->comment("Trigger $args{method}...");

    if ( $args{method} eq "takeover" ) {
        NACL::STask::StorageFailover->takeover(
            command_interface            => $node,
            node                         => $node,
            partner                      => $partner,
            takeovertype                 => "takeover_command",
            nacltask_verify              => 1,
            nacltask_expect_takeover     => 1,
            nacltask_check_partner_state => 1,
        );
        $Log->comment("SFO: takeover is complete!");

    }
    else {
        $Log->comment("\n SFO Giveback is triggered ...");
        NACL::STask::StorageFailover->giveback(
            command_interface          => $node,
            node                       => $node,
            partner                    => $partner,
            'require-partner-waiting'  => "true",
            'override-vetoes'          => "true",
            'only-cfo-aggregates'      => "false",
            wait_for_partner           => 1,
            auto_giveback_expected     => 0,
            'wait-for-asup-generation' => 0,
            'failure-reason'           => 'giveback_failed_due_to_failed_disk',
        );
        $Log->comment(" Wait for few seconds ...");
        NACL::STask::StorageFailover->wait_for_state(
            command_interface => $node,
            node              => $node,
            partner           => $partner,
            node_state        => "up",
            'method-timeout'  => 120,
        );
        $Log->comment(" Giveback completed!");
    }

}

sub cifsCleanup {

    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        vserver => {
            type     => SCALAR,
            required => 1
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }
    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[1];

    $Log->comment("\n CIFS cleanup...");
    $Log->comment(" Deleting CIFS map drive on client...");
    my $cifs_mounts = $self->{cifs_mounts};
    foreach my $mp (@$cifs_mounts) {
        NACL::STask::Client::Mount->purge(
            command_interface => $self->{win_client},
            mount_point       => $mp->{'mount_point'},
            nacltask_verify   => 1,
        );
    }
    $Log->comment(" Deleting CIFS user on the source data vserver...");

    NACL::C::VserverCifs->delete(
        'command_interface' => $nodeobj,
        'vserver'           => $args{vserver},
        'username'          => $self->{cifs_username},
        'password'          => $self->{cifs_password}
    );
}

sub enableSis {

    my ( $self, %args ) = @_;
    my @input_params = (%args);

    #validate_params is the arguments to validate
    my %validate_params = (
        vserver => {
            type     => SCALAR,
            required => 1
        },
        volumes => {
            type     => ARRAYREF,
            required => 1
        },
    );
    eval {
        (%args) =
          Params::Validate::validate( @input_params, \%validate_params );
    };
    if ($@) {
        $Log->error( "failed to validate parameter: " . $@ );
        return 0;
    }
    my $nodes   = $self->{nodes};
    my $nodeobj = $nodes->[1];
    my $vols    = $args{volumes};
    $Log->comment("\nDedupe and Compression config");
    foreach my $v (@$vols) {
        my $vol = $v->{volume};

        my $Sis = NACL::STask::VolumeEfficiency->on(
            command_interface => $nodeobj,
            vserver           => $args{vserver},
            volume            => $vol,
        );
        $Log->comment('Issue \"sis config\" on the volume');
        $Sis->config(
            'inline-compression'     => "true",
            'background-compression' => "true"
        );

        $Log->comment('Issue sis start on the volume');
        $Sis->start( nacltask_wait => '1' );
	$Sis->wait_for_status(status => "ACIVE");


    }
}

## Accessor methods

sub getNodes {

    my $self = shift(@_);
    return $self->{nodes};
}

sub getFirstNode {

    my $self = shift(@_);
    return $self->{nodes}->[0];

}

sub _random_string {
    my %args = @_;
    $args{length} ||= 8;
    $args{chars} ||= [ 'a' .. 'z', 'A' .. 'Z', '0' .. '9' ];

    my $str = '';
    foreach ( 1 .. $args{length} ) {
        $str .= $args{chars}->[ rand @{ $args{chars} } ];
    }
    return $str;
}

sub _populate_nodes {

    my $opts  = shift;
    my @nodes = NACL::C::Node->find();
    foreach $nodesobj (@nodes) {
        $node    = $nodesobj->node();
        $hostobj = host($node);
        my $aggr = $node . '_aggr1';
        $aggr =~ s/-/_/g;
        $Log->debug(
" Getting all the inforamtion of ports and ips from tharnhost file of $node"
        );
        $node_details{$node}{data_port1}      = $hostobj->{DATA_PORT1},
          $node_details{$node}{data_port2}    = $hostobj->{DATA_PORT2},
          $node_details{$node}{data_ip1}      = $hostobj->{DATA_IP1},
          $node_details{$node}{data_ip2}      = $hostobj->{DATA_IP2},
          $node_details{$node}{data_netmask1} = $hostobj->{DATA_MASK1},
          $node_details{$node}{data_netmask2} = $hostobj->{DATA_MASK2},
          $node_details{$node}{data_gateway1} = $hostobj->{DATA_GATEWAY1},
          $node_details{$node}{data_gateway2} = $hostobj->{DATA_GATEWAY2},
          $node_details{$node}{mgmt_port}     = $hostobj->{MGMT_PORT},
          $node_details{$node}{mgmt_mask}     = $hostobj->{MGMT_MASK},
          $node_details{$node}{mgmt_gateway}  = $hostobj->{MGMT_GATEWAY},
          $node_details{$node}{data_protocol} = "nfs,cifs,fcache",
          $node_details{$node}{lif_role}      = "data",
          $node_details{$node}{aggr}          = $aggr,
          $node_details{$node}{igroup}        = "igroup_" . _random_string();
        if ( $opts->{ENABLE_V6} ) {
            my $filtered_network;
            my @filter = (
                { netid => "inet6_data1" },
                { netid => "inet6_data2" },
                { netid => "inet6_data3" },
                { netid => "inet6_data4" }
            );
            my @filtered_networks =
              $nodesobj->{hostrec}->net_filter( include => \@filter );
            if ( scalar(@filtered_networks) <= 0 ) {
                $Log->warn(
"Error : No inet_6 net records found in the tharnhosts.Please update tharnhosts with inet6 records"
                );
                return;
            }
            foreach $filtered_network (@filtered_networks) {
                my $subkey = $1
                  if $filtered_network->{netid} =~ /(inet6_data[1-4])/;
                $node_details{$node}{"data_ip_$subkey"} =
                  $filtered_network->{address};
                $node_details{$node}{"data_gateway_$subkey"} =
                  $filtered_network->{gateway};
                $node_details{$node}{"data_netmask_$subkey"} =
                  $filtered_network->{netmask};
                $node_details{$node}{"data_netmask_$subkey"} =~ s/\///g;
            }
        }

    }
}

sub releaseIPs {

    # if addresses queried from DB
    $Log->comment("=== cleanup: $job_id");
    my ( $cleanupv4, $errorv4 ) =
      $DataIPV4->ip_address_return( 'job_id' => $job_id );
    my ( $cleanupv6, $errorv6 ) =
      $DataIPV6->ip_address_return( 'job_id' => $job_id );
    if   ($cleanupv4) { $Log->comment("=== V4 cleanup: $cleanupv4 addresses"); }
    else              { $Log->comment("=== cleanup error: $errorv4"); }
    if   ($cleanupv6) { $Log->comment("=== V6 cleanup: $cleanupv6 addresses"); }
    else              { $Log->comment("=== cleanup error: $errorv6"); }
    $Log->comment("- Un-Reserved IPs successfully");

    
}
1;
