# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Switch.pm#1 $
#
## @summary Switch - Provide a hardware independent switch class
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Switch - Provide a hardware independent switch class

=head1 SYNOPSIS

  use Network::Switch;

  # Switch constructor
  $switch_obj = Network::Switch->new(hostspec => $hostspec);

=head1 DESCRIPTION

Provides a hardware independent switch class.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id: //depot/prod/test/Rsierranevada/lib/Network/Switch.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Switch;

use strict;
use warnings;

use Carp qw(croak);
use Network::Utils qw(get_ifconfig_info run_ping);
use Params::Validate qw(validate SCALAR ARRAYREF OBJECT);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $VERSION = '1.0';
our $AUTOLOAD;
our @ISA;

#---------------------------------------------------------------------

=head2 new

  $switch_obj = Network::Switch->new(hostspec => $hostspec
                                     [, connspec => $connspec ]);

Creates an independent hardware switch object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostspec

(Required) : Hostspec of switch.

=item connspec

(Optional) : Connection specification for the switch.
Defaults to '#console'.

=item timeout

(Optional) : Timeout for commands to the switch.
Defaults to 120 secs.

=back

=item Returns

Switch object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            hostspec => { type => SCALAR },
            connspec => { type => SCALAR, default => '#telnet' },
            timeout  => { type => SCALAR, default => 120 },
        }
    );

    # Get connectrec object to switch
    Tharn::host($opts{hostspec});
    $opts{conn} = Tharn::connect($opts{connspec});
    $opts{conn}->timeout($opts{timeout});

    my $type = $opts{conn}->hostp->hosttype;
    if ( $type =~ /switch-cisco/ ) {
        require Network::Switch::Cisco;
        @ISA = qw(Network::Switch::Cisco);
    } else {
        croak "Switch type is not supported: $type";
    }

    return $class->SUPER::new(%opts);
}

#---------------------------------------------------------------------

=head2 get_if_port

  $port = $switch_obj->get_if_port(conn => $conn,
                                   ifname => $ifname,
                                   [, retry => $retry ]);

Get the switch port that is connected to the specified interface.

The following named arguments are accepted:

=over

=item Arguments

=over

=item conn

(Required) : Connectrec object of filer.

=item ifname

(Required) : Interface name on filer.

=item retry

(Optional) : Number of times to retry operation.
Defaults to 3.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut


#---------------------------------------------------------------------
sub get_if_port {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            conn   => { type => OBJECT },
            ifname => { type => SCALAR },
            retry  => { type => SCALAR, default => 3 },
        }
    );
    my @args = qw(conn ifname retry);

    # Get the port for interface
    my ($port) = ($self->_get_if_port_and_mac(@opts{@args}))[0];

    return $port;
}

#---------------------------------------------------------------------

=head2 verify_if_port

  $out = $switch_obj->verify_if_port(conn => $conn,
                                     ifname => $ifname,
                                     port => $port,
                                     [, retry => $retry ]);

Verify the interface on the filer is connected to the port on the switch.

The following named arguments are accepted:

=over

=item Arguments

=over

=item conn

(Required) : Connectrec object of filer.

=item ifname

(Required) : Interface name on filer.

=item port

(Required) : Port on switch.

=item retry

(Optional) : Number of times to retry operation.
Defaults to 3.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub verify_if_port {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            conn   => { type => OBJECT },
            ifname => { type => SCALAR },
            port   => { type => SCALAR },
            retry  => { type => SCALAR, default => 3 },
        }
    );
    my @args = qw(conn ifname retry);

    # Get the port and mac address for interface
    my ($port, $mac_addr) = $self->_get_if_port_and_mac(@opts{@args});

    # Verify given port is attached to the interface
    if ( $port ne $opts{port} ) {
        croak "MAC address $mac_addr on interface $opts{ifname} " .
              "is attached to port $port instead of $opts{port}";
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 save_port_settings

  $out = $switch_obj->save_port_settings(port => $port | \@port_list);

Save settings for all ports specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : Port(s) to save information.
This could be a single port (scalar) or a list of ports (array reference).

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub save_port_settings {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => SCALAR|ARRAYREF },
        }
    );
    my ($port_h, $port, $key, $port_gvrp);
    my %extra_h = (
        jumbo_mtu => 0,
    );

    if ( !ref $opts{port} ) {
        # Change the single port to an array
        $opts{port} = [ $opts{port} ];
    }

    if ( !defined $self->{_save_gvrp} ) {
        # Get the GVRP configuration for all ports
        $self->{_save_gvrp} = $self->show_gvrp_conf;
    }

    foreach $port ( @{$opts{port}} ) {
        if ( !defined $self->{_save_ports}{$port} ) {
            $port_gvrp = $self->{_save_gvrp}{ports}{$port};
            # Get the port settings and save it
            $port_h = $self->show_port(port => $port, jumbo => 1);
            foreach $key ( keys %extra_h ) {
                $port_h->{$key} = $extra_h{$key} if !exists $port_h->{$key};
            }
            foreach $key ( keys %$port_gvrp ) {
                $port_h->{"gvrp_$key"} = $port_gvrp->{$key};
            }
            $self->{_save_ports}{$port} = $port_h;
        }
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 restore_port_settings

  $out = $switch_obj->restore_port_settings();

Restore settings for all ports saved with save_port_settings.
Restore settings one port at a time, it does not fail on first
failure to allow the setting of as many ports as possible, the
errors are collected and croak at the end.

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub restore_port_settings {
    my $self = shift;
    my (%args, $port_h, $port, $key, $value, $gvrp_h, $check, @port_list);
    my ($mod_ch_proto, $module);
    my $err = '';
    # Options for set_port in the order in which they will be set
    my @set_port_list = qw(
        speed
        duplex
        flowctl_snd
        flowctl_rec
        channel_mode
        jumbo
        enable
    );
    # Options given by show_port/show_gvrp_conf which need
    # to be translated to correct set_port/set_gvrp options
    my @list = qw(
        speed_mode
        duplex_mode
        jumbo_mtu
        status
        gvrp_status
    );
    my %check = map { $_ => 1 } @set_port_list, @list;

    # Get the GVRP configuration for all ports
    $gvrp_h = $self->show_gvrp_conf;

    # Get the module channel protocols
    $mod_ch_proto = $self->show_channel_protocol;

    # Restore settings one port at a time, it does not fail on first
    # failure to allow the setting of as many ports as possible, the
    # errors are collected and croak at the end
    foreach $port ( keys %{$self->{_save_ports}} ) {
        eval {
            # Get module for port
            $module = $self->get_module_from_port(port => $port);
            # Get current port settings
            $port_h = $self->show_port(port => $port, jumbo => 1);
            foreach $key ( keys %{$gvrp_h->{ports}{$port}} ) {
                $port_h->{"gvrp_$key"} = $gvrp_h->{ports}{$port}{$key};
            }

            %args = ();
            # Get all the settings which were modified
            foreach $key ( keys %{$self->{_save_ports}{$port}} ) {
                # Skip keys not listed in check hash
                next if !$check{$key};
                # Compare current setting with saved setting
                if ( ! exists $port_h->{$key}
                  || $port_h->{$key} ne $self->{_save_ports}{$port}{$key} ) {
                    $args{$key} = $self->{_save_ports}{$port}{$key};
                }
            }

            # Set the correct speed setting
            if ( defined $args{speed_mode} || defined $args{speed} ) {
                $args{speed} = $self->{_save_ports}{$port}{speed_mode};
            }

            # Set the correct duplex setting
            if ( defined $args{duplex_mode} || defined $args{duplex} ) {
                $args{duplex} = $self->{_save_ports}{$port}{duplex_mode};
                # No need to set the duplex when speed will be set to 'auto'
                delete $args{duplex}
                  if (defined $args{speed} && $args{speed} =~ /auto/);
            }

            # Set the correct jumbo setting
            if ( defined $args{jumbo_mtu} ) {
                $args{jumbo} = ($args{jumbo_mtu} ? 1 : 0);
            }

            if ( defined $args{status} ) {
                # Change in status -- most likely the port was enabled/disabled
                $args{enable} = ($args{status} eq 'connected' ? 1 : 0);
            }

            # Reset all settings to their original values
            # in the order specified by @set_port_list
            @port_list = grep { exists $args{$_} } @set_port_list;

            if ( @port_list && $port_list[-1] eq 'channel_mode'
              && $mod_ch_proto->{$module}{protocol} eq 'LACP' ) {
                # Make sure 'channel_mode' is not the last setting to be reset:
                # Set the jumbo frames for port to its current value after
                # restoring the channel_mode to make sure to wait for port
                # to become connected or disabled
                $args{jumbo} = ($port_h->{jumbo_mtu} ? 1 : 0);
                push @port_list, 'jumbo';
            }

            foreach $key ( @port_list ) {
                $check = 0;
                if ( $key eq $port_list[-1]
                  && ($port_h->{status} ne 'disabled' || $key eq 'enable') ) {
                    $check = 1;
                }
                eval {
                    if ( $key eq 'channel_mode'
                      && $mod_ch_proto->{$module}{protocol} eq 'LACP' ) {
                        # Port is in a LACP module
                        $self->set_port_lacp_channel(
                            port => [$port],
                            mode => $args{$key},
                        );
                    } else {
                        # Port is in a PAGP module
                        $self->set_port(
                            port => $port,
                            $key => $args{$key},
                            check_status => $check,
                            curr_status  => $port_h->{status},
                        );
                    }
                };
                $err .= "$@\n" if $@;
            }

            if ( defined $args{gvrp_status} ) {
                $value = ($args{gvrp_status} eq 'Enabled' ? 1 : 0);
                eval {$self->set_gvrp(port => $port, enable => $value)};
                $err .= "$@\n" if $@;
            }
        };
        $err .= "$@\n" if $@;
    }

    croak $err if $err;

    return 1;
}

sub clear_port_counter{
        my $self = shift;
        my (%opts) = validate (
            @_,
             {
                port => {type => SCALAR },
             }
         );

        my ($match,$endtime,$out);

        ($match) = $self->say("clear counter $opts{poprt}",'-noprompt','-wait',
                        quotemeta('Do you want to continue (y/n) [n]?'));

        $out = $self->_say("y");
        croak $out if $out !~ /MAC and port counters cleared/;

        return 1;
}





#---------------------------------------------------------------------
# Internal method
# Get the port and mac address of the given network interface
#---------------------------------------------------------------------
sub _get_if_port_and_mac {
    my ($self, $conn, $ifname, $retry) = @_;
    my ($port, $ifinfo, $mac_addr, $ip_addr);

    # Get interface information
    $ifinfo = get_ifconfig_info($conn, $ifname);

    # Get the MAC address of interface
    $mac_addr = $ifinfo->{$ifname}{mac_addr};

    # Get IP address to ping -- this is the IP address of the interface
    # with the last octet replaced by '5'
    $ip_addr = $ifinfo->{$ifname}{ip_addr};
    $ip_addr =~ s/\.\d+$/\.235/;

    while ( !defined $port && $retry-- > 0 ) {
        eval {
            # Ping the interface from the filer to have the MAC address
            # cached in the CAM
            run_ping($conn, $ip_addr, count => 1);

            # Get the port attached to the MAC address
            $port = $self->get_port_from_mac(mac => $mac_addr);
        };
        sleep(5) if (!defined $port);
    }
    croak $@ if $@;
    croak 'Could not determine host switch port' if (!defined $port);

    return ($port, $mac_addr);
}

#---------------------------------------------------------------------
# AUTOLOAD to exposed all keys in the blessed object as read-only methods
# Do not expose private keys (the ones which start with a '_')
#---------------------------------------------------------------------
sub AUTOLOAD {
    my $self = shift;
    my $method = $AUTOLOAD;
    $method =~ s/.*:://;
    if ( $method =~ /^_/ || !exists $self->{$method} ) {
        croak qq(Can't locate object method "$method" via package "Network::Switch");
    }
    return $self->{$method};
}

#---------------------------------------------------------------------
# Destructor
#---------------------------------------------------------------------
sub DESTROY {
    my $self = shift;

    # Call destructor of base class if it exists
    if ( $self->can('SUPER::DESTROY') ) {
        $self->SUPER::DESTROY;
    }
}

1;

