# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Switch/Cisco.pm#1 $
#
## @summary Cisco - Provide a class for the cisco switch
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Cisco - Provide a class for the cisco switch

=head1 SYNOPSIS

  use Network::Switch::Cisco;

  # Switch constructor
  $switch_obj = Network::Switch->new(hostspec => $hostspec);

=head1 DESCRIPTION

Provides a class for the cisco switch.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id: //depot/prod/test/Rsierranevada/lib/Network/Switch/Cisco.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Switch::Cisco;

use strict;
use warnings;

use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $VERSION = '1.0';

# Global key map
my %key_map_h = (
    # Key map for show_port
    status           => 'Status',
    speed            => 'Speed',
    duplex           => 'Duplex',
    type             => 'Type',
    channel_mode     => 'Channel Mode',
    admin_group      => 'Admin Group',
    channel_id       => 'ChId',
    port_id          => 'Port-ID',
    flowctl_snd      => 'Send Flowadmin',
    flowctl_rec      => 'Receive Fladmin',
    flowctl_snd_oper => 'Control oper',
    flowctl_rec_oper => 'owControl oper',
    rx_pause         => 'RxPause',
    tx_pause         => 'TxPause',
    total_drop       => 'Total-Drop',
    action           => 'Action',
    align_err        => 'Align-Err',
    fcs_err          => 'FCS-Err',
    xmit_err         => 'Xmit-Err',
    rcv_err          => 'Rcv-Err',
    undersize        => 'UnderSize',
    # Key map for show_vlan
    ports            => 'Mod/Ports, Vlans',
    parent           => 'Parent',
    SAID             => 'SAID',
    trans1           => 'Trans1',
    trans2           => 'Trans2',
    brdg_no          => 'BrdgNo',
    brdg_mode        => 'BrdgMode',
    ring_no          => 'RingNo',
    dyn_created      => 'DynCreated',
    if_index         => 'IfIndex',
    mtu              => 'MTU',
    RSPAN            => 'RSPAN',
    name             => 'Name',
    MISTP_inst       => 'MISTP-Inst',
    stp              => 'Stp',
);

# Units conversion
my %units_h = (
    G => 1000
);

# SYSLOG date format
my $SYSLOG_DATE = qr/\d+ (?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+\d{1,2} \d\d\:\d\d\:\d\d [A-Z]{3}/;

#---------------------------------------------------------------------

=head2 new

  $switch_obj = Network::Switch::Cisco->new(hostspec => $hostspec
                                            [, connspec => $connspec ]);

Creates a cisco switch object. This is an abstract class, so do not use
this constructor directly, use Network::Switch->new() instead.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new {
    my $class = shift;
    return bless { @_ }, $class;
}

#---------------------------------------------------------------------

=head2 get_port_from_mac

  $port = $switch_obj->get_port_from_mac(mac => $mac_addr);

Get the port attached to the MAC address specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mac

(Required) : MAC address

=back

=item Returns

The port attached to the MAC address specified.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_port_from_mac {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            mac => { type => SCALAR },
        }
    );
    my ($ret, $out, $port);

    # Send command to switch
    $out = $self->_say("show cam $opts{mac}");

    # Remove extra information
    $out =~ s/\* = Static Entry\. \+ = Permanent Entry\. # = System Entry\. R = Router Entry\.//;
    $out =~ s/X = Port Security Entry \$ = Dot1x Security Entry M = Mac-Auth-Bypass Entry//;
    if ( $out =~ s/Total Matching CAM Entries Displayed = (\d+)// ) {
        croak "Could not find the port attached to MAC address: $opts{mac}"
          if $1 == 0;
    } else {
        croak $out;
    }

    # Process the output by converting it to a hash, the key will be the VLAN
    # and its value is a hash with the port information
    $ret = _process_output(\$out);

    # Just get the first hash
    $ret = (values %$ret)[0];

     # Get the port only
    ($port) = ($ret->{'Destination Ports orVCs / [Protocol Type]'} =~ m{(\d+/\d+)});

    return $port;
}

#---------------------------------------------------------------------

=head2 get_module_from_port

  $module = $switch_obj->get_module_from_port(port => $port);

Get the module the specified port belongs to.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : Port (given by module#/port#, e.g., 3/32) to get module for.

=back

=item Returns

The module the port belongs to.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_module_from_port {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => SCALAR },
        }
    );
    my ($module) = ($opts{port} =~ m{^(\d+)/\d+$});

    return $module;
}

#---------------------------------------------------------------------

=head2 show_module

  $module_hash = $switch_obj->show_module([ module => $module|\@module_list ]);

Get information about the modules in the switch.

The following named arguments are accepted:

=over

=item Arguments

=over

=item module

(Optional) : Module number to get information for. If this argument is a list,
the information for all modules specified is returned.
Defaults to get the information about all modules in the switch.

=back

=item Returns

When argument is a scalar (a single module) a hash reference of the module
information is returned. Otherwise, if argument is not specified or
is a list of modules, a hash reference having as a key the module
and the value is a hash reference to the information for each module.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_module {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            module => { type => SCALAR|ARRAYREF, default => '' },
        }
    );
    my ($module, $ret, $out);
    my %key_map = (
        status     => 'Status',
        slot       => 'Slot',
        model      => 'Model',
        serial     => 'Serial-Num',
        type       => 'Module-Type',
        hw         => 'Hw',
        sw         => 'Sw',
        fw         => 'Fw',
        sub_model  => 'Sub-Model',
        sub_serial => 'Sub-Serial',
        sub_type   => 'Sub-Type',
        sub_hw     => 'Sub-Hw',
        sub_sw     => 'Sub-Sw',
        name       => 'Module-Name',
        ports      => 'Ports',
        mac_addrs  => 'MAC-Address(es)',
    );

    if ( ref $opts{module} eq 'ARRAY' ) {
        # Module argument is an array, so get the module info for all
        # modules specified in the list
        foreach $module ( @{$opts{module}} ) {
            $ret->{$module} = $self->show_module(%opts, module => $module);
        }
        return $ret;
    }

    # Send command to switch
    $out = $self->_say("show module $opts{module}");

    # Process the output by converting it to a hash
    $ret = _convert_hash(_process_output(\$out), \%key_map);

    if ( $opts{module} ne '' ) {
        # Return just the module info if module is specified
        $ret = $ret->{$opts{module}};
    }

    return $ret;
}

#---------------------------------------------------------------------

=head2 reset_module

  $out = $switch_obj->reset_module(module => $module
                                   [ timeout => $timeout ]);

Reset the specified module.

The following named arguments are accepted:

=over

=item Arguments

=over

=item module

(Required) : Module number to reset.

=item timeout

(Optional) : Timeout waiting for command to complete.
Defaults to 300.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub reset_module {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            module  => { type => SCALAR },
            timeout => { type => SCALAR, default => 300 },
        }
    );
    my ($match, $endtime, $out);

    # Send command to switch
    ($match) = $self->_say("reset $opts{module}", '-noprompt',
      '-wait', quotemeta('Do you want to continue (y/n) [n]? '));

    # Confirm resetting module
    $out = $self->_say("y");
    croak $out if $out !~ /Resetting module/;

    # Wait until module is reset
    $endtime = time + $opts{timeout};
    do {
        sleep 10;
        $out = $self->show_module(module => $opts{module});
    } while ($endtime > time && $out->{status} ne 'ok');
    croak "Timeout waiting for module to reset"
      if $out->{status} ne 'ok';

    return 1;
}

#---------------------------------------------------------------------

=head2 show_port

  $port_hash = $switch_obj->show_port([ port => $port|\@port_list ]
                                      [, jumbo => 0|1 ]);

Get information about the port(s) in the switch.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Optional) : Port (given by module#/port#, e.g., 3/32) to get information for.
If this argument is a list, the information for all ports specified is returned.
Defaults to get the information about all ports in the switch.

=back

=item Returns

When argument is a scalar (a single port) a hash reference of the port
information is returned. Otherwise, if argument is not specified or
is a list of ports, a hash reference having as a key the port
and the value is a hash reference to the information for each port.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_port {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port  => { type => SCALAR|ARRAYREF, default => '' },
            jumbo => { type => SCALAR, default => 0 },
        }
    );
    my ($out, $ret, $port, $type, $value, $units, $auto, $jumbo);

    if ( ref $opts{port} eq 'ARRAY' ) {
        # Port argument is an array, so get the port info for all
        # ports specified in the list
        foreach $port ( @{$opts{port}} ) {
            $ret->{$port} = $self->show_port(%opts, port => $port);
        }
        return $ret;
    }

    # Send command to switch
    $out = $self->_say("show port $opts{port}");

    # Remove extra information
    $out =~ s/\* = Configured MAC Address//;

    # Process the output by converting it to a hash
    $ret = _convert_hash(_process_output(\$out));

    if ( $opts{jumbo} ) {
        $jumbo = $self->show_port_jumbo;
    }

    # Translate speed and duplex information
    foreach $port ( keys %$ret ) {
        if ( defined $ret->{$port}{speed} ) {
            # Process speed value
            ($type, $value, $units) = ($ret->{$port}{speed} =~ /^((?:(?:\S+\s+)?a-)?)(\d+|auto.*)(.?)/);
            $value *= $units_h{$units} if exists $units_h{$units};
            $ret->{$port}{speed} = $value;
            $ret->{$port}{speed_mode} = ($type =~ /(\S+)?\s*a-/ ?
                                        (defined $1 ? $1 : 'auto') :
                                        ($value eq 'auto' ? 'auto' : $value));

            # Process duplex value
            ($type, $value) = ($ret->{$port}{duplex} =~ /^((?:a-)?)(\w+)/);
            $ret->{$port}{duplex} = $value;
            $ret->{$port}{duplex_mode} = ($type eq 'a-' || $value eq 'auto' ? 'auto' : $value);
        }

        # Add the jumbo frame MTU if jumbo frames are enabled on the port
        if ( $opts{jumbo} && defined $jumbo->{$port} ) {
            $ret->{$port}{jumbo_mtu} = $jumbo->{$port}{jumbo_mtu};
        }
    }

    if ( $opts{port} ne '' ) {
        # Return just the port info if port is specified
        $ret = $ret->{$opts{port}};
    }

    return $ret;
}

#---------------------------------------------------------------------

=head2 show_port_jumbo

  $port_hash = $switch_obj->show_port_jumbo([ port => $port ]);

Get the MTU of the ports enabled for jumbo frames.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Optional) : Port (given by module#/port#, e.g., 3/32) to get information for.
Defaults to get the information about all ports in the switch.

=back

=item Returns

Hash reference having as a key the port and the value is a hash reference
having the MTU if the port is enabled for jumbo frames

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_port_jumbo {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => SCALAR, default => '' },
        }
    );
    my (%ret, $mod, $port, $start, $end, $mtu, $out);

    # Send command to switch
    $out = $self->_say("show port jumbo");

    # Get the MTU size and all the ports enabled for jumbo frames
    if ( $out !~ /Jumbo frames MTU size is (\d+) bytes.\nJumbo frames enabled on port\(s\) (.*)\./ ) {
        croak $out;
    }

    $mtu = $1;
    # Process each port range at a time
    foreach ( split /,/, $2 ) {
        # Get the module and the range
        ($mod, $start, $end) = m{(\d+)/(\d+)(?:-(\d+))?};
        $end = $start if !defined $end;
        foreach $port ( $start..$end ) {
            # For each port enabled for jumbo frames, return the MTU
            $ret{"$mod/$port"} = { jumbo_mtu => $mtu };
        }
    }

    if ( $opts{port} ) {
       # Set the return hash to include only the specified port
       $port = $opts{port};
       %ret = (exists $ret{$port} ? ( $port => $ret{$port} ) : ());
    }

    return \%ret;
}

#---------------------------------------------------------------------

=head2 set_port

  $out = $switch_obj->set_port(port => $port
                               [, speed => $speed ]
                               [, duplex => 'full'|'half' ]
                               [, flowctl_rec => 'on'|'off'|'desired' ]
                               [, flowctl_snd => 'on'|'off'|'desired' ]
                               [, channel_mode => 'on'|'off' ]
                               [, jumbo => 0|1 ]
                               [, enable => 0|1 ]
                               [, timeout => $timeout ]
                               [, check_status => 0|1 ]
                               [, curr_status => $status ]);

Set port attributes in the switch.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : Port to set the attributes for.

=item speed

(Optional) : Set port transmission speed.

=item duplex

(Optional) : Set port transmission type.

=item flowctl_rec

(Optional) : Set port traffic flowcontrol for receive.

=item flowctl_snd

(Optional) : Set port traffic flowcontrol for send.

=item channel_mode

(Optional) : Set port channel.

=item jumbo

(Optional) : Enable/disable jumbo frames.

=item enable

(Optional) : Enable/disable port.

=item timeout

(Optional) : Timeout waiting for command to complete.
Defaults to the connection's default timeout.

=item check_status

(Optional) : Wait for port to become connected or disabled
based on the initial status of the port and which options are set.
Defaults to 1.

=item curr_status

(Optional) : Wait for this port status instead of querying the switch
before setting the options.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_port {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port         => { type => SCALAR },
            speed        => { type => SCALAR, optional => 1 },
            duplex       => { type => SCALAR, optional => 1 },
            flowctl_rec  => { type => SCALAR, optional => 1 },
            flowctl_snd  => { type => SCALAR, optional => 1 },
            channel_mode => { type => SCALAR, optional => 1 },
            jumbo        => { type => SCALAR, optional => 1 },
            enable       => { type => SCALAR, optional => 1 },
            timeout      => { type => SCALAR, optional => 1 },
            check_status => { type => SCALAR, default => 1 },
            curr_status  => { type => SCALAR, optional => 1 },
        }
    );
    my ($out, $key, $type, $match, %args, $endtime, $enable);
    my $port = $opts{port};
    my $wait_time = 0; # Initial wait time
    my %flowctl = (
        flowctl_rec => 'receive',
        flowctl_snd => 'send',
    );

    if ( $opts{check_status} ) {
        if ( defined $opts{curr_status} ) {
            # Set the expected value of the status to the one provided
            $enable = $opts{curr_status};
        } else {
            # Based on the status of the port set the expected value of the
            # status after changing the switch parameter.
            $out = $self->show_port(port => $port);
            $enable = $out->{status};
        }
    }

    if ( defined $opts{timeout} ) {
        # Overwrite connection's default timeout on the say command
        %args = ( '-timeout', $opts{timeout} );
    } else {
        # Get default timeout
        $opts{timeout} = $self->{conn}->timeout;
    }

    if ( defined $opts{speed} ) {
        $wait_time = 10;
        $out = $self->_say("set port speed $port $opts{speed}", %args);
        if ( $out =~ m{Usage: set port speed <mod/port> (<.*>)} ) {
            croak "Port $port does not support speed $opts{speed}, " .
                  "valid speed values are $1";
        }
        croak $out if $out !~ /Port\(s\)\s+\S+\s+speed set to/;
    }

    if ( defined $opts{duplex} ) {
        $out = $self->_say("set port duplex $port $opts{duplex}", %args);
        croak $out if $out !~ /(Port\(s\)\s+\S+\s+set to|cannot be changed)/;
    }

    foreach $key ( keys %flowctl ) {
        if ( defined $opts{$key} ) {
            $wait_time = 5;
            $type = $flowctl{$key};
            $out = $self->_say("set port flowcontrol $port $type $opts{$key}", %args);
            croak $out if $out !~ /Port\s+\S+\s+flow control $type administration status set to/;
        }
    }

    if ( defined $opts{channel_mode} ) {
        $out = $self->_say("set port channel $port mode $opts{channel_mode}", %args);
        croak $out if $out !~ /Port\(s\)\s+\S+\s+channel mode set to/;
    }

    if ( defined $opts{jumbo} ) {
        $type = $opts{jumbo} ? 'enable' : 'disable';
        $out = $self->_say("set port jumbo $port $type", %args);
        croak $out if $out !~ /Jumbo frames ${type}d on port/;
    }

    if ( defined $opts{enable} ) {
        if ( $opts{enable} ) {
            # Enable port
            $type = 'enable';
            $out = $self->_say("set port $type $port", %args);
            $enable = 'connected';
            $wait_time = 10;
        } else {
            # Disable port
            $type = 'disable';
            ($match) = $self->_say("set port $type $port", '-noprompt',
              '-wait', quotemeta('Do you want to continue (y/n) [n]? '), %args);
            # Confirm disabling port
            $out = $self->_say("y");
            $enable = 'disabled';
            $wait_time = 5;
        }
        croak $out if $out !~ /Port\s+\S+\s+(en|dis)abled/;
    }

    if ( $opts{check_status} ) {
        # Wait until port is connected or disabled
        $endtime = time + $opts{timeout};
        do {
            sleep $wait_time;
            $out = $self->show_port(port => $port);
            $wait_time = 10;
        } while ($endtime > time && $out->{status} ne $enable);
        croak "Timeout waiting for port to become $enable"
          if $out->{status} ne $enable;
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 show_port_channel

  $channel_hash = $switch_obj->show_port_channel([ port => $port ]);

Get information on the channel the given port belongs to.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Optional) : Port (given by module#/port#, e.g., 3/32) to get channel
information for.
Defaults to get the information about all etherchannels in the switch.

=back

=item Returns

Hash reference having as a key the channel id and the value is a hash reference
having the channel information for all ports belonging to the channel id.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_port_channel {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => SCALAR, default => '' },
        }
    );
    my (%ret, $info, $port, $out);

    # Send command to switch
    $out = $self->_say("show port channel $opts{port}");
    return if $out =~ /No ports channell?ing/;

    # Check for correct output
    if ( $out !~ /Port\s+Status\s+Channel\s+/ ) {
        croak $out;
    }

    # Process the output by converting it to a hash
    $info = _convert_hash(_process_output(\$out));

    # Group output by the channel_id
    foreach $port ( keys %$info ) {
        $ret{$info->{$port}{channel_id}}{$port} = { %{$info->{$port}} };
    }

    return \%ret;
}

#---------------------------------------------------------------------

=head2 set_port_channel

  $out = $switch_obj->set_port_channel(port => \@port_list
                                       mode => $mode);

Create an etherchannel with all ports specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : List of ports to include in the etherchannel.

=item mode

(Required) : Etherchannel mode ('on', 'off', 'desirable' or 'auto').

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_port_channel {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => ARRAYREF },
            mode => { type => SCALAR },
        }
    );
    my $out;
    my $ports = join(',', @{$opts{port}});

    # Send command to switch
    $out = $self->_say("set port channel $ports mode $opts{mode}");

    # Check for correct output
    if ( $out !~ /Port\(s\).*channel mode set to/ ) {
        croak $out;
    }

    return 1;
}


#---------------------------------------------------------------------
#The subroutine sets an Admin group to the switch ports.

sub set_port_channel_admin {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => ARRAYREF },
        }
    );
    my $out;
    my $ports = join(',', @{$opts{port}});

    # Send command to switch
    $out = $self->_say("set port channel $ports" );

    # Check for correct output
    if ( $out !~ /admin group/g ) {
        croak $out;
    }

    return 1;
}

#---------------------------------------------------------------------


#---------------------------------------------------------------------

=head2 show_channel_protocol

  $module_hash = $switch_obj->show_channel_protocol();

Get information on channel protocol by module.

=over

=item Returns

Hash reference having as a key the module and the value is a hash reference
having the channel protocol information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_channel_protocol {
    my $self = shift;
    my $out;
    my %key_map = (
        protocol => 'ChannelProtocol',
    );

    # Send command to switch
    $out = $self->_say("show channelprotocol");

    # Check for correct output
    if ( $out !~ /Module\s+Protocol/ ) {
        croak $out;
    }

    # Process the output by converting it to a hash
    return _convert_hash(_process_output(\$out), \%key_map);
}

#---------------------------------------------------------------------

=head2 set_channel_protocol

  $out = $switch_obj->set_channel_protocol(module => $module,
                                           protocol => $protocol);

Set channeling protocol for a module.

The following named arguments are accepted:

=over

=item Arguments

=over

=item module

(Required) : Module number.

=item protocol

(Required) : Type of channeling protocol to set.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_channel_protocol {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            module   => { type => SCALAR },
            protocol => { type => SCALAR },
        }
    );
    my $out;

    # Send command to switch
    $out = $self->_say("set channelprotocol $opts{protocol} $opts{module}");

    # Check for correct output
    if ( $out !~ /Mod \S+ is set to/ ) {
        croak $out;
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 show_port_lacp_channel

  $port_hash = $switch_obj->show_port_lacp_channel(port => $port);

Get LACP channel information for port.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : Port (given by module#/port#, e.g., 3/32) to get LACP channel
information for.

=back

=item Returns

Hash reference of the LACP channel information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_port_lacp_channel {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => SCALAR },
        }
    );
    my ($ret, $out);
    my %key_map = (
        channel_mode  => 'Channel Mode',
        admin_key     => 'Admin key',
        channel_id    => 'Ch id',
        lacp_priority => 'LACP Port Priority',
    );

    # Send command to switch
    $out = $self->_say("show port lacp-channel $opts{port}");
    croak $out if $out =~ /No ports channell?ing/;

    # Check for correct output
    if ( $out !~ /Port\s+Admin/ ) {
        croak $out;
    }

    # Process the output by converting it to a hash
    $ret = _convert_hash(_process_output(\$out), \%key_map);

    return $ret->{$opts{port}};
}

#---------------------------------------------------------------------

=head2 set_port_lacp_channel

  $out = $switch_obj->set_port_lacp_channel(port => \@port_list,
                                            mode => $mode);

Set port LACP channel parameters. If the mode is not specified
then it will just assign the ports to the LACP channel.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : List of ports to set LACP channel parameters.

=item admin_key

(Optional) : Admin key to assign to ports. If not specified the switch
will select a value. This value is ignored if the 'mode' parameter
is specified.

=item mode

(Optional) : Mode ('on', 'off', 'active' or 'passive').

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_port_lacp_channel {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port      => { type => ARRAYREF },
            admin_key => { type => SCALAR, default => '' },
            mode      => { type => SCALAR, optional => 1 },
        }
    );
    my ($cmd, $ports, $out);

    # Create list of ports
    $ports = join(',', @{$opts{port}});

    $cmd = "set port lacp-channel $ports";
    if (defined($opts{mode})) {
        $cmd .= " mode $opts{mode}";
    } else {
        $cmd .= " $opts{admin_key}";
    }

    # Send command to switch
    $out = $self->_say($cmd);

    # Check for correct output
    if ( defined($opts{mode}) ) {
        croak $out if ( $out !~ /Port\(s\) \S+ channel mode set to/ );
    } else {
        croak $out if ( $out !~ /Port\(s\) \S+ are assigned to admin key/ );
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 show_vlan

  $out = $switch_obj->show_vlan([ vlan => $vlan ]);

Get information about the vlan(s) in the switch.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vlan

(Optional) : Vlan to get information for.
If this argument is a list, the information for all vlans specified is returned.
Defaults to get the information about all vlans in the switch.

=back

=item Returns

When argument is a scalar (a single vlan) a hash reference of the vlan
information is returned. Otherwise, if argument is not specified or
is a list of vlans, a hash reference having as a key the vlan
and the value is a hash reference to the information for each vlan.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_vlan {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            vlan => { type => SCALAR|ARRAYREF, default => '' },
        }
    );
    my ($ret, $vlan, $out);

    if ( ref $opts{vlan} eq 'ARRAY' ) {
        # Vlan argument is an array, so get the vlan info for all
        # vlans specified in the list
        foreach $vlan ( @{$opts{vlan}} ) {
            $ret->{$vlan} = $self->show_vlan(%opts, vlan => $vlan);
        }
        return $ret;
    }

    # Send command to switch
    $out = $self->_say("show vlan $opts{vlan}");

    # Merge data from last column
    _merge_last_column(\$out);

    # Process output of command
    $ret = _convert_hash(_process_output(\$out));

    # Expand ports
    foreach $vlan ( keys %$ret ) {
        if ( defined $ret->{$vlan}{ports} ) {
            if ( $ret->{$vlan}{ports} eq 'internal' ) {
                delete $ret->{$vlan}{ports};
            } else {
                $ret->{$vlan}{ports} = _expand_ports($ret->{$vlan}{ports});
            }
        }
    }

    if ( $opts{vlan} ne '' ) {
        # Return just the vlan info if vlan is specified
        $ret = $ret->{$opts{vlan}};
    }

    return $ret;
}

#---------------------------------------------------------------------

=head2 show_gvrp_conf

  $out = $switch_obj->show_gvrp_conf();

Get the GVRP configuration on the switch.

=over

=item Returns

Hash reference of the GVRP configuration.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub show_gvrp_conf {
    my $self = shift;
    my (%ret, $item, $port, $ports, $global, $conf, $part, $out);

    # Send command to switch
    $out = $self->_say("show gvrp conf");

    # Separate sections of output
    ($global, $conf, $part) = ($out =~ /(.*)\nPort based GVRP Configuration:\n(.*)\nGVRP Participants running on (.*)\./s);

    # Process global section
    ($ret{gvrp_feature}) = ($global =~ /GVRP Feature is currently (\S+)/);
    ($ret{dynamic_vlan}) = ($global =~ /GVRP dynamic VLAN creation is (\S+)/);
    $ret{timers}         = { $global =~ /(\S+) = (\d+)/g };

    # Process the port based GVRP configuration
    # Start by making each entry into a single line
    _merge_last_column(\$conf);
    # Process section and create an array
    $conf = _process_output(\$conf, 1);

    # Process each configuration entry
    foreach $item ( @$conf ) {
        # Expand all ports in the entry and create a hash
        # for each port with the configuration information
        $ports = _expand_ports($item->{'Port(s)'});
        foreach $port ( @$ports ) {
            $ret{ports}{$port} = {
                status       => $item->{'GVRP-Status'},
                registration => $item->{'Registration'},
                applicant    => $item->{'Applicant'},
                participant  => 0,
            };
        }
    }

    # Process the GVRP participants
    $ports = _expand_ports($part);
    foreach $port ( @$ports ) {
        $ret{ports}{$port}{participant} = 1;
    }

    return \%ret;
}

#---------------------------------------------------------------------

=head2 set_gvrp

  $out = $switch_obj->set_gvrp(port => $port
                               [, enable => 0|1 ]);

Set GVRP options on a port.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : Port to set GVRP options.

=item enable

(Optional) : Enable/disable GVRP on the port.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_gvrp {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port   => { type => SCALAR },
            enable => { type => SCALAR, optional => 1 },
        }
    );
    my ($type, $out);

    if ( defined $opts{enable} ) {
        if ( $opts{enable} ) {
            # Enable gvrp on port
            $type = 'enable';
        } else {
            # Disable gvrp on port
            $type = 'disable';
        }

        # Send command to switch
        $out = $self->_say("set gvrp $type $opts{port}");
        croak $out if $out !~ /GVRP\s+(en|dis)abled/;
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 verify_no_interconnect

  $out = $switch_obj->verify_no_interconnect(port => $port|\@port_list);

Verify that all ports listed do not belong to a module which is used
for the interconnect.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Optional) : Port (given by module#/port#, e.g., 3/32) to check.
If this argument is a list, all ports specified are checked.

=back

=item Returns

1 on success, otherwise it croaks with error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub verify_no_interconnect {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port => { type => SCALAR|ARRAYREF },
        }
    );
    my (@ports, $info, $chid, $port, $module, %mod_hash);
    my $err = '';

    # Get list of ports to check
    if ( ref $opts{port} eq 'ARRAY' ) {
        @ports = @{$opts{port}};
    } else {
        @ports = ($opts{port});
    }

    # Get port channel information which includes the ports
    # used for the interconnect
    $info = $self->show_port_channel();

    # Get all modules used for the interconnect
    foreach $chid (keys %$info) {
        foreach $port (keys %{$info->{$chid}}) {
            if ($info->{$chid}{$port}{port_id}) {
                # This port has the port_id defined, so its module
                # is used for the interconnect
                $module = $self->get_module_from_port(port => $port);
                $mod_hash{$module} = 1 if defined $module;
            }
        }
    }

    # Check all specified ports
    foreach $port (@ports) {
        # Get module the port belongs to
        $module = $self->get_module_from_port(port => $port);
        if ($mod_hash{$module}) {
            # This port belongs to a module used for the interconnect,
            # save the port information so all ports are checked
            $err .= "\nPort $port belongs to module $module " .
                    "which is used for the interconnect";
        }
    }
    croak 'The following ports on switch "' . $self->{conn}->hostp->id .
          qq(" belong to a module used for the interconnect:$err) if $err;

    return 1;
}

#---------------------------------------------------------------------
# Internal method
# Process the output of command using fixed-width columns. The width
# of each column is determined by the column header '---'
#---------------------------------------------------------------------
sub _process_output {
    my ($out, $mode) = @_;
    my (%ret, @ret_list, @fkeys, @cols, @vals, @list, @fhead, $line, $i);
    my ($offset, $len, $item, $auto, $prev_key);

    # Process each line at a time
    foreach $line ( split /\n/, $$out ) {
        # Process speed legend if found
        if ( $line =~ /\* - config speed is (.*)\./ ) {
            $auto = defined $1 ? $1 : '';
            foreach $item ( keys %ret ) {
                if ( exists $ret{$item}{Speed} ) {
                    # Add the legend to the value
                    $ret{$item}{Speed} =~ s/\*/$auto/;
                }
            }
            next;
        }
        if ( !@fkeys ) {
            # This line is part of the header
            if ( $line =~ /^-/ ) {
                # This line has the column widths
                # Get column widths (size and offset)
                @cols = ();
                $offset = 0;
                # Include spaces after the dashes in the width calculation
                @list = ($line =~ /(-+\s*)/g);
                while ( @list ) {
                    # Create list of column information, include size and offset
                    $len = length(shift @list);
                    push @cols, { size => $len, offset => $offset };
                    $offset += $len;
                }

                # Process header information, create list of column headers
                # Concatenate multi-line column headers
                foreach $item ( @fhead ) {
                    foreach $i ( 0..@cols-1 ) {
                        last if $cols[$i]{offset} >= length($item);
                        # Append this column header
                        $fkeys[$i] .= substr(
                            $item,
                            $cols[$i]{offset},
                            $cols[$i]{size},
                        );
                    }
                }

                # Remove any unnecessary spaces
                foreach ( @fkeys ) {
                    s/\s+/ /g;
                    s/^\s+//;
                    s/\s+$//;
                }
            } else {
                # This line is part of the header information
                push @fhead, $line if $line;
            }
        } elsif ( !$line ) {
            # An empty line marks the end of current column format
            @fkeys = ();
            @fhead = ();
            $prev_key = undef;
        } else {
            # This line is the actual info
            @vals = ();

            # Do not truncate data on the last column
            $cols[-1]{size} = 99999;

            foreach $i ( 0..@cols-1 ) {
                if ( $cols[$i]{offset} > length($line) ) {
                    # This line has no more data for this column
                    $vals[$i] = '';
                } else {
                    # Get data given by the current offset and size
                    $vals[$i] = substr(
                        $line,
                        $cols[$i]{offset},
                        $cols[$i]{size},
                    );
                    # Remove any unnecessary spaces
                    $vals[$i] =~ s/^\s+//;
                    $vals[$i] =~ s/\s+$//;
                }
            }

            if ( $mode ) {
                @ret{@fkeys} = @vals;
                push @ret_list, { %ret };
                %ret = ();
            } else {
                # The first column is the key, while the rest of the columns
                # are the information associated with the first column
                $item = $vals[0];
                # Remove gratuitous heading dividers
                next if $item =~ /^-+$/;
                # Add information for this line into the return hash
                foreach $i ( 1..@vals-1 ) {
                    if ( defined $prev_key && $item eq '' ) {
                        $ret{$prev_key}{$fkeys[$i]} .= ", $vals[$i]"
                          if $vals[$i] ne '';
                    } else {
                        $ret{$item}{$fkeys[$i]} = $vals[$i];
                    }
                }
                $prev_key = $item;
            }
        }
    }

    if ( $mode ) {
        croak $$out if !@ret_list;
        return \@ret_list;
    } else {
        croak $$out if !%ret;
        return \%ret;
    }
}

#---------------------------------------------------------------------
# Internal method
# Merge all lines into the last column so output like:
# VLAN Name      Mod/Ports
# ---- --------- ----------------
# 560  VLAN0560  1/1-4
#                4/1-3
#                15/1
#
# can be converted into:
# VLAN Name      Mod/Ports
# ---- --------- ----------------
# 560  VLAN0560  1/1-4,4/1-3,15/1
#
# So the output can be parsed by _process_output
#---------------------------------------------------------------------
sub _merge_last_column {
    my $out = shift;
    my ($line, @ret);
    my $in_header = 1;

    # Process each line at a time
    foreach $line ( split /\n/, $$out ) {
        if ( $line eq '' ) {
            # Done with info, next header follows
            $in_header = 1;
        } elsif ( $line =~ /^-+\s/ ) {
            # End of header
            $in_header = 0;
        } elsif ( !$in_header && $line =~ s/^\s+/,/ ) {
            # Append current data to last column
            $ret[-1] .= $line;
            next;
        }
        push @ret, $line;
    }

    # Replace with processed data
    $$out = join("\n", @ret);

    return;
}

#---------------------------------------------------------------------
# Internal method
# Expand ports given in a scalar of the form ('1/1-4,2/3,6/1') into
# an array ['1/1', '1/2', '1/3', '1/4', '2/3', '6/1']
#---------------------------------------------------------------------
sub _expand_ports {
    my $port_list = shift;
    my (@ports, $port, $mod, $start, $end);

    foreach ( split /,/, $port_list ) {
        # Get the module and the range
        ($mod, $start, $end) = m{(\d+)/(\d+)(?:-(\d+))?};
        next if (!defined $mod || !defined $start);
        $end = $start if !defined $end;

        # Append range of ports to list
        foreach $port ( $start..$end ) {
            push @ports, "$mod/$port";
        }
    }

    return \@ports;
}

#---------------------------------------------------------------------
# Internal method
# Process input hash to change the keys given in the global key map
# and return a new hash reference with new keys
# All keys not found in the global key map are discarded
#---------------------------------------------------------------------
sub _convert_hash {
    my ($hash, $kmap) = @_;
    my (%ret, $key, $item);
    $kmap = \%key_map_h if !defined $kmap;

    foreach $item ( keys %$hash ) {
        foreach $key ( keys %$kmap ) {
            if ( exists $hash->{$item}{$kmap->{$key}} ) {
                $ret{$item}{$key} = $hash->{$item}{$kmap->{$key}};
            }
        }
    }
    return \%ret;
}

#---------------------------------------------------------------------
# Wrapper for Tharn::say
#---------------------------------------------------------------------
sub _say {
    my ($self, $cmd, @args) = @_;
    my (@list, $qcmd);

    if ( !$self->{_say_enable} ) {
        # Change to admin mode
        Tharn::say($self->{conn}, "enable\n");
        $self->{_say_enable} = 1;
    }

    # Send the command to the switch
    @list = Tharn::say($self->{conn}, $cmd, '-swallow', @args);

    if ( defined $list[2] ) {
        # Remove all '\r' from output
        $list[2] =~ s/\r+//g;

        # Remove echo of command from output and everything before it
        $qcmd = quotemeta($cmd);
        $list[2] =~ s/.*$qcmd\n*//s;

        # Remove console messages
        $list[2] =~ s/^$SYSLOG_DATE.*\r?\n//mg;
    }

    if ( wantarray ) {
        return @list;
    } else {
        return $list[2];
    }
}

#---------------------------------------------------------------------

=head2 set_gvrp_reg

  $out = $switch_obj->set_gvrp_reg(port => $port
                               [, register => $register ]);

Set GVRP registration options on a port.

The following named arguments are accepted:

=over

=item Arguments

=over

=item port

(Required) : Port to set GVRP options.

=item register

(Required) : fixed/normal GVRP registration on the port.

=back

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_gvrp_reg {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            port   => { type => SCALAR },
            register => { type => SCALAR },
        }
    );
    my $out;

    #Send command to switch
    $out = $self->_say("set gvrp registration $opts{register} $opts{port}");

    if ( $out !~ /Registrar Administrative Control set/i)
    {
       return 0;
    }
       
   return 1;

} #End of set_gvrp_reg

#---------------------------------------------------------------------

1;

