# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Toe.pm#1 $
#
## @summary	Toe - Library routines for interface to the filer toe command
## @author	dl-spe-nicqa@netapp.com
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

Ttcp - Library routines for interface to the filer toe command

=head1 SYNOPSIS

  use Network::Toe;

  # TOE port offload list
  $info = toe_port_list(fc => $fc);

  # TOE status
  $info = toe_status(fc => $fc);

=head1 DESCRIPTION

Provides routines to access diagnostic command line interface to the TOE cards.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-spe-nicqa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Network/Toe.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Toe;

use strict;
use warnings;
use base qw(Exporter);
use Carp qw(croak);
use SayWrap qw(sayEOF);
use Params::Validate qw(validate OBJECT);

our @EXPORT_OK = qw(
    toe_port_list
    toe_status
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub toe_port_list(%;);
sub toe_status(%;);
sub _change_keys($;);

#---------------------------------------------------------------------

=head2 toe_port_list

  $info = toe_port_list(fc => $fc)

Get information about the TOE port offload list. The associated
toe command has to be run at the test priveledge level which typically
means that a debug version of ONTAP needs to be loaded on the filer.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

A hash reference of the TOE port offload list.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub toe_port_list(%;) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my (%hash, @lines, $line, $fhead, @fkeys, $toe, %ret, $out, $cmd);

    # Build the toe command. The port list has to be run at the advanced
    # priveledge level.
    $cmd = "priv set -q advanced; toe port list ";

    # Get the messages
    $out = sayEOF($opts{fc}, $cmd);
    if ($out =~ /(toe not found|toe status|Usage|unrecognized privilege)/i) {
        croak ("Filer does not support toe command");
    }

    # Split each of the messages
    @lines = split /\n/, $out;

    # Move the TOE headings
    $fhead = shift @lines;

    # Split up the headings based on 1 or more spaces between each one
    @fkeys = split /\s{1,}/, $fhead;

    # Parse out the TOE port list information
    foreach $line (@lines) {
        # Take out leading spaces
        $line =~ s/^\s+//;

        # Take out hyphens
        $line =~ s/\-//g;

        next if !$line;

        # Split TOE information using the headings as keys
        @hash{@fkeys} = split /\s{2,}/, $line;

        # Load up all the TOE card information into an array
        push @{$toe}, _change_keys(\%hash);
    }

    # Add TOE port information to return hash
    push @{$ret{port_list}}, $toe;

    return \%ret;
}

#---------------------------------------------------------------------

=head2 toe_status

  $info = toe_status(fc => $fc);

Get information about the TOE NIC card(s) status.

=over

=item Arguments

=over

=item fc

(Required) : Connection object to a filer.

=back

=item Returns

An array of hashes containing the TOE card(s) status.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub toe_status(%;) {
    my (%opts) = validate (
        @_,
        {
            fc   => { type => OBJECT },
        }
    );
    my (%hash, @list, @lines, $line, $fhead, @fkeys, $toe, %ret, $out, $item,
        $type, $connections, $slot, @ifcs, $i, @fields, $status, $cmd, @ret,
        $info);

    # Build the toe command
    $cmd = "priv set -q advanced; toe status -n";

    # Get the messages
    $out = sayEOF($opts{fc}, $cmd);
    if ($out !~ /Interface/i) {
        croak ("The 'toe status' command found no offload connections");
    }

    # Put the messages per NIC card into a list (array)
    @list = split /Interface\s+/i, $out;

    # Each $item contains 1 NIC card of messages
    foreach $item (@list) {
        next if !$item;
        next if ($item =~ /no offload connection/i && $item !~ /TOE/i);

        # Take out the semi-colons from the interface message
        $item =~ s/://g;

        # Take out leading spaces
        $item =~ s/^\s+//;

        # Split each of the message lines
        @lines = split /\n/, $item;

        # The first line has the interface(s) and status
        $line = shift @lines;

        @fields = split /\s{1,}/, $line;

        undef @ifcs;

        # Get the list of interfaces for the selected NIC card
        for ($i=0; $i<scalar(@fields); $i++) {
            if ($fields[$i] ne "TOE") {
                push @ifcs, $fields[$i];
            } else {
                $type = $fields[$i];
                $status = $fields[$i+2];
                $i = scalar(@fields);
            }
        }

        undef $toe;

        # Load up interface(s)
        push @{$toe}, @ifcs;

        $slot->{ifc} = $toe;

        # Add TOE card information to the return hash
        $slot->{status} = $status;

        # The second line has the TOE headings
        # Move the TOE headings
        $fhead = shift @lines;

        # Split the headings
        ($fkeys[0], $fkeys[1], $fkeys[2], $fkeys[3], $fkeys[4], $fkeys[5]) =
          ($fhead =~ /\s*(\w+)\s+(\w+)\s+(\w+\s+\w+)\s+(\w+\s+\w+)\s+(\w+\s+\w+)\s+(\w+\s+\w+)/g);

        undef $info;
        undef $connections;

        # The third line starts the TOE values and then the totals
        foreach $line (@lines) {
            next if !$line;
            next if $line =~ /no offload connection/i;
            if ($line !~ /Total/i) {
                # Split the card information
                @hash{@fkeys} = split /\s{1,}/, $line;

                # Load up all the TOE card information into an array
                push @{$info}, _change_keys(\%hash);

                # Add TOE card information to the return hash
                $slot->{info} = $info;
            } else {
                # Parse the card information
                %hash = ($line =~ /\s*(.+)\s+(.+)/g);

                # Load up all the TOE card information into an array
                push @{$connections}, _change_keys(\%hash);

                # Add TOE card information to the return hash
                $slot->{connections} = $connections;
            }
        }

        # Add per slot messages to the return hash
        push @ret, {$type => $slot} if $slot;
        undef $slot;
    }

    return \@ret;
}

#---------------------------------------------------------------------
# Internal function to change spaces on the keys of the specified hash
#---------------------------------------------------------------------
sub _change_keys($;) {
    my $hash = shift;
    my (%ret, $key);

    # Process all keys in the hash
    foreach ( keys %$hash ) {
        $key = $_;
        # Change spaces to underscore
        $key =~ s/\s/_/g;
        # Change periods to underscore
        $key =~ s/\./_/g;
        $ret{$key} = $hash->{$_};
    }

    return \%ret;
}

1;
