
#
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Trace.pm#3 $
#

## @summary Network-Trace - a simple network tracing class
## @author  ewing@netapp.com
## @pod here

=head1 NAME

Network::Trace - a simple network tracing class

=head1 SYNPOSIS

{
    my $trace = use Network::Trace(address => '1.2.3.4');

    # do something you want to trace
}

=head1 DESCRIPTION

This class will trace network activity to the specified endpoint from the
node on which it was called.  The trace can be cancelled, stopped, and 
restarted.  If an object reference falls out of scope, the trace will be
stopped gracefully.

=head1 METHODS

=cut

package Network::Trace;

use strict;
use warnings;

use Tharn;
use TracePkg;
use Options;
use Params::Validate qw(ARRAYREF UNDEF SCALAR);
use Module::Common qw(validate_mixed);
use Sys::Hostname;
use Data::Dumper;
use File::Path;
use File::Copy;

use constant DATA => 1<<2;
use constant MGMT => 1<<1;
use constant CLUSTER => 1<<0;

# Only load if this has been subtested out to windows. We are issuing the 'use' 
# here as opposed to a 'require' in the *tshark() subroutines because the use
# is necessary to resolve the constant variables at compile time.

use if ($^O eq "MSWin32"), 'Win32::Process';
use if ($^O eq "MSWin32"), 'Win32::SearchPath';
use if ($^O ne "MSWin32"), 'POSIX';
use if ($^O ne "MSWin32"), 'POSIX' => ':sys_wait_h';
use if ($^O ne "MSWin32"), 'constant' => 'NORMAL_PRIORITY_CLASS';

my @Capture_Thread;
my $logdir_is_tempdir = 0;

my $Trace = new TracePkg();
$Trace->set_options(separateNewline => 1);
my $Options = new Options();
$Options->parameters(
    definitions => [
    {
        name => 'LOGDIR',
        description => "The location of the logs",
    },
    {
        name => 'NETWORK_TRACE',
        description => "Enable/disable network tracing",
        default => 0,
    },
    {
        name => 'TRACE_ID',
        description => "The trace id",
        default => undef(),
    },
    {
        name => 'SUDO',
        description => "The location of the sudo binary to use.",
        default => undef(),
    },
    {
        name => 'USE_TCPDUMP',
        description => "Tshark is the default trace capture that is used, ".
                       "however, the user can flag to use TCPDUMP instead",
        spec => "boolean",
        default => "false",
    },
    {
        name => 'TCPDUMP',
        description => "The location of the tcpdump binary to use.",
        default => undef(),
    },
    {
        name => 'TSHARK',
        description => "The location of the tshark executable to use.",
        default => undef(),
    },
    {
        name => 'TRACE_FILTER',
        description => "The filter to use when capturing data.",
        default => undef(),
    },
    {
        name => 'TRACE_SNAPLEN',
        description => "The amount of data to capture from each packet. (default: 256 Bytes)",
        default => undef(),
    },
    {
        name => 'TRACE_INTERFACE',
        description => "The interface to capture from. (default: eth1 for linux clients, ".
                       "all for windows clients)",
        default => undef(),
    },
    {
        name => 'TRACE_NETWORK',
        description => "The network in which to capture from. (default: data)",
        default => "data",
        valid_values => ["data", "mgmt"],
    },
    {
        name => 'TRACE_FILES',
        description => "The number of capture files to use in the rotating buffer. (default: 10)",
        default => undef(),
    },
    {
        name => 'TRACE_FILES_SIZE',
        description => "The size of each file in the rotating buffer. (default: 10)",
        default => undef(),
    },
    {
        name => 'TRACE_FILES_OWNER',
        description => "The owner of the trace files. (default: root)",
        default => undef(),
    },
    {
        name => 'TRACE_FATALITY',
        description => "The fatality of a network tracing problem.",
        default => undef(),
    },
    {
        name => 'TRACE_DURATION',
        description => "Length of time for tshark to capture packets. Setting it to 0 ".
                       "will cause it to run until a stop has been called",
        spec        => "i",
        greater_equal => 0,
        default => undef(),
    },
    {
        name => 'TRACE_ADDRS',
        description => "A parameter for traceable addresses.  Not used directly by this module",
        spec => 's@',
        delimiter => ",",
        default => undef(),
    },
    {
    	name => 'TRACE_REMOVE',
    	description => "A flag to control the removal of dump files, (default: 1)",
        default => 1,
    },
    {
        name        => "LOCAL_TEMP_DIR",
        description => "Local directory on client to write packet captures to. Set to ".
                       "string 'LOGDIR' to log to user's actual LOGDIR instead of the ".
                       "client's local disk",
        default     => undef(),
    },
    {
        name        => "TSHARK_WAIT",
        description => "The amount of time (in seconds) to wait for tshark to ".
                       "finish on a windows client",
        default     => "60",
    },
    {
	name        => "PCAP_UNC_PATH",
	description => "UNC path where pcaps should be written",
	default     => undef(),
    },
    ]
);
$Options->compatability(definitions=>{ 
    SUDO_PATH            => "SUDO",
    });

$Options->process();

=head2 new

This is the constructor for the Network::Trace package. The following 
parameters are availalble:

=over

=item address

(required) : network address to use for tracing

=item id

(optional) : an identifier added to the trace, defaults to the local hostname

=item filter

(optional) : an additional filter to be used when capturing network traffic

=item sudo

(optional) : path to sudo (defaults to /usr/local/bin/sudo)

=item tcpdump

(optional) : path to tcpdump (defaults to the location under 'which tcpdump')

=item tshark

(optional) : path to tshark (defaults to C:\Program Files\Wireshark\tshark)

=item snaplen

(optional) : the number of bytes to capture from each packet

=item interface

(optional) : the network interface from which to capture packets 

=item files

(optional) : the number of files to use in the circular buffer

=item size

(optional) : the size of each file in the circular buffer

=item owner

(optional) : the user that will own the capture files 

=item duration

(optional) : length of time to capture packets

=back

=for html <hr>

=cut

sub new {

    $Trace->entry();
    my $class = shift();

    my %p = @_;
    my $self = ();

    # Get options from what was passed in, parameters & some defaults.

    $self->{address}   = $p{address} || $Trace->fatal(msg => "An address is required for tracing.", result => 'SCRIPT');
    $self->{tcpdump}   = $Options->get(option => [qw(TCPDUMP)]) ||            $p{tcpdump};
    $self->{tshark}    = $Options->get(option => [qw(TSHARK)]) ||             $p{tshark};
    $self->{sudo}      = $Options->get(option => [qw(SUDO)]) ||               $p{sudo}     || "/usr/local/bin/sudo";
    $self->{id}        = $Options->get(option => [qw(TRACE_ID)]) ||           $p{id}        || hostname();
    $self->{filter}    = $Options->get(option => [qw(TRACE_FILTER)]) ||       $p{filter}    || undef();
    $self->{snaplen}   = $Options->get(option => [qw(TRACE_SNAPLEN)]) ||      $p{snaplen}   || 500;
    $self->{interface} = $Options->get(option => [qw(TRACE_INTERFACE)]) ||    $p{interface} || 'eth1';
    $self->{files}     = $Options->get(option => [qw(TRACE_FILES)]) ||        $p{files}     || 100;
    $self->{size}      = $Options->get(option => [qw(TRACE_FILES_SIZE)]) ||   $p{size}      || 10240;
    $self->{run_as}    = $Options->get(option => [qw(TRACE_FILES_OWNER)]) ||  $p{owner}     || 'root';
    $self->{fatality}  = $Options->get(option => [qw(TRACE_FATALITY)]) ||     $p{fatality}  || 0;
    $self->{duration}  = $Options->get(option => [qw(TRACE_DURATION)]) ||     $p{duration}  || 0;

    # Save off additional command line parameters (that have default values)

    $self->{use_tcpdump} = $Options->get(option => qw[USE_TCPDUMP]);
    $self->{network}  = $Options->get(option => [qw(TRACE_NETWORK)]);
    $self->{remove}   = $Options->get(option => [qw(TRACE_REMOVE)]);
    $self->{wait}     = $Options->get(option => [qw(TSHARK_WAIT)]);

    #
    # If tracing is disabled, return a reference to an EmptyTrace subclass object.
    #
    if ($Options->get(option => [qw(NETWORK_TRACE)]) == 0) {
        $Trace->exit();
        return new Network::EmptyTrace();
    }

    # Find the host information so that we will know which tracing subroutines to call 
    # later

    $self->{hostname} = hostname();
    my @splithost = split(/\./, $self->{hostname});
    $self->{hostname} = $splithost[0];
    eval {
        $self->{host} = host($self->{hostname});
    };
    if ($@) {
        $Trace->fatal(msg => "Encountered unexpected error while getting the ".
                      "host information for $self->{hostname}. Failed with: $@");
    }
    if (!defined($self->{host})) {
        $Trace->fatal(msg => "Unable to get host information for $self->{hostname}.",
                      result => "CONFIG");
    }

    # Set the global temp directory that exists locally on the client.
    # If the user has specified a particular temporary directory to use, we can
    # use that. If they specified the string LOGDIR, then we will use that. Otherwise, 
    # we'll use what tempdir() returns (different per platform).

    my $unc_path = $Options->get(option => [qw(PCAP_UNC_PATH)]);
    my $temp_dir = $Options->get(option => [qw(LOCAL_TEMP_DIR)]);
    if (defined($temp_dir) && $temp_dir =~ /^LOGDIR$/) {
        $self->{temp_dir} = $Options->get(option => [qw(LOGDIR)]);

        # CAUTION! When running with 'LOCAL_TEMP_DIR=LOGDIR' the user should
        # be aware of 2 things:
        #     1. The permissions on their LOGDIR need to be opened up.
        #     2. Occassionally, Tshark on windows clients will fail to write
        #        to LOGDIR. Reason is still TBD. The safest way to run is to
        #        let it capture locally on the client, and then copy the files
        #        to the LOGDIR.
        #
        # We are using this flag to show that the logdir is the temp_dir.
        # This will be used for various checks down the road.

        $logdir_is_tempdir = 1;

    } elsif ((defined($unc_path)) && ($self->{host}{hosttype} =~ /windows/i)) { 
        $self->{temp_dir} = $unc_path;
    } elsif (!$temp_dir) {
        $self->{temp_dir} = File::Temp::tempdir();
    } else {
        $self->{temp_dir} = $temp_dir;
    }
    $Trace->comment(msg => "Using $self->{temp_dir} for local file storage");

    # Bless the class

    bless($self, $class);

    #
    # If we are supposed to find the TRACE_INTERFACE that is part of the TRACE_NETWORK
    # we will do that here (only if we are running on a linux client. Tracing is enabled
    # on all interfaces for windows clients).
    #
    if (($self->{interface} eq "find") && (($^O ne "MSWin32"))) {
        $self->_find_trace_interface();
    }

    # Initiate tracing

    $self->_start_trace();

    # Return (tracing will still be running in the background)

    $Trace->exit();
    return $self;
}

#
# _stop_trace
#
# Calls the appropiate tracing subroutine depending on the client's hosttype.
#
sub _stop_trace {

    $Trace->entry();
    my $self = shift();

    my %p = (
        remove => 0,
        @_
    );

    # Since all linux clients are managed via subtest (no matter if we are
    # running tcpdump or tshark), we will just call _stop_linux_trace. Otherwise
    # call _stop_tshark_win to handle stopping tshark on windows.

    if ($self->{host}{hosttype} =~ /windows/i) {
        $self->_stop_tshark_win(remove => $p{remove});
    } else {
        $self->_stop_linux_trace(remove => $p{remove});
    }

    $Trace->exit();
}

#
# _stop_linux_trace
#
# Does the actual killing of the trace capture subtest as well as the optional 
# removal/copy of the trace file.
#
sub _stop_linux_trace {

    $Trace->entry();
    my $self = shift();

    my %p = (
        remove => 0,
        @_
    );

    # A little extra debugging may be needed here.

    local $Tharn::THARN_VERBOSE = 2;

    # Verify the trace is running and remove the trace file if requested.
    # There is probably some more checking that should be done here.

    if (defined($self->{subtest})) {
        if ($self->{subtest}->is_alive()) {

            $Trace->debug(msg => "Stopping active trace subtest (".
                          $self->{subtest}->pid().")");
            $self->{subtest}->kill(message => 'ok');

            # Get a list of all of the files that have been written
            # out to the temp directory. If tcpdump was used the file
            # will be in the following format: <filename>.pcap<unique id>

            opendir(DIR, $self->{temp_dir}) or 
                $Trace->warn(msg => "Cannot open directory $self->{temp_dir}");
            my @capture_files = grep(/\.pcap/, readdir(DIR));
            closedir(DIR);
            
            if (scalar(@capture_files) < 1) {
                $Trace->warn(msg => "No *.pcap files found in $self->{temp_dir}");
            }

            # If we have been flagged to remove, then we will just
            # skip the copy. If the temp_dir is the same as the LOGDIR, 
            # then we can also skip the copy.
            
            my $logdir = $Options->get(option => [qw(LOGDIR)]);
            if (!$p{remove} && (!$logdir_is_tempdir)) {

                # Copy each file

                foreach my $capture_file (@capture_files) {

                    # Catch the [0] interface which doesn't exist in the array at all
                    
                    next if ( !defined $capture_file || $capture_file eq "" );

                    # Copy the rest of the .pcap files to the destination log directory.
                    
                    if ( -e "$self->{temp_dir}/$capture_file" ) {
                        $Trace->comment(msg => "Copying capture file from '$self->{temp_dir}/$capture_file' to ".
                                        "'$logdir/$capture_file'");
                        eval {
                            system("sudo cp $self->{temp_dir}/$capture_file $logdir/$capture_file");
                        };
                        if ($@) {
                            $Trace->fatal(msg => "Copy failed: $@");
                        }

                    } else {
                        $Trace->warn(msg => "Capture didn't create a file '$capture_file'");
                    }
                }
            }

            # Remove the contents of the temp_dir. This could be accompolished with
            # specifing 'CLEANUP => 1' to tempdir() but that does not seem to be working
            # so we will explicity call the delete here. If it fails, we will only 
            # throw a warning. Files left in a temporary directory will get cleaned-up
            # on reboot anyways. Only delete the entire temp_dir if it is not the logdir.

            if (!$logdir_is_tempdir) {

                eval {
                    rmtree($self->{temp_dir});
                    $Trace->comment(msg => "Temporary directory $self->{temp_dir} has been ".
                                    "deleted.");
                };
                if ($@) {
                    $Trace->warn(msg => "Failed to remove temporary directory: ".
                                 "$self->{temp_dir}");
                }

            } else {

                # If we are removing .pcap files from the logdir, we can just do the
                # following.

                if ($p{remove}) {

                    foreach my $file (@capture_files) {
                        unlink($self->{temp_dir}."/".$file);
                    }
                    
                    $Trace->comment(msg => "Packet captures in $self->{temp_dir} have ".
                                    "been successfully removed");
                } else {
                    $Trace->comment(msg => "Flagged to keep packet captures in LOGDIR.");
                }
            }

        } else {
            $Trace->debug(msg => "Trace subtest no longer active.");
        }

    } else {
        $Trace->debug(msg => "Network trace subtest isn't defined.");
    }

    $Trace->exit();
}

#
# _stop_tshark_win
#
# Does the actual stopping of tshark on windows clients including the optional 
# removal/copy of the trace file.
#
sub _stop_tshark_win {

    $Trace->entry();
    my $self = shift();

    my %p = (
        remove => 0,
        @_,
    );

    require Win32API::File;

    # Handle the captures

    eval {

        # Grab each thread that we started (one per interface)

        foreach my $thread (@Capture_Thread) {

            if (defined $thread) {

                # When the duration is set to 0, we will have to kill the thread

                if ($self->{duration} == 0) {
                    $Trace->trace(msg => "Killing packet capture to save its file.");
                    $thread->Kill(1); 
                }

                # Translate the timeout to milliseconds
                
                my $timeout = $self->{wait} * 1000;

                # Wait on the child to finish in case we finished up before 
                # the trace timeout (or the kill completed above). Note that
                # if the user specifies a duration that is longer than the test
                # run PLUS the timeout, then Wait() will fail. INFINITE was
                # also not used because this could cause Wait() to wait forever.
                # Running with a duration should be avoided unless the user is
                # only trying to capture the first X seconds of a run.

                my $pid = $thread->GetProcessID();
                $Trace->trace(msg => "Waiting $timeout milliseconds for packet ".
                              "capture '$pid' to finish.");
                $thread->Wait($timeout);

                my $exit_code;
                Win32::Process::GetExitCode($thread, $exit_code);
                $Trace->trace(msg => "ExitCode for packet trace '$exit_code'");
            }
        }
    };
    if ($@) {
        $Trace->warn(msg => "Failed to wait on packet captures '$@'");
    }

    # Since tshark appends the timestamp onto the end of the capture file name,
    # we don't know the exact name of the files (they aren't just what we passed
    # in to 'tshark -w ...'). Find those files here.
    
    opendir(DIR, $self->{temp_dir}) or die "Cannot open directory $self->{tempdir}";
    my @capture_files = grep(/\.pcap/, readdir(DIR));
    closedir(DIR);
    
    if (scalar(@capture_files) < 1) {
        $Trace->warn(msg => "No *.pcap files found in $self->{temp_dir}");
    }
    
    # If we have been flagged to remove, then we will just
    # skip the copy. If the temp_dir is the same as the LOGDIR, 
    # then we can also skip the copy.

    my $logdir = $Options->get(option => [qw(LOGDIR)]);    
    if (!$p{remove}  && (!$logdir_is_tempdir)) {

        # Copy the files

        foreach my $capture_file (@capture_files) {

            # Catch the [0] interface which doesn't exist in the array at all
            
            next if ( !defined $capture_file || $capture_file eq "" );

            # Copy the rest of the .pcap files to the destination log directory.

            if ( -e "$self->{temp_dir}/$capture_file" ) {
                $Trace->comment(msg => "Copying capture file from '$self->{temp_dir}/$capture_file' to ".
                              "'$logdir/$capture_file'");
                Win32API::File::CopyFile("$self->{temp_dir}/$capture_file", "$logdir/$capture_file", 1);
            } else {
                $Trace->warn(msg => "Capture didn't create a file '$capture_file'");
            }
        }
    }

    # Remove the contents of the temp_dir. This could be accomplished with
    # specifing 'CLEANUP => 1' to tempdir() but that does not seem to be working
    # so we will explicity call the delete here. If it fails, we will only 
    # throw a warning. Files left in a temporary directory will get cleaned-up
    # on reboot anyways. Only delete the entire temp_dir if it is not the logdir.

    if (!$logdir_is_tempdir && !$p{remove}) {

        eval {
            rmtree($self->{temp_dir});
            $Trace->comment(msg => "Temporary directory $self->{temp_dir} has been ".
                            "deleted.");
        };
        if ($@) {
            $Trace->warn(msg => "Failed to remove temporary directory: ".
                         "$self->{temp_dir}");
        }

    } else {
        
        # If we are removing .pcap files from the logdir, we can just do the
        # following.

        if ($p{remove}) {

            foreach my $file (@capture_files) {
                unlink($self->{temp_dir}."/".$file);
            }

            $Trace->comment(msg => "Packet captures in $self->{temp_dir} have ".
                            "been successfully removed");
        } else {
            $Trace->comment(msg => "Flagged to keep packet captures in LOGDIR.");
        }
    }

    $Trace->exit();
}

#
# _start_trace
#
# Calls the appropiate tracing subroutine depending on the client's hosttype
# and any user specified parameters.
#
sub _start_trace {

    $Trace->entry();
    my $self = shift();
    
    # Call the proper subroutine to start tracing

    if ($self->{host}{hosttype} =~ /windows/i) {

        # If we are running on a windows client, we will be running tshark 
        # regardless of any user specified parameters.

        $self->_start_tshark_win();

    } elsif (($self->{host}{hosttype} !~ /windows/i) && ($self->{use_tcpdump})) {

        # If we are not running on windows, and the user explictly flagged to run
        # tcpdump, we will kick that off here.
        
        $self->_start_tcpdump();

    } else {
        
        # Otherwise, we will start tshark on the linux client

        $self->_start_tshark_lin();
    }

    $Trace->exit();
}

#
# _start_tcpdump
#
# Start tcpdump on linux clients
#
sub _start_tcpdump {

    $Trace->entry();
    my $self = shift();

    # If the location of tcpdump is still undefined, find it.

    if (!defined($self->{tcpdump})) {
        $self->{tcpdump} = `which tcpdump`;
        chomp($self->{tcpdump});
    }
    $Trace->comment(msg => "Using '$self->{tcpdump}' for the network capture");

    # Dump to the specified trace_file

    $self->{trace_file} = $self->{temp_dir}."/".$self->{id}."_tcpdump_".
                          $self->{interface}."_".$self->{address}.".pcap";
    $Trace->comment(msg => "Network trace for $self->{address} is in $self->{trace_file}");

    # Check to see if a subtest exists and is running.

    if(!defined($self->{subtest}) || (!$self->{subtest}->is_alive())) {

        # Form arrays containing the tcpdump options and the filter to
        # be used when capturing data.

        my @filter;
        my @options;

        if (defined($self->{filter})) {
            @filter = split(/\s+/, $self->{filter});
            $Trace->debug(msg => "Overriding default filter with: ".join(" ", @filter)); 
        } else {
            @filter = ("host", $self->{address});
            $Trace->debug(msg => "Using default filter: ".join(" ", @filter)); 
        }

        # Loop through all of the possible flags that we could have for
        # tcpdump. If the user has defined any of these parameters, then we
        # will pass them along to tcpdump.

        foreach my $flag (
            ['-s', $self->{snaplen}],
            ['-i', $self->{interface}],
            ['-W', $self->{files}],
            ['-C', $self->{size}],
            ['-Z', $self->{run_as}],
            ['-w', $self->{trace_file}]
        ) {
            if (!length($flag->[1])) {
                next;
            }
            push(@options, ($flag->[0], $flag->[1]));
        }
        $Trace->debug(msg => "Set non-filter options: ".join(" ", @options)); 

        # Set what we should do if tcpdump fails (warn or fatal).

        my $onexit;
        if ($self->{fatality}) {
            $onexit = sub {
                my $st = shift(); 
                if (scalar($st->own_results())) {
                    $Trace->fatal(msg => "Network trace subtest failed prematurely (fatal).");
                }
            }
        } else {
            $onexit = sub {
                my $st = shift(); 
                if (scalar($st->own_results())) {
                    $Trace->warn(msg => "Network trace subtest failed prematurely (warn).");
                }
            }
        }

        # Subtest the tcpdump process.

        $self->{subtest} = Tharn::subtest(
            $self->{sudo},
            -bg,
            -onexit, $onexit,
            -runid, "tcpdump_$self->{id}\_$self->{address}",
            '--',
            $self->{tcpdump},
            @options,
            @filter
        );
        $Trace->debug(msg => "Network trace for $self->{address} is in $self->{trace_file}");

    } else {
        $Trace->debug(msg => "Network trace for $self->{address} is already running.");
    }

    $Trace->exit();
}

# _start_tshark_win
#
# Start tshark on windows clients
#
sub _start_tshark_win {

    $Trace->entry();
    my $self = shift();

    # If the user did not specify the TSHARK command line option or the tshark 
    # parameter to new(), we will find the location of the tshark executable 
    # here.

    if (!defined($self->{tshark})) {

        if ( -e "C:/Progra~1/Wireshark/tshark.exe" ) {
            #$self->{tshark} = "C:/Progra~1/Wireshark/tshark.exe";
            $self->{tshark} = "C:\\Progra~1\\Wireshark\\tshark.exe";
        } elsif ( -e "C:/Progra~2/Wireshark/tshark.exe" ) {
            #$self->{tshark} = "C:/Progra~2/Wireshark/tshark.exe";
            $self->{tshark} = "C:\\Progra~2\\Wireshark\\tshark.exe";
        }
        else {
            $Trace->fatal(msg => "Unable to find the tshark executable on ".
                          "$self->{hostname}",
                          result => "CONFIG");
        }
    }
    $Trace->comment(msg => "Using '$self->{tshark}' for the network capture");

    # Decide how we should fail

    my $failure = "warn";
    if ($self->{fatality}) {
        $failure = "fatal";
    }
 
    # Since it is difficult to cross reference the output from 'tshark -D' with
    # ipconfig to find the correct interface to use, we will just enable tracing
    # on all interfaces (1 - 4). No harm is really done here as the one on 
    # the mgmt network really won't be capturing anything.

    foreach my $interface (1,2,3,4) {

        # Set where the captures will be saved locally. Note that tshark appends a
        # timestamp onto the output file name. So even though we are specifying 
        # a particular format here, they will really look like: 
        # <given format>_<unique id>_<timestamp>.pcap.

	my $capture_file = $self->{temp_dir}.'\\'.$self->{id}.'_tshark_'.
                        $interface.'_'.$self->{address}.'.pcap';
       
        $Trace->comment(msg => "Using capture file: $capture_file");

        # Setting the following tshark options:
        # -i <interface>  Name or idx of interface (def: first non-loopback)
        # -f <filter>     Source IP address to filter on
        # -s <snaplen>    Packet snapshot length (def: 65535)
        # -w <outfile|->  Set the output filename (or '-' for stdout)
        # -q              Be more quiet on stdout (e.g. when using statistics)
        # -z <statistics> Various statistics, see the man page for details
        # -b filesize:NUM Switch to next file after NUM KB
        # -b files:NUM    Ringbuffer: replace after NUM files
        # -a duration:NUM Stop after NUM seconds

        my $tshark_opts = ' -i '.$interface.' -s '.$self->{snaplen}.' -w '.$capture_file.' -q -z smb,srt'.
                          ' -z http,tree -z io,phs -f "host ('.$self->{address}.')"';
        if ($self->{duration} > 0) {
            $tshark_opts .= ' -a duration:'.$self->{duration}.' ';
        }
        $tshark_opts .= ' -b filesize:'.$self->{size}.' -b files:'.$self->{files};
        
        # Start the process. Note that we are using the Win32::Process manager here
        # instead of subtest. This is because of the issues with subtest on windows
        # platforms and backwards compatibility with Nate2. 

        $Trace->comment(msg => "Starting tshark with: $self->{tshark} $tshark_opts");
        
        my $process;
        eval {
            Win32::Process::Create($process, $self->{tshark}, $tshark_opts, 0, 
                                   NORMAL_PRIORITY_CLASS, ".") || $self->_report_error_win();
            $Trace->debug(msg => Dumper(\$process));
        };
        if ($@) {

            $Trace->$failure(msg => "Encountered unexpected error while starting ".
                             "tshark: $@");

        } elsif (defined $process) {

            $Trace->comment(msg => "Packet Capturing started successfully");
            $Capture_Thread[$interface] = $process;

        } else {
            $Trace->$failure(msg => "Unable to start capture for '$capture_file' '$@'.");
        }
        $Trace->debug(msg => Dumper(\@Capture_Thread));
    }

    $Trace->exit();
}

# _report_error_win
#
# Report Win32 Process errors
#
sub _report_error_win {
    my $self = shift();
    my $failure = "warn";
    if ($self->{fatality}) {
        $failure = "fatal";
    }
    
    my $err = Win32::FormatMessage( Win32::GetLastError() );
    $Trace->$failure(msg => $err);
}

# _start_tshark_lin
#
# Start tshark on linux clients
#
sub _start_tshark_lin {

    $Trace->entry();
    my $self = shift();

    # If the location of tshark is still undefined, find it.

    if (!defined($self->{tshark})) {
        $self->{tshark} = `which tshark`;
        chomp($self->{tshark});
    }
    $Trace->comment(msg => "Using '$self->{tshark}' for the network capture");

    # Dump to the specified trace_file

    $self->{trace_file}    = $self->{temp_dir}."/".$self->{id}."_tshark_".
                             $self->{interface}."_".$self->{address};
    $Trace->comment(msg => "Network trace for $self->{address} is in $self->{trace_file}");

    # Check to see if a subtest exists and is running.

    if(!defined($self->{subtest}) || (!$self->{subtest}->is_alive())) {

        my $capture_file = "$self->{trace_file}.pcap";

        # Setting the following tshark options:
        # -i <interface>  Name or idx of interface (def: first non-loopback)
        # -s <snaplen>    Packet snapshot length (def: 65535)
        # -w <outfile|->  Set the output filename (or '-' for stdout)
        # -q              Be more quiet on stdout (e.g. when using statistics)
        # -z <statistics> Various statistics, see the man page for details
        # -b filesize:NUM Switch to next file after NUM KB
        # -b files:NUM    Ringbuffer: replace after NUM files
        # -a duration:NUM Stop after NUM seconds
        # src host <ip>   Filter option

        my $tshark_opts = "-i $self->{interface} -s $self->{snaplen} -w ".
                          "$capture_file -q -z smb,srt -z http,tree -z io,phs ".
                          "-b filesize:$self->{size} -b files:$self->{files}";
        if ($self->{duration} > 0) {
            $tshark_opts .= " -a duration:$self->{duration} ";
        }
        $tshark_opts .= " src host $self->{address}";
        
        my $host = hostname();
        my @splithost = split(/\./, $host);
        
        $Trace->comment(msg => "Starting tshark on ".$splithost[0]." with: $self->{sudo} $self->{tshark} $tshark_opts");

        # Set what we should do if tshark fails (warn or fatal).

        my $onexit;
        if ($self->{fatality}) {
            $onexit = sub {
                my $st = shift(); 
                if (scalar($st->own_results())) {
                    $Trace->fatal(msg => "Network trace subtest failed prematurely (fatal).");
                }
            }
        } else {
            $onexit = sub {
                my $st = shift(); 
                if (scalar($st->own_results())) {
                    $Trace->warn(msg => "Network trace subtest failed prematurely (warn).");
                }
            }
        }

        # Subtest the tshark process.
        $self->{subtest} = Tharn::subtest(
            $self->{sudo},
            -client => $splithost[0],
            -bg,
            -onexit, $onexit,
            -runid, "tshark_$self->{id}\_$self->{address}",
            "--",
            $self->{tshark},
            split(' ', $tshark_opts)
        );
        $Trace->debug(msg => "Network trace for $self->{address} started");        

    } else {
        $Trace->comment(msg => "Network trace for $self->{address} is already running.");
    }

    $Trace->exit();
}

#
# _find_trace_interface
#
# When TRACE_INTERFACE is set to 'find' this subroutine is called to find the client's
# interface that is part of the TRACE_NETWORK (data network or mgmt network). This 
# interface is where tracing will be enabled. 
#

sub _find_trace_interface {

    $Trace->entry();
    my $self = shift();

    # There are basically two networks that the client can be on. The data network
    # or the mgmt network. We will query the client for it's ifconfig output and 
    # then parse out the interfaces and addresses appropiately.

    my $ifconfig_bin    = `which ifconfig`;
    my $ifconfig_output = `$ifconfig_bin`;

    # The @interfaces array will hold each interfaces ifconfig output (split 
    # based on at least 2 new lines

    my @interfaces = split(/\n{2,}/, $ifconfig_output);

    # Loop through each interface to extract the interface name and IP address

    my %if_ip_map;
    my @tagged_ifs;
    foreach my $if_info (@interfaces) {

        # Continue if we've hit the loopback interface. We will not be enabling
        # tracing here.

        if ($if_info =~ /^lo\s+/) {
            $Trace->debug(msg => "Skipping the loopback interface.");
            next;
        }

        # Skip over tagged interfaces. There is not a good way for the
        # library to select a tagged interface as the data interface.
        # It is only responsible for the parent interfaces (i.e. eth2
        # and not eth2:100).

        if ($if_info =~ /^(eth\d+:\d+)/) {
            push(@tagged_ifs, $1);
            next;
        }

        # Parse out the ethernet interface name

        my $if;
        if ($if_info =~ /^(eth\d+)/) {
            $if = $1;
        }

        # Parse out the IPv4 address

        my $ip;
        if ($if_info =~ /inet addr:(\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3})\s/) {
            $ip = $1;
        }

        # Save the mapping

        $if_ip_map{$if} = $ip
    }

    # Warn if needed

    if (scalar(@tagged_ifs) >= 1) {
        $Trace->comment(msg => "Skipping the following tagged interfaces: ".
                        Dumper(\@tagged_ifs)."\nIf you would like to enable ".
                        "tracing on a tagged interface, please specify it directly ".
                        "to TRACE_INTERFACE.");
    }

    # Get the management IP from the nate host file
    my $host = hostname();
    my @splithost = split(/\./, $host);
    $host = host($splithost[0]);
    my $mgmt_ip = $host->{default_ip};

    # Now that we have built the mapping, we can pull out the interface that is
    # either on the data network or on the management network.

    my $trace_if = undef();
    if ($self->{network} eq "mgmt") {
        
        # Loop through and find the interface hosting the management IP

        foreach my $mgmt_if (keys(%if_ip_map)) {
            if ($if_ip_map{$mgmt_if} eq $mgmt_ip) {
                $trace_if = $mgmt_if;
            }
        }

    } elsif ($self->{network} eq "data") {

        # Loop through and find the interface hosting the IP that is not
        # the mgmt IP

        foreach my $data_if (keys(%if_ip_map)) {
            if ($if_ip_map{$data_if} ne $mgmt_ip) {
                $trace_if = $data_if;
            }
        }

    } else {
        $Trace->fatal(msg => "Unrecognized network: $self->{network}. Expected ".
                      "data or mgmt. Exiting.");
    }

    # Save the interface 

    if (!defined($trace_if)) {
        $Trace->fatal(msg => "Unable to find an up interface on the ".
                      "$self->{network}");
    }
    $self->{interface} = $trace_if;

    $Trace->exit(msg => "Found $self->{network} interface $self->{interface}.");
    return;

} ## end sub _find_trace_interface

=head2 cancel

Cancels an active trace.  Can also remove trace files.

=over

=item remove

(optional) : remove the accumulated trace file(s).

=back

=for html <hr>

=cut

sub cancel
{
    $Trace->entry();
    my $self = shift();
    my (%p) = validate_mixed(
        @_,
        [],
        {
            remove => {type => SCALAR, default => $self->{remove}},
        }
    );

    # Check and cancel the trace.
    if ($self->{subtest}->is_alive()) {
        $self->_stop_trace(remove => $p{remove});
    }

    $Trace->exit();
}

=head2 restart

Restarts an active trace.  Can also remove trace files.

=over

=item remove

(optional) : remove the accumulated trace file(s).

=back

=for html <hr>

=cut

sub restart
{
    $Trace->entry();
    my $self = shift();
    my (%p) = validate_mixed(
        @_,
        [],
        {
            remove => {type => SCALAR, default => $self->{remove}},
        }
    );
   
    # Stop and start the trace.
    $self->_stop_trace(remove => $p{remove});
    $self->_start_trace();

    $Trace->exit();
}

sub DESTROY
{
    $Trace->entry();
    my $self = shift();

    my $eval_error = $@;

    # Check to see if something bad has happened (i.e., $@ is set). If this is the
    # case, then we will not remove the files, no matter what the user set TRACE_REMOVE
    # to.
    if ($eval_error) {
        # if $@ is set, then don't remove the trace files, the may be needed to 
        # diganosing whatever caused the error.
        $Trace->debug(msg => "Unsetting remove flag since the eval error is set.");
        $self->{remove} = 0;
    }

    $self->_stop_trace(remove => $self->{remove});

    # calling _stop_trace() ends up calling kill() on a subtest. This seems to be causing the 
    #  value of $@ to change to the status of that subtest rather than what we had coming into
    #  this routine. So, set the $@ back to the original value.
    $@ = $eval_error;
    $Trace->exit();
}

#
# Class (static) methods
#

=head2 get_traces

Returns a list of references to trace objects for either the supplied list of 
addresses or the requested filer/interface(s)

This is a static method.

You must supply either a list of addresses or a filer and interface 
specification.

=over 

=item addresses 

(optional) : a reference to a list of addresses for which to create trace objects.

=item filers

(optional) : a reference to a list of filers

=item interfaces

(optional) : a mask showing which interfaces to use 
(i.e., Network::Trace::DATA|Network::Trace::CLUSTER)

=back

=for html <hr>

=cut

sub get_traces
{
    $Trace->entry();
    my (%p) = validate_mixed(
        @_,
        [],
        {
            addresses => {type => ARRAYREF | UNDEF, default => undef()},
            filers => {type => ARRAYREF | UNDEF, default => undef(), depends => ['interfaces']},
            interfaces => {type => SCALAR | UNDEF, default => undef(), depends => ['filers']}, 
        }
    );
    
    my %traces;
    my @traces;

    #
    # Look a the list of filers passed in, get each filer's host infomration 
    # and add the appropriate group of ports to the list of addresses to trace.
    #
    if (defined($p{filers})) {
        $Trace->comment(msg => "Filer and interface passed in, ignoring address list.");

        $p{addresses} = undef();

        foreach my $f (@{$p{filers}}) {
            my $h = host($f);
            if (DATA & $p{interfaces}) {
                $Trace->debug(msg => "Adding data interfaces for $f.");
                $p{interfaces} = $p{interfaces} & ~DATA;
                $traces{$h->{DATA_IP1}} = 1;
                $traces{$h->{DATA_IP2}} = 1;
            } 
            if (CLUSTER & $p{interfaces}) {
                $Trace->debug(msg => "Adding cluster interfaces for $f.");
                $p{interfaces} = $p{interfaces} & ~CLUSTER;
                $traces{$h->{CLUSTER_IP1}} = 1;
                $traces{$h->{CLUSTER_IP2}} = 1;
            } 
            if (MGMT & $p{interfaces}) {
                $Trace->debug(msg => "Adding management interfaces for $f.");
                $p{interfaces} = $p{interfaces} & ~MGMT;
                $traces{$h->{MGMT_IP}} = 1;
            }
            if ($p{interfaces}) {
                $Trace->fatal(msg => "Invalid interface specified: $p{interfaces}", result => 'SCRIPT');
            }
        }
    } else {
        # Weed out duplicate tracing requests.
        foreach my $addr (@{$p{addresses}}) {
            $traces{$addr} = 1;
        }
    }

    # 
    # Start all the requested traces.
    #
    $Trace->comment(msg => "Tracing addresses: ".join(", ", keys(%traces)));
    foreach my $address (keys(%traces)) {
        push(@traces, new Network::Trace(address => $address));
    }
    
    $Trace->exit();
    return \@traces;
}

=head1 PARAMETERS

=over

=item NETWORK_TRACE

Enable/disable all Network tracing.  This parameter allows an aplication to be
coded with the assumption that a network trace will always be taken, but won't
actually cause one.  In other words, this is a global parameter enabling or 
disabling this network trace.

=item TRACE_ID

The trace id.  This value is used to uniquify the trace output files.  It's 
safe to leave this as its default (which is the local hostname).

=item SUDO

The location of the sudo binary to use.

=item TCPDUMP

The location of the tcpdump binary to use.

=item USE_TCPDUMP

Tshark is the default trace capture that is used, however, the user can flag 
to use TCPDUMP instead.

=item TSHARK

The location of the tshark executable to use.

=item TRACE_FILTER

The filter to use when capturing data.  This parameter overrides the default
filter and probably shouldn't be used in a scalable environment.

=item TRACE_SNAPLEN

The amount of data to capture from each packet.  Default to 256 Bytes.

=item TRACE_INTERFACE

The interface to capture from. (default: eth1 for linux clients, all for 
windows clients). When this parameter is set to 'find' for linux clients, 
the library will go out and examine the client's interface configuration 
and cross reference that with the TRACE_NETWORK. The interface that is on 
the specified network will be the one used for tracing. This is useful when 
the TRACE_INTERFACE is different per client when running with more than one 
client.

=item TRACE_NETWORK

When TRACE_INTERFACE is set to 'find', the library will find the interface that
belongs to this network, and use that for the TRACE_INTERFACE. (default: data)

=item TRACE_FILES

The number of capture files to use in the rotating buffer. (default: 10)

=item TRACE_FILES_SIZE

The size of each file in the rotating buffer. (default: 10)

=item TRACE_FILES_OWNER

The owner of the trace files. (default: root)

=item TRACE_FATALITY

The fatality of a network tracing problem.

=item TRACE_ADDRS

A parameter for traceable addresses. Not used directly by this module

=item TRACE_REMOVE (default: 1)

A flag to control the removal of dump files, default is to delete the files

=item TRACE_DURATION 

Length of time to capture packets (default: until killed)

=item LOCAL_TEMP_DIR

Local directory on client to use as temp directory

=item TSHARK_WAIT

The amount of time (in seconds) to wait for tshark to finish on a windows client

=back

=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    This library only works semi-automatically for GX.

    There should be finer control of the filter and tcpdump options.

=head1 STATUS

    in-progress

=head1 ATTRIBUTES

    None

=head1 AUTHOR

    Steven Ewing (ewing@netapp.com)
    $Id: //depot/prod/test/Rsierranevada/lib/Network/Trace.pm#3 $
    Last revised $Date: 2013/04/11 $

=head1 COPYRIGHT

    Copyright 2006, Network Appliance Inc., All Rights Reserved.

=cut

#
# Empty package implementing disabled tracing.
#
package Network::EmptyTrace;

use base qw(Network::Trace);

sub new ()
{
    $Trace->entry();
    my $class = shift();
    my $self = [];
    bless($self, $class);
    $Trace->trace(msg => "Network trace disabled by parameter.");
    $Trace->exit();
    return $self;
}

sub cancel {}
sub restart {}
sub DESTROY {}

1;
