# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Ttcp.pm#1 $
#
## @summary Ttcp - Provide hardware independent access to ttcp
## @author	dl-platform-qa@netapp.com
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Ttcp - Provide hardware independent access to ttcp

=head1 SYNOPSIS

  use Network::Ttcp qw(ttcp_is_installed ttcp_transmitter ttcp_receiver);

  # Verify that a client has ttcp installed in the expected path
  if (!ttcp_is_installed(conn => $conn)) {
      die config_result('TTCP is not installed on client');
  }

  # Ttcp in receive mode
  $info = ttcp_receiver(conn => $conn);

  # Ttcp in transmit mode
  $info = ttcp_transmitter(conn => $conn, host => '10.0.2.50', count => 15000);

=head1 DESCRIPTION

Provides hardware independent access to ttcp.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-platform-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Network/Ttcp.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2008, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Ttcp;

use strict;
use warnings;
use base qw(Exporter);

use Carp qw(croak);
use SayWrap qw(sayEOF);
use Params::Validate qw(validate SCALAR OBJECT);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp( @msgs );
                         Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $VERSION = '1.0';
our @EXPORT_OK = qw(
    ttcp_is_installed
    ttcp_receiver
    ttcp_transmitter
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub ttcp_is_installed(%;);
sub ttcp_receiver(%;);
sub ttcp_transmitter(%;);
sub _get_ttcp_cmd($;);
sub _ttcp_send(%;);

#---------------------------------------------------------------------

=head2 ttcp_is_installed

  $ret = ttcp_is_installed(conn => $conn);

Test whether ttcp is installed on the host specified in
the connection.

=over

=item Arguments

=over

=item conn

(Required) : Connectrec object of client where ttcp will run.

=back

=item Returns

Returns 1 if the ttcp command is found on the host specified through
the connection.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ttcp_is_installed(%;) {
    my (%opts) = validate (
        @_,
        {
            conn       => { type => OBJECT },
        }
    );

    my ( $cmd, $out );
    my $hosttype = $opts{conn}->hostp->hosttype;

    $cmd = _get_ttcp_cmd($hosttype);
    $cmd .= ' -xXxX';

    # If ttcp is there then we expect to see a usage message
    $out = sayEOF($opts{conn}, $cmd);
    if ($out =~ /usage/i) {
        return ( 1 );
    } else {
        return ( 0 );
    }
}

#---------------------------------------------------------------------

=head2 ttcp_receiver

  $info = ttcp_receiver(conn => $conn,
                        [, size => $size ]
                        [, port => $port ]
                        [, proto => 'tcp'|'udp' ]
                        [, timeout => 3600 ]);

Starts ttcp in receive mode.

=over

=item Arguments

=over

=item conn

(Required) : Connectrec object of client where ttcp will run.

=item bind_retry

(Optional) : Number of times to retry the ttcp command if it returns
"bind: Address already in use" error.
Defaults to 5.

=item size

(Optional) : Length of buffers to use.
Defaults to 1024.

=item port

(Optional) : Port number to listen on.
Defaults to 5001.

=item af

(Optional) : address family(inet or inet6).
Defaults to "inet".

=item proto

(Optional) : Protocol to use ('tcp' or 'udp').
Defaults to 'tcp'.

=item timeout

(Optional) : Timeout to use for the say command that runs ttcp. This value
should be larger than the time you think it will take to run the ttcp test.
If the value is too small then say will timeout and an error returned before
the ttcp tests completes.
Defaults to 3600.

=back

=item Returns

A hash reference having the statistics of the ttcp run.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ttcp_receiver(%;) {
    my (%opts) = validate (
        @_,
        {
            bind_retry => { type => OBJECT, default => 5 },
            conn       => { type => OBJECT },
            size       => { type => SCALAR, default => 1024 },
            port       => { type => SCALAR, default => 5001 },
            af         => { type => SCALAR, default => "inet" },
            proto      => { type => SCALAR, default => 'tcp',
                            regex => qr/^(tcp|udp)$/ },
            timeout    => { type => SCALAR, default => 3600,
                            regex => qr/^(\d+)$/ },
        }
    );

    # Send command
    return _ttcp_send(%opts);
}

#---------------------------------------------------------------------

=head2 ttcp_transmitter

  $info = ttcp_transmitter(conn => $conn,
                           host => $host
                           [, count => $count ]
                           [, size => $size ]
                           [, port => $port ]
                           [, proto => 'tcp'|'udp' ]
                           [, obsize => $obsize ]
                           [, timeout => 3600 ]);

Starts ttcp in transmit mode.

=over

=item Arguments

=over

=item conn

(Required) : Connectrec object of client where ttcp will run.

=item host

(Required) : Hostname or IP address of host where ttcp is running
in receive mode.

=item bind_retry

(Optional) : Number of times to retry the ttcp command if it returns
"bind: Address already in use" or "connect: Connection refused" error.
Defaults to 10.

=item count

(Optional) : Number of source buffers to transmit.
Defaults to 2048.

=item size

(Optional) : Length of buffers to use.
Defaults to 1024.

=item port

(Optional) : Port number to send to.
Defaults to 5001.

=item proto

(Optional) : Protocol to use ('tcp' or 'udp').
Defaults to 'tcp'.

=item af

(Optional) : address family(inet or inet6).
Defaults to "inet".

=item obsize

(Optional) : Output buffer size in bytes.
This option is only supported on a filer.

=item timeout

(Optional) : Timeout to use for the say command that runs ttcp. This value
should be larger than the time you think it will take to run the ttcp test.
If the value is too small then say will timeout and an error returned before
the ttcp tests completes.
Defaults to 3600.

=back

=item Returns

A hash reference having the statistics of the ttcp run. The rate is
in 'M = mega" format.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub ttcp_transmitter(%;) {
    my (%opts) = validate (
        @_,
        {
            bind_retry => { type => OBJECT, default => 10 },
            conn       => { type => OBJECT },
            host       => { type => SCALAR },
            count      => { type => SCALAR, default => 2048 },
            size       => { type => SCALAR, default => 1024 },
            port       => { type => SCALAR, default => 5001 },
            af         => { type => SCALAR, default => 'inet' },
            proto      => { type => SCALAR, default => 'tcp',
                            regex => qr/^(tcp|udp)$/ },
            obsize     => { type => SCALAR, optional => 1 },
            timeout    => { type => SCALAR, default => 3600,
                            regex => qr/^(\d+)$/ },
        }
    );

    # Send command
    return _ttcp_send(%opts);
}

#---------------------------------------------------------------------
# Internal function
# Return the command path of ttcp for the type of host
#---------------------------------------------------------------------
sub _get_ttcp_cmd($;) {
    my ( $hosttype ) = @_;
    my $cmd;

    # Set correct command according to OS type
    if ( $hosttype =~ /filer/ ) {
        $cmd = "priv set -q diag; ttcp";
    } elsif ( $hosttype =~ /linux/ ) {
        $cmd = "/usr/bin/ttcp";
    } else {
        croak "TTCP not supported on Hosttype: $hosttype";
    }

    return ( $cmd );
}

#---------------------------------------------------------------------
# Internal function
# Send the proper ttcp command to the client/filer for both the
# receive or the transmit mode
#---------------------------------------------------------------------
sub _ttcp_send(%;) {
    my %opts = @_;
    my ($cmd, $out, $bytes, $sec, $rate);
    my $hosttype = $opts{conn}->hostp->hosttype;

    $cmd = _get_ttcp_cmd($hosttype);

    # Check valid options
    if ( $hosttype !~ /^filer/ && defined $opts{obsize} ) {
        croak "Option obsize is only supported on a filer";
    }

    # Use correct options for transmit and receive
    if ( defined $opts{count} ) {
        $cmd .= " -t -s -n $opts{count}";
        $cmd .= " -S $opts{obsize}" if defined $opts{obsize};
    } else {
        $cmd .= " -r -s";
    }

    # Add length of buffers to use
    $cmd .= " -l $opts{size}";

    # Add port to use
    $cmd .= " -p $opts{port}";

    # Add protocol to use
    $cmd .= " -u" if $opts{proto} ne 'tcp';

    if ( $hosttype =~ /filer/ ) {
        $cmd .= " -P inet6" if ($opts{af} =~ m/^inet6$/i);
    } elsif ( $hosttype =~ /linux/ ) { 
        $cmd .= " -6" if ($opts{af} =~ m/^inet6$/i);
    }

    if ( defined $opts{host} ) {
        # Add the host to connect to when on transmit mode
        $cmd .= " $opts{host}";
    }

    # Make sure bind_retry has a valid value;
    $opts{bind_retry} = 1 if ($opts{bind_retry} < 1);

    while ( $opts{bind_retry} > 0 ) {
        # Send command to filer/client
        $out = sayEOF($opts{conn}, $cmd, timeout => $opts{timeout});
        if ( $out =~
               /(bind: Address already in use|connect: Connection refused)/i ) {
            sleep(15);
            $opts{bind_retry}--;
        } else {
            $opts{bind_retry} = 0;
        }
    }

    # Process output of command
    if ( $hosttype =~ /filer/ ) {
        ($bytes, $sec, $rate) = ($out =~ m{(\d+) bytes processed.*\s(\d+) real sec =\s+(\d+) KB/real sec}s);
    } elsif ( $hosttype =~ /linux/ ) {
        ($bytes, $sec, $rate) = ($out =~ m{(\d+) bytes in (\S+) real seconds = (\S+) KB/sec}s);
    }

    if ( !defined $sec || !defined $rate ) {
        croak "Unable to process request:\n$out";
    }

    return {
        bytes   => $bytes,
        seconds => $sec,
        rate    => $rate,
    };
}

1;

