# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Utils.pm#2 $
#
# Copyright (c) 2004 Network Appliance Inc.
# All right reserved.
#
## @summary Routines for running network utilities
##
## @description
##         These functions are used to run various networking
##         utilities on devices independent of the device type.
##        These functions help in abstracting the test scripts
##        from subtle differences in command usage and command
##        outputs on various platforms.
##
## @dependencies
##
## use Tharn;
##
## @change 2004-10-05 gajria: fixed error match string in config_ipfw_*
## and added return eval.
##
## @change 2005-03-19 jcorbin: added function get_subnet.
##
## @author  dl-maryv (FSS Protocols QA)

package Network::Utils;

BEGIN {
require Exporter;
#use base qw(Exporter);
use vars qw (@ISA @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT_OK = qw (run_netstat
                 run_ping
                 config_ipfw_pipe
                 config_ipfw_rule
                 ipfw_enable
                 ipfw_disable
                 switch_get_type
                 run_getXXbyYY
                 get_nsswitch_conf
                 set_nsswitch_conf
                 get_subnet
                 fill_volume
                 setup_interfaces
                 do_ifconfig
                 do_ifconfig_v6
                 do_ifconfig_speed_duplex
                 get_ifconfig_info
                 get_mac_filter_entries
                 verify_ifconfig_info
                 get_ifstat_counter
                 parse_ifstat_counter
                 run_sio
                 switch_port_config
                 switch_intf_config
                 switch_exec
                 switch_config
                 get_nis_binding
                 switch_enable
                 vfiler_run_vfiler_setup_ipv6
                 tcp_udp_server_create
                 tcp_udp_client_create
                 check_comm_status
                 get_vserver_id
                 get_vserver_id_bsd
                 remove_config_interface
                 get_ifconfig_info_bsd
                 verify_ifconfig_info_bsd
                 run_BSD_ping
                 ndp
                 config_ipv4_rem_ipv6
            );
}
use strict;
use Tharn;
use Data::Dumper;
use Carp;
use SayWrap qw(sayEOF);
use Carp qw(croak);
use Network::IPv6 qw(_ipv6_address_is_configured __check_pattern);
use SPARSE::Util::Logger qw(logStep logDbg logTrc logWarn);
use filer qw(filer_if_netmask_get);
use RegistryCmds qw(registryGet registrySet);
use Params::Validate qw(validate_pos SCALAR OBJECT);
use Utils::Units qw(:constants);
use Socket;
use filer;
use Volume qw(vol_disk_free);
use Utils::Units qw(convert_units strip_unit_spec guess_unit);
use Server;
use Network::Routes qw(login_systemshell);
my $Server;

## @name ::run_netstat()
##
## @summary runs netstat on specified device to extract information
##
## @description
##
## this function takes a host object, a connection object
## mode and interface as arguments and runs the netstat cmd
## on the specified host. the mode arg can be "pktcount"
## or ...(TBD). 
## The function runs the netstat command
## with the required switches to extract network related
## information. The return value depends on the mode.
## In "pktcount" mode the return value is an array consisting
## of the in and out pkt count on the specified interface.
##
## returns hash containing "in" and "out" counts for "pktcount"
##
## @arg $conn  Required: connectspec of device
## @arg $mode  Required: describes type of info required
##                       "pktcount" - give in and out pkt count on iface 
## @arg %args  Optional: Hash of arguments
##             iface => $iface Required name of interface for "pktcount" mode
##
## @return for 'pktcount' mode returns hash containing
##          'in' => number of  packets received
##          'out' => number of packets sent
##          (TBD)
##

sub run_netstat ($$%) {

    my ($conn, $mode, %args) = @_;
    my %result;

    my $hosttype = $conn->hostp->hosttype();

    # packet count mode. netstat is run with "-i  <interface>"
    if ($mode eq "pktcount" ) {

    # make sure that interface has been provided
    my $iface = $args{'iface'}
        or die "Interface is needed for pktcount";

    if ( $hosttype =~ /linux/ ) {
        my (undef, undef, $line) = say ($conn,"netstat -i");
        if ( $line =~
         /$iface\s+\d+\s+(\d+)\s+\d+\s+\d+\s+(\d+)/s ){

        %result = ( 'in' => $1,
                'out' => $2,
              );
        #push  @result,$1;
        #push  @result,$2;
        }
    } elsif ( $hosttype =~ /solaris|filer/ ) {
        my (undef, undef, $line) = say ($conn,"netstat -I $iface");
        if ($line =~
        /$iface\s+\d+\s+\S+\s+\S+\s+(\d+)\s+\d+\s+(\d+)/s) {
        %result = ( 'in' => $1,
                'out' => $2,
              );
        #push @result,$1;
        #push @result,$2;
        }
    }
    } else {
    # other modes like sack, retransmissions etc
    # code to be added later
    }
    return (%result);
}

## @name ::run_ping()
## @summary Pings any ip address from any specifed device
##
## @description
##
## This function pings the specified ip address from any device. It hides
## platform specific differences in ping from the user script.
##
## @arg $conn  Required: connectspec of device
## @arg $destn Required: ip address to be pinged
## 
## @arg %args Optional: hash containing the following key = value pairs
##          confirm_only => 1 or 0 - return 1 if ping succeeds ; else return 0
##          count => <num> number of icmp messages to send
##          broadcast => 1 or 0 - Does a broadcast ping if set to 1 which
##              only works on Linux.
##          packet_size => Number of bytes to send in icmp message
##          interface => Interface to use for sending out the ping
##
## @return for 'confirm_only'= 1 return 0 for Failure, 1 for Success
##          else return entire output of ping command,

sub run_ping ($$%) {
    my ($conn,$destn,%args) = @_;
    my $confirm_only = $args{'confirm_only'};
    my $count = $args{'count'};
    my $interface = undef;	
    my $broadcast = $args{'broadcast'};
    my $packet_size = $args{'packet_size'};

    my $prepend ="";
    my @output;
    my $flags = "";
    my $hosttype = $conn->hostp->hosttype();
    my $pktt_size = "";

    my $cmd = 'ping';
    if ($hosttype =~ /filer/) {
        $prepend = "priv set -q advanced ;";
    } elsif ($hosttype =~ /solaris/) {
        # If $conn is a localhost connection on Solaris then path
        # is restricted for rsh and it will not find ping.
        $cmd = '/usr/sbin/ping';
    } elsif ($hosttype =~ /windows/) {
        $prepend = "cmd /c ";
    }

    # check variable args and set flags
    if ( $confirm_only == 1) {
        if ($hosttype =~ /filer/) {
            # The filer ping only allows count or packet size
            # so reset the packet_size.
            $packet_size = 1;
        } elsif ( $hosttype =~ /linux/ ) {
            $flags = "-c 1";
        } elsif ($hosttype =~ /windows/) {
            $flags = "-n 1";
        }
    } else {
        if (defined($count)) {
            if ($hosttype =~ /filer/) {
                # The filer ping only allows count or packet size
                # so reset the packet_size and let count have priority
                $packet_size = $count;
            } elsif ( $hosttype =~ /linux/ ) {
                $flags = "-c $count";
            }
            # packet count can be defined for solaris and filer
            # but the output is the same eg "1.1.1.1 is alive"
            # so being skipped here - gajria
        }
    }

    # packet size only works with linux machines, filers, and solaris
    # clients. If packet size is specified and machine is not supported,
    # ping will use the default packets size for that system.
    if (defined($packet_size)) {
        if ( $hosttype =~ /linux/ ) {
            $pktt_size = "-s $packet_size";
        } elsif ( $hosttype =~ /(filer|solaris)/) {
            $pktt_size = $packet_size;
        } elsif ($hosttype =~ /windows/) {
            $pktt_size = "-l $packet_size";
        }
    }

    # broadcast '-b' option is only for linux machines. Cannot
    # ping the broadcast address continuously. Say command will
    # timeout.
    if ( $broadcast == 1) {
        if ( $hosttype =~ /linux/ ) {
            if (defined($count)) {
                $flags .= " -b ";
            } else {
                $flags .= "-c 1 -b ";
            }
        }
    }

    if (defined($args{interface})) {
        if ( $hosttype =~ /linux/ ) {
            $interface = "-I $args{interface}";
        } elsif ( $hosttype =~ /(filer|solaris)/) {
            $interface = "-i $args{interface}";
        }
    } else {
            $interface = "";
    }

    # run ping with flags and return output to to array
     @output = say($conn,"$prepend $cmd $interface $flags $destn $pktt_size");

    # if confirm_only mode , return only 1 or 0
    # else return the entire buffer for processing outside

    if ( $confirm_only == 1 ) {
        if ($hosttype =~ /windows/) {
            if ( $output[2] =~ /Reply from/ ) {
                return (1);
            } else {
                return (0);
            }
        } else {
            if ( $output[2] =~ /is alive/ || $output[2] =~ /1 received/ ) {
                return (1);
            } else {
                return (0);
            }
        }
    } else {
        return $output[2];
    }
}



## @name ::config_ipfw_pipe()
##
## @summary Add/Delete an ipfw pipe on specified device
##
## @description
## This subroutine takes optional pipe config params as arguments and
## configures an ipfw pipe. If no arguments are provided, a pipe with
## default bandwidth is configured.
##
## @arg $conn    Required Connectspec of device
## @arg $cmd     Required 'add' or 'delete'
## @arg $pipe_id Required id to pipe to be added or deleted
## @arg %args    Optional Hash of optional arguments
##                        bw => pipe bandwidth configured in bits,Kbits,Mbits
##                              Bytes,KBytes or MBytes per sec
##                              default is 800Mbits/s
##                              NOTE: no space between value and unit 
##                                    eg 384Kbits/s
##                        plr =>  pipe packet loss rate as fraction of 1
##                              default is 0  
##                        delay => pipe delay in milliseconds
##                               default is 0.
##                        queue => num of slots or size in Bytes/KBytes
##                                 (max 1024 slots or 4096KBytes)
##                                default is 50 slots. 
##

sub config_ipfw_pipe ($$$%) {

    my ($conn, $cmd, $pipe_id, %args) = @_;

    my $bw = $args{'bw'};
    my $plr = $args{'plr'};
    my $delay = $args{'delay'};
    my $queue = $args{'queue'};

    my $bw_string = "";
    my $plr_string = "";
    my $queue_string = "";
    my $delay_string = "";

    my $prepend = "";
    my $line;

    # consider hosttype here !!

    my $hosttype = $conn->hostp->hosttype();

    ipfw_enable($conn);

    if ($hosttype =~ /filer/) {
    $prepend = "priv set -q test; "
    }

    return eval {

    if ( $cmd eq 'add' ) {
        if (defined($bw)) {
        
        #qualify bw argument per ipfw requirements
        if ( $bw !~ /\d+(M|m|K|k)?(bits|Bytes)\/s/ ) {
            die "Bandwidth is not in required format: $bw";
        }
        $bw_string = "bw $bw";
        } else {
        $bw_string = "bw 800Mbits/s";
        }
    
        # set value of plr delay and bw to specified argument or default.
        $plr_string = defined($plr) ? "plr $plr" : "";
        $queue_string = defined($queue) ? "queue $queue" : "";
        $delay_string = defined($delay) ? "delay $delay" : "";
    
        (undef,undef,$line)=_say_correctly($conn,"$prepend ipfw pipe $pipe_id config".
                    " $bw_string $plr_string".
                    " $queue_string $delay_string");
    } elsif ( $cmd eq 'delete' ) {
        (undef,undef,$line)=_say_correctly($conn,
                    "$prepend ipfw pipe $pipe_id delete");
    } else {
        die "Incorrect command: $cmd";
    }
    
    # added "setsockopt(IP_" to the output error string. 
    # This string is displayed if
    # the rule/pipe being deleted does not exist.
    
    if ( $line =~ /usage|error|Command not found|setsockopt\(IP_/) {
        # pipe config failed . Return Failure
        die "ipfw command returned error";
    } else {
      return 1;
    }
    };
}




## @name ::config_ipfw_rule()
##
## @summary Add/Delete ipfw rule on specified device
##
## @description
##
## This subroutine takes optional rule config params as arguments and
## configures an ipfw rule.
##
## @arg $conn    Required Connectspec of device
## @arg $cmd     Required 'add' or 'delete'
## @arg $rule_id Required id of rule to be added or deleted
## @arg %args    Required Hash of rule parameteres
##                        'action' => deny/permit
##                        'proto' => ip/tcp/udp etc
##                        'from_ip' =>  source ip address / any
##                        'from_port' => source port for tcp or udp
##                        'to_ip' => destination ip address / any
##                        'to_port' => destination port port for tcp or udp
##                        'pipe_id'=> id of configured pipe to add to rule
##

sub config_ipfw_rule ($$$%) {

    my ($conn, $cmd, $rule_id, %args) = @_;
    my $line;
    my $prepend ="";

    # consider hosttype here !!
    my $hosttype = $conn->hostp->hosttype();

    if ($hosttype =~ /filer/) {
    $prepend = "priv set -q test; "
    }

    ipfw_enable($conn);

    return eval {

    if ( $cmd eq 'add' ) {

        my $action = $args{'action'};
        my $proto = $args{'proto'};
        my $from_ip = $args{'from_ip'};
        my $from_port = $args{'from_port'};
        my $to_ip = $args{'to_ip'};
        my $to_port = $args{'to_port'};
        my $pipe_id = $args{'pipe_id'};


        if (defined($action)) {

        # set values to passed arguments or defaults.
        $proto = defined($proto) ? $proto : "all";
        $from_ip = defined($from_ip) ? $from_ip : "any";
        $from_port = defined($from_port) ? $from_port : "";
        $to_ip = defined($to_ip) ? $to_ip : "any";
        $to_port = defined($to_port) ? $to_port : "";

        # this is a deny/permit rule config
        (undef, undef, $line) = _say_correctly($conn,"$prepend ipfw add $rule_id".
                        " $action $proto from".
                        " $from_ip $from_port".
                        " to $to_ip $to_port");

        } elsif (defined($pipe_id)) {
        # this is a pipe rule config
        (undef, undef, $line) = _say_correctly($conn,"$prepend ipfw add $rule_id".
                        " pipe $pipe_id $proto".
                        " from $from_ip $from_port".
                        " to $to_ip $to_port");
        } else {
        die "Neither Action nor pipe id were defined";
        }

    } elsif ( $cmd eq 'delete' ) {
        (undef, undef, $line) = _say_correctly($conn,"$prepend ipfw delete $rule_id");
    } else {
        die "Incorrect command: $cmd";
    }

    # added "setsockopt(IP_" to the output error string. 
    # This string is displayed if the rule/pipe being deleted 
    # does not exist.

    if ( $line =~ /usage|error|Command not found|setsockopt\(IP_/ ) {
        # pipe config failed . Return 0
        die "ipfw command returned error";
    } else {
      return 1;
    }

    };
}

## @name :: switch_get_type()
##
## @summary Get Type of Switch. Currently checks for Cisco/Extreme
##
## @description
##
## @arg $conn    Required Connectspec of device
##
## @return "CISCO" if switch is a Cisco CAT 6xxx  switch
##         "EXTREME" if switch if Extreme Black Diamond
##         0 - if switch is neither of above.
##

sub switch_get_type( $ ) {

    my $conn = shift;

    say($conn,"set length 0");
    my $output = (say($conn,"show version"))[2];

    if ( $output =~ /cisco/i ) {
    return "CISCO";
    } elsif ( $output =~ /extreme/i ) {
    return "EXTREME";
    } else {
    # Unknown switch type
    return 0;
    }
}


## @name :: run_getXXbyYY()
##
## @summary run getXXbyYY command with proper sub command
##
## @description
## This subroutine runs the getXXbyYY command with the provided 
## subcommand
## and value.
##
## @arg $conn    Required Connectspec of device
## @arg $command Required Sub command for getXXbyYY eg gethostbyname_r
## @arg $name    Required name to lookup
## @args %options Optional.
##       vfiler_name => 'vfiler_name', Defaults to 'undef'.
##		    			 Used to run getXXbyYY on the vfiler.
##
## @return output of getXXbyYY command
## 
## @note croaks on error
##

sub run_getXXbyYY ($$$;%) {
  
	my ($conn,$command,$name,%options) = @_;
	my $vfiler_name = $options{'vfiler_name'} || undef;
	my $output;

	if ($vfiler_name) {
		$output = sayEOF($conn,"priv set -q diag ;".
		"vfiler run $vfiler_name getXXbyYY $command $name");
		if ($output =~ m/vfiler not found/i) {
			croak ("vfiler $vfiler_name does not exist");
		}
	} else {
		$output = sayEOF($conn,"priv set -q diag ;".
		"getXXbyYY $command $name");
	}

	if ($output =~ m/unknown command/s) {
		croak "Command $command is unknown";
	}

	return($output);
}


use constant OPTIONS_PREFIX => "options.nsswitch";

##
## @name ::set_nsswitch_conf
##
## @summary sets the registry values for parameters in /etc/nsswitch.conf
## If no parameter is passed, sets values for all parameters.
## Returns 0 if any error is encountered or returns 1 upon success.
##
## @arg $fc - required. filer connection object
## @arg %nsswitch_params - Optional. list of parameters to be changed;
##
## example: set_nsswitch_conf($fc,  passwd => 'nis', hosts => 'nis');
##
sub set_nsswitch_conf
{
    my ($fc, %nsswitch_values) = @_;

    my %nsswitch_default_values = ( hosts => "files    nis    dns",
				    passwd => "files    nis    ldap",
				    netgroup => "files nis ldap",
				    group => "files nis ldap",
				    shadow => "files nis" );


    # If list of parameters to be changed is not passed, lets use all params
    if ( !%nsswitch_values ) {
        logDbg("nsswitch_values not defined, settnig default values");
        %nsswitch_values = %nsswitch_default_values;
    }
    # If value is not valid, return fail.
    foreach my $service ( keys %nsswitch_values ) {
        if ( $service !~ m/(hosts|passwd|group|netgroup|shadow)/ ) {
            logStep("set_nsswitch_conf: $service is not one of ".
                    "expected values");
            return 0
        }
    }

    # Set the registry values .  This modifies /etc/passwd.
    foreach my $service (keys %nsswitch_values) {
        logDbg("Modifying registry values for $service in /etc/nsswitch.conf ");
        registrySet($fc, OPTIONS_PREFIX . ".$service", 
            $nsswitch_values{$service}) or return 0;
    }

    my %nsswitch_new_values = get_nsswitch_conf($fc);

    foreach my $service (keys %nsswitch_new_values) {
        if ( $nsswitch_new_values{$service} !~ /^$nsswitch_values{$service}/ ) {
            logStep("set_nsswitch_conf: value for $service is not correct");
            return 0;
        }
    }
    logDbg("registry modified for entries in /etc/nsswitch.conf");
    return 1;
}

##
## @name ::get_nsswitch_conf
##
## @summary gets the registry values for parameters in /etc/nsswitch.conf
## If no parameter is passed, gets values for all parameters.
## Returns 0 if any error is encountered or returns all paramets and values
## in a hash.
##
## @arg  $fc - required. filer connection object
## @arg  @nsswitch_params - Optional. list of parameters to be changed;
##
##     example: %nsswitch_values = get_nsswitch_conf($fc, passwd, hosts);
##
sub get_nsswitch_conf
{
    my ($fc, @nsswitch_params) = @_;
    my @nsswitch_all_params = ( "hosts", "passwd", "netgroup", "group",
         "shadow" );
    my %nsswitch_lines = ();

    # If lilst of parameters not passwd, use all the parameters.
    if ( !@nsswitch_params) {
        logDbg("get_nsswitch_conf: getting values for all parameters");
        @nsswitch_params =  @nsswitch_all_params;
    }

    # Retrieve the registry values for the list of parameters.
    foreach my $param (@nsswitch_params) {
        logTrc("Retrieving registry values for $param in /etc/nsswitch.conf ");
        $nsswitch_lines{$param} = registryGet($fc, OPTIONS_PREFIX . ".$param") 
            or return 0;
    }

    return %nsswitch_lines;
}

# name: get_if_netmask_linux
# summary:
#    Return host netmask. Specific to Linux.
#
# arg $cc    Required: Host connection object for the command line interface
# arg $ipaddr    Required: IP address associated with interface to get
#        netmask of.
#
# return netmask in format xxx.xxx.xxx.xxx/yy or croaks if netmask not found
#
# notes IP V4 only
sub get_if_netmask_linux ( $$; ) {
    my ($cc, $ipaddr) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR }
    );
    my $netmask;
    my $cmd = "/sbin/ifconfig -a";

    my(undef, undef, $result) = say($cc, $cmd, '-swallow', '-eval_error');
    croak $@ if $@;

    my @info = split(/\n/, $result);
    foreach my $line (@info) {
        # Only process output associated with the given IP address
        if ($line =~ /$ipaddr/) {
            # inet addr:192.168.1.3  Bcast:192.168.1.255  Mask:255.255.255.0
            if ($line =~ /\sMask:(\d+\.\d+\.\d+\.\d+)/) {
                my $mask = $1;
                my $netmask_bin = inet_aton($mask);
                # Convert netmask to binary form "1111...10000"
                my $netmask_bits = unpack('B32', $netmask_bin);
                # Append subnet bits
                $netmask = "$mask/" . ($netmask_bits =~ tr/1/1/);
                last;
            }
        }
    }
    if (!defined($netmask)) {
        croak "get_if_netmask_linux: netmask not found for ipaddr: $ipaddr";
    }
    return $netmask;
}

# name ::get_if_netmask_solaris
#
# summary
#    Return host netmask. Specific to Solaris.
#
# arg $cc    Required: Host connection object for the command line interface
# arg $ipaddr    Required: IP address associated with interface to get
#        netmask of.
#
# return netmask in format xxx.xxx.xxx.xxx/yy or croaks if netmask not found
#
# notes IP V4 only
sub get_if_netmask_solaris ( $$; ) {
    my ($cc, $ipaddr) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR }
    );
    my $netmask;
    my $cmd = "/sbin/ifconfig -a";

    my(undef, undef, $result) = say($cc, $cmd, '-swallow', '-eval_error');
    croak $@ if $@;

    my @info = split(/\n/, $result);
    foreach my $line (@info) {
        # Only process output associated with the given IP address
        if ($line =~ / $ipaddr /) {
            # inet 10.60.153.24 netmask 0xffffff00 broadcast 10.60.153.255
            if ($line =~ /\snetmask\s(........)\sbroadcast/) {
                # convert netmask from hex string to binary rep to ascii form
                my $netmask_bin = pack('H8', $1);
                my $mask = inet_ntoa($netmask_bin);
                # Convert netmask to binary form "1111...10000"
                my $netmask_bits = unpack('B32', $netmask_bin);
                # Append subnet bits
                $netmask = "$mask/" . ($netmask_bits =~ tr/1/1/);
                last;
            }
        }
    }
    if (!defined($netmask)) {
        croak "get_if_netmask_solaris: netmask not found for ipaddr: $ipaddr";
    }
    return $netmask;
}

## @name ::get_subnet
##
## @description
##  Get the subnet for the given IP address. Returns the
##  subnet associated with the IP address.
##
## @synopsis
##  $subnet = get_subnet($cc, $ipaddr);
##
## @arg $cc
##  Required: Host connection object.
##
## @arg $ipaddr
##  Required: IP address in the ascii form xxx.xxx.xxx.xxx
##
## @return
##  subnet in format xxx.xxx.xxx.xxx/yy returns,
##  undef if the subnet could not be determined.
##
## @notes
## 1) IP V4 only
## 2) Only works for filer, Solaris, and Linux.
sub get_subnet( $$; ) {
    my ($cc, $ipaddr_ascii) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR }
    );
    my ($htype, $ipaddr, $netmask, $netmask_len, $subnet);

    # Get the ip address associated with the connection
    $ipaddr = inet_aton($ipaddr_ascii);

    # Get the netmask for the host
    $htype = $cc->hostp()->hosttype();
    if ($htype =~ /filer/) {
        $netmask = filer_if_netmask_get($cc, $ipaddr_ascii);
    } elsif ($htype =~ 'linux') {
        $netmask = get_if_netmask_linux($cc, $ipaddr_ascii);
    } elsif ($htype =~ 'solaris') {
        $netmask = get_if_netmask_solaris($cc, $ipaddr_ascii);
    } else {
        croak("get_subnet: Don't know how to get subnet for hosttype: $htype");
    }

    # At this point $netmask is in the form "xxx.xxx.xxx.xxx/nn"
    ($netmask, $netmask_len) = split('/', $netmask);

    # Continue processing if the netmask is good
    if ($netmask_len > 0 and $netmask_len < 32) {
        my $host_len = (32 - $netmask_len);    # Number of host bits

        # Convert netmask bits to binary representation
        my $netmask_bin = pack('B32', '1' x $netmask_len . '0' x $host_len);

        # Apply netmask to host IP address to get host subnet
        $subnet = inet_ntoa($ipaddr & $netmask_bin) . "/$netmask_len";
    }
    return $subnet;
}


## @name ::fill_volume
##
## @description
##  fill a volume to desired percentage of total capacity or upto desired 
## remainder space
##  
## @synopsis
##   fill_volume($fc, %options);
##         %options = (volume => <volume>,
##                     directory => <directory>,
##                     percentage => <percentage>,
##                     remainder => <remainder space>,
##                     file_size => <file size>, 
##                     );
##
## @arg $fc Required: Host connection object.
##
## @arg %options Required 
##               volume => name of volume to fill
##               directory => name of optional directory under volume to fill
##               percentage => Optional Percentage of capacity to fill to
##               remainder => Optional remaining space to fill upto
##               file_size => Optional file size 
##
##
## @return 
##      1 on success, croaks on failure 
##
## @notes Either percentage or remainder space should be defined. If directory
## is specified, it should exist under the volume.
## If The filer conn object is not of type serial(console) or telnet, then
## the a new conn object of type telnet will be created because unsparse is 
## available only through telnet or console (not rsh)
##


sub fill_volume($;%) {

    my ($conn,%options) = @_;

    logcomment(Dumper(\%options));

    my $volume = $options{'volume'};
    my $directory = $options{'directory'};
    my $remainder = $options{'remainder'};
    my $percentage = $options{'percentage'}; 
    my $file_size = $options{'file_size'} || "10m";
    my $num_files;

    # process $conn, We need to use either telnet or console for 
    # unsparse command
    # It is not available via rsh.
    # If $conn is rsh, then create a new one using telnet
    if ($conn->conntype() !~ /telnet|serial/) {
	my $connh = host $conn->hostp()->hostname();
	$conn = connect("#telnet",-timeout,600);
    }

    if (!$volume) {
	die "Volume not specified";
    }
    if (!$percentage and !$remainder) {
	die "Either remainder or percentage should be specified";
    }
    if ($percentage > 100) {
	die " invalid value of percentage";
    }

    my $path;
    if ($directory) {
	$path = "/vol/$volume/$directory";
    }else {
	$path = "/vol/$volume";
    }

    my $file_name = "file".time();
    my $file_ext = 0;
    my ($total,$available,$capacity,$output);

    # get vol free space
    ($total,undef,$available) = vol_disk_free($conn, 
					      path => "/vol/$volume")
		or croak "vol disk free failed $@";
    # if percentage is specified convert to size units
    if (!$remainder) {
	$remainder = $total - int($total * $percentage/100);
	$remainder .= "k";
    }

    # convert units to KB
    $remainder = convert_units(strip_unit_spec($remainder),
			       guess_unit($remainder),UNIT_KB);
    $file_size = convert_units(strip_unit_spec($file_size),
			       guess_unit($file_size),UNIT_KB);

    # return if volume is already full to desired level
    if ($available < $remainder) {
	return 1;
    }

    sayEOF($conn,"priv set test");
    logTrc("Available = $available KB, Remainder = $remainder KB");
    sleep 4;

    while ( $available > $remainder) {

	my $free_space_to_fill = $available - $remainder;
	#throttle free_space_to_fill to 90% its size
	$free_space_to_fill = int($free_space_to_fill*90/100);
	# in case the value is rounded to 0, make it 1 KB
	if ($free_space_to_fill == 0) {
	    $free_space_to_fill = 1;
	}
	logTrc("Throttled Free space = $free_space_to_fill KB");
	sleep 5;
	if ( $free_space_to_fill < $file_size) {
	    logTrc("Creating 1 file of size $free_space_to_fill KB");
	    $output = sayEOF($conn,
			     "unsparse -v -l $free_space_to_fill"."k".
			     " -d $path $file_name.$file_ext");
	    if ($output =~ m/No such file or directory/i) {
		croak "file creation failed. Missing volume or".
		    " directory: $output";
	    }
		if ($output =~ m/unsparse not found/i) {
		croak "unsparse command not found - $output";
		}
	    ++$file_ext;
	} else {
	    $num_files = int($free_space_to_fill/$file_size);
	    logTrc("Number of files being created = $num_files");
	    sleep 5;
	    for (1..$num_files) {
		$output = sayEOF($conn,"unsparse -v -l $file_size"."k".
				 " -d $path $file_name.$file_ext");
		if ($output =~ m/No such file or directory/i) {
		    croak "file creation failed. Missing volume or".
			" directory: $output";
		}
		if ($output =~ m/There is not enough free space/i) {
		    last;
		}
		++$file_ext;
	    }
	}
	sleep 30;
	# get vol free space
	(undef,undef,$available,$capacity) = 
	    vol_disk_free($conn, path => "/vol/$volume")
		or die $@;
	logTrc("Available = $available");

    }
    return 1;
}

## @name ::setup_interfaces
##
## @arg 
## $fc - filer connection object
## $cc - client connection object 
## %args - hash of the following variables
## ip_addr - starting ip address
## filer_ifc - string of filer interfaces, ex: "e3a,e3b,e3c"
## client_ifc - string of client interfaces, ex: "hme1,hme2";
##
## @description 
## 
## invokes do_ifconfig() on every interface with an ip address.
## uses the host portion of the starting address to compute subsequent
## addresses.
##
## @return 
## list of references to hash. ifconfig_info_filer and ifconfig_info_client

sub setup_interfaces
{
    my ($fc, $cc, %args)  = @_;

    my @filer_ifc = split(/,/, $args{filer_ifc});
    my @client_ifc = split(/,/, $args{client_ifc});

    my $ip_addr = $args{ip_addr};

    foreach my $ifc ( @filer_ifc ) {
        logDbg("Configuring $ifc with ipaddr $ip_addr");

        logDbg("Configuring filer interface $ifc with $ip_addr");
        do_ifconfig($fc, $ifc, ip_addr => $ip_addr) or
            die $@;

        # Get next ipaddr by incrementing the host portion.
        $ip_addr =~ m/((\d+.){3})(\d+)/;
        my $network_part = $1;
        my $host_part = $3 + 1;
        $ip_addr = "$network_part$host_part";
    }

    logStep("Configuring the client side interfaces starting with ".
		"ipaddr $ip_addr");
    foreach my $ifc ( @client_ifc ) {

        logDbg("Configuring filer interface $ifc with $ip_addr");
    	do_ifconfig($cc, $ifc,  ip_addr => $ip_addr) or
        	die $@;

        # Get next ipaddr by incrementing the host portion.
        $ip_addr =~ m/((\d+.){3})(\d+)/;
        my $network_part = $1;
        my $host_part = $3 + 1;
        $ip_addr = "$network_part$host_part";

	}

    my $ifconfig_info_filer = get_ifconfig_info($fc) or
        die $@;
    my $ifconfig_info_client = get_ifconfig_info($cc) or
        die $@;

    return($ifconfig_info_filer, $ifconfig_info_client );

}




##
## @name :: do_ifconfig
## 
## @description invokes ifconfig on the specified connection.  It chooses
## appropriate ifconfig based on the hosttype. 
##
## @arg $conn - connection object
## @arg $interface - name of the interface
## @arg %options - hash of various options
##            "ip_addr"    - 'ip address to configure'
##            "alias"      - 'ip address to configure as alias'
##            "flowcontrol"- 'flowcontrol setting none|receive|send|full'
##            "netmask"    - 'netmask to configure for address'
##            "media_type" - 'media type to set to, if it contains the word
##                            auto then it is set to auto'
##            "mtusize"    - 'mtusize to configure for interface'
##            "state"      - 'up' or 'down' (state is also overloaded for alias
##                            addition and deletion)
##
## @return 1 upon success, if failures croak().
## example: do_ifconfig($fc, 'e3a', 
##                      ip_addr => '172.16.100.101',
##                      state => 'up',
##			mtusize => '1500',
##                      media_type => 'auto-1000t-fd-up',
##                      alias => '1.1.1.1',
##                      netmask => '255.255.255.0');
##
## @note If alias is specified without an ip address then the function will 
## croak
## if the interface does not already have an ip address. If both ip_addr and
## alias options are specified, the function will configure the ip_Addr first 
## and then configure the alias
##
##
##

use constant IFCONFIG_DEFAULT_FLAGS => "<BROADCAST,RUNNING,MULTICAST>";

sub do_ifconfig
{
    my ($conn, $interface, %options) = @_;
    my $hostp = $conn->hostp;
    my $hostname = $hostp->hostname;
    my $ifconfig_cmd = undef;
    my $ipaddr = $options{ip_addr};
    my $netmask = $options{'netmask'} || "255.255.255.0";
logcomment("inside do_ifconfig");
    my $ifconfig_cmd = get_ifconfig_path($conn) or
        croak "can't find ifconfig path";

    if (defined($options{flowcontrol}) &&
        $options{flowcontrol} !~ /^(none|receive|send|full)$/) {
        croak "$options{flowcontrol} is invalid value for flowcontrol";
    }

    # If state is down, invoke ifconfig <ifc> down
    if ( defined($options{state})) {
	if ($options{state} eq 'down'
	    && $ipaddr ne '0.0.0.0'
	    && !defined($options{'alias'})) {
            logDbg(" Inside state down and some IP ($ipaddr), action: down the interface");
	    sayEOF($conn, "$ifconfig_cmd $interface down");
            logcomment(" Where am I: Inside state down and some IP ($ipaddr), action: down the interface");
	    say($conn, "$ifconfig_cmd $interface down");
            logDbg(" Where am I: Inside state down and some IP ($ipaddr), action: down the interface");
	    sayEOF($conn, "$ifconfig_cmd $interface down");
	    verify_ifconfig_info($conn,  $interface, 
				 flags => 'DOWN' )
		and croak "verify_ifconfig_info failed - $@";
	    return 1;
	} elsif ($options{state} eq 'up' && !defined($ipaddr)
		&& !defined($options{'alias'})) {
            logDbg("Where am I: Inside state UP but no IP ($ipaddr), action: Up the interface");
	    my $cmd_out = sayEOF($conn, "$ifconfig_cmd $interface up");
	    verify_ifconfig_info($conn,  $interface,
				 flags => 'UP' )
		or croak "verify_ifconfig_info failed - $cmd_out";
	}
    }
    if (defined($ipaddr)) {
	# If ip_addr is specified, then set the ip addr on the interface
	logDbg("Configuring interface $interface with ip address $ipaddr ".
	       "on $hostname");
	my $cmd_str = undef;
	if ( $ipaddr eq "0.0.0.0" ) {
            # Set state to whatever caller specified defaulting to down
            my $state;
            if (defined($options{state})) {
                $state = $options{state};
                logDbg("State check: IP addr is ZERO, but state is defined ($state)");
            } else {
                $state = 'down';
                logDbg("State check: IP addr is ZERO, but state is not defined ($state)");
            }
            logDbg("Inside if IP addr ZERO, executing the following command:");
	    my (undef,undef,$cmd_out) = say($conn,"$ifconfig_cmd $interface $ipaddr $state");  #Change: removed $state at the end -Karthik
	    verify_ifconfig_info($conn,  $interface, ip_addr => "0.0.0.0")
		or croak "verify_ifconfig_info failed - $cmd_out";
	} else {
	    $cmd_str = 
		" $interface $ipaddr netmask $netmask $options{state}";  #Change:$options{'state'} to $options{state}
            logDbg("State Check inside IF ip addr,$options{'state'}"); 	
	    logDbg("invoking the command: $ifconfig_cmd $cmd_str");
	    my $cmd_out = sayEOF($conn, "$ifconfig_cmd $cmd_str");
	    if ( $cmd_out =~ m/\w/ ) {# cmd is not expected to print any output.
	        #logWarn("ifconfig printed some output: $cmd_out");
	    }
	    verify_ifconfig_info( $conn, $interface,ip_addr => $ipaddr,
			          netmask => $netmask )
	        or croak "verify_ifconfig_info failed - $cmd_out";
	}
    }

    my $cmd_str = undef;
    # If alias is specified, configure alias
    if ( defined($options{alias})) {
	# assume state is up if undefined
	if (!defined($options{'state'}) or $options{'state'} =~ /^up/i) {
	    $cmd_str =
		sayEOF($conn, 
		       "$ifconfig_cmd $interface alias $options{'alias'}".
		       " netmask $netmask");
	    verify_ifconfig_info( $conn, $interface,
				  alias => $options{'alias'})
		or croak "verify_ifconfig_info failed";
	} elsif ($options{'state'} =~ /^down/i) {
	    $cmd_str = 
		my $cmd_out = sayEOF($conn,
				     "$ifconfig_cmd $interface -alias".
				     " $options{'alias'}");
	    verify_ifconfig_info( $conn, $interface,alias => $options{'alias'})
		and croak "verify_ifconfig_info failed - $cmd_out";
	}
    }

    # get_ifconfig_info() uses the key 'mtu' which doesn't match the param
    # mtusize in do_ifconfig() so if 'mtu' is set then use it.
    $options{mtusize} = $options{mtu} if (defined($options{mtu}));

    # If mtusize is specified, configure mtusize
    if (defined($options{mtusize})) {
        # Set the MTU option name based on the hosttype.
        my $mtu_opt;
	if ( $hostp->hosttype =~ /(solaris|linux)/ ) {
	    $mtu_opt = 'mtu';
	} else {
	    # Filer and everybody else
	    $mtu_opt = 'mtusize';
	}
    	$cmd_str = sayEOF($conn,
		"$ifconfig_cmd $interface $mtu_opt $options{'mtusize'}");

        # Return the error on failure to allow tests to see if the MTU
        # is out of the valid range and to see what the valid range is.
	croak "do_ifconfig failed - $cmd_str" if ($cmd_str =~ /ifconfig\:/i);

        # Sleep to give interface time to start switch negotiations
        sleep(10);
    }

    # If media_type is specified, configure media_type only on the filer
    if (defined($options{media_type}) && ($hostp->hosttype =~ /filer/)) {
	$cmd_str = "mediatype $options{'media_type'}";
	my $cmd_out = sayEOF($conn, "$ifconfig_cmd $interface $cmd_str");
        if ($cmd_out =~ /no such media type/i) {
	    croak "Filer ifconfig does not support mediatype " .
                  "$options{'media_type'}";
        }

	# Changing media type can take a little while to take effect so
        # poll for up to 60 seconds to do the verification.
        my $retry = 12;
        while ($retry > 0) {
	    sleep(5);
	    if (verify_ifconfig_info($conn, $interface,
			             media_type => $options{'media_type'})) {
		$retry = -1;
            } else {
		$retry--;
	    }
	}
	croak "verify_ifconfig_info failed setting mediatype " .
	      "$options{media_type}" if ($retry == 0);
    }

    # If flowcontrol is specified, configure flowcontrol
    if (defined($options{flowcontrol})) {
	$cmd_str = "flowcontrol $options{'flowcontrol'}";
	my $cmd_out = sayEOF($conn, "$ifconfig_cmd $interface $cmd_str");
	verify_ifconfig_info($conn, $interface,
			     flowcontrol => $options{'flowcontrol'})
	    or croak "verify_ifconfig_info failed - $cmd_out";
    }

    # Some versions of Linux have a wierd problem where if one interface
    # starts negotiating with a switch then all interfaces are temporarily
    # stopped until negotiations are done (or further along). To work around
    # this sleep here if it is a linux client to let things settle down.
    sleep(20) if ( $hostp->hosttype =~ /linux/ );

    return 1;
}

sub do_ifconfig_speed_duplex          #section added by Rity.
    {
    my ($conn, $interface, %options) = @_;
    my $hostp = $conn->hostp;
    my $hostname = $hostp->hostname;
    my $ifconfig_cmd = undef;
    my $ipaddr = $options{ip_addr};
    my $netmask = $options{'netmask'} || "255.255.255.0";

    my $ifconfig_cmd = get_ifconfig_path($conn) or
        croak "can't find ifconfig path";

    if (defined($options{flowcontrol}) &&
        $options{flowcontrol} !~ /^(none|receive|send|full)$/) {
        croak "$options{flowcontrol} is invalid value for flowcontrol";
    }

    # If state is down, invoke ifconfig <ifc> down
    if ( defined($options{state})) {
        if ($options{state} eq 'down'
            && $ipaddr ne '0.0.0.0'
            && !defined($options{'alias'})) {
            sayEOF($conn, "$ifconfig_cmd $interface down");

            verify_ifconfig_info($conn,  $interface,
                                 flags => 'DOWN' )
                and croak "verify_ifconfig_info failed - $@";

            return 1;
        } elsif ($options{state} eq 'up' && !defined($ipaddr)
                && !defined($options{'alias'})) {
            my $cmd_out = sayEOF($conn, "$ifconfig_cmd $interface up");
            verify_ifconfig_info($conn,  $interface,
                                 flags => 'UP' )
                or croak "verify_ifconfig_info failed - $cmd_out";
        }
    }

    if (defined($ipaddr)) {
        # If ip_addr is specified, then set the ip addr on the interface
        logDbg("Configuring interface $interface with ip address $ipaddr ".
               "on $hostname");

        my $cmd_str = undef;
        if ( $ipaddr eq "0.0.0.0" ) {
            # Set state to whatever caller specified defaulting to down
            my $state;
            if (defined($options{state})) {
                $state = $options{state};
            } else {
                $state = 'down';
            }
            my $cmd_out = sayEOF($conn,"$ifconfig_cmd $interface $ipaddr $state");

            verify_ifconfig_info($conn,  $interface, ip_addr => "0.0.0.0")
                or croak "verify_ifconfig_info failed - $cmd_out";
        } else {
            $cmd_str =
                " $interface $ipaddr netmask $netmask $options{'state'}";

            logDbg("invoking the command: $ifconfig_cmd $cmd_str");
            my $cmd_out = sayEOF($conn, "$ifconfig_cmd $cmd_str");
            if ( $cmd_out =~ m/\w/ ) {# cmd is not expected to print any output.
                logWarn("ifconfig printed some output: $cmd_out");
            }

            verify_ifconfig_info( $conn, $interface,ip_addr => $ipaddr,
                                  netmask => $netmask )
                or croak "verify_ifconfig_info failed - $cmd_out";
        }
    }

    my $cmd_str = undef;
    # If alias is specified, configure alias
    if ( defined($options{alias})) {
        # assume state is up if undefined
        if (!defined($options{'state'}) or $options{'state'} =~ /^up/i) {
            $cmd_str =
                sayEOF($conn,
                       "$ifconfig_cmd $interface alias $options{'alias'}".
                       " netmask $netmask");

            verify_ifconfig_info( $conn, $interface,
                                  alias => $options{'alias'})
                or croak "verify_ifconfig_info failed";
        } elsif ($options{'state'} =~ /^down/i) {
            $cmd_str =
                my $cmd_out = sayEOF($conn,
                                     "$ifconfig_cmd $interface -alias".
                                     " $options{'alias'}");
            verify_ifconfig_info( $conn, $interface,alias => $options{'alias'})
                and croak "verify_ifconfig_info failed - $cmd_out";
        }

    }

    # get_ifconfig_info() uses the key 'mtu' which doesn't match the param

    # get_ifconfig_info() uses the key 'mtu' which doesn't match the param
    # mtusize in do_ifconfig() so if 'mtu' is set then use it.
    $options{mtusize} = $options{mtu} if (defined($options{mtu}));

    # If mtusize is specified, configure mtusize
    if (defined($options{mtusize})) {
        # Set the MTU option name based on the hosttype.
        my $mtu_opt;
        if ( $hostp->hosttype =~ /(solaris|linux)/ ) {
            $mtu_opt = 'mtu';
        } else {
            # Filer and everybody else
            $mtu_opt = 'mtusize';
        }
        $cmd_str = sayEOF($conn,
                "$ifconfig_cmd $interface $mtu_opt $options{'mtusize'}");

        # Return the error on failure to allow tests to see if the MTU
        # is out of the valid range and to see what the valid range is.
        croak "do_ifconfig failed - $cmd_str" if $cmd_str;

        # Sleep to give interface time to start switch negotiations
        sleep(10);
    }

    # If media_type is specified, configure media_type only on the filer
    if (defined($options{media_type}) && ($hostp->hosttype =~ /filer/)) {
        $cmd_str = "mediatype $options{'media_type'}";
        my $cmd_out = sayEOF($conn, "$ifconfig_cmd $interface $cmd_str");
        if ($cmd_out =~ /no such media type/i) {
            croak "Filer ifconfig does not support mediatype " .
                  "$options{'media_type'}";
        }

        # Changing media type can take a little while to take effect so
        # poll for up to 60 seconds to do the verification.
        #my $retry = 12;
        #while ($retry > 0) {
        #    sleep(5);
        #    if (verify_ifconfig_info($conn, $interface,
        #                            media_type => $options{'media_type'})) {
        #       $retry = -1;
        #    } else {
        #       $retry--;
        #    }
        #}
        #croak "verify_ifconfig_info failed setting mediatype " .
        #     "$options{media_type}" if ($retry == 0);
    }

    # If flowcontrol is specified, configure flowcontrol
    if (defined($options{flowcontrol})) {
        $cmd_str = "flowcontrol $options{'flowcontrol'}";
        my $cmd_out = sayEOF($conn, "$ifconfig_cmd $interface $cmd_str");

        verify_ifconfig_info($conn, $interface,
                             flowcontrol => $options{'flowcontrol'})
            or croak "verify_ifconfig_info failed - $cmd_out";
    }

    # Some versions of Linux have a wierd problem where if one interface
    # starts negotiating with a switch then all interfaces are temporarily
    # stopped until negotiations are done (or further along). To work around
    # this sleep here if it is a linux client to let things settle down.
    sleep(20) if ( $hostp->hosttype =~ /linux/ );

       return 1;
}

##
## @name :: get_ifconfig_info
## 
## @description invokes ifconfig command and gathers the information about
## specified interface or all interfaces in a hash and returns the reference
## to hash.
## @arg $conn - connection object
## @arg $interface - name of the interface (optional, defaults to all interface)
## 
## @return reference to hash upon success else croaks.
## example: get_ifconfig_info($conn);
##

sub get_ifconfig_info
{
    my ($conn, $interface) = @_;
    my %ifconfig_info = ();
    my $ifconfig_cmd = undef;

    my $ifconfig_cmd = get_ifconfig_path($conn) or
        croak "can't find ifconfig path";

    if ( ! defined($interface) ) {
        $ifconfig_cmd .= " -a";
    }

#    my ($filer_out) = say($conn, "$ifconfig_cmd $interface");
    my $filer_out = sayEOF($conn, "$ifconfig_cmd $interface");
    if ($filer_out =~ /(no such interface|device not found)/i) {
        return 0;#croak $conn->hostp()->hostname . ": $filer_out";
    }

    my @ifconfig_out = split(/\n/, $filer_out);

    my $interface = undef;
    foreach my $line ( @ifconfig_out ) {
        # Skip blank lines
        next if ( $line =~ m/^\s*?$/ );
        if ( $line =~ m/^(.*):\sflags=(.*>)\smtu\s(\d+)/ ) {
            # This 'if' works for Filer and Solaris
	    $interface = $1;
	    $ifconfig_info{$interface}{flags} = $2;
	    $ifconfig_info{$interface}{mtu} = $3;
        } elsif ( $line =~ m/^(.*?)\s+Link\sencap:(.*?)\s+(HWaddr\s+((\w+:){5}(\w+)))?/ ) {
            # This 'if' works for Linux
	    $interface = $1;
	    $ifconfig_info{$interface}{link_encap} = $2;
	    $ifconfig_info{$interface}{mac_addr} = $4 if defined($4);
        } elsif ( $line =~ m/^\s+(.*)\s+MTU:(\d+)\s+Metric:/ ) {
            # This 'if' works for Linux
	    $ifconfig_info{$interface}{flags} = $1;
	    $ifconfig_info{$interface}{mtu} = $2;
        } elsif ( $line =~ m/inet\s((\d+\.){3}(\d+))\snetmask(?:-or-prefix)?\s(.*?)\s(broadcast\s((\d+\.){3}(\d+)))?/ ) {
            # This 'if' works for Filer and Solaris

	    #\XXX - if $ifconfig_info{$interface}{'ip_addr'} already
	    # exists this must be an alias. 
	    if (!exists($ifconfig_info{$interface}{'ip_addr'})) {
		$ifconfig_info{$interface}{ip_addr} = $1;
		$ifconfig_info{$interface}{netmask} = $4;
		$ifconfig_info{$interface}{broadcast} = $6 if defined($6);
	    } else {
		# alias line 
		$ifconfig_info{$interface}{'alias'}{$1}{'netmask'} = $4;
		$ifconfig_info{$interface}{'alias'}{$1}{'broadcast'} = $6
                  if defined($6);
	    }
        }elsif ($line =~ m/inet6\s((\d+\:){2}(:\d+))\sprefixlen\s((\d+))/ ){
            #for IPv6 Global Unicast IP Addresses ONLY 
	    $ifconfig_info{$interface}{ipv6_global_addr} =$1;
	    $ifconfig_info{$interface}{ipv6_global_prefixlen} =$4;
	} elsif ($line =~ m/inet6\s((fe80::)(\d+\:){3})\sprefixlen\s(\d+)/ ) {
            #for Link Local IPV6 address
            $ifconfig_info{$interface}{ip6_addr_LLA} = $1;
            $ifconfig_info{$interface}{ip6_addr_LLA_prefixlen} =$2;
        } elsif ($line =~ m/inet6\s((fd20:)(\d+\:){6}(\d+))\sprefixlen\s(\d+)/){
            #for IPv6 Reserved Address 
            $ifconfig_info{$interface}{ip6_addr_RA} =$1;
            $ifconfig_info{$interface}{ip6_addr_RA_prefixlen} =$2;
        } elsif ( $line =~ m/\s+inet\saddr:((\d+\.){3}(\d+))\s+(Bcast:((\d+\.){3}(\d+)))?\s+Mask:(.*)/ ) {
            # This 'if' works for Linux
	    $ifconfig_info{$interface}{ip_addr} = $1;
            my $netmask;
            if (defined($8)) {
                $netmask = $8;
                $ifconfig_info{$interface}{broadcast} = $5;
            } else {
                $netmask = $5;
            }
            $ifconfig_info{$interface}{netmask} = $netmask;

        } elsif ( $line =~ m/ether\s((\w+:){5}(\w+))\s\((.*)\)\sflowcontrol\s(\w+)/ ) {
            # This 'if' works for Filer
	    $ifconfig_info{$interface}{mac_addr} = $1;
            $ifconfig_info{$interface}{media_type} = $4;
            $ifconfig_info{$interface}{flowcontrol} = $5;
            # Parse media_type into sub fields
            my @values = split('-', $ifconfig_info{$interface}{media_type});
            foreach my $val (@values) {
                if ($val eq 'auto') {
                    $ifconfig_info{$interface}{auto} = 1;
                } elsif ($val eq 'tp') {
                    $ifconfig_info{$interface}{speed} = 10;
                    $ifconfig_info{$interface}{media} = $val;
                } elsif ($val eq 'up') {
                    $ifconfig_info{$interface}{up} = 1;
                } elsif ($val =~ /down/) {
                    $ifconfig_info{$interface}{up} = 0;
                } elsif ($val eq 'fd') {
                    $ifconfig_info{$interface}{duplex} = 'full';
                } elsif ($val =~ /^(\d+)(.*)/) {
                    $ifconfig_info{$interface}{speed} = $1;
                    $ifconfig_info{$interface}{media} = $2;
                }
            }
            my $tmp = $ifconfig_info{$interface}{speed};
            if ((($tmp == 10) || ($tmp == 100)) &&
                !defined($ifconfig_info{$interface}{duplex})) {
                $ifconfig_info{$interface}{duplex} = 'half';
            }
            # Set media type to a value that the filer would accept on
            # an ifconfig command. Strip auto and up/down status.
            $tmp = $ifconfig_info{$interface}{media_type};
            $tmp =~ s/^auto-//;
            $tmp =~ s/-(up|down)//;
            $ifconfig_info{$interface}{media_type}= $tmp;
        } elsif ( $line =~ m/partner\s+inet\s+((\d+\.){3}(\d+))\s+\((.*)\)/ ) {
            # This 'if' works for Filer
	    $ifconfig_info{$interface}{partner_ip_addr} = $1;
	    $ifconfig_info{$interface}{partner_ifc} = $2;
        } elsif ( $line =~ m/ether\s+((\w+:){5}(\w+))/ ) {
            # This 'if' works for Filer and Solaris
	    $ifconfig_info{$interface}{mac_addr} = $1;
        } elsif ( $line =~ m/mac-filter\s+((\w+:){5}(\w+))/ ) {
            # This 'if' works for Filer with Virtual MAC enabled
	    push(@{$ifconfig_info{$interface}{mac_filter}}, $1);
        } else {
            logDbg("get_ifconfig_info: unrecognized line in output: $line\n");
            next;
        }
    }

    if ( !%ifconfig_info) {
        return 0;
    }
    return(\%ifconfig_info);
}

##
## @name :: get_mac_filter_entries
## 
## @description invokes ifconfig command and gathers the list of MAC
## address filter entries for the given interface.
## @arg $conn - connection object
## @arg $interface - name of the enterface
##

sub get_mac_filter_entries {
    my ($conn, $interface) = @_;
    my %filter_entries = ();

    my $ifconfig_cmd = get_ifconfig_path($conn);
    if (!$ifconfig_cmd) {
        logDbg("get_mac_filter_entries: can't find ifconfig path");
	return 0;
    }

    my $filer_out = sayEOF($conn, "$ifconfig_cmd $interface");
    if ($filer_out =~ /(no such interface|device not found)/i) {
        logDbg("get_mac_filter_entries: invalid interface " .
		"(" . $conn->hostp()->hostname . ", $interface): $filer_out");
	return 0;
    }

    my @ifconfig_out = split(/\n/, $filer_out);

    my $interface = undef;
    foreach my $line ( @ifconfig_out ) {
        # Skip blank lines
        next if ( $line =~ m/^\s*?$/ );
        if ( $line =~ m/mac-filter\s((\w+:){5}(\w+))/ ) {
	    $filter_entries{$1} = 1;
	}
    }

    return %filter_entries;
}


##
## @name :: verify_ifconfig_info
## 
## @description gets the inconfig_info using get_ifconfig_info() and compares
## the parameters passed for the specified connection.
##
## @arg $conn - connection object
## @arg $interface - name of the interface 
## @arg %args - hash of various options
##            "ip_addr" - ip address to configure
##            "alias"   - 'ip address to configure as alias'
##            "flowcontrol"- 'flowcontrol setting none|receive|send|full'
##            "media_type" - 'media type to set to, if it contains the word
##                            auto then it is set to auto'
##            "netmask" - 'netmask to configure for address'
##            "state"   - 'up' or 'down' 
##
## @return 1 upon success or 0 upon failure.
## example: verify_ifconfig_info($conn, 'e3a', 
##                               flags => '<BROADCAST,RUNNING,MULTICAST>',
##                               ip_addr => '172.16.102.101',
##                               alias => '1.1.1.1'
##                               media_type => 'auto-1000t-fd-up',
##                               netmask => '255.255.255.0');
##

sub verify_ifconfig_info
{
    my($fc, $interface, %args) = @_;

    sleep(10);
    my $ifconfig_info = get_ifconfig_info($fc, $interface) or
        croak "get_ifconfig_info failed - $@";

    if ( exists($ifconfig_info->{$interface}) ) {
        # Verify the state
        if ( defined($args{state}) ) {
            if ( $ifconfig_info->{$interface}{state} ne $args{state} ) {
                logDbg("ifconfig_info contained invalid state");
                logDbg(Dumper($ifconfig_info->{$interface}));
                return 0;
            }
            logDbg("ifconfig_info has valid value for state");
        }
	
        # Verify the ip addr
        if ( defined($args{ip_addr}) ) {
	    if ( $args{ip_addr} eq "0.0.0.0") {
		
		if ( exists($ifconfig_info->{$interface}{'ip_addr'})) {
		    logDbg("ifconfig_info contained ip addr");
		    return 0;
		}
	    } else {

		# XXX\gajria- burt184737 prevents the tagging of ip addresses 
		# as base address or alias because the addresses are displayed 
		# unordered. So here we compare ip_addr to all addresses 
		# returned by ifconfig.

		if ($ifconfig_info->{$interface}{'ip_addr'} ne $args{ip_addr} 
		    && !(exists 
			 ($ifconfig_info->{$interface}{alias}{$args{ip_addr}}))){ 
	       	    logDbg("ifconfig_info contained invalid ip addr");
		    logDbg(Dumper($ifconfig_info->{$interface}));
		    return 0;
		}
		logDbg("ifconfig_info has valid value for ip_addr");
	    }
	}
	
        # Verify the netmask
        if ( defined($args{netmask}) ) {
            my $netmask_af = convert_netmask($args{netmask});

	    # Solaris returns netmask without 0x, filer returns netmaks with 0x
	    # Linux returns decimal format
            if (( $ifconfig_info->{$interface}{netmask} !~ $netmask_af ) &&
                ( $ifconfig_info->{$interface}{netmask} ne $args{netmask} )) {
                logDbg("ifconfig_info contained invalid netmask, ".
                    "expecting $netmask_af");
                logDbg(Dumper($ifconfig_info->{$interface}));
                return 0;
            }
            logDbg("ifconfig_info has valid value for netmask");
        }

        # Verify the flags
        if ( defined($args{flags}) ) {
            if ( $ifconfig_info->{$interface}{flags} !~ m/$args{'flags'}/i ) {
                logDbg("ifconfig_info does not contain expected flags");
                logDbg(Dumper($ifconfig_info->{$interface}));
                return 0;
            }
            logDbg("ifconfig_info has valid value for flags");
        }

	# XXX\gajria- burt184737 prevents the tagging of ip addresses 
	# as base address or alias because the addresses are displayed 
	# unordered. So here we compare ip_addr to all addresses 
	# returned by ifconfig.
	
	if (defined($args{alias})) {	    
	    if (!exists($ifconfig_info->{$interface}{'alias'}{$args{'alias'}}) 
		and $ifconfig_info->{$interface}{'ip_addr'} ne $args{'alias'}){
		return 0;
	    }
	    logDbg("ifconfig_info has information of alias");
	}
	
	if (defined($args{media_type})) {
	    if (exists($ifconfig_info->{$interface}{'media_type'})) {
		if ($args{'media_type'} eq 'auto') {
		    if ($ifconfig_info->{$interface}{'auto'} == 0) {
			return 0;
		    }
		} elsif ($ifconfig_info->{$interface}{'media_type'} !~
			/$args{'media_type'}/){
		    return 0;
		}
	    }
	    logDbg("ifconfig_info has valid value for mediatype");
	}

	if (defined($args{flowcontrol})) {
	    if (exists($ifconfig_info->{$interface}{'flowcontrol'})) {
		if ($ifconfig_info->{$interface}{'flowcontrol'} !~
			/$args{'flowcontrol'}/){
		    return 0;
		}
	    }
	    logDbg("ifconfig_info has valid value for flowcontrol");
	}
    }
    
    return 1;
}


## converts the netmask value from decimal format to hex values or
## convert it from hex to decimal format depending on what form it is in
sub convert_netmask
{
	my ($str) = @_;
	my ($i, $num, @num_array, $output);

	# Remove a leading 0x from value
	$str =~ s/^0x//i;

	# Figure out if the number is in decimal format by counting decimals
	$i = ($str =~ tr/.//);
	if ($i == 0) {
		# Input was a hex number so convert it decimal then use
		# sprintf to make sure it is an 8 digit hex number.
		$i = hex($str);
		$str = sprintf "%08x", $i;

		# Convert hex number to decimal format
		$str =~ s/(..)/0x$1,/g;
		$output = sprintf "%d.%d.%d.%d", eval($str);
	} else {
		# Input was in decimal format so convert it to hex
		@num_array = split(/\./, $str);
		$output="";
		foreach $num ( @num_array ) {
			$output .= sprintf("%02x", $num);
		}
	}
	logcomment("Output is $output");
	return($output);
}


## gets the ifconfig path based on the hosttype.
sub get_ifconfig_path
{
    my($conn) = @_;    
    my $ifconfig_cmd = undef;

    # Set the ifconfig path based onthe host type
    my $hosttype = $conn->hostp->hosttype;
    if ( $hosttype =~ /filer/ ) {
        $ifconfig_cmd = "ifconfig";
    } elsif ( $hosttype =~ /solaris/ ) {
        $ifconfig_cmd = "/usr/local/bin/sudo /usr/sbin/ifconfig";
    } elsif ( $hosttype =~ /linux/ ) {
        $ifconfig_cmd = "/usr/local/bin/sudo /sbin/ifconfig";
    } else {
        logStep("unknown hosttype");
        return 0;
    }
    return($ifconfig_cmd);
}

sub ipfw_enable {
    my ($conn) = @_;

    my $fh = $conn->hostp();
    if ($fh->hosttype() =~ /-ng/i) {

        # This is 10 Mode, hence use Server to send the command over.
        my $filer = $fh->id();
        $Server = new Server(nodes => $filer) if (!$Server);

        if($Server->{_utils_node}{$filer}{ipfw_enabled}) {
            $Log->comment("The node $filer has ipfw already enabled");
            return 1;
        }

        # Set values in registry
        my $cmd = "priv set -q advanced; registry set options.ipfw.enable on";
        $Server->system_node_run(node    => $filer,
                                 command => "\"$cmd\"");

        $cmd = "priv set -q advanced;  registry get options.ipfw.enable";
        my $out =
          $Server->system_node_run(node    => $filer,
                                   command => "\"$cmd\"");

        die "registry set options.ipfw.enable failed: $out"
          if ($out !~ /options.ipfw.enable=on/i);

        # Set options value
        $cmd = "priv set -q diag; options ipfw.enable on";
        $Server->system_node_run(node    => $filer,
                                 command => "\"$cmd\"");

        $cmd = " priv set -q diag; options ipfw.enable";
        $out = $Server->system_node_run(node    => $filer,
                                        command => "\"$cmd\"");

        die(" options ipfw . enable failed : $out ")
          if ($out !~ /ipfw\.enable\s+on\s+/i);

        $Server->{_utils_node}{$filer}{ipfw_enabled} = 1;
    } else {

        # This option is not enabled  by default, also there is
        # no registry entry for this.
        registrySet($conn, "options.ipfw.enable", "on");
        filer_options_set($conn, "ipfw.enable", "on");
    }

    return 1;
} ## end sub ipfw_enable

sub ipfw_disable {
    my ($conn) = @_;

    my $fh = $conn->hostp();
    if ($fh->hosttype() =~ /-ng/i) {

        # This is 10 Mode, hence user Server to send the command over.
        my $filer = $fh->id();
        $Server = new Server(nodes => $filer) if (!$Server);

        if(! $Server->{_utils_node}{$filer}{ipfw_enabled}) {
            $Log->comment("The node $filer has ipfw already disabled");
            return 1;
        }

        # Set values in registry
        my $cmd = "priv set -q advanced; registry set options.ipfw.enable off";
        $Server->system_node_run(node    => $filer,
                                 command => "\"$cmd\"");

        $cmd = "priv set -q advanced;  registry get options.ipfw.enable";
        my $out =
          $Server->system_node_run(node    => $filer,
                                   command => "\"$cmd\"");

        die "registry set options.ipfw.enable failed: $out"
          if ($out !~ /options.ipfw.enable=off/i);

        # Set options value
        $cmd = "priv set -q diag; options ipfw.enable off";
        $Server->system_node_run(node    => $filer,
                                 command => "\"$cmd\"");

        $cmd = " priv set -q diag; options ipfw.enable";
        $out = $Server->system_node_run(node    => $filer,
                                        command => "\"$cmd\"");

        die(" options ipfw . enable failed : $out ")
          if ($out !~ /ipfw\.enable\s+off\s+/i);

        $Server->{_utils_node}{$filer}{ipfw_enabled} = 0;
    } else {

        # This option is not enabled  by default, also there is
        # no registry entry for this.
        registrySet($conn, "options.ipfw.enable", "on");
        filer_options_set($conn, "ipfw.enable", "on");
    }

    return 1;
} ## end sub ipfw_disable

sub _say_correctly {
    my ($conn, $cmd) = @_;

    my $out;
    my $fh = $conn->hostp();
    if ($fh->hosttype() =~ /-ng/i) {

        # This is 10 Mode, hence user Server to send the command over.
        my $filer = $fh->id();
        $Server = new Server(nodes => $filer) if (!$Server);

        $out = $Server->system_node_run(node    => $filer,
                                        command => "\"$cmd\"");
    } else {
        (undef, undef, $out) = say($conn, $cmd);
    }

    return $out;
} ## end sub _say_correctly
##-----------------------------------------------------------------------------------------
## Description:
## This proc returns the value of a specified counter from "ifstat -v $intf" output
##
## Parameter Description:
## fc : Filer connection handle
## counter: The counter for which the value is to retrieved
## intf : interface name of the filer whose counter is to be returned
## Usage Example:
##
## get_ifstat_counter (
##	{
##	fc => $filer_conn,
##	counter => "Good TCP6 cksum",
##	intf => $intf
##	} )
##
sub get_ifstat_counter {
    my $args = shift;
    my $fc = $args->{fc};
    my $counter = $args->{counter};
    my $intf = $args->{intf};
    my $count_val = -1;
    logcomment("Checking for $counter value");
    sayEOF($fc, "priv set test");
    my $ifstat_out = sayEOF ($fc, "ifstat -v $intf");
    if ( $ifstat_out =~  /$counter:[ ]*([0-9]+[k]*)/) {
	$count_val = $1;
	print "\n$counter Value is $count_val \n";
	#if the counter value is in kilos
	if ($count_val =~ /([0-9]+)k/) {
	    $count_val = $1*1000;
	    print "\nExact $counter value in decimal is $count_val \n";
	}
	return $count_val;
    } else {
	print "Counter $counter not found";
	return -1;
    }
}
#-----------------------------------------------------------------------------------------
## sio <read> <rand> <blksz> <start_byte> <file_size> <secs> <threads> <dev>
##<read>:       percentage of accesses that are reads.  Range [0,100].
##              'random' keyword makes the percentage random.
##              BEWARE, writing to a file is unchecked and will trash files
##<rand>:       percentage of acceses that are random.  Range [0,100].
##              Sequential accesses = 0random
##<blksz>:      size of I/O's. Example: 2k, 4k, 1m
##              'random' keyword makes the I/O size random 512bytes to 256k.
##<start_byte>:  lower bound for access location in each file
##<file_size>:  total bytes accessed in each file (e.g. 100m, 2g, 1000k)
##<seconds>:    Runtime for test.  Counting starts AFTER all threads have started
##<threads>: Concurrent I/O generators.  Uses real individual threads.
##<dev>:        Device to access.  May be file (foo.out) or device (/dev/dsk/etc)
## @Example
## run_sio (
##    {
##      sio_path=> $sio_path,
##      read=> $read,
##      rand=> $rand
##      blksz=> $blksz,
##      start_byte=>$start_byte,
##      file_size=>$file_size,
##      secs=>$secs,
##      threads=>$threads,
##      dev=>$dev
##    }
## )
sub run_sio {
    my $args = shift;
    my $client = $args->{client};
    my $sio_path = $args->{sio_path};
    my $read = $args->{read};
    my $rand = $args->{rand};
    my $blksz = $args->{blksz};
    my $start_byte = $args->{start_byte};
    my $file_size= $args->{file_size};
    my $secs = $args->{secs};
    my $threads = $args->{threads};
    my $dev = $args->{dev};
    subtest(
	"$sio_path $read $rand $blksz $start_byte $file_size $secs $threads $dev",
	-bg,
	-client, $client,
    );
}
##-----------------------------------------------------------------------------------------
## Description:
## This proc is similar to get_ifstat_counter except that it takes the ifstat -v output as an argument and parses the same
## instead of executing the ifstat -v output command on the filer itself
##
## Returns the value of a specified counter if successful. If the counter is not found returns -1
## Parameter Description:
## fc : Filer connection handle
## counter: The counter for which the value is to retrieved
## intf : interface name of the filer whose counter is to be returned
## Usage Example:
##
##
## my $ifstat_v_output = sayEOF ($fc, "ifstat -v e0b");
## parse_ifstat_counter (
##      {
##      ifstat_v_output => $ifstat_v_output,
##      counter => "Good TCP6 cksum",
##      }
## )
##
sub parse_ifstat_counter {
    my $args = shift;
    my $ifstat_v_output = $args->{ifstat_v_output};
    my $counter = $args->{counter};
    my $count_val = -1;
    if ( $ifstat_v_output =~  /$counter:[ ]*([0-9]+[k]*)/) {
        $count_val = $1;
        print "\n$counter Value is $count_val \n";
        #if the counter value is in kilos
        if ($count_val =~ /([0-9]+)k/) {
            $count_val = $1*1000;
            print "\nExact $counter value in decimal is $count_val \n";
        }
        return $count_val;
    } else {
        print "Counter $counter not found";
        return -1;
    }
}

##-----------------------------------------------------------------------------------------
## Description:
## Proc to do switch configurations for vlan and vif (Multi and lacp);
## Returns the value of a specified counter if successful. If the counter is not found returns -1
## Parameter Description:
## Hash refernce is used
## Required parameters:-
## $args->{conn} :- Switch Connection handel
## $args->{port_type} :- Port type(Access,trunk,Etherchannel)
## $args->{interface} :- Interface for which the configurations need to be applied
## $args->{default_vlan} :- Default vlan id of the interface
## $args->{channel_group} :- Channel Group number for etherchannel porttype.
## Optional Parameters
## $args->{allowed_vlans} :- Comma seperated list of vlan ids that are allowed on than interface
## $args->{native_vlan} :- Native vlan id for untagged valns.
## $args->{switchport_mode} :- trunk or access for etherchannel configuration
## $args->{channel_protocol} :- Channel protocol type for lacp etherchannel configuration
## Examples
## Access Mode Configuration
##--------------------------
##switch_port_config({conn=>$sc,
##                    port_type=>"access",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan
##                    });
##say($sc,"sh run int $interface");
##
## Trunk mode configuration
##-------------------------
##switch_port_config({conn=>$sc,
##                    port_type=>"trunk",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan
##                    });
## Trunk mode configuration for Untagged Vlans
##-------------------------------------------
##switch_port_config({conn=>$sc,
##                    port_type=>"trunk",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan,
##                    native_vlan=>"10"
##                    });
##switch_port_config({conn=>$sc,
##                    port_type=>"trunk",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan,
##                    allowed_vlans=>"10,20,30,40"
##                    });
##
## Multimode vif configuration
## ---------------------------
##switch_port_config({conn=>$sc,
##                    port_type=>"etherchannel",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan,
##                    channel_group=>158,
##                    $switchport_mode=>"trunk"
##                    });
##
##switch_port_config({conn=>$sc,
##                    port_type=>"etherchannel",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan,
##                    channel_group=>158
##                  });
## Lacp configuration
##-------------------
##switch_port_config({conn=>$sc,
##                    port_type=>"etherchannel",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan,
##                    channel_group=>158,
##                    channel_protocol=>"lacp"
##                    });
##switch_port_config({conn=>$sc,
##                    port_type=>"etherchannel",
##                    interface=>$interface,
##                    default_vlan=>$default_vlan,
##                    channel_group=>158,
##                    channel_protocol=>"lacp",
##                    switchport_mode=>"trunk"
##                    });

sub switch_port_config
{   
	my $args = shift ;
	my $interface;
	my $conn = $args->{conn};
	if(!defined($args->{interface}))
	{
		logcomment("Interface vlan need to be specified");
		return 0;
	}
	if(!defined($args->{default_vlan}))
	{
		logcomment("Defualt vlan need to be specified");
		return 0;
	}
	if(!defined($args->{port_type}))
	{
		logcomment("port type need to be specified");
		return 0;
	}
	my $port_type = $args -> {port_type};
	my $interface =$args->{interface};
	my $default_vlan = $args->{default_vlan};
	if($port_type =~ m/access/)
	{
		say($args->{conn},"configure terminal");
		say($args->{conn},"default interface $interface");
		say($args->{conn},"interface $interface");
		say($args->{conn},"switchport");
		say($args->{conn},"switchport mode access");
		say($args->{conn},"switchport access vlan $default_vlan");
		say($args->{conn},"no shutdown");
		say($args->{conn},"exit");
		say($args->{conn},"exit");
	}
	elsif($port_type =~ m/trunk/)
	{
		my $allowed_vlans =  $args->{allowed_vlans};
		my $native_vlan = $args->{native_vlan};
		say($args->{conn},"configure terminal");
		say($args->{conn},"default interface $interface");
		say($args->{conn},"interface $interface");
		say($args->{conn},"switchport");
		say($args->{conn},"switchport trunk encapsulation dot1q");
		say($args->{conn},"switchport mode trunk");
		if($allowed_vlans)
		{
			say($args->{conn},"switchport trunk allowed vlan $allowed_vlans");
		} else {
		        logcomment ("Allowed vlans \(allowed_vlans\) parameter not passed with trunk mode");
                }
		if($native_vlan)
		{
			say($args->{conn},"switchport trunk native vlan $native_vlan");

		}
		say($args->{conn},"no shutdown");
		say($args->{conn},"exit");
		say($args->{conn},"exit");
	}
	elsif($port_type =~ m/etherchannel/)
	{
		if(!defined($args->{channel_group}))
		{
			logcomment("Channel group is not specified");
			return 0;
		}
		my $interface =$args->{interface};
		my $default_vlan = $args->{default_vlan};
		my $channel_group = $args->{channel_group}; 
		my $allowed_vlans =  $args->{allowed_vlans};
		my $native_vlan = $args->{native_vlan};
		my $switchport_mode = $args->{switchport_mode} || "access";
		say($args->{conn},"configure terminal");
		say($args->{conn},"default interface $interface");
		say($args->{conn},"interface $interface");
		say($args->{conn},"switchport");
		say($args->{conn},"switchport trunk encapsulation dot1q") if ($switchport_mode =~ m/trunk/);
		say($args->{conn},"switchport mode $switchport_mode");
		say($args->{conn},"switchport access vlan $default_vlan");
		if($switchport_mode =~ m/trunk/)
		{
			if($allowed_vlans)
			{
				say($args->{conn},"switchport trunk allowed vlan $allowed_vlans");
			}
			if($native_vlan)
			{
				say($args->{conn},"switchport trunk native vlan $native_vlan");
			}
		}
		if(!defined($args->{channel_protocol}))
		{
			say($args->{conn},"channel-group $channel_group mode on");
		}
		else
		{
			say($args->{conn},"channel-group $channel_group mode active");
			say($args->{conn},"channel-protocol lacp");
		}
		say($args->{conn},"no shutdown");
		say ($args->{conn},"interface port-channel $channel_group");
		say($args->{conn},"no shutdown");
		say($args->{conn},"exit");
		say($args->{conn},"exit");
	}

	else
	{
		logcomment("Only trunk ,access and etherchannel configurations are supported");
		return 0;
	}
	return 1;
}
#------------------------------------------------------------------------------------------
sub do_ifconfig_v6 {
    my $args = shift;
    my $fc = $args->{fc};
    my $intf = $args->{intf};
    my $ipv6_addr = $args->{ipv6_addr};
    my $prefixlen = $args->{prefixlen};
    my $alias = $args->{alias} || "";
    
    #if the ipv6 address has to be removed
    if ($ipv6_addr eq "0::0") {
        logcomment ("Exiting do_ifconfig_v6 after removing ipv6 addresses");
        say($fc, "ifconfig $intf $ipv6_addr");
	return 1;
    }

    if($args->{alias})
    {
	say($fc, "ifconfig $intf alias $alias prefixlen $prefixlen");
    if (_ipv6_address_is_configured({fc => $fc,intfc => $intf,ipv6 => $alias})) {
	    {
#logWarn ("IPv6 alias address is successfully configured on the filer interface");
		    return 1;

	    }
    }
    else
    {
	    logWarn ("IPv6 address is not successfully configured on the filer interface");
	     return 0;

    }
    }
    #Configure IPv6 address on the filer interface
    
    say($fc, "ifconfig $intf $ipv6_addr prefixlen $prefixlen");

    #check if the ipv6 address is successfully configured on the filer interface
    if (!_ipv6_address_is_configured({fc => $fc,intfc => $intf,ipv6 => $ipv6_addr})) {
	logWarn ("IPv6 address is not successfully configured on the filer interface");
	return 0;
    }
    return 1;
}

## @name :: get_nis_binding
##
## @summary
##   Check if NIS binding exists and return name/ip of bound nis server
##
## @args $conn Required:  connection object
##

sub get_nis_binding( $ ) {

    my $conn = shift;

    my $line = sayEOF($conn,"ypwhich");

    if ($line =~ m/not bound to any nis server/i) {
        logcomment ("NIS client not bound to any NIS server");
        return 0;
    }

    if ($line =~ m/nis is not enabled/i) {
        logcomment ("NIS not enabled");
        return 0;
    }

    return($line);
}

#------------------------------------------------------------------------------------------
## @name ::tcp_udp_server_create()
## @summary create tcp_udp_server in sk or bsd  to communicate with tcp_udp_client for interstack communication.
##
## @description
##
##
## @arg $conn           Required Connectspec of device
## @arg $stack          Required to create tcp_udp_server, either 'SK' or 'BSD'
## @arg $addr_family    Required to create tcp_udp_server,4 for IPv4, 6 for IPv6
## @arg $ip_addr        Required to create tcp_udp_server,
## @arg $port_no        Required to create tcp_udp_server ,
## @arg %args           Optional Hash of optional arguments
##                          <vfiler_name>: vfiler name, if to create tcp_server in vfiler context in sk
##                          <vserver>: Vserver-id (number OR * for Any-Vserver OR "" for No-Vserver
##                          <lif-role>: Number as VFILERLIF(1), DATALIF(2),
##                                      CLUSTERLIF(3), LOOPBACKLIF(4), NODEMGMTLIF(5),
##                                      CSERVERMGMTLIF(6), INTERCLUSTERLIF(7),
##                                      VSERVERMGMTLIF(8), ANYLIF(9), CMODELIF(10), ALLLIF(11)
##                                      OR "" for 'Not mentioning LIF' OR * for ANYLIF
##
## @return for         return 0 for Failure, 1 for Success
##Usage for tcp_server create in 'SK' of the filer
##tcp_udp_server_create($fc1,"tcp_server","SK","$ADDR_FAMILY",""$FILER1_IPV6_ADDR1"","$SERVER_PORT_NO",vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE);
##Usage for tcp_server create in 'BSD' of the filer
##tcp_udp_server_create($fc1,"tcp_server","BSD","$ADDR_FAMILY","$FILER1_IPV6_ADDR1"","$SERVER_PORT_NO",vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE)
##Usage for udp_server create in 'SK' of the filer
##tcp_udp_server_create($fc1,"udp_server","SK","$ADDR_FAMILY","$FILER1_IPV6_ADDR1","$SERVER_PORT_NO",vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE)
##Usage for udp_server create in 'BSD' of the filer
##tcp_udp_server_create($fc1,"udp_server","BSD","$ADDR_FAMILY","$FILER1_IPV6_ADDR1","$SERVER_PORT_NO",vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE)
sub tcp_udp_server_create($$$$$$%) {

    my ($conn, $server,$stack, $addr_family, $ip_addr, $port_no,%args) = @_;
    my $vserver_id = defined($args{'vserver_id'}) ? $args{'vserver_id'} : undef;
    my $lif_role = $args{'lif_role'} || undef;;
    my $prepend = defined($args{'vfiler_name'}) ? "vfiler run $args{'vfiler_name'}" : undef;
     
    if($stack eq "BSD"){
        undef($prepend);
        logcomment("creating $server in BSD");
        logcomment("login to systemshell");
        login_systemshell($conn);
        }
    else {
        logcomment("creating $server in sk");
        logcomment("changing previlage to diag");
        say($conn,"priv set diag",-swallow,-blind);
        undef($vserver_id);
        undef($lif_role);
    }
    
    my ($output,undef,undef)=say($conn,"$prepend $server $addr_family $ip_addr $port_no $vserver_id $lif_role", -noprompt, '-wait',"\[lL\]istening on",".*failed.*",-timeout, 40 ); 
    
    if($output=~/listening on/sgi){
        logcomment("$server is created succesfully");
        return 1;
    }else{
        logcomment("$server creation failed");
        say($conn,"exit");
        return 0;
    }
}

#------------------------------------------------------------------------------------------
 
## @name ::tcp_udp_client_create()
## @summary create tcp_udp_client in sk or bsd  to communicate with tcp_udp_server for interstack communication.
##
## @description
##
##
## @arg $conn           Required Connectspec of device
## @arg $stack          Required to create tcp_udp_client, either 'SK' or 'BSD'
## @arg $addr_family    Required to create tcp_udp_client,4 for IPv4, 6 for IPv6
## @arg $ip_addr        Required to create tcp_udp_client,
## @arg $port_no        Required to create tcp_udp_client ,
## @arg %args           Optional Hash of optional arguments
##                          <vfiler_name>: vfiler name, if to create tcp_client in vfiler context in sk
##                          <vserver>: Vserver-id (number OR * for Any-Vserver OR "" for No-Vserver
##                          <lif-role>: Number as VFILERLIF(1), DATALIF(2),
##                                      CLUSTERLIF(3), LOOPBACKLIF(4), NODEMGMTLIF(5),
##                                      CSERVERMGMTLIF(6), INTERCLUSTERLIF(7),
##                                      VSERVERMGMTLIF(8), ANYLIF(9), CMODELIF(10), ALLLIF(11)
##                                      OR "" for 'Not mentioning LIF' OR * for ANYLIF
##                          <pkt_count>: [-c packetcount]: number of times to send packet [1 .. 100]
##                          <payload_size>: [-s payloadsize]: specify payload size [1 .. 65535]
## $args{'payload_size'}  optional parameter and Payload length should be in range from 1 to 65535 and for "Ipv6 udp" its value is limited to 8112.
##
## @return for         return 0 for Failure, 1 for Success
##
##Usage for tcp_client create in 'SK' of the filer
##tcp_udp_client_create($fc2,"tcp_client","SK","$ADDR_FAMILY","$FILER2_IPV6_ADDR1","$CLIENT_PORT_NO","$FILER1_IPV6_ADDR1","$SERVER_PORT_NO",payload_size=>$PKT_SIZE1,vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE)
##Usage for tcp_client create in 'BSD' of the filer
##tcp_udp_client_create($fc2,"tcp_client","BSD","$ADDR_FAMILY","$FILER2_IPV6_ADDR1","$CLIENT_PORT_NO","$FILER1_IPV6_ADDR1","$SERVER_PORT_NO",payload_size=>$PKT_SIZE1,vserver_id=>$ra11_vid,lif_role=>$LIF_ROLE)
##Usage for udp_client create in 'SK' of the filer
##tcp_udp_client_create($fc2,"udp_client","SK","$ADDR_FAMILY","$FILER2_IPV6_ADDR1","$CLIENT_PORT_NO","$FILER1_IPV6_ADDR1","$SERVER_PORT_NO",payload_size=>$PKT_SIZE1,vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE)
##Usage for udp_client create in 'BSD' of the filer
##tcp_udp_client_create($fc2,"udp_client","BSD","$ADDR_FAMILY","$FILER2_IPV6_ADDR1","$CLIENT_PORT_NO","$FILER1_IPV6_ADDR1","$SERVER_PORT_NO",payload_size=>$PKT_SIZE1,vserver_id=>$ip1_vid,lif_role=>$LIF_ROLE)
##
sub tcp_udp_client_create($$$$$$$$%) {

    my ($conn, $client,$stack, $addr_family, $client_ip, $client_port,$server_ip, $server_port,%args) = @_;
    my $vserver_id = defined($args{'vserver_id'}) ? $args{'vserver_id'} : undef;
    my $lif_role = $args{'lif_role'} || undef;
    my $prepend = defined($args{'vfiler_name'}) ? "vfiler run $args{'vfiler_name'}" : undef;
    my $pkt_count = defined($args{'pkt_count'}) ? "-c $args{'pkt_count'}" : undef;
    my $payload_size = defined($args{'payload_size'}) ? "-s $args{'payload_size'}" : "-s 25";
    
    if($stack eq "BSD"){
        undef($prepend);
        logcomment("creating $client in BSD");
        logcomment("login to systemshell");
        login_systemshell($conn);
    }else {
        logcomment("creating $client in sk");
        logcomment("changing previlage to diag");
        say($conn,"priv set diag",-swallow,-blind);
        undef($vserver_id);
        undef($lif_role);
    }
    
    my ($output,undef,undef)= say($conn,"$prepend $client $addr_family $client_ip $client_port $server_ip $server_port $vserver_id $lif_role $pkt_count $payload_size", -noprompt, '-wait',".*failed.*","[1-9][0-9]* bytes (received|sent)",".*57 bytes.*",-timeout, 100);
    if($output=~/[1-9][0-9]* bytes (received|sent)/sgi){
        logcomment("Client is sending data to server succesfully");
        return 1;
    }
    elsif($output=~/bind failed/sgi){
        sleep(60);
        my ($output1,undef,undef)=say($conn,"$prepend $client $addr_family $client_ip $client_port $server_ip $server_port $vserver_id $lif_role $pkt_count $payload_size", -noprompt, '-wait',".*failed.*","[1-9][0-9]* bytes (received|sent)",".*57 bytes.*",-timeout, 100 );
        if($output1=~/[1-9][0-9]* bytes (received|sent)/sgi){
            logcomment("Client is sending data to server succesfully");
            return 1;
        }
        else{
            logcomment("$client creation failed");
            say($conn,"\x03");
            say($conn,"exit");
            return 0;
        }
    }else{
        logcomment("$client creation failed");
        say($conn,"\x03");
        say($conn,"exit");
        return 0;
    }
}

#------------------------------------------------------------------------------------------
## @name ::check_comm_status()
## @summary This is to verify the communication is happening or not between server and client.
## @description
##
##
## @arg $conn           Required to have "server" Connectspec of device
## $args{'payload_size'}  optional parameter and Payload length should be in range from 1 to 65535 for tcp and for UDP its valuse is limited to 8112.
## $args{'pkt_count'}  optional parameter 
sub check_comm_status($%){
    my ($conn,%args)=@_;
    my $count=defined($args{'pkt_count'}) ? $args{'pkt_count'} : 10;
    my $payload_size = defined($args{'payload_size'}) ? "$args{'payload_size'}" :'25';
    my $cnt;
     my $Hostname=$conn->hostp->hostname();
    my (undef,undef,$result)=say($conn,"\x03",-nocr,-prompt,"$Hostname.*",-timeout,60*2+60);
    if($payload_size ne 25 ){
        if($result=~/bytes received/sgi) {
             return 1;
        }
        else{
            return 0;
        }
    }    
    else
    {
        while($result=~/25 bytes received/sgi){
            $cnt+=1;
        }          
#    $count!~$cnt? return 0 : return 1; 
	        if($cnt!~ $count){
			            return 0;
				            }
		        else {
				            return 1;
					            }
			
    }
}

#-------------------------------------------------------------------------------------------
## @name ::get_vserver_id()
## @summary This is to get vserver id associated with specific ip address of interface.
## @description
##
##
##  $conn           Required to have connectspec of device
##  $interface      Required, interface on which ip is configured.
##  $ip             required,IP address for which we are finding vserver id.
sub get_vserver_id{
    my ($conn, $interface,$ip) = @_;
    my %ifconfig_info = ();
    my $ifconfig_cmd = "ifconfig";
    my $vserver_id= undef;
    sayEOF($conn,"priv set test");
    my $filer_out = sayEOF($conn, "$ifconfig_cmd $interface");
    if ($filer_out =~ /(no such interface|device not found)/i) {
        logcomment("failed to execute ifconfig command on $interface");
         return $TCD::FAIL;
    }
    my @ifconfig_out = split(/\n/, $filer_out);
    $interface = undef;
    foreach my $line ( @ifconfig_out ) {
        # Skip blank lines
        next if ( $line =~ m/^\s*?$/ );
        #if ($line =~ m/inet*\s($ip) .* vFiler ID:\s+(\d+)/){
         if($line =~m/inet*(\d*)\s($ip).*\s*vFiler ID:\s*(\d+)/i){
                $vserver_id=$3;
        }
    }
    logcomment("Vserver id is $vserver_id");
    if (!defined($vserver_id)) {
        logcomment("Not able to get vserver_id");
        return $TCD::FAIL;
    }
    return($vserver_id);
}       
#----------------------------------------------------------------------------------------------
sub get_vserver_id_bsd {
    my ($conn, $ips,$ip,$user_name,$user_password) = @_;
    my %ifconfig_info = ();
    my $ifconfig_cmd = "ifconfig";
    my $vserver_id= undef;
    logcomment("login to systemshell");
    login_systemshell($conn);

    my $filer_out = sayEOF($conn, "$ifconfig_cmd $ips");
    if ($filer_out =~ /(no such interface|device not found)/i) {
        logcomment("failed to execute ifconfig command on $ips");
         return 0;
    }
    my @ifconfig_out = split(/\n/, $filer_out);
    $ips = undef;
    foreach my $line ( @ifconfig_out ) {
        # Skip blank lines
        next if ( $line =~ m/^\s*?$/ );
        #if ($line =~ m/inet*\s($ip) .* Vserver ID:\s+(\d+)/){
        if($line =~m/inet*(\d*)\s($ip).* Vserver ID:\s+(\d+)/){
            $vserver_id=$3;
        }
    }
    if (!defined($vserver_id)) {
        logcomment("Not able to get vserver_id");
        return 0;
    }
        say($conn, "exit" );
    return($vserver_id);
}
#----------------------------------------------------------------------------------------------
sub remove_config_interface {
    my ($conn,%args)   = @_;
    my $interface = $args{interface} if($args{interface});
    my $orig_ifconfig_info;
#logDbg(Dumper($conn));
my $if_default = $conn->{'hostp'}->{'IFC'};
if ( $interface =~ m/$if_default/i ) {
    logcomment("Interface is used to connect to filer so not removing the configuration");
#       return 1;
    }
    logcomment("Interface is $interface and  $if_default ");
#    say($conn,"vlan delete -q $interface");
    $orig_ifconfig_info = get_ifconfig_info( $conn, $interface ) or die fail_result("Getting ifconfig failed");
    if ( exists( $orig_ifconfig_info->{$interface}{'ip_addr'} ) ) {
        if ( exists( $orig_ifconfig_info->{$interface}{'alias'} ) ) {
            foreach my $addr (
                keys( %{ $orig_ifconfig_info->{$interface}{'alias'} } ) )
            {
                do_ifconfig(
                    $conn, $interface,
                    alias => $addr,
                    state => "down"
                ) or return 0;
            }
        }
    }
   do_ifconfig($conn,$interface,mtusize=>1500); 
    do_ifconfig( $conn, $interface, ip_addr => '0.0.0.0' ) or return 0;
    do_ifconfig_v6( {fc=>$conn, intf=>$interface, ipv6_addr => '0::0'} ) or return 0;
    do_ifconfig( $conn, $interface, state   => "down" ) or return 0;
    return 1;
}

sub get_ifconfig_info_bsd{
        
      my ($conn, $interface) = @_;
    my %ifconfig_info = ();
    my $ifconfig_cmd = undef;

    my $ifconfig_cmd = get_ifconfig_path($conn) or
        croak "can't find ifconfig path";
logDbg("login to systemshell");
      login_systemshell($conn);
    if ( ! defined($interface) ) {
        $ifconfig_cmd .= " -a";
    }
my $hostname = $conn->hostp()->hostname();
    my (undef,undef,$filer_out) = say($conn, "$ifconfig_cmd $interface",-swallow,-prompt,"$hostname%");
    if ($filer_out =~ /(no such interface|device not found|does not exist)/i) {
        logcomment("Interface does not exists");
#	logcomment("XXXX retruning 0... $filer_out");
        sayEOF($conn, "exit");
        return 0;
    }

    my @ifconfig_out = split(/\n/, $filer_out);
#logcomment("ifconfig o/p is $filer_out");
    my $interface = undef;
    foreach my $line ( @ifconfig_out ) {
        # Skip blank lines
#	    logcomment("Line is $line");
	next if ( $line =~ m/^\s*?$/ ) or ( $line =~ m/^\s*(options|ether|media)/ );
        if ( $line =~ m/^(.*):\sflags=(.*>).*\smtu\s(\d+)/ ) {
            # This 'if' works for Filer and Solaris
	    $interface = $1;
	    $ifconfig_info{$interface}{flags} = $2;
	    $ifconfig_info{$interface}{mtu} = $3;
        }  elsif ( $line =~ m/inet\s((\d+\.){3}(\d+))\snetmask\s(.*?)\s(broadcast\s((\d+\.){3}(\d+)))?\s(.*LIF)\sVserver ID:\s*(\d+)?/i ) {
            #inet 192.168.1.131 netmask 0xffffff00 broadcast 192.168.1.255 VFILERLIF Vserver ID: 27
                $ifconfig_info{$interface}{$1}{netmask} = $4;
		$ifconfig_info{$interface}{$1}{broadcast} = $6 if defined($6);
                $ifconfig_info{$interface}{$1}{lif_type} = $9 if defined($9);
                $ifconfig_info{$interface}{$1}{vserver_id} = $10 if defined($10);
        }elsif ($line =~ m/inet6\s((fe80::)((\d|\w)+\:){3}(\d+|\w+))%(.*)\sprefixlen\s(\d+)\sautoconf\sscopeid\s0x(\d|\w)+\s+(.*LIF)\s+Vserver ID:\s*(\d+)?/){
            #for Link Local IPV6 address
            #$ifconfig_info{$interface}{ip6_addr_LLA} = $1;
            #$ifconfig_info{$interface}{$1}{interface
            $ifconfig_info{$interface}{$1}{prefixlen} =$7;
            $ifconfig_info{$interface}{$1}{lif_type} = $9 if defined($9);
            $ifconfig_info{$interface}{$1}{vserver_id} = $10 if defined($10);
            $ifconfig_info{$interface}{$1}{type} = "LLA";
            
        }elsif ($line =~ m/inet6\s(((\d+:|\w+:){4}|(\d+::|\w+::){1})((\d|\w)+\:){3}(\d|\w)+)\sprefixlen\s(\d+)\sautoconf\s\s+(.*LIF)\s+Vserver ID:\s*(\d+)/){#\sprefixlen\s(\d+)/){
            #inet6 5059::207:43ff:fe05:9fb7 prefixlen 64 autoconf  VFILERLIF Vserver ID: 0
             #inet6 6ffe::2a0:98ff:fe08:53c6 prefixlen 64 autoconf  VFILERLIF Vserver ID: 0
#inet6 6ffe::82 prefixlen 64  VFILERLIF Vserver ID: 0
            #$ifconfig_info{$interface}{$1} =$1;
            $ifconfig_info{$interface}{$1}{prefixlen} =$8;
            $ifconfig_info{$interface}{$1}{lif_type} = $9 if defined($9);
            $ifconfig_info{$interface}{$1}{vserver_id} = $10 if defined($10);
             $ifconfig_info{$interface}{$1}{type} = "RA";
        }elsif($line=~m/inet6\s(((\d+|\w+):)+.*(\d+|\w+)*)\sprefixlen\s(\d+)\s+(.*LIF)\s+Vserver ID:\s*(\d+)/)
        {
            $ifconfig_info{$interface}{$1}{prefixlen} =$5;
            $ifconfig_info{$interface}{$1}{lif_type} = $6 if defined($6);
            $ifconfig_info{$interface}{$1}{vserver_id} = $7 if defined($7);
            $ifconfig_info{$interface}{$1}{type} = "global";
        }elsif($line=~m/inet6\s(::.*)\sprefixlen\s(\d+)\s+(.*LIF)\s+Vserver ID:\s*(\d+)/)
        {
            $ifconfig_info{$interface}{$1}{prefixlen}=$2;
            $ifconfig_info{$interface}{$1}{lif_type} = $3 if defined($3);
            $ifconfig_info{$interface}{$1}{vserver_id} = $4 if defined($4);
            $ifconfig_info{$interface}{$1}{type} = "loopback";
             #prefixlen 128  LOOPBACKLIF Vserver ID: 0
        }
        else {
            logDbg("get_ifconfig_info: unrecognized line in output: $line\n");
            next;
        }
    }
sayEOF($conn, "exit");
    if ( !%ifconfig_info) {

	logcomment("YYYY retruning 0... $filer_out");
        return 0;
    }

#logcomment("ifconfig info".Dumper(\%ifconfig_info));
    return(\%ifconfig_info);
}

sub verify_ifconfig_info_bsd
{
    my($fc, $interface, %args) = @_;
    logcomment("Arguments are".Dumper(\%args));
    sleep(10);
    my $ifconfig_info = get_ifconfig_info_bsd($fc, $interface);
   if(!$ifconfig_info)
   {
#logcomment("return value from $ifconfig_info");
	   return 0;
   }
    
logcomment(Dumper($ifconfig_info));
    if ( exists($ifconfig_info->{$interface}) ) {
        # Verify the state
        if ( defined($args{flags}) ) {
            if ( $ifconfig_info->{$interface}{flags} =~ m/$args{flags}/ ) {
                logcomment("ifconfig_info has valid value for flags");
                        
            }
            else
            {
                logcomment("ifconfig_info  does not valid value for flags");
                 return 0;
            }
            
        }

    if ( defined($args{ip_addr}) ) {
	    if ( !exists($ifconfig_info->{$interface}{"$args{ip_addr}"})) {
		    logcomment("Ipaddress does not exist");
		    return 0;
		}
                   logcomment("Ipaddress exists in BSD");
                
	    }
   	
        # Verify the netmask
    if ( defined($args{netmask}))  {
            my $netmask_af = convert_netmask($args{netmask});
            if (( $ifconfig_info->{$interface}{"$args{ip_addr}"}{netmask} !~ $netmask_af ))
            {
                logDbg("ifconfig_info contained invalid netmask, ".
                    "expecting $netmask_af");
                logDbg(Dumper($ifconfig_info->{$interface}));
                return 0;
            }
            logDbg("ifconfig_info has valid value for netmask");
        }
    if ( defined($args{prefixlen}))  {
            if (( $ifconfig_info->{$interface}{"$args{ip_addr}"}{prefixlen} !~ $args{prefixlen}))
            {
                logDbg("ifconfig_info contained invalid prefixlen, ");
                 return 0;
            }
            logDbg("ifconfig_info has valid value for prefixlen");
        }
    if (defined($args{vserver_id})){
            if (( $ifconfig_info->{$interface}{"$args{ip_addr}"}{vserver_id} !~ m/$args{vserver_id}/))
            {
                logcomment("Invalid vserver id");
                 return 0;
            }
            logDbg("ifconfig_info has valid value for Vserver ID");
        }
     if (defined($args{lif_type})){
            if (( $ifconfig_info->{$interface}{"$args{ip_addr}"}{lif_type} !~ $args{lif_type}))
            {
                logcomment("Lif type is different");
                 return 0;
            }
            logDbg("ifconfig_info has valid value for Vserver ID");
        } 
        
     if (defined($args{mtu})){
            if ( $ifconfig_info->{$interface}{mtu} !~ $args{mtu})
            {
                logcomment("mtu is different");
                 return 0;
            }
            logDbg("ifconfig_info has valid value for Vserver ID");
    }
	return 1;
}
}
#-----------------------------------------------------------------------------------------
##run_BSD_ping($fc,"192.168.1.1","ping",count=>4);
##run_BSD_ping($fc,"6ffe::1","ping6",count=>4,vserver_id=>1);
sub run_BSD_ping($$$%) {
    my ($conn,$destn,$ping,%args) = @_;
    $destn=defined($args{'scope'}) ? "$destn%$args{'scope'}" :$destn; 
    my $vserver_id = defined($args{'vserver_id'}) ? $args{'vserver_id'} : undef;
    my $pkt_size = defined($args{'pkt_size'})? $args{'pkt_size'} : undef;
    my $buff_size = defined($args{'buff_size'})? $args{'buff_size'}: undef;
    my $output;
    my $count= defined($args{'count'})? $args{'count'}: 2;
    my $cmd = "";
    my $hosttype = $conn->hostp->hosttype();
    my $hostname = $conn->hostp->hostname();
    
    if($hosttype =~ /filer/){
        if( defined($pkt_size) ){
            if(($pkt_size > 8112) and ($buff_size)) {
                $cmd =defined($vserver_id)? "sudo vcontext -v $vserver_id /sbin/$ping $ping -c $count -s $pkt_size -b $buff_size $destn" : "sudo $ping -c $count -s $pkt_size -b $buff_size $destn";
            }
            else{
                $cmd =defined($vserver_id)? "sudo vcontext -v $vserver_id /sbin/$ping $ping -c $count -s $pkt_size $destn" : "sudo $ping -c $count -s $pkt_size $destn"; 
            }
        }
        else{
            $cmd = defined($vserver_id)? "sudo vcontext -v $vserver_id /sbin/$ping $ping -c $count $destn" : "sudo $ping -c $count $destn";    
        }
        logcomment("login to systemshell");
        login_systemshell($conn);
        (undef,undef,$output) = say( $conn,"$cmd",-prompt,"$hostname%");
        if (
            __check_pattern(
                {
                    variable => $output,
                    pattern  => "(?:bytes from|is alive)",
                    action   => 'check'
                }
            )
        )
        {
		if(__check_pattern({variable=>$output,pattern=>"100\.0% packet loss",action=>'check'}))
		{

            say($conn,"exit",-swallow,-prompt,"$hostname%");
            return 0;
		}
            say($conn,"exit",-swallow,-prompt,"$hostname%");
            return 1;
        }
        else {
            say($conn,"exit",-swallow,-prompt,"$hostname%");
            return 0;
        }
    }
    else{
       logcomment("Make sure you are using this subroutine to ping FROM the filer");
    }
}
#-------------------------------------------------------------------------------------------
sub ndp
{
my $fc=shift;
my %val = ();
my ( undef, undef, $cmd_out ) = say( $fc, "ndp -an" );
         my @lines = split( /\n/, $cmd_out );
         foreach my $a (@lines) {
             my @array = split( /\s+/, $a );
             $val{$array[0]}{Linklayer_Address} = $array[1];
             $val{$array[0]}{Interface} = $array[2];
         }
     
 return ( \%val );
}

#Bringing the switch to enable mode
#Enable the switch
#switch_enable ($sc, "password");
#
sub switch_enable ($$) {
my $conn = shift;
my $pwd = shift;
say ($conn, "enable");
say ($conn, $pwd);
}

#for switch exec mode commands e.g. show commands
sub switch_exec ($$$) {
my $conn = shift;
my $cmd = shift;
#pass Ctrl+c in order to bring the switch to the exec prompt if it is not in it
say ($conn,"\x03");
my (undef, undef,$output) = say($conn,$cmd);
say($conn,"exit");
say($conn,"exit");
return $output;
}

#for switch configuration commands
sub switch_config ($$$) {
my $conn = shift;
my $cmd = shift;
say ($conn,"\x03");
say($conn,"configure terminal");
say($conn,$cmd);
say($conn,"exit");
say($conn,"exit");
}

sub switch_intf_config ($$$) {
my $conn = shift;
my $interface = shift;
my $cmd = shift;
say($conn,"configure terminal");
say($conn,"interface $interface");
say($conn,$cmd);
say($conn,"exit");
say($conn,"exit");
}

sub config_ipv4_rem_ipv6 ($$%) {
my ($fc,$fh,%args) = @_;
my $rem_ipv6 = defined($args{'rem_ipv6'}) ? $args{'rem_ipv6'} : undef;
my $netmask = $fh->{'DATA_MASK1'};
my $n = 1;
my @dataport_names;
my $port_name;
logcomment('Reconfig the data ports');
my $dataport = 'DATA_PORT' . $n;
$port_name = $fh->{$dataport};
my $ifc_ip = 'IP_' . $port_name;
$ifc_ip = $fh->{$ifc_ip};

while (($port_name ne '') && ($ifc_ip ne '')) {
    logcomment("dataport is $dataport ");
    push(@dataport_names, $port_name);
    logcomment("dataport_name is $port_name");
    ++$n;
    $dataport = 'DATA_PORT' . $n;
    $port_name = $fh->{$dataport};
$ifc_ip = 'IP_' . $port_name;
$ifc_ip = $fh->{$ifc_ip};

}

foreach my $portname (@dataport_names) {
    my $ifc_port = 'IFC_' . $portname;
    logcomment("ifc_port is $ifc_port");
    my $ifc_ip = 'IP_' . $portname;
    $ifc_ip = $fh->{$ifc_ip};
    logcomment("ifc_ip is $ifc_ip");
    do_ifconfig($fc, $portname,
                ip_addr => $ifc_ip,
                state   => 'up',
                netmask => $netmask);

    if ($rem_ipv6) {
        logcomment('Remove the ipv6 address');
        say($fc, "ifconfig $portname 0::0 ");
        filer_options_set($fc, "ip.v6.enable", "off");
        logcomment('Removed the ipv6 address');
    }
}
}

1;
