# Perl Module
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Utils/IPv6.pm#1 $ 
# vim: ft=perl tabstop=4 expandtab
#
# Purpose: IPV6 utilities
#          Use NACL to perform IPv6 checks of common functions
#          specific to IPv6
#
## @summary  
## NACL based IPv6 utilities 
##
## @description 
## NACL based utilities to check IPv6 support an if v6 is enabled.
## Methods provided to gather info and then check and enable.
##
## @author      parsons@netapp.com
##
## @pod here
#
# POD at end of file
#


package Network::Utils::IPv6;

use strict;

# NATE
use Options;
use Params::Common;
#use Tharn qw($Log);
use NATE::Log qw(log_global);

# NACL
use NACL::C::NetworkOptionsIpv6;


my $Log = log_global(1);


BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter);
}

# Define TAGs
our @TAG_1 = qw();
our @TAG_2 = qw();

# Exports

our @EXPORT  = (
   @TAG_1,
   @TAG_2,
);

our @EXPORT_OK  = qw();
our %EXPORT_TAGS = (
   'tag_1'   => [@TAG_1],
   'tag_2'   => [@TAG_2],
);

our $ERROR;
#
# IPV6_SUPPORTED defined here for use in thpls, not used in this lib
# IPV6_ENABLED is used in the lib to determine if v6 should be enabled
#
my $Options = new Options;
$Options->parameters(definitions=> [
    {   'name'        => 'IPV6_SUPPORTED',
        'description' => 'IPv6 support flag',
    },
    {   'name'        => 'IPV6_ENABLED',
        'description' => 'Enable IPv6 if not enabled',
        'default'     => 1,
    },
    ]
);
$Options->process();
my $options = $Options->get_hash();



#=============================================================================
#
#  Constructor
#  Object to store information pertaining to IPv6
#  Constructor can perform the checks if the arg is set, will call
#  methods the same way a user could call thme individually
#
#  Args
#   nacl_c_node   => <ref>    ref NACL::C::Node->find() list  (required)  
#   check         => <0|1>    run the check (optional)
#
# Return
#   wantarray ? (ref, error) : ref );
#
sub new {
    my $_proto = shift;
    my $_class = ref($_proto) || $_proto;
    my $_self  = {};
    # bless object
    bless($_self, $_class);

    $ERROR = '';
    # get object arguments
    my %_arg = @_;
    my ($_a, $_n);
    foreach $_a (keys %_arg) {
      if    ($_a =~ /^-?nacl_c_node$/i)    { $_self->{'input'}{'nacl_nodes'}    = delete($_arg{$_a}); }
    }

    # Check Args
    #
    # check given nacl_c_node arg
    if ( !defined($_self->{'input'}{'nacl_nodes'}) ) {
        $ERROR = "option required:  nacl_c_node";
        return( wantarray ? (undef, $ERROR) : undef );
    }
    # validate it is a NACL::C::Node object
    # If CLI is given as FILER_LIST=NODES NODES=f1
    # this will still work in list context 
    foreach $_n ( @{$_self->{'input'}{'nacl_nodes'}} ) {
        if ( !Scalar::Util::blessed($_n) || !$_n->isa('NACL::C::Node')) {
            $ERROR = "nacl_c_node element not a NACL::C::Node isa " . ref($_n);
            return( wantarray ? (undef, $ERROR) : undef );
        }
    }

   # check if constructor should run checks
   # not required as can be called indivudally by user
   if ( $_arg{'check'} ) {
      $_self->check_ipv6_support;
      $_self->check_ipv6_enabled;
   }

    return( wantarray ? ($_self, $ERROR) : $_self )
}

#-----------------------------------------------------------------------------
#
# Method to check if IPv6 is supported on each node
# As this uses NACL::C::_Version.pm, only supports CMode
#
# Populate object
#    $_obj->{'check'}{'v6_support'}{<nodename>} = <0|1>
#
# Args
#   none   work on object
#
# Return 
#  wantarray ? (status, ERROR) : status );
#
#
#
sub check_ipv6_support {

   my $_obj = shift;
   my ($_node, $_nodename, $_bool);
   my $_c = 0;
   $ERROR = '';

   $Log->comment("Network::Utils::IPv6::check_ipv6_support checking nodes");
   # loop through each node and see if ipv6 is supported
   foreach $_node (  @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = \$_node->hostrec->{'hostname'};
      $Log->comment("Network::Utils::IPv6::check_ipv6_support node: $$_nodename");
      $_bool = $_node->is_feature_supported( 'feature' => 'ipv6_feature');

      # set object 
      if ( $_bool ) { $_obj->{'check'}{'v6_support'}{$$_nodename} = 1; }
      else          { $_obj->{'check'}{'v6_support'}{$$_nodename} = 0; }

      # log settings
      $Log->comment("Network::Utils::IPv6::check_ipv6_support: $$_nodename  " 
                    . ($_obj->{'check'}{'v6_support'}{$$_nodename} ? 'supported' : 'not supported')
      );
      $_c++;
   }

   return( wantarray ? ($_c, $ERROR) : $_c );
}

#---------------------------------------------------------------------------
#
# Check object content to see if node supports ipv6
# If object does not have the information, will get it
#
# Args
#  node  => node name
#
# Return
# wantarray ? (status, ERROR) : status );
#

sub is_ipv6_supported {
   my $_obj = shift;
   my ($_node, $_nodename, $_bool);
   my $_r = 0;

   # get object arguments
   my %_arg = @_;

   # make sure we have v6_support info in object
   if (!exists($_obj->{'check'}{'v6_support'}) ){
      $_obj->check_ipv6_support;
   }

   $ERROR = "node $_arg{'node'} not found in objects node list";

   # loop thru nodes in object
   foreach $_node (  @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = \$_node->hostrec->{'hostname'};
      # object node name matchs arg
      if (  $_arg{'node'} eq $$_nodename ) {
         # object support setting indicating support
         if ( exists($_obj->{'check'}{'v6_support'}{$$_nodename}) ) {
            if (  $_obj->{'check'}{'v6_support'}{$$_nodename} ) {
               $_r++;
               $ERROR = '';
            }
            else {
               $ERROR = "node $_arg{'node'} ipv6 not supported";
            }
         }
         last;
      }
   }

   return( wantarray ? ($_r, $ERROR) : $_r );
}



#-----------------------------------------------------------------------------
#
# Method to check if IPv6 functionality (enable/disable) on a node
# IPv6 enabling is cluster-wide but this does not assume
# all nodes passed to the method are in same cluster, so will
# individually check all nodes
#
# Populate Object
#   $_obj->{'check'}{'v6_enabled'}{<nodename>} = <0|1>;
#   
# Args
#   none   work on object
#   
# Return 
#  wantarray ? (status, ERROR) : status );
#

sub check_ipv6_enabled {

   my $_obj = shift;
   my ($_node, $_nodename, $_ipv6);
   my $_c = 0;
   $ERROR = '';

   $Log->comment("Network::Utils::IPv6::check_ipv6_enabled checking nodes");
   # loop through each node and see if ipv6 is enabled
   foreach $_node (  @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = \$_node->hostrec->{'hostname'};
      $Log->comment("Network::Utils::IPv6::check_ipv6_enabled node: $$_nodename");
      $_ipv6 = NACL::CS::NetworkOptionsIpv6->fetch(
         'command_interface' => $_node
      );

      # set objects
      if ( $_ipv6->{'enabled'} =~ /true/) { $_obj->{'check'}{'v6_enabled'}{$$_nodename} = 1; }
      else                                { $_obj->{'check'}{'v6_enabled'}{$$_nodename} = 0; }

      # log settings
      $Log->comment("Network::Utils::IPv6::check_ipv6_enabled: node $$_nodename  "
                    . ($_obj->{'check'}{'v6_enabled'}{$$_nodename} ? 'enabled' : 'not enabled')
      );
      $_c++; 
   } 
   return( wantarray ? ($_c, $ERROR) : $_c );
}

#---------------------------------------------------------------------------
#
# Check object content to see if node has ipv6 enabled
#
# Args
#  node  => node name
#
# Return
# wantarray ? (status, ERROR) : status );
#

sub is_ipv6_enabled {
   my $_obj = shift;
   my ($_node, $_nodename, $_bool);
   my $_r = 0;

   # get object arguments
   my %_arg = @_;

   # make sure we have v6_enabled info in object
   if (!exists($_obj->{'check'}{'v6_enabled'}) ){
      $_obj->check_ipv6_enabled;
   }

   $ERROR = "node $_arg{'node'} not found in objects node list";

   # loop thru nodes in object
   foreach $_node (  @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = \$_node->hostrec->{'hostname'};
      # object node name matchs arg
      if (  $_arg{'node'} eq $$_nodename ) {
         # object enable setting exists
         if ( exists($_obj->{'check'}{'v6_enabled'}{$$_nodename}) ) {
            # v6 is enabled
            if (  $_obj->{'check'}{'v6_enabled'}{$$_nodename} ) {
               $_r++;
               $ERROR = '';
            }
            # v6 not enabled
            else {
               $ERROR = "node $_arg{'node'} ipv6 not enabled";
            }
         }
         last;
      }
   }
   return( wantarray ? ($_r, $ERROR) : $_r );
}


#-----------------------------------------------------------------------------
#
# Method to enable IPv6 based on IPV6_ENABLE value
# If enabled and desired, log message
# if enabled and not desired, return error
# if not enabled and desired, enable and update object
# if not enabled and not desired, log message 
#
# Note enabling is cluster wide but all nodes are checked as
# not all node may be in same cluster
#
# Args 
#   none required
#   node => <str>    optional, if given just check this node
#
# Return
#  wantarray ? (status, error) : status );
#
sub enable_ipv6 {
   my $_obj = shift;
   my ($_node, $_nodename, $_enabled, $_ipv6_modify);
   my $_r = 0;
   $ERROR = '';

   # get object arguments
   my %_arg = @_;

   # make sure we have v6_enabled info in object
   if (!exists($_obj->{'check'}{'v6_enabled'}) ){
      $_obj->check_ipv6_enabled;
   }


   $Log->comment("Network::Utils::enable_ipv6: checking ipv6 enabling "
               . "IPV6_ENABLED=$options->{'IPV6_ENABLED'}"
   );

   # loop thru nodes in object
   foreach $_node (  @{$_obj->{'input'}{'nacl_nodes'}} ) {
      $_nodename = \$_node->hostrec->{'hostname'};

      $Log->comment("Network::Utils::enable_ipv6: node $$_nodename");

      if ( defined($_arg{'node'}) && $_arg{'node'} ne $$_nodename  ) {
         $Log->comment("Network::Utils::enable_ipv6: only enabling $_arg{'node'} "
                      . "skipping enabling of $$_nodename"
         );
         next;
      }

      # object enable setting exists
      if ( exists($_obj->{'check'}{'v6_enabled'}{$$_nodename}) ) {
         $_enabled = \$_obj->{'check'}{'v6_enabled'}{$$_nodename};
         # v6 enabled, desired
         if (  $$_enabled && $options->{'IPV6_ENABLED'} ) {
            $Log->comment("Network::Utils::IPv6::enable_ipv6: enabled as desired");
            $_r++;
         }
         # v6 enabled, not desired  ERROR out
         elsif (  $$_enabled && !$options->{'IPV6_ENABLED'}) {
            $ERROR = "v6 enabled not desired on $$_nodename";
            $Log->comment("Network::Utils::IPv6::enable_ipv6: $ERROR");
            $_r = 0;
            last;
         }
         # v6 not enabled, desired   perform enabling
         elsif ( !$$_enabled && $options->{'IPV6_ENABLED'} ) {
            $Log->comment("Network::Utils::IPv6::enable_ipv6: not enabled but desired");
            # enable ipv6
            $_ipv6_modify = NACL::C::NetworkOptionsIpv6->modify(
                 'command_interface' => $_node,
                 'enabled'           => 'true'
            );
            # update object
            $_obj->check_ipv6_enabled;
            $_r++;
         }
         # v6 not enabled, not desired
         elsif (!$$_enabled && !$options->{'IPV6_ENABLED'} ) {
            $Log->comment("Network::Utils::IPv6::enable_ipv6: not enabled not desired");
            $_r++; 
         }
         # assert
         else {
            $ERROR = "Network::Utils::IPv6::enable_ipv6: could not resolve enabling "
                  . "enable: $$_enabled IPV6_ENABLED: $options->{'IPV6_ENABLED'}";
            $_r = 0;
            last;
         }
      }   
   } 
   return( wantarray ? ($_r, $ERROR) : $_r);

}



1;

__END__

=pod

=head1 NAME Network::Utils::IPv6

=head1 DESCRIPTION

    Perl Object to provide common IPv6 functions

    Object is given a refence to a NACL Nodes list. Methods will contact
    the known nodes and get IPv6 feature support status and if IPv6 is enabled
    or not. Methods can then be used to get the respective support or enabling
    status. Method is provided to enable IPv6.
   

=head1 SYNOPSIS

    use Network::Utils::IPv6;

    @Nodes = NACL::C::Node->find();
    ($IPv6, $error) = Network::Utils::IPv6->new(
        'nacl_c_node'  => \@Nodes,
        'check'        => 1,
    );

    $IPv6->check_ipv6_support;

    $IPv6->check_ipv6_enabled;


    ($ok, $error) = $IPv6->is_ipv6_supported('node' => $nodename);
    ($ok, $error) = $IPv6->is_ipv6_enabled('node' => $nodename);
    ($ok, $error) = $IPv6->enable_ipv6('node' => $nodename);

=head1 PARAMS

=over

=item IPV6_SUPPORTED=<0|1>
Defined in this lib for user to make use of

=item IPV6_ENABLED=<0|1>
Defined in this lib and made use of by the lib. If non-zero
and IPv6 is not enabled, then method will enable IPv6 on the
respective node and update the object.

=back

=head1 Methods

=head2 new

   Object constructor

   ($IPv6, $error) = Network::Utils::IPv6->new(
        'nacl_c_node'  => \@Nodes,
        'check'        => 1,
   );

=over

=item nacl_c_node => <ref to NACL Node list>

    Required. Reference to a NACL Node list as made with NACL::C::Node->find();
    The reference will be check to contain NACL::C::Node object.

=item check => <0|1>

   Optional. If set the constructor will load the object by calling the 
   check_ipv6_support and check_ipv6_enabled methods. otherwise the 
   user must run these when desired. Note that either of these methods 
   can be called by the user at any time

=back

=head2 check_ipv6_support

    $IPv6->check_ipv6_support;
    $ok = $IPv6->check_ipv6_support;
    ($ok, $error) = $IPv6->check_ipv6_support;

    Will use each node object stored in the object and use the 
    NACL is_feature_supported to check for IPv6 support.

        $_bool = $_node->is_feature_supported( 'feature' => 'ipv6_feature');

    Boolean (0|1) value will be stored in the object.

        $_obj->{'check'}{'v6_support'}{<nodename>} 

    Will return in list context the count of nodes checked and error string.
    In scalar context will just return the count of nodes checked.
    Currently there is no errored condition checked to make an error string.


=head2 is_ipv6_supported

     ($ok, $error) = $IPv6->is_ipv6_supported( 'node' => <nodename>);

=over

=item node => <nodename>

    Name of node to return IPv6 support status

=back

    Will look for the given node name in the object and return
    a the status of IPv6 support. If object does not have
    status info then check_ipv6_support method is used to first
    obtain the needed info.


    If node is found and IPv6 is supported then status = 1
    If node is found and IPv6 is not supported then status = 0 and error returned
    If node is not found then status = 0 and error string indicating
    node not found


=head2 check_ipv6_enabled

    $IPv6->check_ipv6_enabled;
    $ok = $IPv6->check_ipv6_enabled;
    ($ok, $error) = $IPv6->check_ipv6_enabled;

    Will use each node object stored in the object and 
    pass to NACL::CS::NetworkOptionsIpv6->fetch() to get
    IPv6 enabl/disable status.


    Boolean (0|1) value will be stored in the object.

        $_obj->{'check'}{'v6_enabled'}{<nodename>} 

    Will return in list context the count of nodes checked and error string.
    In scalar context will just return the count of nodes checked.
    Currently there is no errored condition checked to make an error string.


=head2 is_ipv6_enabled

     ($ok, $error) = $IPv6->is_ipv6_enabled( 'node' => <nodename>);

=over

=item node => <nodename>

    Name of node to return IPv6 enabled status

=back

    Will look for the given node name in the object and return
    a the status of IPv6 enable. If object does not have
    status info then check_ipv6_enabled method is used to first
    obtain the needed info.

    If node is found and IPv6 is enabled then status = 1
    If node is found and IPv6 is not enabled then status = 0 and error returned
    If node is not found then status = 0 and error string indicating
    node not found

=head2 enable_ipv6

    $IPv6->enable_ipv6;
    $IPv6->enable_ipv6( 'node' => <nodename> );
    $ok = $IPv6->enable_ipv6;
    $ok = $IPv6->enable_ipv6( 'node' => <nodename> );
    ($ok, $error) = $IPv6->enable_ipv6;
    ($ok, $error) = $IPv6->enable_ipv6( 'node' => <nodename> );

    Will check IPv6 enable status, if not enabled  and desired
    then will use NACL::C::NetworkOptionsIpv6->modify() to enable
    and update the object
    Though Ipv6 eanbling is cluster wide, each node will be check
    as the passed in node list may not be a single cluster.
    If a node in a cluster is enabled by the object, then the 
    check_ipv6_enabled method is called to update the object.

=over

=item node => <nodename>

    Optional: Name of node to enable IPv6 on.

    Will enable IPv6 on all nodes if IPV6_ENABLED=1 or any
    non-zero value. if a node name is given, then will only
    enable IPv6 on that node if nodename is found in the object.

=back

Operation

=over

=item If IPv6 is enabled and IPV6_ENABLED=1, 

    Log message indicating, return status=1, blank error string

=item If IPv6 is enabled and IPV6_ENABLED=0,

    Log message indicating, return status=0, error string

=item If IPv6 is not enabled and IPV6_ENABLED=1, 

    Log message, enable with NACL::C::NetworkOptionsIpv6->modify() 
    and update object with check_ipv6_enabled method. 
    Return status=1, blank error string
    
=item If IPv6 is not enabled and IPV6_ENABLED=0, 

    Log message indicating, return status=1, blank error string 

=back
    

=head1 Functions

No exported function

=head1 Example

    use strict;
    use Tharn qw($Log);
    use Options;
    use Network::Utils::IPv6;
    use NACL::C::Node;
    
    my ( $Options, @Nodes, $IPv6,
         $options, $ok, $error, $node,
    );
    
    $Options = Options->new();
    $Options->process();
    $options = $Options->get_hash();
    
    # Get a list of NACL Nodes
    @Nodes = NACL::C::Node->find();
    
    ($IPv6, $error) = Network::Utils::IPv6->new(
            'nacl_c_node'  => \@Nodes,
            'check'        => 1,
    );
    
    # check support and enable status
    foreach $node ( @Nodes ) {
        $Log->comment("Check support on node: " . $node->hostrec->{'hostname'} );
        ($ok, $error) = $IPv6->is_ipv6_supported('node' => $node->hostrec->{'hostname'});
        $Log->comment("support results: $ok " . ($ok ? 'supported' : "not supported $error") );
    
        $Log->comment("Check enabled on node: " . $node->hostrec->{'hostname'} );
        ($ok, $error) = $IPv6->is_ipv6_enabled('node' => $node->hostrec->{'hostname'});
        $Log->comment("enabled results: $ok " . ($ok ? 'enabled' : "not enabled: $error") );
    
    }
    
    # enable
    ($ok, $error) = $IPv6->enable_ipv6;
    $Log->comment("enabling results: $ok " . ($ok ? 'enabled' : "not enabled: $error") );
    
    exit(0)


=head1 STATUS

   Current 

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2013, Network Appliance., All Rights Reserved

=cut















