#
# $Id: //depot/prod/test/main/lib/VifMgr/Failover.pm#3 $
# $Revision: #3 $ submitted on $Date: 2006/10/04 $ By $Author: ewing $
#
## @summary     VifMgr Failover Superclass
## @author      ewing@netapp.com
## @pod here

=head1 NAME

VifMgr::Failover

=head1 SYNOPSIS

none

=head1 DESCRIPTION

Use one of the subclasses inheriting from this class to perform VIF Failover.

=over

C<VifMgr::Failover::Migrate>

=over

or

=back

C<VifMgr::Failover::PortFail>

=back

=head1 SEE ALSO

L<VifMgr::Failover::Migrate>
L<VifMgr::Failover::PortFail>
L<Options>
L<TracePkg>
L<TestUtils>

=head1 BUGS

Yes sir!

=head1 STATUS

  In-progress

=head1 AUTHOR

  Steven Ewing (ewing@netapp.com)
  $Id: //depot/prod/test/main/lib/VifMgr/Failover.pm#3 $
  Last revised $Date: 2006/10/04 $

=cut

package Network::VIFMgr::Failover;

use strict;
use warnings;

#####NACL##########
use NACL::C::Node;

use NATE::Process;
use NATE::Log qw(log_global);
use NATE::BaseException qw(:try);

use TestUtils;
use File::Find;
use Options;

use Tharn;
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF UNDEF OBJECT);
use Carp qw(confess);
use Cwd qw(abs_path);

use constant LOCAL       => "LOCAL";
use constant REMOTE      => "REMOTE";
use constant RANDOM      => "RANDOM";
use constant POWER_CYCLE => "POWER_CYCLE";
use constant DOWN_PORT   => "DOWN_PORT";

#$Log->set_options( separateNewline => 1 );
my %find_in_natelib;

my $CLIENT;
my $Log = log_global();

=head2 new

The subclasses of VifMgr::Failover accept these

=over

ping_checker  - ping checking binary/script
ping_count    - count of pings
ping_interval - interval of pings

=back 

The interface 

=cut

# Create and define the options

my $Options = new Options();

$Options->parameters(
			  definitions => [
				  {
					name        => 'GARP_CHECK',
					description =>
					  'Used to enable or disable checking of gratuitous ARPs.',
					spec => 's',
				  },
			  ]
);
$Options->process();

#############################################################################
# Subroutine: new
# Parameters: Initialization Parameters
# Returns: Blessed object
# Purpose: Object constructor
##############################################################################

sub new {

	$Log->enter();

	my $class = shift();
	my %p     = (@_);

	my $self = ();

	# common Failver class information. Name of the OS under which this
	# copy of perl was built.

	$self->{_os} = $^O;

	# these aren't really user-servicable
	# Set options or their defaults if not specified.

	$self->{_garp_check_count}    = $p{garp_count}    || 3;
	$self->{_garp_check_interval} = $p{garp_interval} || 2;
	$self->{_garp_check_fatality} = $p{garp_fatality} || 1;

	$self->{_ping_checker} = $p{ping_checker}
	  || $class->whereis_binary("ping");
	$self->{_ping_check_count} = $p{ping_count} || 5;
	$self->{_ping_check_count_req} = $p{ping_count_req}
	  || $self->{_ping_check_count};
	$self->{_ping_unreach_allow}  = $p{ping_unreach_allow} || 0;
	$self->{_ping_check_protocol} = $p{ping_protocol}      || 'external';
	$self->{_ping_check_interval} = $p{ping_interval}      || 0.5;
	$self->{_ping_check_timeout}  = $p{ping_timeout}       || 1;
	$self->{_ping_check_bytes}    = $p{ping_bytes}         || 1024;

	bless( $self, $class );

	# Store references to useful objects

	$self->{_node}    = NACL::C::Node->find();
	$self->{_options} = new Options();

	my $o = $self->{_options}->get_hash();

	$CLIENT = $o->{CLIENT}[0];

	if ( $o->{GARP_CHECK} ) {
		$self->{_garp_checker} = $p{garp_checker}
		  || $self->find_in_natelib(
							"nblade/vifmgr/net/util/garpfilter/$^O/garpfilter");
	}
	else {
		$Log->debug(   "Gratuitous arp checking disabled by parameter. "
					 . "Don't look for a full path." );
	}

	local $Data::Dumper::Maxdepth = 2;
	$Log->debug(   "Initialized a new VifMgr::Failover object with "
				 . "this config:\n"
				 . Dumper($self) );

	$Log->exit("Returning blessed reference: $self");
	return $self;

} ## end sub new

#############################################################################
# Subroutine: whereis_binary
# Parameters: Binary to look for
# Returns: Path to specified binary
# Purpose: Searches some common paths for binaries used by this class
#############################################################################

sub whereis_binary {

	$Log->enter();

	# Shift off the class reference and check for the binary we are searching
	# for

	shift();
	my $binary = shift()
	  || $Log->error("A binary is required");

	# Define paths to search

	my @bin_dir = qw(
	  /bin
	  /sbin
	  /usr/bin
	  /usr/sbin
	  /local/bin
	);

	# Iterate through each directory, searching for the binary

	foreach my $dir (@bin_dir) {
		if ( -x "$dir/$binary" ) {
			$Log->comment("Found $dir/$binary");
			return "$dir/$binary";
		}
		else {
			$Log->debug("$dir/$binary not fou-nd.");
		}
	}

	# Exit

	$Log->comment("Couldn't find $binary in any path");
	return undef();
} ## end sub whereis_binary

#############################################################################
# Subroutine: find_in_nate_lib
# Parameters: Binary to search for
# Returns: Natelib path to binary
# Purpose: To be able to search the NateLib for binaries
##############################################################################

sub find_in_natelib {

	$Log->enter();

	# Shift off the class ref and the binary we are searching for

	my $self   = shift();
	my $binary = shift();

	# Get the absolute path of the binary and return if found

	my $found_bin = abs_path($binary);
	$Log->debug(   "Found Cwd::abs_path returned "
				 . ( defined($found_bin) ? $found_bin : "an empty string" )
				 . "." );

	if ( defined($found_bin) ) {
		$Log->exit("Found $binary at $found_bin (Cwd::abs_path method).");
		return $found_bin;
	}

	# Otherwise, continue checking if we previously found it

	if ( exists( $find_in_natelib{$binary} ) ) {
		$Log->exit(   "Found $binary at $find_in_natelib{$binary} "
					. "(Previously cached File::Find::find method)." );
		return $find_in_natelib{$binary};
	}

	# What are the NATE_LIB paths we can check

	my @available_paths =
	  grep( { -d $_ }
			split( /;/, $self->{_options}->get( option => 'NATE_LIB' ) ) );

	# Search

	my $pat = qr/$binary\z/;

	find(
		sub {

			# Don't remove the braces below!
			# P4 keeps expanding the $File$ as if it were a keyword.
			if ( ${File::Find::name} =~ m/$pat/s ) {
				$found_bin = $File::Find::name;
				$find_in_natelib{$binary} = $found_bin;
			}
		},
		@available_paths
	);

	$Log->exit("Found $binary at $found_bin");
	return $found_bin;
} ## end sub find_in_natelib

#############################################################################
# Subroutine: failover_event
# Parameters: none
# Returns: none
# Purpose: This is the method that actually implements the failover event for a
# particular class.  The VifMgr::Failover implementation is only here to serve
# as a template for subclasses.
#############################################################################

sub failover_event {

	$Log->enter();

	confess("Pure virtual method called.");

	$Log->exit();
}

#############################################################################
# Subroutine: garp_check
# Parameters: IP to ARP for
# Returns: Exit status
# Purpose: This is a utility useful to subclasses which checks for gratuitous
#          ARPs emitted by a failed interface.
#############################################################################

sub garp_check {

	$Log->enter();
	my $self = shift();

	# Define the params and their defaults

	my %p = (
			  id   => 'garp_check',
			  ip   => $self->{_vif_address},
			  mask => $self->{_vif_netmask},
			  @_,
	);

	my $o = $self->{_options}->get_hash();

	# Are we flagged to skip this?

	if ( !$o->{GARP_CHECK} ) {
		$Log->exit("Gratuitous arp checking disabled by parameter.");
		return 0;
	}

	# Find a network device to use if one hasn't already been found.

	if ( !defined( $self->{_network_device} ) ) {
		eval {
			$self->{_network_device} = $self->_select_network_device(    ip   => $p{ip},
																mask => $p{mask} );
		};
		if ($@) {
			$Log->warn(
				  "No Network device selection is availalble on this platform");
			$Log->exit();
			return;
		}
	} ## end if (!defined($self->{_network_device...

	$Log->comment(
				 "Network device selected for check: $self->{_network_device}");

	# Subtest the binary which checks for Gratuitous ARPs.

	my $st;

	try {
		$st = NATE::Process->new(
			codespec => "/usr/local/bin/sudo",
			runid    => $p{id},
			args     => [
				 $self->{_garp_checker},
				 $p{ip},
				 $self->{_garp_check_count},
				 $self->{_network_device},
				 (
				   $self->{_garp_check_interval} * $self->{_garp_check_count} *
					 10
				 ),
				 "1"
			],
		);
		
		$st->start();
		$st->wait();
	  }
	  otherwise{
	  	$Log->warn("Failed while checking for gratuitous ARPs.");
	  }

	  #eval {
	  #	$st = Tharn::subtest(
	  #		"/usr/local/bin/sudo",
	  #		-runid,
	  #		$p{id},
	  #		'--',
	  #		$self->{_garp_checker},
	  #		$p{ip},
	  #		$self->{_garp_check_count},
	  #		$self->{_network_device},
	  #		( $self->{_garp_check_interval} * $self->{_garp_check_count} * 10 ),
	  #		"1"
	  #	);
	  #};

	  # Was there an error?

	  #if ($@) {
	  #	if ( $self->{_garp_check_fatality} ) {
	  #		$Log->error( "Failed while checking for gratuitous ARPs.", );
	  #	}
	  #	else {
	  #		$Log->warn("Failed while checking for gratuitous ARPs.");
	  #	}
	  #	} ## end if ($@)

	  $Log->comment( "Exit Status: " . $st->{exit_status});
	return $st->results();
} ## end sub garp_check

#############################################################################
# Subroutine: add_vif_watch
# Parameters: Vif/Vsersever or Address to watch
# Returns: void
# Purpose: Add to vif_watch_list
#############################################################################

sub add_vif_watch {

	$Log->enter();

	# Define the parameters

	my ( $self, %p ) = validate_mixed(
									   @_,
									   [ { type => OBJECT } ],
									   {
										 address => {
													  type    => SCALAR,
													  default => undef()
										 },
										 vserver => {
													  type    => SCALAR,
													  default => undef(),
													  depends => [qw(vif)]
										 },
										 vif => {
												  type    => SCALAR,
												  default => undef(),
												  depends => [qw(vserver)]
										 },
									   }
	);

	my %show_parameters;

	if ( !defined( $p{address} ) && !defined( $p{vif} ) ) {
		$Log->error("Vserver/VIF or VIF address must be specified.");
	}

	# Save the show_parameters

	if ( defined( $p{address} ) ) {
		$Log->debug("Preferring any supplied vif/vserver over address.");
		$show_parameters{address} = $p{address};
	}
	else {
		$show_parameters{vif}    = $p{vif};
		$show_parameters{server} = $p{vserver};
	}

	# Issue the show, overwriting the params

	$Log->debug( "Getting config/checking for interface "
				 . Dumper( \%show_parameters ) );

	#my $vif = $self->{_node}->network_interface_show(%show_parameters);
	my $vif = NACL::CS::NetworkInterface->fetch(
											command_interface => $self->{_node},
											filter => {%show_parameters}, );
	$p{vif}     = $vif->{vif};
	$p{address} = $vif->{address};
	$p{vserver} = $vif->{server};
	$p{netmask} = $vif->{netmask};

	# Add it to the _vif_watch_list if its not there already

	if (
		 !grep( { ( $_->{vif} eq $p{vif} ) && ( $_->{vserver} eq $p{vserver} ) }
				@{ $self->{_vif_watch_list} } ) )
	{
		$Log->comment("Added $p{vif}\:$p{vserver} to the watch list.");
		push(
			  @{ $self->{_vif_watch_list} },
			  {
				 vif     => $p{vif},
				 vserver => $p{vserver},
				 address => $p{address},
				 netmask => $p{netmask}
			  }
		);
	}
	else {
		$Log->comment(   "Interface $p{vif}\:$p{vserver} is already on "
					   . "the watch list." );
	}

	# Print watch list and return

	$Log->debug(
				 "Watchlist: "
				   . join(
						   ", ",
						   map( {"$_->{vserver}\:$_->{vif}"}
								@{ $self->{_vif_watch_list} } )
				   )
	);
	$Log->exit();
} ## end sub add_vif_watch

#############################################################################
# Subroutine: delete_vif_watch
# Parameters: Vif/Vsersever or Address to no longer watch
# Returns: void
# Purpose: Delete from vif_watch_list
#############################################################################

sub delete_vif_watch {

	$Log->enter();

	# Define parameters

	my ( $self, %p ) = validate_mixed(
									   @_,
									   [ { type => OBJECT } ],
									   {
										 address => {
													  type    => SCALAR,
													  default => undef()
										 },
										 vserver => {
													  type    => SCALAR,
													  default => undef(),
													  depends => [qw(vif)]
										 },
										 vif => {
												  type    => SCALAR,
												  default => undef(),
												  depends => [qw(vserver)]
										 },
									   }
	);

	if ( !defined( $p{address} ) && !defined( $p{vif} ) ) {
		$Log->error("Vserver/VIF or VIF address must be specified.");
	}

	my $removed = 0;

	if ( defined( $p{address} ) ) {
		if ( defined( $p{vif} ) || defined( $p{vserver} ) ) {
			$Log->debug("Ignoring supplied vif/vserver over address.");
		}

		$Log->debug("Getting config for address $p{address}.");

		# Perform the show based on the specified address and overwrite the
		# parameters

		#  my $vif =
		#  $self->{_node}->network_interface_show(address => $p{address});

		my $vif = NACL::CS::NetworkInterface->fetch(
											command_interface => $self->{_node},
											filter => { address => $p{address} } );
		$p{vif}     = $vif->{vif};
		$p{vserver} = $vif->{server};
		$p{address} = $vif->{address};
	} ## end if (defined($p{address...

	# Remove it from the watch list if it exists

	if (
		 !grep( { ( $_->{vif} eq $p{vif} ) && ( $_->{vserver} eq $p{vserver} ) }
				@{ $self->{_vif_watch_list} } ) )
	{
		$Log->debug("Interface $p{vif}\:$p{vserver} wasn't on the watch list.");

	}
	else {

		$Log->debug("Removing $p{vif}\:$p{vserver} from the watch list.");

		for ( my $i = 0 ; $i <= $#{ $self->{_vif_watch_list} } ; $i++ ) {
			$Log->debug( "Checking ${$self->{_vif_watch_list}}[$i]->{vserver}\:${$self->{_vif_watch_list}}[$i]->{vif}"
			);

			if (   ${ $self->{_vif_watch_list} }[$i]->{vif} eq $p{vif}
				&& ${ $self->{_vif_watch_list} }[$i]->{vserver} eq $p{vserver} )
			{

				$Log->comment( "Removing index $i " . "("
						   . ${ $self->{_vif_watch_list} }[$i]->{vserver} . "\:"
						   . ${ $self->{_vif_watch_list} }[$i]->{vif}
						   . ") from watch list!." );
				splice( @{ $self->{_vif_watch_list} }, $i, 1 );
				$removed = 1;
				last;
			} ## end if (${$self->{_vif_watch_list...
		} ## end for (my $i = 0 ; $i <= ...
	} ## end else [ if (!grep({($_->{vif} ...

	# Did we not remove anything?

	if ( !$removed ) {
		confess(
				     "Couldn't find anything to remove from the watchlist.\n"
				   . "Watchlist: "
				   . join(
						   ", ",
						   map( {"$_->{vserver}\:$_->{vif}"}
								@{ $self->{_vif_watch_list} } )
				   )
		);
	}

	# Return

	$Log->debug(
				 "Watchlist: "
				   . join(
						   ", ",
						   map( {"$_->{vserver}\:$_->{vif}"}
								@{ $self->{_vif_watch_list} } )
				   )
	);
	$Log->exit();
} ## end sub delete_vif_watch

#############################################################################
# Subroutine: store_vif_state
# Parameters: None
# Returns: void
# Purpose: This is a convenience routine to store the configuration of an
#          interface that is about to be failed.
##############################################################################

sub store_vif_state {

	$Log->enter();
	my $self = shift();

	# Loop through each VIF in the vif_watch_list and save its current state

	foreach my $vif ( @{ $self->{_vif_watch_list} } ) {

		# TODO: make this work for multiple interfaces
		my $v = NACL::CS::NetworkInterface->fetch(
					command_interface => $self->{_node},
					filter => { lif => $vif->{vif}, vserver => $vif->{vserver} } );

	#  my $v = $self->{_node}->network_interface_show(    vif    => $vif->{vif},
	#											   server => $vif->{vserver} );
		$vif->{_vif_state} = $v;
	}

	$Log->exit();
} ## end sub store_vif_state

#############################################################################
# Subroutine: ping_check
# Parameters: Run ID for subtest
# Returns: void
# Purpose: Performs a ping of the failed over interface.
#############################################################################

sub ping_check {

	$Log->enter();
	my $self = shift();

	# Initialise parameters

	my %p = (
			  id => 'ping_check',
			  @_,
	);

	# Subtest the selected ping binary.

	#my $ping = NATE::Process->new {
	#				codespec => "/VifMgrNew/Vifconfig/"
	#}

	my $ping_args = " -c "
	  . $self->{_ping_check_count} . " -i "
	  . $self->{_ping_check_interval} . " "
	  . $self->{_vif_address};

	my $proc = NATE::Process->new(
								   codespec => $self->{_ping_checker},
								   args     => ["$ping_args"],
								   runid    => $p{id},
	);

	$proc->start;
	$proc->wait;

	#$Log->comment("##############################proc:".Dumper($proc));
	if ( $proc->{exit_status} ) {
		$Log->error( "Ping check of $self->{_vif_address} failed.", );
	}
	$proc->destroy;
	$Log->exit();
} ## end sub ping_check

#############################################################################
# Subroutine: garp_check_watch_list
# Parameters: Flag to skip non failable interfaces
# Returns: void
# Purpose: Perform a Gratuitous ARP check for VIFs in vif_watch_list
##############################################################################

sub garp_check_watch_list {

	$Log->enter();

	# Initialize parameters

	my ( $self, %p ) = validate_mixed(
									   @_,
									   [ { type => OBJECT } ],
									   {
										 skip_nonfailable => {
															  type    => SCALAR,
															  default => undef()
										 },
									   }
	);

	local $self->{_vif_address};

	# Iterate through VIF watch list

	foreach my $vif ( @{ $self->{_vif_watch_list} } ) {
		$Log->debug(   "Trying to start a Gratuitous ARP check for "
					 . "$vif->{vserver}\:$vif->{vif}" );

		# Check to see if there are availabl nodes/ports to fail to

		my ( $fail_node, $fail_port );
		if ( $p{skip_nonfailable} ) {

			( $fail_node, $fail_port ) = $self->failport_available(
													 vserver => $vif->{vserver},
													 vif     => $vif->{vif} );

			if ( !( defined($fail_node) || defined($fail_port) ) ) {
				$Log->comment(   "Unable to find an availalble failport for "
							   . "$vif->{vserver}\:$vif->{vif}" );
				next();
			}
		} ## end if ($p{skip_nonfailable...

		# Since we have available nodes/ports to fail to, we can now build the
		# array of subtests

	my $gpcheck = NATE::Process->new(
			codespec => "/usr/local/bin/sudo",
			runid    =>"garp_checker__$vif->{address}",
			args     => [
				 $self->{_garp_checker},
				 $vif->{address},
				 $self->{_garp_check_count},
				 $self->{_network_device},
				 (
				   $self->{_garp_check_interval} * $self->{_garp_check_count} *
					 10
				 ),
				 "1"
			],
		);

	
		push(
			@{ $self->{watch_list_garp_subtest} }, $gpcheck,
		);

# XXX This relies on magic, beware
#$self->{watch_list_garp_subtest}->[$#{$self->{watch_list_garp_subtest}}]->{ip_address} = $vif->{address};
	} ## end foreach my $vif (@{$self->{...

	$Log->exit();
} ## end sub garp_check_watch_list

#############################################################################
# Subroutine: wait_garp_check_watch_list
# Parameters: None
# Returns: void
# Purpose: Waits for the Gratuitous ARP subtests to finish
#############################################################################

sub wait_garp_check_watch_list {

	$Log->enter();
	my $self = shift();

	# Flag for the wait

	# Iterate through VIF watch list

	foreach my $garp_checker ( @{ $self->{watch_list_garp_subtest} } ) {

	$garp_checker->wait();
#if (NATE::Result::worst($garp_checker->own_results())->severity() >= NATE::Result::Fail::severity()) {

		# Check exit status of this current subtest

		if ( $garp_checker->exit_status() ) {
			$Log->warn(   "Failed to detect all required gratuitous "
						. "ARPs.  Found "
						. $garp_checker->exit_status()
						. ", should have got "
						. $self->{_garp_check_count}
						. "." );
		}
		else {
			$Log->comment(   "Found the desired number of gratuitous ARPs "
						   . "(exit: "
						   . $garp_checker->exit_status()
						   . ")" );
		}
	} ## end foreach my $garp_checker (@...

	$Log->exit();
} ## end sub wait_garp_check_watch_list

#############################################################################
# Subroutine: ping_check_watch_list
# Parameters: None
# Returns: void
# Purpose: Call ping_check for each VIF in the vif watch list
#############################################################################

sub ping_vif_watch_list {

	$Log->enter();
	my $self = shift();

	local $self->{_vif_address};

	# TODO: perform pings in parallel

	foreach my $vif ( map( { $_->{address} } @{ $self->{_vif_watch_list} } ) ) {
		$self->{_vif_address} = $vif;
		$self->ping_check();
	}

	$Log->exit();
} ## end sub ping_vif_watch_list

#############################################################################
# Subroutine: check_vif_state
# Parameters: None
# Returns: void
# Purpose: Check the stored vif state and see if the interfaces have moved as
#          appropriate.  This is effectively a vif state diff utiltity.
#############################################################################

sub check_vif_state {

	$Log->enter();
	my $self = shift();

	my %return_diff;
	my $differences = 0;

	# Get the current config

	foreach my $vif ( @{ $self->{_vif_watch_list} } ) {

	#my $state = $self->{_node}->network_interface_show(  vif    => $vif->{vif},
	#											   server => $vif->{vserver} );

		my $state = NACL::CS::NetworkInterface->fetch(
					command_interface => $self->{_node},
					filter => { lif => $vif->{vif}, vserver => $vif->{vserver} } );
		my %comp_h;
		my %diff;

		# Store this VIF information in the "new" section of the
		# comparison hash.

		while ( my ( $k, $v ) = each( %{$state} ) ) {
			$comp_h{$k}->{new} = $v;
		}

		# Store each old value in a hash. comp_h will look like:
		# server -> new -> <new server>, server -> old -> <old server>
		# _vif_state was saved during the call to store_vif_state.
		# This MUST be called for comparison to work.

		while ( my ( $k, $v ) = each( %{ $vif->{_vif_state} } ) ) {
			$comp_h{$k}->{old} = $v;
		}

		# Compare the hashes and store the new values of the changed
		# keys in a third hash.

		foreach my $key ( keys(%comp_h) ) {

			if ( $comp_h{$key}->{new} ne $comp_h{$key}->{old} ) {
				$diff{$key} = $comp_h{$key}->{new};
				$differences++;
			}
		}

		# Create a hash to return of the differences

		if ( scalar( keys(%diff) ) ) {
			$return_diff{"$vif->{vserver}\:$vif->{vif}"} = \%diff;
		}
	} ## end foreach my $vif (@{$self->{...

	# Return the new values of the differences

	$Log->exit(
				"Differences "
				  . (
					  $differences
					  ? "($differences) "
					  : "not "
				  )
				  . "found"
	);
	return \%return_diff;
} ## end sub check_vif_state

#############################################################################
# Subroutine: _select_network_device
# Parameters: None
# Returns: Selected Network Device
# Purpose: Helper to pick a network device for a particular network.
#############################################################################

sub _select_network_device {

	$Log->enter();
	my $self = shift();
	my $dev;

	# Call the OS-specific method.

	if ( $self->can("_select_network_device_$self->{_os}") ) {
		$dev = eval("\$self->_select_network_device_$self->{_os}(\@_)");
		if ($@) {
			die($@);
		}
	}
	else {
		$Log->error(   "No network device selection is possible " . "on $self->{_os}",
		);
	}

	$Log->exit();
	return $dev;
} ## end sub _select_network_device

#############################################################################
# Subroutine: _select_network_device_linux
# Parameters:
# Returns:
# Purpose: OS-specific network device selection
#############################################################################

sub _select_network_device_linux {

	$Log->enter();

	# Initialize parameters

	my ( $self, %p ) = validate_mixed(
									   @_,
									   [ { type => OBJECT } ],
									   {
										 mask          => { type => SCALAR },
										 ip            => { type => SCALAR },
										 route_command => {
														type    => SCALAR,
														default => '/sbin/route'
										 },
									   }
	);

	my @ip;
	my @mask;

	# Break IP and mask into octets

	@ip   = split( /\./, $p{ip} );
	@mask = split( /\./, $p{mask} );

	# Generate a network address from IP and mask

	my $maskedip = sprintf( "%d.%d.%d.%d",
							int( $ip[0] ) & int( $mask[0] ),
							int( $ip[1] ) & int( $mask[1] ),
							int( $ip[2] ) & int( $mask[2] ),
							int( $ip[3] ) & int( $mask[3] ) );

	$Log->debug("IP: \"$p{ip}\" Mask: \"$p{mask}\" Network: \"$maskedip\"\n");

	my $netpart;

	# Get our routing table so we can look for a device

	my $out = `$p{route_command}`;

	foreach my $l ( split( /\n/, $out ) ) {

		# Check each line for the network address.

		if ( $l =~
			m/$maskedip\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/
		  )
		{

			# When a line with the network address is found, check the mask.

			if ( $2 eq $p{mask} ) {

				# If the mask is correct, then we have our device.

				$Log->trace("Device $7 ($maskedip/$2)\n");
				$self->{_network_device} = $7;
				$Log->exit();
				return $7;
			}
		} ## end if ($l =~ ...
	} ## end foreach my $l (split(/\n/, ...
	$Log->error(
			"No interface on this host for network: " . "$maskedip/$p{mask}.\n", );
	$self->{_network_device} = undef();
	$Log->exit();
	return undef();
} ## end sub _select_network_device_linux

#############################################################################
#                                  DESTRUCTOR
#############################################################################

sub DESTROY {

	$Log->enter();

	my $self = shift();

	$Log->exit();
}

1;

__END__

#use Net::Ping;
 
    ############################################################################ 
    Parameter validation

    my ($class, %p) = validate_mixed(
        @_,
        [ {type => SCALAR } ],
        {
            # common: gratuitous arp config
            garp_checker => {type => SCALAR, default => "nblade/vifmgr/net/util/garpfilter/$^O/garpfilter"},
            garp_count => {type => SCALAR, default => 3},
            garp_interval => {type => SCALAR, default => 2},
            # common: ping config
            ping => {type => SCALAR, default => undef()},
            ping_count => {type => SCALAR, default => 5},
            ping_interval => {type => SCALAR, default => 0.5},
   #
            # common: vif configuration
            configure_vif => {type => HASHREF, default => undef()},
   #
            vif => {type => SCALAR, default => undef(), depends => ['vserver']},
            vserver => {type => SCALAR, default => undef(), depends => ['vif']},
            address => {type => SCALAR, default => undef()},
            migrate_type => {type => SCALAR, default => undef()},
            migrate_node => {type => ARRAYREF, default => undef()},
        }
    );

    ##########################################################################
    from ping_check
    Net::Ping version of ping check:

    my $time_unreach = 0;
    my $pings = 0;
 
    for (my $i = 0; $i < $self->{_ping_check_count}; $i++) {
        my ($rv, $time, $host) = $self->{_ping}->ping($self->{_vif_address});

        # There's probably a bug here relating to time_unreach
        if ($rv) {
            $pings ++;
        } else {
            $time_unreach += $time;
        }
        $Log->debug(  "Ping from $host in $time\s, result is $rv");
    }

    if ($pings < $self->{_ping_check_count_req}) {
        $Log->error(  "Recieved $pings while ".
            "$self->{_ping_check_count_req} pings should have been received.");
    }

    if ($time_unreach > $self->{_ping_unreach_allow}) {
        $Log->error(  "Interface was unreachable for too long following ".
            "migration.  It was unreachable for $time_unreach");
    }

    ###########################################################################
