#
# $Id: //depot/prod/test/main/lib/VifMgr/Failover/Migrate.pm#3 $
# $Revision: #3 $ submitted on $Date: 2009/02/19 $ By $Author: kratsas $
#
## @summary     Network::VIFMgr Migrate Class
## @author      ewing@netapp.com
## @pod here

=head1 NAME

Network::VIFMgr::Failover::Migrate

=head1 SYNOPSIS

use Network::VIFMgr::Failover::Migrate;

my $vfm = new Vif(address => '1.2.3.4');

$vfm->failover_event();

=head1 DESCRIPTION

This is a class for VIF Migrate testing.  It combines several features shared
with Network::VIFMgr::Failover::PortFail through the parent class: Network::VIFMgr::Failover.
These features include ping validation, gratuitous ARP checking and pre/post
failover state checks.

=head1 METHODS

=cut

package Network::VIFMgr::Failover::Migrate;

use strict;
use warnings;

use Options;
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR ARRAYREF OBJECT BOOLEAN UNDEF);
use Carp qw(confess);
use Tharn;

use base 'Network::VIFMgr::Failover';

############NACL#############
use NACL::C::Node;
use NACL::CS::NetworkInterface;
use NACL::STask::NetworkInterface;
use NACL::CS::NetworkPort;

use NATE::Log qw(log_global);
use NATE::Process;
my $Log = log_global();

#$Log->set_options( separateNewline => 1 );

=head2 new

my $vfm = new Network::VIFMgr::Failover::Migrate(vif => 'vif0', vserver => 'vs0');

C<new> is the constructor for Network::VIFMgr::Failover::Migrate objects.  This class
provides a simple interface for testing VIF Migrate failover operations.  
For VIF Migrate testing, a vif and vserver or address is required.  This vif's
configuration can be supplied in the constructor call if the interface is to 
be created.  Aside from interface selection, you can also select the kind of
migrations that will be performed on the interface.

These are the options that can be supplied:

=over

Interface Selection:

vif => <vif_name>,
vserver => <vserver_name>

=back

address => <ip_addres>

Migration Selection:

migrate_type => (local|remote|any)

configure_vif => {vif => <vif_name>...netmask => '255.255.255.0'}
=back

There are additional options that may be specified, but are documented as 
part of VifMgr::Failover's documentation.

=cut

#############################################################################
# Subroutine: new
# Parameters: Any defaults the user wants to set
# Returns: Object Reference
# Purpose: Constructor
#############################################################################

sub new {

	$Log->enter();

	# Call the parent class constructor

	my $class = shift(@_);
	my $self  = $class->SUPER::new(@_);

	my %p = (@_);

	# Define the defaults

	$self->{_vif_name}     = $p{vif};
	$self->{_vif_vserver}  = $p{vserver};
	$self->{_vif_address}  = $p{address};
	$self->{_migrate_type} = $p{migrate_type} || $self->RANDOM;
	$self->{_migrate_node} = $p{migrate_node};
	$self->{_last_node}    = undef();
	$self->{_last_port}    = undef();

	my $node = $self->{_options}->get( option => 'FILER' );

	# Configure the vif is requested.

	if ( defined( $p{configure_vif} ) ) {

		NACL::STask::NetworkInterface->create(    command_interface => $self->{_node},
											%{ $p{configure_vif} } );

		#$self->{_node}->network_interface_new( %{ $p{configure_vif} } );

		$self->{_vif_name}    = $p{configure_vif}->{vif};
		$self->{_vif_vserver} = $p{configure_vif}->{server};
		$self->{_vif_address} = $p{configure_vif}->{address};
		$self->{_vif_created} = 1;
	}
	else {
		$self->{_vif_created} = 0;
	}

	# Get the vif config for the interface we want to migrate.

	if ( defined( $self->{_vif_name} ) && defined( $self->{_vif_vserver} ) ) {
		my $vif_info = NACL::CS::NetworkInterface->fetch(
			command_interface => $self->{_node},
			filter            => {
				lif     => $self->{_vif_name},
				vserver => $self->{_vif_vserver},

			}
		);
		$self->{_vif_address} = $vif_info->{address};
		$self->{_vif_netmask} = $vif_info->{netmask};

		$self->{_vif_role} = $vif_info->{role};

		##special attention: should use '_' instead of '-'!!
		$self->{_last_node} = $vif_info->{'curr_node'};

		$self->{_last_port} = $vif_info->{'curr_port'};

	}
	elsif ( defined( $self->{_vif_address} ) ) {

		my $vif_info = NACL::CS::NetworkInterface->fetch(
								  command_interface => $self->{_node},
								  filter => { address => $self->{_vif_address} } );
		$self->{_vif_name}    = $vif_info->{vif};
		$self->{_vif_vserver} = $vif_info->{server};
		$self->{_vif_netmask} = $vif_info->{netmask};
		$self->{_last_node}   = $vif_info->{'curr_node'};
		$self->{_vif_role}    = $vif_info->{role};
		$self->{_last_port}   = $vif_info->{'curr_port'};

	}
	else {
		$Log->error( "vif/vserver or address is required.", );
	}

	# Verify that that there are some filers to use.

	if ( !scalar( @{$node} ) ) {
		$Log->error( "There are no nodes available for this test.", );
	}

	# Collect port information and store the available ports for this test.

	my @port = NACL::CS::NetworkPort->fetch(
											command_interface => $self->{_node},
											filter            => {
												   role => $self->{_vif_role},
												   link => 'up',
												   node => join( "|", @{$node} )
											}
	);

	$self->{_ports} = [ map( {"$_->{node}\:$_->{port}"} @port ) ];
	$Log->comment(
		   "Migrate using these ports: " . join( ", ", @{ $self->{_ports} } ) );

	bless( $self, $class );

	$self->add_vif_watch(    vif     => $self->{_vif_name},
						  vserver => $self->{_vif_vserver} );

	$Log->comment("Created a new $class object: $self");
	return $self;
} ## end sub new

#############################################################################
# Subroutine: select_port
# Parameters: none
# Returns: Node/Port
# Purpose: Selcts ports the the next migration based upon object configuration.
#############################################################################

sub select_port {

	$Log->enter();
	my $self = shift();

	my $node;
	my $port;

	# Make a temporary copy of our ports list.

	my @ports = @{ $self->{_ports} };
	$Log->comment( "Ports: " . join( ", ", @ports ) );

	# Loop until we find what we're looking for

	while (1) {

		if ( !scalar(@ports) ) {
			$Log->error( "There aren't any ports available for migration", );
		}

		# Select a random port from the list & break the node and port name out.

		my $index = int( rand( scalar(@ports) ) );
		( $node, $port ) = split( /:/, $ports[$index] );

		# Print the port selected, its index # out of array size, followed by
		# the current node/port. You'll see the array shrink as loop iterates.

		$Log->comment( "Selected $port on $node ($index/$#ports).  "
				 . "Previous port: $self->{_last_node}\:$self->{_last_port}." );

		$Log->comment(
"##################Migrate_type: $self->{_migrate_type}, LastNode: $self->{_last_node}"
		);

		# If we've got undefined values, something bad happened.

		if (    !defined( $ports[$index] )
			 || !defined($node)
			 || !defined($port) )
		{
			$Log->debug(   "\$ports[\$index]=$ports[$index], \$node=$node, "
						 . "\$port=$port" );
			confess();
		}

		# Check node/port names to determine if we have a suitable port.
		# when an unsuitable port is selected, delete it from the list and try
		# again.

		# First check is if the selected node/port is the current node/port

		if (    ( defined($node) && ( $self->{_last_node} eq $node ) )
			 && ( defined($port) && ( $self->{_last_port} eq $port ) ) )
		{

			$Log->comment(
					 "Deleted index $index: $node\:$port from list: same port");
			splice( @ports, $index, 1 );
			next;

		# If we picked a remote node, but the migrate type is local, we can also
		# remove that.

		}
		elsif (    ( $self->{_migrate_type} eq $self->LOCAL )
				&& ( $self->{_last_node} ne $node ) )
		{
			$Log->comment(
				"Deleted index $index: $node\:$port from list: not a local port" );
			splice( @ports, $index, 1 );
			next;

		# If we picked a local node, but the migrate type is remote, we can also
		# remove that.

		}
		elsif (    ( $self->{_migrate_type} eq $self->REMOTE )
				&& ( $self->{_last_node} eq $node ) )
		{
			$Log->comment(
"Deleted index $index: $node\:$port from list: not a remote port" );
			splice( @ports, $index, 1 );
			next;

		}
		else {

			# Make sure that the link is up and we've selected a data port

			my $port_info;
			eval {
				$port_info = NACL::CS::NetworkPort->fetch(
											command_interface => $self->{_node},
											filter            => {
														node => $node,
														port => $port
											}
				);
			};
			if ($@) {

				# TODO: handle offline nodes correctly here.
				$Log->error($@);
			}

			if (    ( $port_info->{link} ne 'up' )
				 && ( $port_info->{role} ne 'data' ) )
			{
				$Log->debug(   "Deleted index $index: $node\:$port from "
							 . "list port isn't up." );
				delete( $ports[$index] );
				next;
			}

			last;
		} ## end else [ if ((defined($node) &&...
	} ## end while (1)

	# Return the port matching all criteria.

	$Log->comment("Selected port $node\:$port");
	return ( $node, $port );
} ## end sub select_port

=head2 failover_event

$vfm->failover_event

C<failover_event> is the main testing method for VifMgr::Failover::Migrate 
objects.  The C<failover_event> method will select a port, verify that the
interface is up, listen for gratuitous ARPs, migrate the interface and finally
verify that the interface is once again pingable.

There are no options for this method (yet).

=cut

sub failover_event {

	my $self = shift();
	$Log->enter();

	# Store the current state of our interface and select a new port

	$self->store_vif_state();
	my ( $node, $port ) = $self->select_port();

	# Verify the interface is up before migration
	# TODO: use the watch list

#	$Log->warn("#############Not conducting pre/post migrate ping check.###########");
	$Log->comment("Starting ping check pre migrate");
	$self->ping_check( id => "pre-migrate_check" );
	$Log->comment("Pre-migrate ping check succeeded.");

	# Start a gratuitous arp check in the background before migration
	# TODO: use the watch list
	$Log->comment("Starting garp_check:");
	my $garp_checker = $self->garp_check(id=>"garp_checker",ip=>$self->{_vif_address},mask=>$self->{_vif_netmask},);

	$Log->comment("##garp_checker is : $garp_checker");
	#my $garp_checker = Tharn::subtest( sub { $self->garp_check(); },
	#								   -bg, -onexit, sub { }, -runid,
	#								   "garp_checker" );

	# Migrate

	$Log->comment("Attempting migrate to node: $node, port: $port.");

	NACL::STask::NetworkInterface->migrate(
											command_interface => $self->{_node},
											lif     => $self->{_vif_name},
											vserver => $self->{_vif_vserver},
											'dest-node' => $node,
											'dest-port' => $port
	);

	$Log->comment("Sleep to allow LIF move");
	sleep 3;

	# Check reachability following the migrate.

	$Log->comment("Starting ping check post migrate");
	$self->ping_check( id => "post-migrate_check" );
	$Log->comment("Post-migrate ping check succeeded.");

	# Let the gratuitous ARP check finish & check its return value.

	#$garp_checker->wait_finish();

	if ( $garp_checker == 0 ) {
		$Log->comment("Found the desired number of gratuitous ARPs.");
	}
	elsif ( $garp_checker> 0 ) {
		$Log->warn("Failed to detect all required gratuitous ARPs.  Found "
				 . ( $self->{_garp_check_count} - $garp_checker )
				 . ", should have got "
				 . $self->{_garp_check_count}
				 . "." );
	}

	# Check the vif state and make sure there are some differences.

	my $difference = $self->check_vif_state();

	if ( !scalar( keys(%$difference) ) ) {
		$Log->error("Migration didn't cause any change in VIF state.");
	}
	else {
		while ( my ( $k, $v ) = each( %{$difference} ) ) {
			$Log->debug("\t$k -> $v");
		}
	}

	# Record the new port for the next migration.

	$self->{_last_node} = $node;
	$self->{_last_port} = $port;

	$Log->exit();
} ## end sub failover_event

# DESTROY
#
# Destructor for this class
##############################################################################

sub DESTROY {

	$Log->enter();
	my $self = shift();

	# Remove the vif under test if we created one.

	if ( $self->{_vif_created} ) {
		NACL::STask::NetworkInterface->purge(
											command_interface => $self->{_node},
											lif     => $self->{_vif_name},
											vserver => $self->{_vif_vserver}
		);
	}

	$Log->exit();
} ## end sub DESTROY

=head1 SEE ALSO

L<Network::VIFMgr::Failover>
L<Network::VIFMgr::Failover::PortFail>
L<Options>
L<TracePkg>
L<TestUtils>

=head1 BUGS

None known, but many suspected.

=head1 STATUS

  In-progress

=head1 AUTHOR

  Steven Ewing (ewing@netapp.com)
  $Id: //depot/prod/test/main/lib/Network::VIFMgr/Failover/Migrate.pm#3 $
  Last revised $Date: 2009/02/19 $

=cut

1;
