# this is basically -*- perl -*-
#
# $Id: //depot/prod/test/main/lib/Network::VIFMgr/VifConfig/VifDns.pm#13 $
#
# Copyright (c) 2008 Network Appliance, Inc., All Rights Reserved
# Permission is granted to use this script on a Network Appliance Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from Network Appliance, Inc
#
## @summary     Class to handle the DNS Zone configuration/verification of a VIF
## @author      dl-cluster-qa
## @pod         here

=head1 NAME

VifDns.pm

=head1 AUTHOR

dl-cluster-qa

=head1 DESCRIPTION

C<VifDns> will handle all any DNS related configuration for a Virtual Interface. 
This inclues, but is not limited to:
    - Creating a VIF with a DNS Zone
    - Adding a DNS zone to an existing VIF
    - Modifying a DNS Zone
    - Disabling a DNS Zone
    - Deleting a DNS Zone
    - Verifying IPs are registered as part of their configured zones or not

=head1 SYNOPSIS

    use Network::VIFMgr::VifConfig::VifDns;
    my $Vif_Cfg = new Network::VIFMgr::VifConfig::VifDns();
    my $CURR_VIF = "this_vif";
    my $CURR_VS = "vs0";
    my $deleted = $Vif_Cfg->delete_vif_dns(vif => $CURR_VIF, server => $CURR_VS);
    if (!$Vif_Cfg->modify_vif_dns(vif => $CURR_VIF, zone => $zone)) {
        $Test->nlog("- DNS could not be enabled on the following VIFs ".
            " $CURR_VIF to be part of $zone");
        return $TCD::FAIL;
    }

=head1 ARGUMENTS

=over 4

=item ZONE
(Optional)

Usage: [ZONE="ZoneTest.com"]

Description: Default DNS Zone name to use

=back

=head1 PUBLIC METHODS

=over 4

=item create_vif_dns

Parameters: Attributes to create the VIF (defaults to object settings).
Returns: Number of VIFs created. Purpose: To create a VIF on the filer 
with this specified configuration, and then verify that VIF was in fact 
created with the expected parameters. The route for the VIF will be added 
as well. If the user would like, they can also verify that the VIF's IP 
address is properly resolved as part of the DNS Zone

=item modify_vif_dns

Parameters: DNS Zone or Allow-lb-migrate and VIF name to modify (optional
server). Returns: Number of VIFs modified, 0 on Failure. Purpose: To modify 
(or disable by setting to 'none') the dns zone of a VIF, or array of VIFs.

=item delete_vif_dns

Parameters: DNS Zone and VIF name of VIF to be deleted (optional server).
Returns: Number of VIFs whose DNS zones were deleted, 0 on failure.
Purpose: To delete the DNS zone attribute of a VIF (end result will be
changing its DNS zone to 'none'.

=item verify_lb_value

Parameters: List of VIFs to verify that -allow-lb-migrate/lb-weight has been 
flagged properly, attribute to verify (either allow_lb_migr or lb_weight), 
and expected value. Returns: 1 on success, 0 on failure. Purpose: To verify 
that the -allow-lb-migrate/lb-weight attribute can be successfully toggled for 
VIFs

=back

=cut

package Network::VIFMgr::VifConfig::VifDns;

# Standard Libs

use strict;
use warnings;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);

# Netapp Specific Libs

use NATE::Log qw(log_global);
use Options;
use Server;
use Network::DnsResolver;
use Tharn;

#NACL

use NACL::C::Node;

# Globals

my $Options = new Options();
my $Server  = new Server();
my $Log = log_global();
my @Nodes        = NACL::C::Node->find();
my $Dns_Resolver = new Network::DnsResolver();

# Parent Class

use base qw(Network::VIFMgr::VifConfig);

############################################################################
# Subroutine: new
# Parameters: None
# Returns: Reference to this object
# Purpose: To instantiate and configure the VifDns object
############################################################################

sub new {

	$Log->enter();

	# Call the parent method and then extend from it below.

	my $class = shift(@_);
	my $self  = $class->SUPER::new(@_);

	# Get the command line options. These will become the defaults for
	# VIF configuration. However, the calling method is still capable of
	# passing in values to override these to each of the relative subroutines.

	$Options->parameters(
						  definitions => [
							   {
								 name        => "ZONE",
								 description => "Default DNS Zone name to use",
								 default     => "ZoneTest.com",
							   },
						  ]
	);
	$Options->process();
	$Log->debug( "Options = " . Dumper($Options) );

# Configure the DNS object (copying over options locally)
# Note that the default value for lb-weight on an OnTap system is "load".
# Setting it to a specific weight here, so that we can get an even distribution,
# which is what most tests will need.

	my $opts = $Options->get_hash();
	$self->{_zone}          = $opts->{ZONE};
	$self->{_allow_lb_migr} = "false";
	$self->{_lb_weight}     = "100";

	bless( $self, $class );

	$Log->debug(   "Initialized Network::VIFMgr::VifConfig::VifDns object with "
				 . "the following configuration:\n"
				 . Dumper($self) );

	$Log->comment("Returning blessed reference: $self");
	return $self;

} ## end sub new

###########################################################################
# Subroutine: create_vif_dns
# Parameters: Attributes to create the VIF (defaults to object settings)
# Returns: Number of VIFs created
# Purpose: To create a VIF on the filer with this specified configuration,
#          and then verify that VIF was in fact created with the expected
#          parameters. The route for the VIF will be added as well. If the
#          user would like, they can also verify that the VIF's IP address
#          is properly resolved as part of the DNS Zone
###########################################################################

sub create_vif_dns {

	$Log->enter();
	my $self = shift(@_);

	# Initialize parameters with defaults set during object creation
	# Can be overriden if parameter specified by calling routine.
	# Note that the number of vifs created will be based on the num_vifs
	# and not the size of the address array. Index denotes the index to
	# start pulling IP addresses out of $params{address}. This is useful
	# if the calling procedure is adding VIFs at a later point, and does not
	# want to start at the beggining of the array each time. The user MUST
	# specify enough IPs on the ntest command line to account for additional
	# VIFs created later on (via DATA_INTERFACES)

	my %params = validate(
						   @_,
						   {
							  server => {
										  type     => SCALAR,
										  optional => 1,
										  default  => $self->{_server}
							  },
							  vif => {
									   type     => SCALAR,
									   optional => 1,
									   default  => $self->{_vif_name}
							  },
							  vif_name_incr => {
											  type     => SCALAR,
											  optional => 1,
											  default => $self->{_vif_name_incr}
							  },
							  role => {
										type     => SCALAR,
										optional => 1,
										default  => "data"
							  },
							  status_admin => {
												type     => SCALAR,
												optional => 1,
												default  => "up"
							  },
							  address => {
										   type     => ARRAYREF,
										   optional => 1,
										   default  => $self->{_data_ifs}
							  },
							  if_netmask => {
											  type     => SCALAR,
											  optional => 1,
											  default  => $self->{_data_if_mask}
							  },
							  home_node => {
											 type     => SCALAR,
											 optional => 1,
											 default  => undef,
							  },
							  home_port => {
											 type     => SCALAR,
											 optional => 1,
											 default  => undef,
							  },
							  use_failover_group => {
													  type     => SCALAR,
													  optional => 1,
													  default  => undef,
							  },
							  auto_revert => {
											   type     => SCALAR,
											   optional => 1,
											   default  => "false",
							  },
							  allow_lb_migr => {
											  type     => SCALAR,
											  optional => 1,
											  default => $self->{_allow_lb_migr}
							  },
							  lb_weight => {
											 type     => SCALAR,
											 optional => 1,
											 default  => $self->{_lb_weight}
							  },
							  num_vifs => {
											type     => SCALAR,
											optional => 1,
											default  => 1
							  },
							  destination => {
											   type     => SCALAR,
											   optional => 1,
											   default  => "0.0.0.0"
							  },
							  route_mask => {
											  type     => SCALAR,
											  optional => 1,
											  default  => "0.0.0.0"
							  },
							  group_mask => {
											  type     => SCALAR,
											  optional => 1,
											  default  => "0"
							  },
							  gateway => {
										   type     => SCALAR,
										   optional => 1,
										   default  => $self->{_gateway}
							  },
							  metric => {
										  type     => SCALAR,
										  optional => 1,
										  default  => $self->{_metric}
							  },
							  zone => {
										type     => SCALAR,
										optional => 1,
										default  => $self->{_zone}
							  },
							  index => {
										 type     => SCALAR,
										 optional => 1,
										 default  => 0
							  },
							  rtn_name => {
											type     => SCALAR,
											optional => 1,
											default  => 0
							  },
							  verify_ip => {
											 type     => SCALAR,
											 optional => 1,
											 default  => 1,
							  },
							  create_route => {
												type     => SCALAR,
												optional => 1,
												default  => 1,
							  },
							  server_args => {
											   type     => ARRAYREF,
											   optional => 1,
											   default  => undef,
							  }
						   }
	);

	# Verify server exists

	if ( !Network::VIFMgr::VifConfig::_server_exists( server => $params{server} ) ) {
		$Log->warn("Server: $params{server} does not exist. Returning...");
		return 0;
	}

	my $end = $params{index} + $params{num_vifs};
	my @vif_names;

	# params{address} should have enough addresses to create num_vifs.

	if ( scalar( @{ $params{address} } ) < $end ) {
		$Log->warn(   "Not enough addresses neccessary to create VIFs. "
					. "Needed: $end, Found: "
					. scalar( @{ $params{address} } )
					. ". Will only "
					. "create up to "
					. scalar( @{ $params{address} } ) );
		$end = scalar( @{ $params{address} } );
	}

	# Check to make sure we are creating something

	if ( $end < 1 ) {
		$Log->fatal("Not creating any LIFs. Exiting.");
		return 0;
	}

	for ( my $i = $params{index} ; $i < $end ; $i++ ) {

		# Only get the node and/or port if the calling method did not specify
		# it.

		my $node;
		my $port;

		if ( !defined( $params{home_node} ) ) {
			$node = Network::VIFMgr::VifConfig::_get_home_node($self);
		}
		else {
			$node = $params{home_node};
		}

		if ( !defined( $params{home_port} ) ) {
			$port =
			  Network::VIFMgr::VifConfig::_get_home_port( $self, node => $node );
		}
		else {
			$port = $params{home_port};
		}

		# If the user passed in 'use-failover-groups' then we will specify
		# that as well. Otherwise, let the value default.

		my $use_failover_group     = undef;
		my $use_failover_group_val = undef;

		if ( defined( $params{use_failover_group} ) ) {
			$use_failover_group     = 'use-failover-group';
			$use_failover_group_val = $params{use_failover_group};
		}

		# Verify that the IP does not currently exist

		if ( $self->{_delete_ip} ) {
			my @lif;
			try {
				@lif = NACL::CS::NetworkInterface->fetch(
								command_interface => $Nodes[0],
								filter => { address => $params{address}->[$i], } );
			  }

			  catch NACL::Exceptions::NoElementsFound with {
				$Log->comment(
					   "IP $params{address}->[$i] does not " . "yet exist..." );
			  }
			  otherwise {
				$Log->warn("Unexpected error: $@. Returning...");
				return 0;
			  };

		}

		# Perform create

		#$Server->network_interface_new(
		NACL::STask::NetworkInterface->create(
								 command_interface=>$Nodes[0],
								 vserver => $params{server},
								 lif  => $params{vif} . $params{vif_name_incr},
								 role => $params{role},
								 'status-admin'      => $params{status_admin},
								 'home-node'         => $node,
								 'home-port'         => $port,
								 address             => $params{address}->[$i],
								 netmask             => $params{if_netmask},
								 'auto-revert'       => $params{auto_revert},
								 'dns-zone'          => $params{zone},
								 'allow-lb-migrate'  => $params{allow_lb_migr},
								 'lb-weight'         => $params{lb_weight},
								 $use_failover_group => $use_failover_group_val,
								 'nacltask_if_exists' => 'purge',
		);

		# Add the route if specified

		if ( $params{create_route} ) {
			Network::VIFMgr::VifConfig::create_route(
				$self,
				server      => $params{server},
				vif         => $params{vif} . $params{vif_name_incr},
				gateway     => $params{gateway},
				destination => $params{destination},
				route_mask  => $params{route_mask},
				group_mask  => $params{group_mask},
				gateway     => $params{gateway},
				metric      => $params{metric},

				#		   server_args => $params{server_args},
			);
		} ## end if ($params{create_route...

		# Save the VIF name for future reference

		push( @vif_names, "$params{vif}$params{vif_name_incr}" );

		# If we are flagged to, we will verify this IP is allocated for this
		# particular zone if the VIF is up and has a valid domain name

		if ( $params{verify_ip} ) {

			if (    ( $params{zone} !~ /none/ )
				 && ( $params{status_admin} =~ /up/ ) )
			{

				# Since we have an UP VIF and a valid dns zone, we can verify
				# the IP is handed out

				if (
					 !$Dns_Resolver->verify_ip(                ips => $params{address}->[$i],
												domain => $params{zone} )
				  )
				{
					$Log->warn(   "DNS Zone $params{zone} did not resolve "
								. $params{address}->[$i]
								. " when expected to." );
					return 0;
				} ## end if (!$Dns_Resolver->verify_ip...

				$Log->comment( "Zone: $params{zone} successfully resolved "
							   . $params{address}->[$i] );

			}
			elsif ( $params{zone} !~ /none/ ) {

		   # Since either the VIF is DOWN we can verify the IP is not handed out

				if (
					 !$Dns_Resolver->verify_ip(
												ips => $params{address}->[$i],
												domain => $params{zone},
												state  => 0
					 )
				  )
				{
					$Log->warn(   "DNS Zone $params{zone} resolved IP address "
								. $params{address}->[$i]
								. " when expected not to." );
					return 0;
				} ## end if (!$Dns_Resolver->verify_ip...

				$Log->comment(   "Zone: $params{zone} did not resolve "
							   . $params{address}->[$i]
							   . " as expected" );

			}
			else {
				$Log->comment( "IP $params{address}->[$i] does not belong "
					. "to a dns-zone (dns-zone = $params{zone}, nothing to verify.)"
				);
			}
		} ## end if ($params{verify_ip})

		# Only increment this if its a numerical value (i.e. was not
		# initialized to "")

		if ( $params{vif_name_incr} ne "" ) {
			$params{vif_name_incr}++;
		}

	} ## end for (my $i = $params{index...

	# Copy the $params{vif_name_incr} back over to the object so it can
	# be re-initialized for the next use (so we are not using the same values
	# over and over)

	$self->{_vif_name_incr} = $params{vif_name_incr};

	$Log->exit("Successfully created $params{num_vifs} VIFs");

	# There may be a case when we want this method to return the VIF name
	# as opposed to the number of VIFs created. That will be flagged here
	# Also note that we are returning the number of VIFs the user requested
	# to create as opposed to the number of VIFs that were actually created.
	# We don't want the upper level tests to fail just because we were not
	# allocated enough IP addresses. We are still warning with the number
	# of VIFs that were really created and that should be sufficient.

	if ( $params{rtn_name} ) {
		return \@vif_names;
	}
	else {
		return $params{num_vifs};
	}

} ## end sub create_vif_dns

###############################################################################
# Subroutine: modify_vif_dns
# Parameters: DNS Zone or Allow-lb-migrate and VIF name to modify (optional
#             server)
# Returns: Number of VIFs modified, 0 on Failure
# Purpose: To modify any parameter of a VIF and verify that it is still capable
#          of serving DNS traffic
###############################################################################

sub modify_vif_dns {

	$Log->enter();
	my $self = shift(@_);

	# Initialize the parameters to perform the VIF modification

	my %params = validate(
						   @_,
						   {
							  server => {
										  type     => SCALAR,
										  optional => 1,
										  default  => $self->{_server}
							  },
							  vif => {
									   type     => SCALAR | ARRAYREF,
									   optional => 1,
									   default  => $self->{_vif_name}
							  },
							  zone => {
										type     => SCALAR,
										optional => 1,
										default  => undef
							  },
							  allow_lb_migr => {
												 type     => SCALAR,
												 optional => 1,
												 default  => undef
							  },
							  lb_weight => {
											 type     => SCALAR,
											 optional => 1,
											 default  => undef
							  },
							  home_node => {
											 type     => SCALAR,
											 optional => 1,
											 default  => undef,
							  },
							  home_port => {
											 type     => SCALAR,
											 optional => 1,
											 default  => undef,
							  },
							  address => {
										   type     => SCALAR,
										   optional => 1,
										   default  => undef,
							  },
							  netmask => {
										   type     => SCALAR,
										   optional => 1,
										   default  => undef,
							  },
							  failover_policy => {
												   type     => SCALAR,
												   optional => 1,
												   default  => undef,
							  },
							  auto_revert => {
											   type     => SCALAR,
											   optional => 1,
											   default  => undef,
							  },
							  use_failover_group => {
													  type     => SCALAR,
													  optional => 1,
													  default  => undef,
							  },
							  firewall_policy => {
												   type     => SCALAR,
												   optional => 1,
												   default  => undef,
							  },
							  rename => {
										  type     => SCALAR,
										  optional => 1,
										  default  => undef,
							  },
							  verify_ip => {
											 type     => SCALAR,
											 optional => 1,
											 default  => 1,
							  },
						   }
	);

	# Verify server exists

	if ( !Network::VIFMgr::VifConfig::_server_exists( server => $params{server} ) ) {
		$Log->warn("Server: $params{server} does not exist. Returning...");
		return 0;
	}

	# Verify calling proc specificed correct params.

	if (    ( !defined( $params{zone} ) )
		 && ( !defined( $params{lb_weight} ) )
		 && ( !defined( $params{allow_lb_migr} ) )
		 && ( !defined( $params{home_node} ) )
		 && ( !defined( $params{home_port} ) )
		 && ( !defined( $params{address} ) )
		 && ( !defined( $params{netmask} ) )
		 && ( !defined( $params{failover_policy} ) )
		 && ( !defined( $params{firewall_policy} ) )
		 && ( !defined( $params{auto_revert} ) )
		 && ( !defined( $params{use_failover_group} ) )
		 && ( !defined( $params{rename} ) ) )
	{
		$Log->warn(   "Must specify at least one the interface modify "
					. "parameters. Exiting..." );
		return 0;
	} ## end if ((!defined($params{...

	my $vif_count = 0;
	my %addr_old_zone;
	my @show;

  # If the calling method passed in a wildcard character (or single vif name)
  # instead of a list of vifs, we need to extract out the original IP/zone combo
  # before the bulk modification occurrs. This is for verification purposes
  # later on.

	if ( ref( $params{vif} ) ne 'ARRAY' ) {

		@show = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $params{vif},
												 }
		);

		#$Server->network_interface_show(       server => $params{server},
		#										 vif    => $params{vif}, );

		foreach my $vif_ref (@show) {

			# Get the IP addreses and dns-zone

			my %vif = %{$vif_ref};
			$addr_old_zone{ $vif{address} } = $vif{'dns-zone'};
			$vif_count++;
		}

	 # Since calling method can specify an array of VIF names to modify, cast to
	 # array if not already.

		$params{vif} = [ $params{vif} ];

	}
	else {

		# If we already have a list of vif names, we can do a show on each
		# one and parse out the ip and zone from there

		foreach my $vif ( @{ $params{vif} } ) {
			@show = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $params{vif},
												 }
			);

		   #$Server->network_interface_show(          server => $params{server},
		   #										 vif    => $vif, );

			my %vif = %{ $show[0] };
			$addr_old_zone{ $vif{address} } = $vif{zone};
			$vif_count++;
		} ## end foreach my $vif (@{$params{...
	} ## end else [ if (ref($params{vif}) ...

	# Perform the modify for each VIF specified

	foreach my $vif ( @{ $params{vif} } ) {

		if ( defined $params{zone} ) {
			my $result = NACL::STask::NetworkInterface->modify(
				command_interface => $Nodes[0],

				# $Server->network_interface_modify(
				vserver    => $params{server},
				lif        => $vif,
				'dns-zone' => $params{zone}
			);
		}

		if ( defined $params{allow_lb_migr} ) {
			my $result = NACL::STask::NetworkInterface->modify(
				command_interface => $Nodes[0],

				#$Server->network_interface_modify(
				verver             => $params{server},
				lif                => $vif,
				'allow-lb-migrate' => $params{allow_lb_migr}
			);
		}

		if ( defined $params{lb_weight} ) {
			my $result = NACL::STask::NetworkInterface->modify(
				command_interface => $Nodes[0],

				#$Server->network_interface_modify(
				vserver     => $params{server},
				lif         => $vif,
				'lb-weight' => $params{lb_weight}
			);
		}

		if ( defined $params{home_node} ) {

			# If the calling method has specified to find the home-node,
			# we will do that here.

			if ( $params{home_node} eq "find" ) {

				my $num_tries = 0;

				# Get the current value for home-node

				my @output = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $vif,
												 }
				);
				my $curr_node = $output[0]->{'home-node'};

				# Loop through the node list until we find a value
				# that is not our current home-node

				do {
					$params{home_node} =
					  Network::VIFMgr::VifConfig::_get_home_node($self);
					$num_tries++;
				  } while (    ( $params{home_node} eq $curr_node )
							&& ( $num_tries <= 10 ) );

				# Make sure we found a value

				if (    ( $num_tries == 10 )
					 && ( $params{home_node} eq $curr_node ) )
				{
					$Log->warn(   "Unable to find a value to modify the "
								. "home-node to. Exiting..." );
					return 0;
				}
			} ## end if ($params{home_node}...

			# Perform the modify

			my $result = NACL::STask::NetworkInterface->modify(
											   command_interface => $Nodes[0],
											   vserver     => $params{server},
											   lif         => $vif,
											   'home-node' => $params{home_node}
			);
		} ## end if (defined $params{home_node...

		if ( defined $params{home_port} ) {

			# If the calling method has specified to find the home-port,
			# we will do that here.

			if ( $params{home_port} eq "find" ) {

				my $num_tries = 0;

				# Get the current value for home-port

				my @output = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $vif,
												 }
				);
				my $curr_port = $output[0]->{'home-port'};
				my $curr_node = $output[0]->{'home-node'};

				# Loop through the node list until we find a value
				# that is not our current home-node

				do {
					$params{home_port} =
					  Network::VIFMgr::VifConfig::_get_home_port( $self,
														   node => $curr_node );
					$num_tries++;
				  } while (    ( $params{home_port} eq $curr_port )
							&& ( $num_tries <= 10 ) );

				# Make sure we found a value

				if (    ( $num_tries == 10 )
					 && ( $params{home_port} eq $curr_port ) )
				{
					$Log->warn(   "Unable to find a value to modify the "
								. "home-port to. Exiting..." );
					return 0;
				}
			} ## end if ($params{home_port}...

			# Perform the modify

			my $result = NACL::STask::NetworkInterface->modify(
											   command_interface => $Nodes[0],
											   vserver     => $params{server},
											   lif         => $vif,
											   'home-port' => $params{home_port}
			);
		} ## end if (defined $params{home_port...

		if ( defined $params{address} ) {

			# Are we just using the next IP available in DATA_INTERFACES?

			if ( $params{address} eq "next" ) {
				$params{address} =
				  $self->{_data_ifs}->[ $self->{_vif_name_incr}++ ];
			}

			# Now we need to get the old address for this VIF and overwrite
			# it in the addr_old_zone array so that when we check for the DNS
			# zone, we are checking for the new IP. The old one is no longer
			# returned.

			my @show = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $vif,
												 }
			);
			my $old_addr = $show[0]{address};

			# Perform the modify

			my $result = NACL::STask::NetworkInterface->modify(
												 command_interface => $Nodes[0],
												 vserver => $params{server},
												 lif     => $vif,
												 address => $params{address}
			);

			# Now that the modify occurred, we can update the address/zone
			# hash.

			delete $addr_old_zone{$old_addr};
			$addr_old_zone{ $params{address} } = $show[0]{'dns-zone'};

		} ## end if (defined $params{address...

		if ( defined $params{netmask} ) {
			my $result = NACL::STask::NetworkInterface->modify(
												 command_interface => $Nodes[0],
												 vserver => $params{server},
												 lif     => $vif,
												 netmask => $params{netmask}
			);
		}

		if ( defined $params{failover_policy} ) {
			my $result = $Server->network_interface_modify(
								   server            => $params{server},
								   vif               => $vif,
								   'failover-policy' => $params{failover_policy}
			);
		}

		if ( defined $params{auto_revert} ) {
			my $result = $Server->network_interface_modify(
										   server        => $params{server},
										   vif           => $vif,
										   'auto-revert' => $params{auto_revert}
			);
		}

		if ( defined $params{use_failover_group} ) {
			my $result = NACL::STask::NetworkInterface->modify(
							 command_interface    => $Nodes[0],
							 vserver              => $params{server},
							 lif                  => $vif,
							 'use-failover-group' => $params{use_failover_group}
			);
		}

		if ( defined $params{firewall_policy} ) {
			my $result = NACL::STask::NetworkInterface->modify(
								   command_interface => $Nodes[0],
								   vserver           => $params{server},
								   lif               => $vif,
								   'firewall-policy' => $params{firewall_policy}
			);
		}

		if ( defined $params{rename} ) {
			my $result = NACL::STask::NetworkInterface->modify(
												 command_interface => $Nodes[0],
												 vserver   => $params{server},
												 lif       => $vif,
												 'newname' => $params{rename}
			);
		}

	} ## end foreach my $vif (@{$params{...

	# If flagged to, first verify that the old zone is no longer resolving IPs.
	# Then we can verify the IP is properly resolved (in the case where
	# $params{zone} is a valid domain) via the new zone. We do a 1:1 verify
	# here because the IPs could have originally belonged to different zones.

	if ( $params{verify_ip} ) {

		foreach my $address ( keys %addr_old_zone ) {

		   # First - verify the old zone is no longer resolved if they were part
		   # of a DNS zone and are not anymore.

			if (    ( defined( $params{zone} ) )
				 && ( $addr_old_zone{$address} !~ /^none$/ )
				 && ( $addr_old_zone{$address} !~ /$params{zone}/ ) )
			{
				if (
					 !$Dns_Resolver->verify_ip(
											 ips    => $address,
											 domain => $addr_old_zone{$address},
											 state  => 0,
											 sleep  => 10
					 )
				  )
				{
					$Log->warn(   "DNS Zone $addr_old_zone{$address} resolved "
								. "IP address $address when expected not to." );
					return 0;
				} ## end if (!$Dns_Resolver->verify_ip...

				$Log->comment(
						   "IP Address $address is no longer resolved via zone "
							 . $addr_old_zone{$address} );
			} ## end if ((defined($params{zone...

		   # Now, if we have a valid new zone, we will verify that it is able to
		   # resolve the VIF IP's

			if ( ( defined( $params{zone} ) ) && ( $params{zone} !~ /^none$/ ) )
			{
				if (
					 !$Dns_Resolver->verify_ip(
												ips    => $address,
												domain => $params{zone},
												state  => 1
					 )
				  )
				{
					$Log->warn(   "DNS Zone $params{zone} did not resolve "
								. "IP address $address when expected to." );
					return 0;
				} ## end if (!$Dns_Resolver->verify_ip...

				$Log->comment(
						  "IP Address $address was correctly resolved via zone "
							. $params{zone} );
			} ## end if ((defined($params{zone...

  # If we did not modify the actual zone (case where we modifed allow_lb_migr or
  # lb_weight, just check to make sure the previous zone is still recognized.

			if (    ( !defined( $params{zone} ) )
				 && ( $addr_old_zone{$address} !~ /^none$/ ) )
			{
				if (
					 !$Dns_Resolver->verify_ip(
											 ips    => $address,
											 domain => $addr_old_zone{$address},
											 state  => 1
					 )
				  )
				{
					$Log->warn(
							"DNS Zone $addr_old_zone{$address} did not resolve "
							  . "IP address $address when expected to." );
					return 0;
				} ## end if (!$Dns_Resolver->verify_ip...
			} ## end if ((!defined($params{...
		} ## end foreach my $address (keys %addr_old_zone)
	} ## end if ($params{verify_ip})

	$Log->exit(   "Successfully modified $vif_count VIFS and their "
				. "zones are resolving accordingly." );
	return $vif_count;

} ## end sub modify_vif_dns

###############################################################################
# Subroutine: delete_vif_dns
# Parameters: DNS Zone and VIF name of VIF to be deleted (optional server)
# Returns: Number of VIFs whose DNS zones were deleted, 0 on failure
# Purpose: To delete the DNS zone attribute of a VIF (end result will be
#          changing its DNS zone to 'none'.
###############################################################################

sub delete_vif_dns {

	$Log->enter();
	my $self = shift(@_);

	# Initialize the parameters to perform the VIF modification

	my %params = validate(
						   @_,
						   {
							  server => {
										  type     => SCALAR,
										  optional => 1,
										  default  => $self->{_server}
							  },
							  vif       => { type => SCALAR | ARRAYREF, },
							  verify_ip => {
											 type     => SCALAR,
											 optional => 1,
											 default  => 1,
							  },
							  delete_group => {
												type     => SCALAR,
												optional => 1,
												default  => 1,
							  },
						   }
	);

	# Verify server exists

	if ( !Network::VIFMgr::VifConfig::_server_exists( server => $params{server} ) ) {
		$Log->warn("Server: $params{server} does not exist. Returning...");
		return 0;
	}

	my $vif_count = 0;
	my %addr_zone;
	my @show;

	# If the calling method passed in a wildcard character (or single vif name)
	# instead of a list of vifs, we need to extract out each IP/zone before the
	# bulk delete occurrs. This is for verification purposes later on.

	if ( ref( $params{vif} ) ne 'ARRAY' ) {

		@show = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $params{vif},
												 }
		);

		foreach my $vif_ref (@show) {

			# Get the IP addreses and dns-zone

			my %vif = %{$vif_ref};
			$addr_zone{ $vif{address} } = $vif{'dns-zone'};
			$vif_count++;
		}

	 # Since calling method can specify an array of VIF names to delete, cast to
	 # array if not already.

		$params{vif} = [ $params{vif} ];

	}
	else {

		# If we already have a list of vif names, we can do a show on each
		# one and parse out the ip and zone from there

		foreach my $vif ( @{ $params{vif} } ) {
			@show = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $params{vif},
												 }
			);

			my %vif = %{ $show[0] };
			$addr_zone{ $vif{address} } = $vif{zone};
			$vif_count++;
		} ## end foreach my $vif (@{$params{...
	} ## end else [ if (ref($params{vif}) ...

# Delete the VIF (we do not increment vif_count here becasuse it could be a bulk
# delete - with the wildcard (*) character).

	foreach my $vif ( @{ $params{vif} } ) {

		# First, check to see if we are deleting the routing-group. If we are,
		# we need that group name after we delete the VIF.

		my $group = undef;
		if (    ( $Server->has_change( change => "routing-groups" ) )
			 && ( $params{delete_group} ) )
		{
			my @show = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     =>$vif,
												 }
			);
			$group = $show[0]->{'routing-group'};
		}

		# Now delete the VIF

		NACL::STask::NetworkInterface->purge(
											  'command_interface' => $Nodes[0],
											  'lif'               => $vif,
											  'vserver' => $params{server}
		);

		# Now, before we delete the routing-group, let's make sure this is the
		# last VIF in the group. Again, only do if we are running with 'routing-
		# groups' (which we can tell by check if $group is set).

		if ( defined($group) ) {

			try {
				NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 'routing-group' => $group,
												 }
				);
			}
			 catch NACL::Exceptions::NoElementsFound with {
				if (
					 !Network::VIFMgr::VifConfig::delete_routing_group(
													  $self,
													  server => $params{server},
													  group  => $group
					 )
				  )
				{
					$Log->warn("No group to delete for vif $vif");
					return 0;
				} ## end if (!VifMgr::VifConfig::delete_routing_group...
			} ;## end if ($output =~ /There are no entries matching your query./)

		} ## end if (defined($group))
	} ## end foreach my $vif (@{$params{...

   # If flagged to, verify the IP is no longer resolved. We do a 1:1 verify here
   # because the IPs could belong to different zones.

	if ( $params{verify_ip} ) {

		foreach my $address ( keys %addr_zone ) {

			# We will continue if this VIF's IP did not belong to a zone
			# Must also check for the empty string due to burt 291322

			if (    ( $addr_zone{$address} =~ /none/ )
				 || ( $addr_zone{$address} =~ /^$/ ) )
			{
				$Log->comment("No zone to resolve for IP $address");
				next;
			}

			if (
				 !$Dns_Resolver->verify_ip(
											ips    => $address,
											domain => $addr_zone{$address},
											state  => 0,
											sleep  => 10
				 )
			  )
			{
				$Log->warn(   "DNS Zone $addr_zone{$address} resolved "
							. "IP address $address when expected not to." );
				return 0;

			} ## end if (!$Dns_Resolver->verify_ip...

			$Log->comment( "IP Address $address is no longer resolved via zone "
						   . $addr_zone{$address} );
		} ## end foreach my $address (keys %addr_zone)
	} ## end if ($params{verify_ip})

	$Log->exit(   "Successfully deleted the DNS Zone ($params{zone}) for "
				. "$vif_count VIF(s)" );
	return $vif_count;

} ## end sub delete_vif_dns

##############################################################################
# Subroutine: verify_lb_value
# Parameters: List of VIFs to verify that -allow-lb-migrate/lb-weight/-dns-zone
#	      has been flagged properly, attribute to verify (either
#	      allow_lb_migr, zone or lb_weight), and expected value
# Returns: 1 on success, 0 on failure
# Purpose: To verify that the -allow-lb-migrate/lb-weight/-dns-zone attribute
#	   can be successfully toggled for VIFs
###############################################################################

sub verify_lb_value {

	$Log->enter();
	my $self = shift(@_);

	# Initialize the parameters to perform the VIF modification

	my %params = validate(
						   @_,
						   {
							  server => {
										  type     => SCALAR,
										  optional => 1,
										  default  => $self->{_server}
							  },
							  vif       => { type => SCALAR | ARRAYREF, },
							  attribute => {
											 type     => SCALAR,
											 optional => 1,
											 default  => "allow-lb-migrate"
							  },
							  value => {
										 type     => SCALAR,
										 optional => 1,
										 default  => "false"
							  },
						   }
	);

	# Verify server exists

	if ( !Network::VIFMgr::VifConfig::_server_exists( server => $params{server} ) ) {
		$Log->warn("Server: $params{server} does not exist. Returning...");
		return 0;
	}

	# Calling method can specify an array of VIF names to verify, cast to
	# array if not already.

	if ( ref( $params{vif} ) ne 'ARRAY' ) {
		$params{vif} = [ $params{vif} ];
	}

	# Perform the show and verify on a per-VIF basis

	foreach my $vif ( @{ $params{vif} } ) {

		# Show based on the state

		my @output = NACL::CS::NetworkInterface->fetch(
												 command_interface => $Nodes[0],
												 filter            => {
													 vserver => $params{server},
													 lif     => $vif,
													 $params{attribute} => $params{value}
												 }
			);
											

		# Verify that attributes are displayed correctly in the output
		# Can use $otuput[0] because we are viewing on a per-vif basis.

		my %vif_out = %{ $output[0] };
		if ( $vif_out{'vif'} ne $vif ) {
			$Log->warn(   "VIF name displayed incorrectly. Found: "
						. $vif_out{'vif'}
						. ", Expected: $vif" );
			return 0;
		}
		if ( $vif_out{ $params{attribute} } ne $params{value} ) {
			$Log->warn(   "$params{attribute} displayed incorrectly. "
						. "Found: $vif_out{$params{attribute}}, "
						. "Expected: $params{value}" );
			return 0;
		}
	} ## end foreach my $vif (@{$params{...

	$Log->exit(   "Successfully verified that -$params{attribute} has "
				. "been flagged to $params{value} for VIFs: "
				. Dumper( $params{vif} ) );
	return 1;

} ## end sub verify_lb_value

1;
