# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/lib/Network/Vlan.pm#2 $
#
## @summary	Vlan.pm - VLAN interface for VLAN commands on hosts
## @author	dl-spe-nicqa@netapp.com
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

Vlan - Routines for VLAN commands on hosts

=head1 SYNOPSIS

  use Vlan qw(vlan_create vlan_add vlan_delete vlan_modify vlan_stat);

  my $vlan = vlan_create(fc      => $fc,
                         gvrp    => 'on',
                         if_name => $ifc,
                         vlanid  => '501');

  vlan_add(fc      => $fc,
           if_name => $ifc2,
           vlanid  => '501');

  my $state = vlan_modify(fc      => $fc,
                          gvrp    => 'on',
                          if_name => $ifc);

  my $stat_href = vlan_stat(fc      => $fc,
                              if_name => $ifc,
                              vlanid  => '501');

  print Dumper($stat_href);
  $VAR1 = {
        'e0a' => {
                    'gvrp' => 'enabled',
                    'receive' => {
                                   'total_frames' => '14',
                                   'multi_broadcast' => '16',
                                   'total_bytes' => '15',
                                   'vlan_tag_drops' => '18',
                                   'untag_drops' => '17'
                                 },
                    'vlan ids' => '501',
                    'vlan' => {
                                'receive' => {
                                               'total_frames' => '21',
                                               'multi_broadcast' => '23',
                                               'total_bytes' => '22'
                                             },
                                'mac address' => '02:a0:98:08:44:33',
                                'name' => 'e0a-501',
                                'id' => '501',
                                'transmit' => {
                                              'queue_overflows' => '27',
                                              'total_frames' => '24',
                                              'multi_broadcast' => '26',
                                              'total_bytes' => '25'
                                              }
                              },
                    'transmit' => {
                                    'total_frames' => '19',
                                    'total_bytes' => '20'
                                  }
                  }
      };

  vlan_delete(fc      => $fc,
              if_name => $ifc,
              vlanid  => '501');

=head1 DESCRIPTION

Provides an API to VLAN calls used on filers and linux clients.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-spe-nicqa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Network/Vlan.pm#2 $
  Last revised $Date: 2013/04/30 $

head1 COPYRIGHT

  Copyright 2007 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Network::Vlan;

use strict;
use warnings;

use Carp qw(croak);
use Network::Utils qw(do_ifconfig);
use Params::Validate qw(validate OBJECT SCALAR UNDEF);
use SPARSE::Util::Logger qw(logStep logWarn);
use SayWrap qw(sayEOF);
use Tharn;
use base qw(Exporter);
use Data::Dumper;
our @EXPORT_OK = qw(
    vlan_create
    vlan_add
    vlan_delete
    vlan_modify
    vlan_stat
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

my $sudo = '/usr/local/bin/sudo';
my $vconfig = '/sbin/vconfig';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub vlan_create(%);
sub vlan_add(%);
sub vlan_delete(%);
sub vlan_modify(%);
sub vlan_stat(%);
sub _change_keys($);
sub _parse_stat_info($);

#---------------------------------------------------------------------

=head2 vlan_create

    my $vlan = vlan_create(fc       => $fc,
                           gvrp     => 'on',
                           if_name  => $ifc,
                           vlanid   => '501',
                           vlan_idx => '1',
                           mac_addr => '00:17:43:05:ac:45');

This routine creates a VLAN using the interface specified by if_name
and the VLAN id. If the reuse flag is set then if the VLAN
was already created it is used, otherwise the routine would croak
with an error. Currently only supported for the filer and Linux.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of host.

=item gvrp

(Optional) : Flag to enable/disable GVRP on the interface. This option
is for the filer. Defaults to 'off'.

=item if_name

(Required) : Network interface to use for VLAN.

=item reuse

(Optional) : The VLAN can be reused if it already exists.
Defaults to 'no'.

=item vlanid

(Required) : VLAN id to use when performing VLAN operations.

=item vlan_idx

(Optional) : Index number to use for naiming the vlan (e.g., e0c-<index>). Defaults to the requested vlanid.

=item mac_addr

(Optional): MAC Address for the requested VLAN interface. By default, inherits the MAC of the parent interface.

=back

=item Returns

VLAN created or croaks on error

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vlan_create(%) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            gvrp    => { type => SCALAR, default => 'off' },
            if_name => { type => SCALAR },
            reuse   => { type => SCALAR, default => 'no' },
            vlanid  => { type => SCALAR },
            mac_addr => { type => SCALAR, default => '', regex => qr/^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$/ },
            vlan_idx => { type => SCALAR, default => '', regex => qr/^[0-9]+$/ },
        }
    );
    my ($cmd, $out, $vlan_ifc, $vlan_idx);

    my $mac_option = '';
    my $vlanid_option = '';
    my $gvrp_option = '';

    if ($opts{fc}->hostp()->hosttype =~ /filer/) {
        # Bring the filer interface down
        #do_ifconfig($opts{fc}, $opts{if_name}, state => 'down');
        logStep("Vlan check: assigning the ip addr 0.0.0.0");
        do_ifconfig($opts{fc}, $opts{if_name}, ip_addr=>"0.0.0.0", 'state' => 'up');#change 1: addition for [7mode] -Karthik
        #do_ifconfig($opts{fc}, $opts{if_name}, state => 'down');                   #change 2: swapped the order[7 mode]-Alur 

	if ($opts{gvrp} eq 'off') {
	    $gvrp_option = '-g off';
	}
	elsif ($opts{gvrp} eq 'on') {
	    $gvrp_option = '-g on';
	}
	if ($opts{mac_addr} ne '') {
	    $mac_option = '-m ' . $opts{mac_addr};
	}
	if ($opts{vlan_idx} ne '') {
	    $vlanid_option = '-t ' . $opts{vlanid};
	    $vlan_idx = $opts{vlan_idx};
	}
	else {
	    $vlan_idx = $opts{vlanid}; # legacy command format (index replaces id in VLAN name)
	}
	$cmd = "vlan create $gvrp_option $mac_option $vlanid_option $opts{if_name} $vlan_idx";

        #do_ifconfig($opts{fc}, $opts{if_name}, state => 'up');
        $out = sayEOF($opts{fc}, $cmd);

        if ($out !~ /($opts{if_name}\-$vlan_idx)\s+has\s+been\s+created/i) {
            if (($opts{reuse} eq 'no') ||
                ($out !~ /already\s+configured\s+for\s+VLAN\s+tagging/i)) {
#croak "vlan_create error: $out";
		    return 0;
            }
        }
        $vlan_ifc = $1;
    } elsif ($opts{fc}->hostp()->hosttype =~ /linux/) {
        # Linux doesn't return a conclusive error if the VLAN is already
        # configured, it returns 'error: Invalid argument'. If the reuse
        # option is set then try deleting the VLAN first ignoring any
        # returned error.
        if ($opts{reuse} eq 'yes') {
            eval {
                vlan_delete(fc => $opts{fc}, if_name => $opts{if_name},
                            vlanid => $opts{vlanid});
            };
        }

        # Bring the interface up, vconfig will fail if the interface is down
        do_ifconfig($opts{fc}, $opts{if_name}, state => 'up');

        $cmd = "$sudo $vconfig add $opts{if_name} $opts{vlanid}";

        $out = sayEOF($opts{fc}, $cmd);
        if ($out !~ /Added\sVLAN\swith\sVID\s==\s$opts{vlanid}/i) {
            croak "vconfig error: $out";
        }
        $vlan_ifc = "$opts{if_name}.$opts{vlanid}";
    } else {
        croak "vlan_create doesn't now how to work with " .
              $opts{fc}->hostp()->hosttype;
    }
    return $vlan_ifc;
}

#---------------------------------------------------------------------

=head2 vlan_add

    vlan_add(fc       => $fc,
             if_name  => $ifc,
             vlanid   => '501');
             vlan_idx => '1',
             mac_addr => '00:17:43:05:ac:45');

This routine adds an interface specified by if_name
to the vlanid specified. Currently only supported for the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of host.

=item if_name

(Required) : Ethernet interface to use for VLAN.

=item vlanid

(Required) : VLAN id to use when performing VLAN operations.

=item vlan_idx

(Optional) : Index number to use for naiming the vlan (e.g., e0c-<index>). Defaults to the requested vlanid.

=item mac_addr

(Optional): MAC Address for the requested VLAN interface. By default, inherits the MAC of the parent interface.

=back

=item Returns

Nothing or croaks on error

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vlan_add(%) {
    my (%opts) = validate (
        @_,
        {
            fc       => { type => OBJECT },
            if_name  => { type => SCALAR },
            vlanid   => { type => SCALAR },
            mac_addr => { type => SCALAR, default => '', regex => qr/^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$/ },
            vlan_idx => { type => SCALAR, default => '', regex => qr/^[0-9]+$/ },
        }
    );
    my ($cmd, $out, $vlan_idx, $vlan_ifc);
    my $mac_option = '';
    my $vlanid_option = '';

    if ($opts{fc}->hostp()->hosttype =~ /filer/) {
        if ($opts{mac_addr} ne '') {
            $mac_option = '-m ' . $opts{mac_addr};
        }
        if ($opts{vlan_idx} ne '') {
            $vlanid_option = '-t ' . $opts{vlanid};
            $vlan_idx = $opts{vlan_idx};
        }
        else {
            $vlan_idx = $opts{vlanid}; # legacy command format (index replaces id in VLAN name)
        }
        $cmd = "vlan add $mac_option $vlanid_option $opts{if_name} $vlan_idx";

        $out = sayEOF($opts{fc}, $cmd);
        if ($out !~ /($opts{if_name}\-$vlan_idx)\s+has\s+been\s+created/i) {
            croak "vlan_add error: $out";
        }
	$vlan_ifc = $1;
    } else {
        croak "vlan_add doesn't know how to work with " .
              $opts{fc}->hostp()->hosttype;
    }

    return $vlan_ifc;
}

#---------------------------------------------------------------------

=head2 vlan_delete

    vlan_delete(fc      => $fc,
                if_name => $ifc,
                vlanid  => '501');

This routine deletes the VLAN interface specified by if_name
and vlanid. Currently only supported for the filer and Linux.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of host.

=item if_name

(Required) : Ethernet interface to use for VLAN.

=item vlanid

(Required) : VLAN id to use when performing VLAN operations.

=back

=item Returns

Nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vlan_delete(%) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            if_name => { type => SCALAR },
            vlanid  => { type => SCALAR },
        }
    );
    my ($cmd, $out);

    if ($opts{fc}->hostp()->hosttype =~ /filer/) {
        # Bring the filer VLAN interface down
        do_ifconfig($opts{fc}, "$opts{if_name}-$opts{vlanid}", 'ip_addr' => '0.0.0.0');
        do_ifconfig($opts{fc}, "$opts{if_name}-$opts{vlanid}", state => 'down');

        $cmd = "vlan delete -q $opts{if_name} $opts{vlanid}";

        $out = sayEOF($opts{fc}, $cmd);
        if ($out =~ /Delete\s+failed/i) {
            croak "vlan_delete error: $out";
        }
    } elsif ($opts{fc}->hostp()->hosttype =~ /linux/) {
        $cmd = "$sudo $vconfig rem $opts{if_name}.$opts{vlanid}";

        $out = sayEOF($opts{fc}, $cmd);
        if ($out !~ /Removed\s+VLAN/i) {
            croak "vlan_delete error: $out";
        }
    } else {
        croak "vlan_delete doesn't now how to work with " .
              $opts{fc}->hostp()->hosttype;
    }

    return;
}

#---------------------------------------------------------------------

=head2 vlan_modify

    my $state = vlan_modify(fc      => $fc,
                            gvrp    => 'on',
                            if_name => $ifc);

This routine allows GVRP to be enabled or disabled on the interface
if_name. Currently only supported for the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of host.

=item gvrp

(Optional) : Flag to enable/disable GVRP on the interface.
Default is 'off'.

=item vlanid

(Optional) : VLAN id to use when performing VLAN operations.

=item vlan_idx

(Optional) : Index number to use for naiming the vlan (e.g., e0c-<index>). Defaults to the requested vlanid.

=item if_name

(Required) : Ethernet interface to use for VLAN.

=back

=item Returns

State of if_name (enabled|disabled) or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vlan_modify(%) {
    my (%opts) = validate (
        @_,
        {
            fc       => { type => OBJECT },
            gvrp     => { type => SCALAR, default => '' },
            if_name  => { type => SCALAR },
	    vlanid   => { type => SCALAR, default => '', regex => qr/^[0-9]+$/ },
	    vlan_idx => { type => SCALAR, default => '', regex => qr/^[0-9]+$/ },
        }
    );
    my ($cmd, $href, $out, $status);

    my $gvrp_option = '';
    my $vlanid_option = '';

    # preserve original default (prior to vlanid option)
    if (($opts{vlan_idx} eq '') and ($opts{gvrp} eq '')) {
        $opts{gvrp} = 'off';
    }

    if ($opts{fc}->hostp()->hosttype =~ /filer/) {
        if ($opts{gvrp} eq 'off') {
	    $gvrp_option = '-g off';
            $status = 'disabled';
        } elsif ($opts{gvrp} eq 'on') {
	    $gvrp_option = '-g on';
            $status = 'enabled';
	} elsif ($opts{vlan_idx} eq '') {
	    croak "GVRP option is either on (enabled) or off (disabled) per ".
		  "vlan modify [ -g {on|off} ] <if_name>";
	}
	if ($opts{vlanid} ne '') {
	    $vlanid_option = "-t " . $opts{vlanid};
	}

	$cmd = "vlan modify $gvrp_option $vlanid_option $opts{if_name} $opts{vlan_idx}";
        $out = sayEOF($opts{fc}, $cmd);
        if ($out =~ /Modify\s+failed/) {
            croak "vlan_modify error: $out";
        }

        $href = vlan_stat(fc => $opts{fc});
	if ($gvrp_option) {
	    if(!defined($href) ||($href->{$opts{if_name}}{gvrp} ne $status)) {
		croak "vlan_modify error: $out";
	    }
	}
	if ($vlanid_option) {
	    if(!defined($href) ||($href->{$opts{if_name}}{vlan}{'vlan id'} ne $opts{vlanid})) {
		croak "vlan_modify error: $out";
	    }
	}
    } else {
        croak "vlan_modify doesn't now how to work with " .
              $opts{fc}->hostp()->hosttype;
    }

    return $href->{$opts{if_name}}{gvrp};
}

#---------------------------------------------------------------------

=head2 vlan_stat

    my $stat_href = vlan_stat(fc      => $fc,
                              if_name => $ifc,
                              vlanid  => '501');

This routine displays the number of packets received and transmitted on
each link that makes up the VLAN interface.
Currently only supported for the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of host.

=item if_name

(Optional) : Ethernet interface to use for VLAN.

=item vlanid

(Optional) : VLAN id to use when performing VLAN operations.

=back

=item Returns

Hash reference containing VLAN information for the if_name and vlanid

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub vlan_stat(%) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            if_name => { type => SCALAR|UNDEF, optional => 1 },
            vlanid  => { type => SCALAR|UNDEF, optional => 1 },
        }
    );
    my ($cmd, $out, $stat_href);

    if ($opts{fc}->hostp()->hosttype =~ /filer/) {
        if (!defined($opts{if_name}) || !defined($opts{vlanid})) {
            $cmd = "vlan stat";
        } else {
            $cmd = "vlan stat $opts{if_name} $opts{vlanid}";
        }

        $out = sayEOF($opts{fc}, $cmd);
        if ($out =~ /Display\s+stat\s+failed/i) {
            croak "vlan_stat error: $out";
        } else {
            $stat_href = _parse_stat_info($out);
        }
    } else {
        croak "vlan_stat doesn't now how to work with " .
              $opts{fc}->hostp()->hosttype;
    }

    return $stat_href;
}

#---------------------------------------------------------------------
# Internal function to parse vlan stat information
#---------------------------------------------------------------------

sub _parse_stat_info($) {
    my $output = shift;
    my ($dev, %h, %hash, $header, $href, $key, @list, $name,
        $pname, $phys, $stat, @stats, @sub_stat);

    # do some processing on the output returned from the filer
    $output =~ s/^\s+//mg;
    $output =~ s/\s+$//mg;
    $href = \%h;

    # split stats information by every physical interface and remove
    # the first invalid entry
    @stats = split (/Vlan\s+Physical\s+Interface:\s+/i, $output);
    shift @stats;

    # Do some processing on every physical interface entry and
    # the information
    foreach $dev (@stats) {
        $phys = 1;

        # Each physical entry has VLAN interface information so we split the
        # physical RECEIVE/TRANSMIT stats from the VLAN interface
        # RECEIVE/TRANSMIT stats
        @sub_stat = split (/Vlan\s+Interface:\s+/i, $dev);
        foreach $stat ( @sub_stat ) {

            # convert everything to lowercase here so we can access the
            # members of the hash created using lowercase keys.
            $stat = "\L$stat\E";

            # The first member of each array during the split is the name of
            # the physical device. Once we have this information remove it
            # it from the array and process the RECEIVE/TRANSMIT stats.
            ($name) = ($stat =~ /^(\S+)/);
            $stat =~ s/^.*--\n?//;
            ($header, @list) = split
                            (/(RECEIVE|TRANSMIT) STATISTICS\s*\n?/i, $stat);

            %h = ($header =~ /([^:]+):\s+(\S+)\n?/g);

            # Remove all white space and / from the values used as keys in the
            # hash
            $href = _change_keys(\%h);

            while ( $key = shift @list ) {
                $h{$key} = { (shift @list) =~
                                        /([^:]+):\s+(\S+)\s*(?:\n|\|?)\s*/g };
                $h{$key} = _change_keys(\%{$h{$key}});
            }
            if ($phys) {
                $hash{$name} = { %h };
                $pname = $name;
            } else {
                $hash{$pname}{vlan} = { name => $name, %h };
            }
#	    $hash{$pname}{'vlan indices'} = $hash{$pname}{'vlan ids'};
	    $hash{$pname}{'vlan ids'} = $hash{$pname}{'vlan indices'};
            $phys = 0;
	    
        }
    }
    logcomment(Dumper(\%hash));
##For IC compatibility
    return (\%hash);
}

#---------------------------------------------------------------------
# Internal function to change spaces on the keys of the specified hash
#---------------------------------------------------------------------
sub _change_keys($) {
    my $hash = shift;
    my (%ret, $key);

    # Process all keys in the hash
    foreach ( keys %$hash ) {
        $key = $_;
        # Change spaces to underscore
        $key =~ s/\s+/_/g;
        $key =~ s/\//_/g;
        $ret{$key} = $hash->{$_};
    }
    return \%ret;
}

1;
