# $Id$
#
#  Tharndoc comments:
#

## @summary Perl module used to for interaction with filer NICs
## @status reviewed
## @author Michael Eddy
## @pod here

=head1 NAME

    Nic.pm


=head1 DESCRIPTION

    This module provides NIC related functionality

=head1 SYNOPSIS

    use Nic;
    see EXAMPLES for method usage examples

=head1 EXAMPLES

    get_nic_model_info example:
        $info{$filer} = Nic::get_nic_model_info($filer);


=head1 COPYRIGHT

    "Copyright 2008, Network Appliance Inc., All Rights Reserved."

=head1 STATUS
    Shared

=head1 ATTRIBUTES

=head1 AUTHOR

    Michael Eddy

=head1 ROUTINES

=cut

#=head1 ROUTINES
#    
#    get_nic_model_info
#      inputs: node name
#      output: hash containing nic info for the given filer
#

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDE_NIC}) {
    my $newModule = $ENV{OVERRIDE_NIC};
    delete $ENV{OVERRIDE_NIC};
    print "The main Nic module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Nic;
use strict;
use TracePkg;
use Data::Dumper;
use Server;
use Params::Common;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF UNDEF);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
		 get_nic_model_info
		 );
our @EXPORT_OK = qw(
		 get_nic_model_info
		 );


my $Trace = new TracePkg();

#
# Nic::get_nic_model_info
# 
#

=head2 get_nic_model_info 

    $info{$filer} = get_nic_model_info(node => $filer);

This method will query the filer and return a hash containing the NIC model info.

Fatal is called on error.

=over

=item node

 (Required) : Hostname of filer

=back

=cut

sub get_nic_model_info
{
    $Trace->entry();
    my $sysctl_cmd = 'sysctl -a hw.pcidev';
    my $output;
    my $sysconfigtab;
    my %nic_info;
    my $server = new Server();
    my %params = Params::Validate::validate (
	@_,
	{
	    node => { type => SCALAR },
	    sysconfigtab => { type => SCALAR },
	}
    );
    
    # Set the execution server
    $Trace->debug(msg=>"Setting execution server to "..$params{node}."...");
    $server->set_execution_server(-Server => $params{node});

    # Look up the netapp model number from sysconfigtab
    $Trace->debug(msg=>"Using the following location for sysconfigtab: $params{sysconfigtab}");
    $server->shell(cmd=>"cat $params{sysconfigtab}", output=>\$sysconfigtab);
    $Trace->debug(msg=>"Dump of sysconfigtab:\n".Dumper($sysconfigtab));
    
    # Get the sysctl info
    $Trace->trace(msg=>"Getting nic info from sysctl...");
    $server->shell(cmd=>$sysctl_cmd, output=>\$output);
    $Trace->debug(msg=>"Dump of sysctl command:\n".Dumper($output));
    # Organize sysctl info by "real" slot
    my $pcislot = 0;
    while ($output =~ /hw.pcidev.$pcislot/){
	my $realslot;
	my $device;
	my $devicesubid = 0;
	my $vendor;
	my $model;
	$Trace->trace(msg=>"Found info for pci slot $pcislot...");
	if ($output =~ /hw.pcidev.$pcislot.class:\snetwork/){
	    $Trace->trace(msg=>"The device in pci slot $pcislot is a nic.  Getting info...");
	    if ($output =~ /hw.pcidev.$pcislot.slot:\s(\d)/) {
		$realslot = $1;
	    } else {
		$Trace->fatal(msg=>"Did not find a system slot number for the device ".$params{node}."\'s pci slot $pcislot!");
	    }
	    # Put the device's realslot in the hash
	    $nic_info{"system_slot_$realslot"}{realslot} = $realslot;
	    # Put the device's pci slot in the hash
	    $nic_info{"system_slot_$realslot"}{pcislot} = $pcislot;
	    # Put the device's device ID in the hash
	    if ($output =~ /hw.pcidev.$pcislot.device:\s(\d*)/) {
		$device = $1;
		# convert device id to hex
		$device = sprintf("%X", $device);
		$nic_info{"system_slot_$realslot"}{device} = $device;
	    } else {
		$Trace->fatal(msg=>"Did not find a device ID for the device in ".$params{node}."\'s pci slot $pcislot!");
	    }
	    # Put the device's devicesubid in the hash
	    if ($output =~ /hw.pcidev.$pcislot.devicesubid:\s(\d*)/) {
		$devicesubid = $1;
		$nic_info{"system_slot_$realslot"}{devicesubid} = $devicesubid;
	    } else {
		$Trace->warn(msg=>"Did not find a device sub ID for the device in ".$params{node}."\'s pci slot $pcislot!\n Using Default devicesubid of $devicesubid.\n System Image may be prior to change 743242.");
	    }
	    # Put the device's vendor ID in the hash
	    if ($output =~ /hw.pcidev.$pcislot.vendor:\s(\d*)/) {
		$vendor = $1;
		# convert vendor id to hex
		$vendor = sprintf("%X", $vendor);
		$nic_info{"system_slot_$realslot"}{vendor} =  $vendor;
	    } else {
		$Trace->fatal(msg=>"Did not find a device ID for the device in ".$params{node}."\'s pci slot $pcislot!");
	    }
	    # Put the device's model number in the has
	    if ($realslot == 0) {
		$model = "onboard";
	    } else {
		if ($sysconfigtab =~ /card\s+(\w+)\s+\d*\s+0x$vendor\s+0x$device\s+$devicesubid.*/i){
		    $model = $1;
		} else {
		    $Trace->warn(msg=>"Did not find a model number for ".$params{node}."\'s $vendor $device in pci slot $pcislot (real slot $realslot)!  This may not be a supported card in this release.");
		    $model = "unknown";
		}
	    }
	    $nic_info{"system_slot_$realslot"}{model} = $model;
	} else {
	    $Trace->trace(msg=>"The device in pci slot $pcislot is NOT a nic.");
	}

	$pcislot ++;
    }

   
    $Trace->debug(msg=>"Dumping nic model info:\n".Dumper(\%nic_info));

    $Trace->exit();
    return \%nic_info;
    
} # end get_nic_model_info

1;
