#***********************************************************************
#
# $Id$
#
# NAME
#   Node.pm - remote programmatic administration of a Filer.
#
# SYNOPSIS
#   Enables remote programmatic administration of a Filer
#
# DESCRIPTION
#   This Perl module is meant to be programmatic way of administering a
#   Filer.
#
# AUTHOR
#    dl-central-libs
#
## @summary    Enables remote programmatic administration of a Filer
##
## @description This Perl module is meant to be programmatic way of administering a Filer.
##
## @author dl-central-libs
##
## @pod here
# POD at end of file

#***********************************************************************

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDENODE}) {
    my $newModule = $ENV{OVERRIDENODE};
    delete $ENV{OVERRIDENODE};
    print "The main Node module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);
package Node;
use TracePkg;
use Params::Volume;
use Params::Common; # bring in aggregate related params
use Options;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper;
use Storable qw(dclone);
use Time::HiRes qw(gettimeofday);
use boot();
use constant FAILURE => undef;
use constant SUCCESS => 1;
use constant SUCCESS_EXISTS => 2;
use constant RETRYOP => 3;

#use constant ONTAP_SCRIMSHAW => 1;
use constant ONTAP_CLASSIC => 2;
use constant ONTAP_NG => 3;

# The following constants are used for the Transit code
#  This is hopefully temporary and we will get one piece of 
#  transit code for both 7 and 10 mode
use constant UNKNOWN  => 'UNKNOWN';
use constant PANIC    => 'PANIC';
use constant NO_MROOT => 'NO_MROOT';
use constant HANG     => 'HANG';
use constant HALT     => 'HALT';
use constant CLI      => 'CLI';
use constant SHELL    => 'SHELL';
use constant RLM      => 'RLM';
use constant CFE      => 'CFE';
use constant LOGIN    => 'LOGIN';
use constant IBM      => 'IBM';
use constant REBOOT   => 'REBOOT';
use constant SHUTDOWN => 'SHUTDOWN';
use constant NETBOOT  => 'NETBOOT';
use constant SINGLEUSER    => 'SINGLEUSER';
use constant SHELL_REBOOT  => 'SHELL_REBOOT';
use constant CFE_REBOOT    => 'CFE_REBOOT';
use constant BSD_REBOOT    => 'BSD_REBOOT';
#use constant BSD           => 'BSD';
use constant DBLADECLI     => 'DBLADECLI';
use constant SIMSETUP => 'SIMSETUP';
use constant RESET_ALL => 'REST_ALL';

use constant NORMAL   => 'NORMAL';
use constant INSTALL  => 'INSTALL';
use constant SETUP    => 'SETUP';
use constant DISKINIT => 'INIT';
use constant MAINT    => 'MAINT';

# End of Transit constants

my $Trace;
my $Options;

# Declaring class variable $priv_string for use with Node Scope commands
my $priv_string = q{};

# This is a list of hand-created cdef library files which will be loaded by all objects
my @Modules = qw(Cfe);

# This array contains the parameters that are for internal use only; not
# for the command itself
my @Internal_params = qw(AUTOLOAD_COMMAND -Verification -Errors timeout -Timeout -SkipINFO maxidle internal_verification -Autoloaded -Help);

$Trace = new TracePkg;
$Options = new Options;
$Options->parameters(definitions=>
                     [
                      {
                       name=>"SHELL_USER",
                       spec=>"s",
                       description=>"the name to use when logging into a server shell",
                      },
                      {
                       name=>"SHELL_PASSWORD",
                       spec=>"s",
                       description=>"the password to use when logging into a server shell",
                      },
                      {
                       name=>"CLI_USER",
                       spec=>"s",
                       description=>"the name to use when logging into the server CLI",
                      },
                      {
                       name=>"CLI_PASSWORD",
                       spec=>"s",
                       description=>"the password to use when logging into a server CLI",
                      },
                      {
                       name=>"cmdnode",
                       spec=>"s@",
                       description=>"the management nodes to configure",
                      },
                      {
                       name=>"skipcmdnode",
                       spec=>"s@",
                       description=>"the management nodes to not skip during configuration",
                      },
                      {
                       name=>"CLI_PARAMETER_VALIDATION",
                       description=>"Should the object validate the parameters prior to sending them to the CLI",
                       default=>0,
                      },
                      {
                       name=>"NATE_LIB",
                       description=>"The list of nate directories to search",
                      },
                      {
                       name=>"CDEFS_DIRECTORY",
                       spec=>"s",
                       description=>"The directory to find the cdefs",
                      },
                      {
                       name=>"TOPOLOGY",
                       spec=>"s@",
                       description=>"The file that defines the topology",
                      },
                      {
                       name=>"timeout",
                       description=>"The timeout value (in seconds) to be used for CLI connections and CLI commands",
                      },
                      {
                       name=>"maxidle",
                       description=>"The maximum amount of time to allow a CLI connections or CLI command to remain idle",
                      },
                      {
                       name=>"CMDPROF",
                       description=>"Enable cli command profiling and dump hash block to given text file",
                       default => undef(),
                      },
                      {
                       name=>"CMDPROF_CSV",
                       description=>"Enable cli command profiling and dump vars to given csv file",
                       default => undef(),
                      },
                      {
                       name=>"CMDPROF_RT",
                       description=>"Enable cli command profiling in realtime, last N profiles stored in server object",
                       default => 0,
                      },
                      {
                       name=>"WARN_TIMEOUT",
                       description=>"Take the timeout value, double it, and then WARN at the origional timeout and FAIL on the double of it",
                       default=>1
                      },
                      {
                       name=>"TIMEOUT_MULTIPLIER",
                       description=>"Take the timeout value, multiply it by this value and use that as the new timeout on all say_wrapper calls",
                       default=>1
                      },
                      {
                       name=>"SAY_SWALLOW",
                       description=>
                        "Flag indicating that '-swallow' should be passed to say() (except on console) " . 
                        "Enabling swallow adds about 2 seconds per say() call.  Console connections will " .
                        "always have swallow enabled on them to prevent errors based on our ".
                        "experience consoles really need to have swallow enabled for proper operation.",
                       spec=>"boolean",
                       default=>0,
                      },
                      {
                       name=>"SKIP_MROOT_AGGREGATE",
                       description=>"Indicates that aggregates containing the mroot should be skipped",
                       spec=>"boolean",
                       default=>0,
                      },
                      {
                       name=>"POKE_CLI",
                       spec=>"boolean",
                       description=>"Flag to indicate whether or not to poke (i.e. send an extra carriage return to) the CLI",
                       default=>0,
                      },
                      {
                       name=>"CONNECTION_MAXBUFSIZE",
                       spec=>"e",
                       description=>"The value for the connection maxbufsize",
                      },
                      {
                       name=>"RETRY_SAY_FAILURE",
                       spec=>"time",
                       description=>"Try to reconnect and send the command again on a say() failure. The value is the amount of time to continue retrying.",
                      },
                      {
                       name=>"RETRY_SAY_FAILURE_INTERVAL",
                       spec=>"time",
                       default => 30,
                       description=>"How often should the retry be attempted when RETRY_SAY_FAILURE is active",
                      },
                     ]);

#***********************************************************************

sub get_release {
    my $self = shift;
    my $rel;
    $Trace->entry();
    $Trace->exit(msg=>"return $self->{release}");
    return $self->{release};
}

#***********************************************************************

sub get_build {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"build=$self->{build}");
    return $self->{build};
}

#***********************************************************************

sub compare_version {
    my $self = shift;
    $Trace->entry();
    $Trace->fatal(msg=>"compare_version is not implemented by the $self object\n", result=>"SCRIPT");
    $Trace->exit(msg=>"build=$self->{build}");
    return undef;
}

#***********************************************************************

# This routine is called at the beginning of Server::delete_topology() and allows for each
#  node type to execute code prior to anything actually being deleted. It is
#  not marked as private but only delete_topology()_should need it
sub pre_delete_topology {
    my $self = shift;
    $Trace->entry();
    $Trace->exit();
}

#***********************************************************************

# The keys are expected to be of the form text.alphanumeric
# Sort so that numbers are sorted numerically with an alpha
#  component greater than a numeric value
sub _alpha_numerically { 
    my ($a_alpha, $a_numeric) = split(/\./, $a);
    my ($b_alpha, $b_numeric) = split(/\./, $b);

    if ($a_numeric =~ /^\d+$/ && $b_numeric =~ /^\D+$/) {
	return -1;
    } elsif ($a_numeric =~ /^\D+$/ && $b_numeric =~ /^\d+$/) {
	return 1;
    } else {
	$a_numeric <=> $b_numeric;
    }
}


sub _find_cdef_path {
    my ($self, $dir) = @_;
    $Trace->entry(msg=>"dir=$dir\n");

    # $dir will contain the relative path to the cdef directory (i.e. Node/NG/cdefs/R10.0.1x)

    my ($cdefs_dir) = $self->{options_object}->get(option=>["CDEFS_DIRECTORY"]);

    if (!$cdefs_dir) {
        # use NATE to find the real path to the directory
        $cdefs_dir = NATE::Inc::find_file($dir);
    } else {
        # verify that the user-specified directory exists
        my $dir = NATE::Inc::find_file($cdefs_dir);
        if (! $dir) {
            $Trace->fatal(msg=>"The CDEFS_DIRECTORY ($cdefs_dir) does not exist");
        } else {
	    # If they provided a CDEFS_DIRECTORY, and it exists, don't play games and return
	    $Trace->exit(msg=>"return $cdefs_dir");
	    return $cdefs_dir;
	}
    }

    # Check to see if the derived directory is empty.  This is now a possibility once all the
    # CDEFs are moved to /usr/software (and p4 doesn't clean up directories)
    # see if the entries can be loaded with NATE calls
    my $file = NATE::Inc::get_file("$cdefs_dir/debug_trace.cdefs");
    if (! $file) {
	$Trace->trace(msg=>"Basic cdef debug_trace.cdefs not found.  Missing or empty directory.");
	$cdefs_dir = undef;
    }

    $Trace->exit(msg=>"return $cdefs_dir");
    return $cdefs_dir;
}


sub _setup_commands {
    my ($self, $class, $cdefs_dir, @modules) = @_;
    my $server = $self->{server};
    $Trace->entry();

    # now that we have connected to the servers and figured out Releases, setup the command structure
    if ($cdefs_dir) {

        # look at each entry in the cdefs directory and skip non-files. For the files, strip the
        #  .cdefs or .changelist extension and storage all version of the base name in an array
        #  for each basename. So volume.cdefs and volume.123456 will be stored as follows:
        #  %files = {
        #            volume => ["$cdefs_dir/volume.cdefs", "$cdefs_dir/volume.123456"],
        #           };
        my %files;


        # The following conditional is needed since the call to get_file() when providing a directory
        #  is not consistent when the path is http or NFS/CIFS
        if (-d $cdefs_dir) {
            # just use opendir and readdir to get the entries
            opendir(DIR, $cdefs_dir) || $Trace->fatal(msg=>"Could not open $cdefs_dir: $!", result=>"SCRIPT");
            foreach my $file (readdir(DIR)) {
                next if ($file !~ /.cdefs$/ && $file !~ /.\d+$/);
                my $base = $file;
                $base =~ s/\..+$//;
                $file = "$cdefs_dir/$file";
                push(@{$files{$base}}, $file);
            }
        } else {
            # see if the entries can be loaded with NATE calls
            my $dirs = NATE::Inc::get_file($cdefs_dir);
            if (! $dirs) {
                $Trace->fatal(msg=>"Could not open $cdefs_dir: $!", result=>"SCRIPT");
            }

            # For HTTP paths, the entries are returned in HTML so just parse it
            foreach my $dir_entries (split(/\n/, $dirs)) {
                if ($dir_entries =~ /a href=\"(.+\.cdefs)\"/) {
                    my $file = $1;
                    my $base = $file;
                    $base =~ s/\..+$//;
                    $file = "$cdefs_dir/$file";
                    push(@{$files{$base}}, $file);
                }
            }
        }

        if (! scalar keys %files) {
            $Trace->fatal(msg=>"Did not locate any cdefs files in the directory '$cdefs_dir'");
        }

	$Trace->comment(msg=>"Loading CDEFs from directory: '$cdefs_dir'");

        while (my ($base, $file_array) = each %files) {
            # sort the filenames so that the changelist numbers are in increasing order, with
            #  the .cdef last
            foreach my $file (sort _alpha_numerically @$file_array) {
                # if the file includes a changelist number then compare the changelist to the 
                #  currently running changelist. If the current build is newer than the file
                #  then skip this file. 
                if ($file =~ /$base\.(\d+)/) {
                    my $build_change = $1;
                    if ($self->compare_version(build=>$build_change)) {
                        $Trace->debug(msg=>"skip cdef '$file' for $base");
                        next;
                    }
                }

                # If this point is reached then this file contains the correct definitions. Read
                #  the file and bail out of this list. This could be the actual .cdef which is the
                #  the latest file or a changelist which is the latest
                our %CommandDefinitions = ();
                our %Compatability = ();
                our %Options = ();
                my $return;
                my $code = NATE::Inc::get_file($file);
                eval $code;
                if ($@) {
                    $Trace->fatal(result=>"CONFIG", msg=>"Could not eval $file: $@\n");
                }
		
		# Now merge any 'additional_options' if present.  These are fields that didn't make it
		#  into the Options hash on smfgen's first pass (picked when processing extrinsics)
		foreach my $cmd (keys %CommandDefinitions) {
		    if ( defined $CommandDefinitions{$cmd}{additional_options} ) {
			foreach my $opt (keys %{$CommandDefinitions{$cmd}{additional_options}}) {
			    $CommandDefinitions{$cmd}{options}{$opt} = $CommandDefinitions{$cmd}{additional_options}{$opt};
			}
		    }
		}

                # this code is the same as in _define_commands but cuts down on the amount of 'call' tracing
                grep($self->{Commands}->{$_} = $CommandDefinitions{$_}, keys %CommandDefinitions);
                grep($self->{Compatability}->{$_} = $Compatability{$_}, keys %Compatability);
                last;
            }
        }
    }

    # call _load_modules() which is defined in each of the node objects since the code is specific to each
    $self->_load_modules($class, @modules);

    $Trace->exit();
} 

# intentionally not using Trace because this code has been around for a long time with
# no problem and causes needless trace
sub _define_commands {
    my $self = shift;
    my %params = @_;
    foreach my $param (qw(Definitions Compatability)) {
        $Trace->fatal(msg=>"requires a $param parameter", result=>"SCRIPT") if (!$params{$param});
    }
    # these two lines are also in _setup_commands
    grep($self->{Commands}->{$_} = $params{Definitions}->{$_}, keys %{$params{Definitions}});
    grep($self->{Compatability}->{$_} = $params{Compatability}->{$_}, keys %{$params{Compatability}});
}

#***********************************************************************
#
# Method:      open_connection
#
# Description: Open a new connection to the server. This method
#              has to be called before any commands can be executed.
#
# Arguments:
#              type: cli_connection, shell_connection, rlm_connection, etc
#              conntype: the connection has to use this conntype if specified
#              abortive: the routine will die if if cannot create the connection
#
# Returns:     A reference to the connection
#              Will die on failure.
#
#***********************************************************************

sub open_connection {
    my $self = shift;
    my %params = (
		  type => undef,
		  conntype => undef,
		  abortive => undef,
		  @_, # overlay any parameters that were explicitly passed
		  );

    $Trace->entry(msg=>"entered for $self->{server}, type=>$params{type}, conntype=>$params{conntype}, abortive=>$params{abortive}");
    if (!$params{type}) {
	$Trace->fatal(msg=>"requires a type parameter", result=>"SCRIPT");
    }

    my $conn_object;

    my $connections = $self->{connections}; # just for convenience sake
    my $connect_str; # the string that will be passed to connect()

    # if a conntype has been passed in then compare its value to the conntype found in the hostrec and
    #  make sure to override the value from the hostrec with this one before calling connect()
    if ($params{conntype}) {
	my $conn_field = $params{type};
	$conn_field =~ s/_connection//; # make rlm_connection be rlm 
	my $conn;
	if (($conn = $self->{host_object}->{connection_fields}->{$conn_field}) &&
	    ($conn->{conntype} ne $params{conntype})) {
	    $Trace->trace(msg=>"changing the conntype for '$params{type}' from '$conn->{conntype}' to '$params{conntype}'");
	    $conn->{conntype} = $params{conntype};
	}
    }

    # use the connspec that was created by the caller
    if ($connections->{$params{type}}{conn_specs}) {
	$Trace->trace(msg=>"building the connspec for $params{type}");
	$connect_str = $connections->{$params{type}}{conn_specs};
    } else {
	$Trace->fatal(msg=>"could not determine a connspec for type '$params{type}'", result=>"SCRIPT");
    }
    $Trace->trace(msg=>"connect($connect_str)");

    # pass the connspec to connect. wrap this in eval to catch any errors. If this was an
    #  abortive call then provide a better error message than connect...
    eval {
	$conn_object = main::connect($connect_str);
	$connections->{$params{type}}{object} = $conn_object;
        my $type = $conn_object->conntype();

        # If the user has specified a connection buffer size, then 
        #   make the appropriate call on the connection object to change it
        #   to the new value.
        if (my $maxbufsize = $Options->get(option=>'CONNECTION_MAXBUFSIZE')) {
            $conn_object->maxbufsize($maxbufsize);
        }

        # If this is the rlm connection and it is a "persay" connection, then try
        #  to handle ssh key problems. This is not handled by NATE in persay connections
        #  like it is in persist. Find the current command that will be executed and
        #  add on additional options to the ssh command so that it will relax some of the
        #  security checks it is making.
        if ($type eq "ssh-persay" && $params{type} eq "rlm_connection") {
            # find the current command
            my @cmd = $conn_object->conncmd();

            # modify the command that will be run
            if (!grep(/CheckHostIP/, @cmd)) {
                push(@cmd, qw(-o CheckHostIP=no));
            }
            if (!grep(/StrictHostKeyChecking/, @cmd)) {
                push(@cmd, qw(-o StrictHostKeyChecking=no));
            }
            # update the connection with the new command we want to run
            $conn_object->conncmd(@cmd);
        }
       
        # saving default prompts
        if (!defined $connections->{$params{type}}{default_prompts}){
            push ( @{$connections->{$params{type}}{default_prompts}},  $conn_object->prompts() );
        }
    };
    if ($@) {
	if ($params{abortive}) {
	    $Trace->fatal(msg=>"Could not open a '$params{type}' connection: \n$@", result=>"CONFIG");
	} else {
	    $Trace->comment(msg=>"Could not open a '$params{type}' connection: $@");
	}
    }
    
  done: 
    $Trace->exit(msg=>"type = $conn_object");
    return $conn_object;
} # end open_connection

#***********************************************************************
#
# Method:      close_connection
#
# Description: Close a connection
#
# Arguments:   type: the type of connection
#              keep_state: setting this to 1 will cause the method to
#                 only delete the connection object but not delete any
#                 of the other state for this connection
#
# Returns:     Always returns 1 for success.
#
#***********************************************************************

sub close_connection {
    my $self = shift;
    my %params = (
		  type => undef,
                  keep_state => undef,
		  @_, # overlay any parameters that were explicitly passed
		  );
    $Trace->entry(msg=>"entered with type = $params{type}, server = $self->{server}");

    if (exists($self->{connections}{$params{type}}{conn_specs})) {
        $Trace->debug(msg=>"disconnect($self->{connections}{$params{type}}{conn_specs}, -ok)");
    	main::disconnect($self->{connections}{$params{type}}{conn_specs}, '-ok');
    }  

    $Trace->debug(msg=>"Deleting $params{type} key from object...");
    if ($params{keep_state}) {
        delete($self->{connections}{$params{type}}{object});
    } else {
        delete($self->{connections}{$params{type}});
    }
    $Trace->exit();
    return 1;
} # end close_connection

#***********************************************************************
#
# Method:      say_wrapper
#
# Description: Execute a given command on the server.
#
# Arguments:   cmd is the actual command to execute. Required.
#              output: a reference to a variable for returning output. Default is none.
#              timeout: how long to wait on output from Command. Default is 20 seconds. (Soon to be depraecated)
#              -Timeout: how long to wait on output from Command. Default is 20 seconds, 
#		but if the command completes before 2 * 20 then there will be no FAIL. (Preferred method)
#              -Errors: should errors be printed out
#	       nocr: If set, do not append a carriage return to end of command
#              poke: send the poke command after running the command, before looking for
#                    the prompt
#	       noprompt: Issue the command without waiting for a prompt
#              exit_code: a scalar reference to store the exit code for a shell command
#
# Returns:     the output on success, undef on failure.
#              The output of the command executed, stripped of any "\r" characters
#              is returned in output if it was defined.
#
#***********************************************************************

sub say_wrapper {
    my $self = shift;
    my %params = (cmd => "", # required
		  output => undef, # optional
		  timeout => undef,
		  -Timeout => undef,
		  -Errors => 1, # internal
		  OverrideServer => undef,
		  type => "shell_connection",
		  nocr => undef,
                  eval_error => undef,
		  poke => undef,
          ignore => undef,
		  exit_code => undef,
                  maxidle => undef,
		  noprompt => 0,
                  -Warn_timeout => undef,
		  -Swallow => 1,
		  -Noverbose => 0,
                  -Ignore_errors => 0,
                  conn => undef,
		  @_,
		  );
    my $output = $params{output};
    my $server = $self->{server};
    my $type = $params{type};
    my $conn = $self->{connections}{$type}{object};
    if ($params{conn}) { $conn = $params{conn} };
    my $exit_code = $params{exit_code};

    my $_begin_micro = 0;
    my $_end_micro   = 0;

    $Trace->entry(msg=>"entered for $server.");
    $Trace->debug(msg=>"***The command we are going to run is:\n$params{cmd}");
    #########################################################
    # The timeout option will default to 20 seconds.
    # If a value is passed to the method then the timeout will
    #  be set to it. 
    # 
    # If a command-line timeout is also provided, set $timeout to
    #  the max of the option and the method param.
    #
    # If WARN_TIMEOUT is true then $Timeout will be doubled and 
    # 2*$timout will be the new timeout.  1*$timeout will be 
    # $warntimeout.
    #########################################################
    if ($params{timeout} && $params{cmd} !~ /^storage failover show/ ) {
        $Trace->warn(msg=>"The timeout parameter to say_wrapper will be deprecated.\nPlease use -Timeout instead");
        $params{-Timeout} = $params{timeout};
    }

    my $timeout = $params{-Timeout};
    my $timeout_option = $self->{options_object}->get(option=>"timeout");
    if (defined($timeout_option)) {
	if ($timeout_option > $timeout) {
	    $timeout = $timeout_option;
	}
    } elsif (!defined($timeout)) {
	$timeout = 20;
    }

    # Multiply the timeout before we setup the WARN_TIMEOUT so that we still get the 'desired' WARN/FAIL path
    my $timeout_multiplier = $self->{options_object}->get(option=>"TIMEOUT_MULTIPLIER");
    $timeout *= $timeout_multiplier;  # This should always work since the default multiplier is 1

    if (defined($self->{options_object}->get(option=>"WARN_TIMEOUT"))) {
    	$params{'-Warn_timeout'} = $self->{options_object}->get(option=>"WARN_TIMEOUT");
    }

    my $ignoreErrors = $params{-Ignore_errors};
    
    my $warntimeout;
    if ($params{'-Warn_timeout'}) {
    	$warntimeout = $timeout;
    	$timeout = ($timeout * 2);
    }
    if (defined($params{additional_prompts_timeout})) {
    	$warntimeout = $params{additional_prompts_timeout};
    	$timeout = ($params{additional_prompts_timeout} * 2);
    }
    # same behavior for the maxidle param
    my $warnmaxidle = undef;
    my $maxidle = $params{maxidle};
    my $maxidle_option = $self->{options_object}->get(option=>"maxidle");
    if (defined($maxidle_option)) {
	if ($maxidle_option > $maxidle) {
	    $maxidle = $maxidle_option;
	}
    }
    if ((defined($maxidle)) && ($params{'-Warn_timeout'})) {
    	$warnmaxidle = $maxidle;
    	$maxidle = ($maxidle * 2);
    }
    
    $Trace->trace(msg=>"Say timeout params:: warntimeout: $warntimeout, timeout: $timeout, warnmaxidle: $warnmaxidle, maxidle: $maxidle");
    
    if ($exit_code) {
	if (ref($exit_code) ne "SCALAR") {
	    $Trace->fatal(msg=>"exit_code is not a ref:$exit_code", result=>"SCRIPT");
	} elsif ($type ne "shell_connection") {
	    $Trace->fatal(msg=>"exit_code is only used with a 'shell' connection: $type", result=>"SCRIPT");
	}
    }
    if (defined($output)) {
	if (!ref($output)) {
	    $Trace->fatal(msg=>"requires 'output' parameter be a scalar reference: $output", result=>"SCRIPT");
	} else {
	    # if an output reference is passed in then set it to undef to clear any previous content.
	    #  this will allow the caller to reuse a variable.
	    $$output = undef;
	}
    }	
    if ( $params{cmd} eq "" ) {
	$Trace->fatal(msg=>"The 'cmd' parameter is required", result=>"SCRIPT");
    } else {
	$Trace->trace(msg=>"The 'cmd' on $server is '$params{cmd}'");
    }

    if ( defined($params{ignore}) ) {
        if ( ref($params{ignore}) ne 'ARRAY' ) {
            $Trace->fatal(msg=>"requires 'ignore' parameter be a array reference: " .
                "$params{ignore}", result=>"SCRIPT");
        }
    }
    
    my %profile;
    my $prof_file = $Options->get(option=>'CMDPROF');
    my $prof_file_csv = $Options->get(option=>'CMDPROF_CSV');
    if (defined($prof_file) || defined($prof_file_csv)  || $Options->get(option=>'CMDPROF_RT') ) {
        my @command_parts = split(/\s+/, $params{cmd});
        my @command_name;
        my @command_args;
        my $j;
        while (defined($command_parts[$j])) {
            if ($command_parts[$j] =~ m/^-/) {
                push(@command_args, $command_parts[$j++],  $command_parts[$j]);
            } else {
                push(@command_name, $command_parts[$j]);
            }
            $j++;
        }

       $profile{'args'}   = join(" ", @command_args);
       $profile{'cmd'}    = join(" ", @command_name);
       $profile{'server'} = $self->{'server'};
    }

    # if this connection type does not exist, then try to establish it now
    if (!defined($conn)) {
	$self->establish_connections($type=>1);
	$conn = $self->{connections}{$type}{object};
    }

    my $cmd = $params{cmd};
    my $cmd_exp;
    if ($self->{connections}{$type}{cli_from_shell}) {
        if ($self->{connections}{$type}{cli_from_shell_preamble}) {
            $cmd = "$self->{connections}{$type}{cli_from_shell_preamble}; $cmd";
        } elsif ($self->{connections}{$type}{node_run}) {
            # this is the new Node Scope in BR and beyond
            $cmd = "run -node $self->{server} -command \"$priv_string $cmd\"";
        }
	$cmd = "$self->{connections}{$type}{cli_from_shell} '$cmd'";
        $cmd_exp = $cmd;
    } else {
        $cmd_exp = quotemeta $cmd;
    }
    my ($match, $before, $after, $prompt, $retry) = ();

    my @say_params = ($conn, $cmd);
    if (($self->{options_object}->get(option=>"SAY_SWALLOW") && $params{-Swallow}) ||
        $type eq "console_connection") {
        push(@say_params, -swallow);
    }
    push(@say_params, -noprompt) if ($params{noprompt});
    push(@say_params, -nocr) if ($params{nocr});
    push(@say_params, -eval_error) if ($params{eval_error});
    push(@say_params, -timeout, $timeout) if (defined($timeout));
    push(@say_params, -maxidle, $maxidle) if (defined($maxidle));
    push(@say_params, -noverbose) if ($params{-Noverbose});
    $Trace->debug(msg=>"\@say_params:\n". Dumper(@say_params));
    ($profile{'begin'}, $_begin_micro) = gettimeofday;
    $profile{'begin_micro'} = $profile{'begin'} + ($_begin_micro * .000001);
    my $path;
    if ($params{additional_prompts}){
        if ((my $reftype = ref($params{additional_prompts})) eq "HASH"){
            # Not sure if this is always going to work, but it does in the single case that is using this feature so far
            $Trace->debug(msg=>"The \%params hash contains:\n". Dumper(%params));
            my @wait_prompts = keys %{$params{additional_prompts}};
            my @fatal_prompts;
            my @final_prompts;
            foreach my $prompt (keys %{$params{additional_prompts}}) {
                if ($params{additional_prompts}->{$prompt} eq 'FATAL_PROMPT') {
                    push @fatal_prompts , $prompt;
                } elsif ($params{additional_prompts}->{$prompt} eq 'FINAL_PROMPT') {
                    push @final_prompts , $prompt;
                }
            }
            $Trace->debug(msg=>"\@fatal_prompts:\n". Dumper(@fatal_prompts));
            $Trace->debug(msg=>"\@final_prompts:\n". Dumper(@final_prompts));
            $Trace->trace(msg=>"****\nThis multi-prompt command needs to complete within $params{additional_prompts_timeout} seconds.\n****");
            my $timeleft;
            if ($params{additional_prompts_timeout}) {
                $timeleft = $params{additional_prompts_timeout};
            } else {
                $timeleft = $timeout;
            }
            my $begtime = time();
            $Trace->debug(msg=>"****\nBEGTIME: $begtime\n****");
            $Trace->debug(msg=>"****\nTIMELEFT: $timeleft\n****");
            my $curtime;
            
            dowait:
            {
                $Trace->trace(msg=>"There are $timeleft seconds to finish this multi-prompt command.");
                $Trace->debug(msg=>"waiting for ". Dumper(\@wait_prompts));
                $Trace->debug(msg=>"fatal prompts ". Dumper(\@fatal_prompts));
                $Trace->debug(msg=>"final prompts ". Dumper(\@final_prompts));
                my @prompt_args;
                @prompt_args = (-prompt => \@wait_prompts) if @wait_prompts;
                push( @prompt_args, "-ignore", @{$params{ignore}} ) if ($params{ignore});
                $Trace->debug(msg=>"\@prompt_args:\n". Dumper(@prompt_args));
                ($match, $before, $after, $prompt, $retry) = $self->_say($params{type}, @say_params, @prompt_args);
                if ($self->{connections}{$type}{errors} && grep($after =~ /$_/ , @{$self->{connections}{$type}{errors}})) {
                    $Trace->fatal(msg=>"$params{callingfunc} failed: after = $after");
                }
                if (grep({$prompt =~ /$_$/} @fatal_prompts)) {
                    $Trace->fatal(msg=>"$params{callingfunc} failed because: $prompt");
                } elsif (grep({$prompt =~ /$_$/} @final_prompts)) {
                    $Trace->trace(msg=>"****\nFinished multi-prompt command with prompt:\n$prompt\n****");
                    last;
                }
                my @regex;
                my $found;
                my $response;                
                if (@regex = grep({$prompt =~ /^$_$/} @wait_prompts)) {
                    $Trace->debug(msg=>"after say before we remove the last seen prompt ". Dumper(\@wait_prompts));
                    unless ($params{repeat}) {
                        @wait_prompts = grep({!($prompt =~ /^$_$/)} @wait_prompts);
                    }
                    $Trace->debug(msg=>"after we remove the last seen prompt ". Dumper(\@wait_prompts));
                    $found = $regex[0];
                    $response = $params{additional_prompts}->{$found};
                    if ($found ne $prompt) {
                        $Trace->trace(msg=>"Found prompt:'$prompt' with regex:'$found', will respond with:'$response'.");
                    } else {
                        $Trace->trace(msg=>"Found prompt:'$prompt', will respond with:'$response'.");
                    }
                    @say_params = ($conn, $response);
                    $curtime = time();
                    $timeleft = $params{additional_prompts_timeout} - ($curtime - $begtime);
                    $Trace->debug(msg=>"****\nCURTIME: $curtime\nTIMELEFT: $timeleft\n****");                  
                    if ($timeleft <= 0) { 
                        $Trace->fatal(msg=>"$params{callingfunc} failed because we ran out of time!");
                    }
                    if (($self->{options_object}->get(option=>"SAY_SWALLOW") && $params{-Swallow}) ||
                        $type eq "console_connection") {
                        push(@say_params, -swallow);
                    }
                    push(@say_params, -nocr) if ($params{nocr});
                    push(@say_params, -timeout, $timeleft) if (defined($timeleft));
                    push(@say_params, -maxidle, $maxidle) if (defined($maxidle));
                    $Trace->debug(msg=>"\@say_params:\n". Dumper(@say_params));                   
                    redo dowait;
                } else {
                    $Trace->comment(msg=>"We should have never arrived here!");
                    $Trace->comment(msg=>"NATE probably matched a string to a prompt\n" .
                        "try setting 'ignore' option");

                    $Trace->fatal(msg=>"MATCH: $match\n".
                                       "BEFORE: $before\n".
                                       "AFTER: $after\n".
                                       "PROMPT: $prompt\n", result => "FATAL");
                }
            }
            while (my ($key, $value) = each(%{$params{additional_prompts}})){
                if ((!defined($value) || !length($value)) && grep(/^$key$/, @wait_prompts)){
                    $Trace->fatal(msg=>"Expected to see string '$key', but did not");
                }
            }
        } elsif ($reftype eq "ARRAY"){
            foreach my $addprompt (@{$params{additional_prompts}}){
                $Trace->debug(msg=>"waiting for: ". Dumper($addprompt));
                $Trace->debug(msg=>"Say params are: ". Dumper(\@say_params));
                ($match, $before, $after, $prompt, $retry) = $self->_say($params{type}, @say_params, -prompt, $addprompt->{prompt});
                if ($self->{connections}{$type}{errors} && grep($after =~ /$_/ , @{$self->{connections}{$type}{errors}})) {
                    $Trace->fatal(msg=>"$params{callingfunc} failed: after = $after");
                }
                my $command = "";
                if (defined($addprompt->{response}) && length($addprompt->{response})){
                    $command = $addprompt->{response};
                }
                @say_params = ($conn, $command);
                if (($self->{options_object}->get(option=>"SAY_SWALLOW") && $params{-Swallow}) ||
                    $type eq "console_connection") {
                    push(@say_params, -swallow);
                }
                push(@say_params, -nocr) if ($params{nocr});
                if (defined($addprompt->{timeout})){
                    my $timeout = $addprompt->{timeout};
                    if ($params{'-Warn_timeout'}) {
			$warntimeout = $timeout;
			$timeout = ($timeout * 2);
		    }
                    push(@say_params, -timeout, $timeout);
                } elsif (defined($timeout)){
                    push(@say_params, -timeout, $timeout);
                }
                push(@say_params, -maxidle, $maxidle) if (defined($maxidle));
            }
            # Once we're done here, we may need to wait for the real prompt, but we'll check that out
        }
        $path = "Additional prompts";
    } elsif ($params{poke}) {
        $self->_say($params{type}, @say_params, -blind);
        ($match, $before, $after, $prompt, $retry) = $self->_say($params{type}, $conn, $params{poke}, -timeout, $timeout);
        $path = "Poke";
    } else {
        ($match, $before, $after, $prompt, $retry) = $self->_say($params{type}, @say_params);
        $path = "Normal";
    }

    $profile{command_retried} = $retry;
    ($profile{'end'}, $_end_micro) = gettimeofday;
    $profile{'end_micro'} = $profile{'end'} + ($_end_micro * .000001);
    
    my $cmdtime = ($profile{'end'} - $profile{'begin'});
    if (($params{'-Warn_timeout'}) && ($cmdtime > $warntimeout)) {
        $Trace->warn(msg=>"Command [$cmd] took longer then $warntimeout seconds; it actually took $cmdtime seconds to complete. Timeout was $timeout.");
    } else {
        $Trace->trace(msg=>"Command [$cmd] took $cmdtime seconds to complete; warntimeout was $warntimeout, timeout was $timeout.");
    }
    # dump profile hash block to text file
    if (defined($prof_file)) {
        $profile{'cmdtime'} = $cmdtime;   # since we already calculated it
        $Trace->debug(msg => "Writing profiling data:\n".Dumper(\%profile)."\n to file ".$prof_file);
        open(PROFDATA, ">>".$prof_file);
        print PROFDATA Dumper(\%profile);
        print PROFDATA "xxEND_RECORDxx\n";
        close(PROFDATA);
    }
    # dump profile to csv file
    # the 'args' value may have comma's in it, so it is last
    if (defined($prof_file_csv)) {
       $profile{'cmdtime'} = $cmdtime;   # since we already calculated it
       $Trace->debug(msg => "Writing profiling data to csv file $prof_file_csv\n");
       open(PROFDATACSV, ">>".$prof_file_csv);
       printf PROFDATACSV ("%s,%s,%s,%s,%s,%s,%s,%s,%s\n",
          $profile{'server'},
          $profile{'begin'},       $profile{'end'},       $profile{'cmdtime'},
          $profile{'begin_micro'}, $profile{'end_micro'},
          $profile{command_retried},
          $profile{'cmd'},         $profile{'args'},
       );
       close(PROFDATACSV);
    }
    # store info in the server object that ran this command 
    if ( $Options->get(option=>'CMDPROF_RT') ) {
       # put current profile hash on list   $server->{'cmdprof'}[0]->{key} = value
       unshift(@{$self->{'server_object'}{'cmdprof'}}, {%profile});

       # keep list desired size
       if ( scalar(@{$self->{'server_object'}{'cmdprof'}}) > $Options->get(option=>'CMDPROF_RT') )  {
          pop(@{$self->{'server_object'}{'cmdprof'}});
          $Trace->debug('msg' => "CMDPROF_RT popped list\n");
       }
       $Trace->debug('msg' => sprintf("CMDPROF_RT size %u: max %u",
                                      scalar(@{$self->{'server_object'}{'cmdprof'}}),
                                      $Options->get(option=>'CMDPROF_RT'),
                              ),
       );
    }

    $Trace->debug(msg=>"match='$match'\nbefore='$before'\nafter='$after'\nprompt='$prompt' via method '$path'");

    if ($after) {
	$cmd_exp = quotemeta $cmd;
	
	# The NGSH wraps inside the command echo. Luckily, this wrap is a little different so is easy
	#  to remove. The normal wrap is a '\r\n\r'. Replace this combination with a simple \n.
	#  The strange wrap is a '\n\r'. Just remove it so the command echo is easy to remove.
	#  We still need to remove any single '\r' characters as well
	$after =~ s/\r\n\r/\n/g; # weird wrapping that some CLIs' perform
	$after =~ s/\n\r\n/\n/g; # weird wrapping that some CLIs' perform
	$after =~ s/\n\r/\n/g; # weird wrapping that some CLIs' perform
        $after =~ s/\s+\cH//g; # weird wrapping that some CLIs' perform
	$after =~ s/^(.*)$cmd_exp\s*(\n|\z)//m;
	if ($self->{connections}{$type}{cli_from_shell}) {
	    # in this mode, the "shell -c cmd" has been stripped but the output usually also contains just msg
	    $cmd_exp = quotemeta($params{cmd});
	    $after =~ s/^(.*)$cmd_exp\s*(\n|\z)//m;
	}
	$after =~ s/\r//g;
	
	$after =~ s/\s*$//; # chop trailing whitespace
	
	# Capture the output from the command, and return it in a string.
	$$output .= $after if ($output);

	# first, see if the output contains any of the acceptable strings. If it does not, then fail if it has any of the error strings
	if ($self->{connections}{$type}{acceptable} && grep($after =~ /$_/s, @{$self->{connections}{$type}{acceptable}})) {
	    $Trace->debug(msg=>"The output was found in the acceptable errors list");
	} elsif ($params{-Idempotent} && ($self->{connections}{$type}{idempotent} && grep($after =~ /$_/s , @{$self->{connections}{$type}{idempotent}}))) {
	    $Trace->debug(msg=>"The output was found in the idempotent errors list");
	} elsif ($params{-IgnoreTransient} && ($self->{connections}{$type}{transient} && grep($after =~ /$_/s , @{$self->{connections}{$type}{transient}}))) {
	    $Trace->debug(msg=>"The output was found in the transient errors list.  Calling myself again.");
            $self->say_wrapper(%params);
	} elsif ($self->{connections}{$type}{errors} && grep($after =~ /$_/ , @{$self->{connections}{$type}{errors}}) && $ignoreErrors == 0) {
	    $Trace->fatal(msg=>"$params{callingfunc} failed: \nCommand='$params{cmd}'\nResponse='$after'");
	}
    } elsif (!defined($match) && !$params{noprompt}) {
        # some CLIs can return nothing
        $Trace->fatal(msg=>"encountered an error with the command \"$params{cmd}\"");
    }

    if ($exit_code) {
	# ask for the exit code and then strip out everything except the actual code
	my $cmd = "echo \$?";
	($match, $before, $after, $prompt) = $self->_say($params{type}, $conn, $cmd, -timeout, 30);
	$cmd = quotemeta $cmd;
	$after =~ s/$cmd//;
	$after =~ s/\s+//g;
	if ($after !~ /^\d+$/) {
	    $Trace->fatal(msg=>"The exit code for '$params{cmd}' is not a number: '$after'", result=>"FATAL");
	}
	$$exit_code = $after;
    }

  done:
    $Trace->exit(msg=>"return '$after'");
    return $after;
} # end say_wrapper

#***********************************************************************

sub server_type {
    my ($self) = @_;
    $Trace->entry();
    $Trace->exit(msg=>"returning $self->{type}");
    return $self->{type};
}

#***********************************************************************
#
# This method is used to automatically load new methods which are not defined
#   in the namespace. It basically just takes the name of the method being
#   called and passes it into run_command as a special paramater. It has to
#   be called via an anonymous function to work correctly.
#
#***********************************************************************

sub AUTOLOAD {
    my $method = our $AUTOLOAD;
    $Trace->entry();
    return if ($method =~ /::DESTROY/);
    
    no strict 'refs';
    $Trace->trace(msg=>"Node::AUTOLOAD looking for $method");
    push(@_, AUTOLOAD_COMMAND=>$method);
    *$AUTOLOAD = sub { run_command(@_, -Autoloaded => 1); };
    goto &$AUTOLOAD;
    $Trace->exit();
} # end AUTOLOAD

sub _deep_copy {
    my $this = shift;
    if (not ref $this) {
	$this;
    } elsif (ref $this eq "ARRAY") {
	[map _deep_copy($_), @$this];
    } elsif (ref $this eq "HASH") {
	{map { $_ => _deep_copy($this->{$_}) } keys %$this};
    } else { die "what type is $_?" }
}

#***********************************************************************
#
# This method is only called from AUTOLOAD to handle methods which are
#    not defined. It will look up the method name in the commands 
#    structure for this filer ($self->{server}), verify the paramaters,
#    generate the command, execute said command and then verify the
#    output if necessary.
#
#***********************************************************************
sub run_command {
    my $self = shift;
    my $rc = undef;
    my %params = (
		  -Verification => 1, #intentionally CAPS
		  -Errors => 1,
		  -SkipINFO => undef,
		  -Autoloaded => 0,
                  -Help => 0,
		  @_,
		  );
    my @rc = ();
    my %command_params;
    $Trace->entry(msg=>"params = " . join(" ", @_));

    no strict 'refs';
    my $MethodName = $params{AUTOLOAD_COMMAND}; # internal parameter only
    my ($Command, $NameSpace);
    my $outputRef = $params{output};

    my $output;

    if ($outputRef && !ref($outputRef)) {
	$Trace->fatal(msg=>"'output' param is not a ref:$outputRef", result=>"SCRIPT");
    }

    # separate the Method into its Namespace and command parts (i.e. Node:: and ArpServer)
    if (!defined($MethodName)) {
	print Dumper \%params;
	$Trace->fatal(msg=>"does not have a MethodName defined", result=>"SCRIPT");
    } elsif ($MethodName eq "") {
	print Dumper \%params;
	$Trace->fatal(msg=>"has an empty MethodName defined", result=>"SCRIPT");
    } elsif ($MethodName =~ /^(.+\:\:)(.+)$/) {
	$NameSpace = $1;
	$Command = $2;
    } else {
	$Trace->fatal(msg=>"do not know how to parse MethodName: $MethodName", result=>"SCRIPT");
    }	

    # remove our name from the namespace to force the call back through the autoloader
    # this has to occur very early before a die is called because it might get caught and this entry would still remain in the namespace
    if ($params{-Autoloaded} && exists($NameSpace->{$Command})) {
	*{$NameSpace->{$Command}} = undef;
	delete $NameSpace->{$Command};
    }

    # if no_cli=0 establish cli connections up front and load appropriate cdefs in NG.pm
    # if no_cli=1 do NOT establish cli connections up front, but load non-gen cdefs in NG.pm
    # if no_cli exists but is not '1' then establish cli connections on the fly
    # and load the correct gen'ed cdefs for the build running on those nodes
    $Trace->trace(msg=>"entered for $MethodName"); 
    my %Commands = %{$self->{Commands}} if ($self->{Commands});
    if ((!scalar(keys %Commands))&&($self->{no_cli})&&($self->{no_cli} != 1)) {
        $self->establish_connections(cli_connection => 1);
        %Commands = %{$self->{Commands}} if ($self->{Commands});
    }
    if (!scalar(keys %Commands)) {
        $Trace->fatal(msg=>"$MethodName could not find the command hash for $self->{server}", result=>"SCRIPT");
    }
    
    # Handle the old-style Compatability map (still used by Classic
    if ( defined $self->{Compatability} ) {
	$Command = $self->{Compatability}{$Command} if (defined $self->{Compatability}{$Command});
    }
    
    # now look up the Command in the Compatability map in case it is a deprecated method name
    #   the old name has to be deleted from the namespace so that it will go through the autoloader
    #   again when called
    # Only look for an exact match in the global Compatability map.  Delay translating partial map entries
    # until we know we don't have a true match (saves looping time).
    my $oldCommand;
    if ( defined $self->{compatability} ) {
	if ( defined $self->{compatability}{$Command}) {
	    $oldCommand = $Command;
	    if ( defined $self->{compatability}{$Command}{new_command} ) {
		$Command = $self->{compatability}{$Command}{new_command};
	    }
	    if (!$Command || $Command eq "") {
		$Trace->fatal(msg=>"could not figure out the name of the compatable command for $oldCommand", 
			      result=>"SCRIPT");
	    }
	    $Trace->debug(msg=>"Translated old command: $oldCommand, to new command: " .
			  $Command);
	    $MethodName = "${NameSpace}::$Command";
	    $params{AUTOLOAD_COMMAND} = $MethodName;
	} 
    }
    # Check to see if there's a partial entry in the Compatability hash that can be used to convert 
    # this command's command string.  If $Command and $oldCommand are the same, we found a match in the compat hash.
    if ( ($Command ne $oldCommand) && defined $self->{compatability} ) {
	foreach my $key ( @{$self->{compat_ordered}} ) {
	    $Trace->debug(msg=>"Looking for $Command in  $key");
	    if ( $Command =~ /^$key/ || $key =~ /_ALL_COMMANDS/ ) {
		if ( defined $self->{compatability}{$key}{new_command} ) {
                    # define a temp variable and make the conversion on it
                    my $temp_command = $Command;
		    $temp_command =~ s/$key/$self->{compatability}{$key}{new_command}/;
                    # now check to see if this new version of the command is valid
                    if ( defined $Commands{$temp_command} ) {
                        $Command = $temp_command;
                        $oldCommand = $key;
                        $Trace->trace(msg=>"Compatability match on: $key  Command is now: $Command");
                        last;
                    }
		}
	    }
	}
    }	
    if (!$Commands{$Command}) {
	$Trace->fatal(msg=>"could not find the command definition for $Command on $self->{server}", result=>"SCRIPT");
    }

    my %definition = $self->_convert_definition($Commands{$Command});
    my $internal_eval = $definition{internal_eval};
    $Trace->debug(msg=>"***Dumper of '$Command'\n".Dumper($Commands{$Command}));
    $Trace->debug(msg=>"***Dumper of \%definition\n".Dumper(\%definition));

    if (!$definition{command}) {
	$Trace->fatal(msg=>"could not find a command for $Command", result=>"SCRIPT");
    }

    # figure out what the command line option is for each parameter
    my %cmdOptions = %{$definition{options}};
    %cmdOptions = (%cmdOptions, %{$definition{ReplacementOptions}}) if ($definition{ReplacementOptions});

    # if defaults exist, merge the defaults with the passed in params
    if ($definition{defaults}) {
	%command_params = (%{$definition{defaults}}, %params);
    } else { # just use the passed in values
        %command_params = %params;
    }

    # Build a list of all the conversions, so a results hash can contain all compat fields too
    my %all_conversions;
    # For specific commands, convert the parameter names from deprecated names to valid names
    # Conversion definitions exist in the static command definition on non-GX definitions.
    if ( $definition{conversion} ) {
	%command_params = $self->_convert_params($definition{conversion}, %command_params); 
    }

    # GX conversion defs are defined in the Node/NG/Compatability.pm file
    if ( defined $self->{compatability} && defined $self->{compatability}{$oldCommand} && 
	 defined $self->{compatability}{$oldCommand}{conversion} && $oldCommand ne "_ALL_COMMANDS") {
	$Trace->debug(msg=>"Substituting deprecated fields for: $oldCommand");
	%command_params = $self->_convert_params($self->{compatability}{$oldCommand}{conversion}, %command_params);
	%all_conversions = %{$self->{compat_mirror}{$oldCommand}{conversion}};
    } 

    # For ANY command, convert the parameter names from deprecated names to valid names
    if ( defined $self->{compatability}{_ALL_COMMANDS}{conversion} ) {
	$Trace->debug(msg=>"Generically substituting deprecated fields for: _ALL_COMMANDS");
	%command_params = $self->_convert_params($self->{compatability}{_ALL_COMMANDS}{conversion}, %command_params);
	%all_conversions = (%all_conversions, %{$self->{compat_mirror}{_ALL_COMMANDS}{conversion}});
    }
#    $Trace->debug(msg=>"***1Dumper of \%command_params\n". Dumper(\%command_params));
#    $Trace->debug(msg=>"***1Dumper of \%all_conversions\n". Dumper(\%all_conversions));
#    $Trace->debug(msg=>"***1Dumper of \%Commands\n". Dumper(\%Commands));

    # setup the command we will be calling
    my $command = $definition{command};
    $Trace->debug(msg=>"***1\$command is:\n$command");

    if (!$command) {
	$Trace->fatal(msg=>"could not find command information", result=>"SCRIPT");
    }

    # call the setup method if it is defined for this routine
    #  InternalSetup is a special flag in case the same routine is used for setup and verification
    my $setup = $definition{Setup};
    if ($setup) {    
	%command_params = &$setup($self, %command_params, InternalSetup=>1, Output=>$output);
	goto done if (%command_params == ());
	if ($command_params{InternalSetup} == SUCCESS_EXISTS()) {
	    goto done;
	}
	delete $command_params{InternalSetup};
    }

    my %input_params;
    my $cli_validation = $self->{options_object}->get(option=>"CLI_PARAMETER_VALIDATION") || $self->{cli_parameter_validation};
    if ($cli_validation && !(defined($definition{no_validation} && $definition{no_validation}))) {
	$command .= $self->_validate_cli_parameters(definition=>\%definition,
						    cmd_options=>\%cmdOptions,
						    cmd_params=>\%command_params,
						    input_params=>\%input_params,
						    );
	$Trace->debug(msg=>"Back from _validate:\n" . Dumper( \%definition ) .
		      Dumper( \%cmdOptions ) . Dumper( \%command_params ) . Dumper( \%input_params ));
    } else {
	my @add_params;
	if (exists $definition{additional_prompts}){
	    if ((my $reftype = ref($definition{additional_prompts})) eq "HASH"){
		@add_params = grep(/^-/, values %{$definition{additional_prompts}});
		
	    } elsif ($reftype eq "ARRAY"){
		foreach my $ref (@{$definition{additional_prompts}}){
		    push(@add_params, $ref->{response}) if ($ref->{response} =~ /^-/);
		}
	    }
	}
	$Trace->debug(msg=>"Additional fields: ". Dumper(\@add_params));
        $Trace->debug(msg=>"***1\$command is:\n$command");
 
        while (my ($param, $value) = each %command_params) {
            # special-case the 'Help' and pass the '?' to the command
            if ($param eq "-Help" && $value) {
                $command .= " ?";
                next;
            }
            if ($self->_valid_param(cdef=>\%definition, param=>$param, additional=>\@add_params)) {
                $Trace->debug(msg=>"***$param : $value");
                $input_params{$param} = $value;
                ### some cdefs may not have an %Options hash defined, so
                ### we just use the parameter name ($param) in that case.
                my $parameter;
                if ($cmdOptions{$param}){
                    $parameter = $cmdOptions{$param};
                } else {
                    $parameter = $param;
                }
                $Trace->debug(msg=>"***$param : $parameter : $cmdOptions{$param}");
                if (defined($value)) {
                    # if the parameter is passed in as an array ref, then break it up and send it as a list to the CLI
                    if (ref($value) eq "ARRAY") {
                        $command .= " $parameter \"".join(" ", @$value)."\"";
                    } else {
                        if ($Commands{$Command}{param_value_seperator}) {
                            $command .= " $parameter$Commands{$Command}{param_value_seperator}$value";
                        } else {
                            $command .= " $parameter $value";
                        }
                    }
                } else {
                    $command .= " $parameter";
                }
            }
        }
    }
#    $Trace->debug(msg=>"***2Dumper of \%command_params\n". Dumper(\%command_params));
#    $Trace->debug(msg=>"***2Dumper of \%cmdOptions\n". Dumper(\%cmdOptions));
#    $Trace->debug(msg=>"***2Dumper of \%all_conversions\n". Dumper(\%all_conversions));
    $Trace->debug(msg=>"***2\$command is:\n$command");

    # figure out which connection to use for the command. The default is the cli.
    my $conntype = "cli_connection";
    $conntype = ($definition{connection_type}) if ($definition{connection_type});
    if (!grep($conntype, keys %{$self->{connections}})) {
	$Trace->fatal(result=>"SCRIPT", "'$conntype' is not a supported connection type\n");

    }

    my @additional_say_args;
    if (!$params{-Help} && exists $definition{additional_prompts}){
	# Need to do more work since apparently the command that we're calling is going to present us with prompts
	if ((my $reftype = ref($definition{additional_prompts})) eq "HASH"){
	    $Trace->debug(msg=>"\%definitions: ".Dumper(%definition));
	    my %prompts = %{dclone($definition{additional_prompts})};

	    $Trace->debug(msg=>"Additional prompt definition: ".Dumper(\%prompts));
	    while (my ($key, $value) = each %prompts){
		if ($value =~ /^(-.*)/){
		    # Need to pull the value from the parameters passed into the command otherwise the value is a literal
		    $value = $1;
		    $prompts{$key} = $params{$value};
		}
	    }
	    push(@additional_say_args, additional_prompts => \%prompts);
	    push(@additional_say_args, additional_prompts_timeout => $definition{additional_prompts_timeout});
	} elsif ($reftype eq "ARRAY"){
	    # Need to copy since this is an array of hash refs
	    my @prompts = @{dclone($definition{additional_prompts})};

	    $Trace->debug(msg=>"Additional prompt definition: ".Dumper(\@prompts));
	    foreach my $prompt (@prompts){
		if ($prompt->{response} =~ /^(-.*)/){
		    # Need to pull the value from the parameters passed into the command otherwise the value is a literal
		    my $value = $1;
		    $prompt->{response} = $command_params{$value};
		}
	    }
	    push(@additional_say_args, additional_prompts => \@prompts);
	}
	$Trace->debug(msg=>"Additional prompt values: ".Dumper(\@additional_say_args));
    }

    # execute the command, retry if verification tells us to try again
    my $retries = 3;
    my $retry = 1;
  retry:

    eval {
	if ($params{timeout} && $command !~ /^storage failover show/ ) { 
            $Trace->warn(msg=>"The timeout parameter to say_wrapper will be deprecated.\nPlease use -Timeout instead");
            $output = $self->say_wrapper(type        => $conntype, 
                                         cmd         => $command, 
                                         output      => $outputRef, 
                                         -Timeout    => $params{timeout},
	 		                 nocr        => ($params{-Help}) ? 1 : $params{nocr}, 
	 		                 noprompt    => $params{noprompt}, 
	 		                 poke        => $params{poke}, 
	 		                 exit_code   => $params{exit_code},
					 -Errors     => $params{-Errors}, 
					 -Idempotent => $params{-Idempotent}, 
					 -Noverbose  => $params{-Noverbose},
					 -Ignore_errors => $params{-Ignore_errors},
				         callingfunc => $MethodName, 
				         @additional_say_args
				        );
	} else {
	    $output = $self->say_wrapper(type        => $conntype, 
	                                 cmd         => $command, 
	                                 output      => $outputRef, 
	                                 -Timeout    => $params{-Timeout},
	 		                 noprompt    => $params{noprompt}, 
	 		                 nocr        => ($params{-Help}) ? 1 : $params{nocr}, 
                                         poke        => $params{poke}, 
                                         exit_code   => $params{exit_code},
                                         -Errors     => $params{-Errors},
                                         -Idempotent => $params{-Idempotent}, 
					 -Noverbose  => $params{-Noverbose},
					 -Ignore_errors => $params{-Ignore_errors},
                                         callingfunc => $MethodName, 
                                         @additional_say_args
                                    );
	}
    };
    if ($params{-Help} && $self->server_type(type=>$self->ONTAP_NG()) &&
        (!$@ || ($@ && $@ =~ /^(ERROR|Error)/m))) {
        # Since -Help was passed in, the command was sent with the '? character
        # which will leave the command at the prompt. This needs to be cleared.
        # This code should really be in Node::NG
        eval {
            $self->_say($conntype, $self->{connections}{$conntype}{object}, "", -nocr);
        }
    }        
    if ($internal_eval && $@) {
#        if (defined($@) && ($@ !~ m/There are no entries matching your query./)) {
#            print "***$@***\n";
#            die($@);
#        }
        $rc = undef;
	goto done;
    } elsif ($@) {
	die $@;
    }
    
    $params{Caller} = $MethodName if (!$params{Caller});
    ###########################################################
    # Check to see if we have hit one of the FailureOutput 
    # text messages 
    ###########################################################
    if ($definition{FailureOutput}) {
	foreach my $message (@{$definition{FailureOutput}}) {
	    if ($output =~ /$message/) {
		$Trace->debug(msg=>"Failure Message Seen: $output");
		$@ = $output;
		$rc = undef;
		goto done;
	    }
	}
    }
    ###########################################################
    # We did not hit a failure condition
    ###########################################################
    if ($output =~ /usage\:/s) {
	$Trace->fatal(msg=>"Invalid value specified for the command:$output", result=>"SCRIPT");
    } elsif ($definition{SuccessfulOutput}) {
	if ($output !~ /^\s+$/) {
	    my $found=0;
	    $definition{SuccessfulOutput} = [$definition{SuccessfulOutput}] if (ref($definition{SuccessfulOutput}) ne "ARRAY");
	    foreach my $message (@{$definition{SuccessfulOutput}}) {
                if ($output =~ /$message/) {
		    $Trace->debug(msg=>"Success Message Seen: $output");
		    $found=1;
                    last;
		}
            }
	    if (!$found) {
	        $Trace->fatal(msg=>"output = $output", result=>"SCRIPT");
	    }
	}
    }
    
    # run the verification method if one is defined for this routine
    my $negative_test = undef;
    my $verification = $definition{verification};
    $verification = $definition{output_handler} if ($definition{output_handler});

    # if we are trying to ask for help on the command, then setup the help handler
    if ($params{-Help}) {
	$verification = $self->{connections}{$conntype}{help_handler};
        if (!$verification) {
            $Trace->fatal(msg=>"Could not find a help_handler routine defined for the $conntype connection",
                          result => "SCRIPT");
        }
    } elsif ($params{-Idempotent} && ($self->{connections}{$conntype}{idempotent} && 
				 grep($output =~ /$_/s , @{$self->{connections}{$conntype}{idempotent}}))) {
	$Trace->trace(msg=>"Turning verification off since it was an idempotent call");
	$verification = undef;
    } elsif ($definition{negative_verification}) {
	$verification = $definition{negative_verification};
	$negative_test = 1;
    }
    if ($params{-Verification} && $verification) {
	$params{OutputConversion} = $definition{OutputConversion};
	$params{SuccessfulOutput} = $definition{SuccessfulOutput};
	my $internalVerification = (!$definition{output_handler} && $verification ne "_ReturnOutputVerification") 
	|| $params{internal_verification};
	if (wantarray) {
	    $rc = $self->$verification(%input_params, output_in=>$output, 
				       internal_verification=>$internalVerification, 
				       -SkipINFO=>$params{-SkipINFO},
				       want_array=>\@rc,
				       converted_params=>\%all_conversions);
	} else {
	    $rc = $self->$verification(%input_params, output_in=>$output, 
				       -SkipINFO=>$params{-SkipINFO},
				       internal_verification=>$internalVerification,
				       converted_params=>\%all_conversions);
	}
	if ((!$negative_test && !$rc) || ($negative_test && $rc)) {
	    $Trace->fatal(msg=>"$MethodName failed the verification routine", result=>"SCRIPT");
	}
	
    } else {
	$rc = $output;
    }

  done:
    if (!$params{internal_verification} && !$internal_eval && $outputRef) {
	$$outputRef = $output;
    }

    if (wantarray && $rc) {
	$Trace->exit(msg=>"exited with an array of output");
	return @rc;
    } elsif ($rc =~ /^\d$/) {
	$Trace->exit(msg=>"exited with rc = $rc");
    } elsif ($rc) {
	$Trace->exit(msg=>"exited with output");
    } elsif (!defined($rc)) {
	$Trace->exit(msg=>"exited with undef");
    } else {
	$Trace->exit(msg=>"exited with an unknown rc ($rc)");
    }	
    return $rc;
} # run_command

#***********************************************************************
#
# Method:      _convert_params
#
# Description: Convert the old-named parameter names 
#
# Arguments:   
#              the paramaters
#
# Returns:     Returns the new paramater hash
#
#***********************************************************************

sub _convert_params {
    my ($self, $conversion, %params) = @_;
    $Trace->entry();

    foreach my $oldkey (keys %$conversion) {
	if ($params{$oldkey}) {
	    my $newKey = $$conversion{$oldkey};
	    $params{$newKey} = $params{$oldkey} unless (defined($params{$newKey}));
	    delete $params{$oldkey};
	}
    }

    $Trace->exit();
    return %params;
} # _convert_params


# only pass the parameter to the CLI if it exists in our optional_options or required_options
#  lists. This is done because the full show() output hash is sometimes passed back into a
#  command such as a new() or a show() and not all fields are input parameters.
sub _valid_param {
    my $self = shift;
    my (%params) = validate_mixed(
				  @_,
				  [ ],
				  {
                                   cdef => {type => HASHREF},
                                   param => {type => SCALAR},
                                   additional => {type => ARRAYREF|UNDEF, optional => 1},
				  }
				  );
    my $definition = $params{cdef};
    my $param = $params{param};
    my @additional = @{$params{additional}} if ($params{additional});
    if (!$$definition{no_validation}) {
        return ($$definition{options}{$param} && 
                ($$definition{optional_options} && grep($param eq $_, @{$$definition{optional_options}})) || 
                ($$definition{required_options} && grep($param eq $_, @{$$definition{required_options}})));
    } else {
        foreach my $p (@Internal_params, @additional) {
            if ($param eq $p){
                return undef;
            }
        }
        $Trace->debug(msg=>"Returning unverified param: $param.");
        return $param;
    }
}

#***********************************************************************

sub _check_value {
    my $self = shift;
    my ($label, $value, $regexp) = @_;
    my $rc = undef;
    $Trace->entry();

    if (!defined($value) || $value eq "") {
	$Trace->fatal(msg=>"Missing value for $label", result=>"SCRIPT");	
    } elsif ($regexp && $value !~ /$regexp/) {
	$Trace->fatal(msg=>"Error with $label: $value", result=>"SCRIPT");
    } else {
	$rc = 1;
    }
    $Trace->exit(msg=>"rc = $rc");
    return $rc;
} # _check_value


#***********************************************************************

sub _check_host {
    my $self = shift;
    my ($host, $msg) = @_;
    my $rc = undef;
    my $addr = undef;
    $Trace->entry();
    
    if (!($addr = scalar(gethostbyname($host)))) {
	$Trace->fatal(msg=>"Could not determine IP address for $host", result=>"SCRIPT");
    }
    if (!($rc = Socket::inet_ntoa($addr))) {
	$Trace->fatal(msg=>"Could not get $msg for $host\n", result=>"SCRIPT");
    }
    $Trace->exit(msg=>"rc = $rc");
    return $rc;
} # _check_host


#***********************************************************************

sub _compare_hashes {
    my $self = shift;
    my %params = (
		  Skip => undef,
		  Only => undef,
		  Message => "unknown",
		  debug	=> undef,
		  @_);
    my $rc = 1;
    my %Skip;
    my @keys;
    
    $Trace->entry();
    if (!$params{Only}) {
	@keys = keys %{$params{Hash1}};
    }
    if ($params{Skip}) {
	foreach my $key (@{$params{Skip}}) {
	    $Skip{$key} = 1;
	}
    }

    foreach my $key (@keys) {
	next if ($Skip{$key});
	$Trace->debug(msg=>"comparing key \"$key\": ${$params{Hash1}}{$key} ${$params{Hash2}}{$key}") if $params{debug};
	if (${$params{Hash1}}{$key} ne ${$params{Hash2}}{$key}) {
	    $Trace->fatal(msg=>"found a problem with $params{Message} for key $key: '${$params{Hash1}}{$key}' != '${$params{Hash2}}{$key}'", result=>"SCRIPT");
	}
    }
    $Trace->exit(msg=>"rc = $rc");
    return $rc;
} # _compare_hashes

#***********************************************************************

sub _compare_arrays {
    my ($array1, $array2, $debug) = @_;
    $Trace->entry();
    print "$array1, $array2\n" if ($debug);
    if (!$array1 || !$array2) {
	$Trace->exit(msg=>"return undef");
	return undef;
    }
    if ($debug) {
	print "array1=>", join(":", @$array1), "<=\n";
	print "array2=>", join(":", @$array2), "<=\n";
    }
    my $rc = join(":", @$array1) eq join(":", @$array2);
    $Trace->exit(msg=>"rc=$rc\n");
    return $rc;
} # _compare_arrays

#***********************************************************************

sub _process_cmdnode_params {
    my $self = shift;
    my @supportedNodes = @{$self->{cmd_nodes}};
    my @nodes = ();
    $Trace->entry();
    my ($cmd_node, $skip_cmd_node) = $self->{options_object}->get(option=>[qw(cmdnode skipcmdnode)]);
    if ($cmd_node && @{$cmd_node} != ()) {
	foreach my $node (@{$cmd_node}) {
	    if ($node eq "?") {
		my $nodes = join(" ", @supportedNodes);
		$Trace->fatal(msg=>"supported nodes are: $nodes", result=>"SCRIPT");
	    }
	    my @nodenames = grep({$_ eq $node} @supportedNodes);
	    if (scalar @nodenames == 1) {
		push(@nodes, @nodenames);
	    } else {
		$Trace->warn(msg=>"does not support node $node");
		@nodes = ();
	    }
	}
    } else {
	@nodes = @supportedNodes;
    }
    if ($skip_cmd_node && @{$skip_cmd_node} != ()) {
	foreach my $node (@{$skip_cmd_node}) {
	    if ($node eq "?") {
		my $nodes = join(" ", @supportedNodes);
		$Trace->fatal(msg=>"supported nodes are: $nodes", result=>"SCRIPT");
	    }
	    my @nodenames = grep({$_ eq $node} @supportedNodes);
	    if (scalar @nodenames == 1) {
		@nodes = grep($nodenames[0] ne $_, @nodes);
	    } else {
		$Trace->fatal(msg=>"does not support node $node", result=>"SCRIPT");
	    }
	}
    }
done:
    $Trace->exit(msg=>"nodes = @nodes");
    return @nodes;
}

#***********************************************************************
# since the {discover,create,delete}_topology() methods are now in Server.pm
# this package needs to export routines to provide access to the internal
# cmd_node_definitions for each node type.
sub get_topology_definitions {
    my $self = shift;
    my (%params) = validate_mixed(
				  @_,
				  [ ],
				  {
				      definition => {type => SCALAR},
				  }
				  );
    $Trace->entry();

    my $definitions = $self->{cmd_node_definitions}{$params{definition}};
    if (!$definitions) {
        $Trace->fatal(msg=>"Could not locate a cmd_node_definition for '$params{definition}'",
                      result=>"SCRIPT");
    }
    $Trace->exit();
    return %$definitions;
}

#***********************************************************************
# since the {discover,create,delete}_topology() methods are now in Server.pm
# this package needs to export routines to provide access to the internal
# Commands definition (cdef) for each node type.
sub get_command_definition {
    my $self = shift;
    my (%params) = validate_mixed(
				  @_,
				  [ ],
				  {
				      command => {type => SCALAR},
				  }
				  );
    $Trace->entry();

    my $cdef = $self->{Commands}{$params{command}};
    # intentionally not checking for an undef
    $Trace->exit();
    return $cdef;
}

sub get_connection {
    my ($self, %params) = @_;
    
    if (!defined($params{type})) {
	$Trace->fatal(msg=>"a 'type' param is required", result=>"SCRIPT");
    }

    if (!exists($self->{connections}{$params{type}})) {
	$Trace->fatal(msg=>"a '$params{type}' connection has not been created", result=>"SCRIPT");
    }

    return $self->{connections}{$params{type}}{object};
}

# POD for this routine is in Server.pm
sub pick_aggregates {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'Node' } ],
					 {
                                          state => {type => SCALAR|UNDEF, optional => 1, default => "online" },
                                          voltype => {type => SCALAR|UNDEF, optional => 1, default => "flex"},
                                          voltype_discovery_mode => {type => UNDEF|0|1|2|3, optional => 1, default => 0},
                                          avail_size => {type => SCALAR|UNDEF, optional => 1},
                                          topology => {type => HASHREF|UNDEF, optional => 1},
                                          aggregate_list => {type => ARRAYREF|UNDEF, optional => 1},
                                          flexcache_involved => {type => SCALAR|UNDEF, optional => 1, default => undef},
					  block_type => {type => SCALAR|UNDEF, optional => 1, default => undef},
                                          blocktype_discovery_mode => {type => UNDEF|0|1|2|3, optional => 1, default => 0},
					 }
					 );
    my @regexp_aggrs =();
    my @named_aggrs = ();
    my @excluded_aggrs =();
    my @aggrs = ();
    $Trace->entry(msg=>"state=>$params{state}, availsize=>$params{avail_size}");
    if ($self->get_type() == $self->ONTAP_NG() &&
        $params{avail_size} && ($params{avail_size} =~ /^\d+$/)){
	# There isn't a uichange for the data units change.
	# This is a hack since there isn't a cdef for set.
	# We also probably need to evaluate the practice of passing around values
	#   in params that have no units unless we normalize everything to bytes
	#   ourselves.......
	my $output = $self->cli(cmd => "set");
	# This next line is a *huge* hack since Tricky actually only shows some of
	# the fields in bytes.  This will need to be revisited if/when availsize is
	# shown in bytes.

        # The branch is being used here since the version numbers for mainng
        # looks like Tricky.3.
        my $version = $self->get_release(format=>"cooked");
	if ($version !~ /^tricky\.(1|2|3|4)/i && ($output =~ /Data Units/)){
	    # We need to translate the size here since Tricky.2 shows all values in
	    # bytes when -showallfields is true.  The input size is assumes to be
	    # MB
	    $params{avail_size} .= "M";
	}
    }
    if ($params{avail_size} && ($params{avail_size} !~ /^\d+$/)){
	$Options->unit_to_base($params{avail_size});
    }

    my $options = $Options->get_hash();
    my $type;
    if ($params{voltype_discovery_mode} == 1) {
        $type = $params{voltype} || $options->{VOLUME_STYLE};
    } elsif ($params{voltype_discovery_mode} == 2) {
        $type = $options->{VOLUME_STYLE} || "*";
    } elsif ($params{voltype_discovery_mode} == 3) {
        $type = $params{voltype};
    } else {
        $type = $options->{VOLUME_STYLE} || $params{voltype};
    }
    my $blocktype;
    if ($params{blocktype_discovery_mode} == 1) {
        $blocktype = $params{block_type} || $options->{AGGR_BLOCK_TYPE};
    } elsif ($params{blocktype_discovery_mode} == 2) {
        $blocktype = $options->{AGGR_BLOCK_TYPE} || "*";
    } elsif ($params{blocktype_discovery_mode} == 3) {
        $blocktype = $params{block_type};
    } else {
        $blocktype = $options->{AGGR_BLOCK_TYPE} || $params{block_type};
    }

    my $move_member = $options->{MOVE_MEMBER} || 0;
    if ($move_member) {
        $Trace->comment(msg=>"MOVE_MEMBER is true, looking for member destination aggregate");
    }

    my @aggregates;
    my $topology = $self->{server_object}->get_topology();
    if (!$params{aggregate_list}) {
        # verify that the topology has been discovered, if not then discover it
        if (!exists($topology->{cluster}{aggregate}) ||
            !exists($topology->{node})) {
            $topology = $self->{server_object}->discover_topology(cmd_nodes=>["aggregate"]);
        }
        # Please see the NOTE in Server.pm regarding the Topology hack
        if ($self->get_type() == $self->ONTAP_NG() || !$self->{server_object}->get_clustered_nodes()) {
            @aggregates = @{$topology->{cluster}{aggregate}};
        } else { # node
            if ($topology->{node}) {
                foreach my $node (keys %{$topology->{node}}) {
                    if ($topology->{node}{$node}{aggregate}) {
                        push(@aggregates, @{$topology->{node}{$node}{aggregate}});
                    }
                }
            }
        }
        push(@aggregates, @{$params{topology}{aggregate}}) if ($params{topology} && $params{topology}{aggregate});
    } else {
        push(@aggregates, @{$params{aggregate_list}});
    }
    if (!scalar(@aggregates)) {
	$Trace->fatal(result=>"CONFIG", msg=>"No aggregates were found");
    }
    # Separate Includes and Excludes in separate arrays.
    foreach my $tmp_user_aggr ( @{$options->{aggregate}}) {
        # Any aggregate names preceeded with a ! and with the / character around
        #  a regular expression will exclude any aggregates that match the regexp
	if ($tmp_user_aggr =~ /^!\/(.+)\//) {
	    push(@excluded_aggrs, $1);
        } elsif ($tmp_user_aggr =~ /^(!|~)(.+)/) {
            # this is the original way to exclude a specific aggregate name
	    push(@excluded_aggrs, "^$2\$");
	} else {
            # If the param is surrounded by // then it is a regular expression. Do
            #  nothing to the param and just pass it to the later code which will
            #  compare with =~
            if ($tmp_user_aggr =~ m!^/(.+)/$!) {
                push(@regexp_aggrs, $1);
            } else {
                # If MOVE_MEMBER is true then we are doing a member move and must
                # match on member aggregate names that end in _000, _001, etc.
                if ($move_member && $type eq 'striped') {
                    $tmp_user_aggr .= "_[0-9]{3}";
                    push(@regexp_aggrs, $tmp_user_aggr);
                } else {
                    push(@named_aggrs, $tmp_user_aggr);
                }
            }
	}
    }

    # get the list of available servers
    my $server_obj = $self->{server_object};
    my @servers = $server_obj->servers();
    $Trace->trace(msg=>"Specified filers=".join(",", @servers));

    $Trace->debug(msg=>"regexp aggregates=".Dumper(\@regexp_aggrs));
    $Trace->debug(msg=>"named aggregates=".Dumper(\@named_aggrs));
    $Trace->debug(msg=>"excluded aggregates=".Dumper(\@excluded_aggrs));
    $Trace->debug(msg=>"\@aggregates=".Dumper \@aggregates);
    # Examine each aggr that has been discovered and only look at ones that meet the input params.
    # Skip any private test aggrs (those that start with p_).
    aggrloop: 
    foreach my $aggr (@aggregates) {

        $Trace->debug(msg=>"check aggregate '$aggr->{aggregate}'\n");
        # standardize $filer to be an array ref
	my $filer = $aggr->{filers};
	if (ref($filer) ne "ARRAY"){
	    # If it's not an array, turn it into one
	    my @a = split(/,/, $filer);

	    # Only use unique keys though since a filer can be listed twice and that would screw up the aggr check
            $self->_uniq(\@a);
	    $filer = \@a;
	}

        #
        # Remove aggregates that aren't hosted on nodes in the known list
        #
	$Trace->debug(msg => "Servers => " . Dumper(\@servers) . ", Nodes => " . Dumper($filer));

	foreach my $ifiler (@$filer){
	    if (!grep(/^$ifiler$/i, @servers)){
		$Trace->comment(msg => "Skipping aggregate $aggr->{aggregate} because it isn't on one of the specified filers\n");
		next aggrloop;
	    }
	}

        if ($options->{SKIP_MROOT_AGGREGATE} && 
            ( $aggr->{'has-mroot'} eq "true" || $aggr->{'has-partner-mroot'} eq "true" )) {
            $Trace->comment(msg => "Skipping aggregate $aggr->{aggregate} because it contains the mroot\n");
            next;
        }

        if ($params{flexcache_involved}) {
            # if this aggregate is on a flexcache, then mark it as such
            foreach my $ifiler (@$filer){
                if ($server_obj->is_cache(-Server=>$ifiler)) {
                    $aggr->{flexcache} = 1;
                    $Trace->trace( msg=>"marking $aggr->{aggregate} as a flexcache aggregate");
                }
            }

            # don't return trad volumes when caches are involved
            if ($aggr->{type} eq "trad") { #!!!!! This may not be right !!!!!!!!
                $Trace->comment(msg=>"skipping $aggr->{aggregate} on ".((ref($aggr->{filers}) eq "ARRAY") ? join(', ', @{$aggr->{filers}}) : $aggr->{filers})." because it is not an aggregate\n and this test involves a flexcache. The type was '$aggr->{type}'");
                next;
            }
        }

        # skip an aggr if it is private and not specified by the user
	if (!@named_aggrs && !@regexp_aggrs && ($aggr->{aggregate} =~ /^p_/)) {
	    $Trace->comment(msg=>"skipping private aggregate '$aggr->{aggregate}'");
	    next;
	}
        
        # skip aggregates which match the excluded regular expression
        if (scalar(@excluded_aggrs) && grep($aggr->{aggregate} =~ /$_/, @excluded_aggrs)) {
            $Trace->comment(msg=>"skipping aggregate '$aggr->{aggregate}' because it is set to be excluded in the input list");
            next;
        }

	# skip this aggr if the user specified some and it is not in the list
	if (scalar(@regexp_aggrs)) {
	    if (!grep($aggr->{aggregate} =~ /$_(\_\d\d\d)?$/, @regexp_aggrs)) {
		$Trace->comment(msg=>"skipping aggregate '$aggr->{aggregate}' because it did not match the regular expression");
		next;
	    }
	}
        # if the user provided specific names of aggregates, skip any aggregates that do not match the names and
        # flag that we found a match
        my $found_named_aggrs = undef;
        if (scalar(@named_aggrs)) {
	    if (!grep($aggr->{aggregate} eq $_, @named_aggrs)) {
		$Trace->comment(msg=>"skipping aggregate '$aggr->{aggregate}' because it is not in the specific list");
		next;
	    } else {
                $found_named_aggrs = 1;
            }
	}            

        # The default value of voltype is 'flex' so a user might provide a named striped aggregate. The code
        # should not force the user to also specify the volume type.
        # Skip an aggregate if the style/match does not match but only if we did not find specific matches.
	if ($type && ($type ne "*") && !$found_named_aggrs){
              if (!$aggr->{flexcache} && ($aggr->{voltype} && $aggr->{voltype} !~ /$type/) ) {
                  $Trace->comment(msg=>"Skipping aggregate '$aggr->{aggregate}' because the volume style '$aggr->{voltype}' was not '$type'");
                  next;
              }
	}

	# now only add the aggr if it matches the input params. 
	# if the user specified a list, then issue a warning if the criteria is not met
        ## HACK ko ##
        # If the output of the size field is in GB, adjust it to MB (the test code expected values
        ## HACK ko ##
	## AND YES, I'M MAKING THIS AS UGLY AS POSSIBLE
	## AND YES, I'M MAKING THIS AS UGLY AS POSSIBLE
        my $size_mb = $aggr->{availsize};
        if ( $aggr->{availsize} =~ /GB/ ) {
            $Trace->debug(msg=>"Returned availsize is $aggr->{availsize}, and will be converted to megabytes");
            $size_mb =~ s/GB//;
            $size_mb *= 1024;
        } elsif ( $aggr->{availsize} =~ /TB/ ) {
            $Trace->debug(msg=>"Returned availsize is $aggr->{availsize}, and will be converted to megabytes");
            $size_mb =~ s/TB//;
            $size_mb *= 1024;
            $size_mb *= 1024;
        } elsif ($self->get_release(format=>"cooked") =~ /tricky.2/i) {
            # this is a hack until we have a real feature test. The availsize field in tricky.2 defaults to
            #  just regular bytes. Prior to this release, no mb or gb implied MB. This value needs to be
            #  converted to MB.
            $Trace->debug(msg=>"Returned availsize is $aggr->{availsize}, and will be converted to megabytes");
            $size_mb /= 1024;
            $size_mb /= 1024;
            $size_mb = int($aggr->{availsize});
        }

	if (defined($params{avail_size}) && $size_mb < $params{avail_size}) {
	    if ($found_named_aggrs) {
		$Trace->warn(msg=>"$aggr->{aggregate} needs $params{avail_size} space available but only has $size_mb");
	    } else {
		$Trace->comment(msg=>"$aggr->{aggregate} needs $params{avail_size} space available but only has $size_mb");
	    }
            next;
	}
	if (defined($params{state}) && $aggr->{state} ne $params{state}) {
	    if ($found_named_aggrs) {
		$Trace->warn(msg=>"$aggr->{aggregate} should be '$params{state}' but it is in the state '$aggr->{state}'");
	    } else {
		$Trace->comment( msg=>"$aggr->{aggregate} should be '$params{state}' but it is in the state '$aggr->{state}'");
	    }
            next;
	}
	if(defined($blocktype) && $blocktype ne "*" && $blocktype ne $aggr->{'block-type'}) {
            if(!defined($aggr->{'block-type'})) {
                if($blocktype ne "32-bit") {
                    $Trace->warn(msg=>"This Ontap version supports 32-bit block type only");
                    next;
                }
            } else {
                $Trace->comment( msg=>"Skipping aggregate '$aggr->{aggregate}' because the block type was not '$blocktype'");
                next;
            }
        }

	$Trace->trace(msg => "Aggregate $aggr->{aggregate} meets the criteria.  Adding to available list");
	# this aggr meets all critera so save it and remove it from the input list
	push(@aggrs, $aggr);
    }
    if (scalar(@named_aggrs) && !scalar @aggrs) {
        $Trace->comment(msg=>"Aggregate list=".Dumper(\@aggregates));
	$Trace->fatal(result=>"CONFIG", msg=>"Could not find the aggregates that you specified(".join(" ", @named_aggrs).")");
    } elsif (!scalar @aggrs) {
        $Trace->comment(msg=>"Filtered aggregate list=".Dumper(\@aggrs));
        $Trace->comment(msg=>"Aggregate list=".Dumper(\@aggregates));
	$Trace->fatal(result=>"CONFIG", msg=>"Could not find any aggregates with state='$params{state}' and $params{avail_size} space available. Aggregates=");
    }

    $Trace->exit(msg=>"return ".join(" ", map $_->{aggregate}, @aggrs));
    @{$params{aggregate_list}} = @aggrs if ($params{aggregate_list});

    return @aggrs;
}

#***********************************************************************
# SUBROUTINE: CreateDoc
#
# PURPOSE: Walk through the command hash and formulate the POD
# 	   documentation
# OUTPUT: Places POD text in the /x/eng/spin-cores/documentation/
#	  directory. Also executes "pod2html" on the file and creates
#	  the doc webpage
#***********************************************************************
sub CreateDoc {
    my $self = shift;
    my %Commands = %{$self->{Commands}{$self->{server}}};
    my $implements = "";
    my $docPath = "/x/eng/spin-cores/documentation/";
    my $docFile="";
    my $rc = undef;

    $Trace->entry();
    if ($self->{SpinFS}) {
	$docFile = "SpinServer";
    } elsif ($self->{ONTAP_CLASSIC}) {
	$docFile = "Server";
    } elsif ($self->{ONTAP_NG}) {
	$docFile = "OntapNG";
    } 

    ############################################################
    # Open the target file
    ############################################################
    open(OUT, "> $docPath/$docFile.podtext");

    foreach my $command (sort(keys(%Commands))) {
	print OUT "############################################################\n";
	print OUT "\n";
	############################################################
	# The first POD header is the command name
	############################################################ 
	print OUT "=head1 C<".$command."()>\n";
	print OUT "\n";

	############################################################
	# Obtain the command definiton for this command
	############################################################
	my %definition = %{$Commands{$command}};

	############################################################
	# Next print OUT the CLI command that this command implements
	############################################################
	print OUT "=over 4\n";
	print OUT "\n";
	print OUT "=item Description\n";
	print OUT "\n";
	print OUT "This method implements the '".$definition{Command}."' command\n";
	print OUT "\n";

	############################################################
	# Print list of Arguments
	############################################################
	print OUT "=item Arguments\n";
	print OUT "\n";

	if ( (scalar(@{$definition{required_options}}) == 0) &&
	     (scalar(@{$definition{optional_options}}) == 0) &&
	     (scalar(@{$definition{PostCommandOptions}}) == 0) 
	   ) {
	    print OUT "None\n";
	    print OUT "\n";
	} else {
	    print OUT "=over 4\n";
	    print OUT "\n";
	    foreach my $required (@{$definition{required_options}}) {
	        print OUT "=item $required\n";
		print OUT "(Required)\n";
		print OUT "\n";
		############################################################
		# Determine which -option this Option implements
		# Check to see if this option has a replacment option 
		# defined. If so use it. If not, 
		############################################################
		$implements = (defined($definition{ReplacementOptions}{$required})) ? $definition{ReplacementOptions}{$required} : $definition{options}{$required};
		print OUT "Implements the required '$implements' flag\n";
		print OUT "\n";
		############################################################
		# Get list of supported values
		############################################################
		if (defined($definition{SupportedValues}{$required})) {
		    my @values = @{$definition{SupportedValues}{$required}};
		    print OUT "Supported Values: @values\n";
		    print OUT "\n";
		}
		############################################################
		# Get default value
		############################################################
		if (defined($definition{Defaults}{$required})) {
		    print OUT "Default value: $definition{Defaults}{$required}\n";
		    print OUT "\n";
		}
	    }
	    foreach my $optional ((@{$definition{optional_options}}), (@{$definition{PostCommandOptions}})) {
                print OUT "=item $optional\n";
                print OUT "(Optional)\n";
                print OUT "\n";
		############################################################
                # Determine which -option this Option implements
                # Check to see if this option has a replacment option
                # defined. If so use it. If not,
                ############################################################
                $implements = (defined($definition{ReplacementOptions}{$optional})) ? $definition{ReplacementOptions}{$optional} : $definition{options}{$optional};
                print OUT "Implements the optional '$implements' flag\n";
                print OUT "\n";
		############################################################
                # Get list of supported values
                ############################################################
                if (defined($definition{SupportedValues}{$optional})) {
		    my @values = @{$definition{SupportedValues}{$optional}};
		    print OUT "Supported Values: @values\n";
                    print OUT "\n";
                }
		############################################################
                # Get default value
                ############################################################
                if (defined($definition{Defaults}{$optional})) {
                    print OUT "Default value: $definition{Defaults}{$optional}\n";
                    print OUT "\n";
                }
            }
	    print OUT "=back\n";
	    print OUT "\n";
	}
	############################################################
	# Print any release changes
	############################################################
	if (scalar(keys(%{$definition{Release}})) > 0) {
	    ############################################################
	    # We don't care about OutputConversion
	    ############################################################
	    my $header = 0;
	    my $found = 0;
	    foreach my $release (keys(%{$definition{Release}})) {
	        foreach my $item (keys(%{$definition{Release}{$release}})) {
		    if ($item ne "OutputConversion") {    
			$found = 1;
		        if (!$header) {
			    print OUT "=item Release Changes\n";
			    print OUT "\n";
		 	    print OUT "=over 4\n";
		 	    print OUT "\n";
			    print OUT "=item Release $release\n";
			    print OUT "\n";
			    print OUT "=over 4\n";
			    print OUT "\n";
			    $header = 1;
			}
		        print OUT "=item $item\n";
		        print OUT "\n";
		        if ($item eq "ReplacementOptions") {
			    foreach my $option (keys(%{$definition{Release}{$release}{ReplacementOptions}})) { 
			        print OUT "The $option option now implements the '$definition{Release}{$release}{ReplacementOptions}{$option}' flag\n";
			        print OUT "\n";
			    }
			}
			if ($item eq "Command") {
                            print OUT "This method now implements the '$definition{Release}{$release}{Command}' command\n";
                            print OUT "\n";
                        }
			if ($item eq "OptionalOptions") {
			    print OUT "The following options are available in this release:\n";
			    print OUT "\n";
			    print OUT "=over 4\n";
			    print OUT "\n";
			    foreach my $optional ((@{$definition{Release}{$release}{OptionalOptions}})) {
	                        print OUT "=item $optional\n";
	                        print OUT "(Optional)\n";
                		print OUT "\n";
		                ############################################################
               			# Determine which -option this Option implements
		                # Ignore any replacement options for these Release specific
               		        # options. This may need to be implmented later
		                ############################################################
		                $implements = $definition{options}{$optional};
		                print OUT "Implements the optional '$implements' flag\n";
               		        print OUT "\n";
		                ############################################################
		                # Get list of supported values
		                ############################################################
		                if (defined($definition{SupportedValues}{$optional})) {
	                            my @values = @{$definition{SupportedValues}{$optional}};
               			    print OUT "Supported Values: @values\n";
		                    print OUT "\n";
                		}
		                ############################################################
		                # Get default value
		                ############################################################
		                if (defined($definition{Defaults}{$optional})) {
               		            print OUT "Default value: $definition{Defaults}{$optional}\n";
		                    print OUT "\n";
                		}
		            }
			    print OUT "=back\n";
			    print OUT "\n";
		        }
		    }
		}
		if ($found) {
		    print OUT "=back\n";
		    print OUT "\n";
		    print OUT "=back\n";
		    print OUT "\n";
		}
	    }
	}
	############################################################
	# Print the return values
	############################################################
	print OUT "=item Returns\n";
	print OUT "\n";
	print OUT "None\n";
	print OUT "\n";
	print OUT "=back\n";
	print OUT "\n";
	print OUT "=cut\n";
	print OUT "\n";
    }
    print OUT "############################################################\n";
    close(OUT);

    ############################################################
    # Create the corresponding webpage
    ############################################################
    my $filename = "$docPath/$docFile.podtext";
    if (-e $filename) {
	system("pod2html $filename > $docPath/$docFile.html 2>/dev/null");
	$rc = 1;
    } else {
	$Trace->fatal(msg=>"$docPath/$docFile.podtext does not exist", result=>"SCRIPT");
    } 	

  done:
    $Trace->exit(msg=>"rc = $rc");
    return $rc;
}

#***********************************************************************

sub ontap_classic {
    $Trace->entry();
    $Trace->exit();
    return ONTAP_CLASSIC();
}

sub ontap_ng {
    $Trace->entry();
    $Trace->exit();
    return ONTAP_NG();
}

#***********************************************************************

sub GetVirtualServers {
    $Trace->entry();
    $Trace->exit();
    return "unknown";
}

#***********************************************************************
sub _check_connection {
    $Trace->entry();
    $Trace->exit();
}

sub _convert_definition {
    my ($self, $def_in) = @_;
    my %definition = %$def_in;

    my %conversion = (OptionalOptions => "optional_options",
		      RequiredOptions => "required_options",
		      );
    $Trace->entry();
    foreach my $key (qw(Command Conversion Options OptionalOptions RequiredOptions Verification)) {
	if ($$def_in{$key}) {
	    my $new_key = ($conversion{$key}) ? $conversion{$key} : lc $key;
	    $definition{$new_key} = $definition{$key};
	    delete $definition{$key};
	}
    }
    $Trace->exit();
    return %definition;
}

sub _cmd_deprecated {
    my ($self, $old_call, $new_call) = @_;
    print "*** The $old_call method is being deprecated in the next release of the Server object.\n";
    print "*** Please switch to calling $new_call instead.\n";
}

sub _read_topology_files {
    my ($self) = @_;
    my $return;
    $Trace->entry();
    $Trace->fatal(msg=>"This code is not supported yet");
    my $files = $Options->get(option=>"TOPOLOGY");
    goto done if (!$files);

    foreach my $file (@$files) {
	our %topology = ();
	$Trace->debug(msg=>"reading $file\n");
	unless ($return = do $file) {
	    $Trace->fatal(result=>"CONFIG", msg=>"Could not parse $file: $@\n") if ($@);
	    $Trace->fatal(result=>"CONFIG", msg=>"Could not do $file: $!") unless (defined $return);
	    $Trace->fatal(result=>"CONFIG", msg=>"Could not run $file\n") unless ($return);
	}
	
        my $topology = $self->{server_object}->get_topology();
	foreach my $cmdnode (keys %topology) {
	    $topology->{$cmdnode} = [] if (!exists($topology->{$cmdnode}));
	    if (ref($topology{$cmdnode}) eq "ARRAY") {
		push(@{$topology->{$cmdnode}}, @{$topology{$cmdnode}});
	    } else {
		$Trace->fatal(msg=>"The '$cmdnode' key in file $file is unknown", result=>"CONFIG");
	    }
	}
    }

  done:
    $Trace->exit();
}

sub _write_topology_files {
    my ($self, $topology) = @_;
    $Trace->entry();
    my $files = $Options->get(option=>"TOPOLOGY");
    goto done if (!scalar @$files);
    my $file = $$files[0];
    if (scalar @$files > 1) {
	$Trace->comment(msg=>"Multiple files have been specified but only '$file' will be written");
    }

    if (!open(FILE, ">$file")) {
	$Trace->fatal(result=>"CONFIG", msg=>"Could not open $file: $!");
    }
    print FILE Data::Dumper->Dump([$topology], [qw(*topology)]);
    $Trace->exit();
}

sub _find_method_definition {
    my ($self, $method) = @_;
    $Trace->entry();
    my %method_params = ();

    # The specific methods for each command node are specified in a hash created by the derived class.
    #  The value is stored as either just the name of the routine to call for the method, or an anonymous array
    #  containing the name of the routine as well as the named parameters to pass to it.

    # if $method is an array ref, then find the actual method name and %params inside the array ref
    if (ref($method) eq "ARRAY") {
	($method, %method_params) = @{$method};

	# evaluate the method params since they might contain variable names
	map $method_params{$_} = eval "$method_params{$_}", keys %method_params;
    }
    # else, the method is just the name of the method, and there are no %params so return the empty list
    $Trace->exit();
    return ($method, %method_params);
}

sub _say {
    my ($self, $type, $conn, $cmd, @sayparams) = @_;
    my ($match, $before, $after, $prompt);
    my $retry_performed = 0;
    $Trace->entry(msg=>join(",", $cmd, @sayparams));

    my $poke_cli = $Options->get(option=>"POKE_CLI");
    my ($retry_timeout, $retry_interval) = $Options->get(option=>[qw(RETRY_SAY_FAILURE RETRY_SAY_FAILURE_INTERVAL)]);

    $Trace->debug(msg=>"Connection: $conn");
    $Trace->debug(msg=>"Say Params: $cmd, @sayparams");
    if (($type eq "cli_connection") && ($poke_cli)) {
    	($match, $before, $after, $prompt) = main::say($conn, $cmd, @sayparams, -eval_error);
        $Trace->debug(msg=>"Match:  $match");
        $Trace->debug(msg=>"Before: $before");
        $Trace->debug(msg=>"After:  $after");
        $Trace->debug(msg=>"Prompt: $prompt");
        if ($@) {
            if ($@ =~ /Err=TIMEOUT/i) {
                $Trace->trace(msg=>"Before poke: Match:  $match");
                $Trace->trace(msg=>"Before poke: Before: $before");
                $Trace->trace(msg=>"Before poke: After:  $after");
                $Trace->trace(msg=>"Before poke: Prompt: $prompt");
                $Trace->warn(msg=>"!!!COMMAND TAKING TOO LONG...POKING AT NGSH WITH A SINGLE CARRIAGE RETURN!!!");
                ($match, $before, $after, $prompt) = main::say($conn,"\r",'-nocr',
                                                               -fatal, "Error sending the command \"\r\" to the node");
                $Trace->trace(msg=>"After poke:  Match:  $match");
                $Trace->trace(msg=>"After poke:  Before: $before");
                $Trace->trace(msg=>"After poke:  After:  $after");
                $Trace->trace(msg=>"After poke:  Prompt: $prompt");
            } else {
                $Trace->fatal(msg=>"$@", result=>"FAIL");
            }
    	}
    } elsif ($retry_timeout) {
        my $start_time = time;
        my $retry_error;
        my $elapsed_time = undef;
        my $say_ok = undef;
        $Trace->comment(msg=>"Wait up to $retry_timeout seconds for the call to complete");
        do {
            $retry_error = undef;
            eval {
                if (!$conn) {
                    $self->establish_connections($type=>1);
                    $conn = $self->{connections}{$type}{object};
                    main::say($conn, "", -nocr, -swallow);
                }
                ($match, $before, $after, $prompt) = main::say($conn, $cmd, @sayparams);
            };
            if ($@) {
                $retry_error = $@;
                $Trace->trace(msg=>"Encountered an error with the connection");
                $Trace->trace(msg=>"close the connections");
                for my $conn_type (qw(shell cli dblade cfe)) {
                    $self->close_connection(type => "${conn_type}_connection", keep_state => 1);
                }
                $conn = undef;

                $retry_performed = 1;
                sleep $retry_interval;
            } else {
                $say_ok = 1;
            }
            $elapsed_time = time - $start_time;
            $Trace->comment(msg=>"say_ok=$say_ok, elapsed_time=$elapsed_time, retry_timeout=$retry_timeout");
        } until ($say_ok || ( $elapsed_time >= $retry_timeout ));
        
        if ($retry_error) {
            $Trace->fatal(msg=>"Failed to retry the command \"$cmd\" after $retry_timeout seconds: $retry_error");
        } elsif ($retry_performed) {
            $Trace->comment(msg=>"Successfully retried the command \"$cmd\" after $elapsed_time seconds");
        }
    } else {	
    	($match, $before, $after, $prompt) = main::say($conn, $cmd, @sayparams,
                                                       -fatal, "Error sending the command \"$cmd\" to the node");
    }
    if (grep('-nocr', @sayparams)) {
        main::say($conn, "", -nocr, -swallow);
    }
    $Trace->exit();
    return($match, $before, $after, $prompt, $retry_performed);
}

#
# volume: the volume hash for the volume that will be cached
# volume_name: the name of the volume that will be cached
# source: the name of the source filer (or nblade IP)
# cache: the name of the cache node
# topology: a pointer to the has containing the current Test topology to be created
sub _define_flexcache_volume {
    my ($self, $volume, $volume_name, $source, $cache, $topology) = @_;
    $Trace->entry(msg=>"$volume_name");
    $Trace->debug(msg=>"Volume=>".Dumper($volume));
    $Trace->debug(msg=>"cache=>$cache");
    $Trace->debug(msg=>"source=>source");

    # setup the flexcache volume definitions for create_topology
    my $cache_volume_name = $volume_name . "_fc";

    # see if this volume already exists in the topology
    my @vols = $self->{server_object}->search_topology(cmd_node => "cache_volume",
				      key_pairs => ["volume:$cache_volume_name",
						    "-Filer:$cache"],
				      topology => $topology);

    # skip this volume if it is already been added to the list
    if (!scalar @vols) {
	# find available aggrs
	my @aggrs = $self->{server_object}->search_topology(cmd_node => "aggregate",
					   key_pairs => ["filers:$cache",
							 "type:!trad"]);

	if (! scalar @aggrs) {
	    $Trace->fatal(msg=>"Could not find any aggregates on $cache");
	}
	# pick one at random
	my $cache_aggr = $aggrs[int rand(scalar @aggrs)];
	$Trace->debug(msg=>"cache_aggr=>".Dumper($cache_aggr));
	my $size = $volume->{size};
	$size =~ s/KB/k/i;
	$size =~ s/MB/m/i;
	$size =~ s/GB/g/i;
	$size =~ s/TB/t/i;
	# create the volume on the cache
        $Trace->trace(msg=>"Adding cache volume '$cache_volume_name' to cache $cache with source=$source:$volume_name");
	push(@{$topology->{node}{$cache}{cache_volume}},
	     {
              volume => "$cache_volume_name",
              size => $size,
              ruleset => $volume->{ruleset},
              vserver => $volume->{vserver},
              aggregate => $cache_aggr->{aggregate},
              source => "$source:$volume_name",
              -Filer => $cache,
	     });
	push(@{$topology->{node}{$cache}{exportruleset_rules}},
             {
              ruleindex => 1,
              vserver => $volume->{vserver},
              rulesetname => $volume->{ruleset},
              anon => 0,
              protocol => "any",
              clientmatch => "0.0.0.0/0",
              rorule => "any",
              rwrule => "any",
              superuser => "any",
             });

	$Trace->debug(msg=>"cache_volume=>".Dumper($topology->{cluster}{cache_volume}));
    }
    $Trace->exit();
}


# globally define this routine to return false. Cache objects will define their own to return true
sub is_cache {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"no");
    return undef;
}

# globally define this routine to return false. The individual node objects which support specifying
#  initial mode bits on a volume will return true
sub supports_mode_in_volume_create {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"no");
    return undef;
}

# globally define this routine to return false. The individual node objects which support 
#  a Global Name Space (GNS) will return true
sub supports_GNS {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"no");
    return undef;
}

sub get_type {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>$self->{type});
    return $self->{type};
}
    

# stub routine (to cover Classic)
sub _find_all_servers {
    my $self = shift;
    return ($self->{server});
}

sub _uniq {
    my $self = shift;
    my $array = shift;
    my %items;
    foreach my $key (@$array) {
        $items{$key} = 1;
    }
    @$array = keys %items;
}

# Convenience routines
sub shell {
    my $self = shift;
    $self->say_wrapper(type=>"shell_connection", @_);
}

sub cli {
    my $self = shift;
    $self->say_wrapper(type=>"cli_connection", @_);
}

sub dblade {
    my $self = shift;
    $self->say_wrapper(type=>"dblade_connection", @_);
}

sub dblade_set_privilege {
    my $self = shift;
    my %params = @_;

    # Validate argument parameters
    my @level_choices = qw(admin advanced test diag);
    my $priv_level_usage = "Choices for level are one of the following:".join(",", @level_choices);

    if (!grep($params{level} eq $_, @level_choices)) {
        $Trace->fatal(msg=> "$params{level} is an incorrect value.\n".$priv_level_usage);
    }

    my @quiet_choices = qw(on off);
    my $quiet_usage = "Choices for quiet are one of the following:".join(",", @quiet_choices);

    if (!grep($params{quiet} eq $_, @quiet_choices)) {
        $Trace->fatal(msg=> "$params{quiet} is an incorrect value.\n".$quiet_usage);
    }

    #build the privilege set string
    my $separator = q{ }; # space
    my $set_string = q{priv set};
    my $q_option = q{}; # empty string
    my $p_string = qq{$params{level}};

    #if quiet mode on, use "-q" option
    if ($params{quiet} eq "on") { $q_option = "-q" };

    #assign privilege level and join the string
    $priv_string = $set_string.$separator.$q_option.$separator.$p_string.";";
}

sub console {
    my $self = shift;
    $self->say_wrapper(type=>"console_connection", @_);
}

sub cfe {
    my $self = shift;
    $self->say_wrapper(type=>"console_connection", @_);
}

sub rlm {
    my $self = shift;
    $self->say_wrapper(type=>"rlm_connection", @_);
}

sub servers {
    my $self = shift;
    
    return $self->{server_object}->servers();
}

# this routine is used by the objects to find the list of cdef modules to
# load. This is a hack to have a common list loaded for all objects.
sub _common_modules {
    my $self = shift;

    return @Modules;
}


#
# enable diag user
#
# This routine will enable the special 'diag' user account.  It is used to login into bsd from the
# console or through an ngsh session
#
sub enable_diag {
    my $self = shift;
    my %params = (
		  node => $self->{server},
		  password => $self->{CLI_PASSWORD},
		  mode => "clusterm",
		  @_,
		  );

    my $node = $params{node};
    my $password = $params{password};
    my $mode = $params{mode};

    $Trace->entry(msg=>"Enabling 'diag' on node: $node with password: $password, in mode: $mode");
    eval {
	if ( $mode eq "7m" ) {
	    $self->console(cmd=>"priv set diag");
	    $self->console(cmd=>"useradmin diaguser unlock");
	    $self->console(cmd=>"useradmin diaguser password",
			   additional_prompts_timeout => "60",
			   additional_prompts => {
			       "(E|e)nter a new password:" => $password,
			       "(E|e)nter it again:" => $password,
                               '\S+::' => "FINAL_PROMPT",
                           });
	} else {
	    $self->cli(cmd=>"security login unlock -username diag");
	    $self->cli(cmd=>"security login password -username diag",
		       additional_prompts_timeout => "20",
		       additional_prompts => {
			   "(E|e)nter a new password:" => $password,
			   "(E|e)nter it again:" => $password,
			   "$node::.?>" => "FINAL_PROMPT",
		       });
	}
    };
    if ($@) {
	$Trace->fatal(msg=>"Error while enabling diag user:\n$@",
		      result => "FATAL");
    }
    $Trace->exit();
}



#start sub agnostic_transit_to
sub agnostic_transit_to {
    $Trace->entry();
    my $self = shift;
    my %params = (
          node => undef,
          clustered => undef,
          #might be needed at some point if we find that 7mode transits are 
          #timeing out; so far they are not
          #timeout => 120,
          @_,
          );   
    $Trace->debug(msg => "Sub agnostic_transit_to \$params dumper:\n".
                  Dumper(\%params));

    #catch all states and translate them into 7mode transit states if
    #working in 7mode
    my %state_map = (
        CFE => "FIRMWARE",
        IBM => "BOOT_MENU",
        LOGIN => "UP",
        REBOOT => "REBOOT",
        CLI => "UP",
        MAINT => "MAINT",
        SHELL => "UP",
    );
	unless (defined($params{clustered})) {
		my $mode = mode_detector($self, %params);
		if ($mode) {
			if ($mode == 10) {
				$params{clustered} = 1;
			} elsif ($mode == 7) {
				$params{clustered} = 0;
			}
		} else {
			$Trace->fatal(msg=>"Unable to determine the mode {7 or 10} of '$params{node}'!\n".
							   "It must be HUNG.", result=>"FATAL");
		}
	}
    #if operating in 10mode then simple Node::NG::Transit.pm calls
    if ($params{clustered}) {
        $Trace->comment(msg=>"Running 10mode transit! Trying to get to '$params{state}'.");
        delete($params{node});
        delete($params{clustered});
        $self->transit_to(%params, RR_FLAG => main::param('RR_FLAG', -default, 0));
    #if operating in 7mode then translate the states into 7mode 
    #states and use boot.pm transit
    } else {
        my $attempt=1;
        my $max_attempts=3;
        #loop around 3 times since boot.pm transit_to does not poke more then 
        #once it is possibel to miss a state, thus handle here
        while ($attempt <= $max_attempts) {
            $Trace->comment(msg=>"Running 7mode transit! Trying to get to '$state_map{$params{state}}'.");
            $Trace->comment(msg=>"State translation: '$params{state}'=>'$state_map{$params{state}}'");
            my ($success, $state);
            #since boot.pm has a seperate method for REBOOT we must catch here and call that method
            if ($params{state} =~ /REBOOT/) {
                ($success, $state) = boot::reboot_to($params{node}."#console","UP");
            #since boot.pm does not handle poking in the CTL-C to get to BOOT_MENU
            #we must catcht he call and make multiple state and transit calls
            } elsif ($params{state} =~ /IBM/) {
                my $current_state = agnostic_state($self,node=>$params{node},clustered=>$params{clustered});
                if ($current_state !~ /IBM/) {
                    boot::transit_to($params{node}."#console","CTRL_C",timeout=>$params{timeout},username=>'diag',password=>'netapp1!');
                    $self->console(cmd=>"\cc",noprompt=>1);
                    ($success, $state) = boot::transit_to($params{node}."#console",$state_map{$params{state}},timeout=>$params{timeout},username=>'diag',password=>'netapp1!');
                } else {
                    $success = 1;
                }
            } else {
                ($success, $state) = boot::transit_to($params{node}."#console",$state_map{$params{state}},timeout=>$params{timeout},username=>'diag',password=>'netapp1!');
            }
            if (!$success) {
                $Trace->warn(msg=>"We failed to reach '$state_map{$params{state}}' becasue:\n'$@'");
                $Trace->comment(msg=>"This failure was on attempt '$attempt' out of '$max_attempts'.");
                ++$attempt;               
            } else {
                if ($params{REBOOT}) {
                    $Trace->comment(msg=>"We succesfully rebooted.");
                    last;
                } else {
                    $Trace->comment(msg=>"We succesfully reached this state: '$state'.");
                    last;
                }
            }
        }
        if ($attempt > 3) {
            $Trace->warn(msg=>"We failed to reach '$state_map{$params{state}}'!");
        }
    }
    $Trace->exit();
}
#end sub agnostic_transit_to

#start sub agnostic_state
sub agnostic_state {
    $Trace->entry();
    my $self = shift;
    my %params = (
          node => undef,
          clustered => undef,
          @_,
          );   
    my $connRec;
    $Trace->debug(msg => "Sub agnostic_state \$params dumper:\n".
                         Dumper(\%params));

	unless (defined($params{clustered})) {
		my $mode = mode_detector($self, %params);
		if ($mode) {
			if ($mode == 10) {
				$params{clustered} = 1;
			} elsif ($mode == 7) {
				$params{clustered} = 0;
			}
		} else {
			$Trace->warn(msg=>"Unable to determine the mode {7 or 10} of '$params{node}'!\n".
							  "Returning state HUNG.");
			$Trace->exit();
			return("HUNG");
		}
	}

    #catch all states and translate them into 10mode transit states if
    #working in 7mode
    my %state_map = (
        FIRMWARE => "CFE",
        BOOT_MENU => "IBM",
        USERNAME => "LOGIN",
        PASSWORD => "LOGIN",
        REBOOT => "REBOOT",
        UP => "CLI",
        CFE_BANNER => "CFE",
        LOADER_BANNER => "CFE",
        SIMLOADER_BANNER => "CFE",
        MAINT => "MAINT",
    );
    my $state;

    #if operating in 10mode use the simple Node::NG::Transit.pm state() method
    if ($params{clustered}) {
        $Trace->comment(msg=>"Running 10mode state call.");
        delete($params{node});
        delete($params{clustered});
        $state = $self->state(%params);
        

    #if operating in 7mode use the boot.pm boot_state function
    } else {
        $Trace->comment(msg=>"Running 7mode state call.");
        $connRec = Tharn::connect("$params{node};#console");
        $state = boot::boot_state($connRec);
        Tharn::disconnect($connRec, '-ok');
        
        eval {
            $self->console(cmd=>"\r", 
                    additional_prompts_timeout => "10",
                    additional_prompts => {
                      '{takeover}>' => "cf giveback" });

        };

        #since boot.pm has no concept of finding an UNKONW state emulate that here
        unless ($state) {
            $Trace->comment(msg=>"We failed to find a valid state. Marking node in an UNKNOWN state.");
            $state = "UNKNOWN";
        } else {
            $Trace->comment(msg=>"We succesfully found '$params{node}' in this 7mode state: '$state'.");
        }
        #now that we have what boot_state() returned translate that in to a 10mode state
        if ($state_map{$state}) {
            $Trace->comment(msg=>"State translation: '$state'=>'$state_map{$state}'");
            $state = $state_map{$state};
        } else {
            $Trace->warn(msg=>"Could not find a translation from 7mode-to-10mode state thus returning 7mode sate!");
        }
        $Trace->comment(msg=>"Returning state '$state'");
    }
    $Trace->exit();
    return($state);
}
#end sub agnostic_state

#start sub mode_detector
sub mode_detector {
    my $connRec;
    $Trace->entry();
    my $self = shift;
    my %params = (
          node => undef,
          @_,
          );   
    $Trace->debug(msg => "Sub mode_detector \$params dumper:\n".
                         Dumper(\%params));
	
	$Trace->comment(msg=>"Trying to determine if '$params{node}' is running in 7 or 10 mode.");
    my $state;
    if ($Tharn::Hosts{$params{node}}->hosttype() =~ /ng/) {
        #trying Node::NG::Transit.pm state() method
        $Trace->comment(msg=>"Checking to see if this is a 10 mode node first.");
        my $mode;
        my %state_params = %params;
        delete($state_params{node});
        delete($state_params{clustered});
        delete($state_params{panic_string});
        $state = $self->state(timeout => 30,
                              maxidle => 15,
                              cr_interval => 2,
                              %state_params,);
        $Trace->comment(msg=>"STATE:'$state'");
        unless ($state =~ /UNKNOWN|HANG/) {
            if ($state =~ /CFE/) {
                eval {
                    $self->console(cmd=>"printenv bootarg.init.boot_clustered",
                        additional_prompts_timeout => "10",
                        noprompt => 1,
                        additional_prompts => {
                            "printenv not found" => "FATAL_PROMPT",
                            "\"true\"" => "FATAL_PROMPT",
                            "Undefined " => "FATAL_PROMPT",
                            "true" => "FINAL_PROMPT",
                            "TRUE" => "FINAL_PROMPT",
                            "1"	=> "FINAL_PROMPT",});
                };
                if ($@ =~ "failed because: printenv not found") {
                    eval {
                        $self->console(cmd=>"show bootarg.init.boot_clustered",
                            additional_prompts_timeout => "10",
                            noprompt => 1,
                            additional_prompts => {
                                "true" => "FINAL_PROMPT",
                                "TRUE" => "FINAL_PROMPT",
                                "1"	=> "FINAL_PROMPT",});
                    };
                    if ($@ =~ "say: Err=TIMEOUT") {
                        $mode = 7;
                    } else {
                        $mode = 10;
                    }
                } elsif ($@ =~ /say: Err=TIMEOUT|failed because: \"true\"/) {
                    $mode = 7;
                } elsif ($@ =~ /failed because: Undefined/) { #most likely a Tricky build
                    $mode = 10,
                } else {
                    $mode = 10;
                }
            } elsif ($state =~ /LOGIN/) {
                eval {
                    $self->console(cmd=>"\n\n",
                        additional_prompts_timeout => "10",
                        noprompt => 1,
                        additional_prompts => {
                            "login:" => "FINAL_PROMPT",});
                };
                if ($@ =~ "say: Err=TIMEOUT") {
                    $mode = 7;
                } else {
                    $mode = 10;
                }
            } elsif ($state =~ /DBLADECLI/) {
                $self->console(cmd=>"\cd", noprompt => 1,);
                eval {
                    $self->console(cmd=>" ",
                               additional_prompts_timeout => "2",
                                   noprompt => 1,
                                   additional_prompts => {
                                #"::" => "FINAL_PROMPT",
                                "::" => "FINAL_PROMPT",});
                };
                if ($@) {
                    $mode = 7;
                } else {
                    $mode = 10;
                }

            } else {
                $mode = 10;
            }
            $Trace->exit();
            $Trace->comment(msg=>"Found '$params{node}' to be a $mode mode node.");
            return($mode);
        } else {
            $Trace->comment(msg=>"'$params{node}' is *not* a 10 mode node.");
        }
    }

    #trying the boot.pm boot_state function
	$Trace->comment(msg=>"Checking to see if this is a 7 mode node now.");
        $connRec = Tharn::connect("$params{node};#console");
        $state = boot::boot_state($connRec,timeout => 30,maxidle => 15);

	if ($state) {
		$Trace->exit();
		$Trace->comment(msg=>"Found '$params{node}' to be a 7 mode node.");
                Tharn::disconnect($connRec, '-ok');
		return(7);	
	} else {
		$Trace->warn(msg=>"Unable to determine the mode '$params{node}' is running in!");
                Tharn::disconnect($connRec, '-ok');
		return(0);
	}
}
#end sub mode_detector

# NFS_GATHER logs from the requested filer
# To test the functionality of nfs_gather(). Select Action as Intialization Only and specify gx_boot_NO_ADD_TASKS=1 
# as ntest parameters. This will drop the nodes in 7 mode.
 
sub gsi_nfs_gather {
    my $self = shift;
    my %params = (
        dest_server => undef,
        dest_user   => "root",
        mount_ip    => undef,
        delete_on_xfer => undef,
        timeout     => 600,
        server_ref   => undef,
        expect_core  => undef,
        operation    => undef,
        archive_age => 14,
        tar_archive_dir => 0,
        tar_exclude_dirs => undef,
        @_);

    $Trace->comment(msg=>"Begining the gsi_nfs_gather sub routine.");
    ##$Trace->debug(msg=>"Dumper of \%params:\n".Dumper(\%params));
    my $rc = undef;
    my $cmd;
    my $match;

    my $dest_dir=Testinfo::dir();
    if (! defined($dest_dir) ) {
        $Trace->fail(msg=>"unable to get a directory to write to.  Check the logs from Testinfo::dir() to see why.");
        goto done;  #not sure if the fail will kill us or if I have to goto? mg
    }
    $Trace->comment(msg=>"dir() from Testinfo gave us this directory '$dest_dir'.");

    my $mroot = 1;   # assume mroot is mounted
    my $local_dir = "/mnt/$params{server_ref}->{server}";

    # create a connection to the client we will mount from
    my $h = main::host("$params{dest_server}");
    my $client = main::connect("#ssh");
    
    # mount the filer
    main::say($client,"sudo mkdir -p $local_dir");
    $match = main::say($client,"sudo mount $params{mount_ip}:/ $local_dir", '-wait', "failed", '-eval_error');
    if ($match eq "failed") {
        #some 7G tests apparently are too agressive in what the cleanup
        #we have seen cases where the mroot is not longer exported via 7mode
        #if the mount failed assume that is what happened and try and correct
        $Trace->warn(msg=>"Initial mount to '$params{mount_ip}' failed!  Let's go export vol0 on the filer and try again.");
        $params{server_ref}->console(cmd => "exportfs", noprompt => 1);
        $params{server_ref}->console(cmd => "exportfs -p anon=0,nosuid /vol/vol0", noprompt => 1);
        $params{server_ref}->console(cmd => "exportfs", noprompt => 1);
        main::say($client,"sudo mount $params{mount_ip}:/ $local_dir");
    }

    #If you are modifying this list, know that it will only affect 7-mode runs.  To 
    #add the same exclusion to c-mode, you need to go to lib/Node/NG/Misc.pm and look at gsi_gather
    my $excludes =  "--exclude etc/swapfile " .
                    "--exclude etc/.snapshot " .
                    "--exclude etc/java " .
                    "--exclude etc/.onboot/java " .
                    "--exclude etc/zoneinfo " .
                    "--exclude etc/http " .
                    "--exclude etc/lang " .
                    "--exclude etc/templates " .
                    "--exclude etc/man " .
                    "--exclude etc/shelf_fw " .
                    "--exclude etc/disk_fw " .
                    "--exclude etc/firmware " .
                    "--exclude etc/clihelp ".
                    "--exclude etc/acpp_fw ".
		    "--exclude etc/sp_fw ".
                    "--exclude etc/cluster_config/mdb ".
                    ""; 

    # Don't tar archive dir unless told to.
    if ( !$params{tar_archive_dir} ) {
        $excludes .= " --exclude oldcores --exclude oldkcores " 
    }

    # exclude anything else the caller doesn't want.
    if (defined($params{tar_exclude_dirs}) && scalar(@{$params{tar_exclude_dirs}})) {
        if (ref($params{tar_exclude_dirs}) ne 'ARRAY') {
            $Trace->fatal(msg=>"tar_exclude_dir was not an array reference.", result=>'SCRIPT');
        }

        $excludes .= " --exclude ".join(" --exclude ", @{$params{tar_exclude_dirs}})." ";
    }

    # Create the directory paths for:
    #   *testinfo/$node
    #   *testinfo/$node/vmware (if a vsim)
    #
    Testinfo::make_open_dir("$dest_dir/$self->{server}");
    Testinfo::make_open_dir("$dest_dir/$self->{server}/mroot");
    Testinfo::make_open_dir("$dest_dir/$self->{server}/sim");
    Testinfo::make_open_dir("$dest_dir/$self->{server}/vmware") if ($params{server_ref}->{server}=~/vsim/);

    $dest_dir .= "/$self->{server}";
    
    # Verify the target dir has been created
    $cmd = "ls $dest_dir";    
    $match = undef;
    $match = main::say($client,$cmd,'-wait',"Permission denied",'-eval_error','-timeout',$params{timeout});
    if ($match =~ /Permission denied/) {
        $Trace->warn(msg=>"Permission error writing to destination directory: $dest_dir");
        goto done;
    } else {
        $cmd = "sudo chmod -R 777 $dest_dir";
        eval {
            main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
        };
        if ($@) {
            $Trace->warn(msg=>"Error trying to chmod the files");
        }
    }

    my $tar_files = "etc";
    my $tar_ball_path = $dest_dir;
    my $tar_ball = "testInfo.tar.gz";


    sub nodename2simdir {
        my $nodename=shift;
        if ($nodename=~/vsim/) {
            my ($user,undef,$id) = split /[_-]/,$nodename;
            return "/u/$user/vsims/${user}_vsim_$id";
        } else {
            return 0;
        }
    }

    if (my $simdir=nodename2simdir($params{server_ref}->{server})) {
        my $vmware_dir="$dest_dir/vmware";
        my $sudo_cmd = Tharn::param('SUDO_PATH', '-mesg','(NATE is expected to have set this)');
        my $command="$sudo_cmd -u $params{dest_user} cp $simdir/vmware*.log $vmware_dir";
        $Trace->comment(msg=>$command);
        system $command;

        $command="$sudo_cmd -u $params{dest_user} cat /dev/null > $simdir/vmware*.log";
        $Trace->comment(msg=>$command);
        system $command;

        $Trace->comment(msg=>"Copying vmdkstats.log..");
        my $command="$sudo_cmd -u $params{dest_user} cp $simdir/vmdkstats.log $vmware_dir";
        $Trace->comment(msg=>$command);
        system $command;
    }



    my $corelocation="$local_dir/etc";
    my @coredirs = qw(crash mcores);
    my $kcoredir = "crash";
    if ($self->{host_object}->{MODEL} eq 'VSIM' ) {
        $corelocation="$local_dir/sim";
        @coredirs = qw(cores kcores);
        $kcoredir = "kcores";
        $tar_files .= " sim";

	my $server_name = $params{server_ref}->{server};
        eval {
		$Trace->comment(msg=>"I am here which means now I am going to try telnet");
                my $num_retries = 3;
		my $current_attempt = 0;		
                main::host($server_name);
    		my $c = main::connect("#telnet");
		
		# Go to the system shell and list all the cores
		my $output;  
		my $line ;
		(undef, undef, $output, $line) = main::say($c, "priv set advanced", '-prompt', "$server_name*");
		
		sleep(10);
		while ($current_attempt <= $num_retries) {
        		eval {
                		(undef, undef, $output, $line) = main::say($c, "systemshell", '-prompt', "login*", "login:*", "login: *", "-timeout", 120);
	                	(undef, undef, $output, $line) = main::say($c, "\n", '-prompt', "login*", "login:*", "login: *");
        	        	(undef, undef, $output, $line) = main::say($c, "diag", '-prompt', "Password*");
                		(undef, undef, $output, $line) = main::say($c, "$self->{SHELL_PASSWORD}", '-prompt', "$server_name", "-timeout", 120);
        		};

        		if($@) {
                		$Trace->comment(msg=>"Failed attempt $current_attempt of $num_retries");
                		$current_attempt++;
        		} else {
                		last;
        		}
               }

               my @core_files = ();
               GET_CORE:
               while (1) {
                 $Trace->comment(msg=>"Here we are checking for kernel and ucores, if they are in middle of saving we need to wait");

                  (undef, undef, $output, $line) = main::say($c, "ls -lh /sim/kcores /sim/cores /sim/crash /mroot/ucores /mroot/etc/crash | awk '{print \$9}'", '-prompt', "$server_name\%");
                  @core_files = split(/\n/, $output);

                  foreach my $core (@core_files) {
                          if($core =~ /saving/) {
                                  $Trace->comment(msg=>"We have cores in middle of saving so we need to wait");
                                  sleep 60;
                                  @core_files = ();
                                  next GET_CORE;
                           }
                  }
                  $Trace->comment(msg=>"We have no cores in middle of saving so we will not wait");
                  last GET_CORE;
                }

       	};

        $tar_files .= " $dest_dir/sim/kcores";
        $tar_files .= " $dest_dir/sim/cores";
        $tar_files .= " $dest_dir/sim/crash";
	$tar_files .= " $dest_dir/sim/mroot";

        #$tar_ball_path = '/sim';
        $excludes .= " --exclude sim/dev".
                     " --exclude sim/testInfo.tar.gz".
                     " --exclude sim/bin ".
                     " --exclude sim/scratch ".
                     " --exclude sim/updates ";
    }

    if ( $mroot ) {
        # Look to see if we're in the middle of grabbing a kernel core
        $cmd = "ls -l $corelocation/$kcoredir/core.saving.nz";
        my (undef, undef, $after) = main::say($client,$cmd,'-timeout',$params{timeout});
	my (@fields) = split(/\s+/, $after);

        # We'll wait for 13 or so minutes to have the kcore created, longer for a 6070
        $Trace->debug(msg=>"The host_object is: " . Dumper($self->{host_object})); 
        my $max_attempts = 10;
        my $cur_attempts = $max_attempts;
        my $temp_kcore_size = 0;
        
        if ( $fields[4] =~ /\d+/ && $fields[8] =~ /core/ ) {
            GET_KCORE: {
                while ( $cur_attempts-- ) {
                    # We found a core file getting saved.  Sleep for a few to give the saving
                    # process some time.
                    sleep 60;
                    main::say($client,"ls -l $corelocation/$kcoredir",'-timeout',$params{timeout});
                    (undef, undef, $after) = main::say($client,$cmd,'-timeout',$params{timeout});
                    # -rw-------  1 root  wheel  285308474 Jun  9 10:28 /mroot/kcores/core.saving.nz
        	    (@fields) = split(/\s+/, $after);
        	    if ( $fields[4] =~ /\d+/ && $fields[8] =~ /core/ ) {
                        if ($temp_kcore_size < $fields[4]) {
                            $cur_attempts = $max_attempts;
                            $temp_kcore_size = $fields[4];
                            $Trace->comment(msg=>"The core is not done saving, but it is still making progress.");
                        }
                        $Trace->comment(msg=>"Temp kcore still exists, attempts left: $cur_attempts");
                    } else {
                        $Trace->debug(msg=>"Can't find the temp kcore, assume complete");
                        last GET_KCORE;
                    }           
                }
                $Trace->warn(msg=>"Still found a core.saving.nz file, not waiting any longer.");
            }
        }

        # We were told to look for a kcore file, make sure we have one, warn otherwise.
        if ( $params{expect_core} ) {
            $cmd = "ls -l $corelocation/core.*";    
            $match = undef;
            $match = main::say($client,$cmd,'-wait',"No such file or directory",'-eval_error','-timeout',$params{timeout});
            if ($match =~ /No such file or directory/) {
                $Trace->warn(msg=>"Expected to see a kcore, but didn't, continuing.");
            }
        }

        # Tar up /mroot, /var/log/messages
        $cmd = "cd $local_dir ; sudo tar $excludes -vzcf $tar_ball_path\/$tar_ball $tar_files";
        eval {
            main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
        };
        if ($@) {
            $Trace->warn(msg=>"Error trying to create tarball.");
        }

        # Drive on, even if we had a tar error to try to get as much as we can off of the box.
        foreach my $dir (@coredirs) {
            # Create Archive Directory if it doesn't exist
            $cmd = "sudo mkdir -p $corelocation/old$dir";
            eval {
                main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
            };
            if ($@) {
                $Trace->warn(msg=>"Error while creating directory: $corelocation/old$dir/");
            }

            $cmd = "ls $corelocation/old$dir/;";
            $match = undef;
            $match = main::say($client,$cmd,'-wait',"No such file or directory",'-eval_error','-timeout',$params{timeout});
            if ($match =~ /No such file or directory/) {
                $Trace->warn(msg=>"$corelocation/old$dir/ was not found. Skipping Archive.");
                next;
            }

            # delete any files in archive older than archive_age
            $cmd = "find $corelocation/old$dir -type f -mtime +$params{archive_age} " .
            "-execdir rm '{}' ';'";
            eval {
                main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
            };
            if ($@) {
                $Trace->warn(msg=>"Error while attempting to find/remove old files from $corelocation/old$dir/");
            }

            # Move files to archive directory 
            $cmd = "find $corelocation/$dir/ -type f -execdir mv '{}' $corelocation/old$dir/ ';'";
            eval {
                main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
            };
            if ($@) {
                $Trace->warn(msg=>"Error while moving files from $corelocation/$dir/ to " . 
                                  "$corelocation/old$dir/");
            }

            if ( $dir eq "cores" ) {
                # Move backtrace files to archive directory
                $cmd = "find $corelocation/ -type f -maxdepth 1 -name '*.backtrace.log'" .
                " -execdir mv '{}' $corelocation/oldcores/ ';'";
                eval {
                    main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
                };
                if ($@) {
                    $Trace->warn(msg=>"Error while attempting to find/move backtrace log files");
                }
            }
        }
        
        # Untar  /mroot, /var/log/messages, ignore errors
        $cmd = "cd $dest_dir\/mroot ; sudo tar -zxof $dest_dir\/$tar_ball; sudo chmod -R 777 .";
        eval {
            main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
        };
        if ($@) {
            $Trace->warn(msg=>"Error untaring testInfo tarball, continuing.");
        }
    }
    
    if ( $params{delete_on_xfer} && $mroot ) {
                $Trace->debug(msg=>"Clean up the tarballs after transfer");
		$cmd = "sudo rm -fr $dest_dir\/$tar_ball";
                eval {
	            main::say($client,$cmd,'-eval_error','-timeout',$params{timeout});
        	};
       		if ($@) {
            		$Trace->warn(msg=>"Error deleting testInfo tarball.");
        	}

    }
    


    
    # if we haven't blown up by now, everything went well
    $rc = 1;

    done:
    main::say($client,"pwd");
    main::say($client,"sudo umount $local_dir");
    $Trace->exit(msg=>"Exited with rc: $rc");
    return $rc;
}

1;

__END__



=pod

=head1 NAME

Node.pm - remote programmatic administration of a Filer.

=head1 DESCRIPTION

This Perl module is meant to be programmatic way of administering a Filer.

=head1 PARAMS

=over

=item - SHELL_USER

    spec - s
    description - the name to use when logging into a server shell

=item - SHELL_PASSWORD

     spec - s
     description - the password to use when logging into a server shell

=item - CLI_USER

     spec - s
     description - the name to use when logging into the server CLI

=item - CLI_PASSWORD

     spec - s
     description - the password to use when logging into a server CLI

=item - cmdnode

     spec - s@
     description - the management nodes to configure

=item - skipcmdnode

     spec - s@
     description - the management nodes to not skip during configuration

=item - CLI_PARAMETER_VALIDATION

     description - Should the object validate the parameters prior to sending them to the CLI
     default - 0

=item - NATE_LIB

     description - The list of nate directories to search

=item - CDEFS_DIRECTORY

     spec - s
     description - The directory to find the cdefs

=item - TOPOLOGY

     spec - s@
     description - The file that defines the topology

=item - timeout

     description - The timeout value (in seconds) to be used for CLI connections and CLI commands

=item - maxidle

     description - The maximum amount of time to allow a CLI connections or CLI command to remain idle

=item - CMDPROF

     description - given a file will enable cli command profiling to dump hash block to the file, 
     default is undef.
     Time values are stored as epoch seconds
     Hash block example
     {
          'server'      => 'rtpgx-3050-15',      # name of filer
          'begin'       => 1213733229,           # begin time in epoch seconds
          'end'         => 1213733232,           # end time in epoch seconds
          'cmdtime'     => 3,                    # calculated time (end-begin)
          'begin_micro' => '1213733229.92422',   # begin time in epoch seconds and microseconds
          'end_micro'   => '1213733232.09315',   # end  time in epoch seconds and microseconds
          'cmd'         => 'storage disk show',  # filer command for this profile
          'args'        => '',                   # command arguments
     };
     xxEND_RECORDxx


=item - CMDPROF_CSV

     description - given a file will enable cli command profiling to dump values to the csv file
     default is undef.
     CSV fields - server,begin,end,cmdtime,begin_micro,end_micro,cmd,args
     Note that the last field may (command arguments) may contain ','.
     Account for this in your parsing of the csv file.
     See CMDPROF for description of values

=item - CMDPROF_RT

     description - if non-zero, will keep N amount of profiles in the server object, default = 0
     All profiles are stored in a list in the server object under the 'cmdprof' key.
     Each profile is an array element, most recent profile is the 0th element of the array.
     The array only keeps N elements as defined by the CMDPROF_RT param. Each element is a hash
     block, as exampled below.
     See CMDPROF for description of values
     $server->{'cmdprof'}[0]{key} = value
     {
          'server'      => 'rtpgx-3050-15', 
          'begin'       => 1213733229,
          'end'         => 1213733232,
          'cmdtime'     => 3,
          'begin_micro' => '1213733229.92422',
          'end_micro'   => '1213733232.09315',
          'cmd'         => 'storage disk show',
          'args'        => '',
     };

=item - WARN_TIMEOUT

     description - Take the timeout value, double it, and then WARN at the origional timeout and FAIL on the double of it
     default - 1

=item - TIMEOUT_MULTIPLIER

     description - Take the timeout value, multiply it by this value and use that as the new timeout on all say_wrapper calls
     default - 1

=item - SAY_SWALLOW

     description - Flag indicating that '-swallow' should be passed to say() (except on console)
     spec - boolean
     default - 1

=item - SKIP_MROOT_AGGREGATE

     description - Indicates that aggregates containing the mroot should be skipped
     spec - boolean
     default - 0

=item - POKE_CLI

     spec - boolean
     description - Flag to indicate whether or not to poke (i.e. send an extra carriage return to) the CLI
     default - 0

=back

=head1 AUTHOR

  Bill snider (bill.snider@netapp.com)



