#***********************************************************************
# This is the comment section for tharndoc
#***********************************************************************
## @summary Enables remote programmatic administration of a Filer running
##          Ontap
## @author dl-central-libs
## @pod here


=head1 NAME

Classic.pm

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

This module is meant to be programmatic way of administering Classic

=head1 SYNOPSIS

This module is instantiated by Server.pm and is not directly called.
See documentation below for help on individual methods 

=head1 EXAMPLES

No examples needed

=cut

#***********************************************************************

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDECLASSIC}) {
    my $newModule = $ENV{OVERRIDECLASSIC};
    delete $ENV{OVERRIDECLASSIC};
    print "The main Node::Classic module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Node::Classic;
use strict;
use base qw(Node);
use Data::Dumper;
use Options;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

use TracePkg;
my $Trace = new TracePkg();
use Carp qw(confess);

my @Classic_modules = qw(Aggr Arp ASUP Backup Cf Config Date Df Disk DiskFWUpdate DNS Download Echo
			 EMS Environment Exportfs Fcp Fcstat File Filestats Fpolicy Ftp
			 Halt Help Hostname Httpstat Ifconfig Ifstat Igroup Ipsec Ipspace
			 Iscsi VFiler Qtree Vol FlexCacheVol); 

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
	server => undef,
	server_object => undef,
	Trace => $Trace,
	@_,
	};

    bless($self, $class);

    $self->{Trace}->entry(msg=>"Entered for $self->{server}\n");

    if (!$self->{server}) {
	$self->{Trace}->fatal(msg=>"A server must be passed in.", result=>'SCRIPT');
    }

    $self->{host_object} = main::host($self->{server});
    $self->{options_object} = new Options;
    $self->{type} = $self->ontap_classic();

    # create a connections hash which is required by the establish_connections routine. 
    #  go ahead an establish a cli connection since it is needed to get the version
    $self->{connections} = {};

    if (!$self->{no_cli}) {
        $self->establish_connections(cli_connection=>1);
    }

    $self->_define_cmd_nodes();
    $self->_define_modules();

    # The following calls will load the hand-coded cdef modules. This
    # is a hack because we should have a better mechanism for modules
    # being loaded to be defined in Node.pm and not in the children...
    $self->_setup_commands("Node", undef, $self->_common_modules());
    $self->_setup_commands("Node::Classic", undef, @Classic_modules);
    $self->_setup_commands($class, undef, @{$self->{modules}});

    $self->_modify_commands();
    # turn on the more strict CLI validation for this object;
    $self->{cli_parameter_validation} = 1;

  done:
    $self->{Trace}->exit(msg=>"Exited for $self->{server} with $self\n");
    return $self;
}

sub establish_connections {
    my $self = shift;
    my %params = (
		  console_connection =>undef,
		  cli_connection => undef,
		  dblade_connection => undef,
		  rlm_connection => undef,
		  abortive => 1,
		  @_);
    my $connections = $self->{connections};

    $self->{Trace}->entry(msg=>join(" ", @_));

    if ($params{console_connection}) {
	# try to setup and create the console connection
	$connections->{console_connection}{conn_specs} = "$self->{server}\#console ; conntype=pm_cons ; shell_type=sh";
	$self->open_connection(type=>"console_connection", abortive=>$params{abortive});
    }
    my $ip = $self->{host_object}->default_ip();
    
    if ($params{cli_connection}) {
	# setup and create the cli connection
	$connections->{cli_connection}{conn_specs} = "$self->{server}\#cli ; conntype=rsh-persay ; connhost=$ip";
	$connections->{cli_connection}{conn_specs} .= " ; username=root";
#	$connections->{cli_connection}{conn_specs} .= " ; prompts='$self->{server}#'";
	$self->open_connection(type=>"cli_connection", abortive=>$params{abortive});
#	$connections->{cli_connection}{errors} = ["ERROR:", "There are no entries matching your query",
#						  "This table is currently empty", "Usage:"];
#	$connections->{cli_connection}{object}->prompts(["$self->{server}::>","$self->{server}::\\*>"]);
	# call _get_version_info prior to loading the commands because the build information is a prequisite
	$self->_get_version_info();
    }
    
    if ($params{shell_connection}) {
	$connections->{dblade_connection} = $connections->{cli_connection};
    }
    
    if ($params{dblade_connection}) {
	$connections->{dblade_connection} = $connections->{cli_connection};
    }
    
    if ($params{rlm_connection}) {
	# setup and create the rlm connection
	$connections->{rlm_connection}{conn_specs} = "$self->{server}\#rlm ; shell_type=none";
	$self->open_connection(type=>"rlm_connection", conntype=>"ssh-persay", abortive=>$params{abortive});
    }
    $self->{Trace}->exit();
}

# this method is called after all of the cdefs have been read in from the files.
sub _modify_commands {
    my $self = shift;
    my %params = @_;
    $self->{Trace}->entry();
    $self->{Trace}->exit();
}

sub compare_version {
    my $self = shift;
    my %params = (release=>undef,
		  build=>undef,
		  operator=>">=",
		  @_);
    my $rc = 1;
    $self->{Trace}->entry(msg=>"release=$params{release}, build=>$params{build}, operator=>'$params{operator}'");

    if ($params{release}) {
	$rc = version_compare($self->{release}, $params{release});# 1 if v1 > v2, 0 if equal, -1 if v1 < v2
	if ($params{operator} eq ">=" && ($rc < 0) ) {
	    $rc = undef;
	} elsif ($params{operator} eq "<" && $rc < 0) {
	    $rc = undef;
	}
    }
    $self->{Trace}->exit(msg=>"rc=$rc");
    return $rc;
}

sub _get_version_info {
    my $self = shift;
    my %params = (server=>$self->{server},
		  @_);
    $self->{build} = "unknown";
    $self->{release} = "unknown";
    $self->{Trace}->entry(msg=>"server=$params{server}");

    eval "use Filer::Version qw(get_ontap_version);";
    die $@ if $@;

    $self->{release} = $self->{options_object}->get(option=>"RELEASE") || get_ontap_version($self->{host_object});
    if ($self->{release} ne "unknown") {
	$self->{options_object}->set(RELEASE=>$self->{release});
    }
    $self->{Trace}->exit(msg=>"build=$self->{build}, release=$self->{release}");
}

sub _validate_cli_parameters {
    my $self = shift;
    my %params = (definition => undef,
		  cmd_options => undef,
		  cmd_params => undef,
		  input_params => undef,
		  @_);
    
    $self->{Trace}->entry();
    my %definition = %{$params{definition}};
    my %cmdoptions = %{$params{cmd_options}};
    my $input_params = $params{input_params};
    %params = %{$params{cmd_params}};
    my $command = "";
    ################################################################################
    # Now fill out the rest of the command with the option and paramaters
    # We will complain if the required option does not exist in the cmdOption hash
    # We will also complain if the required paramater is not defined
    #
    # In ONTAP, optional options must preceed the required params
    #
    # Also in ONTAP, there are alot of flags that do not have associated values
    # with them. In this case, if the value of the option is blank, just print the
    # flag
    #
    # Also, ONTAP has several commands that use options after the required options.
    # These will be referred to as post_command_options
    #
    # Check for mutual exclusions. ONTAP has several commands where you can choose
    # from one of many options, but may only chose one of those options. Mutexs
    # should be setup in the individual command.pm files and checked for on 
    # optional, required, and post command options
    #
    # Sometimes, an ONTAP command uses a keyword, rather than a hyphenated flag.
    # In these cases, check if cmdOption starts with "-". If it does, include
    # value, otherwise, just print keyword
    #
    # Since parameter values can be "0", must check that param is defined rather
    # than that it contains a value. If we just check that it contains a value,
    # than conditional will fail if value is "0"
    ################################################################################
    foreach my $option (@{$definition{optional_options}}) {
	if (defined($cmdoptions{$option}) && defined($params{$option})) {
	    ################################################################################
	    # Check that value is supported if only certain values are supported for this
	    # option
	    ################################################################################
	    my @supported = ($definition{supported_values}{$option}) ? (@{$definition{supported_values}{$option}}) : ();
	    if (scalar(@supported > 0)) {
		my $valid = 0;
		foreach my $supportedvalue (@supported) {
		    if ($params{$option} eq $supportedvalue) {
			$valid = 1;
		        last;
		    }
	        }
		if (!$valid) {
		    $self->{Trace}->fatal(msg=>"$option contains invalid value $params{$option}. Supported values are: @supported", result=>"SCRIPT");
		}
	    }
	    my %dependency = ($definition{dependencies}->{$option}) ? %{$definition{dependencies}->{$option}} : ();
            if (scalar(keys(%dependency)) ) {
                foreach my $dependant (keys(%dependency)) {
                    if ( $params{$dependant} ) {
                        if ($dependency{$dependant} ne "presence") {
                            if ($params{$dependant} eq $dependency{$dependant} ) {
#				$self->{Trace}->debug(msg=>"$option depends on $dependant having value $dependency{$dependant}. This was met");
                            } else {
				$self->{Trace}->fatal(msg=>"$option depends on $dependant having value $dependency{$dependant}. This was not met", result=>"SCRIPT");
                            }
                        } else {
#			    $self->{Trace}->debug(msg=>"$option does not require $dependant to have a certain value");
                        }
                    } else {
			$self->{Trace}->fatal(msg=>"$option has a dependency on $dependant which was not specified", result=>"SCRIPT");
                    }
                }
            }
            if ($params{$option} ne "false") {
                my @mutex = ($definition{mutual_exclusions}{$option}) ? @{$definition{mutual_exclusions}{$option}} : ();
                if (scalar(@mutex) > 0) {
                    foreach my $mutexop (@mutex) {
                        if ($params{$mutexop} && $params{$mutexop} ne "false") {
			    $self->{Trace}->fatal(msg=>"$option and $mutexop are mutually exclusive", result=>"SCRIPT");
                        }
                    }
                }
            }
	    $input_params->{$option} = $params{$option} if ($input_params);
	    if ( ($cmdoptions{$option} =~ /^-/) && ($params{$option} eq "true") ) {
		$command .= " $cmdoptions{$option}";
	    } elsif ( ($cmdoptions{$option} =~ /^-/) && ($params{$option} eq "false") ) {
		next;
	    } elsif (($params{$option} eq "true") || ($params{$option} eq "false")) { 
		$command .= " $cmdoptions{$option}";	
	    } elsif ($cmdoptions{$option} eq " ") {
		$command .= " $params{$option}";
	    } else {
		$command .= " $cmdoptions{$option} $params{$option}";
	    }
	} elsif (!exists($cmdoptions{$option}) && !(grep($option ne $_, qw(Wait Verify)))) { # hack
	    $self->{Trace}->fatal(msg=>"Optional '$option' does not exist as a Parameter", result=>"SCRIPT");
	}
    }
    foreach my $option (@{$definition{required_options}}) {
	if (defined($cmdoptions{$option}) && defined($params{$option})) {
	    ################################################################################
	    # Check that value is supported if only certain values are supported for this
	    # option
	    ################################################################################
	    my @supported = ($definition{supported_values}{$option}) ? (@{$definition{supported_values}{$option}}) : ();
	    if (scalar(@supported > 0)) {
		my $valid = 0;
		foreach my $supportedvalue (@supported) {
		    if ($params{$option} eq $supportedvalue) {
			$valid = 1;
		        last;
		    }
	        }
		if (!$valid) {
		    $self->{Trace}->fatal(msg=>"$option contains invalid value $params{$option}. Supported values are: @supported", result=>"SCRIPT");
		}
	    }
	    my %dependency = ($definition{dependencies}->{$option}) ? %{$definition{dependencies}->{$option}} : ();
            if (scalar(keys(%dependency))) {
	        foreach my $dependant (keys(%dependency)) {
		    if ( $params{$dependant} ) {
	                if ($dependency{$dependant} ne "presence") {
		            if ($params{$dependant} eq $dependency{$dependant} ) {
#				$self->{Trace}->debug(msg=>"$option depends on $dependant having value $dependency{$dependant}. This was met");
		            } else {
				$self->{Trace}->fatal(msg=>"$option depends on $dependant having value $dependency{$dependant}. This was not met", result=>"SCRIPT");
		            }
	                } else {
#			    $self->{Trace}->debug(msg=>"$option does not require $dependant to have a certain value");
		        }
	            } else {
			$self->{Trace}->fatal(msg=>"$option has a dependency on $dependant which was not specified", result=>"SCRIPT");
                    }
		}
	    }
            if ($params{$option} ne "false") {
                my @mutex = ($definition{mutual_exclusions}{$option}) ? @{$definition{mutual_exclusions}{$option}} : ();
                if (scalar(@mutex) > 0) {
                    foreach my $mutexop (@mutex) {
            	        if ($params{$mutexop} && $params{$mutexop} ne "false") {
			    $self->{Trace}->fatal(msg=>"$option and $mutexop are mutually exclusive", result=>"SCRIPT");
			}
		    }
	    	}
	    }
        } elsif (!$cmdoptions{$option}) {
	    $self->{Trace}->fatal(msg=>"Required '$option' does not exist as a Parameter", result=>"SCRIPT");
	#######################################################################################
	# If a required parameter was not specified, see if one its mutually exclusive
	# pairs was specified. If so, skip the fact the this param was not specified. If no
	# mutex pairs, or no mutex partners were specified, fail
	#######################################################################################
	} else {
	    my @mutex = ($definition{mutual_exclusions}{$option}) ? @{$definition{mutual_exclusions}{$option}} : ();
            if (scalar(@mutex) > 0) {
		my $found=0;
                foreach my $mutexop (@mutex) {
                    if ($params{$mutexop} && $params{$mutexop} ne "false") {
#			$self->{Trace}->debug(msg=>"$option was not specified but $mutexop was and the two are mutually exclusive");
			$found=1;
			last;
                    }
                }
		if (!$found) {
		    $self->{Trace}->fatal(msg=>"$option (or one of its mutex partners: @mutex) is a required argument", result=>"SCRIPT");
		} else {
		    next;
		}
            } else {
		$self->{Trace}->fatal(msg=>"$option is a required argument", result=>"SCRIPT");
	    }
        }
	$input_params->{$option} = $params{$option} if ($input_params);
        if ( ($cmdoptions{$option} =~ /^-/) && ($params{$option} eq "true") ) {
            $command .= " $cmdoptions{$option}";
        } elsif ( ($cmdoptions{$option} =~ /^-/) && ($params{$option} eq "false") ) {
            next;
        } elsif (($params{$option} eq "true") || ($params{$option} eq "false")) {
            $command .= " $cmdoptions{$option}";
	} elsif ($cmdoptions{$option} eq " ") {
            $command .= " $params{$option}";
        } else {
            $command .= " $cmdoptions{$option} $params{$option}";
        }
    }
    foreach my $option (@{$definition{post_command_options}}) {
        if (defined($cmdoptions{$option}) && defined($params{$option})) {
	    ################################################################################
	    # Check that value is supported if only certain values are supported for this
	    # option
	    ################################################################################
	    my @supported = ($definition{supported_values}{$option}) ? (@{$definition{supported_values}{$option}}) : ();
	    if (scalar(@supported > 0)) {
		my $valid = 0;
		foreach my $supportedvalue (@supported) {
		    if ($params{$option} eq $supportedvalue) {
			$valid = 1;
		        last;
		    }
	        }
		if (!$valid) {
		    $self->{Trace}->fatal(msg=>"$option contains invalid value $params{$option}. Supported values are: @supported", result=>"SCRIPT");
		}
	    }
	    my %dependency = ($definition{dependencies}->{$option}) ? %{$definition{dependencies}->{$option}} : ();
            if (scalar(keys(%dependency))) {
                foreach my $dependant (keys(%dependency)) {
                    if ( $params{$dependant} ) {
                        if ($dependency{$dependant} ne "presence") {
                            if ($params{$dependant} eq $dependency{$dependant} ) {
#				$self->{Trace}->debug(msg=>"$option depends on $dependant having value $dependency{$dependant}. This was met");
                            } else {
				$self->{Trace}->fatal(msg=>"$option depends on $dependant having value $dependency{$dependant}. This was not met", result=>"SCRIPT");
                            }
                        } else {
#			    $self->{Trace}->debug(msg=>"$option does not require $dependant to have a certain value");
                        }
                    } else {
			$self->{Trace}->fatal(msg=>"$option has a dependency on $dependant which was not specified", result=>"SCRIPT");
                    }
                }
            }
            if ($params{$option} ne "false") {
		my @mutex = ($definition{mutual_exclusions}{$option}) ? @{$definition{mutual_exclusions}{$option}} : ();
                if (scalar(@mutex) > 0) {
                    foreach my $mutexop (@mutex) {
                        if ($params{$mutexop} && $params{$mutexop} ne "false") {
			    $self->{Trace}->fatal(msg=>"$option and $mutexop are mutually exclusive", result=>"SCRIPT");
                        }
                    }
                }
            }
	    $input_params->{$option} = $params{$option} if ($input_params);
            if ( ($cmdoptions{$option} =~ /^-/) && ($params{$option} eq "true") ) {
                $command .= " $cmdoptions{$option}";
            } elsif ( ($cmdoptions{$option} =~ /^-/) && ($params{$option} eq "false") ) {
                next;
            } elsif (($params{$option} eq "true") || ($params{$option} eq "false")) {
                $command .= " $cmdoptions{$option}";
            } elsif ($cmdoptions{$option} eq " ") {
                $command .= " $params{$option}";
            } else {
                $command .= " $cmdoptions{$option} $params{$option}";
            }
        } elsif (!$cmdoptions{$option}) {
	    $self->{Trace}->fatal(msg=>"Postcommand '$option' does not exist as a Parameter", result=>"SCRIPT");
	} else {
#	    $self->{Trace}->debug(msg=>"$option was not specified");
	}
    }
    $self->{Trace}->exit(msg=>"return '$command'");
    return $command;
}		  

sub vserver_show {
    my $self = shift;
    
    $self->{Trace}->entry();
    $self->{Trace}->exit();
    return [{
	vserver => "N/A",
	}];
}    

sub convert_export_path {
    my $self = shift;
    my %params = (export => undef,		  
		  volume => undef,
		  aggregate => undef,
		  @_);
    my $export;

    $self->{Trace}->entry(msg=>"export=$params{export}, volume=>$params{volume}");
    my @aggrs = $self->{server_object}->search_topology(cmd_node=>"aggregate", key=>"aggregate", value=>$params{aggregate});
    my $aggrh = $aggrs[0];
    if ($aggrh->{type} eq "trad") {
	$export = "/vol/$params{aggregate}/$params{volume}";
    } else {
	$export = "/vol/$params{volume}";
    }
    $self->{Trace}->exit(msg=>"return $export");
    return ($export, $export);
}

# this routine is called from Node::_setup_commands() and needs to be defined in this namespace
# so that the routines found inside the libraries being loaded will also be in this namespace
sub _load_modules {
    my ($self, $class, @modules) = @_;
    $Trace->entry(msg => "class => '$class' modules => '@modules'");
    no strict 'refs';

    foreach my $module (@modules) {
        eval "use ${class}::$module";
        if ($@) {
            $Trace->fatal(msg => "Failed to load module '${class}::$module' '$@'", result => "SCRIPT");
        }
        my $setup = "${class}::Setup${module}Commands";
        $Trace->debug(msg => "Attempting to access '$setup'");
        if (!&$setup($self)) {
            $Trace->trace(msg=>"Could not find the method '$setup'");
        }
    }
    $Trace->exit();

}

1; # Do NOT remove this. It is needed by the package.

