#***********************************************************************
# This is the comment section for tharndoc
#***********************************************************************
## @summary Enables remote programmatic administration of a Filer running
##          7.X
## @author dl-central-libs
## @pod here

=head1 NAME

7_X.pm

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

This module is meant to be programmatic way of administering 7.X filers

=head1 SYNOPSIS

This module is instantiated by Server.pm and is not directly called.
See documentation below for help on individual methods 

=head1 EXAMPLES

No examples needed

=cut

#***********************************************************************

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDECLASSIC_7_X}) {
    my $newModule = $ENV{OVERRIDECLASSIC_7_X};
    delete $ENV{OVERRIDECLASSIC_7_X};
    print "The main Classic::7_X module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Node::Classic::7_X;
use strict;
use base qw(Node::Classic);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper;

use TracePkg;
my $Trace = new TracePkg();

our @Cmd_nodes = qw(aggregate network_ports network_interfaces exportruleset_rules volume cache_volume qtree);
our %Cmd_node_definitions = (
			     discover => {
				 aggregate => "aggr_status",
				 network_ports => "network_ports_show",
				 network_interfaces => "network_interfaces_show",
				 volume => "vol_status",
				 cache_volume => "flexcache_vol_status",
				 vserver => "vfiler_status",
 				 qtree => "qtree_status",
				 exportruleset_rules => "protocol_exportruleset_rule_show",
			     },
			     create => {
				 aggregate => "aggr_create",
				 network_interfaces => "network_interfaces_new",
				 volume => "volume_new",
				 cache_volume => "flexcache_vol_create",
				 qtree => "qtree_create",
				 exportruleset_rules => "protocol_exportruleset_rule_new",
			     },
			     exists => {
				 aggregate => "aggr_exists",
				 network_interfaces => "network_interfaces_exists",
				 volume => "vol_exists",
				 cache_volume => "vol_exists",
				 qtree => "qtree_exists",
				 exportruleset_rules => "protocol_exportruleset_rule_exists",
			     },
			     delete => {
				 aggregate => "aggr_destroy",
				 network_interfaces => "network_interfaces_delete",
				 volume => "volume_delete",
				 cache_volume => "volume_delete",
				 qtree => "qtree_delete",
				 exportruleset_rules => "protocol_exportruleset_rule_delete",
			     },
			     );
my @Modules = qw();

sub filter_vifs
{
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT } ],
					 {
					     volume => {type => HASHREF},
					     interfaces => {type => ARRAYREF},
                                             volume_nodes => {type => ARRAYREF},
					     traffic => {type => SCALAR|UNDEF}, # GX-specific
					     topology => {type => HASHREF},
                                             flexcache_involved => {type => SCALAR|UNDEF,
                                                                    optional => 1,
                                                                    default => 0},
					 }
					 );
    my $cache_involved = $params{flexcache_involved};
    my $volume = $params{volume};
    my $volume_name = $params{volume}->{volume};
    my @vifs = @{$params{interfaces}};
    my $topology = $params{topology};
    my $node = shift @{$params{volume_nodes}};

    $Trace->entry(msg => "Filtering vifs for $volume_name (cache traffic: $cache_involved)");

    my $server = $self->{server_object};

    my @filtered_vifs;

    if ($cache_involved) {
        $Trace->trace(msg => "Selecting interfaces on the cache");
	foreach my $vif (@vifs) {
            $Trace->trace(msg=>"Checking vif '$vif->{address}': cache=$vif->{flexcache},($vif->{'curr-filer'} != $node)");
	    if ($vif->{'curr-filer'} ne $node && $vif->{flexcache}) {
                $Trace->trace(msg=>"Create flexcache volume '$volume_name' on cache $vif->{'curr-filer'} with source=$node");
		push(@filtered_vifs, $vif);
		$self->_define_flexcache_volume($volume, $volume_name, $node, $vif->{'curr-filer'}, $topology);
	    }
	}
    } else {
        $Trace->debug(msg => "Selecting interfaces on the origin");
        @filtered_vifs = grep({ $_->{'curr-filer'} eq $node } @vifs);
    }	

    if (!scalar(@filtered_vifs)) {
        $Trace->fatal(msg => "No vifs satisfied your filter request (cache=$cache_involved) for $volume_name:\n".Dumper(\@vifs), result=>'CONFIG');
    }

    $Trace->exit(msg => "Returning cache=$cache_involved network traffic for volume $volume_name on filer $node\n".join("\n", (map( { "IP: ". $_->{address}." with Home-Filer: ".$_->{'curr-filer'} } @filtered_vifs))));
    return map( { $_->{address} } @filtered_vifs);
}

# This routine is called at the beginning of Server::delete_topology() and allows for 7G specific
#  code to be run before anything is actually deleted
sub pre_delete_topology {
    my $self = shift;
    $Trace->entry();
    $Trace->exit();
}

sub _define_modules {
    my $self = shift;
    $self->{Trace}->entry();
    $self->{modules} = \@Modules;
    $self->{Trace}->exit();
}

sub _define_cmd_nodes {
    my $self = shift;
    $self->{Trace}->entry();
    $self->{cmd_nodes} = \@Cmd_nodes;
    $self->{cmd_node_definitions} = \%Cmd_node_definitions;
    $self->{Trace}->exit();
}

1; # Do NOT remove this. It is needed by the package.

