#***********************************************************************
# This is the comment section for tharndoc
#***********************************************************************
## @summary Perl module used by Filer.pm to implement all "exportfs" commands in ONTAP
## @author dl-central-libs
## @pod here

=head1 NAME

Exportfs.pm

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

This module implements all of the "exportfs" commands in ONTAP

=head1 SYNOPSIS

This module is instantiated by all test scripts that use the Filer.pm module
and is not directly called.
See the descriptions below if you need the syntax for using a specific method

=head1 EXAMPLES

No examples needed

=cut

#***********************************************************************
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDECLASSIC_EXPORTFS}) {
    my $newModule = $ENV{OVERRIDECLASSIC_EXPORTFS};
    delete $ENV{OVERRIDECLASSIC_EXPORTFS};
    print "The main Classic::Exportfs module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

our %Exportfsconversion = (
			);
my %options = (
		   All => "-a",
		   options => "-io",
		   Verbose => "-v",
		   Pathname => " ",
		   Enable => "enable",
		   Disable => "disable",	
		   Save => "save",
		   NoSave => "nosave",
		   AllHosts => "allhosts",
		   Hosts => " ",
		   AllPaths => "allpaths",
		   Paths => " ",
		   Host => " ",
		   ReadOnly => "ro",
		   ReadWrite => "rw",
		   Root => "root",
		   Mount => " ",
		   Persistent => "-p",
		    );

our @Export_rules;

#************************************************************************

=head1 C<Exportfs()>

=over 4

=item Description

This method implements the "exportfs" command

=item Arguments

=over 4

=item All
(Optional)

Implements the "-a" flag, and takes list of path names from /etc/exports file

Mutually Exclusive with Pathname 

Supported Values: "true" or "false"

Default: "false"

=item options
(Optional)

Implements the "-o" flag, and represents a comma-separated list of options
that describe how a file or directory should be exported

Supported Values: "true" or "false"

Default: "false"

=item Verbose
(Optional)

Implements the "-v" flag, and prints each path name as it is exported

Supported Values: "true" or "false"

Default: "false"

=item Pathname 
(Optional)

Specific path to export

Mutually Exclusive with All

=back

=item Returns

None

=back

=cut

#************************************************************************

=head1 C<ExportfsFence()>

=over 4

=item Description

This method implements the "exportfs -b" command

=item Arguments

=over 4

=item Enable
(Optional)

Implements the "enable" keyword

Mutually Exclusive with Disable 

Supported Values: "true" or "false"

Default: "false"

=item Disable
(Optional)

Implements the "disable" keyword

Mutually Exclusive with Enable

Supported Values: "true" or "false"

Default: "false"

=item Save 
(Optional)

Implements the "save" keyword

Mutually Exclusive with NoSave 

Supported Values: "true" or "false"

Default: "false"

=item NoSave 
(Optional)

Implements the "nosave" keyword

Mutually Exclusive with Save 

Supported Values: "true" or "false"

Default: "false"

=item AllHosts 
(Optional)

Implements the "allhosts" keyword

Mutually Exclusive with Hosts 

Supported Values: "true" or "false"

Default: "false"

=item Hosts 
(Optional)

Implements the "hosts" keyword, and accepts a colon separated list of hostnames

Mutually Exclusive with AllHosts 

=item AllPaths
(Optional)

Implements the "allpaths" keyword

Mutually Exclusive with Paths

Supported Values: "true" or "false"

Default: "false"

=item Paths
(Optional)

Implements the "paths" keyword, and accepts a comma separated list of pathnames

Mutually Exclusive with AllPaths

=back

=item Returns

None

=back

=cut

#************************************************************************

my %command_definitions = (
			       Exportfs => {
                                   command => "exportfs",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
				   optional_options => [qw(All options Persistent Verbose Pathname)],
				   mutual_exclusions => {
                                     All => [qw(Pathname)],
                                   },
				   failure_output => [("no such directory")],
				   supported_values => {
                                     All => [qw(true false)],
                                   },
                               },
			       ExportfsFence => {
                                   command => "exportfs -b",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   required_options => [qw(Enable Disable Save NoSave AllHosts Hosts AllPaths Paths)],
                                   mutual_exclusions => {
                                     Enable => [qw(Disable)],
                                     Disable => [qw(Enable)],
                                     Save => [qw(NoSave)],
                                     NoSave => [qw(Save)],
                                     AllHosts => [qw(Hosts)],
                                     Hosts => [qw(AllHosts)],
                                     AllPaths => [qw(Paths)],
                                     Paths => [qw(AllPaths)],
                                   },
				   supported_values => {
                                     Enable => [qw(true false)],
                                     Disable => [qw(true false)],
                                     Save => [qw(true false)],
                                     NoSave => [qw(true false)],
                                     AllHosts => [qw(true false)],
                                     AllPaths => [qw(true false)],
                                   },
                               },
			       ExportfsCheck => {
                                   command => "exportfs -c",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   required_options => [qw(Host Pathname ReadOnly ReadWrite Root Mount)],
                                   mutual_exclusions => {
                                     ReadOnly => [qw(ReadWrite Root Mount)],
                                     ReadWrite => [qw(ReadOnly Root Mount)],
                                     Root => [qw(ReadOnly ReadWrite Mount)],
                                     Mount => [qw(ReadOnly ReadWrite Root)],
                                   },
                               },
			       ExportfsRevert => {
                                   command => "exportfs -d",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                               },
			       ExportfsFlush => {
                                   command => "exportfs -f",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   optional_options => [qw(Pathname)],
                               },
			       ExportfsPersist => {
                                   command => "exportfs -p",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   required_options => [qw(Pathname)],
                                   optional_options => [qw(options)],
                               },
			       exportfs_reload => {
                                   command => "exportfs -r",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                               },
			       ExportfsStorage => {
                                   command => "exportfs -s",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   required_options => [qw(Pathname)],
				   failure_output => [("There is no storage associated with")],
				   successful_output => [("The storage of")],
                               },
			       ExportfsUnexport => {
                                   command => "exportfs -u",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   required_options => [qw(All Pathname)],
                                   optional_options => [qw(options Verbose)],
				   mutual_exclusions => {
                                     All => [qw(Pathname)],
                                     Pathname => [qw(All)],
                                   },
                               },
			       ExportfsWrite => {
                                   command => "exportfs -w",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                                   required_options => [qw(Pathname)],
                               },
			       ExportfsHelp => {
                                   command => "exportfs -h",
                                   conversion => \%Exportfsconversion,
                                   options => \%options,
                               },
			       );
			       
my %Compatability = (
		     );

sub SetupExportfsCommands {
    my ($self, @servers) = @_;
    $self->_define_commands(Definitions => \%command_definitions,
			    Compatability => \%Compatability,
			    Servers => \@servers);
    return 1;
}

sub protocol_exportruleset_rule_show {
    my $self = shift;
    my %params = @_;
    $self->{Trace}->entry();

    my %export;
    foreach my $export (@Export_rules) {
	%export = %$export,last if ($export->{rulesetname} eq $params{rulesetname} && $export->{vserver} eq $params{vserver});
    }
    $self->{Trace}->fatal(msg=>"Could not find an export rule for '$params{rulesetname}'") if (%export == ());

    $self->{Trace}->exit();
    return %export;
}

sub protocol_exportruleset_rule_new {
    my $self = shift;
    my %params = @_;
    $self->{Trace}->entry();

    # this is just a placeholder in the object since export rules don't exist in classic apart from the volume.
    # the volume code will look in this location of the object if it is told to use an export rule for a volume.
    my $exists = undef;
    foreach my $export (@Export_rules) {
	$exists=1 if ($export->{rulesetname} eq $params{rulesetname} && $export->{vserver} eq $params{vserver});
    }
    delete $params{Topology};
    push(@Export_rules, \%params) if (!$exists);

    $self->{Trace}->exit();
}

sub protocol_exportruleset_rule_exists {
    my $self = shift;
    my %params = @_;
    $self->{Trace}->entry();

    # this is just a placeholder in the object since export rules don't exist in classic apart from the volume.
    # the volume code will look in this location of the object if it is told to use an export rule for a volume.
    my $exists = undef;
    foreach my $export (@Export_rules) {
	$exists=1 if ($export->{rulesetname} eq $params{rulesetname} && $export->{vserver} eq $params{vserver});
    }

    $self->{Trace}->exit(msg=>"return $exists");
    return $exists;
}

sub protocol_exportruleset_rule_delete {
    my $self = shift;
    my %params = @_;
    $self->{Trace}->entry();

    # this was just a placeholder in the object since export rules don't exist in classic apart from the volume.
    # the volume code will look in this location of the object if it is told to use an export rule for a volume.
    my $index = 0;
    foreach my $export (@Export_rules) {
	if ($export->{rulesetname} eq $params{rulesetname} && $export->{vserver} eq $params{vserver}) {
	    splice(@{$self->{export_rules}}, $index, 1); # remove this array entry
	    last;
	}
	$index++;
    }

    $self->{Trace}->exit();
}

1;
