#***********************************************************************
# This is the comment section for tharndoc
#***********************************************************************
## @summary Perl module used by Filer.pm to implement all "ifconfig" commands in ONTAP
## @author dl-central-libs
## @pod here

=head1 NAME

Ifconfig.pm

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

This module implements all of the "ifconfig" commands in ONTAP

=head1 SYNOPSIS

This module is instantiated by all test scripts that use the Filer.pm module
and is not directly called.
See the descriptions below if you need the syntax for using a specific method

=head1 EXAMPLES

No examples needed

=cut

#***********************************************************************

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDECLASSIC_IFCONFIG}) {
    my $newModule = $ENV{OVERRIDECLASSIC_IFCONFIG};
    delete $ENV{OVERRIDECLASSIC_IFCONFIG};
    print "The main Classic::Ifconfig module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

use strict;
use Network::Utils qw(get_ifconfig_info);

our %Ifconfigconversion = (
			   address => "Alias",
			   "home-port" => "Interface",
			   "netmask" => "Netmask",
			);
my %options = (
		   All => "-a",
		   Interface => " ",
		   AddressFamily => " ",
		   Broadcast => "broadcast",
		   Down => "down",
		   Up => "up",
		   Mediatype => "mediatype",
		   Flowcontrol => "flowcontrol",
		   Mtusize => "mtusize",
		   Netmask => "netmask",
		   PrefixLen => "prefixlen",
		   Alias => "alias",
		   RemoveAlias => "-alias",
		   Trusted => "trusted",
		   Untrusted => "untrusted",
		   WINS => "wins",
		   NoWINS => "-wins",
		   Partner => "partner",
		   RemovePartner => "-partner",
		   AnyCast => "anycast",
		    );
#************************************************************************

=head1 C<Ifconfig()>

=over 4

=item Description

This method implements the "ifconfig" command

=item Arguments

=over 4

=item All
(Required)

Implements the "-a" flag and displays info for all interfaces

Supported Values: "true" or "false"

Mutually Exclusive with "Interface"

=item Interface
(Required)

The interface to configure

Mutually Exclusive with "All"

=item AddressFamily
(Optional)

Specifies the address family which affects interpretation of remainng arguments

Supported Values: "inet" or "inet6"

=item Broadcast
(Optional)

Implements the "broadcast" keyword and is sued to specify broadcast address

=item Down
(Optional)

Implements the "down" keyword and brings the interface down

Supported Values: "true" or "false"

Mutually Exclusive with "Up"

=item Up 
(Optional)

Implements the "up" keyword and brings the interface up 

Supported Values: "true" or "false"

Mutually Exclusive with "Down"

=item Mediatype
(Optional)

Implements the "mediatype" keyword

Supported Values: "tp", "tp-fd", "100tx", "100tx-fd", "1000fx", "auto"

=item Flowcontrol
(Optional)

Implements the "flowcontrol" keyword 

Supported Values: "none", "receive", "send", "full"

=item Mtusize
(Optional)

Implements the "mtusize" keyword

=item Netmask
(Optional)

Implements the "netmask" keyword

Mutually Exclusive with "PrefixLen"

=item PrefixLen
(Optional)

Implements the "prefixlen" keyword

Mutually Exclusive with "Netmask"

=item Alias
(Optional)

Implements the "alias" keyword

Mutually Exclusive with "RemoveAlias"

Supported Values: "true" or "false"

=item RemoveAlias
(Optional)

Implements the "-alias" keyword

Mutually Exclusive with "Alias"

Supported Values: "true" or "false"

=item Trusted
(Optional)

Implements the "trusted" keyword

Mutually Exclusive with "Untrusted"

Supported Values: "true" or "false"

=item Untrusted
(Optional)

Implements the "untrusted" keyword

Mutually Exclusive with "Trusted"

Supported Values: "true" or "false"

=item WINS
(Optional)

Implements the "wins" keyword

Mutually Exclusive with "NoWINS"

Supported Values: "true" or "false"

=item NoWINS
(Optional)

Implements the "-wins" keyword

Mutually Exclusive with "WINS"

Supported Values: "true" or "false"

=item Partner 
(Optional)

Implements the "partner" keyword and specifies the failover interface

Mutually Exclusive with "RemovePartner"

=item RemovePartner
(Optional)

Implements the "-partner" keyword and removes failover interface

Mutually Exclusive with "Partner"

Supported Values: "true" or "false"

=item Anycast
(Optional)

Implements the "anycast" keyword

=back

=item Returns

None

=back

=cut

#************************************************************************

my %command_definitions = (
			       ifconfig => {
                                   command => "ifconfig",
                                   conversion => \%Ifconfigconversion,
                                   options => \%options,
				   optional_options => [qw(Role)],
				   required_options => [qw(All Interface)],
				   post_command_options => [qw(AddressFamily Broadcast Down Up Mediatype Flowcontrol Mtusize Netmask PrefixLen Alias RemoveAlias Trusted Untrusted WINS NoWINS Partner RemovePartner AnyCast)],
                                   verification => "_process_ifconfig_output",
				   defaults => {
				       All => " ",
				   },
				   mutual_exclusions => {
                                     All => [qw(Interface)],
                                     Interface => [qw(All)],
				     Down => [qw(Up)],
				     Up => [qw(Down)],
				     Netmask => [qw(PrefixLen)],
				     PrefixLen => [qw(Netmask)],
				     Alias => [qw(RemoveAlias)],
				     RemoveAlias => [qw(Alias)],
				     Trusted => [qw(Untrusted)],
				     Untrusted => [qw(Trusted)],
				     WINS => [qw(NoWINS)],
				     NoWINS => [qw(WINS)],
				     Partner => [qw(RemovePartner)],
				     RemovePartner => [qw(Partner)],
                                   },
			  	   supported_values => {
                                     AddressFamily => [qw(inet inet6)],
                                     Mediatype => [qw(tp tp-fd 100tx 100tx-fd 1000fx auto)],
			 	     Flowcontrol => [qw(none receive send full)],
			 	     Down => [qw(true false)],
			 	     Up => [qw(true false)],
			 	     Trusted => [qw(true false)],
			 	     Untrusted => [qw(true false)],
			 	     WINS => [qw(true false)],
			 	     NoWINS => [qw(true false)],
			 	     Partner => [qw(true false)],
			 	     RemovePartner => [qw(true false)],
                                   },
                               },
			       );
			       
my %Compatability = (
		     network_interfaces_show => "ifconfig",
		     network_ports_show => "ifconfig",
		     );

sub SetupIfconfigCommands {
    my ($self, @servers) = @_;
    $self->_define_commands(Definitions => \%command_definitions,
			    Compatability => \%Compatability,
			    Servers => \@servers);
    return 1;
}

sub _process_ifconfig_output {
    my ($self, %params) = @_;
    $self->{Trace}->entry(msg => "params => '" . (join ", " , (keys %params)) . "'");
    my $want_array = $params{want_array};
    $want_array = [] if (!$want_array);

    my $more_options = undef;
    my $port;
    my $status;
    my $mtu;
    my %vif;
    foreach my $line (split(/\n/, $params{output_in})) {
	if ($line =~ /^(.+):.+mtu (\d+)/) {
	    %vif = ();
	    $port = $1;
	    $mtu = $2;
	    $status = "down";
	}
	next if ($params{Interface} && $params{Interface} ne $port);
	$status = "up" if ($line =~ /<UP,/);
	if ($line =~ /^\s+inet (\S+)/ || $line =~ /^\s+inet6 (\S+)/) {
            # For inet6, not sure how to handle sitelocal address such as fe80::2a0 (autoconf )
            # Should we ignore them?
	    next if ($1 eq "127.0.0.1" || $1 eq "fe80::1" || $1 eq "::1");
	    $vif{address} = $1;
	    $vif{server} = "N/A"; # hack
	    $vif{port} = $port;
	    $vif{"status-admin"} = $status;
	    $vif{"status-oper"} = $status;
	    $vif{link} = $status; # for network_ports_show
	    $vif{mtu} = $mtu;
            $vif{role} = "data";
	    $vif{"curr-filer"} = $self->{server};
	    next if ( $params{Role} && $params{Role} ne $vif{role} );
	    push(@$want_array, {%vif}) if (%vif != ());
	}
    }
    $self->{Trace}->exit(msg => "Exiting with $want_array");
    return $want_array;
}

sub network_interfaces_exists {
    my $self = shift;
    my %params = (
		  @_);

    $self->{Trace}->entry(msg=>"ip=>$params{address}");
    my $interfaces = get_ifconfig_info($self->get_connection(type=>"cli_connection"));
    my $found = 0;
    while (my ($name, $interfaceh) = each %$interfaces) {
	$found = 1 if ($interfaceh->{ip_addr} eq $params{address});
	if ($interfaceh->{alias}) {
	    foreach my $aliash (@{$interfaceh->{alias}}) {
		$found = 1 if ($aliash->{ip_addr} eq $params{address});
	    }
	}
    }
    $self->{Trace}->exit(msg=>"return $found");
    return $found;
}

sub network_interfaces_new {
    my $self = shift;
    my %params = (
		  @_);
    my $output;
    $self->{Trace}->entry(msg=>"ip=>$params{address}");
    my $rc = $self->ifconfig(All=>undef,Alias=>$params{address}, Interface=>$params{"home-port"},
			     Netmask=>$params{netmask}, output=>\$output, -Errors=>undef);
    if ($output && $output !~ /already exists/) {
	$self->{Trace}->fatal(msg=>$output);
    }
    $self->{Trace}->exit();
}

sub network_interfaces_delete {
    my $self = shift;
    my %params = (
		  @_);
    my $output;
    $self->{Trace}->entry(msg=>"ip=>$params{address}");
    my $rc = $self->ifconfig(All=>undef,RemoveAlias=>$params{address}, Interface=>$params{"home-port"},
			     Netmask=>$params{netmask}, output=>\$output, -Errors=>undef);
    if ($output && $output !~ /already exists/) {
	$self->{Trace}->fatal(msg=>$output);
    }
    $self->{Trace}->exit();
}

1;

