
#***********************************************************************
# This is the comment section for tharndoc
#***********************************************************************
#
# $Id$
#
## @summary Perl module used to setup command modules for interraction with ONTAP filer 
## @author dl-central-libs
## @pod here

=head1 NAME

NG.pm

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

This module sets up the command descriptions that are listed in each
of the individual Perl modules located within the NG/ directory.
Each ONTAP command has its own Perl module that automates the invocation 
of that command on an ONTAP filer.

=head1 SYNOPSIS

This module is instantiated by Server.pm and is not directly called.

=head1 EXAMPLES

No examples needed

=cut

#***********************************************************************

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDENG}) {
    my $newModule = $ENV{OVERRIDENG};
    delete $ENV{OVERRIDENG};
    print "The main Node::NG module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Node::NG;
use Node::NG::Misc;
use Node::NG::Topology;
use Node::NG::Transit;
use Node::NG::Compatability;
use Node::NG::Os;
use Params::Common qw(parse_release_info $Release_separator);
use strict;
use Carp qw(croak confess);
use base qw(Node);
use TracePkg;
use Data::Dumper qw(Dumper);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);

our @Cmd_nodes; # defined in Topology.pm
our %Cmd_node_definitions; # defined in Topology.pm
my @Objects; # keep track of all of the created objects for END()

# the Cfe module was moved up a level since it is common with Classic as well
my @Modules = qw(Storage Network Volume Bootmenu);

my $Options;
my $CLI_separator = "X;X"; 

my $Trace = new TracePkg();

END {
    # This code will call any POST commands to perform cleanup on the filers.
    # The code is executed inside an END block rather than DESTROY because
    #  there is no guarantee that this class DESTROY will be called prior
    #  to the NATE connection DESTROY's being called
    $Trace->entry();
    foreach my $object (@Objects) {
        # make sure the cli connection is still open
        if ($object->{connections}{cli_connection}{object}) {    
            if (my $cmd = $object->{options_object}->get(option=>"POST_CLI_CMD")) {
                # execute the command
                $object->cli(cmd=>$cmd);
            }
        }
    }
    $Trace->exit();
}

$Options = new Options;
$Options->parameters(definitions=>
                     [
                      {
                       name => "CLUSTER_TRAFFIC",
                       description => "Which kind of cluster traffic to cause",
                       default => 'round-robin',
                       valid_values => [qw(local remote round-robin)],
                      },
                      {
                       name=>"SETVOLPERMS", # temporary hack
                       spec=>"i",
                       description=>"Allow setting of volume permissions",
                       default=>0,
                      },
                      {
                       name=>"CLI_FROM_SHELL",
                       spec=>"i",
                       description=>"run dbladecli/ngsh -c",
                      },
                      {
                       name=>"RLM_CONNTYPE",
                       description=>"the type of connection to make to the RLM",
                       default=>"ssh-persay",
                      },
                      {
                       name=>"LIBSSH_OVERRIDE",
                       description=>"Overrides the ssh-persist connection type hardcoded in Node/NG.pm",
                       default=>0,
                      },
#                      {
#                       name=>"PRE_CLI_CMD",
#                       description=>"CLI commands to run as soon as the CLI is opened",
#                      },
#                      {
#                       name=>"POST_CLI_CMD",
#                       description=>"CLI commands to run upon CLI exit",
#                      },
                     ]);

my $Libssh_override = 0;

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
	server => undef,
	server_object => undef,
	@_,
	};

    bless($self, $class);

    $Trace->entry(msg=>"Entered for $self->{server}\n");

    if (!$self->{server}) {
	$Trace->fatal(msg=>"A server must be passed in.", result=>'SCRIPT');
    }

    $self->{host_object} = main::host($self->{server});
    $self->{options_object} = $Options;
    $self->{CLI} = "/sbin/ngsh";
    $self->{DBLADECLI} = "/usr/sbin/dbladecli";
    $self->{cli_from_shell} = $self->{options_object}->get(option=>"CLI_FROM_SHELL");
    $self->{type} = $self->ontap_ng();

    # create a connections hash which is required by the establish_connections routine. 
    # go ahead an establish a cli connection since it is needed to get the version
    # if no_cli=0 establish cli connections up front and load appropriate cdefs
    # if no_cli=1 do NOT establish cli connections up front, but load non-gen cdefs
    # if no_cli exists but is not '1' then establish cli connections on the fly
    # and load the correct gen'ed cdefs for the build running on those nodes
    $self->{connections} = {};
    if (!$self->{no_cli}) {
        $self->establish_connections(cli_connection => 1);
        $self->_platform_compatibility($self);   
    } elsif ($self->{no_cli} == 1) {
        $Trace->comment(msg => "CLI connection not established. Loading \"non-generated\" cdefs.");

        # The following calls will load the hand-coded cdef modules. This
        # is a hack because we should have a better mechanism for modules
        # being loaded to be defined in Node.pm and not in the children...
        $self->_setup_commands("Node", undef, $self->_common_modules());
        $self->_setup_commands(ref($self), undef , @Modules);

        $self->_modify_commands();
        $self->_platform_compatibility($self);
    } elsif (($self->{no_cli})&&($self->{no_cli} != 1)) {
        $self->_platform_compatibility($self);        
    }

    $self->{cmd_nodes} = \@Cmd_nodes;
    $self->{cmd_node_definitions} = \%Cmd_node_definitions;

    # add this new object to the global list
    push(@Objects, $self);

  done:
    $Trace->exit(msg=>"Exited for $self->{server} with $self\n");
    return $self;
}

sub establish_connections {
    my $self = shift;
    my %params = (
		  console_connection =>undef,
		  shell_connection => undef,
		  cli_connection => undef,
		  dblade_connection => undef,
		  rlm_connection => undef,
		  cfe_connection => undef,
		  abortive => 1,
		  @_);
    my $state;
    my $connections = $self->{connections};

    $Trace->entry(msg=>"console=$params{console_connection},shell=$params{shell_connection},cli=$params{cli_connection},".
		  "dblade=$params{dblade_connection},rlm=$params{rlm_connection},cfe=$params{cfe_connection}");

    $self->_setup_login_credentials() if (!$self->{CLI_USER});

    if ($params{console_connection}) {
        # try to setup and create the console connection
        $connections->{console_connection}{conn_specs} = "$self->{server}\#console ; conntype=pm_cons ; shell_type=sh";
        $self->open_connection(type=>"console_connection", abortive=>$params{abortive});

	# After the connection is setup, we need to add a new prompt.  This is a pain if you don't know what all the default
	# prompts should be.  So get them, update them, and save them (IN TWO PLACES, ugh)
	my @console_prompts = $connections->{console_connection}{object}->prompts();
	push( @console_prompts, "^$self->{server}\%\\s*\$" );
	$connections->{console_connection}{object}->prompts(@console_prompts);
	# Now save in the Connections hash, so Tranist does the right thing
	@{$connections->{console_connection}{default_prompts}} = $connections->{console_connection}{object}->prompts();
    }
    
    my $ip = $self->{host_object}->default_ip();

    $Libssh_override = $self->{options_object}->get(option=>"LIBSSH_OVERRIDE");

    if ($state ne "cfe") {
	if ($params{shell_connection}) {
	    
	    # Get passed or discovered release info
            my ($branch,$release,$build,$buildopts,$buildpath) = parse_release_info(node=>$self->{server});
	    my $options = $Options->get_hash();

	    # Pre-BR releases can directly log into shell
	    if ($self->has_change(change => "node-scope") || $self->{RR_FLAG}) {
		$Trace->trace(msg=>"We're here because node-scope: " . $self->has_change(change => 'node-scope')  .
				" or \$self->{RR_FLAG}: $self->{RR_FLAG}");
		# From BR on up we need to use the systemshell method of getting to shell

		$self->{RR_FLAG} = 1;  # ACK ACK HACK, save the fact we're using the newer RR methods
		eval {
		    $self->_shell_through_ngsh( connections => $connections, ip => $ip );
		};
		if ($@) {
		    # Error either because 'diag' isn't setup, or on an older build		    
		    my $warn_msg = "The account 'diag' is now required to gain BSD shell access.  The node does not seem to\n";
		    $warn_msg .= "have the diag account configured.  Setting up the account now using the same password as the\n";
		    $warn_msg .= "cli admin account, typically 'netapp1!'";
		    $Trace->comment(msg=>"Not able to login in to shell through ngsh:\n$@\n$warn_msg");
		    $self->close_connection( type => "shell_connection" );

		    # Try setting up the 'diag' user.  If we still can't login we're done.
		    $self->enable_diag();		    
		    # Try again
		    $self->_shell_through_ngsh( connections => $connections, ip => $ip );

		    # If we fail again, just fall out, we're out of tricks.
		}
            } else {
		$Trace->trace(msg=>"We're here because node-scope: " . $self->has_change(change => 'node-scope')  .
				" and params{RR_FLAG}: $params{RR_FLAG}");
		# Older method of getting into shell
		# setup and create the shell connection

                # If Libssh_override is set, use libssh-persist
                if (!$Libssh_override) {
		    $connections->{shell_connection}{conn_specs} = "$self->{server}\#shell ; conntype=ssh-persist ; connhost=$ip";
                } else {
		    $connections->{shell_connection}{conn_specs} = "$self->{server}\#shell ; conntype=libssh-persist ; connhost=$ip";
                }
		$connections->{shell_connection}{conn_specs} .= " ; username=$self->{SHELL_USER} ; " . 
		    "password=$self->{SHELL_PASSWORD} ; shell_type=sh";
		$self->open_connection(type=>"shell_connection", abortive=>$params{abortive});
            }
	}

	if ($params{cli_connection}) {
	    # setup and create the cli connection

            # If Libssh_override is set, use libssh-persist
            if (!$Libssh_override) {
	        $connections->{cli_connection}{conn_specs} = "$self->{server}\#$self->{CLI_CONNECTION} ; conntype=ssh-persist";
            } else {
	        $connections->{cli_connection}{conn_specs} = "$self->{server}\#$self->{CLI_CONNECTION} ; conntype=libssh-persist";
            }
	    $connections->{cli_connection}{conn_specs} .= " ; connhost=$ip ; username=$self->{CLI_USER} ; " . 
		"password=$self->{CLI_PASSWORD} ; shell_type=none";
            if (!$self->{cli_from_shell}) {
                $connections->{cli_connection}{conn_specs} .= " ; prompts='^\\S+::>\\s'";
            } else {
                $connections->{cli_connection}{conn_specs} .= " ; prompts='^\\S+#\\s'";
            }
            # Try logging in directly to the CLI. This should work on more recent installs but older ones
            #  are not setup properly for direct login. In that case, fall back to logging in to the shell
            #  and starting ngsh
            eval {
                $self->open_connection(type=>"cli_connection", abortive=>$params{abortive});
            };
            if ($@ && $connections->{cli_connection}{setup}) {
                die $@;
            }
            if ($@) {
                # If we had tried logging in directly to the CLI, and it had failed,then try connecting to the
                # shell and running ngsh. 
                # Close the old connection object
                $self->close_connection(type=>"cli_connection");
                my $warn_msg = "The account might not have been configured in the initial setup to allow direct login to the CLI.\n";
                $warn_msg .= "If this is a Tricky node and the nate_host file does not have the proper\n";
                $warn_msg .= "cli password in the cli connspec, then either update it or set CLI_PASSWORD=EMPTY_STRING.\n";
                $warn_msg .= "Please see the Server documentation for more information\n";
                $warn_msg .= "Falling back to logging in as root and executing ngsh(ngshell)";
                $Trace->warn(msg=>$warn_msg);
                
                # Update the connection specs
                $self->{CLI_CONNECTION} = "ngshell";
                $self->{CLI_USER} = $self->{SHELL_USER};
                $self->{CLI_PASSWORD} = $self->{SHELL_PASSWORD};

                # If Libssh_override is set, use libssh-persist
                if (!$Libssh_override) {
                    $connections->{cli_connection}{conn_specs} = "$self->{server}\#$self->{CLI_CONNECTION} ; conntype=ssh-persist";
                } else {
                    $connections->{cli_connection}{conn_specs} = "$self->{server}\#$self->{CLI_CONNECTION} ; conntype=libssh-persist";
                }
                $connections->{cli_connection}{conn_specs} .= " ; connhost=$ip ; username=$self->{CLI_USER} ; " .
                "password=$self->{CLI_PASSWORD} ; shell_type=none" . " ; prompts='^\\S+#\\s'";

                # Open connection to the shell
                $self->open_connection(type=>"cli_connection", abortive=>$params{abortive});
            }
            if ($self->{CLI_CONNECTION} eq "ngshell") {
                $connections->{cli_connection}{object}->prompts(['^\S+::>\s',
                                                                 '^\S+::\*>\s',
                                                                 '^\S+::sfo-waiting::\*>\s',
                                                                 '^\S+#\s',
                                                                 '^\S+::sfo-waiting::>\s',
                                                                ]);
                
                if (!$self->{cli_from_shell}) {
                    # now start the ngsh process
                    $self->cli(cmd=>$self->{CLI});
                }
            }
            
            $connections->{cli_connection}{errors} = ["(ERROR|Error):", "There are no entries matching your query",
                                                      "This table is currently empty", "Usage:",
					              "(ERROR|Error):", "command failed:", "duplicate entry",		
                                                      "command failed: duplicate entry",
                                                      "server closed connection unexpectedly"];
            $connections->{cli_connection}{acceptable} = [];
            $connections->{cli_connection}{idempotent} = [
                                                          "creation failed: duplicate entry", 
                                                          "deletion failed: entry doesn't exist",
                                                          "command failed:.+entry doesn't exist",
                                                          "creation failed.+duplicate key",
                                                          "creation failed: Duplicate vif id",
                                                          "A load sharing mirror exists",
                                                          "duplicate configuration state",
                                                          "(ERROR|Error):.+Reason: duplicate entry",
                                                          "(ERROR|Error): command failed: duplicate entry",
                                                         ];
            $connections->{cli_connection}{transient} = ["Reason: VLDB could not initiate a replication transaction"];
            # If called as cli_from_shell login to shell & then run command as ngsh -c
	    if (!$self->{cli_from_shell}) {
                $connections->{cli_connection}{object}->prompts(['^\S+::>\s',
                                                                 '^\S+::\*>\s',
                                                                 '^\S+::sfo-waiting::\*>\s',
                                                                 '^\S+#\s',
                                                                 '^\S+::sfo-waiting::>\s',
								 ]);
                $self->cli(cmd=>"set -confirmations off");
                $self->cli(cmd=>"set -showallfields true -showseparator \"$CLI_separator\" -privilege diagnostic ; system timeout modify -timeout 0 ; rows 0");
           } else {
                $connections->{cli_connection}{object}->prompts('^\S+#\s');
                $connections->{cli_connection}{cli_from_shell_preamble} = "set -confirmations off -showallfields true -showseparator \"$CLI_separator\" -privilege diagnostic ; system timeout modify -timeout 0";
                $connections->{cli_connection}{cli_from_shell} = "ngsh -c";
            }
            
            $connections->{cli_connection}{help_handler} = "_cli_help_handler";
            
            # If there is a pre command, then execute it now
            #                if (my $cmd = $self->{options_object}->get(option=>"PRE_CLI_CMD")) {
            #                    $self->cli(cmd=>$cmd);
            #                }
            
            # call _get_version_info prior to loading the commands because the build information is a prequisite
            $self->_get_version_info();

	    # Leave the shell connection open.  Costs an ssh session, save on 5-10 seconds and code complexity

            # Locate the cdefs directory. The top-most directory is based upon the name of this class. So,
            #  for example, Node::NG will become Node/NG. Then add on the name of the current branch or default
            #  to 'main'. Then add on "/cdefs". So a path for a 10.0.1 node would be:
            #  Node/NG/R10.0.1x/cdefs
            my $cdef_dir = ref($self);
            $cdef_dir =~ s!::!/!;
            
            # Provide a default for missing branch info
            $self->{branch} = "main" if $self->{branch} eq "";
            
            $cdef_dir .= $self->{branch} ? "/cdefs/".$self->{branch} : "/cdefs";
            
            # Now lookup the absolute path based on our relative path
            my $found_dir = $self->_find_cdef_path($cdef_dir);
            
            # if the current branch was not found, check to see if this was an R10
            #  branch.  If R10, default to Rtricky, if not then default 'main' as the branch
            if ( ! $found_dir ) {
                $Trace->debug(msg=>"Didn't find cdef dir: $cdef_dir");
                if ( $self->{branch} =~ /R10/ ) {
                    $Trace->debug(msg=>"This is an R10 branch, default to Rtricky");
                    $cdef_dir = ref($self) . "/cdefs/Rtricky";
                } elsif ( $self->{branch} =~ /R8.0/ || $self->{release} =~ /Boilermaker/ ) {
                    $Trace->debug(msg=>"This is an Rboilermaker branch, default to Rboilermaker");
                    $cdef_dir = ref($self) . "/cdefs/Rboilermaker";
                } else {
                    $Trace->debug(msg=>"This is not an R10 branch or Rboilermaker branch, default to main");
                    $cdef_dir = ref($self) . "/cdefs/main";
                    if ( exists $self->{backing_branch} ) {
                        $Trace->debug(msg =>
                            "Sparse branches are not supported for public cdefs and currently default to using main.");
                    }
                }    
                $cdef_dir =~ s!::!/!;
                $found_dir = $self->_find_cdef_path($cdef_dir);
                $Trace->debug(msg=>"Instead using: $found_dir");
            }
            
            # The following calls will load the hand-coded cdef modules. This
            # is a hack because we should have a better mechanism for modules
            # being loaded to be defined in Node.pm and not in the children...
            $self->_setup_commands("Node", undef, $self->_common_modules());
            $self->_setup_commands(ref($self), $found_dir, @Modules);
            $self->_modify_commands();
        }
        $connections->{cli_connection}{setup} = 1;
        
	if ($params{dblade_connection}) {
            # setup and create the dbladecli connection

            if ($self->has_change(change => "node-scope")) {
                # In Converged Ontap, dbladecli has been replaced with NodeScope in the UI so run
                #  the command inside of ngsh and set a flag to tell say_wrapper that it needs to
                #  run the command inside of Node Scope
		# But wait, first we have to set up a shell connection through ngsh/systemshell
		$self->_shell_through_ngsh( connections => $connections, ip => $ip, type => "dblade_connection", tag => "dblade" );
		# the shell connection is now setup, now set the flags that trigger the ngsh -c 'node run <cmd>' stuff
                $connections->{dblade_connection}{cli_from_shell} = "$self->{CLI} -c ";
                $connections->{dblade_connection}{node_run} = 1;
            } else {
                # This is the old-school dbladecli connection so run the command via it
                $connections->{dblade_connection}{cli_from_shell} = "$self->{DBLADECLI} -c ";	    

                # If Libssh_override is set, use libssh-persist
                if (!$Libssh_override) {
		    $connections->{dblade_connection}{conn_specs} = "$self->{server}\#dblade ; conntype=ssh-persist ; connhost=$ip";
                } else {
		    $connections->{dblade_connection}{conn_specs} = "$self->{server}\#dblade ; conntype=libssh-persist ; connhost=$ip";
                }
		$connections->{dblade_connection}{conn_specs} .= " ; username=$self->{SHELL_USER} ; " . 
		    "password=$self->{SHELL_PASSWORD} ; shell_type=none";
		$connections->{dblade_connection}{conn_specs} .= " ; prompts='^\\S+\[#\%\]\\s'";
		$self->open_connection(type=>"dblade_connection", abortive=>$params{abortive});
            }
	}
    
	if ($params{rlm_connection}) {
	    # setup and create the rlm connection
	    my $rlm_conntype = $self->{options_object}->get(option=>"RLM_CONNTYPE");
	    $Trace->debug(msg=>"Using persistent RLM connection.") if ($rlm_conntype eq "ssh-persist");
	    $connections->{rlm_connection}{conn_specs} = "$self->{server}\#rlm ;prompts='(RLM|bmc|SP).*>' ; shell_type=none";
	    $self->open_connection(type=>"rlm_connection", conntype=>$rlm_conntype, abortive=>$params{abortive});
	    $self->rlm(cmd=>"priv set diag") if ($rlm_conntype eq "ssh-persist");
	}
    }
    $Trace->exit();
}


#
# Setup the connectrec for logging into the cli
#
sub _setup_connection {
    my $self = shift;
    my %params = (
		  connections => undef,
		  ip          => undef,
		  tag         => "cli",
		  type        => undef,
		  user        => undef,
		  password    => undef,
		  prompts     => undef,
		  @_);

    my $connections = $params{connections};
    my $ip = $params{ip};
    my $tag = $params{tag};
    my $type = $params{type};
    my $user = $params{user};
    my $password = $params{password};
    my $prompts = $params{prompts};

    $Trace->entry(msg => "ip: $ip, tag: $tag, type: $type " .
		  "user: $user, password: $password" );


    # If Libssh_override is set, use libssh-persist
    if (!$Libssh_override) {
        $connections->{$type}{conn_specs} = "$self->{server}\#$tag ; conntype=ssh-persist ; connhost=$ip";
    } else {
        $connections->{$type}{conn_specs} = "$self->{server}\#$tag ; conntype=libssh-persist ; connhost=$ip";
    }
    $connections->{$type}{conn_specs} .= " ; username=$user ; " . 
	"password=$password ; shell_type=none";

    if ( $type =~ /cli_connection|dblade_connection/ ) {
	if (!$self->{cli_from_shell}) {
	    $connections->{$type}{conn_specs} .= " ; prompts='^\\S+::'";
	} else {
	    $connections->{$type}{conn_specs} .= " ; prompts='^\\S*\[#\%\]'";
	}
    }
    if ( $type eq "shell_connection" ) {
	$connections->{$type}{conn_specs} .= " ; prompts="."^\\S+::";
    }
    
    $Trace->exit();
}



#
# _shell_through_ngsh
#
# In later releases of BR, direct BSD shell login has been removed.  We now log into ngsh and
# use the 'systemshell' command and the 'diag' userid.
#
sub _shell_through_ngsh {
    my $self = shift;
    my %params = (
		  connections => undef,
		  ip          => undef,
		  type        => "shell_connection",
		  tag         => "shell",
		  abortive    => 1,
		  @_);

    my $connections = $params{connections};
    my $ip = $params{ip};
    my $type = $params{type};
    my $tag = $params{tag};

    $Trace->entry(msg => "ip: $ip ");

    # BR has removed the direct access to the bsd shell.  Now one must log into ngsh and then invoke the 'systemshell'
    # command.  This invokes login, where we use the 'diag' user.  This user is assumed to be already
    # setup prior to getting here (such as in gx_boot).
    
    # Setup connection for ngsh
    $self->_setup_connection( connections => $connections, ip => $ip, tag => $tag, type => $type,
			      user => $self->{CLI_USER}, password => $self->{CLI_PASSWORD}, prompts => "'^\\S+::'" );
    
    # Log into ngsh

    # If the connections have already been established, then we need to use the existing connectrec.  NATE doesn't
    # allow for deleting them and will always reuse.
    if ( defined $self->{connections}{saved_connectrecs}{$type} ) {
	$Trace->trace(msg=>"We already have a saved $tag connection, reset it's prompts");
	$self->{connections}{saved_connectrecs}{$type}->prompts(['^\S+::>',
								'^\S+::\*>',
								'^\S+::sfo-waiting::\*>',
								'^\S+::sfo-waiting::>',
								]);
    }

    $self->open_connection(type=>$type, abortive=>$params{abortive});

    # OK, we now have a shell connection.  It happens to look like a CLI, but it's the connection we'll use for shell
    # none the less.  We need to stash it away so we can reuse it later.  NATE doesn't delete it, so we can't either.
    $self->{connections}{saved_connectrecs}{$type} = $self->{connections}{$type}{object};

    # Prep ngsh for script friendly commands
    $self->$tag(cmd => "set -confirmations off");
    $self->$tag(cmd => "set -privilege diagnostic");
    
    # Login to the diag user to establish the shell session
    $self->$tag(cmd => "systemshell -node local",
		additional_prompts_timeout => 20,
		additional_prompts => {
		    'login:'    => 'diag',
		    'Password:' => $self->{CLI_PASSWORD}, 
		    "Login incorrect" => "FATAL_PROMPT",
		    '^\S*#'    => "set prompt='".$self->{server}."#'",
		    '^\S*%'    => "set prompt='".$self->{server}."%'",
		    '^'.$self->{server}.'[#%]'    => "FINAL_PROMPT",			     
		}
		);
    # Set up the typical shell prompt (now that we're out of ngsh)
    $connections->{$type}{object}->prompts(['^' . $self->{server} . '\#\s*$',
					    '^' . $self->{server} . '\%\s*$',
					    '^' . $self->{server} . '\$\s*$',
					    ]);
    $self->$tag(cmd => "exec sh -c 'unset ENV; PS1=" . $self->{server} . "%; export PS1; exec /bin/sh -i' -");
    $Trace->exit();
}


# filter_vifs tries to locate the set of VIFS that honor the traffic direction
#  based upon a given volume
# volume: the volume hash describing the volume
# interfaces: the list of interface hashes to filter
# volume_nodes: reference to an array of nodes hosting the volume
# traffic: see the description of CLUSTER_TRAFFIC
# topology: the test_topology is updated when flexcaches are involved
# flexcache_involved: a flag indicating that the configuration involves caches
#
# returns the list of IP addresses for the filtered vifs
sub filter_vifs
{
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT } ],
					 {
					     volume => {type => HASHREF},
					     interfaces => {type => ARRAYREF},
                                             volume_nodes => {type => ARRAYREF},
					     traffic => {type => SCALAR|UNDEF},
					     topology => {type => HASHREF},
                                             flexcache_involved => {type => SCALAR|UNDEF,
                                                                    optional => 1,
                                                                    default => undef},
					 }
					 );
    my $options = $Options->get_hash();
    my $traffic = $options->{CLUSTER_TRAFFIC} || $params{traffic};
    my $volume = $params{volume};
    my $volume_name = $params{volume}->{volume};
    my @vifs = @{$params{interfaces}};
    my $topology = $params{topology};
    
    my $filter_error = 0;

    $Trace->entry(msg => "Filtering vifs for $volume_name (cluster traffic: $traffic".
                  ", flexcache_involved: $params{flexcache_involved})");

    if (! @vifs) {
        $Trace->fatal(msg=>"There were no vifs passed in to be filtered");
    }

    # filtered_vifs will contain the final list of filtered vifs that are returned
    my @filtered_vifs;

    # volume_nodes is used to handle the case when filers is sometimes not an array
    my @volume_nodes = @{$params{volume_nodes}};

    # if caches are involved, then extra work is needed to setup the cache volumes and 
    #  cache-to-source relationships
    if ($params{flexcache_involved}) {
        $Trace->trace(msg => "Selecting interfaces on the cache");

        # since a cache is involved, all of the interfaces on the caches will be added to the
        # filtered_vifs
        @filtered_vifs = grep($_->{flexcache}, @vifs);
        if (@filtered_vifs == @vifs) {
            $Trace->comment(msg=>Dumper \@vifs);
            $Trace->fatal(msg=>"There were no non-flexcache vifs found in the vif list. The full list is in the log",
                         result=>"CONFIG");
        } elsif ( !scalar @filtered_vifs ) {
            $Trace->comment(msg=>Dumper \@vifs);
            $Trace->fatal(msg=>"There were no flexcache vifs found in the vif list.  The full list in in the log",
                         result=>"CONFIG");
        }
        
        # used to track the vifs that will be the source IPs for the flexcache volumes
        my @source_vifs; 

        if ($traffic eq 'local') {
            # In this case, the caching vifs will be those addresses on the nblade local to the volume
            # and the client_facing vifs will be those addresses on the caches themselves

            $Trace->debug(msg => "Selecting local N to D traffic vif for the cache relationship");
            # Match if the vif being checked in the grep is on a filer that contains the volume
            foreach my $vif (@vifs) {
                if (!$vif->{flexcache} && grep($vif->{'curr-filer'} eq $_, @volume_nodes)) {
                    # this vif is local to the volume/aggregate so use it for the flexcache volume
                    $Trace->trace(msg=>"adding 'local' '$vif->{address}' to \@source_vifs because it is not a flexcache vif and '$vif->{'curr-filer'}' is in ".join(",", @volume_nodes));
                    push(@source_vifs, $vif);
                }
            }

        } elsif ($traffic eq 'remote') {
            # In this case, the caching vifs will be those addresses on the nblade not local to the volume
            # and the client_facing vifs will be those addresses on the caches themselves

            $Trace->debug(msg => "Selecting remote N to D traffic vif for the cache relationship");
            foreach my $vif (@vifs) {
                if (!$vif->{flexcache} && !grep($vif->{'curr-filer'} eq $_, @volume_nodes)) {
                    # this vif is local to the volume/aggregate so use it for the flexcache volume
                    $Trace->trace(msg=>"adding 'remote' '$vif->{address}' to \@source_vifs because it is not a flexcache vif and '$vif->{'curr-filer'}' is not in ".join(",", @volume_nodes));
                    push(@source_vifs, $vif);
                }
            }
        } elsif ($traffic eq "round-robin") {
            # In this case, find all of the non-flexcache vifs as the source vifs
            $Trace->debug(msg => "Selecting $traffic vif");
            foreach my $vif (@vifs) {
                if (!$vif->{flexcache}) {
                    $Trace->trace(msg=>"adding 'round-robin' '$vif->{address}' to \@source_vifs because it is not a flexcache vif");
                    push(@source_vifs, $vif);
                }
            }
        }
        
        # neither of these should be that case, but just in case we'll want to throw a pretty
        # verbose error later in the code.
        if (!scalar(@source_vifs)) {
            $filter_error = 1;
        }
        if ( !scalar(@filtered_vifs)) {
            $filter_error = 1;
        }

        # now we need to define the flexcache volumes that need to be created as well
        # pick a random source filer vif for this volume on the cache
        my $source = $source_vifs[int rand(scalar @source_vifs)]->{address};

        foreach my $cache_vif (@filtered_vifs) {
            my $cache = $cache_vif->{'curr-filer'};
            $Trace->trace(msg=>"Create flexcache volume '$volume_name' on cache $cache with source=$source");
            $self->_define_flexcache_volume($volume, $volume_name, $source, $cache, $topology);
        }
            
    } else {
        # this is the non-cache path so just pick vifs based upon the traffic rules
        if ($traffic eq 'local') {
            $Trace->debug(msg => "Selecting local N to D traffic vif");
            # Match if the vif being checked in the grep is on a filer that contains the volume
            @filtered_vifs = grep({ my $vr = $_; grep(/^$vr->{'curr-filer'}$/, @volume_nodes) } @vifs);
        } elsif ($traffic eq 'remote') {
            $Trace->debug(msg => "Selecting N to remote D traffic vif");
            # Match if the vif being checked in the grep is not on a filer that contains the volume
            @filtered_vifs = grep({ my $vr = $_; !grep(/^$vr->{'curr-filer'}$/, @volume_nodes) } @vifs);
        } elsif ($traffic eq "round-robin") {
            $Trace->debug(msg => "Selecting $traffic vif");
            @filtered_vifs = @vifs;
        }
        if (!scalar(@filtered_vifs)) {
            $filter_error = 1;
        }
    }

    if ($filter_error) {
        $Trace->comment(msg=>Dumper(\@vifs));
        my @vif_nodes = map {$_->{'curr-filer'}} grep(!$_->{flexcache}, @vifs);
        $self->_uniq(\@vif_nodes);
        $Trace->fatal(msg => "A request was made (via CLUSTER_TRAFFIC=$traffic) to find\n".
                      "vifs which are $traffic to the node hosting volume '$volume_name'\n".
                      "but no $traffic vifs could not found.\n".
                      "The full list of vifs was dumped into the log as a comment.\n".
                      "The request for CLUSTER_TRAFFIC was '$traffic'\n".
                      "The value of flexcache_involved is '$params{flexcache_involved}'\n".
                      "The node[s] hosting the volume: ".join(",", @volume_nodes)."\n".
                      "The node[s] hosting the vifs[s]: ".join(",", @vif_nodes)."\n",
                      result=>'CONFIG');
    }

    $Trace->exit(msg => "Returning $traffic N to D traffic for volume $volume_name on filer @volume_nodes}\n".join("\n", (map( { "IP: ". $_->{address}." with Current Filer: ".$_->{'curr-filer'} } @filtered_vifs))));
    return (map( { $_->{address} } @filtered_vifs));
}

sub server_type {
    my ($self, %params) = @_;
    $Trace->entry();
    my $type = $params{type};
    my $is_type = ($type == $self->ONTAP_NG());
    $Trace->exit(msg=>"returning $is_type");
    return $is_type;
}

# this method is called after all of the cdefs have been read in from the files.
sub _modify_commands {
    my $self = shift;
    
    my %params = @_;

    $Trace->entry();

    my %new_commands = (); # new command defs to be added to our current list
    while (my ($cmd, $cmd_hash) = each %{$self->{Commands}}) {
	if ($cmd =~ /.+((_show)|(_status)|(_node_autosupport_destinations)|(_stat))/) {
	    # if this is a show command, then setup an output handler for it
	    $$cmd_hash{output_handler} = "_cli_output_handler";
	    if ($cmd eq "cluster_identity_show") {
		$$cmd_hash{valid_params} = [],
	    }
	} elsif ($cmd =~ /(.+)_(new|create)/) {
	    # if this is a new or create command then setup a verification routine for it
	    if ($cmd =~  /(volume)_(new|create)/) {
		##$$cmd_hash{output_handler} = "$1_show";
		#$Trace->trace(msg=>"Skipping verification of new volumes due to burt221288");
	    } elsif ( $cmd eq "protocol_nfs_server_new" ||
		      $cmd eq "vserver_nfs_new") {
		$Trace->trace(msg=>"Skipping verification of protocol nfs server new");
	    } elsif ($cmd eq "protocol_exportruleset_rule_new" ||
		     $cmd eq "protocol_export_policy_new") {
		# this is a special case which will be verified outside of the object
		$$cmd_hash{output_handler} = "$1_show";
            } elsif (grep($cmd eq $_, qw(volume_clone_create volume_clone_new))) {
                $$cmd_hash{verification} = "_volop_verify";
	    } elsif ( $cmd =~ /volume_snapshot_(create|new)/ ) {
    		if (!($self->has_change(change=>"snapshot-create-delete-bypass-job"))) {
	       	    $$cmd_hash{verification} = "_volop_verify";
	        } else {
	    	    # Skip verification (because Job Manager is being removed from
	    	    # snapshot create/new/delete)
	    	    $Trace->trace(msg=>"Skipping verification of volume snapshot new/create");
	    	}
	    } else {
		# this is a regular new() so use the corresponding show() to verify it
		$$cmd_hash{verification} = "$1_show";
		if ($cmd =~ /protocols?_cifs_server_new/ || 
		    $cmd =~ /vserver_cifs_(new|create)/) {
		    # This command presents us with additional prompts
		    # Using a hash since we may not get all of the prompts depending
		    # on how the routine was called.
		    $$cmd_hash{additional_prompts} = {
            # Tricky
			"Please enter your userid:" => "-userid",
			"Please enter your password:" => "-password",
            # >= BR
			"Enter the user name:" => "-userid",
			"Enter the password:" => "-password",
			"Ok to reuse this account(.+):" => "-reuse_machine_account",
	                "::.?>" => "FINAL_PROMPT",
			};
	            $$cmd_hash{additional_prompts_timeout} = 90;
		}elsif ($cmd =~ /security_login_new/) {
		    $$cmd_hash{additional_prompts} = {
			"(E|e)nter a password for user " => "-password",
			"(E|e)nter it again:" => "-password",
                        "::.?>" => "FINAL_PROMPT",
			};
	            $$cmd_hash{additional_prompts_timeout} = 30;
		}
            }
    # The 'antivirus engine modify' command, and more specifically its '-sophos-http-user' attribute
    # will provoke the UI to prompt for the sophos http user password twice. Adding the additional
    # prompts to catch it when this occurs.
    # Usage Example:
    # antivirus engine modify -vendor sophos -sophos-http-user uk09130201 -password sophos_password
    } elsif ($cmd =~ /antivirus_engine_(option_)?modify/) {
        $$cmd_hash{additional_prompts_timeout} = 60;
        $$cmd_hash{additional_prompts} = {
                '(Please enter|Enter the) password:' => "-password",
                '(Please enter|Enter the) password again:' => "-password",
                "::.?>" => "FINAL_PROMPT",
        };
    } elsif ($cmd eq "antivirus_on_demand_report_upload") {
        $$cmd_hash{additional_prompts_timeout} = 60;
        $$cmd_hash{additional_prompts} = {
                'Enter User: ' => "-user",
                'Enter Password: ' => "-password",
                "::.?>" => "FINAL_PROMPT",
        };
    # When the End User Agreement is prompted by the antivirus_engine_enable command
    # the value of the "-eula" option will be sent back to the filer. Possible values
    # are "yes", and "no". If the user wants the EULA text printed to screen, use the
    # value "text\ryes", or "text\rno". This ensures that after the text is printed,
    # it is also confirmed.
    # Usage Example:
    # $Server->antivirus_engine_enable('vendor'=>"trendmicro", '-eula'=>"yes",);
    } elsif ($cmd eq "antivirus_engine_enable") {
        $$cmd_hash{additional_prompts_timeout} = 3600;
        $$cmd_hash{additional_prompts} = {
                '{yes|no|text}:' => "-eula",
                "::.?>" => "FINAL_PROMPT",
        };
 	} elsif ($cmd =~ /(.+)_modify/) {
	    #$Trace->trace(msg=>"Skipping verification of modified commands due to burt298304");
	} elsif (grep($cmd eq $_, qw(volume_copy volume_copy_start volume_move volume_move_start volume_member_move volume_member_move_start volume_mirror_replicate volume_clone_split volume_clone_split_start))) {
	    $$cmd_hash{verification} = "_volop_verify";
	} elsif ($cmd =~ /(.+)_delete/) {
	    # we will have an exists method, so set the verifier for delete
	    $$cmd_hash{negative_verification} = "$1_exists";

	    # if this is a delete() command then create a new cdef for an exists() routine
	    #  since the delete() params should be all that is needed to uniquely identify this resource
	    my $exists_cmd = "$1_exists";
	    %{$new_commands{$exists_cmd}} = %{$cmd_hash};
	    $new_commands{$exists_cmd}{command} =~ s/delete(-all)?/show/;
	    delete $new_commands{$exists_cmd}{optional_options};
	    delete $new_commands{$exists_cmd}{negative_verification};
	    $new_commands{$exists_cmd}{internal_eval} = 1;
	    $new_commands{$exists_cmd}{output_handler} = "_cli_output_handler";

	    # Deleting the cifs server will cause a userid/passwd prompt
	    if ($cmd =~ /vserver_cifs_delete/) {
		$$cmd_hash{additional_prompts} = {
                    # Tricky
                    "Please enter your userid:" => "-userid",
		    "Please enter your password:" => "-password",
                    # >= BR
		    "Enter the user name:" => "-userid",
		    "Enter the password:" => "-password",
		    "::.?>" => "FINAL_PROMPT",
		};
		$$cmd_hash{additional_prompts_timeout} = 60;
	    }
        } else {
            if ($cmd eq "storage_disk_setstate") {
	        $$cmd_hash{additional_prompts} = {
	            "y or n" => "-yesno",
                    "::.?>"    => "FINAL_PROMPT",
	            };
            $$cmd_hash{additional_prompts_timeout} = 30;
	    }
	}
    }
    # add the new cdefs to the master list
    grep($self->{Commands}->{$_} = $new_commands{$_}, keys %new_commands);
    
    # Define the Compatability hash (Server-wide, not command centric)
    _setup_compatability($self);

    $Trace->exit();
}

# This routine is used when '-Help' is passed into a cdef-driven method.
#  It parses the output of a command when the '?' is passed into it.
#  It will return an array of hashes. Each hash represents a field that
#  can be passed into the command. It will look like:
#  [
#   {
#    field => "volume",
#    value => "volume name",
#    description => "Volume Name",
#   },
#   ]
sub _cli_help_handler {
    my ($self, %params) = @_;
    $Trace->entry();
    my $want_array = $params{want_array};
    $want_array = [] if (!$want_array);
    my $output = $params{output_in};
    my $prev_output;
    foreach my $line (split(/\n/, $output)) {
        my %output = ();
        if ($line =~ /^\s*\[\[-(\S+)\] \<(.+)\>\]\s+(.+)$/) {
            $output{field} = $1;
            $output{value} = $2;
            $output{description} = $3;
        } elsif ($line =~ /^\s*\[\[-(\S+)\] \((.+)\)\]\s+(.+)$/) {
            $output{field} = $1;
            $output{value} = $2;
            $output{description} = $3;
        } elsif ($line =~ /^\s*\[\[-(\S+)\] \<(.+)\>\]$/) {
            $output{field} = $1;
            $output{value} = $2;
        } elsif ($line =~ /^\s*\[\[-(\S+)\] \((.+)\)\]$/) {
            $output{field} = $1;
            $output{value} = $2;
        } elsif ($line =~ /^\s*\[-(\S+)\]\s+$/) {
            $output{field} = $1;
        } elsif ($line =~ /^\s+(.+)$/) {
            $prev_output->{description} = $1;
            next;
        }
        $prev_output = \%output;
        push(@$want_array, \%output);
    }
    $Trace->entry();
    return $want_array;
}

#***********************************************************************
#
# This is a generic verification routine for the show commands. This method
#   will compare the paramaters to the output. It uses the OutputConversion
#   hash to determine the relationship between the output text and the 
#   paramater names
#
#***********************************************************************

sub _cli_output_handler {
    my ($self, %params) = @_;
    $Trace->entry();
    my $want_array = $params{want_array};
    $want_array = [] if (!$want_array);
    my $output = $params{output_in};
    my $converted_params = $params{converted_params};

    
    my @tmp_output = split(/\n/, $output);
    my @fields = split($CLI_separator, shift @tmp_output);
    my @heading = split($CLI_separator, shift @tmp_output);
    my @return_output = ();

    # Some output still contains embedded newlines in the fields so we cannot
    #  totally depend on the @tmp_output containing a single line for each row in
    #  in the CLI. So, loop through it once and find any elements of @array
    #  (the row) which do not contain the right number of elements (values
    #  for the row). The right number should be the number of elements of 
    #  @heading.
    #
    # For any discrepencies, try to reform the row by combining multiple
    #  elements of @tmp_output into a new @output.
    my @output;
    my $partial_line = "";
    foreach my $line (@tmp_output) {
        # just skip blank lines
        next if ($line =~ /^\s*$/);
	if ($line =~ /^(INFO|Info)/) {
	    if (!$params{-SkipINFO}) {
		$Trace->warn(msg=>"Found debug line. Ignoring: \"$line\"");
	    }
	    next;
	}
    #Ignore any separator entry that is returned via the debug_log_show cmd
    elsif ($line =~ /-showseparator \"X;X\"/) { 
	    if (!$params{-SkipINFO}) {            
		$Trace->debug(msg=>"Found output separator config line. Ignoring: \"$line\"");
	    }
	    next;
	}
        my @values = split($CLI_separator, $line);
        if (@values != @heading) {
            $partial_line .= $line;
            
            # make sure we have seen the end of the row
            next if ($partial_line !~ /$CLI_separator$/);

            my @partial = split($CLI_separator, $partial_line);
            if (@partial == @heading) {
                $Trace->debug(msg=>"done with merge: $partial_line");
                push(@output, $partial_line);
                $partial_line = "";
            } elsif (@partial > @heading) {
                $Trace->comment(msg=>"line=$line");
                $Trace->comment(msg=>"partial_line=$partial_line");
                $Trace->fatal(msg=>"Failed to handle embedded newlines in the CLI output; details in the log");
            }
        } else {
            push(@output, $line);
        }
    }

    foreach my $line (@output) {
        my %output = ();
        my $index=0;
        foreach my $field (split($CLI_separator, $line)) {
            ##################################################################
            # NOTE: The splitting of a list field on a comma will break
            # if the field is of type list AND accepts a list of text values
            # (or some other field type) that supports a comma.
            # We have not seen instances of this type of field yet but will 
            # evaluate if problems occur
            ##################################################################
            
            ##################################################################
            # If the field starts with quotes, split the individual strings
            # separated by comma-space. Otherwise, treat entire thing as one 
            # string. Some job output is missing the trialing ".
            ##################################################################
            if ($field =~ /^\".+\"?$/ || $field =~ /^\{.+\}$/) {
                if ($field =~ /, \"/) {
                    my $array_ref = [split(/, /, $field)];
                    $Trace->debug(msg=>"array formed because of '$field'");
                    $output{$fields[$index]} = $array_ref;
                } else {
                    $output{$fields[$index]} = $field;
                }
            } elsif ($field =~ /, ?/) {
                my $array_ref = [split(/, ?/, $field)];
                $Trace->debug(msg=>"array formed because of '$field'");
                $output{$fields[$index]} = $array_ref;
            } else {
                $output{$fields[$index]} = $field;
            }
            # If this param was converted from an old param, stuff the old one too
            if ( defined $converted_params->{$fields[$index]} ) {
                $Trace->debug(msg=>"Field: $fields[$index] was converted from field: " . 
                              $converted_params->{$fields[$index]} . ", stuffing the old field into the output");
                $output{ $converted_params->{$fields[$index]} } = $field;
            }
            $index++;
        }
        push(@$want_array, \%output);
    }
    $Trace->debug(msg=>"want_array is:\n" . Dumper( $want_array ) );

    # we have output and we are supposed to verify the output
    if ($output ne "" && $params{internal_verification}) {
	if (scalar @$want_array > 1) {
	    $Trace->fatal(msg=>"Found more than 1 matching result during verification",
			  result=>"FATAL");
	}
	delete $params{output_in};
	delete $params{want_array};
	delete $params{internal_verification};
	delete $params{converted_params};
	while (my ($key, $value) = each %params) {
	    ##################################################################
	    # If the field input to the server object was a scalar reference
	    # or had no reference, but the output was an array, convert the 
	    # output array into a scalar for comparision
	    #
	    # To ensure that the delimiters match, change ",\s*(\S)" to
	    # only contain a single space ", $1"  
	    # 
	    # To ensure the same sorting, convert input string to an array
	    # sort it, and then write it back
	    ##################################################################
	    if ( ((ref($value) eq "SCALAR") || (not ref $value)) && ref($$want_array[0]->{$key}) eq "ARRAY") {
		my $output_string = join(", ", sort(@{$$want_array[0]->{$key}}));
		my $input_string = $value;
		$input_string =~ s/,\s*(\S)/, $1/g;
		my @input_array = split(/, /, $input_string);
		$input_string = join(", ", sort(@input_array));
		if ($input_string ne $output_string) {
		    $Trace->fatal(msg=>"Verification failure for list field '$key': '$output_string' is not set to '$input_string'",
                              result=>"FATAL");
		}
	    ##################################################################
            # If the field input to the server object was an array reference
            # and the output is an array reference, convert both arrays
            # into a scalar for comparision
            ##################################################################
	    } elsif (ref($value) eq "ARRAY" && ref($$want_array[0]->{$key}) eq "ARRAY") {
                my $output_string = join(", ", sort(@{$$want_array[0]->{$key}}));
                my $input_string = join(", ", sort(@{$value}));
                if ($input_string ne $output_string) {
                    $Trace->fatal(msg=>"Verification failure for list field '$key': '$output_string' is not set to '$input_string'",
                              result=>"FATAL");
                }
	    ##################################################################
	    # If neither input nor output were array references, than use the
	    # values 'as is'
	    ##################################################################
	    } elsif ($value ne $$want_array[0]->{$key}) {
		$Trace->fatal(msg=>"Verification failure for field '$key': '$$want_array[0]->{$key}' is not set to '$value'",
			      result=>"FATAL");
	    }
	}
    }
  done: 
    $Trace->exit(msg => "Exiting with $want_array");
    return $want_array;
} # cli_output_handler

sub compare_version {
    my $self = shift;
    my %params = (release=>undef,
		  build=>undef,
		  operator=>">=",
		  @_);
    my $rc = 1;
    $Trace->entry(msg=>"release=$params{release}, build=>$params{build}, operator=>'$params{operator}'");

    if ($params{release}) {
	$Trace->fatal(msg=>"the 'release' parameter is not supported yet", result=>"SCRIPT");
    }
    if (defined($params{build})) {
	if ($params{operator} eq ">=") {
	    $rc = ($self->{build} >= $params{build});
	} elsif ($params{operator} eq "<") {
	    $rc = ($self->{build} < $params{build});
	}
    }
    $Trace->exit(msg=>"rc=$rc");
    return $rc;
}


# Return 1 if the filer has some passed uichange, else return undef
sub has_change {
    my $self = shift;
    my %params = ( change => undef,
		   @_ );

    for my $change (@{$self->{uichanges}}) {
	if ( $change eq $params{change} ) {
	    return 1;
	}
    }
    return undef;
}


# Return 1 if the filer has debug bits loaded
sub isDebug {
    my $self = shift;

    if ( $self->{buildopts} eq "debug" ) {
	return 1;
    }

    return undef;
}

sub _get_local_node_info {
    my $self = shift;
    my $local_node_info;
    my @local_node_output;
    my $local_node_output;

    $local_node_output = $self->cli(cmd=>"node show local");

    $self->_cli_output_handler(output_in=>$local_node_output, want_array=>\@local_node_output);

    $local_node_info = $local_node_output[0];
    $Trace->debug(msg=>"This is the local_node: $local_node_info->{node}");

    return $local_node_info;
    $Trace->exit();
}

sub _get_version_info {
    my $self = shift;
 
    $self->{build} = "unknown";
    $self->{branch} = "unknown";
    $self->{release} = "unknown";
    $self->{buildopts} = "unknown";
    $self->{buildpath} = "unknown";
    $Trace->entry();

    # Call these directly since it occurs prior to the cdefs being loaded
    my @output;
    my $output;
    my $ident_output;

    # Grab reference to local node info hash
    my $local_node_info = $self->_get_local_node_info();

    # Now get the known UI changes
    eval {
        $output = $self->cli(cmd=>"system uichanges -node $local_node_info->{node}");
    };
    if ($@ || $output =~ /(ERROR|Error)/) {
        eval {
            $output = $self->cli(cmd=>"node uichanges -node $local_node_info->{node}");
        };
    }
    if ($@ || $output =~ /(ERROR|Error): invalid argument \"-node\"/ ) {
	    $output = $self->cli(cmd=>"filer uichanges -filer $local_node_info->{node}");
    }
    eval {
	    $self->_cli_output_handler(output_in=>$output, want_array=>\@output);
    };
    # An error may mean we're in some pre-uichanges build
    if ( $@ ) { 
	if ( $@ =~ /not a recognized command/ ) {
	    $Trace->debug(msg=>"uichanges not supported in this build");
	} elsif ( $@ =~ /This table is currently empty/ ) {
	    # Or that there are no changes to be had, odd, but true
	    $Trace->debug(msg=>"uichanges reports no changes");
	} else {
	    # Or that we just had some error
	    $Trace->fatal(msg=>"Unknown error getting uichanges: $@",
			  result=>"CONFIG");
	}
    } else {
	# For each returned line
	foreach my $uichanges_ref (@output) {
	    next if ( ! $$uichanges_ref{uichanges} );
	    @{$self->{uichanges}} = @{$$uichanges_ref{uichanges}};
        }
        if (ref($self->{uichanges})) {
            $Trace->debug( msg=>"This build has uichanges: " . 
                join(" ", @{$self->{uichanges}}) );
        }
    }

    my $find_release_info = 1;

    # Find the specifics of this version of the product. Look in Options first and
    #  fall back on querying the node
    #
    my ($branch,$release,$build,$buildopts,$buildpath) = parse_release_info(node=>$local_node_info->{node});
    if ($branch) {
        # The parameter validation made sure we had the following 5 fields present
        $self->{branch} = $branch;
        $self->{release} = $release;
        $self->{build} = $build;
        $self->{buildopts} = $buildopts;
        $self->{buildpath} = $buildpath;
        $find_release_info = undef; # we do not need to query the node
    }

        
    if ($find_release_info) { # query the node for the information
        my $do_legacy_discover = 0;
        #
        # :BUG:05/27/2011 12:13:08 PM:dps:
        # This code should check the current priv level so that we can switch it back to whatever the
        # current level is rather than just assuming 'diag'.  Internally to Server this isn't a
        # problem because we only ever call this function on a connect - which will always default
        # to "diag" priv level.
        #
        eval {
            $self->cli(cmd => "set -privilege test");
        };
        # If we are unable to set test privilege then fall back to legacy discovery
        if ($@ && $@ =~ /(ERROR|Error): "test" is an invalid value/) {
            $Trace->trace(msg =>
                "Cannot set privilege level to test, must do legacy release discovery");
            $do_legacy_discover = 1;
        } else {
            my @output;
            @output=();
            eval {
                # This call has to go through cli() rather than system_image_show()
                # because the latter depends on cdefs and this routine we are in is
                # run prior to cdefs being loaded
                my $output = $self->cli(cmd=>"system image show -node $local_node_info->{node} -iscurrent true");
                $self->_cli_output_handler(output_in=>$output, want_array=>\@output);           
            };          
            if ($@){
                $Trace->fatal(msg=>"Unable to retrieve the node release: $@",
                      result=>"CONFIG");
            }
            # push the CLI back to diag mode
            $self->cli(cmd => "set -privilege diag");
            #try to get the release information from the image show command
            #otherwise probe some files through legacy discovery
            if (exists($output[0]->{'build-info'}) && $output[0]->{'build-info'} =~ /BRANCH/) {
                $Trace->trace(msg => "Using image information for node discovery");
                $Trace->trace(msg => Dumper($output[0]));
                $self->{release} = ( defined $output[0]->{version} ) ? $output[0]->{version} : "unknown";
                $self->{build} = ( defined $output[0]->{change} ) ? $output[0]->{change} : "unknown";
                $self->{buildopts} = ( defined $output[0]->{'build-type'} ) ? $output[0]->{'build-type'} : "unknown";
                $self->{buildpath} = ( defined $output[0]->{'build-path'} ) ? $output[0]->{'build-path'} : "unknown";
                $find_release_info = undef;
                my %build_info;
                my $build_info_raw = $output[0]->{'build-info'};
                $build_info_raw =~ s/"//g;

                if ( $build_info_raw =~ /BRANCH: (.+) backingBranch: (.+) maxBBChg:/ ) {  # this build is from a sparse branch
                    $Trace->debug(msg => "BRANCH = '$1' backingBranch = '$2'");
                    $self->{branch} = $1;
                    $self->{backing_branch} = $1;
                } elsif (  $build_info_raw =~ /BRANCH: (\S+)( maxBBChg:)?/ ) { 
                    $Trace->debug(msg => "BRANCH = '$1'");
                    $self->{branch} = $1;
                } else {
                    $Trace->warn(msg =>
                        "Unable to determine the branch from the build-info returned by DOT '$build_info_raw'");
                }
            } else {
                $do_legacy_discover = 1;
            }
        }
        if ($do_legacy_discover) {
            my $ident_output;
            $Trace->trace(msg => "Performing legacy node discovery");
            # Get the information from the /BUILD file since we may want to use it
            $output = $self->shell(cmd => "cat /BUILD");
            my %build_info;
            foreach my $line (split(/\n/, $output)){
                chomp($line);
                my ($key, $value) = split(/:\s+/, $line);
                if (defined($value) && length($value) && ($value !~ /(unknown|p4down)/)){
                    $build_info{$key} = $value;
                }
            }

            $Trace->debug(msg => "Got info " . Dumper(\%build_info));
        
            # Now get the branch info
            if (exists($build_info{BRANCH})){
                $self->{branch} = $build_info{BRANCH};
                $Trace->debug(msg => "Using /BUILD info for branch");
            }

            # if we have a backing branch - squirrel that away too for later
            if (exists($build_info{backingBranch})){
                $Trace->debug(msg => "Using /BUILD info for backingBranch");
                $self->{backing_branch} = $build_info{backingBranch};
            }
                
            # Now get the release info
            # try getting it from /BUILD first then ident
            # :BUG:05/27/2011 03:29:21 PM:dps: This needs to look in /VERSION, not /BUILD
            if (exists($build_info{RELEASE})){
                $self->{release} = $build_info{RELEASE};
                $Trace->debug(msg => "Using /BUILD info for release");
            } else {
                $Trace->debug(msg => "Using ngsh for release");
                # Retrieve product release info from mgmt
                eval {
                    @output=();
                    # This call has to go through cli() rather than system_image_show() 
                    # because the latter depends on cdefs and this routine we are in is
                    # run prior to cdefs being loaded
                    my $output = $self->cli(cmd=>"system image show -node $local_node_info->{node} -iscurrent true");
                    $self->_cli_output_handler(output_in=>$output, want_array=>\@output);
                    $self->{release}=$output[0]{version};
                };
                if ($@){
                    $Trace->fatal(msg=>"Unable to retrieve the node release: $@",
                        result=>"CONFIG");
                }
            }

            # Now get the build info
            if (exists($build_info{CHANGE})){
                $self->{build} = $build_info{CHANGE};
                $Trace->debug(msg => "Using /BUILD info for change");
            }
                
            # Now get the BuildOpts info
            if (exists($build_info{BUILD_OPTS})){
                $self->{buildopts} = $build_info{BUILD_OPTS};
                $Trace->debug(msg => "Using /BUILD info for build_opts");
            } else {
                $ident_output = $self->shell(cmd=>"ident /boot/modules/maytag.ko | grep '\$Ntap'",-Timeout => 60) if !$ident_output;
                if ($ident_output =~ /\$NtapBuildOpts:\s+(\S+)\s+\$\s*$/m){ 
                    $self->{buildopts} = $1;
                    $Trace->debug(msg => "Using ident info for build_opts");
                }
            }

            # Now get the BuildPath info
            $ident_output = $self->shell(cmd=>"ident /boot/modules/maytag.ko | grep '\$Ntap'",-Timeout => 60) if !$ident_output;
            if ($ident_output =~ /\$NtapBuildPath:\s+(\S+)\s+\$\s*$/m){ 
                $self->{buildpath} = $1;
                $Trace->debug(msg => "Using ident info for build_path");
            }
        }
        my $options = $Options->get_hash();
        my $release_info_array = $options->{NODE_RELEASE_INFO};

        push(@$release_info_array, join($Release_separator, $local_node_info->{node}, $self->{branch}, 
                                        $self->{release}, $self->{build}, $self->{buildopts}, $self->{buildpath}));

        my $release_info = join(",", @$release_info_array);

        # throw a PARAM result so that followup scripts don't have to find the same information
        main::logresult("PARAM", "NODE_RELEASE_INFO=\"$release_info\"");

        # Update Options so that all nodes will pick up this change
        $Options->set(NODE_RELEASE_INFO=>$release_info_array);
    }

    $Trace->exit(msg=>"build=$self->{build}, release=$self->{release}, branch=$self->{branch}, buildopts=$self->{buildopts}, buildpath=$self->{buildpath}");
}

sub _validate_cli_parameters {
    my $self = shift;
    my %params = (definition => undef,
		  cmd_options => undef,
		  cmd_params => undef,
		  @_);

    my %definition = %{$params{definition}};
    my %cmdOptions = %{$params{cmd_options}};
    my %cmdParams = %{$params{cmd_params}};
    my $command = "";
    
    ################################################################################
    # Now fill out the rest of the command with the option and paramaters
    # We will complain if the required option does not exist in the cmdOption hash
    # We will also complain if the required paramater is not defined
    #
    # Since parameter values can be "0", must check that param is defined rather
    # than that it contains a value. If we just check that it contains a value,
    # than conditional will fail if value is "0"
    ################################################################################
    foreach my $option (@{$definition{optional_options}}) {
	if (defined($cmdOptions{$option}) && defined($params{$option})) {
	    if ( ($cmdOptions{$option} =~ /^-/) && ($params{$option} eq "true") ) {
		$command .= " $cmdOptions{$option}";
	    } elsif ( ($cmdOptions{$option} =~ /^-/) && ($params{$option} eq "false") ) {
		next;
	    } elsif (($params{$option} eq "true") || ($params{$option} eq "false")) { 
		$command .= " $cmdOptions{$option}";
	    } elsif ($cmdOptions{$option} eq " ") {
		$command .= " $params{$option}";
	    } else {
		$command .= " $cmdOptions{$option} $params{$option}";
	    }
	} elsif (!exists($cmdOptions{$option}) && !(grep($option ne $_, qw(Wait Verify)))) { # hack
	    $Trace->fatal(msg=>"Optional '$option' does not exist as a Parameter", result=>"SCRIPT");
	} else {
	    $Trace->debug(msg=>"$option was not specified");
	}
    }
    foreach my $option (@{$definition{required_options}}) {
	if (!$cmdOptions{$option}) {
	    $Trace->fatal(msg=>"Required '$option' does not exist as a Parameter", result=>"SCRIPT");
        }
        if ( ($cmdOptions{$option} =~ /^-/) && ($params{$option} eq "true") ) {
            $command .= " $cmdOptions{$option}";
        } elsif ( ($cmdOptions{$option} =~ /^-/) && ($params{$option} eq "false") ) {
            next;
        } elsif (($params{$option} eq "true") || ($params{$option} eq "false")) {
            $command .= " $cmdOptions{$option}";
	} elsif ($cmdOptions{$option} eq " ") {
            $command .= " $params{$option}";
        } else {
            $command .= " $cmdOptions{$option} $params{$option}";
        }
    }
    # Loop through all the user passed options, make sure they're defined
    my ($found, $user_option);
    foreach $user_option (keys %cmdParams) {
	if ( $user_option =~ /^-/ ||
	     $user_option eq 'AUTOLOAD_COMMAND' ) {
	    next;
	}
	$found = 0;
	foreach my $req_option (@{$definition{required_options}}) {
	    if ( $user_option eq $req_option ) {
		$found = 1;
		last;
	    }
	}
	if ( ! $found ) {
	    foreach my $opt_option (@{$definition{optional_options}}) {
		if ( $user_option eq $opt_option ) {
		    $found = 1;
		    last;
		}
	    }
        }
	# If we didn't find the user passed option in the definition, fail the script.
	if ( ! $found ) {
	    $Trace->fatal(msg=>"Option: '$user_option' does not exist as a field for this command", result=>"SCRIPT");
	}
    }

    return $command;
}

sub convert_export_path {
    my $self = shift;
    my %params = (export => undef,
		  @_);

    my $export;
    my $ro_export = $params{export};
    $Trace->entry(msg=>"export=$params{export}");

    # handle mirrors 
    $export = "/.admin" if ($params{export} !~ m!^/.admin!);

    # make sure we have the leading /
    $export .= "/" if ($params{export} !~ m!^/!);
    
    $export .= $params{export};
    $Trace->exit(msg=>"return $export, $ro_export");
    return ($export, $ro_export);
}

# Temporary permission set method
sub volume_perms {
  my $self = shift;
  my %params = (
		permissions => 777,
		zsmtimeout => 5,
		@_
	       );
  my %ret;

  $Trace->entry();
  if (!$self->{options_object}->get(option=>"SETVOLPERMS")){
    $Trace->exit();
    return;
  }

  # These parameters aren't always displayed the same way that they are passed which screws up the show
  delete $params{user};
  delete $params{group};
  delete $params{size};

  my $ret = $self->volume_show(%params);
  foreach my $volume (@$ret){
    # We need to run the zsmcli command against the proper machine so track down the
    # DBlade holding the volume
    my $vidhex = sprintf("%08x", $volume->{dsid});

    my $info = $self->shell(cmd => "vldbtest");
    my $vldbtest = "vldbtest";
    $vldbtest .= " -I" if ($info =~ /-I/);
    $vldbtest .= " 0";
    $vldbtest .= " -H" if ($info =~ /-H/);
    $vldbtest .= " localhost";
    # Look up the volume
    $info = $self->shell(cmd => "$vldbtest getrecord -d $volume->{dsid}");
    my $aggrid;
    if ($info =~ /aggrID:\s(\S+)/){
      $aggrid = $1;
    } else {
      push @{$ret{"noaggr"}}, "$volume->{vserver}.$volume->{volume}";
      next;
    }

    # Look up the aggregate
    $info = $self->shell(cmd => "$vldbtest getaggrlocation -a $aggrid");
    my $bladeid;
    if ($info =~ /bladeID:\s(\S+)/){
      $bladeid = $1;
    } else {
      push @{$ret{"nobladeid"}}, "$volume->{vserver}.$volume->{volume}";
      next;
    }

    # look up the blade
    $info = $self->shell(cmd => "$vldbtest getbladeinfo -b $bladeid");
    my $filer;
    if ($info =~ /hostIPAddr1:\s(\S+)/){
      $filer = $1;
    } else {
      push @{$ret{"nohostname"}}, "$volume->{vserver}.$volume->{volume}";
      next;
    }

    # We now have the filer name so light up the virtual chmod
    $Trace->comment(msg => "$volume->{vserver}.$volume->{volume} lives on $filer");
    my $retry = 10;
  setperm:
    {
      my $cmd = "zsmcli -t $params{zsmtimeout} $filer root \"\" d-volume-set-info id=0x$vidhex volume-attrs='[d-volume-info=[security=[d-security-info=[unix=[d-security-unix-info=[permissions=$params{permissions}]]]]]]' desired-attrs=security-unix-permissions";
      my $cmdret = $self->shell(cmd => $cmd, -Timeout => 105);
      if ($cmdret =~ /status=\"([^\"]+)\"/){
	my $result = $1;
	if ($result !~ /passed/i){
	  redo setperm if $retry--;
	  $Trace->fatal(msg => "Command result is $result");
	}
      } else {
	$Trace->fatal(msg => "Unexpected command result => $cmdret");
      }
    }
  }
  $Trace->exit();
  %ret;
}

# Overridden methods

sub vserver_new {
  my $self = shift;

  # Set up the parameters
  my %params = (AUTOLOAD_COMMAND => ref($self)."::vserver_new",
		@_);
  $Trace->entry();
  
  # Check for SETVOLPERMS and if it exists and unix-permissions is not defined then set to 777
  my $setvolperms = $self->{options_object}->get(option=>"SETVOLPERMS");
  my $uichange = (grep({"unix-permissions" eq $_} @{$self->{uichanges}}));
  $Trace->trace(msg=>"UICHANGE=$uichange");
  $Trace->trace(msg=>"SETVOLPERMS=$setvolperms");
  $Trace->debug(msg=>Dumper(\%params));
  if (($self->{options_object}->get(option=>"SETVOLPERMS"))&&(!$params{"unix-permissions"})&&($uichange)) {
      $params{"unix-permissions"} = 777;
      $Trace->trace(msg=>"unix-permissions=".$params{"unix-permissions"});       
  }

  # Call the command
  my $rv = $self->run_command(%params);

  # if the build does not support unix-permissions then set the permissions using back-door
  $Trace->debug(msg=>"This build had these UI change support: ".Dumper($self->{uichanges}));
  unless ($uichange) {
      $Trace->comment(msg=>"This build does not support unix-permissions on vserver_new so using back-door");
      my %ret = $self->volume_perms(vserver     => $params{"vserver"},
                                    volume      => $params{"rootvolume"},
                                    permissions => $params{"unix-permissions"});  
  }

  $Trace->exit();
  $rv;
}

sub volume_new {
  my $self = shift;

  # Set up the parameters
  my %params = (AUTOLOAD_COMMAND => ref($self)."::volume_new",
		@_);
  $Trace->entry();

  # Check for SETVOLPERMS and if it exists and unix-permissions is not defined then set to 777
  my $setvolperms = $self->{options_object}->get(option=>"SETVOLPERMS");
  my $uichange = (grep({"unix-permissions" eq $_} @{$self->{uichanges}}));
  $Trace->trace(msg=>"UICHANGE=$uichange");
  $Trace->trace(msg=>"SETVOLPERMS=$setvolperms");
  $Trace->debug(msg=>Dumper(\%params));
  if (($self->{options_object}->get(option=>"SETVOLPERMS"))&&
      (!$params{"unix-permissions"})&&
      ($uichange) && 
      ($params{"security-style"} !~ /^ntfs$/i) &&
      ($params{"type"} !~ /^DP$/i)) {
      $params{"unix-permissions"} = 777;
      $Trace->comment(msg=>"unix-permissions=".$params{"unix-permissions"});       
  }

  # Call the real command
  my $rv = $self->run_command(%params);
  
  # if the build does not support unix-permissions then set the permissions using back-door
  $Trace->debug(msg=>"This build had these UI change support: ".Dumper($self->{uichanges}));
  unless ($uichange) {
      $Trace->trace(msg=>"This build does not support unix-permissions on volume_new so using back-door");
      $self->volume_perms(%params);
  }

  $Trace->exit();
  $rv;
}

sub volume_modify {
  my $self = shift;

  # Set up the parameters
  my %params = (@_);
  my $rv = 0;
  $Trace->entry();
  # The following if condition has been commented to resolve Burt 340263.
  # if (grep (!/^-/, keys %params) > 2){
    $params{AUTOLOAD_COMMAND} = ref($self)."::volume_modify";
    $rv = $self->run_command(%params);
  # }
  $Trace->exit();
  $rv;
}

sub get_release {
    my $self = shift;
    my %params = (
		  format => "raw",
		  @_, # overlay any parameters that were explicitly passed
		  );
    my $rel;
    $Trace->entry();
    # The following table allows for a normalization of both internal and external release
    # strings into the internal (code name). For example, a release of "10.0.RC3" will be
    # translated into "Tricky.0". The default behavior, however, is to not do any tranlation.
    # Calling this subroutine with a parameter of 'format => "cooked"' will translate it.
    my %release_table = (
			 "Tricky.0" => "Tricky.0",
			 "10.0RC" => "Tricky.0",

			 "Tricky.1" => "Tricky.1",
			 "10.0.1" => "Tricky.1",

			 "Tricky.2" => "Tricky.2",
			 "10.0.2" => "Tricky.2",

			 "Tricky.3" => "Tricky.3",
			 "10.0.3" => "Tricky.3",

			 "Tricky.4" => "Tricky.4",
			 "10.0.4" => "Tricky.4",

			 "Boilermaker.0" => "Boilermaker.0",
			 "Boilermaker__8.0.0" => "Boilermaker.0",
 			 "8.0X" => "Boilermaker.0", 
 			 "8.0RC" => "Boilermaker.0",

			 "Boilermaker.1" => "Boilermaker.1",
			 "Boilermaker__8.0.1" => "Boilermaker.1",
			 "8.0.1X" => "Boilermaker.1",
			 "8.0.1RC" => "Boilermaker.1",
             
			 "Boilermaker.2" => "Boilermaker.2",
			 "Boilermaker__8.0.2" => "Boilermaker.2",
			 "8.0.2X" => "Boilermaker.2",
			 "8.0.2RC" => "Boilermaker.2",

			 "RollingRock__8.1.0" => "Rollingrock.0",
                         # This will need to change once we get a branch for Rollingrock
			 "MainNG__10.0.99" => "Rollingrock.0",
			 "RollingRock.0" => "Rollingrock.0",
			 "fsbN" => "Rollingrock.0",

			 "Boilermaker.3" => "Boilermaker.3",
			 "Boilermaker__8.0.3" => "Boilermaker.3",
			 "8.0.3X" => "Boilermaker.3",
			 "8.0.3RC" => "Boilermaker.3",

			 );
    if ( $params{format} =~ m/cooked/i ) {
	foreach my $key (keys %release_table) {
	    if ( substr($self->{release},0,length($key)) eq $key ) {
		$rel = $release_table{$key};
	    }
	}
    } else {
	$rel = $self->{release};
    }
    $Trace->exit(msg=>"build=$self->{release}, release=$rel");
    return $rel;
}

# The NG object supports the specification of initial mode bits on 
# a volume so this method should return true.
sub supports_mode_in_volume_create {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"yes");
    return 1;
}

# The NG object supports a GNS or submounted volumes so this method 
# returns true
sub supports_GNS {
    my $self = shift;
    $Trace->entry();
    $Trace->exit(msg=>"yes");
    return 1;
}

sub _find_all_servers {
    my $self = shift;
    $Trace->entry();
    
    my $cluster_hash = $self->cluster_show();
    my @nodes;

    foreach my $node (@$cluster_hash) {
	if ($node->{eligibility} =~ /true/i) {
	    push(@nodes, $node->{node});
	} else {
	    $Trace->warn(msg=>"Skipping '$node->{node}' because its eligibility is marked as '$node->{eligibility}'");
	}
    }

    $Trace->exit(msg=>"return @nodes");
    return @nodes;
}


# this routine is called from Node::_setup_commands() and needs to be defined in this namespace
# so that the routines found inside the libraries being loaded will also be in this namespace
sub _load_modules {
    my ($self, $class, @modules) = @_;
    $Trace->entry(msg => "class => '$class' modules => '@modules'");
    no strict 'refs';

    foreach my $module (@modules) {
        eval "use ${class}::$module";
        if ($@) {
            $Trace->fatal(msg => "Failed to load module '${class}::$module' '$@'", result => "SCRIPT");
        } 
        my $setup = "${class}::Setup${module}Commands";
        $Trace->debug(msg => "Attempting to access '$setup'");
        if (!&$setup($self)) {
            $Trace->trace(msg=>"Could not find the method '$setup'");
        }
    }
    $Trace->exit();
}

sub _setup_login_credentials {
    my $self = shift;
    $Trace->entry();

    my $options = $Options->get_hash();
    my ($branch,$release,$build,$buildopts,$buildpath) = parse_release_info(node=>$self->{server});

    # find the username/password for the shell connection
    # Look for the username/password as nate params first
    if ( defined($options->{SHELL_USER})) {
        $self->{SHELL_USER} = $options->{SHELL_USER};
        $Trace->trace(msg=>"set SHELL_USER to '$self->{SHELL_USER}' from the NATE param\n");
    }
    if ( defined($options->{SHELL_PASSWORD})) {
        $self->{SHELL_PASSWORD} = $options->{SHELL_PASSWORD};
        $Trace->trace(msg=>"set SHELL_PASSWORD to '$self->{SHELL_PASSWORD}' from the NATE param\n");
    }

    # next, look for a connection definition in the nate_hosts file. ssh is ok but shell is preferred
    foreach my $elem ("ssh", "shell") {
        my $user = $self->{host_object}->{connection_fields}{$elem}{username};
        my $pass = $self->{host_object}->{connection_fields}{$elem}{password};
        if ( !defined($self->{SHELL_USER}) && defined($user)) {
            $self->{SHELL_USER} = $user;
            $Trace->trace(msg=>"set SHELL_USER to '$self->{SHELL_USER}' from the $elem entry in the host object\n");
        }
        if ( !defined($self->{SHELL_PASSWORD}) && defined($pass)) {
            $self->{SHELL_PASSWORD} = $pass;
            $Trace->trace(msg=>"set SHELL_PASSWORD to '$self->{SHELL_PASSWORD}' from the $elem entry in the host object\n");
        }
    }
    
    # If we still did not get values, use the defaults
    if ( !defined($self->{SHELL_USER})) {
        if ($options->{RELEASE} =~ /tricky|^10.0/ || $release =~ /tricky|^10.0/) {
            $self->{SHELL_USER} = "root";
        } else {
            $self->{SHELL_USER} = "diag";
        }
        $Trace->trace(msg=>"set SHELL_USER to the default value of '$self->{SHELL_USER}'");
        
    }
    if ( !defined($self->{SHELL_PASSWORD})) {
        if ($options->{RELEASE} =~ /tricky|^10.0/ || $release =~ /tricky|^10.0/) {
            $self->{SHELL_PASSWORD} = "admin";
        } else {
            $self->{SHELL_PASSWORD} = "netapp1!";
        }
        $Trace->trace(msg=>"set SHELL_PASSWORD to the default value of '$self->{SHELL_PASSWORD}'\n");
    }

    $self->{CLI_CONNECTION} = "cli";
    # find the username/password for the CLI connection
    # Unless the CLI_FROM_SHELL param has been set, identify the username and password for the CLI connection
    if (!$self->{cli_from_shell}) {
        # Look for the username/password as nate params first
        if ( defined($options->{CLI_USER})) {
            $self->{CLI_USER} = $options->{CLI_USER};
            $Trace->trace(msg=>"set CLI_USER to '$self->{CLI_USER}' from the NATE param\n");
        }
        if ( defined($options->{CLI_PASSWORD})) {
            $self->{CLI_PASSWORD} = $options->{CLI_PASSWORD};
            # The default password for Tricky was an empty string. However, there is no way to pass an 
            # empty string via param() and detect that it is not a param being set. So, make a special
            # string for this param which will allow this to be set.
            if ($self->{CLI_PASSWORD} eq "EMPTY_STRING") {
                $self->{CLI_PASSWORD} = "";
            }
            $Trace->trace(msg=>"set CLI_PASSWORD to '$self->{CLI_PASSWORD}' from the NATE param\n");
        }

        # If we did not get values for the username/password, look for it in the nate_hosts file
        my $user = $self->{host_object}->{connection_fields}{cli}{username};
        my $pass = $self->{host_object}->{connection_fields}{cli}{password};
        if ( !defined($self->{CLI_USER}) && $user) {
            $self->{CLI_USER} = $user;
            $Trace->trace(msg=>"set CLI_USER to '$self->{CLI_USER}' from the cli entry in the host object\n");
        }
        if ( !defined($self->{CLI_PASSWORD}) && defined($pass)) {
            $self->{CLI_PASSWORD} = $pass;
            $Trace->trace(msg=>"set CLI_PASSWORD to '$self->{CLI_PASSWORD}' from the cli entry in the host object\n");
        }

        # If we still did not get values, use the defaults
        if ( !defined($self->{CLI_USER})) {
            $self->{CLI_USER} = "admin";
            $Trace->trace(msg=>"set CLI_USER to the default value of '$self->{CLI_USER}'");
        }
        if ( !defined($self->{CLI_PASSWORD})) {
            # BR introduced new password rules so gx_boot sets the password to this well-known password. So,
            # if the password was not found in the nate_host file or the command-line, fall back on the default
            # for the release. For Tricky, it is "". Otherwise it is "netapp1!". Try to determine what release
            # we are running to pick the correct default.
            # The RELEASE param is set by a testbed or by the build_version.thpl.
            # The parse_release_info() will find information set by deux_init.pl or other previous scripts that
            #  have run a Server->new();
            # If this information is not set to tricky, then fall back on the new default.
            if ($options->{RELEASE} =~ /tricky|^10.0/ || $release =~ /tricky|^10.0/) {
                $self->{CLI_PASSWORD} = "";
            } else {
                $self->{CLI_PASSWORD} = "netapp1!";
            }
            $Trace->trace(msg=>"set CLI_PASSWORD to the default value of '$self->{CLI_PASSWORD}'\n");
        }
    } else {
        # in this case, use "ngsh -c" type commands
        $Trace->trace(msg=>"Set CLI username/password to match shell username/password");
        $self->{CLI_USER} = $self->{SHELL_USER};
        $self->{CLI_PASSWORD} = $self->{SHELL_PASSWORD};
    }
    $Trace->exit();
}
#Subroutine to take care of debug smdb commands which
#do not have CDEFs

sub _exec_debug_smdb_table
{
	 my $self = shift();
	 my %p = @_;

	 $Trace->entry();

	 my $table_name = delete $p{table};
	 my $op = delete $p{operation};
	 delete $p{'-Output'};
	 
	 my $args = join(" ", map( { "-$_ $p{$_}" } keys(%p)));
	 $Trace->comment(msg => Dumper(\%p));

	 my $output = $self->cli(
		 cmd => "debug smdb table $table_name $op $args",
     );

	 my $out_table;
	 my @lines = split(/\n/, $output);
	 
	 # create a list of headers
	 my $line = shift(@lines);
	 my @headers = split($CLI_separator, $line);

	 # remove descriptive headers
	 shift(@lines);

	 foreach my $line (@lines) {
		 my $hline;
		 my @fields = split($CLI_separator, $line);
		 foreach my $h (@headers) {
		     $hline->{$h} = shift(@fields);
	     }
		 push (@{$out_table}, $hline);
	 }

	  $Trace->debug(msg => "debug smdb table results: \n".Dumper($out_table));

	  $Trace->exit();
	  return $out_table;
 }
	 
# Return a name for the current build (in the form
# "mainngN_010203_0405<x86>")

sub get_build_name {
    my $self = shift;

    $Trace->entry();

    # Force _get_version_info to be called
    $self->establish_connections(cli_connection => 1);
    my $buildpath = $self->{buildpath};
    my $buildopts = $self->{buildopts};
    my $version = $buildpath;
    if ($version eq "unknown") {
        my $out = $self->dblade(cmd => "version -v" );
        if ($out =~ /(?:NetApp|Data ONTAP) Release\s+(.*):\s+/) {
            $version = $1;
        } else {
            $Trace->warn(msg => "Could not determine build name" );
        }
    } else {
        # "/abc/def/xyz/" -> "/abc/def/xyz"
        $version =~ s#/$##g;
        # "/abc/def/xyz/bundledefs/.." -> "/abc/def/xyz"
        $version =~ s#/[^/]+/\.\.$##g;
        # "/abc/def/xyz" -> "xyz"
        $version =~ s#.*/##g;
    }
    if ($buildopts) {
        $version .= "<$buildopts>";
    }
    $Trace->exit(msg=>"version=$version");
    return $version;
}

#***********************************************************************
#  Convenience routine to call private _cli_output_handler() which will
#  parse X;X separated CLI output and parse into an array of hashes.
#  This is automatically called for "show" commands, but some commands
#  can be coaxed into outputting in "show" format, but will not 
#  automatically be parsed.
#  
#  Example:
#     # "cluster peer ping" does not output in show format by default, 
#     # but by giving the -fields option, we can cause it to output in
#     # show-compatible format
#     my $ping = $Server->cluster_peer_ping(
#                 'destination-cluster' => $peer_cluster,
#                 'fields' => 'status',
#                );
#
#     $Server->_cli_output_handler(output_in=>$ping, want_array=>\@ping);
#
#  Options:
#      output_in  - string with imbedded X;X separators
#      want_array - array of hashes, keyed by field names
#
#***********************************************************************
sub parse_cli_output {
    my $self = shift;
    return $self->_cli_output_handler( @_ );
}

1;




