
#
# NAME
#   Misc.pm - remote programmatic administration of a NetApp Filer.
#
# DESCRIPTION
#   This Perl module is meant to include only the vol commands
#
# METHODS
#
#************************************************************************

## @author dl-central-libs
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDENG_MISC}) {
	my $newModule = $ENV{OVERRIDENG_MISC};
	delete $ENV{OVERRIDENG_MISC};
	print "The main Node::NG::Misc module will be replaced with $newModule\n";
	require($newModule);
	$LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

use strict;
use TracePkg;
use Options;
use Server;
use Socket;
use Testinfo;

my $Trace = new TracePkg;

sub mk_file {
	my $self = shift;
	my %params = @_;
	my $output;
	my $size = $params{size};
	my $filename = $params{file};
	my $count = $params{count};
	my $step_size = $params{step_size};
	my $timeout = $params{timeout};

	$Trace->entry(msg=>"size=$size file=$filename count=$count step_size=$step_size timeout=$timeout");

	my $cmd = "priv set diag; mkfile $size $filename $count $step_size; priv set";
	$self->dblade(cmd=>$cmd, timeout=>$timeout, output=>\$output);
	if (($output =~ /file\s+creation\s+failed/) ||
		($output =~ /write\s+failed/)) {
		$Trace->fatal(msg=>"file creation failed: $output",result=>"FATAL");
	}
	$Trace->exit(msg=>"exiting with 1");
	return 1;
}

sub rm_file {
	my $self = shift;
	my %params = @_;
	my $output;
	my $filename = $params{file};

	$Trace->entry(msg=>"file=$filename");

	$self->dblade(cmd=>"priv set diag; rm $filename; priv set", 
		output=>\$output);
	if ($output =~ /No\s+such\s+file\s+or\s+directory/) {
		$Trace->fatal(msg=>"file creation failed: output=$output",result=>"FATAL");
	}
	$Trace->exit(msg=>"exiting with 1");
	return 1;
}

sub rd_file {
	my $self = shift;
	my %params = @_;
	my $output;
	my $oldfilename = $params{file};
	my $filename;

	$Trace->entry(msg=>"file=$oldfilename");
	$filename = $self->path_to_UUIDpath(file=>$oldfilename);
	$self->dblade(cmd=>"priv set -q diag; rdfile $filename; priv set",
		output=>\$output);
	if ($output =~ /No\s+such\s+file\s+or\s+directory/) {
		$Trace->fatal(msg=>"cannot rdfile: output=$output",result=>"FATAL");
	}
	$Trace->exit(msg=>"exiting with output=$output");
	return $output;
}

sub path_to_UUIDpath {
	my $self = shift;
	my %params = @_;
	my $item;
	my $vol;
	my $rest;
	my $output;
	my $filename = $params{file};

	$Trace->entry(msg=>"file=$filename");

	$self->dblade(cmd=>"priv set -q diag; df; priv set",
		output=>\$output);

	if ( $filename =~ m!/vol/(\w+)/(.+)! ) {
		$vol = $1;
		$rest = $2;
		$Trace->debug(msg=>"vol=$vol rest=$rest");
	}
	elsif ( $filename =~ m!/vol/(\w+)! ) {
		$vol = $1;
		$rest = "";
		$Trace->debug(msg=>"vol=$vol rest=$rest");
	}
	else {
		$Trace->fatal(msg=>"$filename is not of the appropriate format: /vol/volname/rest",result=>"SCRIPT");
	}
	my @df = split /\n/, $output;
	my $UUIDvol = "";
	for $item ( @df ) {
		if ( $item =~ m!^/vol/(UUID_.+_$vol)/\s! ) {
			$UUIDvol = $1;
			last;
		}
	}
	my $returnstring = "/vol/$UUIDvol/$rest";
	$Trace->exit(msg=>"exiting with returnstring=$returnstring");
	return $returnstring;

}

sub vol_to_UUIDvol {
	my $self = shift;
	my %params = @_;
	my $item;
	my $output;
	my $vol = $params{vol};

	$Trace->entry(msg=>"vol=$vol");

	$self->dblade(cmd=>"priv set -q diag; df; priv set",
		output=>\$output);

	my @df = split /\n/, $output;
	my $UUIDvol = "";
	for $item ( @df ) {
		if ( $item =~ m!^/vol/(UUID_.+_$vol)/\s! ) {
			$UUIDvol = $1;
			last;
		}
	}
	$Trace->exit(msg=>"exiting with UUIDvol=$UUIDvol");
	return $UUIDvol;

}

sub sync {
	my $self = shift;
	my %params = @_;
	my $output;
	my $timeout = $params{timeout};

	$Trace->entry(msg=>"timeout=$timeout");

	$self->dblade(cmd=>"priv set test; sync; sync; priv set",
		timeout=>$timeout, output=>\$output);
	$Trace->exit(msg=>"exiting with 1");
	return 1;
}

# Install a script needed later to snap some diagnostic info from the node
sub gsi_install {
	$Trace->entry(msg=>"Installing the gsi.sh script for later gathering of diag info");
	my $self = shift;
	my %params = (
		dest_server	=> undef,
		dest_user	=> "root",
		timeout     => 60,
		server_ref   => undef,
		scriptspec   => $Tharn::Scriptspec,
		@_);
	my ($cmd, $value, $out, $output, $exit_code);
	my $rc = 0;
	my @install_files = qw(gsi.sh);

	# What's the target of the off filer commands?
	if ( !$params{dest_server} ) {
		$params{dest_server} = "10.98.200.111";
		$Trace->debug(msg=>"Userid root doesn't work on the ftp host, changing to spincores.");
		$params{dest_user} = "spincores";	
	} else {
		$params{dest_server} = gethostbyname($params{dest_server});
		$params{dest_server} = inet_ntoa( $params{dest_server} );
	}

	# Find the directory that the gsi.thpl lived in (to get at the gsi.sh)
	if ( !$params{scriptspec} ) {
		$Trace->fatal("Can't install gsi.sh, scriptspec not defined", result=>"SCRIPT");
	}
	if ($params{scriptspec} !~ /\/gsi\.thpl$/) {
		$Trace->comment(msg => "Script is : $params{scriptspec}");
		$Trace->fatal("Can't install gsi.sh, $params{scriptspec} does not match expected pattern: $value", result=>"SCRIPT");
	}
	$Trace->debug( msg=>"scriptspec is: $params{scriptspec}");
	$params{scriptspec} =~ /(.*)\/gsi.thpl$/;
	my $gsi_loc = $1;
	$Trace->debug( msg=>"gsi.sh will be copied from: $gsi_loc");

	# Now install the gsi.sh files into /var
	foreach my $install_file ( @install_files ) {

		# This only works for files that aren't http or ftp paths which we found out during the
		# golden client sync testing.
		if ($gsi_loc !~ m!^(http|ftp)://!){
			$cmd = "rcp $params{dest_user}\@$params{dest_server}:$gsi_loc\/$install_file /var/$install_file";
			$out = $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, 
				output=>\$output, exit_code => \$exit_code);
			$Trace->comment(msg => "Exit code for '$cmd' is $exit_code");
			if ($out || !$self->file_exists_on_server(path => "/var/$install_file")){
				$Trace->warn(msg=>"Error trying to install file: $install_file.\n$output\n\n$out");
				goto done;
			}
		} else {
			# This code sucks, but I can't think of a better way to track down the right path here.
			# The goal is to find out the path of $params{scriptspec} relative to the NATE_LIB path
			# in which it was found.  We need to do this since we don't have access to the relative
			# path which was in theory used to call this script.  NOTE that this is broken for
			# http:/ftp: absolute paths that aren't part of NATE_LIB.  This code is really only being
			# added for the goldenSync process so that's probably OK for now.

			# We need to check each of the paths in NATE_LIB to see if it matches $params{scriptspec}
			# If NATE::Inc::find_file returned more information....probably a burt
			foreach my $lib_path (split(/;/,$self->{options_object}->get(option=>"NATE_LIB"))){
				my $check_loc = $gsi_loc;

				# Move on to the next NATE_LIB path if we didn't match.  In addition, substitute out
				# the current lib path so that we wind up with a relative path.
				next if ($check_loc !~ s/$lib_path\///);

				# We matched so retrieve the contents of the file.
				$Trace->debug(msg => "$gsi_loc Location = $check_loc, $lib_path");
				my $filedata = NATE::Inc::get_file("$check_loc/$install_file");

				# Move on to the next path if there wasn't any data in the file
				next if (!length($filedata));

				# Got some data in the file so let's send it over to the node

				# Define our HERE document tag
				my $EOF_MARKER = "GSI_EOF_MARKER";

				# Start up the command.  Don't expect a prompt and don't swallow which would cause
				# this to take at least a second.
				$cmd = "cat > /var/$install_file << $EOF_MARKER";
				$out = $params{server_ref}->shell(cmd => $cmd, noprompt => 1, "-Swallow" => 0);

				# Now send each of the lines across to the cat command
				foreach my $line (split(/\n/, $filedata)){
					if (!length($line)){
						# Server->shell doesn't like "empty" commands, so we'll make it a space
						# just so that it doesn't complain
						$line = " ";
					}
					# We also need to escapt $s and `s.  quotemeta may be in order here instead.
					$line =~ s/\$/\\\$/g;
					$line =~ s/\`/\\\`/g;

					# Send the line over to the node.  Remember that it's in a HERE document
					# so there still won't be a prompt.  Swallowing is also still bad.
					$out = $params{server_ref}->shell(cmd => $line, noprompt => 1, "-Swallow" => 0);
				}
				# We're done sending the script, so send the final HERE document tag.  We do expect
				# a prompt and swallowing is OK.
				$out = $params{server_ref}->shell(cmd => $EOF_MARKER, "-Timeout" => $params{timeout}, 
					output => \$output, exit_code => \$exit_code);
				# Don't know the mode bits on the file since it's an http:/ftp: path so we'll just assume
				# that it's an executable.
				$out = $params{server_ref}->shell(cmd => "chmod +x /var/$install_file", "-Timeout" => $params{timeout}, 
					exit_code => \$exit_code);
				# We sent this file so we don't need to keep looking at NATE_LIB entries for this file
				# so bail on that loop.
				last;
			}
		}
	}

	if ($params{server_ref}->file_exists_on_server(path => "/mroot/gsi_data")){
		# Clean up the old gsi_data file
		$cmd = "rm -fr /mroot/gsi.log";
		if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
			$Trace->warn(msg=>"Error trying to clean up old gsi_data file.\n$output");
			goto done;
		}
	}

	$rc = 1;

	done:
	$Trace->exit(msg=>"Script install complete");
	return $rc;
}

sub file_exists_on_server {
	my $self = shift;
	my %params = (@_);
	my $retval = 0;
	if (length($params{path})){
		my ($cmd, $output, $exit_code);
		# Look to see if $params{path} exists, and use if it does
		$cmd = "ls -l $params{path}";
		$self->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
			exit_code=>\$exit_code);
		# yes, this check could be a little tighter.
		if ( $exit_code ) {
			# Does not exit
			$retval = 0;
		} else {
			# Exists
			$retval = 1;
		}
		$Trace->comment(msg => "$params{path} exists = $retval");
	}
	$retval;
}

sub is_mroot_mounted {
	my $self = shift;
	my %params = (@_);
	my ($cmd, $output);
	my $retval = 0;

	# Check to see if /mroot is even mounted
	$cmd = "/sbin/mount";
	if ( $self->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
		if ( $output !~ /\/mroot \(spin\)/g ) {
			$retval = 0;
		} else {
			$retval = 1;
		}
	}
	$retval;
}


# Take a snap shot of some diagnostic info 
sub gsi_snap {
	$Trace->entry(msg=>"Gathering diag info");

	my $self = shift;
	my %params = (
		timeout     => 60,
		server_ref  => undef,
		gsi_args    => "-m",
		bg          => 1,
		hide_output => 0,
		@_);
	my ($cmd, $value, $out, $output);
	my $rc = 1;

	# Note that running these processes in the background *can* be an adventure.
	# They may not finish by the time that the next snap starts or by the time
	# that the gather tries to pull everything off the box.  The running gsi
	# may also interfere with tests that may be running on the server.  We're taking
	# the reckless approach here on the basis that some information is better than
	# no information.

	my $nohup_out_dir = "/var";
	if ($self->is_mroot_mounted()){
		$nohup_out_dir = "/mroot";
	}
	if ($params{use_product_gsi} && $self->file_exists_on_server(path => "/usr/bin/gsi.sh")){
		# Take a snapshot of some diag information
		$cmd = "cd $nohup_out_dir;";
		$cmd .= "nohup " if $params{bg};
		$cmd .= "/usr/bin/gsi.sh $params{gsi_args} </dev/null";
		$cmd .= " >/dev/null 2>&1" if $params{hide_output};
		$cmd .= " &" if $params{bg};
		if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
			$Trace->warn(msg=>"Error trying to snap diag info.\n$output");
		}
		goto done;
	}
	if ($self->file_exists_on_server(path => "/var/gsi.sh")){
		# Take a snapshot of some diag information
		$cmd = "cd $nohup_out_dir;";
		$cmd .= "nohup " if $params{bg};
		$cmd .= "/var/gsi.sh </dev/null";
		$cmd .= " >/dev/null 2>&1" if $params{hide_output};
		$cmd .= " &" if $params{bg};
		if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
			$Trace->warn(msg=>"Error trying to snap diag info.\n$output");
		}
		goto done;
	}

	$rc = 0;

	done:
	$Trace->exit(msg=>"Info gathered");
	$rc;
}

# Establish a shell connection first.  This is lame.  If there already exists a connection
# this is painless.  If not, we may need a shell to a tricky system, or one to a 7m or 10m
# system.  As far as I can tell, I can't tell, and must revert to try one, if that doesn't
# try another.
sub establish_shell {
    $Trace->entry();

    my $self = shift;

    # Try to establish a cluster session
    eval {
	$Trace->comment(msg=>"About to open older type connection");
	$Trace->comment(msg=>"HERE CONNECTION $self->{RR_FLAG}");
	$self->establish_connections(shell_connection => 1);
    };
    if ($@) {
	$Trace->comment(msg=>"Current connections" . Dumper( $self->{connections}));
#	$self->close_connection( type => "shell_connection", keep_state => 1 );
#	$self->close_connection( type => "shell_connection");
	delete $Tharn::Connections{"$self->{server}#shell"};
	$Trace->comment(msg=>"About to open newer type connection");
	#
	# NEED TO CHANGE THIS.  TRY NEW WAY FIRST, AND GET RR_FLAG OUT OF THE PARAM SPACE
	#
	$self->{RR_FLAG} = 1;
	$self->establish_connections(shell_connection => 1);
    }

    $Trace->exit();
}


# Find the sudo executable on a filer (doesn't exist in tricky)
sub sudo_filer_find {
    $Trace->entry();

    my $self = shift;
    my %params = (
		  server_ref => undef,
		  @_);

    $self->establish_shell();

    my $sudo_cmd_filer;
    my $output = $params{server_ref}->shell(cmd=>"which sudo");
    $Trace->comment(msg=>"SUDO FILER @_,$params{server_ref}, $output");
    if ( $output =~ /Command not found/ ) {
	$sudo_cmd_filer = undef;
    } else {
	$sudo_cmd_filer = $output;
    }
    $Trace->exit();
    return $sudo_cmd_filer;
}
    


# GATHER logs from the requested filer
sub gsi_gather {
	my $self = shift;
	my %params = (
		dest_server	=> undef,
		dest_user	=> "root",
		delete_on_xfer => undef,
		timeout     => 600,
		server_ref   => undef,
		expect_core  => undef,
		operation    => undef,
		archive_age => 14,
		tar_archive_dir => 0,
		tar_exclude_dirs => undef,
		@_);

	my $rc = undef;
	my $output = "";
	my $cmd;
	my $out;
	my $exit_code;
	my $mroot = 1;   # assume mroot is mounted
	my $forget_firewall = 0;


    $Trace->comment(msg=>"Begining the gsi_gather sub routine.");
    my $dest_dir=Testinfo::dir();
    if (! defined($dest_dir) ) {
        $Trace->fail(msg=>"unable to get a directory to write to.  Check the logs from Testinfo::dir() to see why.");
        goto done;  #not sure if the fail will kill us or if I have to goto? mg
    }
    $Trace->entry(msg=>"dir() from Testinfo gave us this directory '$dest_dir'.");


    
	# Get the sudo path (needed for br and beyond)
	my $sudo_filer = $self->sudo_filer_find( server_ref=>$params{server_ref} );

	$self->gsi_snap(%params, gsi_args => "", bg => 0);

	# What's the target of the off filer commands?
	if ( !$params{dest_server} ) {
		$params{dest_server} = "ftp.cnc.spin.eng.netapp.com";
		$Trace->debug(msg=>"Userid root doesn't work on the ftp host, changing to spincores.");
		$params{dest_user} = "spincores";	
	}

    #If you are modifying this list, know that it will only affect c-mode runs.  To 
    #add the same exclusion to 7-mode, you need to go to lib/Node.pm and look at gsi_nfs_gather
	my $excludes = "--exclude /mroot/testInfo.tar.gz " .
	"--exclude /mroot/swapfile " .
	"--exclude /mroot/etc/swapfile " .
	"--exclude /mroot/perl " .
	"--exclude /mroot/.snapshot " .
	"--exclude /mroot/etc/.snapshot " .
	"--exclude /mroot/etc/java " .
	"--exclude /mroot/etc/.onboot/java " .
	"--exclude /mroot/etc/zoneinfo " .
	"--exclude /mroot/etc/http " .
	"--exclude /mroot/etc/lang " .
	"--exclude /mroot/etc/templates " .
	"--exclude /mroot/etc/man " .
	"--exclude /mroot/etc/shelf_fw " .
	"--exclude /mroot/etc/disk_fw " .
    "--exclude /mroot/etc/crash/aggregates " .
	"--exclude /mroot/feedback " .
    "--exclude /mroot/etc/clihelp ".
    "--exclude /mroot/etc/acpp_fw ".
     "--exclude /mroot/etc/firmware " .
        "--exclude /mroot/etc/sp_fw " .
    "--exclude /mroot/etc/cluster_config/mdb ".
	""; 



	# Don't tar archive dir unless told to.
	if ( !$params{tar_archive_dir} ) {
		$excludes .= " --exclude /mroot/oldcores --exclude /mroot/oldkcores " 
	}

	# exclude anything else the caller doesn't want.
	if (defined($params{tar_exclude_dirs}) && scalar(@{$params{tar_exclude_dirs}})) {
		if (ref($params{tar_exclude_dirs}) ne 'ARRAY') {
			$Trace->fatal(msg=>"tar_exclude_dir was not an array reference.", result=>'SCRIPT');
		}

		$excludes .= " --exclude ".join(" --exclude ", @{$params{tar_exclude_dirs}})." ";
	}

	my $tar_files = "/mroot /var /tmp/*.core /cfcard/*.core";
	my $tar_ball = "testInfo.tar.gz";


    $Trace->comment(msg=>"Run establish_connections() on this server, and hopefully turn on ssh access.");
    $params{server_ref}->establish_connections(cli_connection => 1);
    $Trace->comment(msg=>"establish_connections() is done.");


    $cmd='sudo ipfstat -io';
	$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output, exit_code=>\$exit_code);


	# Turn off IP filters
    #Programming note:  I'm using a few direct cli calls here.  If ssh isn't up, they will fail.
    #                   The assumption is that once you get to gsi_gather(), the filer is up
    #                   and operational.  The gsi() subroutine in Server.pm does that for us.  It 
    #                   takes a filer in any state, and hopefully gets it back on its feet.  If true,
    #                   that will allow gsi_gather() to simply login to the filer via ssh (Server's 
    #                   cli() calls, and copy the files off.  If that wasn't the case, I'm going to
    #                   make a leap of faith to say that none of gsi_gather() would have worked anyhow.  
    #                   And thus, this change is ok.  I hope that is the case.  This does represent a 
    #                   new philosophy in gsi_gather which formerly contained some "get well" routines.
    #                   I hope this change doesn't impact direct users of gsi_gather() if there were
    #                   any, and if so, I hope this note explains why the change was made.  If you are
    #                   a long time direct user of gsi_gather, please try gsi() from Server.pm or
    #                   contact mgarrett with concerns.  Otherwise, I hope this change allows 
    #                   programming here to be easier and less frought with peril.  mgarrett - 9/7/2010
    #
	$Trace->comment(msg=>"Checking to see if the firewall is on or off.");
    my $original_firewall_state='unknown';
    my $max_firewall_off_attempts=5;
    my $firewall_off_counter=0;

    #The following loop is because of a suspected bug.  We (smoke team) think we have seen occasions where
    #we tell the firewall to turn off, but it doesn't.  This loop is meant to both prove that, and work
    #around it for the time being.  Not a permenant solution!  (9/16/2010)
    FW_OFF_LOOP: while (1) {
        # ask for the status of the firewall on this node only
        my ($answer)=$params{server_ref}->system_service_firewall_show(node=>$params{server_ref}->{server});

        #save the original state of the firewall so we can return it to that state later
        $original_firewall_state=$answer->{enabled} if $firewall_off_counter==0;

        #if the firewall is on, turn it off
        if ($answer->{enabled} eq 'true') {
            $Trace->comment(msg=>"Found that the firewall was turned on for this node.  We must turn it off to allow rsh/rcp to work.  It will be turned back on at the end of this script.");
            $original_firewall_state='true';
            $params{server_ref}->system_service_firewall_modify(node=>$params{server_ref}->{server}, enabled=>'false');
            $firewall_off_counter++;
            if  ($firewall_off_counter == $max_firewall_off_attempts) {
                $Trace->warn(msg=>"Reached the maximum number of attempts to turn off the firewall.  It's still on!");
                last FW_OFF_LOOP 
            } else {
                $Trace->comment(msg=>"Attempt $firewall_off_counter of $max_firewall_off_attempts to turn the firewall off");
                Tharn::sleep(2);
            }
        } else {
            $Trace->comment(msg=>"The firewall is off.");
            last FW_OFF_LOOP;
        }
    }

    #If the firewall is off, you should see this output from the ipfstat -io command
    #p6070-1% sudo ipfstat -io
    #empty list for ipfilter(out)
    #empty list for ipfilter(in)

	$params{server_ref}->shell(cmd=>'sudo ipfstat -io', "-Timeout"=>$params{timeout}, output=>\$output, exit_code=>\$exit_code);
    if ($output =~ /empty list for/) { #good, this means the firewall is really off at the lowest layer
        $Trace->comment(msg=>"Since I saw 'empty list for' in the ipfstat command, that tells me that the firewall is really off.  That's good.");
    } else {
        $Trace->warn(msg=>"Since I didn't see 'empty list for' in the ipfstat command, that tells me that the firewall didn't turn off when we told it to.  That's bad. (probably burt464359)  Taking evasive action and turning the firewall off the workaround way.");
       
        #Notes from Mike Beattie on howto workaround this problem:
        #smoke3-vsim-pbc4b7c% sudo ipf -IFa
        #smoke3-vsim-pbc4b7c% sudo ipf -If /etc/ipf.user.rules
        #smoke3-vsim-pbc4b7c% sudo ipf -y
        #smoke3-vsim-pbc4b7c% sudo ipf -s
        #
        #In this case, I pushed ipf.user.rules just because it is an empty file. You could also create a new emtpy file and install that.
        $params{server_ref}->shell(cmd=>"sudo ipf -IFa", "-Timeout"=>$params{timeout});
        $params{server_ref}->shell(cmd=>"sudo ls -l /etc/ipf.user.rules", "-Timeout"=>$params{timeout});
        $params{server_ref}->shell(cmd=>"sudo ipf -If /etc/ipf.user.rules", "-Timeout"=>$params{timeout});
        $params{server_ref}->shell(cmd=>"sudo ipf -y", "-Timeout"=>$params{timeout});
        $params{server_ref}->shell(cmd=>"sudo ipf -s", "-Timeout"=>$params{timeout});

        #now that we tried the workaround, show the ipfstat output to see if it worked.
        $params{server_ref}->shell(cmd=>'sudo ipfstat -io', "-Timeout"=>$params{timeout}, output=>\$output, exit_code=>\$exit_code);
        if ($output =~ /empty list for/) { #good, this means the firewall is really off at the lowest layer
            $Trace->comment(msg=>"Great!  I now see 'empty list' in the ipfstat output.  That's good.");
        } else {
            $Trace->warn(msg=>"Uhoh!  Even after taking evasive actions, the firewall appears to still be on.  That's bad.");
        }
    }


    # Create the directory paths for:
    #   *testinfo/$node
    #   *testinfo/$node/vmware (if a vsim)
    #
    Testinfo::make_open_dir("$dest_dir/$self->{server}");
    Testinfo::make_open_dir("$dest_dir/$self->{server}/vmware") if ($params{server_ref}->{server}=~/vsim/);

	# Create the directory path
	$dest_dir .= "/$self->{server}";

	# Verify the target dir has been created
	$cmd = "$sudo_filer rsh -l $params{dest_user} $params{dest_server} 'ls $dest_dir;echo rc::\$?'";    
	$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
		exit_code=>\$exit_code);
	# Can only check for rsh errors here
	if ( $exit_code || $output !~ /rc::0/g ) {
		$Trace->warn(msg=>"Error verifying the created destination directory");
		goto done;
	}


    if (my $simdir=nodename2simdir($params{server_ref}->{server})) {
        $Trace->comment(msg=>"*"x20 . "Copying vmware logs and cores to the testinfo dir...");

        $Trace->comment(msg=>"simdir=$simdir");
        $simdir=~m|/vsims/(.+?)_vsim|;
        my $simuser=$1;
        $Trace->comment(msg=>"simuser=$simuser");

        my $vmware_dir="$dest_dir/vmware";
        my $sudo_cmd = param('SUDO_PATH', '-mesg','(NATE is expected to have set this)');
        my $command="$sudo_cmd -u $params{dest_user} cp $simdir/vmware*.log $vmware_dir";
        $Trace->comment(msg=>$command);
        system $command;

        for my $onefile (<$simdir/vmware*.log>) {
            $Trace->comment(msg=>"ls -l $onefile");
            system "ls -l $onefile";

            $command="$sudo_cmd -u $params{dest_user} chmod 666 $onefile";
            $Trace->comment(msg=>$command);
            system $command;
 

            $command="$sudo_cmd -u $params{dest_user} /bin/cat /dev/null > $onefile";
            $Trace->comment(msg=>$command);
            system $command;
        }

        $command="$sudo_cmd -u $params{dest_user} cp $simdir/*core* $vmware_dir";
        $Trace->comment(msg=>$command);
        system $command;

        for my $onefile (<$simdir/*core*>) {
            $command="$sudo_cmd -u $params{dest_user} rm -f $onefile";
            $Trace->comment(msg=>$command);
            system $command;
        }

    }
    
    sub nodename2simdir {
        my $nodename=shift;
        if ($nodename=~/vsim/) {
            my ($user,undef,$id) = split /[_-]/,$nodename;
            return "/u/$user/vsims/${user}_vsim_$id";
        } else {
            return 0;
        }
    }

    my $locations= {
        mroot => {
            description => 'This is either 10.x (tricky/NG/GX) or old converged before Aug 2008)',
            kcores => '/mroot/kcores',
            cores => '/mroot/cores',
            tar_path=>"/mroot",
        },
        sim => {
            description => 'This is a vsim that is newer than Aug 2008',
            kcores => '/sim/kcores',
            cores => '/sim/cores',
            tar_path=>"/sim",
        },
        converged_hw => {
            description => 'This is a converged hardware filer that is newer than Aug 2008',
            kcores => '/mroot/etc/crash',
            cores => '/mroot/etc/mcores',
            tar_path=>"/mroot",
        },
    };

    $Trace->trace(msg=>"Attempting to determine which set of paths to use.  (where are kcores/cores located?)");
    my $locationkey="converged_hw";
    
    $params{server_ref}->shell(cmd=>"ls -l /mroot/etc/crash", "-Timeout"=>$params{timeout}, output=>\$output, exit_code=>\$exit_code);
    if ($output =~ /\/sim\/kcores/) {
       $locationkey = "sim";
    }
    #COREPATH: for my $key (keys %{ $locations }) {
    #    $params{server_ref}->shell(cmd=>"ls -ld $locations->{$key}{kcores}", "-Timeout"=>$params{timeout}, output=>\$output, exit_code=>\$exit_code);
    #    if ($output=~/root\s+wheel/) {
    #        $locationkey=$key;
    #        last COREPATH;
    #    }
    #}
    if (!defined($locationkey)) {
        $Trace->warn(msg=>"Failed to determine kcore and core paths for this node!");
    } else {
        $Trace->comment(msg=>"locationkey=$locationkey");
        $Trace->debug(msg=>"Using this set of paths to get at the cores...\n" . Dumper($locations->{$locationkey}));
    }


	if ($self->{host_object}->{MODEL} eq 'VSIM' ) {
        $tar_files .= " /sim";
        $excludes .= " --exclude /sim/dev --exclude /sim/testInfo.tar.gz --exclude /sim/bin --exclude /sim/swapfile ";
		$mroot=1;
	} else {
		$mroot = $self->is_mroot_mounted();
	}

	if ( $mroot ) {
		# Look to see if we're in the middle of grabbing a kernel core
        $cmd = "ls -l $locations->{$locationkey}{kcores}/core.saving.nz";    
		$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
			exit_code=>\$exit_code);

		# We'll wait for 13 or so minutes to have the kcore created, longer for a 6070
		$Trace->debug(msg=>"The host_object is: " . Dumper($self->{host_object})); 
		my $max_attempts = 10;
        my $cur_attempts = $max_attempts;
        my $temp_kcore_size = 0;
        
		if ( $output =~ /root  wheel\s+\d+ / ) {
			GET_KCORE: {
				while ( $cur_attempts-- ) {
					# We found a core file getting saved.  Sleep for a few to give the saving
					# process some time.
					sleep 60;
					$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, 
						output=>\$output, exit_code=>\$exit_code);
                    # -rw-------  1 root  wheel  285308474 Jun  9 10:28 /mroot/kcores/core.saving.nz
					if ( $output =~ /root  wheel\s+(\d+) / ) {
                        if ($temp_kcore_size < $1) {
                            $cur_attempts = $max_attempts;
                            $temp_kcore_size = $1;
                            $Trace->comment(msg=>"The core is not done saving, but it is still making progress.");
                        }
						$Trace->comment(msg=>"Temp kcore still exists, attempts left: $cur_attempts");
					} else {
						$Trace->debug(msg=>"Can't find the temp kcore, assume complete");
						last GET_KCORE;
					}		    
				}
				$Trace->warn(msg=>"Still found a core.saving.nz file, not waiting any longer.  " .
					"File info:\n $output");
			}
		}

		# Look to see if there are any application core files, then grab a backtrace
		my @core_names = ();
		$cmd = "ls -l $locations->{$locationkey}{cores}";    
		$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
			exit_code=>\$exit_code);
		foreach my $line (split /\n/, $output) {
			$Trace->debug(msg=>"Looking at: $line");
			if ( $line =~ /(\S+)\.core/ ) {
				push @core_names, ( "$1" );
				$Trace->debug(msg=>"Will process core file: $1.core");
			}
		}   
		foreach ( @core_names ) {
			my $gdbCmd;
			# Look to see if /usr/bin/ugdb exists, and use if it does
			$cmd = "ls -l /usr/bin/ugdb";    
			$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
				exit_code=>\$exit_code);
			# yes, this check could be a little tighter.
			if ( $exit_code ) {
				$gdbCmd = "/usr/bin/gdb";
			} else {
				$gdbCmd = "/usr/bin/ugdb";
			}
			
			if ( /(\S+)\.\d+/ ) {
				$cmd = "echo 'bt' > /$locations->{$locationkey}{cores}/gdb.cmds ; " .
				"$gdbCmd /sbin/$1 /$locations->{$locationkey}{cores}/cores/$_.core < /$locations->{$locationkey}{cores}/gdb.cmds > " . 
				"/$locations->{$locationkey}{cores}/$_.backtrace.log";
			}
			 
			if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
					exit_code=>\$exit_code)) {
				$Trace->warn(msg=>"Error attempting to get backtrace on: $_\n$output");
			}
		}

		# Dump the boottimes
		$cmd = "$sudo_filer /sbin/sysctl sysvar.boottimes > /mroot/etc/mlog/boottimes.log";    
		$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
			exit_code=>\$exit_code);
		if ( $exit_code ) {
			$Trace->warn(msg=>"Error trying to create boottimes file\n$output");
		}


        #Ontap changes moved this directory from /mroot/log to /mroot/etc/log
        #This chunk of code finds which is active.
        my $mroot_logdir;
        # Look to see if /usr/bin/ugdb exists, and use if it does
        $cmd = "ls -ld /mroot/log";
        $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
            exit_code=>\$exit_code);
        if ( $exit_code ) { #if the ls failed
            $mroot_logdir= "/mroot/etc/log";
        } else {
            $mroot_logdir= "/mroot/log";
        }



		# Clean up the old tarball if any
		$cmd = "$sudo_filer rm -fr $locations->{$locationkey}{tar_path}\/$tar_ball";    
		if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
			$Trace->warn(msg=>"Error trying to clean up old testInfo tarball.\n$output");
			goto done;
		}

		# We were told to look for a kcore file, make sure we have one, warn otherwise.
		if ( $params{expect_core} ) {
			$cmd = "ls -l /$locations->{$locationkey}{kcores}/core.*";    
			$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output,
				exit_code=>\$exit_code);
			if ( $exit_code ) {
				$Trace->warn(msg=>"Expected to see a kcore, but didn't, continuing.\n$output");
			}
		}

		# Tar up /mroot, /var/log/messages
		$cmd = "set nonomatch ; $sudo_filer tar $excludes -vzcf $locations->{$locationkey}{tar_path}\/$tar_ball $tar_files";
		$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, 
			output=>\$output, exit_code=>\$exit_code);
		if ( $exit_code ) {
			$Trace->warn(msg=>"Error trying to create tarball, rc: $exit_code\n$output");
		}

		# Drive on, even if we had a tar error to try to get as much as we can off of the box.
		# I'm not sure that we'll always have local space to write the tarball, so perhaps
		# we should be trying to pipe the tar through the rsh command.
		foreach my $coretype (qw(cores kcores)) {
			# Create Archive Directory if it doesn't exist
		        $cmd = "$sudo_filer mkdir -p $locations->{$locationkey}{$coretype}.old/";
			$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout},
				output=>\$output,  exit_code=>\$exit_code);
			if ( $exit_code ) {
				$Trace->warn(msg=>"Error while creating directory: $locations->{$locationkey}{$coretype}.old/");
				next;
			}

			$cmd = "ls $locations->{$locationkey}{$coretype}.old/;";
			$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout},
				output=>\$output,  exit_code=>\$exit_code);
			if ( $exit_code ) {
				$Trace->warn(msg=>"$locations->{$locationkey}{$coretype}.old/ was not found. Skipping Archive" .
					"\n$output\n");
				next;
			}

			# delete any files in archive older than archive_age
			$cmd = "find $locations->{$locationkey}{$coretype}.old -type f -mtime +$params{archive_age} " .
			"-execdir $sudo_filer rm '{}' ';'";
			$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout},
				output=>\$output,  exit_code=>\$exit_code);
			if ( $exit_code ) {
				$Trace->warn(msg=>"Error while attempting to find/remove old files from $locations->{$locationkey}{$coretype}.old/" .
					"\n$output\n");
			}

			# Move files to archive directory 
			$cmd = "find $locations->{$locationkey}{$coretype}/ -type f -execdir $sudo_filer mv '{}' $locations->{$locationkey}{$coretype}.old/ ';'";
			$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout},
				output=>\$output,  exit_code=>\$exit_code);
			if ( $exit_code ) {
				$Trace->warn(msg=>"Error while moving files from $locations->{$locationkey}{$coretype}.$coretype/ to " . 
					"$locations->{$locationkey}{$coretype}.old/  \n$output\n");
			}

			if ( $coretype eq "cores" ) {
				# Move backtrace files to archive directory
				$cmd = "find $locations->{$locationkey}{$coretype}/ -type f -maxdepth 1 -name '*.backtrace.log'" .
				" -execdir $sudo_filer mv '{}' $locations->{$locationkey}{$coretype}.old/ ';'";

				$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout},
					output=>\$output,  exit_code=>\$exit_code);
				if ( $exit_code ) {
					$Trace->warn(msg=>"Error while attempting to find/move backtrace log files" .
						"\n$output\n\n$out");
				}
			}
		}
		# Transfer it down
		$cmd = "$sudo_filer rcp $locations->{$locationkey}{tar_path}\/$tar_ball $params{dest_user}\@$params{dest_server}:$dest_dir ";
		if ( $out = $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, 
				output=>\$output)) {
			$Trace->warn(msg=>"Error trying to transfer the file.\n$output\n\n$out");
			goto done;
		}
		# Untar  /mroot, /var/log/messages, ignore errors
		$cmd = "$sudo_filer rsh -l $params{dest_user} $params{dest_server} " . 
		"'cd $dest_dir ; tar -zxof $dest_dir\/$tar_ball ; echo rc::\$?; sudo chmod -R 777 .'";
		$params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, 
			output=>\$output, exit_code=>\$exit_code);
		if ( $exit_code || $output !~ /rc::0/g ) {
			$Trace->comment(msg=>"Error untaring testInfo tarball, continuing.");
		}
	} else {
		$Trace->warn(msg=>"/mroot isn't mounted.  Just grabbing /var/log/* /cfcard/*.core and /tmp/*.core.");
		# mroot not mounted, just grab some files from /var/log /cfcard and /tmp
		for my $onedir ('/var/log/*', '/cfcard/*.core', '/tmp/*.core') {
			$cmd = "$sudo_filer rcp $onedir $params{dest_user}\@$params{dest_server}:$dest_dir ";
			if ( $out = $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, 
					output=>\$output)) {
				$Trace->warn(msg=>"Error trying to transfer the file.\n$output\n\n$out");
				goto done;
			}
		}
	}

	if ( $params{delete_on_xfer} && $mroot ) {
		$Trace->debug(msg=>"Clean up the tarballs after transfer");
		$cmd = "$sudo_filer rm -fr $locations->{$locationkey}{tar_path}\/$tar_ball";    
		if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
			$Trace->warn(msg=>"Error trying clean up testInfo tarball, on server\n$output");
		}
		$cmd = "$sudo_filer rsh -l $params{dest_user} $params{dest_server} " . 
		"rm -fr $dest_dir\/$tar_ball";
		if ( $params{server_ref}->shell(cmd=>$cmd, "-Timeout"=>$params{timeout}, output=>\$output)) {
			$Trace->warn(msg=>"Error trying clean up testInfo tarball, in filesystem\n$output");
		}
	}


	# if we haven't blown up by now, everything went well
	$rc = 1;

	done:

    if ($original_firewall_state ne 'false') {
		# Turn on IP filters
		$Trace->comment(msg=>"All done with rsh.  Can turn the firewall back on. (it was on when we started gsi_gather).");
        $params{server_ref}->system_service_firewall_modify(node=>$self->{server}, enabled=>"$original_firewall_state ");
	}

	$Trace->exit(msg=>"Exited with rc: $rc");
	return $rc;

}

#***********************************************************************

=head1 C<FTPFile()>

=over 4

=item Description

FTP a file(s) that is on the NG-Filer(s) to a remote location.

=item Arguments

=over 4

=item Source
(Required)

Full path to source file

=item Destination
(Required)

Full path to destination file 

=item Servers
(Optional)

The NG-Filer(s) where the source file is located

Default: $self->{Filers}

=item DestServer
(Optional)

FTP server to place file on

Default: "ftp.cnc.spin.eng.netapp.com"

=item DestUser
(Optional)

User name for FTP server

Default: "spincores"

=item DestPass
(Optional)

Password for FTP server

Default: "spincores"

=item Verify
(Optional)

Perform checksum on Source and Destination to ensure correct transfer

Supported Values: "1" or "0"

Default: "0", meaning don't verify 

=item DeleteOnVerify
(Optional)

Remove Source if checksum passes

Supported Values: "1" or "0"

Default: "0", meaning don't delete the source file

=item Algorithm
(Optional)

Type of checksum to perform

Default: "md5"

=item Timeout
(Optional)

How long to wait, in seconds, for file transfer to complete

Default: "1000"

=back

=item Returns

1 on success

undef on failure

=back

=cut

#***********************************************************************
sub FTPFile {
	my $self = shift;
	my %params = (
		Servers	=> $self->{Servers},
		Source	=> undef,
		Destination	=> undef,
		DestServer	=> "ftp.cnc.spin.eng.netapp.com",
		DestUser	=> "spincores",
		DestPass	=> "spincores",
		Verify      => undef,
		DeleteOnVerify => undef,
		Algorithm   => "md5",
		Timeout     => 1000,
		@_);

	my $rc = undef;
	my $output = "";
	my @servers;
	my %serversum;
	my %localsum;

	$Trace->entry(msg=>"entered Filer:$self->{Servers} server:$params{DestServer} algorithm:$params{Algorithm}");

	# make sure the server(s) passed in make sense
	if ( $params{Servers} ) {
		if (ref($params{Servers}) ne "ARRAY") {
			$Trace->fatal(result=>'FATAL', msg=>"was passed a Server parameter that was not an array reference");	    
		} else {
			@servers = @{$params{Servers}};
		}
	} else {
		$Trace->fatal(result=>'FATAL', msg=>"was passed an empty Server list");	    
	}

	# validate the rest of the options
	if ( !defined $params{Source} || $params{Source} eq "" ) {
		$Trace->fatal(result=>'FATAL', msg=>"requires a Source.");
	} elsif ( !defined $params{Destination} || $params{Destination} eq "" ) {
		$Trace->fatal(result=>'FATAL', msg=>"requires a Destination.");
	} 

	# copy the files from all the servers we have
	foreach my $server ( @servers ) {

		if ($params{Verify}) {
			$serversum{$server} = $self->_fileSum( Server => $server, File => $params{Source}, Location => "remote", Method => $params{Algorithm} );
			$Trace->trace(msg=>"Completed server $params{Algorithm} of $params{Source} on $server ($serversum{$server})");
		}

		if ( !$self->$server->shell(cmd=>"ftp -u $params{DestUser}.$params{DestPass}fpt:\/\/$params{DestServer}.$params{Destination} $params{Source}", timeout=>$params{Timeout}, output=>\$output)) {
			$Trace->fatal(msg=>"encountered an error trying to FTP the file.\n\"$output\"");
			goto done;
		}
		# there is little point in parsing the output.  ftp will always return a bad error code
		# when there is a problem so the above check always catches failure.

		if ($params{Verify}) {
			my $base = `basename $params{Source}`;
			chomp($base);

			$Trace->comment(msg=>"Started local $params{Algorithm} of $base from $server.");

			$localsum{$server} = $self->_fileSum( File => "$params{Destination}/$base", Location => "local", Method => $params{Algorithm} );

			if ($localsum{$server} ne $serversum{$server}) {
				$Trace->warn(msg=>"checksum mismatch! ($server\: $serversum{$server} Local: $localsum{$server})");
			} else {
				if ($params{DeleteOnVerify}) {
					$Trace->comment(msg=>"Removing $params{Source} on $server since it was not corrupted in transfer.");
					$self->$server->shell(cmd=>"rm -rf $params{Source}", timeout=>300);
				} else {
					$Trace->comment(msg=>"$params{Source} on $server was not corrupted in transfer ($server\: $serversum{$server} Local: $localsum{$server}).");
				}
			}

			$Trace->comment(msg=>"Completed local $params{Algorithm} of $params{Source}.");
		}
	}

	# if we haven't blown up by now, everything went well
	$rc = 1;

	done:
	$Trace->exit(msg=>"exited with rc \"$rc\"");
	return $rc;

} # FTPFile

#***********************************************************************

=head1 C<_fileSum()>

=over 4

=item Description

Does the work for computing checksums

=item Arguments

=over 4

=item Server
(Required)

The NG-Filer where the file is located

Default: $self->{server} 

=item File
(Required)

The file to perform checksum on

=item Method
(Optional)

Type of checksum to run

Default: "md5"

=item Location
(Optional)

Is file on local server or remote

Supported Values: "local" or "remote"

Default: "remote"

=item Timeout
(Optional)

Time (in seconds) to wait for checksum to complete

Default: "1200" seconds

=item Caller
(Optional)

The name of this method

Default: "_fileSum"

=back

=item Returns

A string representing the checksum for the requested file

undef on failure

=back

=cut

#***********************************************************************
sub _fileSum
{
	my $self = shift();

	my %param = (
		Server => $self->{server},
		File => undef(),
		Method => "md5",
		Location => "remote",
		Timeout => 1200,
		Caller => "_fileSum",
		@_
	);

	my $localMethod = $param{Method};

	$Trace->enter(msg=>"entered.");

	my $out; 

	if (($param{Method} eq "md5") || ($param{Method} eq "sum")) {
		if ($param{Location} eq "remote") {
			$Trace->trace(msg=>"executing remote $param{Method} checksum on $param{File} (on $param{Server}).");
			$self->shell(cmd=>"$param{Method} $param{File}", output => \$out, timeout => $param{Timeout});
		} elsif ($param{Location} eq "local") {
			if ($localMethod eq "md5") { $localMethod = $localMethod."sum" }
			$Trace->trace(msg=>"Spinfs::$param{Caller} executing local $localMethod checksum on $param{File}.");
			my $out = `$localMethod $param{File}`;
		}

		if ($out =~ m/(\S+)\s+\S+/) {
			$Trace->trace(msg=>"$param{Method} of $param{File} is $1.");
			$Trace->exit(msg=>"Exited.\n");
			return $1;
		} else {
			$Trace->exit(msg=>"Exited (a failure occurred).");
			return undef;
		}

	} else {
		$Trace->exit(msg=>"Exited (an unknown method was specified).\n");
		return undef;
	}

	$Trace->exit(msg=>"exited.\n");
}

1;
