# $Id$
## @summary remote programmatic administration of Storage functions
## @author dl-central-libs
## @pod here

=head1 NAME

Storage.pm - remote programmatic administration of Storage functions

=head1 DESCRIPTION

C<Storage> provides routines to handle all aspects of Storage
such as takeover, sendhome, and querying Storage status.

=head1 SYNOPSIS

    All methods are AUTOLOADED by the Server package, so no 
    additional "use" line is necessary.

=head1 ROUTINES

=cut

# The following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one).
my $LibraryLoaded;
if ($ENV{OVERRIDENG_STORAGE}) {
    my $newModule = $ENV{OVERRIDENG_STORAGE};
    delete $ENV{OVERRIDENG_STORAGE};
    print "The main Node::NG::Storage module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

use TracePkg;
use Tharn;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(CODEREF ARRAYREF SCALAR UNDEF OBJECT);
use Benchmark;

my $Trace = new TracePkg;

our %StorageConversion = (
                 );
my %Options = (
          );

my %CommandDefinitions = (
                 );

my %Compatability = (
            );

sub SetupStorageCommands {
    my ($self, @servers) = @_;
    $Trace->entry();

    $self->_define_commands(Definitions => \%CommandDefinitions,
                            Compatability => \%Compatability,
                            Servers => \@servers);
    $Trace->exit(msg=>"exiting with 1");
    return 1;
}

#***********************************************************************

=head2 wait_aggregates

    wait_aggregates(takeover_node           => qdeux-7,
                surrender_node              => qdeux-6,
                original_aggr_config        => $Orig_config,
                original_aggr_member_config => $Orig_member_config,
                mode                        => "Takeover",
                release                     => $RELEASE,
                max_aggr_show_commands  => 20,
                allow_RPC_errors        => 1,
                );

C<wait_aggregates> is used to wait for all specified aggregates to come online on the 
specified node

=over

=item takeover_node

(Required) The node on which all aggregates must be sent to

=item surrender_node

(Required) The node that gave up its aggregates. We are only going to wait
on aggregates that actually failed over

=item original_aggr_config

(Required) Pointer to array of hashes detailing the original aggregate 
configuration 

=item original_aggr_member_config

(Optional) Pointer to array of hashes detailing the original member aggregate 
configuration. Only used when a striped aggregate is seen.

=item original_volume_config

(Optional) Pointer to array of hashes detailing the original volume 
configuration. 

=item mode 

(Optional) The type of mode we are in. Valid values are "Takeover" and "Sendhome"

=item max_aggr_show_commands 

(Optional) The maximum number of 'storage aggregate show' commands to issue while waiting for aggregates
to come online on the correct destination node. Default is 100.

=item allow_RPC_errors 

(Optional) If set, allows RPC errors of 'Program not registered' to be skipped. This is used if we are 
performing show commands on a node that has just come back up from a reboot. Default is undef.

=back

=cut

sub wait_aggregates {

    my $output = "";
    my $out_put = "";
    my $current_config = "";
    my $dblade_state;
    my $ng_state;
    my $Command = "";
    my $Hash = "";
    my $destination_node = "";
    my $destination_node = "";
    my $exp_state = "";
    my @is_root;

    $Trace->entry();
    ########################################################################
    # NOTE: original_aggr_member_config must allow SCALAR type to allow
    # for the case where there are no striped aggregates
    ########################################################################
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
            takeover_node               => { type => SCALAR },
            surrender_node              => { type => SCALAR },
            mode                        => { type => SCALAR, default => "Takeover" },
            original_aggr_config        => { type => ARRAYREF },
            original_aggr_member_config => { type => (ARRAYREF | SCALAR | UNDEF), optional => 1 },
            original_volume_config      => { type => ARRAYREF | UNDEF, optional => 1, default => [] },
            max_aggr_show_commands      => { type => SCALAR, default => 30 },
            retry_sleep                 => { type => SCALAR, default => 10 },
            allow_RPC_errors            => { type => SCALAR, default => undef() },
            release                     => { type => SCALAR, default => "Tricky" },
            dblade_only                 => { type => SCALAR, default => 0 },
        }
    );
    my $server = $self->{server_object};
    #$Trace->comment(msg => "\%params Dumper:\n".Dumper(%params));
    
    ########################################################################
    # Ensure aggregate array is populated
    ########################################################################
    if (scalar(@{$params{original_aggr_config}}) == 0) {
    $Trace->fatal(msg    => "The array of aggregates passed in was empty",
                      result => "FATAL",
                     );
    }

    ########################################################################
    # Loop waiting for aggregates to all be owned by takeover_node.
    # Give up after five minutes
    #
    # BURT 186280 saw a problem where "storage aggregate show" after a
    # takeover was missing an aggregate, so be sure to iterate through
    # all of the aggregates of the Original config.
    #
    # ALso be sure to make sure that aggregate is online, also part of burt
    # 186280
    #
    # Be sure we're only checking aggregates previously owned by
    # surrender_node
    ########################################################################

    my $timeout = $params{max_aggr_show_commands};
    ########################################################################
    # Catch burt 196440. Must eval the storage aggregate show command to
    # catch the error message and retry if needed
    ########################################################################
    my $burt196440_sightings = 10;
    my $burt209066_sightings = 10;
    my $burt214269_sightings = 5;
    my $burt200767_retries = 5;
    checkowntake: for (my $i=0; $i<=(scalar(@{$params{original_aggr_config}}) - 1); $i++) {
        ########################################################################
        # The filers/nodes param of 'storage aggregate show' is an array
        ########################################################################
        ########################################################################
        # If we are in Early_sendhome/giveback mode, then we expect all aggregates homed
        # on takeover node to come online. Aggregates on the surrender node are
        # in the "unknown" state until the surrender node is brought back up
        ########################################################################
        next if (!grep($_ eq $params{surrender_node}, split(/,/,$params{original_aggr_config}[$i]{filers})));
        next if ($params{original_aggr_config}[$i]{state}  eq "unknown");
        next if ($params{original_aggr_config}[$i]{state}  eq "restricted");
        if ($timeout and $burt196440_sightings and $burt209066_sightings) {
            if (!$params{dblade_only}) {
                eval {
                    $current_config = $server->storage_aggregate_show(aggregate => $params{original_aggr_config}[$i]{aggregate},
                                                                    -Timeout  => 300, 
                                                                    -SkipINFO => 1, 
                                                                    output    => \$output,
                                                                  );
                };
                $Trace->debug(msg => "***'Dumper of: \$current_config:\n".Dumper(\$current_config));
            } elsif ($params{dblade_only}) {
                if ($params{release} eq "Tricky") {
                   $server->shell(cmd => "dbladecli -cF 'priv set diag'");
                   $dblade_state = $server->shell(cmd => "dbladecli -cF 'aggr status $params{original_aggr_config}[$i]{aggregate}'",
                                             output => \$output);
                   $Trace->debug(msg => "Dbladecli aggr status:". Dumper($dblade_state));
                } ### if ($params{release} eq "Tricky")
                else {   ### Boilermaker
                   my $node=$params{takeover_node};
                   my $cmd="ngsh -c 'system node run -node $node -command aggr status'";
                   $dblade_state = $server->shell(cmd => "$cmd", \$output); 
                   $Trace->comment(msg => "$cmd:". Dumper($dblade_state));
                }
            } ###
            if ($@ =~ /(ERROR|Error): Failed to get the option information of aggregate/) {
                $Trace->warn(msg => "Burt 196440 seen on the storage aggregate show command.".
                    " Sleeping before retrying ...",
                );
                $burt196440_sightings--;
                sleep 10;
                redo checkowntake;
            } elsif ($@ =~ /Failed to get the information/) {
                $Trace->warn(msg => "Burt 214269 seen on the storage aggregate show command.".
                    " Sleeping before retrying ...",
                );
                $burt214269_sightings--;
                sleep 10;
                redo checkowntake;
            } elsif ($params{allow_RPC_errors} && (($output =~ /RPC: Program not registered/) ||
                                                   ($output =~ /RPC: Timed out/))) {
                $Trace->warn(msg => "RPC errors seen, but allowed. Sleeping before retrying ...");
                sleep 10;
                redo checkowntake;
            ########################################################################
            # Catch burt209066: Due to VLDB slowness, VLDB may report that aggregate
            # is on the wrong dblade, and thus 'storage aggregate show' will
            # report no matches to query. Fix will be in tricky.1, so for now, just
            # retry the command.
            ########################################################################
            } elsif ($@ && $output =~ /There are no entries matching your query/) {
                $Trace->warn(msg => "Burt 209066 seen on the storage aggregate show command.".
                    " Sleeping before retrying ...",
                );
                $burt209066_sightings--;
                sleep 10;
                redo checkowntake;
            } elsif ($@) {
                $Trace->fatal(msg    => "an unexpected error occured on the storage aggregate show command - $@\n$output\n",
                              result => "FATAL",
                             );
            } else {
                ########################################################################
                # Coral aggregates have the voltype param set to "striped". If this 
                # aggregate is a striped aggregate, call wait_member_aggregates
                # to wait for all member aggregates of this aggregate to  
                # failover
                ########################################################################
                if ($params{original_aggr_config}[$i]{voltype} eq "striped") {
                    $server->wait_member_aggregates(takeover_node                 => $params{takeover_node},
                                                  surrender_node                => $params{surrender_node},
                                                  aggregate                     => $params{original_aggr_config}[$i]{aggregate},
                                                  original_aggr_member_config   => $params{original_aggr_member_config},
                                                  mode                          => $params{mode},
                                                  max_aggr_member_show_commands => $params{max_aggr_show_commands},
                                                  release                     =>  $params{release},
                                                  dblade_only                   => $params{dblade_only},
                                                 );
                }

                #
                # "No_Takeover" mode is the case of a no failover, reboot...
                # with inhibit-takeover=true
                #
                if ($params{mode} eq "Takeover") {
                    $destination_node = $params{takeover_node};
                    $exp_state = "online";
                } elsif ($params{mode} eq "No_Takeover") {
                    $destination_node = $params{surrender_node};
                    $exp_state = "online";
                } elsif ($params{mode} eq "sendhome") {
                    $destination_node = $params{surrender_node};
                    $exp_state = "online";
                } elsif ($params{mode} eq "Sendhome") {
                    $destination_node = $params{surrender_node};
                    $exp_state = "online";
                } elsif ($params{mode} eq "giveback") {
                    $destination_node = $params{surrender_node};
                    $exp_state = "online";
                } elsif ($params{mode} eq "Early_giveback") {
                    $destination_node = $params{takeover_node};
                    $exp_state = "unknown";
                } elsif ($params{mode} eq "Early_sendhome") {
                    $destination_node = $params{takeover_node};
                    $exp_state = "unknown";
                } else {
                    $Trace->fatal(msg => "Invalid value '$params{mode}' passed in for the mode parameter",
                                  result => "FATAL",
                                 );
                }
                
                my $success = 0;
                my $state;
                my $filer;
                my $answer;
                my $ref = \@is_root;   ### pointer points to @is_root;
                $Trace->comment(msg => "Release = $params{release}");
                
                if ($params{release} eq "BR") {
                    ## BURT-317974  check for backend ha in takeover mode
                    ## History added by Renuka Puttappa on 10/10/2008
                    my $backend_ha = $server->cluster_ha_show(-Timeout=>300, -SkipINFO=> 1);
                    $Trace->comment(msg=>"Checking for Backend HA in Takeover mode\n");
                    $Trace->comment(msg=>"Dumper of cluster_ha_show:\n".Dumper($backend_ha));
                }
                if (!$params{dblade_only}) {
                    $Trace->comment(msg => "=== destination: '$destination_node' - source: '$$current_config[0]{filers}'===");
                    $Trace->comment(msg => "Mode: $params{mode} - Aggregate: $params{original_aggr_config}[$i]{aggregate}");
                    if (grep($_ eq $destination_node, split(/,/,$$current_config[0]{filers}))) {
                       if ($params{mode} ne "Early_sendhome" && $params{mode} ne "Early_giveback") {
                          ## Below is a work around for burt286122 for BR-MS6
                          ## Also note that hash does not work well w/the
                          ## command $command, it returns a string w/multiple
                          ## spaces between fields, \s+ =>multiple spaces.
                          ## All code begin from if ($params{mode} eq "Takeover") { to } must be
                          ## removed once burt286122 is fixed.
#                         if ($params{mode} eq "Takeover" && $params{release} eq "BR") {
#                            my $aa = $params{original_aggr_config}[$i]{aggregate}; 
#                            my $bb = $$current_config[0]{state}; 
#                            $Trace->comment(msg => "'$params{release}' - aa=$aa - bb=$bb");
#                            $server->set_execution_server(-Server => $destination_node);
#                            $answer = is_root_aggr($self,$aa,$destination_node,$ref);
#                            $Trace->comment(msg => "is_root='$is_root[0]' - '$is_root[1]'");
#                            if ($answer eq "y") {
#                               $Trace->comment(msg => "$aa is the root-aggregate");
#                               $Trace->comment(msg => "info => $is_root[0] - $is_root[1]");
#                               $$current_config[0]{state} = $is_root[1];
#                               if ($$current_config[0]{state} eq "online") {
#                                  $Trace->comment(msg => "$aa is now online");
#                                  $success = 1;
#                               }
#                            }  ### if ($is_root eq "y")
#                            elsif ($answer eq "n" && $bb eq "online") {
#                                $Trace->comment(msg => "The non-root aggregate $aa is '$bb'");
#                                $success = 1;
#                            }
#                            if ($$current_config[0]{state} eq $params{original_aggr_config}[$i]{state}) {
#                               $success = 1;
#                            }
#                         }  ### if ($params{mode} eq "Takeover") && $params{release} eq "BR" 

#                         if ($params{mode} eq "Takeover") {
                             if ($$current_config[0]{state} eq $params{original_aggr_config}[$i]{state}) {
                                 $success = 1;
                             }
                          ### the state of original_aggr_config might have '-' instead of
                          ### 'online' as it expected- we should proceed as long as
                          ### the current_config_state is 'online', that's what we care
                          ### A warning message is sent to warn the user.
                          ### burt232577 and 240949.
#                         else {
                             my $orig_aggr_config_state = $params{original_aggr_config}[$i]{state};
                             my $orig_aggr_config_aggr = $params{original_aggr_config}[$i]{aggregate};
                             if ($orig_aggr_config_aggr ne "online") {
                                ## Modified by Renuka 03/02/2010 Burt:403040 
                                $Trace->comment(msg=>"The original state '$orig_aggr_config_aggr' was '$orig_aggr_config_state' Sleeping before retrying.."); 
                                sleep 10;
                             }
                             if ($$current_config[0]{state} eq "online") {
                                $success = 1;
                             }
#                        } ##
#                       } ## 
                       }
                       else {
                           if ($$current_config[0]{state} eq "unknown") {
                               $success = 1;
                           }                           
                       }
                    }

                    $state = $$current_config[0]{state};
                    $filer = $$current_config[0]{filers};
                } elsif ($params{dblade_only}) {
                    if ( (grep(/$params{original_aggr_config}[$i]{aggregate} $params{original_aggr_config}[$i]{state}/, $dblade_state)) or
                       ($params{original_aggr_config}[$i]{voltype} eq "striped") ) {
                        $state = "NA";
                        $filer = "NA";
                        $success = 1;
                        if ($params{original_aggr_config}[$i]{voltype} eq "striped") {
                            $Trace->comment(msg => "Since we are only checking the dblade status of \n".
                                                   "aggrs and this is a logical striped aggr \n".
                                                   "we can only inspect the member aggrs not \n".
                                                   "the abstract coral aggr.");
                        }
                    }
                }
                if ($success) {
                    $Trace->comment(msg => "Aggregate $params{original_aggr_config}[$i]{aggregate} is now on '$destination_node' with a state of '$state'\n".
                       "Issued ".(($params{max_aggr_show_commands}-$timeout)+1)." 'storage aggregate show commands'",
                   );
                $timeout = $params{max_aggr_show_commands};    ### reset counter
                } else {
                    $Trace->comment(msg => "Need to go again on $params{original_aggr_config}[$i]{aggregate}.\n".
                                           "It is currently owned by '$filer' with a state of '$state'.\n".
                                           "It should be owned by node '$destination_node' with expected status '$exp_state'.\n".
                                           "Issued ".(($params{max_aggr_show_commands}-$timeout)+1)." 'storage aggregate show commands'",
                                   );
                    $timeout--;
                    sleep 10;
                    redo checkowntake;
                }
            }
        }
      }
    if (!$timeout) {
       my $tmp_msg;
       if ($params{mode} eq "Takeover") {
         $tmp_msg = "failover";
         $Trace->comment(msg=>"Timedout waiting for all aggregates to $tmp_msg to $destination_node");
       }
       else {
         $tmp_msg = "sendhome/giveback";
         $Trace->comment(msg=>"Timedout waiting for all aggregates to $tmp_msg to $destination_node");
       }
        my $aggr_output;
        if ($params{release} eq "Tricky") {
           $server->dblade(cmd => "priv set diag");
           $server->dblade(cmd    => "cf readfw",output => \$output,);
           $server->dblade(cmd    => "aggr status", output => \$aggr_output,);
        }
        else {  ### Boilermaker
#          my $Filer = $$current_config[0]{filers};
           my $Filer = $destination_node;
           my $cmd1="ngsh -c system node run -node $Filer -command 'priv set advanced; cf readfw'";
           my $cmd2="ngsh -c system node run -node $Filer -command 'aggr status'";
           $server->shell(cmd => "$cmd1", \$output);
           $server->shell(cmd => "$cmd2", \$aggr_output);
        }
        $Trace->warn(msg  => "Timed out waiting for all aggregates to $tmp_msg to $destination_node\n".
                 "'cf readfw' debug info \n$output\n".
                             "'aggr show' debug info \n$aggr_output",
                    );
    ########################################################################
    # Get list of sendhome events that have happened recently
    ########################################################################
    if ($params{release} eq "Tricky") {
       $server->shell(cmd      => "ems_logviewer /mroot/log/messages.ems | grep sendhome",
                     output   => \$output,-Timeout => 300,);
       $Trace->warn(msg  => "'sendhome EMS events' \n$output\n",);
    }
    else {  ### Boilermaker
       $server->shell(cmd      => "ems_logviewer /mroot/log/messages.ems | grep giveback",
                     output   => \$output,-Timeout => 300,);
       $Trace->warn(msg  => "'giveback EMS events' \n$output\n",);
    }


    if ($params{mode} eq "sendhome") {
            ########################################################################
        # BURT 200767: On a timeout, try issuing another manual sendhome
            ########################################################################
        $Trace->comment(msg => "Issuing subsequent manual sendhome...");
        my $Send_output = "";
        eval {
                $Hash = $server->storage_failover_send_home(fromfiler => $params{takeover_node},
                                  -Timeout  => 300,output    => \$Send_output,);
            };
            if ($@) {
                $Trace->warn(msg     => "BURT 200767 debug\n".
                    "The manual sendhome command failed\n$Send_output\n".
                                        "Obtaining additional debug info...",
                            );
        if ($params{release} eq "Tricky") {
           $server->dblade(cmd    => "priv set diag",);
           $server->dblade(cmd    => "cf readfw",output => \$output,);
        }
        else {   ### Boilermaker
           my $Filer = $$current_config[0]{filers};
           my $cmd1="ngsh -c system node run -node $Filer -command 'priv set advanced; cf readfw'";
           $server->shell(cmd => "$cmd1", \$output);
        }
        $Trace->warn(msg    => "BURT 200767: Additional debug info\n$output\n".
                           "'cf readfw' debug info \n$output\n",);
        ########################################################################
        # Get list of sendhome events that have happened recently
        ########################################################################
        if ($params{release} eq "Tricky") {
           $server->shell(cmd      => "ems_logviewer /mroot/log/messages.ems | grep sendhome",
                    output   => \$output, -Timeout => 300,);
           $Trace->warn(msg  => "BURT 200767 debug\n".
                             "'sendhome EMS events' \n$output\n",);
        }
        else {  ### Boilermaker
           $server->shell(cmd      => "ems_logviewer /mroot/log/messages.ems | grep giveback",
                    output   => \$output, -Timeout => 300,);
           $Trace->warn(msg  => "BURT 200767 debug\n".
                             "'giveback EMS events' \n$output\n",);
        }
        $Trace->fatal(msg    => "manual sendhome to workaround burt 200767 failed\n$Send_output\n",result => "FATAL",);
            } else {
        ########################################################################
        # Start the for loop all over again. Decrement the $burt200767_retries
        # param to avoid an infinite loop 
        ########################################################################
        if ($burt200767_retries) {
            $burt200767_retries--;
                $Trace->warn(msg     => "BURT 200767 debug: The subsequent manual sendhome command succeeded.\n".
                                "Waiting for aggrs to be sent-home...",
                    );
            $timeout = $params{max_aggr_show_commands};
            $burt196440_sightings = 10;
            goto checkowntake;
        } else {
            $Trace->fatal(msg    => "BURT 200767 debug: We retried the manual sendhome/wait routine five times. Aborting",
                                  result => "FATAL",
                                 );
        }       
            }
    } else {
            $Trace->fatal(msg    => "Timed out waiting for all aggregates to failover to $destination_node",
                          result => "FATAL",
                         );
    }
    } elsif (!$burt196440_sightings) {
        $Trace->fatal(msg    => "saw burt196440 ten times. Aborting",
                      result => "FATAL",
                     );
    } elsif (!$burt214269_sightings) {
        $Trace->fatal(msg    => "saw burt214269 five times. Aborting",
                      result => "FATAL",
                     );
    } elsif (!$burt209066_sightings) {
        $Trace->fatal(msg    => "saw burt209066 ten times. Aborting",
                      result => "FATAL",
                     );
    } elsif ($params{mode} eq "Early_sendhome") {
        $Trace->comment(msg => "***Since in early_sendhome all aggrs are unknown as expected!***");
    } elsif ($params{mode} eq "Early_giveback") {
        $Trace->comment(msg => "***Since in early_giveback all aggrs are unknown as expected!***");
    } else {
        $Trace->comment(msg => "***All aggregates are now on $destination_node!***");
    }

    ########################################################################
    # Ensure that all volumes are online on each aggregate involved in the
    # SFO failover. 
    # Multi-layer iteration: 
    # 1) First loop on all aggregates, weeding out all aggregates not originally 
    # on the sendhome node.
    # 2) Then do a vol show -aggregate <this aggr>" to obtain list of volumes on
    # this aggregate
    # 3) Loop through the volumes in the original_volume_config array to obtain
    # the orignal state of this volume
    # 4) Loop until the volume returns to this state
    ########################################################################
    aggr_loop: for (my $i=0; $i<=(scalar(@{$params{original_aggr_config}}) - 1); $i++) {
        ########################################################################
        # The filers/nodes param of 'storage aggregate show' is an array
        ########################################################################
        ########################################################################
        # If we are in Early_sendhome/giveback mode, then we expect all aggregates homed
        # on takeover node to come online. Aggregates on the surrender node are
        # in the "unknown" state until the surrender node is brought back up
        ########################################################################
        if ($params{mode} eq "Early_sendhome") {
            next if (grep($_ eq $params{surrender_node}, split(/,/,$params{original_aggr_config}[$i]{filers})));
        } elsif ($params{mode} eq "Early_giveback") {
            next if (grep($_ eq $params{surrender_node}, split(/,/,$params{original_aggr_config}[$i]{filers})));
        } else {
            next if (!grep($_ eq $params{surrender_node}, split(/,/,$params{original_aggr_config}[$i]{filers})));        
        }
        next if ($params{original_aggr_config}[$i]{state}  eq "unknown");
        next if ($params{original_aggr_config}[$i]{state}  eq "restricted");
        ########################################################################
        # Get list of volumes on this aggregate. Eval the command since
        # it's OK for an aggregate not to have volumes
        ########################################################################
        eval {
            $Hash = $server->volume_show(aggregate => $params{original_aggr_config}[$i]{aggregate},
                                       -Timeout  => 300,
                                       -SkipINFO => 1,
                                       output    => \$output,
                                      );
        };
        if ($@ && $output =~ /[This table is currently empty|There are no entries matching your query]/) {
            $Trace->comment(msg=> "we did not find any volumes on aggregate $params{original_aggr_config}[$i]{aggregate}");
        next aggr_loop;
        } elsif ($@) {
            $Trace->fatal(msg=>"Unexpected error on 'volume show' command - $@\n$output\n",
                          result=>"FATAL",
                         );
        } else {
            $Trace->comment(msg=>"we found volumes on aggregate $params{original_aggr_config}[$i]{aggregate}");
            my $timeout = 90;
            volonline: for (my $j=0; $j<=(scalar(@{$Hash}) - 1); $j++) {
                if ($timeout) {
            ########################################################################
            # Determine this volume's original state
            # This needs to be enhanced. We will be calling this loop on each retry.
            # Is there a way we can get the matching volume once and then use it
            # on each retry?
            ########################################################################
            vol_loop: for (my $k=0; $k<=(scalar(@{$params{original_volume_config}}) - 1); $k++) {
                if ($params{original_volume_config}[$k]{volume} eq $$Hash[$j]{volume}) {
                            if ($$Hash[$j]{state} eq $params{original_volume_config}[$k]{state}) {
                                $Trace->comment(msg=>"we discovered that volume $$Hash[$j]{volume} on $$Hash[$j]{vserver} returned to the ".
                                 "$params{original_volume_config}[$k]{state} after a $params{mode} ".
                             "of aggregate $params{original_aggr_config}[$i]{aggregate}",
                           );
                                next volonline;
                } elsif ($$Hash[$j]{state} eq "online") {
                ########################################################################
                # We need to catch cases where we did not originally see the volume in
                # the online state, but we now do. This is a warning, as the "online"
                # state is not what we expect, but it is valid
                ########################################################################
                $Trace->comment(msg => "Expected volume $$Hash[$j]{volume} on $$Hash[$j]{vserver} to return to the ".
                               "$params{original_volume_config}[$k]{state}, but instead it is online"
                           );
                        } else {
                        $Trace->comment(msg=>"Need to go again on volume $$Hash[$j]{volume} since it is in state $$Hash[$j]{state}\n".
                                     "Issued ".(30-$timeout)." 'volume show' commands",
                                   );
                ########################################################################
                # We know that the aggregate has volumes, so we don't need to eval 
                # this subsequent call. If it fails, let our parent eval catch the failure
                # jackm 043009 - added Timeout 300 to allow for large number of volumes
                ########################################################################
            		$Hash = $server->volume_show(aggregate => $params{original_aggr_config}[$i]{aggregate},
                                       -Timeout  => 300);
                                sleep 10;
                                $timeout--;
                                redo volonline;
                        }
                        }
                }
            }
        }
            if (!$timeout) {
                $Trace->fatal(msg=>"timedout waiting for all volumes to return to their original state on aggregate $params{original_aggr_config}[$i]{aggregate} after a $params{mode}",
                              result=>"FATAL",
                             );
            } else {
                $Trace->comment(msg=>"we verified that all volumes on aggregate $params{original_aggr_config}[$i]{aggregate} are in their original state after a $params{mode}\n",
                       );
            next aggr_loop;
            }
    }
    }
    $Trace->comment(msg=>"we verified that all volumes on all aggregates involved in the $params{mode} are in their original states");
    $Trace->exit();
}


#***********************************************************************

=head2 wait_member_aggregates

    wait_member_aggregates(takeover_node                 => qdeux-7,
                       surrender_node                => qdeux-6,
                       aggregate                     => "coral_aggregate",
                       original_aggr_member_config   => $original_aggr_member_config,
                       mode                          => "Takeover",
                       release                       => "Tricky|Boilermaker" },
               max_aggr_member_show_commands => 100,
                          );

C<wait_member_aggregates> is used to wait for all members of a Coral (a.k.a. striped)
aggregate to come online on the specified node

=over

=item takeover_node

(Required) The node on which all member aggregates must be sent to

=item surrender_node

(Required) The node that gave up its member aggregates. We are only going to wait
on member aggregates that actually failed over

=item aggregate

(Required) The coral aggregate whose members we will wait for. We error out of no 
member aggregates are found

=item original_aggr_member_config

(Required) Pointer to array of hashes detailing the original member aggregate
configuration.

=item mode 

(Optional) The type of mode we are in. Valid values are "Takeover" and "Sendhome"

=item max_aggr_member_show_commands

(Optional) The maximum number of 'storage aggregate member show' commands to issue while waiting for aggregates
to come online on the correct destination node. Default is 100.

=back

=cut

sub wait_member_aggregates {

    $Trace->entry();
    ########################################################################
    # NOTE: Unlike wait_aggregates(), original_aggr_member_config must 
    # ber a valid ARRAYREF here
    ########################################################################
    
    my ($self, %params) = validate_mixed(
                                         @_,
                                         [ { type => OBJECT, isa => 'Node::NG' } ],
                                         {
                                          takeover_node         => { type => SCALAR },
                                          surrender_node        => { type => SCALAR },
                                          mode              => { type => SCALAR, default => "Takeover" },
                                          aggregate             => { type => SCALAR },
                                          original_aggr_member_config   => { type => ARRAYREF },
                                          max_aggr_member_show_commands => { type => SCALAR, default => 100 },
                                          release               => { type => SCALAR, default => "Tricky" },
                                          dblade_only           => { type => SCALAR, default => 0 },
                                         }
                                        );
    
    $Trace->debug(msg => "Dumper of \%params:\n".Dumper(%params));
    
    my $output = "";
    my $current_config = "";
    my $Hash = "";
    my $destination_node = "";
    my $found = 0;
    my $dblade_state;
    my $exp_state = "";
    
    ########################################################################
    # Ensure aggregate array is populated
    ########################################################################
    if (scalar(@{$params{original_aggr_member_config}}) == 0) {
        $Trace->fatal(msg    => "The array of member aggregates passed in was empty",
                      result => "FATAL",
                     );
    }
    
    ########################################################################
    # Get hash of all of the member aggregates of the specified striped
    # aggregate. By default, aggregate member names start with the
    # striped aggregate name followed by "_NNN" where NNN is a number.
    # Search in the original member aggregate config that was passed to us.
    # Abort if there aren't any member aggregates
    ########################################################################
    for (my $i=0; $i<=(scalar(@{$params{original_aggr_member_config}}) - 1); $i++) {
        ($found = 1),last if ($params{original_aggr_member_config}[$i]{aggregate} =~ /^($params{aggregate})/);
    }
    if (!$found) {
        $Trace->fatal(msg    => "No member aggregates found for aggregate $params{aggregate}",
                      result => "FATAL",
                     );
    }
    
    ########################################################################
    # Loop waiting for member aggregates to all be owned by takeover_node.
    #
    # Use the original config hash so we know which members are expected
    # to failover.
    #
    # Also be sure to make sure that the member aggregate is online
    #
    # Be sure we're only checking member aggregates previously owned 
    # by surrender_node
    ########################################################################
    my $timeout = $params{max_aggr_member_show_commands};
  checkownmembertake: for (my $i=0; $i<=(scalar(@{$params{original_aggr_member_config}}) - 1); $i++) {
        ########################################################################
        # NOTE: member aggregate uses "filer" not "filers"
        ########################################################################
        ########################################################################
        # If we are in Early_sendhome/giveback mode, then we expect all aggregates homed
        # on takeover node to come online. Aggregates on the surrender node are
        # in the "unknown" state until the surrender node is brought back up
        ########################################################################
        next if (!grep($_ eq $params{surrender_node}, $params{original_aggr_member_config}[$i]{filer}));
        next if ($params{original_aggr_member_config}[$i]{state}  eq "unknown");
        next if ($params{original_aggr_member_config}[$i]{state}  eq "restricted");
        if ($timeout) {
            if (!$params{dblade_only}) {
                $current_config = $self->storage_aggregate_member_show(aggregate => $params{original_aggr_member_config}[$i]{aggregate},
                                                                       -Timeout  => 300, 
                                                                       -SkipINFO => 1, 
                                                                       output    => \$output,
                                                                      );
                
            } elsif ($params{dblade_only}) {
                if ($params{release} eq "Tricky") {
                   $self->shell(cmd => "dbladecli -cF 'priv set diag'");
                   $dblade_state = $self->shell(cmd => "dbladecli -cF 'aggr status $params{original_aggr_member_config}[$i]{aggregate}'",
                                             output => \$output);
                   $Trace->debug(msg => "Dbladecli aggr status:". Dumper($dblade_state));
                } ### if ($params{release} eq "Tricky")
                else {
                   my $node=$params{takeover_node};
                   my $cmd="ngsh -c 'system node run -node $node -command aggr status'";
                   $dblade_state = $self->shell(cmd => "$cmd", \$output);
                   $Trace->comment(msg => "$cmd:". Dumper($dblade_state));
                }
            }  ###
        ########################################################################
        # Determine correct destination node for this member aggregate based
        # on our mode
        ########################################################################
        if ($params{mode} eq "Takeover") {
            $destination_node = $params{takeover_node};
            $exp_state = "online";
        } elsif ($params{mode} eq "No_Takeover") {
            $destination_node = $params{surrender_node};
            $exp_state = "online";
        } elsif ($params{mode} eq "Sendhome") {
            $destination_node = $params{surrender_node};
            $exp_state = "online";
        } elsif ($params{mode} eq "sendhome") {
            $destination_node = $params{surrender_node};
            $exp_state = "online";
        } elsif ($params{mode} eq "giveback") {
            $destination_node = $params{surrender_node};
            $exp_state = "online";
        } elsif ($params{mode} eq "Early_giveback") {
            $destination_node = $params{takeover_node};
            $exp_state = "unknown";
        } elsif ($params{mode} eq "Early_sendhome") {
            $destination_node = $params{takeover_node};
            $exp_state = "unknown";
        } else {
            $Trace->fatal(msg => "Invalid value '$params{mode}' passed in for the mode parameter",
                              result => "FATAL",
                             );
            }
            
            my $success = 0;
            my $state;
            my $filer;
            if (!$params{dblade_only}) {
		$Trace->debug(msg=>"Dumping \$current_config:\n".Dumper(\$current_config));
                $state = $current_config->[0]{state};
                $filer = $current_config->[0]{filer};
                if ($destination_node eq $filer) {
                    if ($params{mode} ne "Early_sendhome" && $params{mode} ne "Early_giveback") {
                        if ($state eq $params{original_aggr_member_config}[$i]{state}) {
                            $success = 1;
                        }
                    } else {
                        if ($state eq "unknown") {
                            $success = 1;
                        }                           
                    }
                }
            } elsif ($params{dblade_only}) {
                if (grep(/$params{original_aggr_member_config}[$i]{aggregate} $params{original_aggr_member_config}[$i]{state}/, $dblade_state)) {
                    $state = "NA";
                    $filer = "NA";
                    $success = 1;
                }
            } else {
		$Trace->trace(msg=>"Still waiting..");
                $state = $current_config->[0]{state};
                $filer = $current_config->[0]{filer};
	    }
            if ($success) {
                $Trace->comment(msg => "Member aggregate $params{original_aggr_member_config}[$i]{aggregate} is now on $destination_node with a state of '$state'\n".
                                "Issued ".(($params{max_aggr_member_show_commands}-$timeout)+1)." 'storage aggregate member show commands'",
                               );
                $timeout = $params{max_aggr_member_show_commands};    ### reset counter.
            } else {
                $Trace->comment(msg => "Need to go again on $params{original_aggr_member_config}[$i]{aggregate}.\n".
                                       "It is currently owned by '$filer' with a state of '$state'.\n".
                                       "It should be owned by node '$destination_node' with expected status '$exp_state'.\n".
                       "Issued ".(($params{max_aggr_member_show_commands}-$timeout)+1)." 'storage aggregate member show commands'",
                   );
                sleep 10;
                $timeout--;
                redo checkownmembertake;
            }
        }
    }
    if (!$timeout) {
        $Trace->fatal(msg    => "Timed out waiting for all member aggregates of $params{aggregate} to failover to '$destination_node'",
                      result => "FATAL",
                     );
    } else {
    $Trace->comment(msg => "All member aggregates for $params{aggregate} are now on '$destination_node'\nIssued ".
                               (($params{max_aggr_member_show_commands}-$timeout)+1).
                               " 'storage aggregate member show' commands",
                       );
    }
    $Trace->exit();
}

=head2 choose_takeover_node

    my ($Takeover_filer,$Surrender_filer) = 
        choose_takeover_node(
                             filers = \@Filers,
                                );

C<choose_takeover_node> is used to select the node to use as the takeover node (i.e.
the one that stays up), and the node to use as the surrender node (i.e. the node
that goes down). By default, the first node listed in the FILER param is the 
takeover node, however, if the cluster size is only two nodes then the takeover node will
be assigned to the node WITHOUT epsilon to ensure that quorum is preserved.

Returns an array of nodes in the format ($takeover,$surrender)

=over

=item filers 

(Required) Reference to array of filers passed in on FILER param

=back

=cut

sub choose_takeover_node {

    $Trace->entry();

    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
        filers => { type => ARRAYREF },
            takeover_filer => { type => SCALAR, default => "first" },
        }
    );

    my $output = "";
    my $Hash = "";
    my $takeover_filer = "";
    my $surrender_filer = "";

    ########################################################################
    # If we have a two-node cluster, and one of the filers in FILER has 
    # epsilon, make him the takeover filer, otherwise, the first node
    # listed in FILER is the takeover_node
    ########################################################################
    my $Cluster = $self->cluster_show();
    if (scalar(@{$Cluster}) == 2) {
        for (my $i=0; $i<=(scalar(@{$Cluster}) - 1); $i++) {
            if (grep($_ eq $$Cluster[$i]{filer},@{$params{filers}})) {
                if ($$Cluster[$i]{epsilon} eq "true") {
                    if ($params{filers}[0] eq $$Cluster[$i]{filer}) {
                        $takeover_filer = $params{filers}[0];
                        $surrender_filer = $params{filers}[1];
            last;
                    } else {
                        $takeover_filer = $params{filers}[1];
                        $surrender_filer = $params{filers}[0];
            last;
                    }
        }
            }
        }
    }
    $Trace->comment(msg=>"DEBUG: takeover_filer is: '$params{takeover_filer}'");
    if ($params{takeover_filer} eq "first") {
        ($takeover_filer  = $params{filers}[0]) if ($takeover_filer  eq "");
        ($surrender_filer = $params{filers}[1]) if ($surrender_filer eq "");
    } elsif ($params{takeover_filer} eq "random") {
        my $random = int(rand(2));
        if ($random) {
            ($takeover_filer  = $params{filers}[0]) if ($takeover_filer  eq "");
            ($surrender_filer = $params{filers}[1]) if ($surrender_filer eq "");
        } else {
            ($surrender_filer  = $params{filers}[0]) if ($surrender_filer  eq "");
            ($takeover_filer = $params{filers}[1]) if ($takeover_filer eq "");
        }        
        $Trace->comment(msg=>"random is:'$random', takeover_filer is: '$takeover_filer', surrender_filer is: '$surrender_filer'");
    } elsif ($params{takeover_filer} eq "second") {
        ($surrender_filer  = $params{filers}[0]) if ($takeover_filer  eq "");
        ($takeover_filer = $params{filers}[1]) if ($surrender_filer eq "");        
    } else {
        $Trace->fatal(msg => "takeover_filer must be set to 'first', 'second', or 'random'",
                      result => "FATAL");        
    }

    ########################################################################
    # Warn if the surrender node has epsilon
    ########################################################################
    $Cluster = $self->cluster_show(node => $surrender_filer);
    if ($$Cluster[0]{epsilon} eq "true") {
    $Trace->warn(msg => "Surrender node $surrender_filer has epsilon, which may cause issues if this node is takenover while other".
                " cluster member are down"
            );
    }
    $Trace->exit(msg=>"takeover_filer is $takeover_filer, surrender_filer is $surrender_filer");

    return ($takeover_filer,$surrender_filer);
}

=head2 get_mroot_aggregates

    my %Mroots = get_mroot_aggregates(filers = \@Filers);

C<get_mroot_aggregates> is used to determine which aggregate holds the node's mroot volume.
Until burt 188189 is resolved, the only way to get this info is via dbladecli's aggr status
command.

Returns a hash of the form filer => aggregate for each node passed in
via the FILER parameter to ntest

=over

=item filers

(Required) Reference to array of filers passed in on FILER param

=back

=cut

sub get_mroot_aggregates {

    $Trace->entry();

    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
        filers => { type => ARRAYREF },
        release=> { type => SCALAR, default => "Tricky" },
        }
    );

    my $output = "";
    my %Hash = ();

    foreach my $node (@{$params{filers}}) {    
        ########################################################################    
    # Use "aggr status" command in dbladecli of each node to  get the 
    # aggregate with the root option set
    ########################################################################
    my $server = $self->{server_object};
        $server->set_execution_server(-Server => $node);
        if ($params{release} eq "Tricky") {
           $server->dblade(cmd    => "aggr status",output => \$output,);
        }
        else {  ### Boilermaker
           my $cmd="ngsh -c 'system node run -node $node -command aggr status'";
           $server ->shell(cmd => "$cmd", output => \$output);
        }

    ########################################################################
    # say_wrapper doesn't parse the output of a dbladecli command so let's
    # get busy with regex's
    #
    # Sample output from "aggr status" command:
    # dblade> aggr status
    # aggr status
        #        Aggr State      Status            Options
    #   aggr0_qdeux_6 online     raid4, aggr       root
        # p_ericb_qdeux_6_aggr online     raid_dp, aggr
        # dblade>
    # 
    # Quick way to do it is to look for the word "root" and grab the first 
    # argument on the row. This will break if the aggregate is named "root"
    # which is a retarded thing to do. We are sort of protected from this 
    # possibility by the auto-renaming of root aggregates on a cluster join,
    # but with the "storage aggregate rename" command being available, 
    # someone could rename their aggregate to whatever their heart desires.
    ########################################################################
    if ($output =~ /(\w+)\s+(\w+)\s+(\w+,\s+\w+)\s+root/) {
        my $aggregate = $1;
        my $state     = $2;
        my $status    = $3;
        $Trace->comment(msg => "The root aggregate for $node is $aggregate");
        $Hash{$node}=$aggregate;
    }
    if ($Hash{$node} eq "") {
        $Trace->fatal(msg    => "Could not find root aggregate for node $node",
                          result => "FATAL",
             );
    }
    }
    return %Hash;
}

=head2 core_dump_state

    my %Mroots = core_dump_state(nodes = \@Filers);

C<core_dump_state> is used to disabled or enable coredump on specifc nodes
passed in as an array ref.

=over

=item nodes

(Required) Reference to array of nodes to act on

=item state

(Required) disable to turn off coredump or enable to turn on coredump 

=back

=cut

sub core_dump_state {
    $Trace->entry();

    my $self = shift;
    my %params = (state => undef, # required
                  release => undef, # required
                  nodes => undef, # required
                  @_,
                  );

    my $output;
    my $output_hash;
    my $core_attempts = undef;

    foreach my $node (@{$params{nodes}}) {    
        if ($params{release} eq "Tricky") {
           $output_hash = $self->system_coredump_config_show(node => $node,
                        output => \$output,);
        }
        else {
           $output_hash = $self->system_node_coredump_config_show(node => $node,
                        output => \$output,);
        }
        if ($params{state} eq "disable") {
            if (!$$output_hash[0]->{"coredump-attempts"} == 0) {
                $core_attempts = "0";
            } else {
                $Trace->comment(msg=>"$node already has coredumps disabled.");
                next;
            }
        } elsif ($params{state} eq "enable") {
            if ($$output_hash[0]->{"coredump-attempts"} == 0) {
                $core_attempts = "2";
            } else {
                $Trace->comment(msg=>"$node already has coredumps enabled.");
                next;
            }
        } else {
            my $Message = "Current coredump state = $params{state}";
            $Trace->comment(msg => "output_hash= Dumper:\n".Dumper($output_hash));
            $Trace->fatal(msg=>"$Message. Coredump state must be set to 'disabled' or 'enabled'",
                          result=>"FATAL",
                        );
        }
        if ($params{release} eq "Tricky") {
           $self->system_coredump_config_modify(node => $node,
               "coredump-attempts" => $core_attempts,output => \$output,);
        }
        else {   ### Boilermaker
           $self->system_node_coredump_config_modify(node => $node,
               "coredump-attempts" => $core_attempts,output => \$output,);
        }
    }
    
    $Trace->exit();
}

=head2 discover_config

    my @conf = $Server->discover_config(node     => $Takeover_filer,
                                        discover => [qw(sfo aggr member_aggr vol vif)],
                                       );

C<discover_config> is used to obtain various configuration data structures.  This
method can be used for pre and post test comparisons.  It should help with understanding
that objects (aggrs, vols, vifs, etc.) do the correct thing in the test.

Returns an array of an arrayref of hash.

=over

=item node

(Required) The node to execute the commands on.

=back

=cut

sub discover_config {

    $Trace->entry();

    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
            node           => { type => SCALAR },
            retry_attempts => { type => SCALAR, default => 12 },
            retry_sleep    => { type => SCALAR, default => 10 },
            discover       => { type => ARRAYREF, default => [qw(sfo aggr member_aggr vol)] },
            vs             => { type => SCALAR, default => "vs0" },
        }
    );
    $Trace->debug(msg => "\%params debug Dumper:\n".Dumper(%params));

    my $test;
    my $output;
    my @conf;
    my %discovery_tasks = (
        sfo         => {
                command => "storage_failover_show",
                params  => {'-Timeout' => 300, '-SkipINFO' => 1, 'output' => $output},
                conf  => undef,
                },
        aggr        => {
                command => "storage_aggregate_show",
                params  => {'-Timeout' => 300, '-SkipINFO' => 1, 'output' => $output},
                conf  => undef,
                },
        member_aggr => {
                command => "storage_aggregate_member_show",
                params  => {'-Timeout' => 300, '-SkipINFO' => 1, 'output' => $output},
                conf  => undef,
                },
        vol         => {
                command => "volume_show",
                params  => {'vserver' => $params{vs}, '-Timeout' => 1000, '-SkipINFO' => 1, 'output' => $output},
                conf  => undef,
                },
        vif         => {
                command => "network_interface_show",
                params  => {'-Timeout' => 1000, '-SkipINFO' => 1, 'output' => $output},
                conf  => undef,
                },
    );
    $Trace->comment(msg=>"Going to discover configuration from these areas:\n @{$params{discover}}");
    my $server = $self->{server_object};
    $server->set_execution_server(-Server=>$params{node});
    foreach my $task (@{$params{discover}}) {
        $Trace->comment(msg=>"Attempting to discover $task config.");
        my $i = 1;
        my $retry_attempts = $params{retry_attempts};
        while ($retry_attempts) {
            $Trace->comment(msg=>"Attempt $i out of $retry_attempts to discover $task config.");
            my $server_cmd = $discovery_tasks{$task}{command};
            eval {
                $discovery_tasks{$task}{conf} = $self->$server_cmd(%{$discovery_tasks{$task}{params}});
            };
            $Trace->debug(msg=>"$task conf Dumper:\n".Dumper($discovery_tasks{$task}{conf}));
            if (($@) && ($@ =~ /This table is currently empty/) && ($task eq "member_aggr")) {
                $Trace->comment(msg=> "We did not find any striped aggregate members");
                last;
            } elsif (($@) && ($@ =~ /[This table is currently empty|There are no entries matching your query]/) && ($task eq "vol")) {
                $Trace->comment(msg=> "We did not find any volumes in the cluster");
                last;
            } elsif ($@) {
                $retry_attempts--;
                $Trace->warn(msg=>"An unexpected error occured on the '$discovery_tasks{$task}{command}' command.\n$@\n");
            } else {
                $Trace->comment(msg=>"Successfully executed the '$discovery_tasks{$task}{command}' command.");
                last;
            }
            ++$i;
            sleep $params{retry_sleep};
        }
        if (!$retry_attempts) {
            $Trace->fatal(msg=>"Tried to get a successful '$discovery_tasks{$task}{command}' $params{retry_attempts} times\n".
                               "every $params{retry_sleep} seconds.  All attempts failed.",
                          result=>"FATAL",
                         );
        } else {
            push(@conf,$discovery_tasks{$task}{conf});
        }
    }
    $Trace->debug(msg=>"Configuration to return Dumper:\n".Dumper(@conf));
    return (@conf);
    $Trace->exit();
} # end discover_config

#***********************************************************************

=head2 takeover

    takeover(takeover_node         => $takeover_node,
             surrender_node        => $surrender_node,
             takeover_reason       => "reboot",
             takeover_option       => "normal",
             autoboot              => 1,
             orig_sfo_conf         => $orig_sfo_conf,
             orig_aggr_conf        => $orig_aggr_conf,
             orig_aggr_member_conf => $orig_aggr_member_conf,
             orig_vol_conf         => $orig_vol_conf,
             retry_attempts        => 60,
             retry_sleep           => 10,
             wait_aggr_dblade_only => 0
             wait_aggr_mode        => "Takeover"
           );

C<takeover> is used to takeover the aggregates of the surrender node on
takeover node

=over

=item takeover_node

(Required) The node hostname on which all member aggregates must be sent to

=item surrender_node

(Required) The node hostname that gave up its member aggregates.

=item takeover_reason

(Optional) The event that causes the takeover [takeover, reboot, halt, power_failure, panic].
        The default is 'takeover'.
        takeover      = ngsh> storage failover takeover -bynode takeover_node [-option normal|immediate|allow-version|force]
        reboot        = ngsh> reboot -node surrender_node [-inhibit-takeover true|false]
        halt          = ngsh> halt -node surrender_node [-inhibit-takeover true|false]
        power_failure = rlm>  system power cycle
        panic         = bsd>  sysctl debug.panic=1

=item takeover_option

(Optional) Options to 'storage failover takeover' or reboot/halt  [normal, immediate, allow-version, force, inhibit].
        The default is 'normal'.
        normal        = ngsh> storage failover takeover -node takeover_node -option normal
        immediate     = ngsh> storage failover takeover -node takeover_node -option immediate
        allow-version = ngsh> storage failover takeover -node takeover_node -option allow-version
        force         = ngsh> storage failover takeover -node takeover_node -option force
        inhibit       = ngsh> reboot|halt -inhibit-takeover true
        *NOTE: Can only be 'inhibit' if takeover_reason is 'reboot' or 'halt' 

=item autoboot

(Optional) If the test will reboot the surrender node immediately (1) or stop (0) 
at the CFE/LOADER level [0 1].
        The default is '0'.

=item orig_sfo_conf
(Required) An array refference of the output from 'storage failover show'.
It is an array of hash.

=item orig_aggr_conf

(Required) An array refference of the output from 'storage aggr show'.
It is an array of hash.

=item orig_aggr_member_conf

(Required) An array refference of the output from 'storage aggr member show'.
It is an array of hash.

=item orig_vol_conf

(Required) An array refference of the output from 'volume show'.
It is an array of hash.

=item retry_attempts

(Optional) The number of times a 'storage aggr show' will be performed before the aggr is 
determined to have failed to have been taken over.  The retry_sleep param is used to 
determine how many seconds to sleep between each retry.
        The default is '60'.

=item retry_sleep

(Optional) How many seconds to sleep between each 'storage aggr show' retry.
        The default is '10'.

=item wait_aggr_mode

(Optional) The mode (Takeover|Sendhome) that we must test the wait_aggregates method.
        The default is 'Takeover'.

=item wait_aggr

(Optional) Should the state of the aggrs be confirmed after a takeover? (0=no|1=yes).
        The default is '1'.

=back

=cut

sub takeover {
    $Trace->entry();
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
            takeover_node         => { type => SCALAR },
            surrender_node        => { type => SCALAR },
            takeover_reason       => { type => SCALAR, default => "takeover" },
            takeover_option       => { type => SCALAR, default => "normal" },
            autoboot              => { type => SCALAR, default => 0 },
            orig_sfo_conf         => { type => ARRAYREF },
            orig_aggr_conf        => { type => ARRAYREF },
            orig_aggr_member_conf => { type => ARRAYREF | UNDEF, optional => 1 },
            orig_vol_conf         => { type => ARRAYREF | UNDEF, optional => 1, default => [] },
            retry_attempts        => { type => SCALAR, default => 60 },
            retry_sleep           => { type => SCALAR, default => 10 },
            wait_aggr_dblade_only => { type => SCALAR, default => 0 },
            wait_aggr_mode        => { type => SCALAR, default => "Takeover" },
            wait_aggr             => { type => SCALAR, default => 1 },
            release               => { type => SCALAR, default => "Tricky" },
        }
    );
    $Trace->debug(msg => "\%params debug Dumper:\n".Dumper(\%params));
    my $inhibit;
    my $server = $self->{server_object};
    my @Filers = $server->servers(); 
    foreach my $node (@Filers) {
        $params{$node} = host($node);   
    }

    $server->set_execution_server(-Server=>$params{takeover_node});
    my $takeover_start_time = Benchmark->new();
    $Trace->comment(msg=>"***TAKEOVER STARTED - REASON = $params{takeover_reason}***");
    
    if ($params{takeover_option} eq "inhibit") {
        $inhibit = "true";
    } else {
        $inhibit = "false";
    }
    
    if ($params{takeover_reason} eq "takeover") {
        $server->storage_failover_takeover(byfiler=>$params{takeover_node}, option=>$params{takeover_option});
        
    } elsif ($params{takeover_reason} eq "reboot") {
        $server->filer_reboot(filer=>$params{surrender_node}, 'inhibit-takeover'=>"$inhibit");
        
    } elsif ($params{takeover_reason} eq "halt") {
        $server->filer_halt(filer=>$params{surrender_node}, 'inhibit-takeover'=>"$inhibit");
        
    } elsif ($params{takeover_reason} eq "power_failure") {
        $server->set_execution_server(-Server=>$params{surrender_node});
        #eval and prime of the ssh key
        eval {
            $server->rlm(cmd=>"system power cycle", -Timeout=>60);
        };
        if ($@) {
            $Trace->comment(msg => "FAILED FIRST CONNECT TRYING TO PRIME THE SSH HOST KEY");
            system("ssh -o StrictHostKeyChecking=no -l naroot: $params{takeover_node}{RLM_IP} date");
            $server->rlm(cmd=>"system power cycle", -Timeout=>60);
        }
    } elsif ($params{takeover_reason} eq "panic") {
        $server->storage_failover_show(-options,); 
        if ($params{release} eq "BR" || $params{release} eq "RR") {
           $server->shell(cmd=>"sudo sysctl debug.panic=1", -Server=>$params{surrender_node}, noprompt=> 1, -Errors=>0);
        }
        else {      ### Tricky
           $server->shell(cmd=>"sysctl debug.panic=1", -Server=>$params{surrender_node}, noprompt=> 1, -Errors=>0);
        }

    } else {
        $Trace->fatal(msg=>"takeover_reason must be set to takeover, reboot, halt, power_failure, or panic.",
                      result=>"FATAL",
                     );
    }

    $server->set_execution_server(-Server => $params{surrender_node});
    eval {
        if (!$params{autoboot}) {
            $server->transit_to(state    => $server->CFE,
                              wait_for => $server->CFE,
                            );
            $Trace->comment(msg=>"Sucessfully made it down!\nNow let's go check on the aggrs.");
        }
    };

    if ($@) {
        if (($params{takeover_reason} eq "panic")&&($@ =~ /The filer appears to have panicked/)) {
            $Trace->comment(msg=>"Anticipated panic was seen");
            $server->transit_to(state => $server->CFE,
                              wait_for => $server->CFE,
                              wait_timeout => 900,
                            );
        } else {
            $Trace->fatal(msg=>"Transit_to failed, but not as expected\n$@\n",
                          result=>"FATAL",
                         );
        }
    }

    my $take_node = 0;
    foreach (@{$params{orig_sfo_conf}}) {
        unless ($params{orig_sfo_conf}->[$take_node]{node} eq $params{takeover_node}) {
            ++$take_node;
            next;
        }
        last;
    }

    $server->set_execution_server(-Server=>$params{takeover_node});

    ########################################################################
    # If we do not expect any aggrs to fail-over, verify they did not.
    # No need to check the state of volumes if the aggr they are in are offline.
    ########################################################################
    $Trace->comment(msg=> "***takeover reason: $params{takeover_reason}***");
    if ($params{takeover_reason} eq "panic") {
        $Trace->comment(msg=> "***onpanic: $params{orig_sfo_conf}[$take_node]{onpanic}***");
    } elsif (($params{takeover_reason} eq "reboot")||($params{takeover_reason} eq "halt")) {
        $Trace->comment(msg=> "***inhibit-takeover: $inhibit***");
    } elsif ($params{takeover_reason} eq "takeover") {
        $Trace->comment(msg=> "***takeover option: $params{takeover_option}***");
    }
    if ((($params{takeover_reason} eq "panic")&&($params{orig_sfo_conf}[$take_node]{onpanic} eq "false"))||
        ((($params{takeover_reason} eq "reboot")||($params{takeover_reason} eq "halt"))&&($params{takeover_option} eq "inhibit"))) {
        if ($params{autoboot}) {
            sleep 10;
            $Trace->comment(msg=> "***Sleeping 10s until collecting aggr configuration.***");
        }
        my @conf = $server->discover_config(node     => $params{takeover_node},
                                            discover => [qw(aggr member_aggr)],
                                           );
        $Trace->debug(msg=> "Dumper of \@conf:\n".Dumper(@conf));
        
        my @aggr_conf;
        my @member_conf;
        push(@aggr_conf,($params{orig_aggr_conf}, $conf[0]));
        push(@member_conf,($params{orig_aggr_member_conf}, $conf[1]));
        
        my $n = 0;
        my $home;
        foreach my $config (\@aggr_conf, \@member_conf) {
            my $orig_conf = $config->[0];
            my $take_conf = $config->[1];
            if ($n) {
                $Trace->comment(msg=> "***Comparing striped-vol aggrs!***");
                $home = "node";
            } else {
                $Trace->comment(msg=> "***Comparing flex-vol aggrs!***");
                $home = "nodes";
            }
            ++$n;
            for (my $j=0; $j<=(scalar(@{$orig_conf}) - 1); $j++) {
                next if (!grep($_ eq $params{surrender_node}, split(/,/,$$orig_conf[$j]{$home})));
                next if ($$orig_conf[$j]{state}  eq "unknown");    #only interested in aggrs that  
                next if ($$orig_conf[$j]{state}  eq "restricted"); #were owned by surrender node
                ##########################################################################################
                # If no takeover occured than we will get INFO messages on the aggregates
                # that are still owned by the surrender filer. This INFO message
                # causes the aggregate's row to be removed from the display, so we
                # need to iterate through all aggregates making sure that we are
                # looking at the same aggregate in both Orig_confgi and Takeover_config
                ##########################################################################################
                for (my $k=0; $k<=(scalar(@{$take_conf}) - 1); $k++) {
                    if ($$orig_conf[$j]{aggregate} eq $$take_conf[$k]{aggregate}) {
                        $Trace->debug(msg=> "***$$orig_conf[$j]{filers} : $$take_conf[$k]{filers} : $$take_conf[$k]{state}***");
                        if (($$take_conf[$k]{filers} eq $$orig_conf[$j]{filers})&&
                            (($$take_conf[$k]{state} eq '-')||
                             ($$take_conf[$k]{state} eq 'unknown')||
                             ($$take_conf[$k]{state} eq 'offline'))) {
                            $Trace->comment(msg=>"As expected aggr $$take_conf[$k]{aggregate} did not failover on $params{takeover_reason}.\n".
                                                 "The state is '$$take_conf[$k]{state}'."
                                           );
                        } elsif (($$take_conf[$k]{filers} eq $$orig_conf[$j]{filers})&&
                                 ($$take_conf[$k]{state} eq 'online')) {
                            $Trace->fatal(msg=>"Aggr $$take_conf[$k]{aggregate} did NOT fail over on $params{takeover_reason}.\n".
                                               "However, it is somehow still $$take_conf[$k]{state} even though \n".
                                               "$params{takeover_node} went offline becasue of a $params{takeover_reason}.",
                                          result=>"FATAL",
                                         );                                
                        } else {
                            $Trace->fatal(msg=>"Aggr $$take_conf[$k]{aggregate} DID fail over on $params{takeover_reason}.\n ".
                                               "This was unexpected!",
                                          result=>"FATAL",
                                         );
                        }
                        last;
                    }
                }
            }
        }       
    } else {
        if ($params{orig_sfo_conf}[$take_node]{"detection-time"} > 15) {
            ########################################################################
            # Verify that takeover doesn't start until after the designated
            # TAKEOVER_DETECT_TIME (minus 10 seconds)
            ########################################################################
	    $Trace->trace(msg=>"Verifying that takeover starts at proper time...");
            my $start = time();
            my $not_yet = 0;
            while (!$not_yet) {
                my @conf = $server->discover_config(node     => $params{takeover_node},
                                                    discover => [qw(sfo)],
                                                   );
                $Trace->comment(msg=> "Dumper of \@conf:\n".Dumper(@conf));
                my $take_sfo_conf = $conf[0];

                if ((time() - $start) < ($params{orig_sfo_conf}[$take_node]{'detection-time'} - 10)) {
                    if ($$take_sfo_conf[$take_node]{state} ne "connected") {
                        $not_yet = 1;
                        $Trace->fatal(msg=>"After reboot, takeover started before detection time ".
                                           "($params{orig_sfo_conf}[$take_node]{'detection-time'} seconds) expired",
                                      result=>"FATAL",
                                     );
                    } else {
                        $Trace->comment(msg=>"After reboot, takeover has not yet occured since waiting for takeover ".
                                             "detection time to expire. ". 
                                              (($params{orig_sfo_conf}[$take_node]{"detection-time"} - 10) - (time() - $start)).
                                             " seconds remain");
                    }
                } else {
                    $not_yet = 1;
                    $Trace->comment(msg=>"After reboot, no takeover occured before the detection time timeout of ".
                                          ($params{orig_sfo_conf}[$take_node]{"detection-time"} - 10)." seconds");
                }   
            }
        
        }
	    #$Trace->debug(msg => "\%params debug Dumper:\n".Dumper(\%params));
        if ($params{wait_aggr}) {
            $Trace->comment(msg=>"***Now let's make sure the aggrs made it to $params{takeover_node}.");
            ### commented out dblade_only, passing a value of "1" to wait_aggregates
            ### July 11 2008 - TPho
            $server->wait_aggregates(takeover_node               => $params{takeover_node},
                                     surrender_node              => $params{surrender_node},
                                     original_aggr_config        => $params{orig_aggr_conf},
                                     original_aggr_member_config => $params{orig_aggr_member_conf},
                                     original_volume_config      => $params{orig_vol_conf},
                                     mode                        => $params{wait_aggr_mode},
                                     max_aggr_show_commands      => $params{retry_attempts},
                                     retry_sleep                 => $params{retry_sleep},
#                                    dblade_only                 => $params{wait_aggr_dblade_only},
                                     release                     => $params{release},
                                    );
        } else {
            $Trace->comment(msg=>"***Confirmation of aggr state after takeover explicitly skipped...sleeping 50s.*** ");
            sleep 50;
        }
    }

    my $takeover_end_time = Benchmark->new();
    my $max_time = ($params{retry_attempts} * $params{retry_sleep});
    my $takeover_elapsed_time = timestr(timediff($takeover_end_time, $takeover_start_time)); 
    $takeover_elapsed_time =~ s/wallclock sec.*$/wallclock sec/;

    $Trace->comment(msg=>"***takeover reason: '$params{takeover_reason}'\n".
                         "***onpanic:         '$params{orig_sfo_conf}[$take_node]{onpanic}'\n".
                         "***inhibit:         '$inhibit'");
    $Trace->comment(msg=>"***retry_attempts: $params{retry_attempts}");
    $Trace->comment(msg=>"***retry_sleep: $params{retry_sleep} secs");
    $Trace->comment(msg=>"***Max time for wait_aggregates: $max_time secs");

    if ((($params{takeover_reason} eq "panic")&&($params{orig_sfo_conf}[$take_node]{onpanic} eq "false"))||
        ((($params{takeover_reason} eq "reboot")||($params{takeover_reason} eq "halt"))&&($inhibit))) {
        $Trace->comment(msg=>"***As expected after a $params{takeover_reason} of $params{surrender_node}, NO aggregates failed over.");
    } else {
        $Trace->comment(msg=>"***TAKEOVER COMPLETE. TAKEOVER TIME: $takeover_elapsed_time***");
        $Trace->comment(msg=>"***After a $params{takeover_reason} of $params{surrender_node}, all aggregates are now owned by $params{takeover_node}");
    }

    #NEED WORK SINCE autoboot IS A NEW CONCEPT FOR THE THPLS!!!!!!!!
    #if ($params{autoboot}) {
    #    eval {
    #        $self->transit_to(state => $server->LOGIN);
    #    };
    #    if ($@) {
    #        if (($params{takeover_reason} eq "panic")&&($@ =~ /The filer appears to have panicked/)) {
    #            $Trace->comment(msg=>"Anticipated panic was seen");
    #            $self->transit_to(state => $server->LOGIN);
    #        } else {
    #            $Trace->fatal(msg=>"Transit_to failed, but not as expected\n$@\n",
    #                          result=>"FATAL",
    #                         );
    #        }
    #    }
    #    $Trace->comment(msg=>"Sucessfully made it down and back up on $params{surrender_node} with AUTOBOOT=true!");
    #}

    my @stats = ($takeover_start_time, $takeover_end_time);
    $Trace->debug(msg=>"takeover is returning the following stats:
	takeover_start_time: $stats[0]
	takeover_elapsed_time: $stats[1]");
    return @stats;
    $Trace->exit();
} # end takeover

## This function verifies the aggr has root-volume
## Can't use storage_aggregate_show because burt286122, partner root-aggr shows "unknown"
## in Takeover mode. BR - tpho 05/02/2008
sub is_root_aggr {
   my $self = shift;
   my $aggr = shift;
   my $node = shift;
   my $ref_is_root = shift;        # Pointer to arrays of is_root
   my $answer = "n";
   my $hash = "";
   my $cmd = "";
   my $numb;
   my @tmp;
   my $server = $self->{server_object};

   $cmd = "system node run -node $node -command partner aggr status $aggr";
   $hash = $server->shell(cmd => "ngsh -c '$cmd'"); 
   $Trace->comment(msg=>"Dumper of \$hash:\n".Dumper($hash));

   @tmp = split(/\n/,$hash);
   $numb = scalar(@tmp);
   for (my $i=0; $i<$numb; $i++) {
      if ($i == 1) {
         my $tmp1;
         ($tmp1 = $tmp[1]) =~ s/^\s*//;
         my @stat=split(/\s+/,$tmp1);
         $Trace->comment(msg=>"tmp1='$tmp1' === '@stat'"); 
         $Trace->comment(msg=>"$stat[0] - $stat[1]");
         if ($stat[4] eq "root") {
            $answer = "y";
            push (@${ref_is_root}, $stat[0]);    ### aggr name
            push (@${ref_is_root}, $stat[1]);   ### online?
         }
      }
   }

   return($answer);
}

sub check_sfo_status {
    $Trace->entry();
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
            filers         => { type => ARRAYREF },
        }
    );
    $Trace->debug(msg => "\%params debug Dumper:\n".Dumper(\%params));
    my %hash;
    my $result = 0;

    foreach my $node (@{$params{filers}}) {
	$Trace->trace(msg=>"Getting sfo status for $node...");
	my $output = $self->storage_failover_show(node=>$node);
	$Trace->debug(msg=>"storage failover show for $node: ".Dumper($output));
	$hash{$node}{'enabled'} = $$output[0]{'enabled'} ;
	if (!$$output[0]{'enabled'} eq "true"){
	   $Trace->comment(msg=>"SFO takeover is not enabled on $node");
	   $ result = 1;
	}
	$hash{$node}{'possible'} = $$output[0]{'possible'} ;
	if (!$$output[0]{'possible'} eq "true"){
	   $Trace->comment(msg=>"SFO takeover is not possible on $node");
	   $ result = 1;
	}
    }
    if ($result){
	$Trace->comment(msg=>"SFO is not in an operational state...");
    }
    $hash{'result'} = $result;
   
    $Trace->debug(msg=>"check_sfo_status hash to be returned: ".Dumper(\%hash));
    $Trace->exit();
    return \%hash;
} # end check_sfo_status

sub find_sfo_partner {
    $Trace->entry();
    my ($self, %params) = validate_mixed(
	@_,
	[ { type => OBJECT, isa => 'Node::NG' } ],
	{
	    filer => { type => SCALAR },
	}
    );
    $Trace->trace(msg=>"Finding SFO partner for $params{'filer'}...");
    if (!$params{'filer'}) {$Trace->fatal(msg=>"find_sfo_partner was not passed a filer!")};
    my $topology = $self->{server_object}->discover_topology(cmd_nodes => [qw(sfo)]);
    $Trace->debug(msg=>"Dumping contents of discover_topology: \n".Dumper(\$topology));
    my @sfo_topo = $self->{server_object}->search_topology (
	topology => $topology,
	cmd_node => "sfo",
	key => "filer",
	value => $params{'filer'},
    );
    $Trace->debug(msg=>"DEBUG: \@sfo_topo has ".scalar(@sfo_topo)." elements.");
    $Trace->debug(msg=>"sfo topology array for $params{'filer'}: ".Dumper(@sfo_topo));
    my $partner = $sfo_topo[0]->{partnername};
    if ($partner){
	$Trace->trace(msg=>"find_sfo_partner returning $partner.");
    } else {
	$Trace->fatal(msg=>"find_sfo_partner did not find a partner for $params{'filer'}!");
    }
    $Trace->exit();
    return $partner;
} # end find_sfo_partner

sub verify_aggregate_state {
    $Trace->entry();
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
            state                       => { type => SCALAR, default => undef(), optional => 0 },
            aggregate                   => { type => SCALAR, default => undef(), optional => 0 },
            nodes                       => { type => SCALAR, default => undef(), optional => 1 },
            wait_for                    => { type => SCALAR, default => 40,      optional => 1 },
        }
    );

    my $state_target = $params{state};
    my $state_current = "";
    my @aggregate_array;

    my $timeout = $params{wait_for};
    my $aggregate = $params{aggregate};
    my $nodes = $params{nodes};
    my $retries = int($timeout/10);

    $Trace->trace(msg=>"This is the timeout...$timeout.\nThis is the retries...$retries.\nThis is the target state...$state_target.");

    if (defined $nodes) {
        eval {
        @aggregate_array = $self->storage_aggregate_show(
                                                          aggregate=>$aggregate,
                                                          nodes    =>$nodes,
                                                         );
        };
    } else {
        eval {
        @aggregate_array = $self->storage_aggregate_show(
                                                          aggregate=>$aggregate,
                                                         );
        };
    }

    if ($@) {
        $Trace->fatal(msg=>"Could not find the requested aggregate. ".$@);
    } else {
        $state_current = $aggregate_array[0]->{'state'};
    }

    while (($state_current eq 'creating') && ($state_target ne $state_current)) {
        $Trace->entry();
        if ($retries < 0) {
            $Trace->fatal(msg=>"Failed to successfully create the aggregate in the duration specified by the -wait_for parameter...$timeout");
        }

            if (defined $nodes) {
                eval {
                @aggregate_array = $self->storage_aggregate_show(
                                                                  aggregate=>$aggregate,
                                                                  nodes    =>$nodes,
                                                                 );
                };
            } else {
                eval {
                @aggregate_array = $self->storage_aggregate_show(
                                                                  aggregate=>$aggregate,
                                                                 );
                };
            }

        if ($@) {
        $Trace->fatal(msg=>"Could not find the requested aggregate. ".$@);
        } else {
            $state_current = $aggregate_array[0]->{'state'};

        }

        if ($state_current eq 'creating') {
            $Trace->trace(msg=>"This is the current state...$state_current\nWait and try again...");
            $retries--;
            sleep 10;
        }
    $Trace->exit;
    }

    if ($state_target ne $state_current) {
           $Trace->fatal(msg=>"The target state...$state_target did not match the current state...$state_current.");
    }
$Trace->exit;
} # end verify_aggregate_state

1;

