# $Id$

#
# Tharndoc comments:
#

## @summary Perl module used to setup command modules for interraction with ONTAP filer 
## @author dl-central-libs
## @pod here

=head1 NAME

Transit.pm

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

This module contains implementations of filer state and state transition methods.

=head1 SYNOPSIS

This module is invoked through an OntapNG object directly in the OntapNG namespace.

Member methods:
transit_to
state
handle_connections

=head1 EXAMPLES

To get to a node's CFE prompt:

$node->transit_to(state => $node->CFE());

To get a node booted and running ngsh:

$node->transit_to(state => $node->CLI());

To get the state of a node:

$node->state();

To open and close connections on a filer:

$node->handle_connections(
    open  => [qw(console_connecton cli_connection)],
    close => [qw(shell_connection)],
    force => 1
    );

=head1 TODO

Fix maintenance mode support.

Detect more cases when a filer panics.

=cut

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDEPNG_TRANSIT}) {
    my $newModule = $ENV{OVERRIDENG_TRANSIT};
    delete $ENV{OVERRIDENG_TRANSIT};
    print "The main NG::Transit module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

use strict;
use warnings;
use TracePkg;
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR ARRAYREF OBJECT BOOLEAN UNDEF);
use Params::Common qw(parse_release_info);
use NATE::TestInfo;
my $Trace = new TracePkg();

$Trace->set_options(separateNewline=>1);

# The constante declarations are now defined in Node.pm

=head1 METHODS

=head2 Node::NG::transit_to

$node->transit_to(state => 'CLI');

C<transit_to> allows a user to change the state of an OntapGX filer 
arbitrarily.  It will determine what state the filer is in and take the 
necessary actions to put the filer in the desired state.  In addition, 
it will try to handle connections to the filer in an intelligent fashion.

This interface reference is incomplete.  Please refer to the code for a
complete reference.

=over

=item state 

C<state> allows users to select the desired state for the node.  The 
follow list contains all the allowable values for state.  Other values
will cause a run-time error.

=over

=item Level 0 (CFE level)

=over

=item $node->CFE();

=item $node->SIMSETUP();

=item $node->RESET_ALL();

=back 

=item Level 1 (Boot Menu level)

=over

=item $node->IBM();

=item $node->NETBOOT();

=back

=item Level 2 (BSD level)

=over

=item $node->SINGLEUSER();

#=item $node->BSD();

=back

=item Level 3 (Ontap level) 

=over

=item $node->NORMAL();

=item $node->INSTALL();

=item $node->SETUP();

=item $node->DISKINIT();

=item $node->MAINT();

=item $node->LOGIN();

=item $node->SHELL();

=item $node->CLI();

=back

=item level 4 (Fake levels)

=over

=item $node->CFE_REBOOT();

=item $node->BSD_REBOOT();

=item $node->SHELL_REBOOT();

=item $node->REBOOT();

=item $node->SHUTDOWN();

=item $node->POWERCYCLE();

=back 

=back

=item timeout

C<timeout> is the number of seconds that we will wait for node communications
on a per-communication basis.  This is a pass-through to Tharn::say().

=item maxidle

C<maxidle> is the number of seconds a communcation channel can be idle on a 
per-communication basis.  This is a pass-through to Tharn::say().

=item maxcycle

C<maxcycle> is a integer indicating the number of times we will try to change 
the state of a node before giving up.

=item powercycle

C<powercycle> indicates whether or not the user wants to force a reboot before
changing the filer's state.  Using this is funcitonally equivalent (although 
time requriements may be different) to these two calls:

=over

$node->transit_to(state => $node->REBOOT());

$node->transit_to(state => $node->any_state());

=back

=item force_all_reconnect

C<force_all_reconnect> allows a user to force connection changes.  This should
not be needed in most circumstances and unnecssary use will result in costly
connection establishment.

=item wait_for

C<wait_for> allows a user to call transit_to and require it to wait until a 
forthcoming state is reached before taking actions to transit_to the desired
state.  This is useful when you need to transit to an up state after a 
non-transit_to command shuts the filer down.  It allows you to avoid the
resulting race condition. For example:

=item no_connections

C<no_connections> allows the user to totally bypass the opening and closing 
of connections associated with certain state.  Not usually used.

=over

$node->command_that_reboots_node();

$node->transit_to(state => LOGIN, wait_for => CFE);

=back

=back

=head2 Node::NG::state 

my $where_i_am = $node->state();

C<state> will determine what state a Node::NG filer is in and return a string
constant representing that state.  This method is used heavily by transit_to.

=over

=item passive

C<passive> indicates whether or not to perform passive state checking.  This
defaults to 0 as it can take much longer to determine what state we are in 
when simply waiting for a prompt to come by normally.  It is useful when 
trying we may be in or entering a carriage return sensitive state.

=item timeout

C<timeout> is a pass through to Tharn::say() and indicates how long to wait
for a matching prompt before giving up.

=item newlines

C<newlines> indicates the maximum number of newlines to send when performing
active prompt checking.

=item maxidle

C<maxidle> is a pass through to Tharn::say() and indicates how long a 
connection can be completely idle before giving up.

=item transient

C<transient> allow state to report a transient state.  Basically, don't wait 
until a non-transient state has been detected before returning.

=item append_to

C<append_to> is a scalar reference to which the state will be appended.

=item type

C<type> allows the use to select the connection to use to check state.

=item cr_interval

C<cr_interval> an integer number of seconds to wait between the sent carriage
returns.

=item try_passive

C<try_passive> allows a user to try passive checking for one C<cr_interval>.

=item panic_string

C<panic_string> is a scalar reference in which any panic output will be placed
when a panic is detected.

=item Node::NG::handle_connections

C<handle_connections> provides a simple, clean interfaced to open and close 
connections.  Its goal is to do the right thing in as many cases as possible.
It accepts up to two lists, one of connections to open and one of connections
to close as well as a flag indicating whether or not to force connections
open.

=item open

C<open> takes an array reference holding connection names to open.

=item close

C<close> takes an array reference holding connection names to close.

=item force

C<force> takes a scalar indicating whether or not to force connections open.

=back

=cut

#
# Node::NG::transit_to
#
# This method will change the state of an OntapGX filer.  
#
# Note: this method calls itself recusively.  Please keep this in mind when making changes.
#
sub transit_to
{
    my $self = shift();
    my %opt = (
	state => "CLI",          # Where do you want to go?
        timeout => undef(),      # How long to wait for something recognizable
        dialog_timeout => 600,   # how long to wait for a dialog operations
        state_timeout => 600,    # how long to wait when determining state
        wait_timeout => 600,     # how long to wait for the upcoming state
        maxidle => 120,          # The max time to wait for anything before bailing out
        maxcycle => 8,           # The max number of boot states
        powercycle => 1,         # if transit_to fails once, power cycle and try again, uses the power_cmd 
                                 # and possibly other goodies
        reboot => 0,             # perform a reboot no matter what
        netboot => 0,            # ugh...
        boot_backup => 0,        # boot the backup image from CFE (does not change the default image)
        boot_ontap => 0,         # boot the primary image from CFE via boot_ontap command {needed for excelsior}
        shell_path => '/bin/sh', # path to sh; singleuser mode prompts for this info
        #rootfs_path => "10.98.200.111://tftpboot/pub/emichael/n2/hide.rootfs.img",       # path to the rootfs image; used if netboot scripts prompt for a path
        netdev => "e4a",         # device used for cnc network when netbooting
        kernelurl => "tftp://10.98.200.111//tftpboot/tricky/nightly/kernel",  # url for a kernel to be netbooted
        connections => ['shell_connection', 'cli_connection'],
        force_all_reconnect => 0,
        no_connections => 0,     # if set to true then handle_connections (open or closed) is never used
        wait_for => undef(),     # a state to wait for before making the requested state change.

        state_passive => 0,      # Don't send carriage returns to get a prompt
        state_try_passive => 0,  # Check to see if anything comes up before hitting enter
        state_newlines => 10,    # max number of newlines to send when determining state.
        state_cr_interval => 3,  # interval between carrriage return for non-passive
        inhibit_takeover => 0,   # When rebooting, should SFO be disabled
        @_,
    );

    if (!defined($opt{RR_FLAG})) {
       ##Default value of RR_FLAG is 1.
       $opt{RR_FLAG} = 1;
    } else {
       $Trace->comment(msg => "transit_to() is invoked with 'RR_FLAG = $opt{RR_FLAG}' option");
    }

    ##Checking if RR_FLAG was set as nate param.
    my @own_rlst = own_results();
    my @prev_rlst = prev_results();
    my @all_results = NATE::TestInfo->root()->all_results();
    my $nate_rr_flag = 0;
    foreach my $result (@own_rlst, @prev_rlst, @all_results) {
      my $type = $result->type();
      next if $type ne 'PARAM';
      my @msgs = $result->messages();
      foreach my $str (@msgs) {
         my ($value) = ($str =~ /RR_FLAG\s*=\s*(\d)/);
         if (defined($value)) {
            $nate_rr_flag = 1;
            $Trace->comment(msg => "Setting RR_FLAG from nate results");
            $opt{RR_FLAG} = $value;
         }
      }
    }

    ##Checking if RR_FLAG was specified in command line.
    if (defined($ENV{RR_FLAG})) {
       $Trace->comment(msg => "Setting RR_FLAG from nate command line param (RR_FLAG = $ENV{RR_FLAG}).");
       $opt{RR_FLAG} = $ENV{RR_FLAG};
    }

    ## Checking if RR_FLAG was specified in the object.
    if (defined($self->{RR_FLAG})) {
       $Trace->comment(msg => "Setting RR_FLAG from self object handle (RR_FLAG = $self->{RR_FLAG}).");
       $opt{RR_FLAG} = $self->{RR_FLAG};
    }

    $Trace->comment(msg => "RR_FLAG = $opt{RR_FLAG}");
    
    my @callinfo = caller(1);
    my $callmsg = "";
    if ($callinfo[3] ne "Node::NG::transit_to") {
        $callmsg = "Destination state is $opt{state}.\nScript entry point detected. Called from $callinfo[3] at $callinfo[1]\:$callinfo[2]\n";
    }

    $Trace->entry(msg => "$callmsg $opt{maxcycle} attempts remain.");
    $Trace->debug(msg => "Entered Node::NG::transit_to with:\n".Dumper(\%opt));

    if (defined($opt{timeout})) {
        $Trace->warn(msg => "The 'timeout' parameter is deprecated.\n".
            "Please use one or more of the specific timeout indicators: dialog_timeout, wait_timeout and state_timeout");
    }

    #
    # Make a hash of the options we should be passing to recursive invocations and Node::NG::transit_to
    #
    # XXX If you need recursively invoked transit_to instances to receive options passed to the 
    # script-level entry point of transit_to, they should (probably) be placed in this hash.
    #
    my %pass_options = (
        timeout             => $opt{timeout},
        dialog_timeout      => $opt{dialog_timeout},
        wait_timeout        => $opt{wait_timeout},
        state_timeout       => $opt{state_timeout},
        maxidle             => $opt{maxidle},
        force_all_reconnect => $opt{force_all_reconnect},
        no_connections      => $opt{no_connections},
        connections         => $opt{connections},
        netdev              => $opt{netdev},
        state_passive       => $opt{state_passive},
        state_try_passive   => $opt{state_try_passive},
        state_newlines      => $opt{state_newlines},
        state_cr_interval   => $opt{state_cr_interval},
	inhibit_takeover    => $opt{inhibit_takeover},
        RR_FLAG             => $opt{RR_FLAG},
    );

    if (defined($opt{wait_for})) {
        if (!grep({$opt{wait_for} eq $_} $self->state(list => 1))) {
            $Trace->fatal(msg => "The requested wait_for state (\"$opt{wait_for}\") isn't a valid, detectable state.");
        } else {
            $Trace->debug(msg => "$opt{wait_for} is a valid wait_for state.");
        }
    }

    #
    # open the console if necessary and if no_connection param is false
    #
    unless ($opt{no_connections}) {
        $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
        $self->handle_connections(open => [qw(console_connection)], force => $opt{force_all_reconnect} ); 
    }
    #
    # Reboot if requested, then finish a normal transit_to.
    #
    if ($opt{reboot}) {
        $self->transit_to(state => $self->REBOOT(), wait_for => $self->CFE(), %pass_options);
        $opt{maxcycle} --;
    }

    #
    # Handle booting the "backup" image from CFE
    #
    my $boot_command;
    if ($opt{boot_backup}) {
      $Trace->trace(msg => "Backup OntapNG image will be used...");
      $boot_command = 'boot_backup';
    } elsif ($opt{boot_ontap}) {
      $Trace->trace(msg => "Primary OntapNG image will be used via boot_ontap command...");
      $boot_command = 'boot_ontap';      
    } else {
      $Trace->debug(msg => "Primary OntapNG image will be used...");
      $boot_command = 'boot_primary';
    }

    #
    # Should SFO be disabled for a reboot cycle?
    #
    my $reboot_options = '';
    if ($opt{inhibit_takeover}) {
	$Trace->trace(msg => "SFO will not be enabled for this reboot instance.");
	$reboot_options = ' -inhibit-takeover true';
    }	

    #
    # Handle options required for netbooting.
    #
    my $netboot_command = "";
    if ($opt{state} eq "NETBOOT"){
	my $ip = $opt{ip} || $self->{connections}->{console_connection}->{object}->{hostp}->{MGMT_IP};
	my $mask = $opt{mask} || $self->{connections}->{console_connection}->{object}->{hostp}->{MGMT_MASK};
	my $gateway = $opt{gateway} || $self->{connections}->{console_connection}->{object}->{hostp}->{MGMT_GATEWAY};
	my $netdev = $opt{netdev} || $self->{connections}->{console_connection}->{object}->{hostp}->{MGMT_PORT};
	if ($opt{dhcp}) {
	    $netboot_command = "ifconfig -auto $netdev ; netboot $opt{kernelurl}";
	} else {
	    $netboot_command = "ifconfig -addr=$ip -mask=$mask -gw=$gateway $netdev; netboot $opt{kernelurl}";
	}
	if (!$ip || !$mask || !$gateway || !$netdev) {
	    $Trace->fatal(msg => "Interface config info is not complete! ip:$ip, mask:$mask, gateway:$gateway, netdev:$netdev", result => 'CONFIG');
	}

	$Trace->debug(msg => "Using the following netboot command: $netboot_command");
    }

    #
    # This array holds all the state information.  Each index is a level at 
    # which there are one or more unique states.
    #
    my @NgStates = (
        #
        # CFE level
        #
        {
            CFE => { 
                close_connections => ['cli_connection', 'shell_connection'],
                transient => 0,
                accessible_states => { 
                    LOGIN => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => "$boot_command\r"},
                               {qr/(^Starting program at |ONTAP|^mroot is now available|^filter sync|^Welcome to the cluster|^Configuring syscons|^login:)/m => "" } ],
                    CFE_REBOOT => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => "bye\r"} ], 
                    IBM => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => "$boot_command\r"},
                             {qr/^(\* )?Press Ctrl-C for/m => "\cc"},
                             # we want to match on the boot menu header and footer from RR and SN
                             # here as well as the normal "boot menu available message"
                             {qr/^Please choose one of the following:|^Selection \(1-\d\)|Boot Menu will be available/m => ""} ],
                    SIMSETUP => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => "boot_simsetup\r"}, {qr/^Enter selection:/m => "" } ],
                    SINGLEUSER => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => "$boot_command -s\r"}, {qr/^Enter full pathname of shell or RETURN for /m => "$opt{shell_path}\r"} ],
                    NETBOOT => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => "$netboot_command\r"}, {qr/login:/m => "root" }, {qr/^Skipped mounting the CFCARD/m => ""}, {qr/How would you like to continue booting/m => ""}], 
                    SHELL => [ {states => [$self->LOGIN()]} ], 
                    SHELL_REBOOT => [ {} ],
                    #SHELL_REBOOT => [ {states => [CFE_REBOOT]} ],
                    CLI => [ {states => [$self->LOGIN()]} ], 
                    INSTALL => [ {states => [$self->IBM]} ],
                    DISKINIT => [ {states => [$self->IBM]} ],
                    SETUP => [ {states => [$self->IBM]} ],
                    NORMAL => [ {states => [$self->IBM]} ],
                    MAINT => [ {states => [$self->IBM]} ],
                    FAKEINIT => [ {states => [$self->IBM]} ],
		            DBLADECLI => [ {states => [$self->SHELL] } ],
                    RESET_ALL => [ {states => [$self->SIMSETUP] } ],
                },
            },
            {
                SIMSETUP => {
                    transient => 0,
                    accessible_states => {
                        RESET_ALL => [ {qr/^Enter selection:/m => "8\r"},{qr/like to reboot/m => "y"} ],
                        REBOOT => [ {qr/^Enter selection:/m => "10\r"} ],
                        CFE => [ {qr/^Enter selection:/m => "10\r"} ],
                    }
                }
            },
            {
                RESET_ALL => {
                    transient => 1,
                    accessible_states => {},
                }
            },
        },
        #
        # IBM level
        #
        {
            IBM => {
                transient => 0,
                close_connections => ['cli_connection', 'shell_connection'],

                # We are matching against the BR or the RR converged boot menu prompt, then sending
                # the BR selection AND the converged selection.  _executeDialogue() uses -blind on the
                # say option for the response so the extra selection should not be a problem for the
                # non-converged boot menu.  The reason I am giving the older command first is
                # that the non-converged bootmenu sometimes seems to accept numeric input (i.e. 1
                # seems to be the same as "normal" even though 22/7 (or hidden) doesn't seem to display
                # 1 as a valid selection.)  Entering any of the alphabetic selections in the converged
                # menu will cause the "Selection (1-8)?" prompt to be redisplayed.
                accessible_states => {
                    NORMAL   => [ {qr/Please make a selection: |Selection \(1-\d\)\?/m => "normal\r1\r"}, ],
                    INSTALL  => [ {qr/Please make a selection: |Selection \(1-\d\)\?/m => "install\r7\r"}, ],
                    SETUP    => [ {qr/Please make a selection: /m => "setup\r"} ],
                    DISKINIT => [ {qr/Please make a selection: |Selection \(1-\d\)\?/m => "init\r4\r"}, ],
                    MAINT    => [ {qr/Please make a selection: |Selection \(1-\d\)\?/m => "maint\r5\r"}, ],
                    REBOOT   => [ {qr/Please make a selection: |Selection \(1-\d\)\?/m => "reboot\r8\r"},
                                  {qr/Reboot now\? (?:{y\|n}:|\(y or n\))/m => "y\r"} ],
                    FAKEINIT => [ {qr/Please make a selection: |Selection \(1-\d\)\?/m => "fake_init\r24/6\r"} ], 
                }
            }
        },
	{
	    NETBOOT => { 
		transient => 1,
		#entry_prompts => {qr/Enter the NFS path for rootfs.img/m => "$opt{rootfs_path}\r"},
		#entry_prompts are currently broken.  If RC prompts us for a rootfs path, this script will fail
		accessible_states => {},
	    }
	},
        
	#
        # BSD Levels
        #
        {
            SINGLEUSER => {
                transient => 0,
                accessible_states => {
                    SHELL_REBOOT => [ {qr/^#\s/m => "shutdown -r now\r"} ],
                    REBOOT => [ {qr/^#\s/m => "shutdown -r now\r"} ],
                }
            }
        },
        #{
        #    BSD => {
        #        transient => 0,
        #        accessible_states => {
        #            SHELL_REBOOT => [ {qr/#\s/m => "shutdown -r now\r"} ],
        #            REBOOT => [ {qr/#\s/m => "shutdown -r now\r"} ],
        #            CFE => [ {qr/#\s/m => "shutdown -r now\r"} ],
        #        }
        #    }
        #},
        
        #
        # Ontap Levels
        # 
        {
            NORMAL => {
                transient => 0,
                accessible_states => {},
            },
            INSTALL => {
                transient => 1,
                accessible_states => {},
            },
            SETUP => {
                transient => 1,
                accessible_states => {},
            },
            DISKINIT => {
                transient => 1,
                accessible_states => {},
            },
            MAINT => {
                transient => 0,
                accessible_states => {},
            },
            FAKEINIT => {
                transient => 1,
                accessible_states => {},
            },
            LOGIN => $opt{RR_FLAG} == 0 ? {
                transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    CFE => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => '', states => [$self->HALT]} ],
                    SHELL => [ {qr/^login:\s/m => "$self->{SHELL_USER}\r"}, {qr/^\s*Password:/m => "$self->{SHELL_PASSWORD}\r"} ],
                    SHELL_REBOOT => [ {states => [$self->SHELL]} ],
                    CLI => [ {states => [$self->SHELL]} ],
                    REBOOT => [ {qr/^login:\s/m => "$self->{SHELL_USER}\r"}, {qr/^\s*Password:/m => "$self->{SHELL_PASSWORD}\r"}, {qr/^$self->{server}(#|%)\s/m => "ngsh -c \'set -confirmations off; system reboot $self->{server} $reboot_options\'\r"} ],
                    HALT => [ {qr/^login:\s/m => "$self->{SHELL_USER}\r"}, {qr/^\s*Password:/m => "$self->{SHELL_PASSWORD}\r"}, {qr/^$self->{server}(#|%)\s/m => "ngsh -c \'set -confirmations off; system halt $self->{server} $reboot_options\'\r"} ],
                    DBLADECLI => [ {states => [$self->SHELL]} ],
                },
            } : {
                transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    CFE => [ {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => '', states => [$self->HALT]} ],
                    SHELL => [ {states => [$self->CLI]} ],
                    SHELL_REBOOT => [ {states => [$self->SHELL]} ],
                    CLI => [ {qr/^login:\s/m => "admin\r"}, {qr/^\s*Password:/m => "$self->{CLI_PASSWORD}\r"} ],
                    REBOOT => [ {qr/^login:\s/m => "admin\r"}, {qr/^\s*Password:/m => "$self->{CLI_PASSWORD}\r"}, {qr/^\S+::[^>#%]*\*?[>#%]\s/m  => "set -confirmations off; system reboot $self->{server} $reboot_options\r"} ],
                    HALT => [ {qr/^login:\s/m => "admin\r"}, {qr/^\s*Password:/m => "$self->{CLI_PASSWORD}\r"}, {qr/^\S+::[^>#%]*\*?[>#%]\s/m  => "set -confirmations off; system halt $self->{server} $reboot_options\r"} ],
                    DBLADECLI => [ {states => [$self->CLI]} ],
                },
            },
            SHELL => $opt{RR_FLAG} == 0 ? {
                transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    CLI => [ {qr/^$self->{server}#\s/m => "ngsh\r"} ],
                    LOGIN => [ {qr/^$self->{server}#\s/m => "logout\r"} ],
                    REBOOT => [ {qr/^$self->{server}#\s/m => "ngsh -c \'set -confirmations off; node reboot $self->{server} $reboot_options\'\r"} ],
                    SHELL_REBOOT => [ {qr/^$self->{server}#\s/m => "shutdown -r now\r"} ],
                    HALT => [ {qr/^$self->{server}#\s/m => "shutdown -h\r" } ],
		            DBLADECLI => [ {qr/^$self->{server}#\s/m => "dbladecli\r" } ],
                }
            } : {
                transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    CLI => [ {qr/^$self->{server}(#|%)\s/m => "exit\r"} ],
                    LOGIN => [ {states => [$self->CLI]} ],
                    REBOOT => [ {states => [$self->CLI]} ],
                    SHELL_REBOOT => [ {qr/^$self->{server}(#|%)\s/m => "shutdown -r now\r"} ],
                    HALT => [ {qr/^$self->{server}(#|%)\s/m => "shutdown -h\r" } ],
		    DBLADECLI => [ {qr/^$self->{server}(#|%)\s/m => "dbladecli\r" } ],
                }
            },
            CLI => $opt{RR_FLAG} == 0 ? {
                transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    SHELL => [ {qr/^$self->{server}::(\w|\s)*\*?>\s/m  => "exit\r"} ], 
                    LOGIN => [ {states => [$self->SHELL]} ],
                    REBOOT => [ {qr/^$self->{server}::(\w|\s)*\*?>\s/m => "set -confirmations off; reboot -node $self->{server} $reboot_options\r"} ],
                    SHELL_REBOOT => [ {states => [$self->SHELL]} ],
                    DBLADECLI => [ {states => [$self->SHELL]} ],
                }
            } : {
                transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    SHELL => [ {qr/^\S+::[^>#%]*\*?[>#%]\s/m  => "set -confirmations off; set -privilege diagnostic; systemshell -node local\r"}, {qr/^login:\s/m => "diag\r"}, {qr/^\s*Password:/m => "$self->{CLI_PASSWORD}\r"}, {qr/(%|^%)/m => "exec sudo sh\r"}, {qr/^#/m => "export PS1='$self->{server}# '\r"} ],
                    LOGIN => [ {qr/^\S+::[^>#%]*\*?[>#%]\s/m => "exit\r"} ], 
                    REBOOT => [ {qr/^\S+::[^>#%]*\*?[>#%]\s/m => "set -confirmations off; reboot -node $self->{server} $reboot_options\r"} ],
                    SHELL_REBOOT => [ {states => [$self->SHELL]} ],
                    DBLADECLI => [ {states => [$self->SHELL]} ],
                }
            },
	    DBLADECLI => {
		# there are currently two ways to access the dbladecli - via bsd and
		# via the cli.  Unfortunatly, each has it's own way to exit.  We only
		# currently deal with the "non CLI" dbladecli for now, but may have
		# to add another "STATE" to deal with the other...
		transient => 0,
                open_connections => $opt{connections},
                accessible_states => {
                    SHELL => [{qr/($self->{server}|dblade)\*?>/m => "\cD\r" }],
                    LOGIN => [ {states => [$self->SHELL]} ],
                    SHELL_REBOOT => [ {states => [$self->SHELL]} ],
                    CLI => [ {states => [$self->SHELL]} ],
                    REBOOT => [ {states => [$self->SHELL]} ],
                    HALT => [ {states => [$self->SHELL]} ],
		},
	    },
        },

        #
        # Semi-fake levels to help us out
        #
        {
            HALT => {
                transient => 1,
               accessible_states => {
                    CFE => [ {qr/^Please press any key to reboot/m => "y"},
                    {qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?> /m => ''}],
                },
            },
            CFE_REBOOT => {
                transient => 0,
                equivalent_state => $self->CFE,
            },
            BSD_REBOOT => {
                transient => 1,
            },
            SHELL_REBOOT => {
                transient => 1,
                close_connections => ['cli_connection', 'shell_connection'],
            },
            REBOOT => {
                transient => 1,
                close_connections => ['cli_connection', 'shell_connection'],
            },
            SHUTDOWN => {
                transient => 0,
                close_connections => ['cli_connection', 'shell_connection'],
            },
            POWERCYCLE => {
                transient => 0,
                close_connections => ['cli_connection', 'shell_connection'],
            },
        }
    );

    #
    # Figure out where we are and wait for the specified state (if any).
    #
    my $from;
    my $wait_for_time = time() + $opt{wait_timeout};
    do {
        if (defined($opt{wait_for})) {
            $Trace->debug(msg => "Waiting for state ".$opt{wait_for}." to come along.\n".
                ($wait_for_time - time())." seconds of waiting remain.");
            if ($wait_for_time < time()) {
                $Trace->fatal(msg => "We were requested to wait for the state ".
                    $opt{state}." before continuing and did not reach this state ".
                    "before the wait_timeout (".$opt{wait_timeout}."s)");
            }
            select(undef(), undef(), undef(), 1.0);
        }

        # check the state
        $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
        $from = $self->state(timeout        => $opt{state_timeout},
                             passive        => $opt{state_passive},
                             try_passive    => $opt{state_try_passive},
                             newlines       => $opt{state_newlines},
                             cr_interval    => $opt{state_cr_interval},
                             no_connections => $opt{no_connections},
                             maxidle        => $opt{maxidle},
                            );

    } while (defined($opt{wait_for}) && ($from ne $opt{wait_for}));

    #
    # Figure out exactly where we are in the state structure.
    #
    my $start_state;
    my $start_level = 0;

    my $end_state;
    my $end_level = 0;

    if ($from eq $self->UNKNOWN()) {
        $Trace->fatal(msg => "Can't determine the filer's state!")
    } elsif($from eq $self->HANG()) {
        $Trace->fatal(msg => "The filer has hung.");
    }

    for (my $i = 0; $i < scalar(@NgStates); $i ++) {
        if (defined(${$NgStates[$i]}{$from})) { 
            $start_state = $from;
            $start_level = $i;
        }

        if (defined(${$NgStates[$i]}{$opt{state}})) { 
            $end_state = $opt{state};
            $end_level = $i;
        }
    }

    if (!defined($start_state)) {
        $Trace->fatal(msg => "Start State \"$start_state\" not defined!");
    }
    if (!defined($end_state)) {
        $Trace->fatal(msg => "End State \"$end_state\" not defined!");
    }
    
    #
    # Close connections that should be closed in this state only if no_connection param is false
    #
    unless ($opt{no_connections}) {
        $Trace->trace(msg => "Closing connections for $start_state");
        $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
        $self->handle_connections(close => ${NgStates[$start_level]}{$start_state}{close_connections}); 
    }
    #
    # Make sure we aren't already where we want to be.
    #
    if ($start_state eq $end_state) {
        $Trace->trace(msg => "We are already in the end state: $end_state.");
        # make sure the appropriate connections are open even if we're in the proper state
        # but only if no_connection param is false
        unless ($opt{no_connections}) {
            $Trace->trace(msg => "Opening connections for $end_state");
            $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
            $self->handle_connections(open => ${NgStates[$end_level]}{$end_state}{open_connections}, force => $opt{force_all_reconnect}); 
        }
        $Trace->exit(msg => "Node is in $end_state state.  Unwinding...");
        return $end_state;
    } elsif ($opt{maxcycle} == 0) {
        #
        # If too many cycles have been made, bail out.
        # Note: cycles = state changes (this is the terminology used in boot.pm)
        #
        $Trace->fatal(msg => "Cycled too many times.");
    }

    #
    # Check for a direct transition. If no direct transtion exists, reboot and 
    # transit_to again.
    #
    my $level_ref = ${$NgStates[$start_level]}{$start_state};
    my $end_ref = ${NgStates[$end_level]}{$end_state};
    my @accessible_states = keys(%{$$level_ref{accessible_states}});

    $Trace->trace(msg => "Currently at $start_state/$level_ref, can get to: ".join(", ", @accessible_states));

    #
    # If there are intermediate states between the current and out destination 
    # transit_to them.
    #
    foreach my $intermediate_state (@{${${$$level_ref{accessible_states}}{$end_state}}[0]{states}}) {
        $Trace->trace(msg => "Moving to intermediate state $intermediate_state on the way to $end_state.");
        $self->transit_to(state => $intermediate_state, maxcycle => --$opt{maxcycle}, boot_backup => $opt{boot_backup}, %pass_options);
        if ($intermediate_state eq $self->HALT) {
            sleep 10;
        }
        #
        # set a new start level
        #
        for (my $i = 0; $i < scalar(@NgStates); $i ++) {
            if (defined(${$NgStates[$i]}{$intermediate_state})) { 
                $start_state = $intermediate_state;
                $start_level = $i;
            }
        }

        $level_ref = ${$NgStates[$start_level++]}{$start_state};
    }

    $Trace->trace(msg => "Transition from $start_state ($start_level) to $end_state ($end_level) requested.");

    my $dialogue;
    my $entryprompts;

    #
    # If we can get to the next state, go there now.
    # Otherwise reboot and try again.
    #
    if (grep(/$end_state/, @accessible_states)) {
        $Trace->trace(msg => "The end state $end_state is reachable from this state.");

        $dialogue = ${$$level_ref{accessible_states}}{$end_state};
	#$entryprompts = ${$level_ref}{entry_prompts};
	$entryprompts = ${$end_ref}{entry_prompts};
        my $extracrs = 0;

        #
        # hack for console garbage characters problem
        #
        if ($start_state eq 'LOGIN') {
	    $extracrs = 0;
    	    # We've been experiencing some problems getting logged in, so we're going to be a little more
	    # forgiving of timeouts
	    my $curretry = 1;
	    my $maxretry = 5;
	    while ($curretry <= $maxretry) {
	        eval {
		    $self->_executeDialogue($dialogue, $entryprompts, $opt{dialog_timeout}, $opt{maxidle}, $extracrs);
	        };
	        if ($@){
		    $Trace->warn(msg => "Attempt $curretry for login resulted in a LOGIN PROMPT SEQUENCE TIMEOUT.");
		    ++$curretry;
		    next;
	        } else {
		    last;
	        }
	    }
        } else {
            $self->_executeDialogue($dialogue, $entryprompts, $opt{dialog_timeout}, $opt{maxidle}, $extracrs);
        }
        #
        # This is a hack to keep transit_to(state => REBOOT) working. A more complete fix for this requires
        # linking REBOOT with CFE.
        #
        if ($end_state eq 'REBOOT') {
            $self->transit_to(state => $self->CFE, wait_for => $self->CFE, %pass_options);
        }
	
    } else {
        $Trace->trace(msg => "The end state $end_state is not reachable from this state, we must reboot.");
        $self->transit_to(state => $self->REBOOT, maxcycle => --$opt{maxcycle}, %pass_options);
        $self->transit_to(state => $opt{state}, wait_for => $self->CFE, maxcycle => --$opt{maxcycle}, boot_backup => $opt{boot_backup}, %pass_options);
    }

    my $new_state = $end_state;

    #
    # Once all the transitions are done, and we're not entering a cr-sensitive 
    # state, check to see if we are where we should be.
    #
    if (0 == $NgStates[$end_level]{$end_state}{transient}) {

        $Trace->trace(msg => "Entered a stable state; try to verify that we're where we should be.");
        $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
        $new_state = $self->state(timeout        => $opt{state_timeout}, 
                                  passive        => $opt{state_passive},
                                  try_passive    => $opt{state_try_passive},
                                  newlines       => $opt{state_newlines},
                                  cr_interval    => $opt{state_cr_interval},
                                  no_connections => $opt{no_connections},
                                  maxidle        => $opt{maxidle},
                                );

        my @allowable_states = ($opt{state},$NgStates[$end_level]{$end_state}{equivalent_state});
        if (!grep($_ && $new_state eq $_, @allowable_states)) {
             if ($opt{powercycle}) {
                #
                # TODO We need a way to powercycle machines.
                #
                $opt{maxcycle}--;
                $opt{powercycle} = 0;
                $Trace->trace(msg=>"$new_state ne $opt{state}. Run transit_to() again");
                $new_state = $self->transit_to(state => $opt{state}, maxcycle => $opt{maxcycle},powercycle => $opt{powercycle},boot_backup => $opt{boot_backup},%pass_options);
            }
         }
        
         if (!grep($_ && $new_state eq $_, @allowable_states)) {
            #
            # We can't seem to transit to the correct state.  Either something is wrong with our 
            # directions or something for which we haven't accounted happened on the filer.
            # If we have not re-tried it yet, we will try again after inverting the RR_FLAG value.
            #
            if (!$nate_rr_flag && !defined($self->{RR_FLAG})) {
               if ($opt{RR_FLAG}) {
                  $self->{RR_FLAG} = 0;
               } else {
                  $self->{RR_FLAG} = 1;
               }
               $Trace->comment(msg => "Re-trying with RR_FLAG = $self->{RR_FLAG}");
               $new_state = $self->transit_to(state => $opt{state},maxcycle => --$opt{maxcycle}, boot_backup => $opt{boot_backup},%pass_options);
               if (!grep($_ && $new_state eq $_, @allowable_states)) {
                  $Trace->fatal(msg => "Even after inverting RR_FLAG, $new_state not one of: ".join(",", @allowable_states));
               }
            } else {
               $Trace->fatal(msg => "$new_state not one of: ".join(",", @allowable_states));
            }
        }
        # If requested RR_FLAG, then this is a more recent node than Tricky.  Set the flag for others to enjoy.
        $self->{RR_FLAG} = $opt{RR_FLAG};
    } else {
        $Trace->trace(msg => "Entered a transient (or cr-sensitive) state; don't try to verify anything.");
    }
   
    # 
    # I don't think this should be needed.  Obviously, I'm wrong.
    #
    unless ($opt{no_connections}) {
        $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
        $self->handle_connections(open => ${NgStates[$end_level]}{$end_state}{open_connections}, force => $opt{force_all_reconnect}); 
    }
    
    $Trace->exit(msg => "The filer is in the $new_state state.  $opt{maxcycle} attempts remain.");
    return $new_state;
}

#
# handle_connections does the work of opening and closing our connections
#
#    This is how we handle connections normally and when forcing.  An 
#    'x' means nothing will be done, an 'o' means the operation will
#    occur.
#
#    Conn.   |       |      | force | force 
#    Ref.    | close | open | close | open
#    --------+-------+------+-------+-------
#    undef   |   x   |  o   |  x    |  o     
#    --------+-------+------+-------+-------
#    defined |   o   |  x   |  o    |  o    
#
#    Note: forcing a defined connection to reopen causes the connection to
#    be closed first.
#
#    Note: you must always reconnect the shell connection first when the CLI 
#    connection needs to be reconnected.
#
sub handle_connections {
    $Trace->entry();
    my $actions;

    my ($self, %p) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Node::NG' } ],
        {
            open => { type => (ARRAYREF | UNDEF), optional => 1 },
            close => { type => (ARRAYREF | UNDEF), optional => 1 },
            force => { type => BOOLEAN, default => undef() },
        }
    );

    $Trace->trace(msg => "The following operations were requested: ".(defined($p{open}) ? "OPEN: ".join(", ", @{$p{open}}).($p{force} ? " (forcefully)" : " (if necessary)")."\n" : "").(defined($p{close}) ? " CLOSE: ".join(", ", @{$p{close}})."\n" : ""));

    # If someone is calling handle_connections and the NODE_RELEASE_INFO param has already been
    #  set, then we need to clear it here so that the release information will be
    #  refreshed when the node comes back up. This is necessary as the node might
    #  have gotten new software installed.
    if ( parse_release_info(node=>$self->{server})) {
        main::logresult("PARAM", "NODE_RELEASE_INFO=");
        $self->{options_object}->set(NODE_RELEASE_INFO=>undef);
    }

    if ($p{force}) {
        $Trace->comment(
            msg => " *** Please evaluate the need to force connection changes. You ***\n".
                   " *** should never need to force connection changes in normal   ***\n".
                   " *** circumstances and by doing so, you may be masking a bug.  ***");
    }

    #
    # Close connections
    #
    if (defined($p{close})) {
        foreach my $conn (@{$p{close}}) {
            if (defined($self->{connections}->{$conn})) {
                $Trace->trace(msg => "Closing $conn");
                $self->close_connection(type => $conn);
                $actions ++;
            } else { 
                $Trace->debug(msg => "Skipped closing $conn, the connection reference wasn't defined.");
            }
        }
    }

    #  
    # Open connections
    #
    if (defined($p{open})) {
        foreach my $conn (@{$p{open}}) {
            # only force if requested
            if ($p{force} || !defined($self->{connections}->{$conn})) {
                $Trace->trace(msg => "Opening $conn");
                if (defined($self->{connections}->{$conn})) {
                    $Trace->trace(msg => "Forcing connection $conn closed before reopening.");
                    $self->close_connection(type => $conn);
                }
                $self->establish_connections($conn => 1);
                $actions ++;
            } else {
                $Trace->trace(msg => "$conn already established and force not requested.");
            }
        }
    }
    $Trace->exit();
    return $actions;
} # end handle_connections

#
# Executes a dialogue desgined to get us between states.
#
sub _executeDialogue
{
    my $self = shift();
    my $dialogue = shift();
    my $entryprompts = shift();
    my $to = shift();
    my $maxidle = shift();
    my $extracrs = shift();

    my @preprompts;

    $Trace->entry();
    #
    # Optional prompts that may show up will be caught by separate subtests.
    # Note: These are UNUSED at the moment
    #
    while (my ($pprompt, $answer) = each (%{$entryprompts})) {
        push(@preprompts, Tharn::subtest(
                sub {
		    $Trace->debug(msg => "Subtest looking for preprompt: $pprompt");
                    Tharn::say(
                        $self->{connections}->{console_connection}->{object}, 
                        "", 
                        '-wait', $pprompt, 
                        -noprompt,
			-nocr, 
                        -timeout, $to, 
                    );
                    $Trace->debug(msg => "Saw preprompt: $pprompt");
                    Tharn::say(
                        $self->{connections}->{console_connection}->{object}, 
                        $answer."\r", 
                        -nocr, 
                        -blind
                    );
                    $Trace->debug(msg => "Sent reply: \"$answer\"");
                }, 
                -bg, 
                -runid, 'entry_prompt_catcher'
            )
        );
    }
    
    my $sentcr = 0;

    # Get a local copy of the the dynamic Tharn::THARN_VERBOSE to keep
    # scope our modification within this block only.
    local $::THARN_VERBOSE = 1 if ($extracrs);
    #
    # advance through each prompt and response in the dialogue
    #
    foreach my $pair (@{$dialogue}) {
        while (my ($regex, $resp) = each(%{$pair})) {
            if ($regex eq 'states') {
                next();
            }

	    $Trace->debug(msg => "Looking for regex: $regex, and replying with $resp");

            if (!$sentcr) {
                $Trace->debug(msg => "Sent \\r");
                Tharn::say($self->{connections}->{console_connection}->{object}, "\r", -nocr, -blind);
                $sentcr = 1;
            }
	    if ($extracrs){
                $Trace->debug(msg => "Going extracr route");
		
		my $sawit = 0;
		my $finishtime = time() + $to;
		while (time() < $finishtime){
		    my $match = eval {
			Tharn::say(
				  $self->{connections}->{console_connection}->{object}, 
				  "\r", 
				  '-wait', $regex, 
				  -noprompt,
				  -nocr, 
				  -timeout, 2, 
				  -maxidle, $maxidle
				  );
		      };
		    if ($@){
			# More than likely expected since we haven't hit out real timeout
			if ($@ =~ /TIMEOUT/){
			    $Trace->debug(msg => "Short time out: $@");
			} else {
			    $Trace->fatal(msg => "Caught say failure: $@");
			}
		    } else {
			# We must've seen the regex
			$Trace->debug(msg => "Saw match $match");
			if ($match && length($match)){
			    $sawit = 1;
			    last;
			}
		    }
		}
		if (!$sawit){
		    $Trace->fatal(msg => "Did not see regex $regex in $to seconds");
		}
		$Trace->debug(msg => "Saw $regex => $sawit");
		$extracrs--;
	    } else {
                my $match;
                eval {
                    $match = Tharn::say(
                                        $self->{connections}->{console_connection}->{object}, 
                                        "", 
                                        '-wait', $regex, 
                                        -noprompt,
                                        -nocr, 
                                        -timeout, $to, 
                                        -maxidle, $maxidle,
                                        -swallow,
                                       );
                };
                $Trace->fatal(msg=>$@) if ($@);
		$Trace->debug(msg => "Saw $regex");
	    }

            Tharn::say($self->{connections}->{console_connection}->{object}, $resp, -nocr, -blind);
            $Trace->debug(msg => "Sent $resp");
        }
    }

    #
    # stop all optional prompt processing
    #
    $::THARN_VERBOSE = 0;
    foreach my $pprompt (@preprompts) {
        $pprompt->kill(message => 'ok');
    }
}


sub state 
{
    my $self = shift();
    my %opt = (
        passive => 0,          # Don't send carriage returns to get a prompt
        timeout => 600,        # How long to wait for something recognizable
        newlines => 1,         # Specifies the number of newlines to send when determining
                               # state.
        maxidle => 120,        # The max time to wait for anything before bailing out
        transient => 0,        # Return a transient state or wait for something stable
        append_to => undef(),  # A scalar reference to which the state should be appended
        type => 'console',     # the type of connection to use
        cr_interval => '3',    # the interval between carrriage return presses for non-passive
                               # checks
        try_passive => 0,      # Check to see if anything comes up before hitting enter
        panic_string => undef(), # a scalar reference to hold panic output
        list => undef(),
        no_connections => 0,   # is true then do not open/close connections
        @_
    );

    if (defined($opt{panic_string}) && (ref($opt{panic_string}) ne 'SCALAR')) {
        $Trace->fatal(msg => "The provided panic_string can't be used to store panic output, it must be a scalar reference.", result => 'SCRIPT');
    }

    $Trace->entry(msg=>"Entered Node::NG::state with:\n".Dumper(\%opt));
    #
    # Some info about what states we might be in
    #
    my %OntapNgStates = (
        HANG     => { prompts => [],                                           transient => 0, booted => 0, setprompts => [],},
        UNKNOWN  => { prompts => [],                                           transient => 0, booted => 0, setprompts => [],},
        PANIC    => { prompts => [ qr/panic:.*^DUMPCORE: END/ms, qr/DUMPCORE: END/m ], transient => 1, booted => 1, setprompts => ['CFE(?:-\w+)?\*?>'],},
        CFE_REBOOT => { dialogue => {qr/^Starting AUTOBOOT.+/m => "\cc" },   transient => 1, booted => 0, setprompts => [],},
        NO_MROOT => { prompts => [ qr/^WARNING: Giving up waiting for mroot/], transient => 1, booted => 1, setprompts => [],}, 
        CLI      => { prompts => [ qr/^\S+::[^>#%]*\*?[>#%]\s/m ],    transient => 0, booted => 1, setprompts => [],},
        MAINT    => { prompts => [ qr/^(\s*\r*\n*\s*)*\*?>\s/m ],              transient => 0, booted => 1, setprompts => [],},
        SHELL    => { prompts => [ qr/^$self->{server}(#|%)\s/m ],                 transient => 0, booted => 1, setprompts => [],},
        RLM      => { prompts => [ qr/^RLM $self->{server}\*?>\s/m ],          transient => 0, booted => 2, setprompts => [],},
        CFE      => { dialogue => {qr/^Starting AUTOBOOT.+|or any other key for command/m => "\cc" }, prompts => [ qr/^(CFE|LOADER|SIMLOADER|VLOADER)(?:-\w+)?\*?>\s$/m ],       transient => 0, booted => 0, setprompts => ['CFE(?:-\w+)?\*?>','LOADER(?:-\w+)?\*?>','SIMLOADER(?:-\w+)?\*?>','VLOADER(?:-\w+)?\*?>'], },
        LOGIN    => { prompts => [ qr/^login:\s/m, qr/^\s*Password:/m ],       transient => 0, booted => 2, setprompts => [],},
        IBM      => { prompts => [ qr/Please make a selection:|Please choose one of the following:|Please choose a selection|Selection \(1-\d\)\?/ ],             transient => 0, booted => 0, setprompts => [],},
        '24/6'   => {  prompts => [ qr/Install new file system without zeroing disks/ ], transient => 0, booted => 0, setprompts => [],},
        SINGLEUSER => { prompts => [ qr/^#\s/m ],                              transient => 0, booted => 0, setprompts => ['#'],},
        #NO_KMOD    => { prompts => [ qr/#\s/m ],                               transient => 0, booted => 0, setprompts => ['#'],},
        #BSD    => { prompts => [ qr/#\s/m ],                               transient => 0, booted => 1, setprompts => ['#'],},
        DBLADECLI  => { prompts => [ qr/($self->{server}|dblade)\*?>/m ],                    transient => 0, booted => 1, setprompts => ["$self->{server}\*?>",'dblade\*?>'],},
        SIMSETUP   => { prompts => [ qr/^Enter selection\:/m ],                transient => 0, booted => 0, setprompts => ['Enter selection\:'],},
        HALT       => { prompts => [ qr/^The operating system has halted/m,
                                     qr/^Please press any key to reboot/m ], transient => 0, booted => 0, setprompts => [], },
    );

    if ($opt{list}) {
        return keys(%OntapNgStates);
    }

    #
    # TODO Add maintenance mode support.  Prompts, setprompts, etc.
    #
    unless ($opt{no_connections}) {
        $Trace->debug(msg => "Param no_connections is set to: $opt{no_connections}");
        $self->handle_connections(open => [$opt{type}."_connection"]);
    }
    #
    # combine the prompts, matches, and other strings into an array for say()
    #
    my @waitstrings;
    while (my ($state, $info) = each(%OntapNgStates)) {
        if (exists $info->{prompts} && @{$info->{prompts}}) {
                push (@waitstrings, @{$info->{prompts}});
            }
        if (exists $info->{dialogue} && %{$info->{dialogue}}) {
                push (@waitstrings, keys %{$info->{dialogue}});
            }
    }

    $Trace->debug(msg => "Waiting for any of these: ".join(", ", @waitstrings));

    my $match;
    my $before;
    my $after;
    my $prompt;

    my $st;
    my $our_state;

    if ($opt{passive}) {
        $Trace->trace(msg => "Passively examining output from the filer.");
    } else {
        #
        # send carriage returns to the connection while we look for a reasonable prompt
        #
        if (defined($SIG{INT}) && ($SIG{INT} ne 'DEFAULT')) {
            $Trace->trace(msg => "The default signal handler for INT has been overridden.\n".
                "It is currently \"$SIG{INT}\"\n".
                "The default signal handler has been restored.\n");
            $SIG{INT} = 'DEFAULT';
        } else {
            $Trace->debug(msg => "The default signal handler for INT is ".(defined($SIG{INT}) ? $SIG{INT} : 'undef')."\n");
        }
        $st = Tharn::subtest( 
            sub {
                $Trace->trace(msg => "Poking at the filer to determine our prompt.");
                $Trace->debug(msg => "\\r generating subtest started");

                while (1) {
                    if (!$opt{try_passive}) {
                        Tharn::say($self->{connections}->{$opt{type}."_connection"}->{object}, "\r", -nocr, -blind);
                    } else {
                        $Trace->trace(msg => "Trying passive mode once...");
                        $opt{try_passive} = 0;
                    }
                    sleep($opt{cr_interval});
                }
            }, 
            -bg, 
            -runid, 'active_prompt_checking'
        );
    }

    #
    # watch the connection for anything we might recognize
    #
    $Trace->debug(msg => "checking output for matches...");
    
    eval {
        $Trace->debug(msg => "CONNECTION:".Dumper($self->{connections}->{console_connection}));
        ($match, $before, $after, $prompt) = Tharn::say(
            $self->{connections}->{$opt{type}.'_connection'}->{object}, 
            "", 
            '-wait', @waitstrings, 
            -noprompt,
            -nocr, 
            -swallow,
            -timeout, $opt{timeout}, 
            -maxidle, $opt{maxidle}
        );
        $Trace->debug(msg => "Got match \"$match\" from the filer.\n");
    };
    my $say_error = $@;

    if (!$opt{passive}) {
        #
        # stop our \r generating subtest
        #
        $Trace->trace(msg => "Stopping subtest: ".$st->runid());

        # Get a local copy of the the dynamic Tharn::THARN_VERBOSE to keep
        # scope our modification within this block only.
        local $Tharn::THARN_VERBOSE = 2;

        $st->kill(message => 'ok');
    }

    if ($say_error =~ m/Err=IDLE/) {
        $Trace->warn(msg => "Didn't see any output within the idle timeout. The filer is most likely hung or crashed.");
        $our_state = $self->HANG;
    } elsif ($say_error =~ m/Err=TIMEOUT/) {
        $Trace->warn(msg => "Didn't see any recognizable output within the timeout. The filer did appear to be running.");
        $our_state = $self->UNKNOWN;
    } elsif ($say_error) {
        $Trace->fatal(msg => "An unexpected error occurred while trying to determine state: \"$say_error\"");
    } else {
        #
        # Check to see which prompt we got back and return the corresponding state
        #
        while (my ($state, $info) = each(%OntapNgStates)) {
            foreach my $regex (@{$$info{prompts}}) {
                $Trace->debug(msg => "Checking $regex");
                if ($match =~ $regex) {
                    $our_state = $state;
                }
            }
            foreach my $regex (keys %{$$info{dialogue}}) {
                $Trace->trace(msg => "Checking dialogue '$regex'");
                my $send = $info->{dialogue}{$regex};
                if ($match =~ $regex) {
                    $Trace->trace(msg => "Found dialogue '$regex': Send '$send'");
                    Tharn::say($self->{connections}->{$opt{type}."_connection"}->{object}, $send, -nocr, -blind);
                    $our_state = $self->state(@_);
                }
            }
        }
    }
    
    #
    # TODO Add maintenance mode support.  We need a secondary check of ngsh's prompt in some cases to determine
    # where we actually are.  
    #
    
    #
    # If a panic was detected, bail out.
    #

    if ($our_state eq 'PANIC') {
        $Trace->debug(msg => "panic: ".$match);
        if (defined($opt{panic_string})) {
            ${$opt{panic_string}} = $match;
        }
        $Trace->fatal(msg => "The filer appears to have panicked.");
    }

    if ($our_state eq 'NO_MROOT') {
        $Trace->fatal(msg => "The filer gave up mounting mroot.");
    }

    #
    # Append any special prompts needed by the current state
    #
    
    if ($our_state){
        push (my @newprompts, 
            @{ $self->{connections}->{$opt{type}.'_connection'}->{default_prompts} },
            @{ $OntapNgStates{$our_state}{setprompts} }
        );
        $self->{connections}->{$opt{type}.'_connection'}->{object}->prompts(@newprompts);
        $Trace->debug( msg => "Prompt array now contains: ".join ",",
            ($self->{connections}->{$opt{type}.'_connection'}->{object}->prompts()),
        );
    }

    $Trace->exit(msg => "The filer's state is $our_state.");
    return $our_state;
}

1;
