#***********************************************************************
#
# NAME
#   Volume.pm - remote programmatic administration of a NetApp Volume.
#
# DESCRIPTION
#   This Perl module is meant to include only the vol commands
#
# AUTHOR
# 	dl-central-libs
#
# METHODS
#
#************************************************************************

use strict;
# The following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one).
my $LibraryLoaded;
if ($ENV{OVERRIDENG_VOLUME}) {
    my $newModule = $ENV{OVERRIDENG_VOLUME};
    delete $ENV{OVERRIDENG_VOLUME};
    print "The main Node::NG::Volume module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF);
use Data::Dumper;

our %VolumeConversion = (
			 VFSName => "volume",
			 Name => "volume",
			 SPName => "aggregate",
			 StoragePool => "aggregate",
			 MaxQuota => "size",
			);
my %Options = (
	       volume => " ",
	       force => "-f",
		    );

my %CommandDefinitions = (
			  );

my %Compatability = (
		     VFSShow => "volume_show",
		     );

our @Jobs;

use TracePkg;
my $Trace = new TracePkg;

sub SetupVolumeCommands {
    my ($self, @servers) = @_;
    $Trace->entry();

    $self->_define_commands(Definitions => \%CommandDefinitions,
			    Compatability => \%Compatability,
			    Servers => \@servers);
    $Trace->exit(msg=>"exiting with 1");
    return 1;
}

sub volume_offline {
    my $self = shift;
    my %params = @_;
    my $vserver = $params{vserver};

    $Trace->entry();

    # if we don't have a vserver, see if we can find one
    if ( !defined $vserver ) {
        $vserver = $self->volumes_vserver(%params);
    }

    # we want to know where the volume lives in case the call to set it offline fails.
    my %params_copy = %params;
    delete $params_copy{foreground} if defined $params_copy{foreground};
    my $volume = $self->volume_show(%params_copy, 'vserver' => $vserver);
    my $aggregate = $self->storage_aggregate_show('aggregate' => $$volume[0]->{'aggregate'});
    my $node = $$aggregate[0]->{nodes};

    my $volume_online = 1;
    my $offline_retries = 5;
    while ( $volume_online && $offline_retries ) {
        $Trace->trace(msg => "Offlining volume '$params{volume}' in VServer '$params{vserver}'  " .
            "volume_online = '$volume_online' and offline_retries = '$offline_retries' on node '$node'");
        eval {
            $self->volume_modify(%params, "vserver" => $vserver, "state"=>"offline");
        };
        if ( $@ ) {
            # if the volume is busy we should wait 20 seconds and retry.
            if ( $@ =~ /Error: command failed: Unable to set volume attributes for .+ Reason: Volume busy./ ) {
                my $output;
                # striped volumes are EOL - but someone could still have one.  rather than making
                # this code more complex by adding additional checks to find all the constituent
                # volume names  - we are just skipping them.
                if ( $$volume[0]->{'volume-style'} eq 'striped' ) {
                    $output = "Unable to grab debug on striped volumes at this time.";
                } else {
                    # this is somewhat bad because is doesn't use a CDEF, but it is debug output
                    $self->cli(output => \$output, cmd => "node run -node $node \"priv set diag ; wafl vref status -V $params{volume}\"");
                }
                
                
                # run a node scoped command to log the state of the volume - we do that rather than
                # potentially creating a new dblade connection on a different node which takes too
                # much time
                $Trace->warn(msg => "Encountered a volume busy error attempting to offline the " .
                    "volume - retrying the command after a 20 second sleep.  This is related to burt " .
                    "594213 which is currently being worked out.  Output from 'wafl vref " .
                    "status -V $params{volume}' is '$output'" );
                $offline_retries--; # decrement the counter so we eventually fail out.
                Tharn::sleep(20); # wait 20 seconds before we retry the offline
            } else {
                $Trace->fatal(msg => "Failure detected setting the volume offline: '$@'");
            }
        } else {
            $Trace->trace(msg => "Volume offline after '" . ( 6 - $offline_retries ) . "' attempts to set it offline.");
            $volume_online = 0;
        }
    } # the volume should be offline or we would have died
    if ( $volume_online ) {
        $Trace->fatal(msg =>
            "Volume is still online after attempting to offline it 6 times over 2 minutes.");
    }

    $Trace->exit(msg=>"exiting with 1");
    return 1;
}

# 
# This code is meant to be called to properly remove a volume from a C-Mode system.
# Essentially an analog to Volume::vol_purge()  The minor difference for C-Mode is
# that we must unmount the volume from the name space before we can offline it.
#
sub volume_purge {
    my $self = shift;
    my %params = @_;

    $Trace->entry();
    $Trace->trace(msg => Dumper(\%params));

    # We need to grab only some of the params that might be coming in to us.  
    # Otherwise we could be passing in garbage to the commands to some of the
    # subcommands.  Normally these extranious params would be striped by the 
    # cdefs, so we have to do it here ourselves
    my %params_screened;
    foreach my $param ( qw/-Idempotent -Verification -Errors -SkipINFO -Timeout -Noverbose
        -Ignore_errors -Help noprompt poke exit_code volume vserver force foreground/ ) {
        if ( defined $params{$param} ) {
            $params_screened{$param} = $params{$param};
        };
    }
    # For non-root volumes do an unmount before offline and delete.
    if (!$self->is_root_volume(volume=>$params{volume}, vserver=>$params{vserver})) {
        $Trace->trace(msg => "Unmounting volume '$params{volume}' in VServer '$params{vserver}'");
        eval {
            $self->volume_unmount(%params_screened);
        };
        if ($@) {
            if ($@ =~ /This command does not support queries/) {
                $Trace->fatal(msg => "Volume unmount: You cannot use wildcards while trying to unmount volumes: $params{volume} in $params{vserver}", result=>"SCRIPT");
            } else {
                $Trace->fatal(msg => "Volume unmount: Failure to unmount volume: $@");
            }
        }
    }
    $Trace->trace(msg => "Offlining volume '$params{volume}' in VServer '$params{vserver}'");
    $self->volume_offline(%params_screened);

    $Trace->trace(msg => "Deleting volume '$params{volume}' in VServer '$params{vserver}'");
    $self->volume_delete(%params_screened);
    
  done:
    $Trace->exit(msg=>"exiting with 1");
    return 1;
}

sub get_volumes {
    my $self = shift;
    my %params = @_;
    my @volumes;

    $Trace->entry();

    my $output;
    my %params_copy = %params;
    delete $params_copy{foreground} if defined $params_copy{foreground};
    foreach my $vol_hash ($self->volume_show(%params_copy)) {
	$Trace->debug(msg=>"found volume '$$vol_hash{volume}'");
	push(@volumes, $$vol_hash{volume});
    }
  done:
    if (wantarray) {
	$Trace->exit(msg=>"Exiting with @volumes");
	return @volumes;
    } else {
	$Trace->exit(msg=>"Exiting with $volumes[0]");
	return $volumes[0];
    }
}

sub volume_mount_point {
    my $self = shift;
    my %params = @_;
    my $vol = $params{volume};
    my $volref;
    my $vol_path = 0;
    my @showlist; 
    
    $Trace->entry();

    eval {
        $volref = $self->volume_show(volume=>$vol);
    };

    if ($@) {
        if ( $self->is_root_volume(volume=>$vol) ) {
            $vol_path = "/";
        } else {
            $vol_path = "/".$vol;
        }
    } else {
        $vol_path = $$volref[0]{"junction-path"};
    }
    
    $Trace->exit(msg=>"exiting with vol_path = '$vol_path' for volume '$vol'");
    return $vol_path; 
    
}

sub is_root_volume {
# temporary until volume show contains junction path
# return 1 if the volume is a root volume
# return 0 otherwise
    my $self = shift;
    my %params = @_;
    my $vol = $params{volume};
    my $item;
    my $found = 0;
    my @vslist;

    $Trace->entry();

    my $rc = eval {
        @vslist = $self->vserver_show();
    };

    for $item ( @vslist ) {
        # Check if the volume is a root volume of the vserver provided.
        if (defined $params{vserver}) {
            $found = 1 if ( $item->{rootvolume} eq $vol && $item->{vserver} eq $params{vserver} );
        } else {
            $found = 1 if ( $item->{rootvolume} eq $vol );
        }
        last if ( $found eq 1 );
    }

    $Trace->exit(msg=>"exiting with found = $found");
    return $found; 

}

sub create_volume_from_ip {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => HASHREF } ],
					 {
					     ipaddress => { type => SCALAR },
					     "junction-path" => { type => SCALAR, optional => 1},
					     aggregate => { type => SCALAR, optional => 1 },
					     user => { type => SCALAR, optional => 1 },
					     group => { type => SCALAR, optional => 1 },
					     volume => { type => SCALAR, optional => 1 },
					     parent => { type => SCALAR, optional => 1 },
					     parent_path => { type => SCALAR, optional => 1 },
					     'unix-permissions' => {type => SCALAR, optional => 1},
					 }
					 );
    $Trace->entry(msg=>join(",", @_));

    # find a virtual server owning the provided IP
    my $out = $self->network_interfaces_show(address=>$params{ipaddress});
    my $vserver = $$out[0]{server};

    # find the parent volume and mount
    my $parent_volume = $params{parent};
    if (!$parent_volume) {
	$out = $self->vserver_show(vserver=>$vserver);
	$parent_volume = $$out[0]{rootvolume};
    }

    my $parent_path = $params{parent_path} || $$out[0]{"junction-path"};

    # find an aggregate that has at least 100MB free
    my $aggregate = $params{aggregate};
    if (!$aggregate) {
	$out = $self->storage_aggregate_show(size=>">100");
	$aggregate = $$out[0]{aggregate};
    }

    # find the volume name
    my $clean_ip = $params{ipaddress};
    $clean_ip =~ s/\./_/g;
    my $volume = $params{volume} || "create_vfs_from_$clean_ip";

    # find the junction-path
    my $junction_path = $params{'junction-path'} || $params{parent_dir}."/$volume";

    my $unix_permissions = $params{'unix-permissions'};

    my %volume = (vserver=>$vserver, 
		  volume=>$volume, 
		  aggregate=>$aggregate,
		  parent=>$parent_volume,
		  "junction-path"=>$junction_path,
		  'unix-permissions'=>"$unix_permissions",
		  );
    $volume{user} = $params{user} if ($params{user});
    $volume{group} = $params{group} if ($params{group});

    $self->volume_new(-Verification=>undef, %volume);
    $Trace->exit(msg=>"volume=$volume, path=$junction_path");
    return ($vserver, $volume, $junction_path);
}

sub _volop_verify {
    my $self = shift;
    my %params = @_;
    my $rc;

    $Trace->entry();
    if ($params{foreground} =~ /false/i) {
        my $job;
        if ($params{output_in} =~ /\[Job (\d+)]/) {
            $job = $1;
            $Trace->trace(msg=>"Add $job to the jobs list");
            push(@Jobs, $job);
        } else {
            $Trace->fatal(msg=>"Error looking up job id in '$params{output_in}'\n");
        }
        $rc = $job;
    } else {
        if ($params{output_in} !~ /(Job succeeded|Validation succeeded)/) {
            $Trace->fatal(msg=>"Volume operation did not succeed: '$params{output_in}'\n");
        }
        $rc = 1;
    }
    
    $Trace->exit(msg=>"return $rc");
    return $rc;
}

sub wait_on_jobs {
    my $self = shift;
    my %params = (
                  sleep => 10,
                  timeout => 600, # 10 minutes
                  @_);
    $Trace->entry(msg=>"wait up to $params{timeout} seconds");

    # wait for the jobs to complete
    
    my $timeout = time + $params{timeout}; 

    my %jobs;
#    my @errors;

    #    Initial=0       "Initializing"
    #    Queued          "Queued"
    #    Running         "Running"
    #    Waiting         "Waiting For Another Job"
    #    Pausing         "Entering Paused State"
    #    Paused          "Paused"
    #    Quitting        "Entering Quit State"
    #    Success         "Succeeded"
    #    Failure         "Failed"
    #    Reschedule      "Forcing Reschedule"
    #    Error           "Internal Error"
    #    Quit            "Quit"
    #    Dead            "Died"
    #    Unknown         "Unknown" 
    #Anything at or after Success is done.  This is the best thing to check for.
    my @final_states = qw(Success Failure Reschedule Error Quit Dead); # Unknown);

    $Trace->debug(msg=>"Jobs=".Dumper(\@Jobs));
    my $count=1;
    while (time < $timeout) {
        my @jobs = @Jobs;
        foreach my $job (@jobs) {
            $Trace->debug(msg=>"check on $job\n");
            my $job_hash;
            my $job_retries=5;
            my $retry=0;
            my $error;
           
            do {
                eval {
                    $job_hash = $self->job_show(id=>$job);
                    $job_hash = $job_hash->[0] if ($job_hash);
                    $jobs{$job} = $job_hash;
                };
                $error = $@;
                if (!$@ && $retry) {
                    $Trace->comment(msg=>"burt220651: Found the job on retry $retry");
                }
                if ($error) {
                    $Trace->trace(msg=>"job_show $job failed with '$error'");
                }
                sleep(5) if ($error && $retry++ < $job_retries); # only sleep if there was an error
            } until (!$error || $retry > $job_retries);

            if ($error =~ /There are no entries matching your query/) {
                # we somehow missed the job, so look it up in EMS
                $Trace->comment(msg=>"Could not lookup job $job, look in EMS");
                eval {
                    my $event = $self->event_log_show(event=>"*$job*", 
                                                      -Timeout=>180, # give it up to 3 minutes since EMS is large
                                                      messagename=>"mgmtgwd.jobmgr.jobcomplete");
                    $event = $event->[0];
                    # make up a job record since we could not find it
                    $jobs{$job} = {
                                   jobid => $job,
                                   completion => $event->{event},
                                   state => $event->{event},
                                  };
                    foreach my $state (@final_states) {
                        $jobs{$job}{state} = $state if (grep($state, $event->{event}));
                    }
                };
                if ($@) {
                    $Trace->fatal(msg=>"Could not locate job $job in jobmgr or EMS");
                }
                
            } elsif ($@) {
                $Trace->fatal(msg=>"An error occurred looking up the job: $@");
            }
            if (grep($_ =~ /$job_hash->{state}/i, @final_states)) {
                @Jobs = grep($_ ne $job, @Jobs);
            }
        }

        last if (scalar @Jobs == 0);
        
        # simple mechanism to sleep for 5 seconds the first 6 attempts (30 seconds)
        # and then back off to 10 seconds. This could be enhanced if we have really
        # large jobs but maybe poking it every 10 seconds isn't that bad
        my $sleep = 5;
        $sleep=10 if ($count++ > 6);
        sleep $sleep;
    }
    if (scalar @Jobs) {
        # assume a timeout occurred
        print Dumper \@Jobs;
        print Dumper \%jobs;
        $Trace->fatal(msg=>"Timed out after $params{timeout} seconds waiting on the op to occur");
    }
#    if (scalar @errors) {
#        $Trace->fatal(msg=>"An error occurred on some of the jobs ".Dumper \@errors);
#    }
    
    $Trace->exit();
    return \%jobs;
}

# This code is really error prone.  If you are passing in data that results in multiple volumes
# matching the query, you are going to get the first volume matched which may not be what you want.
sub volumes_vserver {
    my $self = shift;
    my %params = @_;
    my $vol = $params{volume};
    my $vserver = 0;
   
    $Trace->entry();

#    my $rc = eval {
#        $vol = $self->volume_show(%params);
#    };

    my %params_copy = %params;
    delete $params_copy{foreground} if defined $params_copy{foreground};
    $vol = $self->volume_show(%params_copy);


    $vserver = $$vol[0]{vserver};

    # warn if we match more than 1 volume
    if ( scalar @$vol > 1) {
        $Trace->warn(msg => "Found more than 1 volume that matched the params you searched " .
            "with to find the vserver of. This code returns the VServer on the first volume " .
            "it matches. You may be working with the incorrect volume now.  Consider rewriting " .
            "your code to be more specific with the params it is passing to volumes_vserver()\n" .
            "Volumes matched: " . Dumper(\$vol) . "\nParams: " . Dumper(\%params)
        );
    }

    $Trace->exit(msg=>"exiting with vserver = $vserver");
    return $vserver;

}


sub striped_vol_create_size_includes_mdv {
    #Identifies Software versions that calculate vol_size /(mdv + dv_count)
    #Checks the node that is the current connection
    my $self = shift;
    my @old_calc_versions = qw(Tricky.0 Tricky.1);
    my $release;
    my $mdv_calc;
    $Trace->entry();
    $release = $self->get_release(format => 'cooked');
    if (grep(/^$release$/,@old_calc_versions)) {
	$mdv_calc = 1;
    } 
    else {
	$mdv_calc = 0;
    }
    $Trace->exit();
    return $mdv_calc;

}    

sub find_min_max_volume_size {
    # Determine various sizes that can be used for volume creation.
    #
    # find_min_max_volume_size(aggr_name, sizetype)
    #
    # sizetype can be one of minperm, maxperm, aggrsize, maxavail
    # minperm size = (1228.8MB * (#Constituent_Aggregates + 1)) 
    # maxperm size = (16TB * (#Constituent_Aggregates + 1))
    # aggrsize = total size of aggregate
    # maxavail  = available size of aggregate
    # If the aggregate is type=flex, return the defaults for minperm/maxperm

    my $outString;
    my ($self, $aggr_name, $sizetype) = @_;
    my $size;
    my $minsize;
    my $dv_count;
    my $mdv_divisor;
    my $release = $self->get_release(format => 'cooked');
    my $min_vol_size;
    if ($release !~ /tricky/i && $self->has_change(change=>"min_dv_size_100mb")) {
        $min_vol_size = 100 #100MB in MB
    } else {
        $min_vol_size = 1228.8; #1.2GB in MB
    }
    my $size_adjust;

    $Trace->entry();
    $Trace->debug(msg=>"Determining Allowed Volume Size");
    if (!defined $sizetype || $sizetype eq "") {
        $Trace->comment(msg=>"Volume size type not defined. Defaulting to minimum permitted");
        $sizetype = "minperm";
    }

    my $outAggrArray = $self->storage_aggregate_show(
                                                       -SkipINFO => 1,
                                                       aggregate => $aggr_name,
                                                       output => \$outString );
    my $nodes = $$outAggrArray[0]{nodes};

    if (ref($nodes) ne "ARRAY"){
        # If it's not an array, turn it into one
        my @a = split(/,/, $nodes);
        $nodes = \@a;
    }

    #Allow for old versions
    #We are using these vars to change sizes/values based upon software version
    if ($self->striped_vol_create_size_includes_mdv()){	
        $Trace->debug(msg=>"Including the mdv in space calculations");
        $dv_count = scalar(@$nodes)+1; # Account for the mdv in volume size 
        $mdv_divisor = 2; #used to cut the available aggr space in half since mdv lives with a dv
        $size_adjust = 1; #Don't convert aggregate size 
    } 
    else {
        $Trace->debug(msg=>"Excluding the mdv in space calculations");
        $dv_count = scalar(@$nodes); # Only Account for DV's in vol size
        $mdv_divisor = 1; #mdv has guarantee of none
        $size_adjust = 2**20; # Used to Convert Size to MB since newer versions express sizes in Bytes
    }

    if ($sizetype eq "minperm") {
        if ($$outAggrArray[0]{voltype} eq "striped") {
            $size = ($dv_count*$min_vol_size);
        } else {
            $size = "20";
        }
    } elsif ($sizetype eq "maxperm") {
        if ($$outAggrArray[0]{voltype} eq "striped") {
            $size = ($dv_count*(15069184/$mdv_divisor));
        } else {
            $size = "15069184";
        }
    } elsif ($sizetype eq "aggrsize") {
        $size = ($$outAggrArray[0]{size} / $size_adjust);
    } elsif ($sizetype eq "maxavail") {
        if ($$outAggrArray[0]{voltype} eq "striped") {
            my $outAggrMemberArray = $self->storage_aggregate_member_show(
                                      -SkipINFO => 1,
                                      aggregate => $aggr_name."*" );
            foreach my $memberAggr (@$outAggrMemberArray) {
                next if ($$memberAggr{aggregate} !~ /${aggr_name}_[0-9]{3}/);
                #Filter out aggregates that may have a similar name
                # Find size of the smallest stripe
		if ($release !~ /tricky/i && $self->has_change(change=>"min_dv_size_100mb")) {
                    if ($$memberAggr{availsize} < 100) {
                        $Trace->fatal(msg=>"Insufficient space on member aggregate: $$memberAggr{aggregate} ($$memberAggr{availsize}MB Available)");
                    }
		} else {
                    if ($$memberAggr{availsize} < 1200) {
                        $Trace->fatal(msg=>"Insufficient space on member aggregate: $$memberAggr{aggregate} ($$memberAggr{availsize}MB Available)");
                    }
		}
                if ($$memberAggr{availsize} < $minsize || $minsize == undef) {
                    $minsize = $$memberAggr{availsize};
                    $minsize = $minsize / $size_adjust; #Convert size to MB if necessary
                }
            }
            # adjust size based on version
            # In older releases we need to cut min stripe size in half
            # because we don't know where the MDV + DV will go.
            # Newer releases don't set space guarantee for MDV 
            # Need to reduce max size a bit due to burt201842.
            $size = (($dv_count)*(($minsize/$mdv_divisor)-10));
            
        } else {
            # Need to reduce max size a bit due to burt201842.
            # Adjust size to MB if newer build
            $size = (($$outAggrArray[0]{availsize} / $size_adjust)-5);
        }
    } else {
    $Trace->fatal(msg=>"Invalid Sizetype");
    }
    $Trace->debug(msg=>"Returning Volume Size : $size");
    $Trace->exit();
    return $size."MB";

}


sub volume_size {
# equivalent of classic VolCmds/volSize
# returns volume size and used size in blocks
    my $self = shift;
    my %params = @_;
    my $vol = $params{vol};
    my $blocksize = 4096;
    my $size = 0;
    my $used = 0;

    $Trace->entry(msg=>"volume_size: vol is $vol");

    my $rc = eval {
        $vol = $self->volume_show(volume=>$vol);
    };

    $Trace->debug(msg=>"volumes_size vol is $vol size $$vol[0]{size} used $$vol[0]{used}");
    $size = convert_size($$vol[0]{size});
    $used = convert_size($$vol[0]{used});

    # convert to blocks
    $size = $size / 4096;
    $used = $used / 4096;

    # buffer so max_inodes test won't fail
    $size = $size * 0.87;

    $Trace->exit(msg=>"exiting with size $size and used $used");
    return ($blocksize,$size,$used);

}

# a replacement for volume_new to additionally set values for keys which are
# not supported solely in volume_new()
sub volume_new_and_modify {
    my ($self, %params) = @_;

    $Trace->entry();
    
    my $volume = $self->volume_new(%params);

    # perform and modifications to the volume
    if ($params{files}) {
	eval {
	    $volume = $self->volume_modify(volume=>$params{volume},
					   vserver=>$params{vserver},
					   files=>$params{files});
	};
	if ($@) {
	    if ($@ =~ /limit can only be increased/) {
		$Trace->comment(msg=>"The current volume '$params{volume}' already has a higher 'files' value so it will be used instead of '$params{files}'");
	    } else {
		die $@;
	    }
	}
    }
    $Trace->exit();
    $volume;
}


1;
