# $Id: //depot/prod/test/Rsierranevada/lib/OntapSSH.pm#2 $
#
# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.
#
## @summary Library of functions to configure and test Ontap SSH administration
## @author Dave Anderson (davea@netapp.com)
## @status public
## @pod here


package OntapSSH;
use strict;
use misc;
use boot qw( reboot );
require "tharnit_lib.pl";
use Tharn;
use Data::Dumper;

use vars qw(@ISA @EXPORT $AUTOLOAD $FILER $FILER_CONN $CLIENT $dbg);

use base qw(Exporter);
@ISA = ('Exporter');

BEGIN {
our @EXPORT_OK = qw(
    get_config
    backup_config
    change_keysize
    check_mount_point
    compare_config
    create_ssh_sessions
    get_client
    get_client_name
    get_debug_module_verbose
    get_debug_ssh_verbose
    get_filer_ipv6
    get_hostname
    get_password
    get_ssh1_enable
    get_ssh1_hostkeysize
    get_ssh1_serverkeysize
    get_ssh2_enable
    get_ssh2_hostkeysize
    get_ssh_access
    get_ssh_connect_auth_method
    get_ssh_connect_protocol
    get_ssh_enable
    get_ssh_idle_timeout
    get_ssh_passphrase
    get_ssh_passwd_auth_enable
    get_ssh_port
    get_ssh_pubkey_auth_enable
    get_stale_opt
    make_sure_ssh_connect_fails_and_log_result
    make_sure_ssh_connect_succeeds_and_log_result
    restore_config
    run_command_over_ssh
    run_command_after_login_to_filer_using_ssh
    run_command_debug_enable
    set_debug_module_verbose
    set_debug_ssh_verbose
    set_filer_ipv6
    set_password
    set_ssh1_enable
    set_ssh1_hostkeysize
    set_ssh1_serverkeysize
    set_ssh2_enable
    set_ssh2_hostkeysize
    set_ssh_access
    set_ssh_connect_auth_method
    set_ssh_connect_protocol
    set_ssh_enable
    set_ssh_idle_timeout
    set_ssh_passphrase
    set_ssh_passwd_auth_enable
    set_ssh_port
    set_ssh_pubkey_auth_enable
    set_ssh_alternate_username
    set_ssh_alternate_password
    set_ssh_alternate_ip
    set_ssh_sessions
    ssh_can_connect
    ssh_client_config_backup
    ssh_client_config_restore
    ssh_client_rm_known_hosts
    ssh_keygen
    verify_max_ssh_sessions
);
}


=head1 NAME

OntapSSH - Functions for setting SSH parameters on filer, including key
generation, and for testing SSH connectivity.

=head1 SYNOPSIS

Note: C<$CLIENT> B<MUST> be a UNIX machine.

See the example scripts linked in the "SEE ALSO" section beginning with
"_sample_*" for example test scripts and basic instructions on writing tests.
This document is primarily intended for library maintainers.

Note that this is an abstract base class. Only the three methods below should
be called directly, from a wrapper test script, all others should be called
from one of the subclasses, OntapSSH::Options.pm or OntapSSH::SecureAdmin.pm,
and they should be called from a subtest. The methods backup_config() and
restore_config() are only used in the wrapper script to preserve the
configuration on the filer prior to invoking subtests, and to restore the
configuration after subtests are complete.

    use OntapSSH qw( get_config backup_config restore_config );

    $cfg = get_config($FILER, $CLIENT);
    $rv = $cfg->backup_config();
    $rv = $cfg->restore_config();



=head1 DESCRIPTION

Note: C<$CLIENT> B<MUST> be a UNIX machine.

This module defines routines used to modify SSH-related settings on the filer
and validate whether SSH may be used to connect to the filer using various
authentication schemes and protocol versions.

This module is still in beta status, testers welcome, but please be warned
that the behavior and interface may change. Please see the links at the bottom
for sample scripts that are to be used to test this library.

Note that this is an abstract base class. Only the three methods below should
be called directly, all others should be called from one of the subclasses,
OntapSSH::Options.pm or OntapSSH::SecureAdmin.pm. The methods backup_config()
and restore_config() are only used in the wrapper script to preserve the
configuration on the filer prior to invoking subtests, and to restore the
configuration after subtests are complete.

This form is used in the wrapper script:

    use OntapSSH qw( get_config backup_config restore_config );

    $cfg = get_config($FILER, $CLIENT);
    $rv = $cfg->backup_config();
    $rv = $cfg->restore_config();


This form is used in the subtest scripts:

    use OntapSSH::SecureAdmin qw(
        get_config
        backup_config
        .
        .
        .
        get_debug_ssh_verbose
        get_debug_module_verbose
    );


This is another form used in the subtest scripts:

    use OntapSSH::Options qw(
        get_config
        backup_config
        .
        .
        .
        get_debug_ssh_verbose
        get_debug_module_verbose
    );

    # For working examples,  please see the sample scripts mentioned in the
    # "SEE-ALSO" section.

Some cautions about using this module:

Do not directly access the local host.  Any access to the local host should be
done through a $CLIENT connection.  This will allow your script to run with
different clients.  The following are examples of local host access to avoid.

            if ( -e /tmp/myTmpFile ) { ...  };
            system ( "mkdir -p ~/.ssh/root" );

Avoid creating 2 objects of this class with the same filer and client pair
within your test wrapper or within your subtest.  This is due to the temp file
that is used to keep track of the keys-changed condition.  The name of this
temp file will be the same for all objects that are created with the same
filer and client pair.  This may confuse the object when it needs to use this
file to determine if the keys are changed.  This limitation is due to the fact
that we need to have this temp file shared between the test wrapper and the
subtests.  If you have to created 2 objects with the same filer and client
pair within a subtest, you have to be very careful on which object creates the
temp file (calling keys_changed() with the $key_type parameter), and which
object uses this temp file.



=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success, and
false on failure.

=for html
<hr><b>Accessor Methods:</b><br>
The accessor methods below(most of the get_/set_ functions) only change the
specified object attribute and mark the object as having "stale" keys. A call
to the base class <code>change_keysize()</code> method will be required to
propogate option changes to the filer and resetthe "stale" flag. If a bogus
value is entered in an object attribute, a flag will be set named <code>&lt;
attribute name&gt;_bogus</code> for future use with negative test cases.
Subsequent entry of a valid attribute value will set the <code>&lt;attribute
name&gt;_bogus</code> flag to "".<br><hr>

=cut


my $tarfile;
my $umount;
my $filesystem_errors = ("error|cannot|permission denied");

logcomment("Initializing OntapSSH.pm");

{
    # Private Class attribute
    my $_keys_changed = 0;

    # The purpose of this function is to keep track, from one subtest to
    # another, whether the SSH keys on the filer have been changed. If they
    # have, we will need to restore them and reboot the filer after all tests
    # are done. If any value that can be interpreted as "TRUE" is passed to
    # this method, a file will be touched to indicate that the keys have
    # changed. Invoking this method without agrguments will return "TRUE" (1)
    # if the keys have been changed or "FALSE" (0) if they have not. So far,
    # this function is only used by change_keysize() and restore_config().

    sub keys_changed {
        my ($self) = shift;
        my $client = $self->{'_client'};
        my $c = connect("$client#telnet");

        logcomment("In keys_changed sub, \@_ is:\n" . Dumper(@_) . "\nCaller is: " . caller());

        my $filename = $self->{'_keys_changed_file'};
        my $output;
        if ( @_ ) {
            (undef,undef,$output) = say($c,"touch $filename");
        }

        else {
            if ( $self->_exists_file_or_dir($c,$filename) ) {
                logcomment("keys_changed() will return '1' because $filename exists");
                return 1;
            }

            else {
                logcomment("keys_changed() will return '0' because $filename does not exist.");
                return 0;
            }
        }
    }
}


=head2 get_config($FILER, $CLIENT)

Creates a configuration object in memory.

This method takes regular Tharn C<$FILER> and C<$CLIENT> objects, and returns a
configuration object containing details of how C<$FILER> and C<$CLIENT> are
configured.  Without an object created in this way, most of the other methods
in this module will not work properly or at all. Below is an example of a
typical configuration object as printed by Data::Dumper:

 $VAR1 = bless( {
                  '_ssh2_hostkeysize' => '768',
                  '_ssh_passwd_auth_enable' => 'on',
                  '_ssh1_hostkeysize' => '768',
                  '_ssh_enable' => 'on',
                  '_ssh1_enable' => 'on',
                  '_client_name' => 'grumpy',
                  '_keys_changed_file' => '/tmp/keyschanged_mgt-u22_grumpy',
                  '_ssh_port' => '22',
                  '_prompt' => [
                                 'mgt-u6\\*?> '
                               ],
                  '_password' => 'sundance',
                  '_filer' => 'mgt-u6; hosttype=filer; default_password=sundance; conn console { conntype=pm_cons ; connhost=mgt-u6 } ',
                  '_client' => 'grumpy; hosttype=unix-solaris; prompt_norm="grumpy.* " ; conn console { conntype=telnet ; connhost=grumpy } conn telnet { username=root; password=tryme; timeout=200 } ',
                  '_stale_key' => 0,
                  '_ssh_access' => '*',
                  '_ssh_pubkey_auth_enable' => 'off',
                  '_ssh2_enable' => 'on',
                  '_filer_ipv6' => 'ffff:ffff:ffff:ffff:ffff:ffff',
                  '_ssh_idle_timeout' => '600',
                  '_hostname' => 'mgt-u6',
                  '_ssh1_serverkeysize' => '512',
                  '_stale_opt' => 0
                }, 'OntapSSH' );


B<Usage:> C<$cfg = get_config($FILER, $CLIENT);>

B<Returns:> C<$cfg> configuration object

=cut

my %_name_map = (
    'ssh.enable'                =>  '_ssh_enable',
    'ssh.access'                =>  '_ssh_access',
    'ssh1.enable'               =>  '_ssh1_enable',
    'ssh2.enable'               =>  '_ssh2_enable',
    'ssh1.serverkeysize'        =>  '_ssh1_serverkeysize',
    'ssh1.hostkeysize'          =>  '_ssh1_hostkeysize',
    'ssh2.hostkeysize'          =>  '_ssh2_hostkeysize',
    'ssh.idle.timeout'          =>  '_ssh_idle_timeout',
    'ssh.passwd_auth.enable'    =>  '_ssh_passwd_auth_enable',
    'ssh.pubkey_auth.enable'    =>  '_ssh_pubkey_auth_enable',
    'ssh.port'                  =>  '_ssh_port',
);

sub get_config () {

    my $_class = shift;
    my $_filer = shift;
    my $_client = shift;

    logcomment("In get_config()");

    my $_host_obj   =   host($_filer);
    my $_hostname   =   $_host_obj->hostname();
    my $_password   =   $_host_obj->default_password();
    my $_prompt     =   $_host_obj->prompt_norm();

    my $_client_obj =   host($_client);
    my $_client_name =   $_client_obj->hostname();

    my $self = {
        _filer      =>  $_filer,
        _hostname   =>  $_hostname,
        _password   =>  $_password,
        _prompt     =>  $_prompt,
        _client     =>  $_client,
        _client_name =>  $_client_name,
        _debug_ssh_verbose => 0,
        _debug_module_verbose => 0,
        _filer_ipv6 => undef,
        _ssh_alternate_username   =>  undef,
        _ssh_alternate_password   =>  undef,
        _ssh_alternate_ip         =>  undef,
        _ssh_sessions             =>  undef,
    };

    my %_cfg = ();      # Temporary hash to store filer config details

    # Get SSH options via registry
    #
    my $regvals = cli($self->{_filer}, "priv set advanced;registry walk options.secureadmin");
    my @a = split(/\n/, $regvals);

    # Map registry names to Perl-safe internal names
    #
    foreach (@a) {

        if ( $_ =~ m/^options/ ) {

            s/options.secureadmin.//;
            s/\r//;
            my @v = split(/\=/);
            $_cfg{$v[0]} = $v[1];   # Begin loading hash
        }
    }

    # Get SSH options via options command
    #
    my $optvals = cli($self->{_filer}, "priv set advanced;options ssh");

    @a = split(/\n/, $optvals);

    foreach (@a) {

        if ( $_ =~ m/^ssh/ ) {

            s/ +/ /gs;
            my @v = split(/ /);
            $_cfg{$v[0]} = $v[1];   # Begin loading hash
        }
    }


    # Load info from temporary hash into $self
    #
    foreach my $key (keys %_cfg) {
        # We only use keys that are predefined in %_name_map.
        #
        $self->{$_name_map{$key}} = $_cfg{$key} if (defined $_name_map{$key});
    }


    # Flags to keep track of whether config matches state of filer.
    # Note that these are ignored by restore_config(), since it's not
    # particularly expensive to just restore the config after finishing up
    # all subtests.

    $self->{'_stale_opt'} = 0;
    $self->{'_stale_key'} = 0;

    # The '_umount' element is used to keep track of the need to umount
    # the filer from /t/$filer/root.  It will be set to "sim_umount
    # /t/$filer/root" if a umount is needed during clean up.  The clean up
    # operation will reset this to undef.
    #
    $self->{'_umount'} = undef;

    # Make an unique number for backup filename prefix:
    #
    my $uniqueID = time / -1000000;
    $uniqueID **= 2;

    # Keep track of the tar filename used for saving and restoring the filer
    # ssh key info.  The name of this file is $uniqueID.FILER.$filer_name.tar.
    #
    $self->{'_filer_keys_tar_file'} = "/tmp/$uniqueID.FILER."
        . $self->{'_hostname'}
        . ".tar";


    # Keep track of the tar filename used for saving and restoring the client
    # ssh info.  The name of this file is $uniqueID.CLIENT.$client_name.tar.
    # We will use the upper case version of the client name.  The lower case
    # of the client name is usually the client prompt. Using the lower case
    # client name makes "say" think the filename is a prompt.  This may cause
    # commands not to work correctly.
    #
    my $uc_name = uc($self->{'_client_name'});
    if ($uc_name eq $self->{'_client_name'}) {
        # if converting to upper case does not make the client name different,
        # insert a "_" before the last character of the name.
        #
        $uc_name = substr($uc_name,0,length($uc_name)-1) . "_"
            . substr($uc_name,length($uc_name)-1,1);
    }
    $self->{'_client_ssh_tar_file'} = "/tmp/$uniqueID.CLIENT.$uc_name.tar";

    # A temp file is used to keep track of the status of keys being changed.
    # This status is shared by both the wrapper and the subtests.  Because of
    # this sharing, we can not simply use a class varible to store the status.
    # The name of this file must be the same for the object created by the
    # test wrapper and the objects created by the subtests.  This filename is
    # keys_changed_FILER.$filer_name.CLIENT.$client_name.
    #
    $self->{'_keys_changed_file'} = "/tmp/keys_changed.FILER."
        . $self->{'_hostname'}
        . ".CLIENT.$uc_name";


    bless $self, ref($_class) || $_class;

    # Output config backup info to log

    logcomment("Configuration Retrieved:\n" . Dumper($self));
    return $self;

}


=head2 backup_config()

Backs up Filer settings, including existing SSH Keys on filer and root user
.ssh directory on filer. These key files and other files are stored on the
client in C</tmp/<uniquenameE<gt>.tar>.  This client is the client that is
used to construct the C<$cfg> object.  Root user .ssh folder on the filer is
also cleared.

C<backup_config()> should only be called once, ideally from a top-level thpl
script which calls a series of subtests. The reason for this limitation is that
there is no mechanism to guarantee that the first "backup" of the filer state
will be the one restored. After the subtests are run a call to
C<$cfg-E<gt>restore_config()> will restore the filer to the state it was in
prior to running the thpl script. It is very important that no changes are
made to the C<$cfg> object between invoking C<backup_config()> and
C<restore_config()>. If you wish to use C<backup_config()> repeatedly in a
script make sure you are using a different C<$cfg> object than was used for
the first invocation of C<backup_config()> or you will lose track of the
original state of the filer -- though if you have
C<$cfg-E<gt>get_debug_module_verbose()> set TRUE you may be able to figure it
out from the log files.

B<Usage:> C<$cfg-E<gt>backup_config()>

B<Returns:> C<$cfg> object, with location of tarred up keys added to object.

I<NOTE: It is strongly advised that the C<$cfg> object used for
C<backup_config()> be given a unique and different name from the C<$cfg>
object used in subtests so that you do not confuse them.>

=cut

sub backup_config () {

    my $self = shift;
    my $client = $self->{'_client'};
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    my $c = connect("$client#telnet");
    logcomment("In backup_config() on client: $self->{'_client_name'}.");
    $dbg ? logcomment(Dumper($self)) : undef;

    # Who is the user?  Just for debugging info.
    my $output;
    $dbg ? (undef,undef,$output) = say($c,"id") : undef;
    $dbg ? logcomment("Who am I?: $output") : undef;

    # check mount point.
    my $mount_ok = $self->check_mount_point($c);
    if ( $mount_ok <= 0 ) {  # The mount point is bad.
        logcomment("Failed!!! : not able to sim_mount $self->{'_hostname'}");

        # sim_umount to clean up if needed.
        $self->_do_sim_umount($c);

        # Mount point is bad.  Raise a FATAL error and exit.
        my $e = NATE::Result::Fatal->new("Fatal error, mount point /t/$self->{'_hostname'}/root not usable.");
        $Log->log($e);
        return $self;
    }

    # Back up SSH keys
    $tarfile = $self->{'_filer_keys_tar_file'};

    my $cmd = "/usr/local/bin/sudo tar cvf $tarfile "
        . "/t/$self->{'_hostname'}/root/etc/sshd/s*[yb] "
        . "/t/$self->{'_hostname'}/root/etc/sshd/root/.ssh";
    $dbg ? logcomment($cmd) : undef;
    (undef,undef,$output) = say($c,$cmd);
    $dbg ? logcomment("The output of $cmd is:\n $output") : undef;

    $cmd = "/usr/local/bin/sudo rm -Rf /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/*";
    $dbg ? logcomment($cmd) : undef;
    (undef,undef,$output) = say($c,$cmd);
    $dbg ? logcomment("The output of $cmd is:\n $output") : undef;

    # sim_umount to clean up if needed.
    $self->_do_sim_umount($c);

    logcomment("ssh keys backed up to $tarfile");

    # Output config backup info to log

    $dbg ? logcomment(Dumper($self)): undef;
    logcomment("backup_config() complete.");

    return $self;
}

=head2 change_keysize()

Changes the SSH keys used on the filer represented by the passed configuration
object, setting them to sizes specified by other methods if applicable.
Normally you will not need to call this method directly, since it is called by
the C<update_filer_settings()> method in either of this module's subclasses.

The SSH keys stored on the filer will be regenerated using the key sizes set
with C<set_ssh1_hostkeysize()>, C<set_ssh1_serverkeysize()>, and
C<set_ssh2_hostkeysize()>. If you have not called any of these methods,
C<change_keysize()> will be a nonop(unless debugging is enabled, in which case
the one thing it will do is log the fact that it was a nonop). We don't want
to change the keys on the filer unless explicitly told to do so, since
restoring the keys properly involves rebooting the filer after all tests are
complete. See the documentation on those three methods for techniques to
regenerate keys without altering their size.

The config object is updated to reflect that the keysizes stored in it are no
longer "stale", implying that they are in sync with the filer. Additionally,
while the three C<set...keysize()> methods mentioned above do their own
validation to ensure that either the sizes set are within acceptable limits or
a flag is set to indicate that they are bogus settings(for future negative
test case use), this method performs one additional step and verifies that the
difference between the SSH1 host & server key sizes is at least 128. If the
difference is less than 128, then both values are flagged as being bogus.

B<Usage:> C<$cfg-E<gt>change_keysize()>

B<Returns:> The configuration object used to invoke this method, with changes
made to the object as described above.

=cut

sub change_keysize () {

    my $self = shift;
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    $dbg ? logcomment("In change_keysize()\n" . Dumper($self)) : undef;

    if( $self->{'_stale_key'} ) {

        # Record fact that keys have changed so we restore them later
        #
        $self->keys_changed(1);

        # Detect insufficient difference between keysizes, flag as
        # negative test case if < 128

        if ( abs( $self->{'_ssh1_hostkeysize'} - $self->{'_ssh1_serverkeysize'} ) < 128 ) {

            $dbg ? logcomment("Setting flag to indicate bogus setting for _ssh1_hostkeysize and _ssh1_serverkeysize (negative testcase, should differ by 128 bits)") : undef;

            $self->{'_ssh1_hostkeysize_bogus'} = 1;
            $self->{'_ssh1_serverkeysize_bogus'} =1;

            $dbg ? logcomment("\$self is now: " . Dumper($self)) : undef;
        }

        cli($self->{'_filer'}, "secureadmin disable all");

        my $c = connect($self->{'_filer'} . "#telnet");

        # When "secureadmin setup" runs, it may return one of two prompts,
        # depending on whether or not "secureadmin setup" has already
        # been run.  If it has not been run before, the first prompt is
        # "Please enter the size of host key for ssh1.* protocol [768] :".
        # If "secureadmin setup" was run successfully before, the first prompt
        # is "Do you want to proceed? [.*]".  We need to handle both of these
        # prompts when running "secureadmin setup".
        #
        my $proceed = "Do you want to proceed\\? \\[.*\\]";
        my $host_key =
          "Please enter the size of host key for ssh1\\.x protocol \\[768\\] :";
        my $prompt;
        my $before;
        (undef, undef, $before, $prompt) =
            say($c, "secureadmin setup -f ssh", "-prompt",
                [$proceed, $host_key]);  # wait for either prompt.
        if ($prompt =~ /$proceed/) {
            say($c, "y", "-noprompt", "-wait", "768");
        }
        say($c, $self->{'_ssh1_hostkeysize'}, "-noprompt", "-wait", "512");
        say($c, $self->{'_ssh1_serverkeysize'}, "-noprompt", "-wait", "768");
        say($c, $self->{'_ssh2_hostkeysize'}, "-noprompt", "-wait", "yes");

        # The last prompt for tt and older is "Setup is complete.".
        # The last prompt for GB and IC is "automatically.".
        #
        say($c, "y", "automatically.|Setup is complete.");

        # "secureadmin disable all" turned off "ssh1.enable" and
        # "secureadmin setup -f ssh" did not turn this on.  Therefore,
        # we need to set "ssh1.enable off" in the hash so that this value
        # of the hash matches the filer setting.
        #
        $self->set_ssh1_enable("off");

        disconnect($c);

        # Mark object as having key settings in sync with filer
        $self->{'_stale_key'} = "";
    }

    else {

        # Keys were not changed -- drop through & log comment
        $dbg ? logcomment("Keys are same as filer settings, will not change") : undef;
    }


    $dbg ? logcomment(Dumper($self)) : undef;
    logcomment("change_keysize() complete.");


    return $self;
}

=head2 check_mount_point($connection_obj)

Check the mount point to see if the filer is mounted.  If not mounted, mount
the filer using sim_mount.  It will try the mount a few times if for some
reason it fails.  If a new mount point is created in this routine,
$self->{'_umount'} will be set to "sim_umount /t/$self->{'_hostname'}/root"
signaling the calling process to handle the proper cleanup.

input:  connection to client.

return: 1=/t/$filername/root is a valid mount point that is good to use.
        0=/t/$filername/root is not a good mount point, and the attempt to
sim_mount failed.

B<Usage:> C<$cfg-E<gt>check_mount_point($connection_obj)>

=cut

sub check_mount_point()
{
    my ($self,$conn) = @_;
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag
    my $output;

    logcomment("In check_mount_point().");

    # To prevent problem with a stale mount point, first do a sim_umount
    # if the filer is already mounted.
    #
    my $cmd = sprintf ("mount | grep /t/%s/root", $self->{'_hostname'});
    (undef, undef,$output) = say($conn,$cmd);
    $dbg ? logcomment("Output of $cmd: $output") : undef;
    if ( $output =~ /^\/t\/$self->{'_hostname'}\/root on $self->{'_hostname'}/m ) {
        # filer is already mounted, do a sim_umount.
        logcomment("Do a sim_umount to prevent stale mount point problem.");
        $cmd = ("sim_umount /t/$self->{'_hostname'}/root");
        say($conn, $cmd);
    } else {
        # A new sim_mount will be created.  Set $self->{'_umount'} to signal
        # the caller to do a sim_umount to cleanup.
        $self->{'_umount'} = "sim_umount /t/$self->{'_hostname'}/root";
        logcomment("Set the '_umount' element in hash to: $self->{'_umount'}.");
    }

    my $mount_count = 0;
    my $max_tries = 2;  # Max tries before quiting.
    my $mount_ok = 0;
    while ($mount_ok == 0 && $mount_count<$max_tries) {
        $mount_count++;
        # Run sim_mount to mount filer.
        logcomment("Sim_mount $self->{'_hostname'} try #$mount_count.");
        $cmd = ("sim_mount $self->{'_hostname'}");
        say($conn, $cmd);

        # Check the mount point to verify that the mount is good.
        $cmd = sprintf ("mount | grep /t/%s/root", $self->{'_hostname'});
        (undef, undef,$output) = say($conn,$cmd);
        $dbg ? logcomment("Output of $cmd: $output") : undef;
       logcomment("OUTPUT : $output");
      if (( $output =~ /^\/t\/$self->{'_hostname'}\/root on $self->{'_hostname'}/m )
          || ($output =~ /$self->{'_hostname'}:\/\son\s\/t\/$self->{'_hostname'}\/root/)) {
            # Filer is mounted successfully, exit the while loop.
            logcomment("/t/$self->{'_hostname'}/root is a mount point that is good to use.");
            $mount_ok = 1;
        }
    }

    logcomment("Leaving check_mount_point(), returning $mount_ok.");
    return $mount_ok;
}

=head2 compare_config([$cfg2])

Function to compare two configuration objects. If a second config object is
not passed as an argument, then the first config object is compared to the
current state of the filer.

Used primarily to verify that settings were configured as desired on the filer.
Not that the various flags that may or may not be set to indicate stale
attributes or bogus values are ignored. If only one configuration object is
passed to this function, the C<get_config()> method will be called to obtain
the current settings on the filer, which will be compared to the settings in
the passed object.

B<Usage:> C<$cfg-E<gt>compare_config([$cfg2])>

B<Returns:> Boolean, TRUE ('1') if configurations match, FALSE ('0') if they
do not. Also sends result table to log.

=cut

sub compare_config () {

    my $cfg1 = shift;
    my $cfg2 = shift;
    my $rslt = "1";
    my $dbg = $cfg1->{'_debug_module_verbose'}; # Set Debug flag


    if ( ! $cfg2 ) {

        $dbg ?  logcomment("compare_config() passed single argument; Getting filer settings.") : undef;

        $cfg2 = $cfg1->get_config($cfg1->{'_filer'},$cfg1->{'_client'});
    }

    my %_attrs = _get_attributes();
    my $results = "Results of compare_config();\n\n\tAttribute\t\t\$cfg1\t"
        ."Results\t\$cfg2\n\t=========\t\t=====\t=======\t=====\n";

    foreach my $_attribute ( keys %_attrs ) {

        my $variabletab;

        if ( $_attribute !~ m/_auth_enable|serverkey|hostkey|idle/ ) {

            $variabletab = "\t\t";
        }

        else {

            $variabletab = "\t";
        }

        if ( $_attribute =~ m/debug|filer_ipv6/ ) {

            next;
        }

        if ( $cfg1->{$_attribute} eq $cfg2->{$_attribute} ) {

            $results = $results . "\t$_attribute $variabletab$cfg1->{$_attribute}\tMatches\t $cfg2->{$_attribute}\n";
            next;
        }

        else {

            $results = $results . "\t$_attribute$variabletab$cfg1->{$_attribute}\tDiffers\t$cfg2->{$_attribute}\n";
            $rslt = 0;
            next;
        }
    }

    $dbg ? logcomment($results) : undef;

    return $rslt;
}

=head2 get_client()

Retrieves client parameter (host specification of the client) in configuration
object.  Returns parameter value.

B<Usage:> C<$cfg-E<gt>get_client()>

=cut


=head2 get_client_name()

Retrieves client name parameter in configuration object.  Returns parameter
value.

B<Usage:> C<$cfg-E<gt>get_client_name()>

=cut

=head2 get_debug_module_verbose()

Retrieves _debug_module_verbose parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_debug_module_verbose()>

=cut

=head2 get_debug_ssh_verbose()

Retrieves _debug_ssh_verbose parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_debug_ssh_verbose()>

=cut

=head2 get_filer_ipv6()

Retrieves _filer_ipv6 parameter in configuration object.  Returns parameter
value. If this parameter is defined, ssh_can_connect() will attempt to
connect to this ipv6 address.

B<Usage:> C<$cfg-E<gt>get_filer_ipv6()>

=cut

=head2 get_hostname()

Retrieves hostname filer parameter in configuration object.  Returns parameter
value.

B<Usage:> C<$cfg-E<gt>get_hostname()>

=cut

=head2 get_password()

Retrieves password parameter in configuration object.  Returns parameter value.

B<Usage:> C<$cfg-E<gt>get_password()>

=cut

=head2 get_ssh1_enable()

Retrieves ssh1.enable filer parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh1_enable()>

=cut

=head2 get_ssh1_hostkeysize()

Retrieves ssh1 host key size filer parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh1_hostkeysize()>

=cut

=head2 get_ssh1_serverkeysize()

Retrieves ssh1 server key size filer parameter in configuration object.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh1_serverkeysize()>

=cut

=head2 get_ssh2_enable()

Retrieves ssh2.enable filer parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh2_enable()>

=cut

=head2 get_ssh2_hostkeysize()

Retrieves ssh2 host key size filer parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh2_hostkeysize()>

=cut

=head2 get_ssh_access()

Retrieves ssh.access filer parameter from configuration object. Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_access()>

=cut

=head2 get_ssh_connect_auth_method()

Retrieves ssh connect auth method filer parameter in configuration object.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_connect_auth_method()>

=cut

=head2 get_ssh_connect_protocol()

Retrieves ssh connect protocol parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_connect_protocol()>

=cut

=head2 get_ssh_enable()

Retrieves ssh.enable filer parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_enable()>

=cut

=head2 get_ssh_idle_timeout()

Retrieves ssh idle timeout filer parameter in configuration object.  Returns
parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_idle_timeout()>

=cut

=head2 get_ssh_passphrase()

Need to write function, Not implemented yet.

More about Whatever.

B<Usage:> C<$cfg-E<gt>get_ssh_passphrase()>

=cut

=head2 get_ssh_passwd_auth_enable()

Retrieves ssh passwd auth enable filer parameter in configuration object.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_passwd_auth_enable()>

=cut

=head2 get_ssh_port()

Returns ssh.port filer parameter in configuration object.

B<Usage:> C<$cfg-E<gt>get_ssh_port($setting)>

=cut

=head2 get_ssh_pubkey_auth_enable()

Retrieves ssh pubkey auth enable filer parameter in configuration object.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>get_ssh_pubkey_auth_enable()>

=cut


=head2 get_stale_opt()

Retrieves _stale_opt parameter in configuration object.  Returns parameter
value. Used to test whether update_filer_settings() succeeeded.

B<Usage:> C<$cfg-E<gt>get_stale_opt()>

=cut

sub get_stale_opt () {
    my $self = shift;
    return $self->{'_stale_opt'};
}

=head2 make_sure_ssh_connect_fails_and_log_result($log_text, [$CLIENT])

This method attempts to make an SSH connection to the filer and will log the
result of that attempt(failed connect means "PASS" result) along with text
that is passed to the method, followed by a statement about whether the
connection itself succeeded or not..

C<$CLIENT> is optional.  If it is specified, the connection will be created
with this specific client.  If C<$CLIENT> is not specified, the connection
will be using the client that is used to construct the object of this class.


B<Usage:> C<$cfg-E<gt>make_sure_ssh_cannot_connect_and_log_result($log_text,
[$CLIENT])>

B<Returns:> Boolean, TRUE ('1') if connection does not succeed, FALSE ('0') if
it does connect successfully.

=cut


sub make_sure_ssh_connect_fails_and_log_result {
        my $self = shift;
        my $log_text = shift;
        my $_client = shift;
        $_client = $self->{'_client'} if (! defined $_client);

        if ( $self->ssh_can_connect($_client) )  {
            logresult("FAIL", "Connection succeeded: " . $log_text);
            return undef;
        }
        else {
            logresult("PASS", "Connection failed: " . $log_text);
            return 1;
        }
}

=head2 make_sure_ssh_connect_succeeds_and_log_result($log_text, [$CLIENT])

This method attempts to make an SSH connection to the filer and will log the
result of that attempt(successful connect means "PASS" result) along with text
that is passed to the method, followed by a statement about whether the
connection itself succeeded or not..

C<$CLIENT> is optional.  If it is specified, the connection will be created
with this specific client.  If C<$CLIENT> is not specified, the connection
will be using the client that is used to construct the object of this class.


B<Usage:> C<$cfg-E<gt>make_sure_ssh_can_connect_and_log_result($log_text,
[$CLIENT])>

B<Returns:> Boolean, TRUE ('1') if connection succeeds, FALSE ('0') if it does
not.

=cut


sub make_sure_ssh_connect_succeeds_and_log_result {
        my $self = shift;
        my $log_text = shift;
        my $_client = shift;
        $_client = $self->{'_client'} if (! defined $_client);

        if ( $self->ssh_can_connect($_client) )  {
            logresult("PASS", "Connection succeeded: " . $log_text);
            return 1;
        }
        else {
            logresult("FAIL", "Connection failed: " . $log_text);
            return undef;
        }
}

=head2 restore_config()

Unconditionally restores backed up ssh options after completion of subtests,
including SSH Keys on filer and root user .ssh directory on filer to what they
were when C<backup_config()> was called with the passed C<$cfg> object.
Conditionally restore original keys(only if they have been changed by a
subtest) as properly restoring them involves a reboot. These key files and
other files are restored from C</tmp/<uniquenameE<gt>.tar> on the client.  This
client is the client that is used to construct the C<$cfg> object.

As with C<backup_config()>, C<restore_config()> should only be called once,
ideally from a top-level thpl script which calls a series of subtests. After
all subtests are run a call to C<$cfg-E<gt>restore_config()> will restore the
filer to the state it was in prior to running the thpl script. If you wish to
use C<restore_config()> repeatedly in a script make sure you are using a
different C<$cfg> object than was used for the first invocation of
C<backup_config()> or you will lose track of the original state of the filer
-- though if you have C<$cfg-E<gt>get_debug_module_verbose()> set TRUE you may
be able to figure out the original state and name of the tarfile containing
original keys by examining the log files.

Also note that if the keys on the filer are changed by invoking
C<change_keysize()> then C<restore_config()> will reboot the filer after
restoring the keys to ensure that the original keys are back in effect. This
function will not return until the filer has finished booting.

B<Usage:> C<$cfg-E<gt>restore_config()>

B<Returns:> Boolean, TRUE ('1') if successful, FALSE ('0') otherwise.

=cut

sub restore_config () {

    my $self = shift;
    my $client = $self->{'_client'};
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    logcomment("In restore_config() on client: $self->{'_client_name'}.");
    $dbg ? logcomment(Dumper($self)) : undef;

    my $c = connect("$client#telnet");

    cli($self->{'_filer'}, "secureadmin disable all;secureadmin enable all");
    cli($self->{'_filer'}, "options ssh.access $self->{'_ssh_access'}");
    cli($self->{'_filer'}, "options ssh.passwd_auth.enable $self->{'_ssh_passwd_auth_enable'}");
    cli($self->{'_filer'}, "options ssh.pubkey_auth.enable $self->{'_ssh_pubkey_auth_enable'}");
    cli($self->{'_filer'}, "options ssh1.enable $self->{'_ssh1_enable'}");
    cli($self->{'_filer'}, "options ssh2.enable $self->{'_ssh2_enable'}");
    cli($self->{'_filer'}, "options ssh.idle.timeout $self->{'_ssh_idle_timeout'}");
    cli($self->{'_filer'}, "options ssh.port $self->{'_ssh_port'}");
    cli($self->{'_filer'}, "options ssh.enable off");
    cli($self->{'_filer'}, "options ssh.enable on");
    cli($self->{'_filer'}, "options ssh.enable $self->{'_ssh_enable'}");

    my $cmd;
    my $output;
    if ( $self->keys_changed() ) {

        logcomment("Keys changed, will restore and reboot filer.");
        $self->change_keysize();

        if ( $self->{'_filer_keys_tar_file'} ) {

            # Who is the user?  Just for debug info.
            $dbg ? (undef,undef,$output) = say($c,"id") : undef;
            $dbg ? logcomment("Who am I?: $output") : undef;

            # check mount point.
            my $mount_ok = $self->check_mount_point($c);
            if ( $mount_ok <= 0 ) {  # The mount point is bad.
                logcomment("Failed!!! : not able to sim_mount $self->{'_hostname'}");
                # sim_umount to clean up if needed.
                $self->_do_sim_umount($c);

                # Mount point is bad.  Raise a FATAL error and exit.
                #
                my $e = NATE::Result::Fatal->new("Fatal error, mount point /t/$self->{'_hostname'}/root not usable.");
                $Log->log($e);
                return $self;
            }

            # Need to send the following commands with 2 say() calls.
            # Combining them makes the command line too long.
            #
            $cmd = "cd /;"
                . "/usr/local/bin/sudo rm -Rf /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/*;"
                . "/usr/local/bin/sudo tar -xvf $self->{'_filer_keys_tar_file'}";
            (undef, undef, $output) = say($c,$cmd);
            logcomment("Output of command $cmd:\n $output");

            $cmd = "/usr/local/bin/sudo rm $self->{'_filer_keys_tar_file'};"
                . "/usr/local/bin/sudo rm $self->{'_keys_changed_file'}";
            (undef, undef, $output) = say($c,$cmd);
            logcomment("Output of command $cmd:\n $output");

            # sim_umount to clean up if needed.
            $self->_do_sim_umount($c);

            my $b = connect($self->{'_filer'} . "#console");
            reboot($b) or logresult('WARN', $@);
            disconnect($b);
            return 0;
        }
    }

    else {

        $dbg ? logcomment("Keys were *not* Changed, no need to reboot filer.\nWill remove backup keys in $self->{'_filer_keys_tar_file'}") : undef;

        # Need to send the following commands with 2 say() calls.
        # Combining them may make the command line too long.
        #
        $cmd = "cd /;"
            . "/usr/local/bin/sudo rm -Rf /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/*";
        (undef, undef, $output) = say($c,$cmd);
        logcomment("Output of command $cmd:\n $output" );

        $cmd = "/usr/local/bin/sudo tar -xvf $self->{'_filer_keys_tar_file'};"
            . "/usr/local/bin/sudo rm $self->{'_filer_keys_tar_file'}";
        (undef, undef, $output) = say($c,$cmd);
        logcomment("Output of command $cmd:\n $output" );

        # sim_umount to clean up if needed.
        $self->_do_sim_umount($c);

    }

    $dbg ? logcomment(Dumper($self)) : undef;
    logcomment("restore_config() complete.");

    return 1;
}


# This is to enable debug log messages. Default is set to 0.

=head2 run_command_over_ssh()

This subroutine runs the interactive and non interactive commands over SSH.
It takes two arguments as input. First one is command to be executed and other
is hash consisting of user prompts for interactive commands. If user does not
specify the command then it returns undef. It also looks for IP address if user
has specified. If not specified then default ip is used from the config object.
If username and password is not specified then default user name and password
of config object will be used. SSH connect protocol is set to -2 by default
if user has not specified. SSH authentication method is passwd by default if
user has not specified. SSH port is taken from the config object if user has not
specified.Default port is "22".

B<Input:> Command to be executed on the filer. If command is non interactive
then user has to specify the expected prompt followed by input value. If you
don't want to specify any input value and just leave it as blank "" as in
example. This will press return key when prompt is encounterd.

B<Returns:> Output of the commands which is executed on the filer. If there are
            more than one commands then all the output will be returned

B<Usage:> C<$cfg-E<gt>run_command_over_ssh("comamnd to be executed")>

    Running command using default values
    $cfg->run_command_over_ssh("version");

    Running non interactive command using non default values
    $cfg->set_ssh_connect_protocol("SSH1");
    $cfg->set_ssh_connect_auth_method("pubkey");
    $cfg->run_command_over_ssh("version");

    Running interactive command
    my %user_prompts = (
        "Country Name.*"  => "US",
        "State or Province Name.*" => "AA",
        "Locality Name.*" => "",
        "Organization Name.*" => "",
        "Organization Unit Name.*" => "",
        "Common Name.*" => "",
        "Administrator email.*" => "",
        "Days until expires.*" => "",
        "Key length.*" => "",
    );
    $cfg->run_command_over_ssh("secureadmin setup -f ssl",\%user_prompts);

You can also run two non interactive commands over ssh. For example command
"version;options ssh" can be executed using same subroutine

    $cfg->run_command_over_ssh("version;options ssh");

Long running commands are also supported

    $cfg->run_command_over_ssh("sysstat 1");
    $cfg->run_command_over_ssh("version;sysstat 1");

I<NOTE:> You can run more then one non interactive commands over ssh using this
function. Also you can run non interactive and interactive command over ssh.
This function also supports long running command like sysstat, nfsstat etc.
All long running commands are executed for 10 seconds and sessions will be
killed by sending control + c.

=cut

sub run_command_over_ssh {
    my $self = shift;
    my $command=shift;
    my $hash_ref=shift;
    my %user_prompts;
    my $cmd;
    my $password;
    my $set_ssh_ver;
    my $method;
    $dbg ? $Log->comment("## Inside run_command_over_ssh() ") : "";

    $dbg ? $Log->comment("## Doing telnet connection to client") : "";

    my $client_conn=connect("$self->{'_client'}#telnet");

    if (!$command) {
        $Log->comment("You have not specified the command. Exiting...");
        return undef;
    }

    if($hash_ref) {
      $dbg ? $Log->comment("## Interactive command is getting executed over ssh") : "";
      %user_prompts=%$hash_ref;
    }

    if ($self->{'_ssh_alternate_username'}) {
        $cmd="ssh -l $self->{'_ssh_alternate_username'} ";
    } else {
        $cmd="ssh -l ". $client_conn->username ." ";
        $dbg ? $Log->comment("## User name not specified. Using default as root") : "";
    }

    if ($self->{'_ssh_alternate_ip'}) {
        $cmd=$cmd . $self->{'_ssh_alternate_ip'};
    } else {
        $cmd=$cmd . $self->{'_hostname'};
    }


    # If there is blank password then default case will be taken
    if ($self->{'_ssh_alternate_password'}) {
        $password= $self->{'_ssh_alternate_password'};
    } else {
        $password= $self->{'_password'};
        $dbg ? $Log->comment("## Password is not entered taking default") : "";
    }
    $dbg ? $Log->comment("## Password: $password") : "";

    if ($self->{'_ssh_connect_protocol'} eq "SSH2") {
        $set_ssh_ver="-2";
    } elsif ($self->{'_ssh_connect_protocol'} eq "SSH1") {
        $set_ssh_ver="-1";
    } else {
        $dbg ? $Log->comment("## SSH version not specified.Using default".
                          "version as SSH2") : "";
        $set_ssh_ver="-2";
    }
    $dbg ? $Log->comment("## SSH Version: $set_ssh_ver") : "";

    $cmd=$cmd." $set_ssh_ver";

    $method=$self->{'_ssh_connect_auth_method'};
    $dbg ? $Log->comment("## SSH Authentication method: $method") : "";

    if ( $self->{'_ssh_port'} ) {
        $cmd=$cmd." -p $self->{'_ssh_port'} ";
        $dbg ? $Log->comment("## SSH port: $self->{'_ssh_port'}") : "";
    }
    $cmd=$cmd." \"$command\" ";

    my $client_prompt=$self->{'_client_name'};

    $dbg ? $Log->comment("## Command to be executed $cmd") : "";

    my @_prompt_array = (".*continue connecting.*",
                         ".*password:",
                         ".*Permission denied.*",
                         ".*closed by remote host.*",
                           $client_prompt);

    my @keys= keys %user_prompts;
    push(@_prompt_array,@keys);

    $dbg ? $Log->comment("## Prompt_array: @_prompt_array") : "";

    my @r = say($client_conn,$cmd,
        '-noprompt', '-wait', @_prompt_array,'-eval_error','-timeout',"15");
    $dbg ? $Log->comment("##  $r[0] :$r[1]:$r[2]:$r[3]") : "";
    ## Following if is used to check in case wrong port is entered
    if ($r[1] =~ /Connection refused/ ) {
        $dbg ? $Log->comment("## Cannot connect : $r[0]") : "";
        disconnect($client_conn);
        return undef;
    }

    if ( $r[0] =~ /(yes\/no)/) {
       @r = say($client_conn, "yes", '-noprompt', '-wait', @_prompt_array,
                '-eval_error','-timeout',"10");
       $dbg ? $Log->comment("## Waiting for (yes/no) prompt. Found ".
           "prompt: $r[0]") : "";
    }

    if ($r[0] =~ /password/ && $method=~ /pubkey/) {
       $Log->comment("Waiting for the password. Probable issue is either ".
        "public key based authentication is disabled on the filer or public ".
        "key based authentication is not configured properly or you have ".
        "given public key as one of the options for this subroutine:$r[0]");
        disconnect($client_conn);
        return undef;
    }

    if ( $r[0] =~ /password/ && $method=~ /passwd/) {
        @r = say($client_conn,$password,'-noprompt','-wait', @_prompt_array,
                       '-eval_error','-timeout',"15");
         $dbg ? $Log->comment("## Waiting for password prompt. Found ".
        "prompt: $r[0] ") : "";
        sleep(2);
        ## follwoing condition will be executed it password entered in previous
        ## case is wrong and password prompt arrives again
        if ( $r[0] =~ /password/) {
            $Log->comment("Incorrect username or password. Please check the ".
                "username and password. SSH exiting test");
            disconnect($client_conn);
            return undef;
        }
    }

    if($r[1]=~ /.*Permission denied.*/ || $r[1]=~ /.*closed by remote host.*/
        || $r[1] =~ /.*Protocol major versions.*/
        || $r[1] =~ /.*server supports no auth methods.*/) {
        $dbg ? $Log->comment("## Exiting ssh run: $r[1]") : "";
        disconnect($client_conn);
        return undef;
    }
    ## Following condition is executed if command is non interactive and after
    ## executing command client prompt will come. So it is looking for client
    ## prompt.
    if($r[0] =~ /$client_prompt/) {

        $dbg ? $Log->comment("## The command you have executed is non ".
             "interactive") : "";
        disconnect($client_conn);
        return $r[1];
    }

    my $send_string;
    my $get_key;

    $dbg ? $Log->comment("## Loop for ssh interactive command. This loop will ".
        "executed if ssh command is interactive") : "";
    ## If not client prompt arrives after executing command the it is
    ## considered as interactive and loop will executed till the interactive
    ## session is completed.
    ## $count is to check number of time this while loop will be repeated.
    ## count is increased by 2 just to make sure that we are getting client
    ## prompt once we are done with running command.

    my $count=@keys;
    $count=$count +2;
    my $array_out;

    do {
        foreach $get_key (@keys) {
            if($r[0] =~ /$get_key/) {
                $dbg ? $Log->comment("## r[0]: $r[0] contains $get_key") : "";
                $send_string=$user_prompts{$get_key};
                $dbg ? $Log->comment("## From input Array : $get_key =>
                $send_string ") : "";

                if(!$send_string) {
                    $dbg ? $Log->comment("## User input is not specified sending \\r") : "";
                    $send_string="\r";
                }
                last;
           } else {
                $send_string=undef;
           }
        }
        if($send_string) {
            @r = say($client_conn,$send_string, '-noprompt', '-wait',
                 @_prompt_array,    '-eval_error','-timeout',"10");
            ## Appending all the output so that user can get other messages like
            ## connection closed etc .. in some cases
            $array_out=$array_out .$r[0].$r[1].$r[2]."\n";
            $dbg ? $Log->comment("## After sending user input : $array_out") :
            "";
            sleep(2);
        } else {
            $dbg ? $Log->comment("## User prompt not found: $r[0]") : "";
            # This will return output for long running commands or multiple
            # commands in case of interactive
            @r=say($client_conn,"\cC");
            $dbg ? $Log->comment("## Output is $r[0] : $r[1] :$r[2] :$r[3]") :
            "";
            $array_out=$array_out .$r[2];
            $dbg ? $Log->comment("## Array out is $array_out") : "";
            disconnect($client_conn);
            return $array_out;
        }

        if($r[0] =~ /$client_prompt/) {
            $dbg ? $Log->comment("## SSH interactive session ended successfully
            ".
                "$r[0]") : "";
            disconnect($client_conn);
            $array_out=$array_out .$r[1];
            $dbg ? $Log->comment("## Array out is $array_out") : "";
            return $array_out;
            last;
        }
      $dbg ? $Log->comment("## Value of count is $count") : "";
      $count--;
    } while($count >=0);
    ## This will be executed in case if user gives wrong input and prompt
    ## waits till correct input is given
    return undef;
}

=head2 run_command_after_login_to_filer_using_ssh()

This function runs command on the filer console. It logs in to the filer console
using SSH and runs the command on the filer console. It takes the filer and
the client information from the C<get_config()> object. If user does not specify
the command it will return undef. Other optional arguments are username,
password, ssh authentication method, ssh version. Username and password is taken
from the config object if user does not specify it. SSH authentication method is
password by default unless user specifies. SSH version is set to 2 by deafult.
You can even run long running commands on the console.

B<Input:> Command to be executed on the filer console.

B<Returns:> Ouput of the commands which is executed on the filer console

B<Usage:> C<$cfg-E<gt>_run_command_after_login_to_filer_using_ssh("command")>

    Example with default values
    $cfg->run_command_after_login_to_filer_using_ssh("version");

    Example running command using non default values
    $cfg->set_ssh_connect_protocol("SSH2");
    $cfg->set_ssh_connect_auth_method("passwd");
    $cfg->set_ssh_passwd_auth_enable("on");
    $cfg->set_ssh_pubkey_auth_enable("off");
    $cfg->update_filer_settings();
    my $out= $cfg->run_command_after_login_to_filer_using_ssh("version");

More then one command can be executed on the filer console.

    $out= $cfg->run_command_after_login_to_filer_using_ssh("version;options ssh");
    $out= $cfg->run_command_after_login_to_filer_using_ssh("version;sysstat 1");

In the second case output of version command will be printed followed by
output of sysstat 1, Output os sysstat will be printed for 15 seconds only

I<NOTE:> This function does not support interactive commands.

=cut

sub run_command_after_login_to_filer_using_ssh {
    my $self = shift;
    my $command=shift;
    my $port;
    my $method;
    my $set_ssh_ver;
    my $password;
    my $cmd;
    my $client_prompt=$self->{'_client_name'};

    my @_prompt_array = (  ".*continue connecting.*",
                           ".*Permission denied.*",
                           ".*closed by remote host.*",
                           ".*password:",
                           ".*Connection.*",
                           ".*Shell not supported on vfilers.*",
                           $client_prompt);

    $dbg ? $Log->comment("## Entering run_command_after_login_to_filer_using_ssh
    ") : "";

    if (!$command) {
        $Log->comment("You have not specified the command. Exiting...");
        return undef;
    }

    if ($self->{'_ssh_alternate_username'}) {
        $cmd="ssh -l $self->{'_ssh_alternate_username'} ";
    } else {
        $cmd="ssh -l root ";
        $dbg ? $Log->comment("## User name not specified. Using user from get_config()") : "";
    }

    #Sujatha: Modified _ssh_connect_ip to _ssh_alternate_ip
    if ($self->{'_ssh_alternate_ip'}) {
        $cmd=$cmd . $self->{'_ssh_alternate_ip'};
    } else {
        $cmd=$cmd . $self->{'_hostname'};
    }


    # If there is blank password then default case will be taken
    if ($self->{'_ssh_alternate_password'}) {
        $password= $self->{'_ssh_alternate_password'};
    } else {
        $password= $self->{'_password'};
        $dbg ? $Log->comment("## Password is not entered taking default") : "";
    }
    $dbg ? $Log->comment("## Password: $password") : "";

    if ($self->{'_ssh_connect_protocol'} eq "SSH2") {
        $set_ssh_ver="-2";
    } elsif ($self->{'_ssh_connect_protocol'} eq "SSH1") {
        $set_ssh_ver="-1";
    } else {
        $dbg ? $Log->comment("## SSH version not specified.Using default".
                          "version as SSH2") : "";
        $set_ssh_ver="-2";
    }
    $dbg ? $Log->comment("## SSH Version: $set_ssh_ver") : "";

    $cmd=$cmd." $set_ssh_ver";

    $method=$self->{'_ssh_connect_auth_method'};
    $dbg ? $Log->comment("## SSH Authentication method: $method") : "";

    if ( $self->{'_ssh_port'} ) {
        $cmd=$cmd." -p $self->{'_ssh_port'} ";
        $dbg ? $Log->comment("## SSH port: $self->{'_ssh_port'}") : "";
    }

    $dbg ? $Log->comment("## Command to be executed $cmd") : "";

    $dbg ? $Log->comment("## Doing telnet connection to client") : "";

    my $client_conn=connect("$self->{'_client'}#telnet");
    my @r = say($client_conn,$cmd,
        '-noprompt', '-wait', @_prompt_array,'-eval_error','-timeout',"15");

    ## Following if is used to check in case wrong port is entered
    if ($r[1] =~ /Connection refused/ ) {
        $dbg ? $Log->comment("## Cannot connect : $r[0]") : "";
        disconnect($client_conn);
        return undef;
    }

    if ( $r[0] =~ /(yes\/no)/) {
       @r = say($client_conn, "yes", '-noprompt', '-wait', @_prompt_array,
                '-eval_error','-timeout',"10");
    }

    if ($r[0] =~ /password/ && $method=~ /pubkey/) {
       $Log->comment("Waiting for the password. Probable issue is either ".
        "public key based authentication is disabled on the filer or public ".
        "key based authentication is not configured properly or you have ".
        "given public key as one of the options for this subroutine:$r[0]");
        disconnect($client_conn);
        return undef;
    }

    if ( $r[0] =~ /password/ && $method=~ /passwd/) {
        @r = say($client_conn,$password,'-noprompt','-wait', @_prompt_array,
                       '-eval_error','-timeout',"15");
           $dbg ? $Log->comment("## Waiting for password prompt. Found ".
        "prompt: $r[0]") : "";

        if ( $r[0] =~ /password/) {
            $Log->comment("Incorrect username or password. Please check the ".
                "username and password. SSH exiting test");
            disconnect($client_conn);
            return undef;
        }
    }
    $Log->comment("Output: $r[0]:$r[1]:$r[2]:$r[3]");

    if($r[0] =~ /.*Connection .* closed.*/) {
        $Log->comment("Exiting ssh run. Probable issue will be you have already".
            " logged to ssh and session is still active: $r[0]");
        disconnect($client_conn);
        return undef;

    }
    if($r[1]=~ /.*Permission denied.*/ || $r[1]=~ /.*closed by remote host.*/
        || $r[1] =~ /.*Protocol major versions.*/
        || $r[1] =~ /.*server supports no auth methods.*/
        || $r[0]=~ /.*Permission denied.*/
        || $r[0]=~ /.*Shell not supported on vfilers.*/) {
        $Log->comment("Exiting ssh run: $r[0]");
        disconnect($client_conn);
        return undef;
    }
    # This is done to make sure that there are no previous session running
    say( $client_conn, "\cC", '-blind' );
    sleep(1);
    say( $client_conn, "\r", '-blind' );
    sleep(1);

    @r = say($client_conn,$command, '-noprompt', '-wait',
                @_prompt_array,    '-eval_error','-timeout',"10");
    @r=say($client_conn,"\cC", '-blind');
    sleep(3);
    @r = say($client_conn,"logout telnet",'-wait', @_prompt_array,
             '-timeout',"10");
    $dbg ? logcomment("## After executing command $r[1]") : "";
    disconnect($client_conn);
    $r[1] =~ s/.*logout telnet//;
    return $r[1];
}

=head2 run_command_debug_enable()

This function prints the debug log when argument is passed. It takes "TRUE" as
the argument to print the debug logs. When passed as FALSE it will stop printing
the debug messages from that point. The default value set is false;

B<Input:> TRUE or FALSE

B<Returns:> None

B<Usage:>

    $cfg->run_command_debug_enable("TRUE");
    $cfg->run_command_over_ssh("version");
    $cfg->run_command_debug_enable("FALSE");

This will print the debug messages for run_command_over_ssh() subroutine
and later it will disable the debug messages. If you want to print debug
messages then you can call this module at the start of script. In the above
example it will print debug messages only for run_command_over_ssh().

=cut

sub run_command_debug_enable() {
    my $self=shift;
    my $setting = shift;

    if ($setting =~ /TRUE/) {
        $dbg=1;
        $dbg ? $Log->comment("Setting debug level to 1.Debug messages will be ".
            "printed All debug messages will start with ##") : "";
    } else {
        $Log->comment("Setting debug level to 0.Debug messages will not be
        printed");
        $dbg=undef;
    }
    return $self;
}

=head2 set_ssh_alternate_username($setting)

This function is used to specify the username. This username will be used to
login to filer using SSH. If not specified it will take the default user as root
This will be applicable only when you call following functions run_command_over_ssh ,
run_command_after_login_to_filer_using_ssh

B<Input:> Username

B<Returns:> config object with username which is specified

B<Usage:> C<$cfg-E<gt>set_ssh_alternate_username("testuser")>

=cut

=head2 set_ssh_alternate_password($setting)

This function is used to set the password for the user.If not called default
password for root is used. This password will be used along with the user which
is specified for login to filer using ssh.This will be applicable only when you
call following functions run_command_over_ssh ,run_command_after_login_to_filer_using_ssh

B<Input:> Password for the user

B<Returns:> config object with password which is specified

B<Usage:> C<$cfg-E<gt>set_ssh_alternate_password("password for user")>

    Sample to use in script
    $cfg->set_ssh_alternate_username("testuser");
    $cfg->set_ssh_alternate_password("test123");

=cut

=head2 set_ssh_alternate_ip($setting)

This function is used to specify the SSH IP for connecting to filer.

B<Input:> IP address

B<Returns:> config object with ip address  which is specified

B<Usage:> C<$cfg-E<gt>set_ssh_alternate_ip("10.1.1.1")>

=cut

=head2 set_ssh_sessions($setting)

This function is used to specify the no of SSH connection to the filer

B<Input:> No of ssh connections

B<Returns:> config object with no. of SSH connections which is specified

B<Usage:> C<$cfg-E<gt>set_ssh_sessions("4")>

=cut

=head2 set_debug_module_verbose($setting)

Updates module debug parameter in configuration object. Accepts Boolean. May
be used to print out debug info for an entire script or a single method,
depending on where in your script the parameter is set.

B<Usage:> C<$cfg-E<gt>set_debug_module_verbose($setting)>

=cut

=head2 set_debug_ssh_verbose($setting)

Updates ssh debug parameter in configuration object. Accepts Boolean. Causes
the actual ssh connect command called by C<$cfg-E<gt>ssh_can_connect($CLIENT)>
to operate in verbose mode for debugging.

B<Usage:> C<$cfg-E<gt>set_debug_ssh_verbose($setting)>

=cut

=head2 set_filer_ipv6($setting)

Sets _filer_ipv6 parameter in configuration object.  Returns parameter
value. If this parameter is defined, ssh_can_connect() and other methods
which call it will attempt to connect to this ipv6 address instead of the
hostname specified in the config object. If you wish to make subsequent
connections to the hostname instead of this IPv6 address, pass "undef"
to this function.

B<Usage:> C<$cfg-E<gt>set_filer_ipv6($setting)>

=cut

=head2 set_password($setting[,$bogus])

Sets the password parameter in configuration object to be used for making SSH
connections to C<$FILER>. Normally this is set to the password defined for
C<$FILER> in the testbed setup and no changes are necessary. If the password
passed to this method is expected to be invalid, then setting a value for
C<$bogus> that evaluates to C<TRUE> will set a flag indicating this. This
"bogus" flag is for planned negative testcase handling in a future version of
this module.

B<Usage:> C<$cfg-E<gt>set_password($setting[,$bogus])>

=cut

sub set_password () {

    my $self = shift;
    my $setting = shift;
    my $bogus = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    my $anon = sub {
        $self->{'_password'} = $setting;
    };

    if ( ! $bogus ) {

        $dbg ? logcomment("Valid value assumed for _password") : undef;

        $self->{'_password_bogus'} ? # Change only if exists
            $self->{'_password_bogus'} = "" : undef;

        $anon->();
    }

    else {
        $dbg ? logcomment("Setting flag to indicate bogus setting for _password (negative testcase)") : undef;
        $self->{'_password_bogus'} = 1;
        $anon->();
    }

    return $self;
}

=head2 set_ssh1_enable($setting)

Updates ssh1.enable filer parameter in configuration object. Accepts "on" or
"off", other values will cause "bogus" flag to be set for this parameter.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>set_ssh1_enable($setting)>

=cut

=head2 set_ssh1_hostkeysize($setting)

Updates SSH1 hostkey size parameter in configuration object. This function is
used in preparation for invoking the C<change_keysize()> method or (preferably)
the C<update_filer_settings()> method in either of this module's subclasses,
and does not by itself make any changes on the filer. A flag is set to indicate
that the keysizes in the configuration object are "stale", meaning that they
do not match SSH key sizes on the filer. If one keysize is changed, all three
keysizes are considered "stale" since it is impossible to change one without
changing all three. Additionally, a flag will be set if C<$setting> is a bogus
value, bogus meaning that the following is not true: C<(384 <= $setting <=
2048)> If the value used for C<$setting> does not differ by at least 128 bits
from the value used for SSH1 server key size, then C<change_keysize()> will
set flags for both values indicating that they are both bogus. This "bogus"
flag is for planned negative testcase handling in a future version of this
module. Note that passing no argument to this method will change the parameter
in the config object to the value that ONTAP uses as a default -- in this case
"768".

If the desired action is to regenerate SSH keys with the same keysize as was
used previously, use something like one of the following examples to set the
"stale key" flag or C<change_keysize()> will ignore your attempts to
regenerate keys on the filer. Even if you do this, you will end up with I<new>
keys of the same size, the filer will not regenerate the same exact keys that
it already has because a new random seed will be used.

Examples of preparation to regenerate keys of the same size as those already
existing on the filer:

C<$cfg-E<gt>set_ssh1_hostkeysize($cfg-E<gt>get_ssh1_hostkeysize())>  (if
current setting on C<$FILER> is not known)

C<$cfg-E<gt>set_ssh1_hostkeysize($setting)> (if current setting on C<$FILER>
is known by other means))

B<Usage:> C<$cfg-E<gt>set_ssh1_hostkeysize($setting)>

=cut

sub set_ssh1_hostkeysize () {

    my $self = shift;
    my $setting = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    # Anonymous subroutine to update object
    my $anon = sub {
        $self->{'_ssh1_hostkeysize'} = $setting;
        $self->{'_stale_key'} = 1;
    };

    # Set to default if no argument passed
    $setting = $setting ? $setting : "768";

    if ( "384" <= $setting && $setting <= "2048" ) {

        $dbg ? logcomment("Valid value detected for _ssh1_hostkeysize") : undef;

        $self->{'_ssh1_hostkeysize_bogus'} ? # Change only if exists
            $self->{'_ssh1_hostkeysize_bogus'} = "" : undef;
        $anon->();
    }

    else {

        $dbg ? logcomment("Setting flag to indicate bogus setting for _ssh1_hostkeysize (negative testcase)") : undef;

        $self->{'_ssh1_hostkeysize_bogus'} = 1;
        $anon->();
    }

    return $self;

}

=head2 set_ssh1_serverkeysize($setting)

Updates SSH1 serverkey size parameter in configuration object. This function
is used in preparation for invoking the C<change_keysize()> method or
(preferably) the C<update_filer_settings()> method in either of this module's
subclasses, and does not by itself make any changes on the filer. A flag is
set to indicate that the keysizes in the configuration object are "stale",
meaning that they do not match SSH key sizes on the filer. If one keysize is
changed, all three keysizes are considered "stale" since it is impossible to
change one without changing all three. Additionally, a flag will be set if
C<$setting> is a bogus value, bogus meaning that the following is not true:
C<(384 <= $setting <= 2048)> If the value used for C<$setting> does not differ
by at least 128 bits from the value used for SSH1 host key size, then
C<change_keysize()> will set flags for both values indicating that they are
both bogus. This "bogus" flag is for planned negative testcase handling in a
future version of this module. Note that passing no argument to this method
will change the parameter in the config object to the value that ONTAP uses as
a default -- in this case "512".

If the desired action is to regenerate SSH keys with the same keysize as was
used previously, use something like one of the following examples to set the
"stale key" flag or C<change_keysize()> will ignore your attempts to
regenerate keys on the filer. Even if you do this, you will end up with I<new>
keys of the same size, the filer will not regenerate the same exact keys that
it already has because a new random seed will be used.

Examples of preparation to regenerate keys of the same size as those already
existing on the filer:

C<$cfg-E<gt>set_ssh1_serverkeysize($cfg-E<gt>get_ssh1_serverkeysize())>  (if
current setting on C<$FILER> is not known)

C<$cfg-E<gt>set_ssh1_serverkeysize($setting)> (if current setting on C<$FILER>
is known by other means))

B<Usage:> C<$cfg-E<gt>set_ssh1_serverkeysize($setting)>

=cut

sub set_ssh1_serverkeysize () {

    my $self = shift;
    my $setting = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    # Anonymous subroutine to update object
    my $anon = sub {
        $self->{'_ssh1_serverkeysize'} = $setting;
        $self->{'_stale_key'} = 1;
    };

    # Set to default if no argument passed
    $setting = $setting ? $setting : "512";

    if ( "384" <= $setting && $setting <= "2048" ) {

        $dbg ? logcomment("Valid value detected for _ssh1_serverkeysize") : undef;

        $self->{'_ssh1_serverkeysize_bogus'} ? # Change only if exists
            $self->{'_ssh1_serverkeysize_bogus'} = "" : undef;
        $anon->();
    }

    else {

        $dbg ? logcomment("Setting flag to indicate bogus setting for _ssh1_serverkeysize (negative testcase)") : undef;

        $self->{'_ssh1_serverkeysize_bogus'} = 1;
        $anon->();
    }

    return $self;

}

=head2 set_ssh2_enable($setting)

Updates ssh2.enable filer parameter in configuration object. Accepts "on" or
"off", other values will cause "bogus" flag to be set for this parameter.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>set_ssh2_enable($setting)>

=cut

=head2 set_ssh2_hostkeysize($setting)

Updates SSH2 hostkey size parameter in configuration object. This function is
used in preparation for invoking the C<change_keysize()> method or (preferably)
the C<update_filer_settings()> method in either of this module's subclasses,
and does not by itself make any changes on the filer. A flag is set to indicate
that the keysizes in the configuration object are "stale", meaning that they
do not match SSH key sizes on the filer. If one keysize is changed, all three
keysizes are considered "stale" since it is impossible to change one without
changing all three. Additionally, a flag will be set if C<$setting> is a bogus
value, bogus meaning that the following is not true: C<(768 <= $setting <=
2048)> This "bogus" flag is for planned negative testcase handling in a future
version of this module. Note that passing no argument to this method will
change the parameter in the config object to the value that ONTAP uses as a
default -- in this case "768".

If the desired action is to regenerate SSH keys with the same keysize as was
used previously, use something like one of the following examples to set the
"stale key" flag or C<change_keysize()> will ignore your attempts to regenerate
keys on the filer. Even if you do this, you will end up with I<new> keys of
the same size, the filer will not regenerate the same exact keys that it
already has because a new random seed will be used.

Examples of preparation to regenerate keys of the same size as those already
existing on the filer:

C<$cfg-E<gt>set_ssh2_hostkeysize($cfg-E<gt>get_ssh2_hostkeysize())>  (if
current setting on C<$FILER> is not known)

C<$cfg-E<gt>set_ssh2_hostkeysize($setting)> (if current setting on C<$FILER>
is known by other means))

B<Usage:> C<$cfg-E<gt>set_ssh2_hostkeysize($setting)>

=cut

sub set_ssh2_hostkeysize () {

    my $self = shift;
    my $setting = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    # Anonymous subroutine to update object
    my $anon = sub {
        $self->{'_ssh2_hostkeysize'} = $setting;
        $self->{'_stale_key'} = 1;
    };

    # Set to default if no argument passed
    $setting = $setting ? $setting : "768";

    if ( "768" <= $setting && $setting <= "2048" ) {

        $dbg ? logcomment("Valid value detected for _ssh2_hostkeysize") : undef;

        $self->{'_ssh2_hostkeysize_bogus'} ? # Change only if exists
            $self->{'_ssh2_hostkeysize_bogus'} = "" : undef;
        $anon->();
    }

    else {

        $dbg ? logcomment("Setting flag to indicate bogus setting for _ssh2_hostkeysize (negative testcase)") : undef;

        $self->{'_ssh2_hostkeysize_bogus'} = 1;
        $anon->();
    }

    return $self;

}

=head2 set_ssh_access($setting[,$bogus])

Sets the access string parameter in configuration object to be used on the
filer. A flag is set to indicate that the value in the configuration object
for "ssh.access" is "stale", meaning that it does not match the value on the
filer.

If the string passed as C<$setting> is not a valid string, the user must
explicitly pass an argument for C<$bogus> that evaluates to C<TRUE>. This
"bogus" flag is for planned negative testcase handling in a future version of
this module. Perhaps in the future this library can be enhanced to detect for
itself whether the setting is bogus or not. Currently it's too big of a mess
to deal with, for example the filer accepts "2.3.&.4.5" as a hostname.

B<Usage:> C<$cfg-E<gt>set_ssh_access($setting[,$bogus])>

=cut

sub set_ssh_access () {



    my $self = shift;
    my $setting = shift;
    my $bogus = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    my $anon = sub {
        $self->{'_ssh_access'} = $setting;
        $self->{'_stale_opt'} = 1;
    };

    if ( ! $bogus ) {

        $dbg ? logcomment("Valid value assumed for _ssh_access") : undef;

        $self->{'_ssh_access_bogus'} ? # Change only if exists
            $self->{'_ssh_access_bogus'} = "" : undef;

        $anon->();
    }

    else {
        $dbg ? logcomment("Setting flag to indicate bogus setting for _ssh_access (negative testcase)") : undef;
        $self->{'_ssh_access_bogus'} = 1;
        $anon->();
    }

    return $self;
}

=head2 set_ssh_connect_auth_method($setting)

Updates the ssh connect auth method parameter in configuration object. This
controls how C<ssh_can_connect()> will behave; if C<$setting> is C<"passwd">
or C<undef> then the connection will be attempted using password
authentication. If C<$setting> is C<"pubkey"> then the connection will be
attempted using public key authentication.

B<Usage:> C<$cfg-E<gt>set_ssh_connect_auth_method($setting)>

=cut

sub set_ssh_connect_auth_method () {

    my $self = shift;
    my $setting = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    if ( $setting eq "passwd" || $setting eq undef ) {

        $dbg ? logcomment("Setting _ssh_connect_auth_method to \"passwd\"") : undef;

        $self->{'_ssh_connect_auth_method'} = "passwd";
    }

    elsif ( $setting eq "pubkey" ) {

        $dbg ? logcomment("Setting _ssh_connect_auth_method to \"pubkey\"") : undef;

        $self->{'_ssh_connect_auth_method'} = "pubkey";
    }

    else {

        $dbg ? logcomment("\"$setting\" Invalid for _ssh_connect_method") : undef;
    }
    return $self;
}

=head2 set_ssh_connect_protocol($setting)

Updates the ssh connect protocol parameter in configuration object. This
controls how C<ssh_can_connect()> will behave; if C<$setting> is C<"SSH2"> or
C<undef> then the connection will be attempted using SSH2. If C<$setting> is
C<"SSH1"> then the connection will be attempted using SSH1.

B<Usage:> C<$cfg-E<gt>set_ssh_connect_protocol($setting)>

=cut

sub set_ssh_connect_protocol () {

    my $self = shift;
    my $setting = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    if ( $setting eq "SSH2" || $setting eq undef ) {

        $dbg ? logcomment("Setting _ssh_connect_protocol to $setting") : undef;

        $self->{'_ssh_connect_protocol'} = $setting;
    }

    elsif ( $setting eq "SSH1" ) {

        $dbg ? logcomment("Setting _ssh_connect_protocol to $setting") : undef;

        $self->{'_ssh_connect_protocol'} = $setting;
    }

    else {

        $dbg ? logcomment("\"$setting\" Invalid for _ssh_connect_protocol") : undef;
    }
    return $self;
}

=head2 set_ssh_enable($setting)

Updates ssh.enable filer parameter in configuration object. Accepts "on" or
"off", other values will cause "bogus" flag to be set for this parameter.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>set_ssh_enable($setting)>

=cut

=head2 set_ssh_idle_timeout($setting)

Updates the ssh idle timeout parameter in configuration object. Currently there
are no methods in this module that exercise this, but the ability to set/get
this value both in the configuration object and on the filer are implemented.

B<Usage:> C<$cfg-E<gt>set_ssh_idle_timeout($setting)>

=cut

sub set_ssh_idle_timeout () {

    my $self = shift;
    my $setting = shift;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    my $anon = sub {
        $self->{'_ssh_idle_timeout'} = $setting;
        $self->{'_stale_opt'} = 1;
    };

    if ( "1" <= $setting && $setting <= "31536000" ) {

        $dbg ? logcomment("Valid value detected for _ssh_idle_timeout") : undef;

        $self->{'_ssh_idle_timeout_bogus'} ? # Change only if exists
            $self->{'_ssh_idle_timeout_bogus'} = "" : undef;

        $anon->();
    }

    else {

        $dbg ? logcomment("Setting flag to indicate bogus setting for _ssh_idle_timeout (negative testcase)") : undef;

        $self->{'_ssh_idle_timeout_bogus'} = 1;
        $anon->();
    }

    return $self;
}

=head2 set_ssh_passphrase($setting)

Need to write function. Not implemented yet.

More about Whatever.

B<Usage:> C<$cfg-E<gt>set_ssh_passphrase()>

=cut

=head2 set_ssh_passwd_auth_enable($setting)

Updates ssh.passwd_auth.enable filer parameter in configuration object.
Accepts "on" or "off", other values will cause "bogus" flag to be set for this
parameter. Returns parameter value.

B<Usage:> C<$cfg-E<gt>set_ssh_passwd_auth_enable($setting)>

=cut

=head2 set_ssh_port($setting)

Updates ssh.port filer parameter in configuration object. Accepts a number.
Range checking ensures that a port between 1 and 65536(inclusive) is chosen.
Returns parameter value.

B<Usage:> C<$cfg-E<gt>set_ssh_port($setting)>

=cut

=head2 set_ssh_pubkey_auth_enable($setting)

Updates ssh.pubkey_auth.enable filer parameter in configuration object. Accepts
"on" or "off", other values will cause "bogus" flag to be set for this
parameter. Returns parameter value.

B<Usage:> C<$cfg-E<gt>set_ssh_pubkey_auth_enable($setting)>

=cut

=head2 ssh_can_connect([$CLIENT])

This method attempts to make an SSH connection to the filer.

C<$CLIENT> is optional.  If it is specified, the connection will be created
with this specific client.  If C<$CLIENT> is not specified, the connection
will be using the client that is used to construct the object of this class.

If the configuration object passed to this method has flags set to indicate
either stale options or stale keys (settings that do not match the settings on
the filer) then a "FAIL" result is logged and the attempt to connect is
aborted. Otherwise, various elements of the configuration object are used by
this method to determine how it will go about making the connection. After the
connection is made(determined by the presence of a filer prompt) the
connection is terminated. Primarily, these elements are set using the
following methods; please see their respective sections for details:

=over 8

=item *

C<set_ssh_connect_protocol($setting)> - This setting determines whether the
connection will be attempted using SSH1 or SSH2.

=item *

C<set_ssh_connect_auth_method($setting)> - This setting determines whether
connection will be attempted using Password Authentication or Public Key
Authentication.

=item *

C<set_password($setting[,$bogus])> - This is the password that will be used
for Password Authentication(default is obtained from the testbed setup for
C<$FILER>).

=item *

C<set_ssh_passphrase($setting)> - Currently not used, this element will be
used in the future to generate public keys with passphrases, and to connect
using PubKey/Passphrase Authentication.


=back

B<Usage:> C<$cfg-E<gt>ssh_can_connect([$CLIENT])>

B<Returns:> Boolean, TRUE ('1') if connection succeeds, FALSE ('0') if it does
not.

=cut

sub ssh_can_connect () {

    ######## Need check for idle timeout and access function
    ######## If stale opts or keys, abort connect
    my $self = shift;
    my $client = shift;
    $client = $self->{'_client'} if (! defined $client);

    my $_verbose = undef;
    my $_protocol_version = undef;
    my $_userstring = undef;
    my $_connect_host;

    if ($self->{'_filer_ipv6'}) {
        $_connect_host = $self->{'_filer_ipv6'};
    } else {
        $_connect_host = $self->{'_hostname'};
    }

    my $_temp_array = $self->{'_prompt'};
    my @_prompt_array = @$_temp_array;
    my @r;
    my $rslt = undef;
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    $dbg ? logcomment("in ssh_can_connect()") : undef;

    my $c = connect("$client#telnet");
    $c->timeout(100);

    $self->{'_ssh_connect_auth_method'} ? "" : logcomment('FAIL', "You must specify _ssh_connect_auth_method.");

    if ( $self->{'_stale_opts'} || $self->{'_stale_keys'} ) {

        $dbg ? logcomment('FAIL', "Attempted to connect to filer with stale options or keys -- call update_filer_settings() before attempting to connect.") : undef;
    }


    $dbg ? logcomment("\@_temp_array is: " . Dumper($_temp_array)) : undef;
    $dbg ? logcomment("\@_prompt_array is: " . Dumper(@_prompt_array)) : undef;

    if ( $self->{'_ssh_connect_auth_method'} eq "passwd" ) {

        # Ensure we connect to filer using same username that we are using
        # to connect to client
        $_userstring = "-l " . $c->username;
    }

    if ( $self->{'_debug_ssh_verbose'} && $self->{'_debug_ssh_verbose'} ne "0" && $self->{'_debug_ssh_verbose'} ne "off" ) {

        $dbg ? logcomment("Will use ssh in verbose mode, arg will be: -v") : undef;

        $_verbose = "-v";
    }


    if ( $self->{'_ssh_connect_protocol'} eq "SSH2" || $self->{'_ssh_connect_protocol'} eq undef ) {

        $dbg ? logcomment("Will use SSH2, arg will be: -2") : undef;

        $_protocol_version = "-2";
    }

    elsif ( $self->{'_ssh_connect_protocol'} eq "SSH1" ) {

        $dbg ? logcomment("Will use SSH1, arg will be: -1") : undef;

        $_protocol_version = "-1";
    }


    my $_connect_string = "ssh $_protocol_version $_verbose $_connect_host -p $self->{'_ssh_port'} $_userstring";

    $dbg ? logcomment("\$connect_string is $_connect_string") : undef;

    # Build an array of things we want say() to match on besides just the prompt

    #Sujatha: Added Conenction.*closed and shell not supported to prompt_array
    push(@_prompt_array, "(yes\/no)", "assword:", "refused", "NASTY", "denied","takeover", "Connection.*closed", "Shell not supported");

    $dbg ? logcomment("\@_prompt_array is: " . Dumper(@_prompt_array)) : undef;

    @r = say($c, $_connect_string, '-noprompt', '-wait', @_prompt_array,
             '-eval_error');
    if ($@ =~ /TIMEOUT/) {
         logcomment("Unable to connect");
         $rslt = undef;
         return $rslt;
    } elsif ($@) {
         die "ssh did not work, but for reasons we didn't expect: $@";
    }

    $dbg ? logcomment("\@r is: " . Dumper(@r)) : undef;

    if ( $r[0] =~ /(yes\/no)/ ) {

        $dbg ? logcomment("Found yes/no!!!") : undef;

        @r = say($c, "yes", '-noprompt', '-wait', @_prompt_array);

        $dbg ? logcomment("\@r is: " . Dumper(@r)) : undef;

        if ( $self->{'_ssh_connect_auth_method'} eq "passwd" ) {

            if ( $r[0] =~ /assword/ ) {

                @r = say($c, $self->{'_password'}, '-noprompt', '-wait', @_prompt_array);

                $dbg ? logcomment("\@r is: " . Dumper(@r)) : undef;
                $dbg ? logcomment("Going to look for $_prompt_array[0] in \@r") : undef;

                if ( $r[0] =~ /$_prompt_array[0]/  || $r[0] =~ /takeover/ ) {

                    $dbg ? logcomment("Connected successfully! Found $_prompt_array[0]") : undef;

                    say($c, "\cD"); # Disconnect

                    $rslt = 1;
                }
            }

            elsif ( $r[0] =~ /$_prompt_array[0]/  || $r[0] =~ /takeover/ ) {

                $dbg ? logcomment("Connected successfully! Found $_prompt_array[0]") : undef;

                say($c, "\cD"); # Disconnect

                $rslt = 1;
            }
        }

        elsif ( $self->{'_ssh_connect_auth_method'} eq "pubkey" ) {

            my $tmpval = $r[0];

            if ( $r[0] =~ /$_prompt_array[0]/ || $r[0] =~ /takeover/ ) {

                    if ( $r[0] =~ /assword:/ ) {

                        $dbg ? logcomment("Got to password!!") : undef;

                        my @r1 = say($c, $self->{'_password'}, '-noprompt', '-wait', @_prompt_array);

                        if ( $r1[0] =~ /$_prompt_array[0]/  || $r1[0] =~ /takeover/ ) {
                            $dbg ? logcomment("Found Prompt! ") : undef;
                            say($c, "\cD"); # Disconnect
                            $rslt = "1";
                        }
                    }

                $dbg ? logcomment("Connected successfully! Found $_prompt_array[0]") : undef;

                say($c, "\cD"); # Disconnect

                $rslt = 1;
            }

            if ( $tmpval =~ /ogin/ ) {
                logresult("FAIL", "Should not see login prompt here!!!(burt 235446)");
            }
        }

    }

    elsif ( $r[0] =~ /assword:/ ) {

        $dbg ? logcomment("Got to password!!") : undef;

        my @r1 = say($c, $self->{'_password'}, '-noprompt', '-wait', @_prompt_array);

        if ( $r1[0] =~ /$_prompt_array[0]/  || $r1[0] =~ /takeover/ ) {
            $dbg ? logcomment("Found Prompt! ") : undef;
            say($c, "\cD"); # Disconnect
            $rslt = "1";
        }
        #sujatha: added connection closed and shell not aupported for finding match.
        elsif ( $r1[0] =~ /refused|NASTY|denied|Connection.*closed|Shell not supported/ ) {

            $dbg ? logcomment("Could not connect!") : undef;

            $dbg ? logcomment("\@r is: " . Dumper(@r)) : undef;

            $rslt = undef;

        }

        $dbg ? logcomment("\@r1 is: " . Dumper(@r)) : undef;

    }

    elsif ( $r[0] =~ /$_prompt_array[0]/ || $r[0] =~ /takeover/ ) {

        $dbg ? logcomment("Found Prompt!") : undef;

        $dbg ? logcomment("\@r is: " . Dumper(@r)) : undef;

        say($c, "\cD"); # Disconnect

        $rslt = "1";

    }
    #sujatha: added connection closed and shell not aupported for finding match.
    elsif ( $r[0] =~ /refused|NASTY|denied|Connection.*closed|Shell not supported/ ) {

        $dbg ? logcomment("Could not connect!") : undef;

        $dbg ? logcomment("\@r is: " . Dumper(@r)) : undef;

        $rslt = undef;

    }

    else {

        $dbg ? logcomment("Not Found!!! \@r is: " . Dumper(@r) . " and prompt string to look for is: " . $_prompt_array[0] ) : undef;

    }

    $rslt ? logcomment("Successfully Connected") : logcomment("Could not connect");

    disconnect($c);

    $dbg ? logcomment("leaving ssh_can_connect()") : undef;

    return $rslt;
}

=head2 ssh_client_config_backup()

Backs up ssh-related configuration files on the client.  This client is the
client that is used to construct the object of this class.

If C<~/.ssh> exists it will tar up the contents in a tar file on the client,
then clear all files under C<~/.ssh>.  The tar filename is
$self->{'_client_ssh_tar_file'}.

B<Usage:> C<$cfg-E<gt>ssh_client_config_backup()>

=cut

sub ssh_client_config_backup () {

    my $self = shift;
    my $rslt = 1;
    my $c = connect("$self->{'_client'}#telnet");
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    $dbg ? logcomment("In ssh_client_config_backup()") : undef;
    $dbg ? logcomment("\$c is: " . Dumper($c)) : undef;

    # Who is the user?  Just for debug info.
    my $output;
    $dbg ? (undef,undef,$output) = say($c,"id") : undef;
    $dbg ? logcomment("Who am I?: $output") : undef;

    my @r = say($c, "cd;tar -cvf $self->{'_client_ssh_tar_file'} .ssh && rm -Rf .ssh/*");

    if ( $r[2] =~ /$filesystem_errors/i ) {
        $dbg ? logcomment('WARN', "File Error in ssh_client_config_backup()" . Dumper(@r)) : undef;
        $rslt = "0";
    }

    if ($dbg) {
        my @t = say($c, "ls -l .ssh");
        logcomment("The output of ls -l .ssh is:\n ". Dumper(@t). "\n");
    }

    disconnect($c);

    $dbg ? logcomment("leaving ssh_client_config_backup()") : undef;

    return $rslt;
}

=head2 ssh_client_config_restore()

Restores ssh-related configuration files on the client that were created with
C<$cfg-E<gt>ssh_client_config_backup()>.  This client is the client that is
used to construct the C<$cfg> object.

This method will clear the contents of C<~/.ssh>, then extract the contents of
the tar file to C<~/.ssh> and remove the tar file.  The name of the tar file
is $self->{'_client_ssh_tar_file'}.

B<Usage:> C<$cfg-E<gt>ssh_client_config_restore()>

B<Returns:> No return value.

B<Returns:> Boolean, TRUE ('1') if restore succeeds, FALSE ('0') if it does not.

=cut

sub ssh_client_config_restore () {

    my $self = shift;
    my $rslt = 1;
    my $c = connect("$self->{'_client'}#telnet");

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    $dbg ? logcomment("in ssh_client_config_restore()") : undef;

    # Who is the user?  Just for debug info.
    my $output;
    $dbg ? (undef,undef,$output) = say($c,"id") : undef;
    $dbg ? logcomment("Who am I?: $output") : undef;

    my $tar_file = $self->{'_client_ssh_tar_file'};
    my @r = say($c, "cd;rm -Rf .ssh/* && tar -xvf $tar_file && rm $tar_file");

    if ($dbg) {
        my @t = say($c, "ls -l .ssh");
        logcomment("The output of ls -l .ssh is:\n ". Dumper(@t). "\n");
    }

    if ( $r[2] =~ /$filesystem_errors/i ) {
        $dbg ? logcomment('WARN', "File Error in ssh_client_config_restore()" . Dumper(@r)) : undef;
        $rslt = "0";
    }

    disconnect($c);

    $dbg ? logcomment("leaving ssh_client_config_restore()") : undef;

    return $rslt;
}

=head2 ssh_client_rm_known_hosts()

Removes C<~/.ssh/known_hosts> file from the client.  This client is the client
that is used to construct the object of this class.

This method will delete the C<~/.ssh/known_hosts> file.

B<Usage:> C<$cfg-E<gt>ssh_client_rm_known_hosts()>

B<Returns:> Boolean, TRUE ('1') if delete succeeds, FALSE ('0') if it does not.

=cut

sub ssh_client_rm_known_hosts () {

    my $self = shift;
    my $rslt = 1;
    my $c = connect("$self->{'_client'}#telnet");

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    $dbg ? logcomment("in ssh_client_rm_known_hosts()") : undef;

    my @r = say($c, "cd;rm .ssh/known_hosts*");

    if ( $r[2] =~ /$filesystem_errors/i ) {
        $dbg ? logcomment('WARN', "File Error in ssh_client_rm_known_hosts()" . Dumper(@r)) : undef;
        $rslt = "0";
    }

    disconnect($c);

    $dbg ? logcomment("leaving ssh_client_rm_known_hosts()") : undef;

    return $rslt;
}

=head2 ssh_keygen($keytype)

Generates RSA & DSA keys for public key authentication, or clears existing
keys depending on C<$keytype>.

=over 8

=item C<$keytype = "1">

In this case, C<ssh-keygen> is run with the C<"-t rsa1"> argument to create an
RSA1 key for SSH1 Public Key authentication. The RSA1 key is then concatenated
to the filer's C</etc/sshd/root/.ssh/authorized_keys> file.

=item C<$keytype = "2">

In this case, C<ssh-keygen> is run once with the C<"-t rsa"> argument and once
with the C<"-t dsa"> argument to create both RSA and DSA keys for SSH2 Public
Key authentication. The keys are then concatenated to the filer's
C</etc/sshd/root/.ssh/authorized_keys> file.

=item C<$keytype = "clear">

In this case the filer's C</etc/sshd/root/.ssh/authorized_keys> file will be
removed.

=back


B<Usage:> C<$cfg-E<gt>ssh_keygen($keytype)>

B<Returns:> Boolean, TRUE ('1') if requested operation succeeds, FALSE ('0')
if it does not.

=cut

sub ssh_keygen () {

    # Create key(s) of requested type, default is keys for SSH2,
    # or clear keys from filer if requested

    my $self = shift;
    my $_key_type = shift;
    my $client = $self->{'_client'};
    my $c = connect("$client#telnet");
    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag
    my @_result_array;
    my $rslt = "1";

    $dbg ? logcomment("in ssh_keygen()") : undef;

    $dbg ? logcomment("Who am I?") : undef;
    $dbg ? say($c, "id") : undef;

    my $mount_ok = $self->check_mount_point($c);
    if ( $mount_ok <= 0 ) {  # The mount point is bad.
        $dbg ? logcomment("Failed!!! : not able to sim_mount $self->{'_hostname'}") : undef;
        # sim_umount to clean up if needed.
        $self->_do_sim_umount($c);
        return "0";
    }

    my $filename = "/t/$self->{'_hostname'}/root/etc/sshd/root/.ssh";
    if ( ! $self->_exists_file_or_dir($c, $filename) ) {

        $dbg ? logcomment("$filename does not exist! Will create it.") : undef;

        say($c, "mkdir -p $filename");

        if ( ! $self->_exists_file_or_dir($c, $filename) ) {

            $dbg ? logcomment("Failed!!! :( $filename does not exist.") : undef;

            $rslt = "0";
        }

    }

    # Drop out here if any problems so far
    #
    if ( $rslt eq "0" ) {
        logcomment("Exiting ssh_keygen() early due to failure.");
        # sim_umount to clean up if needed.
        $self->_do_sim_umount($c);
        return "0";
    }


    # Control logic for different key types, default is keys for SSH2,
    # or clear keys from filer if requested

    if ( $_key_type eq "1" ) {

        my @r = say($c, "rm \$HOME/.ssh/identity*");
        push(@_result_array, $r[2]); # Push significant part of result into array

        @r = say($c, "ssh-keygen -q -t rsa1 -b 1024 -f \$HOME/.ssh/identity -N '$self->{_ssh_passphrase}'");
        push(@_result_array, $r[2]);

        @r = say($c, "cat \$HOME/.ssh/identity.pub >> /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/authorized_keys" );
        push(@_result_array, $r[2]);

        $dbg ? logcomment("Generated SSH1(RSA1) key on filer") : undef;

    }

    elsif ( $_key_type eq "2" || $_key_type eq undef ) {

        my @r = say($c, "rm \$HOME/.ssh/id_[r,d]sa*");
        push(@_result_array, $r[2]); # Push significant part of result into array

        @r = say($c, "ssh-keygen -q -t rsa -b 1024 -f \$HOME/.ssh/id_rsa -N '$self->{_ssh_passphrase}'");
        push(@_result_array, $r[2]);

        @r = say($c, "ssh-keygen -q -t dsa -b 1024 -f \$HOME/.ssh/id_dsa -N '$self->{_ssh_passphrase}'");
        push(@_result_array, $r[2]);

        @r = say($c, "cat \$HOME/.ssh/id_rsa.pub >> /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/authorized_keys" );
        push(@_result_array, $r[2]);

        @r = say($c, "cat \$HOME/.ssh/id_dsa.pub >> /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/authorized_keys" );
        push(@_result_array, $r[2]);

        $dbg ? logcomment("Generated SSH2(RSA & DSA) keys on filer") : undef;

    }

    elsif ( $_key_type eq "clear" ) {

        my @r = say($c, "rm /t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/authorized_keys");
        push(@_result_array, $r[2]); # Push significant part of result into array

        $dbg ? logcomment("Removed authorized_keys file from filer") : undef;

        say($c, "rm \$HOME/.ssh/*");
        push(@_result_array, $r[2]);

        $dbg ? logcomment("Removed key files from client config") : undef;

    }

    else {

        $dbg ? logcomment("Unrecognized key type: $_key_type") : undef;

        $rslt = "0"
    }

    if ( $self->_exists_file_or_dir($c, "/t/$self->{'_hostname'}/root/etc/sshd/root/.ssh/authorized_keys" )) {

        $dbg ? logcomment("authorized_keys file exists on filer") : undef;
    }

    foreach ( @_result_array ) {

        if ( $_ =~ /$filesystem_errors/i ) {
            $dbg ? logcomment('WARN', "File Error in ssh_keygen()" . Dumper($_)) : undef;
            $rslt = "0";
        }
    }

    # sim_umount to clean up if needed.
    $self->_do_sim_umount($c);

    disconnect($c);
    $dbg ? logcomment("leaving ssh_keygen()") : undef;
    return $rslt;
}


=head2 _do_sim_umount($client_connection)

Perform a sim_umount if $self->{'_umount'} is set. If a sim_umount is done,
reset  $self->{'_umount'} to undef. Private function.

input:  connection to the client that will be used to do sim_umount.

return: 1=a sim_umount was performed.
        0=no action was taken.

B<Usage:> C<$cfg-E<gt>_do_sim_umount($client_connection)>

=cut

sub _do_sim_umount()
{
    my ($self, $conn) = @_;
    # sim_umount to clean up if needed.
    if ( defined $self->{'_umount'} ) {
        say($conn,$self->{'_umount'}) ;
        $self->{'_umount'} = undef;
        return 1;
    }
    return 0;
}


=head2 _exists_file_or_dir($client_connection, $filename)

Verify whether or not a file or directory exists on the client using "ls
$filename". Private function.

inputs:

    $client_connection: Connection to the client.

    $filename: The full path and filename of the file to check.  This may be a
name of a directory.

return: 1=file or directory exists.
        0=file or directory does not exist.

B<Usage:> C<$cfg-E<gt>_exists_file_or_dir($client_connection, $filename)>

=cut

sub _exists_file_or_dir()
{
    my ($self, $conn, $file) = @_;
    my ($output);

    (undef,undef,$output) = say($conn,"ls $file");
    if ( $output =~ / No such file or directory/ ) {
        return 0;  # file does not exits.
    } else {
        return 1;  # file exists.
    }
}


# Use AUTOLOAD for similar functions which get/set attributes.
# First, define attributes for which AUTOLOAD is valid

{
    my %_attrs = (
        _debug_ssh_verbose      =>  'debug',
        _debug_module_verbose   =>  'debug',
        _ssh_enable             =>  'read/write',
        _ssh_port               =>  'read/write',
        _ssh1_enable            =>  'read/write',
        _ssh2_enable            =>  'read/write',
        _ssh_passwd_auth_enable =>  'read/write',
        _ssh_pubkey_auth_enable =>  'read/write',
        _password               =>  'read/write',
        _filer_ipv6             =>  'read/write',
        _client                 =>  'read',
        _client_name            =>  'read',
        _hostname               =>  'read',
        _prompt                 =>  'read',
        _ssh_access             =>  'read',
        _ssh_idle_timeout       =>  'read',
        _ssh1_hostkeysize       =>  'read',
        _ssh1_serverkeysize     =>  'read',
        _ssh2_hostkeysize       =>  'read',
        _ssh_connect_protocol   =>  'read',
        _ssh_connect_auth_method=>  'read',
        _ssh_alternate_username   =>  'read/write',
        _ssh_alternate_password   =>  'read/write',
        _ssh_alternate_ip         =>  'read/write',
        _ssh_sessions             =>  'read/write',
    );

    sub _accessible {
        my ($self, $attr, $mode) = @_;
        $_attrs{$attr} =~ /$mode/;
    }

    # Extra accessor function to make this hash available elsewhere.
    # Additional logic added so that certain elements of the %_attr hash
    # which are not attributes of the filer configuration(such as
    # debug and connect directives) are not returned.

    sub _get_attributes {

        my %_h;

        foreach my $_key ( keys %_attrs ) {

            my $_value = $_attrs{$_key};

            if ( $_key !~ m/_debug|_connect|hostname|prompt/ ) {

                $_h{$_key} = $_value;
            }
        }

        return %_h;
    }
}

sub AUTOLOAD {

    no strict "refs";
    my ($self, $setting) = @_;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag

    if ( $AUTOLOAD =~ /.*::set(_\w+)/
                    && $self->_accessible($1, 'write' ) ) {
        my $_attr_name = $1;

        # Anonymous subroutine to update object
        my $anon = sub {
            $self->{$_attr_name} = $setting;
            if ($_attr_name !~ /connect/) {
                $self->{'_stale_opt'} = 1;
            }
        };

        if ($setting eq "on" || $setting eq "off") {

            $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

            if ($_attr_name eq "_ssh_enable" || $_attr_name eq "_ssh2_enable") {
                # need to set _ssh_enable and _ssh2_enable to the same values:
                #
                $self->{"_ssh_enable_bogus"} = "" if ($self->{"_ssh_enable_bogus"});
                $self->{"_ssh2_enable_bogus"} = "" if ($self->{"_ssh2_enable_bogus"});
                $_attr_name = "_ssh_enable";
                $anon->();
                $_attr_name = "_ssh2_enable";
                $anon->();
            } else {
                $self->{$_attr_name . "_bogus"} ? # Chg only if exists
                    $self->{$_attr_name . "_bogus"} = "" : undef;

                $anon->();
            }
        }

        elsif ($_attr_name eq "_ssh_port") {

            if  ($setting > 0 && $setting < 65537) {

                $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

                $self->{"_ssh_port"} = $setting;

            } else {

                  logresult("CONFIG", "Invalid value \"$setting\" detected for $_attr_name, will NOT update object");

            }

        }

        elsif ($_attr_name eq "_filer_ipv6") {

            $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

            $self->{"_filer_ipv6"} = $setting;
        }

        elsif ($_attr_name eq "_ssh_alternate_username") {

            $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

            $self->{"_ssh_alternate_username"} = $setting;
        }

        elsif ($_attr_name eq "_ssh_alternate_password") {

            $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

            $self->{"_ssh_alternate_password"} = $setting;
        }

        elsif ($_attr_name eq "_ssh_alternate_ip") {

            $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

            $self->{"_ssh_alternate_ip"} = $setting;
        }

        elsif ($_attr_name eq "_ssh_sessions") {

                 $dbg ? logcomment("Valid value \"$setting\" detected for $_attr_name, will update object") : undef;

                 $self->{"_ssh_sessions"} = $setting;
        }

        else {

            $dbg ? logcomment("Setting flag to indicate bogus setting \"$setting\" for $_attr_name (negative testcase)") : undef;
            if ($_attr_name eq "_ssh_enable" || $_attr_name eq "_ssh2_enable") {
                # need to set _ssh_enable and _ssh2_enable to the same values:
                #
                $self->{"_ssh_enable_bogus"} = 1;
                $self->{"_ssh2_enable_bogus"} = 1;
                $_attr_name = "_ssh_enable";
                $anon->();
                $_attr_name = "_ssh2_enable";
                $anon->();
            } else {
                $self->{$_attr_name . "_bogus"} = 1;
                $anon->();
            }
        }

    return $setting;

    }

    elsif ( $AUTOLOAD =~ /.*::get(_\w+)/ && $self->_accessible($1, 'read' ) ) {

        my $_attr_name = $1;
        return $self->{$_attr_name};

    }

    elsif ( $AUTOLOAD =~ /.*::get(_\w+)/ && $self->_accessible($1, 'debug' ) ) {

        my $_attr_name = $1;
        return $self->{$_attr_name};

    }

    elsif ( $AUTOLOAD =~ /.*::set(_\w+)/ && $self->_accessible($1, 'debug' ) ) {

        my $_attr_name = $1;

        if ( $setting =~ /on|1|true|TRUE|True/ ) {

            $setting = "1";
        }

        else {

            $setting = undef;
        }

        $self->{$_attr_name} = $setting;
        return $self->{$_attr_name};
    }
    elsif ( $AUTOLOAD =~ /DESTROY/ ) {

        return 1;
    }
    else {

        logresult("SCRIPT", "No such attribute : $1");
    }

return 0;
}

#sub DESTROY {
#   my $self = shift;
#
#   logcomment("Configuration Restore started at ".localtime()."\n");
#   my $begtime = time;
#
#   logcomment("DESTROY(called from base class):\n" . Dumper($self) . "ref is: " . ref($self) . "\ncaller is: " . caller());
#
#   # Don't destroy objects we didn't create ourselves
#   caller() =~ m/main/ ? $self->restore_config() : logcomment("Ignoring DESTROY() call from subtest");
#
#
#   my $endtime = time;
#   my $elapsedtime = $endtime - $begtime;
#
#   logcomment("Configuration Restore ended at " . localtime() . " Elapsed time (seconds): $elapsedtime.");
#
#}

=head2 verify_max_ssh_sessions()

This subroutine creates the ssh sessions which is specified by user. This
subroutine is designed for public key based authentication method only. User
need to enter the maximum number of session which is required. Script
takes the filer and client information from the C<get_config()> object.
Username and password is taken from the config object. SSH authentication method
is pubkey. SSH version is set to 2 by deafult unless user specify it. SSH port
is set to 22 by default unless user specify it. If user does not specify any
sessions then function will retunr undef

B<Input:> None

B<Returns:> Number of ssh sessions created on the filer

B<Usage:> C<$cfg-E<gt>verify_max_ssh_sessions()>

    $cfg->verify_max_ssh_sessions()

=cut

sub verify_max_ssh_sessions {
    my $self = shift;
    my $set_ssh_ver;

    $dbg ? $Log->comment("## Entering verify_max_ssh_sessions ") : "";

    $self->set_ssh_connect_auth_method("pubkey");

    if ($self->{'_ssh_connect_protocol'} eq "SSH2") {
        $set_ssh_ver="2";
    } elsif ($self->{'_ssh_connect_protocol'} eq "SSH1") {
        $set_ssh_ver="1";
    } else {
        $dbg ? $Log->comment("## SSH version not specified.Using default ".
        "version as SSH2") : "";
        $set_ssh_ver="2";
    }
    $dbg ? $Log->comment("## ssh_key for $set_ssh_ver") : "";
    my $val=$self->ssh_keygen($set_ssh_ver);
    my $c;

    my $count= $self->{'_ssh_sessions'};

    if(!$count) {
       $Log->comment("SSH sessions is not entered");
       return undef;
    }
    $dbg ? $Log->comment("## ssh sessions are $count") : "";

    if($val) {
        my $out=$self->run_command_over_ssh("help");
        $c=connect("$self->{'_client'}#telnet", -fail, "Telnet connection failed" );

        if($out =~ /hostname/) {
           while($count) {
            say($c,"ssh $self->{'_hostname'} \-$set_ssh_ver -p $self->{'_ssh_port'} " .
             "nfsstat 60 &",'-noprompt');
            sleep(5);
            $count--;
           }
        } else {
           $Log->comment("SSH connection using pubkey is not successful");
           return undef;
        }
        sleep(10);
    } else {
        $dbg ? $Log->comment("## SSH keygen not successful"):
        return undef;
    }
    $Log->comment("Running one more command to check the max session");

    my $client_c=connect("$self->{'_client'}#telnet", -fail, "Telnet connection failed" );
    my @out1=say($client_c,"ssh $self->{'_hostname'} \-$set_ssh_ver -p " .
     "$self->{'_ssh_port'} options ssh",'-timeout',"15");
    disconnect($client_c);
    disconnect($c);
    $dbg ? $Log->comment("## Output $out1[0]:$out1[1]:$out1[2]:$out1[3]") : "";
    return $out1[2];
}


=head1 SEE ALSO

=for html
<b>
<a href="./lib-OntapSSH-SecureAdmin-pm.html">OntapSSH::SecureAdmin</a><br>
<a href="./lib-OntapSSH-Options-pm.html">OntapSSH::Options</a><br>
<a href="./SSH-_sample_wrapper-thpl.html">../SSH/_sample_wrapper.thpl</a><br>
<a href="./SSH-_sample_Options-thpl.html">../SSH/_sample_Options.thpl</a><br>
<a href="./SSH-_sample_SecureAdmin-thpl.html">../SSH/_sample_SecureAdmin.thpl</a><br>
<a href="./SSH-_sample_run_command-thpl.html">../SSH/_sample_run_command.thpl</a><br>
<a href="./SSH-_sample_tests-thpl.html">../SSH/_sample_tests.thpl</a><br>
<b>

=head1 HISTORY

=head1 TODO

=over

=item * Add handling for negative test cases via <attr>_bogus flags

=item * Need check for idle timeout and access function

=item * Implement capability to make SSH connection using passphrase.

=item * TBD after code review

=back

=cut

=head1 COPYRIGHT

Copyright (c) 2006 Network Appliance, Inc.

=cut

1;

