#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Options.pm#2 $

## @summary     Common library to handle option processing
## @author dl-central-libs
## @pod here

=head1 NAME

Options - Library to handle all input params to a test

=head1 SYNOPSIS

    use Options;
    my $options = new Options;
    $options->parameters(definitions=>
    [
        {
            name => "foo",
            description => "the foo variable",
        },
        {
            name => "bar",
            description => "the bar variable with a default of 1",
        },
        {
            name => "a",
            description => "test out dependencies",
            dependencies => [qw(foo)],
        },
        {
            name => "fubar",
            spec => "s",
            description => "you know what this stands for..",
            default => "yes",
        },
        {
            name => "server",
            spec => "s@",
            description => "an array of servers",
        },
    ],
    require_one_of => [["foo", "bar"]]);

    $options->parameters(definitions=>
    [
        {
            name => "testdir",
            description => "a directory that the test will rely on",
            required => "this is a required parameter, must be a directory and writeable",
            file_tests => "dw",
        },
    ]);

    $options->test_cases(definitions => {
        tc1 => {
            TC_description => "This string has a \ in it",
            foo=>1,
            bar=>2,
            fubar=>"3",
        }
    });

    $options->process();

    my ($foo, $bar) = $options->get(option=>[qw(foo bar fubar)]);
    my $servers = $options->get(option=>"FILER");

    my $opts = $options->get_hash(option=>[qw(foo bar server fubar)]);

=head1 AUTHOR

dl-central-libs

=head1 DESCRIPTION

C<Options> implements an object which is used for all option/parameter
handling by both libraries and scripts. Each script or library creates
an Options object and calls the parameters method to actually define the
parameters that will be used by the code. This is simply a definition and
does not actually process the params. The idea is that each piece of code
handles the definitions that it will use. Once all of the parameters have
been defined, the C<process> method is called to actually process the
params that have been defined. The actual values for the parameters are
stored in the object and can be accessed via calls to C<get>.

The object is actually implemented as a "singleton"; there is only ever
one version of the object in the code that gets created by the first
caller. Each successive new call returns this same object. This design
allows for all Options objects to contain the same values for the params.

The following "global" options can be defined when the package is included:

=over

=item Allow_empty_testcases

use Options qw(-Allow_empty_testcases);

This sets C<Allow_empty_testcases> before any processing gets done by any
instances of Options. Setting this flag will cause C<process> to not
complain about empty testcase definitions.

=back

=head1 METHODS

=cut

package Options;
use strict;
# we don't care if Getopt doesn't know about the option, and case matters to us
use Getopt::Long qw(:config pass_through no_ignore_case);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper;
use TracePkg;


# If the single version of this object isn't already set up, go ahead and create it.
my $Object;
if (!$Object){
    # All of the scalar variables that used to be outside of the object are now inside.
    # This is *required* for Allow_empty_testcases and to keep things a little cleaner
    # all of the other scalars were moved in as well.  In order to keep the code a little
    # cleaner, I've left the Hashes and Arrays outside of the object for now but that may
    # change at some point.
    $Object = bless {
        Allow_empty_testcases => 0,
        Exit_after_list => 1,
        Delayed_help => 0,
        ntest => 0,
        Parameter_Prefix => "",
    }, __PACKAGE__;
}
my $Trace;
$Trace = new TracePkg if (!$Trace);

# Decide if we are running inside of NATE.
# We need to detect that we are running inside of ntest so look for a 
#  value of the global $::RUNID.
$Object->{ntest} = 1 if (defined($::RUNID) && $::RUNID);

my %Callers;
my %Defaults = (
                NUFFO_MULTITEST => "true",
                );
my %Test_cases;
my %Options;
my $Options;
my %Dependencies;
my %Comparisons;
my @Require_one_of;
my %Valid_values = ();
my %Subroutines = (
                   );
my %File_tests = (
                  );

my %Compat = (
              help => "HELP",
              );

# SysPrefix and Parameter_Prefix are related to the sys_prefix option (see POD)
my %SysPrefix = (
                );
my %Parameter_Prefix = (
                );
my %Definitions = (TESTCASE=>"s",
                   LISTTESTCASES=>"boolean",
                   LISTTESTPLAN=>"boolean",
                   HELP=>"boolean",
                   NTAP_QA_TRACE=>"s",
                   NUFFO_MULTITEST=>"boolean",
                   LIST_EMPTY_TESTCASES=>"boolean",
                   HOSTRECS=>"s", # this is used internall as a string so do not change to boolean
		   SKIP_EXECUTION=>"boolean",
                  );
my %Descriptions = (TESTCASE=>"the name of a specific testcase to run",
                    LISTTESTCASES=>"list all available testcase names and their descriptions",
                    LISTTESTPLAN=>"list all testcases using the syntax for a .plan file",
                    HELP=>"print this help information",
                    NTAP_QA_TRACE=>"the TraceSpec",
                    NUFFO_MULTITEST=>"perform the necessary steps to make nuffo.thpl work with the script",
                    LIST_EMPTY_TESTCASES=>"If set, allow the listing of empty testcases",
                    HOSTRECS=>"a hash containing all host records, DO NOT set this param",
                    #HOSTRECS is used internal to the options object and is only here to prevent
                    #the slim chance of someone reusing this PARAM name in a thpl
		    SKIP_EXECUTION=>"Cause a test to stop execution after option processing",
                   );
my %Requirements = (
                    );
my %Nate_params;

my %Delimiters;

# This hash will track to origins of the param definitions
my %Origins;

# This following block will find the absolute path to this library for the 
# options defined here
if (! keys %Origins) {
    _get_caller_info();
}

sub _get_caller_info {
    my @caller= caller(0);
    # set the origin of the params defined here as this filename
    $Origins{library}{$caller[1]} = [keys %Definitions];
}

=head2 new

    my $Options = new Options;

C<new> is the invocant for the Options object. The data structures for 
the object are all stored in the class so they share the same information.

It returns the object on success and dies on failure

=cut

sub new {
    $Trace->entry();
    shift;  # Take off the package/object
    my %params = @_;
    while (my ($k, $v) = each %params){
        if (exists($Object->{$k})){
            $Trace->comment(msg => "Changing value of $k from $Object->{$k} to $v");
        } else {
            $Trace->comment(msg => "Defining value of $k as $v");
        }
        $Object->{$k} = $v;
    }
    $Options = \%Options if (!defined($Options));
    $Trace->exit(msg=>"returning $Object");
    return $Object;
} # end new

=head2 parameters

    $Options = $Options->parameters(definitions=>[{},{},...], 
                                    require_one_of=>[[..],[..],..]);

The C<parameters> method provides definitions for the options that should be processed. 
The definitions can also provide further testing that should occur on the option as it
is processed. The call returns nothing and will die on error.

=over

=item definitions

(Required) List of option definitions, specified as hashes. example:

   {
    name => name,
    spec => "s@",
    description => "This is a parameter",
    required => "You must specify this basic parameter",
    default => [],
    subroutine => [\&routine, arg1, arg2,...],
    file_tests => "dw",
    dependencies => [qw(option1 option2)],
    delimiter => ",",
    less => 5,
    greater_equal => 10,
    valid_values => [6, 7, 8, 9],
    sys_prefix => 1,
   }

=over

=item name

(Required) the name for the option

=item spec

(Optional) Specifies the option specifier as defined in Getopt::Long. The 
values that are currently supported are:

=over

=item undef 

signifies a simple scalar value

=item 'boolean' 

signifies that the value is a boolean.
Possible values that equate to true: true, 1, on, yes, defined
Possible values that equate to false: false, 0, off, no, undef

Internally, a true value is represented by 1 and a false value is represented
by undef.

=item 'evaluate' 

=item 'e' 

=item 'e@' 

signifies that the parameter value 
should be evaluated, further described below.
The param should be separated by ':' or ','. Internally, the options is stored
in an anonymous array. If the C<spec> specifies an array, an empty anonymous array
will be created for options that have not been specified.

The 'e' or 'evaluate' specifier is used to enable evaluation of strings to perform evaluation
of the parameter to perform math functions (really could be any perl function)
as well as common size substitutions like kilo, mega, etc.  For example, specifying
'1k-2' as a value would result in the param being set to 1022.  The common substitutions are:
   k => 1024, m => 1024^2, g => 1024^3, t => 1024^4, p => 1024^5, e => 1024^6
   K => 1000, M => 1000^2, G => 1000^3, T => 1000^4, P => 1000^5, E => 1000^6

=item time

Signifies that the value is a unit of time and should be converted to the base number of seconds.
The available time units are 'm' for minutes, 'h' for hours, 'd' for days.

1m will convert into 60
1h will convert into 3600
1d will convert into 86400

=item 's@' 

which signifies that the option is for a list

=back

=item description

(Required) The text that describes the usage of the option. This text is printed
when '-help' is specified.

=item required

(Optional) The text that is printed when the option is required but is not specified.

=item default

(Optional) The default value for the option

=item subroutine

(Optional) This param indicates that a subroutine should be called if this option is
specified. The value of this parameter is a list. The first element is a reference to the
routine to call. The remainder of the list will be passed as params to the routine. The value
of the parameter is prepended to the list prior to calling the routine. 

=item file_tests

(Optional) A list of perl file-test operators to perform on the option if it is specified.
Since tests will run from Windows and unix, the only allowable values are (e d f r w x z s).
The perl file test operator will not be applied when an http is specified via your NATE_LIB
definition.

Example: 'dr' specifies that the option must be a directory and must be readable by the 
caller.

=item dependencies

(Optional) a list of additional options which are required if this option is specified. If
any of these dependencies are also not specified, an error is generated.

=item delimiter

(Optional) This option allows a list to be split on a specific user-defined delimiter. 

Example:

POUND_LIST="abcd:efgh#ijkl:mnop#qrst:uvwx"

$Options->parameters(definitions=>
   [
    {
       name         => "POUND_LIST",
       spec         => "s@",
       default      => undef,
       delimiter    => "#",
       description  => "This is a list I've created to delimit by the pound sign",
    },
]);

$Options->process();

Which yields the indexes of the POUND_LIST array...

20080715 160115 [01_define_spec] T Options::process:1121 POUND_LIST = abcd:efgh
20080715 160115 [01_define_spec] T Options::process:1121 POUND_LIST = ijkl:mnop
20080715 160115 [01_define_spec] T Options::process:1121 POUND_LIST = qrst:uvwx

=item less

(Optional) the option must be less than this value, otherwise an error is generated.

=item greater_equal

(Optional) the option must be greater than or equal to this value, otherwise an
error is generated

=item valid_values

(Optional) a list of valid values for this option. If the option is not one of these
values, an error is generated.

=back

=item require_one_of

(Optional) a list of lists of options, each of which specifies that at least one of its
elements must be present; otherwise, an error is generated.

example:
   require_one_of => [[qw(option1 option2)], [qw(option3 option4)]],

=back

=item sys_prefix

(Optional) if true, not only will a parameter of name "NAME" be retrieved,
but Cx_NAME and Sx_NAME will also be loaded following the same parameter
definition (where x is an integer starting at 1, limited by the parameters
listed below:

    The number of systems are limited by the following parameters:

        CMode - MAX_CMODE_SYSTEMS (default: 4)
        7Mode - MAX_7MODE_SYSTEMS (default: 3)

This facility is being provided so that you can easily specify that you intend a given 
parameter to be processed in the context of a given system.

For example:

    Say you have a test where you want to create NUM_VOLS on 2 different clusters, but you want
    a different number of volumes on each cluster.  You could define:

    definitions => [
        {
            name        => "NUM_VOLS",
            spec        => "s",
            default     => 2,
            description => "Number of volumes to create",
            sys_prefix  => 1,
        },

    Then when you call your script, you could provide the following parameters

    C1_NUM_VOLS=5 C2_NUM_VOLS=10 myscript.thpl
        or 
    NUM_VOLS=5 C2_NUM_VOLS=10 myscript.thpl

    And retrieve the parameter which corresponds to each cluster.  The prefixed
    NUM_VOLS parameters (ex: C2_NUM_VOLS) are processed using the same parameter 
    definition as the non-prefixed NUM_VOLS parameter.

    In the first case, both NUM_VOLS parameters are prefaced by a system id.  In the
    second case (useful when there are many systems), there is a generic NUM_VOLS
    (which it is assumed that all systems other than C2 would use) and a system
    specific C2_NUM_VOLS which only C2 would be expected to use.  Applying these
    parameters to a given system is left to a higher layer library (see lib/SysInit.pm).  
    Options just provides the facility to process these parameters (apply defaults,
    validity checking, etc...)

    (Note: logical system ids can be mapped to clusters or 7Mode systems using the 
           classes in NACL::MTask::Group).

=cut

sub parameters {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
            duplication_ok => {type => SCALAR | UNDEF, optional => 1},
            definitions => {type => ARRAYREF},
            require_one_of => {type => ARRAYREF, optional => 1},
        }
    );

    $Trace->entry(msg=>"duplication_ok='" . 
        (defined $params{duplication_ok} ? $params{duplication_ok} : "undef()") . "'");

    # determine where the parameter is being defined. It will be stored in $file.
    my ($pkg, $file, $line, $subroutine) = caller(0);
    my ($pkg1, $file1, $line1, $subroutine1) = caller(1);
    if ($subroutine1 =~ /TestManager::parameters/) {
        # This is a special case since TestManager.pm has a front-end to this
        #  parameters(). This will cause the original caller of the
        #  TestManager->parameters to be the origin.
        $file = $file1;
    }

    foreach my $param (@{$params{definitions}}) {
        my $name = $$param{name};

        if (!defined($name)) {
            $Trace->fatal(msg=>"A name is missing from the parameter\n", result=>"SCRIPT");
        }
        $Trace->trace(msg=>"defining '$name'");

        # the default value for the spec is "s" so make sure it is there
        if (!$param->{spec}) {
            $param->{spec} = "s";
        }
        # if there is a need to specify type of delimiter used to parse a parameter...grab it
        if (defined $$param{delimiter}) {
            my $delimiter = $$param{delimiter};
            $Delimiters{$name} = $delimiter;
        }

        # if this parameter supports system prefixes
        if (defined $$param{sys_prefix}) {
            my $sys_prefix = $$param{sys_prefix};
            $SysPrefix{$name} = $sys_prefix;
        }

        # determine if the definition already exists
        if ($Definitions{$name}) {
            # this option was already seen. Determine if this is the old way of setting a new default
            if ($Defaults{$name} && $param->{default} && !defined($Options{$name}) &&
                ( !$param->{spec} || ($param->{spec} eq $Definitions{$name}))) {
                my ($package, $file, $line) = caller;
                $Trace->comment(msg=>"current definition for $name is in package '$package' at line $line of $file:".Dumper($param));
                ($package, $file, $line) = @{$Callers{$name}};
                $Trace->comment(msg=>"A previous definition for '$name' was made in package '$package' at line $line of $file\n");
                $Trace->warn(msg=>"It appears as though the code is trying to set a new default for the parameter '$name' and it is not using the new set_default() method. More information is available in the log.");
            } elsif (!$params{duplication_ok} && ($Definitions{$name} ne $param->{spec})) {
                my ($package, $file, $line) = caller;
                $Trace->comment(msg=>"current definition for $name is in package '$package' at line $line of $file:".Dumper($param));
                $Trace->comment(msg=>"previous definition of the spec: $Definitions{$name}");

                my $message;
                if ($Callers{$name}) {
                    ($package, $file, $line) = @{$Callers{$name}};
                    $message = "A previous definition for '$name' with a different specification was made in package '$package' at line $line of $file. More information is available in the log.\n";
                } else {
                    $message = "A previous definition for '$name' with a different specification was made. More information is available in the log.\n";
                }

                # a different spec was found for the two definitions. Determine the severity.
                if (grep(/s@/, ($Definitions{$name}, $param->{spec}))) {
                    # This is bad since one is an array and the other is not.
                    $Trace->fatal(msg => $message, result => "SCRIPT");
                } else {
                    $Trace->warn(msg => $message);
                }
            }
        } else {
            $Definitions{$name} = "$param->{spec}";
            $Trace->debug(msg=>"Definition for '$name' set to '$Definitions{$name}'");

            if (!defined($$param{description})) {
                $Trace->fatal(msg=>"A description is missing for the '$name' parameter\n", result=>"SCRIPT");
            }
        }

        # store the origin of this parameter
        if ($file =~ /\.pm$/) {
            # this is a library so store it separately 
            push (@{$Origins{library}{$file}}, $name);
        } else {
            # this is a script so store it separately 
            push (@{$Origins{script}{$file}}, $name);
        }

        $Callers{$name} = [caller];

        if (exists($$param{required})) {
            $Requirements{$name} = $$param{required};
            $Trace->debug(msg=>"Requirement for '$name' set to '$Requirements{$name}'");
        }
        if (exists($$param{default})) {
            $Defaults{$name} = $$param{default} ;
            my $default = (ref($Defaults{$name}) eq "ARRAY") ? join(",", @{$Defaults{$name}}) : $Defaults{$name};
            if ( defined $default ) {
                $Trace->debug(msg=>"Defaults for '$name' set to '$default'");
            } else {
                $Trace->debug(msg=>"Defaults for '$name' set to 'undef()'");
            }
        }
        if (exists($$param{subroutine})) {
            $Subroutines{$name} = $$param{subroutine};
            $Trace->debug(msg=>"Subroutine for '$name' set to '$Subroutines{$name}'");
        }
        if (exists($$param{description})) {
            $Descriptions{$name} = $$param{description};
            $Trace->debug(msg=>"Description for '$name' set to '$Descriptions{$name}'");
        }
        if (my $tests = $$param{file_tests}) {
            my @operators;
            my ($package, $file, $line) = @{$Callers{$name}};
            foreach my $operator ( split(//, $tests) ) {
                if (!grep($operator eq $_, qw(e d f r w x z s))) {
                    $Trace->warn(msg=>"The file_test operator '$operator' in the option definition of '$name' is not supported at line $line of $file");
                } else {
                    push(@operators, $operator);
                }
            }
            if (scalar(@operators)) {
                $File_tests{$name} = \@operators;
                $Trace->debug(msg=>"File_test for '$name' set to '".join("",@operators)."'");
            }
        }
        if (exists($$param{dependencies})) {
            if (ref($$param{dependencies}) ne "ARRAY") {
                $Dependencies{$name} = [$$param{dependencies}];
            } else {
                $Dependencies{$name} = $$param{dependencies};
            }
            $Trace->debug(msg=>"Dependencies for '$name' set to '".join(" ", $Dependencies{$name})."'");
        }
        if (exists($$param{less})) {
            if ($$param{less} !~ /\d+/) {
                $Trace->fatal(msg=>"A The value for the 'less' parameter must be a number for '$name'\n", result=>"SCRIPT");
            }
            $Comparisons{less}{$name} = $$param{less};
            $Trace->debug(msg=>"Comparisons for '$name' set to '$Comparisons{less}{$name}'");
        }
        if (exists($$param{greater_equal})) {
            if ($$param{greater_equal} !~ /\d+/) {
                $Trace->fatal(msg=>"A The value for the 'greater_equal' parameter must be a number for '$name'\n", result=>"SCRIPT");
            }
            $Comparisons{greater_equal}{$name} = $$param{greater_equal};
            $Trace->debug(msg=>"greater_equal for '$name' set to '$Comparisons{greater_equal}{$name}'");
        }
        if (exists($$param{valid_values})) {
            if (ref($$param{valid_values}) ne "ARRAY") {
                $Valid_values{$name} = [$$param{valid_values}];
            } else {
                $Valid_values{$name} = $$param{valid_values};
            }
            $Trace->debug(msg=>"valid_values for '$name' set to '$Valid_values{$name}'");
        }
    }
    if (exists($params{require_one_of})) {
        foreach my $requirement (@{$params{require_one_of}}) {
            if (ref($requirement) ne "ARRAY") {
                $Trace->fatal(msg=>"A The 'require_one_of' parameter must be an array ref of array refs\n", result=>"SCRIPT");
            } else {
                push(@Require_one_of, $requirement);
            }
            $Trace->debug(msg=>"Require_one_of appended with [".join(" ", @{$params{require_one_of}})."]");
        }
    }
    $Trace->exit();
}

=head2 test_cases

    $Options->test_cases(definitions=>
        {
         "testcase1" => {
             TC_description => "this is testcase1 which sets the values of 'foo' and 'bar' to 1",
             foo => 1,
             bar => 1,
         },
         "testcase2" => {
             TC_description => "this is testcase2 which sets the values of 'foo' and 'bar' to 2",
             foo => 2,
             bar => 2,
         },
        },
        default=>"testcase1");

C<test_cases> specifies a list of hashes which define test cases. A test case
has a name and a list of parameter assignments for that test case. The idea is that
if the script has multiple parameters and a test case is specified as a combination of
those params with certain values, then name the test case and set the values. The test case
can then be executed by passing the TESTCASE=testcase param.

=over

=item definitions

(Required) This is the hash of hashes that define the test cases. The hash key is the name
of the testcase. The value is another hash. This hash specifies the names of the parameters 
and the values they should be set to. In addition, the following keys may be specified:

=over

=item TC_description

(Optional) The value of this key is a description of the testcase. This is used for storing nuffo
results.

=back

=item default

(Optional) This param specifies the name of the default C<TESTCASE> to use for the test.

=back

=cut

sub test_cases {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
            definitions => {type => HASHREF},
            default => {type => SCALAR, optional => 1},
            exit_after_list => {type => SCALAR | UNDEF, optional => 1},
            start_testcase => {type => SCALAR | UNDEF, optional => 1, default => 1},
        }
    );

    $Trace->entry(msg=>"exit_after_list=$params{exit_after_list}\n");

    $self->{testcase_start} = $params{start_testcase};

    if ($self->{ntest} && $self->{testcase_start}) {
        eval "use Burt qw(testcase_start);";
        die $@ if ($@);
    }

    %Test_cases = %{$params{definitions}} if %{$params{definitions}} != ();
    if (exists($params{exit_after_list})) {
        $self->{Exit_after_list} = $params{exit_after_list};
    }

    my $badTestCase = undef;
    # look for illegal characters in the testcase name. nip and other tools won't like them
    foreach my $tc (keys %Test_cases) {
        $Trace->debug(msg=>"defining testcase '$tc'");
        foreach my $illegal (qw(- /)) {
            my $qillegal = quotemeta($illegal);
            if ($tc =~ /$qillegal/) {
                 $badTestCase .= "encountered an illegal character '$illegal' in TestCase name '$tc'\n";
            }
        }
    }
    $Trace->fatal(msg=>$badTestCase, result=>"SCRIPT") if ($badTestCase);

    if (defined($Options->{TESTCASE})) {
        $Trace->debug(msg=>"TESTCASE already set to $Options->{TESTCASE}");
    }
    if ($params{default}) {
        if (!exists($Test_cases{$params{default}})) {
            $Trace->fatal(msg=>"the default testcase '$params{default}' is not a defined testcase", result=>"SCRIPT");
        }
        if ($self->{Default_testcase} && ($self->{Default_testcase} ne $params{default})) {
            $Trace->warn(msg=>"a default testcase was defined in test_cases() as $params{default} and as a default option($self->{Default_testcase}). Setting default to $params{default}");
        }
        $Trace->debug(msg=>"storing TESTCASE as $params{default}");
        # TESTCASE is a very special internally-used param. Store the value of the default so that it
        # can be checked in the proper code locations. Setting it in %$Options prevents the nate_params
        # hash from being updated properly in process(). This, in turn, will cause process_test_cases()
        # not to work properly. The default value in Options will be set in process_test_cases().
        if (!defined($Options->{TESTCASE})) {
            $self->{testcase} = $params{default};
        }
        $self->{default_testcase} = $params{default};
    }
    
    $Trace->exit();
}

=head2 compatability

    $Options->compatability(definitions=>{param=>compatable_param});

C<compatability> specifies a list of params and their compatable params. For example,

  $Options->compatability(definitions=>{ 
                                        FILER => "NODES" 
                                       });

indicates that the param C<NODE> is now an acceptable NATE parameter, but internally
the C<FILER> param will be set to the value specfied for C<NODES> as well. As a result,
the following two commands will both set C<FILER> to C<filer1> internally to the object.

  ntest NODES=filer1 test.thpl
  ntest FILER=filer1 test.thpl

test.thpl will internally access C<NODES> to obtain the value C<filer1>.

Values for the normal parameter override any values specified by the compatible option.
So passing "NODES=filer1 FILER=filer2" will cause NODES to internally be set to filer1.

=cut

sub compatability {
    my $self = shift;
    my %params = @_;

    $Trace->entry();
    if (!defined($params{definitions})) {
        $Trace->fatal(msg=>"The definitions parameter is required\n", result=>"SCRIPT");
    }
    if (ref($params{definitions}) ne "HASH") {
        $Trace->fatal(msg=>"The definitions parameter is not a HASH reference\n", result=>"SCRIPT");
    }
    while (my ($key, $value) = each  %{$params{definitions}}) {
        $Trace->debug(msg=>"$key = $value");
        $Compat{$key} = $value;
    }
    $Trace->exit();
}

=head2 process

    $Options->process();

C<process> is the method that will actually examine all sources of
options and store the values internally. Once stored, the options
will be available via the C<get> and C<get_hash> methods. The C<process>
method can be called multiple times and has to be called at least once
before options are available internally. When called, it will only
process the options which have been defined via C<parameters> and 
C<compatability>. Processing of the options occurs as follows:

  Look for values stored as defaults in the definition of the options.
  Look for values stored in the hostrec for the given FILER.
  Look for values stored in the testcase hash if the C<TESTCASE> param 
     has been set.
  Look for values specified as NATE params (using calls to param).
  Look for values specified as script options (the '--option' format).
     This processing is not normally used while running inside of NATE.

If hostrec is specified then C<process> will also look for the param
C<DATA_IP1> and set the param C<default_data_interface> to its value.
If C<DATA_IP1> is not found then it will set C<default_data_interface>
to the default_ip as specified in the host record.

All values for the known options will be stored in NATE's params.

In the cases of options that have been specified as spec C<s@>, process 
will insure that the internal option is stored as an array. It will
create an empty array if the option does not have a value.

After all options have been processed, C<process> will then follow all 
of the other directives as specified in the options definitions:

  Execute any subroutines as specified by C<subroutine>
  Check for requirements as specified by C<required> and C<require_one_of>
  Check for dependencies as specified by C<dependencies>
  Check for delimiters as specified by C<delimiter>
  Check for the comparisons as specified by C<less> and C<greater_equal>
  Perform the perl file test operators as specified by C<file_tests>
  Check for valid values of the options as specified by C<valid_values>

=over

=item test_cases

(Optional) Controls whether or not C<TESTCASE> is processed. This is on by default
and will cause process() to look at the test cases that were defined by
C<test_cases> and load the params stored there if C<TESTCASE> is found.

=item hostrec

(Optional) Controls whether or not params will be processed out of the 
nate_hosts file for each filer that has been specified. The param must
pass in a reference to an array of node names that should be loaded.

=back

=cut

# I am not ready to document run_help or sfopairs because I do not know that
# they should really be exposed to everyone
sub process {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
         test_cases => {type => SCALAR | UNDEF, default => 1}, # should the test cases be processed?
         run_help => {type => SCALAR | UNDEF, default => undef}, # should the help be processed?
         sfopairs => {type => SCALAR | UNDEF, default => undef},
         hostrec => {type => ARRAYREF | SCALAR | UNDEF, default => undef}, # should options be loaded from hostrec
        }
    );
    $Trace->entry();
    # when running inside of nate, $0 is set to 'full script options' so
    #  we need to first split this by spaces and split the first element
    #  (which is the actual script) by the '/' character and take the
    #  last element which will leave us with the real script name
    $self->{Full_script} = (split(/\s+/, $0))[0];
    my @program = split("/", $self->{Full_script});
    $self->{Program} = pop @program;
    $self->{ScriptName} = $self->{Program};
    $self->{ScriptName} =~ s/\.thpl//;

    my %cl_options;
    my @getopt;
    my @eval_options;
    my @booleans;
    my @time_options;
    $Trace->trace(msg=>"Setting up eval params");

    # examine the specs and save the options to the appropriate lists for
    #  Getopt::Long::GetOptions, booleans and eval_options
    # and in a list 
    foreach (keys %Definitions){
        my $spec = $Definitions{$_};
        if ($spec =~ /evaluate|^e$/){
            push @eval_options, $_;
            $spec =~ s/evaluate|^e$/s/;
        } elsif ($spec =~ /boolean/){
            push @booleans, $_;
            $spec =~ s/boolean/s/;
        } elsif ($spec =~ /time/){
            push @time_options, $_;
            $spec =~ s/time/s/;
        }
        if ($spec eq "!") {
            push(@getopt, "$_$spec");
        } else {
            push(@getopt, "$_=$spec");
        }
    }
    
    $Trace->debug(msg=>"eval_options = " . join(', ', @eval_options));
    $Trace->debug(msg=>"booleans = " . join(', ', @booleans));
    $Trace->debug(msg=>"time_options = " . join(', ', @time_options));
    $Trace->debug(msg=>"getopt = " . join(', ', @getopt));

    ##################################################################################################
    #
    # Find the values of the given parameters from the command line
    ##################################################################################################
    $Trace->debug(msg=>"calling GetOptions: " . Dumper \@getopt);
    my $rc = Getopt::Long::GetOptions(\%cl_options,
                                      @getopt,
                                      );
    if (!$rc) {
        $Trace->fatal(msg=>"Unknown options found by Getopt::Long::GetOptions\n", result=>"SCRIPT");
    }

    ##################################################################################################
    #
    # Find the values of the given parameters from the environment/NATE
    #
    # Nate parameters are normally in the ENV; however, calls to param prior to calling this method can
    #  change the value of the parameters (internal to NATE) without changing the value in the ENV. For
    #  this reason, the values should be found via a call to param rather than looking in %ENV.
    ##################################################################################################
    
    # create a reverse mapping of the compatible parameters. This is needed since the mapping provided
    # by the scripts are keyed on the old name
    my %compat;
    while (my ($key, $value) = each %Compat) {
        $compat{$value} = [] if (!exists($compat{$value}));
        push(@{$compat{$value}}, $key)
    }
    
    foreach my $option (keys %Definitions) {
        # see if this option has any compatible options names to deal with
        my @options = ($compat{$option}) ? @{$compat{$option}} : ();
        push(@options, $option);
        $Trace->debug(msg => "Look for value of $option from any of:".join(" ", @options));

        foreach my $check_option (@options) {
            # If we are running inside of nate then find the value of the option via param(), 
            #  otherwise just grab it from the %ENV (outside of NATE)
            if ($self->{ntest}) {
                # Look for the specific script version of the option first
                # This is strange but param() will return '' rather than undef if it was not defined
                my $value_script = main::param( $self->{ScriptName} . "_" . $check_option );
                my $value = main::param( $check_option );
                $Trace->debug(msg => "Option '$option', value = '" . 
                    ( defined $value ? $value : "undef()" ) . "', script value = '" .
                    ( defined $value_script ? $value_script : "undef()") . "'");
                if (defined($value_script) && $value_script ne "") {
                    $Nate_params{$option} = $value_script;
                    $Trace->debug(msg=>"storing $option => $value_script as a NATE param (from value of $check_option)");
                } elsif (defined($value) && $value ne "") { 
                    $Nate_params{$option} = $value;
                    $Trace->debug(msg=>"storing $option => $value as a NATE param (from value of $check_option)");
                }
 
                if ( $SysPrefix{$check_option} ) {
                    my %system_limit;
                    # Allow max number of systems to be overridden
                    $system_limit{C} = $self->get_max_cmode_systems();
                    $system_limit{S} = $self->get_max_7mode_systems();
                    foreach my $prefix ( "C", "S" ) {
                        my $index = 1;
                        my $done = 0;

                        while ( !$done ) {

                            my $found = 1;
                            my $full_preface = $prefix . $index . "_" ;
                            my $prefixed_option = $full_preface . $check_option;

                            # Look for the specific script version of the option first
                            # This is strange but param() will return '' rather than undef if it was not defined
                            my $value_script = main::param( $self->{ScriptName} . "_" . $prefixed_option );
                            my $value = main::param( $prefixed_option );
                            $Trace->debug(msg => "Prefixed Option '$prefixed_option', value = '" . 
                                ( defined $value ? $value : "undef()" ) . "', script value = '" .
                                ( defined $value_script ? $value_script : "undef()") . "'");
                            if (defined($value_script) && $value_script ne "") {
                                $Nate_params{$prefixed_option} = $value_script;
                                $Trace->debug(msg=>"storing $option => $value_script as a NATE param (from script value of $prefixed_option)");
                            } elsif (defined($value) && $value ne "") { 
                                $Nate_params{$prefixed_option} = $value;
                                $Trace->debug(msg=>"storing $option => $value as a NATE param (from value of $prefixed_option)");
                            } else {
                                $found = 0;
                            }

                            # If there is a known option (say OPTION) that is prefixed by a system 
                            # prefix (say C2_OPTION) clone the definitions from
                            # OPTION to a new entry for C2_OPTION
                            if ( $found ) {
                                $Definitions{$prefixed_option}      = $Definitions{$option};
                                $Parameter_Prefix{$prefixed_option} = $full_preface;

                                if ( $Subroutines{$option} ) {
                                    $Subroutines{$prefixed_option}  = $Subroutines{$option};
                                }
                                if ( $Dependencies{$option} ) {
                                    $Dependencies{$prefixed_option} = $Dependencies{$option};
                                }
                                if ( $Descriptions{$option} ) {
                                    $Descriptions{$prefixed_option} = $Descriptions{$option};
                                }
                                if ( $Requirements{$option} ) {
                                    $Requirements{$prefixed_option} = $Requirements{$option};
                                }
                                if ( $Delimiters{$option} ) {
                                    $Delimiters{$prefixed_option}   = $Delimiters{$option};
                                }
                                if ( $Origins{$option} ) {
                                    $Origins{$prefixed_option}      = $Origins{$option};
                                }
                                if ( $Valid_values{$option} ) {
                                    $Valid_values{$prefixed_option} = $Valid_values{$option};
                                }
                            }

                            $index++;

                            if ( $index > $system_limit{$prefix} ) {
                                $done = 1;
                            }
                        }
                    }
                }
            } else {
                if ( exists $ENV{$option} ) {
                    $Trace->debug(msg=>"storing $option => $ENV{$option} as a NATE param");
                    $Nate_params{$option} = $ENV{$option};
                }
            }
        }
    }

    # if Options has been created with the Delayed_help flag then help() and requirement
    #  checking should not occur until the caller says it is ready
    if ( ( ($self->{Delayed_help} && $params{run_help}) || !$self->{Delayed_help} )
         && ( $cl_options{HELP} || $Nate_params{HELP} )) {
        $self->_help();
    }

    my %option_arrays; # a convenience hash for tracking which options are supposed to be arrays
    my %option_definitions; # a convenience hash for checking for the existence of an option definition

    # fill on our convenience array: if the key exists, then it is a valid option
    while (my ($key, $type) = each %Definitions) {
        $option_definitions{$key}=1;
        # now check if this is supposed to be an array
        if ($type && $type =~ /\@/) {
            $option_arrays{$key}=1; # a sign that this key is an array
            $Trace->debug(msg=>"defining '$key' as an array");
        }
    }
 
    # if the delimiter parameter is used on an option that is not defined as an array throw a WARN
    foreach my $delimiter_key (keys %Delimiters) {
        if ( !defined $option_arrays{$delimiter_key} ) {
            $Trace->warn(msg=>"Option $delimiter_key is not defined as an array, or list. Using a delimiter parameter with this option is incorrect.");
        }
    }
    
    # process testcases unless otherwise told not to
    my $tc_options;
    if ($params{test_cases}) {
        if (defined($tc_options = $self->_process_test_cases(\%cl_options, \%Nate_params, $Options))){
            $Trace->trace(msg => "Allow_empty_testcases = $self->{Allow_empty_testcases}");
            if (($Options->{TESTCASE} ne $self->{default_testcase}) # Not the default testcase
                && !(grep(!/TC_description/, keys %{$tc_options})) # Doesn't have any keys other than TC_description
                && !$self->{Allow_empty_testcases}){ # And we're not supposed to allow empty testcases
                $Trace->fatal(msg => "No useful options defined in testcase $Options->{TESTCASE}", result => "SCRIPT");
            }
        }
    }

    my $host_object;
    # if the filer is required and we are not being told to block the help routine
    #  then pull in the values from the filer. If Delayed_help is active, then 
    #  don't process FILER until the run_help is enabled. This will prevent FILER
    #  errors when HELP is true.
    if ($self->{ntest} && $params{hostrec} &&
        ( ($self->{Delayed_help} && $params{run_help}) || !$self->{Delayed_help})) {
        
        # FILER used to be defined here in Options and was a required parameter. This
        # is now moved over to Server.pm since it is the main consumer of the param.
        # Some scripts still rely on the hostrec processing and will not have the list
        # of nodes yet so detect that and provide a workaround.
        my @filers = ();
        if (ref($params{hostrec})) { # already checked via validate()
            @filers = @{$params{hostrec}};
        } elsif ($Nate_params{NODES}) {
            @filers = split(/[,:]/, $Nate_params{NODES});
        } else {
            # should we fatal here?
        }
        my %sfo_pairs;
        foreach my $hostrec (@filers) {
            my $node_object = main::host($hostrec);
            my $node_name = $node_object->hostname();
            $host_object->{HOSTRECS}{$node_name} = $node_object; #create a key for each filer's host_rec
        }            
        if ($params{sfopairs}) {  # only grab the sfopair info if the test calls for it
            foreach my $node (keys %{$host_object->{HOSTRECS}}) {
                push @{$sfo_pairs{$host_object->{HOSTRECS}{$node}{SFOPAIR}}}, $node; #create a tmp % for the sfo pairs associated with FILERS
                $Trace->debug(msg=>"$node:SFOPAIR: $host_object->{HOSTRECS}{$node}{SFOPAIR}");
                if (scalar(@{$sfo_pairs{$host_object->{HOSTRECS}{$node}{SFOPAIR}}}) > 2) {
                        my $num = scalar(@{$sfo_pairs{$host_object->{HOSTRECS}{$node}{SFOPAIR}}});
                        $Trace->trace(msg=>"SFOPAIR:\n'$num':'@{$sfo_pairs{$host_object->{HOSTRECS}{$node}{SFOPAIR}}}'");
                        $Trace->trace(msg=>"SFOPAIR:\n'$host_object->{HOSTRECS}{$node}{SFOPAIR}'");
                        $Trace->fatal(msg=>"There are greather then 2 nodes defined in the same SFO PAIR!\n".
                                      "Please confirm that SFOPAIR is defined in the nate_host \n".
                                      "file of all FILERs.",
                                      result=>"CONFIG");
                }
            }
            $Trace->debug(msg=>"\%sfo_pairs hash data dumper:\n".Dumper(%sfo_pairs));
            my $i=0;
            foreach my $key (keys %sfo_pairs) {                            #put the SFO pairs in the $host_object
                my $value = $sfo_pairs{$key};                              #SFOPAIRS=>[[node1,node2],[node3,node4]...]
                $host_object->{SFOPAIRS}[$i] = $value;
                ++$i;
            }
            $Trace->debug(msg=>"SFOPAIRS:\n".Dumper($host_object->{SFOPAIRS}));
        }
        $Trace->debug(msg=>"\$host_object hash data dumper:\n".Dumper($host_object));
    } else {
        $Trace->trace(msg=>"Not calling host: Nate = '$self->{ntest}' Hostrec = '" .
            (defined $params{hostrec} ? $params{hostrec} : "undef()") . "'");
    }

    # merge our default options, environment and command-line options
    $self->_process_options(options_hash=>\%Defaults, text=>"Defaults", override=>0);
    $self->_process_options(options_hash=>$tc_options, text=>"TestCase");
    $self->_process_options(options_hash=>\%Nate_params, text=>"Params");
    $self->_process_options(options_hash=>\%cl_options, text=>"Options");

    # In order to have parameters defined by a testcase to be automatically passed via 
    # a call to subtest, the paramter value from the testcase had to be set in NATE via
    # a call to param(). This cannot be just blindly done because it could override the
    # value passed in via the command line. So, only call param(-set) if there is not
    # a value defined previously by nate. (it's value would have been stored in %Nate_params.
    while (my ($option, $value) = each %$tc_options) {
        if (!defined($Nate_params{$option})) {
            $Trace->debug(msg=>"Setting the NATE param: param($option, -set, $value)");
            main::param($option, '-set', $value);
        }
    }

    if ($host_object && $params{hostrec}) {
        # Now bring in any other generic values that were stored in the host object.
        foreach my $key (keys %$host_object){
            next if ($key ne uc $key); # only process params that are all CAPS
            $Trace->debug(msg=>"auto-setting $key to " . $host_object->{$key});
            # parse the values of Compat first and then override with the normal values
            if (my $compat_key=$Compat{$key}) {
                $Options->{$compat_key} = $host_object->{$key};
                $Trace->debug(msg=>"setting $compat_key to $host_object->{$key} ($compat_key was translated from $key)");
            }
            $Options->{$key} = $host_object->{$key};
        }
    }

    # I don't like having this special option in this module but the default_ip is different between Classic and NG.
    #  This is the only common module used by all code so it pretty much has to be here. For now, the user cannot
    #  specify this option; it is only found in the hostspec. Scripts should only rely on this option if they don't
    #  get a data_interfaces param.

    #  If Delayed_help is active, then don't process interfaces until the run_help is enabled. 
    #    This will prevent interface errors when HELP is true.
    if ($self->{ntest} && $host_object && 
        ( ($self->{Delayed_help} && $params{run_help}) || !$self->{Delayed_help})) {
        my $source;
        my $first_node = (keys(%{$host_object->{HOSTRECS}}))[0];
        $Trace->trace(msg=>"NODE: $first_node");
        if ($host_object->{HOSTRECS}{$first_node}{"DATA_IP1"}){
            $Options->{default_data_interface} = $host_object->{HOSTRECS}{$first_node}{"DATA_IP1"};
            $source = "DATA_IP1";
        } elsif ($host_object->{HOSTRECS}{$first_node}{"default_ip"}) {
            $Options->{default_data_interface} = $host_object->{HOSTRECS}{$first_node}{"default_ip"};
            $source = "default_ip";
        } elsif (!defined($Options->{default_data_interface})) {
            $Trace->fatal(result=>"SCRIPT", msg=>"Did not set a default_data_interface param\n".Dumper($Options));
        }
        $Trace->debug(msg=>"auto-setting default_data_interface to " . $Options->{default_data_interface} . " from $source of $first_node.");
    }

    # go through the list of array-type options and split out comma-separated values
    foreach my $key (keys %option_arrays) {
        $Trace->trace(msg=>"processing option '$key' as an array");
        my @values = ();
        # make sure this is an array in case someone wants to push onto it
        if (!defined($Options->{$key})) {
            $Options->{$key} = [];
            next;
        } elsif (ref($Options->{$key}) ne "ARRAY") {
            $Options->{$key} = [$Options->{$key}];
        }

        foreach my $value (@{$Options->{$key}}) {
            my $delimiter;
            if (defined $Delimiters{$key}) {
                $delimiter = $Delimiters{$key};
            } else {
            $delimiter = ",:";
            }
            foreach my $value (split(/[$delimiter]/, $value)) {
                push(@values, $value);
            }
        }
        $Options->{$key} = \@values;
        $Trace->debug(msg=>"options '$key' set to [".join(" ", @values)."]");
    }

    foreach my $option (keys %Subroutines) {
        next if (!$Options->{$option});
        my ($routine, @args) = @{$Subroutines{$option}};
        $Trace->trace(msg=>"call '$routine($Options->{$option})' for option '$option'");

        # If this option is one that is prefixed by a system id (i.e. C1, C2, S1, etc...)
        # we need to set the Parameter_Prefix so that any sets in the
        # subroutine will set the correct Parameter
        if ( $Parameter_Prefix{$option} ) {
            $self->{Parameter_Prefix} = $Parameter_Prefix{$option};
        }

        &$routine(@args, $Options->{$option});

        $self->{Parameter_Prefix} = "";
    }

    # booleans must be dealt with before requirement so that anything like "false" is turned
    #  into undef()
    $Trace->trace(msg=>"Convert the boolean values: @booleans");
    foreach my $option (@booleans) {
        next if (!exists($Options->{$option}));
        if (ref($Options->{$option}) eq "ARRAY"){
            $self->convert_boolean($option, @{$Options->{$option}});
        } else {
            $self->convert_boolean($option, $Options->{$option});
        }
    }

    #  If Delayed_help is active, then don't process these until the run_help is enabled. 
    #    This will prevent errors when HELP is true.
    if ($self->{Exit_after_list} ||
        ($self->{Delayed_help} && $params{run_help}) ||
        !$self->{Delayed_help}) {
        $self->_check_requirements();
        $self->_check_dependencies();
        _check_comparisons();
        _file_test();
        _check_valid_values();
    }

    foreach (@eval_options){
        # Go through the options we're supposed to eval and do the substitutions
        if ( !defined $Options->{$_} ) {
            $Trace->trace(msg => "Skipping evaluation of undef option '$_'");
            next;
        } else {
            $Trace->trace(msg=>"Evaluating: '$_' -> '$Options->{$_}'"); 
        }
        
        if (ref($Options->{$_}) eq "ARRAY"){
            $self->unit_to_base(@{$Options->{$_}});
        } else {
            $self->unit_to_base($Options->{$_});
        }
    }

    foreach (@time_options){
        # Go through the options which need to have a time conversion
        if ( !defined $Options->{$_} ) {
            $Trace->trace(msg => "Skipping time conversion of undef option '$_'");
            next;
        } else {
            $Trace->trace(msg=>"Converting time for '$_' -> '$Options->{$_}'"); 
        }
        
        if (ref($Options->{$_}) eq "ARRAY"){
            $self->_normalize_time($_, @{$Options->{$_}});
        } else {
            $self->_normalize_time($_, $Options->{$_});
        }
    }

    # Echo the parameters that were passed.
    foreach my $key (keys %$Options) {
        next if (!exists($Options->{$key}));
        if ( ref($Options->{$key}) eq "ARRAY" ) {
            foreach my $member (@{$Options->{$key}}) {
                $Trace->trace(msg=>"$key = $member\n");
            }
        } else {
            $Trace->trace(msg=>"$key = '" . 
                (defined $Options->{$key} ? $Options->{$key} : "undef()'") . "'");
        }
    }

    if ($self->{ntest}) {
        # Code used to exist here to take the value that was found via Options and
        # to set the NATE param value to be the same. But that code was causing problems
        # when a script was running a mixture of code which relied on both Options and param()
        # and the param was used differently. Specifically, NOTE used VOL_SIZE and assumed it
        # was defined in KB whereas Params::Volume assumed it was in bytes.
        
        # Now that all of the options have been processed, check to see if a TESTCASE has been defined.
        # If so, then log the param result if necessary.
        my $testcase = $Options->{TESTCASE};
        if ($testcase && $Test_cases{$testcase} && $self->{testcase_start}){
            # call testcase_start() so that nuffo.thpl will properly find the testcase
            # name and description
            Burt::testcase_start($testcase, ${$Test_cases{$testcase}}{TC_description});
            $self->{testcase_start} = undef;

            if ($Options->{NUFFO_MULTITEST} && !$self->{nuffo_multitest}) {
                main::logresult("PARAM", "NUFFO_MULTITEST_LOGS=yes");
                # call logresult to set the NUFFO_MULTITEST_LOGS param for nuffo.thpl
                $self->{nuffo_multitest} = 1;
            }
        }
        if ($Nate_params{'SKIP_EXECUTION'}){
	    $Trace->comment(msg => "Script exiting since SKIP_EXECUTION is set");
	    exit();
	}
    }

    $self->{Processed} = 1;
    $Trace->exit();
}

=head2 defined

    $Options->defined(option => option_name);

C<defined> returns 1 if the provided option is defined. It returns undef
otherwise. For options that have been defined as arrays (spec = C<s@>),
the method will return 1 only if the array is not empty.

=cut

sub defined {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
            option => {type => SCALAR},
        }
    );
    my $key = $params{option};

    if (exists($Options->{$key})) {
        if (ref($Options->{$key}) eq "ARRAY") {
            # this is an array. If it is empty, return undef
            # otherwise return the array
            return (scalar @{$Options->{$key}}) ? Dumper($Options->{$key}) : undef;
        } else {
            # it is defined but not an array so return its value
            return defined($Options->{$key});
        }
    } else {
        # it is not defined
        return undef;
    }
}

# this is old code for use by SpinTestTool
sub define {
    my $self = shift;
    my %params = @_;

    $Trace->entry();
    if ($params{options}) {
        if (ref($params{options}) ne "HASH") {
            $Trace->fatal(msg=>"parameter 'options' is required to be a HASH ref: $params{options}\n", result=>"SCRIPT");
        }
        $Options = $params{options};
        # save a default testcase for backwards-compatability
        $self->{Default_testcase} = $Options->{TESTCASE};
    }
    $Trace->exit();
}

=head2 get

    $option1 = $Options->get(option=>"option1");
    ($option1, $option2) = $Options->get(option=>[qw(option1 option2)]);
    $testcase_hash = $Options->get(testcase=>"testcase_name");

C<get> is used to obtain the value for the specified options[s]. If a 
reference to a list is passed in, then values for each of the options
specified in the list are returned in the same order. Options which 
are stored as arrays will return a reference to the array.

=over

=item option

(Optional) The name of the option. This can also be a reference to a list
of options.
If C<option> is a scalar, then the value of the option is returned. If the
option is an array, then a reference to the array is returned. 

If C<option> is an array ref, then it contains a list of options to get. In
this case, an array is returned with the values for the specified options.
The list is in the same order as the options. If any of the options are 
arrays, then the returned value is an array ref.

=item testcase

(Optional) the name of the testcase.
If C<testcase> is specified then a reference to the hash of options/values
is returned.

=back

If neither C<option> or C<testcase> are specified, then undef is returned.

=cut

sub get {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
            option => {type => SCALAR | ARRAYREF | UNDEF, optional => 1},
            testcase => {type => SCALAR | UNDEF, optional => 1},
        }
    );

    $Trace->entry(msg=>"option='" . (defined $params{option} ? $params{option} : "undef()") . 
        "',testcase='" . (defined $params{testcase} ? $params{testcase} : "undef()") . "'");

    my @values;
    if (defined($params{testcase})) {
        $Trace->exit(msg=>"return hash for $params{testcase}");
        return $Test_cases{$params{testcase}};
    } elsif (defined($params{option})) {
        my @options = (ref($params{option}) eq "ARRAY") ? @{$params{option}} : ($params{option});
        $Trace->debug(msg=>"option=" . join(" ", @options));
        foreach my $option (@options) {
            push(@values, $Options->{$option});
        }
        if ( scalar(@values) >= 1 ) {
            $Trace->exit(msg=>"return " . join(" ", @values));
        } else {
            $Trace->exit(msg=>"return undef()");
        }
        return (wantarray) ? @values : $values[0];
    } else {
        return undef;
    }
}

=head2 set

    $Options->set(option1=>"value1", option2=>"value2",...);

C<set> is used to set values for options. The options do not have to
be predefined via calls to C<parameters>. If the value is specified as
undef, then the corresponding entry for option in the internal hash is 
deleted.

This method is not currently setting the value in NATE's params.

=cut

sub set {
    my $self = shift;
    my %params = (@_);
    my $parameter_prefix = $self->{Parameter_Prefix};

    $Trace->entry(msg=>"params: " . Dumper(\%params));

    if ( $parameter_prefix ) {
        my %prefixed_params;
        foreach my $key (keys %params){
            $prefixed_params{$parameter_prefix.$key} = $params{$key};
        }
        %params = %prefixed_params;
    }

    foreach my $key (keys %params){
        if (!defined($params{$key})){
            $Trace->debug(msg=>"Unsetting option $key, value was $Options->{$key}");
            # the option is being set to undef so delete it from the current Options hash
            delete($Options->{$key});

            if ($self->{ntest}) {
              # now unset it inside of nate so that it won't pass it inside of a subtest
              main::param($key, '-unset', "");
              # the -unset does not remove it from the ENV so remote it from there as well
              #  else a call to param inside of a subtest will also find it
              delete $ENV{$key};
            }
        } else {
            $Options->{$key} = $params{$key};
            $Trace->debug(msg=>"");
        }
    }
    $Trace->exit();
}

=head2 set_default

    $Options->set_default(option1=>"default1", option2=>"default2",...);

C<set_default> is used to set new default values for options. The new
default value is only applied if a value is not provided via the other 
sources (nate/env, testcases). The last call to C<set_default> will 
override previous calls. So, if multiple libraries are trying to set the
default with this call, the last one to run will set the default.

=cut

sub set_default {
    my $self = shift;
    my %params = (@_);

    $Trace->entry(msg=>"params: " . Dumper(\%params));
    foreach my $option (keys %params){
        $Defaults{$option} = $params{$option};
        $Trace->trace(msg=>"Setting the default value for option '$option' to '$params{$option}'");
    }
    $Trace->exit();
}

=head2 get_hash

    $hash_ref = $Options->get_hash(option=>"option1");
    $hash_ref = $Options->get_hash(option=>[qw(option1 option2)]);
    $hash_ref = $Options->get_hash();

C<get_hash> is used to obtain the value for the specified option[s] and to
return them in a hash form. If a reference to a list is passed in, then the
values for all of the options will be returned in the hash reference. The
method returns a reference to the hash. If C<option> is not specified, then
the returned hash will contain all options.

=over

=item option

(Optional) The name of the option. This can also be a reference to a list
of options.

=back

=cut

sub get_hash {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
            option => {type => SCALAR | ARRAYREF | UNDEF, optional => 1},
        }
    );

    if ( defined $params{option} ) {
        $Trace->entry(msg=>"option=$params{option}");
    } else {
        $Trace->entry(msg=>"option='undef'");
    }

    if (!defined($params{option})) {
        $Trace->exit();
        return $Options;
    }

    my %values;
    my @options = (ref($params{option}) eq "ARRAY") ? @{$params{option}} : ($params{option});
    $Trace->debug(msg=>"option=" . join(" ", @{$params{option}}));
    foreach my $option (@options) {
        $values{$option} = $Options->{$option};
    }
    $Trace->exit();
    return \%values;
}

=head2 print

    $Options->print();

C<print> is a debugging method. Calling it will print out the internal
hash representing all options and their values. This hash will only
contain the options that have been defined and processed.

=cut

sub print {
    my $self = shift;

    $Trace->entry();
    print Dumper($Options);
    $Trace->exit();
}

=head2 get_interface_table

    @interface_table = $Options->get_interface_table(nodes=>['node0'], interface_type=>"CLUSTER,DATA,MGMT");
    @interface_table = $Options->get_interface_table(nodes=>'node0', interface_type=>[qw(CLUSTER DATA MGMT)]);
    @interface_table = $Options->get_interface_table(nodes=>'node0,node1,node2');

C<get_interface_table> is used to retrieve the network interface configuration from the
tharnhost files for nodes passed in as C<nodes> either in the "old" one-line-per-config-item
style or the "new" all-interface-on-one-line style.  The retrieved information
is returned in an array of hashes with the interface_type and node information
stored as keys in the hash.  Errors are thrown if not all required keys are
specified by in the tharnhost file.  By default, C<get_interface_table> retrieves information
for all interfaces types on all nodes.  This can be narrowed down by passing in
either the C<nodes> or C<interface_type> parameters.

=over

=item interface_type

(Optional) The type of interfaces for which to return data.  Valid values are:
CLUSTER, DATA, MGMT.  This param can be passed as a reference to an array
containing the desired types or as a comma or colon separated string.  If unspecified,
data for all interface types is returned.

=item nodes

(Required) The nodes for which to return data.  This param can be passed as
a reference to an array containing the desired nodes or as a comma or colon
separated string.

=item required_keys

(Optional) The keys to be considered as required for each interface.  This
param can be passed as a reference to an array containing the desired keys
or as a comma or colon separated string.  If unspecified, the following keys
are considered to be required: name, port, address, netmask

=item ports

(Optional) The ports to alternate through for each interface if they aren't
specified in the tharnhost file.  This param can be passed as a reference
to an array containing the desired ports or as a comma or colon separated
string.

=back

=cut

sub get_interface_table {
    my ($self, %params) = validate_mixed(
        @_,
        [ { type => OBJECT, isa => 'Options' } ],
        {
         interface_type => {type => SCALAR | UNDEF, default => ["CLUSTER", "DATA", "MGMT"]},
         required_keys => {type => ARRAYREF | SCALAR, default => ["name", "port", "address", "netmask"]},
         nodes => {type => ARRAYREF | SCALAR},
         ports => {type => ARRAYREF | SCALAR, default => undef},
        }
    );
    my @retval;

    # Convert incoming params that are expected to be array refs if they are not
    if (ref($params{nodes}) eq ""){
	my @nodes = split(/[,:]/, $params{nodes});
	$params{nodes} = \@nodes;
    }
    if (ref($params{required_keys}) eq ""){
	my @required_keys = split(/[,:]/, $params{required_keys});
	$params{required_keys} = \@required_keys;
    }
    if (ref($params{interface_type}) eq ""){
	my @interface_type = split(/[,:]/, $params{interface_type});
	$params{interface_type} = \@interface_type;
    }

    # Loop through all the nodes
    foreach my $node (@{$params{nodes}}){
	# Get the hostrec information for the node
	my $node_info = main::host($node);
	next unless $node_info;

	# Loop through all the interface types
	foreach my $interface_type (@{$params{interface_type}}){
	    # Grab all the keys for the current interface type
	    my @keys = grep(/^${interface_type}_(IP|CONFIG)(\d+)?$/, keys %{$node_info});

	    # @keys should now have everything that we need to operate over, but do them in order
	    foreach my $key (sort _portsort @keys){
		$key =~ /(\d+)?$/;
		my $instance = $1 || "";

		# Set up some defaults
		my %data = (
			    mtu => 1500,
			    node => $node,
			    if_type => $interface_type,
			);

		# A default name
		if ($interface_type =~ /cluster/i){
		    $data{name} = "clus$instance",
		} else {
		    $data{name} = $node . "_".(lc $interface_type)."$instance",
		}

		# Grab the values for this instance from the host entry
		if ($key =~ /IP/){
		    # Old format where everything is in individual keys
		    $data{port} = $node_info->{$interface_type."_PORT$instance"};
		    $data{address} = $node_info->{$interface_type."_IP$instance"};
		    $data{netmask} = $node_info->{$interface_type."_MASK$instance"};
		    $data{gateway} = $node_info->{$interface_type."_GATEWAY$instance"};
		    if (exists($node_info->{$interface_type."_MTU$instance"})
			&& defined($node_info->{$interface_type."_MTU$instance"})){
			$data{mtu} = $node_info->{$interface_type."_MTU$instance"};
		    }
		} elsif ($key =~ /CONFIG/){
		    # New format where the config is more like what is used in CFE
		    foreach my $pair (split(/\s+/,$node_info->{$key})){
			my ($k, $v) = split(/=/, $pair);
			$data{$k} = $v;
		    }
		} # No else since we matched the keys before

		$Trace->trace(msg => "Data gleaned from config is " . Dumper(\%data));
		my @missing;
		foreach my $required_key (@{$params{required_keys}}){
		    if (!exists($data{$required_key}) || !defined($data{$required_key}) || !length($data{$required_key})){
			if (($required_key =~ /port/) && exists($params{ports})){
			    $data{$required_key} = $params{ports}->[$instance % @{$params{ports}}];
			} else {
			    push @missing, $required_key;
			}
		    }
		}
		if (@missing){
		    $Trace->fatal(msg => "Missing required keys for $node->$key => " . join(',', @missing) . Dumper(\%params));
		}
		push (@retval, \%data);
	    }
	}
    }
    @retval
}

#***********************************************************************

#
# The following routines are internal, so they don't need POD
#

sub _portsort {
    $a =~ /(\d+)$/;
    my $an = $1;
    $b =~ /(\d+)$/;
    my $bn = $1;

    $an <=> $bn;
}

sub _normalize_time {
    my $self = shift;
    my $option = shift;
    $Trace->entry();
    foreach (@_) {
        next if (!defined($_));
        s/\^/\*\*/;
        if (/^\d+$/) {
            next;
        } elsif (/^(\d+)m$/i) {
            $_ = $1 * 60;
        } elsif (/^(\d+)h$/i) {
            $_= $1 * 3600;
        } elsif (/^(\d+)d$/i) {
            $_ = $1 * 3600 * 24;
        } else {
            $Trace->fatal(msg=>"Could not normalize the time for '$option=$_'");
        }
    }
    $Trace->exit();
}

my %multipliers = (
                   'k' => 1024**1,
                   'K' => 1000**1,
                   'm' => 1024**2,
                   'M' => 1000**2,
                   'g' => 1024**3,
                   'G' => 1000**3,
                   't' => 1024**4,
                   'T' => 1000**4,
                   'p' => 1024**5,
                   'P' => 1000**5,
                   'e' => 1024**6,
                   'E' => 1000**6,
                  );
my $multipliers = join('', keys %multipliers);

#
# Sub to handle expanding the multipliers into expressions that will eval to the proper
# number if we find a multiplier in our table
#
sub _expand_multiplier {
  my ($num, $mfactor) = @_;
  if (exists($multipliers{$mfactor})){
    return "($num * $multipliers{$mfactor})";
  }
  return "$num$mfactor";
}

#
# Handle the substitution of values
#
=head2 unit_to_base

    my $kbyte = "1k";
    my $mbyte = "1m";
    unit_to_base($kbyte, $mbyte);

C<unit_to_base> is used to enable evaluation of strings to perform evaluation
of the parameter to perform math functions (really could be any perl function)
as well as common size substitutions like kilo, mega, etc.  For example, specifying
'1k-2' as a value would result in the param being set to 1022.  The common 
substitutions are:
   k => 1024, m => 1024^2, g => 1024^3, t => 1024^4, p => 1024^5, e => 1024^6
   K => 1000, M => 1000^2, G => 1000^3, T => 1000^4, P => 1000^5, E => 1000^6

=cut

#sub unit_to_base (\@) {
sub unit_to_base {
    my $self;
    if (ref($_[0]) eq "Options"){
        $self = shift;
    }
    foreach (@_){
        next if (!defined($_));
        s/\^/\*\*/;
        next if ($_ =~ /^(\d+\.?\d*|\d*\.?\d+)$/); # leave a raw number alone
        # Substitute each expansion string with the expanded string.
        # For example 1k in the string gets matched and the substitution
        # calls expand_multiplier(1, 'k') which returns (1 * 1024)
        $_ =~ s/(\d+\.?\d*|\d*\.?\d+)([$multipliers])b?/_expand_multiplier($1,$2)/gei;

        # At this point, the string has been fully substituted and we can
        # finally do the full eval of the resulting string
        my $val = eval $_;
        if ($@){
            $Trace->fatal(msg=>"Error evaluating $_: $@\n", result=>"SCRIPT");
        }
        $_ = int($val);
    }
}

#
# Convert the booleans
#
sub convert_boolean {
    my $self = shift;
    my $option = shift;
    foreach my $boolean (@_){
        if (!defined $boolean || $boolean =~ /no|off|0|false|undef|^$/i) {
            $boolean = undef;
        } elsif ($boolean =~ /yes|on|1|true|defined/i) {
            $boolean = 1;
        } else {
            $Trace->fatal(result=>"SCRIPT",
                          msg=>"Do not know how to convert the value of option '$option' ('$boolean') to a boolean value");
        }
    }
}

sub _process_options {
    my $self = shift;
    my %params = (
                  override => 1,
                  text => undef,
                  @_);
    $Trace->entry(msg=>"load params from $params{text}");

    my $hr = $params{options_hash};
    $Trace->debug(msg=>"options from $params{text} are: " . Dumper($hr));
    if ($hr) {
        foreach my $key (keys %{$hr}) {
            my $option;
            if (!$params{override} && ( $option=$self->defined(option=>$key) ) ) {
                $Trace->debug(msg=>"Not setting '$key' because it is already defined as '$option'");
                next;
            }
            # parse the values of Compat first and then override with the normal values
            if (my $compat_key=$Compat{$key}) {
                $Options->{$compat_key} = $hr->{$key} if ($hr->{$key});
                $Trace->debug(msg=>"setting $compat_key to $hr->{$key} ($compat_key was translated from $key)");
            }
            if ($Definitions{$key}) {
                $Options->{$key} = $hr->{$key};
                # the following is only for the debug call
                my $value = (ref($hr->{$key}) eq "ARRAY") ? join(",", @{$hr->{$key}}) : $hr->{$key};
                $value = "undef()" if (!defined($value));
                $Trace->debug(msg=>"setting $key to $value from source \"$params{text}\"");
            }
        }
    }
    $Trace->exit();
}

sub _check_dependencies {
    my $self = shift;
    my $failed = undef;
    $Trace->entry();
    # loop through the required options array and test each one
    while (my ($option, $dependencies) = each %Dependencies) {
        $Trace->debug(msg=>"check '$option':".join(" ", @$dependencies));
        if (!defined($option)) {
            $Trace->fatal(msg=>"Undefined option\n", result=>"SCRIPT");
        }
        # this is complicated because the option may be a scalar or an array. an empty array needs to be
        # treated as a false in this case, which isn't the way a blanket if of a variable works - IE if ( [] ) returns true.
        # we could have made a blanket statement that all options shouldn't be predefined (IE undef), but if someone would
        # mistakenly define the variable as an empty array then they wouldn't even realize that this test would fail.
        if ($self->defined(option=>$option)) {
            foreach my $dependency (@$dependencies) {
                if ( !$self->defined(option=>$dependency)) {
                    $failed .= "\tOption '$dependency' is required because option '$option'='";
                    $failed .= (ref($Options->{$option}) eq "ARRAY") ? join(" ", @{$Options->{$option}}) : $Options->{$option};
                    $failed .= "'\n";
                } else {
                    $Trace->trace(msg=>"Option '$option' makes '$dependency' required and both are correctly set.");
                }
            }
        }
    }
    $Trace->fatal(msg=>"Failure during Options dependancy verification:\n$failed", result=>"CONFIG") if ($failed);
    $Trace->exit();
}

sub _check_comparisons {
    my $failed = undef;
    $Trace->entry();

    while (my ($option, $value) = each %{$Comparisons{less}}) {
        next if (!defined($Options->{$option}));
        $Trace->debug(msg=>"check '$option' $Options->{$option} < $value");
        unless ( $Options->{$option} < $value ) {
            $failed .= "option '$option' ($Options->{$option}) must be < '$value'\n";
        }
    }
    while (my ($option, $value) = each %{$Comparisons{greater_equal}}) {
        next if (!defined($Options->{$option}));
        $Trace->debug(msg=>"check '$option' $Options->{$option} >= $value");
        unless ( $Options->{$option} >= $value ) {
            $failed .= "option '$option' ($Options->{$option}) must be >= '$value'\n";
        }
    }
    $Trace->fatal(msg=>$failed, result=>"SCRIPT") if ($failed);
    $Trace->exit();
}

sub _check_requirements {
    my $self = shift;
    my $failed = undef;
    $Trace->entry();

    # check the required arguments
    foreach my $option (keys %Requirements) {
        $Trace->debug(msg => "Checking for required parameter $option");
        if (!defined($Options->{$option}) || 
            ($Options->{$option} && (ref($Options->{$option}) eq "ARRAY") && (!scalar(@{$Options->{$option}})))) {
            $Trace->fatal(msg=>"option '$option' is required: $Requirements{$option}\n", result=>"SCRIPT");
        }
    }

    foreach my $list (@Require_one_of) {
        my $optionSet = 0;
        # loop through all the options and see if one of them is set
        $Trace->debug(msg=>"check that one of the following is specified: ".join(" ", @$list));
        foreach my $opt ( @$list ) {
            if ($self->defined(option=>$opt)) {
                $optionSet = 1;
                last;
            }
        }
        if ( !$optionSet ) {
            $failed .= "one of \"" . join("\" or \"", @$list) . "\" are required.\n";
            next;
        } else {
            $Trace->trace(msg=>"\"".join("\" or \"",  @$list)."\" are set as required.\n");
        }
    }
    $Trace->fatal(msg=>$failed, result=>"SCRIPT") if ($failed);
    $Trace->exit();
}

sub _check_valid_values {
        my $failed = undef;
        $Trace->entry();

        while (my ($option, $value) = each %Valid_values) {
                next if (!defined($Options->{$option}));
                $Trace->debug(msg=>"check that the value of '$option' is one of: ".join(" ", @{$Valid_values{$option}}));
                my @option_array;
                if ( ref $Options->{$option} eq "ARRAY" ) {
                        @option_array = @{$Options->{$option}};
                } else {
                        push @option_array, $Options->{$option};
                }
                foreach my $value ( @option_array ) {
                        unless (grep( $_ eq $value, @{$Valid_values{$option}} ) ) {
                                $failed .= "\toption '$option' ($value) must be one of <".join(" ", @{$Valid_values{$option}}).">\n";
                        }
                }
        }

        $Trace->fatal(msg=>"\n$failed", result=>"SCRIPT") if ($failed);
        $Trace->exit();
}

# this routine now only supports a limited number of file_test operators since we are running
#  tests on both unix and windows. Since windows does not support all of the calls and the http
#  path does not support all of the calls, this list is now limited in scope. If a test tries
#  to use an unsupported test, a warning will be thrown.

sub _file_test {
    my $failed = undef;
    $Trace->entry();
    while (my ($option, $tests) = each %File_tests) {
        my $tfile = $Options->{$option};
        next if (!$tfile); # The option was not defined

        my $file = $Object->{ntest} ? NATE::Inc::find_file($tfile) : $tfile;
        if (!$file) {
            $Trace->fatal(msg=>"While trying to find the location of the option '$option', I could not find '$tfile' anywhere in the NATE_LIB path");
        }
        $Trace->debug(msg=>"find_file($tfile) returned $file");

        if ($file =~ /^http:.+/) {
            $Trace->warn(msg=>"The value of $option ($tfile) was found via an http path. The file_test operators\
do not work against http paths. You must either change your NATE_LIB to\
not start with an http path or not run a file test on this file");
            $Trace->exit();
            return;
        }

        my $file_param = "$file (from param $option)";
        # go through the list of operators that we were passed and test them all.  if someone passed in
        # "erw" and the file doesn't exists then they should expect to get back a failure and see that 
        foreach my $operator ( @$tests ) {
            $Trace->debug(msg=>"test -$operator $file");
            if ($operator eq 'e' && ! -e $file) {
                $failed .= "$file_param does not exist\n";
            }
            if ($operator eq 'd' && ! -d $file) {
                $failed .= "$file_param is not a directory\n";
            }

            if ($operator eq 'f' && ! -f $file) {
                $failed .= "$file_param is not a file\n";
            }
            if ($operator eq 'r' && ! -r $file) {
                $failed .= "$file_param is not readable\n";
                $failed .= $operator;
            }
            if ($operator eq 'w' && ! -w $file) {
                $failed .= "$file_param is not writable\n";
            }
            if ($operator eq 'x' && ! -x $file) {
                $failed .= "$file_param is not executable\n";
            }
            if ($operator eq 'z' && ! -z $file) {
                $failed .= "$file_param does not exist or is not zero length\n";
            }
            if ($operator eq 's' && ! -s $file) {
                $failed .= "$file_param does not exist or is zero length\n";
            }
        }
    }

    $Trace->fatal(msg=>$failed, result=>"CONFIG") if ($failed);
    $Trace->exit();
}

sub _help {
    my $self = shift;
    $Trace->entry();

    # find the maximum length of the params for pretty printing
    my $max=0;
    foreach my $key (keys %Descriptions) {
        my $length = length $key;
        $max = $length if ($length > $max);
    }

    # print out the params from the file[s] first
    foreach my $file (keys %{$Origins{script}}) {
        print "$file:\n";
        _print_help_info($Origins{script}{$file}, $max);
    }

    # now print out the params from the libaries
    foreach my $file (keys %{$Origins{library}}) {
        print "$file:\n";
        _print_help_info($Origins{library}{$file}, $max);
    }

    # now find out any of the required params
    my @required = keys %Requirements;
    if (scalar @required) {
        print "Required parameters: ", join(",", @required), "\n";
    }

    # now find out any of the require_one_of params
    if (scalar @Require_one_of) {
        print "Require_one_of parameters: ";
        my @required;
        foreach my $list (@Require_one_of) {
            my $string = "[";
            $string .= join(",", @$list);
            $string .= "]";
            push(@required, $string);
        }
        print join(",", @required);
        print "\n";
    }
    print "\n";

    sub _print_help_info {
        my $options = shift @_;
        my $max = shift @_;
        
        foreach my $option (sort @{$options}){
            # determine the amount of '.'s to print after the param
            my $num = $max - length($option) - 2;
            my $str;
            if ($num > 0){
                $str = ' 'x2 . '.'x$num;
            }

            # find any default values for the param
            my $default = "";
            if ($Defaults{$option}) {
                $default = " (default: ";
                if (ref($Defaults{$option}) eq "ARRAY") {
                    $default .= join(",", @{$Defaults{$option}});
                } else {
                    $default .= $Defaults{$option};
                }
                $default .= ")";
            }

            # find the valid values for the param
            my $valid = "";
            if ($Valid_values{$option}) {
                $valid = " (valid_values: ";
                $valid .= join(",", @{$Valid_values{$option}});
                $valid .= ")";
            }

            # find the dependencies for this param
            my $depends = "";
            if ($Dependencies{$option}) {
                $depends = " (dependencies: ";
                $depends .= join(",", @{$Dependencies{$option}});
                $depends .= ")";
            }

            # is this a boolean?
            my $boolean = "";
            if ($Definitions{$option} =~ /boolean/) {
                $boolean = " (boolean value)";
            }

            # is this an eval?
            my $eval = "";
            if ($Definitions{$option} =~ /eval/) {
                $eval = " (eval value)";
            }

            # print out all of the information for this param
            printf( "  %-${max}s: $Descriptions{$option}%s%s%s%s%s\n", $option . $str, 
                    $default, $valid, $depends, $boolean, $eval);
        }
        print "\n";
    }
    $Trace->exit();
    exit 0;
}

sub _process_test_cases {
    my $self = shift;
    my @hashes = @_;
    my %tc_options;

    $Trace->entry();
    if ( keys %Test_cases == 0 ) {
        $Trace->trace(msg=>"no test cases are defined for this test");
        $Trace->exit();
        return undef;
    }

    # WARNING:  There are nip dependencies on LISTTESTPLAN and LISTTESTCASES that are not defined
    #           here.  Changes should be made with caution

    foreach my $hash (@hashes) {
        # If the caller is asking for a listing of the test cases, loop through each key of the
        #  Test_cases hash and print out the testcase name and description (if it exists)
        if (defined($$hash{LISTTESTCASES} && $$hash{LISTTESTCASES} =~ /yes/i)){
            $Trace->trace(msg=>"found LISTTESTCASES");
            my @tc;
            if (defined($$hash{TESTCASE})) {
                push(@tc, $$hash{TESTCASE}) if ($Test_cases{$$hash{TESTCASE}});
            } else {
                @tc = sort keys %Test_cases;
            }
            foreach my $testcase (@tc){
                if (${$Test_cases{$testcase}}{TC_multiplier}) {
                    foreach my $multiplier (@{$Test_cases{$testcase}{TC_multiplier}}) {
                        print "testcase : ".$multiplier."_".$testcase." : ";
                        if (${$Test_cases{$testcase}}{TC_description}) {
                           print "($multiplier) ${$Test_cases{$testcase}}{TC_description}"; 
                        } else {
                            print "no description specified";
                        }
                        print "\n";
                    }
                } else {
                    $Trace->trace( msg => "tc = $testcase, dt = $self->{default_testcase}, let = $Options->{LIST_EMPTY_TESTCASES}");
                    next if (($testcase ne $self->{default_testcase}) # Not the default testcase
                             && !(grep(!/TC_description/, keys %{$Test_cases{$testcase}})) # Doesn't have any keys other than TC_description
                             && defined($Options->{LIST_EMPTY_TESTCASES}) && ($Options->{LIST_EMPTY_TESTCASES} =~ /^\s*(false|0|no)\s*$/i)); # And empty testcases aren't allowed
                    print "testcase : $testcase : ";
                    if (${$Test_cases{$testcase}}{TC_description}) {
                        print "${$Test_cases{$testcase}}{TC_description}";
                    } else {
                        print "no description specified";
                    }
                    print "\n";
                }

            }
            exit 0 if ($self->{Exit_after_list});
        }
        # If the caller is asking for a listing of the test cases for a .plan file, loop through 
        #  each key of the Test_cases hash and print out the plan file information
        if (defined($$hash{LISTTESTPLAN} && $$hash{LISTTESTPLAN} =~ /yes/i)){
            $Trace->trace(msg=>"found LISTTESTPLAN");
            my @tc;
            if (defined($$hash{TESTCASE})) {
                push(@tc, $$hash{TESTCASE}) if ($Test_cases{$$hash{TESTCASE}});
            } else {
                @tc = sort keys %Test_cases;
            }
            eval "use Utils qw(relative_testname);";
            die $@ if $@;
            my $testname = relative_testname();
            my $testset;
            foreach my $testcase (@tc){
                if (${$Test_cases{$testcase}}{TC_multiplier}) {
                    foreach my $multiplier (@{$Test_cases{$testcase}{TC_multiplier}}) {
                        if (${$Test_cases{$testcase}}{TESTSET}) {
                            $testset = " TESTSET=${$Test_cases{$testcase}}{TESTSET}/$multiplier";
                        } elsif (${$Test_cases{$testcase}}{TC_testset}) {
                            $testset = " TESTSET=${$Test_cases{$testcase}}{TC_testset}/$multiplier";
                            $Trace->warn(msg=>"'TC_testset'is being deprecated. Please use 'TESTSET' in your testcase definitions");
                        } else {
                            $testset = "";
                        }
                        print $testname." TESTCASE=".$testcase."_".$multiplier.$testset."\n";
                    }
                } else {
                    if (${$Test_cases{$testcase}}{TESTSET}) {
                        $testset = " TESTSET=${$Test_cases{$testcase}}{TESTSET}";
                    } elsif (${$Test_cases{$testcase}}{TC_testset}) {
                        $Trace->warn(msg=>"'TC_testset'is being deprecated. Please use 'TESTSET' in your testcase definitions");
                        $testset = " TESTSET=${$Test_cases{$testcase}}{TC_testset}";
                    } else {
                        $testset = "";
                    }
                    print $testname." TESTCASE=".$testcase.$testset."\n";
                }
            }
            exit 0 if ($self->{Exit_after_list});
        }
    }

    # We have now passed up processing of LISTTESTCASES. Now define TESTCASE in the Options hash.
    # It will be found in one of the hash references in the following loop and processed correctly.
    $Options->{TESTCASE} = $self->{testcase} if (!defined($Options->{TESTCASE}));

    # loop again now that we did not find LISTTESTCASES or LISTTESTPLAN
    foreach my $hash (@hashes) {
        # If the caller passed in a specific testcase, look for it in the TestCase hash. If it
        #  exists, save the options hash so that it can be merged with the other options later.
        #  If it does not exist, then exit with an error
        if (defined($$hash{TESTCASE})) {
            my $test_case = $$hash{TESTCASE};
            $Trace->trace(msg=>"found $test_case in $hash");
            if ($Test_cases{$test_case}) {
                if (${$Test_cases{$test_case}}{TC_multiplier}) {
                    foreach my $multiplier (@{$Test_cases{$test_case}{TC_multiplier}}) {
                        %tc_options = %{$Test_cases{$test_case}};
                        # set the current TESTCASE in the options hash to the one specified since 
                        # some scripts will want to know
                        $Options->{TESTCASE} = $multiplier."_".$test_case;
                    }
                } else {
                    %tc_options = %{$Test_cases{$test_case}};
                    # set the current TESTCASE in the options hash to the one specified since 
                    # some scripts will want to know
                    $Options->{TESTCASE} = $test_case;
                }
                delete $tc_options{Description};
                last;
            } else {
                my $text = join(", ", sort keys %Test_cases)."\n";
                $Trace->fatal(msg=>"Testcase \"$test_case\" NOT defined\nAvailable testcases: $text", result=>"SCRIPT");
            }
        }
    }
    $Trace->exit();
    return \%tc_options;
}


# for backwards compatability
sub processed {
    $Trace->entry();

    $Trace->exit(msg=>"returning $Object->{Processed}");
    return $Object->{Processed};
}

sub get_max_cmode_systems {
    my $self = shift;
    my $cmax = main::param( "MAX_CMODE_SYSTEMS" );
    return defined( $cmax ) ? $cmax : 4;
}

sub get_max_7mode_systems {
    my $self = shift;
    my $smax = main::param( "MAX_7MODE_SYSTEMS" );
    return defined( $smax ) ? $smax : 3;
}

sub import {
    # Import gets called with a package and a list of items
    # We want to pick off the ones that start with - and
    # handle.  Other cases need to be handled as they are
    # encountered
    my $package = shift;

    $Trace = new TracePkg unless defined($Trace);

    $Trace->entry();

    my @params = @_;
    foreach my $item (@params){
        if ($item =~ s/^-//){
            my $varname = $item;
            $varname =~ s/^no_//i;

            my $oldvalue = $Object->{$varname};
            if ($item =~ /^no_/i){
                $Object->{$varname} = 0;
            } else {
                $Object->{$varname} = 1;
            }
            $Trace->trace(msg => "Handled param $varname.  Changed from $oldvalue to $Object->{$varname}.");
        } else {
            $Trace->warn(msg => "Did not handle param $item");
        }
    }

    $Trace->exit();
}

=head1 PARAMS 

=over

=item TESTCASE

This param defines which testcase the test should be running with. 
If there are no test cases defined via a call to C<test_cases>, then
nothing changes. See C<test_cases>.

=item LISTTESTCASES

Print out a list of defined test cases and their descriptions. The test
will exit immmediately afterwards.

=item LISTTESTPLAN

Print out a list of test cases in the format of a .plan file. The test
will exit immmediately afterwards.

=item HELP

Print out a list of all defined options and their descriptions. The test
will exit immmediately afterwards.

=item NTAP_QA_TRACE

The param is a definition for the ENV variable used by L<TracePkg>.

=item NUFFO_MULTITEST

This param is a boolean and true by default. If the param is true and the
Options object has test cases defined and the param C<TESTCASE> is defined
then the code will set the nate param NUFFO_MULTITEST_LOGS to be 1 so that
nuffo.thpl will parse the log files looking for results. Setting this param
to false will cause the code to do nothing with the NUFFO_MULTITEST_LOGS
param.

=back

=head1 SEE ALSO

L<TracePkg>

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/Options.pm#2 $
  Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
