# Hey emacs, this is basically -*- perl -*-
# vim: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:
#
###############################################################################
# Copyright (c) 2013 NetApp, Inc.
# All rights reserved.
###############################################################################

=pod

=head1 NAME

Os::BSDExecutor

=for html <hr>

=head1 DESCRIPTION

This library provides standalone methods to run commands in the FreeBSD
shell (systemshell) of a node, along with executing binaries in the shell.

=for html <hr>

=head1 SYNOPSIS

# Step-1: Install a tool's binary in the FreeBSD shell
my $filer_path = Os::BSDExecutor::install(
                    node        => $node_object,
                    source_path => "/path/to/tool_binary",
                );
$Test->nlog("Output is:\n$output");

# Step-2: Run the installed binary on the filer using the filer path
my $output = Os::BSDExecutor::execute(
                node    => $node_object,
                command => $filer_path . " -arg1 -arg2",
                timeout => 600,     # Option supported by the
                                    # execute_raw_command() method
            );

# Step-3: Delete the tool's binary from the filer by binary name
Os::BSDExecutor::delete_binary(
    node       => $node_object,
    filer_path => "tool_binary",
);

# Run a command "ls /clus"
my $output = Os::BSDExecutor::execute(
                    node    => $node,
                    command => "ls -laR /clus/",
                );
$Test->nlog("Output is:\n$output");

=cut

package Os::BSDExecutor;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Params::Validate qw(:all);
use Data::Dumper qw(Dumper);
use File::Basename;

use Tharn qw($Log);
use TestUtils qw(get_site_info guess_site);
use NATE::BaseException;

# constants
use constant Binary_Filer_Path => "/mroot/";
use constant Default_Tools_Location_BSD =>
                '/x/eng/localtest/arch/amd64-bsd-freebsd8/bin/';

# NACL
use NACL::C::Node;

# private globals
my %Cache;

# process options to make TestUtils happy
my $Options = Options->new();
$Options->process();


###############################################################################
#                   Helper methods for routine tasks
###############################################################################

=for html <hr>

=head2 execute

Runs a command in the FreeBSD shell of a node.

=over

=item Options

=over

=item node => $node_object

 (Required) NACL::C::Node object for the node that will execute the command.

=item command => $cmd_to_execute

 (Required) A command string for the command to be executed.

=item %other_args

 (Optional) any additional args supported by the systemshell APIset's
            execute_raw_command() method.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # example for getting just the raw output
 my $output = Os::BSDExecutor::execute(
                    node    => $node,
                    command => "ls -laR /clus/",
                );
 $Test->nlog("Raw output is:$output\n");

=back

=cut

sub execute {
    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        node        => { isa  => 'NACL::C::Node' },
                        command     => { type => SCALAR },
                    },
                    # pass extra args as-is to the execute_raw_command() method
                    allow_extra => 1,
                );

    # get inputs
    my $node    = delete $args{node};
    my $command = delete $args{command};

    # run the command
    return $node->get_systemshell_apiset()->execute_raw_command(
                               command => $command,
                               %args,
                           );
}

=for html <hr>

=head2 install_binary

Installs a binary on a node by copying it to a node, and returning the
node specific path to the caller. This path can be used to run binaries
on the node via the execute() method.
Installation should be a 1-time operation.
If the binary already exists on the filer, install_binary() will fail.
To use multiple instances of the binary, give each instance an alternate
name using the 'rename_at_destination' parameter.

=over

=item Options

=over

=item node => $node_object

 (Required) NACL::C::Node object for the node to use

=item source_path => $path_to_find_binary

 (Required) A string containing the path of the binary's off-box location.
            The binary will be picked up from this location to copy to the filer.
            E.g. - '/some/path/iozone'
            Alternatively, specifying just the binary's name will make this method
            look at the standard location:
            E.g. - 'iozone' => '/x/eng/localtest/arch/amd64-bsd-freebsd8/bin/iozone'

=item rename_at_destination => 'alternate_name'

 (Optional) Give the binary an alternate name while copying it to the filer.
            E.g. - if the alternate name 'hammer_1' is provided, 'hammer'
            will be copied to the filer as 'hammer_1'.
 [Default]: Use the same name. E.g. - 'hammer' will be copied as 'hammer'.

=back

=item Returns

 SCALAR, the path of the binary on the filer, or
 exception on failures (including the binary already exists)

=item Examples

 # install a binary on the node
 my $binary_path_on_filer = Os::BSDExecutor::install_binary(
                    node        => $node,
                    source_path => "/path/to/tool_binary",
                );

 # install a binary on the node using a different target name
 my $binary_path_on_filer = Os::BSDExecutor::install_binary(
                    node        => $node,
                    source_path => "/path/to/tool_binary",
                    rename_at_destination => 'tool_binary_1',
                );

=back

=cut

sub install_binary {
    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        node        => { isa  => 'NACL::C::Node' },
                        source_path => { type => SCALAR },
                        rename_at_destination => {
                            type    => SCALAR,
                            default => '',
                        },
                    },
                );

    my $node                = delete $args{node};
    my $node_name           = $node->node();
    my $binary_source_path  = delete $args{source_path};
    my $alternate_name      = delete $args{rename_at_destination};

    # check if the caller has specified an alternate name for the
    # binary, to use after copying it to the filer
    # else,
    # extract just the binary name, without any path prefixed to it
    my $binary_name_short;
    if ($alternate_name ne '') {
        $binary_name_short = $alternate_name;
    } else {
        $binary_name_short = File::Basename::fileparse($binary_source_path);
    }

    # build the filer path for the binary
    my $binary_filer_path = Binary_Filer_Path() . $binary_name_short;
    $Log->debug("Binary filer path to use: [$binary_filer_path]");

    # check if the binary already exists on the filer
    # if yes, throw an exception
    $Log->debug("\nChecking if the binary exists on the filer...\n");
    if (_does_filename_exist_on_filer(
                node => $node,
                path => $binary_filer_path)) {

        throw NATE::BaseException(
            "Binary $binary_filer_path already exists on the filer!\n" .
            "Use an alternate name if you wish to use multiple instances.\n"
        );
    }

    # we need to copy the binary

    # first, check to see if the caller has provided its own path to the
    # binary, else get the binary from the default location
    if ($binary_source_path =~ /\//) {
        # looks like a path is provided
        # use it to get the binary instead of the default location
        $Log->comment("\nGetting the binary from user specified source " .
                      "location: $binary_source_path");
    } else {
        # no source path was provided, try to get the binary from the
        # standard location that we know of
        $binary_source_path = Default_Tools_Location_BSD() . $binary_source_path;
        $Log->comment("\nGetting the binary from the default source " .
                      "location: $binary_source_path");
    }

    # first check if the binary actually exists (or is accessible)
    if (! -e $binary_source_path or ! -r $binary_source_path) {
        throw NATE::BaseException(
            "Source path [$binary_source_path] is nonexistent/unreadable!\n"
        );
    }

    # get the web server root for this node/site
    my $web_server_root = &_get_web_server_root(node => $node);
    $Log->debug("Web server root: $web_server_root\n");

    # get a 'systemshell' apiset object for the target node, and
    # use 'wget' to copy the binary to the filer
    my $cmd = "sudo wget "
            . $web_server_root
            . $binary_source_path
            . " -O "
            . $binary_filer_path;
    &execute(
        node    => $node,
        command => $cmd,
    );

    # check if 'wget' actually copied the binary to the filer
    if (!_does_filename_exist_on_filer(
                node => $node,
                path => $binary_filer_path)) {

        throw NATE::BaseException(
            "Did not find the binary on the filer!\n"
        );
    }

    # save the filer_path to cache
    $Cache{$node_name}{tools}{$binary_filer_path} = 1;
    _cache_node(node => $node);

    # change the permissions on the binary, just in case
    my $output = &execute(
                    node    => $node,
                    command => "sudo chmod 755 $binary_filer_path",
                );
    if ($output ne '') {
        throw NATE::BaseException(
            "Failed to change permissions for the binary [$binary_filer_path] " .
            "on the filer. \nOutput: $output\n"
        );
    }

    # return the string containing the filer_path
    return $binary_filer_path;

}

=for html <hr>

=head2 delete_binary

Deletes a binary from the filer.
This method is also internally used by Os::BSDExecutor::cleanup()
for all binaries copied to the filer.

=over

=item Options

=over

=item node => $node_object

 (Required) NACL::C::Node object for the node to use

=item filer_path => $filer_path_for_binary

 (Required) A string containing the full filer path of the binary,
            returned by Os::BSDExecutor::install_binary()
            E.g. - (filer_path => '/mroot/iozone')

=back

=item Returns

 NONE if successful, throws an exception on failure.

=item Examples

 Os::BSDExecutor::delete_binary(
    node       => $node,
    filer_path => "/filer_path/to/tool_binary",
);

=back

=cut

sub delete_binary {
    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        node       => { isa  => 'NACL::C::Node' },
                        filer_path => { type => SCALAR },
                    },
                );

    my $node              = delete $args{node};
    my $node_name         = $node->node();
    my $binary_filer_path = delete $args{filer_path};

    # for protecting the binary, first see if we installed it
    if (!defined $Cache{$node_name}{tools}{$binary_filer_path}) {
        throw NATE::BaseException(
            "Os::BSDExecutor::delete_binary() did not install " .
            "the binary [$binary_filer_path], cannot delete it!"
        );
    }

    # run the "rm" command to delete the binary from the filer
    my $output = &execute(
                    node    => $node,
                    command => "sudo rm $binary_filer_path"
                );
    if ($output ne '') {
        throw NATE::BaseException(
            "Failed to delete the binary!\nOutput:$output\n"
        );
    }

    # check if the binary was actually deleted
    if (_does_filename_exist_on_filer(
                node => $node,
                path => $binary_filer_path)) {

        throw NATE::BaseException(
            "Failed to delete the binary [$binary_filer_path]!\n"
        );
    }
    
    # delete the entry from our cache
    delete $Cache{$node_name}{tools}{$binary_filer_path};

    return;
}

=for html <hr>

=head2 cleanup

Cleanup all binaries using the FreeBSD shell of a node.

=over

=item Options

=over

=item none

 =back

=item Returns

 NONE if successful, throws an exception on failure.

=item Examples

 Os::BSDExecutor::cleanup();

=back

=cut

sub cleanup {

    my $me = (caller(0))[3];

    # delete all binaries from the BSD shell
    foreach my $node_name (keys %Cache) {

        $Log->debug(sub { $me . ": Cleaning up node $node_name..." });
        # delete all binaries for this node
        foreach my $tool (keys %{$Cache{$node_name}{tools}}) {

            $Log->debug(sub { $me .
                        ": Deleting tool $tool from node $node_name..." });
            delete_binary(
                node       => $Cache{$node_name}{node},
                filer_path => $tool,
            );
        }
    }

    return;
}


###############################################################################
#                   Private methods for internal use
###############################################################################

#-----------------------------------------------------------------------------#
#
# PRIVATE METHOD
#
# Method    : _get_web_server_root
#
# Purpose   : Gets the web server root path for builds by identifying the
#             site at which the node is located at.
#
# Parameters:
#   node    : an object of type NACL::C::Node
#
# Returns:
#   SCALAR  : the string containing the web server root, or an
#             exception for failure
#
#-----------------------------------------------------------------------------#

sub _get_web_server_root
{
    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        node => {isa => 'NACL::C::Node' },
                    },
                );

    my $node      = delete $args{node};
    my $node_name = $node->node();
    _cache_node(node => $node);

    # check our local cache first
    if (!defined $Cache{$node_name}{web_server}) {

        #=============================================================
        # Step-1: find out which site we are in
        my $host = $node->hostrec();

        # 1-a: first, look into the tharn info
        my $site = $host->{SITE};
        if (!defined $site or
            (defined $site and $site eq '')) {

            $Log->debug("Hostrec did not have site info, guessing it...");
            # 1-b: try and guess the site
            $site = TestUtils::guess_site($host);

            if (!defined $site or
                (defined $site and $site eq '')) {
                # no site was returned?
                throw NATE::BaseException(
                    "TestUtils::guess_site() returned nothing for 'site'!\n" .
                    "I don't know where to get the binary from.\n"
                );
            }
        }
        $Log->debug("\nNode $node_name seems to be at site [$site]\n");

        #=============================================================
        # Step-2: get general site info
        my %SiteInfo = TestUtils::get_site_info();
        my @sites = keys %SiteInfo;
        $Log->debug(sub {
            "\nList of available sites:\n" . Dumper(\@sites) . "\n"
        });

        if (!@sites) {
            # no sites were returned?
            throw NATE::BaseException(
                "TestUtils::get_site_info() returned nothing!\n"
            );
        }

        #=============================================================
        # Step-3: pull the web server info
        my $local_site_data = $SiteInfo{$site};
        $Log->debug(sub {
            "\nSite info for site $site:\n" . Dumper($local_site_data) . "\n"
        });

        if (!defined $$local_site_data{install_proto} or
            !defined $$local_site_data{install_server}) {

            throw NATE::BaseException(
                "TestUtils::get_site_info() returned 'undefined' for " .
                "'install_proto' and/or 'install_server'.\n" .
                "Please check TestUtils/site_info.dat for these values for site [$site]\n"
            );
        }
        
        #=============================================================
        # Step-4: build the web server string and save it
        my $web_server = $local_site_data->{install_proto} . "://";
        if (defined $local_site_data->{install_user} and
            defined $local_site_data->{install_pass}) {
            $web_server .= $local_site_data->{install_user} . ":"
                        .  $local_site_data->{install_pass} . "@";
        }
        $web_server .= $local_site_data->{install_server};
        $Log->debug("\nTestUtils says the web server root is [$web_server]\n");

        if (defined $web_server and $web_server ne '') {
            # save to cache
            $Cache{$node_name}{web_server} = $web_server;
        } else {
            throw NATE::BaseException(
                "BSDExecutor::_get_web_server_root() - failed to get web server root" .
                " for node [$node_name] at site [$site]\n"
            );
        }
    }

    return $Cache{$node_name}{web_server};
}


#-----------------------------------------------------------------------------#
#
# PRIVATE METHOD
#
# Method    : _does_filename_exist_on_filer
#
# Purpose   : Checks if a file by a name, exists on the filer or not.
#
# Parameters:
#   node    : an object of type NACL::C::Node
#   path    : SCALAR, the file path to check for.
#
# Returns:
#   SCALAR  : 1 (yes, it exists) OR 0 (no, it does not exist)
#
#-----------------------------------------------------------------------------#

sub _does_filename_exist_on_filer {

    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        node => { isa  => 'NACL::C::Node' },
                        path => { type => SCALAR },
                    },
                );

    my $node        = delete $args{node};
    my $binary_name = _get_binary_filer_path_string(
                            binary_name => $args{path}
                        );

    # run the "ls" command to see if the binary already exists or not
    my $output = &execute(
                    node    => $node,
                    command => "ls -l $binary_name"
                );
    if ($output =~ /No such file or directory/) {
        return 0;   # false
    }

    return 1; # true
}

#-----------------------------------------------------------------------------#
#
# PRIVATE METHOD
#
# Method    : _cache_node
#
# Purpose   : Cache a node object
#
# Parameters:
#   node    : [Required]: an object of type NACL::C::Node
#
# Returns:
#   NONE
#
#-----------------------------------------------------------------------------#

sub _cache_node {

    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        node        => { isa  => 'NACL::C::Node' },
                    },
                );

    my $node      = delete $args{node};
    my $node_name = $node->node();
    $Cache{$node_name}{node} = $node
                    if (!defined $Cache{$node_name}{node});
    return;
}

#-----------------------------------------------------------------------------#
#
# PRIVATE METHOD
#
# Method    : _get_binary_filer_path_string
#
# Purpose   : Prefix the standard filer path to a binary name if required
#
# Parameters:
#   binary  : [Required]: the binary name
#
# Returns:
#   SCALAR, a string containing the filer-path + binary name
#
#-----------------------------------------------------------------------------#

sub _get_binary_filer_path_string {

    my %args = Params::Validate::validate_with(
                    params => \@_,
                    spec   => {
                        binary_name => { type => SCALAR },
                    },
                );

    my $binary_name = delete $args{binary_name};
    if ($binary_name !~ /\//) {
        # no path was provided, probably just the binary's name
        # build the correct path using the default location
        $binary_name = Binary_Filer_Path() . $binary_name;
    }

    return $binary_name;
}

1;
