# Hey emacs, this is basically -*- perl -*-
# vim: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:
#
###############################################################################
# Copyright (c) 2013 NetApp, Inc.
# All rights reserved.
###############################################################################

=pod

=head1 NAME

Os::ClusfsUtils

=for html <hr>

=head1 DESCRIPTION

This library provides standalone utility methods to aid clusfs testing.
This includes verification of access and contents in /clus, setting
testpoints, collecting stats and debug information, etc.

=for html <hr>

=head1 SYNOPSIS

# List "/clus" for only 1 volume on a vserver
my $output = Os::ClusfsUtils::list_clus_basic(
                    node    => $node,
                    vserver => $vserver,
                    volume  => $volume,
                );

# List "/clus" using all possible variations
my $output = Os::ClusfsUtils::list_clus_all(
                    node    => $node,
                );

# List "/clus" for vserver "vs_test1" and volume "testvol"
my $output = Os::ClusfsUtils::list_clus_all(
                    node    => $node,
                    vserver => "vs_test1",
                    volume  => "testvol",
                );

# Run svstat -e
my $raw_output = Os::ClusfsUtils::svstat(
                    node           => $node,
                    return_data    => \%parsed_data,
                    'extended-rpc' => 1,
                );

# Run svstat -w
my $raw_output = Os::ClusfsUtils::svstat(
                    node        => $node,
                    return_data => \%parsed_data,
                    interval    => "0.1",
                );

# View SpinVFS stats (run 'sysctl spinvfs')
my $raw_output = Os::ClusfsUtils::spinvfs_stats(node => $node);
$Test->nlog("Raw output is:$raw_output\n");

# View ClusFS stats (run 'sysctl clusfs')
my $raw_output = Os::ClusfsUtils::clusfs_stats(node => $node);
$Test->nlog("Raw output is:$raw_output\n");

# View the volcache (run 'sysctl clusfs.clusfs_volcache)
my $raw_output = Os::ClusfsUtils::clusfs_volcache(node => $node);
$Test->nlog("Raw output is:$raw_output\n");

=cut

package Os::ClusfsUtils;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;
use Data::Dumper qw(Dumper);

use Options;
use Params::Validate qw(:all);

use Tharn;
use NATE::BaseException;

# NACL
use NACL::C::Node;

###############################################################################
#                   Helper methods for routine tasks
###############################################################################

=for html <hr>

=head2 list_clus_basic

This is a simple listing of a single vserver and/or single volume in clusfs.
For an exhaustive listing of everything under clusfs for multiple vservers and
volumes by name/ID/snapshot/etc., please refer to the list_clus_all() method.

Runs the "ls" command against either:
1) vserver or cserver. 
2) volume or vsadmin along with it's corresponding vserver or cserver (volume only)

=over

=item Options

=over

=item node => $node

 (Required) NACL::C::Node object for the node that will execute the command.

=item vserver => NACL::C::Vserver

 (Optional) Filter: An object of type NACL::C::Vserver

=item volume => NACL::C::Volume|NACL::STask::Volume

 (Optional) Filter: An object of type NACL::C::Volume or NACL::STask::Volume
 With this parameter, the 'vserver' parameter is REQUIRED.

=item 'long-listing' => 1|0

 (Optional) The "-l" flag for "ls", default behavior is to ALWAYS ADD this flag.

=item 'list-all' => 1|0

 (Optional) The "-a" flag for "ls", default behavior is to skip this flag.

=item recursive => 1|0

 (Optional) The "-R" flag for "ls", default behavior is to skip this flag.

=item 'sort-by-time' => 1|0

 (Optional) The "-t" flag for "ls", default behavior is to skip this flag.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # Validate /clus ONLY
 my $output = Os::ClusfsUtils::list_clus_basic(
                    node    => $node,
                 );

 # Validate /clus for a vserver
 my $output = Os::ClusfsUtils::list_clus_basic(
                    node    => $node,
                    vserver => $vserver,
                 );

 # Validate /clus for a volume on a vserver
 my $output = Os::ClusfsUtils::list_clus_basic(
                    node    => $node,
                    vserver => $vserver,
                    volume  => $volume,
                 );

=back

=cut

sub list_clus_basic {

    my %args = Params::Validate::validate(
        @_,
        {   node           => {isa => 'NACL::C::Node',},
            'long-listing' => {

                # ls -l
                type    => SCALAR,
                default => 1,
            },
            'list-all' => {

                # ls -a
                type    => SCALAR,
                default => 0,
            },
            'recursive' => {

                # ls -R
                type    => SCALAR,
                default => 0,
            },
            'sort-by-time' => {

                # ls -t
                type    => SCALAR,
                default => 0,
            },
            vserver => {
                isa      => 'NACL::C::Vserver',
                optional => 1,
            },
            volume => {
                isa      => 'NACL::C::Volume',
                optional => 1,
            },
            cserver => {
                type    => SCALAR,
                default => 0,
            },
            vsadmin => {
                type    => SCALAR,
                default => 0,
            },
        }
    );

    # the volume/vsadmin parameter needs a vserver/cserver parameter
    if ((defined $args{volume} or $args{vsadmin})
        and !(defined $args{vserver} or $args{cserver})) {
        throw NATE::BaseException(
            "Parameter 'volume' or 'vsadmin' was provided, but 'vserver' or 'cserver' wasn't!"
        );
    }

    # only one of vserver and cserver can be specified to this basic method
    if ($args{cserver} and $args{vserver}) {
        throw NATE::BaseException(
            "Specify either 'vserver' or 'cserver', not both!\n");
    }

    # only one of volume and vsadmin can be specified along with the vserver
    if (defined $args{volume} and $args{vsadmin}) {
        throw NATE::BaseException(
            "Specify either 'volume' or 'vsadmin', not both!\n");
    }

 #----------------------------------------------------------------------------
 # get inputs
    my $node    = delete $args{node};
    my $vserver = delete $args{vserver}
        if (defined $args{vserver});
    my $volume = delete $args{volume}
        if (defined $args{volume});
    my $cserver = delete $args{cserver};
    my $vsadmin = delete $args{vsadmin};

 #----------------------------------------------------------------------------
 # build the "ls" command
 # add flags for "ls" if requested
    my $cmd = "ls" . _build_ls_flags(%args) . " /clus";

    # append vserver if needed
    if (defined $vserver) {
        $cmd .= "/" . $vserver->vserver();

        # append cserver if needed
        if ($vsadmin) {
            $cmd .= "/.vsadmin";

        } elsif (defined $volume) {

            # append volume if needed
            $cmd .= $volume->state()->junction_path();
        }
    } elsif ($cserver) {
        $cmd = "sudo ls" . _build_ls_flags(%args) . " /clus";
        $cmd .= "/.cserver/";

        if (defined $volume) {

            # append volume if needed
            $cmd .= $volume->volume();
        }

    }

 #----------------------------------------------------------------------------
 # make the "ls" call in the node's systemshell
    $Log->comment("\nValidating clusfs...\n");
    my $output = $node->get_systemshell_apiset()
        ->execute_raw_command(command => $cmd,);

    # validate the output and check for errors
    _validate_listing(output => $output);
    $Log->debug("No known Clusfs errors were found!\n");

    return $output;

} ## end sub list_clus_basic

=for html <hr>

=head2 list_clus_all

This is an exhaustive listing of everything possible in clusfs, defined by
the scope of the vservers and volumes provided. For a simpler listing of 1
vserver and/or 1 volume, please refer to the list_clus_basic() method

Runs the "ls" command against "/clus/..." in the FreeBSD shell of a node.
By listing [and validating] contents by vserver, vserver-id, cserver contents,
listings for a specific volume, list snapshots, this method validates access
and the contents of clusfs.

Use the same method to list the contents of .cserver (or) .cserver/MDV's combination.

=over

=item Options

=over

=item node => $node

 (Required) NACL::C::Node object for the node that will execute the command.

=item vserver_list => ARRAYREF of NACL::C::Vserver objects

 (Optional) Filter: an array reference containing a list of vserver components to list

=item volume_list => ARRAYREF of NACL::C::Volume objects

 (Optional) Filter: an array reference containing a list of volume components to list
            The vserver names will be extracted from the volume component.
            Note: If both, volume_list and vserver_list are provided, then the vservers
            from the volumes, as well as the provided vserver will be looked at.

=item 'long-listing' => 1|0

 (Optional) The "-l" flag for "ls", default behavior is to ALWAYS ADD this flag.

=item 'list-all' => 1|0

 (Optional) The "-a" flag for "ls", default behavior is to skip this flag.

=item recursive => 1|0

 (Optional) The "-R" flag for "ls", default behavior is to skip this flag.

=item 'sort-by-time' => 1|0

 (Optional) The "-t" flag for "ls", default behavior is to skip this flag.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # Validate /clus for a list of vservers
 Os::ClusfsUtils::list_clus_all(
    node         => $node,
    vserver_list => \@vserver_components,
 );

 # Validate /clus for a list of volumes
 Os::ClusfsUtils::list_clus_all(
    node        => $node,
    volume_list => \@volume_components,
 );

=back

=cut

sub list_clus_all {

    my %args = Params::Validate::validate(
        @_,
        {   node           => {isa => 'NACL::C::Node',},
            'long-listing' => {

                # ls -l
                type    => SCALAR,
                default => 1,
            },
            'list-all' => {

                # ls -a
                type    => SCALAR,
                default => 0,
            },
            'recursive' => {

                # ls -R
                type    => SCALAR,
                default => 0,
            },
            'sort-by-time' => {

                # ls -t
                type    => SCALAR,
                default => 0,
            },
            vserver_list => {
                type      => ARRAYREF,
                default   => [],
                callbacks => {
                    'NACL::C::Vserver' => sub {
                        my @vservers = @{$_[0]};
                        scalar @vservers > 0
                            && (@vservers ==
                            grep { UNIVERSAL::isa($_, 'NACL::C::Vserver') }
                            @vservers);
                    },
                },
            },
            volume_list => {
                type      => ARRAYREF,
                default   => [],
                callbacks => {
                    'NACL::C::Volume' => sub {
                        my @volumes = @{$_[0]};
                        scalar @volumes > 0
                            && (@volumes ==
                            grep { UNIVERSAL::isa($_, 'NACL::C::Volume') }
                            @volumes);
                    },
                },
            },
        }
    );

 #----------------------------------------------------------------------------
 # get inputs
 # get the node object
    my $node = delete $args{node};

 #----------------------------------------------------------------------------
 # build path map using the vserver filter (if provided)
    my %path_map;
    my $vservers;

    if (@{$args{vserver_list}}) {

        # get the volume list
        $vservers = delete $args{vserver_list};

        _build_path_map_from_vservers(
            node         => $node,
            vserver_list => $vservers,
            return_data  => \%path_map
        );
    } else {

        # simply delete the keys with empty lists
        delete $args{vserver_list};
    }

 #----------------------------------------------------------------------------
 # build path map using the volume filter (if provided)
    my $volumes;
    if (@{$args{volume_list}}) {

        # get the volume list
        $volumes = delete $args{volume_list};

        # build a path map based on the vservers an junction-paths
        _build_path_map_from_volumes(
            node        => $node,
            volume_list => $volumes,
            return_data => \%path_map
        );
    } else {

        # simply delete the keys with empty lists
        delete $args{volume_list};
    }

 #----------------------------------------------------------------------------
 # make the "ls" call in the node's systemshell
 # add flags for "ls" if requested
    my $ls_flags = _build_ls_flags(%args);

    foreach my $vserver (keys %path_map) {

        $Log->comment("\nValidating clusfs for vserver $vserver...\n");

        my $apiset = $node->get_systemshell_apiset();

        # for each path associated with this vserver
        foreach my $path (@{$path_map{$vserver}}) {
            my $cmd = "ls" . $ls_flags . " " . $path;
            my $output = $apiset->execute_raw_command(command => $cmd,);

            # validate the output and check for errors
            _validate_listing(output => $output);
            $Log->comment("Clusfs validation passed for vserver $vserver!\n");
        }
    }
    return;

} ## end sub list_clus_all

=for html <hr>

=head2 svstat

Runs the "svstat" command in the FreeBSD shell of a node.
Wrapper around the 'NACL::APISet::Host::CLI::Freebsd::default::svstat' package.
For a full list of options for svstat(), please see the above package.
Some common options are listed below.

=over

=item Options

=over

=item node => $node

 (Required) NACL::C::Node object for the node that will execute the command.

=item display-spinvfs => 1

 (Optional) Run svstat with "-v"

=item rpc-stats => 1

 (Optional) Run svstat with "-r"

=item extended-rpc => 1

 (Optional) Run svstat with "-e"

=item interval => SCALAR

 (Optional) Run svstat with "-w"

=item %other_args

 (Optional) any additional args supported by the systemshell APIset's
            execute_raw_command() method.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # Run svstat -v
 my $raw_output = Os::ClusfsUtils::svstat(
                    node              => $node,
                    return_data       => \%parsed_data,
                    'display-spinvfs' => 1,
                );
 $Test->nlog("Raw output is:$raw_output\n");

=back

=cut

sub svstat {

    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            node        => {isa => 'NACL::C::Node'},
            return_data => {
                type    => HASHREF,
                default => [],
            },
        },

        # pass extra args as-is to the svstat() method
        allow_extra => 1,
    );

 #----------------------------------------------------------------------------
 # Steps:
 # 1. Call svstat on the systemshell APIset extracted from the node component:
 #    my $response = $node->get_systemshell_apiset()->svstat(...);
 # 2. Extract and return the raw output:
 #    $response->get_processed_output()
 # 3. If requested, return the parsed output as well:
 #    $response->get_parsed_output()

 #----------------------------------------------------------------------------
 # get the node object
    my $node = delete $args{node};

    # get the hash reference for the return data
    my $ref_return_data;
    my $want_parsed_data = 0;
    if (defined $args{return_data}) {
        $ref_return_data  = delete $args{return_data};
        $want_parsed_data = 1;
    }

 #----------------------------------------------------------------------------
 # run svstat() - pass the rest of the args as-is, let svstat() validate them
    my $response = $node->get_systemshell_apiset()->svstat(%args);

    # is parsed data required?
    if ($want_parsed_data) {

        # parsed output is an array of hashes
        my @parsed_data = @{$response->get_parsed_output()};

        # there is only 1 element in the array, a multi-level hash
        # set the caller's hash reference to this
        %{$ref_return_data} = %{$parsed_data[0]};

       # NOTE:
       # - this step can be further expanded to validate the parsed data.
       # - additional inputs may be required to know what needs to be checked.
    }

    # return the raw output
    return $response->get_processed_output();

} ## end sub svstat

=for html <hr>

=head2 spinvfs_stats

Runs the "sysctl spinvfs" command in the FreeBSD shell of a node.

=over

=item Options

=over

=item node => $node

 (Required) NACL::C::Node object for the node that will execute the command.

=item %other_args

 (Optional) any additional args supported by the systemshell APIset's
            execute_raw_command() method.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # Run 'sysctl spinvfs'
 my $raw_output = Os::ClusfsUtils::spinvfs_stats(node => $node);
 $Test->nlog("Raw output is:$raw_output\n");

=back

=cut

sub spinvfs_stats {

    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {node => {isa => 'NACL::C::Node'},},

        # pass extra args as-is to the execute_raw_command()
        # method in the systemshell APIset
        allow_extra => 1,
    );

 #----------------------------------------------------------------------------
 # get the node object
    my $node = delete $args{node};

 #----------------------------------------------------------------------------
 # run 'sysctl spinvfs' - pass the rest of the args as-is
    return $node->get_systemshell_apiset()->execute_raw_command(
        command => 'sysctl spinvfs',
        %args,
    );
} ## end sub spinvfs_stats

=for html <hr>

=head2 clusfs_stats

Runs the "sysctl clusfs" command in the FreeBSD shell of a node.

=over

=item Options

=over

=item node => $node

 (Required) NACL::C::Node object for the node that will execute the command.

=item %other_args

 (Optional) any additional args supported by the systemshell APIset's
            execute_raw_command() method.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # Run 'sysctl clusfs'
 my $raw_output = Os::ClusfsUtils::clusfs_stats(node => $node);
 $Test->nlog("Raw output is:$raw_output\n");

=back

=cut

sub clusfs_stats {

    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {node => {isa => 'NACL::C::Node'},},

        # pass extra args as-is to the execute_raw_command()
        # method in the systemshell APIset
        allow_extra => 1,
    );

 #----------------------------------------------------------------------------
 # get the node object
    my $node = delete $args{node};

 #----------------------------------------------------------------------------
 # run 'sysctl clusfs' - pass the rest of the args as-is
    return $node->get_systemshell_apiset()->execute_raw_command(
        command => 'sysctl clusfs',
        %args,
    );
} ## end sub clusfs_stats

=for html <hr>

=head2 clusfs_volcache

Runs the "sysctl clusfs.clusfs_volcache" command in the FreeBSD shell
of a node to view the ClusFS volcache.

=over

=item Options

=over

=item node => $node

 (Required) NACL::C::Node object for the node that will execute the command.

=item return_data => $hash_ref

 (Optional) a reference to an empty HASH if parsed data is required.

=item %other_args

 (Optional) any additional args supported by the systemshell APIset's
            execute_raw_command() method.

=back

=item Returns

 RAW output if successful, throws an exception on failure.

=item Examples

 # simple call to view volcache
 my $raw_output = Os::ClusfsUtils::clusfs_volcache(node => $node);
 $Test->nlog("Raw output is:$raw_output\n");

 # request parsed data
 my %parsed_data;
 my $raw_output = Os::ClusfsUtils::clusfs_volcache(
                    node        => $node,
                    return_data => \%parsed_data,
                );
 $Test->nlog("Raw output is:$raw_output\n");
 $Test->nlog("Parsed output for vserver 'vs0':\n" .
             Dumper($parsed_data{'vs0'}));

=back

=cut

sub clusfs_volcache {

    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            node        => {isa => 'NACL::C::Node'},
            return_data => {
                type     => HASHREF,
                optional => 1,
            },
        },

        # pass extra args as-is to the execute_raw_command()
        # method in the systemshell APIset
        allow_extra => 1,
    );

 #----------------------------------------------------------------------------
 # get the node object
    my $node        = delete $args{node};
    my $return_data = delete $args{return_data};

 #----------------------------------------------------------------------------
 # run 'sysctl clusfs' - pass the rest of the args as-is
    my $output = $node->get_systemshell_apiset()->execute_raw_command(
        command => 'sysctl clusfs.clusfs_volcache',
        %args,
    );

 #----------------------------------------------------------------------------
 # parse the output if requested
    if (defined $return_data) {

# sample output:
#
# clusfs.clusfs_volcache:
# Name                          flags VsID    DSID     MSID    RefCnt    Gen           bladeID                     State
# .cserver                         2  -1        0          0     1        1  00000000-0000-0000-0000-000000000000   RDY
# osk-svt34                        1  -1        0          0     1       19  00000000-0000-0000-0000-000000000000   RDY
# vs_pepsi                         1   3     1025 2147484673     1       19  3a7df17b-1bd1-11e3-bd36-9d87c82d86c7   RDY

        my @lines = split(/\n/, $output);
        my @keys;
        foreach my $line (@lines) {
            next if ($line =~ /clusfs_volcache/);    # ignore the 1st line
            next if ($line eq '');                   # ignore empty lines

            # build hash keys based on the header line
            if ($line =~ /RefCnt/) {
                @keys = split(/\s+/, $line);

            } else {

                # process this line
                my @values = split(/\s+/, $line);
                foreach my $i (1 .. (scalar(@keys) - 1)) {
                    $return_data->{$values[0]}->{$keys[$i]} = $values[$i];
                }
            }
        }    # end of main 'foreach' loop

    }    # end of 'parse return_data'

    return $output;
} ## end sub clusfs_volcache

###############################################################################
#                       Internal helper methods
###############################################################################

#-----------------------------------------------------------------------------#
# METHOD  : _build_path_map_from_volumes
#
# SUMMARY : Takes a list of volume STask/component objects, extracts their
#           vservers (if associated with any), and builds a path to access
#           them via "ls" using their junction paths. If the volume is not
#           mounted, then there is no point in trying to list them.
#
# INPUTS  :
#           volume_list => \@array
#               [Required] A reference to an array containing NACL::C::Volume
#               or NACL::STask/Volume objects.
#           return_data => \%hash
#               [Required]  A reference to an empty hash. It will have the
#               following structure (vserver => @array_of_paths):
#               {
#                   'vs0' => [
#                       '/clus/vs0', '/clus/vs0/vol1', '/clus/vs0/vol2'
#                       '/clus/vs0/vol1/vol3', ...
#                   ]
#               }
#
# RETURNS : NONE if successful, exception for errors.
#
#-----------------------------------------------------------------------------#

sub _build_path_map_from_volumes {
    my %args = Params::Validate::validate(
        @_,
        {   node        => {isa  => 'NACL::C::Node',},
            return_data => {type => HASHREF,},
            volume_list => {
                type      => ARRAYREF,
                callbacks => {
                    'NACL::C::Volume' => sub {
                        my @volumes = @{$_[0]};
                        scalar @volumes > 0
                            && (@volumes ==
                            grep { UNIVERSAL::isa($_, 'NACL::C::Volume') }
                            @volumes);
                    },
                },
            },
        }
    );

    # get inputs
    my $node        = $args{node};
    my @volumes     = @{$args{volume_list}};
    my $return_data = $args{return_data};

    #----------------------------------------------------------------------
    # build a map of the following structure
    #
    # {
    #   'vs0' => [
    #       '/clus/vs0/vol1',
    #       '/clus/vs0/vol1/.snapshot',
    #       '/clus/vs0/vol2',
    #       '/clus/vs0/vol2/.snapshot',
    #   ],
    #   '__vserver_ID__3' => [      # if vserver ID for vs0 = 3
    #       '/clus/.vserver_3/vol1',
    #       '/clus/.vserver_3/vol1/.snapshot',
    #       '/clus/.vserver_3/vol2',
    #       '/clus/.vserver_3/vol2/.snapshot',
    #   ],
    #   'vs1' => [
    #       '/clus/vs1/vol1',
    #       '/clus/vs0/vol1/.snapshot',
    #       '/clus/vs1/vol2',
    #       '/clus/vs0/vol2/.snapshot',
    #   ],
    #   '__vserver_ID__4' => [      # if vserver ID for vs1 = 4
    #       '/clus/.vserver_4/vol1',
    #       '/clus/.vserver_4/vol1/.snapshot',
    #       '/clus/.vserver_4/vol2',
    #       '/clus/.vserver_4/vol2/.snapshot',
    #   ],
    #   'cserver' => [
    #       '/clus/.cserver',
    #   ],
    # }
    #
    foreach my $volume (@volumes) {

       #----------------------------------------------------------------------
       # get the junction path (throw an exception if not found)
        my $junction_path = $volume->state()->junction_path();

        # check if the junction-path exists and is a valid one
        if (!defined $junction_path or $junction_path eq '-') {
            throw NATE::BaseException("Volume "
                    . $volume->volume()
                    . " is not mounted to a junction path!\n");
        }

       #----------------------------------------------------------------------
       # get the vserver
        my $vserver_name = $volume->state()->vserver();
        if (!defined $vserver_name or $vserver_name eq '-') {
            throw NATE::BaseException("Volume "
                    . $volume->volume()
                    . " is not associated with any vserver!\n");
        }

       #----------------------------------------------------------------------
       # build the volume's path and add it to the path map
        my $path = '/clus/' . $vserver_name . $junction_path;
        push(@{$return_data->{$vserver_name}}, $path);
        $Log->debug("Volume path for clusfs test: $path");

       #----------------------------------------------------------------------
       # build the volume's snapshot path and add it to the path map
        $path .= '/.snapshot';
        push(@{$return_data->{$vserver_name}}, $path);
        $Log->debug("Volume snapshot path for clusfs test: $path");

       #----------------------------------------------------------------------
       # get the vserver ID
        my $vserver_id = _get_vserver_id_by_name(
            node         => $node,
            vserver_name => $vserver_name
        );

        # build a path using the vserver ID as well
        my $path_by_id;
        my $key_vs_id = '__vserver_ID__' . $vserver_id;
        if (!defined $return_data->{$key_vs_id}) {
            $path_by_id = '/clus/.vserver_' . $vserver_id . $junction_path;
            push(@{$return_data->{$key_vs_id}}, $path_by_id);

        } else {
            $path_by_id = $$return_data{$key_vs_id}->[0];
        }
        $Log->debug(
                  "Volume path for clusfs test for vserver $vserver_name by "
                . "vserver ID: $path_by_id");

      #----------------------------------------------------------------------
      # build the volume's snapshot path by vserver ID, add it to the path map
        $path_by_id .= '/.snapshot';
        push(@{$return_data->{$vserver_name}}, $path_by_id);
        $Log->debug("Volume snapshot path for clusfs test for vserver "
                . "$vserver_name by vserver ID: $path_by_id: $path_by_id");

       #----------------------------------------------------------------------
       # build the path for .cserver and add it to the path map
        if (!defined $return_data->{'cserver'}) {

            # add this entry only if it hasn't been added yet, say,
            # by _build_path_map_from_volumes()
            my $path_for_cserver = '/clus/.cserver';

            push(@{$return_data->{'cserver'}}, $path_for_cserver);
            $Log->debug(
                "Path for clusfs test for .cserver: $path_for_cserver");
        }

    }

    $Log->debug(sub { "Path map:\n" . Dumper($return_data) });

    return;
} ## end sub _build_path_map_from_volumes

#-----------------------------------------------------------------------------#
# METHOD  : _build_path_map_from_vservers
#
# SUMMARY : Takes a list of vserver components, and builds a path to access
#           them via "ls" using their junction paths.
#
# INPUTS  :
#           vsrever_list => \@array
#               [Required] A reference to an array containing NACL::C::Vserver
#               components.
#           return_data => \%hash
#               [Required]  A reference to an empty hash. It will have the
#               following structure (vserver => @array_of_paths):
#               {
#                   'vs0' => [
#                       '/clus/vs0', '/clus/vs1', ...
#                   ]
#               }
#
# RETURNS : NONE if successful, exception for errors.
#
#-----------------------------------------------------------------------------#

sub _build_path_map_from_vservers {
    my %args = Params::Validate::validate(
        @_,
        {   node         => {isa  => 'NACL::C::Node',},
            return_data  => {type => HASHREF,},
            vserver_list => {
                type      => ARRAYREF,
                callbacks => {
                    'NACL::C::Vserver' => sub {
                        my @vservers = @{$_[0]};
                        scalar @vservers > 0
                            && (@vservers ==
                            grep { UNIVERSAL::isa($_, 'NACL::C::Vserver') }
                            @vservers);
                    },
                },
            },
        }
    );

    # get inputs
    my $node        = $args{node};
    my @vservers    = @{$args{vserver_list}};
    my $return_data = $args{return_data};

    #----------------------------------------------------------------------
    # build a map of the following structure
    #
    # {
    #   'vs0' => [
    #       '/clus/vs0',
    #       '/clus/vs0/.snapshot',
    #   ],
    #   '__vserver_ID__3' => [      # if vserver ID for vs0 = 3
    #       '/clus/.vserver_3',
    #       '/clus/.vserver_3/.snapshot',
    #   ],
    #   'vs1' => [
    #       '/clus/vs1',
    #       '/clus/vs1/.snapshot',
    #   ],
    #   '__vserver_ID__4' => [      # if vserver ID for vs1 = 4
    #       '/clus/.vserver_4',
    #       '/clus/.vserver_4/.snapshot',
    #   ],
    #   'cserver' => [
    #       '/clus/.cserver',
    #   ],
    # }
    #
    foreach my $vserver (@vservers) {

       #----------------------------------------------------------------------
       # get the vserver's name
        my $vserver_name = $vserver->vserver();

        # build the path and add it to the path map
        my $path = '/clus/' . $vserver_name;
        push(@{$return_data->{$vserver_name}}, $path);
        $Log->debug("Path for clusfs test by vserver name: $path");

       #----------------------------------------------------------------------
       # build the vserver's snapshot path and add it to the path map
        $path .= '/.snapshot';
        push(@{$return_data->{$vserver_name}}, $path);
        $Log->debug("Vserver snapshot path for clusfs test: $path");

       #----------------------------------------------------------------------
       # get the vserver ID
        my $vserver_id = _get_vserver_id_by_name(
            node         => $node,
            vserver_name => $vserver_name
        );

        # build a path using the vserver ID as well
        my $path_by_id;
        my $key_vs_id = '__vserver_ID__' . $vserver_id;
        if (!defined $return_data->{$key_vs_id}) {
            $path_by_id = '/clus/.vserver_' . $vserver_id;
            push(@{$return_data->{$key_vs_id}}, $path_by_id);

        } else {
            $path_by_id = ${$return_data->{$key_vs_id}}->[0];

        }
        $Log->debug("Path for clusfs test by vserver ID: $path_by_id");

       #----------------------------------------------------------------------
       # build the vserver's snapshot path by ID and add it to the path map
        $path_by_id .= '/.snapshot';
        push(@{$return_data->{$vserver_name}}, $path_by_id);
        $Log->debug("Vserver snapshot path for clusfs test: $path_by_id");

       #----------------------------------------------------------------------
       # build the path for .cserver and add it to the path map
        if (!defined $return_data->{'cserver'}) {

            # add this entry only if it hasn't been added yet, say,
            # by _build_path_map_from_volumes()
            my $path_for_cserver = '/clus/.cserver';

            push(@{$return_data->{'cserver'}}, $path_for_cserver);
            $Log->debug(
                "Path for clusfs test for .cserver: $path_for_cserver");
        }

    }

    $Log->debug(sub { "Path map:\n" . Dumper($return_data) });

    return;
} ## end sub _build_path_map_from_vservers

#-----------------------------------------------------------------------------#
# METHOD  : _validate_listing
#
# SUMMARY : Takes output for "ls" and checks it for known errors.
#
# INPUTS  :
#           output => SCALAR
#               [Required] The output string for "ls".
#
# RETURNS : NONE if successful, exception for errors.
#
#-----------------------------------------------------------------------------#

sub _validate_listing {

    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {output => {type => SCALAR},},
    );

    # canned set of errors
    my @errors = (
        'Socket not connected',
        'Stale NFS file handle',
        'No such file or directory',
        'Unknown error: 10008'
    );

    # split the output into multiple lines
    my @lines = split(/\n/, $args{output});

    $Log->comment("Checking for errors...\n");

    my @errors_found;

    # check for errors
    foreach my $error (@errors) {
        my @lines_found = grep(/$error/, @lines);
        if (scalar @lines_found > 0) {

            # error matched - throw an exception!
            throw NATE::BaseException(
                "\nClusfs validation failed! An error string was matched:\n"
                    . "Error string  : $error\n"
                    . "Lines matched : "
                    . join("\n", @lines_found)
                    . "\n");
        }
    }

    $Log->comment("...error check passed!\n");

    return;
} ## end sub _validate_listing

#-----------------------------------------------------------------------------#
# METHOD  : _build_ls_flags
#
# SUMMARY : Reads options, translates them to flags accepted by "ls", and
#           returns a string that can be appended to "ls".
#
# INPUTS  : [Required] HASH containing flags (see POD for list_clus_basic()
#           or list_clus_all()).
#
# RETURNS : SCALAR (string)
#
#-----------------------------------------------------------------------------#

sub _build_ls_flags {

    my %args = Params::Validate::validate(
        @_,
        {   'long-listing' => {

                # ls -l
                type    => SCALAR,
                default => 1,
            },
            'list-all' => {

                # ls -a
                type    => SCALAR,
                default => 0,
            },
            'recursive' => {

                # ls -R
                type    => SCALAR,
                default => 0,
            },
            'sort-by-time' => {

                # ls -t
                type    => SCALAR,
                default => 0,
            },
        }
    );

 #----------------------------------------------------------------------------
 # build the flags' string
    my $ls_flags = "";

    if ($args{'long-listing'} == 1) {

        # ls -l
        $ls_flags .= " -l";
    }
    if ($args{'list-all'} == 1) {

        # ls -a
        $ls_flags .= " -a";
    }
    if ($args{'recursive'} == 1) {

        # ls -R
        $ls_flags .= " -R";
    }
    if ($args{'sort-by-time'} == 1) {

        # ls -l
        $ls_flags .= " -t";
    }

    return $ls_flags;
} ## end sub _build_ls_flags

#-----------------------------------------------------------------------------#
# METHOD  : _get_vserver_id_by_name
#
# SUMMARY : Takes a vserver's name and returns the ID (not UUID) using
#           'debug smdb table vserver_by_name show -vserver xxx'
#
# INPUTS  :
#           vserver => SCALAR
#               [Required] The name of the vserver whose ID is to be found.
#
# RETURNS : SCALAR (vserver ID) if successful, exception for errors of type
#           'NACL::APISet::Exceptions::NoMatchingEntriesException'.
#
#-----------------------------------------------------------------------------#

sub _get_vserver_id_by_name {

    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            node         => {isa  => 'NACL::C::Node'},
            vserver_name => {type => SCALAR},
        },
    );

    my $node         = $args{node};
    my $vserver_name = $args{vserver_name};

    # get the vserver ID and store it in cache
    $Log->debug(
        "Running 'debug smdb table vserver_by_name show' for vserver $vserver_name...\n"
    );
    my $response = $node->apiset()->debug_smdb_table_show(
        table   => 'vserver_by_name',
        vserver => $vserver_name,
    );
    my %data = %{${$response->get_parsed_output()}[0]};
    $Log->debug(
        sub {
            "\nSMDB Table Data:\n" . Dumper(\%data) . "\n";
        }
    );

    my $vserver_id = $data{'id-unsigned'};

    $Log->comment("\nVserver ID for vserver $vserver_name is: $vserver_id\n");

    return $vserver_id;
} ## end sub _get_vserver_id_by_name

1;
