#Hey emacs, this is basically -*- perl -*-
#$Id$
## @summary Interface implementation for coredump functions
## @author dl-kernel-qa
## @status private
## @pod ./coreMgr.pm

package Os::CoreMgr;
use strict;

use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);

use Data::Dumper qw(Dumper);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use POSIX;
use Time::Local;
use Tharn qw($Log);
use Utils qw(make_host make_conn);

################################################################################
# Subroutine Name: new()
# Objective:       Create a new Coredump object
# Details:         See POD documentation at the end of this file
################################################################################
sub new {
    my ($class, %args) = @_;
    my (
		$hosttype,
        $self,
        );
 
    ### bless instance into class
    $self= bless{}, $class;

	### Determine type of node provided and load
	### the appropriate library implementation
	die "Missing required parameter 'Node'" unless (defined $args{node});
	
	$self->{thost_of} = Utils::make_host($args{node});
	$hosttype         = $self->{thost_of}->{hosttype};
	
	if ($hosttype =~ /^filer-cfe$/i) {
		require Os::CoreMgr::CoreMgr_7m;
		$self = Os::CoreMgr::CoreMgr_7m->_new(%args);
	}
	elsif ($hosttype =~ /^filer-ng-cfe/i) {
		require Os::CoreMgr::CoreMgr_Cm;
		$self = Os::CoreMgr::CoreMgr_Cm->_new(%args);
	}
	else {
		die "Host platform '".$hosttype."' is not ",
			"currently supported by Os::Coredump";   
	}
	
    return $self;
}

###############################################################################
# Method:    ems_check
# Objective: verify ems messages 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub ems_check {
	my ($self,%args)=@_;
	my (
		$fail_cnt,
		%event_of,
		@lines,
		@messages,
	);
	
	### Validate input
	($self,%args) = validate_mixed (
		@_,
		[ { type => OBJECT} ],
		{
			events    => {type => ARRAYREF},
			messages  => {type => ARRAYREF,
						 optional=>1, default=>$self->{messages}}
		}
	);
		
	### Initialize variables
	$fail_cnt=0;
	
	%event_of=(
		"coredump_dump_disk_failed"   => "coredump.dump.disk.failed",
		"coredump_dump_blocks"        => "coredump.dump.blocks",
		"coredump_dump_failed"        => "coredump.dump.failed",
		"coredump_findcore_partial"   => "coredump.findcore.partial",
		"coredump_dump_raid_notready" => "coredump.dump.raid.notready",
		"coredump_spare_dump"         => "coredump.spare.dump",
		"coredump_spare_abort"        => "coredump.spare.abort",
		"coredump_sparecore_killed"   => "coredump.sparecore.killed",
		"coredump_dump_started"       => "coredump.dump.started",
		"coredump_dump_time"          => "coredump.dump.time",
		"coredump_dump_write_error"   => "coredump.dump.write.error",
		"coredump_dump_find_disks_failed" => "coredump.dump.find.disks.failed",
		"coredump_dump_find_spare_failed" => "coredump.dump.find.spare.failed",
		"coredump_save_aborted"       => "coredump.save.aborted",
		"mgr_boot_reason_ok"          => "mgr.boot.reason_ok.*",
		"mgr_boot_reason_abnormal" => "mgr.boot.reason_abnormal.* after a panic",
		"coredump_save_attempts_count"=> "coredump.save.attempts.count",
		"coredump_save_create_failed" => "coredump.save.create.failed",
		"coredump_save_error"         => "coredump.save.error",
		"coredump_save_nospace"       => "coredump.save.nospace",
		"coredump_save_started"       => "coredump.save.started",
		"coredump_save_completed"     => "coredump.save.completed",
		"sk_panic"                    => "sk.panic",
	);
	
	### Verify that legal ems values were passed in
	foreach my $event (@{$args{events}}) {
		unless (grep (/^$event$/, (keys %event_of))) {
			$self->_log(msg=>"Invalid ems value passed in: ".$event);
			return 0;
		}
	}
	
	### Check for ems events in messages file
	foreach my $event (@{$args{events}}) {
	
		@lines = grep (/$event_of{$event}/, @{$args{messages}});
		$fail_cnt++ unless (@lines);
		@messages=(@lines) ? @lines : "ERROR: EMS message not found ($event)";
		if ($self->_get_loghdl()) {
			if (@lines){
				foreach my $line (@lines) {
					my ($ems)=$line=~/(\[.*\])/;
					$self->_log(msg=>"ems message found: ".$event,indent=>3);
				}
			}
			else {
				$self->_log(msg=> "ERROR: ems message not found ($event)");
			}
		}
		else {
			$Tharn::Log->comment(@messages);	
		}
	}
	
	return ($fail_cnt==0) ? 1 : 0;
	
}

###############################################################################
# Method:    get_coretool_info
# Objective: return output from running coretool on core
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_coretool_info {
    my ($self,%args) = @_;
    my (
        $bsdObj,
        $cmd,
		$core_of,
        $filter,
		$name,
		$node,
        $output,
        );
    
    ($self,%args) = validate_mixed (
		@_,
		[ { type => OBJECT} ],
		{
			cid     => {type => SCALAR, 
							callbacks=>{
								'legal values'=>sub{shift()=~
								/\d+/i
							}}},     
		} 
	);

    ### Verify that a core name exists with that cid
	$node=$self->get_node();
	$name = $self->{cores_of}{$node}{$args{cid}}{name};
	unless ($name) {
		die "Unable to find corefile with cid '$args{cid}'";
	}
	
	### Retreive bsd connection
    $bsdObj=$self->get_conn_obj(type=>"bsd");
    
	### Execute and store coretool output in hash
    if (not defined $self->{coretool}{$args{cid}}){
        $cmd="/mroot/etc/crash/coretool /mroot/etc/crash/".$name;
        $self->{coretool}{$args{cid}}=$bsdObj->execute_raw_command(command=>$cmd);
    }
	
	$output=$self->{coretool}{$args{cid}};
	($core_of->{filename})=$output=~/filename\s+(\S+)/;
	($core_of->{file_size})=$output=~/file size\s+(\S+)/;
	($core_of->{nr_of_segments})=$output=~/nr of segments\s+(\d+)/;
	($core_of->{missing_blocks})=$output=~/missing blocks\s+(\d+)/;
	($core_of->{model})=$output=~/model=(\S+)/;
	($core_of->{nvram_systemid})=$output=~/nvram_systemid=(\S+)/;
    ($core_of->{hostname})=$output=~/hostname=(\S+)/;
	($core_of->{serial_number})=$output=~/serialnumber=(\S+)/;
	($core_of->{sw_arch})=$output=~/sw_arch=(\S+)/;
	
    return $core_of;
}

###############################################################################
# Method:    get_num_cpus
# Objective: retrieve number of active cpus
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_num_cpus {
	my ($self) = @_;
	my (
		$cmd,
		$num,
		);
	
	$num = undef;
	
	$cmd  = "sysctl -a | egrep \"^hw.ncpu\" | awk '{print \$2}'";
	$num = $self->{apiObjBsd}->execute_raw_command(command=>$cmd);
	
	return $num;
}

###############################################################################
# Method:    get_memory
# Objective: retrieve ontap memory
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_memory {
	my ($self) = @_;
	my (
		$cmd,
		$size,
		@lines,
		);
	
	$size = undef;
	
	### If not restricted get actual memory
	$cmd  = "sysctl -a | egrep \"^hw.ontap_physmem\" | awk '{print \$2}'";
	$size = $self->{apiObjBsd}->execute_raw_command(command=>$cmd);
	$size = POSIX::ceil($size/1024**3)."G";
	
	return $size;
}
###############################################################################
# Method:    get_node
# Objective: return node name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_node {
	my ($self)=@_;
	
	### strip off any metadata appended to the node name
	if ($self->{node}=~/;/){
		$self->{node}=~s/;.*//g;
	}
	
	return $self->{node};
}

###############################################################################
# Method:    get_version
# Objective: retrieve ontap version
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_version {
	my ($self)=@_;
	return $self->{version_of}->{ontap_version};	
}


###############################################################################
# Method:    is_debug_build
# Objective: determine whether a debug build is loaded
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_debug_build {
	my ($self)=@_;
	
	return ($self->{version_of}{build_info}=~/d/i) ? 1 : 0;
}
###############################################################################
#   P  R  I  V  A  T  E     M  E  T  H  O  D  S 
###############################################################################
# Method:    get_arch
# Objective: return architecture type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_arch{
	my ($self)=@_;
	return $self->{arch};
}

###############################################################################
# Method:    _get_hosttype()
# Objective: Get client host type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_hosttype{
    my($self) = @_;
    return $self->{hostObj}{hosttype};
}

###############################################################################
# Method:    _get_loghdl
# Objective: return log handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_loghdl {
	my ($self)=@_;
	return $self->{loghdl};
}

###############################################################################
# Method: _log
# Objective: display library messages to stdout and natelog
################################################################################
sub _log{
    my ($self,%args) = @_;
    my (
        $bFirst,
        $dash,
        @lines,
        $indent,
        $nocr,
    );

    $indent=0;

    @lines=(ref $args{msg} eq "ARRAY") ? @{$args{msg}} : ($args{msg});
    $indent=(defined $args{indent}) ? $args{indent} : 0;
    $nocr=(defined $args{nocr}) ? $args{nocr} : 0;
    $bFirst=1;
    
    if (defined $self->_get_loghdl()){
        my $loghdl=$self->_get_loghdl();
        if (not $indent){
            foreach my $line (@lines){
                print {$$loghdl} $line.
                (($nocr==1) ? " " : "\n");
            }
        }
        else {
            foreach my $line (@lines){
                $dash = ($bFirst == 1)
                    ? " "x($indent-1)."- "
                    : " "x$indent;
                $bFirst=0;
                print {$$loghdl} $dash.$line.
                (($nocr==1) ? " " : "\n");
            }
        }
        $Log->comment(join("\n",@lines));
      
    }

}

################################################################
# Subroutine Name: _size_to_bytes()
# Objective:       convert size strings to bytes.
# Details:         See POD documentation at the end of this file
################################################################
sub _size_to_bytes {
    my ($self,$size)=@_;
    my (
        $bytes,
        $nbr,
        $unit
    );
    
    ### If no units then byte 
    $size .= "b" if ($size=~/^\d+$/);
   
    ### Separate number and units
    ($nbr,$unit)=$size=~/^(\d+\.?\d*)\s{0,}(\w+)/;
    return $nbr if ($unit=~/b/i);

    if ($unit =~/^k/i) {
     $bytes = $nbr*1024;
    }
    elsif ($unit =~/^m/i) {
     $bytes = $nbr * 1024**2;
    }
    elsif ($unit=~/^g/i){
     $bytes = $nbr * 1024**3;
    }
    elsif ($unit=~/^t/i){
     $bytes = $nbr * 1024**4;
    }
    return int($bytes);
}


###############################################################################
# Method:    _wait_for_dump
# Objective: wait for dumpcore operation to complete
###############################################################################
sub _wait_for_dump {
	my ($self, %args)=@_;
	my (
		$dot_cnt,
		$host_of,
		$start_flag,
		$timeout,
		$t1,
		$t2,
		$wait,
		%result_of
	);
	
	### Validate input arguments
	($self,%args) = validate_mixed (
		@_,
		[ { type => OBJECT} ],
		{
			wait   => {type => SCALAR, optional=>1, default=>1,
							callbacks=>{
								'legal values'=>sub{shift()=~
								/1|0/i
							}}},                     
		}
	);
	
	### initialize variables
	$host_of = $self->{console_conn}->{hostp};
	$start_flag = 0;
	$timeout = 30;
	$wait=$args{wait};
	
	%result_of = (
		disks     => undef,
		diskerrs  => 0,
		compressed => undef,
		duration => undef,
		message => "dump did not complete ",
		recursive => 0,
		starts    => 0,
		status    => 0,
		timestamp => undef,
	);
	
	### monitor console messages 
	while (1) {
		my $line =
			&Tharn::say(
				$self->{console_conn}, "",
				".*\r",
				-noprompt,
				-nocr,
				-timeout=>$timeout,
				-eval_error);
		
		if ($@ && $@ =~ /Err=TIMEOUT/i) {
			$self->_log(msg=>"timed out",indent=>5);
			$result_of{status} = 0;
			$result_of{message} = "dump did not complete successfully";
			last;
		}
		elsif ($line =~ /DUMPCORE: START/) {
			$t1=timelocal(localtime());
			$self->_log(msg=>$line,indent=>5);
			$result_of{starts}++;
			$start_flag = 1;
		}
		elsif ($line =~ /coredump written/) {
			$t2=timelocal(localtime());
			$self->_log(msg=>$line,indent=>5);
			if (defined $t2 and defined $t1){
				$result_of{duration}=$t2-$t1;
				$self->_log(msg=>"dumpcore duration: ".
							$result_of{duration}."s",indent=>5);
			}
			$result_of{status} = 1;
			$result_of{message} = "dump did complete successfully";
			last;
		}
		elsif ($line =~ /coredump \*NOT\* written/) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{status} = 0;
			$result_of{message} = "dump did not complete successfully";
			last;
		}
		elsif ($line =~ /Partner has aborted Sparecore/) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{status} = 0;
			$result_of{message} = "dump aborted by partner";
			last;
		}
		elsif ($line =~ /^HALT: cluster partner has taken over/) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{status} = 3;
			$result_of{message} = "cluster partner has take over";
			last;
		}
		elsif ($line =~ /(\d+) blocks are needed .* dump/) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{dump_blocks}=$1;
		}
		elsif ($line =~ /must do compression now/i){
			$self->_log(msg=>$line,indent=>5);
			$result_of{compressed}=1;
			if ($wait==0) {
				$result_of{status}=1;
				#$result_of{message} = "dump did complete successfully";
				$result_of{message} = "dump success unknown due to param wait=>0";
				last;
			}
		}
		elsif ($line =~ /compression deferred/i){
			$self->_log(msg=>$line,indent=>5);
			$result_of{compressed}=0;
			if ($wait==0) {
				$result_of{status}=1;
				#$result_of{message} = "dump did complete successfully";
				$result_of{message} = "dump success unknown due to param wait=>0";				
				last;
			}
		}
		elsif ($line =~ /No eligible disks found|Writing panic info/i or
		        $line =~ /Sparecore restart not allowed/i or
				$line =~ /Raid assimilation incomplete/i or
				$line =~ /Testpoint hit/i
				) {
			
			$self->_log(msg=>$line,indent=>5);
		}
		elsif ($line =~ /PANIC:.* on release .* on (.*)/) {
			$self->_log(msg=>$line,indent=>5);
			my $buf=$1;
			$buf=~s/\r|\n//g;
			$result_of{timestamp} = $buf;
		}
		elsif ($line =~ /recursive PANIC:/i) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{recursive}++;
			$timeout = 10;
		}
		elsif ($line =~ /(Forcing dump write error on disk.*)/) {
			$self->_log(msg=>$1,indent=>5);
			$result_of{diskerrs}++;
		}
		elsif ($line =~ /Dumping to disks: (.*)/) {
			$self->_log(msg=>$line,indent=>5);
			my $buf=$1;
			$buf=~s/\r|\n//g;
			$result_of{starts}++ if (!$start_flag);
			$start_flag = 0;
			$result_of{disks} = $buf;
			$timeout = -1;
		}		
		elsif ($line =~ /coredump.dump.started:info\]: ((\S+).*)/){
			$self->_log(msg=>"LINE: ".$line);
			$self->_log(msg=>$1,indent=>5);
			my $dump_type=$2;
			$self->_log(msg=>"DUMPTYPE:".$dump_type);
			$dump_type=~s/core//g;
			$result_of{dump_type}=lc($dump_type);
		}
		elsif ($line =~ /Testpoint hit:.*\[dumpcore, watchdog_timeout\]/) {
			$self->_log(msg=>$line,indent=>5);
			halt($self->{console_conn}) or logresult("FATAL", $@);
			$result_of{status} = 0;
			$result_of{msg} = "watchdog timeout";
			$result_of{watchdog} = 1;
			last;
		}
		elsif ($line =~ /Testpoint hit/) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{status} = 0;
		}
		elsif ($line =~ /Transitioning to SF_SPARECORE state/) {
			$self->_log(msg=>$line,indent=>5);
			$result_of{sparecore} = 1;
		}
	}
	
	### return as hash reference.
	delete $result_of{dump_duration} if (not defined $result_of{dump_duration});
	return %result_of;
}

1;

############################################################################
#  P O D
############################################################################

=pod

=head1 Package Name

Os::CoreMgr 

=head1 Author

dl-kernel-qa@netapp.com

=head1 Description

Os::CoreMgr provides a API interface for issuing common ONTAP commands
related to dumping and saving cores.  

=head1 See Also

Package instantiates one of the following modules depending on whether
its 7m or cmode hosttype. See these modules for implementation detail.

B<CoreMgr_7m.pm> for nodes running ONTAP 7m 

OR 

B<CoreMgr_Cm.pm> for nodes running ONTAP Cmode

=head1 Public Methods

=head2 B<new( B<node>, I<loghdl, ctool_loc, save_onboot, sparse> )>

Creates a new OS::CoreMgr object. 

=over 4

=item Inputs

=over 4

=item * B<node>

Name of node to use for executing library routines 

=item * I<loghdl> 

Log handle from test to use for logging messages to both STDOUT and NATE from 
library function 

=item * I<ctool_loc> (default: /x/eng/localtest/noarch/share/coretool)

Location of coretool executable to upload to node filesystem directory /mroot/etc/crash 

=item * I<save_onboot> (default: 1)

Flag indicating whether unsaved cores should automatically be saved on reboot.

=item * I<sparse> (default: 1)

Flag indicating whether cores should be sparse or non sparse

=back

=item Outputs

Returns a Os::CoreMgr object.

=item Examples

 ### Create a PTS object
 use TCD;
 use CoreMgr;
 ...
 try {
   $Test->nlog("- Create Os::CoreMgr object");
   my $cmObj=Os::CoreMgr->new(
	   loghdl  => $Test->get_loghdl(),
	   node    => $node_name,
	   timeout => 60
   );
 }
 catch NATE::Error with {
 	my $ex=$@;
  	$Test->nlog("ERROR:".Dumper($ex));
	$rc=0;
 };

=back


=head2 B<delete_unsaved_cores( I<partner,cid> )>

Removes any unsaved cores from disk core area

=over 4 

=item Inputs

=over 4

=item * I<cid> 

Core id of core to remove

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->delete_unsaved_cores();

=back


=head2 B<delete_saved_cores( I<partner> )>

Removes any saved cores stored on mroot filessytem

=over 4

=item Inputs

=over 4

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->delete_saved_cores();

=back


=head2 B<destroy()>

Destroy the Os::CoreMgr instance. As part of destroying the instance it
will change the following node parameters back to their defaults;
save_onboot, sparse_cores, auto_giveback, minfree_limit, and
dump_attempts.

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

none

=item Examples

 $CMObj->destroy();

=back



=head2 B<dump_attempts( I<value> )>

Sets or retrieves the number of dump attempts value

=over 4 

=item Inputs

=over 4

=item * I<value>

Number of attempts to try when dumping core

=back

=item Outputs

Returns back the number of attempts to try 

=item Examples

 $CMObj->dump_attempts(value=>2);

=back


=head2 B<dumpcore( I<partner, reason, wait> )>

Dumps system core

=over 4 

=item Inputs

=over 4

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=item * I<reason> (default: 'test')

message tag to include with panic string

=item * I<wait> (default: 1)

flag indicating whether to wait for dumpcore to complete before
returning control back to test

=back

=item Outputs

Returns the core_id on success and 0 on failure

=item Examples

 $CMObj->dumpcore();

=back


=head2 B<ems_check( B<events>, I<messages> )>

Check whether specified ems message events occured 

=over 4 

=item Inputs

=over 4

=item * B<events> 

An array reference of ems events to look for.  Current supported
events are;

=begin html

<pre><small>
coredump_dump_disk_failed,
coredump_dump_blocks,
coredump_dump_failed,
coredump_findcore_partial,
coredump_dump_raid_notready,
coredump_spare_dump,
coredump_spare_abort,
coredump_sparecore_killed,
coredump_dump_started,
coredump_dump_time,
coredump_dump_write_error,
coredump_dump_find_disks_failed,
coredump_dump_find_spare_failed,
coredump_save_aborted,
mgr_boot_reason_ok,
mgr_boot_reason_abnormal,
coredump_save_attempts_count,
coredump_save_create_failed,
coredump_save_error,
coredump_save_nospace,ore
coredump_save_started,
coredump_save_completed,
sk_panic
</pre></small>

=end html

=item * I<messages> (default: all messages since log_marker() call)

An array reference of messages 

=back

=item Outputs

Returns 1 on success and 0 on failure

=item Examples

	$marker=$CMObj->log_marker(name=>"TEST");
 	@messages=$CMObj->read_messages(marker=>$marker);
 
	### Verify ems messages   
	$CMObj->ems_check(
		events=>[qw(
			sk_panic
			coredump_dump_started
			coredump_dump_blocks
			coredump_dump_time
			mgr_boot_reason_abnormal
			coredump_save_started
			coredump_save_completed
				 )],
		messages=>[@messages]
	) or $fail_cnt++;

=back


=head2 B<find_dump_type( I<partner> )>

Type of core that will be produced when core is dumped

=over 4 

=item Inputs

=over 4

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns back the number of attempts to try 

=item Examples

 $CMObj->find_dump_type();

=back 


=head2 B<get_available_space()>

Amount of available space in mroot in bytes

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns available space in bytes

=item Examples

 my $bytes=$CMObj->get_available_space();

=back 


=head2 B<get_conn_obj( B<type>, I<nologin> )>

Retrieve connection handle of requested type

=over 4 

=item Inputs

=over 4

=item * B<type> (values: cli|bsd|console)

Type of connection to return 

=item * I<nologin> (default: 0)

Determine whether to log back into connection

=back

=item Outputs

Returns connection handle

=item Examples

 my $connObj=$CMObj->get_core_objs(type=>"cli");

=back 


=head2 B<get_core_cnt( I<status> )>

Number of cores on the node

=over 4 

=item Inputs

=over 4

=item * I<status> (values: saved|unsaved)

Returns the total of all saved and unsaved cores unless status
filter is applied

=back

=item Outputs

Returns core count

=item Examples

 my $cnt=$CMObj->get_core_cnt(status=>"saved");

=back



=head2 B<get_core_ids(  )>

Retrieve a list of all core ids (saved and unsaved)

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns array or arrayref of core ids

=item Examples

 my @cids=$CMObj->get_core_ids();

=back


=head2 B<get_core_info( B<cid> )>

Retrieve a hash of core metadata.  

=over 4 

=item Inputs

=over 4

=item * B<cid> 

The core id 

=back

=item Outputs

Returns a hash reference of dumpcore metadata.  The following data is
returned; 

=begin html

<pre>
        name
        status
        coreid
        timestamp
		 version
        panic_string
        size
</pre>

=end html

=item Examples

 my $cid=$CMObj->get_last_cid();
 my $core_hsh=$CMObj->get_core_info(cid=>$cid);
 $Test->nlog("Core name: ".$core_hsh->{name});

=back


=head2 B<get_coretool_info( B<cid> )>

Retrieve core information using coretool

=over 4 

=item Inputs

=over 4

=item * B<cid>

Core id

=back

=item Outputs

Returns a hash reference of coretool metadata.  The following data is
returned; 

=begin html

<pre>
        filename
        file_size
        nr_of_segments
        missing_blocks
        model
        nvram_systemid
        hostname
        serial_number
        sw_arch
</pre>

=end html 

=item Examples

 $ctool_of = $CMObj->get_coretool_info(cid=>$cid);	
 missing_blocks = $ctool_of->{missing_blocks};

=back



=head2 B<get_dump_info( cid )>

Retrieve a hash of dumpcore metadata.  

=over 4 

=item Inputs

=over 4

=item * I<cid> (default: last dumpcore operation)

The core id produced by dumpcore operation.  If one is not specified
then the last core id generated is used.  

=back

=item Outputs

Returns a hash reference of dumpcore metadata.  The following data is
returned; 

=begin html

<pre>
        compressed
        diskerrs
        disks
        duration
        message
        recursive
        starts
        status
        timestamp
        type
</pre>

=end html

=item Examples

 my $cid=$CMObj->get_last_cid();
 my $dump_hsh=$CMObj->get_dump_info(cid=>$cid);
 $Test->nlog("Number of dump starts: ".$dump_hsh->{starts});

=back


=head2 B<get_last_cid()>

The core id of the last core dumped by the node

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns a scalar value

=item Examples

 my $cid=$CMObj->get_last_cid();

=back


=head2 B<get_local_disks()>

Local disks associated with node

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns an array or array ref of local disks

=item Examples

 my @disks=$CMObj->get_local_disks();
 foreach my $disk (@disks) {
	$Test->nlog("Disk: ".$disk);
 }

=back

=head2 B<get_memory()>

Amount of physical memory being used for dumping core

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns a scalar value

=item Examples

 my $memory=$CMObj->get_memory();

=back


=head2 B<get_node()>

Node being used by CoreMgr object

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns a scalar value

=item Examples

 my $name = $CMObj->get_node();

=back


=head2 B<get_save_info( I<partner> )>

Returnes savecore information 

=over 4 

=item Inputs

=over 4

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns back a hash reference to save core data containing the
following keys below ...

=begin html

<pre>
        blocks,
        disk_free_cnt,
        disk_used_cnt,
        dump_type,
        unsaved_cnt
</pre>


=end html

=item Examples

	my $hsh=$CMObj->get_save_info();
	$Test->nlog("disk_free_cnt: ".$hsh->{disk_free_cnt});

=back 


=head2 B<get_version()>

Version of ontap installed

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns a scalar value

=item Examples

 my $version = $CMObj->get_version);

=back


=head2 B<giveback( I<force> )>

Issues a storage giveback

=over 4 

=item Inputs

=over 4

=item * I<force> (default: 0)   

Option to override-vetoes that would otherwise prevent giveback from occuring

=back

=item Outputs

Returns 1 on success and 0 on failure

=item Examples

    # Example: Perfom storage givback
    $Test->nlog("- Perfom storage giveback");
    $CMObj->giveback(force=>1);

=back


=head2 B<info_dump( I<cid> )>

Display dump information metatdata 

=over 4 

=item Inputs

=over 4

=item * B<cid>

Core id of dump to display

=back

=item Outputs

Returns an array of dump data to display

=item Examples

    # Example: Perfom storage givback
    $Test->nlog("- Display dump data");
    $CMObj->info_dump(cid=>$CMObj->get_last_cid());
	
=begin html

  <B>Result:</B>
<small><pre>
      Display dump information
        core id:    1288636613
        compressed: 1
        diskerrs:   0
        duration:   254
        message:    dump did complete successfully
        recursive:  0
        starts:     1
        status:     1
        timestamp:  Mon Nov  1 14:36:53 EDT 2010
        disks:      0a.27
<pre></small>

=end html

=back


=head2 B< info_cores( I<cid, partner> ) >

Retrieve information for 1 or more cores in formated form to be
displayed

=over 4 

=item Inputs

=over 4

=item * I<cid>

Core id

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns an array of core data to display

=item Examples

    $Test->nlog("- Display core metadata");
    my @lines = $CMObj->info_cores();
    $Test->nlog(@lines);
	
=begin html


  <B>Result:</B>
<pre>
        - Display core metadata 
        Name                                     Status   Coreid     Size   Timestamp                   
        ---------------------------------------  -------- ---------- ------ ----------------------------
        core.118046026.2010-11-02.05_49_07.nz    unsaved  1288676947     0  Tue Nov  2 01:49:07 EDT 2010
        core.118046026.2010-11-02.05_54_22.nz    unsaved  1288677262     0  Tue Nov  2 01:54:22 EDT 2010
</pre>

=end html
	
	
=back


=head2 B< is_auto_giveback() >

Determine whether auto-giveback is on or off

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns 1 if auto-giveback is on, otherwise 0

=item Examples
    
    my $state=$CMObj->is_auto_giveback();
	
=back


=head2 B< is_core_file( B<cid>, I<partner>) >

Determine if a core file exists for a given core id

=over 4 

=item Inputs

=over 4

=item * B<cid>

Core id

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns 1 if a core file exists, otherwise 0

=item Examples
    
    my $exists=$CMObj->is_core_file(cid=>$cid);
	
=back



=head2 B< is_failover() >

Determine if node is setup for storage failover

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns 1 if failover is enabled, otherwise 0

=item Examples
    
    my $state=$CMObj->is_failover();
    $Test->nlog("failover state: ".$state);
	
=back


=head2 B< is_save_onboot() >

Determine if automatic savecore is enabled on boot

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns 1 if 'save on boot' is enabled, otherwise 0

=item Examples
    
    my $state=$CMObj->is_save_onboot();
    $Test->nlog("save on boot: ".$state);
	
=back


=head2 B<is_sparse_core()>

Determine if sparse core is enabled

=over 4 

=item Inputs

=over 4

none

=back

=item Outputs

Returns 1 if 'spare core' is enabled, otherwise 0

=item Examples
    
    my $state=$CMObj->is_sparse_core();
    $Test->nlog("sparse core ".$state);
	
=back


=head2 B<log_marker( I<marker> )>

Creates a unique log marker with timestamp in node's messages file
for marking where start of test begins

=over 4 

=item Inputs

=over 4

=item * I<marker> (default: 'NATE coredump test')

Marker value to prefix timestamp with (I<marker> - I<localtime>) for
locating starting point of interest in /mroot/etc/log/messages file

=back

=item Outputs

Returns log marker if successful, otherwise 0

=item Examples
    
    $marker_timestamp=$CMObj->log_marker(name=>"test1");
	
=back


=head2 B<mroot_minfree_limit( I<bytes> )>

Retrieve or sets the minimum free bytes that must be available in
in mroot to enable an unsaved core to be saved 

=over 4 

=item Inputs

=over 4

=item * I<bytes> 

Sets the minimum free byte limit in mroot.  For 7m the value is
automatically converted to blocks.  Specifying a value of 0, sets 
the minfree limit to 0 for 7mode and 250m for Cmode.  

=back

=item Outputs

Returns the minfree limit

=item Examples
    
    my $bytes=$CMObj->mroot_minfree_limit();
	$Test->nlog("mroot free space minimum limit: ".$bytes);
	
=back


=head2 B<read_messages( B<marker> )>

Retrieves all messages after a specified marker in the
/mroot/etc/log/messages file.

=over 4 

=item Inputs

=over 4

=item * B<marker> 

Marker value starting point for retrieving messages

=back

=item Outputs

Returns an array or array reference of messages

=item Examples

    $marker=$CMObj->log_marker(name=>'test1');
    @messages=$CMObj->read_messages(marker=>$marker);
    $Test->nlog("Messages: ",@messages);
	
=back


=head2 B<refresh( I<partner> )>

Refresh current core metadata

=over 4 

=item Inputs

=over 4

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->refresh();

=back 


=head2 B<restrict_memory( I<size> )>

Restricts the amount of memory to use for the node. Unit should
be provided with size; M, G.

=over 4 

=item Inputs

=over 4

=item * I<size> (default: 4G)

Retricted memory size to use

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->restrict_memory(memory=>"8G");

=back 


=head2 B< savecore( I<abort, cid, force, partner, synchronous, wait> )>

Save all unsaved cores unless a specific core id (cid) is specified.
B<IMPORTANT> wait_for_savecore() will automatically be called unless 
the option I<wait=>0> is specified.

=over 4

=item Inputs

=over 4

=item * I<abort> 

Value indicating what savecore percentage to abort savecore operation.
For example, abort=>20 says abort savecore operation when savecore
status shows core as 20% or more completed

=item * I<cid> 

Core id to specifically save

=item * I<force> (default: 0)

If savecore fails to save an  unsaved  core too  many times, it will
stop trying.  This flag tells savecore to ignore the  previous
attempts, and try again.  This attempt will be made synchronously.

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=item * I<synchronous> (default: 0)

Starts  savecore  in  a  synchronous  mode, which  blocks  other
processing  until the save  has  completed.   Periodic   progress
updates will be displayed on the console.  This option is only
supported in 7Mode, while in CMode all save core requests are
asynchronous

=item * I<wait> (default: 1)

Flag indicating whether to wait for savecore operation to complete
before returning control back to the test

=back

=item Outputs

Returns 1 on success and 0 on failure.  A success is determined by
whether a core was saved 

=item Examples

 if ($CMObj->savecore()){
	$Test->nlog("savecore operation was successful);
 };

=back 


=head2 B<set_auto_giveback( $bool )>

Set whether auto-giveback is enabled or not 

=over 4 

=item Inputs

=over 4

=item * B<bool> (values: 1|0)

Flag indicating whether to enable or disable auto-giveback

=back

=item Outputs

None

=item Examples
    
  ### Disable failover	
  $CMObj->set_auto_giveback(1);
	
=back


=head2 B<set_failover( $bool )>

Set whether failover is enabled on not 

=over 4 

=item Inputs

=over 4

=item * B<$bool> (values: 1|0)

Flag indicating whether to enable or disable failover

=back

=item Outputs

None

=item Examples
    
  ### Disable failover	
  $CMObj->set_failover(0);
	
=back


=head2 B<set_save_onboot( B<$bool> )>

Set whether savecore is automatically run on reboot

=over 4 

=item Inputs

=over 4

=item * B<$bool> (values: 1|0)

Flag indicating whether to enable or disable auto savecore

=back

=item Outputs

None

=item Examples
    
  ### Enable auto savecore
  $CMObj->set_save_onboot(1);
	
=back

=head2 B<set_sparse_core( B<$bool> )>

Set whether sparse cores are enabled or not

=over 4 

=item Inputs

=over 4

=item * B<$bool> (values: 1|0)

Flag indicating whether to enable or disable sparse cores

=back

=item Outputs

None

=item Examples
    
  ### Enable sparse cores
  $CMObj->set_sparse_core(1);
	
=back


=head2 B< tp_clear() >

Clear all existing test points

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns back object handle ($self) 

=item Examples
    
  ### Clear testpoints
  $CMObj->tp_clear();
	
=back


=head2 B<tp_error( I<module>, B<id, cnt, parm1>, I<parm2> )>

Set a testpoint

=over 4 

=item Inputs

=over 4

=item * I<module> (default: dumpcore)

Module family containing testpoints

=item * B<id> 

Name of testpoint

=item * B<cnt> 

Number of times testpoint may be hit

=item * B<parm1> 

Parameter value 1

=item * I<parm2> 

Parameter value 2

=back

=item Outputs

Returns back object handle ($self) 

=item Examples
    
 $Test->nlog("- Set test point to cause disk io error");
	$CMObj->tp_error(
			module=>"dumpcore",
			id=>"write_error",
			cnt=>1,
			parm1=>1,
			);
	
=back


=head2 B<tp_list()>

List current test points that are set

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns back object handle ($self) 

=item Examples
    
  	my @tps=$CMObj->tp_list();
	$Test->nlog(@tps);
	
=back


=head2 B<unrestrict_memory()>

Removes memory restriction on the node if one exists

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->unrestrict_memory();

=back


=head2 B<wait_for_boot()>

Wait for the node to bootup 

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->wait_for_bootup();

=back 


=head2 B<wait_for_fmdisk()>

Wait for the auto savecore operation to start following a reboot.  The
save core start is determined by when the ems messages 'fmdisk_reserve_spare'
is displayed to the console. This option is for HA configurations only.

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->wait_for_fmdisk();

=back 


=head2 B<wait_for_giveback()>

Waits until giveback of partner node's storage is complete 

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns back object handle ($self)  

=item Examples

 $CMObj->wait_for_fmdisk();

=back 


=head2 B<wait_for_savecore( I<abort, partner, retries, sleep, timeout, timeout_retries, wait> )>

wait for savecore to complete

=over 4 

=item Inputs

=over 4

=item * I<abort> 

Value indicating what savecore percentage to abort savecore operation.
For example, abort=>20 says abort savecore operation when savecore
status shows core as 20% or more completed

=item * I<partner> (default: 0)

Flag indicating whether commmand should be executed against
partner node

=item * I<retries> (default: 1)

Number of times to retry looking for savecore operation in progress.
By default it will try 1 more time after waiting the specified
sleep time.  

=item * I<sleep> (default: 5)

Amount of time to wait in seconds before requerying savecore 
status operation

=item * I<timeout> (default: 30)

Amount of time to wait in seconds for data to be returned from a
savecore status operation

=item * I<timeout_retries> (default: 10)

Number of times to retry operation due to a timeout condition

=item * I<wait> (default: 1)

Flag indicating whether to wait for savecore operation to complete
before returning control back to the test

=back

=item Outputs

Returns 1 on success and 0 on failure.  A success is determined by
whether a core was saved 

=item Examples

 if ($CMObj->wait_for_savecore()){
	$Test->nlog("savecore operation was successful);
 };

=back 

=head2 B<wait_for_takeover()>

wait for taken over node to be ready for giveback 

=over 4 

=item Inputs

=over 4

None

=back

=item Outputs

Returns 1 on success and 0 on failure.  A success is determined by
whether a core was saved 

=item Examples

 if ($CMObj->wait_for_takeover()){
	$Test->nlog("takeover operation was successful);
 };

=back 

=cut

