#Hey emacs, this is basically -*- perl -*-
#$Id$
## @summary Interface implementation for 10m coredump functions
## @author dl-kernel-qa
## @status private
## @pod ../CoreMgr.pm

package Os::CoreMgr::CoreMgr_Cm;

# Imports
use strict;
use Data::Dumper qw(Dumper);
use Module::Common qw(validate_mixed);
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::Transit;
use Options;
use Params::Validate qw(SCALAR SCALARREF BOOLEAN OBJECT ARRAYREF UNDEF);
use Time::Local;
use Time::ParseDate qw(parsedate);
use TestUtils qw(get_site_info);
use Tharn qw($Log);
use base qw(Os::CoreMgr);

# Import exceptions
use NACL::Transit::Exceptions::TransitException qw(:try);
use NACL::Transit::Exceptions::RetriesExceeded qw(:try);

# Constants
use constant CORETOOL_LOC =>
    "/x/eng/localtest/noarch/share/coretool/freebsd8";
use constant MEMORY_DFLT => "4G";

# Globals
my $Core_of = {};

###############################################################################
# Method:    clear core area
# Objective: removed all unsaved cores
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_unsaved_cores {
    my ($self, %args) = @_;
    my ($cmd, $cliObj, $node, $output, $partner);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },

        }
    );

    ### Initialize variables
    $partner = ($args{partner}) ? "partner" : "";
    $cliObj = $self->get_conn_obj(type => "cli");
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    ### Execute command
    $cliObj->system_node_coredump_delete_all(node => $node);

    ### Clean up global core hash
    if ($args{cid}) {
        $self->_log(
            msg => "removing unsaved core "
                . $self->{cores_of}{$node}{$args{cid}}{name},
            indent => 3
        );
        delete $self->{cores_of}{$node}{$args{cid}};
    } else {
        foreach my $cid (keys %{$self->{cores_of}{$node}}) {
            my $core_of = $self->{cores_of}{$node}{$cid};
            if ($core_of->{status} =~ /^unsaved$/) {
                $self->_log(
                    msg    => "removing unsaved core " . $core_of->{name},
                    indent => 3
                );
                delete $self->{cores_of}{$node}{$cid};
            }
        }
    }
    return $self;

} ## end sub delete_unsaved_cores

###############################################################################
# Method:    delete_saved_cores
# Objective: delete all core files
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_saved_cores {
    my ($self, %args) = @_;
    my ($cliObj, $cmd, $node, $output, $partner,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $partner = "partner " if $args{partner};
    $cliObj = $self->get_conn_obj(type => "cli");
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

### BUG:  cdefGet parser not creating the command definition.

=comment

	### Execute command and parse output
	$cliObj->system_node_coredump_delete(
		node=>$node,
		delete=>'*.nz'
	);
=cut

    $cmd = "node run -node " . $node . " \"priv set test;rmcores\"";
    $cliObj->execute_raw_command(command => $cmd);

    foreach my $key (keys %{$self->{cores_of}{$node}}) {
        my $core_of = $self->{cores_of}{$node}{$key};
        if ($core_of->{status} =~ /^saved$/) {
            $self->_log(
                msg    => "removing saved core " . $core_of->{name},
                indent => 3
            );
            delete $self->{cores_of}{$node}{$key};
        }
    }

    return $self;
} ## end sub delete_saved_cores

###############################################################################
# Method:    destroy coremgr_10m object
# Objective: log unique marker in messages file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self) = @_;

    ### Set back to ONTAP defaults
    $self->set_save_onboot(1);
    $self->set_sparse_core(1);
    $self->set_auto_giveback(1);
    $self->set_auto_giveback_panic(1);
    $self->dump_attempts(value => 2);
    $self->mroot_minfree_limit(bytes => 262144000);
    $self->_set_log_level(level => "INFO");

} ## end sub destroy

###############################################################################
# Method:    disk_fail
# Objective: fail a disk
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disk_fail {
    my ($self, %args) = @_;
    my ($cmd, $cliObj, $rc,);

    ### Validate input arguments
    ($self, %args)
        = validate_mixed(@_, [{type => OBJECT}],
        {disk => {type => SCALAR,},});

    ### Initialize variables
#$cmd="node run -node local -command \"priv set test; dumpcore_fail -d $args{disk}";
    $cmd
        = "node run -node local -command \"priv set test; disk fail $args{disk}\"";
    $rc = 1;
    $cliObj = $self->get_conn_obj(type => "cli");
    ### Execute command and parse output
    try {
        $cliObj->execute_raw_command(command => "$cmd");
    }
    catch NATE::Error with {
        $rc = 0;
    };

    return $rc;
} ## end sub disk_fail

###############################################################################
# Method:    dump_attempts
# Objective: set number of attempts if dump fails
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_attempts {
    my ($self, %args) = @_;
    my ($cliObj, $recs, $retry_cnt, $value,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   value => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $retry_cnt = $args{value};

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

#config modify -node csiqa-6070-41 -sparsecore-enabled true -min-free 250MB -coredump-attempts 2 -save-attempts 2 -save-onstartup true
    if ($retry_cnt) {
        $self->_log(
            msg    => "setting dump_attempts to " . $retry_cnt,
            indent => 3
        );

        $cliObj->system_node_coredump_config_modify(
            'node'              => $self->get_node(),
            'coredump-attempts' => $retry_cnt,
        );
    }

    $recs = $cliObj->system_node_coredump_config_show(
        'node' => $self->get_node(),)->get_parsed_output();

    $retry_cnt = $recs->[0]->{'coredump-attempts'};

    return $retry_cnt;

} ## end sub dump_attempts

###############################################################################
# Method:    dumpcore
# Objective: panic node to dump a core
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dumpcore {
    my ($self, %args) = @_;
    my ($cid,    $cmd,     $coreObj, $console_conn,
        $output, $partner, $prompt,  $sp_conn,
        $t1,     $t2,      %dump_of,
    );

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   sp => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            reason => {type => SCALAR, optional => 1, default => "test"},
            wait   => {type => SCALAR, optional => 1, default => 1},
        }
    );

    if ($args{sp}) {

        ## Initialize variables
        #$partner="partner " if $args{partner};
        $cmd = "system core";
        $sp_conn = $self->get_conn_obj(type => "sp");

        ### Execute command to panic system
        (undef, undef, $output, $prompt)
            = &Tharn::say($sp_conn, $cmd, '-prompt',
            "This will cause a dirty shutdown of your appliance. *");

        if ($prompt =~ /shutdown/i) {
            (undef, undef, $output, undef)
                = &Tharn::say($sp_conn, "y", -blind);
        }
    }

    else {
        ### Initialize variables
        $partner = "partner " if $args{partner};
        $console_conn = $self->get_conn_obj(type => "console");

		# let's make sure we get a CLI prompt to execute the command
		$self->{transit}->change_state(to => "CLI");
        #$cmd=$partner."reboot -d $args{reason}";

        #my $cliObj = $self->get_conn_obj(type=>"cli");
        my $node = ($args{partner}) ? $self->{partner} : $self->get_node();

        $cmd
            = "node run -command \"priv set test; panic -f\" " 
            . "-node "
            . $node;

        &Tharn::say($console_conn, "$cmd", -blind);

    }
    ### Set connections to non exist
    $self->{conn_cli_exists}     = 0;
    $self->{conn_bsd_exists}     = 0;
    $self->{conn_console_exists} = 0;
    $self->{cid}                 = undef;

    ### Wait for dump to complete
    $self->_log(msg => "dumping core", indent => 3);
    %dump_of = $self->_wait_for_dump(wait => $args{wait});

    ### Determine core id from timestamp and
    ### save with dump hash record
    $cid = parsedate($dump_of{timestamp});
    $dump_of{cid} = $cid;
    my @disks = split(/\s+/, $dump_of{disks});
    $dump_of{disks} = [@disks];

    ### Store the dump record in the object
    $self->{dumpinfo}{$cid} = \%dump_of;

    my ($cnt, $compressed, $type);
    $cnt = scalar(@{$self->get_dump_info(cid => $cid)->{disks}});
    $compressed = $self->get_dump_info(cid => $cid)->{compressed};

    $type = "none"       if ($cnt == 0 and not defined $compressed);
    $type = "spray"      if ($cnt > 1  and not $compressed);
    $type = "spare"      if ($cnt == 1);
    $type = "compressed" if ($cnt > 1  and $compressed);

    $self->{dumpinfo}{$cid}{type} = $type;

    ### Store the last dump cid in the object
    $self->{last_cid} = $cid;

    return $cid;
} ## end sub dumpcore

###############################################################################
# Method:    find_dump_type
# Objective: get type of core dump will create
# Details:   See POD documentation at the end of this file.
###############################################################################
sub find_dump_type {
    my ($self, %args) = @_;
    my ($cmd, $cliObj, $node, $output, $partner, $rec,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $partner = "partner " if $args{partner};
    $cliObj = $self->get_conn_obj(type => "cli");
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    ### Execute command and parse output
    $rec = $cliObj->system_node_coredump_status(node => $node,)
        ->get_parsed_output()->[0];

    return "spray"      if ($rec->{type} =~ /spray/);
    return "spare"      if ($rec->{type} =~ /spare/);
    return "compressed" if ($rec->{type} =~ /compress|zipp/);
    return "none";
} ## end sub find_dump_type

###############################################################################
# Method:    get_available_space
# Objective: min bytes avaiable in mroot needed to save core
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_available_space {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd, $size,);

    $bsdObj = $self->get_conn_obj(type => "bsd");
    $cmd    = "df -ha | egrep /mroot\$ | awk '{print \$4}'";
    $size   = $bsdObj->execute_raw_command(command => "$cmd");

    return $self->_size_to_bytes($size);

} ## end sub get_available_space
###############################################################################
# Method:    get_conn_obj
# Objective: create a connection to node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_conn_obj {
    my ($self, %args) = @_;
    my ($connObj, $errmsg, $hostObj, $retry_cnt, $retry_limit, $set_mode,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   type => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /cli|bsd|console|sp/i;
                        }
                }
            },
            nologin => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    Tharn::host($self->get_node());
    $set_mode = ($self->_get_hosttype() =~ /filer-ng/i) ? "CMode" : "7Mode";
    $retry_cnt = 0;
    $retry_limit = 5;

    if ($args{type} =~ /cli/i) {

        while (1) {
            eval {
                if ($self->{apiObjCli}) {
                    if (not $self->{apiObjCli}->api_is_session_alive()) {
                        $self->_log(
                            msg =>
                                "reconnecting APISet object ($set_mode-SSH) for "
                                . $self->get_node(),
                            indent => 5
                        );
                        $self->{apiObjCli}->api_reconnect_session();
                        $self->{apiObjCli}
                            ->set_privilege('privilege-level' => 'diag');
                    }
                } else {
                    $self->_log(
                        msg => "creating APISet object ($set_mode-SSH) for "
                            . $self->get_node(),
                        indent => 5
                    );

                    $self->{apiObjCli} = NACL::APISet->new(
                        hostobj   => $self->{hostObj},
                        category  => "Node",
                        interface => "CLI",
                        set       => $set_mode,
                    );
                    $self->{apiObjCli}->set_timeout(
                        'connectrec-timeout' => $self->{timeout});
                    $self->{apiObjCli}
                        ->set_privilege('privilege-level' => 'diag');

                }
                $self->{conn_cli_exists} = 1;
                last;
            };
            if ($@) {
                my $exception_object = shift;
                if ($retry_cnt++ < $retry_limit) {
                    $self->_log(
                        msg    => "unable to connect, retrying again",
                        indent => 5
                    );
                    redo;
                }
                die $exception_object;
            }
        }

        $connObj = $self->{apiObjCli};

    } elsif ($args{type} =~ /bsd/i) {

        while (1) {
            eval {
                if ($self->{apiObjBsd}) {
                    if (   not $self->{conn_bsd_exists}
                        or not $self->{apiObjBsd}->api_is_session_alive()) {
                        $self->_log(
                            msg =>
                                "connecting APISet object (Systemshell-SSH) for "
                                . $self->get_node(),
                            indent => 5
                        );
                        $self->{apiObjBsd}->api_reconnect_session();
                    }
                } else {
                    $self->_log(
                        msg => "creating APISet object (Systemshell-SSH) for "
                            . $self->get_node(),
                        indent => 5
                    );
                    $self->{apiObjBsd} = NACL::APISet->new(
                        hostobj    => $self->{hostObj},
                        category   => "Node",
                        interface  => "CLI",
                        set        => "Systemshell",
                        shell_type => "CMode",
                    );
                    $self->{apiObjBsd}->set_timeout(
                        'connectrec-timeout' => $self->{timeout});
                }
                ### set connection to true
                $self->{conn_bsd_exists} = 1;
                last;
            };
            if ($@) {
                my $exception_object = shift;
                if ($retry_cnt++ < $retry_limit) {
                    $self->_log(
                        msg    => "unable to connect, retrying again",
                        indent => 5
                    );
                    redo;
                }
                die $exception_object;
            }
        }

        $connObj = $self->{apiObjBsd};

    } elsif ($args{type} =~ /sp/) {

        #if (not $self->{conn_sp_exists}) {
        my $node = $self->get_node();
	my $conn_type = '';
	
	if (exists $self->{hostObj}->{conn}->{sp_login}) {
	    $conn_type = "sp_login";
	} elsif (exists $self->{hostObj}->{conn}->{rlm}) {
	    $conn_type = "rlm";	    
	} else {
	    die "Tharnhost file does not have sp_login or rlm entry.";
	}
	
        $self->_log(
            msg    => "creating a $conn_type connection for " . $node,
            indent => 5
        );	

        $self->{sp_conn} = Tharn::connect(
            $self->{$node} . "#"."$conn_type",
            "-login",
            -eval => sub {
                if (@_) {
                    $errmsg = (@_)[0]->{messages};
                    die $errmsg;
                }
            }
        );	
	
        $self->{conn_sp_exists} = 1;

        $connObj = $self->{sp_conn};

    } elsif ($args{type} =~ /console/) {
        if (not $self->{conn_console_exists}) {
            $self->_log(
                msg => "creating a console connection for "
                    . $self->get_node(),
                indent => 5
            );
            $self->{console_conn} = Tharn::connect(
                $self->{$self->{node}} . "#console",
                ($args{nologin}) ? "-nologin" : "-login",
                -eval => sub {
                    if (@_) {
                        $errmsg = (@_)[0]->{messages};
                        die $errmsg;
                    }
                }
            );

            $self->{conn_console_exists} = 1;
        }

        $connObj = $self->{console_conn};

    } else {
        die "connection type '" . $args{conn} . "' unknown";
    }

    return $connObj;

} ## end sub get_conn_obj

###############################################################################
# Method:    get_conn_obj
# Objective: create a connection to node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_conn_obj_OLD {
    my ($self, %args) = @_;
    my ($connObj, $errmsg, $hostObj, $retry_cnt, $set_mode,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   type => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /cli|bsd|console/i;
                        }
                }
            },
            nologin => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    Tharn::host($self->get_node());
    $set_mode = ($self->_get_hosttype() =~ /filer-ng/i) ? "CMode" : "7Mode";
    $retry_cnt = 0;

    if ($args{type} =~ /cli/i) {
        if (not $self->{conn_cli_exists}) {
            $self->_log(
                msg => "creating APISet object ($set_mode-SSH) for "
                    . $self->get_node(),
                indent => 5
            );

            while (1) {
                try {
                    $self->{apiObjCli} = NACL::APISet->new(
                        hostobj   => $self->{hostObj},
                        category  => "Node",
                        interface => "CLI",
                        set       => $set_mode,
                    );
                }
                catch NACL::APISet::Exceptions::ConnectionFailedException
                    with {
                    my $exception_object = shift;
                    if ($retry_cnt++ < 1) {
                        $self->_log(
                            msg    => "unable to connect, retrying again",
                            indent => 5
                        );
                        redo;
                    }
                    $exception_object->throw();
                    };
                last;
            }

            ### set connection to true
            $self->{conn_cli_exists} = 1;
        }

        $connObj = $self->{apiObjCli};

    } elsif ($args{type} =~ /bsd/i) {
        if (not $self->{conn_bsd_exists}) {
            $self->_log(
                msg => "creating APISet object (Systemshell-SSH) for "
                    . $self->get_node(),
                indent => 5
            );
            while (1) {
                try {
                    $self->{apiObjBsd} = NACL::APISet->new(
                        hostobj    => $self->{hostObj},
                        category   => "Node",
                        interface  => "CLI",
                        set        => "Systemshell",
                        shell_type => "CMode",
                    );
                }
                catch NACL::APISet::Exceptions::ConnectionFailedException
                    with {
                    my $exception_object = shift;
                    if ($retry_cnt++ < 1) {
                        $self->_log(
                            msg    => "unable to connect, retrying again",
                            indent => 5
                        );
                        redo;
                    }
                    $exception_object->throw();
                    };
                last;
            }
            ### set connection to true
            $self->{conn_bsd_exists} = 1;
        }

        $connObj = $self->{apiObjBsd};

    } elsif ($args{type} =~ /console/) {
        if (not $self->{conn_console_exists}) {
            $self->_log(
                msg => "creating a console connection for "
                    . $self->get_node(),
                indent => 5
            );
            $self->{console_conn} = Tharn::connect(
                $self->{$self->{node}} . "#console",
                ($args{nologin}) ? "-nologin" : "-login",
                -eval => sub {
                    if (@_) {
                        $errmsg = (@_)[0]->{messages};
                        die $errmsg;
                    }
                }
            );

            $self->{conn_console_exists} = 1;
        }

        $connObj = $self->{console_conn};

    } else {
        die "connection type '" . $args{conn} . "' unknown";
    }

    return $connObj;

} ## end sub get_conn_obj_OLD

###############################################################################
# Method:    get_core_cnt
# Objective: retrieve core count
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_core_cnt {
    my ($self, %args) = @_;
    my ($cnt, $node, $status, @coreObjs,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            status => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /^saved$|^unsaved$/i;
                        }
                }
            },
        }
    );

    ### Initialize variables;
    $cnt    = 0;
    $status = $args{status};
    $node   = ($args{partner}) ? $self->{partner} : $self->get_node();

    if ($status) {
        foreach my $key (keys %{$self->{cores_of}{$node}}) {
            if ($self->{cores_of}{$node}{$key}{status} =~ /^$status$/) {
                $cnt++;
            }
        }
    } else {
        $cnt = scalar(keys %{$self->{cores_of}{$node}});
    }

    return $cnt;

} ## end sub get_core_cnt

###############################################################################
# Method:    get_core_ids
# Objective: retrieve list of core ids
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_core_ids {
    my ($self, %args) = @_;
    my ($node, @cids,);
    $node = $self->get_node();
    push @cids, keys %{$self->{cores_of}{$node}};

    return (wantarray()) ? @cids : [@cids];

} ## end sub get_core_ids

###############################################################################
# Method:    get_core_info
# Objective: retrieve core metadata in hsh
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_core_info {
    my ($self, %args) = @_;
    my ($node, @cids,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {type => SCALAR},

            #name      => {type => SCALAR, optional=>1},
            #status    => {type => SCALAR, optional=>1},
        }
    );

    $node = $self->get_node();
    @cids = (keys %{$self->{cores_of}{$node}});

    unless (grep $args{cid}, @cids) {
        die "Unable to find core record with coreid '" . $args{cid} . "'";
    }

    return $self->{cores_of}{$node}{$args{cid}};

} ## end sub get_core_info

###############################################################################
# Method:    get_dump_info
# Objective: return whether dump was compressed or not
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_dump_info {
    my ($self, %args) = @_;

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {
                type      => SCALAR,
                optional  => 1,
                default   => $self->{last_cid},
                callbacks => {
                    'legal values' => sub { shift() >= 1 }
                }
            }
        },
    );

    return $self->{dumpinfo}{$args{cid}};
} ## end sub get_dump_info

###############################################################################
# Method:    get_last_cid
# Objective: retrieve core id of last core dumped
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_last_cid {
    my ($self) = @_;

    return $self->{last_cid};
}

###############################################################################
# Method:    get_save_info
# Objective: return savecore info in hsh structure
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_save_info {
    my ($self, %args) = @_;
    my ($cliObj, $console_conn, $cmd, $hsh_of, $node, $partner, $rec,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    ### Initialize parameters
    $cliObj = $self->get_conn_obj(type => "cli");

    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    ### Execute command
    #$self->_log(msg=>"retrieving save core info");
    #$self->_log(msg=>"TIMEOUT: ".$cliObj->get_timeout());
    $rec = $cliObj->system_node_coredump_status(node => $node,)
        ->get_parsed_output()->[0];

    $hsh_of = {
        blocks        => $rec->{'total-blocks'},
        disk_free_cnt => (defined $rec->{numdisks}) ? $rec->{numdisks} : 0,
        disk_used_cnt => (defined $rec->{'disks-used'}) ? $rec->{'disks-used'}
        : 0,
        dump_type   => $rec->{type},
        unsaved_cnt => (defined $rec->{'unsaved-cores'})
        ? $rec->{'unsaved-cores'}
        : 0
    };

    return $hsh_of;
} ## end sub get_save_info

###############################################################################
# Method:    get_local_disks
# Objective: return array of local disks
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_local_disks {
    my ($self) = @_;
    my ($cliObj, @disks, $output,);

    $cliObj = $self->get_conn_obj(type => "cli");

    $output = $cliObj->storage_disk_show(owner => $self->get_node());

    foreach my $disks_of (@{$output->{parsed_output}}) {
        my ($disk) = $disks_of->{disk} =~ /:(\S+)/;
        push @disks, $disk;
    }

    return (wantarray()) ? sort @disks : [sort @disks];
} ## end sub get_local_disks

###############################################################################
# Method:    giveback
# Objective: initiates storage giveback to taken over node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub giveback {
    my ($self, %args) = @_;
    my ($cmd, $cliObj, $force, $resObj, $waiting, $warning,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   force => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            waiting => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    #storage failover giveback -fromnode csiqa-6070-41 -override-vetoes

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");
    $force   = ($args{force})   ? "true"  : "false";
    $waiting = ($args{waiting}) ? "false" : "true";

	# a list of known giveback errors to check for before
	# reporting a test failure
	my @known_giveback_errors = (
		'Partner has not fully booted for giveback of data aggregates',
		'A giveback is already in progress',
	);

	try {
		$resObj = $cliObj->storage_failover_giveback(
			fromnode                  => $self->get_node(),
			'override-vetoes'         => $force,
			'require-partner-waiting' => $waiting,
		);

		#&Tharn::say($console_conn, "$cmd", '-blind');
		$self->_log(msg => "command: " . $resObj->get_command(), indent => 3);
	} catch NACL::APISet::Exceptions::ResponseException with {
		my $exception = shift;

		# check if [auto-]giveback was already in progress
		foreach my $known_error (@known_giveback_errors) {
			if (!grep(/$known_error/, $exception->text())) {
				# if no, then rethrow the exception
				$exception->throw();
			}
		}
		# else, ignore the exception and move on
	};

    return 1;
} ## end sub giveback

###############################################################################
# Method:    info_dump
# Objective: return info about dump
# Details:   See POD documentation at the end of this file
################################################################################
sub info_dump {
    my ($self, %args) = @_;
    my ($cid, @bufs, %hsh_of,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
        }
    );

    $hsh_of{cid} = $args{cid} if (defined $args{cid});

    push @bufs,
        "    core id:    " . $args{cid},
        "    compressed: " . $self->get_dump_info(%hsh_of)->{compressed},
        "    diskerrs:   " . $self->get_dump_info(%hsh_of)->{diskerrs},
        "    duration:   " . $self->get_dump_info(%hsh_of)->{duration},
        "    message:    " . $self->get_dump_info(%hsh_of)->{message},
        "    recursive:  " . $self->get_dump_info(%hsh_of)->{recursive},
        "    starts:     " . $self->get_dump_info(%hsh_of)->{starts},
        "    status:     " . $self->get_dump_info(%hsh_of)->{status},
        "    timestamp:  " . $self->get_dump_info(%hsh_of)->{timestamp},
        "    disks:      @{$self->get_dump_info(%hsh_of)->{disks}}",
        "";

    return (wantarray()) ? @bufs : [@bufs];
} ## end sub info_dump

###############################################################################
# Method:    info_cores
# Objective: display formatted metadata
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info_cores {
    my ($self, %args) = @_;
    my ($first, $node, @bufs, %cores_of,);

    ### Parse input
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },

            detail => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $first = 1;

    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    #$self->_log(msg=>"NODE: ".$node);
    %cores_of
        = (exists $self->{cores_of}{$node})
        ? %{$self->{cores_of}{$node}}
        : {};

    foreach my $key (keys %cores_of) {

        next if (defined $args{cid} and $key !~ /^$args{cid}$/i);

        if ($args{detail}) {
            push @bufs,

                #  "Ref:               ".$self,
                "    Name:              " . $cores_of{$key}->{name},
                "    Status:            " . $cores_of{$key}->{status},
                "    Coreid:            " . $cores_of{$key}->{core_id},
                "    Timestamp:         " . $cores_of{$key}->{timestamp},
                "    Os Version:        " . $cores_of{$key}->{version},
                "    Panic:             " . $cores_of{$key}->{panic_string},
                "    Size:              " . $cores_of{$key}->{size};
        } else {

            if ($first) {
                push @bufs,
                    sprintf(
                    "    %-39.39s  %-8.8s %-10.10s %-6.6s %-28.28s\n"
                        . "    %-39.39s  %-8.8s %-10.10s %-6.6s %-28.28s",
                    "Name",      "Status", "Coreid", "Size",
                    "Timestamp", "-" x 39, "-" x 8,  "-" x 10,
                    "-" x 6,     "-" x 28
                    );
                $first = 0;
            }

            push @bufs,
                sprintf(
                "    %-39.39s  %-8.8s %10.10s %5.6s  %-s",
                $cores_of{$key}->{name},    $cores_of{$key}->{status},
                $cores_of{$key}->{core_id}, $cores_of{$key}->{size},
                $cores_of{$key}->{timestamp},
                );
        }
    }
    return (wantarray()) ? @bufs : [@bufs];
} ## end sub info_cores

###############################################################################
# Method:    is_auto_giveback
# Objective: detemine whether 'save on boot' is set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_auto_giveback {
    my ($self) = @_;
    return $self->{auto_giveback};
}

###############################################################################
# Method:    is_auto_giveback_panic
# Objective: detemine whether 'autogiveback after panic' is set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_auto_giveback_panic {
    my ($self) = @_;
    return $self->{auto_giveback_panic};
}

###############################################################################
# Method:    is_core_file
# Objective: verify if core file exists
###############################################################################
sub is_core_file {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd, $exists, $name, $node, $output,);

    ### Parse input
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    #$self->_log(msg=>Dumper($self->{cores_of}));
    $name = $self->{cores_of}{$node}{$args{cid}}{name};
    if ($args{partner}) {
        $bsdObj->execute_raw_command(command => "sudo mount_partner");
    }

    $cmd
        = ($args{partner})
        ? "ls -l1 /partner/etc/crash/" . $name
        : "ls -l1 /mroot/etc/crash/" . $name;

    $exists = 1;

    ### Execute command
    $output = $bsdObj->execute_raw_command(command => $cmd);
    $exists = 0 if ($output =~ /no such file/i);

    return $exists;
} ## end sub is_core_file

###############################################################################
# Method:    is_failover
# Objective: Determine failover state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_failover {
    my ($self, %args) = @_;
    my ($cliObj, $rc, $recs,);

    ### Initialize variables

    $cliObj = $self->get_conn_obj(type => "cli");

    $recs = $cliObj->storage_failover_show(node => $self->get_node(),)
        ->get_parsed_output();

    if ($recs->[0]->{enabled} =~ /true/i) {
        $rc = 1;
    } else {
        $rc = 0;
    }

    return $rc;
} ## end sub is_failover

###############################################################################
# Method:    is_save_on_boot
# Objective: detemine whether 'save on boot' is set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_save_onboot {
    my ($self) = @_;
    return $self->{save_onboot};
}

###############################################################################
# Method:    is_sparse_core
# Objective: determine whether sparse core is set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_sparse_core {
    my ($self) = @_;
    return $self->{sparse_core};
}

###############################################################################
# Method:    log_marker
# Objective: log unique marker in messages file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub log_marker {
    my ($self, %args) = @_;
    my ($cmd, $cliObj, $marker, $rc,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   name => {
                type     => SCALAR,
                optional => 1,
                default  => "NATE coredump test"
            },
        }
    );

    ### Initialize variables
    $marker = $args{name} . " - " . localtime();
    $cmd    = "echo \"$marker\" >> /mroot/etc/log/messages";
    $cliObj = $self->get_conn_obj(type => "bsd");
    ## $self->{cookie}=$cookie;

    ### Execute command and parse output
    $cliObj->execute_raw_command(command => "$cmd");

    $rc = $self->_verify_marker(marker => $marker);

    return ($rc == 0) ? $rc : $marker;
} ## end sub log_marker

###############################################################################
# Method:    log_marker
# Objective: log unique marker in messages file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub log_marker_OLD {
    my ($self, %args) = @_;
    my ($cmd, $cliObj, $marker, $rc,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   marker => {
                type     => SCALAR,
                optional => 1,
                default  => "NATE coredump test"
            },
        }
    );

    ### Initialize variables
    $marker = $args{marker} . " - " . localtime();
    $cmd    = "node run local \"priv set test; log \"$marker\"\"";
    $cliObj = $self->get_conn_obj(type => "cli");
    ## $self->{cookie}=$cookie;

    ### Execute command and parse output
    $cliObj->execute_raw_command(command => "$cmd");

    $rc = $self->_verify_marker(marker => $marker);

    ### Make sure cookie gets logged
    ### $self->read_messages(marker=>$cookie);

    return ($rc == 0) ? $rc : $marker;
} ## end sub log_marker_OLD

###############################################################################
# Method:    mroot_minfree_limit
# Objective: min bytes avaiable in mroot needed to save core
# Details:   See POD documentation at the end of this file.
###############################################################################
sub mroot_minfree_limit {
    my ($self, %args) = @_;
    my ($cliObj, $bytes, $cmd,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   bytes => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $bytes = $args{bytes};
    $cliObj = $self->get_conn_obj(type => "cli");

    if (defined $args{bytes}) {
        if ($args{bytes} == 0) {
            $bytes = 262144000;    ### 250M*1024**2
        } else {
            $bytes = $args{bytes};
        }

        $self->_log(
            msg => "setting mroot minfree limit to "
                . int($bytes / 1024**2) . "M",
            indent => 3
        );
        my $resObj = $cliObj->system_node_coredump_config_modify(
            node       => $self->get_node(),
            'min-free' => $bytes,
        )->get_parsed_output()->[0];

    } else {
        my $resObj = $cliObj->system_node_coredump_config_show(
            node => $self->get_node(),)->get_parsed_output()->[0];

        $bytes = $resObj->{'min-free'};
    }

    return $bytes;
} ## end sub mroot_minfree_limit

###############################################################################
# Method:    read_messages
# Objective: read /mroot/etc/log/messages file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub read_messages {
    my ($self, %args) = @_;
    my ($bsdObj, $check_marker, $cmd, $output, $retry_cnt, @lines,);

    ### Validate input arguments
    ($self, %args)
        = validate_mixed(@_, [{type => OBJECT}],
        {marker => {type => SCALAR},});

    ### Verify that messages log is up to date by
    ### logging a marker and waiting for it to be
    ### written.
    $check_marker = $self->log_marker(name => "sync check");
    unless ($check_marker) {
        die "/mroot/etc/log/messages is not current";
    }

    ### Retrieve messages since 'log_marker'
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $cmd    = "grep -A 9999 '" . $args{marker} . "' /mroot/etc/log/messages";
    $output = $bsdObj->execute_raw_command(command => "$cmd");

    @lines = split(/\r*\n+/, $output);
    map(s/^\s+//, @lines);
    map(s/\s+$//, @lines);
    map(s/\cM//g, @lines);
    @lines = grep (/\S+/, @lines);

    $self->{messages} = [@lines];

    return (wantarray()) ? @{$self->{messages}} : $self->{messages};
} ## end sub read_messages

###############################################################################
# Method:    refresh
# Objective: retrieve core file size
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;
    my ($cid, $cores_of, $found, $partner, @coreObjs,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
        }
    );

    ### Retrieve the hash list of all core types
    $self->_log(msg => "refreshing core data", indent => 3);

    $cores_of = $self->_load_cores(partner => $args{partner});

    return $self;
} ## end sub refresh

################################################################################
# Method: restrict_memory
# Objective: set memory restriction
################################################################################
sub restrict_memory {
    my ($self, %args) = @_;
    my ($current_restrict_size, $memory_restrict_size,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {size => {type => SCALAR, optional => 1, default => MEMORY_DFLT},}
    );

    $memory_restrict_size = $args{size};

    ### Update restrict memsize if MEMORY param passed into
    ### test differs from the restrict memsize bootarg. Otherwise,
    ### if no MEMORY param passed in and a restrict memsize bootarg
    ### exists then remove the restrict memsize bootarg.

    $current_restrict_size = $self->_get_restricted_memsize();

    if ($memory_restrict_size) {
        if ($current_restrict_size !~ /^$memory_restrict_size$/i) {
            $self->_log(
                msg => "updating bootarg to restrict memsize to "
                    . $memory_restrict_size,
                indent => 3
            );
            $self->_set_restricted_memsize(memory => $memory_restrict_size);
        } else {
            $self->_log(
                msg => "memory already restricted to "
                    . $memory_restrict_size,
                indent => 3
            );
        }
    }

    1;
} ## end sub restrict_memory

###############################################################################
# Method:    savecore
# Objective: savecore
# Details:   See POD documentation at the end of this file.
###############################################################################
sub savecore {
    my ($self, %args) = @_;
    my ($cid,    $cliObj,  $cmd,     $conn, $ha,  $node,
        $output, $partner, @options, $rc,   $rec, $resObj,
    );

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   abort => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
            cid => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/;
                        }
                }
            },
            force => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
            giveback => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
            giveback_force => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
            synchronous => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
            wait => {
                type      => SCALAR,
                optional  => 1,
                default   => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $rc      = 0;
    $cliObj  = $self->get_conn_obj(type => "cli");
    $partner = $self->{partner};

    ### Retrieve failover record for node
    $rec = $cliObj->storage_failover_show(node => $self->get_node())
        ->get_parsed_output()->[0];

    ### If node is in takeover state then only the
    ### local node can do the save of unsaved cores
    ### on the downed partner
    if ($rec->{state} =~ /takeover/) {
        $node = $self->get_node();
        $args{partner} = 0;
    } else {
        $node = ($args{partner}) ? $self->{partner} : $self->get_node();
    }

    ### Execute the command
    if ($args{force} or $args{cid}) {
        $resObj = $cliObj->system_node_coredump_save(
            node     => $node,
            corename => $self->{cores_of}{$node}{$args{cid}}{name}
        );
    } else {
        $resObj = $cliObj->system_node_coredump_save_all(node => $node);
    }

    ### ret
    $self->_log(msg => "command: " . $resObj->get_command(), indent => 3);

=commment
    if ($args{invalidate}) {
		if ($args{cid}){
			delete $self->{cores_of}{$node}{$args{cid}};
		} else {
			$self->delete_unsaved_cores(partner=>$args{partner});
		}
	}
=cut

    ### Wait for save to complete
    $self->_log(msg => "waiting for save asynchronously", indent => 3);
    $rc = $self->wait_for_savecore(
        wait           => $args{wait},
        abort          => $args{abort},
        partner        => $args{partner},
        giveback_force => $args{giveback_force},
        giveback       => $args{giveback},
    );

    $self->refresh(partner => $args{partner});    #if ($cmd=~/^savecore$/);

    return $rc;
} ## end sub savecore

###############################################################################
# Method:    set_auto_giveback
# Objective: turn auto giveback on or off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_auto_giveback {
    my ($self, $value) = @_;
    my ($cliObj, $state,);

    unless ($value =~ /1|0|on|off|true|false/) {
        die "Invalid value '$value' for setting auto-giveback state";
    }

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

    $state = ($value =~ /1|t|on/) ? "true" : "false";

#config modify -node csiqa-6070-41 -sparsecore-enabled true -min-free 250MB -coredump-attempts 2 -save-attempts 2 -save-onstartup true

    $self->_log(
        msg    => "setting auto-giveback " . $state,
        indent => 3
    );

    $cliObj->storage_failover_modify(
        'node'          => $self->get_node(),
        'auto-giveback' => $state,
    );

    $self->{auto_giveback} = $value;

    return 1;
} ## end sub set_auto_giveback

###############################################################################
# Method:    set_auto_giveback_panic
# Objective: turn auto giveback after panic on or off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_auto_giveback_panic {
    my ($self, $value) = @_;
    my ($cliObj, $state,);

    unless ($value =~ /1|0|on|off|true|false/) {
        die
            "Invalid value '$value' for setting auto-giveback after panic state";
    }

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

    $state = ($value =~ /1|t|on/) ? "true" : "false";

    $self->_log(
        msg    => "setting auto-giveback after panic " . $state,
        indent => 3
    );

    my $prompts_answers
        = ['.*Do you want to disable auto-giveback-after-panic*' => 'y'];

    $cliObj->storage_failover_modify(
        'node'                      => $self->get_node(),
        'auto-giveback-after-panic' => $state,
        'connectrec-match_table'    => $prompts_answers,
    );

    $self->{auto_giveback_panic} = $value;

    return 1;
} ## end sub set_auto_giveback_panic

###############################################################################
# Method:    set_failover
# Objective: define whether failover is on or off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_failover {
    my ($self, $value) = @_;
    my ($cliObj, $state,);

    unless ($value =~ /1|0|on|off|true|false/) {
        die "Invalid value '$value' for setting set-failover state";
    }

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

    $state = ($value =~ /1|on|t/i) ? "true" : "false";

    $self->_log(
        msg    => "setting failover " . $state,
        indent => 3
    );

    $cliObj->cluster_ha_modify('configured' => 'false',);

    $cliObj->storage_failover_modify(
        'node'    => $self->get_node(),
        'enabled' => $state,
        'onpanic' => $state,
    );

    return 1;
} ## end sub set_failover

###############################################################################
# Method:    set_save_onboot
# Objective: define whether core is automatically saved on bootup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_save_onboot {
    my ($self, $value) = @_;
    my ($cliObj, $cmd, $state);

    unless ($value =~ /1|0|on|off|true|false/) {
        die "Invalid value '$value' for setting save-onboot state";
    }

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

    $state = ($value =~ /1|t|on/) ? "true" : "false";

    $self->_log(
        msg    => "setting save-onboot " . $state,
        indent => 3
    );

    $cliObj->system_node_coredump_config_modify(
        'node'           => $self->get_node(),
        'save-onstartup' => $state,
    );

    ### BURT 360745
    ### In HA 10m, setting autosave to false results in system
    ### running whatever the savecore setting is in /mroot/etc/rc.
    ### Use fiji to enforce autosave state (avoid /mroot/etc/rc).
    $cmd = "node run -node local ";
    $cmd .=
        ($value =~ /1|t|on/)
        ? "\"fiji clear all\""
        : "\"fiji filter '(core.inhibit.takeover.savecore (nil) (return 1))'\"";

    $cliObj->execute_raw_command(command => $cmd);

    $self->{save_onboot} = ($value =~ /1|t|on/) ? 1 : 0;

    return 1;
} ## end sub set_save_onboot

###############################################################################
# Method:    set_sparse_core
# Objective: sparse_cre
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_sparse_core {
    my ($self, $value) = @_;
    my ($cliObj, $state,);

    unless ($value =~ /1|0|on|off|true|false/) {
        die "Invalid value '$value' for setting sparse-core state";
    }

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

    $state = ($value =~ /1|t|on/) ? "true" : "false";

    $self->_log(
        msg    => "setting sparse core " . $state,
        indent => 3
    );

    $cliObj->system_node_coredump_config_modify(
        'node'               => $self->get_node(),
        'sparsecore-enabled' => $state,
    );

    $self->{sparse_core} = ($value =~ /1|t|on/) ? 1 : 0;

    return 1;
} ## end sub set_sparse_core
###############################################################################
# Method:    tp_clear
# Objective: clear test points
# Details:   See POD documentation at the end of this file.
###############################################################################
sub tp_clear {
    my ($self) = @_;
    my ($cliObj, $cmd, $output,);

    $cliObj = $self->get_conn_obj(type => "cli");
    $cmd    = "node run -node local -command \"priv set test; tp clear\"";
    $output = $cliObj->execute_raw_command(command => "$cmd");

    return $self;
} ## end sub tp_clear

###############################################################################
# Method:    tp_error
# Objective: set a test point
# Details:   See POD documentation at the end of this file.
###############################################################################
sub tp_error {
    my ($self, %args) = @_;
    my ($cliObj, $cmd, $output,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   module => {type => SCALAR, optional => 1, default => "dumpcore"},
            id     => {type => SCALAR},
            cnt    => {type => SCALAR},
            parm1  => {type => SCALAR},
            parm2 => {type => SCALAR, optional => 1},
        }
    );

    $cliObj = $self->get_conn_obj(type => "cli");
    $cmd = "node run -node local -command ";
    $cmd
        .= "\"priv set test; tp error "
        . $args{module} . " "
        . $args{id} . " "
        . $args{cnt} . " "
        . $args{parm1} . " "
        . $args{parm2} . "\"";

    $output = $cliObj->execute_raw_command(command => "$cmd");
    return $self;
} ## end sub tp_error

###############################################################################
# Method:    tp_list
# Objective: list test points that are currently set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub tp_list {
    my ($self) = @_;
    my ($cliObj, $cmd, $output, @lines,);
    $cliObj = $self->get_conn_obj(type => "cli");
    $cmd    = "node run -node local -command \"priv set test; tp list\"";
    $output = $cliObj->execute_raw_command(command => "$cmd");

    @lines = split(/\r*\n+/, $output);
    map(s/^\s+//, @lines);
    map(s/\s+$//, @lines);
    map(s/\cM//g, @lines);
    @lines = grep (/^\[|No testpoints/i, @lines);

    return (wantarray()) ? @lines : [@lines];
} ## end sub tp_list

################################################################################
# Method: _unrestrict_memory
# Objective: remove memory restriction
################################################################################
sub unrestrict_memory {
    my ($self) = @_;

    if ($self->_has_restricted_memory()) {
        $self->_set_restricted_memsize(memory => -1);
    } else {
        $self->_log(msg => "no restrictions to remove", indent => 3);
    }
} ## end sub unrestrict_memory

###############################################################################
# Method:    wait_for_boot
# Objective: boot node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wait_for_boot {
    my ($self, %args) = @_;
    my ($cmd, $t1, $t2, $duration,);

    ### Boot up system and measure duration
    $self->_log(msg => "booting ontap", indent => 3);
    $t1 = timelocal(localtime());
    $self->{transit}->change_state(to => "CLI");
    $t2 = timelocal(localtime());

    ### Store boot duration in dump hash
    $duration = ($t2 - $t1);
    $self->_log(msg => "boot duration: " . $duration . "s", indent => 5);

    ### Re-establish connections
    $self->_log(
        msg    => "re-establishing connections to " . $self->get_node(),
        indent => 3
    );
    $self->get_conn_obj(type => "bsd");
    $self->get_conn_obj(type => "cli");

    $self->refresh();
    return $self;
} ## end sub wait_for_boot_BURT

###############################################################################
# Method:    wait_for_fmdisk
# Objective: wait for auto savecore operation to start
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wait_for_fmdisk {
    my ($self, %args) = @_;
    my ($console_conn, $done_filter, $output, $preoutput, $timeout, $rc,
        @lines, @values,);

    ###Initialize variables
    $rc      = 1;
    $timeout = 600;

    #$console_conn = $self->get_conn_obj(type => "console");

    push @values, 'fmdisk_reserve_spare', 'coredump.save.started';

    $done_filter = join('|', @values);

    $self->_log(msg => "Use console_conn to wait for msgs $done_filter");    
    while (1) {
        my (undef, $preoutput, $output) = &Tharn::say(
            $self->{console_conn}, "",
            ".*\r",
            -noprompt,
            -nocr,
            -swallow,
            -timeout => $timeout,
            -maxidle => $timeout,
            -eval_error
        );

        if ($@ && $@ =~ /Err=TIMEOUT/i) {
            $self->_log(
                msg => "Timed out waiting for fmdisk (timeout=$timeout)");
            $rc = 0;
            last;
        }

#since some ems messages are not preface by a carriage return need to expand the match.
        $output = $preoutput . $output;

		if ($output =~ /fmdisk_reserve_spare/ or
			$output =~ /coredump\.save\.started/) {

            $self->{fmdisk} = [@values];
            $rc = 1;
            last;
        }

    }

    return $rc;
} ## end sub wait_for_fmdisk

###############################################################################
# Method:    wait_for_giveback
# Objective: wait until giveback of other node storage is complete
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wait_for_giveback {
    my ($self, %args) = @_;

    $args{msg} = "SF_UP";

    eval { $self->_wait(%args); };
    if ($@) {
        $self->_log(msg => $@, indent => 3);
        return 0;
    }

    return 1;
} ## end sub wait_for_giveback
###############################################################################
# Method:    _wait_for_savecore
# Objective: wait for core to be saved
###############################################################################
sub wait_for_savecore {
    my ($self, %args) = @_;
    my ($cmd,          $cnt,       $complete,    $core,     $core_cnt,
        $console_conn, $core_msg,  $done_filter, $fail_cnt, $first,
        $gbf_set,      $last_core, $last_pct,    $loop_cnt, $node,
        $output,       $pct,       $pct_last,    $rc,       $retry_cnt,
        $save_cnt,     $tp_set,    $t1,          $t2,       $value,
        $wait,         @lines,     @values,
    );

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   abort => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            giveback => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            giveback_force => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            retries => {
                type      => SCALAR,
                optional  => 1,
                default   => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            sleep => {
                type      => SCALAR,
                optional  => 1,
                default   => 5,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            timeout => {
                type      => SCALAR,
                optional  => 1,
                default   => 30,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            timeout_retries => {
                type      => SCALAR,
                optional  => 1,
                default   => 10,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            wait => {
                type      => SCALAR,
                optional  => 1,
                default   => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },

            _async => {
                type      => SCALAR,
                optional  => 1,
                default   => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    #$partner=($args{partner}) ? "partner" : "";

    ### BUG: Missing NACL cdef.  Use 7m cli equivalent for now
    $cmd
        = "node run -node "
        . $self->get_node() . " "
        . "-command \"priv set test; savecore -s\"";

    $complete  = 0;
    $fail_cnt  = 0;
    $first     = 1;
    $loop_cnt  = 0;
    $pct       = 0;
    $pct_last  = 0;
    $rc        = 0;
    $save_cnt  = 0;
    $tp_set    = 0;
    $retry_cnt = 0;
    $wait      = $args{wait};

    ### Retrieve console handle
    $console_conn = $self->get_conn_obj(type => "console");

    ### Start duration timer since there's a core
    ### to be saved
    $t1 = timelocal(localtime());

    ### Messages that will end waiting for savecore
    ### operation
    push @values,
        "No core needs to be saved",
        "No savecore in progress",
        "Savecore completed",
        "Previous attempts to save core failed";
    $done_filter = join('|', @values);

    ### Sit in while loop until one of the savecore completion
    ### messages is seen.
    while (1) {

        ### Determine what say call to use depending on
        ### whether an async or sync operation is specified
        if ($args{_async}) {
            (undef, undef, $output) = &Tharn::say(
                $console_conn,
                $cmd,
                '-swallow',
                '-timeout' => $args{timeout},
                '-eval_error'
            );
        } else {
            $output = &Tharn::say(
                $console_conn,
                "",
                ".*\r",
                '-noprompt',
                '-swallow',
                '-timeout' => $args{timeout},
                '-eval_error'
            );

        }

=comment
		my $obj=$self->get_conn_obj(type=>"cli")->
			system_node_coredump_status(
				node=>$self->get_node()
			)->get_parsed_output()->[0];

		my $total_blocks=$obj->{'total-blocks'};
		my $blocks_saved=$obj->{'blocks-saved'};
		my $state=$obj->{state};
		my $corename =$obj->{corename};

		$obj=$self->get_conn_obj(type=>"cli")->
			system_node_coredump_show(
				node=>$self->get_node(),
				corename=>$corename,
			)->get_parsed_output()->[0];

		my $saved=$obj->{'is-saved'};

		$self->_log(msg=>"\nCORENAME:     ".$corename);
		$self->_log(msg=>"TOTAL BLOCKS: ".$total_blocks);
		$self->_log(msg=>"BLOCKS SAVED: ".$blocks_saved);
		$self->_log(msg=>"STATE:        ".$state);
		#$self->_log(msg=>"SAVED:        ".$saved);
		$self->_log(msg=>"% DONE:       ".(($total_blocks=~/\-/)
					? 0
					: int(($blocks_saved/$total_blocks)*100)
					));


		$self->_log(msg=>"");
=cut

        ### Exit if no filer response is seen after retry limit for
        ### timeout period has expired
        if ($@ && $@ =~ /Err=TIMEOUT/i) {
            ### Lets try again
            if ($retry_cnt++ < $args{timeout_retries}) {
                $self->_log(msg => "");
                $self->_log(
                    msg => "WARNING: timed out after waiting "
                        . $args{timeout}
                        . "s for savecore response "
                        . "(retry: $retry_cnt)",
                    indent => 5
                );
                redo;
            } else {
                $self->_log(msg => "");
                $self->_log(msg => "ERROR: timed out after waiting "
                        . $args{timeout_retries}
                        . "times for "
                        . $args{timeout}
                        . "s for savecore response");
                $rc = 0;
                last;
            }
        }

        ### Parse the output that may contain 1 or more lines
        ### due to busy console
        @lines = split(/\r*\n+/, $output);
        map(s/^\s+//, @lines);
        map(s/\s+$//, @lines);
        map(s/\cM//g, @lines);
        ### Iterate over the output displaying selected
        ### messages of interest
        foreach my $line (@lines) {

            #$self->_log(msg=>"LINE: ".$line);

#[csiqa-3070-63:coredump.save.started:info]: Saving 497M to core.118046071.2010-09-21.21_12_20.nz (
            if ($line =~ /coredump.save.started.*]: (.* to (\S+))/) {
                $core     = $2;
                $core_msg = $1;
            } elsif ($line =~ /coredump.save.complete.*\]: (.*)/) {
                $self->_log(msg => "$1", indent => 3);
                $save_cnt += 1;
            } elsif ($line =~ /(Testpoint .*)/) {
                $self->_log(msg => "");
                $self->_log(msg => "$1", indent => 3);
                $rc = 0;
            } elsif ($line =~ /$done_filter/si) {
                ### Done filter looks to only occur once whether
                ### there is 1 or more unsaved cores to save.
                ### Assume success if at least one core was saved
                ### otherwise retry again to make sure its not
                ### a case where savecore operation hadn't started
                ### yet
                $complete = ($save_cnt > 0) ? 1 : 0;
                if ($complete == 0 and $loop_cnt++ >= $args{retries}) {
                    $fail_cnt++;
                    $rc = 0;
                } else {
                    $rc = 1;
                }
                last;
            } elsif ($line =~ /coredump.save.create.failed.*\]: (.*)/
                or $line =~ /coredump.save.error.*\]: (.*)/
                or $line =~ /coredump.save.aborted.*\]: (.*)/
                or $line =~ /coredump.save.nospace.*\]: (.*)/
                or $line =~ /coredump.save.attempts.count.*\]: (.*)/) {
                $self->_log(msg => "WARNING: $1");
                $fail_cnt++;
                $rc = 0;
                last;
            }
        }

        ### If a complete or fail_cnt flag is set
        ### then exit the while loop
        last if ($complete or $fail_cnt);

        ### Filter out the line showing status of save
        ### for whether savecore operation is async or
        ### sync
        my $filter;
        if ($args{_async}) {
            @lines = grep (/Saving core/, @lines);
            $filter = '(\S+): (\d+)';
        } else {
            @lines = grep (/done\)/, @lines);
            $filter = '\((\d+)\% done\)';
        }

        ### Parse the status line and display new status
        ### output format to STDOUT
        if ($lines[0] =~ /$filter/) {
            if ($args{_async}) {
                $core  = $1;
                $value = $2;
            } else {
                $value = $1;
            }

            next if ($value == 0);

            #next if ($core !~ /^core/ and $value !~ /\d+/);
            $first = 1 if ($last_core !~ /$core/);

            if ($first) {

                #$self->_log(msg=>"  - Done") if ($cnt++>0);
                $self->_log(msg => $core_msg, indent => 3)
                    if (defined $core_msg);
                $self->_log(
                    msg    => "Saving $core ...",
                    nocr   => 1,
                    indent => 3
                );
            }

            ### Only display out once it reaches
            ### one of the completion stage
            $pct = 10 if ($value >= 10 and $value <= 19);
            $pct = 20 if ($value >= 20 and $value <= 29);
            $pct = 30 if ($value >= 30 and $value <= 39);
            $pct = 40 if ($value >= 40 and $value <= 49);
            $pct = 50 if ($value >= 50 and $value <= 59);
            $pct = 60 if ($value >= 60 and $value <= 69);
            $pct = 70 if ($value >= 70 and $value <= 79);
            $pct = 80 if ($value >= 80 and $value <= 89);
            $pct = 90 if ($value >= 90 and $value <= 99);

            $self->_log(msg => $pct . "%", nocr => 1) if ($pct != $last_pct);
            $last_pct  = $pct;
            $last_core = $core;
            $first     = 0;

            ### Exit while loop if 'wait' option is set to false
            if (defined $pct and not $wait) {
                $self->_log(
                    msg    => "aborting wait for save core completion",
                    indent => 3
                );
                $rc = 0;
                last;
            }

            ### Set testpoint of abort option is set to true
            if ($args{abort} and $pct >= $args{abort} and not $tp_set) {
                $core_cnt
                    = ($self->get_core_cnt() > 0) ? $self->get_core_cnt() : 1;
                $self->tp_error(
                    module => "dumpcore",
                    id     => "abort_save",
                    cnt    => $core_cnt,
                    parm1  => 1,
                );
                $tp_set = 1;
            }

            ### Issue forced giveback if requested
            if (    $args{giveback_force}
                and $pct >= $args{giveback_force}
                and not $gbf_set) {
                $self->giveback(force => 1, waiting => 1);
                $gbf_set = 1;
            }

            ### Issue giveback if requested
            if ($args{giveback} and $pct >= $args{giveback} and not $gbf_set)
            {
                $self->giveback(force => 0, waiting => 1);
                $gbf_set = 1;
            }

        }
        ### Sleep before requesting
        ### next savecore status
        sleep($args{sleep});
    }

    ### Display duration time if a
    ### a core was saved
    if ($save_cnt) {
        $t2 = timelocal(localtime());
        if (defined $t2 and defined $t1) {
            $self->_log(
                msg => "save duration: " . ($t2 - $t1) . "s",
                indent => 5
            );
        }
    }

    #$self->_log(msg=>"rc=$rc");

    return $rc;
} ## end sub wait_for_savecore

###############################################################################
# Method:    wait_for_takeover
# Objective: wait until takeover of other node storage is complete
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wait_for_takeover {
    my ($self, %args) = @_;

    $args{msg} = "SF_MBWAIT";

    eval { $self->_wait(%args); };
    if ($@) {
        $self->_log(msg => $@, indent => 3);
        return 0;
    }

    return 1;
} ## end sub wait_for_takeover

###############################################################################
#  P  R  I  V  A  T  E
###############################################################################

###############################################################################
# Method:    _ems_check
# Objective: Check for existance of ems event in messages file
###############################################################################
sub _ems_check {
    my ($self, %args) = @_;
    my ($event, @lines, @messages,);

    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   event    => {type => SCALAR},
            messages => {
                type     => ARRAYREF,
                optional => 1,
                default  => $self->{messages}
            }
        }
    );

    @lines = grep (/$args{event}/, @{$args{messages}});
    @messages
        = (@lines) ? @lines : "ERROR: EMS message not found ($args{event})";
    if ($self->_get_loghdl()) {
        if (@lines) {
            foreach my $line (@lines) {
                my ($ems) = $line =~ /(\[.*\])/;
                $self->_log(msg => "ems message found: " . $ems, indent => 3);
            }
        } else {
            $self->_log(msg => "ERROR: ems message not found ($args{event})");
        }
    } else {
        $Tharn::Log->comment(@messages);
    }

    return (@lines) ? 1 : 0;
} ## end sub _ems_check

###############################################################################
# Method:    _get_partner
# Objective: retrieve partner name
###############################################################################
sub _get_partner {
    my ($self) = @_;
    my ($cliObj, $cmd, $name, $result, @lines,);

    $cliObj = $self->get_conn_obj(type => "cli");

    # UP 01:24:01, partner 'csiqa-3070-64', cluster monitor disabled
    ### Get partner name
    $cmd = "node run local cf monitor";
    $result = $cliObj->execute_raw_command(command => $cmd);
    ($name) = $result =~ /partner \'(\S+)\'/si;

    return $name;
} ## end sub _get_partner

###############################################################################
# Method:    _get_restricted_memsize
# Objective: retrieve restricted memsize bootarg parameter
###############################################################################
sub _get_restricted_memsize {
    my ($self) = @_;
    my ($bsdObj, $cmd, $result, $size, @lines,);

    $size = undef;
    $bsdObj = $self->get_conn_obj(type => "bsd");
    eval {
        ### Check if memory is restricted
        $cmd = "kenv | egrep \"\^bootarg.init.restricted_memsize=\"";
        $result = $bsdObj->execute_raw_command(command => $cmd);
        ($size) = $result =~ /\=\"(\S+)\"/;
    };

    return $size;
} ## end sub _get_restricted_memsize

###############################################################################
# Method:    _get_size
# Objective: retrieve core file size
###############################################################################
sub _get_size {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd, $name, $node, $parent, $size, @cmds,);

    ### Parse input
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   cid => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
            partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /\d+/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();
    $bsdObj = $self->get_conn_obj(type => "bsd");

    #$self->_log(msg=>Dumper($self->{cores_of}));

    $name = $self->{cores_of}{$node}{$args{cid}}{name};
    if ($args{partner}) {
        $bsdObj->execute_raw_command(command => 'sudo mount_partner');
    }

    $cmd
        = ($args{partner})
        ? "ls -lh /partner/etc/crash/" . $name . " | awk '{print \$5}'"
        : "ls -lh /mroot/etc/crash/" . $name . " | awk '{print \$5}'";

    ### Execute command
    $size = $bsdObj->execute_raw_command(command => $cmd);
    $size = 0 if ($size =~ /no such file/i);
    $size =~ s/\r*\n+//g;

    return $size;
} ## end sub _get_size

###############################################################################
# Method:    get_state
# Objective: Return machine connection state
###############################################################################
sub get_state {
    my ($self) = @_;
    my ($state);

    $self->_log(
        msg    => "retrieving " . $self->get_node() . " state",
        indent => 5
    );
    $state = $self->{transit}->get_state();

    ### return state
    return $state;
} ## end sub get_state

###############################################################################
# Method:    _has_restricted_memory
# Objective: retrieve restricted memsize bootarg parameter
###############################################################################
sub _has_restricted_memory {
    my ($self) = @_;
    my ($bsdObj, $cmd, $rc, $result,);

    $rc = 0;
    $bsdObj = $self->get_conn_obj(type => "bsd");

    eval {
        ### Check if memory is restricted
        $cmd    = "kenv | egrep \"\^bootarg.init.restricted_memsize=\"";
        $result = $bsdObj->execute_raw_command(command => $cmd);
        $rc     = 1 if ($result =~ /\=\"(\S+)\"/);
    };

    return $rc;
} ## end sub _has_restricted_memory

###############################################################################
# Method:    _init
# Objective: Iniitialize set of core objects
###############################################################################
sub _init_OLD {
    my ($self) = @_;
    my ($cores_of,);

    ### Retrieve the hash list of all core types
    $cores_of = $self->_list_cores();

    ### Iterate through list and create core objects
    foreach my $core_of (@{$cores_of}) {
        $core_of->{parent} = $self;
        push @{$self->{coreObjs}}, Os::Core->new(%$core_of);
    }

    return (wantarray()) ? @{$self->{coreObjs}} : $self->{coreObjs};
} ## end sub _init_OLD

##############################################################################
# Method:    _install_coretool
# Objective: Install coretool on filer
###############################################################################
sub _install_coretool {
    my ($self) = @_;
    my ($arch,   $bsdObj,  $cmd, $dir,     $ctool_loc,
        $errmsg, $options, $os,  %site_of, $site,
    );

    ### BUG: get_site_info() doesn't know where site file lives unless an
    ### Option option is instantiated.  The Option instance doesn't see site
    ### variable unless a Server object is created (I'll get site info from
    ### the env myself).
    $options = new Options();
    $options->process();
    $site    = $ENV{"SITE"};
    %site_of = get_site_info();

    $dir       = $self->{dir};
    $arch      = $self->_get_arch();
    $ctool_loc = $self->{ctool_loc};

    ### Copy over coretool file to the host directory under test
    ### using wget.
    unless ($site_of{$site}->{install_server}) {
        die "ERROR: No site_info file found using SITE parameter '$site'\n"
            . "Verify that a SITE parameter is defined in ~/.ntest file\n";
    }

    $bsdObj = $self->get_conn_obj(type => "bsd");

    $cmd
        = "wget "
        . $site_of{$site}->{install_proto} . '://'
        . $site_of{$site}->{install_user} . ':'
        . $site_of{$site}->{install_pass} . '@'
        . $site_of{$site}->{install_server}
        . $ctool_loc . "/"
        . $arch
        . "/coretool";

    $self->_log(
        msg    => ["upload 'CTOOL_LOC:$arch/coretool' to '$dir'"],
        indent => 5
    );
    $bsdObj->execute_raw_command(command => $cmd);

    ### Change permissions on ftest binary so it can be executed
    $bsdObj->execute_raw_command(
        command => "chmod 777 " . $self->{dir} . "/coretool");

    return 1;

} ## end sub _install_coretool

##############################################################################
# Method:    _is_coretool_installed
# Objective: Verify if coretool is installed on node
###############################################################################
sub _is_coretool_installed {
    my ($self) = @_;
    my ($bsdObj, $cmd, $exist, $output,);

    ### Initialize variables
    $cmd = "ls -lh /mroot/etc/crash/coretool";

    $bsdObj = $self->get_conn_obj(type => "bsd");

    ### Execute command
    $output = $bsdObj->execute_raw_command(command => $cmd);
    $exist = ($output =~ /no such file/i) ? 0 : 1;
    return $exist;
} ## end sub _is_coretool_installed

###############################################################################
# Method:    _load_cores
# Objective: retrieve list of cores
###############################################################################
sub _load_cores {
    my ($self, %args) = @_;
    my ($cliObj,
        $cmd,
        $console_conn,
        $core_of,
        $filter,
        $found,
        $node,
        $output,
        $partner,
        @lines,

        #%core_of,
        @cores_of,
    );

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   partner => {
                type      => SCALAR,
                optional  => 1,
                default   => 0,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0/i;
                        }
                }
            },
            status => {
                type      => SCALAR,
                optional  => 1,
                default   => "all",
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /all|^saved$|^unsaved$/i;
                        }
                }
            },
            name => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /.*\.nz/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $node = ($args{partner}) ? $self->{partner} : $self->get_node();

    $cliObj = $self->get_conn_obj(type => "cli");

=commment
	BUG:  CDEF is MISSING
	my @recs=$cliObj->system_node_coredump_show(
		node=>$node
	)->get_parsed_output();

	$self->_log(msg=>"CORES: \n".Dumper(@recs));
=cut

    $cmd    = "node run -node " . $node . " \"priv set test; savecore -l\"";
    $filter = "" if ($args{status} =~ /all/);
    $filter = ".* will be created by" if ($args{status} =~ /^unsaved$/i);
    $filter = "\$" if ($args{status} =~ /^saved$/i);

    $output = $cliObj->execute_raw_command(command => "$cmd");

    #$Tharn::Log->comment("OUTPUT ($cmd):\n".$output);

    @lines = split(/\r*\n+/, $output);
    map(s/^\s+//, @lines);
    map(s/\s+$//, @lines);
    map(s/\cM//g, @lines);
    @lines = grep (/\S+/, @lines);

    ### Construct hash structure for core data
    foreach my $line (@lines) {
        if ($line =~ /(core.*.nz)$filter/) {
            $core_of            = {};
            $core_of->{name}    = $1;
            $core_of->{status}  = "saved";
            $core_of->{core_id} = "";
            next
                if (defined $args{name}
                and $core_of->{name} !~ /^$args{name}$/);
            $found = 1;
        }
        next unless ($found == 1);
        if ($line =~ /Panic string: (.*)/) {
            $core_of->{panic_string} = $1;
        } elsif ($line =~ /Core ID: (\d+)/) {
            $core_of->{core_id} = $1;
            $core_of->{status}  = "unsaved";
        } elsif ($line =~ /Panic time: (.*)/) {
            $core_of->{timestamp} = $1;
            $core_of->{core_id}   = parsedate($core_of->{timestamp})
                if ($core_of->{status} =~ /^saved$/i);
        } elsif ($line =~ /OS version: (.*)/) {
            $core_of->{version} = $1;

            #push @{$self->{cores_of}{$core_of->{core_id}}}, $core_of;
            my $cid = parsedate($core_of->{timestamp});

            $Core_of->{$node}{$cid}{name}         = $core_of->{name};
            $Core_of->{$node}{$cid}{status}       = $core_of->{status};
            $Core_of->{$node}{$cid}{core_id}      = $core_of->{core_id};
            $Core_of->{$node}{$cid}{panic_string} = $core_of->{panic_string};
            $Core_of->{$node}{$cid}{timestamp}    = $core_of->{timestamp};
            $Core_of->{$node}{$cid}{version}      = $core_of->{version};
            $Core_of->{$node}{$cid}{size}
                = $self->_get_size(cid => $cid, partner => $args{partner});

            #$self->{cores_of}{$cid}{size} = $self->_get_size(cid=>$cid);
            #$self->{cores_of}

            $found = 0;
        }
    }

    #$self->_log(msg=>"DUMPER:\n".Dumper($self->{cores_of}));

    ### Return the result
    return $self->{cores_of};

} ## end sub _load_cores

###############################################################################
# Method:    _modify_rc
# Objective: Modify /mroot/etc/rc file
###############################################################################
sub _modify_rc {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd, $from, $to, $output, $rc,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   from => {type => SCALAR},
            to   => {type => SCALAR},
        }
    );
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $rc     = "/mroot/etc/rc";
    $from   = $args{from};
    $to     = $args{to};

    $cmd = "sed -i.orig 's/$from/$to/g' " . $rc;
    $output = $bsdObj->execute_raw_command(command => $cmd);

    return 1;
} ## end sub _modify_rc

################################################################################
# Method: _new()
# Objective: Create a new Coredump_7m object
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($bsdObj,  $cliObj,          $consObj, $errmsg,
        $hostObj, $memory_restrict, $options, $self,
        %site_of, $spObj,
    );

    ### bless instance into class
    $self = bless {}, $class;

    ### Validate input arguments
    ($class, %args) = validate_mixed(
        @_,
        [{type => SCALAR}],
        {

            #client     =>  {type => SCALAR},
            loghdl      => {type => SCALARREF,      optional => 1},
            ctool_loc   => {type => SCALAR | UNDEF, optional => 1},
            node        => {type => SCALAR},
            save_onboot => {
                type      => SCALAR,
                optional  => 1,
                default   => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0|on|off/i;
                        }
                }
            },
            sparse => {
                type      => SCALAR,
                optional  => 1,
                default   => 1,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0|on|off/i;
                        }
                }
            },
            timeout => {type => SCALAR, optional => 1}
        }
    );

    ### Store metadata in object
    #$self->{client}        = $args{client};
    $self->{hostObj}     = Tharn::host($args{node});
    $self->{loghdl}      = $args{loghdl};
    $self->{node}        = $args{node};
    $self->{timeout}     = $args{timeout};
    $self->{coreObjs}    = [];
    $self->{sparse_core} = $args{sparse};
    $self->{save_onboot} = $args{save_onboot};
    $self->{cores_of}    = $Core_of;
    ### Create connections to node
    $self->_log(
        msg    => "creating connections to " . $self->get_node(),
        indent => 3
    );
    push @{$self->{hostObj}->{prompt_norm}}, $args{node}
        . '\d{10}(?:|\\(takeover\\)|::[^>#%]*|__prompt__)\\*?[>#%]\\s';

    ### Create a transit object
    $self->_log(msg => "creating a transit object", indent => 3);
    $self->{transit} = NACL::Transit->new(name => $self->get_node());
    $self->{hostObj}{timeout} = 120;

    $cliObj  = $self->get_conn_obj(type => "cli");
    $bsdObj  = $self->get_conn_obj(type => "bsd");
    $consObj = $self->get_conn_obj(type => "console");

=comment
	### Retrieve ontap version
	my $rec=$cliObj->system_node_image_show(
		node=>$self->get_node(),
		isdefault=>"true"
	)->get_parsed_output()->[0];

	$self->_log(msg=>Dumper($rec));
	($self->{version_of}{ontap_version})=$rec->{'build-path'};
	($self->{version_of}{build_info})=$rec->{'build-type'};
=cut

    ### Retrieve ontap version
    my $output = $cliObj->execute_raw_command(
        command => "node run local \"version -v\"");
    ($self->{version_of}{ontap_version}) = $output =~ /(.*)\:/si;
    ($self->{version_of}{build_info})    = $output =~ /\:.*\<(\S+)\>/si;

    ### Make sure save on boot and sparse are set
    $self->_set_autoboot(enable => 1);
    $self->_set_gdb_recursive_panic(enable => 0);
    $self->_set_qos_classify_panic(enable => 0);
    $self->_set_log_level(level => "DEBUG");
    $self->set_save_onboot($self->{save_onboot});
    $self->set_sparse_core($self->{sparse_core});
    $self->set_failover(0);
    $self->set_auto_giveback(0);
    $self->set_auto_giveback_panic(0);
    $self->dump_attempts(value => 2);
    $self->mroot_minfree_limit(bytes => 262144000);
    $self->{partner} = $self->_get_partner();
    $self->{ctool_loc}
        = (defined $args{ctool_loc}) ? $args{ctool_loc} : CORETOOL_LOC;

    ### Set directory for bsd connection to /mroot/etc/crash
    $self->{dir} = "/mroot/etc/crash";
    $self->_log(msg => "setting directory to " . $self->{dir}, indent => 3);
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $bsdObj->execute_raw_command(command => "cd " . $self->{dir});

    ### Retrieve architecture to locate correct coretool
    ### version
    $self->_log(msg => "retrieving host architecture ", indent => 3);
    $self->{arch} = $bsdObj->execute_raw_command(command => "uname -p");
    $self->{arch} =~ s/\n//g;
    $self->_log(msg => $self->{arch}, indent => 5);

    ### Upload coretool to node if it doesn't exist on node already
    $self->_log(msg => "upload coretool to node", indent => 3);
    if ($self->_is_coretool_installed()) {
        $self->_log(msg => "coretool already installed", indent => 5);
    } else {
        $self->_install_coretool();
    }

    ### Create objects for any existing cores on node
    $self->_log(
        msg    => "creating core objects for any existing cores",
        indent => 3
    );

    my @cores = $self->_load_cores();

    ### Return handle to object
    return $self;
} ## end sub _new

################################################################################
# Method: _set_autoboot()
# Objective: Set autoboot state
################################################################################
sub _set_autoboot {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   enable => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0|on|off|true|false/i;
                        }
                }
            },
        }
    );

    $self->_log(
        msg => "setting autoboot "
            . (($args{enable} =~ /1|on|t/i) ? "true" : "false"),
        indent => 3
    );

    ### Initialize variables
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $cmd
        = ($args{enable} =~ /1|on|t/)
        ? "sudo sysctl kern.bootargs=AUTOBOOT=true"
        : "sudo sysctl kern.bootargs=AUTOBOOT=false";

    ### Execute command and parse output
    $bsdObj->execute_raw_command(command => "$cmd");

    return 1;
} ## end sub _set_autoboot

###############################################################################
# Method:    _set_log_level
# Objective: define whether core is automatically saved on bootup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _set_log_level {
    my ($self, %args) = @_;
    my ($cliObj, $value,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   level => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift()
                            =~ /NODE_FAULT|SVC_FAULT|NODE_ERROR|SVC_ERROR|WARNING|NOTICE|INFO|DEBUG|VAR/i;
                        }
                }
            },
        }
    );

    ### Initialize variables
    $cliObj = $self->get_conn_obj(type => "cli");

    $self->_log(
        msg    => "setting log level to " . $args{level},
        indent => 3
    );

    $cliObj->event_config_modify(consoleloglevel => $args{level});

    return 1;
} ## end sub _set_log_level

###############################################################################
# Method:    _set_gdb_recursive_panic
# Objective: enable/disable gdb entrance on recursive panic
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _set_gdb_recursive_panic {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   enable => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0|on|off|true|false/i;
                        }
                }
            },
        }
    );

    $self->_log(
        msg => "setting gdb-on-recursive-panic "
            . (($args{enable} =~ /1|on|t/i) ? "true" : "false"),
        indent => 3
    );

    ### Initialize variables
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $cmd
        = ($args{enable} =~ /1|on|t/)
        ? "sudo sysctl kern.bootargs=gdb-on-recursive-panic?=true"
        : "sudo sysctl kern.bootargs=gdb-on-recursive-panic?=false";

    ### Execute command and parse output
    $bsdObj->execute_raw_command(command => "$cmd");

    return 1;
} ## end sub _set_gdb_recursive_panic

################################################################################
# Method: _set_qos_classify_panic()
# Objective: Set qos_classify_panic state
################################################################################
sub _set_qos_classify_panic {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   enable => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /1|0|on|off|true|false/i;
                        }
                }
            },
        }
    );

    $self->_log(
        msg => "setting qos.classify.panic_how "
            . (($args{enable} =~ /1|on|t/i) ? 1 : 0),
        indent => 3
    );

    ### Initialize variables
    $bsdObj = $self->get_conn_obj(type => "bsd");
    $cmd
        = ($args{enable} =~ /1|on|t/)
        ? "sudo sysctl kern.bootargs=qos.classify.panic_how=1"
        : "sudo sysctl kern.bootargs=qos.classify.panic_how=0";

    ### Execute command and parse output
    $bsdObj->execute_raw_command(command => "$cmd");

    return 1;
} ## end sub _set_qos_classify_panic

###############################################################################
# Method:    _set_restricted_memsize()
# Objective: Set up firmware environment
###############################################################################
sub _set_restricted_memsize {
    my ($self, %args) = @_;
    my ($cmd, $console_conn, $errmsg,);

    ### Transit to firmware
    unless ($self->{transit}->get_state() =~ /FIRMWARE/) {
        $self->_log(msg => "changing state to FIRMWARE", indent => 3);
        $self->{transit}->change_state(to => 'FIRMWARE');
        $self->{conn_cli_exists} = 0;
        $self->{conn_bsd_exists} = 0;
    }

    ### Change environment
    $console_conn = $self->get_conn_obj(type => "console");
    $cmd
        = (length $args{memory} and $args{memory} != -1)
        ? "setenv bootarg.init.restricted_memsize" . " " . $args{memory}
        : "unsetenv bootarg.init.restricted_memsize";
    $self->_log(msg => "command: " . $cmd, indent => 3);
    &Tharn::say($console_conn, $cmd, -noprompt);

    ### Transit to firmware
    unless ($self->{transit}->get_state() =~ /CLI/) {
        $self->_log(msg => "changing state to CLI", indent => 3);
        $self->{transit}->change_state(to => "CLI", from => "FIRMWARE");
    }

    $self->_log(msg => "re-establishing connection", indent => 3);

    ### Re-establish connections
    $self->get_conn_obj(type => "cli");
    $self->get_conn_obj(type => "bsd");

    return $self;
} ## end sub _set_restricted_memsize

################################################################################
# Subroutine Name: _transit_to
# Objective: transition machine to requested state
################################################################################
sub _transit_to {
    my ($self, %args) = @_;
    my ($console_conn, $rc, $t1, $t2,);

    use boot;

    ### Validate input parameters
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   mode => {
                type      => SCALAR,
                callbacks => {
                    'legal values' => sub {
                        shift() =~ /FIRMWARE|BOOT_MENU|MAINT|UP/i;
                        }
                }
            },
            timeout => {
                type     => SCALAR,
                optional => 1,
                default  => 60,
                regex    => qr/^\d+$/
            },
        }
    );

    ### Transit to the requested state
    $console_conn = $self->get_conn_obj(type => "console", nologin => 1);
    $self->_log(
        msg    => "transitioning to " . uc($args{mode}) . " mode",
        indent => 7
    );

    $rc = boot::transit_to($console_conn, $args{mode},
        timeout => $args{timeout});

    return $rc;
} ## end sub _transit_to

###############################################################################
# Method:    _verify_marker
# Objective: verify unique marker was written to /mroot/etc/log/messages file
###############################################################################
sub _verify_marker {
    my ($self, %args) = @_;
    my ($bsdObj, $cmd, $duration, $output, $retry_cnt, $t1, $t2, @lines,);

    ### Validate input arguments
    ($self, %args)
        = validate_mixed(@_, [{type => OBJECT}],
        {marker => {type => SCALAR},});

    $self->_log(
        msg    => "Waiting for message cookie: '" . $args{marker} . "'",
        indent => 3
    );

    $bsdObj = $self->get_conn_obj(type => "bsd");
    $cmd = "grep '" . $args{marker} . "' /mroot/etc/log/messages";

    $t1 = timelocal(localtime());
    while (1) {
        $output = $bsdObj->execute_raw_command(command => "$cmd");
        if ($output !~ /$args{marker}/) {

            #if ($output !~ /sync check/i){
            if ($retry_cnt++ < 120) {
                sleep(1);
                redo;
            }
            return 0;
        }
        last;
    }

    $t2       = timelocal(localtime());
    $duration = $t2 - $t1;

    ### Display the time duration waiting for 'messages file' to
    ### recieve expected log markder
    $self->_log(msg => "WARNING: /mroot/etc/log/messages took "
            . $retry_cnt . "s "
            . "to get log marker")
        if ($retry_cnt);

    $self->_log(msg => "duration: " . $duration . "s") if ($duration >= 2);

    return 1;
} ## end sub _verify_marker

##############################################################################
# Method:    _wait
# Objective: wait until operation is complete
###############################################################################
sub _wait {
    my ($self, %args) = @_;
    my ($cliObj, $cmd, $console_conn, $partner_state,);

    ### Validate input arguments
    ($self, %args) = validate_mixed(
        @_,
        [{type => OBJECT}],
        {   msg   => {type => SCALAR},
            sleep => {
                type      => SCALAR,
                default   => $self->{sleep},
                callbacks => {
                    'Is An Integer>0' => sub { shift() > 0 }
                }
            },
            timeout => {
                type      => SCALAR,
                default   => $self->{timeout},
                callbacks => {
                    'Is An Integer>0' => sub {
                        shift() > 0;
                        }
                }
            },
        }
    );

    $cliObj = $self->get_conn_obj(type => "cli");

    $self->_log(msg => "waiting for msg '$args{msg}'", indent => 3);
    ### Execute command in loop until state or
    ### timeout met
    eval {
        local $SIG{ALRM}
            = sub { die "command timeout (" . $args{timeout} . "s)"; };
        alarm $args{timeout};
        while (1) {
            eval {

                my $resObj = $cliObj->storage_failover_show(
                    node => $self->get_node(),)->get_parsed_output()->[0];

                $partner_state = $resObj->{'partner-firmware-state'};

                $Tharn::Log->comment("Partner state: " . $partner_state);
            };
            if ($@) {
                next if ($@ =~ /error sending/i);
                die $@;
            }
            last if ($partner_state =~ /$args{msg}/si);

            Tharn::snooze($args{sleep});
        }
        alarm 0;
    };
    if ($@) {
        die $@;
    }

    return $self;
} ## end sub _wait

1;
