
package Os::EZIO_BASE;

use strict;

use NATE::STAF qw( staf_global staf_wrap );
use Carp;

sub new {
    my ( $class, %args ) = @_;
    my $self = bless {
        _client         => $args{client}         || undef,
        _servername     => $args{servername}     || undef,
        _driveletter    => $args{driveletter}    || "N:",
        _rtparameters   => $args{rtparameters}   || undef,
        _workdir        => $args{workdir}        || undef,
        _sharename      => $args{sharename}      || "ROOTSHARE",
        _cifsdomain     => $args{cifsdomain}     || "CSIQA",
        _username       => $args{username}       || "Administrator",
        _password       => $args{password}       || "netapp",
        _runtimeseconds => $args{runtimeseconds} || 0,
        _restorefile    => $args{restorefile},	
        _source_file    => undef,
        _source_lib     => undef,
        _dest_file      => undef,
        _dest_lib       => undef,
        _is_deployed    => 0,
        _is_setup       => 0,
        _ezio_handle    => "NONE;0;0",
        ,
        _output       => "No Output Captured\n",
        _datafilename => "",
        _dispose      => $args{dispose} || 1,
        _is_deployed  => 0,
        _is_started   => 0,
    }, $class;
    $self->_init( %args );
    return $self;
}

sub _init {
    my ( $self, %args ) = @_;
    if (defined($self->{_restorefile})) {
	$self->restoreme($self->{_restorefile});
    }
    $self->validate_parameters();
}

sub setup_client {
    my ($self) = @_;
    if ( $self->{_is_setup} ) { return }

    my $result = $self->_mount_cifs_drive(
        $self->{_client},    $self->{_driveletter}, $self->{_servername},
        $self->{_sharename}, $self->{_cifsdomain},  $self->{_username},
        $self->{_password},
    );

    croak( "Failed to setup client " . $self->{_client} ) if ( $result != 0 );
    $self->{_is_setup} = 1;
}

sub cleanup_client {
    my ($self) = @_;

    my $result =
      $self->_unmount_cifs_drive( $self->{_client}, $self->{_driveletter} );

    croak( "Failed to setup client " . $self->{_client} ) if ( $result != 0 );
}

sub deploy {
    my ($self) = @_;
    if ( $self->{_is_deployed} ) { return; }
    my $result = 0;

    $result = $self->_mkdir( $self->{_client}, $self->{_workdir} );
    croak( "Failed to make work directory on client " . $self->{_client} )
      if ( $result != 0 );

    $result =
      $self->_deploy_file( $self->{_client}, $self->{_source_file},
        $self->{_dest_file} );
    croak( "Faied to deploy executable files to client " . $self->{_client} )
      if ( $result != 0 );

    if ( defined( $self->{_source_lib} ) ) {
        $result =
          $self->_deploy_file( $self->{_client}, $self->{_source_lib},
            $self->{_dest_lib} );
        croak( "Failed to deploy library files to client " . $self->{_client} )
          if ( $result != 0 );
    }

    $self->{_is_deployed} = 1;
}

sub isstarted {
    my ($self) = @_;

    return $self->{_is_started};
}

sub init {
    croak("INIT not implemented");
}

sub pause {
    croak("PAUSE not implemented");
}

sub continue {
    croak("CONTINUE not implemented");
}

sub abort {
    croak("ABORT not implemented");
}

sub start {
    my $self = shift;
    if ( $self->{_is_started} ) { return; }
    $self->deploy();
    $self->setup_client();

    $self->{_ezio_handle} = $self->_process_start(
        $self->{_client},       $self->{_dest_file},
        $self->{_rtparameters}, $self->{_workdir},
        $self->{_runtimeseconds}
    );

    $self->{_is_started} = 1;

}

sub isdone {
    my ( $self, $ezio_handle ) = @_;
    return $self->_isdone( $self->{_ezio_handle} );
}

sub wait_done {
    my ( $self, $timeout ) = @_;
    return $self->_wait_done( $self->{_ezio_handle}, $timeout );
}

sub finish {
    my ($self) = @_;
    $self->_wait_done( $self->{_ezio_handle}, 10000 );
    return $self->end($self);
}

sub stop {
    my ($self) = @_;
    return $self->end();
}

sub end {
    my ($self) = @_;

    my $results = $self->_process_stop( $self->{_ezio_handle} );
    $self->{_output} =
      $self->_getoutput( $self->{_ezio_handle}, $self->{_workdir} );
    $self->cleanup_client();
    return $results;
}

sub validate_parameters {
    my ($self) = @_;
    my $drive = $self->{_driveletter};

    my $time     = $self->{_runtimeseconds};
    my $dtparams = $self->{_rtparameters};

    if ( ( $drive lt "F" ) || ( $drive gt "Z" ) ) {
        $self->usage();
        croak( "Invalid drive letter: [" . $drive . "]\n" );
    }
    if ( $time != 0 ) {
        if ( ( $time < 100 ) || ( $time > 100000 ) ) {
            $self->usage();
            croak("Invalid runtimeseconds parameter");
        }
    }
}

sub usage {
    my $usage = "USAGE:\n";
    $usage .= "\tUsage message (multi-line) goes here. \n";
    $usage .= "\tLine 1\n";
    $usage .= "\n";
    print $usage;
    return $usage;
}

sub printme {
    my ($self) = @_;
    my $prme =
        "EZIO_BASE"
      . "\n\t_client $self->{_client}"
      . "\n\t_servername $self->{_servername}"
      . "\n\t_driveletter $self->{_driveletter}"
      . "\n\t_rtparameters $self->{_rtparameters}"
      . "\n\t_filecount $self->{_filecount}"
      . "\n\t_workdir $self->{_workdir}"
      . "\n\t_sharename $self->{_sharename}"
      . "\n\t_cifsdomain $self->{_cifsdomain}"
      . "\n\t_username $self->{_username}"
      . "\n\t_password $self->{_password}"
      . "\n\t_source_file $self->{_source_file}"
      . "\n\t_source_lib $self->{_source_lib}"
      . "\n\t_dest_file $self->{_dest_file}"
      . "\n\t_dest_lib $self->{_dest_lib}"
      . "\n\t_is_deployed $self->{_is_deployed}"
      . "\n\t_is_setup $self->{_is_setup}"
      . "\n\t_output $self->{_output}" . "\n";
    print($prme);
    return $prme;
}

sub persistme {
    my ( $self, $fname ) = @_;
    if ( !defined($fname) ) {
        $fname = "dump_ezio_base_" . ${$} . ".persist";
    }
    my $dd = Data::Dumper->new( [$self], [qw(interim)] );
    open DUMP, ">", $fname or die "Cant open file";
    print DUMP $dd->Dump();
    close DUMP;
}

sub restoreme {
    my ( $self, $fname ) = @_;
    my $interim;

    if ( !defined($fname) ) {
        $fname = "dump_ezio_base_" . ${$} . ".persist";
    }
    if (-e $fname) {
	my $data = do {
		if ( open DUMP, '<', $fname ){
		    local $/; <DUMP> ;
		}
		else  { undef  }
	    };
	close DUMP;
	eval $data;

	    $self->{_client}         = $interim->{_client};
	    $self->{_servername}     = $interim->{_servername};
	    $self->{_driveletter}    = $interim->{_driveletter};
	    $self->{_rtparameters}   = $interim->{_rtparameters};
	    $self->{_workdir}        = $interim->{_workdir};
	    $self->{_sharename}      = $interim->{_sharename};
	    $self->{_cifsdomain}     = $interim->{_cifsdomain};
	    $self->{_username}       = $interim->{_username};
	    $self->{_runtimeseconds} = $interim->{_runtimeseconds};
	    $self->{_source_file}    = $interim->{_source_file};
	    $self->{_source_lib}     = $interim->{_source_lib};
	    $self->{_dest_file}      = $interim->{_dest_file};
	    $self->{_dest_lib}       = $interim->{_dest_lib};
	    $self->{_is_deployed}    = $interim->{_is_deployed};
	    $self->{_is_setup}       = $interim->{_is_setup};
	    $self->{_ezio_handle}    = $interim->{_ezio_handle};
	    $self->{_output}         = $interim->{_output};
	    $self->{_datafilename}   = $interim->{_datafilename};
	    $self->{_dispose}        = $interim->{_dispose};
	    $self->{_is_deployed}    = $interim->{_is_deployed};
	    $self->{_is_started}     = $interim->{_is_started};

    }
}

sub _transfer_file {
    my ( $self, $from_machine, $destination_machine, $sourcefile,
        $destination_file )
      = @_;

    my $staf = staf_global();
    my $cmd  = " COPY FILE $sourcefile "
      . " TOMACHINE $destination_machine TOFILE $destination_file";
    my $result = $staf->submit(
        endpoint => $from_machine,
        service  => "FS",
        request  => $cmd
    );
    if ( defined($result) ) {
        return $result;
    }
    else {
        return 0;
    }
}

sub _mount_cifs_drive {
    my (
        $self,      $endpoint,   $driveletter, $servername,
        $sharename, $cifsdomain, $username,    $password
    ) = @_;

    my $unmount_status = _unmount_cifs_drive( $self, $endpoint, $driveletter );
    $driveletter =~ s/://g;
    my $cmd = "net ";
    my $params =
      " use " . $driveletter . ": " . "\\\\" . $servername . "\\" . $sharename;
    my $auth = " /USER:$cifsdomain\\$username  $password ";
    $params .= $auth;
    my $timeout = 30;

    my $ezio_handle =
      _process_start( $self, $endpoint, $cmd, $params, $timeout );
    my $ezio_status = _wait_done( $self, $ezio_handle, 33 );
    my $output = _getoutput( $self, $ezio_handle );
    if ( $output =~ /completed successfully/ ) {
        return "\n\t" . $params . "\n\t" . $sharename . "\n\t" . $output . " 1";
        return 0;
    }
    return "\n\t" . $params . "\n\t" . $sharename . "\n\t" . $output . " 1";
}

sub _unmount_cifs_drive {
    my ( $self, $endpoint, $driveletter ) = @_;

    $driveletter =~ s/://g;
    my $cmd     = "net ";
    my $params  = " use " . $driveletter . ": " . "/d";
    my $timeout = 10;

    my $ezio_handle =
      _process_start( $self, $endpoint, $cmd, $params, $timeout );
    my $ezio_status = _wait_done( $self, $ezio_handle, 33 );
    my $output = _getoutput( $self, $ezio_handle );
    if ( $output =~ /completed successfully/ ) {
        return "\n\t" . $params . "\n\t" . $output . " 1";
        return 0;
    }
    return "\n\t" . $params . "\n\t" . $output . " 1";
    return 1;
}

sub _deploy_file {
    my ( $self, $destination_machine, $sourcefile, $destination_file ) = @_;
    return _transfer_file( $self, "local", $destination_machine, $sourcefile,
        $destination_file );
}

sub _process_start {
    my ( $self, $endpoint, $command, $params, $workdir, $timeout ) = @_;

    my $staf       = staf_global();
    my $uniquifier = int( rand(100000) );
    my $outfile    = $workdir . "/out_" . $uniquifier . ".txt";

    my $trailer =
      " WORKDIR $workdir NOTIFY ONEND STDERRTOSTDOUT  STDOUT " . $outfile;
    my $cmd =
        "START COMMAND "
      . staf_wrap($command)
      . " PARMS "
      . staf_wrap($params)
      . $trailer;

    my $staf_handle = $staf->submit(
        endpoint => $endpoint,
        service  => 'PROCESS',
        request  => $cmd
    );
    my $ezio_handle = $endpoint . ";" . $staf_handle . ";" . $uniquifier;
    return $ezio_handle;
}

sub _delete_file {
    my ( $self, $endpoint, $sourcefile ) = @_;

    my $staf   = staf_global();
    my $cmd    = " DELETE ENTRY $sourcefile CONFIRM ";
    my $result = $staf->submit(
        endpoint => $endpoint,
        service  => "FS",
        request  => $cmd
    );
    if ( defined($result) ) {
        return $result;
    }
    else {
        return 0;
    }
}

sub _mkdir {
    my ( $self, $endpoint, $pathname ) = @_;

    my $staf   = staf_global();
    my $cmd    = " CREATE DIRECTORY $pathname FULLPATH ";
    my $result = $staf->submit(
        endpoint => $endpoint,
        service  => "FS",
        request  => $cmd
    );
    if ( defined($result) ) {
        return $result;
    }
    else {
        return 0;
    }
}

sub _getoutput {
    my ( $self, $ezio_handle, $workdir ) = @_;
    my ( $endpoint, $idno, $unique ) = split( ';', $ezio_handle );
    my $results =
      "ERROR: Invalid results because process not done.  Wait process done.";

    if ( _isdone( $self, $ezio_handle ) ) {
        $results = "";
        eval {
            my $staf = staf_global();

            my $filename = $workdir . "/out_" . $unique . ".txt";
            my $cmd      = " GET FILE " . $filename;
            $results = $staf->submit(
                endpoint => $endpoint,
                service  => 'FS',
                request  => $cmd
            );

            _delete_file( $self, $endpoint, $filename );
        };
    }
    return $results;
}

sub _isdone {
    my ( $self, $ezio_handle ) = @_;
    my $results = _process_status( $self, $ezio_handle );

    return $results->{'isdone'};
}

sub _wait_done {
    my ( $self, $ezio_handle, $timeout ) = @_;
    my ( $endpoint, $idno, $unique ) = split( ';', $ezio_handle );
    my $status;
    my $debug = "debug\n\t";

    $status = _process_status( $self, $ezio_handle );
    if ( !$status->{isdone} ) {
        if ( $timeout > 0 ) {
          EARLYEXIT: {
                while ( $timeout > 0 ) {
                    $status = _process_status( $self, $ezio_handle );
                    if ( $status->{isdone} ) {
                        last EARLYEXIT;
                    }
                    else {
                        if ( $timeout >= 15 ) {
                            sleep(15);
                            $timeout -= 15;
                            $debug .= "sleep " . $timeout . "\n\t";
                        }
                        else {
                            sleep($timeout);
                            $timeout = 0;
                            $status = _process_status( $self, $ezio_handle );
                        }
                    }
                }
            }
        }
    }
    if ( defined($status) ) {
        if ( !$status->{isdone} ) {

            #             $status = undef;
        }
    }
    $status->{'debug'} = $debug;
    return $status;

}

sub _process_stop {
    my ( $self, $ezio_handle ) = @_;

    return _process_kill( $self, $ezio_handle );
}

sub DESTROY {
    ## Put what you need in here.  Currently nothing to do.
    ## Remember that these objects are allowed to persist
    ## themselves for later resurrection so you can't wontonly
    ## stop the I/O when the object goes out of existance.
}

## Process status returns this stucture
## {
##          'workdir' => undef,
##          'focus' => 'Background',
##          'staf-map-class-name' => 'STAF/Service/Process/ProcessInfo',
##          'key' => undef,
##          'pid' => '2188',
##          'startTimestamp' => '20120216-13:54:48',
##          'startMode' => 'Async',
##          'userName' => undef,
##          'shell' => undef,
##          'handleName' => undef,
##          'command' => 'ls.exe',
##          'rc' => '1',
##          'parms' => '-l',
##          'handle' => '65',
##          'title' => undef,
##          'endTimestamp' => '20120216-13:54:48',
##          'workload' => undef
##          ------- ADDED ---------------
##          'isdone' => 1 as true or 0 as false
##          'ezio_handle' => ezio_handle
##          'staf_handle' => same as 'handle' above, for clarity
##        };
##
sub _process_status {
    my ( $self, $ezio_handle ) = @_;
    my ( $endpoint, $idno, $unique ) = split( ';', $ezio_handle );

    my $staf        = staf_global();
    my $ezio_status = $staf->submit(
        endpoint => $endpoint,
        service  => " PROCESS ",
        request  => " QUERY HANDLE " . $idno
    );
    my $rc = $ezio_status->{'rc'};
    if ( defined($rc) ) {
        $ezio_status->{'isdone'} = 1;
    }
    else {
        $ezio_status->{'isdone'} = 0;
    }
    $ezio_status->{'ezio_handle'} = $ezio_handle;
    $ezio_status->{'staf_handle'} = $ezio_status->{'handle'};
    return $ezio_status;

}

sub _process_kill {
    my ( $self, $ezio_handle ) = @_;
    my ( $endpoint, $idno, $unique ) = split( ';', $ezio_handle );
    my $status = _process_status( $self, $ezio_handle );
    my $pid    = $status->{'pid'};
    my $info   = "unassigned";

    my $staf = staf_global();

    eval {
        $info = $staf->submit(
            endpoint => $endpoint,
            service  => "PROCESS",
            request  => " KILL PID " . $pid . " CONFIRM USING SIGKILLALL "
        );
    };

    sleep(1);
    return _process_status( $self, $ezio_handle );

}

sub _timestamp {

    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) =
      localtime(time);
    my $format = "%04d%02d%02d %02d%02d%02d";
    my $timestamp =
      sprintf( $format, $year + 1900, $mon + 1, $mday, $hour, $min, $sec );
    return $timestamp;

}

1;
