#Hey emacs, this is basically -*- perl -*-
#$Id$
## @summary Interface implementation for Fixed FreeBSD Memory Partitioning functions
## @author dl-kernel-qa
## @status private
## @pod POD is vailable./FixedFreeBSDMemory.pm
## v.1 Inital check-in
## v.2 Let the get_default_freebsd_size return 512MB instead of 512000000 to avoid confusion.
## v.3  (08/09/2012)
##     1. Modified the check condition in set_bootarg_init_low_mem so it only look for invalid switch
##     2. Modified High end default memory to 1024 MB in get_default_freebsd_size
##     3. Cleaned up unused lines.

package Os::FixedFreeBSDMemory;
use strict;
use NACL::APISet;
use Params::Validate qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use Tharn qw($Log);

################################################################################
# Method   : new()
# Objective: Create a new FixedFreeBSDMemory object
# Details  : See POD documentation at the end of this file
################################################################################
sub new {
    my $class = shift;
    my (
        $self,
        );

    ## Validate the input
    my %args = validate_with (
                 params => \@_,
                 spec => {
                     api_set_obj     => {isa  => 'NACL::APISet'},
                     api_firmware_obj => {isa => 'NACL::APISet'},
                     transit_obj     => {isa  => 'NACL::Transit'}
                 }
         );


    ### bless instance into class
    $self= bless{}, $class;

    die "Missing required parameter 'api_set_obj'" unless (defined $args{api_set_obj});
    

    $self->{api_set_obj} = $args{api_set_obj};
    $self->{api_firmware_obj} = $args{api_firmware_obj};
    $self->{transit_obj} = $args{transit_obj} if (defined $args{transit_obj});
    #$self->{verbose_mode} = defined $args{verbose_mode} ?  $args{verbose_mode}:0 ;

    return $self;
}   #End new()

###############################################################################
# Method:    get_freebsd_memory
# Objective: get the size of memory allocated for FreeBSD on the system
# Return:    return size or 0 for failed.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_freebsd_memory {
    my ($self, %args) = @_;
    my $resp;
    my $api_obj= $self->{api_set_obj};
    if ($self->go_to_systemshell()) {
        
        my $cmd = "sysctl -a hw.physmem";
	$resp = $api_obj->execute_raw_command(command=> $cmd);
	if ($resp =~ /hw.physmem: (\w+)/) {
            my $size = $1;
	     $Log->comment("FreeBSD memory size is $size");
             return $size;
        } else {
              $Log->comment ("Error: run_raw_command failed!");
              return 0;
        }
    } else {
        $Log->comment("Error: Could not get to FreeBSD shell...");
        return 0;
    }
}   


###############################################################################
# Method:    get_actual_physical_memory
# Objective: get the size of the physical memory on the system.
# Return:    Return either size or 0 for failed.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_actual_physical_memory {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    my $api_obj= $self->{api_set_obj};
    if ($self->go_to_systemshell()) {
        my $cmd = "sysctl -a hw.total_physmem";
        $resp = $api_obj->execute_raw_command (command=> $cmd);
	if ($resp=~ /hw.total_physmem: (\w+)/ ) {
	    my $size = $1;
	    return $size;
	} else {
	    $Log->comment ("Could not get physical memory of the system!");
	    return 0;
	}
    } else {
	$Log->comment("Error: Could not get to FreeBSD shell...");
	return 0;
    }
}

###############################################################################
# Method:    get_bootarg_init_low_mem
# Objective: get value of bootarg.init.low_mem
# Return:    Return either size, 0 for undefined, -1 for error
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_bootarg_init_low_mem {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    my $size;
     
    my $api_firmware_obj = $self->{api_firmware_obj};
    
    if ($self->go_to_firmware()) {
        my $cmd = "printenv bootarg.init.low_mem";
        $resp = $api_firmware_obj->execute_raw_command (command=> $cmd);
	if ($resp =~ /Undefined/) {
	    return 0;
	} else {
	    if ($resp =~ /bootarg.init.low_mem (\w+)/) {
		$Log->comment("got $resp");
		$size = $1;
                $Log->comment (" bootarg.init.low_mem is set to: $size");
		return $size;
	    } elsif ($resp =~ /(\w+)/) {
                $size = $1;
                $Log->comment ("for vsim bootarg.init.low_mem return $size");
                return $size;
            }
            else {
		$Log->comment ("Could not get value of bootarg.init.low_mem of the system!");
		return -1;
	    }
	}
    } else {
	$Log->comment("Error: Could not get to firmware...");
	return -1;
    }
}

###############################################################################
# Method:    set_bootarg_init_low_mem
# Objective: set bootarg.init.low_mem from ok prompt
# Return:    -1 for invalide value, 1 for success, 0 for fail
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_bootarg_init_low_mem {
    my $ret;
    my $resp;
    my ($self, %args) = @_;
    my $api_firmware_obj = $self->{api_firmware_obj};
    my $size = $args{low_mem_size};
    $Log->comment ("got bootarg.init.low_mem value is $size bytes");
    if ($self->go_to_firmware) {
        my $cmd = "setenv bootarg.init.low_mem $size";
        $resp = $api_firmware_obj->execute_raw_command (command=> $cmd);
        ### Invalid value is given
        if (($resp =~ /Invalid switch/)) {
            return -1;
        } else {
            ## make sure that the value is set
            $ret = $self->get_bootarg_init_low_mem();
            if ($ret == $size) {
                $Log->comment ("Successfully set bootarg.init.low_mem");
                return 1;
            } else {
                $Log->comment ("Failed to set bootarg.init.low_mem");
                return 0;
            }
        }
    } else {
        $Log->comment("Error: Could not get to firmware...");
        return 0;
    }
}


###############################################################################
# Method:    set_bootarg_init_restricted_mem
# Objective: set bootarg.init.restricted_memsize from ok prompt
# Return:    -1 for invalide value, 1 for success, 0 for fail
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_bootarg_init_restricted_mem {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    #my $api_obj= $self->{api_set_obj};
    my $api_firmware_obj= $self->{api_firmware_obj};
    my $size = $args{restricted_mem_size};
    $Log->comment ("got bootarg.init.restricted_mem value is $size");
    if ($self->go_to_firmware) {
        my $cmd = "setenv bootarg.init.restricted_memsize $size";
        $resp = $api_firmware_obj->execute_raw_command (command=> $cmd);
        ### Invalid value is given
        if (($resp =~ /Invalid switch/) && ($resp =~ /command status = -8/)) {
            return -1;
        } else {
            ## make sure that the value is set
            $ret = $self->get_bootarg_init_restricted_mem();
            if ($ret == $size) {
                $Log->comment ("Successfully set bootarg.init.restricted_memsize");
                return 1;
            } else {
                $Log->comment ("Failed to set bootarg.init.restricted_memsize");
                return 0;
            }
        }
    } else {
        $Log->comment("Error: Could not get to firmware...");
        return 0;
    }
}

###############################################################################
# Method:    get_bootarg_init_restricted_mem
# Objective: get value of bootarg.init.restricted_mem
# Return:    Return either size or 0 for undefined, -1 for error
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_bootarg_init_restricted_mem {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    my $size;
     
    my $api_firmware_obj= $self->{api_firmware_obj};
    if ($self->go_to_firmware()) {
        my $cmd = "printenv bootarg.init.restricted_memsize";
        $resp = $api_firmware_obj->execute_raw_command (command=> $cmd);
	if ($resp =~ /Undefined/) {
	    return 0;
	} else {
	    if ($resp =~ /bootarg.init.restricted_memsize (\w+)/) {
		$size = $1;
                $Log->comment ("bootarg.init.restricted_memsize is $size");
		return $size;
	    } elsif ($resp =~ /(\w+)/) {
                $size = $1;
                $Log->comment ("for vsim bootarg.init.restricted_memsize return $size bytes");
                return $size;
            }
            else {
		$Log->comment ("Could not get value of bootarg.init.restricted_memsize of the system!");
		return -1;
	    }
	}
    } else {
	$Log->comment("Error: Could not get to firmware...");
	return -1;
    }
}
###############################################################################
# Method:    unset_bootarg_init_low_mem
# Objective: unset bootarg.init.low_mem from ok prompt
# Return:    0 for fail and 1 for success
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unset_bootarg_init_low_mem {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    #my $api_obj= $self->{api_set_obj};
    my $api_firmware_obj= $self->{api_firmware_obj};
    if ($self->go_to_firmware()) {
	$Log->comment("unsetting bootarg.init.low_mem");
	my $cmd = "unsetenv bootarg.init.low_mem";
	$resp = $api_firmware_obj->execute_raw_command (command=> $cmd);
	if ($resp) {
	     $Log->comment ("failed to unset bootarg.init.low_mem");
	     return 0;
	} else {
	    ## double check and make sure that bootarg_init_low_mem is unset
	    if (!($self->get_bootarg_init_low_mem)) {
		$Log->comment ("successfully unset bootarg.init.low_mem");
		return 1;
	    }
	}
    } else {
	$Log->comment("Error: Could not get to Firmware...");
	return 0;
    }
}
     
###############################################################################
# Method:    unset_bootarg_init_restricted_memsize
# Objective: unset bootarg.init.restricted_memsize from ok prompt
# Return:    0 for fail and 1 for success
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unset_bootarg_init_restricted_memsize {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    my $api_firmware_obj= $self->{api_firmware_obj};
    if ($self->go_to_firmware()) {
	$Log->comment("unsetting bootarg.init.restricted_memsize");
	my $cmd = "unsetenv bootarg.init.restricted_memsize";
	$resp = $api_firmware_obj->execute_raw_command (command=> $cmd);
	if ($resp) {
	     $Log->comment ("failed to unset bootarg.init.restricted_memsize");
	     return 0;
	} else {
	    ## double check and make sure that bootarg_init_low_mem is unset
	    if (!($self->get_bootarg_init_restricted_mem)) {
		$Log->comment ("successfully unset bootarg.init.restricted_mem");
		return 1;
	    }
	}
    } else {
	$Log->comment("Error: Could not get to Firmware...");
	return 0;
    }
}

###############################################################################
# Method:    get_reboot_output
# Objective: use transit->reboot obj to get the output of the reboot session
# Return:    Return the console output from a reboot session
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_reboot_output {
    my ($self, %args) = @_;
    my $resp;
    my $ret;
    my $api_obj= $self->{api_set_obj};
    my $transit_obj = $self->{transit_obj};
    $resp = $transit_obj->reboot (append_output_to=>\$ret, inhibit_takeover=>1, iftakenover=>"wait", maxcycle=>40, timeout=>-1);
    return $ret;
}

###############################################################################
# Method:    go_to_systemshell
# Objective: transit console to systemshell
# Details:   See POD documentation at the end of this file.
###############################################################################
sub go_to_systemshell {
    my ($self, %args) = @_;
    my $transit_obj = $self->{transit_obj};
    my $state = $transit_obj->get_state();
    if ($state ne "SYSTEMSHELL") {
        sleep 60;
        if ($transit_obj->change_state(to=>'SYSTEMSHELL', inhibit_takeover=>1, iftakenover=>"wait", maxcycle=>40, timeout=>-1)) {
            $Log->comment("System has been transit to systemshell");
            return 1;
        } else {
            return 0;
            }
        
    } else {
        $Log->comment("System is at Systemshell");
        return 1;
    }
}
 
###############################################################################
# Method:    go_to_firmware
# Objective: transit console to firmware
# Details:   See POD documentation at the end of this file.
###############################################################################
sub go_to_firmware {
    my ($self, %args) = @_;
    my $transit_obj = $self->{transit_obj};
    my $state = $transit_obj->get_state();
    if ($state ne "FIRMWARE") {
        if ($transit_obj->change_state(to=>'FIRMWARE',inhibit_takeover=>1, iftakenover=>"wait", maxcycle=>40, timeout=>-1)) {
            $Log->comment("System has been transit to LOADER> prompt");
            return 1;
        } else {
            return 0;
        }
    } else {
        $Log->comment("System is at ok prompt");
        return 1;
    }
}

###############################################################################
# Method:    go_to_cli
# Objective: transit console to command line
# Details:   See POD documentation at the end of this file.
###############################################################################
sub go_to_cli {
    my ($self, %args) = @_;
    my $transit_obj = $self->{transit_obj};
    my $state = $transit_obj->get_state();
    if ($state ne "CLI") {
        if ($transit_obj->change_state(to=>'CLI', inhibit_takeover=>1, iftakenover=>"wait", maxcycle=>40, timeout=>-1)) {
            $Log->comment("System has been transit to command line");
            return 1;
        } else {
            return 0;
        }
    } else {
        $Log->comment("System is at command line");
        return 1;
    }
}


##########################################################
# Method: get_default_freebsd_size                       #
# Description: get the size of FreeBSD memory based on   #
#              the physical memory available on system   #
# Args:  N/A                                             #
# Return: return default FreeBSD memory size             #
# Dependencies : None                                    #
#########################################################
sub get_default_freebsd_size {
    my ($self, %args) = @_;
    my $size;
    my $platform;
     
    my $low_end_freebsd_mem = "521MB";
    my $middle_end_freebsd_mem = "760MB";
    my $middle_to_high_end_freebsd_mem = "825MB";
    my $high_end_freebsd_mem = "1024MB";

    
    if (defined $args{size}) {
        $size = $args{size};
    }
    if (defined $args{platform}) {
        $platform = $args{platform};
    }
        ### Vsim default size is 760MB^M
        # 2G is equal to 2 * 1024 *1024 * 1024 = 2147483638 bytes^M
        # 4G is equal to 4 * 1024 *1024 * 1024 = 4294967296 bytes^M
        # 8G is equal to 8 * 1024 *1024 * 1024 = 8589934592 bytes^M
        # 16G is equal to 16 * 1024 * 1024 * 1024 = 17179869841 bytes^M

    if ($platform eq "VSim") {
           return $middle_end_freebsd_mem;
    } else {
        if ( $size < 4294541312 ) {
            return $middle_end_freebsd_mem;
        } elsif ((4294541312 <= $size ) && ($size < 17179443200)) {
            return $middle_to_high_end_freebsd_mem;
        } else {
            return $high_end_freebsd_mem;
        }
    }
}

##########################################################
# Method: compare_size                       #
# Description: get the percentage difference between the #
#              size of default freeBSD memory and the    #
#              actual size get from sysctl.              #
# Args:  N/A                                             #
# Return: return the number of difference                #
# Dependencies :                                    #
#########################################################
sub compare_size {
    my ($self, %args) = @_;
    my $bsd_size;
    my $actual_bsd_size;
    if (defined $args{bsdsize}) {
        $bsd_size = $args{bsdsize};
    }
    if (defined $args{actualsize}) {
        $actual_bsd_size = $args{actualsize};
    }
    $Log->comment ("bsd_size is $bsd_size and actual size is $actual_bsd_size");
    my $temp = ($bsd_size - $actual_bsd_size)/$bsd_size;
    return $temp;
    
}

sub cleanup_APISet {
    my ( $self, %args ) = @_;
    $Log->enter();

    $self->{session}->cleanup();
    delete $self->{session};
    delete $self->{schema};

    $Log->exit();
} ## end sub cleanup_APISet


1;

############################################################################
#  P O D
############################################################################

=pod

=head1 Package Name

Os::FixedFreeBSDMemory

=head1 Author

hongling@netapp.com

=head1 Description

Os::FixedFreeBSDMemory package provides an API interface for general methods that will be used for Fixed FreeBSD memory paritioning
project. Most of the methods in this API interface can be used by Flexmem project in FS.0 release as well.

=head1 Public Methods

=head2 B<new> 

Constructor that will be used to create a new Os::FixedFreeBSDMemory object.

=over 6

=item Inputs

=over 6

=item * B<api_set_obj>

Name of api_set_obj that will be used for CLI level operation

=item * B<api_firmware_obj>

Name of api_firmware_obj that will be used for the firmware level operation, like issue commands from LOADER>.

=item * B<transit_obj>

Name of the transit object that will be used to transfer filer from one state to another

=back

=item Outputs

Returns a Os::FixedFreeBSDMemory object.

=item Examples

 ### Create a FixedFreeBSDmemory object
 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$MSObj = Os::FixedFreeBSDMemory->new (
					api_set_obj => $Api_set_obj,
                                        api_firmware_obj => $api_firmware_obj,
			                transit_obj => $Transit_obj
	);


=back


=head2 B<get_freebsd_memory>

Method for getting the size of freeBSD memory.

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Returns the size of FreeBSD memory or 0 for failures.

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->get_freebsd_memory();
	);


=back

=head2 B<get_actual_physcial_memory>

Method for getting the size of the physical memory available on the system.

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Returns the size of total physical memory or 0 for failures.

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->get_actual_physical_memory();
	);


=back

=head2 B<get_bootarg_init_low_mem>

Method for getting the status of bootarg.init.low_mem from boot loader.

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Returns the current value of bootarg.init.low_mem, 0 for undefined, -1 for failures.

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->get_bootarg_init_low_mem();
	);


=back

=head2 B<get_bootarg_init_restricted_mem>

Method for getting the status of bootarg.init.restricted_memsize from boot loader.

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Returns the current value of bootarg.init.low_mem, 0 for undefined, -1 for failures.

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->get_bootarg_init_restricted_mem();
	);


=back

=head2 B<set_bootarg_init_low_mem>

Method for setting the status of bootarg.init.low_mem from boot loader.

=over 6

=item Inputs

=over 6

=item * low_mem_size

=back

=item Outputs

Return -1 for invalide value, 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->set_bootarg_init_low_mem(low_mem_size=>$size);
	);


=back

=head2 B<set_bootarg_init_restricted_mem>

Method for setting the status of bootarg.init.restricted_mem from boot loader.

=over 6

=item Inputs

=over 6

=item * restricted_mem_size

=back

=item Outputs

Return -1 for invalide value, 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->set_bootarg_init_low_mem(restricted_mem_size=>$size);
	);


=back

=head2 B<unset_bootarg_init_low_mem>

Method for unsetting the value of bootarg.init.low_mem from boot loader if it is defined

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Return 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->unset_bootarg_init_low_mem();
	);


=back

=head2 B<unset_bootarg_init_retricted_mem>

Method for unsetting the value of bootarg.init.restricted_mem from boot loader if it is defined

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Return 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->unset_bootarg_init_restricted_mem();
	);


=back

=head2 B<get_reboot_output>

Method for getting the console output from a reboot

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Return console output

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->get_reboot_output();
	);


=back

=head2 B<go_to_systemshell>

Method for transfering the system to the systemshell

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Return 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->go_to_systemshell();
	);


=back

=head2 B<go_to_firmware>

Method for transfering the system to the firmware

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Return 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->go_to_firmware();
	);


=back

=head2 B<go_to_cli>

Method for transfering the system to the CLI command line

=over 6

=item Inputs

=over 6

=item * No

=back

=item Outputs

Return 1 for success, 0 for fail

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->go_to_cli();
	);


=back

=head2 B<get_default_freebsd_size>

Method for getting the default freebsd memory size for system

=over 6

=item Inputs

=over 6

=item * size

=back

=item Outputs

Return size for each system

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->get_default_freebsd_size(size=>$size);
	);


=back

=head2 B<compare_size>

Method for compareing the default FreeBSD memory size and the acutal one
(Note: this is because the actual memory size is a little off the default value, for example
        760 MB may be 735MB actually which is correct).

=over 6

=item Inputs

=over 6

=item bsdsize actualsize

=back

=item Outputs

Return size for each system

=item Examples

 use TCD;
 use Os::FixedFreeBSDMemory;
 ...
	$ret = $BSDmem_obj->compare_size(bsdsize=>$size1, actualsize=>$size2);
	);


=back
=cut
