# Hey emacs, this is basically -*- perl -*-
# vim: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:
#
###############################################################################
# Copyright (c) 2013 NetApp, Inc.
# All rights reserved.
###############################################################################

=pod

=head1 NAME

Os::KernelTestSetup

=for html <hr>

=head1 DESCRIPTION

This library's purpose is to aid OS Kernel QA (and similar) automation,
by providing methods for setup and cleanup of common test resources
such as volume layouts with nested junctions, vservers, etc.

=for html <hr>

=head1 SYNOPSIS

# Create a new object instance
my $oslib = Os::KernelTestSetup->new();

# Setup test volumes
$oslib->setup_volumes();

# Setup test vservers
$oslib->setup_vservers();

# Cleanup only test volumes
$oslib->cleanup_volumes();

# Cleanup ALL test resources
$oslib->cleanup();

=head1 ATTRIBUTES

=head2 volumes

A list of volumes created by this library

=head2 vservers

A list of vservers created by this library

=head2 volume_name_prefix

SCALAR: the default name prefix to use for volume names

=head2 vserver_name_prefix

SCALAR: the default name prefix to use for vserver names

=head2 cleanup

NACL cleanup object

=cut


package Os::KernelTestSetup;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use Data::Dumper qw(Dumper);

use Tharn;
use NATE::BaseException;

# NACL
use NACL::C::Node;
use NACL::C::StorageAggregate;
use NACL::STask::Volume;
use NACL::STask::Vserver;
use NACL::GeneralUtils;

# Method generator for set/get
use Class::MethodMaker [
        array => [
            { -default => [] },
            'volumes'
        ],
        array => [
            { -default => [] },
            'vservers'
        ],
        scalar => 'volume_name_prefix',
        scalar => 'vserver_name_prefix',
        scalar => [
            {   '-default_ctor' =>
                    sub { NACL::Realm->realm()->cleanup() },
                '-type' => 'NACL::Cleanup'
            },
            'cleanup_manager',
       ],
       new => 'new',
    ];

=for html <hr>

=head1 METHODS

=head2 new

=over

Instantiates a new Os::KernelTestSetup object.

=item Options

=over

=item none

=back

=item Returns

 Returns a new Os::KernelTestSetup object instance.

=item Example

 my $oslib = Os::KernelTestSetup->new();

=back

=cut


###############################################################################
#           Primary setup / cleanup methods
###############################################################################

=for html <hr>

=head2 setup_volumes

Create test volumes based on a user-defined layout.
Uses (and supports args for) NACL::STask::Volume. Actual cleanup is done by the STask.

=over

=item Options

=over

=item nodes => \@node_objects

 (Required) an array ref to the NACL::C::Node components
 of the nodes whose aggregates should be used for this setup step.

=item count => SCALAR

 (Required) how many volumes to create

=item name_prefix => SCALAR

 (Optional) a name prefix for the volume
 Default: the library will generate a name of the format:
            kernel_test_volume_<random-word>_<timestamp>_<x>
 where, x = the number of resource being created (1, 2, .. 100)

=item nested_junctions => 1|0

 (Optional) should nested junctions be created? if yes, the first volume
 will be the parent of the nested junction.
 [Default: 0 (false - don't create nested junctions)]

=item spread_evenly => 1|0

 (Optional) should the volumes be spread evenly across all SFO aggregates?
 '0' means use only one aggregate, '1' means spread evenly across SFO aggregates.
 [Default: 1 (true)]

=item %other_args

 (Optional) any additional args supported by the NACL::STask::Volume task.
 Note that these args will be applied to *ALL* volumes.
 For example, the arg (vserver => 'vs123') can be specified for creating
 all the volumes on that vserver.

=back

=item Returns

 An array of volume STask objects if successful, throws an exception on failure.

=item Examples

 # general parameters
 my @nodes = ($node_a_object, $node_b_object);

 # create new volumes on aggregates owned by these nodes
 my @volumes = $oslib->setup_volumes(
        nodes            => \@nodes,
        vserver          => 'vs123',
        volume_count     => 5,      # how many volumes to create
        nested_junctions => 1,      # default = 0 (false)
    );

 # reuse existing volumes on aggregates owned by these nodes
 $oslib->setup_volumes(
        nodes              => \@nodes,
        size               => '100M',       # size for all volumes = 100MB
        volume_count       => 5,            # how many volumes to create
        nested_junctions   => 1,            # default = 0 (false)
        nacltask_if_exists => 'reuse',      # NACL: reuse the volume if it already exists
    );

=back

=cut

sub setup_volumes {

    my $self = shift;
    my %args = Params::Validate::validate_with(
                params => \@_,
                spec   => {
                    nodes => {
                        type     => ARRAYREF,
                    },
                    count => {
                        type     => SCALAR,
                    },
                    name_prefix => {
                        type     => SCALAR,
                        optional => 1,
                    },
                    nested_junctions => {
                        type     => SCALAR,
                        default  => 0,
                    },
                    spread_evenly => {
                        type     => SCALAR,
                        default  => 1,
                    },
                },
                # allow extra args to pass them through to NACL as-is
                allow_extra => 1,
            );

    #----------------------------------------------------------------------
    # create the test volumes
    my @volumes_created = $self->_create_resources(
                                resource_type => 'volume',
                                %args,
                            );

    #----------------------------------------------------------------------
    # perform any additional tasks such as:
    # - create a few files, links, etc.
    # - create a few snapshots
    # TO-DO

    return @volumes_created;
}


=for html <hr>

=head2 setup_vservers

Create test vservers based on a user-defined layout.
Uses (and supports args for) NACL::STask::Vserver. Actual cleanup is done by the STask.

=over

=item Options

=over

=item nodes => \@node_objects

 (Required) an array ref to the NACL::C::Node components
 of the nodes whose aggregates should be used for this setup step.

=item count => SCALAR

 (Required) how many vservers to create

=item name_prefix => SCALAR

 (Optional) a name prefix for the volume
 Default: the library will generate a name of the format:
            kernel_test_volume_<random-word>_<timestamp>_<x>
 where, x = the number of resource being created (1, 2, .. 100)

=item spread_evenly => 1|0

 (Optional) should the vserver root volumes be spread evenly across all SFO aggregates?
 '0' means use only one SFO aggregate, '1' means spread evenly across SFO aggregates.
 [Default: 1 (true)]

=item %other_args

 (Optional) any additional args supported by the NACL::STask::Vserver task.
 Note that these args will be applied to *ALL* vservers.

=back

=item Returns

 An array of vserver STask objects if successful, throws an exception on failure.

=item Examples

 # general parameters
 my @nodes = ($node_a_object, $node_b_object);

 # create new volumes on aggregates owned by these nodes
 my @vservers = $oslib->setup_vservers(
        nodes => \@nodes,
        count => 5,      # how many volumes to create
    );

 # reuse existing vservers
 $oslib->setup_vservers(
        nodes              => \@nodes,
        count              => 5,            # how many vservers to create
        nacltask_if_exists => 'reuse',      # NACL: reuse the vserver if it already exists
    );

=back

=cut

sub setup_vservers {

    my $self = shift;
    my %args = Params::Validate::validate_with(
                params => \@_,
                spec   => {
                    nodes => {
                        type     => ARRAYREF,
                    },
                    count => {
                        type     => SCALAR,
                    },
                    name_prefix => {
                        type     => SCALAR,
                        optional => 1,
                    },
                    spread_evenly => {
                        type     => SCALAR,
                        default  => 1,
                    },
                },
                # allow extra args to pass them through to NACL as-is
                allow_extra => 1,
            );

    #----------------------------------------------------------------------
    # create the test vservers
    my @vservers_created = $self->_create_resources(
                                resource_type => 'vserver',
                                %args,
                            );

    #----------------------------------------------------------------------
    # perform any additional tasks if requested:
    # TO-DO

    return @vservers_created;
}

=for html <hr>

=head2 cleanup

Cleanup test resources; created by this library.

=over

=item Options

=over

=item NONE

=back

=item Returns

 Nothing if successful, throws an exception on failure.

=item Examples

 $oslib->cleanup();

=back

=cut

sub cleanup {

    my $self = shift;

    # clean up all test resources created by this library
    $self->cleanup_manager()->run();

    return;

}

=for html <hr>

=head2 cleanup_volumes

Cleanup test volumes created by this library.

=over

=item Options

=over

=item NONE

=back

=item Returns

 Nothing if successful, throws an exception on failure.

=item Examples

 $oslib->cleanup_volumes();

=back

=cut

sub cleanup_volumes {

    my $self = shift;

    # clean up test volumes
    my @volumes_to_cleanup = $self->volumes();
    if (@volumes_to_cleanup) {

        # cleanup volume STask objects by invoking their cleanup
        $self->cleanup_manager()->run(objects => \@volumes_to_cleanup);

        # reset the setup flag to avoid duplicate cleanup in case
        # the test calls cleanup() multiple times by accident
        $self->volumes( [] );
    }

    return;

}

=for html <hr>

=head2 cleanup_vservers

Cleanup test vservers created by this library.

=over

=item Options

=over

=item NONE

=back

=item Returns

 Nothing if successful, throws an exception on failure.

=item Examples

 $oslib->cleanup_vservers();

=back

=cut

sub cleanup_vservers {

    my $self = shift;

    # clean up test vservers
    my @vservers_to_cleanup = $self->vservers();
    if (@vservers_to_cleanup) {

        # cleanup vserver STask objects by invoking their cleanup
        $self->cleanup_manager()->run(objects => \@vservers_to_cleanup);

        # reset the setup flag to avoid duplicate cleanup in case
        # the test calls cleanup() multiple times by accident
        $self->vservers( [] );
    }

    return;

}

###############################################################################
#                 Utility methods for helper tasks & tests
###############################################################################

=for html <hr>

=head2 find_aggregates_for_volumes

Takes node objects, finds usable SFO aggregates on them, and then
arranges them by node, followed by ordering them based on whether
even-spreading across all nodes is requested or if we should
simply exhaust all aggregates on each node before proceeding to
the next node's aggregates.

=over

=item Options

=over

=item nodes => \@node_objects

 (Required) an array ref to the NACL::C::Node components
 of the nodes whose aggregates should be used for this setup step.

=item spread_evenly => 1|0

 (Optional) should the volumes be spread evenly across all SFO aggregates?
 '0' means use all SFO aggregates on 1 node, then the next node, and so on
 '1' means spread evenly across SFO aggregates
 [Default: 1 (true)]

=back

=item Returns

 An array of aggregate names if successful,
 throws an exception on failure.

=item Examples

 # general parameters
 my @nodes = ($node_a_object, $node_b_object);

 # create new volumes on aggregates owned by these nodes
 # (using default behavior for 'spread_evenly')
 my @aggregates = $oslib->find_aggregates_for_volumes(
                    nodes => \@nodes,
                );

 # find aggregates, spread evenly across all nodes
 my @aggregates = $oslib->find_aggregates_for_volumes(
                    nodes         => \@nodes,
                    spread_evenly => 1,
                );

 # find aggregates, NOT spread evenly across all nodes,
 # but listed by aggregates per node
 my @aggregates = $oslib->find_aggregates_for_volumes(
                    nodes         => \@nodes,
                    spread_evenly => 1,
                );

=back

=cut

sub find_aggregates_for_volumes {
    my $self = shift;
    my %args = Params::Validate::validate(
                @_,
                {
                    spread_evenly => {
                        type     => SCALAR,
                        default  => 1,
                    },
                    nodes => {
                        type      => ARRAYREF,
                        callbacks => {
                            'NACL::C::Node' =>
                                sub {
                                    my @nodes = @{$_[0]};
                                    scalar @nodes > 0 &&
                                    (@nodes ==
                                        grep { UNIVERSAL::isa($_, 'NACL::C::Node') } @nodes);
                                },
                        },
                    },
                });

    # get inputs
    my @nodes         = @{$args{nodes}};
    my $spread_evenly = $args{spread_evenly};
    my @return_aggrs  = ();

    # setup a filter string to filter aggregates owned only
    # by the nodes provided to us. e.g. - 'node1|node2'
    my $node_filter = join("|", map {$_->node()} @nodes);

    $Log->comment("\nLooking for online SFO aggregates using node filter " .
                  "{node => $node_filter}...\n");
    # find all usable aggregates across the cluster
    my @aggregates = NACL::C::StorageAggregate->find_online_non_root_arranged_by_most_availsize(
        command_interface => $nodes[0],
        filter            => { nodes => [ $node_filter ] },
    );

    $Log->debug(sub {
                    "\nAll usable aggregates across the cluster:\n" .
                    join ("\n", map{$_->aggregate()} @aggregates)
                }
    );

    # the following post-processing is done to avoid multiple NGSH calls
    # for "storage aggregate show"
    my %aggr_map;
    foreach my $aggregate (@aggregates) {
        my @owner_nodes = $aggregate->get_nodes();
        my $owner_node  = $owner_nodes[0]->node();
        push (@{$aggr_map{$owner_node}}, $aggregate->aggregate());
    }

    if ($spread_evenly == 1) {
        # spread evenly if requested
        $Log->comment("\nProcessing aggregates for spreading volumes evenly across all nodes...\n");
        while (scalar keys %aggr_map > 0) {
            foreach my $node_name (keys %aggr_map) {
                # delete the node's hash entry if its aggregate list is depleted
                delete $aggr_map{$node_name}
                    if (scalar @{$aggr_map{$node_name}} < 1);

                # take the next element in the list for each node and push it
                # onto the processed aggregates' list
                push (@return_aggrs, shift(@{$aggr_map{$node_name}}))
                    if (defined $aggr_map{$node_name});
                    
            }
        }
        # end of even-spreading

    } else {
        # don't spread evenly - simply append each nodes' aggregate list
        # the processed aggregates' list
        $Log->comment("\nProcessing aggregates serially by each node...\n");
        foreach my $node_name (keys %aggr_map) {
            push (@return_aggrs, @{$aggr_map{$node_name}});
        }
    }

    # DEBUG: print the list of aggregates processed by this method
    $Log->debug(sub {
                    "\nProcessed list of aggregates (in-order):\n" .
                    join("\n", @return_aggrs)
                }
    );

    return @return_aggrs;
}

=for html <hr>

=head2 volumes

Return an array of volumes created by this test library.

=over

=item Options

=over

=item NONE

=back

=item Returns

 Array of objects of type NACL::STask::Volume

=item Examples

 # get a list of all volumes
 my @volumes = $oslib->volumes();
 
=back

=cut

#----------------------------------------------------------#
# volumes() : Above method generated by Class::MethodMaker #
#----------------------------------------------------------#

=for html <hr>

=head2 vservers

Return an array of vservers created by this test library.

=over

=item Options

=over

=item NONE

=back

=item Returns

 Array of objects of type NACL::STask::Vserver

=item Examples

 # get a list of all vservers
 my @vservers = $oslib->vservers();
 
=back

=cut


#-----------------------------------------------------------#
# vservers() : Above method generated by Class::MethodMaker #
#-----------------------------------------------------------#


###############################################################################
#                 Private helper methods for internal purposes
###############################################################################

#-----------------------------------------------------------------------------#
# METHOD  : _create_resources
#
# SUMMARY : Takes node objects, a count of resources to create, and all other
#           supported arguments for creating volumes and vservers.
#           This private method is a common driver method for setup_volumes()
#           and setup_vservers(), that loads the internal attributes for the
#           created volumes and vservers once created.
#
# INPUTS  :
#           Please see POD for setup_volumes() and setup_vservers()
#
# RETURNS : ARRAY of resources created if successful, exception for errors.
#
#-----------------------------------------------------------------------------#

sub _create_resources {

    my $self = shift;
    my %args = Params::Validate::validate_with(
                params => \@_,
                spec   => {
                    nodes => {
                        type     => ARRAYREF,
                        callbacks => {
                            'NACL::C::Node' =>
                                sub {
                                    my @nodes = @{$_[0]};
                                    scalar @nodes > 0 &&
                                    (@nodes ==
                                        grep { UNIVERSAL::isa($_, 'NACL::C::Node') } @nodes);
                                },
                        },
                    },
                    resource_type => {
                        type     => SCALAR,
                        regex    => qr/volume|vserver/,
                    },
                    count => {
                        type     => SCALAR,
                    },
                    name_prefix => {
                        type     => SCALAR,
                        optional => 1,
                    },
                    spread_evenly => {
                        type     => SCALAR,
                        default  => 1,
                    },
                },
                # allow extra args to pass them through to NACL,
                # or volume/vserver specific parameters that have
                # already been validated by setup_[volumes|vservers]
                # e.g. - nested_junctions > 1|0, for volumes
                allow_extra => 1,
            );

    # get inputs
    # pass the 'vserver' arg as-is to NACL, if defined; don't copy it here
    my @nodes = @{$args{nodes}};
    delete $args{nodes};
    my $resource_type       = delete $args{resource_type};
    my $count               = delete $args{count};
    my $spread_evenly       = delete $args{spread_evenly};
    my $nested_junctions;
    my $stask_method_name;
    my @aggregates;
    my @resources_created;

    # defaults based on the resource type
    if ($resource_type =~ /volume/) {
        $stask_method_name = 'NACL::STask::Volume';
        # volume specific options
        $nested_junctions  = delete $args{nested_junctions};
    } else {
        $stask_method_name = 'NACL::STask::Vserver';
    }

    # find all usable aggregates (if no aggregate was provided)
    # check for the 'aggregate' parameter that will be passed
    # down to NACL::STask::Volume->create(), skip aggregate
    # lookup if this is already provided.

    if (defined $args{aggregate}) {
        $Log->comment("Using user provided aggregate: $args{aggregate}\n"
                    . "Will not look for other aggregates.");
        push (@aggregates, delete $args{aggregate});

    } else {
        @aggregates = $self->find_aggregates_for_volumes(
            nodes         => \@nodes,
            spread_evenly => $spread_evenly,
        );
        $Log->comment("\nThe following aggregates will be used:\n" .
                        join("\n", @aggregates));
    }

    # resource name
    # a suffix like 1, 2, 3, etc for the resource count will be appended
    # in the foreach loop below
    my $name_prefix;
    if (defined $args{name_prefix}) {
        # use the user specified name prefix
        $name_prefix = delete $args{name_prefix};

    } else {
        my $prefix_key = $resource_type . "_name_prefix";
        if (defined $self->{$prefix_key}) {
            # use the already created name prefix
            $name_prefix = $self->{$prefix_key};
        } else {
            # generate a meaningful random name
            # the name will be of the format:
            #       kernel_test_volume_<random-word>_timestamp_
            #       OR
            #       kernel_test_vserver_<random-word>_timestamp_
            #
            $name_prefix = NACL::GeneralUtils::random_name_generator(
                            prefix => "kernel_test_$resource_type",
                            size   => 1,    # number of random words to use
                            suffix => "_",
                        );
            # save it for using it the next time
            $self->{$prefix_key} = $name_prefix;
        }
    }

    #----------------------------------------------------------------------
    # VOLUME/VSERVER CREATION

    # For VOLUMES only:
    # this junction-path will either be reset to "" if nested paths are not
    # requested, or will be appended to, if nested paths are requested
    my $junction_path  = "";
    my $aggr_index     = 0;

    # start creating the resources
    foreach my $i (1..$count) {

        #----------------------------------------------------------------------
        # volume/vserver name & junction-path
        my $resource_name = $name_prefix;

        if ($resource_type =~ /volume/) {
            # append 
            $resource_name .= scalar @{$self->volumes()};

            # args for NACL::STask::Volume
            $args{volume} = $resource_name;

            # volume junction-path
            if ($nested_junctions == 1) {
                # append to the previous junction-path
                $args{'junction-path'} .= "/" . $resource_name;
            } else {
                # reset the junction-path to mount the volume at
                # the vserver's root
                $args{'junction-path'} = "/" . $resource_name;
            }
        } else {
            # appending this will help keep the name's tail unique in case
            # multiple calls are made to create volumes/vservers.
            $resource_name .= scalar @{$self->vservers()};

            # args for NACL::STask::Vserver
            $args{vserver} = $resource_name;
        }

        #----------------------------------------------------------------------
        # call NACL::STask::Volume/Vserver->create(...);
        # MAX number of retries = the number of aggregates available
        my $num_of_retries = scalar @aggregates;
        my $resource;

        # loop while retry > 0, the terminating condition for this loop
        while ($num_of_retries > 0) {

            try {
                $Log->comment("\nCreating $resource_type $resource_name using " .
                              "aggregate $aggregates[$aggr_index]");
                $resource = $stask_method_name->create(
                    command_interface   => $nodes[0],
                    aggregate           => $aggregates[$aggr_index],
                    nacltask_to_cleanup => 1,
                    %args,
                );

                # looks like we succeeded
                # save and return the list of vserver objects and
                # jump out of the loop
                push (@{$self->{$resource_type . "s"}}, $resource);
                push (@resources_created, $resource);
                $Log->comment("\nSuccessfully created $resource_type $resource_name");
                last;

            } catch NACL::APISet::Exceptions::ResponseException with {
                my $exception = shift;

                if ($exception->text() =~
                    /No space left on device/i) {

                    # the aggregate does not have enough space
                    # delete this aggregate from the list
                    $Log->comment("\nThe aggregate does not have enough space. " .
                                  "\nDeleting this aggregate from the usable aggregates' list.");
                    splice(@aggregates, $aggr_index, 1);

                    # try the next aggregate until we've exhausted all of them
                    # the while loop will take care of the retry
                    if (scalar @aggregates > 0) {
                        $Log->comment("\nRetrying with the next aggregate if available...");
                    } else {
                        $Log->comment("\nNo more usable aggregates left!\n");
                    }

                } else {
                    # looks like some other error that needs to be looked at
                    # rethrow the exception and bail
                    $exception->throw();
                }
            };

        } # end of 'while' loop for retried

        # check if the resource was created
        if (!defined $resource) {
            throw NATE::BaseException (
                "Failed to create $resource_type!"
            );
        }
        
        #----------------------------------------------------------------------
        # POST RESOURCE CREATION PROCESSING
        if ($spread_evenly == 1) {
            # continue to the next aggregate
            $Log->debug("\nProceeding to the next aggregate...\n");
            $aggr_index++;

            # increment/adjust the aggregate index if the next aggr_index
            # will cross the array boundary
            if ($aggr_index >= scalar @aggregates) {
                # cycle back to the first aggregate
                $Log->debug("\nCycling back to the first aggregate...\n");
                $aggr_index = 0;
            }

        } else {
            # if we don't plan to spread the volumes/vservers evenly, don't
            # increment the aggr_index; reuse the existing aggregate until
            # we run out of space on it, thus forcing us to go the next
            # available aggregate
        }


    } # end of resource creation


    return @resources_created;
}

1;
